// Copyright (c) 2019-2023 Intel Corporation
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include <set>

#include "av1_scd.h"

#if defined(MFX_ENABLE_ADAPTIVE_ENCODE)

namespace aenc {
mfxU16 AGOPSelect1(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1311) {
if (Rs < 2424) {
if (Cs < 533) {
if (negBalance < 14) {
if (mu_mv_mag_sq < 115) {
if (Cs < 179) {
if (mcTcor < 99) {
if (negBalance < 6) {
if (ssDCval < 63) {
if (refDCval < 27) {
if (SC < 47) {
if (mu_mv_mag_sq < 46) {
if (tcor < 67) {
if (diffMVdiffVal < 25) {
return 8;
}
else if (diffMVdiffVal >= 25) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (tcor >= 67) {
return 2;
}
}
else if (mu_mv_mag_sq >= 46) {
if (ssDCval < 17) {
return 4;
}
else if (ssDCval >= 17) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (diffMVdiffVal < 12) {
return 8;
}
else if (diffMVdiffVal >= 12) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
}
}
}
}
else if (SC >= 47) {
if (tcor < 76) {
if (negBalance < 2) {
if (Rs < 192) {
return 4;
}
else if (Rs >= 192) {
return 1;
}
}
else if (negBalance >= 2) {
return 2;
}
}
else if (tcor >= 76) {
return 4;
}
}
}
else if (refDCval >= 27) {
if (Rs < 10) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
else if (Rs >= 10) {
if (refDCval < 59) {
if (AFD < 16) {
if (Rs < 218) {
return 8;
}
else if (Rs >= 218) {
if (ssDCval < 47) {
if (Rs < 285) {
return 4;
}
else if (Rs >= 285) {
return 8;
}
}
else if (ssDCval >= 47) {
return 8;
}
}
}
else if (AFD >= 16) {
return 1;
}
}
else if (refDCval >= 59) {
return 2;
}
}
}
}
else if (ssDCval >= 63) {
if (AFD < 5) {
if (MVDiff < 34) {
if (Rs < 32) {
return 2;
}
else if (Rs >= 32) {
return 1;
}
}
else if (MVDiff >= 34) {
return 4;
}
}
else if (AFD >= 5) {
if (Rs < 114) {
return 1;
}
else if (Rs >= 114) {
return 8;
}
}
}
}
else if (negBalance >= 6) {
if (ssDCval < 83) {
if (refDCval < 27) {
if (mcTcor < 91) {
if (refDCval < 25) {
if (TSC < 2) {
if (posBalance < 10) {
return 4;
}
else if (posBalance >= 10) {
return 8;
}
}
else if (TSC >= 2) {
if (Rs < 152) {
return 2;
}
else if (Rs >= 152) {
return 1;
}
}
}
else if (refDCval >= 25) {
if (MVDiff < 5) {
if (Rs < 53) {
return 4;
}
else if (Rs >= 53) {
return 8;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
}
else if (mcTcor >= 91) {
if (tcor < 96) {
if (Rs < 2) {
return 2;
}
else if (Rs >= 2) {
if (posBalance < 15) {
return 4;
}
else if (posBalance >= 15) {
return 1;
}
}
}
else if (tcor >= 96) {
if (refDCval < 25) {
return 1;
}
else if (refDCval >= 25) {
if (negBalance < 12) {
return 4;
}
else if (negBalance >= 12) {
return 1;
}
}
}
}
}
else if (refDCval >= 27) {
if (SC < 180) {
if (refDCval < 32) {
if (MVDiff < 9) {
if (negBalance < 11) {
if (ssDCval < 28) {
return 1;
}
else if (ssDCval >= 28) {
return 4;
}
}
else if (negBalance >= 11) {
return 4;
}
}
else if (MVDiff >= 9) {
return 1;
}
}
else if (refDCval >= 32) {
if (MVDiff < 33) {
return 8;
}
else if (MVDiff >= 33) {
return 4;
}
}
}
else if (SC >= 180) {
if (negBalance < 12) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (AFD < 4) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (Rs < 114) {
return 8;
}
else if (Rs >= 114) {
return 1;
}
}
}
else if (AFD >= 4) {
return 2;
}
}
}
else if (negBalance >= 12) {
if (mu_mv_mag_sq < 74) {
if (Rs < 1146) {
return 8;
}
else if (Rs >= 1146) {
return 2;
}
}
else if (mu_mv_mag_sq >= 74) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (tcor < 95) {
if (SC < 217) {
return 4;
}
else if (SC >= 217) {
return 1;
}
}
else if (tcor >= 95) {
if (Cs < 75) {
if (Rs < 471) {
return 1;
}
else if (Rs >= 471) {
return 4;
}
}
else if (Cs >= 75) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 83) {
if (refDCval < 88) {
if (Cs < 49) {
if (Rs < 585) {
return 1;
}
else if (Rs >= 585) {
return 8;
}
}
else if (Cs >= 49) {
return 1;
}
}
else if (refDCval >= 88) {
if (ssDCval < 101) {
if (Rs < 364) {
return 4;
}
else if (Rs >= 364) {
return 2;
}
}
else if (ssDCval >= 101) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 27) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
if (posBalance < 45) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (negBalance < 11) {
return 4;
}
else if (negBalance >= 11) {
if (Rs < 32) {
return 4;
}
else if (Rs >= 32) {
if (Rs < 50) {
return 1;
}
else if (Rs >= 50) {
return 4;
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 17) {
return 8;
}
else if (Rs >= 17) {
return 1;
}
}
}
}
else if (ssDCval >= 27) {
if (MVDiff < 8) {
if (Rs < 186) {
if (posBalance < 22) {
if (mcTcor < 102) {
return 8;
}
else if (mcTcor >= 102) {
return 4;
}
}
else if (posBalance >= 22) {
if (ssDCval < 47) {
if (Rs < 110) {
if (posBalance < 23) {
return 4;
}
else if (posBalance >= 23) {
return 8;
}
}
else if (Rs >= 110) {
return 1;
}
}
else if (ssDCval >= 47) {
if (tcor < 100) {
if (posBalance < 29) {
return 1;
}
else if (posBalance >= 29) {
if (Rs < 152) {
return 4;
}
else if (Rs >= 152) {
return 2;
}
}
}
else if (tcor >= 100) {
return 2;
}
}
}
}
else if (Rs >= 186) {
if (mu_mv_mag_sq < 18) {
if (Rs < 645) {
return 4;
}
else if (Rs >= 645) {
return 8;
}
}
else if (mu_mv_mag_sq >= 18) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (Rs < 1301) {
return 8;
}
else if (Rs >= 1301) {
return 1;
}
}
}
}
}
else if (MVDiff >= 8) {
if (posBalance < 59) {
if (posBalance < 22) {
if (Rs < 60) {
return 2;
}
else if (Rs >= 60) {
return 8;
}
}
else if (posBalance >= 22) {
if (Cs < 174) {
return 4;
}
else if (Cs >= 174) {
if (Rs < 418) {
return 8;
}
else if (Rs >= 418) {
return 4;
}
}
}
}
else if (posBalance >= 59) {
if (AFD < 10) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (ssDCval < 133) {
if (mcTcor < 104) {
return 8;
}
else if (mcTcor >= 104) {
return 4;
}
}
else if (ssDCval >= 133) {
return 4;
}
}
}
}
else if (AFD >= 10) {
return 1;
}
}
}
}
}
}
else if (Cs >= 179) {
if (SC < 869) {
if (tcor < 94) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
if (refDCval < 48) {
if (mu_mv_mag_sq < 36) {
if (posBalance < 95) {
return 2;
}
else if (posBalance >= 95) {
if (Rs < 437) {
return 4;
}
else if (Rs >= 437) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 36) {
if (posBalance < 30) {
if (negBalance < 11) {
return 4;
}
else if (negBalance >= 11) {
if (Cs < 242) {
return 8;
}
else if (Cs >= 242) {
if (MVDiff < 1) {
if (Rs < 322) {
return 4;
}
else if (Rs >= 322) {
return 1;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
else if (posBalance >= 30) {
if (SC < 459) {
if (diffRsCsdiff < -1) {
if (Rs < 219) {
return 1;
}
else if (Rs >= 219) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (SC >= 459) {
if (mcTcor < 95) {
if (Rs < 567) {
return 1;
}
else if (Rs >= 567) {
return 4;
}
}
else if (mcTcor >= 95) {
if (Rs < 378) {
return 4;
}
else if (Rs >= 378) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 48) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 45) {
return 2;
}
else if (mu_mv_mag_sq >= 45) {
return 1;
}
}
}
}
}
else if (tcor >= 94) {
if (ssDCval < 36) {
if (mu_mv_mag_sq < 103) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 532) {
if (Rs < 275) {
return 8;
}
else if (Rs >= 275) {
return 1;
}
}
else if (Rs >= 532) {
return 4;
}
}
}
else if (ssDCval >= 36) {
if (posBalance < 33) {
if (SC < 561) {
return 1;
}
else if (SC >= 561) {
if (posBalance < 23) {
if (Rs < 503) {
return 1;
}
else if (Rs >= 503) {
return 8;
}
}
else if (posBalance >= 23) {
if (SC < 841) {
if (diffMVdiffVal < -26) {
return 8;
}
else if (diffMVdiffVal >= -26) {
if (Rs < 502) {
if (Cs < 459) {
return 4;
}
else if (Cs >= 459) {
return 1;
}
}
else if (Rs >= 502) {
return 8;
}
}
}
else if (SC >= 841) {
return 1;
}
}
}
}
else if (posBalance >= 33) {
if (Rs < 189) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (ssDCval < 79) {
if (Rs < 145) {
return 2;
}
else if (Rs >= 145) {
return 4;
}
}
else if (ssDCval >= 79) {
return 4;
}
}
}
else if (Rs >= 189) {
if (negBalance < 7) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 1;
}
}
else if (negBalance >= 7) {
if (TSC < 2) {
if (mu_mv_mag_sq < 102) {
if (Rs < 289) {
return 1;
}
else if (Rs >= 289) {
return 4;
}
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 276) {
if (Rs < 219) {
return 8;
}
else if (Rs >= 219) {
return 4;
}
}
else if (Rs >= 276) {
if (Rs < 350) {
return 1;
}
else if (Rs >= 350) {
return 8;
}
}
}
}
else if (TSC >= 2) {
if (Cs < 297) {
return 8;
}
else if (Cs >= 297) {
return 1;
}
}
}
}
}
}
}
}
else if (SC >= 869) {
if (RsCsDiff < 1) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
if (Cs < 396) {
if (ssDCval < 152) {
if (mu_mv_mag_sq < 89) {
return 1;
}
else if (mu_mv_mag_sq >= 89) {
return 8;
}
}
else if (ssDCval >= 152) {
return 1;
}
}
else if (Cs >= 396) {
if (negBalance < 13) {
return 8;
}
else if (negBalance >= 13) {
if (Rs < 988) {
return 8;
}
else if (Rs >= 988) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (Rs < 2071) {
if (mcTcor < 97) {
if (negBalance < 8) {
if (mu_mv_mag_sq < 76) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (mu_mv_mag_sq >= 76) {
if (Cs < 473) {
return 4;
}
else if (Cs >= 473) {
return 8;
}
}
}
else if (negBalance >= 8) {
if (SC < 1253) {
if (Rs < 468) {
return 4;
}
else if (Rs >= 468) {
return 1;
}
}
else if (SC >= 1253) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 495) {
if (AFD < 2) {
if (Rs < 585) {
return 8;
}
else if (Rs >= 585) {
return 4;
}
}
else if (AFD >= 2) {
return 1;
}
}
else if (Cs >= 495) {
return 8;
}
}
}
else if (Rs >= 2071) {
if (Cs < 374) {
return 1;
}
else if (Cs >= 374) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (SC < 699) {
if (refDCval < 47) {
if (Cs < 98) {
if (MVDiff < 1076) {
if (ssDCval < 39) {
if (ssDCval < 22) {
if (ssDCval < 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (ssDCval >= 22) {
if (Cs < 59) {
if (posBalance < 61) {
if (Rs < 13) {
return 4;
}
else if (Rs >= 13) {
if (posBalance < 14) {
return 8;
}
else if (posBalance >= 14) {
return 4;
}
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (Cs >= 59) {
return 8;
}
}
}
else if (ssDCval >= 39) {
return 8;
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (Cs >= 98) {
if (Rs < 203) {
if (negBalance < 3) {
return 8;
}
else if (negBalance >= 3) {
if (SC < 422) {
if (Cs < 180) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 169) {
return 1;
}
else if (Rs >= 169) {
return 8;
}
}
}
else if (Cs >= 180) {
return 4;
}
}
else if (SC >= 422) {
return 1;
}
}
}
else if (Rs >= 203) {
if (refDCval < 39) {
if (Rs < 306) {
return 8;
}
else if (Rs >= 306) {
if (mu_mv_mag_sq < 128) {
if (SC < 541) {
return 8;
}
else if (SC >= 541) {
return 4;
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 280) {
return 8;
}
else if (Cs >= 280) {
return 4;
}
}
}
}
else if (refDCval >= 39) {
if (Rs < 285) {
return 4;
}
else if (Rs >= 285) {
if (Rs < 301) {
return 1;
}
else if (Rs >= 301) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 47) {
if (ssDCval < 82) {
if (Cs < 245) {
if (ssDCval < 62) {
if (Rs < 16) {
return 2;
}
else if (Rs >= 16) {
if (MVDiff < 1) {
if (negBalance < 6) {
if (Rs < 183) {
return 8;
}
else if (Rs >= 183) {
return 4;
}
}
else if (negBalance >= 6) {
return 8;
}
}
else if (MVDiff >= 1) {
if (Rs < 220) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 4;
}
}
}
else if (Rs >= 220) {
return 4;
}
}
}
}
else if (ssDCval >= 62) {
if (Cs < 130) {
if (Rs < 172) {
if (Rs < 42) {
return 2;
}
else if (Rs >= 42) {
if (Rs < 103) {
return 4;
}
else if (Rs >= 103) {
return 2;
}
}
}
else if (Rs >= 172) {
return 8;
}
}
else if (Cs >= 130) {
return 2;
}
}
}
else if (Cs >= 245) {
if (Rs < 251) {
return 4;
}
else if (Rs >= 251) {
if (SC < 646) {
if (Rs < 325) {
return 1;
}
else if (Rs >= 325) {
return 4;
}
}
else if (SC >= 646) {
return 8;
}
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (Cs < 328) {
if (negBalance < 11) {
if (Cs < 301) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (posBalance < 13) {
return 8;
}
else if (posBalance >= 13) {
return 4;
}
}
}
else if (Cs >= 301) {
return 2;
}
}
else if (negBalance >= 11) {
return 2;
}
}
else if (Cs >= 328) {
return 4;
}
}
}
}
}
else if (SC >= 699) {
if (diffTSC < -3) {
if (posBalance < 5) {
if (SC < 744) {
return 8;
}
else if (SC >= 744) {
return 4;
}
}
else if (posBalance >= 5) {
if (ssDCval < 146) {
if (diffAFD < -7) {
if (posBalance < 21) {
return 2;
}
else if (posBalance >= 21) {
return 1;
}
}
else if (diffAFD >= -7) {
if (Rs < 801) {
if (ssDCval < 127) {
if (Rs < 569) {
return 4;
}
else if (Rs >= 569) {
return 8;
}
}
else if (ssDCval >= 127) {
return 2;
}
}
else if (Rs >= 801) {
return 1;
}
}
}
else if (ssDCval >= 146) {
return 8;
}
}
}
else if (diffTSC >= -3) {
if (mu_mv_mag_sq < 128) {
if (Rs < 711) {
if (ssDCval < 141) {
if (Rs < 348) {
return 1;
}
else if (Rs >= 348) {
if (ssDCval < 42) {
if (negBalance < 5) {
return 8;
}
else if (negBalance >= 5) {
if (Cs < 409) {
return 4;
}
else if (Cs >= 409) {
return 1;
}
}
}
else if (ssDCval >= 42) {
if (SC < 938) {
if (posBalance < 24) {
return 1;
}
else if (posBalance >= 24) {
return 8;
}
}
else if (SC >= 938) {
if (posBalance < 35) {
if (gchDC < 1) {
if (SC < 944) {
if (SC < 941) {
return 8;
}
else if (SC >= 941) {
return 1;
}
}
else if (SC >= 944) {
return 8;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (posBalance >= 35) {
if (Rs < 598) {
if (Rs < 515) {
return 8;
}
else if (Rs >= 515) {
return 1;
}
}
else if (Rs >= 598) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 141) {
if (Rs < 496) {
return 1;
}
else if (Rs >= 496) {
return 8;
}
}
}
else if (Rs >= 711) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Cs < 480) {
if (Rs < 714) {
return 2;
}
else if (Rs >= 714) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (posBalance < 41) {
if (mu_mv_mag_sq < 125) {
return 1;
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
else if (posBalance >= 41) {
return 8;
}
}
}
}
else if (Cs >= 480) {
return 1;
}
}
else if (ssDCval >= 60) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
if (Rs < 1471) {
return 2;
}
else if (Rs >= 1471) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (Cs < 250) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
return 2;
}
}
else if (Cs >= 250) {
if (Cs < 315) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Rs < 1026) {
return 1;
}
else if (Rs >= 1026) {
return 8;
}
}
}
else if (Cs >= 315) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (posBalance < 39) {
if (ssDCval < 66) {
if (Rs < 1093) {
return 8;
}
else if (Rs >= 1093) {
return 2;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (posBalance >= 39) {
if (Rs < 560) {
return 8;
}
else if (Rs >= 560) {
if (Rs < 795) {
return 4;
}
else if (Rs >= 795) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 35) {
if (tcor < 64) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (Cs < 387) {
if (negBalance < 46) {
if (Cs < 98) {
if (negBalance < 19) {
if (Cs < 72) {
if (Rs < 53) {
return 8;
}
else if (Rs >= 53) {
return 1;
}
}
else if (Cs >= 72) {
return 2;
}
}
else if (negBalance >= 19) {
if (diffTSC < -1) {
if (diffMVdiffVal < 6) {
return 1;
}
else if (diffMVdiffVal >= 6) {
return 4;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
}
else if (Cs >= 98) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (Rs < 139) {
if (Cs < 166) {
return 8;
}
else if (Cs >= 166) {
return 4;
}
}
else if (Rs >= 139) {
if (refDCval < 26) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
return 8;
}
}
else if (refDCval >= 26) {
if (AFD < 5) {
if (Rs < 215) {
return 2;
}
else if (Rs >= 215) {
return 1;
}
}
else if (AFD >= 5) {
if (AFD < 22) {
if (posBalance < 76) {
return 1;
}
else if (posBalance >= 76) {
if (Rs < 265) {
return 8;
}
else if (Rs >= 265) {
return 1;
}
}
}
else if (AFD >= 22) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 46) {
if (Rs < 236) {
if (MVDiff < 276) {
return 8;
}
else if (MVDiff >= 276) {
return 4;
}
}
else if (Rs >= 236) {
if (Cs < 31) {
if (Rs < 377) {
return 1;
}
else if (Rs >= 377) {
return 8;
}
}
else if (Cs >= 31) {
return 1;
}
}
}
}
else if (Cs >= 387) {
if (RsDiff < 4) {
if (negBalance < 26) {
if (TSC < 3) {
if (Cs < 491) {
return 8;
}
else if (Cs >= 491) {
return 4;
}
}
else if (TSC >= 3) {
return 1;
}
}
else if (negBalance >= 26) {
return 2;
}
}
else if (RsDiff >= 4) {
return 8;
}
}
}
}
else if (tcor >= 64) {
if (posBalance < 27) {
if (AFD < 2) {
if (mu_mv_mag_sq < 115) {
if (MVDiff < 3) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (MVDiff >= 3) {
if (posBalance < 21) {
if (ssDCval < 22) {
if (MVDiff < 245) {
return 8;
}
else if (MVDiff >= 245) {
return 4;
}
}
else if (ssDCval >= 22) {
if (ssDCval < 31) {
if (refDCval < 26) {
if (Rs < 8) {
return 4;
}
else if (Rs >= 8) {
return 1;
}
}
else if (refDCval >= 26) {
return 4;
}
}
else if (ssDCval >= 31) {
return 8;
}
}
}
else if (posBalance >= 21) {
if (ssDCval < 30) {
return 1;
}
else if (ssDCval >= 30) {
if (Rs < 239) {
return 8;
}
else if (Rs >= 239) {
if (Rs < 467) {
return 1;
}
else if (Rs >= 467) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
else if (AFD >= 2) {
if (negBalance < 58) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 86) {
if (ssDCval < 32) {
return 1;
}
else if (ssDCval >= 32) {
if (Rs < 119) {
return 1;
}
else if (Rs >= 119) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 86) {
return 4;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (negBalance >= 58) {
if (mu_mv_mag_sq < 4) {
if (Rs < 84) {
return 8;
}
else if (Rs >= 84) {
return 1;
}
}
else if (mu_mv_mag_sq >= 4) {
if (refDCval < 20) {
return 4;
}
else if (refDCval >= 20) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
if (Cs < 1) {
if (mu_mv_mag_sq < 36) {
return 8;
}
else if (mu_mv_mag_sq >= 36) {
return 4;
}
}
else if (Cs >= 1) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 27) {
if (ssDCval < 30) {
if (Cs < 351) {
if (MVDiff < 63) {
if (mu_mv_mag_sq < 50) {
if (posBalance < 71) {
if (Cs < 249) {
if (MVDiff < 11) {
if (Cs < 25) {
if (Rs < 71) {
return 1;
}
else if (Rs >= 71) {
return 4;
}
}
else if (Cs >= 25) {
if (diffRsCsdiff < -3) {
if (Rs < 761) {
return 1;
}
else if (Rs >= 761) {
return 4;
}
}
else if (diffRsCsdiff >= -3) {
return 1;
}
}
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (Cs >= 249) {
if (MVDiff < 34) {
return 1;
}
else if (MVDiff >= 34) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
}
}
else if (posBalance >= 71) {
return 2;
}
}
else if (mu_mv_mag_sq >= 50) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (SC < 300) {
if (SC < 255) {
return 1;
}
else if (SC >= 255) {
if (MVDiff < 25) {
return 8;
}
else if (MVDiff >= 25) {
return 1;
}
}
}
else if (SC >= 300) {
return 1;
}
}
}
}
else if (MVDiff >= 63) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (Rs < 49) {
return 1;
}
else if (Rs >= 49) {
return 8;
}
}
}
}
else if (Cs >= 351) {
if (Rs < 680) {
if (posBalance < 31) {
return 1;
}
else if (posBalance >= 31) {
return 4;
}
}
else if (Rs >= 680) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
else if (ssDCval >= 30) {
if (posBalance < 41) {
if (Cs < 260) {
if (SC < 493) {
if (mcTcor < 97) {
if (AFD < 4) {
if (posBalance < 28) {
if (Cs < 67) {
return 1;
}
else if (Cs >= 67) {
return 4;
}
}
else if (posBalance >= 28) {
if (diffMVdiffVal < -14) {
return 1;
}
else if (diffMVdiffVal >= -14) {
if (Rs < 145) {
return 8;
}
else if (Rs >= 145) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (mcTcor < 96) {
if (Cs < 119) {
return 4;
}
else if (Cs >= 119) {
return 8;
}
}
else if (mcTcor >= 96) {
return 4;
}
}
}
}
}
}
else if (AFD >= 4) {
if (posBalance < 28) {
return 2;
}
else if (posBalance >= 28) {
if (mu_mv_mag_sq < 61) {
if (negBalance < 27) {
return 1;
}
else if (negBalance >= 27) {
if (negBalance < 31) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (negBalance >= 31) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (Rs < 175) {
return 8;
}
else if (Rs >= 175) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 22) {
if (Rs < 105) {
return 8;
}
else if (Rs >= 105) {
return 4;
}
}
else if (mu_mv_mag_sq >= 22) {
if (negBalance < 15) {
return 8;
}
else if (negBalance >= 15) {
return 1;
}
}
}
}
else if (SC >= 493) {
if (mcTcor < 92) {
if (AFD < 4) {
if (Rs < 496) {
return 8;
}
else if (Rs >= 496) {
return 1;
}
}
else if (AFD >= 4) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (mcTcor >= 92) {
return 4;
}
}
}
else if (Cs >= 260) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (posBalance >= 41) {
if (RsDiff < 2) {
if (mcTcor < 89) {
if (RsDiff < 1) {
if (SC < 46) {
return 4;
}
else if (SC >= 46) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (RsDiff >= 1) {
if (Cs < 298) {
if (mu_mv_mag_sq < 3) {
if (Cs < 213) {
return 2;
}
else if (Cs >= 213) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (Cs >= 298) {
if (Rs < 284) {
return 1;
}
else if (Rs >= 284) {
return 2;
}
}
}
}
else if (mcTcor >= 89) {
if (posBalance < 64) {
if (gchDC < 3) {
if (negBalance < 43) {
if (ssDCval < 35) {
return 1;
}
else if (ssDCval >= 35) {
if (Rs < 95) {
return 2;
}
else if (Rs >= 95) {
if (Rs < 533) {
return 1;
}
else if (Rs >= 533) {
return 4;
}
}
}
}
else if (negBalance >= 43) {
return 2;
}
}
else if (gchDC >= 3) {
if (Cs < 98) {
return 1;
}
else if (Cs >= 98) {
return 4;
}
}
}
else if (posBalance >= 64) {
if (Rs < 60) {
return 2;
}
else if (Rs >= 60) {
if (Rs < 102) {
return 1;
}
else if (Rs >= 102) {
return 8;
}
}
}
}
}
else if (RsDiff >= 2) {
if (Cs < 501) {
return 1;
}
else if (Cs >= 501) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 35) {
if (Rs < 300) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 52) {
if (mcTcor < 90) {
if (Cs < 66) {
if (RsDiff < 1) {
if (TSC < 3) {
if (Rs < 123) {
return 4;
}
else if (Rs >= 123) {
return 2;
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (RsDiff >= 1) {
if (diffAFD < 3) {
if (mu_mv_mag_sq < 44) {
return 1;
}
else if (mu_mv_mag_sq >= 44) {
if (Rs < 91) {
return 8;
}
else if (Rs >= 91) {
return 1;
}
}
}
else if (diffAFD >= 3) {
if (Cs < 36) {
return 2;
}
else if (Cs >= 36) {
return 4;
}
}
}
}
else if (Cs >= 66) {
if (posBalance < 54) {
if (ssDCval < 33) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 1;
}
}
else if (ssDCval >= 33) {
if (MVDiff < 28) {
if (Cs < 281) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 162) {
return 2;
}
else if (Rs >= 162) {
return 8;
}
}
}
else if (Cs >= 281) {
return 2;
}
}
else if (MVDiff >= 28) {
if (diffTSC < -1) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 124) {
return 4;
}
else if (Rs >= 124) {
return 2;
}
}
}
else if (diffTSC >= -1) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
}
}
else if (posBalance >= 54) {
if (negBalance < 34) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 2;
}
}
else if (negBalance >= 34) {
if (tcor < 67) {
if (mu_mv_mag_sq < 14) {
if (Rs < 234) {
if (negBalance < 45) {
return 4;
}
else if (negBalance >= 45) {
if (mu_mv_mag_sq < 3) {
if (Cs < 294) {
return 4;
}
else if (Cs >= 294) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (Rs >= 234) {
return 4;
}
}
else if (mu_mv_mag_sq >= 14) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Rs < 51) {
return 2;
}
else if (Rs >= 51) {
return 8;
}
}
}
}
else if (tcor >= 67) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (CsDiff < 2) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
else if (CsDiff >= 2) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (negBalance < 78) {
if (tcor < 93) {
if (posBalance < 68) {
if (ssDCval < 38) {
if (posBalance < 46) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (posBalance >= 46) {
if (Cs < 215) {
if (AFD < 5) {
if (Rs < 83) {
if (Rs < 72) {
return 1;
}
else if (Rs >= 72) {
return 2;
}
}
else if (Rs >= 83) {
if (Rs < 127) {
return 4;
}
else if (Rs >= 127) {
return 8;
}
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 215) {
if (Cs < 303) {
return 8;
}
else if (Cs >= 303) {
return 1;
}
}
}
}
else if (ssDCval >= 38) {
if (ssDCval < 42) {
if (Rs < 137) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (Cs < 23) {
return 2;
}
else if (Cs >= 23) {
if (diffRsCsdiff < 1) {
return 4;
}
else if (diffRsCsdiff >= 1) {
if (Cs < 142) {
return 4;
}
else if (Cs >= 142) {
return 1;
}
}
}
}
}
else if (Rs >= 137) {
if (CsDiff < 2) {
if (AFD < 9) {
if (mcTcor < 97) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (Rs < 193) {
return 2;
}
else if (Rs >= 193) {
return 1;
}
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
}
else if (ssDCval >= 42) {
if (SC < 457) {
if (MVDiff < 33) {
if (AFD < 12) {
if (SC < 441) {
return 1;
}
else if (SC >= 441) {
return 2;
}
}
else if (AFD >= 12) {
return 2;
}
}
else if (MVDiff >= 33) {
if (ssDCval < 43) {
if (diffAFD < -1) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
return 2;
}
}
else if (diffAFD >= -1) {
if (Rs < 108) {
return 2;
}
else if (Rs >= 108) {
return 1;
}
}
}
else if (ssDCval >= 43) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (Rs < 179) {
return 2;
}
else if (Rs >= 179) {
return 4;
}
}
}
}
}
}
else if (SC >= 457) {
if (Cs < 343) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 4;
}
}
else if (Cs >= 343) {
if (Rs < 261) {
return 1;
}
else if (Rs >= 261) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 68) {
return 2;
}
}
else if (tcor >= 93) {
if (diffRsCsdiff < -1) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 104) {
if (Rs < 49) {
return 4;
}
else if (Rs >= 49) {
return 1;
}
}
else if (Rs >= 104) {
if (Cs < 380) {
return 2;
}
else if (Cs >= 380) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -1) {
if (Rs < 272) {
return 1;
}
else if (Rs >= 272) {
if (Rs < 291) {
return 4;
}
else if (Rs >= 291) {
return 1;
}
}
}
}
}
else if (negBalance >= 78) {
if (Rs < 230) {
return 8;
}
else if (Rs >= 230) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (SC < 530) {
if (negBalance < 27) {
if (mu_mv_mag_sq < 108) {
if (refDCval < 47) {
if (ssDCval < 39) {
if (SC < 388) {
if (mu_mv_mag_sq < 98) {
if (SC < 373) {
return 1;
}
else if (SC >= 373) {
if (Rs < 144) {
return 1;
}
else if (Rs >= 144) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 98) {
return 4;
}
}
else if (SC >= 388) {
if (tcor < 91) {
if (posBalance < 38) {
if (Rs < 193) {
return 8;
}
else if (Rs >= 193) {
return 4;
}
}
else if (posBalance >= 38) {
return 1;
}
}
else if (tcor >= 91) {
return 8;
}
}
}
else if (ssDCval >= 39) {
if (Cs < 187) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (negBalance < 23) {
return 1;
}
else if (negBalance >= 23) {
if (Rs < 58) {
return 1;
}
else if (Rs >= 58) {
return 8;
}
}
}
}
else if (Cs >= 187) {
return 1;
}
}
}
else if (refDCval >= 47) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (Rs < 129) {
return 2;
}
else if (Rs >= 129) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (tcor < 92) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
else if (tcor >= 92) {
if (posBalance < 37) {
if (ssDCval < 43) {
if (Cs < 75) {
return 4;
}
else if (Cs >= 75) {
if (Cs < 206) {
return 1;
}
else if (Cs >= 206) {
return 8;
}
}
}
else if (ssDCval >= 43) {
if (Rs < 227) {
return 8;
}
else if (Rs >= 227) {
return 4;
}
}
}
else if (posBalance >= 37) {
return 8;
}
}
}
}
else if (negBalance >= 27) {
if (gchDC < 1) {
if (Cs < 51) {
if (Rs < 68) {
return 8;
}
else if (Rs >= 68) {
if (Rs < 111) {
return 4;
}
else if (Rs >= 111) {
if (Rs < 146) {
return 2;
}
else if (Rs >= 146) {
return 1;
}
}
}
}
else if (Cs >= 51) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (Rs < 165) {
return 1;
}
else if (Rs >= 165) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
if (posBalance < 23) {
if (mcTcor < 105) {
if (Rs < 224) {
if (AFD < 30) {
return 8;
}
else if (AFD >= 30) {
return 1;
}
}
else if (Rs >= 224) {
return 4;
}
}
else if (mcTcor >= 105) {
if (Rs < 161) {
return 4;
}
else if (Rs >= 161) {
return 8;
}
}
}
else if (posBalance >= 23) {
if (negBalance < 44) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (refDCval < 40) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (Cs < 35) {
return 2;
}
else if (Cs >= 35) {
return 4;
}
}
}
else if (refDCval >= 40) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (Rs < 190) {
return 4;
}
else if (Rs >= 190) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 44) {
if (RsDiff < 3) {
if (mu_mv_mag_sq < 73) {
if (diffMVdiffVal < 0) {
if (Rs < 27) {
return 8;
}
else if (Rs >= 27) {
return 2;
}
}
else if (diffMVdiffVal >= 0) {
return 8;
}
}
else if (mu_mv_mag_sq >= 73) {
if (Rs < 41) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 2;
}
}
}
else if (Rs >= 41) {
return 4;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 286) {
return 1;
}
else if (Rs >= 286) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 530) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
if (mu_mv_mag_sq < 57) {
if (Rs < 274) {
return 2;
}
else if (Rs >= 274) {
return 1;
}
}
else if (mu_mv_mag_sq >= 57) {
if (SC < 537) {
if (Rs < 261) {
return 8;
}
else if (Rs >= 261) {
return 1;
}
}
else if (SC >= 537) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 50) {
if (refDCval < 70) {
if (posBalance < 34) {
if (Cs < 48) {
if (Rs < 167) {
return 2;
}
else if (Rs >= 167) {
return 8;
}
}
else if (Cs >= 48) {
if (refDCval < 59) {
if (SC < 426) {
if (posBalance < 31) {
if (Cs < 239) {
return 8;
}
else if (Cs >= 239) {
return 1;
}
}
else if (posBalance >= 31) {
if (Rs < 130) {
return 4;
}
else if (Rs >= 130) {
return 8;
}
}
}
else if (SC >= 426) {
return 4;
}
}
else if (refDCval >= 59) {
if (tcor < 98) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (Rs < 170) {
return 1;
}
else if (Rs >= 170) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (posBalance < 30) {
if (Rs < 166) {
return 8;
}
else if (Rs >= 166) {
if (MVDiff < 28) {
return 4;
}
else if (MVDiff >= 28) {
return 8;
}
}
}
else if (posBalance >= 30) {
return 2;
}
}
}
}
}
else if (posBalance >= 34) {
if (TSC < 4) {
if (TSCindex < 2) {
if (refDCval < 66) {
if (MVDiff < 7) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Rs < 245) {
if (Cs < 190) {
if (ssDCval < 57) {
if (Rs < 152) {
return 4;
}
else if (Rs >= 152) {
return 1;
}
}
else if (ssDCval >= 57) {
if (posBalance < 41) {
if (Rs < 146) {
return 4;
}
else if (Rs >= 146) {
return 1;
}
}
else if (posBalance >= 41) {
return 1;
}
}
}
else if (Cs >= 190) {
return 4;
}
}
else if (Rs >= 245) {
if (negBalance < 18) {
return 1;
}
else if (negBalance >= 18) {
if (Rs < 255) {
return 1;
}
else if (Rs >= 255) {
return 4;
}
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 242) {
return 1;
}
else if (Rs >= 242) {
return 2;
}
}
}
else if (refDCval >= 66) {
if (Cs < 239) {
return 2;
}
else if (Cs >= 239) {
if (Cs < 379) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (Cs >= 379) {
return 1;
}
}
}
}
else if (TSCindex >= 2) {
if (RsCsDiff < 1) {
if (Cs < 182) {
if (MVDiff < 3) {
if (ssDCval < 65) {
if (Rs < 78) {
return 2;
}
else if (Rs >= 78) {
if (Rs < 200) {
if (negBalance < 20) {
return 4;
}
else if (negBalance >= 20) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (Rs < 132) {
return 2;
}
else if (Rs >= 132) {
return 1;
}
}
}
}
else if (Rs >= 200) {
return 4;
}
}
}
else if (ssDCval >= 65) {
if (Rs < 101) {
return 1;
}
else if (Rs >= 101) {
return 2;
}
}
}
else if (MVDiff >= 3) {
if (mcTcor < 95) {
if (AFD < 5) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (AFD >= 5) {
if (Rs < 21) {
return 4;
}
else if (Rs >= 21) {
return 1;
}
}
}
else if (mcTcor >= 95) {
if (tcor < 94) {
if (Cs < 32) {
return 1;
}
else if (Cs >= 32) {
if (ssDCval < 60) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
else if (ssDCval >= 60) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (AFD < 8) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (Rs < 158) {
return 1;
}
else if (Rs >= 158) {
return 4;
}
}
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
else if (tcor >= 94) {
if (negBalance < 21) {
if (mu_mv_mag_sq < 93) {
if (MVDiff < 28) {
return 1;
}
else if (MVDiff >= 28) {
return 8;
}
}
else if (mu_mv_mag_sq >= 93) {
return 4;
}
}
else if (negBalance >= 21) {
if (posBalance < 39) {
if (Rs < 110) {
if (Rs < 63) {
return 1;
}
else if (Rs >= 63) {
return 8;
}
}
else if (Rs >= 110) {
return 4;
}
}
else if (posBalance >= 39) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (posBalance < 45) {
if (MVDiff < 4) {
if (Rs < 152) {
return 4;
}
else if (Rs >= 152) {
return 1;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (posBalance >= 45) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 182) {
if (negBalance < 30) {
if (ssDCval < 60) {
if (MVDiff < 1) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (TSCindex < 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (TSCindex >= 3) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 192) {
if (Rs < 144) {
return 4;
}
else if (Rs >= 144) {
return 1;
}
}
else if (Cs >= 192) {
if (Rs < 92) {
return 1;
}
else if (Rs >= 92) {
return 4;
}
}
}
}
else if (ssDCval >= 60) {
if (MVDiff < 5) {
if (negBalance < 23) {
if (Cs < 343) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 187) {
return 2;
}
else if (Rs >= 187) {
return 4;
}
}
}
else if (Cs >= 343) {
if (diffAFD < 1) {
if (Cs < 344) {
return 1;
}
else if (Cs >= 344) {
if (Cs < 402) {
return 4;
}
else if (Cs >= 402) {
if (Rs < 202) {
return 4;
}
else if (Rs >= 202) {
return 1;
}
}
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (negBalance >= 23) {
if (posBalance < 47) {
if (ssDCval < 69) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (Rs < 177) {
if (Cs < 311) {
return 4;
}
else if (Cs >= 311) {
if (Cs < 388) {
return 1;
}
else if (Cs >= 388) {
return 4;
}
}
}
else if (Rs >= 177) {
if (SC < 493) {
if (Rs < 215) {
return 4;
}
else if (Rs >= 215) {
return 1;
}
}
else if (SC >= 493) {
return 4;
}
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 166) {
return 4;
}
else if (Rs >= 166) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (mu_mv_mag_sq < 119) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
if (SC < 448) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 1;
}
}
else if (SC >= 448) {
if (mu_mv_mag_sq < 110) {
if (posBalance < 48) {
if (Rs < 226) {
return 4;
}
else if (Rs >= 226) {
return 1;
}
}
else if (posBalance >= 48) {
return 1;
}
}
else if (mu_mv_mag_sq >= 110) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (Cs < 304) {
return 1;
}
else if (Cs >= 304) {
if (Rs < 175) {
return 1;
}
else if (Rs >= 175) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 5) {
if (Cs < 220) {
if (Rs < 243) {
return 1;
}
else if (Rs >= 243) {
return 4;
}
}
else if (Cs >= 220) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
}
else if (negBalance >= 30) {
if (tcor < 98) {
if (negBalance < 46) {
if (tcor < 97) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (posBalance < 52) {
if (mu_mv_mag_sq < 109) {
return 4;
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 195) {
return 4;
}
else if (Rs >= 195) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (SC < 424) {
return 1;
}
else if (SC >= 424) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 196) {
return 4;
}
else if (Rs >= 196) {
return 1;
}
}
}
}
}
}
else if (tcor >= 97) {
if (negBalance < 32) {
if (diffAFD < 1) {
if (MVDiff < 2) {
if (Cs < 329) {
return 1;
}
else if (Cs >= 329) {
return 4;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (negBalance >= 32) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 1;
}
}
}
}
else if (negBalance >= 46) {
return 2;
}
}
else if (tcor >= 98) {
if (posBalance < 43) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (posBalance >= 43) {
if (Cs < 491) {
if (posBalance < 48) {
if (MVDiff < 4) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
return 4;
}
}
else if (MVDiff >= 4) {
if (negBalance < 32) {
if (Rs < 180) {
return 1;
}
else if (Rs >= 180) {
return 4;
}
}
else if (negBalance >= 32) {
return 1;
}
}
}
else if (posBalance >= 48) {
return 1;
}
}
else if (Cs >= 491) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (MVDiff < 2) {
if (Cs < 404) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 118) {
if (Rs < 163) {
if (Rs < 141) {
return 4;
}
else if (Rs >= 141) {
return 1;
}
}
else if (Rs >= 163) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 118) {
return 1;
}
}
else if (ssDCval >= 69) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 81) {
return 1;
}
else if (mu_mv_mag_sq >= 81) {
return 4;
}
}
}
}
}
else if (Cs >= 404) {
return 1;
}
}
else if (MVDiff >= 2) {
if (Rs < 37) {
return 1;
}
else if (Rs >= 37) {
if (RsDiff < 1) {
if (MVDiff < 35) {
if (tcor < 105) {
if (ssDCval < 53) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
}
else if (ssDCval >= 53) {
if (Cs < 368) {
if (mu_mv_mag_sq < 79) {
if (diffRsCsdiff < 1) {
if (Cs < 361) {
if (Rs < 76) {
if (gchDC < 1) {
if (Cs < 184) {
return 1;
}
else if (Cs >= 184) {
return 4;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (Rs >= 76) {
return 4;
}
}
else if (Cs >= 361) {
return 1;
}
}
else if (diffRsCsdiff >= 1) {
if (Cs < 183) {
if (Cs < 155) {
return 4;
}
else if (Cs >= 155) {
return 1;
}
}
else if (Cs >= 183) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 79) {
return 1;
}
}
else if (Cs >= 368) {
return 4;
}
}
}
else if (tcor >= 105) {
return 8;
}
}
else if (MVDiff >= 35) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (RsDiff >= 1) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (AFD < 9) {
if (MVDiff < 4) {
if (Cs < 320) {
return 1;
}
else if (Cs >= 320) {
return 4;
}
}
else if (MVDiff >= 4) {
if (RsDiff < 2) {
if (mcTcor < 97) {
if (ssDCval < 62) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (Cs < 407) {
if (SC < 471) {
if (Cs < 72) {
return 2;
}
else if (Cs >= 72) {
return 4;
}
}
else if (SC >= 471) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 2;
}
}
}
else if (Cs >= 407) {
return 4;
}
}
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (mcTcor >= 97) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (Rs < 254) {
if (refDCval < 68) {
if (SC < 626) {
if (mu_mv_mag_sq < 89) {
return 4;
}
else if (mu_mv_mag_sq >= 89) {
return 1;
}
}
else if (SC >= 626) {
if (Rs < 209) {
return 4;
}
else if (Rs >= 209) {
return 8;
}
}
}
else if (refDCval >= 68) {
if (Cs < 191) {
return 8;
}
else if (Cs >= 191) {
return 4;
}
}
}
else if (Rs >= 254) {
if (diffRsCsdiff < -2) {
if (Rs < 278) {
return 2;
}
else if (Rs >= 278) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
if (refDCval < 63) {
return 4;
}
else if (refDCval >= 63) {
if (Cs < 243) {
return 4;
}
else if (Cs >= 243) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 2) {
return 2;
}
}
}
else if (AFD >= 9) {
if (tcor < 87) {
if (Cs < 187) {
return 4;
}
else if (Cs >= 187) {
if (MVDiff < 44) {
return 1;
}
else if (MVDiff >= 44) {
if (Rs < 240) {
return 4;
}
else if (Rs >= 240) {
return 2;
}
}
}
}
else if (tcor >= 87) {
return 1;
}
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (RsCsDiff < 1) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (SC < 121) {
if (Rs < 58) {
return 1;
}
else if (Rs >= 58) {
return 2;
}
}
else if (SC >= 121) {
return 8;
}
}
}
else if (RsCsDiff >= 1) {
if (tcor < 56) {
if (refDCval < 64) {
if (Cs < 214) {
if (posBalance < 93) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
if (Rs < 148) {
if (Rs < 99) {
return 2;
}
else if (Rs >= 99) {
return 4;
}
}
else if (Rs >= 148) {
return 1;
}
}
}
else if (posBalance >= 93) {
if (Rs < 151) {
return 4;
}
else if (Rs >= 151) {
return 2;
}
}
}
else if (Cs >= 214) {
if (tcor < 53) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 2;
}
}
}
else if (tcor >= 53) {
return 1;
}
}
}
else if (refDCval >= 64) {
if (mcTcor < 76) {
if (RsDiff < 3) {
if (refDCval < 66) {
return 8;
}
else if (refDCval >= 66) {
return 4;
}
}
else if (RsDiff >= 3) {
if (Rs < 174) {
return 8;
}
else if (Rs >= 174) {
return 1;
}
}
}
else if (mcTcor >= 76) {
if (Rs < 221) {
return 8;
}
else if (Rs >= 221) {
return 2;
}
}
}
}
else if (tcor >= 56) {
if (mu_mv_mag_sq < 40) {
if (MVDiff < 47) {
if (mu_mv_mag_sq < 37) {
if (negBalance < 36) {
if (Cs < 146) {
if (Rs < 150) {
return 1;
}
else if (Rs >= 150) {
return 8;
}
}
else if (Cs >= 146) {
if (Rs < 222) {
return 4;
}
else if (Rs >= 222) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (AFD < 13) {
if (diffMVdiffVal < -11) {
if (SC < 415) {
return 2;
}
else if (SC >= 415) {
return 1;
}
}
else if (diffMVdiffVal >= -11) {
return 1;
}
}
else if (AFD >= 13) {
if (Cs < 331) {
if (Rs < 174) {
return 1;
}
else if (Rs >= 174) {
return 4;
}
}
else if (Cs >= 331) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
return 2;
}
}
else if (MVDiff >= 47) {
if (diffAFD < -5) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (diffAFD >= -5) {
if (Cs < 35) {
return 1;
}
else if (Cs >= 35) {
if (Cs < 324) {
if (diffMVdiffVal < 13) {
if (AFD < 24) {
if (refDCval < 69) {
return 2;
}
else if (refDCval >= 69) {
return 4;
}
}
else if (AFD >= 24) {
return 1;
}
}
else if (diffMVdiffVal >= 13) {
if (SC < 413) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (SC >= 413) {
return 2;
}
}
}
else if (Cs >= 324) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (AFD < 7) {
if (MVDiff < 10) {
if (Rs < 239) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (Rs >= 239) {
return 1;
}
}
else if (MVDiff >= 10) {
return 2;
}
}
else if (AFD >= 7) {
if (negBalance < 59) {
if (gchDC < 6) {
return 4;
}
else if (gchDC >= 6) {
if (Cs < 50) {
return 1;
}
else if (Cs >= 50) {
return 8;
}
}
}
else if (negBalance >= 59) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 105) {
if (TSC < 2) {
if (ssDCval < 79) {
if (mu_mv_mag_sq < 73) {
return 1;
}
else if (mu_mv_mag_sq >= 73) {
if (Cs < 181) {
if (ssDCval < 75) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
return 2;
}
}
else if (ssDCval >= 75) {
if (Cs < 102) {
if (SC < 227) {
return 2;
}
else if (SC >= 227) {
return 1;
}
}
else if (Cs >= 102) {
if (Rs < 167) {
return 1;
}
else if (Rs >= 167) {
return 2;
}
}
}
}
else if (Cs >= 181) {
if (Rs < 109) {
return 4;
}
else if (Rs >= 109) {
return 1;
}
}
}
}
else if (ssDCval >= 79) {
if (Cs < 137) {
if (ssDCval < 118) {
if (MVDiff < 15) {
if (Rs < 94) {
if (Rs < 76) {
return 4;
}
else if (Rs >= 76) {
return 2;
}
}
else if (Rs >= 94) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (MVDiff >= 15) {
if (Cs < 127) {
if (ssDCval < 110) {
if (refDCval < 106) {
return 1;
}
else if (refDCval >= 106) {
if (Rs < 49) {
return 1;
}
else if (Rs >= 49) {
return 2;
}
}
}
else if (ssDCval >= 110) {
return 2;
}
}
else if (Cs >= 127) {
if (Rs < 65) {
return 2;
}
else if (Rs >= 65) {
return 1;
}
}
}
}
else if (ssDCval >= 118) {
if (diffAFD < -2) {
if (Cs < 105) {
return 2;
}
else if (Cs >= 105) {
return 8;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
}
else if (Cs >= 137) {
if (ssDCval < 132) {
if (Rs < 180) {
if (MVDiff < 35) {
if (MVDiff < 6) {
if (Rs < 133) {
if (Cs < 153) {
return 4;
}
else if (Cs >= 153) {
if (Cs < 212) {
return 1;
}
else if (Cs >= 212) {
return 8;
}
}
}
else if (Rs >= 133) {
return 4;
}
}
else if (MVDiff >= 6) {
if (Rs < 149) {
return 4;
}
else if (Rs >= 149) {
if (Cs < 256) {
return 2;
}
else if (Cs >= 256) {
return 4;
}
}
}
}
else if (MVDiff >= 35) {
return 2;
}
}
else if (Rs >= 180) {
if (Cs < 288) {
if (Rs < 242) {
return 2;
}
else if (Rs >= 242) {
if (Rs < 264) {
return 4;
}
else if (Rs >= 264) {
return 1;
}
}
}
else if (Cs >= 288) {
return 4;
}
}
}
else if (ssDCval >= 132) {
if (Cs < 196) {
return 8;
}
else if (Cs >= 196) {
return 2;
}
}
}
}
}
else if (TSC >= 2) {
if (ssDCval < 84) {
if (Cs < 239) {
if (MVDiff < 4) {
if (Cs < 157) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (Cs >= 157) {
if (TSC < 3) {
if (AFD < 3) {
if (Rs < 228) {
return 2;
}
else if (Rs >= 228) {
if (Rs < 256) {
return 8;
}
else if (Rs >= 256) {
return 1;
}
}
}
else if (AFD >= 3) {
if (Cs < 224) {
return 1;
}
else if (Cs >= 224) {
return 2;
}
}
}
else if (TSC >= 3) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
}
}
else if (MVDiff >= 4) {
if (AFD < 9) {
if (mu_mv_mag_sq < 41) {
if (Cs < 88) {
return 1;
}
else if (Cs >= 88) {
if (AFD < 6) {
if (diffAFD < -1) {
if (Rs < 108) {
return 4;
}
else if (Rs >= 108) {
return 1;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (AFD >= 6) {
if (tcor < 92) {
if (Cs < 200) {
return 1;
}
else if (Cs >= 200) {
return 4;
}
}
else if (tcor >= 92) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (AFD < 7) {
if (Rs < 130) {
if (mu_mv_mag_sq < 63) {
return 1;
}
else if (mu_mv_mag_sq >= 63) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
return 2;
}
}
}
else if (Rs >= 130) {
if (posBalance < 48) {
if (SC < 230) {
return 8;
}
else if (SC >= 230) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (posBalance >= 48) {
return 1;
}
}
}
else if (AFD >= 7) {
if (Rs < 231) {
return 8;
}
else if (Rs >= 231) {
return 2;
}
}
}
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 47) {
if (ssDCval < 82) {
if (Cs < 160) {
if (TSC < 9) {
if (SC < 274) {
if (AFD < 10) {
if (Rs < 109) {
return 2;
}
else if (Rs >= 109) {
return 1;
}
}
else if (AFD >= 10) {
return 2;
}
}
else if (SC >= 274) {
return 1;
}
}
else if (TSC >= 9) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 1;
}
}
}
else if (Cs >= 160) {
if (diffMVdiffVal < 21) {
if (Rs < 287) {
return 2;
}
else if (Rs >= 287) {
return 4;
}
}
else if (diffMVdiffVal >= 21) {
if (Rs < 127) {
return 1;
}
else if (Rs >= 127) {
return 4;
}
}
}
}
else if (ssDCval >= 82) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
if (Rs < 207) {
return 8;
}
else if (Rs >= 207) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (TSC < 6) {
if (MVDiff < 527) {
return 1;
}
else if (MVDiff >= 527) {
return 4;
}
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 53) {
return 8;
}
else if (mu_mv_mag_sq >= 53) {
if (Cs < 111) {
return 4;
}
else if (Cs >= 111) {
if (Cs < 192) {
return 8;
}
else if (Cs >= 192) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Cs >= 239) {
if (mcTcor < 97) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (Rs < 277) {
if (AFD < 27) {
if (SC < 632) {
if (mu_mv_mag_sq < 15) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 15) {
if (MVDiff < 46) {
return 1;
}
else if (MVDiff >= 46) {
return 8;
}
}
}
else if (SC >= 632) {
return 8;
}
}
else if (AFD >= 27) {
if (Rs < 202) {
return 8;
}
else if (Rs >= 202) {
return 2;
}
}
}
else if (Rs >= 277) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 76) {
if (Rs < 114) {
return 4;
}
else if (Rs >= 114) {
if (CsDiff < 1) {
if (ssDCval < 72) {
if (mu_mv_mag_sq < 84) {
if (Rs < 120) {
return 4;
}
else if (Rs >= 120) {
return 1;
}
}
else if (mu_mv_mag_sq >= 84) {
if (Rs < 216) {
return 4;
}
else if (Rs >= 216) {
return 1;
}
}
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (CsDiff >= 1) {
if (MVDiff < 6) {
if (Rs < 136) {
return 4;
}
else if (Rs >= 136) {
return 1;
}
}
else if (MVDiff >= 6) {
if (Cs < 435) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 1;
}
}
else if (Cs >= 435) {
return 1;
}
}
}
}
}
else if (ssDCval >= 76) {
if (SC < 522) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (ssDCval < 78) {
if (AFD < 8) {
if (Rs < 194) {
return 8;
}
else if (Rs >= 194) {
return 2;
}
}
else if (AFD >= 8) {
return 4;
}
}
else if (ssDCval >= 78) {
return 2;
}
}
}
else if (SC >= 522) {
return 1;
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 113) {
if (negBalance < 43) {
if (MVDiff < 28) {
if (ssDCval < 96) {
if (SC < 366) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
return 1;
}
}
else if (SC >= 366) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (ssDCval >= 96) {
if (posBalance < 62) {
if (MVDiff < 20) {
return 1;
}
else if (MVDiff >= 20) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
return 1;
}
}
}
else if (posBalance >= 62) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
return 1;
}
}
}
}
else if (MVDiff >= 28) {
if (Cs < 178) {
if (TSC < 4) {
if (Rs < 90) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 4;
}
}
else if (Rs >= 90) {
return 2;
}
}
else if (TSC >= 4) {
if (RsDiff < 2) {
if (Rs < 44) {
return 1;
}
else if (Rs >= 44) {
return 8;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (Cs >= 178) {
if (diffRsCsdiff < -3) {
return 8;
}
else if (diffRsCsdiff >= -3) {
return 2;
}
}
}
}
else if (negBalance >= 43) {
if (mcTcor < 71) {
return 4;
}
else if (mcTcor >= 71) {
if (Rs < 102) {
if (negBalance < 57) {
if (ssDCval < 107) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (refDCval < 95) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
return 2;
}
}
else if (refDCval >= 95) {
return 2;
}
}
}
else if (ssDCval >= 107) {
if (Rs < 95) {
return 2;
}
else if (Rs >= 95) {
return 4;
}
}
}
else if (negBalance >= 57) {
if (Cs < 302) {
return 1;
}
else if (Cs >= 302) {
return 2;
}
}
}
else if (Rs >= 102) {
if (RsDiff < 2) {
if (mcTcor < 97) {
if (Cs < 167) {
if (tcor < 50) {
if (Rs < 146) {
if (MVDiff < 44) {
if (diffMVdiffVal < -16) {
return 2;
}
else if (diffMVdiffVal >= -16) {
return 8;
}
}
else if (MVDiff >= 44) {
return 2;
}
}
else if (Rs >= 146) {
return 2;
}
}
else if (tcor >= 50) {
if (Rs < 144) {
return 1;
}
else if (Rs >= 144) {
if (Cs < 57) {
return 1;
}
else if (Cs >= 57) {
return 8;
}
}
}
}
else if (Cs >= 167) {
if (SC < 576) {
if (SC < 538) {
if (Cs < 291) {
if (MVDiff < 7) {
if (CsDiff < 1) {
if (diffRsCsdiff < -1) {
if (Rs < 149) {
return 2;
}
else if (Rs >= 149) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
return 2;
}
}
else if (CsDiff >= 1) {
if (Cs < 203) {
return 2;
}
else if (Cs >= 203) {
return 8;
}
}
}
else if (MVDiff >= 7) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (Cs >= 291) {
return 8;
}
}
else if (SC >= 538) {
if (Rs < 280) {
return 2;
}
else if (Rs >= 280) {
return 1;
}
}
}
else if (SC >= 576) {
if (diffMVdiffVal < -5) {
return 2;
}
else if (diffMVdiffVal >= -5) {
if (Cs < 321) {
return 2;
}
else if (Cs >= 321) {
if (posBalance < 58) {
if (Rs < 255) {
return 8;
}
else if (Rs >= 255) {
return 2;
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 15) {
if (tcor < 96) {
if (posBalance < 70) {
return 2;
}
else if (posBalance >= 70) {
return 1;
}
}
else if (tcor >= 96) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (Rs < 209) {
if (refDCval < 106) {
return 1;
}
else if (refDCval >= 106) {
if (Cs < 147) {
return 2;
}
else if (Cs >= 147) {
return 1;
}
}
}
else if (Rs >= 209) {
if (Cs < 68) {
return 1;
}
else if (Cs >= 68) {
return 8;
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 242) {
if (Cs < 98) {
if (tcor < 51) {
if (Cs < 34) {
return 2;
}
else if (Cs >= 34) {
return 4;
}
}
else if (tcor >= 51) {
if (Rs < 177) {
return 2;
}
else if (Rs >= 177) {
return 1;
}
}
}
else if (Cs >= 98) {
if (negBalance < 53) {
if (Cs < 227) {
if (Rs < 186) {
return 8;
}
else if (Rs >= 186) {
return 2;
}
}
else if (Cs >= 227) {
return 1;
}
}
else if (negBalance >= 53) {
if (diffAFD < -2) {
if (Rs < 226) {
return 4;
}
else if (Rs >= 226) {
return 2;
}
}
else if (diffAFD >= -2) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
return 1;
}
}
}
}
}
else if (Rs >= 242) {
if (tcor < 67) {
if (ssDCval < 97) {
if (Cs < 49) {
return 8;
}
else if (Cs >= 49) {
return 2;
}
}
else if (ssDCval >= 97) {
return 1;
}
}
else if (tcor >= 67) {
if (tcor < 86) {
if (posBalance < 69) {
return 1;
}
else if (posBalance >= 69) {
return 2;
}
}
else if (tcor >= 86) {
if (Cs < 152) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 2;
}
}
else if (Cs >= 152) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (refDCval < 160) {
if (tcor < 101) {
if (TSC < 3) {
if (Rs < 6) {
return 1;
}
else if (Rs >= 6) {
return 2;
}
}
else if (TSC >= 3) {
if (Cs < 256) {
if (RsCsDiff < 3) {
if (Cs < 40) {
if (mcTcor < 96) {
if (MVDiff < 101) {
if (MVDiff < 55) {
return 2;
}
else if (MVDiff >= 55) {
return 1;
}
}
else if (MVDiff >= 101) {
return 2;
}
}
else if (mcTcor >= 96) {
if (Rs < 119) {
return 8;
}
else if (Rs >= 119) {
if (Rs < 211) {
return 1;
}
else if (Rs >= 211) {
return 2;
}
}
}
}
else if (Cs >= 40) {
return 2;
}
}
else if (RsCsDiff >= 3) {
if (MVDiff < 60) {
if (RsCsDiff < 7) {
return 2;
}
else if (RsCsDiff >= 7) {
return 4;
}
}
else if (MVDiff >= 60) {
if (MVDiff < 106) {
if (Cs < 124) {
return 8;
}
else if (Cs >= 124) {
return 1;
}
}
else if (MVDiff >= 106) {
if (Rs < 294) {
return 8;
}
else if (Rs >= 294) {
return 2;
}
}
}
}
}
else if (Cs >= 256) {
if (Cs < 414) {
if (AFD < 16) {
if (ssDCval < 119) {
if (Cs < 323) {
return 2;
}
else if (Cs >= 323) {
return 8;
}
}
else if (ssDCval >= 119) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (AFD >= 16) {
if (Rs < 287) {
return 1;
}
else if (Rs >= 287) {
return 2;
}
}
}
else if (Cs >= 414) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
return 8;
}
}
}
}
}
else if (tcor >= 101) {
return 8;
}
}
else if (refDCval >= 160) {
if (mcTcor < 98) {
if (ssDCval < 185) {
if (mu_mv_mag_sq < 45) {
if (posBalance < 63) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
if (Rs < 222) {
return 2;
}
else if (Rs >= 222) {
return 1;
}
}
}
else if (posBalance >= 63) {
if (Rs < 101) {
return 8;
}
else if (Rs >= 101) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (Rs < 4) {
return 1;
}
else if (Rs >= 4) {
return 2;
}
}
}
else if (ssDCval >= 185) {
return 2;
}
}
else if (mcTcor >= 98) {
if (Rs < 256) {
return 2;
}
else if (Rs >= 256) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Cs < 121) {
if (refDCval < 86) {
if (Rs < 97) {
return 4;
}
else if (Rs >= 97) {
return 8;
}
}
else if (refDCval >= 86) {
if (Cs < 104) {
return 1;
}
else if (Cs >= 104) {
return 8;
}
}
}
else if (Cs >= 121) {
if (Cs < 239) {
if (TSCindex < 3) {
if (Cs < 164) {
if (Cs < 125) {
if (SC < 260) {
return 2;
}
else if (SC >= 260) {
return 8;
}
}
else if (Cs >= 125) {
if (Rs < 102) {
return 4;
}
else if (Rs >= 102) {
if (Rs < 206) {
if (negBalance < 20) {
return 2;
}
else if (negBalance >= 20) {
if (mu_mv_mag_sq < 120) {
return 2;
}
else if (mu_mv_mag_sq >= 120) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
}
}
else if (Rs >= 206) {
return 8;
}
}
}
}
else if (Cs >= 164) {
if (MVDiff < 1) {
if (ssDCval < 81) {
if (Cs < 232) {
return 2;
}
else if (Cs >= 232) {
if (Rs < 246) {
return 1;
}
else if (Rs >= 246) {
return 2;
}
}
}
else if (ssDCval >= 81) {
return 1;
}
}
else if (MVDiff >= 1) {
if (Rs < 178) {
if (Cs < 167) {
return 1;
}
else if (Cs >= 167) {
return 2;
}
}
else if (Rs >= 178) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
if (mu_mv_mag_sq < 118) {
return 2;
}
else if (mu_mv_mag_sq >= 118) {
return 1;
}
}
}
}
}
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 113) {
return 1;
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 131) {
if (Cs < 221) {
return 1;
}
else if (Cs >= 221) {
return 8;
}
}
else if (Rs >= 131) {
if (Rs < 205) {
return 2;
}
else if (Rs >= 205) {
return 1;
}
}
}
}
}
else if (Cs >= 239) {
if (ssDCval < 78) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 120) {
return 1;
}
else if (Rs >= 120) {
return 8;
}
}
}
else if (ssDCval >= 78) {
if (Cs < 314) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (Cs >= 314) {
if (Rs < 246) {
return 4;
}
else if (Rs >= 246) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 300) {
if (RsDiff < 4) {
if (refDCval < 75) {
if (mcTcor < 87) {
if (AFD < 10) {
if (negBalance < 32) {
if (negBalance < 26) {
if (posBalance < 61) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 8;
}
}
else if (posBalance >= 61) {
if (Rs < 423) {
if (Rs < 348) {
return 1;
}
else if (Rs >= 348) {
return 4;
}
}
else if (Rs >= 423) {
return 8;
}
}
}
else if (negBalance >= 26) {
if (Cs < 346) {
if (Cs < 158) {
return 1;
}
else if (Cs >= 158) {
return 8;
}
}
else if (Cs >= 346) {
return 2;
}
}
}
else if (negBalance >= 32) {
if (Cs < 126) {
return 1;
}
else if (Cs >= 126) {
if (negBalance < 51) {
if (diffTSC < 1) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
return 4;
}
}
else if (diffTSC >= 1) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (Rs < 543) {
return 2;
}
else if (Rs >= 543) {
return 8;
}
}
}
}
else if (negBalance >= 51) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Cs < 210) {
return 8;
}
else if (Cs >= 210) {
return 4;
}
}
}
}
}
}
else if (AFD >= 10) {
if (MVDiff < 60) {
if (ssDCval < 56) {
if (MVDiff < 44) {
if (TSC < 6) {
if (refDCval < 41) {
return 8;
}
else if (refDCval >= 41) {
return 4;
}
}
else if (TSC >= 6) {
if (RsCsDiff < 7) {
if (Rs < 327) {
return 8;
}
else if (Rs >= 327) {
return 4;
}
}
else if (RsCsDiff >= 7) {
return 8;
}
}
}
else if (MVDiff >= 44) {
if (tcor < 62) {
if (Rs < 685) {
return 4;
}
else if (Rs >= 685) {
return 1;
}
}
else if (tcor >= 62) {
return 1;
}
}
}
else if (ssDCval >= 56) {
if (Cs < 213) {
return 2;
}
else if (Cs >= 213) {
return 4;
}
}
}
else if (MVDiff >= 60) {
if (refDCval < 51) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 40) {
if (Rs < 406) {
return 4;
}
else if (Rs >= 406) {
return 1;
}
}
else if (ssDCval >= 40) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
return 1;
}
}
else if (refDCval >= 51) {
if (mcTcor < 79) {
return 8;
}
else if (mcTcor >= 79) {
if (negBalance < 54) {
return 4;
}
else if (negBalance >= 54) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 87) {
if (Rs < 1294) {
if (tcor < 101) {
if (CsDiff < 1) {
if (ssDCval < 58) {
if (TSC < 1) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (Rs < 728) {
return 4;
}
else if (Rs >= 728) {
return 1;
}
}
}
else if (TSC >= 1) {
if (mcTcor < 98) {
if (RsDiff < 2) {
if (negBalance < 31) {
if (mu_mv_mag_sq < 114) {
if (Cs < 440) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
if (ssDCval < 47) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (ssDCval >= 47) {
if (Rs < 415) {
return 4;
}
else if (Rs >= 415) {
return 1;
}
}
}
}
else if (Cs >= 440) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (negBalance >= 31) {
if (AFD < 5) {
if (diffAFD < -1) {
if (Rs < 448) {
if (Rs < 369) {
return 8;
}
else if (Rs >= 369) {
return 4;
}
}
else if (Rs >= 448) {
return 1;
}
}
else if (diffAFD >= -1) {
if (Cs < 225) {
return 8;
}
else if (Cs >= 225) {
return 4;
}
}
}
else if (AFD >= 5) {
if (tcor < 90) {
if (Rs < 418) {
return 4;
}
else if (Rs >= 418) {
return 8;
}
}
else if (tcor >= 90) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
return 4;
}
}
}
}
}
else if (RsDiff >= 2) {
if (TSC < 6) {
if (Rs < 935) {
return 1;
}
else if (Rs >= 935) {
return 2;
}
}
else if (TSC >= 6) {
if (Rs < 572) {
return 8;
}
else if (Rs >= 572) {
return 1;
}
}
}
}
else if (mcTcor >= 98) {
if (MVDiff < 5) {
if (negBalance < 21) {
if (SC < 1078) {
if (ssDCval < 54) {
if (Rs < 659) {
return 1;
}
else if (Rs >= 659) {
return 8;
}
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (SC >= 1078) {
if (Rs < 1152) {
return 1;
}
else if (Rs >= 1152) {
return 8;
}
}
}
else if (negBalance >= 21) {
if (mu_mv_mag_sq < 87) {
if (Cs < 357) {
return 1;
}
else if (Cs >= 357) {
return 8;
}
}
else if (mu_mv_mag_sq >= 87) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (AFD < 3) {
if (Cs < 412) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (Cs < 197) {
return 4;
}
else if (Cs >= 197) {
return 1;
}
}
}
else if (Cs >= 412) {
return 8;
}
}
else if (AFD >= 3) {
if (Rs < 1095) {
if (ssDCval < 56) {
if (MVDiff < 44) {
return 1;
}
else if (MVDiff >= 44) {
return 4;
}
}
else if (ssDCval >= 56) {
if (Rs < 428) {
return 4;
}
else if (Rs >= 428) {
return 1;
}
}
}
else if (Rs >= 1095) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 58) {
if (Rs < 980) {
if (Cs < 434) {
if (AFD < 3) {
if (mu_mv_mag_sq < 106) {
if (Cs < 270) {
if (refDCval < 70) {
if (MVDiff < 1) {
if (Rs < 776) {
return 2;
}
else if (Rs >= 776) {
return 1;
}
}
else if (MVDiff >= 1) {
if (SC < 882) {
return 4;
}
else if (SC >= 882) {
if (Rs < 719) {
return 8;
}
else if (Rs >= 719) {
return 4;
}
}
}
}
else if (refDCval >= 70) {
return 8;
}
}
else if (Cs >= 270) {
if (negBalance < 27) {
if (Rs < 663) {
if (Rs < 349) {
return 2;
}
else if (Rs >= 349) {
return 8;
}
}
else if (Rs >= 663) {
if (Rs < 754) {
return 2;
}
else if (Rs >= 754) {
return 1;
}
}
}
else if (negBalance >= 27) {
if (posBalance < 46) {
if (Rs < 798) {
return 1;
}
else if (Rs >= 798) {
return 4;
}
}
else if (posBalance >= 46) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (ssDCval < 72) {
if (mcTcor < 99) {
if (Cs < 342) {
if (negBalance < 22) {
return 4;
}
else if (negBalance >= 22) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (Rs < 598) {
return 4;
}
else if (Rs >= 598) {
return 2;
}
}
}
}
else if (Cs >= 342) {
return 1;
}
}
else if (mcTcor >= 99) {
if (posBalance < 40) {
if (SC < 1317) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 417) {
return 4;
}
else if (Rs >= 417) {
if (refDCval < 69) {
return 1;
}
else if (refDCval >= 69) {
if (Rs < 573) {
return 2;
}
else if (Rs >= 573) {
return 1;
}
}
}
}
}
else if (SC >= 1317) {
return 8;
}
}
else if (posBalance >= 40) {
if (Cs < 366) {
if (Rs < 518) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
if (Rs < 436) {
return 4;
}
else if (Rs >= 436) {
return 8;
}
}
}
else if (Rs >= 518) {
return 2;
}
}
else if (Cs >= 366) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
}
else if (ssDCval >= 72) {
if (Rs < 556) {
if (Rs < 444) {
return 8;
}
else if (Rs >= 444) {
return 1;
}
}
else if (Rs >= 556) {
return 4;
}
}
}
}
else if (AFD >= 3) {
if (Cs < 182) {
if (Rs < 497) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
return 8;
}
}
else if (Rs >= 497) {
if (TSC < 6) {
if (Rs < 558) {
if (Cs < 84) {
return 1;
}
else if (Cs >= 84) {
return 4;
}
}
else if (Rs >= 558) {
return 1;
}
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (Cs >= 182) {
if (mcTcor < 98) {
if (MVDiff < 7) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 2;
}
}
else if (MVDiff >= 7) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 55) {
if (Cs < 199) {
return 8;
}
else if (Cs >= 199) {
return 4;
}
}
else if (mu_mv_mag_sq >= 55) {
if (ssDCval < 70) {
if (negBalance < 26) {
if (Rs < 651) {
return 8;
}
else if (Rs >= 651) {
return 1;
}
}
else if (negBalance >= 26) {
return 1;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
}
}
}
else if (Cs >= 434) {
if (SC < 1244) {
if (MVDiff < 2) {
if (refDCval < 69) {
if (Cs < 486) {
return 1;
}
else if (Cs >= 486) {
if (Rs < 565) {
if (Rs < 463) {
return 1;
}
else if (Rs >= 463) {
return 8;
}
}
else if (Rs >= 565) {
return 1;
}
}
}
else if (refDCval >= 69) {
if (Rs < 443) {
return 2;
}
else if (Rs >= 443) {
if (Rs < 676) {
return 1;
}
else if (Rs >= 676) {
return 2;
}
}
}
}
else if (MVDiff >= 2) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (SC >= 1244) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 825) {
return 2;
}
else if (Rs >= 825) {
return 1;
}
}
}
}
}
else if (Rs >= 980) {
if (SC < 1573) {
return 1;
}
else if (SC >= 1573) {
if (Rs < 1246) {
return 1;
}
else if (Rs >= 1246) {
return 8;
}
}
}
}
}
else if (CsDiff >= 1) {
if (TSCindex < 2) {
if (negBalance < 18) {
return 8;
}
else if (negBalance >= 18) {
return 4;
}
}
else if (TSCindex >= 2) {
if (mu_mv_mag_sq < 85) {
if (posBalance < 34) {
if (ssDCval < 55) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
if (Rs < 550) {
return 4;
}
else if (Rs >= 550) {
if (Rs < 696) {
return 8;
}
else if (Rs >= 696) {
return 2;
}
}
}
}
else if (ssDCval >= 55) {
if (negBalance < 22) {
if (Cs < 313) {
return 4;
}
else if (Cs >= 313) {
return 2;
}
}
else if (negBalance >= 22) {
return 1;
}
}
}
else if (posBalance >= 34) {
if (refDCval < 58) {
if (negBalance < 38) {
if (Rs < 657) {
if (mcTcor < 94) {
if (RsCsDiff < 4) {
if (Cs < 302) {
return 4;
}
else if (Cs >= 302) {
if (negBalance < 33) {
return 1;
}
else if (negBalance >= 33) {
if (Rs < 425) {
return 4;
}
else if (Rs >= 425) {
return 1;
}
}
}
}
else if (RsCsDiff >= 4) {
if (mu_mv_mag_sq < 13) {
if (MVDiff < 29) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
return 1;
}
}
else if (MVDiff >= 29) {
return 1;
}
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 318) {
return 2;
}
else if (Rs >= 318) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (refDCval < 54) {
if (SC < 987) {
if (SC < 887) {
if (refDCval < 44) {
if (RsDiff < 2) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
return 1;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (refDCval >= 44) {
if (SC < 805) {
return 4;
}
else if (SC >= 805) {
if (RsCsDiff < 4) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 4;
}
}
else if (RsCsDiff >= 4) {
return 4;
}
}
}
}
else if (SC >= 887) {
if (MVDiff < 7) {
if (Rs < 569) {
return 4;
}
else if (Rs >= 569) {
return 1;
}
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (SC >= 987) {
if (SC < 1079) {
if (negBalance < 24) {
return 1;
}
else if (negBalance >= 24) {
return 4;
}
}
else if (SC >= 1079) {
return 8;
}
}
}
else if (refDCval >= 54) {
if (Rs < 439) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (MVDiff < 7) {
if (Rs < 369) {
return 2;
}
else if (Rs >= 369) {
return 8;
}
}
else if (MVDiff >= 7) {
if (Rs < 308) {
return 2;
}
else if (Rs >= 308) {
return 4;
}
}
}
}
else if (Rs >= 439) {
if (AFD < 5) {
if (Rs < 581) {
return 8;
}
else if (Rs >= 581) {
return 4;
}
}
else if (AFD >= 5) {
if (MVDiff < 12) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 1;
}
}
else if (MVDiff >= 12) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 657) {
if (Cs < 410) {
if (tcor < 79) {
if (Cs < 184) {
return 2;
}
else if (Cs >= 184) {
if (Rs < 746) {
return 8;
}
else if (Rs >= 746) {
return 1;
}
}
}
else if (tcor >= 79) {
if (diffAFD < 3) {
return 1;
}
else if (diffAFD >= 3) {
return 2;
}
}
}
else if (Cs >= 410) {
if (mu_mv_mag_sq < 67) {
if (posBalance < 45) {
if (mu_mv_mag_sq < 53) {
return 2;
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 735) {
return 2;
}
else if (Rs >= 735) {
return 1;
}
}
}
else if (posBalance >= 45) {
if (Cs < 456) {
return 8;
}
else if (Cs >= 456) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 67) {
if (RsDiff < 3) {
if (Cs < 516) {
return 1;
}
else if (Cs >= 516) {
if (Rs < 934) {
return 4;
}
else if (Rs >= 934) {
return 2;
}
}
}
else if (RsDiff >= 3) {
return 4;
}
}
}
}
}
else if (negBalance >= 38) {
if (SC < 924) {
if (SC < 594) {
if (Cs < 193) {
if (RsDiff < 2) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 4;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Cs >= 193) {
if (Cs < 201) {
return 8;
}
else if (Cs >= 201) {
return 4;
}
}
}
else if (SC >= 594) {
if (MVDiff < 43) {
if (AFD < 9) {
if (Rs < 360) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (Cs < 321) {
return 8;
}
else if (Cs >= 321) {
return 1;
}
}
}
else if (Rs >= 360) {
if (posBalance < 58) {
if (mcTcor < 94) {
if (MVDiff < 28) {
return 8;
}
else if (MVDiff >= 28) {
return 1;
}
}
else if (mcTcor >= 94) {
if (Rs < 456) {
if (Rs < 404) {
return 1;
}
else if (Rs >= 404) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
return 8;
}
}
}
else if (Rs >= 456) {
return 1;
}
}
}
else if (posBalance >= 58) {
if (Cs < 381) {
return 1;
}
else if (Cs >= 381) {
return 4;
}
}
}
}
else if (AFD >= 9) {
if (MVDiff < 12) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
return 8;
}
}
else if (MVDiff >= 12) {
if (Cs < 282) {
if (MVDiff < 20) {
return 1;
}
else if (MVDiff >= 20) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
return 8;
}
}
}
else if (Cs >= 282) {
if (MVDiff < 34) {
return 4;
}
else if (MVDiff >= 34) {
if (Rs < 393) {
return 1;
}
else if (Rs >= 393) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 43) {
if (CsDiff < 3) {
if (diffAFD < -2) {
if (Cs < 246) {
return 8;
}
else if (Cs >= 246) {
if (Rs < 342) {
return 1;
}
else if (Rs >= 342) {
return 4;
}
}
}
else if (diffAFD >= -2) {
return 2;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
}
}
else if (SC >= 924) {
if (Cs < 315) {
if (SC < 963) {
if (Rs < 704) {
return 8;
}
else if (Rs >= 704) {
return 2;
}
}
else if (SC >= 963) {
return 1;
}
}
else if (Cs >= 315) {
if (mu_mv_mag_sq < 26) {
if (mcTcor < 92) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (Cs < 500) {
return 8;
}
else if (Cs >= 500) {
return 4;
}
}
}
else if (mcTcor >= 92) {
if (Rs < 577) {
return 1;
}
else if (Rs >= 577) {
if (ssDCval < 38) {
return 1;
}
else if (ssDCval >= 38) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
if (Rs < 860) {
return 1;
}
else if (Rs >= 860) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (Rs < 559) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
else if (Rs >= 559) {
if (ssDCval < 42) {
return 2;
}
else if (ssDCval >= 42) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 58) {
if (ssDCval < 61) {
if (SC < 1073) {
if (tcor < 97) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (AFD < 6) {
if (ssDCval < 59) {
if (RsCsDiff < 7) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 2;
}
}
else if (RsCsDiff >= 7) {
return 2;
}
}
else if (ssDCval >= 59) {
return 2;
}
}
else if (AFD >= 6) {
if (AFD < 9) {
if (Cs < 249) {
return 1;
}
else if (Cs >= 249) {
if (mu_mv_mag_sq < 28) {
if (tcor < 93) {
return 2;
}
else if (tcor >= 93) {
return 4;
}
}
else if (mu_mv_mag_sq >= 28) {
if (Cs < 345) {
return 4;
}
else if (Cs >= 345) {
return 1;
}
}
}
}
else if (AFD >= 9) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (Cs < 165) {
return 1;
}
else if (Cs >= 165) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 97) {
if (mu_mv_mag_sq < 60) {
return 4;
}
else if (mu_mv_mag_sq >= 60) {
if (Rs < 423) {
return 4;
}
else if (Rs >= 423) {
return 8;
}
}
}
}
else if (SC >= 1073) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (tcor < 90) {
if (Rs < 805) {
return 4;
}
else if (Rs >= 805) {
return 8;
}
}
else if (tcor >= 90) {
return 1;
}
}
}
}
else if (ssDCval >= 61) {
if (Rs < 747) {
if (RsDiff < 3) {
if (Rs < 511) {
if (refDCval < 70) {
if (posBalance < 67) {
if (posBalance < 52) {
if (mcTcor < 96) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
return 2;
}
}
else if (mcTcor >= 96) {
if (AFD < 7) {
if (MVDiff < 4) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (MVDiff >= 4) {
if (Rs < 312) {
return 2;
}
else if (Rs >= 312) {
return 4;
}
}
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (negBalance < 47) {
if (Rs < 367) {
return 1;
}
else if (Rs >= 367) {
if (posBalance < 58) {
if (Rs < 411) {
return 1;
}
else if (Rs >= 411) {
return 4;
}
}
else if (posBalance >= 58) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (Cs < 400) {
return 4;
}
else if (Cs >= 400) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 67) {
if (AFD < 10) {
if (Rs < 367) {
return 4;
}
else if (Rs >= 367) {
return 2;
}
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (refDCval >= 70) {
if (refDCval < 73) {
if (Cs < 450) {
return 1;
}
else if (Cs >= 450) {
return 4;
}
}
else if (refDCval >= 73) {
if (Rs < 316) {
if (Rs < 307) {
return 2;
}
else if (Rs >= 307) {
return 8;
}
}
else if (Rs >= 316) {
if (Rs < 393) {
return 1;
}
else if (Rs >= 393) {
return 4;
}
}
}
}
}
else if (Rs >= 511) {
if (refDCval < 67) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 1060) {
return 1;
}
else if (SC >= 1060) {
return 4;
}
}
}
}
else if (refDCval >= 67) {
return 4;
}
}
}
else if (RsDiff >= 3) {
if (SC < 939) {
if (refDCval < 71) {
if (TSC < 8) {
if (diffRsCsdiff < -11) {
return 8;
}
else if (diffRsCsdiff >= -11) {
return 1;
}
}
else if (TSC >= 8) {
if (SC < 920) {
return 4;
}
else if (SC >= 920) {
return 8;
}
}
}
else if (refDCval >= 71) {
return 8;
}
}
else if (SC >= 939) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Cs < 337) {
return 1;
}
else if (Cs >= 337) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
return 4;
}
}
}
}
}
}
else if (Rs >= 747) {
if (mcTcor < 96) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Cs < 321) {
return 2;
}
else if (Cs >= 321) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
if (tcor < 97) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (negBalance < 42) {
if (Rs < 798) {
return 8;
}
else if (Rs >= 798) {
return 1;
}
}
else if (negBalance >= 42) {
if (Rs < 930) {
return 1;
}
else if (Rs >= 930) {
return 4;
}
}
}
}
else if (tcor >= 97) {
if (mu_mv_mag_sq < 56) {
if (SC < 1552) {
return 4;
}
else if (SC >= 1552) {
return 2;
}
}
else if (mu_mv_mag_sq >= 56) {
return 1;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 477) {
if (ssDCval < 64) {
if (MVDiff < 3) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 114) {
return 1;
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (MVDiff >= 3) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
if (posBalance < 45) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (Rs < 372) {
return 4;
}
else if (Rs >= 372) {
return 2;
}
}
}
else if (posBalance >= 45) {
if (Cs < 326) {
if (Rs < 366) {
return 2;
}
else if (Rs >= 366) {
return 4;
}
}
else if (Cs >= 326) {
return 1;
}
}
}
}
}
else if (ssDCval >= 64) {
if (Cs < 410) {
if (mcTcor < 99) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
else if (mcTcor >= 99) {
return 8;
}
}
else if (Cs >= 410) {
if (negBalance < 31) {
if (Rs < 368) {
return 1;
}
else if (Rs >= 368) {
return 8;
}
}
else if (negBalance >= 31) {
return 2;
}
}
}
}
else if (Rs >= 477) {
if (TSC < 3) {
if (Rs < 551) {
if (MVDiff < 5) {
if (Cs < 527) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 506) {
return 1;
}
else if (Rs >= 506) {
return 8;
}
}
}
else if (Cs >= 527) {
return 2;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (Rs >= 551) {
if (mu_mv_mag_sq < 99) {
if (tcor < 98) {
if (posBalance < 29) {
return 8;
}
else if (posBalance >= 29) {
return 1;
}
}
else if (tcor >= 98) {
if (posBalance < 41) {
if (Rs < 679) {
if (Rs < 644) {
return 8;
}
else if (Rs >= 644) {
return 2;
}
}
else if (Rs >= 679) {
return 1;
}
}
else if (posBalance >= 41) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (posBalance < 47) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (diffRsCsdiff < 1) {
return 4;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 654) {
return 1;
}
else if (Rs >= 654) {
return 4;
}
}
}
}
else if (posBalance >= 47) {
if (Rs < 642) {
if (Rs < 600) {
return 1;
}
else if (Rs >= 600) {
if (Rs < 612) {
return 8;
}
else if (Rs >= 612) {
return 4;
}
}
}
else if (Rs >= 642) {
return 2;
}
}
}
}
}
else if (TSC >= 3) {
if (TSC < 5) {
if (posBalance < 42) {
if (RsDiff < 3) {
if (MVDiff < 3) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
return 1;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (posBalance >= 42) {
if (Rs < 621) {
if (Rs < 583) {
if (Rs < 491) {
if (AFD < 4) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (AFD >= 4) {
return 1;
}
}
else if (Rs >= 491) {
return 8;
}
}
else if (Rs >= 583) {
if (Rs < 593) {
return 2;
}
else if (Rs >= 593) {
return 1;
}
}
}
else if (Rs >= 621) {
if (MVDiff < 1) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
if (Rs < 863) {
return 8;
}
else if (Rs >= 863) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 528) {
if (Rs < 1012) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
return 4;
}
}
else if (Rs >= 1012) {
return 4;
}
}
else if (Cs >= 528) {
return 4;
}
}
}
}
}
else if (TSC >= 5) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
}
}
}
}
else if (tcor >= 101) {
if (RsDiff < 1) {
if (Rs < 407) {
if (Cs < 115) {
if (Rs < 348) {
return 1;
}
else if (Rs >= 348) {
return 8;
}
}
else if (Cs >= 115) {
return 4;
}
}
else if (Rs >= 407) {
if (Rs < 823) {
return 8;
}
else if (Rs >= 823) {
return 4;
}
}
}
else if (RsDiff >= 1) {
if (ssDCval < 47) {
if (Rs < 667) {
return 1;
}
else if (Rs >= 667) {
return 8;
}
}
else if (ssDCval >= 47) {
if (Cs < 163) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
else if (Cs >= 163) {
if (AFD < 6) {
if (Rs < 974) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (Rs >= 974) {
return 8;
}
}
else if (AFD >= 6) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1294) {
if (SC < 1587) {
if (mu_mv_mag_sq < 59) {
if (ssDCval < 51) {
return 2;
}
else if (ssDCval >= 51) {
return 1;
}
}
else if (mu_mv_mag_sq >= 59) {
return 8;
}
}
else if (SC >= 1587) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1751) {
if (mu_mv_mag_sq < 62) {
if (negBalance < 29) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (AFD < 6) {
if (RsDiff < 3) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (negBalance >= 29) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 62) {
if (TSCindex < 2) {
if (Rs < 1444) {
if (Rs < 1375) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (Rs < 1354) {
return 8;
}
else if (Rs >= 1354) {
return 1;
}
}
}
else if (Rs >= 1375) {
if (Cs < 446) {
return 1;
}
else if (Cs >= 446) {
return 2;
}
}
}
else if (Rs >= 1444) {
return 1;
}
}
else if (TSCindex >= 2) {
if (tcor < 98) {
if (negBalance < 26) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (Cs < 309) {
return 2;
}
else if (Cs >= 309) {
return 8;
}
}
}
else if (negBalance >= 26) {
return 1;
}
}
else if (tcor >= 98) {
if (ssDCval < 63) {
if (Cs < 496) {
return 8;
}
else if (Cs >= 496) {
return 1;
}
}
else if (ssDCval >= 63) {
if (mu_mv_mag_sq < 85) {
return 8;
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 1449) {
if (SC < 1863) {
if (Rs < 1317) {
return 8;
}
else if (Rs >= 1317) {
return 1;
}
}
else if (SC >= 1863) {
if (Rs < 1409) {
return 8;
}
else if (Rs >= 1409) {
return 2;
}
}
}
else if (Rs >= 1449) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 1751) {
if (Cs < 434) {
if (refDCval < 68) {
return 2;
}
else if (refDCval >= 68) {
if (RsCsDiff < 10) {
return 2;
}
else if (RsCsDiff >= 10) {
return 1;
}
}
}
else if (Cs >= 434) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (Rs < 1407) {
return 8;
}
else if (Rs >= 1407) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (Cs < 142) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 76) {
if (refDCval < 82) {
if (Rs < 839) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 1;
}
}
else if (Rs >= 839) {
return 1;
}
}
else if (refDCval >= 82) {
if (diffRsCsdiff < -2) {
if (Cs < 64) {
return 8;
}
else if (Cs >= 64) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (refDCval < 80) {
if (Rs < 1222) {
return 8;
}
else if (Rs >= 1222) {
return 1;
}
}
else if (refDCval >= 80) {
if (Cs < 58) {
return 8;
}
else if (Cs >= 58) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
}
else if (RsDiff >= 2) {
if (tcor < 58) {
if (mu_mv_mag_sq < 21) {
return 8;
}
else if (mu_mv_mag_sq >= 21) {
return 2;
}
}
else if (tcor >= 58) {
if (posBalance < 7) {
return 8;
}
else if (posBalance >= 7) {
if (CsDiff < 2) {
if (SC < 364) {
if (Rs < 347) {
return 1;
}
else if (Rs >= 347) {
return 8;
}
}
else if (SC >= 364) {
return 1;
}
}
else if (CsDiff >= 2) {
if (Rs < 412) {
return 2;
}
else if (Rs >= 412) {
return 8;
}
}
}
}
}
}
else if (Cs >= 142) {
if (SC < 1916) {
if (refDCval < 81) {
if (AFD < 5) {
if (MVDiff < 1) {
if (ssDCval < 78) {
if (Rs < 1096) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Rs >= 1096) {
return 1;
}
}
else if (ssDCval >= 78) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (Rs < 721) {
return 1;
}
else if (Rs >= 721) {
return 8;
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 32) {
if (tcor < 98) {
if (Rs < 688) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (Rs < 604) {
return 8;
}
else if (Rs >= 604) {
if (Rs < 641) {
return 1;
}
else if (Rs >= 641) {
return 8;
}
}
}
}
else if (Rs >= 688) {
return 4;
}
}
else if (tcor >= 98) {
if (Rs < 961) {
return 8;
}
else if (Rs >= 961) {
if (Rs < 1278) {
return 1;
}
else if (Rs >= 1278) {
return 8;
}
}
}
}
else if (negBalance >= 32) {
if (mcTcor < 97) {
if (Rs < 649) {
return 8;
}
else if (Rs >= 649) {
return 4;
}
}
else if (mcTcor >= 97) {
return 8;
}
}
}
}
else if (AFD >= 5) {
if (tcor < 77) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (SC < 760) {
return 4;
}
else if (SC >= 760) {
return 2;
}
}
}
else if (tcor >= 77) {
if (mu_mv_mag_sq < 47) {
if (Cs < 255) {
if (mu_mv_mag_sq < 12) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 12) {
return 1;
}
}
else if (Cs >= 255) {
if (Cs < 527) {
return 1;
}
else if (Cs >= 527) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (gchDC < 4) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
}
}
else if (refDCval >= 81) {
if (mcTcor < 92) {
if (MVDiff < 23) {
if (negBalance < 63) {
if (mcTcor < 82) {
if (Rs < 316) {
return 8;
}
else if (Rs >= 316) {
return 2;
}
}
else if (mcTcor >= 82) {
if (Cs < 267) {
if (Cs < 255) {
return 8;
}
else if (Cs >= 255) {
return 1;
}
}
else if (Cs >= 267) {
return 8;
}
}
}
else if (negBalance >= 63) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (MVDiff >= 23) {
if (Cs < 205) {
return 1;
}
else if (Cs >= 205) {
if (tcor < 73) {
if (Rs < 315) {
return 1;
}
else if (Rs >= 315) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 4;
}
}
}
else if (tcor >= 73) {
if (Rs < 452) {
if (tcor < 79) {
if (diffMVdiffVal < 8) {
return 2;
}
else if (diffMVdiffVal >= 8) {
return 8;
}
}
else if (tcor >= 79) {
return 8;
}
}
else if (Rs >= 452) {
if (ssDCval < 96) {
if (Rs < 492) {
return 8;
}
else if (Rs >= 492) {
return 1;
}
}
else if (ssDCval >= 96) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (SC < 1081) {
if (Cs < 225) {
if (ssDCval < 93) {
if (SC < 509) {
if (SC < 488) {
return 4;
}
else if (SC >= 488) {
return 1;
}
}
else if (SC >= 509) {
if (mu_mv_mag_sq < 78) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (refDCval < 91) {
return 4;
}
else if (refDCval >= 91) {
if (Rs < 412) {
return 2;
}
else if (Rs >= 412) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (SC < 512) {
if (Rs < 311) {
return 2;
}
else if (Rs >= 311) {
return 1;
}
}
else if (SC >= 512) {
return 2;
}
}
}
}
else if (ssDCval >= 93) {
if (ssDCval < 110) {
if (CsDiff < 2) {
if (negBalance < 56) {
return 1;
}
else if (negBalance >= 56) {
return 4;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (ssDCval >= 110) {
if (MVDiff < 93) {
return 8;
}
else if (MVDiff >= 93) {
return 2;
}
}
}
}
else if (Cs >= 225) {
if (mu_mv_mag_sq < 63) {
if (MVDiff < 4) {
if (Cs < 311) {
if (Rs < 380) {
return 1;
}
else if (Rs >= 380) {
return 4;
}
}
else if (Cs >= 311) {
if (Cs < 473) {
if (posBalance < 50) {
if (Cs < 411) {
return 8;
}
else if (Cs >= 411) {
return 1;
}
}
else if (posBalance >= 50) {
return 1;
}
}
else if (Cs >= 473) {
if (Rs < 487) {
return 1;
}
else if (Rs >= 487) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (Rs < 518) {
return 2;
}
else if (Rs >= 518) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (mcTcor < 99) {
if (ssDCval < 116) {
if (negBalance < 48) {
if (mu_mv_mag_sq < 52) {
return 1;
}
else if (mu_mv_mag_sq >= 52) {
return 4;
}
}
else if (negBalance >= 48) {
if (diffAFD < 1) {
if (posBalance < 49) {
if (Cs < 413) {
return 4;
}
else if (Cs >= 413) {
return 8;
}
}
else if (posBalance >= 49) {
if (MVDiff < 21) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (MVDiff >= 21) {
if (refDCval < 86) {
return 1;
}
else if (refDCval >= 86) {
if (Cs < 319) {
if (Rs < 397) {
return 1;
}
else if (Rs >= 397) {
return 2;
}
}
else if (Cs >= 319) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
}
}
}
}
else if (diffAFD >= 1) {
if (posBalance < 58) {
if (mu_mv_mag_sq < 19) {
return 2;
}
else if (mu_mv_mag_sq >= 19) {
return 8;
}
}
else if (posBalance >= 58) {
if (Rs < 378) {
return 1;
}
else if (Rs >= 378) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 116) {
if (refDCval < 130) {
if (tcor < 79) {
return 4;
}
else if (tcor >= 79) {
if (Rs < 480) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (Rs >= 480) {
if (SC < 944) {
if (Rs < 503) {
if (Rs < 487) {
return 8;
}
else if (Rs >= 487) {
return 4;
}
}
else if (Rs >= 503) {
return 1;
}
}
else if (SC >= 944) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
}
else if (refDCval >= 130) {
if (RsCsDiff < 7) {
if (Cs < 305) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 4;
}
}
else if (Cs >= 305) {
if (TSC < 4) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
else if (TSC >= 4) {
return 4;
}
}
}
else if (RsCsDiff >= 7) {
if (Rs < 326) {
return 1;
}
else if (Rs >= 326) {
return 2;
}
}
}
}
}
else if (mcTcor >= 99) {
if (MVDiff < 51) {
if (TSC < 4) {
if (gchDC < 1) {
if (negBalance < 33) {
return 1;
}
else if (negBalance >= 33) {
return 2;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (TSC >= 4) {
if (Rs < 426) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (Rs >= 426) {
return 1;
}
}
}
else if (MVDiff >= 51) {
if (Rs < 438) {
return 4;
}
else if (Rs >= 438) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (ssDCval < 96) {
if (refDCval < 94) {
if (negBalance < 39) {
if (Cs < 238) {
return 2;
}
else if (Cs >= 238) {
if (MVDiff < 18) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
else if (MVDiff >= 18) {
return 8;
}
}
}
else if (negBalance >= 39) {
if (Cs < 278) {
return 2;
}
else if (Cs >= 278) {
if (SC < 1015) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (SC >= 1015) {
if (Cs < 512) {
return 1;
}
else if (Cs >= 512) {
return 8;
}
}
}
}
}
else if (refDCval >= 94) {
return 8;
}
}
else if (ssDCval >= 96) {
if (Rs < 592) {
if (ssDCval < 160) {
if (negBalance < 22) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (Rs < 386) {
return 1;
}
else if (Rs >= 386) {
return 8;
}
}
}
else if (negBalance >= 22) {
if (ssDCval < 116) {
if (Rs < 524) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (Rs >= 524) {
if (Rs < 581) {
return 8;
}
else if (Rs >= 581) {
if (Rs < 588) {
return 4;
}
else if (Rs >= 588) {
return 8;
}
}
}
}
else if (ssDCval >= 116) {
if (ssDCval < 123) {
if (Rs < 424) {
return 1;
}
else if (Rs >= 424) {
return 2;
}
}
else if (ssDCval >= 123) {
if (Rs < 323) {
if (Cs < 392) {
return 8;
}
else if (Cs >= 392) {
return 2;
}
}
else if (Rs >= 323) {
return 8;
}
}
}
}
}
else if (ssDCval >= 160) {
return 4;
}
}
else if (Rs >= 592) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
}
}
}
else if (SC >= 1081) {
if (refDCval < 86) {
if (mu_mv_mag_sq < 120) {
if (tcor < 88) {
if (Rs < 852) {
return 2;
}
else if (Rs >= 852) {
return 8;
}
}
else if (tcor >= 88) {
if (MVDiff < 8) {
if (SC < 1910) {
if (posBalance < 23) {
return 8;
}
else if (posBalance >= 23) {
return 1;
}
}
else if (SC >= 1910) {
return 8;
}
}
else if (MVDiff >= 8) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 1034) {
return 1;
}
else if (Rs >= 1034) {
return 2;
}
}
}
else if (refDCval >= 86) {
if (MVDiff < 3) {
if (diffRsCsdiff < -4) {
return 2;
}
else if (diffRsCsdiff >= -4) {
if (SC < 1090) {
if (Rs < 617) {
return 8;
}
else if (Rs >= 617) {
if (Rs < 803) {
return 4;
}
else if (Rs >= 803) {
return 8;
}
}
}
else if (SC >= 1090) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 136) {
if (ssDCval < 124) {
if (RsDiff < 2) {
if (ssDCval < 115) {
return 8;
}
else if (ssDCval >= 115) {
if (Rs < 577) {
return 8;
}
else if (Rs >= 577) {
return 2;
}
}
}
else if (RsDiff >= 2) {
if (mcTcor < 98) {
if (AFD < 12) {
if (Cs < 292) {
return 1;
}
else if (Cs >= 292) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (Rs < 905) {
return 1;
}
else if (Rs >= 905) {
return 8;
}
}
}
}
else if (AFD >= 12) {
if (Rs < 1465) {
return 1;
}
else if (Rs >= 1465) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 1388) {
return 1;
}
else if (Rs >= 1388) {
return 8;
}
}
}
}
else if (ssDCval >= 124) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
}
}
else if (ssDCval >= 136) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
}
}
}
}
else if (SC >= 1916) {
if (Cs < 297) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
if (Rs < 1969) {
return 8;
}
else if (Rs >= 1969) {
return 4;
}
}
}
else if (Cs >= 297) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 82) {
if (tcor < 98) {
if (negBalance < 30) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
else if (negBalance >= 30) {
return 1;
}
}
else if (tcor >= 98) {
if (Cs < 306) {
return 1;
}
else if (Cs >= 306) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (SC < 2520) {
if (negBalance < 20) {
if (Rs < 1808) {
return 8;
}
else if (Rs >= 1808) {
return 1;
}
}
else if (negBalance >= 20) {
if (Cs < 497) {
if (ssDCval < 128) {
return 1;
}
else if (ssDCval >= 128) {
return 4;
}
}
else if (Cs >= 497) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
return 1;
}
}
}
}
else if (SC >= 2520) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
if (Cs < 312) {
return 8;
}
else if (Cs >= 312) {
if (Cs < 483) {
if (Rs < 2405) {
return 2;
}
else if (Rs >= 2405) {
return 1;
}
}
else if (Cs >= 483) {
return 1;
}
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 73) {
if (Cs < 243) {
if (mu_mv_mag_sq < 53) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (posBalance < 50) {
if (ssDCval < 180) {
return 1;
}
else if (ssDCval >= 180) {
return 2;
}
}
else if (posBalance >= 50) {
if (negBalance < 43) {
if (diffRsCsdiff < -35) {
if (AFD < 24) {
if (Rs < 1513) {
return 1;
}
else if (Rs >= 1513) {
return 4;
}
}
else if (AFD >= 24) {
return 2;
}
}
else if (diffRsCsdiff >= -35) {
return 1;
}
}
else if (negBalance >= 43) {
if (refDCval < 99) {
if (refDCval < 62) {
if (AFD < 18) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (refDCval >= 62) {
if (AFD < 30) {
if (refDCval < 93) {
if (AFD < 21) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
else if (AFD >= 21) {
return 2;
}
}
else if (refDCval >= 93) {
if (Rs < 490) {
return 8;
}
else if (Rs >= 490) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 651) {
return 2;
}
else if (Rs >= 651) {
return 8;
}
}
}
}
}
else if (AFD >= 30) {
if (ssDCval < 74) {
if (Cs < 188) {
return 2;
}
else if (Cs >= 188) {
return 4;
}
}
else if (ssDCval >= 74) {
return 1;
}
}
}
}
else if (refDCval >= 99) {
if (Cs < 127) {
if (diffRsCsdiff < -8) {
if (Rs < 1071) {
return 4;
}
else if (Rs >= 1071) {
return 8;
}
}
else if (diffRsCsdiff >= -8) {
return 1;
}
}
else if (Cs >= 127) {
if (MVDiff < 158) {
return 1;
}
else if (MVDiff >= 158) {
if (Rs < 884) {
return 2;
}
else if (Rs >= 884) {
return 1;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (tcor < 45) {
if (AFD < 21) {
if (Rs < 1079) {
return 4;
}
else if (Rs >= 1079) {
return 1;
}
}
else if (AFD >= 21) {
if (diffRsCsdiff < -8) {
if (Rs < 727) {
return 8;
}
else if (Rs >= 727) {
if (Rs < 1140) {
return 1;
}
else if (Rs >= 1140) {
return 2;
}
}
}
else if (diffRsCsdiff >= -8) {
return 2;
}
}
}
else if (tcor >= 45) {
if (posBalance < 56) {
if (diffRsCsdiff < 58) {
if (Cs < 45) {
return 1;
}
else if (Cs >= 45) {
return 2;
}
}
else if (diffRsCsdiff >= 58) {
return 1;
}
}
else if (posBalance >= 56) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 2;
}
}
}
}
}
else if (Cs >= 243) {
if (negBalance < 46) {
if (RsDiff < 12) {
if (Rs < 324) {
if (CsDiff < 5) {
if (Rs < 311) {
return 1;
}
else if (Rs >= 311) {
return 8;
}
}
else if (CsDiff >= 5) {
if (Rs < 314) {
return 2;
}
else if (Rs >= 314) {
return 4;
}
}
}
else if (Rs >= 324) {
if (mu_mv_mag_sq < 69) {
if (posBalance < 54) {
if (diffRsCsdiff < -36) {
if (Rs < 1612) {
return 1;
}
else if (Rs >= 1612) {
return 2;
}
}
else if (diffRsCsdiff >= -36) {
if (negBalance < 35) {
if (tcor < 69) {
return 1;
}
else if (tcor >= 69) {
return 2;
}
}
else if (negBalance >= 35) {
return 1;
}
}
}
else if (posBalance >= 54) {
if (MVDiff < 42) {
if (mcTcor < 86) {
if (Cs < 435) {
return 8;
}
else if (Cs >= 435) {
if (Rs < 788) {
return 1;
}
else if (Rs >= 788) {
return 2;
}
}
}
else if (mcTcor >= 86) {
if (RsCsDiff < 22) {
return 2;
}
else if (RsCsDiff >= 22) {
return 1;
}
}
}
else if (MVDiff >= 42) {
if (negBalance < 32) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (negBalance >= 32) {
if (Cs < 351) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
return 2;
}
}
}
else if (Cs >= 351) {
if (diffRsCsdiff < -76) {
return 8;
}
else if (diffRsCsdiff >= -76) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (Rs < 592) {
return 1;
}
else if (Rs >= 592) {
return 8;
}
}
}
}
else if (RsDiff >= 12) {
if (MVDiff < 42) {
return 2;
}
else if (MVDiff >= 42) {
if (mcTcor < 89) {
if (MVDiff < 59) {
if (Rs < 1670) {
return 2;
}
else if (Rs >= 1670) {
return 1;
}
}
else if (MVDiff >= 59) {
return 2;
}
}
else if (mcTcor >= 89) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 1;
}
}
}
}
}
else if (negBalance >= 46) {
if (MVDiff < 31) {
if (RsDiff < 7) {
if (mu_mv_mag_sq < 7) {
if (diffTSC < 2) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
else if (diffTSC >= 2) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
if (RsDiff < 5) {
if (mcTcor < 89) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 2;
}
}
else if (mcTcor >= 89) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (RsDiff >= 5) {
if (ssDCval < 124) {
if (posBalance < 65) {
if (mu_mv_mag_sq < 34) {
if (mu_mv_mag_sq < 12) {
if (refDCval < 97) {
return 8;
}
else if (refDCval >= 97) {
if (Rs < 908) {
return 8;
}
else if (Rs >= 908) {
if (Cs < 513) {
return 2;
}
else if (Cs >= 513) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 34) {
if (Rs < 735) {
return 1;
}
else if (Rs >= 735) {
return 4;
}
}
}
else if (posBalance >= 65) {
if (Rs < 810) {
return 4;
}
else if (Rs >= 810) {
return 1;
}
}
}
else if (ssDCval >= 124) {
return 2;
}
}
}
}
else if (RsDiff >= 7) {
if (Rs < 1094) {
if (Rs < 815) {
return 8;
}
else if (Rs >= 815) {
return 4;
}
}
else if (Rs >= 1094) {
if (Cs < 412) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (CsDiff < 2) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 8;
}
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (Cs >= 412) {
if (SC < 1743) {
if (AFD < 29) {
return 4;
}
else if (AFD >= 29) {
return 2;
}
}
else if (SC >= 1743) {
if (diffRsCsdiff < -11) {
return 4;
}
else if (diffRsCsdiff >= -11) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 31) {
if (AFD < 23) {
if (SC < 2354) {
if (posBalance < 49) {
if (AFD < 15) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (Rs < 891) {
return 2;
}
else if (Rs >= 891) {
return 1;
}
}
}
else if (AFD >= 15) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (RsDiff < 6) {
if (TSC < 6) {
if (MVDiff < 93) {
return 4;
}
else if (MVDiff >= 93) {
return 8;
}
}
else if (TSC >= 6) {
if (ssDCval < 59) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (ssDCval < 56) {
return 2;
}
else if (ssDCval >= 56) {
return 8;
}
}
}
else if (ssDCval >= 59) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (MVDiff < 74) {
return 4;
}
else if (MVDiff >= 74) {
if (Rs < 457) {
return 4;
}
else if (Rs >= 457) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 6) {
if (Cs < 363) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
if (mu_mv_mag_sq < 58) {
if (RsDiff < 16) {
if (posBalance < 66) {
return 2;
}
else if (posBalance >= 66) {
return 1;
}
}
else if (RsDiff >= 16) {
return 1;
}
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 681) {
return 8;
}
else if (Rs >= 681) {
return 4;
}
}
}
}
else if (Cs >= 363) {
if (Rs < 1000) {
if (Rs < 963) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (Rs >= 963) {
return 1;
}
}
else if (Rs >= 1000) {
if (mcTcor < 82) {
if (Rs < 1799) {
return 4;
}
else if (Rs >= 1799) {
return 8;
}
}
else if (mcTcor >= 82) {
if (Cs < 520) {
if (Rs < 1286) {
return 8;
}
else if (Rs >= 1286) {
return 2;
}
}
else if (Cs >= 520) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 2354) {
if (AFD < 17) {
if (Rs < 1956) {
return 4;
}
else if (Rs >= 1956) {
return 1;
}
}
else if (AFD >= 17) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
else if (AFD >= 23) {
if (refDCval < 47) {
return 8;
}
else if (refDCval >= 47) {
if (gchDC < 9) {
if (diffAFD < 6) {
if (CsDiff < 5) {
if (RsCsDiff < 94) {
if (MVDiff < 117) {
if (RsCsDiff < 62) {
if (negBalance < 61) {
return 1;
}
else if (negBalance >= 61) {
return 2;
}
}
else if (RsCsDiff >= 62) {
if (AFD < 25) {
if (Rs < 966) {
return 1;
}
else if (Rs >= 966) {
return 8;
}
}
else if (AFD >= 25) {
return 2;
}
}
}
else if (MVDiff >= 117) {
if (TSC < 9) {
if (MVDiff < 165) {
return 1;
}
else if (MVDiff >= 165) {
return 8;
}
}
else if (TSC >= 9) {
if (Rs < 535) {
return 8;
}
else if (Rs >= 535) {
if (Rs < 1445) {
return 4;
}
else if (Rs >= 1445) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 94) {
if (diffTSC < 2) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (diffTSC >= 2) {
if (ssDCval < 122) {
return 4;
}
else if (ssDCval >= 122) {
return 1;
}
}
}
}
else if (CsDiff >= 5) {
if (diffRsCsdiff < -21) {
return 2;
}
else if (diffRsCsdiff >= -21) {
if (posBalance < 58) {
if (ssDCval < 116) {
if (Rs < 329) {
return 2;
}
else if (Rs >= 329) {
return 4;
}
}
else if (ssDCval >= 116) {
return 2;
}
}
else if (posBalance >= 58) {
return 1;
}
}
}
}
else if (diffAFD >= 6) {
if (ssDCval < 69) {
if (Rs < 413) {
return 1;
}
else if (Rs >= 413) {
return 8;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
else if (gchDC >= 9) {
if (ssDCval < 110) {
if (refDCval < 74) {
return 2;
}
else if (refDCval >= 74) {
if (Cs < 262) {
return 1;
}
else if (Cs >= 262) {
return 8;
}
}
}
else if (ssDCval >= 110) {
if (mu_mv_mag_sq < 11) {
if (Rs < 766) {
if (Rs < 643) {
return 1;
}
else if (Rs >= 643) {
return 4;
}
}
else if (Rs >= 766) {
return 8;
}
}
else if (mu_mv_mag_sq >= 11) {
return 1;
}
}
}
}
}
}
}
}
}
else if (tcor >= 73) {
if (mu_mv_mag_sq < 78) {
if (RsCsDiff < 90) {
if (refDCval < 139) {
if (Cs < 290) {
if (refDCval < 78) {
if (diffAFD < -10) {
if (Rs < 745) {
return 8;
}
else if (Rs >= 745) {
return 1;
}
}
else if (diffAFD >= -10) {
if (mcTcor < 101) {
if (Cs < 204) {
return 1;
}
else if (Cs >= 204) {
if (mcTcor < 96) {
if (Rs < 1870) {
if (ssDCval < 76) {
if (Rs < 568) {
if (Rs < 483) {
return 1;
}
else if (Rs >= 483) {
return 4;
}
}
else if (Rs >= 568) {
return 1;
}
}
else if (ssDCval >= 76) {
return 4;
}
}
else if (Rs >= 1870) {
if (Rs < 1980) {
return 2;
}
else if (Rs >= 1980) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 30) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
else if (mu_mv_mag_sq >= 30) {
if (Cs < 258) {
return 2;
}
else if (Cs >= 258) {
return 8;
}
}
}
}
}
else if (mcTcor >= 101) {
if (Rs < 875) {
return 8;
}
else if (Rs >= 875) {
return 2;
}
}
}
}
else if (refDCval >= 78) {
if (MVDiff < 18) {
if (Rs < 1114) {
return 1;
}
else if (Rs >= 1114) {
if (posBalance < 55) {
if (diffAFD < -1) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (diffAFD >= -1) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 121) {
return 2;
}
else if (ssDCval >= 121) {
return 1;
}
}
}
}
else if (MVDiff >= 18) {
if (SC < 841) {
if (ssDCval < 120) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (MVDiff < 69) {
return 1;
}
else if (MVDiff >= 69) {
if (Rs < 345) {
return 1;
}
else if (Rs >= 345) {
return 8;
}
}
}
}
else if (ssDCval >= 120) {
if (MVDiff < 148) {
return 8;
}
else if (MVDiff >= 148) {
return 2;
}
}
}
else if (SC >= 841) {
if (diffRsCsdiff < -14) {
if (ssDCval < 95) {
if (diffRsCsdiff < -30) {
return 2;
}
else if (diffRsCsdiff >= -30) {
if (posBalance < 61) {
if (Rs < 1904) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 8;
}
}
else if (Rs >= 1904) {
return 2;
}
}
else if (posBalance >= 61) {
return 2;
}
}
}
else if (ssDCval >= 95) {
if (mcTcor < 98) {
if (TSC < 7) {
if (Cs < 229) {
return 2;
}
else if (Cs >= 229) {
return 1;
}
}
else if (TSC >= 7) {
if (SC < 2268) {
return 1;
}
else if (SC >= 2268) {
return 8;
}
}
}
else if (mcTcor >= 98) {
return 1;
}
}
}
else if (diffRsCsdiff >= -14) {
if (diffRsCsdiff < 39) {
if (MVDiff < 34) {
if (refDCval < 99) {
if (MVDiff < 32) {
return 1;
}
else if (MVDiff >= 32) {
if (Rs < 940) {
return 4;
}
else if (Rs >= 940) {
return 1;
}
}
}
else if (refDCval >= 99) {
if (Cs < 138) {
return 2;
}
else if (Cs >= 138) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
if (Rs < 1711) {
return 2;
}
else if (Rs >= 1711) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 34) {
return 1;
}
}
else if (diffRsCsdiff >= 39) {
return 2;
}
}
}
}
}
}
else if (Cs >= 290) {
if (Cs < 428) {
if (tcor < 96) {
if (Rs < 647) {
if (mu_mv_mag_sq < 10) {
if (Rs < 576) {
if (Rs < 549) {
return 1;
}
else if (Rs >= 549) {
return 2;
}
}
else if (Rs >= 576) {
return 4;
}
}
else if (mu_mv_mag_sq >= 10) {
if (SC < 788) {
if (Rs < 337) {
if (Rs < 321) {
return 2;
}
else if (Rs >= 321) {
return 8;
}
}
else if (Rs >= 337) {
return 1;
}
}
else if (SC >= 788) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 506) {
return 4;
}
else if (Rs >= 506) {
if (Rs < 570) {
return 8;
}
else if (Rs >= 570) {
return 1;
}
}
}
}
}
}
else if (Rs >= 647) {
if (negBalance < 54) {
if (ssDCval < 96) {
if (SC < 2667) {
if (posBalance < 65) {
if (Cs < 413) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
if (Rs < 1436) {
return 2;
}
else if (Rs >= 1436) {
return 1;
}
}
}
else if (Cs >= 413) {
if (Rs < 796) {
return 8;
}
else if (Rs >= 796) {
if (negBalance < 30) {
if (Rs < 2074) {
return 1;
}
else if (Rs >= 2074) {
return 4;
}
}
else if (negBalance >= 30) {
return 1;
}
}
}
}
else if (posBalance >= 65) {
if (Rs < 1503) {
if (Rs < 843) {
return 4;
}
else if (Rs >= 843) {
return 1;
}
}
else if (Rs >= 1503) {
return 2;
}
}
}
else if (SC >= 2667) {
if (Cs < 388) {
return 1;
}
else if (Cs >= 388) {
return 4;
}
}
}
else if (ssDCval >= 96) {
if (TSC < 5) {
if (Rs < 2221) {
return 8;
}
else if (Rs >= 2221) {
return 2;
}
}
else if (TSC >= 5) {
if (ssDCval < 115) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (ssDCval >= 115) {
return 1;
}
}
}
}
else if (negBalance >= 54) {
if (mcTcor < 96) {
if (ssDCval < 62) {
if (Rs < 1629) {
return 1;
}
else if (Rs >= 1629) {
return 2;
}
}
else if (ssDCval >= 62) {
if (Cs < 397) {
return 8;
}
else if (Cs >= 397) {
if (Rs < 1961) {
return 2;
}
else if (Rs >= 1961) {
if (Rs < 2354) {
return 8;
}
else if (Rs >= 2354) {
return 1;
}
}
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
}
}
else if (tcor >= 96) {
if (refDCval < 107) {
if (posBalance < 56) {
if (Rs < 1308) {
if (Rs < 1055) {
return 8;
}
else if (Rs >= 1055) {
return 4;
}
}
else if (Rs >= 1308) {
return 1;
}
}
else if (posBalance >= 56) {
return 4;
}
}
else if (refDCval >= 107) {
return 1;
}
}
}
else if (Cs >= 428) {
if (mcTcor < 90) {
if (posBalance < 57) {
if (refDCval < 50) {
if (posBalance < 36) {
return 8;
}
else if (posBalance >= 36) {
return 2;
}
}
else if (refDCval >= 50) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 1;
}
}
}
else if (posBalance >= 57) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (refDCval < 65) {
if (negBalance < 31) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
return 1;
}
}
}
else if (negBalance >= 31) {
if (negBalance < 47) {
if (ssDCval < 59) {
if (AFD < 12) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 4;
}
}
else if (AFD >= 12) {
if (mcTcor < 97) {
return 2;
}
else if (mcTcor >= 97) {
return 1;
}
}
}
else if (ssDCval >= 59) {
if (Rs < 1493) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
return 2;
}
}
else if (Rs >= 1493) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (ssDCval < 62) {
if (AFD < 12) {
if (Rs < 1199) {
return 8;
}
else if (Rs >= 1199) {
return 1;
}
}
else if (AFD >= 12) {
return 1;
}
}
else if (ssDCval >= 62) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
return 4;
}
}
}
}
}
else if (refDCval >= 65) {
if (ssDCval < 123) {
if (CsDiff < 4) {
if (TSC < 8) {
if (tcor < 97) {
if (RsDiff < 6) {
if (MVDiff < 3) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (RsDiff >= 6) {
if (posBalance < 54) {
if (Rs < 1189) {
return 2;
}
else if (Rs >= 1189) {
return 1;
}
}
else if (posBalance >= 54) {
if (ssDCval < 102) {
if (Rs < 1562) {
return 1;
}
else if (Rs >= 1562) {
return 2;
}
}
else if (ssDCval >= 102) {
return 1;
}
}
}
}
else if (tcor >= 97) {
if (diffAFD < 1) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 1;
}
}
}
else if (diffAFD >= 1) {
return 4;
}
}
}
else if (TSC >= 8) {
if (RsDiff < 6) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 1268) {
return 4;
}
else if (Rs >= 1268) {
return 1;
}
}
}
else if (RsDiff >= 6) {
if (SC < 2859) {
return 1;
}
else if (SC >= 2859) {
return 8;
}
}
}
}
else if (CsDiff >= 4) {
if (RsCsDiff < 57) {
if (ssDCval < 92) {
if (RsCsDiff < 37) {
if (Rs < 1260) {
return 4;
}
else if (Rs >= 1260) {
return 1;
}
}
else if (RsCsDiff >= 37) {
return 1;
}
}
else if (ssDCval >= 92) {
return 8;
}
}
else if (RsCsDiff >= 57) {
if (RsDiff < 8) {
if (ssDCval < 77) {
return 1;
}
else if (ssDCval >= 77) {
return 2;
}
}
else if (RsDiff >= 8) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
return 2;
}
}
}
}
}
else if (ssDCval >= 123) {
if (TSC < 7) {
if (Cs < 487) {
if (AFD < 10) {
if (Rs < 2003) {
return 2;
}
else if (Rs >= 2003) {
return 1;
}
}
else if (AFD >= 10) {
if (diffRsCsdiff < 19) {
return 1;
}
else if (diffRsCsdiff >= 19) {
return 2;
}
}
}
else if (Cs >= 487) {
if (ssDCval < 133) {
return 2;
}
else if (ssDCval >= 133) {
return 8;
}
}
}
else if (TSC >= 7) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 139) {
if (ssDCval < 146) {
if (gchDC < 2) {
if (mcTcor < 98) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 1;
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 35) {
return 1;
}
else if (mu_mv_mag_sq >= 35) {
if (Rs < 1243) {
return 2;
}
else if (Rs >= 1243) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
if (gchDC < 12) {
return 1;
}
else if (gchDC >= 12) {
return 2;
}
}
}
}
else if (ssDCval >= 146) {
if (ssDCval < 177) {
if (mcTcor < 93) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (Rs < 597) {
return 1;
}
else if (Rs >= 597) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 828) {
if (AFD < 16) {
if (Rs < 628) {
return 2;
}
else if (Rs >= 628) {
return 8;
}
}
else if (AFD >= 16) {
if (MVDiff < 162) {
return 1;
}
else if (MVDiff >= 162) {
return 8;
}
}
}
else if (Rs >= 828) {
if (Rs < 2347) {
if (tcor < 78) {
if (Rs < 1490) {
return 1;
}
else if (Rs >= 1490) {
return 8;
}
}
else if (tcor >= 78) {
return 1;
}
}
else if (Rs >= 2347) {
return 2;
}
}
}
}
else if (ssDCval >= 177) {
if (refDCval < 179) {
return 8;
}
else if (refDCval >= 179) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 90) {
if (refDCval < 111) {
if (MVDiff < 45) {
if (AFD < 12) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
}
}
}
else if (AFD >= 12) {
if (ssDCval < 70) {
if (Cs < 100) {
return 1;
}
else if (Cs >= 100) {
return 2;
}
}
else if (ssDCval >= 70) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
if (Cs < 266) {
if (ssDCval < 103) {
return 2;
}
else if (ssDCval >= 103) {
return 1;
}
}
else if (Cs >= 266) {
if (diffMVdiffVal < -11) {
return 1;
}
else if (diffMVdiffVal >= -11) {
if (posBalance < 67) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
if (tcor < 79) {
if (negBalance < 51) {
if (Rs < 1639) {
return 2;
}
else if (Rs >= 1639) {
return 1;
}
}
else if (negBalance >= 51) {
return 2;
}
}
else if (tcor >= 79) {
if (mcTcor < 90) {
return 1;
}
else if (mcTcor >= 90) {
return 2;
}
}
}
}
else if (posBalance >= 67) {
if (posBalance < 72) {
return 1;
}
else if (posBalance >= 72) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 45) {
if (mcTcor < 96) {
if (gchDC < 2) {
if (SC < 2439) {
return 1;
}
else if (SC >= 2439) {
return 4;
}
}
else if (gchDC >= 2) {
if (RsDiff < 15) {
if (gchDC < 16) {
if (ssDCval < 97) {
return 1;
}
else if (ssDCval >= 97) {
if (Rs < 1870) {
return 1;
}
else if (Rs >= 1870) {
return 8;
}
}
}
else if (gchDC >= 16) {
return 8;
}
}
else if (RsDiff >= 15) {
return 2;
}
}
}
else if (mcTcor >= 96) {
if (MVDiff < 68) {
if (ssDCval < 96) {
if (MVDiff < 52) {
return 1;
}
else if (MVDiff >= 52) {
return 2;
}
}
else if (ssDCval >= 96) {
return 1;
}
}
else if (MVDiff >= 68) {
if (Rs < 362) {
return 1;
}
else if (Rs >= 362) {
return 2;
}
}
}
}
}
else if (refDCval >= 111) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (refDCval < 122) {
if (MVDiff < 58) {
if (mu_mv_mag_sq < 31) {
return 1;
}
else if (mu_mv_mag_sq >= 31) {
if (Cs < 370) {
return 1;
}
else if (Cs >= 370) {
return 2;
}
}
}
else if (MVDiff >= 58) {
if (Rs < 984) {
return 8;
}
else if (Rs >= 984) {
if (mcTcor < 96) {
if (refDCval < 120) {
return 8;
}
else if (refDCval >= 120) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
if (Rs < 2333) {
return 8;
}
else if (Rs >= 2333) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (CsDiff < 3) {
if (Rs < 1960) {
return 1;
}
else if (Rs >= 1960) {
return 8;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
}
}
}
else if (refDCval >= 122) {
if (MVDiff < 136) {
if (TSC < 7) {
if (mcTcor < 97) {
if (ssDCval < 134) {
return 2;
}
else if (ssDCval >= 134) {
return 1;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (TSC >= 7) {
if (mcTcor < 94) {
if (Cs < 487) {
return 1;
}
else if (Cs >= 487) {
return 8;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
}
else if (MVDiff >= 136) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (Cs < 464) {
return 1;
}
else if (Cs >= 464) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (negBalance < 51) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
if (SC < 820) {
return 1;
}
else if (SC >= 820) {
if (Cs < 408) {
if (Rs < 648) {
return 8;
}
else if (Rs >= 648) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (Cs >= 408) {
if (ssDCval < 64) {
if (posBalance < 42) {
if (Rs < 690) {
return 8;
}
else if (Rs >= 690) {
return 4;
}
}
else if (posBalance >= 42) {
return 2;
}
}
else if (ssDCval >= 64) {
if (Cs < 524) {
return 8;
}
else if (Cs >= 524) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 51) {
if (posBalance < 25) {
return 1;
}
else if (posBalance >= 25) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Cs >= 533) {
if (negBalance < 14) {
if (Rs < 702) {
if (ssDCval < 58) {
if (ssDCval < 29) {
return 8;
}
else if (ssDCval >= 29) {
if (Cs < 905) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 437) {
return 1;
}
else if (Rs >= 437) {
return 8;
}
}
}
else if (Cs >= 905) {
if (Cs < 1019) {
if (Rs < 685) {
return 2;
}
else if (Rs >= 685) {
return 8;
}
}
else if (Cs >= 1019) {
return 1;
}
}
}
}
else if (ssDCval >= 58) {
if (Rs < 652) {
if (Rs < 515) {
if (SC < 1431) {
if (posBalance < 50) {
if (Rs < 445) {
return 4;
}
else if (Rs >= 445) {
if (Rs < 473) {
return 8;
}
else if (Rs >= 473) {
return 2;
}
}
}
else if (posBalance >= 50) {
if (Cs < 636) {
return 1;
}
else if (Cs >= 636) {
if (Rs < 504) {
return 8;
}
else if (Rs >= 504) {
return 4;
}
}
}
}
else if (SC >= 1431) {
return 8;
}
}
else if (Rs >= 515) {
if (ssDCval < 90) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
return 2;
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (Rs >= 652) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (Cs < 679) {
if (Rs < 694) {
return 8;
}
else if (Rs >= 694) {
return 1;
}
}
else if (Cs >= 679) {
return 1;
}
}
}
}
}
else if (Rs >= 702) {
if (diffAFD < -3) {
if (Rs < 1580) {
if (mcTcor < 99) {
if (diffAFD < -20) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (diffAFD >= -20) {
if (mcTcor < 79) {
if (Rs < 1225) {
return 1;
}
else if (Rs >= 1225) {
return 8;
}
}
else if (mcTcor >= 79) {
if (diffRsCsdiff < 4) {
if (diffAFD < -16) {
if (Cs < 821) {
return 4;
}
else if (Cs >= 821) {
return 8;
}
}
else if (diffAFD >= -16) {
return 8;
}
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
}
}
else if (mcTcor >= 99) {
if (diffRsCsdiff < -38) {
if (diffMVdiffVal < 34) {
if (Rs < 1297) {
return 8;
}
else if (Rs >= 1297) {
if (diffAFD < -18) {
return 4;
}
else if (diffAFD >= -18) {
if (Rs < 1365) {
if (Cs < 1043) {
return 4;
}
else if (Cs >= 1043) {
return 1;
}
}
else if (Rs >= 1365) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 34) {
return 1;
}
}
else if (diffRsCsdiff >= -38) {
if (ssDCval < 73) {
if (Rs < 1244) {
return 8;
}
else if (Rs >= 1244) {
return 1;
}
}
else if (ssDCval >= 73) {
if (Cs < 729) {
if (Rs < 1073) {
return 2;
}
else if (Rs >= 1073) {
return 1;
}
}
else if (Cs >= 729) {
if (MVDiff < 37) {
if (MVDiff < 6) {
if (Rs < 1220) {
return 1;
}
else if (Rs >= 1220) {
return 8;
}
}
else if (MVDiff >= 6) {
if (Rs < 929) {
return 1;
}
else if (Rs >= 929) {
return 4;
}
}
}
else if (MVDiff >= 37) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1580) {
if (diffRsCsdiff < -161) {
if (diffAFD < -20) {
return 4;
}
else if (diffAFD >= -20) {
return 1;
}
}
else if (diffRsCsdiff >= -161) {
if (SC < 2474) {
return 4;
}
else if (SC >= 2474) {
return 8;
}
}
}
}
else if (diffAFD >= -3) {
if (Rs < 1810) {
if (Rs < 967) {
if (posBalance < 12) {
if (ssDCval < 152) {
if (Cs < 1112) {
return 8;
}
else if (Cs >= 1112) {
return 2;
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (posBalance >= 12) {
if (Cs < 1141) {
if (ssDCval < 67) {
if (ssDCval < 31) {
return 8;
}
else if (ssDCval >= 31) {
if (Rs < 882) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Rs >= 882) {
if (Cs < 840) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
return 2;
}
}
else if (Cs >= 840) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (CsDiff < 3) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (posBalance < 29) {
if (Rs < 775) {
if (SC < 1368) {
if (SC < 1328) {
return 1;
}
else if (SC >= 1328) {
return 8;
}
}
else if (SC >= 1368) {
return 1;
}
}
else if (Rs >= 775) {
return 8;
}
}
else if (posBalance >= 29) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 794) {
return 2;
}
else if (Rs >= 794) {
return 1;
}
}
}
}
else if (Cs >= 1141) {
if (Rs < 805) {
return 2;
}
else if (Rs >= 805) {
if (Rs < 864) {
return 1;
}
else if (Rs >= 864) {
return 8;
}
}
}
}
}
else if (Rs >= 967) {
if (SC < 1801) {
return 8;
}
else if (SC >= 1801) {
if (negBalance < 9) {
if (Cs < 1193) {
if (diffMVdiffVal < -3) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (ssDCval < 124) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (ssDCval >= 124) {
if (Cs < 917) {
if (Rs < 1137) {
return 4;
}
else if (Rs >= 1137) {
return 8;
}
}
else if (Cs >= 917) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -3) {
if (ssDCval < 152) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 1358) {
return 8;
}
else if (Rs >= 1358) {
return 1;
}
}
}
else if (ssDCval >= 152) {
return 2;
}
}
}
else if (Cs >= 1193) {
if (negBalance < 2) {
if (ssDCval < 131) {
if (Rs < 1348) {
return 8;
}
else if (Rs >= 1348) {
return 4;
}
}
else if (ssDCval >= 131) {
return 1;
}
}
else if (negBalance >= 2) {
return 8;
}
}
}
else if (negBalance >= 9) {
if (MVDiff < 3) {
if (ssDCval < 57) {
if (Cs < 588) {
return 1;
}
else if (Cs >= 588) {
if (mcTcor < 99) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Cs < 906) {
return 8;
}
else if (Cs >= 906) {
if (Rs < 1415) {
return 1;
}
else if (Rs >= 1415) {
return 8;
}
}
}
}
else if (mcTcor >= 99) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (Cs < 1029) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
if (mu_mv_mag_sq < 128) {
if (mu_mv_mag_sq < 121) {
return 8;
}
else if (mu_mv_mag_sq >= 121) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (posBalance < 40) {
if (Rs < 1535) {
return 8;
}
else if (Rs >= 1535) {
return 4;
}
}
else if (posBalance >= 40) {
return 4;
}
}
}
}
else if (Cs >= 1029) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (gchDC < 2) {
if (diffRsCsdiff < -11) {
return 8;
}
else if (diffRsCsdiff >= -11) {
if (posBalance < 51) {
if (Rs < 1665) {
if (MVDiff < 4) {
if (Cs < 890) {
if (SC < 1816) {
return 1;
}
else if (SC >= 1816) {
return 8;
}
}
else if (Cs >= 890) {
return 1;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (Rs >= 1665) {
return 8;
}
}
else if (posBalance >= 51) {
if (Rs < 1432) {
return 4;
}
else if (Rs >= 1432) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (diffMVdiffVal < -15) {
return 4;
}
else if (diffMVdiffVal >= -15) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1810) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (ssDCval < 79) {
if (SC < 3013) {
if (Cs < 989) {
if (Cs < 654) {
return 2;
}
else if (Cs >= 654) {
if (Rs < 2009) {
return 1;
}
else if (Rs >= 2009) {
if (Rs < 2015) {
return 4;
}
else if (Rs >= 2015) {
return 1;
}
}
}
}
else if (Cs >= 989) {
if (Cs < 1078) {
if (negBalance < 8) {
return 1;
}
else if (negBalance >= 8) {
return 4;
}
}
else if (Cs >= 1078) {
return 1;
}
}
}
else if (SC >= 3013) {
if (Rs < 2256) {
return 4;
}
else if (Rs >= 2256) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 79) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (Rs < 1948) {
return 4;
}
else if (Rs >= 1948) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 45) {
if (refDCval < 47) {
if (Rs < 1815) {
if (mu_mv_mag_sq < 51) {
if (tcor < 88) {
if (Rs < 496) {
if (Rs < 454) {
if (Rs < 354) {
return 1;
}
else if (Rs >= 354) {
return 8;
}
}
else if (Rs >= 454) {
return 1;
}
}
else if (Rs >= 496) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (negBalance < 32) {
if (SC < 1632) {
return 2;
}
else if (SC >= 1632) {
return 1;
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
}
else if (tcor >= 88) {
if (negBalance < 37) {
if (Rs < 642) {
return 8;
}
else if (Rs >= 642) {
return 4;
}
}
else if (negBalance >= 37) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 51) {
if (refDCval < 29) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (refDCval >= 29) {
if (Rs < 1496) {
if (ssDCval < 43) {
if (Cs < 651) {
if (Rs < 797) {
if (Rs < 742) {
return 8;
}
else if (Rs >= 742) {
return 2;
}
}
else if (Rs >= 797) {
return 1;
}
}
else if (Cs >= 651) {
if (Rs < 556) {
return 1;
}
else if (Rs >= 556) {
return 8;
}
}
}
else if (ssDCval >= 43) {
if (RsCsDiff < 16) {
if (Rs < 1061) {
return 4;
}
else if (Rs >= 1061) {
return 4;
}
}
else if (RsCsDiff >= 16) {
if (Rs < 1332) {
return 8;
}
else if (Rs >= 1332) {
if (Cs < 1243) {
return 2;
}
else if (Cs >= 1243) {
return 8;
}
}
}
}
}
else if (Rs >= 1496) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (MVDiff < 13) {
if (Cs < 953) {
return 1;
}
else if (Cs >= 953) {
return 8;
}
}
else if (MVDiff >= 13) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1815) {
return 4;
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 86) {
if (Cs < 761) {
if (Rs < 1581) {
if (refDCval < 73) {
if (TSC < 5) {
if (negBalance < 37) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 489) {
if (Cs < 594) {
return 1;
}
else if (Cs >= 594) {
return 4;
}
}
else if (Rs >= 489) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
else if (negBalance >= 37) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Cs < 652) {
return 1;
}
else if (Cs >= 652) {
return 8;
}
}
}
}
else if (TSC >= 5) {
if (mcTcor < 91) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
else if (mcTcor >= 91) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
if (tcor < 84) {
return 2;
}
else if (tcor >= 84) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 1325) {
return 1;
}
else if (Rs >= 1325) {
if (Rs < 1435) {
return 2;
}
else if (Rs >= 1435) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 73) {
if (posBalance < 25) {
return 8;
}
else if (posBalance >= 25) {
if (diffMVdiffVal < 7) {
if (diffMVdiffVal < -7) {
return 8;
}
else if (diffMVdiffVal >= -7) {
if (Cs < 691) {
return 2;
}
else if (Cs >= 691) {
if (Cs < 714) {
return 8;
}
else if (Cs >= 714) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 7) {
if (Rs < 1089) {
return 8;
}
else if (Rs >= 1089) {
return 1;
}
}
}
}
}
else if (Rs >= 1581) {
if (tcor < 85) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Cs < 688) {
return 1;
}
else if (Cs >= 688) {
return 4;
}
}
}
else if (tcor >= 85) {
if (MVDiff < 39) {
if (negBalance < 29) {
if (Rs < 1830) {
return 2;
}
else if (Rs >= 1830) {
return 1;
}
}
else if (negBalance >= 29) {
return 1;
}
}
else if (MVDiff >= 39) {
return 2;
}
}
}
}
else if (Cs >= 761) {
if (MVDiff < 8) {
if (refDCval < 53) {
if (Rs < 997) {
if (Rs < 547) {
return 8;
}
else if (Rs >= 547) {
return 4;
}
}
else if (Rs >= 997) {
if (Rs < 1421) {
return 1;
}
else if (Rs >= 1421) {
return 2;
}
}
}
else if (refDCval >= 53) {
if (negBalance < 27) {
if (Rs < 1423) {
if (Rs < 765) {
return 4;
}
else if (Rs >= 765) {
return 1;
}
}
else if (Rs >= 1423) {
return 4;
}
}
else if (negBalance >= 27) {
if (Cs < 954) {
if (ssDCval < 73) {
if (tcor < 92) {
if (negBalance < 45) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
return 4;
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (tcor >= 92) {
if (TSC < 5) {
if (Rs < 807) {
return 4;
}
else if (Rs >= 807) {
return 1;
}
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (ssDCval >= 73) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 1164) {
return 2;
}
else if (Rs >= 1164) {
if (AFD < 10) {
if (RsCsDiff < 16) {
return 4;
}
else if (RsCsDiff >= 16) {
return 1;
}
}
else if (AFD >= 10) {
return 2;
}
}
}
}
}
else if (Cs >= 954) {
if (ssDCval < 139) {
if (tcor < 89) {
if (Rs < 1905) {
return 8;
}
else if (Rs >= 1905) {
if (Rs < 2193) {
return 2;
}
else if (Rs >= 2193) {
return 4;
}
}
}
else if (tcor >= 89) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (Rs < 723) {
return 8;
}
else if (Rs >= 723) {
return 4;
}
}
}
}
else if (ssDCval >= 139) {
return 8;
}
}
}
}
}
else if (MVDiff >= 8) {
if (refDCval < 70) {
if (AFD < 8) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (diffAFD < -2) {
if (Rs < 1104) {
return 4;
}
else if (Rs >= 1104) {
return 8;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
}
}
else if (AFD >= 8) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (mcTcor < 93) {
if (Cs < 1142) {
if (negBalance < 35) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (negBalance >= 35) {
if (SC < 2195) {
return 8;
}
else if (SC >= 2195) {
if (Cs < 1132) {
return 4;
}
else if (Cs >= 1132) {
return 8;
}
}
}
}
else if (Cs >= 1142) {
return 2;
}
}
else if (mcTcor >= 93) {
if (negBalance < 53) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
else if (negBalance >= 53) {
return 4;
}
}
}
}
}
else if (refDCval >= 70) {
if (ssDCval < 155) {
if (Rs < 1863) {
if (posBalance < 42) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1046) {
return 2;
}
else if (Rs >= 1046) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (posBalance >= 42) {
if (Cs < 1049) {
if (Rs < 1169) {
return 4;
}
else if (Rs >= 1169) {
if (Rs < 1464) {
return 8;
}
else if (Rs >= 1464) {
return 2;
}
}
}
else if (Cs >= 1049) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
}
else if (Rs >= 1863) {
if (MVDiff < 19) {
if (Cs < 867) {
return 1;
}
else if (Cs >= 867) {
return 8;
}
}
else if (MVDiff >= 19) {
if (Rs < 2293) {
return 2;
}
else if (Rs >= 2293) {
if (Rs < 2305) {
return 8;
}
else if (Rs >= 2305) {
return 2;
}
}
}
}
}
else if (ssDCval >= 155) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (refDCval < 58) {
if (TSC < 3) {
if (Rs < 613) {
if (Rs < 416) {
return 2;
}
else if (Rs >= 416) {
return 1;
}
}
else if (Rs >= 613) {
if (CsDiff < 2) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
if (Rs < 1129) {
return 4;
}
else if (Rs >= 1129) {
return 8;
}
}
}
else if (CsDiff >= 2) {
return 4;
}
}
}
else if (TSC >= 3) {
if (posBalance < 42) {
if (mu_mv_mag_sq < 113) {
if (CsDiff < 3) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 113) {
if (Cs < 1074) {
return 4;
}
else if (Cs >= 1074) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (Cs < 886) {
if (Cs < 605) {
if (SC < 1589) {
return 8;
}
else if (SC >= 1589) {
return 1;
}
}
else if (Cs >= 605) {
return 8;
}
}
else if (Cs >= 886) {
return 1;
}
}
}
}
else if (refDCval >= 58) {
if (mu_mv_mag_sq < 115) {
if (ssDCval < 125) {
if (ssDCval < 116) {
if (negBalance < 30) {
if (MVDiff < 2) {
if (Rs < 1146) {
if (Rs < 636) {
return 2;
}
else if (Rs >= 636) {
return 4;
}
}
else if (Rs >= 1146) {
if (Rs < 1473) {
return 8;
}
else if (Rs >= 1473) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
return 8;
}
}
}
else if (negBalance >= 30) {
if (tcor < 75) {
if (Rs < 1512) {
return 4;
}
else if (Rs >= 1512) {
return 2;
}
}
else if (tcor >= 75) {
if (SC < 1385) {
if (Rs < 608) {
return 2;
}
else if (Rs >= 608) {
return 8;
}
}
else if (SC >= 1385) {
if (Rs < 2249) {
if (Rs < 845) {
if (Rs < 686) {
return 8;
}
else if (Rs >= 686) {
return 4;
}
}
else if (Rs >= 845) {
return 8;
}
}
else if (Rs >= 2249) {
if (Cs < 810) {
if (Cs < 648) {
return 4;
}
else if (Cs >= 648) {
if (Rs < 2268) {
return 4;
}
else if (Rs >= 2268) {
return 8;
}
}
}
else if (Cs >= 810) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 116) {
return 4;
}
}
else if (ssDCval >= 125) {
if (mu_mv_mag_sq < 90) {
if (TSC < 4) {
if (Rs < 1549) {
return 8;
}
else if (Rs >= 1549) {
return 4;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 90) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 2396) {
return 8;
}
else if (Rs >= 2396) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 45) {
if (mu_mv_mag_sq < 18) {
if (negBalance < 47) {
if (refDCval < 51) {
if (Rs < 883) {
if (RsCsDiff < 57) {
if (ssDCval < 36) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (ssDCval >= 36) {
if (ssDCval < 43) {
if (RsCsDiff < 11) {
if (Rs < 565) {
return 2;
}
else if (Rs >= 565) {
return 8;
}
}
else if (RsCsDiff >= 11) {
if (TSC < 6) {
if (refDCval < 35) {
return 8;
}
else if (refDCval >= 35) {
return 2;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (ssDCval >= 43) {
if (TSC < 5) {
if (Rs < 682) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Rs >= 682) {
return 4;
}
}
else if (TSC >= 5) {
if (diffRsCsdiff < 6) {
if (Rs < 855) {
return 2;
}
else if (Rs >= 855) {
return 8;
}
}
else if (diffRsCsdiff >= 6) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 57) {
return 1;
}
}
else if (Rs >= 883) {
if (Cs < 1118) {
if (posBalance < 50) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (posBalance >= 50) {
return 8;
}
}
else if (Cs >= 1118) {
if (tcor < 79) {
if (Cs < 1245) {
return 8;
}
else if (Cs >= 1245) {
return 1;
}
}
else if (tcor >= 79) {
if (Rs < 1126) {
return 4;
}
else if (Rs >= 1126) {
return 8;
}
}
}
}
}
else if (refDCval >= 51) {
if (mcTcor < 91) {
if (AFD < 10) {
if (ssDCval < 75) {
if (MVDiff < 11) {
if (Rs < 702) {
return 2;
}
else if (Rs >= 702) {
if (AFD < 8) {
if (Cs < 984) {
return 8;
}
else if (Cs >= 984) {
return 4;
}
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (ssDCval >= 75) {
if (ssDCval < 155) {
return 8;
}
else if (ssDCval >= 155) {
return 2;
}
}
}
else if (AFD >= 10) {
if (MVDiff < 13) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (Rs < 871) {
return 8;
}
else if (Rs >= 871) {
return 1;
}
}
}
else if (MVDiff >= 13) {
if (ssDCval < 69) {
if (Rs < 694) {
return 8;
}
else if (Rs >= 694) {
if (Rs < 1589) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
return 4;
}
}
else if (Rs >= 1589) {
return 4;
}
}
}
else if (ssDCval >= 69) {
if (tcor < 73) {
return 8;
}
else if (tcor >= 73) {
if (AFD < 12) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (AFD >= 12) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (SC < 2764) {
if (RsDiff < 2) {
if (negBalance < 38) {
if (Rs < 660) {
if (Rs < 335) {
return 2;
}
else if (Rs >= 335) {
return 4;
}
}
else if (Rs >= 660) {
if (Rs < 871) {
return 1;
}
else if (Rs >= 871) {
return 8;
}
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (RsDiff >= 2) {
if (Cs < 688) {
if (ssDCval < 84) {
if (posBalance < 54) {
if (posBalance < 47) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (Rs < 714) {
return 2;
}
else if (Rs >= 714) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
}
}
else if (posBalance >= 54) {
if (Cs < 547) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 1;
}
}
else if (Cs >= 547) {
if (SC < 1246) {
if (SC < 1165) {
return 1;
}
else if (SC >= 1165) {
return 8;
}
}
else if (SC >= 1246) {
if (Cs < 565) {
if (Rs < 1023) {
return 1;
}
else if (Rs >= 1023) {
return 2;
}
}
else if (Cs >= 565) {
if (Cs < 590) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 4;
}
}
else if (Cs >= 590) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (RsCsDiff < 43) {
if (Rs < 1840) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
return 1;
}
}
else if (Rs >= 1840) {
return 8;
}
}
else if (RsCsDiff >= 43) {
return 1;
}
}
}
else if (Cs >= 688) {
if (MVDiff < 6) {
if (refDCval < 65) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
if (Cs < 935) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (Cs >= 935) {
if (Cs < 1202) {
return 8;
}
else if (Cs >= 1202) {
return 4;
}
}
}
}
else if (refDCval >= 65) {
if (Cs < 1046) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
if (Rs < 804) {
if (Rs < 731) {
return 4;
}
else if (Rs >= 731) {
return 1;
}
}
else if (Rs >= 804) {
return 8;
}
}
}
else if (Cs >= 1046) {
if (RsDiff < 4) {
if (Rs < 564) {
return 1;
}
else if (Rs >= 564) {
return 8;
}
}
else if (RsDiff >= 4) {
if (Rs < 1110) {
return 2;
}
else if (Rs >= 1110) {
return 4;
}
}
}
}
}
else if (MVDiff >= 6) {
if (tcor < 85) {
if (negBalance < 41) {
if (ssDCval < 136) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (ssDCval >= 136) {
return 2;
}
}
else if (negBalance >= 41) {
if (Rs < 588) {
return 4;
}
else if (Rs >= 588) {
if (Rs < 1380) {
return 8;
}
else if (Rs >= 1380) {
return 2;
}
}
}
}
else if (tcor >= 85) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 148) {
if (ssDCval < 109) {
if (gchDC < 1) {
if (Cs < 951) {
if (Rs < 1821) {
return 4;
}
else if (Rs >= 1821) {
return 2;
}
}
else if (Cs >= 951) {
if (Rs < 1033) {
return 2;
}
else if (Rs >= 1033) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (diffAFD < -2) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (diffAFD >= -2) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (posBalance < 58) {
if (TSC < 7) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (Rs < 1616) {
return 1;
}
else if (Rs >= 1616) {
return 2;
}
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (posBalance >= 58) {
if (TSC < 6) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
return 4;
}
}
else if (TSC >= 6) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 109) {
if (ssDCval < 138) {
if (diffRsCsdiff < -8) {
return 1;
}
else if (diffRsCsdiff >= -8) {
return 8;
}
}
else if (ssDCval >= 138) {
return 4;
}
}
}
else if (ssDCval >= 148) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 852) {
return 1;
}
else if (Rs >= 852) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 2764) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (refDCval < 70) {
if (mcTcor < 93) {
if (Cs < 799) {
return 2;
}
else if (Cs >= 799) {
return 4;
}
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (refDCval >= 70) {
if (ssDCval < 93) {
if (Cs < 1157) {
if (posBalance < 57) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
return 4;
}
}
else if (posBalance >= 57) {
return 2;
}
}
else if (Cs >= 1157) {
return 8;
}
}
else if (ssDCval >= 93) {
if (TSC < 7) {
if (Rs < 2043) {
if (negBalance < 41) {
return 2;
}
else if (negBalance >= 41) {
if (Rs < 1926) {
return 8;
}
else if (Rs >= 1926) {
return 2;
}
}
}
else if (Rs >= 2043) {
return 8;
}
}
else if (TSC >= 7) {
if (Rs < 2297) {
return 8;
}
else if (Rs >= 2297) {
return 4;
}
}
}
}
}
}
}
}
}
else if (negBalance >= 47) {
if (MVDiff < 18) {
if (Rs < 978) {
if (RsCsDiff < 11) {
if (MVDiff < 4) {
if (posBalance < 55) {
if (Cs < 575) {
return 2;
}
else if (Cs >= 575) {
return 8;
}
}
else if (posBalance >= 55) {
if (Cs < 958) {
if (refDCval < 111) {
return 1;
}
else if (refDCval >= 111) {
return 2;
}
}
else if (Cs >= 958) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (RsCsDiff >= 11) {
if (ssDCval < 65) {
if (mcTcor < 74) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
else if (mcTcor >= 74) {
if (Rs < 812) {
if (tcor < 79) {
if (TSC < 6) {
if (negBalance < 52) {
return 4;
}
else if (negBalance >= 52) {
if (Rs < 717) {
return 4;
}
else if (Rs >= 717) {
return 1;
}
}
}
else if (TSC >= 6) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (Rs < 551) {
return 4;
}
else if (Rs >= 551) {
return 8;
}
}
}
}
else if (tcor >= 79) {
if (Rs < 566) {
return 1;
}
else if (Rs >= 566) {
return 8;
}
}
}
else if (Rs >= 812) {
if (posBalance < 57) {
if (Cs < 1137) {
return 8;
}
else if (Cs >= 1137) {
return 4;
}
}
else if (posBalance >= 57) {
if (refDCval < 48) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (mcTcor < 79) {
if (Rs < 863) {
return 1;
}
else if (Rs >= 863) {
return 8;
}
}
else if (mcTcor >= 79) {
return 1;
}
}
}
}
else if (refDCval >= 48) {
if (MVDiff < 7) {
if (Cs < 695) {
return 8;
}
else if (Cs >= 695) {
return 4;
}
}
else if (MVDiff >= 7) {
if (AFD < 12) {
if (Cs < 843) {
if (Rs < 902) {
return 1;
}
else if (Rs >= 902) {
return 8;
}
}
else if (Cs >= 843) {
return 1;
}
}
else if (AFD >= 12) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 5) {
if (ssDCval < 117) {
if (RsDiff < 4) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 786) {
return 8;
}
else if (Rs >= 786) {
return 1;
}
}
}
else if (RsDiff >= 4) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (diffAFD < -1) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
}
}
else if (ssDCval >= 117) {
if (Rs < 935) {
return 4;
}
else if (Rs >= 935) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Cs < 604) {
if (Rs < 709) {
return 2;
}
else if (Rs >= 709) {
return 4;
}
}
else if (Cs >= 604) {
return 1;
}
}
}
}
}
else if (Rs >= 978) {
if (Rs < 1304) {
if (tcor < 82) {
if (posBalance < 63) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
if (MVDiff < 3) {
if (Rs < 1133) {
return 1;
}
else if (Rs >= 1133) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (Cs < 610) {
return 2;
}
else if (Cs >= 610) {
return 8;
}
}
}
else if (tcor >= 82) {
if (Cs < 864) {
if (diffRsCsdiff < 11) {
return 8;
}
else if (diffRsCsdiff >= 11) {
if (Rs < 1110) {
return 4;
}
else if (Rs >= 1110) {
return 8;
}
}
}
else if (Cs >= 864) {
if (ssDCval < 59) {
if (Cs < 913) {
return 8;
}
else if (Cs >= 913) {
return 4;
}
}
else if (ssDCval >= 59) {
if (Cs < 868) {
return 4;
}
else if (Cs >= 868) {
if (tcor < 95) {
if (diffRsCsdiff < 13) {
return 8;
}
else if (diffRsCsdiff >= 13) {
if (Rs < 1188) {
return 8;
}
else if (Rs >= 1188) {
return 1;
}
}
}
else if (tcor >= 95) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1304) {
if (posBalance < 52) {
if (mcTcor < 93) {
if (posBalance < 50) {
if (Rs < 1883) {
return 8;
}
else if (Rs >= 1883) {
return 4;
}
}
else if (posBalance >= 50) {
return 1;
}
}
else if (mcTcor >= 93) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (AFD < 10) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 4;
}
}
else if (AFD >= 10) {
return 2;
}
}
}
}
else if (posBalance >= 52) {
if (Cs < 737) {
if (ssDCval < 80) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (ssDCval >= 80) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
}
else if (Cs >= 737) {
if (posBalance < 54) {
if (ssDCval < 76) {
if (Rs < 1592) {
return 8;
}
else if (Rs >= 1592) {
return 1;
}
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (posBalance >= 54) {
if (ssDCval < 95) {
if (RsDiff < 9) {
if (refDCval < 68) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (RsCsDiff < 69) {
return 8;
}
else if (RsCsDiff >= 69) {
if (Rs < 1534) {
return 4;
}
else if (Rs >= 1534) {
return 8;
}
}
}
}
else if (refDCval >= 68) {
if (MVDiff < 8) {
if (Rs < 1609) {
return 8;
}
else if (Rs >= 1609) {
if (diffRsCsdiff < -5) {
if (Cs < 940) {
return 1;
}
else if (Cs >= 940) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
return 1;
}
}
}
else if (MVDiff >= 8) {
if (tcor < 89) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
return 4;
}
}
}
else if (tcor >= 89) {
return 8;
}
}
}
}
else if (RsDiff >= 9) {
if (Rs < 2219) {
return 2;
}
else if (Rs >= 2219) {
return 4;
}
}
}
else if (ssDCval >= 95) {
if (ssDCval < 99) {
if (SC < 2343) {
return 2;
}
else if (SC >= 2343) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (Rs < 1766) {
return 2;
}
else if (Rs >= 1766) {
return 8;
}
}
}
}
}
else if (ssDCval >= 99) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 18) {
if (Cs < 768) {
if (mcTcor < 93) {
if (Cs < 591) {
return 8;
}
else if (Cs >= 591) {
if (tcor < 86) {
if (posBalance < 62) {
if (negBalance < 54) {
if (SC < 1914) {
if (RsCsDiff < 19) {
return 8;
}
else if (RsCsDiff >= 19) {
if (mu_mv_mag_sq < 5) {
return 1;
}
else if (mu_mv_mag_sq >= 5) {
return 2;
}
}
}
else if (SC >= 1914) {
if (Cs < 697) {
return 1;
}
else if (Cs >= 697) {
return 8;
}
}
}
else if (negBalance >= 54) {
if (Cs < 605) {
return 4;
}
else if (Cs >= 605) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (diffRsCsdiff < 11) {
return 8;
}
else if (diffRsCsdiff >= 11) {
return 4;
}
}
}
else if (tcor >= 86) {
if (ssDCval < 79) {
return 4;
}
else if (ssDCval >= 79) {
return 1;
}
}
}
}
else if (mcTcor >= 93) {
if (Cs < 638) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (negBalance < 55) {
return 2;
}
else if (negBalance >= 55) {
if (Rs < 1670) {
return 1;
}
else if (Rs >= 1670) {
return 2;
}
}
}
}
else if (Cs >= 638) {
if (Rs < 1642) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Rs < 869) {
return 2;
}
else if (Rs >= 869) {
return 8;
}
}
}
else if (Rs >= 1642) {
return 2;
}
}
}
}
else if (Cs >= 768) {
if (Rs < 781) {
if (MVDiff < 25) {
if (diffRsCsdiff < -5) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
return 4;
}
}
else if (diffRsCsdiff >= -5) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (MVDiff >= 25) {
return 4;
}
}
else if (Rs >= 781) {
if (gchDC < 3) {
if (tcor < 79) {
if (posBalance < 64) {
if (diffAFD < -1) {
if (Rs < 1025) {
return 4;
}
else if (Rs >= 1025) {
if (Rs < 1475) {
return 2;
}
else if (Rs >= 1475) {
return 8;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (posBalance >= 64) {
if (Cs < 898) {
return 8;
}
else if (Cs >= 898) {
return 4;
}
}
}
else if (tcor >= 79) {
if (diffAFD < -1) {
if (diffAFD < -10) {
return 4;
}
else if (diffAFD >= -10) {
if (Cs < 801) {
if (Rs < 1436) {
return 8;
}
else if (Rs >= 1436) {
return 4;
}
}
else if (Cs >= 801) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (ssDCval < 149) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
if (Rs < 1187) {
return 4;
}
else if (Rs >= 1187) {
return 8;
}
}
}
else if (ssDCval >= 149) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
if (Rs < 1605) {
return 2;
}
else if (Rs >= 1605) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (Rs < 1607) {
if (ssDCval < 47) {
if (mu_mv_mag_sq < 101) {
if (Rs < 688) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (mu_mv_mag_sq < 33) {
if (Rs < 658) {
return 2;
}
else if (Rs >= 658) {
return 4;
}
}
else if (mu_mv_mag_sq >= 33) {
if (Cs < 536) {
return 4;
}
else if (Cs >= 536) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
}
else if (Rs >= 688) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (tcor < 83) {
if (MVDiff < 13) {
if (Cs < 661) {
if (Cs < 653) {
return 8;
}
else if (Cs >= 653) {
return 2;
}
}
else if (Cs >= 661) {
return 8;
}
}
else if (MVDiff >= 13) {
if (ssDCval < 40) {
return 8;
}
else if (ssDCval >= 40) {
if (negBalance < 35) {
if (Rs < 1132) {
return 1;
}
else if (Rs >= 1132) {
return 8;
}
}
else if (negBalance >= 35) {
if (Rs < 997) {
if (TSC < 7) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
else if (TSC >= 7) {
if (MVDiff < 15) {
if (Rs < 782) {
return 8;
}
else if (Rs >= 782) {
return 1;
}
}
else if (MVDiff >= 15) {
return 8;
}
}
}
else if (Rs >= 997) {
return 8;
}
}
}
}
}
else if (tcor >= 83) {
if (TSC < 6) {
if (RsCsDiff < 19) {
if (Rs < 878) {
if (Cs < 648) {
return 8;
}
else if (Cs >= 648) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
else if (Rs >= 878) {
if (Cs < 643) {
return 2;
}
else if (Cs >= 643) {
return 8;
}
}
}
else if (RsCsDiff >= 19) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (SC < 2245) {
if (Rs < 898) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (Rs >= 898) {
if (tcor < 87) {
if (posBalance < 57) {
if (Rs < 1029) {
return 8;
}
else if (Rs >= 1029) {
if (RsCsDiff < 38) {
return 1;
}
else if (RsCsDiff >= 38) {
if (Cs < 768) {
return 1;
}
else if (Cs >= 768) {
return 8;
}
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (tcor >= 87) {
if (Rs < 1143) {
return 8;
}
else if (Rs >= 1143) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 8;
}
}
}
}
}
else if (SC >= 2245) {
return 8;
}
}
}
else if (ssDCval >= 47) {
if (negBalance < 44) {
if (ssDCval < 78) {
if (Rs < 605) {
if (Cs < 857) {
if (tcor < 85) {
if (diffMVdiffVal < -10) {
return 8;
}
else if (diffMVdiffVal >= -10) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (TSC < 5) {
if (Rs < 467) {
return 1;
}
else if (Rs >= 467) {
return 4;
}
}
else if (TSC >= 5) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
else if (tcor >= 85) {
if (mu_mv_mag_sq < 103) {
if (mu_mv_mag_sq < 55) {
if (Cs < 576) {
return 8;
}
else if (Cs >= 576) {
return 4;
}
}
else if (mu_mv_mag_sq >= 55) {
if (Rs < 319) {
if (Cs < 608) {
return 4;
}
else if (Cs >= 608) {
return 1;
}
}
else if (Rs >= 319) {
if (Cs < 620) {
if (Rs < 509) {
if (tcor < 96) {
return 2;
}
else if (tcor >= 96) {
return 1;
}
}
else if (Rs >= 509) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (Cs >= 620) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 445) {
return 1;
}
else if (Rs >= 445) {
if (Cs < 590) {
return 8;
}
else if (Cs >= 590) {
return 4;
}
}
}
}
}
else if (Cs >= 857) {
if (ssDCval < 53) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (Rs < 537) {
return 2;
}
else if (Rs >= 537) {
return 4;
}
}
}
else if (ssDCval >= 53) {
return 1;
}
}
}
else if (Rs >= 605) {
if (refDCval < 64) {
if (TSCindex < 3) {
if (RsDiff < 2) {
if (Cs < 1009) {
return 8;
}
else if (Cs >= 1009) {
return 1;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (TSCindex >= 3) {
if (Rs < 1133) {
if (RsDiff < 3) {
if (Rs < 769) {
if (refDCval < 55) {
if (refDCval < 51) {
if (Rs < 706) {
return 1;
}
else if (Rs >= 706) {
return 8;
}
}
else if (refDCval >= 51) {
if (Cs < 565) {
return 8;
}
else if (Cs >= 565) {
return 4;
}
}
}
else if (refDCval >= 55) {
if (negBalance < 31) {
if (Cs < 722) {
if (Rs < 659) {
return 8;
}
else if (Rs >= 659) {
return 2;
}
}
else if (Cs >= 722) {
return 4;
}
}
else if (negBalance >= 31) {
if (Rs < 742) {
return 8;
}
else if (Rs >= 742) {
return 4;
}
}
}
}
else if (Rs >= 769) {
if (CsDiff < 4) {
if (diffRsCsdiff < -5) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (ssDCval < 59) {
if (Cs < 796) {
return 8;
}
else if (Cs >= 796) {
return 4;
}
}
else if (ssDCval >= 59) {
return 8;
}
}
}
else if (diffRsCsdiff >= -5) {
if (negBalance < 20) {
return 1;
}
else if (negBalance >= 20) {
return 8;
}
}
}
else if (CsDiff >= 4) {
return 4;
}
}
}
else if (RsDiff >= 3) {
if (SC < 1500) {
if (TSC < 5) {
if (Rs < 727) {
return 4;
}
else if (Rs >= 727) {
return 8;
}
}
else if (TSC >= 5) {
return 1;
}
}
else if (SC >= 1500) {
if (posBalance < 58) {
if (CsDiff < 4) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
if (Rs < 981) {
return 4;
}
else if (Rs >= 981) {
if (Rs < 1094) {
return 8;
}
else if (Rs >= 1094) {
if (Cs < 928) {
return 8;
}
else if (Cs >= 928) {
return 4;
}
}
}
}
}
else if (CsDiff >= 4) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Rs < 1063) {
if (SC < 1980) {
return 8;
}
else if (SC >= 1980) {
return 4;
}
}
else if (Rs >= 1063) {
return 8;
}
}
}
}
}
else if (posBalance >= 58) {
if (mu_mv_mag_sq < 86) {
if (Rs < 1090) {
if (Cs < 577) {
return 4;
}
else if (Cs >= 577) {
return 8;
}
}
else if (Rs >= 1090) {
return 4;
}
}
else if (mu_mv_mag_sq >= 86) {
if (SC < 1843) {
return 4;
}
else if (SC >= 1843) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1133) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 115) {
if (posBalance < 50) {
if (Rs < 1445) {
return 4;
}
else if (Rs >= 1445) {
return 8;
}
}
else if (posBalance >= 50) {
if (TSC < 5) {
if (refDCval < 61) {
if (ssDCval < 52) {
if (Rs < 1252) {
return 8;
}
else if (Rs >= 1252) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (ssDCval >= 52) {
return 8;
}
}
else if (refDCval >= 61) {
return 1;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
else if (MVDiff >= 3) {
if (MVDiff < 9) {
if (Rs < 1300) {
if (RsCsDiff < 19) {
if (Cs < 745) {
return 8;
}
else if (Cs >= 745) {
if (diffRsCsdiff < -2) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
else if (RsCsDiff >= 19) {
if (diffRsCsdiff < 5) {
return 8;
}
else if (diffRsCsdiff >= 5) {
if (Rs < 1212) {
return 4;
}
else if (Rs >= 1212) {
return 8;
}
}
}
}
else if (Rs >= 1300) {
if (ssDCval < 61) {
if (ssDCval < 48) {
return 2;
}
else if (ssDCval >= 48) {
if (negBalance < 23) {
return 1;
}
else if (negBalance >= 23) {
return 8;
}
}
}
else if (ssDCval >= 61) {
if (diffRsCsdiff < -3) {
if (Cs < 1214) {
return 8;
}
else if (Cs >= 1214) {
return 1;
}
}
else if (diffRsCsdiff >= -3) {
return 4;
}
}
}
}
else if (MVDiff >= 9) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
if (posBalance < 60) {
if (TSC < 5) {
if (Rs < 1156) {
return 1;
}
else if (Rs >= 1156) {
return 8;
}
}
else if (TSC >= 5) {
if (SC < 2367) {
if (ssDCval < 54) {
if (Rs < 1412) {
return 8;
}
else if (Rs >= 1412) {
return 1;
}
}
else if (ssDCval >= 54) {
if (SC < 1870) {
return 1;
}
else if (SC >= 1870) {
return 2;
}
}
}
else if (SC >= 2367) {
return 1;
}
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 64) {
if (Cs < 1076) {
if (SC < 1899) {
if (RsDiff < 3) {
if (Cs < 762) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 762) {
if (Rs < 823) {
if (mu_mv_mag_sq < 99) {
return 1;
}
else if (mu_mv_mag_sq >= 99) {
if (Rs < 692) {
return 8;
}
else if (Rs >= 692) {
if (Cs < 840) {
return 4;
}
else if (Cs >= 840) {
return 2;
}
}
}
}
else if (Rs >= 823) {
return 2;
}
}
}
else if (RsDiff >= 3) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
if (Cs < 697) {
return 1;
}
else if (Cs >= 697) {
return 8;
}
}
}
}
else if (SC >= 1899) {
if (Cs < 908) {
if (posBalance < 58) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
}
else if (Cs >= 908) {
if (diffMVdiffVal < -2) {
if (mcTcor < 93) {
if (Rs < 1254) {
return 8;
}
else if (Rs >= 1254) {
return 4;
}
}
else if (mcTcor >= 93) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
}
else if (Cs >= 1076) {
if (TSC < 5) {
if (Rs < 1032) {
if (Rs < 804) {
return 8;
}
else if (Rs >= 804) {
return 4;
}
}
else if (Rs >= 1032) {
if (negBalance < 38) {
if (mu_mv_mag_sq < 115) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (ssDCval < 69) {
if (Cs < 1243) {
return 4;
}
else if (Cs >= 1243) {
if (Cs < 1253) {
return 1;
}
else if (Cs >= 1253) {
return 4;
}
}
}
else if (ssDCval >= 69) {
return 8;
}
}
}
else if (negBalance >= 38) {
if (diffRsCsdiff < 6) {
if (mu_mv_mag_sq < 106) {
if (Cs < 1167) {
return 8;
}
else if (Cs >= 1167) {
return 1;
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 1302) {
if (Rs < 1274) {
return 8;
}
else if (Rs >= 1274) {
return 1;
}
}
else if (Rs >= 1302) {
return 8;
}
}
}
else if (diffRsCsdiff >= 6) {
return 1;
}
}
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 46) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
else if (mu_mv_mag_sq >= 46) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (posBalance < 58) {
if (mu_mv_mag_sq < 89) {
if (Rs < 1551) {
return 4;
}
else if (Rs >= 1551) {
return 1;
}
}
else if (mu_mv_mag_sq >= 89) {
if (posBalance < 55) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 4;
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (diffRsCsdiff < 6) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (diffRsCsdiff >= 6) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 78) {
if (refDCval < 130) {
if (Cs < 957) {
if (posBalance < 54) {
if (tcor < 92) {
if (ssDCval < 92) {
if (Rs < 446) {
return 8;
}
else if (Rs >= 446) {
return 2;
}
}
else if (ssDCval >= 92) {
return 8;
}
}
else if (tcor >= 92) {
if (MVDiff < 5) {
if (negBalance < 40) {
if (diffAFD < 1) {
if (diffMVdiffVal < -1) {
if (Rs < 569) {
return 8;
}
else if (Rs >= 569) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (negBalance >= 40) {
if (Rs < 626) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
else if (Rs >= 626) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (Rs < 1477) {
return 8;
}
else if (Rs >= 1477) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
if (Rs < 727) {
if (TSC < 4) {
if (Rs < 454) {
return 8;
}
else if (Rs >= 454) {
return 2;
}
}
else if (TSC >= 4) {
if (diffMVdiffVal < -7) {
if (diffMVdiffVal < -9) {
return 8;
}
else if (diffMVdiffVal >= -9) {
if (Rs < 591) {
return 2;
}
else if (Rs >= 591) {
return 8;
}
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
}
else if (Rs >= 727) {
if (SC < 1999) {
if (diffRsCsdiff < -3) {
return 1;
}
else if (diffRsCsdiff >= -3) {
if (Rs < 740) {
return 1;
}
else if (Rs >= 740) {
return 8;
}
}
}
else if (SC >= 1999) {
return 8;
}
}
}
}
else if (Cs >= 957) {
if (RsCsDiff < 38) {
if (posBalance < 50) {
if (Rs < 1273) {
if (Cs < 1180) {
return 8;
}
else if (Cs >= 1180) {
return 4;
}
}
else if (Rs >= 1273) {
return 8;
}
}
else if (posBalance >= 50) {
if (negBalance < 40) {
if (posBalance < 51) {
if (Rs < 1085) {
return 4;
}
else if (Rs >= 1085) {
return 8;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (negBalance >= 40) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (SC < 1942) {
if (Rs < 677) {
return 8;
}
else if (Rs >= 677) {
return 4;
}
}
else if (SC >= 1942) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 1473) {
return 4;
}
else if (Rs >= 1473) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 38) {
if (ssDCval < 104) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (RsCsDiff < 46) {
if (Rs < 1532) {
return 8;
}
else if (Rs >= 1532) {
return 2;
}
}
else if (RsCsDiff >= 46) {
return 2;
}
}
}
else if (ssDCval >= 104) {
return 8;
}
}
}
}
else if (refDCval >= 130) {
if (tcor < 88) {
if (ssDCval < 145) {
if (Rs < 1099) {
if (Rs < 746) {
return 8;
}
else if (Rs >= 746) {
return 1;
}
}
else if (Rs >= 1099) {
if (Cs < 1127) {
return 4;
}
else if (Cs >= 1127) {
return 1;
}
}
}
else if (ssDCval >= 145) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
else if (tcor >= 88) {
if (MVDiff < 25) {
if (posBalance < 61) {
if (Rs < 357) {
return 2;
}
else if (Rs >= 357) {
if (ssDCval < 160) {
if (mu_mv_mag_sq < 124) {
return 8;
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 1039) {
return 8;
}
else if (Rs >= 1039) {
return 4;
}
}
}
else if (ssDCval >= 160) {
return 2;
}
}
}
else if (posBalance >= 61) {
if (Cs < 993) {
if (Rs < 1347) {
return 8;
}
else if (Rs >= 1347) {
return 1;
}
}
else if (Cs >= 993) {
return 4;
}
}
}
else if (MVDiff >= 25) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 421) {
if (ssDCval < 96) {
if (RsDiff < 2) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (ssDCval >= 96) {
if (Cs < 662) {
return 2;
}
else if (Cs >= 662) {
return 4;
}
}
}
else if (Rs >= 421) {
if (mcTcor < 95) {
if (MVDiff < 3) {
if (Rs < 463) {
return 1;
}
else if (Rs >= 463) {
return 8;
}
}
else if (MVDiff >= 3) {
if (tcor < 77) {
if (SC < 1846) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (mcTcor < 84) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (AFD < 11) {
if (Cs < 912) {
return 8;
}
else if (Cs >= 912) {
return 4;
}
}
else if (AFD >= 11) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (mcTcor >= 84) {
return 4;
}
}
}
else if (SC >= 1846) {
if (ssDCval < 96) {
if (Cs < 1286) {
if (negBalance < 45) {
if (Cs < 989) {
return 8;
}
else if (Cs >= 989) {
return 1;
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (Cs >= 1286) {
return 1;
}
}
else if (ssDCval >= 96) {
if (Rs < 1344) {
return 1;
}
else if (Rs >= 1344) {
return 4;
}
}
}
}
else if (tcor >= 77) {
if (posBalance < 56) {
if (RsDiff < 3) {
if (AFD < 11) {
if (ssDCval < 129) {
return 8;
}
else if (ssDCval >= 129) {
return 4;
}
}
else if (AFD >= 11) {
if (Rs < 820) {
return 2;
}
else if (Rs >= 820) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (AFD < 8) {
if (mu_mv_mag_sq < 63) {
if (Cs < 856) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (Rs < 1254) {
return 4;
}
else if (Rs >= 1254) {
if (Rs < 1472) {
return 8;
}
else if (Rs >= 1472) {
return 4;
}
}
}
}
else if (Cs >= 856) {
return 4;
}
}
else if (mu_mv_mag_sq >= 63) {
return 8;
}
}
else if (AFD >= 8) {
if (ssDCval < 75) {
if (ssDCval < 60) {
if (MVDiff < 7) {
if (Rs < 1258) {
return 4;
}
else if (Rs >= 1258) {
return 8;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (ssDCval >= 60) {
if (tcor < 85) {
if (TSC < 7) {
if (Cs < 671) {
return 4;
}
else if (Cs >= 671) {
return 8;
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (tcor >= 85) {
if (Cs < 804) {
return 4;
}
else if (Cs >= 804) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
}
}
}
else if (ssDCval >= 75) {
if (diffMVdiffVal < -16) {
return 2;
}
else if (diffMVdiffVal >= -16) {
if (refDCval < 155) {
if (diffMVdiffVal < 7) {
return 8;
}
else if (diffMVdiffVal >= 7) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 1536) {
return 4;
}
else if (Rs >= 1536) {
return 2;
}
}
}
}
else if (refDCval >= 155) {
if (Rs < 1436) {
return 2;
}
else if (Rs >= 1436) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 56) {
if (Cs < 787) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 70) {
if (diffRsCsdiff < 10) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (Cs < 785) {
return 8;
}
else if (Cs >= 785) {
return 4;
}
}
}
else if (diffRsCsdiff >= 10) {
if (Rs < 887) {
return 4;
}
else if (Rs >= 887) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
return 8;
}
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (Cs >= 787) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (mu_mv_mag_sq < 47) {
if (negBalance < 47) {
if (Cs < 865) {
return 8;
}
else if (Cs >= 865) {
return 4;
}
}
else if (negBalance >= 47) {
if (Cs < 1226) {
return 8;
}
else if (Cs >= 1226) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (Cs < 1256) {
if (negBalance < 45) {
if (ssDCval < 71) {
if (Rs < 1381) {
return 4;
}
else if (Rs >= 1381) {
return 8;
}
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (Cs >= 1256) {
return 2;
}
}
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (Rs < 935) {
if (Cs < 708) {
if (Rs < 845) {
return 8;
}
else if (Rs >= 845) {
return 4;
}
}
else if (Cs >= 708) {
if (ssDCval < 133) {
if (RsCsDiff < 12) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
return 8;
}
}
else if (RsCsDiff >= 12) {
if (SC < 1896) {
return 8;
}
else if (SC >= 1896) {
if (Cs < 1238) {
return 1;
}
else if (Cs >= 1238) {
return 8;
}
}
}
}
else if (ssDCval >= 133) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
else if (Rs >= 935) {
if (diffMVdiffVal < -1) {
if (SC < 1860) {
if (Rs < 1022) {
return 8;
}
else if (Rs >= 1022) {
return 4;
}
}
else if (SC >= 1860) {
if (Cs < 939) {
if (posBalance < 48) {
if (Rs < 1261) {
return 8;
}
else if (Rs >= 1261) {
return 4;
}
}
else if (posBalance >= 48) {
return 8;
}
}
else if (Cs >= 939) {
if (refDCval < 72) {
return 8;
}
else if (refDCval >= 72) {
if (Rs < 1491) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 1;
}
}
}
else if (Rs >= 1491) {
return 8;
}
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (Cs < 1123) {
if (MVDiff < 28) {
return 8;
}
else if (MVDiff >= 28) {
if (Rs < 1009) {
return 8;
}
else if (Rs >= 1009) {
return 4;
}
}
}
else if (Cs >= 1123) {
if (AFD < 9) {
if (Rs < 1245) {
return 1;
}
else if (Rs >= 1245) {
return 8;
}
}
else if (AFD >= 9) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1607) {
if (refDCval < 72) {
if (refDCval < 62) {
if (Cs < 889) {
if (tcor < 91) {
if (mcTcor < 91) {
if (tcor < 77) {
if (Cs < 803) {
return 8;
}
else if (Cs >= 803) {
return 2;
}
}
else if (tcor >= 77) {
return 4;
}
}
else if (mcTcor >= 91) {
if (Cs < 854) {
return 2;
}
else if (Cs >= 854) {
return 4;
}
}
}
else if (tcor >= 91) {
if (negBalance < 29) {
if (Rs < 2147) {
return 1;
}
else if (Rs >= 2147) {
return 8;
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
else if (Cs >= 889) {
if (mu_mv_mag_sq < 116) {
if (negBalance < 39) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 1759) {
return 8;
}
else if (Rs >= 1759) {
return 4;
}
}
}
else if (negBalance >= 39) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
if (Cs < 1163) {
return 2;
}
else if (Cs >= 1163) {
return 8;
}
}
}
}
else if (refDCval >= 62) {
if (Rs < 2234) {
if (mu_mv_mag_sq < 84) {
if (negBalance < 45) {
if (MVDiff < 14) {
if (Cs < 1110) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (Cs >= 1110) {
if (Rs < 1820) {
return 1;
}
else if (Rs >= 1820) {
return 8;
}
}
}
else if (MVDiff >= 14) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
if (Rs < 1705) {
return 4;
}
else if (Rs >= 1705) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (Cs < 1023) {
if (Rs < 1661) {
return 8;
}
else if (Rs >= 1661) {
return 2;
}
}
else if (Cs >= 1023) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (TSC < 8) {
if (Cs < 1035) {
if (SC < 2549) {
return 8;
}
else if (SC >= 2549) {
return 2;
}
}
else if (Cs >= 1035) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (Rs < 1690) {
return 8;
}
else if (Rs >= 1690) {
return 2;
}
}
}
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (Rs >= 2234) {
if (refDCval < 70) {
return 2;
}
else if (refDCval >= 70) {
return 4;
}
}
}
}
else if (refDCval >= 72) {
if (RsDiff < 9) {
if (refDCval < 81) {
if (negBalance < 47) {
if (mu_mv_mag_sq < 109) {
if (Rs < 2204) {
if (refDCval < 78) {
if (RsCsDiff < 47) {
return 1;
}
else if (RsCsDiff >= 47) {
return 8;
}
}
else if (refDCval >= 78) {
if (SC < 3053) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (Rs < 1680) {
return 1;
}
else if (Rs >= 1680) {
return 8;
}
}
}
else if (SC >= 3053) {
return 8;
}
}
}
else if (Rs >= 2204) {
if (Rs < 2253) {
if (Rs < 2221) {
return 4;
}
else if (Rs >= 2221) {
return 1;
}
}
else if (Rs >= 2253) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
else if (negBalance >= 47) {
if (TSC < 7) {
if (Rs < 1821) {
return 8;
}
else if (Rs >= 1821) {
return 1;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (negBalance < 49) {
if (ssDCval < 98) {
if (RsCsDiff < 23) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (RsCsDiff >= 23) {
if (ssDCval < 87) {
if (Cs < 1263) {
return 8;
}
else if (Cs >= 1263) {
if (Rs < 2109) {
return 1;
}
else if (Rs >= 2109) {
return 8;
}
}
}
else if (ssDCval >= 87) {
if (Cs < 958) {
if (Rs < 1932) {
return 8;
}
else if (Rs >= 1932) {
return 1;
}
}
else if (Cs >= 958) {
if (TSC < 8) {
if (ssDCval < 90) {
if (Rs < 1863) {
return 8;
}
else if (Rs >= 1863) {
if (Cs < 1275) {
return 8;
}
else if (Cs >= 1275) {
return 2;
}
}
}
else if (ssDCval >= 90) {
return 2;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
else if (ssDCval >= 98) {
if (posBalance < 68) {
if (tcor < 94) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (Rs < 1713) {
return 2;
}
else if (Rs >= 1713) {
return 8;
}
}
}
else if (tcor >= 94) {
if (Rs < 2338) {
if (negBalance < 45) {
if (diffRsCsdiff < 15) {
return 8;
}
else if (diffRsCsdiff >= 15) {
if (Rs < 1720) {
return 8;
}
else if (Rs >= 1720) {
return 2;
}
}
}
else if (negBalance >= 45) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
}
else if (Rs >= 2338) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 8;
}
}
}
}
else if (posBalance >= 68) {
if (SC < 3048) {
if (Cs < 870) {
return 2;
}
else if (Cs >= 870) {
return 1;
}
}
else if (SC >= 3048) {
if (Rs < 1963) {
return 4;
}
else if (Rs >= 1963) {
return 8;
}
}
}
}
}
else if (negBalance >= 49) {
if (MVDiff < 5) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (diffRsCsdiff < -12) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
if (Rs < 1870) {
return 1;
}
else if (Rs >= 1870) {
return 2;
}
}
}
else if (diffRsCsdiff >= -12) {
if (diffAFD < -1) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (posBalance < 52) {
if (Cs < 1173) {
return 1;
}
else if (Cs >= 1173) {
return 8;
}
}
else if (posBalance >= 52) {
return 4;
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (refDCval < 114) {
if (RsCsDiff < 51) {
if (Rs < 812) {
if (ssDCval < 76) {
if (posBalance < 59) {
if (RsCsDiff < 26) {
if (diffMVdiffVal < 7) {
if (Rs < 605) {
if (Rs < 169) {
return 1;
}
else if (Rs >= 169) {
return 2;
}
}
else if (Rs >= 605) {
if (Rs < 736) {
return 4;
}
else if (Rs >= 736) {
return 1;
}
}
}
else if (diffMVdiffVal >= 7) {
if (Cs < 556) {
return 1;
}
else if (Cs >= 556) {
return 4;
}
}
}
else if (RsCsDiff >= 26) {
if (Cs < 836) {
if (diffRsCsdiff < -19) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
return 8;
}
}
else if (diffRsCsdiff >= -19) {
if (diffMVdiffVal < 8) {
if (Cs < 645) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 1;
}
}
else if (Cs >= 645) {
if (Rs < 650) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 4;
}
}
else if (Rs >= 650) {
return 1;
}
}
}
else if (diffMVdiffVal >= 8) {
if (MVDiff < 31) {
return 8;
}
else if (MVDiff >= 31) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
}
else if (Cs >= 836) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
if (Cs < 847) {
return 4;
}
else if (Cs >= 847) {
return 2;
}
}
}
}
}
else if (posBalance >= 59) {
if (AFD < 15) {
if (CsDiff < 3) {
if (Rs < 615) {
return 2;
}
else if (Rs >= 615) {
return 1;
}
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 5) {
if (diffRsCsdiff < -9) {
if (SC < 1162) {
return 8;
}
else if (SC >= 1162) {
return 4;
}
}
else if (diffRsCsdiff >= -9) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
if (mcTcor < 85) {
return 4;
}
else if (mcTcor >= 85) {
return 8;
}
}
}
}
else if (AFD >= 15) {
if (tcor < 55) {
if (ssDCval < 49) {
if (Rs < 492) {
return 2;
}
else if (Rs >= 492) {
return 8;
}
}
else if (ssDCval >= 49) {
if (tcor < 50) {
return 4;
}
else if (tcor >= 50) {
if (Rs < 301) {
return 4;
}
else if (Rs >= 301) {
if (Rs < 330) {
return 2;
}
else if (Rs >= 330) {
return 8;
}
}
}
}
}
else if (tcor >= 55) {
if (RsCsDiff < 37) {
if (MVDiff < 58) {
return 4;
}
else if (MVDiff >= 58) {
if (Rs < 467) {
return 2;
}
else if (Rs >= 467) {
return 1;
}
}
}
else if (RsCsDiff >= 37) {
if (Rs < 721) {
return 2;
}
else if (Rs >= 721) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 76) {
if (ssDCval < 99) {
if (MVDiff < 68) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (SC < 1416) {
return 8;
}
else if (SC >= 1416) {
return 1;
}
}
}
else if (MVDiff >= 68) {
if (refDCval < 99) {
if (refDCval < 83) {
return 1;
}
else if (refDCval >= 83) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 4;
}
}
}
else if (refDCval >= 99) {
return 8;
}
}
}
else if (ssDCval >= 99) {
if (negBalance < 54) {
if (SC < 1264) {
if (Cs < 615) {
if (Rs < 423) {
return 4;
}
else if (Rs >= 423) {
if (Rs < 537) {
return 8;
}
else if (Rs >= 537) {
return 2;
}
}
}
else if (Cs >= 615) {
return 2;
}
}
else if (SC >= 1264) {
if (AFD < 17) {
if (Cs < 642) {
return 2;
}
else if (Cs >= 642) {
return 4;
}
}
else if (AFD >= 17) {
if (Rs < 613) {
return 1;
}
else if (Rs >= 613) {
return 8;
}
}
}
}
else if (negBalance >= 54) {
return 1;
}
}
}
}
else if (Rs >= 812) {
if (ssDCval < 88) {
if (TSC < 10) {
if (SC < 1759) {
if (negBalance < 44) {
if (Rs < 836) {
return 8;
}
else if (Rs >= 836) {
return 1;
}
}
else if (negBalance >= 44) {
if (negBalance < 52) {
if (SC < 1699) {
return 4;
}
else if (SC >= 1699) {
if (Rs < 941) {
return 2;
}
else if (Rs >= 941) {
if (Rs < 1027) {
return 4;
}
else if (Rs >= 1027) {
return 1;
}
}
}
}
else if (negBalance >= 52) {
if (Cs < 690) {
if (mu_mv_mag_sq < 30) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
else if (mu_mv_mag_sq >= 30) {
return 4;
}
}
else if (Cs >= 690) {
if (SC < 1720) {
return 4;
}
else if (SC >= 1720) {
if (Rs < 949) {
return 1;
}
else if (Rs >= 949) {
return 4;
}
}
}
}
}
}
else if (SC >= 1759) {
if (mcTcor < 86) {
if (refDCval < 52) {
if (Cs < 822) {
return 8;
}
else if (Cs >= 822) {
return 4;
}
}
else if (refDCval >= 52) {
return 8;
}
}
else if (mcTcor >= 86) {
if (refDCval < 78) {
if (mu_mv_mag_sq < 7) {
if (TSC < 9) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (MVDiff < 19) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
return 4;
}
}
else if (MVDiff >= 19) {
return 4;
}
}
}
else if (TSC >= 9) {
if (Rs < 1257) {
if (Rs < 875) {
return 8;
}
else if (Rs >= 875) {
return 2;
}
}
else if (Rs >= 1257) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (negBalance < 42) {
if (Rs < 1314) {
return 8;
}
else if (Rs >= 1314) {
if (Cs < 959) {
return 2;
}
else if (Cs >= 959) {
return 1;
}
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
else if (refDCval >= 78) {
if (Cs < 903) {
if (Rs < 1409) {
if (Rs < 1033) {
return 8;
}
else if (Rs >= 1033) {
return 2;
}
}
else if (Rs >= 1409) {
return 1;
}
}
else if (Cs >= 903) {
return 4;
}
}
}
}
}
else if (TSC >= 10) {
if (Cs < 558) {
return 4;
}
else if (Cs >= 558) {
if (diffRsCsdiff < -19) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
return 2;
}
}
else if (diffRsCsdiff >= -19) {
if (mu_mv_mag_sq < 49) {
return 8;
}
else if (mu_mv_mag_sq >= 49) {
if (Rs < 1232) {
return 1;
}
else if (Rs >= 1232) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (AFD < 15) {
if (tcor < 77) {
if (Rs < 871) {
return 1;
}
else if (Rs >= 871) {
return 8;
}
}
else if (tcor >= 77) {
if (TSC < 9) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 4;
}
}
else if (TSC >= 9) {
if (Cs < 915) {
return 2;
}
else if (Cs >= 915) {
return 8;
}
}
}
}
else if (AFD >= 15) {
if (mu_mv_mag_sq < 2) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (Rs < 1564) {
return 2;
}
else if (Rs >= 1564) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (Rs < 863) {
return 1;
}
else if (Rs >= 863) {
return 8;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 51) {
if (negBalance < 45) {
if (mcTcor < 84) {
if (diffMVdiffVal < 3) {
if (tcor < 40) {
return 4;
}
else if (tcor >= 40) {
if (Cs < 837) {
return 4;
}
else if (Cs >= 837) {
if (Rs < 1393) {
return 1;
}
else if (Rs >= 1393) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 3) {
if (ssDCval < 57) {
if (Cs < 1177) {
return 2;
}
else if (Cs >= 1177) {
return 1;
}
}
else if (ssDCval >= 57) {
if (diffRsCsdiff < 79) {
if (SC < 1438) {
return 4;
}
else if (SC >= 1438) {
if (AFD < 40) {
return 8;
}
else if (AFD >= 40) {
return 1;
}
}
}
else if (diffRsCsdiff >= 79) {
if (Rs < 1349) {
if (Rs < 892) {
return 2;
}
else if (Rs >= 892) {
return 4;
}
}
else if (Rs >= 1349) {
if (Rs < 1519) {
return 8;
}
else if (Rs >= 1519) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 84) {
if (Cs < 817) {
if (RsCsDiff < 110) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Cs < 736) {
if (diffMVdiffVal < 3) {
if (CsDiff < 4) {
if (ssDCval < 78) {
if (Rs < 2140) {
return 8;
}
else if (Rs >= 2140) {
return 4;
}
}
else if (ssDCval >= 78) {
return 1;
}
}
else if (CsDiff >= 4) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
}
else if (diffMVdiffVal >= 3) {
if (Cs < 576) {
return 8;
}
else if (Cs >= 576) {
if (Cs < 650) {
return 4;
}
else if (Cs >= 650) {
return 8;
}
}
}
}
else if (Cs >= 736) {
return 1;
}
}
}
else if (RsCsDiff >= 110) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
if (Rs < 1952) {
return 1;
}
else if (Rs >= 1952) {
return 4;
}
}
}
}
}
else if (Cs >= 817) {
if (posBalance < 49) {
if (refDCval < 72) {
if (RsDiff < 14) {
return 2;
}
else if (RsDiff >= 14) {
return 1;
}
}
else if (refDCval >= 72) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 61) {
if (diffRsCsdiff < 17) {
return 2;
}
else if (diffRsCsdiff >= 17) {
return 4;
}
}
else if (posBalance >= 61) {
if (Rs < 2020) {
return 2;
}
else if (Rs >= 2020) {
if (Rs < 2184) {
return 8;
}
else if (Rs >= 2184) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (MVDiff < 27) {
if (posBalance < 67) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
else if (posBalance >= 67) {
if (Rs < 1262) {
return 1;
}
else if (Rs >= 1262) {
if (Rs < 1718) {
return 8;
}
else if (Rs >= 1718) {
return 2;
}
}
}
}
else if (MVDiff >= 27) {
if (Cs < 957) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 8;
}
}
else if (Cs >= 957) {
if (mu_mv_mag_sq < 27) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (Rs < 1483) {
return 4;
}
else if (Rs >= 1483) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 27) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (RsDiff < 12) {
if (SC < 2443) {
if (MVDiff < 10) {
if (diffAFD < -1) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (posBalance < 58) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (posBalance >= 58) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 8;
}
}
}
}
else if (MVDiff >= 10) {
if (mu_mv_mag_sq < 5) {
if (mcTcor < 86) {
if (MVDiff < 44) {
if (AFD < 16) {
if (tcor < 56) {
return 4;
}
else if (tcor >= 56) {
if (refDCval < 48) {
return 2;
}
else if (refDCval >= 48) {
if (Rs < 1519) {
return 4;
}
else if (Rs >= 1519) {
return 2;
}
}
}
}
else if (AFD >= 16) {
if (Rs < 1192) {
if (diffRsCsdiff < 17) {
return 2;
}
else if (diffRsCsdiff >= 17) {
return 1;
}
}
else if (Rs >= 1192) {
if (Rs < 1291) {
if (Rs < 1214) {
return 8;
}
else if (Rs >= 1214) {
return 2;
}
}
else if (Rs >= 1291) {
return 4;
}
}
}
}
else if (MVDiff >= 44) {
if (mu_mv_mag_sq < 3) {
if (diffMVdiffVal < -12) {
if (Rs < 691) {
return 1;
}
else if (Rs >= 691) {
if (Rs < 1120) {
return 8;
}
else if (Rs >= 1120) {
return 2;
}
}
}
else if (diffMVdiffVal >= -12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1380) {
return 4;
}
else if (Rs >= 1380) {
return 8;
}
}
}
}
else if (mcTcor >= 86) {
if (CsDiff < 6) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 3) {
if (negBalance < 73) {
return 2;
}
else if (negBalance >= 73) {
if (Rs < 1164) {
return 1;
}
else if (Rs >= 1164) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (posBalance >= 54) {
if (ssDCval < 76) {
if (negBalance < 52) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (negBalance >= 52) {
if (Cs < 868) {
return 8;
}
else if (Cs >= 868) {
if (Rs < 1332) {
return 1;
}
else if (Rs >= 1332) {
return 8;
}
}
}
}
else if (ssDCval >= 76) {
if (MVDiff < 77) {
if (MVDiff < 25) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
return 4;
}
}
else if (MVDiff >= 25) {
return 1;
}
}
else if (MVDiff >= 77) {
if (Rs < 1442) {
return 2;
}
else if (Rs >= 1442) {
return 8;
}
}
}
}
}
else if (CsDiff >= 6) {
if (refDCval < 68) {
if (tcor < 67) {
if (refDCval < 62) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
return 1;
}
}
else if (refDCval >= 62) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 2;
}
}
}
else if (tcor >= 67) {
if (ssDCval < 56) {
if (Rs < 1022) {
return 2;
}
else if (Rs >= 1022) {
return 8;
}
}
else if (ssDCval >= 56) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 8;
}
}
}
}
else if (refDCval >= 68) {
if (diffAFD < -2) {
if (diffRsCsdiff < -29) {
if (Rs < 850) {
return 4;
}
else if (Rs >= 850) {
if (Rs < 897) {
return 1;
}
else if (Rs >= 897) {
return 8;
}
}
}
else if (diffRsCsdiff >= -29) {
if (MVDiff < 26) {
return 2;
}
else if (MVDiff >= 26) {
if (TSC < 10) {
if (MVDiff < 80) {
return 1;
}
else if (MVDiff >= 80) {
return 2;
}
}
else if (TSC >= 10) {
return 2;
}
}
}
}
else if (diffAFD >= -2) {
if (refDCval < 96) {
if (Rs < 1210) {
if (gchDC < 8) {
if (diffRsCsdiff < 36) {
if (CsDiff < 7) {
if (mcTcor < 92) {
if (Cs < 1212) {
if (RsDiff < 7) {
return 2;
}
else if (RsDiff >= 7) {
return 4;
}
}
else if (Cs >= 1212) {
return 4;
}
}
else if (mcTcor >= 92) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
return 2;
}
}
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (diffRsCsdiff >= 36) {
return 4;
}
}
else if (gchDC >= 8) {
return 4;
}
}
else if (Rs >= 1210) {
if (SC < 2331) {
if (Rs < 1221) {
return 8;
}
else if (Rs >= 1221) {
return 4;
}
}
else if (SC >= 2331) {
if (Rs < 1384) {
return 2;
}
else if (Rs >= 1384) {
return 8;
}
}
}
}
else if (refDCval >= 96) {
if (mcTcor < 93) {
if (Rs < 1349) {
return 8;
}
else if (Rs >= 1349) {
return 2;
}
}
else if (mcTcor >= 93) {
if (Rs < 1121) {
return 4;
}
else if (Rs >= 1121) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (tcor < 49) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (TSC < 9) {
if (diffRsCsdiff < 35) {
if (Cs < 998) {
return 4;
}
else if (Cs >= 998) {
return 2;
}
}
else if (diffRsCsdiff >= 35) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
}
else if (TSC >= 9) {
if (refDCval < 53) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (refDCval >= 53) {
if (ssDCval < 90) {
if (posBalance < 59) {
if (Rs < 751) {
return 8;
}
else if (Rs >= 751) {
if (MVDiff < 57) {
return 4;
}
else if (MVDiff >= 57) {
return 1;
}
}
}
else if (posBalance >= 59) {
if (tcor < 41) {
if (diffRsCsdiff < 32) {
return 8;
}
else if (diffRsCsdiff >= 32) {
return 2;
}
}
else if (tcor >= 41) {
return 2;
}
}
}
else if (ssDCval >= 90) {
if (Cs < 1142) {
return 4;
}
else if (Cs >= 1142) {
return 1;
}
}
}
}
}
}
else if (tcor >= 49) {
if (AFD < 26) {
if (Rs < 1000) {
if (Rs < 630) {
if (ssDCval < 85) {
if (diffMVdiffVal < -7) {
return 4;
}
else if (diffMVdiffVal >= -7) {
if (Rs < 545) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
return 2;
}
}
else if (Rs >= 545) {
if (Cs < 1221) {
return 4;
}
else if (Cs >= 1221) {
return 8;
}
}
}
}
else if (ssDCval >= 85) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 1;
}
}
}
else if (Rs >= 630) {
if (ssDCval < 76) {
if (tcor < 70) {
if (posBalance < 62) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Rs < 733) {
return 8;
}
else if (Rs >= 733) {
return 2;
}
}
}
else if (posBalance >= 62) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 769) {
return 4;
}
else if (Rs >= 769) {
return 1;
}
}
}
}
else if (tcor >= 70) {
if (Rs < 801) {
if (Rs < 768) {
return 1;
}
else if (Rs >= 768) {
return 8;
}
}
else if (Rs >= 801) {
return 1;
}
}
}
else if (ssDCval >= 76) {
if (AFD < 20) {
if (diffAFD < -3) {
if (Rs < 908) {
return 2;
}
else if (Rs >= 908) {
return 8;
}
}
else if (diffAFD >= -3) {
if (TSC < 7) {
if (Cs < 870) {
return 2;
}
else if (Cs >= 870) {
return 8;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (AFD >= 20) {
if (RsDiff < 7) {
if (negBalance < 53) {
if (Rs < 883) {
return 2;
}
else if (Rs >= 883) {
return 4;
}
}
else if (negBalance >= 53) {
if (Rs < 652) {
return 1;
}
else if (Rs >= 652) {
return 8;
}
}
}
else if (RsDiff >= 7) {
if (ssDCval < 100) {
return 1;
}
else if (ssDCval >= 100) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1000) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 19) {
if (diffAFD < -1) {
if (SC < 1948) {
return 2;
}
else if (SC >= 1948) {
return 8;
}
}
else if (diffAFD >= -1) {
if (SC < 2265) {
if (Rs < 1100) {
return 4;
}
else if (Rs >= 1100) {
return 8;
}
}
else if (SC >= 2265) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 19) {
if (ssDCval < 68) {
if (Cs < 791) {
return 1;
}
else if (Cs >= 791) {
return 4;
}
}
else if (ssDCval >= 68) {
if (TSC < 8) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
else if (TSC >= 8) {
return 1;
}
}
}
}
else if (posBalance >= 55) {
if (Cs < 1090) {
if (negBalance < 57) {
if (tcor < 56) {
return 4;
}
else if (tcor >= 56) {
if (ssDCval < 87) {
if (SC < 2383) {
if (tcor < 68) {
return 8;
}
else if (tcor >= 68) {
if (ssDCval < 85) {
if (posBalance < 66) {
return 4;
}
else if (posBalance >= 66) {
return 1;
}
}
else if (ssDCval >= 85) {
if (Rs < 1400) {
return 8;
}
else if (Rs >= 1400) {
return 4;
}
}
}
}
else if (SC >= 2383) {
if (Rs < 1498) {
return 1;
}
else if (Rs >= 1498) {
return 8;
}
}
}
else if (ssDCval >= 87) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (TSC < 10) {
if (Rs < 1465) {
return 8;
}
else if (Rs >= 1465) {
return 4;
}
}
else if (TSC >= 10) {
if (Rs < 1618) {
return 1;
}
else if (Rs >= 1618) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 57) {
if (Cs < 859) {
if (diffMVdiffVal < -6) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 4;
}
}
else if (diffMVdiffVal >= -6) {
if (RsCsDiff < 155) {
return 4;
}
else if (RsCsDiff >= 155) {
return 8;
}
}
}
else if (Cs >= 859) {
return 8;
}
}
}
else if (Cs >= 1090) {
if (Rs < 1086) {
return 1;
}
else if (Rs >= 1086) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
}
else if (AFD >= 26) {
if (MVDiff < 135) {
if (Cs < 688) {
if (RsCsDiff < 117) {
if (ssDCval < 87) {
return 4;
}
else if (ssDCval >= 87) {
return 1;
}
}
else if (RsCsDiff >= 117) {
if (Rs < 766) {
return 8;
}
else if (Rs >= 766) {
return 4;
}
}
}
else if (Cs >= 688) {
if (negBalance < 65) {
if (tcor < 51) {
return 2;
}
else if (tcor >= 51) {
if (gchDC < 3) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (tcor < 63) {
if (ssDCval < 88) {
return 4;
}
else if (ssDCval >= 88) {
return 1;
}
}
else if (tcor >= 63) {
return 4;
}
}
}
else if (gchDC >= 3) {
if (Rs < 649) {
return 4;
}
else if (Rs >= 649) {
return 2;
}
}
}
}
else if (negBalance >= 65) {
if (Rs < 434) {
return 4;
}
else if (Rs >= 434) {
return 1;
}
}
}
}
else if (MVDiff >= 135) {
if (refDCval < 91) {
return 8;
}
else if (refDCval >= 91) {
if (Rs < 1240) {
return 1;
}
else if (Rs >= 1240) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 2443) {
if (Cs < 872) {
if (MVDiff < 27) {
if (ssDCval < 77) {
if (refDCval < 56) {
return 4;
}
else if (refDCval >= 56) {
return 2;
}
}
else if (ssDCval >= 77) {
if (posBalance < 56) {
if (Cs < 762) {
return 2;
}
else if (Cs >= 762) {
return 8;
}
}
else if (posBalance >= 56) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (Rs < 1724) {
return 4;
}
else if (Rs >= 1724) {
return 1;
}
}
}
}
}
else if (MVDiff >= 27) {
if (posBalance < 58) {
if (Rs < 1648) {
return 2;
}
else if (Rs >= 1648) {
return 8;
}
}
else if (posBalance >= 58) {
if (mcTcor < 90) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
return 1;
}
}
else if (mcTcor >= 90) {
if (Rs < 2097) {
return 2;
}
else if (Rs >= 2097) {
return 8;
}
}
}
}
}
else if (Cs >= 872) {
if (AFD < 19) {
if (TSC < 10) {
if (Rs < 1818) {
if (gchDC < 1) {
if (negBalance < 53) {
if (tcor < 81) {
if (MVDiff < 11) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (Rs < 1508) {
return 4;
}
else if (Rs >= 1508) {
return 2;
}
}
}
else if (MVDiff >= 11) {
return 2;
}
}
else if (tcor >= 81) {
if (Cs < 1266) {
return 8;
}
else if (Cs >= 1266) {
return 4;
}
}
}
else if (negBalance >= 53) {
if (Rs < 1713) {
if (diffRsCsdiff < 22) {
if (MVDiff < 43) {
return 8;
}
else if (MVDiff >= 43) {
return 4;
}
}
else if (diffRsCsdiff >= 22) {
return 4;
}
}
else if (Rs >= 1713) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (MVDiff < 6) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 2) {
if (AFD < 16) {
if (refDCval < 67) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (refDCval >= 67) {
if (Cs < 1115) {
return 1;
}
else if (Cs >= 1115) {
return 8;
}
}
}
else if (AFD >= 16) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsDiff < 10) {
if (diffAFD < -3) {
if (Rs < 1566) {
return 1;
}
else if (Rs >= 1566) {
return 8;
}
}
else if (diffAFD >= -3) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 1470) {
return 8;
}
else if (Rs >= 1470) {
return 1;
}
}
}
}
else if (RsDiff >= 10) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1818) {
if (negBalance < 60) {
if (negBalance < 53) {
if (CsDiff < 6) {
if (Cs < 881) {
return 8;
}
else if (Cs >= 881) {
return 4;
}
}
else if (CsDiff >= 6) {
if (refDCval < 94) {
return 8;
}
else if (refDCval >= 94) {
if (Rs < 1868) {
return 8;
}
else if (Rs >= 1868) {
return 2;
}
}
}
}
else if (negBalance >= 53) {
if (tcor < 80) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 4;
}
}
else if (tcor >= 80) {
if (Rs < 2170) {
return 4;
}
else if (Rs >= 2170) {
return 8;
}
}
}
}
else if (negBalance >= 60) {
if (Rs < 2044) {
if (Cs < 1076) {
return 8;
}
else if (Cs >= 1076) {
return 2;
}
}
else if (Rs >= 2044) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (Cs < 995) {
if (Rs < 1591) {
return 8;
}
else if (Rs >= 1591) {
return 1;
}
}
else if (Cs >= 995) {
if (RsCsDiff < 115) {
if (mu_mv_mag_sq < 56) {
return 8;
}
else if (mu_mv_mag_sq >= 56) {
return 1;
}
}
else if (RsCsDiff >= 115) {
if (mu_mv_mag_sq < 7) {
if (tcor < 76) {
return 4;
}
else if (tcor >= 76) {
if (Rs < 1794) {
return 8;
}
else if (Rs >= 1794) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
}
}
}
else if (AFD >= 19) {
if (mu_mv_mag_sq < 30) {
if (RsCsDiff < 129) {
if (mcTcor < 90) {
if (tcor < 58) {
return 8;
}
else if (tcor >= 58) {
if (Cs < 1041) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
if (Cs < 890) {
return 1;
}
else if (Cs >= 890) {
return 8;
}
}
}
else if (Cs >= 1041) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
else if (mcTcor >= 90) {
if (refDCval < 98) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 1;
}
}
else if (refDCval >= 98) {
if (negBalance < 59) {
return 4;
}
else if (negBalance >= 59) {
if (Rs < 1903) {
return 8;
}
else if (Rs >= 1903) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 129) {
if (AFD < 23) {
if (refDCval < 76) {
if (MVDiff < 17) {
if (Rs < 1882) {
return 8;
}
else if (Rs >= 1882) {
return 2;
}
}
else if (MVDiff >= 17) {
if (negBalance < 59) {
if (Cs < 1042) {
if (Rs < 1573) {
return 2;
}
else if (Rs >= 1573) {
return 4;
}
}
else if (Cs >= 1042) {
return 4;
}
}
else if (negBalance >= 59) {
return 8;
}
}
}
else if (refDCval >= 76) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Cs < 1075) {
return 8;
}
else if (Cs >= 1075) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 2;
}
}
}
}
}
}
else if (AFD >= 23) {
if (SC < 2921) {
if (gchDC < 4) {
if (SC < 2842) {
return 4;
}
else if (SC >= 2842) {
if (Rs < 1659) {
return 4;
}
else if (Rs >= 1659) {
return 8;
}
}
}
else if (gchDC >= 4) {
return 8;
}
}
else if (SC >= 2921) {
if (Rs < 1856) {
return 2;
}
else if (Rs >= 1856) {
if (Rs < 2156) {
return 4;
}
else if (Rs >= 2156) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (MVDiff < 35) {
return 2;
}
else if (MVDiff >= 35) {
if (Cs < 988) {
return 2;
}
else if (Cs >= 988) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 12) {
if (refDCval < 86) {
if (mcTcor < 85) {
if (Rs < 1818) {
return 4;
}
else if (Rs >= 1818) {
if (Rs < 2296) {
if (Cs < 639) {
return 1;
}
else if (Cs >= 639) {
return 8;
}
}
else if (Rs >= 2296) {
return 2;
}
}
}
else if (mcTcor >= 85) {
if (Cs < 883) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
if (Rs < 1418) {
return 2;
}
else if (Rs >= 1418) {
if (Rs < 1634) {
return 4;
}
else if (Rs >= 1634) {
return 8;
}
}
}
}
else if (Cs >= 883) {
if (RsCsDiff < 255) {
if (mu_mv_mag_sq < 20) {
if (Rs < 1703) {
if (Rs < 1407) {
return 8;
}
else if (Rs >= 1407) {
return 2;
}
}
else if (Rs >= 1703) {
if (Rs < 2194) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (Rs >= 2194) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 994) {
return 2;
}
else if (Cs >= 994) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 20) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
if (Rs < 1930) {
return 4;
}
else if (Rs >= 1930) {
return 1;
}
}
}
}
else if (RsCsDiff >= 255) {
if (AFD < 31) {
return 2;
}
else if (AFD >= 31) {
return 4;
}
}
}
}
}
else if (refDCval >= 86) {
if (TSC < 14) {
if (posBalance < 57) {
if (RsCsDiff < 200) {
if (Rs < 1606) {
return 4;
}
else if (Rs >= 1606) {
return 8;
}
}
else if (RsCsDiff >= 200) {
if (posBalance < 47) {
if (Rs < 2369) {
return 1;
}
else if (Rs >= 2369) {
return 8;
}
}
else if (posBalance >= 47) {
return 2;
}
}
}
else if (posBalance >= 57) {
if (negBalance < 54) {
if (Cs < 877) {
if (Rs < 2208) {
return 2;
}
else if (Rs >= 2208) {
return 4;
}
}
else if (Cs >= 877) {
return 8;
}
}
else if (negBalance >= 54) {
if (tcor < 58) {
return 4;
}
else if (tcor >= 58) {
if (mu_mv_mag_sq < 9) {
if (Rs < 2285) {
return 1;
}
else if (Rs >= 2285) {
return 2;
}
}
else if (mu_mv_mag_sq >= 9) {
if (posBalance < 60) {
if (gchDC < 2) {
if (Rs < 1646) {
return 2;
}
else if (Rs >= 1646) {
return 1;
}
}
else if (gchDC >= 2) {
if (Cs < 1014) {
return 4;
}
else if (Cs >= 1014) {
return 2;
}
}
}
else if (posBalance >= 60) {
return 4;
}
}
}
}
}
}
else if (TSC >= 14) {
if (refDCval < 96) {
return 1;
}
else if (refDCval >= 96) {
if (mu_mv_mag_sq < 26) {
return 2;
}
else if (mu_mv_mag_sq >= 26) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (MVDiff < 25) {
if (tcor < 84) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
if (negBalance < 53) {
if (RsDiff < 11) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (ssDCval < 152) {
return 1;
}
else if (ssDCval >= 152) {
return 8;
}
}
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (negBalance >= 53) {
if (posBalance < 62) {
if (Cs < 856) {
if (CsDiff < 5) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (Cs >= 856) {
if (RsCsDiff < 201) {
return 8;
}
else if (RsCsDiff >= 201) {
if (Rs < 1941) {
return 2;
}
else if (Rs >= 1941) {
return 8;
}
}
}
}
else if (posBalance >= 62) {
if (negBalance < 57) {
if (Rs < 1872) {
return 4;
}
else if (Rs >= 1872) {
return 8;
}
}
else if (negBalance >= 57) {
if (Cs < 846) {
return 8;
}
else if (Cs >= 846) {
return 2;
}
}
}
}
}
}
else if (tcor >= 84) {
if (MVDiff < 12) {
if (ssDCval < 124) {
if (Rs < 1203) {
return 2;
}
else if (Rs >= 1203) {
return 8;
}
}
else if (ssDCval >= 124) {
if (Cs < 1159) {
return 2;
}
else if (Cs >= 1159) {
return 8;
}
}
}
else if (MVDiff >= 12) {
if (negBalance < 52) {
if (Rs < 1882) {
if (TSC < 10) {
if (MVDiff < 13) {
if (Rs < 1091) {
return 2;
}
else if (Rs >= 1091) {
return 8;
}
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (TSC >= 10) {
return 1;
}
}
else if (Rs >= 1882) {
return 1;
}
}
else if (negBalance >= 52) {
if (TSC < 8) {
if (Rs < 1629) {
return 2;
}
else if (Rs >= 1629) {
return 1;
}
}
else if (TSC >= 8) {
if (ssDCval < 112) {
return 8;
}
else if (ssDCval >= 112) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 25) {
if (SC < 1568) {
if (MVDiff < 126) {
if (gchDC < 4) {
if (posBalance < 71) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 1;
}
}
else if (posBalance >= 71) {
return 8;
}
}
else if (gchDC >= 4) {
return 2;
}
}
else if (MVDiff >= 126) {
if (diffMVdiffVal < 27) {
return 1;
}
else if (diffMVdiffVal >= 27) {
if (Rs < 628) {
return 8;
}
else if (Rs >= 628) {
return 4;
}
}
}
}
else if (SC >= 1568) {
if (refDCval < 142) {
if (RsCsDiff < 87) {
if (tcor < 50) {
return 1;
}
else if (tcor >= 50) {
if (SC < 2430) {
if (AFD < 14) {
if (Rs < 1112) {
return 8;
}
else if (Rs >= 1112) {
return 4;
}
}
else if (AFD >= 14) {
return 8;
}
}
else if (SC >= 2430) {
if (TSC < 12) {
return 1;
}
else if (TSC >= 12) {
return 8;
}
}
}
}
else if (RsCsDiff >= 87) {
if (tcor < 79) {
if (tcor < 43) {
if (AFD < 36) {
return 1;
}
else if (AFD >= 36) {
if (Rs < 925) {
return 1;
}
else if (Rs >= 925) {
return 2;
}
}
}
else if (tcor >= 43) {
if (mcTcor < 88) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
if (negBalance < 52) {
if (refDCval < 128) {
if (Rs < 1545) {
return 8;
}
else if (Rs >= 1545) {
return 4;
}
}
else if (refDCval >= 128) {
return 1;
}
}
else if (negBalance >= 52) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
}
}
else if (mcTcor >= 88) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Cs < 611) {
return 1;
}
else if (Cs >= 611) {
if (gchDC < 5) {
if (Cs < 631) {
return 2;
}
else if (Cs >= 631) {
return 1;
}
}
else if (gchDC >= 5) {
return 2;
}
}
}
}
}
}
else if (tcor >= 79) {
if (posBalance < 46) {
if (Rs < 2335) {
if (Cs < 1099) {
return 2;
}
else if (Cs >= 1099) {
return 8;
}
}
else if (Rs >= 2335) {
return 1;
}
}
else if (posBalance >= 46) {
return 1;
}
}
}
}
else if (refDCval >= 142) {
if (negBalance < 55) {
if (MVDiff < 32) {
return 1;
}
else if (MVDiff >= 32) {
if (diffAFD < 3) {
if (SC < 3137) {
return 8;
}
else if (SC >= 3137) {
if (Rs < 2328) {
return 2;
}
else if (Rs >= 2328) {
return 1;
}
}
}
else if (diffAFD >= 3) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
if (Rs < 1103) {
return 1;
}
else if (Rs >= 1103) {
return 8;
}
}
}
}
}
else if (negBalance >= 55) {
if (mcTcor < 87) {
if (CsDiff < 5) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 1;
}
}
else if (CsDiff >= 5) {
if (diffMVdiffVal < -50) {
if (Rs < 1882) {
return 1;
}
else if (Rs >= 1882) {
return 8;
}
}
else if (diffMVdiffVal >= -50) {
return 4;
}
}
}
else if (mcTcor >= 87) {
if (diffMVdiffVal < -9) {
if (SC < 2161) {
return 2;
}
else if (SC >= 2161) {
return 1;
}
}
else if (diffMVdiffVal >= -9) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (ssDCval < 82) {
if (Cs < 1006) {
if (mu_mv_mag_sq < 122) {
if (Rs < 872) {
if (TSCindex < 2) {
if (negBalance < 19) {
return 1;
}
else if (negBalance >= 19) {
if (Rs < 857) {
return 8;
}
else if (Rs >= 857) {
return 1;
}
}
}
else if (TSCindex >= 2) {
if (MVDiff < 1) {
if (Rs < 524) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (gchDC < 1) {
if (ssDCval < 71) {
if (Rs < 463) {
return 2;
}
else if (Rs >= 463) {
if (Rs < 475) {
return 4;
}
else if (Rs >= 475) {
return 2;
}
}
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
else if (Rs >= 524) {
if (refDCval < 64) {
if (refDCval < 54) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
return 4;
}
}
else if (refDCval >= 54) {
return 8;
}
}
else if (refDCval >= 64) {
if (Rs < 803) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (Rs < 637) {
return 8;
}
else if (Rs >= 637) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (Rs < 767) {
return 1;
}
else if (Rs >= 767) {
return 4;
}
}
}
}
}
}
else if (Rs >= 803) {
if (mu_mv_mag_sq < 92) {
return 8;
}
else if (mu_mv_mag_sq >= 92) {
if (Cs < 639) {
return 2;
}
else if (Cs >= 639) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (refDCval < 68) {
if (AFD < 9) {
if (mu_mv_mag_sq < 79) {
if (tcor < 95) {
if (posBalance < 51) {
if (ssDCval < 58) {
if (Cs < 620) {
if (Rs < 389) {
return 4;
}
else if (Rs >= 389) {
return 1;
}
}
else if (Cs >= 620) {
if (negBalance < 36) {
return 4;
}
else if (negBalance >= 36) {
if (Rs < 554) {
return 1;
}
else if (Rs >= 554) {
return 8;
}
}
}
}
else if (ssDCval >= 58) {
if (Cs < 889) {
return 1;
}
else if (Cs >= 889) {
return 8;
}
}
}
else if (posBalance >= 51) {
if (RsCsDiff < 12) {
if (posBalance < 62) {
if (Rs < 637) {
return 1;
}
else if (Rs >= 637) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (posBalance >= 62) {
return 2;
}
}
else if (RsCsDiff >= 12) {
return 4;
}
}
}
else if (tcor >= 95) {
if (Rs < 761) {
if (posBalance < 61) {
if (diffRsCsdiff < 7) {
if (diffMVdiffVal < -6) {
if (Cs < 555) {
return 1;
}
else if (Cs >= 555) {
return 4;
}
}
else if (diffMVdiffVal >= -6) {
return 4;
}
}
else if (diffRsCsdiff >= 7) {
if (Rs < 692) {
return 4;
}
else if (Rs >= 692) {
return 1;
}
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (Rs >= 761) {
if (Cs < 721) {
if (AFD < 4) {
if (Rs < 787) {
return 2;
}
else if (Rs >= 787) {
return 1;
}
}
else if (AFD >= 4) {
return 8;
}
}
else if (Cs >= 721) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (mu_mv_mag_sq < 106) {
if (Cs < 589) {
if (tcor < 98) {
if (Rs < 662) {
return 1;
}
else if (Rs >= 662) {
return 2;
}
}
else if (tcor >= 98) {
if (posBalance < 53) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
return 4;
}
}
else if (posBalance >= 53) {
return 1;
}
}
}
else if (Cs >= 589) {
if (SC < 1566) {
if (TSC < 4) {
if (RsDiff < 1) {
if (ssDCval < 53) {
if (Rs < 537) {
return 4;
}
else if (Rs >= 537) {
return 1;
}
}
else if (ssDCval >= 53) {
if (negBalance < 19) {
return 4;
}
else if (negBalance >= 19) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (posBalance < 43) {
if (ssDCval < 59) {
if (TSC < 3) {
if (Cs < 732) {
return 2;
}
else if (Cs >= 732) {
if (Rs < 660) {
return 1;
}
else if (Rs >= 660) {
return 8;
}
}
}
else if (TSC >= 3) {
return 1;
}
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (posBalance >= 43) {
if (AFD < 4) {
if (diffMVdiffVal < 1) {
return 1;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 422) {
return 1;
}
else if (Rs >= 422) {
return 8;
}
}
}
else if (AFD >= 4) {
if (ssDCval < 66) {
if (Rs < 844) {
return 8;
}
else if (Rs >= 844) {
return 1;
}
}
else if (ssDCval >= 66) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (SC >= 1566) {
if (ssDCval < 53) {
if (Rs < 728) {
return 1;
}
else if (Rs >= 728) {
return 8;
}
}
else if (ssDCval >= 53) {
if (Cs < 929) {
return 4;
}
else if (Cs >= 929) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (MVDiff < 3) {
if (Rs < 696) {
if (negBalance < 26) {
if (ssDCval < 60) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
return 2;
}
}
else if (ssDCval >= 60) {
if (Rs < 361) {
return 8;
}
else if (Rs >= 361) {
return 4;
}
}
}
else if (negBalance >= 26) {
return 1;
}
}
else if (Rs >= 696) {
if (Rs < 763) {
return 4;
}
else if (Rs >= 763) {
if (Cs < 563) {
return 4;
}
else if (Cs >= 563) {
return 2;
}
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (diffRsCsdiff < -4) {
return 2;
}
else if (diffRsCsdiff >= -4) {
return 4;
}
}
}
}
}
}
else if (AFD >= 9) {
if (Rs < 269) {
return 1;
}
else if (Rs >= 269) {
if (negBalance < 44) {
if (CsDiff < 4) {
if (SC < 1381) {
return 1;
}
else if (SC >= 1381) {
if (Rs < 743) {
return 2;
}
else if (Rs >= 743) {
return 4;
}
}
}
else if (CsDiff >= 4) {
if (RsCsDiff < 68) {
return 2;
}
else if (RsCsDiff >= 68) {
return 1;
}
}
}
else if (negBalance >= 44) {
if (diffRsCsdiff < -14) {
if (Cs < 906) {
return 2;
}
else if (Cs >= 906) {
return 1;
}
}
else if (diffRsCsdiff >= -14) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 68) {
if (Rs < 775) {
if (Cs < 839) {
if (posBalance < 49) {
if (MVDiff < 11) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (Rs < 637) {
if (Rs < 512) {
return 4;
}
else if (Rs >= 512) {
return 8;
}
}
else if (Rs >= 637) {
if (Rs < 767) {
return 2;
}
else if (Rs >= 767) {
return 1;
}
}
}
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (posBalance >= 49) {
if (posBalance < 56) {
if (mu_mv_mag_sq < 75) {
if (Rs < 559) {
return 8;
}
else if (Rs >= 559) {
return 4;
}
}
else if (mu_mv_mag_sq >= 75) {
return 8;
}
}
else if (posBalance >= 56) {
if (TSC < 4) {
if (Cs < 664) {
return 4;
}
else if (Cs >= 664) {
return 1;
}
}
else if (TSC >= 4) {
if (Rs < 650) {
return 8;
}
else if (Rs >= 650) {
return 4;
}
}
}
}
}
else if (Cs >= 839) {
if (diffMVdiffVal < 1) {
if (gchDC < 1) {
if (Rs < 683) {
return 2;
}
else if (Rs >= 683) {
return 8;
}
}
else if (gchDC >= 1) {
if (Cs < 981) {
return 1;
}
else if (Cs >= 981) {
return 2;
}
}
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
}
else if (Rs >= 775) {
if (tcor < 97) {
if (RsCsDiff < 16) {
return 4;
}
else if (RsCsDiff >= 16) {
if (Rs < 861) {
return 2;
}
else if (Rs >= 861) {
return 8;
}
}
}
else if (tcor >= 97) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 872) {
if (Rs < 1786) {
if (refDCval < 69) {
if (SC < 2441) {
if (negBalance < 42) {
if (SC < 1616) {
if (refDCval < 65) {
if (tcor < 99) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (diffAFD < -1) {
if (Rs < 962) {
return 8;
}
else if (Rs >= 962) {
return 1;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (tcor >= 99) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 2;
}
}
}
else if (refDCval >= 65) {
return 2;
}
}
else if (SC >= 1616) {
if (MVDiff < 12) {
if (TSC < 4) {
if (ssDCval < 51) {
if (Rs < 1201) {
if (Cs < 592) {
return 1;
}
else if (Cs >= 592) {
return 8;
}
}
else if (Rs >= 1201) {
return 1;
}
}
else if (ssDCval >= 51) {
if (Rs < 1358) {
if (diffMVdiffVal < -10) {
return 8;
}
else if (diffMVdiffVal >= -10) {
if (posBalance < 58) {
if (SC < 1766) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
if (Cs < 666) {
return 1;
}
else if (Cs >= 666) {
return 2;
}
}
}
else if (SC >= 1766) {
return 1;
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
else if (Rs >= 1358) {
if (Cs < 771) {
if (Rs < 1364) {
return 8;
}
else if (Rs >= 1364) {
if (diffRsCsdiff < -7) {
if (Rs < 1640) {
return 1;
}
else if (Rs >= 1640) {
return 2;
}
}
else if (diffRsCsdiff >= -7) {
return 1;
}
}
}
else if (Cs >= 771) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
return 8;
}
}
}
}
}
else if (TSC >= 4) {
if (ssDCval < 67) {
if (negBalance < 28) {
if (Rs < 1001) {
return 4;
}
else if (Rs >= 1001) {
if (Cs < 906) {
return 1;
}
else if (Cs >= 906) {
return 8;
}
}
}
else if (negBalance >= 28) {
if (mu_mv_mag_sq < 105) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (MVDiff < 4) {
if (Rs < 1350) {
return 8;
}
else if (Rs >= 1350) {
return 4;
}
}
else if (MVDiff >= 4) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
return 8;
}
}
}
else if (ssDCval >= 67) {
return 8;
}
}
}
else if (MVDiff >= 12) {
if (posBalance < 45) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (posBalance >= 45) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
}
else if (negBalance >= 42) {
if (RsCsDiff < 42) {
if (Rs < 1213) {
if (SC < 1650) {
if (Rs < 949) {
if (Rs < 885) {
return 1;
}
else if (Rs >= 885) {
return 4;
}
}
else if (Rs >= 949) {
return 1;
}
}
else if (SC >= 1650) {
if (MVDiff < 15) {
if (Rs < 1168) {
return 8;
}
else if (Rs >= 1168) {
return 4;
}
}
else if (MVDiff >= 15) {
return 4;
}
}
}
else if (Rs >= 1213) {
if (Cs < 844) {
return 8;
}
else if (Cs >= 844) {
return 1;
}
}
}
else if (RsCsDiff >= 42) {
if (Cs < 804) {
return 2;
}
else if (Cs >= 804) {
return 4;
}
}
}
}
else if (SC >= 2441) {
if (Cs < 803) {
if (Cs < 785) {
return 4;
}
else if (Cs >= 785) {
return 1;
}
}
else if (Cs >= 803) {
if (negBalance < 21) {
if (Rs < 1712) {
return 1;
}
else if (Rs >= 1712) {
return 4;
}
}
else if (negBalance >= 21) {
return 8;
}
}
}
}
else if (refDCval >= 69) {
if (SC < 1713) {
if (refDCval < 73) {
if (ssDCval < 69) {
if (Rs < 911) {
return 8;
}
else if (Rs >= 911) {
return 4;
}
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (refDCval >= 73) {
return 8;
}
}
else if (SC >= 1713) {
if (Cs < 792) {
if (mu_mv_mag_sq < 120) {
if (diffRsCsdiff < 13) {
if (MVDiff < 23) {
if (refDCval < 70) {
if (Rs < 1291) {
return 2;
}
else if (Rs >= 1291) {
return 1;
}
}
else if (refDCval >= 70) {
return 1;
}
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (diffRsCsdiff >= 13) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (Cs < 643) {
return 8;
}
else if (Cs >= 643) {
return 1;
}
}
}
else if (Cs >= 792) {
if (mu_mv_mag_sq < 63) {
if (Rs < 985) {
if (Cs < 920) {
return 8;
}
else if (Cs >= 920) {
return 2;
}
}
else if (Rs >= 985) {
if (MVDiff < 105) {
return 1;
}
else if (MVDiff >= 105) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (CsDiff < 1) {
if (Rs < 1365) {
return 1;
}
else if (Rs >= 1365) {
return 8;
}
}
else if (CsDiff >= 1) {
if (TSCindex < 3) {
if (Rs < 1301) {
return 4;
}
else if (Rs >= 1301) {
return 1;
}
}
else if (TSCindex >= 3) {
if (negBalance < 67) {
if (diffRsCsdiff < -11) {
if (Rs < 1220) {
return 8;
}
else if (Rs >= 1220) {
return 1;
}
}
else if (diffRsCsdiff >= -11) {
return 8;
}
}
else if (negBalance >= 67) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 1786) {
if (posBalance < 49) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 113) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
else if (ssDCval >= 56) {
if (Rs < 2088) {
if (Cs < 932) {
if (Rs < 1792) {
return 8;
}
else if (Rs >= 1792) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Rs < 2026) {
return 1;
}
else if (Rs >= 2026) {
return 4;
}
}
}
}
else if (Cs >= 932) {
if (ssDCval < 75) {
if (ssDCval < 57) {
if (Rs < 1957) {
return 1;
}
else if (Rs >= 1957) {
return 4;
}
}
else if (ssDCval >= 57) {
return 4;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
}
else if (Rs >= 2088) {
if (posBalance < 34) {
if (diffTSC < -1) {
if (Rs < 2312) {
return 8;
}
else if (Rs >= 2312) {
return 8;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (posBalance >= 34) {
if (mu_mv_mag_sq < 87) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (RsCsDiff < 12) {
return 1;
}
else if (RsCsDiff >= 12) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 49) {
if (negBalance < 36) {
if (mu_mv_mag_sq < 71) {
if (Cs < 865) {
return 1;
}
else if (Cs >= 865) {
return 4;
}
}
else if (mu_mv_mag_sq >= 71) {
if (tcor < 97) {
return 2;
}
else if (tcor >= 97) {
if (Rs < 1919) {
return 8;
}
else if (Rs >= 1919) {
return 1;
}
}
}
}
else if (negBalance >= 36) {
if (Rs < 1944) {
return 8;
}
else if (Rs >= 1944) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 1126) {
if (ssDCval < 64) {
if (Rs < 533) {
if (ssDCval < 49) {
if (posBalance < 17) {
return 8;
}
else if (posBalance >= 17) {
if (Cs < 753) {
return 1;
}
else if (Cs >= 753) {
if (Rs < 497) {
return 4;
}
else if (Rs >= 497) {
return 1;
}
}
}
}
else if (ssDCval >= 49) {
if (Cs < 894) {
return 4;
}
else if (Cs >= 894) {
if (Rs < 325) {
return 2;
}
else if (Rs >= 325) {
return 4;
}
}
}
}
else if (Rs >= 533) {
if (SC < 1640) {
return 8;
}
else if (SC >= 1640) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (RsDiff < 1) {
if (Rs < 1028) {
return 1;
}
else if (Rs >= 1028) {
return 8;
}
}
else if (RsDiff >= 1) {
if (Cs < 643) {
return 1;
}
else if (Cs >= 643) {
if (Rs < 742) {
return 1;
}
else if (Rs >= 742) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (Rs < 804) {
if (Rs < 701) {
return 8;
}
else if (Rs >= 701) {
if (Cs < 629) {
return 2;
}
else if (Cs >= 629) {
return 1;
}
}
}
else if (Rs >= 804) {
if (Cs < 786) {
if (Rs < 1072) {
if (AFD < 1) {
if (Rs < 966) {
return 2;
}
else if (Rs >= 966) {
return 8;
}
}
else if (AFD >= 1) {
return 2;
}
}
else if (Rs >= 1072) {
if (Cs < 676) {
if (Rs < 1124) {
return 1;
}
else if (Rs >= 1124) {
return 2;
}
}
else if (Cs >= 676) {
return 4;
}
}
}
else if (Cs >= 786) {
if (ssDCval < 73) {
if (Rs < 924) {
return 4;
}
else if (Rs >= 924) {
return 1;
}
}
else if (ssDCval >= 73) {
if (negBalance < 28) {
return 8;
}
else if (negBalance >= 28) {
if (Rs < 996) {
return 4;
}
else if (Rs >= 996) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1126) {
if (posBalance < 43) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 59) {
if (Cs < 971) {
if (Cs < 647) {
return 1;
}
else if (Cs >= 647) {
if (ssDCval < 56) {
if (ssDCval < 48) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Rs < 1385) {
return 4;
}
else if (Rs >= 1385) {
if (Rs < 1503) {
return 8;
}
else if (Rs >= 1503) {
return 4;
}
}
}
}
else if (ssDCval >= 48) {
return 8;
}
}
else if (ssDCval >= 56) {
return 1;
}
}
}
else if (Cs >= 971) {
return 4;
}
}
else if (ssDCval >= 59) {
if (mcTcor < 99) {
if (posBalance < 36) {
if (Rs < 1789) {
return 8;
}
else if (Rs >= 1789) {
return 1;
}
}
else if (posBalance >= 36) {
return 1;
}
}
else if (mcTcor >= 99) {
if (Cs < 594) {
return 8;
}
else if (Cs >= 594) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (ssDCval < 66) {
if (ssDCval < 62) {
if (Rs < 1546) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (Rs >= 1546) {
return 1;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (ssDCval >= 66) {
if (Rs < 1272) {
return 4;
}
else if (Rs >= 1272) {
return 8;
}
}
}
}
}
else if (posBalance >= 43) {
if (ssDCval < 73) {
if (Cs < 703) {
return 8;
}
else if (Cs >= 703) {
if (posBalance < 47) {
if (Rs < 1151) {
return 4;
}
else if (Rs >= 1151) {
if (ssDCval < 48) {
return 4;
}
else if (ssDCval >= 48) {
return 8;
}
}
}
else if (posBalance >= 47) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 73) {
if (Cs < 771) {
return 1;
}
else if (Cs >= 771) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1006) {
if (AFD < 3) {
if (RsDiff < 1) {
if (Rs < 461) {
if (ssDCval < 60) {
return 2;
}
else if (ssDCval >= 60) {
return 8;
}
}
else if (Rs >= 461) {
if (mu_mv_mag_sq < 101) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 4;
}
}
else if (mu_mv_mag_sq >= 101) {
if (Cs < 1093) {
if (negBalance < 23) {
if (ssDCval < 50) {
if (Rs < 1087) {
return 4;
}
else if (Rs >= 1087) {
return 4;
}
}
else if (ssDCval >= 50) {
if (Rs < 1838) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
else if (Rs >= 1838) {
if (negBalance < 15) {
if (Rs < 1920) {
return 1;
}
else if (Rs >= 1920) {
return 4;
}
}
else if (negBalance >= 15) {
return 4;
}
}
}
}
else if (negBalance >= 23) {
return 8;
}
}
else if (Cs >= 1093) {
if (Rs < 1450) {
return 1;
}
else if (Rs >= 1450) {
if (Rs < 1665) {
return 8;
}
else if (Rs >= 1665) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 96) {
return 4;
}
else if (mu_mv_mag_sq >= 96) {
if (mu_mv_mag_sq < 126) {
if (Rs < 1721) {
if (MVDiff < 1) {
if (negBalance < 33) {
if (TSC < 2) {
if (Rs < 1207) {
return 1;
}
else if (Rs >= 1207) {
if (Cs < 1306) {
return 8;
}
else if (Cs >= 1306) {
return 4;
}
}
}
else if (TSC >= 2) {
if (tcor < 97) {
if (Rs < 475) {
return 2;
}
else if (Rs >= 475) {
return 1;
}
}
else if (tcor >= 97) {
if (Cs < 1113) {
if (Rs < 835) {
if (Rs < 685) {
return 4;
}
else if (Rs >= 685) {
return 1;
}
}
else if (Rs >= 835) {
if (posBalance < 30) {
return 1;
}
else if (posBalance >= 30) {
if (RsDiff < 2) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
if (Rs < 1354) {
return 4;
}
else if (Rs >= 1354) {
return 8;
}
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (Cs >= 1113) {
if (ssDCval < 78) {
if (Cs < 1175) {
if (negBalance < 21) {
if (Cs < 1159) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (Cs < 1136) {
return 8;
}
else if (Cs >= 1136) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 1700) {
return 8;
}
else if (Rs >= 1700) {
return 4;
}
}
}
}
}
else if (Cs >= 1159) {
return 4;
}
}
else if (negBalance >= 21) {
return 4;
}
}
else if (Cs >= 1175) {
return 8;
}
}
else if (ssDCval >= 78) {
return 4;
}
}
}
}
}
else if (negBalance >= 33) {
return 1;
}
}
else if (MVDiff >= 1) {
if (diffMVdiffVal < -11) {
return 4;
}
else if (diffMVdiffVal >= -11) {
return 1;
}
}
}
else if (Rs >= 1721) {
if (diffMVdiffVal < 1) {
if (ssDCval < 69) {
if (tcor < 99) {
if (Rs < 1877) {
return 4;
}
else if (Rs >= 1877) {
if (Rs < 1930) {
return 1;
}
else if (Rs >= 1930) {
if (Cs < 1054) {
return 4;
}
else if (Cs >= 1054) {
if (Rs < 1937) {
return 4;
}
else if (Rs >= 1937) {
return 1;
}
}
}
}
}
else if (tcor >= 99) {
return 4;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (SC < 2957) {
if (Rs < 1340) {
if (ssDCval < 51) {
if (SC < 1993) {
if (Rs < 745) {
return 8;
}
else if (Rs >= 745) {
return 1;
}
}
else if (SC >= 1993) {
return 8;
}
}
else if (ssDCval >= 51) {
if (Cs < 1092) {
if (Rs < 621) {
return 4;
}
else if (Rs >= 621) {
return 8;
}
}
else if (Cs >= 1092) {
if (Cs < 1196) {
return 4;
}
else if (Cs >= 1196) {
if (Rs < 1315) {
return 4;
}
else if (Rs >= 1315) {
return 1;
}
}
}
}
}
else if (Rs >= 1340) {
if (Rs < 1399) {
return 4;
}
else if (Rs >= 1399) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (ssDCval < 47) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
return 4;
}
}
else if (ssDCval >= 47) {
if (SC < 2801) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (Cs < 1122) {
if (diffRsCsdiff < 1) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (Cs < 1092) {
if (mu_mv_mag_sq < 128) {
return 4;
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (Cs >= 1092) {
if (Rs < 1668) {
if (Rs < 1646) {
return 4;
}
else if (Rs >= 1646) {
return 4;
}
}
else if (Rs >= 1668) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 1) {
return 4;
}
}
else if (Cs >= 1122) {
return 4;
}
}
}
else if (SC >= 2801) {
if (Rs < 1679) {
return 8;
}
else if (Rs >= 1679) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1772) {
if (Rs < 1749) {
if (negBalance < 19) {
if (Cs < 1121) {
return 4;
}
else if (Cs >= 1121) {
return 4;
}
}
else if (negBalance >= 19) {
return 4;
}
}
else if (Rs >= 1749) {
if (Cs < 1144) {
if (Rs < 1762) {
if (Rs < 1755) {
return 8;
}
else if (Rs >= 1755) {
return 4;
}
}
else if (Rs >= 1762) {
return 8;
}
}
else if (Cs >= 1144) {
return 8;
}
}
}
else if (Rs >= 1772) {
return 4;
}
}
else if (mu_mv_mag_sq >= 128) {
if (SC < 2839) {
if (Cs < 1087) {
if (SC < 2812) {
return 4;
}
else if (SC >= 2812) {
if (Rs < 1760) {
return 8;
}
else if (Rs >= 1760) {
return 8;
}
}
}
else if (Cs >= 1087) {
return 8;
}
}
else if (SC >= 2839) {
if (SC < 2865) {
if (SC < 2846) {
if (Rs < 1764) {
return 4;
}
else if (Rs >= 1764) {
return 4;
}
}
else if (SC >= 2846) {
if (Cs < 1158) {
if (SC < 2856) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
if (Rs < 1709) {
return 8;
}
else if (Rs >= 1709) {
return 4;
}
}
}
else if (SC >= 2856) {
if (Rs < 1780) {
return 8;
}
else if (Rs >= 1780) {
return 8;
}
}
}
else if (Cs >= 1158) {
return 4;
}
}
}
else if (SC >= 2865) {
if (SC < 2915) {
if (Cs < 1137) {
if (Rs < 1802) {
if (SC < 2870) {
return 4;
}
else if (SC >= 2870) {
if (Rs < 1766) {
return 4;
}
else if (Rs >= 1766) {
if (Rs < 1778) {
return 4;
}
else if (Rs >= 1778) {
return 4;
}
}
}
}
else if (Rs >= 1802) {
return 4;
}
}
else if (Cs >= 1137) {
return 4;
}
}
else if (SC >= 2915) {
if (Rs < 1751) {
return 4;
}
else if (Rs >= 1751) {
if (SC < 2954) {
if (diffAFD < 1) {
if (Rs < 1810) {
return 4;
}
else if (Rs >= 1810) {
return 4;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (SC >= 2954) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (SC >= 2957) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (Rs < 1911) {
return 1;
}
else if (Rs >= 1911) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 3) {
if (Rs < 891) {
if (RsCsDiff < 12) {
if (mu_mv_mag_sq < 122) {
if (posBalance < 45) {
if (Rs < 655) {
return 8;
}
else if (Rs >= 655) {
return 2;
}
}
else if (posBalance >= 45) {
if (negBalance < 35) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (negBalance >= 35) {
if (Cs < 1111) {
return 1;
}
else if (Cs >= 1111) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
return 4;
}
}
else if (RsCsDiff >= 12) {
if (tcor < 93) {
if (RsCsDiff < 25) {
return 1;
}
else if (RsCsDiff >= 25) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 8;
}
}
}
}
else if (tcor >= 93) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 4;
}
}
}
}
}
}
else if (Rs >= 891) {
if (RsCsDiff < 7) {
if (negBalance < 34) {
if (refDCval < 81) {
if (diffRsCsdiff < -7) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
else if (diffRsCsdiff >= -7) {
if (mu_mv_mag_sq < 119) {
return 8;
}
else if (mu_mv_mag_sq >= 119) {
if (negBalance < 33) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 4;
}
}
else if (negBalance >= 33) {
return 4;
}
}
}
}
else if (refDCval >= 81) {
return 4;
}
}
else if (negBalance >= 34) {
if (mu_mv_mag_sq < 99) {
if (Rs < 935) {
return 1;
}
else if (Rs >= 935) {
return 8;
}
}
else if (mu_mv_mag_sq >= 99) {
if (ssDCval < 72) {
if (mcTcor < 98) {
if (RsCsDiff < 4) {
return 1;
}
else if (RsCsDiff >= 4) {
return 8;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (ssDCval >= 72) {
if (Rs < 1247) {
return 4;
}
else if (Rs >= 1247) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 7) {
if (MVDiff < 4) {
if (Rs < 1436) {
if (mu_mv_mag_sq < 117) {
if (Rs < 964) {
if (ssDCval < 72) {
if (SC < 2135) {
return 4;
}
else if (SC >= 2135) {
return 2;
}
}
else if (ssDCval >= 72) {
if (Cs < 1184) {
return 8;
}
else if (Cs >= 1184) {
return 4;
}
}
}
else if (Rs >= 964) {
if (Rs < 1049) {
if (negBalance < 26) {
if (Rs < 991) {
return 8;
}
else if (Rs >= 991) {
return 4;
}
}
else if (negBalance >= 26) {
return 8;
}
}
else if (Rs >= 1049) {
if (Rs < 1153) {
if (Cs < 1056) {
return 8;
}
else if (Cs >= 1056) {
return 4;
}
}
else if (Rs >= 1153) {
if (AFD < 7) {
if (SC < 2599) {
if (diffTSC < -1) {
if (Cs < 1127) {
return 8;
}
else if (Cs >= 1127) {
return 4;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (SC >= 2599) {
return 4;
}
}
else if (AFD >= 7) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (negBalance < 39) {
return 4;
}
else if (negBalance >= 39) {
return 1;
}
}
}
}
else if (Rs >= 1436) {
if (posBalance < 44) {
if (ssDCval < 65) {
if (Rs < 1799) {
return 2;
}
else if (Rs >= 1799) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
return 1;
}
}
}
else if (ssDCval >= 65) {
if (Rs < 2162) {
return 8;
}
else if (Rs >= 2162) {
return 1;
}
}
}
else if (posBalance >= 44) {
if (posBalance < 56) {
if (Cs < 1275) {
return 8;
}
else if (Cs >= 1275) {
return 1;
}
}
else if (posBalance >= 56) {
if (Rs < 1695) {
return 2;
}
else if (Rs >= 1695) {
return 8;
}
}
}
}
}
else if (MVDiff >= 4) {
if (tcor < 96) {
if (ssDCval < 70) {
if (Cs < 1225) {
if (RsCsDiff < 11) {
if (Rs < 1156) {
return 4;
}
else if (Rs >= 1156) {
return 8;
}
}
else if (RsCsDiff >= 11) {
return 4;
}
}
else if (Cs >= 1225) {
if (Rs < 1969) {
return 2;
}
else if (Rs >= 1969) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (Cs < 1270) {
return 2;
}
else if (Cs >= 1270) {
if (Rs < 1736) {
return 4;
}
else if (Rs >= 1736) {
return 8;
}
}
}
}
else if (tcor >= 96) {
if (posBalance < 38) {
if (Rs < 2151) {
return 1;
}
else if (Rs >= 2151) {
return 4;
}
}
else if (posBalance >= 38) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 1142) {
return 4;
}
else if (Rs >= 1142) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 82) {
if (MVDiff < 2) {
if (SC < 1556) {
if (mu_mv_mag_sq < 116) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Rs < 691) {
return 4;
}
else if (Rs >= 691) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 116) {
return 2;
}
}
else if (SC >= 1556) {
if (Cs < 1036) {
if (RsDiff < 3) {
if (Rs < 2226) {
return 4;
}
else if (Rs >= 2226) {
return 8;
}
}
else if (RsDiff >= 3) {
if (Rs < 1298) {
return 4;
}
else if (Rs >= 1298) {
return 1;
}
}
}
else if (Cs >= 1036) {
if (Rs < 1328) {
return 4;
}
else if (Rs >= 1328) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (Rs < 652) {
if (ssDCval < 85) {
if (SC < 1234) {
if (Rs < 515) {
return 8;
}
else if (Rs >= 515) {
return 4;
}
}
else if (SC >= 1234) {
if (Rs < 627) {
if (Cs < 712) {
return 2;
}
else if (Cs >= 712) {
return 1;
}
}
else if (Rs >= 627) {
return 2;
}
}
}
else if (ssDCval >= 85) {
if (MVDiff < 3) {
if (Rs < 600) {
return 2;
}
else if (Rs >= 600) {
return 1;
}
}
else if (MVDiff >= 3) {
return 2;
}
}
}
else if (Rs >= 652) {
if (Cs < 855) {
if (Cs < 692) {
if (refDCval < 85) {
return 8;
}
else if (refDCval >= 85) {
if (refDCval < 87) {
if (RsCsDiff < 19) {
return 4;
}
else if (RsCsDiff >= 19) {
if (Rs < 1424) {
return 4;
}
else if (Rs >= 1424) {
return 1;
}
}
}
else if (refDCval >= 87) {
if (Rs < 1787) {
return 1;
}
else if (Rs >= 1787) {
return 2;
}
}
}
}
else if (Cs >= 692) {
if (TSC < 6) {
if (ssDCval < 86) {
return 1;
}
else if (ssDCval >= 86) {
if (Rs < 1308) {
return 4;
}
else if (Rs >= 1308) {
return 1;
}
}
}
else if (TSC >= 6) {
if (Rs < 1632) {
return 8;
}
else if (Rs >= 1632) {
return 2;
}
}
}
}
else if (Cs >= 855) {
if (mu_mv_mag_sq < 65) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
if (ssDCval < 84) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (Cs < 1007) {
return 4;
}
else if (Cs >= 1007) {
return 2;
}
}
}
else if (ssDCval >= 84) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 65) {
if (refDCval < 83) {
if (Rs < 1899) {
return 1;
}
else if (Rs >= 1899) {
return 8;
}
}
else if (refDCval >= 83) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 92) {
if (SC < 1302) {
if (negBalance < 72) {
if (MVDiff < 5) {
if (Cs < 725) {
if (negBalance < 26) {
if (Rs < 615) {
return 1;
}
else if (Rs >= 615) {
return 4;
}
}
else if (negBalance >= 26) {
if (refDCval < 114) {
if (SC < 1224) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
else if (SC >= 1224) {
if (Rs < 594) {
return 2;
}
else if (Rs >= 594) {
return 4;
}
}
}
else if (refDCval >= 114) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (mcTcor < 98) {
if (Rs < 559) {
return 2;
}
else if (Rs >= 559) {
return 8;
}
}
else if (mcTcor >= 98) {
return 2;
}
}
}
}
}
else if (Cs >= 725) {
if (SC < 1248) {
return 4;
}
else if (SC >= 1248) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (Rs < 600) {
if (SC < 1148) {
if (posBalance < 51) {
if (Rs < 537) {
return 2;
}
else if (Rs >= 537) {
return 8;
}
}
else if (posBalance >= 51) {
if (negBalance < 39) {
if (ssDCval < 133) {
return 2;
}
else if (ssDCval >= 133) {
if (Rs < 333) {
return 2;
}
else if (Rs >= 333) {
return 8;
}
}
}
else if (negBalance >= 39) {
if (posBalance < 57) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (posBalance >= 57) {
if (Rs < 476) {
return 4;
}
else if (Rs >= 476) {
if (Rs < 504) {
return 1;
}
else if (Rs >= 504) {
return 8;
}
}
}
}
}
}
else if (SC >= 1148) {
return 2;
}
}
else if (Rs >= 600) {
if (TSC < 4) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
else if (TSC >= 4) {
return 2;
}
}
}
}
else if (negBalance >= 72) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (SC >= 1302) {
if (RsCsDiff < 67) {
if (MVDiff < 8) {
if (TSC < 6) {
if (Cs < 973) {
if (SC < 1350) {
if (refDCval < 99) {
return 2;
}
else if (refDCval >= 99) {
if (Rs < 675) {
return 1;
}
else if (Rs >= 675) {
return 2;
}
}
}
else if (SC >= 1350) {
if (negBalance < 26) {
if (posBalance < 44) {
if (Rs < 2029) {
return 4;
}
else if (Rs >= 2029) {
return 8;
}
}
else if (posBalance >= 44) {
if (SC < 2773) {
return 1;
}
else if (SC >= 2773) {
return 2;
}
}
}
else if (negBalance >= 26) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (Rs < 761) {
return 4;
}
else if (Rs >= 761) {
if (negBalance < 57) {
if (Rs < 2042) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (Rs >= 2042) {
return 4;
}
}
else if (negBalance >= 57) {
if (Rs < 911) {
return 4;
}
else if (Rs >= 911) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 5) {
if (RsDiff < 4) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (RsDiff >= 4) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (TSC >= 5) {
if (MVDiff < 4) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
if (Rs < 1459) {
if (Rs < 839) {
return 1;
}
else if (Rs >= 839) {
return 8;
}
}
else if (Rs >= 1459) {
return 1;
}
}
}
else if (MVDiff >= 4) {
if (diffAFD < -1) {
if (negBalance < 45) {
if (Rs < 1432) {
return 2;
}
else if (Rs >= 1432) {
return 4;
}
}
else if (negBalance >= 45) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (Cs < 964) {
return 8;
}
else if (Cs >= 964) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 973) {
if (Rs < 1584) {
if (MVDiff < 5) {
if (Rs < 754) {
if (ssDCval < 114) {
if (Rs < 636) {
return 1;
}
else if (Rs >= 636) {
return 4;
}
}
else if (ssDCval >= 114) {
return 8;
}
}
else if (Rs >= 754) {
return 4;
}
}
else if (MVDiff >= 5) {
if (Cs < 982) {
return 1;
}
else if (Cs >= 982) {
return 8;
}
}
}
else if (Rs >= 1584) {
if (RsCsDiff < 16) {
if (Rs < 1973) {
return 1;
}
else if (Rs >= 1973) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (Rs < 2057) {
return 2;
}
else if (Rs >= 2057) {
return 4;
}
}
}
}
}
else if (TSC >= 6) {
if (tcor < 93) {
if (gchDC < 1) {
if (Cs < 923) {
return 8;
}
else if (Cs >= 923) {
return 4;
}
}
else if (gchDC >= 1) {
if (Rs < 802) {
return 2;
}
else if (Rs >= 802) {
return 8;
}
}
}
else if (tcor >= 93) {
if (mu_mv_mag_sq < 84) {
return 8;
}
else if (mu_mv_mag_sq >= 84) {
return 1;
}
}
}
}
else if (MVDiff >= 8) {
if (ssDCval < 103) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (RsCsDiff < 38) {
if (Rs < 1581) {
return 8;
}
else if (Rs >= 1581) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
return 8;
}
}
}
else if (RsCsDiff >= 38) {
if (Cs < 1074) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
return 1;
}
}
else if (Cs >= 1074) {
if (Rs < 1568) {
return 2;
}
else if (Rs >= 1568) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 103) {
if (ssDCval < 164) {
if (posBalance < 68) {
if (Rs < 2170) {
if (Cs < 861) {
if (tcor < 94) {
if (CsDiff < 4) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
if (Cs < 598) {
if (Rs < 884) {
return 2;
}
else if (Rs >= 884) {
return 8;
}
}
else if (Cs >= 598) {
return 1;
}
}
}
else if (CsDiff >= 4) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
return 8;
}
}
}
else if (tcor >= 94) {
if (negBalance < 37) {
if (Cs < 709) {
return 1;
}
else if (Cs >= 709) {
return 8;
}
}
else if (negBalance >= 37) {
if (ssDCval < 150) {
return 8;
}
else if (ssDCval >= 150) {
return 2;
}
}
}
}
else if (Cs >= 861) {
return 8;
}
}
else if (Rs >= 2170) {
if (AFD < 13) {
if (Cs < 941) {
return 2;
}
else if (Cs >= 941) {
return 8;
}
}
else if (AFD >= 13) {
return 1;
}
}
}
else if (posBalance >= 68) {
if (diffMVdiffVal < 13) {
return 2;
}
else if (diffMVdiffVal >= 13) {
if (Rs < 944) {
return 8;
}
else if (Rs >= 944) {
return 4;
}
}
}
}
else if (ssDCval >= 164) {
return 2;
}
}
}
}
else if (RsCsDiff >= 67) {
if (CsDiff < 6) {
if (TSC < 6) {
if (ssDCval < 102) {
if (SC < 2976) {
return 1;
}
else if (SC >= 2976) {
return 4;
}
}
else if (ssDCval >= 102) {
if (Rs < 1307) {
return 2;
}
else if (Rs >= 1307) {
return 8;
}
}
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 46) {
if (Rs < 1647) {
if (TSC < 8) {
if (posBalance < 76) {
return 2;
}
else if (posBalance >= 76) {
return 1;
}
}
else if (TSC >= 8) {
if (Rs < 1497) {
return 1;
}
else if (Rs >= 1497) {
return 8;
}
}
}
else if (Rs >= 1647) {
if (tcor < 83) {
if (Rs < 2236) {
return 1;
}
else if (Rs >= 2236) {
if (Rs < 2279) {
return 8;
}
else if (Rs >= 2279) {
return 2;
}
}
}
else if (tcor >= 83) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (Cs < 687) {
return 2;
}
else if (Cs >= 687) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (mu_mv_mag_sq < 21) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (ssDCval < 106) {
if (negBalance < 41) {
if (Rs < 2089) {
return 2;
}
else if (Rs >= 2089) {
return 8;
}
}
else if (negBalance >= 41) {
if (Cs < 851) {
return 2;
}
else if (Cs >= 851) {
return 1;
}
}
}
else if (ssDCval >= 106) {
if (RsDiff < 12) {
if (Cs < 1223) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
if (Cs < 819) {
return 2;
}
else if (Cs >= 819) {
return 8;
}
}
}
else if (Cs >= 1223) {
return 8;
}
}
else if (RsDiff >= 12) {
if (Rs < 1671) {
return 2;
}
else if (Rs >= 1671) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (AFD < 23) {
if (refDCval < 90) {
return 4;
}
else if (refDCval >= 90) {
return 8;
}
}
else if (AFD >= 23) {
if (tcor < 70) {
if (Rs < 1306) {
return 4;
}
else if (Rs >= 1306) {
return 8;
}
}
else if (tcor >= 70) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (ssDCval < 133) {
if (Cs < 1220) {
if (Rs < 343) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 343) {
if (mu_mv_mag_sq < 103) {
if (negBalance < 38) {
if (ssDCval < 107) {
if (Rs < 2053) {
if (Cs < 670) {
if (Rs < 1295) {
return 1;
}
else if (Rs >= 1295) {
return 4;
}
}
else if (Cs >= 670) {
return 8;
}
}
else if (Rs >= 2053) {
return 1;
}
}
else if (ssDCval >= 107) {
if (ssDCval < 115) {
if (Rs < 981) {
if (Rs < 410) {
return 4;
}
else if (Rs >= 410) {
return 1;
}
}
else if (Rs >= 981) {
return 8;
}
}
else if (ssDCval >= 115) {
if (tcor < 99) {
if (AFD < 3) {
if (Cs < 576) {
return 2;
}
else if (Cs >= 576) {
return 4;
}
}
else if (AFD >= 3) {
if (Rs < 1639) {
return 8;
}
else if (Rs >= 1639) {
return 2;
}
}
}
else if (tcor >= 99) {
if (diffMVdiffVal < 2) {
if (posBalance < 58) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (Rs < 552) {
return 1;
}
else if (Rs >= 552) {
return 2;
}
}
}
else if (posBalance >= 58) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
return 2;
}
}
}
}
}
else if (negBalance >= 38) {
if (RsDiff < 2) {
if (MVDiff < 5) {
if (diffRsCsdiff < -2) {
if (Rs < 639) {
return 8;
}
else if (Rs >= 639) {
return 4;
}
}
else if (diffRsCsdiff >= -2) {
return 4;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (RsDiff >= 2) {
if (ssDCval < 90) {
if (Rs < 1039) {
return 4;
}
else if (Rs >= 1039) {
return 8;
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (Cs < 917) {
if (refDCval < 97) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1069) {
if (diffAFD < -1) {
if (Rs < 590) {
return 8;
}
else if (Rs >= 590) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Rs >= 1069) {
if (ssDCval < 92) {
if (Rs < 2147) {
return 8;
}
else if (Rs >= 2147) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
else if (ssDCval >= 92) {
if (mu_mv_mag_sq < 127) {
return 4;
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 694) {
return 4;
}
else if (Cs >= 694) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 2181) {
if (posBalance < 26) {
return 4;
}
else if (posBalance >= 26) {
return 8;
}
}
else if (Rs >= 2181) {
return 4;
}
}
}
else if (refDCval >= 97) {
if (SC < 2489) {
if (Rs < 560) {
if (ssDCval < 102) {
if (Cs < 790) {
return 1;
}
else if (Cs >= 790) {
return 8;
}
}
else if (ssDCval >= 102) {
if (negBalance < 24) {
if (Rs < 481) {
return 4;
}
else if (Rs >= 481) {
return 1;
}
}
else if (negBalance >= 24) {
if (Cs < 902) {
if (Rs < 549) {
if (mu_mv_mag_sq < 108) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 108) {
return 8;
}
}
else if (Rs >= 549) {
return 4;
}
}
else if (Cs >= 902) {
if (Rs < 369) {
return 4;
}
else if (Rs >= 369) {
return 8;
}
}
}
}
}
else if (Rs >= 560) {
if (ssDCval < 125) {
if (ssDCval < 97) {
return 1;
}
else if (ssDCval >= 97) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (Rs < 623) {
return 8;
}
else if (Rs >= 623) {
return 4;
}
}
}
}
else if (ssDCval >= 125) {
if (Rs < 977) {
return 8;
}
else if (Rs >= 977) {
return 1;
}
}
}
}
else if (SC >= 2489) {
if (mu_mv_mag_sq < 113) {
return 8;
}
else if (mu_mv_mag_sq >= 113) {
return 1;
}
}
}
}
else if (Cs >= 917) {
if (ssDCval < 121) {
if (diffMVdiffVal < -1) {
if (tcor < 97) {
if (Rs < 1054) {
return 4;
}
else if (Rs >= 1054) {
if (Rs < 1595) {
return 8;
}
else if (Rs >= 1595) {
return 2;
}
}
}
else if (tcor >= 97) {
if (tcor < 98) {
if (ssDCval < 95) {
if (Rs < 1247) {
return 1;
}
else if (Rs >= 1247) {
return 2;
}
}
else if (ssDCval >= 95) {
if (mu_mv_mag_sq < 104) {
return 1;
}
else if (mu_mv_mag_sq >= 104) {
return 8;
}
}
}
else if (tcor >= 98) {
return 8;
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (ssDCval >= 121) {
if (Rs < 1401) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 1401) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1220) {
if (Rs < 665) {
if (ssDCval < 120) {
return 1;
}
else if (ssDCval >= 120) {
return 8;
}
}
else if (Rs >= 665) {
if (ssDCval < 95) {
return 2;
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
if (Rs < 2380) {
return 8;
}
else if (Rs >= 2380) {
return 1;
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 145) {
return 1;
}
else if (ssDCval >= 145) {
if (Cs < 858) {
return 2;
}
else if (Cs >= 858) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2424) {
if (MVDiff < 12) {
if (ssDCval < 109) {
if (posBalance < 48) {
if (Rs < 3736) {
if (RsCsDiff < 7) {
if (Cs < 686) {
if (ssDCval < 82) {
if (Cs < 405) {
return 8;
}
else if (Cs >= 405) {
if (Cs < 487) {
return 4;
}
else if (Cs >= 487) {
return 1;
}
}
}
else if (ssDCval >= 82) {
return 4;
}
}
else if (Cs >= 686) {
if (Rs < 3079) {
if (tcor < 99) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (tcor >= 99) {
return 8;
}
}
else if (Rs >= 3079) {
if (Rs < 3270) {
if (ssDCval < 90) {
if (Rs < 3179) {
return 4;
}
else if (Rs >= 3179) {
return 8;
}
}
else if (ssDCval >= 90) {
return 2;
}
}
else if (Rs >= 3270) {
if (Cs < 1286) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
if (Rs < 3503) {
return 4;
}
else if (Rs >= 3503) {
return 8;
}
}
}
else if (Cs >= 1286) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 7) {
if (refDCval < 95) {
if (negBalance < 51) {
if (RsCsDiff < 23) {
if (SC < 3493) {
if (Cs < 515) {
return 8;
}
else if (Cs >= 515) {
if (MVDiff < 8) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (MVDiff >= 8) {
return 4;
}
}
}
else if (SC >= 3493) {
if (refDCval < 63) {
if (diffRsCsdiff < -14) {
return 4;
}
else if (diffRsCsdiff >= -14) {
return 8;
}
}
else if (refDCval >= 63) {
if (SC < 4362) {
if (tcor < 99) {
return 2;
}
else if (tcor >= 99) {
return 8;
}
}
else if (SC >= 4362) {
if (Rs < 3551) {
return 4;
}
else if (Rs >= 3551) {
if (Rs < 3639) {
return 2;
}
else if (Rs >= 3639) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 23) {
if (refDCval < 66) {
if (Rs < 2573) {
if (SC < 3682) {
if (mu_mv_mag_sq < 56) {
if (Rs < 2452) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (Rs >= 2452) {
return 4;
}
}
else if (mu_mv_mag_sq >= 56) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 2438) {
return 2;
}
else if (Rs >= 2438) {
return 8;
}
}
}
}
else if (SC >= 3682) {
if (mu_mv_mag_sq < 99) {
return 8;
}
else if (mu_mv_mag_sq >= 99) {
return 4;
}
}
}
else if (Rs >= 2573) {
if (MVDiff < 1) {
if (Cs < 1156) {
return 2;
}
else if (Cs >= 1156) {
return 8;
}
}
else if (MVDiff >= 1) {
if (SC < 4421) {
if (SC < 3664) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (Rs < 2697) {
return 1;
}
else if (Rs >= 2697) {
return 4;
}
}
}
else if (SC >= 3664) {
if (CsDiff < 6) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (RsCsDiff < 49) {
if (Rs < 2881) {
return 2;
}
else if (Rs >= 2881) {
return 4;
}
}
else if (RsCsDiff >= 49) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
return 2;
}
}
}
}
else if (CsDiff >= 6) {
if (Rs < 3150) {
return 2;
}
else if (Rs >= 3150) {
return 4;
}
}
}
}
else if (SC >= 4421) {
if (mu_mv_mag_sq < 52) {
return 2;
}
else if (mu_mv_mag_sq >= 52) {
if (Cs < 1131) {
return 2;
}
else if (Cs >= 1131) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 66) {
if (Rs < 2877) {
if (TSC < 6) {
if (mu_mv_mag_sq < 80) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (Rs < 2734) {
return 2;
}
else if (Rs >= 2734) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
return 2;
}
}
else if (TSC >= 6) {
if (RsCsDiff < 62) {
if (Cs < 876) {
if (Rs < 2606) {
return 8;
}
else if (Rs >= 2606) {
return 8;
}
}
else if (Cs >= 876) {
return 4;
}
}
else if (RsCsDiff >= 62) {
if (Rs < 2833) {
if (RsDiff < 12) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 1;
}
}
else if (RsDiff >= 12) {
return 4;
}
}
else if (Rs >= 2833) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
}
}
else if (Rs >= 2877) {
if (ssDCval < 77) {
if (TSC < 7) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
if (diffAFD < -2) {
if (MVDiff < 6) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
return 2;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (diffAFD >= -2) {
if (TSC < 5) {
if (Rs < 3203) {
return 1;
}
else if (Rs >= 3203) {
if (Cs < 1269) {
return 4;
}
else if (Cs >= 1269) {
return 1;
}
}
}
else if (TSC >= 5) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 4;
}
}
}
}
}
}
else if (TSC >= 7) {
if (RsCsDiff < 121) {
return 2;
}
else if (RsCsDiff >= 121) {
if (Rs < 3396) {
if (diffRsCsdiff < 63) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (diffRsCsdiff >= 63) {
if (Cs < 815) {
return 1;
}
else if (Cs >= 815) {
return 4;
}
}
}
else if (Rs >= 3396) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
}
}
else if (ssDCval >= 77) {
if (mu_mv_mag_sq < 66) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
if (gchDC < 3) {
if (Cs < 1138) {
if (Rs < 3099) {
if (Rs < 3007) {
return 8;
}
else if (Rs >= 3007) {
return 8;
}
}
else if (Rs >= 3099) {
if (Rs < 3414) {
return 4;
}
else if (Rs >= 3414) {
if (Cs < 1063) {
return 4;
}
else if (Cs >= 1063) {
return 4;
}
}
}
}
else if (Cs >= 1138) {
return 4;
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 66) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 51) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (refDCval >= 95) {
if (mu_mv_mag_sq < 111) {
if (negBalance < 44) {
if (TSC < 5) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (TSC >= 5) {
if (Rs < 3619) {
if (Cs < 1054) {
if (MVDiff < 9) {
if (Cs < 972) {
return 8;
}
else if (Cs >= 972) {
return 4;
}
}
else if (MVDiff >= 9) {
if (gchDC < 1) {
if (Cs < 881) {
return 4;
}
else if (Cs >= 881) {
if (Rs < 3435) {
return 4;
}
else if (Rs >= 3435) {
if (Rs < 3508) {
return 4;
}
else if (Rs >= 3508) {
return 4;
}
}
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (Cs >= 1054) {
return 8;
}
}
else if (Rs >= 3619) {
return 2;
}
}
}
else if (negBalance >= 44) {
return 2;
}
}
else if (mu_mv_mag_sq >= 111) {
return 4;
}
}
}
}
else if (Rs >= 3736) {
if (Cs < 1257) {
if (refDCval < 91) {
if (Cs < 1139) {
if (diffAFD < -2) {
if (Rs < 4227) {
return 1;
}
else if (Rs >= 4227) {
return 4;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Cs >= 1139) {
if (ssDCval < 71) {
if (AFD < 7) {
if (Rs < 3799) {
return 8;
}
else if (Rs >= 3799) {
return 1;
}
}
else if (AFD >= 7) {
if (Cs < 1160) {
return 2;
}
else if (Cs >= 1160) {
return 4;
}
}
}
else if (ssDCval >= 71) {
if (mu_mv_mag_sq < 62) {
if (Cs < 1188) {
return 4;
}
else if (Cs >= 1188) {
if (Rs < 4264) {
return 4;
}
else if (Rs >= 4264) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 62) {
return 4;
}
}
}
}
else if (refDCval >= 91) {
if (negBalance < 38) {
if (tcor < 96) {
if (diffMVdiffVal < -2) {
if (Rs < 3748) {
return 4;
}
else if (Rs >= 3748) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 4110) {
if (Cs < 1099) {
if (Rs < 3800) {
return 8;
}
else if (Rs >= 3800) {
return 8;
}
}
else if (Cs >= 1099) {
return 8;
}
}
else if (Rs >= 4110) {
if (mu_mv_mag_sq < 55) {
return 4;
}
else if (mu_mv_mag_sq >= 55) {
if (Rs < 4601) {
if (Cs < 1176) {
return 4;
}
else if (Cs >= 1176) {
return 4;
}
}
else if (Rs >= 4601) {
return 8;
}
}
}
}
}
else if (tcor >= 96) {
if (Cs < 1092) {
return 2;
}
else if (Cs >= 1092) {
return 4;
}
}
}
else if (negBalance >= 38) {
return 2;
}
}
}
else if (Cs >= 1257) {
if (mu_mv_mag_sq < 127) {
if (negBalance < 33) {
if (Rs < 4087) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
else if (Rs >= 4087) {
if (Rs < 4115) {
return 1;
}
else if (Rs >= 4115) {
if (Rs < 4127) {
if (Rs < 4122) {
return 8;
}
else if (Rs >= 4122) {
return 4;
}
}
else if (Rs >= 4127) {
return 8;
}
}
}
}
else if (negBalance >= 33) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 1271) {
if (Rs < 4445) {
return 4;
}
else if (Rs >= 4445) {
return 2;
}
}
else if (Cs >= 1271) {
return 8;
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (SC < 3823) {
if (CsDiff < 1) {
if (Cs < 882) {
return 1;
}
else if (Cs >= 882) {
return 8;
}
}
else if (CsDiff >= 1) {
if (ssDCval < 77) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 2;
}
}
else if (ssDCval >= 77) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
else if (SC >= 3823) {
if (Cs < 757) {
return 4;
}
else if (Cs >= 757) {
if (ssDCval < 63) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (ssDCval >= 63) {
return 2;
}
}
}
}
else if (CsDiff >= 3) {
if (negBalance < 38) {
if (posBalance < 54) {
if (tcor < 95) {
if (negBalance < 32) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (negBalance >= 32) {
if (tcor < 90) {
if (TSC < 8) {
if (Cs < 582) {
return 1;
}
else if (Cs >= 582) {
return 4;
}
}
else if (TSC >= 8) {
if (Cs < 884) {
return 1;
}
else if (Cs >= 884) {
return 2;
}
}
}
else if (tcor >= 90) {
if (diffMVdiffVal < -2) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
if (diffRsCsdiff < 4) {
return 2;
}
else if (diffRsCsdiff >= 4) {
if (Cs < 916) {
return 2;
}
else if (Cs >= 916) {
return 4;
}
}
}
}
}
}
else if (tcor >= 95) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (Rs < 3043) {
return 1;
}
else if (Rs >= 3043) {
return 2;
}
}
}
}
else if (posBalance >= 54) {
if (RsDiff < 5) {
if (Rs < 2793) {
return 1;
}
else if (Rs >= 2793) {
return 4;
}
}
else if (RsDiff >= 5) {
if (AFD < 17) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (ssDCval < 83) {
return 2;
}
else if (ssDCval >= 83) {
if (MVDiff < 8) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (SC < 4268) {
return 1;
}
else if (SC >= 4268) {
return 2;
}
}
}
else if (MVDiff >= 8) {
if (Rs < 3502) {
return 1;
}
else if (Rs >= 3502) {
return 2;
}
}
}
}
}
else if (AFD >= 17) {
if (Rs < 3289) {
return 4;
}
else if (Rs >= 3289) {
return 2;
}
}
}
}
}
else if (negBalance >= 38) {
if (refDCval < 61) {
if (Cs < 963) {
return 2;
}
else if (Cs >= 963) {
return 4;
}
}
else if (refDCval >= 61) {
if (refDCval < 65) {
if (SC < 3561) {
return 2;
}
else if (SC >= 3561) {
if (Rs < 2611) {
return 4;
}
else if (Rs >= 2611) {
if (RsCsDiff < 142) {
if (diffRsCsdiff < -53) {
return 4;
}
else if (diffRsCsdiff >= -53) {
if (RsCsDiff < 38) {
return 4;
}
else if (RsCsDiff >= 38) {
return 2;
}
}
}
else if (RsCsDiff >= 142) {
return 4;
}
}
}
}
else if (refDCval >= 65) {
if (diffRsCsdiff < -191) {
return 4;
}
else if (diffRsCsdiff >= -191) {
if (Rs < 2453) {
if (Cs < 1074) {
return 2;
}
else if (Cs >= 1074) {
return 4;
}
}
else if (Rs >= 2453) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
if (Cs < 930) {
return 2;
}
else if (Cs >= 930) {
if (Rs < 2672) {
return 2;
}
else if (Rs >= 2672) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsCsDiff < 121) {
if (Cs < 1018) {
if (tcor < 96) {
if (SC < 3116) {
return 2;
}
else if (SC >= 3116) {
if (mu_mv_mag_sq < 59) {
return 8;
}
else if (mu_mv_mag_sq >= 59) {
if (Rs < 2734) {
return 8;
}
else if (Rs >= 2734) {
if (Rs < 2858) {
return 4;
}
else if (Rs >= 2858) {
return 4;
}
}
}
}
}
else if (tcor >= 96) {
if (mcTcor < 100) {
return 2;
}
else if (mcTcor >= 100) {
return 8;
}
}
}
else if (Cs >= 1018) {
if (diffAFD < -3) {
if (Rs < 3151) {
return 4;
}
else if (Rs >= 3151) {
return 2;
}
}
else if (diffAFD >= -3) {
if (Rs < 4214) {
if (ssDCval < 108) {
if (negBalance < 31) {
if (Rs < 3079) {
return 1;
}
else if (Rs >= 3079) {
return 8;
}
}
else if (negBalance >= 31) {
return 8;
}
}
else if (ssDCval >= 108) {
return 2;
}
}
else if (Rs >= 4214) {
return 2;
}
}
}
}
else if (RsCsDiff >= 121) {
if (refDCval < 106) {
if (negBalance < 52) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
return 2;
}
}
else if (negBalance >= 52) {
if (SC < 4536) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (Cs < 1041) {
return 4;
}
else if (Cs >= 1041) {
return 2;
}
}
}
else if (SC >= 4536) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
return 4;
}
}
}
}
else if (refDCval >= 106) {
if (Cs < 856) {
return 2;
}
else if (Cs >= 856) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
if (Rs < 3490) {
return 2;
}
else if (Rs >= 3490) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 2911) {
if (mcTcor < 97) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (tcor < 72) {
if (Cs < 922) {
return 2;
}
else if (Cs >= 922) {
return 4;
}
}
else if (tcor >= 72) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (tcor < 91) {
if (Cs < 1155) {
return 1;
}
else if (Cs >= 1155) {
return 8;
}
}
else if (tcor >= 91) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 117) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Rs < 2703) {
return 1;
}
else if (Rs >= 2703) {
return 4;
}
}
}
else if (refDCval >= 117) {
if (SC < 3377) {
if (ssDCval < 128) {
if (Rs < 2742) {
return 1;
}
else if (Rs >= 2742) {
return 2;
}
}
else if (ssDCval >= 128) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (Cs < 640) {
return 1;
}
else if (Cs >= 640) {
return 2;
}
}
}
}
else if (SC >= 3377) {
if (Cs < 1238) {
return 2;
}
else if (Cs >= 1238) {
return 1;
}
}
}
}
}
else if (Rs >= 2911) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 31) {
if (negBalance < 26) {
if (Rs < 4114) {
return 8;
}
else if (Rs >= 4114) {
return 2;
}
}
else if (negBalance >= 26) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 141) {
return 2;
}
else if (ssDCval >= 141) {
if (Rs < 4164) {
return 1;
}
else if (Rs >= 4164) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 10) {
if (Rs < 3525) {
return 4;
}
else if (Rs >= 3525) {
return 8;
}
}
else if (RsDiff >= 10) {
if (Rs < 3317) {
return 4;
}
else if (Rs >= 3317) {
if (ssDCval < 137) {
return 2;
}
else if (ssDCval >= 137) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 31) {
if (refDCval < 142) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
if (diffRsCsdiff < 44) {
return 2;
}
else if (diffRsCsdiff >= 44) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (Rs < 5944) {
return 1;
}
else if (Rs >= 5944) {
return 2;
}
}
}
}
}
else if (refDCval >= 142) {
if (Rs < 3894) {
if (SC < 4946) {
if (mu_mv_mag_sq < 119) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
if (Cs < 1126) {
return 2;
}
else if (Cs >= 1126) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 119) {
return 1;
}
}
else if (SC >= 4946) {
if (Rs < 3725) {
return 2;
}
else if (Rs >= 3725) {
return 1;
}
}
}
else if (Rs >= 3894) {
return 2;
}
}
}
}
else if (MVDiff >= 6) {
if (TSC < 5) {
if (ssDCval < 115) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Rs < 3190) {
return 8;
}
else if (Rs >= 3190) {
return 2;
}
}
}
else if (ssDCval >= 115) {
if (mu_mv_mag_sq < 84) {
return 8;
}
else if (mu_mv_mag_sq >= 84) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
}
}
else if (TSC >= 5) {
if (ssDCval < 130) {
if (Rs < 3586) {
if (Cs < 1112) {
if (refDCval < 113) {
if (Rs < 3263) {
if (Rs < 2997) {
return 2;
}
else if (Rs >= 2997) {
return 4;
}
}
else if (Rs >= 3263) {
return 2;
}
}
else if (refDCval >= 113) {
if (diffAFD < -6) {
return 1;
}
else if (diffAFD >= -6) {
return 2;
}
}
}
else if (Cs >= 1112) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (Rs >= 3586) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
return 4;
}
}
}
else if (ssDCval >= 130) {
if (mcTcor < 94) {
if (posBalance < 59) {
if (Rs < 4971) {
if (Rs < 3858) {
return 8;
}
else if (Rs >= 3858) {
return 1;
}
}
else if (Rs >= 4971) {
return 2;
}
}
else if (posBalance >= 59) {
if (Rs < 3217) {
return 1;
}
else if (Rs >= 3217) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (mcTcor < 97) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (ssDCval < 138) {
if (Cs < 908) {
return 2;
}
else if (Cs >= 908) {
return 1;
}
}
else if (ssDCval >= 138) {
if (posBalance < 39) {
return 2;
}
else if (posBalance >= 39) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (diffMVdiffVal < -13) {
return 1;
}
else if (diffMVdiffVal >= -13) {
if (ssDCval < 164) {
if (mu_mv_mag_sq < 3) {
if (Rs < 3842) {
if (Cs < 725) {
return 2;
}
else if (Cs >= 725) {
return 1;
}
}
else if (Rs >= 3842) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffTSC < -3) {
return 1;
}
else if (diffTSC >= -3) {
return 2;
}
}
}
else if (ssDCval >= 164) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 12) {
if (mcTcor < 68) {
if (MVDiff < 51) {
return 1;
}
else if (MVDiff >= 51) {
if (refDCval < 98) {
if (Rs < 3172) {
return 8;
}
else if (Rs >= 3172) {
return 1;
}
}
else if (refDCval >= 98) {
return 4;
}
}
}
else if (mcTcor >= 68) {
if (RsDiff < 9) {
if (refDCval < 92) {
if (refDCval < 69) {
if (RsCsDiff < 53) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
if (MVDiff < 22) {
if (Rs < 2574) {
return 4;
}
else if (Rs >= 2574) {
return 8;
}
}
else if (MVDiff >= 22) {
return 4;
}
}
}
else if (RsCsDiff >= 53) {
if (Cs < 689) {
if (Rs < 3044) {
return 8;
}
else if (Rs >= 3044) {
return 4;
}
}
else if (Cs >= 689) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 69) {
if (posBalance < 42) {
if (Cs < 355) {
return 8;
}
else if (Cs >= 355) {
if (Rs < 2823) {
return 1;
}
else if (Rs >= 2823) {
if (Rs < 3090) {
return 2;
}
else if (Rs >= 3090) {
return 4;
}
}
}
}
else if (posBalance >= 42) {
if (SC < 3425) {
return 1;
}
else if (SC >= 3425) {
if (diffTSC < -2) {
if (Rs < 2799) {
return 8;
}
else if (Rs >= 2799) {
return 2;
}
}
else if (diffTSC >= -2) {
if (ssDCval < 82) {
if (diffAFD < -1) {
if (MVDiff < 13) {
return 1;
}
else if (MVDiff >= 13) {
return 4;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (ssDCval >= 82) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 92) {
if (Cs < 1198) {
if (Rs < 4318) {
if (diffRsCsdiff < -22) {
if (MVDiff < 37) {
if (refDCval < 130) {
if (diffMVdiffVal < 4) {
if (MVDiff < 15) {
if (Rs < 2821) {
return 2;
}
else if (Rs >= 2821) {
return 8;
}
}
else if (MVDiff >= 15) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 8;
}
}
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
else if (refDCval >= 130) {
if (gchDC < 2) {
if (ssDCval < 131) {
return 2;
}
else if (ssDCval >= 131) {
return 1;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (MVDiff >= 37) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -22) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (SC < 3463) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
if (MVDiff < 34) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
return 1;
}
}
else if (MVDiff >= 34) {
if (Cs < 13) {
return 8;
}
else if (Cs >= 13) {
if (SC < 3053) {
return 1;
}
else if (SC >= 3053) {
return 8;
}
}
}
}
}
else if (SC >= 3463) {
if (posBalance < 37) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
else if (posBalance >= 37) {
if (posBalance < 61) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
return 8;
}
}
else if (posBalance >= 61) {
if (Cs < 763) {
return 2;
}
else if (Cs >= 763) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 4318) {
return 2;
}
}
else if (Cs >= 1198) {
if (Rs < 3815) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 1;
}
}
else if (Rs >= 3815) {
return 4;
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 133) {
if (refDCval < 78) {
if (TSC < 8) {
if (Rs < 2746) {
if (refDCval < 70) {
if (Rs < 2551) {
return 2;
}
else if (Rs >= 2551) {
if (diffTSC < -1) {
if (Rs < 2595) {
return 4;
}
else if (Rs >= 2595) {
return 2;
}
}
else if (diffTSC >= -1) {
if (tcor < 84) {
return 4;
}
else if (tcor >= 84) {
return 2;
}
}
}
}
else if (refDCval >= 70) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (SC < 3186) {
return 1;
}
else if (SC >= 3186) {
return 2;
}
}
}
}
else if (Rs >= 2746) {
if (MVDiff < 101) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
if (posBalance < 56) {
if (Cs < 355) {
if (Rs < 3771) {
return 1;
}
else if (Rs >= 3771) {
return 4;
}
}
else if (Cs >= 355) {
if (Rs < 2860) {
if (Rs < 2802) {
return 4;
}
else if (Rs >= 2802) {
return 2;
}
}
else if (Rs >= 2860) {
if (Cs < 1251) {
return 4;
}
else if (Cs >= 1251) {
if (Rs < 3442) {
return 4;
}
else if (Rs >= 3442) {
return 1;
}
}
}
}
}
else if (posBalance >= 56) {
if (Rs < 3768) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (Rs >= 3768) {
return 4;
}
}
}
}
else if (MVDiff >= 101) {
return 2;
}
}
}
else if (TSC >= 8) {
if (Rs < 3662) {
if (MVDiff < 25) {
if (mu_mv_mag_sq < 34) {
if (negBalance < 43) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
return 2;
}
}
else if (negBalance >= 43) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
if (Rs < 2708) {
return 1;
}
else if (Rs >= 2708) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (tcor < 82) {
if (refDCval < 70) {
if (RsCsDiff < 136) {
return 4;
}
else if (RsCsDiff >= 136) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
return 2;
}
}
}
else if (refDCval >= 70) {
return 4;
}
}
else if (tcor >= 82) {
if (Rs < 3537) {
return 2;
}
else if (Rs >= 3537) {
return 1;
}
}
}
}
else if (MVDiff >= 25) {
if (CsDiff < 8) {
if (mcTcor < 85) {
if (diffRsCsdiff < 75) {
if (SC < 4411) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (SC >= 4411) {
return 2;
}
}
else if (diffRsCsdiff >= 75) {
if (RsDiff < 16) {
return 4;
}
else if (RsDiff >= 16) {
return 2;
}
}
}
else if (mcTcor >= 85) {
if (diffMVdiffVal < -9) {
if (SC < 3798) {
return 1;
}
else if (SC >= 3798) {
return 2;
}
}
else if (diffMVdiffVal >= -9) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 4;
}
}
}
}
else if (CsDiff >= 8) {
if (Cs < 1110) {
return 8;
}
else if (Cs >= 1110) {
return 1;
}
}
}
}
else if (Rs >= 3662) {
if (refDCval < 76) {
if (Rs < 5986) {
return 4;
}
else if (Rs >= 5986) {
return 8;
}
}
else if (refDCval >= 76) {
return 2;
}
}
}
}
else if (refDCval >= 78) {
if (MVDiff < 30) {
if (ssDCval < 111) {
if (Cs < 1168) {
if (ssDCval < 98) {
if (TSC < 13) {
if (Rs < 4350) {
if (negBalance < 40) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 2;
}
}
else if (negBalance >= 40) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (SC < 3654) {
return 2;
}
else if (SC >= 3654) {
if (RsCsDiff < 213) {
return 4;
}
else if (RsCsDiff >= 213) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 4350) {
if (Cs < 684) {
return 4;
}
else if (Cs >= 684) {
return 1;
}
}
}
else if (TSC >= 13) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 4;
}
}
}
else if (ssDCval >= 98) {
if (Cs < 569) {
return 2;
}
else if (Cs >= 569) {
if (SC < 3601) {
if (diffMVdiffVal < -7) {
return 1;
}
else if (diffMVdiffVal >= -7) {
return 2;
}
}
else if (SC >= 3601) {
if (Rs < 2882) {
if (TSC < 11) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (Rs < 2682) {
return 8;
}
else if (Rs >= 2682) {
return 2;
}
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (Rs >= 2882) {
if (RsCsDiff < 299) {
if (Rs < 3076) {
return 2;
}
else if (Rs >= 3076) {
if (mcTcor < 97) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (Rs < 3679) {
return 2;
}
else if (Rs >= 3679) {
return 8;
}
}
}
else if (mcTcor >= 97) {
return 2;
}
}
}
else if (RsCsDiff >= 299) {
if (mcTcor < 90) {
if (Rs < 3695) {
if (Rs < 3423) {
return 4;
}
else if (Rs >= 3423) {
return 8;
}
}
else if (Rs >= 3695) {
return 2;
}
}
else if (mcTcor >= 90) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 1168) {
if (Rs < 3109) {
if (AFD < 19) {
if (negBalance < 54) {
return 4;
}
else if (negBalance >= 54) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
}
else if (AFD >= 19) {
if (AFD < 32) {
return 4;
}
else if (AFD >= 32) {
return 2;
}
}
}
else if (Rs >= 3109) {
if (mcTcor < 90) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 2;
}
}
else if (mcTcor >= 90) {
if (CsDiff < 5) {
if (Cs < 1208) {
return 4;
}
else if (Cs >= 1208) {
if (Rs < 4192) {
return 2;
}
else if (Rs >= 4192) {
return 8;
}
}
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
else if (ssDCval >= 111) {
if (Rs < 3048) {
if (mu_mv_mag_sq < 42) {
if (mcTcor < 95) {
if (Cs < 780) {
if (MVDiff < 26) {
return 2;
}
else if (MVDiff >= 26) {
return 1;
}
}
else if (Cs >= 780) {
if (Rs < 2880) {
return 1;
}
else if (Rs >= 2880) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (diffRsCsdiff < -153) {
if (Rs < 2850) {
return 2;
}
else if (Rs >= 2850) {
return 1;
}
}
else if (diffRsCsdiff >= -153) {
if (MVDiff < 14) {
if (Rs < 2594) {
return 2;
}
else if (Rs >= 2594) {
return 1;
}
}
else if (MVDiff >= 14) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (Rs >= 3048) {
if (mu_mv_mag_sq < 39) {
if (RsDiff < 23) {
if (mu_mv_mag_sq < 26) {
if (Cs < 1141) {
if (diffRsCsdiff < -154) {
return 2;
}
else if (diffRsCsdiff >= -154) {
if (tcor < 80) {
if (mcTcor < 94) {
if (Rs < 4318) {
return 1;
}
else if (Rs >= 4318) {
return 8;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (tcor >= 80) {
if (RsCsDiff < 238) {
if (negBalance < 54) {
return 2;
}
else if (negBalance >= 54) {
if (AFD < 13) {
if (Rs < 4402) {
return 2;
}
else if (Rs >= 4402) {
return 1;
}
}
else if (AFD >= 13) {
return 8;
}
}
}
else if (RsCsDiff >= 238) {
if (Cs < 519) {
return 2;
}
else if (Cs >= 519) {
if (diffAFD < -3) {
if (Rs < 3898) {
return 1;
}
else if (Rs >= 3898) {
return 2;
}
}
else if (diffAFD >= -3) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1141) {
if (ssDCval < 122) {
return 2;
}
else if (ssDCval >= 122) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (RsDiff >= 23) {
if (diffTSC < -3) {
if (Cs < 913) {
return 2;
}
else if (Cs >= 913) {
return 8;
}
}
else if (diffTSC >= -3) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 2;
}
}
}
}
}
else if (MVDiff >= 30) {
if (RsDiff < 16) {
if (mcTcor < 95) {
if (AFD < 24) {
if (Rs < 2698) {
if (mu_mv_mag_sq < 10) {
if (Cs < 297) {
return 4;
}
else if (Cs >= 297) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
if (negBalance < 55) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (SC < 3171) {
return 1;
}
else if (SC >= 3171) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
}
}
else if (negBalance >= 55) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (posBalance < 46) {
if (Rs < 2543) {
return 8;
}
else if (Rs >= 2543) {
return 2;
}
}
else if (posBalance >= 46) {
if (Rs < 2663) {
return 1;
}
else if (Rs >= 2663) {
if (Rs < 2680) {
return 8;
}
else if (Rs >= 2680) {
return 1;
}
}
}
}
}
else if (Rs >= 2698) {
if (diffRsCsdiff < -87) {
if (RsCsDiff < 197) {
if (MVDiff < 42) {
return 2;
}
else if (MVDiff >= 42) {
return 8;
}
}
else if (RsCsDiff >= 197) {
if (Cs < 1108) {
return 4;
}
else if (Cs >= 1108) {
return 1;
}
}
}
else if (diffRsCsdiff >= -87) {
if (MVDiff < 100) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (TSC < 11) {
if (diffRsCsdiff < -23) {
return 8;
}
else if (diffRsCsdiff >= -23) {
if (SC < 3741) {
return 8;
}
else if (SC >= 3741) {
if (Rs < 3040) {
return 2;
}
else if (Rs >= 3040) {
return 1;
}
}
}
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (MVDiff >= 100) {
if (Rs < 3595) {
return 4;
}
else if (Rs >= 3595) {
return 2;
}
}
}
}
}
else if (AFD >= 24) {
if (Cs < 1129) {
if (Rs < 2782) {
if (Rs < 2553) {
return 1;
}
else if (Rs >= 2553) {
return 2;
}
}
else if (Rs >= 2782) {
return 1;
}
}
else if (Cs >= 1129) {
if (Rs < 2784) {
if (Rs < 2646) {
return 2;
}
else if (Rs >= 2646) {
return 8;
}
}
else if (Rs >= 2784) {
return 4;
}
}
}
}
else if (mcTcor >= 95) {
if (Rs < 3352) {
if (posBalance < 64) {
if (RsCsDiff < 193) {
if (Cs < 380) {
if (tcor < 90) {
if (Rs < 2612) {
return 1;
}
else if (Rs >= 2612) {
return 8;
}
}
else if (tcor >= 90) {
if (Rs < 2942) {
return 2;
}
else if (Rs >= 2942) {
return 4;
}
}
}
else if (Cs >= 380) {
if (TSC < 10) {
if (MVDiff < 34) {
return 1;
}
else if (MVDiff >= 34) {
if (diffAFD < -9) {
return 1;
}
else if (diffAFD >= -9) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
return 8;
}
}
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (RsCsDiff >= 193) {
if (MVDiff < 40) {
return 2;
}
else if (MVDiff >= 40) {
return 1;
}
}
}
else if (posBalance >= 64) {
if (mu_mv_mag_sq < 5) {
return 1;
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 2597) {
return 1;
}
else if (Rs >= 2597) {
return 2;
}
}
}
}
else if (Rs >= 3352) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
if (posBalance < 54) {
if (Cs < 486) {
return 4;
}
else if (Cs >= 486) {
return 2;
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
}
}
else if (RsDiff >= 16) {
if (Cs < 391) {
if (RsCsDiff < 291) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 2728) {
return 1;
}
else if (Rs >= 2728) {
return 4;
}
}
}
else if (RsCsDiff >= 291) {
return 2;
}
}
else if (Cs >= 391) {
if (mcTcor < 83) {
if (Cs < 1041) {
if (Rs < 2936) {
return 1;
}
else if (Rs >= 2936) {
return 2;
}
}
else if (Cs >= 1041) {
if (diffRsCsdiff < -142) {
return 2;
}
else if (diffRsCsdiff >= -142) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 2;
}
}
}
}
else if (mcTcor >= 83) {
if (gchDC < 4) {
if (Cs < 857) {
if (RsCsDiff < 398) {
if (refDCval < 112) {
if (Cs < 787) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 2712) {
return 2;
}
else if (Rs >= 2712) {
return 4;
}
}
}
else if (Cs >= 787) {
if (Rs < 3286) {
return 1;
}
else if (Rs >= 3286) {
return 8;
}
}
}
else if (refDCval >= 112) {
if (Rs < 3009) {
return 1;
}
else if (Rs >= 3009) {
if (Rs < 4430) {
if (ssDCval < 128) {
return 8;
}
else if (ssDCval >= 128) {
return 1;
}
}
else if (Rs >= 4430) {
return 2;
}
}
}
}
else if (RsCsDiff >= 398) {
if (Cs < 775) {
return 2;
}
else if (Cs >= 775) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
if (Rs < 3428) {
return 2;
}
else if (Rs >= 3428) {
return 1;
}
}
}
}
}
else if (Cs >= 857) {
if (ssDCval < 105) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (ssDCval >= 105) {
if (TSC < 12) {
if (tcor < 66) {
return 8;
}
else if (tcor >= 66) {
if (MVDiff < 41) {
if (TSC < 10) {
if (ssDCval < 127) {
return 8;
}
else if (ssDCval >= 127) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
return 2;
}
}
}
else if (TSC >= 10) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
if (Rs < 4131) {
return 1;
}
else if (Rs >= 4131) {
return 8;
}
}
}
}
else if (MVDiff >= 41) {
if (refDCval < 118) {
return 2;
}
else if (refDCval >= 118) {
if (RsDiff < 22) {
return 8;
}
else if (RsDiff >= 22) {
if (TSC < 10) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
if (MVDiff < 44) {
return 2;
}
else if (MVDiff >= 44) {
return 8;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
}
else if (TSC >= 12) {
if (ssDCval < 114) {
return 2;
}
else if (ssDCval >= 114) {
return 1;
}
}
}
}
}
else if (gchDC >= 4) {
if (TSC < 9) {
if (diffMVdiffVal < 9) {
if (MVDiff < 31) {
return 2;
}
else if (MVDiff >= 31) {
return 8;
}
}
else if (diffMVdiffVal >= 9) {
return 2;
}
}
else if (TSC >= 9) {
if (negBalance < 52) {
if (mcTcor < 96) {
if (TSC < 12) {
if (diffMVdiffVal < -3) {
if (Rs < 4315) {
return 1;
}
else if (Rs >= 4315) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (negBalance >= 52) {
if (mu_mv_mag_sq < 94) {
if (ssDCval < 85) {
if (Rs < 4151) {
return 2;
}
else if (Rs >= 4151) {
return 1;
}
}
else if (ssDCval >= 85) {
if (negBalance < 53) {
if (Rs < 2871) {
return 8;
}
else if (Rs >= 2871) {
return 1;
}
}
else if (negBalance >= 53) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (Rs < 3464) {
return 4;
}
else if (Rs >= 3464) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 133) {
if (mcTcor < 93) {
if (AFD < 19) {
if (posBalance < 54) {
if (Rs < 4970) {
if (Rs < 3334) {
return 8;
}
else if (Rs >= 3334) {
return 1;
}
}
else if (Rs >= 4970) {
return 2;
}
}
else if (posBalance >= 54) {
if (Cs < 1028) {
return 4;
}
else if (Cs >= 1028) {
if (AFD < 14) {
if (Rs < 2930) {
return 8;
}
else if (Rs >= 2930) {
return 4;
}
}
else if (AFD >= 14) {
return 1;
}
}
}
}
else if (AFD >= 19) {
if (Rs < 4932) {
if (mu_mv_mag_sq < 27) {
if (MVDiff < 85) {
if (refDCval < 141) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
if (Rs < 3367) {
return 1;
}
else if (Rs >= 3367) {
return 4;
}
}
}
else if (refDCval >= 141) {
return 1;
}
}
else if (MVDiff >= 85) {
if (posBalance < 70) {
return 8;
}
else if (posBalance >= 70) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (Rs < 2921) {
return 1;
}
else if (Rs >= 2921) {
return 2;
}
}
}
else if (Rs >= 4932) {
if (AFD < 28) {
return 2;
}
else if (AFD >= 28) {
return 8;
}
}
}
}
else if (mcTcor >= 93) {
if (CsDiff < 4) {
if (SC < 3586) {
if (AFD < 25) {
return 1;
}
else if (AFD >= 25) {
if (Rs < 2533) {
return 1;
}
else if (Rs >= 2533) {
return 2;
}
}
}
else if (SC >= 3586) {
if (mcTcor < 95) {
if (Cs < 768) {
return 2;
}
else if (Cs >= 768) {
return 4;
}
}
else if (mcTcor >= 95) {
if (diffRsCsdiff < 10) {
if (posBalance < 53) {
if (mcTcor < 97) {
return 1;
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 16) {
return 1;
}
else if (mu_mv_mag_sq >= 16) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (ssDCval < 150) {
return 2;
}
else if (ssDCval >= 150) {
return 1;
}
}
}
else if (diffRsCsdiff >= 10) {
return 2;
}
}
}
}
else if (CsDiff >= 4) {
if (tcor < 84) {
if (Rs < 4904) {
if (SC < 3622) {
return 1;
}
else if (SC >= 3622) {
if (posBalance < 44) {
if (gchDC < 17) {
return 2;
}
else if (gchDC >= 17) {
return 1;
}
}
else if (posBalance >= 44) {
if (TSC < 9) {
if (Rs < 3340) {
return 2;
}
else if (Rs >= 3340) {
return 8;
}
}
else if (TSC >= 9) {
if (MVDiff < 83) {
if (posBalance < 61) {
return 1;
}
else if (posBalance >= 61) {
if (posBalance < 64) {
if (Rs < 3191) {
return 1;
}
else if (Rs >= 3191) {
return 2;
}
}
else if (posBalance >= 64) {
return 1;
}
}
}
else if (MVDiff >= 83) {
if (Cs < 932) {
return 8;
}
else if (Cs >= 932) {
return 2;
}
}
}
}
}
}
else if (Rs >= 4904) {
if (Cs < 996) {
return 2;
}
else if (Cs >= 996) {
return 8;
}
}
}
else if (tcor >= 84) {
if (SC < 7278) {
if (SC < 4215) {
return 1;
}
else if (SC >= 4215) {
if (negBalance < 49) {
if (RsDiff < 20) {
return 1;
}
else if (RsDiff >= 20) {
return 2;
}
}
else if (negBalance >= 49) {
if (refDCval < 140) {
if (AFD < 16) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 1;
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (refDCval >= 140) {
if (Cs < 882) {
return 2;
}
else if (Cs >= 882) {
return 1;
}
}
}
}
}
else if (SC >= 7278) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1311) {
if (mu_mv_mag_sq < 111) {
if (RsDiff < 9) {
if (CsDiff < 7) {
if (ssDCval < 90) {
if (negBalance < 40) {
if (Rs < 819) {
if (Rs < 314) {
return 4;
}
else if (Rs >= 314) {
if (MVDiff < 4) {
if (ssDCval < 69) {
if (negBalance < 32) {
return 1;
}
else if (negBalance >= 32) {
return 8;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (MVDiff >= 4) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 742) {
if (Cs < 1923) {
return 2;
}
else if (Cs >= 1923) {
return 1;
}
}
else if (Rs >= 742) {
return 8;
}
}
}
}
}
else if (Rs >= 819) {
if (SC < 10923) {
if (posBalance < 45) {
if (ssDCval < 64) {
if (Rs < 2917) {
if (RsCsDiff < 16) {
if (Rs < 1429) {
if (Rs < 956) {
return 8;
}
else if (Rs >= 956) {
return 4;
}
}
else if (Rs >= 1429) {
if (MVDiff < 51) {
return 8;
}
else if (MVDiff >= 51) {
return 1;
}
}
}
else if (RsCsDiff >= 16) {
if (ssDCval < 45) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (ssDCval >= 45) {
if (posBalance < 43) {
if (mu_mv_mag_sq < 99) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
return 8;
}
}
else if (mu_mv_mag_sq >= 99) {
if (SC < 2966) {
return 4;
}
else if (SC >= 2966) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (Rs < 2182) {
return 4;
}
else if (Rs >= 2182) {
return 8;
}
}
}
}
}
else if (Rs >= 2917) {
if (Cs < 2396) {
if (mu_mv_mag_sq < 106) {
return 4;
}
else if (mu_mv_mag_sq >= 106) {
if (Cs < 2252) {
return 8;
}
else if (Cs >= 2252) {
if (Cs < 2308) {
return 4;
}
else if (Cs >= 2308) {
return 4;
}
}
}
}
else if (Cs >= 2396) {
if (negBalance < 22) {
return 8;
}
else if (negBalance >= 22) {
return 8;
}
}
}
}
else if (ssDCval >= 64) {
if (refDCval < 73) {
if (MVDiff < 6) {
if (Rs < 3347) {
if (mu_mv_mag_sq < 88) {
if (Cs < 1790) {
return 2;
}
else if (Cs >= 1790) {
return 1;
}
}
else if (mu_mv_mag_sq >= 88) {
return 8;
}
}
else if (Rs >= 3347) {
return 4;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (refDCval >= 73) {
if (Rs < 3747) {
if (mu_mv_mag_sq < 94) {
if (posBalance < 40) {
if (Cs < 1465) {
return 4;
}
else if (Cs >= 1465) {
return 8;
}
}
else if (posBalance >= 40) {
if (SC < 3628) {
return 8;
}
else if (SC >= 3628) {
if (Cs < 2010) {
if (AFD < 6) {
if (Rs < 2191) {
return 1;
}
else if (Rs >= 2191) {
return 8;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 2010) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (Rs < 3011) {
if (Rs < 2021) {
return 8;
}
else if (Rs >= 2021) {
if (tcor < 91) {
if (Rs < 2636) {
return 4;
}
else if (Rs >= 2636) {
return 8;
}
}
else if (tcor >= 91) {
return 1;
}
}
}
else if (Rs >= 3011) {
if (Cs < 1648) {
return 2;
}
else if (Cs >= 1648) {
if (Cs < 1951) {
return 4;
}
else if (Cs >= 1951) {
return 8;
}
}
}
}
}
else if (Rs >= 3747) {
if (CsDiff < 5) {
if (ssDCval < 84) {
if (Cs < 2009) {
if (Rs < 4330) {
return 4;
}
else if (Rs >= 4330) {
return 8;
}
}
else if (Cs >= 2009) {
return 4;
}
}
else if (ssDCval >= 84) {
if (Rs < 3923) {
return 8;
}
else if (Rs >= 3923) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (SC < 5745) {
return 4;
}
else if (SC >= 5745) {
if (RsDiff < 7) {
if (Rs < 4886) {
return 8;
}
else if (Rs >= 4886) {
return 1;
}
}
else if (RsDiff >= 7) {
if (mu_mv_mag_sq < 90) {
return 4;
}
else if (mu_mv_mag_sq >= 90) {
return 2;
}
}
}
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2897) {
if (SC < 2903) {
if (mu_mv_mag_sq < 88) {
if (Cs < 1600) {
if (posBalance < 60) {
if (RsCsDiff < 19) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Rs < 1263) {
return 2;
}
else if (Rs >= 1263) {
return 4;
}
}
}
}
else if (RsCsDiff >= 19) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
if (Rs < 1298) {
return 2;
}
else if (Rs >= 1298) {
return 8;
}
}
}
}
else if (posBalance >= 60) {
if (mcTcor < 86) {
if (Rs < 1437) {
return 1;
}
else if (Rs >= 1437) {
return 8;
}
}
else if (mcTcor >= 86) {
if (Rs < 932) {
return 2;
}
else if (Rs >= 932) {
if (negBalance < 37) {
return 8;
}
else if (negBalance >= 37) {
if (Rs < 1135) {
return 8;
}
else if (Rs >= 1135) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1600) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 88) {
if (Rs < 1027) {
return 8;
}
else if (Rs >= 1027) {
if (mu_mv_mag_sq < 106) {
if (ssDCval < 64) {
if (negBalance < 10) {
return 4;
}
else if (negBalance >= 10) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (diffRsCsdiff < -12) {
return 4;
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
}
}
else if (ssDCval >= 64) {
return 4;
}
}
else if (mu_mv_mag_sq >= 106) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
if (Rs < 1120) {
return 4;
}
else if (Rs >= 1120) {
return 8;
}
}
}
}
}
}
else if (SC >= 2903) {
if (SC < 4403) {
if (Cs < 1337) {
if (TSC < 6) {
if (Rs < 2627) {
if (Rs < 2035) {
return 8;
}
else if (Rs >= 2035) {
if (Cs < 1318) {
return 8;
}
else if (Cs >= 1318) {
return 2;
}
}
}
else if (Rs >= 2627) {
return 4;
}
}
else if (TSC >= 6) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (Cs >= 1337) {
if (Rs < 1493) {
if (negBalance < 35) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (Rs < 1086) {
if (Cs < 2444) {
return 1;
}
else if (Cs >= 2444) {
return 8;
}
}
else if (Rs >= 1086) {
return 4;
}
}
}
else if (negBalance >= 35) {
if (Rs < 1170) {
return 8;
}
else if (Rs >= 1170) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (Rs >= 1493) {
if (Cs < 1578) {
if (Cs < 1559) {
if (Rs < 2343) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
else if (Rs >= 2343) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 8;
}
}
}
}
else if (Cs >= 1559) {
if (Rs < 1968) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (Rs >= 1968) {
return 2;
}
}
}
else if (Cs >= 1578) {
if (mcTcor < 89) {
if (Cs < 1861) {
return 8;
}
else if (Cs >= 1861) {
if (MVDiff < 11) {
if (Rs < 1938) {
return 8;
}
else if (Rs >= 1938) {
return 4;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
}
else if (mcTcor >= 89) {
if (RsDiff < 7) {
if (Cs < 1586) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
else if (Cs >= 1586) {
return 8;
}
}
else if (RsDiff >= 7) {
return 1;
}
}
}
}
}
}
else if (SC >= 4403) {
if (CsDiff < 3) {
if (Rs < 1703) {
return 1;
}
else if (Rs >= 1703) {
return 8;
}
}
else if (CsDiff >= 3) {
if (Cs < 2486) {
if (tcor < 97) {
if (diffRsCsdiff < 25) {
if (SC < 4417) {
return 4;
}
else if (SC >= 4417) {
if (posBalance < 48) {
if (Rs < 2688) {
return 8;
}
else if (Rs >= 2688) {
return 1;
}
}
else if (posBalance >= 48) {
return 8;
}
}
}
else if (diffRsCsdiff >= 25) {
return 4;
}
}
else if (tcor >= 97) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (Cs >= 2486) {
if (Rs < 1762) {
return 1;
}
else if (Rs >= 1762) {
return 4;
}
}
}
}
}
}
else if (Rs >= 2897) {
if (Cs < 2165) {
if (RsCsDiff < 38) {
if (tcor < 98) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (Cs < 1920) {
return 4;
}
else if (Cs >= 1920) {
return 8;
}
}
}
}
else if (tcor >= 98) {
if (Rs < 6077) {
return 4;
}
else if (Rs >= 6077) {
return 2;
}
}
}
else if (RsCsDiff >= 38) {
if (tcor < 92) {
if (Cs < 1613) {
return 2;
}
else if (Cs >= 1613) {
if (Rs < 3349) {
if (Cs < 1819) {
return 8;
}
else if (Cs >= 1819) {
return 1;
}
}
else if (Rs >= 3349) {
if (Rs < 4466) {
return 4;
}
else if (Rs >= 4466) {
return 4;
}
}
}
}
else if (tcor >= 92) {
if (Rs < 3259) {
if (Cs < 1462) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
else if (Cs >= 1462) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
if (Rs < 2940) {
return 1;
}
else if (Rs >= 2940) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
}
}
else if (Rs >= 3259) {
if (mu_mv_mag_sq < 59) {
if (TSC < 6) {
if (ssDCval < 81) {
return 4;
}
else if (ssDCval >= 81) {
if (Rs < 4194) {
return 1;
}
else if (Rs >= 4194) {
return 2;
}
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (mu_mv_mag_sq >= 59) {
return 2;
}
}
}
}
}
else if (Cs >= 2165) {
if (Rs < 5724) {
if (ssDCval < 55) {
if (negBalance < 31) {
if (Rs < 3814) {
return 8;
}
else if (Rs >= 3814) {
return 4;
}
}
else if (negBalance >= 31) {
return 4;
}
}
else if (ssDCval >= 55) {
if (RsDiff < 8) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 4772) {
return 8;
}
else if (Rs >= 4772) {
return 4;
}
}
}
else if (RsDiff >= 8) {
return 1;
}
}
}
else if (Rs >= 5724) {
return 2;
}
}
}
}
}
else if (SC >= 10923) {
return 2;
}
}
}
else if (negBalance >= 40) {
if (SC < 7201) {
if (mcTcor < 94) {
if (SC < 2862) {
if (mcTcor < 86) {
if (ssDCval < 88) {
if (ssDCval < 54) {
if (Rs < 1410) {
return 4;
}
else if (Rs >= 1410) {
return 8;
}
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (ssDCval >= 88) {
if (Rs < 1370) {
return 8;
}
else if (Rs >= 1370) {
return 4;
}
}
}
else if (mcTcor >= 86) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 37) {
if (Cs < 1353) {
return 4;
}
else if (Cs >= 1353) {
if (Rs < 1151) {
return 1;
}
else if (Rs >= 1151) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 37) {
return 8;
}
}
else if (MVDiff >= 3) {
if (AFD < 11) {
if (Rs < 1265) {
if (negBalance < 46) {
if (Rs < 816) {
return 2;
}
else if (Rs >= 816) {
return 4;
}
}
else if (negBalance >= 46) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (Rs >= 1265) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
return 8;
}
}
}
else if (AFD >= 11) {
if (mcTcor < 90) {
if (Cs < 1346) {
return 2;
}
else if (Cs >= 1346) {
if (Rs < 1269) {
return 8;
}
else if (Rs >= 1269) {
return 2;
}
}
}
else if (mcTcor >= 90) {
if (negBalance < 48) {
if (SC < 2424) {
return 2;
}
else if (SC >= 2424) {
return 8;
}
}
else if (negBalance >= 48) {
if (mu_mv_mag_sq < 3) {
if (Rs < 757) {
return 8;
}
else if (Rs >= 757) {
if (Cs < 1361) {
if (Rs < 1281) {
return 8;
}
else if (Rs >= 1281) {
return 4;
}
}
else if (Cs >= 1361) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1223) {
return 8;
}
else if (Rs >= 1223) {
return 2;
}
}
}
}
}
}
}
}
else if (SC >= 2862) {
if (Cs < 2870) {
if (RsCsDiff < 48) {
if (SC < 2886) {
if (SC < 2877) {
return 8;
}
else if (SC >= 2877) {
return 1;
}
}
else if (SC >= 2886) {
return 8;
}
}
else if (RsCsDiff >= 48) {
if (refDCval < 77) {
if (negBalance < 51) {
if (TSC < 12) {
if (Cs < 1845) {
if (RsCsDiff < 57) {
if (Rs < 1499) {
if (Rs < 1393) {
return 2;
}
else if (Rs >= 1393) {
return 4;
}
}
else if (Rs >= 1499) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (RsCsDiff >= 57) {
return 8;
}
}
else if (Cs >= 1845) {
if (diffRsCsdiff < 21) {
if (Cs < 1852) {
return 4;
}
else if (Cs >= 1852) {
if (diffRsCsdiff < -47) {
return 4;
}
else if (diffRsCsdiff >= -47) {
return 8;
}
}
}
else if (diffRsCsdiff >= 21) {
if (Rs < 3350) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
return 4;
}
}
else if (Rs >= 3350) {
return 4;
}
}
}
}
else if (TSC >= 12) {
if (Rs < 2285) {
return 1;
}
else if (Rs >= 2285) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (mu_mv_mag_sq < 106) {
return 8;
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 1551) {
return 4;
}
else if (Rs >= 1551) {
return 8;
}
}
}
}
else if (refDCval >= 77) {
if (Rs < 1684) {
if (AFD < 15) {
if (negBalance < 47) {
if (Cs < 1402) {
return 8;
}
else if (Cs >= 1402) {
return 4;
}
}
else if (negBalance >= 47) {
return 8;
}
}
else if (AFD >= 15) {
return 4;
}
}
else if (Rs >= 1684) {
if (SC < 4129) {
if (negBalance < 45) {
if (Rs < 1989) {
return 8;
}
else if (Rs >= 1989) {
return 2;
}
}
else if (negBalance >= 45) {
if (diffAFD < -3) {
if (Rs < 2120) {
return 8;
}
else if (Rs >= 2120) {
return 2;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
}
else if (SC >= 4129) {
if (refDCval < 83) {
if (MVDiff < 6) {
if (diffRsCsdiff < 55) {
return 8;
}
else if (diffRsCsdiff >= 55) {
return 4;
}
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 68) {
if (Rs < 2940) {
return 8;
}
else if (Rs >= 2940) {
return 4;
}
}
else if (mu_mv_mag_sq >= 68) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
return 4;
}
}
}
}
else if (refDCval >= 83) {
return 8;
}
}
}
}
}
}
else if (Cs >= 2870) {
if (Rs < 3116) {
return 4;
}
else if (Rs >= 3116) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (Cs < 1605) {
if (Rs < 2722) {
if (tcor < 89) {
if (diffRsCsdiff < -36) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (diffRsCsdiff >= -36) {
if (Rs < 830) {
if (Rs < 581) {
return 1;
}
else if (Rs >= 581) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (Rs >= 830) {
if (TSC < 7) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
if (RsCsDiff < 51) {
if (Rs < 1420) {
return 4;
}
else if (Rs >= 1420) {
return 8;
}
}
else if (RsCsDiff >= 51) {
return 2;
}
}
}
else if (TSC >= 7) {
if (ssDCval < 84) {
if (posBalance < 64) {
if (posBalance < 57) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (posBalance >= 57) {
return 4;
}
}
else if (posBalance >= 64) {
return 8;
}
}
else if (ssDCval >= 84) {
return 8;
}
}
}
}
}
else if (tcor >= 89) {
if (mcTcor < 97) {
if (RsCsDiff < 51) {
if (negBalance < 51) {
if (SC < 2394) {
return 8;
}
else if (SC >= 2394) {
if (ssDCval < 80) {
if (Rs < 1254) {
if (Rs < 1013) {
return 2;
}
else if (Rs >= 1013) {
return 4;
}
}
else if (Rs >= 1254) {
if (Rs < 2109) {
if (negBalance < 41) {
if (Rs < 1716) {
return 8;
}
else if (Rs >= 1716) {
return 1;
}
}
else if (negBalance >= 41) {
return 8;
}
}
else if (Rs >= 2109) {
if (Rs < 2307) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
return 8;
}
}
else if (Rs >= 2307) {
return 8;
}
}
}
}
else if (ssDCval >= 80) {
if (mu_mv_mag_sq < 79) {
if (SC < 3143) {
return 4;
}
else if (SC >= 3143) {
return 8;
}
}
else if (mu_mv_mag_sq >= 79) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
if (Rs < 2228) {
return 8;
}
else if (Rs >= 2228) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 51) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 1;
}
}
}
else if (RsCsDiff >= 51) {
if (MVDiff < 5) {
if (RsCsDiff < 69) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (RsCsDiff >= 69) {
if (Rs < 1874) {
return 4;
}
else if (Rs >= 1874) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (TSC < 7) {
if (Rs < 2362) {
return 2;
}
else if (Rs >= 2362) {
return 8;
}
}
else if (TSC >= 7) {
if (Cs < 1357) {
if (Cs < 1352) {
return 8;
}
else if (Cs >= 1352) {
return 1;
}
}
else if (Cs >= 1357) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
if (CsDiff < 5) {
if (Rs < 1032) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
return 2;
}
}
else if (Rs >= 1032) {
if (Cs < 1510) {
if (diffRsCsdiff < -18) {
if (Rs < 2402) {
return 1;
}
else if (Rs >= 2402) {
return 8;
}
}
else if (diffRsCsdiff >= -18) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (Cs >= 1510) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (refDCval < 77) {
if (Rs < 704) {
return 8;
}
else if (Rs >= 704) {
return 2;
}
}
else if (refDCval >= 77) {
return 8;
}
}
}
}
}
else if (Rs >= 2722) {
if (tcor < 91) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
else if (tcor >= 91) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
if (Rs < 3476) {
return 4;
}
else if (Rs >= 3476) {
return 1;
}
}
}
}
}
else if (Cs >= 1605) {
if (Rs < 2115) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (Cs < 2009) {
if (RsDiff < 4) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
if (mu_mv_mag_sq < 100) {
return 8;
}
else if (mu_mv_mag_sq >= 100) {
return 1;
}
}
}
else if (RsDiff >= 4) {
if (AFD < 9) {
if (Rs < 1471) {
return 2;
}
else if (Rs >= 1471) {
return 8;
}
}
else if (AFD >= 9) {
if (diffMVdiffVal < 2) {
if (Cs < 1752) {
return 4;
}
else if (Cs >= 1752) {
if (MVDiff < 7) {
if (Rs < 1670) {
return 8;
}
else if (Rs >= 1670) {
if (Cs < 1963) {
return 4;
}
else if (Cs >= 1963) {
return 8;
}
}
}
else if (MVDiff >= 7) {
return 2;
}
}
}
else if (diffMVdiffVal >= 2) {
if (RsCsDiff < 56) {
return 4;
}
else if (RsCsDiff >= 56) {
return 8;
}
}
}
}
}
else if (Cs >= 2009) {
if (ssDCval < 83) {
if (diffRsCsdiff < -12) {
if (Cs < 2267) {
return 2;
}
else if (Cs >= 2267) {
return 8;
}
}
else if (diffRsCsdiff >= -12) {
return 1;
}
}
else if (ssDCval >= 83) {
if (Rs < 1762) {
return 8;
}
else if (Rs >= 1762) {
if (Cs < 2200) {
return 8;
}
else if (Cs >= 2200) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2115) {
if (SC < 5094) {
if (refDCval < 83) {
if (MVDiff < 5) {
if (Rs < 2129) {
if (Rs < 2122) {
return 8;
}
else if (Rs >= 2122) {
return 4;
}
}
else if (Rs >= 2129) {
return 8;
}
}
else if (MVDiff >= 5) {
if (TSC < 8) {
if (Rs < 2128) {
return 8;
}
else if (Rs >= 2128) {
return 4;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (refDCval >= 83) {
if (TSC < 8) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (mu_mv_mag_sq < 77) {
return 1;
}
else if (mu_mv_mag_sq >= 77) {
if (diffRsCsdiff < 5) {
return 8;
}
else if (diffRsCsdiff >= 5) {
if (Rs < 2437) {
return 8;
}
else if (Rs >= 2437) {
return 1;
}
}
}
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (SC >= 5094) {
if (Rs < 2776) {
return 4;
}
else if (Rs >= 2776) {
if (posBalance < 45) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
if (Rs < 3384) {
return 8;
}
else if (Rs >= 3384) {
return 2;
}
}
}
else if (posBalance >= 45) {
if (RsCsDiff < 60) {
return 8;
}
else if (RsCsDiff >= 60) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
if (posBalance < 55) {
if (Rs < 3496) {
return 8;
}
else if (Rs >= 3496) {
return 4;
}
}
else if (posBalance >= 55) {
if (Rs < 3999) {
return 8;
}
else if (Rs >= 3999) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (SC >= 7201) {
if (Cs < 2664) {
return 4;
}
else if (Cs >= 2664) {
return 8;
}
}
}
}
else if (ssDCval >= 90) {
if (ssDCval < 163) {
if (Rs < 4425) {
if (ssDCval < 109) {
if (mcTcor < 96) {
if (Rs < 1434) {
if (refDCval < 95) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (Cs < 1526) {
return 8;
}
else if (Cs >= 1526) {
return 4;
}
}
}
else if (refDCval >= 95) {
if (MVDiff < 4) {
if (Rs < 1341) {
return 8;
}
else if (Rs >= 1341) {
return 4;
}
}
else if (MVDiff >= 4) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
else if (Rs >= 1434) {
if (Rs < 2149) {
if (mcTcor < 88) {
if (tcor < 64) {
return 4;
}
else if (tcor >= 64) {
if (Rs < 1520) {
if (Rs < 1503) {
return 8;
}
else if (Rs >= 1503) {
return 4;
}
}
else if (Rs >= 1520) {
return 8;
}
}
}
else if (mcTcor >= 88) {
if (AFD < 12) {
if (RsCsDiff < 80) {
if (ssDCval < 100) {
if (Rs < 1858) {
if (RsCsDiff < 67) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
else if (RsCsDiff >= 67) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (Rs >= 1858) {
if (TSC < 7) {
if (Cs < 1749) {
return 2;
}
else if (Cs >= 1749) {
return 8;
}
}
else if (TSC >= 7) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 2;
}
}
}
}
else if (ssDCval >= 100) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 2;
}
}
}
else if (RsCsDiff >= 80) {
return 2;
}
}
else if (AFD >= 12) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 1671) {
return 8;
}
else if (Rs >= 1671) {
return 1;
}
}
}
}
}
else if (Rs >= 2149) {
if (mu_mv_mag_sq < 99) {
if (refDCval < 107) {
if (negBalance < 55) {
if (diffRsCsdiff < -57) {
if (Rs < 3076) {
return 4;
}
else if (Rs >= 3076) {
return 8;
}
}
else if (diffRsCsdiff >= -57) {
if (negBalance < 34) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (negBalance >= 34) {
if (Cs < 2734) {
if (diffMVdiffVal < -14) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 8;
}
}
else if (diffMVdiffVal >= -14) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
if (MVDiff < 29) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
if (Rs < 3231) {
return 4;
}
else if (Rs >= 3231) {
return 8;
}
}
}
else if (MVDiff >= 29) {
if (Cs < 1801) {
return 4;
}
else if (Cs >= 1801) {
return 8;
}
}
}
}
}
else if (Cs >= 2734) {
return 4;
}
}
}
}
else if (negBalance >= 55) {
if (Cs < 1964) {
return 8;
}
else if (Cs >= 1964) {
return 4;
}
}
}
else if (refDCval >= 107) {
if (TSC < 7) {
if (Rs < 2847) {
return 8;
}
else if (Rs >= 2847) {
return 4;
}
}
else if (TSC >= 7) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Rs < 2765) {
if (Cs < 1408) {
return 1;
}
else if (Cs >= 1408) {
return 8;
}
}
else if (Rs >= 2765) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (ssDCval < 103) {
if (mcTcor < 93) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (mcTcor >= 93) {
if (Rs < 2300) {
return 8;
}
else if (Rs >= 2300) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
}
}
else if (ssDCval >= 103) {
return 8;
}
}
}
}
}
else if (mcTcor >= 96) {
if (RsCsDiff < 55) {
if (MVDiff < 4) {
if (posBalance < 40) {
if (Rs < 3180) {
if (negBalance < 29) {
if (Rs < 2393) {
if (Rs < 2259) {
return 8;
}
else if (Rs >= 2259) {
if (negBalance < 8) {
return 1;
}
else if (negBalance >= 8) {
return 4;
}
}
}
else if (Rs >= 2393) {
return 8;
}
}
else if (negBalance >= 29) {
if (Rs < 2273) {
return 4;
}
else if (Rs >= 2273) {
return 1;
}
}
}
else if (Rs >= 3180) {
if (SC < 5657) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (SC >= 5657) {
if (Rs < 4195) {
return 8;
}
else if (Rs >= 4195) {
return 4;
}
}
}
}
else if (posBalance >= 40) {
if (Rs < 925) {
return 4;
}
else if (Rs >= 925) {
if (mcTcor < 98) {
if (SC < 4250) {
if (Cs < 2058) {
if (SC < 3485) {
if (Rs < 1853) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
else if (Rs >= 1853) {
return 2;
}
}
else if (SC >= 3485) {
return 8;
}
}
else if (Cs >= 2058) {
return 2;
}
}
else if (SC >= 4250) {
if (Rs < 3581) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (Rs < 2804) {
return 8;
}
else if (Rs >= 2804) {
return 4;
}
}
}
else if (Rs >= 3581) {
if (mu_mv_mag_sq < 85) {
return 2;
}
else if (mu_mv_mag_sq >= 85) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
}
}
else if (mcTcor >= 98) {
if (Cs < 1530) {
if (Rs < 3139) {
return 8;
}
else if (Rs >= 3139) {
return 2;
}
}
else if (Cs >= 1530) {
if (TSC < 1) {
return 2;
}
else if (TSC >= 1) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (tcor < 93) {
if (posBalance < 55) {
if (RsDiff < 6) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
else if (RsDiff >= 6) {
if (ssDCval < 106) {
return 8;
}
else if (ssDCval >= 106) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (Cs < 1399) {
return 2;
}
else if (Cs >= 1399) {
return 8;
}
}
}
else if (tcor >= 93) {
if (CsDiff < 1) {
if (diffAFD < -1) {
if (Rs < 2342) {
if (Rs < 2258) {
return 2;
}
else if (Rs >= 2258) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 4;
}
}
}
else if (Rs >= 2342) {
return 8;
}
}
else if (diffAFD >= -1) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
return 4;
}
}
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 106) {
if (diffRsCsdiff < -29) {
if (Rs < 2132) {
return 2;
}
else if (Rs >= 2132) {
return 8;
}
}
else if (diffRsCsdiff >= -29) {
if (ssDCval < 107) {
if (SC < 2206) {
return 4;
}
else if (SC >= 2206) {
if (posBalance < 20) {
if (tcor < 99) {
if (Rs < 2276) {
return 2;
}
else if (Rs >= 2276) {
return 1;
}
}
else if (tcor >= 99) {
return 8;
}
}
else if (posBalance >= 20) {
if (Rs < 4125) {
return 8;
}
else if (Rs >= 4125) {
if (Rs < 4136) {
return 4;
}
else if (Rs >= 4136) {
return 8;
}
}
}
}
}
else if (ssDCval >= 107) {
if (negBalance < 40) {
if (TSC < 6) {
if (Rs < 2338) {
return 4;
}
else if (Rs >= 2338) {
return 8;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (negBalance >= 40) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 2048) {
return 2;
}
else if (Rs >= 2048) {
if (Rs < 2933) {
return 1;
}
else if (Rs >= 2933) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 55) {
if (Rs < 2220) {
if (gchDC < 1) {
if (SC < 3595) {
if (Cs < 1531) {
return 1;
}
else if (Cs >= 1531) {
return 8;
}
}
else if (SC >= 3595) {
if (Rs < 1997) {
return 2;
}
else if (Rs >= 1997) {
return 1;
}
}
}
else if (gchDC >= 1) {
if (refDCval < 115) {
return 2;
}
else if (refDCval >= 115) {
return 8;
}
}
}
else if (Rs >= 2220) {
if (negBalance < 38) {
if (diffRsCsdiff < 14) {
if (MVDiff < 12) {
if (SC < 6227) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
else if (SC >= 6227) {
return 8;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (diffRsCsdiff >= 14) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (negBalance >= 38) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (Cs < 2815) {
if (RsDiff < 6) {
if (Rs < 2508) {
return 8;
}
else if (Rs >= 2508) {
return 4;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (Cs >= 2815) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 1263) {
if (ssDCval < 148) {
if (SC < 2651) {
if (Rs < 645) {
if (Rs < 586) {
return 4;
}
else if (Rs >= 586) {
return 1;
}
}
else if (Rs >= 645) {
if (posBalance < 50) {
if (ssDCval < 116) {
return 4;
}
else if (ssDCval >= 116) {
if (mu_mv_mag_sq < 104) {
return 8;
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 1105) {
return 4;
}
else if (Rs >= 1105) {
return 8;
}
}
}
}
else if (posBalance >= 50) {
if (Rs < 1020) {
return 8;
}
else if (Rs >= 1020) {
if (mu_mv_mag_sq < 95) {
if (mcTcor < 87) {
return 1;
}
else if (mcTcor >= 87) {
return 8;
}
}
else if (mu_mv_mag_sq >= 95) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
}
}
else if (SC >= 2651) {
if (negBalance < 45) {
if (refDCval < 133) {
if (diffAFD < -1) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
else if (diffAFD >= -1) {
if (RsCsDiff < 18) {
return 4;
}
else if (RsCsDiff >= 18) {
if (tcor < 92) {
if (Rs < 1214) {
return 4;
}
else if (Rs >= 1214) {
return 8;
}
}
else if (tcor >= 92) {
return 8;
}
}
}
}
else if (refDCval >= 133) {
return 8;
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (ssDCval >= 148) {
if (Rs < 1180) {
if (Rs < 1044) {
return 8;
}
else if (Rs >= 1044) {
return 4;
}
}
else if (Rs >= 1180) {
return 2;
}
}
}
else if (Rs >= 1263) {
if (mcTcor < 92) {
if (refDCval < 140) {
if (ssDCval < 128) {
if (TSC < 11) {
if (Rs < 1755) {
if (refDCval < 117) {
return 8;
}
else if (refDCval >= 117) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
}
}
else if (Rs >= 1755) {
return 8;
}
}
else if (TSC >= 11) {
if (ssDCval < 111) {
return 1;
}
else if (ssDCval >= 111) {
return 8;
}
}
}
else if (ssDCval >= 128) {
if (MVDiff < 11) {
if (Cs < 2121) {
if (negBalance < 52) {
if (diffMVdiffVal < 3) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Rs < 1336) {
return 8;
}
else if (Rs >= 1336) {
return 1;
}
}
}
else if (diffMVdiffVal >= 3) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (Cs >= 2121) {
return 8;
}
}
else if (MVDiff >= 11) {
if (SC < 3332) {
if (Cs < 1525) {
return 8;
}
else if (Cs >= 1525) {
return 1;
}
}
else if (SC >= 3332) {
return 8;
}
}
}
}
else if (refDCval >= 140) {
if (mu_mv_mag_sq < 27) {
if (Rs < 2135) {
if (Cs < 1380) {
return 1;
}
else if (Cs >= 1380) {
return 2;
}
}
else if (Rs >= 2135) {
return 8;
}
}
else if (mu_mv_mag_sq >= 27) {
if (Rs < 3452) {
return 8;
}
else if (Rs >= 3452) {
return 1;
}
}
}
}
else if (mcTcor >= 92) {
if (AFD < 12) {
if (posBalance < 35) {
if (mu_mv_mag_sq < 107) {
if (tcor < 93) {
if (Cs < 1510) {
return 1;
}
else if (Cs >= 1510) {
if (Rs < 1744) {
return 2;
}
else if (Rs >= 1744) {
return 8;
}
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 2768) {
if (Rs < 2587) {
if (Rs < 2184) {
return 2;
}
else if (Rs >= 2184) {
return 8;
}
}
else if (Rs >= 2587) {
return 4;
}
}
else if (Rs >= 2768) {
if (Cs < 3660) {
return 2;
}
else if (Cs >= 3660) {
return 8;
}
}
}
}
else if (posBalance >= 35) {
if (mu_mv_mag_sq < 37) {
if (ssDCval < 113) {
if (Rs < 2023) {
return 2;
}
else if (Rs >= 2023) {
if (Cs < 1418) {
return 8;
}
else if (Cs >= 1418) {
return 4;
}
}
}
else if (ssDCval >= 113) {
if (Rs < 1702) {
if (ssDCval < 144) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 8;
}
}
else if (ssDCval >= 144) {
return 2;
}
}
else if (Rs >= 1702) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (Rs < 2819) {
return 8;
}
else if (Rs >= 2819) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (Cs < 1363) {
if (Rs < 2823) {
if (negBalance < 52) {
if (Rs < 2352) {
return 8;
}
else if (Rs >= 2352) {
if (SC < 3723) {
return 1;
}
else if (SC >= 3723) {
return 8;
}
}
}
else if (negBalance >= 52) {
if (Rs < 2075) {
return 2;
}
else if (Rs >= 2075) {
return 8;
}
}
}
else if (Rs >= 2823) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
else if (Cs >= 1363) {
if (negBalance < 48) {
if (refDCval < 123) {
if (diffTSC < 2) {
if (diffAFD < -2) {
if (Rs < 2044) {
if (Rs < 1825) {
return 8;
}
else if (Rs >= 1825) {
return 2;
}
}
else if (Rs >= 2044) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (diffTSC >= 2) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (refDCval >= 123) {
if (MVDiff < 7) {
if (ssDCval < 136) {
if (ssDCval < 135) {
if (diffRsCsdiff < -23) {
if (ssDCval < 128) {
return 1;
}
else if (ssDCval >= 128) {
return 8;
}
}
else if (diffRsCsdiff >= -23) {
if (diffMVdiffVal < -11) {
if (Rs < 1846) {
return 1;
}
else if (Rs >= 1846) {
return 8;
}
}
else if (diffMVdiffVal >= -11) {
return 8;
}
}
}
else if (ssDCval >= 135) {
if (Rs < 2755) {
return 1;
}
else if (Rs >= 2755) {
return 8;
}
}
}
else if (ssDCval >= 136) {
return 8;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (tcor < 95) {
if (tcor < 92) {
if (Rs < 2996) {
return 8;
}
else if (Rs >= 2996) {
if (SC < 5609) {
return 1;
}
else if (SC >= 5609) {
return 8;
}
}
}
else if (tcor >= 92) {
return 1;
}
}
else if (tcor >= 95) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 12) {
if (Rs < 2421) {
if (posBalance < 61) {
if (mcTcor < 96) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
return 4;
}
}
else if (mcTcor >= 96) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
if (Cs < 1456) {
return 1;
}
else if (Cs >= 1456) {
return 8;
}
}
}
}
else if (posBalance >= 61) {
if (tcor < 91) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 1;
}
}
else if (tcor >= 91) {
if (Rs < 2031) {
return 2;
}
else if (Rs >= 2031) {
return 8;
}
}
}
}
else if (Rs >= 2421) {
if (SC < 4558) {
if (mu_mv_mag_sq < 49) {
return 8;
}
else if (mu_mv_mag_sq >= 49) {
if (Cs < 1676) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
else if (Cs >= 1676) {
return 1;
}
}
}
else if (SC >= 4558) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 4425) {
if (CsDiff < 4) {
if (Cs < 2777) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
else if (Cs >= 2777) {
return 8;
}
}
else if (CsDiff >= 4) {
if (Rs < 6255) {
if (Cs < 1846) {
if (tcor < 96) {
if (SC < 6142) {
return 1;
}
else if (SC >= 6142) {
if (Rs < 5058) {
return 8;
}
else if (Rs >= 5058) {
return 1;
}
}
}
else if (tcor >= 96) {
if (RsDiff < 6) {
if (Cs < 1812) {
return 8;
}
else if (Cs >= 1812) {
return 2;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (Cs >= 1846) {
if (tcor < 87) {
return 1;
}
else if (tcor >= 87) {
return 8;
}
}
}
else if (Rs >= 6255) {
if (Rs < 6634) {
return 4;
}
else if (Rs >= 6634) {
return 8;
}
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2311) {
if (Rs < 1101) {
return 1;
}
else if (Rs >= 1101) {
return 2;
}
}
else if (Rs >= 2311) {
return 8;
}
}
}
}
else if (CsDiff >= 7) {
if (Rs < 3687) {
if (Rs < 1512) {
if (TSC < 5) {
if (mu_mv_mag_sq < 87) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 4;
}
}
else if (mu_mv_mag_sq >= 87) {
if (Rs < 1082) {
return 8;
}
else if (Rs >= 1082) {
return 1;
}
}
}
else if (TSC >= 5) {
if (ssDCval < 93) {
if (ssDCval < 60) {
if (mcTcor < 72) {
return 4;
}
else if (mcTcor >= 72) {
if (SC < 2967) {
if (negBalance < 45) {
if (AFD < 10) {
if (Rs < 1218) {
return 4;
}
else if (Rs >= 1218) {
return 8;
}
}
else if (AFD >= 10) {
return 2;
}
}
else if (negBalance >= 45) {
if (Rs < 1439) {
if (posBalance < 58) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
if (diffRsCsdiff < 6) {
if (Cs < 1887) {
return 4;
}
else if (Cs >= 1887) {
return 2;
}
}
else if (diffRsCsdiff >= 6) {
return 2;
}
}
}
else if (posBalance >= 58) {
return 4;
}
}
else if (Rs >= 1439) {
return 8;
}
}
}
else if (SC >= 2967) {
if (tcor < 80) {
if (Cs < 1918) {
return 1;
}
else if (Cs >= 1918) {
if (MVDiff < 129) {
return 2;
}
else if (MVDiff >= 129) {
return 1;
}
}
}
else if (tcor >= 80) {
if (Rs < 1064) {
return 8;
}
else if (Rs >= 1064) {
return 1;
}
}
}
}
}
else if (ssDCval >= 60) {
if (mu_mv_mag_sq < 40) {
if (tcor < 88) {
if (gchDC < 4) {
if (mcTcor < 93) {
if (TSC < 10) {
if (negBalance < 39) {
if (Rs < 1076) {
return 8;
}
else if (Rs >= 1076) {
return 1;
}
}
else if (negBalance >= 39) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
if (Rs < 1287) {
return 8;
}
else if (Rs >= 1287) {
return 4;
}
}
}
}
else if (TSC >= 10) {
if (Rs < 1284) {
if (MVDiff < 32) {
if (Rs < 1247) {
return 4;
}
else if (Rs >= 1247) {
return 2;
}
}
else if (MVDiff >= 32) {
if (gchDC < 3) {
if (Cs < 1400) {
return 4;
}
else if (Cs >= 1400) {
return 2;
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (Rs >= 1284) {
if (RsDiff < 7) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 1340) {
return 8;
}
else if (Rs >= 1340) {
return 4;
}
}
}
else if (RsDiff >= 7) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 23) {
if (Cs < 1364) {
return 8;
}
else if (Cs >= 1364) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
if (diffRsCsdiff < 55) {
return 4;
}
else if (diffRsCsdiff >= 55) {
if (Cs < 2418) {
return 2;
}
else if (Cs >= 2418) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (Rs < 1410) {
return 1;
}
else if (Rs >= 1410) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 1480) {
return 8;
}
else if (Rs >= 1480) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 88) {
if (SC < 2743) {
if (diffAFD < 4) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Rs < 945) {
return 8;
}
else if (Rs >= 945) {
return 2;
}
}
}
else if (diffAFD >= 4) {
if (Rs < 737) {
return 1;
}
else if (Rs >= 737) {
return 2;
}
}
}
else if (SC >= 2743) {
if (AFD < 19) {
if (Rs < 1057) {
return 4;
}
else if (Rs >= 1057) {
if (MVDiff < 15) {
if (Rs < 1429) {
return 1;
}
else if (Rs >= 1429) {
return 4;
}
}
else if (MVDiff >= 15) {
return 8;
}
}
}
else if (AFD >= 19) {
if (RsCsDiff < 172) {
return 2;
}
else if (RsCsDiff >= 172) {
return 8;
}
}
}
}
else if (ssDCval >= 88) {
if (MVDiff < 7) {
if (Rs < 1437) {
return 1;
}
else if (Rs >= 1437) {
return 8;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
}
else if (gchDC >= 4) {
if (refDCval < 64) {
if (diffRsCsdiff < 14) {
if (Rs < 1178) {
return 1;
}
else if (Rs >= 1178) {
return 4;
}
}
else if (diffRsCsdiff >= 14) {
return 8;
}
}
else if (refDCval >= 64) {
if (diffAFD < 5) {
return 2;
}
else if (diffAFD >= 5) {
return 8;
}
}
}
}
else if (tcor >= 88) {
if (CsDiff < 8) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 991) {
return 8;
}
else if (Rs >= 991) {
return 2;
}
}
}
}
else if (CsDiff >= 8) {
if (Rs < 1240) {
return 2;
}
else if (Rs >= 1240) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (negBalance < 40) {
if (tcor < 75) {
return 1;
}
else if (tcor >= 75) {
if (Rs < 1297) {
return 8;
}
else if (Rs >= 1297) {
return 4;
}
}
}
else if (negBalance >= 40) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 93) {
if (AFD < 25) {
if (MVDiff < 24) {
if (posBalance < 54) {
if (Cs < 2052) {
if (mu_mv_mag_sq < 25) {
if (mcTcor < 95) {
return 2;
}
else if (mcTcor >= 95) {
if (Rs < 1117) {
return 2;
}
else if (Rs >= 1117) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
return 8;
}
}
}
else if (Cs >= 2052) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (posBalance >= 54) {
if (AFD < 17) {
if (negBalance < 45) {
if (mcTcor < 94) {
if (diffRsCsdiff < 11) {
return 8;
}
else if (diffRsCsdiff >= 11) {
if (Rs < 1414) {
return 2;
}
else if (Rs >= 1414) {
if (Rs < 1472) {
return 4;
}
else if (Rs >= 1472) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 1183) {
return 2;
}
else if (Rs >= 1183) {
return 4;
}
}
}
else if (negBalance >= 45) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 2;
}
}
}
else if (AFD >= 17) {
if (mu_mv_mag_sq < 5) {
if (negBalance < 55) {
if (Cs < 1818) {
return 4;
}
else if (Cs >= 1818) {
if (CsDiff < 12) {
return 1;
}
else if (CsDiff >= 12) {
return 4;
}
}
}
else if (negBalance >= 55) {
if (Cs < 2192) {
return 8;
}
else if (Cs >= 2192) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 2;
}
}
}
}
}
else if (MVDiff >= 24) {
if (mcTcor < 91) {
if (gchDC < 40) {
if (mcTcor < 80) {
if (TSC < 14) {
return 1;
}
else if (TSC >= 14) {
return 8;
}
}
else if (mcTcor >= 80) {
return 8;
}
}
else if (gchDC >= 40) {
return 4;
}
}
else if (mcTcor >= 91) {
if (Rs < 1236) {
if (Rs < 853) {
return 8;
}
else if (Rs >= 853) {
return 1;
}
}
else if (Rs >= 1236) {
if (Cs < 1490) {
return 8;
}
else if (Cs >= 1490) {
return 4;
}
}
}
}
}
else if (AFD >= 25) {
if (TSC < 11) {
if (Cs < 2235) {
return 4;
}
else if (Cs >= 2235) {
return 2;
}
}
else if (TSC >= 11) {
if (negBalance < 58) {
if (AFD < 38) {
return 8;
}
else if (AFD >= 38) {
return 2;
}
}
else if (negBalance >= 58) {
if (Cs < 2049) {
return 2;
}
else if (Cs >= 2049) {
if (CsDiff < 15) {
return 1;
}
else if (CsDiff >= 15) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 1512) {
if (Cs < 2730) {
if (AFD < 17) {
if (negBalance < 38) {
if (ssDCval < 100) {
if (MVDiff < 6) {
if (Rs < 2123) {
if (Rs < 1942) {
return 8;
}
else if (Rs >= 1942) {
return 4;
}
}
else if (Rs >= 2123) {
return 1;
}
}
else if (MVDiff >= 6) {
if (ssDCval < 69) {
if (Cs < 2254) {
if (AFD < 12) {
if (mu_mv_mag_sq < 82) {
return 1;
}
else if (mu_mv_mag_sq >= 82) {
return 8;
}
}
else if (AFD >= 12) {
return 2;
}
}
else if (Cs >= 2254) {
return 4;
}
}
else if (ssDCval >= 69) {
if (Cs < 2584) {
return 8;
}
else if (Cs >= 2584) {
return 4;
}
}
}
}
else if (ssDCval >= 100) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (negBalance >= 38) {
if (ssDCval < 52) {
if (Cs < 1621) {
if (tcor < 57) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 1617) {
return 4;
}
else if (Rs >= 1617) {
return 8;
}
}
}
else if (tcor >= 57) {
return 8;
}
}
else if (Cs >= 1621) {
if (negBalance < 39) {
return 8;
}
else if (negBalance >= 39) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (MVDiff < 5) {
if (Cs < 1998) {
return 8;
}
else if (Cs >= 1998) {
return 4;
}
}
else if (MVDiff >= 5) {
if (mu_mv_mag_sq < 27) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 4;
}
}
else if (mu_mv_mag_sq >= 27) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 52) {
if (Cs < 2299) {
if (gchDC < 2) {
if (TSC < 7) {
if (diffTSC < -2) {
if (Rs < 2099) {
return 1;
}
else if (Rs >= 2099) {
return 4;
}
}
else if (diffTSC >= -2) {
if (Rs < 1958) {
return 8;
}
else if (Rs >= 1958) {
if (ssDCval < 105) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (ssDCval >= 105) {
return 8;
}
}
}
}
else if (TSC >= 7) {
if (AFD < 16) {
if (Rs < 2745) {
if (negBalance < 43) {
if (MVDiff < 20) {
if (SC < 4777) {
return 8;
}
else if (SC >= 4777) {
if (Rs < 2654) {
return 1;
}
else if (Rs >= 2654) {
return 8;
}
}
}
else if (MVDiff >= 20) {
if (Rs < 1823) {
return 1;
}
else if (Rs >= 1823) {
return 8;
}
}
}
else if (negBalance >= 43) {
if (RsDiff < 6) {
if (Cs < 2222) {
return 8;
}
else if (Cs >= 2222) {
return 4;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (Rs >= 2745) {
if (AFD < 14) {
if (mcTcor < 96) {
if (Rs < 3623) {
return 8;
}
else if (Rs >= 3623) {
if (Rs < 3659) {
return 1;
}
else if (Rs >= 3659) {
return 8;
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (AFD >= 14) {
if (SC < 4965) {
return 4;
}
else if (SC >= 4965) {
return 8;
}
}
}
}
else if (AFD >= 16) {
if (Rs < 1900) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
if (Rs < 1880) {
return 8;
}
else if (Rs >= 1880) {
return 1;
}
}
}
}
else if (Rs >= 1900) {
if (diffRsCsdiff < -35) {
if (Rs < 2093) {
return 1;
}
else if (Rs >= 2093) {
return 8;
}
}
else if (diffRsCsdiff >= -35) {
return 8;
}
}
}
}
}
else if (gchDC >= 2) {
if (Rs < 1947) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (negBalance < 45) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
else if (negBalance >= 45) {
return 2;
}
}
}
else if (Rs >= 1947) {
if (ssDCval < 82) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
return 8;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
}
}
else if (Cs >= 2299) {
if (RsCsDiff < 80) {
if (Rs < 1869) {
return 4;
}
else if (Rs >= 1869) {
return 1;
}
}
else if (RsCsDiff >= 80) {
if (Rs < 3026) {
if (posBalance < 59) {
if (Rs < 2267) {
if (MVDiff < 3) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
else if (MVDiff >= 3) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 1865) {
return 1;
}
else if (Rs >= 1865) {
return 8;
}
}
}
}
else if (Rs >= 2267) {
if (Rs < 2891) {
if (mu_mv_mag_sq < 2) {
if (SC < 5146) {
return 8;
}
else if (SC >= 5146) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
else if (Rs >= 2891) {
if (Rs < 2951) {
return 4;
}
else if (Rs >= 2951) {
return 8;
}
}
}
}
else if (posBalance >= 59) {
if (negBalance < 56) {
if (diffAFD < 1) {
if (MVDiff < 2) {
if (Rs < 2781) {
if (Rs < 2347) {
return 1;
}
else if (Rs >= 2347) {
return 2;
}
}
else if (Rs >= 2781) {
return 8;
}
}
else if (MVDiff >= 2) {
if (RsCsDiff < 92) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (RsCsDiff >= 92) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (RsCsDiff < 115) {
return 4;
}
else if (RsCsDiff >= 115) {
if (Cs < 2354) {
return 4;
}
else if (Cs >= 2354) {
if (Cs < 2718) {
return 8;
}
else if (Cs >= 2718) {
return 2;
}
}
}
}
}
else if (negBalance >= 56) {
if (Rs < 2154) {
return 2;
}
else if (Rs >= 2154) {
return 4;
}
}
}
}
else if (Rs >= 3026) {
if (Rs < 3117) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 4;
}
}
else if (Rs >= 3117) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (ssDCval < 107) {
if (mu_mv_mag_sq < 68) {
return 4;
}
else if (mu_mv_mag_sq >= 68) {
return 8;
}
}
else if (ssDCval >= 107) {
return 8;
}
}
}
}
}
}
}
}
}
else if (AFD >= 17) {
if (MVDiff < 53) {
if (negBalance < 52) {
if (tcor < 81) {
if (ssDCval < 80) {
if (negBalance < 47) {
if (Rs < 1749) {
return 8;
}
else if (Rs >= 1749) {
return 2;
}
}
else if (negBalance >= 47) {
if (Cs < 1675) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 4;
}
}
else if (Cs >= 1675) {
if (posBalance < 65) {
return 1;
}
else if (posBalance >= 65) {
return 2;
}
}
}
}
else if (ssDCval >= 80) {
if (mcTcor < 93) {
if (negBalance < 45) {
if (Rs < 1533) {
return 8;
}
else if (Rs >= 1533) {
return 4;
}
}
else if (negBalance >= 45) {
if (Rs < 2332) {
return 8;
}
else if (Rs >= 2332) {
return 4;
}
}
}
else if (mcTcor >= 93) {
return 4;
}
}
}
else if (tcor >= 81) {
if (ssDCval < 87) {
return 2;
}
else if (ssDCval >= 87) {
return 1;
}
}
}
else if (negBalance >= 52) {
if (ssDCval < 120) {
if (TSC < 10) {
if (MVDiff < 15) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 2;
}
}
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (TSC >= 10) {
if (tcor < 72) {
if (mu_mv_mag_sq < 5) {
if (AFD < 18) {
if (Rs < 2095) {
return 8;
}
else if (Rs >= 2095) {
return 1;
}
}
else if (AFD >= 18) {
if (AFD < 23) {
if (Rs < 2167) {
return 4;
}
else if (Rs >= 2167) {
return 8;
}
}
else if (AFD >= 23) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 1636) {
return 8;
}
else if (Rs >= 1636) {
if (Rs < 1938) {
return 1;
}
else if (Rs >= 1938) {
return 8;
}
}
}
}
else if (tcor >= 72) {
if (RsDiff < 7) {
if (Cs < 2056) {
return 4;
}
else if (Cs >= 2056) {
return 8;
}
}
else if (RsDiff >= 7) {
if (CsDiff < 10) {
if (diffRsCsdiff < -55) {
return 4;
}
else if (diffRsCsdiff >= -55) {
if (Rs < 1977) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (Rs >= 1977) {
return 8;
}
}
}
else if (CsDiff >= 10) {
if (SC < 3768) {
if (Rs < 1641) {
return 8;
}
else if (Rs >= 1641) {
return 1;
}
}
else if (SC >= 3768) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 120) {
if (negBalance < 59) {
if (MVDiff < 18) {
if (Rs < 1696) {
return 4;
}
else if (Rs >= 1696) {
if (Rs < 2200) {
return 2;
}
else if (Rs >= 2200) {
return 8;
}
}
}
else if (MVDiff >= 18) {
return 8;
}
}
else if (negBalance >= 59) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 2;
}
}
}
}
}
else if (MVDiff >= 53) {
if (Cs < 1521) {
return 8;
}
else if (Cs >= 1521) {
return 1;
}
}
}
}
else if (Cs >= 2730) {
if (ssDCval < 103) {
if (negBalance < 48) {
if (Rs < 3028) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (Rs < 2692) {
return 4;
}
else if (Rs >= 2692) {
return 8;
}
}
}
else if (Rs >= 3028) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 99) {
if (Cs < 3771) {
return 1;
}
else if (Cs >= 3771) {
return 4;
}
}
else if (mu_mv_mag_sq >= 99) {
if (Cs < 3391) {
return 4;
}
else if (Cs >= 3391) {
return 8;
}
}
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (negBalance >= 48) {
if (refDCval < 87) {
if (mu_mv_mag_sq < 5) {
if (MVDiff < 5) {
if (AFD < 15) {
if (Cs < 2893) {
return 2;
}
else if (Cs >= 2893) {
if (Rs < 2809) {
return 4;
}
else if (Rs >= 2809) {
return 1;
}
}
}
else if (AFD >= 15) {
if (Cs < 3818) {
return 1;
}
else if (Cs >= 3818) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (tcor < 68) {
return 4;
}
else if (tcor >= 68) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (refDCval < 81) {
if (Cs < 3613) {
return 8;
}
else if (Cs >= 3613) {
return 4;
}
}
else if (refDCval >= 81) {
if (Rs < 2256) {
return 8;
}
else if (Rs >= 2256) {
return 4;
}
}
}
}
else if (refDCval >= 87) {
if (Rs < 1995) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
return 4;
}
}
else if (Rs >= 1995) {
if (Cs < 3920) {
if (tcor < 78) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (tcor >= 78) {
return 4;
}
}
else if (Cs >= 3920) {
if (Cs < 5131) {
return 8;
}
else if (Cs >= 5131) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 103) {
if (tcor < 91) {
if (diffMVdiffVal < 1) {
if (CsDiff < 12) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (Rs < 3186) {
return 1;
}
else if (Rs >= 3186) {
return 4;
}
}
}
else if (CsDiff >= 12) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (tcor >= 91) {
return 8;
}
}
}
}
}
else if (Rs >= 3687) {
if (mu_mv_mag_sq < 89) {
if (AFD < 12) {
if (CsDiff < 8) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (diffAFD < 2) {
if (MVDiff < 5) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 4486) {
return 4;
}
else if (Rs >= 4486) {
return 1;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
}
else if (CsDiff >= 8) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
if (MVDiff < 8) {
if (Cs < 2446) {
if (Cs < 1650) {
return 8;
}
else if (Cs >= 1650) {
if (RsCsDiff < 115) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
return 1;
}
}
else if (RsCsDiff >= 115) {
return 1;
}
}
}
else if (Cs >= 2446) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 89) {
if (Cs < 2824) {
if (Cs < 2648) {
if (mu_mv_mag_sq < 92) {
if (diffRsCsdiff < 30) {
return 1;
}
else if (diffRsCsdiff >= 30) {
return 8;
}
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
else if (Cs >= 2648) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (Rs < 4318) {
if (Cs < 2726) {
return 8;
}
else if (Cs >= 2726) {
return 4;
}
}
else if (Rs >= 4318) {
return 1;
}
}
}
}
else if (Cs >= 2824) {
if (Cs < 3701) {
if (ssDCval < 130) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Rs < 3970) {
return 4;
}
else if (Rs >= 3970) {
if (Cs < 2883) {
return 1;
}
else if (Cs >= 2883) {
return 8;
}
}
}
}
else if (ssDCval >= 130) {
if (Rs < 4158) {
return 8;
}
else if (Rs >= 4158) {
return 1;
}
}
}
else if (Cs >= 3701) {
if (ssDCval < 79) {
if (Rs < 5477) {
return 8;
}
else if (Rs >= 5477) {
return 2;
}
}
else if (ssDCval >= 79) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (Cs < 1922) {
if (Rs < 3341) {
if (ssDCval < 113) {
if (mcTcor < 87) {
if (AFD < 18) {
if (MVDiff < 16) {
if (Cs < 1676) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
if (TSC < 10) {
if (Rs < 2353) {
return 4;
}
else if (Rs >= 2353) {
return 2;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (Cs >= 1676) {
if (gchDC < 1) {
if (Cs < 1886) {
return 4;
}
else if (Cs >= 1886) {
return 8;
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 40) {
return 4;
}
else if (mu_mv_mag_sq >= 40) {
if (negBalance < 35) {
if (Rs < 2307) {
return 8;
}
else if (Rs >= 2307) {
return 1;
}
}
else if (negBalance >= 35) {
return 8;
}
}
}
}
}
else if (MVDiff >= 16) {
if (ssDCval < 59) {
if (tcor < 59) {
if (CsDiff < 9) {
if (Rs < 1594) {
return 8;
}
else if (Rs >= 1594) {
return 1;
}
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (tcor >= 59) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
if (Rs < 1816) {
return 1;
}
else if (Rs >= 1816) {
return 8;
}
}
}
}
else if (ssDCval >= 59) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2753) {
return 4;
}
else if (Rs >= 2753) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 1;
}
}
}
}
}
else if (AFD >= 18) {
if (RsCsDiff < 267) {
if (Cs < 1743) {
if (negBalance < 53) {
if (diffAFD < 1) {
if (CsDiff < 11) {
return 2;
}
else if (CsDiff >= 11) {
return 4;
}
}
else if (diffAFD >= 1) {
if (AFD < 28) {
if (gchDC < 1) {
if (Rs < 2067) {
return 1;
}
else if (Rs >= 2067) {
return 2;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (AFD >= 28) {
return 2;
}
}
}
else if (negBalance >= 53) {
if (RsCsDiff < 190) {
if (Rs < 1675) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 1551) {
return 8;
}
else if (Rs >= 1551) {
return 2;
}
}
}
else if (Rs >= 1675) {
if (Rs < 2655) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 8;
}
}
else if (Rs >= 2655) {
if (Rs < 2989) {
return 2;
}
else if (Rs >= 2989) {
return 4;
}
}
}
}
else if (RsCsDiff >= 190) {
if (mu_mv_mag_sq < 38) {
if (AFD < 31) {
return 4;
}
else if (AFD >= 31) {
return 8;
}
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 1994) {
if (Rs < 1254) {
return 1;
}
else if (Rs >= 1254) {
return 4;
}
}
else if (Rs >= 1994) {
return 8;
}
}
}
}
}
else if (Cs >= 1743) {
if (tcor < 59) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 1;
}
}
else if (tcor >= 59) {
if (refDCval < 86) {
if (negBalance < 57) {
return 4;
}
else if (negBalance >= 57) {
return 8;
}
}
else if (refDCval >= 86) {
if (Rs < 2167) {
return 1;
}
else if (Rs >= 2167) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 267) {
if (CsDiff < 12) {
if (refDCval < 77) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 8;
}
}
else if (refDCval >= 77) {
if (AFD < 24) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
if (mu_mv_mag_sq < 13) {
return 8;
}
else if (mu_mv_mag_sq >= 13) {
return 4;
}
}
}
else if (AFD >= 24) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (SC < 3439) {
return 4;
}
else if (SC >= 3439) {
return 2;
}
}
}
}
}
else if (CsDiff >= 12) {
if (diffTSC < -1) {
if (Rs < 1336) {
if (Rs < 1053) {
return 2;
}
else if (Rs >= 1053) {
return 8;
}
}
else if (Rs >= 1336) {
if (Rs < 1696) {
return 1;
}
else if (Rs >= 1696) {
return 4;
}
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
}
else if (mcTcor >= 87) {
if (negBalance < 48) {
if (refDCval < 89) {
if (posBalance < 49) {
if (negBalance < 40) {
if (Cs < 1609) {
if (Rs < 2455) {
return 1;
}
else if (Rs >= 2455) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Cs < 1411) {
return 4;
}
else if (Cs >= 1411) {
return 2;
}
}
}
}
else if (Cs >= 1609) {
return 1;
}
}
else if (negBalance >= 40) {
if (Rs < 1884) {
return 8;
}
else if (Rs >= 1884) {
if (RsCsDiff < 138) {
if (ssDCval < 74) {
if (Rs < 2875) {
return 8;
}
else if (Rs >= 2875) {
return 4;
}
}
else if (ssDCval >= 74) {
return 2;
}
}
else if (RsCsDiff >= 138) {
return 2;
}
}
}
}
else if (posBalance >= 49) {
if (Rs < 2991) {
if (mu_mv_mag_sq < 50) {
if (SC < 3379) {
if (Cs < 1419) {
if (Rs < 1743) {
return 4;
}
else if (Rs >= 1743) {
return 1;
}
}
else if (Cs >= 1419) {
return 8;
}
}
else if (SC >= 3379) {
if (negBalance < 44) {
if (RsCsDiff < 133) {
return 4;
}
else if (RsCsDiff >= 133) {
return 2;
}
}
else if (negBalance >= 44) {
if (SC < 3541) {
return 2;
}
else if (SC >= 3541) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (Rs >= 2991) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
return 2;
}
}
}
}
else if (refDCval >= 89) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 93) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
else if (ssDCval >= 93) {
if (diffMVdiffVal < -8) {
if (Rs < 2654) {
return 4;
}
else if (Rs >= 2654) {
return 1;
}
}
else if (diffMVdiffVal >= -8) {
if (diffAFD < 2) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 9) {
if (Rs < 2950) {
return 4;
}
else if (Rs >= 2950) {
return 2;
}
}
else if (AFD >= 9) {
if (tcor < 80) {
if (MVDiff < 34) {
return 2;
}
else if (MVDiff >= 34) {
return 8;
}
}
else if (tcor >= 80) {
if (Rs < 1690) {
return 4;
}
else if (Rs >= 1690) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 48) {
if (RsDiff < 14) {
if (mcTcor < 95) {
if (refDCval < 85) {
if (ssDCval < 71) {
if (Rs < 1638) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (AFD < 25) {
return 1;
}
else if (AFD >= 25) {
return 4;
}
}
}
else if (Rs >= 1638) {
return 8;
}
}
else if (ssDCval >= 71) {
if (SC < 4636) {
if (mu_mv_mag_sq < 11) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (CsDiff < 11) {
if (negBalance < 55) {
if (Cs < 1709) {
if (posBalance < 64) {
return 4;
}
else if (posBalance >= 64) {
if (Rs < 1886) {
return 4;
}
else if (Rs >= 1886) {
return 8;
}
}
}
else if (Cs >= 1709) {
if (Rs < 2139) {
return 8;
}
else if (Rs >= 2139) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
}
else if (negBalance >= 55) {
if (TSC < 13) {
if (Cs < 1600) {
if (ssDCval < 81) {
return 2;
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (Cs >= 1600) {
if (RsCsDiff < 154) {
return 8;
}
else if (RsCsDiff >= 154) {
return 4;
}
}
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (CsDiff >= 11) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (ssDCval < 82) {
if (mu_mv_mag_sq < 50) {
return 8;
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 2626) {
return 1;
}
else if (Rs >= 2626) {
return 8;
}
}
}
else if (ssDCval >= 82) {
return 4;
}
}
}
else if (SC >= 4636) {
if (Rs < 2917) {
return 8;
}
else if (Rs >= 2917) {
return 2;
}
}
}
}
else if (refDCval >= 85) {
if (TSC < 10) {
if (tcor < 84) {
if (posBalance < 57) {
if (MVDiff < 25) {
return 2;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (posBalance >= 57) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (posBalance < 60) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
else if (posBalance >= 60) {
if (Rs < 2572) {
return 2;
}
else if (Rs >= 2572) {
return 4;
}
}
}
}
}
else if (tcor >= 84) {
return 8;
}
}
else if (TSC >= 10) {
if (Rs < 1657) {
if (Rs < 1547) {
return 2;
}
else if (Rs >= 1547) {
return 4;
}
}
else if (Rs >= 1657) {
if (Rs < 2935) {
if (AFD < 23) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 1;
}
}
else if (AFD >= 23) {
if (refDCval < 90) {
return 1;
}
else if (refDCval >= 90) {
if (Rs < 1891) {
return 4;
}
else if (Rs >= 1891) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
if (SC < 3810) {
return 4;
}
else if (SC >= 3810) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2935) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (AFD < 17) {
if (Rs < 2787) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
else if (Rs >= 2787) {
if (Rs < 3298) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 4;
}
}
else if (Rs >= 3298) {
return 4;
}
}
}
else if (AFD >= 17) {
if (negBalance < 59) {
if (Rs < 2271) {
return 2;
}
else if (Rs >= 2271) {
return 1;
}
}
else if (negBalance >= 59) {
if (tcor < 75) {
return 2;
}
else if (tcor >= 75) {
if (diffAFD < 6) {
return 4;
}
else if (diffAFD >= 6) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 14) {
if (diffMVdiffVal < -31) {
return 1;
}
else if (diffMVdiffVal >= -31) {
if (ssDCval < 100) {
if (negBalance < 58) {
if (diffRsCsdiff < 12) {
if (RsCsDiff < 387) {
return 4;
}
else if (RsCsDiff >= 387) {
return 2;
}
}
else if (diffRsCsdiff >= 12) {
if (refDCval < 81) {
if (RsDiff < 15) {
return 4;
}
else if (RsDiff >= 15) {
return 2;
}
}
else if (refDCval >= 81) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 2;
}
}
}
}
else if (negBalance >= 58) {
if (Rs < 2402) {
if (Cs < 1607) {
return 8;
}
else if (Cs >= 1607) {
return 4;
}
}
else if (Rs >= 2402) {
return 8;
}
}
}
else if (ssDCval >= 100) {
if (AFD < 27) {
if (refDCval < 105) {
return 2;
}
else if (refDCval >= 105) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 4;
}
}
}
else if (AFD >= 27) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (AFD < 18) {
if (RsCsDiff < 183) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 2516) {
if (CsDiff < 8) {
if (ssDCval < 158) {
return 8;
}
else if (ssDCval >= 158) {
return 2;
}
}
else if (CsDiff >= 8) {
return 2;
}
}
else if (Rs >= 2516) {
return 8;
}
}
}
else if (RsCsDiff >= 183) {
if (Rs < 2905) {
if (posBalance < 38) {
if (Rs < 2371) {
return 2;
}
else if (Rs >= 2371) {
return 1;
}
}
else if (posBalance >= 38) {
if (TSC < 12) {
if (Rs < 1403) {
return 1;
}
else if (Rs >= 1403) {
return 8;
}
}
else if (TSC >= 12) {
if (Rs < 2105) {
return 8;
}
else if (Rs >= 2105) {
return 1;
}
}
}
}
else if (Rs >= 2905) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
}
}
else if (AFD >= 18) {
if (mu_mv_mag_sq < 14) {
if (RsDiff < 13) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
if (diffAFD < 2) {
if (negBalance < 62) {
if (Cs < 1702) {
return 1;
}
else if (Cs >= 1702) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
else if (negBalance >= 62) {
return 8;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (RsDiff >= 13) {
if (Rs < 2887) {
if (diffMVdiffVal < -22) {
if (AFD < 25) {
if (Rs < 2239) {
return 4;
}
else if (Rs >= 2239) {
return 8;
}
}
else if (AFD >= 25) {
return 2;
}
}
else if (diffMVdiffVal >= -22) {
return 8;
}
}
else if (Rs >= 2887) {
if (Rs < 3166) {
return 1;
}
else if (Rs >= 3166) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (AFD < 34) {
if (mu_mv_mag_sq < 49) {
if (gchDC < 5) {
if (ssDCval < 160) {
if (diffAFD < 10) {
return 8;
}
else if (diffAFD >= 10) {
return 2;
}
}
else if (ssDCval >= 160) {
return 2;
}
}
else if (gchDC >= 5) {
if (Cs < 1606) {
if (Rs < 2865) {
return 8;
}
else if (Rs >= 2865) {
return 4;
}
}
else if (Cs >= 1606) {
if (Cs < 1646) {
return 4;
}
else if (Cs >= 1646) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
if (Rs < 2498) {
return 1;
}
else if (Rs >= 2498) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (Cs < 1742) {
return 2;
}
else if (Cs >= 1742) {
return 4;
}
}
}
else if (AFD >= 34) {
if (RsCsDiff < 503) {
if (Rs < 1627) {
return 4;
}
else if (Rs >= 1627) {
return 2;
}
}
else if (RsCsDiff >= 503) {
return 1;
}
}
}
}
}
}
else if (Rs >= 3341) {
if (tcor < 86) {
if (RsCsDiff < 183) {
if (posBalance < 52) {
if (refDCval < 87) {
if (RsCsDiff < 175) {
return 4;
}
else if (RsCsDiff >= 175) {
return 2;
}
}
else if (refDCval >= 87) {
if (ssDCval < 124) {
if (Rs < 4325) {
return 8;
}
else if (Rs >= 4325) {
return 1;
}
}
else if (ssDCval >= 124) {
return 4;
}
}
}
else if (posBalance >= 52) {
if (diffMVdiffVal < -13) {
if (Rs < 3475) {
return 8;
}
else if (Rs >= 3475) {
return 1;
}
}
else if (diffMVdiffVal >= -13) {
if (ssDCval < 77) {
if (Rs < 3495) {
return 8;
}
else if (Rs >= 3495) {
return 2;
}
}
else if (ssDCval >= 77) {
return 8;
}
}
}
}
else if (RsCsDiff >= 183) {
if (mu_mv_mag_sq < 10) {
if (refDCval < 101) {
return 2;
}
else if (refDCval >= 101) {
if (TSC < 14) {
if (RsCsDiff < 283) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (RsCsDiff >= 283) {
if (SC < 6412) {
if (RsCsDiff < 473) {
if (mu_mv_mag_sq < 2) {
if (RsCsDiff < 332) {
return 8;
}
else if (RsCsDiff >= 332) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
else if (RsCsDiff >= 473) {
if (ssDCval < 143) {
return 2;
}
else if (ssDCval >= 143) {
return 1;
}
}
}
else if (SC >= 6412) {
if (Rs < 6316) {
return 8;
}
else if (Rs >= 6316) {
if (ssDCval < 132) {
return 8;
}
else if (ssDCval >= 132) {
if (RsDiff < 25) {
if (ssDCval < 138) {
if (Rs < 6489) {
return 8;
}
else if (Rs >= 6489) {
return 2;
}
}
else if (ssDCval >= 138) {
return 8;
}
}
else if (RsDiff >= 25) {
return 2;
}
}
}
}
}
}
else if (TSC >= 14) {
if (Rs < 3432) {
return 2;
}
else if (Rs >= 3432) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (refDCval < 125) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (Cs < 1730) {
if (MVDiff < 63) {
if (Cs < 1590) {
if (MVDiff < 45) {
return 2;
}
else if (MVDiff >= 45) {
if (Cs < 1397) {
return 2;
}
else if (Cs >= 1397) {
return 4;
}
}
}
else if (Cs >= 1590) {
if (RsDiff < 18) {
if (mu_mv_mag_sq < 19) {
if (Rs < 4291) {
return 4;
}
else if (Rs >= 4291) {
if (Rs < 5647) {
return 8;
}
else if (Rs >= 5647) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 19) {
return 1;
}
}
else if (RsDiff >= 18) {
return 2;
}
}
}
else if (MVDiff >= 63) {
if (Rs < 5718) {
if (Rs < 4635) {
return 4;
}
else if (Rs >= 4635) {
return 2;
}
}
else if (Rs >= 5718) {
return 1;
}
}
}
else if (Cs >= 1730) {
if (mu_mv_mag_sq < 76) {
if (RsDiff < 19) {
if (tcor < 83) {
if (ssDCval < 89) {
if (diffTSC < -1) {
if (Cs < 1907) {
return 8;
}
else if (Cs >= 1907) {
return 2;
}
}
else if (diffTSC >= -1) {
if (Cs < 1886) {
return 4;
}
else if (Cs >= 1886) {
return 8;
}
}
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (tcor >= 83) {
return 2;
}
}
else if (RsDiff >= 19) {
if (AFD < 35) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 4587) {
return 4;
}
else if (Rs >= 4587) {
return 2;
}
}
}
else if (AFD >= 35) {
if (diffTSC < 3) {
return 1;
}
else if (diffTSC >= 3) {
if (Rs < 4544) {
return 8;
}
else if (Rs >= 4544) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (RsDiff < 13) {
if (Rs < 3616) {
return 8;
}
else if (Rs >= 3616) {
return 4;
}
}
else if (RsDiff >= 13) {
return 1;
}
}
}
}
}
else if (refDCval >= 125) {
if (gchDC < 3) {
if (Rs < 3875) {
if (SC < 5139) {
return 1;
}
else if (SC >= 5139) {
if (Rs < 3652) {
return 2;
}
else if (Rs >= 3652) {
return 8;
}
}
}
else if (Rs >= 3875) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
}
else if (gchDC >= 3) {
if (TSC < 12) {
if (Rs < 5516) {
return 2;
}
else if (Rs >= 5516) {
return 4;
}
}
else if (TSC >= 12) {
return 1;
}
}
}
}
}
}
else if (tcor >= 86) {
if (refDCval < 100) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (ssDCval < 62) {
if (Rs < 7647) {
if (Rs < 3684) {
return 8;
}
else if (Rs >= 3684) {
return 4;
}
}
else if (Rs >= 7647) {
return 2;
}
}
else if (ssDCval >= 62) {
if (mcTcor < 93) {
if (negBalance < 56) {
if (posBalance < 43) {
if (Cs < 1598) {
return 4;
}
else if (Cs >= 1598) {
return 2;
}
}
else if (posBalance >= 43) {
if (ssDCval < 96) {
return 2;
}
else if (ssDCval >= 96) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
}
}
else if (negBalance >= 56) {
if (Cs < 1519) {
return 1;
}
else if (Cs >= 1519) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (CsDiff < 4) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 4;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
}
else if (refDCval >= 100) {
if (RsDiff < 11) {
if (MVDiff < 5) {
if (tcor < 93) {
if (Rs < 3455) {
return 4;
}
else if (Rs >= 3455) {
return 1;
}
}
else if (tcor >= 93) {
return 2;
}
}
else if (MVDiff >= 5) {
if (mu_mv_mag_sq < 86) {
return 8;
}
else if (mu_mv_mag_sq >= 86) {
return 1;
}
}
}
else if (RsDiff >= 11) {
if (MVDiff < 17) {
if (Rs < 3989) {
if (Cs < 1680) {
return 2;
}
else if (Cs >= 1680) {
return 8;
}
}
else if (Rs >= 3989) {
if (Cs < 1327) {
if (Rs < 5495) {
if (Rs < 4464) {
return 4;
}
else if (Rs >= 4464) {
return 1;
}
}
else if (Rs >= 5495) {
return 2;
}
}
else if (Cs >= 1327) {
return 2;
}
}
}
else if (MVDiff >= 17) {
if (ssDCval < 137) {
if (ssDCval < 111) {
if (Cs < 1411) {
return 1;
}
else if (Cs >= 1411) {
return 2;
}
}
else if (ssDCval >= 111) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
return 1;
}
}
}
else if (ssDCval >= 137) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1922) {
if (posBalance < 50) {
if (ssDCval < 78) {
if (SC < 10214) {
if (mu_mv_mag_sq < 57) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (ssDCval < 65) {
if (posBalance < 46) {
if (Rs < 3389) {
if (refDCval < 85) {
return 1;
}
else if (refDCval >= 85) {
return 8;
}
}
else if (Rs >= 3389) {
if (RsDiff < 16) {
return 8;
}
else if (RsDiff >= 16) {
return 4;
}
}
}
else if (posBalance >= 46) {
if (mcTcor < 76) {
return 4;
}
else if (mcTcor >= 76) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 2;
}
}
}
}
else if (ssDCval >= 65) {
if (CsDiff < 15) {
if (refDCval < 78) {
return 2;
}
else if (refDCval >= 78) {
if (RsDiff < 16) {
return 8;
}
else if (RsDiff >= 16) {
return 2;
}
}
}
else if (CsDiff >= 15) {
if (Cs < 3565) {
return 4;
}
else if (Cs >= 3565) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (tcor < 87) {
if (diffRsCsdiff < 207) {
if (negBalance < 47) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (Rs < 5008) {
return 1;
}
else if (Rs >= 5008) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (Cs < 2511) {
return 8;
}
else if (Cs >= 2511) {
if (Cs < 3089) {
return 4;
}
else if (Cs >= 3089) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 207) {
return 8;
}
}
else if (tcor >= 87) {
if (Cs < 3365) {
if (Rs < 4406) {
if (negBalance < 41) {
if (Rs < 3081) {
return 4;
}
else if (Rs >= 3081) {
return 1;
}
}
else if (negBalance >= 41) {
return 8;
}
}
else if (Rs >= 4406) {
if (Rs < 6471) {
return 4;
}
else if (Rs >= 6471) {
return 8;
}
}
}
else if (Cs >= 3365) {
return 8;
}
}
}
}
else if (SC >= 10214) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (RsDiff < 20) {
return 2;
}
else if (RsDiff >= 20) {
return 8;
}
}
}
}
else if (ssDCval >= 78) {
if (SC < 9080) {
if (mcTcor < 86) {
if (Cs < 3822) {
if (negBalance < 48) {
if (SC < 8694) {
return 8;
}
else if (SC >= 8694) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (diffMVdiffVal < 1) {
if (RsDiff < 12) {
return 1;
}
else if (RsDiff >= 12) {
if (TSC < 16) {
return 8;
}
else if (TSC >= 16) {
return 2;
}
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
else if (Cs >= 3822) {
if (Rs < 3745) {
return 1;
}
else if (Rs >= 3745) {
return 2;
}
}
}
else if (mcTcor >= 86) {
if (ssDCval < 122) {
if (Cs < 2626) {
if (mcTcor < 94) {
if (RsCsDiff < 176) {
if (Rs < 5286) {
return 8;
}
else if (Rs >= 5286) {
return 4;
}
}
else if (RsCsDiff >= 176) {
if (diffMVdiffVal < -3) {
if (negBalance < 46) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 2;
}
}
else if (negBalance >= 46) {
if (Rs < 3125) {
return 4;
}
else if (Rs >= 3125) {
return 8;
}
}
}
else if (diffMVdiffVal >= -3) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
}
else if (mcTcor >= 94) {
if (SC < 4743) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
else if (SC >= 4743) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (Rs < 3914) {
return 2;
}
else if (Rs >= 3914) {
return 1;
}
}
}
}
}
}
else if (Cs >= 2626) {
if (SC < 8423) {
if (mcTcor < 88) {
if (Rs < 4689) {
return 4;
}
else if (Rs >= 4689) {
return 8;
}
}
else if (mcTcor >= 88) {
if (mu_mv_mag_sq < 98) {
if (TSC < 15) {
return 2;
}
else if (TSC >= 15) {
return 8;
}
}
else if (mu_mv_mag_sq >= 98) {
return 4;
}
}
}
else if (SC >= 8423) {
if (Rs < 5136) {
if (Rs < 5063) {
return 8;
}
else if (Rs >= 5063) {
return 2;
}
}
else if (Rs >= 5136) {
return 1;
}
}
}
}
else if (ssDCval >= 122) {
if (negBalance < 19) {
return 8;
}
else if (negBalance >= 19) {
if (negBalance < 55) {
if (Rs < 4126) {
if (RsCsDiff < 193) {
if (posBalance < 48) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (posBalance >= 48) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
return 2;
}
}
}
else if (RsCsDiff >= 193) {
if (ssDCval < 175) {
return 4;
}
else if (ssDCval >= 175) {
return 1;
}
}
}
else if (Rs >= 4126) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 4;
}
}
}
}
else if (negBalance >= 55) {
if (Cs < 2008) {
return 2;
}
else if (Cs >= 2008) {
return 8;
}
}
}
}
}
}
else if (SC >= 9080) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
if (ssDCval < 125) {
if (ssDCval < 86) {
if (Rs < 4668) {
return 8;
}
else if (Rs >= 4668) {
return 2;
}
}
else if (ssDCval >= 86) {
if (mu_mv_mag_sq < 77) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
if (posBalance < 47) {
if (Rs < 6873) {
return 2;
}
else if (Rs >= 6873) {
return 8;
}
}
else if (posBalance >= 47) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 6273) {
return 8;
}
else if (Rs >= 6273) {
if (Rs < 6521) {
return 4;
}
else if (Rs >= 6521) {
return 2;
}
}
}
}
}
else if (ssDCval >= 125) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (AFD < 26) {
if (gchDC < 9) {
return 8;
}
else if (gchDC >= 9) {
return 1;
}
}
else if (AFD >= 26) {
return 2;
}
}
}
}
}
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 23) {
if (ssDCval < 150) {
if (negBalance < 45) {
if (SC < 10185) {
if (MVDiff < 5) {
if (Cs < 2673) {
return 2;
}
else if (Cs >= 2673) {
return 4;
}
}
else if (MVDiff >= 5) {
if (Rs < 2081) {
if (gchDC < 4) {
if (Rs < 1472) {
return 2;
}
else if (Rs >= 1472) {
return 1;
}
}
else if (gchDC >= 4) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 4;
}
}
}
else if (Rs >= 2081) {
if (posBalance < 56) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (diffMVdiffVal < -1) {
if (Cs < 2935) {
return 2;
}
else if (Cs >= 2935) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 5666) {
return 8;
}
else if (Rs >= 5666) {
return 1;
}
}
}
}
else if (posBalance >= 56) {
if (Cs < 2021) {
return 2;
}
else if (Cs >= 2021) {
if (gchDC < 3) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (RsDiff < 15) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 8;
}
}
}
else if (RsDiff >= 15) {
return 2;
}
}
}
else if (gchDC >= 3) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
if (Cs < 2253) {
if (Rs < 4833) {
return 4;
}
else if (Rs >= 4833) {
return 8;
}
}
else if (Cs >= 2253) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 10185) {
return 1;
}
}
else if (negBalance >= 45) {
if (AFD < 21) {
if (ssDCval < 71) {
if (CsDiff < 13) {
if (tcor < 71) {
if (mcTcor < 76) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
else if (mcTcor >= 76) {
if (Cs < 2737) {
return 4;
}
else if (Cs >= 2737) {
return 1;
}
}
}
else if (tcor >= 71) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
if (Rs < 3084) {
return 4;
}
else if (Rs >= 3084) {
return 2;
}
}
}
}
else if (CsDiff >= 13) {
if (mcTcor < 90) {
if (negBalance < 58) {
return 4;
}
else if (negBalance >= 58) {
return 2;
}
}
else if (mcTcor >= 90) {
if (Rs < 3362) {
return 1;
}
else if (Rs >= 3362) {
return 2;
}
}
}
}
else if (ssDCval >= 71) {
if (refDCval < 100) {
if (refDCval < 79) {
if (tcor < 59) {
if (Rs < 1852) {
return 2;
}
else if (Rs >= 1852) {
return 4;
}
}
else if (tcor >= 59) {
if (Rs < 3317) {
if (posBalance < 64) {
if (diffAFD < -2) {
if (Rs < 3027) {
return 1;
}
else if (Rs >= 3027) {
return 2;
}
}
else if (diffAFD >= -2) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
if (diffRsCsdiff < -9) {
return 8;
}
else if (diffRsCsdiff >= -9) {
if (Rs < 2947) {
return 1;
}
else if (Rs >= 2947) {
return 4;
}
}
}
}
}
else if (posBalance >= 64) {
if (Cs < 2647) {
return 2;
}
else if (Cs >= 2647) {
return 4;
}
}
}
else if (Rs >= 3317) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
if (Rs < 3699) {
return 4;
}
else if (Rs >= 3699) {
return 2;
}
}
}
}
}
else if (refDCval >= 79) {
if (SC < 8261) {
if (mcTcor < 94) {
if (mcTcor < 77) {
return 4;
}
else if (mcTcor >= 77) {
if (Rs < 2743) {
if (negBalance < 57) {
if (AFD < 15) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (Rs < 2572) {
return 8;
}
else if (Rs >= 2572) {
return 4;
}
}
}
else if (AFD >= 15) {
if (gchDC < 2) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (Cs < 2492) {
return 8;
}
else if (Cs >= 2492) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (Rs < 2094) {
return 8;
}
else if (Rs >= 2094) {
return 4;
}
}
}
}
else if (negBalance >= 57) {
if (posBalance < 61) {
if (Cs < 2236) {
return 4;
}
else if (Cs >= 2236) {
return 8;
}
}
else if (posBalance >= 61) {
return 4;
}
}
}
else if (Rs >= 2743) {
if (RsCsDiff < 156) {
return 8;
}
else if (RsCsDiff >= 156) {
if (CsDiff < 13) {
if (AFD < 16) {
if (MVDiff < 14) {
if (tcor < 84) {
if (Rs < 2934) {
return 4;
}
else if (Rs >= 2934) {
if (Rs < 3885) {
return 2;
}
else if (Rs >= 3885) {
return 4;
}
}
}
else if (tcor >= 84) {
if (SC < 6930) {
return 4;
}
else if (SC >= 6930) {
if (Rs < 4038) {
return 2;
}
else if (Rs >= 4038) {
if (Rs < 4631) {
return 4;
}
else if (Rs >= 4631) {
return 2;
}
}
}
}
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (AFD >= 16) {
if (diffAFD < 3) {
if (negBalance < 56) {
if (Cs < 3657) {
if (Rs < 2928) {
if (Cs < 2152) {
return 4;
}
else if (Cs >= 2152) {
return 8;
}
}
else if (Rs >= 2928) {
if (CsDiff < 12) {
return 4;
}
else if (CsDiff >= 12) {
if (Rs < 3635) {
return 8;
}
else if (Rs >= 3635) {
return 4;
}
}
}
}
else if (Cs >= 3657) {
return 8;
}
}
else if (negBalance >= 56) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (diffAFD >= 3) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
return 2;
}
}
}
}
else if (CsDiff >= 13) {
if (negBalance < 57) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (negBalance >= 57) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (refDCval < 85) {
if (Rs < 2752) {
return 8;
}
else if (Rs >= 2752) {
if (Rs < 2957) {
return 2;
}
else if (Rs >= 2957) {
return 1;
}
}
}
else if (refDCval >= 85) {
if (tcor < 89) {
if (Rs < 3858) {
if (negBalance < 58) {
return 4;
}
else if (negBalance >= 58) {
return 2;
}
}
else if (Rs >= 3858) {
return 2;
}
}
else if (tcor >= 89) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
}
}
else if (SC >= 8261) {
if (RsDiff < 16) {
return 8;
}
else if (RsDiff >= 16) {
if (Rs < 5855) {
return 2;
}
else if (Rs >= 5855) {
return 8;
}
}
}
}
}
else if (refDCval >= 100) {
if (tcor < 69) {
return 8;
}
else if (tcor >= 69) {
if (diffRsCsdiff < 80) {
if (MVDiff < 5) {
if (ssDCval < 121) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
if (Rs < 3590) {
if (ssDCval < 110) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
else if (ssDCval >= 110) {
return 8;
}
}
else if (Rs >= 3590) {
if (Rs < 4949) {
return 8;
}
else if (Rs >= 4949) {
if (Rs < 5540) {
return 4;
}
else if (Rs >= 5540) {
return 2;
}
}
}
}
}
else if (ssDCval >= 121) {
return 8;
}
}
else if (MVDiff >= 5) {
if (RsCsDiff < 223) {
if (Rs < 3483) {
if (Cs < 2403) {
if (ssDCval < 148) {
return 8;
}
else if (ssDCval >= 148) {
return 1;
}
}
else if (Cs >= 2403) {
if (RsCsDiff < 183) {
return 4;
}
else if (RsCsDiff >= 183) {
if (Cs < 2848) {
return 8;
}
else if (Cs >= 2848) {
if (SC < 6377) {
return 4;
}
else if (SC >= 6377) {
return 8;
}
}
}
}
}
else if (Rs >= 3483) {
return 8;
}
}
else if (RsCsDiff >= 223) {
if (refDCval < 105) {
if (Rs < 3670) {
if (Rs < 2289) {
return 8;
}
else if (Rs >= 2289) {
return 2;
}
}
else if (Rs >= 3670) {
return 8;
}
}
else if (refDCval >= 105) {
if (SC < 5758) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
else if (SC >= 5758) {
if (AFD < 15) {
if (Cs < 2190) {
return 4;
}
else if (Cs >= 2190) {
return 8;
}
}
else if (AFD >= 15) {
if (posBalance < 61) {
if (mu_mv_mag_sq < 3) {
if (Rs < 6138) {
return 4;
}
else if (Rs >= 6138) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (posBalance >= 61) {
return 8;
}
}
}
}
}
}
}
else if (diffRsCsdiff >= 80) {
if (AFD < 16) {
if (Rs < 7544) {
return 8;
}
else if (Rs >= 7544) {
return 1;
}
}
else if (AFD >= 16) {
if (Cs < 2724) {
if (diffRsCsdiff < 144) {
return 1;
}
else if (diffRsCsdiff >= 144) {
if (Rs < 4601) {
return 8;
}
else if (Rs >= 4601) {
return 2;
}
}
}
else if (Cs >= 2724) {
if (Rs < 7782) {
return 4;
}
else if (Rs >= 7782) {
return 1;
}
}
}
}
}
}
}
}
else if (AFD >= 21) {
if (Cs < 3069) {
if (refDCval < 110) {
if (tcor < 76) {
if (RsDiff < 21) {
if (negBalance < 50) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (diffRsCsdiff < -26) {
if (Cs < 2191) {
return 2;
}
else if (Cs >= 2191) {
return 8;
}
}
else if (diffRsCsdiff >= -26) {
return 2;
}
}
}
else if (negBalance >= 50) {
if (tcor < 71) {
if (Rs < 3081) {
if (posBalance < 63) {
if (RsDiff < 13) {
if (Rs < 1678) {
if (CsDiff < 15) {
if (Rs < 1407) {
return 4;
}
else if (Rs >= 1407) {
return 8;
}
}
else if (CsDiff >= 15) {
return 2;
}
}
else if (Rs >= 1678) {
if (ssDCval < 91) {
return 1;
}
else if (ssDCval >= 91) {
return 4;
}
}
}
else if (RsDiff >= 13) {
if (ssDCval < 71) {
return 4;
}
else if (ssDCval >= 71) {
if (SC < 4815) {
return 2;
}
else if (SC >= 4815) {
return 8;
}
}
}
}
else if (posBalance >= 63) {
if (RsDiff < 11) {
if (mcTcor < 88) {
return 4;
}
else if (mcTcor >= 88) {
return 8;
}
}
else if (RsDiff >= 11) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 8;
}
}
}
}
else if (Rs >= 3081) {
if (ssDCval < 97) {
return 4;
}
else if (ssDCval >= 97) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (Rs < 3211) {
return 4;
}
else if (Rs >= 3211) {
if (Rs < 4613) {
return 1;
}
else if (Rs >= 4613) {
return 8;
}
}
}
}
}
}
else if (tcor >= 71) {
if (mcTcor < 90) {
if (Cs < 2191) {
if (Rs < 3231) {
return 8;
}
else if (Rs >= 3231) {
return 4;
}
}
else if (Cs >= 2191) {
if (RsCsDiff < 356) {
return 1;
}
else if (RsCsDiff >= 356) {
if (Rs < 4092) {
return 8;
}
else if (Rs >= 4092) {
return 2;
}
}
}
}
else if (mcTcor >= 90) {
if (diffRsCsdiff < -3) {
if (Cs < 2317) {
return 8;
}
else if (Cs >= 2317) {
return 4;
}
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
}
}
}
else if (RsDiff >= 21) {
if (posBalance < 61) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 2;
}
}
else if (posBalance >= 61) {
if (Rs < 4742) {
return 1;
}
else if (Rs >= 4742) {
return 2;
}
}
}
}
else if (tcor >= 76) {
if (mcTcor < 94) {
if (TSC < 14) {
if (ssDCval < 107) {
return 8;
}
else if (ssDCval >= 107) {
return 2;
}
}
else if (TSC >= 14) {
if (Rs < 3080) {
return 1;
}
else if (Rs >= 3080) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
if (Rs < 1851) {
return 4;
}
else if (Rs >= 1851) {
return 2;
}
}
}
}
}
else if (refDCval >= 110) {
if (negBalance < 58) {
if (TSC < 14) {
if (AFD < 24) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (AFD >= 24) {
return 2;
}
}
else if (TSC >= 14) {
if (RsCsDiff < 201) {
return 2;
}
else if (RsCsDiff >= 201) {
if (TSC < 17) {
return 8;
}
else if (TSC >= 17) {
return 1;
}
}
}
}
else if (negBalance >= 58) {
if (RsDiff < 15) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (RsDiff < 10) {
return 2;
}
else if (RsDiff >= 10) {
if (ssDCval < 137) {
if (Rs < 1613) {
return 8;
}
else if (Rs >= 1613) {
return 1;
}
}
else if (ssDCval >= 137) {
return 8;
}
}
}
}
else if (RsDiff >= 15) {
if (Rs < 3013) {
if (Cs < 2894) {
return 1;
}
else if (Cs >= 2894) {
return 8;
}
}
else if (Rs >= 3013) {
return 8;
}
}
}
}
}
else if (Cs >= 3069) {
if (mcTcor < 92) {
if (SC < 9919) {
if (CsDiff < 16) {
if (Cs < 3822) {
if (MVDiff < 5) {
if (Rs < 3534) {
return 8;
}
else if (Rs >= 3534) {
if (Cs < 3572) {
return 4;
}
else if (Cs >= 3572) {
return 2;
}
}
}
else if (MVDiff >= 5) {
if (refDCval < 122) {
if (RsCsDiff < 247) {
return 4;
}
else if (RsCsDiff >= 247) {
if (Rs < 6001) {
return 8;
}
else if (Rs >= 6001) {
return 4;
}
}
}
else if (refDCval >= 122) {
return 4;
}
}
}
else if (Cs >= 3822) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
if (Rs < 5115) {
return 1;
}
else if (Rs >= 5115) {
return 2;
}
}
}
}
else if (CsDiff >= 16) {
if (ssDCval < 116) {
if (refDCval < 98) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (Rs < 5105) {
if (TSC < 16) {
return 8;
}
else if (TSC >= 16) {
if (SC < 7923) {
return 4;
}
else if (SC >= 7923) {
return 8;
}
}
}
else if (Rs >= 5105) {
return 4;
}
}
}
else if (refDCval >= 98) {
if (Rs < 2183) {
return 4;
}
else if (Rs >= 2183) {
return 2;
}
}
}
else if (ssDCval >= 116) {
if (Rs < 4515) {
return 4;
}
else if (Rs >= 4515) {
if (TSC < 16) {
return 4;
}
else if (TSC >= 16) {
return 8;
}
}
}
}
}
else if (SC >= 9919) {
if (AFD < 40) {
if (RsCsDiff < 633) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 4;
}
}
else if (RsCsDiff >= 633) {
if (SC < 15586) {
return 4;
}
else if (SC >= 15586) {
return 8;
}
}
}
else if (AFD >= 40) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (posBalance < 56) {
if (Rs < 7762) {
if (Rs < 5391) {
return 2;
}
else if (Rs >= 5391) {
return 4;
}
}
else if (Rs >= 7762) {
return 1;
}
}
else if (posBalance >= 56) {
if (mu_mv_mag_sq < 3) {
return 8;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 3139) {
return 8;
}
else if (Rs >= 3139) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 150) {
if (ssDCval < 168) {
if (Rs < 3335) {
return 8;
}
else if (Rs >= 3335) {
if (posBalance < 55) {
if (Rs < 6482) {
return 8;
}
else if (Rs >= 6482) {
return 4;
}
}
else if (posBalance >= 55) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
if (Rs < 7252) {
return 2;
}
else if (Rs >= 7252) {
return 4;
}
}
}
}
}
else if (ssDCval >= 168) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (Cs < 3478) {
if (negBalance < 43) {
if (CsDiff < 8) {
if (tcor < 89) {
if (mcTcor < 83) {
if (Rs < 3668) {
return 1;
}
else if (Rs >= 3668) {
if (Rs < 3881) {
return 4;
}
else if (Rs >= 3881) {
return 8;
}
}
}
else if (mcTcor >= 83) {
return 8;
}
}
else if (tcor >= 89) {
if (Cs < 2255) {
if (RsCsDiff < 112) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (Rs < 4851) {
return 8;
}
else if (Rs >= 4851) {
return 4;
}
}
}
else if (RsCsDiff >= 112) {
return 2;
}
}
else if (Cs >= 2255) {
if (ssDCval < 95) {
return 4;
}
else if (ssDCval >= 95) {
if (negBalance < 41) {
if (Cs < 2316) {
return 8;
}
else if (Cs >= 2316) {
return 2;
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
}
}
else if (CsDiff >= 8) {
if (ssDCval < 66) {
if (RsDiff < 13) {
if (refDCval < 59) {
return 4;
}
else if (refDCval >= 59) {
return 8;
}
}
else if (RsDiff >= 13) {
if (Rs < 3601) {
return 2;
}
else if (Rs >= 3601) {
return 1;
}
}
}
else if (ssDCval >= 66) {
if (MVDiff < 58) {
if (Rs < 5238) {
if (ssDCval < 95) {
if (MVDiff < 4) {
if (Rs < 4119) {
if (Rs < 3750) {
return 4;
}
else if (Rs >= 3750) {
return 8;
}
}
else if (Rs >= 4119) {
return 2;
}
}
else if (MVDiff >= 4) {
if (Cs < 2557) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
return 4;
}
}
else if (Cs >= 2557) {
if (Rs < 3516) {
if (Rs < 2480) {
return 1;
}
else if (Rs >= 2480) {
return 4;
}
}
else if (Rs >= 3516) {
if (Rs < 4271) {
return 8;
}
else if (Rs >= 4271) {
return 4;
}
}
}
}
}
else if (ssDCval >= 95) {
if (negBalance < 33) {
if (refDCval < 122) {
if (Rs < 4960) {
return 8;
}
else if (Rs >= 4960) {
return 4;
}
}
else if (refDCval >= 122) {
return 2;
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
else if (Rs >= 5238) {
if (mcTcor < 92) {
if (Rs < 6500) {
return 8;
}
else if (Rs >= 6500) {
return 2;
}
}
else if (mcTcor >= 92) {
if (Cs < 2945) {
return 2;
}
else if (Cs >= 2945) {
if (Rs < 8001) {
return 4;
}
else if (Rs >= 8001) {
return 1;
}
}
}
}
}
else if (MVDiff >= 58) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (Rs < 2073) {
return 2;
}
else if (Rs >= 2073) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 43) {
if (mu_mv_mag_sq < 57) {
if (tcor < 69) {
if (ssDCval < 67) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
else if (ssDCval >= 67) {
if (gchDC < 6) {
if (Rs < 5866) {
if (AFD < 31) {
if (diffRsCsdiff < 24) {
if (Cs < 3212) {
if (Cs < 2653) {
if (RsDiff < 18) {
if (SC < 4506) {
return 1;
}
else if (SC >= 4506) {
return 8;
}
}
else if (RsDiff >= 18) {
return 4;
}
}
else if (Cs >= 2653) {
return 4;
}
}
else if (Cs >= 3212) {
if (Rs < 4647) {
return 1;
}
else if (Rs >= 4647) {
return 8;
}
}
}
else if (diffRsCsdiff >= 24) {
if (MVDiff < 17) {
if (RsDiff < 18) {
return 8;
}
else if (RsDiff >= 18) {
return 4;
}
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (AFD >= 31) {
if (CsDiff < 20) {
return 1;
}
else if (CsDiff >= 20) {
return 8;
}
}
}
else if (Rs >= 5866) {
if (Cs < 3307) {
return 4;
}
else if (Cs >= 3307) {
return 2;
}
}
}
else if (gchDC >= 6) {
if (Rs < 4057) {
return 8;
}
else if (Rs >= 4057) {
return 2;
}
}
}
}
else if (tcor >= 69) {
if (Cs < 2567) {
if (ssDCval < 101) {
if (diffRsCsdiff < 18) {
if (tcor < 76) {
if (mu_mv_mag_sq < 42) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 4;
}
}
else if (mu_mv_mag_sq >= 42) {
if (Rs < 2814) {
if (Rs < 1993) {
return 2;
}
else if (Rs >= 1993) {
return 1;
}
}
else if (Rs >= 2814) {
if (Rs < 3317) {
return 4;
}
else if (Rs >= 3317) {
return 8;
}
}
}
}
else if (tcor >= 76) {
if (MVDiff < 17) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (gchDC < 1) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
else if (gchDC >= 1) {
if (ssDCval < 92) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (ssDCval >= 92) {
return 2;
}
}
}
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (diffRsCsdiff >= 18) {
if (mu_mv_mag_sq < 39) {
return 8;
}
else if (mu_mv_mag_sq >= 39) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (posBalance < 62) {
if (RsCsDiff < 183) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 4;
}
}
else if (RsCsDiff >= 183) {
return 8;
}
}
else if (posBalance >= 62) {
if (Rs < 3556) {
return 4;
}
else if (Rs >= 3556) {
if (Rs < 4077) {
return 8;
}
else if (Rs >= 4077) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 101) {
if (AFD < 24) {
if (Rs < 4976) {
if (posBalance < 51) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
if (Rs < 3607) {
return 4;
}
else if (Rs >= 3607) {
return 8;
}
}
}
else if (posBalance >= 51) {
if (diffAFD < 12) {
if (ssDCval < 106) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
else if (ssDCval >= 106) {
return 8;
}
}
else if (diffAFD >= 12) {
if (Rs < 3300) {
return 1;
}
else if (Rs >= 3300) {
return 8;
}
}
}
}
else if (Rs >= 4976) {
if (SC < 7519) {
return 1;
}
else if (SC >= 7519) {
return 4;
}
}
}
else if (AFD >= 24) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 3368) {
return 1;
}
else if (Rs >= 3368) {
return 4;
}
}
}
}
}
else if (Cs >= 2567) {
if (ssDCval < 81) {
if (Rs < 4916) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
if (AFD < 20) {
if (refDCval < 78) {
return 8;
}
else if (refDCval >= 78) {
if (Rs < 3160) {
return 8;
}
else if (Rs >= 3160) {
return 1;
}
}
}
else if (AFD >= 20) {
return 1;
}
}
}
else if (Rs >= 4916) {
return 2;
}
}
else if (ssDCval >= 81) {
if (ssDCval < 116) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (TSC < 11) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
}
else if (TSC >= 11) {
if (diffTSC < 1) {
if (MVDiff < 27) {
if (Rs < 5347) {
if (Cs < 2602) {
return 8;
}
else if (Cs >= 2602) {
return 4;
}
}
else if (Rs >= 5347) {
return 8;
}
}
else if (MVDiff >= 27) {
return 8;
}
}
else if (diffTSC >= 1) {
if (Cs < 2639) {
return 4;
}
else if (Cs >= 2639) {
return 8;
}
}
}
}
}
else if (ssDCval >= 116) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
if (ssDCval < 134) {
return 4;
}
else if (ssDCval >= 134) {
if (Rs < 5144) {
return 8;
}
else if (Rs >= 5144) {
if (Rs < 5858) {
return 4;
}
else if (Rs >= 5858) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (ssDCval < 137) {
if (tcor < 68) {
if (AFD < 34) {
return 4;
}
else if (AFD >= 34) {
return 1;
}
}
else if (tcor >= 68) {
if (Rs < 4109) {
if (refDCval < 81) {
if (MVDiff < 8) {
if (posBalance < 57) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
if (Rs < 3566) {
return 8;
}
else if (Rs >= 3566) {
return 1;
}
}
}
else if (posBalance >= 57) {
if (negBalance < 48) {
if (Cs < 2859) {
return 8;
}
else if (Cs >= 2859) {
return 1;
}
}
else if (negBalance >= 48) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (TSC < 13) {
if (mu_mv_mag_sq < 68) {
if (Rs < 2814) {
return 8;
}
else if (Rs >= 2814) {
return 4;
}
}
else if (mu_mv_mag_sq >= 68) {
if (negBalance < 45) {
return 4;
}
else if (negBalance >= 45) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
return 4;
}
}
}
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (Cs < 2616) {
if (ssDCval < 89) {
if (Rs < 3416) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
return 4;
}
}
else if (Rs >= 3416) {
return 4;
}
}
else if (ssDCval >= 89) {
if (gchDC < 1) {
if (negBalance < 53) {
return 8;
}
else if (negBalance >= 53) {
if (Rs < 2994) {
return 4;
}
else if (Rs >= 2994) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (ssDCval < 104) {
if (Rs < 3469) {
if (Rs < 2319) {
return 2;
}
else if (Rs >= 2319) {
return 8;
}
}
else if (Rs >= 3469) {
return 2;
}
}
else if (ssDCval >= 104) {
if (Rs < 3111) {
if (Cs < 2112) {
return 8;
}
else if (Cs >= 2112) {
return 4;
}
}
else if (Rs >= 3111) {
return 8;
}
}
}
}
}
else if (Cs >= 2616) {
if (refDCval < 107) {
if (SC < 7014) {
if (Cs < 2952) {
if (diffAFD < 1) {
if (AFD < 12) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
return 4;
}
}
else if (AFD >= 12) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (Rs < 4017) {
return 8;
}
else if (Rs >= 4017) {
return 4;
}
}
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (Cs >= 2952) {
return 4;
}
}
else if (SC >= 7014) {
if (posBalance < 58) {
if (Rs < 3977) {
return 1;
}
else if (Rs >= 3977) {
return 4;
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
else if (refDCval >= 107) {
if (tcor < 87) {
if (ssDCval < 127) {
if (Cs < 3020) {
return 1;
}
else if (Cs >= 3020) {
return 8;
}
}
else if (ssDCval >= 127) {
return 4;
}
}
else if (tcor >= 87) {
return 8;
}
}
}
}
}
else if (Rs >= 4109) {
if (ssDCval < 104) {
if (Cs < 2416) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
if (Cs < 2152) {
return 2;
}
else if (Cs >= 2152) {
return 8;
}
}
}
else if (Cs >= 2416) {
if (diffTSC < 3) {
if (SC < 7045) {
return 4;
}
else if (SC >= 7045) {
if (diffAFD < -5) {
if (Rs < 5054) {
return 4;
}
else if (Rs >= 5054) {
return 2;
}
}
else if (diffAFD >= -5) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (CsDiff < 11) {
return 4;
}
else if (CsDiff >= 11) {
return 8;
}
}
}
}
}
else if (diffTSC >= 3) {
return 1;
}
}
}
else if (ssDCval >= 104) {
return 8;
}
}
}
}
else if (ssDCval >= 137) {
if (tcor < 86) {
if (ssDCval < 148) {
return 4;
}
else if (ssDCval >= 148) {
return 8;
}
}
else if (tcor >= 86) {
return 1;
}
}
}
}
}
else if (Cs >= 3478) {
if (Rs < 7741) {
if (negBalance < 52) {
if (AFD < 14) {
if (ssDCval < 92) {
if (negBalance < 47) {
if (Cs < 4593) {
if (Rs < 6527) {
return 4;
}
else if (Rs >= 6527) {
return 2;
}
}
else if (Cs >= 4593) {
return 8;
}
}
else if (negBalance >= 47) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 4;
}
}
}
else if (ssDCval >= 92) {
if (Cs < 3571) {
if (Rs < 5227) {
return 8;
}
else if (Rs >= 5227) {
if (Rs < 5396) {
return 1;
}
else if (Rs >= 5396) {
if (Rs < 6105) {
return 8;
}
else if (Rs >= 6105) {
return 4;
}
}
}
}
else if (Cs >= 3571) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
return 8;
}
}
}
}
else if (AFD >= 14) {
if (tcor < 71) {
if (Rs < 4455) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
if (SC < 7634) {
if (RsCsDiff < 611) {
return 8;
}
else if (RsCsDiff >= 611) {
return 1;
}
}
else if (SC >= 7634) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
if (Rs < 4162) {
return 8;
}
else if (Rs >= 4162) {
return 1;
}
}
}
}
}
else if (Rs >= 4455) {
if (RsDiff < 18) {
return 8;
}
else if (RsDiff >= 18) {
return 4;
}
}
}
else if (tcor >= 71) {
if (Rs < 5732) {
if (RsDiff < 17) {
if (ssDCval < 123) {
if (SC < 8196) {
if (TSC < 13) {
if (Cs < 3723) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (Cs >= 3723) {
return 2;
}
}
else if (TSC >= 13) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
return 2;
}
}
}
else if (SC >= 8196) {
if (Rs < 3597) {
return 2;
}
else if (Rs >= 3597) {
return 8;
}
}
}
else if (ssDCval >= 123) {
return 4;
}
}
else if (RsDiff >= 17) {
return 4;
}
}
else if (Rs >= 5732) {
if (TSC < 15) {
if (MVDiff < 5) {
if (Rs < 7297) {
return 4;
}
else if (Rs >= 7297) {
return 8;
}
}
else if (MVDiff >= 5) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
return 4;
}
}
}
else if (TSC >= 15) {
if (Rs < 7131) {
return 2;
}
else if (Rs >= 7131) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 52) {
if (negBalance < 63) {
if (Rs < 3465) {
if (mcTcor < 82) {
return 4;
}
else if (mcTcor >= 82) {
if (Cs < 5183) {
return 8;
}
else if (Cs >= 5183) {
return 2;
}
}
}
else if (Rs >= 3465) {
if (tcor < 62) {
return 4;
}
else if (tcor >= 62) {
if (posBalance < 54) {
if (diffTSC < 2) {
if (Rs < 4309) {
return 1;
}
else if (Rs >= 4309) {
return 8;
}
}
else if (diffTSC >= 2) {
return 4;
}
}
else if (posBalance >= 54) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
if (mu_mv_mag_sq < 30) {
if (Rs < 5984) {
return 8;
}
else if (Rs >= 5984) {
return 4;
}
}
else if (mu_mv_mag_sq >= 30) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 63) {
return 4;
}
}
}
else if (Rs >= 7741) {
if (AFD < 16) {
if (Rs < 8555) {
return 8;
}
else if (Rs >= 8555) {
return 1;
}
}
else if (AFD >= 16) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1638) {
if (Rs < 3098) {
if (Rs < 1455) {
if (ssDCval < 122) {
if (negBalance < 2) {
if (Rs < 1194) {
return 2;
}
else if (Rs >= 1194) {
return 8;
}
}
else if (negBalance >= 2) {
if (refDCval < 63) {
if (negBalance < 27) {
if (mu_mv_mag_sq < 128) {
if (AFD < 3) {
if (refDCval < 53) {
if (Cs < 1387) {
if (Cs < 1353) {
if (Rs < 1135) {
return 8;
}
else if (Rs >= 1135) {
return 4;
}
}
else if (Cs >= 1353) {
return 8;
}
}
else if (Cs >= 1387) {
if (RsDiff < 2) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (Rs < 982) {
return 1;
}
else if (Rs >= 982) {
return 4;
}
}
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (refDCval >= 53) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
else if (AFD >= 3) {
if (SC < 2563) {
return 8;
}
else if (SC >= 2563) {
if (SC < 2821) {
return 4;
}
else if (SC >= 2821) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1482) {
if (Cs < 1343) {
return 4;
}
else if (Cs >= 1343) {
return 8;
}
}
else if (Cs >= 1482) {
return 1;
}
}
}
else if (negBalance >= 27) {
if (Rs < 1258) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (Rs >= 1258) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 4;
}
}
}
}
}
else if (refDCval >= 63) {
if (ssDCval < 77) {
if (MVDiff < 3) {
if (Cs < 1417) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (diffRsCsdiff < -4) {
if (Rs < 1426) {
return 4;
}
else if (Rs >= 1426) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
if (posBalance < 55) {
if (ssDCval < 69) {
if (Rs < 1062) {
return 2;
}
else if (Rs >= 1062) {
return 8;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (posBalance >= 55) {
return 1;
}
}
}
}
else if (Cs >= 1417) {
return 8;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (ssDCval >= 77) {
if (Rs < 986) {
if (Rs < 905) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Rs >= 905) {
return 4;
}
}
else if (Rs >= 986) {
return 8;
}
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 2) {
if (diffAFD < -1) {
if (negBalance < 22) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (negBalance >= 22) {
if (Rs < 1379) {
return 4;
}
else if (Rs >= 1379) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (Rs < 1383) {
return 4;
}
else if (Rs >= 1383) {
return 8;
}
}
}
else if (AFD >= 2) {
if (posBalance < 40) {
if (tcor < 76) {
if (Rs < 848) {
return 2;
}
else if (Rs >= 848) {
return 8;
}
}
else if (tcor >= 76) {
return 1;
}
}
else if (posBalance >= 40) {
if (Cs < 1467) {
if (mcTcor < 96) {
if (TSC < 5) {
if (ssDCval < 129) {
return 4;
}
else if (ssDCval >= 129) {
return 8;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (mcTcor >= 96) {
return 8;
}
}
else if (Cs >= 1467) {
if (refDCval < 128) {
if (Rs < 1255) {
return 4;
}
else if (Rs >= 1255) {
return 8;
}
}
else if (refDCval >= 128) {
if (negBalance < 41) {
return 8;
}
else if (negBalance >= 41) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1455) {
if (ssDCval < 140) {
if (ssDCval < 86) {
if (tcor < 98) {
if (Rs < 2208) {
if (negBalance < 61) {
if (RsCsDiff < 11) {
if (MVDiff < 7) {
if (Rs < 1476) {
return 4;
}
else if (Rs >= 1476) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (Rs < 1861) {
return 4;
}
else if (Rs >= 1861) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (RsCsDiff >= 11) {
return 8;
}
}
else if (negBalance >= 61) {
return 4;
}
}
else if (Rs >= 2208) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (ssDCval < 45) {
return 4;
}
else if (ssDCval >= 45) {
if (tcor < 97) {
if (Rs < 2930) {
if (Rs < 2756) {
return 8;
}
else if (Rs >= 2756) {
return 8;
}
}
else if (Rs >= 2930) {
return 2;
}
}
else if (tcor >= 97) {
if (Rs < 2787) {
return 4;
}
else if (Rs >= 2787) {
return 2;
}
}
}
}
}
}
else if (tcor >= 98) {
if (negBalance < 19) {
if (SC < 4462) {
if (Cs < 1488) {
if (Cs < 1344) {
if (Cs < 1342) {
if (Cs < 1330) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (mu_mv_mag_sq < 127) {
if (Cs < 1326) {
if (Cs < 1317) {
return 2;
}
else if (Cs >= 1317) {
if (mu_mv_mag_sq < 123) {
return 2;
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
}
else if (Cs >= 1326) {
return 2;
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
}
else if (Cs >= 1330) {
return 8;
}
}
else if (Cs >= 1342) {
if (posBalance < 31) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
return 2;
}
}
else if (posBalance >= 31) {
return 2;
}
}
}
else if (Cs >= 1344) {
return 8;
}
}
else if (Cs >= 1488) {
if (Rs < 1902) {
if (AFD < 2) {
if (Cs < 1589) {
return 1;
}
else if (Cs >= 1589) {
return 8;
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 1902) {
return 8;
}
}
}
else if (SC >= 4462) {
return 1;
}
}
else if (negBalance >= 19) {
if (Rs < 1501) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
if (Rs < 1474) {
return 8;
}
else if (Rs >= 1474) {
return 4;
}
}
}
else if (Rs >= 1501) {
if (Rs < 1833) {
if (ssDCval < 81) {
return 4;
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (Rs >= 1833) {
if (Rs < 2446) {
if (CsDiff < 3) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (CsDiff >= 3) {
return 4;
}
}
else if (Rs >= 2446) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
if (Rs < 3009) {
return 4;
}
else if (Rs >= 3009) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 117) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (MVDiff < 103) {
if (ssDCval < 93) {
if (mu_mv_mag_sq < 115) {
return 2;
}
else if (mu_mv_mag_sq >= 115) {
if (Cs < 1324) {
return 2;
}
else if (Cs >= 1324) {
return 8;
}
}
}
else if (ssDCval >= 93) {
return 8;
}
}
else if (MVDiff >= 103) {
return 4;
}
}
}
else if (ssDCval >= 117) {
if (CsDiff < 2) {
if (diffAFD < -19) {
if (Cs < 1559) {
if (Rs < 2408) {
return 4;
}
else if (Rs >= 2408) {
return 1;
}
}
else if (Cs >= 1559) {
return 8;
}
}
else if (diffAFD >= -19) {
if (Rs < 2902) {
return 8;
}
else if (Rs >= 2902) {
if (Rs < 3070) {
return 2;
}
else if (Rs >= 3070) {
return 8;
}
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 120) {
return 8;
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 2233) {
return 1;
}
else if (Rs >= 2233) {
if (Rs < 2905) {
return 8;
}
else if (Rs >= 2905) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (Rs < 1779) {
if (Rs < 1710) {
return 8;
}
else if (Rs >= 1710) {
return 2;
}
}
else if (Rs >= 1779) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Rs < 2545) {
return 8;
}
else if (Rs >= 2545) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3098) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 125) {
if (SC < 4924) {
if (AFD < 3) {
if (Cs < 1527) {
return 1;
}
else if (Cs >= 1527) {
return 4;
}
}
else if (AFD >= 3) {
return 2;
}
}
else if (SC >= 4924) {
if (posBalance < 45) {
if (AFD < 4) {
if (negBalance < 21) {
if (Rs < 5520) {
if (ssDCval < 61) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (Rs >= 5520) {
return 8;
}
}
else if (negBalance >= 21) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 123) {
if (Rs < 4163) {
return 1;
}
else if (Rs >= 4163) {
return 8;
}
}
else if (mu_mv_mag_sq >= 123) {
return 4;
}
}
}
}
else if (AFD >= 4) {
return 4;
}
}
else if (posBalance >= 45) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1568) {
if (Rs < 3673) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (Rs < 3539) {
return 1;
}
else if (Rs >= 3539) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (Rs >= 3673) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (Cs >= 1568) {
if (Rs < 3238) {
return 8;
}
else if (Rs >= 3238) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 104) {
if (posBalance < 42) {
if (negBalance < 28) {
if (Rs < 4478) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 4;
}
}
else if (Rs >= 4478) {
return 2;
}
}
else if (negBalance >= 28) {
if (posBalance < 38) {
if (Rs < 4493) {
return 4;
}
else if (Rs >= 4493) {
return 2;
}
}
else if (posBalance >= 38) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (ssDCval < 84) {
return 2;
}
else if (ssDCval >= 84) {
if (SC < 5720) {
if (Rs < 3369) {
if (Rs < 3139) {
return 4;
}
else if (Rs >= 3139) {
return 2;
}
}
else if (Rs >= 3369) {
return 4;
}
}
else if (SC >= 5720) {
if (Cs < 1498) {
return 2;
}
else if (Cs >= 1498) {
return 8;
}
}
}
}
}
else if (ssDCval >= 104) {
if (tcor < 81) {
if (SC < 4798) {
return 2;
}
else if (SC >= 4798) {
return 4;
}
}
else if (tcor >= 81) {
return 2;
}
}
}
}
}
else if (Cs >= 1638) {
if (ssDCval < 102) {
if (posBalance < 42) {
if (Rs < 3558) {
if (refDCval < 61) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (negBalance < 16) {
if (MVDiff < 2) {
if (posBalance < 6) {
return 4;
}
else if (posBalance >= 6) {
return 1;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (negBalance >= 16) {
if (SC < 3090) {
if (Rs < 1021) {
if (Rs < 830) {
return 1;
}
else if (Rs >= 830) {
return 8;
}
}
else if (Rs >= 1021) {
return 4;
}
}
else if (SC >= 3090) {
if (negBalance < 17) {
if (Cs < 1947) {
return 8;
}
else if (Cs >= 1947) {
return 1;
}
}
else if (negBalance >= 17) {
return 8;
}
}
}
}
}
else if (refDCval >= 61) {
if (negBalance < 48) {
if (ssDCval < 69) {
if (RsCsDiff < 7) {
if (ssDCval < 65) {
if (Rs < 2974) {
if (Rs < 2645) {
return 8;
}
else if (Rs >= 2645) {
if (Cs < 1944) {
return 8;
}
else if (Cs >= 1944) {
return 4;
}
}
}
else if (Rs >= 2974) {
return 8;
}
}
else if (ssDCval >= 65) {
if (Cs < 1741) {
return 8;
}
else if (Cs >= 1741) {
if (Cs < 1765) {
return 1;
}
else if (Cs >= 1765) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 7) {
if (SC < 4710) {
return 8;
}
else if (SC >= 4710) {
if (negBalance < 27) {
if (gchDC < 1) {
if (SC < 4858) {
if (negBalance < 25) {
if (Rs < 3029) {
if (Rs < 2956) {
return 4;
}
else if (Rs >= 2956) {
if (Rs < 2975) {
return 4;
}
else if (Rs >= 2975) {
return 4;
}
}
}
else if (Rs >= 3029) {
return 8;
}
}
else if (negBalance >= 25) {
return 4;
}
}
else if (SC >= 4858) {
if (Cs < 1820) {
return 4;
}
else if (Cs >= 1820) {
if (Cs < 1925) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 3108) {
if (Rs < 2988) {
return 8;
}
else if (Rs >= 2988) {
if (Rs < 3043) {
return 4;
}
else if (Rs >= 3043) {
return 8;
}
}
}
else if (Rs >= 3108) {
return 8;
}
}
}
else if (Cs >= 1925) {
return 4;
}
}
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (negBalance >= 27) {
return 4;
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 127) {
if (TSC < 1) {
if (Rs < 2391) {
if (ssDCval < 92) {
if (Rs < 1934) {
return 8;
}
else if (Rs >= 1934) {
if (Rs < 2371) {
return 1;
}
else if (Rs >= 2371) {
return 2;
}
}
}
else if (ssDCval >= 92) {
if (negBalance < 8) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
return 2;
}
}
else if (negBalance >= 8) {
if (posBalance < 20) {
if (mu_mv_mag_sq < 118) {
if (Rs < 2356) {
if (Rs < 2294) {
return 2;
}
else if (Rs >= 2294) {
return 4;
}
}
else if (Rs >= 2356) {
return 2;
}
}
else if (mu_mv_mag_sq >= 118) {
return 4;
}
}
else if (posBalance >= 20) {
return 4;
}
}
}
}
else if (Rs >= 2391) {
if (Cs < 2216) {
return 8;
}
else if (Cs >= 2216) {
if (Rs < 3076) {
return 1;
}
else if (Rs >= 3076) {
return 8;
}
}
}
}
else if (TSC >= 1) {
if (negBalance < 22) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
if (Rs < 3397) {
return 8;
}
else if (Rs >= 3397) {
if (Rs < 3469) {
return 2;
}
else if (Rs >= 3469) {
return 8;
}
}
}
}
else if (negBalance >= 22) {
if (Rs < 3059) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (Cs < 1827) {
return 4;
}
else if (Cs >= 1827) {
return 8;
}
}
}
else if (Rs >= 3059) {
if (Cs < 2367) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
else if (Cs >= 2367) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (SC < 4499) {
if (SC < 4443) {
if (posBalance < 40) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (posBalance >= 40) {
if (Rs < 2481) {
return 8;
}
else if (Rs >= 2481) {
return 4;
}
}
}
else if (SC >= 4443) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
return 4;
}
}
}
else if (SC >= 4499) {
if (Cs < 3351) {
return 8;
}
else if (Cs >= 3351) {
return 4;
}
}
}
}
}
else if (negBalance >= 48) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
return 8;
}
}
}
}
}
else if (Rs >= 3558) {
if (Cs < 2196) {
if (Rs < 3843) {
if (posBalance < 22) {
return 8;
}
else if (posBalance >= 22) {
return 2;
}
}
else if (Rs >= 3843) {
if (Rs < 8473) {
if (CsDiff < 3) {
if (ssDCval < 66) {
if (Rs < 4631) {
return 8;
}
else if (Rs >= 4631) {
return 4;
}
}
else if (ssDCval >= 66) {
return 4;
}
}
else if (CsDiff >= 3) {
if (negBalance < 29) {
if (ssDCval < 86) {
if (posBalance < 23) {
return 4;
}
else if (posBalance >= 23) {
if (Cs < 1741) {
return 4;
}
else if (Cs >= 1741) {
if (tcor < 96) {
if (Cs < 1976) {
return 4;
}
else if (Cs >= 1976) {
if (posBalance < 36) {
if (Rs < 4155) {
return 8;
}
else if (Rs >= 4155) {
if (Cs < 2007) {
return 4;
}
else if (Cs >= 2007) {
if (negBalance < 13) {
return 4;
}
else if (negBalance >= 13) {
if (gchDC < 1) {
if (CsDiff < 4) {
if (Rs < 4339) {
if (Rs < 4242) {
return 4;
}
else if (Rs >= 4242) {
return 4;
}
}
else if (Rs >= 4339) {
return 4;
}
}
else if (CsDiff >= 4) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (Rs < 4340) {
return 8;
}
else if (Rs >= 4340) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 36) {
if (Rs < 4262) {
return 8;
}
else if (Rs >= 4262) {
return 8;
}
}
}
}
else if (tcor >= 96) {
return 8;
}
}
}
}
else if (ssDCval >= 86) {
if (Cs < 1877) {
return 4;
}
else if (Cs >= 1877) {
return 2;
}
}
}
else if (negBalance >= 29) {
if (ssDCval < 91) {
if (Rs < 4685) {
return 2;
}
else if (Rs >= 4685) {
return 8;
}
}
else if (ssDCval >= 91) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
else if (Rs >= 8473) {
return 2;
}
}
}
else if (Cs >= 2196) {
if (Cs < 2719) {
if (AFD < 2) {
if (Cs < 2294) {
if (Rs < 4549) {
return 8;
}
else if (Rs >= 4549) {
return 4;
}
}
else if (Cs >= 2294) {
return 8;
}
}
else if (AFD >= 2) {
if (Rs < 4427) {
return 8;
}
else if (Rs >= 4427) {
if (diffMVdiffVal < 1) {
if (ssDCval < 75) {
if (mu_mv_mag_sq < 114) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (mu_mv_mag_sq >= 114) {
if (negBalance < 22) {
if (Rs < 5112) {
if (mu_mv_mag_sq < 115) {
return 8;
}
else if (mu_mv_mag_sq >= 115) {
if (diffRsCsdiff < 4) {
if (refDCval < 54) {
return 8;
}
else if (refDCval >= 54) {
return 4;
}
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
}
else if (Rs >= 5112) {
return 4;
}
}
else if (negBalance >= 22) {
return 8;
}
}
}
else if (ssDCval >= 75) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
if (Rs < 4893) {
if (diffAFD < 1) {
if (negBalance < 14) {
if (Rs < 4637) {
return 4;
}
else if (Rs >= 4637) {
return 4;
}
}
else if (negBalance >= 14) {
if (Rs < 4762) {
return 8;
}
else if (Rs >= 4762) {
return 8;
}
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (Rs >= 4893) {
return 8;
}
}
}
}
}
else if (Cs >= 2719) {
if (Rs < 4149) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (Rs >= 4149) {
if (posBalance < 41) {
if (Cs < 2796) {
if (Rs < 4900) {
return 8;
}
else if (Rs >= 4900) {
if (Cs < 2725) {
return 8;
}
else if (Cs >= 2725) {
if (posBalance < 34) {
return 4;
}
else if (posBalance >= 34) {
return 4;
}
}
}
}
else if (Cs >= 2796) {
return 8;
}
}
else if (posBalance >= 41) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
}
}
}
}
else if (posBalance >= 42) {
if (Rs < 6807) {
if (SC < 8342) {
if (ssDCval < 81) {
if (SC < 2930) {
if (negBalance < 25) {
if (Cs < 1735) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Cs >= 1735) {
return 1;
}
}
else if (negBalance >= 25) {
if (SC < 2928) {
if (ssDCval < 78) {
if (negBalance < 27) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
else if (negBalance >= 27) {
return 8;
}
}
else if (ssDCval >= 78) {
return 1;
}
}
else if (SC >= 2928) {
return 1;
}
}
}
else if (SC >= 2930) {
if (negBalance < 34) {
if (RsDiff < 3) {
if (SC < 5269) {
if (ssDCval < 66) {
if (MVDiff < 5) {
if (Cs < 2099) {
if (refDCval < 53) {
return 4;
}
else if (refDCval >= 53) {
if (SC < 4217) {
return 8;
}
else if (SC >= 4217) {
if (mu_mv_mag_sq < 128) {
if (tcor < 97) {
return 4;
}
else if (tcor >= 97) {
return 8;
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 2686) {
return 1;
}
else if (Rs >= 2686) {
return 4;
}
}
}
}
}
else if (Cs >= 2099) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (ssDCval >= 66) {
if (Cs < 1913) {
if (ssDCval < 69) {
if (Rs < 2506) {
return 1;
}
else if (Rs >= 2506) {
return 8;
}
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (Cs >= 1913) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
}
else if (SC >= 5269) {
if (negBalance < 4) {
return 4;
}
else if (negBalance >= 4) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (Cs < 2649) {
if (SC < 4340) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (SC >= 4340) {
if (Cs < 2156) {
if (RsDiff < 6) {
if (negBalance < 26) {
if (Rs < 4166) {
if (Rs < 3440) {
return 8;
}
else if (Rs >= 3440) {
return 4;
}
}
else if (Rs >= 4166) {
return 4;
}
}
else if (negBalance >= 26) {
if (Cs < 1739) {
return 8;
}
else if (Cs >= 1739) {
return 4;
}
}
}
else if (RsDiff >= 6) {
if (Rs < 3601) {
return 4;
}
else if (Rs >= 3601) {
return 2;
}
}
}
else if (Cs >= 2156) {
if (Rs < 2617) {
return 4;
}
else if (Rs >= 2617) {
if (Rs < 3658) {
return 8;
}
else if (Rs >= 3658) {
if (posBalance < 46) {
if (Rs < 4782) {
if (Rs < 4644) {
return 8;
}
else if (Rs >= 4644) {
return 4;
}
}
else if (Rs >= 4782) {
return 8;
}
}
else if (posBalance >= 46) {
if (Cs < 2634) {
return 4;
}
else if (Cs >= 2634) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 2649) {
if (Rs < 4920) {
return 8;
}
else if (Rs >= 4920) {
return 4;
}
}
}
}
else if (negBalance >= 34) {
if (diffMVdiffVal < -4) {
if (Cs < 1947) {
return 8;
}
else if (Cs >= 1947) {
return 4;
}
}
else if (diffMVdiffVal >= -4) {
if (posBalance < 62) {
if (MVDiff < 5) {
if (Rs < 2912) {
if (Cs < 3094) {
if (negBalance < 37) {
if (Cs < 2306) {
return 8;
}
else if (Cs >= 2306) {
if (Rs < 2471) {
return 4;
}
else if (Rs >= 2471) {
return 8;
}
}
}
else if (negBalance >= 37) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 8;
}
}
}
else if (Cs >= 3094) {
return 4;
}
}
else if (Rs >= 2912) {
if (diffAFD < -2) {
if (Cs < 3629) {
return 8;
}
else if (Cs >= 3629) {
return 4;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (Cs < 2189) {
return 4;
}
else if (Cs >= 2189) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
if (Rs < 3501) {
return 4;
}
else if (Rs >= 3501) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (Cs < 2615) {
if (Rs < 3704) {
if (Cs < 1932) {
if (SC < 4098) {
if (Rs < 1097) {
return 1;
}
else if (Rs >= 1097) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (SC >= 4098) {
if (MVDiff < 1) {
if (Cs < 1819) {
return 4;
}
else if (Cs >= 1819) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
if (mu_mv_mag_sq < 125) {
if (Rs < 2412) {
return 4;
}
else if (Rs >= 2412) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
return 4;
}
}
}
}
else if (MVDiff >= 1) {
if (Rs < 3092) {
return 8;
}
else if (Rs >= 3092) {
return 2;
}
}
}
}
else if (Cs >= 1932) {
if (Cs < 2518) {
if (posBalance < 43) {
if (Cs < 2119) {
if (Rs < 2301) {
return 8;
}
else if (Rs >= 2301) {
return 2;
}
}
else if (Cs >= 2119) {
return 8;
}
}
else if (posBalance >= 43) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
return 8;
}
}
}
else if (Cs >= 2518) {
if (negBalance < 45) {
if (Rs < 2312) {
return 4;
}
else if (Rs >= 2312) {
return 8;
}
}
else if (negBalance >= 45) {
return 4;
}
}
}
}
else if (Rs >= 3704) {
if (AFD < 8) {
if (RsDiff < 6) {
if (SC < 6290) {
if (mu_mv_mag_sq < 112) {
return 2;
}
else if (mu_mv_mag_sq >= 112) {
if (Cs < 1715) {
return 4;
}
else if (Cs >= 1715) {
return 8;
}
}
}
else if (SC >= 6290) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (Cs >= 2615) {
if (refDCval < 98) {
if (refDCval < 86) {
if (Cs < 3084) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (Cs >= 3084) {
if (Rs < 1697) {
return 2;
}
else if (Rs >= 1697) {
if (diffMVdiffVal < -2) {
if (Rs < 3799) {
return 8;
}
else if (Rs >= 3799) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
}
else if (refDCval >= 86) {
if (mu_mv_mag_sq < 124) {
if (Cs < 3533) {
if (Rs < 3140) {
if (negBalance < 50) {
if (Rs < 2522) {
return 4;
}
else if (Rs >= 2522) {
return 8;
}
}
else if (negBalance >= 50) {
return 4;
}
}
else if (Rs >= 3140) {
return 4;
}
}
else if (Cs >= 3533) {
if (ssDCval < 94) {
if (SC < 7579) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (SC >= 7579) {
return 8;
}
}
else if (ssDCval >= 94) {
if (Cs < 3554) {
return 8;
}
else if (Cs >= 3554) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (SC < 7365) {
if (Rs < 3395) {
return 8;
}
else if (Rs >= 3395) {
if (mcTcor < 99) {
if (Cs < 3729) {
if (ssDCval < 95) {
if (Cs < 2747) {
return 8;
}
else if (Cs >= 2747) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
if (Rs < 3539) {
return 8;
}
else if (Rs >= 3539) {
return 4;
}
}
}
}
else if (ssDCval >= 95) {
if (Rs < 3622) {
return 4;
}
else if (Rs >= 3622) {
return 8;
}
}
}
else if (Cs >= 3729) {
return 8;
}
}
else if (mcTcor >= 99) {
if (Rs < 3493) {
return 4;
}
else if (Rs >= 3493) {
if (Rs < 4133) {
return 8;
}
else if (Rs >= 4133) {
return 4;
}
}
}
}
}
else if (SC >= 7365) {
return 4;
}
}
}
}
else if (refDCval >= 98) {
if (AFD < 9) {
if (Cs < 4024) {
return 8;
}
else if (Cs >= 4024) {
return 4;
}
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 123) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (Cs < 4077) {
return 8;
}
else if (Cs >= 4077) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
}
}
}
}
else if (SC >= 8342) {
if (MVDiff < 3) {
if (negBalance < 24) {
if (Rs < 4577) {
return 1;
}
else if (Rs >= 4577) {
return 8;
}
}
else if (negBalance >= 24) {
if (Cs < 3666) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (Cs >= 3666) {
if (negBalance < 29) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Rs < 5276) {
return 1;
}
else if (Rs >= 5276) {
if (Rs < 6117) {
return 2;
}
else if (Rs >= 6117) {
return 8;
}
}
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 116) {
if (Rs < 5473) {
return 8;
}
else if (Rs >= 5473) {
return 2;
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
else if (Rs >= 6807) {
if (ssDCval < 95) {
return 4;
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
else if (ssDCval >= 102) {
if (posBalance < 53) {
if (ssDCval < 117) {
if (refDCval < 105) {
if (negBalance < 42) {
if (Cs < 1991) {
if (AFD < 5) {
if (Rs < 3027) {
return 2;
}
else if (Rs >= 3027) {
return 4;
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (Cs >= 1991) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (negBalance >= 42) {
if (Rs < 4897) {
return 8;
}
else if (Rs >= 4897) {
return 2;
}
}
}
else if (refDCval >= 105) {
if (diffTSC < -2) {
if (ssDCval < 108) {
if (Cs < 2017) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
return 8;
}
}
else if (Cs >= 2017) {
return 8;
}
}
else if (ssDCval >= 108) {
return 8;
}
}
else if (diffTSC >= -2) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (diffTSC < -6) {
if (tcor < 100) {
return 8;
}
else if (tcor >= 100) {
if (MVDiff < 10) {
if (Rs < 2671) {
return 8;
}
else if (Rs >= 2671) {
if (ssDCval < 135) {
return 1;
}
else if (ssDCval >= 135) {
return 8;
}
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
else if (diffTSC >= -6) {
if (Rs < 2602) {
if (Cs < 3652) {
if (tcor < 93) {
if (Cs < 1840) {
return 1;
}
else if (Cs >= 1840) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 1;
}
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (Cs >= 3652) {
return 1;
}
}
else if (Rs >= 2602) {
if (Cs < 1681) {
if (Rs < 3894) {
return 8;
}
else if (Rs >= 3894) {
return 2;
}
}
else if (Cs >= 1681) {
if (ssDCval < 200) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 35) {
if (Cs < 3017) {
if (SC < 5736) {
if (tcor < 97) {
return 4;
}
else if (tcor >= 97) {
if (AFD < 2) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (AFD >= 2) {
return 4;
}
}
}
else if (SC >= 5736) {
return 2;
}
}
else if (Cs >= 3017) {
return 8;
}
}
else if (posBalance >= 35) {
if (Rs < 6509) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
return 4;
}
}
else if (Rs >= 6509) {
if (Rs < 7159) {
return 4;
}
else if (Rs >= 7159) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (diffMVdiffVal < -7) {
if (posBalance < 19) {
return 8;
}
else if (posBalance >= 19) {
return 1;
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
}
else if (ssDCval >= 200) {
if (Rs < 5683) {
return 4;
}
else if (Rs >= 5683) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 53) {
if (Rs < 1705) {
if (Cs < 1872) {
if (mu_mv_mag_sq < 128) {
if (negBalance < 34) {
if (refDCval < 130) {
if (Rs < 814) {
return 1;
}
else if (Rs >= 814) {
return 4;
}
}
else if (refDCval >= 130) {
return 8;
}
}
else if (negBalance >= 34) {
return 8;
}
}
else if (mu_mv_mag_sq >= 128) {
return 2;
}
}
else if (Cs >= 1872) {
if (mcTcor < 92) {
if (Rs < 886) {
return 8;
}
else if (Rs >= 886) {
return 4;
}
}
else if (mcTcor >= 92) {
return 2;
}
}
}
else if (Rs >= 1705) {
if (ssDCval < 137) {
if (refDCval < 105) {
if (Cs < 2821) {
if (Rs < 4889) {
if (Rs < 3796) {
return 8;
}
else if (Rs >= 3796) {
return 1;
}
}
else if (Rs >= 4889) {
return 2;
}
}
else if (Cs >= 2821) {
return 8;
}
}
else if (refDCval >= 105) {
if (RsDiff < 2) {
if (ssDCval < 113) {
if (Rs < 2107) {
return 2;
}
else if (Rs >= 2107) {
return 8;
}
}
else if (ssDCval >= 113) {
return 8;
}
}
else if (RsDiff >= 2) {
if (posBalance < 63) {
if (diffAFD < -4) {
if (Rs < 5958) {
if (Rs < 4027) {
return 4;
}
else if (Rs >= 4027) {
return 2;
}
}
else if (Rs >= 5958) {
return 8;
}
}
else if (diffAFD >= -4) {
if (gchDC < 2) {
if (negBalance < 55) {
if (ssDCval < 105) {
if (Rs < 4823) {
return 8;
}
else if (Rs >= 4823) {
return 2;
}
}
else if (ssDCval >= 105) {
if (Cs < 1777) {
if (Cs < 1775) {
if (SC < 3494) {
if (Rs < 1821) {
return 8;
}
else if (Rs >= 1821) {
return 1;
}
}
else if (SC >= 3494) {
return 8;
}
}
else if (Cs >= 1775) {
return 1;
}
}
else if (Cs >= 1777) {
if (diffAFD < -3) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (diffAFD >= -3) {
if (negBalance < 38) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 6013) {
return 8;
}
else if (Rs >= 6013) {
return 4;
}
}
}
else if (negBalance >= 38) {
return 8;
}
}
}
}
}
else if (negBalance >= 55) {
if (Rs < 4187) {
if (Rs < 2472) {
return 8;
}
else if (Rs >= 2472) {
return 2;
}
}
else if (Rs >= 4187) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (posBalance >= 63) {
if (Cs < 1945) {
if (Rs < 2322) {
return 8;
}
else if (Rs >= 2322) {
return 1;
}
}
else if (Cs >= 1945) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
if (mcTcor < 96) {
if (Rs < 5949) {
return 1;
}
else if (Rs >= 5949) {
if (Rs < 8158) {
return 8;
}
else if (Rs >= 8158) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 5475) {
if (Rs < 3624) {
return 8;
}
else if (Rs >= 3624) {
return 1;
}
}
else if (Rs >= 5475) {
return 8;
}
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect2(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1313) {
if (Rs < 2587) {
if (Cs < 538) {
if (negBalance < 14) {
if (MVDiff < 7) {
if (SC < 792) {
if (refDCval < 47) {
if (SC < 277) {
if (negBalance < 8) {
if (Rs < 2) {
if (ssDCval < 22) {
if (ssDCval < 10) {
return 4;
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (Rs >= 2) {
return 8;
}
}
else if (negBalance >= 8) {
if (SC < 198) {
if (ssDCval < 40) {
if (refDCval < 25) {
if (mu_mv_mag_sq < 88) {
return 1;
}
else if (mu_mv_mag_sq >= 88) {
if (Rs < 3) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (Rs >= 3) {
return 8;
}
}
}
else if (refDCval >= 25) {
if (mu_mv_mag_sq < 77) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
return 8;
}
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 90) {
return 4;
}
else if (Rs >= 90) {
if (posBalance < 24) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (posBalance >= 24) {
return 8;
}
}
}
}
}
else if (ssDCval >= 40) {
if (diffMVdiffVal < -4) {
if (Rs < 54) {
return 8;
}
else if (Rs >= 54) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (SC >= 198) {
if (ssDCval < 35) {
if (ssDCval < 27) {
return 1;
}
else if (ssDCval >= 27) {
if (posBalance < 38) {
if (diffMVdiffVal < -213) {
return 4;
}
else if (diffMVdiffVal >= -213) {
return 8;
}
}
else if (posBalance >= 38) {
if (Cs < 111) {
return 8;
}
else if (Cs >= 111) {
return 1;
}
}
}
}
else if (ssDCval >= 35) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
}
else if (SC >= 277) {
if (Rs < 204) {
if (ssDCval < 36) {
if (tcor < 97) {
if (ssDCval < 28) {
return 1;
}
else if (ssDCval >= 28) {
return 8;
}
}
else if (tcor >= 97) {
if (Cs < 187) {
return 8;
}
else if (Cs >= 187) {
return 4;
}
}
}
else if (ssDCval >= 36) {
if (SC < 422) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 1;
}
}
else if (SC >= 422) {
if (tcor < 90) {
return 4;
}
else if (tcor >= 90) {
return 1;
}
}
}
}
else if (Rs >= 204) {
if (mu_mv_mag_sq < 118) {
if (diffMVdiffVal < -4) {
if (Rs < 313) {
return 2;
}
else if (Rs >= 313) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
if (RsDiff < 1) {
if (Cs < 270) {
if (Cs < 154) {
return 4;
}
else if (Cs >= 154) {
return 1;
}
}
else if (Cs >= 270) {
return 4;
}
}
else if (RsDiff >= 1) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (tcor < 101) {
if (diffMVdiffVal < -3) {
if (Cs < 153) {
return 4;
}
else if (Cs >= 153) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (TSC < 1) {
if (Cs < 409) {
return 4;
}
else if (Cs >= 409) {
return 1;
}
}
else if (TSC >= 1) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
else if (tcor >= 101) {
return 8;
}
}
}
}
}
else if (refDCval >= 47) {
if (ssDCval < 84) {
if (SC < 86) {
return 2;
}
else if (SC >= 86) {
if (ssDCval < 67) {
if (Cs < 281) {
if (negBalance < 13) {
if (diffMVdiffVal < 2) {
if (negBalance < 12) {
if (negBalance < 4) {
if (Rs < 268) {
return 8;
}
else if (Rs >= 268) {
if (Rs < 350) {
return 4;
}
else if (Rs >= 350) {
return 8;
}
}
}
else if (negBalance >= 4) {
return 8;
}
}
else if (negBalance >= 12) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (diffMVdiffVal >= 2) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (SC < 139) {
return 2;
}
else if (SC >= 139) {
return 8;
}
}
}
}
else if (negBalance >= 13) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (AFD < 2) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (Cs >= 281) {
if (Rs < 252) {
return 4;
}
else if (Rs >= 252) {
if (Rs < 344) {
return 1;
}
else if (Rs >= 344) {
return 8;
}
}
}
}
else if (ssDCval >= 67) {
if (Cs < 118) {
if (Rs < 145) {
return 4;
}
else if (Rs >= 145) {
return 8;
}
}
else if (Cs >= 118) {
if (Cs < 241) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
if (Rs < 385) {
return 8;
}
else if (Rs >= 385) {
return 2;
}
}
}
else if (Cs >= 241) {
if (ssDCval < 72) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
return 4;
}
}
else if (ssDCval >= 72) {
if (Rs < 318) {
return 1;
}
else if (Rs >= 318) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
if (AFD < 2) {
if (Rs < 467) {
return 4;
}
else if (Rs >= 467) {
return 1;
}
}
else if (AFD >= 2) {
if (Rs < 389) {
return 8;
}
else if (Rs >= 389) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 792) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 33) {
if (Cs < 289) {
return 4;
}
else if (Cs >= 289) {
if (Rs < 573) {
return 1;
}
else if (Rs >= 573) {
return 8;
}
}
}
else if (ssDCval >= 33) {
if (Cs < 267) {
if (SC < 1450) {
return 8;
}
else if (SC >= 1450) {
if (ssDCval < 87) {
if (Rs < 1366) {
return 2;
}
else if (Rs >= 1366) {
return 1;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
}
else if (Cs >= 267) {
if (Rs < 711) {
if (Rs < 409) {
if (mu_mv_mag_sq < 126) {
return 1;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 403) {
return 8;
}
else if (Rs >= 403) {
return 1;
}
}
}
else if (Rs >= 409) {
if (RsCsDiff < 1) {
if (Cs < 397) {
if (Rs < 622) {
if (negBalance < 4) {
return 8;
}
else if (negBalance >= 4) {
if (Rs < 487) {
return 8;
}
else if (Rs >= 487) {
return 1;
}
}
}
else if (Rs >= 622) {
return 8;
}
}
else if (Cs >= 397) {
if (SC < 967) {
if (posBalance < 24) {
return 1;
}
else if (posBalance >= 24) {
return 8;
}
}
else if (SC >= 967) {
if (Cs < 418) {
if (Rs < 602) {
return 1;
}
else if (Rs >= 602) {
return 4;
}
}
else if (Cs >= 418) {
if (Rs < 491) {
if (Cs < 508) {
return 1;
}
else if (Cs >= 508) {
return 8;
}
}
else if (Rs >= 491) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (Rs < 496) {
return 1;
}
else if (Rs >= 496) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
return 8;
}
}
}
}
}
else if (Rs >= 711) {
if (negBalance < 9) {
if (refDCval < 54) {
return 8;
}
else if (refDCval >= 54) {
if (ssDCval < 68) {
if (Cs < 317) {
return 1;
}
else if (Cs >= 317) {
if (Rs < 861) {
return 8;
}
else if (Rs >= 861) {
if (Rs < 1004) {
return 1;
}
else if (Rs >= 1004) {
return 4;
}
}
}
}
else if (ssDCval >= 68) {
return 8;
}
}
}
else if (negBalance >= 9) {
if (mu_mv_mag_sq < 127) {
if (posBalance < 47) {
if (ssDCval < 68) {
if (Cs < 457) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (Cs >= 457) {
if (Rs < 1395) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Rs < 793) {
return 2;
}
else if (Rs >= 793) {
return 8;
}
}
}
else if (Rs >= 1395) {
return 2;
}
}
}
else if (ssDCval >= 68) {
if (Cs < 333) {
return 2;
}
else if (Cs >= 333) {
return 8;
}
}
}
else if (posBalance >= 47) {
if (mu_mv_mag_sq < 114) {
if (Rs < 1988) {
return 1;
}
else if (Rs >= 1988) {
if (Rs < 2090) {
return 8;
}
else if (Rs >= 2090) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 10) {
return 1;
}
else if (negBalance >= 10) {
if (posBalance < 35) {
if (Rs < 1068) {
if (Cs < 287) {
return 1;
}
else if (Cs >= 287) {
return 8;
}
}
else if (Rs >= 1068) {
return 1;
}
}
else if (posBalance >= 35) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (Cs < 243) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
return 2;
}
}
else if (Cs >= 243) {
if (ssDCval < 61) {
if (Cs < 273) {
if (Rs < 736) {
return 8;
}
else if (Rs >= 736) {
return 1;
}
}
else if (Cs >= 273) {
return 8;
}
}
else if (ssDCval >= 61) {
if (ssDCval < 67) {
if (Cs < 386) {
return 1;
}
else if (Cs >= 386) {
if (Rs < 833) {
return 8;
}
else if (Rs >= 833) {
return 2;
}
}
}
else if (ssDCval >= 67) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (SC < 1092) {
return 4;
}
else if (SC >= 1092) {
if (Cs < 262) {
return 2;
}
else if (Cs >= 262) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (tcor < 94) {
if (negBalance < 5) {
if (MVDiff < 182) {
if (refDCval < 15) {
return 4;
}
else if (refDCval >= 15) {
if (Rs < 58) {
if (ssDCval < 51) {
if (mu_mv_mag_sq < 39) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 39) {
if (mcTcor < 61) {
if (MVDiff < 48) {
return 4;
}
else if (MVDiff >= 48) {
return 8;
}
}
else if (mcTcor >= 61) {
return 8;
}
}
}
else if (ssDCval >= 51) {
return 2;
}
}
else if (Rs >= 58) {
if (negBalance < 4) {
if (negBalance < 2) {
if (diffAFD < 5) {
if (diffMVdiffVal < 35) {
if (mu_mv_mag_sq < 56) {
return 4;
}
else if (mu_mv_mag_sq >= 56) {
if (Cs < 130) {
return 8;
}
else if (Cs >= 130) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 35) {
return 8;
}
}
else if (diffAFD >= 5) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 4;
}
}
}
else if (negBalance >= 2) {
return 8;
}
}
else if (negBalance >= 4) {
if (ssDCval < 87) {
if (Cs < 54) {
return 2;
}
else if (Cs >= 54) {
return 4;
}
}
else if (ssDCval >= 87) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 182) {
if (diffMVdiffVal < 36) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (AFD < 10) {
if (Rs < 196) {
return 8;
}
else if (Rs >= 196) {
return 4;
}
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (diffMVdiffVal >= 36) {
if (ssDCval < 36) {
if (negBalance < 1) {
if (ssDCval < 11) {
return 1;
}
else if (ssDCval >= 11) {
return 4;
}
}
else if (negBalance >= 1) {
if (Rs < 89) {
return 2;
}
else if (Rs >= 89) {
return 1;
}
}
}
else if (ssDCval >= 36) {
return 1;
}
}
}
}
else if (negBalance >= 5) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (posBalance < 40) {
if (ssDCval < 29) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
if (Rs < 192) {
return 1;
}
else if (Rs >= 192) {
return 4;
}
}
}
else if (ssDCval >= 29) {
if (tcor < 90) {
if (Cs < 458) {
return 8;
}
else if (Cs >= 458) {
return 1;
}
}
else if (tcor >= 90) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 2;
}
}
}
}
else if (posBalance >= 40) {
if (MVDiff < 13) {
if (refDCval < 49) {
if (posBalance < 81) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 8;
}
}
else if (posBalance >= 81) {
return 2;
}
}
else if (refDCval >= 49) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 4;
}
}
}
else if (MVDiff >= 13) {
if (diffMVdiffVal < -9) {
if (posBalance < 109) {
return 2;
}
else if (posBalance >= 109) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 958) {
return 8;
}
else if (Rs >= 958) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -9) {
if (refDCval < 135) {
if (Rs < 10) {
if (TSCindex < 2) {
return 2;
}
else if (TSCindex >= 2) {
return 8;
}
}
else if (Rs >= 10) {
if (MVDiff < 28) {
if (Rs < 1292) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
return 1;
}
}
else if (Rs >= 1292) {
return 8;
}
}
else if (MVDiff >= 28) {
if (tcor < 86) {
if (diffMVdiffVal < 22) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
return 2;
}
}
else if (diffMVdiffVal >= 22) {
if (refDCval < 42) {
return 1;
}
else if (refDCval >= 42) {
if (Rs < 1126) {
return 8;
}
else if (Rs >= 1126) {
return 1;
}
}
}
}
else if (tcor >= 86) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (Rs < 1245) {
return 8;
}
else if (Rs >= 1245) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 135) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 94) {
if (negBalance < 4) {
if (ssDCval < 125) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
return 8;
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (ssDCval < 29) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
return 8;
}
}
else if (ssDCval >= 29) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (Cs < 210) {
if (ssDCval < 109) {
return 4;
}
else if (ssDCval >= 109) {
if (Rs < 180) {
return 4;
}
else if (Rs >= 180) {
return 8;
}
}
}
else if (Cs >= 210) {
return 8;
}
}
}
}
}
else if (ssDCval >= 125) {
if (SC < 1309) {
if (MVDiff < 33) {
if (diffMVdiffVal < -11) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (Rs < 344) {
return 8;
}
else if (Rs >= 344) {
return 4;
}
}
}
else if (diffMVdiffVal >= -11) {
if (posBalance < 18) {
if (Rs < 582) {
return 2;
}
else if (Rs >= 582) {
return 8;
}
}
else if (posBalance >= 18) {
return 8;
}
}
}
else if (MVDiff >= 33) {
if (Rs < 132) {
if (MVDiff < 445) {
return 2;
}
else if (MVDiff >= 445) {
return 1;
}
}
else if (Rs >= 132) {
if (diffAFD < -6) {
if (ssDCval < 129) {
return 2;
}
else if (ssDCval >= 129) {
if (MVDiff < 45) {
return 8;
}
else if (MVDiff >= 45) {
return 4;
}
}
}
else if (diffAFD >= -6) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
}
else if (SC >= 1309) {
if (diffAFD < -8) {
return 2;
}
else if (diffAFD >= -8) {
return 1;
}
}
}
}
else if (negBalance >= 4) {
if (SC < 923) {
if (Cs < 116) {
if (Cs < 3) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (Cs >= 3) {
if (negBalance < 5) {
if (Cs < 100) {
return 8;
}
else if (Cs >= 100) {
return 4;
}
}
else if (negBalance >= 5) {
if (mu_mv_mag_sq < 19) {
if (Rs < 36) {
return 1;
}
else if (Rs >= 36) {
return 4;
}
}
else if (mu_mv_mag_sq >= 19) {
if (SC < 136) {
return 1;
}
else if (SC >= 136) {
if (ssDCval < 82) {
if (mcTcor < 97) {
return 1;
}
else if (mcTcor >= 97) {
return 8;
}
}
else if (ssDCval >= 82) {
return 1;
}
}
}
}
}
}
else if (Cs >= 116) {
if (mu_mv_mag_sq < 72) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 333) {
return 1;
}
else if (Rs >= 333) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (ssDCval < 88) {
if (Rs < 215) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (AFD < 1) {
if (Rs < 129) {
return 1;
}
else if (Rs >= 129) {
return 2;
}
}
else if (AFD >= 1) {
return 1;
}
}
}
else if (Rs >= 215) {
if (ssDCval < 37) {
return 4;
}
else if (ssDCval >= 37) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
}
else if (ssDCval >= 88) {
if (Rs < 422) {
return 4;
}
else if (Rs >= 422) {
if (ssDCval < 161) {
if (Cs < 394) {
return 1;
}
else if (Cs >= 394) {
return 2;
}
}
else if (ssDCval >= 161) {
return 4;
}
}
}
}
}
}
else if (SC >= 923) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (ssDCval < 143) {
return 1;
}
else if (ssDCval >= 143) {
if (Cs < 401) {
return 1;
}
else if (Cs >= 401) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (mcTcor < 82) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (tcor < 59) {
if (SC < 854) {
if (negBalance < 44) {
if (Cs < 317) {
if (posBalance < 75) {
if (TSCindex < 2) {
if (Rs < 31) {
return 1;
}
else if (Rs >= 31) {
return 2;
}
}
else if (TSCindex >= 2) {
return 1;
}
}
else if (posBalance >= 75) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (Cs >= 317) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (negBalance >= 44) {
if (Rs < 509) {
if (Rs < 16) {
return 4;
}
else if (Rs >= 16) {
return 8;
}
}
else if (Rs >= 509) {
return 1;
}
}
}
else if (SC >= 854) {
return 8;
}
}
else if (tcor >= 59) {
if (posBalance < 36) {
if (refDCval < 30) {
if (Rs < 275) {
if (posBalance < 6) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
return 4;
}
}
else if (posBalance >= 6) {
return 1;
}
}
else if (Rs >= 275) {
return 4;
}
}
else if (refDCval >= 30) {
if (Rs < 166) {
return 2;
}
else if (Rs >= 166) {
return 8;
}
}
}
else if (posBalance >= 36) {
if (ssDCval < 26) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (ssDCval >= 26) {
if (mcTcor < 72) {
return 2;
}
else if (mcTcor >= 72) {
if (TSC < 2) {
if (Rs < 486) {
return 1;
}
else if (Rs >= 486) {
return 4;
}
}
else if (TSC >= 2) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 82) {
if (posBalance < 27) {
if (posBalance < 6) {
if (mu_mv_mag_sq < 103) {
if (mu_mv_mag_sq < 28) {
if (refDCval < 25) {
if (mcTcor < 99) {
return 4;
}
else if (mcTcor >= 99) {
if (MVDiff < 67) {
return 4;
}
else if (MVDiff >= 67) {
return 8;
}
}
}
else if (refDCval >= 25) {
return 8;
}
}
else if (mu_mv_mag_sq >= 28) {
return 4;
}
}
else if (mu_mv_mag_sq >= 103) {
return 8;
}
}
else if (posBalance >= 6) {
if (MVDiff < 7) {
if (TSCindex < 2) {
if (tcor < 102) {
if (ssDCval < 27) {
return 1;
}
else if (ssDCval >= 27) {
return 4;
}
}
else if (tcor >= 102) {
return 8;
}
}
else if (TSCindex >= 2) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (Rs < 201) {
return 4;
}
else if (Rs >= 201) {
return 2;
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 7) {
if (MVDiff < 26) {
return 4;
}
else if (MVDiff >= 26) {
return 8;
}
}
else if (Rs >= 7) {
if (ssDCval < 31) {
if (negBalance < 58) {
return 1;
}
else if (negBalance >= 58) {
if (Rs < 93) {
return 8;
}
else if (Rs >= 93) {
return 1;
}
}
}
else if (ssDCval >= 31) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Rs < 443) {
return 1;
}
else if (Rs >= 443) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 27) {
if (refDCval < 30) {
if (Cs < 333) {
if (MVDiff < 34) {
if (posBalance < 69) {
if (tcor < 90) {
if (posBalance < 42) {
if (mu_mv_mag_sq < 41) {
if (Cs < 30) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
else if (Cs >= 30) {
if (diffRsCsdiff < -4) {
if (RsCsDiff < 5) {
return 4;
}
else if (RsCsDiff >= 5) {
return 1;
}
}
else if (diffRsCsdiff >= -4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (ssDCval < 30) {
return 1;
}
else if (ssDCval >= 30) {
if (Rs < 208) {
return 4;
}
else if (Rs >= 208) {
return 1;
}
}
}
}
else if (posBalance >= 42) {
if (negBalance < 20) {
if (AFD < 6) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (negBalance >= 20) {
return 1;
}
}
}
else if (tcor >= 90) {
return 1;
}
}
else if (posBalance >= 69) {
return 2;
}
}
else if (MVDiff >= 34) {
if (Cs < 245) {
if (negBalance < 27) {
if (CsDiff < 1) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
return 1;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (negBalance >= 27) {
if (negBalance < 30) {
if (Rs < 89) {
return 1;
}
else if (Rs >= 89) {
return 4;
}
}
else if (negBalance >= 30) {
return 1;
}
}
}
else if (Cs >= 245) {
if (Rs < 187) {
return 2;
}
else if (Rs >= 187) {
return 4;
}
}
}
}
else if (Cs >= 333) {
if (mu_mv_mag_sq < 61) {
if (mu_mv_mag_sq < 25) {
if (Rs < 381) {
return 2;
}
else if (Rs >= 381) {
return 8;
}
}
else if (mu_mv_mag_sq >= 25) {
return 4;
}
}
else if (mu_mv_mag_sq >= 61) {
return 1;
}
}
}
else if (refDCval >= 30) {
if (posBalance < 41) {
if (Cs < 257) {
if (mcTcor < 97) {
if (AFD < 4) {
if (ssDCval < 33) {
if (posBalance < 28) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 28) {
if (Rs < 122) {
return 8;
}
else if (Rs >= 122) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (MVDiff < 4) {
if (Rs < 494) {
return 8;
}
else if (Rs >= 494) {
return 1;
}
}
else if (MVDiff >= 4) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
if (Rs < 298) {
return 8;
}
else if (Rs >= 298) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 33) {
return 4;
}
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 42) {
if (tcor < 74) {
return 1;
}
else if (tcor >= 74) {
return 4;
}
}
else if (mu_mv_mag_sq >= 42) {
if (Rs < 214) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (Rs >= 214) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 342) {
if (mu_mv_mag_sq < 17) {
if (Rs < 103) {
if (Rs < 82) {
return 1;
}
else if (Rs >= 82) {
return 8;
}
}
else if (Rs >= 103) {
return 4;
}
}
else if (mu_mv_mag_sq >= 17) {
if (posBalance < 30) {
if (AFD < 2) {
if (Rs < 121) {
return 8;
}
else if (Rs >= 121) {
return 1;
}
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 30) {
return 1;
}
}
}
else if (Rs >= 342) {
return 4;
}
}
}
else if (Cs >= 257) {
if (negBalance < 15) {
if (Rs < 280) {
return 1;
}
else if (Rs >= 280) {
return 4;
}
}
else if (negBalance >= 15) {
return 1;
}
}
}
else if (posBalance >= 41) {
if (tcor < 62) {
if (mu_mv_mag_sq < 10) {
return 8;
}
else if (mu_mv_mag_sq >= 10) {
if (Rs < 164) {
if (Rs < 119) {
return 8;
}
else if (Rs >= 119) {
return 4;
}
}
else if (Rs >= 164) {
return 1;
}
}
}
else if (tcor >= 62) {
if (posBalance < 69) {
if (Rs < 293) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (mcTcor < 88) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
return 2;
}
}
else if (mcTcor >= 88) {
if (ssDCval < 34) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
if (Rs < 166) {
return 8;
}
else if (Rs >= 166) {
if (Rs < 204) {
return 2;
}
else if (Rs >= 204) {
return 1;
}
}
}
}
else if (ssDCval >= 34) {
return 4;
}
}
}
}
else if (Rs >= 293) {
return 1;
}
}
else if (posBalance >= 69) {
if (mu_mv_mag_sq < 14) {
return 2;
}
else if (mu_mv_mag_sq >= 14) {
if (Rs < 51) {
return 4;
}
else if (Rs >= 51) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 300) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 52) {
if (mcTcor < 88) {
if (MVDiff < 40) {
if (tcor < 67) {
if (ssDCval < 42) {
if (TSC < 4) {
if (Cs < 191) {
return 1;
}
else if (Cs >= 191) {
return 4;
}
}
else if (TSC >= 4) {
if (gchDC < 1) {
if (Rs < 200) {
return 8;
}
else if (Rs >= 200) {
return 2;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (ssDCval >= 42) {
return 4;
}
}
else if (tcor >= 67) {
if (posBalance < 55) {
if (TSC < 4) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Cs < 202) {
return 8;
}
else if (Cs >= 202) {
return 2;
}
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (MVDiff >= 40) {
if (posBalance < 47) {
if (tcor < 67) {
if (Cs < 293) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
else if (Cs >= 293) {
return 2;
}
}
else if (tcor >= 67) {
if (Cs < 52) {
if (ssDCval < 36) {
return 1;
}
else if (ssDCval >= 36) {
return 8;
}
}
else if (Cs >= 52) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
return 8;
}
}
}
}
else if (posBalance >= 47) {
if (posBalance < 75) {
if (mu_mv_mag_sq < 4) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 4) {
if (MVDiff < 227) {
if (mcTcor < 81) {
if (CsDiff < 2) {
if (Rs < 101) {
return 8;
}
else if (Rs >= 101) {
if (MVDiff < 69) {
return 8;
}
else if (MVDiff >= 69) {
return 4;
}
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (mcTcor >= 81) {
if (MVDiff < 48) {
return 1;
}
else if (MVDiff >= 48) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
else if (MVDiff >= 227) {
return 1;
}
}
}
else if (posBalance >= 75) {
return 2;
}
}
}
}
else if (mcTcor >= 88) {
if (posBalance < 8) {
if (posBalance < 1) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
else if (posBalance >= 1) {
if (Rs < 231) {
return 8;
}
else if (Rs >= 231) {
return 2;
}
}
}
else if (posBalance >= 8) {
if (mcTcor < 98) {
if (ssDCval < 43) {
if (TSC < 4) {
if (AFD < 9) {
if (AFD < 4) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (Rs < 190) {
return 8;
}
else if (Rs >= 190) {
return 2;
}
}
}
else if (AFD >= 4) {
if (refDCval < 38) {
if (tcor < 82) {
if (Cs < 215) {
if (Rs < 74) {
if (Cs < 34) {
return 2;
}
else if (Cs >= 34) {
return 1;
}
}
else if (Rs >= 74) {
return 2;
}
}
else if (Cs >= 215) {
return 8;
}
}
else if (tcor >= 82) {
if (Rs < 53) {
return 2;
}
else if (Rs >= 53) {
if (posBalance < 44) {
if (Rs < 163) {
return 1;
}
else if (Rs >= 163) {
return 4;
}
}
else if (posBalance >= 44) {
return 1;
}
}
}
}
else if (refDCval >= 38) {
if (Rs < 101) {
if (Cs < 50) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
if (Rs < 25) {
return 8;
}
else if (Rs >= 25) {
return 4;
}
}
}
else if (Cs >= 50) {
if (Rs < 58) {
return 1;
}
else if (Rs >= 58) {
if (negBalance < 47) {
return 4;
}
else if (negBalance >= 47) {
return 1;
}
}
}
}
else if (Rs >= 101) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
else if (AFD >= 9) {
if (posBalance < 62) {
if (AFD < 10) {
if (Rs < 128) {
return 4;
}
else if (Rs >= 128) {
return 1;
}
}
else if (AFD >= 10) {
return 1;
}
}
else if (posBalance >= 62) {
return 4;
}
}
}
else if (TSC >= 4) {
if (diffMVdiffVal < 9) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (RsDiff < 2) {
if (negBalance < 41) {
if (Cs < 265) {
return 8;
}
else if (Cs >= 265) {
return 1;
}
}
else if (negBalance >= 41) {
return 1;
}
}
else if (RsDiff >= 2) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 9) {
if (Cs < 80) {
return 1;
}
else if (Cs >= 80) {
if (Cs < 223) {
return 2;
}
else if (Cs >= 223) {
return 4;
}
}
}
}
}
else if (ssDCval >= 43) {
if (posBalance < 67) {
if (SC < 488) {
if (Rs < 35) {
if (Rs < 22) {
return 4;
}
else if (Rs >= 22) {
return 2;
}
}
else if (Rs >= 35) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (diffAFD < -1) {
if (mcTcor < 90) {
return 2;
}
else if (mcTcor >= 90) {
return 1;
}
}
else if (diffAFD >= -1) {
if (Rs < 192) {
return 2;
}
else if (Rs >= 192) {
return 4;
}
}
}
}
}
else if (SC >= 488) {
if (Rs < 253) {
if (Cs < 488) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
return 1;
}
}
else if (Cs >= 488) {
return 4;
}
}
else if (Rs >= 253) {
return 4;
}
}
}
else if (posBalance >= 67) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (MVDiff < 34) {
return 1;
}
else if (MVDiff >= 34) {
if (SC < 233) {
return 2;
}
else if (SC >= 233) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (Cs < 303) {
if (negBalance < 29) {
if (MVDiff < 2) {
if (Rs < 230) {
if (refDCval < 38) {
if (Cs < 226) {
if (Rs < 150) {
return 4;
}
else if (Rs >= 150) {
if (Rs < 164) {
return 1;
}
else if (Rs >= 164) {
return 4;
}
}
}
else if (Cs >= 226) {
if (Rs < 110) {
return 1;
}
else if (Rs >= 110) {
return 8;
}
}
}
else if (refDCval >= 38) {
if (Cs < 144) {
return 8;
}
else if (Cs >= 144) {
if (Rs < 185) {
return 1;
}
else if (Rs >= 185) {
return 8;
}
}
}
}
else if (Rs >= 230) {
if (Rs < 292) {
return 4;
}
else if (Rs >= 292) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (refDCval < 47) {
if (posBalance < 29) {
if (refDCval < 39) {
if (mu_mv_mag_sq < 89) {
if (Rs < 194) {
return 1;
}
else if (Rs >= 194) {
return 2;
}
}
else if (mu_mv_mag_sq >= 89) {
return 4;
}
}
else if (refDCval >= 39) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
if (Rs < 148) {
return 8;
}
else if (Rs >= 148) {
return 4;
}
}
}
}
else if (posBalance >= 29) {
if (mu_mv_mag_sq < 112) {
if (mcTcor < 97) {
if (tcor < 91) {
if (mu_mv_mag_sq < 65) {
if (Cs < 162) {
return 1;
}
else if (Cs >= 162) {
return 8;
}
}
else if (mu_mv_mag_sq >= 65) {
if (tcor < 87) {
return 1;
}
else if (tcor >= 87) {
if (Rs < 203) {
return 4;
}
else if (Rs >= 203) {
return 1;
}
}
}
}
else if (tcor >= 91) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (SC < 476) {
return 1;
}
else if (SC >= 476) {
if (Rs < 246) {
return 8;
}
else if (Rs >= 246) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (Rs < 46) {
return 8;
}
else if (Rs >= 46) {
return 4;
}
}
}
}
}
else if (refDCval >= 47) {
if (Rs < 176) {
return 8;
}
else if (Rs >= 176) {
return 4;
}
}
}
}
else if (negBalance >= 29) {
if (negBalance < 42) {
if (posBalance < 46) {
if (mu_mv_mag_sq < 78) {
if (Rs < 264) {
return 1;
}
else if (Rs >= 264) {
return 4;
}
}
else if (mu_mv_mag_sq >= 78) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (Rs < 102) {
return 8;
}
else if (Rs >= 102) {
return 1;
}
}
}
}
else if (posBalance >= 46) {
if (AFD < 8) {
if (Rs < 79) {
if (Rs < 70) {
return 8;
}
else if (Rs >= 70) {
return 1;
}
}
else if (Rs >= 79) {
return 8;
}
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (negBalance >= 42) {
if (mcTcor < 96) {
if (Rs < 152) {
if (refDCval < 43) {
if (AFD < 6) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 2;
}
}
else if (AFD >= 6) {
if (ssDCval < 43) {
if (posBalance < 53) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 37) {
return 8;
}
else if (Rs >= 37) {
return 4;
}
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (ssDCval >= 43) {
return 2;
}
}
}
else if (refDCval >= 43) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (ssDCval < 42) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
return 8;
}
}
else if (ssDCval >= 42) {
return 2;
}
}
}
}
else if (Rs >= 152) {
if (diffMVdiffVal < -8) {
return 1;
}
else if (diffMVdiffVal >= -8) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 207) {
return 1;
}
else if (Rs >= 207) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (SC < 72) {
return 4;
}
else if (SC >= 72) {
if (MVDiff < 10) {
if (negBalance < 68) {
return 8;
}
else if (negBalance >= 68) {
return 4;
}
}
else if (MVDiff >= 10) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 303) {
if (refDCval < 50) {
if (negBalance < 34) {
if (Cs < 312) {
if (Cs < 307) {
return 1;
}
else if (Cs >= 307) {
return 8;
}
}
else if (Cs >= 312) {
return 1;
}
}
else if (negBalance >= 34) {
if (Rs < 253) {
return 8;
}
else if (Rs >= 253) {
if (Rs < 282) {
return 2;
}
else if (Rs >= 282) {
return 1;
}
}
}
}
else if (refDCval >= 50) {
if (Rs < 239) {
return 4;
}
else if (Rs >= 239) {
return 2;
}
}
}
}
}
else if (ssDCval >= 50) {
if (refDCval < 70) {
if (TSCindex < 2) {
if (refDCval < 62) {
if (posBalance < 30) {
if (Rs < 28) {
return 2;
}
else if (Rs >= 28) {
if (Rs < 205) {
return 8;
}
else if (Rs >= 205) {
return 1;
}
}
}
else if (posBalance >= 30) {
if (TSC < 1) {
if (diffMVdiffVal < -26) {
return 8;
}
else if (diffMVdiffVal >= -26) {
return 4;
}
}
else if (TSC >= 1) {
if (tcor < 99) {
if (Rs < 266) {
if (negBalance < 21) {
if (mcTcor < 99) {
if (negBalance < 16) {
return 1;
}
else if (negBalance >= 16) {
return 4;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (negBalance >= 21) {
if (tcor < 96) {
if (Rs < 55) {
return 1;
}
else if (Rs >= 55) {
return 4;
}
}
else if (tcor >= 96) {
return 1;
}
}
}
else if (Rs >= 266) {
if (Rs < 291) {
return 1;
}
else if (Rs >= 291) {
return 2;
}
}
}
else if (tcor >= 99) {
if (Rs < 246) {
if (Cs < 190) {
if (refDCval < 57) {
if (refDCval < 54) {
return 1;
}
else if (refDCval >= 54) {
return 4;
}
}
else if (refDCval >= 57) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 4;
}
}
}
else if (Cs >= 190) {
return 4;
}
}
else if (Rs >= 246) {
return 1;
}
}
}
}
}
else if (refDCval >= 62) {
if (SC < 381) {
if (MVDiff < 5) {
if (mcTcor < 97) {
return 1;
}
else if (mcTcor >= 97) {
return 2;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (SC >= 381) {
if (mu_mv_mag_sq < 104) {
if (Rs < 182) {
return 2;
}
else if (Rs >= 182) {
return 4;
}
}
else if (mu_mv_mag_sq >= 104) {
if (Cs < 378) {
return 4;
}
else if (Cs >= 378) {
if (Rs < 205) {
return 1;
}
else if (Rs >= 205) {
return 4;
}
}
}
}
}
}
else if (TSCindex >= 2) {
if (RsCsDiff < 1) {
if (SC < 342) {
if (tcor < 100) {
if (MVDiff < 2) {
if (Cs < 126) {
if (ssDCval < 58) {
return 2;
}
else if (ssDCval >= 58) {
if (Rs < 123) {
return 1;
}
else if (Rs >= 123) {
return 8;
}
}
}
else if (Cs >= 126) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
if (posBalance < 27) {
return 1;
}
else if (posBalance >= 27) {
return 2;
}
}
}
}
else if (MVDiff >= 2) {
if (mcTcor < 95) {
if (negBalance < 50) {
if (refDCval < 55) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
else if (refDCval >= 55) {
if (Cs < 100) {
return 2;
}
else if (Cs >= 100) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (AFD < 6) {
if (Rs < 9) {
return 4;
}
else if (Rs >= 9) {
return 1;
}
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (Cs < 189) {
if (negBalance < 46) {
if (mu_mv_mag_sq < 87) {
if (diffRsCsdiff < -1) {
if (refDCval < 63) {
return 4;
}
else if (refDCval >= 63) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (Cs < 150) {
if (mcTcor < 99) {
if (Rs < 121) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Rs < 85) {
return 2;
}
else if (Rs >= 85) {
return 8;
}
}
}
else if (Rs >= 121) {
if (Rs < 160) {
if (AFD < 2) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (AFD >= 2) {
return 1;
}
}
else if (Rs >= 160) {
return 4;
}
}
}
else if (mcTcor >= 99) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (Cs >= 150) {
return 1;
}
}
}
else if (negBalance >= 46) {
if (Rs < 79) {
if (Rs < 28) {
return 4;
}
else if (Rs >= 28) {
return 1;
}
}
else if (Rs >= 79) {
return 8;
}
}
}
else if (Cs >= 189) {
if (ssDCval < 62) {
if (Rs < 143) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
else if (Rs >= 143) {
return 1;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
}
}
}
else if (tcor >= 100) {
if (Rs < 40) {
return 2;
}
else if (Rs >= 40) {
return 8;
}
}
}
else if (SC >= 342) {
if (negBalance < 30) {
if (ssDCval < 60) {
if (MVDiff < 1) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (posBalance < 46) {
if (Rs < 254) {
return 4;
}
else if (Rs >= 254) {
return 1;
}
}
else if (posBalance >= 46) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 191) {
return 1;
}
else if (Cs >= 191) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 4;
}
}
}
}
else if (ssDCval >= 60) {
if (mu_mv_mag_sq < 80) {
if (Cs < 295) {
return 4;
}
else if (Cs >= 295) {
if (Rs < 178) {
return 2;
}
else if (Rs >= 178) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (mu_mv_mag_sq < 124) {
if (SC < 363) {
if (posBalance < 44) {
if (Rs < 189) {
return 2;
}
else if (Rs >= 189) {
return 4;
}
}
else if (posBalance >= 44) {
return 1;
}
}
else if (SC >= 363) {
if (TSC < 2) {
if (Rs < 136) {
if (Rs < 128) {
return 2;
}
else if (Rs >= 128) {
return 1;
}
}
else if (Rs >= 136) {
if (MVDiff < 5) {
if (Cs < 377) {
return 4;
}
else if (Cs >= 377) {
if (Cs < 394) {
return 1;
}
else if (Cs >= 394) {
return 4;
}
}
}
else if (MVDiff >= 5) {
return 1;
}
}
}
else if (TSC >= 2) {
if (SC < 515) {
if (posBalance < 48) {
if (Rs < 128) {
return 1;
}
else if (Rs >= 128) {
if (Cs < 336) {
return 4;
}
else if (Cs >= 336) {
return 1;
}
}
}
else if (posBalance >= 48) {
if (Rs < 233) {
if (Rs < 206) {
if (mu_mv_mag_sq < 100) {
return 1;
}
else if (mu_mv_mag_sq >= 100) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (Rs < 181) {
if (Rs < 150) {
return 1;
}
else if (Rs >= 150) {
return 4;
}
}
else if (Rs >= 181) {
return 1;
}
}
}
}
else if (Rs >= 206) {
return 4;
}
}
else if (Rs >= 233) {
return 1;
}
}
}
else if (SC >= 515) {
if (ssDCval < 69) {
if (Rs < 240) {
if (Rs < 180) {
if (Rs < 174) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
return 1;
}
}
else if (Rs >= 174) {
return 1;
}
}
else if (Rs >= 180) {
return 4;
}
}
else if (Rs >= 240) {
if (Rs < 280) {
return 1;
}
else if (Rs >= 280) {
return 2;
}
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 161) {
return 4;
}
else if (Rs >= 161) {
if (posBalance < 46) {
if (refDCval < 69) {
if (Rs < 237) {
return 4;
}
else if (Rs >= 237) {
return 1;
}
}
else if (refDCval >= 69) {
if (Cs < 325) {
return 4;
}
else if (Cs >= 325) {
return 1;
}
}
}
else if (posBalance >= 46) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 30) {
if (tcor < 97) {
if (ssDCval < 59) {
if (Cs < 200) {
return 8;
}
else if (Cs >= 200) {
return 4;
}
}
else if (ssDCval >= 59) {
if (mu_mv_mag_sq < 90) {
if (AFD < 4) {
if (SC < 389) {
if (Rs < 210) {
return 1;
}
else if (Rs >= 210) {
return 4;
}
}
else if (SC >= 389) {
return 4;
}
}
else if (AFD >= 4) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
if (Cs < 182) {
return 4;
}
else if (Cs >= 182) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (Cs < 320) {
return 1;
}
else if (Cs >= 320) {
return 4;
}
}
}
}
else if (tcor >= 97) {
if (Cs < 333) {
if (posBalance < 46) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (posBalance < 44) {
if (refDCval < 61) {
return 1;
}
else if (refDCval >= 61) {
return 4;
}
}
else if (posBalance >= 44) {
return 1;
}
}
}
else if (posBalance >= 46) {
if (ssDCval < 64) {
if (mu_mv_mag_sq < 102) {
if (SC < 505) {
if (SC < 466) {
return 1;
}
else if (SC >= 466) {
return 4;
}
}
else if (SC >= 505) {
return 1;
}
}
else if (mu_mv_mag_sq >= 102) {
return 1;
}
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (Cs >= 333) {
if (Rs < 163) {
return 1;
}
else if (Rs >= 163) {
if (ssDCval < 58) {
if (Cs < 410) {
return 4;
}
else if (Cs >= 410) {
return 1;
}
}
else if (ssDCval >= 58) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (TSC < 4) {
if (mcTcor < 91) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
if (Cs < 68) {
return 2;
}
else if (Cs >= 68) {
return 8;
}
}
}
}
else if (mcTcor >= 91) {
if (Rs < 35) {
return 1;
}
else if (Rs >= 35) {
if (MVDiff < 2) {
if (Cs < 404) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 117) {
if (tcor < 95) {
if (Cs < 391) {
return 1;
}
else if (Cs >= 391) {
return 4;
}
}
else if (tcor >= 95) {
return 4;
}
}
else if (mu_mv_mag_sq >= 117) {
return 1;
}
}
else if (ssDCval >= 69) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 165) {
return 1;
}
else if (Rs >= 165) {
return 4;
}
}
}
}
}
else if (Cs >= 404) {
return 1;
}
}
else if (MVDiff >= 2) {
if (RsDiff < 1) {
if (MVDiff < 46) {
if (negBalance < 51) {
if (SC < 577) {
if (diffRsCsdiff < 1) {
if (refDCval < 52) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
else if (refDCval >= 52) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
if (Rs < 157) {
return 1;
}
else if (Rs >= 157) {
return 4;
}
}
}
}
}
else if (diffRsCsdiff >= 1) {
if (MVDiff < 19) {
if (Rs < 93) {
return 1;
}
else if (Rs >= 93) {
if (Rs < 226) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
return 1;
}
}
}
else if (Rs >= 226) {
return 1;
}
}
}
else if (MVDiff >= 19) {
return 2;
}
}
}
else if (SC >= 577) {
return 4;
}
}
else if (negBalance >= 51) {
if (Cs < 198) {
return 4;
}
else if (Cs >= 198) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (MVDiff >= 46) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (RsDiff >= 1) {
if (refDCval < 61) {
if (RsDiff < 2) {
if (TSC < 3) {
if (AFD < 5) {
if (Cs < 112) {
return 8;
}
else if (Cs >= 112) {
return 1;
}
}
else if (AFD >= 5) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
else if (TSC >= 3) {
if (CsDiff < 2) {
if (posBalance < 64) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (Rs < 208) {
return 4;
}
else if (Rs >= 208) {
return 8;
}
}
}
else if (posBalance >= 64) {
return 2;
}
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (RsDiff >= 2) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (refDCval >= 61) {
if (negBalance < 38) {
if (posBalance < 52) {
if (tcor < 96) {
if (Cs < 129) {
return 8;
}
else if (Cs >= 129) {
return 1;
}
}
else if (tcor >= 96) {
return 4;
}
}
else if (posBalance >= 52) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
if (Rs < 201) {
return 8;
}
else if (Rs >= 201) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 38) {
if (gchDC < 2) {
if (SC < 759) {
if (Rs < 291) {
return 4;
}
else if (Rs >= 291) {
if (Rs < 296) {
return 1;
}
else if (Rs >= 296) {
return 4;
}
}
}
else if (SC >= 759) {
return 1;
}
}
else if (gchDC >= 2) {
if (AFD < 9) {
if (Cs < 272) {
return 4;
}
else if (Cs >= 272) {
return 1;
}
}
else if (AFD >= 9) {
return 1;
}
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 40) {
if (mcTcor < 77) {
if (MVDiff < 120) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (MVDiff >= 120) {
if (negBalance < 53) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
return 1;
}
}
else if (negBalance >= 53) {
if (Rs < 99) {
return 2;
}
else if (Rs >= 99) {
return 4;
}
}
}
}
else if (mcTcor >= 77) {
if (MVDiff < 61) {
if (RsCsDiff < 17) {
if (negBalance < 41) {
if (CsDiff < 2) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
if (Rs < 33) {
return 1;
}
else if (Rs >= 33) {
return 2;
}
}
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (negBalance >= 41) {
if (Rs < 259) {
if (Cs < 471) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
if (AFD < 9) {
if (Rs < 134) {
return 1;
}
else if (Rs >= 134) {
return 2;
}
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (Cs >= 471) {
return 2;
}
}
else if (Rs >= 259) {
if (tcor < 75) {
if (Cs < 412) {
return 4;
}
else if (Cs >= 412) {
return 1;
}
}
else if (tcor >= 75) {
if (ssDCval < 61) {
if (Rs < 298) {
return 1;
}
else if (Rs >= 298) {
return 2;
}
}
else if (ssDCval >= 61) {
if (Rs < 264) {
return 4;
}
else if (Rs >= 264) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 17) {
if (Rs < 158) {
return 4;
}
else if (Rs >= 158) {
if (MVDiff < 44) {
return 2;
}
else if (MVDiff >= 44) {
return 8;
}
}
}
}
else if (MVDiff >= 61) {
if (AFD < 13) {
if (posBalance < 75) {
if (Cs < 44) {
return 1;
}
else if (Cs >= 44) {
if (posBalance < 70) {
return 2;
}
else if (posBalance >= 70) {
if (Rs < 171) {
return 1;
}
else if (Rs >= 171) {
return 2;
}
}
}
}
else if (posBalance >= 75) {
if (Rs < 77) {
return 2;
}
else if (Rs >= 77) {
return 1;
}
}
}
else if (AFD >= 13) {
if (negBalance < 54) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
if (ssDCval < 67) {
if (Rs < 267) {
return 8;
}
else if (Rs >= 267) {
return 2;
}
}
else if (ssDCval >= 67) {
if (Rs < 111) {
return 8;
}
else if (Rs >= 111) {
return 1;
}
}
}
}
else if (negBalance >= 54) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (AFD < 19) {
if (negBalance < 82) {
if (AFD < 8) {
if (MVDiff < 7) {
if (Rs < 257) {
return 4;
}
else if (Rs >= 257) {
return 1;
}
}
else if (MVDiff >= 7) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
return 8;
}
}
}
}
}
else if (AFD >= 8) {
if (mu_mv_mag_sq < 103) {
return 4;
}
else if (mu_mv_mag_sq >= 103) {
return 1;
}
}
}
else if (negBalance >= 82) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (AFD >= 19) {
if (mcTcor < 75) {
if (Rs < 203) {
return 1;
}
else if (Rs >= 203) {
return 2;
}
}
else if (mcTcor >= 75) {
if (AFD < 28) {
if (SC < 136) {
return 8;
}
else if (SC >= 136) {
return 1;
}
}
else if (AFD >= 28) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 70) {
if (TSCindex < 2) {
if (ssDCval < 93) {
if (Rs < 102) {
if (refDCval < 73) {
return 1;
}
else if (refDCval >= 73) {
return 4;
}
}
else if (Rs >= 102) {
if (MVDiff < 6) {
if (Cs < 270) {
if (Rs < 286) {
return 2;
}
else if (Rs >= 286) {
if (Cs < 205) {
return 1;
}
else if (Cs >= 205) {
return 2;
}
}
}
else if (Cs >= 270) {
if (Rs < 221) {
return 1;
}
else if (Rs >= 221) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (SC < 235) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Rs < 126) {
return 1;
}
else if (Rs >= 126) {
return 2;
}
}
}
else if (SC >= 235) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
if (Rs < 148) {
return 4;
}
else if (Rs >= 148) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 93) {
if (Cs < 115) {
if (ssDCval < 118) {
return 1;
}
else if (ssDCval >= 118) {
return 8;
}
}
else if (Cs >= 115) {
if (MVDiff < 29) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (MVDiff >= 29) {
if (Cs < 209) {
return 1;
}
else if (Cs >= 209) {
return 2;
}
}
}
}
}
else if (TSCindex >= 2) {
if (mcTcor < 97) {
if (Rs < 118) {
if (mu_mv_mag_sq < 66) {
if (Cs < 7) {
if (mcTcor < 68) {
if (ssDCval < 144) {
return 4;
}
else if (ssDCval >= 144) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (mcTcor >= 68) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
if (MVDiff < 42) {
if (Rs < 15) {
return 2;
}
else if (Rs >= 15) {
return 1;
}
}
else if (MVDiff >= 42) {
return 8;
}
}
}
}
}
else if (Cs >= 7) {
if (ssDCval < 108) {
if (ssDCval < 98) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (mu_mv_mag_sq < 27) {
if (diffMVdiffVal < 14) {
if (SC < 109) {
return 1;
}
else if (SC >= 109) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
return 1;
}
}
}
else if (diffMVdiffVal >= 14) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (SC < 203) {
if (diffMVdiffVal < 4) {
if (MVDiff < 35) {
if (Cs < 25) {
return 1;
}
else if (Cs >= 25) {
return 8;
}
}
else if (MVDiff >= 35) {
return 2;
}
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
else if (SC >= 203) {
return 1;
}
}
}
}
else if (ssDCval >= 98) {
if (MVDiff < 108) {
return 1;
}
else if (MVDiff >= 108) {
return 4;
}
}
}
else if (ssDCval >= 108) {
if (mcTcor < 78) {
return 1;
}
else if (mcTcor >= 78) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (negBalance < 60) {
if (Cs < 138) {
if (TSC < 23) {
if (Rs < 105) {
return 4;
}
else if (Rs >= 105) {
if (Cs < 47) {
return 2;
}
else if (Cs >= 47) {
return 4;
}
}
}
else if (TSC >= 23) {
return 8;
}
}
else if (Cs >= 138) {
if (Rs < 96) {
return 8;
}
else if (Rs >= 96) {
return 1;
}
}
}
else if (negBalance >= 60) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
else if (Rs >= 118) {
if (RsCsDiff < 3) {
if (mcTcor < 95) {
if (ssDCval < 86) {
if (mu_mv_mag_sq < 21) {
return 1;
}
else if (mu_mv_mag_sq >= 21) {
if (Rs < 268) {
return 8;
}
else if (Rs >= 268) {
return 1;
}
}
}
else if (ssDCval >= 86) {
if (MVDiff < 24) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
if (Cs < 81) {
if (Rs < 134) {
return 4;
}
else if (Rs >= 134) {
return 8;
}
}
else if (Cs >= 81) {
if (Cs < 365) {
if (tcor < 67) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (Rs < 175) {
if (Rs < 131) {
return 2;
}
else if (Rs >= 131) {
return 8;
}
}
else if (Rs >= 175) {
if (Rs < 204) {
return 2;
}
else if (Rs >= 204) {
return 8;
}
}
}
}
}
else if (tcor >= 67) {
if (Cs < 268) {
return 2;
}
else if (Cs >= 268) {
if (RsDiff < 1) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (Rs < 230) {
return 2;
}
else if (Rs >= 230) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (Rs < 240) {
return 2;
}
else if (Rs >= 240) {
if (Rs < 254) {
return 8;
}
else if (Rs >= 254) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 365) {
return 8;
}
}
}
}
else if (MVDiff >= 24) {
if (TSC < 5) {
if (Rs < 140) {
if (diffAFD < 1) {
if (Cs < 83) {
return 8;
}
else if (Cs >= 83) {
return 2;
}
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (Rs >= 140) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
if (Rs < 264) {
return 1;
}
else if (Rs >= 264) {
return 2;
}
}
}
}
else if (TSC >= 5) {
if (gchDC < 2) {
if (Cs < 26) {
return 1;
}
else if (Cs >= 26) {
return 4;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
}
else if (mcTcor >= 95) {
if (Cs < 64) {
if (refDCval < 90) {
return 1;
}
else if (refDCval >= 90) {
return 8;
}
}
else if (Cs >= 64) {
if (Rs < 137) {
if (ssDCval < 79) {
if (Cs < 159) {
return 8;
}
else if (Cs >= 159) {
return 4;
}
}
else if (ssDCval >= 79) {
return 1;
}
}
else if (Rs >= 137) {
if (diffTSC < 1) {
if (RsCsDiff < 2) {
if (posBalance < 46) {
if (Rs < 233) {
return 8;
}
else if (Rs >= 233) {
return 1;
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (RsCsDiff >= 2) {
if (mu_mv_mag_sq < 6) {
return 2;
}
else if (mu_mv_mag_sq >= 6) {
if (gchDC < 1) {
if (Cs < 276) {
return 1;
}
else if (Cs >= 276) {
if (Cs < 418) {
return 8;
}
else if (Cs >= 418) {
return 2;
}
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (diffTSC >= 1) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 3) {
if (mcTcor < 69) {
if (tcor < 17) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (tcor >= 17) {
return 4;
}
}
else if (mcTcor >= 69) {
if (SC < 609) {
if (tcor < 58) {
if (AFD < 26) {
if (SC < 314) {
if (AFD < 18) {
if (mcTcor < 81) {
return 2;
}
else if (mcTcor >= 81) {
if (negBalance < 53) {
if (Rs < 215) {
return 2;
}
else if (Rs >= 215) {
return 8;
}
}
else if (negBalance >= 53) {
return 8;
}
}
}
else if (AFD >= 18) {
if (Rs < 186) {
return 2;
}
else if (Rs >= 186) {
return 1;
}
}
}
else if (SC >= 314) {
return 2;
}
}
else if (AFD >= 26) {
if (RsCsDiff < 23) {
if (Rs < 164) {
if (Rs < 153) {
return 4;
}
else if (Rs >= 153) {
return 1;
}
}
else if (Rs >= 164) {
return 8;
}
}
else if (RsCsDiff >= 23) {
return 1;
}
}
}
else if (tcor >= 58) {
if (gchDC < 2) {
if (diffAFD < -1) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
if (Rs < 229) {
if (Rs < 201) {
return 4;
}
else if (Rs >= 201) {
return 1;
}
}
else if (Rs >= 229) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (MVDiff < 53) {
return 1;
}
else if (MVDiff >= 53) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (negBalance < 47) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
return 1;
}
}
else if (negBalance >= 47) {
if (mu_mv_mag_sq < 9) {
if (Cs < 198) {
return 8;
}
else if (Cs >= 198) {
return 2;
}
}
else if (mu_mv_mag_sq >= 9) {
if (Cs < 233) {
if (Rs < 195) {
return 2;
}
else if (Rs >= 195) {
if (Cs < 155) {
return 1;
}
else if (Cs >= 155) {
return 2;
}
}
}
else if (Cs >= 233) {
if (Cs < 347) {
return 1;
}
else if (Cs >= 347) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 609) {
if (TSC < 8) {
if (ssDCval < 110) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (Rs < 233) {
return 4;
}
else if (Rs >= 233) {
return 1;
}
}
}
else if (ssDCval >= 110) {
if (Cs < 484) {
return 2;
}
else if (Cs >= 484) {
return 8;
}
}
}
else if (TSC >= 8) {
if (Cs < 528) {
return 2;
}
else if (Cs >= 528) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 232) {
if (Rs < 90) {
if (ssDCval < 97) {
if (Cs < 126) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (MVDiff < 71) {
if (diffMVdiffVal < -10) {
if (Rs < 72) {
return 1;
}
else if (Rs >= 72) {
return 4;
}
}
else if (diffMVdiffVal >= -10) {
return 1;
}
}
else if (MVDiff >= 71) {
if (Cs < 87) {
return 4;
}
else if (Cs >= 87) {
return 8;
}
}
}
}
else if (Cs >= 126) {
if (MVDiff < 35) {
if (tcor < 98) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 4;
}
}
else if (tcor >= 98) {
if (Rs < 73) {
return 1;
}
else if (Rs >= 73) {
return 2;
}
}
}
else if (MVDiff >= 35) {
return 2;
}
}
}
else if (ssDCval >= 97) {
if (AFD < 3) {
if (Rs < 75) {
return 4;
}
else if (Rs >= 75) {
if (MVDiff < 27) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 2;
}
}
else if (MVDiff >= 27) {
return 1;
}
}
}
else if (AFD >= 3) {
if (posBalance < 78) {
if (negBalance < 40) {
if (posBalance < 45) {
if (Rs < 52) {
return 4;
}
else if (Rs >= 52) {
return 1;
}
}
else if (posBalance >= 45) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (negBalance >= 40) {
if (Rs < 1) {
return 1;
}
else if (Rs >= 1) {
if (ssDCval < 112) {
if (AFD < 6) {
if (Rs < 63) {
return 2;
}
else if (Rs >= 63) {
return 1;
}
}
else if (AFD >= 6) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
return 1;
}
}
}
else if (ssDCval >= 112) {
return 2;
}
}
}
}
else if (posBalance >= 78) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
}
else if (Rs >= 90) {
if (Cs < 78) {
if (refDCval < 98) {
if (Cs < 42) {
if (Cs < 11) {
return 2;
}
else if (Cs >= 11) {
return 1;
}
}
else if (Cs >= 42) {
if (posBalance < 47) {
if (Rs < 132) {
return 1;
}
else if (Rs >= 132) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
return 1;
}
}
}
}
else if (refDCval >= 98) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
return 1;
}
}
}
}
else if (Cs >= 78) {
if (MVDiff < 2) {
if (Cs < 210) {
if (Cs < 125) {
if (Cs < 87) {
return 1;
}
else if (Cs >= 87) {
if (Rs < 139) {
if (Cs < 121) {
return 8;
}
else if (Cs >= 121) {
return 2;
}
}
else if (Rs >= 139) {
return 8;
}
}
}
else if (Cs >= 125) {
if (mu_mv_mag_sq < 96) {
if (diffMVdiffVal < -1) {
if (diffRsCsdiff < -1) {
if (Rs < 245) {
return 4;
}
else if (Rs >= 245) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
if (RsDiff < 1) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 116) {
return 2;
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
else if (mcTcor >= 98) {
return 2;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
else if (Cs >= 210) {
if (ssDCval < 74) {
if (refDCval < 72) {
return 2;
}
else if (refDCval >= 72) {
if (Rs < 177) {
return 2;
}
else if (Rs >= 177) {
return 1;
}
}
}
else if (ssDCval >= 74) {
return 2;
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 106) {
if (AFD < 2) {
if (ssDCval < 75) {
if (Rs < 194) {
return 2;
}
else if (Rs >= 194) {
return 1;
}
}
else if (ssDCval >= 75) {
if (Cs < 211) {
if (mu_mv_mag_sq < 105) {
if (Cs < 99) {
return 2;
}
else if (Cs >= 99) {
if (posBalance < 40) {
if (Rs < 166) {
return 1;
}
else if (Rs >= 166) {
return 2;
}
}
else if (posBalance >= 40) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 199) {
return 2;
}
else if (Rs >= 199) {
return 8;
}
}
}
else if (Cs >= 211) {
return 2;
}
}
}
else if (AFD >= 2) {
if (negBalance < 46) {
if (refDCval < 91) {
if (ssDCval < 73) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
if (MVDiff < 3) {
if (Rs < 203) {
return 2;
}
else if (Rs >= 203) {
return 1;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (ssDCval >= 73) {
if (mu_mv_mag_sq < 91) {
if (posBalance < 56) {
if (negBalance < 31) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 129) {
if (Rs < 113) {
return 1;
}
else if (Rs >= 113) {
return 2;
}
}
else if (Rs >= 129) {
return 1;
}
}
}
else if (negBalance >= 31) {
return 1;
}
}
else if (posBalance >= 56) {
if (TSC < 4) {
if (ssDCval < 76) {
if (Rs < 200) {
return 8;
}
else if (Rs >= 200) {
return 1;
}
}
else if (ssDCval >= 76) {
if (negBalance < 32) {
return 1;
}
else if (negBalance >= 32) {
return 2;
}
}
}
else if (TSC >= 4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (Rs < 189) {
return 2;
}
else if (Rs >= 189) {
return 8;
}
}
}
}
else if (refDCval >= 91) {
if (Rs < 137) {
if (Cs < 210) {
return 1;
}
else if (Cs >= 210) {
return 4;
}
}
else if (Rs >= 137) {
if (Rs < 203) {
if (Cs < 183) {
return 2;
}
else if (Cs >= 183) {
return 8;
}
}
else if (Rs >= 203) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (Rs < 238) {
return 2;
}
else if (Rs >= 238) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (ssDCval < 86) {
if (Cs < 175) {
if (diffMVdiffVal < 3) {
if (Rs < 106) {
return 4;
}
else if (Rs >= 106) {
return 1;
}
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
else if (Cs >= 175) {
if (Rs < 233) {
return 2;
}
else if (Rs >= 233) {
return 1;
}
}
}
else if (ssDCval >= 86) {
if (mu_mv_mag_sq < 58) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 2;
}
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 115) {
return 1;
}
else if (Rs >= 115) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 106) {
if (posBalance < 19) {
return 8;
}
else if (posBalance >= 19) {
if (mu_mv_mag_sq < 90) {
if (RsCsDiff < 7) {
return 2;
}
else if (RsCsDiff >= 7) {
if (Rs < 241) {
return 2;
}
else if (Rs >= 241) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 232) {
if (refDCval < 76) {
if (Rs < 114) {
return 4;
}
else if (Rs >= 114) {
if (tcor < 97) {
if (Rs < 168) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 1;
}
}
else if (Rs >= 168) {
if (mu_mv_mag_sq < 89) {
if (diffMVdiffVal < -1) {
if (Rs < 222) {
return 4;
}
else if (Rs >= 222) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 89) {
if (Rs < 180) {
return 4;
}
else if (Rs >= 180) {
return 8;
}
}
}
}
else if (tcor >= 97) {
if (Rs < 129) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
}
else if (Rs >= 129) {
return 1;
}
}
}
}
else if (refDCval >= 76) {
if (ssDCval < 110) {
if (RsDiff < 1) {
if (posBalance < 47) {
if (AFD < 2) {
if (Rs < 207) {
return 4;
}
else if (Rs >= 207) {
return 2;
}
}
else if (AFD >= 2) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
return 2;
}
}
}
else if (posBalance >= 47) {
if (Rs < 184) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
else if (Rs >= 184) {
if (Cs < 295) {
return 2;
}
else if (Cs >= 295) {
if (Cs < 371) {
return 4;
}
else if (Cs >= 371) {
if (Rs < 261) {
return 2;
}
else if (Rs >= 261) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 1) {
if (RsDiff < 2) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 22) {
return 2;
}
else if (mu_mv_mag_sq >= 22) {
if (Rs < 201) {
if (Rs < 99) {
return 2;
}
else if (Rs >= 99) {
return 1;
}
}
else if (Rs >= 201) {
return 2;
}
}
}
}
else if (RsDiff >= 2) {
if (AFD < 20) {
if (refDCval < 104) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
else if (refDCval >= 104) {
return 4;
}
}
else if (AFD >= 20) {
if (refDCval < 88) {
return 1;
}
else if (refDCval >= 88) {
return 8;
}
}
}
}
}
else if (ssDCval >= 110) {
if (ssDCval < 128) {
if (Cs < 340) {
if (Rs < 258) {
return 4;
}
else if (Rs >= 258) {
return 8;
}
}
else if (Cs >= 340) {
return 2;
}
}
else if (ssDCval >= 128) {
if (Rs < 292) {
return 2;
}
else if (Rs >= 292) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 300) {
if (RsDiff < 4) {
if (refDCval < 75) {
if (mcTcor < 89) {
if (tcor < 71) {
if (negBalance < 47) {
if (posBalance < 55) {
if (Rs < 385) {
return 1;
}
else if (Rs >= 385) {
if (Cs < 458) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
return 8;
}
}
else if (Cs >= 458) {
if (Rs < 670) {
return 4;
}
else if (Rs >= 670) {
return 1;
}
}
}
}
else if (posBalance >= 55) {
if (diffMVdiffVal < 4) {
if (ssDCval < 57) {
if (tcor < 56) {
if (Rs < 365) {
return 8;
}
else if (Rs >= 365) {
return 4;
}
}
else if (tcor >= 56) {
return 2;
}
}
else if (ssDCval >= 57) {
if (refDCval < 61) {
return 8;
}
else if (refDCval >= 61) {
if (Rs < 349) {
if (Rs < 323) {
return 4;
}
else if (Rs >= 323) {
return 2;
}
}
else if (Rs >= 349) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 4) {
if (diffAFD < 7) {
if (negBalance < 33) {
if (Rs < 464) {
return 1;
}
else if (Rs >= 464) {
return 8;
}
}
else if (negBalance >= 33) {
return 8;
}
}
else if (diffAFD >= 7) {
return 1;
}
}
}
}
else if (negBalance >= 47) {
if (ssDCval < 53) {
if (posBalance < 45) {
if (gchDC < 7) {
return 1;
}
else if (gchDC >= 7) {
return 8;
}
}
else if (posBalance >= 45) {
if (mu_mv_mag_sq < 51) {
if (mcTcor < 79) {
if (refDCval < 38) {
if (mcTcor < 70) {
return 4;
}
else if (mcTcor >= 70) {
return 8;
}
}
else if (refDCval >= 38) {
return 8;
}
}
else if (mcTcor >= 79) {
if (MVDiff < 45) {
if (MVDiff < 24) {
if (Rs < 637) {
return 4;
}
else if (Rs >= 637) {
return 8;
}
}
else if (MVDiff >= 24) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 8;
}
}
}
else if (MVDiff >= 45) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 51) {
return 4;
}
}
}
else if (ssDCval >= 53) {
if (MVDiff < 63) {
if (refDCval < 54) {
if (Rs < 423) {
return 1;
}
else if (Rs >= 423) {
return 8;
}
}
else if (refDCval >= 54) {
return 4;
}
}
else if (MVDiff >= 63) {
if (Rs < 361) {
if (Rs < 318) {
return 8;
}
else if (Rs >= 318) {
return 4;
}
}
else if (Rs >= 361) {
if (SC < 1157) {
return 2;
}
else if (SC >= 1157) {
return 4;
}
}
}
}
}
}
else if (tcor >= 71) {
if (refDCval < 44) {
if (ssDCval < 35) {
if (Rs < 353) {
return 8;
}
else if (Rs >= 353) {
return 1;
}
}
else if (ssDCval >= 35) {
if (negBalance < 31) {
if (RsDiff < 2) {
if (Rs < 469) {
return 1;
}
else if (Rs >= 469) {
return 8;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (negBalance >= 31) {
return 8;
}
}
}
else if (refDCval >= 44) {
if (mu_mv_mag_sq < 102) {
if (ssDCval < 51) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 343) {
return 1;
}
else if (Rs >= 343) {
return 8;
}
}
}
else if (ssDCval >= 51) {
if (diffRsCsdiff < -11) {
return 2;
}
else if (diffRsCsdiff >= -11) {
if (posBalance < 56) {
if (mcTcor < 87) {
if (Cs < 505) {
return 2;
}
else if (Cs >= 505) {
return 4;
}
}
else if (mcTcor >= 87) {
if (Rs < 446) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (Rs >= 446) {
if (SC < 951) {
return 4;
}
else if (SC >= 951) {
return 1;
}
}
}
}
else if (posBalance >= 56) {
if (Rs < 520) {
if (Rs < 372) {
return 1;
}
else if (Rs >= 372) {
return 4;
}
}
else if (Rs >= 520) {
if (Cs < 363) {
return 4;
}
else if (Cs >= 363) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
return 8;
}
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1292) {
if (tcor < 101) {
if (MVDiff < 2) {
if (tcor < 96) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (mcTcor < 97) {
if (posBalance < 31) {
return 1;
}
else if (posBalance >= 31) {
if (ssDCval < 67) {
if (negBalance < 28) {
if (tcor < 91) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (tcor >= 91) {
if (Rs < 886) {
return 8;
}
else if (Rs >= 886) {
return 1;
}
}
}
else if (negBalance >= 28) {
if (Rs < 1000) {
return 8;
}
else if (Rs >= 1000) {
return 1;
}
}
}
else if (ssDCval >= 67) {
if (ssDCval < 73) {
if (SC < 1104) {
if (Rs < 491) {
return 2;
}
else if (Rs >= 491) {
return 1;
}
}
else if (SC >= 1104) {
return 4;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 428) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (SC < 943) {
if (Rs < 486) {
return 1;
}
else if (Rs >= 486) {
return 4;
}
}
else if (SC >= 943) {
return 1;
}
}
}
}
else if (Cs >= 428) {
if (Cs < 483) {
return 4;
}
else if (Cs >= 483) {
if (Rs < 792) {
return 8;
}
else if (Rs >= 792) {
return 1;
}
}
}
}
}
}
else if (tcor >= 96) {
if (ssDCval < 69) {
if (Rs < 443) {
if (Cs < 266) {
if (Rs < 383) {
return 4;
}
else if (Rs >= 383) {
return 8;
}
}
else if (Cs >= 266) {
if (negBalance < 19) {
if (SC < 810) {
return 8;
}
else if (SC >= 810) {
return 1;
}
}
else if (negBalance >= 19) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 2;
}
}
}
}
else if (Rs >= 443) {
if (posBalance < 42) {
if (SC < 1081) {
if (ssDCval < 60) {
if (mcTcor < 99) {
if (SC < 1057) {
return 1;
}
else if (SC >= 1057) {
return 8;
}
}
else if (mcTcor >= 99) {
if (AFD < 3) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (ssDCval < 56) {
if (Rs < 466) {
return 8;
}
else if (Rs >= 466) {
return 1;
}
}
else if (ssDCval >= 56) {
return 8;
}
}
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (ssDCval >= 60) {
if (Cs < 503) {
if (mu_mv_mag_sq < 106) {
if (negBalance < 20) {
return 8;
}
else if (negBalance >= 20) {
if (Cs < 472) {
return 1;
}
else if (Cs >= 472) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 106) {
return 1;
}
}
else if (Cs >= 503) {
return 8;
}
}
}
else if (SC >= 1081) {
if (Cs < 512) {
if (mu_mv_mag_sq < 120) {
return 1;
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 1284) {
if (Rs < 956) {
if (SC < 1236) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
else if (SC >= 1236) {
if (Cs < 362) {
return 1;
}
else if (Cs >= 362) {
return 8;
}
}
}
else if (Rs >= 956) {
return 1;
}
}
else if (Rs >= 1284) {
return 8;
}
}
}
else if (Cs >= 512) {
if (Rs < 1002) {
return 8;
}
else if (Rs >= 1002) {
return 1;
}
}
}
}
else if (posBalance >= 42) {
if (Rs < 673) {
if (ssDCval < 67) {
if (Cs < 479) {
if (Cs < 328) {
return 8;
}
else if (Cs >= 328) {
if (mu_mv_mag_sq < 81) {
if (Rs < 557) {
if (Cs < 404) {
return 1;
}
else if (Cs >= 404) {
return 8;
}
}
else if (Rs >= 557) {
return 1;
}
}
else if (mu_mv_mag_sq >= 81) {
return 1;
}
}
}
else if (Cs >= 479) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (SC < 1024) {
if (SC < 909) {
if (Rs < 509) {
return 1;
}
else if (Rs >= 509) {
return 8;
}
}
else if (SC >= 909) {
return 2;
}
}
else if (SC >= 1024) {
return 1;
}
}
}
else if (Rs >= 673) {
if (mu_mv_mag_sq < 99) {
if (Rs < 780) {
return 8;
}
else if (Rs >= 780) {
return 1;
}
}
else if (mu_mv_mag_sq >= 99) {
if (Cs < 284) {
if (ssDCval < 62) {
if (Cs < 223) {
return 1;
}
else if (Cs >= 223) {
return 2;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (Cs >= 284) {
if (SC < 1251) {
return 2;
}
else if (SC >= 1251) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 883) {
if (Rs < 597) {
if (SC < 953) {
if (tcor < 98) {
if (Rs < 403) {
return 8;
}
else if (Rs >= 403) {
return 1;
}
}
else if (tcor >= 98) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (SC < 910) {
return 2;
}
else if (SC >= 910) {
return 1;
}
}
}
}
else if (SC >= 953) {
if (Cs < 387) {
return 2;
}
else if (Cs >= 387) {
return 1;
}
}
}
else if (Rs >= 597) {
if (ssDCval < 72) {
if (Cs < 337) {
return 4;
}
else if (Cs >= 337) {
if (Rs < 620) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (Rs >= 620) {
return 2;
}
}
}
else if (ssDCval >= 72) {
if (tcor < 99) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (tcor >= 99) {
return 4;
}
}
}
}
else if (Rs >= 883) {
if (Cs < 277) {
return 1;
}
else if (Cs >= 277) {
if (refDCval < 72) {
return 8;
}
else if (refDCval >= 72) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 124) {
if (mu_mv_mag_sq < 57) {
if (MVDiff < 10) {
if (refDCval < 69) {
if (diffMVdiffVal < 4) {
return 1;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 424) {
return 4;
}
else if (Rs >= 424) {
return 1;
}
}
}
else if (refDCval >= 69) {
return 4;
}
}
else if (MVDiff >= 10) {
if (Rs < 1275) {
if (posBalance < 56) {
if (Rs < 561) {
if (negBalance < 39) {
if (ssDCval < 38) {
return 1;
}
else if (ssDCval >= 38) {
return 8;
}
}
else if (negBalance >= 39) {
if (Rs < 327) {
if (Rs < 310) {
return 1;
}
else if (Rs >= 310) {
return 8;
}
}
else if (Rs >= 327) {
return 1;
}
}
}
else if (Rs >= 561) {
return 1;
}
}
else if (posBalance >= 56) {
return 1;
}
}
else if (Rs >= 1275) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (TSC < 3) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (TSC >= 3) {
if (Cs < 86) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 86) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 2;
}
}
}
}
}
else if (Cs >= 124) {
if (ssDCval < 71) {
if (refDCval < 58) {
if (TSC < 4) {
if (TSCindex < 2) {
if (AFD < 2) {
if (Rs < 814) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 4;
}
}
else if (Rs >= 814) {
return 1;
}
}
else if (AFD >= 2) {
return 4;
}
}
else if (TSCindex >= 2) {
if (Cs < 420) {
if (Rs < 525) {
if (ssDCval < 46) {
if (posBalance < 38) {
if (mu_mv_mag_sq < 87) {
if (mcTcor < 96) {
if (Rs < 425) {
return 4;
}
else if (Rs >= 425) {
return 2;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (mu_mv_mag_sq >= 87) {
return 1;
}
}
else if (posBalance >= 38) {
if (mcTcor < 92) {
if (SC < 557) {
return 8;
}
else if (SC >= 557) {
if (Rs < 493) {
if (Rs < 313) {
return 4;
}
else if (Rs >= 313) {
return 1;
}
}
else if (Rs >= 493) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 71) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (MVDiff < 5) {
if (Rs < 359) {
return 1;
}
else if (Rs >= 359) {
return 4;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 71) {
return 1;
}
}
}
}
else if (ssDCval >= 46) {
if (MVDiff < 3) {
if (Rs < 510) {
return 1;
}
else if (Rs >= 510) {
return 8;
}
}
else if (MVDiff >= 3) {
if (CsDiff < 1) {
if (AFD < 3) {
if (Rs < 313) {
return 2;
}
else if (Rs >= 313) {
return 4;
}
}
else if (AFD >= 3) {
if (Rs < 339) {
if (Cs < 137) {
if (Rs < 330) {
return 1;
}
else if (Rs >= 330) {
return 2;
}
}
else if (Cs >= 137) {
return 4;
}
}
else if (Rs >= 339) {
if (diffRsCsdiff < -5) {
return 4;
}
else if (diffRsCsdiff >= -5) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (refDCval < 48) {
return 1;
}
else if (refDCval >= 48) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 1;
}
}
}
}
}
}
}
else if (CsDiff >= 1) {
if (SC < 597) {
if (ssDCval < 56) {
if (Rs < 310) {
return 2;
}
else if (Rs >= 310) {
return 4;
}
}
else if (ssDCval >= 56) {
return 2;
}
}
else if (SC >= 597) {
if (gchDC < 3) {
if (mu_mv_mag_sq < 60) {
if (MVDiff < 6) {
if (Cs < 301) {
return 4;
}
else if (Cs >= 301) {
return 1;
}
}
else if (MVDiff >= 6) {
if (AFD < 5) {
if (SC < 717) {
return 4;
}
else if (SC >= 717) {
return 1;
}
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 60) {
return 4;
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
}
}
else if (Rs >= 525) {
if (posBalance < 30) {
return 2;
}
else if (posBalance >= 30) {
if (mcTcor < 99) {
if (negBalance < 44) {
if (MVDiff < 42) {
return 1;
}
else if (MVDiff >= 42) {
if (Rs < 1047) {
return 4;
}
else if (Rs >= 1047) {
return 1;
}
}
}
else if (negBalance >= 44) {
if (posBalance < 56) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (RsCsDiff < 5) {
return 1;
}
else if (RsCsDiff >= 5) {
return 2;
}
}
}
}
}
else if (Cs >= 420) {
if (Cs < 437) {
if (ssDCval < 51) {
if (MVDiff < 6) {
if (Rs < 494) {
return 1;
}
else if (Rs >= 494) {
return 4;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (ssDCval >= 51) {
if (negBalance < 28) {
if (Rs < 614) {
return 8;
}
else if (Rs >= 614) {
return 1;
}
}
else if (negBalance >= 28) {
if (Cs < 433) {
return 1;
}
else if (Cs >= 433) {
if (Rs < 610) {
return 2;
}
else if (Rs >= 610) {
return 8;
}
}
}
}
}
else if (Cs >= 437) {
if (ssDCval < 43) {
if (diffRsCsdiff < 3) {
if (Rs < 657) {
if (diffRsCsdiff < -4) {
return 1;
}
else if (diffRsCsdiff >= -4) {
return 8;
}
}
else if (Rs >= 657) {
if (Rs < 777) {
return 2;
}
else if (Rs >= 777) {
return 8;
}
}
}
else if (diffRsCsdiff >= 3) {
return 1;
}
}
else if (ssDCval >= 43) {
if (RsDiff < 2) {
if (negBalance < 21) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
if (Cs < 489) {
return 8;
}
else if (Cs >= 489) {
if (Rs < 423) {
return 4;
}
else if (Rs >= 423) {
return 1;
}
}
}
}
else if (negBalance >= 21) {
if (Rs < 887) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (Rs < 440) {
if (CsDiff < 2) {
if (Cs < 509) {
return 1;
}
else if (Cs >= 509) {
return 4;
}
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (Rs >= 440) {
return 4;
}
}
}
else if (Rs >= 887) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (AFD < 7) {
if (Rs < 513) {
if (mu_mv_mag_sq < 72) {
return 2;
}
else if (mu_mv_mag_sq >= 72) {
return 1;
}
}
else if (Rs >= 513) {
if (TSC < 3) {
if (Rs < 659) {
return 4;
}
else if (Rs >= 659) {
return 2;
}
}
else if (TSC >= 3) {
if (diffMVdiffVal < -7) {
return 8;
}
else if (diffMVdiffVal >= -7) {
return 4;
}
}
}
}
else if (AFD >= 7) {
if (mu_mv_mag_sq < 24) {
return 1;
}
else if (mu_mv_mag_sq >= 24) {
return 8;
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (SC < 1030) {
if (MVDiff < 54) {
if (mcTcor < 92) {
if (Rs < 375) {
if (Cs < 277) {
return 4;
}
else if (Cs >= 277) {
return 1;
}
}
else if (Rs >= 375) {
if (refDCval < 37) {
return 1;
}
else if (refDCval >= 37) {
if (refDCval < 43) {
return 8;
}
else if (refDCval >= 43) {
if (Cs < 347) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 8;
}
}
else if (Cs >= 347) {
return 1;
}
}
}
}
}
else if (mcTcor >= 92) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Rs < 466) {
if (negBalance < 45) {
return 4;
}
else if (negBalance >= 45) {
if (MVDiff < 38) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (MVDiff >= 38) {
return 1;
}
}
}
else if (Rs >= 466) {
if (MVDiff < 7) {
if (Rs < 532) {
return 8;
}
else if (Rs >= 532) {
return 4;
}
}
else if (MVDiff >= 7) {
if (posBalance < 52) {
if (Rs < 556) {
return 4;
}
else if (Rs >= 556) {
return 8;
}
}
else if (posBalance >= 52) {
if (diffMVdiffVal < 11) {
return 1;
}
else if (diffMVdiffVal >= 11) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 54) {
if (mu_mv_mag_sq < 7) {
return 2;
}
else if (mu_mv_mag_sq >= 7) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
else if (SC >= 1030) {
if (negBalance < 32) {
if (diffMVdiffVal < 1) {
return 1;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 823) {
return 4;
}
else if (Rs >= 823) {
return 2;
}
}
}
else if (negBalance >= 32) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (posBalance < 58) {
if (Rs < 824) {
return 8;
}
else if (Rs >= 824) {
return 1;
}
}
else if (posBalance >= 58) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 58) {
if (mu_mv_mag_sq < 78) {
if (refDCval < 62) {
if (mcTcor < 97) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (SC < 1395) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (diffAFD < 2) {
if (negBalance < 39) {
if (Rs < 328) {
return 4;
}
else if (Rs >= 328) {
if (MVDiff < 6) {
if (SC < 1039) {
return 1;
}
else if (SC >= 1039) {
if (Rs < 789) {
return 2;
}
else if (Rs >= 789) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (mcTcor < 93) {
if (Rs < 647) {
return 1;
}
else if (Rs >= 647) {
return 2;
}
}
else if (mcTcor >= 93) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (SC < 1082) {
return 2;
}
else if (SC >= 1082) {
return 4;
}
}
}
else if (diffAFD >= 2) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (Rs < 409) {
return 2;
}
else if (Rs >= 409) {
return 1;
}
}
}
}
}
else if (SC >= 1395) {
if (Rs < 1068) {
return 4;
}
else if (Rs >= 1068) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (tcor < 95) {
if (tcor < 81) {
return 2;
}
else if (tcor >= 81) {
return 1;
}
}
else if (tcor >= 95) {
if (negBalance < 24) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (negBalance >= 24) {
if (Rs < 708) {
if (mu_mv_mag_sq < 28) {
return 4;
}
else if (mu_mv_mag_sq >= 28) {
return 1;
}
}
else if (Rs >= 708) {
if (Rs < 1051) {
return 1;
}
else if (Rs >= 1051) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 62) {
if (posBalance < 41) {
if (mcTcor < 100) {
if (diffMVdiffVal < 3) {
if (MVDiff < 19) {
return 1;
}
else if (MVDiff >= 19) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (Rs < 617) {
return 8;
}
else if (Rs >= 617) {
return 1;
}
}
}
}
else if (mcTcor >= 100) {
return 2;
}
}
else if (posBalance >= 41) {
if (negBalance < 47) {
if (Cs < 194) {
if (Rs < 491) {
return 8;
}
else if (Rs >= 491) {
return 1;
}
}
else if (Cs >= 194) {
if (gchDC < 2) {
if (Rs < 320) {
if (TSC < 3) {
if (Cs < 327) {
return 2;
}
else if (Cs >= 327) {
return 4;
}
}
else if (TSC >= 3) {
return 1;
}
}
else if (Rs >= 320) {
if (tcor < 96) {
if (posBalance < 47) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (refDCval < 70) {
return 4;
}
else if (refDCval >= 70) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (Rs < 586) {
if (posBalance < 58) {
if (diffMVdiffVal < -1) {
if (SC < 813) {
return 1;
}
else if (SC >= 813) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
if (SC < 940) {
if (refDCval < 70) {
return 4;
}
else if (refDCval >= 70) {
return 1;
}
}
else if (SC >= 940) {
return 1;
}
}
}
else if (posBalance >= 58) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
else if (Rs >= 586) {
return 1;
}
}
}
else if (tcor >= 96) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (mu_mv_mag_sq < 38) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (Rs < 419) {
return 1;
}
else if (Rs >= 419) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 467) {
return 2;
}
else if (Rs >= 467) {
return 8;
}
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 847) {
if (diffMVdiffVal < 16) {
if (mu_mv_mag_sq < 40) {
if (Rs < 316) {
return 1;
}
else if (Rs >= 316) {
return 4;
}
}
else if (mu_mv_mag_sq >= 40) {
if (Rs < 436) {
return 4;
}
else if (Rs >= 436) {
if (Rs < 452) {
return 2;
}
else if (Rs >= 452) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 16) {
if (Rs < 427) {
return 1;
}
else if (Rs >= 427) {
return 2;
}
}
}
else if (Rs >= 847) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (Cs < 274) {
if (posBalance < 37) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (Cs < 253) {
return 4;
}
else if (Cs >= 253) {
return 8;
}
}
}
else if (posBalance >= 37) {
if (Cs < 217) {
if (Rs < 714) {
return 2;
}
else if (Rs >= 714) {
return 1;
}
}
else if (Cs >= 217) {
return 4;
}
}
}
else if (Cs >= 274) {
if (tcor < 99) {
if (negBalance < 36) {
if (SC < 676) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
if (Rs < 319) {
return 1;
}
else if (Rs >= 319) {
return 4;
}
}
}
else if (SC >= 676) {
if (mcTcor < 97) {
if (Cs < 468) {
if (Rs < 486) {
return 1;
}
else if (Rs >= 486) {
return 4;
}
}
else if (Cs >= 468) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 28) {
if (Cs < 439) {
if (Rs < 665) {
if (mu_mv_mag_sq < 102) {
if (mu_mv_mag_sq < 82) {
return 1;
}
else if (mu_mv_mag_sq >= 82) {
return 8;
}
}
else if (mu_mv_mag_sq >= 102) {
return 1;
}
}
else if (Rs >= 665) {
return 2;
}
}
else if (Cs >= 439) {
if (posBalance < 48) {
if (ssDCval < 60) {
return 2;
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (posBalance >= 48) {
if (Rs < 580) {
return 8;
}
else if (Rs >= 580) {
return 2;
}
}
}
}
else if (negBalance >= 28) {
if (tcor < 94) {
if (Rs < 457) {
return 8;
}
else if (Rs >= 457) {
return 1;
}
}
else if (tcor >= 94) {
return 1;
}
}
}
}
}
else if (negBalance >= 36) {
if (mu_mv_mag_sq < 100) {
if (MVDiff < 6) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 1;
}
}
else if (MVDiff >= 6) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 595) {
return 1;
}
else if (Rs >= 595) {
return 4;
}
}
}
}
else if (tcor >= 99) {
if (mu_mv_mag_sq < 103) {
if (Rs < 414) {
return 4;
}
else if (Rs >= 414) {
return 8;
}
}
else if (mu_mv_mag_sq >= 103) {
if (negBalance < 32) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
return 2;
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (mu_mv_mag_sq < 96) {
if (Rs < 555) {
if (mu_mv_mag_sq < 20) {
if (RsDiff < 3) {
if (Rs < 472) {
return 1;
}
else if (Rs >= 472) {
if (Cs < 314) {
return 4;
}
else if (Cs >= 314) {
return 1;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 392) {
return 4;
}
else if (Rs >= 392) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 20) {
if (MVDiff < 4) {
if (Rs < 451) {
return 8;
}
else if (Rs >= 451) {
return 1;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (Rs >= 555) {
if (Rs < 710) {
return 4;
}
else if (Rs >= 710) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
if (Rs < 983) {
return 4;
}
else if (Rs >= 983) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (posBalance < 49) {
if (Rs < 514) {
return 8;
}
else if (Rs >= 514) {
return 4;
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
}
}
else if (tcor >= 101) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
if (TSCindex < 3) {
if (ssDCval < 39) {
return 1;
}
else if (ssDCval >= 39) {
return 8;
}
}
else if (TSCindex >= 3) {
if (negBalance < 34) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
if (Cs < 323) {
return 8;
}
else if (Cs >= 323) {
return 2;
}
}
}
else if (negBalance >= 34) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (AFD < 5) {
if (Cs < 163) {
return 8;
}
else if (Cs >= 163) {
return 4;
}
}
else if (AFD >= 5) {
if (mu_mv_mag_sq < 27) {
if (Rs < 610) {
return 1;
}
else if (Rs >= 610) {
return 4;
}
}
else if (mu_mv_mag_sq >= 27) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1292) {
if (SC < 1587) {
if (mu_mv_mag_sq < 64) {
return 1;
}
else if (mu_mv_mag_sq >= 64) {
return 8;
}
}
else if (SC >= 1587) {
if (mu_mv_mag_sq < 126) {
if (posBalance < 60) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 24) {
if (Rs < 1977) {
return 1;
}
else if (Rs >= 1977) {
return 4;
}
}
else if (mu_mv_mag_sq >= 24) {
return 1;
}
}
else if (ssDCval >= 56) {
if (Rs < 1804) {
if (mu_mv_mag_sq < 54) {
if (negBalance < 26) {
if (Cs < 384) {
return 4;
}
else if (Cs >= 384) {
return 2;
}
}
else if (negBalance >= 26) {
if (tcor < 97) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
else if (tcor >= 97) {
if (Rs < 1612) {
return 2;
}
else if (Rs >= 1612) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (TSCindex < 2) {
if (ssDCval < 64) {
if (Rs < 1454) {
return 2;
}
else if (Rs >= 1454) {
return 1;
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (TSCindex >= 2) {
if (ssDCval < 63) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (ssDCval >= 63) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
if (AFD < 4) {
if (Cs < 200) {
return 1;
}
else if (Cs >= 200) {
if (SC < 1811) {
return 8;
}
else if (SC >= 1811) {
if (SC < 1891) {
return 1;
}
else if (SC >= 1891) {
if (Rs < 1672) {
return 8;
}
else if (Rs >= 1672) {
return 1;
}
}
}
}
}
else if (AFD >= 4) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1804) {
if (SC < 2314) {
return 2;
}
else if (SC >= 2314) {
if (Rs < 1883) {
return 4;
}
else if (Rs >= 1883) {
return 8;
}
}
}
}
}
else if (posBalance >= 60) {
if (Cs < 211) {
if (MVDiff < 235) {
return 1;
}
else if (MVDiff >= 235) {
return 2;
}
}
else if (Cs >= 211) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (ssDCval < 62) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (ssDCval >= 62) {
if (refDCval < 65) {
if (negBalance < 23) {
return 2;
}
else if (negBalance >= 23) {
return 8;
}
}
else if (refDCval >= 65) {
if (Rs < 1412) {
return 8;
}
else if (Rs >= 1412) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (Cs < 142) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 72) {
if (ssDCval < 82) {
if (Rs < 819) {
if (TSC < 4) {
if (Cs < 4) {
return 1;
}
else if (Cs >= 4) {
return 8;
}
}
else if (TSC >= 4) {
if (Cs < 79) {
return 1;
}
else if (Cs >= 79) {
return 8;
}
}
}
else if (Rs >= 819) {
return 1;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
else if (mu_mv_mag_sq >= 72) {
if (posBalance < 56) {
if (posBalance < 31) {
if (Cs < 105) {
return 8;
}
else if (Cs >= 105) {
return 1;
}
}
else if (posBalance >= 31) {
if (AFD < 3) {
if (MVDiff < 1) {
if (Rs < 1359) {
return 8;
}
else if (Rs >= 1359) {
return 1;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (AFD >= 3) {
if (Rs < 1002) {
return 8;
}
else if (Rs >= 1002) {
return 1;
}
}
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (tcor < 59) {
if (negBalance < 50) {
if (mcTcor < 85) {
return 2;
}
else if (mcTcor >= 85) {
return 8;
}
}
else if (negBalance >= 50) {
if (Cs < 61) {
return 8;
}
else if (Cs >= 61) {
if (Cs < 108) {
return 2;
}
else if (Cs >= 108) {
return 8;
}
}
}
}
else if (tcor >= 59) {
if (gchDC < 12) {
if (MVDiff < 15) {
if (posBalance < 63) {
if (Rs < 878) {
return 2;
}
else if (Rs >= 878) {
return 4;
}
}
else if (posBalance >= 63) {
return 1;
}
}
else if (MVDiff >= 15) {
return 1;
}
}
else if (gchDC >= 12) {
return 8;
}
}
}
}
else if (Cs >= 142) {
if (SC < 1873) {
if (SC < 899) {
if (mcTcor < 97) {
if (MVDiff < 23) {
if (negBalance < 60) {
if (TSC < 6) {
if (AFD < 9) {
if (Cs < 164) {
return 1;
}
else if (Cs >= 164) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (Rs < 563) {
if (posBalance < 56) {
if (tcor < 85) {
if (refDCval < 88) {
return 1;
}
else if (refDCval >= 88) {
return 8;
}
}
else if (tcor >= 85) {
return 8;
}
}
else if (posBalance >= 56) {
if (refDCval < 85) {
return 8;
}
else if (refDCval >= 85) {
if (MVDiff < 19) {
if (SC < 892) {
return 2;
}
else if (SC >= 892) {
return 8;
}
}
else if (MVDiff >= 19) {
return 8;
}
}
}
}
else if (Rs >= 563) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 1;
}
}
}
}
}
else if (AFD >= 9) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
if (Rs < 368) {
return 1;
}
else if (Rs >= 368) {
if (Rs < 534) {
return 2;
}
else if (Rs >= 534) {
return 8;
}
}
}
}
}
else if (TSC >= 6) {
if (Rs < 485) {
if (Cs < 191) {
if (Rs < 350) {
return 8;
}
else if (Rs >= 350) {
return 2;
}
}
else if (Cs >= 191) {
return 8;
}
}
else if (Rs >= 485) {
if (Rs < 566) {
if (Rs < 513) {
return 4;
}
else if (Rs >= 513) {
return 8;
}
}
else if (Rs >= 566) {
return 1;
}
}
}
}
else if (negBalance >= 60) {
if (tcor < 90) {
if (Cs < 306) {
return 1;
}
else if (Cs >= 306) {
return 2;
}
}
else if (tcor >= 90) {
if (Cs < 284) {
return 2;
}
else if (Cs >= 284) {
return 8;
}
}
}
}
else if (MVDiff >= 23) {
if (Cs < 256) {
if (Rs < 393) {
if (Rs < 332) {
return 8;
}
else if (Rs >= 332) {
return 2;
}
}
else if (Rs >= 393) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (Cs >= 256) {
if (MVDiff < 101) {
if (refDCval < 99) {
if (AFD < 10) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
return 8;
}
}
else if (AFD >= 10) {
if (mcTcor < 89) {
if (Rs < 317) {
return 1;
}
else if (Rs >= 317) {
return 2;
}
}
else if (mcTcor >= 89) {
if (mu_mv_mag_sq < 10) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (mu_mv_mag_sq >= 10) {
if (Rs < 432) {
return 1;
}
else if (Rs >= 432) {
if (Rs < 494) {
return 8;
}
else if (Rs >= 494) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 99) {
if (diffRsCsdiff < -5) {
if (Rs < 346) {
return 8;
}
else if (Rs >= 346) {
return 4;
}
}
else if (diffRsCsdiff >= -5) {
if (Cs < 264) {
return 4;
}
else if (Cs >= 264) {
return 2;
}
}
}
}
else if (MVDiff >= 101) {
if (Rs < 342) {
if (Rs < 323) {
return 4;
}
else if (Rs >= 323) {
return 8;
}
}
else if (Rs >= 342) {
if (Rs < 361) {
return 2;
}
else if (Rs >= 361) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 104) {
if (refDCval < 78) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (AFD < 11) {
if (Rs < 486) {
if (Cs < 387) {
return 2;
}
else if (Cs >= 387) {
return 4;
}
}
else if (Rs >= 486) {
if (Cs < 237) {
return 4;
}
else if (Cs >= 237) {
return 1;
}
}
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (refDCval >= 78) {
if (ssDCval < 116) {
if (mu_mv_mag_sq < 77) {
if (refDCval < 92) {
if (negBalance < 38) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (negBalance >= 38) {
if (ssDCval < 82) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
if (Rs < 604) {
if (Rs < 497) {
return 8;
}
else if (Rs >= 497) {
return 4;
}
}
else if (Rs >= 604) {
return 1;
}
}
}
else if (ssDCval >= 82) {
return 1;
}
}
}
else if (refDCval >= 92) {
if (AFD < 14) {
if (MVDiff < 24) {
return 1;
}
else if (MVDiff >= 24) {
if (Rs < 525) {
if (Cs < 248) {
return 4;
}
else if (Cs >= 248) {
return 8;
}
}
else if (Rs >= 525) {
return 1;
}
}
}
else if (AFD >= 14) {
if (Rs < 374) {
return 2;
}
else if (Rs >= 374) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
if (ssDCval < 96) {
if (tcor < 98) {
if (Rs < 512) {
return 1;
}
else if (Rs >= 512) {
return 8;
}
}
else if (tcor >= 98) {
if (Cs < 237) {
return 2;
}
else if (Cs >= 237) {
return 8;
}
}
}
else if (ssDCval >= 96) {
return 1;
}
}
}
else if (ssDCval >= 116) {
if (mcTcor < 99) {
if (AFD < 5) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (ssDCval < 124) {
return 2;
}
else if (ssDCval >= 124) {
if (Cs < 370) {
return 2;
}
else if (Cs >= 370) {
return 4;
}
}
}
}
else if (AFD >= 5) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
if (Cs < 361) {
return 2;
}
else if (Cs >= 361) {
if (negBalance < 35) {
if (Rs < 457) {
return 2;
}
else if (Rs >= 457) {
return 4;
}
}
else if (negBalance >= 35) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (Rs < 365) {
if (ssDCval < 126) {
return 2;
}
else if (ssDCval >= 126) {
return 4;
}
}
else if (Rs >= 365) {
if (ssDCval < 154) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 4;
}
}
else if (ssDCval >= 154) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Cs < 237) {
if (ssDCval < 94) {
if (Rs < 313) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (Rs >= 313) {
return 2;
}
}
else if (ssDCval >= 94) {
return 1;
}
}
else if (Cs >= 237) {
if (Rs < 444) {
return 8;
}
else if (Rs >= 444) {
return 1;
}
}
}
}
}
else if (SC >= 899) {
if (CsDiff < 2) {
if (negBalance < 32) {
if (ssDCval < 94) {
if (SC < 1240) {
if (refDCval < 77) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 4;
}
}
else if (refDCval >= 77) {
if (MVDiff < 5) {
if (posBalance < 23) {
return 8;
}
else if (posBalance >= 23) {
return 1;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (SC >= 1240) {
if (mu_mv_mag_sq < 117) {
if (Cs < 395) {
return 1;
}
else if (Cs >= 395) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 117) {
return 2;
}
}
}
else if (ssDCval >= 94) {
if (MVDiff < 3) {
if (Cs < 466) {
if (ssDCval < 99) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (ssDCval >= 99) {
return 8;
}
}
else if (Cs >= 466) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 2;
}
}
}
else if (MVDiff >= 3) {
if (posBalance < 54) {
if (mcTcor < 99) {
if (negBalance < 17) {
if (ssDCval < 156) {
return 1;
}
else if (ssDCval >= 156) {
return 8;
}
}
else if (negBalance >= 17) {
return 8;
}
}
else if (mcTcor >= 99) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (Rs < 561) {
return 2;
}
else if (Rs >= 561) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Rs < 744) {
return 1;
}
else if (Rs >= 744) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 32) {
if (RsCsDiff < 2) {
if (Rs < 539) {
if (RsCsDiff < 1) {
return 8;
}
else if (RsCsDiff >= 1) {
if (Rs < 506) {
return 1;
}
else if (Rs >= 506) {
return 2;
}
}
}
else if (Rs >= 539) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
return 8;
}
}
}
else if (RsCsDiff >= 2) {
if (mu_mv_mag_sq < 2) {
if (Cs < 385) {
if (ssDCval < 89) {
if (MVDiff < 21) {
if (refDCval < 76) {
return 4;
}
else if (refDCval >= 76) {
return 1;
}
}
else if (MVDiff >= 21) {
if (Rs < 807) {
return 4;
}
else if (Rs >= 807) {
return 8;
}
}
}
else if (ssDCval >= 89) {
if (ssDCval < 130) {
return 8;
}
else if (ssDCval >= 130) {
return 4;
}
}
}
else if (Cs >= 385) {
if (ssDCval < 112) {
return 1;
}
else if (ssDCval >= 112) {
if (ssDCval < 125) {
if (Rs < 493) {
return 2;
}
else if (Rs >= 493) {
return 8;
}
}
else if (ssDCval >= 125) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 258) {
if (SC < 1511) {
return 1;
}
else if (SC >= 1511) {
return 4;
}
}
else if (Cs >= 258) {
if (Rs < 674) {
if (posBalance < 53) {
if (Cs < 402) {
if (SC < 1022) {
return 4;
}
else if (SC >= 1022) {
return 8;
}
}
else if (Cs >= 402) {
if (ssDCval < 118) {
if (mu_mv_mag_sq < 89) {
return 2;
}
else if (mu_mv_mag_sq >= 89) {
if (Cs < 517) {
return 1;
}
else if (Cs >= 517) {
return 2;
}
}
}
else if (ssDCval >= 118) {
if (Cs < 491) {
return 1;
}
else if (Cs >= 491) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
if (Rs < 401) {
return 4;
}
else if (Rs >= 401) {
if (SC < 922) {
if (SC < 913) {
return 8;
}
else if (SC >= 913) {
return 1;
}
}
else if (SC >= 922) {
return 8;
}
}
}
}
}
else if (Rs >= 674) {
if (posBalance < 63) {
if (ssDCval < 123) {
if (Rs < 693) {
if (Cs < 428) {
return 8;
}
else if (Cs >= 428) {
return 1;
}
}
else if (Rs >= 693) {
return 8;
}
}
else if (ssDCval >= 123) {
return 1;
}
}
else if (posBalance >= 63) {
return 1;
}
}
}
}
}
}
}
else if (CsDiff >= 2) {
if (mcTcor < 96) {
if (mu_mv_mag_sq < 27) {
if (negBalance < 50) {
if (Rs < 747) {
if (diffMVdiffVal < -2) {
if (Rs < 518) {
return 2;
}
else if (Rs >= 518) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
else if (Rs >= 747) {
if (Rs < 936) {
return 4;
}
else if (Rs >= 936) {
return 1;
}
}
}
else if (negBalance >= 50) {
if (mcTcor < 92) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
return 2;
}
}
else if (mcTcor >= 92) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
if (TSC < 6) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (ssDCval < 79) {
if (Cs < 494) {
return 1;
}
else if (Cs >= 494) {
if (Rs < 438) {
return 2;
}
else if (Rs >= 438) {
return 4;
}
}
}
else if (ssDCval >= 79) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 63) {
if (Cs < 506) {
if (MVDiff < 26) {
if (diffRsCsdiff < -22) {
return 4;
}
else if (diffRsCsdiff >= -22) {
if (SC < 1728) {
if (negBalance < 60) {
if (Cs < 375) {
if (ssDCval < 94) {
return 1;
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (Cs >= 375) {
return 1;
}
}
else if (negBalance >= 60) {
if (Rs < 688) {
if (Rs < 597) {
return 1;
}
else if (Rs >= 597) {
return 4;
}
}
else if (Rs >= 688) {
return 8;
}
}
}
else if (SC >= 1728) {
return 8;
}
}
}
else if (MVDiff >= 26) {
if (negBalance < 58) {
if (Rs < 687) {
return 8;
}
else if (Rs >= 687) {
return 4;
}
}
else if (negBalance >= 58) {
return 2;
}
}
}
else if (Cs >= 506) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
if (gchDC < 1) {
if (Rs < 479) {
return 8;
}
else if (Rs >= 479) {
return 2;
}
}
else if (gchDC >= 1) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 508) {
return 2;
}
else if (Rs >= 508) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
}
else if (SC >= 1873) {
if (Rs < 2046) {
if (refDCval < 91) {
if (Cs < 487) {
if (mu_mv_mag_sq < 2) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsCsDiff < 5) {
return 1;
}
else if (RsCsDiff >= 5) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
return 1;
}
}
}
}
else if (Cs >= 487) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (MVDiff < 2) {
if (Rs < 1859) {
return 1;
}
else if (Rs >= 1859) {
return 8;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
else if (refDCval >= 91) {
if (ssDCval < 118) {
if (ssDCval < 96) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (Rs < 2021) {
return 8;
}
else if (Rs >= 2021) {
return 1;
}
}
}
else if (ssDCval >= 96) {
if (Rs < 2006) {
return 8;
}
else if (Rs >= 2006) {
if (Rs < 2019) {
return 4;
}
else if (Rs >= 2019) {
return 8;
}
}
}
}
else if (ssDCval >= 118) {
if (Rs < 1814) {
if (Cs < 421) {
return 4;
}
else if (Cs >= 421) {
return 2;
}
}
else if (Rs >= 1814) {
return 1;
}
}
}
}
else if (Rs >= 2046) {
if (posBalance < 54) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
if (Rs < 2104) {
return 8;
}
else if (Rs >= 2104) {
return 1;
}
}
}
else if (posBalance >= 54) {
if (Cs < 345) {
return 1;
}
else if (Cs >= 345) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (RsDiff < 14) {
if (Cs < 242) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 69) {
if (Rs < 1398) {
if (diffTSC < 6) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (diffRsCsdiff < -60) {
if (Rs < 646) {
return 2;
}
else if (Rs >= 646) {
return 1;
}
}
else if (diffRsCsdiff >= -60) {
return 1;
}
}
}
else if (diffTSC >= 6) {
return 8;
}
}
else if (Rs >= 1398) {
if (MVDiff < 38) {
if (refDCval < 64) {
if (Rs < 1882) {
return 1;
}
else if (Rs >= 1882) {
return 4;
}
}
else if (refDCval >= 64) {
return 2;
}
}
else if (MVDiff >= 38) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (Rs < 575) {
if (AFD < 25) {
return 4;
}
else if (AFD >= 25) {
return 8;
}
}
else if (Rs >= 575) {
if (Cs < 44) {
return 1;
}
else if (Cs >= 44) {
return 2;
}
}
}
}
else if (ssDCval >= 84) {
if (gchDC < 4) {
if (mcTcor < 84) {
if (ssDCval < 126) {
if (MVDiff < 76) {
return 8;
}
else if (MVDiff >= 76) {
if (Cs < 153) {
return 4;
}
else if (Cs >= 153) {
return 2;
}
}
}
else if (ssDCval >= 126) {
if (Rs < 418) {
return 2;
}
else if (Rs >= 418) {
return 1;
}
}
}
else if (mcTcor >= 84) {
if (ssDCval < 138) {
if (posBalance < 64) {
if (mcTcor < 92) {
if (diffRsCsdiff < 10) {
if (Cs < 114) {
return 8;
}
else if (Cs >= 114) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
return 2;
}
}
}
else if (diffRsCsdiff >= 10) {
return 2;
}
}
else if (mcTcor >= 92) {
return 2;
}
}
else if (posBalance >= 64) {
if (Rs < 819) {
if (Cs < 107) {
return 2;
}
else if (Cs >= 107) {
return 1;
}
}
else if (Rs >= 819) {
return 2;
}
}
}
else if (ssDCval >= 138) {
if (mcTcor < 90) {
if (Rs < 996) {
return 2;
}
else if (Rs >= 996) {
return 1;
}
}
else if (mcTcor >= 90) {
return 4;
}
}
}
}
else if (gchDC >= 4) {
if (mu_mv_mag_sq < 50) {
return 1;
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 920) {
if (Cs < 89) {
return 1;
}
else if (Cs >= 89) {
return 2;
}
}
else if (Rs >= 920) {
return 1;
}
}
}
}
}
else if (Cs >= 242) {
if (gchDC < 2) {
if (posBalance < 46) {
if (RsDiff < 10) {
if (AFD < 13) {
if (Rs < 930) {
return 8;
}
else if (Rs >= 930) {
return 1;
}
}
else if (AFD >= 13) {
if (MVDiff < 35) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 1;
}
}
else if (MVDiff >= 35) {
return 1;
}
}
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (posBalance >= 46) {
if (MVDiff < 20) {
if (mcTcor < 90) {
if (SC < 830) {
return 1;
}
else if (SC >= 830) {
if (Cs < 525) {
return 8;
}
else if (Cs >= 525) {
if (Rs < 870) {
return 4;
}
else if (Rs >= 870) {
return 8;
}
}
}
}
else if (mcTcor >= 90) {
if (MVDiff < 10) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (Rs < 1406) {
return 1;
}
else if (Rs >= 1406) {
return 2;
}
}
}
else if (MVDiff >= 10) {
return 1;
}
}
}
else if (MVDiff >= 20) {
if (ssDCval < 59) {
if (TSC < 6) {
if (negBalance < 52) {
if (mcTcor < 87) {
if (Rs < 482) {
return 1;
}
else if (Rs >= 482) {
return 2;
}
}
else if (mcTcor >= 87) {
return 1;
}
}
else if (negBalance >= 52) {
return 4;
}
}
else if (TSC >= 6) {
if (tcor < 39) {
if (Rs < 493) {
return 2;
}
else if (Rs >= 493) {
return 1;
}
}
else if (tcor >= 39) {
if (diffAFD < -1) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (diffAFD >= -1) {
if (SC < 819) {
if (Rs < 477) {
return 8;
}
else if (Rs >= 477) {
return 2;
}
}
else if (SC >= 819) {
return 8;
}
}
}
}
}
else if (ssDCval >= 59) {
if (Rs < 1842) {
if (RsCsDiff < 74) {
if (mcTcor < 77) {
if (AFD < 18) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
return 4;
}
}
else if (AFD >= 18) {
return 8;
}
}
else if (mcTcor >= 77) {
if (tcor < 60) {
if (diffTSC < -1) {
if (Cs < 437) {
return 4;
}
else if (Cs >= 437) {
if (Rs < 1101) {
return 8;
}
else if (Rs >= 1101) {
return 2;
}
}
}
else if (diffTSC >= -1) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 1;
}
}
}
else if (tcor >= 60) {
if (posBalance < 60) {
if (Cs < 442) {
return 1;
}
else if (Cs >= 442) {
return 8;
}
}
else if (posBalance >= 60) {
if (MVDiff < 37) {
return 8;
}
else if (MVDiff >= 37) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 74) {
if (AFD < 26) {
if (posBalance < 65) {
return 4;
}
else if (posBalance >= 65) {
return 2;
}
}
else if (AFD >= 26) {
if (MVDiff < 45) {
if (Rs < 851) {
return 8;
}
else if (Rs >= 851) {
return 4;
}
}
else if (MVDiff >= 45) {
return 1;
}
}
}
}
else if (Rs >= 1842) {
if (TSC < 8) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (TSC >= 8) {
if (ssDCval < 128) {
return 8;
}
else if (ssDCval >= 128) {
if (Rs < 2470) {
return 4;
}
else if (Rs >= 2470) {
return 8;
}
}
}
}
}
}
}
}
else if (gchDC >= 2) {
if (diffAFD < 4) {
if (ssDCval < 49) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 39) {
return 1;
}
else if (mu_mv_mag_sq >= 39) {
if (Rs < 868) {
return 2;
}
else if (Rs >= 868) {
return 8;
}
}
}
else if (negBalance >= 44) {
if (Cs < 265) {
return 4;
}
else if (Cs >= 265) {
if (TSC < 5) {
if (Rs < 599) {
return 4;
}
else if (Rs >= 599) {
return 8;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
}
else if (ssDCval >= 49) {
if (TSC < 6) {
if (Cs < 495) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
}
else if (Cs >= 495) {
return 4;
}
}
else if (TSC >= 6) {
if (SC < 751) {
if (MVDiff < 185) {
return 8;
}
else if (MVDiff >= 185) {
return 4;
}
}
else if (SC >= 751) {
if (MVDiff < 21) {
if (Rs < 774) {
return 8;
}
else if (Rs >= 774) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 4;
}
}
}
}
else if (MVDiff >= 21) {
if (refDCval < 84) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
if (Cs < 407) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (negBalance < 50) {
return 4;
}
else if (negBalance >= 50) {
if (CsDiff < 4) {
if (refDCval < 76) {
if (AFD < 30) {
return 1;
}
else if (AFD >= 30) {
return 2;
}
}
else if (refDCval >= 76) {
return 2;
}
}
else if (CsDiff >= 4) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 1;
}
}
}
}
}
else if (Cs >= 407) {
if (ssDCval < 72) {
if (MVDiff < 64) {
if (AFD < 14) {
if (Rs < 833) {
return 1;
}
else if (Rs >= 833) {
return 8;
}
}
else if (AFD >= 14) {
if (Rs < 1221) {
if (MVDiff < 45) {
return 2;
}
else if (MVDiff >= 45) {
return 4;
}
}
else if (Rs >= 1221) {
return 4;
}
}
}
else if (MVDiff >= 64) {
if (Rs < 640) {
if (Rs < 457) {
return 2;
}
else if (Rs >= 457) {
return 1;
}
}
else if (Rs >= 640) {
return 8;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 597) {
return 4;
}
else if (Rs >= 597) {
return 1;
}
}
}
}
}
else if (refDCval >= 84) {
if (AFD < 29) {
if (RsDiff < 11) {
if (gchDC < 10) {
if (tcor < 66) {
return 2;
}
else if (tcor >= 66) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (posBalance < 72) {
return 8;
}
else if (posBalance >= 72) {
return 2;
}
}
}
}
else if (gchDC >= 10) {
return 1;
}
}
else if (RsDiff >= 11) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 4;
}
}
}
else if (AFD >= 29) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
}
}
}
}
}
else if (diffAFD >= 4) {
if (diffTSC < 9) {
if (tcor < 67) {
if (Cs < 457) {
if (Rs < 304) {
return 8;
}
else if (Rs >= 304) {
return 1;
}
}
else if (Cs >= 457) {
if (tcor < 56) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 1;
}
}
else if (tcor >= 56) {
if (Rs < 708) {
if (Rs < 583) {
return 2;
}
else if (Rs >= 583) {
return 4;
}
}
else if (Rs >= 708) {
return 8;
}
}
}
}
else if (tcor >= 67) {
if (refDCval < 117) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 8;
}
}
else if (refDCval >= 117) {
return 1;
}
}
}
else if (diffTSC >= 9) {
if (Rs < 547) {
return 2;
}
else if (Rs >= 547) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 14) {
if (MVDiff < 23) {
if (negBalance < 48) {
if (posBalance < 54) {
if (Cs < 389) {
return 1;
}
else if (Cs >= 389) {
return 2;
}
}
else if (posBalance >= 54) {
return 2;
}
}
else if (negBalance >= 48) {
return 2;
}
}
else if (MVDiff >= 23) {
if (posBalance < 49) {
if (SC < 2644) {
if (ssDCval < 48) {
return 2;
}
else if (ssDCval >= 48) {
return 1;
}
}
else if (SC >= 2644) {
return 2;
}
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 61) {
if (tcor < 59) {
if (SC < 2461) {
if (MVDiff < 73) {
if (Rs < 1968) {
return 1;
}
else if (Rs >= 1968) {
return 2;
}
}
else if (MVDiff >= 73) {
if (mu_mv_mag_sq < 51) {
if (diffRsCsdiff < -89) {
if (Rs < 1536) {
return 2;
}
else if (Rs >= 1536) {
return 4;
}
}
else if (diffRsCsdiff >= -89) {
return 2;
}
}
else if (mu_mv_mag_sq >= 51) {
return 4;
}
}
}
else if (SC >= 2461) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (tcor >= 59) {
if (AFD < 30) {
if (RsCsDiff < 203) {
return 1;
}
else if (RsCsDiff >= 203) {
if (gchDC < 2) {
if (AFD < 23) {
if (Rs < 1994) {
return 2;
}
else if (Rs >= 1994) {
return 4;
}
}
else if (AFD >= 23) {
return 1;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (AFD >= 30) {
if (MVDiff < 94) {
return 1;
}
else if (MVDiff >= 94) {
if (Rs < 1979) {
return 8;
}
else if (Rs >= 1979) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (diffAFD < 11) {
return 2;
}
else if (diffAFD >= 11) {
return 1;
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 77) {
if (RsCsDiff < 87) {
if (Cs < 260) {
if (ssDCval < 140) {
if (ssDCval < 84) {
if (negBalance < 37) {
if (posBalance < 57) {
if (Rs < 1366) {
if (Rs < 643) {
return 2;
}
else if (Rs >= 643) {
return 1;
}
}
else if (Rs >= 1366) {
if (Rs < 1847) {
return 2;
}
else if (Rs >= 1847) {
return 1;
}
}
}
else if (posBalance >= 57) {
return 1;
}
}
else if (negBalance >= 37) {
if (CsDiff < 2) {
if (tcor < 100) {
if (refDCval < 84) {
return 1;
}
else if (refDCval >= 84) {
if (Rs < 1662) {
return 1;
}
else if (Rs >= 1662) {
return 2;
}
}
}
else if (tcor >= 100) {
if (Cs < 193) {
return 1;
}
else if (Cs >= 193) {
return 2;
}
}
}
else if (CsDiff >= 2) {
if (SC < 791) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (SC < 758) {
return 1;
}
else if (SC >= 758) {
return 4;
}
}
}
else if (SC >= 791) {
if (Rs < 1761) {
return 1;
}
else if (Rs >= 1761) {
if (Rs < 1975) {
return 2;
}
else if (Rs >= 1975) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (MVDiff < 57) {
if (RsDiff < 5) {
if (MVDiff < 41) {
return 1;
}
else if (MVDiff >= 41) {
return 8;
}
}
else if (RsDiff >= 5) {
if (ssDCval < 115) {
if (AFD < 15) {
if (RsDiff < 8) {
if (tcor < 89) {
if (diffAFD < 2) {
if (Cs < 76) {
return 1;
}
else if (Cs >= 76) {
return 2;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (tcor >= 89) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
return 1;
}
}
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (AFD >= 15) {
if (RsDiff < 9) {
if (ssDCval < 111) {
return 1;
}
else if (ssDCval >= 111) {
return 2;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
}
else if (ssDCval >= 115) {
if (gchDC < 9) {
return 1;
}
else if (gchDC >= 9) {
return 2;
}
}
}
}
else if (MVDiff >= 57) {
if (SC < 958) {
if (MVDiff < 128) {
return 8;
}
else if (MVDiff >= 128) {
if (posBalance < 75) {
return 2;
}
else if (posBalance >= 75) {
return 1;
}
}
}
else if (SC >= 958) {
if (mu_mv_mag_sq < 6) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (TSC < 9) {
if (Rs < 1716) {
if (Rs < 1542) {
return 8;
}
else if (Rs >= 1542) {
return 1;
}
}
else if (Rs >= 1716) {
return 8;
}
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 6) {
if (diffRsCsdiff < -134) {
return 2;
}
else if (diffRsCsdiff >= -134) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (gchDC < 2) {
if (posBalance < 53) {
if (SC < 1148) {
return 2;
}
else if (SC >= 1148) {
return 1;
}
}
else if (posBalance >= 53) {
if (Rs < 1561) {
return 4;
}
else if (Rs >= 1561) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (refDCval < 173) {
return 1;
}
else if (refDCval >= 173) {
if (Rs < 684) {
return 1;
}
else if (Rs >= 684) {
return 2;
}
}
}
}
}
else if (Cs >= 260) {
if (Rs < 822) {
if (gchDC < 3) {
if (negBalance < 31) {
if (negBalance < 28) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
return 1;
}
}
else if (negBalance >= 28) {
return 2;
}
}
else if (negBalance >= 31) {
if (diffTSC < -1) {
if (Rs < 732) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (Rs < 603) {
return 2;
}
else if (Rs >= 603) {
return 1;
}
}
}
else if (Rs >= 732) {
return 1;
}
}
else if (diffTSC >= -1) {
if (Cs < 509) {
if (MVDiff < 23) {
if (diffMVdiffVal < -10) {
return 4;
}
else if (diffMVdiffVal >= -10) {
if (mu_mv_mag_sq < 59) {
if (Cs < 413) {
if (negBalance < 61) {
if (ssDCval < 55) {
if (Rs < 653) {
return 4;
}
else if (Rs >= 653) {
return 1;
}
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (negBalance >= 61) {
if (Rs < 707) {
return 4;
}
else if (Rs >= 707) {
return 8;
}
}
}
else if (Cs >= 413) {
if (diffAFD < 1) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 59) {
return 4;
}
}
}
else if (MVDiff >= 23) {
if (RsCsDiff < 46) {
return 4;
}
else if (RsCsDiff >= 46) {
return 1;
}
}
}
else if (Cs >= 509) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
}
else if (gchDC >= 3) {
if (diffAFD < -4) {
if (Cs < 401) {
return 8;
}
else if (Cs >= 401) {
return 2;
}
}
else if (diffAFD >= -4) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (mu_mv_mag_sq < 10) {
return 1;
}
else if (mu_mv_mag_sq >= 10) {
if (mu_mv_mag_sq < 46) {
return 8;
}
else if (mu_mv_mag_sq >= 46) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 822) {
if (Cs < 463) {
if (ssDCval < 98) {
if (SC < 2676) {
if (negBalance < 26) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (negBalance >= 26) {
if (tcor < 79) {
if (Cs < 437) {
if (SC < 1231) {
if (Rs < 880) {
return 8;
}
else if (Rs >= 880) {
return 4;
}
}
else if (SC >= 1231) {
return 1;
}
}
else if (Cs >= 437) {
if (Rs < 1226) {
return 4;
}
else if (Rs >= 1226) {
return 1;
}
}
}
else if (tcor >= 79) {
if (mu_mv_mag_sq < 71) {
if (AFD < 19) {
if (ssDCval < 66) {
if (AFD < 12) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
if (mcTcor < 97) {
if (diffRsCsdiff < -50) {
if (Rs < 1432) {
return 1;
}
else if (Rs >= 1432) {
return 2;
}
}
else if (diffRsCsdiff >= -50) {
return 1;
}
}
else if (mcTcor >= 97) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
if (refDCval < 64) {
return 8;
}
else if (refDCval >= 64) {
return 2;
}
}
}
}
}
else if (AFD >= 12) {
if (SC < 1632) {
return 1;
}
else if (SC >= 1632) {
return 2;
}
}
}
else if (ssDCval >= 66) {
if (negBalance < 57) {
return 1;
}
else if (negBalance >= 57) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
else if (AFD >= 19) {
return 2;
}
}
else if (mu_mv_mag_sq >= 71) {
return 4;
}
}
}
}
else if (SC >= 2676) {
if (ssDCval < 70) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (ssDCval >= 70) {
if (ssDCval < 95) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 95) {
return 2;
}
}
}
}
else if (ssDCval >= 98) {
if (AFD < 11) {
if (Cs < 400) {
if (TSC < 5) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (Rs < 1162) {
return 4;
}
else if (Rs >= 1162) {
return 8;
}
}
}
else if (TSC >= 5) {
if (MVDiff < 32) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (diffAFD < -1) {
if (Rs < 1257) {
return 4;
}
else if (Rs >= 1257) {
return 1;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (MVDiff >= 32) {
return 1;
}
}
}
else if (Cs >= 400) {
if (gchDC < 2) {
if (Rs < 1448) {
return 8;
}
else if (Rs >= 1448) {
return 2;
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (AFD >= 11) {
if (ssDCval < 112) {
if (Cs < 393) {
if (Rs < 1591) {
return 2;
}
else if (Rs >= 1591) {
return 1;
}
}
else if (Cs >= 393) {
if (diffMVdiffVal < -20) {
if (Rs < 2354) {
return 8;
}
else if (Rs >= 2354) {
return 1;
}
}
else if (diffMVdiffVal >= -20) {
return 8;
}
}
}
else if (ssDCval >= 112) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (ssDCval < 122) {
if (tcor < 91) {
if (posBalance < 72) {
if (SC < 2502) {
return 8;
}
else if (SC >= 2502) {
if (Rs < 2280) {
return 1;
}
else if (Rs >= 2280) {
return 8;
}
}
}
else if (posBalance >= 72) {
return 1;
}
}
else if (tcor >= 91) {
return 1;
}
}
else if (ssDCval >= 122) {
return 1;
}
}
}
}
}
}
else if (Cs >= 463) {
if (refDCval < 70) {
if (posBalance < 35) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 2;
}
}
else if (posBalance >= 35) {
if (Rs < 1326) {
if (negBalance < 36) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 2;
}
}
else if (negBalance >= 36) {
if (refDCval < 63) {
if (Rs < 872) {
return 2;
}
else if (Rs >= 872) {
return 8;
}
}
else if (refDCval >= 63) {
if (Rs < 1214) {
return 4;
}
else if (Rs >= 1214) {
return 8;
}
}
}
}
else if (Rs >= 1326) {
if (ssDCval < 56) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 1;
}
}
else if (ssDCval >= 56) {
if (gchDC < 3) {
if (diffMVdiffVal < -5) {
if (Rs < 1441) {
return 2;
}
else if (Rs >= 1441) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
return 2;
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
}
else if (refDCval >= 70) {
if (Rs < 1670) {
if (ssDCval < 90) {
if (mcTcor < 91) {
if (RsDiff < 7) {
return 2;
}
else if (RsDiff >= 7) {
return 4;
}
}
else if (mcTcor >= 91) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 1;
}
}
}
else if (ssDCval >= 90) {
if (negBalance < 44) {
if (Rs < 1257) {
return 1;
}
else if (Rs >= 1257) {
return 2;
}
}
else if (negBalance >= 44) {
if (RsCsDiff < 32) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 1163) {
return 2;
}
else if (Rs >= 1163) {
return 1;
}
}
}
else if (RsCsDiff >= 32) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
}
else if (Rs >= 1670) {
if (negBalance < 53) {
if (diffTSC < -2) {
if (Rs < 1949) {
return 1;
}
else if (Rs >= 1949) {
return 2;
}
}
else if (diffTSC >= -2) {
if (MVDiff < 45) {
if (Rs < 1757) {
if (Rs < 1756) {
return 1;
}
else if (Rs >= 1756) {
return 8;
}
}
else if (Rs >= 1757) {
return 1;
}
}
else if (MVDiff >= 45) {
return 4;
}
}
}
else if (negBalance >= 53) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 87) {
if (refDCval < 105) {
if (MVDiff < 34) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (Rs < 1961) {
if (TSC < 8) {
if (tcor < 78) {
return 1;
}
else if (tcor >= 78) {
if (gchDC < 3) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
if (Cs < 228) {
return 2;
}
else if (Cs >= 228) {
return 1;
}
}
}
else if (gchDC >= 3) {
return 2;
}
}
}
else if (TSC >= 8) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (Cs < 285) {
return 2;
}
else if (Cs >= 285) {
return 1;
}
}
}
}
else if (Rs >= 1961) {
if (negBalance < 40) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
if (Cs < 196) {
return 2;
}
else if (Cs >= 196) {
return 1;
}
}
}
else if (negBalance >= 40) {
if (posBalance < 48) {
if (negBalance < 51) {
if (Cs < 348) {
return 1;
}
else if (Cs >= 348) {
if (Rs < 2363) {
return 4;
}
else if (Rs >= 2363) {
return 2;
}
}
}
else if (negBalance >= 51) {
return 2;
}
}
else if (posBalance >= 48) {
return 2;
}
}
}
}
}
else if (MVDiff >= 34) {
if (TSC < 7) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
return 2;
}
}
}
else if (TSC >= 7) {
if (Cs < 238) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (diffRsCsdiff < -133) {
return 2;
}
else if (diffRsCsdiff >= -133) {
return 1;
}
}
}
else if (Cs >= 238) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
if (RsCsDiff < 182) {
if (diffRsCsdiff < -57) {
return 2;
}
else if (diffRsCsdiff >= -57) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (TSC < 9) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (TSC < 8) {
if (Rs < 976) {
return 2;
}
else if (Rs >= 976) {
return 1;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (TSC >= 9) {
if (RsCsDiff < 141) {
return 1;
}
else if (RsCsDiff >= 141) {
if (refDCval < 79) {
return 2;
}
else if (refDCval >= 79) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 182) {
if (posBalance < 74) {
return 2;
}
else if (posBalance >= 74) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 105) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (mcTcor < 95) {
if (AFD < 21) {
if (Rs < 1890) {
if (gchDC < 3) {
if (Cs < 254) {
return 4;
}
else if (Cs >= 254) {
return 2;
}
}
else if (gchDC >= 3) {
return 8;
}
}
else if (Rs >= 1890) {
return 1;
}
}
else if (AFD >= 21) {
if (AFD < 36) {
if (mcTcor < 91) {
if (Rs < 1954) {
return 1;
}
else if (Rs >= 1954) {
return 8;
}
}
else if (mcTcor >= 91) {
return 1;
}
}
else if (AFD >= 36) {
return 2;
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 88) {
if (Rs < 1105) {
return 8;
}
else if (Rs >= 1105) {
return 2;
}
}
else if (ssDCval >= 88) {
if (AFD < 12) {
if (SC < 2253) {
return 2;
}
else if (SC >= 2253) {
return 1;
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 67) {
if (CsDiff < 3) {
if (refDCval < 110) {
return 2;
}
else if (refDCval >= 110) {
if (SC < 2608) {
return 1;
}
else if (SC >= 2608) {
if (MVDiff < 54) {
if (mu_mv_mag_sq < 31) {
return 1;
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
else if (MVDiff >= 54) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
}
else if (CsDiff >= 3) {
if (SC < 3069) {
if (AFD < 34) {
if (TSC < 7) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (AFD >= 34) {
if (Rs < 1863) {
return 1;
}
else if (Rs >= 1863) {
return 2;
}
}
}
else if (SC >= 3069) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 67) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
if (negBalance < 53) {
if (Cs < 408) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (SC < 1016) {
if (Cs < 46) {
return 2;
}
else if (Cs >= 46) {
return 1;
}
}
else if (SC >= 1016) {
return 2;
}
}
}
else if (Cs >= 408) {
if (ssDCval < 58) {
if (Rs < 1163) {
return 4;
}
else if (Rs >= 1163) {
return 1;
}
}
else if (ssDCval >= 58) {
if (Cs < 519) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (Cs >= 519) {
if (posBalance < 51) {
if (Rs < 2251) {
return 1;
}
else if (Rs >= 2251) {
return 8;
}
}
else if (posBalance >= 51) {
return 2;
}
}
}
}
}
else if (negBalance >= 53) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Cs >= 538) {
if (RsCsDiff < 43) {
if (mcTcor < 96) {
if (AFD < 15) {
if (posBalance < 41) {
if (posBalance < 29) {
if (diffAFD < -16) {
if (Cs < 1219) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (Cs >= 1219) {
return 8;
}
}
else if (diffAFD >= -16) {
if (diffRsCsdiff < 11) {
if (tcor < 82) {
if (posBalance < 20) {
if (diffAFD < 2) {
if (diffMVdiffVal < -5) {
return 1;
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (posBalance >= 20) {
if (tcor < 75) {
return 1;
}
else if (tcor >= 75) {
if (Rs < 1030) {
return 4;
}
else if (Rs >= 1030) {
return 8;
}
}
}
}
else if (tcor >= 82) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 121) {
if (negBalance < 8) {
if (Rs < 1430) {
return 1;
}
else if (Rs >= 1430) {
return 8;
}
}
else if (negBalance >= 8) {
return 8;
}
}
else if (mu_mv_mag_sq >= 121) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 11) {
if (diffAFD < 3) {
if (Rs < 674) {
return 1;
}
else if (Rs >= 674) {
if (Rs < 1939) {
return 8;
}
else if (Rs >= 1939) {
return 4;
}
}
}
else if (diffAFD >= 3) {
return 1;
}
}
}
}
else if (posBalance >= 29) {
if (ssDCval < 47) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (negBalance < 32) {
if (negBalance < 21) {
if (ssDCval < 27) {
return 4;
}
else if (ssDCval >= 27) {
if (Rs < 1890) {
return 1;
}
else if (Rs >= 1890) {
return 4;
}
}
}
else if (negBalance >= 21) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (mcTcor < 80) {
return 8;
}
else if (mcTcor >= 80) {
return 4;
}
}
}
}
else if (negBalance >= 32) {
if (Rs < 1257) {
return 8;
}
else if (Rs >= 1257) {
return 4;
}
}
}
}
else if (ssDCval >= 47) {
if (refDCval < 94) {
if (mu_mv_mag_sq < 117) {
if (TSC < 5) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
if (MVDiff < 16) {
if (Rs < 793) {
if (Cs < 649) {
return 1;
}
else if (Cs >= 649) {
return 2;
}
}
else if (Rs >= 793) {
if (CsDiff < 4) {
if (Rs < 2235) {
if (Cs < 791) {
return 1;
}
else if (Cs >= 791) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
}
else if (Rs >= 2235) {
return 2;
}
}
else if (CsDiff >= 4) {
if (Rs < 1772) {
return 8;
}
else if (Rs >= 1772) {
return 2;
}
}
}
}
else if (MVDiff >= 16) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
else if (TSC >= 5) {
if (Rs < 1243) {
return 8;
}
else if (Rs >= 1243) {
if (RsDiff < 5) {
if (Rs < 1786) {
return 4;
}
else if (Rs >= 1786) {
return 8;
}
}
else if (RsDiff >= 5) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
return 8;
}
}
}
else if (refDCval >= 94) {
if (negBalance < 30) {
if (Cs < 803) {
if (Rs < 520) {
return 2;
}
else if (Rs >= 520) {
return 8;
}
}
else if (Cs >= 803) {
if (mu_mv_mag_sq < 100) {
return 4;
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 1045) {
if (Rs < 732) {
return 4;
}
else if (Rs >= 732) {
return 1;
}
}
else if (Rs >= 1045) {
return 8;
}
}
}
}
else if (negBalance >= 30) {
if (ssDCval < 155) {
if (SC < 1546) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
return 8;
}
}
else if (SC >= 1546) {
return 8;
}
}
else if (ssDCval >= 155) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 41) {
if (Rs < 421) {
if (ssDCval < 97) {
if (ssDCval < 82) {
if (mcTcor < 83) {
if (Cs < 570) {
return 2;
}
else if (Cs >= 570) {
if (Rs < 264) {
return 8;
}
else if (Rs >= 264) {
return 4;
}
}
}
else if (mcTcor >= 83) {
if (ssDCval < 61) {
if (mu_mv_mag_sq < 3) {
if (Rs < 346) {
return 2;
}
else if (Rs >= 346) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
return 1;
}
}
else if (ssDCval >= 61) {
if (AFD < 7) {
if (Cs < 657) {
return 4;
}
else if (Cs >= 657) {
return 8;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (ssDCval >= 82) {
if (MVDiff < 6) {
if (Cs < 834) {
return 1;
}
else if (Cs >= 834) {
return 8;
}
}
else if (MVDiff >= 6) {
if (Rs < 305) {
return 1;
}
else if (Rs >= 305) {
return 8;
}
}
}
}
else if (ssDCval >= 97) {
if (diffTSC < 1) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (diffTSC >= 1) {
if (Rs < 354) {
return 2;
}
else if (Rs >= 354) {
return 1;
}
}
}
}
else if (Rs >= 421) {
if (ssDCval < 157) {
if (RsDiff < 3) {
if (Rs < 816) {
if (ssDCval < 62) {
if (negBalance < 32) {
if (negBalance < 26) {
if (tcor < 87) {
if (Cs < 753) {
return 4;
}
else if (Cs >= 753) {
return 1;
}
}
else if (tcor >= 87) {
return 1;
}
}
else if (negBalance >= 26) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (diffAFD < 1) {
if (Cs < 557) {
return 1;
}
else if (Cs >= 557) {
return 4;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
else if (negBalance >= 32) {
if (MVDiff < 5) {
if (Rs < 702) {
if (Rs < 454) {
return 1;
}
else if (Rs >= 454) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 4;
}
}
}
else if (Rs >= 702) {
if (ssDCval < 49) {
if (Rs < 813) {
return 1;
}
else if (Rs >= 813) {
return 4;
}
}
else if (ssDCval >= 49) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
else if (MVDiff >= 5) {
if (Rs < 516) {
if (mu_mv_mag_sq < 41) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 41) {
if (Rs < 445) {
return 8;
}
else if (Rs >= 445) {
return 1;
}
}
}
else if (Rs >= 516) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (Rs < 775) {
if (diffMVdiffVal < 8) {
if (TSC < 6) {
if (MVDiff < 7) {
if (Rs < 688) {
return 8;
}
else if (Rs >= 688) {
return 2;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (diffMVdiffVal >= 8) {
return 2;
}
}
else if (Rs >= 775) {
if (Rs < 791) {
return 1;
}
else if (Rs >= 791) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 62) {
if (MVDiff < 7) {
if (negBalance < 45) {
if (mu_mv_mag_sq < 55) {
return 2;
}
else if (mu_mv_mag_sq >= 55) {
if (Cs < 1042) {
if (TSC < 5) {
if (Cs < 650) {
return 8;
}
else if (Cs >= 650) {
if (tcor < 95) {
if (Cs < 679) {
if (Rs < 539) {
return 2;
}
else if (Rs >= 539) {
return 8;
}
}
else if (Cs >= 679) {
return 2;
}
}
else if (tcor >= 95) {
return 8;
}
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (Cs >= 1042) {
if (ssDCval < 119) {
if (Rs < 695) {
if (Rs < 581) {
return 1;
}
else if (Rs >= 581) {
return 8;
}
}
else if (Rs >= 695) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 4;
}
}
}
else if (ssDCval >= 119) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (refDCval < 89) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 679) {
return 1;
}
else if (Rs >= 679) {
return 8;
}
}
}
else if (refDCval >= 89) {
if (MVDiff < 3) {
if (Cs < 883) {
if (refDCval < 111) {
return 1;
}
else if (refDCval >= 111) {
return 2;
}
}
else if (Cs >= 883) {
return 8;
}
}
else if (MVDiff >= 3) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (Rs < 689) {
return 8;
}
else if (Rs >= 689) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (posBalance < 45) {
if (Cs < 938) {
return 2;
}
else if (Cs >= 938) {
return 8;
}
}
else if (posBalance >= 45) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 816) {
if (refDCval < 63) {
if (CsDiff < 4) {
if (CsDiff < 2) {
if (diffRsCsdiff < -5) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
return 8;
}
}
else if (CsDiff >= 2) {
if (MVDiff < 3) {
if (tcor < 88) {
if (Rs < 926) {
return 4;
}
else if (Rs >= 926) {
if (Rs < 1077) {
return 8;
}
else if (Rs >= 1077) {
return 1;
}
}
}
else if (tcor >= 88) {
if (RsCsDiff < 7) {
if (Rs < 968) {
return 8;
}
else if (Rs >= 968) {
return 1;
}
}
else if (RsCsDiff >= 7) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 54) {
if (posBalance < 44) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (posBalance >= 44) {
return 8;
}
}
else if (mu_mv_mag_sq >= 54) {
if (negBalance < 32) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
if (Rs < 1059) {
return 4;
}
else if (Rs >= 1059) {
return 8;
}
}
}
else if (negBalance >= 32) {
if (mu_mv_mag_sq < 107) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
return 1;
}
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 1321) {
return 1;
}
else if (Rs >= 1321) {
return 8;
}
}
}
}
}
}
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (refDCval >= 63) {
if (Rs < 1123) {
if (Rs < 904) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 853) {
return 2;
}
else if (Rs >= 853) {
return 8;
}
}
}
}
else if (Rs >= 904) {
if (MVDiff < 19) {
if (Cs < 1185) {
if (Cs < 677) {
return 4;
}
else if (Cs >= 677) {
if (Rs < 911) {
if (Cs < 982) {
return 8;
}
else if (Cs >= 982) {
return 2;
}
}
else if (Rs >= 911) {
return 8;
}
}
}
else if (Cs >= 1185) {
return 4;
}
}
else if (MVDiff >= 19) {
if (Rs < 1112) {
return 4;
}
else if (Rs >= 1112) {
return 8;
}
}
}
}
else if (Rs >= 1123) {
if (ssDCval < 135) {
if (tcor < 96) {
if (diffRsCsdiff < -9) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (diffRsCsdiff >= -9) {
return 8;
}
}
else if (tcor >= 96) {
return 1;
}
}
else if (ssDCval >= 135) {
if (Rs < 1905) {
return 1;
}
else if (Rs >= 1905) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 3) {
if (negBalance < 44) {
if (Rs < 828) {
if (mu_mv_mag_sq < 66) {
if (negBalance < 36) {
if (Rs < 648) {
if (ssDCval < 72) {
if (MVDiff < 17) {
if (Rs < 561) {
return 4;
}
else if (Rs >= 561) {
return 8;
}
}
else if (MVDiff >= 17) {
if (Cs < 697) {
return 1;
}
else if (Cs >= 697) {
if (Rs < 542) {
return 8;
}
else if (Rs >= 542) {
return 2;
}
}
}
}
else if (ssDCval >= 72) {
return 2;
}
}
else if (Rs >= 648) {
if (mcTcor < 91) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
if (negBalance < 28) {
return 2;
}
else if (negBalance >= 28) {
return 8;
}
}
}
else if (mcTcor >= 91) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
else if (negBalance >= 36) {
if (mu_mv_mag_sq < 48) {
if (AFD < 12) {
if (posBalance < 72) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
if (posBalance < 65) {
return 2;
}
else if (posBalance >= 65) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 2;
}
}
}
}
else if (posBalance >= 72) {
return 8;
}
}
else if (AFD >= 12) {
if (Rs < 612) {
return 4;
}
else if (Rs >= 612) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (Cs < 891) {
return 4;
}
else if (Cs >= 891) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (MVDiff < 8) {
if (Rs < 816) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
else if (Rs >= 816) {
return 1;
}
}
else if (MVDiff >= 8) {
if (Rs < 792) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 4;
}
}
else if (Rs >= 792) {
return 2;
}
}
}
}
else if (Rs >= 828) {
if (refDCval < 44) {
if (Rs < 1548) {
if (Cs < 668) {
if (refDCval < 39) {
if (Rs < 1224) {
return 8;
}
else if (Rs >= 1224) {
return 4;
}
}
else if (refDCval >= 39) {
if (Cs < 618) {
return 8;
}
else if (Cs >= 618) {
return 2;
}
}
}
else if (Cs >= 668) {
return 8;
}
}
else if (Rs >= 1548) {
return 4;
}
}
else if (refDCval >= 44) {
if (SC < 2489) {
if (posBalance < 55) {
if (AFD < 8) {
if (ssDCval < 47) {
if (Rs < 1071) {
return 8;
}
else if (Rs >= 1071) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (ssDCval >= 47) {
if (Cs < 718) {
if (MVDiff < 5) {
if (tcor < 93) {
if (posBalance < 43) {
if (Rs < 912) {
return 2;
}
else if (Rs >= 912) {
return 4;
}
}
else if (posBalance >= 43) {
return 4;
}
}
else if (tcor >= 93) {
if (Rs < 1042) {
return 8;
}
else if (Rs >= 1042) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (Cs < 580) {
if (Rs < 1236) {
return 8;
}
else if (Rs >= 1236) {
if (Rs < 1325) {
return 1;
}
else if (Rs >= 1325) {
return 2;
}
}
}
else if (Cs >= 580) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 8;
}
}
}
}
else if (Cs >= 718) {
if (ssDCval < 79) {
if (RsCsDiff < 23) {
if (mcTcor < 95) {
if (TSC < 5) {
if (negBalance < 28) {
if (Rs < 1242) {
return 4;
}
else if (Rs >= 1242) {
if (Rs < 1543) {
return 8;
}
else if (Rs >= 1543) {
return 8;
}
}
}
else if (negBalance >= 28) {
return 8;
}
}
else if (TSC >= 5) {
if (Cs < 860) {
return 4;
}
else if (Cs >= 860) {
if (SC < 2205) {
return 8;
}
else if (SC >= 2205) {
if (Rs < 1487) {
return 1;
}
else if (Rs >= 1487) {
return 8;
}
}
}
}
}
else if (mcTcor >= 95) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (tcor < 97) {
if (Rs < 1453) {
return 4;
}
else if (Rs >= 1453) {
return 1;
}
}
else if (tcor >= 97) {
return 8;
}
}
}
}
else if (RsCsDiff >= 23) {
if (mcTcor < 92) {
if (Cs < 911) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
return 8;
}
}
else if (Cs >= 911) {
return 8;
}
}
else if (mcTcor >= 92) {
if (tcor < 92) {
if (diffMVdiffVal < 1) {
if (Rs < 1462) {
return 2;
}
else if (Rs >= 1462) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (tcor >= 92) {
return 4;
}
}
}
}
else if (ssDCval >= 79) {
if (Rs < 835) {
return 4;
}
else if (Rs >= 835) {
return 8;
}
}
}
}
}
else if (AFD >= 8) {
if (ssDCval < 63) {
if (Cs < 843) {
if (Rs < 931) {
return 1;
}
else if (Rs >= 931) {
return 2;
}
}
else if (Cs >= 843) {
if (Rs < 923) {
return 4;
}
else if (Rs >= 923) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 1206) {
return 8;
}
else if (Rs >= 1206) {
return 2;
}
}
}
}
}
else if (ssDCval >= 63) {
if (TSC < 7) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (diffMVdiffVal < 8) {
return 1;
}
else if (diffMVdiffVal >= 8) {
return 8;
}
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 53) {
if (mu_mv_mag_sq < 111) {
if (Cs < 1045) {
if (Rs < 1341) {
if (TSC < 5) {
if (ssDCval < 46) {
return 2;
}
else if (ssDCval >= 46) {
return 8;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (Rs >= 1341) {
if (Rs < 1411) {
return 2;
}
else if (Rs >= 1411) {
return 8;
}
}
}
else if (Cs >= 1045) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 920) {
if (Rs < 1115) {
return 8;
}
else if (Rs >= 1115) {
return 1;
}
}
else if (Cs >= 920) {
return 8;
}
}
}
else if (ssDCval >= 53) {
if (ssDCval < 72) {
if (Cs < 1052) {
if (refDCval < 64) {
if (mu_mv_mag_sq < 51) {
if (Cs < 664) {
if (Rs < 1079) {
return 2;
}
else if (Rs >= 1079) {
return 1;
}
}
else if (Cs >= 664) {
if (negBalance < 33) {
return 2;
}
else if (negBalance >= 33) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 51) {
if (negBalance < 35) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
else if (negBalance >= 35) {
return 4;
}
}
}
else if (refDCval >= 64) {
if (mu_mv_mag_sq < 65) {
if (diffRsCsdiff < 8) {
return 1;
}
else if (diffRsCsdiff >= 8) {
if (Cs < 828) {
return 1;
}
else if (Cs >= 828) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 65) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 1427) {
return 4;
}
else if (Rs >= 1427) {
return 1;
}
}
}
}
}
else if (Cs >= 1052) {
if (negBalance < 36) {
if (Rs < 994) {
return 8;
}
else if (Rs >= 994) {
if (Rs < 1097) {
return 1;
}
else if (Rs >= 1097) {
return 4;
}
}
}
else if (negBalance >= 36) {
if (SC < 2429) {
return 4;
}
else if (SC >= 2429) {
if (SC < 2445) {
return 1;
}
else if (SC >= 2445) {
return 4;
}
}
}
}
}
else if (ssDCval >= 72) {
if (ssDCval < 130) {
if (mu_mv_mag_sq < 3) {
if (Cs < 977) {
if (ssDCval < 81) {
return 4;
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (Cs >= 977) {
if (Rs < 946) {
return 4;
}
else if (Rs >= 946) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffAFD < -2) {
if (Rs < 1179) {
return 1;
}
else if (Rs >= 1179) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (ssDCval >= 130) {
if (AFD < 10) {
if (diffMVdiffVal < -3) {
if (Rs < 1168) {
return 4;
}
else if (Rs >= 1168) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
return 4;
}
}
else if (AFD >= 10) {
if (Rs < 1134) {
if (Rs < 898) {
return 1;
}
else if (Rs >= 898) {
return 2;
}
}
else if (Rs >= 1134) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 2489) {
if (ssDCval < 75) {
if (MVDiff < 6) {
if (ssDCval < 64) {
if (AFD < 6) {
if (Cs < 1044) {
if (Cs < 1018) {
if (Rs < 1652) {
return 8;
}
else if (Rs >= 1652) {
if (Rs < 1868) {
return 8;
}
else if (Rs >= 1868) {
return 4;
}
}
}
else if (Cs >= 1018) {
return 2;
}
}
else if (Cs >= 1044) {
return 8;
}
}
else if (AFD >= 6) {
if (posBalance < 45) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (posBalance >= 45) {
if (refDCval < 62) {
if (Rs < 1317) {
return 4;
}
else if (Rs >= 1317) {
if (Cs < 743) {
return 1;
}
else if (Cs >= 743) {
return 8;
}
}
}
else if (refDCval >= 62) {
if (Rs < 1860) {
return 1;
}
else if (Rs >= 1860) {
return 4;
}
}
}
}
}
else if (ssDCval >= 64) {
if (tcor < 94) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
if (Rs < 1548) {
return 8;
}
else if (Rs >= 1548) {
return 2;
}
}
}
else if (tcor >= 94) {
if (Rs < 1336) {
return 1;
}
else if (Rs >= 1336) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
if (Cs < 1124) {
if (refDCval < 68) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (refDCval >= 68) {
if (Cs < 959) {
return 2;
}
else if (Cs >= 959) {
return 4;
}
}
}
else if (Cs >= 1124) {
if (ssDCval < 60) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
return 8;
}
}
else if (ssDCval >= 60) {
return 1;
}
}
}
}
else if (ssDCval >= 75) {
if (refDCval < 78) {
if (Rs < 1586) {
return 8;
}
else if (Rs >= 1586) {
return 1;
}
}
else if (refDCval >= 78) {
if (Rs < 1712) {
if (mu_mv_mag_sq < 57) {
if (SC < 2563) {
if (Rs < 1531) {
return 4;
}
else if (Rs >= 1531) {
return 8;
}
}
else if (SC >= 2563) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (diffMVdiffVal < -3) {
if (Rs < 1573) {
return 8;
}
else if (Rs >= 1573) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
else if (Rs >= 1712) {
if (ssDCval < 93) {
if (refDCval < 92) {
if (Cs < 697) {
return 1;
}
else if (Cs >= 697) {
if (SC < 2980) {
return 8;
}
else if (SC >= 2980) {
if (Rs < 1863) {
return 1;
}
else if (Rs >= 1863) {
return 8;
}
}
}
}
else if (refDCval >= 92) {
if (Rs < 2108) {
return 2;
}
else if (Rs >= 2108) {
return 4;
}
}
}
else if (ssDCval >= 93) {
if (mu_mv_mag_sq < 126) {
if (SC < 3642) {
if (mu_mv_mag_sq < 5) {
if (Rs < 1870) {
return 2;
}
else if (Rs >= 1870) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
return 8;
}
}
else if (SC >= 3642) {
if (Rs < 2408) {
return 1;
}
else if (Rs >= 2408) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 126) {
return 1;
}
}
}
}
}
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 1146) {
if (MVDiff < 18) {
if (ssDCval < 71) {
if (ssDCval < 52) {
if (AFD < 12) {
if (TSC < 7) {
if (negBalance < 52) {
if (SC < 2051) {
if (Rs < 817) {
if (mu_mv_mag_sq < 2) {
if (mcTcor < 83) {
if (Rs < 777) {
return 8;
}
else if (Rs >= 777) {
return 1;
}
}
else if (mcTcor >= 83) {
if (Cs < 575) {
return 1;
}
else if (Cs >= 575) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (AFD < 11) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (Rs >= 817) {
if (RsCsDiff < 33) {
if (RsDiff < 5) {
if (Rs < 894) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (Rs < 885) {
return 8;
}
else if (Rs >= 885) {
return 1;
}
}
}
else if (Rs >= 894) {
return 8;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (RsCsDiff >= 33) {
if (mu_mv_mag_sq < 3) {
return 8;
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 1730) {
return 2;
}
else if (SC >= 1730) {
return 1;
}
}
}
}
}
else if (SC >= 2051) {
return 4;
}
}
else if (negBalance >= 52) {
if (posBalance < 58) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
return 4;
}
}
else if (posBalance >= 58) {
if (mcTcor < 79) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Cs < 600) {
return 1;
}
else if (Cs >= 600) {
return 8;
}
}
}
else if (mcTcor >= 79) {
return 1;
}
}
}
}
else if (TSC >= 7) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (SC < 1660) {
if (Rs < 816) {
return 8;
}
else if (Rs >= 816) {
if (Rs < 944) {
return 1;
}
else if (Rs >= 944) {
return 8;
}
}
}
else if (SC >= 1660) {
return 8;
}
}
}
}
else if (AFD >= 12) {
if (refDCval < 52) {
return 4;
}
else if (refDCval >= 52) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (mcTcor < 82) {
return 4;
}
else if (mcTcor >= 82) {
if (ssDCval < 61) {
if (negBalance < 53) {
if (Cs < 940) {
if (Rs < 903) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
return 8;
}
}
}
else if (Rs >= 903) {
return 8;
}
}
else if (Cs >= 940) {
return 4;
}
}
else if (negBalance >= 53) {
if (refDCval < 61) {
return 8;
}
else if (refDCval >= 61) {
return 4;
}
}
}
else if (ssDCval >= 61) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (AFD < 10) {
if (diffAFD < 2) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 4;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
else if (AFD >= 10) {
if (Cs < 819) {
return 8;
}
else if (Cs >= 819) {
if (Rs < 1101) {
return 1;
}
else if (Rs >= 1101) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (posBalance < 52) {
if (Rs < 1088) {
if (MVDiff < 6) {
if (Rs < 1017) {
return 2;
}
else if (Rs >= 1017) {
return 8;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (Rs >= 1088) {
return 4;
}
}
else if (posBalance >= 52) {
if (mu_mv_mag_sq < 2) {
if (TSC < 7) {
if (Cs < 575) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (Rs < 841) {
return 1;
}
else if (Rs >= 841) {
return 8;
}
}
}
else if (Cs >= 575) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (TSC >= 7) {
if (MVDiff < 13) {
if (RsCsDiff < 38) {
return 8;
}
else if (RsCsDiff >= 38) {
if (Rs < 1031) {
return 2;
}
else if (Rs >= 1031) {
return 4;
}
}
}
else if (MVDiff >= 13) {
if (Rs < 824) {
return 8;
}
else if (Rs >= 824) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (Cs < 542) {
if (Rs < 798) {
return 2;
}
else if (Rs >= 798) {
return 8;
}
}
else if (Cs >= 542) {
if (Rs < 1107) {
return 8;
}
else if (Rs >= 1107) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Cs < 736) {
return 8;
}
else if (Cs >= 736) {
return 1;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 18) {
if (refDCval < 114) {
if (diffMVdiffVal < -3) {
if (Cs < 1016) {
if (Cs < 572) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (Cs >= 572) {
if (MVDiff < 19) {
if (Rs < 757) {
return 1;
}
else if (Rs >= 757) {
return 8;
}
}
else if (MVDiff >= 19) {
if (tcor < 58) {
return 4;
}
else if (tcor >= 58) {
if (tcor < 69) {
return 8;
}
else if (tcor >= 69) {
if (diffAFD < 1) {
if (posBalance < 53) {
if (tcor < 75) {
return 4;
}
else if (tcor >= 75) {
return 8;
}
}
else if (posBalance >= 53) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 2;
}
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1016) {
if (Rs < 884) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 2;
}
}
else if (Rs >= 884) {
return 8;
}
}
}
else if (diffMVdiffVal >= -3) {
if (AFD < 11) {
if (Rs < 672) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (Rs >= 672) {
if (negBalance < 45) {
if (Rs < 929) {
return 2;
}
else if (Rs >= 929) {
return 8;
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (AFD >= 11) {
if (mu_mv_mag_sq < 25) {
if (tcor < 69) {
return 8;
}
else if (tcor >= 69) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (mcTcor < 89) {
return 1;
}
else if (mcTcor >= 89) {
if (AFD < 13) {
if (Rs < 724) {
return 1;
}
else if (Rs >= 724) {
return 4;
}
}
else if (AFD >= 13) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (Cs < 742) {
return 4;
}
else if (Cs >= 742) {
return 1;
}
}
}
}
}
else if (refDCval >= 114) {
if (ssDCval < 143) {
return 4;
}
else if (ssDCval >= 143) {
if (Cs < 917) {
return 2;
}
else if (Cs >= 917) {
return 1;
}
}
}
}
}
else if (Rs >= 1146) {
if (Rs < 1586) {
if (mcTcor < 91) {
if (ssDCval < 97) {
if (Cs < 1111) {
if (Cs < 543) {
return 2;
}
else if (Cs >= 543) {
return 8;
}
}
else if (Cs >= 1111) {
if (ssDCval < 62) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (Cs < 1205) {
if (Rs < 1401) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
return 4;
}
}
else if (Rs >= 1401) {
return 8;
}
}
else if (Cs >= 1205) {
return 1;
}
}
}
else if (ssDCval >= 62) {
return 8;
}
}
}
else if (ssDCval >= 97) {
if (tcor < 74) {
if (ssDCval < 120) {
if (MVDiff < 22) {
if (Cs < 962) {
return 2;
}
else if (Cs >= 962) {
return 8;
}
}
else if (MVDiff >= 22) {
return 8;
}
}
else if (ssDCval >= 120) {
if (Rs < 1227) {
return 1;
}
else if (Rs >= 1227) {
return 4;
}
}
}
else if (tcor >= 74) {
if (ssDCval < 98) {
if (TSC < 7) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (ssDCval >= 98) {
if (mu_mv_mag_sq < 114) {
return 8;
}
else if (mu_mv_mag_sq >= 114) {
if (Rs < 1289) {
return 8;
}
else if (Rs >= 1289) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (MVDiff < 4) {
if (AFD < 6) {
if (Rs < 1339) {
return 1;
}
else if (Rs >= 1339) {
return 8;
}
}
else if (AFD >= 6) {
if (refDCval < 79) {
if (TSC < 6) {
if (Cs < 1030) {
return 8;
}
else if (Cs >= 1030) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (refDCval >= 79) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
}
}
}
else if (MVDiff >= 4) {
if (AFD < 9) {
if (Cs < 888) {
if (ssDCval < 71) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (ssDCval >= 71) {
if (ssDCval < 136) {
return 8;
}
else if (ssDCval >= 136) {
return 4;
}
}
}
else if (Cs >= 888) {
if (Rs < 1310) {
if (TSC < 6) {
if (Cs < 948) {
return 4;
}
else if (Cs >= 948) {
return 8;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (Rs >= 1310) {
return 4;
}
}
}
else if (AFD >= 9) {
if (MVDiff < 36) {
if (tcor < 85) {
if (tcor < 84) {
if (Cs < 657) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 8;
}
}
else if (Cs >= 657) {
return 8;
}
}
else if (tcor >= 84) {
return 4;
}
}
else if (tcor >= 85) {
if (negBalance < 47) {
if (Rs < 1469) {
return 1;
}
else if (Rs >= 1469) {
return 8;
}
}
else if (negBalance >= 47) {
if (Cs < 1205) {
return 8;
}
else if (Cs >= 1205) {
return 1;
}
}
}
}
else if (MVDiff >= 36) {
if (ssDCval < 69) {
if (MVDiff < 47) {
return 4;
}
else if (MVDiff >= 47) {
return 8;
}
}
else if (ssDCval >= 69) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1586) {
if (ssDCval < 99) {
if (posBalance < 59) {
if (ssDCval < 87) {
if (mu_mv_mag_sq < 85) {
if (SC < 3048) {
if (mu_mv_mag_sq < 17) {
if (ssDCval < 66) {
if (AFD < 10) {
if (Cs < 918) {
return 4;
}
else if (Cs >= 918) {
return 8;
}
}
else if (AFD >= 10) {
return 1;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (mu_mv_mag_sq >= 17) {
return 8;
}
}
else if (SC >= 3048) {
if (TSC < 8) {
if (Cs < 1174) {
return 8;
}
else if (Cs >= 1174) {
return 1;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (ssDCval < 68) {
if (refDCval < 62) {
return 8;
}
else if (refDCval >= 62) {
return 2;
}
}
else if (ssDCval >= 68) {
if (Rs < 1837) {
return 8;
}
else if (Rs >= 1837) {
if (Rs < 2090) {
return 1;
}
else if (Rs >= 2090) {
return 8;
}
}
}
}
}
else if (ssDCval >= 87) {
if (RsCsDiff < 23) {
if (Rs < 1963) {
return 8;
}
else if (Rs >= 1963) {
return 2;
}
}
else if (RsCsDiff >= 23) {
if (Cs < 1301) {
if (tcor < 76) {
if (Cs < 1166) {
return 2;
}
else if (Cs >= 1166) {
return 8;
}
}
else if (tcor >= 76) {
return 2;
}
}
else if (Cs >= 1301) {
return 8;
}
}
}
}
else if (posBalance >= 59) {
if (diffMVdiffVal < -10) {
if (Rs < 1659) {
return 2;
}
else if (Rs >= 1659) {
return 8;
}
}
else if (diffMVdiffVal >= -10) {
return 8;
}
}
}
else if (ssDCval >= 99) {
if (mu_mv_mag_sq < 109) {
if (MVDiff < 27) {
return 8;
}
else if (MVDiff >= 27) {
return 4;
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 1833) {
return 1;
}
else if (Rs >= 1833) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 157) {
if (Rs < 832) {
return 8;
}
else if (Rs >= 832) {
if (Rs < 1948) {
return 2;
}
else if (Rs >= 1948) {
return 8;
}
}
}
}
}
}
else if (AFD >= 15) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 88) {
if (SC < 1690) {
if (negBalance < 56) {
if (AFD < 21) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (diffRsCsdiff < -18) {
if (Rs < 873) {
return 1;
}
else if (Rs >= 873) {
return 4;
}
}
else if (diffRsCsdiff >= -18) {
return 4;
}
}
}
else if (AFD >= 21) {
return 1;
}
}
else if (negBalance >= 56) {
if (diffAFD < -1) {
if (Rs < 620) {
return 4;
}
else if (Rs >= 620) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (SC >= 1690) {
if (Rs < 822) {
return 2;
}
else if (Rs >= 822) {
if (MVDiff < 53) {
return 8;
}
else if (MVDiff >= 53) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
}
else if (ssDCval >= 88) {
if (Cs < 827) {
if (AFD < 22) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
else if (AFD >= 22) {
return 1;
}
}
else if (Cs >= 827) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (SC < 1380) {
if (posBalance < 73) {
if (ssDCval < 96) {
if (mcTcor < 84) {
if (diffMVdiffVal < -8) {
if (Rs < 284) {
return 4;
}
else if (Rs >= 284) {
return 8;
}
}
else if (diffMVdiffVal >= -8) {
return 4;
}
}
else if (mcTcor >= 84) {
if (gchDC < 2) {
if (Rs < 321) {
return 2;
}
else if (Rs >= 321) {
if (MVDiff < 35) {
return 8;
}
else if (MVDiff >= 35) {
if (Rs < 552) {
return 2;
}
else if (Rs >= 552) {
if (Rs < 680) {
return 4;
}
else if (Rs >= 680) {
return 8;
}
}
}
}
}
else if (gchDC >= 2) {
if (RsDiff < 4) {
if (mu_mv_mag_sq < 10) {
return 4;
}
else if (mu_mv_mag_sq >= 10) {
if (Rs < 505) {
return 1;
}
else if (Rs >= 505) {
return 2;
}
}
}
else if (RsDiff >= 4) {
if (TSC < 7) {
if (Rs < 542) {
return 1;
}
else if (Rs >= 542) {
return 8;
}
}
else if (TSC >= 7) {
if (Rs < 450) {
return 2;
}
else if (Rs >= 450) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 96) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
if (MVDiff < 159) {
if (refDCval < 99) {
if (Rs < 441) {
return 2;
}
else if (Rs >= 441) {
return 8;
}
}
else if (refDCval >= 99) {
return 2;
}
}
else if (MVDiff >= 159) {
if (Rs < 427) {
return 1;
}
else if (Rs >= 427) {
return 8;
}
}
}
}
}
else if (posBalance >= 73) {
return 4;
}
}
else if (SC >= 1380) {
if (mu_mv_mag_sq < 7) {
if (posBalance < 64) {
if (Rs < 730) {
return 2;
}
else if (Rs >= 730) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 1031) {
return 4;
}
else if (Rs >= 1031) {
return 8;
}
}
}
}
else if (posBalance >= 64) {
return 4;
}
}
else if (mu_mv_mag_sq >= 7) {
if (tcor < 75) {
if (posBalance < 57) {
if (ssDCval < 141) {
return 1;
}
else if (ssDCval >= 141) {
return 8;
}
}
else if (posBalance >= 57) {
if (AFD < 23) {
if (Cs < 1238) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
if (Rs < 1158) {
return 4;
}
else if (Rs >= 1158) {
return 8;
}
}
}
else if (Cs >= 1238) {
return 1;
}
}
else if (AFD >= 23) {
return 1;
}
}
}
else if (tcor >= 75) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (negBalance < 13) {
if (Rs < 702) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 387) {
return 1;
}
else if (Rs >= 387) {
return 8;
}
}
}
else if (ssDCval >= 56) {
if (Cs < 693) {
if (RsCsDiff < 1) {
if (Rs < 528) {
return 1;
}
else if (Rs >= 528) {
return 8;
}
}
else if (RsCsDiff >= 1) {
if (Cs < 632) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Cs >= 632) {
return 8;
}
}
}
else if (Cs >= 693) {
if (posBalance < 36) {
if (Rs < 416) {
return 4;
}
else if (Rs >= 416) {
return 1;
}
}
else if (posBalance >= 36) {
if (SC < 1173) {
return 4;
}
else if (SC >= 1173) {
if (Cs < 796) {
if (Rs < 565) {
return 8;
}
else if (Rs >= 565) {
if (Rs < 650) {
return 2;
}
else if (Rs >= 650) {
return 1;
}
}
}
else if (Cs >= 796) {
return 8;
}
}
}
}
}
}
else if (Rs >= 702) {
if (MVDiff < 6) {
if (Rs < 1841) {
if (Rs < 967) {
if (diffMVdiffVal < -1) {
if (negBalance < 3) {
return 2;
}
else if (negBalance >= 3) {
if (ssDCval < 43) {
if (Rs < 737) {
return 8;
}
else if (Rs >= 737) {
return 2;
}
}
else if (ssDCval >= 43) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 11) {
if (ssDCval < 136) {
return 8;
}
else if (ssDCval >= 136) {
return 2;
}
}
else if (posBalance >= 11) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (Rs < 745) {
if (posBalance < 27) {
if (posBalance < 26) {
return 1;
}
else if (posBalance >= 26) {
if (Rs < 732) {
return 1;
}
else if (Rs >= 732) {
return 8;
}
}
}
else if (posBalance >= 27) {
return 8;
}
}
else if (Rs >= 745) {
if (ssDCval < 73) {
if (Rs < 882) {
if (Cs < 927) {
return 2;
}
else if (Cs >= 927) {
return 1;
}
}
else if (Rs >= 882) {
return 8;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
}
}
}
}
else if (Rs >= 967) {
if (ssDCval < 53) {
if (TSCindex < 2) {
if (Cs < 549) {
return 1;
}
else if (Cs >= 549) {
return 8;
}
}
else if (TSCindex >= 2) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
}
else if (ssDCval >= 53) {
if (mu_mv_mag_sq < 127) {
if (Cs < 1042) {
if (SC < 1872) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (Cs < 771) {
return 8;
}
else if (Cs >= 771) {
return 2;
}
}
}
else if (SC >= 1872) {
if (diffMVdiffVal < 3) {
if (Cs < 1008) {
return 8;
}
else if (Cs >= 1008) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
}
else if (Cs >= 1042) {
if (Cs < 1241) {
return 1;
}
else if (Cs >= 1241) {
if (Rs < 1506) {
return 1;
}
else if (Rs >= 1506) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (diffMVdiffVal < -4) {
if (ssDCval < 135) {
return 1;
}
else if (ssDCval >= 135) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (diffMVdiffVal >= -4) {
if (ssDCval < 62) {
if (SC < 2268) {
return 8;
}
else if (SC >= 2268) {
return 1;
}
}
else if (ssDCval >= 62) {
if (SC < 2856) {
if (Cs < 679) {
if (diffRsCsdiff < -1) {
if (Rs < 1077) {
return 8;
}
else if (Rs >= 1077) {
return 4;
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
else if (Cs >= 679) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
}
else if (SC >= 2856) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 1841) {
if (Rs < 2054) {
if (Cs < 996) {
if (ssDCval < 81) {
if (Cs < 754) {
return 8;
}
else if (Cs >= 754) {
if (SC < 2952) {
return 1;
}
else if (SC >= 2952) {
if (Rs < 2008) {
return 1;
}
else if (Rs >= 2008) {
return 4;
}
}
}
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (Cs >= 996) {
if (ssDCval < 69) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 4;
}
}
else if (ssDCval >= 69) {
if (Rs < 1907) {
return 1;
}
else if (Rs >= 1907) {
return 8;
}
}
}
}
else if (Rs >= 2054) {
if (Cs < 1289) {
if (posBalance < 42) {
if (Cs < 681) {
return 4;
}
else if (Cs >= 681) {
return 8;
}
}
else if (posBalance >= 42) {
if (Cs < 1037) {
return 1;
}
else if (Cs >= 1037) {
return 8;
}
}
}
else if (Cs >= 1289) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 84) {
if (Rs < 1465) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (Rs < 1094) {
if (Rs < 780) {
return 2;
}
else if (Rs >= 780) {
return 8;
}
}
else if (Rs >= 1094) {
if (Cs < 1117) {
return 1;
}
else if (Cs >= 1117) {
return 8;
}
}
}
}
else if (Rs >= 1465) {
return 8;
}
}
else if (ssDCval >= 84) {
if (ssDCval < 93) {
return 4;
}
else if (ssDCval >= 93) {
if (ssDCval < 134) {
if (diffTSC < -5) {
if (diffMVdiffVal < -42) {
return 4;
}
else if (diffMVdiffVal >= -42) {
if (ssDCval < 111) {
return 8;
}
else if (ssDCval >= 111) {
if (diffTSC < -8) {
if (Rs < 1546) {
return 8;
}
else if (Rs >= 1546) {
return 1;
}
}
else if (diffTSC >= -8) {
return 1;
}
}
}
}
else if (diffTSC >= -5) {
if (diffRsCsdiff < -12) {
if (Rs < 1200) {
return 2;
}
else if (Rs >= 1200) {
return 8;
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
}
else if (ssDCval >= 134) {
if (diffRsCsdiff < -47) {
if (diffMVdiffVal < -27) {
return 4;
}
else if (diffMVdiffVal >= -27) {
if (diffAFD < -28) {
return 4;
}
else if (diffAFD >= -28) {
if (Rs < 795) {
return 1;
}
else if (Rs >= 795) {
if (Rs < 1461) {
return 8;
}
else if (Rs >= 1461) {
if (Rs < 1527) {
return 4;
}
else if (Rs >= 1527) {
return 8;
}
}
}
}
}
}
else if (diffRsCsdiff >= -47) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (refDCval < 88) {
if (Rs < 1745) {
if (Rs < 1002) {
if (refDCval < 64) {
if (AFD < 1) {
if (Rs < 847) {
return 1;
}
else if (Rs >= 847) {
return 8;
}
}
else if (AFD >= 1) {
if (Rs < 832) {
if (Rs < 532) {
if (negBalance < 22) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
return 4;
}
}
else if (ssDCval >= 50) {
if (MVDiff < 8) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (Rs < 511) {
return 4;
}
else if (Rs >= 511) {
return 2;
}
}
}
else if (MVDiff >= 8) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (tcor < 94) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 2) {
if (Cs < 756) {
return 2;
}
else if (Cs >= 756) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsCsDiff < 23) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 7) {
if (Rs < 479) {
return 4;
}
else if (Rs >= 479) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
return 1;
}
}
}
else if (RsCsDiff >= 23) {
if (Cs < 676) {
return 8;
}
else if (Cs >= 676) {
if (Rs < 382) {
return 2;
}
else if (Rs >= 382) {
return 4;
}
}
}
}
}
}
else if (tcor >= 94) {
if (ssDCval < 56) {
if (posBalance < 42) {
if (Cs < 542) {
if (Rs < 301) {
return 1;
}
else if (Rs >= 301) {
return 8;
}
}
else if (Cs >= 542) {
if (Cs < 1026) {
return 4;
}
else if (Cs >= 1026) {
return 2;
}
}
}
else if (posBalance >= 42) {
if (Rs < 460) {
if (Rs < 330) {
if (Rs < 302) {
return 1;
}
else if (Rs >= 302) {
return 4;
}
}
else if (Rs >= 330) {
return 1;
}
}
else if (Rs >= 460) {
if (diffAFD < -1) {
if (Cs < 1011) {
return 1;
}
else if (Cs >= 1011) {
return 4;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
}
else if (ssDCval >= 56) {
if (ssDCval < 59) {
if (Cs < 611) {
if (Rs < 433) {
return 4;
}
else if (Rs >= 433) {
return 2;
}
}
else if (Cs >= 611) {
return 4;
}
}
else if (ssDCval >= 59) {
if (Rs < 276) {
return 4;
}
else if (Rs >= 276) {
if (Cs < 631) {
return 1;
}
else if (Cs >= 631) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 532) {
if (MVDiff < 1) {
if (Cs < 938) {
if (ssDCval < 54) {
if (mu_mv_mag_sq < 114) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
return 4;
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (Cs >= 938) {
if (Rs < 616) {
if (Cs < 1062) {
return 1;
}
else if (Cs >= 1062) {
return 4;
}
}
else if (Rs >= 616) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (Rs < 817) {
return 1;
}
else if (Rs >= 817) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 17) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
return 1;
}
}
else if (negBalance >= 17) {
if (ssDCval < 62) {
if (TSC < 3) {
if (ssDCval < 48) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (ssDCval >= 48) {
if (SC < 1415) {
if (negBalance < 29) {
if (SC < 1265) {
return 4;
}
else if (SC >= 1265) {
return 1;
}
}
else if (negBalance >= 29) {
if (Rs < 758) {
return 8;
}
else if (Rs >= 758) {
return 4;
}
}
}
else if (SC >= 1415) {
if (TSCindex < 3) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (TSCindex >= 3) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
return 8;
}
}
}
}
}
else if (TSC >= 3) {
if (MVDiff < 15) {
if (RsCsDiff < 16) {
if (negBalance < 29) {
if (RsCsDiff < 5) {
if (Cs < 644) {
return 4;
}
else if (Cs >= 644) {
if (Rs < 640) {
return 8;
}
else if (Rs >= 640) {
return 1;
}
}
}
else if (RsCsDiff >= 5) {
return 4;
}
}
else if (negBalance >= 29) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
if (negBalance < 39) {
if (posBalance < 53) {
if (diffAFD < 1) {
if (AFD < 5) {
if (Rs < 700) {
return 4;
}
else if (Rs >= 700) {
return 8;
}
}
else if (AFD >= 5) {
return 4;
}
}
else if (diffAFD >= 1) {
if (Cs < 611) {
return 8;
}
else if (Cs >= 611) {
if (Rs < 582) {
return 4;
}
else if (Rs >= 582) {
return 1;
}
}
}
}
else if (posBalance >= 53) {
if (Rs < 653) {
if (Rs < 588) {
return 4;
}
else if (Rs >= 588) {
return 1;
}
}
else if (Rs >= 653) {
return 8;
}
}
}
else if (negBalance >= 39) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Cs < 822) {
return 1;
}
else if (Cs >= 822) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 16) {
if (MVDiff < 8) {
if (SC < 1401) {
return 4;
}
else if (SC >= 1401) {
return 8;
}
}
else if (MVDiff >= 8) {
if (Rs < 733) {
if (Cs < 779) {
return 4;
}
else if (Cs >= 779) {
return 2;
}
}
else if (Rs >= 733) {
return 4;
}
}
}
}
else if (MVDiff >= 15) {
if (RsDiff < 2) {
if (Rs < 753) {
return 8;
}
else if (Rs >= 753) {
return 4;
}
}
else if (RsDiff >= 2) {
if (diffRsCsdiff < 3) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 4;
}
}
else if (diffRsCsdiff >= 3) {
return 8;
}
}
}
}
}
else if (ssDCval >= 62) {
if (RsDiff < 2) {
if (AFD < 4) {
if (Rs < 719) {
return 2;
}
else if (Rs >= 719) {
return 4;
}
}
else if (AFD >= 4) {
return 8;
}
}
else if (RsDiff >= 2) {
return 2;
}
}
}
}
}
}
else if (Rs >= 832) {
if (refDCval < 49) {
if (tcor < 95) {
if (AFD < 6) {
if (Rs < 897) {
return 4;
}
else if (Rs >= 897) {
return 1;
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (tcor >= 95) {
return 8;
}
}
else if (refDCval >= 49) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 120) {
if (mcTcor < 97) {
if (posBalance < 48) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Cs < 954) {
return 1;
}
else if (Cs >= 954) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (refDCval < 59) {
return 8;
}
else if (refDCval >= 59) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 44) {
if (diffAFD < -1) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (Rs < 983) {
return 4;
}
else if (Rs >= 983) {
return 1;
}
}
}
else if (diffAFD >= -1) {
if (TSC < 5) {
if (Rs < 999) {
return 1;
}
else if (Rs >= 999) {
if (Rs < 1001) {
return 8;
}
else if (Rs >= 1001) {
return 1;
}
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (negBalance >= 44) {
if (Cs < 652) {
return 4;
}
else if (Cs >= 652) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (SC < 1728) {
return 8;
}
else if (SC >= 1728) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Cs < 1109) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (Cs >= 1109) {
return 1;
}
}
}
}
}
else if (MVDiff >= 8) {
if (Cs < 779) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 1;
}
}
}
else if (Cs >= 779) {
if (Cs < 896) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Cs >= 896) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 64) {
if (negBalance < 28) {
if (Cs < 697) {
if (SC < 1236) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Cs < 612) {
return 1;
}
else if (Cs >= 612) {
return 4;
}
}
}
else if (SC >= 1236) {
if (posBalance < 38) {
if (ssDCval < 65) {
if (Rs < 874) {
return 2;
}
else if (Rs >= 874) {
return 1;
}
}
else if (ssDCval >= 65) {
if (Cs < 594) {
return 2;
}
else if (Cs >= 594) {
return 8;
}
}
}
else if (posBalance >= 38) {
if (MVDiff < 7) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 2;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
else if (Cs >= 697) {
if (ssDCval < 71) {
if (Rs < 894) {
if (RsDiff < 1) {
if (Rs < 689) {
if (SC < 1250) {
return 4;
}
else if (SC >= 1250) {
return 8;
}
}
else if (Rs >= 689) {
if (mu_mv_mag_sq < 112) {
return 8;
}
else if (mu_mv_mag_sq >= 112) {
if (Cs < 730) {
if (Rs < 710) {
return 4;
}
else if (Rs >= 710) {
return 8;
}
}
else if (Cs >= 730) {
return 4;
}
}
}
}
else if (RsDiff >= 1) {
if (Cs < 699) {
return 1;
}
else if (Cs >= 699) {
return 4;
}
}
}
else if (Rs >= 894) {
if (Cs < 811) {
return 2;
}
else if (Cs >= 811) {
if (Cs < 956) {
return 1;
}
else if (Cs >= 956) {
if (Rs < 928) {
return 4;
}
else if (Rs >= 928) {
return 8;
}
}
}
}
}
else if (ssDCval >= 71) {
if (Rs < 866) {
if (Rs < 652) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
if (Cs < 743) {
if (Rs < 603) {
return 2;
}
else if (Rs >= 603) {
return 1;
}
}
else if (Cs >= 743) {
return 2;
}
}
}
else if (Rs >= 652) {
if (mu_mv_mag_sq < 109) {
if (Cs < 789) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
else if (Cs >= 789) {
return 2;
}
}
else if (mu_mv_mag_sq >= 109) {
return 1;
}
}
}
else if (Rs >= 866) {
return 8;
}
}
}
}
else if (negBalance >= 28) {
if (ssDCval < 84) {
if (RsDiff < 2) {
if (TSC < 4) {
if (CsDiff < 1) {
if (Rs < 716) {
return 8;
}
else if (Rs >= 716) {
if (Rs < 911) {
return 4;
}
else if (Rs >= 911) {
return 8;
}
}
}
else if (CsDiff >= 1) {
if (MVDiff < 1) {
if (mu_mv_mag_sq < 80) {
return 8;
}
else if (mu_mv_mag_sq >= 80) {
if (ssDCval < 71) {
if (refDCval < 65) {
return 4;
}
else if (refDCval >= 65) {
if (Cs < 603) {
return 2;
}
else if (Cs >= 603) {
if (Cs < 1098) {
if (mu_mv_mag_sq < 126) {
return 1;
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
else if (Cs >= 1098) {
return 2;
}
}
}
}
else if (ssDCval >= 71) {
if (Cs < 906) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Cs >= 906) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 4;
}
}
}
}
}
else if (MVDiff >= 1) {
if (mcTcor < 98) {
if (ssDCval < 69) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (ssDCval < 67) {
if (Rs < 763) {
return 8;
}
else if (Rs >= 763) {
return 1;
}
}
else if (ssDCval >= 67) {
return 1;
}
}
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (mcTcor >= 98) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
if (Rs < 736) {
if (Cs < 685) {
return 8;
}
else if (Cs >= 685) {
return 1;
}
}
else if (Rs >= 736) {
if (Rs < 867) {
return 4;
}
else if (Rs >= 867) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (posBalance < 44) {
if (Cs < 684) {
return 4;
}
else if (Cs >= 684) {
return 8;
}
}
else if (posBalance >= 44) {
if (negBalance < 33) {
if (ssDCval < 68) {
if (Rs < 870) {
return 8;
}
else if (Rs >= 870) {
return 2;
}
}
else if (ssDCval >= 68) {
if (Rs < 756) {
return 1;
}
else if (Rs >= 756) {
return 4;
}
}
}
else if (negBalance >= 33) {
if (Rs < 320) {
return 2;
}
else if (Rs >= 320) {
if (mu_mv_mag_sq < 102) {
return 8;
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 704) {
return 8;
}
else if (Rs >= 704) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (refDCval < 75) {
if (RsCsDiff < 23) {
if (refDCval < 69) {
if (tcor < 94) {
if (diffMVdiffVal < 1) {
if (MVDiff < 6) {
if (Cs < 862) {
return 2;
}
else if (Cs >= 862) {
return 8;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
else if (tcor >= 94) {
if (CsDiff < 2) {
if (Cs < 771) {
if (Rs < 827) {
return 8;
}
else if (Rs >= 827) {
return 1;
}
}
else if (Cs >= 771) {
return 2;
}
}
else if (CsDiff >= 2) {
if (Rs < 508) {
return 1;
}
else if (Rs >= 508) {
if (MVDiff < 16) {
if (ssDCval < 69) {
if (Cs < 1095) {
return 4;
}
else if (Cs >= 1095) {
if (SC < 2037) {
if (negBalance < 38) {
return 2;
}
else if (negBalance >= 38) {
return 4;
}
}
else if (SC >= 2037) {
return 4;
}
}
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (MVDiff >= 16) {
return 8;
}
}
}
}
}
else if (refDCval >= 69) {
if (negBalance < 35) {
if (Rs < 781) {
return 1;
}
else if (Rs >= 781) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (negBalance >= 35) {
if (Cs < 852) {
if (refDCval < 70) {
if (Cs < 745) {
return 1;
}
else if (Cs >= 745) {
return 2;
}
}
else if (refDCval >= 70) {
if (tcor < 87) {
if (Rs < 340) {
return 1;
}
else if (Rs >= 340) {
return 8;
}
}
else if (tcor >= 87) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 733) {
return 1;
}
else if (Rs >= 733) {
return 4;
}
}
}
}
}
else if (Cs >= 852) {
if (SC < 1947) {
return 1;
}
else if (SC >= 1947) {
if (Rs < 855) {
return 8;
}
else if (Rs >= 855) {
if (Rs < 897) {
return 1;
}
else if (Rs >= 897) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 23) {
if (Rs < 818) {
if (Cs < 1192) {
return 2;
}
else if (Cs >= 1192) {
return 1;
}
}
else if (Rs >= 818) {
if (Cs < 907) {
return 8;
}
else if (Cs >= 907) {
if (Rs < 942) {
return 2;
}
else if (Rs >= 942) {
if (Rs < 967) {
return 8;
}
else if (Rs >= 967) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 75) {
if (Cs < 1078) {
if (Cs < 820) {
if (SC < 1434) {
return 8;
}
else if (SC >= 1434) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (Cs >= 820) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (ssDCval < 83) {
return 2;
}
else if (ssDCval >= 83) {
return 8;
}
}
}
}
else if (Cs >= 1078) {
if (Rs < 781) {
if (Rs < 545) {
return 1;
}
else if (Rs >= 545) {
return 2;
}
}
else if (Rs >= 781) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (posBalance < 48) {
if (posBalance < 37) {
if (mu_mv_mag_sq < 76) {
return 2;
}
else if (mu_mv_mag_sq >= 76) {
return 1;
}
}
else if (posBalance >= 37) {
return 2;
}
}
else if (posBalance >= 48) {
if (SC < 1468) {
if (mcTcor < 98) {
if (SC < 1462) {
return 8;
}
else if (SC >= 1462) {
return 2;
}
}
else if (mcTcor >= 98) {
if (AFD < 5) {
if (Cs < 734) {
return 1;
}
else if (Cs >= 734) {
return 2;
}
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (SC >= 1468) {
if (Rs < 904) {
return 4;
}
else if (Rs >= 904) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1002) {
if (Cs < 1081) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 43) {
if (ssDCval < 51) {
if (Rs < 1233) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (Rs >= 1233) {
if (MVDiff < 2) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Rs < 1646) {
return 8;
}
else if (Rs >= 1646) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (Rs < 1639) {
return 1;
}
else if (Rs >= 1639) {
return 4;
}
}
}
}
else if (ssDCval >= 51) {
if (Cs < 726) {
if (RsDiff < 3) {
if (ssDCval < 70) {
if (SC < 1578) {
return 2;
}
else if (SC >= 1578) {
if (negBalance < 40) {
if (Cs < 713) {
return 1;
}
else if (Cs >= 713) {
if (Cs < 714) {
return 8;
}
else if (Cs >= 714) {
return 1;
}
}
}
else if (negBalance >= 40) {
return 4;
}
}
}
else if (ssDCval >= 70) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
}
else if (RsDiff >= 3) {
if (CsDiff < 3) {
if (ssDCval < 70) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (ssDCval >= 70) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (Rs < 1390) {
return 8;
}
else if (Rs >= 1390) {
return 4;
}
}
}
}
else if (CsDiff >= 3) {
if (Rs < 1451) {
if (Rs < 1182) {
return 1;
}
else if (Rs >= 1182) {
return 8;
}
}
else if (Rs >= 1451) {
return 1;
}
}
}
}
else if (Cs >= 726) {
if (negBalance < 22) {
if (posBalance < 32) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
return 1;
}
}
else if (posBalance >= 32) {
if (Rs < 1332) {
if (SC < 1800) {
return 2;
}
else if (SC >= 1800) {
return 1;
}
}
else if (Rs >= 1332) {
if (SC < 2299) {
return 8;
}
else if (SC >= 2299) {
if (posBalance < 35) {
if (Rs < 1707) {
return 8;
}
else if (Rs >= 1707) {
return 1;
}
}
else if (posBalance >= 35) {
return 1;
}
}
}
}
}
else if (negBalance >= 22) {
if (mu_mv_mag_sq < 86) {
if (refDCval < 81) {
return 1;
}
else if (refDCval >= 81) {
if (Rs < 1414) {
return 8;
}
else if (Rs >= 1414) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (ssDCval < 60) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
if (Rs < 1293) {
return 1;
}
else if (Rs >= 1293) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (negBalance < 58) {
return 8;
}
else if (negBalance >= 58) {
if (Rs < 1132) {
return 8;
}
else if (Rs >= 1132) {
return 4;
}
}
}
}
}
}
}
}
else if (posBalance >= 43) {
if (Rs < 1386) {
if (mcTcor < 99) {
if (MVDiff < 11) {
if (SC < 2049) {
if (TSC < 3) {
if (SC < 1783) {
if (Rs < 1051) {
return 1;
}
else if (Rs >= 1051) {
if (Cs < 668) {
return 2;
}
else if (Cs >= 668) {
return 4;
}
}
}
else if (SC >= 1783) {
return 1;
}
}
else if (TSC >= 3) {
if (posBalance < 56) {
if (ssDCval < 48) {
if (Rs < 1188) {
return 8;
}
else if (Rs >= 1188) {
if (Rs < 1357) {
return 4;
}
else if (Rs >= 1357) {
return 4;
}
}
}
else if (ssDCval >= 48) {
if (CsDiff < 3) {
if (diffRsCsdiff < 6) {
if (diffMVdiffVal < -2) {
if (Rs < 1132) {
return 8;
}
else if (Rs >= 1132) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
else if (diffRsCsdiff >= 6) {
if (Rs < 1121) {
return 8;
}
else if (Rs >= 1121) {
return 1;
}
}
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 71) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
return 8;
}
}
else if (mu_mv_mag_sq >= 71) {
return 8;
}
}
}
}
else if (posBalance >= 56) {
if (MVDiff < 4) {
if (Cs < 973) {
return 8;
}
else if (Cs >= 973) {
return 4;
}
}
else if (MVDiff >= 4) {
if (Rs < 1186) {
return 8;
}
else if (Rs >= 1186) {
return 1;
}
}
}
}
}
else if (SC >= 2049) {
if (ssDCval < 68) {
if (ssDCval < 58) {
if (MVDiff < 6) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
return 8;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (ssDCval >= 58) {
if (Rs < 1326) {
if (SC < 2193) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (Rs < 1268) {
if (Rs < 1091) {
return 4;
}
else if (Rs >= 1091) {
return 8;
}
}
else if (Rs >= 1268) {
return 1;
}
}
}
else if (SC >= 2193) {
if (TSC < 5) {
if (SC < 2256) {
if (Cs < 1058) {
if (Rs < 1257) {
return 8;
}
else if (Rs >= 1257) {
return 1;
}
}
else if (Cs >= 1058) {
return 1;
}
}
else if (SC >= 2256) {
return 8;
}
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (Rs >= 1326) {
if (AFD < 6) {
if (Rs < 1361) {
return 8;
}
else if (Rs >= 1361) {
return 1;
}
}
else if (AFD >= 6) {
return 4;
}
}
}
}
else if (ssDCval >= 68) {
if (Rs < 1282) {
if (Rs < 1083) {
if (Rs < 1010) {
return 8;
}
else if (Rs >= 1010) {
if (Rs < 1058) {
return 4;
}
else if (Rs >= 1058) {
return 2;
}
}
}
else if (Rs >= 1083) {
return 8;
}
}
else if (Rs >= 1282) {
if (Cs < 932) {
return 1;
}
else if (Cs >= 932) {
return 4;
}
}
}
}
}
else if (MVDiff >= 11) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 56) {
if (Cs < 657) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 1;
}
}
else if (Cs >= 657) {
if (Rs < 1309) {
return 2;
}
else if (Rs >= 1309) {
if (Rs < 1342) {
return 8;
}
else if (Rs >= 1342) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 56) {
return 8;
}
}
}
}
else if (mcTcor >= 99) {
if (SC < 1806) {
if (mu_mv_mag_sq < 109) {
if (Cs < 612) {
return 8;
}
else if (Cs >= 612) {
return 1;
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 1143) {
return 2;
}
else if (Rs >= 1143) {
return 1;
}
}
}
else if (SC >= 1806) {
if (Cs < 968) {
if (tcor < 100) {
return 1;
}
else if (tcor >= 100) {
if (Rs < 1352) {
return 8;
}
else if (Rs >= 1352) {
return 1;
}
}
}
else if (Cs >= 968) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (Rs < 1011) {
return 1;
}
else if (Rs >= 1011) {
if (Rs < 1146) {
return 4;
}
else if (Rs >= 1146) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1386) {
if (mu_mv_mag_sq < 95) {
if (ssDCval < 61) {
if (negBalance < 24) {
return 1;
}
else if (negBalance >= 24) {
return 8;
}
}
else if (ssDCval >= 61) {
if (Cs < 821) {
if (mcTcor < 97) {
if (diffTSC < -1) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (Rs < 1558) {
return 8;
}
else if (Rs >= 1558) {
return 2;
}
}
}
else if (diffTSC >= -1) {
if (ssDCval < 70) {
if (Rs < 1558) {
return 8;
}
else if (Rs >= 1558) {
return 2;
}
}
else if (ssDCval >= 70) {
return 1;
}
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (Cs >= 821) {
if (negBalance < 37) {
if (Rs < 1558) {
return 1;
}
else if (Rs >= 1558) {
return 2;
}
}
else if (negBalance >= 37) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (ssDCval < 74) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Rs < 1650) {
return 8;
}
else if (Rs >= 1650) {
return 2;
}
}
}
else if (ssDCval >= 74) {
if (Rs < 1489) {
return 1;
}
else if (Rs >= 1489) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1318) {
if (posBalance < 41) {
if (Rs < 1143) {
return 8;
}
else if (Rs >= 1143) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (negBalance < 17) {
if (posBalance < 38) {
if (Rs < 1242) {
if (Rs < 1180) {
return 8;
}
else if (Rs >= 1180) {
return 1;
}
}
else if (Rs >= 1242) {
return 8;
}
}
else if (posBalance >= 38) {
return 1;
}
}
else if (negBalance >= 17) {
return 1;
}
}
}
}
else if (posBalance >= 41) {
if (Rs < 1168) {
return 4;
}
else if (Rs >= 1168) {
if (Cs < 972) {
return 8;
}
else if (Cs >= 972) {
return 4;
}
}
}
}
else if (Rs >= 1318) {
if (Rs < 1662) {
return 8;
}
else if (Rs >= 1662) {
if (Cs < 1067) {
if (Rs < 1715) {
return 4;
}
else if (Rs >= 1715) {
return 4;
}
}
else if (Cs >= 1067) {
return 8;
}
}
}
}
}
else if (Cs >= 1081) {
if (tcor < 98) {
if (MVDiff < 6) {
if (Rs < 1335) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 113) {
return 8;
}
else if (mu_mv_mag_sq >= 113) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 107) {
if (AFD < 8) {
if (refDCval < 68) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Cs < 1194) {
return 8;
}
else if (Cs >= 1194) {
return 4;
}
}
}
}
}
else if (refDCval >= 68) {
if (Cs < 1205) {
return 8;
}
else if (Cs >= 1205) {
if (Cs < 1265) {
return 2;
}
else if (Cs >= 1265) {
return 8;
}
}
}
}
else if (AFD >= 8) {
return 4;
}
}
else if (mu_mv_mag_sq >= 107) {
if (negBalance < 37) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (mu_mv_mag_sq < 121) {
if (Rs < 1240) {
if (Cs < 1184) {
return 4;
}
else if (Cs >= 1184) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 4;
}
}
}
else if (Rs >= 1240) {
return 8;
}
}
else if (mu_mv_mag_sq >= 121) {
if (AFD < 3) {
if (Rs < 1163) {
return 4;
}
else if (Rs >= 1163) {
return 1;
}
}
else if (AFD >= 3) {
return 4;
}
}
}
}
else if (negBalance >= 37) {
if (Rs < 1114) {
return 4;
}
else if (Rs >= 1114) {
return 1;
}
}
}
}
}
else if (Rs >= 1335) {
if (posBalance < 44) {
if (Rs < 1485) {
return 8;
}
else if (Rs >= 1485) {
if (diffRsCsdiff < 2) {
return 4;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 1701) {
if (Rs < 1670) {
return 4;
}
else if (Rs >= 1670) {
return 4;
}
}
else if (Rs >= 1701) {
return 8;
}
}
}
}
else if (posBalance >= 44) {
if (SC < 2498) {
if (Rs < 1353) {
return 8;
}
else if (Rs >= 1353) {
return 4;
}
}
else if (SC >= 2498) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
if (AFD < 7) {
if (Cs < 1228) {
if (Rs < 1152) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (Rs >= 1152) {
if (Rs < 1554) {
if (Cs < 1197) {
if (Cs < 1104) {
return 8;
}
else if (Cs >= 1104) {
return 1;
}
}
else if (Cs >= 1197) {
return 8;
}
}
else if (Rs >= 1554) {
return 8;
}
}
}
else if (Cs >= 1228) {
return 4;
}
}
else if (AFD >= 7) {
if (diffMVdiffVal < 0) {
return 4;
}
else if (diffMVdiffVal >= 0) {
if (Cs < 1125) {
return 8;
}
else if (Cs >= 1125) {
return 2;
}
}
}
}
}
else if (tcor >= 98) {
if (RsDiff < 1) {
if (MVDiff < 8) {
if (ssDCval < 55) {
if (Rs < 1583) {
return 8;
}
else if (Rs >= 1583) {
return 4;
}
}
else if (ssDCval >= 55) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
if (Rs < 1132) {
return 4;
}
else if (Rs >= 1132) {
return 8;
}
}
}
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (RsDiff >= 1) {
if (ssDCval < 58) {
if (mu_mv_mag_sq < 123) {
return 8;
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1192) {
return 8;
}
else if (Rs >= 1192) {
if (refDCval < 47) {
if (Rs < 1547) {
return 8;
}
else if (Rs >= 1547) {
return 4;
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1742) {
if (ssDCval < 49) {
if (SC < 2839) {
if (ssDCval < 48) {
if (CsDiff < 2) {
if (AFD < 2) {
if (Rs < 1683) {
return 8;
}
else if (Rs >= 1683) {
if (Rs < 1709) {
return 4;
}
else if (Rs >= 1709) {
return 8;
}
}
}
else if (AFD >= 2) {
if (diffRsCsdiff < 2) {
if (Cs < 1092) {
return 8;
}
else if (Cs >= 1092) {
if (Cs < 1103) {
return 4;
}
else if (Cs >= 1103) {
if (Cs < 1156) {
if (Cs < 1147) {
if (Cs < 1116) {
return 4;
}
else if (Cs >= 1116) {
if (Cs < 1137) {
return 4;
}
else if (Cs >= 1137) {
return 4;
}
}
}
else if (Cs >= 1147) {
return 8;
}
}
else if (Cs >= 1156) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
}
else if (CsDiff >= 2) {
if (Rs < 1672) {
return 8;
}
else if (Rs >= 1672) {
return 8;
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (SC >= 2839) {
if (Rs < 1728) {
return 4;
}
else if (Rs >= 1728) {
if (Rs < 1731) {
if (Cs < 1132) {
return 4;
}
else if (Cs >= 1132) {
return 4;
}
}
else if (Rs >= 1731) {
return 4;
}
}
}
}
else if (ssDCval >= 49) {
if (Rs < 1463) {
return 4;
}
else if (Rs >= 1463) {
return 8;
}
}
}
else if (Rs >= 1742) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 58) {
if (MVDiff < 8) {
if (refDCval < 70) {
if (mu_mv_mag_sq < 126) {
return 1;
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1267) {
return 4;
}
else if (Cs >= 1267) {
return 1;
}
}
}
else if (refDCval >= 70) {
if (Rs < 1530) {
if (Cs < 1109) {
if (SC < 2127) {
return 4;
}
else if (SC >= 2127) {
return 8;
}
}
else if (Cs >= 1109) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
}
}
else if (Rs >= 1530) {
return 8;
}
}
}
else if (MVDiff >= 8) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1745) {
if (TSC < 4) {
if (ssDCval < 82) {
if (ssDCval < 48) {
if (TSCindex < 3) {
if (Cs < 1187) {
if (SC < 2954) {
if (negBalance < 15) {
return 8;
}
else if (negBalance >= 15) {
if (Rs < 1820) {
if (negBalance < 20) {
if (posBalance < 40) {
if (posBalance < 39) {
if (diffAFD < 1) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (Rs < 1758) {
if (Rs < 1751) {
return 4;
}
else if (Rs >= 1751) {
return 4;
}
}
else if (Rs >= 1758) {
if (posBalance < 33) {
return 4;
}
else if (posBalance >= 33) {
if (diffAFD < -1) {
if (Rs < 1766) {
return 4;
}
else if (Rs >= 1766) {
return 4;
}
}
else if (diffAFD >= -1) {
if (Rs < 1770) {
return 8;
}
else if (Rs >= 1770) {
if (posBalance < 36) {
return 4;
}
else if (posBalance >= 36) {
return 8;
}
}
}
}
}
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (posBalance >= 39) {
return 8;
}
}
else if (posBalance >= 40) {
if (Rs < 1778) {
return 4;
}
else if (Rs >= 1778) {
return 4;
}
}
}
else if (negBalance >= 20) {
if (SC < 2877) {
return 4;
}
else if (SC >= 2877) {
if (posBalance < 36) {
if (Rs < 1756) {
return 8;
}
else if (Rs >= 1756) {
if (SC < 2920) {
if (Cs < 1096) {
return 4;
}
else if (Cs >= 1096) {
return 4;
}
}
else if (SC >= 2920) {
return 4;
}
}
}
else if (posBalance >= 36) {
return 8;
}
}
}
}
else if (Rs >= 1820) {
return 4;
}
}
}
else if (SC >= 2954) {
return 8;
}
}
else if (Cs >= 1187) {
if (Rs < 2032) {
return 4;
}
else if (Rs >= 2032) {
return 8;
}
}
}
else if (TSCindex >= 3) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 8;
}
}
}
else if (ssDCval >= 48) {
if (ssDCval < 71) {
if (mu_mv_mag_sq < 100) {
if (posBalance < 38) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (tcor < 96) {
if (Cs < 793) {
return 4;
}
else if (Cs >= 793) {
return 2;
}
}
else if (tcor >= 96) {
return 4;
}
}
}
else if (posBalance >= 38) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (SC < 2588) {
return 1;
}
else if (SC >= 2588) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 938) {
if (gchDC < 1) {
if (MVDiff < 6) {
if (ssDCval < 51) {
if (Rs < 2348) {
return 8;
}
else if (Rs >= 2348) {
return 1;
}
}
else if (ssDCval >= 51) {
if (negBalance < 32) {
return 1;
}
else if (negBalance >= 32) {
if (Rs < 1775) {
return 8;
}
else if (Rs >= 1775) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 109) {
return 1;
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
}
else if (Cs >= 938) {
if (Cs < 1195) {
if (Rs < 1965) {
if (Cs < 993) {
if (Rs < 1776) {
return 8;
}
else if (Rs >= 1776) {
return 1;
}
}
else if (Cs >= 993) {
if (SC < 2919) {
if (CsDiff < 1) {
if (Rs < 1846) {
return 1;
}
else if (Rs >= 1846) {
return 4;
}
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (SC >= 2919) {
if (MVDiff < 2) {
if (negBalance < 19) {
if (Rs < 1844) {
return 1;
}
else if (Rs >= 1844) {
if (SC < 2950) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (SC >= 2950) {
return 4;
}
}
}
else if (negBalance >= 19) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
if (Rs < 1870) {
return 1;
}
else if (Rs >= 1870) {
if (Cs < 1085) {
return 4;
}
else if (Cs >= 1085) {
return 1;
}
}
}
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
else if (Rs >= 1965) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
return 4;
}
}
}
else if (Cs >= 1195) {
if (Cs < 1272) {
return 8;
}
else if (Cs >= 1272) {
return 2;
}
}
}
}
}
else if (ssDCval >= 71) {
if (Cs < 985) {
return 1;
}
else if (Cs >= 985) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
else if (ssDCval >= 82) {
if (MVDiff < 1) {
if (diffMVdiffVal < -1) {
if (SC < 2599) {
return 4;
}
else if (SC >= 2599) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (MVDiff >= 1) {
if (MVDiff < 3) {
if (Rs < 1875) {
return 4;
}
else if (Rs >= 1875) {
return 8;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 932) {
if (refDCval < 70) {
if (negBalance < 32) {
if (SC < 3090) {
return 1;
}
else if (SC >= 3090) {
return 2;
}
}
else if (negBalance >= 32) {
return 2;
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 89) {
if (diffRsCsdiff < -45) {
return 2;
}
else if (diffRsCsdiff >= -45) {
if (posBalance < 34) {
if (Rs < 2022) {
return 1;
}
else if (Rs >= 2022) {
if (Rs < 2241) {
return 4;
}
else if (Rs >= 2241) {
return 8;
}
}
}
else if (posBalance >= 34) {
if (diffMVdiffVal < -8) {
return 2;
}
else if (diffMVdiffVal >= -8) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
if (Rs < 2169) {
return 2;
}
else if (Rs >= 2169) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (SC < 2944) {
return 4;
}
else if (SC >= 2944) {
if (Rs < 2262) {
return 8;
}
else if (Rs >= 2262) {
if (Rs < 2278) {
return 4;
}
else if (Rs >= 2278) {
return 4;
}
}
}
}
}
}
else if (Cs >= 932) {
if (RsCsDiff < 33) {
if (ssDCval < 75) {
if (diffRsCsdiff < -6) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 1;
}
}
else if (diffRsCsdiff >= -6) {
if (TSC < 5) {
if (Rs < 2511) {
return 4;
}
else if (Rs >= 2511) {
return 4;
}
}
else if (TSC >= 5) {
if (Rs < 2295) {
return 8;
}
else if (Rs >= 2295) {
return 4;
}
}
}
}
else if (ssDCval >= 75) {
if (Cs < 1202) {
if (Rs < 2193) {
if (Cs < 988) {
return 1;
}
else if (Cs >= 988) {
return 8;
}
}
else if (Rs >= 2193) {
return 1;
}
}
else if (Cs >= 1202) {
if (Rs < 2030) {
return 1;
}
else if (Rs >= 2030) {
return 4;
}
}
}
}
else if (RsCsDiff >= 33) {
if (Rs < 2236) {
return 8;
}
else if (Rs >= 2236) {
if (Cs < 1215) {
return 2;
}
else if (Cs >= 1215) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 88) {
if (Rs < 410) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 104) {
if (ssDCval < 101) {
if (Cs < 745) {
return 1;
}
else if (Cs >= 745) {
if (Rs < 382) {
return 4;
}
else if (Rs >= 382) {
return 8;
}
}
}
else if (ssDCval >= 101) {
if (Rs < 254) {
return 2;
}
else if (Rs >= 254) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 331) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 331) {
if (Rs < 369) {
if (ssDCval < 109) {
return 4;
}
else if (ssDCval >= 109) {
return 8;
}
}
else if (Rs >= 369) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 111) {
if (Rs < 342) {
return 2;
}
else if (Rs >= 342) {
return 8;
}
}
else if (ssDCval >= 111) {
if (Cs < 556) {
if (Rs < 379) {
return 8;
}
else if (Rs >= 379) {
return 4;
}
}
else if (Cs >= 556) {
return 2;
}
}
}
}
else if (Rs >= 410) {
if (mu_mv_mag_sq < 128) {
if (Rs < 630) {
if (Cs < 1212) {
if (mu_mv_mag_sq < 102) {
if (MVDiff < 4) {
if (mcTcor < 99) {
if (AFD < 4) {
if (Rs < 471) {
return 8;
}
else if (Rs >= 471) {
if (Rs < 526) {
return 4;
}
else if (Rs >= 526) {
return 2;
}
}
}
else if (AFD >= 4) {
if (Cs < 578) {
if (diffTSC < -1) {
if (Cs < 565) {
return 2;
}
else if (Cs >= 565) {
return 1;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (Cs >= 578) {
if (negBalance < 32) {
if (Rs < 572) {
return 8;
}
else if (Rs >= 572) {
return 2;
}
}
else if (negBalance >= 32) {
if (Rs < 540) {
return 1;
}
else if (Rs >= 540) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (Rs < 575) {
return 1;
}
else if (Rs >= 575) {
if (MVDiff < 3) {
if (Cs < 594) {
return 2;
}
else if (Cs >= 594) {
return 1;
}
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
else if (MVDiff >= 4) {
if (mcTcor < 98) {
if (negBalance < 24) {
if (Rs < 571) {
return 4;
}
else if (Rs >= 571) {
return 8;
}
}
else if (negBalance >= 24) {
if (AFD < 9) {
if (Cs < 628) {
if (negBalance < 35) {
return 2;
}
else if (negBalance >= 35) {
if (Rs < 597) {
return 8;
}
else if (Rs >= 597) {
return 2;
}
}
}
else if (Cs >= 628) {
return 8;
}
}
else if (AFD >= 9) {
if (tcor < 86) {
return 1;
}
else if (tcor >= 86) {
if (Rs < 490) {
return 8;
}
else if (Rs >= 490) {
return 2;
}
}
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 69) {
if (TSC < 4) {
if (mu_mv_mag_sq < 100) {
if (CsDiff < 3) {
if (posBalance < 71) {
if (ssDCval < 123) {
return 2;
}
else if (ssDCval >= 123) {
return 1;
}
}
else if (posBalance >= 71) {
return 1;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 100) {
return 4;
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 67) {
if (ssDCval < 163) {
return 2;
}
else if (ssDCval >= 163) {
if (Rs < 571) {
return 4;
}
else if (Rs >= 571) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 450) {
return 2;
}
else if (Rs >= 450) {
return 8;
}
}
}
}
else if (negBalance >= 69) {
if (TSC < 3) {
if (Rs < 586) {
return 8;
}
else if (Rs >= 586) {
return 1;
}
}
else if (TSC >= 3) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (SC < 1221) {
if (Cs < 707) {
if (SC < 1100) {
if (refDCval < 114) {
return 1;
}
else if (refDCval >= 114) {
return 8;
}
}
else if (SC >= 1100) {
return 8;
}
}
else if (Cs >= 707) {
if (ssDCval < 117) {
return 1;
}
else if (ssDCval >= 117) {
return 8;
}
}
}
else if (SC >= 1221) {
if (Cs < 794) {
if (mu_mv_mag_sq < 108) {
if (Rs < 526) {
return 2;
}
else if (Rs >= 526) {
if (Rs < 598) {
return 4;
}
else if (Rs >= 598) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 108) {
return 8;
}
}
else if (Cs >= 794) {
return 8;
}
}
}
}
else if (Cs >= 1212) {
if (ssDCval < 120) {
return 1;
}
else if (ssDCval >= 120) {
return 8;
}
}
}
else if (Rs >= 630) {
if (ssDCval < 154) {
if (diffRsCsdiff < -6) {
if (ssDCval < 117) {
if (Cs < 797) {
if (refDCval < 90) {
return 1;
}
else if (refDCval >= 90) {
if (Rs < 798) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (Rs >= 798) {
if (SC < 2773) {
if (RsDiff < 5) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 8;
}
}
else if (RsDiff >= 5) {
if (Rs < 1602) {
return 1;
}
else if (Rs >= 1602) {
return 8;
}
}
}
else if (SC >= 2773) {
if (Rs < 2283) {
return 1;
}
else if (Rs >= 2283) {
return 2;
}
}
}
}
}
else if (Cs >= 797) {
if (mu_mv_mag_sq < 124) {
if (ssDCval < 104) {
if (Rs < 1061) {
if (Cs < 936) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 4;
}
}
else if (Cs >= 936) {
return 4;
}
}
else if (Rs >= 1061) {
if (gchDC < 1) {
if (negBalance < 31) {
if (mu_mv_mag_sq < 79) {
return 1;
}
else if (mu_mv_mag_sq >= 79) {
return 8;
}
}
else if (negBalance >= 31) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Rs < 1430) {
return 4;
}
else if (Rs >= 1430) {
return 1;
}
}
}
}
else if (gchDC >= 1) {
if (ssDCval < 98) {
if (diffRsCsdiff < -29) {
return 4;
}
else if (diffRsCsdiff >= -29) {
if (Rs < 2192) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
if (SC < 2497) {
return 8;
}
else if (SC >= 2497) {
return 1;
}
}
}
else if (Rs >= 2192) {
return 4;
}
}
}
else if (ssDCval >= 98) {
if (diffRsCsdiff < -16) {
if (Rs < 1924) {
return 2;
}
else if (Rs >= 1924) {
return 1;
}
}
else if (diffRsCsdiff >= -16) {
return 4;
}
}
}
}
}
else if (ssDCval >= 104) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (negBalance < 33) {
if (Rs < 705) {
return 1;
}
else if (Rs >= 705) {
return 8;
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (negBalance < 37) {
if (Rs < 1189) {
return 8;
}
else if (Rs >= 1189) {
return 2;
}
}
else if (negBalance >= 37) {
return 1;
}
}
}
}
else if (ssDCval >= 117) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1103) {
return 4;
}
else if (Rs >= 1103) {
if (tcor < 93) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Cs < 1036) {
return 8;
}
else if (Cs >= 1036) {
return 2;
}
}
}
else if (tcor >= 93) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (refDCval < 134) {
if (Rs < 2368) {
if (negBalance < 26) {
if (Rs < 1881) {
return 2;
}
else if (Rs >= 1881) {
return 1;
}
}
else if (negBalance >= 26) {
if (Cs < 750) {
return 2;
}
else if (Cs >= 750) {
if (ssDCval < 133) {
return 8;
}
else if (ssDCval >= 133) {
if (Rs < 1631) {
return 2;
}
else if (Rs >= 1631) {
return 8;
}
}
}
}
}
else if (Rs >= 2368) {
if (refDCval < 129) {
return 1;
}
else if (refDCval >= 129) {
return 2;
}
}
}
else if (refDCval >= 134) {
if (posBalance < 47) {
if (TSC < 4) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (posBalance >= 47) {
if (diffAFD < -1) {
if (TSC < 6) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
if (RsCsDiff < 16) {
if (Rs < 1289) {
return 4;
}
else if (Rs >= 1289) {
return 8;
}
}
else if (RsCsDiff >= 16) {
return 1;
}
}
}
else if (TSC >= 6) {
if (SC < 1650) {
return 2;
}
else if (SC >= 1650) {
return 1;
}
}
}
else if (diffAFD >= -1) {
if (Rs < 2331) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 1;
}
}
else if (Rs >= 2331) {
return 1;
}
}
}
}
}
}
}
else if (diffRsCsdiff >= -6) {
if (Cs < 1041) {
if (MVDiff < 5) {
if (ssDCval < 97) {
if (SC < 2651) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (MVDiff < 2) {
if (mu_mv_mag_sq < 124) {
if (SC < 1643) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (Rs < 659) {
return 2;
}
else if (Rs >= 659) {
return 4;
}
}
}
else if (SC >= 1643) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (MVDiff >= 2) {
if (gchDC < 2) {
if (Rs < 850) {
return 1;
}
else if (Rs >= 850) {
if (Rs < 1458) {
return 8;
}
else if (Rs >= 1458) {
return 1;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (SC >= 2651) {
if (refDCval < 89) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 2011) {
return 8;
}
else if (Rs >= 2011) {
return 1;
}
}
}
}
else if (refDCval >= 89) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
}
}
else if (ssDCval >= 97) {
if (ssDCval < 122) {
if (negBalance < 46) {
if (diffTSC < -2) {
if (Rs < 1096) {
return 1;
}
else if (Rs >= 1096) {
return 8;
}
}
else if (diffTSC >= -2) {
if (diffRsCsdiff < -4) {
if (Cs < 839) {
return 8;
}
else if (Cs >= 839) {
return 1;
}
}
else if (diffRsCsdiff >= -4) {
return 8;
}
}
}
else if (negBalance >= 46) {
if (diffRsCsdiff < 1) {
if (diffTSC < -1) {
if (Rs < 716) {
return 8;
}
else if (Rs >= 716) {
return 1;
}
}
else if (diffTSC >= -1) {
if (Cs < 745) {
return 4;
}
else if (Cs >= 745) {
return 8;
}
}
}
else if (diffRsCsdiff >= 1) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
return 1;
}
}
}
}
else if (ssDCval >= 122) {
if (Cs < 640) {
if (ssDCval < 133) {
if (Rs < 834) {
return 2;
}
else if (Rs >= 834) {
return 1;
}
}
else if (ssDCval >= 133) {
return 8;
}
}
else if (Cs >= 640) {
if (ssDCval < 140) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (tcor < 93) {
if (Rs < 1706) {
return 4;
}
else if (Rs >= 1706) {
return 8;
}
}
else if (tcor >= 93) {
return 2;
}
}
}
else if (ssDCval >= 140) {
if (Rs < 710) {
if (Rs < 672) {
return 1;
}
else if (Rs >= 672) {
return 4;
}
}
else if (Rs >= 710) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 5) {
if (negBalance < 35) {
if (ssDCval < 131) {
if (mu_mv_mag_sq < 97) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (CsDiff < 3) {
if (Cs < 594) {
return 2;
}
else if (Cs >= 594) {
return 1;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 97) {
return 8;
}
}
else if (ssDCval >= 131) {
if (Rs < 1718) {
return 8;
}
else if (Rs >= 1718) {
return 4;
}
}
}
else if (negBalance >= 35) {
if (tcor < 94) {
if (negBalance < 50) {
if (RsDiff < 4) {
if (Rs < 880) {
if (Cs < 759) {
return 2;
}
else if (Cs >= 759) {
return 1;
}
}
else if (Rs >= 880) {
if (ssDCval < 128) {
return 8;
}
else if (ssDCval >= 128) {
if (mcTcor < 97) {
return 4;
}
else if (mcTcor >= 97) {
return 8;
}
}
}
}
else if (RsDiff >= 4) {
if (Rs < 754) {
return 4;
}
else if (Rs >= 754) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (Cs < 848) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (Cs >= 848) {
if (Rs < 1129) {
if (Rs < 951) {
return 4;
}
else if (Rs >= 951) {
return 2;
}
}
else if (Rs >= 1129) {
return 8;
}
}
}
}
else if (tcor >= 94) {
if (RsDiff < 5) {
if (posBalance < 45) {
if (Cs < 687) {
if (Cs < 560) {
return 8;
}
else if (Cs >= 560) {
return 1;
}
}
else if (Cs >= 687) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
return 4;
}
}
}
else if (posBalance >= 45) {
if (diffAFD < -2) {
if (Rs < 1165) {
return 2;
}
else if (Rs >= 1165) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (RsDiff >= 5) {
if (Rs < 2257) {
if (Rs < 1539) {
return 8;
}
else if (Rs >= 1539) {
return 2;
}
}
else if (Rs >= 2257) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1041) {
if (AFD < 12) {
if (mu_mv_mag_sq < 106) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (Rs < 1949) {
if (Rs < 1722) {
if (refDCval < 111) {
if (mu_mv_mag_sq < 75) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 75) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
}
else if (refDCval >= 111) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (Rs >= 1722) {
if (negBalance < 39) {
if (MVDiff < 5) {
if (Rs < 1882) {
return 2;
}
else if (Rs >= 1882) {
if (Rs < 1919) {
return 8;
}
else if (Rs >= 1919) {
return 4;
}
}
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (negBalance >= 39) {
return 8;
}
}
}
else if (Rs >= 1949) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 2343) {
return 8;
}
else if (Rs >= 2343) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
}
}
else if (AFD >= 12) {
return 4;
}
}
}
}
else if (ssDCval >= 154) {
if (Rs < 790) {
if (ssDCval < 171) {
return 4;
}
else if (ssDCval >= 171) {
return 2;
}
}
else if (Rs >= 790) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 131) {
if (Rs < 2155) {
if (AFD < 4) {
if (posBalance < 23) {
if (Rs < 1210) {
return 8;
}
else if (Rs >= 1210) {
return 4;
}
}
else if (posBalance >= 23) {
if (MVDiff < 6) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (Rs < 1955) {
return 1;
}
else if (Rs >= 1955) {
return 8;
}
}
}
else if (MVDiff >= 6) {
return 1;
}
}
}
else if (AFD >= 4) {
return 1;
}
}
else if (Rs >= 2155) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
return 8;
}
}
}
else if (ssDCval >= 131) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 43) {
if (Rs < 2117) {
if (refDCval < 41) {
if (Rs < 700) {
if (posBalance < 62) {
if (diffRsCsdiff < -5) {
return 1;
}
else if (diffRsCsdiff >= -5) {
if (Rs < 353) {
return 1;
}
else if (Rs >= 353) {
return 8;
}
}
}
else if (posBalance >= 62) {
return 2;
}
}
else if (Rs >= 700) {
if (Rs < 2008) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (ssDCval < 42) {
return 8;
}
else if (ssDCval >= 42) {
if (Rs < 1133) {
return 1;
}
else if (Rs >= 1133) {
return 8;
}
}
}
}
else if (Rs >= 2008) {
if (Rs < 2106) {
return 1;
}
else if (Rs >= 2106) {
return 8;
}
}
}
}
else if (refDCval >= 41) {
if (Cs < 868) {
if (tcor < 74) {
if (CsDiff < 5) {
if (MVDiff < 15) {
if (RsCsDiff < 62) {
return 8;
}
else if (RsCsDiff >= 62) {
if (tcor < 41) {
return 8;
}
else if (tcor >= 41) {
if (tcor < 51) {
return 4;
}
else if (tcor >= 51) {
if (Rs < 1176) {
return 4;
}
else if (Rs >= 1176) {
return 8;
}
}
}
}
}
else if (MVDiff >= 15) {
if (posBalance < 59) {
if (tcor < 57) {
if (SC < 2728) {
return 1;
}
else if (SC >= 2728) {
return 2;
}
}
else if (tcor >= 57) {
if (diffAFD < 2) {
if (Rs < 1030) {
if (Rs < 932) {
return 1;
}
else if (Rs >= 932) {
return 4;
}
}
else if (Rs >= 1030) {
return 8;
}
}
else if (diffAFD >= 2) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
if (Rs < 1741) {
return 2;
}
else if (Rs >= 1741) {
return 8;
}
}
}
}
}
else if (posBalance >= 59) {
if (Cs < 723) {
if (SC < 1929) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 924) {
return 1;
}
else if (Rs >= 924) {
return 4;
}
}
}
else if (SC >= 1929) {
if (mu_mv_mag_sq < 19) {
if (Cs < 555) {
return 4;
}
else if (Cs >= 555) {
return 8;
}
}
else if (mu_mv_mag_sq >= 19) {
if (MVDiff < 31) {
if (SC < 2641) {
return 8;
}
else if (SC >= 2641) {
return 1;
}
}
else if (MVDiff >= 31) {
if (MVDiff < 51) {
return 4;
}
else if (MVDiff >= 51) {
if (Rs < 1733) {
return 4;
}
else if (Rs >= 1733) {
return 8;
}
}
}
}
}
}
else if (Cs >= 723) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (Rs < 1489) {
return 1;
}
else if (Rs >= 1489) {
return 8;
}
}
}
}
}
}
else if (CsDiff >= 5) {
if (tcor < 57) {
if (MVDiff < 28) {
if (mu_mv_mag_sq < 43) {
if (Cs < 853) {
if (negBalance < 51) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
return 2;
}
}
else if (negBalance >= 51) {
return 4;
}
}
else if (Cs >= 853) {
return 8;
}
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 726) {
return 2;
}
else if (Rs >= 726) {
if (Rs < 941) {
return 8;
}
else if (Rs >= 941) {
return 4;
}
}
}
}
else if (MVDiff >= 28) {
if (TSC < 10) {
if (refDCval < 45) {
if (Cs < 721) {
return 2;
}
else if (Cs >= 721) {
return 1;
}
}
else if (refDCval >= 45) {
if (Rs < 662) {
if (diffRsCsdiff < -32) {
if (diffTSC < -1) {
if (Rs < 462) {
return 8;
}
else if (Rs >= 462) {
return 1;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (diffRsCsdiff >= -32) {
return 4;
}
}
else if (Rs >= 662) {
if (negBalance < 56) {
if (gchDC < 2) {
if (diffAFD < 0) {
if (Rs < 863) {
return 8;
}
else if (Rs >= 863) {
if (Rs < 1371) {
return 2;
}
else if (Rs >= 1371) {
return 4;
}
}
}
else if (diffAFD >= 0) {
return 1;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (negBalance >= 56) {
if (negBalance < 60) {
if (gchDC < 1) {
if (Rs < 817) {
return 4;
}
else if (Rs >= 817) {
return 8;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (negBalance >= 60) {
return 8;
}
}
}
}
}
else if (TSC >= 10) {
if (mcTcor < 73) {
if (tcor < 26) {
if (SC < 1055) {
if (Rs < 201) {
return 4;
}
else if (Rs >= 201) {
return 8;
}
}
else if (SC >= 1055) {
return 4;
}
}
else if (tcor >= 26) {
if (mu_mv_mag_sq < 2) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
}
else if (mcTcor >= 73) {
if (ssDCval < 91) {
if (Cs < 703) {
if (Rs < 808) {
return 8;
}
else if (Rs >= 808) {
if (Rs < 1101) {
return 4;
}
else if (Rs >= 1101) {
return 2;
}
}
}
else if (Cs >= 703) {
return 2;
}
}
else if (ssDCval >= 91) {
if (TSC < 12) {
return 1;
}
else if (TSC >= 12) {
if (Rs < 1800) {
return 8;
}
else if (Rs >= 1800) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 57) {
if (Rs < 1472) {
if (SC < 1324) {
if (MVDiff < 37) {
if (tcor < 69) {
return 2;
}
else if (tcor >= 69) {
if (SC < 1195) {
return 8;
}
else if (SC >= 1195) {
return 4;
}
}
}
else if (MVDiff >= 37) {
if (diffAFD < 4) {
if (Cs < 845) {
if (diffAFD < -2) {
if (Rs < 474) {
return 1;
}
else if (Rs >= 474) {
return 2;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (Cs >= 845) {
return 8;
}
}
else if (diffAFD >= 4) {
if (Rs < 297) {
return 4;
}
else if (Rs >= 297) {
if (Rs < 569) {
return 2;
}
else if (Rs >= 569) {
return 8;
}
}
}
}
}
else if (SC >= 1324) {
if (negBalance < 48) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Cs < 824) {
if (Rs < 979) {
return 2;
}
else if (Rs >= 979) {
return 4;
}
}
else if (Cs >= 824) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (diffMVdiffVal < 34) {
if (RsDiff < 13) {
if (diffAFD < 2) {
if (Rs < 839) {
if (TSC < 8) {
if (posBalance < 63) {
return 2;
}
else if (posBalance >= 63) {
return 1;
}
}
else if (TSC >= 8) {
if (Cs < 778) {
return 8;
}
else if (Cs >= 778) {
return 4;
}
}
}
else if (Rs >= 839) {
if (negBalance < 52) {
if (refDCval < 60) {
if (Rs < 1179) {
return 4;
}
else if (Rs >= 1179) {
return 8;
}
}
else if (refDCval >= 60) {
if (Cs < 649) {
return 8;
}
else if (Cs >= 649) {
return 1;
}
}
}
else if (negBalance >= 52) {
if (mu_mv_mag_sq < 15) {
if (diffAFD < -2) {
if (Cs < 621) {
return 4;
}
else if (Cs >= 621) {
return 8;
}
}
else if (diffAFD >= -2) {
if (MVDiff < 64) {
return 4;
}
else if (MVDiff >= 64) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 4;
}
}
}
}
}
else if (diffAFD >= 2) {
if (posBalance < 60) {
if (gchDC < 2) {
if (diffRsCsdiff < -10) {
return 4;
}
else if (diffRsCsdiff >= -10) {
return 2;
}
}
else if (gchDC >= 2) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
}
else if (posBalance >= 60) {
if (Rs < 804) {
return 2;
}
else if (Rs >= 804) {
return 1;
}
}
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (diffMVdiffVal >= 34) {
return 8;
}
}
}
}
else if (Rs >= 1472) {
if (TSC < 9) {
if (Rs < 1516) {
if (Cs < 707) {
return 2;
}
else if (Cs >= 707) {
if (Rs < 1504) {
return 8;
}
else if (Rs >= 1504) {
return 4;
}
}
}
else if (Rs >= 1516) {
return 2;
}
}
else if (TSC >= 9) {
if (posBalance < 59) {
if (Rs < 1956) {
return 8;
}
else if (Rs >= 1956) {
return 2;
}
}
else if (posBalance >= 59) {
if (AFD < 28) {
return 4;
}
else if (AFD >= 28) {
if (Rs < 1823) {
return 2;
}
else if (Rs >= 1823) {
return 8;
}
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (refDCval < 90) {
if (TSC < 7) {
if (ssDCval < 72) {
if (posBalance < 46) {
if (refDCval < 54) {
if (Rs < 1726) {
if (Rs < 971) {
return 8;
}
else if (Rs >= 971) {
return 1;
}
}
else if (Rs >= 1726) {
if (Cs < 829) {
return 4;
}
else if (Cs >= 829) {
return 1;
}
}
}
else if (refDCval >= 54) {
if (AFD < 9) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
if (Rs < 1580) {
return 8;
}
else if (Rs >= 1580) {
return 2;
}
}
}
else if (AFD >= 9) {
if (negBalance < 45) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
return 2;
}
}
else if (negBalance >= 45) {
if (CsDiff < 5) {
if (Rs < 1415) {
return 2;
}
else if (Rs >= 1415) {
return 1;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
}
else if (posBalance >= 46) {
if (mcTcor < 96) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (tcor < 90) {
if (gchDC < 2) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
return 8;
}
}
else if (gchDC >= 2) {
if (Rs < 1738) {
return 2;
}
else if (Rs >= 1738) {
return 4;
}
}
}
else if (tcor >= 90) {
if (Rs < 1646) {
return 4;
}
else if (Rs >= 1646) {
return 2;
}
}
}
}
else if (mcTcor >= 96) {
if (posBalance < 68) {
return 2;
}
else if (posBalance >= 68) {
return 1;
}
}
}
}
else if (ssDCval >= 72) {
if (Rs < 1450) {
if (negBalance < 49) {
if (mu_mv_mag_sq < 29) {
return 1;
}
else if (mu_mv_mag_sq >= 29) {
if (Rs < 790) {
return 4;
}
else if (Rs >= 790) {
return 8;
}
}
}
else if (negBalance >= 49) {
return 2;
}
}
else if (Rs >= 1450) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (RsCsDiff < 83) {
if (MVDiff < 14) {
if (SC < 2449) {
if (AFD < 18) {
if (Cs < 593) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
else if (Cs >= 593) {
return 8;
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (SC >= 2449) {
if (Cs < 838) {
if (SC < 2906) {
return 2;
}
else if (SC >= 2906) {
return 8;
}
}
else if (Cs >= 838) {
if (Rs < 1828) {
return 8;
}
else if (Rs >= 1828) {
return 4;
}
}
}
}
else if (MVDiff >= 14) {
if (gchDC < 2) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (diffAFD < -1) {
if (diffAFD < -3) {
if (mu_mv_mag_sq < 10) {
return 4;
}
else if (mu_mv_mag_sq >= 10) {
if (Rs < 1396) {
return 1;
}
else if (Rs >= 1396) {
return 8;
}
}
}
else if (diffAFD >= -3) {
if (Cs < 633) {
if (Rs < 1046) {
return 8;
}
else if (Rs >= 1046) {
if (Rs < 1340) {
return 4;
}
else if (Rs >= 1340) {
return 1;
}
}
}
else if (Cs >= 633) {
return 2;
}
}
}
else if (diffAFD >= -1) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
if (Rs < 1162) {
return 1;
}
else if (Rs >= 1162) {
return 4;
}
}
}
}
}
else if (gchDC >= 2) {
if (gchDC < 3) {
if (posBalance < 54) {
if (RsCsDiff < 69) {
return 4;
}
else if (RsCsDiff >= 69) {
return 2;
}
}
else if (posBalance >= 54) {
if (AFD < 18) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (AFD >= 18) {
return 4;
}
}
}
else if (gchDC >= 3) {
if (mcTcor < 93) {
if (mu_mv_mag_sq < 7) {
if (Rs < 941) {
return 1;
}
else if (Rs >= 941) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
else if (mcTcor >= 93) {
if (diffRsCsdiff < 6) {
if (RsCsDiff < 67) {
if (mcTcor < 94) {
if (Cs < 803) {
return 4;
}
else if (Cs >= 803) {
return 2;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (RsCsDiff >= 67) {
return 1;
}
}
else if (diffRsCsdiff >= 6) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 83) {
if (refDCval < 86) {
if (mcTcor < 90) {
if (diffAFD < 2) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 4;
}
}
else if (diffAFD >= 2) {
if (Cs < 608) {
return 2;
}
else if (Cs >= 608) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (gchDC < 3) {
if (Cs < 837) {
return 2;
}
else if (Cs >= 837) {
return 4;
}
}
else if (gchDC >= 3) {
if (SC < 2404) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
if (Cs < 801) {
return 8;
}
else if (Cs >= 801) {
return 2;
}
}
}
else if (SC >= 2404) {
if (Rs < 2093) {
return 2;
}
else if (Rs >= 2093) {
return 8;
}
}
}
}
}
else if (refDCval >= 86) {
if (gchDC < 2) {
if (Rs < 1084) {
return 2;
}
else if (Rs >= 1084) {
return 4;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
}
else if (refDCval >= 90) {
if (Rs < 1379) {
if (RsCsDiff < 73) {
if (MVDiff < 34) {
if (ssDCval < 113) {
if (Rs < 1065) {
if (Rs < 956) {
return 8;
}
else if (Rs >= 956) {
return 4;
}
}
else if (Rs >= 1065) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
}
else if (ssDCval >= 113) {
return 8;
}
}
else if (MVDiff >= 34) {
if (negBalance < 50) {
if (Rs < 1190) {
return 2;
}
else if (Rs >= 1190) {
return 4;
}
}
else if (negBalance >= 50) {
if (tcor < 86) {
return 1;
}
else if (tcor >= 86) {
if (Cs < 594) {
return 1;
}
else if (Cs >= 594) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 73) {
if (tcor < 75) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (RsCsDiff < 123) {
return 1;
}
else if (RsCsDiff >= 123) {
return 2;
}
}
}
else if (tcor >= 75) {
if (diffAFD < 3) {
if (diffRsCsdiff < -20) {
if (diffAFD < -1) {
if (Rs < 1315) {
return 1;
}
else if (Rs >= 1315) {
return 8;
}
}
else if (diffAFD >= -1) {
if (Cs < 763) {
return 2;
}
else if (Cs >= 763) {
return 8;
}
}
}
else if (diffRsCsdiff >= -20) {
if (ssDCval < 97) {
if (ssDCval < 94) {
if (Rs < 464) {
return 8;
}
else if (Rs >= 464) {
return 2;
}
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (ssDCval >= 97) {
if (posBalance < 64) {
if (AFD < 15) {
if (Rs < 860) {
return 2;
}
else if (Rs >= 860) {
return 1;
}
}
else if (AFD >= 15) {
return 2;
}
}
else if (posBalance >= 64) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
}
}
else if (diffAFD >= 3) {
if (Cs < 606) {
return 1;
}
else if (Cs >= 606) {
return 8;
}
}
}
}
}
else if (Rs >= 1379) {
if (TSC < 5) {
if (Rs < 1764) {
return 8;
}
else if (Rs >= 1764) {
return 1;
}
}
else if (TSC >= 5) {
if (MVDiff < 42) {
if (TSC < 9) {
if (mcTcor < 94) {
if (ssDCval < 127) {
return 2;
}
else if (ssDCval >= 127) {
if (Rs < 1641) {
return 8;
}
else if (Rs >= 1641) {
return 1;
}
}
}
else if (mcTcor >= 94) {
if (SC < 2289) {
if (mu_mv_mag_sq < 10) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 2;
}
}
else if (mu_mv_mag_sq >= 10) {
return 8;
}
}
else if (SC >= 2289) {
if (Rs < 1944) {
if (tcor < 90) {
if (Cs < 707) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (Cs >= 707) {
if (Cs < 823) {
return 2;
}
else if (Cs >= 823) {
return 1;
}
}
}
else if (tcor >= 90) {
if (Rs < 1910) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
if (Rs < 1821) {
return 2;
}
else if (Rs >= 1821) {
return 1;
}
}
}
else if (Rs >= 1910) {
return 2;
}
}
}
else if (Rs >= 1944) {
if (Rs < 2093) {
return 1;
}
else if (Rs >= 2093) {
return 4;
}
}
}
}
}
else if (TSC >= 9) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
if (Rs < 1524) {
return 8;
}
else if (Rs >= 1524) {
return 2;
}
}
}
}
else if (MVDiff >= 42) {
if (Cs < 606) {
return 1;
}
else if (Cs >= 606) {
if (negBalance < 49) {
if (Rs < 2009) {
return 1;
}
else if (Rs >= 2009) {
return 2;
}
}
else if (negBalance >= 49) {
if (AFD < 14) {
if (Rs < 1679) {
return 8;
}
else if (Rs >= 1679) {
if (Rs < 1931) {
return 2;
}
else if (Rs >= 1931) {
return 1;
}
}
}
else if (AFD >= 14) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 868) {
if (AFD < 19) {
if (Rs < 1299) {
if (ssDCval < 63) {
if (negBalance < 20) {
if (Rs < 570) {
return 1;
}
else if (Rs >= 570) {
return 2;
}
}
else if (negBalance >= 20) {
if (MVDiff < 18) {
if (gchDC < 2) {
if (ssDCval < 47) {
if (Rs < 1068) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 4;
}
}
else if (Rs >= 1068) {
return 8;
}
}
else if (ssDCval >= 47) {
if (ssDCval < 61) {
if (tcor < 56) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (tcor >= 56) {
return 4;
}
}
else if (ssDCval >= 61) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (MVDiff >= 18) {
if (mu_mv_mag_sq < 21) {
if (ssDCval < 57) {
if (TSC < 8) {
if (MVDiff < 33) {
if (Cs < 913) {
return 8;
}
else if (Cs >= 913) {
return 2;
}
}
else if (MVDiff >= 33) {
if (Rs < 655) {
return 4;
}
else if (Rs >= 655) {
if (Rs < 778) {
return 8;
}
else if (Rs >= 778) {
return 2;
}
}
}
}
else if (TSC >= 8) {
if (gchDC < 4) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 2;
}
}
else if (gchDC >= 4) {
if (Rs < 444) {
return 2;
}
else if (Rs >= 444) {
return 8;
}
}
}
}
else if (ssDCval >= 57) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (negBalance < 54) {
if (mcTcor < 69) {
if (Rs < 876) {
return 2;
}
else if (Rs >= 876) {
if (Rs < 1139) {
return 8;
}
else if (Rs >= 1139) {
return 1;
}
}
}
else if (mcTcor >= 69) {
return 8;
}
}
else if (negBalance >= 54) {
if (Rs < 861) {
return 4;
}
else if (Rs >= 861) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 63) {
if (negBalance < 36) {
if (ssDCval < 137) {
if (SC < 2187) {
if (SC < 2035) {
if (RsCsDiff < 59) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (RsCsDiff >= 59) {
return 1;
}
}
else if (SC >= 2035) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 1;
}
}
}
else if (SC >= 2187) {
if (Rs < 1271) {
return 1;
}
else if (Rs >= 1271) {
return 2;
}
}
}
else if (ssDCval >= 137) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (tcor < 68) {
if (Rs < 1086) {
if (RsCsDiff < 52) {
return 8;
}
else if (RsCsDiff >= 52) {
return 4;
}
}
else if (Rs >= 1086) {
if (tcor < 66) {
return 8;
}
else if (tcor >= 66) {
if (Rs < 1271) {
return 4;
}
else if (Rs >= 1271) {
return 8;
}
}
}
}
else if (tcor >= 68) {
if (mu_mv_mag_sq < 3) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (MVDiff < 8) {
if (diffRsCsdiff < 1) {
if (Rs < 1007) {
return 1;
}
else if (Rs >= 1007) {
return 8;
}
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
else if (MVDiff >= 8) {
if (refDCval < 84) {
if (TSC < 9) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
if (negBalance < 51) {
if (diffMVdiffVal < 3) {
if (SC < 2033) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 1;
}
}
else if (SC >= 2033) {
if (Rs < 1056) {
return 1;
}
else if (Rs >= 1056) {
return 2;
}
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 913) {
return 4;
}
else if (Rs >= 913) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (Rs < 1121) {
return 8;
}
else if (Rs >= 1121) {
return 4;
}
}
}
}
else if (TSC >= 9) {
return 2;
}
}
else if (refDCval >= 84) {
if (negBalance < 55) {
if (diffRsCsdiff < -23) {
return 2;
}
else if (diffRsCsdiff >= -23) {
if (tcor < 78) {
if (Rs < 971) {
return 4;
}
else if (Rs >= 971) {
return 8;
}
}
else if (tcor >= 78) {
return 1;
}
}
}
else if (negBalance >= 55) {
if (diffRsCsdiff < 7) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1197) {
return 4;
}
else if (Rs >= 1197) {
return 2;
}
}
}
else if (diffRsCsdiff >= 7) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (refDCval < 81) {
if (Rs < 818) {
if (Rs < 744) {
return 2;
}
else if (Rs >= 744) {
if (SC < 2023) {
return 1;
}
else if (SC >= 2023) {
return 4;
}
}
}
else if (Rs >= 818) {
if (MVDiff < 15) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
else if (MVDiff >= 15) {
if (Cs < 931) {
return 8;
}
else if (Cs >= 931) {
if (refDCval < 67) {
if (Rs < 1051) {
return 2;
}
else if (Rs >= 1051) {
return 8;
}
}
else if (refDCval >= 67) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 81) {
if (tcor < 91) {
if (tcor < 84) {
if (diffRsCsdiff < 27) {
if (mu_mv_mag_sq < 47) {
return 8;
}
else if (mu_mv_mag_sq >= 47) {
if (Cs < 1220) {
return 2;
}
else if (Cs >= 1220) {
return 8;
}
}
}
else if (diffRsCsdiff >= 27) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (tcor >= 84) {
if (mu_mv_mag_sq < 7) {
return 2;
}
else if (mu_mv_mag_sq >= 7) {
if (ssDCval < 156) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (ssDCval >= 156) {
return 2;
}
}
}
}
else if (tcor >= 91) {
if (Cs < 1198) {
return 8;
}
else if (Cs >= 1198) {
if (Rs < 1134) {
return 4;
}
else if (Rs >= 1134) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1299) {
if (TSC < 10) {
if (refDCval < 111) {
if (negBalance < 41) {
if (MVDiff < 10) {
if (CsDiff < 5) {
if (mcTcor < 87) {
return 2;
}
else if (mcTcor >= 87) {
if (ssDCval < 62) {
if (Cs < 1198) {
return 4;
}
else if (Cs >= 1198) {
return 8;
}
}
else if (ssDCval >= 62) {
if (ssDCval < 80) {
if (Rs < 1880) {
return 1;
}
else if (Rs >= 1880) {
return 2;
}
}
else if (ssDCval >= 80) {
return 4;
}
}
}
}
else if (CsDiff >= 5) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (refDCval < 84) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
if (Cs < 899) {
return 1;
}
else if (Cs >= 899) {
if (RsCsDiff < 56) {
if (Rs < 1664) {
return 8;
}
else if (Rs >= 1664) {
return 4;
}
}
else if (RsCsDiff >= 56) {
return 8;
}
}
}
}
else if (refDCval >= 84) {
if (Rs < 1813) {
return 2;
}
else if (Rs >= 1813) {
return 4;
}
}
}
}
}
else if (MVDiff >= 10) {
if (diffRsCsdiff < -36) {
if (RsDiff < 7) {
if (tcor < 84) {
if (Rs < 1791) {
return 2;
}
else if (Rs >= 1791) {
return 8;
}
}
else if (tcor >= 84) {
return 1;
}
}
else if (RsDiff >= 7) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
}
else if (diffRsCsdiff >= -36) {
if (negBalance < 40) {
if (RsDiff < 10) {
if (mu_mv_mag_sq < 57) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 2;
}
}
}
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (negBalance >= 40) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 2;
}
}
}
}
}
else if (negBalance >= 41) {
if (RsDiff < 8) {
if (ssDCval < 62) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (MVDiff < 6) {
if (Cs < 1096) {
return 8;
}
else if (Cs >= 1096) {
return 4;
}
}
else if (MVDiff >= 6) {
if (Cs < 917) {
return 4;
}
else if (Cs >= 917) {
if (TSC < 7) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 1;
}
}
else if (TSC >= 7) {
if (Cs < 1294) {
return 8;
}
else if (Cs >= 1294) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 62) {
if (RsCsDiff < 48) {
if (Rs < 1628) {
if (Rs < 1520) {
return 1;
}
else if (Rs >= 1520) {
return 8;
}
}
else if (Rs >= 1628) {
return 2;
}
}
else if (RsCsDiff >= 48) {
if (MVDiff < 3) {
if (diffRsCsdiff < 28) {
return 8;
}
else if (diffRsCsdiff >= 28) {
if (Rs < 1882) {
return 8;
}
else if (Rs >= 1882) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (CsDiff < 5) {
if (diffMVdiffVal < 3) {
if (ssDCval < 94) {
if (ssDCval < 91) {
if (negBalance < 49) {
if (AFD < 10) {
if (Cs < 1167) {
return 4;
}
else if (Cs >= 1167) {
return 8;
}
}
else if (AFD >= 10) {
return 8;
}
}
else if (negBalance >= 49) {
if (Cs < 1023) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
return 2;
}
}
else if (Cs >= 1023) {
return 8;
}
}
}
else if (ssDCval >= 91) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
if (MVDiff < 27) {
if (Cs < 954) {
return 8;
}
else if (Cs >= 954) {
return 4;
}
}
else if (MVDiff >= 27) {
if (Rs < 1801) {
return 8;
}
else if (Rs >= 1801) {
return 2;
}
}
}
}
else if (CsDiff >= 5) {
if (Cs < 1044) {
if (Rs < 1486) {
return 8;
}
else if (Rs >= 1486) {
if (ssDCval < 77) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (ssDCval >= 77) {
if (ssDCval < 94) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (ssDCval >= 94) {
return 8;
}
}
}
}
else if (Cs >= 1044) {
if (Cs < 1204) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Cs < 1073) {
return 1;
}
else if (Cs >= 1073) {
return 8;
}
}
}
else if (Cs >= 1204) {
if (ssDCval < 87) {
if (Rs < 1351) {
return 4;
}
else if (Rs >= 1351) {
return 8;
}
}
else if (ssDCval >= 87) {
if (Rs < 1554) {
return 2;
}
else if (Rs >= 1554) {
if (diffAFD < 2) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Rs < 1737) {
return 2;
}
else if (Rs >= 1737) {
return 8;
}
}
}
else if (diffAFD >= 2) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (Cs < 1133) {
if (Rs < 1836) {
if (TSC < 8) {
if (MVDiff < 11) {
if (Rs < 1749) {
return 4;
}
else if (Rs >= 1749) {
return 8;
}
}
else if (MVDiff >= 11) {
return 2;
}
}
else if (TSC >= 8) {
if (diffMVdiffVal < 2) {
if (tcor < 71) {
return 4;
}
else if (tcor >= 71) {
if (Rs < 1430) {
return 4;
}
else if (Rs >= 1430) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
if (diffRsCsdiff < 27) {
return 1;
}
else if (diffRsCsdiff >= 27) {
if (Rs < 1565) {
return 8;
}
else if (Rs >= 1565) {
return 2;
}
}
}
}
}
else if (Rs >= 1836) {
if (gchDC < 4) {
if (Cs < 1105) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (ssDCval < 60) {
return 2;
}
else if (ssDCval >= 60) {
if (diffAFD < -7) {
return 8;
}
else if (diffAFD >= -7) {
return 4;
}
}
}
}
else if (Cs >= 1105) {
if (Cs < 1129) {
return 2;
}
else if (Cs >= 1129) {
return 4;
}
}
}
else if (gchDC >= 4) {
if (Rs < 2023) {
return 8;
}
else if (Rs >= 2023) {
return 4;
}
}
}
}
else if (Cs >= 1133) {
if (refDCval < 79) {
if (MVDiff < 23) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1809) {
return 4;
}
else if (Rs >= 1809) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
}
else if (MVDiff >= 23) {
if (MVDiff < 25) {
if (Rs < 1538) {
return 8;
}
else if (Rs >= 1538) {
return 1;
}
}
else if (MVDiff >= 25) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
}
}
else if (refDCval >= 79) {
if (tcor < 81) {
if (Rs < 1988) {
return 2;
}
else if (Rs >= 1988) {
return 4;
}
}
else if (tcor >= 81) {
if (MVDiff < 8) {
if (Rs < 1928) {
return 1;
}
else if (Rs >= 1928) {
return 4;
}
}
else if (MVDiff >= 8) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 111) {
if (mu_mv_mag_sq < 99) {
if (RsCsDiff < 156) {
if (MVDiff < 6) {
if (mcTcor < 95) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
else if (mcTcor >= 95) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (refDCval < 166) {
if (Cs < 883) {
if (Cs < 876) {
return 8;
}
else if (Cs >= 876) {
return 2;
}
}
else if (Cs >= 883) {
return 8;
}
}
else if (refDCval >= 166) {
return 2;
}
}
}
else if (RsCsDiff >= 156) {
if (negBalance < 48) {
if (Rs < 1446) {
return 8;
}
else if (Rs >= 1446) {
return 2;
}
}
else if (negBalance >= 48) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
return 2;
}
}
}
}
else if (TSC >= 10) {
if (ssDCval < 96) {
if (Rs < 1406) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 1377) {
return 4;
}
else if (Rs >= 1377) {
return 2;
}
}
}
else if (Rs >= 1406) {
return 8;
}
}
else if (ssDCval >= 96) {
if (Cs < 1091) {
if (MVDiff < 47) {
return 1;
}
else if (MVDiff >= 47) {
return 4;
}
}
else if (Cs >= 1091) {
if (Rs < 1857) {
if (posBalance < 61) {
if (Rs < 1567) {
if (Rs < 1383) {
return 8;
}
else if (Rs >= 1383) {
return 1;
}
}
else if (Rs >= 1567) {
if (gchDC < 1) {
if (Rs < 1694) {
return 8;
}
else if (Rs >= 1694) {
return 1;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (mu_mv_mag_sq < 11) {
if (Rs < 1632) {
return 8;
}
else if (Rs >= 1632) {
return 4;
}
}
else if (mu_mv_mag_sq >= 11) {
return 1;
}
}
}
else if (Rs >= 1857) {
if (tcor < 75) {
return 4;
}
else if (tcor >= 75) {
if (TSC < 11) {
if (Rs < 1951) {
return 1;
}
else if (Rs >= 1951) {
return 8;
}
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
}
}
}
else if (AFD >= 19) {
if (RsDiff < 9) {
if (ssDCval < 87) {
if (MVDiff < 55) {
if (mu_mv_mag_sq < 27) {
if (SC < 2391) {
if (ssDCval < 70) {
if (TSC < 9) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (TSC >= 9) {
if (Rs < 1218) {
if (diffTSC < 2) {
return 4;
}
else if (diffTSC >= 2) {
return 1;
}
}
else if (Rs >= 1218) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (negBalance < 58) {
if (RsDiff < 8) {
if (mu_mv_mag_sq < 5) {
return 2;
}
else if (mu_mv_mag_sq >= 5) {
if (Cs < 1307) {
return 4;
}
else if (Cs >= 1307) {
return 2;
}
}
}
else if (RsDiff >= 8) {
if (Rs < 1071) {
return 1;
}
else if (Rs >= 1071) {
if (Rs < 1220) {
return 4;
}
else if (Rs >= 1220) {
return 8;
}
}
}
}
else if (negBalance >= 58) {
if (Cs < 948) {
return 8;
}
else if (Cs >= 948) {
return 2;
}
}
}
}
else if (SC >= 2391) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (Cs < 983) {
if (Rs < 1120) {
return 4;
}
else if (Rs >= 1120) {
return 2;
}
}
else if (Cs >= 983) {
return 2;
}
}
}
else if (MVDiff >= 55) {
if (Rs < 906) {
if (AFD < 27) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
if (MVDiff < 67) {
return 8;
}
else if (MVDiff >= 67) {
return 4;
}
}
}
else if (AFD >= 27) {
if (Cs < 943) {
return 1;
}
else if (Cs >= 943) {
if (Rs < 644) {
return 8;
}
else if (Rs >= 644) {
return 2;
}
}
}
}
else if (Rs >= 906) {
if (tcor < 60) {
if (MVDiff < 62) {
if (Rs < 1184) {
return 4;
}
else if (Rs >= 1184) {
return 8;
}
}
else if (MVDiff >= 62) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
}
else if (tcor >= 60) {
if (Rs < 1442) {
return 8;
}
else if (Rs >= 1442) {
if (Rs < 1900) {
return 1;
}
else if (Rs >= 1900) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (TSC < 10) {
if (TSC < 9) {
if (refDCval < 95) {
if (diffAFD < 4) {
return 1;
}
else if (diffAFD >= 4) {
if (Rs < 831) {
return 1;
}
else if (Rs >= 831) {
return 2;
}
}
}
else if (refDCval >= 95) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (MVDiff < 23) {
if (Rs < 829) {
return 1;
}
else if (Rs >= 829) {
return 8;
}
}
else if (MVDiff >= 23) {
return 4;
}
}
}
}
else if (TSC >= 9) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (ssDCval < 96) {
if (RsDiff < 7) {
if (Rs < 1255) {
return 4;
}
else if (Rs >= 1255) {
return 1;
}
}
else if (RsDiff >= 7) {
return 8;
}
}
else if (ssDCval >= 96) {
return 1;
}
}
}
}
else if (TSC >= 10) {
if (negBalance < 58) {
if (ssDCval < 129) {
if (Rs < 1198) {
if (Cs < 1030) {
return 2;
}
else if (Cs >= 1030) {
if (Rs < 1017) {
return 1;
}
else if (Rs >= 1017) {
return 2;
}
}
}
else if (Rs >= 1198) {
return 1;
}
}
else if (ssDCval >= 129) {
if (Rs < 1219) {
if (Rs < 1052) {
return 1;
}
else if (Rs >= 1052) {
return 8;
}
}
else if (Rs >= 1219) {
return 4;
}
}
}
else if (negBalance >= 58) {
if (SC < 1591) {
if (MVDiff < 118) {
return 1;
}
else if (MVDiff >= 118) {
if (Rs < 490) {
return 8;
}
else if (Rs >= 490) {
return 4;
}
}
}
else if (SC >= 1591) {
if (Cs < 1050) {
if (Rs < 1022) {
return 2;
}
else if (Rs >= 1022) {
if (Cs < 982) {
return 2;
}
else if (Cs >= 982) {
return 8;
}
}
}
else if (Cs >= 1050) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 121) {
if (gchDC < 4) {
if (AFD < 23) {
if (RsCsDiff < 163) {
if (diffMVdiffVal < 2) {
if (SC < 2785) {
if (posBalance < 65) {
if (Cs < 978) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
return 8;
}
}
else if (Cs >= 978) {
return 4;
}
}
else if (posBalance >= 65) {
return 8;
}
}
else if (SC >= 2785) {
if (negBalance < 50) {
if (Rs < 1933) {
return 4;
}
else if (Rs >= 1933) {
return 2;
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
return 8;
}
}
else if (RsCsDiff >= 163) {
if (ssDCval < 78) {
if (RsCsDiff < 189) {
if (Rs < 1916) {
return 4;
}
else if (Rs >= 1916) {
return 8;
}
}
else if (RsCsDiff >= 189) {
return 8;
}
}
else if (ssDCval >= 78) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
}
else if (AFD >= 23) {
if (RsCsDiff < 264) {
if (tcor < 54) {
if (MVDiff < 26) {
return 4;
}
else if (MVDiff >= 26) {
if (posBalance < 53) {
if (Cs < 1155) {
return 2;
}
else if (Cs >= 1155) {
return 4;
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
else if (tcor >= 54) {
if (SC < 2065) {
return 2;
}
else if (SC >= 2065) {
if (RsCsDiff < 124) {
if (Rs < 1540) {
return 4;
}
else if (Rs >= 1540) {
return 1;
}
}
else if (RsCsDiff >= 124) {
if (diffRsCsdiff < 24) {
if (Cs < 954) {
return 8;
}
else if (Cs >= 954) {
return 4;
}
}
else if (diffRsCsdiff >= 24) {
if (MVDiff < 29) {
return 8;
}
else if (MVDiff >= 29) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 264) {
if (Cs < 1162) {
return 2;
}
else if (Cs >= 1162) {
return 8;
}
}
}
}
else if (gchDC >= 4) {
if (RsDiff < 12) {
if (SC < 2594) {
if (Cs < 906) {
return 2;
}
else if (Cs >= 906) {
if (diffAFD < 4) {
if (posBalance < 56) {
if (MVDiff < 161) {
return 4;
}
else if (MVDiff >= 161) {
if (Rs < 1081) {
return 1;
}
else if (Rs >= 1081) {
return 8;
}
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (diffAFD >= 4) {
return 1;
}
}
}
else if (SC >= 2594) {
if (diffRsCsdiff < -102) {
return 2;
}
else if (diffRsCsdiff >= -102) {
return 8;
}
}
}
else if (RsDiff >= 12) {
if (mcTcor < 84) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 4;
}
}
else if (mcTcor >= 84) {
if (gchDC < 5) {
return 4;
}
else if (gchDC >= 5) {
return 2;
}
}
}
}
}
else if (refDCval >= 121) {
if (mcTcor < 94) {
if (ssDCval < 133) {
if (diffTSC < -2) {
if (Rs < 1537) {
return 8;
}
else if (Rs >= 1537) {
if (Rs < 1885) {
return 4;
}
else if (Rs >= 1885) {
return 2;
}
}
}
else if (diffTSC >= -2) {
return 1;
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 12) {
return 4;
}
else if (mu_mv_mag_sq >= 12) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 1431) {
return 8;
}
else if (Rs >= 1431) {
if (CsDiff < 8) {
if (TSC < 9) {
if (tcor < 84) {
return 2;
}
else if (tcor >= 84) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (CsDiff >= 8) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2117) {
if (ssDCval < 77) {
if (mcTcor < 92) {
if (RsCsDiff < 117) {
if (ssDCval < 69) {
if (diffMVdiffVal < 4) {
if (TSC < 6) {
if (Rs < 2158) {
if (Rs < 2144) {
return 8;
}
else if (Rs >= 2144) {
return 2;
}
}
else if (Rs >= 2158) {
return 4;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (diffMVdiffVal >= 4) {
if (ssDCval < 57) {
if (tcor < 75) {
if (Rs < 2373) {
return 2;
}
else if (Rs >= 2373) {
return 4;
}
}
else if (tcor >= 75) {
return 4;
}
}
else if (ssDCval >= 57) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
else if (ssDCval >= 69) {
if (MVDiff < 29) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (MVDiff >= 29) {
return 8;
}
}
}
else if (RsCsDiff >= 117) {
if (Cs < 1053) {
if (TSC < 10) {
if (mcTcor < 85) {
if (RsCsDiff < 246) {
if (Rs < 2215) {
return 8;
}
else if (Rs >= 2215) {
return 4;
}
}
else if (RsCsDiff >= 246) {
return 2;
}
}
else if (mcTcor >= 85) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
return 2;
}
}
}
else if (TSC >= 10) {
if (mu_mv_mag_sq < 5) {
if (Rs < 2295) {
return 4;
}
else if (Rs >= 2295) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
if (refDCval < 76) {
return 1;
}
else if (refDCval >= 76) {
return 8;
}
}
}
}
else if (Cs >= 1053) {
if (mu_mv_mag_sq < 16) {
if (AFD < 19) {
if (Rs < 2493) {
return 8;
}
else if (Rs >= 2493) {
return 4;
}
}
else if (AFD >= 19) {
return 4;
}
}
else if (mu_mv_mag_sq >= 16) {
if (mu_mv_mag_sq < 41) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Rs < 2311) {
return 2;
}
else if (Rs >= 2311) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (negBalance < 46) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 2;
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (Cs < 849) {
if (TSC < 5) {
if (Cs < 687) {
return 2;
}
else if (Cs >= 687) {
return 1;
}
}
else if (TSC >= 5) {
if (diffMVdiffVal < 24) {
return 2;
}
else if (diffMVdiffVal >= 24) {
return 4;
}
}
}
else if (Cs >= 849) {
if (ssDCval < 61) {
if (RsDiff < 8) {
if (posBalance < 42) {
if (Rs < 2175) {
if (Rs < 2150) {
return 4;
}
else if (Rs >= 2150) {
return 1;
}
}
else if (Rs >= 2175) {
if (Cs < 1123) {
return 2;
}
else if (Cs >= 1123) {
return 8;
}
}
}
else if (posBalance >= 42) {
return 8;
}
}
else if (RsDiff >= 8) {
if (refDCval < 56) {
return 4;
}
else if (refDCval >= 56) {
return 2;
}
}
}
else if (ssDCval >= 61) {
if (posBalance < 52) {
if (refDCval < 71) {
if (posBalance < 30) {
return 1;
}
else if (posBalance >= 30) {
if (ssDCval < 66) {
if (CsDiff < 5) {
if (diffTSC < 2) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 4;
}
}
else if (diffTSC >= 2) {
return 2;
}
}
else if (CsDiff >= 5) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (refDCval >= 71) {
if (Rs < 2495) {
return 4;
}
else if (Rs >= 2495) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (MVDiff < 24) {
if (diffTSC < -2) {
if (Rs < 2381) {
return 8;
}
else if (Rs >= 2381) {
return 2;
}
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (MVDiff >= 24) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 77) {
if (Cs < 872) {
if (ssDCval < 126) {
if (TSC < 12) {
if (tcor < 85) {
if (RsCsDiff < 141) {
if (diffTSC < 1) {
if (refDCval < 79) {
return 2;
}
else if (refDCval >= 79) {
return 8;
}
}
else if (diffTSC >= 1) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (Rs < 2300) {
return 2;
}
else if (Rs >= 2300) {
return 4;
}
}
}
}
else if (RsCsDiff >= 141) {
if (MVDiff < 31) {
if (mcTcor < 93) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (Cs < 812) {
return 4;
}
else if (Cs >= 812) {
return 2;
}
}
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (MVDiff >= 31) {
if (diffMVdiffVal < -23) {
return 2;
}
else if (diffMVdiffVal >= -23) {
if (posBalance < 69) {
if (tcor < 47) {
return 4;
}
else if (tcor >= 47) {
return 8;
}
}
else if (posBalance >= 69) {
return 2;
}
}
}
}
}
else if (tcor >= 85) {
if (Rs < 2265) {
if (gchDC < 2) {
if (ssDCval < 106) {
if (Cs < 615) {
return 1;
}
else if (Cs >= 615) {
return 4;
}
}
else if (ssDCval >= 106) {
if (Cs < 774) {
return 1;
}
else if (Cs >= 774) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (Rs >= 2265) {
if (Cs < 811) {
if (RsDiff < 12) {
if (posBalance < 53) {
if (posBalance < 51) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
return 1;
}
}
else if (posBalance >= 51) {
if (Rs < 2399) {
return 1;
}
else if (Rs >= 2399) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (ssDCval < 95) {
if (Rs < 2295) {
return 1;
}
else if (Rs >= 2295) {
return 2;
}
}
else if (ssDCval >= 95) {
return 1;
}
}
}
else if (RsDiff >= 12) {
if (refDCval < 104) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (refDCval >= 104) {
return 1;
}
}
}
else if (Cs >= 811) {
if (tcor < 86) {
if (Cs < 827) {
return 1;
}
else if (Cs >= 827) {
return 8;
}
}
else if (tcor >= 86) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
return 1;
}
}
}
}
}
}
else if (TSC >= 12) {
if (tcor < 45) {
if (Rs < 2302) {
return 4;
}
else if (Rs >= 2302) {
return 2;
}
}
else if (tcor >= 45) {
if (Rs < 2584) {
return 1;
}
else if (Rs >= 2584) {
return 8;
}
}
}
}
else if (ssDCval >= 126) {
if (tcor < 69) {
if (Rs < 2298) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (MVDiff < 44) {
return 8;
}
else if (MVDiff >= 44) {
return 1;
}
}
}
else if (Rs >= 2298) {
return 1;
}
}
else if (tcor >= 69) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
else if (Cs >= 872) {
if (RsDiff < 11) {
if (refDCval < 107) {
if (MVDiff < 25) {
if (negBalance < 37) {
if (Cs < 1223) {
return 1;
}
else if (Cs >= 1223) {
return 8;
}
}
else if (negBalance >= 37) {
if (RsDiff < 8) {
if (ssDCval < 89) {
if (tcor < 92) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (tcor >= 92) {
return 2;
}
}
else if (ssDCval >= 89) {
if (Rs < 2383) {
if (TSC < 8) {
if (SC < 3565) {
return 4;
}
else if (SC >= 3565) {
return 8;
}
}
else if (TSC >= 8) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
}
else if (Rs >= 2383) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (tcor < 87) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
else if (tcor >= 87) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
return 8;
}
}
}
}
}
else if (MVDiff >= 25) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 1;
}
}
}
}
else if (refDCval >= 107) {
if (AFD < 14) {
if (mcTcor < 97) {
if (posBalance < 40) {
if (Rs < 2403) {
return 4;
}
else if (Rs >= 2403) {
return 8;
}
}
else if (posBalance >= 40) {
return 8;
}
}
else if (mcTcor >= 97) {
if (Cs < 974) {
if (Rs < 2507) {
return 1;
}
else if (Rs >= 2507) {
return 2;
}
}
else if (Cs >= 974) {
return 8;
}
}
}
else if (AFD >= 14) {
if (MVDiff < 32) {
return 1;
}
else if (MVDiff >= 32) {
if (Rs < 2518) {
return 8;
}
else if (Rs >= 2518) {
return 1;
}
}
}
}
}
else if (RsDiff >= 11) {
if (tcor < 52) {
if (CsDiff < 7) {
if (Rs < 2279) {
return 8;
}
else if (Rs >= 2279) {
return 1;
}
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (tcor >= 52) {
if (ssDCval < 114) {
if (MVDiff < 40) {
if (Cs < 1147) {
if (TSC < 12) {
if (Rs < 2158) {
return 4;
}
else if (Rs >= 2158) {
return 2;
}
}
else if (TSC >= 12) {
if (Rs < 2366) {
return 4;
}
else if (Rs >= 2366) {
return 1;
}
}
}
else if (Cs >= 1147) {
if (Cs < 1274) {
if (tcor < 79) {
return 4;
}
else if (tcor >= 79) {
return 8;
}
}
else if (Cs >= 1274) {
return 2;
}
}
}
else if (MVDiff >= 40) {
if (MVDiff < 70) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
if (Rs < 2399) {
return 1;
}
else if (Rs >= 2399) {
return 2;
}
}
}
else if (MVDiff >= 70) {
if (Rs < 2275) {
return 8;
}
else if (Rs >= 2275) {
return 2;
}
}
}
}
else if (ssDCval >= 114) {
if (tcor < 69) {
return 8;
}
else if (tcor >= 69) {
if (ssDCval < 130) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (ssDCval >= 130) {
if (Cs < 945) {
return 1;
}
else if (Cs >= 945) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2587) {
if (MVDiff < 12) {
if (ssDCval < 109) {
if (posBalance < 47) {
if (negBalance < 21) {
if (SC < 4856) {
if (Cs < 813) {
if (Cs < 676) {
return 1;
}
else if (Cs >= 676) {
return 4;
}
}
else if (Cs >= 813) {
if (negBalance < 20) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 3180) {
return 8;
}
else if (Rs >= 3180) {
return 4;
}
}
}
else if (negBalance >= 20) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
return 2;
}
}
}
}
else if (SC >= 4856) {
if (SC < 5384) {
if (posBalance < 34) {
if (Cs < 1291) {
return 8;
}
else if (Cs >= 1291) {
return 4;
}
}
else if (posBalance >= 34) {
if (Cs < 1269) {
return 4;
}
else if (Cs >= 1269) {
if (Rs < 3892) {
return 4;
}
else if (Rs >= 3892) {
return 8;
}
}
}
}
else if (SC >= 5384) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
}
else if (negBalance >= 21) {
if (Rs < 3731) {
if (negBalance < 51) {
if (CsDiff < 2) {
if (ssDCval < 92) {
if (diffRsCsdiff < -2) {
if (RsDiff < 3) {
if (Cs < 1049) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (Cs >= 1049) {
return 8;
}
}
else if (RsDiff >= 3) {
if (Rs < 2709) {
return 1;
}
else if (Rs >= 2709) {
return 2;
}
}
}
else if (diffRsCsdiff >= -2) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 2888) {
return 2;
}
else if (Rs >= 2888) {
return 8;
}
}
}
}
else if (ssDCval >= 92) {
if (Cs < 1161) {
return 2;
}
else if (Cs >= 1161) {
if (Rs < 3114) {
return 8;
}
else if (Rs >= 3114) {
return 4;
}
}
}
}
else if (CsDiff >= 2) {
if (refDCval < 96) {
if (RsCsDiff < 23) {
if (Cs < 518) {
return 8;
}
else if (Cs >= 518) {
if (Cs < 1155) {
if (mu_mv_mag_sq < 64) {
if (Rs < 3557) {
return 4;
}
else if (Rs >= 3557) {
return 2;
}
}
else if (mu_mv_mag_sq >= 64) {
return 2;
}
}
else if (Cs >= 1155) {
return 8;
}
}
}
else if (RsCsDiff >= 23) {
if (refDCval < 77) {
if (CsDiff < 6) {
if (diffAFD < 2) {
if (ssDCval < 65) {
if (MVDiff < 3) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
else if (MVDiff >= 3) {
if (SC < 4809) {
return 4;
}
else if (SC >= 4809) {
return 2;
}
}
}
else if (ssDCval >= 65) {
if (posBalance < 41) {
if (diffTSC < 1) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (diffTSC >= 1) {
if (Cs < 1091) {
if (Rs < 3146) {
return 8;
}
else if (Rs >= 3146) {
return 4;
}
}
else if (Cs >= 1091) {
return 2;
}
}
}
else if (posBalance >= 41) {
if (Rs < 2919) {
if (Cs < 1055) {
return 4;
}
else if (Cs >= 1055) {
if (Rs < 2853) {
return 2;
}
else if (Rs >= 2853) {
if (Rs < 2908) {
return 4;
}
else if (Rs >= 2908) {
return 2;
}
}
}
}
else if (Rs >= 2919) {
if (TSC < 7) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (TSC >= 7) {
if (diffRsCsdiff < 10) {
if (Rs < 3376) {
return 2;
}
else if (Rs >= 3376) {
return 1;
}
}
else if (diffRsCsdiff >= 10) {
return 4;
}
}
}
}
}
}
else if (diffAFD >= 2) {
if (diffRsCsdiff < 59) {
if (gchDC < 1) {
if (Rs < 2741) {
return 2;
}
else if (Rs >= 2741) {
return 8;
}
}
else if (gchDC >= 1) {
if (diffAFD < 3) {
return 1;
}
else if (diffAFD >= 3) {
return 4;
}
}
}
else if (diffRsCsdiff >= 59) {
if (Cs < 1187) {
return 4;
}
else if (Cs >= 1187) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
if (Rs < 3287) {
return 2;
}
else if (Rs >= 3287) {
return 1;
}
}
}
}
}
else if (refDCval >= 77) {
if (Rs < 2925) {
if (mu_mv_mag_sq < 39) {
return 1;
}
else if (mu_mv_mag_sq >= 39) {
if (Cs < 867) {
return 4;
}
else if (Cs >= 867) {
return 2;
}
}
}
else if (Rs >= 2925) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (Rs < 3169) {
if (Rs < 3099) {
return 2;
}
else if (Rs >= 3099) {
return 1;
}
}
else if (Rs >= 3169) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (Rs < 3229) {
return 8;
}
else if (Rs >= 3229) {
if (Cs < 1063) {
if (Rs < 3604) {
return 4;
}
else if (Rs >= 3604) {
return 4;
}
}
else if (Cs >= 1063) {
return 4;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 96) {
if (mu_mv_mag_sq < 95) {
if (tcor < 94) {
if (Cs < 929) {
return 1;
}
else if (Cs >= 929) {
if (Rs < 3379) {
return 8;
}
else if (Rs >= 3379) {
return 8;
}
}
}
else if (tcor >= 94) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (MVDiff < 9) {
if (Rs < 3499) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 3363) {
return 4;
}
else if (Rs >= 3363) {
return 8;
}
}
}
else if (Rs >= 3499) {
return 4;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (ssDCval < 103) {
return 4;
}
else if (ssDCval >= 103) {
if (Rs < 3014) {
return 8;
}
else if (Rs >= 3014) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 51) {
if (Rs < 2813) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
return 1;
}
}
else if (Rs >= 2813) {
return 2;
}
}
}
else if (Rs >= 3731) {
if (negBalance < 46) {
if (Cs < 1035) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (RsDiff < 11) {
if (Rs < 3798) {
if (Rs < 3758) {
return 4;
}
else if (Rs >= 3758) {
return 8;
}
}
else if (Rs >= 3798) {
return 1;
}
}
else if (RsDiff >= 11) {
return 4;
}
}
}
else if (Cs >= 1035) {
if (negBalance < 33) {
if (ssDCval < 68) {
if (Rs < 4771) {
return 1;
}
else if (Rs >= 4771) {
return 4;
}
}
else if (ssDCval >= 68) {
if (SC < 5325) {
if (mcTcor < 96) {
if (Rs < 4157) {
if (Rs < 3916) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (SC < 4906) {
return 8;
}
else if (SC >= 4906) {
return 4;
}
}
}
else if (Rs >= 3916) {
return 8;
}
}
else if (Rs >= 4157) {
return 4;
}
}
else if (mcTcor >= 96) {
if (Cs < 1290) {
return 4;
}
else if (Cs >= 1290) {
if (Rs < 3895) {
return 4;
}
else if (Rs >= 3895) {
return 8;
}
}
}
}
else if (SC >= 5325) {
if (Rs < 4429) {
if (Cs < 1298) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
else if (Cs >= 1298) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (Rs >= 4429) {
return 2;
}
}
}
}
else if (negBalance >= 33) {
if (ssDCval < 92) {
if (mu_mv_mag_sq < 25) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
else if (mu_mv_mag_sq >= 25) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
if (Rs < 4218) {
return 4;
}
else if (Rs >= 4218) {
return 4;
}
}
}
}
}
else if (ssDCval >= 92) {
return 2;
}
}
}
}
else if (negBalance >= 46) {
return 2;
}
}
}
}
else if (posBalance >= 47) {
if (ssDCval < 90) {
if (refDCval < 60) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (refDCval >= 60) {
if (CsDiff < 3) {
if (Cs < 735) {
if (ssDCval < 79) {
if (SC < 3204) {
return 2;
}
else if (SC >= 3204) {
return 4;
}
}
else if (ssDCval >= 79) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (SC < 3234) {
if (Rs < 2767) {
return 4;
}
else if (Rs >= 2767) {
return 2;
}
}
else if (SC >= 3234) {
if (Rs < 2764) {
return 8;
}
else if (Rs >= 2764) {
return 8;
}
}
}
}
}
else if (Cs >= 735) {
if (gchDC < 2) {
if (refDCval < 63) {
if (Cs < 1028) {
return 2;
}
else if (Cs >= 1028) {
return 4;
}
}
else if (refDCval >= 63) {
return 2;
}
}
else if (gchDC >= 2) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
else if (CsDiff >= 3) {
if (posBalance < 50) {
if (negBalance < 41) {
if (mu_mv_mag_sq < 78) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
if (ssDCval < 68) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 4;
}
}
else if (ssDCval >= 68) {
if (Rs < 3193) {
return 1;
}
else if (Rs >= 3193) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
return 2;
}
}
else if (negBalance >= 41) {
if (diffRsCsdiff < -156) {
if (Rs < 3295) {
return 4;
}
else if (Rs >= 3295) {
return 2;
}
}
else if (diffRsCsdiff >= -156) {
if (Rs < 3533) {
return 2;
}
else if (Rs >= 3533) {
if (Rs < 3577) {
return 4;
}
else if (Rs >= 3577) {
return 2;
}
}
}
}
}
else if (posBalance >= 50) {
if (refDCval < 79) {
if (ssDCval < 63) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 2715) {
return 2;
}
else if (Rs >= 2715) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
}
else if (ssDCval >= 63) {
if (negBalance < 19) {
return 8;
}
else if (negBalance >= 19) {
if (negBalance < 39) {
if (posBalance < 55) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (Cs < 938) {
if (posBalance < 52) {
if (Rs < 2901) {
return 2;
}
else if (Rs >= 2901) {
return 4;
}
}
else if (posBalance >= 52) {
return 2;
}
}
else if (Cs >= 938) {
return 2;
}
}
}
else if (posBalance >= 55) {
return 2;
}
}
else if (negBalance >= 39) {
if (RsCsDiff < 49) {
if (Cs < 1085) {
return 2;
}
else if (Cs >= 1085) {
return 4;
}
}
else if (RsCsDiff >= 49) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
if (Rs < 3142) {
return 2;
}
else if (Rs >= 3142) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 79) {
if (RsCsDiff < 85) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
return 2;
}
}
}
else if (RsCsDiff >= 85) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 90) {
if (RsCsDiff < 102) {
if (tcor < 96) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (TSC < 6) {
if (Rs < 3937) {
return 8;
}
else if (Rs >= 3937) {
return 2;
}
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 53) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (mu_mv_mag_sq >= 53) {
if (negBalance < 30) {
if (ssDCval < 101) {
if (Rs < 4229) {
if (Cs < 496) {
return 4;
}
else if (Cs >= 496) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (Rs >= 4229) {
return 4;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (negBalance >= 30) {
return 8;
}
}
}
}
}
else if (tcor >= 96) {
if (posBalance < 57) {
if (Rs < 3180) {
return 8;
}
else if (Rs >= 3180) {
if (SC < 4972) {
return 2;
}
else if (SC >= 4972) {
return 4;
}
}
}
else if (posBalance >= 57) {
return 2;
}
}
}
else if (RsCsDiff >= 102) {
if (negBalance < 49) {
if (RsCsDiff < 142) {
if (diffRsCsdiff < 46) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 4;
}
}
else if (diffRsCsdiff >= 46) {
return 8;
}
}
else if (RsCsDiff >= 142) {
return 2;
}
}
else if (negBalance >= 49) {
if (Cs < 846) {
return 2;
}
else if (Cs >= 846) {
if (ssDCval < 103) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (Cs < 988) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
return 4;
}
}
else if (Cs >= 988) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (Rs < 3033) {
return 8;
}
else if (Rs >= 3033) {
return 2;
}
}
}
}
}
else if (ssDCval >= 103) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 3363) {
if (tcor < 92) {
if (RsDiff < 9) {
if (Rs < 3314) {
return 8;
}
else if (Rs >= 3314) {
return 2;
}
}
else if (RsDiff >= 9) {
if (Cs < 768) {
if (diffMVdiffVal < -25) {
return 1;
}
else if (diffMVdiffVal >= -25) {
return 2;
}
}
else if (Cs >= 768) {
if (Rs < 3044) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 1;
}
}
else if (Rs >= 3044) {
if (Cs < 1177) {
return 4;
}
else if (Cs >= 1177) {
return 1;
}
}
}
}
}
else if (tcor >= 92) {
if (MVDiff < 8) {
if (TSC < 3) {
if (tcor < 99) {
if (SC < 3430) {
return 2;
}
else if (SC >= 3430) {
return 8;
}
}
else if (tcor >= 99) {
if (Rs < 2724) {
return 1;
}
else if (Rs >= 2724) {
return 2;
}
}
}
else if (TSC >= 3) {
if (Cs < 1202) {
return 2;
}
else if (Cs >= 1202) {
if (Rs < 2942) {
if (Cs < 1293) {
return 8;
}
else if (Cs >= 1293) {
return 2;
}
}
else if (Rs >= 2942) {
return 2;
}
}
}
}
else if (MVDiff >= 8) {
if (SC < 3816) {
if (negBalance < 47) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (negBalance >= 47) {
return 2;
}
}
else if (SC >= 3816) {
if (mu_mv_mag_sq < 85) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (mu_mv_mag_sq >= 85) {
return 2;
}
}
}
}
}
else if (Rs >= 3363) {
if (CsDiff < 3) {
if (ssDCval < 142) {
if (SC < 4285) {
if (SC < 4265) {
return 2;
}
else if (SC >= 4265) {
return 1;
}
}
else if (SC >= 4285) {
return 2;
}
}
else if (ssDCval >= 142) {
if (Rs < 3836) {
if (SC < 4958) {
if (Cs < 1289) {
return 2;
}
else if (Cs >= 1289) {
return 8;
}
}
else if (SC >= 4958) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 2;
}
}
}
else if (Rs >= 3836) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 4;
}
}
}
}
else if (CsDiff >= 3) {
if (Rs < 3933) {
if (ssDCval < 142) {
if (MVDiff < 7) {
if (Cs < 701) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
else if (Cs >= 701) {
if (tcor < 91) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Rs < 3582) {
return 4;
}
else if (Rs >= 3582) {
return 8;
}
}
}
else if (tcor >= 91) {
return 2;
}
}
}
else if (MVDiff >= 7) {
if (TSC < 7) {
if (ssDCval < 117) {
return 1;
}
else if (ssDCval >= 117) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 2;
}
}
}
else if (TSC >= 7) {
if (SC < 4921) {
return 2;
}
else if (SC >= 4921) {
if (Rs < 3758) {
return 8;
}
else if (Rs >= 3758) {
return 1;
}
}
}
}
}
else if (ssDCval >= 142) {
if (Cs < 1107) {
if (Rs < 3619) {
return 1;
}
else if (Rs >= 3619) {
return 2;
}
}
else if (Cs >= 1107) {
return 1;
}
}
}
else if (Rs >= 3933) {
if (refDCval < 136) {
if (tcor < 67) {
if (Rs < 4173) {
return 4;
}
else if (Rs >= 4173) {
return 2;
}
}
else if (tcor >= 67) {
if (RsDiff < 5) {
if (Rs < 4450) {
return 1;
}
else if (Rs >= 4450) {
return 2;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (refDCval >= 136) {
if (mcTcor < 97) {
if (RsDiff < 19) {
if (SC < 5246) {
if (Rs < 4221) {
return 4;
}
else if (Rs >= 4221) {
return 2;
}
}
else if (SC >= 5246) {
if (ssDCval < 153) {
return 1;
}
else if (ssDCval >= 153) {
return 2;
}
}
}
else if (RsDiff >= 19) {
return 2;
}
}
else if (mcTcor >= 97) {
if (diffTSC < -2) {
if (Rs < 4455) {
return 1;
}
else if (Rs >= 4455) {
return 2;
}
}
else if (diffTSC >= -2) {
return 2;
}
}
}
}
}
}
}
}
else if (MVDiff >= 12) {
if (mcTcor < 69) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
if (tcor < 41) {
if (Rs < 3625) {
return 8;
}
else if (Rs >= 3625) {
return 2;
}
}
else if (tcor >= 41) {
return 4;
}
}
}
else if (mcTcor >= 69) {
if (refDCval < 132) {
if (refDCval < 98) {
if (refDCval < 78) {
if (Rs < 3290) {
if (TSC < 8) {
if (AFD < 10) {
if (posBalance < 42) {
if (posBalance < 38) {
if (Rs < 2702) {
return 4;
}
else if (Rs >= 2702) {
return 1;
}
}
else if (posBalance >= 38) {
return 8;
}
}
else if (posBalance >= 42) {
if (RsCsDiff < 38) {
return 8;
}
else if (RsCsDiff >= 38) {
return 4;
}
}
}
else if (AFD >= 10) {
if (Rs < 2900) {
if (posBalance < 56) {
if (diffRsCsdiff < 23) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (gchDC < 2) {
if (Rs < 2848) {
return 4;
}
else if (Rs >= 2848) {
return 2;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (diffRsCsdiff >= 23) {
if (Cs < 1108) {
return 4;
}
else if (Cs >= 1108) {
return 1;
}
}
}
else if (posBalance >= 56) {
return 2;
}
}
else if (Rs >= 2900) {
if (Rs < 3286) {
if (diffTSC < -3) {
if (Rs < 3107) {
return 4;
}
else if (Rs >= 3107) {
return 2;
}
}
else if (diffTSC >= -3) {
return 4;
}
}
else if (Rs >= 3286) {
return 1;
}
}
}
}
else if (TSC >= 8) {
if (TSC < 12) {
if (RsCsDiff < 128) {
if (diffRsCsdiff < -16) {
return 2;
}
else if (diffRsCsdiff >= -16) {
if (Cs < 1216) {
return 4;
}
else if (Cs >= 1216) {
return 1;
}
}
}
else if (RsCsDiff >= 128) {
if (posBalance < 54) {
if (Rs < 2730) {
if (refDCval < 69) {
return 2;
}
else if (refDCval >= 69) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 2;
}
}
}
else if (Rs >= 2730) {
if (mcTcor < 93) {
if (diffRsCsdiff < 43) {
if (SC < 4231) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
else if (SC >= 4231) {
if (Cs < 1311) {
return 4;
}
else if (Cs >= 1311) {
return 2;
}
}
}
else if (diffRsCsdiff >= 43) {
if (Cs < 1168) {
return 4;
}
else if (Cs >= 1168) {
return 2;
}
}
}
else if (mcTcor >= 93) {
return 2;
}
}
}
else if (posBalance >= 54) {
return 2;
}
}
}
else if (TSC >= 12) {
if (Cs < 498) {
return 2;
}
else if (Cs >= 498) {
return 1;
}
}
}
}
else if (Rs >= 3290) {
if (posBalance < 60) {
if (Cs < 946) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (Cs < 533) {
if (Rs < 4843) {
return 1;
}
else if (Rs >= 4843) {
return 8;
}
}
else if (Cs >= 533) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
else if (Cs >= 946) {
if (AFD < 8) {
if (Rs < 7126) {
return 8;
}
else if (Rs >= 7126) {
return 2;
}
}
else if (AFD >= 8) {
if (posBalance < 56) {
if (posBalance < 35) {
if (Rs < 3452) {
return 2;
}
else if (Rs >= 3452) {
return 1;
}
}
else if (posBalance >= 35) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (diffMVdiffVal < -6) {
if (Rs < 3592) {
return 4;
}
else if (Rs >= 3592) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
return 4;
}
}
}
}
else if (posBalance >= 56) {
return 2;
}
}
}
}
else if (posBalance >= 60) {
if (Cs < 670) {
if (Rs < 5197) {
return 4;
}
else if (Rs >= 5197) {
return 2;
}
}
else if (Cs >= 670) {
return 2;
}
}
}
}
else if (refDCval >= 78) {
if (RsDiff < 8) {
if (AFD < 7) {
if (SC < 3210) {
return 2;
}
else if (SC >= 3210) {
return 1;
}
}
else if (AFD >= 7) {
if (Rs < 3190) {
return 8;
}
else if (Rs >= 3190) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (MVDiff < 36) {
if (RsCsDiff < 124) {
if (diffRsCsdiff < -28) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 2;
}
}
else if (diffRsCsdiff >= -28) {
if (Rs < 3844) {
if (SC < 3797) {
return 1;
}
else if (SC >= 3797) {
return 8;
}
}
else if (Rs >= 3844) {
return 4;
}
}
}
else if (RsCsDiff >= 124) {
if (TSC < 13) {
if (negBalance < 40) {
if (posBalance < 59) {
if (Rs < 2893) {
return 1;
}
else if (Rs >= 2893) {
return 4;
}
}
else if (posBalance >= 59) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
if (Rs < 2617) {
return 2;
}
else if (Rs >= 2617) {
return 1;
}
}
}
}
else if (negBalance >= 40) {
if (Rs < 4255) {
if (refDCval < 94) {
if (Cs < 349) {
if (Rs < 3575) {
return 2;
}
else if (Rs >= 3575) {
return 4;
}
}
else if (Cs >= 349) {
if (negBalance < 41) {
if (Cs < 1173) {
return 2;
}
else if (Cs >= 1173) {
return 4;
}
}
else if (negBalance >= 41) {
return 2;
}
}
}
else if (refDCval >= 94) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (Cs < 662) {
return 2;
}
else if (Cs >= 662) {
return 4;
}
}
}
}
}
else if (Rs >= 4255) {
if (Cs < 1130) {
if (Cs < 710) {
return 4;
}
else if (Cs >= 710) {
return 1;
}
}
else if (Cs >= 1130) {
return 2;
}
}
}
}
else if (TSC >= 13) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
if (Rs < 2806) {
return 1;
}
else if (Rs >= 2806) {
return 2;
}
}
}
}
}
else if (MVDiff >= 36) {
if (Cs < 422) {
if (gchDC < 2) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
return 2;
}
}
else if (gchDC >= 2) {
if (mcTcor < 80) {
return 4;
}
else if (mcTcor >= 80) {
return 2;
}
}
}
else if (Cs >= 422) {
if (CsDiff < 6) {
if (SC < 4086) {
if (AFD < 14) {
if (Rs < 2901) {
return 1;
}
else if (Rs >= 2901) {
return 4;
}
}
else if (AFD >= 14) {
if (diffRsCsdiff < 71) {
return 8;
}
else if (diffRsCsdiff >= 71) {
if (Rs < 2899) {
return 1;
}
else if (Rs >= 2899) {
if (SC < 3741) {
return 8;
}
else if (SC >= 3741) {
return 1;
}
}
}
}
}
else if (SC >= 4086) {
if (diffRsCsdiff < 123) {
if (diffAFD < -3) {
if (Rs < 4228) {
return 1;
}
else if (Rs >= 4228) {
return 8;
}
}
else if (diffAFD >= -3) {
return 4;
}
}
else if (diffRsCsdiff >= 123) {
if (Rs < 3679) {
return 2;
}
else if (Rs >= 3679) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (diffRsCsdiff < -176) {
return 1;
}
else if (diffRsCsdiff >= -176) {
if (negBalance < 41) {
if (Rs < 4639) {
return 8;
}
else if (Rs >= 4639) {
return 2;
}
}
else if (negBalance >= 41) {
if (TSC < 17) {
return 2;
}
else if (TSC >= 17) {
return 4;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 98) {
if (RsDiff < 20) {
if (Rs < 4013) {
if (RsCsDiff < 220) {
if (SC < 3452) {
if (CsDiff < 3) {
if (negBalance < 50) {
if (RsCsDiff < 82) {
if (Rs < 2799) {
if (Cs < 12) {
return 8;
}
else if (Cs >= 12) {
return 1;
}
}
else if (Rs >= 2799) {
return 8;
}
}
else if (RsCsDiff >= 82) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (Rs < 2939) {
return 8;
}
else if (Rs >= 2939) {
return 1;
}
}
}
}
else if (negBalance >= 50) {
if (AFD < 14) {
if (Rs < 2843) {
return 1;
}
else if (Rs >= 2843) {
return 4;
}
}
else if (AFD >= 14) {
if (ssDCval < 100) {
if (Rs < 2752) {
return 1;
}
else if (Rs >= 2752) {
return 8;
}
}
else if (ssDCval >= 100) {
return 8;
}
}
}
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 12) {
if (refDCval < 110) {
if (Rs < 2803) {
if (Cs < 439) {
return 8;
}
else if (Cs >= 439) {
return 2;
}
}
else if (Rs >= 2803) {
return 1;
}
}
else if (refDCval >= 110) {
return 1;
}
}
else if (mu_mv_mag_sq >= 12) {
if (ssDCval < 124) {
if (Rs < 3092) {
return 8;
}
else if (Rs >= 3092) {
return 1;
}
}
else if (ssDCval >= 124) {
return 1;
}
}
}
}
else if (SC >= 3452) {
if (AFD < 8) {
if (Cs < 944) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Rs < 3366) {
return 2;
}
else if (Rs >= 3366) {
return 4;
}
}
}
else if (Cs >= 944) {
return 8;
}
}
else if (AFD >= 8) {
if (mcTcor < 95) {
if (MVDiff < 71) {
if (negBalance < 54) {
if (MVDiff < 29) {
if (negBalance < 25) {
if (Cs < 709) {
return 8;
}
else if (Cs >= 709) {
return 8;
}
}
else if (negBalance >= 25) {
if (Rs < 3071) {
return 4;
}
else if (Rs >= 3071) {
return 2;
}
}
}
else if (MVDiff >= 29) {
if (Cs < 720) {
return 8;
}
else if (Cs >= 720) {
if (Cs < 1188) {
return 1;
}
else if (Cs >= 1188) {
return 8;
}
}
}
}
else if (negBalance >= 54) {
if (Cs < 382) {
return 4;
}
else if (Cs >= 382) {
if (ssDCval < 98) {
if (Rs < 2827) {
return 2;
}
else if (Rs >= 2827) {
return 8;
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
else if (MVDiff >= 71) {
if (refDCval < 106) {
if (Rs < 2808) {
return 2;
}
else if (Rs >= 2808) {
if (Cs < 478) {
return 8;
}
else if (Cs >= 478) {
return 4;
}
}
}
else if (refDCval >= 106) {
if (Cs < 487) {
return 4;
}
else if (Cs >= 487) {
return 8;
}
}
}
}
else if (mcTcor >= 95) {
if (MVDiff < 16) {
if (Rs < 3565) {
if (mu_mv_mag_sq < 3) {
if (Rs < 3015) {
if (Rs < 2713) {
return 1;
}
else if (Rs >= 2713) {
return 4;
}
}
else if (Rs >= 3015) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
return 2;
}
}
else if (Rs >= 3565) {
if (CsDiff < 3) {
if (Rs < 3815) {
return 2;
}
else if (Rs >= 3815) {
return 4;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (MVDiff >= 16) {
if (refDCval < 126) {
if (Rs < 3176) {
if (SC < 4158) {
if (Rs < 2714) {
return 8;
}
else if (Rs >= 2714) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (SC < 3557) {
return 1;
}
else if (SC >= 3557) {
if (ssDCval < 105) {
if (Rs < 2858) {
return 1;
}
else if (Rs >= 2858) {
return 2;
}
}
else if (ssDCval >= 105) {
return 8;
}
}
}
}
}
else if (SC >= 4158) {
return 2;
}
}
else if (Rs >= 3176) {
if (gchDC < 4) {
if (diffMVdiffVal < -24) {
return 4;
}
else if (diffMVdiffVal >= -24) {
return 8;
}
}
else if (gchDC >= 4) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 2;
}
}
}
}
else if (refDCval >= 126) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (Rs < 3436) {
return 1;
}
else if (Rs >= 3436) {
return 8;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 220) {
if (MVDiff < 37) {
if (mcTcor < 92) {
if (Cs < 750) {
return 2;
}
else if (Cs >= 750) {
if (tcor < 72) {
if (AFD < 28) {
return 8;
}
else if (AFD >= 28) {
return 2;
}
}
else if (tcor >= 72) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
if (Rs < 3237) {
return 1;
}
else if (Rs >= 3237) {
return 8;
}
}
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 104) {
return 2;
}
else if (ssDCval >= 104) {
if (TSC < 7) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 8;
}
}
else if (TSC >= 7) {
if (Cs < 1123) {
if (tcor < 83) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (Cs < 529) {
return 2;
}
else if (Cs >= 529) {
return 1;
}
}
}
else if (tcor >= 83) {
if (diffRsCsdiff < -4) {
return 1;
}
else if (diffRsCsdiff >= -4) {
if (diffRsCsdiff < 49) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
return 2;
}
}
else if (diffRsCsdiff >= 49) {
return 1;
}
}
}
}
else if (Cs >= 1123) {
if (Rs < 2821) {
return 1;
}
else if (Rs >= 2821) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 37) {
if (mcTcor < 91) {
if (CsDiff < 5) {
if (ssDCval < 119) {
return 4;
}
else if (ssDCval >= 119) {
return 8;
}
}
else if (CsDiff >= 5) {
if (ssDCval < 109) {
if (Cs < 1196) {
return 8;
}
else if (Cs >= 1196) {
if (Cs < 1290) {
return 2;
}
else if (Cs >= 1290) {
return 8;
}
}
}
else if (ssDCval >= 109) {
if (Rs < 2591) {
return 4;
}
else if (Rs >= 2591) {
return 1;
}
}
}
}
else if (mcTcor >= 91) {
if (AFD < 16) {
if (Rs < 3051) {
return 4;
}
else if (Rs >= 3051) {
return 2;
}
}
else if (AFD >= 16) {
if (Rs < 3514) {
if (tcor < 73) {
if (Rs < 2917) {
if (Rs < 2702) {
return 1;
}
else if (Rs >= 2702) {
return 2;
}
}
else if (Rs >= 2917) {
return 8;
}
}
else if (tcor >= 73) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
if (diffAFD < 1) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
if (Cs < 427) {
return 1;
}
else if (Cs >= 427) {
return 8;
}
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
else if (Rs >= 3514) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 4013) {
if (negBalance < 42) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
if (Rs < 5510) {
return 1;
}
else if (Rs >= 5510) {
return 2;
}
}
}
else if (negBalance >= 42) {
if (Cs < 648) {
return 2;
}
else if (Cs >= 648) {
if (posBalance < 60) {
if (Cs < 1086) {
if (Rs < 4222) {
return 2;
}
else if (Rs >= 4222) {
if (AFD < 18) {
if (negBalance < 53) {
if (Rs < 5284) {
return 8;
}
else if (Rs >= 5284) {
return 2;
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (AFD >= 18) {
return 2;
}
}
}
else if (Cs >= 1086) {
if (mu_mv_mag_sq < 12) {
if (diffAFD < -3) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 2;
}
}
else if (diffAFD >= -3) {
if (Rs < 4645) {
return 2;
}
else if (Rs >= 4645) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 12) {
return 1;
}
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
}
}
else if (RsDiff >= 20) {
if (Cs < 847) {
if (RsDiff < 23) {
if (mcTcor < 98) {
if (gchDC < 9) {
if (Cs < 791) {
return 2;
}
else if (Cs >= 791) {
if (Cs < 802) {
return 1;
}
else if (Cs >= 802) {
return 2;
}
}
}
else if (gchDC >= 9) {
return 1;
}
}
else if (mcTcor >= 98) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
return 1;
}
}
}
else if (RsDiff >= 23) {
if (AFD < 34) {
return 2;
}
else if (AFD >= 34) {
if (Rs < 3955) {
return 2;
}
else if (Rs >= 3955) {
return 8;
}
}
}
}
else if (Cs >= 847) {
if (CsDiff < 8) {
if (TSC < 12) {
if (tcor < 74) {
if (Rs < 3667) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (Rs >= 3667) {
if (mu_mv_mag_sq < 7) {
if (mu_mv_mag_sq < 4) {
if (Rs < 6547) {
return 2;
}
else if (Rs >= 6547) {
return 8;
}
}
else if (mu_mv_mag_sq >= 4) {
if (MVDiff < 31) {
if (Rs < 4336) {
return 2;
}
else if (Rs >= 4336) {
return 8;
}
}
else if (MVDiff >= 31) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (gchDC < 1) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 2;
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
else if (tcor >= 74) {
if (ssDCval < 120) {
return 2;
}
else if (ssDCval >= 120) {
if (Rs < 5243) {
return 1;
}
else if (Rs >= 5243) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 2;
}
}
}
}
}
else if (TSC >= 12) {
if (Cs < 1215) {
return 2;
}
else if (Cs >= 1215) {
if (Rs < 3685) {
return 2;
}
else if (Rs >= 3685) {
return 4;
}
}
}
}
else if (CsDiff >= 8) {
if (MVDiff < 38) {
if (Cs < 1099) {
if (AFD < 28) {
return 1;
}
else if (AFD >= 28) {
return 2;
}
}
else if (Cs >= 1099) {
return 2;
}
}
else if (MVDiff >= 38) {
if (AFD < 27) {
if (Rs < 5470) {
return 2;
}
else if (Rs >= 5470) {
return 8;
}
}
else if (AFD >= 27) {
if (tcor < 45) {
return 2;
}
else if (tcor >= 45) {
if (mu_mv_mag_sq < 112) {
return 1;
}
else if (mu_mv_mag_sq >= 112) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 132) {
if (mcTcor < 93) {
if (AFD < 19) {
if (ssDCval < 150) {
if (ssDCval < 134) {
if (Cs < 608) {
return 4;
}
else if (Cs >= 608) {
return 1;
}
}
else if (ssDCval >= 134) {
return 4;
}
}
else if (ssDCval >= 150) {
if (Rs < 4410) {
return 8;
}
else if (Rs >= 4410) {
return 2;
}
}
}
else if (AFD >= 19) {
if (RsCsDiff < 677) {
if (mu_mv_mag_sq < 19) {
if (RsCsDiff < 256) {
return 8;
}
else if (RsCsDiff >= 256) {
if (Rs < 4932) {
if (mcTcor < 85) {
if (Rs < 3407) {
return 1;
}
else if (Rs >= 3407) {
return 4;
}
}
else if (mcTcor >= 85) {
if (diffMVdiffVal < -36) {
return 8;
}
else if (diffMVdiffVal >= -36) {
return 1;
}
}
}
else if (Rs >= 4932) {
if (AFD < 26) {
if (Cs < 1053) {
return 2;
}
else if (Cs >= 1053) {
return 1;
}
}
else if (AFD >= 26) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 19) {
if (posBalance < 55) {
if (Rs < 2921) {
return 1;
}
else if (Rs >= 2921) {
return 2;
}
}
else if (posBalance >= 55) {
return 1;
}
}
}
else if (RsCsDiff >= 677) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (CsDiff < 4) {
if (SC < 4054) {
if (AFD < 9) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (AFD >= 9) {
if (diffAFD < -2) {
if (Rs < 3287) {
return 1;
}
else if (Rs >= 3287) {
return 8;
}
}
else if (diffAFD >= -2) {
if (negBalance < 57) {
return 2;
}
else if (negBalance >= 57) {
if (RsDiff < 13) {
return 2;
}
else if (RsDiff >= 13) {
return 1;
}
}
}
}
}
else if (SC >= 4054) {
if (mcTcor < 95) {
if (Cs < 770) {
return 2;
}
else if (Cs >= 770) {
return 4;
}
}
else if (mcTcor >= 95) {
if (CsDiff < 3) {
if (Rs < 3345) {
return 8;
}
else if (Rs >= 3345) {
return 2;
}
}
else if (CsDiff >= 3) {
if (negBalance < 44) {
if (diffMVdiffVal < 5) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
else if (negBalance >= 44) {
if (Cs < 764) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
else if (Cs >= 764) {
return 1;
}
}
}
}
}
}
else if (CsDiff >= 4) {
if (RsDiff < 21) {
if (ssDCval < 148) {
if (posBalance < 40) {
if (Rs < 3707) {
return 1;
}
else if (Rs >= 3707) {
return 2;
}
}
else if (posBalance >= 40) {
if (Rs < 4698) {
return 1;
}
else if (Rs >= 4698) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
if (AFD < 22) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 1;
}
}
else if (AFD >= 22) {
return 8;
}
}
}
}
}
else if (ssDCval >= 148) {
if (RsDiff < 13) {
if (Cs < 726) {
return 1;
}
else if (Cs >= 726) {
return 8;
}
}
else if (RsDiff >= 13) {
if (Rs < 4937) {
return 1;
}
else if (Rs >= 4937) {
return 8;
}
}
}
}
else if (RsDiff >= 21) {
if (Rs < 4818) {
if (diffRsCsdiff < 165) {
if (RsCsDiff < 472) {
return 1;
}
else if (RsCsDiff >= 472) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 2;
}
}
}
else if (diffRsCsdiff >= 165) {
if (TSC < 14) {
return 1;
}
else if (TSC >= 14) {
return 2;
}
}
}
else if (Rs >= 4818) {
if (tcor < 82) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
if (Cs < 996) {
return 2;
}
else if (Cs >= 996) {
return 8;
}
}
}
else if (tcor >= 82) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1313) {
if (mu_mv_mag_sq < 105) {
if (RsCsDiff < 101) {
if (ssDCval < 93) {
if (negBalance < 38) {
if (Rs < 811) {
if (RsDiff < 1) {
if (Rs < 335) {
return 4;
}
else if (Rs >= 335) {
if (Rs < 505) {
return 1;
}
else if (Rs >= 505) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (MVDiff < 107) {
return 1;
}
else if (MVDiff >= 107) {
return 8;
}
}
}
else if (Rs >= 811) {
if (SC < 10510) {
if (negBalance < 32) {
if (posBalance < 45) {
if (ssDCval < 64) {
if (Rs < 2997) {
if (ssDCval < 54) {
if (diffAFD < -2) {
if (Rs < 2121) {
return 2;
}
else if (Rs >= 2121) {
return 1;
}
}
else if (diffAFD >= -2) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 1;
}
}
}
else if (ssDCval >= 54) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
if (Rs < 1543) {
return 4;
}
else if (Rs >= 1543) {
return 8;
}
}
}
}
else if (Rs >= 2997) {
if (Cs < 3315) {
return 4;
}
else if (Cs >= 3315) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (ssDCval < 71) {
if (MVDiff < 1) {
if (Rs < 3077) {
return 8;
}
else if (Rs >= 3077) {
return 4;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 71) {
if (TSC < 3) {
if (Rs < 2347) {
if (Rs < 2165) {
return 8;
}
else if (Rs >= 2165) {
return 1;
}
}
else if (Rs >= 2347) {
if (Cs < 2179) {
return 4;
}
else if (Cs >= 2179) {
if (Rs < 3803) {
return 1;
}
else if (Rs >= 3803) {
return 4;
}
}
}
}
else if (TSC >= 3) {
if (SC < 7250) {
if (ssDCval < 74) {
if (Rs < 2856) {
return 8;
}
else if (Rs >= 2856) {
return 4;
}
}
else if (ssDCval >= 74) {
if (Rs < 2415) {
if (Rs < 1861) {
return 8;
}
else if (Rs >= 1861) {
return 1;
}
}
else if (Rs >= 2415) {
return 8;
}
}
}
else if (SC >= 7250) {
return 4;
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2653) {
if (Cs < 1623) {
if (Rs < 1055) {
if (ssDCval < 51) {
if (Rs < 978) {
return 1;
}
else if (Rs >= 978) {
return 8;
}
}
else if (ssDCval >= 51) {
return 2;
}
}
else if (Rs >= 1055) {
if (SC < 2756) {
return 4;
}
else if (SC >= 2756) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (diffRsCsdiff < 10) {
if (MVDiff < 2) {
if (Rs < 2266) {
return 8;
}
else if (Rs >= 2266) {
return 4;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (diffRsCsdiff >= 10) {
if (Rs < 1334) {
return 8;
}
else if (Rs >= 1334) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1623) {
if (mu_mv_mag_sq < 79) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 1347) {
return 2;
}
else if (Rs >= 1347) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (ssDCval < 54) {
if (Rs < 1872) {
return 1;
}
else if (Rs >= 1872) {
return 4;
}
}
else if (ssDCval >= 54) {
return 8;
}
}
}
}
else if (Rs >= 2653) {
if (diffRsCsdiff < -32) {
if (Cs < 1921) {
return 2;
}
else if (Cs >= 1921) {
return 8;
}
}
else if (diffRsCsdiff >= -32) {
if (MVDiff < 3) {
if (Rs < 2963) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 1;
}
}
else if (Rs >= 2963) {
if (ssDCval < 84) {
return 2;
}
else if (ssDCval >= 84) {
if (Rs < 3422) {
return 4;
}
else if (Rs >= 3422) {
return 2;
}
}
}
}
else if (MVDiff >= 3) {
if (posBalance < 55) {
if (MVDiff < 6) {
if (Rs < 3678) {
return 8;
}
else if (Rs >= 3678) {
return 8;
}
}
else if (MVDiff >= 6) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (Cs < 1728) {
return 1;
}
else if (Cs >= 1728) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 32) {
if (RsDiff < 7) {
if (Cs < 1609) {
if (Rs < 3187) {
if (mu_mv_mag_sq < 77) {
if (ssDCval < 71) {
if (ssDCval < 49) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (ssDCval >= 49) {
if (tcor < 77) {
if (Rs < 1074) {
return 8;
}
else if (Rs >= 1074) {
return 1;
}
}
else if (tcor >= 77) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 1398) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Cs >= 1398) {
if (MVDiff < 2) {
if (Rs < 1787) {
return 4;
}
else if (Rs >= 1787) {
return 8;
}
}
else if (MVDiff >= 2) {
if (mcTcor < 96) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (Rs < 1740) {
return 4;
}
else if (Rs >= 1740) {
return 2;
}
}
}
else if (mcTcor >= 96) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (Rs < 2124) {
if (Cs < 1335) {
return 1;
}
else if (Cs >= 1335) {
return 8;
}
}
else if (Rs >= 2124) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 77) {
if (diffRsCsdiff < -21) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
return 2;
}
}
else if (diffRsCsdiff >= -21) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (Rs < 1478) {
return 4;
}
else if (Rs >= 1478) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
if (Rs < 2705) {
return 4;
}
else if (Rs >= 2705) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 3187) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (ssDCval < 81) {
if (Cs < 1338) {
return 2;
}
else if (Cs >= 1338) {
return 4;
}
}
else if (ssDCval >= 81) {
if (Rs < 4253) {
return 2;
}
else if (Rs >= 4253) {
return 8;
}
}
}
}
}
else if (Cs >= 1609) {
if (AFD < 7) {
if (Rs < 2931) {
if (SC < 4739) {
if (MVDiff < 6) {
if (Cs < 2158) {
return 8;
}
else if (Cs >= 2158) {
return 1;
}
}
else if (MVDiff >= 6) {
if (CsDiff < 5) {
if (mu_mv_mag_sq < 91) {
return 1;
}
else if (mu_mv_mag_sq >= 91) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (SC >= 4739) {
if (ssDCval < 82) {
if (AFD < 5) {
if (Rs < 2696) {
return 8;
}
else if (Rs >= 2696) {
return 4;
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (ssDCval >= 82) {
return 4;
}
}
}
else if (Rs >= 2931) {
if (ssDCval < 81) {
if (Rs < 3285) {
return 4;
}
else if (Rs >= 3285) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 8;
}
}
}
else if (ssDCval >= 81) {
return 8;
}
}
}
else if (AFD >= 7) {
if (mcTcor < 94) {
if (mcTcor < 93) {
if (mcTcor < 89) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
return 8;
}
}
else if (mcTcor >= 89) {
return 8;
}
}
else if (mcTcor >= 93) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Rs < 2717) {
return 2;
}
else if (Rs >= 2717) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (ssDCval < 86) {
if (Rs < 3304) {
if (negBalance < 35) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 62) {
return 4;
}
else if (mu_mv_mag_sq >= 62) {
if (MVDiff < 5) {
if (Rs < 2758) {
return 8;
}
else if (Rs >= 2758) {
return 1;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
}
}
else if (negBalance >= 35) {
if (diffMVdiffVal < 1) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 3110) {
return 1;
}
else if (Rs >= 3110) {
return 8;
}
}
}
else if (diffMVdiffVal >= 1) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Rs < 1649) {
return 4;
}
else if (Rs >= 1649) {
return 2;
}
}
}
}
}
else if (Rs >= 3304) {
if (Rs < 4829) {
if (SC < 5723) {
return 4;
}
else if (SC >= 5723) {
return 8;
}
}
else if (Rs >= 4829) {
if (Rs < 5085) {
return 1;
}
else if (Rs >= 5085) {
return 4;
}
}
}
}
else if (ssDCval >= 86) {
if (Rs < 2560) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (Rs >= 2560) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (Cs < 1800) {
if (Rs < 2640) {
if (Cs < 1701) {
if (Cs < 1367) {
return 2;
}
else if (Cs >= 1367) {
return 1;
}
}
else if (Cs >= 1701) {
return 8;
}
}
else if (Rs >= 2640) {
if (MVDiff < 6) {
if (ssDCval < 92) {
if (mu_mv_mag_sq < 65) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Cs < 1518) {
return 4;
}
else if (Cs >= 1518) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 65) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
if (Rs < 3283) {
return 4;
}
else if (Rs >= 3283) {
return 2;
}
}
}
}
else if (ssDCval >= 92) {
return 8;
}
}
else if (MVDiff >= 6) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 2884) {
return 2;
}
else if (Rs >= 2884) {
return 4;
}
}
}
}
}
else if (Cs >= 1800) {
if (ssDCval < 87) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (Rs < 3674) {
return 1;
}
else if (Rs >= 3674) {
if (Rs < 4641) {
return 2;
}
else if (Rs >= 4641) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 3881) {
return 8;
}
else if (Rs >= 3881) {
return 2;
}
}
}
}
}
}
}
else if (SC >= 10510) {
if (Rs < 5682) {
return 8;
}
else if (Rs >= 5682) {
return 2;
}
}
}
}
else if (negBalance >= 38) {
if (Rs < 1384) {
if (ssDCval < 56) {
if (MVDiff < 18) {
if (posBalance < 56) {
if (Rs < 1044) {
if (Rs < 890) {
return 4;
}
else if (Rs >= 890) {
return 2;
}
}
else if (Rs >= 1044) {
if (Rs < 1235) {
return 8;
}
else if (Rs >= 1235) {
if (Rs < 1287) {
return 4;
}
else if (Rs >= 1287) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
}
}
}
else if (posBalance >= 56) {
return 4;
}
}
else if (MVDiff >= 18) {
if (Cs < 1343) {
return 4;
}
else if (Cs >= 1343) {
return 2;
}
}
}
else if (ssDCval >= 56) {
if (AFD < 13) {
if (ssDCval < 83) {
if (Rs < 1250) {
if (mcTcor < 96) {
if (Cs < 1507) {
if (negBalance < 43) {
if (refDCval < 59) {
if (AFD < 9) {
if (Rs < 1104) {
return 4;
}
else if (Rs >= 1104) {
return 8;
}
}
else if (AFD >= 9) {
return 2;
}
}
else if (refDCval >= 59) {
return 4;
}
}
else if (negBalance >= 43) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (MVDiff < 9) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
return 1;
}
}
else if (MVDiff >= 9) {
if (ssDCval < 65) {
if (Rs < 1034) {
return 2;
}
else if (Rs >= 1034) {
return 8;
}
}
else if (ssDCval >= 65) {
return 4;
}
}
}
}
}
else if (Cs >= 1507) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
if (TSC < 5) {
if (Rs < 1072) {
return 1;
}
else if (Rs >= 1072) {
if (Rs < 1147) {
return 4;
}
else if (Rs >= 1147) {
return 8;
}
}
}
else if (TSC >= 5) {
if (MVDiff < 21) {
if (MVDiff < 6) {
if (diffRsCsdiff < -2) {
if (Rs < 783) {
return 2;
}
else if (Rs >= 783) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (MVDiff >= 6) {
if (Rs < 569) {
return 1;
}
else if (Rs >= 569) {
return 2;
}
}
}
else if (MVDiff >= 21) {
return 8;
}
}
}
}
else if (Rs >= 1250) {
if (AFD < 10) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (AFD >= 10) {
if (Rs < 1345) {
return 4;
}
else if (Rs >= 1345) {
return 8;
}
}
}
}
else if (ssDCval >= 83) {
if (mu_mv_mag_sq < 47) {
if (diffRsCsdiff < -8) {
if (Rs < 1223) {
return 4;
}
else if (Rs >= 1223) {
return 2;
}
}
else if (diffRsCsdiff >= -8) {
if (tcor < 88) {
if (Rs < 1101) {
return 4;
}
else if (Rs >= 1101) {
return 8;
}
}
else if (tcor >= 88) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 47) {
return 8;
}
}
}
else if (AFD >= 13) {
if (ssDCval < 74) {
if (RsCsDiff < 73) {
if (tcor < 76) {
return 8;
}
else if (tcor >= 76) {
if (ssDCval < 69) {
if (Rs < 912) {
return 2;
}
else if (Rs >= 912) {
return 8;
}
}
else if (ssDCval >= 69) {
return 2;
}
}
}
else if (RsCsDiff >= 73) {
if (tcor < 81) {
if (RsCsDiff < 99) {
return 4;
}
else if (RsCsDiff >= 99) {
return 2;
}
}
else if (tcor >= 81) {
return 8;
}
}
}
else if (ssDCval >= 74) {
if (SC < 2525) {
if (RsCsDiff < 71) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 4;
}
}
}
else if (RsCsDiff >= 71) {
return 4;
}
}
else if (SC >= 2525) {
if (TSC < 11) {
if (diffAFD < -6) {
return 2;
}
else if (diffAFD >= -6) {
if (MVDiff < 5) {
if (Rs < 1162) {
return 8;
}
else if (Rs >= 1162) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1384) {
if (RsDiff < 8) {
if (SC < 7013) {
if (SC < 5264) {
if (mcTcor < 94) {
if (SC < 2883) {
if (AFD < 16) {
if (RsDiff < 7) {
if (ssDCval < 64) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (RsDiff >= 7) {
if (Rs < 1416) {
return 1;
}
else if (Rs >= 1416) {
return 4;
}
}
}
else if (AFD >= 16) {
if (ssDCval < 68) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
else if (ssDCval >= 68) {
if (refDCval < 90) {
return 4;
}
else if (refDCval >= 90) {
return 8;
}
}
}
}
else if (SC >= 2883) {
if (diffRsCsdiff < 29) {
if (diffTSC < -3) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
else if (diffTSC >= -3) {
if (posBalance < 57) {
if (SC < 3408) {
if (SC < 3400) {
if (tcor < 63) {
return 1;
}
else if (tcor >= 63) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (SC >= 3400) {
if (Rs < 1903) {
return 8;
}
else if (Rs >= 1903) {
return 1;
}
}
}
else if (SC >= 3408) {
if (RsCsDiff < 73) {
if (diffAFD < -4) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 1;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (RsCsDiff >= 73) {
if (mu_mv_mag_sq < 92) {
if (TSC < 8) {
if (Cs < 1915) {
return 8;
}
else if (Cs >= 1915) {
if (Rs < 2356) {
return 4;
}
else if (Rs >= 2356) {
return 8;
}
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (mu_mv_mag_sq >= 92) {
if (diffRsCsdiff < 7) {
if (AFD < 11) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (AFD >= 11) {
return 8;
}
}
else if (diffRsCsdiff >= 7) {
return 8;
}
}
}
}
}
else if (posBalance >= 57) {
if (negBalance < 45) {
if (ssDCval < 82) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
return 4;
}
}
}
else if (ssDCval >= 82) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (negBalance >= 45) {
if (ssDCval < 60) {
if (Rs < 1656) {
if (mcTcor < 69) {
if (Cs < 1576) {
return 8;
}
else if (Cs >= 1576) {
return 4;
}
}
else if (mcTcor >= 69) {
if (Rs < 1414) {
return 2;
}
else if (Rs >= 1414) {
return 4;
}
}
}
else if (Rs >= 1656) {
if (SC < 3523) {
return 8;
}
else if (SC >= 3523) {
if (Rs < 2027) {
if (tcor < 68) {
return 8;
}
else if (tcor >= 68) {
return 4;
}
}
else if (Rs >= 2027) {
return 8;
}
}
}
}
else if (ssDCval >= 60) {
if (mu_mv_mag_sq < 46) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
if (negBalance < 57) {
if (negBalance < 49) {
if (Cs < 1457) {
return 8;
}
else if (Cs >= 1457) {
return 4;
}
}
else if (negBalance >= 49) {
if (AFD < 19) {
if (Rs < 1617) {
if (Rs < 1615) {
return 8;
}
else if (Rs >= 1615) {
return 4;
}
}
else if (Rs >= 1617) {
return 8;
}
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (negBalance >= 57) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (Rs < 2212) {
return 1;
}
else if (Rs >= 2212) {
return 8;
}
}
}
}
}
}
}
}
else if (diffRsCsdiff >= 29) {
if (ssDCval < 87) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Cs < 1768) {
return 8;
}
else if (Cs >= 1768) {
return 4;
}
}
}
else if (ssDCval >= 87) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 3153) {
if (mu_mv_mag_sq < 90) {
if (MVDiff < 5) {
if (TSC < 7) {
if (Cs < 1535) {
if (ssDCval < 84) {
if (RsDiff < 5) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 1871) {
return 4;
}
else if (Rs >= 1871) {
return 1;
}
}
}
else if (RsDiff >= 5) {
if (diffMVdiffVal < -2) {
if (Rs < 2550) {
return 8;
}
else if (Rs >= 2550) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 89) {
if (Rs < 2207) {
return 2;
}
else if (Rs >= 2207) {
return 4;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
}
else if (Cs >= 1535) {
if (Cs < 2588) {
if (AFD < 11) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (Cs < 1594) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 2585) {
return 1;
}
else if (Rs >= 2585) {
return 4;
}
}
}
else if (Cs >= 1594) {
return 8;
}
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (Cs >= 2588) {
return 4;
}
}
}
else if (TSC >= 7) {
if (MVDiff < 2) {
if (SC < 4654) {
if (Rs < 1651) {
return 8;
}
else if (Rs >= 1651) {
if (ssDCval < 78) {
return 4;
}
else if (ssDCval >= 78) {
return 1;
}
}
}
else if (SC >= 4654) {
return 8;
}
}
else if (MVDiff >= 2) {
if (refDCval < 81) {
if (tcor < 92) {
if (Cs < 1904) {
return 4;
}
else if (Cs >= 1904) {
return 8;
}
}
else if (tcor >= 92) {
if (Rs < 2030) {
return 4;
}
else if (Rs >= 2030) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (refDCval < 91) {
if (ssDCval < 88) {
if (refDCval < 84) {
if (negBalance < 45) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
return 1;
}
}
else if (negBalance >= 45) {
if (Rs < 2149) {
return 1;
}
else if (Rs >= 2149) {
return 8;
}
}
}
else if (refDCval >= 84) {
return 1;
}
}
else if (ssDCval >= 88) {
return 8;
}
}
else if (refDCval >= 91) {
if (Rs < 2478) {
return 8;
}
else if (Rs >= 2478) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 5) {
if (refDCval < 81) {
if (mu_mv_mag_sq < 66) {
if (Cs < 1570) {
return 8;
}
else if (Cs >= 1570) {
if (Cs < 1720) {
return 4;
}
else if (Cs >= 1720) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (gchDC < 1) {
if (AFD < 10) {
if (AFD < 9) {
if (Rs < 1620) {
return 1;
}
else if (Rs >= 1620) {
return 8;
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (AFD >= 10) {
return 4;
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (refDCval >= 81) {
if (posBalance < 60) {
if (Cs < 1479) {
if (tcor < 92) {
if (MVDiff < 10) {
if (negBalance < 51) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
else if (negBalance >= 51) {
return 8;
}
}
else if (MVDiff >= 10) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
return 2;
}
}
}
else if (tcor >= 92) {
return 8;
}
}
else if (Cs >= 1479) {
return 8;
}
}
else if (posBalance >= 60) {
if (Rs < 1947) {
return 2;
}
else if (Rs >= 1947) {
if (SC < 4258) {
return 8;
}
else if (SC >= 4258) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (diffRsCsdiff < 19) {
if (Cs < 1742) {
if (Rs < 2068) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (refDCval < 76) {
if (Rs < 1679) {
return 4;
}
else if (Rs >= 1679) {
return 1;
}
}
else if (refDCval >= 76) {
return 8;
}
}
}
else if (Rs >= 2068) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (negBalance < 42) {
if (RsCsDiff < 37) {
return 8;
}
else if (RsCsDiff >= 37) {
return 2;
}
}
else if (negBalance >= 42) {
if (Rs < 2556) {
return 1;
}
else if (Rs >= 2556) {
return 8;
}
}
}
}
}
else if (Cs >= 1742) {
if (Cs < 2645) {
if (tcor < 93) {
if (RsCsDiff < 46) {
return 4;
}
else if (RsCsDiff >= 46) {
return 8;
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (Cs >= 2645) {
return 4;
}
}
}
else if (diffRsCsdiff >= 19) {
if (Cs < 1631) {
return 2;
}
else if (Cs >= 1631) {
if (Rs < 1976) {
return 8;
}
else if (Rs >= 1976) {
return 4;
}
}
}
}
}
else if (Rs >= 3153) {
if (RsDiff < 7) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (RsDiff >= 7) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 4;
}
}
}
}
}
else if (SC >= 5264) {
if (Rs < 2723) {
if (Cs < 2815) {
return 2;
}
else if (Cs >= 2815) {
return 4;
}
}
else if (Rs >= 2723) {
if (posBalance < 46) {
if (Rs < 3453) {
return 8;
}
else if (Rs >= 3453) {
if (posBalance < 30) {
return 2;
}
else if (posBalance >= 30) {
return 4;
}
}
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 74) {
if (CsDiff < 8) {
if (SC < 6105) {
return 8;
}
else if (SC >= 6105) {
if (SC < 6604) {
return 1;
}
else if (SC >= 6604) {
return 8;
}
}
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (mu_mv_mag_sq >= 74) {
if (refDCval < 88) {
if (MVDiff < 2) {
if (Cs < 3436) {
return 8;
}
else if (Cs >= 3436) {
return 1;
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 90) {
return 8;
}
else if (mu_mv_mag_sq >= 90) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
if (Rs < 3640) {
return 8;
}
else if (Rs >= 3640) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 88) {
if (Cs < 2721) {
return 8;
}
else if (Cs >= 2721) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 7013) {
if (Rs < 5747) {
return 8;
}
else if (Rs >= 5747) {
return 4;
}
}
}
else if (RsDiff >= 8) {
if (tcor < 92) {
if (ssDCval < 84) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (Cs < 1878) {
if (TSC < 9) {
if (diffMVdiffVal < 2) {
if (SC < 4298) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (SC >= 4298) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 1360) {
return 8;
}
else if (Cs >= 1360) {
return 4;
}
}
}
else if (TSC >= 9) {
if (posBalance < 53) {
if (Rs < 2271) {
return 8;
}
else if (Rs >= 2271) {
return 4;
}
}
else if (posBalance >= 53) {
if (negBalance < 49) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (Rs < 1994) {
return 1;
}
else if (Rs >= 1994) {
return 4;
}
}
}
else if (negBalance >= 49) {
return 8;
}
}
}
}
else if (Cs >= 1878) {
return 8;
}
}
}
else if (ssDCval >= 84) {
if (mcTcor < 94) {
if (TSC < 9) {
if (Rs < 2427) {
return 2;
}
else if (Rs >= 2427) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 2763) {
if (Cs < 1351) {
return 1;
}
else if (Cs >= 1351) {
return 2;
}
}
else if (Rs >= 2763) {
return 8;
}
}
}
}
else if (tcor >= 92) {
if (AFD < 10) {
if (Cs < 2142) {
return 2;
}
else if (Cs >= 2142) {
return 4;
}
}
else if (AFD >= 10) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 93) {
if (ssDCval < 163) {
if (Rs < 4342) {
if (Rs < 1439) {
if (mu_mv_mag_sq < 38) {
if (ssDCval < 145) {
if (AFD < 20) {
if (mcTcor < 88) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (Rs < 1197) {
return 1;
}
else if (Rs >= 1197) {
return 8;
}
}
}
else if (mcTcor >= 88) {
if (ssDCval < 122) {
if (RsCsDiff < 74) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (mcTcor < 93) {
if (SC < 2714) {
return 8;
}
else if (SC >= 2714) {
return 4;
}
}
else if (mcTcor >= 93) {
if (SC < 2201) {
if (Rs < 683) {
return 8;
}
else if (Rs >= 683) {
return 4;
}
}
else if (SC >= 2201) {
return 8;
}
}
}
}
else if (RsCsDiff >= 74) {
if (posBalance < 55) {
if (Rs < 1367) {
return 2;
}
else if (Rs >= 1367) {
return 8;
}
}
else if (posBalance >= 55) {
if (diffAFD < 3) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
return 4;
}
}
else if (diffAFD >= 3) {
return 2;
}
}
}
}
else if (ssDCval >= 122) {
if (CsDiff < 4) {
if (Cs < 1358) {
return 8;
}
else if (Cs >= 1358) {
return 1;
}
}
else if (CsDiff >= 4) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
}
else if (AFD >= 20) {
if (mu_mv_mag_sq < 18) {
return 4;
}
else if (mu_mv_mag_sq >= 18) {
return 2;
}
}
}
else if (ssDCval >= 145) {
return 2;
}
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 1262) {
if (MVDiff < 19) {
if (Rs < 724) {
return 1;
}
else if (Rs >= 724) {
if (ssDCval < 145) {
if (negBalance < 44) {
if (mcTcor < 94) {
if (Cs < 1367) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 1367) {
if (diffRsCsdiff < 6) {
return 4;
}
else if (diffRsCsdiff >= 6) {
if (Rs < 1102) {
return 4;
}
else if (Rs >= 1102) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (MVDiff < 6) {
if (posBalance < 60) {
if (ssDCval < 132) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (Rs < 1006) {
return 8;
}
else if (Rs >= 1006) {
return 4;
}
}
}
}
else if (ssDCval >= 132) {
return 8;
}
}
else if (posBalance >= 60) {
return 8;
}
}
else if (MVDiff >= 6) {
if (Cs < 1319) {
return 4;
}
else if (Cs >= 1319) {
return 8;
}
}
}
}
else if (negBalance >= 44) {
if (ssDCval < 102) {
return 4;
}
else if (ssDCval >= 102) {
return 8;
}
}
}
else if (ssDCval >= 145) {
return 4;
}
}
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (Rs >= 1262) {
if (SC < 3132) {
if (SC < 2595) {
return 4;
}
else if (SC >= 2595) {
if (Cs < 1315) {
return 2;
}
else if (Cs >= 1315) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (SC >= 3132) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Cs < 2222) {
return 8;
}
else if (Cs >= 2222) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1439) {
if (refDCval < 110) {
if (mcTcor < 96) {
if (Cs < 2700) {
if (negBalance < 56) {
if (TSC < 10) {
if (Rs < 2225) {
if (ssDCval < 101) {
if (RsDiff < 8) {
if (TSC < 7) {
if (Rs < 1871) {
return 8;
}
else if (Rs >= 1871) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (TSC >= 7) {
if (Cs < 2420) {
if (posBalance < 49) {
if (Rs < 1787) {
return 8;
}
else if (Rs >= 1787) {
return 1;
}
}
else if (posBalance >= 49) {
return 8;
}
}
else if (Cs >= 2420) {
if (Rs < 1693) {
return 2;
}
else if (Rs >= 1693) {
return 8;
}
}
}
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (Rs >= 2225) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (refDCval < 103) {
if (negBalance < 55) {
if (diffAFD < -4) {
if (Rs < 2714) {
return 8;
}
else if (Rs >= 2714) {
return 4;
}
}
else if (diffAFD >= -4) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
return 8;
}
}
}
}
else if (negBalance >= 55) {
return 4;
}
}
else if (refDCval >= 103) {
if (diffAFD < 4) {
if (MVDiff < 12) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
return 8;
}
}
else if (MVDiff >= 12) {
if (Rs < 3039) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (Rs >= 3039) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
return 4;
}
}
}
}
else if (diffAFD >= 4) {
return 4;
}
}
}
}
}
else if (TSC >= 10) {
if (Rs < 1814) {
if (Cs < 1708) {
return 8;
}
else if (Cs >= 1708) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (Rs >= 1814) {
if (negBalance < 52) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
if (Cs < 1527) {
return 1;
}
else if (Cs >= 1527) {
if (CsDiff < 8) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (Rs < 2474) {
return 8;
}
else if (Rs >= 2474) {
return 4;
}
}
}
else if (CsDiff >= 8) {
return 1;
}
}
}
}
else if (negBalance >= 52) {
return 8;
}
}
}
}
else if (negBalance >= 56) {
if (MVDiff < 11) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 4;
}
}
else if (MVDiff >= 11) {
if (TSC < 12) {
if (Rs < 1648) {
return 2;
}
else if (Rs >= 1648) {
return 8;
}
}
else if (TSC >= 12) {
return 1;
}
}
}
}
else if (Cs >= 2700) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Cs < 2729) {
return 4;
}
else if (Cs >= 2729) {
if (MVDiff < 3) {
if (Rs < 2918) {
return 4;
}
else if (Rs >= 2918) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1916) {
if (TSC < 6) {
if (tcor < 94) {
if (Rs < 1839) {
return 8;
}
else if (Rs >= 1839) {
return 4;
}
}
else if (tcor >= 94) {
if (Rs < 1896) {
return 2;
}
else if (Rs >= 1896) {
return 8;
}
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (Rs >= 1916) {
if (tcor < 92) {
if (ssDCval < 98) {
if (ssDCval < 97) {
return 1;
}
else if (ssDCval >= 97) {
return 2;
}
}
else if (ssDCval >= 98) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 2948) {
return 1;
}
else if (Rs >= 2948) {
return 8;
}
}
}
}
else if (tcor >= 92) {
if (Rs < 3553) {
if (MVDiff < 3) {
if (posBalance < 43) {
if (CsDiff < 6) {
if (Rs < 3212) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (Rs >= 3212) {
return 4;
}
}
else if (CsDiff >= 6) {
return 1;
}
}
else if (posBalance >= 43) {
if (mcTcor < 97) {
if (Rs < 2660) {
return 8;
}
else if (Rs >= 2660) {
return 4;
}
}
else if (mcTcor >= 97) {
if (Rs < 3302) {
return 8;
}
else if (Rs >= 3302) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
}
else if (MVDiff >= 3) {
if (refDCval < 106) {
if (CsDiff < 1) {
if (Rs < 2287) {
return 1;
}
else if (Rs >= 2287) {
return 8;
}
}
else if (CsDiff >= 1) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Rs < 2091) {
if (Rs < 1922) {
return 1;
}
else if (Rs >= 1922) {
if (Cs < 1726) {
return 8;
}
else if (Cs >= 1726) {
return 2;
}
}
}
else if (Rs >= 2091) {
if (Rs < 2930) {
return 8;
}
else if (Rs >= 2930) {
return 1;
}
}
}
}
}
else if (refDCval >= 106) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2352) {
return 2;
}
else if (Rs >= 2352) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (MVDiff < 16) {
if (AFD < 10) {
if (Rs < 2807) {
return 8;
}
else if (Rs >= 2807) {
return 4;
}
}
else if (AFD >= 10) {
return 8;
}
}
else if (MVDiff >= 16) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 3553) {
if (Cs < 1940) {
if (ssDCval < 100) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
else if (ssDCval >= 100) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Rs < 3960) {
return 8;
}
else if (Rs >= 3960) {
return 4;
}
}
}
}
else if (Cs >= 1940) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 110) {
if (ssDCval < 140) {
if (ssDCval < 125) {
if (Rs < 1754) {
if (AFD < 13) {
if (RsCsDiff < 57) {
return 8;
}
else if (RsCsDiff >= 57) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 8;
}
}
}
}
else if (AFD >= 13) {
if (negBalance < 47) {
return 4;
}
else if (negBalance >= 47) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
return 2;
}
}
}
}
else if (Rs >= 1754) {
if (Cs < 1376) {
if (tcor < 96) {
if (Rs < 4149) {
return 8;
}
else if (Rs >= 4149) {
return 1;
}
}
else if (tcor >= 96) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
return 2;
}
}
}
else if (Cs >= 1376) {
if (TSC < 11) {
if (Rs < 4300) {
if (tcor < 106) {
if (mcTcor < 101) {
if (diffMVdiffVal < -3) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
if (Rs < 1948) {
return 2;
}
else if (Rs >= 1948) {
if (Cs < 2752) {
return 8;
}
else if (Cs >= 2752) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (mcTcor >= 101) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 8;
}
}
}
else if (tcor >= 106) {
return 2;
}
}
else if (Rs >= 4300) {
if (Rs < 4316) {
return 1;
}
else if (Rs >= 4316) {
return 8;
}
}
}
else if (TSC >= 11) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
}
else if (ssDCval >= 125) {
if (mcTcor < 92) {
if (Cs < 2041) {
if (AFD < 12) {
if (CsDiff < 5) {
if (SC < 3776) {
return 1;
}
else if (SC >= 3776) {
return 8;
}
}
else if (CsDiff >= 5) {
if (ssDCval < 138) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Cs < 1642) {
return 1;
}
else if (Cs >= 1642) {
return 8;
}
}
}
else if (ssDCval >= 138) {
return 1;
}
}
}
else if (AFD >= 12) {
if (Rs < 1501) {
return 1;
}
else if (Rs >= 1501) {
return 8;
}
}
}
else if (Cs >= 2041) {
if (Rs < 3656) {
return 8;
}
else if (Rs >= 3656) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (refDCval < 127) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (mcTcor < 94) {
if (Cs < 1682) {
return 1;
}
else if (Cs >= 1682) {
return 8;
}
}
else if (mcTcor >= 94) {
if (diffMVdiffVal < -3) {
if (Rs < 1961) {
return 2;
}
else if (Rs >= 1961) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
}
}
else if (refDCval >= 127) {
if (negBalance < 29) {
if (Rs < 2639) {
return 8;
}
else if (Rs >= 2639) {
return 4;
}
}
else if (negBalance >= 29) {
if (mcTcor < 97) {
if (ssDCval < 128) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 3418) {
return 1;
}
else if (Rs >= 3418) {
return 8;
}
}
}
else if (ssDCval >= 128) {
return 8;
}
}
else if (mcTcor >= 97) {
if (SC < 3936) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 1;
}
}
else if (SC >= 3936) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (negBalance < 24) {
return 2;
}
else if (negBalance >= 24) {
if (mu_mv_mag_sq < 54) {
if (Rs < 1689) {
if (mu_mv_mag_sq < 22) {
return 2;
}
else if (mu_mv_mag_sq >= 22) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (Rs >= 1689) {
if (diffMVdiffVal < -15) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (diffMVdiffVal >= -15) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (diffRsCsdiff < -25) {
if (Rs < 3054) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 4;
}
}
else if (Rs >= 3054) {
if (Rs < 3207) {
return 4;
}
else if (Rs >= 3207) {
if (Cs < 2173) {
return 8;
}
else if (Cs >= 2173) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= -25) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 4342) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 25) {
if (Cs < 1395) {
return 2;
}
else if (Cs >= 1395) {
return 8;
}
}
else if (mu_mv_mag_sq >= 25) {
if (CsDiff < 7) {
if (Rs < 5810) {
if (tcor < 89) {
return 1;
}
else if (tcor >= 89) {
if (Cs < 1543) {
return 2;
}
else if (Cs >= 1543) {
return 8;
}
}
}
else if (Rs >= 5810) {
return 4;
}
}
else if (CsDiff >= 7) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (Cs < 2662) {
return 2;
}
else if (Cs >= 2662) {
if (Rs < 5961) {
return 8;
}
else if (Rs >= 5961) {
return 4;
}
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2264) {
if (Rs < 1101) {
return 1;
}
else if (Rs >= 1101) {
return 2;
}
}
else if (Rs >= 2264) {
return 8;
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1874) {
if (Rs < 3355) {
if (tcor < 78) {
if (mcTcor < 86) {
if (Rs < 1635) {
if (gchDC < 3) {
if (mcTcor < 74) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
if (Rs < 1306) {
return 4;
}
else if (Rs >= 1306) {
return 1;
}
}
}
}
else if (mcTcor >= 74) {
if (negBalance < 55) {
if (Cs < 1443) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1083) {
return 4;
}
else if (Rs >= 1083) {
return 2;
}
}
}
else if (Cs >= 1443) {
if (Cs < 1534) {
if (Rs < 966) {
return 4;
}
else if (Rs >= 966) {
return 1;
}
}
else if (Cs >= 1534) {
return 4;
}
}
}
else if (negBalance >= 55) {
if (ssDCval < 87) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
if (mu_mv_mag_sq < 13) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
else if (mu_mv_mag_sq >= 13) {
return 4;
}
}
}
else if (ssDCval >= 87) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 2;
}
}
}
}
}
else if (gchDC >= 3) {
if (negBalance < 66) {
if (MVDiff < 41) {
if (Rs < 1115) {
return 8;
}
else if (Rs >= 1115) {
if (Rs < 1401) {
return 2;
}
else if (Rs >= 1401) {
return 1;
}
}
}
else if (MVDiff >= 41) {
return 8;
}
}
else if (negBalance >= 66) {
if (Cs < 1367) {
return 4;
}
else if (Cs >= 1367) {
if (AFD < 31) {
return 2;
}
else if (AFD >= 31) {
return 1;
}
}
}
}
}
else if (Rs >= 1635) {
if (mu_mv_mag_sq < 30) {
if (ssDCval < 121) {
if (RsDiff < 12) {
if (CsDiff < 8) {
if (Cs < 1329) {
return 2;
}
else if (Cs >= 1329) {
if (Cs < 1779) {
return 8;
}
else if (Cs >= 1779) {
if (Rs < 2446) {
return 8;
}
else if (Rs >= 2446) {
return 4;
}
}
}
}
else if (CsDiff >= 8) {
if (mu_mv_mag_sq < 2) {
if (refDCval < 80) {
if (RsCsDiff < 154) {
if (Cs < 1812) {
return 8;
}
else if (Cs >= 1812) {
return 4;
}
}
else if (RsCsDiff >= 154) {
return 4;
}
}
else if (refDCval >= 80) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
if (CsDiff < 9) {
if (Cs < 1725) {
if (Rs < 1722) {
return 1;
}
else if (Rs >= 1722) {
return 8;
}
}
else if (Cs >= 1725) {
return 1;
}
}
else if (CsDiff >= 9) {
if (CsDiff < 10) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
return 8;
}
}
else if (CsDiff >= 10) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 12) {
if (RsCsDiff < 255) {
if (CsDiff < 10) {
if (posBalance < 64) {
if (negBalance < 55) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 8;
}
}
else if (negBalance >= 55) {
return 4;
}
}
else if (posBalance >= 64) {
if (Rs < 2133) {
return 8;
}
else if (Rs >= 2133) {
return 1;
}
}
}
else if (CsDiff >= 10) {
return 1;
}
}
else if (RsCsDiff >= 255) {
if (posBalance < 52) {
if (Rs < 2630) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
else if (Rs >= 2630) {
if (Rs < 2884) {
return 2;
}
else if (Rs >= 2884) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (mcTcor < 82) {
if (MVDiff < 61) {
if (diffMVdiffVal < 8) {
return 8;
}
else if (diffMVdiffVal >= 8) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
return 8;
}
}
}
else if (MVDiff >= 61) {
return 2;
}
}
else if (mcTcor >= 82) {
if (gchDC < 1) {
if (Cs < 1630) {
return 8;
}
else if (Cs >= 1630) {
return 2;
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 121) {
if (posBalance < 56) {
if (Cs < 1442) {
return 2;
}
else if (Cs >= 1442) {
return 8;
}
}
else if (posBalance >= 56) {
if (ssDCval < 144) {
if (Rs < 1829) {
return 8;
}
else if (Rs >= 1829) {
return 1;
}
}
else if (ssDCval >= 144) {
if (CsDiff < 10) {
if (Rs < 2457) {
return 2;
}
else if (Rs >= 2457) {
return 1;
}
}
else if (CsDiff >= 10) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (RsCsDiff < 201) {
if (mu_mv_mag_sq < 57) {
if (AFD < 22) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (AFD >= 22) {
return 4;
}
}
else if (mu_mv_mag_sq >= 57) {
if (Rs < 2260) {
if (Cs < 1739) {
if (Rs < 2110) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
return 4;
}
}
else if (Rs >= 2110) {
return 4;
}
}
else if (Cs >= 1739) {
return 4;
}
}
else if (Rs >= 2260) {
return 8;
}
}
}
else if (RsCsDiff >= 201) {
if (TSC < 9) {
if (Cs < 1406) {
return 8;
}
else if (Cs >= 1406) {
return 1;
}
}
else if (TSC >= 9) {
if (TSC < 22) {
if (tcor < 77) {
if (Rs < 3215) {
return 8;
}
else if (Rs >= 3215) {
return 2;
}
}
else if (tcor >= 77) {
return 1;
}
}
else if (TSC >= 22) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 86) {
if (posBalance < 50) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (negBalance < 54) {
if (AFD < 13) {
if (Rs < 2638) {
return 8;
}
else if (Rs >= 2638) {
return 4;
}
}
else if (AFD >= 13) {
if (Cs < 1830) {
return 2;
}
else if (Cs >= 1830) {
return 8;
}
}
}
else if (negBalance >= 54) {
if (MVDiff < 114) {
return 8;
}
else if (MVDiff >= 114) {
return 4;
}
}
}
}
else if (posBalance >= 50) {
if (RsCsDiff < 263) {
if (refDCval < 88) {
if (TSC < 13) {
if (ssDCval < 79) {
if (RsDiff < 6) {
if (diffMVdiffVal < -10) {
return 2;
}
else if (diffMVdiffVal >= -10) {
return 4;
}
}
else if (RsDiff >= 6) {
if (negBalance < 58) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (RsDiff < 11) {
if (MVDiff < 13) {
if (Rs < 2201) {
return 8;
}
else if (Rs >= 2201) {
if (Rs < 2499) {
return 4;
}
else if (Rs >= 2499) {
return 8;
}
}
}
else if (MVDiff >= 13) {
if (RsDiff < 9) {
if (gchDC < 2) {
if (SC < 3047) {
if (Cs < 1428) {
return 8;
}
else if (Cs >= 1428) {
return 4;
}
}
else if (SC >= 3047) {
return 1;
}
}
else if (gchDC >= 2) {
if (Rs < 789) {
return 8;
}
else if (Rs >= 789) {
return 4;
}
}
}
else if (RsDiff >= 9) {
if (diffRsCsdiff < -9) {
return 8;
}
else if (diffRsCsdiff >= -9) {
if (RsCsDiff < 173) {
if (negBalance < 54) {
return 4;
}
else if (negBalance >= 54) {
return 2;
}
}
else if (RsCsDiff >= 173) {
return 8;
}
}
}
}
}
else if (RsDiff >= 11) {
if (mcTcor < 90) {
if (mcTcor < 87) {
if (Cs < 1533) {
return 8;
}
else if (Cs >= 1533) {
return 1;
}
}
else if (mcTcor >= 87) {
return 2;
}
}
else if (mcTcor >= 90) {
return 8;
}
}
}
}
else if (negBalance >= 58) {
if (RsCsDiff < 154) {
return 2;
}
else if (RsCsDiff >= 154) {
if (MVDiff < 21) {
if (Cs < 1597) {
return 2;
}
else if (Cs >= 1597) {
return 8;
}
}
else if (MVDiff >= 21) {
return 4;
}
}
}
}
}
else if (ssDCval >= 79) {
if (refDCval < 85) {
if (Rs < 1596) {
if (Rs < 1130) {
return 4;
}
else if (Rs >= 1130) {
if (Cs < 1505) {
return 2;
}
else if (Cs >= 1505) {
return 8;
}
}
}
else if (Rs >= 1596) {
return 4;
}
}
else if (refDCval >= 85) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (Rs < 2446) {
return 8;
}
else if (Rs >= 2446) {
return 4;
}
}
}
}
}
else if (TSC >= 13) {
if (Rs < 1981) {
if (negBalance < 59) {
return 1;
}
else if (negBalance >= 59) {
return 8;
}
}
else if (Rs >= 1981) {
return 4;
}
}
}
else if (refDCval >= 88) {
if (negBalance < 54) {
if (AFD < 18) {
if (RsDiff < 12) {
if (ssDCval < 103) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (Rs < 2450) {
return 8;
}
else if (Rs >= 2450) {
return 2;
}
}
}
else if (ssDCval >= 103) {
if (Rs < 3164) {
return 8;
}
else if (Rs >= 3164) {
if (Rs < 3319) {
return 1;
}
else if (Rs >= 3319) {
return 8;
}
}
}
}
else if (RsDiff >= 12) {
return 4;
}
}
else if (AFD >= 18) {
if (Rs < 1689) {
if (diffRsCsdiff < 53) {
return 4;
}
else if (diffRsCsdiff >= 53) {
return 8;
}
}
else if (Rs >= 1689) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
if (Rs < 2867) {
return 2;
}
else if (Rs >= 2867) {
return 4;
}
}
}
}
}
else if (negBalance >= 54) {
if (Rs < 1368) {
if (SC < 2280) {
return 1;
}
else if (SC >= 2280) {
if (Cs < 1457) {
return 8;
}
else if (Cs >= 1457) {
return 2;
}
}
}
else if (Rs >= 1368) {
if (tcor < 62) {
if (SC < 4111) {
if (tcor < 54) {
return 4;
}
else if (tcor >= 54) {
if (Rs < 2020) {
return 8;
}
else if (Rs >= 2020) {
return 2;
}
}
}
else if (SC >= 4111) {
return 1;
}
}
else if (tcor >= 62) {
if (ssDCval < 162) {
if (diffMVdiffVal < -2) {
if (Cs < 1802) {
if (MVDiff < 32) {
return 8;
}
else if (MVDiff >= 32) {
if (mu_mv_mag_sq < 9) {
return 1;
}
else if (mu_mv_mag_sq >= 9) {
return 8;
}
}
}
else if (Cs >= 1802) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 3015) {
return 8;
}
else if (Rs >= 3015) {
return 4;
}
}
}
else if (ssDCval >= 162) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 263) {
if (RsCsDiff < 379) {
if (RsDiff < 15) {
if (refDCval < 87) {
if (negBalance < 57) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
if (SC < 3403) {
return 2;
}
else if (SC >= 3403) {
return 8;
}
}
}
else if (negBalance >= 57) {
if (Rs < 2030) {
return 4;
}
else if (Rs >= 2030) {
return 8;
}
}
}
else if (refDCval >= 87) {
if (Rs < 961) {
return 2;
}
else if (Rs >= 961) {
if (AFD < 32) {
return 8;
}
else if (AFD >= 32) {
return 2;
}
}
}
}
else if (RsDiff >= 15) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Rs < 2603) {
if (Cs < 1679) {
return 2;
}
else if (Cs >= 1679) {
return 4;
}
}
else if (Rs >= 2603) {
return 4;
}
}
}
}
else if (RsCsDiff >= 379) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
if (posBalance < 59) {
if (Cs < 1470) {
return 1;
}
else if (Cs >= 1470) {
if (RsDiff < 20) {
return 8;
}
else if (RsDiff >= 20) {
return 2;
}
}
}
else if (posBalance >= 59) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
}
}
}
}
else if (tcor >= 78) {
if (refDCval < 100) {
if (Rs < 3015) {
if (mcTcor < 96) {
if (refDCval < 59) {
if (mu_mv_mag_sq < 68) {
return 1;
}
else if (mu_mv_mag_sq >= 68) {
return 8;
}
}
else if (refDCval >= 59) {
if (Cs < 1549) {
if (mu_mv_mag_sq < 9) {
if (TSC < 9) {
if (ssDCval < 93) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 1814) {
return 8;
}
else if (Rs >= 1814) {
return 2;
}
}
}
else if (ssDCval >= 93) {
if (Rs < 2342) {
return 2;
}
else if (Rs >= 2342) {
return 4;
}
}
}
else if (TSC >= 9) {
if (RsCsDiff < 156) {
if (SC < 3330) {
if (RsDiff < 9) {
if (Cs < 1458) {
return 8;
}
else if (Cs >= 1458) {
return 2;
}
}
else if (RsDiff >= 9) {
if (Rs < 1540) {
return 2;
}
else if (Rs >= 1540) {
return 1;
}
}
}
else if (SC >= 3330) {
return 8;
}
}
else if (RsCsDiff >= 156) {
if (diffRsCsdiff < 8) {
if (Cs < 1342) {
return 8;
}
else if (Cs >= 1342) {
return 4;
}
}
else if (diffRsCsdiff >= 8) {
if (TSC < 12) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (TSC >= 12) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (SC < 4299) {
if (posBalance < 44) {
if (RsDiff < 15) {
return 2;
}
else if (RsDiff >= 15) {
return 8;
}
}
else if (posBalance >= 44) {
if (Cs < 1326) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
else if (Cs >= 1326) {
if (TSC < 10) {
if (Rs < 1984) {
if (RsCsDiff < 121) {
if (Rs < 1825) {
return 8;
}
else if (Rs >= 1825) {
return 1;
}
}
else if (RsCsDiff >= 121) {
return 8;
}
}
else if (Rs >= 1984) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
return 8;
}
}
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
else if (SC >= 4299) {
if (Rs < 2869) {
return 4;
}
else if (Rs >= 2869) {
return 2;
}
}
}
}
else if (Cs >= 1549) {
if (ssDCval < 87) {
if (RsDiff < 10) {
if (mu_mv_mag_sq < 5) {
if (posBalance < 52) {
if (refDCval < 80) {
if (Cs < 1620) {
return 8;
}
else if (Cs >= 1620) {
return 1;
}
}
else if (refDCval >= 80) {
return 8;
}
}
else if (posBalance >= 52) {
if (diffAFD < -1) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
else if (diffAFD >= -1) {
if (ssDCval < 86) {
if (diffMVdiffVal < -5) {
if (Rs < 1463) {
return 8;
}
else if (Rs >= 1463) {
return 4;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (ssDCval >= 86) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (diffRsCsdiff < -52) {
return 1;
}
else if (diffRsCsdiff >= -52) {
return 8;
}
}
}
else if (RsDiff >= 10) {
if (negBalance < 48) {
if (Rs < 2465) {
if (Cs < 1641) {
return 4;
}
else if (Cs >= 1641) {
return 8;
}
}
else if (Rs >= 2465) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (MVDiff < 19) {
if (Cs < 1618) {
return 8;
}
else if (Cs >= 1618) {
if (Rs < 2748) {
return 4;
}
else if (Rs >= 2748) {
if (Cs < 1680) {
return 4;
}
else if (Cs >= 1680) {
return 8;
}
}
}
}
else if (MVDiff >= 19) {
return 8;
}
}
}
}
else if (ssDCval >= 87) {
if (Rs < 1655) {
return 2;
}
else if (Rs >= 1655) {
if (Rs < 2778) {
if (gchDC < 3) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
if (refDCval < 87) {
if (Rs < 2283) {
return 2;
}
else if (Rs >= 2283) {
return 8;
}
}
else if (refDCval >= 87) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (SC < 4051) {
return 8;
}
else if (SC >= 4051) {
return 2;
}
}
}
else if (Rs >= 2778) {
if (Rs < 2783) {
return 2;
}
else if (Rs >= 2783) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (RsDiff < 7) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (RsDiff >= 7) {
if (mu_mv_mag_sq < 27) {
if (MVDiff < 33) {
if (diffRsCsdiff < 66) {
return 2;
}
else if (diffRsCsdiff >= 66) {
if (Rs < 1504) {
return 1;
}
else if (Rs >= 1504) {
return 8;
}
}
}
else if (MVDiff >= 33) {
return 1;
}
}
else if (mu_mv_mag_sq >= 27) {
if (RsDiff < 10) {
if (Rs < 2814) {
return 4;
}
else if (Rs >= 2814) {
return 1;
}
}
else if (RsDiff >= 10) {
return 8;
}
}
}
}
}
else if (Rs >= 3015) {
if (negBalance < 54) {
if (AFD < 13) {
if (Cs < 1712) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (Cs >= 1712) {
return 8;
}
}
else if (AFD >= 13) {
if (AFD < 15) {
if (Rs < 3306) {
return 4;
}
else if (Rs >= 3306) {
return 8;
}
}
else if (AFD >= 15) {
if (Rs < 3193) {
return 2;
}
else if (Rs >= 3193) {
return 4;
}
}
}
}
else if (negBalance >= 54) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
}
}
else if (refDCval >= 100) {
if (AFD < 17) {
if (Rs < 1539) {
if (TSC < 7) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (Rs >= 1539) {
if (diffRsCsdiff < -35) {
if (ssDCval < 109) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (Rs < 3143) {
return 4;
}
else if (Rs >= 3143) {
return 2;
}
}
}
}
else if (ssDCval >= 109) {
if (tcor < 80) {
if (Rs < 2136) {
return 8;
}
else if (Rs >= 2136) {
return 2;
}
}
else if (tcor >= 80) {
return 8;
}
}
}
else if (diffRsCsdiff >= -35) {
if (RsCsDiff < 161) {
return 8;
}
else if (RsCsDiff >= 161) {
if (mu_mv_mag_sq < 62) {
if (posBalance < 48) {
if (Rs < 2347) {
return 2;
}
else if (Rs >= 2347) {
return 8;
}
}
else if (posBalance >= 48) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 62) {
if (Rs < 2023) {
return 2;
}
else if (Rs >= 2023) {
return 4;
}
}
}
}
}
}
else if (AFD >= 17) {
if (SC < 4331) {
if (MVDiff < 20) {
if (mu_mv_mag_sq < 7) {
if (mu_mv_mag_sq < 5) {
if (Rs < 1660) {
return 4;
}
else if (Rs >= 1660) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
else if (MVDiff >= 20) {
if (tcor < 86) {
if (SC < 4120) {
if (gchDC < 5) {
if (MVDiff < 23) {
if (Cs < 1488) {
return 1;
}
else if (Cs >= 1488) {
return 2;
}
}
else if (MVDiff >= 23) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
return 8;
}
}
}
else if (gchDC >= 5) {
if (Rs < 1500) {
return 2;
}
else if (Rs >= 1500) {
return 8;
}
}
}
else if (SC >= 4120) {
if (RsCsDiff < 187) {
return 2;
}
else if (RsCsDiff >= 187) {
return 8;
}
}
}
else if (tcor >= 86) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
return 8;
}
}
}
}
else if (SC >= 4331) {
if (CsDiff < 7) {
if (Cs < 1498) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (Cs >= 1498) {
return 4;
}
}
else if (CsDiff >= 7) {
if (Rs < 3213) {
return 8;
}
else if (Rs >= 3213) {
if (Rs < 3234) {
return 1;
}
else if (Rs >= 3234) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 3355) {
if (refDCval < 101) {
if (mcTcor < 92) {
if (RsDiff < 13) {
if (negBalance < 41) {
if (posBalance < 49) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
if (Rs < 4118) {
return 4;
}
else if (Rs >= 4118) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (Rs < 3533) {
return 8;
}
else if (Rs >= 3533) {
return 2;
}
}
}
else if (negBalance >= 41) {
return 8;
}
}
else if (RsDiff >= 13) {
if (mcTcor < 87) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
if (mu_mv_mag_sq < 21) {
return 2;
}
else if (mu_mv_mag_sq >= 21) {
if (ssDCval < 94) {
if (Rs < 3397) {
return 8;
}
else if (Rs >= 3397) {
return 2;
}
}
else if (ssDCval >= 94) {
if (Rs < 3675) {
return 8;
}
else if (Rs >= 3675) {
return 4;
}
}
}
}
}
else if (mcTcor >= 87) {
if (MVDiff < 102) {
if (Rs < 3388) {
return 4;
}
else if (Rs >= 3388) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (MVDiff >= 102) {
return 8;
}
}
}
}
else if (mcTcor >= 92) {
if (posBalance < 45) {
if (negBalance < 34) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (negBalance >= 34) {
if (mcTcor < 94) {
if (diffMVdiffVal < 2) {
if (Rs < 3870) {
return 4;
}
else if (Rs >= 3870) {
if (Rs < 4262) {
return 2;
}
else if (Rs >= 4262) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 1409) {
return 1;
}
else if (Cs >= 1409) {
return 2;
}
}
}
else if (mcTcor >= 94) {
return 2;
}
}
}
else if (posBalance >= 45) {
if (tcor < 81) {
if (TSC < 9) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 4;
}
}
else if (TSC >= 9) {
if (Rs < 3550) {
return 8;
}
else if (Rs >= 3550) {
return 2;
}
}
}
else if (tcor >= 81) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (diffAFD < -7) {
if (Rs < 3640) {
return 8;
}
else if (Rs >= 3640) {
return 2;
}
}
else if (diffAFD >= -7) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 101) {
if (RsDiff < 13) {
if (mu_mv_mag_sq < 86) {
if (posBalance < 43) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (posBalance < 40) {
if (Rs < 3556) {
return 4;
}
else if (Rs >= 3556) {
if (Rs < 4424) {
return 2;
}
else if (Rs >= 4424) {
return 1;
}
}
}
else if (posBalance >= 40) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (ssDCval < 137) {
if (diffRsCsdiff < -87) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (mu_mv_mag_sq < 3) {
if (Rs < 4779) {
return 1;
}
else if (Rs >= 4779) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (diffRsCsdiff >= -87) {
return 8;
}
}
else if (ssDCval >= 137) {
if (AFD < 13) {
if (Rs < 3929) {
return 8;
}
else if (Rs >= 3929) {
return 1;
}
}
else if (AFD >= 13) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
return 1;
}
}
else if (RsDiff >= 13) {
if (MVDiff < 17) {
if (tcor < 77) {
if (posBalance < 61) {
if (mu_mv_mag_sq < 13) {
if (Rs < 3516) {
return 4;
}
else if (Rs >= 3516) {
return 8;
}
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 3967) {
return 2;
}
else if (Rs >= 3967) {
return 4;
}
}
}
else if (posBalance >= 61) {
return 2;
}
}
else if (tcor >= 77) {
if (posBalance < 59) {
if (negBalance < 64) {
return 2;
}
else if (negBalance >= 64) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
else if (posBalance >= 59) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
if (Cs < 1469) {
return 4;
}
else if (Cs >= 1469) {
return 8;
}
}
}
}
}
else if (MVDiff >= 17) {
if (tcor < 72) {
if (AFD < 23) {
if (diffRsCsdiff < 60) {
return 8;
}
else if (diffRsCsdiff >= 60) {
if (Rs < 4137) {
return 1;
}
else if (Rs >= 4137) {
return 4;
}
}
}
else if (AFD >= 23) {
if (ssDCval < 119) {
if (refDCval < 103) {
return 4;
}
else if (refDCval >= 103) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (ssDCval >= 119) {
if (AFD < 34) {
if (Cs < 1410) {
if (Rs < 4378) {
return 2;
}
else if (Rs >= 4378) {
return 8;
}
}
else if (Cs >= 1410) {
return 4;
}
}
else if (AFD >= 34) {
if (AFD < 40) {
if (AFD < 37) {
if (Rs < 4618) {
return 1;
}
else if (Rs >= 4618) {
return 4;
}
}
else if (AFD >= 37) {
return 2;
}
}
else if (AFD >= 40) {
return 1;
}
}
}
}
}
else if (tcor >= 72) {
if (TSC < 15) {
if (SC < 7205) {
if (diffRsCsdiff < -71) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 3589) {
return 8;
}
else if (Rs >= 3589) {
return 1;
}
}
}
else if (diffRsCsdiff >= -71) {
if (negBalance < 51) {
if (ssDCval < 135) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
return 8;
}
}
else if (ssDCval >= 135) {
if (ssDCval < 149) {
return 4;
}
else if (ssDCval >= 149) {
if (Rs < 4405) {
return 1;
}
else if (Rs >= 4405) {
return 8;
}
}
}
}
else if (negBalance >= 51) {
if (Cs < 1627) {
if (MVDiff < 27) {
return 1;
}
else if (MVDiff >= 27) {
return 8;
}
}
else if (Cs >= 1627) {
return 1;
}
}
}
}
else if (SC >= 7205) {
if (Cs < 1527) {
if (diffMVdiffVal < 7) {
if (diffAFD < -4) {
if (Rs < 6480) {
return 2;
}
else if (Rs >= 6480) {
if (Rs < 6608) {
return 8;
}
else if (Rs >= 6608) {
return 2;
}
}
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (diffMVdiffVal >= 7) {
return 2;
}
}
else if (Cs >= 1527) {
return 8;
}
}
}
else if (TSC >= 15) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1874) {
if (AFD < 12) {
if (RsDiff < 9) {
if (Rs < 3843) {
if (refDCval < 70) {
if (mcTcor < 93) {
if (Cs < 2753) {
if (TSC < 8) {
if (Rs < 2735) {
if (Rs < 2049) {
return 1;
}
else if (Rs >= 2049) {
return 4;
}
}
else if (Rs >= 2735) {
return 1;
}
}
else if (TSC >= 8) {
if (Cs < 2235) {
return 4;
}
else if (Cs >= 2235) {
return 8;
}
}
}
else if (Cs >= 2753) {
return 4;
}
}
else if (mcTcor >= 93) {
if (SC < 6420) {
return 1;
}
else if (SC >= 6420) {
return 8;
}
}
}
else if (refDCval >= 70) {
if (SC < 5476) {
if (ssDCval < 89) {
if (TSC < 10) {
if (SC < 4820) {
if (SC < 3939) {
return 1;
}
else if (SC >= 3939) {
return 8;
}
}
else if (SC >= 4820) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (SC >= 5476) {
if (Cs < 1968) {
return 1;
}
else if (Cs >= 1968) {
if (MVDiff < 5) {
if (ssDCval < 81) {
if (SC < 5757) {
if (SC < 5576) {
return 8;
}
else if (SC >= 5576) {
return 4;
}
}
else if (SC >= 5757) {
return 8;
}
}
else if (ssDCval >= 81) {
if (ssDCval < 113) {
if (Cs < 3958) {
if (posBalance < 55) {
if (Rs < 2792) {
return 8;
}
else if (Rs >= 2792) {
return 4;
}
}
else if (posBalance >= 55) {
return 4;
}
}
else if (Cs >= 3958) {
return 8;
}
}
else if (ssDCval >= 113) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (posBalance < 39) {
if (Cs < 4808) {
return 4;
}
else if (Cs >= 4808) {
return 8;
}
}
else if (posBalance >= 39) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3843) {
if (ssDCval < 99) {
if (TSC < 7) {
if (Rs < 5907) {
if (Rs < 4547) {
return 8;
}
else if (Rs >= 4547) {
return 1;
}
}
else if (Rs >= 5907) {
return 2;
}
}
else if (TSC >= 7) {
if (SC < 7194) {
return 4;
}
else if (SC >= 7194) {
return 8;
}
}
}
else if (ssDCval >= 99) {
if (TSC < 9) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (diffTSC < -1) {
if (Rs < 4109) {
return 8;
}
else if (Rs >= 4109) {
return 1;
}
}
else if (diffTSC >= -1) {
if (Rs < 5817) {
return 8;
}
else if (Rs >= 5817) {
return 4;
}
}
}
}
else if (TSC >= 9) {
if (mu_mv_mag_sq < 80) {
return 8;
}
else if (mu_mv_mag_sq >= 80) {
if (diffRsCsdiff < 31) {
return 1;
}
else if (diffRsCsdiff >= 31) {
if (mu_mv_mag_sq < 92) {
if (Rs < 4666) {
return 8;
}
else if (Rs >= 4666) {
return 1;
}
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 113) {
if (tcor < 90) {
if (negBalance < 35) {
if (ssDCval < 74) {
if (CsDiff < 13) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (TSC < 9) {
if (Rs < 4130) {
return 1;
}
else if (Rs >= 4130) {
return 4;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (CsDiff >= 13) {
return 8;
}
}
else if (ssDCval >= 74) {
if (tcor < 85) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (Rs < 4681) {
return 2;
}
else if (Rs >= 4681) {
return 4;
}
}
}
else if (tcor >= 85) {
return 2;
}
}
}
else if (negBalance >= 35) {
if (ssDCval < 80) {
if (negBalance < 42) {
if (Cs < 3770) {
if (mu_mv_mag_sq < 71) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (mu_mv_mag_sq >= 71) {
return 4;
}
}
else if (Cs >= 3770) {
return 8;
}
}
else if (negBalance >= 42) {
if (mu_mv_mag_sq < 75) {
if (ssDCval < 73) {
if (Rs < 3754) {
return 1;
}
else if (Rs >= 3754) {
return 8;
}
}
else if (ssDCval >= 73) {
return 4;
}
}
else if (mu_mv_mag_sq >= 75) {
return 8;
}
}
}
else if (ssDCval >= 80) {
if (MVDiff < 8) {
if (Rs < 4033) {
if (posBalance < 62) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
return 4;
}
}
else if (posBalance >= 62) {
return 4;
}
}
else if (Rs >= 4033) {
if (Cs < 3255) {
return 1;
}
else if (Cs >= 3255) {
if (Rs < 5262) {
return 4;
}
else if (Rs >= 5262) {
return 8;
}
}
}
}
else if (MVDiff >= 8) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 8;
}
}
}
}
}
else if (tcor >= 90) {
if (negBalance < 43) {
if (refDCval < 72) {
if (RsCsDiff < 269) {
return 4;
}
else if (RsCsDiff >= 269) {
if (Rs < 7774) {
return 8;
}
else if (Rs >= 7774) {
return 2;
}
}
}
else if (refDCval >= 72) {
if (Cs < 4732) {
if (tcor < 91) {
if (RsCsDiff < 176) {
if (Cs < 1917) {
return 2;
}
else if (Cs >= 1917) {
return 8;
}
}
else if (RsCsDiff >= 176) {
return 2;
}
}
else if (tcor >= 91) {
if (Cs < 3175) {
if (RsCsDiff < 111) {
if (Rs < 4650) {
return 8;
}
else if (Rs >= 4650) {
return 2;
}
}
else if (RsCsDiff >= 111) {
return 2;
}
}
else if (Cs >= 3175) {
if (MVDiff < 3) {
if (Rs < 5291) {
return 4;
}
else if (Rs >= 5291) {
return 1;
}
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
else if (Cs >= 4732) {
return 8;
}
}
}
else if (negBalance >= 43) {
if (Cs < 2441) {
if (Rs < 4348) {
if (Cs < 1952) {
if (Rs < 3618) {
return 4;
}
else if (Rs >= 3618) {
return 2;
}
}
else if (Cs >= 1952) {
if (diffTSC < 1) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (Rs >= 4348) {
if (Rs < 5649) {
if (Cs < 2201) {
return 2;
}
else if (Cs >= 2201) {
return 1;
}
}
else if (Rs >= 5649) {
return 4;
}
}
}
else if (Cs >= 2441) {
if (Rs < 4123) {
return 4;
}
else if (Rs >= 4123) {
if (Rs < 5829) {
return 8;
}
else if (Rs >= 5829) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (negBalance < 47) {
if (Rs < 4966) {
if (negBalance < 32) {
if (mu_mv_mag_sq < 58) {
if (Cs < 2998) {
return 2;
}
else if (Cs >= 2998) {
if (Cs < 3948) {
return 4;
}
else if (Cs >= 3948) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (Cs < 3526) {
return 8;
}
else if (Cs >= 3526) {
return 2;
}
}
}
else if (negBalance >= 32) {
if (diffRsCsdiff < -177) {
if (Rs < 3755) {
return 1;
}
else if (Rs >= 3755) {
return 2;
}
}
else if (diffRsCsdiff >= -177) {
if (RsCsDiff < 151) {
if (Cs < 2785) {
return 8;
}
else if (Cs >= 2785) {
return 1;
}
}
else if (RsCsDiff >= 151) {
return 8;
}
}
}
}
else if (Rs >= 4966) {
if (negBalance < 28) {
return 8;
}
else if (negBalance >= 28) {
if (MVDiff < 2) {
if (Rs < 6965) {
if (Rs < 5775) {
return 8;
}
else if (Rs >= 5775) {
return 4;
}
}
else if (Rs >= 6965) {
return 2;
}
}
else if (MVDiff >= 2) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
if (Rs < 8427) {
return 4;
}
else if (Rs >= 8427) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 47) {
if (ssDCval < 136) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
if (Rs < 7068) {
return 2;
}
else if (Rs >= 7068) {
return 1;
}
}
}
else if (ssDCval >= 136) {
if (Cs < 3374) {
return 1;
}
else if (Cs >= 3374) {
return 8;
}
}
}
}
}
}
else if (AFD >= 12) {
if (AFD < 18) {
if (RsDiff < 15) {
if (Rs < 1455) {
if (tcor < 81) {
if (Cs < 2891) {
if (ssDCval < 89) {
if (TSC < 8) {
if (Rs < 1089) {
return 4;
}
else if (Rs >= 1089) {
return 8;
}
}
else if (TSC >= 8) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (ssDCval >= 89) {
if (Rs < 1062) {
if (Rs < 875) {
return 1;
}
else if (Rs >= 875) {
return 2;
}
}
else if (Rs >= 1062) {
return 8;
}
}
}
else if (Cs >= 2891) {
if (Rs < 1284) {
return 8;
}
else if (Rs >= 1284) {
if (Rs < 1362) {
return 1;
}
else if (Rs >= 1362) {
return 2;
}
}
}
}
else if (tcor >= 81) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (RsCsDiff < 190) {
if (Rs < 863) {
return 1;
}
else if (Rs >= 863) {
return 8;
}
}
else if (RsCsDiff >= 190) {
return 2;
}
}
}
}
else if (Rs >= 1455) {
if (refDCval < 52) {
if (negBalance < 56) {
if (CsDiff < 9) {
if (Rs < 3511) {
if (Rs < 3080) {
return 2;
}
else if (Rs >= 3080) {
return 8;
}
}
else if (Rs >= 3511) {
return 4;
}
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (negBalance >= 56) {
if (Rs < 2053) {
return 2;
}
else if (Rs >= 2053) {
return 4;
}
}
}
else if (refDCval >= 52) {
if (mcTcor < 90) {
if (SC < 6344) {
if (tcor < 64) {
if (Rs < 2602) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 2;
}
}
else if (Rs >= 2602) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
else if (tcor >= 64) {
if (TSC < 15) {
if (mcTcor < 86) {
if (negBalance < 57) {
if (Rs < 1798) {
return 4;
}
else if (Rs >= 1798) {
if (Cs < 3351) {
if (Rs < 3060) {
if (Rs < 2127) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Rs < 2060) {
if (Cs < 2336) {
return 8;
}
else if (Cs >= 2336) {
return 2;
}
}
else if (Rs >= 2060) {
return 4;
}
}
}
else if (Rs >= 2127) {
return 8;
}
}
else if (Rs >= 3060) {
if (tcor < 81) {
if (mu_mv_mag_sq < 58) {
return 8;
}
else if (mu_mv_mag_sq >= 58) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (tcor >= 81) {
if (Cs < 2766) {
return 2;
}
else if (Cs >= 2766) {
return 8;
}
}
}
}
else if (Cs >= 3351) {
return 4;
}
}
}
else if (negBalance >= 57) {
if (posBalance < 59) {
if (Rs < 3038) {
return 8;
}
else if (Rs >= 3038) {
return 4;
}
}
else if (posBalance >= 59) {
if (ssDCval < 79) {
if (Cs < 2184) {
return 8;
}
else if (Cs >= 2184) {
return 2;
}
}
else if (ssDCval >= 79) {
return 4;
}
}
}
}
else if (mcTcor >= 86) {
if (Rs < 3119) {
if (posBalance < 44) {
if (Rs < 2313) {
return 4;
}
else if (Rs >= 2313) {
return 2;
}
}
else if (posBalance >= 44) {
if (MVDiff < 3) {
if (Rs < 2334) {
return 1;
}
else if (Rs >= 2334) {
if (Cs < 3168) {
return 8;
}
else if (Cs >= 3168) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 2) {
if (Cs < 2215) {
if (ssDCval < 82) {
if (Rs < 2724) {
return 8;
}
else if (Rs >= 2724) {
return 4;
}
}
else if (ssDCval >= 82) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
return 8;
}
}
}
else if (Cs >= 2215) {
if (posBalance < 64) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (negBalance < 52) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 8;
}
}
else if (negBalance >= 52) {
return 8;
}
}
}
else if (posBalance >= 64) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (mu_mv_mag_sq < 53) {
if (diffMVdiffVal < -5) {
if (Rs < 2234) {
return 1;
}
else if (Rs >= 2234) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 53) {
if (RsCsDiff < 238) {
if (mu_mv_mag_sq < 63) {
return 8;
}
else if (mu_mv_mag_sq >= 63) {
if (tcor < 84) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
if (refDCval < 83) {
return 8;
}
else if (refDCval >= 83) {
return 4;
}
}
}
else if (tcor >= 84) {
return 8;
}
}
}
else if (RsCsDiff >= 238) {
return 4;
}
}
}
}
}
}
else if (Rs >= 3119) {
if (Rs < 3529) {
if (ssDCval < 93) {
if (negBalance < 49) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
else if (negBalance >= 49) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (Rs < 3314) {
return 4;
}
else if (Rs >= 3314) {
return 1;
}
}
}
}
else if (ssDCval >= 93) {
if (ssDCval < 103) {
return 2;
}
else if (ssDCval >= 103) {
if (Cs < 2308) {
return 8;
}
else if (Cs >= 2308) {
return 4;
}
}
}
}
else if (Rs >= 3529) {
if (ssDCval < 140) {
if (negBalance < 54) {
if (Cs < 2577) {
return 8;
}
else if (Cs >= 2577) {
return 4;
}
}
else if (negBalance >= 54) {
if (Rs < 3833) {
return 4;
}
else if (Rs >= 3833) {
return 8;
}
}
}
else if (ssDCval >= 140) {
return 4;
}
}
}
}
}
else if (TSC >= 15) {
if (Rs < 3094) {
return 4;
}
else if (Rs >= 3094) {
return 8;
}
}
}
}
else if (SC >= 6344) {
if (ssDCval < 123) {
if (refDCval < 101) {
if (negBalance < 50) {
if (SC < 7438) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
if (Rs < 3774) {
return 1;
}
else if (Rs >= 3774) {
if (Rs < 4042) {
return 4;
}
else if (Rs >= 4042) {
return 8;
}
}
}
}
else if (SC >= 7438) {
if (Cs < 4421) {
if (MVDiff < 19) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (Cs >= 4421) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (ssDCval < 97) {
if (mu_mv_mag_sq < 65) {
if (tcor < 78) {
if (Cs < 3249) {
return 8;
}
else if (Cs >= 3249) {
return 4;
}
}
else if (tcor >= 78) {
if (Rs < 5261) {
if (mu_mv_mag_sq < 63) {
return 8;
}
else if (mu_mv_mag_sq >= 63) {
return 1;
}
}
else if (Rs >= 5261) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 65) {
return 8;
}
}
else if (ssDCval >= 97) {
return 4;
}
}
}
else if (refDCval >= 101) {
if (mu_mv_mag_sq < 72) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
return 4;
}
}
else if (mu_mv_mag_sq >= 72) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
return 1;
}
}
}
}
else if (ssDCval >= 123) {
if (tcor < 77) {
if (mu_mv_mag_sq < 15) {
return 8;
}
else if (mu_mv_mag_sq >= 15) {
if (negBalance < 43) {
if (Cs < 2724) {
return 1;
}
else if (Cs >= 2724) {
return 8;
}
}
else if (negBalance >= 43) {
if (mu_mv_mag_sq < 41) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 8;
}
}
else if (mu_mv_mag_sq >= 41) {
if (Cs < 3776) {
if (diffRsCsdiff < -11) {
if (Rs < 3838) {
return 8;
}
else if (Rs >= 3838) {
return 1;
}
}
else if (diffRsCsdiff >= -11) {
return 8;
}
}
else if (Cs >= 3776) {
return 1;
}
}
}
}
}
else if (tcor >= 77) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
}
else if (mcTcor >= 90) {
if (RsCsDiff < 195) {
if (ssDCval < 96) {
if (MVDiff < 7) {
if (negBalance < 56) {
if (TSC < 11) {
if (AFD < 16) {
if (mu_mv_mag_sq < 27) {
if (Cs < 2521) {
if (Rs < 3815) {
return 4;
}
else if (Rs >= 3815) {
return 1;
}
}
else if (Cs >= 2521) {
if (RsDiff < 8) {
if (Rs < 2664) {
return 4;
}
else if (Rs >= 2664) {
return 1;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (ssDCval < 86) {
if (Rs < 4815) {
return 1;
}
else if (Rs >= 4815) {
return 8;
}
}
else if (ssDCval >= 86) {
return 4;
}
}
}
else if (AFD >= 16) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 2231) {
return 2;
}
else if (Rs >= 2231) {
return 1;
}
}
}
}
}
else if (TSC >= 11) {
if (Rs < 3014) {
return 8;
}
else if (Rs >= 3014) {
if (Rs < 3916) {
return 4;
}
else if (Rs >= 3916) {
return 8;
}
}
}
}
else if (negBalance >= 56) {
return 8;
}
}
else if (MVDiff >= 7) {
if (Rs < 2669) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
}
else if (Rs >= 2669) {
if (Cs < 2374) {
if (tcor < 82) {
if (SC < 5249) {
return 2;
}
else if (SC >= 5249) {
if (Rs < 5614) {
return 8;
}
else if (Rs >= 5614) {
return 4;
}
}
}
else if (tcor >= 82) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
return 2;
}
}
}
}
else if (Cs >= 2374) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
if (Rs < 2927) {
return 8;
}
else if (Rs >= 2927) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 96) {
if (negBalance < 41) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
return 8;
}
}
else if (negBalance >= 41) {
if (Cs < 2734) {
if (diffRsCsdiff < 137) {
if (mu_mv_mag_sq < 80) {
if (Rs < 3743) {
if (Cs < 2386) {
if (negBalance < 43) {
if (Rs < 2198) {
return 1;
}
else if (Rs >= 2198) {
return 4;
}
}
else if (negBalance >= 43) {
if (tcor < 94) {
if (ssDCval < 102) {
if (SC < 3832) {
return 4;
}
else if (SC >= 3832) {
return 8;
}
}
else if (ssDCval >= 102) {
return 8;
}
}
else if (tcor >= 94) {
return 2;
}
}
}
else if (Cs >= 2386) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (negBalance < 51) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (mu_mv_mag_sq < 64) {
return 4;
}
else if (mu_mv_mag_sq >= 64) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
}
}
else if (Rs >= 3743) {
return 8;
}
}
else if (mu_mv_mag_sq >= 80) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (Rs < 3568) {
return 2;
}
else if (Rs >= 3568) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 137) {
if (Rs < 2805) {
return 8;
}
else if (Rs >= 2805) {
return 1;
}
}
}
else if (Cs >= 2734) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (diffAFD < 9) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
if (Rs < 4322) {
return 4;
}
else if (Rs >= 4322) {
return 8;
}
}
}
else if (diffAFD >= 9) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 195) {
if (Rs < 5310) {
if (Rs < 3505) {
if (Cs < 2799) {
if (SC < 4651) {
return 8;
}
else if (SC >= 4651) {
if (tcor < 90) {
if (Rs < 3138) {
if (CsDiff < 11) {
if (posBalance < 53) {
if (Rs < 3004) {
return 4;
}
else if (Rs >= 3004) {
return 2;
}
}
else if (posBalance >= 53) {
if (SC < 4840) {
if (RsDiff < 12) {
return 2;
}
else if (RsDiff >= 12) {
return 8;
}
}
else if (SC >= 4840) {
return 8;
}
}
}
else if (CsDiff >= 11) {
if (Cs < 2138) {
return 8;
}
else if (Cs >= 2138) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
}
else if (Rs >= 3138) {
if (gchDC < 6) {
return 4;
}
else if (gchDC >= 6) {
return 8;
}
}
}
else if (tcor >= 90) {
return 8;
}
}
}
else if (Cs >= 2799) {
if (mcTcor < 94) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (Rs < 3236) {
return 4;
}
else if (Rs >= 3236) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 3042) {
return 8;
}
else if (Rs >= 3042) {
return 1;
}
}
}
}
else if (Rs >= 3505) {
if (mu_mv_mag_sq < 76) {
if (negBalance < 45) {
if (tcor < 82) {
if (Rs < 5042) {
return 4;
}
else if (Rs >= 5042) {
return 8;
}
}
else if (tcor >= 82) {
if (Rs < 5115) {
return 2;
}
else if (Rs >= 5115) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (MVDiff < 5) {
if (Cs < 3343) {
if (SC < 7532) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
else if (SC >= 7532) {
if (Rs < 5086) {
return 8;
}
else if (Rs >= 5086) {
return 4;
}
}
}
else if (Cs >= 3343) {
return 2;
}
}
else if (MVDiff >= 5) {
if (Cs < 2067) {
return 2;
}
else if (Cs >= 2067) {
if (mu_mv_mag_sq < 50) {
return 8;
}
else if (mu_mv_mag_sq >= 50) {
if (tcor < 86) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 4;
}
}
else if (tcor >= 86) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (TSC < 11) {
if (ssDCval < 86) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 4;
}
}
else if (ssDCval >= 86) {
if (Rs < 3895) {
return 1;
}
else if (Rs >= 3895) {
return 4;
}
}
}
else if (TSC >= 11) {
return 8;
}
}
}
}
else if (Rs >= 5310) {
if (mcTcor < 94) {
if (RsDiff < 14) {
if (mu_mv_mag_sq < 56) {
return 8;
}
else if (mu_mv_mag_sq >= 56) {
return 4;
}
}
else if (RsDiff >= 14) {
return 4;
}
}
else if (mcTcor >= 94) {
return 4;
}
}
}
}
}
}
}
else if (RsDiff >= 15) {
if (tcor < 80) {
if (AFD < 14) {
if (Rs < 3035) {
return 2;
}
else if (Rs >= 3035) {
return 8;
}
}
else if (AFD >= 14) {
if (tcor < 68) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (tcor >= 68) {
if (MVDiff < 8) {
if (tcor < 79) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (tcor >= 79) {
return 4;
}
}
else if (MVDiff >= 8) {
if (diffTSC < -4) {
return 8;
}
else if (diffTSC >= -4) {
if (diffAFD < 2) {
if (TSC < 15) {
return 4;
}
else if (TSC >= 15) {
return 2;
}
}
else if (diffAFD >= 2) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 80) {
if (CsDiff < 11) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (MVDiff < 13) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
if (AFD < 17) {
if (Rs < 5507) {
return 1;
}
else if (Rs >= 5507) {
return 8;
}
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (MVDiff >= 13) {
if (diffRsCsdiff < -35) {
return 2;
}
else if (diffRsCsdiff >= -35) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 4;
}
}
}
}
}
else if (CsDiff >= 11) {
if (Rs < 8415) {
if (mcTcor < 85) {
if (Cs < 2928) {
return 4;
}
else if (Cs >= 2928) {
return 2;
}
}
else if (mcTcor >= 85) {
if (Rs < 5410) {
if (refDCval < 84) {
if (tcor < 83) {
if (SC < 8221) {
if (Rs < 3934) {
if (Rs < 3489) {
return 4;
}
else if (Rs >= 3489) {
return 8;
}
}
else if (Rs >= 3934) {
return 2;
}
}
else if (SC >= 8221) {
return 8;
}
}
else if (tcor >= 83) {
return 8;
}
}
else if (refDCval >= 84) {
if (diffMVdiffVal < 3) {
if (Cs < 2924) {
return 2;
}
else if (Cs >= 2924) {
return 1;
}
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
}
else if (Rs >= 5410) {
if (SC < 9713) {
if (mu_mv_mag_sq < 67) {
if (Rs < 5658) {
return 8;
}
else if (Rs >= 5658) {
return 1;
}
}
else if (mu_mv_mag_sq >= 67) {
return 4;
}
}
else if (SC >= 9713) {
if (RsCsDiff < 420) {
return 4;
}
else if (RsCsDiff >= 420) {
return 8;
}
}
}
}
}
else if (Rs >= 8415) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 4;
}
}
}
}
}
}
else if (AFD >= 18) {
if (refDCval < 156) {
if (TSC < 21) {
if (RsCsDiff < 639) {
if (Cs < 3573) {
if (refDCval < 107) {
if (ssDCval < 67) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (negBalance < 50) {
if (TSC < 13) {
if (diffRsCsdiff < 191) {
return 2;
}
else if (diffRsCsdiff >= 191) {
return 1;
}
}
else if (TSC >= 13) {
return 4;
}
}
else if (negBalance >= 50) {
if (MVDiff < 26) {
if (refDCval < 65) {
if (Rs < 3701) {
return 4;
}
else if (Rs >= 3701) {
return 2;
}
}
else if (refDCval >= 65) {
if (Rs < 2273) {
return 1;
}
else if (Rs >= 2273) {
return 4;
}
}
}
else if (MVDiff >= 26) {
if (diffRsCsdiff < 201) {
if (TSC < 12) {
if (MVDiff < 30) {
if (Rs < 1541) {
return 8;
}
else if (Rs >= 1541) {
return 1;
}
}
else if (MVDiff >= 30) {
return 2;
}
}
else if (TSC >= 12) {
if (AFD < 20) {
if (mu_mv_mag_sq < 2) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
return 2;
}
}
else if (AFD >= 20) {
return 4;
}
}
}
else if (diffRsCsdiff >= 201) {
return 8;
}
}
}
}
}
else if (ssDCval >= 67) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 2) {
if (Rs < 4872) {
return 4;
}
else if (Rs >= 4872) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 2973) {
if (tcor < 69) {
return 4;
}
else if (tcor >= 69) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
return 4;
}
}
}
else if (Cs >= 2973) {
if (Rs < 6040) {
return 8;
}
else if (Rs >= 6040) {
return 4;
}
}
}
}
else if (MVDiff >= 5) {
if (negBalance < 57) {
if (posBalance < 51) {
if (ssDCval < 86) {
if (mu_mv_mag_sq < 54) {
if (Cs < 2106) {
return 4;
}
else if (Cs >= 2106) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
if (Rs < 6507) {
return 2;
}
else if (Rs >= 6507) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 54) {
return 4;
}
}
else if (ssDCval >= 86) {
return 1;
}
}
else if (posBalance >= 51) {
if (mcTcor < 94) {
if (refDCval < 78) {
if (posBalance < 61) {
if (diffRsCsdiff < -135) {
return 2;
}
else if (diffRsCsdiff >= -135) {
if (RsCsDiff < 326) {
if (RsCsDiff < 243) {
if (mu_mv_mag_sq < 42) {
return 8;
}
else if (mu_mv_mag_sq >= 42) {
return 1;
}
}
else if (RsCsDiff >= 243) {
if (MVDiff < 18) {
if (Cs < 3010) {
if (TSC < 14) {
return 1;
}
else if (TSC >= 14) {
return 8;
}
}
else if (Cs >= 3010) {
return 8;
}
}
else if (MVDiff >= 18) {
return 8;
}
}
}
else if (RsCsDiff >= 326) {
if (gchDC < 2) {
if (Rs < 2370) {
return 4;
}
else if (Rs >= 2370) {
if (Rs < 5273) {
return 8;
}
else if (Rs >= 5273) {
return 2;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (posBalance >= 61) {
if (SC < 4884) {
if (posBalance < 65) {
if (mcTcor < 92) {
return 2;
}
else if (mcTcor >= 92) {
return 8;
}
}
else if (posBalance >= 65) {
if (SC < 4318) {
if (Rs < 1831) {
return 2;
}
else if (Rs >= 1831) {
return 4;
}
}
else if (SC >= 4318) {
return 1;
}
}
}
else if (SC >= 4884) {
if (RsCsDiff < 605) {
if (mu_mv_mag_sq < 12) {
if (CsDiff < 13) {
return 2;
}
else if (CsDiff >= 13) {
if (Rs < 2791) {
return 4;
}
else if (Rs >= 2791) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 12) {
return 8;
}
}
else if (RsCsDiff >= 605) {
if (Cs < 2740) {
return 1;
}
else if (Cs >= 2740) {
return 4;
}
}
}
}
}
else if (refDCval >= 78) {
if (TSC < 12) {
if (ssDCval < 96) {
if (mu_mv_mag_sq < 12) {
if (SC < 5903) {
if (diffRsCsdiff < 69) {
if (SC < 3913) {
if (Cs < 2150) {
return 8;
}
else if (Cs >= 2150) {
return 4;
}
}
else if (SC >= 3913) {
if (mcTcor < 91) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (mcTcor >= 91) {
return 4;
}
}
}
else if (diffRsCsdiff >= 69) {
return 8;
}
}
else if (SC >= 5903) {
return 8;
}
}
else if (mu_mv_mag_sq >= 12) {
if (RsDiff < 16) {
return 8;
}
else if (RsDiff >= 16) {
return 4;
}
}
}
else if (ssDCval >= 96) {
if (diffAFD < 2) {
if (Rs < 5951) {
return 8;
}
else if (Rs >= 5951) {
return 4;
}
}
else if (diffAFD >= 2) {
if (Cs < 2611) {
return 2;
}
else if (Cs >= 2611) {
if (Rs < 3456) {
return 4;
}
else if (Rs >= 3456) {
return 8;
}
}
}
}
}
else if (TSC >= 12) {
if (Rs < 2979) {
if (ssDCval < 87) {
if (mu_mv_mag_sq < 7) {
if (posBalance < 62) {
if (Rs < 2616) {
return 8;
}
else if (Rs >= 2616) {
return 4;
}
}
else if (posBalance >= 62) {
return 4;
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 2534) {
return 8;
}
else if (Cs >= 2534) {
if (Rs < 1860) {
return 2;
}
else if (Rs >= 1860) {
return 1;
}
}
}
}
else if (ssDCval >= 87) {
if (AFD < 22) {
if (Rs < 2945) {
return 8;
}
else if (Rs >= 2945) {
return 2;
}
}
else if (AFD >= 22) {
if (MVDiff < 21) {
return 1;
}
else if (MVDiff >= 21) {
return 8;
}
}
}
}
else if (Rs >= 2979) {
if (diffAFD < 4) {
if (mcTcor < 81) {
if (Rs < 5838) {
return 8;
}
else if (Rs >= 5838) {
return 1;
}
}
else if (mcTcor >= 81) {
if (CsDiff < 13) {
if (mcTcor < 91) {
if (ssDCval < 107) {
if (mu_mv_mag_sq < 49) {
return 4;
}
else if (mu_mv_mag_sq >= 49) {
return 2;
}
}
else if (ssDCval >= 107) {
return 1;
}
}
else if (mcTcor >= 91) {
if (Cs < 2779) {
return 8;
}
else if (Cs >= 2779) {
return 4;
}
}
}
else if (CsDiff >= 13) {
if (tcor < 72) {
return 4;
}
else if (tcor >= 72) {
return 8;
}
}
}
}
else if (diffAFD >= 4) {
if (CsDiff < 15) {
return 8;
}
else if (CsDiff >= 15) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 1697) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1054) {
return 2;
}
else if (Rs >= 1054) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (Rs < 1116) {
return 2;
}
else if (Rs >= 1116) {
return 4;
}
}
}
}
else if (Rs >= 1697) {
if (Cs < 2506) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (ssDCval < 93) {
if (Rs < 2020) {
return 2;
}
else if (Rs >= 2020) {
return 8;
}
}
else if (ssDCval >= 93) {
return 2;
}
}
}
else if (Cs >= 2506) {
if (Cs < 2741) {
return 8;
}
else if (Cs >= 2741) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 57) {
if (mcTcor < 79) {
if (MVDiff < 35) {
return 4;
}
else if (MVDiff >= 35) {
if (Rs < 2558) {
return 1;
}
else if (Rs >= 2558) {
return 8;
}
}
}
else if (mcTcor >= 79) {
if (tcor < 65) {
if (Rs < 1656) {
if (Rs < 917) {
if (refDCval < 91) {
return 8;
}
else if (refDCval >= 91) {
return 2;
}
}
else if (Rs >= 917) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
if (Rs < 1435) {
return 2;
}
else if (Rs >= 1435) {
return 8;
}
}
}
}
else if (Rs >= 1656) {
if (negBalance < 62) {
if (MVDiff < 24) {
return 4;
}
else if (MVDiff >= 24) {
if (Cs < 1941) {
if (Rs < 2697) {
return 8;
}
else if (Rs >= 2697) {
return 4;
}
}
else if (Cs >= 1941) {
return 1;
}
}
}
else if (negBalance >= 62) {
if (AFD < 25) {
return 4;
}
else if (AFD >= 25) {
return 1;
}
}
}
}
else if (tcor >= 65) {
if (mu_mv_mag_sq < 13) {
if (Rs < 2035) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
else if (Rs >= 2035) {
if (TSC < 15) {
if (posBalance < 60) {
if (mcTcor < 93) {
if (RsDiff < 17) {
if (TSC < 14) {
if (CsDiff < 13) {
if (SC < 5541) {
if (SC < 5431) {
if (Rs < 2206) {
return 8;
}
else if (Rs >= 2206) {
return 4;
}
}
else if (SC >= 5431) {
return 8;
}
}
else if (SC >= 5541) {
return 4;
}
}
else if (CsDiff >= 13) {
return 1;
}
}
else if (TSC >= 14) {
if (Rs < 4239) {
return 1;
}
else if (Rs >= 4239) {
return 4;
}
}
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (posBalance >= 60) {
if (Rs < 2141) {
return 4;
}
else if (Rs >= 2141) {
return 8;
}
}
}
else if (TSC >= 15) {
if (Rs < 2301) {
return 4;
}
else if (Rs >= 2301) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 3276) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 4;
}
}
else if (Rs >= 3276) {
if (diffMVdiffVal < -7) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 107) {
if (SC < 5647) {
if (posBalance < 58) {
if (MVDiff < 21) {
if (Rs < 2849) {
if (Rs < 2674) {
if (SC < 3962) {
return 1;
}
else if (SC >= 3962) {
return 8;
}
}
else if (Rs >= 2674) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
}
else if (Rs >= 2849) {
return 8;
}
}
else if (MVDiff >= 21) {
return 8;
}
}
else if (posBalance >= 58) {
if (negBalance < 61) {
if (mu_mv_mag_sq < 14) {
if (posBalance < 62) {
if (diffAFD < -3) {
if (Rs < 2760) {
return 2;
}
else if (Rs >= 2760) {
return 1;
}
}
else if (diffAFD >= -3) {
if (AFD < 27) {
if (ssDCval < 105) {
return 2;
}
else if (ssDCval >= 105) {
return 8;
}
}
else if (AFD >= 27) {
return 2;
}
}
}
else if (posBalance >= 62) {
if (negBalance < 56) {
if (ssDCval < 115) {
if (Rs < 2342) {
return 8;
}
else if (Rs >= 2342) {
return 2;
}
}
else if (ssDCval >= 115) {
if (Cs < 1952) {
return 2;
}
else if (Cs >= 1952) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (Rs < 2596) {
return 2;
}
else if (Rs >= 2596) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (mcTcor < 92) {
if (diffMVdiffVal < 31) {
return 8;
}
else if (diffMVdiffVal >= 31) {
if (Rs < 2561) {
return 8;
}
else if (Rs >= 2561) {
return 2;
}
}
}
else if (mcTcor >= 92) {
if (Rs < 2520) {
return 8;
}
else if (Rs >= 2520) {
return 2;
}
}
}
}
else if (negBalance >= 61) {
if (RsCsDiff < 223) {
if (ssDCval < 118) {
if (Rs < 2559) {
return 8;
}
else if (Rs >= 2559) {
return 1;
}
}
else if (ssDCval >= 118) {
return 2;
}
}
else if (RsCsDiff >= 223) {
return 1;
}
}
}
}
else if (SC >= 5647) {
if (gchDC < 2) {
if (SC < 7789) {
if (AFD < 25) {
if (Cs < 3332) {
if (mu_mv_mag_sq < 16) {
if (Cs < 2260) {
return 1;
}
else if (Cs >= 2260) {
if (Rs < 4702) {
return 8;
}
else if (Rs >= 4702) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (diffRsCsdiff < -14) {
if (TSC < 15) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 1;
}
}
else if (TSC >= 15) {
return 8;
}
}
else if (diffRsCsdiff >= -14) {
if (SC < 7768) {
return 8;
}
else if (SC >= 7768) {
return 4;
}
}
}
}
else if (Cs >= 3332) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 4;
}
}
}
else if (AFD >= 25) {
if (RsDiff < 15) {
return 1;
}
else if (RsDiff >= 15) {
if (Rs < 3532) {
return 8;
}
else if (Rs >= 3532) {
return 4;
}
}
}
}
else if (SC >= 7789) {
return 8;
}
}
else if (gchDC >= 2) {
if (Cs < 3041) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
if (Cs < 2188) {
return 1;
}
else if (Cs >= 2188) {
if (Cs < 2937) {
return 8;
}
else if (Cs >= 2937) {
return 1;
}
}
}
}
else if (Cs >= 3041) {
if (Rs < 5239) {
return 2;
}
else if (Rs >= 5239) {
return 8;
}
}
}
}
}
}
else if (Cs >= 3573) {
if (posBalance < 54) {
if (RsCsDiff < 321) {
if (CsDiff < 16) {
return 8;
}
else if (CsDiff >= 16) {
return 2;
}
}
else if (RsCsDiff >= 321) {
if (mu_mv_mag_sq < 42) {
if (Rs < 6181) {
return 2;
}
else if (Rs >= 6181) {
return 4;
}
}
else if (mu_mv_mag_sq >= 42) {
if (tcor < 76) {
if (Rs < 5501) {
if (Rs < 4410) {
return 2;
}
else if (Rs >= 4410) {
return 4;
}
}
else if (Rs >= 5501) {
return 8;
}
}
else if (tcor >= 76) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
if (mu_mv_mag_sq < 29) {
if (negBalance < 55) {
if (RsCsDiff < 396) {
if (SC < 4916) {
return 2;
}
else if (SC >= 4916) {
return 8;
}
}
else if (RsCsDiff >= 396) {
if (SC < 10390) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (MVDiff < 19) {
if (RsCsDiff < 583) {
if (RsCsDiff < 469) {
if (AFD < 20) {
if (Cs < 3801) {
if (Rs < 5578) {
return 2;
}
else if (Rs >= 5578) {
return 8;
}
}
else if (Cs >= 3801) {
return 4;
}
}
else if (AFD >= 20) {
return 8;
}
}
else if (RsCsDiff >= 469) {
return 4;
}
}
else if (RsCsDiff >= 583) {
return 2;
}
}
else if (MVDiff >= 19) {
if (CsDiff < 20) {
if (Rs < 4369) {
return 2;
}
else if (Rs >= 4369) {
return 8;
}
}
else if (CsDiff >= 20) {
if (Rs < 1756) {
return 8;
}
else if (Rs >= 1756) {
return 4;
}
}
}
}
}
else if (SC >= 10390) {
return 8;
}
}
}
else if (negBalance >= 55) {
if (Cs < 3916) {
if (Cs < 3623) {
return 8;
}
else if (Cs >= 3623) {
return 4;
}
}
else if (Cs >= 3916) {
if (ssDCval < 109) {
if (refDCval < 94) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
else if (refDCval >= 94) {
if (Cs < 3989) {
return 8;
}
else if (Cs >= 3989) {
return 4;
}
}
}
else if (ssDCval >= 109) {
if (Rs < 6314) {
if (Rs < 3566) {
return 2;
}
else if (Rs >= 3566) {
return 1;
}
}
else if (Rs >= 6314) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (ssDCval < 107) {
if (Rs < 7130) {
if (RsDiff < 19) {
return 8;
}
else if (RsDiff >= 19) {
if (Rs < 5229) {
return 8;
}
else if (Rs >= 5229) {
return 2;
}
}
}
else if (Rs >= 7130) {
return 4;
}
}
else if (ssDCval >= 107) {
if (Rs < 3917) {
return 8;
}
else if (Rs >= 3917) {
if (Rs < 4232) {
return 1;
}
else if (Rs >= 4232) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 639) {
if (mcTcor < 95) {
if (SC < 10615) {
if (AFD < 30) {
if (tcor < 57) {
if (negBalance < 55) {
if (diffAFD < 2) {
if (posBalance < 48) {
if (Rs < 3607) {
return 4;
}
else if (Rs >= 3607) {
return 2;
}
}
else if (posBalance >= 48) {
return 4;
}
}
else if (diffAFD >= 2) {
if (Cs < 2926) {
return 8;
}
else if (Cs >= 2926) {
return 4;
}
}
}
else if (negBalance >= 55) {
if (MVDiff < 41) {
return 4;
}
else if (MVDiff >= 41) {
if (Rs < 2591) {
return 4;
}
else if (Rs >= 2591) {
return 1;
}
}
}
}
else if (tcor >= 57) {
if (negBalance < 56) {
if (CsDiff < 16) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
if (SC < 10178) {
if (MVDiff < 34) {
if (CsDiff < 14) {
return 2;
}
else if (CsDiff >= 14) {
return 8;
}
}
else if (MVDiff >= 34) {
if (Cs < 3006) {
return 8;
}
else if (Cs >= 3006) {
return 1;
}
}
}
else if (SC >= 10178) {
return 1;
}
}
}
else if (CsDiff >= 16) {
if (MVDiff < 57) {
if (MVDiff < 29) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 55) {
if (Rs < 4734) {
return 2;
}
else if (Rs >= 4734) {
return 8;
}
}
else if (posBalance >= 55) {
if (Rs < 5484) {
return 4;
}
else if (Rs >= 5484) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 6471) {
return 4;
}
else if (Rs >= 6471) {
return 2;
}
}
}
else if (MVDiff >= 29) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
return 8;
}
}
}
else if (MVDiff >= 57) {
if (Rs < 864) {
return 2;
}
else if (Rs >= 864) {
return 1;
}
}
}
}
else if (negBalance >= 56) {
if (mu_mv_mag_sq < 13) {
if (refDCval < 76) {
return 2;
}
else if (refDCval >= 76) {
if (Cs < 2850) {
return 8;
}
else if (Cs >= 2850) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 13) {
return 8;
}
}
}
}
else if (AFD >= 30) {
if (MVDiff < 29) {
if (posBalance < 63) {
if (Rs < 6400) {
return 2;
}
else if (Rs >= 6400) {
if (SC < 10341) {
return 4;
}
else if (SC >= 10341) {
return 2;
}
}
}
else if (posBalance >= 63) {
return 1;
}
}
else if (MVDiff >= 29) {
if (CsDiff < 22) {
if (ssDCval < 108) {
return 1;
}
else if (ssDCval >= 108) {
if (Rs < 4736) {
return 4;
}
else if (Rs >= 4736) {
return 8;
}
}
}
else if (CsDiff >= 22) {
return 8;
}
}
}
}
else if (SC >= 10615) {
if (mu_mv_mag_sq < 50) {
if (TSC < 11) {
if (Cs < 4323) {
return 4;
}
else if (Cs >= 4323) {
return 8;
}
}
else if (TSC >= 11) {
if (AFD < 41) {
return 4;
}
else if (AFD >= 41) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 6420) {
return 8;
}
else if (Rs >= 6420) {
if (gchDC < 2) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
return 8;
}
}
else if (gchDC >= 2) {
if (Cs < 4725) {
return 2;
}
else if (Cs >= 4725) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 128) {
if (ssDCval < 94) {
if (Rs < 4864) {
return 2;
}
else if (Rs >= 4864) {
return 4;
}
}
else if (ssDCval >= 94) {
if (Rs < 3731) {
return 2;
}
else if (Rs >= 3731) {
return 1;
}
}
}
else if (ssDCval >= 128) {
if (Cs < 2690) {
return 2;
}
else if (Cs >= 2690) {
return 8;
}
}
}
}
}
else if (TSC >= 21) {
if (RsDiff < 24) {
return 8;
}
else if (RsDiff >= 24) {
if (mcTcor < 68) {
if (Rs < 6778) {
return 1;
}
else if (Rs >= 6778) {
return 4;
}
}
else if (mcTcor >= 68) {
if (tcor < 60) {
if (Rs < 8576) {
return 8;
}
else if (Rs >= 8576) {
return 4;
}
}
else if (tcor >= 60) {
return 4;
}
}
}
}
}
else if (refDCval >= 156) {
if (posBalance < 33) {
if (Rs < 1455) {
return 1;
}
else if (Rs >= 1455) {
return 8;
}
}
else if (posBalance >= 33) {
if (refDCval < 166) {
if (Rs < 3129) {
return 8;
}
else if (Rs >= 3129) {
if (Cs < 1992) {
if (Rs < 5658) {
return 2;
}
else if (Rs >= 5658) {
return 4;
}
}
else if (Cs >= 1992) {
return 4;
}
}
}
else if (refDCval >= 166) {
if (RsDiff < 20) {
return 4;
}
else if (RsDiff >= 20) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Cs < 1644) {
if (Rs < 3092) {
if (Rs < 1455) {
if (ssDCval < 122) {
if (negBalance < 1) {
return 2;
}
else if (negBalance >= 1) {
if (refDCval < 64) {
if (negBalance < 27) {
if (posBalance < 27) {
if (Rs < 572) {
return 4;
}
else if (Rs >= 572) {
if (Cs < 1568) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1411) {
return 1;
}
else if (Rs >= 1411) {
return 8;
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (Cs >= 1568) {
return 1;
}
}
}
else if (posBalance >= 27) {
if (SC < 2385) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
return 8;
}
}
}
else if (SC >= 2385) {
if (posBalance < 42) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
if (Rs < 1057) {
if (Rs < 1034) {
return 1;
}
else if (Rs >= 1034) {
return 8;
}
}
else if (Rs >= 1057) {
return 1;
}
}
}
else if (posBalance >= 42) {
if (SC < 2524) {
return 8;
}
else if (SC >= 2524) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 27) {
if (SC < 2549) {
if (Rs < 771) {
return 1;
}
else if (Rs >= 771) {
return 8;
}
}
else if (SC >= 2549) {
if (ssDCval < 59) {
if (refDCval < 54) {
if (posBalance < 49) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (posBalance >= 49) {
if (Rs < 1170) {
return 4;
}
else if (Rs >= 1170) {
if (Rs < 1435) {
return 8;
}
else if (Rs >= 1435) {
return 4;
}
}
}
}
else if (refDCval >= 54) {
return 4;
}
}
else if (ssDCval >= 59) {
return 8;
}
}
}
}
else if (refDCval >= 64) {
if (ssDCval < 77) {
if (RsDiff < 3) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
if (Rs < 1308) {
if (Rs < 1091) {
return 2;
}
else if (Rs >= 1091) {
return 4;
}
}
else if (Rs >= 1308) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (MVDiff < 2) {
if (Rs < 1376) {
return 1;
}
else if (Rs >= 1376) {
if (Rs < 1427) {
return 4;
}
else if (Rs >= 1427) {
return 8;
}
}
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (ssDCval >= 77) {
if (Rs < 944) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Cs < 1358) {
return 1;
}
else if (Cs >= 1358) {
return 4;
}
}
}
else if (Rs >= 944) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 2) {
if (posBalance < 28) {
if (Cs < 1349) {
return 1;
}
else if (Cs >= 1349) {
return 8;
}
}
else if (posBalance >= 28) {
if (Rs < 1373) {
return 4;
}
else if (Rs >= 1373) {
return 8;
}
}
}
else if (AFD >= 2) {
if (mcTcor < 96) {
if (posBalance < 44) {
if (diffAFD < -1) {
if (Rs < 894) {
return 2;
}
else if (Rs >= 894) {
return 8;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (posBalance >= 44) {
if (Rs < 951) {
return 8;
}
else if (Rs >= 951) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (Rs < 990) {
return 8;
}
else if (Rs >= 990) {
if (Rs < 1004) {
return 4;
}
else if (Rs >= 1004) {
if (Cs < 1518) {
return 8;
}
else if (Cs >= 1518) {
if (Rs < 1056) {
return 4;
}
else if (Rs >= 1056) {
return 8;
}
}
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (Cs < 1405) {
return 8;
}
else if (Cs >= 1405) {
if (Rs < 1223) {
if (Rs < 1045) {
return 8;
}
else if (Rs >= 1045) {
return 4;
}
}
else if (Rs >= 1223) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1455) {
if (ssDCval < 86) {
if (mcTcor < 97) {
if (posBalance < 44) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
if (ssDCval < 46) {
return 4;
}
else if (ssDCval >= 46) {
if (AFD < 7) {
if (Rs < 2456) {
return 8;
}
else if (Rs >= 2456) {
if (Rs < 2862) {
return 4;
}
else if (Rs >= 2862) {
return 8;
}
}
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (posBalance >= 44) {
if (Rs < 2831) {
return 8;
}
else if (Rs >= 2831) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 39) {
if (Rs < 2965) {
if (negBalance < 19) {
if (Rs < 1993) {
if (Cs < 1427) {
return 8;
}
else if (Cs >= 1427) {
if (Rs < 1603) {
return 1;
}
else if (Rs >= 1603) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 1746) {
return 8;
}
else if (Rs >= 1746) {
return 1;
}
}
}
}
}
else if (Rs >= 1993) {
if (Cs < 1350) {
if (Rs < 2342) {
if (Rs < 2302) {
if (Rs < 2300) {
if (SC < 3638) {
if (SC < 3619) {
if (Rs < 2295) {
if (Cs < 1315) {
return 2;
}
else if (Cs >= 1315) {
if (Rs < 2287) {
return 8;
}
else if (Rs >= 2287) {
if (Rs < 2290) {
return 2;
}
else if (Rs >= 2290) {
return 8;
}
}
}
}
else if (Rs >= 2295) {
return 2;
}
}
else if (SC >= 3619) {
return 8;
}
}
else if (SC >= 3638) {
return 2;
}
}
else if (Rs >= 2300) {
if (Cs < 1341) {
return 2;
}
else if (Cs >= 1341) {
return 8;
}
}
}
else if (Rs >= 2302) {
return 8;
}
}
else if (Rs >= 2342) {
if (posBalance < 35) {
return 2;
}
else if (posBalance >= 35) {
return 8;
}
}
}
else if (Cs >= 1350) {
return 8;
}
}
}
else if (negBalance >= 19) {
if (Rs < 1566) {
return 1;
}
else if (Rs >= 1566) {
if (MVDiff < 1) {
if (Rs < 1800) {
return 4;
}
else if (Rs >= 1800) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
else if (Rs >= 2965) {
return 1;
}
}
else if (posBalance >= 39) {
if (negBalance < 34) {
if (Rs < 2469) {
if (ssDCval < 77) {
if (Rs < 1922) {
if (diffRsCsdiff < -2) {
if (Rs < 1505) {
return 8;
}
else if (Rs >= 1505) {
return 4;
}
}
else if (diffRsCsdiff >= -2) {
return 4;
}
}
else if (Rs >= 1922) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
return 4;
}
}
}
else if (ssDCval >= 77) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (Rs >= 2469) {
if (Rs < 2988) {
return 4;
}
else if (Rs >= 2988) {
return 2;
}
}
}
else if (negBalance >= 34) {
if (RsCsDiff < 29) {
return 8;
}
else if (RsCsDiff >= 29) {
return 4;
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (CsDiff < 2) {
if (diffAFD < -20) {
if (Cs < 1578) {
return 4;
}
else if (Cs >= 1578) {
return 8;
}
}
else if (diffAFD >= -20) {
if (Rs < 2988) {
return 8;
}
else if (Rs >= 2988) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (Rs < 3084) {
return 2;
}
else if (Rs >= 3084) {
return 8;
}
}
}
}
}
else if (CsDiff >= 2) {
if (ssDCval < 117) {
if (ssDCval < 94) {
if (MVDiff < 2) {
if (Cs < 1360) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 1360) {
if (refDCval < 89) {
if (mu_mv_mag_sq < 124) {
if (Cs < 1496) {
return 8;
}
else if (Cs >= 1496) {
return 2;
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (refDCval >= 89) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (posBalance < 54) {
if (Rs < 1871) {
return 8;
}
else if (Rs >= 1871) {
return 2;
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (ssDCval >= 94) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (ssDCval >= 117) {
if (mu_mv_mag_sq < 120) {
if (diffMVdiffVal < 6) {
if (Rs < 2813) {
return 8;
}
else if (Rs >= 2813) {
return 2;
}
}
else if (diffMVdiffVal >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 120) {
if (SC < 3519) {
return 1;
}
else if (SC >= 3519) {
if (Rs < 2930) {
return 8;
}
else if (Rs >= 2930) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (MVDiff < 3) {
if (Rs < 2008) {
return 2;
}
else if (Rs >= 2008) {
return 1;
}
}
else if (MVDiff >= 3) {
if (Rs < 2768) {
return 8;
}
else if (Rs >= 2768) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3092) {
if (refDCval < 78) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 44) {
if (Rs < 3641) {
if (ssDCval < 63) {
if (Cs < 1344) {
return 8;
}
else if (Cs >= 1344) {
return 4;
}
}
else if (ssDCval >= 63) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
if (Rs < 3309) {
return 2;
}
else if (Rs >= 3309) {
return 4;
}
}
}
}
else if (Rs >= 3641) {
if (AFD < 4) {
if (negBalance < 21) {
if (Rs < 5520) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (Rs >= 5520) {
return 8;
}
}
else if (negBalance >= 21) {
if (MVDiff < 1) {
if (AFD < 3) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (AFD >= 4) {
if (Rs < 4375) {
return 4;
}
else if (Rs >= 4375) {
return 8;
}
}
}
}
else if (posBalance >= 44) {
if (mcTcor < 99) {
if (RsDiff < 5) {
if (Rs < 3880) {
return 4;
}
else if (Rs >= 3880) {
return 2;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (mcTcor >= 99) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1568) {
if (AFD < 3) {
if (Rs < 3678) {
if (Rs < 3649) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (Rs >= 3649) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 8;
}
}
}
else if (Rs >= 3678) {
if (RsCsDiff < 12) {
return 8;
}
else if (RsCsDiff >= 12) {
if (Rs < 4335) {
return 4;
}
else if (Rs >= 4335) {
return 8;
}
}
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (Cs >= 1568) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 4;
}
}
}
}
else if (refDCval >= 78) {
if (ssDCval < 113) {
if (posBalance < 42) {
if (Rs < 4121) {
if (Rs < 3207) {
return 8;
}
else if (Rs >= 3207) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 4;
}
}
}
else if (Rs >= 4121) {
if (ssDCval < 96) {
if (posBalance < 36) {
return 2;
}
else if (posBalance >= 36) {
return 8;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (posBalance >= 42) {
if (refDCval < 90) {
if (Cs < 1625) {
return 2;
}
else if (Cs >= 1625) {
if (Rs < 3381) {
return 4;
}
else if (Rs >= 3381) {
return 2;
}
}
}
else if (refDCval >= 90) {
if (Cs < 1530) {
if (Rs < 4483) {
if (Rs < 3263) {
return 2;
}
else if (Rs >= 3263) {
return 4;
}
}
else if (Rs >= 4483) {
return 2;
}
}
else if (Cs >= 1530) {
if (ssDCval < 101) {
return 8;
}
else if (ssDCval >= 101) {
if (mu_mv_mag_sq < 117) {
return 1;
}
else if (mu_mv_mag_sq >= 117) {
if (Rs < 3733) {
return 2;
}
else if (Rs >= 3733) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (mcTcor < 94) {
if (Rs < 3256) {
return 1;
}
else if (Rs >= 3256) {
return 4;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
}
}
}
else if (Cs >= 1644) {
if (refDCval < 102) {
if (posBalance < 3) {
if (refDCval < 65) {
if (Rs < 8850) {
return 8;
}
else if (Rs >= 8850) {
return 2;
}
}
else if (refDCval >= 65) {
if (diffTSC < -1) {
if (diffRsCsdiff < -1) {
if (ssDCval < 69) {
if (Rs < 3548) {
return 4;
}
else if (Rs >= 3548) {
return 8;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
return 4;
}
}
else if (diffTSC >= -1) {
if (ssDCval < 90) {
if (Rs < 3473) {
return 8;
}
else if (Rs >= 3473) {
return 4;
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (refDCval < 67) {
if (Cs < 3314) {
if (posBalance < 36) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 3272) {
if (SC < 3676) {
if (Cs < 1697) {
if (Rs < 863) {
return 1;
}
else if (Rs >= 863) {
return 4;
}
}
else if (Cs >= 1697) {
return 8;
}
}
else if (SC >= 3676) {
if (ssDCval < 62) {
if (Rs < 3257) {
return 1;
}
else if (Rs >= 3257) {
if (Rs < 3262) {
return 4;
}
else if (Rs >= 3262) {
return 1;
}
}
}
else if (ssDCval >= 62) {
if (Rs < 2325) {
return 8;
}
else if (Rs >= 2325) {
return 1;
}
}
}
}
else if (Rs >= 3272) {
return 4;
}
}
}
else if (posBalance >= 36) {
if (Rs < 1663) {
if (ssDCval < 57) {
if (Rs < 1049) {
if (Rs < 791) {
return 1;
}
else if (Rs >= 791) {
return 8;
}
}
else if (Rs >= 1049) {
return 4;
}
}
else if (ssDCval >= 57) {
if (mcTcor < 98) {
if (tcor < 98) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 1024) {
return 8;
}
else if (Rs >= 1024) {
return 1;
}
}
}
else if (tcor >= 98) {
return 1;
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 116) {
return 8;
}
else if (mu_mv_mag_sq >= 116) {
return 1;
}
}
}
}
else if (Rs >= 1663) {
if (negBalance < 20) {
if (Cs < 2083) {
if (ssDCval < 66) {
if (posBalance < 43) {
if (SC < 4013) {
return 8;
}
else if (SC >= 4013) {
if (Rs < 2615) {
return 1;
}
else if (Rs >= 2615) {
return 4;
}
}
}
else if (posBalance >= 43) {
return 8;
}
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (Cs >= 2083) {
return 4;
}
}
else if (negBalance >= 20) {
if (Cs < 1860) {
if (ssDCval < 65) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (Rs < 2939) {
return 4;
}
else if (Rs >= 2939) {
if (Rs < 3937) {
return 8;
}
else if (Rs >= 3937) {
return 8;
}
}
}
}
else if (ssDCval >= 65) {
if (SC < 4086) {
return 8;
}
else if (SC >= 4086) {
return 1;
}
}
}
else if (Cs >= 1860) {
return 8;
}
}
}
}
}
else if (Cs >= 3314) {
return 8;
}
}
else if (refDCval >= 67) {
if (mu_mv_mag_sq < 122) {
if (posBalance < 36) {
if (refDCval < 92) {
if (negBalance < 7) {
if (Rs < 2382) {
return 1;
}
else if (Rs >= 2382) {
return 8;
}
}
else if (negBalance >= 7) {
if (posBalance < 34) {
if (posBalance < 11) {
if (Rs < 2370) {
return 1;
}
else if (Rs >= 2370) {
return 2;
}
}
else if (posBalance >= 11) {
if (mu_mv_mag_sq < 112) {
if (Cs < 2176) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 2176) {
return 1;
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
}
else if (posBalance >= 34) {
return 2;
}
}
}
else if (refDCval >= 92) {
if (TSC < 1) {
if (negBalance < 8) {
return 2;
}
else if (negBalance >= 8) {
if (posBalance < 14) {
if (mu_mv_mag_sq < 117) {
return 2;
}
else if (mu_mv_mag_sq >= 117) {
return 4;
}
}
else if (posBalance >= 14) {
if (Rs < 2297) {
return 2;
}
else if (Rs >= 2297) {
return 4;
}
}
}
}
else if (TSC >= 1) {
if (Rs < 2319) {
if (Rs < 2308) {
return 1;
}
else if (Rs >= 2308) {
return 4;
}
}
else if (Rs >= 2319) {
return 8;
}
}
}
}
else if (posBalance >= 36) {
if (Cs < 2110) {
if (Rs < 3109) {
if (Cs < 2072) {
return 8;
}
else if (Cs >= 2072) {
return 2;
}
}
else if (Rs >= 3109) {
return 2;
}
}
else if (Cs >= 2110) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (SC < 6962) {
if (Rs < 1104) {
if (Rs < 619) {
return 8;
}
else if (Rs >= 619) {
return 1;
}
}
else if (Rs >= 1104) {
if (diffRsCsdiff < -210) {
if (Rs < 3479) {
return 8;
}
else if (Rs >= 3479) {
return 4;
}
}
else if (diffRsCsdiff >= -210) {
if (SC < 4544) {
if (SC < 4407) {
return 8;
}
else if (SC >= 4407) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
if (SC < 4456) {
return 4;
}
else if (SC >= 4456) {
if (Rs < 2093) {
return 4;
}
else if (Rs >= 2093) {
return 8;
}
}
}
}
}
else if (SC >= 4544) {
if (mu_mv_mag_sq < 125) {
if (Cs < 2137) {
if (Rs < 3019) {
return 8;
}
else if (Rs >= 3019) {
return 2;
}
}
else if (Cs >= 2137) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
}
}
}
else if (SC >= 6962) {
if (SC < 7530) {
if (SC < 7183) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (SC >= 7183) {
return 4;
}
}
else if (SC >= 7530) {
if (diffRsCsdiff < -5) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
if (Rs < 6655) {
return 8;
}
else if (Rs >= 6655) {
return 4;
}
}
}
else if (diffRsCsdiff >= -5) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (posBalance < 48) {
if (posBalance < 17) {
if (tcor < 71) {
return 4;
}
else if (tcor >= 71) {
return 8;
}
}
else if (posBalance >= 17) {
if (ssDCval < 72) {
if (posBalance < 41) {
if (tcor < 99) {
if (RsCsDiff < 87) {
if (MVDiff < 4) {
if (Rs < 2893) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (Rs >= 2893) {
if (negBalance < 26) {
if (ssDCval < 65) {
if (CsDiff < 4) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (MVDiff < 3) {
if (MVDiff < 2) {
if (RsDiff < 5) {
if (Rs < 3003) {
if (Rs < 2935) {
return 8;
}
else if (Rs >= 2935) {
return 8;
}
}
else if (Rs >= 3003) {
if (posBalance < 39) {
if (refDCval < 65) {
if (negBalance < 24) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Cs < 1856) {
return 4;
}
else if (Cs >= 1856) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (Rs < 3603) {
return 8;
}
else if (Rs >= 3603) {
if (posBalance < 34) {
return 4;
}
else if (posBalance >= 34) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 24) {
return 8;
}
}
else if (refDCval >= 65) {
return 4;
}
}
else if (posBalance >= 39) {
if (Rs < 3070) {
return 4;
}
else if (Rs >= 3070) {
return 4;
}
}
}
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Rs < 4776) {
return 8;
}
else if (Rs >= 4776) {
return 4;
}
}
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (CsDiff >= 4) {
if (negBalance < 24) {
if (diffRsCsdiff < -1) {
if (Cs < 2485) {
if (diffTSC < -1) {
if (diffAFD < -1) {
if (posBalance < 28) {
return 4;
}
else if (posBalance >= 28) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (diffTSC >= -1) {
if (negBalance < 20) {
return 4;
}
else if (negBalance >= 20) {
if (Rs < 4429) {
return 4;
}
else if (Rs >= 4429) {
if (Rs < 4707) {
return 4;
}
else if (Rs >= 4707) {
return 4;
}
}
}
}
}
else if (Cs >= 2485) {
if (Cs < 2700) {
if (Rs < 4818) {
return 4;
}
else if (Rs >= 4818) {
if (Rs < 4990) {
return 4;
}
else if (Rs >= 4990) {
return 4;
}
}
}
else if (Cs >= 2700) {
return 4;
}
}
}
else if (diffRsCsdiff >= -1) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
if (negBalance < 14) {
return 4;
}
else if (negBalance >= 14) {
if (Cs < 2024) {
return 4;
}
else if (Cs >= 2024) {
if (diffAFD < 1) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 4716) {
return 4;
}
else if (Rs >= 4716) {
return 4;
}
}
}
else if (diffAFD >= 1) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 24) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 65) {
if (Rs < 2948) {
return 4;
}
else if (Rs >= 2948) {
return 4;
}
}
}
else if (negBalance >= 26) {
if (posBalance < 37) {
if (Cs < 1790) {
return 4;
}
else if (Cs >= 1790) {
if (Cs < 2774) {
if (Rs < 4520) {
if (Cs < 1913) {
return 4;
}
else if (Cs >= 1913) {
return 4;
}
}
else if (Rs >= 4520) {
return 4;
}
}
else if (Cs >= 2774) {
return 8;
}
}
}
else if (posBalance >= 37) {
return 4;
}
}
}
}
else if (MVDiff >= 4) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
return 8;
}
}
}
else if (RsCsDiff >= 87) {
return 8;
}
}
else if (tcor >= 99) {
if (Rs < 7273) {
if (Cs < 1670) {
return 4;
}
else if (Cs >= 1670) {
return 8;
}
}
else if (Rs >= 7273) {
return 2;
}
}
}
else if (posBalance >= 41) {
if (RsDiff < 5) {
if (refDCval < 71) {
if (negBalance < 17) {
if (negBalance < 14) {
return 4;
}
else if (negBalance >= 14) {
return 4;
}
}
else if (negBalance >= 17) {
if (posBalance < 42) {
if (tcor < 98) {
if (Rs < 3107) {
if (Rs < 2972) {
return 8;
}
else if (Rs >= 2972) {
return 4;
}
}
else if (Rs >= 3107) {
return 4;
}
}
else if (tcor >= 98) {
if (RsCsDiff < 7) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (RsCsDiff >= 7) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (MVDiff < 6) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
if (Rs < 2669) {
return 4;
}
else if (Rs >= 2669) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (Rs < 1018) {
return 8;
}
else if (Rs >= 1018) {
return 4;
}
}
}
}
}
else if (refDCval >= 71) {
if (Rs < 3856) {
if (Rs < 2463) {
return 4;
}
else if (Rs >= 2463) {
return 8;
}
}
else if (Rs >= 3856) {
return 1;
}
}
}
else if (RsDiff >= 5) {
if (RsCsDiff < 80) {
if (TSC < 5) {
if (refDCval < 55) {
if (negBalance < 12) {
return 4;
}
else if (negBalance >= 12) {
if (Rs < 4731) {
if (Rs < 3686) {
return 4;
}
else if (Rs >= 3686) {
return 4;
}
}
else if (Rs >= 4731) {
if (Rs < 4876) {
return 8;
}
else if (Rs >= 4876) {
return 8;
}
}
}
}
else if (refDCval >= 55) {
return 8;
}
}
else if (TSC >= 5) {
return 4;
}
}
else if (RsCsDiff >= 80) {
if (negBalance < 32) {
if (Cs < 4431) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
else if (Cs >= 4431) {
return 1;
}
}
else if (negBalance >= 32) {
return 1;
}
}
}
}
}
else if (ssDCval >= 72) {
if (refDCval < 94) {
if (SC < 6868) {
if (Rs < 3085) {
if (Cs < 1908) {
if (MVDiff < 1) {
if (Rs < 2554) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
else if (Rs >= 2554) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (Cs < 1894) {
return 4;
}
else if (Cs >= 1894) {
if (Rs < 2670) {
return 4;
}
else if (Rs >= 2670) {
return 8;
}
}
}
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (Cs >= 1908) {
if (AFD < 7) {
if (Cs < 1940) {
if (Cs < 1932) {
return 8;
}
else if (Cs >= 1932) {
if (Rs < 2809) {
return 4;
}
else if (Rs >= 2809) {
return 8;
}
}
}
else if (Cs >= 1940) {
return 8;
}
}
else if (AFD >= 7) {
if (Rs < 2619) {
return 4;
}
else if (Rs >= 2619) {
return 8;
}
}
}
}
else if (Rs >= 3085) {
if (Cs < 2265) {
if (RsCsDiff < 23) {
if (ssDCval < 86) {
if (Cs < 1917) {
if (Rs < 4509) {
return 4;
}
else if (Rs >= 4509) {
return 8;
}
}
else if (Cs >= 1917) {
return 8;
}
}
else if (ssDCval >= 86) {
return 2;
}
}
else if (RsCsDiff >= 23) {
if (ssDCval < 91) {
if (Cs < 2042) {
if (Rs < 4647) {
if (Rs < 3623) {
if (Rs < 3215) {
return 2;
}
else if (Rs >= 3215) {
if (Cs < 1648) {
return 2;
}
else if (Cs >= 1648) {
return 4;
}
}
}
else if (Rs >= 3623) {
return 2;
}
}
else if (Rs >= 4647) {
return 8;
}
}
else if (Cs >= 2042) {
return 8;
}
}
else if (ssDCval >= 91) {
if (diffRsCsdiff < -16) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
return 2;
}
}
else if (diffRsCsdiff >= -16) {
if (refDCval < 93) {
return 8;
}
else if (refDCval >= 93) {
return 2;
}
}
}
}
}
else if (Cs >= 2265) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
}
else if (SC >= 6868) {
if (SC < 9357) {
if (SC < 8129) {
if (Cs < 2004) {
return 8;
}
else if (Cs >= 2004) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (SC >= 8129) {
if (Rs < 5774) {
return 8;
}
else if (Rs >= 5774) {
return 4;
}
}
}
else if (SC >= 9357) {
if (mu_mv_mag_sq < 119) {
if (Cs < 3969) {
return 8;
}
else if (Cs >= 3969) {
return 2;
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
}
}
else if (refDCval >= 94) {
if (Cs < 1917) {
if (Rs < 3423) {
return 8;
}
else if (Rs >= 3423) {
return 4;
}
}
else if (Cs >= 1917) {
if (negBalance < 57) {
if (Cs < 2023) {
if (Rs < 3554) {
return 8;
}
else if (Rs >= 3554) {
return 2;
}
}
else if (Cs >= 2023) {
if (RsCsDiff < 7) {
if (Rs < 3875) {
return 8;
}
else if (Rs >= 3875) {
return 4;
}
}
else if (RsCsDiff >= 7) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
if (Rs < 5241) {
return 4;
}
else if (Rs >= 5241) {
return 8;
}
}
}
}
}
else if (negBalance >= 57) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 90) {
if (Rs < 6123) {
if (SC < 7443) {
if (mu_mv_mag_sq < 117) {
if (Rs < 3691) {
if (SC < 4345) {
if (refDCval < 66) {
if (diffRsCsdiff < -10) {
return 4;
}
else if (diffRsCsdiff >= -10) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 1151) {
return 8;
}
else if (Rs >= 1151) {
return 4;
}
}
}
}
else if (refDCval >= 66) {
if (Rs < 887) {
return 1;
}
else if (Rs >= 887) {
if (gchDC < 2) {
if (Cs < 2221) {
return 8;
}
else if (Cs >= 2221) {
if (Cs < 2228) {
return 1;
}
else if (Cs >= 2228) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (SC >= 4345) {
if (Rs < 2124) {
return 4;
}
else if (Rs >= 2124) {
if (Cs < 2073) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 3201) {
if (Rs < 2569) {
return 8;
}
else if (Rs >= 2569) {
if (mcTcor < 95) {
if (Rs < 2727) {
return 4;
}
else if (Rs >= 2727) {
return 8;
}
}
else if (mcTcor >= 95) {
return 4;
}
}
}
else if (Rs >= 3201) {
if (Rs < 3461) {
return 2;
}
else if (Rs >= 3461) {
return 8;
}
}
}
}
else if (Cs >= 2073) {
if (posBalance < 62) {
if (Cs < 2637) {
if (Rs < 3277) {
return 8;
}
else if (Rs >= 3277) {
if (Rs < 3288) {
return 4;
}
else if (Rs >= 3288) {
return 8;
}
}
}
else if (Cs >= 2637) {
if (Rs < 2786) {
return 4;
}
else if (Rs >= 2786) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (Rs < 3274) {
return 8;
}
else if (Rs >= 3274) {
return 4;
}
}
}
}
}
else if (posBalance >= 62) {
if (Rs < 2186) {
return 8;
}
else if (Rs >= 2186) {
return 4;
}
}
}
}
}
}
else if (Rs >= 3691) {
if (Cs < 2276) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (Cs >= 2276) {
if (Cs < 3258) {
if (ssDCval < 82) {
if (SC < 6613) {
return 4;
}
else if (SC >= 6613) {
if (SC < 7408) {
return 8;
}
else if (SC >= 7408) {
return 4;
}
}
}
else if (ssDCval >= 82) {
return 4;
}
}
else if (Cs >= 3258) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (refDCval < 81) {
if (Cs < 3523) {
if (diffMVdiffVal < -4) {
if (tcor < 95) {
if (Cs < 2992) {
return 8;
}
else if (Cs >= 2992) {
return 4;
}
}
else if (tcor >= 95) {
return 4;
}
}
else if (diffMVdiffVal >= -4) {
if (negBalance < 36) {
if (MVDiff < 3) {
if (SC < 5169) {
if (Cs < 2307) {
if (Rs < 2635) {
return 8;
}
else if (Rs >= 2635) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (Cs >= 2307) {
return 4;
}
}
else if (SC >= 5169) {
return 8;
}
}
else if (MVDiff >= 3) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (negBalance >= 36) {
return 8;
}
}
}
else if (Cs >= 3523) {
if (Rs < 3347) {
return 4;
}
else if (Rs >= 3347) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (Rs < 3259) {
if (tcor < 98) {
if (posBalance < 73) {
if (TSC < 4) {
if (Cs < 2429) {
return 8;
}
else if (Cs >= 2429) {
return 4;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (posBalance >= 73) {
return 4;
}
}
else if (tcor >= 98) {
if (Cs < 1879) {
if (SC < 4650) {
return 4;
}
else if (SC >= 4650) {
return 8;
}
}
else if (Cs >= 1879) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
return 8;
}
}
}
}
else if (Rs >= 3259) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
if (SC < 6403) {
if (SC < 6199) {
return 4;
}
else if (SC >= 6199) {
return 8;
}
}
else if (SC >= 6403) {
return 4;
}
}
}
}
}
}
else if (SC >= 7443) {
if (negBalance < 24) {
return 1;
}
else if (negBalance >= 24) {
if (negBalance < 29) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
}
else if (Rs >= 6123) {
return 4;
}
}
else if (ssDCval >= 90) {
if (Cs < 2515) {
if (Rs < 3622) {
if (Rs < 1885) {
if (Cs < 2131) {
return 8;
}
else if (Cs >= 2131) {
if (Rs < 1707) {
return 2;
}
else if (Rs >= 1707) {
return 4;
}
}
}
else if (Rs >= 1885) {
return 8;
}
}
else if (Rs >= 3622) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (Cs >= 2515) {
if (SC < 8144) {
if (negBalance < 33) {
if (Rs < 4127) {
if (Cs < 3365) {
return 8;
}
else if (Cs >= 3365) {
return 4;
}
}
else if (Rs >= 4127) {
return 4;
}
}
else if (negBalance >= 33) {
if (Cs < 3290) {
if (Cs < 2736) {
if (negBalance < 45) {
if (Cs < 2647) {
return 8;
}
else if (Cs >= 2647) {
if (negBalance < 41) {
return 8;
}
else if (negBalance >= 41) {
if (Rs < 3274) {
return 8;
}
else if (Rs >= 3274) {
return 4;
}
}
}
}
else if (negBalance >= 45) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Rs < 3783) {
return 8;
}
else if (Rs >= 3783) {
return 1;
}
}
}
}
else if (Cs >= 2736) {
if (ssDCval < 98) {
if (Cs < 3238) {
return 4;
}
else if (Cs >= 3238) {
if (Rs < 3360) {
return 8;
}
else if (Rs >= 3360) {
return 4;
}
}
}
else if (ssDCval >= 98) {
if (mu_mv_mag_sq < 122) {
if (Rs < 3806) {
return 4;
}
else if (Rs >= 3806) {
return 1;
}
}
else if (mu_mv_mag_sq >= 122) {
return 8;
}
}
}
}
else if (Cs >= 3290) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (negBalance < 53) {
if (negBalance < 45) {
if (CsDiff < 8) {
if (Cs < 3532) {
return 4;
}
else if (Cs >= 3532) {
return 8;
}
}
else if (CsDiff >= 8) {
if (Rs < 1507) {
return 2;
}
else if (Rs >= 1507) {
return 4;
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (negBalance >= 53) {
return 4;
}
}
}
}
}
else if (SC >= 8144) {
if (Rs < 6944) {
if (RsCsDiff < 79) {
if (SC < 8642) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (SC >= 8642) {
return 8;
}
}
else if (RsCsDiff >= 79) {
return 8;
}
}
else if (Rs >= 6944) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 102) {
if (mcTcor < 96) {
if (ssDCval < 137) {
if (refDCval < 105) {
if (negBalance < 42) {
if (Cs < 1836) {
return 4;
}
else if (Cs >= 1836) {
return 8;
}
}
else if (negBalance >= 42) {
if (SC < 7207) {
if (Rs < 3734) {
return 8;
}
else if (Rs >= 3734) {
return 1;
}
}
else if (SC >= 7207) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
if (Cs < 2893) {
return 1;
}
else if (Cs >= 2893) {
return 8;
}
}
}
}
}
else if (refDCval >= 105) {
if (MVDiff < 4) {
if (diffAFD < 4) {
if (posBalance < 63) {
if (diffAFD < -3) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
}
else if (diffAFD >= -3) {
if (diffMVdiffVal < -1) {
if (negBalance < 54) {
if (refDCval < 123) {
return 8;
}
else if (refDCval >= 123) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 1;
}
}
}
else if (negBalance >= 54) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Rs < 2923) {
return 8;
}
else if (Rs >= 2923) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (diffMVdiffVal < 2) {
if (tcor < 85) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (tcor >= 85) {
return 8;
}
}
else if (diffMVdiffVal >= 2) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
return 8;
}
}
}
}
}
else if (posBalance >= 63) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (Cs < 1933) {
return 1;
}
else if (Cs >= 1933) {
return 8;
}
}
}
}
else if (diffAFD >= 4) {
if (Cs < 2051) {
return 4;
}
else if (Cs >= 2051) {
if (Rs < 3174) {
return 1;
}
else if (Rs >= 3174) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (gchDC < 1) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
if (Rs < 3210) {
return 8;
}
else if (Rs >= 3210) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (Rs < 3627) {
if (negBalance < 51) {
if (diffRsCsdiff < -44) {
return 2;
}
else if (diffRsCsdiff >= -44) {
return 8;
}
}
else if (negBalance >= 51) {
if (Rs < 3192) {
if (Rs < 2055) {
return 4;
}
else if (Rs >= 2055) {
return 1;
}
}
else if (Rs >= 3192) {
return 2;
}
}
}
else if (Rs >= 3627) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
if (negBalance < 39) {
if (ssDCval < 200) {
if (diffAFD < 6) {
return 8;
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (ssDCval >= 200) {
return 4;
}
}
else if (negBalance >= 39) {
if (refDCval < 142) {
return 1;
}
else if (refDCval >= 142) {
if (negBalance < 46) {
if (Rs < 3583) {
return 8;
}
else if (Rs >= 3583) {
return 1;
}
}
else if (negBalance >= 46) {
if (Rs < 8232) {
return 8;
}
else if (Rs >= 8232) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1909) {
if (ssDCval < 115) {
if (Cs < 1951) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (AFD < 5) {
if (Rs < 1640) {
return 2;
}
else if (Rs >= 1640) {
return 8;
}
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (Cs >= 1951) {
if (diffAFD < -1) {
if (Rs < 1653) {
return 2;
}
else if (Rs >= 1653) {
return 8;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
}
else if (ssDCval >= 115) {
if (Rs < 1259) {
if (Rs < 1176) {
if (Rs < 923) {
return 1;
}
else if (Rs >= 923) {
return 8;
}
}
else if (Rs >= 1176) {
if (Cs < 1798) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (Cs >= 1798) {
return 8;
}
}
}
else if (Rs >= 1259) {
return 8;
}
}
}
else if (Rs >= 1909) {
if (refDCval < 117) {
if (Rs < 4963) {
if (posBalance < 10) {
if (mu_mv_mag_sq < 112) {
if (Rs < 2096) {
return 2;
}
else if (Rs >= 2096) {
if (Rs < 2623) {
return 1;
}
else if (Rs >= 2623) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
else if (posBalance >= 10) {
if (diffTSC < -2) {
if (posBalance < 57) {
if (Cs < 2125) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 2346) {
return 8;
}
else if (Rs >= 2346) {
return 4;
}
}
}
else if (Cs >= 2125) {
return 8;
}
}
else if (posBalance >= 57) {
if (Rs < 2899) {
return 8;
}
else if (Rs >= 2899) {
return 2;
}
}
}
else if (diffTSC >= -2) {
if (diffRsCsdiff < 14) {
return 8;
}
else if (diffRsCsdiff >= 14) {
if (RsCsDiff < 37) {
if (mcTcor < 97) {
return 4;
}
else if (mcTcor >= 97) {
return 8;
}
}
else if (RsCsDiff >= 37) {
return 8;
}
}
}
}
}
else if (Rs >= 4963) {
if (Cs < 2403) {
return 2;
}
else if (Cs >= 2403) {
return 8;
}
}
}
else if (refDCval >= 117) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 35) {
if (ssDCval < 146) {
if (Rs < 2774) {
if (Cs < 2936) {
if (Rs < 2562) {
return 8;
}
else if (Rs >= 2562) {
if (Cs < 2899) {
return 4;
}
else if (Cs >= 2899) {
if (posBalance < 13) {
return 8;
}
else if (posBalance >= 13) {
if (negBalance < 8) {
return 8;
}
else if (negBalance >= 8) {
return 4;
}
}
}
}
}
else if (Cs >= 2936) {
return 8;
}
}
else if (Rs >= 2774) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
if (Cs < 2232) {
return 2;
}
else if (Cs >= 2232) {
return 8;
}
}
}
}
else if (ssDCval >= 146) {
return 8;
}
}
else if (posBalance >= 35) {
if (Rs < 5094) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 2744) {
return 8;
}
else if (Rs >= 2744) {
return 2;
}
}
}
else if (Rs >= 5094) {
if (Cs < 3886) {
if (Cs < 2991) {
return 2;
}
else if (Cs >= 2991) {
return 4;
}
}
else if (Cs >= 3886) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 4837) {
if (Cs < 3891) {
if (AFD < 6) {
if (diffTSC < -6) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
if (Rs < 2386) {
return 8;
}
else if (Rs >= 2386) {
return 1;
}
}
}
else if (diffTSC >= -6) {
if (negBalance < 39) {
return 8;
}
else if (negBalance >= 39) {
if (Rs < 3403) {
return 8;
}
else if (Rs >= 3403) {
return 1;
}
}
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 3891) {
if (Rs < 4086) {
if (ssDCval < 168) {
return 1;
}
else if (ssDCval >= 168) {
return 4;
}
}
else if (Rs >= 4086) {
return 8;
}
}
}
else if (Rs >= 4837) {
if (Cs < 1975) {
return 2;
}
else if (Cs >= 1975) {
return 8;
}
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect3(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1313) {
if (Rs < 2599) {
if (Cs < 534) {
if (negBalance < 14) {
if (mu_mv_mag_sq < 115) {
if (Cs < 179) {
if (posBalance < 48) {
if (mcTcor < 99) {
if (ssDCval < 74) {
if (refDCval < 27) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
if (negBalance < 7) {
return 4;
}
else if (negBalance >= 7) {
if (tcor < 95) {
return 1;
}
else if (tcor >= 95) {
if (Cs < 82) {
if (posBalance < 15) {
return 4;
}
else if (posBalance >= 15) {
return 1;
}
}
else if (Cs >= 82) {
return 4;
}
}
}
}
}
else if (refDCval >= 27) {
if (negBalance < 9) {
if (Rs < 199) {
if (Rs < 10) {
return 2;
}
else if (Rs >= 10) {
return 8;
}
}
else if (Rs >= 199) {
return 4;
}
}
else if (negBalance >= 9) {
if (ssDCval < 31) {
if (posBalance < 38) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
return 4;
}
}
else if (posBalance >= 38) {
return 1;
}
}
else if (ssDCval >= 31) {
if (SC < 238) {
if (Rs < 47) {
return 4;
}
else if (Rs >= 47) {
return 8;
}
}
else if (SC >= 238) {
return 4;
}
}
}
}
}
else if (ssDCval >= 74) {
if (ssDCval < 91) {
return 1;
}
else if (ssDCval >= 91) {
return 4;
}
}
}
else if (mcTcor >= 99) {
if (MVDiff < 9) {
if (SC < 351) {
if (ssDCval < 48) {
if (Cs < 40) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 1;
}
}
}
else if (Cs >= 40) {
if (TSCindex < 1) {
return 8;
}
else if (TSCindex >= 1) {
if (ssDCval < 32) {
if (Rs < 75) {
return 4;
}
else if (Rs >= 75) {
return 1;
}
}
else if (ssDCval >= 32) {
return 8;
}
}
}
}
else if (ssDCval >= 48) {
if (negBalance < 10) {
if (Rs < 32) {
return 2;
}
else if (Rs >= 32) {
return 1;
}
}
else if (negBalance >= 10) {
if (Rs < 152) {
return 4;
}
else if (Rs >= 152) {
return 2;
}
}
}
}
else if (SC >= 351) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
}
else if (MVDiff >= 9) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
if (diffAFD < -1) {
if (ssDCval < 36) {
return 1;
}
else if (ssDCval >= 36) {
if (Rs < 29) {
return 2;
}
else if (Rs >= 29) {
return 4;
}
}
}
else if (diffAFD >= -1) {
if (MVDiff < 10) {
if (Rs < 99) {
if (Rs < 13) {
return 4;
}
else if (Rs >= 13) {
return 1;
}
}
else if (Rs >= 99) {
return 4;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 48) {
if (MVDiff < 12) {
if (Rs < 13) {
if (ssDCval < 42) {
if (ssDCval < 24) {
return 8;
}
else if (ssDCval >= 24) {
return 4;
}
}
else if (ssDCval >= 42) {
return 2;
}
}
else if (Rs >= 13) {
if (posBalance < 50) {
if (MVDiff < 5) {
if (Cs < 157) {
return 8;
}
else if (Cs >= 157) {
return 2;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (posBalance >= 50) {
if (Rs < 1400) {
if (negBalance < 1) {
if (Cs < 28) {
return 2;
}
else if (Cs >= 28) {
return 8;
}
}
else if (negBalance >= 1) {
return 8;
}
}
else if (Rs >= 1400) {
if (Rs < 1580) {
return 1;
}
else if (Rs >= 1580) {
return 8;
}
}
}
}
}
else if (MVDiff >= 12) {
if (negBalance < 7) {
if (TSC < 7) {
if (Rs < 115) {
if (refDCval < 50) {
if (ssDCval < 17) {
return 4;
}
else if (ssDCval >= 17) {
if (mu_mv_mag_sq < 47) {
if (negBalance < 1) {
if (Rs < 52) {
return 8;
}
else if (Rs >= 52) {
return 4;
}
}
else if (negBalance >= 1) {
if (mcTcor < 75) {
if (Rs < 15) {
return 8;
}
else if (Rs >= 15) {
return 2;
}
}
else if (mcTcor >= 75) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (mu_mv_mag_sq < 103) {
return 8;
}
else if (mu_mv_mag_sq >= 103) {
return 4;
}
}
}
}
else if (refDCval >= 50) {
if (Rs < 62) {
return 2;
}
else if (Rs >= 62) {
return 4;
}
}
}
else if (Rs >= 115) {
if (AFD < 5) {
if (MVDiff < 38) {
if (ssDCval < 34) {
if (Cs < 58) {
return 4;
}
else if (Cs >= 58) {
return 1;
}
}
else if (ssDCval >= 34) {
return 8;
}
}
else if (MVDiff >= 38) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
if (mu_mv_mag_sq < 58) {
return 4;
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 215) {
return 8;
}
else if (Rs >= 215) {
return 1;
}
}
}
}
}
else if (AFD >= 5) {
if (Cs < 153) {
return 8;
}
else if (Cs >= 153) {
return 1;
}
}
}
}
else if (TSC >= 7) {
if (gchDC < 12) {
if (Rs < 497) {
if (Rs < 267) {
return 8;
}
else if (Rs >= 267) {
return 1;
}
}
else if (Rs >= 497) {
return 2;
}
}
else if (gchDC >= 12) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 2;
}
}
}
}
else if (negBalance >= 7) {
if (negBalance < 10) {
if (AFD < 6) {
if (ssDCval < 34) {
if (diffMVdiffVal < -45) {
return 2;
}
else if (diffMVdiffVal >= -45) {
return 1;
}
}
else if (ssDCval >= 34) {
if (Rs < 29) {
return 1;
}
else if (Rs >= 29) {
return 4;
}
}
}
else if (AFD >= 6) {
if (mu_mv_mag_sq < 7) {
return 2;
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 99) {
return 1;
}
else if (Cs >= 99) {
return 2;
}
}
}
}
else if (negBalance >= 10) {
if (Cs < 106) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
if (Cs < 28) {
return 2;
}
else if (Cs >= 28) {
if (Rs < 360) {
return 8;
}
else if (Rs >= 360) {
return 1;
}
}
}
}
else if (Cs >= 106) {
if (Rs < 142) {
return 8;
}
else if (Rs >= 142) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 179) {
if (SC < 861) {
if (tcor < 93) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
if (mu_mv_mag_sq < 41) {
if (posBalance < 103) {
if (MVDiff < 194) {
return 2;
}
else if (MVDiff >= 194) {
return 1;
}
}
else if (posBalance >= 103) {
if (refDCval < 53) {
if (Rs < 344) {
return 2;
}
else if (Rs >= 344) {
return 4;
}
}
else if (refDCval >= 53) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (mcTcor < 94) {
if (negBalance < 9) {
if (refDCval < 39) {
return 4;
}
else if (refDCval >= 39) {
if (Cs < 284) {
return 8;
}
else if (Cs >= 284) {
return 1;
}
}
}
else if (negBalance >= 9) {
if (Cs < 416) {
if (Rs < 150) {
return 8;
}
else if (Rs >= 150) {
if (diffTSC < -1) {
if (Rs < 280) {
return 8;
}
else if (Rs >= 280) {
return 1;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
}
else if (Cs >= 416) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (SC < 451) {
if (Rs < 180) {
return 8;
}
else if (Rs >= 180) {
return 1;
}
}
else if (SC >= 451) {
if (ssDCval < 40) {
return 4;
}
else if (ssDCval >= 40) {
return 1;
}
}
}
}
}
}
else if (tcor >= 93) {
if (mu_mv_mag_sq < 97) {
if (ssDCval < 114) {
if (AFD < 2) {
if (posBalance < 30) {
if (Rs < 134) {
return 1;
}
else if (Rs >= 134) {
return 4;
}
}
else if (posBalance >= 30) {
return 4;
}
}
else if (AFD >= 2) {
if (Cs < 253) {
return 4;
}
else if (Cs >= 253) {
if (Rs < 451) {
return 1;
}
else if (Rs >= 451) {
return 8;
}
}
}
}
else if (ssDCval >= 114) {
if (MVDiff < 25) {
if (ssDCval < 161) {
if (Rs < 249) {
return 4;
}
else if (Rs >= 249) {
return 1;
}
}
else if (ssDCval >= 161) {
return 4;
}
}
else if (MVDiff >= 25) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (ssDCval < 84) {
if (refDCval < 44) {
if (ssDCval < 37) {
if (Rs < 126) {
return 8;
}
else if (Rs >= 126) {
if (posBalance < 34) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (posBalance >= 34) {
return 1;
}
}
}
else if (ssDCval >= 37) {
if (Rs < 228) {
return 1;
}
else if (Rs >= 228) {
if (ssDCval < 41) {
return 4;
}
else if (ssDCval >= 41) {
return 1;
}
}
}
}
else if (refDCval >= 44) {
if (posBalance < 40) {
if (Rs < 136) {
return 2;
}
else if (Rs >= 136) {
return 8;
}
}
else if (posBalance >= 40) {
if (Rs < 253) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (Rs >= 253) {
return 1;
}
}
}
}
else if (ssDCval >= 84) {
if (posBalance < 27) {
if (Rs < 501) {
return 1;
}
else if (Rs >= 501) {
return 8;
}
}
else if (posBalance >= 27) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
}
}
}
else if (SC >= 861) {
if (refDCval < 34) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
return 4;
}
}
else if (refDCval >= 34) {
if (Cs < 396) {
if (Cs < 342) {
if (Rs < 1566) {
if (posBalance < 121) {
if (AFD < 11) {
if (MVDiff < 1) {
if (Rs < 743) {
return 8;
}
else if (Rs >= 743) {
return 1;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (posBalance >= 121) {
return 4;
}
}
else if (Rs >= 1566) {
if (diffMVdiffVal < 1) {
if (Rs < 1797) {
return 1;
}
else if (Rs >= 1797) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
else if (Cs >= 342) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
return 1;
}
}
}
else if (Cs >= 396) {
if (RsCsDiff < 1) {
return 8;
}
else if (RsCsDiff >= 1) {
if (refDCval < 49) {
if (SC < 1199) {
return 1;
}
else if (SC >= 1199) {
return 2;
}
}
else if (refDCval >= 49) {
if (tcor < 99) {
if (Rs < 1857) {
if (Rs < 876) {
return 8;
}
else if (Rs >= 876) {
if (Rs < 1097) {
return 1;
}
else if (Rs >= 1097) {
if (Cs < 416) {
return 4;
}
else if (Cs >= 416) {
return 8;
}
}
}
}
else if (Rs >= 1857) {
if (Rs < 2215) {
return 1;
}
else if (Rs >= 2215) {
return 2;
}
}
}
else if (tcor >= 99) {
return 1;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (SC < 792) {
if (refDCval < 47) {
if (Cs < 98) {
if (diffRsCsdiff < -1) {
return 8;
}
else if (diffRsCsdiff >= -1) {
if (ssDCval < 39) {
if (ssDCval < 22) {
if (MVDiff < 821) {
if (ssDCval < 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (MVDiff >= 821) {
return 4;
}
}
else if (ssDCval >= 22) {
if (Cs < 56) {
if (posBalance < 61) {
if (Rs < 31) {
return 4;
}
else if (Rs >= 31) {
if (Cs < 37) {
return 8;
}
else if (Cs >= 37) {
return 4;
}
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (Cs >= 56) {
return 8;
}
}
}
else if (ssDCval >= 39) {
return 8;
}
}
}
else if (Cs >= 98) {
if (Rs < 202) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
if (Cs < 232) {
if (ssDCval < 38) {
if (Rs < 140) {
return 8;
}
else if (Rs >= 140) {
return 4;
}
}
else if (ssDCval >= 38) {
if (Rs < 80) {
return 8;
}
else if (Rs >= 80) {
return 1;
}
}
}
else if (Cs >= 232) {
return 1;
}
}
}
else if (Rs >= 202) {
if (posBalance < 12) {
if (tcor < 71) {
if (Rs < 437) {
return 4;
}
else if (Rs >= 437) {
return 8;
}
}
else if (tcor >= 71) {
if (ssDCval < 26) {
return 8;
}
else if (ssDCval >= 26) {
if (Cs < 266) {
return 8;
}
else if (Cs >= 266) {
return 4;
}
}
}
}
else if (posBalance >= 12) {
if (TSCindex < 1) {
if (Rs < 416) {
return 4;
}
else if (Rs >= 416) {
return 8;
}
}
else if (TSCindex >= 1) {
if (mu_mv_mag_sq < 127) {
if (Cs < 446) {
if (posBalance < 25) {
if (Cs < 136) {
return 4;
}
else if (Cs >= 136) {
if (Rs < 282) {
return 8;
}
else if (Rs >= 282) {
return 1;
}
}
}
else if (posBalance >= 25) {
return 4;
}
}
else if (Cs >= 446) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 47) {
if (ssDCval < 84) {
if (Rs < 23) {
return 2;
}
else if (Rs >= 23) {
if (ssDCval < 66) {
if (Cs < 279) {
if (MVDiff < 1) {
if (posBalance < 28) {
if (Rs < 213) {
return 8;
}
else if (Rs >= 213) {
return 4;
}
}
else if (posBalance >= 28) {
return 8;
}
}
else if (MVDiff >= 1) {
if (gchDC < 2) {
if (MVDiff < 3) {
if (Cs < 166) {
return 4;
}
else if (Cs >= 166) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (Cs >= 279) {
if (Rs < 252) {
return 4;
}
else if (Rs >= 252) {
if (Rs < 318) {
return 1;
}
else if (Rs >= 318) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
if (Rs < 81) {
return 4;
}
else if (Rs >= 81) {
if (Cs < 237) {
return 2;
}
else if (Cs >= 237) {
if (Cs < 265) {
return 4;
}
else if (Cs >= 265) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (ssDCval < 95) {
if (Rs < 404) {
return 2;
}
else if (Rs >= 404) {
return 1;
}
}
else if (ssDCval >= 95) {
if (Cs < 353) {
if (Cs < 310) {
if (diffMVdiffVal < -8) {
if (Rs < 306) {
return 8;
}
else if (Rs >= 306) {
if (Rs < 338) {
return 4;
}
else if (Rs >= 338) {
return 2;
}
}
}
else if (diffMVdiffVal >= -8) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (Cs >= 310) {
return 2;
}
}
else if (Cs >= 353) {
return 4;
}
}
}
}
}
}
else if (SC >= 792) {
if (diffTSC < -3) {
if (posBalance < 3) {
return 4;
}
else if (posBalance >= 3) {
if (diffAFD < -8) {
if (ssDCval < 147) {
if (MVDiff < 13) {
return 1;
}
else if (MVDiff >= 13) {
return 2;
}
}
else if (ssDCval >= 147) {
return 8;
}
}
else if (diffAFD >= -8) {
if (Rs < 827) {
if (negBalance < 4) {
return 4;
}
else if (negBalance >= 4) {
if (Rs < 576) {
return 2;
}
else if (Rs >= 576) {
return 8;
}
}
}
else if (Rs >= 827) {
return 1;
}
}
}
}
else if (diffTSC >= -3) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Cs < 498) {
if (Rs < 1314) {
if (ssDCval < 42) {
if (negBalance < 6) {
return 8;
}
else if (negBalance >= 6) {
if (Rs < 503) {
return 1;
}
else if (Rs >= 503) {
return 4;
}
}
}
else if (ssDCval >= 42) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 855) {
return 8;
}
else if (Rs >= 855) {
return 1;
}
}
}
}
else if (Rs >= 1314) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (Cs >= 498) {
return 1;
}
}
else if (ssDCval >= 60) {
if (Cs < 457) {
if (Rs < 473) {
return 8;
}
else if (Rs >= 473) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Rs < 1444) {
return 2;
}
else if (Rs >= 1444) {
return 1;
}
}
}
}
else if (Cs >= 457) {
return 2;
}
}
}
else if (ssDCval >= 67) {
if (refDCval < 141) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
if (refDCval < 78) {
if (Rs < 641) {
return 4;
}
else if (Rs >= 641) {
return 8;
}
}
else if (refDCval >= 78) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
return 8;
}
}
}
}
else if (refDCval >= 141) {
if (Rs < 496) {
return 1;
}
else if (Rs >= 496) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 137) {
if (Cs < 317) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (ssDCval < 87) {
if (Cs < 250) {
return 2;
}
else if (Cs >= 250) {
return 1;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
}
else if (Cs >= 317) {
if (SC < 1882) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (AFD < 1) {
if (Rs < 602) {
return 8;
}
else if (Rs >= 602) {
if (Rs < 801) {
return 4;
}
else if (Rs >= 801) {
return 8;
}
}
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (SC >= 1882) {
if (SC < 1891) {
return 2;
}
else if (SC >= 1891) {
if (Rs < 1452) {
if (Rs < 1425) {
return 8;
}
else if (Rs >= 1425) {
return 2;
}
}
else if (Rs >= 1452) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 35) {
if (mcTcor < 82) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
if (ssDCval < 23) {
if (tcor < 55) {
if (Rs < 12) {
return 8;
}
else if (Rs >= 12) {
return 1;
}
}
else if (tcor >= 55) {
return 4;
}
}
else if (ssDCval >= 23) {
if (Cs < 377) {
if (negBalance < 45) {
if (Cs < 209) {
if (posBalance < 67) {
if (TSCindex < 2) {
if (Rs < 26) {
if (Rs < 4) {
return 4;
}
else if (Rs >= 4) {
return 1;
}
}
else if (Rs >= 26) {
return 2;
}
}
else if (TSCindex >= 2) {
if (ssDCval < 39) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (ssDCval >= 39) {
return 4;
}
}
}
else if (posBalance >= 67) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (Cs >= 209) {
if (AFD < 6) {
if (mu_mv_mag_sq < 53) {
return 2;
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 235) {
return 8;
}
else if (Rs >= 235) {
return 1;
}
}
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (Rs < 419) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (Rs < 53) {
return 4;
}
else if (Rs >= 53) {
return 2;
}
}
}
else if (Rs >= 419) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (Cs >= 377) {
if (RsDiff < 4) {
if (negBalance < 26) {
if (Rs < 452) {
return 1;
}
else if (Rs >= 452) {
return 8;
}
}
else if (negBalance >= 26) {
return 2;
}
}
else if (RsDiff >= 4) {
if (Rs < 1164) {
return 8;
}
else if (Rs >= 1164) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 82) {
if (posBalance < 27) {
if (MVDiff < 5) {
if (tcor < 102) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
if (Rs < 228) {
return 1;
}
else if (Rs >= 228) {
return 8;
}
}
}
else if (tcor >= 102) {
if (mu_mv_mag_sq < 106) {
if (ssDCval < 31) {
return 4;
}
else if (ssDCval >= 31) {
return 8;
}
}
else if (mu_mv_mag_sq >= 106) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (negBalance < 57) {
if (AFD < 2) {
if (posBalance < 20) {
if (mcTcor < 95) {
if (Rs < 132) {
return 8;
}
else if (Rs >= 132) {
return 1;
}
}
else if (mcTcor >= 95) {
if (ssDCval < 21) {
return 8;
}
else if (ssDCval >= 21) {
return 4;
}
}
}
else if (posBalance >= 20) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
if (Rs < 239) {
return 8;
}
else if (Rs >= 239) {
return 1;
}
}
}
}
else if (AFD >= 2) {
if (mu_mv_mag_sq < 77) {
if (ssDCval < 33) {
return 1;
}
else if (ssDCval >= 33) {
return 8;
}
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 256) {
if (Rs < 205) {
return 1;
}
else if (Rs >= 205) {
return 4;
}
}
else if (Rs >= 256) {
return 2;
}
}
}
}
else if (negBalance >= 57) {
if (mu_mv_mag_sq < 74) {
if (diffMVdiffVal < -30) {
return 4;
}
else if (diffMVdiffVal >= -30) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
if (tcor < 122) {
return 8;
}
else if (tcor >= 122) {
if (mcTcor < 101) {
if (negBalance < 77) {
return 4;
}
else if (negBalance >= 77) {
if (Rs < 207) {
return 8;
}
else if (Rs >= 207) {
return 4;
}
}
}
else if (mcTcor >= 101) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 74) {
return 4;
}
}
}
}
else if (posBalance >= 27) {
if (posBalance < 69) {
if (Cs < 82) {
if (tcor < 89) {
if (posBalance < 42) {
if (mu_mv_mag_sq < 9) {
if (tcor < 85) {
return 1;
}
else if (tcor >= 85) {
return 4;
}
}
else if (mu_mv_mag_sq >= 9) {
if (tcor < 77) {
if (Rs < 150) {
if (AFD < 9) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (Rs >= 150) {
return 4;
}
}
else if (tcor >= 77) {
return 1;
}
}
}
else if (posBalance >= 42) {
return 1;
}
}
else if (tcor >= 89) {
return 1;
}
}
else if (Cs >= 82) {
if (SC < 576) {
if (refDCval < 30) {
if (posBalance < 46) {
if (MVDiff < 23) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
if (diffMVdiffVal < -8) {
return 4;
}
else if (diffMVdiffVal >= -8) {
if (SC < 332) {
if (Rs < 175) {
return 1;
}
else if (Rs >= 175) {
return 4;
}
}
else if (SC >= 332) {
return 1;
}
}
}
}
else if (MVDiff >= 23) {
if (Cs < 90) {
return 8;
}
else if (Cs >= 90) {
if (SC < 474) {
return 1;
}
else if (SC >= 474) {
return 4;
}
}
}
}
else if (posBalance >= 46) {
if (tcor < 68) {
return 8;
}
else if (tcor >= 68) {
if (mu_mv_mag_sq < 81) {
if (MVDiff < 41) {
return 1;
}
else if (MVDiff >= 41) {
if (Cs < 194) {
return 8;
}
else if (Cs >= 194) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 81) {
return 8;
}
}
}
}
else if (refDCval >= 30) {
if (Cs < 261) {
if (mcTcor < 92) {
if (CsDiff < 1) {
if (Cs < 108) {
return 1;
}
else if (Cs >= 108) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Rs < 207) {
return 2;
}
else if (Rs >= 207) {
return 4;
}
}
}
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 3) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
if (Rs < 298) {
return 8;
}
else if (Rs >= 298) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (gchDC < 3) {
if (RsDiff < 1) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 8;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 109) {
if (Rs < 76) {
return 1;
}
else if (Rs >= 76) {
return 8;
}
}
else if (Rs >= 109) {
if (Cs < 144) {
if (diffAFD < 3) {
if (SC < 207) {
return 8;
}
else if (SC >= 207) {
if (tcor < 80) {
if (negBalance < 25) {
return 4;
}
else if (negBalance >= 25) {
return 1;
}
}
else if (tcor >= 80) {
if (posBalance < 31) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 31) {
return 1;
}
}
}
}
else if (diffAFD >= 3) {
return 2;
}
}
else if (Cs >= 144) {
if (MVDiff < 8) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (MVDiff >= 8) {
if (SC < 347) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
return 2;
}
}
else if (SC >= 347) {
if (Rs < 318) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 48) {
return 1;
}
else if (mu_mv_mag_sq >= 48) {
return 4;
}
}
}
else if (Rs >= 318) {
if (Cs < 220) {
return 4;
}
else if (Cs >= 220) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 261) {
return 1;
}
}
}
else if (SC >= 576) {
if (TSC < 3) {
if (posBalance < 39) {
if (mu_mv_mag_sq < 61) {
if (SC < 715) {
if (Rs < 350) {
return 4;
}
else if (Rs >= 350) {
return 1;
}
}
else if (SC >= 715) {
return 4;
}
}
else if (mu_mv_mag_sq >= 61) {
if (ssDCval < 32) {
return 1;
}
else if (ssDCval >= 32) {
if (Cs < 278) {
return 4;
}
else if (Cs >= 278) {
return 1;
}
}
}
}
else if (posBalance >= 39) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 607) {
return 8;
}
else if (Rs >= 607) {
return 1;
}
}
}
}
}
else if (TSC >= 3) {
if (Cs < 409) {
if (Rs < 422) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Rs >= 422) {
return 1;
}
}
else if (Cs >= 409) {
if (tcor < 81) {
return 1;
}
else if (tcor >= 81) {
if (Rs < 796) {
return 8;
}
else if (Rs >= 796) {
return 1;
}
}
}
}
}
}
}
else if (posBalance >= 69) {
if (mu_mv_mag_sq < 14) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 14) {
if (Rs < 59) {
return 4;
}
else if (Rs >= 59) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 35) {
if (Rs < 252) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 52) {
if (negBalance < 39) {
if (tcor < 91) {
if (posBalance < 53) {
if (Cs < 194) {
if (tcor < 81) {
if (SC < 130) {
return 2;
}
else if (SC >= 130) {
return 1;
}
}
else if (tcor >= 81) {
if (Rs < 43) {
return 1;
}
else if (Rs >= 43) {
if (AFD < 5) {
if (Rs < 128) {
return 4;
}
else if (Rs >= 128) {
return 2;
}
}
else if (AFD >= 5) {
return 4;
}
}
}
}
else if (Cs >= 194) {
return 1;
}
}
else if (posBalance >= 53) {
if (refDCval < 38) {
if (MVDiff < 17) {
if (Rs < 81) {
return 2;
}
else if (Rs >= 81) {
return 8;
}
}
else if (MVDiff >= 17) {
if (mu_mv_mag_sq < 5) {
if (Rs < 42) {
return 1;
}
else if (Rs >= 42) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
return 1;
}
}
}
else if (refDCval >= 38) {
if (diffMVdiffVal < 2) {
if (Cs < 101) {
return 1;
}
else if (Cs >= 101) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 92) {
return 4;
}
else if (Rs >= 92) {
return 2;
}
}
}
}
}
}
else if (tcor >= 91) {
if (RsCsDiff < 5) {
return 1;
}
else if (RsCsDiff >= 5) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (AFD < 5) {
if (mcTcor < 107) {
if (Cs < 175) {
if (MVDiff < 20) {
if (Rs < 106) {
return 1;
}
else if (Rs >= 106) {
if (Cs < 13) {
return 1;
}
else if (Cs >= 13) {
return 8;
}
}
}
else if (MVDiff >= 20) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
return 2;
}
}
}
else if (Cs >= 175) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (mcTcor >= 107) {
return 4;
}
}
else if (AFD >= 5) {
if (SC < 325) {
if (Rs < 123) {
if (posBalance < 6) {
if (mu_mv_mag_sq < 40) {
return 8;
}
else if (mu_mv_mag_sq >= 40) {
return 4;
}
}
else if (posBalance >= 6) {
if (ssDCval < 38) {
if (posBalance < 44) {
if (AFD < 16) {
if (MVDiff < 109) {
if (Cs < 13) {
return 2;
}
else if (Cs >= 13) {
return 1;
}
}
else if (MVDiff >= 109) {
return 2;
}
}
else if (AFD >= 16) {
return 1;
}
}
else if (posBalance >= 44) {
if (Rs < 101) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (Rs >= 101) {
return 4;
}
}
}
else if (ssDCval >= 38) {
if (tcor < 79) {
if (ssDCval < 47) {
if (mu_mv_mag_sq < 34) {
if (MVDiff < 99) {
if (Cs < 124) {
return 4;
}
else if (Cs >= 124) {
if (Rs < 114) {
return 1;
}
else if (Rs >= 114) {
return 4;
}
}
}
else if (MVDiff >= 99) {
if (ssDCval < 42) {
return 8;
}
else if (ssDCval >= 42) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 34) {
return 8;
}
}
else if (ssDCval >= 47) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (Rs < 95) {
return 8;
}
else if (Rs >= 95) {
return 4;
}
}
}
}
else if (tcor >= 79) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
}
}
}
else if (Rs >= 123) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
if (Cs < 195) {
if (ssDCval < 47) {
if (AFD < 35) {
if (diffMVdiffVal < -28) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
else if (diffMVdiffVal >= -28) {
return 1;
}
}
else if (AFD >= 35) {
return 8;
}
}
else if (ssDCval >= 47) {
if (Rs < 187) {
return 2;
}
else if (Rs >= 187) {
return 1;
}
}
}
else if (Cs >= 195) {
return 4;
}
}
}
}
else if (SC >= 325) {
if (AFD < 10) {
if (Rs < 166) {
if (Rs < 123) {
return 1;
}
else if (Rs >= 123) {
return 8;
}
}
else if (Rs >= 166) {
if (TSC < 4) {
if (mu_mv_mag_sq < 19) {
return 2;
}
else if (mu_mv_mag_sq >= 19) {
if (Rs < 221) {
return 1;
}
else if (Rs >= 221) {
return 8;
}
}
}
else if (TSC >= 4) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
if (Rs < 175) {
return 2;
}
else if (Rs >= 175) {
return 8;
}
}
}
}
}
else if (AFD >= 10) {
if (TSC < 5) {
if (mu_mv_mag_sq < 23) {
if (Rs < 160) {
return 1;
}
else if (Rs >= 160) {
return 4;
}
}
else if (mu_mv_mag_sq >= 23) {
if (Rs < 113) {
return 8;
}
else if (Rs >= 113) {
return 1;
}
}
}
else if (TSC >= 5) {
if (ssDCval < 36) {
if (ssDCval < 35) {
if (Rs < 206) {
return 8;
}
else if (Rs >= 206) {
return 2;
}
}
else if (ssDCval >= 35) {
return 4;
}
}
else if (ssDCval >= 36) {
if (tcor < 56) {
return 8;
}
else if (tcor >= 56) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (negBalance < 28) {
if (MVDiff < 2) {
if (Cs < 318) {
if (refDCval < 38) {
if (Cs < 217) {
return 4;
}
else if (Cs >= 217) {
if (Rs < 109) {
return 1;
}
else if (Rs >= 109) {
return 8;
}
}
}
else if (refDCval >= 38) {
if (Cs < 139) {
return 8;
}
else if (Cs >= 139) {
if (Rs < 192) {
return 1;
}
else if (Rs >= 192) {
if (Rs < 230) {
return 8;
}
else if (Rs >= 230) {
return 4;
}
}
}
}
}
else if (Cs >= 318) {
return 1;
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 110) {
if (ssDCval < 39) {
if (posBalance < 30) {
if (Cs < 222) {
return 4;
}
else if (Cs >= 222) {
return 1;
}
}
else if (posBalance >= 30) {
if (gchDC < 1) {
if (Cs < 294) {
if (mcTcor < 93) {
return 1;
}
else if (mcTcor >= 93) {
return 8;
}
}
else if (Cs >= 294) {
return 1;
}
}
else if (gchDC >= 1) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
}
else if (ssDCval >= 39) {
if (Cs < 46) {
return 8;
}
else if (Cs >= 46) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 110) {
if (mu_mv_mag_sq < 114) {
if (Rs < 47) {
return 8;
}
else if (Rs >= 47) {
return 4;
}
}
else if (mu_mv_mag_sq >= 114) {
if (Rs < 48) {
return 1;
}
else if (Rs >= 48) {
return 8;
}
}
}
}
}
else if (negBalance >= 28) {
if (gchDC < 1) {
if (SC < 204) {
if (SC < 135) {
return 8;
}
else if (SC >= 135) {
return 1;
}
}
else if (SC >= 204) {
return 8;
}
}
else if (gchDC >= 1) {
if (Rs < 184) {
if (TSCindex < 3) {
if (posBalance < 23) {
if (Rs < 8) {
return 2;
}
else if (Rs >= 8) {
return 8;
}
}
else if (posBalance >= 23) {
if (Cs < 186) {
return 1;
}
else if (Cs >= 186) {
return 8;
}
}
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 101) {
if (mcTcor < 103) {
if (tcor < 78) {
if (Rs < 104) {
if (tcor < 22) {
return 2;
}
else if (tcor >= 22) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (Rs >= 104) {
if (Rs < 149) {
if (Cs < 199) {
return 2;
}
else if (Cs >= 199) {
return 4;
}
}
else if (Rs >= 149) {
return 4;
}
}
}
else if (tcor >= 78) {
return 8;
}
}
else if (mcTcor >= 103) {
return 4;
}
}
else if (mu_mv_mag_sq >= 101) {
if (refDCval < 40) {
if (posBalance < 29) {
return 4;
}
else if (posBalance >= 29) {
if (Cs < 47) {
if (Rs < 68) {
return 8;
}
else if (Rs >= 68) {
return 2;
}
}
else if (Cs >= 47) {
return 8;
}
}
}
else if (refDCval >= 40) {
if (Rs < 35) {
return 1;
}
else if (Rs >= 35) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
return 4;
}
}
}
}
}
}
else if (Rs >= 184) {
if (diffAFD < 1) {
if (mu_mv_mag_sq < 94) {
if (refDCval < 40) {
return 8;
}
else if (refDCval >= 40) {
if (AFD < 28) {
return 1;
}
else if (AFD >= 28) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 94) {
return 8;
}
}
else if (diffAFD >= 1) {
if (diffMVdiffVal < 0) {
return 4;
}
else if (diffMVdiffVal >= 0) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 50) {
if (TSCindex < 2) {
if (refDCval < 66) {
if (Rs < 108) {
if (posBalance < 34) {
if (Cs < 40) {
return 2;
}
else if (Cs >= 40) {
return 8;
}
}
else if (posBalance >= 34) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
return 1;
}
}
}
else if (Rs >= 108) {
if (posBalance < 33) {
if (AFD < 1) {
if (Cs < 125) {
return 8;
}
else if (Cs >= 125) {
return 4;
}
}
else if (AFD >= 1) {
if (SC < 283) {
return 1;
}
else if (SC >= 283) {
return 8;
}
}
}
else if (posBalance >= 33) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
if (Rs < 200) {
if (posBalance < 44) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 1;
}
}
}
else if (posBalance >= 44) {
if (Cs < 190) {
if (refDCval < 57) {
if (Cs < 149) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (Cs >= 149) {
return 4;
}
}
else if (refDCval >= 57) {
return 1;
}
}
else if (Cs >= 190) {
return 4;
}
}
}
else if (Rs >= 200) {
return 4;
}
}
}
}
}
else if (refDCval >= 66) {
if (ssDCval < 93) {
if (Cs < 259) {
if (SC < 256) {
if (SC < 200) {
if (Rs < 88) {
return 4;
}
else if (Rs >= 88) {
return 1;
}
}
else if (SC >= 200) {
if (Rs < 125) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
return 2;
}
}
else if (Rs >= 125) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (Cs < 99) {
return 2;
}
else if (Cs >= 99) {
return 1;
}
}
}
}
}
else if (SC >= 256) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
if (MVDiff < 2) {
if (Rs < 200) {
return 2;
}
else if (Rs >= 200) {
return 8;
}
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
else if (Cs >= 259) {
if (SC < 483) {
if (refDCval < 69) {
if (Cs < 273) {
return 4;
}
else if (Cs >= 273) {
return 2;
}
}
else if (refDCval >= 69) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
return 1;
}
}
}
else if (SC >= 483) {
if (Rs < 164) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (Rs >= 164) {
if (Cs < 381) {
if (Cs < 321) {
return 1;
}
else if (Cs >= 321) {
return 4;
}
}
else if (Cs >= 381) {
return 1;
}
}
}
}
}
else if (ssDCval >= 93) {
if (Cs < 139) {
if (posBalance < 46) {
if (diffMVdiffVal < -19) {
return 8;
}
else if (diffMVdiffVal >= -19) {
return 1;
}
}
else if (posBalance >= 46) {
if (MVDiff < 25) {
return 4;
}
else if (MVDiff >= 25) {
return 1;
}
}
}
else if (Cs >= 139) {
if (MVDiff < 41) {
return 4;
}
else if (MVDiff >= 41) {
return 2;
}
}
}
}
}
else if (TSCindex >= 2) {
if (ssDCval < 71) {
if (Cs < 186) {
if (tcor < 94) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 76) {
if (tcor < 56) {
if (MVDiff < 118) {
if (Rs < 118) {
return 8;
}
else if (Rs >= 118) {
return 2;
}
}
else if (MVDiff >= 118) {
if (mcTcor < 43) {
if (Rs < 148) {
return 4;
}
else if (Rs >= 148) {
return 1;
}
}
else if (mcTcor >= 43) {
if (diffAFD < -7) {
return 2;
}
else if (diffAFD >= -7) {
return 1;
}
}
}
}
else if (tcor >= 56) {
if (mu_mv_mag_sq < 7) {
if (Rs < 58) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
else if (Rs >= 58) {
if (tcor < 72) {
if (Cs < 161) {
return 8;
}
else if (Cs >= 161) {
return 1;
}
}
else if (tcor >= 72) {
if (negBalance < 36) {
if (Cs < 142) {
if (Cs < 19) {
return 1;
}
else if (Cs >= 19) {
return 2;
}
}
else if (Cs >= 142) {
return 4;
}
}
else if (negBalance >= 36) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffMVdiffVal < 2) {
if (diffMVdiffVal < -7) {
if (Rs < 116) {
return 8;
}
else if (Rs >= 116) {
if (Cs < 82) {
return 1;
}
else if (Cs >= 82) {
return 4;
}
}
}
else if (diffMVdiffVal >= -7) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 4;
}
}
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (MVDiff < 6) {
if (Rs < 79) {
return 2;
}
else if (Rs >= 79) {
return 1;
}
}
else if (MVDiff >= 6) {
if (mcTcor < 56) {
if (AFD < 47) {
return 1;
}
else if (AFD >= 47) {
return 4;
}
}
else if (mcTcor >= 56) {
if (refDCval < 60) {
return 8;
}
else if (refDCval >= 60) {
if (Rs < 59) {
return 8;
}
else if (Rs >= 59) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (TSC < 4) {
if (TSCindex < 3) {
if (mu_mv_mag_sq < 13) {
return 1;
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 76) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Rs >= 76) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 1;
}
}
}
}
}
else if (TSCindex >= 3) {
if (MVDiff < 6) {
if (posBalance < 56) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (posBalance >= 56) {
if (Rs < 105) {
return 1;
}
else if (Rs >= 105) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (SC < 172) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
if (Rs < 21) {
if (MVDiff < 70) {
return 1;
}
else if (MVDiff >= 70) {
return 4;
}
}
else if (Rs >= 21) {
if (Rs < 85) {
return 4;
}
else if (Rs >= 85) {
return 1;
}
}
}
}
else if (SC >= 172) {
if (refDCval < 69) {
if (AFD < 6) {
if (ssDCval < 59) {
if (Rs < 158) {
return 1;
}
else if (Rs >= 158) {
return 4;
}
}
else if (ssDCval >= 59) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
if (Rs < 168) {
return 1;
}
else if (Rs >= 168) {
return 4;
}
}
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (refDCval >= 69) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (tcor < 87) {
if (mu_mv_mag_sq < 52) {
if (posBalance < 57) {
if (Rs < 135) {
return 1;
}
else if (Rs >= 135) {
return 2;
}
}
else if (posBalance >= 57) {
return 1;
}
}
else if (mu_mv_mag_sq >= 52) {
if (Rs < 88) {
return 8;
}
else if (Rs >= 88) {
return 4;
}
}
}
else if (tcor >= 87) {
return 2;
}
}
}
}
else if (tcor >= 94) {
if (negBalance < 49) {
if (Rs < 208) {
if (refDCval < 66) {
if (MVDiff < 1) {
if (Cs < 120) {
return 1;
}
else if (Cs >= 120) {
if (refDCval < 62) {
if (Rs < 133) {
return 8;
}
else if (Rs >= 133) {
return 1;
}
}
else if (refDCval >= 62) {
return 2;
}
}
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 100) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (tcor < 98) {
if (Cs < 167) {
return 4;
}
else if (Cs >= 167) {
return 1;
}
}
else if (tcor >= 98) {
if (ssDCval < 57) {
if (Rs < 126) {
return 1;
}
else if (Rs >= 126) {
return 4;
}
}
else if (ssDCval >= 57) {
return 1;
}
}
}
}
}
else if (refDCval >= 66) {
if (SC < 276) {
if (Rs < 73) {
return 1;
}
else if (Rs >= 73) {
if (Rs < 117) {
if (Rs < 88) {
return 4;
}
else if (Rs >= 88) {
return 2;
}
}
else if (Rs >= 117) {
return 1;
}
}
}
else if (SC >= 276) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (Rs >= 208) {
if (ssDCval < 60) {
if (Rs < 233) {
return 8;
}
else if (Rs >= 233) {
return 1;
}
}
else if (ssDCval >= 60) {
return 4;
}
}
}
else if (negBalance >= 49) {
if (SC < 144) {
if (Cs < 48) {
return 2;
}
else if (Cs >= 48) {
return 4;
}
}
else if (SC >= 144) {
if (RsCsDiff < 3) {
return 8;
}
else if (RsCsDiff >= 3) {
if (Rs < 115) {
return 2;
}
else if (Rs >= 115) {
return 1;
}
}
}
}
}
}
else if (Cs >= 186) {
if (TSC < 4) {
if (RsCsDiff < 1) {
if (negBalance < 29) {
if (Rs < 170) {
if (mu_mv_mag_sq < 82) {
if (Cs < 279) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (Cs >= 279) {
return 2;
}
}
else if (mu_mv_mag_sq >= 82) {
if (SC < 319) {
return 2;
}
else if (SC >= 319) {
if (Cs < 321) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 1;
}
}
else if (Cs >= 321) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 160) {
return 1;
}
else if (Rs >= 160) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 170) {
if (MVDiff < 1) {
if (ssDCval < 61) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 1;
}
}
}
else if (ssDCval >= 61) {
if (ssDCval < 69) {
if (Rs < 176) {
if (Cs < 330) {
return 4;
}
else if (Cs >= 330) {
return 1;
}
}
else if (Rs >= 176) {
return 4;
}
}
else if (ssDCval >= 69) {
if (Cs < 336) {
return 4;
}
else if (Cs >= 336) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 68) {
if (MVDiff < 5) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
if (Cs < 226) {
return 1;
}
else if (Cs >= 226) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (ssDCval >= 68) {
if (Cs < 354) {
if (Rs < 179) {
return 1;
}
else if (Rs >= 179) {
return 4;
}
}
else if (Cs >= 354) {
if (Rs < 224) {
return 1;
}
else if (Rs >= 224) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 29) {
if (tcor < 98) {
if (MVDiff < 6) {
if (Rs < 145) {
return 1;
}
else if (Rs >= 145) {
if (SC < 456) {
if (Cs < 210) {
if (Rs < 212) {
return 1;
}
else if (Rs >= 212) {
return 4;
}
}
else if (Cs >= 210) {
return 4;
}
}
else if (SC >= 456) {
if (negBalance < 34) {
if (diffAFD < 1) {
if (mu_mv_mag_sq < 119) {
return 4;
}
else if (mu_mv_mag_sq >= 119) {
return 1;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (negBalance >= 34) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (posBalance < 46) {
if (Rs < 118) {
if (Rs < 108) {
return 2;
}
else if (Rs >= 108) {
return 8;
}
}
else if (Rs >= 118) {
return 4;
}
}
else if (posBalance >= 46) {
if (tcor < 95) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
else if (tcor >= 95) {
return 1;
}
}
}
}
else if (tcor >= 98) {
if (posBalance < 49) {
if (Cs < 249) {
if (Rs < 143) {
return 8;
}
else if (Rs >= 143) {
return 1;
}
}
else if (Cs >= 249) {
if (mu_mv_mag_sq < 121) {
if (negBalance < 34) {
if (ssDCval < 64) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (ssDCval >= 64) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (diffAFD < 1) {
if (Cs < 344) {
return 1;
}
else if (Cs >= 344) {
return 4;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
}
}
else if (negBalance >= 34) {
if (Cs < 271) {
return 1;
}
else if (Cs >= 271) {
if (Rs < 231) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (Rs >= 231) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
return 1;
}
}
}
else if (posBalance >= 49) {
return 1;
}
}
}
}
else if (RsCsDiff >= 1) {
if (Rs < 65) {
return 1;
}
else if (Rs >= 65) {
if (MVDiff < 2) {
if (Cs < 404) {
if (ssDCval < 69) {
if (Rs < 167) {
return 1;
}
else if (Rs >= 167) {
if (mu_mv_mag_sq < 116) {
return 4;
}
else if (mu_mv_mag_sq >= 116) {
return 1;
}
}
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (Cs >= 404) {
return 1;
}
}
else if (MVDiff >= 2) {
if (negBalance < 51) {
if (gchDC < 1) {
if (Cs < 203) {
if (mu_mv_mag_sq < 67) {
return 4;
}
else if (mu_mv_mag_sq >= 67) {
return 1;
}
}
else if (Cs >= 203) {
if (ssDCval < 58) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Rs < 216) {
return 4;
}
else if (Rs >= 216) {
return 8;
}
}
}
}
else if (ssDCval >= 58) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (Rs < 232) {
if (ssDCval < 53) {
if (Cs < 326) {
return 1;
}
else if (Cs >= 326) {
return 4;
}
}
else if (ssDCval >= 53) {
if (posBalance < 41) {
if (Rs < 187) {
return 1;
}
else if (Rs >= 187) {
return 8;
}
}
else if (posBalance >= 41) {
if (tcor < 56) {
return 8;
}
else if (tcor >= 56) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
if (MVDiff < 10) {
if (Rs < 184) {
return 1;
}
else if (Rs >= 184) {
return 4;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
}
}
}
}
else if (Rs >= 232) {
if (RsCsDiff < 5) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
return 4;
}
}
else if (RsCsDiff >= 5) {
return 2;
}
}
}
}
else if (negBalance >= 51) {
if (tcor < 94) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (tcor >= 94) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 38) {
if (ssDCval < 67) {
if (CsDiff < 4) {
if (Rs < 213) {
if (diffAFD < 4) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (Rs < 122) {
return 2;
}
else if (Rs >= 122) {
return 8;
}
}
}
}
else if (diffAFD >= 4) {
return 2;
}
}
else if (Rs >= 213) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (Cs < 413) {
if (tcor < 77) {
return 1;
}
else if (tcor >= 77) {
if (SC < 482) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (SC >= 482) {
return 2;
}
}
}
else if (Cs >= 413) {
return 1;
}
}
}
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 67) {
if (negBalance < 59) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
return 2;
}
}
else if (negBalance >= 59) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (Rs < 76) {
return 2;
}
else if (Rs >= 76) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (diffTSC < 2) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 2;
}
}
}
}
else if (diffTSC >= 2) {
if (SC < 513) {
return 1;
}
else if (SC >= 513) {
if (Rs < 177) {
return 8;
}
else if (Rs >= 177) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (posBalance < 48) {
if (posBalance < 4) {
return 8;
}
else if (posBalance >= 4) {
if (MVDiff < 4) {
if (Cs < 211) {
if (Cs < 125) {
if (ssDCval < 88) {
if (Rs < 97) {
return 4;
}
else if (Rs >= 97) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 139) {
return 2;
}
else if (Rs >= 139) {
return 8;
}
}
}
}
else if (ssDCval >= 88) {
return 1;
}
}
else if (Cs >= 125) {
if (negBalance < 29) {
if (Cs < 161) {
if (posBalance < 46) {
if (Cs < 133) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
return 8;
}
}
else if (Cs >= 133) {
return 2;
}
}
else if (posBalance >= 46) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Rs < 201) {
return 2;
}
else if (Rs >= 201) {
return 8;
}
}
}
}
else if (Cs >= 161) {
if (Cs < 209) {
return 2;
}
else if (Cs >= 209) {
if (Rs < 184) {
return 2;
}
else if (Rs >= 184) {
return 1;
}
}
}
}
else if (negBalance >= 29) {
if (Rs < 191) {
if (ssDCval < 101) {
return 2;
}
else if (ssDCval >= 101) {
return 4;
}
}
else if (Rs >= 191) {
return 1;
}
}
}
}
else if (Cs >= 211) {
if (mu_mv_mag_sq < 104) {
if (mu_mv_mag_sq < 98) {
return 2;
}
else if (mu_mv_mag_sq >= 98) {
if (SC < 432) {
return 1;
}
else if (SC >= 432) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (SC < 400) {
return 2;
}
else if (SC >= 400) {
return 1;
}
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 109) {
if (TSC < 2) {
if (Cs < 89) {
return 1;
}
else if (Cs >= 89) {
if (Rs < 109) {
if (MVDiff < 13) {
if (Rs < 70) {
return 4;
}
else if (Rs >= 70) {
if (Cs < 236) {
return 2;
}
else if (Cs >= 236) {
return 4;
}
}
}
else if (MVDiff >= 13) {
if (Cs < 107) {
return 1;
}
else if (Cs >= 107) {
return 2;
}
}
}
else if (Rs >= 109) {
if (Cs < 256) {
if (SC < 349) {
if (SC < 228) {
return 2;
}
else if (SC >= 228) {
if (AFD < 7) {
if (ssDCval < 74) {
if (Cs < 148) {
return 2;
}
else if (Cs >= 148) {
return 1;
}
}
else if (ssDCval >= 74) {
if (negBalance < 18) {
return 2;
}
else if (negBalance >= 18) {
return 1;
}
}
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (SC >= 349) {
if (Rs < 240) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
return 2;
}
}
else if (Rs >= 240) {
return 1;
}
}
}
else if (Cs >= 256) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
return 4;
}
}
}
}
}
else if (TSC >= 2) {
if (TSC < 5) {
if (mcTcor < 100) {
if (MVDiff < 5) {
if (SC < 403) {
if (Cs < 107) {
return 8;
}
else if (Cs >= 107) {
return 1;
}
}
else if (SC >= 403) {
return 4;
}
}
else if (MVDiff >= 5) {
if (ssDCval < 106) {
if (refDCval < 74) {
if (Rs < 130) {
if (Rs < 119) {
if (Rs < 81) {
return 1;
}
else if (Rs >= 81) {
if (Cs < 201) {
return 2;
}
else if (Cs >= 201) {
return 1;
}
}
}
else if (Rs >= 119) {
if (Cs < 232) {
return 2;
}
else if (Cs >= 232) {
return 4;
}
}
}
else if (Rs >= 130) {
return 1;
}
}
else if (refDCval >= 74) {
return 1;
}
}
else if (ssDCval >= 106) {
if (Rs < 51) {
return 1;
}
else if (Rs >= 51) {
return 2;
}
}
}
}
else if (mcTcor >= 100) {
if (AFD < 6) {
if (Rs < 138) {
return 1;
}
else if (Rs >= 138) {
return 8;
}
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (TSC >= 5) {
if (MVDiff < 144) {
if (mu_mv_mag_sq < 10) {
if (Rs < 205) {
return 1;
}
else if (Rs >= 205) {
return 2;
}
}
else if (mu_mv_mag_sq >= 10) {
return 8;
}
}
else if (MVDiff >= 144) {
return 1;
}
}
}
}
else if (ssDCval >= 109) {
if (MVDiff < 117) {
if (Rs < 1) {
return 1;
}
else if (Rs >= 1) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 2;
}
}
}
}
else if (MVDiff >= 117) {
if (Rs < 82) {
if (tcor < 84) {
if (MVDiff < 394) {
return 2;
}
else if (MVDiff >= 394) {
return 1;
}
}
else if (tcor >= 84) {
return 1;
}
}
else if (Rs >= 82) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 48) {
if (mcTcor < 97) {
if (mcTcor < 71) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
if (Cs < 359) {
return 1;
}
else if (Cs >= 359) {
return 8;
}
}
}
else if (mcTcor >= 71) {
if (MVDiff < 6) {
if (Rs < 120) {
return 1;
}
else if (Rs >= 120) {
if (diffMVdiffVal < -8) {
return 2;
}
else if (diffMVdiffVal >= -8) {
if (tcor < 77) {
if (SC < 446) {
if (Rs < 163) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 2;
}
}
else if (Rs >= 163) {
return 2;
}
}
else if (SC >= 446) {
return 8;
}
}
else if (tcor >= 77) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
if (AFD < 3) {
if (Rs < 54) {
return 4;
}
else if (Rs >= 54) {
return 1;
}
}
else if (AFD >= 3) {
if (refDCval < 80) {
if (refDCval < 74) {
if (refDCval < 71) {
if (Rs < 101) {
return 1;
}
else if (Rs >= 101) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
return 2;
}
}
}
else if (refDCval >= 71) {
return 1;
}
}
else if (refDCval >= 74) {
if (Rs < 150) {
if (negBalance < 48) {
if (Cs < 122) {
return 1;
}
else if (Cs >= 122) {
return 4;
}
}
else if (negBalance >= 48) {
if (Cs < 178) {
return 2;
}
else if (Cs >= 178) {
return 1;
}
}
}
else if (Rs >= 150) {
if (diffTSC < 1) {
if (Cs < 288) {
return 1;
}
else if (Cs >= 288) {
return 4;
}
}
else if (diffTSC >= 1) {
if (Rs < 195) {
return 4;
}
else if (Rs >= 195) {
return 8;
}
}
}
}
}
else if (refDCval >= 80) {
if (posBalance < 62) {
if (AFD < 4) {
if (Cs < 162) {
return 1;
}
else if (Cs >= 162) {
return 2;
}
}
else if (AFD >= 4) {
if (negBalance < 56) {
if (ssDCval < 86) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (ssDCval >= 86) {
if (Rs < 239) {
if (negBalance < 54) {
if (ssDCval < 107) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
if (posBalance < 58) {
if (Rs < 180) {
if (Rs < 139) {
return 8;
}
else if (Rs >= 139) {
return 2;
}
}
else if (Rs >= 180) {
return 8;
}
}
else if (posBalance >= 58) {
if (diffMVdiffVal < -4) {
if (posBalance < 60) {
return 2;
}
else if (posBalance >= 60) {
if (Rs < 132) {
return 2;
}
else if (Rs >= 132) {
return 8;
}
}
}
else if (diffMVdiffVal >= -4) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (Cs < 14) {
return 2;
}
else if (Cs >= 14) {
if (Rs < 94) {
return 2;
}
else if (Rs >= 94) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 107) {
return 2;
}
}
else if (negBalance >= 54) {
if (MVDiff < 29) {
return 8;
}
else if (MVDiff >= 29) {
if (Cs < 14) {
return 8;
}
else if (Cs >= 14) {
return 2;
}
}
}
}
else if (Rs >= 239) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (posBalance < 58) {
if (mu_mv_mag_sq < 9) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 2;
}
}
else if (mu_mv_mag_sq >= 9) {
if (posBalance < 51) {
if (Cs < 102) {
return 8;
}
else if (Cs >= 102) {
if (Rs < 155) {
return 1;
}
else if (Rs >= 155) {
return 2;
}
}
}
else if (posBalance >= 51) {
if (mu_mv_mag_sq < 82) {
return 1;
}
else if (mu_mv_mag_sq >= 82) {
if (Rs < 111) {
return 8;
}
else if (Rs >= 111) {
return 2;
}
}
}
}
}
else if (posBalance >= 58) {
if (diffAFD < 3) {
if (Rs < 242) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
return 4;
}
}
else if (Rs >= 242) {
return 1;
}
}
else if (diffAFD >= 3) {
return 1;
}
}
}
}
}
else if (posBalance >= 62) {
if (MVDiff < 82) {
if (ssDCval < 116) {
if (refDCval < 82) {
return 8;
}
else if (refDCval >= 82) {
if (TSC < 3) {
if (Cs < 56) {
return 1;
}
else if (Cs >= 56) {
return 2;
}
}
else if (TSC >= 3) {
if (Rs < 214) {
return 1;
}
else if (Rs >= 214) {
if (Cs < 70) {
return 1;
}
else if (Cs >= 70) {
if (Rs < 232) {
return 8;
}
else if (Rs >= 232) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 116) {
if (mcTcor < 86) {
if (Rs < 220) {
return 4;
}
else if (Rs >= 220) {
return 1;
}
}
else if (mcTcor >= 86) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 2;
}
}
}
}
else if (MVDiff >= 82) {
if (MVDiff < 219) {
if (posBalance < 65) {
if (mu_mv_mag_sq < 15) {
return 2;
}
else if (mu_mv_mag_sq >= 15) {
if (Rs < 127) {
return 1;
}
else if (Rs >= 127) {
return 8;
}
}
}
else if (posBalance >= 65) {
return 2;
}
}
else if (MVDiff >= 219) {
return 1;
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 231) {
if (TSC < 2) {
if (Rs < 77) {
if (SC < 146) {
if (Cs < 91) {
if (ssDCval < 117) {
return 1;
}
else if (ssDCval >= 117) {
return 4;
}
}
else if (Cs >= 91) {
return 2;
}
}
else if (SC >= 146) {
return 4;
}
}
else if (Rs >= 77) {
if (ssDCval < 75) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
if (Cs < 217) {
return 2;
}
else if (Cs >= 217) {
return 1;
}
}
}
else if (ssDCval >= 75) {
if (Rs < 143) {
if (ssDCval < 104) {
if (Cs < 210) {
return 1;
}
else if (Cs >= 210) {
return 4;
}
}
else if (ssDCval >= 104) {
if (Cs < 117) {
return 2;
}
else if (Cs >= 117) {
return 4;
}
}
}
else if (Rs >= 143) {
if (Rs < 239) {
return 2;
}
else if (Rs >= 239) {
return 1;
}
}
}
}
}
else if (TSC >= 2) {
if (ssDCval < 98) {
if (Cs < 191) {
if (refDCval < 76) {
if (negBalance < 30) {
if (Cs < 149) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (Cs >= 149) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (negBalance >= 30) {
if (diffMVdiffVal < -11) {
return 4;
}
else if (diffMVdiffVal >= -11) {
if (TSCindex < 3) {
return 2;
}
else if (TSCindex >= 3) {
return 1;
}
}
}
}
else if (refDCval >= 76) {
if (TSC < 3) {
if (Rs < 198) {
if (MVDiff < 222) {
return 1;
}
else if (MVDiff >= 222) {
return 4;
}
}
else if (Rs >= 198) {
return 4;
}
}
else if (TSC >= 3) {
if (refDCval < 93) {
if (refDCval < 83) {
if (Rs < 91) {
return 4;
}
else if (Rs >= 91) {
if (TSC < 4) {
if (RsCsDiff < 2) {
return 2;
}
else if (RsCsDiff >= 2) {
return 1;
}
}
else if (TSC >= 4) {
return 1;
}
}
}
else if (refDCval >= 83) {
return 1;
}
}
else if (refDCval >= 93) {
if (tcor < 93) {
if (Rs < 165) {
return 1;
}
else if (Rs >= 165) {
return 4;
}
}
else if (tcor >= 93) {
return 2;
}
}
}
}
}
else if (Cs >= 191) {
if (ssDCval < 75) {
if (Rs < 168) {
return 2;
}
else if (Rs >= 168) {
return 1;
}
}
else if (ssDCval >= 75) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
}
}
else if (ssDCval >= 98) {
if (gchDC < 5) {
if (mu_mv_mag_sq < 74) {
if (negBalance < 57) {
if (TSCindex < 3) {
if (Rs < 51) {
return 1;
}
else if (Rs >= 51) {
return 2;
}
}
else if (TSCindex >= 3) {
return 2;
}
}
else if (negBalance >= 57) {
if (negBalance < 63) {
return 1;
}
else if (negBalance >= 63) {
if (Rs < 44) {
return 2;
}
else if (Rs >= 44) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 74) {
return 1;
}
}
else if (gchDC >= 5) {
if (Rs < 110) {
if (MVDiff < 196) {
return 8;
}
else if (MVDiff >= 196) {
return 4;
}
}
else if (Rs >= 110) {
if (Cs < 161) {
return 1;
}
else if (Cs >= 161) {
return 4;
}
}
}
}
}
}
else if (Cs >= 231) {
if (refDCval < 76) {
if (Rs < 114) {
if (Rs < 82) {
return 2;
}
else if (Rs >= 82) {
return 4;
}
}
else if (Rs >= 114) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (Rs < 174) {
return 4;
}
else if (Rs >= 174) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 4;
}
}
}
}
}
else if (refDCval >= 76) {
if (RsDiff < 1) {
if (Cs < 376) {
if (Rs < 182) {
if (ssDCval < 89) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
return 4;
}
}
else if (ssDCval >= 89) {
if (mu_mv_mag_sq < 60) {
return 4;
}
else if (mu_mv_mag_sq >= 60) {
if (Rs < 52) {
return 2;
}
else if (Rs >= 52) {
return 1;
}
}
}
}
else if (Rs >= 182) {
if (Cs < 303) {
return 2;
}
else if (Cs >= 303) {
return 4;
}
}
}
else if (Cs >= 376) {
return 2;
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 25) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
if (diffRsCsdiff < -2) {
if (Rs < 245) {
return 2;
}
else if (Rs >= 245) {
return 4;
}
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (diffMVdiffVal < 3) {
if (Cs < 448) {
if (MVDiff < 45) {
return 1;
}
else if (MVDiff >= 45) {
if (Rs < 120) {
return 2;
}
else if (Rs >= 120) {
if (Rs < 153) {
return 8;
}
else if (Rs >= 153) {
return 1;
}
}
}
}
else if (Cs >= 448) {
return 2;
}
}
else if (diffMVdiffVal >= 3) {
if (mu_mv_mag_sq < 43) {
return 8;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 111) {
return 1;
}
else if (Rs >= 111) {
if (Rs < 164) {
return 8;
}
else if (Rs >= 164) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 252) {
if (RsDiff < 4) {
if (ssDCval < 75) {
if (mcTcor < 89) {
if (refDCval < 41) {
if (tcor < 60) {
if (negBalance < 45) {
if (refDCval < 39) {
return 1;
}
else if (refDCval >= 39) {
return 8;
}
}
else if (negBalance >= 45) {
if (Cs < 270) {
return 8;
}
else if (Cs >= 270) {
return 4;
}
}
}
else if (tcor >= 60) {
if (diffRsCsdiff < -6) {
if (Cs < 323) {
return 1;
}
else if (Cs >= 323) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (diffRsCsdiff >= -6) {
if (Cs < 152) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Cs >= 152) {
if (negBalance < 27) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
else if (negBalance >= 27) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (negBalance < 45) {
if (Rs < 393) {
return 8;
}
else if (Rs >= 393) {
return 2;
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 41) {
if (MVDiff < 59) {
if (mu_mv_mag_sq < 101) {
if (posBalance < 58) {
if (Cs < 346) {
if (negBalance < 47) {
if (Cs < 283) {
if (posBalance < 52) {
if (Rs < 395) {
return 8;
}
else if (Rs >= 395) {
return 2;
}
}
else if (posBalance >= 52) {
return 2;
}
}
else if (Cs >= 283) {
if (Rs < 690) {
return 8;
}
else if (Rs >= 690) {
return 1;
}
}
}
else if (negBalance >= 47) {
if (posBalance < 53) {
if (ssDCval < 67) {
if (Rs < 385) {
return 8;
}
else if (Rs >= 385) {
return 1;
}
}
else if (ssDCval >= 67) {
return 1;
}
}
else if (posBalance >= 53) {
if (Cs < 161) {
return 1;
}
else if (Cs >= 161) {
if (Rs < 376) {
return 4;
}
else if (Rs >= 376) {
return 8;
}
}
}
}
}
else if (Cs >= 346) {
if (Rs < 383) {
if (diffMVdiffVal < 1) {
if (Rs < 354) {
return 2;
}
else if (Rs >= 354) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
else if (Rs >= 383) {
if (ssDCval < 58) {
if (mu_mv_mag_sq < 48) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
else if (mu_mv_mag_sq >= 48) {
if (RsCsDiff < 16) {
return 4;
}
else if (RsCsDiff >= 16) {
return 1;
}
}
}
else if (ssDCval >= 58) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
else if (posBalance >= 58) {
if (TSC < 6) {
if (tcor < 81) {
return 4;
}
else if (tcor >= 81) {
if (Cs < 309) {
return 1;
}
else if (Cs >= 309) {
if (Rs < 645) {
return 4;
}
else if (Rs >= 645) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (refDCval < 61) {
if (posBalance < 59) {
if (Rs < 309) {
return 8;
}
else if (Rs >= 309) {
return 4;
}
}
else if (posBalance >= 59) {
if (Rs < 774) {
if (MVDiff < 47) {
return 8;
}
else if (MVDiff >= 47) {
return 4;
}
}
else if (Rs >= 774) {
return 2;
}
}
}
else if (refDCval >= 61) {
if (AFD < 20) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (AFD >= 20) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
else if (MVDiff >= 59) {
if (gchDC < 4) {
if (tcor < 51) {
return 4;
}
else if (tcor >= 51) {
if (diffMVdiffVal < 39) {
return 2;
}
else if (diffMVdiffVal >= 39) {
return 4;
}
}
}
else if (gchDC >= 4) {
if (Cs < 124) {
return 1;
}
else if (Cs >= 124) {
if (refDCval < 50) {
if (Cs < 509) {
return 1;
}
else if (Cs >= 509) {
return 8;
}
}
else if (refDCval >= 50) {
if (refDCval < 67) {
if (TSC < 9) {
if (Rs < 716) {
return 2;
}
else if (Rs >= 716) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (refDCval >= 67) {
if (Rs < 438) {
return 2;
}
else if (Rs >= 438) {
return 4;
}
}
}
}
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1286) {
if (posBalance < 26) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
if (tcor < 101) {
if (Cs < 373) {
if (posBalance < 5) {
return 8;
}
else if (posBalance >= 5) {
if (Cs < 38) {
return 4;
}
else if (Cs >= 38) {
if (SC < 442) {
return 4;
}
else if (SC >= 442) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 1;
}
}
}
}
}
else if (Cs >= 373) {
if (posBalance < 24) {
return 8;
}
else if (posBalance >= 24) {
if (Cs < 438) {
if (Rs < 911) {
return 1;
}
else if (Rs >= 911) {
return 2;
}
}
else if (Cs >= 438) {
return 8;
}
}
}
}
else if (tcor >= 101) {
if (AFD < 8) {
if (ssDCval < 37) {
return 1;
}
else if (ssDCval >= 37) {
if (CsDiff < 1) {
if (SC < 838) {
if (posBalance < 19) {
if (mcTcor < 106) {
return 8;
}
else if (mcTcor >= 106) {
return 4;
}
}
else if (posBalance >= 19) {
return 4;
}
}
else if (SC >= 838) {
return 8;
}
}
else if (CsDiff >= 1) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (AFD >= 8) {
return 1;
}
}
}
}
else if (posBalance >= 26) {
if (mu_mv_mag_sq < 107) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Cs < 123) {
if (AFD < 5) {
if (Cs < 83) {
if (Rs < 556) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
return 1;
}
}
else if (Rs >= 556) {
if (Cs < 22) {
return 4;
}
else if (Cs >= 22) {
return 8;
}
}
}
else if (Cs >= 83) {
if (negBalance < 33) {
if (diffMVdiffVal < -4) {
if (RsDiff < 1) {
if (Rs < 570) {
return 1;
}
else if (Rs >= 570) {
return 4;
}
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
if (MVDiff < 18) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (MVDiff >= 18) {
return 4;
}
}
}
else if (negBalance >= 33) {
return 1;
}
}
}
else if (AFD >= 5) {
if (MVDiff < 7) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (MVDiff >= 7) {
if (diffTSC < -1) {
if (tcor < 85) {
if (Rs < 446) {
return 1;
}
else if (Rs >= 446) {
if (Rs < 617) {
return 8;
}
else if (Rs >= 617) {
return 1;
}
}
}
else if (tcor >= 85) {
return 1;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
}
}
else if (Cs >= 123) {
if (mu_mv_mag_sq < 85) {
if (mcTcor < 97) {
if (ssDCval < 45) {
if (MVDiff < 4) {
if (negBalance < 40) {
if (ssDCval < 39) {
return 1;
}
else if (ssDCval >= 39) {
if (Rs < 502) {
if (Rs < 409) {
return 1;
}
else if (Rs >= 409) {
return 4;
}
}
else if (Rs >= 502) {
if (Rs < 859) {
return 8;
}
else if (Rs >= 859) {
return 2;
}
}
}
}
else if (negBalance >= 40) {
return 8;
}
}
else if (MVDiff >= 4) {
if (Cs < 480) {
if (posBalance < 37) {
if (Rs < 414) {
return 4;
}
else if (Rs >= 414) {
return 2;
}
}
else if (posBalance >= 37) {
if (mcTcor < 90) {
if (Rs < 305) {
return 8;
}
else if (Rs >= 305) {
return 4;
}
}
else if (mcTcor >= 90) {
if (tcor < 62) {
return 4;
}
else if (tcor >= 62) {
if (gchDC < 3) {
if (tcor < 65) {
if (Rs < 674) {
return 1;
}
else if (Rs >= 674) {
return 8;
}
}
else if (tcor >= 65) {
if (MVDiff < 6) {
if (diffAFD < 2) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 4;
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
}
else if (gchDC >= 3) {
if (refDCval < 40) {
return 1;
}
else if (refDCval >= 40) {
if (Rs < 447) {
return 2;
}
else if (Rs >= 447) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 480) {
if (negBalance < 42) {
if (mu_mv_mag_sq < 35) {
return 8;
}
else if (mu_mv_mag_sq >= 35) {
return 2;
}
}
else if (negBalance >= 42) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
}
else if (ssDCval >= 45) {
if (TSC < 4) {
if (RsDiff < 1) {
if (Cs < 365) {
return 4;
}
else if (Cs >= 365) {
if (Rs < 276) {
return 4;
}
else if (Rs >= 276) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (mcTcor < 93) {
if (negBalance < 38) {
if (diffTSC < 1) {
if (Rs < 816) {
return 1;
}
else if (Rs >= 816) {
if (Rs < 1130) {
return 4;
}
else if (Rs >= 1130) {
return 1;
}
}
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (negBalance >= 38) {
if (AFD < 6) {
if (Rs < 576) {
return 2;
}
else if (Rs >= 576) {
return 1;
}
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (refDCval < 57) {
if (Rs < 482) {
if (Rs < 304) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (refDCval < 48) {
if (Rs < 273) {
return 1;
}
else if (Rs >= 273) {
return 4;
}
}
else if (refDCval >= 48) {
if (Cs < 188) {
return 4;
}
else if (Cs >= 188) {
return 2;
}
}
}
}
else if (Rs >= 304) {
if (MVDiff < 8) {
if (tcor < 89) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
else if (tcor >= 89) {
return 1;
}
}
else if (MVDiff >= 8) {
return 4;
}
}
}
else if (Rs >= 482) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
}
else if (refDCval >= 57) {
if (ssDCval < 62) {
if (Rs < 816) {
if (refDCval < 60) {
if (mu_mv_mag_sq < 10) {
if (Rs < 280) {
return 8;
}
else if (Rs >= 280) {
return 2;
}
}
else if (mu_mv_mag_sq >= 10) {
if (Cs < 267) {
if (Rs < 331) {
return 4;
}
else if (Rs >= 331) {
return 2;
}
}
else if (Cs >= 267) {
if (Cs < 479) {
if (diffMVdiffVal < -3) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
if (SC < 632) {
return 4;
}
else if (SC >= 632) {
return 1;
}
}
}
else if (Cs >= 479) {
return 2;
}
}
}
}
else if (refDCval >= 60) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
else if (Rs >= 816) {
if (Cs < 341) {
return 1;
}
else if (Cs >= 341) {
return 4;
}
}
}
else if (ssDCval >= 62) {
if (diffMVdiffVal < 3) {
if (posBalance < 55) {
if (gchDC < 1) {
if (MVDiff < 7) {
if (Cs < 222) {
return 4;
}
else if (Cs >= 222) {
return 1;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (posBalance >= 55) {
if (Cs < 409) {
return 4;
}
else if (Cs >= 409) {
return 2;
}
}
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
}
}
}
}
else if (TSC >= 4) {
if (MVDiff < 51) {
if (ssDCval < 58) {
if (MVDiff < 12) {
if (Rs < 445) {
if (SC < 856) {
return 4;
}
else if (SC >= 856) {
return 8;
}
}
else if (Rs >= 445) {
if (refDCval < 45) {
return 1;
}
else if (refDCval >= 45) {
return 8;
}
}
}
else if (MVDiff >= 12) {
if (ssDCval < 52) {
if (Cs < 159) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
else if (Cs >= 159) {
if (diffAFD < -3) {
if (Rs < 383) {
return 1;
}
else if (Rs >= 383) {
return 8;
}
}
else if (diffAFD >= -3) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 8;
}
}
}
}
else if (ssDCval >= 52) {
if (mcTcor < 93) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
else if (mcTcor >= 93) {
if (tcor < 84) {
if (ssDCval < 55) {
if (MVDiff < 27) {
return 1;
}
else if (MVDiff >= 27) {
return 4;
}
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (tcor >= 84) {
if (Cs < 298) {
return 1;
}
else if (Cs >= 298) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 58) {
if (ssDCval < 72) {
if (diffRsCsdiff < -9) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (Rs < 808) {
return 1;
}
else if (Rs >= 808) {
return 2;
}
}
}
else if (diffRsCsdiff >= -9) {
if (RsCsDiff < 12) {
if (SC < 1093) {
if (tcor < 84) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
if (refDCval < 66) {
return 2;
}
else if (refDCval >= 66) {
return 4;
}
}
}
else if (tcor >= 84) {
if (refDCval < 70) {
if (Rs < 553) {
if (MVDiff < 5) {
if (Rs < 444) {
if (Rs < 425) {
return 1;
}
else if (Rs >= 425) {
return 2;
}
}
else if (Rs >= 444) {
return 4;
}
}
else if (MVDiff >= 5) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (Rs < 359) {
return 4;
}
else if (Rs >= 359) {
if (Rs < 426) {
return 1;
}
else if (Rs >= 426) {
return 2;
}
}
}
}
}
else if (Rs >= 553) {
if (Rs < 874) {
return 1;
}
else if (Rs >= 874) {
return 2;
}
}
}
else if (refDCval >= 70) {
if (Rs < 573) {
return 1;
}
else if (Rs >= 573) {
return 4;
}
}
}
}
else if (SC >= 1093) {
return 4;
}
}
else if (RsCsDiff >= 12) {
if (SC < 896) {
return 1;
}
else if (SC >= 896) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (Cs < 358) {
return 1;
}
else if (Cs >= 358) {
if (CsDiff < 4) {
if (Rs < 524) {
if (Rs < 468) {
return 4;
}
else if (Rs >= 468) {
return 2;
}
}
else if (Rs >= 524) {
return 4;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 72) {
if (diffTSC < 1) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
if (Cs < 528) {
return 8;
}
else if (Cs >= 528) {
return 4;
}
}
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
else if (MVDiff >= 51) {
if (diffMVdiffVal < -18) {
if (Rs < 383) {
return 1;
}
else if (Rs >= 383) {
return 2;
}
}
else if (diffMVdiffVal >= -18) {
if (diffRsCsdiff < 4) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 1;
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 411) {
return 1;
}
else if (Rs >= 411) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 717) {
if (TSC < 4) {
if (MVDiff < 2) {
if (diffMVdiffVal < -2) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
if (SC < 954) {
return 1;
}
else if (SC >= 954) {
if (Cs < 370) {
return 1;
}
else if (Cs >= 370) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
return 8;
}
}
}
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 24) {
if (Cs < 296) {
if (negBalance < 45) {
if (SC < 502) {
return 1;
}
else if (SC >= 502) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
if (Rs < 260) {
return 1;
}
else if (Rs >= 260) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 45) {
if (Cs < 237) {
return 4;
}
else if (Cs >= 237) {
return 8;
}
}
}
else if (Cs >= 296) {
if (MVDiff < 7) {
if (Cs < 372) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (Rs < 561) {
return 4;
}
else if (Rs >= 561) {
return 1;
}
}
}
else if (Cs >= 372) {
return 4;
}
}
else if (MVDiff >= 7) {
if (Cs < 484) {
if (ssDCval < 70) {
if (posBalance < 64) {
if (negBalance < 51) {
if (Rs < 679) {
return 4;
}
else if (Rs >= 679) {
if (Rs < 711) {
return 2;
}
else if (Rs >= 711) {
return 4;
}
}
}
else if (negBalance >= 51) {
return 2;
}
}
else if (posBalance >= 64) {
return 2;
}
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (Cs >= 484) {
if (Cs < 506) {
return 1;
}
else if (Cs >= 506) {
if (Rs < 429) {
return 2;
}
else if (Rs >= 429) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (TSC < 3) {
if (AFD < 6) {
if (Cs < 419) {
if (refDCval < 69) {
if (RsCsDiff < 2) {
if (SC < 664) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Rs < 256) {
return 1;
}
else if (Rs >= 256) {
return 4;
}
}
}
else if (SC >= 664) {
if (AFD < 4) {
if (Rs < 506) {
return 4;
}
else if (Rs >= 506) {
return 1;
}
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (RsCsDiff >= 2) {
if (ssDCval < 48) {
return 2;
}
else if (ssDCval >= 48) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
if (Rs < 341) {
return 2;
}
else if (Rs >= 341) {
return 4;
}
}
}
}
}
else if (refDCval >= 69) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 8;
}
}
}
else if (Cs >= 419) {
if (MVDiff < 9) {
if (refDCval < 60) {
if (mu_mv_mag_sq < 62) {
if (Rs < 401) {
return 4;
}
else if (Rs >= 401) {
if (Rs < 459) {
return 1;
}
else if (Rs >= 459) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 62) {
return 8;
}
}
else if (refDCval >= 60) {
if (Rs < 467) {
return 4;
}
else if (Rs >= 467) {
return 1;
}
}
}
else if (MVDiff >= 9) {
if (MVDiff < 14) {
if (Rs < 550) {
return 4;
}
else if (Rs >= 550) {
return 2;
}
}
else if (MVDiff >= 14) {
return 1;
}
}
}
}
else if (AFD >= 6) {
if (Rs < 657) {
return 2;
}
else if (Rs >= 657) {
return 1;
}
}
}
else if (TSC >= 3) {
if (MVDiff < 19) {
if (Cs < 391) {
if (SC < 502) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (SC >= 502) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 432) {
return 1;
}
else if (Rs >= 432) {
return 4;
}
}
}
}
else if (Cs >= 391) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
}
else if (MVDiff >= 19) {
if (ssDCval < 68) {
if (negBalance < 47) {
if (SC < 809) {
if (negBalance < 27) {
return 1;
}
else if (negBalance >= 27) {
return 4;
}
}
else if (SC >= 809) {
return 1;
}
}
else if (negBalance >= 47) {
return 8;
}
}
else if (ssDCval >= 68) {
return 8;
}
}
}
}
}
}
else if (TSC >= 4) {
if (MVDiff < 3) {
if (Rs < 522) {
return 8;
}
else if (Rs >= 522) {
return 4;
}
}
else if (MVDiff >= 3) {
if (tcor < 98) {
if (RsCsDiff < 9) {
return 1;
}
else if (RsCsDiff >= 9) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
return 8;
}
}
}
else if (tcor >= 98) {
if (TSC < 5) {
if (Rs < 572) {
return 1;
}
else if (Rs >= 572) {
return 4;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (Rs >= 717) {
if (mu_mv_mag_sq < 57) {
if (tcor < 97) {
if (ssDCval < 55) {
if (Rs < 768) {
return 1;
}
else if (Rs >= 768) {
if (Cs < 321) {
return 1;
}
else if (Cs >= 321) {
return 2;
}
}
}
else if (ssDCval >= 55) {
if (tcor < 87) {
if (Rs < 917) {
return 1;
}
else if (Rs >= 917) {
return 4;
}
}
else if (tcor >= 87) {
return 1;
}
}
}
else if (tcor >= 97) {
if (Cs < 162) {
return 1;
}
else if (Cs >= 162) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (SC < 1594) {
return 4;
}
else if (SC >= 1594) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (tcor < 99) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (tcor >= 99) {
if (Cs < 261) {
return 1;
}
else if (Cs >= 261) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (ssDCval < 68) {
if (Cs < 265) {
if (ssDCval < 48) {
if (negBalance < 32) {
return 1;
}
else if (negBalance >= 32) {
return 8;
}
}
else if (ssDCval >= 48) {
if (posBalance < 39) {
if (negBalance < 23) {
return 8;
}
else if (negBalance >= 23) {
if (Rs < 564) {
return 4;
}
else if (Rs >= 564) {
return 1;
}
}
}
else if (posBalance >= 39) {
if (diffMVdiffVal < -6) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
if (negBalance < 32) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 414) {
return 4;
}
else if (Rs >= 414) {
if (Rs < 533) {
return 1;
}
else if (Rs >= 533) {
return 2;
}
}
}
}
else if (negBalance >= 32) {
if (Cs < 206) {
return 4;
}
else if (Cs >= 206) {
return 1;
}
}
}
}
}
}
else if (Cs >= 265) {
if (Rs < 447) {
if (MVDiff < 4) {
if (SC < 878) {
if (diffTSC < -1) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 1;
}
}
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (SC >= 878) {
if (Rs < 415) {
return 1;
}
else if (Rs >= 415) {
if (Rs < 437) {
return 8;
}
else if (Rs >= 437) {
return 2;
}
}
}
}
else if (MVDiff >= 4) {
if (AFD < 3) {
if (Rs < 282) {
return 1;
}
else if (Rs >= 282) {
if (Rs < 415) {
return 2;
}
else if (Rs >= 415) {
return 8;
}
}
}
else if (AFD >= 3) {
if (MVDiff < 9) {
if (Rs < 320) {
return 1;
}
else if (Rs >= 320) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
return 2;
}
}
}
else if (MVDiff >= 9) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
return 8;
}
}
}
}
}
else if (Rs >= 447) {
if (Rs < 939) {
if (SC < 1278) {
if (tcor < 98) {
if (negBalance < 39) {
if (TSCindex < 2) {
if (Rs < 525) {
return 8;
}
else if (Rs >= 525) {
return 4;
}
}
else if (TSCindex >= 2) {
if (diffRsCsdiff < -2) {
if (Rs < 520) {
return 1;
}
else if (Rs >= 520) {
if (Rs < 612) {
return 8;
}
else if (Rs >= 612) {
return 2;
}
}
}
else if (diffRsCsdiff >= -2) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (MVDiff < 6) {
if (ssDCval < 58) {
if (Cs < 312) {
return 4;
}
else if (Cs >= 312) {
return 1;
}
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
}
}
}
else if (negBalance >= 39) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (tcor >= 98) {
if (refDCval < 57) {
if (Cs < 442) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
else if (Cs >= 442) {
if (Rs < 588) {
return 8;
}
else if (Rs >= 588) {
return 4;
}
}
}
else if (refDCval >= 57) {
if (Rs < 678) {
if (RsCsDiff < 1) {
if (SC < 887) {
if (Rs < 458) {
return 2;
}
else if (Rs >= 458) {
return 8;
}
}
else if (SC >= 887) {
if (SC < 1024) {
return 1;
}
else if (SC >= 1024) {
if (Rs < 626) {
return 8;
}
else if (Rs >= 626) {
return 1;
}
}
}
}
else if (RsCsDiff >= 1) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (Rs >= 678) {
if (Rs < 803) {
return 1;
}
else if (Rs >= 803) {
return 2;
}
}
}
}
}
else if (SC >= 1278) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
}
else if (Rs >= 939) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
}
}
}
else if (ssDCval >= 68) {
if (posBalance < 50) {
if (Cs < 379) {
if (TSC < 4) {
if (Cs < 276) {
if (Rs < 865) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (Rs >= 865) {
return 1;
}
}
else if (Cs >= 276) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
if (Rs < 580) {
return 1;
}
else if (Rs >= 580) {
if (Rs < 718) {
return 2;
}
else if (Rs >= 718) {
return 8;
}
}
}
}
}
else if (TSC >= 4) {
return 2;
}
}
else if (Cs >= 379) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (negBalance < 22) {
if (MVDiff < 3) {
if (Rs < 503) {
return 8;
}
else if (Rs >= 503) {
return 4;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (negBalance >= 22) {
if (MVDiff < 3) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 304) {
return 8;
}
else if (Rs >= 304) {
return 2;
}
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (ssDCval < 68) {
if (Rs < 479) {
if (Cs < 253) {
if (Rs < 332) {
return 4;
}
else if (Rs >= 332) {
return 8;
}
}
else if (Cs >= 253) {
if (negBalance < 21) {
if (Cs < 450) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 124) {
if (Cs < 281) {
return 4;
}
else if (Cs >= 281) {
return 1;
}
}
else if (mu_mv_mag_sq >= 124) {
if (Cs < 372) {
return 1;
}
else if (Cs >= 372) {
return 4;
}
}
}
else if (posBalance >= 49) {
return 4;
}
}
else if (Cs >= 450) {
if (Cs < 472) {
return 8;
}
else if (Cs >= 472) {
return 1;
}
}
}
else if (negBalance >= 21) {
if (diffMVdiffVal < 2) {
if (ssDCval < 66) {
if (diffMVdiffVal < -3) {
if (Rs < 320) {
return 1;
}
else if (Rs >= 320) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (ssDCval >= 66) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 381) {
if (Rs < 356) {
return 2;
}
else if (Rs >= 356) {
return 8;
}
}
else if (Rs >= 381) {
return 1;
}
}
}
}
}
else if (Rs >= 479) {
if (TSCindex < 3) {
if (Cs < 317) {
if (posBalance < 42) {
if (Rs < 789) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
if (Rs < 701) {
return 8;
}
else if (Rs >= 701) {
return 1;
}
}
}
else if (Rs >= 789) {
return 1;
}
}
else if (posBalance >= 42) {
if (Rs < 622) {
return 8;
}
else if (Rs >= 622) {
return 2;
}
}
}
else if (Cs >= 317) {
if (SC < 879) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (SC >= 879) {
if (posBalance < 38) {
if (mu_mv_mag_sq < 115) {
return 1;
}
else if (mu_mv_mag_sq >= 115) {
if (Cs < 402) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (Cs >= 402) {
return 8;
}
}
}
else if (posBalance >= 38) {
if (refDCval < 62) {
return 1;
}
else if (refDCval >= 62) {
if (negBalance < 17) {
return 8;
}
else if (negBalance >= 17) {
return 1;
}
}
}
}
}
}
else if (TSCindex >= 3) {
if (mcTcor < 97) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (mcTcor >= 97) {
if (SC < 1075) {
if (Cs < 504) {
return 1;
}
else if (Cs >= 504) {
return 2;
}
}
else if (SC >= 1075) {
if (ssDCval < 57) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 68) {
if (SC < 1001) {
if (Cs < 398) {
if (mcTcor < 99) {
if (Rs < 581) {
return 1;
}
else if (Rs >= 581) {
return 4;
}
}
else if (mcTcor >= 99) {
if (posBalance < 32) {
return 4;
}
else if (posBalance >= 32) {
if (SC < 999) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
else if (SC >= 999) {
return 4;
}
}
}
}
else if (Cs >= 398) {
if (posBalance < 52) {
if (TSC < 3) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
if (Rs < 433) {
return 2;
}
else if (Rs >= 433) {
return 1;
}
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (posBalance >= 52) {
return 2;
}
}
}
else if (SC >= 1001) {
if (MVDiff < 1) {
if (mu_mv_mag_sq < 119) {
if (Rs < 798) {
return 2;
}
else if (Rs >= 798) {
return 8;
}
}
else if (mu_mv_mag_sq >= 119) {
if (SC < 1069) {
if (ssDCval < 72) {
if (Rs < 673) {
return 1;
}
else if (Rs >= 673) {
return 4;
}
}
else if (ssDCval >= 72) {
return 4;
}
}
else if (SC >= 1069) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 72) {
if (negBalance < 19) {
if (Rs < 760) {
return 1;
}
else if (Rs >= 760) {
return 8;
}
}
else if (negBalance >= 19) {
if (Rs < 606) {
return 1;
}
else if (Rs >= 606) {
return 2;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 653) {
return 4;
}
else if (Rs >= 653) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1286) {
if (SC < 1603) {
if (RsCsDiff < 2) {
if (Cs < 222) {
return 8;
}
else if (Cs >= 222) {
return 1;
}
}
else if (RsCsDiff >= 2) {
return 1;
}
}
else if (SC >= 1603) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1751) {
if (mu_mv_mag_sq < 123) {
if (mu_mv_mag_sq < 56) {
if (negBalance < 29) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (AFD < 6) {
if (Cs < 304) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
else if (Cs >= 304) {
return 2;
}
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (negBalance >= 29) {
return 1;
}
}
else if (mu_mv_mag_sq >= 56) {
if (posBalance < 70) {
if (Cs < 409) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (negBalance < 36) {
if (mu_mv_mag_sq < 85) {
return 8;
}
else if (mu_mv_mag_sq >= 85) {
if (SC < 1626) {
return 8;
}
else if (SC >= 1626) {
return 1;
}
}
}
else if (negBalance >= 36) {
return 1;
}
}
}
else if (Cs >= 409) {
if (negBalance < 37) {
if (AFD < 1) {
if (Rs < 1426) {
return 2;
}
else if (Rs >= 1426) {
return 1;
}
}
else if (AFD >= 1) {
return 1;
}
}
else if (negBalance >= 37) {
return 8;
}
}
}
else if (posBalance >= 70) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (TSCindex < 2) {
if (ssDCval < 60) {
if (negBalance < 16) {
return 1;
}
else if (negBalance >= 16) {
return 8;
}
}
else if (ssDCval >= 60) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (TSCindex >= 2) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
return 8;
}
}
}
}
else if (Rs >= 1751) {
if (Cs < 434) {
if (Rs < 2386) {
if (MVDiff < 8) {
if (refDCval < 74) {
return 2;
}
else if (refDCval >= 74) {
return 8;
}
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (Rs >= 2386) {
return 8;
}
}
else if (Cs >= 434) {
if (posBalance < 46) {
if (SC < 2946) {
return 4;
}
else if (SC >= 2946) {
return 1;
}
}
else if (posBalance >= 46) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 62) {
if (Cs < 517) {
return 8;
}
else if (Cs >= 517) {
return 1;
}
}
else if (ssDCval >= 62) {
if (Rs < 1409) {
return 8;
}
else if (Rs >= 1409) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (Cs < 141) {
if (ssDCval < 108) {
if (tcor < 58) {
if (Rs < 327) {
if (Cs < 28) {
return 4;
}
else if (Cs >= 28) {
return 8;
}
}
else if (Rs >= 327) {
if (diffMVdiffVal < -6) {
return 2;
}
else if (diffMVdiffVal >= -6) {
if (mcTcor < 85) {
return 2;
}
else if (mcTcor >= 85) {
if (tcor < 49) {
return 8;
}
else if (tcor >= 49) {
if (mu_mv_mag_sq < 21) {
return 8;
}
else if (mu_mv_mag_sq >= 21) {
return 2;
}
}
}
}
}
}
else if (tcor >= 58) {
if (RsDiff < 2) {
if (negBalance < 38) {
if (Cs < 49) {
if (Rs < 518) {
return 1;
}
else if (Rs >= 518) {
return 8;
}
}
else if (Cs >= 49) {
if (posBalance < 59) {
if (ssDCval < 106) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 106) {
return 8;
}
}
else if (posBalance >= 59) {
if (Rs < 297) {
return 1;
}
else if (Rs >= 297) {
return 8;
}
}
}
}
else if (negBalance >= 38) {
if (Rs < 1460) {
if (Cs < 66) {
if (posBalance < 11) {
return 8;
}
else if (posBalance >= 11) {
return 1;
}
}
else if (Cs >= 66) {
return 8;
}
}
else if (Rs >= 1460) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
}
else if (ssDCval >= 108) {
if (tcor < 97) {
if (Rs < 357) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
return 2;
}
}
else if (Rs >= 357) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 2;
}
}
}
}
else if (tcor >= 97) {
if (diffAFD < -6) {
if (Cs < 4) {
return 8;
}
else if (Cs >= 4) {
return 1;
}
}
else if (diffAFD >= -6) {
return 8;
}
}
}
}
else if (Cs >= 141) {
if (SC < 856) {
if (mcTcor < 97) {
if (MVDiff < 22) {
if (Cs < 163) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Cs >= 163) {
if (AFD < 9) {
if (Rs < 325) {
if (TSC < 6) {
if (posBalance < 57) {
if (ssDCval < 123) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
if (Rs < 298) {
return 4;
}
else if (Rs >= 298) {
return 2;
}
}
}
else if (ssDCval >= 123) {
return 8;
}
}
else if (posBalance >= 57) {
if (negBalance < 33) {
return 2;
}
else if (negBalance >= 33) {
if (Cs < 361) {
if (MVDiff < 18) {
if (Cs < 281) {
return 8;
}
else if (Cs >= 281) {
return 2;
}
}
else if (MVDiff >= 18) {
return 8;
}
}
else if (Cs >= 361) {
return 8;
}
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (Rs >= 325) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (Rs < 562) {
if (Rs < 402) {
if (negBalance < 71) {
return 8;
}
else if (negBalance >= 71) {
if (Rs < 360) {
return 8;
}
else if (Rs >= 360) {
return 2;
}
}
}
else if (Rs >= 402) {
if (Rs < 407) {
return 1;
}
else if (Rs >= 407) {
if (diffRsCsdiff < -1) {
return 1;
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
}
else if (Rs >= 562) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
}
else if (AFD >= 9) {
if (TSC < 6) {
if (posBalance < 52) {
if (Rs < 294) {
return 1;
}
else if (Rs >= 294) {
return 2;
}
}
else if (posBalance >= 52) {
return 1;
}
}
else if (TSC >= 6) {
if (ssDCval < 109) {
return 8;
}
else if (ssDCval >= 109) {
if (Cs < 398) {
return 1;
}
else if (Cs >= 398) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 22) {
if (Cs < 174) {
if (mu_mv_mag_sq < 30) {
return 1;
}
else if (mu_mv_mag_sq >= 30) {
if (Cs < 145) {
return 2;
}
else if (Cs >= 145) {
return 8;
}
}
}
else if (Cs >= 174) {
if (Rs < 284) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (MVDiff < 37) {
return 2;
}
else if (MVDiff >= 37) {
return 1;
}
}
}
else if (Rs >= 284) {
if (ssDCval < 81) {
if (Rs < 382) {
if (Rs < 315) {
return 4;
}
else if (Rs >= 315) {
return 8;
}
}
else if (Rs >= 382) {
if (Rs < 512) {
return 4;
}
else if (Rs >= 512) {
return 1;
}
}
}
else if (ssDCval >= 81) {
if (AFD < 12) {
if (diffMVdiffVal < 34) {
if (Rs < 381) {
if (Cs < 326) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
else if (Cs >= 326) {
if (Rs < 347) {
return 8;
}
else if (Rs >= 347) {
if (Rs < 365) {
return 2;
}
else if (Rs >= 365) {
return 8;
}
}
}
}
else if (Rs >= 381) {
return 2;
}
}
else if (diffMVdiffVal >= 34) {
if (Rs < 489) {
return 4;
}
else if (Rs >= 489) {
return 8;
}
}
}
else if (AFD >= 12) {
if (MVDiff < 29) {
if (MVDiff < 27) {
if (Rs < 307) {
return 2;
}
else if (Rs >= 307) {
return 1;
}
}
else if (MVDiff >= 27) {
return 4;
}
}
else if (MVDiff >= 29) {
if (diffMVdiffVal < -93) {
return 4;
}
else if (diffMVdiffVal >= -93) {
if (Rs < 456) {
return 2;
}
else if (Rs >= 456) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 80) {
if (Cs < 314) {
if (AFD < 4) {
if (Rs < 302) {
return 2;
}
else if (Rs >= 302) {
return 8;
}
}
else if (AFD >= 4) {
if (Rs < 515) {
return 2;
}
else if (Rs >= 515) {
return 4;
}
}
}
else if (Cs >= 314) {
if (Rs < 430) {
return 8;
}
else if (Rs >= 430) {
return 1;
}
}
}
else if (ssDCval >= 80) {
if (mu_mv_mag_sq < 104) {
if (negBalance < 30) {
if (ssDCval < 97) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 4;
}
}
else if (ssDCval >= 97) {
if (ssDCval < 127) {
return 1;
}
else if (ssDCval >= 127) {
if (Cs < 252) {
if (Rs < 398) {
return 2;
}
else if (Rs >= 398) {
return 8;
}
}
else if (Cs >= 252) {
if (Cs < 398) {
if (Cs < 307) {
return 1;
}
else if (Cs >= 307) {
return 4;
}
}
else if (Cs >= 398) {
return 1;
}
}
}
}
}
else if (negBalance >= 30) {
if (mu_mv_mag_sq < 10) {
if (MVDiff < 30) {
if (Rs < 464) {
if (posBalance < 63) {
if (Rs < 279) {
if (Rs < 273) {
return 4;
}
else if (Rs >= 273) {
return 2;
}
}
else if (Rs >= 279) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (posBalance >= 63) {
return 2;
}
}
else if (Rs >= 464) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
if (Rs < 468) {
return 4;
}
else if (Rs >= 468) {
return 1;
}
}
}
}
else if (MVDiff >= 30) {
if (Cs < 286) {
return 4;
}
else if (Cs >= 286) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (negBalance < 39) {
return 2;
}
else if (negBalance >= 39) {
if (posBalance < 52) {
if (RsDiff < 1) {
if (Rs < 351) {
return 4;
}
else if (Rs >= 351) {
return 8;
}
}
else if (RsDiff >= 1) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
if (Rs < 393) {
return 8;
}
else if (Rs >= 393) {
return 4;
}
}
}
}
else if (posBalance >= 52) {
if (diffMVdiffVal < 1) {
if (posBalance < 65) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (Cs < 280) {
return 1;
}
else if (Cs >= 280) {
return 4;
}
}
}
else if (posBalance >= 65) {
if (Rs < 382) {
return 2;
}
else if (Rs >= 382) {
return 8;
}
}
}
else if (diffMVdiffVal >= 1) {
if (gchDC < 1) {
if (Rs < 490) {
return 4;
}
else if (Rs >= 490) {
return 8;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
if (Rs < 313) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
return 2;
}
}
else if (Rs >= 313) {
if (ssDCval < 94) {
if (SC < 506) {
if (Cs < 189) {
return 2;
}
else if (Cs >= 189) {
return 1;
}
}
else if (SC >= 506) {
return 2;
}
}
else if (ssDCval >= 94) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 856) {
if (SC < 1873) {
if (ssDCval < 81) {
if (tcor < 98) {
if (mu_mv_mag_sq < 82) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (Cs < 267) {
if (Rs < 888) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 1;
}
}
else if (Rs >= 888) {
return 1;
}
}
else if (Cs >= 267) {
if (Rs < 1436) {
if (refDCval < 73) {
return 4;
}
else if (refDCval >= 73) {
return 1;
}
}
else if (Rs >= 1436) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (negBalance < 32) {
if (Rs < 480) {
return 8;
}
else if (Rs >= 480) {
if (diffMVdiffVal < -2) {
if (Rs < 764) {
return 4;
}
else if (Rs >= 764) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
else if (tcor >= 98) {
if (negBalance < 28) {
if (Cs < 218) {
return 8;
}
else if (Cs >= 218) {
if (ssDCval < 80) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (ssDCval >= 80) {
if (Rs < 730) {
return 1;
}
else if (Rs >= 730) {
return 8;
}
}
}
}
else if (negBalance >= 28) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Cs < 381) {
if (Rs < 1284) {
return 8;
}
else if (Rs >= 1284) {
return 1;
}
}
else if (Cs >= 381) {
return 1;
}
}
}
}
}
else if (ssDCval >= 81) {
if (refDCval < 94) {
if (MVDiff < 4) {
if (Cs < 249) {
if (Rs < 909) {
return 1;
}
else if (Rs >= 909) {
return 2;
}
}
else if (Cs >= 249) {
if (tcor < 89) {
if (negBalance < 44) {
if (Cs < 462) {
return 1;
}
else if (Cs >= 462) {
return 2;
}
}
else if (negBalance >= 44) {
return 1;
}
}
else if (tcor >= 89) {
if (SC < 1188) {
if (Rs < 451) {
return 8;
}
else if (Rs >= 451) {
if (Cs < 525) {
return 1;
}
else if (Cs >= 525) {
return 8;
}
}
}
else if (SC >= 1188) {
if (refDCval < 85) {
if (posBalance < 23) {
return 8;
}
else if (posBalance >= 23) {
if (refDCval < 82) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (refDCval >= 82) {
return 1;
}
}
}
else if (refDCval >= 85) {
return 8;
}
}
}
}
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 35) {
if (mcTcor < 85) {
return 2;
}
else if (mcTcor >= 85) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (diffMVdiffVal < -6) {
return 1;
}
else if (diffMVdiffVal >= -6) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (Rs < 1273) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 580) {
return 8;
}
else if (Rs >= 580) {
return 1;
}
}
}
else if (Rs >= 1273) {
if (Rs < 1373) {
return 1;
}
else if (Rs >= 1373) {
return 8;
}
}
}
}
}
else if (refDCval >= 94) {
if (ssDCval < 115) {
if (mu_mv_mag_sq < 84) {
if (CsDiff < 2) {
if (posBalance < 62) {
if (mu_mv_mag_sq < 37) {
return 8;
}
else if (mu_mv_mag_sq >= 37) {
if (Cs < 313) {
return 8;
}
else if (Cs >= 313) {
if (Cs < 476) {
return 1;
}
else if (Cs >= 476) {
return 8;
}
}
}
}
else if (posBalance >= 62) {
if (SC < 955) {
return 8;
}
else if (SC >= 955) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (mcTcor < 94) {
if (negBalance < 58) {
return 8;
}
else if (negBalance >= 58) {
if (TSC < 6) {
if (Rs < 557) {
return 2;
}
else if (Rs >= 557) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (refDCval < 99) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 559) {
return 1;
}
else if (Rs >= 559) {
return 4;
}
}
}
else if (refDCval >= 99) {
if (Cs < 482) {
if (Rs < 1126) {
return 1;
}
else if (Rs >= 1126) {
return 8;
}
}
else if (Cs >= 482) {
if (tcor < 87) {
return 1;
}
else if (tcor >= 87) {
if (Rs < 659) {
return 2;
}
else if (Rs >= 659) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (Cs < 452) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (Cs >= 452) {
if (Cs < 466) {
if (Cs < 463) {
return 8;
}
else if (Cs >= 463) {
return 4;
}
}
else if (Cs >= 466) {
return 8;
}
}
}
}
else if (ssDCval >= 115) {
if (mcTcor < 98) {
if (MVDiff < 20) {
if (diffMVdiffVal < -4) {
if (diffRsCsdiff < 4) {
if (Rs < 411) {
return 8;
}
else if (Rs >= 411) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
return 4;
}
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 971) {
return 2;
}
else if (Rs >= 971) {
return 1;
}
}
}
else if (diffMVdiffVal >= -4) {
if (ssDCval < 161) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (Rs < 914) {
return 8;
}
else if (Rs >= 914) {
if (Rs < 1031) {
return 1;
}
else if (Rs >= 1031) {
return 8;
}
}
}
}
else if (ssDCval >= 161) {
return 1;
}
}
}
else if (MVDiff >= 20) {
if (tcor < 92) {
return 2;
}
else if (tcor >= 92) {
if (Cs < 402) {
return 4;
}
else if (Cs >= 402) {
return 1;
}
}
}
}
else if (mcTcor >= 98) {
if (refDCval < 119) {
if (Cs < 432) {
return 1;
}
else if (Cs >= 432) {
if (Rs < 411) {
if (AFD < 4) {
if (Rs < 372) {
return 4;
}
else if (Rs >= 372) {
return 1;
}
}
else if (AFD >= 4) {
return 2;
}
}
else if (Rs >= 411) {
return 2;
}
}
}
else if (refDCval >= 119) {
if (ssDCval < 158) {
if (TSC < 5) {
if (mu_mv_mag_sq < 68) {
if (refDCval < 120) {
if (Cs < 496) {
return 1;
}
else if (Cs >= 496) {
return 2;
}
}
else if (refDCval >= 120) {
return 1;
}
}
else if (mu_mv_mag_sq >= 68) {
if (Rs < 565) {
return 8;
}
else if (Rs >= 565) {
return 1;
}
}
}
else if (TSC >= 5) {
if (Cs < 505) {
if (SC < 1177) {
return 4;
}
else if (SC >= 1177) {
return 8;
}
}
else if (Cs >= 505) {
return 2;
}
}
}
else if (ssDCval >= 158) {
if (RsCsDiff < 3) {
return 4;
}
else if (RsCsDiff >= 3) {
return 1;
}
}
}
}
}
}
}
}
else if (SC >= 1873) {
if (SC < 2500) {
if (Cs < 500) {
if (ssDCval < 100) {
if (mu_mv_mag_sq < 82) {
if (Cs < 306) {
return 1;
}
else if (Cs >= 306) {
if (AFD < 5) {
if (mu_mv_mag_sq < 80) {
return 8;
}
else if (mu_mv_mag_sq >= 80) {
if (Rs < 1844) {
return 1;
}
else if (Rs >= 1844) {
return 8;
}
}
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (Rs < 2046) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (Rs >= 2046) {
return 2;
}
}
}
else if (ssDCval >= 100) {
if (ssDCval < 112) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
else if (ssDCval >= 112) {
if (Cs < 421) {
if (Rs < 1813) {
return 4;
}
else if (Rs >= 1813) {
return 1;
}
}
else if (Cs >= 421) {
return 2;
}
}
}
}
else if (Cs >= 500) {
if (AFD < 4) {
if (ssDCval < 106) {
return 4;
}
else if (ssDCval >= 106) {
return 1;
}
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (SC >= 2500) {
if (posBalance < 53) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (Cs < 359) {
return 1;
}
else if (Cs >= 359) {
if (Rs < 2073) {
return 1;
}
else if (Rs >= 2073) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (mu_mv_mag_sq < 78) {
if (tcor < 74) {
if (Cs < 333) {
if (SC < 2160) {
if (ssDCval < 83) {
if (Rs < 651) {
if (SC < 590) {
return 1;
}
else if (SC >= 590) {
if (TSC < 5) {
if (Rs < 352) {
return 8;
}
else if (Rs >= 352) {
return 2;
}
}
else if (TSC >= 5) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
if (negBalance < 54) {
if (diffRsCsdiff < -5) {
if (Cs < 245) {
return 4;
}
else if (Cs >= 245) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
return 4;
}
}
else if (negBalance >= 54) {
if (mu_mv_mag_sq < 14) {
if (Cs < 319) {
return 8;
}
else if (Cs >= 319) {
return 2;
}
}
else if (mu_mv_mag_sq >= 14) {
if (SC < 722) {
if (Rs < 380) {
return 4;
}
else if (Rs >= 380) {
if (Rs < 454) {
return 8;
}
else if (Rs >= 454) {
return 2;
}
}
}
else if (SC >= 722) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 651) {
if (RsDiff < 11) {
if (tcor < 23) {
return 8;
}
else if (tcor >= 23) {
if (MVDiff < 21) {
if (Rs < 1174) {
return 1;
}
else if (Rs >= 1174) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
return 1;
}
}
}
else if (MVDiff >= 21) {
if (Cs < 320) {
return 1;
}
else if (Cs >= 320) {
if (Rs < 738) {
return 4;
}
else if (Rs >= 738) {
return 1;
}
}
}
}
}
else if (RsDiff >= 11) {
if (diffRsCsdiff < -126) {
if (CsDiff < 2) {
if (Rs < 1239) {
return 1;
}
else if (Rs >= 1239) {
return 4;
}
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (diffRsCsdiff >= -126) {
if (negBalance < 55) {
if (mu_mv_mag_sq < 30) {
return 1;
}
else if (mu_mv_mag_sq >= 30) {
if (negBalance < 37) {
return 1;
}
else if (negBalance >= 37) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 2;
}
}
}
}
else if (negBalance >= 55) {
if (Rs < 1947) {
return 2;
}
else if (Rs >= 1947) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 83) {
if (diffAFD < 5) {
if (ssDCval < 94) {
if (RsDiff < 7) {
if (AFD < 28) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 1517) {
return 2;
}
else if (Rs >= 1517) {
return 1;
}
}
}
else if (AFD >= 28) {
return 1;
}
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (ssDCval >= 94) {
if (mu_mv_mag_sq < 33) {
if (mcTcor < 90) {
if (mcTcor < 80) {
if (Rs < 677) {
return 1;
}
else if (Rs >= 677) {
if (Rs < 777) {
return 8;
}
else if (Rs >= 777) {
if (Rs < 956) {
return 4;
}
else if (Rs >= 956) {
return 1;
}
}
}
}
else if (mcTcor >= 80) {
if (ssDCval < 113) {
if (Rs < 486) {
return 8;
}
else if (Rs >= 486) {
if (MVDiff < 22) {
if (Rs < 1369) {
return 8;
}
else if (Rs >= 1369) {
return 2;
}
}
else if (MVDiff >= 22) {
return 2;
}
}
}
else if (ssDCval >= 113) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (Rs < 616) {
return 1;
}
else if (Rs >= 616) {
return 2;
}
}
}
}
}
else if (mcTcor >= 90) {
if (AFD < 19) {
if (ssDCval < 108) {
if (posBalance < 62) {
if (Rs < 492) {
return 8;
}
else if (Rs >= 492) {
return 2;
}
}
else if (posBalance >= 62) {
return 8;
}
}
else if (ssDCval >= 108) {
if (Rs < 808) {
return 1;
}
else if (Rs >= 808) {
return 4;
}
}
}
else if (AFD >= 19) {
if (MVDiff < 167) {
if (mu_mv_mag_sq < 6) {
return 2;
}
else if (mu_mv_mag_sq >= 6) {
return 1;
}
}
else if (MVDiff >= 167) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (tcor < 31) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 4;
}
}
else if (tcor >= 31) {
if (negBalance < 77) {
return 2;
}
else if (negBalance >= 77) {
return 1;
}
}
}
}
}
else if (diffAFD >= 5) {
if (MVDiff < 35) {
return 2;
}
else if (MVDiff >= 35) {
if (SC < 1824) {
return 1;
}
else if (SC >= 1824) {
if (Rs < 1638) {
return 8;
}
else if (Rs >= 1638) {
return 1;
}
}
}
}
}
}
else if (SC >= 2160) {
if (ssDCval < 109) {
if (gchDC < 9) {
if (MVDiff < 24) {
return 2;
}
else if (MVDiff >= 24) {
if (posBalance < 52) {
if (Cs < 206) {
return 1;
}
else if (Cs >= 206) {
return 2;
}
}
else if (posBalance >= 52) {
return 2;
}
}
}
else if (gchDC >= 9) {
if (Rs < 2380) {
return 1;
}
else if (Rs >= 2380) {
return 8;
}
}
}
else if (ssDCval >= 109) {
if (Cs < 292) {
return 1;
}
else if (Cs >= 292) {
return 8;
}
}
}
}
else if (Cs >= 333) {
if (AFD < 24) {
if (negBalance < 48) {
if (negBalance < 40) {
if (RsCsDiff < 27) {
if (Cs < 446) {
if (SC < 1047) {
return 8;
}
else if (SC >= 1047) {
return 1;
}
}
else if (Cs >= 446) {
if (Rs < 616) {
return 4;
}
else if (Rs >= 616) {
return 2;
}
}
}
else if (RsCsDiff >= 27) {
if (Rs < 2099) {
if (diffMVdiffVal < -44) {
return 8;
}
else if (diffMVdiffVal >= -44) {
return 1;
}
}
else if (Rs >= 2099) {
return 2;
}
}
}
else if (negBalance >= 40) {
if (SC < 972) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (SC >= 972) {
if (mcTcor < 86) {
if (posBalance < 58) {
if (tcor < 65) {
return 2;
}
else if (tcor >= 65) {
if (Rs < 866) {
return 8;
}
else if (Rs >= 866) {
return 1;
}
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (mcTcor >= 86) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
if (diffMVdiffVal < -15) {
return 1;
}
else if (diffMVdiffVal >= -15) {
if (Rs < 2441) {
if (ssDCval < 39) {
return 1;
}
else if (ssDCval >= 39) {
return 2;
}
}
else if (Rs >= 2441) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 48) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Cs < 387) {
if (RsCsDiff < 37) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (ssDCval < 100) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
return 8;
}
}
}
else if (ssDCval >= 100) {
if (Rs < 542) {
return 2;
}
else if (Rs >= 542) {
return 1;
}
}
}
}
else if (RsCsDiff >= 37) {
if (mu_mv_mag_sq < 15) {
if (ssDCval < 60) {
if (Cs < 373) {
return 2;
}
else if (Cs >= 373) {
return 1;
}
}
else if (ssDCval >= 60) {
if (posBalance < 48) {
if (Rs < 2102) {
return 1;
}
else if (Rs >= 2102) {
return 2;
}
}
else if (posBalance >= 48) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (CsDiff < 4) {
if (Rs < 1399) {
return 2;
}
else if (Rs >= 1399) {
return 4;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
else if (Cs >= 387) {
if (tcor < 39) {
if (ssDCval < 123) {
if (Cs < 490) {
return 8;
}
else if (Cs >= 490) {
if (Cs < 518) {
return 4;
}
else if (Cs >= 518) {
return 8;
}
}
}
else if (ssDCval >= 123) {
return 1;
}
}
else if (tcor >= 39) {
if (mcTcor < 81) {
if (Rs < 560) {
if (Cs < 472) {
if (Rs < 484) {
return 4;
}
else if (Rs >= 484) {
return 8;
}
}
else if (Cs >= 472) {
return 2;
}
}
else if (Rs >= 560) {
if (ssDCval < 126) {
if (diffMVdiffVal < -5) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (ssDCval >= 126) {
return 4;
}
}
}
else if (mcTcor >= 81) {
if (gchDC < 1) {
if (Rs < 1038) {
return 8;
}
else if (Rs >= 1038) {
if (MVDiff < 32) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (negBalance < 57) {
return 2;
}
else if (negBalance >= 57) {
return 4;
}
}
}
else if (MVDiff >= 32) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (Cs < 435) {
if (mu_mv_mag_sq < 23) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (mu_mv_mag_sq >= 23) {
if (Rs < 633) {
return 8;
}
else if (Rs >= 633) {
return 1;
}
}
}
else if (Cs >= 435) {
if (Rs < 675) {
if (AFD < 20) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (AFD >= 20) {
return 4;
}
}
else if (Rs >= 675) {
if (Rs < 1158) {
if (Rs < 940) {
return 8;
}
else if (Rs >= 940) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
else if (Rs >= 1158) {
if (diffAFD < -1) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (diffAFD >= -1) {
if (negBalance < 54) {
return 2;
}
else if (negBalance >= 54) {
if (Rs < 1215) {
return 2;
}
else if (Rs >= 1215) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (AFD >= 24) {
if (MVDiff < 43) {
if (SC < 2308) {
if (posBalance < 63) {
return 4;
}
else if (posBalance >= 63) {
if (Rs < 1321) {
return 2;
}
else if (Rs >= 1321) {
return 1;
}
}
}
else if (SC >= 2308) {
if (RsCsDiff < 269) {
return 1;
}
else if (RsCsDiff >= 269) {
return 2;
}
}
}
else if (MVDiff >= 43) {
if (tcor < 71) {
if (refDCval < 107) {
if (mu_mv_mag_sq < 43) {
if (Rs < 359) {
if (CsDiff < 6) {
if (Rs < 321) {
return 8;
}
else if (Rs >= 321) {
if (Rs < 341) {
return 1;
}
else if (Rs >= 341) {
return 8;
}
}
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (Rs >= 359) {
if (diffRsCsdiff < 23) {
if (CsDiff < 6) {
if (AFD < 26) {
if (Cs < 354) {
return 2;
}
else if (Cs >= 354) {
return 1;
}
}
else if (AFD >= 26) {
return 1;
}
}
else if (CsDiff >= 6) {
if (Rs < 496) {
return 2;
}
else if (Rs >= 496) {
return 4;
}
}
}
else if (diffRsCsdiff >= 23) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
if (Rs < 500) {
return 2;
}
else if (Rs >= 500) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (tcor < 30) {
if (Rs < 1095) {
return 8;
}
else if (Rs >= 1095) {
return 4;
}
}
else if (tcor >= 30) {
return 2;
}
}
}
}
else if (refDCval >= 107) {
if (mcTcor < 84) {
if (refDCval < 161) {
if (AFD < 34) {
return 4;
}
else if (AFD >= 34) {
return 8;
}
}
else if (refDCval >= 161) {
return 1;
}
}
else if (mcTcor >= 84) {
return 1;
}
}
}
else if (tcor >= 71) {
if (AFD < 30) {
return 8;
}
else if (AFD >= 30) {
return 1;
}
}
}
}
}
}
else if (tcor >= 74) {
if (RsCsDiff < 115) {
if (Cs < 309) {
if (refDCval < 139) {
if (ssDCval < 85) {
if (Rs < 562) {
if (Cs < 133) {
return 1;
}
else if (Cs >= 133) {
if (diffRsCsdiff < 6) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (MVDiff < 113) {
return 1;
}
else if (MVDiff >= 113) {
if (Rs < 470) {
return 2;
}
else if (Rs >= 470) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 6) {
return 8;
}
}
}
else if (Rs >= 562) {
if (gchDC < 3) {
if (TSC < 8) {
if (mu_mv_mag_sq < 38) {
return 1;
}
else if (mu_mv_mag_sq >= 38) {
if (diffTSC < -1) {
if (Cs < 222) {
return 1;
}
else if (Cs >= 222) {
if (Rs < 1092) {
return 8;
}
else if (Rs >= 1092) {
return 2;
}
}
}
else if (diffTSC >= -1) {
return 1;
}
}
}
else if (TSC >= 8) {
return 2;
}
}
else if (gchDC >= 3) {
if (Cs < 266) {
if (MVDiff < 182) {
if (MVDiff < 17) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
if (Rs < 2089) {
return 1;
}
else if (Rs >= 2089) {
return 2;
}
}
}
else if (MVDiff >= 17) {
if (diffRsCsdiff < -21) {
if (Cs < 123) {
return 1;
}
else if (Cs >= 123) {
return 2;
}
}
else if (diffRsCsdiff >= -21) {
return 1;
}
}
}
else if (MVDiff >= 182) {
if (Rs < 727) {
return 8;
}
else if (Rs >= 727) {
if (Rs < 924) {
return 2;
}
else if (Rs >= 924) {
return 1;
}
}
}
}
else if (Cs >= 266) {
if (Cs < 280) {
return 4;
}
else if (Cs >= 280) {
if (posBalance < 85) {
return 1;
}
else if (posBalance >= 85) {
if (Rs < 2194) {
return 2;
}
else if (Rs >= 2194) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (MVDiff < 21) {
if (ssDCval < 117) {
if (ssDCval < 95) {
if (diffAFD < -1) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (diffAFD >= -1) {
if (posBalance < 57) {
if (negBalance < 65) {
return 1;
}
else if (negBalance >= 65) {
return 2;
}
}
else if (posBalance >= 57) {
if (Rs < 1111) {
return 1;
}
else if (Rs >= 1111) {
return 2;
}
}
}
}
else if (ssDCval >= 95) {
if (Cs < 272) {
return 2;
}
else if (Cs >= 272) {
if (Rs < 1878) {
return 8;
}
else if (Rs >= 1878) {
return 4;
}
}
}
}
else if (ssDCval >= 117) {
return 1;
}
}
else if (MVDiff >= 21) {
if (mu_mv_mag_sq < 25) {
if (Rs < 659) {
if (mu_mv_mag_sq < 10) {
return 1;
}
else if (mu_mv_mag_sq >= 10) {
if (Rs < 374) {
return 1;
}
else if (Rs >= 374) {
return 8;
}
}
}
else if (Rs >= 659) {
if (tcor < 77) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
if (Rs < 1552) {
return 1;
}
else if (Rs >= 1552) {
if (Rs < 2035) {
return 8;
}
else if (Rs >= 2035) {
if (Rs < 2068) {
return 1;
}
else if (Rs >= 2068) {
return 8;
}
}
}
}
}
else if (tcor >= 77) {
if (TSC < 7) {
if (mu_mv_mag_sq < 3) {
if (refDCval < 99) {
if (Rs < 1130) {
return 2;
}
else if (Rs >= 1130) {
return 1;
}
}
else if (refDCval >= 99) {
if (MVDiff < 52) {
return 2;
}
else if (MVDiff >= 52) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 965) {
if (Rs < 753) {
return 1;
}
else if (Rs >= 753) {
return 2;
}
}
else if (Rs >= 965) {
if (diffRsCsdiff < -97) {
if (Rs < 1883) {
return 1;
}
else if (Rs >= 1883) {
return 2;
}
}
else if (diffRsCsdiff >= -97) {
return 1;
}
}
}
}
else if (TSC >= 7) {
if (Rs < 681) {
return 2;
}
else if (Rs >= 681) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (AFD < 11) {
if (Rs < 2010) {
return 2;
}
else if (Rs >= 2010) {
return 1;
}
}
else if (AFD >= 11) {
if (Rs < 296) {
return 2;
}
else if (Rs >= 296) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 38) {
return 1;
}
else if (mu_mv_mag_sq >= 38) {
if (posBalance < 89) {
return 2;
}
else if (posBalance >= 89) {
return 1;
}
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 139) {
if (gchDC < 2) {
if (refDCval < 145) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 1;
}
}
else if (refDCval >= 145) {
if (negBalance < 49) {
if (Cs < 196) {
return 4;
}
else if (Cs >= 196) {
if (Rs < 873) {
return 2;
}
else if (Rs >= 873) {
return 1;
}
}
}
else if (negBalance >= 49) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 178) {
if (gchDC < 11) {
return 1;
}
else if (gchDC >= 11) {
return 2;
}
}
else if (ssDCval >= 178) {
return 2;
}
}
}
}
else if (Cs >= 309) {
if (Rs < 826) {
if (ssDCval < 119) {
if (SC < 1227) {
if (diffMVdiffVal < -18) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
return 2;
}
}
else if (diffMVdiffVal >= -18) {
if (diffAFD < 1) {
if (refDCval < 86) {
if (diffMVdiffVal < -4) {
if (Cs < 517) {
return 1;
}
else if (Cs >= 517) {
return 4;
}
}
else if (diffMVdiffVal >= -4) {
if (posBalance < 58) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Rs < 688) {
return 1;
}
else if (Rs >= 688) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 1;
}
}
}
}
else if (refDCval >= 86) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Rs < 766) {
return 8;
}
else if (Rs >= 766) {
return 1;
}
}
}
}
else if (diffAFD >= 1) {
if (RsCsDiff < 33) {
if (TSC < 6) {
if (Rs < 402) {
return 2;
}
else if (Rs >= 402) {
return 4;
}
}
else if (TSC >= 6) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
if (Cs < 449) {
return 8;
}
else if (Cs >= 449) {
return 4;
}
}
}
}
else if (RsCsDiff >= 33) {
return 1;
}
}
}
}
else if (SC >= 1227) {
if (diffRsCsdiff < 6) {
if (Cs < 443) {
return 2;
}
else if (Cs >= 443) {
return 8;
}
}
else if (diffRsCsdiff >= 6) {
return 2;
}
}
}
else if (ssDCval >= 119) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (diffRsCsdiff < -1) {
if (mu_mv_mag_sq < 45) {
return 8;
}
else if (mu_mv_mag_sq >= 45) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
}
}
else if (Rs >= 826) {
if (refDCval < 71) {
if (RsDiff < 8) {
if (mcTcor < 98) {
if (Cs < 456) {
if (Cs < 409) {
if (Rs < 2302) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
else if (Rs >= 2302) {
return 4;
}
}
else if (Cs >= 409) {
if (Rs < 1831) {
if (Rs < 1069) {
if (Cs < 438) {
if (Rs < 883) {
return 1;
}
else if (Rs >= 883) {
return 2;
}
}
else if (Cs >= 438) {
return 4;
}
}
else if (Rs >= 1069) {
if (Cs < 453) {
return 1;
}
else if (Cs >= 453) {
if (Rs < 1342) {
return 2;
}
else if (Rs >= 1342) {
return 1;
}
}
}
}
else if (Rs >= 1831) {
return 4;
}
}
}
else if (Cs >= 456) {
if (Cs < 524) {
if (posBalance < 56) {
if (Rs < 1094) {
if (ssDCval < 54) {
return 2;
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (Rs >= 1094) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (AFD < 11) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (AFD >= 11) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
}
}
}
else if (posBalance >= 56) {
if (Cs < 497) {
return 1;
}
else if (Cs >= 497) {
return 4;
}
}
}
else if (Cs >= 524) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (posBalance < 48) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (Rs < 2287) {
return 1;
}
else if (Rs >= 2287) {
return 4;
}
}
}
}
else if (RsDiff >= 8) {
if (mu_mv_mag_sq < 10) {
if (Rs < 1271) {
return 8;
}
else if (Rs >= 1271) {
return 2;
}
}
else if (mu_mv_mag_sq >= 10) {
return 1;
}
}
}
else if (refDCval >= 71) {
if (mcTcor < 95) {
if (refDCval < 91) {
if (negBalance < 62) {
if (Rs < 1299) {
if (tcor < 82) {
if (CsDiff < 4) {
if (Cs < 415) {
return 1;
}
else if (Cs >= 415) {
return 4;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (tcor >= 82) {
return 1;
}
}
else if (Rs >= 1299) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (Cs < 490) {
if (negBalance < 56) {
return 1;
}
else if (negBalance >= 56) {
return 2;
}
}
else if (Cs >= 490) {
return 2;
}
}
}
}
else if (negBalance >= 62) {
if (Rs < 1444) {
return 2;
}
else if (Rs >= 1444) {
return 8;
}
}
}
else if (refDCval >= 91) {
if (diffRsCsdiff < 8) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (posBalance < 64) {
if (diffTSC < 1) {
if (gchDC < 3) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (MVDiff < 57) {
return 8;
}
else if (MVDiff >= 57) {
if (Cs < 430) {
return 1;
}
else if (Cs >= 430) {
if (Cs < 475) {
return 8;
}
else if (Cs >= 475) {
return 1;
}
}
}
}
}
else if (gchDC >= 3) {
return 1;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (posBalance >= 64) {
if (Rs < 1516) {
return 2;
}
else if (Rs >= 1516) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 8) {
if (diffRsCsdiff < 36) {
if (Rs < 1644) {
return 2;
}
else if (Rs >= 1644) {
return 8;
}
}
else if (diffRsCsdiff >= 36) {
if (Cs < 357) {
return 8;
}
else if (Cs >= 357) {
if (Rs < 1338) {
return 4;
}
else if (Rs >= 1338) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (AFD < 8) {
if (ssDCval < 97) {
if (Rs < 1308) {
return 4;
}
else if (Rs >= 1308) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (ssDCval >= 97) {
if (Cs < 423) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (Cs >= 423) {
if (Rs < 1855) {
return 2;
}
else if (Rs >= 1855) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 2;
}
}
}
}
}
else if (AFD >= 8) {
if (TSC < 4) {
if (ssDCval < 94) {
return 1;
}
else if (ssDCval >= 94) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (TSC >= 4) {
if (tcor < 90) {
if (gchDC < 3) {
if (ssDCval < 84) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
if (Rs < 1995) {
return 1;
}
else if (Rs >= 1995) {
if (Cs < 420) {
return 2;
}
else if (Cs >= 420) {
return 4;
}
}
}
}
else if (ssDCval >= 84) {
if (Cs < 495) {
if (Rs < 2122) {
if (RsDiff < 8) {
if (MVDiff < 23) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (MVDiff >= 23) {
return 2;
}
}
else if (RsDiff >= 8) {
return 1;
}
}
else if (Rs >= 2122) {
if (negBalance < 50) {
if (Cs < 422) {
return 1;
}
else if (Cs >= 422) {
return 2;
}
}
else if (negBalance >= 50) {
if (Cs < 352) {
return 1;
}
else if (Cs >= 352) {
if (refDCval < 121) {
return 8;
}
else if (refDCval >= 121) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 8;
}
}
}
}
}
}
else if (Cs >= 495) {
if (MVDiff < 17) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (MVDiff >= 17) {
if (diffTSC < -2) {
if (Rs < 1703) {
return 8;
}
else if (Rs >= 1703) {
return 1;
}
}
else if (diffTSC >= -2) {
return 2;
}
}
}
}
}
else if (gchDC >= 3) {
if (mu_mv_mag_sq < 43) {
if (mu_mv_mag_sq < 35) {
return 1;
}
else if (mu_mv_mag_sq >= 35) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 43) {
return 2;
}
}
}
else if (tcor >= 90) {
if (gchDC < 1) {
if (mcTcor < 96) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
if (Rs < 1336) {
return 4;
}
else if (Rs >= 1336) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1296) {
return 4;
}
else if (Rs >= 1296) {
if (diffMVdiffVal < 10) {
return 1;
}
else if (diffMVdiffVal >= 10) {
if (Rs < 2051) {
return 1;
}
else if (Rs >= 2051) {
return 4;
}
}
}
}
}
else if (gchDC >= 1) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
if (posBalance < 78) {
return 1;
}
else if (posBalance >= 78) {
if (Cs < 496) {
return 1;
}
else if (Cs >= 496) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 115) {
if (refDCval < 111) {
if (MVDiff < 35) {
if (negBalance < 37) {
if (Rs < 2428) {
return 1;
}
else if (Rs >= 2428) {
return 2;
}
}
else if (negBalance >= 37) {
if (SC < 2542) {
if (negBalance < 51) {
if (tcor < 78) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 2;
}
}
else if (tcor >= 78) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1657) {
return 2;
}
else if (Rs >= 1657) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffAFD < -3) {
if (Rs < 1871) {
return 1;
}
else if (Rs >= 1871) {
return 2;
}
}
else if (diffAFD >= -3) {
return 2;
}
}
}
}
else if (negBalance >= 51) {
if (SC < 2528) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 2;
}
}
else if (SC >= 2528) {
return 1;
}
}
}
else if (SC >= 2542) {
if (RsCsDiff < 128) {
if (Cs < 360) {
return 1;
}
else if (Cs >= 360) {
return 2;
}
}
else if (RsCsDiff >= 128) {
return 2;
}
}
}
}
else if (MVDiff >= 35) {
if (posBalance < 51) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 2;
}
}
else if (posBalance >= 51) {
if (SC < 2592) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
else if (SC >= 2592) {
if (RsDiff < 13) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (RsDiff >= 13) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 111) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (MVDiff < 361) {
if (mcTcor < 95) {
if (refDCval < 120) {
return 8;
}
else if (refDCval >= 120) {
if (RsCsDiff < 128) {
return 8;
}
else if (RsCsDiff >= 128) {
if (posBalance < 77) {
return 1;
}
else if (posBalance >= 77) {
return 2;
}
}
}
}
else if (mcTcor >= 95) {
if (MVDiff < 37) {
if (RsCsDiff < 246) {
if (RsCsDiff < 128) {
if (AFD < 19) {
if (Cs < 440) {
return 2;
}
else if (Cs >= 440) {
return 1;
}
}
else if (AFD >= 19) {
return 1;
}
}
else if (RsCsDiff >= 128) {
return 1;
}
}
else if (RsCsDiff >= 246) {
if (Cs < 445) {
return 1;
}
else if (Cs >= 445) {
return 2;
}
}
}
else if (MVDiff >= 37) {
if (diffRsCsdiff < -251) {
return 8;
}
else if (diffRsCsdiff >= -251) {
if (ssDCval < 119) {
if (ssDCval < 118) {
return 1;
}
else if (ssDCval >= 118) {
if (Rs < 1795) {
return 1;
}
else if (Rs >= 1795) {
return 8;
}
}
}
else if (ssDCval >= 119) {
return 1;
}
}
}
}
}
else if (MVDiff >= 361) {
if (Rs < 2121) {
return 2;
}
else if (Rs >= 2121) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 1530) {
if (tcor < 42) {
if (gchDC < 2) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
if (Rs < 853) {
return 1;
}
else if (Rs >= 853) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
return 8;
}
}
}
else if (tcor >= 42) {
if (Cs < 199) {
if (AFD < 35) {
if (TSC < 5) {
if (MVDiff < 197) {
return 1;
}
else if (MVDiff >= 197) {
return 2;
}
}
else if (TSC >= 5) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
if (AFD < 27) {
return 1;
}
else if (AFD >= 27) {
return 2;
}
}
}
}
else if (AFD >= 35) {
return 1;
}
}
else if (Cs >= 199) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (mu_mv_mag_sq < 119) {
if (refDCval < 62) {
if (RsDiff < 6) {
if (Rs < 938) {
return 2;
}
else if (Rs >= 938) {
return 4;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (refDCval >= 62) {
if (Rs < 1302) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (Rs >= 1302) {
if (Rs < 1467) {
return 2;
}
else if (Rs >= 1467) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (ssDCval < 114) {
return 1;
}
else if (ssDCval >= 114) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1530) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (mcTcor < 95) {
return 2;
}
else if (mcTcor >= 95) {
if (MVDiff < 26) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (MVDiff >= 26) {
if (mu_mv_mag_sq < 134) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Cs < 500) {
return 2;
}
else if (Cs >= 500) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 134) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 534) {
if (RsDiff < 8) {
if (mcTcor < 96) {
if (AFD < 14) {
if (posBalance < 46) {
if (mu_mv_mag_sq < 109) {
if (MVDiff < 8) {
if (ssDCval < 48) {
if (mu_mv_mag_sq < 22) {
if (Rs < 1661) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (Rs >= 1661) {
return 4;
}
}
else if (mu_mv_mag_sq >= 22) {
if (negBalance < 15) {
return 1;
}
else if (negBalance >= 15) {
if (diffMVdiffVal < -4) {
if (RsCsDiff < 22) {
return 4;
}
else if (RsCsDiff >= 22) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 8;
}
}
}
else if (diffMVdiffVal >= -4) {
if (AFD < 8) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (AFD >= 8) {
if (Rs < 1483) {
if (Rs < 1041) {
return 1;
}
else if (Rs >= 1041) {
return 8;
}
}
else if (Rs >= 1483) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 48) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (ssDCval < 60) {
if (ssDCval < 56) {
if (TSC < 5) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (diffRsCsdiff < 3) {
if (Cs < 650) {
return 1;
}
else if (Cs >= 650) {
return 8;
}
}
else if (diffRsCsdiff >= 3) {
return 1;
}
}
}
else if (TSC >= 5) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
return 2;
}
}
}
else if (ssDCval >= 56) {
if (Rs < 827) {
return 4;
}
else if (Rs >= 827) {
if (negBalance < 27) {
return 1;
}
else if (negBalance >= 27) {
return 8;
}
}
}
}
else if (ssDCval >= 60) {
if (ssDCval < 125) {
if (negBalance < 53) {
if (Cs < 836) {
if (Rs < 869) {
if (tcor < 89) {
return 2;
}
else if (tcor >= 89) {
if (Cs < 628) {
return 4;
}
else if (Cs >= 628) {
return 2;
}
}
}
else if (Rs >= 869) {
if (MVDiff < 4) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
if (Rs < 2215) {
return 1;
}
else if (Rs >= 2215) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (Rs < 933) {
return 4;
}
else if (Rs >= 933) {
if (Rs < 978) {
return 2;
}
else if (Rs >= 978) {
if (Rs < 1183) {
return 1;
}
else if (Rs >= 1183) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 836) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
if (mu_mv_mag_sq < 97) {
if (diffAFD < -2) {
if (Rs < 1843) {
return 8;
}
else if (Rs >= 1843) {
return 1;
}
}
else if (diffAFD >= -2) {
if (gchDC < 2) {
if (negBalance < 39) {
if (ssDCval < 101) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (Rs < 1381) {
return 2;
}
else if (Rs >= 1381) {
return 4;
}
}
}
else if (ssDCval >= 101) {
return 4;
}
}
else if (negBalance >= 39) {
return 4;
}
}
else if (gchDC >= 2) {
if (Rs < 853) {
return 2;
}
else if (Rs >= 853) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (Rs < 1855) {
if (Rs < 789) {
return 4;
}
else if (Rs >= 789) {
return 8;
}
}
else if (Rs >= 1855) {
return 2;
}
}
}
}
}
else if (negBalance >= 53) {
if (mu_mv_mag_sq < 34) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (mu_mv_mag_sq >= 34) {
return 8;
}
}
}
else if (ssDCval >= 125) {
if (Rs < 1799) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Cs < 1013) {
return 2;
}
else if (Cs >= 1013) {
if (Rs < 1157) {
return 1;
}
else if (Rs >= 1157) {
return 8;
}
}
}
}
else if (Rs >= 1799) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 8) {
if (refDCval < 83) {
if (tcor < 62) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 525) {
return 1;
}
else if (Rs >= 525) {
return 4;
}
}
}
else if (tcor >= 62) {
if (AFD < 5) {
if (mu_mv_mag_sq < 61) {
if (Rs < 784) {
return 4;
}
else if (Rs >= 784) {
return 8;
}
}
else if (mu_mv_mag_sq >= 61) {
if (mu_mv_mag_sq < 85) {
return 4;
}
else if (mu_mv_mag_sq >= 85) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (Cs < 742) {
if (Rs < 921) {
return 1;
}
else if (Rs >= 921) {
return 8;
}
}
else if (Cs >= 742) {
return 4;
}
}
}
}
}
else if (AFD >= 5) {
if (negBalance < 36) {
if (AFD < 8) {
if (posBalance < 43) {
if (ssDCval < 27) {
return 4;
}
else if (ssDCval >= 27) {
if (negBalance < 29) {
if (diffMVdiffVal < 5) {
return 1;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 1471) {
return 1;
}
else if (Rs >= 1471) {
return 8;
}
}
}
else if (negBalance >= 29) {
if (Rs < 1422) {
if (tcor < 93) {
if (diffRsCsdiff < -8) {
return 4;
}
else if (diffRsCsdiff >= -8) {
return 8;
}
}
else if (tcor >= 93) {
return 4;
}
}
else if (Rs >= 1422) {
if (Rs < 1756) {
return 1;
}
else if (Rs >= 1756) {
return 8;
}
}
}
}
}
else if (posBalance >= 43) {
if (TSC < 4) {
if (Cs < 725) {
return 8;
}
else if (Cs >= 725) {
return 1;
}
}
else if (TSC >= 4) {
if (RsCsDiff < 11) {
return 8;
}
else if (RsCsDiff >= 11) {
return 2;
}
}
}
}
else if (AFD >= 8) {
if (diffAFD < 2) {
if (posBalance < 45) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
if (Rs < 694) {
return 1;
}
else if (Rs >= 694) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 2;
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 1475) {
return 2;
}
else if (Rs >= 1475) {
return 4;
}
}
}
else if (diffAFD >= 2) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
else if (negBalance >= 36) {
if (Cs < 759) {
if (SC < 1776) {
if (diffMVdiffVal < -40) {
if (MVDiff < 65) {
return 2;
}
else if (MVDiff >= 65) {
if (Rs < 597) {
return 4;
}
else if (Rs >= 597) {
return 1;
}
}
}
else if (diffMVdiffVal >= -40) {
if (RsDiff < 4) {
if (Rs < 646) {
return 1;
}
else if (Rs >= 646) {
if (Cs < 564) {
if (Cs < 548) {
return 8;
}
else if (Cs >= 548) {
return 1;
}
}
else if (Cs >= 564) {
return 8;
}
}
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (SC >= 1776) {
if (tcor < 85) {
if (Rs < 1866) {
if (Rs < 1325) {
return 1;
}
else if (Rs >= 1325) {
return 2;
}
}
else if (Rs >= 1866) {
return 4;
}
}
else if (tcor >= 85) {
if (RsCsDiff < 17) {
return 8;
}
else if (RsCsDiff >= 17) {
return 1;
}
}
}
}
else if (Cs >= 759) {
if (ssDCval < 58) {
if (Rs < 802) {
if (AFD < 8) {
if (Rs < 563) {
return 8;
}
else if (Rs >= 563) {
return 4;
}
}
else if (AFD >= 8) {
return 2;
}
}
else if (Rs >= 802) {
if (Cs < 831) {
if (Rs < 1266) {
return 8;
}
else if (Rs >= 1266) {
return 4;
}
}
else if (Cs >= 831) {
return 8;
}
}
}
else if (ssDCval >= 58) {
if (RsCsDiff < 60) {
if (diffTSC < -1) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (diffTSC >= -1) {
if (tcor < 93) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (tcor >= 93) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
else if (RsCsDiff >= 60) {
if (Rs < 2336) {
return 8;
}
else if (Rs >= 2336) {
if (Rs < 2466) {
return 1;
}
else if (Rs >= 2466) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 83) {
if (ssDCval < 154) {
if (negBalance < 43) {
if (Rs < 710) {
return 4;
}
else if (Rs >= 710) {
return 8;
}
}
else if (negBalance >= 43) {
if (negBalance < 64) {
if (negBalance < 52) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 2;
}
}
else if (negBalance >= 52) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Cs < 1049) {
if (Cs < 751) {
return 8;
}
else if (Cs >= 751) {
return 4;
}
}
else if (Cs >= 1049) {
return 8;
}
}
}
}
else if (negBalance >= 64) {
if (tcor < 80) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 2;
}
}
else if (tcor >= 80) {
return 8;
}
}
}
}
else if (ssDCval >= 154) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (SC < 3001) {
if (Rs < 583) {
if (Rs < 380) {
return 2;
}
else if (Rs >= 380) {
if (Rs < 463) {
return 4;
}
else if (Rs >= 463) {
return 1;
}
}
}
else if (Rs >= 583) {
if (mcTcor < 82) {
if (ssDCval < 126) {
if (Cs < 552) {
return 8;
}
else if (Cs >= 552) {
return 1;
}
}
else if (ssDCval >= 126) {
if (MVDiff < 29) {
return 8;
}
else if (MVDiff >= 29) {
return 4;
}
}
}
else if (mcTcor >= 82) {
if (Rs < 1202) {
if (Cs < 927) {
if (posBalance < 7) {
return 4;
}
else if (posBalance >= 7) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 880) {
return 1;
}
else if (Rs >= 880) {
return 8;
}
}
}
}
else if (Cs >= 927) {
if (negBalance < 12) {
if (Rs < 1189) {
return 8;
}
else if (Rs >= 1189) {
return 1;
}
}
else if (negBalance >= 12) {
if (Cs < 1188) {
return 1;
}
else if (Cs >= 1188) {
return 4;
}
}
}
}
else if (Rs >= 1202) {
if (ssDCval < 50) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
return 4;
}
}
else if (ssDCval >= 50) {
if (diffMVdiffVal < -21) {
return 4;
}
else if (diffMVdiffVal >= -21) {
if (MVDiff < 28) {
return 8;
}
else if (MVDiff >= 28) {
if (Rs < 1522) {
return 4;
}
else if (Rs >= 1522) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 3001) {
if (ssDCval < 46) {
return 4;
}
else if (ssDCval >= 46) {
if (SC < 3043) {
if (Rs < 1891) {
return 4;
}
else if (Rs >= 1891) {
return 4;
}
}
else if (SC >= 3043) {
return 8;
}
}
}
}
}
else if (posBalance >= 46) {
if (Rs < 407) {
if (ssDCval < 97) {
if (tcor < 86) {
if (mcTcor < 86) {
if (Cs < 714) {
if (Rs < 295) {
return 1;
}
else if (Rs >= 295) {
return 2;
}
}
else if (Cs >= 714) {
return 4;
}
}
else if (mcTcor >= 86) {
if (RsCsDiff < 94) {
return 1;
}
else if (RsCsDiff >= 94) {
return 4;
}
}
}
else if (tcor >= 86) {
if (CsDiff < 3) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 325) {
return 1;
}
else if (Rs >= 325) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
}
else if (CsDiff >= 3) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
else if (ssDCval >= 97) {
if (Cs < 718) {
if (Rs < 286) {
return 8;
}
else if (Rs >= 286) {
return 1;
}
}
else if (Cs >= 718) {
return 4;
}
}
}
else if (Rs >= 407) {
if (mu_mv_mag_sq < 32) {
if (negBalance < 45) {
if (ssDCval < 58) {
if (Rs < 882) {
if (refDCval < 52) {
if (posBalance < 56) {
if (AFD < 9) {
if (refDCval < 37) {
return 8;
}
else if (refDCval >= 37) {
if (Cs < 666) {
if (RsCsDiff < 11) {
return 8;
}
else if (RsCsDiff >= 11) {
return 2;
}
}
else if (Cs >= 666) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 4;
}
}
}
}
else if (AFD >= 9) {
if (SC < 1489) {
if (Rs < 640) {
return 2;
}
else if (Rs >= 640) {
return 4;
}
}
else if (SC >= 1489) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
else if (posBalance >= 56) {
if (ssDCval < 42) {
if (Cs < 646) {
if (Cs < 582) {
return 2;
}
else if (Cs >= 582) {
return 8;
}
}
else if (Cs >= 646) {
if (Rs < 468) {
return 1;
}
else if (Rs >= 468) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (Rs < 685) {
return 2;
}
else if (Rs >= 685) {
return 8;
}
}
}
}
}
else if (ssDCval >= 42) {
if (gchDC < 2) {
if (Cs < 841) {
return 8;
}
else if (Cs >= 841) {
return 2;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
else if (refDCval >= 52) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
if (Rs < 672) {
return 4;
}
else if (Rs >= 672) {
return 8;
}
}
}
}
}
else if (Rs >= 882) {
if (Cs < 633) {
if (diffMVdiffVal < -11) {
if (Rs < 984) {
return 1;
}
else if (Rs >= 984) {
return 2;
}
}
else if (diffMVdiffVal >= -11) {
return 8;
}
}
else if (Cs >= 633) {
if (refDCval < 47) {
return 8;
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 3) {
if (diffMVdiffVal < 2) {
if (Rs < 1124) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (Rs >= 1124) {
return 8;
}
}
else if (diffMVdiffVal >= 2) {
if (SC < 2833) {
return 4;
}
else if (SC >= 2833) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 687) {
return 4;
}
else if (Cs >= 687) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 58) {
if (posBalance < 61) {
if (SC < 1391) {
return 2;
}
else if (SC >= 1391) {
if (Cs < 820) {
if (refDCval < 64) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Rs < 1457) {
if (Rs < 810) {
return 1;
}
else if (Rs >= 810) {
return 8;
}
}
else if (Rs >= 1457) {
return 2;
}
}
}
else if (refDCval >= 64) {
if (ssDCval < 99) {
if (diffMVdiffVal < -11) {
return 4;
}
else if (diffMVdiffVal >= -11) {
if (ssDCval < 65) {
if (Cs < 625) {
return 1;
}
else if (Cs >= 625) {
return 4;
}
}
else if (ssDCval >= 65) {
return 1;
}
}
}
else if (ssDCval >= 99) {
if (Rs < 1006) {
return 2;
}
else if (Rs >= 1006) {
return 8;
}
}
}
}
else if (Cs >= 820) {
if (negBalance < 37) {
if (refDCval < 137) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
return 1;
}
}
else if (refDCval >= 137) {
if (Rs < 1268) {
return 2;
}
else if (Rs >= 1268) {
return 8;
}
}
}
else if (negBalance >= 37) {
if (RsDiff < 6) {
if (posBalance < 54) {
if (RsCsDiff < 29) {
if (Rs < 1228) {
return 2;
}
else if (Rs >= 1228) {
return 8;
}
}
else if (RsCsDiff >= 29) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (ssDCval < 75) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (posBalance < 54) {
if (SC < 2629) {
return 4;
}
else if (SC >= 2629) {
return 8;
}
}
else if (posBalance >= 54) {
if (ssDCval < 102) {
return 2;
}
else if (ssDCval >= 102) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 61) {
if (Cs < 1215) {
if (diffAFD < -2) {
if (Cs < 618) {
return 1;
}
else if (Cs >= 618) {
if (SC < 1214) {
return 8;
}
else if (SC >= 1214) {
if (diffTSC < -3) {
return 2;
}
else if (diffTSC >= -3) {
if (Rs < 1501) {
return 4;
}
else if (Rs >= 1501) {
return 2;
}
}
}
}
}
else if (diffAFD >= -2) {
if (tcor < 80) {
if (Rs < 686) {
return 4;
}
else if (Rs >= 686) {
if (Cs < 1165) {
return 8;
}
else if (Cs >= 1165) {
return 1;
}
}
}
else if (tcor >= 80) {
if (ssDCval < 137) {
if (ssDCval < 98) {
if (TSC < 7) {
if (SC < 2030) {
if (Rs < 667) {
return 4;
}
else if (Rs >= 667) {
return 8;
}
}
else if (SC >= 2030) {
if (Rs < 2143) {
return 2;
}
else if (Rs >= 2143) {
return 8;
}
}
}
else if (TSC >= 7) {
if (Rs < 712) {
return 8;
}
else if (Rs >= 712) {
return 4;
}
}
}
else if (ssDCval >= 98) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (ssDCval >= 137) {
if (ssDCval < 143) {
return 4;
}
else if (ssDCval >= 143) {
if (Rs < 621) {
return 8;
}
else if (Rs >= 621) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1215) {
if (Rs < 962) {
return 4;
}
else if (Rs >= 962) {
return 1;
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 1001) {
if (AFD < 11) {
if (refDCval < 48) {
if (ssDCval < 45) {
if (mcTcor < 89) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 813) {
return 2;
}
else if (Rs >= 813) {
return 8;
}
}
}
else if (mcTcor >= 89) {
if (SC < 1378) {
return 4;
}
else if (SC >= 1378) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 4;
}
}
}
}
else if (ssDCval >= 45) {
if (negBalance < 54) {
if (posBalance < 59) {
if (Cs < 659) {
return 1;
}
else if (Cs >= 659) {
return 8;
}
}
else if (posBalance >= 59) {
if (Rs < 684) {
return 8;
}
else if (Rs >= 684) {
if (Rs < 977) {
return 1;
}
else if (Rs >= 977) {
return 8;
}
}
}
}
else if (negBalance >= 54) {
if (Cs < 656) {
return 1;
}
else if (Cs >= 656) {
return 8;
}
}
}
}
else if (refDCval >= 48) {
if (RsDiff < 4) {
if (MVDiff < 5) {
if (RsCsDiff < 11) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 668) {
return 8;
}
else if (Rs >= 668) {
return 1;
}
}
}
}
else if (RsCsDiff >= 11) {
if (Cs < 653) {
if (Rs < 733) {
return 2;
}
else if (Rs >= 733) {
return 8;
}
}
else if (Cs >= 653) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (MVDiff < 6) {
if (SC < 1537) {
return 4;
}
else if (SC >= 1537) {
if (Rs < 975) {
return 8;
}
else if (Rs >= 975) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (diffMVdiffVal < 8) {
return 8;
}
else if (diffMVdiffVal >= 8) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (mcTcor < 72) {
return 8;
}
else if (mcTcor >= 72) {
if (Cs < 688) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
if (Rs < 789) {
return 1;
}
else if (Rs >= 789) {
return 2;
}
}
}
else if (Cs >= 688) {
if (AFD < 9) {
if (Rs < 685) {
return 4;
}
else if (Rs >= 685) {
return 8;
}
}
else if (AFD >= 9) {
if (ssDCval < 80) {
return 4;
}
else if (ssDCval >= 80) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 11) {
if (Cs < 840) {
if (TSC < 6) {
if (ssDCval < 58) {
if (Cs < 554) {
return 4;
}
else if (Cs >= 554) {
return 2;
}
}
else if (ssDCval >= 58) {
if (Cs < 769) {
if (negBalance < 56) {
if (Cs < 558) {
return 1;
}
else if (Cs >= 558) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
return 4;
}
}
}
else if (negBalance >= 56) {
return 1;
}
}
else if (Cs >= 769) {
return 2;
}
}
}
else if (TSC >= 6) {
if (negBalance < 51) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 7) {
if (Cs < 740) {
return 4;
}
else if (Cs >= 740) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
if (mu_mv_mag_sq < 20) {
return 8;
}
else if (mu_mv_mag_sq >= 20) {
return 2;
}
}
}
else if (ssDCval >= 50) {
if (Rs < 585) {
return 1;
}
else if (Rs >= 585) {
if (Cs < 548) {
return 2;
}
else if (Cs >= 548) {
if (Cs < 754) {
return 4;
}
else if (Cs >= 754) {
if (Rs < 833) {
return 1;
}
else if (Rs >= 833) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 51) {
if (mcTcor < 86) {
if (Rs < 907) {
if (refDCval < 48) {
if (negBalance < 54) {
if (Rs < 655) {
return 4;
}
else if (Rs >= 655) {
return 8;
}
}
else if (negBalance >= 54) {
return 8;
}
}
else if (refDCval >= 48) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (Rs >= 907) {
return 8;
}
}
else if (mcTcor >= 86) {
if (RsCsDiff < 11) {
return 2;
}
else if (RsCsDiff >= 11) {
if (Cs < 747) {
if (ssDCval < 48) {
return 4;
}
else if (ssDCval >= 48) {
if (ssDCval < 50) {
if (SC < 1450) {
return 8;
}
else if (SC >= 1450) {
return 1;
}
}
else if (ssDCval >= 50) {
return 8;
}
}
}
else if (Cs >= 747) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 840) {
if (negBalance < 50) {
if (MVDiff < 15) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 1;
}
}
}
else if (MVDiff >= 15) {
if (diffAFD < -2) {
if (Rs < 711) {
return 8;
}
else if (Rs >= 711) {
return 4;
}
}
else if (diffAFD >= -2) {
return 2;
}
}
}
else if (negBalance >= 50) {
if (negBalance < 59) {
if (RsDiff < 6) {
if (diffMVdiffVal < -4) {
if (CsDiff < 5) {
if (Rs < 896) {
return 1;
}
else if (Rs >= 896) {
return 8;
}
}
else if (CsDiff >= 5) {
if (ssDCval < 78) {
return 4;
}
else if (ssDCval >= 78) {
return 8;
}
}
}
else if (diffMVdiffVal >= -4) {
if (TSC < 7) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (negBalance >= 59) {
if (Rs < 835) {
return 2;
}
else if (Rs >= 835) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1001) {
if (MVDiff < 5) {
if (ssDCval < 57) {
if (Cs < 1091) {
return 8;
}
else if (Cs >= 1091) {
return 4;
}
}
else if (ssDCval >= 57) {
if (diffMVdiffVal < -3) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
if (posBalance < 62) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (Rs < 1460) {
return 1;
}
else if (Rs >= 1460) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (Rs < 1709) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
else if (Rs >= 1709) {
return 1;
}
}
}
}
}
else if (MVDiff >= 5) {
if (tcor < 79) {
if (ssDCval < 121) {
if (refDCval < 62) {
if (ssDCval < 61) {
if (Rs < 1024) {
if (Cs < 947) {
return 8;
}
else if (Cs >= 947) {
if (Rs < 1018) {
return 4;
}
else if (Rs >= 1018) {
return 2;
}
}
}
else if (Rs >= 1024) {
return 8;
}
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (refDCval >= 62) {
if (ssDCval < 98) {
if (Rs < 1154) {
return 8;
}
else if (Rs >= 1154) {
if (Cs < 655) {
return 2;
}
else if (Cs >= 655) {
if (tcor < 76) {
if (Rs < 1166) {
return 2;
}
else if (Rs >= 1166) {
return 8;
}
}
else if (tcor >= 76) {
if (Cs < 910) {
return 2;
}
else if (Cs >= 910) {
if (Cs < 1094) {
return 8;
}
else if (Cs >= 1094) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 98) {
if (diffMVdiffVal < -6) {
if (Cs < 905) {
return 8;
}
else if (Cs >= 905) {
return 2;
}
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
}
}
else if (ssDCval >= 121) {
if (Cs < 923) {
if (Rs < 1493) {
return 1;
}
else if (Rs >= 1493) {
return 4;
}
}
else if (Cs >= 923) {
return 8;
}
}
}
else if (tcor >= 79) {
if (RsDiff < 5) {
if (ssDCval < 113) {
if (mcTcor < 93) {
if (posBalance < 51) {
if (Cs < 1059) {
return 8;
}
else if (Cs >= 1059) {
return 4;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (mcTcor >= 93) {
if (ssDCval < 62) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 1116) {
return 8;
}
else if (Rs >= 1116) {
return 4;
}
}
}
else if (ssDCval >= 62) {
if (posBalance < 59) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (posBalance >= 59) {
if (Cs < 690) {
return 2;
}
else if (Cs >= 690) {
return 8;
}
}
}
}
}
else if (ssDCval >= 113) {
if (diffRsCsdiff < -8) {
if (Rs < 1198) {
return 8;
}
else if (Rs >= 1198) {
return 4;
}
}
else if (diffRsCsdiff >= -8) {
return 4;
}
}
}
else if (RsDiff >= 5) {
if (negBalance < 47) {
if (SC < 2200) {
if (Rs < 1290) {
return 2;
}
else if (Rs >= 1290) {
if (Rs < 1345) {
return 8;
}
else if (Rs >= 1345) {
return 1;
}
}
}
else if (SC >= 2200) {
if (ssDCval < 92) {
if (SC < 3364) {
return 4;
}
else if (SC >= 3364) {
return 8;
}
}
else if (ssDCval >= 92) {
if (SC < 2883) {
return 8;
}
else if (SC >= 2883) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
else if (negBalance >= 47) {
if (AFD < 11) {
if (CsDiff < 3) {
if (Rs < 1458) {
return 8;
}
else if (Rs >= 1458) {
return 2;
}
}
else if (CsDiff >= 3) {
if (posBalance < 58) {
if (diffRsCsdiff < 29) {
if (negBalance < 50) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (negBalance >= 50) {
if (RsCsDiff < 38) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (RsCsDiff >= 38) {
return 8;
}
}
}
else if (diffRsCsdiff >= 29) {
return 4;
}
}
else if (posBalance >= 58) {
if (Cs < 607) {
return 2;
}
else if (Cs >= 607) {
if (MVDiff < 30) {
return 8;
}
else if (MVDiff >= 30) {
return 2;
}
}
}
}
}
else if (AFD >= 11) {
if (mcTcor < 94) {
if (MVDiff < 23) {
if (posBalance < 54) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 4;
}
}
else if (posBalance >= 54) {
if (RsDiff < 6) {
if (ssDCval < 109) {
return 8;
}
else if (ssDCval >= 109) {
return 4;
}
}
else if (RsDiff >= 6) {
if (ssDCval < 108) {
if (Rs < 1735) {
if (SC < 2215) {
if (diffAFD < -2) {
if (Rs < 1348) {
return 8;
}
else if (Rs >= 1348) {
return 4;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (SC >= 2215) {
if (tcor < 85) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (tcor >= 85) {
return 8;
}
}
}
else if (Rs >= 1735) {
return 1;
}
}
else if (ssDCval >= 108) {
return 8;
}
}
}
}
else if (MVDiff >= 23) {
if (Cs < 867) {
return 1;
}
else if (Cs >= 867) {
if (Cs < 1096) {
if (Rs < 1111) {
return 2;
}
else if (Rs >= 1111) {
return 4;
}
}
else if (Cs >= 1096) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 1299) {
if (diffRsCsdiff < -5) {
if (Cs < 1158) {
return 2;
}
else if (Cs >= 1158) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
return 1;
}
}
else if (Rs >= 1299) {
if (Rs < 2202) {
return 8;
}
else if (Rs >= 2202) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (refDCval < 161) {
if (Rs < 1576) {
if (ssDCval < 47) {
if (mu_mv_mag_sq < 101) {
if (MVDiff < 14) {
if (MVDiff < 6) {
if (Cs < 745) {
if (RsDiff < 4) {
if (diffRsCsdiff < -9) {
if (Rs < 918) {
return 4;
}
else if (Rs >= 918) {
return 2;
}
}
else if (diffRsCsdiff >= -9) {
return 8;
}
}
else if (RsDiff >= 4) {
if (TSC < 7) {
if (Rs < 1351) {
if (Rs < 1077) {
if (Rs < 1019) {
return 2;
}
else if (Rs >= 1019) {
return 8;
}
}
else if (Rs >= 1077) {
return 2;
}
}
else if (Rs >= 1351) {
return 8;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (Cs >= 745) {
return 8;
}
}
else if (MVDiff >= 6) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (Rs < 904) {
if (Cs < 753) {
return 8;
}
else if (Cs >= 753) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
return 8;
}
}
}
else if (Rs >= 904) {
return 8;
}
}
}
}
else if (MVDiff >= 14) {
if (ssDCval < 39) {
if (Rs < 534) {
return 2;
}
else if (Rs >= 534) {
return 8;
}
}
else if (ssDCval >= 39) {
if (SC < 1769) {
if (SC < 1531) {
if (SC < 1425) {
return 2;
}
else if (SC >= 1425) {
return 8;
}
}
else if (SC >= 1531) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (SC < 1639) {
return 1;
}
else if (SC >= 1639) {
return 8;
}
}
}
}
else if (SC >= 1769) {
if (Cs < 1179) {
return 8;
}
else if (Cs >= 1179) {
if (mcTcor < 77) {
if (Rs < 1400) {
return 8;
}
else if (Rs >= 1400) {
return 4;
}
}
else if (mcTcor >= 77) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (SC < 2199) {
if (posBalance < 58) {
if (Rs < 1066) {
if (Rs < 898) {
if (SC < 1553) {
return 8;
}
else if (SC >= 1553) {
if (SC < 1780) {
return 1;
}
else if (SC >= 1780) {
return 8;
}
}
}
else if (Rs >= 898) {
if (tcor < 81) {
return 1;
}
else if (tcor >= 81) {
return 8;
}
}
}
else if (Rs >= 1066) {
if (tcor < 89) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (tcor >= 89) {
return 8;
}
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (SC >= 2199) {
return 8;
}
}
}
else if (ssDCval >= 47) {
if (TSC < 7) {
if (RsDiff < 3) {
if (Rs < 458) {
if (MVDiff < 4) {
if (Rs < 417) {
return 4;
}
else if (Rs >= 417) {
if (TSC < 5) {
if (Cs < 905) {
return 8;
}
else if (Cs >= 905) {
return 1;
}
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (MVDiff >= 4) {
if (posBalance < 49) {
if (diffAFD < 1) {
if (Rs < 417) {
return 8;
}
else if (Rs >= 417) {
return 2;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (posBalance >= 49) {
if (Cs < 1019) {
return 8;
}
else if (Cs >= 1019) {
return 1;
}
}
}
}
else if (Rs >= 458) {
if (refDCval < 67) {
if (negBalance < 29) {
if (Rs < 1002) {
if (negBalance < 24) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
return 1;
}
}
else if (negBalance >= 24) {
if (TSC < 5) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (Rs < 842) {
return 4;
}
else if (Rs >= 842) {
return 8;
}
}
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (Rs >= 1002) {
return 8;
}
}
else if (negBalance >= 29) {
if (MVDiff < 3) {
if (Rs < 855) {
if (Cs < 669) {
return 8;
}
else if (Cs >= 669) {
if (SC < 1556) {
return 4;
}
else if (SC >= 1556) {
if (Rs < 835) {
return 8;
}
else if (Rs >= 835) {
return 4;
}
}
}
}
else if (Rs >= 855) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (Cs < 1080) {
if (ssDCval < 48) {
if (Rs < 966) {
return 1;
}
else if (Rs >= 966) {
if (Rs < 1200) {
return 8;
}
else if (Rs >= 1200) {
return 2;
}
}
}
else if (ssDCval >= 48) {
return 8;
}
}
else if (Cs >= 1080) {
if (Rs < 1308) {
if (Cs < 1209) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (Cs >= 1209) {
return 4;
}
}
else if (Rs >= 1308) {
return 8;
}
}
}
}
}
else if (refDCval >= 67) {
if (ssDCval < 70) {
if (Rs < 971) {
return 2;
}
else if (Rs >= 971) {
if (Rs < 1104) {
return 8;
}
else if (Rs >= 1104) {
return 1;
}
}
}
else if (ssDCval >= 70) {
if (TSC < 5) {
if (ssDCval < 98) {
if (negBalance < 30) {
return 2;
}
else if (negBalance >= 30) {
if (ssDCval < 91) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
return 8;
}
}
else if (ssDCval >= 91) {
if (Rs < 709) {
return 2;
}
else if (Rs >= 709) {
return 1;
}
}
}
}
else if (ssDCval >= 98) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
if (mu_mv_mag_sq < 39) {
return 4;
}
else if (mu_mv_mag_sq >= 39) {
if (Rs < 1421) {
if (diffMVdiffVal < -3) {
if (tcor < 94) {
if (Rs < 1041) {
return 4;
}
else if (Rs >= 1041) {
return 8;
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (Rs >= 1421) {
return 1;
}
}
}
}
}
else if (TSC >= 5) {
if (diffRsCsdiff < -6) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= -6) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 3) {
if (MVDiff < 2) {
if (diffTSC < 2) {
if (Cs < 1233) {
if (CsDiff < 2) {
if (Rs < 1240) {
return 1;
}
else if (Rs >= 1240) {
return 4;
}
}
else if (CsDiff >= 2) {
if (mcTcor < 95) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
return 4;
}
}
else if (mcTcor >= 95) {
if (negBalance < 42) {
if (Cs < 1037) {
return 8;
}
else if (Cs >= 1037) {
if (Cs < 1161) {
return 4;
}
else if (Cs >= 1161) {
if (Cs < 1196) {
return 8;
}
else if (Cs >= 1196) {
return 4;
}
}
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
}
else if (Cs >= 1233) {
if (mu_mv_mag_sq < 80) {
return 4;
}
else if (mu_mv_mag_sq >= 80) {
if (MVDiff < 1) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (Rs < 1106) {
return 4;
}
else if (Rs >= 1106) {
return 8;
}
}
}
}
}
else if (diffTSC >= 2) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 75) {
if (ssDCval < 52) {
if (negBalance < 34) {
if (Rs < 1076) {
return 4;
}
else if (Rs >= 1076) {
if (Cs < 1008) {
return 2;
}
else if (Cs >= 1008) {
return 8;
}
}
}
else if (negBalance >= 34) {
if (diffAFD < 2) {
if (Cs < 1170) {
if (Rs < 796) {
if (Rs < 564) {
return 8;
}
else if (Rs >= 564) {
return 4;
}
}
else if (Rs >= 796) {
return 8;
}
}
else if (Cs >= 1170) {
return 4;
}
}
else if (diffAFD >= 2) {
if (Rs < 992) {
return 4;
}
else if (Rs >= 992) {
if (Rs < 1453) {
return 8;
}
else if (Rs >= 1453) {
return 2;
}
}
}
}
}
else if (ssDCval >= 52) {
if (Cs < 932) {
if (SC < 1831) {
if (negBalance < 23) {
if (Rs < 779) {
return 2;
}
else if (Rs >= 779) {
return 4;
}
}
else if (negBalance >= 23) {
if (negBalance < 43) {
if (RsCsDiff < 15) {
if (tcor < 92) {
if (posBalance < 52) {
if (Rs < 671) {
return 4;
}
else if (Rs >= 671) {
return 1;
}
}
else if (posBalance >= 52) {
return 4;
}
}
else if (tcor >= 92) {
return 8;
}
}
else if (RsCsDiff >= 15) {
if (MVDiff < 14) {
if (Rs < 1089) {
if (Rs < 677) {
return 1;
}
else if (Rs >= 677) {
return 8;
}
}
else if (Rs >= 1089) {
return 1;
}
}
else if (MVDiff >= 14) {
return 1;
}
}
}
else if (negBalance >= 43) {
if (negBalance < 47) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (negBalance >= 47) {
if (Rs < 1102) {
return 8;
}
else if (Rs >= 1102) {
return 4;
}
}
}
}
}
else if (SC >= 1831) {
if (MVDiff < 13) {
if (Cs < 828) {
if (posBalance < 54) {
if (Rs < 1391) {
return 2;
}
else if (Rs >= 1391) {
if (Rs < 1508) {
return 8;
}
else if (Rs >= 1508) {
return 1;
}
}
}
else if (posBalance >= 54) {
return 8;
}
}
else if (Cs >= 828) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
if (mu_mv_mag_sq < 66) {
if (Cs < 918) {
return 1;
}
else if (Cs >= 918) {
return 2;
}
}
else if (mu_mv_mag_sq >= 66) {
return 8;
}
}
}
}
else if (MVDiff >= 13) {
if (ssDCval < 58) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
else if (ssDCval >= 58) {
if (AFD < 12) {
if (Cs < 896) {
return 8;
}
else if (Cs >= 896) {
if (Rs < 1222) {
return 8;
}
else if (Rs >= 1222) {
return 2;
}
}
}
else if (AFD >= 12) {
return 2;
}
}
}
}
}
else if (Cs >= 932) {
if (AFD < 8) {
if (MVDiff < 4) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 91) {
if (Rs < 1538) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 4;
}
}
else if (Rs >= 1538) {
return 8;
}
}
else if (mu_mv_mag_sq >= 91) {
if (RsCsDiff < 22) {
return 1;
}
else if (RsCsDiff >= 22) {
if (Rs < 1351) {
return 4;
}
else if (Rs >= 1351) {
return 8;
}
}
}
}
}
else if (AFD >= 8) {
if (Cs < 1091) {
return 8;
}
else if (Cs >= 1091) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (CsDiff < 4) {
if (Rs < 1204) {
return 4;
}
else if (Rs >= 1204) {
return 1;
}
}
else if (CsDiff >= 4) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (ssDCval < 123) {
if (Rs < 912) {
if (tcor < 87) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (tcor >= 87) {
if (Cs < 1187) {
return 8;
}
else if (Cs >= 1187) {
return 4;
}
}
}
else if (Rs >= 912) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 1253) {
return 8;
}
else if (Rs >= 1253) {
return 4;
}
}
}
}
else if (ssDCval >= 123) {
if (RsCsDiff < 28) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 59) {
if (refDCval < 135) {
if (Rs < 1312) {
return 1;
}
else if (Rs >= 1312) {
return 8;
}
}
else if (refDCval >= 135) {
return 4;
}
}
else if (mu_mv_mag_sq >= 59) {
return 8;
}
}
}
else if (RsCsDiff >= 28) {
return 8;
}
}
}
}
}
}
else if (TSC >= 7) {
if (Rs < 1263) {
if (ssDCval < 81) {
if (mu_mv_mag_sq < 99) {
if (Rs < 1132) {
if (gchDC < 4) {
if (SC < 1852) {
if (mcTcor < 90) {
if (mu_mv_mag_sq < 60) {
if (Cs < 767) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 672) {
return 1;
}
else if (Rs >= 672) {
return 4;
}
}
}
else if (Cs >= 767) {
return 4;
}
}
else if (mu_mv_mag_sq >= 60) {
return 4;
}
}
else if (mcTcor >= 90) {
if (RsCsDiff < 19) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (RsCsDiff >= 19) {
if (Cs < 598) {
return 4;
}
else if (Cs >= 598) {
return 1;
}
}
}
}
else if (SC >= 1852) {
if (SC < 2150) {
if (TSC < 10) {
if (RsDiff < 6) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 4;
}
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (TSC >= 10) {
return 4;
}
}
else if (SC >= 2150) {
if (Rs < 1028) {
return 8;
}
else if (Rs >= 1028) {
return 4;
}
}
}
}
else if (gchDC >= 4) {
return 8;
}
}
else if (Rs >= 1132) {
if (negBalance < 44) {
if (RsCsDiff < 46) {
return 4;
}
else if (RsCsDiff >= 46) {
return 8;
}
}
else if (negBalance >= 44) {
if (Cs < 1168) {
return 8;
}
else if (Cs >= 1168) {
if (Cs < 1179) {
if (Rs < 1199) {
return 4;
}
else if (Rs >= 1199) {
return 1;
}
}
else if (Cs >= 1179) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
return 8;
}
}
else if (ssDCval >= 81) {
if (ssDCval < 130) {
if (posBalance < 117) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
return 2;
}
}
}
else if (posBalance >= 117) {
return 1;
}
}
else if (ssDCval >= 130) {
if (SC < 2118) {
if (Rs < 850) {
return 2;
}
else if (Rs >= 850) {
return 4;
}
}
else if (SC >= 2118) {
return 8;
}
}
}
}
else if (Rs >= 1263) {
if (Cs < 1234) {
if (RsCsDiff < 43) {
return 8;
}
else if (RsCsDiff >= 43) {
if (tcor < 94) {
if (diffAFD < -2) {
if (Rs < 1525) {
if (Cs < 876) {
return 2;
}
else if (Cs >= 876) {
return 8;
}
}
else if (Rs >= 1525) {
return 1;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (tcor >= 94) {
if (Rs < 1464) {
return 2;
}
else if (Rs >= 1464) {
return 1;
}
}
}
}
else if (Cs >= 1234) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (Rs < 1481) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 1455) {
return 8;
}
else if (Rs >= 1455) {
return 1;
}
}
}
else if (Rs >= 1481) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1576) {
if (ssDCval < 99) {
if (Rs < 2317) {
if (RsCsDiff < 23) {
if (diffRsCsdiff < -8) {
if (mu_mv_mag_sq < 80) {
return 1;
}
else if (mu_mv_mag_sq >= 80) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (diffRsCsdiff >= -8) {
if (diffMVdiffVal < -18) {
return 1;
}
else if (diffMVdiffVal >= -18) {
return 8;
}
}
}
else if (RsCsDiff >= 23) {
if (ssDCval < 62) {
if (negBalance < 36) {
if (Cs < 1070) {
if (mcTcor < 94) {
return 2;
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (Cs >= 1070) {
if (MVDiff < 5) {
if (Cs < 1162) {
return 8;
}
else if (Cs >= 1162) {
return 8;
}
}
else if (MVDiff >= 5) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (Cs < 716) {
return 4;
}
else if (Cs >= 716) {
return 8;
}
}
}
else if (ssDCval >= 62) {
if (ssDCval < 70) {
if (TSC < 8) {
if (Cs < 1087) {
if (mu_mv_mag_sq < 44) {
if (Rs < 1874) {
return 4;
}
else if (Rs >= 1874) {
return 1;
}
}
else if (mu_mv_mag_sq >= 44) {
if (SC < 2481) {
return 8;
}
else if (SC >= 2481) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 2;
}
}
}
}
else if (Cs >= 1087) {
return 8;
}
}
else if (TSC >= 8) {
return 2;
}
}
else if (ssDCval >= 70) {
if (ssDCval < 86) {
if (Rs < 1749) {
return 8;
}
else if (Rs >= 1749) {
if (RsDiff < 5) {
if (ssDCval < 81) {
if (Cs < 1297) {
return 1;
}
else if (Cs >= 1297) {
return 8;
}
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (RsDiff >= 5) {
if (diffTSC < 1) {
if (refDCval < 72) {
if (Rs < 1786) {
return 8;
}
else if (Rs >= 1786) {
return 2;
}
}
else if (refDCval >= 72) {
return 8;
}
}
else if (diffTSC >= 1) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (SC < 3068) {
return 1;
}
else if (SC >= 3068) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (TSC < 8) {
if (diffMVdiffVal < 3) {
if (SC < 2801) {
if (RsCsDiff < 49) {
return 8;
}
else if (RsCsDiff >= 49) {
return 2;
}
}
else if (SC >= 2801) {
if (RsDiff < 7) {
return 2;
}
else if (RsDiff >= 7) {
if (Rs < 1815) {
return 2;
}
else if (Rs >= 1815) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2317) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 2593) {
if (mu_mv_mag_sq < 50) {
return 4;
}
else if (mu_mv_mag_sq >= 50) {
if (MVDiff < 24) {
if (RsDiff < 6) {
if (Rs < 2403) {
return 2;
}
else if (Rs >= 2403) {
return 4;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (MVDiff >= 24) {
return 8;
}
}
}
else if (Rs >= 2593) {
return 4;
}
}
}
}
else if (ssDCval >= 99) {
if (negBalance < 47) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (refDCval < 135) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 2100) {
return 2;
}
else if (Rs >= 2100) {
return 8;
}
}
}
else if (refDCval >= 135) {
if (Cs < 1274) {
if (mcTcor < 92) {
if (Rs < 1766) {
return 4;
}
else if (Rs >= 1766) {
return 1;
}
}
else if (mcTcor >= 92) {
return 8;
}
}
else if (Cs >= 1274) {
return 1;
}
}
}
}
else if (negBalance >= 47) {
if (MVDiff < 5) {
if (SC < 2972) {
return 1;
}
else if (SC >= 2972) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
}
else if (refDCval >= 161) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
}
else if (AFD >= 14) {
if (MVDiff < 12) {
if (TSC < 9) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (Rs < 822) {
if (AFD < 17) {
if (Rs < 785) {
return 8;
}
else if (Rs >= 785) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
return 1;
}
}
}
else if (AFD >= 17) {
if (Cs < 628) {
return 8;
}
else if (Cs >= 628) {
if (Rs < 645) {
return 2;
}
else if (Rs >= 645) {
return 8;
}
}
}
}
else if (Rs >= 822) {
if (AFD < 20) {
if (Rs < 1200) {
if (diffMVdiffVal < 1) {
if (tcor < 80) {
if (negBalance < 58) {
if (tcor < 48) {
return 4;
}
else if (tcor >= 48) {
if (Rs < 831) {
return 4;
}
else if (Rs >= 831) {
return 8;
}
}
}
else if (negBalance >= 58) {
return 4;
}
}
else if (tcor >= 80) {
if (RsCsDiff < 46) {
return 4;
}
else if (RsCsDiff >= 46) {
if (Rs < 893) {
return 8;
}
else if (Rs >= 893) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (Rs >= 1200) {
return 8;
}
}
else if (AFD >= 20) {
return 4;
}
}
}
}
else if (TSC >= 9) {
if (Rs < 1376) {
if (CsDiff < 5) {
if (negBalance < 60) {
return 4;
}
else if (negBalance >= 60) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 768) {
return 2;
}
else if (Rs >= 768) {
return 8;
}
}
}
}
else if (CsDiff >= 5) {
if (SC < 2373) {
if (diffRsCsdiff < 28) {
return 8;
}
else if (diffRsCsdiff >= 28) {
return 2;
}
}
else if (SC >= 2373) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
}
else if (Rs >= 1376) {
if (Rs < 1503) {
if (Cs < 1151) {
return 8;
}
else if (Cs >= 1151) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (Rs >= 1503) {
return 8;
}
}
}
}
else if (MVDiff >= 12) {
if (Rs < 1285) {
if (ssDCval < 70) {
if (TSC < 10) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
if (tcor < 49) {
if (AFD < 22) {
if (refDCval < 46) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
return 8;
}
}
else if (refDCval >= 46) {
if (Rs < 1140) {
if (diffRsCsdiff < 51) {
return 4;
}
else if (diffRsCsdiff >= 51) {
if (RsCsDiff < 108) {
return 8;
}
else if (RsCsDiff >= 108) {
return 4;
}
}
}
else if (Rs >= 1140) {
return 1;
}
}
}
else if (AFD >= 22) {
if (CsDiff < 9) {
return 2;
}
else if (CsDiff >= 9) {
return 8;
}
}
}
else if (tcor >= 49) {
if (RsDiff < 4) {
if (diffMVdiffVal < 2) {
if (AFD < 15) {
if (Rs < 498) {
return 1;
}
else if (Rs >= 498) {
return 4;
}
}
else if (AFD >= 15) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
if (MVDiff < 50) {
return 4;
}
else if (MVDiff >= 50) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
return 8;
}
}
}
}
else if (RsDiff >= 4) {
if (diffMVdiffVal < 16) {
if (mu_mv_mag_sq < 32) {
if (mu_mv_mag_sq < 7) {
if (TSC < 9) {
if (AFD < 16) {
if (ssDCval < 56) {
if (RsCsDiff < 94) {
return 4;
}
else if (RsCsDiff >= 94) {
return 8;
}
}
else if (ssDCval >= 56) {
if (refDCval < 60) {
return 1;
}
else if (refDCval >= 60) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 4;
}
}
}
}
}
else if (AFD >= 16) {
if (tcor < 56) {
return 8;
}
else if (tcor >= 56) {
if (diffRsCsdiff < -14) {
if (Rs < 699) {
return 4;
}
else if (Rs >= 699) {
return 1;
}
}
else if (diffRsCsdiff >= -14) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
if (Cs < 695) {
return 4;
}
else if (Cs >= 695) {
if (Rs < 1002) {
return 8;
}
else if (Rs >= 1002) {
return 1;
}
}
}
}
}
}
}
else if (TSC >= 9) {
if (negBalance < 55) {
if (Cs < 711) {
return 1;
}
else if (Cs >= 711) {
return 4;
}
}
else if (negBalance >= 55) {
if (Rs < 667) {
return 1;
}
else if (Rs >= 667) {
if (diffTSC < 2) {
return 2;
}
else if (diffTSC >= 2) {
if (Rs < 1016) {
return 1;
}
else if (Rs >= 1016) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Cs < 1119) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (SC < 1602) {
if (Cs < 671) {
return 8;
}
else if (Cs >= 671) {
return 4;
}
}
else if (SC >= 1602) {
if (Cs < 933) {
return 1;
}
else if (Cs >= 933) {
return 8;
}
}
}
}
else if (Cs >= 1119) {
if (Rs < 832) {
return 1;
}
else if (Rs >= 832) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (mcTcor < 85) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 2;
}
}
else if (mcTcor >= 85) {
if (Rs < 960) {
return 1;
}
else if (Rs >= 960) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 16) {
if (diffMVdiffVal < 32) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Rs < 685) {
return 8;
}
else if (Rs >= 685) {
return 1;
}
}
}
else if (diffMVdiffVal >= 32) {
if (TSC < 8) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
else if (TSC >= 8) {
if (Rs < 780) {
return 2;
}
else if (Rs >= 780) {
return 4;
}
}
}
}
}
}
}
}
else if (TSC >= 10) {
if (RsDiff < 7) {
if (diffRsCsdiff < 9) {
if (CsDiff < 6) {
if (Rs < 713) {
if (SC < 1289) {
return 8;
}
else if (SC >= 1289) {
return 4;
}
}
else if (Rs >= 713) {
return 8;
}
}
else if (CsDiff >= 6) {
if (diffRsCsdiff < -17) {
if (Rs < 1008) {
return 8;
}
else if (Rs >= 1008) {
return 4;
}
}
else if (diffRsCsdiff >= -17) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Rs < 1015) {
return 1;
}
else if (Rs >= 1015) {
return 4;
}
}
}
}
}
else if (diffRsCsdiff >= 9) {
if (mu_mv_mag_sq < 14) {
if (refDCval < 67) {
if (TSC < 11) {
if (refDCval < 48) {
return 8;
}
else if (refDCval >= 48) {
return 4;
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (refDCval >= 67) {
return 4;
}
}
else if (mu_mv_mag_sq >= 14) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
if (Rs < 363) {
return 1;
}
else if (Rs >= 363) {
return 2;
}
}
}
}
}
else if (RsDiff >= 7) {
if (refDCval < 50) {
return 4;
}
else if (refDCval >= 50) {
if (Rs < 933) {
if (negBalance < 58) {
if (Rs < 710) {
return 1;
}
else if (Rs >= 710) {
return 4;
}
}
else if (negBalance >= 58) {
return 8;
}
}
else if (Rs >= 933) {
return 1;
}
}
}
}
}
else if (ssDCval >= 70) {
if (refDCval < 81) {
if (tcor < 81) {
if (TSC < 13) {
if (posBalance < 66) {
if (diffMVdiffVal < 4) {
if (RsCsDiff < 73) {
if (Rs < 934) {
if (Rs < 731) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (Rs < 683) {
return 4;
}
else if (Rs >= 683) {
return 1;
}
}
}
else if (Rs >= 731) {
return 1;
}
}
else if (Rs >= 934) {
if (SC < 2274) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
else if (SC >= 2274) {
return 2;
}
}
}
else if (RsCsDiff >= 73) {
if (diffMVdiffVal < -14) {
return 4;
}
else if (diffMVdiffVal >= -14) {
return 2;
}
}
}
else if (diffMVdiffVal >= 4) {
if (posBalance < 61) {
if (MVDiff < 50) {
return 4;
}
else if (MVDiff >= 50) {
if (Cs < 943) {
return 8;
}
else if (Cs >= 943) {
return 4;
}
}
}
else if (posBalance >= 61) {
if (CsDiff < 6) {
if (ssDCval < 82) {
return 2;
}
else if (ssDCval >= 82) {
return 1;
}
}
else if (CsDiff >= 6) {
return 4;
}
}
}
}
else if (posBalance >= 66) {
if (AFD < 28) {
if (MVDiff < 39) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (MVDiff >= 39) {
return 4;
}
}
else if (AFD >= 28) {
if (Rs < 736) {
return 1;
}
else if (Rs >= 736) {
return 2;
}
}
}
}
else if (TSC >= 13) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
return 1;
}
}
}
else if (tcor >= 81) {
if (tcor < 84) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
if (Rs < 1102) {
return 2;
}
else if (Rs >= 1102) {
return 8;
}
}
}
else if (tcor >= 84) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (SC < 1380) {
if (ssDCval < 97) {
if (Cs < 722) {
if (Rs < 430) {
if (Rs < 280) {
return 8;
}
else if (Rs >= 280) {
if (Rs < 328) {
return 4;
}
else if (Rs >= 328) {
return 1;
}
}
}
else if (Rs >= 430) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (Cs >= 722) {
if (Rs < 520) {
return 4;
}
else if (Rs >= 520) {
return 2;
}
}
}
else if (ssDCval >= 97) {
if (diffRsCsdiff < 18) {
if (tcor < 83) {
if (mcTcor < 94) {
if (refDCval < 92) {
return 4;
}
else if (refDCval >= 92) {
return 2;
}
}
else if (mcTcor >= 94) {
if (ssDCval < 113) {
return 2;
}
else if (ssDCval >= 113) {
return 4;
}
}
}
else if (tcor >= 83) {
if (Rs < 495) {
return 4;
}
else if (Rs >= 495) {
if (Rs < 548) {
return 1;
}
else if (Rs >= 548) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 18) {
if (SC < 1097) {
if (Rs < 442) {
return 1;
}
else if (Rs >= 442) {
return 4;
}
}
else if (SC >= 1097) {
return 8;
}
}
}
}
else if (SC >= 1380) {
if (mu_mv_mag_sq < 13) {
if (diffRsCsdiff < 6) {
if (Rs < 1178) {
if (refDCval < 86) {
if (ssDCval < 80) {
return 4;
}
else if (ssDCval >= 80) {
return 2;
}
}
else if (refDCval >= 86) {
if (refDCval < 155) {
if (tcor < 72) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 4;
}
}
else if (tcor >= 72) {
if (AFD < 15) {
if (Rs < 1115) {
if (Rs < 957) {
return 2;
}
else if (Rs >= 957) {
return 4;
}
}
else if (Rs >= 1115) {
return 8;
}
}
else if (AFD >= 15) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 2;
}
}
}
}
else if (refDCval >= 155) {
if (Rs < 705) {
return 1;
}
else if (Rs >= 705) {
return 2;
}
}
}
}
else if (Rs >= 1178) {
if (negBalance < 56) {
if (Rs < 1220) {
return 2;
}
else if (Rs >= 1220) {
return 1;
}
}
else if (negBalance >= 56) {
return 4;
}
}
}
else if (diffRsCsdiff >= 6) {
if (RsDiff < 7) {
if (gchDC < 2) {
if (Rs < 776) {
return 4;
}
else if (Rs >= 776) {
return 1;
}
}
else if (gchDC >= 2) {
if (RsCsDiff < 48) {
if (Rs < 672) {
return 4;
}
else if (Rs >= 672) {
return 1;
}
}
else if (RsCsDiff >= 48) {
return 2;
}
}
}
else if (RsDiff >= 7) {
if (diffRsCsdiff < 26) {
if (Rs < 1110) {
return 8;
}
else if (Rs >= 1110) {
return 1;
}
}
else if (diffRsCsdiff >= 26) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 1026) {
if (AFD < 23) {
if (posBalance < 58) {
if (mu_mv_mag_sq < 48) {
if (AFD < 17) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
return 2;
}
}
else if (AFD >= 17) {
return 8;
}
}
else if (mu_mv_mag_sq >= 48) {
return 2;
}
}
else if (posBalance >= 58) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
if (ssDCval < 146) {
return 8;
}
else if (ssDCval >= 146) {
if (Rs < 867) {
return 8;
}
else if (Rs >= 867) {
return 1;
}
}
}
}
}
else if (AFD >= 23) {
if (posBalance < 70) {
if (Cs < 1150) {
return 1;
}
else if (Cs >= 1150) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
else if (posBalance >= 70) {
return 8;
}
}
}
else if (Rs >= 1026) {
if (MVDiff < 16) {
if (Cs < 762) {
return 4;
}
else if (Cs >= 762) {
return 8;
}
}
else if (MVDiff >= 16) {
if (mu_mv_mag_sq < 37) {
return 1;
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 1072) {
return 1;
}
else if (Rs >= 1072) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1285) {
if (negBalance < 47) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (diffAFD < 3) {
if (MVDiff < 46) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (Rs < 1421) {
return 4;
}
else if (Rs >= 1421) {
return 1;
}
}
}
else if (MVDiff >= 46) {
if (Rs < 1453) {
return 8;
}
else if (Rs >= 1453) {
if (Rs < 1863) {
return 1;
}
else if (Rs >= 1863) {
return 4;
}
}
}
}
else if (diffAFD >= 3) {
if (posBalance < 61) {
if (Rs < 1355) {
return 2;
}
else if (Rs >= 1355) {
return 1;
}
}
else if (posBalance >= 61) {
return 8;
}
}
}
}
else if (negBalance >= 47) {
if (AFD < 19) {
if (ssDCval < 128) {
if (SC < 2614) {
if (posBalance < 59) {
if (mu_mv_mag_sq < 3) {
if (diffMVdiffVal < -1) {
if (MVDiff < 52) {
return 2;
}
else if (MVDiff >= 52) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (Cs < 1053) {
return 4;
}
else if (Cs >= 1053) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
return 8;
}
}
}
else if (posBalance >= 59) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (ssDCval < 110) {
if (SC < 2212) {
return 8;
}
else if (SC >= 2212) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
return 4;
}
}
}
else if (ssDCval >= 110) {
return 8;
}
}
}
}
else if (SC >= 2614) {
return 8;
}
}
else if (ssDCval >= 128) {
if (Rs < 1368) {
return 8;
}
else if (Rs >= 1368) {
return 1;
}
}
}
else if (AFD >= 19) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Rs < 1414) {
if (MVDiff < 41) {
if (SC < 2258) {
return 8;
}
else if (SC >= 2258) {
return 4;
}
}
else if (MVDiff >= 41) {
return 1;
}
}
else if (Rs >= 1414) {
if (Rs < 1817) {
if (Cs < 785) {
return 2;
}
else if (Cs >= 785) {
return 1;
}
}
else if (Rs >= 1817) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (negBalance < 14) {
if (Rs < 702) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 56) {
if (Rs < 242) {
return 2;
}
else if (Rs >= 242) {
return 1;
}
}
else if (ssDCval >= 56) {
if (Rs < 441) {
if (Cs < 1028) {
return 4;
}
else if (Cs >= 1028) {
return 8;
}
}
else if (Rs >= 441) {
if (SC < 1393) {
if (Rs < 509) {
return 1;
}
else if (Rs >= 509) {
if (Cs < 685) {
return 8;
}
else if (Cs >= 685) {
if (Cs < 713) {
return 1;
}
else if (Cs >= 713) {
return 8;
}
}
}
}
else if (SC >= 1393) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 756) {
if (Rs < 681) {
return 8;
}
else if (Rs >= 681) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (posBalance < 30) {
return 1;
}
else if (posBalance >= 30) {
return 8;
}
}
}
}
else if (Cs >= 756) {
if (ssDCval < 77) {
if (Rs < 303) {
return 2;
}
else if (Rs >= 303) {
return 8;
}
}
else if (ssDCval >= 77) {
return 2;
}
}
}
}
else if (Rs >= 702) {
if (MVDiff < 6) {
if (SC < 2811) {
if (Rs < 965) {
if (diffMVdiffVal < -1) {
if (negBalance < 3) {
return 2;
}
else if (negBalance >= 3) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (Cs < 664) {
return 2;
}
else if (Cs >= 664) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 10) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
return 2;
}
}
else if (posBalance >= 10) {
if (ssDCval < 73) {
if (SC < 1466) {
if (Rs < 882) {
return 2;
}
else if (Rs >= 882) {
return 8;
}
}
else if (SC >= 1466) {
if (MVDiff < 3) {
if (Cs < 1148) {
return 8;
}
else if (Cs >= 1148) {
if (Rs < 864) {
return 1;
}
else if (Rs >= 864) {
return 8;
}
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (ssDCval >= 73) {
if (mcTcor < 99) {
if (Rs < 741) {
return 1;
}
else if (Rs >= 741) {
return 8;
}
}
else if (mcTcor >= 99) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (ssDCval < 137) {
return 8;
}
else if (ssDCval >= 137) {
if (Rs < 835) {
return 1;
}
else if (Rs >= 835) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 965) {
if (ssDCval < 53) {
if (TSC < 2) {
if (Cs < 548) {
return 1;
}
else if (Cs >= 548) {
return 8;
}
}
else if (TSC >= 2) {
if (Rs < 1549) {
return 1;
}
else if (Rs >= 1549) {
return 8;
}
}
}
else if (ssDCval >= 53) {
if (ssDCval < 64) {
if (Cs < 1003) {
if (mu_mv_mag_sq < 118) {
if (Cs < 705) {
return 1;
}
else if (Cs >= 705) {
return 8;
}
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 1595) {
return 8;
}
else if (Rs >= 1595) {
return 1;
}
}
}
else if (Cs >= 1003) {
return 1;
}
}
else if (ssDCval >= 64) {
if (diffMVdiffVal < -4) {
if (Rs < 1134) {
if (Cs < 1019) {
return 1;
}
else if (Cs >= 1019) {
return 8;
}
}
else if (Rs >= 1134) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
if (Cs < 1232) {
if (ssDCval < 152) {
if (Cs < 600) {
if (Rs < 1052) {
return 8;
}
else if (Rs >= 1052) {
return 4;
}
}
else if (Cs >= 600) {
if (diffRsCsdiff < -2) {
if (Cs < 954) {
return 8;
}
else if (Cs >= 954) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (Cs >= 1232) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 2811) {
if (Rs < 2042) {
if (SC < 2906) {
if (Rs < 1810) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 1786) {
return 8;
}
else if (Rs >= 1786) {
return 1;
}
}
}
else if (Rs >= 1810) {
if (negBalance < 7) {
return 8;
}
else if (negBalance >= 7) {
if (Cs < 987) {
return 1;
}
else if (Cs >= 987) {
if (Rs < 1837) {
return 1;
}
else if (Rs >= 1837) {
return 4;
}
}
}
}
}
else if (SC >= 2906) {
if (ssDCval < 68) {
if (RsCsDiff < 2) {
return 4;
}
else if (RsCsDiff >= 2) {
if (Rs < 1896) {
if (Rs < 1830) {
return 8;
}
else if (Rs >= 1830) {
return 1;
}
}
else if (Rs >= 1896) {
return 4;
}
}
}
else if (ssDCval >= 68) {
if (Rs < 1906) {
return 1;
}
else if (Rs >= 1906) {
return 8;
}
}
}
}
else if (Rs >= 2042) {
if (posBalance < 28) {
if (Cs < 709) {
return 4;
}
else if (Cs >= 709) {
if (diffRsCsdiff < -19) {
return 1;
}
else if (diffRsCsdiff >= -19) {
return 8;
}
}
}
else if (posBalance >= 28) {
if (Cs < 1277) {
if (mu_mv_mag_sq < 121) {
if (negBalance < 5) {
return 8;
}
else if (negBalance >= 5) {
if (Cs < 1096) {
return 1;
}
else if (Cs >= 1096) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 121) {
return 8;
}
}
else if (Cs >= 1277) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 83) {
if (negBalance < 11) {
if (tcor < 98) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
return 1;
}
}
else if (tcor >= 98) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (Rs < 1721) {
if (Rs < 1088) {
return 8;
}
else if (Rs >= 1088) {
return 1;
}
}
else if (Rs >= 1721) {
return 8;
}
}
}
}
else if (negBalance >= 11) {
if (Cs < 676) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
return 2;
}
}
else if (Cs >= 676) {
if (Rs < 2059) {
if (MVDiff < 23) {
if (Rs < 1090) {
return 8;
}
else if (Rs >= 1090) {
return 1;
}
}
else if (MVDiff >= 23) {
return 8;
}
}
else if (Rs >= 2059) {
return 8;
}
}
}
}
else if (ssDCval >= 83) {
if (TSCindex < 1) {
if (diffRsCsdiff < -46) {
if (ssDCval < 140) {
if (diffAFD < -15) {
if (diffMVdiffVal < -32) {
return 4;
}
else if (diffMVdiffVal >= -32) {
if (MVDiff < 59) {
return 4;
}
else if (MVDiff >= 59) {
return 1;
}
}
}
else if (diffAFD >= -15) {
if (Rs < 1101) {
if (Rs < 923) {
return 2;
}
else if (Rs >= 923) {
return 1;
}
}
else if (Rs >= 1101) {
if (Rs < 1571) {
return 8;
}
else if (Rs >= 1571) {
return 4;
}
}
}
}
else if (ssDCval >= 140) {
if (diffAFD < -28) {
return 4;
}
else if (diffAFD >= -28) {
return 8;
}
}
}
else if (diffRsCsdiff >= -46) {
if (Cs < 643) {
if (Rs < 1040) {
return 2;
}
else if (Rs >= 1040) {
return 1;
}
}
else if (Cs >= 643) {
if (Cs < 970) {
if (Rs < 794) {
return 8;
}
else if (Rs >= 794) {
return 4;
}
}
else if (Cs >= 970) {
if (ssDCval < 133) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
return 1;
}
}
else if (ssDCval >= 133) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
}
}
}
}
else if (TSCindex >= 1) {
return 8;
}
}
}
}
}
else if (negBalance >= 14) {
if (ssDCval < 88) {
if (Rs < 1793) {
if (Rs < 1096) {
if (refDCval < 64) {
if (Rs < 806) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (Rs < 523) {
if (negBalance < 22) {
if (mu_mv_mag_sq < 97) {
if (Cs < 835) {
return 4;
}
else if (Cs >= 835) {
return 1;
}
}
else if (mu_mv_mag_sq >= 97) {
if (Rs < 445) {
return 2;
}
else if (Rs >= 445) {
if (mu_mv_mag_sq < 121) {
if (Rs < 511) {
return 1;
}
else if (Rs >= 511) {
return 2;
}
}
else if (mu_mv_mag_sq >= 121) {
return 4;
}
}
}
}
else if (negBalance >= 22) {
if (AFD < 7) {
if (Rs < 460) {
if (MVDiff < 13) {
if (tcor < 96) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (tcor >= 96) {
if (tcor < 100) {
if (Rs < 235) {
return 4;
}
else if (Rs >= 235) {
if (negBalance < 35) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
if (Cs < 656) {
return 1;
}
else if (Cs >= 656) {
return 8;
}
}
}
else if (negBalance >= 35) {
return 4;
}
}
}
else if (tcor >= 100) {
if (Cs < 759) {
return 8;
}
else if (Cs >= 759) {
return 4;
}
}
}
}
else if (MVDiff >= 13) {
return 4;
}
}
else if (Rs >= 460) {
if (Cs < 607) {
if (Cs < 592) {
if (Cs < 553) {
return 1;
}
else if (Cs >= 553) {
return 4;
}
}
else if (Cs >= 592) {
return 2;
}
}
else if (Cs >= 607) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (AFD >= 7) {
if (RsCsDiff < 23) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
if (Rs < 401) {
return 1;
}
else if (Rs >= 401) {
return 4;
}
}
}
}
else if (RsCsDiff >= 23) {
if (AFD < 17) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (Cs < 619) {
return 2;
}
else if (Cs >= 619) {
return 8;
}
}
}
else if (AFD >= 17) {
if (Cs < 743) {
return 2;
}
else if (Cs >= 743) {
return 4;
}
}
}
}
}
}
else if (Rs >= 523) {
if (MVDiff < 1) {
if (Cs < 941) {
if (tcor < 97) {
if (ssDCval < 55) {
if (AFD < 4) {
if (Rs < 646) {
return 1;
}
else if (Rs >= 646) {
return 8;
}
}
else if (AFD >= 4) {
return 4;
}
}
else if (ssDCval >= 55) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 648) {
return 4;
}
else if (Rs >= 648) {
return 8;
}
}
}
}
else if (tcor >= 97) {
if (ssDCval < 44) {
return 4;
}
else if (ssDCval >= 44) {
return 8;
}
}
}
else if (Cs >= 941) {
if (Rs < 648) {
if (Cs < 1060) {
return 1;
}
else if (Cs >= 1060) {
return 4;
}
}
else if (Rs >= 648) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 17) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
return 1;
}
}
else if (negBalance >= 17) {
if (Cs < 922) {
if (MVDiff < 14) {
if (negBalance < 28) {
if (posBalance < 46) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (Rs < 787) {
if (posBalance < 32) {
return 1;
}
else if (posBalance >= 32) {
return 4;
}
}
else if (Rs >= 787) {
return 1;
}
}
}
else if (posBalance >= 46) {
if (tcor < 99) {
if (Rs < 536) {
return 2;
}
else if (Rs >= 536) {
if (posBalance < 70) {
return 4;
}
else if (posBalance >= 70) {
return 2;
}
}
}
else if (tcor >= 99) {
return 8;
}
}
}
else if (negBalance >= 28) {
if (mu_mv_mag_sq < 89) {
if (posBalance < 36) {
if (Rs < 599) {
return 2;
}
else if (Rs >= 599) {
return 8;
}
}
else if (posBalance >= 36) {
if (AFD < 10) {
if (Rs < 620) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (Rs >= 620) {
if (SC < 1397) {
if (tcor < 95) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 688) {
return 2;
}
else if (Rs >= 688) {
return 4;
}
}
}
else if (tcor >= 95) {
if (Cs < 714) {
if (diffRsCsdiff < 6) {
return 4;
}
else if (diffRsCsdiff >= 6) {
return 1;
}
}
else if (Cs >= 714) {
return 8;
}
}
}
else if (SC >= 1397) {
if (AFD < 5) {
if (Rs < 767) {
return 8;
}
else if (Rs >= 767) {
return 4;
}
}
else if (AFD >= 5) {
if (ssDCval < 42) {
return 2;
}
else if (ssDCval >= 42) {
if (diffAFD < -2) {
if (Rs < 744) {
return 2;
}
else if (Rs >= 744) {
return 4;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
}
}
}
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (Cs < 784) {
if (Rs < 784) {
return 8;
}
else if (Rs >= 784) {
return 2;
}
}
else if (Cs >= 784) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
}
}
else if (MVDiff >= 14) {
if (refDCval < 55) {
if (ssDCval < 51) {
if (TSC < 4) {
if (Rs < 684) {
return 8;
}
else if (Rs >= 684) {
return 1;
}
}
else if (TSC >= 4) {
return 4;
}
}
else if (ssDCval >= 51) {
if (Rs < 588) {
return 4;
}
else if (Rs >= 588) {
return 8;
}
}
}
else if (refDCval >= 55) {
return 1;
}
}
}
else if (Cs >= 922) {
if (negBalance < 33) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (Rs < 664) {
return 1;
}
else if (Rs >= 664) {
return 2;
}
}
}
else if (negBalance >= 33) {
if (mu_mv_mag_sq < 4) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (mu_mv_mag_sq >= 4) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 806) {
if (refDCval < 49) {
if (RsDiff < 3) {
if (Cs < 558) {
return 1;
}
else if (Cs >= 558) {
return 8;
}
}
else if (RsDiff >= 3) {
if (Rs < 969) {
if (Rs < 889) {
return 4;
}
else if (Rs >= 889) {
return 8;
}
}
else if (Rs >= 969) {
return 1;
}
}
}
else if (refDCval >= 49) {
if (SC < 2149) {
if (mu_mv_mag_sq < 121) {
if (RsDiff < 4) {
if (refDCval < 51) {
if (AFD < 5) {
if (Rs < 894) {
return 1;
}
else if (Rs >= 894) {
return 8;
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (refDCval >= 51) {
if (negBalance < 39) {
if (RsCsDiff < 7) {
if (MVDiff < 7) {
if (Cs < 1080) {
return 1;
}
else if (Cs >= 1080) {
if (Rs < 982) {
return 8;
}
else if (Rs >= 982) {
return 1;
}
}
}
else if (MVDiff >= 7) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (Cs < 706) {
return 1;
}
else if (Cs >= 706) {
return 8;
}
}
}
}
else if (RsCsDiff >= 7) {
if (Cs < 940) {
if (Rs < 874) {
if (Cs < 681) {
return 1;
}
else if (Cs >= 681) {
return 4;
}
}
else if (Rs >= 874) {
return 1;
}
}
else if (Cs >= 940) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (negBalance >= 39) {
if (Cs < 846) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
return 8;
}
}
}
else if (Cs >= 846) {
if (Rs < 831) {
return 8;
}
else if (Rs >= 831) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (AFD < 2) {
if (Cs < 662) {
return 8;
}
else if (Cs >= 662) {
return 1;
}
}
else if (AFD >= 2) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (Rs < 1070) {
return 1;
}
else if (Rs >= 1070) {
return 8;
}
}
}
}
}
else if (SC >= 2149) {
if (MVDiff < 6) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
return 4;
}
}
}
}
}
else if (refDCval >= 64) {
if (SC < 2120) {
if (negBalance < 26) {
if (Cs < 702) {
if (mu_mv_mag_sq < 97) {
if (Rs < 438) {
return 4;
}
else if (Rs >= 438) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (ssDCval < 72) {
if (SC < 1717) {
if (Rs < 583) {
return 4;
}
else if (Rs >= 583) {
return 2;
}
}
else if (SC >= 1717) {
if (posBalance < 47) {
if (posBalance < 41) {
if (Rs < 1073) {
return 8;
}
else if (Rs >= 1073) {
return 1;
}
}
else if (posBalance >= 41) {
return 2;
}
}
else if (posBalance >= 47) {
return 4;
}
}
}
else if (ssDCval >= 72) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
else if (Cs >= 702) {
if (ssDCval < 82) {
if (Rs < 718) {
if (RsDiff < 1) {
if (SC < 1248) {
return 4;
}
else if (SC >= 1248) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
return 8;
}
}
}
else if (RsDiff >= 1) {
return 4;
}
}
else if (Rs >= 718) {
if (RsCsDiff < 2) {
if (ssDCval < 68) {
if (Rs < 891) {
if (Cs < 736) {
return 8;
}
else if (Cs >= 736) {
return 4;
}
}
else if (Rs >= 891) {
if (Cs < 811) {
return 2;
}
else if (Cs >= 811) {
return 1;
}
}
}
else if (ssDCval >= 68) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (Cs < 729) {
return 2;
}
else if (Cs >= 729) {
return 1;
}
}
}
}
else if (RsCsDiff >= 2) {
if (ssDCval < 71) {
return 4;
}
else if (ssDCval >= 71) {
if (CsDiff < 3) {
if (Rs < 803) {
return 1;
}
else if (Rs >= 803) {
return 8;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
}
}
}
else if (ssDCval >= 82) {
if (CsDiff < 1) {
if (Rs < 632) {
return 2;
}
else if (Rs >= 632) {
return 1;
}
}
else if (CsDiff >= 1) {
if (SC < 1361) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (SC >= 1361) {
return 2;
}
}
}
}
}
else if (negBalance >= 26) {
if (refDCval < 75) {
if (CsDiff < 4) {
if (RsDiff < 1) {
if (Rs < 730) {
return 8;
}
else if (Rs >= 730) {
if (Rs < 866) {
return 4;
}
else if (Rs >= 866) {
if (Rs < 981) {
return 8;
}
else if (Rs >= 981) {
return 1;
}
}
}
}
else if (RsDiff >= 1) {
if (mcTcor < 98) {
if (refDCval < 67) {
if (RsDiff < 3) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Cs < 625) {
if (Rs < 672) {
return 1;
}
else if (Rs >= 672) {
return 4;
}
}
else if (Cs >= 625) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 802) {
return 8;
}
else if (Rs >= 802) {
return 4;
}
}
}
}
}
else if (RsDiff >= 3) {
if (Rs < 953) {
return 2;
}
else if (Rs >= 953) {
return 4;
}
}
}
else if (refDCval >= 67) {
if (SC < 1850) {
if (Rs < 775) {
if (SC < 1179) {
if (mu_mv_mag_sq < 77) {
if (negBalance < 47) {
return 4;
}
else if (negBalance >= 47) {
return 8;
}
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 512) {
return 2;
}
else if (Rs >= 512) {
return 8;
}
}
}
else if (SC >= 1179) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
if (Cs < 636) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 4;
}
}
else if (Cs >= 636) {
if (RsDiff < 2) {
if (Rs < 650) {
return 8;
}
else if (Rs >= 650) {
return 1;
}
}
else if (RsDiff >= 2) {
if (SC < 1760) {
return 1;
}
else if (SC >= 1760) {
if (Rs < 713) {
return 1;
}
else if (Rs >= 713) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 775) {
if (negBalance < 38) {
if (posBalance < 68) {
return 4;
}
else if (posBalance >= 68) {
return 2;
}
}
else if (negBalance >= 38) {
if (Rs < 924) {
return 2;
}
else if (Rs >= 924) {
if (SC < 1566) {
return 2;
}
else if (SC >= 1566) {
return 1;
}
}
}
}
}
else if (SC >= 1850) {
if (refDCval < 68) {
if (Rs < 965) {
if (Rs < 883) {
return 2;
}
else if (Rs >= 883) {
return 4;
}
}
else if (Rs >= 965) {
return 8;
}
}
else if (refDCval >= 68) {
if (SC < 1964) {
if (ssDCval < 70) {
if (Rs < 966) {
return 1;
}
else if (Rs >= 966) {
return 8;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (SC >= 1964) {
return 8;
}
}
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 63) {
if (mu_mv_mag_sq < 72) {
if (Cs < 1038) {
if (AFD < 9) {
if (Rs < 855) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (Rs < 547) {
return 1;
}
else if (Rs >= 547) {
return 4;
}
}
}
else if (Rs >= 855) {
return 1;
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (Cs >= 1038) {
if (Rs < 805) {
return 2;
}
else if (Rs >= 805) {
if (Rs < 881) {
return 4;
}
else if (Rs >= 881) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (Cs < 599) {
return 2;
}
else if (Cs >= 599) {
if (Rs < 638) {
return 8;
}
else if (Rs >= 638) {
if (diffTSC < 1) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 986) {
return 1;
}
else if (Rs >= 986) {
return 2;
}
}
}
else if (diffTSC >= 1) {
if (mu_mv_mag_sq < 109) {
return 1;
}
else if (mu_mv_mag_sq >= 109) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 63) {
return 4;
}
}
}
}
else if (CsDiff >= 4) {
if (Rs < 834) {
if (diffMVdiffVal < -6) {
if (RsCsDiff < 33) {
if (Rs < 306) {
return 1;
}
else if (Rs >= 306) {
return 2;
}
}
else if (RsCsDiff >= 33) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
if (mcTcor < 99) {
if (Cs < 1210) {
return 2;
}
else if (Cs >= 1210) {
if (Cs < 1252) {
return 1;
}
else if (Cs >= 1252) {
return 2;
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
else if (Rs >= 834) {
if (mu_mv_mag_sq < 31) {
if (Cs < 870) {
return 1;
}
else if (Cs >= 870) {
if (Cs < 1174) {
if (Rs < 944) {
return 8;
}
else if (Rs >= 944) {
return 2;
}
}
else if (Cs >= 1174) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 31) {
return 8;
}
}
}
}
else if (refDCval >= 75) {
if (ssDCval < 85) {
if (Cs < 774) {
if (ssDCval < 76) {
if (mu_mv_mag_sq < 43) {
if (Rs < 341) {
return 1;
}
else if (Rs >= 341) {
return 8;
}
}
else if (mu_mv_mag_sq >= 43) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (ssDCval >= 76) {
if (AFD < 19) {
if (refDCval < 85) {
return 8;
}
else if (refDCval >= 85) {
if (Rs < 579) {
return 2;
}
else if (Rs >= 579) {
return 8;
}
}
}
else if (AFD >= 19) {
return 1;
}
}
}
else if (Cs >= 774) {
if (mu_mv_mag_sq < 4) {
if (tcor < 91) {
if (Cs < 791) {
return 4;
}
else if (Cs >= 791) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (tcor >= 91) {
return 2;
}
}
else if (mu_mv_mag_sq >= 4) {
if (Rs < 844) {
if (Cs < 959) {
if (TSC < 4) {
if (Rs < 614) {
return 1;
}
else if (Rs >= 614) {
return 2;
}
}
else if (TSC >= 4) {
if (Rs < 795) {
return 8;
}
else if (Rs >= 795) {
return 2;
}
}
}
else if (Cs >= 959) {
if (Cs < 1174) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
}
else if (Cs >= 1174) {
return 1;
}
}
}
else if (Rs >= 844) {
if (SC < 2014) {
if (posBalance < 67) {
if (mu_mv_mag_sq < 123) {
return 8;
}
else if (mu_mv_mag_sq >= 123) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
return 4;
}
}
}
else if (posBalance >= 67) {
return 2;
}
}
else if (SC >= 2014) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (SC < 1574) {
if (TSC < 5) {
if (mu_mv_mag_sq < 96) {
if (Rs < 697) {
return 2;
}
else if (Rs >= 697) {
return 1;
}
}
else if (mu_mv_mag_sq >= 96) {
return 1;
}
}
else if (TSC >= 5) {
if (Rs < 705) {
if (Cs < 721) {
return 8;
}
else if (Cs >= 721) {
return 1;
}
}
else if (Rs >= 705) {
return 4;
}
}
}
else if (SC >= 1574) {
if (Rs < 917) {
return 4;
}
else if (Rs >= 917) {
return 8;
}
}
}
}
}
}
else if (SC >= 2120) {
if (tcor < 98) {
if (Rs < 1003) {
if (RsDiff < 3) {
if (mcTcor < 98) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
if (ssDCval < 81) {
return 4;
}
else if (ssDCval >= 81) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 987) {
return 2;
}
else if (Rs >= 987) {
return 4;
}
}
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (Rs >= 1003) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 111) {
return 8;
}
else if (mu_mv_mag_sq >= 111) {
return 4;
}
}
else if (MVDiff >= 3) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
else if (tcor >= 98) {
if (mu_mv_mag_sq < 122) {
if (Cs < 1111) {
if (Cs < 1101) {
return 4;
}
else if (Cs >= 1101) {
return 8;
}
}
else if (Cs >= 1111) {
return 4;
}
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 931) {
return 1;
}
else if (Rs >= 931) {
if (SC < 2185) {
return 4;
}
else if (SC >= 2185) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1096) {
if (Cs < 979) {
if (mu_mv_mag_sq < 122) {
if (SC < 2452) {
if (TSC < 4) {
if (ssDCval < 47) {
if (Cs < 773) {
if (Cs < 625) {
if (Rs < 1385) {
return 4;
}
else if (Rs >= 1385) {
return 4;
}
}
else if (Cs >= 625) {
return 4;
}
}
else if (Cs >= 773) {
if (Rs < 1311) {
return 8;
}
else if (Rs >= 1311) {
return 1;
}
}
}
else if (ssDCval >= 47) {
if (RsDiff < 4) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 118) {
if (TSCindex < 3) {
if (mu_mv_mag_sq < 97) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 97) {
return 1;
}
}
else if (TSCindex >= 3) {
if (Cs < 765) {
if (Rs < 1109) {
if (Rs < 1104) {
return 1;
}
else if (Rs >= 1104) {
return 2;
}
}
else if (Rs >= 1109) {
if (Rs < 1694) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
if (Rs < 1403) {
return 2;
}
else if (Rs >= 1403) {
return 1;
}
}
}
else if (Rs >= 1694) {
if (Rs < 1700) {
return 2;
}
else if (Rs >= 1700) {
if (mu_mv_mag_sq < 86) {
return 1;
}
else if (mu_mv_mag_sq >= 86) {
if (Rs < 1750) {
return 8;
}
else if (Rs >= 1750) {
return 4;
}
}
}
}
}
}
else if (Cs >= 765) {
if (CsDiff < 2) {
if (mu_mv_mag_sq < 62) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 62) {
if (Rs < 1340) {
return 1;
}
else if (Rs >= 1340) {
return 8;
}
}
}
else if (CsDiff >= 2) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 1464) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
return 1;
}
}
else if (Rs >= 1464) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (SC < 2376) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
if (Rs < 1499) {
return 1;
}
else if (Rs >= 1499) {
return 8;
}
}
}
else if (SC >= 2376) {
if (Rs < 1652) {
if (Cs < 936) {
return 8;
}
else if (Cs >= 936) {
return 1;
}
}
else if (Rs >= 1652) {
return 1;
}
}
}
}
else if (RsDiff >= 4) {
if (Rs < 1649) {
return 1;
}
else if (Rs >= 1649) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (Rs < 1562) {
if (RsCsDiff < 23) {
if (negBalance < 42) {
if (ssDCval < 64) {
if (AFD < 7) {
if (mu_mv_mag_sq < 104) {
return 1;
}
else if (mu_mv_mag_sq >= 104) {
if (Cs < 637) {
return 1;
}
else if (Cs >= 637) {
return 8;
}
}
}
else if (AFD >= 7) {
if (Cs < 836) {
if (Cs < 771) {
return 2;
}
else if (Cs >= 771) {
return 1;
}
}
else if (Cs >= 836) {
if (Rs < 1320) {
return 8;
}
else if (Rs >= 1320) {
return 4;
}
}
}
}
else if (ssDCval >= 64) {
if (MVDiff < 5) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (Rs < 1460) {
return 8;
}
else if (Rs >= 1460) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (negBalance < 33) {
if (Rs < 1296) {
return 8;
}
else if (Rs >= 1296) {
return 1;
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
}
else if (negBalance >= 42) {
if (negBalance < 53) {
if (Cs < 547) {
return 4;
}
else if (Cs >= 547) {
if (CsDiff < 3) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 1;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
}
else if (negBalance >= 53) {
if (Cs < 650) {
return 1;
}
else if (Cs >= 650) {
return 4;
}
}
}
}
else if (RsCsDiff >= 23) {
if (RsDiff < 5) {
if (MVDiff < 13) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (MVDiff >= 13) {
if (diffMVdiffVal < 8) {
return 2;
}
else if (diffMVdiffVal >= 8) {
if (Rs < 1221) {
return 8;
}
else if (Rs >= 1221) {
return 1;
}
}
}
}
else if (RsDiff >= 5) {
if (Rs < 1433) {
if (diffRsCsdiff < -9) {
if (Rs < 1294) {
return 4;
}
else if (Rs >= 1294) {
return 1;
}
}
else if (diffRsCsdiff >= -9) {
if (tcor < 89) {
if (Rs < 1290) {
return 1;
}
else if (Rs >= 1290) {
return 8;
}
}
else if (tcor >= 89) {
return 8;
}
}
}
else if (Rs >= 1433) {
return 1;
}
}
}
}
else if (Rs >= 1562) {
if (refDCval < 71) {
if (MVDiff < 36) {
if (Cs < 757) {
return 1;
}
else if (Cs >= 757) {
return 8;
}
}
else if (MVDiff >= 36) {
return 2;
}
}
else if (refDCval >= 71) {
return 1;
}
}
}
}
else if (SC >= 2452) {
if (mu_mv_mag_sq < 52) {
if (ssDCval < 74) {
if (Cs < 828) {
return 4;
}
else if (Cs >= 828) {
if (Rs < 1707) {
return 8;
}
else if (Rs >= 1707) {
return 2;
}
}
}
else if (ssDCval >= 74) {
return 1;
}
}
else if (mu_mv_mag_sq >= 52) {
if (diffAFD < 2) {
if (diffRsCsdiff < -13) {
if (Rs < 1673) {
return 1;
}
else if (Rs >= 1673) {
return 8;
}
}
else if (diffRsCsdiff >= -13) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 1;
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 1345) {
if (Cs < 732) {
if (posBalance < 47) {
if (Cs < 718) {
return 1;
}
else if (Cs >= 718) {
if (Rs < 1179) {
return 4;
}
else if (Rs >= 1179) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (Rs < 1223) {
return 2;
}
else if (Rs >= 1223) {
if (Rs < 1328) {
return 1;
}
else if (Rs >= 1328) {
return 8;
}
}
}
}
else if (Cs >= 732) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
return 1;
}
}
}
}
}
else if (Rs >= 1345) {
if (Rs < 1721) {
if (ssDCval < 52) {
if (posBalance < 36) {
return 1;
}
else if (posBalance >= 36) {
if (negBalance < 30) {
if (Cs < 756) {
return 4;
}
else if (Cs >= 756) {
if (Rs < 1436) {
return 8;
}
else if (Rs >= 1436) {
return 8;
}
}
}
else if (negBalance >= 30) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (refDCval < 75) {
if (SC < 2547) {
return 8;
}
else if (SC >= 2547) {
return 1;
}
}
else if (refDCval >= 75) {
if (Cs < 713) {
return 1;
}
else if (Cs >= 713) {
return 8;
}
}
}
}
else if (Rs >= 1721) {
return 1;
}
}
}
}
else if (Cs >= 979) {
if (AFD < 4) {
if (ssDCval < 49) {
if (SC < 2927) {
if (Rs < 1483) {
return 8;
}
else if (Rs >= 1483) {
if (Rs < 1687) {
if (Cs < 1122) {
if (Cs < 1060) {
return 1;
}
else if (Cs >= 1060) {
if (negBalance < 19) {
if (Cs < 1105) {
return 4;
}
else if (Cs >= 1105) {
if (Cs < 1114) {
return 4;
}
else if (Cs >= 1114) {
return 4;
}
}
}
else if (negBalance >= 19) {
if (posBalance < 38) {
if (Rs < 1594) {
return 4;
}
else if (Rs >= 1594) {
if (Rs < 1679) {
return 4;
}
else if (Rs >= 1679) {
return 4;
}
}
}
else if (posBalance >= 38) {
return 8;
}
}
}
}
else if (Cs >= 1122) {
return 4;
}
}
else if (Rs >= 1687) {
if (diffTSC < -1) {
if (Rs < 1741) {
return 8;
}
else if (Rs >= 1741) {
if (Cs < 1116) {
if (Rs < 1759) {
return 4;
}
else if (Rs >= 1759) {
return 8;
}
}
else if (Cs >= 1116) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (posBalance < 33) {
if (Rs < 1736) {
return 8;
}
else if (Rs >= 1736) {
if (Cs < 1153) {
if (Cs < 1140) {
if (Cs < 1107) {
if (Cs < 1101) {
if (Rs < 1786) {
return 4;
}
else if (Rs >= 1786) {
return 8;
}
}
else if (Cs >= 1101) {
return 4;
}
}
else if (Cs >= 1107) {
return 8;
}
}
else if (Cs >= 1140) {
return 4;
}
}
else if (Cs >= 1153) {
return 8;
}
}
}
else if (posBalance >= 33) {
if (Cs < 1156) {
if (posBalance < 34) {
if (Rs < 1771) {
return 4;
}
else if (Rs >= 1771) {
return 4;
}
}
else if (posBalance >= 34) {
if (RsDiff < 2) {
if (Cs < 1061) {
return 4;
}
else if (Cs >= 1061) {
if (Rs < 1702) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (Rs >= 1702) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (Rs < 1780) {
if (TSCindex < 3) {
if (Cs < 1101) {
if (Rs < 1739) {
if (Rs < 1735) {
return 8;
}
else if (Rs >= 1735) {
return 4;
}
}
else if (Rs >= 1739) {
return 8;
}
}
else if (Cs >= 1101) {
if (tcor < 99) {
if (Cs < 1134) {
if (posBalance < 39) {
if (SC < 2885) {
return 4;
}
else if (SC >= 2885) {
return 4;
}
}
else if (posBalance >= 39) {
return 4;
}
}
else if (Cs >= 1134) {
if (Rs < 1774) {
if (Rs < 1759) {
return 8;
}
else if (Rs >= 1759) {
return 8;
}
}
else if (Rs >= 1774) {
return 4;
}
}
}
else if (tcor >= 99) {
return 4;
}
}
}
else if (TSCindex >= 3) {
return 8;
}
}
else if (Rs >= 1780) {
if (Rs < 1783) {
return 8;
}
else if (Rs >= 1783) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Cs < 1130) {
return 4;
}
else if (Cs >= 1130) {
if (Cs < 1148) {
if (Rs < 1724) {
return 4;
}
else if (Rs >= 1724) {
return 8;
}
}
else if (Cs >= 1148) {
return 4;
}
}
}
}
}
else if (Cs >= 1156) {
if (Rs < 1745) {
return 4;
}
else if (Rs >= 1745) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 2927) {
return 4;
}
}
else if (ssDCval >= 49) {
if (Cs < 1056) {
if (posBalance < 37) {
if (mu_mv_mag_sq < 127) {
if (Rs < 1280) {
return 8;
}
else if (Rs >= 1280) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
else if (posBalance >= 37) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1716) {
return 1;
}
else if (Rs >= 1716) {
return 8;
}
}
}
}
else if (Cs >= 1056) {
if (Rs < 1481) {
if (refDCval < 70) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
if (negBalance < 36) {
if (mu_mv_mag_sq < 109) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (mu_mv_mag_sq >= 109) {
if (mu_mv_mag_sq < 122) {
if (CsDiff < 2) {
if (Rs < 1385) {
return 1;
}
else if (Rs >= 1385) {
return 8;
}
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 122) {
if (Cs < 1146) {
return 4;
}
else if (Cs >= 1146) {
if (diffRsCsdiff < 1) {
if (Cs < 1152) {
return 1;
}
else if (Cs >= 1152) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (Rs < 1282) {
return 4;
}
else if (Rs >= 1282) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 1) {
return 1;
}
}
}
}
}
else if (negBalance >= 36) {
return 1;
}
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 128) {
if (tcor < 97) {
if (Rs < 1346) {
return 4;
}
else if (Rs >= 1346) {
return 8;
}
}
else if (tcor >= 97) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
if (negBalance < 23) {
return 8;
}
else if (negBalance >= 23) {
if (Rs < 1408) {
return 4;
}
else if (Rs >= 1408) {
if (Rs < 1445) {
return 8;
}
else if (Rs >= 1445) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
}
else if (Rs >= 1481) {
if (negBalance < 20) {
if (Rs < 1685) {
return 8;
}
else if (Rs >= 1685) {
return 1;
}
}
else if (negBalance >= 20) {
return 8;
}
}
}
}
}
else if (AFD >= 4) {
if (AFD < 9) {
if (Rs < 1387) {
if (Cs < 1224) {
if (Rs < 1178) {
if (mu_mv_mag_sq < 106) {
if (MVDiff < 7) {
if (AFD < 7) {
if (Cs < 1009) {
if (Rs < 1123) {
return 8;
}
else if (Rs >= 1123) {
return 1;
}
}
else if (Cs >= 1009) {
return 8;
}
}
else if (AFD >= 7) {
if (Rs < 1131) {
return 8;
}
else if (Rs >= 1131) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Cs < 1096) {
return 1;
}
else if (Cs >= 1096) {
return 4;
}
}
}
else if (Rs >= 1178) {
if (RsDiff < 3) {
if (SC < 2581) {
return 8;
}
else if (SC >= 2581) {
return 1;
}
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 116) {
if (Cs < 1209) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
return 4;
}
}
else if (Cs >= 1209) {
return 4;
}
}
else if (mu_mv_mag_sq >= 116) {
return 4;
}
}
}
}
else if (Cs >= 1224) {
if (ssDCval < 67) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
return 4;
}
}
else if (ssDCval >= 67) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
if (Rs < 1240) {
return 4;
}
else if (Rs >= 1240) {
return 1;
}
}
}
}
}
else if (Rs >= 1387) {
if (posBalance < 48) {
if (CsDiff < 3) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (CsDiff >= 3) {
if (Rs < 1756) {
if (Rs < 1623) {
return 4;
}
else if (Rs >= 1623) {
return 1;
}
}
else if (Rs >= 1756) {
return 2;
}
}
}
else if (posBalance >= 48) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (diffRsCsdiff < 2) {
if (SC < 2933) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
return 8;
}
}
else if (SC >= 2933) {
return 1;
}
}
else if (diffRsCsdiff >= 2) {
if (Cs < 1052) {
return 2;
}
else if (Cs >= 1052) {
return 8;
}
}
}
}
}
}
else if (AFD >= 9) {
if (posBalance < 44) {
if (Rs < 1575) {
return 2;
}
else if (Rs >= 1575) {
return 8;
}
}
else if (posBalance >= 44) {
if (SC < 2795) {
return 4;
}
else if (SC >= 2795) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1793) {
if (TSC < 4) {
if (refDCval < 83) {
if (ssDCval < 48) {
if (posBalance < 39) {
if (mu_mv_mag_sq < 128) {
return 4;
}
else if (mu_mv_mag_sq >= 128) {
if (SC < 2932) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Cs < 1092) {
return 4;
}
else if (Cs >= 1092) {
return 4;
}
}
}
else if (SC >= 2932) {
return 4;
}
}
}
else if (posBalance >= 39) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 4;
}
}
}
else if (ssDCval >= 48) {
if (posBalance < 46) {
if (ssDCval < 71) {
if (mu_mv_mag_sq < 103) {
if (tcor < 96) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Cs < 1019) {
return 4;
}
else if (Cs >= 1019) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 8;
}
}
}
}
else if (tcor >= 96) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (posBalance < 29) {
if (Cs < 1081) {
return 4;
}
else if (Cs >= 1081) {
return 1;
}
}
else if (posBalance >= 29) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (tcor < 99) {
if (SC < 3240) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (posBalance < 35) {
if (posBalance < 34) {
return 1;
}
else if (posBalance >= 34) {
return 8;
}
}
else if (posBalance >= 35) {
return 1;
}
}
}
else if (SC >= 3240) {
if (Rs < 2540) {
return 2;
}
else if (Rs >= 2540) {
if (Rs < 2586) {
return 4;
}
else if (Rs >= 2586) {
return 8;
}
}
}
}
else if (tcor >= 99) {
if (Cs < 1095) {
if (Cs < 993) {
if (SC < 2894) {
return 1;
}
else if (SC >= 2894) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (Rs < 2368) {
return 8;
}
else if (Rs >= 2368) {
return 1;
}
}
}
}
else if (Cs >= 993) {
if (diffRsCsdiff < -1) {
if (Rs < 1912) {
return 1;
}
else if (Rs >= 1912) {
return 4;
}
}
else if (diffRsCsdiff >= -1) {
return 4;
}
}
}
else if (Cs >= 1095) {
if (Rs < 2088) {
return 1;
}
else if (Rs >= 2088) {
return 2;
}
}
}
}
}
else if (ssDCval >= 71) {
if (Cs < 1136) {
return 1;
}
else if (Cs >= 1136) {
return 8;
}
}
}
else if (posBalance >= 46) {
if (SC < 2805) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (SC >= 2805) {
if (Rs < 2361) {
return 8;
}
else if (Rs >= 2361) {
if (Rs < 2446) {
return 1;
}
else if (Rs >= 2446) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 83) {
if (MVDiff < 3) {
if (Rs < 2186) {
return 4;
}
else if (Rs >= 2186) {
return 8;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (TSC >= 4) {
if (Cs < 1034) {
if (refDCval < 71) {
if (Rs < 2031) {
if (TSC < 5) {
if (posBalance < 40) {
if (Rs < 1892) {
return 4;
}
else if (Rs >= 1892) {
return 8;
}
}
else if (posBalance >= 40) {
return 1;
}
}
else if (TSC >= 5) {
if (Cs < 811) {
return 2;
}
else if (Cs >= 811) {
return 8;
}
}
}
else if (Rs >= 2031) {
if (tcor < 89) {
if (Cs < 830) {
return 2;
}
else if (Cs >= 830) {
return 8;
}
}
else if (tcor >= 89) {
if (negBalance < 28) {
return 1;
}
else if (negBalance >= 28) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 71) {
if (posBalance < 33) {
if (AFD < 9) {
if (Cs < 674) {
return 4;
}
else if (Cs >= 674) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (gchDC < 2) {
if (Cs < 728) {
if (Rs < 2341) {
return 4;
}
else if (Rs >= 2341) {
return 4;
}
}
else if (Cs >= 728) {
if (Cs < 874) {
return 8;
}
else if (Cs >= 874) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (AFD >= 9) {
if (Rs < 1940) {
return 2;
}
else if (Rs >= 1940) {
return 1;
}
}
}
else if (posBalance >= 33) {
if (posBalance < 51) {
if (mu_mv_mag_sq < 91) {
if (diffRsCsdiff < -28) {
if (Cs < 706) {
return 1;
}
else if (Cs >= 706) {
return 2;
}
}
else if (diffRsCsdiff >= -28) {
return 1;
}
}
else if (mu_mv_mag_sq >= 91) {
return 2;
}
}
else if (posBalance >= 51) {
if (negBalance < 36) {
if (mu_mv_mag_sq < 84) {
if (Rs < 2441) {
return 1;
}
else if (Rs >= 2441) {
return 4;
}
}
else if (mu_mv_mag_sq >= 84) {
return 2;
}
}
else if (negBalance >= 36) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1034) {
if (Rs < 2190) {
if (negBalance < 46) {
if (negBalance < 42) {
if (Rs < 2045) {
return 8;
}
else if (Rs >= 2045) {
return 4;
}
}
else if (negBalance >= 42) {
if (Cs < 1163) {
return 8;
}
else if (Cs >= 1163) {
return 1;
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
else if (Rs >= 2190) {
if (negBalance < 41) {
if (SC < 3517) {
return 4;
}
else if (SC >= 3517) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (ssDCval < 74) {
if (Rs < 2534) {
return 8;
}
else if (Rs >= 2534) {
return 4;
}
}
else if (ssDCval >= 74) {
if (Rs < 2387) {
return 1;
}
else if (Rs >= 2387) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (Rs < 401) {
if (MVDiff < 8) {
if (ssDCval < 122) {
if (mu_mv_mag_sq < 109) {
if (ssDCval < 103) {
if (Cs < 747) {
return 1;
}
else if (Cs >= 747) {
if (Cs < 834) {
return 4;
}
else if (Cs >= 834) {
return 8;
}
}
}
else if (ssDCval >= 103) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 369) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (Rs >= 369) {
return 8;
}
}
}
else if (ssDCval >= 122) {
if (Rs < 323) {
return 2;
}
else if (Rs >= 323) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (TSC < 5) {
if (Cs < 982) {
return 2;
}
else if (Cs >= 982) {
return 8;
}
}
else if (TSC >= 5) {
if (Rs < 391) {
return 8;
}
else if (Rs >= 391) {
return 1;
}
}
}
}
else if (Rs >= 401) {
if (mu_mv_mag_sq < 128) {
if (mu_mv_mag_sq < 98) {
if (negBalance < 27) {
if (refDCval < 116) {
if (posBalance < 82) {
if (Rs < 444) {
return 2;
}
else if (Rs >= 444) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (MVDiff < 2) {
if (Rs < 1357) {
return 1;
}
else if (Rs >= 1357) {
return 8;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
else if (posBalance >= 82) {
return 8;
}
}
else if (refDCval >= 116) {
if (tcor < 97) {
if (TSC < 4) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
return 2;
}
}
else if (TSC >= 4) {
if (Rs < 2039) {
return 2;
}
else if (Rs >= 2039) {
return 8;
}
}
}
else if (tcor >= 97) {
if (Cs < 602) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
return 2;
}
}
else if (Cs >= 602) {
return 2;
}
}
}
}
else if (negBalance >= 27) {
if (SC < 1559) {
if (MVDiff < 3) {
if (negBalance < 44) {
if (Rs < 487) {
return 8;
}
else if (Rs >= 487) {
if (Rs < 583) {
if (Cs < 573) {
return 2;
}
else if (Cs >= 573) {
return 1;
}
}
else if (Rs >= 583) {
if (Rs < 587) {
return 8;
}
else if (Rs >= 587) {
return 2;
}
}
}
}
else if (negBalance >= 44) {
if (Cs < 553) {
return 2;
}
else if (Cs >= 553) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Rs < 542) {
return 4;
}
else if (Rs >= 542) {
return 1;
}
}
}
}
}
else if (MVDiff >= 3) {
if (mcTcor < 98) {
if (AFD < 9) {
if (ssDCval < 157) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (ssDCval < 114) {
return 8;
}
else if (ssDCval >= 114) {
if (ssDCval < 119) {
return 2;
}
else if (ssDCval >= 119) {
if (Rs < 684) {
return 8;
}
else if (Rs >= 684) {
return 2;
}
}
}
}
}
else if (ssDCval >= 157) {
return 4;
}
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 23) {
if (MVDiff < 7) {
if (SC < 1362) {
return 1;
}
else if (SC >= 1362) {
return 4;
}
}
else if (MVDiff >= 7) {
if (ssDCval < 98) {
if (Rs < 616) {
return 1;
}
else if (Rs >= 616) {
return 4;
}
}
else if (ssDCval >= 98) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (MVDiff < 31) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 8;
}
}
else if (MVDiff >= 31) {
return 2;
}
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 59) {
if (SC < 1188) {
if (Rs < 459) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 410) {
return 4;
}
else if (Rs >= 410) {
return 2;
}
}
}
else if (Rs >= 459) {
if (refDCval < 100) {
if (Rs < 472) {
return 2;
}
else if (Rs >= 472) {
return 8;
}
}
else if (refDCval >= 100) {
if (ssDCval < 149) {
return 2;
}
else if (ssDCval >= 149) {
if (Rs < 503) {
return 1;
}
else if (Rs >= 503) {
return 2;
}
}
}
}
}
else if (SC >= 1188) {
if (AFD < 12) {
if (tcor < 99) {
if (negBalance < 47) {
if (Rs < 518) {
return 1;
}
else if (Rs >= 518) {
return 2;
}
}
else if (negBalance >= 47) {
if (posBalance < 55) {
if (Cs < 625) {
return 1;
}
else if (Cs >= 625) {
return 2;
}
}
else if (posBalance >= 55) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
else if (tcor >= 99) {
if (SC < 1208) {
return 2;
}
else if (SC >= 1208) {
return 1;
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 15) {
return 1;
}
else if (mu_mv_mag_sq >= 15) {
if (Rs < 767) {
return 2;
}
else if (Rs >= 767) {
return 8;
}
}
}
}
}
else if (negBalance >= 59) {
if (Cs < 703) {
if (Rs < 578) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
return 2;
}
}
else if (Rs >= 578) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
return 8;
}
}
}
else if (Cs >= 703) {
return 8;
}
}
}
}
}
else if (SC >= 1559) {
if (refDCval < 101) {
if (TSC < 7) {
if (mu_mv_mag_sq < 66) {
if (TSC < 5) {
if (MVDiff < 29) {
return 1;
}
else if (MVDiff >= 29) {
if (Rs < 2106) {
return 8;
}
else if (Rs >= 2106) {
return 2;
}
}
}
else if (TSC >= 5) {
if (MVDiff < 9) {
if (refDCval < 94) {
if (Cs < 882) {
return 1;
}
else if (Cs >= 882) {
if (Rs < 2158) {
return 4;
}
else if (Rs >= 2158) {
return 1;
}
}
}
else if (refDCval >= 94) {
if (negBalance < 51) {
if (Rs < 2428) {
return 4;
}
else if (Rs >= 2428) {
return 8;
}
}
else if (negBalance >= 51) {
if (Rs < 776) {
return 4;
}
else if (Rs >= 776) {
return 1;
}
}
}
}
else if (MVDiff >= 9) {
if (Cs < 1115) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
return 2;
}
}
else if (Cs >= 1115) {
if (Rs < 1492) {
return 2;
}
else if (Rs >= 1492) {
if (Rs < 1989) {
return 8;
}
else if (Rs >= 1989) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (Rs < 910) {
if (Cs < 897) {
if (Rs < 753) {
return 1;
}
else if (Rs >= 753) {
return 8;
}
}
else if (Cs >= 897) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (Rs >= 910) {
if (Cs < 911) {
if (Rs < 1911) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Rs >= 1911) {
return 1;
}
}
else if (Cs >= 911) {
if (diffAFD < -2) {
if (Rs < 1850) {
return 4;
}
else if (Rs >= 1850) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
}
else if (TSC >= 7) {
if (diffRsCsdiff < -17) {
if (Rs < 2011) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
return 8;
}
}
else if (Rs >= 2011) {
return 4;
}
}
else if (diffRsCsdiff >= -17) {
if (Cs < 1028) {
if (SC < 2597) {
if (Rs < 811) {
return 4;
}
else if (Rs >= 811) {
return 8;
}
}
else if (SC >= 2597) {
return 4;
}
}
else if (Cs >= 1028) {
if (Cs < 1053) {
if (Cs < 1036) {
return 2;
}
else if (Cs >= 1036) {
return 1;
}
}
else if (Cs >= 1053) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
if (Cs < 1137) {
if (Cs < 1098) {
return 8;
}
else if (Cs >= 1098) {
if (Cs < 1122) {
return 2;
}
else if (Cs >= 1122) {
return 1;
}
}
}
else if (Cs >= 1137) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 101) {
if (posBalance < 70) {
if (MVDiff < 3) {
if (SC < 2480) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1447) {
return 4;
}
else if (Rs >= 1447) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (SC < 1706) {
return 1;
}
else if (SC >= 1706) {
return 8;
}
}
}
}
else if (SC >= 2480) {
if (Rs < 1862) {
if (Rs < 1695) {
return 8;
}
else if (Rs >= 1695) {
return 2;
}
}
else if (Rs >= 1862) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (negBalance < 49) {
if (MVDiff < 23) {
if (diffRsCsdiff < -40) {
if (Rs < 2157) {
return 2;
}
else if (Rs >= 2157) {
return 1;
}
}
else if (diffRsCsdiff >= -40) {
if (SC < 1624) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 844) {
return 4;
}
else if (Rs >= 844) {
return 8;
}
}
}
else if (SC >= 1624) {
if (Rs < 2355) {
if (ssDCval < 157) {
if (Rs < 925) {
if (Rs < 896) {
if (Rs < 587) {
return 1;
}
else if (Rs >= 587) {
if (posBalance < 46) {
if (Rs < 661) {
return 1;
}
else if (Rs >= 661) {
return 8;
}
}
else if (posBalance >= 46) {
return 8;
}
}
}
else if (Rs >= 896) {
return 1;
}
}
else if (Rs >= 925) {
if (MVDiff < 5) {
if (posBalance < 62) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
if (Rs < 1787) {
return 8;
}
else if (Rs >= 1787) {
return 2;
}
}
}
else if (posBalance >= 62) {
if (Rs < 1564) {
return 4;
}
else if (Rs >= 1564) {
return 2;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (ssDCval >= 157) {
if (Rs < 1121) {
return 4;
}
else if (Rs >= 1121) {
return 2;
}
}
}
else if (Rs >= 2355) {
if (posBalance < 50) {
if (Cs < 1014) {
return 1;
}
else if (Cs >= 1014) {
return 8;
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 59) {
return 8;
}
else if (mu_mv_mag_sq >= 59) {
if (ssDCval < 137) {
return 1;
}
else if (ssDCval >= 137) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 23) {
if (Rs < 1503) {
if (ssDCval < 130) {
return 8;
}
else if (ssDCval >= 130) {
return 4;
}
}
else if (Rs >= 1503) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (tcor < 98) {
if (Cs < 873) {
if (tcor < 90) {
if (MVDiff < 20) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 2;
}
}
else if (MVDiff >= 20) {
if (Cs < 640) {
return 1;
}
else if (Cs >= 640) {
return 2;
}
}
}
else if (tcor >= 90) {
if (mu_mv_mag_sq < 42) {
if (Rs < 808) {
return 2;
}
else if (Rs >= 808) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (Rs < 1998) {
return 8;
}
else if (Rs >= 1998) {
return 2;
}
}
}
}
else if (Cs >= 873) {
if (refDCval < 142) {
if (Rs < 1195) {
if (Cs < 1021) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 1;
}
}
else if (Cs >= 1021) {
return 8;
}
}
else if (Rs >= 1195) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
if (SC < 2794) {
if (Rs < 1377) {
return 8;
}
else if (Rs >= 1377) {
return 2;
}
}
else if (SC >= 2794) {
return 8;
}
}
}
}
else if (refDCval >= 142) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (Rs < 698) {
return 1;
}
else if (Rs >= 698) {
return 2;
}
}
}
}
}
else if (tcor >= 98) {
return 8;
}
}
}
}
else if (posBalance >= 70) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (refDCval < 97) {
if (Cs < 873) {
if (MVDiff < 1) {
if (diffMVdiffVal < -1) {
if (Rs < 1677) {
return 4;
}
else if (Rs >= 1677) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2162) {
if (AFD < 4) {
if (posBalance < 26) {
return 4;
}
else if (posBalance >= 26) {
if (Cs < 762) {
return 8;
}
else if (Cs >= 762) {
if (Cs < 831) {
return 4;
}
else if (Cs >= 831) {
return 8;
}
}
}
}
else if (AFD >= 4) {
return 4;
}
}
else if (Rs >= 2162) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
}
}
else if (MVDiff >= 1) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
}
else if (Cs >= 873) {
if (tcor < 98) {
if (RsCsDiff < 23) {
if (Rs < 853) {
if (AFD < 5) {
if (Rs < 605) {
return 8;
}
else if (Rs >= 605) {
return 1;
}
}
else if (AFD >= 5) {
return 4;
}
}
else if (Rs >= 853) {
return 8;
}
}
else if (RsCsDiff >= 23) {
if (Rs < 1560) {
return 8;
}
else if (Rs >= 1560) {
return 2;
}
}
}
else if (tcor >= 98) {
return 8;
}
}
}
else if (refDCval >= 97) {
if (RsDiff < 1) {
if (Cs < 1190) {
if (mu_mv_mag_sq < 107) {
if (ssDCval < 119) {
if (Rs < 666) {
return 2;
}
else if (Rs >= 666) {
return 8;
}
}
else if (ssDCval >= 119) {
return 4;
}
}
else if (mu_mv_mag_sq >= 107) {
if (ssDCval < 97) {
return 1;
}
else if (ssDCval >= 97) {
return 8;
}
}
}
else if (Cs >= 1190) {
return 1;
}
}
else if (RsDiff >= 1) {
if (Rs < 521) {
if (ssDCval < 117) {
if (mu_mv_mag_sq < 108) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 744) {
return 1;
}
else if (Cs >= 744) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (Cs < 782) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (Cs >= 782) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (refDCval < 118) {
return 4;
}
else if (refDCval >= 118) {
return 8;
}
}
}
else if (Rs >= 521) {
if (Rs < 1693) {
if (SC < 1221) {
if (mu_mv_mag_sq < 102) {
if (Rs < 582) {
return 1;
}
else if (Rs >= 582) {
return 2;
}
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 630) {
return 8;
}
else if (Rs >= 630) {
return 1;
}
}
}
else if (SC >= 1221) {
if (mu_mv_mag_sq < 100) {
if (Cs < 1173) {
return 8;
}
else if (Cs >= 1173) {
if (Rs < 664) {
return 1;
}
else if (Rs >= 664) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (SC < 1360) {
if (Rs < 651) {
return 8;
}
else if (Rs >= 651) {
return 1;
}
}
else if (SC >= 1360) {
return 8;
}
}
}
}
else if (Rs >= 1693) {
if (SC < 3018) {
if (mu_mv_mag_sq < 111) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1001) {
return 1;
}
else if (Cs >= 1001) {
return 2;
}
}
}
else if (SC >= 3018) {
if (diffMVdiffVal < -3) {
if (Cs < 1190) {
return 8;
}
else if (Cs >= 1190) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 123) {
if (CsDiff < 2) {
if (Rs < 2155) {
if (posBalance < 23) {
return 4;
}
else if (posBalance >= 23) {
return 8;
}
}
else if (Rs >= 2155) {
if (Rs < 2383) {
return 4;
}
else if (Rs >= 2383) {
return 8;
}
}
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 123) {
if (Rs < 1025) {
return 8;
}
else if (Rs >= 1025) {
return 1;
}
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (ssDCval < 91) {
if (mcTcor < 93) {
if (mcTcor < 79) {
if (refDCval < 42) {
return 8;
}
else if (refDCval >= 42) {
if (mcTcor < 75) {
if (Cs < 1210) {
if (ssDCval < 48) {
if (Rs < 1428) {
return 4;
}
else if (Rs >= 1428) {
return 1;
}
}
else if (ssDCval >= 48) {
if (AFD < 31) {
if (mcTcor < 71) {
return 8;
}
else if (mcTcor >= 71) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 1;
}
}
}
else if (AFD >= 31) {
if (Rs < 1288) {
return 1;
}
else if (Rs >= 1288) {
return 4;
}
}
}
}
else if (Cs >= 1210) {
if (mu_mv_mag_sq < 5) {
return 4;
}
else if (mu_mv_mag_sq >= 5) {
return 1;
}
}
}
else if (mcTcor >= 75) {
if (AFD < 25) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (AFD >= 25) {
if (Rs < 1903) {
return 2;
}
else if (Rs >= 1903) {
return 1;
}
}
}
}
}
else if (mcTcor >= 79) {
if (RsDiff < 12) {
if (posBalance < 46) {
if (MVDiff < 14) {
if (Rs < 2100) {
if (AFD < 12) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (AFD >= 12) {
if (Rs < 1922) {
return 2;
}
else if (Rs >= 1922) {
return 1;
}
}
}
else if (Rs >= 2100) {
if (AFD < 15) {
if (ssDCval < 67) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (Rs < 2421) {
return 2;
}
else if (Rs >= 2421) {
return 4;
}
}
}
else if (ssDCval >= 67) {
return 2;
}
}
else if (AFD >= 15) {
return 2;
}
}
}
else if (MVDiff >= 14) {
if (negBalance < 59) {
if (ssDCval < 51) {
if (Rs < 1783) {
return 8;
}
else if (Rs >= 1783) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 1;
}
}
}
else if (ssDCval >= 51) {
if (mu_mv_mag_sq < 30) {
if (Rs < 2293) {
return 2;
}
else if (Rs >= 2293) {
if (Rs < 2365) {
return 8;
}
else if (Rs >= 2365) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (Rs < 2513) {
return 2;
}
else if (Rs >= 2513) {
return 4;
}
}
}
}
else if (negBalance >= 59) {
return 8;
}
}
}
else if (posBalance >= 46) {
if (AFD < 23) {
if (AFD < 16) {
if (CsDiff < 7) {
if (TSC < 7) {
if (ssDCval < 75) {
if (Rs < 1943) {
if (Rs < 978) {
return 1;
}
else if (Rs >= 978) {
return 8;
}
}
else if (Rs >= 1943) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
return 8;
}
}
}
else if (ssDCval >= 75) {
return 1;
}
}
else if (TSC >= 7) {
if (Cs < 911) {
if (MVDiff < 37) {
if (RsCsDiff < 83) {
if (Cs < 624) {
if (Rs < 1547) {
return 2;
}
else if (Rs >= 1547) {
return 1;
}
}
else if (Cs >= 624) {
if (tcor < 82) {
if (Rs < 2144) {
return 8;
}
else if (Rs >= 2144) {
return 4;
}
}
else if (tcor >= 82) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 4;
}
}
}
}
else if (RsCsDiff >= 83) {
if (SC < 2776) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 1;
}
}
}
else if (SC >= 2776) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
return 4;
}
}
}
}
}
else if (MVDiff >= 37) {
return 8;
}
}
else if (Cs >= 911) {
if (diffAFD < -1) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (diffAFD >= -1) {
if (MVDiff < 32) {
if (posBalance < 53) {
if (Cs < 1131) {
return 4;
}
else if (Cs >= 1131) {
return 2;
}
}
else if (posBalance >= 53) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
return 4;
}
}
}
else if (MVDiff >= 32) {
if (Rs < 1904) {
return 1;
}
else if (Rs >= 1904) {
return 2;
}
}
}
}
}
}
else if (CsDiff >= 7) {
if (negBalance < 47) {
if (MVDiff < 17) {
if (Cs < 1255) {
return 8;
}
else if (Cs >= 1255) {
return 1;
}
}
else if (MVDiff >= 17) {
return 1;
}
}
else if (negBalance >= 47) {
if (TSC < 7) {
if (Rs < 1384) {
return 4;
}
else if (Rs >= 1384) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
else if (AFD >= 16) {
if (mu_mv_mag_sq < 24) {
if (diffRsCsdiff < -80) {
if (Rs < 1968) {
return 4;
}
else if (Rs >= 1968) {
if (Rs < 2072) {
return 8;
}
else if (Rs >= 2072) {
return 2;
}
}
}
else if (diffRsCsdiff >= -80) {
if (AFD < 19) {
if (Cs < 1249) {
if (Cs < 552) {
return 2;
}
else if (Cs >= 552) {
return 8;
}
}
else if (Cs >= 1249) {
if (TSC < 10) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 8;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (AFD >= 19) {
if (mu_mv_mag_sq < 9) {
if (ssDCval < 81) {
if (Cs < 1185) {
return 4;
}
else if (Cs >= 1185) {
if (Rs < 1190) {
return 2;
}
else if (Rs >= 1190) {
return 1;
}
}
}
else if (ssDCval >= 81) {
if (mu_mv_mag_sq < 6) {
return 8;
}
else if (mu_mv_mag_sq >= 6) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
if (refDCval < 79) {
if (negBalance < 52) {
return 4;
}
else if (negBalance >= 52) {
return 8;
}
}
else if (refDCval >= 79) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (mcTcor < 90) {
if (Cs < 1151) {
if (Rs < 1067) {
return 8;
}
else if (Rs >= 1067) {
return 4;
}
}
else if (Cs >= 1151) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 2408) {
return 1;
}
else if (Rs >= 2408) {
return 4;
}
}
}
}
else if (mcTcor >= 90) {
if (Rs < 2207) {
return 2;
}
else if (Rs >= 2207) {
return 8;
}
}
}
}
}
else if (AFD >= 23) {
if (refDCval < 70) {
if (mu_mv_mag_sq < 41) {
if (TSC < 12) {
if (diffRsCsdiff < -28) {
if (Rs < 1129) {
return 4;
}
else if (Rs >= 1129) {
return 8;
}
}
else if (diffRsCsdiff >= -28) {
return 4;
}
}
else if (TSC >= 12) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 41) {
return 8;
}
}
else if (refDCval >= 70) {
if (tcor < 64) {
if (diffMVdiffVal < -12) {
if (Rs < 1215) {
return 4;
}
else if (Rs >= 1215) {
return 2;
}
}
else if (diffMVdiffVal >= -12) {
if (Cs < 1068) {
if (SC < 2171) {
if (Cs < 761) {
return 4;
}
else if (Cs >= 761) {
return 2;
}
}
else if (SC >= 2171) {
return 1;
}
}
else if (Cs >= 1068) {
if (tcor < 52) {
return 2;
}
else if (tcor >= 52) {
return 4;
}
}
}
}
else if (tcor >= 64) {
if (MVDiff < 85) {
return 4;
}
else if (MVDiff >= 85) {
if (Rs < 908) {
return 4;
}
else if (Rs >= 908) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 12) {
if (TSC < 11) {
if (Cs < 936) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
if (MVDiff < 30) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
return 2;
}
}
}
else if (MVDiff >= 30) {
if (Rs < 2051) {
return 8;
}
else if (Rs >= 2051) {
return 1;
}
}
}
}
else if (Cs >= 936) {
if (ssDCval < 68) {
if (ssDCval < 42) {
return 1;
}
else if (ssDCval >= 42) {
if (RsDiff < 16) {
return 2;
}
else if (RsDiff >= 16) {
return 8;
}
}
}
else if (ssDCval >= 68) {
if (RsCsDiff < 239) {
if (RsCsDiff < 217) {
if (Rs < 1749) {
return 4;
}
else if (Rs >= 1749) {
if (Rs < 2070) {
return 8;
}
else if (Rs >= 2070) {
return 4;
}
}
}
else if (RsCsDiff >= 217) {
return 8;
}
}
else if (RsCsDiff >= 239) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
if (Rs < 2149) {
return 4;
}
else if (Rs >= 2149) {
return 8;
}
}
}
}
}
}
else if (TSC >= 11) {
if (mu_mv_mag_sq < 21) {
if (RsCsDiff < 223) {
if (CsDiff < 7) {
if (SC < 3546) {
return 1;
}
else if (SC >= 3546) {
return 2;
}
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (RsCsDiff >= 223) {
if (TSC < 13) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (TSC >= 13) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (SC < 3323) {
return 4;
}
else if (SC >= 3323) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (Rs < 2130) {
if (refDCval < 73) {
if (Cs < 740) {
return 2;
}
else if (Cs >= 740) {
if (negBalance < 51) {
if (Rs < 1345) {
if (Cs < 954) {
return 2;
}
else if (Cs >= 954) {
return 4;
}
}
else if (Rs >= 1345) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
if (Rs < 2092) {
return 2;
}
else if (Rs >= 2092) {
return 1;
}
}
}
}
else if (negBalance >= 51) {
if (SC < 2218) {
return 8;
}
else if (SC >= 2218) {
return 4;
}
}
}
}
else if (refDCval >= 73) {
if (MVDiff < 15) {
if (Cs < 924) {
if (AFD < 23) {
return 1;
}
else if (AFD >= 23) {
return 4;
}
}
else if (Cs >= 924) {
if (posBalance < 60) {
if (Cs < 1194) {
return 4;
}
else if (Cs >= 1194) {
if (Rs < 2000) {
return 1;
}
else if (Rs >= 2000) {
return 8;
}
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
else if (MVDiff >= 15) {
if (tcor < 86) {
if (negBalance < 47) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 1554) {
return 8;
}
else if (Rs >= 1554) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
return 2;
}
}
}
}
else if (negBalance >= 47) {
if (RsDiff < 12) {
if (posBalance < 60) {
if (Cs < 998) {
if (TSC < 8) {
if (SC < 2392) {
return 8;
}
else if (SC >= 2392) {
return 2;
}
}
else if (TSC >= 8) {
if (ssDCval < 87) {
return 4;
}
else if (ssDCval >= 87) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 4;
}
}
}
}
else if (Cs >= 998) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (Rs < 1026) {
return 1;
}
else if (Rs >= 1026) {
return 2;
}
}
}
}
else if (posBalance >= 60) {
if (AFD < 25) {
if (RsCsDiff < 157) {
return 1;
}
else if (RsCsDiff >= 157) {
if (Rs < 1381) {
return 8;
}
else if (Rs >= 1381) {
return 4;
}
}
}
else if (AFD >= 25) {
if (Rs < 1040) {
return 8;
}
else if (Rs >= 1040) {
return 2;
}
}
}
}
else if (RsDiff >= 12) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (Rs < 1554) {
return 2;
}
else if (Rs >= 1554) {
return 4;
}
}
}
}
}
else if (tcor >= 86) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 1;
}
}
}
}
}
else if (Rs >= 2130) {
if (MVDiff < 26) {
if (negBalance < 42) {
if (RsCsDiff < 117) {
if (MVDiff < 14) {
if (Rs < 2571) {
if (diffRsCsdiff < 35) {
return 2;
}
else if (diffRsCsdiff >= 35) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
return 1;
}
}
}
else if (Rs >= 2571) {
return 4;
}
}
else if (MVDiff >= 14) {
if (Cs < 1019) {
return 1;
}
else if (Cs >= 1019) {
return 2;
}
}
}
else if (RsCsDiff >= 117) {
if (Cs < 803) {
return 2;
}
else if (Cs >= 803) {
if (Cs < 1030) {
return 4;
}
else if (Cs >= 1030) {
if (Rs < 2426) {
return 4;
}
else if (Rs >= 2426) {
return 2;
}
}
}
}
}
else if (negBalance >= 42) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
if (MVDiff < 3) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 1;
}
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
else if (MVDiff >= 26) {
if (negBalance < 46) {
if (Cs < 611) {
return 4;
}
else if (Cs >= 611) {
return 2;
}
}
else if (negBalance >= 46) {
if (ssDCval < 86) {
return 1;
}
else if (ssDCval >= 86) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (tcor < 68) {
if (mcTcor < 66) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
if (RsCsDiff < 233) {
if (Rs < 2303) {
return 1;
}
else if (Rs >= 2303) {
return 2;
}
}
else if (RsCsDiff >= 233) {
if (Rs < 1236) {
return 8;
}
else if (Rs >= 1236) {
return 4;
}
}
}
}
else if (mcTcor >= 66) {
if (MVDiff < 81) {
if (diffAFD < 2) {
if (AFD < 32) {
if (refDCval < 98) {
if (TSC < 10) {
if (MVDiff < 36) {
return 4;
}
else if (MVDiff >= 36) {
if (Rs < 1673) {
return 1;
}
else if (Rs >= 1673) {
return 8;
}
}
}
else if (TSC >= 10) {
if (Cs < 1124) {
return 1;
}
else if (Cs >= 1124) {
return 8;
}
}
}
else if (refDCval >= 98) {
if (ssDCval < 115) {
if (ssDCval < 97) {
if (Rs < 1548) {
return 4;
}
else if (Rs >= 1548) {
return 8;
}
}
else if (ssDCval >= 97) {
return 4;
}
}
else if (ssDCval >= 115) {
if (tcor < 46) {
if (Rs < 700) {
return 8;
}
else if (Rs >= 700) {
return 1;
}
}
else if (tcor >= 46) {
if (Rs < 2244) {
if (MVDiff < 30) {
if (Cs < 818) {
if (Rs < 1804) {
return 4;
}
else if (Rs >= 1804) {
return 8;
}
}
else if (Cs >= 818) {
return 8;
}
}
else if (MVDiff >= 30) {
return 4;
}
}
else if (Rs >= 2244) {
return 8;
}
}
}
}
}
else if (AFD >= 32) {
if (Cs < 1074) {
return 2;
}
else if (Cs >= 1074) {
return 4;
}
}
}
else if (diffAFD >= 2) {
if (RsCsDiff < 121) {
if (Rs < 1846) {
return 1;
}
else if (Rs >= 1846) {
return 2;
}
}
else if (RsCsDiff >= 121) {
if (diffMVdiffVal < 10) {
if (tcor < 51) {
if (refDCval < 90) {
return 8;
}
else if (refDCval >= 90) {
return 4;
}
}
else if (tcor >= 51) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (diffMVdiffVal >= 10) {
if (AFD < 30) {
if (mcTcor < 77) {
if (RsCsDiff < 156) {
return 8;
}
else if (RsCsDiff >= 156) {
return 1;
}
}
else if (mcTcor >= 77) {
return 8;
}
}
else if (AFD >= 30) {
return 1;
}
}
}
}
}
else if (MVDiff >= 81) {
if (posBalance < 74) {
if (CsDiff < 4) {
if (Rs < 1928) {
return 8;
}
else if (Rs >= 1928) {
return 2;
}
}
else if (CsDiff >= 4) {
if (AFD < 45) {
if (diffRsCsdiff < -65) {
if (Rs < 1935) {
if (Rs < 1112) {
return 8;
}
else if (Rs >= 1112) {
return 1;
}
}
else if (Rs >= 1935) {
return 2;
}
}
else if (diffRsCsdiff >= -65) {
if (CsDiff < 6) {
if (Rs < 1686) {
return 4;
}
else if (Rs >= 1686) {
return 1;
}
}
else if (CsDiff >= 6) {
return 1;
}
}
}
else if (AFD >= 45) {
return 2;
}
}
}
else if (posBalance >= 74) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
return 4;
}
}
}
}
}
else if (tcor >= 68) {
if (ssDCval < 111) {
if (Cs < 1051) {
if (AFD < 16) {
if (TSC < 10) {
if (tcor < 87) {
return 2;
}
else if (tcor >= 87) {
if (SC < 2696) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (SC >= 2696) {
if (SC < 3009) {
return 2;
}
else if (SC >= 3009) {
return 4;
}
}
}
}
else if (TSC >= 10) {
if (refDCval < 94) {
return 4;
}
else if (refDCval >= 94) {
return 1;
}
}
}
else if (AFD >= 16) {
if (RsCsDiff < 167) {
if (posBalance < 61) {
if (MVDiff < 33) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 2;
}
}
else if (MVDiff >= 33) {
if (Cs < 581) {
return 4;
}
else if (Cs >= 581) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (SC < 2042) {
return 8;
}
else if (SC >= 2042) {
return 1;
}
}
}
else if (RsCsDiff >= 167) {
if (MVDiff < 33) {
if (SC < 3127) {
if (refDCval < 105) {
if (gchDC < 3) {
if (Cs < 690) {
return 1;
}
else if (Cs >= 690) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 1;
}
}
}
else if (gchDC >= 3) {
return 2;
}
}
else if (refDCval >= 105) {
return 2;
}
}
else if (SC >= 3127) {
return 2;
}
}
else if (MVDiff >= 33) {
if (mu_mv_mag_sq < 43) {
if (diffAFD < 4) {
if (posBalance < 58) {
if (Rs < 2487) {
if (Cs < 931) {
return 8;
}
else if (Cs >= 931) {
return 2;
}
}
else if (Rs >= 2487) {
return 2;
}
}
else if (posBalance >= 58) {
return 1;
}
}
else if (diffAFD >= 4) {
return 1;
}
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 1367) {
return 4;
}
else if (Rs >= 1367) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1051) {
if (tcor < 82) {
if (Rs < 1639) {
if (AFD < 22) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (AFD >= 22) {
if (Rs < 1156) {
return 1;
}
else if (Rs >= 1156) {
return 4;
}
}
}
else if (Rs >= 1639) {
if (mu_mv_mag_sq < 64) {
if (diffRsCsdiff < 45) {
if (mcTcor < 93) {
return 2;
}
else if (mcTcor >= 93) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 1923) {
return 8;
}
else if (Rs >= 1923) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 45) {
if (Rs < 2154) {
if (Rs < 1890) {
return 8;
}
else if (Rs >= 1890) {
return 1;
}
}
else if (Rs >= 2154) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 64) {
return 4;
}
}
}
else if (tcor >= 82) {
if (MVDiff < 10) {
if (Cs < 1271) {
if (Rs < 1782) {
return 1;
}
else if (Rs >= 1782) {
return 4;
}
}
else if (Cs >= 1271) {
if (Cs < 1291) {
return 2;
}
else if (Cs >= 1291) {
return 8;
}
}
}
else if (MVDiff >= 10) {
if (ssDCval < 92) {
if (Rs < 2360) {
return 4;
}
else if (Rs >= 2360) {
return 8;
}
}
else if (ssDCval >= 92) {
return 8;
}
}
}
}
}
else if (ssDCval >= 111) {
if (Rs < 1815) {
if (Cs < 1036) {
if (Cs < 579) {
if (mcTcor < 94) {
if (Rs < 1712) {
return 8;
}
else if (Rs >= 1712) {
return 1;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (Cs >= 579) {
if (TSC < 10) {
if (mu_mv_mag_sq < 7) {
if (CsDiff < 5) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (CsDiff >= 5) {
if (negBalance < 49) {
if (ssDCval < 147) {
return 2;
}
else if (ssDCval >= 147) {
return 8;
}
}
else if (negBalance >= 49) {
if (ssDCval < 115) {
return 2;
}
else if (ssDCval >= 115) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 15) {
if (Rs < 1727) {
return 8;
}
else if (Rs >= 1727) {
return 1;
}
}
else if (AFD >= 15) {
if (Rs < 914) {
return 8;
}
else if (Rs >= 914) {
if (tcor < 88) {
return 2;
}
else if (tcor >= 88) {
if (Rs < 1194) {
return 2;
}
else if (Rs >= 1194) {
return 8;
}
}
}
}
}
}
else if (TSC >= 10) {
if (MVDiff < 47) {
return 1;
}
else if (MVDiff >= 47) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (RsDiff < 14) {
return 2;
}
else if (RsDiff >= 14) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1036) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Rs < 1590) {
return 4;
}
else if (Rs >= 1590) {
return 2;
}
}
}
}
else if (Rs >= 1815) {
if (mu_mv_mag_sq < 46) {
if (Cs < 834) {
if (tcor < 72) {
if (Rs < 2360) {
if (Cs < 734) {
return 2;
}
else if (Cs >= 734) {
return 8;
}
}
else if (Rs >= 2360) {
return 1;
}
}
else if (tcor >= 72) {
if (mcTcor < 95) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (SC < 2524) {
return 2;
}
else if (SC >= 2524) {
return 1;
}
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
else if (Cs >= 834) {
if (MVDiff < 17) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
if (Rs < 1833) {
return 2;
}
else if (Rs >= 1833) {
return 8;
}
}
}
else if (MVDiff >= 17) {
if (TSC < 11) {
if (CsDiff < 7) {
if (diffRsCsdiff < -32) {
return 1;
}
else if (diffRsCsdiff >= -32) {
if (SC < 2992) {
return 1;
}
else if (SC >= 2992) {
return 8;
}
}
}
else if (CsDiff >= 7) {
if (Rs < 2325) {
return 2;
}
else if (Rs >= 2325) {
return 1;
}
}
}
else if (TSC >= 11) {
if (MVDiff < 66) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
return 1;
}
}
else if (MVDiff >= 66) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (refDCval < 115) {
return 8;
}
else if (refDCval >= 115) {
if (Rs < 2568) {
return 2;
}
else if (Rs >= 2568) {
if (Rs < 2573) {
return 8;
}
else if (Rs >= 2573) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2599) {
if (refDCval < 109) {
if (RsDiff < 5) {
if (SC < 4936) {
if (RsDiff < 2) {
if (ssDCval < 92) {
if (Cs < 849) {
if (Cs < 448) {
return 8;
}
else if (Cs >= 448) {
if (Cs < 676) {
return 1;
}
else if (Cs >= 676) {
return 4;
}
}
}
else if (Cs >= 849) {
return 8;
}
}
else if (ssDCval >= 92) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
if (Rs < 3759) {
return 8;
}
else if (Rs >= 3759) {
return 2;
}
}
}
}
else if (RsDiff >= 2) {
if (MVDiff < 5) {
if (Cs < 1097) {
if (tcor < 97) {
if (CsDiff < 3) {
if (diffRsCsdiff < -8) {
return 2;
}
else if (diffRsCsdiff >= -8) {
if (diffRsCsdiff < -4) {
return 4;
}
else if (diffRsCsdiff >= -4) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 2773) {
return 2;
}
else if (Rs >= 2773) {
return 4;
}
}
}
else if (tcor >= 97) {
if (diffRsCsdiff < -2) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (Rs < 3768) {
if (refDCval < 63) {
return 4;
}
else if (refDCval >= 63) {
return 2;
}
}
else if (Rs >= 3768) {
return 4;
}
}
}
else if (diffRsCsdiff >= -2) {
if (Rs < 3380) {
if (posBalance < 50) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 3014) {
return 2;
}
else if (Rs >= 3014) {
return 8;
}
}
}
else if (posBalance >= 50) {
return 2;
}
}
else if (Rs >= 3380) {
return 2;
}
}
}
}
else if (Cs >= 1097) {
if (diffTSC < -1) {
if (mu_mv_mag_sq < 111) {
return 1;
}
else if (mu_mv_mag_sq >= 111) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
if (Rs < 3082) {
if (Rs < 2876) {
return 4;
}
else if (Rs >= 2876) {
return 1;
}
}
else if (Rs >= 3082) {
return 2;
}
}
}
}
else if (diffTSC >= -1) {
if (ssDCval < 103) {
if (ssDCval < 65) {
if (Rs < 3041) {
return 4;
}
else if (Rs >= 3041) {
return 8;
}
}
else if (ssDCval >= 65) {
if (tcor < 97) {
if (CsDiff < 4) {
if (Rs < 3366) {
if (Rs < 3073) {
return 4;
}
else if (Rs >= 3073) {
return 8;
}
}
else if (Rs >= 3366) {
return 8;
}
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (tcor >= 97) {
return 4;
}
}
}
else if (ssDCval >= 103) {
if (Rs < 3025) {
return 8;
}
else if (Rs >= 3025) {
return 2;
}
}
}
}
}
else if (MVDiff >= 5) {
if (CsDiff < 1) {
if (Cs < 259) {
return 1;
}
else if (Cs >= 259) {
return 2;
}
}
else if (CsDiff >= 1) {
if (tcor < 92) {
if (Rs < 2935) {
return 1;
}
else if (Rs >= 2935) {
if (Rs < 3178) {
return 4;
}
else if (Rs >= 3178) {
return 4;
}
}
}
else if (tcor >= 92) {
if (TSC < 5) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
if (refDCval < 65) {
if (Rs < 3111) {
return 4;
}
else if (Rs >= 3111) {
return 8;
}
}
else if (refDCval >= 65) {
return 8;
}
}
}
else if (TSC >= 5) {
return 2;
}
}
}
}
}
}
else if (SC >= 4936) {
if (Cs < 1282) {
if (Rs < 4554) {
if (negBalance < 18) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
return 4;
}
}
else if (negBalance >= 18) {
return 4;
}
}
else if (Rs >= 4554) {
return 2;
}
}
else if (Cs >= 1282) {
if (Rs < 3892) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 2;
}
}
else if (Rs >= 3892) {
return 8;
}
}
}
}
else if (RsDiff >= 5) {
if (ssDCval < 90) {
if (posBalance < 47) {
if (negBalance < 51) {
if (Rs < 3817) {
if (refDCval < 66) {
if (CsDiff < 6) {
if (MVDiff < 4) {
if (tcor < 93) {
return 2;
}
else if (tcor >= 93) {
if (Rs < 2927) {
return 2;
}
else if (Rs >= 2927) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (posBalance < 45) {
if (Cs < 1270) {
if (refDCval < 65) {
return 4;
}
else if (refDCval >= 65) {
return 1;
}
}
else if (Cs >= 1270) {
if (Rs < 3116) {
return 4;
}
else if (Rs >= 3116) {
if (Rs < 3431) {
return 8;
}
else if (Rs >= 3431) {
return 2;
}
}
}
}
else if (posBalance >= 45) {
if (mu_mv_mag_sq < 43) {
if (diffRsCsdiff < 15) {
return 2;
}
else if (diffRsCsdiff >= 15) {
return 4;
}
}
else if (mu_mv_mag_sq >= 43) {
return 4;
}
}
}
}
else if (CsDiff >= 6) {
if (negBalance < 33) {
return 4;
}
else if (negBalance >= 33) {
return 2;
}
}
}
else if (refDCval >= 66) {
if (Rs < 2869) {
if (Cs < 1054) {
if (mu_mv_mag_sq < 7) {
if (Cs < 478) {
return 2;
}
else if (Cs >= 478) {
if (SC < 3762) {
return 1;
}
else if (SC >= 3762) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 9) {
if (ssDCval < 84) {
if (Cs < 599) {
return 8;
}
else if (Cs >= 599) {
return 1;
}
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (AFD >= 9) {
if (Cs < 241) {
return 1;
}
else if (Cs >= 241) {
return 4;
}
}
}
}
else if (Cs >= 1054) {
if (SC < 3964) {
if (RsDiff < 11) {
return 2;
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (SC >= 3964) {
if (mu_mv_mag_sq < 69) {
if (Rs < 2834) {
return 1;
}
else if (Rs >= 2834) {
if (Rs < 2836) {
return 4;
}
else if (Rs >= 2836) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (Rs < 2820) {
return 1;
}
else if (Rs >= 2820) {
return 2;
}
}
}
}
}
else if (Rs >= 2869) {
if (mu_mv_mag_sq < 81) {
if (posBalance < 45) {
if (ssDCval < 77) {
if (SC < 4704) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (SC < 4184) {
if (TSC < 6) {
if (Cs < 932) {
return 4;
}
else if (Cs >= 932) {
return 1;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (SC >= 4184) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
else if (SC >= 4704) {
if (Rs < 3541) {
return 4;
}
else if (Rs >= 3541) {
return 1;
}
}
}
else if (ssDCval >= 77) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
if (diffRsCsdiff < 3) {
if (Rs < 3414) {
return 4;
}
else if (Rs >= 3414) {
return 4;
}
}
else if (diffRsCsdiff >= 3) {
return 8;
}
}
}
}
else if (posBalance >= 45) {
if (negBalance < 42) {
if (SC < 4611) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
return 2;
}
}
else if (SC >= 4611) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (negBalance >= 42) {
if (tcor < 86) {
return 4;
}
else if (tcor >= 86) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 33) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (negBalance >= 33) {
return 2;
}
}
}
}
}
else if (Rs >= 3817) {
if (mu_mv_mag_sq < 104) {
if (Rs < 5671) {
return 4;
}
else if (Rs >= 5671) {
return 8;
}
}
else if (mu_mv_mag_sq >= 104) {
return 1;
}
}
}
else if (negBalance >= 51) {
if (mcTcor < 93) {
if (diffMVdiffVal < 13) {
if (Cs < 873) {
return 2;
}
else if (Cs >= 873) {
return 8;
}
}
else if (diffMVdiffVal >= 13) {
return 4;
}
}
else if (mcTcor >= 93) {
if (Rs < 4225) {
return 2;
}
else if (Rs >= 4225) {
return 4;
}
}
}
}
else if (posBalance >= 47) {
if (CsDiff < 3) {
if (Rs < 3383) {
if (ssDCval < 76) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (Cs < 694) {
return 4;
}
else if (Cs >= 694) {
return 2;
}
}
}
else if (ssDCval >= 76) {
if (AFD < 10) {
if (diffRsCsdiff < -24) {
if (Cs < 816) {
if (Rs < 2868) {
return 1;
}
else if (Rs >= 2868) {
return 4;
}
}
else if (Cs >= 816) {
return 2;
}
}
else if (diffRsCsdiff >= -24) {
return 1;
}
}
else if (AFD >= 10) {
if (RsDiff < 12) {
if (Rs < 2767) {
return 4;
}
else if (Rs >= 2767) {
return 8;
}
}
else if (RsDiff >= 12) {
return 2;
}
}
}
}
else if (Rs >= 3383) {
if (posBalance < 62) {
if (Cs < 1273) {
return 4;
}
else if (Cs >= 1273) {
return 2;
}
}
else if (posBalance >= 62) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 63) {
if (MVDiff < 17) {
if (Cs < 992) {
if (tcor < 84) {
return 4;
}
else if (tcor >= 84) {
return 2;
}
}
else if (Cs >= 992) {
if (Rs < 3325) {
return 4;
}
else if (Rs >= 3325) {
if (Cs < 1178) {
if (Rs < 7852) {
return 4;
}
else if (Rs >= 7852) {
return 2;
}
}
else if (Cs >= 1178) {
return 2;
}
}
}
}
else if (MVDiff >= 17) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
if (Rs < 2748) {
return 2;
}
else if (Rs >= 2748) {
return 4;
}
}
}
}
else if (refDCval >= 63) {
if (mcTcor < 90) {
if (Rs < 3410) {
if (CsDiff < 8) {
if (TSC < 12) {
if (TSC < 10) {
if (tcor < 83) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Cs < 1159) {
return 2;
}
else if (Cs >= 1159) {
return 8;
}
}
}
else if (tcor >= 83) {
return 1;
}
}
else if (TSC >= 10) {
return 2;
}
}
else if (TSC >= 12) {
if (diffRsCsdiff < 127) {
return 1;
}
else if (diffRsCsdiff >= 127) {
if (Rs < 2922) {
return 2;
}
else if (Rs >= 2922) {
return 8;
}
}
}
}
else if (CsDiff >= 8) {
if (Rs < 2666) {
return 8;
}
else if (Rs >= 2666) {
return 1;
}
}
}
else if (Rs >= 3410) {
if (Cs < 1092) {
return 4;
}
else if (Cs >= 1092) {
return 2;
}
}
}
else if (mcTcor >= 90) {
if (posBalance < 52) {
if (negBalance < 41) {
if (mu_mv_mag_sq < 78) {
if (tcor < 92) {
if (CsDiff < 7) {
if (posBalance < 48) {
if (TSC < 8) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
else if (TSC >= 8) {
return 2;
}
}
else if (posBalance >= 48) {
return 4;
}
}
else if (CsDiff >= 7) {
if (Rs < 2947) {
return 1;
}
else if (Rs >= 2947) {
return 2;
}
}
}
else if (tcor >= 92) {
if (Rs < 3516) {
if (RsDiff < 8) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 1;
}
}
else if (RsDiff >= 8) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 4;
}
}
}
else if (Rs >= 3516) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 78) {
return 2;
}
}
else if (negBalance >= 41) {
if (tcor < 82) {
if (negBalance < 51) {
if (Cs < 1070) {
return 4;
}
else if (Cs >= 1070) {
return 2;
}
}
else if (negBalance >= 51) {
return 2;
}
}
else if (tcor >= 82) {
if (mcTcor < 99) {
if (negBalance < 42) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (negBalance >= 42) {
return 2;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
}
else if (posBalance >= 52) {
if (Rs < 4286) {
if (Cs < 651) {
if (Rs < 3309) {
if (negBalance < 34) {
if (Rs < 3069) {
return 4;
}
else if (Rs >= 3069) {
return 2;
}
}
else if (negBalance >= 34) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Cs < 338) {
return 1;
}
else if (Cs >= 338) {
return 2;
}
}
}
}
else if (Rs >= 3309) {
if (tcor < 83) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 2;
}
}
}
else if (tcor >= 83) {
return 2;
}
}
}
else if (Cs >= 651) {
if (mcTcor < 99) {
if (tcor < 79) {
if (SC < 4172) {
return 2;
}
else if (SC >= 4172) {
if (Rs < 3985) {
return 1;
}
else if (Rs >= 3985) {
return 2;
}
}
}
else if (tcor >= 79) {
if (ssDCval < 65) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (Cs < 1082) {
return 2;
}
else if (Cs >= 1082) {
if (Rs < 3070) {
return 4;
}
else if (Rs >= 3070) {
return 2;
}
}
}
}
else if (ssDCval >= 65) {
if (negBalance < 36) {
if (posBalance < 58) {
if (mu_mv_mag_sq < 65) {
return 1;
}
else if (mu_mv_mag_sq >= 65) {
return 2;
}
}
else if (posBalance >= 58) {
return 2;
}
}
else if (negBalance >= 36) {
return 2;
}
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
else if (Rs >= 4286) {
if (MVDiff < 12) {
if (Rs < 4838) {
return 2;
}
else if (Rs >= 4838) {
return 4;
}
}
else if (MVDiff >= 12) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 90) {
if (mcTcor < 71) {
if (tcor < 43) {
if (Rs < 3271) {
return 8;
}
else if (Rs >= 3271) {
if (Rs < 3594) {
return 4;
}
else if (Rs >= 3594) {
return 2;
}
}
}
else if (tcor >= 43) {
return 1;
}
}
else if (mcTcor >= 71) {
if (RsCsDiff < 183) {
if (mcTcor < 98) {
if (RsCsDiff < 90) {
if (negBalance < 35) {
if (mcTcor < 97) {
if (posBalance < 87) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (gchDC < 4) {
if (diffRsCsdiff < -15) {
return 4;
}
else if (diffRsCsdiff >= -15) {
if (ssDCval < 92) {
if (mu_mv_mag_sq < 57) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
if (Rs < 4128) {
return 4;
}
else if (Rs >= 4128) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (posBalance < 47) {
if (Rs < 4197) {
if (Rs < 4007) {
return 4;
}
else if (Rs >= 4007) {
return 4;
}
}
else if (Rs >= 4197) {
return 4;
}
}
else if (posBalance >= 47) {
if (Cs < 1165) {
return 4;
}
else if (Cs >= 1165) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 4244) {
return 8;
}
else if (Rs >= 4244) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 92) {
if (negBalance < 32) {
if (diffMVdiffVal < -2) {
if (MVDiff < 9) {
if (Rs < 3392) {
return 4;
}
else if (Rs >= 3392) {
if (Rs < 3931) {
if (Cs < 1029) {
return 4;
}
else if (Cs >= 1029) {
return 4;
}
}
else if (Rs >= 3931) {
return 4;
}
}
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
if (diffRsCsdiff < 6) {
if (Cs < 1092) {
if (posBalance < 39) {
if (Rs < 3445) {
return 8;
}
else if (Rs >= 3445) {
return 8;
}
}
else if (posBalance >= 39) {
if (tcor < 94) {
if (Rs < 3739) {
if (Rs < 3577) {
return 8;
}
else if (Rs >= 3577) {
return 8;
}
}
else if (Rs >= 3739) {
if (Rs < 3967) {
return 8;
}
else if (Rs >= 3967) {
return 8;
}
}
}
else if (tcor >= 94) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
return 4;
}
}
}
}
else if (Cs >= 1092) {
return 8;
}
}
else if (diffRsCsdiff >= 6) {
if (AFD < 8) {
if (Rs < 3104) {
return 8;
}
else if (Rs >= 3104) {
return 4;
}
}
else if (AFD >= 8) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Rs < 3770) {
return 4;
}
else if (Rs >= 3770) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (posBalance >= 87) {
return 2;
}
}
else if (mcTcor >= 97) {
if (Cs < 993) {
return 2;
}
else if (Cs >= 993) {
if (SC < 4943) {
return 8;
}
else if (SC >= 4943) {
return 2;
}
}
}
}
else if (negBalance >= 35) {
if (MVDiff < 26) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
if (Rs < 3614) {
return 8;
}
else if (Rs >= 3614) {
if (Rs < 3978) {
return 2;
}
else if (Rs >= 3978) {
return 4;
}
}
}
}
else if (MVDiff >= 26) {
if (refDCval < 105) {
if (TSC < 7) {
if (Rs < 3926) {
return 4;
}
else if (Rs >= 3926) {
return 8;
}
}
else if (TSC >= 7) {
if (Cs < 1021) {
if (Rs < 2677) {
if (Rs < 2607) {
return 1;
}
else if (Rs >= 2607) {
return 8;
}
}
else if (Rs >= 2677) {
return 1;
}
}
else if (Cs >= 1021) {
return 8;
}
}
}
else if (refDCval >= 105) {
return 8;
}
}
}
}
else if (RsCsDiff >= 90) {
if (MVDiff < 24) {
if (ssDCval < 106) {
if (TSC < 10) {
if (TSC < 7) {
if (AFD < 13) {
if (RsCsDiff < 94) {
return 8;
}
else if (RsCsDiff >= 94) {
return 2;
}
}
else if (AFD >= 13) {
return 8;
}
}
else if (TSC >= 7) {
if (gchDC < 2) {
if (Cs < 885) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
return 8;
}
}
else if (Cs >= 885) {
if (Rs < 3013) {
if (Cs < 1247) {
return 4;
}
else if (Cs >= 1247) {
return 8;
}
}
else if (Rs >= 3013) {
if (mcTcor < 95) {
if (Rs < 3923) {
return 8;
}
else if (Rs >= 3923) {
return 4;
}
}
else if (mcTcor >= 95) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (TSC >= 10) {
if (diffMVdiffVal < -8) {
if (Rs < 2839) {
return 2;
}
else if (Rs >= 2839) {
return 4;
}
}
else if (diffMVdiffVal >= -8) {
if (Rs < 2657) {
return 2;
}
else if (Rs >= 2657) {
return 8;
}
}
}
}
else if (ssDCval >= 106) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (MVDiff >= 24) {
if (TSC < 14) {
if (diffAFD < -3) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
if (Rs < 2664) {
return 8;
}
else if (Rs >= 2664) {
return 2;
}
}
}
else if (diffAFD >= -3) {
if (TSC < 8) {
if (posBalance < 59) {
if (Rs < 3275) {
return 4;
}
else if (Rs >= 3275) {
return 8;
}
}
else if (posBalance >= 59) {
return 2;
}
}
else if (TSC >= 8) {
if (RsCsDiff < 107) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
if (MVDiff < 42) {
return 1;
}
else if (MVDiff >= 42) {
return 8;
}
}
}
else if (RsCsDiff >= 107) {
return 8;
}
}
}
}
else if (TSC >= 14) {
if (Rs < 2790) {
return 2;
}
else if (Rs >= 2790) {
if (Cs < 678) {
return 1;
}
else if (Cs >= 678) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (Cs < 978) {
if (Rs < 2876) {
return 8;
}
else if (Rs >= 2876) {
return 2;
}
}
else if (Cs >= 978) {
if (RsCsDiff < 53) {
return 8;
}
else if (RsCsDiff >= 53) {
if (Rs < 3552) {
return 4;
}
else if (Rs >= 3552) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 183) {
if (mcTcor < 93) {
if (Cs < 731) {
if (MVDiff < 50) {
return 2;
}
else if (MVDiff >= 50) {
if (diffRsCsdiff < 125) {
if (AFD < 23) {
return 4;
}
else if (AFD >= 23) {
return 8;
}
}
else if (diffRsCsdiff >= 125) {
return 2;
}
}
}
else if (Cs >= 731) {
if (CsDiff < 7) {
if (negBalance < 51) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
return 2;
}
}
}
else if (negBalance >= 51) {
if (RsDiff < 25) {
if (AFD < 17) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (Rs < 3570) {
return 2;
}
else if (Rs >= 3570) {
return 4;
}
}
}
else if (AFD >= 17) {
if (RsCsDiff < 387) {
if (RsCsDiff < 190) {
if (Rs < 2878) {
return 4;
}
else if (Rs >= 2878) {
return 8;
}
}
else if (RsCsDiff >= 190) {
return 4;
}
}
else if (RsCsDiff >= 387) {
if (Cs < 1187) {
return 8;
}
else if (Cs >= 1187) {
return 4;
}
}
}
}
else if (RsDiff >= 25) {
return 2;
}
}
}
else if (CsDiff >= 7) {
if (AFD < 24) {
if (tcor < 77) {
if (refDCval < 92) {
if (Rs < 2662) {
return 8;
}
else if (Rs >= 2662) {
return 2;
}
}
else if (refDCval >= 92) {
return 8;
}
}
else if (tcor >= 77) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 1;
}
}
}
else if (AFD >= 24) {
if (Rs < 2613) {
return 4;
}
else if (Rs >= 2613) {
return 2;
}
}
}
}
}
else if (mcTcor >= 93) {
if (gchDC < 5) {
if (RsDiff < 15) {
if (ssDCval < 105) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
else if (ssDCval >= 105) {
if (Rs < 3165) {
return 8;
}
else if (Rs >= 3165) {
return 4;
}
}
}
else if (RsDiff >= 15) {
if (AFD < 28) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
}
}
}
else if (AFD >= 28) {
if (RsDiff < 17) {
return 1;
}
else if (RsDiff >= 17) {
return 2;
}
}
}
}
else if (gchDC >= 5) {
if (Cs < 566) {
if (Rs < 2870) {
return 8;
}
else if (Rs >= 2870) {
return 2;
}
}
else if (Cs >= 566) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (MVDiff < 7) {
if (SC < 4269) {
if (mcTcor < 94) {
if (Cs < 913) {
if (Rs < 3575) {
return 4;
}
else if (Rs >= 3575) {
return 2;
}
}
else if (Cs >= 913) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Cs < 573) {
if (CsDiff < 2) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
else if (CsDiff >= 2) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
}
else if (Cs >= 573) {
if (Rs < 2645) {
return 8;
}
else if (Rs >= 2645) {
if (mcTcor < 96) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
}
}
}
else if (SC >= 4269) {
if (refDCval < 138) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (TSC < 6) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
if (Rs < 3943) {
if (Rs < 3613) {
return 2;
}
else if (Rs >= 3613) {
return 8;
}
}
else if (Rs >= 3943) {
return 2;
}
}
}
else if (TSC >= 6) {
if (diffRsCsdiff < -132) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
else if (diffRsCsdiff >= -132) {
return 2;
}
}
}
}
else if (refDCval >= 138) {
if (CsDiff < 3) {
if (Rs < 3963) {
if (mu_mv_mag_sq < 119) {
if (Cs < 1210) {
return 2;
}
else if (Cs >= 1210) {
if (Rs < 3750) {
return 2;
}
else if (Rs >= 3750) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 3670) {
return 2;
}
else if (Rs >= 3670) {
return 1;
}
}
}
else if (Rs >= 3963) {
return 2;
}
}
else if (CsDiff >= 3) {
if (Rs < 3889) {
if (ssDCval < 142) {
return 2;
}
else if (ssDCval >= 142) {
if (Cs < 1107) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (Cs >= 1107) {
return 1;
}
}
}
else if (Rs >= 3889) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (Rs < 4159) {
return 2;
}
else if (Rs >= 4159) {
return 1;
}
}
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (refDCval < 134) {
if (Rs < 3448) {
if (RsCsDiff < 209) {
if (mu_mv_mag_sq < 26) {
if (MVDiff < 17) {
if (mu_mv_mag_sq < 24) {
if (posBalance < 45) {
if (Cs < 360) {
return 1;
}
else if (Cs >= 360) {
return 8;
}
}
else if (posBalance >= 45) {
if (CsDiff < 4) {
if (Rs < 2692) {
return 1;
}
else if (Rs >= 2692) {
if (Rs < 3341) {
return 2;
}
else if (Rs >= 3341) {
return 1;
}
}
}
else if (CsDiff >= 4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 24) {
return 4;
}
}
else if (MVDiff >= 17) {
if (diffAFD < -4) {
if (Cs < 437) {
if (MVDiff < 83) {
return 2;
}
else if (MVDiff >= 83) {
if (Rs < 2931) {
return 1;
}
else if (Rs >= 2931) {
return 8;
}
}
}
else if (Cs >= 437) {
return 1;
}
}
else if (diffAFD >= -4) {
if (tcor < 87) {
if (Rs < 2605) {
return 2;
}
else if (Rs >= 2605) {
return 8;
}
}
else if (tcor >= 87) {
if (gchDC < 4) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 2665) {
return 1;
}
else if (Rs >= 2665) {
return 8;
}
}
}
else if (gchDC >= 4) {
if (diffMVdiffVal < 53) {
return 1;
}
else if (diffMVdiffVal >= 53) {
if (Rs < 2672) {
return 2;
}
else if (Rs >= 2672) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (ssDCval < 127) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 54) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (mu_mv_mag_sq >= 54) {
return 2;
}
}
else if (CsDiff >= 3) {
if (Cs < 805) {
if (Cs < 556) {
return 8;
}
else if (Cs >= 556) {
if (Rs < 2804) {
return 1;
}
else if (Rs >= 2804) {
return 2;
}
}
}
else if (Cs >= 805) {
return 8;
}
}
}
else if (ssDCval >= 127) {
if (diffAFD < -6) {
return 1;
}
else if (diffAFD >= -6) {
if (Cs < 1063) {
return 2;
}
else if (Cs >= 1063) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 209) {
if (tcor < 77) {
if (CsDiff < 8) {
if (AFD < 26) {
if (Cs < 615) {
return 2;
}
else if (Cs >= 615) {
if (refDCval < 118) {
if (Rs < 3339) {
return 8;
}
else if (Rs >= 3339) {
if (Cs < 992) {
return 2;
}
else if (Cs >= 992) {
return 8;
}
}
}
else if (refDCval >= 118) {
if (Rs < 2762) {
return 1;
}
else if (Rs >= 2762) {
return 2;
}
}
}
}
else if (AFD >= 26) {
if (diffRsCsdiff < 62) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 2;
}
}
else if (diffRsCsdiff >= 62) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
else if (CsDiff >= 8) {
if (tcor < 39) {
return 2;
}
else if (tcor >= 39) {
if (Cs < 1274) {
return 1;
}
else if (Cs >= 1274) {
if (Rs < 2956) {
return 4;
}
else if (Rs >= 2956) {
return 1;
}
}
}
}
}
else if (tcor >= 77) {
if (CsDiff < 4) {
if (diffTSC < -1) {
if (SC < 3228) {
return 4;
}
else if (SC >= 3228) {
return 1;
}
}
else if (diffTSC >= -1) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
if (MVDiff < 39) {
return 2;
}
else if (MVDiff >= 39) {
if (Rs < 2925) {
return 2;
}
else if (Rs >= 2925) {
return 8;
}
}
}
}
}
else if (CsDiff >= 4) {
if (Cs < 1250) {
if (Rs < 2626) {
if (Rs < 2608) {
return 1;
}
else if (Rs >= 2608) {
return 2;
}
}
else if (Rs >= 2626) {
return 1;
}
}
else if (Cs >= 1250) {
return 8;
}
}
}
}
}
else if (Rs >= 3448) {
if (MVDiff < 23) {
if (TSC < 7) {
if (Rs < 4168) {
if (MVDiff < 12) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 1276) {
if (SC < 4419) {
return 8;
}
else if (SC >= 4419) {
return 2;
}
}
else if (Cs >= 1276) {
return 8;
}
}
}
else if (MVDiff >= 12) {
if (Cs < 520) {
return 2;
}
else if (Cs >= 520) {
return 8;
}
}
}
else if (Rs >= 4168) {
return 2;
}
}
else if (TSC >= 7) {
if (Rs < 4744) {
if (Rs < 4594) {
if (diffMVdiffVal < -44) {
return 1;
}
else if (diffMVdiffVal >= -44) {
if (AFD < 11) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
if (Rs < 4055) {
return 8;
}
else if (Rs >= 4055) {
return 1;
}
}
}
else if (AFD >= 11) {
if (AFD < 33) {
return 2;
}
else if (AFD >= 33) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
}
}
}
else if (Rs >= 4594) {
if (ssDCval < 116) {
return 2;
}
else if (ssDCval >= 116) {
if (Cs < 741) {
return 2;
}
else if (Cs >= 741) {
return 1;
}
}
}
}
else if (Rs >= 4744) {
return 2;
}
}
}
else if (MVDiff >= 23) {
if (Cs < 621) {
if (Rs < 3676) {
if (tcor < 82) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (tcor >= 82) {
return 1;
}
}
else if (Rs >= 3676) {
return 2;
}
}
else if (Cs >= 621) {
if (RsCsDiff < 411) {
if (posBalance < 47) {
if (AFD < 23) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
if (Rs < 5193) {
return 1;
}
else if (Rs >= 5193) {
if (Rs < 5681) {
return 2;
}
else if (Rs >= 5681) {
return 8;
}
}
}
}
else if (AFD >= 23) {
return 1;
}
}
else if (posBalance >= 47) {
if (negBalance < 39) {
if (Rs < 4747) {
return 2;
}
else if (Rs >= 4747) {
return 1;
}
}
else if (negBalance >= 39) {
if (CsDiff < 6) {
if (negBalance < 64) {
return 8;
}
else if (negBalance >= 64) {
return 1;
}
}
else if (CsDiff >= 6) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
if (Rs < 4320) {
return 2;
}
else if (Rs >= 4320) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 411) {
if (refDCval < 119) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
return 2;
}
}
else if (refDCval >= 119) {
if (TSC < 18) {
if (AFD < 22) {
if (Cs < 851) {
return 2;
}
else if (Cs >= 851) {
return 1;
}
}
else if (AFD >= 22) {
if (mu_mv_mag_sq < 9) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (diffMVdiffVal < -8) {
return 8;
}
else if (diffMVdiffVal >= -8) {
if (MVDiff < 51) {
if (Rs < 3667) {
return 8;
}
else if (Rs >= 3667) {
if (AFD < 31) {
return 2;
}
else if (AFD >= 31) {
return 8;
}
}
}
else if (MVDiff >= 51) {
if (diffMVdiffVal < 25) {
return 8;
}
else if (diffMVdiffVal >= 25) {
if (Rs < 5520) {
return 2;
}
else if (Rs >= 5520) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
return 2;
}
}
}
else if (TSC >= 18) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 134) {
if (mcTcor < 94) {
if (AFD < 19) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
if (posBalance < 55) {
if (Cs < 700) {
return 2;
}
else if (Cs >= 700) {
return 1;
}
}
else if (posBalance >= 55) {
return 4;
}
}
}
else if (AFD >= 19) {
if (CsDiff < 5) {
if (RsDiff < 19) {
if (MVDiff < 29) {
return 2;
}
else if (MVDiff >= 29) {
if (diffRsCsdiff < 109) {
return 8;
}
else if (diffRsCsdiff >= 109) {
return 1;
}
}
}
else if (RsDiff >= 19) {
return 2;
}
}
else if (CsDiff >= 5) {
if (Rs < 4749) {
if (MVDiff < 95) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
if (refDCval < 135) {
return 2;
}
else if (refDCval >= 135) {
return 1;
}
}
}
else if (MVDiff >= 95) {
return 8;
}
}
else if (Rs >= 4749) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
if (Cs < 1245) {
if (RsDiff < 25) {
return 1;
}
else if (RsDiff >= 25) {
return 2;
}
}
else if (Cs >= 1245) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (CsDiff < 4) {
if (TSC < 6) {
if (tcor < 95) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
else if (tcor >= 95) {
if (diffAFD < -6) {
return 1;
}
else if (diffAFD >= -6) {
return 2;
}
}
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 35) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (Cs < 877) {
if (diffAFD < 1) {
if (negBalance < 42) {
if (Cs < 506) {
return 2;
}
else if (Cs >= 506) {
if (Rs < 3892) {
return 1;
}
else if (Rs >= 3892) {
return 2;
}
}
}
else if (negBalance >= 42) {
return 1;
}
}
else if (diffAFD >= 1) {
if (Rs < 2826) {
return 1;
}
else if (Rs >= 2826) {
return 2;
}
}
}
else if (Cs >= 877) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (SC < 3804) {
if (posBalance < 64) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
if (Rs < 3051) {
return 1;
}
else if (Rs >= 3051) {
return 2;
}
}
}
else if (posBalance >= 64) {
return 2;
}
}
else if (SC >= 3804) {
if (diffMVdiffVal < 18) {
if (Cs < 1203) {
return 2;
}
else if (Cs >= 1203) {
return 1;
}
}
else if (diffMVdiffVal >= 18) {
if (Rs < 5987) {
return 1;
}
else if (Rs >= 5987) {
return 2;
}
}
}
}
}
}
else if (CsDiff >= 4) {
if (Rs < 4190) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (TSC < 11) {
if (posBalance < 67) {
if (Rs < 4149) {
return 1;
}
else if (Rs >= 4149) {
return 8;
}
}
else if (posBalance >= 67) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 1;
}
}
}
else if (TSC >= 11) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 2;
}
}
}
}
}
else if (Rs >= 4190) {
if (negBalance < 49) {
if (Rs < 6446) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
return 1;
}
}
else if (Rs >= 6446) {
return 2;
}
}
else if (negBalance >= 49) {
if (Cs < 1063) {
if (refDCval < 135) {
return 8;
}
else if (refDCval >= 135) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
return 2;
}
}
}
else if (Cs >= 1063) {
if (Rs < 5750) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
return 1;
}
}
else if (Rs >= 5750) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1313) {
if (mu_mv_mag_sq < 111) {
if (RsCsDiff < 101) {
if (refDCval < 105) {
if (negBalance < 39) {
if (Rs < 816) {
if (Cs < 1454) {
if (Rs < 301) {
return 4;
}
else if (Rs >= 301) {
if (ssDCval < 70) {
if (Rs < 739) {
return 1;
}
else if (Rs >= 739) {
return 8;
}
}
else if (ssDCval >= 70) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
return 4;
}
}
}
}
else if (Cs >= 1454) {
if (TSC < 5) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (TSC >= 5) {
if (posBalance < 62) {
if (Rs < 659) {
return 2;
}
else if (Rs >= 659) {
return 1;
}
}
else if (posBalance >= 62) {
return 8;
}
}
}
}
else if (Rs >= 816) {
if (posBalance < 43) {
if (ssDCval < 76) {
if (ssDCval < 64) {
if (Rs < 2997) {
if (mu_mv_mag_sq < 97) {
if (ssDCval < 52) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (ssDCval >= 52) {
if (diffAFD < -1) {
if (Cs < 1650) {
return 8;
}
else if (Cs >= 1650) {
return 2;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (SC < 3437) {
if (Rs < 1781) {
if (Rs < 995) {
return 8;
}
else if (Rs >= 995) {
if (SC < 2962) {
return 4;
}
else if (SC >= 2962) {
return 8;
}
}
}
else if (Rs >= 1781) {
return 1;
}
}
else if (SC >= 3437) {
return 8;
}
}
}
else if (Rs >= 2997) {
if (Cs < 2385) {
if (MVDiff < 3) {
if (Rs < 4503) {
return 4;
}
else if (Rs >= 4503) {
if (Rs < 4627) {
return 4;
}
else if (Rs >= 4627) {
return 4;
}
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (Cs >= 2385) {
if (Rs < 4945) {
if (posBalance < 38) {
if (Rs < 4644) {
return 4;
}
else if (Rs >= 4644) {
return 4;
}
}
else if (posBalance >= 38) {
if (Rs < 4731) {
return 8;
}
else if (Rs >= 4731) {
return 8;
}
}
}
else if (Rs >= 4945) {
return 8;
}
}
}
}
else if (ssDCval >= 64) {
if (MVDiff < 5) {
if (Rs < 3514) {
if (Rs < 2642) {
return 8;
}
else if (Rs >= 2642) {
if (Cs < 1903) {
return 2;
}
else if (Cs >= 1903) {
return 1;
}
}
}
else if (Rs >= 3514) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (Rs < 8803) {
if (mu_mv_mag_sq < 21) {
return 2;
}
else if (mu_mv_mag_sq >= 21) {
return 4;
}
}
else if (Rs >= 8803) {
return 2;
}
}
}
}
else if (ssDCval >= 76) {
if (Rs < 2838) {
if (Rs < 1874) {
return 8;
}
else if (Rs >= 1874) {
if (negBalance < 26) {
if (ssDCval < 86) {
return 1;
}
else if (ssDCval >= 86) {
if (Rs < 2405) {
if (mu_mv_mag_sq < 109) {
return 4;
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 2338) {
return 2;
}
else if (Rs >= 2338) {
return 1;
}
}
}
else if (Rs >= 2405) {
return 8;
}
}
}
else if (negBalance >= 26) {
if (negBalance < 37) {
return 1;
}
else if (negBalance >= 37) {
if (Rs < 1983) {
return 1;
}
else if (Rs >= 1983) {
return 8;
}
}
}
}
}
else if (Rs >= 2838) {
if (MVDiff < 7) {
if (SC < 5593) {
if (Rs < 3283) {
if (SC < 4990) {
return 8;
}
else if (SC >= 4990) {
return 1;
}
}
else if (Rs >= 3283) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (negBalance < 37) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (negBalance >= 37) {
return 2;
}
}
}
}
else if (SC >= 5593) {
if (ssDCval < 91) {
if (refDCval < 85) {
if (SC < 7001) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
else if (SC >= 7001) {
if (Cs < 1813) {
return 8;
}
else if (Cs >= 1813) {
return 4;
}
}
}
else if (refDCval >= 85) {
if (CsDiff < 6) {
if (refDCval < 90) {
return 4;
}
else if (refDCval >= 90) {
if (Rs < 4425) {
return 4;
}
else if (Rs >= 4425) {
return 2;
}
}
}
else if (CsDiff >= 6) {
return 1;
}
}
}
else if (ssDCval >= 91) {
if (negBalance < 28) {
if (ssDCval < 96) {
return 2;
}
else if (ssDCval >= 96) {
if (Rs < 3460) {
return 8;
}
else if (Rs >= 3460) {
return 4;
}
}
}
else if (negBalance >= 28) {
if (posBalance < 39) {
return 2;
}
else if (posBalance >= 39) {
return 8;
}
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 4246) {
if (Cs < 1410) {
return 4;
}
else if (Cs >= 1410) {
return 8;
}
}
else if (Rs >= 4246) {
if (Cs < 2228) {
return 4;
}
else if (Cs >= 2228) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 43) {
if (Rs < 3499) {
if (Cs < 1600) {
if (tcor < 75) {
if (TSC < 6) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (tcor >= 75) {
if (RsDiff < 7) {
if (diffRsCsdiff < -19) {
if (mu_mv_mag_sq < 94) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (Cs < 1429) {
return 2;
}
else if (Cs >= 1429) {
if (Rs < 3154) {
return 4;
}
else if (Rs >= 3154) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 94) {
return 2;
}
}
else if (diffRsCsdiff >= -19) {
if (Rs < 1116) {
if (ssDCval < 56) {
return 2;
}
else if (ssDCval >= 56) {
if (posBalance < 56) {
if (MVDiff < 4) {
if (negBalance < 27) {
return 2;
}
else if (negBalance >= 27) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (Rs < 967) {
return 8;
}
else if (Rs >= 967) {
return 4;
}
}
}
}
}
else if (MVDiff >= 4) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
else if (Rs >= 1116) {
if (ssDCval < 87) {
if (Rs < 2619) {
if (tcor < 97) {
if (MVDiff < 11) {
if (posBalance < 55) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Rs < 1569) {
return 4;
}
else if (Rs >= 1569) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 59) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
else if (ssDCval >= 59) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 4;
}
}
}
}
else if (MVDiff >= 11) {
if (Rs < 1595) {
return 4;
}
else if (Rs >= 1595) {
return 8;
}
}
}
else if (tcor >= 97) {
if (negBalance < 29) {
if (Rs < 1462) {
return 4;
}
else if (Rs >= 1462) {
return 8;
}
}
else if (negBalance >= 29) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
else if (Rs >= 2619) {
if (CsDiff < 4) {
if (Rs < 3259) {
return 4;
}
else if (Rs >= 3259) {
if (Rs < 3404) {
return 4;
}
else if (Rs >= 3404) {
return 4;
}
}
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (ssDCval >= 87) {
if (MVDiff < 4) {
if (Rs < 1839) {
return 8;
}
else if (Rs >= 1839) {
if (ssDCval < 99) {
return 2;
}
else if (ssDCval >= 99) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (Rs < 1334) {
return 4;
}
else if (Rs >= 1334) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (MVDiff < 10) {
if (Rs < 2296) {
if (Rs < 2024) {
return 8;
}
else if (Rs >= 2024) {
return 4;
}
}
else if (Rs >= 2296) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (AFD < 13) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
if (Rs < 3248) {
return 4;
}
else if (Rs >= 3248) {
return 2;
}
}
}
else if (AFD >= 13) {
return 4;
}
}
}
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
}
else if (Cs >= 1600) {
if (RsCsDiff < 63) {
if (SC < 4665) {
if (MVDiff < 5) {
if (SC < 4397) {
if (diffAFD < -2) {
if (Rs < 1703) {
return 1;
}
else if (Rs >= 1703) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (SC >= 4397) {
if (mu_mv_mag_sq < 108) {
if (diffTSC < -1) {
if (Cs < 1867) {
if (Rs < 2710) {
return 4;
}
else if (Rs >= 2710) {
return 1;
}
}
else if (Cs >= 1867) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 108) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 83) {
if (negBalance < 29) {
if (Rs < 1033) {
return 8;
}
else if (Rs >= 1033) {
if (negBalance < 14) {
if (SC < 2785) {
return 4;
}
else if (SC >= 2785) {
return 8;
}
}
else if (negBalance >= 14) {
if (mu_mv_mag_sq < 105) {
if (Cs < 2025) {
return 8;
}
else if (Cs >= 2025) {
return 4;
}
}
else if (mu_mv_mag_sq >= 105) {
return 4;
}
}
}
}
else if (negBalance >= 29) {
if (RsCsDiff < 33) {
if (Rs < 1940) {
if (ssDCval < 61) {
if (Rs < 975) {
return 1;
}
else if (Rs >= 975) {
return 8;
}
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (Rs >= 1940) {
return 8;
}
}
else if (RsCsDiff >= 33) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (Cs < 1838) {
return 8;
}
else if (Cs >= 1838) {
if (Rs < 1227) {
return 1;
}
else if (Rs >= 1227) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 83) {
if (diffAFD < -3) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (diffAFD >= -3) {
if (mcTcor < 88) {
if (Rs < 1548) {
return 4;
}
else if (Rs >= 1548) {
return 8;
}
}
else if (mcTcor >= 88) {
return 8;
}
}
}
}
}
else if (SC >= 4665) {
if (negBalance < 26) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
else if (negBalance >= 26) {
if (Cs < 2545) {
if (Cs < 1902) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
else if (Cs >= 1902) {
if (Rs < 2325) {
return 4;
}
else if (Rs >= 2325) {
if (refDCval < 83) {
if (tcor < 96) {
if (RsCsDiff < 57) {
return 8;
}
else if (RsCsDiff >= 57) {
return 1;
}
}
else if (tcor >= 96) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (refDCval >= 83) {
return 8;
}
}
}
}
else if (Cs >= 2545) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Rs < 3238) {
return 1;
}
else if (Rs >= 3238) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 63) {
if (posBalance < 48) {
if (mcTcor < 90) {
if (Rs < 2748) {
return 4;
}
else if (Rs >= 2748) {
return 8;
}
}
else if (mcTcor >= 90) {
if (Cs < 3060) {
return 1;
}
else if (Cs >= 3060) {
return 4;
}
}
}
else if (posBalance >= 48) {
if (Rs < 1399) {
if (MVDiff < 10) {
if (Rs < 933) {
return 8;
}
else if (Rs >= 933) {
return 1;
}
}
else if (MVDiff >= 10) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (Rs >= 1399) {
if (diffRsCsdiff < 30) {
if (TSC < 8) {
if (Rs < 3380) {
return 8;
}
else if (Rs >= 3380) {
return 2;
}
}
else if (TSC >= 8) {
if (posBalance < 51) {
if (Rs < 2773) {
return 8;
}
else if (Rs >= 2773) {
return 1;
}
}
else if (posBalance >= 51) {
if (Cs < 2112) {
return 8;
}
else if (Cs >= 2112) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 30) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 3499) {
if (Cs < 1934) {
if (mu_mv_mag_sq < 64) {
if (ssDCval < 85) {
if (negBalance < 32) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 2;
}
}
else if (negBalance >= 32) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (ssDCval >= 85) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (SC < 5781) {
if (Rs < 3988) {
return 8;
}
else if (Rs >= 3988) {
if (Cs < 1335) {
return 4;
}
else if (Cs >= 1335) {
return 8;
}
}
}
else if (SC >= 5781) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (posBalance < 46) {
if (SC < 5603) {
if (negBalance < 33) {
if (SC < 5516) {
return 4;
}
else if (SC >= 5516) {
return 8;
}
}
else if (negBalance >= 33) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (Rs < 4030) {
return 8;
}
else if (Rs >= 4030) {
return 4;
}
}
}
}
else if (SC >= 5603) {
if (diffRsCsdiff < 14) {
return 2;
}
else if (diffRsCsdiff >= 14) {
if (Rs < 4435) {
return 8;
}
else if (Rs >= 4435) {
return 4;
}
}
}
}
else if (posBalance >= 46) {
return 2;
}
}
}
else if (Cs >= 1934) {
if (Rs < 6019) {
if (ssDCval < 73) {
if (Cs < 2890) {
if (negBalance < 21) {
if (Rs < 4557) {
return 8;
}
else if (Rs >= 4557) {
return 4;
}
}
else if (negBalance >= 21) {
return 4;
}
}
else if (Cs >= 2890) {
return 8;
}
}
else if (ssDCval >= 73) {
if (mu_mv_mag_sq < 101) {
if (RsCsDiff < 99) {
return 8;
}
else if (RsCsDiff >= 99) {
return 2;
}
}
else if (mu_mv_mag_sq >= 101) {
if (RsDiff < 6) {
if (Rs < 4240) {
return 4;
}
else if (Rs >= 4240) {
return 8;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
else if (Rs >= 6019) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 1456) {
if (RsCsDiff < 51) {
if (ssDCval < 83) {
if (TSC < 10) {
if (MVDiff < 5) {
if (ssDCval < 79) {
if (Rs < 916) {
return 8;
}
else if (Rs >= 916) {
if (mu_mv_mag_sq < 72) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (mu_mv_mag_sq < 104) {
if (RsCsDiff < 46) {
if (negBalance < 42) {
if (Rs < 1300) {
return 4;
}
else if (Rs >= 1300) {
return 8;
}
}
else if (negBalance >= 42) {
return 8;
}
}
else if (RsCsDiff >= 46) {
return 4;
}
}
else if (mu_mv_mag_sq >= 104) {
if (MVDiff < 1) {
if (Rs < 1314) {
return 4;
}
else if (Rs >= 1314) {
return 8;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
}
}
}
else if (ssDCval >= 79) {
if (Rs < 1146) {
return 1;
}
else if (Rs >= 1146) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (posBalance < 58) {
if (Rs < 1099) {
if (RsDiff < 3) {
if (tcor < 96) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 622) {
return 2;
}
else if (Rs >= 622) {
return 8;
}
}
}
else if (tcor >= 96) {
return 8;
}
}
else if (RsDiff >= 3) {
if (Rs < 700) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
else if (Rs >= 700) {
if (gchDC < 3) {
if (MVDiff < 25) {
return 2;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (gchDC >= 3) {
if (Rs < 981) {
return 8;
}
else if (Rs >= 981) {
return 4;
}
}
}
}
}
else if (Rs >= 1099) {
if (negBalance < 48) {
if (Cs < 1413) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
return 4;
}
}
else if (Cs >= 1413) {
return 1;
}
}
else if (negBalance >= 48) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (negBalance < 55) {
if (mcTcor < 95) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (mcTcor >= 95) {
if (Rs < 1102) {
if (Cs < 1433) {
return 2;
}
else if (Cs >= 1433) {
return 8;
}
}
else if (Rs >= 1102) {
return 4;
}
}
}
else if (negBalance >= 55) {
return 8;
}
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (ssDCval >= 83) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1209) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
return 4;
}
}
else if (Rs >= 1209) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (posBalance < 50) {
return 4;
}
else if (posBalance >= 50) {
if (SC < 3208) {
return 8;
}
else if (SC >= 3208) {
if (Rs < 1335) {
return 4;
}
else if (Rs >= 1335) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 51) {
if (mcTcor < 97) {
if (ssDCval < 55) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
if (RsCsDiff < 91) {
return 2;
}
else if (RsCsDiff >= 91) {
return 4;
}
}
}
else if (ssDCval >= 55) {
if (refDCval < 71) {
if (RsCsDiff < 73) {
if (posBalance < 59) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 1048) {
return 2;
}
else if (Rs >= 1048) {
return 4;
}
}
}
else if (posBalance >= 59) {
return 8;
}
}
else if (RsCsDiff >= 73) {
if (tcor < 74) {
if (Cs < 1393) {
return 2;
}
else if (Cs >= 1393) {
if (RsCsDiff < 99) {
return 4;
}
else if (RsCsDiff >= 99) {
return 2;
}
}
}
else if (tcor >= 74) {
if (MVDiff < 32) {
return 8;
}
else if (MVDiff >= 32) {
return 4;
}
}
}
}
else if (refDCval >= 71) {
if (negBalance < 52) {
if (TSC < 7) {
if (refDCval < 75) {
return 2;
}
else if (refDCval >= 75) {
return 8;
}
}
else if (TSC >= 7) {
if (posBalance < 63) {
if (mcTcor < 87) {
if (Rs < 1114) {
return 4;
}
else if (Rs >= 1114) {
return 2;
}
}
else if (mcTcor >= 87) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 38) {
return 4;
}
else if (mu_mv_mag_sq >= 38) {
if (Cs < 1703) {
return 8;
}
else if (Cs >= 1703) {
return 4;
}
}
}
}
}
else if (posBalance >= 63) {
if (AFD < 17) {
if (Cs < 1533) {
return 8;
}
else if (Cs >= 1533) {
if (Rs < 753) {
return 8;
}
else if (Rs >= 753) {
return 4;
}
}
}
else if (AFD >= 17) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (ssDCval < 85) {
if (Rs < 1269) {
if (mcTcor < 96) {
if (Rs < 1130) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
if (Rs < 970) {
return 4;
}
else if (Rs >= 970) {
return 8;
}
}
}
else if (Rs >= 1130) {
return 2;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (Rs >= 1269) {
return 8;
}
}
else if (ssDCval >= 85) {
if (posBalance < 55) {
if (Rs < 803) {
return 8;
}
else if (Rs >= 803) {
return 2;
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 42) {
if (Rs < 843) {
return 1;
}
else if (Rs >= 843) {
return 2;
}
}
else if (posBalance >= 42) {
if (tcor < 89) {
if (Rs < 700) {
return 4;
}
else if (Rs >= 700) {
return 8;
}
}
else if (tcor >= 89) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1456) {
if (refDCval < 102) {
if (SC < 7235) {
if (Cs < 2479) {
if (mcTcor < 94) {
if (RsCsDiff < 87) {
if (Rs < 2353) {
if (refDCval < 91) {
if (posBalance < 70) {
if (AFD < 16) {
if (posBalance < 56) {
if (Cs < 2333) {
if (diffAFD < -1) {
if (RsCsDiff < 69) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
if (Rs < 1792) {
return 8;
}
else if (Rs >= 1792) {
if (Rs < 2134) {
return 1;
}
else if (Rs >= 2134) {
return 4;
}
}
}
}
else if (RsCsDiff >= 69) {
if (Rs < 1950) {
return 1;
}
else if (Rs >= 1950) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
}
else if (diffAFD >= -1) {
if (Cs < 1322) {
return 2;
}
else if (Cs >= 1322) {
if (Rs < 2326) {
return 8;
}
else if (Rs >= 2326) {
if (Rs < 2338) {
return 4;
}
else if (Rs >= 2338) {
return 8;
}
}
}
}
}
else if (Cs >= 2333) {
return 4;
}
}
else if (posBalance >= 56) {
if (negBalance < 44) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (Rs < 1896) {
return 8;
}
else if (Rs >= 1896) {
return 2;
}
}
}
else if (negBalance >= 44) {
if (ssDCval < 60) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (SC < 3514) {
if (negBalance < 47) {
return 4;
}
else if (negBalance >= 47) {
return 8;
}
}
else if (SC >= 3514) {
return 4;
}
}
}
else if (ssDCval >= 60) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
}
}
else if (AFD >= 16) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
if (negBalance < 57) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
return 8;
}
}
else if (negBalance >= 57) {
return 4;
}
}
}
}
else if (posBalance >= 70) {
return 4;
}
}
else if (refDCval >= 91) {
if (tcor < 83) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 1;
}
}
else if (tcor >= 83) {
if (diffRsCsdiff < 9) {
if (Cs < 2141) {
return 8;
}
else if (Cs >= 2141) {
if (Rs < 1657) {
return 2;
}
else if (Rs >= 1657) {
return 4;
}
}
}
else if (diffRsCsdiff >= 9) {
if (AFD < 11) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 2110) {
return 2;
}
else if (Rs >= 2110) {
return 8;
}
}
}
else if (AFD >= 11) {
if (Cs < 1940) {
return 8;
}
else if (Cs >= 1940) {
return 4;
}
}
}
}
}
}
else if (Rs >= 2353) {
if (Rs < 4472) {
if (mu_mv_mag_sq < 93) {
return 8;
}
else if (mu_mv_mag_sq >= 93) {
if (mu_mv_mag_sq < 97) {
if (MVDiff < 9) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (mu_mv_mag_sq >= 97) {
return 8;
}
}
}
else if (Rs >= 4472) {
return 4;
}
}
}
else if (RsCsDiff >= 87) {
if (ssDCval < 85) {
if (TSC < 8) {
if (mcTcor < 91) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (TSC < 6) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (mcTcor >= 91) {
if (negBalance < 46) {
if (Rs < 2701) {
return 8;
}
else if (Rs >= 2701) {
return 2;
}
}
else if (negBalance >= 46) {
return 4;
}
}
}
else if (TSC >= 8) {
if (diffAFD < -2) {
if (diffAFD < -4) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (diffAFD >= -4) {
if (SC < 4358) {
return 4;
}
else if (SC >= 4358) {
return 8;
}
}
}
else if (diffAFD >= -2) {
if (Rs < 2512) {
if (diffMVdiffVal < -6) {
if (Cs < 1470) {
if (Cs < 1415) {
return 8;
}
else if (Cs >= 1415) {
return 1;
}
}
else if (Cs >= 1470) {
return 4;
}
}
else if (diffMVdiffVal >= -6) {
if (AFD < 19) {
if (MVDiff < 29) {
if (CsDiff < 6) {
if (Rs < 2018) {
return 4;
}
else if (Rs >= 2018) {
return 8;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (MVDiff >= 29) {
return 4;
}
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (Rs >= 2512) {
if (refDCval < 78) {
return 8;
}
else if (refDCval >= 78) {
if (tcor < 83) {
if (Rs < 2698) {
return 4;
}
else if (Rs >= 2698) {
return 8;
}
}
else if (tcor >= 83) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (Rs < 2214) {
if (TSC < 9) {
if (Rs < 1937) {
return 8;
}
else if (Rs >= 1937) {
return 2;
}
}
else if (TSC >= 9) {
if (Cs < 1647) {
return 8;
}
else if (Cs >= 1647) {
if (Rs < 1753) {
if (negBalance < 56) {
return 4;
}
else if (negBalance >= 56) {
return 8;
}
}
else if (Rs >= 1753) {
if (gchDC < 1) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
}
else if (Rs >= 2214) {
if (Rs < 4632) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (Rs >= 4632) {
return 1;
}
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 2954) {
if (refDCval < 89) {
if (tcor < 88) {
if (Cs < 2367) {
if (Cs < 1538) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (tcor < 86) {
return 2;
}
else if (tcor >= 86) {
if (Cs < 1482) {
return 4;
}
else if (Cs >= 1482) {
if (Rs < 2011) {
return 8;
}
else if (Rs >= 2011) {
return 2;
}
}
}
}
}
else if (Cs >= 1538) {
if (Rs < 2682) {
return 4;
}
else if (Rs >= 2682) {
return 8;
}
}
}
else if (Cs >= 2367) {
return 1;
}
}
else if (tcor >= 88) {
if (Cs < 1538) {
if (MVDiff < 5) {
if (refDCval < 80) {
if (Rs < 2155) {
if (Cs < 1465) {
if (Cs < 1324) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
else if (Cs >= 1324) {
return 8;
}
}
else if (Cs >= 1465) {
if (Rs < 1980) {
return 4;
}
else if (Rs >= 1980) {
return 8;
}
}
}
else if (Rs >= 2155) {
if (RsCsDiff < 38) {
return 1;
}
else if (RsCsDiff >= 38) {
if (Rs < 2442) {
return 8;
}
else if (Rs >= 2442) {
return 2;
}
}
}
}
else if (refDCval >= 80) {
if (ssDCval < 84) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 1;
}
}
else if (ssDCval >= 84) {
if (TSC < 7) {
if (Cs < 1432) {
return 4;
}
else if (Cs >= 1432) {
return 2;
}
}
else if (TSC >= 7) {
if (Rs < 2228) {
return 8;
}
else if (Rs >= 2228) {
return 1;
}
}
}
}
}
else if (MVDiff >= 5) {
if (mu_mv_mag_sq < 68) {
return 8;
}
else if (mu_mv_mag_sq >= 68) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 1;
}
}
}
}
else if (Cs >= 1538) {
if (TSC < 7) {
if (MVDiff < 5) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (SC < 3762) {
return 4;
}
else if (SC >= 3762) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (diffAFD < -1) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 8;
}
}
else if (diffAFD >= -1) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Cs < 1708) {
return 8;
}
else if (Cs >= 1708) {
return 4;
}
}
}
}
}
else if (TSC >= 7) {
if (MVDiff < 2) {
if (SC < 4567) {
if (Rs < 1869) {
if (Rs < 1693) {
return 8;
}
else if (Rs >= 1693) {
return 4;
}
}
else if (Rs >= 1869) {
if (ssDCval < 78) {
return 4;
}
else if (ssDCval >= 78) {
return 1;
}
}
}
else if (SC >= 4567) {
if (Cs < 1941) {
return 1;
}
else if (Cs >= 1941) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 79) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (ssDCval >= 79) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (negBalance < 53) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (negBalance >= 53) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 89) {
if (Rs < 2039) {
if (mcTcor < 96) {
if (RsDiff < 7) {
if (AFD < 7) {
if (Rs < 1794) {
return 8;
}
else if (Rs >= 1794) {
return 2;
}
}
else if (AFD >= 7) {
if (Cs < 2306) {
return 8;
}
else if (Cs >= 2306) {
return 1;
}
}
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (mcTcor >= 96) {
if (negBalance < 45) {
if (posBalance < 61) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
else if (posBalance >= 61) {
return 2;
}
}
else if (negBalance >= 45) {
if (Cs < 1502) {
if (Rs < 1933) {
return 8;
}
else if (Rs >= 1933) {
return 1;
}
}
else if (Cs >= 1502) {
if (Rs < 1984) {
return 2;
}
else if (Rs >= 1984) {
if (Rs < 2034) {
return 8;
}
else if (Rs >= 2034) {
return 2;
}
}
}
}
}
}
else if (Rs >= 2039) {
if (negBalance < 53) {
if (gchDC < 2) {
if (diffMVdiffVal < -12) {
return 2;
}
else if (diffMVdiffVal >= -12) {
return 8;
}
}
else if (gchDC >= 2) {
if (Cs < 2033) {
if (Rs < 2208) {
return 2;
}
else if (Rs >= 2208) {
return 4;
}
}
else if (Cs >= 2033) {
return 8;
}
}
}
else if (negBalance >= 53) {
if (AFD < 11) {
if (Cs < 1824) {
return 8;
}
else if (Cs >= 1824) {
return 4;
}
}
else if (AFD >= 11) {
if (ssDCval < 100) {
return 8;
}
else if (ssDCval >= 100) {
return 2;
}
}
}
}
}
}
else if (Rs >= 2954) {
if (ssDCval < 88) {
if (AFD < 10) {
if (Cs < 1894) {
if (RsDiff < 7) {
if (Cs < 1373) {
return 2;
}
else if (Cs >= 1373) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
else if (RsDiff >= 7) {
if (Rs < 2996) {
return 4;
}
else if (Rs >= 2996) {
return 2;
}
}
}
else if (Cs >= 1894) {
if (SC < 5599) {
return 8;
}
else if (SC >= 5599) {
return 4;
}
}
}
else if (AFD >= 10) {
return 4;
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 73) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
return 1;
}
}
else if (mu_mv_mag_sq >= 73) {
if (Rs < 3313) {
return 4;
}
else if (Rs >= 3313) {
if (Rs < 3577) {
return 8;
}
else if (Rs >= 3577) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 2479) {
if (Rs < 2749) {
if (Cs < 2836) {
if (mu_mv_mag_sq < 90) {
if (mcTcor < 94) {
if (Rs < 2690) {
return 8;
}
else if (Rs >= 2690) {
return 2;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (mu_mv_mag_sq >= 90) {
if (Rs < 2549) {
return 4;
}
else if (Rs >= 2549) {
return 8;
}
}
}
else if (Cs >= 2836) {
return 4;
}
}
else if (Rs >= 2749) {
if (ssDCval < 87) {
if (ssDCval < 51) {
if (Cs < 2865) {
return 4;
}
else if (Cs >= 2865) {
return 8;
}
}
else if (ssDCval >= 51) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (posBalance < 57) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (Rs < 3194) {
return 8;
}
else if (Rs >= 3194) {
return 4;
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
else if (ssDCval >= 87) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
if (Cs < 2756) {
return 8;
}
else if (Cs >= 2756) {
return 4;
}
}
}
}
}
}
else if (SC >= 7235) {
if (Cs < 2539) {
return 4;
}
else if (Cs >= 2539) {
return 8;
}
}
}
else if (refDCval >= 102) {
if (mu_mv_mag_sq < 89) {
if (Rs < 3217) {
if (tcor < 91) {
if (AFD < 10) {
if (Rs < 2804) {
return 4;
}
else if (Rs >= 2804) {
return 8;
}
}
else if (AFD >= 10) {
if (Cs < 1403) {
return 4;
}
else if (Cs >= 1403) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
}
}
else if (tcor >= 91) {
if (SC < 4854) {
if (Rs < 1865) {
if (Rs < 1581) {
return 8;
}
else if (Rs >= 1581) {
return 2;
}
}
else if (Rs >= 1865) {
return 8;
}
}
else if (SC >= 4854) {
return 4;
}
}
}
else if (Rs >= 3217) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (Rs < 4005) {
return 4;
}
else if (Rs >= 4005) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (tcor < 95) {
if (Cs < 2960) {
return 1;
}
else if (Cs >= 2960) {
return 8;
}
}
else if (tcor >= 95) {
if (Rs < 2166) {
if (Rs < 1835) {
return 2;
}
else if (Rs >= 1835) {
return 1;
}
}
else if (Rs >= 2166) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 105) {
if (ssDCval < 163) {
if (Rs < 4657) {
if (Rs < 1489) {
if (mu_mv_mag_sq < 38) {
if (ssDCval < 138) {
if (TSC < 11) {
if (negBalance < 44) {
if (Cs < 1603) {
if (MVDiff < 10) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (Rs < 1161) {
return 1;
}
else if (Rs >= 1161) {
return 8;
}
}
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (Cs >= 1603) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (negBalance >= 44) {
if (diffAFD < 3) {
if (mcTcor < 90) {
if (Cs < 1806) {
return 1;
}
else if (Cs >= 1806) {
return 8;
}
}
else if (mcTcor >= 90) {
if (refDCval < 106) {
return 4;
}
else if (refDCval >= 106) {
return 8;
}
}
}
else if (diffAFD >= 3) {
return 2;
}
}
}
else if (TSC >= 11) {
if (Rs < 1293) {
if (Cs < 2441) {
return 2;
}
else if (Cs >= 2441) {
return 8;
}
}
else if (Rs >= 1293) {
return 4;
}
}
}
else if (ssDCval >= 138) {
if (SC < 2651) {
return 1;
}
else if (SC >= 2651) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 1239) {
if (Rs < 968) {
if (ssDCval < 119) {
if (AFD < 7) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
else if (AFD >= 7) {
return 8;
}
}
else if (ssDCval >= 119) {
if (Cs < 1991) {
return 8;
}
else if (Cs >= 1991) {
return 1;
}
}
}
else if (Rs >= 968) {
if (CsDiff < 6) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (MVDiff < 19) {
if (MVDiff < 3) {
if (Rs < 1091) {
return 8;
}
else if (Rs >= 1091) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (MVDiff >= 19) {
return 8;
}
}
}
else if (CsDiff >= 6) {
if (posBalance < 118) {
return 8;
}
else if (posBalance >= 118) {
return 1;
}
}
}
}
else if (Rs >= 1239) {
if (AFD < 10) {
if (ssDCval < 160) {
return 8;
}
else if (ssDCval >= 160) {
return 2;
}
}
else if (AFD >= 10) {
if (ssDCval < 143) {
if (posBalance < 54) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Rs < 1388) {
return 4;
}
else if (Rs >= 1388) {
return 2;
}
}
}
else if (posBalance >= 54) {
if (Rs < 1463) {
return 8;
}
else if (Rs >= 1463) {
return 1;
}
}
}
else if (ssDCval >= 143) {
return 4;
}
}
}
}
}
else if (Rs >= 1489) {
if (mcTcor < 95) {
if (ssDCval < 142) {
if (refDCval < 125) {
if (TSC < 11) {
if (diffRsCsdiff < -30) {
if (AFD < 17) {
if (diffRsCsdiff < -146) {
if (Rs < 3062) {
return 1;
}
else if (Rs >= 3062) {
return 4;
}
}
else if (diffRsCsdiff >= -146) {
if (diffMVdiffVal < -13) {
if (Rs < 2963) {
return 1;
}
else if (Rs >= 2963) {
return 4;
}
}
else if (diffMVdiffVal >= -13) {
if (Cs < 2464) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
if (Rs < 1710) {
return 8;
}
else if (Rs >= 1710) {
return 4;
}
}
}
else if (Cs >= 2464) {
if (Rs < 3859) {
if (Rs < 3179) {
return 1;
}
else if (Rs >= 3179) {
return 4;
}
}
else if (Rs >= 3859) {
return 8;
}
}
}
}
}
else if (AFD >= 17) {
if (Rs < 1538) {
return 1;
}
else if (Rs >= 1538) {
return 4;
}
}
}
else if (diffRsCsdiff >= -30) {
if (diffRsCsdiff < 80) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (TSC < 7) {
if (posBalance < 51) {
if (Cs < 1792) {
return 4;
}
else if (Cs >= 1792) {
if (Rs < 2903) {
return 2;
}
else if (Rs >= 2903) {
return 8;
}
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (diffRsCsdiff >= 80) {
return 4;
}
}
}
else if (TSC >= 11) {
if (CsDiff < 5) {
if (Cs < 1788) {
return 1;
}
else if (Cs >= 1788) {
return 8;
}
}
else if (CsDiff >= 5) {
if (Rs < 1547) {
return 4;
}
else if (Rs >= 1547) {
return 8;
}
}
}
}
else if (refDCval >= 125) {
if (mcTcor < 92) {
if (Cs < 2166) {
if (AFD < 12) {
if (refDCval < 138) {
if (SC < 3822) {
if (negBalance < 49) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (mu_mv_mag_sq < 51) {
return 1;
}
else if (mu_mv_mag_sq >= 51) {
return 8;
}
}
}
else if (negBalance >= 49) {
return 8;
}
}
else if (SC >= 3822) {
return 8;
}
}
else if (refDCval >= 138) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (AFD >= 12) {
if (Rs < 1545) {
return 4;
}
else if (Rs >= 1545) {
return 8;
}
}
}
else if (Cs >= 2166) {
return 8;
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 106) {
if (tcor < 84) {
if (SC < 4403) {
return 8;
}
else if (SC >= 4403) {
return 4;
}
}
else if (tcor >= 84) {
if (refDCval < 127) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (Cs < 1849) {
return 1;
}
else if (Cs >= 1849) {
return 8;
}
}
}
else if (refDCval >= 127) {
if (RsCsDiff < 22) {
if (Cs < 1771) {
return 8;
}
else if (Cs >= 1771) {
return 1;
}
}
else if (RsCsDiff >= 22) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 2830) {
return 8;
}
else if (Rs >= 2830) {
return 1;
}
}
}
}
}
else if (ssDCval >= 142) {
if (SC < 5778) {
if (TSC < 12) {
if (posBalance < 67) {
if (diffMVdiffVal < -17) {
if (Rs < 1843) {
return 2;
}
else if (Rs >= 1843) {
return 8;
}
}
else if (diffMVdiffVal >= -17) {
return 8;
}
}
else if (posBalance >= 67) {
if (Rs < 2107) {
return 4;
}
else if (Rs >= 2107) {
return 8;
}
}
}
else if (TSC >= 12) {
return 2;
}
}
else if (SC >= 5778) {
return 1;
}
}
}
else if (mcTcor >= 95) {
if (refDCval < 112) {
if (mu_mv_mag_sq < 43) {
if (Rs < 3150) {
if (Cs < 1793) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
if (RsDiff < 5) {
if (Rs < 2817) {
return 4;
}
else if (Rs >= 2817) {
return 8;
}
}
else if (RsDiff >= 5) {
if (Rs < 1778) {
return 2;
}
else if (Rs >= 1778) {
return 8;
}
}
}
}
else if (Cs >= 1793) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
else if (Rs >= 3150) {
return 8;
}
}
else if (mu_mv_mag_sq >= 43) {
if (posBalance < 33) {
if (negBalance < 57) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
if (Rs < 2155) {
return 2;
}
else if (Rs >= 2155) {
if (Rs < 2689) {
return 4;
}
else if (Rs >= 2689) {
if (Rs < 3810) {
return 8;
}
else if (Rs >= 3810) {
return 2;
}
}
}
}
}
else if (negBalance >= 57) {
return 8;
}
}
else if (posBalance >= 33) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (ssDCval < 109) {
if (MVDiff < 2) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (MVDiff >= 2) {
if (AFD < 8) {
if (Cs < 1459) {
return 8;
}
else if (Cs >= 1459) {
return 4;
}
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (ssDCval >= 109) {
return 8;
}
}
}
}
}
else if (refDCval >= 112) {
if (Cs < 1360) {
if (mcTcor < 98) {
if (diffTSC < 2) {
if (RsDiff < 3) {
if (Rs < 2449) {
return 1;
}
else if (Rs >= 2449) {
return 8;
}
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (diffTSC >= 2) {
if (Rs < 3339) {
return 2;
}
else if (Rs >= 3339) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 2413) {
if (Rs < 1658) {
return 2;
}
else if (Rs >= 1658) {
return 8;
}
}
else if (Rs >= 2413) {
return 2;
}
}
}
else if (Cs >= 1360) {
if (Rs < 1753) {
if (mu_mv_mag_sq < 17) {
return 2;
}
else if (mu_mv_mag_sq >= 17) {
if (negBalance < 63) {
return 8;
}
else if (negBalance >= 63) {
if (Rs < 1545) {
return 2;
}
else if (Rs >= 1545) {
return 8;
}
}
}
}
else if (Rs >= 1753) {
if (tcor < 96) {
if (negBalance < 32) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Cs < 1598) {
return 1;
}
else if (Cs >= 1598) {
return 8;
}
}
}
else if (negBalance >= 32) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (Rs < 2008) {
return 1;
}
else if (Rs >= 2008) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 8;
}
}
}
}
}
}
else if (tcor >= 96) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
if (CsDiff < 2) {
if (tcor < 100) {
if (Rs < 2752) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 4;
}
}
else if (Rs >= 2752) {
return 2;
}
}
else if (tcor >= 100) {
return 8;
}
}
else if (CsDiff >= 2) {
if (tcor < 106) {
if (refDCval < 113) {
if (Rs < 1961) {
return 2;
}
else if (Rs >= 1961) {
return 8;
}
}
else if (refDCval >= 113) {
return 8;
}
}
else if (tcor >= 106) {
if (Rs < 2014) {
return 8;
}
else if (Rs >= 2014) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 4657) {
if (Cs < 1733) {
return 2;
}
else if (Cs >= 1733) {
if (Cs < 3807) {
if (Rs < 5647) {
return 1;
}
else if (Rs >= 5647) {
if (Rs < 6055) {
return 2;
}
else if (Rs >= 6055) {
return 4;
}
}
}
else if (Cs >= 3807) {
return 8;
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2283) {
if (Rs < 1064) {
return 1;
}
else if (Rs >= 1064) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (Rs >= 2283) {
return 8;
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 2096) {
if (Rs < 3354) {
if (ssDCval < 87) {
if (AFD < 17) {
if (Cs < 1475) {
if (Rs < 2791) {
if (Rs < 1378) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
if (Rs < 1167) {
return 2;
}
else if (Rs >= 1167) {
return 8;
}
}
}
else if (Rs >= 1378) {
if (refDCval < 45) {
return 8;
}
else if (refDCval >= 45) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (TSC < 9) {
if (SC < 3383) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
if (Rs < 1542) {
return 4;
}
else if (Rs >= 1542) {
return 1;
}
}
}
else if (SC >= 3383) {
if (tcor < 85) {
return 2;
}
else if (tcor >= 85) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
if (Rs < 2130) {
return 2;
}
else if (Rs >= 2130) {
return 4;
}
}
}
}
}
else if (TSC >= 9) {
if (Cs < 1338) {
return 1;
}
else if (Cs >= 1338) {
if (RsCsDiff < 129) {
if (tcor < 83) {
return 8;
}
else if (tcor >= 83) {
return 1;
}
}
else if (RsCsDiff >= 129) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 2791) {
if (mcTcor < 92) {
if (diffRsCsdiff < 7) {
return 4;
}
else if (diffRsCsdiff >= 7) {
if (Rs < 3202) {
return 2;
}
else if (Rs >= 3202) {
return 4;
}
}
}
else if (mcTcor >= 92) {
return 2;
}
}
}
else if (Cs >= 1475) {
if (posBalance < 51) {
if (ssDCval < 70) {
if (Cs < 1763) {
if (negBalance < 52) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (negBalance < 32) {
if (Rs < 1717) {
return 4;
}
else if (Rs >= 1717) {
if (Rs < 2575) {
return 8;
}
else if (Rs >= 2575) {
return 1;
}
}
}
else if (negBalance >= 32) {
return 1;
}
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (Cs >= 1763) {
if (posBalance < 45) {
if (posBalance < 35) {
if (Rs < 2786) {
return 8;
}
else if (Rs >= 2786) {
if (Rs < 3082) {
return 1;
}
else if (Rs >= 3082) {
return 4;
}
}
}
else if (posBalance >= 35) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 4;
}
}
}
else if (posBalance >= 45) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
}
else if (ssDCval >= 70) {
if (refDCval < 75) {
if (Rs < 3036) {
return 2;
}
else if (Rs >= 3036) {
return 1;
}
}
else if (refDCval >= 75) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
return 1;
}
}
}
}
else if (posBalance >= 51) {
if (Rs < 1794) {
if (negBalance < 49) {
if (negBalance < 33) {
if (Rs < 1112) {
return 1;
}
else if (Rs >= 1112) {
return 4;
}
}
else if (negBalance >= 33) {
if (Cs < 1537) {
if (Rs < 1230) {
return 2;
}
else if (Rs >= 1230) {
return 1;
}
}
else if (Cs >= 1537) {
if (RsDiff < 11) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
else if (RsDiff >= 11) {
if (Cs < 1708) {
return 1;
}
else if (Cs >= 1708) {
return 4;
}
}
}
}
}
else if (negBalance >= 49) {
if (mcTcor < 93) {
if (SC < 3691) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (negBalance < 63) {
return 4;
}
else if (negBalance >= 63) {
return 2;
}
}
}
}
else if (SC >= 3691) {
return 2;
}
}
else if (mcTcor >= 93) {
if (Rs < 1074) {
return 8;
}
else if (Rs >= 1074) {
return 1;
}
}
}
}
else if (Rs >= 1794) {
if (SC < 4657) {
if (refDCval < 52) {
if (Cs < 1812) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 4;
}
}
else if (Cs >= 1812) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (refDCval >= 52) {
if (mu_mv_mag_sq < 4) {
if (Rs < 2335) {
if (Cs < 1835) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (Rs < 2134) {
return 8;
}
else if (Rs >= 2134) {
return 4;
}
}
}
}
else if (Cs >= 1835) {
return 8;
}
}
else if (Rs >= 2335) {
if (ssDCval < 74) {
if (SC < 4527) {
return 8;
}
else if (SC >= 4527) {
return 4;
}
}
else if (ssDCval >= 74) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 4) {
if (CsDiff < 7) {
if (Rs < 2779) {
if (Rs < 2274) {
return 8;
}
else if (Rs >= 2274) {
return 4;
}
}
else if (Rs >= 2779) {
return 8;
}
}
else if (CsDiff >= 7) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (MVDiff < 17) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (CsDiff < 8) {
if (Rs < 1997) {
return 1;
}
else if (Rs >= 1997) {
return 8;
}
}
else if (CsDiff >= 8) {
return 1;
}
}
}
else if (MVDiff >= 17) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 4657) {
if (mcTcor < 93) {
if (Cs < 2001) {
if (Rs < 2780) {
return 4;
}
else if (Rs >= 2780) {
return 8;
}
}
else if (Cs >= 2001) {
if (negBalance < 43) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
return 2;
}
}
else if (negBalance >= 43) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
return 8;
}
}
}
}
else if (mcTcor >= 93) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 2;
}
}
}
}
}
}
}
else if (AFD >= 17) {
if (negBalance < 48) {
if (posBalance < 55) {
if (Rs < 1636) {
return 4;
}
else if (Rs >= 1636) {
return 2;
}
}
else if (posBalance >= 55) {
if (diffTSC < 9) {
if (refDCval < 60) {
if (Rs < 828) {
if (Rs < 651) {
return 4;
}
else if (Rs >= 651) {
return 2;
}
}
else if (Rs >= 828) {
return 1;
}
}
else if (refDCval >= 60) {
if (RsDiff < 10) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
return 2;
}
}
else if (RsDiff >= 10) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
return 8;
}
}
}
}
else if (diffTSC >= 9) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (mu_mv_mag_sq < 7) {
if (RsCsDiff < 310) {
if (SC < 3320) {
if (MVDiff < 15) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 8;
}
}
}
else if (MVDiff >= 15) {
if (Rs < 1672) {
if (diffMVdiffVal < 17) {
if (ssDCval < 71) {
if (SC < 2869) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 836) {
return 4;
}
else if (Rs >= 836) {
return 2;
}
}
}
else if (SC >= 2869) {
if (ssDCval < 62) {
if (Rs < 1488) {
return 2;
}
else if (Rs >= 1488) {
return 4;
}
}
else if (ssDCval >= 62) {
return 4;
}
}
}
else if (ssDCval >= 71) {
if (ssDCval < 78) {
if (mcTcor < 92) {
return 2;
}
else if (mcTcor >= 92) {
if (Rs < 1260) {
return 2;
}
else if (Rs >= 1260) {
return 4;
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
if (Rs < 1427) {
return 2;
}
else if (Rs >= 1427) {
return 8;
}
}
}
}
}
else if (diffMVdiffVal >= 17) {
return 1;
}
}
else if (Rs >= 1672) {
if (Rs < 1892) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
return 8;
}
}
else if (Rs >= 1892) {
return 8;
}
}
}
}
else if (SC >= 3320) {
if (Rs < 2544) {
if (MVDiff < 21) {
if (ssDCval < 79) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (MVDiff < 17) {
if (TSC < 10) {
if (Rs < 2222) {
return 8;
}
else if (Rs >= 2222) {
return 4;
}
}
else if (TSC >= 10) {
if (Cs < 1408) {
return 8;
}
else if (Cs >= 1408) {
return 4;
}
}
}
else if (MVDiff >= 17) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 2;
}
}
}
}
else if (ssDCval >= 79) {
if (mcTcor < 94) {
if (AFD < 23) {
if (TSC < 10) {
if (Rs < 2221) {
return 8;
}
else if (Rs >= 2221) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (AFD >= 23) {
return 4;
}
}
else if (mcTcor >= 94) {
if (Rs < 1974) {
return 2;
}
else if (Rs >= 1974) {
return 8;
}
}
}
}
else if (MVDiff >= 21) {
if (RsDiff < 10) {
if (diffMVdiffVal < 9) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 1908) {
return 1;
}
else if (Rs >= 1908) {
return 4;
}
}
}
else if (diffMVdiffVal >= 9) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 1;
}
}
}
else if (RsDiff >= 10) {
if (posBalance < 58) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 1;
}
}
else if (posBalance >= 58) {
if (tcor < 59) {
return 1;
}
else if (tcor >= 59) {
return 4;
}
}
}
}
}
else if (Rs >= 2544) {
if (Cs < 1574) {
return 2;
}
else if (Cs >= 1574) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (mu_mv_mag_sq < 5) {
return 4;
}
else if (mu_mv_mag_sq >= 5) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 310) {
if (AFD < 28) {
if (CsDiff < 13) {
return 2;
}
else if (CsDiff >= 13) {
if (Rs < 2102) {
return 2;
}
else if (Rs >= 2102) {
return 8;
}
}
}
else if (AFD >= 28) {
if (Cs < 1553) {
return 1;
}
else if (Cs >= 1553) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (Rs < 3135) {
if (diffMVdiffVal < -4) {
if (diffRsCsdiff < -61) {
if (Cs < 1474) {
return 2;
}
else if (Cs >= 1474) {
return 8;
}
}
else if (diffRsCsdiff >= -61) {
if (negBalance < 53) {
if (Cs < 1802) {
return 1;
}
else if (Cs >= 1802) {
return 2;
}
}
else if (negBalance >= 53) {
if (diffMVdiffVal < -11) {
if (TSC < 13) {
if (SC < 3592) {
if (Rs < 1515) {
return 4;
}
else if (Rs >= 1515) {
return 8;
}
}
else if (SC >= 3592) {
return 2;
}
}
else if (TSC >= 13) {
return 1;
}
}
else if (diffMVdiffVal >= -11) {
if (negBalance < 55) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (negBalance >= 55) {
if (CsDiff < 15) {
return 4;
}
else if (CsDiff >= 15) {
return 1;
}
}
}
}
}
}
else if (diffMVdiffVal >= -4) {
if (diffTSC < -1) {
if (mcTcor < 90) {
if (CsDiff < 14) {
if (RsCsDiff < 232) {
if (AFD < 18) {
if (Rs < 2018) {
return 1;
}
else if (Rs >= 2018) {
return 8;
}
}
else if (AFD >= 18) {
return 1;
}
}
else if (RsCsDiff >= 232) {
if (Cs < 1905) {
return 8;
}
else if (Cs >= 1905) {
return 2;
}
}
}
else if (CsDiff >= 14) {
return 4;
}
}
else if (mcTcor >= 90) {
if (diffAFD < 3) {
if (AFD < 18) {
if (Rs < 1969) {
return 8;
}
else if (Rs >= 1969) {
return 2;
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (diffAFD >= 3) {
return 2;
}
}
}
else if (diffTSC >= -1) {
if (Rs < 1791) {
if (SC < 2919) {
if (mu_mv_mag_sq < 31) {
if (AFD < 18) {
if (Rs < 1205) {
return 4;
}
else if (Rs >= 1205) {
return 8;
}
}
else if (AFD >= 18) {
return 8;
}
}
else if (mu_mv_mag_sq >= 31) {
if (Rs < 1440) {
return 4;
}
else if (Rs >= 1440) {
return 8;
}
}
}
else if (SC >= 2919) {
if (tcor < 73) {
return 4;
}
else if (tcor >= 73) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 1;
}
}
}
}
else if (Rs >= 1791) {
if (mu_mv_mag_sq < 15) {
if (tcor < 67) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
if (Rs < 2426) {
return 8;
}
else if (Rs >= 2426) {
return 2;
}
}
}
else if (tcor >= 67) {
return 8;
}
}
else if (mu_mv_mag_sq >= 15) {
if (CsDiff < 15) {
if (diffAFD < -1) {
if (Rs < 2706) {
return 8;
}
else if (Rs >= 2706) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (CsDiff >= 15) {
if (Cs < 1870) {
return 8;
}
else if (Cs >= 1870) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 3135) {
if (AFD < 28) {
return 4;
}
else if (AFD >= 28) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (AFD < 18) {
if (refDCval < 100) {
if (Rs < 2782) {
if (TSC < 8) {
if (mcTcor < 97) {
if (MVDiff < 18) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (MVDiff >= 18) {
return 8;
}
}
else if (mcTcor >= 97) {
if (Cs < 1857) {
return 1;
}
else if (Cs >= 1857) {
return 2;
}
}
}
else if (TSC >= 8) {
if (RsCsDiff < 121) {
if (TSC < 13) {
if (CsDiff < 7) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
else if (CsDiff >= 7) {
return 8;
}
}
else if (TSC >= 13) {
return 1;
}
}
else if (RsCsDiff >= 121) {
if (tcor < 75) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (tcor >= 75) {
if (mcTcor < 95) {
if (Cs < 1386) {
if (Cs < 1376) {
return 8;
}
else if (Cs >= 1376) {
return 2;
}
}
else if (Cs >= 1386) {
return 8;
}
}
else if (mcTcor >= 95) {
if (Rs < 2251) {
if (AFD < 17) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 2;
}
}
else if (AFD >= 17) {
return 2;
}
}
else if (Rs >= 2251) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 2782) {
if (MVDiff < 11) {
if (Cs < 1871) {
if (negBalance < 51) {
if (SC < 4394) {
if (SC < 4360) {
return 2;
}
else if (SC >= 4360) {
return 4;
}
}
else if (SC >= 4394) {
return 2;
}
}
else if (negBalance >= 51) {
if (Cs < 1500) {
return 8;
}
else if (Cs >= 1500) {
return 4;
}
}
}
else if (Cs >= 1871) {
if (RsCsDiff < 117) {
return 8;
}
else if (RsCsDiff >= 117) {
return 4;
}
}
}
else if (MVDiff >= 11) {
return 8;
}
}
}
else if (refDCval >= 100) {
if (Rs < 2990) {
if (SC < 3884) {
if (Cs < 1973) {
if (negBalance < 27) {
if (AFD < 9) {
if (Rs < 2115) {
return 8;
}
else if (Rs >= 2115) {
if (Rs < 2217) {
return 4;
}
else if (Rs >= 2217) {
return 1;
}
}
}
else if (AFD >= 9) {
return 2;
}
}
else if (negBalance >= 27) {
if (TSC < 12) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 79) {
return 8;
}
else if (mu_mv_mag_sq >= 79) {
if (Rs < 2036) {
return 8;
}
else if (Rs >= 2036) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (refDCval < 108) {
return 8;
}
else if (refDCval >= 108) {
if (MVDiff < 21) {
if (ssDCval < 121) {
if (diffMVdiffVal < 4) {
return 4;
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
else if (ssDCval >= 121) {
return 2;
}
}
else if (MVDiff >= 21) {
return 8;
}
}
}
}
else if (TSC >= 12) {
if (Rs < 1806) {
return 8;
}
else if (Rs >= 1806) {
return 1;
}
}
}
}
else if (Cs >= 1973) {
if (mcTcor < 94) {
if (Cs < 2091) {
return 8;
}
else if (Cs >= 2091) {
return 1;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
}
else if (SC >= 3884) {
if (SC < 4538) {
if (RsCsDiff < 215) {
return 8;
}
else if (RsCsDiff >= 215) {
if (ssDCval < 111) {
if (SC < 4260) {
return 4;
}
else if (SC >= 4260) {
return 8;
}
}
else if (ssDCval >= 111) {
return 8;
}
}
}
else if (SC >= 4538) {
if (SC < 4551) {
return 4;
}
else if (SC >= 4551) {
if (refDCval < 106) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (refDCval >= 106) {
if (diffTSC < -2) {
if (Rs < 2757) {
return 4;
}
else if (Rs >= 2757) {
return 8;
}
}
else if (diffTSC >= -2) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2990) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 109) {
return 4;
}
else if (ssDCval >= 109) {
if (Cs < 1484) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 1;
}
}
else if (Cs >= 1484) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 2997) {
return 4;
}
else if (Rs >= 2997) {
if (diffRsCsdiff < -28) {
if (ssDCval < 107) {
return 4;
}
else if (ssDCval >= 107) {
if (Rs < 3227) {
return 8;
}
else if (Rs >= 3227) {
return 4;
}
}
}
else if (diffRsCsdiff >= -28) {
return 8;
}
}
}
}
}
}
else if (AFD >= 18) {
if (TSC < 11) {
if (RsCsDiff < 289) {
if (AFD < 26) {
if (mu_mv_mag_sq < 11) {
if (tcor < 83) {
if (negBalance < 54) {
if (mcTcor < 95) {
if (gchDC < 3) {
if (Cs < 1761) {
return 4;
}
else if (Cs >= 1761) {
if (Rs < 2281) {
return 8;
}
else if (Rs >= 2281) {
return 4;
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
else if (negBalance >= 54) {
if (RsCsDiff < 187) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 1;
}
}
else if (RsCsDiff >= 187) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (diffMVdiffVal < -4) {
if (SC < 3282) {
return 4;
}
else if (SC >= 3282) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
if (Rs < 2620) {
return 8;
}
else if (Rs >= 2620) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 83) {
if (mu_mv_mag_sq < 5) {
if (Rs < 2122) {
return 2;
}
else if (Rs >= 2122) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (mcTcor < 83) {
return 4;
}
else if (mcTcor >= 83) {
if (posBalance < 52) {
if (tcor < 80) {
return 2;
}
else if (tcor >= 80) {
return 1;
}
}
else if (posBalance >= 52) {
if (Cs < 1832) {
if (ssDCval < 148) {
if (mu_mv_mag_sq < 23) {
if (Cs < 1371) {
return 4;
}
else if (Cs >= 1371) {
return 2;
}
}
else if (mu_mv_mag_sq >= 23) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
return 8;
}
}
}
else if (ssDCval >= 148) {
return 2;
}
}
else if (Cs >= 1832) {
return 2;
}
}
}
}
}
else if (AFD >= 26) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
if (Rs < 2513) {
return 8;
}
else if (Rs >= 2513) {
return 2;
}
}
}
}
else if (RsCsDiff >= 289) {
if (AFD < 20) {
if (MVDiff < 36) {
if (Rs < 2961) {
return 2;
}
else if (Rs >= 2961) {
return 4;
}
}
else if (MVDiff >= 36) {
return 8;
}
}
else if (AFD >= 20) {
if (Rs < 3306) {
return 2;
}
else if (Rs >= 3306) {
return 1;
}
}
}
}
else if (TSC >= 11) {
if (AFD < 24) {
if (refDCval < 124) {
if (tcor < 80) {
if (ssDCval < 96) {
if (Rs < 2183) {
if (diffRsCsdiff < -20) {
return 8;
}
else if (diffRsCsdiff >= -20) {
if (Cs < 2051) {
return 1;
}
else if (Cs >= 2051) {
return 8;
}
}
}
else if (Rs >= 2183) {
if (ssDCval < 89) {
if (Rs < 2711) {
return 8;
}
else if (Rs >= 2711) {
return 4;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
}
else if (ssDCval >= 96) {
if (TSC < 13) {
if (tcor < 58) {
if (Rs < 1577) {
return 2;
}
else if (Rs >= 1577) {
return 4;
}
}
else if (tcor >= 58) {
if (diffRsCsdiff < 68) {
if (SC < 3262) {
return 4;
}
else if (SC >= 3262) {
if (CsDiff < 6) {
if (Rs < 2959) {
return 2;
}
else if (Rs >= 2959) {
return 4;
}
}
else if (CsDiff >= 6) {
if (diffAFD < -8) {
if (Rs < 1933) {
return 8;
}
else if (Rs >= 1933) {
return 1;
}
}
else if (diffAFD >= -8) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 68) {
return 4;
}
}
}
else if (TSC >= 13) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (tcor < 55) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 8;
}
}
else if (tcor >= 55) {
return 8;
}
}
}
}
}
else if (tcor >= 80) {
if (diffAFD < -2) {
if (RsCsDiff < 220) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
return 8;
}
}
else if (RsCsDiff >= 220) {
if (Rs < 3228) {
return 2;
}
else if (Rs >= 3228) {
return 8;
}
}
}
else if (diffAFD >= -2) {
if (Rs < 1649) {
return 2;
}
else if (Rs >= 1649) {
return 8;
}
}
}
}
else if (refDCval >= 124) {
if (mu_mv_mag_sq < 15) {
if (MVDiff < 19) {
if (Rs < 2097) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
return 2;
}
}
else if (Rs >= 2097) {
return 8;
}
}
else if (MVDiff >= 19) {
if (Rs < 3097) {
return 1;
}
else if (Rs >= 3097) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (ssDCval < 157) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 8;
}
}
}
else if (ssDCval >= 157) {
if (Rs < 2074) {
return 1;
}
else if (Rs >= 2074) {
return 4;
}
}
}
}
}
else if (AFD >= 24) {
if (CsDiff < 8) {
if (Rs < 2321) {
return 8;
}
else if (Rs >= 2321) {
if (RsDiff < 15) {
if (Cs < 1331) {
return 2;
}
else if (Cs >= 1331) {
return 4;
}
}
else if (RsDiff >= 15) {
return 2;
}
}
}
else if (CsDiff >= 8) {
if (TSC < 14) {
if (refDCval < 97) {
if (Cs < 1524) {
if (Cs < 1399) {
return 8;
}
else if (Cs >= 1399) {
return 2;
}
}
else if (Cs >= 1524) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 8;
}
}
}
else if (refDCval >= 97) {
if (diffRsCsdiff < 19) {
if (tcor < 65) {
if (ssDCval < 116) {
if (Rs < 3208) {
return 4;
}
else if (Rs >= 3208) {
return 2;
}
}
else if (ssDCval >= 116) {
return 2;
}
}
else if (tcor >= 65) {
if (mu_mv_mag_sq < 7) {
return 2;
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
}
else if (diffRsCsdiff >= 19) {
if (ssDCval < 115) {
if (ssDCval < 102) {
if (Rs < 2485) {
return 2;
}
else if (Rs >= 2485) {
if (Rs < 2949) {
return 1;
}
else if (Rs >= 2949) {
return 8;
}
}
}
else if (ssDCval >= 102) {
return 2;
}
}
else if (ssDCval >= 115) {
if (diffTSC < 1) {
if (Rs < 2245) {
return 8;
}
else if (Rs >= 2245) {
return 1;
}
}
else if (diffTSC >= 1) {
if (Rs < 1803) {
return 2;
}
else if (Rs >= 1803) {
return 8;
}
}
}
}
}
}
else if (TSC >= 14) {
if (refDCval < 110) {
if (Cs < 1522) {
if (TSC < 19) {
return 2;
}
else if (TSC >= 19) {
return 8;
}
}
else if (Cs >= 1522) {
if (Rs < 1916) {
if (Rs < 1198) {
return 8;
}
else if (Rs >= 1198) {
return 4;
}
}
else if (Rs >= 1916) {
return 8;
}
}
}
else if (refDCval >= 110) {
if (tcor < 37) {
if (Rs < 2152) {
return 8;
}
else if (Rs >= 2152) {
return 1;
}
}
else if (tcor >= 37) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 3354) {
if (tcor < 88) {
if (RsDiff < 12) {
if (refDCval < 86) {
if (negBalance < 45) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
if (mu_mv_mag_sq < 60) {
if (ssDCval < 82) {
return 2;
}
else if (ssDCval >= 82) {
return 4;
}
}
else if (mu_mv_mag_sq >= 60) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
}
else if (negBalance >= 45) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 4049) {
return 4;
}
else if (Rs >= 4049) {
return 8;
}
}
}
}
else if (refDCval >= 86) {
if (tcor < 69) {
return 1;
}
else if (tcor >= 69) {
if (SC < 6834) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 1;
}
}
else if (SC >= 6834) {
return 1;
}
}
}
}
else if (RsDiff >= 12) {
if (MVDiff < 13) {
if (tcor < 83) {
if (RsCsDiff < 306) {
if (ssDCval < 100) {
if (SC < 5320) {
if (Rs < 3419) {
return 8;
}
else if (Rs >= 3419) {
return 2;
}
}
else if (SC >= 5320) {
if (Cs < 1653) {
return 2;
}
else if (Cs >= 1653) {
return 4;
}
}
}
else if (ssDCval >= 100) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (RsCsDiff >= 306) {
if (diffRsCsdiff < -62) {
if (Cs < 1849) {
return 2;
}
else if (Cs >= 1849) {
if (Rs < 3687) {
return 8;
}
else if (Rs >= 3687) {
return 2;
}
}
}
else if (diffRsCsdiff >= -62) {
if (TSC < 10) {
if (tcor < 77) {
return 4;
}
else if (tcor >= 77) {
if (Rs < 5311) {
return 2;
}
else if (Rs >= 5311) {
return 4;
}
}
}
else if (TSC >= 10) {
if (Rs < 3392) {
return 4;
}
else if (Rs >= 3392) {
return 2;
}
}
}
}
}
else if (tcor >= 83) {
if (mu_mv_mag_sq < 76) {
if (diffRsCsdiff < 71) {
return 2;
}
else if (diffRsCsdiff >= 71) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (Rs < 3632) {
return 8;
}
else if (Rs >= 3632) {
return 1;
}
}
}
}
else if (MVDiff >= 13) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (negBalance < 59) {
if (tcor < 76) {
if (SC < 5400) {
if (RsCsDiff < 293) {
if (Cs < 1676) {
if (Rs < 3488) {
return 1;
}
else if (Rs >= 3488) {
return 4;
}
}
else if (Cs >= 1676) {
return 8;
}
}
else if (RsCsDiff >= 293) {
if (ssDCval < 102) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (ssDCval >= 102) {
if (AFD < 23) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 1;
}
}
else if (AFD >= 23) {
if (negBalance < 48) {
if (Rs < 3765) {
return 8;
}
else if (Rs >= 3765) {
return 1;
}
}
else if (negBalance >= 48) {
return 2;
}
}
}
}
}
else if (SC >= 5400) {
if (TSC < 12) {
if (ssDCval < 133) {
if (RsCsDiff < 268) {
return 1;
}
else if (RsCsDiff >= 268) {
if (diffMVdiffVal < -10) {
return 2;
}
else if (diffMVdiffVal >= -10) {
if (Cs < 1363) {
return 2;
}
else if (Cs >= 1363) {
return 8;
}
}
}
}
else if (ssDCval >= 133) {
if (negBalance < 53) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
return 4;
}
}
else if (negBalance >= 53) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
}
else if (TSC >= 12) {
if (gchDC < 6) {
if (negBalance < 56) {
if (tcor < 70) {
return 4;
}
else if (tcor >= 70) {
if (Rs < 5345) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
return 4;
}
}
else if (Rs >= 5345) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (Rs < 4031) {
return 4;
}
else if (Rs >= 4031) {
return 2;
}
}
}
else if (gchDC >= 6) {
if (Rs < 5117) {
return 8;
}
else if (Rs >= 5117) {
return 2;
}
}
}
}
}
else if (tcor >= 76) {
if (refDCval < 143) {
if (MVDiff < 23) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (Cs < 1943) {
if (mcTcor < 95) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (mcTcor >= 95) {
return 8;
}
}
else if (Cs >= 1943) {
if (Rs < 4010) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 8;
}
}
else if (Rs >= 4010) {
return 2;
}
}
}
}
else if (MVDiff >= 23) {
if (RsCsDiff < 619) {
if (negBalance < 50) {
if (Cs < 1755) {
if (AFD < 20) {
if (Rs < 6268) {
return 2;
}
else if (Rs >= 6268) {
return 8;
}
}
else if (AFD >= 20) {
if (Rs < 6489) {
return 8;
}
else if (Rs >= 6489) {
return 2;
}
}
}
else if (Cs >= 1755) {
return 8;
}
}
else if (negBalance >= 50) {
if (ssDCval < 123) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
return 2;
}
}
else if (ssDCval >= 123) {
return 8;
}
}
}
else if (RsCsDiff >= 619) {
return 2;
}
}
}
else if (refDCval >= 143) {
if (CsDiff < 9) {
if (Rs < 5018) {
return 4;
}
else if (Rs >= 5018) {
return 8;
}
}
else if (CsDiff >= 9) {
return 1;
}
}
}
}
else if (negBalance >= 59) {
if (MVDiff < 58) {
if (mcTcor < 89) {
return 2;
}
else if (mcTcor >= 89) {
if (refDCval < 108) {
return 2;
}
else if (refDCval >= 108) {
if (negBalance < 62) {
if (SC < 6763) {
return 1;
}
else if (SC >= 6763) {
return 8;
}
}
else if (negBalance >= 62) {
return 1;
}
}
}
}
else if (MVDiff >= 58) {
if (CsDiff < 8) {
if (Rs < 4411) {
return 4;
}
else if (Rs >= 4411) {
return 8;
}
}
else if (CsDiff >= 8) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
}
}
}
}
}
}
else if (tcor >= 88) {
if (RsDiff < 9) {
if (mu_mv_mag_sq < 74) {
if (Rs < 5095) {
if (ssDCval < 97) {
if (Rs < 4504) {
return 2;
}
else if (Rs >= 4504) {
return 4;
}
}
else if (ssDCval >= 97) {
return 8;
}
}
else if (Rs >= 5095) {
return 1;
}
}
else if (mu_mv_mag_sq >= 74) {
return 1;
}
}
else if (RsDiff >= 9) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (ssDCval < 101) {
if (gchDC < 5) {
if (Cs < 1931) {
if (diffMVdiffVal < -16) {
if (Rs < 3617) {
return 8;
}
else if (Rs >= 3617) {
return 2;
}
}
else if (diffMVdiffVal >= -16) {
return 2;
}
}
else if (Cs >= 1931) {
if (tcor < 89) {
if (Rs < 4695) {
return 8;
}
else if (Rs >= 4695) {
return 2;
}
}
else if (tcor >= 89) {
if (TSC < 6) {
if (refDCval < 92) {
return 2;
}
else if (refDCval >= 92) {
return 8;
}
}
else if (TSC >= 6) {
if (Rs < 3598) {
return 8;
}
else if (Rs >= 3598) {
return 2;
}
}
}
}
}
else if (gchDC >= 5) {
if (Cs < 1647) {
return 1;
}
else if (Cs >= 1647) {
return 8;
}
}
}
else if (ssDCval >= 101) {
if (mu_mv_mag_sq < 87) {
if (SC < 6520) {
if (diffRsCsdiff < -36) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 1;
}
}
else if (diffRsCsdiff >= -36) {
if (ssDCval < 135) {
if (RsDiff < 15) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (MVDiff < 6) {
if (Rs < 4629) {
return 1;
}
else if (Rs >= 4629) {
return 2;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
}
else if (RsDiff >= 15) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
if (Rs < 3947) {
return 8;
}
else if (Rs >= 3947) {
return 1;
}
}
}
}
else if (ssDCval >= 135) {
return 1;
}
}
}
else if (SC >= 6520) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
if (Rs < 6040) {
return 1;
}
else if (Rs >= 6040) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (Rs < 3455) {
return 4;
}
else if (Rs >= 3455) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 2096) {
if (AFD < 12) {
if (ssDCval < 100) {
if (Rs < 5294) {
if (Cs < 2659) {
if (ssDCval < 88) {
if (posBalance < 52) {
if (negBalance < 37) {
if (mcTcor < 93) {
if (Cs < 2369) {
return 1;
}
else if (Cs >= 2369) {
return 4;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
else if (negBalance >= 37) {
if (MVDiff < 5) {
if (Rs < 3565) {
return 4;
}
else if (Rs >= 3565) {
if (Rs < 4371) {
return 8;
}
else if (Rs >= 4371) {
return 2;
}
}
}
else if (MVDiff >= 5) {
return 4;
}
}
}
else if (posBalance >= 52) {
if (ssDCval < 54) {
if (negBalance < 43) {
return 8;
}
else if (negBalance >= 43) {
return 4;
}
}
else if (ssDCval >= 54) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Rs < 2829) {
return 8;
}
else if (Rs >= 2829) {
return 4;
}
}
}
}
}
else if (ssDCval >= 88) {
if (mcTcor < 96) {
if (RsDiff < 11) {
if (Cs < 2526) {
return 8;
}
else if (Cs >= 2526) {
if (Rs < 4087) {
return 4;
}
else if (Rs >= 4087) {
return 8;
}
}
}
else if (RsDiff >= 11) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 8;
}
}
}
else if (mcTcor >= 96) {
return 2;
}
}
}
else if (Cs >= 2659) {
if (refDCval < 83) {
if (negBalance < 49) {
if (SC < 8494) {
if (ssDCval < 55) {
if (Rs < 2342) {
return 1;
}
else if (Rs >= 2342) {
return 8;
}
}
else if (ssDCval >= 55) {
if (Cs < 2746) {
if (posBalance < 50) {
if (Rs < 4034) {
return 1;
}
else if (Rs >= 4034) {
return 4;
}
}
else if (posBalance >= 50) {
return 8;
}
}
else if (Cs >= 2746) {
if (mcTcor < 94) {
if (ssDCval < 70) {
return 4;
}
else if (ssDCval >= 70) {
if (Cs < 3737) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (Cs >= 3737) {
return 8;
}
}
}
else if (mcTcor >= 94) {
return 8;
}
}
}
}
else if (SC >= 8494) {
return 8;
}
}
else if (negBalance >= 49) {
return 8;
}
}
else if (refDCval >= 83) {
if (Cs < 4036) {
if (negBalance < 36) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (negBalance >= 36) {
return 4;
}
}
else if (Cs >= 4036) {
return 8;
}
}
}
}
else if (Rs >= 5294) {
if (ssDCval < 73) {
if (Cs < 3549) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
else if (Cs >= 3549) {
return 8;
}
}
else if (ssDCval >= 73) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (Cs < 4620) {
return 2;
}
else if (Cs >= 4620) {
return 8;
}
}
}
}
}
else if (ssDCval >= 100) {
if (negBalance < 46) {
if (negBalance < 37) {
if (refDCval < 117) {
if (Rs < 5829) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (SC < 8304) {
return 2;
}
else if (SC >= 8304) {
return 8;
}
}
}
else if (Rs >= 5829) {
if (Cs < 2559) {
return 2;
}
else if (Cs >= 2559) {
return 1;
}
}
}
else if (refDCval >= 117) {
if (mu_mv_mag_sq < 62) {
if (ssDCval < 129) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (Rs < 5957) {
return 2;
}
else if (Rs >= 5957) {
return 1;
}
}
}
else if (ssDCval >= 129) {
if (MVDiff < 49) {
if (Rs < 7069) {
return 8;
}
else if (Rs >= 7069) {
return 2;
}
}
else if (MVDiff >= 49) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 62) {
if (mu_mv_mag_sq < 93) {
if (MVDiff < 9) {
if (SC < 6725) {
if (Rs < 3129) {
return 1;
}
else if (Rs >= 3129) {
return 8;
}
}
else if (SC >= 6725) {
return 2;
}
}
else if (MVDiff >= 9) {
if (Rs < 1902) {
return 1;
}
else if (Rs >= 1902) {
if (Rs < 10068) {
return 8;
}
else if (Rs >= 10068) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
return 8;
}
}
}
}
else if (negBalance >= 37) {
if (mcTcor < 93) {
if (mu_mv_mag_sq < 98) {
if (mu_mv_mag_sq < 91) {
return 8;
}
else if (mu_mv_mag_sq >= 91) {
if (Rs < 3736) {
if (Rs < 3648) {
return 4;
}
else if (Rs >= 3648) {
return 2;
}
}
else if (Rs >= 3736) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (mcTcor >= 93) {
if (Rs < 6370) {
if (ssDCval < 108) {
if (Cs < 2796) {
return 8;
}
else if (Cs >= 2796) {
return 4;
}
}
else if (ssDCval >= 108) {
if (Rs < 6218) {
return 8;
}
else if (Rs >= 6218) {
if (Rs < 6334) {
return 4;
}
else if (Rs >= 6334) {
return 8;
}
}
}
}
else if (Rs >= 6370) {
if (Rs < 6970) {
if (Cs < 4110) {
return 4;
}
else if (Cs >= 4110) {
return 8;
}
}
else if (Rs >= 6970) {
if (Rs < 8672) {
return 2;
}
else if (Rs >= 8672) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (tcor < 93) {
if (mu_mv_mag_sq < 84) {
if (MVDiff < 6) {
if (Rs < 3113) {
return 1;
}
else if (Rs >= 3113) {
if (Rs < 3474) {
return 8;
}
else if (Rs >= 3474) {
return 2;
}
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 84) {
if (Cs < 3498) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
if (mu_mv_mag_sq < 92) {
if (Rs < 4645) {
return 1;
}
else if (Rs >= 4645) {
return 8;
}
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
}
else if (Cs >= 3498) {
return 8;
}
}
}
else if (tcor >= 93) {
if (ssDCval < 104) {
return 4;
}
else if (ssDCval >= 104) {
return 8;
}
}
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 76) {
if (RsDiff < 17) {
if (refDCval < 101) {
if (tcor < 64) {
if (refDCval < 63) {
if (negBalance < 57) {
if (AFD < 20) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
return 4;
}
}
else if (AFD >= 20) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 4;
}
}
}
else if (negBalance >= 57) {
if (Cs < 2116) {
return 4;
}
else if (Cs >= 2116) {
return 2;
}
}
}
else if (refDCval >= 63) {
if (negBalance < 54) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
if (Rs < 1573) {
if (mu_mv_mag_sq < 21) {
return 2;
}
else if (mu_mv_mag_sq >= 21) {
if (Rs < 1213) {
return 8;
}
else if (Rs >= 1213) {
return 4;
}
}
}
else if (Rs >= 1573) {
return 8;
}
}
}
else if (negBalance >= 54) {
if (AFD < 28) {
if (AFD < 17) {
if (Rs < 1760) {
return 4;
}
else if (Rs >= 1760) {
return 8;
}
}
else if (AFD >= 17) {
if (diffMVdiffVal < -16) {
if (Rs < 3126) {
return 2;
}
else if (Rs >= 3126) {
return 8;
}
}
else if (diffMVdiffVal >= -16) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
if (Cs < 2899) {
return 8;
}
else if (Cs >= 2899) {
return 4;
}
}
}
}
}
else if (AFD >= 28) {
if (CsDiff < 19) {
if (CsDiff < 13) {
return 1;
}
else if (CsDiff >= 13) {
if (Rs < 968) {
return 2;
}
else if (Rs >= 968) {
return 4;
}
}
}
else if (CsDiff >= 19) {
return 8;
}
}
}
}
}
else if (tcor >= 64) {
if (negBalance < 47) {
if (mu_mv_mag_sq < 58) {
if (RsCsDiff < 193) {
if (mcTcor < 92) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (mcTcor >= 92) {
if (tcor < 82) {
if (Cs < 3570) {
if (Cs < 2138) {
if (Rs < 2731) {
return 8;
}
else if (Rs >= 2731) {
return 2;
}
}
else if (Cs >= 2138) {
return 4;
}
}
else if (Cs >= 3570) {
return 8;
}
}
else if (tcor >= 82) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 8;
}
}
}
}
else if (RsCsDiff >= 193) {
if (Rs < 2822) {
if (diffAFD < -2) {
if (TSC < 10) {
if (Rs < 1640) {
return 1;
}
else if (Rs >= 1640) {
return 8;
}
}
else if (TSC >= 10) {
return 4;
}
}
else if (diffAFD >= -2) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 2;
}
}
}
else if (Rs >= 2822) {
if (ssDCval < 88) {
if (posBalance < 56) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (CsDiff < 18) {
return 2;
}
else if (CsDiff >= 18) {
return 4;
}
}
}
}
else if (posBalance >= 56) {
if (Rs < 3588) {
return 4;
}
else if (Rs >= 3588) {
return 8;
}
}
}
else if (ssDCval >= 88) {
if (MVDiff < 11) {
if (Cs < 2952) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 1;
}
}
else if (Cs >= 2952) {
return 4;
}
}
else if (MVDiff >= 11) {
if (Rs < 5053) {
if (Rs < 3197) {
return 2;
}
else if (Rs >= 3197) {
return 8;
}
}
else if (Rs >= 5053) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (mcTcor < 93) {
if (posBalance < 50) {
if (Rs < 3034) {
return 2;
}
else if (Rs >= 3034) {
return 4;
}
}
else if (posBalance >= 50) {
if (RsCsDiff < 247) {
if (Cs < 2595) {
if (Rs < 2930) {
return 8;
}
else if (Rs >= 2930) {
if (Rs < 3884) {
return 4;
}
else if (Rs >= 3884) {
return 8;
}
}
}
else if (Cs >= 2595) {
if (Rs < 4373) {
return 4;
}
else if (Rs >= 4373) {
return 2;
}
}
}
else if (RsCsDiff >= 247) {
if (Rs < 2058) {
return 1;
}
else if (Rs >= 2058) {
if (Cs < 3453) {
return 8;
}
else if (Cs >= 3453) {
return 4;
}
}
}
}
}
else if (mcTcor >= 93) {
if (negBalance < 37) {
if (Rs < 5207) {
return 8;
}
else if (Rs >= 5207) {
return 4;
}
}
else if (negBalance >= 37) {
return 1;
}
}
}
}
else if (negBalance >= 47) {
if (ssDCval < 91) {
if (TSC < 10) {
if (MVDiff < 5) {
if (RsDiff < 11) {
if (AFD < 15) {
if (mcTcor < 90) {
if (Cs < 2826) {
return 8;
}
else if (Cs >= 2826) {
return 4;
}
}
else if (mcTcor >= 90) {
if (Rs < 2870) {
return 2;
}
else if (Rs >= 2870) {
if (Rs < 3408) {
return 1;
}
else if (Rs >= 3408) {
return 2;
}
}
}
}
else if (AFD >= 15) {
if (ssDCval < 87) {
if (Cs < 3818) {
return 1;
}
else if (Cs >= 3818) {
return 8;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
}
else if (RsDiff >= 11) {
if (Rs < 3193) {
return 8;
}
else if (Rs >= 3193) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (CsDiff < 10) {
if (Rs < 2022) {
return 4;
}
else if (Rs >= 2022) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
if (Rs < 4593) {
if (SC < 5275) {
return 8;
}
else if (SC >= 5275) {
if (CsDiff < 9) {
return 4;
}
else if (CsDiff >= 9) {
return 8;
}
}
}
else if (Rs >= 4593) {
return 8;
}
}
}
}
else if (CsDiff >= 10) {
if (MVDiff < 22) {
if (diffMVdiffVal < 3) {
if (Rs < 1081) {
return 8;
}
else if (Rs >= 1081) {
if (CsDiff < 11) {
if (gchDC < 1) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
return 8;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (CsDiff >= 11) {
return 4;
}
}
}
else if (diffMVdiffVal >= 3) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 2;
}
}
}
else if (MVDiff >= 22) {
if (AFD < 18) {
if (Rs < 1183) {
return 4;
}
else if (Rs >= 1183) {
return 1;
}
}
else if (AFD >= 18) {
return 2;
}
}
}
}
}
else if (TSC >= 10) {
if (Cs < 3793) {
if (SC < 4139) {
if (tcor < 69) {
if (Rs < 977) {
return 8;
}
else if (Rs >= 977) {
return 2;
}
}
else if (tcor >= 69) {
if (Rs < 1002) {
return 2;
}
else if (Rs >= 1002) {
return 4;
}
}
}
else if (SC >= 4139) {
if (MVDiff < 10) {
if (mcTcor < 90) {
if (mcTcor < 86) {
if (ssDCval < 89) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (tcor < 72) {
return 4;
}
else if (tcor >= 72) {
if (negBalance < 50) {
return 4;
}
else if (negBalance >= 50) {
return 8;
}
}
}
}
else if (ssDCval >= 89) {
return 4;
}
}
else if (mcTcor >= 86) {
if (TSC < 12) {
if (Rs < 2803) {
if (ssDCval < 73) {
if (Rs < 2575) {
return 1;
}
else if (Rs >= 2575) {
return 8;
}
}
else if (ssDCval >= 73) {
if (Rs < 2613) {
return 8;
}
else if (Rs >= 2613) {
return 4;
}
}
}
else if (Rs >= 2803) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
return 1;
}
}
}
else if (TSC >= 12) {
if (tcor < 83) {
if (diffMVdiffVal < -1) {
if (AFD < 20) {
if (Rs < 2572) {
return 4;
}
else if (Rs >= 2572) {
return 8;
}
}
else if (AFD >= 20) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (tcor >= 83) {
if (Rs < 4998) {
return 8;
}
else if (Rs >= 4998) {
return 2;
}
}
}
}
}
else if (mcTcor >= 90) {
if (Cs < 3514) {
if (RsCsDiff < 154) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
return 8;
}
}
else if (RsCsDiff >= 154) {
if (TSC < 13) {
if (SC < 4852) {
return 2;
}
else if (SC >= 4852) {
if (ssDCval < 79) {
if (Cs < 3098) {
return 8;
}
else if (Cs >= 3098) {
return 4;
}
}
else if (ssDCval >= 79) {
if (Rs < 2514) {
return 8;
}
else if (Rs >= 2514) {
if (Cs < 2950) {
return 4;
}
else if (Cs >= 2950) {
return 2;
}
}
}
}
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (Cs >= 3514) {
return 2;
}
}
}
else if (MVDiff >= 10) {
if (RsCsDiff < 223) {
if (Rs < 3095) {
if (negBalance < 57) {
if (Rs < 1600) {
return 4;
}
else if (Rs >= 1600) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
return 4;
}
}
}
else if (negBalance >= 57) {
if (mcTcor < 79) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 4;
}
}
else if (mcTcor >= 79) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 2373) {
return 4;
}
else if (Rs >= 2373) {
return 1;
}
}
}
}
}
else if (Rs >= 3095) {
if (SC < 5966) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 4;
}
}
else if (SC >= 5966) {
if (mu_mv_mag_sq < 71) {
return 8;
}
else if (mu_mv_mag_sq >= 71) {
return 4;
}
}
}
}
else if (RsCsDiff >= 223) {
if (posBalance < 58) {
if (Rs < 2873) {
if (TSC < 12) {
if (Rs < 1939) {
return 4;
}
else if (Rs >= 1939) {
return 8;
}
}
else if (TSC >= 12) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (Rs >= 2873) {
if (MVDiff < 15) {
if (negBalance < 57) {
if (Rs < 3798) {
return 4;
}
else if (Rs >= 3798) {
return 1;
}
}
else if (negBalance >= 57) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (MVDiff >= 15) {
if (TSC < 12) {
if (Rs < 4140) {
return 8;
}
else if (Rs >= 4140) {
return 2;
}
}
else if (TSC >= 12) {
if (TSC < 15) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
return 8;
}
}
else if (TSC >= 15) {
return 8;
}
}
}
}
}
else if (posBalance >= 58) {
if (ssDCval < 86) {
if (Cs < 2558) {
if (Cs < 2512) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
if (negBalance < 52) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 2;
}
}
else if (negBalance >= 52) {
return 8;
}
}
}
else if (Cs >= 2512) {
return 2;
}
}
else if (Cs >= 2558) {
if (diffRsCsdiff < 100) {
if (Cs < 2573) {
return 1;
}
else if (Cs >= 2573) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
return 1;
}
}
}
else if (diffRsCsdiff >= 100) {
if (Rs < 3570) {
return 8;
}
else if (Rs >= 3570) {
return 4;
}
}
}
}
else if (ssDCval >= 86) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (TSC < 15) {
return 4;
}
else if (TSC >= 15) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 3793) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (Cs < 4035) {
return 2;
}
else if (Cs >= 4035) {
return 8;
}
}
}
}
}
else if (ssDCval >= 91) {
if (Rs < 4688) {
if (Cs < 2535) {
if (mu_mv_mag_sq < 7) {
if (RsCsDiff < 247) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (Rs < 2484) {
return 1;
}
else if (Rs >= 2484) {
return 8;
}
}
}
}
else if (RsCsDiff >= 247) {
if (Cs < 2151) {
return 8;
}
else if (Cs >= 2151) {
if (negBalance < 60) {
return 4;
}
else if (negBalance >= 60) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (mcTcor < 94) {
if (Cs < 2103) {
return 4;
}
else if (Cs >= 2103) {
if (ssDCval < 99) {
return 8;
}
else if (ssDCval >= 99) {
return 2;
}
}
}
else if (mcTcor >= 94) {
return 2;
}
}
}
else if (Cs >= 2535) {
if (TSC < 10) {
if (Rs < 2875) {
return 8;
}
else if (Rs >= 2875) {
return 4;
}
}
else if (TSC >= 10) {
if (MVDiff < 22) {
if (diffAFD < -3) {
if (Rs < 2394) {
return 4;
}
else if (Rs >= 2394) {
return 8;
}
}
else if (diffAFD >= -3) {
if (SC < 7017) {
if (diffMVdiffVal < -10) {
return 2;
}
else if (diffMVdiffVal >= -10) {
if (diffRsCsdiff < -70) {
if (Rs < 3400) {
return 1;
}
else if (Rs >= 3400) {
return 4;
}
}
else if (diffRsCsdiff >= -70) {
return 4;
}
}
}
else if (SC >= 7017) {
if (RsCsDiff < 255) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (RsCsDiff >= 255) {
return 4;
}
}
}
}
else if (MVDiff >= 22) {
if (RsCsDiff < 326) {
if (mu_mv_mag_sq < 2) {
return 4;
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
else if (RsCsDiff >= 326) {
return 4;
}
}
}
}
}
else if (Rs >= 4688) {
if (diffRsCsdiff < -162) {
return 2;
}
else if (diffRsCsdiff >= -162) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 101) {
if (AFD < 21) {
if (mcTcor < 86) {
if (refDCval < 125) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (CsDiff < 12) {
return 4;
}
else if (CsDiff >= 12) {
return 8;
}
}
}
else if (refDCval >= 125) {
if (Cs < 2619) {
if (diffMVdiffVal < -21) {
return 1;
}
else if (diffMVdiffVal >= -21) {
return 8;
}
}
else if (Cs >= 2619) {
if (mcTcor < 84) {
if (posBalance < 60) {
if (refDCval < 139) {
if (ssDCval < 134) {
if (Cs < 2812) {
return 8;
}
else if (Cs >= 2812) {
return 1;
}
}
else if (ssDCval >= 134) {
if (SC < 7490) {
return 8;
}
else if (SC >= 7490) {
if (mu_mv_mag_sq < 20) {
return 8;
}
else if (mu_mv_mag_sq >= 20) {
if (diffMVdiffVal < 1) {
return 1;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 4186) {
return 8;
}
else if (Rs >= 4186) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 139) {
return 1;
}
}
else if (posBalance >= 60) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
if (Rs < 3260) {
return 2;
}
else if (Rs >= 3260) {
return 4;
}
}
}
}
else if (mcTcor >= 84) {
if (Rs < 4984) {
return 4;
}
else if (Rs >= 4984) {
return 8;
}
}
}
}
}
else if (mcTcor >= 86) {
if (RsDiff < 9) {
if (Rs < 3512) {
if (negBalance < 44) {
if (SC < 4445) {
return 1;
}
else if (SC >= 4445) {
if (diffRsCsdiff < -8) {
if (Rs < 2349) {
return 1;
}
else if (Rs >= 2349) {
return 8;
}
}
else if (diffRsCsdiff >= -8) {
return 8;
}
}
}
else if (negBalance >= 44) {
if (SC < 5456) {
if (RsCsDiff < 121) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (Cs < 2113) {
return 8;
}
else if (Cs >= 2113) {
return 4;
}
}
}
else if (RsCsDiff >= 121) {
return 8;
}
}
else if (SC >= 5456) {
if (ssDCval < 108) {
return 4;
}
else if (ssDCval >= 108) {
if (Rs < 3184) {
return 1;
}
else if (Rs >= 3184) {
return 8;
}
}
}
}
}
else if (Rs >= 3512) {
return 8;
}
}
else if (RsDiff >= 9) {
if (MVDiff < 4) {
if (mcTcor < 95) {
if (Cs < 2428) {
if (Rs < 2977) {
return 1;
}
else if (Rs >= 2977) {
return 8;
}
}
else if (Cs >= 2428) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
if (mu_mv_mag_sq < 70) {
if (RsCsDiff < 191) {
if (Rs < 3694) {
return 4;
}
else if (Rs >= 3694) {
return 8;
}
}
else if (RsCsDiff >= 191) {
return 4;
}
}
else if (mu_mv_mag_sq >= 70) {
return 8;
}
}
}
}
else if (mcTcor >= 95) {
return 4;
}
}
else if (MVDiff >= 4) {
if (refDCval < 158) {
if (negBalance < 46) {
if (CsDiff < 10) {
if (Cs < 2655) {
if (diffAFD < -3) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (Cs >= 2655) {
if (tcor < 88) {
if (Rs < 4425) {
return 2;
}
else if (Rs >= 4425) {
return 4;
}
}
else if (tcor >= 88) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
}
else if (CsDiff >= 10) {
if (CsDiff < 16) {
if (AFD < 18) {
if (tcor < 91) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
if (Rs < 3787) {
return 2;
}
else if (Rs >= 3787) {
return 8;
}
}
}
else if (tcor >= 91) {
return 8;
}
}
else if (AFD >= 18) {
if (CsDiff < 12) {
return 2;
}
else if (CsDiff >= 12) {
if (Rs < 3650) {
return 1;
}
else if (Rs >= 3650) {
return 4;
}
}
}
}
else if (CsDiff >= 16) {
if (Rs < 6699) {
return 2;
}
else if (Rs >= 6699) {
return 1;
}
}
}
}
else if (negBalance >= 46) {
if (Rs < 3676) {
if (ssDCval < 108) {
if (refDCval < 104) {
if (mu_mv_mag_sq < 57) {
if (Cs < 2873) {
return 8;
}
else if (Cs >= 2873) {
return 4;
}
}
else if (mu_mv_mag_sq >= 57) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 2;
}
}
}
else if (refDCval >= 104) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (Rs < 2996) {
return 8;
}
else if (Rs >= 2996) {
return 4;
}
}
}
}
else if (ssDCval >= 108) {
if (mu_mv_mag_sq < 33) {
if (negBalance < 59) {
if (Rs < 1503) {
return 1;
}
else if (Rs >= 1503) {
return 8;
}
}
else if (negBalance >= 59) {
if (Rs < 3326) {
return 1;
}
else if (Rs >= 3326) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (mcTcor < 91) {
if (Cs < 3145) {
return 8;
}
else if (Cs >= 3145) {
return 4;
}
}
else if (mcTcor >= 91) {
if (diffRsCsdiff < 30) {
if (tcor < 81) {
if (Rs < 2622) {
return 8;
}
else if (Rs >= 2622) {
return 4;
}
}
else if (tcor >= 81) {
return 8;
}
}
else if (diffRsCsdiff >= 30) {
return 4;
}
}
}
}
}
else if (Rs >= 3676) {
if (tcor < 86) {
if (diffRsCsdiff < -142) {
return 4;
}
else if (diffRsCsdiff >= -142) {
if (Cs < 2113) {
return 4;
}
else if (Cs >= 2113) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
return 4;
}
}
}
}
else if (tcor >= 86) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (negBalance < 49) {
if (Rs < 5258) {
return 8;
}
else if (Rs >= 5258) {
return 4;
}
}
else if (negBalance >= 49) {
if (negBalance < 55) {
return 1;
}
else if (negBalance >= 55) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 158) {
if (Cs < 3497) {
return 4;
}
else if (Cs >= 3497) {
return 1;
}
}
}
}
}
}
else if (AFD >= 21) {
if (MVDiff < 26) {
if (refDCval < 134) {
if (tcor < 84) {
if (Cs < 2537) {
if (diffAFD < 3) {
if (MVDiff < 11) {
if (Rs < 3239) {
return 2;
}
else if (Rs >= 3239) {
return 8;
}
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (diffAFD >= 3) {
return 8;
}
}
else if (Cs >= 2537) {
if (AFD < 25) {
if (MVDiff < 15) {
if (Cs < 2767) {
if (Rs < 2464) {
return 2;
}
else if (Rs >= 2464) {
return 8;
}
}
else if (Cs >= 2767) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (Cs < 3005) {
return 4;
}
else if (Cs >= 3005) {
return 8;
}
}
}
}
else if (MVDiff >= 15) {
if (Cs < 2901) {
return 8;
}
else if (Cs >= 2901) {
if (Rs < 3496) {
return 2;
}
else if (Rs >= 3496) {
return 8;
}
}
}
}
else if (AFD >= 25) {
if (diffMVdiffVal < -3) {
if (CsDiff < 14) {
return 1;
}
else if (CsDiff >= 14) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
}
}
else if (tcor >= 84) {
return 2;
}
}
else if (refDCval >= 134) {
if (RsDiff < 12) {
if (Rs < 3120) {
return 2;
}
else if (Rs >= 3120) {
return 8;
}
}
else if (RsDiff >= 12) {
return 8;
}
}
}
else if (MVDiff >= 26) {
if (RsCsDiff < 294) {
return 8;
}
else if (RsCsDiff >= 294) {
if (mcTcor < 78) {
return 8;
}
else if (mcTcor >= 78) {
if (Cs < 5119) {
if (mcTcor < 86) {
if (SC < 6358) {
return 8;
}
else if (SC >= 6358) {
return 2;
}
}
else if (mcTcor >= 86) {
if (Rs < 3769) {
return 2;
}
else if (Rs >= 3769) {
return 8;
}
}
}
else if (Cs >= 5119) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 17) {
if (TSC < 8) {
if (ssDCval < 127) {
if (posBalance < 37) {
if (Cs < 2801) {
return 4;
}
else if (Cs >= 2801) {
return 8;
}
}
else if (posBalance >= 37) {
if (Rs < 5527) {
return 4;
}
else if (Rs >= 5527) {
return 1;
}
}
}
else if (ssDCval >= 127) {
if (Cs < 2476) {
return 2;
}
else if (Cs >= 2476) {
return 8;
}
}
}
else if (TSC >= 8) {
if (ssDCval < 139) {
if (TSC < 22) {
if (Rs < 5751) {
if (negBalance < 54) {
if (tcor < 79) {
if (Cs < 2510) {
if (mcTcor < 86) {
if (mu_mv_mag_sq < 7) {
if (Rs < 4082) {
return 8;
}
else if (Rs >= 4082) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
else if (mcTcor >= 86) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (RsDiff < 19) {
if (refDCval < 87) {
return 8;
}
else if (refDCval >= 87) {
if (Rs < 3096) {
return 1;
}
else if (Rs >= 3096) {
return 4;
}
}
}
else if (RsDiff >= 19) {
return 4;
}
}
}
}
else if (Cs >= 2510) {
if (tcor < 50) {
return 8;
}
else if (tcor >= 50) {
if (Rs < 5154) {
if (posBalance < 62) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 46) {
if (Rs < 3565) {
if (Cs < 2717) {
return 2;
}
else if (Cs >= 2717) {
return 4;
}
}
else if (Rs >= 3565) {
if (Rs < 4981) {
return 2;
}
else if (Rs >= 4981) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 46) {
return 4;
}
}
else if (posBalance >= 55) {
if (SC < 6061) {
return 8;
}
else if (SC >= 6061) {
return 4;
}
}
}
else if (posBalance >= 62) {
if (diffAFD < 5) {
if (Rs < 4550) {
if (TSC < 19) {
return 8;
}
else if (TSC >= 19) {
return 4;
}
}
else if (Rs >= 4550) {
return 4;
}
}
else if (diffAFD >= 5) {
if (Rs < 3644) {
return 1;
}
else if (Rs >= 3644) {
return 4;
}
}
}
}
else if (Rs >= 5154) {
if (CsDiff < 17) {
if (Cs < 3203) {
return 8;
}
else if (Cs >= 3203) {
return 2;
}
}
else if (CsDiff >= 17) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
return 4;
}
}
}
}
}
}
else if (tcor >= 79) {
if (MVDiff < 19) {
if (RsCsDiff < 674) {
if (posBalance < 60) {
if (TSC < 14) {
if (CsDiff < 9) {
if (Rs < 5090) {
return 8;
}
else if (Rs >= 5090) {
if (Rs < 5362) {
return 2;
}
else if (Rs >= 5362) {
return 4;
}
}
}
else if (CsDiff >= 9) {
return 2;
}
}
else if (TSC >= 14) {
return 4;
}
}
else if (posBalance >= 60) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (RsCsDiff >= 674) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 4;
}
}
}
else if (MVDiff >= 19) {
if (Cs < 3127) {
if (SC < 5691) {
return 4;
}
else if (SC >= 5691) {
return 8;
}
}
else if (Cs >= 3127) {
return 1;
}
}
}
}
else if (negBalance >= 54) {
if (AFD < 31) {
if (tcor < 58) {
if (Rs < 3125) {
if (Cs < 2130) {
return 8;
}
else if (Cs >= 2130) {
return 1;
}
}
else if (Rs >= 3125) {
return 4;
}
}
else if (tcor >= 58) {
if (TSC < 16) {
if (SC < 7003) {
if (RsCsDiff < 439) {
if (gchDC < 1) {
if (Cs < 2395) {
return 2;
}
else if (Cs >= 2395) {
return 8;
}
}
else if (gchDC >= 1) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 4;
}
}
}
else if (RsCsDiff >= 439) {
if (ssDCval < 67) {
if (Cs < 2187) {
return 8;
}
else if (Cs >= 2187) {
return 4;
}
}
else if (ssDCval >= 67) {
return 8;
}
}
}
else if (SC >= 7003) {
if (mcTcor < 90) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
if (Rs < 4271) {
return 8;
}
else if (Rs >= 4271) {
return 2;
}
}
}
else if (mcTcor >= 90) {
if (Rs < 3743) {
return 2;
}
else if (Rs >= 3743) {
return 8;
}
}
}
}
else if (TSC >= 16) {
if (refDCval < 137) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
if (Rs < 4327) {
return 8;
}
else if (Rs >= 4327) {
return 4;
}
}
}
else if (refDCval >= 137) {
return 2;
}
}
}
}
else if (AFD >= 31) {
if (mu_mv_mag_sq < 11) {
if (RsDiff < 21) {
return 1;
}
else if (RsDiff >= 21) {
return 2;
}
}
else if (mu_mv_mag_sq >= 11) {
if (Cs < 2856) {
return 1;
}
else if (Cs >= 2856) {
return 8;
}
}
}
}
}
else if (Rs >= 5751) {
if (negBalance < 46) {
if (negBalance < 39) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
else if (negBalance >= 39) {
if (ssDCval < 81) {
if (MVDiff < 21) {
if (Rs < 6587) {
return 4;
}
else if (Rs >= 6587) {
return 1;
}
}
else if (MVDiff >= 21) {
return 2;
}
}
else if (ssDCval >= 81) {
if (Cs < 3106) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
else if (Cs >= 3106) {
return 1;
}
}
}
}
else if (negBalance >= 46) {
if (AFD < 40) {
if (negBalance < 48) {
if (RsCsDiff < 689) {
return 2;
}
else if (RsCsDiff >= 689) {
return 4;
}
}
else if (negBalance >= 48) {
return 4;
}
}
else if (AFD >= 40) {
return 8;
}
}
}
}
else if (TSC >= 22) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (Cs < 2447) {
if (Rs < 3256) {
return 8;
}
else if (Rs >= 3256) {
return 1;
}
}
else if (Cs >= 2447) {
return 8;
}
}
}
}
else if (ssDCval >= 139) {
if (refDCval < 166) {
if (gchDC < 2) {
if (diffRsCsdiff < 365) {
if (AFD < 33) {
if (Rs < 4677) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
return 8;
}
}
}
else if (Rs >= 4677) {
return 4;
}
}
else if (AFD >= 33) {
if (Rs < 7091) {
return 2;
}
else if (Rs >= 7091) {
return 4;
}
}
}
else if (diffRsCsdiff >= 365) {
if (CsDiff < 19) {
return 8;
}
else if (CsDiff >= 19) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 157) {
return 8;
}
else if (ssDCval >= 157) {
if (Rs < 4336) {
return 8;
}
else if (Rs >= 4336) {
return 4;
}
}
}
}
else if (refDCval >= 166) {
if (Cs < 3720) {
return 2;
}
else if (Cs >= 3720) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (SC < 6984) {
if (negBalance < 54) {
if (Rs < 4099) {
if (Rs < 3215) {
if (diffTSC < -3) {
return 2;
}
else if (diffTSC >= -3) {
if (mu_mv_mag_sq < 100) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 1;
}
}
else if (mu_mv_mag_sq >= 100) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
}
}
else if (Rs >= 3215) {
if (refDCval < 81) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (SC < 6817) {
return 8;
}
else if (SC >= 6817) {
if (Rs < 3824) {
return 1;
}
else if (Rs >= 3824) {
return 8;
}
}
}
}
else if (refDCval >= 81) {
if (CsDiff < 9) {
if (SC < 6174) {
return 8;
}
else if (SC >= 6174) {
return 4;
}
}
else if (CsDiff >= 9) {
return 4;
}
}
}
}
else if (Rs >= 4099) {
return 1;
}
}
else if (negBalance >= 54) {
if (Cs < 2611) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
if (Rs < 3784) {
return 8;
}
else if (Rs >= 3784) {
return 1;
}
}
}
else if (Cs >= 2611) {
if (SC < 6765) {
return 4;
}
else if (SC >= 6765) {
return 8;
}
}
}
}
else if (SC >= 6984) {
if (negBalance < 51) {
if (tcor < 85) {
if (TSC < 15) {
if (ssDCval < 81) {
if (SC < 8513) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 4;
}
}
else if (SC >= 8513) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
return 4;
}
}
}
else if (ssDCval >= 81) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
else if (TSC >= 15) {
if (ssDCval < 122) {
return 2;
}
else if (ssDCval >= 122) {
if (RsDiff < 21) {
return 8;
}
else if (RsDiff >= 21) {
return 4;
}
}
}
}
else if (tcor >= 85) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
if (ssDCval < 92) {
if (SC < 7955) {
return 4;
}
else if (SC >= 7955) {
return 8;
}
}
else if (ssDCval >= 92) {
return 8;
}
}
}
}
else if (negBalance >= 51) {
if (negBalance < 63) {
if (Rs < 7760) {
if (Rs < 3179) {
return 4;
}
else if (Rs >= 3179) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
if (ssDCval < 122) {
if (diffRsCsdiff < 59) {
return 8;
}
else if (diffRsCsdiff >= 59) {
if (diffRsCsdiff < 72) {
if (Cs < 4589) {
return 8;
}
else if (Cs >= 4589) {
return 4;
}
}
else if (diffRsCsdiff >= 72) {
return 8;
}
}
}
else if (ssDCval >= 122) {
return 4;
}
}
}
}
else if (Rs >= 7760) {
return 4;
}
}
else if (negBalance >= 63) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1661) {
if (Rs < 3097) {
if (Rs < 1454) {
if (refDCval < 122) {
if (negBalance < 1) {
if (Rs < 730) {
return 1;
}
else if (Rs >= 730) {
return 2;
}
}
else if (negBalance >= 1) {
if (refDCval < 63) {
if (AFD < 3) {
if (mu_mv_mag_sq < 128) {
if (Cs < 1369) {
if (ssDCval < 55) {
if (Rs < 1154) {
return 8;
}
else if (Rs >= 1154) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 4;
}
}
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (Cs >= 1369) {
if (negBalance < 24) {
return 1;
}
else if (negBalance >= 24) {
if (Rs < 1386) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (Rs >= 1386) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (SC < 2901) {
if (Cs < 1343) {
if (Rs < 1074) {
return 8;
}
else if (Rs >= 1074) {
return 4;
}
}
else if (Cs >= 1343) {
return 8;
}
}
else if (SC >= 2901) {
return 1;
}
}
}
else if (AFD >= 3) {
if (SC < 2582) {
return 8;
}
else if (SC >= 2582) {
if (tcor < 96) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (tcor >= 96) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (Rs < 1200) {
return 8;
}
else if (Rs >= 1200) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 63) {
if (ssDCval < 77) {
if (Cs < 1433) {
if (CsDiff < 3) {
if (Rs < 1274) {
if (negBalance < 30) {
if (Rs < 313) {
return 4;
}
else if (Rs >= 313) {
return 8;
}
}
else if (negBalance >= 30) {
return 2;
}
}
else if (Rs >= 1274) {
return 1;
}
}
else if (CsDiff >= 3) {
if (Rs < 1388) {
return 4;
}
else if (Rs >= 1388) {
if (Cs < 1383) {
return 8;
}
else if (Cs >= 1383) {
return 4;
}
}
}
}
else if (Cs >= 1433) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (ssDCval >= 77) {
if (Rs < 953) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Cs < 1358) {
return 1;
}
else if (Cs >= 1358) {
return 4;
}
}
}
else if (Rs >= 953) {
return 8;
}
}
}
}
}
else if (refDCval >= 122) {
if (AFD < 2) {
if (posBalance < 28) {
if (ssDCval < 132) {
return 8;
}
else if (ssDCval >= 132) {
if (Rs < 1327) {
return 1;
}
else if (Rs >= 1327) {
return 8;
}
}
}
else if (posBalance >= 28) {
if (Rs < 1382) {
return 4;
}
else if (Rs >= 1382) {
return 8;
}
}
}
else if (AFD >= 2) {
if (mcTcor < 96) {
if (ssDCval < 133) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 933) {
return 8;
}
else if (Rs >= 933) {
return 4;
}
}
}
else if (ssDCval >= 133) {
return 1;
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 120) {
return 8;
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 1022) {
return 8;
}
else if (Rs >= 1022) {
if (Rs < 1270) {
return 4;
}
else if (Rs >= 1270) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1454) {
if (ssDCval < 86) {
if (mcTcor < 98) {
if (Rs < 2218) {
if (CsDiff < 2) {
if (Cs < 1499) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
if (Rs < 1829) {
return 4;
}
else if (Rs >= 1829) {
return 8;
}
}
}
else if (Cs >= 1499) {
if (Rs < 1846) {
return 4;
}
else if (Rs >= 1846) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (refDCval < 86) {
return 8;
}
else if (refDCval >= 86) {
return 4;
}
}
}
else if (Rs >= 2218) {
if (Cs < 1381) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (Cs >= 1381) {
if (SC < 4171) {
return 8;
}
else if (SC >= 4171) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (Rs < 2930) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (Rs >= 2930) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 18) {
if (SC < 4462) {
if (Cs < 1350) {
if (Rs < 2296) {
if (Cs < 1314) {
return 2;
}
else if (Cs >= 1314) {
return 8;
}
}
else if (Rs >= 2296) {
if (Rs < 2297) {
return 2;
}
else if (Rs >= 2297) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (SC < 3638) {
if (AFD < 1) {
if (negBalance < 14) {
return 2;
}
else if (negBalance >= 14) {
return 8;
}
}
else if (AFD >= 1) {
return 8;
}
}
else if (SC >= 3638) {
if (negBalance < 13) {
return 8;
}
else if (negBalance >= 13) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1350) {
if (SC < 3298) {
if (Cs < 1443) {
return 8;
}
else if (Cs >= 1443) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (SC >= 3298) {
return 8;
}
}
}
else if (SC >= 4462) {
return 1;
}
}
else if (negBalance >= 18) {
if (Cs < 1338) {
if (Rs < 2327) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (Rs < 1833) {
return 4;
}
else if (Rs >= 1833) {
return 8;
}
}
}
else if (Rs >= 2327) {
return 2;
}
}
else if (Cs >= 1338) {
if (ssDCval < 78) {
if (ssDCval < 69) {
if (negBalance < 25) {
if (Rs < 1954) {
return 1;
}
else if (Rs >= 1954) {
return 8;
}
}
else if (negBalance >= 25) {
if (diffTSC < 1) {
if (Rs < 2252) {
return 4;
}
else if (Rs >= 2252) {
return 4;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (ssDCval >= 78) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (CsDiff < 2) {
if (diffRsCsdiff < -202) {
if (Cs < 1434) {
if (Rs < 2447) {
return 4;
}
else if (Rs >= 2447) {
return 1;
}
}
else if (Cs >= 1434) {
return 8;
}
}
else if (diffRsCsdiff >= -202) {
return 8;
}
}
else if (CsDiff >= 2) {
if (ssDCval < 117) {
if (TSC < 10) {
if (diffAFD < -2) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 2244) {
return 2;
}
else if (Rs >= 2244) {
return 8;
}
}
}
else if (diffAFD >= -2) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Cs < 1445) {
if (Rs < 1904) {
return 8;
}
else if (Rs >= 1904) {
return 2;
}
}
else if (Cs >= 1445) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (Rs < 2292) {
return 8;
}
else if (Rs >= 2292) {
return 1;
}
}
}
else if (ssDCval >= 117) {
if (mu_mv_mag_sq < 125) {
if (diffAFD < -1) {
if (Cs < 1443) {
return 8;
}
else if (Cs >= 1443) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
if (ssDCval < 123) {
if (Rs < 2410) {
return 1;
}
else if (Rs >= 2410) {
return 8;
}
}
else if (ssDCval >= 123) {
if (Rs < 2797) {
return 8;
}
else if (Rs >= 2797) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3097) {
if (refDCval < 77) {
if (mu_mv_mag_sq < 126) {
if (posBalance < 44) {
if (Rs < 3673) {
if (ssDCval < 63) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (Rs >= 3673) {
if (RsDiff < 3) {
if (posBalance < 33) {
if (negBalance < 21) {
return 4;
}
else if (negBalance >= 21) {
return 8;
}
}
else if (posBalance >= 33) {
if (Cs < 1558) {
if (Cs < 1324) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (Cs >= 1324) {
return 8;
}
}
else if (Cs >= 1558) {
return 4;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 4322) {
if (Cs < 1339) {
return 1;
}
else if (Cs >= 1339) {
return 4;
}
}
else if (Rs >= 4322) {
if (Rs < 4948) {
return 8;
}
else if (Rs >= 4948) {
return 4;
}
}
}
}
}
else if (posBalance >= 44) {
return 2;
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1578) {
if (RsDiff < 3) {
if (Rs < 3659) {
if (mu_mv_mag_sq < 127) {
if (Rs < 3619) {
return 8;
}
else if (Rs >= 3619) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
else if (Rs >= 3659) {
return 8;
}
}
else if (RsDiff >= 3) {
if (Rs < 3681) {
return 8;
}
else if (Rs >= 3681) {
return 4;
}
}
}
else if (Cs >= 1578) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 4;
}
}
}
}
else if (refDCval >= 77) {
if (ssDCval < 114) {
if (ssDCval < 84) {
if (Rs < 3953) {
return 2;
}
else if (Rs >= 3953) {
return 8;
}
}
else if (ssDCval >= 84) {
if (negBalance < 28) {
if (Rs < 4493) {
if (Cs < 1591) {
return 4;
}
else if (Cs >= 1591) {
if (mcTcor < 99) {
if (Cs < 1604) {
if (Cs < 1595) {
return 4;
}
else if (Cs >= 1595) {
return 8;
}
}
else if (Cs >= 1604) {
return 4;
}
}
else if (mcTcor >= 99) {
return 8;
}
}
}
else if (Rs >= 4493) {
return 2;
}
}
else if (negBalance >= 28) {
if (Cs < 1549) {
if (Rs < 3509) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 2;
}
}
else if (Rs >= 3509) {
return 4;
}
}
else if (Cs >= 1549) {
if (Rs < 4734) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 4;
}
}
else if (Rs >= 4734) {
return 2;
}
}
}
}
}
else if (ssDCval >= 114) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (Rs < 3363) {
return 2;
}
else if (Rs >= 3363) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1661) {
if (ssDCval < 98) {
if (SC < 4241) {
if (ssDCval < 67) {
if (TSC < 2) {
if (mu_mv_mag_sq < 118) {
if (Rs < 992) {
return 8;
}
else if (Rs >= 992) {
return 4;
}
}
else if (mu_mv_mag_sq >= 118) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
if (negBalance < 16) {
return 1;
}
else if (negBalance >= 16) {
if (Rs < 1073) {
return 1;
}
else if (Rs >= 1073) {
if (SC < 3947) {
return 8;
}
else if (SC >= 3947) {
return 1;
}
}
}
}
}
}
}
else if (TSC >= 2) {
if (refDCval < 63) {
if (Rs < 1654) {
if (negBalance < 26) {
if (Rs < 1048) {
return 8;
}
else if (Rs >= 1048) {
return 4;
}
}
else if (negBalance >= 26) {
return 8;
}
}
else if (Rs >= 1654) {
return 8;
}
}
else if (refDCval >= 63) {
if (ssDCval < 65) {
if (Rs < 1150) {
return 1;
}
else if (Rs >= 1150) {
return 8;
}
}
else if (ssDCval >= 65) {
if (SC < 4142) {
return 8;
}
else if (SC >= 4142) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (SC < 2800) {
if (Rs < 615) {
return 8;
}
else if (Rs >= 615) {
return 1;
}
}
else if (SC >= 2800) {
if (negBalance < 64) {
if (diffMVdiffVal < -6) {
return 4;
}
else if (diffMVdiffVal >= -6) {
if (Rs < 2478) {
return 8;
}
else if (Rs >= 2478) {
if (Cs < 1677) {
return 4;
}
else if (Cs >= 1677) {
return 8;
}
}
}
}
else if (negBalance >= 64) {
return 4;
}
}
}
}
else if (SC >= 4241) {
if (posBalance < 3) {
if (diffRsCsdiff < -5) {
return 8;
}
else if (diffRsCsdiff >= -5) {
if (refDCval < 65) {
if (ssDCval < 38) {
return 8;
}
else if (ssDCval >= 38) {
if (Rs < 6330) {
return 4;
}
else if (Rs >= 6330) {
return 2;
}
}
}
else if (refDCval >= 65) {
if (Cs < 2038) {
if (Rs < 2425) {
return 4;
}
else if (Rs >= 2425) {
return 8;
}
}
else if (Cs >= 2038) {
if (Rs < 4498) {
return 4;
}
else if (Rs >= 4498) {
return 8;
}
}
}
}
}
else if (posBalance >= 3) {
if (Rs < 6674) {
if (SC < 8417) {
if (ssDCval < 82) {
if (refDCval < 67) {
if (negBalance < 19) {
if (Rs < 3659) {
if (Cs < 3107) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 2980) {
if (Cs < 1901) {
return 1;
}
else if (Cs >= 1901) {
if (SC < 5019) {
if (Cs < 1978) {
return 4;
}
else if (Cs >= 1978) {
if (Cs < 2083) {
return 8;
}
else if (Cs >= 2083) {
if (Rs < 2755) {
return 8;
}
else if (Rs >= 2755) {
return 4;
}
}
}
}
else if (SC >= 5019) {
return 4;
}
}
}
else if (Rs >= 2980) {
return 1;
}
}
}
else if (Cs >= 3107) {
return 8;
}
}
else if (Rs >= 3659) {
if (ssDCval < 56) {
if (Cs < 2019) {
return 4;
}
else if (Cs >= 2019) {
if (diffRsCsdiff < -8) {
if (mu_mv_mag_sq < 114) {
return 8;
}
else if (mu_mv_mag_sq >= 114) {
if (Rs < 5082) {
if (posBalance < 41) {
if (diffRsCsdiff < -13) {
return 4;
}
else if (diffRsCsdiff >= -13) {
if (MVDiff < 2) {
if (Rs < 4987) {
return 8;
}
else if (Rs >= 4987) {
return 8;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (posBalance >= 41) {
return 8;
}
}
else if (Rs >= 5082) {
return 4;
}
}
}
else if (diffRsCsdiff >= -8) {
if (mu_mv_mag_sq < 117) {
if (tcor < 92) {
if (Rs < 4877) {
return 8;
}
else if (Rs >= 4877) {
return 8;
}
}
else if (tcor >= 92) {
if (MVDiff < 1) {
if (Cs < 2362) {
if (Rs < 4553) {
if (Rs < 4328) {
return 4;
}
else if (Rs >= 4328) {
return 4;
}
}
else if (Rs >= 4553) {
return 4;
}
}
else if (Cs >= 2362) {
return 4;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (Rs < 4148) {
return 8;
}
else if (Rs >= 4148) {
if (Rs < 5125) {
if (Cs < 2729) {
if (TSC < 3) {
if (posBalance < 27) {
return 4;
}
else if (posBalance >= 27) {
if (CsDiff < 5) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (Cs < 2124) {
return 4;
}
else if (Cs >= 2124) {
if (negBalance < 14) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (negBalance >= 14) {
if (mu_mv_mag_sq < 118) {
if (Rs < 4760) {
return 8;
}
else if (Rs >= 4760) {
return 8;
}
}
else if (mu_mv_mag_sq >= 118) {
return 4;
}
}
}
}
}
else if (CsDiff >= 5) {
return 4;
}
}
}
else if (TSC >= 3) {
return 4;
}
}
else if (Cs >= 2729) {
return 4;
}
}
else if (Rs >= 5125) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 56) {
return 4;
}
}
}
else if (negBalance >= 19) {
if (SC < 4932) {
if (Rs < 2778) {
if (RsDiff < 2) {
if (Cs < 1845) {
return 1;
}
else if (Cs >= 1845) {
return 8;
}
}
else if (RsDiff >= 2) {
if (Rs < 2134) {
return 4;
}
else if (Rs >= 2134) {
return 8;
}
}
}
else if (Rs >= 2778) {
if (SC < 4826) {
if (negBalance < 23) {
if (Rs < 2933) {
return 8;
}
else if (Rs >= 2933) {
return 8;
}
}
else if (negBalance >= 23) {
if (SC < 4787) {
if (SC < 4738) {
return 8;
}
else if (SC >= 4738) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 3005) {
return 4;
}
else if (Rs >= 3005) {
return 4;
}
}
}
}
else if (SC >= 4787) {
return 4;
}
}
}
else if (SC >= 4826) {
if (negBalance < 24) {
if (Rs < 3043) {
if (Rs < 2951) {
return 4;
}
else if (Rs >= 2951) {
if (gchDC < 1) {
if (Rs < 2976) {
return 4;
}
else if (Rs >= 2976) {
return 4;
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (Rs >= 3043) {
return 8;
}
}
else if (negBalance >= 24) {
return 4;
}
}
}
}
else if (SC >= 4932) {
if (Rs < 2017) {
return 4;
}
else if (Rs >= 2017) {
if (Cs < 2628) {
if (tcor < 99) {
if (Rs < 2944) {
return 8;
}
else if (Rs >= 2944) {
if (refDCval < 55) {
if (SC < 6170) {
return 8;
}
else if (SC >= 6170) {
if (posBalance < 33) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 4482) {
return 8;
}
else if (Rs >= 4482) {
if (Rs < 4808) {
return 4;
}
else if (Rs >= 4808) {
return 4;
}
}
}
}
else if (posBalance >= 33) {
if (Rs < 4475) {
return 4;
}
else if (Rs >= 4475) {
return 4;
}
}
}
}
else if (refDCval >= 55) {
if (SC < 4957) {
if (Rs < 3110) {
return 8;
}
else if (Rs >= 3110) {
return 8;
}
}
else if (SC >= 4957) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 8;
}
}
}
}
}
else if (tcor >= 99) {
if (ssDCval < 63) {
if (Rs < 2885) {
return 8;
}
else if (Rs >= 2885) {
return 4;
}
}
else if (ssDCval >= 63) {
return 8;
}
}
}
else if (Cs >= 2628) {
return 8;
}
}
}
}
}
else if (refDCval >= 67) {
if (Rs < 2184) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (Rs >= 2184) {
if (Cs < 1936) {
if (posBalance < 37) {
if (Rs < 4762) {
return 8;
}
else if (Rs >= 4762) {
return 4;
}
}
else if (posBalance >= 37) {
if (Rs < 3601) {
if (Rs < 2555) {
return 8;
}
else if (Rs >= 2555) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
else if (Rs >= 3601) {
if (Rs < 5429) {
return 2;
}
else if (Rs >= 5429) {
return 4;
}
}
}
}
else if (Cs >= 1936) {
if (mu_mv_mag_sq < 118) {
if (Cs < 3065) {
if (refDCval < 81) {
if (diffAFD < 2) {
if (gchDC < 2) {
if (Rs < 2212) {
if (Rs < 2201) {
return 8;
}
else if (Rs >= 2201) {
return 4;
}
}
else if (Rs >= 2212) {
if (Cs < 2029) {
if (Rs < 2745) {
return 8;
}
else if (Rs >= 2745) {
return 4;
}
}
else if (Cs >= 2029) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (refDCval >= 81) {
if (Cs < 2105) {
if (Rs < 3553) {
return 8;
}
else if (Rs >= 3553) {
return 2;
}
}
else if (Cs >= 2105) {
return 4;
}
}
}
else if (Cs >= 3065) {
return 8;
}
}
else if (mu_mv_mag_sq >= 118) {
if (diffTSC < -2) {
if (SC < 6006) {
return 8;
}
else if (SC >= 6006) {
if (SC < 7081) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
return 4;
}
}
else if (SC >= 7081) {
return 8;
}
}
}
else if (diffTSC >= -2) {
if (ssDCval < 81) {
if (diffRsCsdiff < 44) {
if (Cs < 1957) {
if (Rs < 3110) {
return 8;
}
else if (Rs >= 3110) {
return 4;
}
}
else if (Cs >= 1957) {
return 8;
}
}
else if (diffRsCsdiff >= 44) {
if (Rs < 3198) {
return 4;
}
else if (Rs >= 3198) {
return 8;
}
}
}
else if (ssDCval >= 81) {
if (Rs < 4408) {
return 8;
}
else if (Rs >= 4408) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 82) {
if (Cs < 2708) {
if (refDCval < 86) {
if (tcor < 98) {
if (Rs < 3741) {
if (posBalance < 58) {
if (Cs < 1685) {
if (Rs < 2751) {
return 8;
}
else if (Rs >= 2751) {
return 4;
}
}
else if (Cs >= 1685) {
return 8;
}
}
else if (posBalance >= 58) {
if (Rs < 2816) {
return 4;
}
else if (Rs >= 2816) {
return 8;
}
}
}
else if (Rs >= 3741) {
if (Cs < 2187) {
return 2;
}
else if (Cs >= 2187) {
return 8;
}
}
}
else if (tcor >= 98) {
if (Cs < 2470) {
if (MVDiff < 1) {
if (Rs < 2605) {
return 1;
}
else if (Rs >= 2605) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (diffRsCsdiff < -6) {
if (Rs < 3827) {
return 4;
}
else if (Rs >= 3827) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
if (mcTcor < 99) {
if (RsCsDiff < 11) {
return 8;
}
else if (RsCsDiff >= 11) {
return 4;
}
}
else if (mcTcor >= 99) {
return 4;
}
}
}
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (Cs >= 2470) {
return 8;
}
}
}
else if (refDCval >= 86) {
if (mu_mv_mag_sq < 127) {
if (tcor < 99) {
if (Rs < 3727) {
if (TSCindex < 1) {
if (Rs < 3044) {
return 2;
}
else if (Rs >= 3044) {
return 8;
}
}
else if (TSCindex >= 1) {
if (negBalance < 58) {
if (Cs < 2618) {
if (Cs < 1974) {
if (Cs < 1946) {
return 8;
}
else if (Cs >= 1946) {
return 2;
}
}
else if (Cs >= 1974) {
return 8;
}
}
else if (Cs >= 2618) {
if (Rs < 3117) {
return 8;
}
else if (Rs >= 3117) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
else if (negBalance >= 58) {
return 4;
}
}
}
else if (Rs >= 3727) {
if (SC < 6371) {
if (RsDiff < 6) {
if (diffRsCsdiff < -10) {
if (Rs < 4236) {
return 2;
}
else if (Rs >= 4236) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (diffRsCsdiff >= -10) {
return 8;
}
}
else if (RsDiff >= 6) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
return 2;
}
}
}
else if (SC >= 6371) {
if (Cs < 2267) {
if (SC < 6455) {
if (SC < 6394) {
return 2;
}
else if (SC >= 6394) {
return 8;
}
}
else if (SC >= 6455) {
return 2;
}
}
else if (Cs >= 2267) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 2;
}
}
}
}
}
else if (tcor >= 99) {
if (ssDCval < 94) {
if (SC < 4457) {
if (Rs < 2320) {
return 2;
}
else if (Rs >= 2320) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
if (ssDCval < 92) {
return 1;
}
else if (ssDCval >= 92) {
return 4;
}
}
}
}
else if (SC >= 4457) {
if (MVDiff < 2) {
if (Rs < 4726) {
if (posBalance < 13) {
if (Rs < 2401) {
return 2;
}
else if (Rs >= 2401) {
return 8;
}
}
else if (posBalance >= 13) {
return 2;
}
}
else if (Rs >= 4726) {
return 4;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (ssDCval >= 94) {
if (Rs < 2552) {
return 4;
}
else if (Rs >= 2552) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 4306) {
if (SC < 4463) {
if (SC < 4445) {
return 8;
}
else if (SC >= 4445) {
return 4;
}
}
else if (SC >= 4463) {
return 8;
}
}
else if (Rs >= 4306) {
if (Rs < 5072) {
return 1;
}
else if (Rs >= 5072) {
return 4;
}
}
}
}
}
else if (Cs >= 2708) {
if (ssDCval < 89) {
if (Cs < 3217) {
if (Rs < 3382) {
if (refDCval < 84) {
return 4;
}
else if (refDCval >= 84) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
}
else if (Rs >= 3382) {
return 4;
}
}
else if (Cs >= 3217) {
if (posBalance < 53) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Rs < 3227) {
return 8;
}
else if (Rs >= 3227) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Rs < 3912) {
if (Rs < 2932) {
return 8;
}
else if (Rs >= 2932) {
return 4;
}
}
else if (Rs >= 3912) {
return 8;
}
}
}
}
}
else if (ssDCval >= 89) {
if (mu_mv_mag_sq < 123) {
if (Cs < 3539) {
if (ssDCval < 96) {
return 4;
}
else if (ssDCval >= 96) {
if (Rs < 2972) {
return 8;
}
else if (Rs >= 2972) {
return 4;
}
}
}
else if (Cs >= 3539) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (SC < 7334) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (Rs < 3456) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (Rs >= 3456) {
if (posBalance < 52) {
if (Rs < 3623) {
return 4;
}
else if (Rs >= 3623) {
if (Rs < 3964) {
return 8;
}
else if (Rs >= 3964) {
return 4;
}
}
}
else if (posBalance >= 52) {
if (Cs < 3575) {
return 4;
}
else if (Cs >= 3575) {
return 8;
}
}
}
}
}
else if (SC >= 7334) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 8417) {
if (negBalance < 31) {
if (Rs < 6002) {
if (Rs < 4456) {
if (posBalance < 50) {
if (Rs < 4338) {
return 8;
}
else if (Rs >= 4338) {
return 1;
}
}
else if (posBalance >= 50) {
return 8;
}
}
else if (Rs >= 4456) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (Rs >= 6002) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (negBalance >= 31) {
if (Cs < 3733) {
if (Rs < 5273) {
return 4;
}
else if (Rs >= 5273) {
return 8;
}
}
else if (Cs >= 3733) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 5982) {
return 8;
}
else if (Rs >= 5982) {
return 2;
}
}
}
}
}
}
else if (Rs >= 6674) {
if (Rs < 9152) {
return 4;
}
else if (Rs >= 9152) {
return 2;
}
}
}
}
}
else if (ssDCval >= 98) {
if (ssDCval < 129) {
if (Rs < 1908) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (tcor < 97) {
if (Rs < 1372) {
return 4;
}
else if (Rs >= 1372) {
if (Cs < 2241) {
return 8;
}
else if (Cs >= 2241) {
return 4;
}
}
}
else if (tcor >= 97) {
if (posBalance < 54) {
if (Cs < 2071) {
return 8;
}
else if (Cs >= 2071) {
return 2;
}
}
else if (posBalance >= 54) {
if (mu_mv_mag_sq < 124) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
if (Rs < 1600) {
return 8;
}
else if (Rs >= 1600) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 2;
}
}
}
}
}
else if (Rs >= 1908) {
if (refDCval < 105) {
if (Cs < 1995) {
if (Rs < 3598) {
return 8;
}
else if (Rs >= 3598) {
if (AFD < 5) {
if (Cs < 1926) {
return 4;
}
else if (Cs >= 1926) {
return 2;
}
}
else if (AFD >= 5) {
if (Rs < 4328) {
return 2;
}
else if (Rs >= 4328) {
return 1;
}
}
}
}
else if (Cs >= 1995) {
if (TSC < 9) {
if (ssDCval < 104) {
if (Rs < 2065) {
return 4;
}
else if (Rs >= 2065) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (AFD < 5) {
if (Rs < 3003) {
return 4;
}
else if (Rs >= 3003) {
return 8;
}
}
else if (AFD >= 5) {
if (ssDCval < 103) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (ssDCval >= 103) {
return 1;
}
}
}
}
}
else if (ssDCval >= 104) {
if (Rs < 4974) {
return 8;
}
else if (Rs >= 4974) {
if (Cs < 2906) {
return 2;
}
else if (Cs >= 2906) {
return 8;
}
}
}
}
else if (TSC >= 9) {
if (mcTcor < 93) {
if (Cs < 2895) {
if (Rs < 4947) {
return 1;
}
else if (Rs >= 4947) {
return 2;
}
}
else if (Cs >= 2895) {
return 8;
}
}
else if (mcTcor >= 93) {
if (Cs < 3168) {
if (Cs < 2595) {
return 8;
}
else if (Cs >= 2595) {
return 4;
}
}
else if (Cs >= 3168) {
return 8;
}
}
}
}
}
else if (refDCval >= 105) {
if (MVDiff < 1) {
if (diffAFD < -7) {
if (Rs < 2648) {
return 8;
}
else if (Rs >= 2648) {
return 1;
}
}
else if (diffAFD >= -7) {
if (diffMVdiffVal < -1) {
if (posBalance < 12) {
if (AFD < 1) {
if (Cs < 1840) {
return 1;
}
else if (Cs >= 1840) {
return 8;
}
}
else if (AFD >= 1) {
return 2;
}
}
else if (posBalance >= 12) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 3040) {
return 8;
}
else if (Rs >= 3040) {
return 2;
}
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 62) {
if (diffTSC < -2) {
if (posBalance < 50) {
if (Rs < 6397) {
if (Cs < 2058) {
return 4;
}
else if (Cs >= 2058) {
return 8;
}
}
else if (Rs >= 6397) {
return 4;
}
}
else if (posBalance >= 50) {
if (diffRsCsdiff < -13) {
return 8;
}
else if (diffRsCsdiff >= -13) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (diffTSC >= -2) {
if (diffRsCsdiff < 19) {
if (Cs < 1679) {
if (Rs < 4148) {
return 8;
}
else if (Rs >= 4148) {
return 2;
}
}
else if (Cs >= 1679) {
return 8;
}
}
else if (diffRsCsdiff >= 19) {
if (RsCsDiff < 91) {
if (ssDCval < 108) {
if (Rs < 2816) {
return 8;
}
else if (Rs >= 2816) {
return 4;
}
}
else if (ssDCval >= 108) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (RsCsDiff < 79) {
return 8;
}
else if (RsCsDiff >= 79) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 91) {
return 8;
}
}
}
}
else if (posBalance >= 62) {
if (Cs < 1912) {
if (Rs < 2462) {
return 8;
}
else if (Rs >= 2462) {
return 1;
}
}
else if (Cs >= 1912) {
return 8;
}
}
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 117) {
if (AFD < 12) {
if (refDCval < 107) {
if (diffRsCsdiff < -106) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (Rs < 3627) {
return 4;
}
else if (Rs >= 3627) {
return 2;
}
}
}
else if (diffRsCsdiff >= -106) {
return 8;
}
}
else if (refDCval >= 107) {
return 8;
}
}
else if (AFD >= 12) {
if (Cs < 3381) {
return 2;
}
else if (Cs >= 3381) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (Cs < 2839) {
if (posBalance < 59) {
if (refDCval < 124) {
return 8;
}
else if (refDCval >= 124) {
if (AFD < 7) {
if (diffAFD < 2) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 2758) {
return 8;
}
else if (Rs >= 2758) {
return 1;
}
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (posBalance >= 59) {
return 1;
}
}
else if (Cs >= 2839) {
if (CsDiff < 1) {
if (Rs < 2769) {
if (posBalance < 16) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (posBalance >= 16) {
return 4;
}
}
else if (Rs >= 2769) {
if (Rs < 3440) {
return 2;
}
else if (Rs >= 3440) {
return 1;
}
}
}
else if (CsDiff >= 1) {
if (diffRsCsdiff < -28) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (diffRsCsdiff >= -28) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 129) {
if (mcTcor < 96) {
if (negBalance < 42) {
if (ssDCval < 200) {
if (posBalance < 66) {
if (diffAFD < 6) {
if (ssDCval < 130) {
if (Cs < 3423) {
return 8;
}
else if (Cs >= 3423) {
return 1;
}
}
else if (ssDCval >= 130) {
return 8;
}
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (posBalance >= 66) {
if (Rs < 1689) {
return 8;
}
else if (Rs >= 1689) {
return 1;
}
}
}
else if (ssDCval >= 200) {
return 4;
}
}
else if (negBalance >= 42) {
if (ssDCval < 143) {
if (Rs < 2101) {
return 8;
}
else if (Rs >= 2101) {
return 1;
}
}
else if (ssDCval >= 143) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 130) {
if (negBalance < 34) {
if (Rs < 1787) {
return 4;
}
else if (Rs >= 1787) {
return 1;
}
}
else if (negBalance >= 34) {
return 8;
}
}
else if (refDCval >= 130) {
if (ssDCval < 144) {
if (Rs < 3820) {
if (refDCval < 131) {
if (posBalance < 60) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (posBalance >= 60) {
return 4;
}
}
else if (refDCval >= 131) {
if (SC < 4979) {
return 8;
}
else if (SC >= 4979) {
if (SC < 4983) {
return 1;
}
else if (SC >= 4983) {
return 8;
}
}
}
}
else if (Rs >= 3820) {
if (Cs < 3137) {
if (Cs < 2786) {
if (Rs < 4405) {
return 8;
}
else if (Rs >= 4405) {
return 2;
}
}
else if (Cs >= 2786) {
return 1;
}
}
else if (Cs >= 3137) {
if (Rs < 3894) {
return 1;
}
else if (Rs >= 3894) {
return 8;
}
}
}
}
else if (ssDCval >= 144) {
if (ssDCval < 201) {
return 8;
}
else if (ssDCval >= 201) {
if (Rs < 2489) {
return 8;
}
else if (Rs >= 2489) {
return 4;
}
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect4(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1313) {
if (Rs < 2339) {
if (Cs < 534) {
if (negBalance < 14) {
if (MVDiff < 7) {
if (SC < 793) {
if (ssDCval < 67) {
if (Cs < 11) {
if (refDCval < 47) {
if (refDCval < 22) {
if (ssDCval < 9) {
return 4;
}
else if (ssDCval >= 9) {
return 8;
}
}
else if (refDCval >= 22) {
if (Rs < 58) {
return 4;
}
else if (Rs >= 58) {
return 8;
}
}
}
else if (refDCval >= 47) {
return 2;
}
}
else if (Cs >= 11) {
if (Cs < 264) {
if (AFD < 1) {
if (SC < 375) {
if (negBalance < 8) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 2;
}
}
else if (negBalance >= 8) {
if (SC < 67) {
if (Cs < 41) {
if (Cs < 13) {
return 8;
}
else if (Cs >= 13) {
return 4;
}
}
else if (Cs >= 41) {
return 8;
}
}
else if (SC >= 67) {
if (posBalance < 43) {
if (ssDCval < 42) {
if (ssDCval < 29) {
return 8;
}
else if (ssDCval >= 29) {
if (ssDCval < 38) {
if (mu_mv_mag_sq < 106) {
return 8;
}
else if (mu_mv_mag_sq >= 106) {
return 4;
}
}
else if (ssDCval >= 38) {
if (Rs < 93) {
return 8;
}
else if (Rs >= 93) {
return 1;
}
}
}
}
else if (ssDCval >= 42) {
return 8;
}
}
else if (posBalance >= 43) {
return 4;
}
}
}
}
else if (SC >= 375) {
if (Cs < 152) {
if (tcor < 104) {
if (Rs < 432) {
return 4;
}
else if (Rs >= 432) {
return 8;
}
}
else if (tcor >= 104) {
return 8;
}
}
else if (Cs >= 152) {
if (SC < 377) {
return 4;
}
else if (SC >= 377) {
return 8;
}
}
}
}
else if (AFD >= 1) {
if (posBalance < 49) {
if (ssDCval < 28) {
if (posBalance < 45) {
if (Rs < 42) {
return 4;
}
else if (Rs >= 42) {
return 1;
}
}
else if (posBalance >= 45) {
return 8;
}
}
else if (ssDCval >= 28) {
if (ssDCval < 39) {
if (Rs < 90) {
return 4;
}
else if (Rs >= 90) {
if (mcTcor < 100) {
if (mu_mv_mag_sq < 115) {
if (Cs < 116) {
if (ssDCval < 31) {
return 4;
}
else if (ssDCval >= 31) {
return 8;
}
}
else if (Cs >= 116) {
if (SC < 238) {
return 1;
}
else if (SC >= 238) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
else if (mcTcor >= 100) {
if (Rs < 96) {
return 8;
}
else if (Rs >= 96) {
return 1;
}
}
}
}
else if (ssDCval >= 39) {
if (posBalance < 40) {
if (Rs < 86) {
return 8;
}
else if (Rs >= 86) {
if (Rs < 207) {
return 1;
}
else if (Rs >= 207) {
if (Cs < 221) {
if (Rs < 467) {
return 8;
}
else if (Rs >= 467) {
return 4;
}
}
else if (Cs >= 221) {
return 1;
}
}
}
}
else if (posBalance >= 40) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 119) {
return 8;
}
else if (Rs >= 119) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 49) {
if (tcor < 98) {
if (AFD < 6) {
if (negBalance < 1) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 2;
}
}
else if (negBalance >= 1) {
return 8;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (tcor >= 98) {
if (Cs < 113) {
return 8;
}
else if (Cs >= 113) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
}
}
}
else if (Cs >= 264) {
if (Rs < 203) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
return 1;
}
}
else if (Rs >= 203) {
if (SC < 679) {
if (TSCindex < 2) {
if (negBalance < 3) {
if (mu_mv_mag_sq < 128) {
return 4;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 324) {
return 8;
}
else if (Rs >= 324) {
if (Rs < 343) {
return 4;
}
else if (Rs >= 343) {
return 8;
}
}
}
}
else if (negBalance >= 3) {
return 4;
}
}
else if (TSCindex >= 2) {
if (ssDCval < 40) {
if (Rs < 206) {
return 2;
}
else if (Rs >= 206) {
return 4;
}
}
else if (ssDCval >= 40) {
return 1;
}
}
}
else if (SC >= 679) {
if (mu_mv_mag_sq < 126) {
if (Rs < 448) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
if (SC < 769) {
return 1;
}
else if (SC >= 769) {
if (Rs < 333) {
return 1;
}
else if (Rs >= 333) {
return 4;
}
}
}
}
else if (Rs >= 448) {
return 8;
}
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (posBalance < 28) {
if (Rs < 179) {
return 1;
}
else if (Rs >= 179) {
if (ssDCval < 129) {
if (Rs < 304) {
return 8;
}
else if (Rs >= 304) {
return 2;
}
}
else if (ssDCval >= 129) {
if (Rs < 463) {
return 4;
}
else if (Rs >= 463) {
return 1;
}
}
}
}
else if (posBalance >= 28) {
if (ssDCval < 94) {
if (Cs < 119) {
if (Rs < 145) {
return 4;
}
else if (Rs >= 145) {
return 8;
}
}
else if (Cs >= 119) {
if (Cs < 241) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
if (ssDCval < 88) {
if (ssDCval < 84) {
if (Rs < 385) {
return 8;
}
else if (Rs >= 385) {
return 2;
}
}
else if (ssDCval >= 84) {
return 1;
}
}
else if (ssDCval >= 88) {
return 2;
}
}
}
else if (Cs >= 241) {
if (Rs < 370) {
if (Cs < 245) {
return 1;
}
else if (Cs >= 245) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
return 4;
}
}
}
else if (Rs >= 370) {
return 8;
}
}
}
}
else if (ssDCval >= 94) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 389) {
return 8;
}
else if (Rs >= 389) {
return 1;
}
}
}
}
}
}
else if (SC >= 793) {
if (mu_mv_mag_sq < 128) {
if (Cs < 268) {
if (Rs < 625) {
if (Rs < 554) {
return 8;
}
else if (Rs >= 554) {
return 4;
}
}
else if (Rs >= 625) {
if (TSCindex < 2) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (mu_mv_mag_sq < 127) {
if (Cs < 208) {
return 8;
}
else if (Cs >= 208) {
if (Rs < 862) {
return 8;
}
else if (Rs >= 862) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 2;
}
}
}
else if (TSCindex >= 2) {
if (SC < 1075) {
return 8;
}
else if (SC >= 1075) {
if (Rs < 1276) {
return 4;
}
else if (Rs >= 1276) {
return 1;
}
}
}
}
}
else if (Cs >= 268) {
if (SC < 945) {
if (ssDCval < 109) {
if (Rs < 396) {
return 1;
}
else if (Rs >= 396) {
if (Rs < 478) {
return 8;
}
else if (Rs >= 478) {
if (Cs < 373) {
return 8;
}
else if (Cs >= 373) {
if (Rs < 530) {
if (Rs < 482) {
return 1;
}
else if (Rs >= 482) {
return 8;
}
}
else if (Rs >= 530) {
return 1;
}
}
}
}
}
else if (ssDCval >= 109) {
if (posBalance < 13) {
return 8;
}
else if (posBalance >= 13) {
if (SC < 937) {
return 1;
}
else if (SC >= 937) {
return 8;
}
}
}
}
else if (SC >= 945) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Rs < 896) {
if (Cs < 456) {
return 8;
}
else if (Cs >= 456) {
if (AFD < 2) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (Rs >= 896) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
if (Rs < 1166) {
return 1;
}
else if (Rs >= 1166) {
return 8;
}
}
}
}
else if (ssDCval >= 60) {
if (Cs < 444) {
return 1;
}
else if (Cs >= 444) {
return 2;
}
}
}
else if (ssDCval >= 67) {
if (Rs < 491) {
return 1;
}
else if (Rs >= 491) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
if (mu_mv_mag_sq < 95) {
if (Rs < 1970) {
return 1;
}
else if (Rs >= 1970) {
return 2;
}
}
else if (mu_mv_mag_sq >= 95) {
if (Rs < 634) {
return 4;
}
else if (Rs >= 634) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (refDCval < 60) {
return 8;
}
else if (refDCval >= 60) {
if (ssDCval < 87) {
if (Cs < 355) {
if (Cs < 258) {
return 2;
}
else if (Cs >= 258) {
return 1;
}
}
else if (Cs >= 355) {
if (ssDCval < 72) {
if (Rs < 1186) {
return 8;
}
else if (Rs >= 1186) {
return 2;
}
}
else if (ssDCval >= 72) {
return 4;
}
}
}
else if (ssDCval >= 87) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (SC < 1092) {
if (Rs < 575) {
return 8;
}
else if (Rs >= 575) {
return 4;
}
}
else if (SC >= 1092) {
return 8;
}
}
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (tcor < 94) {
if (negBalance < 5) {
if (MVDiff < 220) {
if (diffRsCsdiff < 2) {
if (ssDCval < 15) {
return 4;
}
else if (ssDCval >= 15) {
if (mu_mv_mag_sq < 47) {
if (Rs < 58) {
if (mcTcor < 68) {
if (Rs < 10) {
return 8;
}
else if (Rs >= 10) {
if (Rs < 25) {
return 2;
}
else if (Rs >= 25) {
return 1;
}
}
}
else if (mcTcor >= 68) {
if (MVDiff < 110) {
return 2;
}
else if (MVDiff >= 110) {
return 8;
}
}
}
else if (Rs >= 58) {
if (gchDC < 6) {
if (posBalance < 93) {
if (Rs < 205) {
return 8;
}
else if (Rs >= 205) {
return 2;
}
}
else if (posBalance >= 93) {
return 4;
}
}
else if (gchDC >= 6) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (mu_mv_mag_sq < 105) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
if (Cs < 87) {
if (Rs < 106) {
return 2;
}
else if (Rs >= 106) {
return 4;
}
}
else if (Cs >= 87) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (mcTcor < 75) {
return 8;
}
else if (mcTcor >= 75) {
if (Rs < 356) {
return 4;
}
else if (Rs >= 356) {
return 8;
}
}
}
}
}
}
else if (diffRsCsdiff >= 2) {
if (posBalance < 113) {
return 8;
}
else if (posBalance >= 113) {
return 4;
}
}
}
else if (MVDiff >= 220) {
if (AFD < 8) {
if (MVDiff < 673) {
if (diffMVdiffVal < 38) {
return 8;
}
else if (diffMVdiffVal >= 38) {
if (Rs < 50) {
return 2;
}
else if (Rs >= 50) {
return 4;
}
}
}
else if (MVDiff >= 673) {
if (ssDCval < 11) {
return 1;
}
else if (ssDCval >= 11) {
return 4;
}
}
}
else if (AFD >= 8) {
if (Rs < 877) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 4;
}
}
else if (Rs >= 877) {
return 2;
}
}
}
}
else if (negBalance >= 5) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (diffMVdiffVal < -9) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
if (posBalance < 111) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
return 2;
}
}
else if (posBalance >= 111) {
if (Rs < 1186) {
return 4;
}
else if (Rs >= 1186) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -9) {
if (mu_mv_mag_sq < 38) {
if (mcTcor < 96) {
if (AFD < 11) {
if (TSC < 2) {
if (posBalance < 76) {
if (ssDCval < 25) {
return 1;
}
else if (ssDCval >= 25) {
return 2;
}
}
else if (posBalance >= 76) {
return 1;
}
}
else if (TSC >= 2) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (diffMVdiffVal < 27) {
return 2;
}
else if (diffMVdiffVal >= 27) {
if (Rs < 57) {
return 8;
}
else if (Rs >= 57) {
return 4;
}
}
}
}
}
else if (AFD >= 11) {
if (Cs < 421) {
return 1;
}
else if (Cs >= 421) {
return 2;
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (mu_mv_mag_sq >= 38) {
if (refDCval < 44) {
if (posBalance < 66) {
if (Cs < 217) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
else if (Cs >= 217) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
if (MVDiff < 31) {
return 1;
}
else if (MVDiff >= 31) {
return 4;
}
}
}
}
else if (posBalance >= 66) {
if (mcTcor < 88) {
if (Rs < 351) {
return 8;
}
else if (Rs >= 351) {
return 1;
}
}
else if (mcTcor >= 88) {
return 1;
}
}
}
else if (refDCval >= 44) {
if (MVDiff < 37) {
if (posBalance < 108) {
if (ssDCval < 49) {
if (Rs < 247) {
return 8;
}
else if (Rs >= 247) {
return 1;
}
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (posBalance >= 108) {
return 4;
}
}
else if (MVDiff >= 37) {
if (Rs < 226) {
return 1;
}
else if (Rs >= 226) {
return 2;
}
}
}
}
}
}
}
}
else if (tcor >= 94) {
if (diffAFD < -1) {
if (refDCval < 125) {
if (mu_mv_mag_sq < 113) {
if (negBalance < 4) {
if (mcTcor < 96) {
return 4;
}
else if (mcTcor >= 96) {
if (Rs < 19) {
return 2;
}
else if (Rs >= 19) {
return 8;
}
}
}
else if (negBalance >= 4) {
if (Cs < 114) {
if (diffAFD < -3) {
if (Rs < 31) {
return 2;
}
else if (Rs >= 31) {
return 8;
}
}
else if (diffAFD >= -3) {
if (Rs < 182) {
return 1;
}
else if (Rs >= 182) {
return 8;
}
}
}
else if (Cs >= 114) {
if (AFD < 2) {
if (Rs < 494) {
return 4;
}
else if (Rs >= 494) {
return 8;
}
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (MVDiff < 1076) {
if (Rs < 685) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (posBalance < 32) {
if (Rs < 348) {
return 4;
}
else if (Rs >= 348) {
if (Rs < 467) {
return 8;
}
else if (Rs >= 467) {
return 1;
}
}
}
else if (posBalance >= 32) {
return 8;
}
}
}
else if (Rs >= 685) {
if (Rs < 769) {
return 4;
}
else if (Rs >= 769) {
return 8;
}
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
}
else if (refDCval >= 125) {
if (ssDCval < 130) {
if (Rs < 1109) {
if (Rs < 327) {
return 8;
}
else if (Rs >= 327) {
if (diffAFD < -4) {
if (MVDiff < 91) {
return 2;
}
else if (MVDiff >= 91) {
return 4;
}
}
else if (diffAFD >= -4) {
return 4;
}
}
}
else if (Rs >= 1109) {
if (Cs < 530) {
return 1;
}
else if (Cs >= 530) {
return 2;
}
}
}
else if (ssDCval >= 130) {
if (MVDiff < 45) {
if (Rs < 880) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (Rs >= 880) {
return 1;
}
}
else if (MVDiff >= 45) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
}
else if (diffAFD >= -1) {
if (ssDCval < 22) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
return 1;
}
}
else if (ssDCval >= 22) {
if (SC < 877) {
if (posBalance < 63) {
if (mu_mv_mag_sq < 64) {
return 4;
}
else if (mu_mv_mag_sq >= 64) {
if (ssDCval < 133) {
if (Cs < 176) {
if (negBalance < 9) {
if (ssDCval < 27) {
return 4;
}
else if (ssDCval >= 27) {
if (TSCindex < 1) {
return 8;
}
else if (TSCindex >= 1) {
return 4;
}
}
}
else if (negBalance >= 9) {
if (ssDCval < 56) {
if (ssDCval < 37) {
if (Rs < 8) {
return 4;
}
else if (Rs >= 8) {
return 1;
}
}
else if (ssDCval >= 37) {
return 8;
}
}
else if (ssDCval >= 56) {
if (posBalance < 33) {
return 1;
}
else if (posBalance >= 33) {
return 4;
}
}
}
}
else if (Cs >= 176) {
if (mu_mv_mag_sq < 115) {
if (diffMVdiffVal < 3) {
if (Cs < 214) {
if (Cs < 211) {
return 4;
}
else if (Cs >= 211) {
return 1;
}
}
else if (Cs >= 214) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
if (mu_mv_mag_sq < 98) {
if (Cs < 426) {
if (posBalance < 30) {
if (Rs < 241) {
return 1;
}
else if (Rs >= 241) {
return 4;
}
}
else if (posBalance >= 30) {
return 4;
}
}
else if (Cs >= 426) {
return 1;
}
}
else if (mu_mv_mag_sq >= 98) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
}
else if (ssDCval >= 133) {
if (ssDCval < 161) {
return 1;
}
else if (ssDCval >= 161) {
return 4;
}
}
}
}
else if (posBalance >= 63) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (tcor < 101) {
if (SC < 287) {
if (Rs < 26) {
return 8;
}
else if (Rs >= 26) {
return 4;
}
}
else if (SC >= 287) {
return 8;
}
}
else if (tcor >= 101) {
if (Rs < 8) {
return 2;
}
else if (Rs >= 8) {
return 1;
}
}
}
}
}
else if (SC >= 877) {
if (ssDCval < 42) {
return 4;
}
else if (ssDCval >= 42) {
if (Cs < 412) {
if (Cs < 48) {
return 8;
}
else if (Cs >= 48) {
if (Rs < 874) {
if (Rs < 550) {
return 1;
}
else if (Rs >= 550) {
return 8;
}
}
else if (Rs >= 874) {
return 1;
}
}
}
else if (Cs >= 412) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (tcor < 63) {
if (negBalance < 18) {
if (posBalance < 36) {
return 1;
}
else if (posBalance >= 36) {
return 8;
}
}
else if (negBalance >= 18) {
if (negBalance < 45) {
if (CsDiff < 2) {
if (posBalance < 65) {
if (AFD < 3) {
if (MVDiff < 68) {
return 1;
}
else if (MVDiff >= 68) {
return 2;
}
}
else if (AFD >= 3) {
if (ssDCval < 39) {
if (diffTSC < -1) {
if (tcor < 60) {
return 1;
}
else if (tcor >= 60) {
if (Rs < 348) {
return 4;
}
else if (Rs >= 348) {
return 1;
}
}
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (ssDCval >= 39) {
return 4;
}
}
}
else if (posBalance >= 65) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (Rs < 445) {
if (RsCsDiff < 17) {
if (Cs < 287) {
if (Rs < 391) {
return 8;
}
else if (Rs >= 391) {
return 1;
}
}
else if (Cs >= 287) {
return 2;
}
}
else if (RsCsDiff >= 17) {
if (Cs < 493) {
return 1;
}
else if (Cs >= 493) {
return 8;
}
}
}
else if (Rs >= 445) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (Cs < 452) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (Cs >= 452) {
return 2;
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 357) {
if (MVDiff < 284) {
return 8;
}
else if (MVDiff >= 284) {
return 4;
}
}
else if (Rs >= 357) {
if (Cs < 221) {
return 1;
}
else if (Cs >= 221) {
return 8;
}
}
}
}
}
else if (tcor >= 63) {
if (posBalance < 21) {
if (mu_mv_mag_sq < 111) {
if (AFD < 2) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
if (ssDCval < 32) {
return 4;
}
else if (ssDCval >= 32) {
if (Rs < 93) {
return 8;
}
else if (Rs >= 93) {
return 4;
}
}
}
}
else if (AFD >= 2) {
if (Rs < 28) {
if (negBalance < 60) {
return 1;
}
else if (negBalance >= 60) {
if (Rs < 1) {
if (mu_mv_mag_sq < 34) {
if (MVDiff < 450) {
return 8;
}
else if (MVDiff >= 450) {
return 4;
}
}
else if (mu_mv_mag_sq >= 34) {
return 4;
}
}
else if (Rs >= 1) {
return 8;
}
}
}
else if (Rs >= 28) {
if (ssDCval < 18) {
return 4;
}
else if (ssDCval >= 18) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
return 8;
}
}
else if (posBalance >= 21) {
if (posBalance < 46) {
if (refDCval < 30) {
if (ssDCval < 23) {
if (Cs < 313) {
if (MVDiff < 53) {
return 1;
}
else if (MVDiff >= 53) {
return 8;
}
}
else if (Cs >= 313) {
return 4;
}
}
else if (ssDCval >= 23) {
if (mu_mv_mag_sq < 43) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
if (diffMVdiffVal < 23) {
if (Rs < 85) {
return 1;
}
else if (Rs >= 85) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (MVDiff < 29) {
return 1;
}
else if (MVDiff >= 29) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
}
else if (diffMVdiffVal >= 23) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (MVDiff < 27) {
return 1;
}
else if (MVDiff >= 27) {
if (Cs < 242) {
return 1;
}
else if (Cs >= 242) {
return 4;
}
}
}
}
}
else if (refDCval >= 30) {
if (posBalance < 37) {
if (tcor < 95) {
if (Cs < 243) {
if (MVDiff < 17) {
if (mcTcor < 93) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (mcTcor >= 93) {
if (negBalance < 22) {
if (MVDiff < 6) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 8;
}
}
else if (MVDiff >= 6) {
if (Rs < 208) {
return 8;
}
else if (Rs >= 208) {
return 2;
}
}
}
else if (negBalance >= 22) {
if (AFD < 3) {
if (Rs < 206) {
return 1;
}
else if (Rs >= 206) {
return 8;
}
}
else if (AFD >= 3) {
if (Rs < 93) {
return 1;
}
else if (Rs >= 93) {
return 4;
}
}
}
}
}
else if (MVDiff >= 17) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
return 4;
}
}
}
}
else if (Cs >= 243) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (tcor >= 95) {
if (SC < 227) {
if (Cs < 83) {
if (Rs < 4) {
return 4;
}
else if (Rs >= 4) {
return 1;
}
}
else if (Cs >= 83) {
if (AFD < 3) {
if (Rs < 117) {
return 8;
}
else if (Rs >= 117) {
return 1;
}
}
else if (AFD >= 3) {
if (Rs < 84) {
return 2;
}
else if (Rs >= 84) {
return 1;
}
}
}
}
else if (SC >= 227) {
if (Rs < 371) {
if (Rs < 111) {
return 4;
}
else if (Rs >= 111) {
return 1;
}
}
else if (Rs >= 371) {
if (Rs < 787) {
return 4;
}
else if (Rs >= 787) {
return 1;
}
}
}
}
}
else if (posBalance >= 37) {
if (TSCindex < 2) {
if (tcor < 94) {
if (Rs < 108) {
return 8;
}
else if (Rs >= 108) {
return 4;
}
}
else if (tcor >= 94) {
return 1;
}
}
else if (TSCindex >= 2) {
if (mcTcor < 91) {
if (AFD < 4) {
if (Rs < 469) {
return 8;
}
else if (Rs >= 469) {
return 1;
}
}
else if (AFD >= 4) {
return 1;
}
}
else if (mcTcor >= 91) {
if (gchDC < 2) {
if (MVDiff < 2) {
if (Rs < 519) {
return 8;
}
else if (Rs >= 519) {
return 1;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (gchDC >= 2) {
if (Rs < 868) {
if (Rs < 645) {
return 1;
}
else if (Rs >= 645) {
return 4;
}
}
else if (Rs >= 868) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 46) {
if (posBalance < 67) {
if (Rs < 389) {
if (mcTcor < 86) {
if (Cs < 133) {
return 1;
}
else if (Cs >= 133) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
}
else if (mcTcor >= 86) {
if (negBalance < 20) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (negBalance >= 20) {
if (mcTcor < 91) {
if (ssDCval < 33) {
if (MVDiff < 44) {
return 1;
}
else if (MVDiff >= 44) {
if (Rs < 79) {
return 1;
}
else if (Rs >= 79) {
return 2;
}
}
}
else if (ssDCval >= 33) {
if (Rs < 171) {
return 1;
}
else if (Rs >= 171) {
return 2;
}
}
}
else if (mcTcor >= 91) {
return 1;
}
}
}
}
else if (Rs >= 389) {
return 1;
}
}
else if (posBalance >= 67) {
if (mu_mv_mag_sq < 11) {
return 2;
}
else if (mu_mv_mag_sq >= 11) {
if (Rs < 137) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Rs >= 137) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 304) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 49) {
if (mcTcor < 90) {
if (MVDiff < 42) {
if (Rs < 91) {
return 1;
}
else if (Rs >= 91) {
if (tcor < 67) {
if (mcTcor < 84) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
return 8;
}
}
else if (mcTcor >= 84) {
if (MVDiff < 29) {
return 4;
}
else if (MVDiff >= 29) {
if (Rs < 173) {
return 8;
}
else if (Rs >= 173) {
return 1;
}
}
}
}
else if (tcor >= 67) {
if (mcTcor < 84) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (mcTcor >= 84) {
if (Rs < 223) {
if (Rs < 169) {
if (AFD < 6) {
if (Cs < 162) {
return 8;
}
else if (Cs >= 162) {
return 2;
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (Rs >= 169) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (Rs >= 223) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 42) {
if (posBalance < 56) {
if (negBalance < 60) {
if (mu_mv_mag_sq < 6) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 6) {
if (RsCsDiff < 1) {
if (Rs < 47) {
return 1;
}
else if (Rs >= 47) {
if (Rs < 63) {
return 4;
}
else if (Rs >= 63) {
return 2;
}
}
}
else if (RsCsDiff >= 1) {
return 1;
}
}
}
else if (negBalance >= 60) {
if (AFD < 14) {
if (MVDiff < 151) {
return 8;
}
else if (MVDiff >= 151) {
if (Cs < 40) {
return 8;
}
else if (Cs >= 40) {
return 2;
}
}
}
else if (AFD >= 14) {
if (MVDiff < 152) {
if (Rs < 104) {
return 8;
}
else if (Rs >= 104) {
return 4;
}
}
else if (MVDiff >= 152) {
if (Rs < 141) {
return 1;
}
else if (Rs >= 141) {
if (Rs < 183) {
return 8;
}
else if (Rs >= 183) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 56) {
if (mu_mv_mag_sq < 15) {
if (posBalance < 66) {
return 4;
}
else if (posBalance >= 66) {
return 2;
}
}
else if (mu_mv_mag_sq >= 15) {
if (TSC < 7) {
if (diffMVdiffVal < 13) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (diffMVdiffVal >= 13) {
return 8;
}
}
else if (TSC >= 7) {
return 1;
}
}
}
}
}
else if (mcTcor >= 90) {
if (posBalance < 8) {
if (MVDiff < 75) {
return 8;
}
else if (MVDiff >= 75) {
if (refDCval < 42) {
if (Cs < 2) {
return 8;
}
else if (Cs >= 2) {
return 4;
}
}
else if (refDCval >= 42) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (posBalance >= 8) {
if (mcTcor < 98) {
if (Rs < 272) {
if (mu_mv_mag_sq < 3) {
if (Rs < 126) {
if (Rs < 8) {
return 8;
}
else if (Rs >= 8) {
return 1;
}
}
else if (Rs >= 126) {
if (refDCval < 43) {
if (ssDCval < 37) {
if (Cs < 202) {
return 2;
}
else if (Cs >= 202) {
return 8;
}
}
else if (ssDCval >= 37) {
if (Cs < 361) {
return 2;
}
else if (Cs >= 361) {
return 1;
}
}
}
else if (refDCval >= 43) {
if (posBalance < 65) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 2;
}
}
else if (posBalance >= 65) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 115) {
if (Rs < 38) {
if (Cs < 42) {
if (Rs < 18) {
return 8;
}
else if (Rs >= 18) {
return 1;
}
}
else if (Cs >= 42) {
return 4;
}
}
else if (Rs >= 38) {
return 2;
}
}
else if (SC >= 115) {
if (TSC < 4) {
if (mu_mv_mag_sq < 22) {
if (ssDCval < 37) {
return 1;
}
else if (ssDCval >= 37) {
if (negBalance < 45) {
if (refDCval < 49) {
if (Rs < 239) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Rs >= 239) {
return 1;
}
}
else if (refDCval >= 49) {
return 2;
}
}
else if (negBalance >= 45) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 22) {
if (negBalance < 35) {
if (posBalance < 42) {
if (Rs < 158) {
return 1;
}
else if (Rs >= 158) {
return 4;
}
}
else if (posBalance >= 42) {
if (MVDiff < 51) {
return 1;
}
else if (MVDiff >= 51) {
return 4;
}
}
}
else if (negBalance >= 35) {
if (AFD < 8) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
return 2;
}
}
else if (AFD >= 8) {
if (Rs < 118) {
return 2;
}
else if (Rs >= 118) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (diffMVdiffVal < -1) {
if (Cs < 406) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 1;
}
}
else if (Cs >= 406) {
if (Rs < 232) {
return 4;
}
else if (Rs >= 232) {
return 8;
}
}
}
else if (diffMVdiffVal >= -1) {
if (tcor < 100) {
return 1;
}
else if (tcor >= 100) {
return 2;
}
}
}
}
}
}
else if (Rs >= 272) {
if (refDCval < 44) {
return 1;
}
else if (refDCval >= 44) {
if (Cs < 163) {
return 1;
}
else if (Cs >= 163) {
return 4;
}
}
}
}
else if (mcTcor >= 98) {
if (tcor < 91) {
if (tcor < 69) {
if (Rs < 100) {
return 2;
}
else if (Rs >= 100) {
return 4;
}
}
else if (tcor >= 69) {
if (AFD < 6) {
if (Rs < 45) {
return 1;
}
else if (Rs >= 45) {
return 4;
}
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (tcor >= 91) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (SC < 530) {
if (negBalance < 27) {
if (mu_mv_mag_sq < 102) {
if (mcTcor < 97) {
if (tcor < 91) {
if (mu_mv_mag_sq < 59) {
if (Rs < 124) {
return 1;
}
else if (Rs >= 124) {
return 8;
}
}
else if (mu_mv_mag_sq >= 59) {
if (tcor < 78) {
if (posBalance < 27) {
if (Rs < 84) {
return 8;
}
else if (Rs >= 84) {
return 4;
}
}
else if (posBalance >= 27) {
if (Cs < 146) {
return 4;
}
else if (Cs >= 146) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
return 1;
}
}
}
}
else if (tcor >= 78) {
return 1;
}
}
}
else if (tcor >= 91) {
if (ssDCval < 43) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (ssDCval >= 43) {
if (Cs < 176) {
return 4;
}
else if (Cs >= 176) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 28) {
if (Cs < 123) {
return 2;
}
else if (Cs >= 123) {
if (Rs < 125) {
return 1;
}
else if (Rs >= 125) {
return 4;
}
}
}
else if (posBalance >= 28) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (negBalance < 16) {
if (TSCindex < 2) {
if (ssDCval < 39) {
if (Rs < 92) {
return 4;
}
else if (Rs >= 92) {
return 8;
}
}
else if (ssDCval >= 39) {
if (Rs < 185) {
return 8;
}
else if (Rs >= 185) {
return 4;
}
}
}
else if (TSCindex >= 2) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 1;
}
}
}
else if (negBalance >= 16) {
if (posBalance < 68) {
if (Rs < 256) {
if (ssDCval < 35) {
if (Rs < 127) {
return 4;
}
else if (Rs >= 127) {
return 1;
}
}
else if (ssDCval >= 35) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (Rs < 42) {
return 8;
}
else if (Rs >= 42) {
return 1;
}
}
}
}
else if (Rs >= 256) {
return 4;
}
}
else if (posBalance >= 68) {
return 4;
}
}
}
}
else if (negBalance >= 27) {
if (gchDC < 1) {
if (SC < 161) {
if (SC < 133) {
return 8;
}
else if (SC >= 133) {
if (Rs < 76) {
return 1;
}
else if (Rs >= 76) {
if (Rs < 91) {
return 8;
}
else if (Rs >= 91) {
return 4;
}
}
}
}
else if (SC >= 161) {
if (diffRsCsdiff < -1) {
if (mcTcor < 90) {
if (Rs < 194) {
return 2;
}
else if (Rs >= 194) {
return 4;
}
}
else if (mcTcor >= 90) {
if (Cs < 107) {
return 1;
}
else if (Cs >= 107) {
return 8;
}
}
}
else if (diffRsCsdiff >= -1) {
if (Cs < 322) {
return 8;
}
else if (Cs >= 322) {
return 1;
}
}
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 96) {
if (Rs < 232) {
if (mcTcor < 103) {
if (tcor < 78) {
if (Rs < 66) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
else if (Rs >= 66) {
if (posBalance < 49) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (posBalance >= 49) {
if (diffMVdiffVal < 11) {
return 2;
}
else if (diffMVdiffVal >= 11) {
if (Cs < 95) {
return 4;
}
else if (Cs >= 95) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 4;
}
}
}
}
}
}
else if (tcor >= 78) {
if (TSC < 3) {
if (posBalance < 23) {
return 8;
}
else if (posBalance >= 23) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 50) {
return 1;
}
else if (Rs >= 50) {
return 8;
}
}
}
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (mcTcor >= 103) {
return 4;
}
}
else if (Rs >= 232) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (mu_mv_mag_sq < 58) {
return 8;
}
else if (mu_mv_mag_sq >= 58) {
if (SC < 365) {
return 1;
}
else if (SC >= 365) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 269) {
return 4;
}
else if (Rs >= 269) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (negBalance < 61) {
if (posBalance < 53) {
if (Rs < 16) {
return 2;
}
else if (Rs >= 16) {
return 8;
}
}
else if (posBalance >= 53) {
if (Rs < 29) {
return 8;
}
else if (Rs >= 29) {
return 4;
}
}
}
else if (negBalance >= 61) {
if (tcor < 76) {
if (Rs < 21) {
return 2;
}
else if (Rs >= 21) {
return 1;
}
}
else if (tcor >= 76) {
return 4;
}
}
}
}
}
}
else if (SC >= 530) {
if (mu_mv_mag_sq < 58) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (negBalance < 40) {
if (SC < 537) {
if (Rs < 237) {
return 8;
}
else if (Rs >= 237) {
return 1;
}
}
else if (SC >= 537) {
return 1;
}
}
else if (negBalance >= 40) {
if (Rs < 282) {
return 8;
}
else if (Rs >= 282) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 50) {
if (refDCval < 70) {
if (posBalance < 31) {
if (Cs < 8) {
return 2;
}
else if (Cs >= 8) {
if (ssDCval < 58) {
if (Cs < 291) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (Cs >= 291) {
if (Rs < 205) {
return 4;
}
else if (Rs >= 205) {
return 1;
}
}
}
else if (ssDCval >= 58) {
if (negBalance < 20) {
return 1;
}
else if (negBalance >= 20) {
if (SC < 145) {
if (Rs < 60) {
return 1;
}
else if (Rs >= 60) {
return 2;
}
}
else if (SC >= 145) {
if (ssDCval < 66) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
else if (ssDCval >= 66) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 31) {
if (TSCindex < 2) {
if (ssDCval < 65) {
if (Rs < 66) {
return 1;
}
else if (Rs >= 66) {
if (TSC < 1) {
if (diffMVdiffVal < -26) {
if (Rs < 135) {
return 4;
}
else if (Rs >= 135) {
return 8;
}
}
else if (diffMVdiffVal >= -26) {
return 4;
}
}
else if (TSC >= 1) {
if (Rs < 245) {
if (Cs < 190) {
if (MVDiff < 3) {
if (Rs < 146) {
if (Rs < 143) {
return 1;
}
else if (Rs >= 143) {
return 4;
}
}
else if (Rs >= 146) {
return 1;
}
}
else if (MVDiff >= 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 1;
}
}
}
else if (Cs >= 190) {
return 4;
}
}
else if (Rs >= 245) {
if (Rs < 269) {
return 1;
}
else if (Rs >= 269) {
return 4;
}
}
}
}
}
else if (ssDCval >= 65) {
if (SC < 415) {
return 2;
}
else if (SC >= 415) {
if (MVDiff < 2) {
if (Cs < 367) {
return 4;
}
else if (Cs >= 367) {
return 1;
}
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
else if (TSCindex >= 2) {
if (CsDiff < 1) {
if (Cs < 189) {
if (mcTcor < 96) {
if (AFD < 5) {
if (Cs < 71) {
if (Rs < 35) {
return 8;
}
else if (Rs >= 35) {
return 2;
}
}
else if (Cs >= 71) {
if (negBalance < 28) {
return 8;
}
else if (negBalance >= 28) {
return 1;
}
}
}
else if (AFD >= 5) {
if (TSC < 5) {
if (mu_mv_mag_sq < 24) {
if (MVDiff < 42) {
return 1;
}
else if (MVDiff >= 42) {
return 4;
}
}
else if (mu_mv_mag_sq >= 24) {
if (Rs < 155) {
return 8;
}
else if (Rs >= 155) {
if (Rs < 224) {
return 4;
}
else if (Rs >= 224) {
return 1;
}
}
}
}
else if (TSC >= 5) {
if (RsDiff < 1) {
if (Rs < 50) {
return 4;
}
else if (Rs >= 50) {
return 2;
}
}
else if (RsDiff >= 1) {
if (Cs < 95) {
return 1;
}
else if (Cs >= 95) {
return 2;
}
}
}
}
}
else if (mcTcor >= 96) {
if (MVDiff < 3) {
if (refDCval < 63) {
if (Rs < 74) {
return 2;
}
else if (Rs >= 74) {
return 1;
}
}
else if (refDCval >= 63) {
if (Cs < 126) {
return 1;
}
else if (Cs >= 126) {
if (Rs < 198) {
return 2;
}
else if (Rs >= 198) {
return 4;
}
}
}
}
else if (MVDiff >= 3) {
if (Rs < 185) {
if (negBalance < 53) {
if (tcor < 93) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (refDCval < 60) {
if (mu_mv_mag_sq < 48) {
if (MVDiff < 36) {
return 1;
}
else if (MVDiff >= 36) {
if (Rs < 106) {
return 4;
}
else if (Rs >= 106) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 48) {
return 4;
}
}
else if (refDCval >= 60) {
if (Cs < 78) {
return 1;
}
else if (Cs >= 78) {
return 4;
}
}
}
}
else if (tcor >= 93) {
if (refDCval < 57) {
if (negBalance < 29) {
if (Rs < 150) {
if (Rs < 132) {
return 1;
}
else if (Rs >= 132) {
return 4;
}
}
else if (Rs >= 150) {
return 1;
}
}
else if (negBalance >= 29) {
return 1;
}
}
else if (refDCval >= 57) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 1;
}
}
}
}
else if (negBalance >= 53) {
if (Rs < 42) {
return 4;
}
else if (Rs >= 42) {
return 8;
}
}
}
else if (Rs >= 185) {
if (posBalance < 52) {
if (MVDiff < 6) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (posBalance >= 52) {
if (AFD < 6) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 8;
}
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
}
else if (Cs >= 189) {
if (negBalance < 32) {
if (ssDCval < 60) {
if (MVDiff < 1) {
if (mcTcor < 98) {
if (ssDCval < 55) {
if (Rs < 223) {
return 1;
}
else if (Rs >= 223) {
return 4;
}
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 232) {
return 4;
}
else if (Rs >= 232) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
return 4;
}
}
}
}
else if (ssDCval >= 60) {
if (negBalance < 29) {
if (MVDiff < 5) {
if (negBalance < 23) {
if (Cs < 340) {
if (Rs < 128) {
return 2;
}
else if (Rs >= 128) {
if (mu_mv_mag_sq < 76) {
return 2;
}
else if (mu_mv_mag_sq >= 76) {
return 4;
}
}
}
else if (Cs >= 340) {
if (Rs < 195) {
if (mu_mv_mag_sq < 121) {
return 4;
}
else if (mu_mv_mag_sq >= 121) {
if (Rs < 169) {
return 4;
}
else if (Rs >= 169) {
return 1;
}
}
}
else if (Rs >= 195) {
return 4;
}
}
}
else if (negBalance >= 23) {
if (mu_mv_mag_sq < 107) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (mu_mv_mag_sq < 98) {
if (Rs < 278) {
return 1;
}
else if (Rs >= 278) {
return 4;
}
}
else if (mu_mv_mag_sq >= 98) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (posBalance < 47) {
if (ssDCval < 69) {
if (refDCval < 61) {
return 1;
}
else if (refDCval >= 61) {
if (Cs < 356) {
return 4;
}
else if (Cs >= 356) {
if (Rs < 186) {
return 1;
}
else if (Rs >= 186) {
return 4;
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 161) {
return 4;
}
else if (Rs >= 161) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (mu_mv_mag_sq < 119) {
if (mu_mv_mag_sq < 110) {
if (Rs < 182) {
return 1;
}
else if (Rs >= 182) {
return 4;
}
}
else if (mu_mv_mag_sq >= 110) {
if (Cs < 372) {
return 4;
}
else if (Cs >= 372) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 5) {
if (Cs < 221) {
if (Rs < 251) {
return 1;
}
else if (Rs >= 251) {
return 4;
}
}
else if (Cs >= 221) {
if (ssDCval < 65) {
if (Rs < 277) {
return 4;
}
else if (Rs >= 277) {
return 2;
}
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
else if (negBalance >= 29) {
if (posBalance < 48) {
if (mu_mv_mag_sq < 122) {
if (Rs < 142) {
return 1;
}
else if (Rs >= 142) {
return 4;
}
}
else if (mu_mv_mag_sq >= 122) {
return 1;
}
}
else if (posBalance >= 48) {
if (tcor < 98) {
if (ssDCval < 61) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (ssDCval >= 61) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Rs < 156) {
return 1;
}
else if (Rs >= 156) {
return 4;
}
}
}
}
else if (tcor >= 98) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 32) {
if (refDCval < 63) {
if (MVDiff < 1) {
if (Cs < 226) {
return 4;
}
else if (Cs >= 226) {
return 1;
}
}
else if (MVDiff >= 1) {
if (SC < 388) {
if (Rs < 125) {
return 2;
}
else if (Rs >= 125) {
return 1;
}
}
else if (SC >= 388) {
return 4;
}
}
}
else if (refDCval >= 63) {
if (Rs < 183) {
if (negBalance < 33) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (negBalance >= 33) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 1;
}
}
}
else if (Rs >= 183) {
return 1;
}
}
}
}
}
else if (CsDiff >= 1) {
if (TSC < 4) {
if (Rs < 266) {
if (Rs < 51) {
return 1;
}
else if (Rs >= 51) {
if (MVDiff < 3) {
if (posBalance < 47) {
if (gchDC < 1) {
if (Cs < 365) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (Rs < 182) {
return 1;
}
else if (Rs >= 182) {
return 4;
}
}
}
else if (Cs >= 365) {
return 1;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (posBalance >= 47) {
if (Cs < 405) {
if (mcTcor < 98) {
if (Cs < 312) {
return 1;
}
else if (Cs >= 312) {
if (MVDiff < 1) {
if (Rs < 171) {
return 1;
}
else if (Rs >= 171) {
return 4;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 174) {
return 1;
}
else if (Rs >= 174) {
if (Cs < 207) {
return 1;
}
else if (Cs >= 207) {
return 4;
}
}
}
}
else if (Cs >= 405) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Rs < 228) {
return 4;
}
else if (Rs >= 228) {
return 1;
}
}
}
}
}
else if (MVDiff >= 3) {
if (gchDC < 2) {
if (mcTcor < 100) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (ssDCval < 53) {
if (Cs < 329) {
if (Cs < 258) {
return 2;
}
else if (Cs >= 258) {
return 1;
}
}
else if (Cs >= 329) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 2;
}
}
}
}
else if (ssDCval >= 53) {
if (diffTSC < -1) {
if (RsDiff < 1) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
else if (RsDiff >= 1) {
if (negBalance < 41) {
return 8;
}
else if (negBalance >= 41) {
return 4;
}
}
}
else if (diffTSC >= -1) {
if (SC < 408) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (SC < 334) {
return 4;
}
else if (SC >= 334) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 229) {
return 1;
}
else if (Rs >= 229) {
return 4;
}
}
}
}
}
}
else if (SC >= 408) {
return 4;
}
}
}
}
}
else if (mcTcor >= 100) {
if (AFD < 5) {
if (Rs < 166) {
return 8;
}
else if (Rs >= 166) {
return 4;
}
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 69) {
if (RsCsDiff < 5) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (refDCval < 52) {
return 1;
}
else if (refDCval >= 52) {
return 4;
}
}
}
}
else if (RsCsDiff >= 5) {
return 1;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
}
}
}
else if (Rs >= 266) {
if (gchDC < 1) {
if (mu_mv_mag_sq < 83) {
if (tcor < 91) {
if (Cs < 189) {
return 4;
}
else if (Cs >= 189) {
if (posBalance < 50) {
if (Rs < 290) {
return 1;
}
else if (Rs >= 290) {
return 2;
}
}
else if (posBalance >= 50) {
return 1;
}
}
}
else if (tcor >= 91) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 83) {
return 1;
}
}
else if (gchDC >= 1) {
if (tcor < 93) {
if (diffAFD < -1) {
if (ssDCval < 61) {
if (Cs < 404) {
return 8;
}
else if (Cs >= 404) {
return 1;
}
}
else if (ssDCval >= 61) {
if (Rs < 295) {
return 2;
}
else if (Rs >= 295) {
return 1;
}
}
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (tcor >= 93) {
if (MVDiff < 7) {
if (Cs < 276) {
return 1;
}
else if (Cs >= 276) {
return 4;
}
}
else if (MVDiff >= 7) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
}
}
}
}
else if (TSC >= 4) {
if (MVDiff < 122) {
if (MVDiff < 24) {
if (mu_mv_mag_sq < 30) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
if (Rs < 256) {
if (Rs < 240) {
return 2;
}
else if (Rs >= 240) {
return 8;
}
}
else if (Rs >= 256) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (Rs < 123) {
return 8;
}
else if (Rs >= 123) {
if (SC < 523) {
return 4;
}
else if (SC >= 523) {
return 1;
}
}
}
}
}
else if (MVDiff >= 24) {
if (Cs < 272) {
if (tcor < 72) {
if (posBalance < 61) {
if (negBalance < 59) {
if (Rs < 227) {
return 2;
}
else if (Rs >= 227) {
return 4;
}
}
else if (negBalance >= 59) {
return 1;
}
}
else if (posBalance >= 61) {
if (posBalance < 68) {
if (diffAFD < 3) {
if (Cs < 260) {
return 4;
}
else if (Cs >= 260) {
return 1;
}
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (posBalance >= 68) {
return 1;
}
}
}
else if (tcor >= 72) {
if (mu_mv_mag_sq < 28) {
if (diffMVdiffVal < -30) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
return 2;
}
}
else if (diffMVdiffVal >= -30) {
if (Cs < 112) {
if (Rs < 190) {
return 2;
}
else if (Rs >= 190) {
return 4;
}
}
else if (Cs >= 112) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 28) {
if (SC < 445) {
return 4;
}
else if (SC >= 445) {
return 2;
}
}
}
}
else if (Cs >= 272) {
if (tcor < 71) {
if (Cs < 410) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
return 4;
}
}
}
else if (Cs >= 410) {
if (Rs < 262) {
if (Rs < 257) {
return 4;
}
else if (Rs >= 257) {
return 8;
}
}
else if (Rs >= 262) {
return 1;
}
}
}
else if (tcor >= 71) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 122) {
if (negBalance < 59) {
if (TSC < 14) {
if (gchDC < 9) {
if (posBalance < 65) {
if (Rs < 112) {
return 1;
}
else if (Rs >= 112) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (posBalance >= 65) {
return 1;
}
}
else if (gchDC >= 9) {
if (Cs < 141) {
return 8;
}
else if (Cs >= 141) {
return 1;
}
}
}
else if (TSC >= 14) {
if (SC < 316) {
if (Rs < 190) {
return 4;
}
else if (Rs >= 190) {
return 2;
}
}
else if (SC >= 316) {
return 1;
}
}
}
else if (negBalance >= 59) {
if (AFD < 28) {
return 2;
}
else if (AFD >= 28) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 106) {
if (TSCindex < 2) {
if (ssDCval < 81) {
if (ssDCval < 78) {
if (SC < 223) {
return 2;
}
else if (SC >= 223) {
if (ssDCval < 75) {
if (Cs < 158) {
return 2;
}
else if (Cs >= 158) {
return 1;
}
}
else if (ssDCval >= 75) {
return 1;
}
}
}
else if (ssDCval >= 78) {
return 2;
}
}
else if (ssDCval >= 81) {
if (Cs < 140) {
if (Rs < 85) {
if (Rs < 8) {
return 1;
}
else if (Rs >= 8) {
return 4;
}
}
else if (Rs >= 85) {
if (MVDiff < 35) {
return 1;
}
else if (MVDiff >= 35) {
return 8;
}
}
}
else if (Cs >= 140) {
if (MVDiff < 41) {
if (Cs < 163) {
if (Rs < 127) {
return 4;
}
else if (Rs >= 127) {
if (Rs < 243) {
return 2;
}
else if (Rs >= 243) {
return 1;
}
}
}
else if (Cs >= 163) {
return 4;
}
}
else if (MVDiff >= 41) {
return 2;
}
}
}
}
else if (TSCindex >= 2) {
if (mcTcor < 97) {
if (MVDiff < 22) {
if (SC < 279) {
if (AFD < 3) {
if (Rs < 94) {
return 4;
}
else if (Rs >= 94) {
return 8;
}
}
else if (AFD >= 3) {
if (AFD < 9) {
if (Cs < 151) {
return 1;
}
else if (Cs >= 151) {
if (Rs < 92) {
return 4;
}
else if (Rs >= 92) {
return 1;
}
}
}
else if (AFD >= 9) {
if (Cs < 85) {
return 8;
}
else if (Cs >= 85) {
return 1;
}
}
}
}
else if (SC >= 279) {
if (mcTcor < 95) {
if (ssDCval < 85) {
if (MVDiff < 11) {
if (Rs < 276) {
return 8;
}
else if (Rs >= 276) {
return 2;
}
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (ssDCval >= 85) {
if (AFD < 11) {
if (Rs < 238) {
if (tcor < 78) {
if (Cs < 289) {
if (negBalance < 48) {
if (diffRsCsdiff < -1) {
return 8;
}
else if (diffRsCsdiff >= -1) {
return 2;
}
}
else if (negBalance >= 48) {
if (tcor < 67) {
return 8;
}
else if (tcor >= 67) {
if (Cs < 266) {
return 2;
}
else if (Cs >= 266) {
return 8;
}
}
}
}
else if (Cs >= 289) {
return 8;
}
}
else if (tcor >= 78) {
if (SC < 586) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (mcTcor < 81) {
return 8;
}
else if (mcTcor >= 81) {
return 2;
}
}
}
else if (SC >= 586) {
return 8;
}
}
}
else if (Rs >= 238) {
if (mcTcor < 84) {
if (mu_mv_mag_sq < 52) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
else if (mu_mv_mag_sq >= 52) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (mcTcor >= 84) {
return 8;
}
}
}
else if (AFD >= 11) {
if (Cs < 330) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (Cs >= 330) {
if (Rs < 233) {
return 8;
}
else if (Rs >= 233) {
return 2;
}
}
}
}
}
else if (mcTcor >= 95) {
if (diffAFD < 1) {
if (diffRsCsdiff < 1) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (negBalance < 43) {
if (SC < 499) {
return 1;
}
else if (SC >= 499) {
if (Cs < 413) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
else if (Cs >= 413) {
return 2;
}
}
}
else if (negBalance >= 43) {
if (diffMVdiffVal < -4) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 1) {
if (ssDCval < 96) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
if (Rs < 199) {
return 1;
}
else if (Rs >= 199) {
return 8;
}
}
}
else if (ssDCval >= 96) {
if (Rs < 290) {
return 8;
}
else if (Rs >= 290) {
return 2;
}
}
}
}
else if (diffAFD >= 1) {
if (ssDCval < 79) {
if (Rs < 148) {
return 4;
}
else if (Rs >= 148) {
return 8;
}
}
else if (ssDCval >= 79) {
return 1;
}
}
}
}
}
else if (MVDiff >= 22) {
if (mcTcor < 71) {
if (gchDC < 7) {
if (AFD < 28) {
return 4;
}
else if (AFD >= 28) {
return 8;
}
}
else if (gchDC >= 7) {
if (MVDiff < 210) {
if (Cs < 426) {
if (TSC < 20) {
return 2;
}
else if (TSC >= 20) {
return 8;
}
}
else if (Cs >= 426) {
return 1;
}
}
else if (MVDiff >= 210) {
if (Rs < 218) {
return 1;
}
else if (Rs >= 218) {
return 4;
}
}
}
}
else if (mcTcor >= 71) {
if (Cs < 24) {
if (mu_mv_mag_sq < 67) {
if (Rs < 287) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (mu_mv_mag_sq < 35) {
if (ssDCval < 113) {
return 1;
}
else if (ssDCval >= 113) {
if (ssDCval < 169) {
if (Cs < 7) {
return 8;
}
else if (Cs >= 7) {
return 2;
}
}
else if (ssDCval >= 169) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (Cs < 18) {
return 8;
}
else if (Cs >= 18) {
if (Rs < 54) {
return 1;
}
else if (Rs >= 54) {
return 2;
}
}
}
}
}
else if (Rs >= 287) {
return 8;
}
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 148) {
return 8;
}
else if (Rs >= 148) {
return 2;
}
}
}
else if (Cs >= 24) {
if (ssDCval < 133) {
if (negBalance < 45) {
if (diffAFD < -3) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
return 2;
}
}
else if (diffAFD >= -3) {
if (Rs < 9) {
if (Cs < 110) {
return 1;
}
else if (Cs >= 110) {
return 2;
}
}
else if (Rs >= 9) {
if (Cs < 379) {
if (diffMVdiffVal < -33) {
if (Rs < 97) {
return 1;
}
else if (Rs >= 97) {
return 2;
}
}
else if (diffMVdiffVal >= -33) {
return 1;
}
}
else if (Cs >= 379) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (tcor < 97) {
if (mu_mv_mag_sq < 53) {
if (ssDCval < 128) {
if (posBalance < 42) {
if (Cs < 401) {
return 1;
}
else if (Cs >= 401) {
return 2;
}
}
else if (posBalance >= 42) {
if (diffMVdiffVal < -91) {
if (Rs < 290) {
return 1;
}
else if (Rs >= 290) {
return 4;
}
}
else if (diffMVdiffVal >= -91) {
if (mcTcor < 89) {
if (mu_mv_mag_sq < 9) {
if (posBalance < 65) {
if (Cs < 76) {
return 1;
}
else if (Cs >= 76) {
if (gchDC < 8) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (Rs < 214) {
return 8;
}
else if (Rs >= 214) {
return 2;
}
}
}
else if (gchDC >= 8) {
return 1;
}
}
}
else if (posBalance >= 65) {
return 4;
}
}
else if (mu_mv_mag_sq >= 9) {
if (ssDCval < 88) {
if (Rs < 260) {
if (Rs < 94) {
return 2;
}
else if (Rs >= 94) {
return 8;
}
}
else if (Rs >= 260) {
return 2;
}
}
else if (ssDCval >= 88) {
if (Cs < 435) {
return 2;
}
else if (Cs >= 435) {
return 8;
}
}
}
}
else if (mcTcor >= 89) {
if (mcTcor < 91) {
if (Rs < 282) {
return 1;
}
else if (Rs >= 282) {
return 8;
}
}
else if (mcTcor >= 91) {
if (MVDiff < 31) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
return 8;
}
}
else if (MVDiff >= 31) {
if (diffMVdiffVal < -9) {
return 2;
}
else if (diffMVdiffVal >= -9) {
if (tcor < 52) {
return 2;
}
else if (tcor >= 52) {
if (gchDC < 3) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (gchDC >= 3) {
if (negBalance < 46) {
return 1;
}
else if (negBalance >= 46) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 128) {
if (Rs < 213) {
return 2;
}
else if (Rs >= 213) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (AFD < 15) {
if (Cs < 34) {
return 2;
}
else if (Cs >= 34) {
return 8;
}
}
else if (AFD >= 15) {
if (Cs < 128) {
return 4;
}
else if (Cs >= 128) {
if (negBalance < 64) {
return 1;
}
else if (negBalance >= 64) {
if (Rs < 188) {
return 4;
}
else if (Rs >= 188) {
return 2;
}
}
}
}
}
}
else if (tcor >= 97) {
if (Cs < 294) {
return 8;
}
else if (Cs >= 294) {
return 4;
}
}
}
}
else if (ssDCval >= 133) {
if (diffTSC < -3) {
return 1;
}
else if (diffTSC >= -3) {
if (MVDiff < 230) {
return 2;
}
else if (MVDiff >= 230) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 106) {
if (Cs < 240) {
if (Cs < 79) {
if (posBalance < 57) {
if (SC < 185) {
return 1;
}
else if (SC >= 185) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
if (SC < 188) {
return 8;
}
else if (SC >= 188) {
return 1;
}
}
}
}
else if (posBalance >= 57) {
if (ssDCval < 97) {
if (AFD < 14) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (Rs < 144) {
return 1;
}
else if (Rs >= 144) {
return 2;
}
}
}
else if (AFD >= 14) {
return 2;
}
}
else if (ssDCval >= 97) {
if (negBalance < 33) {
if (Rs < 30) {
return 1;
}
else if (Rs >= 30) {
return 4;
}
}
else if (negBalance >= 33) {
return 2;
}
}
}
}
else if (Cs >= 79) {
if (Rs < 55) {
if (Rs < 21) {
return 2;
}
else if (Rs >= 21) {
if (refDCval < 90) {
if (Cs < 95) {
if (Rs < 49) {
return 8;
}
else if (Rs >= 49) {
return 4;
}
}
else if (Cs >= 95) {
return 1;
}
}
else if (refDCval >= 90) {
return 4;
}
}
}
else if (Rs >= 55) {
if (AFD < 2) {
if (ssDCval < 99) {
if (mu_mv_mag_sq < 74) {
return 1;
}
else if (mu_mv_mag_sq >= 74) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Rs < 160) {
if (Cs < 99) {
return 2;
}
else if (Cs >= 99) {
return 1;
}
}
else if (Rs >= 160) {
return 2;
}
}
}
}
else if (ssDCval >= 99) {
return 1;
}
}
else if (AFD >= 2) {
if (Rs < 122) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (mu_mv_mag_sq < 63) {
if (AFD < 8) {
if (MVDiff < 23) {
if (ssDCval < 96) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
if (Rs < 110) {
return 4;
}
else if (Rs >= 110) {
return 1;
}
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (MVDiff >= 23) {
return 1;
}
}
else if (AFD >= 8) {
if (tcor < 85) {
return 1;
}
else if (tcor >= 85) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
}
}
}
else if (Rs >= 122) {
if (refDCval < 91) {
if (TSCindex < 3) {
if (ssDCval < 74) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
else if (ssDCval >= 74) {
if (Cs < 148) {
if (Rs < 156) {
return 1;
}
else if (Rs >= 156) {
return 8;
}
}
else if (Cs >= 148) {
if (Rs < 229) {
return 2;
}
else if (Rs >= 229) {
return 1;
}
}
}
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
if (posBalance < 51) {
if (Rs < 217) {
return 2;
}
else if (Rs >= 217) {
return 1;
}
}
else if (posBalance >= 51) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (CsDiff < 2) {
if (gchDC < 2) {
if (Rs < 280) {
if (Cs < 222) {
if (mu_mv_mag_sq < 101) {
if (MVDiff < 2) {
if (Rs < 234) {
return 8;
}
else if (Rs >= 234) {
return 1;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 101) {
return 2;
}
}
else if (Cs >= 222) {
if (Rs < 231) {
return 8;
}
else if (Rs >= 231) {
return 2;
}
}
}
else if (Rs >= 280) {
return 8;
}
}
else if (gchDC >= 2) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (Rs < 292) {
return 4;
}
else if (Rs >= 292) {
return 2;
}
}
}
}
}
else if (refDCval >= 91) {
if (MVDiff < 12) {
if (MVDiff < 6) {
if (tcor < 99) {
if (Rs < 164) {
return 1;
}
else if (Rs >= 164) {
return 4;
}
}
else if (tcor >= 99) {
return 2;
}
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (MVDiff >= 12) {
if (AFD < 6) {
if (Cs < 110) {
return 8;
}
else if (Cs >= 110) {
return 1;
}
}
else if (AFD >= 6) {
if (mu_mv_mag_sq < 21) {
if (Rs < 162) {
if (Rs < 141) {
return 1;
}
else if (Rs >= 141) {
return 2;
}
}
else if (Rs >= 162) {
return 4;
}
}
else if (mu_mv_mag_sq >= 21) {
if (SC < 343) {
return 2;
}
else if (SC >= 343) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 240) {
if (refDCval < 76) {
if (ssDCval < 72) {
if (Rs < 131) {
if (Cs < 247) {
if (Rs < 127) {
return 1;
}
else if (Rs >= 127) {
return 4;
}
}
else if (Cs >= 247) {
return 4;
}
}
else if (Rs >= 131) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
if (mu_mv_mag_sq < 92) {
return 1;
}
else if (mu_mv_mag_sq >= 92) {
return 4;
}
}
}
}
else if (ssDCval >= 72) {
if (mu_mv_mag_sq < 34) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 34) {
return 1;
}
}
}
else if (refDCval >= 76) {
if (RsDiff < 1) {
if (Cs < 366) {
if (tcor < 96) {
if (refDCval < 78) {
return 4;
}
else if (refDCval >= 78) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 2;
}
}
}
else if (tcor >= 96) {
if (Rs < 224) {
if (refDCval < 98) {
return 4;
}
else if (refDCval >= 98) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
else if (Rs >= 224) {
return 2;
}
}
}
else if (Cs >= 366) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
return 2;
}
}
}
else if (RsDiff >= 1) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 22) {
return 2;
}
else if (mu_mv_mag_sq >= 22) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
}
}
else if (RsDiff >= 2) {
if (negBalance < 53) {
if (SC < 626) {
return 1;
}
else if (SC >= 626) {
return 2;
}
}
else if (negBalance >= 53) {
if (SC < 529) {
return 8;
}
else if (SC >= 529) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 106) {
if (negBalance < 23) {
if (ssDCval < 136) {
return 4;
}
else if (ssDCval >= 136) {
return 2;
}
}
else if (negBalance >= 23) {
if (tcor < 101) {
if (AFD < 16) {
if (mu_mv_mag_sq < 97) {
if (SC < 275) {
if (negBalance < 29) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
else if (negBalance >= 29) {
if (Cs < 186) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (Rs < 62) {
return 1;
}
else if (Rs >= 62) {
return 2;
}
}
}
else if (Cs >= 186) {
return 1;
}
}
}
else if (SC >= 275) {
if (refDCval < 111) {
if (Rs < 157) {
if (Rs < 129) {
return 1;
}
else if (Rs >= 129) {
return 4;
}
}
else if (Rs >= 157) {
return 2;
}
}
else if (refDCval >= 111) {
if (ssDCval < 166) {
return 2;
}
else if (ssDCval >= 166) {
if (Rs < 261) {
return 2;
}
else if (Rs >= 261) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 112) {
return 1;
}
else if (Rs >= 112) {
return 2;
}
}
}
}
else if (AFD >= 16) {
if (MVDiff < 31) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
else if (MVDiff >= 31) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
if (Cs < 48) {
return 8;
}
else if (Cs >= 48) {
return 2;
}
}
}
}
}
else if (tcor >= 101) {
if (ssDCval < 160) {
return 8;
}
else if (ssDCval >= 160) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 130) {
if (refDCval < 75) {
if (Rs < 112) {
return 4;
}
else if (Rs >= 112) {
return 8;
}
}
else if (refDCval >= 75) {
if (mu_mv_mag_sq < 114) {
return 4;
}
else if (mu_mv_mag_sq >= 114) {
if (diffRsCsdiff < -1) {
if (Rs < 83) {
if (Rs < 40) {
return 4;
}
else if (Rs >= 40) {
return 8;
}
}
else if (Rs >= 83) {
if (Cs < 241) {
return 1;
}
else if (Cs >= 241) {
return 2;
}
}
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
}
}
else if (Rs >= 130) {
if (Cs < 239) {
if (negBalance < 25) {
if (Cs < 133) {
if (Cs < 90) {
return 1;
}
else if (Cs >= 90) {
if (MVDiff < 1) {
if (Cs < 107) {
return 8;
}
else if (Cs >= 107) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (Rs < 140) {
return 8;
}
else if (Rs >= 140) {
return 2;
}
}
}
}
else if (MVDiff >= 1) {
if (Rs < 144) {
return 2;
}
else if (Rs >= 144) {
return 8;
}
}
}
}
else if (Cs >= 133) {
if (SC < 407) {
if (diffTSC < -1) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (Rs < 165) {
return 2;
}
else if (Rs >= 165) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (Rs < 230) {
return 2;
}
else if (Rs >= 230) {
return 8;
}
}
}
else if (SC >= 407) {
if (ssDCval < 83) {
if (Rs < 204) {
if (Cs < 223) {
return 1;
}
else if (Cs >= 223) {
return 2;
}
}
else if (Rs >= 204) {
if (diffMVdiffVal < -2) {
if (Rs < 219) {
return 1;
}
else if (Rs >= 219) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
}
else if (ssDCval >= 83) {
return 1;
}
}
}
}
else if (negBalance >= 25) {
if (SC < 381) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
if (RsCsDiff < 2) {
if (Cs < 154) {
return 8;
}
else if (Cs >= 154) {
return 2;
}
}
else if (RsCsDiff >= 2) {
if (Cs < 30) {
return 1;
}
else if (Cs >= 30) {
return 2;
}
}
}
}
}
else if (SC >= 381) {
if (ssDCval < 77) {
return 1;
}
else if (ssDCval >= 77) {
return 2;
}
}
}
}
else if (Cs >= 239) {
if (ssDCval < 78) {
if (Rs < 200) {
return 8;
}
else if (Rs >= 200) {
return 1;
}
}
else if (ssDCval >= 78) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 304) {
if (RsDiff < 4) {
if (refDCval < 73) {
if (mcTcor < 87) {
if (tcor < 74) {
if (negBalance < 32) {
if (refDCval < 50) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
if (MVDiff < 34) {
return 8;
}
else if (MVDiff >= 34) {
if (Rs < 347) {
return 2;
}
else if (Rs >= 347) {
return 1;
}
}
}
}
else if (refDCval >= 50) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
return 8;
}
}
}
else if (negBalance >= 32) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Cs < 204) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (mcTcor < 68) {
return 1;
}
else if (mcTcor >= 68) {
if (Cs < 165) {
if (Cs < 33) {
return 2;
}
else if (Cs >= 33) {
return 8;
}
}
else if (Cs >= 165) {
return 2;
}
}
}
}
else if (Cs >= 204) {
if (diffAFD < 1) {
if (Rs < 686) {
if (ssDCval < 40) {
if (tcor < 58) {
if (Rs < 402) {
return 4;
}
else if (Rs >= 402) {
return 2;
}
}
else if (tcor >= 58) {
if (Rs < 362) {
if (Rs < 337) {
return 8;
}
else if (Rs >= 337) {
return 2;
}
}
else if (Rs >= 362) {
return 8;
}
}
}
else if (ssDCval >= 40) {
if (mcTcor < 76) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
return 4;
}
}
}
else if (mcTcor >= 76) {
if (Cs < 370) {
if (mu_mv_mag_sq < 14) {
return 8;
}
else if (mu_mv_mag_sq >= 14) {
return 4;
}
}
else if (Cs >= 370) {
return 4;
}
}
}
}
else if (Rs >= 686) {
if (Rs < 728) {
return 1;
}
else if (Rs >= 728) {
return 4;
}
}
}
else if (diffAFD >= 1) {
if (tcor < 67) {
if (posBalance < 51) {
if (Rs < 633) {
return 1;
}
else if (Rs >= 633) {
return 8;
}
}
else if (posBalance >= 51) {
if (refDCval < 67) {
if (diffRsCsdiff < -16) {
return 2;
}
else if (diffRsCsdiff >= -16) {
if (diffMVdiffVal < 20) {
return 8;
}
else if (diffMVdiffVal >= 20) {
return 4;
}
}
}
else if (refDCval >= 67) {
if (Rs < 370) {
return 2;
}
else if (Rs >= 370) {
return 4;
}
}
}
}
else if (tcor >= 67) {
return 2;
}
}
}
}
}
}
else if (tcor >= 74) {
if (negBalance < 51) {
if (negBalance < 34) {
if (ssDCval < 43) {
if (Cs < 312) {
return 8;
}
else if (Cs >= 312) {
return 1;
}
}
else if (ssDCval >= 43) {
return 4;
}
}
else if (negBalance >= 34) {
if (Cs < 155) {
if (Cs < 46) {
return 8;
}
else if (Cs >= 46) {
return 1;
}
}
else if (Cs >= 155) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 355) {
return 2;
}
else if (Rs >= 355) {
return 8;
}
}
}
}
}
else if (negBalance >= 51) {
if (posBalance < 54) {
if (Cs < 502) {
return 8;
}
else if (Cs >= 502) {
return 1;
}
}
else if (posBalance >= 54) {
return 4;
}
}
}
}
else if (mcTcor >= 87) {
if (Rs < 1280) {
if (tcor < 101) {
if (TSC < 1) {
if (mcTcor < 99) {
return 4;
}
else if (mcTcor >= 99) {
if (Rs < 1007) {
if (Cs < 115) {
return 4;
}
else if (Cs >= 115) {
if (ssDCval < 71) {
if (diffAFD < -1) {
if (Rs < 766) {
return 8;
}
else if (Rs >= 766) {
if (Rs < 931) {
return 1;
}
else if (Rs >= 931) {
return 8;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (ssDCval >= 71) {
return 4;
}
}
}
else if (Rs >= 1007) {
return 1;
}
}
}
else if (TSC >= 1) {
if (MVDiff < 2) {
if (refDCval < 69) {
if (mcTcor < 97) {
if (Rs < 418) {
if (mu_mv_mag_sq < 115) {
return 1;
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 363) {
return 4;
}
else if (Rs >= 363) {
return 8;
}
}
}
else if (Rs >= 418) {
if (tcor < 90) {
if (negBalance < 37) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (negBalance >= 37) {
return 8;
}
}
else if (tcor >= 90) {
if (posBalance < 48) {
if (mu_mv_mag_sq < 102) {
return 1;
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 983) {
if (RsCsDiff < 1) {
return 1;
}
else if (RsCsDiff >= 1) {
return 8;
}
}
else if (Rs >= 983) {
return 1;
}
}
}
else if (posBalance >= 48) {
if (SC < 942) {
if (SC < 869) {
return 8;
}
else if (SC >= 869) {
return 2;
}
}
else if (SC >= 942) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 40) {
if (negBalance < 25) {
if (posBalance < 43) {
if (Rs < 897) {
if (mu_mv_mag_sq < 127) {
if (refDCval < 56) {
if (mu_mv_mag_sq < 79) {
return 8;
}
else if (mu_mv_mag_sq >= 79) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 738) {
return 1;
}
else if (Rs >= 738) {
return 8;
}
}
}
}
else if (refDCval >= 56) {
if (Cs < 332) {
return 1;
}
else if (Cs >= 332) {
if (Cs < 359) {
return 8;
}
else if (Cs >= 359) {
if (Rs < 595) {
if (mu_mv_mag_sq < 104) {
return 8;
}
else if (mu_mv_mag_sq >= 104) {
if (Cs < 461) {
return 1;
}
else if (Cs >= 461) {
if (Rs < 561) {
return 8;
}
else if (Rs >= 561) {
return 1;
}
}
}
}
else if (Rs >= 595) {
if (ssDCval < 66) {
if (Cs < 422) {
if (Rs < 687) {
return 8;
}
else if (Rs >= 687) {
return 1;
}
}
else if (Cs >= 422) {
return 8;
}
}
else if (ssDCval >= 66) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 1;
}
}
else if (Rs >= 897) {
return 1;
}
}
else if (posBalance >= 43) {
if (Rs < 577) {
if (Cs < 302) {
if (Rs < 418) {
return 4;
}
else if (Rs >= 418) {
return 8;
}
}
else if (Cs >= 302) {
if (diffRsCsdiff < -1) {
if (Rs < 525) {
return 1;
}
else if (Rs >= 525) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
if (diffAFD < -1) {
if (Rs < 550) {
return 1;
}
else if (Rs >= 550) {
return 8;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
else if (Rs >= 577) {
if (AFD < 2) {
if (mu_mv_mag_sq < 122) {
return 2;
}
else if (mu_mv_mag_sq >= 122) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
else if (AFD >= 2) {
if (Cs < 485) {
if (ssDCval < 69) {
if (TSC < 2) {
if (ssDCval < 62) {
if (Rs < 764) {
return 8;
}
else if (Rs >= 764) {
return 2;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (TSC >= 2) {
return 1;
}
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (Cs >= 485) {
return 4;
}
}
}
}
}
else if (negBalance >= 25) {
if (posBalance < 55) {
if (SC < 956) {
if (tcor < 100) {
return 1;
}
else if (tcor >= 100) {
return 4;
}
}
else if (SC >= 956) {
if (Cs < 505) {
if (diffTSC < 1) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (TSC < 3) {
if (diffMVdiffVal < -2) {
if (Rs < 954) {
return 8;
}
else if (Rs >= 954) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (ssDCval < 62) {
if (Rs < 681) {
return 8;
}
else if (Rs >= 681) {
return 2;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
}
else if (Cs >= 505) {
if (RsCsDiff < 1) {
return 1;
}
else if (RsCsDiff >= 1) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
if (refDCval < 67) {
return 1;
}
else if (refDCval >= 67) {
return 2;
}
}
}
}
else if (negBalance >= 40) {
if (AFD < 5) {
if (Rs < 894) {
return 4;
}
else if (Rs >= 894) {
return 1;
}
}
else if (AFD >= 5) {
if (Rs < 875) {
return 8;
}
else if (Rs >= 875) {
return 4;
}
}
}
}
}
else if (refDCval >= 69) {
if (SC < 1325) {
if (Rs < 601) {
if (SC < 977) {
if (Rs < 399) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (Rs >= 399) {
if (tcor < 99) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 4;
}
}
else if (tcor >= 99) {
if (Rs < 524) {
return 1;
}
else if (Rs >= 524) {
return 2;
}
}
}
}
else if (SC >= 977) {
if (Rs < 562) {
return 4;
}
else if (Rs >= 562) {
return 1;
}
}
}
else if (Rs >= 601) {
if (ssDCval < 72) {
if (tcor < 99) {
if (SC < 1056) {
return 4;
}
else if (SC >= 1056) {
return 1;
}
}
else if (tcor >= 99) {
if (Rs < 610) {
return 1;
}
else if (Rs >= 610) {
return 2;
}
}
}
else if (ssDCval >= 72) {
return 4;
}
}
}
else if (SC >= 1325) {
if (SC < 1633) {
return 8;
}
else if (SC >= 1633) {
return 1;
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 126) {
if (RsDiff < 2) {
if (Cs < 80) {
if (TSC < 3) {
if (Rs < 666) {
return 1;
}
else if (Rs >= 666) {
return 4;
}
}
else if (TSC >= 3) {
if (Rs < 433) {
return 1;
}
else if (Rs >= 433) {
return 8;
}
}
}
else if (Cs >= 80) {
if (refDCval < 64) {
return 1;
}
else if (refDCval >= 64) {
if (Rs < 882) {
return 4;
}
else if (Rs >= 882) {
return 1;
}
}
}
}
else if (RsDiff >= 2) {
if (SC < 974) {
if (tcor < 83) {
if (diffMVdiffVal < -24) {
return 8;
}
else if (diffMVdiffVal >= -24) {
if (Cs < 86) {
return 1;
}
else if (Cs >= 86) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
return 8;
}
}
}
}
else if (tcor >= 83) {
if (Rs < 317) {
return 4;
}
else if (Rs >= 317) {
if (refDCval < 69) {
return 1;
}
else if (refDCval >= 69) {
if (tcor < 90) {
return 4;
}
else if (tcor >= 90) {
return 1;
}
}
}
}
}
else if (SC >= 974) {
if (RsDiff < 3) {
if (Rs < 995) {
return 2;
}
else if (Rs >= 995) {
return 1;
}
}
else if (RsDiff >= 3) {
return 1;
}
}
}
}
else if (Cs >= 126) {
if (ssDCval < 71) {
if (refDCval < 44) {
if (negBalance < 32) {
if (posBalance < 34) {
if (mu_mv_mag_sq < 70) {
return 2;
}
else if (mu_mv_mag_sq >= 70) {
if (diffAFD < -1) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (posBalance >= 34) {
if (Cs < 426) {
if (TSC < 2) {
if (RsDiff < 1) {
if (Rs < 372) {
return 1;
}
else if (Rs >= 372) {
return 8;
}
}
else if (RsDiff >= 1) {
return 4;
}
}
else if (TSC >= 2) {
if (MVDiff < 3) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (Cs >= 426) {
if (Rs < 692) {
return 1;
}
else if (Rs >= 692) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
if (Rs < 798) {
return 1;
}
else if (Rs >= 798) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 32) {
if (SC < 1455) {
if (posBalance < 49) {
if (tcor < 93) {
if (Cs < 468) {
return 1;
}
else if (Cs >= 468) {
if (Rs < 431) {
return 8;
}
else if (Rs >= 431) {
return 1;
}
}
}
else if (tcor >= 93) {
if (Rs < 459) {
return 2;
}
else if (Rs >= 459) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (SC < 781) {
if (CsDiff < 1) {
if (Cs < 153) {
return 1;
}
else if (Cs >= 153) {
return 8;
}
}
else if (CsDiff >= 1) {
if (diffTSC < -1) {
if (Rs < 372) {
return 8;
}
else if (Rs >= 372) {
return 2;
}
}
else if (diffTSC >= -1) {
if (negBalance < 35) {
if (Rs < 346) {
return 1;
}
else if (Rs >= 346) {
return 4;
}
}
else if (negBalance >= 35) {
return 1;
}
}
}
}
else if (SC >= 781) {
if (Rs < 899) {
if (ssDCval < 41) {
if (Rs < 752) {
if (diffMVdiffVal < -3) {
if (Rs < 627) {
return 8;
}
else if (Rs >= 627) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (Rs >= 752) {
return 1;
}
}
else if (ssDCval >= 41) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (Rs >= 899) {
return 1;
}
}
}
}
else if (SC >= 1455) {
return 2;
}
}
}
else if (refDCval >= 44) {
if (SC < 709) {
if (Cs < 264) {
if (Cs < 206) {
if (mu_mv_mag_sq < 65) {
if (tcor < 96) {
if (RsDiff < 3) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (Rs < 402) {
if (ssDCval < 45) {
return 1;
}
else if (ssDCval >= 45) {
return 4;
}
}
else if (Rs >= 402) {
return 1;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 465) {
return 8;
}
else if (Rs >= 465) {
return 1;
}
}
}
else if (tcor >= 96) {
if (Cs < 176) {
return 1;
}
else if (Cs >= 176) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 65) {
if (ssDCval < 50) {
if (Rs < 331) {
return 4;
}
else if (Rs >= 331) {
return 1;
}
}
else if (ssDCval >= 50) {
if (Cs < 204) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (Rs < 405) {
return 4;
}
else if (Rs >= 405) {
return 2;
}
}
}
else if (Cs >= 204) {
return 8;
}
}
}
}
else if (Cs >= 206) {
if (posBalance < 55) {
if (ssDCval < 49) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 1;
}
}
}
else if (ssDCval >= 49) {
if (Rs < 305) {
return 2;
}
else if (Rs >= 305) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (ssDCval < 53) {
if (Rs < 390) {
return 8;
}
else if (Rs >= 390) {
return 1;
}
}
else if (ssDCval >= 53) {
if (Cs < 255) {
return 4;
}
else if (Cs >= 255) {
return 2;
}
}
}
}
}
}
else if (Cs >= 264) {
if (mcTcor < 89) {
if (mu_mv_mag_sq < 40) {
if (Cs < 320) {
return 1;
}
else if (Cs >= 320) {
return 4;
}
}
else if (mu_mv_mag_sq >= 40) {
return 8;
}
}
else if (mcTcor >= 89) {
if (Rs < 315) {
if (Rs < 307) {
return 1;
}
else if (Rs >= 307) {
if (ssDCval < 62) {
if (Cs < 361) {
return 2;
}
else if (Cs >= 361) {
return 1;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
}
else if (Rs >= 315) {
if (tcor < 98) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (mcTcor < 96) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (Rs < 327) {
return 1;
}
else if (Rs >= 327) {
if (ssDCval < 67) {
return 2;
}
else if (ssDCval >= 67) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (posBalance < 66) {
return 4;
}
else if (posBalance >= 66) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
}
}
else if (SC >= 709) {
if (mu_mv_mag_sq < 3) {
if (tcor < 97) {
if (refDCval < 64) {
if (negBalance < 37) {
if (Rs < 642) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
if (Rs < 413) {
return 1;
}
else if (Rs >= 413) {
return 2;
}
}
}
else if (Rs >= 642) {
if (mcTcor < 99) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (mcTcor >= 99) {
return 2;
}
}
}
else if (negBalance >= 37) {
if (mcTcor < 96) {
if (AFD < 6) {
if (Rs < 560) {
return 2;
}
else if (Rs >= 560) {
return 1;
}
}
else if (AFD >= 6) {
if (MVDiff < 67) {
if (gchDC < 1) {
if (Rs < 613) {
return 8;
}
else if (Rs >= 613) {
if (Rs < 742) {
return 4;
}
else if (Rs >= 742) {
return 2;
}
}
}
else if (gchDC >= 1) {
if (MVDiff < 26) {
if (RsCsDiff < 12) {
if (Rs < 696) {
return 1;
}
else if (Rs >= 696) {
return 4;
}
}
else if (RsCsDiff >= 12) {
return 4;
}
}
else if (MVDiff >= 26) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
else if (MVDiff >= 67) {
return 2;
}
}
}
else if (mcTcor >= 96) {
if (TSC < 4) {
if (Cs < 377) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (Cs >= 377) {
return 2;
}
}
else if (TSC >= 4) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 1;
}
}
}
}
}
else if (refDCval >= 64) {
if (Rs < 812) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
if (Cs < 443) {
if (Rs < 591) {
return 4;
}
else if (Rs >= 591) {
return 1;
}
}
else if (Cs >= 443) {
return 1;
}
}
}
else if (Rs >= 812) {
if (mcTcor < 95) {
if (Rs < 960) {
return 2;
}
else if (Rs >= 960) {
return 4;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
else if (tcor >= 97) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 99) {
if (mcTcor < 97) {
if (SC < 1192) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (refDCval < 56) {
if (diffMVdiffVal < 11) {
if (diffAFD < 1) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (ssDCval < 52) {
if (diffMVdiffVal < -6) {
return 4;
}
else if (diffMVdiffVal >= -6) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 388) {
return 1;
}
else if (Cs >= 388) {
return 4;
}
}
}
}
else if (ssDCval >= 52) {
if (diffRsCsdiff < -1) {
if (Rs < 596) {
return 1;
}
else if (Rs >= 596) {
return 4;
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
}
else if (diffAFD >= 1) {
if (Rs < 422) {
if (Rs < 352) {
return 1;
}
else if (Rs >= 352) {
return 4;
}
}
else if (Rs >= 422) {
return 1;
}
}
}
else if (diffMVdiffVal >= 11) {
if (SC < 894) {
return 4;
}
else if (SC >= 894) {
return 2;
}
}
}
else if (refDCval >= 56) {
if (refDCval < 62) {
if (mcTcor < 92) {
if (RsCsDiff < 7) {
return 2;
}
else if (RsCsDiff >= 7) {
return 1;
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 85) {
if (posBalance < 63) {
if (tcor < 93) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 446) {
return 2;
}
else if (Rs >= 446) {
return 1;
}
}
}
else if (tcor >= 93) {
if (Cs < 397) {
return 1;
}
else if (Cs >= 397) {
return 2;
}
}
}
else if (posBalance >= 63) {
return 1;
}
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 484) {
if (Rs < 381) {
return 8;
}
else if (Rs >= 381) {
return 1;
}
}
else if (Rs >= 484) {
if (Rs < 579) {
return 8;
}
else if (Rs >= 579) {
return 2;
}
}
}
}
}
else if (refDCval >= 62) {
if (diffRsCsdiff < -13) {
return 8;
}
else if (diffRsCsdiff >= -13) {
if (Cs < 309) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (Cs >= 309) {
if (Rs < 383) {
return 1;
}
else if (Rs >= 383) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (AFD < 11) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
else if (AFD >= 11) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (SC >= 1192) {
if (ssDCval < 55) {
if (Rs < 928) {
return 8;
}
else if (Rs >= 928) {
return 1;
}
}
else if (ssDCval >= 55) {
if (Cs < 489) {
if (mu_mv_mag_sq < 66) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 66) {
return 4;
}
}
else if (Cs >= 489) {
if (ssDCval < 63) {
if (mcTcor < 93) {
if (Rs < 781) {
return 1;
}
else if (Rs >= 781) {
return 2;
}
}
else if (mcTcor >= 93) {
return 8;
}
}
else if (ssDCval >= 63) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 762) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 85) {
if (Cs < 399) {
return 1;
}
else if (Cs >= 399) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (ssDCval < 46) {
return 4;
}
else if (ssDCval >= 46) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (tcor < 97) {
if (negBalance < 41) {
if (CsDiff < 2) {
if (AFD < 3) {
if (Rs < 487) {
return 2;
}
else if (Rs >= 487) {
return 4;
}
}
else if (AFD >= 3) {
if (refDCval < 55) {
if (Cs < 277) {
if (Rs < 499) {
return 2;
}
else if (Rs >= 499) {
return 4;
}
}
else if (Cs >= 277) {
if (gchDC < 2) {
if (ssDCval < 48) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (ssDCval >= 48) {
return 1;
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (refDCval >= 55) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (RsDiff < 2) {
if (Rs < 558) {
return 4;
}
else if (Rs >= 558) {
return 8;
}
}
else if (RsDiff >= 2) {
if (Cs < 457) {
return 4;
}
else if (Cs >= 457) {
return 1;
}
}
}
}
else if (negBalance >= 41) {
if (AFD < 10) {
if (mu_mv_mag_sq < 21) {
return 4;
}
else if (mu_mv_mag_sq >= 21) {
if (Rs < 409) {
return 4;
}
else if (Rs >= 409) {
return 8;
}
}
}
else if (AFD >= 10) {
if (Cs < 385) {
return 4;
}
else if (Cs >= 385) {
return 1;
}
}
}
}
else if (tcor >= 97) {
if (RsCsDiff < 2) {
if (Rs < 476) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Rs >= 476) {
if (Cs < 335) {
if (diffMVdiffVal < -7) {
return 1;
}
else if (diffMVdiffVal >= -7) {
return 4;
}
}
else if (Cs >= 335) {
if (Rs < 643) {
return 1;
}
else if (Rs >= 643) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
if (MVDiff < 14) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
return 4;
}
}
else if (MVDiff >= 14) {
return 1;
}
}
}
}
}
else if (Rs >= 762) {
if (Cs < 503) {
if (diffRsCsdiff < 4) {
if (ssDCval < 52) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
else if (ssDCval >= 52) {
return 1;
}
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
else if (Cs >= 503) {
if (Rs < 884) {
return 1;
}
else if (Rs >= 884) {
return 2;
}
}
}
}
}
else if (mcTcor >= 99) {
if (SC < 982) {
if (Cs < 486) {
if (negBalance < 24) {
if (negBalance < 15) {
return 1;
}
else if (negBalance >= 15) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (SC < 779) {
return 8;
}
else if (SC >= 779) {
return 4;
}
}
}
}
else if (negBalance >= 24) {
if (Cs < 305) {
return 1;
}
else if (Cs >= 305) {
if (AFD < 4) {
if (Rs < 552) {
return 8;
}
else if (Rs >= 552) {
return 2;
}
}
else if (AFD >= 4) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
}
else if (Cs >= 486) {
return 4;
}
}
else if (SC >= 982) {
if (ssDCval < 67) {
if (RsCsDiff < 5) {
if (Cs < 530) {
return 1;
}
else if (Cs >= 530) {
return 4;
}
}
else if (RsCsDiff >= 5) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
if (AFD < 4) {
if (Rs < 1240) {
return 2;
}
else if (Rs >= 1240) {
return 4;
}
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (posBalance < 40) {
if (Rs < 875) {
return 1;
}
else if (Rs >= 875) {
return 8;
}
}
else if (posBalance >= 40) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (tcor < 97) {
if (mcTcor < 97) {
if (Rs < 501) {
return 8;
}
else if (Rs >= 501) {
return 4;
}
}
else if (mcTcor >= 97) {
if (SC < 777) {
if (Rs < 519) {
if (Rs < 412) {
return 1;
}
else if (Rs >= 412) {
return 8;
}
}
else if (Rs >= 519) {
return 4;
}
}
else if (SC >= 777) {
return 1;
}
}
}
else if (tcor >= 97) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 101) {
if (posBalance < 1) {
return 4;
}
else if (posBalance >= 1) {
if (posBalance < 34) {
if (ssDCval < 42) {
return 1;
}
else if (ssDCval >= 42) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
if (Cs < 13) {
return 4;
}
else if (Cs >= 13) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (Cs < 82) {
return 1;
}
else if (Cs >= 82) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 34) {
if (Rs < 585) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
else if (Rs >= 585) {
if (Rs < 735) {
return 4;
}
else if (Rs >= 735) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1280) {
if (RsDiff < 2) {
if (SC < 1650) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
if (Cs < 289) {
return 1;
}
else if (Cs >= 289) {
return 8;
}
}
}
else if (SC >= 1650) {
if (ssDCval < 62) {
if (refDCval < 56) {
if (Cs < 362) {
return 2;
}
else if (Cs >= 362) {
return 1;
}
}
else if (refDCval >= 56) {
if (Cs < 418) {
if (posBalance < 34) {
return 1;
}
else if (posBalance >= 34) {
if (Rs < 1539) {
return 8;
}
else if (Rs >= 1539) {
return 2;
}
}
}
else if (Cs >= 418) {
return 8;
}
}
}
else if (ssDCval >= 62) {
if (ssDCval < 64) {
if (mcTcor < 100) {
if (negBalance < 19) {
return 2;
}
else if (negBalance >= 19) {
if (Rs < 1741) {
if (Rs < 1462) {
return 2;
}
else if (Rs >= 1462) {
return 1;
}
}
else if (Rs >= 1741) {
return 2;
}
}
}
else if (mcTcor >= 100) {
if (posBalance < 29) {
return 2;
}
else if (posBalance >= 29) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (Rs < 1728) {
if (ssDCval < 69) {
if (Cs < 490) {
return 1;
}
else if (Cs >= 490) {
if (Rs < 1463) {
return 8;
}
else if (Rs >= 1463) {
return 1;
}
}
}
else if (ssDCval >= 69) {
if (Rs < 1332) {
return 1;
}
else if (Rs >= 1332) {
return 8;
}
}
}
else if (Rs >= 1728) {
return 2;
}
}
}
}
}
else if (RsDiff >= 2) {
if (posBalance < 59) {
if (TSCindex < 3) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 51) {
if (negBalance < 34) {
if (ssDCval < 58) {
if (Rs < 1967) {
return 2;
}
else if (Rs >= 1967) {
return 4;
}
}
else if (ssDCval >= 58) {
if (AFD < 5) {
if (Cs < 304) {
return 1;
}
else if (Cs >= 304) {
return 2;
}
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (negBalance >= 34) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 51) {
if (tcor < 97) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
else if (tcor >= 97) {
if (Cs < 439) {
return 8;
}
else if (Cs >= 439) {
if (Cs < 521) {
return 1;
}
else if (Cs >= 521) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 59) {
if (Cs < 196) {
return 1;
}
else if (Cs >= 196) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 73) {
if (Rs < 1622) {
if (Cs < 173) {
if (MVDiff < 8) {
if (ssDCval < 88) {
if (Cs < 61) {
return 8;
}
else if (Cs >= 61) {
if (TSC < 3) {
if (mu_mv_mag_sq < 110) {
if (Rs < 324) {
return 8;
}
else if (Rs >= 324) {
return 1;
}
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 782) {
return 2;
}
else if (Rs >= 782) {
return 8;
}
}
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (ssDCval >= 88) {
return 8;
}
}
else if (MVDiff >= 8) {
if (tcor < 58) {
if (mcTcor < 88) {
if (Cs < 38) {
return 4;
}
else if (Cs >= 38) {
return 2;
}
}
else if (mcTcor >= 88) {
if (MVDiff < 25) {
if (Rs < 345) {
return 8;
}
else if (Rs >= 345) {
return 2;
}
}
else if (MVDiff >= 25) {
if (Rs < 381) {
return 8;
}
else if (Rs >= 381) {
return 2;
}
}
}
}
else if (tcor >= 58) {
if (RsDiff < 2) {
if (posBalance < 51) {
if (negBalance < 42) {
if (negBalance < 17) {
return 8;
}
else if (negBalance >= 17) {
if (diffMVdiffVal < -6) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
return 1;
}
}
}
else if (negBalance >= 42) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Cs < 95) {
return 1;
}
else if (Cs >= 95) {
return 8;
}
}
}
}
else if (posBalance >= 51) {
if (Cs < 75) {
if (SC < 835) {
return 1;
}
else if (SC >= 835) {
return 8;
}
}
else if (Cs >= 75) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (AFD < 13) {
if (mcTcor < 99) {
if (ssDCval < 136) {
return 1;
}
else if (ssDCval >= 136) {
if (Rs < 858) {
return 1;
}
else if (Rs >= 858) {
return 4;
}
}
}
else if (mcTcor >= 99) {
if (mu_mv_mag_sq < 16) {
if (Cs < 134) {
if (Rs < 619) {
return 1;
}
else if (Rs >= 619) {
return 8;
}
}
else if (Cs >= 134) {
return 4;
}
}
else if (mu_mv_mag_sq >= 16) {
if (Cs < 1) {
return 8;
}
else if (Cs >= 1) {
return 1;
}
}
}
}
else if (AFD >= 13) {
if (Rs < 484) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 1;
}
}
else if (Rs >= 484) {
return 1;
}
}
}
}
}
}
else if (Cs >= 173) {
if (SC < 855) {
if (mcTcor < 97) {
if (MVDiff < 20) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (diffAFD < 1) {
if (Cs < 486) {
if (RsCsDiff < 11) {
if (Rs < 416) {
if (mcTcor < 82) {
return 2;
}
else if (mcTcor >= 82) {
if (Rs < 306) {
return 2;
}
else if (Rs >= 306) {
if (Cs < 318) {
if (diffMVdiffVal < -9) {
return 2;
}
else if (diffMVdiffVal >= -9) {
if (SC < 668) {
return 8;
}
else if (SC >= 668) {
return 2;
}
}
}
else if (Cs >= 318) {
return 8;
}
}
}
}
else if (Rs >= 416) {
if (diffMVdiffVal < 2) {
if (Rs < 548) {
if (diffMVdiffVal < -7) {
return 1;
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
else if (Rs >= 548) {
return 4;
}
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
}
else if (RsCsDiff >= 11) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
return 8;
}
}
}
else if (Cs >= 486) {
return 2;
}
}
else if (diffAFD >= 1) {
if (gchDC < 2) {
if (AFD < 10) {
if (Cs < 253) {
if (Rs < 462) {
return 1;
}
else if (Rs >= 462) {
return 4;
}
}
else if (Cs >= 253) {
return 8;
}
}
else if (AFD >= 10) {
return 1;
}
}
else if (gchDC >= 2) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
if (Rs < 440) {
return 8;
}
else if (Rs >= 440) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 20) {
if (TSC < 10) {
if (refDCval < 88) {
if (AFD < 14) {
if (SC < 752) {
if (Rs < 334) {
return 8;
}
else if (Rs >= 334) {
return 1;
}
}
else if (SC >= 752) {
return 8;
}
}
else if (AFD >= 14) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 8;
}
}
}
else if (refDCval >= 88) {
if (diffMVdiffVal < 48) {
if (MVDiff < 24) {
if (negBalance < 50) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
else if (negBalance >= 50) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
return 8;
}
}
}
else if (MVDiff >= 24) {
if (Rs < 490) {
if (Rs < 326) {
if (Rs < 321) {
return 2;
}
else if (Rs >= 321) {
return 8;
}
}
else if (Rs >= 326) {
if (RsCsDiff < 16) {
return 2;
}
else if (RsCsDiff >= 16) {
if (Rs < 368) {
return 2;
}
else if (Rs >= 368) {
return 4;
}
}
}
}
else if (Rs >= 490) {
return 8;
}
}
}
else if (diffMVdiffVal >= 48) {
if (Cs < 240) {
return 2;
}
else if (Cs >= 240) {
if (Rs < 331) {
return 8;
}
else if (Rs >= 331) {
if (Rs < 450) {
return 4;
}
else if (Rs >= 450) {
return 8;
}
}
}
}
}
}
else if (TSC >= 10) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 82) {
if (AFD < 4) {
if (Cs < 316) {
return 8;
}
else if (Cs >= 316) {
if (Rs < 425) {
return 8;
}
else if (Rs >= 425) {
return 1;
}
}
}
else if (AFD >= 4) {
if (SC < 762) {
if (AFD < 9) {
if (SC < 557) {
return 1;
}
else if (SC >= 557) {
return 2;
}
}
else if (AFD >= 9) {
return 8;
}
}
else if (SC >= 762) {
if (TSC < 3) {
if (Rs < 540) {
return 1;
}
else if (Rs >= 540) {
return 4;
}
}
else if (TSC >= 3) {
if (SC < 790) {
return 1;
}
else if (SC >= 790) {
return 4;
}
}
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 85) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
if (Rs < 498) {
return 2;
}
else if (Rs >= 498) {
if (Rs < 573) {
return 8;
}
else if (Rs >= 573) {
return 4;
}
}
}
}
else if (ssDCval >= 85) {
if (Rs < 373) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (tcor < 87) {
if (refDCval < 105) {
return 4;
}
else if (refDCval >= 105) {
if (Rs < 354) {
if (Rs < 331) {
return 8;
}
else if (Rs >= 331) {
return 1;
}
}
else if (Rs >= 354) {
return 2;
}
}
}
else if (tcor >= 87) {
if (ssDCval < 124) {
if (Cs < 446) {
return 2;
}
else if (Cs >= 446) {
if (ssDCval < 117) {
if (Rs < 324) {
return 8;
}
else if (Rs >= 324) {
return 1;
}
}
else if (ssDCval >= 117) {
return 2;
}
}
}
else if (ssDCval >= 124) {
if (tcor < 96) {
if (Cs < 356) {
return 2;
}
else if (Cs >= 356) {
return 8;
}
}
else if (tcor >= 96) {
return 4;
}
}
}
}
}
else if (Rs >= 373) {
if (negBalance < 33) {
if (mu_mv_mag_sq < 95) {
if (Cs < 397) {
if (MVDiff < 8) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (MVDiff >= 8) {
if (Cs < 348) {
return 1;
}
else if (Cs >= 348) {
return 4;
}
}
}
else if (Cs >= 397) {
return 1;
}
}
else if (mu_mv_mag_sq >= 95) {
if (Rs < 443) {
return 8;
}
else if (Rs >= 443) {
return 1;
}
}
}
else if (negBalance >= 33) {
if (mu_mv_mag_sq < 13) {
if (MVDiff < 23) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 4;
}
}
else if (MVDiff >= 23) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (mu_mv_mag_sq < 42) {
return 8;
}
else if (mu_mv_mag_sq >= 42) {
if (Cs < 396) {
if (Rs < 450) {
return 1;
}
else if (Rs >= 450) {
return 8;
}
}
else if (Cs >= 396) {
return 2;
}
}
}
}
}
}
}
}
}
else if (SC >= 855) {
if (ssDCval < 116) {
if (refDCval < 94) {
if (negBalance < 30) {
if (mu_mv_mag_sq < 120) {
if (refDCval < 83) {
if (SC < 1159) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (negBalance < 26) {
if (SC < 1045) {
return 8;
}
else if (SC >= 1045) {
return 4;
}
}
else if (negBalance >= 26) {
if (Rs < 672) {
return 1;
}
else if (Rs >= 672) {
return 4;
}
}
}
}
else if (SC >= 1159) {
if (Cs < 327) {
return 1;
}
else if (Cs >= 327) {
if (Cs < 430) {
return 8;
}
else if (Cs >= 430) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
return 8;
}
}
}
}
}
else if (refDCval >= 83) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (Cs < 507) {
return 8;
}
else if (Cs >= 507) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (negBalance < 17) {
if (Rs < 1080) {
if (Cs < 258) {
if (Cs < 221) {
return 8;
}
else if (Cs >= 221) {
return 1;
}
}
else if (Cs >= 258) {
return 2;
}
}
else if (Rs >= 1080) {
return 2;
}
}
else if (negBalance >= 17) {
if (Rs < 464) {
return 8;
}
else if (Rs >= 464) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 80) {
if (Cs < 202) {
return 8;
}
else if (Cs >= 202) {
return 1;
}
}
else if (ssDCval >= 80) {
if (Cs < 256) {
return 2;
}
else if (Cs >= 256) {
if (refDCval < 82) {
return 2;
}
else if (refDCval >= 82) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 1;
}
}
}
}
}
else if (negBalance >= 30) {
if (mu_mv_mag_sq < 38) {
if (tcor < 75) {
if (Rs < 621) {
return 2;
}
else if (Rs >= 621) {
return 8;
}
}
else if (tcor >= 75) {
if (ssDCval < 76) {
if (Cs < 301) {
if (Rs < 1138) {
return 4;
}
else if (Rs >= 1138) {
return 1;
}
}
else if (Cs >= 301) {
if (Cs < 451) {
return 1;
}
else if (Cs >= 451) {
return 8;
}
}
}
else if (ssDCval >= 76) {
if (tcor < 98) {
if (Cs < 470) {
if (negBalance < 64) {
if (Cs < 192) {
if (Rs < 942) {
return 4;
}
else if (Rs >= 942) {
return 1;
}
}
else if (Cs >= 192) {
return 1;
}
}
else if (negBalance >= 64) {
return 2;
}
}
else if (Cs >= 470) {
if (tcor < 89) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (Rs < 410) {
return 1;
}
else if (Rs >= 410) {
return 8;
}
}
}
else if (tcor >= 89) {
return 1;
}
}
}
else if (tcor >= 98) {
if (Rs < 714) {
return 1;
}
else if (Rs >= 714) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (tcor < 98) {
if (negBalance < 41) {
if (mu_mv_mag_sq < 108) {
if (MVDiff < 6) {
if (posBalance < 46) {
if (Rs < 829) {
return 4;
}
else if (Rs >= 829) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
return 1;
}
}
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 102) {
return 1;
}
else if (mu_mv_mag_sq >= 102) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 108) {
return 8;
}
}
else if (negBalance >= 41) {
if (ssDCval < 77) {
if (diffMVdiffVal < 1) {
if (Rs < 614) {
return 2;
}
else if (Rs >= 614) {
if (Rs < 683) {
return 8;
}
else if (Rs >= 683) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
else if (ssDCval >= 77) {
if (Cs < 238) {
return 1;
}
else if (Cs >= 238) {
return 8;
}
}
}
}
else if (tcor >= 98) {
if (Rs < 547) {
if (Rs < 439) {
return 8;
}
else if (Rs >= 439) {
return 1;
}
}
else if (Rs >= 547) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (negBalance < 36) {
if (SC < 1060) {
return 8;
}
else if (SC >= 1060) {
return 1;
}
}
else if (negBalance >= 36) {
if (Rs < 1355) {
return 8;
}
else if (Rs >= 1355) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 94) {
if (mu_mv_mag_sq < 82) {
if (mcTcor < 93) {
if (posBalance < 47) {
if (Rs < 447) {
return 8;
}
else if (Rs >= 447) {
if (Rs < 504) {
return 2;
}
else if (Rs >= 504) {
return 1;
}
}
}
else if (posBalance >= 47) {
return 8;
}
}
else if (mcTcor >= 93) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (negBalance < 60) {
if (mcTcor < 94) {
if (SC < 897) {
if (SC < 884) {
return 1;
}
else if (SC >= 884) {
return 2;
}
}
else if (SC >= 897) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (gchDC < 2) {
if (ssDCval < 111) {
if (Rs < 1304) {
return 1;
}
else if (Rs >= 1304) {
return 8;
}
}
else if (ssDCval >= 111) {
if (Rs < 486) {
return 1;
}
else if (Rs >= 486) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (Rs < 588) {
return 1;
}
else if (Rs >= 588) {
return 8;
}
}
}
}
else if (negBalance >= 60) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (Rs < 478) {
return 8;
}
else if (Rs >= 478) {
if (Rs < 600) {
return 1;
}
else if (Rs >= 600) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (Cs < 466) {
if (ssDCval < 98) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (Cs >= 466) {
return 8;
}
}
}
}
else if (ssDCval >= 116) {
if (mcTcor < 98) {
if (MVDiff < 14) {
if (Rs < 1044) {
if (diffRsCsdiff < 4) {
if (ssDCval < 160) {
if (Rs < 914) {
return 8;
}
else if (Rs >= 914) {
if (Rs < 1031) {
return 1;
}
else if (Rs >= 1031) {
return 8;
}
}
}
else if (ssDCval >= 160) {
return 1;
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 733) {
return 2;
}
else if (Rs >= 733) {
return 1;
}
}
}
else if (Rs >= 1044) {
return 4;
}
}
else if (MVDiff >= 14) {
if (Rs < 1096) {
if (negBalance < 57) {
if (Rs < 412) {
if (Rs < 399) {
return 8;
}
else if (Rs >= 399) {
return 2;
}
}
else if (Rs >= 412) {
return 4;
}
}
else if (negBalance >= 57) {
return 2;
}
}
else if (Rs >= 1096) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (refDCval < 118) {
if (Cs < 452) {
if (Rs < 465) {
return 2;
}
else if (Rs >= 465) {
return 1;
}
}
else if (Cs >= 452) {
if (Rs < 383) {
return 4;
}
else if (Rs >= 383) {
return 2;
}
}
}
else if (refDCval >= 118) {
if (mu_mv_mag_sq < 74) {
if (gchDC < 1) {
if (diffRsCsdiff < 2) {
if (TSC < 4) {
if (Rs < 412) {
return 2;
}
else if (Rs >= 412) {
return 1;
}
}
else if (TSC >= 4) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (diffRsCsdiff >= 2) {
return 2;
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 66) {
if (Rs < 369) {
return 2;
}
else if (Rs >= 369) {
return 1;
}
}
else if (mu_mv_mag_sq >= 66) {
if (Rs < 506) {
return 4;
}
else if (Rs >= 506) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 74) {
if (Rs < 573) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (Cs < 445) {
return 4;
}
else if (Cs >= 445) {
return 8;
}
}
}
else if (Rs >= 573) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1622) {
if (Cs < 318) {
if (ssDCval < 102) {
if (Cs < 307) {
return 1;
}
else if (Cs >= 307) {
if (Cs < 310) {
return 8;
}
else if (Cs >= 310) {
return 1;
}
}
}
else if (ssDCval >= 102) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (Cs < 159) {
return 1;
}
else if (Cs >= 159) {
return 4;
}
}
}
}
else if (Cs >= 318) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 101) {
if (tcor < 98) {
if (negBalance < 31) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
return 2;
}
}
}
else if (negBalance >= 31) {
return 1;
}
}
else if (tcor >= 98) {
if (RsDiff < 2) {
if (Rs < 2093) {
return 1;
}
else if (Rs >= 2093) {
return 8;
}
}
else if (RsDiff >= 2) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (MVDiff < 1) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
return 1;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (refDCval < 82) {
return 1;
}
else if (refDCval >= 82) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
if (Rs < 1766) {
return 8;
}
else if (Rs >= 1766) {
return 1;
}
}
}
}
}
else if (posBalance >= 54) {
if (negBalance < 25) {
if (Rs < 2077) {
return 1;
}
else if (Rs >= 2077) {
return 2;
}
}
else if (negBalance >= 25) {
if (Rs < 1813) {
return 8;
}
else if (Rs >= 1813) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (Cs < 390) {
if (RsDiff < 8) {
if (refDCval < 83) {
if (Rs < 724) {
if (Cs < 131) {
if (negBalance < 38) {
if (ssDCval < 58) {
return 2;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (negBalance >= 38) {
return 1;
}
}
else if (Cs >= 131) {
if (gchDC < 7) {
if (posBalance < 54) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (Rs < 554) {
if (tcor < 46) {
return 1;
}
else if (tcor >= 46) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 422) {
return 8;
}
else if (Rs >= 422) {
return 2;
}
}
}
}
else if (Rs >= 554) {
return 1;
}
}
}
else if (posBalance >= 54) {
if (diffRsCsdiff < -6) {
if (negBalance < 50) {
return 4;
}
else if (negBalance >= 50) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 489) {
if (Rs < 423) {
return 8;
}
else if (Rs >= 423) {
return 4;
}
}
else if (Rs >= 489) {
if (Rs < 583) {
return 2;
}
else if (Rs >= 583) {
return 1;
}
}
}
}
}
else if (diffRsCsdiff >= -6) {
if (diffAFD < 1) {
if (CsDiff < 4) {
if (Rs < 356) {
return 1;
}
else if (Rs >= 356) {
return 4;
}
}
else if (CsDiff >= 4) {
if (Rs < 643) {
return 2;
}
else if (Rs >= 643) {
return 4;
}
}
}
else if (diffAFD >= 1) {
if (SC < 786) {
if (posBalance < 62) {
if (Cs < 323) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 2;
}
}
else if (Cs >= 323) {
if (Rs < 335) {
return 1;
}
else if (Rs >= 335) {
return 8;
}
}
}
else if (posBalance >= 62) {
return 8;
}
}
else if (SC >= 786) {
if (tcor < 63) {
return 1;
}
else if (tcor >= 63) {
if (Rs < 603) {
return 4;
}
else if (Rs >= 603) {
return 2;
}
}
}
}
}
}
}
else if (gchDC >= 7) {
if (Cs < 384) {
return 1;
}
else if (Cs >= 384) {
return 4;
}
}
}
}
else if (Rs >= 724) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
if (Cs < 353) {
return 1;
}
else if (Cs >= 353) {
if (gchDC < 3) {
if (Rs < 767) {
return 4;
}
else if (Rs >= 767) {
return 2;
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
}
else if (refDCval >= 83) {
if (tcor < 68) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
if (tcor < 36) {
if (RsCsDiff < 39) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
else if (RsCsDiff >= 39) {
if (Rs < 845) {
return 1;
}
else if (Rs >= 845) {
return 4;
}
}
}
else if (tcor >= 36) {
if (Cs < 90) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 536) {
return 2;
}
else if (Rs >= 536) {
if (Rs < 684) {
return 1;
}
else if (Rs >= 684) {
return 8;
}
}
}
}
else if (Cs >= 90) {
if (AFD < 13) {
if (MVDiff < 24) {
if (Rs < 886) {
return 8;
}
else if (Rs >= 886) {
return 2;
}
}
else if (MVDiff >= 24) {
return 2;
}
}
else if (AFD >= 13) {
if (mu_mv_mag_sq < 37) {
if (diffAFD < -4) {
if (Cs < 338) {
return 2;
}
else if (Cs >= 338) {
return 4;
}
}
else if (diffAFD >= -4) {
if (MVDiff < 107) {
if (ssDCval < 93) {
if (Rs < 830) {
return 2;
}
else if (Rs >= 830) {
return 1;
}
}
else if (ssDCval >= 93) {
return 1;
}
}
else if (MVDiff >= 107) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 37) {
return 2;
}
}
}
}
}
}
else if (tcor >= 68) {
if (Cs < 239) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 1;
}
}
else if (Cs >= 239) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
}
else if (RsDiff >= 8) {
if (refDCval < 110) {
if (SC < 1972) {
if (ssDCval < 88) {
if (negBalance < 59) {
if (mu_mv_mag_sq < 59) {
if (tcor < 56) {
if (ssDCval < 44) {
return 2;
}
else if (ssDCval >= 44) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (TSC < 10) {
if (Rs < 1017) {
return 2;
}
else if (Rs >= 1017) {
return 4;
}
}
else if (TSC >= 10) {
return 1;
}
}
}
}
else if (tcor >= 56) {
return 1;
}
}
else if (mu_mv_mag_sq >= 59) {
if (MVDiff < 50) {
return 2;
}
else if (MVDiff >= 50) {
return 1;
}
}
}
else if (negBalance >= 59) {
if (gchDC < 13) {
if (Rs < 707) {
return 4;
}
else if (Rs >= 707) {
return 2;
}
}
else if (gchDC >= 13) {
return 1;
}
}
}
else if (ssDCval >= 88) {
if (Cs < 68) {
return 4;
}
else if (Cs >= 68) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
if (diffRsCsdiff < -86) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 2;
}
}
else if (diffRsCsdiff >= -86) {
return 2;
}
}
}
}
}
else if (SC >= 1972) {
if (MVDiff < 28) {
if (diffMVdiffVal < -10) {
if (Rs < 2139) {
return 2;
}
else if (Rs >= 2139) {
return 1;
}
}
else if (diffMVdiffVal >= -10) {
if (diffAFD < 4) {
return 2;
}
else if (diffAFD >= 4) {
return 1;
}
}
}
else if (MVDiff >= 28) {
if (Cs < 227) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
return 2;
}
}
}
else if (Cs >= 227) {
if (ssDCval < 58) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 1;
}
}
else if (ssDCval >= 58) {
if (tcor < 67) {
return 4;
}
else if (tcor >= 67) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 110) {
if (mcTcor < 93) {
if (refDCval < 124) {
if (Rs < 1524) {
if (Cs < 328) {
return 2;
}
else if (Cs >= 328) {
return 4;
}
}
else if (Rs >= 1524) {
return 8;
}
}
else if (refDCval >= 124) {
if (mu_mv_mag_sq < 45) {
if (refDCval < 138) {
return 1;
}
else if (refDCval >= 138) {
if (Rs < 1919) {
if (Rs < 1108) {
return 4;
}
else if (Rs >= 1108) {
return 1;
}
}
else if (Rs >= 1919) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (CsDiff < 3) {
if (SC < 1193) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 1;
}
}
else if (SC >= 1193) {
return 1;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
if (Cs < 72) {
return 2;
}
else if (Cs >= 72) {
return 1;
}
}
}
}
}
else if (Cs >= 390) {
if (negBalance < 48) {
if (Rs < 1675) {
if (diffRsCsdiff < 14) {
if (RsDiff < 5) {
if (diffMVdiffVal < 0) {
if (diffAFD < -1) {
if (Cs < 472) {
return 1;
}
else if (Cs >= 472) {
return 8;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (diffMVdiffVal >= 0) {
return 8;
}
}
else if (RsDiff >= 5) {
if (MVDiff < 49) {
if (Cs < 520) {
if (mu_mv_mag_sq < 9) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 1;
}
}
else if (mu_mv_mag_sq >= 9) {
return 1;
}
}
else if (Cs >= 520) {
return 2;
}
}
else if (MVDiff >= 49) {
if (Rs < 959) {
if (Rs < 484) {
return 8;
}
else if (Rs >= 484) {
return 2;
}
}
else if (Rs >= 959) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 14) {
if (Rs < 1362) {
if (ssDCval < 111) {
return 1;
}
else if (ssDCval >= 111) {
return 8;
}
}
else if (Rs >= 1362) {
return 8;
}
}
}
else if (Rs >= 1675) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (AFD < 23) {
if (ssDCval < 124) {
if (MVDiff < 86) {
if (diffRsCsdiff < 15) {
if (MVDiff < 8) {
if (Rs < 659) {
return 1;
}
else if (Rs >= 659) {
return 8;
}
}
else if (MVDiff >= 8) {
if (posBalance < 58) {
if (gchDC < 3) {
if (mu_mv_mag_sq < 25) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (negBalance < 58) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (negBalance >= 58) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (Rs < 555) {
return 4;
}
else if (Rs >= 555) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (Rs < 1372) {
return 4;
}
else if (Rs >= 1372) {
return 2;
}
}
}
else if (posBalance >= 58) {
if (tcor < 72) {
if (RsCsDiff < 31) {
if (CsDiff < 3) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
else if (CsDiff >= 3) {
if (diffRsCsdiff < 3) {
return 4;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 536) {
return 2;
}
else if (Rs >= 536) {
return 8;
}
}
}
}
else if (RsCsDiff >= 31) {
if (diffMVdiffVal < 10) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
if (diffRsCsdiff < 12) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 988) {
return 4;
}
else if (Rs >= 988) {
return 8;
}
}
}
else if (diffRsCsdiff >= 12) {
return 4;
}
}
}
else if (diffMVdiffVal >= 10) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 4;
}
}
}
}
else if (tcor >= 72) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 15) {
if (SC < 1616) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (MVDiff < 64) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 4;
}
}
else if (MVDiff >= 64) {
return 8;
}
}
}
else if (SC >= 1616) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
return 8;
}
}
}
}
else if (MVDiff >= 86) {
if (CsDiff < 4) {
if (MVDiff < 146) {
return 2;
}
else if (MVDiff >= 146) {
return 4;
}
}
else if (CsDiff >= 4) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
}
else if (ssDCval >= 124) {
if (TSC < 8) {
if (tcor < 40) {
return 1;
}
else if (tcor >= 40) {
if (Rs < 831) {
return 8;
}
else if (Rs >= 831) {
return 2;
}
}
}
else if (TSC >= 8) {
if (Rs < 1066) {
return 2;
}
else if (Rs >= 1066) {
return 4;
}
}
}
}
else if (AFD >= 23) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 44) {
return 4;
}
else if (mu_mv_mag_sq >= 44) {
if (Rs < 1491) {
return 8;
}
else if (Rs >= 1491) {
return 2;
}
}
}
else if (ssDCval >= 69) {
if (tcor < 54) {
if (posBalance < 53) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (posBalance >= 53) {
if (mu_mv_mag_sq < 41) {
if (Rs < 921) {
return 1;
}
else if (Rs >= 921) {
return 2;
}
}
else if (mu_mv_mag_sq >= 41) {
if (Rs < 773) {
return 2;
}
else if (Rs >= 773) {
return 4;
}
}
}
}
else if (tcor >= 54) {
if (mu_mv_mag_sq < 57) {
if (gchDC < 9) {
if (CsDiff < 4) {
if (Rs < 791) {
return 1;
}
else if (Rs >= 791) {
return 8;
}
}
else if (CsDiff >= 4) {
if (MVDiff < 34) {
if (Rs < 942) {
return 1;
}
else if (Rs >= 942) {
return 4;
}
}
else if (MVDiff >= 34) {
return 1;
}
}
}
else if (gchDC >= 9) {
if (RsDiff < 16) {
return 8;
}
else if (RsDiff >= 16) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (ssDCval < 128) {
if (Rs < 1558) {
return 1;
}
else if (Rs >= 1558) {
return 8;
}
}
else if (ssDCval >= 128) {
return 2;
}
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 85) {
if (RsCsDiff < 74) {
if (Cs < 119) {
if (ssDCval < 140) {
if (posBalance < 101) {
if (posBalance < 29) {
if (Rs < 915) {
return 8;
}
else if (Rs >= 915) {
return 1;
}
}
else if (posBalance >= 29) {
if (TSC < 3) {
if (Rs < 1498) {
return 1;
}
else if (Rs >= 1498) {
return 2;
}
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (posBalance >= 101) {
return 2;
}
}
else if (ssDCval >= 140) {
if (Rs < 648) {
return 1;
}
else if (Rs >= 648) {
return 4;
}
}
}
else if (Cs >= 119) {
if (Rs < 599) {
if (Cs < 188) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (Rs < 553) {
return 8;
}
else if (Rs >= 553) {
return 2;
}
}
}
else if (Cs >= 188) {
if (negBalance < 47) {
if (Rs < 481) {
if (MVDiff < 25) {
if (Rs < 359) {
return 2;
}
else if (Rs >= 359) {
return 1;
}
}
else if (MVDiff >= 25) {
return 2;
}
}
else if (Rs >= 481) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
if (Rs < 582) {
return 4;
}
else if (Rs >= 582) {
if (Rs < 596) {
return 1;
}
else if (Rs >= 596) {
return 2;
}
}
}
}
}
else if (negBalance >= 47) {
if (posBalance < 53) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (ssDCval < 115) {
return 4;
}
else if (ssDCval >= 115) {
if (Rs < 571) {
return 2;
}
else if (Rs >= 571) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
if (tcor < 90) {
if (posBalance < 62) {
if (Cs < 426) {
if (MVDiff < 67) {
return 1;
}
else if (MVDiff >= 67) {
return 8;
}
}
else if (Cs >= 426) {
return 1;
}
}
else if (posBalance >= 62) {
return 8;
}
}
else if (tcor >= 90) {
return 2;
}
}
}
}
}
else if (Rs >= 599) {
if (ssDCval < 139) {
if (Cs < 273) {
if (MVDiff < 12) {
if (RsCsDiff < 28) {
if (gchDC < 3) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 1;
}
}
else if (gchDC >= 3) {
return 2;
}
}
else if (RsCsDiff >= 28) {
if (tcor < 81) {
return 2;
}
else if (tcor >= 81) {
if (posBalance < 57) {
if (diffRsCsdiff < -3) {
return 2;
}
else if (diffRsCsdiff >= -3) {
return 1;
}
}
else if (posBalance >= 57) {
return 2;
}
}
}
}
else if (MVDiff >= 12) {
if (mcTcor < 99) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
if (RsCsDiff < 28) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 1;
}
}
}
else if (RsCsDiff >= 28) {
if (refDCval < 144) {
return 1;
}
else if (refDCval >= 144) {
return 2;
}
}
}
}
else if (mcTcor >= 99) {
if (Rs < 836) {
if (Rs < 768) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 1;
}
}
else if (Rs >= 768) {
return 4;
}
}
else if (Rs >= 836) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (ssDCval < 78) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 2;
}
}
else if (ssDCval >= 78) {
return 1;
}
}
}
}
}
}
else if (Cs >= 273) {
if (Cs < 449) {
if (refDCval < 95) {
if (Rs < 1035) {
if (refDCval < 58) {
if (Cs < 417) {
if (mu_mv_mag_sq < 60) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (Rs < 667) {
return 4;
}
else if (Rs >= 667) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 60) {
if (Rs < 778) {
return 8;
}
else if (Rs >= 778) {
return 2;
}
}
}
else if (Cs >= 417) {
if (Cs < 437) {
return 2;
}
else if (Cs >= 437) {
return 4;
}
}
}
else if (refDCval >= 58) {
if (diffMVdiffVal < -3) {
if (AFD < 15) {
if (posBalance < 58) {
if (Cs < 361) {
return 8;
}
else if (Cs >= 361) {
return 1;
}
}
else if (posBalance >= 58) {
return 4;
}
}
else if (AFD >= 15) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
}
else if (Rs >= 1035) {
if (negBalance < 26) {
if (Rs < 1933) {
return 4;
}
else if (Rs >= 1933) {
return 2;
}
}
else if (negBalance >= 26) {
if (SC < 2596) {
if (negBalance < 54) {
if (TSC < 7) {
if (mu_mv_mag_sq < 39) {
return 1;
}
else if (mu_mv_mag_sq >= 39) {
if (tcor < 84) {
if (Rs < 1280) {
if (Rs < 1244) {
return 2;
}
else if (Rs >= 1244) {
return 4;
}
}
else if (Rs >= 1280) {
return 1;
}
}
else if (tcor >= 84) {
return 1;
}
}
}
else if (TSC >= 7) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
if (Rs < 1306) {
return 2;
}
else if (Rs >= 1306) {
return 1;
}
}
}
}
}
else if (negBalance >= 54) {
if (Cs < 348) {
return 1;
}
else if (Cs >= 348) {
if (MVDiff < 29) {
if (Cs < 436) {
return 4;
}
else if (Cs >= 436) {
return 1;
}
}
else if (MVDiff >= 29) {
return 2;
}
}
}
}
else if (SC >= 2596) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
if (SC < 2751) {
return 1;
}
else if (SC >= 2751) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 95) {
if (ssDCval < 104) {
if (tcor < 93) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (Rs < 1468) {
return 2;
}
else if (Rs >= 1468) {
return 1;
}
}
}
else if (tcor >= 93) {
return 4;
}
}
else if (ssDCval >= 104) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (ssDCval < 115) {
if (tcor < 91) {
if (ssDCval < 106) {
if (Rs < 1058) {
return 1;
}
else if (Rs >= 1058) {
return 2;
}
}
else if (ssDCval >= 106) {
if (MVDiff < 50) {
return 8;
}
else if (MVDiff >= 50) {
return 1;
}
}
}
else if (tcor >= 91) {
return 1;
}
}
else if (ssDCval >= 115) {
return 1;
}
}
}
}
}
else if (Cs >= 449) {
if (refDCval < 73) {
if (RsCsDiff < 49) {
if (Rs < 1227) {
if (refDCval < 62) {
if (negBalance < 36) {
if (RsCsDiff < 33) {
if (posBalance < 47) {
if (Rs < 1001) {
return 2;
}
else if (Rs >= 1001) {
return 1;
}
}
else if (posBalance >= 47) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 1;
}
}
}
else if (RsCsDiff >= 33) {
return 2;
}
}
else if (negBalance >= 36) {
if (Cs < 486) {
if (Rs < 878) {
return 8;
}
else if (Rs >= 878) {
if (Rs < 1000) {
return 2;
}
else if (Rs >= 1000) {
return 4;
}
}
}
else if (Cs >= 486) {
return 8;
}
}
}
else if (refDCval >= 62) {
if (tcor < 88) {
if (RsCsDiff < 38) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
return 4;
}
}
else if (RsCsDiff >= 38) {
return 1;
}
}
else if (tcor >= 88) {
return 1;
}
}
}
else if (Rs >= 1227) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 20) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (Rs < 1998) {
return 2;
}
else if (Rs >= 1998) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 20) {
if (Rs < 1386) {
return 8;
}
else if (Rs >= 1386) {
return 4;
}
}
}
else if (MVDiff >= 8) {
if (diffMVdiffVal < 6) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
return 1;
}
}
else if (diffMVdiffVal >= 6) {
if (Cs < 484) {
return 1;
}
else if (Cs >= 484) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 49) {
if (AFD < 13) {
if (diffRsCsdiff < 32) {
if (mu_mv_mag_sq < 54) {
return 2;
}
else if (mu_mv_mag_sq >= 54) {
return 8;
}
}
else if (diffRsCsdiff >= 32) {
return 1;
}
}
else if (AFD >= 13) {
if (Rs < 870) {
return 8;
}
else if (Rs >= 870) {
if (Rs < 1028) {
return 1;
}
else if (Rs >= 1028) {
return 4;
}
}
}
}
}
else if (refDCval >= 73) {
if (ssDCval < 123) {
if (mcTcor < 94) {
if (posBalance < 57) {
if (negBalance < 57) {
return 1;
}
else if (negBalance >= 57) {
return 2;
}
}
else if (posBalance >= 57) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (Cs < 511) {
if (Cs < 451) {
if (Rs < 1688) {
return 1;
}
else if (Rs >= 1688) {
return 2;
}
}
else if (Cs >= 451) {
return 1;
}
}
else if (Cs >= 511) {
if (Rs < 898) {
if (Rs < 767) {
return 4;
}
else if (Rs >= 767) {
return 8;
}
}
else if (Rs >= 898) {
if (AFD < 17) {
if (Cs < 513) {
return 8;
}
else if (Cs >= 513) {
return 1;
}
}
else if (AFD >= 17) {
return 2;
}
}
}
}
}
else if (ssDCval >= 123) {
if (AFD < 12) {
if (diffTSC < -1) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
if (Cs < 495) {
return 1;
}
else if (Cs >= 495) {
return 2;
}
}
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (AFD >= 12) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 139) {
if (ssDCval < 146) {
if (gchDC < 2) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (SC < 1845) {
return 4;
}
else if (SC >= 1845) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (AFD < 28) {
if (Rs < 2173) {
return 1;
}
else if (Rs >= 2173) {
return 2;
}
}
else if (AFD >= 28) {
return 8;
}
}
}
else if (ssDCval >= 146) {
if (Cs < 500) {
if (negBalance < 67) {
if (tcor < 78) {
if (Rs < 1490) {
return 1;
}
else if (Rs >= 1490) {
return 8;
}
}
else if (tcor >= 78) {
return 1;
}
}
else if (negBalance >= 67) {
if (tcor < 90) {
return 2;
}
else if (tcor >= 90) {
return 1;
}
}
}
else if (Cs >= 500) {
if (Rs < 1842) {
return 8;
}
else if (Rs >= 1842) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 74) {
if (refDCval < 109) {
if (MVDiff < 39) {
if (negBalance < 43) {
if (refDCval < 104) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (ssDCval < 76) {
if (SC < 2254) {
return 1;
}
else if (SC >= 2254) {
return 2;
}
}
else if (ssDCval >= 76) {
if (Rs < 1806) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
else if (Rs >= 1806) {
return 1;
}
}
}
}
else if (refDCval >= 104) {
return 2;
}
}
else if (negBalance >= 43) {
if (Rs < 1836) {
if (posBalance < 64) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
if (Cs < 105) {
return 1;
}
else if (Cs >= 105) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
}
}
else if (posBalance >= 64) {
if (Cs < 518) {
return 1;
}
else if (Cs >= 518) {
return 2;
}
}
}
else if (Rs >= 1836) {
if (diffMVdiffVal < -25) {
return 1;
}
else if (diffMVdiffVal >= -25) {
if (AFD < 24) {
if (diffMVdiffVal < -16) {
if (Rs < 1959) {
return 1;
}
else if (Rs >= 1959) {
return 2;
}
}
else if (diffMVdiffVal >= -16) {
return 2;
}
}
else if (AFD >= 24) {
if (Cs < 132) {
return 2;
}
else if (Cs >= 132) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 39) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (MVDiff < 66) {
if (CsDiff < 5) {
if (RsCsDiff < 123) {
if (mu_mv_mag_sq < 25) {
return 1;
}
else if (mu_mv_mag_sq >= 25) {
if (mcTcor < 90) {
return 1;
}
else if (mcTcor >= 90) {
if (Rs < 1870) {
return 2;
}
else if (Rs >= 1870) {
return 8;
}
}
}
}
else if (RsCsDiff >= 123) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (mcTcor < 93) {
if (Rs < 1837) {
return 2;
}
else if (Rs >= 1837) {
return 8;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
}
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (MVDiff >= 66) {
if (Rs < 2112) {
if (RsCsDiff < 198) {
if (Cs < 200) {
return 1;
}
else if (Cs >= 200) {
if (diffRsCsdiff < -22) {
if (gchDC < 11) {
return 1;
}
else if (gchDC >= 11) {
if (Rs < 645) {
return 2;
}
else if (Rs >= 645) {
return 8;
}
}
}
else if (diffRsCsdiff >= -22) {
if (ssDCval < 89) {
if (AFD < 30) {
return 2;
}
else if (AFD >= 30) {
return 8;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
}
}
else if (RsCsDiff >= 198) {
return 2;
}
}
else if (Rs >= 2112) {
if (Rs < 2279) {
return 4;
}
else if (Rs >= 2279) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 109) {
if (AFD < 13) {
if (Rs < 1386) {
return 4;
}
else if (Rs >= 1386) {
if (negBalance < 41) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 1;
}
}
else if (negBalance >= 41) {
return 1;
}
}
}
else if (AFD >= 13) {
if (MVDiff < 61) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (AFD < 19) {
if (tcor < 82) {
if (Rs < 1352) {
return 1;
}
else if (Rs >= 1352) {
return 2;
}
}
else if (tcor >= 82) {
if (mu_mv_mag_sq < 35) {
return 1;
}
else if (mu_mv_mag_sq >= 35) {
if (MVDiff < 34) {
return 1;
}
else if (MVDiff >= 34) {
return 2;
}
}
}
}
else if (AFD >= 19) {
return 1;
}
}
}
else if (MVDiff >= 61) {
if (MVDiff < 116) {
if (AFD < 23) {
if (Cs < 353) {
return 1;
}
else if (Cs >= 353) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (diffAFD < 2) {
if (Rs < 2283) {
return 8;
}
else if (Rs >= 2283) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 8;
}
}
}
else if (diffAFD >= 2) {
return 1;
}
}
}
}
else if (AFD >= 23) {
return 1;
}
}
else if (MVDiff >= 116) {
if (Rs < 1227) {
if (Cs < 352) {
return 1;
}
else if (Cs >= 352) {
return 2;
}
}
else if (Rs >= 1227) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
return 1;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (negBalance < 53) {
if (posBalance < 51) {
if (Cs < 510) {
return 8;
}
else if (Cs >= 510) {
if (Rs < 1421) {
return 4;
}
else if (Rs >= 1421) {
return 1;
}
}
}
else if (posBalance >= 51) {
if (Rs < 732) {
return 1;
}
else if (Rs >= 732) {
if (mcTcor < 96) {
if (Rs < 1366) {
return 8;
}
else if (Rs >= 1366) {
return 1;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
}
}
else if (negBalance >= 53) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 534) {
if (negBalance < 14) {
if (Rs < 702) {
if (ssDCval < 57) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (ssDCval < 30) {
return 8;
}
else if (ssDCval >= 30) {
if (negBalance < 6) {
return 8;
}
else if (negBalance >= 6) {
if (ssDCval < 51) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (ssDCval >= 51) {
if (Cs < 1050) {
return 2;
}
else if (Cs >= 1050) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 57) {
if (negBalance < 10) {
if (Rs < 508) {
if (Cs < 578) {
if (Rs < 444) {
return 2;
}
else if (Rs >= 444) {
return 1;
}
}
else if (Cs >= 578) {
if (refDCval < 68) {
return 4;
}
else if (refDCval >= 68) {
return 8;
}
}
}
else if (Rs >= 508) {
return 8;
}
}
else if (negBalance >= 10) {
if (Rs < 652) {
if (SC < 1103) {
if (SC < 1053) {
return 2;
}
else if (SC >= 1053) {
return 1;
}
}
else if (SC >= 1103) {
if (Cs < 681) {
if (Rs < 553) {
return 1;
}
else if (Rs >= 553) {
return 8;
}
}
else if (Cs >= 681) {
if (Cs < 846) {
if (Rs < 503) {
return 4;
}
else if (Rs >= 503) {
return 2;
}
}
else if (Cs >= 846) {
return 8;
}
}
}
}
else if (Rs >= 652) {
if (TSCindex < 2) {
if (tcor < 100) {
return 1;
}
else if (tcor >= 100) {
return 8;
}
}
else if (TSCindex >= 2) {
return 8;
}
}
}
}
}
else if (Rs >= 702) {
if (diffTSC < -2) {
if (SC < 2469) {
if (diffAFD < -9) {
if (Cs < 578) {
return 2;
}
else if (Cs >= 578) {
if (refDCval < 133) {
if (Rs < 1315) {
if (Rs < 789) {
return 4;
}
else if (Rs >= 789) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
return 8;
}
}
}
else if (Rs >= 1315) {
if (Rs < 1415) {
return 1;
}
else if (Rs >= 1415) {
return 8;
}
}
}
else if (refDCval >= 133) {
if (ssDCval < 140) {
return 4;
}
else if (ssDCval >= 140) {
if (tcor < 88) {
if (Rs < 1342) {
return 8;
}
else if (Rs >= 1342) {
return 4;
}
}
else if (tcor >= 88) {
return 8;
}
}
}
}
}
else if (diffAFD >= -9) {
if (diffRsCsdiff < -2) {
if (MVDiff < 6) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 124) {
if (Rs < 1431) {
return 8;
}
else if (Rs >= 1431) {
return 1;
}
}
else if (mu_mv_mag_sq >= 124) {
if (Cs < 639) {
return 1;
}
else if (Cs >= 639) {
if (diffMVdiffVal < -15) {
if (Cs < 858) {
return 8;
}
else if (Cs >= 858) {
return 4;
}
}
else if (diffMVdiffVal >= -15) {
if (Cs < 1090) {
return 4;
}
else if (Cs >= 1090) {
return 8;
}
}
}
}
}
}
else if (diffRsCsdiff >= -2) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
}
}
else if (SC >= 2469) {
if (MVDiff < 16) {
if (Cs < 1290) {
return 8;
}
else if (Cs >= 1290) {
return 1;
}
}
else if (MVDiff >= 16) {
if (diffMVdiffVal < 9) {
if (diffMVdiffVal < -10) {
return 4;
}
else if (diffMVdiffVal >= -10) {
if (diffAFD < -10) {
return 8;
}
else if (diffAFD >= -10) {
return 4;
}
}
}
else if (diffMVdiffVal >= 9) {
if (Rs < 1656) {
return 1;
}
else if (Rs >= 1656) {
return 8;
}
}
}
}
}
else if (diffTSC >= -2) {
if (Rs < 1829) {
if (Rs < 967) {
if (diffMVdiffVal < -1) {
if (posBalance < 20) {
if (SC < 1569) {
return 8;
}
else if (SC >= 1569) {
return 2;
}
}
else if (posBalance >= 20) {
if (RsDiff < 3) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (SC < 1530) {
if (Rs < 754) {
if (Rs < 726) {
return 1;
}
else if (Rs >= 726) {
return 8;
}
}
else if (Rs >= 754) {
return 2;
}
}
else if (SC >= 1530) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 794) {
return 2;
}
else if (Rs >= 794) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 10) {
if (ssDCval < 147) {
return 8;
}
else if (ssDCval >= 147) {
return 2;
}
}
else if (posBalance >= 10) {
if (MVDiff < 4) {
if (ssDCval < 73) {
if (SC < 1491) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 2;
}
}
else if (SC >= 1491) {
return 8;
}
}
else if (ssDCval >= 73) {
if (Rs < 725) {
if (RsDiff < 1) {
if (Rs < 724) {
return 8;
}
else if (Rs >= 724) {
return 1;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (Rs >= 725) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (gchDC < 3) {
if (ssDCval < 31) {
return 8;
}
else if (ssDCval >= 31) {
if (Rs < 750) {
return 1;
}
else if (Rs >= 750) {
if (Rs < 808) {
return 2;
}
else if (Rs >= 808) {
return 1;
}
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
}
else if (Rs >= 967) {
if (SC < 1787) {
if (tcor < 101) {
return 8;
}
else if (tcor >= 101) {
return 4;
}
}
else if (SC >= 1787) {
if (negBalance < 9) {
if (diffMVdiffVal < -1) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (diffMVdiffVal < -10) {
if (Cs < 1038) {
return 4;
}
else if (Cs >= 1038) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (diffMVdiffVal >= -10) {
if (ssDCval < 137) {
if (posBalance < 12) {
return 1;
}
else if (posBalance >= 12) {
if (Rs < 1105) {
return 1;
}
else if (Rs >= 1105) {
return 8;
}
}
}
else if (ssDCval >= 137) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (tcor < 95) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (tcor < 90) {
if (Rs < 1204) {
return 2;
}
else if (Rs >= 1204) {
return 1;
}
}
else if (tcor >= 90) {
if (Rs < 1136) {
return 1;
}
else if (Rs >= 1136) {
return 8;
}
}
}
}
else if (tcor >= 95) {
return 8;
}
}
}
else if (negBalance >= 9) {
if (MVDiff < 3) {
if (posBalance < 28) {
if (diffMVdiffVal < -3) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (posBalance >= 28) {
if (ssDCval < 57) {
if (Cs < 614) {
return 1;
}
else if (Cs >= 614) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
if (Rs < 1127) {
return 1;
}
else if (Rs >= 1127) {
return 8;
}
}
}
}
else if (ssDCval >= 57) {
if (Cs < 992) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (posBalance < 38) {
if (SC < 2084) {
if (SC < 1870) {
return 8;
}
else if (SC >= 1870) {
return 1;
}
}
else if (SC >= 2084) {
return 8;
}
}
else if (posBalance >= 38) {
return 8;
}
}
}
else if (Cs >= 992) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (diffMVdiffVal < 3) {
if (MVDiff < 6) {
if (Rs < 1506) {
if (posBalance < 30) {
if (Cs < 877) {
return 8;
}
else if (Cs >= 877) {
return 1;
}
}
else if (posBalance >= 30) {
return 1;
}
}
else if (Rs >= 1506) {
return 8;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1226) {
if (Rs < 1098) {
return 8;
}
else if (Rs >= 1098) {
return 4;
}
}
else if (Rs >= 1226) {
if (Cs < 1113) {
return 8;
}
else if (Cs >= 1113) {
if (Rs < 1639) {
return 1;
}
else if (Rs >= 1639) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1829) {
if (Rs < 2020) {
if (refDCval < 79) {
if (SC < 2900) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
if (Cs < 654) {
return 2;
}
else if (Cs >= 654) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
if (SC < 2879) {
return 1;
}
else if (SC >= 2879) {
return 4;
}
}
}
}
}
}
else if (SC >= 2900) {
if (Rs < 1891) {
if (Cs < 1065) {
if (Rs < 1880) {
return 4;
}
else if (Rs >= 1880) {
return 1;
}
}
else if (Cs >= 1065) {
return 1;
}
}
else if (Rs >= 1891) {
if (Cs < 986) {
if (Rs < 2009) {
return 1;
}
else if (Rs >= 2009) {
return 4;
}
}
else if (Cs >= 986) {
return 4;
}
}
}
}
else if (refDCval >= 79) {
if (Cs < 1207) {
return 8;
}
else if (Cs >= 1207) {
return 1;
}
}
}
else if (Rs >= 2020) {
if (Cs < 1289) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (Rs < 2038) {
return 2;
}
else if (Rs >= 2038) {
return 1;
}
}
}
}
else if (Cs >= 1289) {
if (Rs < 2265) {
return 8;
}
else if (Rs >= 2265) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 44) {
if (refDCval < 47) {
if (MVDiff < 9) {
if (Rs < 1482) {
if (ssDCval < 29) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (ssDCval >= 29) {
if (CsDiff < 3) {
if (negBalance < 28) {
if (Cs < 813) {
if (Rs < 1134) {
if (Rs < 797) {
return 2;
}
else if (Rs >= 797) {
return 1;
}
}
else if (Rs >= 1134) {
return 4;
}
}
else if (Cs >= 813) {
return 8;
}
}
else if (negBalance >= 28) {
return 4;
}
}
else if (CsDiff >= 3) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 1;
}
}
}
}
else if (Rs >= 1482) {
if (AFD < 3) {
if (Rs < 1808) {
return 4;
}
else if (Rs >= 1808) {
return 8;
}
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (MVDiff >= 9) {
if (tcor < 61) {
return 8;
}
else if (tcor >= 61) {
if (negBalance < 31) {
if (diffRsCsdiff < 6) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (SC < 1605) {
return 2;
}
else if (SC >= 1605) {
return 1;
}
}
}
else if (diffRsCsdiff >= 6) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (Cs < 648) {
return 2;
}
else if (Cs >= 648) {
return 1;
}
}
}
}
else if (negBalance >= 31) {
if (Rs < 1753) {
if (posBalance < 43) {
if (SC < 2754) {
return 8;
}
else if (SC >= 2754) {
return 1;
}
}
else if (posBalance >= 43) {
return 2;
}
}
else if (Rs >= 1753) {
return 4;
}
}
}
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 87) {
if (negBalance < 62) {
if (Cs < 845) {
if (mcTcor < 92) {
if (Rs < 1279) {
if (RsCsDiff < 29) {
if (diffMVdiffVal < 3) {
if (Rs < 842) {
return 2;
}
else if (Rs >= 842) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 556) {
return 1;
}
else if (Rs >= 556) {
return 8;
}
}
}
else if (RsCsDiff >= 29) {
if (Rs < 444) {
return 2;
}
else if (Rs >= 444) {
return 8;
}
}
}
else if (Rs >= 1279) {
if (AFD < 11) {
if (diffAFD < 0) {
if (Rs < 1710) {
return 2;
}
else if (Rs >= 1710) {
return 1;
}
}
else if (diffAFD >= 0) {
return 1;
}
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (mcTcor >= 92) {
if (Rs < 812) {
if (tcor < 91) {
if (MVDiff < 23) {
if (Cs < 624) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (Cs >= 624) {
if (ssDCval < 48) {
return 8;
}
else if (ssDCval >= 48) {
return 2;
}
}
}
else if (MVDiff >= 23) {
if (Cs < 689) {
return 1;
}
else if (Cs >= 689) {
return 2;
}
}
}
else if (tcor >= 91) {
if (diffRsCsdiff < -2) {
if (Rs < 704) {
return 2;
}
else if (Rs >= 704) {
return 4;
}
}
else if (diffRsCsdiff >= -2) {
return 4;
}
}
}
else if (Rs >= 812) {
if (mu_mv_mag_sq < 72) {
if (MVDiff < 13) {
if (MVDiff < 4) {
if (AFD < 8) {
if (mu_mv_mag_sq < 65) {
return 1;
}
else if (mu_mv_mag_sq >= 65) {
if (Rs < 1758) {
return 8;
}
else if (Rs >= 1758) {
return 1;
}
}
}
else if (AFD >= 8) {
return 8;
}
}
else if (MVDiff >= 4) {
if (TSC < 7) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (MVDiff >= 13) {
if (ssDCval < 69) {
if (Rs < 860) {
return 1;
}
else if (Rs >= 860) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 2;
}
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
if (negBalance < 33) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 1375) {
return 1;
}
else if (Rs >= 1375) {
if (Rs < 1925) {
return 2;
}
else if (Rs >= 1925) {
return 4;
}
}
}
}
else if (negBalance >= 33) {
if (Rs < 2071) {
return 1;
}
else if (Rs >= 2071) {
return 4;
}
}
}
}
}
}
}
}
else if (Cs >= 845) {
if (mu_mv_mag_sq < 66) {
if (mcTcor < 88) {
if (Cs < 1177) {
return 8;
}
else if (Cs >= 1177) {
return 1;
}
}
else if (mcTcor >= 88) {
if (Cs < 1104) {
if (SC < 2938) {
if (Rs < 989) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
return 2;
}
}
else if (Rs >= 989) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
if (ssDCval < 62) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
if (Rs < 1342) {
return 8;
}
else if (Rs >= 1342) {
return 4;
}
}
}
else if (ssDCval >= 62) {
if (mcTcor < 96) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 1577) {
return 8;
}
else if (Rs >= 1577) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
}
else if (SC >= 2938) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (SC < 3166) {
return 4;
}
else if (SC >= 3166) {
return 2;
}
}
}
}
else if (Cs >= 1104) {
if (tcor < 92) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 2;
}
}
else if (tcor >= 92) {
if (Rs < 1301) {
return 4;
}
else if (Rs >= 1301) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (negBalance < 28) {
if (Cs < 1278) {
if (RsDiff < 2) {
if (Rs < 860) {
return 4;
}
else if (Rs >= 860) {
return 1;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (Cs >= 1278) {
return 2;
}
}
else if (negBalance >= 28) {
if (Cs < 947) {
if (tcor < 94) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
else if (tcor >= 94) {
return 1;
}
}
else if (Cs >= 947) {
if (ssDCval < 69) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
return 4;
}
}
else if (ssDCval >= 69) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 62) {
if (tcor < 80) {
if (Rs < 741) {
return 8;
}
else if (Rs >= 741) {
return 2;
}
}
else if (tcor >= 80) {
if (tcor < 90) {
if (ssDCval < 93) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
return 8;
}
}
else if (ssDCval >= 93) {
if (Rs < 576) {
return 8;
}
else if (Rs >= 576) {
return 2;
}
}
}
else if (tcor >= 90) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (ssDCval < 56) {
if (TSC < 3) {
if (Rs < 613) {
if (Rs < 442) {
return 2;
}
else if (Rs >= 442) {
return 1;
}
}
else if (Rs >= 613) {
if (negBalance < 23) {
if (Cs < 1004) {
return 8;
}
else if (Cs >= 1004) {
return 4;
}
}
else if (negBalance >= 23) {
return 4;
}
}
}
else if (TSC >= 3) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 102) {
if (Cs < 1149) {
return 1;
}
else if (Cs >= 1149) {
return 8;
}
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 451) {
return 1;
}
else if (Rs >= 451) {
if (Rs < 2017) {
return 8;
}
else if (Rs >= 2017) {
if (Rs < 2055) {
return 8;
}
else if (Rs >= 2055) {
return 8;
}
}
}
}
}
else if (mcTcor >= 95) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
if (Rs < 1024) {
return 8;
}
else if (Rs >= 1024) {
return 4;
}
}
}
}
}
else if (ssDCval >= 56) {
if (Rs < 734) {
if (Cs < 678) {
if (Rs < 543) {
return 2;
}
else if (Rs >= 543) {
return 8;
}
}
else if (Cs >= 678) {
if (ssDCval < 109) {
if (Cs < 913) {
if (Rs < 654) {
return 2;
}
else if (Rs >= 654) {
return 4;
}
}
else if (Cs >= 913) {
return 8;
}
}
else if (ssDCval >= 109) {
return 4;
}
}
}
else if (Rs >= 734) {
if (CsDiff < 3) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (ssDCval < 60) {
if (Rs < 1089) {
return 8;
}
else if (Rs >= 1089) {
return 1;
}
}
else if (ssDCval >= 60) {
if (Cs < 917) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 8;
}
}
else if (Cs >= 917) {
return 8;
}
}
}
}
else if (CsDiff >= 3) {
if (TSC < 4) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
else if (TSC >= 4) {
if (ssDCval < 89) {
if (AFD < 6) {
if (negBalance < 30) {
if (Cs < 1060) {
return 2;
}
else if (Cs >= 1060) {
return 4;
}
}
else if (negBalance >= 30) {
if (Rs < 1600) {
return 8;
}
else if (Rs >= 1600) {
return 1;
}
}
}
else if (AFD >= 6) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 4;
}
}
}
else if (ssDCval >= 89) {
return 8;
}
}
}
}
}
}
}
}
else if (posBalance >= 44) {
if (Rs < 400) {
if (ssDCval < 96) {
if (ssDCval < 82) {
if (Rs < 350) {
if (MVDiff < 3) {
if (Cs < 806) {
return 8;
}
else if (Cs >= 806) {
if (Rs < 328) {
return 2;
}
else if (Rs >= 328) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 25) {
return 8;
}
else if (ssDCval >= 25) {
if (CsDiff < 7) {
return 1;
}
else if (CsDiff >= 7) {
if (Rs < 308) {
return 2;
}
else if (Rs >= 308) {
return 1;
}
}
}
}
}
else if (Rs >= 350) {
if (mu_mv_mag_sq < 105) {
if (AFD < 10) {
if (ssDCval < 59) {
if (posBalance < 48) {
if (Rs < 385) {
return 8;
}
else if (Rs >= 385) {
return 4;
}
}
else if (posBalance >= 48) {
return 2;
}
}
else if (ssDCval >= 59) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (AFD >= 10) {
return 1;
}
}
else if (mu_mv_mag_sq >= 105) {
return 1;
}
}
}
else if (ssDCval >= 82) {
if (MVDiff < 4) {
if (Cs < 833) {
return 1;
}
else if (Cs >= 833) {
return 8;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (ssDCval >= 96) {
if (Cs < 734) {
if (TSC < 5) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 2;
}
}
else if (TSC >= 5) {
return 1;
}
}
else if (Cs >= 734) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (Rs >= 400) {
if (mu_mv_mag_sq < 39) {
if (negBalance < 44) {
if (ssDCval < 60) {
if (Rs < 808) {
if (tcor < 90) {
if (AFD < 8) {
if (ssDCval < 37) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
return 2;
}
}
else if (ssDCval >= 37) {
if (ssDCval < 41) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
else if (ssDCval >= 41) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
return 8;
}
}
}
}
else if (AFD >= 8) {
if (negBalance < 32) {
if (Rs < 766) {
return 1;
}
else if (Rs >= 766) {
return 4;
}
}
else if (negBalance >= 32) {
if (Cs < 589) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (Cs >= 589) {
if (negBalance < 34) {
if (Rs < 646) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (Rs >= 646) {
if (Rs < 682) {
return 4;
}
else if (Rs >= 682) {
return 2;
}
}
}
else if (negBalance >= 34) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (Cs < 827) {
if (Rs < 586) {
return 2;
}
else if (Rs >= 586) {
if (Rs < 695) {
return 8;
}
else if (Rs >= 695) {
return 2;
}
}
}
else if (Cs >= 827) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 90) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
}
}
else if (Rs >= 808) {
if (Cs < 957) {
if (AFD < 11) {
if (mcTcor < 95) {
if (SC < 1723) {
if (negBalance < 37) {
return 8;
}
else if (negBalance >= 37) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (RsCsDiff < 27) {
return 8;
}
else if (RsCsDiff >= 27) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (SC >= 1723) {
if (Cs < 696) {
return 4;
}
else if (Cs >= 696) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (Rs < 1062) {
return 4;
}
else if (Rs >= 1062) {
return 1;
}
}
}
}
}
}
else if (AFD >= 11) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (Rs < 2002) {
return 2;
}
else if (Rs >= 2002) {
return 4;
}
}
}
}
else if (Cs >= 957) {
if (SC < 2098) {
if (AFD < 9) {
if (AFD < 7) {
if (Rs < 923) {
return 2;
}
else if (Rs >= 923) {
return 8;
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (AFD >= 9) {
return 8;
}
}
else if (SC >= 2098) {
if (Cs < 1276) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1276) {
return 4;
}
}
}
}
}
else if (ssDCval >= 60) {
if (negBalance < 35) {
if (refDCval < 137) {
if (RsDiff < 5) {
if (posBalance < 63) {
if (ssDCval < 72) {
if (Rs < 940) {
return 4;
}
else if (Rs >= 940) {
if (Rs < 1238) {
return 8;
}
else if (Rs >= 1238) {
return 1;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 623) {
return 2;
}
else if (Rs >= 623) {
if (Rs < 1963) {
return 1;
}
else if (Rs >= 1963) {
return 8;
}
}
}
}
else if (posBalance >= 63) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 612) {
return 8;
}
else if (Rs >= 612) {
return 4;
}
}
}
}
}
else if (RsDiff >= 5) {
if (SC < 2456) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (SC >= 2456) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (diffMVdiffVal < 5) {
if (gchDC < 3) {
if (mu_mv_mag_sq < 30) {
return 2;
}
else if (mu_mv_mag_sq >= 30) {
return 1;
}
}
else if (gchDC >= 3) {
return 1;
}
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
}
}
}
else if (refDCval >= 137) {
if (diffAFD < 2) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 2;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (negBalance >= 35) {
if (refDCval < 88) {
if (Rs < 1792) {
if (Rs < 1137) {
if (Cs < 793) {
if (ssDCval < 64) {
if (MVDiff < 38) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (MVDiff >= 38) {
return 1;
}
}
else if (ssDCval >= 64) {
if (SC < 1483) {
if (Cs < 585) {
return 4;
}
else if (Cs >= 585) {
return 2;
}
}
else if (SC >= 1483) {
if (SC < 1632) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (AFD < 8) {
if (Rs < 878) {
return 1;
}
else if (Rs >= 878) {
return 4;
}
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (SC >= 1632) {
return 4;
}
}
}
}
else if (Cs >= 793) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (tcor < 88) {
if (Cs < 1181) {
return 8;
}
else if (Cs >= 1181) {
return 4;
}
}
else if (tcor >= 88) {
if (Rs < 1013) {
return 2;
}
else if (Rs >= 1013) {
return 4;
}
}
}
}
}
else if (Rs >= 1137) {
if (mcTcor < 94) {
if (Cs < 912) {
if (Cs < 623) {
return 8;
}
else if (Cs >= 623) {
return 4;
}
}
else if (Cs >= 912) {
if (Rs < 1332) {
return 1;
}
else if (Rs >= 1332) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (Cs < 1235) {
if (ssDCval < 65) {
if (Rs < 1215) {
return 4;
}
else if (Rs >= 1215) {
if (SC < 2117) {
return 8;
}
else if (SC >= 2117) {
return 1;
}
}
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (Cs >= 1235) {
return 4;
}
}
}
}
else if (Rs >= 1792) {
if (ssDCval < 74) {
if (posBalance < 48) {
if (Rs < 2159) {
return 4;
}
else if (Rs >= 2159) {
return 8;
}
}
else if (posBalance >= 48) {
return 2;
}
}
else if (ssDCval >= 74) {
if (Rs < 2152) {
return 1;
}
else if (Rs >= 2152) {
return 4;
}
}
}
}
else if (refDCval >= 88) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (MVDiff < 25) {
if (diffRsCsdiff < 1) {
if (RsCsDiff < 29) {
if (ssDCval < 144) {
return 8;
}
else if (ssDCval >= 144) {
if (Rs < 774) {
return 8;
}
else if (Rs >= 774) {
return 2;
}
}
}
else if (RsCsDiff >= 29) {
if (MVDiff < 15) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
return 2;
}
}
else if (MVDiff >= 15) {
if (ssDCval < 153) {
return 8;
}
else if (ssDCval >= 153) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 1) {
if (MVDiff < 11) {
if (MVDiff < 10) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
else if (MVDiff >= 10) {
return 2;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
}
else if (MVDiff >= 25) {
if (TSC < 6) {
if (diffRsCsdiff < -19) {
return 8;
}
else if (diffRsCsdiff >= -19) {
return 4;
}
}
else if (TSC >= 6) {
if (Rs < 1756) {
return 8;
}
else if (Rs >= 1756) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 44) {
if (RsCsDiff < 73) {
if (refDCval < 137) {
if (Rs < 1001) {
if (CsDiff < 5) {
if (MVDiff < 6) {
if (posBalance < 59) {
if (ssDCval < 89) {
if (Cs < 640) {
if (ssDCval < 64) {
if (Rs < 721) {
return 4;
}
else if (Rs >= 721) {
return 1;
}
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (Cs >= 640) {
if (diffRsCsdiff < -4) {
if (Cs < 977) {
if (SC < 1504) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (SC >= 1504) {
return 8;
}
}
else if (Cs >= 977) {
return 4;
}
}
else if (diffRsCsdiff >= -4) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
}
}
else if (ssDCval >= 89) {
if (MVDiff < 5) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (ssDCval < 110) {
return 1;
}
else if (ssDCval >= 110) {
return 2;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (posBalance >= 59) {
if (ssDCval < 60) {
if (ssDCval < 49) {
if (Cs < 585) {
return 8;
}
else if (Cs >= 585) {
return 1;
}
}
else if (ssDCval >= 49) {
return 4;
}
}
else if (ssDCval >= 60) {
if (ssDCval < 100) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
return 1;
}
}
else if (ssDCval >= 100) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (AFD < 11) {
if (Rs < 826) {
if (ssDCval < 44) {
if (Rs < 602) {
if (ssDCval < 41) {
return 2;
}
else if (ssDCval >= 41) {
return 4;
}
}
else if (Rs >= 602) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
}
else if (ssDCval >= 44) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
if (MVDiff < 13) {
return 1;
}
else if (MVDiff >= 13) {
if (AFD < 8) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (AFD >= 8) {
return 8;
}
}
}
}
}
}
else if (Rs >= 826) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 54) {
if (SC < 1670) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
}
else if (SC >= 1670) {
return 8;
}
}
else if (ssDCval >= 54) {
if (RsCsDiff < 22) {
if (Cs < 614) {
return 1;
}
else if (Cs >= 614) {
return 8;
}
}
else if (RsCsDiff >= 22) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (tcor < 91) {
if (SC < 1456) {
return 2;
}
else if (SC >= 1456) {
return 8;
}
}
else if (tcor >= 91) {
return 4;
}
}
}
}
else if (AFD >= 11) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (ssDCval < 106) {
if (negBalance < 52) {
if (posBalance < 54) {
if (Rs < 753) {
if (Rs < 594) {
return 2;
}
else if (Rs >= 594) {
return 1;
}
}
else if (Rs >= 753) {
return 4;
}
}
else if (posBalance >= 54) {
if (RsCsDiff < 38) {
if (posBalance < 63) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 8;
}
}
else if (posBalance >= 63) {
if (Rs < 515) {
return 4;
}
else if (Rs >= 515) {
return 8;
}
}
}
else if (RsCsDiff >= 38) {
return 8;
}
}
}
else if (negBalance >= 52) {
if (tcor < 59) {
if (Rs < 917) {
return 4;
}
else if (Rs >= 917) {
return 8;
}
}
else if (tcor >= 59) {
if (posBalance < 63) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 941) {
return 8;
}
else if (Rs >= 941) {
return 2;
}
}
}
else if (posBalance >= 63) {
if (Rs < 790) {
return 8;
}
else if (Rs >= 790) {
return 4;
}
}
}
}
}
else if (ssDCval >= 106) {
return 4;
}
}
}
}
}
else if (CsDiff >= 5) {
if (negBalance < 49) {
if (ssDCval < 48) {
return 2;
}
else if (ssDCval >= 48) {
if (Rs < 833) {
return 1;
}
else if (Rs >= 833) {
if (Rs < 966) {
return 8;
}
else if (Rs >= 966) {
return 4;
}
}
}
}
else if (negBalance >= 49) {
if (diffRsCsdiff < 10) {
if (diffMVdiffVal < 8) {
if (gchDC < 1) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
if (Rs < 896) {
return 4;
}
else if (Rs >= 896) {
return 1;
}
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (diffMVdiffVal >= 8) {
return 8;
}
}
else if (diffRsCsdiff >= 10) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1001) {
if (posBalance < 55) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (AFD < 11) {
if (refDCval < 91) {
if (MVDiff < 7) {
if (mcTcor < 96) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (SC < 2938) {
return 2;
}
else if (SC >= 2938) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1589) {
return 8;
}
else if (Rs >= 1589) {
return 1;
}
}
}
else if (MVDiff >= 7) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (diffAFD < -1) {
if (diffRsCsdiff < -19) {
return 4;
}
else if (diffRsCsdiff >= -19) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (Rs < 1816) {
return 2;
}
else if (Rs >= 1816) {
return 4;
}
}
}
}
else if (diffAFD >= -1) {
if (Cs < 604) {
return 1;
}
else if (Cs >= 604) {
return 4;
}
}
}
}
}
else if (refDCval >= 91) {
if (CsDiff < 3) {
if (Rs < 1570) {
if (Rs < 1255) {
return 1;
}
else if (Rs >= 1255) {
return 8;
}
}
else if (Rs >= 1570) {
return 2;
}
}
else if (CsDiff >= 3) {
if (MVDiff < 9) {
if (Cs < 1027) {
return 1;
}
else if (Cs >= 1027) {
return 8;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
}
}
else if (AFD >= 11) {
if (CsDiff < 5) {
if (TSC < 9) {
if (diffMVdiffVal < -14) {
return 2;
}
else if (diffMVdiffVal >= -14) {
if (Rs < 1960) {
if (Rs < 1300) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (Rs >= 1300) {
if (ssDCval < 106) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (Cs < 723) {
return 1;
}
else if (Cs >= 723) {
return 8;
}
}
}
else if (ssDCval >= 106) {
return 8;
}
}
}
else if (Rs >= 1960) {
return 8;
}
}
}
else if (TSC >= 9) {
if (Cs < 975) {
return 1;
}
else if (Cs >= 975) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 1316) {
if (Cs < 1065) {
if (Rs < 1195) {
return 1;
}
else if (Rs >= 1195) {
return 8;
}
}
else if (Cs >= 1065) {
return 4;
}
}
else if (Rs >= 1316) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 72) {
if (Cs < 1227) {
if (Cs < 565) {
if (Rs < 1081) {
return 1;
}
else if (Rs >= 1081) {
return 4;
}
}
else if (Cs >= 565) {
if (tcor < 86) {
if (Cs < 976) {
return 8;
}
else if (Cs >= 976) {
if (Rs < 1181) {
if (Cs < 1113) {
return 1;
}
else if (Cs >= 1113) {
return 8;
}
}
else if (Rs >= 1181) {
return 8;
}
}
}
else if (tcor >= 86) {
if (tcor < 88) {
if (negBalance < 50) {
if (MVDiff < 14) {
if (Cs < 773) {
return 8;
}
else if (Cs >= 773) {
return 4;
}
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (tcor >= 88) {
return 8;
}
}
}
}
else if (Cs >= 1227) {
if (Rs < 1331) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
else if (Rs >= 1331) {
return 8;
}
}
}
else if (ssDCval >= 72) {
if (MVDiff < 17) {
if (tcor < 90) {
if (refDCval < 99) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (ssDCval < 73) {
if (Cs < 715) {
return 2;
}
else if (Cs >= 715) {
return 8;
}
}
else if (ssDCval >= 73) {
if (posBalance < 59) {
if (TSC < 7) {
if (Cs < 626) {
return 8;
}
else if (Cs >= 626) {
if (diffTSC < -1) {
if (Rs < 1233) {
return 2;
}
else if (Rs >= 1233) {
return 8;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
}
else if (TSC >= 7) {
if (refDCval < 98) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Cs < 998) {
return 8;
}
else if (Cs >= 998) {
return 4;
}
}
}
else if (refDCval >= 98) {
return 2;
}
}
}
else if (posBalance >= 59) {
if (Cs < 577) {
if (Rs < 1098) {
return 8;
}
else if (Rs >= 1098) {
return 2;
}
}
else if (Cs >= 577) {
return 8;
}
}
}
}
}
else if (refDCval >= 99) {
if (Rs < 1044) {
if (Rs < 1032) {
return 8;
}
else if (Rs >= 1032) {
return 4;
}
}
else if (Rs >= 1044) {
return 8;
}
}
}
else if (tcor >= 90) {
if (MVDiff < 6) {
if (RsDiff < 5) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
return 2;
}
}
else if (RsDiff >= 5) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 1542) {
return 2;
}
else if (Rs >= 1542) {
return 8;
}
}
}
}
}
else if (MVDiff >= 17) {
if (tcor < 68) {
return 1;
}
else if (tcor >= 68) {
if (mu_mv_mag_sq < 2) {
if (SC < 1834) {
return 1;
}
else if (SC >= 1834) {
if (RsCsDiff < 63) {
return 8;
}
else if (RsCsDiff >= 63) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (TSC < 7) {
if (ssDCval < 105) {
if (MVDiff < 24) {
return 2;
}
else if (MVDiff >= 24) {
if (Rs < 1822) {
return 8;
}
else if (Rs >= 1822) {
return 2;
}
}
}
else if (ssDCval >= 105) {
return 8;
}
}
else if (TSC >= 7) {
if (SC < 2178) {
return 4;
}
else if (SC >= 2178) {
if (Rs < 1602) {
return 8;
}
else if (Rs >= 1602) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 137) {
if (ssDCval < 142) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
return 4;
}
}
else if (ssDCval >= 142) {
if (CsDiff < 4) {
if (Cs < 985) {
return 2;
}
else if (Cs >= 985) {
return 8;
}
}
else if (CsDiff >= 4) {
if (Rs < 1294) {
return 1;
}
else if (Rs >= 1294) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 73) {
if (Cs < 828) {
if (SC < 2416) {
if (TSC < 8) {
if (Cs < 598) {
return 1;
}
else if (Cs >= 598) {
if (Cs < 762) {
return 8;
}
else if (Cs >= 762) {
return 4;
}
}
}
else if (TSC >= 8) {
if (Rs < 1504) {
if (Cs < 669) {
return 2;
}
else if (Cs >= 669) {
return 1;
}
}
else if (Rs >= 1504) {
return 4;
}
}
}
else if (SC >= 2416) {
if (mcTcor < 90) {
if (Rs < 1816) {
return 1;
}
else if (Rs >= 1816) {
return 4;
}
}
else if (mcTcor >= 90) {
if (Cs < 811) {
return 2;
}
else if (Cs >= 811) {
return 8;
}
}
}
}
else if (Cs >= 828) {
if (posBalance < 51) {
if (ssDCval < 95) {
if (diffTSC < -1) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (diffTSC >= -1) {
if (diffRsCsdiff < 38) {
return 4;
}
else if (diffRsCsdiff >= 38) {
if (Rs < 1086) {
return 4;
}
else if (Rs >= 1086) {
return 1;
}
}
}
}
else if (ssDCval >= 95) {
return 2;
}
}
else if (posBalance >= 51) {
if (Rs < 2194) {
if (Rs < 1035) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (Rs >= 1035) {
if (gchDC < 2) {
if (diffAFD < 2) {
if (posBalance < 52) {
if (Rs < 1610) {
return 2;
}
else if (Rs >= 1610) {
return 8;
}
}
else if (posBalance >= 52) {
return 8;
}
}
else if (diffAFD >= 2) {
if (Rs < 1698) {
return 2;
}
else if (Rs >= 1698) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (Rs >= 2194) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (ssDCval < 158) {
if (mcTcor < 89) {
if (mu_mv_mag_sq < 117) {
if (refDCval < 51) {
if (diffAFD < 3) {
if (Cs < 572) {
if (Cs < 566) {
if (Cs < 535) {
return 4;
}
else if (Cs >= 535) {
return 8;
}
}
else if (Cs >= 566) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
}
else if (Cs >= 572) {
if (Rs < 630) {
if (tcor < 79) {
if (Rs < 463) {
return 1;
}
else if (Rs >= 463) {
if (Rs < 537) {
return 2;
}
else if (Rs >= 537) {
return 4;
}
}
}
else if (tcor >= 79) {
return 8;
}
}
else if (Rs >= 630) {
if (RsDiff < 3) {
if (Rs < 811) {
return 8;
}
else if (Rs >= 811) {
return 1;
}
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 103) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (tcor < 82) {
if (diffMVdiffVal < -7) {
if (Rs < 997) {
return 2;
}
else if (Rs >= 997) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
else if (tcor >= 82) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (SC < 1811) {
return 1;
}
else if (SC >= 1811) {
return 8;
}
}
}
}
}
}
else if (diffAFD >= 3) {
if (Rs < 1268) {
if (Rs < 938) {
return 8;
}
else if (Rs >= 938) {
return 1;
}
}
else if (Rs >= 1268) {
return 2;
}
}
}
else if (refDCval >= 51) {
if (refDCval < 63) {
if (Rs < 1040) {
if (Rs < 704) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (Rs >= 704) {
if (posBalance < 54) {
if (Cs < 654) {
return 8;
}
else if (Cs >= 654) {
if (SC < 1680) {
if (Rs < 842) {
return 1;
}
else if (Rs >= 842) {
return 2;
}
}
else if (SC >= 1680) {
return 4;
}
}
}
else if (posBalance >= 54) {
return 4;
}
}
}
else if (Rs >= 1040) {
if (Cs < 1168) {
if (mu_mv_mag_sq < 80) {
if (posBalance < 70) {
return 8;
}
else if (posBalance >= 70) {
if (Rs < 1327) {
return 4;
}
else if (Rs >= 1327) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (RsCsDiff < 63) {
if (diffMVdiffVal < 1) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 1089) {
return 8;
}
else if (Rs >= 1089) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (RsCsDiff >= 63) {
return 2;
}
}
}
else if (Cs >= 1168) {
if (Rs < 1595) {
if (Rs < 1172) {
return 8;
}
else if (Rs >= 1172) {
return 1;
}
}
else if (Rs >= 1595) {
return 8;
}
}
}
}
else if (refDCval >= 63) {
if (tcor < 73) {
if (TSC < 7) {
if (Rs < 1655) {
return 1;
}
else if (Rs >= 1655) {
return 2;
}
}
else if (TSC >= 7) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (ssDCval < 100) {
return 8;
}
else if (ssDCval >= 100) {
return 4;
}
}
}
}
else if (tcor >= 73) {
if (negBalance < 44) {
if (ssDCval < 73) {
if (Cs < 920) {
return 8;
}
else if (Cs >= 920) {
return 4;
}
}
else if (ssDCval >= 73) {
if (refDCval < 129) {
if (Rs < 2034) {
return 8;
}
else if (Rs >= 2034) {
return 1;
}
}
else if (refDCval >= 129) {
if (Rs < 1218) {
return 8;
}
else if (Rs >= 1218) {
return 4;
}
}
}
}
else if (negBalance >= 44) {
if (Cs < 1231) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
return 8;
}
}
else if (Cs >= 1231) {
if (Cs < 1233) {
return 2;
}
else if (Cs >= 1233) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (posBalance < 55) {
if (ssDCval < 48) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 1330) {
return 1;
}
else if (Rs >= 1330) {
return 8;
}
}
}
else if (ssDCval >= 48) {
return 8;
}
}
else if (posBalance >= 55) {
if (negBalance < 39) {
if (tcor < 86) {
if (Rs < 1361) {
return 1;
}
else if (Rs >= 1361) {
if (Rs < 1612) {
return 4;
}
else if (Rs >= 1612) {
return 2;
}
}
}
else if (tcor >= 86) {
if (Rs < 928) {
return 1;
}
else if (Rs >= 928) {
return 8;
}
}
}
else if (negBalance >= 39) {
return 8;
}
}
}
}
else if (mcTcor >= 89) {
if (refDCval < 78) {
if (RsCsDiff < 28) {
if (Rs < 870) {
if (ssDCval < 56) {
if (refDCval < 37) {
if (Cs < 598) {
return 8;
}
else if (Cs >= 598) {
return 1;
}
}
else if (refDCval >= 37) {
if (Rs < 665) {
if (SC < 1311) {
if (tcor < 90) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (tcor >= 90) {
return 4;
}
}
else if (SC >= 1311) {
if (SC < 1685) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
return 4;
}
}
else if (SC >= 1685) {
return 4;
}
}
}
else if (Rs >= 665) {
if (Cs < 912) {
if (Rs < 757) {
if (AFD < 4) {
if (Rs < 723) {
return 8;
}
else if (Rs >= 723) {
return 4;
}
}
else if (AFD >= 4) {
return 4;
}
}
else if (Rs >= 757) {
if (posBalance < 52) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (posBalance >= 52) {
if (RsCsDiff < 16) {
if (diffTSC < -1) {
if (Rs < 771) {
return 8;
}
else if (Rs >= 771) {
return 4;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (RsCsDiff >= 16) {
return 4;
}
}
}
}
else if (Cs >= 912) {
return 8;
}
}
}
}
else if (ssDCval >= 56) {
if (AFD < 5) {
if (ssDCval < 67) {
if (posBalance < 53) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (Rs < 675) {
return 2;
}
else if (Rs >= 675) {
return 8;
}
}
}
else if (posBalance >= 53) {
if (mu_mv_mag_sq < 98) {
if (Rs < 504) {
return 1;
}
else if (Rs >= 504) {
return 8;
}
}
else if (mu_mv_mag_sq >= 98) {
if (mu_mv_mag_sq < 122) {
return 4;
}
else if (mu_mv_mag_sq >= 122) {
return 8;
}
}
}
}
else if (ssDCval >= 67) {
if (SC < 1644) {
if (MVDiff < 1) {
if (Rs < 599) {
return 2;
}
else if (Rs >= 599) {
return 1;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (SC >= 1644) {
return 2;
}
}
}
else if (AFD >= 5) {
if (Rs < 862) {
if (negBalance < 38) {
if (mu_mv_mag_sq < 87) {
if (CsDiff < 2) {
if (Rs < 658) {
return 8;
}
else if (Rs >= 658) {
return 4;
}
}
else if (CsDiff >= 2) {
if (diffMVdiffVal < 5) {
if (MVDiff < 8) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (Rs < 517) {
return 1;
}
else if (Rs >= 517) {
return 2;
}
}
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (diffMVdiffVal >= 5) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (Rs < 501) {
return 1;
}
else if (Rs >= 501) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
}
}
else if (negBalance >= 38) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (mu_mv_mag_sq < 81) {
if (Cs < 788) {
return 8;
}
else if (Cs >= 788) {
return 1;
}
}
else if (mu_mv_mag_sq >= 81) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
return 2;
}
}
}
}
}
else if (Rs >= 862) {
if (Cs < 1045) {
return 2;
}
else if (Cs >= 1045) {
return 4;
}
}
}
}
}
else if (Rs >= 870) {
if (refDCval < 60) {
if (RsCsDiff < 9) {
if (negBalance < 16) {
return 1;
}
else if (negBalance >= 16) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (ssDCval < 60) {
if (mu_mv_mag_sq < 66) {
if (negBalance < 35) {
if (refDCval < 56) {
return 1;
}
else if (refDCval >= 56) {
return 8;
}
}
else if (negBalance >= 35) {
return 8;
}
}
else if (mu_mv_mag_sq >= 66) {
return 8;
}
}
else if (ssDCval >= 60) {
if (Rs < 1266) {
return 1;
}
else if (Rs >= 1266) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 9) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 68) {
if (negBalance < 38) {
if (Cs < 913) {
return 1;
}
else if (Cs >= 913) {
if (Rs < 1185) {
return 4;
}
else if (Rs >= 1185) {
return 2;
}
}
}
else if (negBalance >= 38) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (negBalance < 30) {
if (Cs < 1177) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (Cs >= 1177) {
if (Rs < 1584) {
return 1;
}
else if (Rs >= 1584) {
return 8;
}
}
}
else if (negBalance >= 30) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
return 4;
}
}
}
}
else if (posBalance >= 49) {
if (Cs < 899) {
if (negBalance < 42) {
if (MVDiff < 2) {
if (tcor < 87) {
return 1;
}
else if (tcor >= 87) {
if (ssDCval < 46) {
if (Rs < 1170) {
return 8;
}
else if (Rs >= 1170) {
return 1;
}
}
else if (ssDCval >= 46) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (Cs < 755) {
if (Rs < 1423) {
return 2;
}
else if (Rs >= 1423) {
return 1;
}
}
else if (Cs >= 755) {
if (Rs < 1262) {
return 8;
}
else if (Rs >= 1262) {
return 2;
}
}
}
}
else if (negBalance >= 42) {
return 8;
}
}
else if (Cs >= 899) {
if (mu_mv_mag_sq < 61) {
if (MVDiff < 12) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (Rs < 1244) {
return 4;
}
else if (Rs >= 1244) {
return 8;
}
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 61) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 60) {
if (TSC < 6) {
if (Rs < 997) {
if (Cs < 1094) {
if (Cs < 651) {
if (Rs < 906) {
return 8;
}
else if (Rs >= 906) {
return 4;
}
}
else if (Cs >= 651) {
if (negBalance < 33) {
if (Rs < 940) {
return 2;
}
else if (Rs >= 940) {
return 8;
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
else if (Cs >= 1094) {
return 4;
}
}
else if (Rs >= 997) {
if (RsCsDiff < 15) {
if (AFD < 4) {
if (mu_mv_mag_sq < 122) {
if (Rs < 1498) {
return 1;
}
else if (Rs >= 1498) {
return 8;
}
}
else if (mu_mv_mag_sq >= 122) {
if (negBalance < 36) {
return 4;
}
else if (negBalance >= 36) {
return 1;
}
}
}
else if (AFD >= 4) {
if (mcTcor < 91) {
if (Rs < 1139) {
return 4;
}
else if (Rs >= 1139) {
return 1;
}
}
else if (mcTcor >= 91) {
if (mu_mv_mag_sq < 116) {
if (AFD < 7) {
if (negBalance < 40) {
if (ssDCval < 71) {
if (tcor < 94) {
if (mcTcor < 96) {
if (ssDCval < 61) {
if (Rs < 1302) {
return 1;
}
else if (Rs >= 1302) {
return 8;
}
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (ssDCval >= 71) {
return 1;
}
}
else if (negBalance >= 40) {
if (Rs < 1334) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (Rs >= 1334) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
return 4;
}
}
}
}
else if (AFD >= 7) {
if (posBalance < 46) {
if (Cs < 724) {
return 2;
}
else if (Cs >= 724) {
return 4;
}
}
else if (posBalance >= 46) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 116) {
if (TSC < 4) {
if (Rs < 1163) {
return 4;
}
else if (Rs >= 1163) {
return 1;
}
}
else if (TSC >= 4) {
if (diffAFD < 2) {
if (diffRsCsdiff < -11) {
if (Rs < 1325) {
return 4;
}
else if (Rs >= 1325) {
return 8;
}
}
else if (diffRsCsdiff >= -11) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 4;
}
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 15) {
if (Cs < 923) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (AFD < 9) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 1;
}
}
}
else if (AFD >= 9) {
if (Rs < 1162) {
return 1;
}
else if (Rs >= 1162) {
return 8;
}
}
}
}
else if (Cs >= 923) {
if (diffMVdiffVal < 3) {
if (negBalance < 34) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (AFD < 5) {
if (MVDiff < 2) {
if (Rs < 1426) {
return 4;
}
else if (Rs >= 1426) {
return 1;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (negBalance >= 34) {
if (Cs < 1257) {
if (ssDCval < 74) {
if (Rs < 1535) {
if (negBalance < 36) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (negBalance >= 36) {
if (mcTcor < 92) {
if (Rs < 1175) {
return 4;
}
else if (Rs >= 1175) {
if (Rs < 1305) {
return 1;
}
else if (Rs >= 1305) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
}
}
else if (Rs >= 1535) {
return 8;
}
}
else if (ssDCval >= 74) {
if (Rs < 1631) {
return 8;
}
else if (Rs >= 1631) {
return 1;
}
}
}
else if (Cs >= 1257) {
if (mu_mv_mag_sq < 104) {
return 1;
}
else if (mu_mv_mag_sq >= 104) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 3) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 4;
}
}
}
}
}
}
else if (TSC >= 6) {
if (MVDiff < 5) {
if (Rs < 1782) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (Rs >= 1782) {
return 1;
}
}
else if (MVDiff >= 5) {
if (Rs < 1418) {
if (Rs < 1032) {
return 8;
}
else if (Rs >= 1032) {
if (refDCval < 73) {
return 4;
}
else if (refDCval >= 73) {
return 8;
}
}
}
else if (Rs >= 1418) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Rs < 1473) {
return 8;
}
else if (Rs >= 1473) {
if (Rs < 1779) {
return 4;
}
else if (Rs >= 1779) {
return 1;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 28) {
if (Cs < 911) {
if (SC < 1937) {
if (ssDCval < 64) {
if (RsDiff < 7) {
if (posBalance < 49) {
if (Cs < 586) {
return 2;
}
else if (Cs >= 586) {
if (Rs < 825) {
return 4;
}
else if (Rs >= 825) {
return 8;
}
}
}
else if (posBalance >= 49) {
return 4;
}
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (ssDCval >= 64) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
return 8;
}
}
}
else if (SC >= 1937) {
if (Rs < 1775) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (posBalance < 52) {
if (AFD < 8) {
if (Rs < 1681) {
return 4;
}
else if (Rs >= 1681) {
return 1;
}
}
else if (AFD >= 8) {
return 8;
}
}
else if (posBalance >= 52) {
if (Rs < 1699) {
if (Rs < 1207) {
return 8;
}
else if (Rs >= 1207) {
return 2;
}
}
else if (Rs >= 1699) {
return 8;
}
}
}
}
else if (Rs >= 1775) {
if (ssDCval < 73) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
return 2;
}
}
else if (ssDCval >= 73) {
return 1;
}
}
}
}
else if (Cs >= 911) {
if (negBalance < 44) {
if (ssDCval < 55) {
if (Rs < 1259) {
if (Rs < 974) {
return 8;
}
else if (Rs >= 974) {
return 4;
}
}
else if (Rs >= 1259) {
if (Rs < 2031) {
return 8;
}
else if (Rs >= 2031) {
return 4;
}
}
}
else if (ssDCval >= 55) {
if (mu_mv_mag_sq < 62) {
if (mu_mv_mag_sq < 49) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
if (Rs < 1725) {
return 4;
}
else if (Rs >= 1725) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 49) {
return 8;
}
}
else if (mu_mv_mag_sq >= 62) {
if (tcor < 95) {
if (ssDCval < 74) {
if (Rs < 665) {
return 1;
}
else if (Rs >= 665) {
if (negBalance < 41) {
if (Rs < 2298) {
if (Cs < 938) {
return 2;
}
else if (Cs >= 938) {
return 4;
}
}
else if (Rs >= 2298) {
return 8;
}
}
else if (negBalance >= 41) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
else if (ssDCval >= 74) {
if (Rs < 2003) {
return 1;
}
else if (Rs >= 2003) {
return 8;
}
}
}
else if (tcor >= 95) {
if (Rs < 1701) {
return 8;
}
else if (Rs >= 1701) {
return 1;
}
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 1586) {
if (ssDCval < 70) {
if (Rs < 1166) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 1017) {
return 1;
}
else if (Rs >= 1017) {
return 4;
}
}
}
else if (Rs >= 1166) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 1341) {
return 4;
}
else if (Rs >= 1341) {
return 8;
}
}
}
}
else if (ssDCval >= 70) {
if (Rs < 1426) {
return 4;
}
else if (Rs >= 1426) {
return 8;
}
}
}
else if (Rs >= 1586) {
if (mu_mv_mag_sq < 90) {
if (Rs < 1857) {
if (Cs < 1236) {
if (MVDiff < 30) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (Cs < 1105) {
return 2;
}
else if (Cs >= 1105) {
return 8;
}
}
}
else if (MVDiff >= 30) {
return 2;
}
}
else if (Cs >= 1236) {
return 2;
}
}
else if (Rs >= 1857) {
return 8;
}
}
else if (mu_mv_mag_sq >= 90) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 78) {
if (MVDiff < 6) {
if (negBalance < 48) {
if (tcor < 86) {
if (mu_mv_mag_sq < 54) {
return 2;
}
else if (mu_mv_mag_sq >= 54) {
if (AFD < 6) {
if (Rs < 847) {
return 8;
}
else if (Rs >= 847) {
return 4;
}
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (tcor >= 86) {
if (diffAFD < 2) {
if (mcTcor < 95) {
if (posBalance < 69) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (ssDCval < 96) {
if (posBalance < 50) {
if (SC < 1597) {
return 1;
}
else if (SC >= 1597) {
return 2;
}
}
else if (posBalance >= 50) {
if (Cs < 1259) {
if (negBalance < 31) {
return 2;
}
else if (negBalance >= 31) {
return 8;
}
}
else if (Cs >= 1259) {
return 1;
}
}
}
else if (ssDCval >= 96) {
return 8;
}
}
}
else if (posBalance >= 69) {
if (Rs < 1186) {
return 8;
}
else if (Rs >= 1186) {
return 1;
}
}
}
else if (mcTcor >= 95) {
if (negBalance < 38) {
if (mu_mv_mag_sq < 75) {
if (Rs < 1717) {
if (Rs < 1067) {
return 2;
}
else if (Rs >= 1067) {
return 8;
}
}
else if (Rs >= 1717) {
return 1;
}
}
else if (mu_mv_mag_sq >= 75) {
if (tcor < 90) {
return 4;
}
else if (tcor >= 90) {
if (negBalance < 24) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (negBalance >= 24) {
return 8;
}
}
}
}
else if (negBalance >= 38) {
if (MVDiff < 3) {
if (ssDCval < 108) {
if (mu_mv_mag_sq < 108) {
if (Rs < 869) {
if (SC < 1765) {
if (tcor < 96) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
return 1;
}
}
else if (tcor >= 96) {
return 8;
}
}
else if (SC >= 1765) {
return 4;
}
}
else if (Rs >= 869) {
if (SC < 2571) {
return 8;
}
else if (SC >= 2571) {
if (Rs < 1861) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (Rs >= 1861) {
if (Rs < 1987) {
return 8;
}
else if (Rs >= 1987) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
return 8;
}
}
else if (ssDCval >= 108) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 66) {
if (Cs < 1136) {
return 8;
}
else if (Cs >= 1136) {
return 4;
}
}
else if (mu_mv_mag_sq >= 66) {
if (tcor < 90) {
return 4;
}
else if (tcor >= 90) {
return 8;
}
}
}
}
}
}
else if (diffAFD >= 2) {
if (Cs < 722) {
return 1;
}
else if (Cs >= 722) {
if (RsDiff < 2) {
if (Rs < 717) {
return 2;
}
else if (Rs >= 717) {
return 1;
}
}
else if (RsDiff >= 2) {
if (posBalance < 51) {
if (TSC < 7) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (posBalance >= 51) {
if (RsCsDiff < 49) {
return 8;
}
else if (RsCsDiff >= 49) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 48) {
if (Rs < 505) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 426) {
return 4;
}
else if (Rs >= 426) {
return 1;
}
}
}
else if (Rs >= 505) {
if (tcor < 91) {
if (diffTSC < 2) {
if (Rs < 2082) {
return 8;
}
else if (Rs >= 2082) {
return 1;
}
}
else if (diffTSC >= 2) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
else if (tcor >= 91) {
if (posBalance < 59) {
if (Cs < 885) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (Cs >= 885) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1025) {
return 8;
}
else if (Rs >= 1025) {
return 1;
}
}
}
}
else if (posBalance >= 59) {
return 1;
}
}
}
}
}
else if (MVDiff >= 6) {
if (MVDiff < 23) {
if (Cs < 1048) {
if (negBalance < 27) {
if (Cs < 707) {
if (Rs < 1064) {
return 2;
}
else if (Rs >= 1064) {
return 1;
}
}
else if (Cs >= 707) {
return 8;
}
}
else if (negBalance >= 27) {
if (refDCval < 140) {
if (diffRsCsdiff < -12) {
if (diffRsCsdiff < -14) {
if (Cs < 646) {
return 1;
}
else if (Cs >= 646) {
return 8;
}
}
else if (diffRsCsdiff >= -14) {
if (Cs < 816) {
if (Rs < 1405) {
return 8;
}
else if (Rs >= 1405) {
return 2;
}
}
else if (Cs >= 816) {
return 1;
}
}
}
else if (diffRsCsdiff >= -12) {
if (tcor < 81) {
return 2;
}
else if (tcor >= 81) {
return 8;
}
}
}
else if (refDCval >= 140) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
if (AFD < 9) {
if (Rs < 920) {
return 2;
}
else if (Rs >= 920) {
return 4;
}
}
else if (AFD >= 9) {
return 1;
}
}
}
}
}
else if (Cs >= 1048) {
if (ssDCval < 105) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (Rs < 1915) {
if (refDCval < 92) {
return 8;
}
else if (refDCval >= 92) {
if (diffMVdiffVal < 1) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 2;
}
}
}
}
else if (Rs >= 1915) {
if (AFD < 9) {
if (mu_mv_mag_sq < 76) {
return 4;
}
else if (mu_mv_mag_sq >= 76) {
if (Rs < 2143) {
return 8;
}
else if (Rs >= 2143) {
if (Rs < 2219) {
return 2;
}
else if (Rs >= 2219) {
return 1;
}
}
}
}
else if (AFD >= 9) {
return 8;
}
}
}
}
else if (ssDCval >= 105) {
if (RsCsDiff < 19) {
if (mcTcor < 95) {
if (ssDCval < 139) {
if (Rs < 711) {
return 8;
}
else if (Rs >= 711) {
return 4;
}
}
else if (ssDCval >= 139) {
return 8;
}
}
else if (mcTcor >= 95) {
return 8;
}
}
else if (RsCsDiff >= 19) {
return 8;
}
}
}
}
else if (MVDiff >= 23) {
if (TSC < 5) {
if (Cs < 1037) {
return 4;
}
else if (Cs >= 1037) {
return 8;
}
}
else if (TSC >= 5) {
if (tcor < 81) {
return 2;
}
else if (tcor >= 81) {
if (mu_mv_mag_sq < 45) {
if (RsCsDiff < 29) {
return 4;
}
else if (RsCsDiff >= 29) {
return 8;
}
}
else if (mu_mv_mag_sq >= 45) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 158) {
if (Rs < 875) {
return 4;
}
else if (Rs >= 875) {
return 2;
}
}
}
}
}
}
else if (AFD >= 13) {
if (refDCval < 114) {
if (Rs < 1903) {
if (MVDiff < 15) {
if (refDCval < 91) {
if (refDCval < 58) {
if (RsDiff < 7) {
if (negBalance < 49) {
if (Rs < 920) {
return 1;
}
else if (Rs >= 920) {
return 8;
}
}
else if (negBalance >= 49) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
}
else if (RsDiff >= 7) {
if (posBalance < 61) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 8;
}
}
else if (posBalance >= 61) {
return 8;
}
}
}
else if (refDCval >= 58) {
if (mcTcor < 94) {
if (refDCval < 75) {
if (negBalance < 51) {
if (TSC < 9) {
if (posBalance < 59) {
if (Rs < 1296) {
return 1;
}
else if (Rs >= 1296) {
if (Rs < 1647) {
return 8;
}
else if (Rs >= 1647) {
return 2;
}
}
}
else if (posBalance >= 59) {
if (mcTcor < 86) {
return 1;
}
else if (mcTcor >= 86) {
return 8;
}
}
}
else if (TSC >= 9) {
if (Rs < 1197) {
return 2;
}
else if (Rs >= 1197) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (mu_mv_mag_sq < 64) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (RsCsDiff < 63) {
return 8;
}
else if (RsCsDiff >= 63) {
if (AFD < 15) {
if (RsCsDiff < 80) {
if (Rs < 1735) {
return 2;
}
else if (Rs >= 1735) {
return 1;
}
}
else if (RsCsDiff >= 80) {
return 8;
}
}
else if (AFD >= 15) {
if (Cs < 1254) {
return 8;
}
else if (Cs >= 1254) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 64) {
return 4;
}
}
}
else if (refDCval >= 75) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1504) {
if (negBalance < 55) {
if (Cs < 1280) {
return 8;
}
else if (Cs >= 1280) {
return 4;
}
}
else if (negBalance >= 55) {
return 4;
}
}
else if (Rs >= 1504) {
if (Rs < 1704) {
return 8;
}
else if (Rs >= 1704) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (AFD < 19) {
if (Cs < 815) {
if (Cs < 685) {
return 8;
}
else if (Cs >= 685) {
return 4;
}
}
else if (Cs >= 815) {
return 8;
}
}
else if (AFD >= 19) {
return 4;
}
}
}
}
}
else if (mcTcor >= 94) {
if (MVDiff < 6) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 4;
}
}
else if (MVDiff >= 6) {
if (posBalance < 65) {
if (Rs < 523) {
return 2;
}
else if (Rs >= 523) {
if (ssDCval < 75) {
if (mu_mv_mag_sq < 2) {
if (Cs < 1018) {
return 8;
}
else if (Cs >= 1018) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
else if (ssDCval >= 75) {
if (mu_mv_mag_sq < 5) {
if (Rs < 890) {
return 8;
}
else if (Rs >= 890) {
if (SC < 2561) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (SC >= 2561) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (diffAFD < -3) {
if (Rs < 1562) {
return 1;
}
else if (Rs >= 1562) {
return 4;
}
}
else if (diffAFD >= -3) {
if (RsCsDiff < 51) {
if (Rs < 871) {
return 8;
}
else if (Rs >= 871) {
return 4;
}
}
else if (RsCsDiff >= 51) {
if (SC < 2858) {
return 8;
}
else if (SC >= 2858) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 65) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Cs < 1242) {
return 1;
}
else if (Cs >= 1242) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 91) {
if (mcTcor < 90) {
if (Rs < 1595) {
if (mcTcor < 79) {
if (Rs < 1197) {
return 8;
}
else if (Rs >= 1197) {
return 4;
}
}
else if (mcTcor >= 79) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
}
else if (Rs >= 1595) {
if (Cs < 1144) {
return 2;
}
else if (Cs >= 1144) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (AFD < 18) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 7) {
if (SC < 2064) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
return 4;
}
}
else if (SC >= 2064) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 1470) {
return 8;
}
else if (Rs >= 1470) {
return 1;
}
}
}
}
}
else if (AFD >= 18) {
if (tcor < 74) {
if (Rs < 687) {
return 1;
}
else if (Rs >= 687) {
return 4;
}
}
else if (tcor >= 74) {
if (Rs < 708) {
return 2;
}
else if (Rs >= 708) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 15) {
if (SC < 2502) {
if (mcTcor < 79) {
if (TSC < 11) {
if (refDCval < 45) {
if (refDCval < 40) {
return 8;
}
else if (refDCval >= 40) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
else if (refDCval >= 45) {
if (MVDiff < 99) {
if (TSC < 9) {
if (diffRsCsdiff < 51) {
return 4;
}
else if (diffRsCsdiff >= 51) {
if (Rs < 477) {
return 8;
}
else if (Rs >= 477) {
return 4;
}
}
}
else if (TSC >= 9) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (Cs < 601) {
if (TSC < 10) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (Cs >= 601) {
if (diffRsCsdiff < -31) {
if (Rs < 521) {
return 2;
}
else if (Rs >= 521) {
return 1;
}
}
else if (diffRsCsdiff >= -31) {
if (RsCsDiff < 27) {
return 8;
}
else if (RsCsDiff >= 27) {
if (mcTcor < 78) {
return 4;
}
else if (mcTcor >= 78) {
if (Rs < 917) {
return 4;
}
else if (Rs >= 917) {
return 1;
}
}
}
}
}
}
}
}
else if (MVDiff >= 99) {
return 8;
}
}
}
else if (TSC >= 11) {
if (CsDiff < 6) {
if (refDCval < 68) {
return 8;
}
else if (refDCval >= 68) {
if (refDCval < 80) {
if (Rs < 1409) {
return 4;
}
else if (Rs >= 1409) {
return 8;
}
}
else if (refDCval >= 80) {
if (Rs < 1404) {
return 8;
}
else if (Rs >= 1404) {
return 1;
}
}
}
}
else if (CsDiff >= 6) {
if (posBalance < 58) {
if (AFD < 31) {
return 8;
}
else if (AFD >= 31) {
return 1;
}
}
else if (posBalance >= 58) {
if (gchDC < 3) {
if (diffRsCsdiff < -7) {
return 2;
}
else if (diffRsCsdiff >= -7) {
return 4;
}
}
else if (gchDC >= 3) {
if (mcTcor < 58) {
if (Rs < 796) {
return 2;
}
else if (Rs >= 796) {
if (Cs < 974) {
return 4;
}
else if (Cs >= 974) {
if (Rs < 1110) {
return 1;
}
else if (Rs >= 1110) {
return 8;
}
}
}
}
else if (mcTcor >= 58) {
if (Rs < 1037) {
return 8;
}
else if (Rs >= 1037) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 79) {
if (tcor < 92) {
if (mu_mv_mag_sq < 3) {
if (mcTcor < 89) {
if (refDCval < 106) {
if (RsDiff < 6) {
if (SC < 1259) {
if (Rs < 565) {
return 4;
}
else if (Rs >= 565) {
return 2;
}
}
else if (SC >= 1259) {
if (Cs < 1031) {
if (diffMVdiffVal < -16) {
if (Rs < 808) {
return 1;
}
else if (Rs >= 808) {
return 4;
}
}
else if (diffMVdiffVal >= -16) {
return 8;
}
}
else if (Cs >= 1031) {
if (tcor < 71) {
return 4;
}
else if (tcor >= 71) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
}
else if (RsDiff >= 6) {
if (MVDiff < 31) {
if (Rs < 1138) {
if (MVDiff < 29) {
return 4;
}
else if (MVDiff >= 29) {
return 2;
}
}
else if (Rs >= 1138) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (MVDiff >= 31) {
if (ssDCval < 60) {
if (Rs < 1110) {
if (Rs < 914) {
return 8;
}
else if (Rs >= 914) {
return 2;
}
}
else if (Rs >= 1110) {
return 4;
}
}
else if (ssDCval >= 60) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (Rs < 885) {
return 2;
}
else if (Rs >= 885) {
if (MVDiff < 38) {
return 4;
}
else if (MVDiff >= 38) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 106) {
return 1;
}
}
else if (mcTcor >= 89) {
if (Rs < 1375) {
if (refDCval < 103) {
if (ssDCval < 73) {
if (Rs < 954) {
if (ssDCval < 65) {
if (Rs < 494) {
if (Rs < 310) {
return 4;
}
else if (Rs >= 310) {
return 2;
}
}
else if (Rs >= 494) {
if (Cs < 1020) {
if (gchDC < 5) {
return 1;
}
else if (gchDC >= 5) {
return 2;
}
}
else if (Cs >= 1020) {
return 2;
}
}
}
else if (ssDCval >= 65) {
if (MVDiff < 37) {
if (refDCval < 69) {
if (Rs < 755) {
return 8;
}
else if (Rs >= 755) {
if (Rs < 805) {
return 4;
}
else if (Rs >= 805) {
return 1;
}
}
}
else if (refDCval >= 69) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 1;
}
}
}
else if (MVDiff >= 37) {
return 4;
}
}
}
else if (Rs >= 954) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
if (Cs < 845) {
return 1;
}
else if (Cs >= 845) {
return 8;
}
}
}
}
else if (ssDCval >= 73) {
if (Cs < 959) {
if (TSC < 11) {
if (SC < 1811) {
if (Rs < 773) {
if (Cs < 926) {
return 2;
}
else if (Cs >= 926) {
return 1;
}
}
else if (Rs >= 773) {
if (MVDiff < 71) {
if (Rs < 837) {
return 8;
}
else if (Rs >= 837) {
return 1;
}
}
else if (MVDiff >= 71) {
return 8;
}
}
}
else if (SC >= 1811) {
if (diffRsCsdiff < -12) {
return 8;
}
else if (diffRsCsdiff >= -12) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 2;
}
}
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (Cs >= 959) {
if (negBalance < 53) {
if (Rs < 868) {
return 4;
}
else if (Rs >= 868) {
if (Rs < 989) {
return 2;
}
else if (Rs >= 989) {
return 1;
}
}
}
else if (negBalance >= 53) {
return 2;
}
}
}
}
else if (refDCval >= 103) {
if (Rs < 1127) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
return 4;
}
}
else if (Rs >= 1127) {
if (Rs < 1135) {
return 1;
}
else if (Rs >= 1135) {
return 8;
}
}
}
}
else if (Rs >= 1375) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
if (diffMVdiffVal < -7) {
if (Cs < 742) {
return 8;
}
else if (Cs >= 742) {
return 2;
}
}
else if (diffMVdiffVal >= -7) {
if (Rs < 1752) {
return 1;
}
else if (Rs >= 1752) {
if (Rs < 1820) {
return 8;
}
else if (Rs >= 1820) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 26) {
if (Rs < 620) {
if (tcor < 51) {
if (MVDiff < 111) {
return 4;
}
else if (MVDiff >= 111) {
return 1;
}
}
else if (tcor >= 51) {
if (ssDCval < 69) {
if (Cs < 593) {
if (Cs < 579) {
if (AFD < 16) {
if (Rs < 539) {
return 1;
}
else if (Rs >= 539) {
return 8;
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (Cs >= 579) {
return 8;
}
}
else if (Cs >= 593) {
if (MVDiff < 28) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 312) {
return 2;
}
else if (Rs >= 312) {
return 4;
}
}
}
else if (MVDiff >= 28) {
if (ssDCval < 56) {
if (SC < 1168) {
if (Rs < 295) {
return 1;
}
else if (Rs >= 295) {
return 4;
}
}
else if (SC >= 1168) {
if (Rs < 554) {
return 2;
}
else if (Rs >= 554) {
return 8;
}
}
}
else if (ssDCval >= 56) {
if (AFD < 15) {
if (Rs < 465) {
return 1;
}
else if (Rs >= 465) {
return 4;
}
}
else if (AFD >= 15) {
return 4;
}
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 307) {
return 4;
}
else if (Rs >= 307) {
if (AFD < 17) {
if (posBalance < 61) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 2;
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (AFD >= 17) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
if (Cs < 820) {
if (Rs < 499) {
return 2;
}
else if (Rs >= 499) {
return 8;
}
}
else if (Cs >= 820) {
if (Cs < 894) {
return 1;
}
else if (Cs >= 894) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 620) {
if (posBalance < 46) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
if (negBalance < 56) {
return 2;
}
else if (negBalance >= 56) {
if (MVDiff < 40) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (MVDiff >= 40) {
return 4;
}
}
}
}
else if (posBalance >= 46) {
if (posBalance < 67) {
if (mu_mv_mag_sq < 55) {
if (mu_mv_mag_sq < 10) {
if (refDCval < 90) {
if (Rs < 1267) {
if (RsCsDiff < 38) {
if (Rs < 901) {
return 8;
}
else if (Rs >= 901) {
return 4;
}
}
else if (RsCsDiff >= 38) {
if (negBalance < 51) {
if (Cs < 573) {
return 2;
}
else if (Cs >= 573) {
if (posBalance < 65) {
return 4;
}
else if (posBalance >= 65) {
if (posBalance < 66) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 4;
}
}
else if (posBalance >= 66) {
return 4;
}
}
}
}
else if (negBalance >= 51) {
if (refDCval < 67) {
if (Cs < 830) {
return 4;
}
else if (Cs >= 830) {
return 1;
}
}
else if (refDCval >= 67) {
if (SC < 1654) {
if (SC < 1493) {
return 1;
}
else if (SC >= 1493) {
if (Rs < 862) {
return 2;
}
else if (Rs >= 862) {
return 4;
}
}
}
else if (SC >= 1654) {
if (diffRsCsdiff < -6) {
if (diffMVdiffVal < 9) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (diffMVdiffVal >= 9) {
return 4;
}
}
else if (diffRsCsdiff >= -6) {
if (Rs < 1199) {
return 4;
}
else if (Rs >= 1199) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1267) {
if (tcor < 82) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 4;
}
}
else if (tcor >= 82) {
if (Rs < 1379) {
return 8;
}
else if (Rs >= 1379) {
return 4;
}
}
}
}
else if (refDCval >= 90) {
if (RsCsDiff < 94) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (RsCsDiff >= 94) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (RsDiff < 9) {
if (tcor < 81) {
if (refDCval < 56) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 8;
}
}
else if (refDCval >= 56) {
if (TSC < 10) {
if (RsCsDiff < 73) {
if (mu_mv_mag_sq < 34) {
if (refDCval < 84) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 721) {
return 8;
}
else if (Rs >= 721) {
return 1;
}
}
}
else if (refDCval >= 84) {
return 8;
}
}
else if (mu_mv_mag_sq >= 34) {
if (Rs < 1077) {
if (tcor < 62) {
return 4;
}
else if (tcor >= 62) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 1;
}
}
}
else if (Rs >= 1077) {
if (AFD < 17) {
if (Rs < 1360) {
return 8;
}
else if (Rs >= 1360) {
return 2;
}
}
else if (AFD >= 17) {
return 4;
}
}
}
}
else if (RsCsDiff >= 73) {
if (Cs < 777) {
if (Rs < 1084) {
return 1;
}
else if (Rs >= 1084) {
return 4;
}
}
else if (Cs >= 777) {
if (Rs < 1055) {
return 2;
}
else if (Rs >= 1055) {
return 1;
}
}
}
}
else if (TSC >= 10) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
return 1;
}
}
}
}
}
else if (tcor >= 81) {
if (diffRsCsdiff < -16) {
return 2;
}
else if (diffRsCsdiff >= -16) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (Cs < 773) {
if (TSC < 8) {
if (Rs < 869) {
return 8;
}
else if (Rs >= 869) {
return 1;
}
}
else if (TSC >= 8) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
return 4;
}
}
}
else if (Cs >= 773) {
if (tcor < 61) {
if (Cs < 911) {
return 2;
}
else if (Cs >= 911) {
if (Rs < 1291) {
return 1;
}
else if (Rs >= 1291) {
return 8;
}
}
}
else if (tcor >= 61) {
if (TSC < 11) {
if (Cs < 978) {
return 8;
}
else if (Cs >= 978) {
if (SC < 2265) {
return 8;
}
else if (SC >= 2265) {
return 4;
}
}
}
else if (TSC >= 11) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (ssDCval < 70) {
if (Rs < 865) {
return 4;
}
else if (Rs >= 865) {
if (Rs < 1090) {
return 8;
}
else if (Rs >= 1090) {
return 2;
}
}
}
else if (ssDCval >= 70) {
return 4;
}
}
}
else if (posBalance >= 67) {
if (Rs < 794) {
return 8;
}
else if (Rs >= 794) {
if (ssDCval < 69) {
if (SC < 1813) {
return 1;
}
else if (SC >= 1813) {
return 4;
}
}
else if (ssDCval >= 69) {
if (Cs < 596) {
return 2;
}
else if (Cs >= 596) {
if (Rs < 1567) {
if (negBalance < 46) {
if (MVDiff < 139) {
return 1;
}
else if (MVDiff >= 139) {
return 8;
}
}
else if (negBalance >= 46) {
if (posBalance < 70) {
if (Rs < 929) {
return 2;
}
else if (Rs >= 929) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
}
else if (posBalance >= 70) {
return 8;
}
}
}
else if (Rs >= 1567) {
return 8;
}
}
}
}
}
}
}
}
else if (AFD >= 26) {
if (Cs < 638) {
if (mcTcor < 88) {
if (Rs < 259) {
return 2;
}
else if (Rs >= 259) {
return 8;
}
}
else if (mcTcor >= 88) {
if (AFD < 36) {
return 4;
}
else if (AFD >= 36) {
return 2;
}
}
}
else if (Cs >= 638) {
if (MVDiff < 209) {
if (tcor < 38) {
return 2;
}
else if (tcor >= 38) {
if (refDCval < 71) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 1;
}
}
else if (refDCval >= 71) {
if (ssDCval < 83) {
return 2;
}
else if (ssDCval >= 83) {
if (negBalance < 59) {
if (SC < 2084) {
if (mcTcor < 95) {
if (MVDiff < 199) {
if (diffAFD < 21) {
return 1;
}
else if (diffAFD >= 21) {
return 4;
}
}
else if (MVDiff >= 199) {
return 4;
}
}
else if (mcTcor >= 95) {
return 2;
}
}
else if (SC >= 2084) {
if (SC < 2355) {
return 2;
}
else if (SC >= 2355) {
return 4;
}
}
}
else if (negBalance >= 59) {
if (diffAFD < -1) {
if (Cs < 732) {
return 4;
}
else if (Cs >= 732) {
return 2;
}
}
else if (diffAFD >= -1) {
if (mu_mv_mag_sq < 5) {
return 2;
}
else if (mu_mv_mag_sq >= 5) {
if (posBalance < 58) {
if (Cs < 904) {
return 4;
}
else if (Cs >= 904) {
return 1;
}
}
else if (posBalance >= 58) {
return 4;
}
}
}
}
}
}
}
}
else if (MVDiff >= 209) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
return 1;
}
}
}
}
}
}
else if (tcor >= 92) {
if (Rs < 544) {
return 1;
}
else if (Rs >= 544) {
return 8;
}
}
}
}
else if (SC >= 2502) {
if (RsDiff < 9) {
if (AFD < 19) {
if (posBalance < 70) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 97) {
if (negBalance < 49) {
if (Rs < 1731) {
return 4;
}
else if (Rs >= 1731) {
return 1;
}
}
else if (negBalance >= 49) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 4;
}
}
}
else if (ssDCval >= 97) {
if (Rs < 1500) {
return 4;
}
else if (Rs >= 1500) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (mu_mv_mag_sq < 50) {
if (mu_mv_mag_sq < 3) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 33) {
return 2;
}
else if (negBalance >= 33) {
if (Cs < 806) {
return 1;
}
else if (Cs >= 806) {
if (diffTSC < 3) {
return 8;
}
else if (diffTSC >= 3) {
if (Rs < 1461) {
return 1;
}
else if (Rs >= 1461) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
return 1;
}
}
}
else if (posBalance >= 70) {
return 2;
}
}
else if (AFD >= 19) {
if (mu_mv_mag_sq < 22) {
if (MVDiff < 57) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (AFD < 23) {
return 1;
}
else if (AFD >= 23) {
if (Rs < 1627) {
return 4;
}
else if (Rs >= 1627) {
return 1;
}
}
}
}
else if (MVDiff >= 57) {
return 8;
}
}
else if (mu_mv_mag_sq >= 22) {
return 2;
}
}
}
else if (RsDiff >= 9) {
if (posBalance < 50) {
if (refDCval < 74) {
if (Cs < 1246) {
return 2;
}
else if (Cs >= 1246) {
return 8;
}
}
else if (refDCval >= 74) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (refDCval < 103) {
if (ssDCval < 94) {
if (AFD < 24) {
if (diffRsCsdiff < -31) {
if (mu_mv_mag_sq < 7) {
return 2;
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 4;
}
}
}
else if (diffRsCsdiff >= -31) {
if (diffRsCsdiff < -25) {
return 4;
}
else if (diffRsCsdiff >= -25) {
if (RsDiff < 10) {
if (refDCval < 87) {
return 8;
}
else if (refDCval >= 87) {
if (Rs < 1790) {
return 4;
}
else if (Rs >= 1790) {
return 8;
}
}
}
else if (RsDiff >= 10) {
if (diffMVdiffVal < -3) {
if (Rs < 1898) {
return 4;
}
else if (Rs >= 1898) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (Cs < 920) {
return 4;
}
else if (Cs >= 920) {
if (diffMVdiffVal < 1) {
if (Rs < 1604) {
return 2;
}
else if (Rs >= 1604) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
}
}
}
}
}
else if (AFD >= 24) {
if (Rs < 1721) {
if (Rs < 1648) {
return 4;
}
else if (Rs >= 1648) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 4;
}
}
}
else if (Rs >= 1721) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 2;
}
}
}
}
else if (ssDCval >= 94) {
if (MVDiff < 36) {
if (Cs < 1094) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
return 1;
}
}
else if (Cs >= 1094) {
return 2;
}
}
else if (MVDiff >= 36) {
return 8;
}
}
}
else if (refDCval >= 103) {
if (diffRsCsdiff < 20) {
if (Cs < 738) {
return 8;
}
else if (Cs >= 738) {
return 4;
}
}
else if (diffRsCsdiff >= 20) {
if (Rs < 1857) {
return 8;
}
else if (Rs >= 1857) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 1903) {
if (refDCval < 71) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (Cs < 1209) {
if (TSC < 12) {
if (RsDiff < 8) {
if (Cs < 689) {
return 4;
}
else if (Cs >= 689) {
return 2;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (TSC >= 12) {
return 4;
}
}
else if (Cs >= 1209) {
if (AFD < 17) {
if (Rs < 2019) {
return 4;
}
else if (Rs >= 2019) {
return 8;
}
}
else if (AFD >= 17) {
return 2;
}
}
}
}
else if (refDCval >= 71) {
if (gchDC < 6) {
if (RsCsDiff < 170) {
if (TSC < 11) {
if (mcTcor < 87) {
if (Cs < 1159) {
return 8;
}
else if (Cs >= 1159) {
if (Rs < 2223) {
return 4;
}
else if (Rs >= 2223) {
return 8;
}
}
}
else if (mcTcor >= 87) {
if (tcor < 86) {
if (MVDiff < 36) {
if (ssDCval < 79) {
if (diffMVdiffVal < 1) {
if (Cs < 737) {
return 2;
}
else if (Cs >= 737) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 857) {
return 2;
}
else if (Cs >= 857) {
if (Rs < 2024) {
return 2;
}
else if (Rs >= 2024) {
return 4;
}
}
}
}
else if (ssDCval >= 79) {
if (TSC < 9) {
if (ssDCval < 80) {
if (Rs < 2003) {
return 1;
}
else if (Rs >= 2003) {
return 4;
}
}
else if (ssDCval >= 80) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
else if (TSC >= 9) {
return 4;
}
}
}
else if (MVDiff >= 36) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (SC < 3129) {
return 2;
}
else if (SC >= 3129) {
return 4;
}
}
}
}
else if (tcor >= 86) {
if (SC < 3226) {
if (MVDiff < 20) {
if (Rs < 2196) {
return 4;
}
else if (Rs >= 2196) {
return 2;
}
}
else if (MVDiff >= 20) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 8;
}
}
}
else if (SC >= 3226) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
}
else if (TSC >= 11) {
if (SC < 3222) {
if (diffAFD < 2) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 1;
}
}
else if (diffAFD >= 2) {
if (Cs < 994) {
return 8;
}
else if (Cs >= 994) {
return 2;
}
}
}
else if (SC >= 3222) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
}
}
else if (RsCsDiff >= 170) {
if (gchDC < 3) {
if (mcTcor < 92) {
if (AFD < 25) {
if (ssDCval < 85) {
if (SC < 3137) {
return 8;
}
else if (SC >= 3137) {
if (Cs < 1276) {
return 4;
}
else if (Cs >= 1276) {
return 2;
}
}
}
else if (ssDCval >= 85) {
if (Rs < 1926) {
return 1;
}
else if (Rs >= 1926) {
if (refDCval < 86) {
return 1;
}
else if (refDCval >= 86) {
return 4;
}
}
}
}
else if (AFD >= 25) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
if (Rs < 2009) {
return 8;
}
else if (Rs >= 2009) {
return 2;
}
}
}
}
else if (mcTcor >= 92) {
if (RsDiff < 14) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 2;
}
}
else if (RsDiff >= 14) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
if (Rs < 1933) {
return 1;
}
else if (Rs >= 1933) {
return 2;
}
}
}
}
}
else if (gchDC >= 3) {
if (mu_mv_mag_sq < 19) {
return 2;
}
else if (mu_mv_mag_sq >= 19) {
if (Rs < 2120) {
if (Rs < 1927) {
return 1;
}
else if (Rs >= 1927) {
return 8;
}
}
else if (Rs >= 2120) {
return 2;
}
}
}
}
}
else if (gchDC >= 6) {
if (Cs < 590) {
return 1;
}
else if (Cs >= 590) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 114) {
if (MVDiff < 20) {
if (tcor < 65) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (TSC < 7) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (tcor >= 65) {
if (RsDiff < 9) {
if (negBalance < 62) {
if (ssDCval < 169) {
if (diffRsCsdiff < -31) {
if (Rs < 970) {
return 8;
}
else if (Rs >= 970) {
return 1;
}
}
else if (diffRsCsdiff >= -31) {
if (MVDiff < 5) {
if (RsDiff < 5) {
if (Rs < 990) {
return 1;
}
else if (Rs >= 990) {
return 4;
}
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (MVDiff >= 5) {
if (diffAFD < -2) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 1573) {
return 1;
}
else if (Rs >= 1573) {
return 8;
}
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
else if (ssDCval >= 169) {
return 2;
}
}
else if (negBalance >= 62) {
if (ssDCval < 119) {
return 8;
}
else if (ssDCval >= 119) {
return 2;
}
}
}
else if (RsDiff >= 9) {
if (tcor < 87) {
if (Cs < 882) {
if (AFD < 29) {
return 2;
}
else if (AFD >= 29) {
return 8;
}
}
else if (Cs >= 882) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (Rs < 1286) {
return 8;
}
else if (Rs >= 1286) {
return 2;
}
}
}
}
else if (tcor >= 87) {
if (Rs < 1620) {
return 2;
}
else if (Rs >= 1620) {
return 1;
}
}
}
}
}
else if (MVDiff >= 20) {
if (SC < 2117) {
if (tcor < 31) {
return 1;
}
else if (tcor >= 31) {
if (mu_mv_mag_sq < 5) {
if (ssDCval < 124) {
if (diffMVdiffVal < -49) {
return 1;
}
else if (diffMVdiffVal >= -49) {
if (AFD < 23) {
if (Rs < 682) {
return 4;
}
else if (Rs >= 682) {
return 2;
}
}
else if (AFD >= 23) {
return 2;
}
}
}
else if (ssDCval >= 124) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
if (gchDC < 6) {
if (refDCval < 149) {
if (AFD < 23) {
if (Cs < 816) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (Rs < 1323) {
return 4;
}
else if (Rs >= 1323) {
return 8;
}
}
}
else if (Cs >= 816) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 8;
}
}
}
else if (AFD >= 23) {
if (Cs < 688) {
if (Rs < 960) {
return 1;
}
else if (Rs >= 960) {
if (Rs < 1152) {
return 2;
}
else if (Rs >= 1152) {
return 8;
}
}
}
else if (Cs >= 688) {
return 2;
}
}
}
else if (refDCval >= 149) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
return 2;
}
}
}
else if (gchDC >= 6) {
if (diffRsCsdiff < 29) {
if (AFD < 36) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
return 8;
}
}
else if (AFD >= 36) {
return 1;
}
}
else if (diffRsCsdiff >= 29) {
if (Cs < 890) {
return 2;
}
else if (Cs >= 890) {
return 8;
}
}
}
}
}
}
else if (SC >= 2117) {
if (tcor < 79) {
if (mu_mv_mag_sq < 179) {
if (mcTcor < 88) {
if (diffTSC < 5) {
if (tcor < 60) {
if (tcor < 36) {
if (Rs < 1503) {
return 2;
}
else if (Rs >= 1503) {
return 1;
}
}
else if (tcor >= 36) {
return 8;
}
}
else if (tcor >= 60) {
if (TSC < 11) {
if (Cs < 1245) {
return 4;
}
else if (Cs >= 1245) {
return 8;
}
}
else if (TSC >= 11) {
if (mcTcor < 85) {
if (TSC < 16) {
return 8;
}
else if (TSC >= 16) {
return 1;
}
}
else if (mcTcor >= 85) {
if (Rs < 1683) {
return 4;
}
else if (Rs >= 1683) {
return 1;
}
}
}
}
}
else if (diffTSC >= 5) {
if (Rs < 1749) {
if (AFD < 23) {
if (Rs < 1505) {
return 2;
}
else if (Rs >= 1505) {
return 8;
}
}
else if (AFD >= 23) {
return 4;
}
}
else if (Rs >= 1749) {
return 1;
}
}
}
else if (mcTcor >= 88) {
if (RsCsDiff < 149) {
if (SC < 2195) {
return 1;
}
else if (SC >= 2195) {
return 8;
}
}
else if (RsCsDiff >= 149) {
if (Cs < 601) {
return 1;
}
else if (Cs >= 601) {
if (Rs < 2104) {
if (Cs < 1237) {
return 2;
}
else if (Cs >= 1237) {
if (Rs < 1884) {
return 8;
}
else if (Rs >= 1884) {
return 2;
}
}
}
else if (Rs >= 2104) {
if (Cs < 830) {
return 1;
}
else if (Cs >= 830) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 179) {
return 1;
}
}
else if (tcor >= 79) {
if (RsCsDiff < 110) {
if (AFD < 17) {
if (ssDCval < 118) {
return 1;
}
else if (ssDCval >= 118) {
if (Cs < 607) {
return 1;
}
else if (Cs >= 607) {
if (Rs < 1053) {
return 1;
}
else if (Rs >= 1053) {
return 8;
}
}
}
}
else if (AFD >= 17) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 4;
}
}
}
else if (RsCsDiff >= 110) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (ssDCval < 131) {
return 1;
}
else if (ssDCval >= 131) {
if (Cs < 1279) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
if (MVDiff < 108) {
if (Rs < 1873) {
if (Rs < 1801) {
return 1;
}
else if (Rs >= 1801) {
return 2;
}
}
else if (Rs >= 1873) {
return 1;
}
}
else if (MVDiff >= 108) {
return 8;
}
}
}
else if (Cs >= 1279) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 88) {
if (SC < 2150) {
if (Rs < 1002) {
if (refDCval < 63) {
if (Rs < 843) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (Rs < 531) {
if (SC < 1069) {
if (ssDCval < 45) {
if (posBalance < 17) {
return 8;
}
else if (posBalance >= 17) {
return 1;
}
}
else if (ssDCval >= 45) {
if (tcor < 94) {
if (Rs < 290) {
return 1;
}
else if (Rs >= 290) {
return 8;
}
}
else if (tcor >= 94) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 59) {
return 4;
}
else if (mu_mv_mag_sq >= 59) {
return 1;
}
}
else if (ssDCval >= 56) {
return 4;
}
}
}
}
else if (SC >= 1069) {
if (CsDiff < 1) {
if (Rs < 268) {
return 2;
}
else if (Rs >= 268) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
return 4;
}
}
}
else if (CsDiff >= 1) {
if (ssDCval < 59) {
if (SC < 1460) {
if (SC < 1091) {
return 4;
}
else if (SC >= 1091) {
if (posBalance < 44) {
if (negBalance < 33) {
return 2;
}
else if (negBalance >= 33) {
return 4;
}
}
else if (posBalance >= 44) {
if (Rs < 517) {
if (SC < 1273) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (SC >= 1273) {
if (Cs < 1015) {
return 1;
}
else if (Cs >= 1015) {
return 2;
}
}
}
else if (Rs >= 517) {
return 2;
}
}
}
}
else if (SC >= 1460) {
if (Cs < 1015) {
return 1;
}
else if (Cs >= 1015) {
return 4;
}
}
}
else if (ssDCval >= 59) {
return 1;
}
}
}
}
else if (Rs >= 531) {
if (MVDiff < 1) {
if (Cs < 931) {
if (tcor < 97) {
if (ssDCval < 55) {
if (Rs < 643) {
return 1;
}
else if (Rs >= 643) {
return 4;
}
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (tcor >= 97) {
return 8;
}
}
else if (Cs >= 931) {
if (Rs < 637) {
if (Cs < 1060) {
return 1;
}
else if (Cs >= 1060) {
return 4;
}
}
else if (Rs >= 637) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (Rs < 835) {
if (negBalance < 27) {
return 1;
}
else if (negBalance >= 27) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
return 1;
}
}
}
else if (Rs >= 835) {
return 8;
}
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 17) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
return 1;
}
}
else if (negBalance >= 17) {
if (negBalance < 35) {
if (RsDiff < 2) {
if (SC < 1382) {
if (tcor < 96) {
if (Cs < 572) {
return 1;
}
else if (Cs >= 572) {
return 8;
}
}
else if (tcor >= 96) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
}
}
else if (SC >= 1382) {
if (refDCval < 58) {
if (Rs < 716) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (Rs < 701) {
return 8;
}
else if (Rs >= 701) {
return 4;
}
}
}
else if (Rs >= 716) {
if (Cs < 933) {
return 8;
}
else if (Cs >= 933) {
return 1;
}
}
}
else if (refDCval >= 58) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 1;
}
}
}
}
else if (RsDiff >= 2) {
if (MVDiff < 11) {
if (tcor < 92) {
if (Rs < 719) {
return 8;
}
else if (Rs >= 719) {
if (Rs < 790) {
return 4;
}
else if (Rs >= 790) {
return 1;
}
}
}
else if (tcor >= 92) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 754) {
return 4;
}
else if (Rs >= 754) {
return 8;
}
}
}
}
else if (MVDiff >= 11) {
if (SC < 1372) {
return 1;
}
else if (SC >= 1372) {
return 4;
}
}
}
}
else if (negBalance >= 35) {
if (AFD < 8) {
if (posBalance < 53) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
if (Rs < 818) {
if (MVDiff < 26) {
if (mu_mv_mag_sq < 92) {
return 8;
}
else if (mu_mv_mag_sq >= 92) {
if (Rs < 732) {
return 4;
}
else if (Rs >= 732) {
return 8;
}
}
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (Rs >= 818) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 8;
}
}
}
else if (AFD >= 8) {
if (Cs < 635) {
return 1;
}
else if (Cs >= 635) {
if (Cs < 866) {
return 8;
}
else if (Cs >= 866) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 843) {
if (ssDCval < 49) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (ssDCval >= 49) {
if (negBalance < 45) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 51) {
if (Rs < 896) {
return 1;
}
else if (Rs >= 896) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (ssDCval >= 51) {
if (diffAFD < -1) {
if (ssDCval < 58) {
if (Cs < 1051) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (Cs >= 1051) {
return 4;
}
}
else if (ssDCval >= 58) {
if (Cs < 703) {
return 8;
}
else if (Cs >= 703) {
return 4;
}
}
}
else if (diffAFD >= -1) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (tcor < 95) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
else if (tcor >= 95) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1101) {
return 8;
}
else if (Cs >= 1101) {
return 1;
}
}
}
else if (negBalance >= 45) {
return 4;
}
}
}
}
else if (refDCval >= 63) {
if (RsCsDiff < 2) {
if (Cs < 710) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
if (ssDCval < 72) {
if (Rs < 721) {
if (Cs < 648) {
return 4;
}
else if (Cs >= 648) {
return 8;
}
}
else if (Rs >= 721) {
if (posBalance < 35) {
if (Rs < 928) {
return 8;
}
else if (Rs >= 928) {
return 1;
}
}
else if (posBalance >= 35) {
return 2;
}
}
}
else if (ssDCval >= 72) {
if (ssDCval < 85) {
if (mu_mv_mag_sq < 88) {
if (Rs < 366) {
return 1;
}
else if (Rs >= 366) {
return 2;
}
}
else if (mu_mv_mag_sq >= 88) {
return 8;
}
}
else if (ssDCval >= 85) {
if (Rs < 631) {
return 2;
}
else if (Rs >= 631) {
return 1;
}
}
}
}
}
else if (Cs >= 710) {
if (ssDCval < 72) {
if (Rs < 728) {
if (ssDCval < 66) {
if (Rs < 676) {
return 8;
}
else if (Rs >= 676) {
return 4;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (Rs >= 728) {
if (Rs < 855) {
if (Rs < 852) {
return 4;
}
else if (Rs >= 852) {
return 8;
}
}
else if (Rs >= 855) {
if (Cs < 811) {
return 2;
}
else if (Cs >= 811) {
return 1;
}
}
}
}
else if (ssDCval >= 72) {
if (Rs < 650) {
if (AFD < 3) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (AFD >= 3) {
if (Cs < 812) {
return 1;
}
else if (Cs >= 812) {
return 2;
}
}
}
else if (Rs >= 650) {
if (TSCindex < 3) {
if (posBalance < 29) {
return 2;
}
else if (posBalance >= 29) {
if (diffMVdiffVal < -2) {
if (Rs < 666) {
return 2;
}
else if (Rs >= 666) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (Rs < 719) {
return 2;
}
else if (Rs >= 719) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 2) {
if (ssDCval < 85) {
if (ssDCval < 75) {
if (RsCsDiff < 23) {
if (Rs < 769) {
if (mu_mv_mag_sq < 89) {
if (negBalance < 45) {
if (gchDC < 2) {
if (Cs < 975) {
if (refDCval < 64) {
if (Cs < 574) {
if (Rs < 292) {
return 4;
}
else if (Rs >= 292) {
return 1;
}
}
else if (Cs >= 574) {
return 2;
}
}
else if (refDCval >= 64) {
if (refDCval < 67) {
if (tcor < 96) {
if (Cs < 569) {
return 4;
}
else if (Cs >= 569) {
return 1;
}
}
else if (tcor >= 96) {
return 4;
}
}
else if (refDCval >= 67) {
if (Cs < 766) {
return 4;
}
else if (Cs >= 766) {
return 1;
}
}
}
}
else if (Cs >= 975) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (negBalance >= 45) {
if (Rs < 496) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (Rs >= 496) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (Rs < 651) {
if (Cs < 592) {
if (Rs < 490) {
if (Rs < 430) {
return 1;
}
else if (Rs >= 430) {
return 2;
}
}
else if (Rs >= 490) {
return 4;
}
}
else if (Cs >= 592) {
return 8;
}
}
else if (Rs >= 651) {
if (MVDiff < 2) {
if (ssDCval < 65) {
if (Rs < 692) {
return 4;
}
else if (Rs >= 692) {
if (Rs < 703) {
return 8;
}
else if (Rs >= 703) {
return 2;
}
}
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 126) {
return 1;
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
return 8;
}
}
}
}
}
else if (Rs >= 769) {
if (tcor < 98) {
if (negBalance < 40) {
if (refDCval < 70) {
if (Cs < 815) {
if (Rs < 854) {
if (Rs < 802) {
return 1;
}
else if (Rs >= 802) {
return 2;
}
}
else if (Rs >= 854) {
return 1;
}
}
else if (Cs >= 815) {
if (tcor < 96) {
if (Rs < 831) {
if (Cs < 1110) {
return 4;
}
else if (Cs >= 1110) {
return 2;
}
}
else if (Rs >= 831) {
return 4;
}
}
else if (tcor >= 96) {
if (mu_mv_mag_sq < 103) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (posBalance < 51) {
if (Rs < 980) {
return 2;
}
else if (Rs >= 980) {
return 1;
}
}
else if (posBalance >= 51) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 103) {
return 4;
}
}
}
}
else if (refDCval >= 70) {
return 4;
}
}
else if (negBalance >= 40) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Cs < 552) {
return 4;
}
else if (Cs >= 552) {
return 1;
}
}
}
}
else if (tcor >= 98) {
if (Cs < 694) {
if (posBalance < 38) {
if (Rs < 898) {
return 4;
}
else if (Rs >= 898) {
return 1;
}
}
else if (posBalance >= 38) {
return 2;
}
}
else if (Cs >= 694) {
if (posBalance < 45) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
if (Rs < 990) {
return 2;
}
else if (Rs >= 990) {
return 4;
}
}
}
else if (posBalance >= 45) {
if (Rs < 989) {
return 1;
}
else if (Rs >= 989) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 23) {
if (Rs < 813) {
if (Cs < 1253) {
return 2;
}
else if (Cs >= 1253) {
return 1;
}
}
else if (Rs >= 813) {
if (Cs < 1079) {
if (tcor < 82) {
if (Rs < 911) {
return 2;
}
else if (Rs >= 911) {
return 1;
}
}
else if (tcor >= 82) {
return 8;
}
}
else if (Cs >= 1079) {
return 2;
}
}
}
}
else if (ssDCval >= 75) {
if (MVDiff < 2) {
if (ssDCval < 81) {
if (RsCsDiff < 4) {
if (negBalance < 28) {
return 8;
}
else if (negBalance >= 28) {
if (refDCval < 80) {
return 4;
}
else if (refDCval >= 80) {
return 8;
}
}
}
else if (RsCsDiff >= 4) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 8;
}
}
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (MVDiff >= 2) {
if (Cs < 880) {
if (negBalance < 37) {
if (Rs < 552) {
return 1;
}
else if (Rs >= 552) {
if (Cs < 867) {
return 2;
}
else if (Cs >= 867) {
return 1;
}
}
}
else if (negBalance >= 37) {
if (gchDC < 3) {
if (AFD < 21) {
if (MVDiff < 4) {
if (Rs < 592) {
return 8;
}
else if (Rs >= 592) {
return 4;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (AFD >= 21) {
return 1;
}
}
else if (gchDC >= 3) {
if (Cs < 831) {
return 4;
}
else if (Cs >= 831) {
return 1;
}
}
}
}
else if (Cs >= 880) {
if (Rs < 488) {
return 1;
}
else if (Rs >= 488) {
return 2;
}
}
}
}
}
else if (ssDCval >= 85) {
if (Cs < 861) {
if (negBalance < 41) {
return 2;
}
else if (negBalance >= 41) {
return 4;
}
}
else if (Cs >= 861) {
return 4;
}
}
}
}
}
else if (Rs >= 1002) {
if (ssDCval < 69) {
if (ssDCval < 52) {
if (AFD < 5) {
if (Cs < 617) {
return 1;
}
else if (Cs >= 617) {
if (mu_mv_mag_sq < 113) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 1299) {
return 8;
}
else if (Rs >= 1299) {
if (Rs < 1345) {
if (Rs < 1326) {
return 8;
}
else if (Rs >= 1326) {
return 8;
}
}
else if (Rs >= 1345) {
return 4;
}
}
}
}
}
else if (AFD >= 5) {
if (SC < 1960) {
return 1;
}
else if (SC >= 1960) {
return 4;
}
}
}
else if (ssDCval >= 52) {
if (negBalance < 41) {
if (posBalance < 43) {
if (mu_mv_mag_sq < 128) {
if (TSC < 4) {
if (MVDiff < 13) {
if (posBalance < 31) {
if (SC < 2027) {
return 1;
}
else if (SC >= 2027) {
return 8;
}
}
else if (posBalance >= 31) {
return 1;
}
}
else if (MVDiff >= 13) {
return 2;
}
}
else if (TSC >= 4) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1210) {
if (Rs < 1118) {
return 8;
}
else if (Rs >= 1118) {
return 1;
}
}
else if (Rs >= 1210) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (mu_mv_mag_sq < 122) {
if (MVDiff < 7) {
if (RsDiff < 4) {
if (SC < 1742) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
else if (SC >= 1742) {
return 1;
}
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (MVDiff >= 7) {
if (AFD < 7) {
if (Cs < 755) {
if (Cs < 669) {
return 8;
}
else if (Cs >= 669) {
return 1;
}
}
else if (Cs >= 755) {
return 8;
}
}
else if (AFD >= 7) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (SC < 1832) {
return 2;
}
else if (SC >= 1832) {
return 4;
}
}
}
}
}
else if (negBalance >= 41) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
if (Rs < 1095) {
return 4;
}
else if (Rs >= 1095) {
return 2;
}
}
}
}
}
else if (ssDCval >= 69) {
if (SC < 1876) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
if (Rs < 1046) {
return 8;
}
else if (Rs >= 1046) {
if (Cs < 698) {
return 1;
}
else if (Cs >= 698) {
return 8;
}
}
}
}
else if (SC >= 1876) {
if (mu_mv_mag_sq < 56) {
if (Cs < 881) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (Cs >= 881) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (Rs < 1164) {
return 8;
}
else if (Rs >= 1164) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (Cs < 638) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 638) {
if (negBalance < 30) {
if (ssDCval < 76) {
if (TSC < 4) {
if (Rs < 1458) {
return 1;
}
else if (Rs >= 1458) {
return 8;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (negBalance >= 30) {
if (tcor < 99) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 4;
}
}
else if (tcor >= 99) {
if (Cs < 893) {
return 8;
}
else if (Cs >= 893) {
return 4;
}
}
}
}
}
}
}
}
}
else if (SC >= 2150) {
if (Rs < 1832) {
if (Cs < 1024) {
if (mu_mv_mag_sq < 128) {
if (RsDiff < 1) {
if (refDCval < 69) {
if (Cs < 999) {
return 1;
}
else if (Cs >= 999) {
return 8;
}
}
else if (refDCval >= 69) {
return 8;
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 87) {
if (negBalance < 34) {
if (ssDCval < 43) {
return 4;
}
else if (ssDCval >= 43) {
if (SC < 2644) {
if (Cs < 620) {
if (Cs < 614) {
return 1;
}
else if (Cs >= 614) {
return 2;
}
}
else if (Cs >= 620) {
return 1;
}
}
else if (SC >= 2644) {
return 4;
}
}
}
else if (negBalance >= 34) {
if (ssDCval < 71) {
if (AFD < 11) {
if (negBalance < 39) {
if (Rs < 1347) {
if (Rs < 1293) {
return 8;
}
else if (Rs >= 1293) {
return 1;
}
}
else if (Rs >= 1347) {
if (Rs < 1474) {
return 4;
}
else if (Rs >= 1474) {
return 2;
}
}
}
else if (negBalance >= 39) {
return 8;
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (ssDCval >= 71) {
if (Rs < 1509) {
return 8;
}
else if (Rs >= 1509) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (mcTcor < 99) {
if (Rs < 1774) {
if (Cs < 763) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (Cs >= 763) {
if (negBalance < 19) {
if (MVDiff < 1) {
if (Rs < 1576) {
return 8;
}
else if (Rs >= 1576) {
return 8;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (negBalance >= 19) {
return 8;
}
}
}
else if (Rs >= 1774) {
if (ssDCval < 78) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (ssDCval >= 78) {
return 4;
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 69) {
if (ssDCval < 61) {
if (mu_mv_mag_sq < 123) {
if (SC < 2409) {
return 1;
}
else if (SC >= 2409) {
return 4;
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1418) {
return 8;
}
else if (Rs >= 1418) {
return 4;
}
}
}
else if (ssDCval >= 61) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Rs < 1649) {
return 8;
}
else if (Rs >= 1649) {
return 4;
}
}
}
}
else if (ssDCval >= 69) {
if (posBalance < 41) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
if (Rs < 1530) {
return 8;
}
else if (Rs >= 1530) {
return 4;
}
}
}
else if (posBalance >= 41) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1713) {
if (Rs < 1259) {
return 1;
}
else if (Rs >= 1259) {
if (TSCindex < 2) {
if (Cs < 943) {
return 8;
}
else if (Cs >= 943) {
return 1;
}
}
else if (TSCindex >= 2) {
return 8;
}
}
}
else if (Rs >= 1713) {
return 1;
}
}
}
else if (Cs >= 1024) {
if (AFD < 3) {
if (ssDCval < 52) {
if (Rs < 1651) {
if (Cs < 1063) {
return 1;
}
else if (Cs >= 1063) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
if (negBalance < 23) {
if (negBalance < 20) {
return 8;
}
else if (negBalance >= 20) {
return 4;
}
}
else if (negBalance >= 23) {
return 8;
}
}
}
}
else if (Rs >= 1651) {
if (Cs < 1163) {
if (Rs < 1807) {
if (diffAFD < 1) {
if (ssDCval < 48) {
if (SC < 2808) {
if (Rs < 1735) {
if (Rs < 1707) {
if (negBalance < 18) {
if (Cs < 1094) {
return 8;
}
else if (Cs >= 1094) {
return 4;
}
}
else if (negBalance >= 18) {
if (Cs < 1102) {
return 8;
}
else if (Cs >= 1102) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (Rs < 1675) {
return 4;
}
else if (Rs >= 1675) {
return 8;
}
}
}
}
}
else if (Rs >= 1707) {
if (Cs < 1067) {
if (RsDiff < 2) {
if (Rs < 1728) {
return 4;
}
else if (Rs >= 1728) {
return 4;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (Cs >= 1067) {
return 4;
}
}
}
else if (Rs >= 1735) {
return 8;
}
}
else if (SC >= 2808) {
if (diffAFD < -1) {
if (posBalance < 36) {
if (Cs < 1115) {
return 8;
}
else if (Cs >= 1115) {
return 8;
}
}
else if (posBalance >= 36) {
return 4;
}
}
else if (diffAFD >= -1) {
if (posBalance < 38) {
if (Cs < 1156) {
if (Cs < 1057) {
return 4;
}
else if (Cs >= 1057) {
if (Rs < 1753) {
if (Rs < 1702) {
if (CsDiff < 2) {
if (Cs < 1133) {
return 4;
}
else if (Cs >= 1133) {
if (RsDiff < 2) {
if (Rs < 1694) {
return 4;
}
else if (Rs >= 1694) {
return 4;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (Rs >= 1702) {
if (mu_mv_mag_sq < 127) {
if (Rs < 1737) {
return 4;
}
else if (Rs >= 1737) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
return 4;
}
}
}
else if (Rs >= 1753) {
if (negBalance < 19) {
return 8;
}
else if (negBalance >= 19) {
if (posBalance < 34) {
if (Cs < 1078) {
return 8;
}
else if (Cs >= 1078) {
if (Cs < 1095) {
return 4;
}
else if (Cs >= 1095) {
if (Cs < 1129) {
if (negBalance < 20) {
return 4;
}
else if (negBalance >= 20) {
if (Rs < 1771) {
return 4;
}
else if (Rs >= 1771) {
if (Rs < 1794) {
return 4;
}
else if (Rs >= 1794) {
return 4;
}
}
}
}
else if (Cs >= 1129) {
return 4;
}
}
}
}
else if (posBalance >= 34) {
return 4;
}
}
}
}
}
else if (Cs >= 1156) {
return 4;
}
}
else if (posBalance >= 38) {
if (negBalance < 19) {
if (Rs < 1745) {
if (Rs < 1720) {
return 4;
}
else if (Rs >= 1720) {
return 4;
}
}
else if (Rs >= 1745) {
if (Rs < 1751) {
return 8;
}
else if (Rs >= 1751) {
if (posBalance < 41) {
if (Cs < 1145) {
return 4;
}
else if (Cs >= 1145) {
return 4;
}
}
else if (posBalance >= 41) {
return 8;
}
}
}
}
else if (negBalance >= 19) {
return 8;
}
}
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (diffAFD >= 1) {
if (mu_mv_mag_sq < 127) {
return 4;
}
else if (mu_mv_mag_sq >= 127) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (Rs < 1753) {
return 8;
}
else if (Rs >= 1753) {
if (SC < 2924) {
if (Rs < 1782) {
return 8;
}
else if (Rs >= 1782) {
return 8;
}
}
else if (SC >= 2924) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1807) {
if (negBalance < 19) {
if (Rs < 1828) {
if (Cs < 1060) {
return 4;
}
else if (Cs >= 1060) {
return 8;
}
}
else if (Rs >= 1828) {
return 8;
}
}
else if (negBalance >= 19) {
return 8;
}
}
}
else if (Cs >= 1163) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
return 8;
}
}
}
}
else if (ssDCval >= 52) {
if (MVDiff < 7) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 70) {
if (diffAFD < -2) {
if (Rs < 1443) {
return 4;
}
else if (Rs >= 1443) {
return 8;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (ssDCval >= 70) {
if (Rs < 1571) {
if (ssDCval < 81) {
if (SC < 2386) {
return 8;
}
else if (SC >= 2386) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (Rs >= 1571) {
if (Rs < 1818) {
return 1;
}
else if (Rs >= 1818) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (AFD >= 3) {
if (Rs < 1420) {
if (RsDiff < 2) {
if (refDCval < 81) {
if (Rs < 1135) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (Rs < 1029) {
return 4;
}
else if (Rs >= 1029) {
return 8;
}
}
}
else if (Rs >= 1135) {
if (ssDCval < 70) {
if (mu_mv_mag_sq < 104) {
return 8;
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 1313) {
return 1;
}
else if (Rs >= 1313) {
return 4;
}
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
else if (refDCval >= 81) {
return 4;
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 120) {
if (Cs < 1120) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
else if (Cs >= 1120) {
if (diffRsCsdiff < 2) {
if (Rs < 1328) {
if (diffMVdiffVal < 1) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (SC < 2172) {
if (Rs < 948) {
return 4;
}
else if (Rs >= 948) {
return 8;
}
}
else if (SC >= 2172) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
if (mu_mv_mag_sq < 81) {
if (Rs < 963) {
return 4;
}
else if (Rs >= 963) {
return 2;
}
}
else if (mu_mv_mag_sq >= 81) {
return 4;
}
}
}
else if (Rs >= 1328) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 1034) {
return 2;
}
else if (Rs >= 1034) {
if (AFD < 4) {
if (Rs < 1224) {
return 4;
}
else if (Rs >= 1224) {
return 1;
}
}
else if (AFD >= 4) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (Cs < 1054) {
return 8;
}
else if (Cs >= 1054) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
if (Rs < 1154) {
return 4;
}
else if (Rs >= 1154) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1420) {
if (posBalance < 47) {
if (tcor < 95) {
return 2;
}
else if (tcor >= 95) {
if (SC < 2857) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (SC >= 2857) {
return 8;
}
}
}
else if (posBalance >= 47) {
if (Cs < 1041) {
if (Rs < 1641) {
return 8;
}
else if (Rs >= 1641) {
return 2;
}
}
else if (Cs >= 1041) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1832) {
if (TSC < 4) {
if (ssDCval < 82) {
if (posBalance < 39) {
if (ssDCval < 72) {
if (posBalance < 16) {
return 8;
}
else if (posBalance >= 16) {
if (mu_mv_mag_sq < 103) {
if (Cs < 1081) {
return 4;
}
else if (Cs >= 1081) {
if (Cs < 1098) {
return 1;
}
else if (Cs >= 1098) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (SC < 2830) {
return 1;
}
else if (SC >= 2830) {
if (Rs < 2180) {
if (Rs < 1936) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (Cs < 1070) {
if (Cs < 1001) {
return 1;
}
else if (Cs >= 1001) {
if (SC < 2921) {
return 4;
}
else if (SC >= 2921) {
if (SC < 2934) {
return 1;
}
else if (SC >= 2934) {
return 4;
}
}
}
}
else if (Cs >= 1070) {
return 1;
}
}
}
else if (Rs >= 1936) {
if (Cs < 887) {
return 1;
}
else if (Cs >= 887) {
return 4;
}
}
}
else if (Rs >= 2180) {
if (Rs < 2265) {
return 8;
}
else if (Rs >= 2265) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (posBalance >= 39) {
if (mu_mv_mag_sq < 109) {
if (Cs < 894) {
return 1;
}
else if (Cs >= 894) {
if (Cs < 1051) {
return 4;
}
else if (Cs >= 1051) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (posBalance < 45) {
if (ssDCval < 53) {
if (Rs < 1986) {
return 4;
}
else if (Rs >= 1986) {
return 8;
}
}
else if (ssDCval >= 53) {
if (Cs < 1083) {
return 1;
}
else if (Cs >= 1083) {
return 8;
}
}
}
else if (posBalance >= 45) {
return 8;
}
}
}
}
else if (ssDCval >= 82) {
if (MVDiff < 2) {
if (Cs < 939) {
if (Rs < 2195) {
return 4;
}
else if (Rs >= 2195) {
return 8;
}
}
else if (Cs >= 939) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Cs < 994) {
return 1;
}
else if (Cs >= 994) {
return 8;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 912) {
if (ssDCval < 71) {
if (posBalance < 45) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (Rs < 2179) {
return 2;
}
else if (Rs >= 2179) {
return 4;
}
}
}
else if (posBalance >= 45) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 2;
}
}
}
else if (ssDCval >= 71) {
if (RsCsDiff < 56) {
if (mu_mv_mag_sq < 93) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (diffRsCsdiff < -16) {
if (diffRsCsdiff < -21) {
return 1;
}
else if (diffRsCsdiff >= -21) {
return 2;
}
}
else if (diffRsCsdiff >= -16) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
if (Rs < 2256) {
return 4;
}
else if (Rs >= 2256) {
return 4;
}
}
}
}
else if (RsCsDiff >= 56) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
if (Rs < 2021) {
return 8;
}
else if (Rs >= 2021) {
return 1;
}
}
}
}
}
else if (Cs >= 912) {
if (negBalance < 45) {
if (Rs < 1974) {
if (Rs < 1864) {
return 4;
}
else if (Rs >= 1864) {
if (Cs < 1171) {
return 8;
}
else if (Cs >= 1171) {
return 2;
}
}
}
else if (Rs >= 1974) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 4;
}
}
}
else if (negBalance >= 45) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 88) {
if (mu_mv_mag_sq < 103) {
if (SC < 1310) {
if (MVDiff < 7) {
if (Cs < 760) {
if (Rs < 522) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (CsDiff < 4) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (AFD < 8) {
if (Cs < 628) {
return 8;
}
else if (Cs >= 628) {
if (Cs < 668) {
return 2;
}
else if (Cs >= 668) {
return 1;
}
}
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (CsDiff >= 4) {
return 4;
}
}
}
else if (Rs >= 522) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (ssDCval < 114) {
if (Cs < 708) {
return 1;
}
else if (Cs >= 708) {
return 2;
}
}
else if (ssDCval >= 114) {
if (ssDCval < 124) {
if (SC < 1274) {
if (Cs < 535) {
return 8;
}
else if (Cs >= 535) {
return 2;
}
}
else if (SC >= 1274) {
return 8;
}
}
else if (ssDCval >= 124) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 664) {
return 2;
}
else if (Rs >= 664) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 760) {
if (Rs < 430) {
return 4;
}
else if (Rs >= 430) {
if (Rs < 433) {
return 1;
}
else if (Rs >= 433) {
return 4;
}
}
}
}
else if (MVDiff >= 7) {
if (negBalance < 70) {
if (Rs < 402) {
return 2;
}
else if (Rs >= 402) {
if (TSC < 4) {
if (ssDCval < 125) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
return 1;
}
}
else if (ssDCval >= 125) {
if (SC < 1184) {
return 8;
}
else if (SC >= 1184) {
return 1;
}
}
}
else if (TSC >= 4) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Cs < 553) {
if (Rs < 418) {
return 4;
}
else if (Rs >= 418) {
return 1;
}
}
else if (Cs >= 553) {
if (SC < 1235) {
if (RsDiff < 2) {
if (Rs < 448) {
return 2;
}
else if (Rs >= 448) {
return 8;
}
}
else if (RsDiff >= 2) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
return 2;
}
}
}
else if (SC >= 1235) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 70) {
if (Rs < 603) {
return 8;
}
else if (Rs >= 603) {
return 1;
}
}
}
}
else if (SC >= 1310) {
if (Cs < 609) {
if (posBalance < 66) {
if (MVDiff < 26) {
if (negBalance < 38) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
return 1;
}
}
else if (negBalance >= 38) {
if (tcor < 95) {
if (mu_mv_mag_sq < 5) {
return 1;
}
else if (mu_mv_mag_sq >= 5) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (Rs < 1983) {
return 8;
}
else if (Rs >= 1983) {
return 1;
}
}
}
}
else if (tcor >= 95) {
return 8;
}
}
}
else if (MVDiff >= 26) {
if (TSC < 7) {
if (Rs < 1524) {
return 2;
}
else if (Rs >= 1524) {
return 8;
}
}
else if (TSC >= 7) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (mu_mv_mag_sq < 31) {
return 1;
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
}
}
}
else if (posBalance >= 66) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 1;
}
}
}
else if (Cs >= 609) {
if (MVDiff < 5) {
if (Rs < 667) {
if (ssDCval < 94) {
if (Rs < 523) {
return 8;
}
else if (Rs >= 523) {
return 2;
}
}
else if (ssDCval >= 94) {
if (TSC < 4) {
if (Rs < 464) {
return 4;
}
else if (Rs >= 464) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (TSC >= 4) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 651) {
return 1;
}
else if (Rs >= 651) {
return 4;
}
}
}
}
}
else if (Rs >= 667) {
if (ssDCval < 122) {
if (Rs < 1038) {
if (Cs < 712) {
return 8;
}
else if (Cs >= 712) {
if (ssDCval < 118) {
if (RsDiff < 2) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (ssDCval >= 118) {
if (Rs < 847) {
return 8;
}
else if (Rs >= 847) {
return 4;
}
}
}
}
else if (Rs >= 1038) {
if (ssDCval < 94) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
if (Cs < 803) {
if (Rs < 2023) {
return 4;
}
else if (Rs >= 2023) {
return 1;
}
}
else if (Cs >= 803) {
return 8;
}
}
}
else if (ssDCval >= 94) {
if (tcor < 91) {
if (Cs < 1082) {
return 8;
}
else if (Cs >= 1082) {
return 4;
}
}
else if (tcor >= 91) {
if (Rs < 1051) {
return 1;
}
else if (Rs >= 1051) {
if (negBalance < 23) {
return 1;
}
else if (negBalance >= 23) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 122) {
if (SC < 2628) {
if (RsCsDiff < 16) {
if (Cs < 1168) {
return 4;
}
else if (Cs >= 1168) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (ssDCval < 127) {
return 4;
}
else if (ssDCval >= 127) {
return 8;
}
}
}
else if (SC >= 2628) {
if (Rs < 1949) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
return 2;
}
}
else if (Rs >= 1949) {
if (Rs < 2051) {
return 4;
}
else if (Rs >= 2051) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 5) {
if (refDCval < 102) {
if (TSC < 5) {
if (mu_mv_mag_sq < 89) {
if (Rs < 619) {
return 2;
}
else if (Rs >= 619) {
return 1;
}
}
else if (mu_mv_mag_sq >= 89) {
if (Cs < 844) {
return 4;
}
else if (Cs >= 844) {
if (Rs < 1556) {
return 8;
}
else if (Rs >= 1556) {
return 1;
}
}
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 54) {
if (RsDiff < 5) {
if (Cs < 919) {
return 8;
}
else if (Cs >= 919) {
if (Cs < 1206) {
return 4;
}
else if (Cs >= 1206) {
return 8;
}
}
}
else if (RsDiff >= 5) {
if (Rs < 2000) {
if (mu_mv_mag_sq < 28) {
if (diffMVdiffVal < 1) {
if (gchDC < 3) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 2;
}
}
else if (gchDC >= 3) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
if (diffRsCsdiff < -12) {
if (Rs < 1608) {
return 1;
}
else if (Rs >= 1608) {
return 2;
}
}
else if (diffRsCsdiff >= -12) {
if (SC < 2068) {
return 1;
}
else if (SC >= 2068) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 28) {
if (ssDCval < 92) {
if (Rs < 1082) {
return 8;
}
else if (Rs >= 1082) {
return 2;
}
}
else if (ssDCval >= 92) {
if (Rs < 1661) {
return 4;
}
else if (Rs >= 1661) {
return 1;
}
}
}
}
else if (Rs >= 2000) {
if (refDCval < 91) {
return 1;
}
else if (refDCval >= 91) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (diffRsCsdiff < -18) {
return 2;
}
else if (diffRsCsdiff >= -18) {
return 8;
}
}
}
}
else if (refDCval >= 102) {
if (AFD < 9) {
if (ssDCval < 161) {
if (negBalance < 28) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (negBalance >= 28) {
if (AFD < 2) {
if (Rs < 854) {
return 8;
}
else if (Rs >= 854) {
return 4;
}
}
else if (AFD >= 2) {
if (SC < 1903) {
if (refDCval < 139) {
if (Cs < 687) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (Cs >= 687) {
if (Cs < 1219) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Cs < 742) {
return 2;
}
else if (Cs >= 742) {
return 8;
}
}
}
else if (Cs >= 1219) {
return 1;
}
}
}
else if (refDCval >= 139) {
return 1;
}
}
else if (SC >= 1903) {
return 8;
}
}
}
}
else if (ssDCval >= 161) {
if (Rs < 737) {
return 4;
}
else if (Rs >= 737) {
return 2;
}
}
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1955) {
if (posBalance < 65) {
if (tcor < 94) {
if (SC < 1924) {
if (diffRsCsdiff < 4) {
return 1;
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
else if (SC >= 1924) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
}
else if (tcor >= 94) {
if (Cs < 1071) {
return 8;
}
else if (Cs >= 1071) {
return 2;
}
}
}
else if (posBalance >= 65) {
if (Rs < 1510) {
return 2;
}
else if (Rs >= 1510) {
return 1;
}
}
}
else if (Rs >= 1955) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 145) {
if (diffRsCsdiff < -16) {
if (MVDiff < 40) {
if (CsDiff < 4) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
else if (CsDiff >= 4) {
if (MVDiff < 21) {
if (CsDiff < 7) {
if (refDCval < 122) {
return 1;
}
else if (refDCval >= 122) {
return 8;
}
}
else if (CsDiff >= 7) {
if (Rs < 1164) {
return 1;
}
else if (Rs >= 1164) {
return 2;
}
}
}
else if (MVDiff >= 21) {
return 2;
}
}
}
else if (MVDiff >= 40) {
if (Cs < 821) {
return 4;
}
else if (Cs >= 821) {
return 8;
}
}
}
else if (diffRsCsdiff >= -16) {
if (AFD < 21) {
if (posBalance < 70) {
if (Rs < 2290) {
if (CsDiff < 7) {
if (RsCsDiff < 19) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
return 2;
}
}
else if (RsCsDiff >= 19) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (CsDiff >= 7) {
if (Cs < 1049) {
return 2;
}
else if (Cs >= 1049) {
return 8;
}
}
}
else if (Rs >= 2290) {
return 2;
}
}
else if (posBalance >= 70) {
return 2;
}
}
else if (AFD >= 21) {
if (Rs < 1956) {
return 2;
}
else if (Rs >= 1956) {
if (Rs < 2169) {
return 1;
}
else if (Rs >= 2169) {
return 8;
}
}
}
}
}
else if (ssDCval >= 145) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (mcTcor < 99) {
if (Rs < 1368) {
return 2;
}
else if (Rs >= 1368) {
return 8;
}
}
else if (mcTcor >= 99) {
if (Rs < 1525) {
return 8;
}
else if (Rs >= 1525) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (ssDCval < 133) {
if (Cs < 1234) {
if (Rs < 350) {
if (ssDCval < 129) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
if (ssDCval < 110) {
return 4;
}
else if (ssDCval >= 110) {
return 8;
}
}
}
else if (ssDCval >= 129) {
return 2;
}
}
else if (Rs >= 350) {
if (Cs < 948) {
if (refDCval < 97) {
if (mu_mv_mag_sq < 127) {
if (Rs < 1069) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (Rs < 932) {
return 1;
}
else if (Rs >= 932) {
return 8;
}
}
}
else if (Rs >= 1069) {
if (ssDCval < 92) {
if (ssDCval < 89) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (ssDCval >= 92) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 1493) {
return 8;
}
else if (Rs >= 1493) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 2155) {
if (posBalance < 25) {
return 4;
}
else if (posBalance >= 25) {
return 8;
}
}
else if (Rs >= 2155) {
if (Cs < 747) {
return 8;
}
else if (Cs >= 747) {
return 4;
}
}
}
}
else if (refDCval >= 97) {
if (SC < 2425) {
if (SC < 1360) {
if (mu_mv_mag_sq < 108) {
if (ssDCval < 124) {
if (Cs < 690) {
return 1;
}
else if (Cs >= 690) {
return 4;
}
}
else if (ssDCval >= 124) {
return 8;
}
}
else if (mu_mv_mag_sq >= 108) {
if (RsDiff < 1) {
if (Cs < 893) {
return 8;
}
else if (Cs >= 893) {
return 4;
}
}
else if (RsDiff >= 1) {
if (Cs < 773) {
if (ssDCval < 120) {
if (diffMVdiffVal < -9) {
return 8;
}
else if (diffMVdiffVal >= -9) {
if (Cs < 554) {
return 8;
}
else if (Cs >= 554) {
return 1;
}
}
}
else if (ssDCval >= 120) {
return 8;
}
}
else if (Cs >= 773) {
return 8;
}
}
}
}
else if (SC >= 1360) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (Rs < 878) {
return 4;
}
else if (Rs >= 878) {
return 1;
}
}
}
}
else if (SC >= 2425) {
if (mu_mv_mag_sq < 114) {
if (Cs < 660) {
return 1;
}
else if (Cs >= 660) {
return 8;
}
}
else if (mu_mv_mag_sq >= 114) {
if (SC < 2762) {
return 1;
}
else if (SC >= 2762) {
return 8;
}
}
}
}
}
else if (Cs >= 948) {
if (diffRsCsdiff < -18) {
return 2;
}
else if (diffRsCsdiff >= -18) {
if (ssDCval < 121) {
if (Cs < 1220) {
if (tcor < 95) {
if (Rs < 1055) {
return 4;
}
else if (Rs >= 1055) {
return 8;
}
}
else if (tcor >= 95) {
return 8;
}
}
else if (Cs >= 1220) {
if (negBalance < 34) {
if (Rs < 1213) {
return 1;
}
else if (Rs >= 1213) {
if (Rs < 1910) {
return 2;
}
else if (Rs >= 1910) {
return 8;
}
}
}
else if (negBalance >= 34) {
return 8;
}
}
}
else if (ssDCval >= 121) {
if (Rs < 1488) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 1488) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1234) {
if (Rs < 659) {
return 1;
}
else if (Rs >= 659) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
if (ssDCval < 145) {
return 1;
}
else if (ssDCval >= 145) {
if (Cs < 858) {
return 2;
}
else if (Cs >= 858) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2339) {
if (MVDiff < 18) {
if (refDCval < 109) {
if (RsDiff < 6) {
if (Rs < 3759) {
if (Cs < 1026) {
if (Cs < 669) {
if (Cs < 515) {
if (CsDiff < 1) {
if (Cs < 250) {
return 1;
}
else if (Cs >= 250) {
if (Cs < 491) {
return 2;
}
else if (Cs >= 491) {
return 8;
}
}
}
else if (CsDiff >= 1) {
if (Rs < 2488) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
}
else if (Rs >= 2488) {
if (mcTcor < 96) {
if (Rs < 2797) {
return 4;
}
else if (Rs >= 2797) {
return 8;
}
}
else if (mcTcor >= 96) {
return 8;
}
}
}
}
else if (Cs >= 515) {
if (ssDCval < 86) {
if (MVDiff < 7) {
if (mu_mv_mag_sq < 123) {
return 1;
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
else if (MVDiff >= 7) {
if (Cs < 582) {
return 4;
}
else if (Cs >= 582) {
return 1;
}
}
}
else if (ssDCval >= 86) {
if (negBalance < 25) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (negBalance >= 25) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
}
else if (Cs >= 669) {
if (MVDiff < 8) {
if (posBalance < 33) {
if (ssDCval < 79) {
if (Rs < 2350) {
return 4;
}
else if (Rs >= 2350) {
if (Cs < 926) {
return 8;
}
else if (Cs >= 926) {
return 8;
}
}
}
else if (ssDCval >= 79) {
if (posBalance < 22) {
return 4;
}
else if (posBalance >= 22) {
if (Rs < 2878) {
return 4;
}
else if (Rs >= 2878) {
return 8;
}
}
}
}
else if (posBalance >= 33) {
if (refDCval < 81) {
if (Rs < 3128) {
if (negBalance < 22) {
if (Cs < 885) {
if (TSC < 6) {
if (Rs < 2644) {
return 8;
}
else if (Rs >= 2644) {
return 8;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (Cs >= 885) {
return 4;
}
}
else if (negBalance >= 22) {
if (mcTcor < 96) {
return 4;
}
else if (mcTcor >= 96) {
if (SC < 3861) {
if (Cs < 790) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
return 2;
}
}
else if (Cs >= 790) {
return 2;
}
}
else if (SC >= 3861) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 4;
}
}
}
}
}
else if (Rs >= 3128) {
return 4;
}
}
else if (refDCval >= 81) {
if (tcor < 96) {
if (Cs < 917) {
return 1;
}
else if (Cs >= 917) {
if (Cs < 923) {
return 4;
}
else if (Cs >= 923) {
return 8;
}
}
}
else if (tcor >= 96) {
if (diffAFD < 2) {
if (RsDiff < 3) {
if (Rs < 3317) {
return 8;
}
else if (Rs >= 3317) {
return 2;
}
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
}
}
else if (MVDiff >= 8) {
if (tcor < 94) {
if (RsCsDiff < 48) {
return 1;
}
else if (RsCsDiff >= 48) {
return 4;
}
}
else if (tcor >= 94) {
if (Cs < 775) {
return 4;
}
else if (Cs >= 775) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (posBalance < 50) {
if (SC < 4297) {
return 4;
}
else if (SC >= 4297) {
if (Rs < 3559) {
if (Rs < 3439) {
return 4;
}
else if (Rs >= 3439) {
return 4;
}
}
else if (Rs >= 3559) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (Rs < 2987) {
return 8;
}
else if (Rs >= 2987) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 1026) {
if (RsCsDiff < 5) {
if (negBalance < 23) {
if (SC < 5003) {
if (negBalance < 20) {
return 8;
}
else if (negBalance >= 20) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Rs < 2998) {
return 8;
}
else if (Rs >= 2998) {
return 2;
}
}
}
}
else if (SC >= 5003) {
return 4;
}
}
else if (negBalance >= 23) {
if (Cs < 1120) {
if (Rs < 2930) {
return 8;
}
else if (Rs >= 2930) {
return 2;
}
}
else if (Cs >= 1120) {
return 8;
}
}
}
else if (RsCsDiff >= 5) {
if (ssDCval < 103) {
if (ssDCval < 102) {
if (ssDCval < 77) {
if (refDCval < 63) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 75) {
return 2;
}
else if (mu_mv_mag_sq >= 75) {
if (mu_mv_mag_sq < 103) {
return 4;
}
else if (mu_mv_mag_sq >= 103) {
if (Cs < 1255) {
return 8;
}
else if (Cs >= 1255) {
return 4;
}
}
}
}
else if (ssDCval >= 56) {
if (Cs < 1136) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
else if (Cs >= 1136) {
return 8;
}
}
}
else if (refDCval >= 63) {
if (posBalance < 44) {
if (mu_mv_mag_sq < 118) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Rs < 2848) {
if (Rs < 2531) {
return 4;
}
else if (Rs >= 2531) {
return 2;
}
}
else if (Rs >= 2848) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 2734) {
return 4;
}
else if (Rs >= 2734) {
return 1;
}
}
}
else if (posBalance >= 44) {
if (negBalance < 20) {
if (Rs < 3073) {
return 8;
}
else if (Rs >= 3073) {
return 8;
}
}
else if (negBalance >= 20) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 4;
}
}
}
}
}
else if (ssDCval >= 77) {
if (ssDCval < 90) {
if (Rs < 3095) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 2;
}
}
else if (Rs >= 3095) {
if (Rs < 3305) {
return 4;
}
else if (Rs >= 3305) {
return 2;
}
}
}
else if (ssDCval >= 90) {
if (SC < 3787) {
return 8;
}
else if (SC >= 3787) {
if (ssDCval < 95) {
if (Rs < 2880) {
return 1;
}
else if (Rs >= 2880) {
return 2;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
else if (ssDCval >= 102) {
return 4;
}
}
else if (ssDCval >= 103) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
return 2;
}
}
}
}
}
else if (Rs >= 3759) {
if (Cs < 1276) {
if (posBalance < 62) {
if (diffRsCsdiff < 17) {
if (SC < 5718) {
return 4;
}
else if (SC >= 5718) {
return 2;
}
}
else if (diffRsCsdiff >= 17) {
return 1;
}
}
else if (posBalance >= 62) {
return 2;
}
}
else if (Cs >= 1276) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 42) {
if (Rs < 4075) {
return 4;
}
else if (Rs >= 4075) {
return 1;
}
}
else if (posBalance >= 42) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 3900) {
return 4;
}
else if (Rs >= 3900) {
if (Rs < 4576) {
return 8;
}
else if (Rs >= 4576) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 6) {
if (refDCval < 90) {
if (posBalance < 48) {
if (negBalance < 51) {
if (Rs < 3734) {
if (refDCval < 65) {
if (MVDiff < 3) {
if (Rs < 2539) {
return 8;
}
else if (Rs >= 2539) {
if (posBalance < 37) {
if (Rs < 2858) {
return 8;
}
else if (Rs >= 2858) {
return 4;
}
}
else if (posBalance >= 37) {
return 2;
}
}
}
else if (MVDiff >= 3) {
if (CsDiff < 5) {
if (negBalance < 43) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Rs < 2359) {
return 2;
}
else if (Rs >= 2359) {
if (negBalance < 41) {
return 4;
}
else if (negBalance >= 41) {
if (Cs < 1095) {
return 4;
}
else if (Cs >= 1095) {
return 8;
}
}
}
}
}
else if (negBalance >= 43) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (Cs < 1194) {
if (AFD < 17) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
else if (AFD >= 17) {
return 2;
}
}
else if (Cs >= 1194) {
if (mu_mv_mag_sq < 59) {
return 2;
}
else if (mu_mv_mag_sq >= 59) {
return 4;
}
}
}
}
}
else if (refDCval >= 65) {
if (mu_mv_mag_sq < 78) {
if (ssDCval < 72) {
if (Cs < 1055) {
if (Rs < 2497) {
return 2;
}
else if (Rs >= 2497) {
if (diffMVdiffVal < 3) {
if (posBalance < 41) {
if (Cs < 911) {
if (Cs < 764) {
return 8;
}
else if (Cs >= 764) {
return 4;
}
}
else if (Cs >= 911) {
return 1;
}
}
else if (posBalance >= 41) {
if (TSC < 9) {
if (diffRsCsdiff < -116) {
if (Rs < 3054) {
return 1;
}
else if (Rs >= 3054) {
return 4;
}
}
else if (diffRsCsdiff >= -116) {
return 4;
}
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (diffMVdiffVal >= 3) {
if (Cs < 968) {
return 1;
}
else if (Cs >= 968) {
return 4;
}
}
}
}
else if (Cs >= 1055) {
if (posBalance < 45) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 42) {
if (Rs < 3319) {
return 4;
}
else if (Rs >= 3319) {
return 2;
}
}
else if (mu_mv_mag_sq >= 42) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
else if (posBalance >= 45) {
return 2;
}
}
}
else if (ssDCval >= 72) {
if (SC < 3779) {
if (Cs < 985) {
if (posBalance < 35) {
if (Rs < 2636) {
return 4;
}
else if (Rs >= 2636) {
return 8;
}
}
else if (posBalance >= 35) {
if (RsCsDiff < 136) {
return 1;
}
else if (RsCsDiff >= 136) {
if (Rs < 2847) {
return 2;
}
else if (Rs >= 2847) {
return 4;
}
}
}
}
else if (Cs >= 985) {
if (Cs < 1183) {
return 4;
}
else if (Cs >= 1183) {
return 2;
}
}
}
else if (SC >= 3779) {
if (negBalance < 44) {
if (TSC < 7) {
if (TSC < 5) {
if (Rs < 3295) {
return 1;
}
else if (Rs >= 3295) {
return 4;
}
}
else if (TSC >= 5) {
return 4;
}
}
else if (TSC >= 7) {
if (Cs < 1147) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
return 4;
}
}
else if (Cs >= 1147) {
return 4;
}
}
}
else if (negBalance >= 44) {
if (RsCsDiff < 59) {
if (Rs < 3113) {
return 8;
}
else if (Rs >= 3113) {
return 4;
}
}
else if (RsCsDiff >= 59) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
if (Cs < 752) {
return 4;
}
else if (Cs >= 752) {
if (MVDiff < 12) {
if (ssDCval < 69) {
if (Cs < 1045) {
return 1;
}
else if (Cs >= 1045) {
return 2;
}
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (MVDiff >= 12) {
return 4;
}
}
}
}
}
}
else if (Rs >= 3734) {
if (mu_mv_mag_sq < 100) {
if (Rs < 3818) {
if (diffMVdiffVal < -3) {
if (Cs < 674) {
return 4;
}
else if (Cs >= 674) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
if (Cs < 1155) {
return 4;
}
else if (Cs >= 1155) {
return 2;
}
}
}
else if (Rs >= 3818) {
return 4;
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 1062) {
return 4;
}
else if (Cs >= 1062) {
return 1;
}
}
}
}
else if (negBalance >= 51) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (CsDiff < 3) {
if (Rs < 3378) {
if (ssDCval < 76) {
if (diffMVdiffVal < -5) {
if (Rs < 2794) {
return 2;
}
else if (Rs >= 2794) {
return 4;
}
}
else if (diffMVdiffVal >= -5) {
return 2;
}
}
else if (ssDCval >= 76) {
if (TSC < 6) {
if (Cs < 632) {
return 1;
}
else if (Cs >= 632) {
return 2;
}
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (Rs >= 3378) {
if (Cs < 688) {
return 4;
}
else if (Cs >= 688) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 60) {
if (Rs < 7113) {
if (Rs < 2604) {
return 2;
}
else if (Rs >= 2604) {
return 4;
}
}
else if (Rs >= 7113) {
return 2;
}
}
else if (refDCval >= 60) {
if (negBalance < 37) {
if (posBalance < 56) {
if (AFD < 8) {
if (Rs < 2587) {
return 1;
}
else if (Rs >= 2587) {
return 2;
}
}
else if (AFD >= 8) {
if (negBalance < 35) {
if (Cs < 937) {
return 4;
}
else if (Cs >= 937) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 2;
}
}
}
}
else if (negBalance >= 35) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 2533) {
return 2;
}
else if (Rs >= 2533) {
if (Rs < 3293) {
return 4;
}
else if (Rs >= 3293) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 56) {
if (Rs < 4604) {
if (mcTcor < 99) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
if (Rs < 2764) {
return 1;
}
else if (Rs >= 2764) {
return 2;
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (Rs >= 4604) {
return 4;
}
}
}
else if (negBalance >= 37) {
if (tcor < 77) {
if (RsCsDiff < 246) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
if (Rs < 3119) {
return 2;
}
else if (Rs >= 3119) {
return 1;
}
}
}
else if (RsCsDiff >= 246) {
if (RsDiff < 29) {
return 2;
}
else if (RsDiff >= 29) {
return 4;
}
}
}
else if (tcor >= 77) {
if (mcTcor < 93) {
if (Rs < 2506) {
if (Cs < 698) {
return 2;
}
else if (Cs >= 698) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
return 4;
}
}
}
else if (Rs >= 2506) {
if (refDCval < 64) {
return 4;
}
else if (refDCval >= 64) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (diffRsCsdiff < -191) {
if (Cs < 1073) {
return 4;
}
else if (Cs >= 1073) {
return 2;
}
}
else if (diffRsCsdiff >= -191) {
if (ssDCval < 84) {
if (Rs < 3782) {
return 2;
}
else if (Rs >= 3782) {
if (Cs < 975) {
return 4;
}
else if (Cs >= 975) {
return 2;
}
}
}
else if (ssDCval >= 84) {
if (negBalance < 42) {
if (Rs < 2869) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (Rs < 2405) {
return 2;
}
else if (Rs >= 2405) {
return 1;
}
}
}
else if (Rs >= 2869) {
return 2;
}
}
else if (negBalance >= 42) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 90) {
if (RsCsDiff < 121) {
if (mcTcor < 98) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (Cs < 726) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 4043) {
return 8;
}
else if (Rs >= 4043) {
return 2;
}
}
}
else if (Cs >= 726) {
if (TSC < 6) {
if (Rs < 4214) {
if (Cs < 882) {
return 4;
}
else if (Cs >= 882) {
return 8;
}
}
else if (Rs >= 4214) {
return 2;
}
}
else if (TSC >= 6) {
if (Cs < 1143) {
if (SC < 4498) {
return 4;
}
else if (SC >= 4498) {
if (MVDiff < 8) {
if (mcTcor < 96) {
if (diffMVdiffVal < 2) {
if (Rs < 4179) {
if (negBalance < 30) {
if (Rs < 4119) {
return 4;
}
else if (Rs >= 4119) {
return 4;
}
}
else if (negBalance >= 30) {
return 4;
}
}
else if (Rs >= 4179) {
return 4;
}
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
else if (mcTcor >= 96) {
return 4;
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
else if (Cs >= 1143) {
if (Rs < 4125) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
}
else if (Rs >= 4125) {
if (MVDiff < 4) {
if (Rs < 4197) {
return 8;
}
else if (Rs >= 4197) {
return 8;
}
}
else if (MVDiff >= 4) {
if (Cs < 1267) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
if (mu_mv_mag_sq < 59) {
if (Rs < 4353) {
if (Rs < 4150) {
return 4;
}
else if (Rs >= 4150) {
return 4;
}
}
else if (Rs >= 4353) {
return 4;
}
}
else if (mu_mv_mag_sq >= 59) {
return 4;
}
}
}
else if (Cs >= 1267) {
return 4;
}
}
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (Cs < 966) {
return 2;
}
else if (Cs >= 966) {
if (Rs < 3538) {
return 8;
}
else if (Rs >= 3538) {
return 2;
}
}
}
}
else if (RsCsDiff >= 121) {
if (mcTcor < 70) {
return 1;
}
else if (mcTcor >= 70) {
if (Cs < 874) {
if (Rs < 2807) {
if (SC < 3278) {
if (tcor < 89) {
return 2;
}
else if (tcor >= 89) {
return 1;
}
}
else if (SC >= 3278) {
if (Cs < 686) {
return 8;
}
else if (Cs >= 686) {
if (Rs < 2684) {
return 4;
}
else if (Rs >= 2684) {
return 2;
}
}
}
}
else if (Rs >= 2807) {
return 2;
}
}
else if (Cs >= 874) {
if (tcor < 81) {
if (Rs < 3340) {
if (gchDC < 2) {
if (RsCsDiff < 143) {
if (Rs < 2660) {
return 4;
}
else if (Rs >= 2660) {
return 8;
}
}
else if (RsCsDiff >= 143) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (Rs < 2620) {
return 4;
}
else if (Rs >= 2620) {
return 2;
}
}
}
else if (Rs >= 3340) {
return 8;
}
}
else if (tcor >= 81) {
if (ssDCval < 106) {
if (diffAFD < -7) {
if (Rs < 3655) {
return 4;
}
else if (Rs >= 3655) {
return 2;
}
}
else if (diffAFD >= -7) {
if (Rs < 2535) {
return 8;
}
else if (Rs >= 2535) {
if (refDCval < 106) {
return 2;
}
else if (refDCval >= 106) {
if (Rs < 3445) {
return 2;
}
else if (Rs >= 3445) {
return 4;
}
}
}
}
}
else if (ssDCval >= 106) {
if (RsDiff < 15) {
return 4;
}
else if (RsDiff >= 15) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 3359) {
if (mcTcor < 98) {
if (RsDiff < 9) {
if (tcor < 92) {
if (posBalance < 40) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
else if (posBalance >= 40) {
return 8;
}
}
else if (tcor >= 92) {
if (RsDiff < 6) {
if (ssDCval < 135) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 2777) {
return 8;
}
else if (Rs >= 2777) {
return 2;
}
}
}
else if (ssDCval >= 135) {
if (refDCval < 137) {
if (Rs < 2370) {
return 8;
}
else if (Rs >= 2370) {
return 1;
}
}
else if (refDCval >= 137) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (Cs < 974) {
if (mu_mv_mag_sq < 59) {
return 1;
}
else if (mu_mv_mag_sq >= 59) {
return 2;
}
}
else if (Cs >= 974) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (Cs < 786) {
if (refDCval < 115) {
if (Rs < 3299) {
if (posBalance < 75) {
return 2;
}
else if (posBalance >= 75) {
return 1;
}
}
else if (Rs >= 3299) {
return 8;
}
}
else if (refDCval >= 115) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (diffMVdiffVal < 3) {
if (AFD < 31) {
if (mu_mv_mag_sq < 53) {
return 1;
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 2823) {
return 1;
}
else if (Rs >= 2823) {
return 2;
}
}
}
else if (AFD >= 31) {
return 2;
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 3235) {
return 2;
}
else if (Rs >= 3235) {
return 1;
}
}
}
}
}
else if (Cs >= 786) {
if (diffMVdiffVal < -6) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
if (negBalance < 48) {
if (Rs < 2873) {
if (Rs < 2611) {
return 1;
}
else if (Rs >= 2611) {
if (Rs < 2730) {
return 8;
}
else if (Rs >= 2730) {
return 4;
}
}
}
else if (Rs >= 2873) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (Rs < 3038) {
if (Cs < 1205) {
return 8;
}
else if (Cs >= 1205) {
return 4;
}
}
else if (Rs >= 3038) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (MVDiff < 4) {
if (refDCval < 115) {
if (SC < 3855) {
return 8;
}
else if (SC >= 3855) {
return 2;
}
}
else if (refDCval >= 115) {
return 2;
}
}
else if (MVDiff >= 4) {
if (Rs < 2697) {
if (ssDCval < 136) {
return 1;
}
else if (ssDCval >= 136) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 1;
}
}
}
else if (Rs >= 2697) {
if (TSC < 5) {
if (CsDiff < 3) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 2913) {
return 8;
}
else if (Rs >= 2913) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (Cs < 340) {
return 1;
}
else if (Cs >= 340) {
return 8;
}
}
}
else if (TSC >= 5) {
if (diffRsCsdiff < -49) {
if (Rs < 2972) {
return 8;
}
else if (Rs >= 2972) {
if (Rs < 3109) {
return 1;
}
else if (Rs >= 3109) {
return 2;
}
}
}
else if (diffRsCsdiff >= -49) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3359) {
if (MVDiff < 7) {
if (refDCval < 142) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (Rs < 3876) {
if (mu_mv_mag_sq < 31) {
if (TSC < 6) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
else if (Rs >= 3876) {
if (refDCval < 136) {
return 2;
}
else if (refDCval >= 136) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (Rs < 4418) {
return 1;
}
else if (Rs >= 4418) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 142) {
if (Rs < 3965) {
if (Rs < 3668) {
return 2;
}
else if (Rs >= 3668) {
if (Cs < 1200) {
if (Rs < 3775) {
if (CsDiff < 3) {
if (Rs < 3749) {
return 2;
}
else if (Rs >= 3749) {
return 1;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (Rs >= 3775) {
return 2;
}
}
else if (Cs >= 1200) {
return 1;
}
}
}
else if (Rs >= 3965) {
if (Rs < 6336) {
return 2;
}
else if (Rs >= 6336) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (mcTcor < 93) {
if (ssDCval < 134) {
if (Cs < 1279) {
return 2;
}
else if (Cs >= 1279) {
return 8;
}
}
else if (ssDCval >= 134) {
if (mu_mv_mag_sq < 3) {
if (Rs < 5047) {
return 1;
}
else if (Rs >= 5047) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
}
else if (mcTcor >= 93) {
if (mu_mv_mag_sq < 50) {
if (RsDiff < 20) {
if (TSC < 7) {
if (RsDiff < 5) {
if (Cs < 1251) {
return 1;
}
else if (Cs >= 1251) {
return 8;
}
}
else if (RsDiff >= 5) {
if (Rs < 4023) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (Rs >= 4023) {
return 2;
}
}
}
else if (TSC >= 7) {
if (negBalance < 54) {
if (ssDCval < 125) {
if (diffMVdiffVal < -10) {
return 1;
}
else if (diffMVdiffVal >= -10) {
return 2;
}
}
else if (ssDCval >= 125) {
if (CsDiff < 4) {
if (Rs < 5201) {
if (Cs < 918) {
return 2;
}
else if (Cs >= 918) {
return 4;
}
}
else if (Rs >= 5201) {
return 1;
}
}
else if (CsDiff >= 4) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (posBalance < 56) {
if (Rs < 4993) {
return 2;
}
else if (Rs >= 4993) {
return 1;
}
}
else if (posBalance >= 56) {
if (Cs < 476) {
return 2;
}
else if (Cs >= 476) {
return 4;
}
}
}
}
}
else if (RsDiff >= 20) {
if (diffMVdiffVal < -10) {
if (Cs < 989) {
if (RsCsDiff < 474) {
return 1;
}
else if (RsCsDiff >= 474) {
return 2;
}
}
else if (Cs >= 989) {
return 8;
}
}
else if (diffMVdiffVal >= -10) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (AFD < 8) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 2;
}
}
else if (AFD >= 8) {
if (posBalance < 65) {
return 2;
}
else if (posBalance >= 65) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 18) {
if (mcTcor < 69) {
if (MVDiff < 43) {
return 1;
}
else if (MVDiff >= 43) {
if (Cs < 951) {
return 8;
}
else if (Cs >= 951) {
if (AFD < 33) {
return 4;
}
else if (AFD >= 33) {
return 2;
}
}
}
}
else if (mcTcor >= 69) {
if (RsCsDiff < 279) {
if (refDCval < 78) {
if (TSC < 8) {
if (Rs < 3010) {
if (AFD < 12) {
if (posBalance < 34) {
if (Cs < 1231) {
return 1;
}
else if (Cs >= 1231) {
return 8;
}
}
else if (posBalance >= 34) {
if (Cs < 1095) {
if (negBalance < 34) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Cs < 348) {
return 8;
}
else if (Cs >= 348) {
return 4;
}
}
}
else if (negBalance >= 34) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
return 1;
}
}
}
else if (Cs >= 1095) {
if (Cs < 1164) {
return 2;
}
else if (Cs >= 1164) {
return 4;
}
}
}
}
else if (AFD >= 12) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 11) {
if (tcor < 79) {
return 4;
}
else if (tcor >= 79) {
if (Rs < 2415) {
return 4;
}
else if (Rs >= 2415) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 11) {
return 4;
}
}
else if (gchDC >= 2) {
if (negBalance < 38) {
return 2;
}
else if (negBalance >= 38) {
return 1;
}
}
}
}
else if (Rs >= 3010) {
if (Cs < 945) {
return 4;
}
else if (Cs >= 945) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (diffTSC < -1) {
if (Rs < 3268) {
return 2;
}
else if (Rs >= 3268) {
return 4;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
}
}
}
else if (TSC >= 8) {
if (Rs < 3586) {
if (refDCval < 71) {
if (TSC < 11) {
if (AFD < 16) {
if (mcTcor < 88) {
if (diffMVdiffVal < 3) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 2441) {
return 8;
}
else if (Rs >= 2441) {
return 2;
}
}
}
else if (mcTcor >= 88) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
return 4;
}
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (TSC >= 11) {
if (Cs < 619) {
return 4;
}
else if (Cs >= 619) {
return 1;
}
}
}
else if (refDCval >= 71) {
if (Rs < 2874) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (Rs < 2737) {
return 2;
}
else if (Rs >= 2737) {
return 1;
}
}
}
else if (Rs >= 2874) {
if (ssDCval < 73) {
return 4;
}
else if (ssDCval >= 73) {
return 2;
}
}
}
}
else if (Rs >= 3586) {
return 4;
}
}
}
else if (refDCval >= 78) {
if (mcTcor < 92) {
if (refDCval < 134) {
if (AFD < 24) {
if (RsDiff < 15) {
if (SC < 3270) {
if (tcor < 80) {
if (CsDiff < 5) {
if (diffRsCsdiff < -53) {
if (Rs < 2699) {
if (Rs < 2634) {
return 1;
}
else if (Rs >= 2634) {
return 4;
}
}
else if (Rs >= 2699) {
return 2;
}
}
else if (diffRsCsdiff >= -53) {
if (Cs < 293) {
return 4;
}
else if (Cs >= 293) {
return 8;
}
}
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (tcor >= 80) {
return 1;
}
}
else if (SC >= 3270) {
if (mu_mv_mag_sq < 47) {
if (diffRsCsdiff < -89) {
if (Cs < 1019) {
if (Rs < 3668) {
return 4;
}
else if (Rs >= 3668) {
return 2;
}
}
else if (Cs >= 1019) {
if (Rs < 3005) {
return 1;
}
else if (Rs >= 3005) {
return 8;
}
}
}
else if (diffRsCsdiff >= -89) {
if (AFD < 13) {
if (Cs < 1039) {
if (Rs < 2627) {
return 2;
}
else if (Rs >= 2627) {
return 1;
}
}
else if (Cs >= 1039) {
return 8;
}
}
else if (AFD >= 13) {
if (Rs < 2393) {
if (Rs < 2370) {
return 8;
}
else if (Rs >= 2370) {
return 2;
}
}
else if (Rs >= 2393) {
if (MVDiff < 32) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (MVDiff >= 32) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (diffRsCsdiff < -24) {
return 8;
}
else if (diffRsCsdiff >= -24) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (Rs < 3225) {
return 1;
}
else if (Rs >= 3225) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 15) {
if (ssDCval < 110) {
if (Rs < 3135) {
if (Rs < 2569) {
return 4;
}
else if (Rs >= 2569) {
if (TSC < 13) {
if (MVDiff < 30) {
return 2;
}
else if (MVDiff >= 30) {
if (Rs < 2734) {
return 2;
}
else if (Rs >= 2734) {
return 4;
}
}
}
else if (TSC >= 13) {
return 4;
}
}
}
else if (Rs >= 3135) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (ssDCval >= 110) {
if (diffMVdiffVal < 3) {
if (diffAFD < 1) {
if (Rs < 3541) {
return 1;
}
else if (Rs >= 3541) {
return 8;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
else if (AFD >= 24) {
if (ssDCval < 114) {
if (SC < 3875) {
if (refDCval < 97) {
if (RsCsDiff < 227) {
return 1;
}
else if (RsCsDiff >= 227) {
return 2;
}
}
else if (refDCval >= 97) {
return 2;
}
}
else if (SC >= 3875) {
if (Rs < 3628) {
return 4;
}
else if (Rs >= 3628) {
if (Rs < 3919) {
return 2;
}
else if (Rs >= 3919) {
return 1;
}
}
}
}
else if (ssDCval >= 114) {
if (mu_mv_mag_sq < 3) {
return 8;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 2900) {
return 8;
}
else if (Rs >= 2900) {
return 1;
}
}
}
}
}
else if (refDCval >= 134) {
if (negBalance < 50) {
if (Cs < 872) {
return 1;
}
else if (Cs >= 872) {
return 8;
}
}
else if (negBalance >= 50) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 8;
}
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 121) {
if (refDCval < 91) {
if (AFD < 14) {
if (SC < 3644) {
if (diffAFD < -2) {
if (Rs < 2376) {
return 1;
}
else if (Rs >= 2376) {
return 2;
}
}
else if (diffAFD >= -2) {
if (SC < 2718) {
return 2;
}
else if (SC >= 2718) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 2;
}
}
}
}
else if (SC >= 3644) {
if (Rs < 3997) {
return 2;
}
else if (Rs >= 3997) {
return 4;
}
}
}
else if (AFD >= 14) {
if (Rs < 3792) {
if (AFD < 23) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 2;
}
}
else if (AFD >= 23) {
if (Rs < 2573) {
return 8;
}
else if (Rs >= 2573) {
return 1;
}
}
}
else if (Rs >= 3792) {
return 4;
}
}
}
else if (refDCval >= 91) {
if (Rs < 5127) {
if (refDCval < 107) {
if (RsCsDiff < 157) {
if (posBalance < 56) {
if (tcor < 94) {
if (Rs < 2379) {
return 1;
}
else if (Rs >= 2379) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
return 8;
}
}
}
else if (tcor >= 94) {
if (Rs < 2942) {
return 8;
}
else if (Rs >= 2942) {
if (Rs < 3215) {
return 4;
}
else if (Rs >= 3215) {
return 4;
}
}
}
}
else if (posBalance >= 56) {
if (diffAFD < -1) {
if (diffTSC < -4) {
return 4;
}
else if (diffTSC >= -4) {
if (RsCsDiff < 54) {
if (Cs < 403) {
return 8;
}
else if (Cs >= 403) {
if (MVDiff < 34) {
if (Rs < 2471) {
return 1;
}
else if (Rs >= 2471) {
return 8;
}
}
else if (MVDiff >= 34) {
return 1;
}
}
}
else if (RsCsDiff >= 54) {
if (Cs < 1226) {
if (mu_mv_mag_sq < 29) {
return 2;
}
else if (mu_mv_mag_sq >= 29) {
if (Rs < 3334) {
return 1;
}
else if (Rs >= 3334) {
return 2;
}
}
}
else if (Cs >= 1226) {
return 1;
}
}
}
}
else if (diffAFD >= -1) {
if (TSC < 9) {
if (SC < 3813) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (SC >= 3813) {
return 8;
}
}
else if (TSC >= 9) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 157) {
if (gchDC < 3) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (MVDiff < 27) {
if (Cs < 1212) {
return 2;
}
else if (Cs >= 1212) {
return 8;
}
}
else if (MVDiff >= 27) {
if (posBalance < 50) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 1;
}
}
else if (posBalance >= 50) {
if (SC < 3101) {
return 2;
}
else if (SC >= 3101) {
return 8;
}
}
}
}
}
else if (gchDC >= 3) {
if (Rs < 3000) {
return 2;
}
else if (Rs >= 3000) {
if (refDCval < 104) {
return 1;
}
else if (refDCval >= 104) {
if (Rs < 3565) {
return 4;
}
else if (Rs >= 3565) {
if (Rs < 3790) {
return 2;
}
else if (Rs >= 3790) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 107) {
if (Rs < 2949) {
if (posBalance < 65) {
if (mu_mv_mag_sq < 12) {
if (RsCsDiff < 151) {
if (AFD < 17) {
if (Cs < 475) {
if (Rs < 2495) {
return 8;
}
else if (Rs >= 2495) {
return 2;
}
}
else if (Cs >= 475) {
return 1;
}
}
else if (AFD >= 17) {
return 8;
}
}
else if (RsCsDiff >= 151) {
if (mcTcor < 93) {
return 2;
}
else if (mcTcor >= 93) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 12) {
if (Cs < 407) {
if (diffTSC < 1) {
if (diffMVdiffVal < 23) {
return 1;
}
else if (diffMVdiffVal >= 23) {
return 8;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (Cs >= 407) {
if (AFD < 26) {
if (MVDiff < 23) {
if (Rs < 2681) {
return 8;
}
else if (Rs >= 2681) {
return 2;
}
}
else if (MVDiff >= 23) {
return 8;
}
}
else if (AFD >= 26) {
return 2;
}
}
}
}
else if (posBalance >= 65) {
if (MVDiff < 69) {
if (Rs < 2790) {
return 1;
}
else if (Rs >= 2790) {
if (diffMVdiffVal < -18) {
return 1;
}
else if (diffMVdiffVal >= -18) {
return 2;
}
}
}
else if (MVDiff >= 69) {
return 4;
}
}
}
else if (Rs >= 2949) {
if (TSC < 8) {
if (Rs < 3337) {
if (SC < 3560) {
if (Rs < 3028) {
return 8;
}
else if (Rs >= 3028) {
return 1;
}
}
else if (SC >= 3560) {
return 8;
}
}
else if (Rs >= 3337) {
if (CsDiff < 5) {
if (ssDCval < 113) {
return 8;
}
else if (ssDCval >= 113) {
return 2;
}
}
else if (CsDiff >= 5) {
if (Rs < 4341) {
return 8;
}
else if (Rs >= 4341) {
return 1;
}
}
}
}
else if (TSC >= 8) {
if (Cs < 382) {
if (Cs < 349) {
return 4;
}
else if (Cs >= 349) {
return 1;
}
}
else if (Cs >= 382) {
if (MVDiff < 31) {
if (Rs < 4795) {
return 1;
}
else if (Rs >= 4795) {
return 8;
}
}
else if (MVDiff >= 31) {
return 8;
}
}
}
}
}
}
else if (Rs >= 5127) {
return 2;
}
}
}
else if (ssDCval >= 121) {
if (RsCsDiff < 87) {
if (diffRsCsdiff < -1) {
if (mu_mv_mag_sq < 36) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
else if (mu_mv_mag_sq >= 36) {
if (Rs < 3805) {
if (Rs < 2681) {
if (Rs < 2516) {
if (Rs < 2407) {
return 2;
}
else if (Rs >= 2407) {
return 8;
}
}
else if (Rs >= 2516) {
return 1;
}
}
else if (Rs >= 2681) {
return 8;
}
}
else if (Rs >= 3805) {
return 2;
}
}
}
else if (diffRsCsdiff >= -1) {
if (refDCval < 119) {
return 1;
}
else if (refDCval >= 119) {
if (Rs < 3706) {
return 8;
}
else if (Rs >= 3706) {
if (Rs < 3999) {
return 2;
}
else if (Rs >= 3999) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 87) {
if (Rs < 3245) {
if (CsDiff < 3) {
if (AFD < 25) {
return 1;
}
else if (AFD >= 25) {
return 2;
}
}
else if (CsDiff >= 3) {
if (mcTcor < 95) {
if (MVDiff < 67) {
if (diffMVdiffVal < 7) {
return 1;
}
else if (diffMVdiffVal >= 7) {
if (Rs < 2666) {
return 1;
}
else if (Rs >= 2666) {
if (Rs < 2980) {
return 8;
}
else if (Rs >= 2980) {
return 2;
}
}
}
}
else if (MVDiff >= 67) {
if (Rs < 2553) {
if (Rs < 2449) {
return 1;
}
else if (Rs >= 2449) {
return 2;
}
}
else if (Rs >= 2553) {
return 8;
}
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
else if (Rs >= 3245) {
if (CsDiff < 5) {
if (mcTcor < 94) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 2;
}
}
else if (mcTcor >= 94) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
if (gchDC < 2) {
if (negBalance < 40) {
if (Rs < 3579) {
return 1;
}
else if (Rs >= 3579) {
return 2;
}
}
else if (negBalance >= 40) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (RsDiff < 15) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 1;
}
}
else if (RsDiff >= 15) {
return 1;
}
}
}
}
}
else if (CsDiff >= 5) {
if (refDCval < 126) {
if (Rs < 5148) {
return 8;
}
else if (Rs >= 5148) {
return 2;
}
}
else if (refDCval >= 126) {
if (posBalance < 67) {
return 1;
}
else if (posBalance >= 67) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 279) {
if (Cs < 675) {
if (RsCsDiff < 532) {
if (refDCval < 126) {
if (MVDiff < 32) {
if (mcTcor < 98) {
if (Rs < 4380) {
return 2;
}
else if (Rs >= 4380) {
if (Rs < 4926) {
return 4;
}
else if (Rs >= 4926) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (MVDiff < 21) {
return 1;
}
else if (MVDiff >= 21) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
}
}
else if (MVDiff >= 32) {
if (ssDCval < 72) {
if (Rs < 2739) {
return 1;
}
else if (Rs >= 2739) {
return 4;
}
}
else if (ssDCval >= 72) {
if (TSC < 11) {
if (mcTcor < 99) {
if (ssDCval < 114) {
if (diffRsCsdiff < 311) {
return 2;
}
else if (diffRsCsdiff >= 311) {
return 4;
}
}
else if (ssDCval >= 114) {
if (Cs < 388) {
return 2;
}
else if (Cs >= 388) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 2;
}
}
}
}
else if (mcTcor >= 99) {
if (Rs < 2468) {
return 1;
}
else if (Rs >= 2468) {
return 8;
}
}
}
else if (TSC >= 11) {
if (SC < 3471) {
return 1;
}
else if (SC >= 3471) {
if (Rs < 3652) {
return 2;
}
else if (Rs >= 3652) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 126) {
if (AFD < 19) {
if (Cs < 598) {
return 2;
}
else if (Cs >= 598) {
return 4;
}
}
else if (AFD >= 19) {
if (Cs < 231) {
return 2;
}
else if (Cs >= 231) {
return 1;
}
}
}
}
else if (RsCsDiff >= 532) {
if (ssDCval < 73) {
if (ssDCval < 68) {
if (Rs < 4409) {
return 2;
}
else if (Rs >= 4409) {
return 8;
}
}
else if (ssDCval >= 68) {
return 4;
}
}
else if (ssDCval >= 73) {
return 2;
}
}
}
else if (Cs >= 675) {
if (CsDiff < 5) {
if (mu_mv_mag_sq < 9) {
if (diffMVdiffVal < -11) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (diffMVdiffVal >= -11) {
if (refDCval < 118) {
if (negBalance < 46) {
if (Rs < 4013) {
return 2;
}
else if (Rs >= 4013) {
if (Rs < 4383) {
return 4;
}
else if (Rs >= 4383) {
return 1;
}
}
}
else if (negBalance >= 46) {
return 2;
}
}
else if (refDCval >= 118) {
if (Rs < 5983) {
if (Cs < 1016) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
else if (Cs >= 1016) {
return 8;
}
}
else if (Rs >= 5983) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (tcor < 46) {
return 2;
}
else if (tcor >= 46) {
if (negBalance < 39) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 2;
}
}
else if (negBalance >= 39) {
if (RsDiff < 25) {
return 4;
}
else if (RsDiff >= 25) {
return 2;
}
}
}
}
}
else if (CsDiff >= 5) {
if (ssDCval < 122) {
if (tcor < 77) {
if (AFD < 26) {
if (SC < 4252) {
if (AFD < 21) {
if (diffMVdiffVal < 7) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
else if (diffMVdiffVal >= 7) {
return 4;
}
}
else if (AFD >= 21) {
if (Cs < 1197) {
if (tcor < 74) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
if (mcTcor < 79) {
if (Rs < 2677) {
return 4;
}
else if (Rs >= 2677) {
return 8;
}
}
else if (mcTcor >= 79) {
return 8;
}
}
}
else if (tcor >= 74) {
if (RsDiff < 17) {
return 4;
}
else if (RsDiff >= 17) {
return 2;
}
}
}
else if (Cs >= 1197) {
return 4;
}
}
}
else if (SC >= 4252) {
if (mu_mv_mag_sq < 4) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 1;
}
}
else if (mu_mv_mag_sq >= 4) {
if (diffMVdiffVal < -6) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
}
}
else if (AFD >= 26) {
if (gchDC < 6) {
if (mcTcor < 93) {
if (ssDCval < 83) {
return 1;
}
else if (ssDCval >= 83) {
if (MVDiff < 79) {
return 2;
}
else if (MVDiff >= 79) {
if (Cs < 1062) {
return 2;
}
else if (Cs >= 1062) {
return 4;
}
}
}
}
else if (mcTcor >= 93) {
if (CsDiff < 9) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (CsDiff >= 9) {
return 1;
}
}
}
else if (gchDC >= 6) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 1;
}
}
}
}
else if (tcor >= 77) {
if (TSC < 13) {
if (mu_mv_mag_sq < 23) {
if (MVDiff < 35) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 2;
}
}
else if (MVDiff >= 35) {
if (CsDiff < 7) {
if (diffAFD < -9) {
return 1;
}
else if (diffAFD >= -9) {
if (mcTcor < 97) {
return 2;
}
else if (mcTcor >= 97) {
return 8;
}
}
}
else if (CsDiff >= 7) {
if (MVDiff < 45) {
return 1;
}
else if (MVDiff >= 45) {
if (Rs < 4788) {
return 2;
}
else if (Rs >= 4788) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (Cs < 880) {
if (ssDCval < 101) {
return 2;
}
else if (ssDCval >= 101) {
return 1;
}
}
else if (Cs >= 880) {
return 1;
}
}
}
else if (TSC >= 13) {
return 1;
}
}
}
else if (ssDCval >= 122) {
if (Rs < 5049) {
if (tcor < 73) {
if (diffAFD < 6) {
if (mcTcor < 92) {
if (Rs < 4029) {
return 1;
}
else if (Rs >= 4029) {
return 8;
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 9) {
if (MVDiff < 41) {
if (Rs < 4507) {
return 8;
}
else if (Rs >= 4507) {
if (Rs < 4680) {
return 2;
}
else if (Rs >= 4680) {
return 8;
}
}
}
else if (MVDiff >= 41) {
return 2;
}
}
else if (mu_mv_mag_sq >= 9) {
return 2;
}
}
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (tcor >= 73) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
if (RsCsDiff < 597) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (RsCsDiff < 287) {
return 4;
}
else if (RsCsDiff >= 287) {
if (MVDiff < 55) {
return 1;
}
else if (MVDiff >= 55) {
if (Rs < 4124) {
return 1;
}
else if (Rs >= 4124) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 597) {
if (RsCsDiff < 623) {
return 2;
}
else if (RsCsDiff >= 623) {
return 1;
}
}
}
}
}
else if (Rs >= 5049) {
if (TSC < 13) {
if (Rs < 5656) {
return 8;
}
else if (Rs >= 5656) {
if (MVDiff < 47) {
if (Rs < 6220) {
if (MVDiff < 33) {
if (Cs < 1238) {
return 8;
}
else if (Cs >= 1238) {
return 2;
}
}
else if (MVDiff >= 33) {
if (MVDiff < 44) {
return 2;
}
else if (MVDiff >= 44) {
return 8;
}
}
}
else if (Rs >= 6220) {
return 8;
}
}
else if (MVDiff >= 47) {
return 2;
}
}
}
else if (TSC >= 13) {
if (Rs < 6019) {
return 1;
}
else if (Rs >= 6019) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1313) {
if (mu_mv_mag_sq < 111) {
if (RsCsDiff < 101) {
if (ssDCval < 105) {
if (negBalance < 38) {
if (Rs < 816) {
if (Cs < 1429) {
if (refDCval < 63) {
return 1;
}
else if (refDCval >= 63) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
if (Rs < 311) {
return 4;
}
else if (Rs >= 311) {
if (Rs < 459) {
return 1;
}
else if (Rs >= 459) {
return 8;
}
}
}
}
}
else if (Cs >= 1429) {
if (TSC < 5) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
if (Rs < 662) {
return 8;
}
else if (Rs >= 662) {
return 4;
}
}
}
else if (TSC >= 5) {
if (Rs < 598) {
return 2;
}
else if (Rs >= 598) {
return 8;
}
}
}
}
else if (Rs >= 816) {
if (SC < 10165) {
if (negBalance < 32) {
if (posBalance < 45) {
if (ssDCval < 64) {
if (Rs < 2997) {
if (ssDCval < 58) {
if (TSC < 5) {
if (ssDCval < 55) {
if (MVDiff < 39) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (MVDiff >= 39) {
return 1;
}
}
else if (ssDCval >= 55) {
if (Rs < 1031) {
return 8;
}
else if (Rs >= 1031) {
return 4;
}
}
}
else if (TSC >= 5) {
if (Rs < 2380) {
return 2;
}
else if (Rs >= 2380) {
return 1;
}
}
}
else if (ssDCval >= 58) {
if (posBalance < 42) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 2;
}
}
else if (posBalance >= 42) {
if (Rs < 1543) {
return 4;
}
else if (Rs >= 1543) {
return 8;
}
}
}
}
else if (Rs >= 2997) {
if (posBalance < 36) {
if (Cs < 2481) {
return 4;
}
else if (Cs >= 2481) {
return 4;
}
}
else if (posBalance >= 36) {
if (RsDiff < 7) {
if (Rs < 4583) {
return 4;
}
else if (Rs >= 4583) {
return 4;
}
}
else if (RsDiff >= 7) {
return 8;
}
}
}
}
else if (ssDCval >= 64) {
if (refDCval < 73) {
if (MVDiff < 6) {
if (Rs < 3032) {
if (Rs < 1062) {
return 1;
}
else if (Rs >= 1062) {
return 8;
}
}
else if (Rs >= 3032) {
return 4;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (refDCval >= 73) {
if (MVDiff < 7) {
if (Rs < 3283) {
if (RsCsDiff < 1) {
if (Rs < 2404) {
if (Rs < 2351) {
if (ssDCval < 92) {
return 1;
}
else if (ssDCval >= 92) {
if (Rs < 2337) {
return 2;
}
else if (Rs >= 2337) {
return 4;
}
}
}
else if (Rs >= 2351) {
return 4;
}
}
else if (Rs >= 2404) {
if (ssDCval < 87) {
return 1;
}
else if (ssDCval >= 87) {
return 8;
}
}
}
else if (RsCsDiff >= 1) {
if (mu_mv_mag_sq < 85) {
if (Rs < 1864) {
return 8;
}
else if (Rs >= 1864) {
return 1;
}
}
else if (mu_mv_mag_sq >= 85) {
if (posBalance < 26) {
return 1;
}
else if (posBalance >= 26) {
return 8;
}
}
}
}
else if (Rs >= 3283) {
if (SC < 5668) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (SC >= 5668) {
if (SC < 7112) {
if (ssDCval < 96) {
if (TSC < 4) {
if (ssDCval < 89) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
else if (ssDCval >= 89) {
if (Rs < 4292) {
return 8;
}
else if (Rs >= 4292) {
return 2;
}
}
}
else if (TSC >= 4) {
if (Rs < 4981) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (Rs >= 4981) {
return 1;
}
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (SC >= 7112) {
return 4;
}
}
}
}
else if (MVDiff >= 7) {
if (TSC < 3) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (TSC >= 3) {
return 8;
}
}
}
}
}
else if (posBalance >= 45) {
if (RsDiff < 5) {
if (diffRsCsdiff < -22) {
if (ssDCval < 71) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (Rs < 1129) {
return 2;
}
else if (Rs >= 1129) {
return 4;
}
}
}
else if (ssDCval >= 71) {
if (Rs < 2657) {
return 8;
}
else if (Rs >= 2657) {
return 4;
}
}
}
else if (diffRsCsdiff >= -22) {
if (Cs < 1483) {
if (tcor < 97) {
if (RsDiff < 4) {
if (Rs < 918) {
return 2;
}
else if (Rs >= 918) {
return 8;
}
}
else if (RsDiff >= 4) {
if (Rs < 1124) {
return 2;
}
else if (Rs >= 1124) {
return 4;
}
}
}
else if (tcor >= 97) {
if (posBalance < 79) {
if (MVDiff < 1) {
if (Rs < 2264) {
return 8;
}
else if (Rs >= 2264) {
if (Rs < 2871) {
return 4;
}
else if (Rs >= 2871) {
return 4;
}
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (posBalance >= 79) {
if (Rs < 3262) {
return 8;
}
else if (Rs >= 3262) {
return 2;
}
}
}
}
else if (Cs >= 1483) {
if (refDCval < 57) {
if (refDCval < 55) {
if (Rs < 3782) {
return 8;
}
else if (Rs >= 3782) {
return 4;
}
}
else if (refDCval >= 55) {
if (Rs < 940) {
return 8;
}
else if (Rs >= 940) {
return 4;
}
}
}
else if (refDCval >= 57) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (SC < 4396) {
if (mu_mv_mag_sq < 76) {
if (Rs < 2103) {
return 1;
}
else if (Rs >= 2103) {
return 8;
}
}
else if (mu_mv_mag_sq >= 76) {
return 8;
}
}
else if (SC >= 4396) {
if (ssDCval < 88) {
if (Rs < 3422) {
return 4;
}
else if (Rs >= 3422) {
return 8;
}
}
else if (ssDCval >= 88) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 5) {
if (Rs < 2896) {
if (posBalance < 68) {
if (mu_mv_mag_sq < 91) {
if (gchDC < 2) {
if (Rs < 2817) {
return 1;
}
else if (Rs >= 2817) {
return 4;
}
}
else if (gchDC >= 2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 91) {
if (MVDiff < 2) {
if (Cs < 1875) {
if (Rs < 2669) {
return 8;
}
else if (Rs >= 2669) {
return 1;
}
}
else if (Cs >= 1875) {
return 4;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (posBalance >= 68) {
if (mu_mv_mag_sq < 100) {
return 8;
}
else if (mu_mv_mag_sq >= 100) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (Rs >= 2896) {
if (AFD < 9) {
if (ssDCval < 64) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
if (Rs < 4039) {
return 8;
}
else if (Rs >= 4039) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (Rs < 3459) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
if (Cs < 1452) {
return 2;
}
else if (Cs >= 1452) {
return 4;
}
}
}
else if (Rs >= 3459) {
if (Rs < 5342) {
return 2;
}
else if (Rs >= 5342) {
return 4;
}
}
}
}
else if (AFD >= 9) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (refDCval < 91) {
return 8;
}
else if (refDCval >= 91) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 32) {
if (RsDiff < 7) {
if (Cs < 1599) {
if (CsDiff < 6) {
if (Rs < 3655) {
if (SC < 2905) {
if (mu_mv_mag_sq < 76) {
if (tcor < 89) {
if (Rs < 1434) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (Rs >= 1434) {
return 8;
}
}
else if (tcor >= 89) {
if (MVDiff < 21) {
return 2;
}
else if (MVDiff >= 21) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (mcTcor < 95) {
if (Rs < 1432) {
return 8;
}
else if (Rs >= 1432) {
return 4;
}
}
else if (mcTcor >= 95) {
if (mu_mv_mag_sq < 106) {
return 4;
}
else if (mu_mv_mag_sq >= 106) {
if (SC < 2600) {
return 4;
}
else if (SC >= 2600) {
return 8;
}
}
}
}
}
else if (SC >= 2905) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (mcTcor < 97) {
if (MVDiff < 19) {
if (SC < 4505) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 2007) {
return 2;
}
else if (Rs >= 2007) {
return 8;
}
}
}
else if (SC >= 4505) {
if (mu_mv_mag_sq < 82) {
if (Cs < 1353) {
return 2;
}
else if (Cs >= 1353) {
return 8;
}
}
else if (mu_mv_mag_sq >= 82) {
if (Rs < 3196) {
return 2;
}
else if (Rs >= 3196) {
return 4;
}
}
}
}
else if (MVDiff >= 19) {
return 4;
}
}
else if (mcTcor >= 97) {
if (RsDiff < 6) {
if (MVDiff < 2) {
if (Rs < 2236) {
if (Rs < 1851) {
return 4;
}
else if (Rs >= 1851) {
return 8;
}
}
else if (Rs >= 2236) {
return 4;
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 98) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
else if (mu_mv_mag_sq >= 98) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
else if (Rs >= 3655) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
if (Rs < 4134) {
return 2;
}
else if (Rs >= 4134) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (Rs < 2050) {
if (posBalance < 41) {
return 2;
}
else if (posBalance >= 41) {
if (TSC < 5) {
if (Rs < 1013) {
return 4;
}
else if (Rs >= 1013) {
return 2;
}
}
else if (TSC >= 5) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (Rs < 1124) {
return 8;
}
else if (Rs >= 1124) {
return 1;
}
}
}
}
}
else if (Rs >= 2050) {
return 1;
}
}
}
else if (Cs >= 1599) {
if (Cs < 2750) {
if (refDCval < 89) {
if (Rs < 2706) {
if (mu_mv_mag_sq < 83) {
if (TSC < 6) {
if (negBalance < 36) {
if (ssDCval < 59) {
return 2;
}
else if (ssDCval >= 59) {
return 1;
}
}
else if (negBalance >= 36) {
if (Cs < 2103) {
return 8;
}
else if (Cs >= 2103) {
return 1;
}
}
}
else if (TSC >= 6) {
if (Cs < 1884) {
if (diffAFD < 2) {
if (ssDCval < 60) {
return 2;
}
else if (ssDCval >= 60) {
return 8;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (Cs >= 1884) {
if (Rs < 1312) {
return 8;
}
else if (Rs >= 1312) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (MVDiff < 6) {
if (RsDiff < 6) {
if (Cs < 2472) {
return 8;
}
else if (Cs >= 2472) {
return 4;
}
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (MVDiff >= 6) {
if (RsCsDiff < 11) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (RsCsDiff >= 11) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 8;
}
}
}
}
}
else if (Rs >= 2706) {
if (tcor < 96) {
if (ssDCval < 72) {
if (Rs < 3369) {
return 1;
}
else if (Rs >= 3369) {
return 4;
}
}
else if (ssDCval >= 72) {
if (TSC < 7) {
if (diffTSC < -2) {
if (Rs < 3006) {
return 8;
}
else if (Rs >= 3006) {
return 1;
}
}
else if (diffTSC >= -2) {
return 8;
}
}
else if (TSC >= 7) {
if (Rs < 3060) {
return 1;
}
else if (Rs >= 3060) {
if (Rs < 3923) {
return 8;
}
else if (Rs >= 3923) {
return 4;
}
}
}
}
}
else if (tcor >= 96) {
if (Rs < 4642) {
if (RsCsDiff < 48) {
return 4;
}
else if (RsCsDiff >= 48) {
if (Rs < 3066) {
return 1;
}
else if (Rs >= 3066) {
if (Rs < 3472) {
return 4;
}
else if (Rs >= 3472) {
return 2;
}
}
}
}
else if (Rs >= 4642) {
if (Rs < 5124) {
return 8;
}
else if (Rs >= 5124) {
return 4;
}
}
}
}
}
else if (refDCval >= 89) {
if (mcTcor < 96) {
if (Rs < 1490) {
return 2;
}
else if (Rs >= 1490) {
return 8;
}
}
else if (mcTcor >= 96) {
if (TSC < 6) {
if (SC < 4532) {
if (posBalance < 47) {
if (Rs < 2367) {
return 1;
}
else if (Rs >= 2367) {
if (Rs < 2470) {
return 8;
}
else if (Rs >= 2470) {
return 1;
}
}
}
else if (posBalance >= 47) {
return 8;
}
}
else if (SC >= 4532) {
if (ssDCval < 91) {
if (Rs < 3852) {
return 8;
}
else if (Rs >= 3852) {
return 2;
}
}
else if (ssDCval >= 91) {
return 8;
}
}
}
else if (TSC >= 6) {
if (ssDCval < 94) {
if (Rs < 4114) {
return 2;
}
else if (Rs >= 4114) {
return 4;
}
}
else if (ssDCval >= 94) {
return 8;
}
}
}
}
}
else if (Cs >= 2750) {
if (MVDiff < 2) {
if (Rs < 4184) {
return 4;
}
else if (Rs >= 4184) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Cs < 4295) {
return 8;
}
else if (Cs >= 4295) {
return 1;
}
}
}
}
}
else if (RsDiff >= 7) {
if (MVDiff < 8) {
if (mcTcor < 94) {
if (RsCsDiff < 94) {
if (Cs < 1450) {
if (Rs < 2462) {
return 8;
}
else if (Rs >= 2462) {
return 4;
}
}
else if (Cs >= 1450) {
return 8;
}
}
else if (RsCsDiff >= 94) {
if (Rs < 2563) {
return 4;
}
else if (Rs >= 2563) {
return 1;
}
}
}
else if (mcTcor >= 94) {
if (ssDCval < 89) {
if (Cs < 1691) {
if (MVDiff < 1) {
if (Rs < 4073) {
return 2;
}
else if (Rs >= 4073) {
return 4;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (Cs >= 1691) {
if (Rs < 3350) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (Rs >= 3350) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (Rs < 4508) {
return 1;
}
else if (Rs >= 4508) {
return 4;
}
}
}
}
}
else if (ssDCval >= 89) {
if (posBalance < 44) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (posBalance >= 44) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
else if (MVDiff >= 8) {
if (Cs < 1399) {
if (Rs < 2340) {
if (Rs < 1863) {
return 1;
}
else if (Rs >= 1863) {
return 2;
}
}
else if (Rs >= 2340) {
return 4;
}
}
else if (Cs >= 1399) {
if (Rs < 5046) {
if (tcor < 94) {
if (mu_mv_mag_sq < 23) {
return 1;
}
else if (mu_mv_mag_sq >= 23) {
if (ssDCval < 76) {
if (refDCval < 63) {
return 8;
}
else if (refDCval >= 63) {
return 4;
}
}
else if (ssDCval >= 76) {
return 8;
}
}
}
else if (tcor >= 94) {
return 2;
}
}
else if (Rs >= 5046) {
return 4;
}
}
}
}
}
}
else if (SC >= 10165) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 2;
}
}
}
}
else if (negBalance >= 38) {
if (Rs < 1455) {
if (ssDCval < 56) {
if (MVDiff < 21) {
if (posBalance < 56) {
if (Rs < 1089) {
if (Cs < 1434) {
return 4;
}
else if (Cs >= 1434) {
return 2;
}
}
else if (Rs >= 1089) {
if (Cs < 1452) {
return 8;
}
else if (Cs >= 1452) {
if (Rs < 1214) {
return 8;
}
else if (Rs >= 1214) {
return 4;
}
}
}
}
else if (posBalance >= 56) {
return 4;
}
}
else if (MVDiff >= 21) {
if (TSC < 8) {
if (Rs < 540) {
return 4;
}
else if (Rs >= 540) {
return 2;
}
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (AFD < 13) {
if (ssDCval < 83) {
if (posBalance < 56) {
if (negBalance < 41) {
if (Cs < 1639) {
if (Rs < 772) {
return 2;
}
else if (Rs >= 772) {
return 4;
}
}
else if (Cs >= 1639) {
if (mu_mv_mag_sq < 89) {
if (diffMVdiffVal < -3) {
if (Rs < 780) {
return 2;
}
else if (Rs >= 780) {
if (Rs < 995) {
return 1;
}
else if (Rs >= 995) {
return 8;
}
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 89) {
if (Rs < 1053) {
return 8;
}
else if (Rs >= 1053) {
return 4;
}
}
}
}
else if (negBalance >= 41) {
if (ssDCval < 72) {
if (ssDCval < 60) {
if (Rs < 909) {
return 2;
}
else if (Rs >= 909) {
return 8;
}
}
else if (ssDCval >= 60) {
if (MVDiff < 3) {
if (Rs < 1421) {
return 8;
}
else if (Rs >= 1421) {
return 1;
}
}
else if (MVDiff >= 3) {
if (diffRsCsdiff < -11) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 8;
}
}
else if (diffRsCsdiff >= -11) {
if (negBalance < 42) {
if (Rs < 905) {
return 2;
}
else if (Rs >= 905) {
return 1;
}
}
else if (negBalance >= 42) {
return 1;
}
}
}
}
}
else if (ssDCval >= 72) {
if (RsCsDiff < 23) {
return 1;
}
else if (RsCsDiff >= 23) {
if (MVDiff < 15) {
if (Rs < 1208) {
return 2;
}
else if (Rs >= 1208) {
return 8;
}
}
else if (MVDiff >= 15) {
if (Rs < 578) {
return 1;
}
else if (Rs >= 578) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 56) {
if (refDCval < 76) {
if (Rs < 1283) {
if (refDCval < 59) {
if (Cs < 1541) {
if (Rs < 1109) {
return 2;
}
else if (Rs >= 1109) {
return 8;
}
}
else if (Cs >= 1541) {
return 4;
}
}
else if (refDCval >= 59) {
if (Rs < 1142) {
if (diffRsCsdiff < 16) {
return 8;
}
else if (diffRsCsdiff >= 16) {
if (Cs < 1474) {
return 1;
}
else if (Cs >= 1474) {
return 8;
}
}
}
else if (Rs >= 1142) {
return 4;
}
}
}
else if (Rs >= 1283) {
if (Rs < 1376) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
else if (Rs >= 1376) {
return 8;
}
}
}
else if (refDCval >= 76) {
if (Rs < 868) {
return 1;
}
else if (Rs >= 868) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
else if (ssDCval >= 83) {
if (ssDCval < 95) {
if (mu_mv_mag_sq < 45) {
if (diffRsCsdiff < -8) {
if (Rs < 1223) {
return 4;
}
else if (Rs >= 1223) {
return 2;
}
}
else if (diffRsCsdiff >= -8) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 45) {
return 8;
}
}
else if (ssDCval >= 95) {
if (Cs < 2194) {
if (MVDiff < 11) {
if (posBalance < 62) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Rs < 1375) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
else if (Rs >= 1375) {
return 8;
}
}
}
else if (posBalance >= 62) {
return 8;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (Cs >= 2194) {
return 2;
}
}
}
}
else if (AFD >= 13) {
if (mcTcor < 72) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
else if (mcTcor >= 72) {
if (MVDiff < 11) {
if (refDCval < 92) {
if (ssDCval < 74) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
}
else if (ssDCval >= 74) {
if (mu_mv_mag_sq < 3) {
if (tcor < 86) {
return 4;
}
else if (tcor >= 86) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
if (Rs < 1218) {
return 8;
}
else if (Rs >= 1218) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 1658) {
return 8;
}
else if (Cs >= 1658) {
return 4;
}
}
}
}
else if (refDCval >= 92) {
if (SC < 3017) {
return 8;
}
else if (SC >= 3017) {
if (TSC < 9) {
if (AFD < 15) {
if (Rs < 1234) {
return 2;
}
else if (Rs >= 1234) {
return 8;
}
}
else if (AFD >= 15) {
return 2;
}
}
else if (TSC >= 9) {
if (Cs < 1826) {
return 4;
}
else if (Cs >= 1826) {
return 8;
}
}
}
}
}
else if (MVDiff >= 11) {
if (ssDCval < 74) {
if (RsCsDiff < 73) {
if (tcor < 78) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Cs < 1561) {
return 2;
}
else if (Cs >= 1561) {
return 8;
}
}
}
else if (tcor >= 78) {
return 2;
}
}
else if (RsCsDiff >= 73) {
if (tcor < 83) {
if (mcTcor < 85) {
if (Rs < 1313) {
return 2;
}
else if (Rs >= 1313) {
return 4;
}
}
else if (mcTcor >= 85) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (tcor >= 83) {
return 8;
}
}
}
else if (ssDCval >= 74) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (TSC < 12) {
if (AFD < 15) {
if (Rs < 1156) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Cs < 1347) {
return 8;
}
else if (Cs >= 1347) {
return 4;
}
}
}
else if (Rs >= 1156) {
return 8;
}
}
else if (AFD >= 15) {
if (MVDiff < 52) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
return 8;
}
}
else if (MVDiff >= 52) {
return 8;
}
}
}
else if (TSC >= 12) {
if (Rs < 1346) {
if (Rs < 1290) {
return 2;
}
else if (Rs >= 1290) {
return 4;
}
}
else if (Rs >= 1346) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1455) {
if (CsDiff < 7) {
if (mcTcor < 94) {
if (refDCval < 91) {
if (RsDiff < 8) {
if (posBalance < 70) {
if (refDCval < 77) {
if (diffRsCsdiff < -44) {
if (RsCsDiff < 69) {
return 8;
}
else if (RsCsDiff >= 69) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (diffRsCsdiff >= -44) {
if (Cs < 1423) {
if (Cs < 1411) {
if (Cs < 1318) {
if (Cs < 1317) {
return 8;
}
else if (Cs >= 1317) {
return 2;
}
}
else if (Cs >= 1318) {
return 8;
}
}
else if (Cs >= 1411) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
return 8;
}
}
}
else if (Cs >= 1423) {
if (Rs < 2091) {
if (negBalance < 51) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (SC < 3099) {
return 4;
}
else if (SC >= 3099) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (RsCsDiff < 67) {
return 8;
}
else if (RsCsDiff >= 67) {
if (Rs < 2003) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 8;
}
}
else if (Rs >= 2003) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 51) {
return 8;
}
}
else if (Rs >= 2091) {
return 8;
}
}
}
}
else if (refDCval >= 77) {
if (negBalance < 57) {
if (Cs < 2505) {
if (diffMVdiffVal < -3) {
if (mu_mv_mag_sq < 93) {
if (SC < 3149) {
return 4;
}
else if (SC >= 3149) {
return 8;
}
}
else if (mu_mv_mag_sq >= 93) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
return 8;
}
}
}
else if (diffMVdiffVal >= -3) {
if (AFD < 16) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Rs < 2051) {
return 1;
}
else if (Rs >= 2051) {
return 8;
}
}
}
else if (AFD >= 16) {
if (ssDCval < 84) {
if (Rs < 1835) {
return 4;
}
else if (Rs >= 1835) {
return 8;
}
}
else if (ssDCval >= 84) {
return 8;
}
}
}
}
else if (Cs >= 2505) {
if (Rs < 2728) {
return 4;
}
else if (Rs >= 2728) {
return 8;
}
}
}
else if (negBalance >= 57) {
if (refDCval < 81) {
return 8;
}
else if (refDCval >= 81) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
}
}
}
else if (posBalance >= 70) {
return 4;
}
}
else if (RsDiff >= 8) {
if (posBalance < 45) {
if (Rs < 3177) {
return 2;
}
else if (Rs >= 3177) {
return 4;
}
}
else if (posBalance >= 45) {
if (RsCsDiff < 93) {
if (TSC < 12) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (posBalance < 61) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (posBalance >= 61) {
return 8;
}
}
}
else if (TSC >= 12) {
return 1;
}
}
else if (RsCsDiff >= 93) {
if (SC < 4023) {
return 8;
}
else if (SC >= 4023) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (mu_mv_mag_sq < 70) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 4;
}
}
else if (mu_mv_mag_sq >= 70) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 91) {
if (Rs < 2106) {
if (mcTcor < 89) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 1725) {
return 8;
}
else if (Rs >= 1725) {
return 1;
}
}
}
else if (mcTcor >= 89) {
if (RsCsDiff < 67) {
if (Rs < 2020) {
if (refDCval < 101) {
return 8;
}
else if (refDCval >= 101) {
return 2;
}
}
else if (Rs >= 2020) {
return 2;
}
}
else if (RsCsDiff >= 67) {
if (AFD < 12) {
if (Rs < 1681) {
return 8;
}
else if (Rs >= 1681) {
return 2;
}
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (Rs >= 2106) {
if (Rs < 4639) {
if (diffRsCsdiff < -124) {
return 4;
}
else if (diffRsCsdiff >= -124) {
return 8;
}
}
else if (Rs >= 4639) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (SC < 7284) {
if (SC < 4774) {
if (Cs < 2327) {
if (Cs < 1479) {
if (ssDCval < 87) {
if (Rs < 2796) {
if (mu_mv_mag_sq < 91) {
if (MVDiff < 9) {
if (refDCval < 82) {
if (AFD < 11) {
if (MVDiff < 3) {
if (Rs < 1981) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 1;
}
}
else if (Rs >= 1981) {
if (RsCsDiff < 57) {
if (CsDiff < 4) {
if (Rs < 2230) {
return 8;
}
else if (Rs >= 2230) {
return 1;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (RsCsDiff >= 57) {
if (Rs < 2325) {
return 8;
}
else if (Rs >= 2325) {
return 2;
}
}
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (refDCval >= 82) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (mu_mv_mag_sq < 52) {
if (Cs < 1351) {
return 1;
}
else if (Cs >= 1351) {
return 8;
}
}
else if (mu_mv_mag_sq >= 52) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (Cs < 1395) {
return 1;
}
else if (Cs >= 1395) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 9) {
if (negBalance < 45) {
if (gchDC < 1) {
if (Rs < 2196) {
return 2;
}
else if (Rs >= 2196) {
return 8;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (negBalance >= 45) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 2408) {
return 1;
}
else if (Rs >= 2408) {
return 8;
}
}
}
}
}
else if (Rs >= 2796) {
if (Cs < 1428) {
return 2;
}
else if (Cs >= 1428) {
return 4;
}
}
}
else if (ssDCval >= 87) {
if (SC < 3550) {
if (TSC < 7) {
if (ssDCval < 91) {
if (Rs < 2147) {
return 8;
}
else if (Rs >= 2147) {
return 2;
}
}
else if (ssDCval >= 91) {
if (Rs < 1672) {
return 8;
}
else if (Rs >= 1672) {
return 2;
}
}
}
else if (TSC >= 7) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (Rs < 1872) {
return 2;
}
else if (Rs >= 1872) {
return 8;
}
}
}
}
}
else if (SC >= 3550) {
if (ssDCval < 90) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (Cs < 1391) {
return 4;
}
else if (Cs >= 1391) {
return 8;
}
}
}
else if (ssDCval >= 90) {
if (diffMVdiffVal < -35) {
return 4;
}
else if (diffMVdiffVal >= -35) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1479) {
if (mu_mv_mag_sq < 44) {
if (tcor < 92) {
if (refDCval < 83) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Cs < 2045) {
return 4;
}
else if (Cs >= 2045) {
return 8;
}
}
}
else if (refDCval >= 83) {
if (ssDCval < 102) {
if (AFD < 16) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
return 2;
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (ssDCval >= 102) {
if (Rs < 3090) {
return 4;
}
else if (Rs >= 3090) {
return 8;
}
}
}
}
else if (tcor >= 92) {
if (posBalance < 58) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
return 8;
}
}
else if (posBalance >= 58) {
if (Rs < 2060) {
return 2;
}
else if (Rs >= 2060) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (posBalance < 13) {
if (diffAFD < 1) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 2;
}
}
else if (diffAFD >= 1) {
if (Rs < 1886) {
return 8;
}
else if (Rs >= 1886) {
return 1;
}
}
}
else if (posBalance >= 13) {
if (negBalance < 52) {
if (refDCval < 80) {
if (AFD < 8) {
if (Rs < 1494) {
if (Rs < 1475) {
return 4;
}
else if (Rs >= 1475) {
return 1;
}
}
else if (Rs >= 1494) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (AFD >= 8) {
if (RsDiff < 6) {
if (Cs < 1898) {
return 4;
}
else if (Cs >= 1898) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
else if (refDCval >= 80) {
if (RsCsDiff < 51) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (ssDCval < 98) {
if (refDCval < 87) {
if (Cs < 1607) {
return 1;
}
else if (Cs >= 1607) {
return 8;
}
}
else if (refDCval >= 87) {
return 8;
}
}
else if (ssDCval >= 98) {
return 2;
}
}
}
else if (RsCsDiff >= 51) {
if (SC < 3903) {
if (negBalance < 46) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
return 1;
}
}
}
else if (negBalance >= 46) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (SC >= 3903) {
if (MVDiff < 2) {
if (Cs < 1715) {
if (Rs < 2712) {
return 1;
}
else if (Rs >= 2712) {
return 4;
}
}
else if (Cs >= 1715) {
if (Rs < 2060) {
return 1;
}
else if (Rs >= 2060) {
if (Rs < 2719) {
return 8;
}
else if (Rs >= 2719) {
return 1;
}
}
}
}
else if (MVDiff >= 2) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Cs < 1595) {
return 1;
}
else if (Cs >= 1595) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 52) {
if (SC < 4417) {
if (Cs < 2192) {
return 8;
}
else if (Cs >= 2192) {
return 1;
}
}
else if (SC >= 4417) {
if (Rs < 2879) {
return 4;
}
else if (Rs >= 2879) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 2327) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (Rs < 1603) {
return 2;
}
else if (Rs >= 1603) {
return 8;
}
}
}
}
else if (SC >= 4774) {
if (Cs < 1734) {
if (ssDCval < 90) {
if (RsDiff < 8) {
if (mu_mv_mag_sq < 92) {
if (posBalance < 49) {
if (Cs < 1719) {
return 4;
}
else if (Cs >= 1719) {
return 1;
}
}
else if (posBalance >= 49) {
if (Rs < 3579) {
return 2;
}
else if (Rs >= 3579) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (Rs < 3866) {
return 2;
}
else if (Rs >= 3866) {
return 4;
}
}
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (ssDCval >= 90) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
return 1;
}
}
}
else if (Cs >= 1734) {
if (CsDiff < 4) {
if (Rs < 4020) {
if (Rs < 2544) {
if (Rs < 2386) {
return 8;
}
else if (Rs >= 2386) {
return 4;
}
}
else if (Rs >= 2544) {
return 8;
}
}
else if (Rs >= 4020) {
if (Cs < 2682) {
return 4;
}
else if (Cs >= 2682) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (mu_mv_mag_sq < 57) {
return 8;
}
else if (mu_mv_mag_sq >= 57) {
if (negBalance < 45) {
if (SC < 5035) {
if (SC < 4825) {
return 1;
}
else if (SC >= 4825) {
return 8;
}
}
else if (SC >= 5035) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
if (TSC < 5) {
if (SC < 6541) {
return 4;
}
else if (SC >= 6541) {
return 8;
}
}
else if (TSC >= 5) {
if (ssDCval < 79) {
if (Cs < 2715) {
return 4;
}
else if (Cs >= 2715) {
if (Rs < 3401) {
return 1;
}
else if (Rs >= 3401) {
return 8;
}
}
}
else if (ssDCval >= 79) {
if (SC < 5242) {
if (Rs < 2847) {
return 4;
}
else if (Rs >= 2847) {
return 8;
}
}
else if (SC >= 5242) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 45) {
if (ssDCval < 88) {
if (posBalance < 55) {
if (Cs < 2805) {
return 8;
}
else if (Cs >= 2805) {
return 4;
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (ssDCval >= 88) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 7284) {
if (Rs < 5565) {
return 8;
}
else if (Rs >= 5565) {
return 4;
}
}
}
}
else if (CsDiff >= 7) {
if (Rs < 3073) {
if (Cs < 2956) {
if (refDCval < 51) {
if (Cs < 1511) {
return 8;
}
else if (Cs >= 1511) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (diffRsCsdiff < 7) {
return 4;
}
else if (diffRsCsdiff >= 7) {
if (Rs < 1688) {
return 4;
}
else if (Rs >= 1688) {
return 8;
}
}
}
}
}
else if (refDCval >= 51) {
if (MVDiff < 42) {
if (diffAFD < -5) {
if (AFD < 11) {
if (Rs < 1859) {
return 1;
}
else if (Rs >= 1859) {
return 8;
}
}
else if (AFD >= 11) {
return 2;
}
}
else if (diffAFD >= -5) {
if (mcTcor < 93) {
if (Rs < 1748) {
if (Cs < 1866) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
if (Rs < 1608) {
return 4;
}
else if (Rs >= 1608) {
return 8;
}
}
}
else if (Cs >= 1866) {
if (Rs < 1673) {
if (Cs < 2002) {
return 4;
}
else if (Cs >= 2002) {
return 8;
}
}
else if (Rs >= 1673) {
return 4;
}
}
}
else if (Rs >= 1748) {
if (gchDC < 2) {
if (diffTSC < -2) {
if (Rs < 2133) {
return 4;
}
else if (Rs >= 2133) {
return 8;
}
}
else if (diffTSC >= -2) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 4;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (mcTcor >= 93) {
if (Cs < 2297) {
if (diffRsCsdiff < 36) {
if (tcor < 87) {
if (SC < 3922) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (SC >= 3922) {
if (Rs < 2941) {
return 4;
}
else if (Rs >= 2941) {
return 8;
}
}
}
else if (tcor >= 87) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (diffRsCsdiff >= 36) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (Cs >= 2297) {
if (Rs < 2329) {
return 1;
}
else if (Rs >= 2329) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
if (diffRsCsdiff < 26) {
return 8;
}
else if (diffRsCsdiff >= 26) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 42) {
if (ssDCval < 96) {
return 1;
}
else if (ssDCval >= 96) {
if (Rs < 2030) {
return 2;
}
else if (Rs >= 2030) {
return 8;
}
}
}
}
}
else if (Cs >= 2956) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (Rs >= 3073) {
if (mu_mv_mag_sq < 88) {
if (Rs < 3414) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 8;
}
}
else if (Rs >= 3414) {
return 8;
}
}
else if (mu_mv_mag_sq >= 88) {
if (Cs < 2573) {
if (Rs < 3592) {
return 4;
}
else if (Rs >= 3592) {
return 1;
}
}
else if (Cs >= 2573) {
if (Cs < 3275) {
return 4;
}
else if (Cs >= 3275) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 105) {
if (ssDCval < 163) {
if (Rs < 4579) {
if (Rs < 1554) {
if (mu_mv_mag_sq < 36) {
if (ssDCval < 138) {
if (mcTcor < 87) {
if (ssDCval < 112) {
return 4;
}
else if (ssDCval >= 112) {
if (Rs < 1537) {
return 1;
}
else if (Rs >= 1537) {
return 8;
}
}
}
else if (mcTcor >= 87) {
if (Cs < 1358) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (Cs >= 1358) {
if (TSC < 8) {
if (negBalance < 40) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
else if (negBalance >= 40) {
return 8;
}
}
else if (TSC >= 8) {
if (negBalance < 52) {
if (Cs < 1967) {
return 4;
}
else if (Cs >= 1967) {
return 1;
}
}
else if (negBalance >= 52) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (Rs < 1012) {
return 8;
}
else if (Rs >= 1012) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 138) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (Cs < 1474) {
return 1;
}
else if (Cs >= 1474) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 36) {
if (Rs < 1239) {
if (Rs < 658) {
if (Rs < 586) {
return 4;
}
else if (Rs >= 586) {
return 1;
}
}
else if (Rs >= 658) {
if (Rs < 990) {
if (ssDCval < 115) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
else if (ssDCval >= 115) {
if (Cs < 2094) {
return 8;
}
else if (Cs >= 2094) {
return 1;
}
}
}
else if (Rs >= 990) {
if (CsDiff < 6) {
if (ssDCval < 127) {
if (posBalance < 57) {
if (Cs < 1406) {
return 8;
}
else if (Cs >= 1406) {
return 4;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (ssDCval >= 127) {
if (negBalance < 38) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (negBalance >= 38) {
if (MVDiff < 20) {
if (Rs < 1012) {
return 8;
}
else if (Rs >= 1012) {
return 4;
}
}
else if (MVDiff >= 20) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (posBalance < 119) {
return 8;
}
else if (posBalance >= 119) {
return 1;
}
}
}
}
}
else if (Rs >= 1239) {
if (Cs < 1745) {
if (ssDCval < 159) {
if (diffMVdiffVal < -14) {
if (Cs < 1470) {
return 8;
}
else if (Cs >= 1470) {
return 1;
}
}
else if (diffMVdiffVal >= -14) {
return 8;
}
}
else if (ssDCval >= 159) {
return 2;
}
}
else if (Cs >= 1745) {
if (ssDCval < 127) {
return 8;
}
else if (ssDCval >= 127) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1554) {
if (mu_mv_mag_sq < 42) {
if (Rs < 1934) {
if (mu_mv_mag_sq < 14) {
if (CsDiff < 5) {
if (ssDCval < 122) {
if (Rs < 1592) {
return 8;
}
else if (Rs >= 1592) {
return 4;
}
}
else if (ssDCval >= 122) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
}
else if (CsDiff >= 5) {
if (gchDC < 1) {
if (negBalance < 56) {
if (ssDCval < 135) {
return 8;
}
else if (ssDCval >= 135) {
if (Rs < 1789) {
return 1;
}
else if (Rs >= 1789) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (Rs < 1604) {
return 8;
}
else if (Rs >= 1604) {
if (Rs < 1787) {
return 2;
}
else if (Rs >= 1787) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
if (Cs < 2024) {
if (Cs < 1482) {
if (mu_mv_mag_sq < 5) {
return 8;
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 1707) {
return 2;
}
else if (Rs >= 1707) {
return 4;
}
}
}
else if (Cs >= 1482) {
if (Rs < 1590) {
return 4;
}
else if (Rs >= 1590) {
return 2;
}
}
}
else if (Cs >= 2024) {
if (Rs < 1866) {
return 8;
}
else if (Rs >= 1866) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
return 8;
}
}
else if (Rs >= 1934) {
if (ssDCval < 110) {
if (TSC < 7) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (Cs < 1572) {
return 4;
}
else if (Cs >= 1572) {
return 8;
}
}
}
}
else if (TSC >= 7) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (Rs < 2142) {
return 8;
}
else if (Rs >= 2142) {
return 1;
}
}
}
}
else if (ssDCval >= 110) {
if (Rs < 2017) {
if (Rs < 1990) {
return 8;
}
else if (Rs >= 1990) {
if (Cs < 1804) {
return 1;
}
else if (Cs >= 1804) {
return 8;
}
}
}
else if (Rs >= 2017) {
if (mcTcor < 98) {
if (Cs < 2777) {
return 8;
}
else if (Cs >= 2777) {
return 2;
}
}
else if (mcTcor >= 98) {
if (Rs < 3427) {
return 8;
}
else if (Rs >= 3427) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (posBalance < 55) {
if (posBalance < 37) {
if (RsCsDiff < 2) {
if (mu_mv_mag_sq < 108) {
if (Rs < 2744) {
if (Cs < 2065) {
return 4;
}
else if (Cs >= 2065) {
return 8;
}
}
else if (Rs >= 2744) {
return 2;
}
}
else if (mu_mv_mag_sq >= 108) {
if (posBalance < 9) {
return 2;
}
else if (posBalance >= 9) {
if (Rs < 3159) {
return 4;
}
else if (Rs >= 3159) {
return 2;
}
}
}
}
else if (RsCsDiff >= 2) {
if (Rs < 4034) {
if (Cs < 1335) {
return 2;
}
else if (Cs >= 1335) {
if (RsCsDiff < 4) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (Rs < 2422) {
return 1;
}
else if (Rs >= 2422) {
return 8;
}
}
}
}
else if (RsCsDiff >= 4) {
return 8;
}
}
}
else if (Rs >= 4034) {
if (Rs < 4377) {
return 1;
}
else if (Rs >= 4377) {
return 2;
}
}
}
}
else if (posBalance >= 37) {
if (refDCval < 109) {
if (Cs < 1527) {
if (MVDiff < 11) {
if (MVDiff < 1) {
if (Rs < 2747) {
return 8;
}
else if (Rs >= 2747) {
return 4;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (MVDiff >= 11) {
if (Rs < 3058) {
return 4;
}
else if (Rs >= 3058) {
if (Rs < 3859) {
return 8;
}
else if (Rs >= 3859) {
return 4;
}
}
}
}
else if (Cs >= 1527) {
if (diffRsCsdiff < -69) {
return 4;
}
else if (diffRsCsdiff >= -69) {
return 8;
}
}
}
else if (refDCval >= 109) {
if (Cs < 1416) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (ssDCval < 129) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 2;
}
}
else if (ssDCval >= 129) {
if (Rs < 2279) {
return 8;
}
else if (Rs >= 2279) {
if (Rs < 2837) {
return 1;
}
else if (Rs >= 2837) {
return 2;
}
}
}
}
}
else if (Cs >= 1416) {
if (diffRsCsdiff < -11) {
if (diffRsCsdiff < -142) {
if (Rs < 3445) {
return 8;
}
else if (Rs >= 3445) {
return 1;
}
}
else if (diffRsCsdiff >= -142) {
if (ssDCval < 130) {
if (ssDCval < 125) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (ssDCval >= 125) {
if (Rs < 3020) {
if (Cs < 1987) {
return 1;
}
else if (Cs >= 1987) {
return 8;
}
}
else if (Rs >= 3020) {
return 8;
}
}
}
else if (ssDCval >= 130) {
return 8;
}
}
}
else if (diffRsCsdiff >= -11) {
if (mu_mv_mag_sq < 51) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 51) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 55) {
if (Rs < 3676) {
if (negBalance < 57) {
if (MVDiff < 8) {
if (refDCval < 123) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (Cs < 1768) {
return 1;
}
else if (Cs >= 1768) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
else if (refDCval >= 123) {
if (ssDCval < 130) {
if (Rs < 1954) {
return 1;
}
else if (Rs >= 1954) {
if (CsDiff < 6) {
if (Cs < 1391) {
return 2;
}
else if (Cs >= 1391) {
return 8;
}
}
else if (CsDiff >= 6) {
if (Rs < 2887) {
return 8;
}
else if (Rs >= 2887) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
else if (ssDCval >= 130) {
if (mu_mv_mag_sq < 96) {
return 8;
}
else if (mu_mv_mag_sq >= 96) {
if (Cs < 1740) {
if (Rs < 2149) {
return 8;
}
else if (Rs >= 2149) {
return 1;
}
}
else if (Cs >= 1740) {
return 8;
}
}
}
}
}
else if (MVDiff >= 8) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 2473) {
if (Cs < 1802) {
return 1;
}
else if (Cs >= 1802) {
return 4;
}
}
else if (Rs >= 2473) {
return 8;
}
}
}
}
}
else if (negBalance >= 57) {
return 4;
}
}
else if (Rs >= 3676) {
if (ssDCval < 138) {
if (Cs < 1452) {
return 2;
}
else if (Cs >= 1452) {
if (diffRsCsdiff < -74) {
return 4;
}
else if (diffRsCsdiff >= -74) {
return 8;
}
}
}
else if (ssDCval >= 138) {
if (Cs < 1679) {
return 8;
}
else if (Cs >= 1679) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 4579) {
if (Cs < 1868) {
if (tcor < 92) {
if (Cs < 1712) {
return 1;
}
else if (Cs >= 1712) {
return 8;
}
}
else if (tcor >= 92) {
return 2;
}
}
else if (Cs >= 1868) {
if (Cs < 3725) {
if (Rs < 5979) {
if (Rs < 4651) {
return 8;
}
else if (Rs >= 4651) {
return 1;
}
}
else if (Rs >= 5979) {
return 4;
}
}
else if (Cs >= 3725) {
return 8;
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2246) {
if (Rs < 1049) {
return 1;
}
else if (Rs >= 1049) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (Rs >= 2246) {
return 8;
}
}
}
}
else if (RsCsDiff >= 101) {
if (CsDiff < 8) {
if (tcor < 88) {
if (RsDiff < 13) {
if (Cs < 1467) {
if (TSC < 9) {
if (SC < 4594) {
if (tcor < 73) {
if (Rs < 1341) {
return 4;
}
else if (Rs >= 1341) {
if (mu_mv_mag_sq < 5) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 1;
}
}
else if (mu_mv_mag_sq >= 5) {
return 8;
}
}
}
else if (tcor >= 73) {
if (MVDiff < 24) {
if (mu_mv_mag_sq < 20) {
if (Rs < 2492) {
return 2;
}
else if (Rs >= 2492) {
if (Rs < 2871) {
return 4;
}
else if (Rs >= 2871) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 20) {
if (diffAFD < -1) {
if (AFD < 12) {
if (Rs < 1513) {
return 2;
}
else if (Rs >= 1513) {
return 8;
}
}
else if (AFD >= 12) {
return 2;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (MVDiff >= 24) {
if (ssDCval < 122) {
if (Rs < 1865) {
return 8;
}
else if (Rs >= 1865) {
return 1;
}
}
else if (ssDCval >= 122) {
return 8;
}
}
}
}
else if (SC >= 4594) {
if (posBalance < 50) {
if (ssDCval < 62) {
if (Rs < 4196) {
return 2;
}
else if (Rs >= 4196) {
return 4;
}
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (posBalance >= 50) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
if (Cs < 1324) {
return 4;
}
else if (Cs >= 1324) {
return 8;
}
}
}
}
}
else if (TSC >= 9) {
if (MVDiff < 21) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (mu_mv_mag_sq < 78) {
if (diffRsCsdiff < 54) {
if (Rs < 1470) {
return 4;
}
else if (Rs >= 1470) {
if (diffRsCsdiff < -36) {
if (Rs < 3067) {
return 8;
}
else if (Rs >= 3067) {
return 4;
}
}
else if (diffRsCsdiff >= -36) {
return 8;
}
}
}
else if (diffRsCsdiff >= 54) {
return 2;
}
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 2298) {
return 4;
}
else if (Rs >= 2298) {
return 8;
}
}
}
}
else if (MVDiff >= 21) {
if (MVDiff < 66) {
if (diffRsCsdiff < 28) {
if (Cs < 1439) {
if (MVDiff < 49) {
if (TSC < 11) {
if (Rs < 1995) {
return 8;
}
else if (Rs >= 1995) {
return 2;
}
}
else if (TSC >= 11) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (MVDiff >= 49) {
return 8;
}
}
else if (Cs >= 1439) {
if (Rs < 1933) {
return 4;
}
else if (Rs >= 1933) {
return 8;
}
}
}
else if (diffRsCsdiff >= 28) {
if (gchDC < 7) {
return 8;
}
else if (gchDC >= 7) {
return 4;
}
}
}
else if (MVDiff >= 66) {
if (Cs < 1363) {
if (Rs < 2793) {
return 2;
}
else if (Rs >= 2793) {
return 8;
}
}
else if (Cs >= 1363) {
return 4;
}
}
}
}
}
else if (Cs >= 1467) {
if (ssDCval < 88) {
if (negBalance < 39) {
if (Cs < 1851) {
if (mcTcor < 89) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (mcTcor >= 89) {
if (RsCsDiff < 147) {
return 4;
}
else if (RsCsDiff >= 147) {
return 2;
}
}
}
else if (Cs >= 1851) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (SC < 5219) {
if (TSC < 12) {
if (ssDCval < 71) {
if (Rs < 1821) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 4;
}
}
else if (Rs >= 1821) {
if (Cs < 1968) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 1;
}
}
else if (Cs >= 1968) {
if (Rs < 2314) {
return 4;
}
else if (Rs >= 2314) {
return 8;
}
}
}
}
else if (ssDCval >= 71) {
if (TSC < 9) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (TSC >= 9) {
if (mcTcor < 89) {
if (mu_mv_mag_sq < 50) {
return 4;
}
else if (mu_mv_mag_sq >= 50) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (Rs < 2760) {
return 4;
}
else if (Rs >= 2760) {
return 8;
}
}
}
}
else if (mcTcor >= 89) {
if (refDCval < 87) {
if (RsCsDiff < 143) {
return 8;
}
else if (RsCsDiff >= 143) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
return 8;
}
}
}
else if (refDCval >= 87) {
return 4;
}
}
}
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (SC >= 5219) {
if (Rs < 2710) {
return 2;
}
else if (Rs >= 2710) {
return 8;
}
}
}
}
else if (ssDCval >= 88) {
if (negBalance < 56) {
if (ssDCval < 100) {
if (SC < 4843) {
if (TSC < 10) {
if (RsCsDiff < 117) {
if (Cs < 1698) {
if (Rs < 2422) {
return 2;
}
else if (Rs >= 2422) {
return 4;
}
}
else if (Cs >= 1698) {
return 8;
}
}
else if (RsCsDiff >= 117) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
return 4;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (SC >= 4843) {
return 8;
}
}
else if (ssDCval >= 100) {
if (Rs < 1658) {
if (Rs < 1560) {
return 4;
}
else if (Rs >= 1560) {
return 2;
}
}
else if (Rs >= 1658) {
if (refDCval < 107) {
if (refDCval < 104) {
return 8;
}
else if (refDCval >= 104) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
if (Rs < 2638) {
return 8;
}
else if (Rs >= 2638) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
return 4;
}
}
}
}
}
else if (refDCval >= 107) {
if (MVDiff < 27) {
if (tcor < 66) {
if (Rs < 3822) {
return 1;
}
else if (Rs >= 3822) {
return 8;
}
}
else if (tcor >= 66) {
if (diffMVdiffVal < -88) {
return 4;
}
else if (diffMVdiffVal >= -88) {
if (Cs < 2435) {
return 8;
}
else if (Cs >= 2435) {
if (Rs < 4017) {
if (Rs < 2498) {
return 2;
}
else if (Rs >= 2498) {
if (Rs < 3369) {
return 8;
}
else if (Rs >= 3369) {
return 4;
}
}
}
else if (Rs >= 4017) {
return 8;
}
}
}
}
}
else if (MVDiff >= 27) {
if (mu_mv_mag_sq < 48) {
if (Rs < 3859) {
return 8;
}
else if (Rs >= 3859) {
return 1;
}
}
else if (mu_mv_mag_sq >= 48) {
if (SC < 4396) {
if (Rs < 2379) {
return 8;
}
else if (Rs >= 2379) {
return 2;
}
}
else if (SC >= 4396) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
}
}
}
}
else if (negBalance >= 56) {
if (RsDiff < 12) {
if (Rs < 2353) {
if (SC < 3872) {
if (Rs < 1929) {
return 2;
}
else if (Rs >= 1929) {
return 8;
}
}
else if (SC >= 3872) {
if (mu_mv_mag_sq < 26) {
return 1;
}
else if (mu_mv_mag_sq >= 26) {
return 8;
}
}
}
else if (Rs >= 2353) {
if (SC < 5735) {
if (tcor < 74) {
if (Cs < 1583) {
return 4;
}
else if (Cs >= 1583) {
return 8;
}
}
else if (tcor >= 74) {
return 8;
}
}
else if (SC >= 5735) {
if (Rs < 4344) {
return 4;
}
else if (Rs >= 4344) {
return 1;
}
}
}
}
else if (RsDiff >= 12) {
return 4;
}
}
}
}
}
else if (RsDiff >= 13) {
if (tcor < 84) {
if (ssDCval < 86) {
if (Cs < 1746) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 2;
}
}
else if (Cs >= 1746) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (ssDCval >= 86) {
if (AFD < 33) {
if (Rs < 5587) {
if (refDCval < 138) {
if (mcTcor < 90) {
if (AFD < 24) {
if (mcTcor < 85) {
if (Cs < 3430) {
if (SC < 4326) {
if (Rs < 2546) {
return 4;
}
else if (Rs >= 2546) {
return 2;
}
}
else if (SC >= 4326) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (Cs >= 3430) {
return 1;
}
}
else if (mcTcor >= 85) {
if (diffMVdiffVal < -1) {
if (Rs < 3369) {
return 1;
}
else if (Rs >= 3369) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (negBalance < 52) {
if (Rs < 3286) {
return 8;
}
else if (Rs >= 3286) {
return 4;
}
}
else if (negBalance >= 52) {
return 4;
}
}
}
}
else if (AFD >= 24) {
if (diffMVdiffVal < -8) {
return 2;
}
else if (diffMVdiffVal >= -8) {
if (mu_mv_mag_sq < 5) {
return 2;
}
else if (mu_mv_mag_sq >= 5) {
return 4;
}
}
}
}
else if (mcTcor >= 90) {
if (refDCval < 101) {
if (mcTcor < 93) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (refDCval >= 101) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (RsDiff < 18) {
return 1;
}
else if (RsDiff >= 18) {
if (Rs < 4970) {
return 2;
}
else if (Rs >= 4970) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 138) {
if (negBalance < 57) {
if (Rs < 3248) {
if (RsDiff < 15) {
return 4;
}
else if (RsDiff >= 15) {
if (RsCsDiff < 299) {
return 2;
}
else if (RsCsDiff >= 299) {
return 1;
}
}
}
else if (Rs >= 3248) {
return 4;
}
}
else if (negBalance >= 57) {
if (Cs < 1979) {
return 8;
}
else if (Cs >= 1979) {
return 4;
}
}
}
}
else if (Rs >= 5587) {
if (TSC < 16) {
if (Cs < 1429) {
if (Cs < 1421) {
return 8;
}
else if (Cs >= 1421) {
return 2;
}
}
else if (Cs >= 1429) {
return 8;
}
}
else if (TSC >= 16) {
return 4;
}
}
}
else if (AFD >= 33) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
if (Rs < 3635) {
return 2;
}
else if (Rs >= 3635) {
return 4;
}
}
}
}
}
else if (tcor >= 84) {
if (CsDiff < 4) {
if (Rs < 8906) {
return 4;
}
else if (Rs >= 8906) {
return 8;
}
}
else if (CsDiff >= 4) {
if (AFD < 11) {
if (Cs < 1745) {
return 4;
}
else if (Cs >= 1745) {
return 8;
}
}
else if (AFD >= 11) {
if (mcTcor < 95) {
if (Cs < 2186) {
if (Cs < 1379) {
if (Cs < 1377) {
return 2;
}
else if (Cs >= 1377) {
return 4;
}
}
else if (Cs >= 1379) {
return 2;
}
}
else if (Cs >= 2186) {
return 4;
}
}
else if (mcTcor >= 95) {
if (RsCsDiff < 227) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (Rs < 3039) {
return 1;
}
else if (Rs >= 3039) {
return 4;
}
}
}
else if (RsCsDiff >= 227) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 88) {
if (ssDCval < 112) {
if (Rs < 3200) {
if (Cs < 1453) {
if (ssDCval < 98) {
if (Rs < 2732) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (Rs < 1972) {
return 1;
}
else if (Rs >= 1972) {
return 8;
}
}
}
else if (Rs >= 2732) {
return 2;
}
}
else if (ssDCval >= 98) {
if (Cs < 1349) {
return 4;
}
else if (Cs >= 1349) {
return 8;
}
}
}
else if (Cs >= 1453) {
if (AFD < 10) {
if (diffMVdiffVal < 1) {
if (MVDiff < 6) {
if (Rs < 2705) {
return 8;
}
else if (Rs >= 2705) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (AFD >= 10) {
if (posBalance < 59) {
if (diffAFD < -4) {
if (AFD < 12) {
if (Rs < 3064) {
return 4;
}
else if (Rs >= 3064) {
return 2;
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (posBalance >= 59) {
if (mu_mv_mag_sq < 29) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
if (Rs < 2448) {
return 2;
}
else if (Rs >= 2448) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (Rs < 2204) {
return 4;
}
else if (Rs >= 2204) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3200) {
if (Cs < 2085) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (ssDCval < 60) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (Rs < 7137) {
return 8;
}
else if (Rs >= 7137) {
return 2;
}
}
}
else if (ssDCval >= 60) {
if (ssDCval < 101) {
if (diffMVdiffVal < -11) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (diffMVdiffVal >= -11) {
if (posBalance < 48) {
if (negBalance < 39) {
if (Cs < 1848) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (Cs >= 1848) {
return 8;
}
}
else if (negBalance >= 39) {
return 2;
}
}
else if (posBalance >= 48) {
return 2;
}
}
}
else if (ssDCval >= 101) {
if (RsDiff < 11) {
if (MVDiff < 6) {
if (Cs < 1777) {
return 1;
}
else if (Cs >= 1777) {
if (Rs < 3945) {
return 4;
}
else if (Rs >= 3945) {
return 2;
}
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (RsDiff >= 11) {
return 2;
}
}
}
}
}
else if (Cs >= 2085) {
if (negBalance < 35) {
if (SC < 6340) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
else if (SC >= 6340) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 2;
}
}
}
else if (negBalance >= 35) {
if (MVDiff < 10) {
if (Cs < 3916) {
if (Rs < 4252) {
if (SC < 6533) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (SC >= 6533) {
return 4;
}
}
else if (Rs >= 4252) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (Cs >= 3916) {
return 8;
}
}
else if (MVDiff >= 10) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (mcTcor < 93) {
if (ssDCval < 136) {
return 8;
}
else if (ssDCval >= 136) {
return 1;
}
}
else if (mcTcor >= 93) {
if (Rs < 5023) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (MVDiff < 57) {
if (Rs < 4464) {
return 8;
}
else if (Rs >= 4464) {
if (Cs < 1566) {
return 1;
}
else if (Cs >= 1566) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 8;
}
}
}
}
else if (MVDiff >= 57) {
return 2;
}
}
}
else if (Rs >= 5023) {
if (Cs < 2886) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
if (Rs < 5736) {
return 1;
}
else if (Rs >= 5736) {
return 8;
}
}
}
else if (Cs >= 2886) {
if (Cs < 3913) {
return 4;
}
else if (Cs >= 3913) {
return 8;
}
}
}
}
}
}
}
else if (CsDiff >= 8) {
if (AFD < 12) {
if (Rs < 3783) {
if (ssDCval < 94) {
if (Cs < 1727) {
if (mu_mv_mag_sq < 41) {
return 1;
}
else if (mu_mv_mag_sq >= 41) {
if (CsDiff < 9) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
else if (CsDiff >= 9) {
if (ssDCval < 45) {
if (mcTcor < 61) {
return 1;
}
else if (mcTcor >= 61) {
return 8;
}
}
else if (ssDCval >= 45) {
if (AFD < 11) {
if (Rs < 1326) {
return 4;
}
else if (Rs >= 1326) {
return 1;
}
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
else if (Cs >= 1727) {
if (posBalance < 55) {
if (ssDCval < 85) {
if (Rs < 1564) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (Rs >= 1564) {
if (posBalance < 49) {
if (MVDiff < 6) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
return 1;
}
}
else if (MVDiff >= 6) {
if (RsDiff < 11) {
if (ssDCval < 73) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (RsDiff >= 11) {
if (RsCsDiff < 267) {
return 8;
}
else if (RsCsDiff >= 267) {
return 1;
}
}
}
}
else if (posBalance >= 49) {
if (TSC < 11) {
if (tcor < 89) {
if (Rs < 2598) {
if (SC < 3672) {
return 8;
}
else if (SC >= 3672) {
return 4;
}
}
else if (Rs >= 2598) {
if (Cs < 3201) {
return 8;
}
else if (Cs >= 3201) {
return 4;
}
}
}
else if (tcor >= 89) {
return 4;
}
}
else if (TSC >= 11) {
return 8;
}
}
}
}
else if (ssDCval >= 85) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (Cs < 2783) {
if (ssDCval < 51) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
return 4;
}
}
else if (ssDCval >= 51) {
if (Rs < 3105) {
return 8;
}
else if (Rs >= 3105) {
if (refDCval < 81) {
return 8;
}
else if (refDCval >= 81) {
return 4;
}
}
}
}
else if (Cs >= 2783) {
if (negBalance < 53) {
if (posBalance < 60) {
if (SC < 6879) {
return 4;
}
else if (SC >= 6879) {
return 1;
}
}
else if (posBalance >= 60) {
if (Rs < 2901) {
return 4;
}
else if (Rs >= 2901) {
return 8;
}
}
}
else if (negBalance >= 53) {
return 8;
}
}
}
}
}
else if (ssDCval >= 94) {
if (negBalance < 34) {
if (RsDiff < 10) {
if (mu_mv_mag_sq < 59) {
if (Rs < 1780) {
return 2;
}
else if (Rs >= 1780) {
if (Cs < 3859) {
return 4;
}
else if (Cs >= 3859) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (Rs < 1727) {
if (Rs < 1526) {
return 2;
}
else if (Rs >= 1526) {
return 1;
}
}
else if (Rs >= 1727) {
if (ssDCval < 105) {
if (Rs < 2541) {
return 2;
}
else if (Rs >= 2541) {
return 1;
}
}
else if (ssDCval >= 105) {
if (mu_mv_mag_sq < 62) {
if (Rs < 3107) {
return 4;
}
else if (Rs >= 3107) {
return 8;
}
}
else if (mu_mv_mag_sq >= 62) {
return 8;
}
}
}
}
}
else if (RsDiff >= 10) {
if (ssDCval < 114) {
if (Cs < 2461) {
return 8;
}
else if (Cs >= 2461) {
return 2;
}
}
else if (ssDCval >= 114) {
if (MVDiff < 28) {
if (Rs < 2692) {
if (Cs < 2604) {
return 8;
}
else if (Cs >= 2604) {
return 1;
}
}
else if (Rs >= 2692) {
return 4;
}
}
else if (MVDiff >= 28) {
return 2;
}
}
}
}
else if (negBalance >= 34) {
if (Rs < 3023) {
if (posBalance < 43) {
if (Rs < 2791) {
return 4;
}
else if (Rs >= 2791) {
return 8;
}
}
else if (posBalance >= 43) {
return 8;
}
}
else if (Rs >= 3023) {
if (ssDCval < 104) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (ssDCval >= 104) {
if (Cs < 2864) {
if (SC < 5388) {
return 8;
}
else if (SC >= 5388) {
return 1;
}
}
else if (Cs >= 2864) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3783) {
if (RsDiff < 10) {
if (TSC < 9) {
if (posBalance < 40) {
if (Rs < 4587) {
return 8;
}
else if (Rs >= 4587) {
return 1;
}
}
else if (posBalance >= 40) {
if (ssDCval < 109) {
if (Rs < 5971) {
if (Cs < 2311) {
if (Cs < 2133) {
return 1;
}
else if (Cs >= 2133) {
return 2;
}
}
else if (Cs >= 2311) {
if (SC < 7092) {
return 4;
}
else if (SC >= 7092) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
else if (Rs >= 5971) {
return 2;
}
}
else if (ssDCval >= 109) {
if (Rs < 5245) {
return 8;
}
else if (Rs >= 5245) {
if (Cs < 3468) {
return 4;
}
else if (Cs >= 3468) {
if (Rs < 6413) {
return 1;
}
else if (Rs >= 6413) {
return 8;
}
}
}
}
}
}
else if (TSC >= 9) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
if (Cs < 3663) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (mu_mv_mag_sq < 92) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
}
else if (Cs >= 3663) {
return 8;
}
}
}
}
else if (RsDiff >= 10) {
if (ssDCval < 100) {
if (ssDCval < 72) {
if (tcor < 88) {
if (Cs < 1971) {
return 1;
}
else if (Cs >= 1971) {
if (SC < 8698) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (SC >= 8698) {
return 8;
}
}
}
else if (tcor >= 88) {
return 8;
}
}
else if (ssDCval >= 72) {
if (TSC < 8) {
if (MVDiff < 23) {
return 2;
}
else if (MVDiff >= 23) {
return 8;
}
}
else if (TSC >= 8) {
if (mu_mv_mag_sq < 76) {
if (Cs < 2756) {
if (Cs < 1681) {
return 4;
}
else if (Cs >= 1681) {
return 2;
}
}
else if (Cs >= 2756) {
if (Rs < 4138) {
return 4;
}
else if (Rs >= 4138) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 76) {
return 4;
}
}
}
}
else if (ssDCval >= 100) {
if (Rs < 5916) {
if (RsCsDiff < 183) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
return 1;
}
}
}
else if (RsCsDiff >= 183) {
if (RsDiff < 15) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 5339) {
return 1;
}
else if (Rs >= 5339) {
return 8;
}
}
}
else if (RsDiff >= 15) {
return 2;
}
}
}
else if (Rs >= 5916) {
if (Cs < 2729) {
return 2;
}
else if (Cs >= 2729) {
if (Cs < 3597) {
return 1;
}
else if (Cs >= 3597) {
if (Rs < 8568) {
if (diffMVdiffVal < -1) {
if (Rs < 6121) {
return 4;
}
else if (Rs >= 6121) {
if (Rs < 6367) {
return 8;
}
else if (Rs >= 6367) {
return 1;
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (Rs >= 8568) {
return 1;
}
}
}
}
}
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 71) {
if (RsDiff < 17) {
if (ssDCval < 101) {
if (tcor < 64) {
if (AFD < 21) {
if (gchDC < 2) {
if (negBalance < 57) {
if (tcor < 56) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
if (Rs < 1636) {
return 4;
}
else if (Rs >= 1636) {
if (Rs < 1784) {
return 1;
}
else if (Rs >= 1784) {
return 2;
}
}
}
}
else if (tcor >= 56) {
if (negBalance < 53) {
if (MVDiff < 34) {
return 8;
}
else if (MVDiff >= 34) {
if (Cs < 3160) {
return 4;
}
else if (Cs >= 3160) {
return 2;
}
}
}
else if (negBalance >= 53) {
if (Rs < 2287) {
if (Cs < 1638) {
if (Rs < 1556) {
return 4;
}
else if (Rs >= 1556) {
if (Rs < 1709) {
return 1;
}
else if (Rs >= 1709) {
return 8;
}
}
}
else if (Cs >= 1638) {
return 4;
}
}
else if (Rs >= 2287) {
if (Cs < 3055) {
return 8;
}
else if (Cs >= 3055) {
return 4;
}
}
}
}
}
else if (negBalance >= 57) {
if (AFD < 19) {
if (Cs < 1853) {
if (Rs < 1246) {
return 2;
}
else if (Rs >= 1246) {
return 4;
}
}
else if (Cs >= 1853) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (TSC < 13) {
if (Rs < 1484) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (Rs >= 1484) {
if (mu_mv_mag_sq < 7) {
if (Rs < 2084) {
return 1;
}
else if (Rs >= 2084) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffAFD < 0) {
return 1;
}
else if (diffAFD >= 0) {
return 8;
}
}
}
}
else if (TSC >= 13) {
if (RsCsDiff < 174) {
return 1;
}
else if (RsCsDiff >= 174) {
if (Cs < 1868) {
return 8;
}
else if (Cs >= 1868) {
return 2;
}
}
}
}
}
else if (AFD >= 21) {
if (SC < 6147) {
if (RsCsDiff < 314) {
if (Rs < 2146) {
if (Cs < 1796) {
if (MVDiff < 27) {
if (CsDiff < 10) {
if (mcTcor < 85) {
if (Rs < 1510) {
return 4;
}
else if (Rs >= 1510) {
return 2;
}
}
else if (mcTcor >= 85) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (CsDiff >= 10) {
if (diffMVdiffVal < -1) {
if (CsDiff < 12) {
return 8;
}
else if (CsDiff >= 12) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
else if (MVDiff >= 27) {
if (posBalance < 65) {
if (negBalance < 57) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
return 8;
}
}
else if (negBalance >= 57) {
if (diffAFD < 1) {
if (diffRsCsdiff < -37) {
return 4;
}
else if (diffRsCsdiff >= -37) {
return 1;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
else if (posBalance >= 65) {
if (tcor < 59) {
if (AFD < 28) {
return 1;
}
else if (AFD >= 28) {
return 4;
}
}
else if (tcor >= 59) {
if (AFD < 27) {
return 4;
}
else if (AFD >= 27) {
return 2;
}
}
}
}
}
else if (Cs >= 1796) {
if (refDCval < 61) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 1;
}
}
else if (refDCval >= 61) {
if (diffAFD < 4) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
if (Cs < 2532) {
return 4;
}
else if (Cs >= 2532) {
return 8;
}
}
}
}
else if (diffAFD >= 4) {
return 2;
}
}
}
}
else if (Rs >= 2146) {
if (diffAFD < 2) {
if (Cs < 1587) {
return 2;
}
else if (Cs >= 1587) {
if (Rs < 2922) {
if (SC < 3901) {
return 8;
}
else if (SC >= 3901) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
if (Rs < 2526) {
return 2;
}
else if (Rs >= 2526) {
return 1;
}
}
}
}
else if (Rs >= 2922) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 2532) {
return 8;
}
else if (Rs >= 2532) {
return 4;
}
}
}
}
else if (RsCsDiff >= 314) {
if (mu_mv_mag_sq < 7) {
if (tcor < 61) {
if (Cs < 1434) {
return 1;
}
else if (Cs >= 1434) {
if (RsCsDiff < 538) {
return 2;
}
else if (RsCsDiff >= 538) {
return 8;
}
}
}
else if (tcor >= 61) {
if (Rs < 2105) {
if (Rs < 1605) {
return 2;
}
else if (Rs >= 1605) {
return 4;
}
}
else if (Rs >= 2105) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (mcTcor < 85) {
if (MVDiff < 29) {
if (Rs < 2377) {
return 2;
}
else if (Rs >= 2377) {
return 1;
}
}
else if (MVDiff >= 29) {
return 8;
}
}
else if (mcTcor >= 85) {
if (AFD < 31) {
if (posBalance < 58) {
if (Rs < 1301) {
return 8;
}
else if (Rs >= 1301) {
return 1;
}
}
else if (posBalance >= 58) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
if (Rs < 2806) {
return 8;
}
else if (Rs >= 2806) {
return 4;
}
}
}
}
else if (AFD >= 31) {
if (Rs < 1510) {
return 2;
}
else if (Rs >= 1510) {
return 1;
}
}
}
}
}
}
else if (SC >= 6147) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
if (Rs < 1747) {
return 8;
}
else if (Rs >= 1747) {
return 2;
}
}
}
}
}
else if (tcor >= 64) {
if (posBalance < 50) {
if (negBalance < 50) {
if (AFD < 18) {
if (Rs < 2350) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 2258) {
return 2;
}
else if (Rs >= 2258) {
return 1;
}
}
}
}
else if (Rs >= 2350) {
if (mcTcor < 89) {
if (diffAFD < 1) {
if (Cs < 4183) {
if (Rs < 2703) {
return 2;
}
else if (Rs >= 2703) {
return 1;
}
}
else if (Cs >= 4183) {
return 4;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (mcTcor >= 89) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
return 2;
}
}
}
}
else if (AFD >= 18) {
if (mu_mv_mag_sq < 56) {
if (Rs < 2379) {
if (RsCsDiff < 212) {
return 2;
}
else if (RsCsDiff >= 212) {
return 8;
}
}
else if (Rs >= 2379) {
if (Rs < 5413) {
return 2;
}
else if (Rs >= 5413) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (negBalance < 41) {
return 4;
}
else if (negBalance >= 41) {
if (Rs < 2521) {
return 1;
}
else if (Rs >= 2521) {
return 2;
}
}
}
}
}
else if (negBalance >= 50) {
if (mu_mv_mag_sq < 22) {
if (RsCsDiff < 321) {
if (Rs < 1309) {
if (CsDiff < 13) {
return 2;
}
else if (CsDiff >= 13) {
return 1;
}
}
else if (Rs >= 1309) {
if (Cs < 2380) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 1;
}
}
}
else if (Cs >= 2380) {
return 4;
}
}
}
else if (RsCsDiff >= 321) {
if (diffAFD < -15) {
return 8;
}
else if (diffAFD >= -15) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 22) {
if (diffAFD < -21) {
return 4;
}
else if (diffAFD >= -21) {
if (SC < 2815) {
if (Rs < 1247) {
return 8;
}
else if (Rs >= 1247) {
return 4;
}
}
else if (SC >= 2815) {
return 8;
}
}
}
}
}
else if (posBalance >= 50) {
if (SC < 8269) {
if (mu_mv_mag_sq < 7) {
if (Rs < 3151) {
if (RsCsDiff < 283) {
if (ssDCval < 72) {
if (MVDiff < 9) {
if (Rs < 2683) {
return 1;
}
else if (Rs >= 2683) {
return 8;
}
}
else if (MVDiff >= 9) {
if (ssDCval < 55) {
if (Rs < 2052) {
return 2;
}
else if (Rs >= 2052) {
return 8;
}
}
else if (ssDCval >= 55) {
if (ssDCval < 67) {
if (negBalance < 47) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (negBalance >= 47) {
if (negBalance < 61) {
if (Cs < 1399) {
return 8;
}
else if (Cs >= 1399) {
if (posBalance < 51) {
if (Rs < 1638) {
return 4;
}
else if (Rs >= 1638) {
return 1;
}
}
else if (posBalance >= 51) {
return 4;
}
}
}
else if (negBalance >= 61) {
return 2;
}
}
}
else if (ssDCval >= 67) {
if (AFD < 18) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (refDCval < 71) {
return 4;
}
else if (refDCval >= 71) {
if (Rs < 2399) {
return 8;
}
else if (Rs >= 2399) {
return 4;
}
}
}
}
else if (AFD >= 18) {
if (posBalance < 55) {
if (Cs < 1676) {
return 2;
}
else if (Cs >= 1676) {
return 8;
}
}
else if (posBalance >= 55) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 72) {
if (mcTcor < 94) {
if (refDCval < 79) {
if (Cs < 2011) {
if (diffRsCsdiff < -3) {
if (negBalance < 60) {
if (Rs < 2844) {
return 4;
}
else if (Rs >= 2844) {
return 2;
}
}
else if (negBalance >= 60) {
return 2;
}
}
else if (diffRsCsdiff >= -3) {
if (posBalance < 58) {
if (MVDiff < 23) {
if (negBalance < 54) {
return 4;
}
else if (negBalance >= 54) {
if (Rs < 1139) {
return 4;
}
else if (Rs >= 1139) {
if (Rs < 1554) {
return 8;
}
else if (Rs >= 1554) {
return 1;
}
}
}
}
else if (MVDiff >= 23) {
if (Rs < 1792) {
return 4;
}
else if (Rs >= 1792) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (ssDCval < 76) {
if (RsCsDiff < 205) {
if (Rs < 1490) {
return 8;
}
else if (Rs >= 1490) {
return 1;
}
}
else if (RsCsDiff >= 205) {
return 2;
}
}
else if (ssDCval >= 76) {
if (Rs < 1572) {
return 2;
}
else if (Rs >= 1572) {
return 8;
}
}
}
}
}
else if (Cs >= 2011) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (CsDiff < 12) {
return 8;
}
else if (CsDiff >= 12) {
if (Cs < 2399) {
return 2;
}
else if (Cs >= 2399) {
return 8;
}
}
}
}
}
else if (refDCval >= 79) {
if (TSC < 15) {
if (ssDCval < 89) {
if (TSC < 10) {
if (RsDiff < 9) {
if (Rs < 2615) {
if (Cs < 1532) {
return 8;
}
else if (Cs >= 1532) {
return 4;
}
}
else if (Rs >= 2615) {
return 8;
}
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (TSC >= 10) {
if (RsDiff < 11) {
if (ssDCval < 86) {
if (Rs < 2708) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (Rs < 2473) {
return 4;
}
else if (Rs >= 2473) {
return 8;
}
}
}
else if (Rs >= 2708) {
if (Cs < 3407) {
return 4;
}
else if (Cs >= 3407) {
return 8;
}
}
}
else if (ssDCval >= 86) {
if (Cs < 3489) {
if (refDCval < 88) {
if (Cs < 2141) {
return 8;
}
else if (Cs >= 2141) {
return 4;
}
}
else if (refDCval >= 88) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 2282) {
return 4;
}
else if (Rs >= 2282) {
return 1;
}
}
}
}
else if (Cs >= 3489) {
return 8;
}
}
}
else if (RsDiff >= 11) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
return 4;
}
}
}
}
else if (ssDCval >= 89) {
if (SC < 3663) {
if (posBalance < 66) {
if (refDCval < 89) {
return 2;
}
else if (refDCval >= 89) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (Rs < 1601) {
return 8;
}
else if (Rs >= 1601) {
if (Rs < 1896) {
return 1;
}
else if (Rs >= 1896) {
return 8;
}
}
}
}
}
else if (posBalance >= 66) {
if (Cs < 1608) {
return 1;
}
else if (Cs >= 1608) {
if (Rs < 1049) {
return 2;
}
else if (Rs >= 1049) {
return 4;
}
}
}
}
else if (SC >= 3663) {
if (tcor < 78) {
if (diffAFD < 5) {
if (mcTcor < 85) {
return 4;
}
else if (mcTcor >= 85) {
return 8;
}
}
else if (diffAFD >= 5) {
return 4;
}
}
else if (tcor >= 78) {
if (RsCsDiff < 221) {
if (mcTcor < 89) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
else if (mcTcor >= 89) {
return 8;
}
}
else if (RsCsDiff >= 221) {
return 4;
}
}
}
}
}
else if (TSC >= 15) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
return 8;
}
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 2988) {
if (mcTcor < 97) {
if (negBalance < 56) {
if (tcor < 84) {
if (Cs < 1392) {
if (Rs < 1299) {
return 1;
}
else if (Rs >= 1299) {
return 8;
}
}
else if (Cs >= 1392) {
if (AFD < 22) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2262) {
return 4;
}
else if (Rs >= 2262) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
else if (AFD >= 22) {
if (Rs < 2091) {
if (Rs < 1717) {
return 1;
}
else if (Rs >= 1717) {
return 2;
}
}
else if (Rs >= 2091) {
return 8;
}
}
}
}
else if (tcor >= 84) {
if (Rs < 1930) {
if (AFD < 16) {
if (Rs < 1247) {
return 8;
}
else if (Rs >= 1247) {
return 1;
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (Rs >= 1930) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
else if (mcTcor >= 97) {
return 2;
}
}
else if (Rs >= 2988) {
if (Rs < 3149) {
return 1;
}
else if (Rs >= 3149) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 283) {
if (refDCval < 92) {
if (negBalance < 55) {
if (diffRsCsdiff < -76) {
if (Rs < 2680) {
return 1;
}
else if (Rs >= 2680) {
return 4;
}
}
else if (diffRsCsdiff >= -76) {
if (posBalance < 65) {
return 2;
}
else if (posBalance >= 65) {
if (Rs < 1815) {
return 2;
}
else if (Rs >= 1815) {
return 4;
}
}
}
}
else if (negBalance >= 55) {
if (TSC < 11) {
if (Rs < 1966) {
return 2;
}
else if (Rs >= 1966) {
return 4;
}
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (refDCval >= 92) {
if (RsDiff < 13) {
if (tcor < 66) {
return 8;
}
else if (tcor >= 66) {
return 4;
}
}
else if (RsDiff >= 13) {
return 2;
}
}
}
}
else if (Rs >= 3151) {
if (negBalance < 52) {
if (AFD < 16) {
if (MVDiff < 15) {
if (Cs < 2431) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (Cs >= 2431) {
if (RsCsDiff < 255) {
if (diffRsCsdiff < -21) {
return 4;
}
else if (diffRsCsdiff >= -21) {
return 2;
}
}
else if (RsCsDiff >= 255) {
if (Cs < 2519) {
return 2;
}
else if (Cs >= 2519) {
return 4;
}
}
}
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (AFD >= 16) {
if (TSC < 12) {
if (Rs < 3570) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 4;
}
}
else if (Rs >= 3570) {
return 2;
}
}
else if (TSC >= 12) {
if (RsDiff < 15) {
if (Cs < 3747) {
return 1;
}
else if (Cs >= 3747) {
return 8;
}
}
else if (RsDiff >= 15) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (ssDCval < 91) {
if (TSC < 11) {
if (ssDCval < 82) {
if (Cs < 3151) {
return 2;
}
else if (Cs >= 3151) {
return 4;
}
}
else if (ssDCval >= 82) {
if (Cs < 3177) {
return 4;
}
else if (Cs >= 3177) {
return 2;
}
}
}
else if (TSC >= 11) {
if (Cs < 2541) {
if (Rs < 3184) {
return 2;
}
else if (Rs >= 3184) {
if (diffMVdiffVal < -5) {
return 8;
}
else if (diffMVdiffVal >= -5) {
return 4;
}
}
}
else if (Cs >= 2541) {
if (Rs < 3804) {
if (ssDCval < 76) {
if (Rs < 3385) {
return 8;
}
else if (Rs >= 3385) {
return 1;
}
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (Rs >= 3804) {
if (Rs < 4058) {
return 2;
}
else if (Rs >= 4058) {
return 4;
}
}
}
}
}
else if (ssDCval >= 91) {
if (Rs < 4503) {
if (Cs < 1618) {
return 8;
}
else if (Cs >= 1618) {
return 4;
}
}
else if (Rs >= 4503) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (mu_mv_mag_sq < 57) {
if (RsDiff < 6) {
if (mu_mv_mag_sq < 37) {
if (Rs < 811) {
if (Rs < 744) {
return 2;
}
else if (Rs >= 744) {
return 8;
}
}
else if (Rs >= 811) {
return 4;
}
}
else if (mu_mv_mag_sq >= 37) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (RsCsDiff < 267) {
if (mcTcor < 79) {
if (SC < 4398) {
if (AFD < 16) {
if (Rs < 2194) {
return 2;
}
else if (Rs >= 2194) {
return 8;
}
}
else if (AFD >= 16) {
return 1;
}
}
else if (SC >= 4398) {
if (Rs < 2835) {
return 4;
}
else if (Rs >= 2835) {
return 8;
}
}
}
else if (mcTcor >= 79) {
if (AFD < 25) {
if (ssDCval < 72) {
if (TSC < 8) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (ssDCval >= 72) {
if (mu_mv_mag_sq < 49) {
if (tcor < 78) {
if (mu_mv_mag_sq < 19) {
if (Cs < 1447) {
return 2;
}
else if (Cs >= 1447) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 19) {
if (ssDCval < 85) {
if (RsCsDiff < 149) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 1;
}
}
else if (RsCsDiff >= 149) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
if (Rs < 1215) {
return 2;
}
else if (Rs >= 1215) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (Rs < 1684) {
return 4;
}
else if (Rs >= 1684) {
return 8;
}
}
}
}
else if (tcor >= 78) {
if (diffAFD < -1) {
if (mu_mv_mag_sq < 41) {
return 8;
}
else if (mu_mv_mag_sq >= 41) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (RsDiff < 8) {
if (Rs < 2100) {
return 4;
}
else if (Rs >= 2100) {
return 8;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
else if (diffAFD >= -1) {
if (negBalance < 60) {
if (Rs < 1603) {
if (Rs < 1520) {
return 8;
}
else if (Rs >= 1520) {
return 1;
}
}
else if (Rs >= 1603) {
return 8;
}
}
else if (negBalance >= 60) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (ssDCval < 82) {
if (tcor < 82) {
if (MVDiff < 40) {
return 1;
}
else if (MVDiff >= 40) {
return 8;
}
}
else if (tcor >= 82) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (ssDCval >= 82) {
if (Cs < 2781) {
if (RsCsDiff < 215) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
if (Rs < 2452) {
return 8;
}
else if (Rs >= 2452) {
return 2;
}
}
}
else if (RsCsDiff >= 215) {
if (posBalance < 63) {
if (Cs < 2564) {
return 8;
}
else if (Cs >= 2564) {
return 2;
}
}
else if (posBalance >= 63) {
return 2;
}
}
}
else if (Cs >= 2781) {
if (Rs < 3471) {
return 4;
}
else if (Rs >= 3471) {
return 8;
}
}
}
}
}
}
else if (AFD >= 25) {
return 4;
}
}
}
else if (RsCsDiff >= 267) {
if (refDCval < 96) {
if (ssDCval < 65) {
if (Rs < 2518) {
return 4;
}
else if (Rs >= 2518) {
if (Rs < 2807) {
return 8;
}
else if (Rs >= 2807) {
return 2;
}
}
}
else if (ssDCval >= 65) {
if (CsDiff < 9) {
if (MVDiff < 35) {
return 4;
}
else if (MVDiff >= 35) {
return 8;
}
}
else if (CsDiff >= 9) {
if (mu_mv_mag_sq < 44) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
if (refDCval < 82) {
return 8;
}
else if (refDCval >= 82) {
if (Rs < 1579) {
return 2;
}
else if (Rs >= 1579) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (diffMVdiffVal < -1) {
if (Cs < 3078) {
return 2;
}
else if (Cs >= 3078) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 54) {
if (Rs < 3966) {
return 4;
}
else if (Rs >= 3966) {
return 2;
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 96) {
if (Cs < 2564) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 4;
}
}
}
else if (Cs >= 2564) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (refDCval < 95) {
if (posBalance < 56) {
if (AFD < 16) {
if (Cs < 3029) {
if (diffRsCsdiff < -2) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
return 4;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (Cs >= 3029) {
return 1;
}
}
else if (AFD >= 16) {
return 1;
}
}
else if (posBalance >= 56) {
if (TSC < 13) {
if (RsDiff < 9) {
if (negBalance < 55) {
if (RsDiff < 8) {
if (posBalance < 61) {
if (Rs < 2061) {
if (Rs < 1381) {
return 1;
}
else if (Rs >= 1381) {
return 8;
}
}
else if (Rs >= 2061) {
return 1;
}
}
else if (posBalance >= 61) {
return 4;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (negBalance >= 55) {
if (Rs < 1916) {
return 4;
}
else if (Rs >= 1916) {
return 2;
}
}
}
else if (RsDiff >= 9) {
if (mcTcor < 87) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 4;
}
}
}
else if (mcTcor >= 87) {
if (Cs < 3789) {
return 4;
}
else if (Cs >= 3789) {
return 2;
}
}
}
}
else if (TSC >= 13) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (mcTcor < 81) {
if (Cs < 2411) {
return 8;
}
else if (Cs >= 2411) {
return 4;
}
}
else if (mcTcor >= 81) {
if (Cs < 2415) {
return 4;
}
else if (Cs >= 2415) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 95) {
if (SC < 6019) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
if (Cs < 2926) {
return 2;
}
else if (Cs >= 2926) {
return 4;
}
}
}
else if (SC >= 6019) {
if (Rs < 4524) {
return 4;
}
else if (Rs >= 4524) {
return 8;
}
}
}
}
}
}
else if (SC >= 8269) {
if (diffTSC < -3) {
return 2;
}
else if (diffTSC >= -3) {
return 8;
}
}
}
}
}
else if (ssDCval >= 101) {
if (AFD < 20) {
if (ssDCval < 110) {
if (mcTcor < 94) {
if (Rs < 3531) {
if (Rs < 2965) {
if (diffAFD < 2) {
if (tcor < 60) {
return 4;
}
else if (tcor >= 60) {
if (Cs < 1453) {
if (Rs < 2071) {
return 8;
}
else if (Rs >= 2071) {
return 4;
}
}
else if (Cs >= 1453) {
if (MVDiff < 5) {
if (Rs < 2717) {
return 8;
}
else if (Rs >= 2717) {
if (Rs < 2826) {
return 4;
}
else if (Rs >= 2826) {
return 8;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (diffAFD >= 2) {
if (SC < 5169) {
return 8;
}
else if (SC >= 5169) {
return 4;
}
}
}
else if (Rs >= 2965) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Cs < 2512) {
if (Rs < 3236) {
if (Rs < 3128) {
return 8;
}
else if (Rs >= 3128) {
return 4;
}
}
else if (Rs >= 3236) {
return 2;
}
}
else if (Cs >= 2512) {
return 4;
}
}
}
}
else if (Rs >= 3531) {
if (diffRsCsdiff < 87) {
return 8;
}
else if (diffRsCsdiff >= 87) {
if (mcTcor < 93) {
if (refDCval < 101) {
return 4;
}
else if (refDCval >= 101) {
return 8;
}
}
else if (mcTcor >= 93) {
if (Rs < 5208) {
return 2;
}
else if (Rs >= 5208) {
return 4;
}
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 4791) {
if (Rs < 1687) {
if (Rs < 1333) {
if (Rs < 1199) {
return 2;
}
else if (Rs >= 1199) {
return 1;
}
}
else if (Rs >= 1333) {
return 4;
}
}
else if (Rs >= 1687) {
if (RsDiff < 14) {
if (mu_mv_mag_sq < 56) {
return 8;
}
else if (mu_mv_mag_sq >= 56) {
return 2;
}
}
else if (RsDiff >= 14) {
return 2;
}
}
}
else if (Rs >= 4791) {
return 4;
}
}
}
else if (ssDCval >= 110) {
if (Rs < 5303) {
if (mcTcor < 84) {
if (Cs < 2840) {
if (negBalance < 58) {
if (MVDiff < 107) {
return 8;
}
else if (MVDiff >= 107) {
return 2;
}
}
else if (negBalance >= 58) {
if (SC < 3529) {
return 8;
}
else if (SC >= 3529) {
if (Rs < 1984) {
return 2;
}
else if (Rs >= 1984) {
return 1;
}
}
}
}
else if (Cs >= 2840) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (negBalance < 45) {
if (Rs < 2721) {
return 1;
}
else if (Rs >= 2721) {
if (Cs < 3933) {
return 8;
}
else if (Cs >= 3933) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (Rs < 3896) {
return 8;
}
else if (Rs >= 3896) {
return 1;
}
}
}
}
}
else if (mcTcor >= 84) {
if (negBalance < 43) {
if (RsDiff < 9) {
if (Cs < 1813) {
return 4;
}
else if (Cs >= 1813) {
if (Cs < 2877) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (Rs < 2415) {
return 1;
}
else if (Rs >= 2415) {
return 4;
}
}
}
else if (Cs >= 2877) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (posBalance < 55) {
if (CsDiff < 11) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (CsDiff >= 11) {
if (RsDiff < 14) {
if (Rs < 4187) {
return 4;
}
else if (Rs >= 4187) {
return 2;
}
}
else if (RsDiff >= 14) {
if (Rs < 4318) {
return 2;
}
else if (Rs >= 4318) {
return 1;
}
}
}
}
else if (posBalance >= 55) {
if (Rs < 1582) {
return 4;
}
else if (Rs >= 1582) {
return 8;
}
}
}
}
else if (negBalance >= 43) {
if (SC < 4953) {
if (TSC < 13) {
if (SC < 3829) {
if (RsCsDiff < 183) {
if (negBalance < 44) {
if (SC < 3522) {
return 8;
}
else if (SC >= 3522) {
if (Rs < 1742) {
return 1;
}
else if (Rs >= 1742) {
return 2;
}
}
}
else if (negBalance >= 44) {
return 8;
}
}
else if (RsCsDiff >= 183) {
if (ssDCval < 125) {
return 1;
}
else if (ssDCval >= 125) {
if (Rs < 1668) {
return 2;
}
else if (Rs >= 1668) {
return 8;
}
}
}
}
else if (SC >= 3829) {
if (posBalance < 50) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 1;
}
}
else if (posBalance >= 50) {
return 8;
}
}
}
else if (TSC >= 13) {
if (Cs < 1809) {
return 1;
}
else if (Cs >= 1809) {
return 8;
}
}
}
else if (SC >= 4953) {
if (ssDCval < 126) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (posBalance < 50) {
if (Rs < 4344) {
return 8;
}
else if (Rs >= 4344) {
if (Rs < 4831) {
return 2;
}
else if (Rs >= 4831) {
return 1;
}
}
}
else if (posBalance >= 50) {
return 8;
}
}
}
else if (ssDCval >= 126) {
if (Cs < 3154) {
if (TSC < 11) {
if (ssDCval < 139) {
if (Rs < 3332) {
if (Rs < 2470) {
return 8;
}
else if (Rs >= 2470) {
return 4;
}
}
else if (Rs >= 3332) {
return 8;
}
}
else if (ssDCval >= 139) {
return 4;
}
}
else if (TSC >= 11) {
if (RsCsDiff < 258) {
return 8;
}
else if (RsCsDiff >= 258) {
if (Cs < 2107) {
return 8;
}
else if (Cs >= 2107) {
return 4;
}
}
}
}
else if (Cs >= 3154) {
return 4;
}
}
}
}
}
}
else if (Rs >= 5303) {
if (ssDCval < 124) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (Rs < 8298) {
return 8;
}
else if (Rs >= 8298) {
return 1;
}
}
}
else if (ssDCval >= 124) {
if (gchDC < 9) {
return 8;
}
else if (gchDC >= 9) {
return 1;
}
}
}
}
}
else if (AFD >= 20) {
if (Rs < 1198) {
if (gchDC < 13) {
if (negBalance < 54) {
if (Cs < 2034) {
if (AFD < 25) {
return 1;
}
else if (AFD >= 25) {
return 4;
}
}
else if (Cs >= 2034) {
if (Rs < 929) {
return 2;
}
else if (Rs >= 929) {
return 8;
}
}
}
else if (negBalance >= 54) {
return 2;
}
}
else if (gchDC >= 13) {
return 8;
}
}
else if (Rs >= 1198) {
if (refDCval < 158) {
if (refDCval < 114) {
if (negBalance < 59) {
if (RsCsDiff < 351) {
if (tcor < 78) {
if (TSC < 13) {
if (RsCsDiff < 290) {
if (tcor < 67) {
return 4;
}
else if (tcor >= 67) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (Rs < 1330) {
return 8;
}
else if (Rs >= 1330) {
return 4;
}
}
}
}
else if (RsCsDiff >= 290) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
if (Cs < 2098) {
if (Rs < 2609) {
return 1;
}
else if (Rs >= 2609) {
return 4;
}
}
else if (Cs >= 2098) {
return 2;
}
}
}
}
else if (TSC >= 13) {
if (MVDiff < 14) {
if (Rs < 4013) {
return 4;
}
else if (Rs >= 4013) {
return 1;
}
}
else if (MVDiff >= 14) {
if (ssDCval < 126) {
if (MVDiff < 28) {
if (TSC < 14) {
if (AFD < 29) {
return 4;
}
else if (AFD >= 29) {
return 2;
}
}
else if (TSC >= 14) {
return 8;
}
}
else if (MVDiff >= 28) {
return 8;
}
}
else if (ssDCval >= 126) {
if (Rs < 1572) {
return 4;
}
else if (Rs >= 1572) {
return 8;
}
}
}
}
}
else if (tcor >= 78) {
if (Rs < 2305) {
if (RsCsDiff < 193) {
return 1;
}
else if (RsCsDiff >= 193) {
return 2;
}
}
else if (Rs >= 2305) {
if (MVDiff < 23) {
if (Rs < 4470) {
return 2;
}
else if (Rs >= 4470) {
return 4;
}
}
else if (MVDiff >= 23) {
return 8;
}
}
}
}
else if (RsCsDiff >= 351) {
if (mcTcor < 80) {
return 8;
}
else if (mcTcor >= 80) {
if (refDCval < 104) {
return 8;
}
else if (refDCval >= 104) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (ssDCval < 106) {
if (SC < 4868) {
return 2;
}
else if (SC >= 4868) {
return 8;
}
}
else if (ssDCval >= 106) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 59) {
if (AFD < 22) {
if (RsDiff < 15) {
return 4;
}
else if (RsDiff >= 15) {
return 8;
}
}
else if (AFD >= 22) {
if (gchDC < 3) {
if (diffAFD < -2) {
if (Cs < 1743) {
return 8;
}
else if (Cs >= 1743) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Cs < 1817) {
return 2;
}
else if (Cs >= 1817) {
if (Cs < 3779) {
return 1;
}
else if (Cs >= 3779) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
if (Cs < 3742) {
return 8;
}
else if (Cs >= 3742) {
return 2;
}
}
}
}
}
else if (refDCval >= 114) {
if (RsDiff < 10) {
if (negBalance < 54) {
if (Cs < 1648) {
return 4;
}
else if (Cs >= 1648) {
return 8;
}
}
else if (negBalance >= 54) {
if (tcor < 74) {
if (CsDiff < 12) {
return 2;
}
else if (CsDiff >= 12) {
return 8;
}
}
else if (tcor >= 74) {
if (CsDiff < 9) {
return 1;
}
else if (CsDiff >= 9) {
return 8;
}
}
}
}
else if (RsDiff >= 10) {
if (refDCval < 140) {
if (negBalance < 48) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
return 8;
}
}
else if (negBalance >= 48) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 14) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
if (posBalance < 60) {
if (TSC < 12) {
if (Rs < 5801) {
return 1;
}
else if (Rs >= 5801) {
return 8;
}
}
else if (TSC >= 12) {
if (Cs < 3664) {
return 8;
}
else if (Cs >= 3664) {
return 4;
}
}
}
else if (posBalance >= 60) {
if (Rs < 4790) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 1;
}
}
else if (Rs >= 4790) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (SC < 6023) {
if (Cs < 1384) {
return 1;
}
else if (Cs >= 1384) {
return 8;
}
}
else if (SC >= 6023) {
if (Rs < 4487) {
return 4;
}
else if (Rs >= 4487) {
return 1;
}
}
}
}
else if (gchDC >= 2) {
if (AFD < 22) {
if (diffMVdiffVal < -39) {
return 4;
}
else if (diffMVdiffVal >= -39) {
return 8;
}
}
else if (AFD >= 22) {
if (RsCsDiff < 358) {
if (tcor < 64) {
return 1;
}
else if (tcor >= 64) {
if (diffAFD < 2) {
if (Rs < 2093) {
if (Cs < 1316) {
return 8;
}
else if (Cs >= 1316) {
return 1;
}
}
else if (Rs >= 2093) {
return 8;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (RsCsDiff >= 358) {
return 8;
}
}
}
}
}
else if (refDCval >= 140) {
if (TSC < 16) {
return 8;
}
else if (TSC >= 16) {
if (MVDiff < 65) {
return 2;
}
else if (MVDiff >= 65) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 158) {
if (Rs < 3118) {
return 8;
}
else if (Rs >= 3118) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 17) {
if (Cs < 3910) {
if (MVDiff < 17) {
if (Cs < 1727) {
return 2;
}
else if (Cs >= 1727) {
if (negBalance < 50) {
if (tcor < 78) {
if (mu_mv_mag_sq < 44) {
if (Cs < 3576) {
if (MVDiff < 16) {
if (gchDC < 7) {
return 2;
}
else if (gchDC >= 7) {
return 4;
}
}
else if (MVDiff >= 16) {
return 4;
}
}
else if (Cs >= 3576) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
if (Rs < 4911) {
return 4;
}
else if (Rs >= 4911) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (tcor >= 78) {
if (Cs < 3137) {
return 2;
}
else if (Cs >= 3137) {
if (RsCsDiff < 611) {
if (Rs < 5081) {
return 2;
}
else if (Rs >= 5081) {
return 1;
}
}
else if (RsCsDiff >= 611) {
return 8;
}
}
}
}
else if (negBalance >= 50) {
if (RsDiff < 24) {
if (SC < 8166) {
if (TSC < 10) {
if (gchDC < 1) {
if (Rs < 4580) {
return 2;
}
else if (Rs >= 4580) {
return 4;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (TSC >= 10) {
if (Rs < 4056) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (AFD < 24) {
if (RsCsDiff < 417) {
return 2;
}
else if (RsCsDiff >= 417) {
return 8;
}
}
else if (AFD >= 24) {
if (tcor < 67) {
if (AFD < 30) {
return 4;
}
else if (AFD >= 30) {
return 1;
}
}
else if (tcor >= 67) {
return 8;
}
}
}
}
else if (Rs >= 4056) {
if (mu_mv_mag_sq < 7) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (TSC < 15) {
return 2;
}
else if (TSC >= 15) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (posBalance < 56) {
if (SC < 7768) {
if (Rs < 4339) {
return 4;
}
else if (Rs >= 4339) {
return 8;
}
}
else if (SC >= 7768) {
return 4;
}
}
else if (posBalance >= 56) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 4;
}
}
}
}
}
}
else if (SC >= 8166) {
if (RsCsDiff < 472) {
if (SC < 8534) {
return 4;
}
else if (SC >= 8534) {
return 2;
}
}
else if (RsCsDiff >= 472) {
return 4;
}
}
}
else if (RsDiff >= 24) {
if (Rs < 5926) {
return 2;
}
else if (Rs >= 5926) {
return 4;
}
}
}
}
}
else if (MVDiff >= 17) {
if (Cs < 1530) {
if (MVDiff < 45) {
if (refDCval < 143) {
if (diffAFD < -2) {
if (posBalance < 60) {
if (diffAFD < -4) {
if (Cs < 1436) {
return 1;
}
else if (Cs >= 1436) {
return 8;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (posBalance >= 60) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Rs < 3207) {
if (Rs < 2431) {
return 2;
}
else if (Rs >= 2431) {
return 1;
}
}
else if (Rs >= 3207) {
return 2;
}
}
}
else if (refDCval >= 143) {
return 1;
}
}
else if (MVDiff >= 45) {
if (Rs < 3086) {
if (Cs < 1415) {
return 1;
}
else if (Cs >= 1415) {
return 8;
}
}
else if (Rs >= 3086) {
return 1;
}
}
}
else if (Cs >= 1530) {
if (gchDC < 2) {
if (ssDCval < 141) {
if (ssDCval < 104) {
if (TSC < 14) {
if (SC < 5416) {
if (Rs < 2728) {
return 4;
}
else if (Rs >= 2728) {
return 2;
}
}
else if (SC >= 5416) {
if (Rs < 4741) {
if (Rs < 3669) {
if (CsDiff < 16) {
return 8;
}
else if (CsDiff >= 16) {
if (Rs < 3148) {
return 2;
}
else if (Rs >= 3148) {
return 4;
}
}
}
else if (Rs >= 3669) {
return 4;
}
}
else if (Rs >= 4741) {
if (TSC < 12) {
if (Rs < 5338) {
return 4;
}
else if (Rs >= 5338) {
return 8;
}
}
else if (TSC >= 12) {
return 2;
}
}
}
}
else if (TSC >= 14) {
if (AFD < 22) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (Rs < 4637) {
if (Rs < 3276) {
return 4;
}
else if (Rs >= 3276) {
return 8;
}
}
else if (Rs >= 4637) {
return 2;
}
}
}
else if (AFD >= 22) {
if (Cs < 2678) {
if (SC < 6545) {
return 2;
}
else if (SC >= 6545) {
return 1;
}
}
else if (Cs >= 2678) {
return 2;
}
}
}
}
else if (ssDCval >= 104) {
if (AFD < 28) {
if (ssDCval < 129) {
return 8;
}
else if (ssDCval >= 129) {
if (negBalance < 51) {
if (RsCsDiff < 473) {
return 8;
}
else if (RsCsDiff >= 473) {
return 2;
}
}
else if (negBalance >= 51) {
if (Rs < 4027) {
return 1;
}
else if (Rs >= 4027) {
if (Rs < 5015) {
return 4;
}
else if (Rs >= 5015) {
return 8;
}
}
}
}
}
else if (AFD >= 28) {
return 2;
}
}
}
else if (ssDCval >= 141) {
if (Rs < 3374) {
return 8;
}
else if (Rs >= 3374) {
if (ssDCval < 167) {
if (posBalance < 55) {
if (AFD < 28) {
return 8;
}
else if (AFD >= 28) {
return 4;
}
}
else if (posBalance >= 55) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
return 2;
}
}
}
else if (ssDCval >= 167) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 75) {
if (negBalance < 43) {
if (CsDiff < 15) {
return 2;
}
else if (CsDiff >= 15) {
return 4;
}
}
else if (negBalance >= 43) {
if (AFD < 21) {
if (Rs < 4179) {
return 8;
}
else if (Rs >= 4179) {
return 4;
}
}
else if (AFD >= 21) {
if (AFD < 31) {
if (refDCval < 67) {
return 4;
}
else if (refDCval >= 67) {
if (RsDiff < 19) {
return 4;
}
else if (RsDiff >= 19) {
return 8;
}
}
}
else if (AFD >= 31) {
if (AFD < 38) {
return 2;
}
else if (AFD >= 38) {
return 4;
}
}
}
}
}
else if (ssDCval >= 75) {
if (mu_mv_mag_sq < 30) {
if (mu_mv_mag_sq < 20) {
if (diffTSC < 2) {
if (mcTcor < 90) {
if (refDCval < 98) {
if (Rs < 4131) {
return 8;
}
else if (Rs >= 4131) {
if (Rs < 5414) {
return 2;
}
else if (Rs >= 5414) {
return 1;
}
}
}
else if (refDCval >= 98) {
return 8;
}
}
else if (mcTcor >= 90) {
if (AFD < 21) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 8;
}
}
else if (AFD >= 21) {
if (negBalance < 56) {
if (Rs < 4488) {
return 2;
}
else if (Rs >= 4488) {
return 4;
}
}
else if (negBalance >= 56) {
if (ssDCval < 99) {
if (Rs < 3593) {
return 8;
}
else if (Rs >= 3593) {
return 2;
}
}
else if (ssDCval >= 99) {
if (Rs < 3578) {
return 8;
}
else if (Rs >= 3578) {
return 1;
}
}
}
}
}
}
else if (diffTSC >= 2) {
if (mcTcor < 77) {
if (AFD < 31) {
return 4;
}
else if (AFD >= 31) {
if (AFD < 37) {
return 1;
}
else if (AFD >= 37) {
return 8;
}
}
}
else if (mcTcor >= 77) {
if (negBalance < 57) {
if (RsDiff < 24) {
if (Rs < 6668) {
return 8;
}
else if (Rs >= 6668) {
return 1;
}
}
else if (RsDiff >= 24) {
if (diffMVdiffVal < -13) {
return 8;
}
else if (diffMVdiffVal >= -13) {
return 2;
}
}
}
else if (negBalance >= 57) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 20) {
if (diffTSC < 3) {
if (TSC < 14) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 4;
}
}
else if (TSC >= 14) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
if (Rs < 4572) {
if (Rs < 2876) {
return 1;
}
else if (Rs >= 2876) {
return 8;
}
}
else if (Rs >= 4572) {
if (AFD < 27) {
return 1;
}
else if (AFD >= 27) {
return 2;
}
}
}
}
}
else if (diffTSC >= 3) {
if (ssDCval < 115) {
return 4;
}
else if (ssDCval >= 115) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (AFD < 31) {
if (Cs < 3837) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (refDCval < 90) {
return 8;
}
else if (refDCval >= 90) {
if (RsCsDiff < 475) {
return 8;
}
else if (RsCsDiff >= 475) {
if (AFD < 28) {
return 4;
}
else if (AFD >= 28) {
return 8;
}
}
}
}
}
else if (Cs >= 3837) {
return 2;
}
}
else if (AFD >= 31) {
if (mu_mv_mag_sq < 45) {
if (Cs < 3226) {
return 8;
}
else if (Cs >= 3226) {
if (Rs < 5352) {
return 2;
}
else if (Rs >= 5352) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (Rs < 3505) {
return 2;
}
else if (Rs >= 3505) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Cs >= 3910) {
if (TSC < 11) {
if (refDCval < 126) {
if (Rs < 7690) {
return 8;
}
else if (Rs >= 7690) {
if (RsDiff < 41) {
return 1;
}
else if (RsDiff >= 41) {
return 8;
}
}
}
else if (refDCval >= 126) {
return 8;
}
}
else if (TSC >= 11) {
if (mu_mv_mag_sq < 5) {
if (SC < 10952) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
if (AFD < 27) {
return 4;
}
else if (AFD >= 27) {
return 8;
}
}
}
else if (SC >= 10952) {
return 4;
}
}
else if (mu_mv_mag_sq >= 5) {
if (TSC < 20) {
if (RsCsDiff < 716) {
if (MVDiff < 22) {
if (posBalance < 47) {
if (Rs < 5463) {
return 4;
}
else if (Rs >= 5463) {
return 1;
}
}
else if (posBalance >= 47) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 4;
}
}
}
else if (MVDiff >= 22) {
return 2;
}
}
else if (RsCsDiff >= 716) {
if (gchDC < 3) {
if (negBalance < 39) {
return 8;
}
else if (negBalance >= 39) {
if (CsDiff < 19) {
if (Rs < 6374) {
return 8;
}
else if (Rs >= 6374) {
return 4;
}
}
else if (CsDiff >= 19) {
return 4;
}
}
}
else if (gchDC >= 3) {
if (CsDiff < 19) {
return 4;
}
else if (CsDiff >= 19) {
return 8;
}
}
}
}
else if (TSC >= 20) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (Cs < 3306) {
if (refDCval < 90) {
if (TSC < 10) {
if (RsDiff < 14) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Rs < 3360) {
return 4;
}
else if (Rs >= 3360) {
return 8;
}
}
}
else if (RsDiff >= 14) {
return 8;
}
}
else if (TSC >= 10) {
if (negBalance < 54) {
if (Cs < 2393) {
if (Cs < 1864) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
return 1;
}
}
else if (Cs >= 1864) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
if (Rs < 3329) {
return 4;
}
else if (Rs >= 3329) {
return 8;
}
}
}
}
else if (Cs >= 2393) {
if (diffRsCsdiff < -76) {
if (refDCval < 80) {
return 8;
}
else if (refDCval >= 80) {
if (Rs < 3695) {
return 4;
}
else if (Rs >= 3695) {
return 1;
}
}
}
else if (diffRsCsdiff >= -76) {
if (SC < 6826) {
return 8;
}
else if (SC >= 6826) {
if (SC < 6845) {
return 1;
}
else if (SC >= 6845) {
return 8;
}
}
}
}
}
else if (negBalance >= 54) {
if (diffAFD < 1) {
if (mu_mv_mag_sq < 81) {
if (TSC < 15) {
return 8;
}
else if (TSC >= 15) {
return 4;
}
}
else if (mu_mv_mag_sq >= 81) {
return 4;
}
}
else if (diffAFD >= 1) {
if (Rs < 3939) {
return 4;
}
else if (Rs >= 3939) {
return 1;
}
}
}
}
}
else if (refDCval >= 90) {
if (Cs < 1893) {
if (refDCval < 109) {
return 2;
}
else if (refDCval >= 109) {
return 4;
}
}
else if (Cs >= 1893) {
if (ssDCval < 100) {
if (Rs < 3887) {
if (Cs < 2583) {
if (Rs < 1594) {
return 4;
}
else if (Rs >= 1594) {
if (Rs < 2606) {
return 8;
}
else if (Rs >= 2606) {
return 2;
}
}
}
else if (Cs >= 2583) {
return 4;
}
}
else if (Rs >= 3887) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 4510) {
return 4;
}
else if (Rs >= 4510) {
return 2;
}
}
}
}
else if (ssDCval >= 100) {
if (Rs < 5411) {
if (negBalance < 50) {
if (RsCsDiff < 347) {
if (diffAFD < -2) {
if (Rs < 2711) {
return 8;
}
else if (Rs >= 2711) {
return 4;
}
}
else if (diffAFD >= -2) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 1;
}
}
}
else if (RsCsDiff >= 347) {
return 1;
}
}
else if (negBalance >= 50) {
if (TSC < 11) {
if (Rs < 3582) {
return 4;
}
else if (Rs >= 3582) {
return 8;
}
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (Rs >= 5411) {
return 4;
}
}
}
}
}
else if (Cs >= 3306) {
if (negBalance < 50) {
if (ssDCval < 81) {
if (CsDiff < 20) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
if (mu_mv_mag_sq < 81) {
return 8;
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
return 8;
}
}
}
}
else if (CsDiff >= 20) {
if (Rs < 4880) {
return 1;
}
else if (Rs >= 4880) {
return 8;
}
}
}
else if (ssDCval >= 81) {
if (AFD < 17) {
if (RsCsDiff < 201) {
if (SC < 7984) {
return 4;
}
else if (SC >= 7984) {
return 8;
}
}
else if (RsCsDiff >= 201) {
if (mu_mv_mag_sq < 74) {
return 4;
}
else if (mu_mv_mag_sq >= 74) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (Rs < 4298) {
return 8;
}
else if (Rs >= 4298) {
return 4;
}
}
}
}
}
else if (AFD >= 17) {
if (Rs < 7322) {
if (TSC < 14) {
if (Cs < 3855) {
return 1;
}
else if (Cs >= 3855) {
return 8;
}
}
else if (TSC >= 14) {
return 2;
}
}
else if (Rs >= 7322) {
return 4;
}
}
}
}
else if (negBalance >= 50) {
if (posBalance < 56) {
if (tcor < 80) {
return 4;
}
else if (tcor >= 80) {
if (Rs < 6433) {
return 2;
}
else if (Rs >= 6433) {
return 8;
}
}
}
else if (posBalance >= 56) {
if (Rs < 3140) {
return 4;
}
else if (Rs >= 3140) {
if (ssDCval < 122) {
if (negBalance < 53) {
if (SC < 10004) {
return 8;
}
else if (SC >= 10004) {
return 4;
}
}
else if (negBalance >= 53) {
if (Rs < 3466) {
if (Cs < 4401) {
return 8;
}
else if (Cs >= 4401) {
return 4;
}
}
else if (Rs >= 3466) {
return 8;
}
}
}
else if (ssDCval >= 122) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1633) {
if (Rs < 3092) {
if (Rs < 1455) {
if (ssDCval < 122) {
if (negBalance < 1) {
return 2;
}
else if (negBalance >= 1) {
if (ssDCval < 63) {
if (negBalance < 27) {
if (mu_mv_mag_sq < 128) {
if (TSCindex < 3) {
if (posBalance < 44) {
if (ssDCval < 47) {
if (Rs < 958) {
return 1;
}
else if (Rs >= 958) {
return 8;
}
}
else if (ssDCval >= 47) {
return 1;
}
}
else if (posBalance >= 44) {
if (Rs < 967) {
return 1;
}
else if (Rs >= 967) {
return 4;
}
}
}
else if (TSCindex >= 3) {
if (Rs < 762) {
return 1;
}
else if (Rs >= 762) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 1202) {
return 8;
}
else if (Rs >= 1202) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1498) {
if (Cs < 1343) {
return 4;
}
else if (Cs >= 1343) {
return 8;
}
}
else if (Cs >= 1498) {
return 1;
}
}
}
else if (negBalance >= 27) {
if (SC < 2777) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (SC >= 2777) {
if (negBalance < 33) {
return 4;
}
else if (negBalance >= 33) {
if (Rs < 1342) {
return 4;
}
else if (Rs >= 1342) {
return 8;
}
}
}
}
}
else if (ssDCval >= 63) {
if (Cs < 1419) {
if (RsDiff < 1) {
if (Rs < 400) {
return 4;
}
else if (Rs >= 400) {
return 1;
}
}
else if (RsDiff >= 1) {
if (MVDiff < 3) {
if (diffTSC < -1) {
if (Rs < 1418) {
return 4;
}
else if (Rs >= 1418) {
return 8;
}
}
else if (diffTSC >= -1) {
if (diffTSC < 1) {
if (Rs < 1038) {
if (Rs < 668) {
return 8;
}
else if (Rs >= 668) {
return 2;
}
}
else if (Rs >= 1038) {
return 8;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (Cs >= 1419) {
if (SC < 1942) {
return 1;
}
else if (SC >= 1942) {
if (negBalance < 41) {
if (diffRsCsdiff < -7) {
return 4;
}
else if (diffRsCsdiff >= -7) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (negBalance >= 41) {
if (Rs < 1126) {
return 1;
}
else if (Rs >= 1126) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 2) {
if (posBalance < 21) {
return 8;
}
else if (posBalance >= 21) {
if (Rs < 1383) {
return 4;
}
else if (Rs >= 1383) {
return 8;
}
}
}
else if (AFD >= 2) {
if (mcTcor < 96) {
if (Rs < 935) {
return 8;
}
else if (Rs >= 935) {
if (Rs < 1200) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (ssDCval < 133) {
return 4;
}
else if (ssDCval >= 133) {
return 8;
}
}
}
else if (Rs >= 1200) {
if (ssDCval < 131) {
return 8;
}
else if (ssDCval >= 131) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (Cs < 1467) {
return 8;
}
else if (Cs >= 1467) {
if (refDCval < 128) {
if (Rs < 1243) {
return 4;
}
else if (Rs >= 1243) {
return 8;
}
}
else if (refDCval >= 128) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1455) {
if (ssDCval < 86) {
if (tcor < 98) {
if (Rs < 2208) {
if (negBalance < 61) {
if (diffMVdiffVal < -29) {
return 1;
}
else if (diffMVdiffVal >= -29) {
return 8;
}
}
else if (negBalance >= 61) {
return 4;
}
}
else if (Rs >= 2208) {
if (mu_mv_mag_sq < 122) {
if (Rs < 2787) {
if (Cs < 1534) {
if (tcor < 92) {
if (Rs < 2456) {
return 8;
}
else if (Rs >= 2456) {
return 8;
}
}
else if (tcor >= 92) {
return 4;
}
}
else if (Cs >= 1534) {
return 8;
}
}
else if (Rs >= 2787) {
return 2;
}
}
else if (mu_mv_mag_sq >= 122) {
if (negBalance < 43) {
return 8;
}
else if (negBalance >= 43) {
return 4;
}
}
}
}
else if (tcor >= 98) {
if (negBalance < 19) {
if (SC < 4469) {
if (Cs < 1350) {
if (Rs < 2287) {
return 8;
}
else if (Rs >= 2287) {
if (negBalance < 12) {
if (SC < 3630) {
if (Rs < 2295) {
return 8;
}
else if (Rs >= 2295) {
if (Rs < 2297) {
return 2;
}
else if (Rs >= 2297) {
if (Rs < 2299) {
return 8;
}
else if (Rs >= 2299) {
return 2;
}
}
}
}
else if (SC >= 3630) {
return 8;
}
}
else if (negBalance >= 12) {
if (Rs < 2339) {
if (SC < 3644) {
if (Rs < 2313) {
if (Rs < 2288) {
return 2;
}
else if (Rs >= 2288) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (Rs < 2302) {
if (Rs < 2293) {
return 8;
}
else if (Rs >= 2293) {
return 2;
}
}
else if (Rs >= 2302) {
return 8;
}
}
}
}
else if (Rs >= 2313) {
return 2;
}
}
else if (SC >= 3644) {
return 8;
}
}
else if (Rs >= 2339) {
return 2;
}
}
}
}
else if (Cs >= 1350) {
if (Rs < 1490) {
if (Cs < 1462) {
return 8;
}
else if (Cs >= 1462) {
return 1;
}
}
else if (Rs >= 1490) {
if (mu_mv_mag_sq < 123) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 123) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
}
}
}
else if (SC >= 4469) {
return 1;
}
}
else if (negBalance >= 19) {
if (ssDCval < 78) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (Cs < 1557) {
if (Cs < 1341) {
if (Rs < 2283) {
return 8;
}
else if (Rs >= 2283) {
if (Rs < 2727) {
return 4;
}
else if (Rs >= 2727) {
return 2;
}
}
}
else if (Cs >= 1341) {
return 4;
}
}
else if (Cs >= 1557) {
if (negBalance < 27) {
if (Rs < 1952) {
return 1;
}
else if (Rs >= 1952) {
return 4;
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
}
else if (ssDCval >= 78) {
if (Rs < 2446) {
return 8;
}
else if (Rs >= 2446) {
return 4;
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 154) {
if (ssDCval < 117) {
if (diffAFD < -3) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
if (MVDiff < 122) {
return 8;
}
else if (MVDiff >= 122) {
return 4;
}
}
}
else if (diffAFD >= -3) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Cs < 1409) {
return 2;
}
else if (Cs >= 1409) {
return 8;
}
}
}
}
else if (ssDCval >= 117) {
if (CsDiff < 2) {
if (Rs < 2828) {
if (mcTcor < 77) {
return 4;
}
else if (mcTcor >= 77) {
return 8;
}
}
else if (Rs >= 2828) {
if (Cs < 1511) {
if (Rs < 2915) {
return 1;
}
else if (Rs >= 2915) {
return 2;
}
}
else if (Cs >= 1511) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 120) {
if (Cs < 1581) {
return 8;
}
else if (Cs >= 1581) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (ssDCval < 120) {
return 1;
}
else if (ssDCval >= 120) {
if (Rs < 2797) {
if (ssDCval < 136) {
return 8;
}
else if (ssDCval >= 136) {
return 1;
}
}
else if (Rs >= 2797) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 154) {
if (diffAFD < -7) {
return 8;
}
else if (diffAFD >= -7) {
return 2;
}
}
}
}
}
else if (Rs >= 3092) {
if (refDCval < 77) {
if (mu_mv_mag_sq < 126) {
if (SC < 4913) {
if (AFD < 3) {
if (Cs < 1527) {
return 1;
}
else if (Cs >= 1527) {
return 4;
}
}
else if (AFD >= 3) {
return 2;
}
}
else if (SC >= 4913) {
if (ssDCval < 63) {
if (negBalance < 21) {
if (Rs < 5514) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (Rs >= 5514) {
return 8;
}
}
else if (negBalance >= 21) {
return 8;
}
}
else if (ssDCval >= 63) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (ssDCval < 74) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (ssDCval >= 74) {
if (Rs < 3985) {
return 2;
}
else if (Rs >= 3985) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1577) {
if (RsCsDiff < 12) {
if (Rs < 3678) {
if (tcor < 100) {
if (SC < 5010) {
return 8;
}
else if (SC >= 5010) {
return 4;
}
}
else if (tcor >= 100) {
return 4;
}
}
else if (Rs >= 3678) {
return 8;
}
}
else if (RsCsDiff >= 12) {
return 4;
}
}
else if (Cs >= 1577) {
if (Rs < 3265) {
return 8;
}
else if (Rs >= 3265) {
return 4;
}
}
}
}
else if (refDCval >= 77) {
if (ssDCval < 113) {
if (posBalance < 42) {
if (Rs < 3986) {
if (Rs < 3310) {
if (Rs < 3197) {
return 8;
}
else if (Rs >= 3197) {
return 2;
}
}
else if (Rs >= 3310) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 4;
}
}
}
else if (Rs >= 3986) {
if (ssDCval < 96) {
if (diffRsCsdiff < 14) {
if (posBalance < 27) {
return 2;
}
else if (posBalance >= 27) {
return 8;
}
}
else if (diffRsCsdiff >= 14) {
return 2;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (posBalance >= 42) {
if (refDCval < 91) {
if (Rs < 3125) {
if (Rs < 3103) {
return 2;
}
else if (Rs >= 3103) {
return 4;
}
}
else if (Rs >= 3125) {
return 2;
}
}
else if (refDCval >= 91) {
if (Cs < 1534) {
if (Cs < 1375) {
if (Cs < 1317) {
return 4;
}
else if (Cs >= 1317) {
return 2;
}
}
else if (Cs >= 1375) {
return 4;
}
}
else if (Cs >= 1534) {
if (Cs < 1614) {
return 8;
}
else if (Cs >= 1614) {
return 4;
}
}
}
}
}
else if (ssDCval >= 113) {
if (diffRsCsdiff < -40) {
if (Rs < 3547) {
return 2;
}
else if (Rs >= 3547) {
return 4;
}
}
else if (diffRsCsdiff >= -40) {
return 2;
}
}
}
}
}
else if (Cs >= 1633) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (refDCval < 65) {
if (Rs < 8850) {
return 8;
}
else if (Rs >= 8850) {
return 2;
}
}
else if (refDCval >= 65) {
if (diffTSC < -1) {
if (diffRsCsdiff < -1) {
if (ssDCval < 74) {
if (diffAFD < -6) {
return 8;
}
else if (diffAFD >= -6) {
return 4;
}
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
return 4;
}
}
else if (diffTSC >= -1) {
if (Rs < 3814) {
if (Cs < 2593) {
return 8;
}
else if (Cs >= 2593) {
return 4;
}
}
else if (Rs >= 3814) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 3319) {
if (posBalance < 36) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 3246) {
if (negBalance < 16) {
return 1;
}
else if (negBalance >= 16) {
if (SC < 3947) {
if (Rs < 1417) {
return 4;
}
else if (Rs >= 1417) {
return 8;
}
}
else if (SC >= 3947) {
if (posBalance < 31) {
if (Rs < 2153) {
return 1;
}
else if (Rs >= 2153) {
return 8;
}
}
else if (posBalance >= 31) {
return 1;
}
}
}
}
else if (Rs >= 3246) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
return 4;
}
}
}
}
else if (posBalance >= 36) {
if (Rs < 1659) {
if (ssDCval < 57) {
if (Rs < 992) {
if (Rs < 800) {
return 1;
}
else if (Rs >= 800) {
return 8;
}
}
else if (Rs >= 992) {
return 4;
}
}
else if (ssDCval >= 57) {
if (negBalance < 27) {
if (Rs < 1148) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
else if (Rs >= 1148) {
return 8;
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (Rs >= 1659) {
if (negBalance < 19) {
if (Cs < 2070) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (Cs >= 2070) {
return 4;
}
}
else if (negBalance >= 19) {
if (refDCval < 66) {
if (posBalance < 43) {
if (Rs < 2314) {
return 8;
}
else if (Rs >= 2314) {
if (negBalance < 22) {
if (Rs < 2694) {
return 1;
}
else if (Rs >= 2694) {
if (Rs < 2910) {
return 4;
}
else if (Rs >= 2910) {
return 4;
}
}
}
else if (negBalance >= 22) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (Cs < 1815) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (Cs >= 1815) {
return 8;
}
}
}
else if (refDCval >= 66) {
if (Rs < 3481) {
return 1;
}
else if (Rs >= 3481) {
return 8;
}
}
}
}
}
}
else if (Cs >= 3319) {
return 8;
}
}
else if (ssDCval >= 67) {
if (mu_mv_mag_sq < 120) {
if (posBalance < 36) {
if (refDCval < 92) {
if (negBalance < 7) {
return 1;
}
else if (negBalance >= 7) {
if (mu_mv_mag_sq < 112) {
if (Rs < 2157) {
return 8;
}
else if (Rs >= 2157) {
return 1;
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
}
else if (refDCval >= 92) {
if (negBalance < 7) {
if (ssDCval < 92) {
return 1;
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (negBalance >= 7) {
if (Rs < 2317) {
return 2;
}
else if (Rs >= 2317) {
if (posBalance < 14) {
return 2;
}
else if (posBalance >= 14) {
return 4;
}
}
}
}
}
else if (posBalance >= 36) {
return 8;
}
}
else if (mu_mv_mag_sq >= 120) {
if (negBalance < 7) {
return 8;
}
else if (negBalance >= 7) {
if (posBalance < 40) {
if (Cs < 2238) {
if (Rs < 2955) {
if (AFD < 1) {
if (Cs < 1766) {
return 8;
}
else if (Cs >= 1766) {
return 4;
}
}
else if (AFD >= 1) {
return 8;
}
}
else if (Rs >= 2955) {
if (Rs < 4318) {
return 2;
}
else if (Rs >= 4318) {
return 4;
}
}
}
else if (Cs >= 2238) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 8;
}
}
}
else if (posBalance >= 40) {
if (Rs < 1097) {
if (Rs < 664) {
return 8;
}
else if (Rs >= 664) {
return 1;
}
}
else if (Rs >= 1097) {
if (Rs < 4302) {
if (negBalance < 33) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (diffTSC < -1) {
if (ssDCval < 81) {
if (Rs < 2850) {
return 4;
}
else if (Rs >= 2850) {
return 8;
}
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (negBalance >= 33) {
if (Rs < 2064) {
return 4;
}
else if (Rs >= 2064) {
return 8;
}
}
}
else if (Rs >= 4302) {
if (ssDCval < 97) {
if (Cs < 4124) {
return 8;
}
else if (Cs >= 4124) {
if (Rs < 4351) {
return 8;
}
else if (Rs >= 4351) {
return 1;
}
}
}
else if (ssDCval >= 97) {
return 4;
}
}
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6660) {
if (SC < 8331) {
if (ssDCval < 81) {
if (negBalance < 34) {
if (RsCsDiff < 15) {
if (SC < 5111) {
if (Rs < 2914) {
if (mu_mv_mag_sq < 114) {
if (Rs < 992) {
return 8;
}
else if (Rs >= 992) {
return 4;
}
}
else if (mu_mv_mag_sq >= 114) {
if (mcTcor < 99) {
if (Cs < 2237) {
return 8;
}
else if (Cs >= 2237) {
return 4;
}
}
else if (mcTcor >= 99) {
if (ssDCval < 80) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (Cs < 1764) {
return 1;
}
else if (Cs >= 1764) {
return 8;
}
}
}
else if (ssDCval >= 80) {
return 4;
}
}
}
}
else if (Rs >= 2914) {
if (negBalance < 25) {
if (ssDCval < 72) {
if (Cs < 1920) {
if (Rs < 2933) {
return 4;
}
else if (Rs >= 2933) {
if (SC < 4871) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 2988) {
return 8;
}
else if (Rs >= 2988) {
return 8;
}
}
}
else if (SC >= 4871) {
if (SC < 4875) {
return 4;
}
else if (SC >= 4875) {
if (Rs < 3026) {
if (Rs < 3000) {
return 8;
}
else if (Rs >= 3000) {
return 8;
}
}
else if (Rs >= 3026) {
if (Cs < 1840) {
if (Rs < 3100) {
return 8;
}
else if (Rs >= 3100) {
return 8;
}
}
else if (Cs >= 1840) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1920) {
return 4;
}
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (negBalance >= 25) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
return 4;
}
}
}
}
else if (SC >= 5111) {
if (posBalance < 32) {
if (Rs < 4989) {
if (RsCsDiff < 7) {
if (Rs < 4081) {
return 1;
}
else if (Rs >= 4081) {
return 8;
}
}
else if (RsCsDiff >= 7) {
return 8;
}
}
else if (Rs >= 4989) {
return 4;
}
}
else if (posBalance >= 32) {
return 8;
}
}
}
else if (RsCsDiff >= 15) {
if (Rs < 3745) {
if (SC < 5365) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
if (SC < 4264) {
if (gchDC < 4) {
if (Rs < 2526) {
return 8;
}
else if (Rs >= 2526) {
return 4;
}
}
else if (gchDC >= 4) {
return 4;
}
}
else if (SC >= 4264) {
if (RsDiff < 7) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
if (Rs < 2868) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Rs >= 2868) {
return 8;
}
}
}
else if (RsDiff >= 7) {
return 8;
}
}
}
}
else if (SC >= 5365) {
return 8;
}
}
else if (Rs >= 3745) {
if (SC < 5917) {
return 2;
}
else if (SC >= 5917) {
if (tcor < 98) {
if (Cs < 2729) {
if (MVDiff < 1) {
if (posBalance < 45) {
if (mcTcor < 96) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 8;
}
}
else if (mcTcor >= 96) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (posBalance < 31) {
if (Rs < 5096) {
if (negBalance < 25) {
if (Cs < 2220) {
return 4;
}
else if (Cs >= 2220) {
return 8;
}
}
else if (negBalance >= 25) {
return 4;
}
}
else if (Rs >= 5096) {
return 4;
}
}
else if (posBalance >= 31) {
if (diffRsCsdiff < -11) {
return 4;
}
else if (diffRsCsdiff >= -11) {
return 4;
}
}
}
}
}
else if (posBalance >= 45) {
return 4;
}
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 119) {
if (MVDiff < 2) {
if (negBalance < 13) {
return 8;
}
else if (negBalance >= 13) {
if (diffRsCsdiff < -19) {
return 8;
}
else if (diffRsCsdiff >= -19) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 4407) {
if (Rs < 4397) {
if (Rs < 4338) {
return 8;
}
else if (Rs >= 4338) {
return 8;
}
}
else if (Rs >= 4397) {
return 8;
}
}
else if (Rs >= 4407) {
if (negBalance < 22) {
if (gchDC < 1) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (negBalance >= 22) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (MVDiff < 4) {
if (diffTSC < 1) {
if (diffRsCsdiff < 6) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
return 4;
}
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (AFD < 4) {
if (RsCsDiff < 22) {
if (negBalance < 13) {
return 4;
}
else if (negBalance >= 13) {
if (Rs < 4327) {
return 4;
}
else if (Rs >= 4327) {
return 4;
}
}
}
else if (RsCsDiff >= 22) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (Cs >= 2729) {
return 8;
}
}
else if (tcor >= 98) {
return 4;
}
}
}
}
}
else if (negBalance >= 34) {
if (MVDiff < 4) {
if (posBalance < 69) {
if (diffMVdiffVal < -4) {
if (diffRsCsdiff < 9) {
return 4;
}
else if (diffRsCsdiff >= 9) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
if (diffTSC < -3) {
if (Rs < 2952) {
return 4;
}
else if (Rs >= 2952) {
return 8;
}
}
else if (diffTSC >= -3) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 117) {
if (Rs < 2946) {
return 4;
}
else if (Rs >= 2946) {
return 2;
}
}
else if (mu_mv_mag_sq >= 117) {
if (posBalance < 24) {
if (Rs < 3085) {
return 4;
}
else if (Rs >= 3085) {
return 8;
}
}
else if (posBalance >= 24) {
if (Cs < 3359) {
return 8;
}
else if (Cs >= 3359) {
if (Rs < 2740) {
return 4;
}
else if (Rs >= 2740) {
return 8;
}
}
}
}
}
else if (posBalance >= 49) {
if (refDCval < 81) {
if (Rs < 2212) {
if (SC < 4415) {
return 8;
}
else if (SC >= 4415) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
else if (Rs >= 2212) {
return 8;
}
}
else if (refDCval >= 81) {
if (Rs < 3329) {
return 4;
}
else if (Rs >= 3329) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 69) {
return 4;
}
}
else if (MVDiff >= 4) {
if (mcTcor < 94) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
return 4;
}
}
else if (mcTcor >= 94) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
}
else if (ssDCval >= 81) {
if (Cs < 2622) {
if (Rs < 3795) {
if (Cs < 1781) {
if (tcor < 97) {
if (Rs < 1106) {
return 1;
}
else if (Rs >= 1106) {
return 8;
}
}
else if (tcor >= 97) {
if (diffMVdiffVal < -1) {
if (Rs < 3172) {
return 8;
}
else if (Rs >= 3172) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 1995) {
if (Rs < 1419) {
return 1;
}
else if (Rs >= 1419) {
return 8;
}
}
else if (Rs >= 1995) {
return 4;
}
}
}
}
else if (Cs >= 1781) {
if (negBalance < 49) {
if (Rs < 3085) {
if (ssDCval < 84) {
if (Cs < 2437) {
if (mcTcor < 98) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
if (Rs < 2658) {
return 8;
}
else if (Rs >= 2658) {
return 4;
}
}
}
else if (mcTcor >= 98) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (Cs >= 2437) {
return 4;
}
}
else if (ssDCval >= 84) {
if (diffAFD < -5) {
if (Rs < 2301) {
return 2;
}
else if (Rs >= 2301) {
return 8;
}
}
else if (diffAFD >= -5) {
return 8;
}
}
}
else if (Rs >= 3085) {
if (Cs < 2202) {
if (negBalance < 22) {
if (Cs < 1803) {
return 4;
}
else if (Cs >= 1803) {
return 8;
}
}
else if (negBalance >= 22) {
if (mcTcor < 97) {
if (Cs < 1963) {
return 2;
}
else if (Cs >= 1963) {
return 8;
}
}
else if (mcTcor >= 97) {
return 2;
}
}
}
else if (Cs >= 2202) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
else if (negBalance >= 49) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (mu_mv_mag_sq < 115) {
return 8;
}
else if (mu_mv_mag_sq >= 115) {
return 4;
}
}
}
}
}
else if (Rs >= 3795) {
if (negBalance < 29) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (posBalance < 41) {
if (Rs < 3864) {
return 2;
}
else if (Rs >= 3864) {
if (diffAFD < -2) {
if (Rs < 5046) {
return 2;
}
else if (Rs >= 5046) {
return 4;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
}
else if (posBalance >= 41) {
if (ssDCval < 96) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (Rs < 3988) {
return 2;
}
else if (Rs >= 3988) {
return 8;
}
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
}
else if (negBalance >= 29) {
if (AFD < 8) {
if (RsDiff < 7) {
if (ssDCval < 91) {
if (Rs < 4651) {
return 2;
}
else if (Rs >= 4651) {
return 8;
}
}
else if (ssDCval >= 91) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (AFD >= 8) {
return 1;
}
}
}
}
else if (Cs >= 2622) {
if (Rs < 2679) {
if (Rs < 1391) {
return 2;
}
else if (Rs >= 1391) {
return 4;
}
}
else if (Rs >= 2679) {
if (Rs < 3342) {
if (tcor < 90) {
return 4;
}
else if (tcor >= 90) {
if (refDCval < 97) {
if (negBalance < 37) {
if (Rs < 3311) {
return 4;
}
else if (Rs >= 3311) {
return 8;
}
}
else if (negBalance >= 37) {
return 8;
}
}
else if (refDCval >= 97) {
if (mu_mv_mag_sq < 118) {
if (Cs < 3463) {
return 4;
}
else if (Cs >= 3463) {
return 8;
}
}
else if (mu_mv_mag_sq >= 118) {
if (Cs < 3856) {
if (refDCval < 98) {
if (Cs < 3163) {
return 4;
}
else if (Cs >= 3163) {
return 8;
}
}
else if (refDCval >= 98) {
if (SC < 5931) {
if (SC < 5891) {
return 8;
}
else if (SC >= 5891) {
return 4;
}
}
else if (SC >= 5931) {
return 8;
}
}
}
else if (Cs >= 3856) {
return 4;
}
}
}
}
}
else if (Rs >= 3342) {
if (ssDCval < 98) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
if (Cs < 3614) {
if (RsDiff < 4) {
if (Cs < 3207) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (Rs < 4150) {
return 8;
}
else if (Rs >= 4150) {
return 4;
}
}
}
else if (Cs >= 3207) {
if (Cs < 3248) {
if (Rs < 4116) {
return 8;
}
else if (Rs >= 4116) {
return 4;
}
}
else if (Cs >= 3248) {
return 4;
}
}
}
else if (RsDiff >= 4) {
if (Rs < 3396) {
if (Rs < 3363) {
return 4;
}
else if (Rs >= 3363) {
return 8;
}
}
else if (Rs >= 3396) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 8;
}
}
}
}
else if (Cs >= 3614) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (Cs < 4212) {
return 8;
}
else if (Cs >= 4212) {
if (Rs < 3590) {
return 4;
}
else if (Rs >= 3590) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
}
}
else if (SC >= 8331) {
if (negBalance < 31) {
if (MVDiff < 3) {
if (CsDiff < 11) {
if (Cs < 4278) {
return 8;
}
else if (Cs >= 4278) {
if (RsDiff < 6) {
if (Cs < 4327) {
return 1;
}
else if (Cs >= 4327) {
return 8;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (CsDiff >= 11) {
return 1;
}
}
else if (MVDiff >= 3) {
if (Rs < 5665) {
return 8;
}
else if (Rs >= 5665) {
if (Cs < 3900) {
return 8;
}
else if (Cs >= 3900) {
return 2;
}
}
}
}
else if (negBalance >= 31) {
if (Cs < 3674) {
if (Rs < 5208) {
return 4;
}
else if (Rs >= 5208) {
return 8;
}
}
else if (Cs >= 3674) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 5939) {
return 8;
}
else if (Rs >= 5939) {
return 2;
}
}
}
}
}
}
else if (Rs >= 6660) {
if (ssDCval < 96) {
if (RsDiff < 44) {
return 4;
}
else if (RsDiff >= 44) {
return 8;
}
}
else if (ssDCval >= 96) {
return 8;
}
}
}
}
}
else if (ssDCval >= 102) {
if (refDCval < 129) {
if (Rs < 1893) {
if (refDCval < 115) {
if (negBalance < 47) {
if (posBalance < 54) {
if (ssDCval < 114) {
if (Rs < 1155) {
return 2;
}
else if (Rs >= 1155) {
return 8;
}
}
else if (ssDCval >= 114) {
return 2;
}
}
else if (posBalance >= 54) {
if (Cs < 1916) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Cs >= 1916) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (Rs < 1494) {
if (Rs < 1247) {
return 4;
}
else if (Rs >= 1247) {
return 2;
}
}
else if (Rs >= 1494) {
return 8;
}
}
}
else if (refDCval >= 115) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (posBalance < 56) {
if (Rs < 1859) {
return 1;
}
else if (Rs >= 1859) {
return 8;
}
}
else if (posBalance >= 56) {
if (SC < 3074) {
if (Rs < 858) {
return 8;
}
else if (Rs >= 858) {
return 4;
}
}
else if (SC >= 3074) {
return 8;
}
}
}
}
}
else if (Rs >= 1893) {
if (refDCval < 105) {
if (negBalance < 43) {
if (Cs < 2353) {
if (Rs < 3868) {
if (Rs < 2118) {
return 2;
}
else if (Rs >= 2118) {
return 8;
}
}
else if (Rs >= 3868) {
if (Rs < 4499) {
return 4;
}
else if (Rs >= 4499) {
return 2;
}
}
}
else if (Cs >= 2353) {
return 8;
}
}
else if (negBalance >= 43) {
if (Cs < 3105) {
if (Rs < 4923) {
if (Rs < 3509) {
return 8;
}
else if (Rs >= 3509) {
return 1;
}
}
else if (Rs >= 4923) {
return 2;
}
}
else if (Cs >= 3105) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (diffRsCsdiff < -999) {
return 1;
}
else if (diffRsCsdiff >= -999) {
if (SC < 5696) {
if (RsCsDiff < 77) {
if (refDCval < 124) {
if (SC < 5668) {
if (Rs < 3265) {
if (diffMVdiffVal < -2) {
if (Rs < 2620) {
return 8;
}
else if (Rs >= 2620) {
if (Rs < 2635) {
return 4;
}
else if (Rs >= 2635) {
if (Cs < 2780) {
return 8;
}
else if (Cs >= 2780) {
if (Rs < 2700) {
return 8;
}
else if (Rs >= 2700) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= -2) {
if (RsDiff < 7) {
if (Rs < 1908) {
if (Rs < 1907) {
return 8;
}
else if (Rs >= 1907) {
return 2;
}
}
else if (Rs >= 1908) {
return 8;
}
}
else if (RsDiff >= 7) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
else if (Rs >= 3265) {
if (Cs < 1967) {
if (Cs < 1744) {
return 8;
}
else if (Cs >= 1744) {
return 4;
}
}
else if (Cs >= 1967) {
return 8;
}
}
}
else if (SC >= 5668) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 8;
}
}
}
else if (refDCval >= 124) {
if (Rs < 2086) {
return 8;
}
else if (Rs >= 2086) {
if (Rs < 2170) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
else if (Rs >= 2170) {
if (SC < 5007) {
return 8;
}
else if (SC >= 5007) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
return 1;
}
}
}
}
else if (SC >= 5696) {
if (Cs < 1885) {
return 2;
}
else if (Cs >= 1885) {
if (mu_mv_mag_sq < 122) {
if (RsDiff < 1) {
if (Cs < 2966) {
return 8;
}
else if (Cs >= 2966) {
return 2;
}
}
else if (RsDiff >= 1) {
if (diffRsCsdiff < -14) {
if (diffRsCsdiff < -20) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
if (RsDiff < 8) {
if (Rs < 3339) {
return 8;
}
else if (Rs >= 3339) {
return 4;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
}
else if (diffRsCsdiff >= -20) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
}
else if (diffRsCsdiff >= -14) {
if (Rs < 6431) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
else if (Rs >= 6431) {
if (Rs < 6723) {
return 4;
}
else if (Rs >= 6723) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (ssDCval < 105) {
if (Cs < 2350) {
return 2;
}
else if (Cs >= 2350) {
return 8;
}
}
else if (ssDCval >= 105) {
if (diffAFD < -2) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (Rs < 5276) {
return 2;
}
else if (Rs >= 5276) {
return 8;
}
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 129) {
if (mcTcor < 96) {
if (AFD < 6) {
if (ssDCval < 199) {
if (ssDCval < 130) {
if (Rs < 1667) {
return 8;
}
else if (Rs >= 1667) {
return 1;
}
}
else if (ssDCval >= 130) {
return 8;
}
}
else if (ssDCval >= 199) {
return 4;
}
}
else if (AFD >= 6) {
if (RsCsDiff < 173) {
if (Rs < 2120) {
return 8;
}
else if (Rs >= 2120) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
return 1;
}
}
}
else if (RsCsDiff >= 173) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 7738) {
if (Rs < 4715) {
return 2;
}
else if (Rs >= 4715) {
return 1;
}
}
else if (Rs >= 7738) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 4780) {
if (Cs < 3939) {
if (Rs < 4577) {
if (SC < 3056) {
if (Rs < 1168) {
return 8;
}
else if (Rs >= 1168) {
if (ssDCval < 132) {
return 4;
}
else if (ssDCval >= 132) {
return 8;
}
}
}
else if (SC >= 3056) {
if (diffAFD < -20) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (diffAFD >= -20) {
return 8;
}
}
}
else if (Rs >= 4577) {
return 1;
}
}
else if (Cs >= 3939) {
if (Rs < 3973) {
if (Cs < 4280) {
return 1;
}
else if (Cs >= 4280) {
return 4;
}
}
else if (Rs >= 3973) {
return 8;
}
}
}
else if (Rs >= 4780) {
if (Cs < 2853) {
return 2;
}
else if (Cs >= 2853) {
return 8;
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect5(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1333) {
if (Rs < 2388) {
if (Cs < 527) {
if (negBalance < 14) {
if (SC < 807) {
if (Cs < 205) {
if (refDCval < 48) {
if (AFD < 1) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
if (ssDCval < 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 10) {
if (Rs < 392) {
if (negBalance < 5) {
return 8;
}
else if (negBalance >= 5) {
if (Cs < 31) {
return 8;
}
else if (Cs >= 31) {
if (Rs < 55) {
return 1;
}
else if (Rs >= 55) {
return 8;
}
}
}
}
else if (Rs >= 392) {
if (Rs < 432) {
return 4;
}
else if (Rs >= 432) {
return 8;
}
}
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (ssDCval < 27) {
if (posBalance < 60) {
if (negBalance < 13) {
if (MVDiff < 16) {
if (posBalance < 26) {
if (Rs < 40) {
return 4;
}
else if (Rs >= 40) {
if (Rs < 135) {
return 8;
}
else if (Rs >= 135) {
return 4;
}
}
}
else if (posBalance >= 26) {
return 1;
}
}
else if (MVDiff >= 16) {
return 4;
}
}
else if (negBalance >= 13) {
if (Rs < 23) {
return 1;
}
else if (Rs >= 23) {
return 8;
}
}
}
else if (posBalance >= 60) {
return 8;
}
}
else if (ssDCval >= 27) {
if (SC < 378) {
if (negBalance < 7) {
if (Rs < 8) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (Rs >= 8) {
return 8;
}
}
else if (negBalance >= 7) {
if (MVDiff < 16) {
if (Cs < 41) {
return 4;
}
else if (Cs >= 41) {
if (Cs < 178) {
if (diffTSC < -1) {
if (Rs < 113) {
if (ssDCval < 31) {
return 4;
}
else if (ssDCval >= 31) {
return 8;
}
}
else if (Rs >= 113) {
return 1;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (Cs >= 178) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
else if (MVDiff >= 16) {
return 4;
}
}
}
else if (SC >= 378) {
if (Rs < 298) {
return 4;
}
else if (Rs >= 298) {
return 8;
}
}
}
}
}
else if (AFD >= 1) {
if (posBalance < 59) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
if (refDCval < 29) {
if (Cs < 162) {
if (mcTcor < 88) {
if (Rs < 134) {
return 8;
}
else if (Rs >= 134) {
return 1;
}
}
else if (mcTcor >= 88) {
if (posBalance < 19) {
return 4;
}
else if (posBalance >= 19) {
return 1;
}
}
}
else if (Cs >= 162) {
if (MVDiff < 21) {
return 1;
}
else if (MVDiff >= 21) {
return 4;
}
}
}
else if (refDCval >= 29) {
if (tcor < 94) {
if (posBalance < 36) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (posBalance >= 36) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 1;
}
}
}
else if (tcor >= 94) {
if (SC < 219) {
if (posBalance < 36) {
if (Cs < 79) {
return 4;
}
else if (Cs >= 79) {
if (mu_mv_mag_sq < 115) {
return 8;
}
else if (mu_mv_mag_sq >= 115) {
return 1;
}
}
}
else if (posBalance >= 36) {
return 8;
}
}
else if (SC >= 219) {
if (SC < 319) {
if (TSC < 1) {
return 8;
}
else if (TSC >= 1) {
return 1;
}
}
else if (SC >= 319) {
if (posBalance < 27) {
return 8;
}
else if (posBalance >= 27) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 59) {
if (mu_mv_mag_sq < 17) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
if (SC < 396) {
if (SC < 331) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 141) {
return 8;
}
else if (Rs >= 141) {
return 1;
}
}
}
else if (SC >= 331) {
return 8;
}
}
else if (SC >= 396) {
if (Rs < 660) {
return 2;
}
else if (Rs >= 660) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (Rs < 235) {
if (Rs < 13) {
if (mcTcor < 69) {
if (Rs < 3) {
if (posBalance < 73) {
return 4;
}
else if (posBalance >= 73) {
return 8;
}
}
else if (Rs >= 3) {
return 1;
}
}
else if (mcTcor >= 69) {
if (posBalance < 80) {
if (Rs < 3) {
if (TSC < 1) {
return 2;
}
else if (TSC >= 1) {
return 4;
}
}
else if (Rs >= 3) {
return 8;
}
}
else if (posBalance >= 80) {
return 4;
}
}
}
else if (Rs >= 13) {
if (mu_mv_mag_sq < 49) {
if (diffMVdiffVal < -66) {
return 8;
}
else if (diffMVdiffVal >= -66) {
if (negBalance < 2) {
return 4;
}
else if (negBalance >= 2) {
if (Cs < 29) {
return 8;
}
else if (Cs >= 29) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (Rs < 22) {
if (Rs < 16) {
return 8;
}
else if (Rs >= 16) {
return 1;
}
}
else if (Rs >= 22) {
return 8;
}
}
}
}
else if (Rs >= 235) {
if (tcor < 62) {
return 1;
}
else if (tcor >= 62) {
if (tcor < 88) {
if (ssDCval < 40) {
return 4;
}
else if (ssDCval >= 40) {
return 8;
}
}
else if (tcor >= 88) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 48) {
if (refDCval < 82) {
if (Rs < 31) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
else if (Rs >= 31) {
if (refDCval < 67) {
if (AFD < 2) {
if (MVDiff < 1) {
if (negBalance < 6) {
if (Rs < 179) {
return 8;
}
else if (Rs >= 179) {
return 4;
}
}
else if (negBalance >= 6) {
return 8;
}
}
else if (MVDiff >= 1) {
if (tcor < 99) {
if (Cs < 11) {
if (Rs < 441) {
return 2;
}
else if (Rs >= 441) {
return 4;
}
}
else if (Cs >= 11) {
return 8;
}
}
else if (tcor >= 99) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 8;
}
}
}
}
}
else if (AFD >= 2) {
if (mu_mv_mag_sq < 47) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (mu_mv_mag_sq >= 47) {
if (refDCval < 60) {
if (Rs < 49) {
return 2;
}
else if (Rs >= 49) {
return 8;
}
}
else if (refDCval >= 60) {
return 1;
}
}
}
}
else if (refDCval >= 67) {
if (Cs < 107) {
if (Rs < 153) {
return 4;
}
else if (Rs >= 153) {
return 8;
}
}
else if (Cs >= 107) {
if (Cs < 130) {
if (mu_mv_mag_sq < 125) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
return 4;
}
}
else if (Cs >= 130) {
return 2;
}
}
}
}
}
else if (refDCval >= 82) {
if (ssDCval < 87) {
return 1;
}
else if (ssDCval >= 87) {
if (SC < 320) {
if (SC < 136) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
if (gchDC < 58) {
return 1;
}
else if (gchDC >= 58) {
if (Rs < 37) {
return 2;
}
else if (Rs >= 37) {
return 8;
}
}
}
}
else if (SC >= 136) {
if (gchDC < 6) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 4;
}
}
else if (gchDC >= 6) {
return 8;
}
}
}
else if (SC >= 320) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
return 2;
}
}
}
}
}
}
else if (Cs >= 205) {
if (mu_mv_mag_sq < 112) {
if (AFD < 2) {
if (mu_mv_mag_sq < 70) {
return 4;
}
else if (mu_mv_mag_sq >= 70) {
if (ssDCval < 82) {
if (Rs < 159) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
return 2;
}
}
else if (Rs >= 159) {
if (ssDCval < 70) {
if (tcor < 95) {
if (Cs < 248) {
return 8;
}
else if (Cs >= 248) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 1;
}
}
}
else if (tcor >= 95) {
if (posBalance < 21) {
return 8;
}
else if (posBalance >= 21) {
return 4;
}
}
}
else if (ssDCval >= 70) {
if (Rs < 293) {
return 1;
}
else if (Rs >= 293) {
return 8;
}
}
}
}
else if (ssDCval >= 82) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
else if (AFD >= 2) {
if (posBalance < 40) {
if (RsDiff < 1) {
if (Cs < 221) {
return 1;
}
else if (Cs >= 221) {
return 4;
}
}
else if (RsDiff >= 1) {
if (ssDCval < 158) {
return 1;
}
else if (ssDCval >= 158) {
return 2;
}
}
}
else if (posBalance >= 40) {
if (negBalance < 3) {
if (refDCval < 51) {
return 4;
}
else if (refDCval >= 51) {
return 8;
}
}
else if (negBalance >= 3) {
if (SC < 498) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 1;
}
}
else if (SC >= 498) {
if (mu_mv_mag_sq < 24) {
if (Cs < 270) {
return 1;
}
else if (Cs >= 270) {
return 2;
}
}
else if (mu_mv_mag_sq >= 24) {
if (TSC < 2) {
if (Cs < 232) {
return 1;
}
else if (Cs >= 232) {
return 4;
}
}
else if (TSC >= 2) {
if (Rs < 463) {
if (Rs < 309) {
return 8;
}
else if (Rs >= 309) {
return 1;
}
}
else if (Rs >= 463) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (ssDCval < 66) {
if (Rs < 378) {
if (Rs < 206) {
if (Cs < 255) {
if (Rs < 166) {
return 4;
}
else if (Rs >= 166) {
return 8;
}
}
else if (Cs >= 255) {
return 1;
}
}
else if (Rs >= 206) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 40) {
return 4;
}
else if (ssDCval >= 40) {
if (Rs < 241) {
return 4;
}
else if (Rs >= 241) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 3) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (negBalance >= 3) {
if (Cs < 254) {
return 8;
}
else if (Cs >= 254) {
return 4;
}
}
}
}
}
else if (Rs >= 378) {
return 8;
}
}
else if (ssDCval >= 66) {
if (Cs < 239) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (Cs >= 239) {
if (SC < 646) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 1;
}
}
else if (SC >= 646) {
if (MVDiff < 40) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
else if (MVDiff >= 40) {
if (Rs < 446) {
return 4;
}
else if (Rs >= 446) {
return 1;
}
}
}
}
}
}
}
}
else if (SC >= 807) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 128) {
if (Cs < 268) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 2;
}
}
}
}
else if (Cs >= 268) {
if (SC < 940) {
if (ssDCval < 122) {
if (Cs < 386) {
if (Cs < 333) {
return 8;
}
else if (Cs >= 333) {
if (Rs < 475) {
return 8;
}
else if (Rs >= 475) {
return 1;
}
}
}
else if (Cs >= 386) {
if (Rs < 349) {
return 1;
}
else if (Rs >= 349) {
return 8;
}
}
}
else if (ssDCval >= 122) {
if (gchDC < 1) {
if (posBalance < 13) {
return 8;
}
else if (posBalance >= 13) {
return 1;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (SC >= 940) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Rs < 1314) {
if (CsDiff < 1) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (refDCval < 54) {
return 1;
}
else if (refDCval >= 54) {
if (mu_mv_mag_sq < 116) {
if (Rs < 715) {
return 8;
}
else if (Rs >= 715) {
return 1;
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
}
}
else if (CsDiff >= 1) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (Rs >= 1314) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (Cs < 445) {
return 1;
}
else if (Cs >= 445) {
if (Rs < 1381) {
return 8;
}
else if (Rs >= 1381) {
if (Rs < 1462) {
return 2;
}
else if (Rs >= 1462) {
return 1;
}
}
}
}
}
else if (ssDCval >= 67) {
if (Rs < 487) {
return 1;
}
else if (Rs >= 487) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (SC < 1063) {
return 4;
}
else if (SC >= 1063) {
if (RsDiff < 1) {
if (Cs < 298) {
return 1;
}
else if (Cs >= 298) {
return 8;
}
}
else if (RsDiff >= 1) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (refDCval < 60) {
return 8;
}
else if (refDCval >= 60) {
if (ssDCval < 84) {
if (Rs < 552) {
return 8;
}
else if (Rs >= 552) {
if (SC < 1204) {
if (Rs < 673) {
return 4;
}
else if (Rs >= 673) {
return 1;
}
}
else if (SC >= 1204) {
if (Rs < 1110) {
return 8;
}
else if (Rs >= 1110) {
return 2;
}
}
}
}
else if (ssDCval >= 84) {
if (negBalance < 13) {
return 8;
}
else if (negBalance >= 13) {
if (Rs < 606) {
return 8;
}
else if (Rs >= 606) {
if (Rs < 1116) {
return 4;
}
else if (Rs >= 1116) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
else if (MVDiff >= 8) {
if (ssDCval < 72) {
if (tcor < 91) {
if (Rs < 1075) {
if (Rs < 473) {
return 8;
}
else if (Rs >= 473) {
if (mu_mv_mag_sq < 71) {
if (Rs < 880) {
return 8;
}
else if (Rs >= 880) {
return 1;
}
}
else if (mu_mv_mag_sq >= 71) {
return 1;
}
}
}
else if (Rs >= 1075) {
if (Cs < 278) {
return 2;
}
else if (Cs >= 278) {
if (Rs < 1309) {
return 8;
}
else if (Rs >= 1309) {
return 4;
}
}
}
}
else if (tcor >= 91) {
if (mu_mv_mag_sq < 105) {
if (SC < 963) {
return 4;
}
else if (SC >= 963) {
if (Rs < 791) {
return 1;
}
else if (Rs >= 791) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Cs < 227) {
return 4;
}
else if (Cs >= 227) {
return 8;
}
}
}
}
else if (ssDCval >= 72) {
if (diffTSC < -3) {
if (ssDCval < 161) {
if (Rs < 926) {
if (MVDiff < 29) {
return 2;
}
else if (MVDiff >= 29) {
if (Rs < 480) {
return 2;
}
else if (Rs >= 480) {
if (MVDiff < 32) {
if (Rs < 578) {
return 4;
}
else if (Rs >= 578) {
return 2;
}
}
else if (MVDiff >= 32) {
return 4;
}
}
}
}
else if (Rs >= 926) {
if (diffRsCsdiff < -17) {
if (SC < 1831) {
return 2;
}
else if (SC >= 1831) {
if (Rs < 1675) {
return 4;
}
else if (Rs >= 1675) {
return 1;
}
}
}
else if (diffRsCsdiff >= -17) {
return 1;
}
}
}
else if (ssDCval >= 161) {
return 8;
}
}
else if (diffTSC >= -3) {
if (Cs < 399) {
if (posBalance < 85) {
if (tcor < 89) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (tcor >= 89) {
return 1;
}
}
else if (posBalance >= 85) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (mu_mv_mag_sq < 51) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 1679) {
return 8;
}
else if (Rs >= 1679) {
return 2;
}
}
}
}
}
else if (Cs >= 399) {
if (Rs < 398) {
return 2;
}
else if (Rs >= 398) {
if (mu_mv_mag_sq < 6) {
return 1;
}
else if (mu_mv_mag_sq >= 6) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (mcTcor < 85) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (tcor < 60) {
if (negBalance < 44) {
if (Cs < 408) {
if (Cs < 108) {
if (negBalance < 19) {
if (Rs < 72) {
if (Rs < 38) {
return 8;
}
else if (Rs >= 38) {
return 2;
}
}
else if (Rs >= 72) {
return 1;
}
}
else if (negBalance >= 19) {
return 1;
}
}
else if (Cs >= 108) {
if (AFD < 7) {
if (ssDCval < 25) {
if (Cs < 166) {
return 1;
}
else if (Cs >= 166) {
return 8;
}
}
else if (ssDCval >= 25) {
if (Cs < 122) {
return 1;
}
else if (Cs >= 122) {
return 2;
}
}
}
else if (AFD >= 7) {
if (Cs < 131) {
return 8;
}
else if (Cs >= 131) {
if (Rs < 96) {
return 4;
}
else if (Rs >= 96) {
return 1;
}
}
}
}
}
else if (Cs >= 408) {
if (Rs < 556) {
return 2;
}
else if (Rs >= 556) {
return 8;
}
}
}
else if (negBalance >= 44) {
if (RsDiff < 5) {
if (MVDiff < 288) {
return 8;
}
else if (MVDiff >= 288) {
return 4;
}
}
else if (RsDiff >= 5) {
if (Cs < 33) {
return 8;
}
else if (Cs >= 33) {
return 1;
}
}
}
}
else if (tcor >= 60) {
if (refDCval < 23) {
if (Rs < 237) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 237) {
return 4;
}
}
else if (refDCval >= 23) {
if (Rs < 5) {
if (ssDCval < 21) {
return 1;
}
else if (ssDCval >= 21) {
return 4;
}
}
else if (Rs >= 5) {
if (mcTcor < 72) {
if (MVDiff < 39) {
if (Rs < 840) {
return 4;
}
else if (Rs >= 840) {
return 1;
}
}
else if (MVDiff >= 39) {
return 2;
}
}
else if (mcTcor >= 72) {
if (Cs < 356) {
if (diffMVdiffVal < -5) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
if (ssDCval < 34) {
return 1;
}
else if (ssDCval >= 34) {
return 8;
}
}
}
else if (diffMVdiffVal >= -5) {
if (MVDiff < 9) {
if (Rs < 464) {
return 1;
}
else if (Rs >= 464) {
return 4;
}
}
else if (MVDiff >= 9) {
return 1;
}
}
}
else if (Cs >= 356) {
if (Rs < 655) {
if (Rs < 468) {
return 2;
}
else if (Rs >= 468) {
return 8;
}
}
else if (Rs >= 655) {
return 1;
}
}
}
}
}
}
}
}
else if (mcTcor >= 85) {
if (posBalance < 27) {
if (posBalance < 11) {
if (Rs < 150) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (mu_mv_mag_sq < 38) {
if (negBalance < 55) {
if (MVDiff < 229) {
return 4;
}
else if (MVDiff >= 229) {
return 8;
}
}
else if (negBalance >= 55) {
return 8;
}
}
else if (mu_mv_mag_sq >= 38) {
return 4;
}
}
}
else if (Rs >= 150) {
if (Rs < 217) {
return 1;
}
else if (Rs >= 217) {
return 4;
}
}
}
else if (posBalance >= 11) {
if (MVDiff < 11) {
if (negBalance < 20) {
if (SC < 202) {
return 4;
}
else if (SC >= 202) {
if (Cs < 113) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (Cs >= 113) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
if (Rs < 348) {
return 8;
}
else if (Rs >= 348) {
return 2;
}
}
}
}
}
else if (negBalance >= 20) {
if (posBalance < 21) {
return 4;
}
else if (posBalance >= 21) {
if (negBalance < 23) {
if (Rs < 66) {
return 1;
}
else if (Rs >= 66) {
return 8;
}
}
else if (negBalance >= 23) {
return 1;
}
}
}
}
else if (MVDiff >= 11) {
if (ssDCval < 30) {
if (Rs < 8) {
return 8;
}
else if (Rs >= 8) {
return 1;
}
}
else if (ssDCval >= 30) {
return 8;
}
}
}
}
else if (posBalance >= 27) {
if (refDCval < 30) {
if (Cs < 300) {
if (gchDC < 1) {
if (diffMVdiffVal < -11) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (diffMVdiffVal >= -11) {
if (mu_mv_mag_sq < 40) {
if (Cs < 255) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
return 1;
}
}
}
else if (Cs >= 255) {
return 4;
}
}
else if (mu_mv_mag_sq >= 40) {
return 1;
}
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 40) {
if (negBalance < 26) {
if (MVDiff < 49) {
if (Cs < 22) {
if (Rs < 52) {
return 2;
}
else if (Rs >= 52) {
return 4;
}
}
else if (Cs >= 22) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (MVDiff >= 49) {
return 8;
}
}
else if (negBalance >= 26) {
if (posBalance < 45) {
if (Rs < 68) {
return 1;
}
else if (Rs >= 68) {
if (Rs < 289) {
return 4;
}
else if (Rs >= 289) {
return 1;
}
}
}
else if (posBalance >= 45) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (diffRsCsdiff < -3) {
if (Rs < 269) {
return 8;
}
else if (Rs >= 269) {
return 1;
}
}
else if (diffRsCsdiff >= -3) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (diffMVdiffVal < -2) {
if (Rs < 311) {
return 8;
}
else if (Rs >= 311) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
}
}
}
}
else if (Cs >= 300) {
if (MVDiff < 20) {
if (ssDCval < 23) {
return 4;
}
else if (ssDCval >= 23) {
return 1;
}
}
else if (MVDiff >= 20) {
if (mu_mv_mag_sq < 40) {
if (Rs < 631) {
return 1;
}
else if (Rs >= 631) {
return 8;
}
}
else if (mu_mv_mag_sq >= 40) {
return 4;
}
}
}
}
else if (refDCval >= 30) {
if (posBalance < 41) {
if (mu_mv_mag_sq < 17) {
if (MVDiff < 8) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (MVDiff < 5) {
if (Rs < 73) {
return 1;
}
else if (Rs >= 73) {
return 4;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (Rs < 145) {
return 8;
}
else if (Rs >= 145) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (Cs < 261) {
if (Rs < 331) {
if (mcTcor < 97) {
if (AFD < 4) {
if (Cs < 101) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (Cs >= 101) {
if (diffMVdiffVal < -7) {
return 1;
}
else if (diffMVdiffVal >= -7) {
if (ssDCval < 33) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
return 8;
}
}
else if (ssDCval >= 33) {
return 4;
}
}
}
}
else if (AFD >= 4) {
if (Cs < 67) {
if (Rs < 223) {
return 4;
}
else if (Rs >= 223) {
return 2;
}
}
else if (Cs >= 67) {
if (mu_mv_mag_sq < 56) {
return 1;
}
else if (mu_mv_mag_sq >= 56) {
if (Rs < 188) {
return 1;
}
else if (Rs >= 188) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 31) {
if (Rs < 117) {
return 8;
}
else if (Rs >= 117) {
return 1;
}
}
else if (posBalance >= 31) {
if (negBalance < 17) {
if (Rs < 104) {
return 8;
}
else if (Rs >= 104) {
return 1;
}
}
else if (negBalance >= 17) {
return 1;
}
}
}
}
else if (Rs >= 331) {
if (TSC < 3) {
if (MVDiff < 4) {
if (Rs < 365) {
return 8;
}
else if (Rs >= 365) {
return 4;
}
}
else if (MVDiff >= 4) {
if (Cs < 142) {
return 2;
}
else if (Cs >= 142) {
return 4;
}
}
}
else if (TSC >= 3) {
if (Rs < 336) {
return 4;
}
else if (Rs >= 336) {
return 1;
}
}
}
}
else if (Cs >= 261) {
if (MVDiff < 22) {
return 1;
}
else if (MVDiff >= 22) {
return 4;
}
}
}
}
else if (posBalance >= 41) {
if (posBalance < 67) {
if (tcor < 74) {
if (posBalance < 60) {
if (mu_mv_mag_sq < 5) {
if (mcTcor < 90) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (tcor < 65) {
return 8;
}
else if (tcor >= 65) {
return 1;
}
}
}
else if (mcTcor >= 90) {
return 1;
}
}
else if (mu_mv_mag_sq >= 5) {
if (ssDCval < 35) {
return 1;
}
else if (ssDCval >= 35) {
if (Cs < 174) {
return 1;
}
else if (Cs >= 174) {
return 4;
}
}
}
}
else if (posBalance >= 60) {
if (Rs < 592) {
return 8;
}
else if (Rs >= 592) {
return 1;
}
}
}
else if (tcor >= 74) {
if (negBalance < 26) {
if (MVDiff < 30) {
if (mcTcor < 88) {
if (Rs < 199) {
return 1;
}
else if (Rs >= 199) {
return 8;
}
}
else if (mcTcor >= 88) {
return 1;
}
}
else if (MVDiff >= 30) {
return 8;
}
}
else if (negBalance >= 26) {
return 1;
}
}
}
else if (posBalance >= 67) {
if (Cs < 15) {
return 2;
}
else if (Cs >= 15) {
if (Rs < 229) {
if (Rs < 51) {
return 4;
}
else if (Rs >= 51) {
return 1;
}
}
else if (Rs >= 229) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 241) {
if (ssDCval < 51) {
if (mu_mv_mag_sq < 53) {
if (mcTcor < 88) {
if (posBalance < 47) {
if (posBalance < 27) {
if (TSC < 8) {
if (Rs < 32) {
return 8;
}
else if (Rs >= 32) {
return 2;
}
}
else if (TSC >= 8) {
return 1;
}
}
else if (posBalance >= 27) {
if (RsDiff < 3) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (mcTcor < 87) {
return 1;
}
else if (mcTcor >= 87) {
if (Rs < 70) {
return 4;
}
else if (Rs >= 70) {
return 1;
}
}
}
}
else if (RsDiff >= 3) {
return 8;
}
}
}
else if (posBalance >= 47) {
if (MVDiff < 43) {
if (ssDCval < 44) {
if (mcTcor < 86) {
if (refDCval < 36) {
if (Rs < 115) {
return 4;
}
else if (Rs >= 115) {
return 8;
}
}
else if (refDCval >= 36) {
return 8;
}
}
else if (mcTcor >= 86) {
if (Rs < 169) {
return 8;
}
else if (Rs >= 169) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
return 2;
}
}
}
}
else if (ssDCval >= 44) {
return 4;
}
}
else if (MVDiff >= 43) {
if (mcTcor < 84) {
if (MVDiff < 65) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (MVDiff >= 65) {
if (Rs < 101) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 4;
}
}
else if (Rs >= 101) {
if (gchDC < 6) {
if (tcor < 34) {
if (Cs < 33) {
return 1;
}
else if (Cs >= 33) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (tcor >= 34) {
return 4;
}
}
else if (gchDC >= 6) {
if (Rs < 171) {
return 1;
}
else if (Rs >= 171) {
return 8;
}
}
}
}
}
else if (mcTcor >= 84) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
}
else if (mcTcor >= 88) {
if (posBalance < 10) {
if (posBalance < 2) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
else if (posBalance >= 2) {
return 8;
}
}
else if (posBalance >= 10) {
if (mcTcor < 97) {
if (refDCval < 42) {
if (TSC < 4) {
if (MVDiff < 28) {
if (tcor < 77) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
if (Cs < 263) {
return 4;
}
else if (Cs >= 263) {
return 1;
}
}
}
else if (tcor >= 77) {
if (refDCval < 38) {
if (posBalance < 54) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Cs < 204) {
return 4;
}
else if (Cs >= 204) {
return 8;
}
}
}
else if (posBalance >= 54) {
if (diffMVdiffVal < -9) {
return 1;
}
else if (diffMVdiffVal >= -9) {
if (Cs < 268) {
return 2;
}
else if (Cs >= 268) {
return 8;
}
}
}
}
else if (refDCval >= 38) {
return 2;
}
}
}
else if (MVDiff >= 28) {
if (SC < 70) {
if (Rs < 30) {
return 8;
}
else if (Rs >= 30) {
return 2;
}
}
else if (SC >= 70) {
if (Rs < 179) {
if (refDCval < 35) {
if (Rs < 88) {
return 2;
}
else if (Rs >= 88) {
return 1;
}
}
else if (refDCval >= 35) {
if (diffMVdiffVal < 18) {
return 1;
}
else if (diffMVdiffVal >= 18) {
if (Rs < 75) {
return 4;
}
else if (Rs >= 75) {
return 1;
}
}
}
}
else if (Rs >= 179) {
return 2;
}
}
}
}
else if (TSC >= 4) {
if (ssDCval < 41) {
if (Rs < 177) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 1;
}
}
else if (Rs >= 177) {
return 1;
}
}
else if (ssDCval >= 41) {
if (MVDiff < 37) {
if (Rs < 203) {
return 1;
}
else if (Rs >= 203) {
return 4;
}
}
else if (MVDiff >= 37) {
return 2;
}
}
}
}
else if (refDCval >= 42) {
if (refDCval < 46) {
if (refDCval < 43) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 2;
}
}
}
else if (refDCval >= 43) {
return 1;
}
}
else if (refDCval >= 46) {
if (negBalance < 57) {
if (posBalance < 51) {
if (AFD < 17) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
else if (AFD >= 17) {
return 2;
}
}
else if (posBalance >= 51) {
if (SC < 334) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
return 2;
}
}
else if (SC >= 334) {
return 1;
}
}
}
else if (negBalance >= 57) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (AFD < 5) {
if (MVDiff < 43) {
return 1;
}
else if (MVDiff >= 43) {
return 4;
}
}
else if (AFD >= 5) {
if (diffMVdiffVal < 30) {
if (posBalance < 54) {
if (negBalance < 41) {
if (Rs < 43) {
return 1;
}
else if (Rs >= 43) {
if (Rs < 160) {
return 4;
}
else if (Rs >= 160) {
if (SC < 556) {
return 1;
}
else if (SC >= 556) {
return 4;
}
}
}
}
else if (negBalance >= 41) {
return 1;
}
}
else if (posBalance >= 54) {
if (mu_mv_mag_sq < 13) {
if (Cs < 149) {
return 2;
}
else if (Cs >= 149) {
return 1;
}
}
else if (mu_mv_mag_sq >= 13) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 30) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (negBalance < 27) {
if (mu_mv_mag_sq < 104) {
if (posBalance < 29) {
if (ssDCval < 37) {
return 4;
}
else if (ssDCval >= 37) {
if (tcor < 85) {
if (Rs < 64) {
return 8;
}
else if (Rs >= 64) {
return 4;
}
}
else if (tcor >= 85) {
if (Rs < 215) {
if (Rs < 48) {
return 2;
}
else if (Rs >= 48) {
return 1;
}
}
else if (Rs >= 215) {
return 8;
}
}
}
}
else if (posBalance >= 29) {
if (mcTcor < 97) {
if (tcor < 91) {
if (ssDCval < 36) {
if (Cs < 226) {
return 8;
}
else if (Cs >= 226) {
return 1;
}
}
else if (ssDCval >= 36) {
if (posBalance < 75) {
if (Rs < 20) {
return 8;
}
else if (Rs >= 20) {
return 1;
}
}
else if (posBalance >= 75) {
return 4;
}
}
}
else if (tcor >= 91) {
if (mu_mv_mag_sq < 66) {
return 1;
}
else if (mu_mv_mag_sq >= 66) {
if (posBalance < 36) {
if (Rs < 185) {
return 1;
}
else if (Rs >= 185) {
return 8;
}
}
else if (posBalance >= 36) {
return 8;
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 50) {
return 1;
}
else if (refDCval >= 50) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (tcor < 94) {
if (posBalance < 68) {
if (diffAFD < 1) {
if (Rs < 90) {
if (Rs < 69) {
return 8;
}
else if (Rs >= 69) {
return 4;
}
}
else if (Rs >= 90) {
return 1;
}
}
else if (diffAFD >= 1) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
else if (posBalance >= 68) {
return 4;
}
}
else if (tcor >= 94) {
if (ssDCval < 35) {
if (Rs < 121) {
return 4;
}
else if (Rs >= 121) {
return 1;
}
}
else if (ssDCval >= 35) {
if (negBalance < 15) {
if (TSC < 1) {
return 8;
}
else if (TSC >= 1) {
return 4;
}
}
else if (negBalance >= 15) {
if (Rs < 229) {
if (diffMVdiffVal < -7) {
if (Cs < 89) {
return 8;
}
else if (Cs >= 89) {
return 1;
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
else if (Rs >= 229) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 27) {
if (gchDC < 1) {
if (ssDCval < 42) {
if (negBalance < 41) {
return 8;
}
else if (negBalance >= 41) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (ssDCval >= 42) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (mu_mv_mag_sq < 93) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 93) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
return 1;
}
}
}
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 96) {
if (mcTcor < 105) {
if (tcor < 76) {
if (Rs < 104) {
if (diffMVdiffVal < -67) {
if (Rs < 27) {
return 8;
}
else if (Rs >= 27) {
return 2;
}
}
else if (diffMVdiffVal >= -67) {
return 8;
}
}
else if (Rs >= 104) {
if (TSC < 5) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 4;
}
}
else if (TSC >= 5) {
if (Rs < 123) {
return 2;
}
else if (Rs >= 123) {
return 1;
}
}
}
}
else if (tcor >= 76) {
if (negBalance < 42) {
if (ssDCval < 35) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
else if (ssDCval >= 35) {
if (AFD < 5) {
if (Cs < 95) {
if (Rs < 30) {
return 8;
}
else if (Rs >= 30) {
return 1;
}
}
else if (Cs >= 95) {
return 8;
}
}
else if (AFD >= 5) {
if (Rs < 175) {
return 1;
}
else if (Rs >= 175) {
return 4;
}
}
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
else if (mcTcor >= 105) {
return 4;
}
}
else if (mu_mv_mag_sq >= 96) {
if (AFD < 2) {
if (Rs < 14) {
return 2;
}
else if (Rs >= 14) {
return 8;
}
}
else if (AFD >= 2) {
if (diffMVdiffVal < 21) {
if (Cs < 50) {
if (Rs < 122) {
return 4;
}
else if (Rs >= 122) {
return 2;
}
}
else if (Cs >= 50) {
if (SC < 107) {
return 8;
}
else if (SC >= 107) {
if (negBalance < 47) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 100) {
return 8;
}
else if (Rs >= 100) {
return 4;
}
}
}
else if (negBalance >= 47) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 21) {
if (Rs < 20) {
return 2;
}
else if (Rs >= 20) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 51) {
if (TSCindex < 2) {
if (Cs < 250) {
if (Rs < 157) {
if (mu_mv_mag_sq < 113) {
if (Cs < 48) {
if (tcor < 101) {
return 1;
}
else if (tcor >= 101) {
return 2;
}
}
else if (Cs >= 48) {
if (ssDCval < 85) {
if (ssDCval < 58) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (Rs < 141) {
return 4;
}
else if (Rs >= 141) {
if (MVDiff < 3) {
if (Rs < 146) {
if (Rs < 143) {
return 1;
}
else if (Rs >= 143) {
return 4;
}
}
else if (Rs >= 146) {
return 1;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
}
}
else if (ssDCval >= 58) {
if (Rs < 121) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
return 2;
}
}
else if (Rs >= 121) {
if (AFD < 2) {
if (posBalance < 46) {
if (diffMVdiffVal < -2) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 152) {
if (Cs < 97) {
return 2;
}
else if (Cs >= 97) {
return 1;
}
}
else if (Rs >= 152) {
return 2;
}
}
}
else if (posBalance >= 46) {
return 1;
}
}
else if (AFD >= 2) {
return 2;
}
}
}
}
else if (ssDCval >= 85) {
if (Rs < 121) {
return 4;
}
else if (Rs >= 121) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (Cs < 120) {
if (Rs < 86) {
return 4;
}
else if (Rs >= 86) {
return 1;
}
}
else if (Cs >= 120) {
if (ssDCval < 66) {
if (ssDCval < 57) {
if (Cs < 169) {
return 4;
}
else if (Cs >= 169) {
return 8;
}
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (ssDCval >= 66) {
return 2;
}
}
}
}
else if (Rs >= 157) {
if (ssDCval < 65) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (ssDCval >= 65) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
if (Rs < 199) {
return 8;
}
else if (Rs >= 199) {
return 1;
}
}
}
}
}
else if (Cs >= 250) {
if (AFD < 2) {
if (ssDCval < 78) {
if (ssDCval < 66) {
if (refDCval < 58) {
return 1;
}
else if (refDCval >= 58) {
return 4;
}
}
else if (ssDCval >= 66) {
if (Rs < 183) {
if (MVDiff < 2) {
if (Cs < 379) {
return 4;
}
else if (Cs >= 379) {
return 1;
}
}
else if (MVDiff >= 2) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 1;
}
}
}
else if (Rs >= 183) {
return 1;
}
}
}
else if (ssDCval >= 78) {
return 4;
}
}
else if (AFD >= 2) {
if (Rs < 115) {
return 4;
}
else if (Rs >= 115) {
return 2;
}
}
}
}
else if (TSCindex >= 2) {
if (refDCval < 71) {
if (Cs < 188) {
if (AFD < 4) {
if (Rs < 207) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 47) {
return 1;
}
else if (mu_mv_mag_sq >= 47) {
if (ssDCval < 66) {
if (negBalance < 35) {
if (tcor < 95) {
if (posBalance < 27) {
return 1;
}
else if (posBalance >= 27) {
if (Cs < 61) {
return 2;
}
else if (Cs >= 61) {
if (Rs < 186) {
return 8;
}
else if (Rs >= 186) {
return 4;
}
}
}
}
else if (tcor >= 95) {
if (negBalance < 21) {
return 4;
}
else if (negBalance >= 21) {
if (Cs < 144) {
return 1;
}
else if (Cs >= 144) {
if (Rs < 142) {
return 4;
}
else if (Rs >= 142) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 35) {
if (refDCval < 61) {
if (Rs < 107) {
return 8;
}
else if (Rs >= 107) {
return 1;
}
}
else if (refDCval >= 61) {
return 2;
}
}
}
else if (ssDCval >= 66) {
if (Rs < 142) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 2;
}
}
else if (Rs >= 142) {
return 2;
}
}
}
}
else if (posBalance >= 49) {
if (MVDiff < 1) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
return 1;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
}
else if (Rs >= 207) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
return 4;
}
}
}
else if (AFD >= 4) {
if (TSC < 5) {
if (mu_mv_mag_sq < 50) {
if (Cs < 31) {
if (Rs < 4) {
return 2;
}
else if (Rs >= 4) {
if (SC < 111) {
return 1;
}
else if (SC >= 111) {
return 2;
}
}
}
else if (Cs >= 31) {
if (MVDiff < 7) {
if (posBalance < 61) {
return 1;
}
else if (posBalance >= 61) {
return 4;
}
}
else if (MVDiff >= 7) {
if (mcTcor < 92) {
if (MVDiff < 12) {
if (Cs < 105) {
return 2;
}
else if (Cs >= 105) {
return 8;
}
}
else if (MVDiff >= 12) {
if (posBalance < 62) {
if (AFD < 6) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
return 4;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (posBalance >= 62) {
if (Rs < 159) {
return 8;
}
else if (Rs >= 159) {
return 2;
}
}
}
}
else if (mcTcor >= 92) {
if (AFD < 6) {
if (MVDiff < 17) {
if (ssDCval < 63) {
if (ssDCval < 59) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (MVDiff >= 17) {
if (Rs < 40) {
return 1;
}
else if (Rs >= 40) {
if (Cs < 97) {
return 1;
}
else if (Cs >= 97) {
return 4;
}
}
}
}
else if (AFD >= 6) {
if (MVDiff < 44) {
if (ssDCval < 71) {
return 4;
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (MVDiff >= 44) {
if (Rs < 161) {
if (Cs < 66) {
if (Rs < 22) {
return 4;
}
else if (Rs >= 22) {
return 1;
}
}
else if (Cs >= 66) {
return 1;
}
}
else if (Rs >= 161) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (posBalance < 50) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (Cs < 90) {
return 2;
}
else if (Cs >= 90) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
}
else if (TSC >= 5) {
if (AFD < 16) {
if (Cs < 34) {
return 1;
}
else if (Cs >= 34) {
if (MVDiff < 114) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
return 2;
}
}
else if (MVDiff >= 114) {
return 1;
}
}
}
else if (AFD >= 16) {
if (AFD < 24) {
if (refDCval < 60) {
if (Rs < 65) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 4;
}
}
else if (Rs >= 65) {
if (diffMVdiffVal < -39) {
return 2;
}
else if (diffMVdiffVal >= -39) {
return 1;
}
}
}
else if (refDCval >= 60) {
if (mu_mv_mag_sq < 22) {
return 8;
}
else if (mu_mv_mag_sq >= 22) {
if (Cs < 49) {
return 8;
}
else if (Cs >= 49) {
return 4;
}
}
}
}
else if (AFD >= 24) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (Rs < 205) {
return 1;
}
else if (Rs >= 205) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 188) {
if (negBalance < 44) {
if (negBalance < 30) {
if (MVDiff < 1) {
if (ssDCval < 63) {
if (Cs < 230) {
return 4;
}
else if (Cs >= 230) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 232) {
return 4;
}
else if (Rs >= 232) {
return 1;
}
}
}
}
else if (ssDCval >= 63) {
if (Cs < 365) {
if (Cs < 220) {
return 2;
}
else if (Cs >= 220) {
if (ssDCval < 69) {
if (Rs < 176) {
if (Cs < 337) {
if (Rs < 153) {
if (mu_mv_mag_sq < 121) {
if (Cs < 283) {
return 1;
}
else if (Cs >= 283) {
if (Rs < 151) {
return 4;
}
else if (Rs >= 151) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 121) {
return 4;
}
}
else if (Rs >= 153) {
return 4;
}
}
else if (Cs >= 337) {
if (mu_mv_mag_sq < 119) {
return 4;
}
else if (mu_mv_mag_sq >= 119) {
return 1;
}
}
}
else if (Rs >= 176) {
return 4;
}
}
else if (ssDCval >= 69) {
return 4;
}
}
}
else if (Cs >= 365) {
if (ssDCval < 69) {
if (Rs < 189) {
return 1;
}
else if (Rs >= 189) {
return 4;
}
}
else if (ssDCval >= 69) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
return 4;
}
}
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 66) {
if (tcor < 85) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (tcor >= 85) {
if (Cs < 192) {
if (Rs < 141) {
return 4;
}
else if (Rs >= 141) {
return 1;
}
}
else if (Cs >= 192) {
if (mu_mv_mag_sq < 41) {
if (Rs < 212) {
return 1;
}
else if (Rs >= 212) {
return 2;
}
}
else if (mu_mv_mag_sq >= 41) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (Rs < 193) {
return 4;
}
else if (Rs >= 193) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 66) {
if (TSC < 2) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 117) {
return 1;
}
else if (Rs >= 117) {
return 2;
}
}
}
else if (TSC >= 2) {
if (Rs < 136) {
return 4;
}
else if (Rs >= 136) {
if (mu_mv_mag_sq < 99) {
if (posBalance < 95) {
return 1;
}
else if (posBalance >= 95) {
return 4;
}
}
else if (mu_mv_mag_sq >= 99) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (Rs < 191) {
if (Cs < 336) {
return 4;
}
else if (Cs >= 336) {
return 1;
}
}
else if (Rs >= 191) {
return 4;
}
}
}
}
}
}
}
}
else if (negBalance >= 30) {
if (tcor < 98) {
if (Rs < 37) {
return 1;
}
else if (Rs >= 37) {
if (mu_mv_mag_sq < 69) {
if (MVDiff < 30) {
if (posBalance < 52) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
if (Rs < 234) {
return 4;
}
else if (Rs >= 234) {
if (Rs < 235) {
return 1;
}
else if (Rs >= 235) {
return 4;
}
}
}
}
else if (posBalance >= 52) {
if (Rs < 235) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
if (negBalance < 35) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
else if (negBalance >= 35) {
if (Cs < 202) {
return 1;
}
else if (Cs >= 202) {
return 4;
}
}
}
}
else if (Rs >= 235) {
return 8;
}
}
}
else if (MVDiff >= 30) {
if (diffMVdiffVal < -6) {
if (Rs < 185) {
return 2;
}
else if (Rs >= 185) {
return 4;
}
}
else if (diffMVdiffVal >= -6) {
if (diffAFD < -6) {
return 8;
}
else if (diffAFD >= -6) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (negBalance < 35) {
if (MVDiff < 1) {
if (SC < 536) {
return 1;
}
else if (SC >= 536) {
if (Rs < 203) {
return 4;
}
else if (Rs >= 203) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 206) {
return 1;
}
else if (Cs >= 206) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
if (mcTcor < 98) {
if (posBalance < 47) {
if (Rs < 203) {
return 1;
}
else if (Rs >= 203) {
return 4;
}
}
else if (posBalance >= 47) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 35) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
if (Rs < 139) {
return 8;
}
else if (Rs >= 139) {
return 4;
}
}
}
}
}
}
else if (tcor >= 98) {
if (Cs < 286) {
if (ssDCval < 62) {
if (diffMVdiffVal < -1) {
if (Rs < 188) {
if (Rs < 145) {
return 8;
}
else if (Rs >= 145) {
return 1;
}
}
else if (Rs >= 188) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
return 4;
}
}
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (Cs >= 286) {
if (MVDiff < 4) {
if (SC < 455) {
return 4;
}
else if (SC >= 455) {
if (ssDCval < 67) {
if (Cs < 493) {
return 1;
}
else if (Cs >= 493) {
return 4;
}
}
else if (ssDCval >= 67) {
if (Rs < 172) {
return 1;
}
else if (Rs >= 172) {
return 4;
}
}
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
}
else if (negBalance >= 44) {
if (mu_mv_mag_sq < 26) {
if (ssDCval < 67) {
if (AFD < 7) {
if (Rs < 102) {
return 1;
}
else if (Rs >= 102) {
return 4;
}
}
else if (AFD >= 7) {
if (diffRsCsdiff < -26) {
return 8;
}
else if (diffRsCsdiff >= -26) {
if (refDCval < 55) {
return 2;
}
else if (refDCval >= 55) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (Rs < 220) {
if (MVDiff < 73) {
return 4;
}
else if (MVDiff >= 73) {
return 2;
}
}
else if (Rs >= 220) {
if (Rs < 225) {
return 1;
}
else if (Rs >= 225) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (diffMVdiffVal < -1) {
if (CsDiff < 5) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
return 2;
}
}
else if (CsDiff >= 5) {
if (Rs < 115) {
return 1;
}
else if (Rs >= 115) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (diffAFD < -1) {
if (Rs < 188) {
return 1;
}
else if (Rs >= 188) {
return 8;
}
}
else if (diffAFD >= -1) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
if (Rs < 152) {
return 2;
}
else if (Rs >= 152) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 71) {
if (Cs < 232) {
if (Rs < 111) {
if (negBalance < 24) {
if (diffMVdiffVal < 6) {
if (Cs < 54) {
return 1;
}
else if (Cs >= 54) {
if (Rs < 103) {
return 4;
}
else if (Rs >= 103) {
return 1;
}
}
}
else if (diffMVdiffVal >= 6) {
if (Cs < 88) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
return 4;
}
}
else if (Cs >= 88) {
return 2;
}
}
}
else if (negBalance >= 24) {
if (ssDCval < 108) {
if (SC < 111) {
if (tcor < 38) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
if (AFD < 41) {
return 1;
}
else if (AFD >= 41) {
return 8;
}
}
}
else if (tcor >= 38) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 4;
}
}
}
else if (SC >= 111) {
if (Rs < 21) {
return 2;
}
else if (Rs >= 21) {
if (Rs < 56) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 1;
}
}
}
else if (Rs >= 56) {
if (negBalance < 43) {
if (ssDCval < 73) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
else if (ssDCval >= 73) {
if (ssDCval < 106) {
if (Cs < 165) {
return 1;
}
else if (Cs >= 165) {
if (mu_mv_mag_sq < 64) {
if (Rs < 76) {
return 4;
}
else if (Rs >= 76) {
return 2;
}
}
else if (mu_mv_mag_sq >= 64) {
return 1;
}
}
}
else if (ssDCval >= 106) {
return 2;
}
}
}
else if (negBalance >= 43) {
if (refDCval < 80) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (posBalance < 58) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (posBalance >= 58) {
if (AFD < 37) {
return 2;
}
else if (AFD >= 37) {
return 4;
}
}
}
}
else if (refDCval >= 80) {
if (Cs < 52) {
if (diffMVdiffVal < 6) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
return 8;
}
}
else if (Cs >= 52) {
if (Rs < 99) {
if (refDCval < 82) {
return 8;
}
else if (refDCval >= 82) {
return 1;
}
}
else if (Rs >= 99) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 108) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 3) {
if (posBalance < 63) {
if (Rs < 1) {
return 1;
}
else if (Rs >= 1) {
if (mu_mv_mag_sq < 33) {
return 2;
}
else if (mu_mv_mag_sq >= 33) {
return 1;
}
}
}
else if (posBalance >= 63) {
return 8;
}
}
else if (Rs >= 3) {
if (AFD < 23) {
if (mu_mv_mag_sq < 52) {
if (RsCsDiff < 3) {
if (Rs < 101) {
if (MVDiff < 394) {
return 2;
}
else if (MVDiff >= 394) {
return 1;
}
}
else if (Rs >= 101) {
return 1;
}
}
else if (RsCsDiff >= 3) {
if (Rs < 93) {
return 1;
}
else if (Rs >= 93) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
return 8;
}
}
}
else if (AFD >= 23) {
if (Rs < 97) {
return 8;
}
else if (Rs >= 97) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 111) {
if (TSC < 2) {
if (Cs < 126) {
if (mu_mv_mag_sq < 94) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (ssDCval < 114) {
if (Cs < 96) {
return 2;
}
else if (Cs >= 96) {
return 1;
}
}
else if (ssDCval >= 114) {
if (Rs < 215) {
return 2;
}
else if (Rs >= 215) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (Cs < 88) {
return 1;
}
else if (Cs >= 88) {
if (negBalance < 22) {
return 8;
}
else if (negBalance >= 22) {
if (Cs < 121) {
return 8;
}
else if (Cs >= 121) {
return 2;
}
}
}
}
}
else if (Cs >= 126) {
if (negBalance < 29) {
if (Cs < 210) {
if (Rs < 233) {
if (diffMVdiffVal < 9) {
return 2;
}
else if (diffMVdiffVal >= 9) {
return 1;
}
}
else if (Rs >= 233) {
if (Rs < 236) {
return 8;
}
else if (Rs >= 236) {
return 2;
}
}
}
else if (Cs >= 210) {
if (ssDCval < 74) {
if (Rs < 181) {
return 2;
}
else if (Rs >= 181) {
return 1;
}
}
else if (ssDCval >= 74) {
return 2;
}
}
}
else if (negBalance >= 29) {
if (Cs < 187) {
if (Rs < 163) {
return 1;
}
else if (Rs >= 163) {
return 8;
}
}
else if (Cs >= 187) {
if (Rs < 156) {
return 2;
}
else if (Rs >= 156) {
return 4;
}
}
}
}
}
else if (TSC >= 2) {
if (negBalance < 84) {
if (refDCval < 97) {
if (MVDiff < 3) {
if (mcTcor < 98) {
if (Cs < 149) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (Cs >= 149) {
if (Rs < 155) {
return 8;
}
else if (Rs >= 155) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 186) {
if (Rs < 129) {
return 1;
}
else if (Rs >= 129) {
return 2;
}
}
else if (Rs >= 186) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
if (Cs < 194) {
return 1;
}
else if (Cs >= 194) {
return 2;
}
}
}
}
}
else if (MVDiff >= 3) {
if (tcor < 78) {
if (ssDCval < 81) {
if (mu_mv_mag_sq < 7) {
if (Rs < 186) {
return 2;
}
else if (Rs >= 186) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 57) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
return 1;
}
}
else if (Cs >= 57) {
return 4;
}
}
}
else if (ssDCval >= 81) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
if (Rs < 233) {
if (CsDiff < 3) {
if (AFD < 18) {
if (MVDiff < 49) {
if (MVDiff < 33) {
if (SC < 161) {
return 8;
}
else if (SC >= 161) {
if (Rs < 175) {
if (Cs < 204) {
if (TSC < 3) {
if (Rs < 153) {
return 2;
}
else if (Rs >= 153) {
if (Rs < 161) {
return 8;
}
else if (Rs >= 161) {
return 2;
}
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (Cs >= 204) {
return 8;
}
}
else if (Rs >= 175) {
return 2;
}
}
}
else if (MVDiff >= 33) {
if (Cs < 77) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
else if (Cs >= 77) {
return 8;
}
}
}
else if (MVDiff >= 49) {
return 8;
}
}
else if (AFD >= 18) {
return 2;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (Rs >= 233) {
return 1;
}
}
}
}
else if (tcor >= 78) {
if (AFD < 11) {
if (mu_mv_mag_sq < 5) {
if (Cs < 73) {
if (ssDCval < 96) {
return 1;
}
else if (ssDCval >= 96) {
return 2;
}
}
else if (Cs >= 73) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 135) {
if (ssDCval < 74) {
if (Rs < 120) {
return 2;
}
else if (Rs >= 120) {
if (Rs < 124) {
return 8;
}
else if (Rs >= 124) {
return 1;
}
}
}
else if (ssDCval >= 74) {
if (MVDiff < 28) {
return 1;
}
else if (MVDiff >= 28) {
return 2;
}
}
}
else if (Rs >= 135) {
if (refDCval < 95) {
if (negBalance < 24) {
return 8;
}
else if (negBalance >= 24) {
if (ssDCval < 70) {
if (Rs < 185) {
return 1;
}
else if (Rs >= 185) {
return 8;
}
}
else if (ssDCval >= 70) {
if (SC < 288) {
if (Cs < 53) {
return 1;
}
else if (Cs >= 53) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Rs < 147) {
return 1;
}
else if (Rs >= 147) {
return 8;
}
}
}
}
else if (SC >= 288) {
return 1;
}
}
}
}
else if (refDCval >= 95) {
return 8;
}
}
}
}
else if (AFD >= 11) {
if (refDCval < 89) {
return 2;
}
else if (refDCval >= 89) {
if (Rs < 158) {
return 8;
}
else if (Rs >= 158) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 97) {
if (TSC < 4) {
if (Cs < 33) {
return 1;
}
else if (Cs >= 33) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (TSC >= 4) {
if (MVDiff < 72) {
if (mu_mv_mag_sq < 9) {
if (Rs < 225) {
return 4;
}
else if (Rs >= 225) {
if (Rs < 230) {
return 8;
}
else if (Rs >= 230) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (MVDiff < 22) {
return 1;
}
else if (MVDiff >= 22) {
if (Cs < 84) {
return 1;
}
else if (Cs >= 84) {
return 2;
}
}
}
}
else if (MVDiff >= 72) {
if (tcor < 44) {
return 4;
}
else if (tcor >= 44) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 84) {
if (AFD < 40) {
return 8;
}
else if (AFD >= 40) {
return 1;
}
}
}
}
}
else if (Cs >= 232) {
if (ssDCval < 77) {
if (negBalance < 36) {
if (Rs < 114) {
return 4;
}
else if (Rs >= 114) {
if (posBalance < 57) {
if (diffRsCsdiff < -1) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (posBalance >= 57) {
if (Rs < 128) {
return 4;
}
else if (Rs >= 128) {
return 1;
}
}
}
}
else if (negBalance >= 36) {
if (Rs < 176) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (Rs >= 176) {
if (Rs < 200) {
if (mcTcor < 97) {
if (Cs < 259) {
return 1;
}
else if (Cs >= 259) {
return 8;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (Rs >= 200) {
if (MVDiff < 5) {
if (Cs < 276) {
return 2;
}
else if (Cs >= 276) {
return 1;
}
}
else if (MVDiff >= 5) {
if (Rs < 214) {
return 4;
}
else if (Rs >= 214) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 77) {
if (AFD < 4) {
if (ssDCval < 135) {
if (posBalance < 44) {
if (Cs < 265) {
return 2;
}
else if (Cs >= 265) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (posBalance >= 44) {
if (mu_mv_mag_sq < 77) {
return 4;
}
else if (mu_mv_mag_sq >= 77) {
if (ssDCval < 89) {
if (Cs < 256) {
return 2;
}
else if (Cs >= 256) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
}
else if (ssDCval >= 89) {
if (Rs < 236) {
return 1;
}
else if (Rs >= 236) {
return 4;
}
}
}
}
}
else if (ssDCval >= 135) {
return 2;
}
}
else if (AFD >= 4) {
if (tcor < 88) {
if (posBalance < 60) {
if (MVDiff < 7) {
if (Rs < 215) {
if (Cs < 249) {
return 2;
}
else if (Cs >= 249) {
return 8;
}
}
else if (Rs >= 215) {
return 2;
}
}
else if (MVDiff >= 7) {
if (mcTcor < 93) {
if (SC < 586) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (Cs < 292) {
return 2;
}
else if (Cs >= 292) {
return 8;
}
}
}
else if (SC >= 586) {
return 8;
}
}
else if (mcTcor >= 93) {
if (MVDiff < 38) {
return 1;
}
else if (MVDiff >= 38) {
return 2;
}
}
}
}
else if (posBalance >= 60) {
if (mu_mv_mag_sq < 19) {
if (Rs < 127) {
return 2;
}
else if (Rs >= 127) {
if (Cs < 240) {
return 8;
}
else if (Cs >= 240) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 19) {
if (SC < 408) {
return 1;
}
else if (SC >= 408) {
if (diffMVdiffVal < -12) {
return 1;
}
else if (diffMVdiffVal >= -12) {
if (Rs < 227) {
return 8;
}
else if (Rs >= 227) {
return 1;
}
}
}
}
}
}
else if (tcor >= 88) {
if (ssDCval < 78) {
if (Rs < 192) {
return 4;
}
else if (Rs >= 192) {
return 2;
}
}
else if (ssDCval >= 78) {
if (TSC < 3) {
if (Rs < 217) {
return 1;
}
else if (Rs >= 217) {
return 2;
}
}
else if (TSC >= 3) {
if (Cs < 242) {
if (Rs < 167) {
return 1;
}
else if (Rs >= 167) {
return 8;
}
}
else if (Cs >= 242) {
if (tcor < 90) {
if (SC < 517) {
return 2;
}
else if (SC >= 517) {
return 1;
}
}
else if (tcor >= 90) {
if (Rs < 228) {
return 2;
}
else if (Rs >= 228) {
if (Rs < 232) {
return 8;
}
else if (Rs >= 232) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 241) {
if (RsDiff < 4) {
if (refDCval < 75) {
if (mcTcor < 89) {
if (refDCval < 41) {
if (tcor < 64) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (Cs < 84) {
return 1;
}
else if (Cs >= 84) {
if (tcor < 37) {
if (Rs < 346) {
return 4;
}
else if (Rs >= 346) {
return 1;
}
}
else if (tcor >= 37) {
if (diffRsCsdiff < -2) {
if (CsDiff < 3) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
}
}
else if (tcor >= 64) {
if (Cs < 152) {
if (ssDCval < 39) {
if (MVDiff < 238) {
return 1;
}
else if (MVDiff >= 238) {
return 4;
}
}
else if (ssDCval >= 39) {
if (Rs < 617) {
return 8;
}
else if (Rs >= 617) {
return 2;
}
}
}
else if (Cs >= 152) {
if (negBalance < 33) {
if (Rs < 321) {
if (SC < 502) {
return 8;
}
else if (SC >= 502) {
return 2;
}
}
else if (Rs >= 321) {
if (posBalance < 61) {
return 1;
}
else if (posBalance >= 61) {
return 8;
}
}
}
else if (negBalance >= 33) {
if (Rs < 245) {
if (Rs < 244) {
return 8;
}
else if (Rs >= 244) {
return 2;
}
}
else if (Rs >= 245) {
if (Rs < 774) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (Rs >= 774) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 41) {
if (mu_mv_mag_sq < 94) {
if (posBalance < 54) {
if (Cs < 148) {
if (mu_mv_mag_sq < 22) {
return 8;
}
else if (mu_mv_mag_sq >= 22) {
return 1;
}
}
else if (Cs >= 148) {
if (Rs < 423) {
if (Cs < 321) {
if (AFD < 28) {
return 8;
}
else if (AFD >= 28) {
return 2;
}
}
else if (Cs >= 321) {
if (tcor < 62) {
return 8;
}
else if (tcor >= 62) {
if (Rs < 280) {
return 8;
}
else if (Rs >= 280) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
else if (Rs >= 423) {
if (diffRsCsdiff < -1) {
if (SC < 930) {
return 2;
}
else if (SC >= 930) {
if (SC < 1075) {
return 1;
}
else if (SC >= 1075) {
if (Rs < 743) {
return 2;
}
else if (Rs >= 743) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= -1) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
return 2;
}
}
}
}
}
else if (posBalance >= 54) {
if (negBalance < 45) {
if (SC < 788) {
if (Cs < 167) {
return 1;
}
else if (Cs >= 167) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
return 2;
}
}
}
else if (SC >= 788) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 642) {
return 1;
}
else if (Rs >= 642) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (MVDiff < 60) {
if (TSC < 6) {
if (Rs < 675) {
return 4;
}
else if (Rs >= 675) {
if (Rs < 758) {
return 1;
}
else if (Rs >= 758) {
return 8;
}
}
}
else if (TSC >= 6) {
if (refDCval < 55) {
if (Rs < 797) {
if (RsCsDiff < 7) {
return 4;
}
else if (RsCsDiff >= 7) {
return 8;
}
}
else if (Rs >= 797) {
return 2;
}
}
else if (refDCval >= 55) {
if (diffMVdiffVal < -9) {
return 1;
}
else if (diffMVdiffVal >= -9) {
return 4;
}
}
}
}
else if (MVDiff >= 60) {
if (Rs < 393) {
if (RsCsDiff < 16) {
if (negBalance < 54) {
return 4;
}
else if (negBalance >= 54) {
if (Rs < 270) {
return 4;
}
else if (Rs >= 270) {
return 8;
}
}
}
else if (RsCsDiff >= 16) {
if (Rs < 359) {
return 2;
}
else if (Rs >= 359) {
return 4;
}
}
}
else if (Rs >= 393) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1273) {
if (posBalance < 28) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
if (tcor < 101) {
if (RsDiff < 1) {
if (SC < 880) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
return 8;
}
}
else if (SC >= 880) {
if (Cs < 388) {
if (Rs < 745) {
return 8;
}
else if (Rs >= 745) {
return 1;
}
}
else if (Cs >= 388) {
if (gchDC < 1) {
if (Rs < 563) {
return 1;
}
else if (Rs >= 563) {
if (Rs < 795) {
return 4;
}
else if (Rs >= 795) {
return 8;
}
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (RsDiff >= 1) {
if (ssDCval < 60) {
if (refDCval < 36) {
return 2;
}
else if (refDCval >= 36) {
if (SC < 1420) {
return 1;
}
else if (SC >= 1420) {
return 2;
}
}
}
else if (ssDCval >= 60) {
return 8;
}
}
}
else if (tcor >= 101) {
if (SC < 844) {
if (negBalance < 44) {
if (ssDCval < 45) {
if (Rs < 284) {
return 2;
}
else if (Rs >= 284) {
return 1;
}
}
else if (ssDCval >= 45) {
return 4;
}
}
else if (negBalance >= 44) {
if (AFD < 8) {
if (Cs < 298) {
return 8;
}
else if (Cs >= 298) {
if (Rs < 302) {
return 8;
}
else if (Rs >= 302) {
return 4;
}
}
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (SC >= 844) {
return 8;
}
}
}
}
else if (posBalance >= 28) {
if (mu_mv_mag_sq < 107) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Cs < 117) {
if (AFD < 5) {
if (Rs < 575) {
if (ssDCval < 47) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 1;
}
}
else if (ssDCval >= 47) {
return 4;
}
}
else if (Rs >= 575) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (AFD >= 5) {
if (CsDiff < 1) {
if (diffMVdiffVal < -31) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
return 1;
}
}
else if (diffMVdiffVal >= -31) {
return 1;
}
}
else if (CsDiff >= 1) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
if (diffTSC < -1) {
if (tcor < 86) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
else if (tcor >= 86) {
return 1;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
}
}
}
else if (Cs >= 117) {
if (mu_mv_mag_sq < 78) {
if (mcTcor < 96) {
if (ssDCval < 57) {
if (Rs < 465) {
if (MVDiff < 43) {
if (RsCsDiff < 7) {
if (Rs < 306) {
if (MVDiff < 6) {
if (posBalance < 50) {
return 4;
}
else if (posBalance >= 50) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 275) {
return 2;
}
else if (Rs >= 275) {
return 4;
}
}
}
}
else if (MVDiff >= 6) {
if (mcTcor < 94) {
if (negBalance < 32) {
if (Rs < 281) {
return 2;
}
else if (Rs >= 281) {
return 4;
}
}
else if (negBalance >= 32) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 281) {
if (Cs < 230) {
return 1;
}
else if (Cs >= 230) {
return 2;
}
}
else if (Rs >= 281) {
return 1;
}
}
}
}
else if (Rs >= 306) {
if (ssDCval < 44) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
return 1;
}
}
else if (ssDCval >= 44) {
if (refDCval < 56) {
if (MVDiff < 3) {
if (Rs < 354) {
return 4;
}
else if (Rs >= 354) {
return 1;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (refDCval >= 56) {
if (Rs < 364) {
return 2;
}
else if (Rs >= 364) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 7) {
if (posBalance < 59) {
if (ssDCval < 52) {
if (TSC < 4) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (TSC >= 4) {
if (diffRsCsdiff < 4) {
return 8;
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
}
else if (ssDCval >= 52) {
return 1;
}
}
else if (posBalance >= 59) {
if (Rs < 311) {
return 4;
}
else if (Rs >= 311) {
return 8;
}
}
}
}
else if (MVDiff >= 43) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 25) {
return 2;
}
else if (mu_mv_mag_sq >= 25) {
if (ssDCval < 43) {
return 2;
}
else if (ssDCval >= 43) {
return 4;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 411) {
return 1;
}
else if (Rs >= 411) {
return 2;
}
}
}
}
else if (Rs >= 465) {
if (posBalance < 52) {
if (ssDCval < 49) {
if (posBalance < 32) {
return 2;
}
else if (posBalance >= 32) {
if (mu_mv_mag_sq < 62) {
return 1;
}
else if (mu_mv_mag_sq >= 62) {
if (Rs < 761) {
return 4;
}
else if (Rs >= 761) {
return 1;
}
}
}
}
else if (ssDCval >= 49) {
if (MVDiff < 11) {
if (Rs < 611) {
return 4;
}
else if (Rs >= 611) {
return 8;
}
}
else if (MVDiff >= 11) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (SC < 1164) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (SC < 925) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
else if (SC >= 925) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (diffRsCsdiff < 3) {
if (mu_mv_mag_sq < 31) {
return 1;
}
else if (mu_mv_mag_sq >= 31) {
if (Rs < 632) {
return 4;
}
else if (Rs >= 632) {
return 1;
}
}
}
else if (diffRsCsdiff >= 3) {
return 8;
}
}
}
}
}
else if (SC >= 1164) {
if (negBalance < 39) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
else if (negBalance >= 39) {
if (Rs < 888) {
return 8;
}
else if (Rs >= 888) {
if (Rs < 953) {
return 4;
}
else if (Rs >= 953) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 57) {
if (SC < 1075) {
if (ssDCval < 62) {
if (RsCsDiff < 9) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (tcor < 93) {
if (mu_mv_mag_sq < 60) {
if (mu_mv_mag_sq < 12) {
if (AFD < 9) {
if (Rs < 370) {
return 2;
}
else if (Rs >= 370) {
return 4;
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (mu_mv_mag_sq >= 12) {
return 2;
}
}
else if (mu_mv_mag_sq >= 60) {
return 4;
}
}
else if (tcor >= 93) {
return 1;
}
}
}
else if (RsCsDiff >= 9) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (SC < 621) {
return 2;
}
else if (SC >= 621) {
return 1;
}
}
}
}
else if (ssDCval >= 62) {
if (AFD < 8) {
if (ssDCval < 70) {
if (mcTcor < 95) {
if (RsDiff < 3) {
if (SC < 734) {
return 1;
}
else if (SC >= 734) {
return 2;
}
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (mcTcor >= 95) {
if (Rs < 494) {
return 4;
}
else if (Rs >= 494) {
return 1;
}
}
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (AFD >= 8) {
if (RsDiff < 3) {
if (diffAFD < -3) {
if (Rs < 364) {
return 8;
}
else if (Rs >= 364) {
return 2;
}
}
else if (diffAFD >= -3) {
if (Rs < 278) {
return 2;
}
else if (Rs >= 278) {
return 4;
}
}
}
else if (RsDiff >= 3) {
if (diffMVdiffVal < -11) {
if (Rs < 828) {
return 1;
}
else if (Rs >= 828) {
return 2;
}
}
else if (diffMVdiffVal >= -11) {
if (negBalance < 58) {
if (MVDiff < 38) {
return 8;
}
else if (MVDiff >= 38) {
return 1;
}
}
else if (negBalance >= 58) {
return 4;
}
}
}
}
}
}
else if (SC >= 1075) {
if (Rs < 1080) {
if (Cs < 411) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Rs < 811) {
return 4;
}
else if (Rs >= 811) {
return 1;
}
}
}
else if (Cs >= 411) {
return 4;
}
}
else if (Rs >= 1080) {
if (Cs < 465) {
return 1;
}
else if (Cs >= 465) {
return 8;
}
}
}
}
}
else if (mcTcor >= 96) {
if (MVDiff < 2) {
if (negBalance < 42) {
if (mcTcor < 97) {
if (refDCval < 68) {
if (Cs < 440) {
return 8;
}
else if (Cs >= 440) {
return 2;
}
}
else if (refDCval >= 68) {
if (Rs < 569) {
return 1;
}
else if (Rs >= 569) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (diffRsCsdiff < 3) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
if (posBalance < 59) {
if (ssDCval < 56) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (ssDCval >= 56) {
return 1;
}
}
else if (posBalance >= 59) {
if (Rs < 468) {
return 1;
}
else if (Rs >= 468) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 3) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (negBalance >= 42) {
if (AFD < 8) {
if (Cs < 158) {
return 1;
}
else if (Cs >= 158) {
return 8;
}
}
else if (AFD >= 8) {
if (Rs < 519) {
return 1;
}
else if (Rs >= 519) {
return 4;
}
}
}
}
else if (MVDiff >= 2) {
if (SC < 1097) {
if (tcor < 98) {
if (TSC < 4) {
if (MVDiff < 7) {
if (Cs < 383) {
if (mu_mv_mag_sq < 54) {
if (CsDiff < 2) {
if (posBalance < 67) {
if (SC < 589) {
if (mu_mv_mag_sq < 47) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (Rs < 291) {
return 1;
}
else if (Rs >= 291) {
return 4;
}
}
}
else if (SC >= 589) {
if (ssDCval < 65) {
if (SC < 721) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 1;
}
}
else if (SC >= 721) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Rs < 474) {
return 1;
}
else if (Rs >= 474) {
return 4;
}
}
}
}
else if (ssDCval >= 65) {
if (tcor < 97) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
return 4;
}
}
}
else if (tcor >= 97) {
return 1;
}
}
}
}
else if (posBalance >= 67) {
return 2;
}
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 54) {
if (tcor < 83) {
return 2;
}
else if (tcor >= 83) {
if (Cs < 150) {
return 2;
}
else if (Cs >= 150) {
if (RsDiff < 1) {
if (Cs < 275) {
return 1;
}
else if (Cs >= 275) {
if (Rs < 305) {
return 4;
}
else if (Rs >= 305) {
return 2;
}
}
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
}
else if (Cs >= 383) {
if (Rs < 244) {
return 1;
}
else if (Rs >= 244) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Rs < 263) {
if (Rs < 246) {
return 4;
}
else if (Rs >= 246) {
return 8;
}
}
else if (Rs >= 263) {
if (mu_mv_mag_sq < 75) {
return 4;
}
else if (mu_mv_mag_sq >= 75) {
if (Rs < 474) {
return 1;
}
else if (Rs >= 474) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (refDCval < 45) {
if (posBalance < 45) {
if (Cs < 236) {
return 4;
}
else if (Cs >= 236) {
if (posBalance < 36) {
return 2;
}
else if (posBalance >= 36) {
if (Cs < 303) {
return 8;
}
else if (Cs >= 303) {
return 1;
}
}
}
}
else if (posBalance >= 45) {
return 1;
}
}
else if (refDCval >= 45) {
if (posBalance < 60) {
if (SC < 776) {
if (gchDC < 1) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (SC < 681) {
if (SC < 519) {
return 4;
}
else if (SC >= 519) {
return 1;
}
}
else if (SC >= 681) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 49) {
if (Rs < 307) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (Cs < 266) {
return 4;
}
else if (Cs >= 266) {
if (Cs < 436) {
return 2;
}
else if (Cs >= 436) {
return 4;
}
}
}
}
else if (Rs >= 307) {
if (diffRsCsdiff < -10) {
return 8;
}
else if (diffRsCsdiff >= -10) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (RsCsDiff < 2) {
if (Cs < 283) {
return 4;
}
else if (Cs >= 283) {
return 1;
}
}
else if (RsCsDiff >= 2) {
if (refDCval < 48) {
return 1;
}
else if (refDCval >= 48) {
return 2;
}
}
}
}
}
else if (SC >= 776) {
if (MVDiff < 22) {
if (Rs < 289) {
return 1;
}
else if (Rs >= 289) {
if (ssDCval < 71) {
if (negBalance < 42) {
if (RsCsDiff < 5) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 460) {
if (SC < 839) {
return 4;
}
else if (SC >= 839) {
return 1;
}
}
else if (Rs >= 460) {
return 1;
}
}
}
else if (RsCsDiff >= 5) {
return 4;
}
}
else if (negBalance >= 42) {
if (refDCval < 65) {
return 1;
}
else if (refDCval >= 65) {
return 4;
}
}
}
else if (ssDCval >= 71) {
return 1;
}
}
}
else if (MVDiff >= 22) {
if (Rs < 446) {
if (Rs < 405) {
return 2;
}
else if (Rs >= 405) {
return 8;
}
}
else if (Rs >= 446) {
return 1;
}
}
}
}
else if (posBalance >= 60) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
return 2;
}
}
}
}
}
else if (TSC >= 4) {
if (RsCsDiff < 7) {
if (MVDiff < 13) {
if (Cs < 138) {
return 1;
}
else if (Cs >= 138) {
if (diffMVdiffVal < 7) {
return 4;
}
else if (diffMVdiffVal >= 7) {
return 1;
}
}
}
else if (MVDiff >= 13) {
if (mcTcor < 97) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 1;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
}
else if (RsCsDiff >= 7) {
if (ssDCval < 69) {
if (Rs < 558) {
return 1;
}
else if (Rs >= 558) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
return 8;
}
}
}
else if (ssDCval >= 69) {
if (Rs < 704) {
if (Cs < 310) {
return 8;
}
else if (Cs >= 310) {
return 4;
}
}
else if (Rs >= 704) {
return 1;
}
}
}
}
}
else if (tcor >= 98) {
if (Rs < 289) {
if (Cs < 203) {
return 2;
}
else if (Cs >= 203) {
return 1;
}
}
else if (Rs >= 289) {
if (mu_mv_mag_sq < 24) {
if (AFD < 6) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 24) {
if (Rs < 411) {
if (ssDCval < 56) {
if (SC < 832) {
return 1;
}
else if (SC >= 832) {
return 8;
}
}
else if (ssDCval >= 56) {
return 4;
}
}
else if (Rs >= 411) {
if (Rs < 470) {
return 8;
}
else if (Rs >= 470) {
if (CsDiff < 1) {
if (SC < 668) {
return 8;
}
else if (SC >= 668) {
return 1;
}
}
else if (CsDiff >= 1) {
if (Cs < 436) {
if (Rs < 479) {
return 4;
}
else if (Rs >= 479) {
return 8;
}
}
else if (Cs >= 436) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
}
}
}
}
else if (SC >= 1097) {
if (Cs < 291) {
if (mu_mv_mag_sq < 44) {
if (Cs < 203) {
return 1;
}
else if (Cs >= 203) {
if (tcor < 96) {
if (mu_mv_mag_sq < 41) {
return 1;
}
else if (mu_mv_mag_sq >= 41) {
return 4;
}
}
else if (tcor >= 96) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 44) {
return 1;
}
}
else if (Cs >= 291) {
if (posBalance < 55) {
if (RsCsDiff < 4) {
return 1;
}
else if (RsCsDiff >= 4) {
if (negBalance < 30) {
return 2;
}
else if (negBalance >= 30) {
if (diffRsCsdiff < 7) {
return 1;
}
else if (diffRsCsdiff >= 7) {
if (Rs < 698) {
return 8;
}
else if (Rs >= 698) {
return 2;
}
}
}
}
}
else if (posBalance >= 55) {
if (Cs < 493) {
return 4;
}
else if (Cs >= 493) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (refDCval < 71) {
if (Cs < 277) {
if (ssDCval < 48) {
if (negBalance < 34) {
if (posBalance < 31) {
if (Rs < 362) {
return 4;
}
else if (Rs >= 362) {
return 2;
}
}
else if (posBalance >= 31) {
return 1;
}
}
else if (negBalance >= 34) {
if (Rs < 294) {
return 1;
}
else if (Rs >= 294) {
return 8;
}
}
}
else if (ssDCval >= 48) {
if (Cs < 256) {
if (Rs < 414) {
if (SC < 549) {
if (Cs < 217) {
if (Rs < 310) {
return 4;
}
else if (Rs >= 310) {
return 2;
}
}
else if (Cs >= 217) {
return 1;
}
}
else if (SC >= 549) {
return 4;
}
}
else if (Rs >= 414) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (AFD < 4) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (AFD >= 4) {
if (Rs < 443) {
return 1;
}
else if (Rs >= 443) {
return 2;
}
}
}
}
}
else if (Cs >= 256) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 105) {
return 1;
}
else if (mu_mv_mag_sq >= 105) {
return 4;
}
}
else if (MVDiff >= 6) {
if (Rs < 250) {
return 1;
}
else if (Rs >= 250) {
return 4;
}
}
}
}
}
else if (Cs >= 277) {
if (refDCval < 58) {
if (negBalance < 38) {
if (Cs < 409) {
if (negBalance < 37) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (mu_mv_mag_sq < 92) {
if (Rs < 605) {
return 4;
}
else if (Rs >= 605) {
return 1;
}
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
}
else if (negBalance >= 37) {
return 4;
}
}
else if (Cs >= 409) {
if (Rs < 417) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (Rs < 329) {
return 2;
}
else if (Rs >= 329) {
return 1;
}
}
}
else if (Rs >= 417) {
if (mcTcor < 99) {
if (SC < 1271) {
if (diffTSC < 1) {
if (Rs < 443) {
if (Cs < 485) {
return 1;
}
else if (Cs >= 485) {
return 8;
}
}
else if (Rs >= 443) {
return 1;
}
}
else if (diffTSC >= 1) {
if (Rs < 599) {
return 1;
}
else if (Rs >= 599) {
if (Rs < 655) {
return 4;
}
else if (Rs >= 655) {
return 2;
}
}
}
}
else if (SC >= 1271) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (Rs < 1008) {
return 4;
}
else if (Rs >= 1008) {
return 1;
}
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 8;
}
}
}
}
}
else if (negBalance >= 38) {
if (Rs < 345) {
return 1;
}
else if (Rs >= 345) {
return 8;
}
}
}
else if (refDCval >= 58) {
if (negBalance < 26) {
if (Rs < 654) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Rs < 352) {
if (Rs < 317) {
return 2;
}
else if (Rs >= 317) {
return 1;
}
}
else if (Rs >= 352) {
if (ssDCval < 70) {
if (diffMVdiffVal < -3) {
if (Rs < 483) {
return 8;
}
else if (Rs >= 483) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (ssDCval >= 70) {
if (Rs < 449) {
return 8;
}
else if (Rs >= 449) {
return 1;
}
}
}
}
}
else if (Rs >= 654) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
if (Rs < 954) {
return 2;
}
else if (Rs >= 954) {
return 8;
}
}
}
}
else if (negBalance >= 26) {
if (mcTcor < 97) {
if (tcor < 93) {
if (Cs < 477) {
if (Rs < 336) {
return 1;
}
else if (Rs >= 336) {
return 4;
}
}
else if (Cs >= 477) {
return 8;
}
}
else if (tcor >= 93) {
if (Rs < 898) {
if (MVDiff < 3) {
if (Rs < 444) {
return 1;
}
else if (Rs >= 444) {
return 2;
}
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (Rs >= 898) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (Cs < 457) {
if (refDCval < 62) {
if (MVDiff < 4) {
if (Cs < 330) {
return 2;
}
else if (Cs >= 330) {
return 1;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 2;
}
}
}
else if (refDCval >= 62) {
if (posBalance < 40) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (posBalance >= 40) {
return 1;
}
}
}
else if (Cs >= 457) {
if (Rs < 591) {
if (Rs < 293) {
return 1;
}
else if (Rs >= 293) {
return 8;
}
}
else if (Rs >= 591) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
}
}
}
}
else if (refDCval >= 71) {
if (Cs < 437) {
if (negBalance < 34) {
if (tcor < 98) {
if (Cs < 271) {
return 1;
}
else if (Cs >= 271) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Cs < 282) {
return 8;
}
else if (Cs >= 282) {
return 1;
}
}
}
}
else if (tcor >= 98) {
if (posBalance < 51) {
if (ssDCval < 73) {
if (Rs < 325) {
return 2;
}
else if (Rs >= 325) {
return 8;
}
}
else if (ssDCval >= 73) {
return 1;
}
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (negBalance >= 34) {
if (Rs < 1040) {
return 8;
}
else if (Rs >= 1040) {
return 1;
}
}
}
else if (Cs >= 437) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (Rs < 378) {
return 2;
}
else if (Rs >= 378) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (ssDCval < 67) {
if (mcTcor < 97) {
if (Cs < 441) {
if (Cs < 235) {
if (Rs < 554) {
return 8;
}
else if (Rs >= 554) {
return 2;
}
}
else if (Cs >= 235) {
if (mu_mv_mag_sq < 113) {
if (Cs < 362) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 4;
}
}
else if (Cs >= 362) {
return 1;
}
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 308) {
if (Rs < 271) {
return 1;
}
else if (Rs >= 271) {
return 4;
}
}
else if (Rs >= 308) {
return 8;
}
}
}
}
else if (Cs >= 441) {
if (mu_mv_mag_sq < 109) {
return 1;
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 1005) {
if (Rs < 484) {
if (Rs < 424) {
return 8;
}
else if (Rs >= 424) {
return 1;
}
}
else if (Rs >= 484) {
return 8;
}
}
else if (Rs >= 1005) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 342) {
if (refDCval < 60) {
if (ssDCval < 54) {
if (mu_mv_mag_sq < 116) {
return 1;
}
else if (mu_mv_mag_sq >= 116) {
if (refDCval < 52) {
return 8;
}
else if (refDCval >= 52) {
if (Rs < 628) {
return 4;
}
else if (Rs >= 628) {
return 1;
}
}
}
}
else if (ssDCval >= 54) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 60) {
if (SC < 1222) {
return 1;
}
else if (SC >= 1222) {
if (Rs < 925) {
return 2;
}
else if (Rs >= 925) {
return 1;
}
}
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
}
else if (refDCval >= 60) {
if (Rs < 684) {
if (ssDCval < 61) {
if (Rs < 286) {
return 1;
}
else if (Rs >= 286) {
return 2;
}
}
else if (ssDCval >= 61) {
if (negBalance < 28) {
if (Rs < 517) {
if (refDCval < 61) {
return 2;
}
else if (refDCval >= 61) {
return 4;
}
}
else if (Rs >= 517) {
return 8;
}
}
else if (negBalance >= 28) {
if (Rs < 290) {
return 1;
}
else if (Rs >= 290) {
return 2;
}
}
}
}
else if (Rs >= 684) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Rs < 1068) {
return 2;
}
else if (Rs >= 1068) {
return 1;
}
}
}
}
}
else if (Cs >= 342) {
if (SC < 995) {
if (SC < 661) {
if (Rs < 266) {
return 4;
}
else if (Rs >= 266) {
return 1;
}
}
else if (SC >= 661) {
if (AFD < 4) {
if (tcor < 101) {
return 1;
}
else if (tcor >= 101) {
return 8;
}
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (SC >= 995) {
if (AFD < 3) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
if (Rs < 909) {
if (mu_mv_mag_sq < 117) {
if (Rs < 565) {
return 8;
}
else if (Rs >= 565) {
return 1;
}
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
else if (Rs >= 909) {
return 1;
}
}
}
else if (AFD >= 3) {
if (Rs < 617) {
return 4;
}
else if (Rs >= 617) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (SC < 953) {
if (negBalance < 25) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
return 2;
}
}
else if (negBalance >= 25) {
if (posBalance < 52) {
if (Rs < 543) {
if (diffRsCsdiff < -1) {
return 2;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (Rs >= 543) {
if (Rs < 622) {
return 2;
}
else if (Rs >= 622) {
return 4;
}
}
}
else if (posBalance >= 52) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
}
}
else if (SC >= 953) {
if (mu_mv_mag_sq < 123) {
if (mcTcor < 99) {
if (ssDCval < 72) {
if (SC < 1187) {
return 1;
}
else if (SC >= 1187) {
if (Rs < 853) {
return 2;
}
else if (Rs >= 853) {
return 8;
}
}
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (mcTcor >= 99) {
if (posBalance < 40) {
if (Rs < 779) {
return 4;
}
else if (Rs >= 779) {
return 8;
}
}
else if (posBalance >= 40) {
if (Rs < 992) {
return 2;
}
else if (Rs >= 992) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (SC < 1068) {
if (Rs < 599) {
return 1;
}
else if (Rs >= 599) {
if (ssDCval < 69) {
if (Rs < 612) {
return 2;
}
else if (Rs >= 612) {
return 1;
}
}
else if (ssDCval >= 69) {
if (Cs < 426) {
return 4;
}
else if (Cs >= 426) {
if (Rs < 620) {
return 1;
}
else if (Rs >= 620) {
return 4;
}
}
}
}
}
else if (SC >= 1068) {
if (MVDiff < 1) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
return 1;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 1273) {
if (RsCsDiff < 2) {
if (ssDCval < 64) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (SC < 1856) {
if (Rs < 1503) {
if (ssDCval < 54) {
return 2;
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (Rs >= 1503) {
if (Rs < 1522) {
return 1;
}
else if (Rs >= 1522) {
return 2;
}
}
}
else if (SC >= 1856) {
if (ssDCval < 62) {
if (Cs < 419) {
if (Rs < 1565) {
return 1;
}
else if (Rs >= 1565) {
return 2;
}
}
else if (Cs >= 419) {
return 8;
}
}
else if (ssDCval >= 62) {
if (refDCval < 64) {
if (mu_mv_mag_sq < 122) {
if (Rs < 1655) {
return 1;
}
else if (Rs >= 1655) {
return 2;
}
}
else if (mu_mv_mag_sq >= 122) {
return 2;
}
}
else if (refDCval >= 64) {
if (Rs < 1403) {
return 8;
}
else if (Rs >= 1403) {
if (Rs < 1475) {
return 2;
}
else if (Rs >= 1475) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (Rs < 1399) {
if (Cs < 300) {
return 1;
}
else if (Cs >= 300) {
if (AFD < 2) {
if (Cs < 457) {
return 1;
}
else if (Cs >= 457) {
return 8;
}
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (Rs >= 1399) {
if (Rs < 1766) {
return 1;
}
else if (Rs >= 1766) {
return 2;
}
}
}
}
else if (RsCsDiff >= 2) {
if (TSCindex < 3) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (TSCindex >= 3) {
if (negBalance < 30) {
if (Rs < 1804) {
if (refDCval < 70) {
if (RsDiff < 3) {
if (Rs < 1551) {
return 1;
}
else if (Rs >= 1551) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 2;
}
}
}
else if (RsDiff >= 3) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (Rs < 1417) {
return 1;
}
else if (Rs >= 1417) {
if (Rs < 1619) {
return 2;
}
else if (Rs >= 1619) {
return 1;
}
}
}
}
}
else if (refDCval >= 70) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (Rs >= 1804) {
if (RsCsDiff < 10) {
return 2;
}
else if (RsCsDiff >= 10) {
if (SC < 2334) {
return 8;
}
else if (SC >= 2334) {
return 4;
}
}
}
}
else if (negBalance >= 30) {
if (posBalance < 59) {
if (mcTcor < 95) {
if (Rs < 1392) {
return 4;
}
else if (Rs >= 1392) {
if (Rs < 1437) {
return 2;
}
else if (Rs >= 1437) {
return 1;
}
}
}
else if (mcTcor >= 95) {
if (Rs < 2199) {
if (refDCval < 74) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
if (Cs < 329) {
return 8;
}
else if (Cs >= 329) {
return 1;
}
}
}
else if (refDCval >= 74) {
if (Rs < 1440) {
return 1;
}
else if (Rs >= 1440) {
return 8;
}
}
}
else if (Rs >= 2199) {
return 4;
}
}
}
else if (posBalance >= 59) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (Cs < 152) {
if (ssDCval < 107) {
if (RsCsDiff < 2) {
if (negBalance < 39) {
if (posBalance < 54) {
if (Cs < 51) {
return 8;
}
else if (Cs >= 51) {
if (negBalance < 18) {
if (Rs < 1237) {
return 8;
}
else if (Rs >= 1237) {
return 1;
}
}
else if (negBalance >= 18) {
if (MVDiff < 38) {
return 1;
}
else if (MVDiff >= 38) {
return 8;
}
}
}
}
else if (posBalance >= 54) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 8;
}
}
}
}
else if (negBalance >= 39) {
if (ssDCval < 87) {
if (Rs < 738) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Cs < 86) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
return 1;
}
}
else if (Cs >= 86) {
return 8;
}
}
}
else if (Rs >= 738) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
return 8;
}
}
}
else if (ssDCval >= 87) {
if (SC < 321) {
if (Rs < 290) {
return 1;
}
else if (Rs >= 290) {
return 8;
}
}
else if (SC >= 321) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
if (tcor < 57) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 288) {
return 8;
}
else if (Rs >= 288) {
if (mu_mv_mag_sq < 21) {
return 8;
}
else if (mu_mv_mag_sq >= 21) {
return 2;
}
}
}
}
else if (tcor >= 57) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (SC < 419) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 2;
}
}
else if (SC >= 419) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 107) {
if (MVDiff < 20) {
if (ssDCval < 133) {
return 8;
}
else if (ssDCval >= 133) {
if (Rs < 822) {
return 2;
}
else if (Rs >= 822) {
return 4;
}
}
}
else if (MVDiff >= 20) {
if (mu_mv_mag_sq < 15) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (mcTcor < 97) {
return 1;
}
else if (mcTcor >= 97) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (Rs < 340) {
if (Rs < 254) {
return 1;
}
else if (Rs >= 254) {
return 2;
}
}
else if (Rs >= 340) {
return 1;
}
}
}
}
}
else if (Cs >= 152) {
if (SC < 855) {
if (mcTcor < 97) {
if (MVDiff < 23) {
if (AFD < 9) {
if (Rs < 334) {
if (TSC < 6) {
if (negBalance < 47) {
if (tcor < 94) {
if (Cs < 255) {
return 2;
}
else if (Cs >= 255) {
return 8;
}
}
else if (tcor >= 94) {
if (Cs < 292) {
if (Cs < 173) {
return 8;
}
else if (Cs >= 173) {
return 4;
}
}
else if (Cs >= 292) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (ssDCval < 87) {
if (SC < 632) {
if (Rs < 254) {
return 8;
}
else if (Rs >= 254) {
return 2;
}
}
else if (SC >= 632) {
return 8;
}
}
else if (ssDCval >= 87) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 2;
}
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (Rs >= 334) {
if (Cs < 390) {
if (ssDCval < 110) {
if (posBalance < 68) {
if (SC < 789) {
if (negBalance < 49) {
if (ssDCval < 85) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (ssDCval >= 85) {
return 1;
}
}
else if (negBalance >= 49) {
return 8;
}
}
else if (SC >= 789) {
return 4;
}
}
else if (posBalance >= 68) {
return 8;
}
}
else if (ssDCval >= 110) {
return 8;
}
}
else if (Cs >= 390) {
return 8;
}
}
}
else if (AFD >= 9) {
if (TSC < 6) {
if (mu_mv_mag_sq < 5) {
return 1;
}
else if (mu_mv_mag_sq >= 5) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (posBalance < 65) {
return 1;
}
else if (posBalance >= 65) {
if (Rs < 334) {
return 8;
}
else if (Rs >= 334) {
return 1;
}
}
}
}
}
else if (TSC >= 6) {
if (RsDiff < 3) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
if (Rs < 344) {
return 1;
}
else if (Rs >= 344) {
return 2;
}
}
}
else if (RsDiff >= 3) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (Cs < 470) {
if (Rs < 447) {
return 8;
}
else if (Rs >= 447) {
return 4;
}
}
else if (Cs >= 470) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 23) {
if (Cs < 179) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (Cs >= 179) {
if (MVDiff < 157) {
if (mcTcor < 93) {
if (SC < 752) {
if (MVDiff < 29) {
if (Rs < 327) {
return 4;
}
else if (Rs >= 327) {
return 8;
}
}
else if (MVDiff >= 29) {
if (tcor < 79) {
if (AFD < 28) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (mu_mv_mag_sq < 56) {
return 2;
}
else if (mu_mv_mag_sq >= 56) {
return 1;
}
}
}
}
else if (AFD >= 28) {
return 1;
}
}
else if (tcor >= 79) {
return 1;
}
}
}
else if (SC >= 752) {
if (TSC < 9) {
if (tcor < 74) {
return 8;
}
else if (tcor >= 74) {
if (Rs < 365) {
return 2;
}
else if (Rs >= 365) {
return 8;
}
}
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 251) {
return 8;
}
else if (Rs >= 251) {
if (diffMVdiffVal < -3) {
if (MVDiff < 34) {
if (diffRsCsdiff < -3) {
return 2;
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
else if (MVDiff >= 34) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
if (SC < 700) {
return 4;
}
else if (SC >= 700) {
if (tcor < 80) {
if (Rs < 368) {
return 2;
}
else if (Rs >= 368) {
return 4;
}
}
else if (tcor >= 80) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 157) {
if (SC < 845) {
return 4;
}
else if (SC >= 845) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 78) {
if (SC < 677) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
if (refDCval < 76) {
return 8;
}
else if (refDCval >= 76) {
return 2;
}
}
}
else if (SC >= 677) {
if (RsDiff < 2) {
if (diffRsCsdiff < -2) {
if (Rs < 475) {
return 8;
}
else if (Rs >= 475) {
return 4;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (ssDCval >= 78) {
if (MVDiff < 2) {
if (Rs < 533) {
if (ssDCval < 91) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
if (Rs < 264) {
return 8;
}
else if (Rs >= 264) {
return 1;
}
}
}
else if (ssDCval >= 91) {
if (Rs < 373) {
if (Cs < 180) {
return 4;
}
else if (Cs >= 180) {
return 2;
}
}
else if (Rs >= 373) {
return 1;
}
}
}
else if (Rs >= 533) {
return 2;
}
}
else if (MVDiff >= 2) {
if (Cs < 183) {
if (Rs < 275) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 268) {
return 8;
}
else if (Rs >= 268) {
return 2;
}
}
}
else if (Rs >= 275) {
return 1;
}
}
else if (Cs >= 183) {
if (tcor < 100) {
if (Rs < 371) {
if (ssDCval < 126) {
if (tcor < 88) {
if (Cs < 294) {
if (Cs < 226) {
return 2;
}
else if (Cs >= 226) {
return 4;
}
}
else if (Cs >= 294) {
return 1;
}
}
else if (tcor >= 88) {
if (TSC < 3) {
if (RsDiff < 1) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (TSC >= 3) {
if (Cs < 504) {
return 2;
}
else if (Cs >= 504) {
return 1;
}
}
}
}
else if (ssDCval >= 126) {
if (tcor < 97) {
if (diffMVdiffVal < 41) {
if (MVDiff < 14) {
if (Rs < 289) {
if (Cs < 321) {
if (Rs < 272) {
return 4;
}
else if (Rs >= 272) {
return 8;
}
}
else if (Cs >= 321) {
return 2;
}
}
else if (Rs >= 289) {
return 8;
}
}
else if (MVDiff >= 14) {
return 2;
}
}
else if (diffMVdiffVal >= 41) {
return 4;
}
}
else if (tcor >= 97) {
return 4;
}
}
}
else if (Rs >= 371) {
if (ssDCval < 93) {
if (MVDiff < 9) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 4;
}
}
else if (MVDiff >= 9) {
if (SC < 721) {
return 2;
}
else if (SC >= 721) {
return 8;
}
}
}
else if (ssDCval >= 93) {
if (posBalance < 61) {
if (ssDCval < 152) {
if (MVDiff < 32) {
if (mu_mv_mag_sq < 45) {
if (SC < 850) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
if (Rs < 418) {
return 1;
}
else if (Rs >= 418) {
return 4;
}
}
}
else if (SC >= 850) {
return 8;
}
}
else if (mu_mv_mag_sq >= 45) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
}
else if (MVDiff >= 32) {
return 2;
}
}
else if (ssDCval >= 152) {
if (Rs < 438) {
return 8;
}
else if (Rs >= 438) {
return 4;
}
}
}
else if (posBalance >= 61) {
if (Cs < 301) {
return 8;
}
else if (Cs >= 301) {
return 4;
}
}
}
}
}
else if (tcor >= 100) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 855) {
if (SC < 1883) {
if (mu_mv_mag_sq < 63) {
if (mcTcor < 93) {
if (mcTcor < 83) {
if (CsDiff < 4) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 2;
}
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (mcTcor >= 83) {
if (ssDCval < 96) {
if (AFD < 8) {
if (Rs < 593) {
return 2;
}
else if (Rs >= 593) {
return 1;
}
}
else if (AFD >= 8) {
if (posBalance < 56) {
if (Rs < 492) {
return 8;
}
else if (Rs >= 492) {
return 1;
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
else if (ssDCval >= 96) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
if (RsCsDiff < 2) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
return 1;
}
}
else if (RsCsDiff >= 2) {
if (Cs < 437) {
if (negBalance < 56) {
if (SC < 1743) {
if (diffMVdiffVal < -2) {
if (ssDCval < 125) {
if (MVDiff < 22) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 1;
}
}
}
else if (MVDiff >= 22) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (Rs < 991) {
return 8;
}
else if (Rs >= 991) {
return 1;
}
}
}
}
else if (ssDCval >= 125) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 166) {
return 4;
}
else if (Cs >= 166) {
if (diffAFD < -2) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
return 1;
}
}
else if (diffAFD >= -2) {
if (gchDC < 3) {
if (SC < 886) {
if (SC < 881) {
return 1;
}
else if (SC >= 881) {
return 8;
}
}
else if (SC >= 886) {
return 1;
}
}
else if (gchDC >= 3) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
}
}
}
}
else if (SC >= 1743) {
return 8;
}
}
else if (negBalance >= 56) {
if (mu_mv_mag_sq < 26) {
if (negBalance < 58) {
if (ssDCval < 107) {
return 1;
}
else if (ssDCval >= 107) {
return 8;
}
}
else if (negBalance >= 58) {
if (SC < 952) {
return 2;
}
else if (SC >= 952) {
if (CsDiff < 3) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (CsDiff >= 3) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
return 8;
}
}
}
else if (Cs >= 437) {
if (posBalance < 52) {
if (mcTcor < 99) {
if (ssDCval < 131) {
if (mu_mv_mag_sq < 2) {
if (posBalance < 46) {
if (Rs < 727) {
return 8;
}
else if (Rs >= 727) {
return 2;
}
}
else if (posBalance >= 46) {
if (tcor < 89) {
return 2;
}
else if (tcor >= 89) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
return 2;
}
}
else if (ssDCval >= 131) {
return 8;
}
}
else if (mcTcor >= 99) {
if (refDCval < 119) {
return 2;
}
else if (refDCval >= 119) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (negBalance < 48) {
if (SC < 894) {
if (ssDCval < 118) {
return 2;
}
else if (ssDCval >= 118) {
if (Rs < 372) {
return 8;
}
else if (Rs >= 372) {
return 4;
}
}
}
else if (SC >= 894) {
if (Rs < 647) {
if (negBalance < 44) {
if (ssDCval < 123) {
if (RsCsDiff < 4) {
return 1;
}
else if (RsCsDiff >= 4) {
if (posBalance < 60) {
return 2;
}
else if (posBalance >= 60) {
return 1;
}
}
}
else if (ssDCval >= 123) {
return 1;
}
}
else if (negBalance >= 44) {
return 2;
}
}
else if (Rs >= 647) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (MVDiff < 4) {
if (diffMVdiffVal < -3) {
if (Rs < 563) {
return 2;
}
else if (Rs >= 563) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (MVDiff >= 4) {
if (Cs < 468) {
if (Cs < 444) {
return 8;
}
else if (Cs >= 444) {
return 1;
}
}
else if (Cs >= 468) {
if (ssDCval < 83) {
return 1;
}
else if (ssDCval >= 83) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (negBalance < 32) {
if (refDCval < 94) {
if (Rs < 1007) {
if (refDCval < 78) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (Rs < 719) {
return 4;
}
else if (Rs >= 719) {
return 8;
}
}
}
else if (refDCval >= 78) {
if (MVDiff < 4) {
if (posBalance < 24) {
return 8;
}
else if (posBalance >= 24) {
return 1;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (Rs >= 1007) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (refDCval >= 94) {
if (MVDiff < 2) {
if (Cs < 452) {
if (refDCval < 98) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (refDCval >= 98) {
return 8;
}
}
else if (Cs >= 452) {
return 8;
}
}
else if (MVDiff >= 2) {
if (tcor < 97) {
if (ssDCval < 158) {
return 8;
}
else if (ssDCval >= 158) {
return 4;
}
}
else if (tcor >= 97) {
if (posBalance < 53) {
if (negBalance < 18) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (negBalance >= 18) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
return 2;
}
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
}
}
else if (negBalance >= 32) {
if (refDCval < 81) {
if (SC < 913) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 479) {
return 8;
}
else if (Rs >= 479) {
if (Rs < 613) {
return 2;
}
else if (Rs >= 613) {
return 4;
}
}
}
}
else if (SC >= 913) {
return 8;
}
}
else if (refDCval >= 81) {
if (MVDiff < 3) {
if (refDCval < 85) {
if (Rs < 474) {
return 8;
}
else if (Rs >= 474) {
if (Rs < 771) {
return 1;
}
else if (Rs >= 771) {
return 8;
}
}
}
else if (refDCval >= 85) {
if (SC < 1062) {
if (Rs < 569) {
if (Rs < 452) {
return 8;
}
else if (Rs >= 452) {
if (Cs < 500) {
return 1;
}
else if (Cs >= 500) {
if (Rs < 498) {
return 8;
}
else if (Rs >= 498) {
return 2;
}
}
}
}
else if (Rs >= 569) {
return 4;
}
}
else if (SC >= 1062) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 109) {
return 8;
}
else if (ssDCval >= 109) {
if (AFD < 4) {
if (Rs < 417) {
return 4;
}
else if (Rs >= 417) {
return 2;
}
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 1883) {
if (Cs < 305) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
if (Rs < 1969) {
return 8;
}
else if (Rs >= 1969) {
return 4;
}
}
}
else if (Cs >= 305) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 85) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (AFD < 6) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 80) {
return 8;
}
else if (mu_mv_mag_sq >= 80) {
if (Rs < 1817) {
return 1;
}
else if (Rs >= 1817) {
return 8;
}
}
}
}
else if (AFD >= 6) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (refDCval < 80) {
return 1;
}
else if (refDCval >= 80) {
if (Rs < 1512) {
return 1;
}
else if (Rs >= 1512) {
if (Rs < 1998) {
if (Cs < 328) {
return 8;
}
else if (Cs >= 328) {
if (MVDiff < 4) {
if (Cs < 524) {
if (refDCval < 81) {
return 1;
}
else if (refDCval >= 81) {
return 4;
}
}
else if (Cs >= 524) {
return 1;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
}
else if (Rs >= 1998) {
return 8;
}
}
}
}
}
else if (posBalance >= 55) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (mu_mv_mag_sq < 64) {
if (tcor < 76) {
if (negBalance < 47) {
if (CsDiff < 3) {
if (MVDiff < 13) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
else if (MVDiff >= 13) {
if (mu_mv_mag_sq < 31) {
if (Rs < 2297) {
if (Cs < 272) {
return 1;
}
else if (Cs >= 272) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
if (Rs < 1396) {
return 4;
}
else if (Rs >= 1396) {
return 2;
}
}
}
}
else if (Rs >= 2297) {
return 4;
}
}
else if (mu_mv_mag_sq >= 31) {
if (mcTcor < 95) {
if (Rs < 628) {
if (gchDC < 4) {
return 1;
}
else if (gchDC >= 4) {
return 2;
}
}
else if (Rs >= 628) {
return 1;
}
}
else if (mcTcor >= 95) {
if (RsDiff < 8) {
if (Cs < 410) {
return 1;
}
else if (Cs >= 410) {
return 2;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 28) {
if (mu_mv_mag_sq < 18) {
if (Rs < 757) {
if (diffMVdiffVal < 20) {
if (Cs < 469) {
return 4;
}
else if (Cs >= 469) {
return 2;
}
}
else if (diffMVdiffVal >= 20) {
if (AFD < 28) {
if (Rs < 464) {
return 2;
}
else if (Rs >= 464) {
return 8;
}
}
else if (AFD >= 28) {
return 1;
}
}
}
else if (Rs >= 757) {
if (mcTcor < 93) {
if (SC < 1431) {
return 2;
}
else if (SC >= 1431) {
if (diffMVdiffVal < -4) {
if (Rs < 1439) {
return 8;
}
else if (Rs >= 1439) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
if (ssDCval < 132) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 1;
}
}
else if (ssDCval >= 132) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (SC < 2764) {
if (MVDiff < 75) {
return 1;
}
else if (MVDiff >= 75) {
if (posBalance < 69) {
return 1;
}
else if (posBalance >= 69) {
if (Rs < 1817) {
return 8;
}
else if (Rs >= 1817) {
return 1;
}
}
}
}
else if (SC >= 2764) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 28) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
if (posBalance < 51) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
return 2;
}
}
else if (posBalance >= 51) {
if (tcor < 64) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
if (Cs < 160) {
return 1;
}
else if (Cs >= 160) {
return 2;
}
}
}
else if (tcor >= 64) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 47) {
if (AFD < 23) {
if (Cs < 332) {
if (MVDiff < 13) {
if (Rs < 960) {
return 1;
}
else if (Rs >= 960) {
return 2;
}
}
else if (MVDiff >= 13) {
if (ssDCval < 46) {
if (SC < 775) {
if (Rs < 439) {
return 1;
}
else if (Rs >= 439) {
return 8;
}
}
else if (SC >= 775) {
return 1;
}
}
else if (ssDCval >= 46) {
if (posBalance < 49) {
if (MVDiff < 35) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (MVDiff >= 35) {
return 1;
}
}
else if (posBalance >= 49) {
if (RsDiff < 12) {
if (diffMVdiffVal < -19) {
if (mcTcor < 85) {
return 1;
}
else if (mcTcor >= 85) {
return 2;
}
}
else if (diffMVdiffVal >= -19) {
if (diffAFD < 3) {
if (diffAFD < 2) {
if (mcTcor < 87) {
if (Cs < 283) {
if (ssDCval < 125) {
if (Rs < 424) {
return 4;
}
else if (Rs >= 424) {
return 8;
}
}
else if (ssDCval >= 125) {
return 1;
}
}
else if (Cs >= 283) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
if (Rs < 996) {
return 1;
}
else if (Rs >= 996) {
return 2;
}
}
}
}
else if (mcTcor >= 87) {
if (posBalance < 57) {
if (tcor < 74) {
return 1;
}
else if (tcor >= 74) {
if (Rs < 1169) {
return 4;
}
else if (Rs >= 1169) {
return 2;
}
}
}
else if (posBalance >= 57) {
if (ssDCval < 97) {
if (Rs < 1372) {
if (MVDiff < 39) {
if (Rs < 892) {
return 8;
}
else if (Rs >= 892) {
return 2;
}
}
else if (MVDiff >= 39) {
return 2;
}
}
else if (Rs >= 1372) {
if (MVDiff < 59) {
return 1;
}
else if (MVDiff >= 59) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (Rs < 1685) {
return 1;
}
else if (Rs >= 1685) {
return 8;
}
}
}
}
}
else if (ssDCval >= 97) {
if (ssDCval < 126) {
return 1;
}
else if (ssDCval >= 126) {
if (Rs < 956) {
return 4;
}
else if (Rs >= 956) {
return 2;
}
}
}
}
}
}
else if (diffAFD >= 2) {
if (AFD < 21) {
if (Cs < 316) {
return 4;
}
else if (Cs >= 316) {
return 2;
}
}
else if (AFD >= 21) {
return 8;
}
}
}
else if (diffAFD >= 3) {
return 1;
}
}
}
else if (RsDiff >= 12) {
if (MVDiff < 87) {
return 2;
}
else if (MVDiff >= 87) {
return 4;
}
}
}
}
}
}
else if (Cs >= 332) {
if (MVDiff < 19) {
if (SC < 1066) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (SC >= 1066) {
if (posBalance < 54) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (ssDCval < 82) {
return 2;
}
else if (ssDCval >= 82) {
return 8;
}
}
}
else if (posBalance >= 54) {
if (diffMVdiffVal < -7) {
if (Rs < 1638) {
return 2;
}
else if (Rs >= 1638) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
if (Rs < 1887) {
return 8;
}
else if (Rs >= 1887) {
return 2;
}
}
}
}
}
else if (MVDiff >= 19) {
if (tcor < 40) {
if (ssDCval < 123) {
if (AFD < 19) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
else if (AFD >= 19) {
if (MVDiff < 47) {
return 8;
}
else if (MVDiff >= 47) {
if (Rs < 844) {
return 1;
}
else if (Rs >= 844) {
return 4;
}
}
}
}
else if (ssDCval >= 123) {
return 1;
}
}
else if (tcor >= 40) {
if (ssDCval < 124) {
if (diffTSC < -2) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
if (Cs < 478) {
return 4;
}
else if (Cs >= 478) {
return 8;
}
}
}
else if (diffTSC >= -2) {
if (posBalance < 53) {
if (Rs < 1256) {
if (mcTcor < 91) {
if (ssDCval < 76) {
if (Cs < 508) {
return 8;
}
else if (Cs >= 508) {
return 2;
}
}
else if (ssDCval >= 76) {
return 1;
}
}
else if (mcTcor >= 91) {
if (Rs < 747) {
return 2;
}
else if (Rs >= 747) {
return 1;
}
}
}
else if (Rs >= 1256) {
return 2;
}
}
else if (posBalance >= 53) {
if (SC < 1315) {
if (TSC < 8) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (posBalance < 62) {
if (diffMVdiffVal < 2) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (Rs < 517) {
return 8;
}
else if (Rs >= 517) {
return 2;
}
}
}
else if (diffMVdiffVal >= 2) {
return 2;
}
}
else if (posBalance >= 62) {
return 8;
}
}
}
else if (TSC >= 8) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (Rs < 419) {
return 8;
}
else if (Rs >= 419) {
return 4;
}
}
}
}
else if (SC >= 1315) {
if (posBalance < 64) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (mcTcor < 86) {
if (Rs < 941) {
return 8;
}
else if (Rs >= 941) {
return 4;
}
}
else if (mcTcor >= 86) {
return 8;
}
}
}
else if (posBalance >= 64) {
if (Rs < 1525) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 8;
}
}
else if (Rs >= 1525) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 124) {
if (Rs < 1356) {
if (Rs < 495) {
return 8;
}
else if (Rs >= 495) {
return 2;
}
}
else if (Rs >= 1356) {
return 4;
}
}
}
}
}
}
else if (AFD >= 23) {
if (ssDCval < 112) {
if (gchDC < 6) {
if (refDCval < 73) {
if (posBalance < 63) {
if (Cs < 177) {
return 1;
}
else if (Cs >= 177) {
return 4;
}
}
else if (posBalance >= 63) {
if (AFD < 29) {
if (Rs < 910) {
return 8;
}
else if (Rs >= 910) {
return 1;
}
}
else if (AFD >= 29) {
return 2;
}
}
}
else if (refDCval >= 73) {
if (mu_mv_mag_sq < 43) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (RsDiff < 8) {
if (Rs < 657) {
if (tcor < 57) {
return 1;
}
else if (tcor >= 57) {
if (Cs < 394) {
return 2;
}
else if (Cs >= 394) {
return 1;
}
}
}
else if (Rs >= 657) {
if (ssDCval < 109) {
return 1;
}
else if (ssDCval >= 109) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (posBalance < 61) {
return 1;
}
else if (posBalance >= 61) {
if (Rs < 819) {
if (SC < 1237) {
return 1;
}
else if (SC >= 1237) {
return 2;
}
}
else if (Rs >= 819) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (Cs < 397) {
return 2;
}
else if (Cs >= 397) {
if (Rs < 736) {
return 8;
}
else if (Rs >= 736) {
return 4;
}
}
}
}
}
else if (gchDC >= 6) {
if (gchDC < 9) {
return 1;
}
else if (gchDC >= 9) {
if (Cs < 353) {
if (diffRsCsdiff < 21) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
else if (diffRsCsdiff >= 21) {
return 2;
}
}
else if (Cs >= 353) {
if (gchDC < 20) {
return 8;
}
else if (gchDC >= 20) {
if (Rs < 811) {
return 2;
}
else if (Rs >= 811) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (RsCsDiff < 151) {
if (Rs < 1788) {
if (Rs < 1023) {
if (SC < 1259) {
return 1;
}
else if (SC >= 1259) {
return 8;
}
}
else if (Rs >= 1023) {
return 1;
}
}
else if (Rs >= 1788) {
return 8;
}
}
else if (RsCsDiff >= 151) {
if (posBalance < 66) {
return 1;
}
else if (posBalance >= 66) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 76) {
if (RsCsDiff < 73) {
if (Cs < 134) {
if (ssDCval < 143) {
if (SC < 341) {
return 8;
}
else if (SC >= 341) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (refDCval < 79) {
return 1;
}
else if (refDCval >= 79) {
if (SC < 804) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
else if (SC >= 804) {
if (TSC < 5) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
return 2;
}
}
else if (TSC >= 5) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 143) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (Cs >= 134) {
if (Rs < 853) {
if (CsDiff < 2) {
if (MVDiff < 71) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 55) {
return 1;
}
else if (mu_mv_mag_sq >= 55) {
return 8;
}
}
else if (mcTcor >= 99) {
if (Rs < 788) {
return 4;
}
else if (Rs >= 788) {
return 2;
}
}
}
else if (MVDiff >= 71) {
if (gchDC < 2) {
if (Rs < 764) {
return 1;
}
else if (Rs >= 764) {
return 4;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (ssDCval < 106) {
if (negBalance < 45) {
if (SC < 1145) {
if (SC < 737) {
if (Cs < 251) {
return 1;
}
else if (Cs >= 251) {
return 2;
}
}
else if (SC >= 737) {
if (MVDiff < 33) {
return 1;
}
else if (MVDiff >= 33) {
return 4;
}
}
}
else if (SC >= 1145) {
return 2;
}
}
else if (negBalance >= 45) {
if (mu_mv_mag_sq < 20) {
if (mcTcor < 94) {
if (posBalance < 54) {
if (Rs < 742) {
if (Rs < 676) {
return 1;
}
else if (Rs >= 676) {
return 8;
}
}
else if (Rs >= 742) {
return 2;
}
}
else if (posBalance >= 54) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 635) {
if (ssDCval < 97) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Cs < 479) {
return 1;
}
else if (Cs >= 479) {
return 2;
}
}
}
else if (ssDCval >= 97) {
return 4;
}
}
else if (Rs >= 635) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 20) {
if (Rs < 373) {
return 1;
}
else if (Rs >= 373) {
return 8;
}
}
}
}
else if (ssDCval >= 106) {
if (posBalance < 44) {
if (Cs < 289) {
return 1;
}
else if (Cs >= 289) {
return 2;
}
}
else if (posBalance >= 44) {
if (MVDiff < 44) {
if (Rs < 561) {
return 4;
}
else if (Rs >= 561) {
return 8;
}
}
else if (MVDiff >= 44) {
if (Rs < 646) {
if (posBalance < 72) {
return 8;
}
else if (posBalance >= 72) {
return 2;
}
}
else if (Rs >= 646) {
return 1;
}
}
}
}
}
}
else if (Rs >= 853) {
if (ssDCval < 138) {
if (SC < 2268) {
if (Cs < 453) {
if (MVDiff < 18) {
if (ssDCval < 86) {
if (RsDiff < 6) {
if (tcor < 83) {
if (mu_mv_mag_sq < 38) {
if (Cs < 409) {
return 1;
}
else if (Cs >= 409) {
if (Cs < 431) {
return 2;
}
else if (Cs >= 431) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 38) {
return 4;
}
}
else if (tcor >= 83) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 1;
}
}
}
}
else if (RsDiff >= 6) {
if (mu_mv_mag_sq < 32) {
if (diffRsCsdiff < -29) {
return 2;
}
else if (diffRsCsdiff >= -29) {
return 1;
}
}
else if (mu_mv_mag_sq >= 32) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
if (Cs < 439) {
return 2;
}
else if (Cs >= 439) {
return 1;
}
}
}
}
}
else if (ssDCval >= 86) {
if (Cs < 272) {
if (Rs < 1365) {
return 1;
}
else if (Rs >= 1365) {
if (Rs < 1702) {
return 2;
}
else if (Rs >= 1702) {
if (Cs < 155) {
return 2;
}
else if (Cs >= 155) {
return 1;
}
}
}
}
else if (Cs >= 272) {
if (RsDiff < 5) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 1;
}
}
else if (RsDiff >= 5) {
return 1;
}
}
}
}
else if (MVDiff >= 18) {
if (diffAFD < -22) {
return 8;
}
else if (diffAFD >= -22) {
if (Rs < 1889) {
if (diffRsCsdiff < -13) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (tcor < 89) {
if (Rs < 1265) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
else if (Rs >= 1265) {
return 1;
}
}
else if (tcor >= 89) {
return 1;
}
}
}
else if (diffRsCsdiff >= -13) {
return 1;
}
}
else if (Rs >= 1889) {
if (Rs < 1900) {
return 2;
}
else if (Rs >= 1900) {
return 1;
}
}
}
}
}
else if (Cs >= 453) {
if (ssDCval < 123) {
if (mu_mv_mag_sq < 25) {
if (mcTcor < 93) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 1604) {
return 4;
}
else if (Rs >= 1604) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (Rs < 1117) {
return 1;
}
else if (Rs >= 1117) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (mu_mv_mag_sq < 40) {
if (ssDCval < 67) {
return 2;
}
else if (ssDCval >= 67) {
if (Cs < 492) {
return 1;
}
else if (Cs >= 492) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (Cs < 476) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (Cs >= 476) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 123) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
}
}
else if (SC >= 2268) {
if (refDCval < 69) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
if (Rs < 2058) {
return 2;
}
else if (Rs >= 2058) {
return 1;
}
}
}
else if (refDCval >= 69) {
if (tcor < 96) {
if (refDCval < 98) {
if (diffRsCsdiff < -53) {
if (Rs < 2114) {
return 1;
}
else if (Rs >= 2114) {
return 4;
}
}
else if (diffRsCsdiff >= -53) {
return 1;
}
}
else if (refDCval >= 98) {
if (ssDCval < 113) {
if (Rs < 2354) {
if (Cs < 227) {
return 1;
}
else if (Cs >= 227) {
return 8;
}
}
else if (Rs >= 2354) {
return 1;
}
}
else if (ssDCval >= 113) {
if (Rs < 1775) {
return 8;
}
else if (Rs >= 1775) {
return 1;
}
}
}
}
else if (tcor >= 96) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (mu_mv_mag_sq < 22) {
if (SC < 2735) {
return 2;
}
else if (SC >= 2735) {
return 1;
}
}
else if (mu_mv_mag_sq >= 22) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 138) {
if (gchDC < 2) {
if (SC < 1738) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 1;
}
}
else if (SC >= 1738) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
return 1;
}
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 73) {
if (ssDCval < 112) {
if (Rs < 2023) {
if (refDCval < 68) {
if (Rs < 1537) {
return 1;
}
else if (Rs >= 1537) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
}
else if (refDCval >= 68) {
if (Cs < 271) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (posBalance < 48) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 2;
}
}
else if (posBalance >= 48) {
if (Cs < 241) {
if (diffMVdiffVal < -1) {
if (Rs < 1632) {
return 2;
}
else if (Rs >= 1632) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (Cs >= 241) {
return 2;
}
}
}
}
else if (Cs >= 271) {
if (tcor < 86) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Cs < 349) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
return 1;
}
}
else if (Cs >= 349) {
if (mu_mv_mag_sq < 12) {
if (RsCsDiff < 128) {
if (tcor < 83) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 1;
}
}
else if (tcor >= 83) {
return 8;
}
}
else if (RsCsDiff >= 128) {
return 2;
}
}
else if (mu_mv_mag_sq >= 12) {
if (mcTcor < 95) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 2;
}
}
else if (mcTcor >= 95) {
if (Cs < 366) {
return 2;
}
else if (Cs >= 366) {
return 1;
}
}
}
}
}
}
else if (tcor >= 86) {
return 1;
}
}
}
}
else if (Rs >= 2023) {
if (ssDCval < 83) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
if (RsDiff < 11) {
if (Cs < 232) {
return 2;
}
else if (Cs >= 232) {
if (RsDiff < 10) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (RsDiff >= 10) {
return 4;
}
}
}
else if (RsDiff >= 11) {
if (RsDiff < 18) {
return 2;
}
else if (RsDiff >= 18) {
if (Rs < 2289) {
return 2;
}
else if (Rs >= 2289) {
return 1;
}
}
}
}
}
else if (ssDCval >= 83) {
if (mcTcor < 95) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 1;
}
}
else if (mcTcor >= 95) {
if (diffAFD < -2) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 1;
}
}
else if (diffAFD >= -2) {
if (gchDC < 4) {
if (ssDCval < 92) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 2;
}
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (gchDC >= 4) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (TSC < 6) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (Rs < 1438) {
return 4;
}
else if (Rs >= 1438) {
return 1;
}
}
}
else if (TSC >= 6) {
if (mcTcor < 94) {
if (SC < 1388) {
return 4;
}
else if (SC >= 1388) {
if (MVDiff < 93) {
return 1;
}
else if (MVDiff >= 93) {
if (MVDiff < 118) {
return 8;
}
else if (MVDiff >= 118) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 908) {
if (RsDiff < 10) {
if (diffAFD < 4) {
return 2;
}
else if (diffAFD >= 4) {
return 1;
}
}
else if (RsDiff >= 10) {
if (diffTSC < -3) {
return 8;
}
else if (diffTSC >= -3) {
return 1;
}
}
}
else if (Rs >= 908) {
if (Rs < 2148) {
if (TSC < 7) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (Rs >= 2148) {
if (MVDiff < 60) {
return 1;
}
else if (MVDiff >= 60) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (AFD < 24) {
if (Rs < 2188) {
if (Rs < 2166) {
return 8;
}
else if (Rs >= 2166) {
return 1;
}
}
else if (Rs >= 2188) {
return 8;
}
}
else if (AFD >= 24) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (Rs < 1517) {
if (AFD < 18) {
if (negBalance < 47) {
if (CsDiff < 3) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (mcTcor < 92) {
if (refDCval < 43) {
return 1;
}
else if (refDCval >= 43) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 892) {
return 1;
}
else if (Rs >= 892) {
return 4;
}
}
}
}
else if (mcTcor >= 92) {
return 1;
}
}
}
else if (CsDiff >= 3) {
if (negBalance < 28) {
if (Rs < 1110) {
if (Rs < 692) {
return 8;
}
else if (Rs >= 692) {
return 2;
}
}
else if (Rs >= 1110) {
return 1;
}
}
else if (negBalance >= 28) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (RsCsDiff < 68) {
if (AFD < 10) {
if (Cs < 439) {
if (Cs < 358) {
return 1;
}
else if (Cs >= 358) {
return 4;
}
}
else if (Cs >= 439) {
return 8;
}
}
else if (AFD >= 10) {
return 4;
}
}
else if (RsCsDiff >= 68) {
if (Rs < 793) {
return 8;
}
else if (Rs >= 793) {
if (Rs < 1255) {
return 2;
}
else if (Rs >= 1255) {
return 1;
}
}
}
}
}
else if (AFD >= 18) {
if (tcor < 53) {
if (ssDCval < 92) {
if (CsDiff < 3) {
if (Rs < 496) {
return 8;
}
else if (Rs >= 496) {
return 2;
}
}
else if (CsDiff >= 3) {
if (AFD < 22) {
if (Rs < 914) {
return 4;
}
else if (Rs >= 914) {
return 1;
}
}
else if (AFD >= 22) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 562) {
return 8;
}
else if (Rs >= 562) {
return 1;
}
}
}
}
}
else if (ssDCval >= 92) {
if (mu_mv_mag_sq < 75) {
if (Rs < 842) {
return 2;
}
else if (Rs >= 842) {
return 4;
}
}
else if (mu_mv_mag_sq >= 75) {
return 1;
}
}
}
else if (tcor >= 53) {
if (diffMVdiffVal < 59) {
if (AFD < 23) {
if (Rs < 475) {
return 1;
}
else if (Rs >= 475) {
if (diffRsCsdiff < 30) {
return 2;
}
else if (diffRsCsdiff >= 30) {
return 8;
}
}
}
else if (AFD >= 23) {
if (Cs < 451) {
return 2;
}
else if (Cs >= 451) {
if (Cs < 467) {
return 4;
}
else if (Cs >= 467) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 59) {
return 1;
}
}
}
}
else if (Rs >= 1517) {
if (ssDCval < 114) {
if (posBalance < 51) {
if (negBalance < 60) {
if (Rs < 2098) {
return 1;
}
else if (Rs >= 2098) {
return 8;
}
}
else if (negBalance >= 60) {
return 2;
}
}
else if (posBalance >= 51) {
if (Cs < 521) {
return 2;
}
else if (Cs >= 521) {
return 8;
}
}
}
else if (ssDCval >= 114) {
if (SC < 2702) {
if (RsDiff < 14) {
if (Rs < 1656) {
if (Rs < 1555) {
return 4;
}
else if (Rs >= 1555) {
if (Rs < 1595) {
return 1;
}
else if (Rs >= 1595) {
return 8;
}
}
}
else if (Rs >= 1656) {
return 1;
}
}
else if (RsDiff >= 14) {
if (AFD < 38) {
return 2;
}
else if (AFD >= 38) {
return 1;
}
}
}
else if (SC >= 2702) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Cs >= 527) {
if (negBalance < 13) {
if (Rs < 701) {
if (ssDCval < 56) {
if (ssDCval < 30) {
return 8;
}
else if (ssDCval >= 30) {
if (mu_mv_mag_sq < 127) {
if (Cs < 925) {
return 1;
}
else if (Cs >= 925) {
if (Cs < 1042) {
return 2;
}
else if (Cs >= 1042) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 380) {
if (Rs < 258) {
return 2;
}
else if (Rs >= 258) {
return 1;
}
}
else if (Rs >= 380) {
return 8;
}
}
}
}
else if (ssDCval >= 56) {
if (Rs < 658) {
if (Rs < 509) {
if (Cs < 613) {
if (mu_mv_mag_sq < 48) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
if (Rs < 413) {
return 8;
}
else if (Rs >= 413) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (Rs < 335) {
return 4;
}
else if (Rs >= 335) {
return 1;
}
}
}
else if (Cs >= 613) {
if (refDCval < 68) {
return 4;
}
else if (refDCval >= 68) {
return 8;
}
}
}
else if (Rs >= 509) {
if (RsDiff < 4) {
if (ssDCval < 89) {
if (Rs < 610) {
return 8;
}
else if (Rs >= 610) {
return 2;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (RsDiff >= 4) {
return 2;
}
}
}
else if (Rs >= 658) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (diffTSC < 1) {
if (ssDCval < 117) {
return 1;
}
else if (ssDCval >= 117) {
if (Cs < 628) {
return 1;
}
else if (Cs >= 628) {
return 8;
}
}
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
}
else if (Rs >= 701) {
if (diffAFD < -3) {
if (SC < 2455) {
if (tcor < 99) {
if (diffAFD < -19) {
if (SC < 2310) {
return 4;
}
else if (SC >= 2310) {
if (Rs < 1588) {
return 1;
}
else if (Rs >= 1588) {
return 8;
}
}
}
else if (diffAFD >= -19) {
if (diffRsCsdiff < 4) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
if (SC < 1946) {
return 8;
}
else if (SC >= 1946) {
return 1;
}
}
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
}
else if (tcor >= 99) {
if (Cs < 628) {
if (Rs < 1184) {
if (Cs < 578) {
return 2;
}
else if (Cs >= 578) {
return 8;
}
}
else if (Rs >= 1184) {
return 1;
}
}
else if (Cs >= 628) {
if (diffRsCsdiff < -57) {
if (SC < 2331) {
if (diffMVdiffVal < 11) {
return 8;
}
else if (diffMVdiffVal >= 11) {
return 4;
}
}
else if (SC >= 2331) {
if (Rs < 1361) {
return 1;
}
else if (Rs >= 1361) {
return 4;
}
}
}
else if (diffRsCsdiff >= -57) {
if (ssDCval < 86) {
if (Rs < 1228) {
return 8;
}
else if (Rs >= 1228) {
return 1;
}
}
else if (ssDCval >= 86) {
if (diffTSC < -6) {
if (Rs < 1402) {
return 1;
}
else if (Rs >= 1402) {
return 8;
}
}
else if (diffTSC >= -6) {
if (Cs < 1005) {
return 4;
}
else if (Cs >= 1005) {
if (ssDCval < 139) {
return 4;
}
else if (ssDCval >= 139) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 2455) {
if (MVDiff < 23) {
if (diffRsCsdiff < -169) {
if (Rs < 2168) {
return 1;
}
else if (Rs >= 2168) {
return 8;
}
}
else if (diffRsCsdiff >= -169) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
return 8;
}
}
}
}
else if (MVDiff >= 23) {
if (MVDiff < 85) {
if (Rs < 1461) {
if (Cs < 1073) {
return 4;
}
else if (Cs >= 1073) {
return 8;
}
}
else if (Rs >= 1461) {
return 4;
}
}
else if (MVDiff >= 85) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 1;
}
}
}
}
}
else if (diffAFD >= -3) {
if (Rs < 1841) {
if (refDCval < 153) {
if (MVDiff < 2) {
if (diffMVdiffVal < -3) {
if (posBalance < 1) {
if (Rs < 1112) {
return 8;
}
else if (Rs >= 1112) {
return 1;
}
}
else if (posBalance >= 1) {
if (ssDCval < 137) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (ssDCval >= 137) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -3) {
if (Cs < 1144) {
if (Rs < 1770) {
if (negBalance < 11) {
if (tcor < 101) {
if (diffMVdiffVal < -2) {
if (posBalance < 29) {
return 8;
}
else if (posBalance >= 29) {
if (Rs < 1072) {
return 8;
}
else if (Rs >= 1072) {
return 1;
}
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (tcor >= 101) {
if (Cs < 631) {
return 4;
}
else if (Cs >= 631) {
return 8;
}
}
}
else if (negBalance >= 11) {
if (Cs < 1044) {
if (Rs < 709) {
return 1;
}
else if (Rs >= 709) {
if (Rs < 1367) {
if (SC < 1464) {
if (Cs < 692) {
if (ssDCval < 81) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
return 2;
}
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (Cs >= 692) {
return 1;
}
}
else if (SC >= 1464) {
return 8;
}
}
else if (Rs >= 1367) {
if (Rs < 1462) {
return 1;
}
else if (Rs >= 1462) {
return 8;
}
}
}
}
else if (Cs >= 1044) {
if (Rs < 993) {
return 8;
}
else if (Rs >= 993) {
return 1;
}
}
}
}
else if (Rs >= 1770) {
if (posBalance < 22) {
return 8;
}
else if (posBalance >= 22) {
if (Rs < 1826) {
return 1;
}
else if (Rs >= 1826) {
return 4;
}
}
}
}
else if (Cs >= 1144) {
if (SC < 2519) {
if (mu_mv_mag_sq < 128) {
if (Rs < 950) {
return 2;
}
else if (Rs >= 950) {
return 1;
}
}
else if (mu_mv_mag_sq >= 128) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
return 1;
}
}
}
else if (SC >= 2519) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (Rs < 796) {
if (tcor < 98) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (tcor >= 98) {
if (ssDCval < 121) {
if (Rs < 709) {
return 1;
}
else if (Rs >= 709) {
return 2;
}
}
else if (ssDCval >= 121) {
return 8;
}
}
}
else if (Rs >= 796) {
if (MVDiff < 3) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (Cs < 719) {
return 1;
}
else if (Cs >= 719) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (refDCval < 43) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (Cs < 1227) {
return 4;
}
else if (Cs >= 1227) {
return 8;
}
}
}
else if (refDCval >= 43) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (Rs < 1049) {
if (diffMVdiffVal < -194) {
return 1;
}
else if (diffMVdiffVal >= -194) {
return 8;
}
}
else if (Rs >= 1049) {
if (SC < 2294) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (diffTSC < 1) {
if (Rs < 1399) {
return 1;
}
else if (Rs >= 1399) {
return 8;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
}
else if (SC >= 2294) {
if (Rs < 1783) {
return 8;
}
else if (Rs >= 1783) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 153) {
if (Rs < 836) {
return 8;
}
else if (Rs >= 836) {
return 2;
}
}
}
else if (Rs >= 1841) {
if (Rs < 2053) {
if (Cs < 1001) {
if (ssDCval < 81) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (Cs < 654) {
return 2;
}
else if (Cs >= 654) {
if (SC < 3004) {
return 1;
}
else if (SC >= 3004) {
return 4;
}
}
}
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (Cs >= 1001) {
if (refDCval < 69) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 4;
}
}
else if (refDCval >= 69) {
if (Rs < 1907) {
return 1;
}
else if (Rs >= 1907) {
return 8;
}
}
}
}
else if (Rs >= 2053) {
if (posBalance < 32) {
if (negBalance < 11) {
if (mu_mv_mag_sq < 94) {
return 4;
}
else if (mu_mv_mag_sq >= 94) {
return 8;
}
}
else if (negBalance >= 11) {
if (Rs < 2287) {
return 8;
}
else if (Rs >= 2287) {
if (SC < 3621) {
return 2;
}
else if (SC >= 3621) {
return 8;
}
}
}
}
else if (posBalance >= 32) {
if (negBalance < 11) {
if (Rs < 2167) {
return 1;
}
else if (Rs >= 2167) {
return 8;
}
}
else if (negBalance >= 11) {
if (Cs < 1324) {
return 2;
}
else if (Cs >= 1324) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 12) {
if (posBalance < 41) {
if (refDCval < 48) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (negBalance < 31) {
if (Cs < 535) {
return 2;
}
else if (Cs >= 535) {
if (RsCsDiff < 9) {
if (Cs < 840) {
if (negBalance < 20) {
if (Rs < 797) {
return 2;
}
else if (Rs >= 797) {
return 1;
}
}
else if (negBalance >= 20) {
return 4;
}
}
else if (Cs >= 840) {
if (Rs < 1589) {
if (Rs < 584) {
return 1;
}
else if (Rs >= 584) {
return 8;
}
}
else if (Rs >= 1589) {
return 4;
}
}
}
else if (RsCsDiff >= 9) {
if (diffAFD < 3) {
if (diffRsCsdiff < 10) {
if (ssDCval < 46) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (ssDCval >= 46) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 4;
}
}
}
else if (diffRsCsdiff >= 10) {
if (RsDiff < 7) {
if (posBalance < 31) {
return 1;
}
else if (posBalance >= 31) {
return 4;
}
}
else if (RsDiff >= 7) {
return 1;
}
}
}
else if (diffAFD >= 3) {
return 1;
}
}
}
}
else if (negBalance >= 31) {
if (Rs < 1467) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 1142) {
return 8;
}
else if (Rs >= 1142) {
return 4;
}
}
}
else if (Rs >= 1467) {
if (Cs < 1243) {
return 4;
}
else if (Cs >= 1243) {
return 1;
}
}
}
}
}
else if (refDCval >= 48) {
if (mu_mv_mag_sq < 73) {
if (ssDCval < 86) {
if (negBalance < 63) {
if (Cs < 784) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (SC < 2744) {
if (ssDCval < 83) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (Rs < 1401) {
return 2;
}
else if (Rs >= 1401) {
return 1;
}
}
}
else if (ssDCval >= 83) {
return 2;
}
}
else if (SC >= 2744) {
if (Rs < 2080) {
return 2;
}
else if (Rs >= 2080) {
return 1;
}
}
}
}
else if (Cs >= 784) {
if (tcor < 89) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
if (Rs < 1723) {
return 8;
}
else if (Rs >= 1723) {
return 1;
}
}
}
else if (tcor >= 89) {
if (Cs < 974) {
if (Rs < 892) {
return 2;
}
else if (Rs >= 892) {
return 1;
}
}
else if (Cs >= 974) {
if (Cs < 1005) {
return 8;
}
else if (Cs >= 1005) {
return 2;
}
}
}
}
}
else if (negBalance >= 63) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 151) {
if (CsDiff < 4) {
if (tcor < 91) {
if (AFD < 7) {
if (Rs < 1306) {
return 8;
}
else if (Rs >= 1306) {
return 1;
}
}
else if (AFD >= 7) {
if (Cs < 1109) {
return 2;
}
else if (Cs >= 1109) {
return 8;
}
}
}
else if (tcor >= 91) {
return 8;
}
}
else if (CsDiff >= 4) {
if (Cs < 1180) {
return 8;
}
else if (Cs >= 1180) {
return 1;
}
}
}
else if (ssDCval >= 151) {
if (Rs < 1850) {
return 2;
}
else if (Rs >= 1850) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 73) {
if (refDCval < 57) {
if (TSC < 3) {
if (negBalance < 24) {
if (Rs < 1097) {
if (Rs < 698) {
return 2;
}
else if (Rs >= 698) {
return 1;
}
}
else if (Rs >= 1097) {
return 8;
}
}
else if (negBalance >= 24) {
return 4;
}
}
else if (TSC >= 3) {
if (RsCsDiff < 23) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 4;
}
}
else if (RsCsDiff >= 23) {
if (Cs < 755) {
return 1;
}
else if (Cs >= 755) {
return 8;
}
}
}
}
else if (refDCval >= 57) {
if (mu_mv_mag_sq < 89) {
if (negBalance < 33) {
if (tcor < 90) {
if (Rs < 1576) {
return 8;
}
else if (Rs >= 1576) {
if (Rs < 1915) {
return 2;
}
else if (Rs >= 1915) {
return 4;
}
}
}
else if (tcor >= 90) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
else if (negBalance >= 33) {
if (Cs < 900) {
if (Cs < 710) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Cs >= 710) {
if (tcor < 94) {
return 2;
}
else if (tcor >= 94) {
return 1;
}
}
}
else if (Cs >= 900) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (Rs < 582) {
if (Cs < 643) {
return 8;
}
else if (Cs >= 643) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (Rs >= 582) {
if (negBalance < 15) {
if (CsDiff < 4) {
if (Rs < 972) {
return 1;
}
else if (Rs >= 972) {
return 8;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (negBalance >= 15) {
if (RsCsDiff < 9) {
if (ssDCval < 58) {
if (Rs < 1344) {
return 8;
}
else if (Rs >= 1344) {
return 1;
}
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (RsCsDiff >= 9) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (MVDiff < 1) {
if (Rs < 1464) {
return 4;
}
else if (Rs >= 1464) {
return 1;
}
}
else if (MVDiff >= 1) {
if (diffMVdiffVal < 7) {
if (Rs < 1984) {
return 8;
}
else if (Rs >= 1984) {
if (negBalance < 31) {
if (Cs < 893) {
return 2;
}
else if (Cs >= 893) {
return 4;
}
}
else if (negBalance >= 31) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 7) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (posBalance >= 41) {
if (refDCval < 156) {
if (mu_mv_mag_sq < 32) {
if (negBalance < 44) {
if (refDCval < 60) {
if (negBalance < 32) {
if (tcor < 80) {
if (ssDCval < 27) {
return 8;
}
else if (ssDCval >= 27) {
if (MVDiff < 12) {
if (Rs < 741) {
return 2;
}
else if (Rs >= 741) {
return 8;
}
}
else if (MVDiff >= 12) {
if (MVDiff < 114) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (MVDiff >= 114) {
return 2;
}
}
}
}
else if (tcor >= 80) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Cs < 625) {
if (Rs < 1175) {
return 4;
}
else if (Rs >= 1175) {
return 1;
}
}
else if (Cs >= 625) {
return 4;
}
}
}
}
else if (negBalance >= 32) {
if (Rs < 845) {
if (negBalance < 36) {
if (refDCval < 47) {
return 8;
}
else if (refDCval >= 47) {
if (posBalance < 57) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (negBalance >= 36) {
if (MVDiff < 10) {
if (SC < 1580) {
if (Rs < 459) {
return 2;
}
else if (Rs >= 459) {
return 8;
}
}
else if (SC >= 1580) {
if (Rs < 718) {
return 8;
}
else if (Rs >= 718) {
return 2;
}
}
}
else if (MVDiff >= 10) {
if (Rs < 375) {
return 1;
}
else if (Rs >= 375) {
if (ssDCval < 46) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
return 2;
}
}
else if (ssDCval >= 46) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (Rs < 633) {
return 4;
}
else if (Rs >= 633) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 845) {
if (refDCval < 48) {
if (mu_mv_mag_sq < 26) {
return 8;
}
else if (mu_mv_mag_sq >= 26) {
return 2;
}
}
else if (refDCval >= 48) {
if (mcTcor < 96) {
if (Cs < 669) {
return 4;
}
else if (Cs >= 669) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 8) {
if (ssDCval < 55) {
if (RsCsDiff < 33) {
return 8;
}
else if (RsCsDiff >= 33) {
return 4;
}
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (MVDiff >= 8) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
if (MVDiff < 10) {
if (Rs < 1096) {
return 4;
}
else if (Rs >= 1096) {
return 2;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (Cs < 911) {
if (Cs < 732) {
return 2;
}
else if (Cs >= 732) {
return 1;
}
}
else if (Cs >= 911) {
if (Cs < 1153) {
return 8;
}
else if (Cs >= 1153) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 60) {
if (RsCsDiff < 37) {
if (ssDCval < 79) {
if (Rs < 1096) {
if (diffAFD < 2) {
if (SC < 1930) {
if (tcor < 86) {
if (negBalance < 40) {
if (Rs < 804) {
return 2;
}
else if (Rs >= 804) {
return 1;
}
}
else if (negBalance >= 40) {
return 8;
}
}
else if (tcor >= 86) {
if (tcor < 94) {
if (diffMVdiffVal < 4) {
if (CsDiff < 4) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (diffMVdiffVal >= 4) {
if (Rs < 640) {
return 8;
}
else if (Rs >= 640) {
return 2;
}
}
}
else if (tcor >= 94) {
if (ssDCval < 72) {
if (Rs < 757) {
return 4;
}
else if (Rs >= 757) {
return 8;
}
}
else if (ssDCval >= 72) {
return 1;
}
}
}
}
else if (SC >= 1930) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
}
else if (diffAFD >= 2) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (Rs >= 1096) {
if (SC < 2642) {
if (ssDCval < 64) {
if (Cs < 1077) {
return 4;
}
else if (Cs >= 1077) {
return 1;
}
}
else if (ssDCval >= 64) {
if (Cs < 1109) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
return 4;
}
}
else if (Cs >= 1109) {
return 4;
}
}
}
else if (SC >= 2642) {
if (Cs < 710) {
return 2;
}
else if (Cs >= 710) {
return 4;
}
}
}
}
else if (ssDCval >= 79) {
if (MVDiff < 4) {
if (Rs < 795) {
return 2;
}
else if (Rs >= 795) {
return 8;
}
}
else if (MVDiff >= 4) {
if (SC < 1831) {
return 8;
}
else if (SC >= 1831) {
if (negBalance < 38) {
if (posBalance < 77) {
if (ssDCval < 139) {
return 1;
}
else if (ssDCval >= 139) {
return 4;
}
}
else if (posBalance >= 77) {
return 2;
}
}
else if (negBalance >= 38) {
if (Rs < 1130) {
if (Rs < 1034) {
return 2;
}
else if (Rs >= 1034) {
return 4;
}
}
else if (Rs >= 1130) {
if (ssDCval < 96) {
return 1;
}
else if (ssDCval >= 96) {
if (Cs < 642) {
return 2;
}
else if (Cs >= 642) {
return 8;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 37) {
if (ssDCval < 72) {
if (mcTcor < 91) {
if (ssDCval < 64) {
if (Rs < 1963) {
if (Cs < 933) {
return 4;
}
else if (Cs >= 933) {
return 8;
}
}
else if (Rs >= 1963) {
return 2;
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (ssDCval >= 72) {
if (ssDCval < 88) {
if (TSC < 7) {
if (Rs < 816) {
return 2;
}
else if (Rs >= 816) {
return 1;
}
}
else if (TSC >= 7) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 1661) {
return 8;
}
else if (Rs >= 1661) {
return 1;
}
}
}
}
else if (ssDCval >= 88) {
if (negBalance < 40) {
if (tcor < 81) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (tcor >= 81) {
if (Cs < 580) {
return 1;
}
else if (Cs >= 580) {
if (SC < 3064) {
return 2;
}
else if (SC >= 3064) {
if (Cs < 1204) {
return 2;
}
else if (Cs >= 1204) {
return 8;
}
}
}
}
}
else if (negBalance >= 40) {
if (Cs < 1068) {
return 4;
}
else if (Cs >= 1068) {
if (Cs < 1179) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
if (Rs < 1254) {
return 2;
}
else if (Rs >= 1254) {
return 8;
}
}
}
else if (Cs >= 1179) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 422) {
if (negBalance < 48) {
if (Rs < 352) {
return 1;
}
else if (Rs >= 352) {
return 2;
}
}
else if (negBalance >= 48) {
if (mcTcor < 93) {
if (Cs < 929) {
return 8;
}
else if (Cs >= 929) {
return 1;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
}
else if (Rs >= 422) {
if (Cs < 724) {
if (Rs < 1746) {
if (posBalance < 47) {
if (SC < 1768) {
if (Cs < 634) {
return 2;
}
else if (Cs >= 634) {
if (Cs < 657) {
return 8;
}
else if (Cs >= 657) {
return 2;
}
}
}
else if (SC >= 1768) {
return 1;
}
}
else if (posBalance >= 47) {
if (refDCval < 73) {
if (Rs < 729) {
if (MVDiff < 7) {
if (Cs < 549) {
return 8;
}
else if (Cs >= 549) {
return 4;
}
}
else if (MVDiff >= 7) {
if (gchDC < 1) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Cs < 629) {
return 8;
}
else if (Cs >= 629) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (MVDiff < 19) {
if (tcor < 65) {
return 8;
}
else if (tcor >= 65) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (MVDiff >= 19) {
if (Rs < 652) {
if (diffMVdiffVal < -2) {
if (Rs < 508) {
return 2;
}
else if (Rs >= 508) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
else if (Rs >= 652) {
return 8;
}
}
}
}
}
else if (Rs >= 729) {
if (Rs < 860) {
if (negBalance < 54) {
if (ssDCval < 45) {
if (Cs < 537) {
return 4;
}
else if (Cs >= 537) {
return 8;
}
}
else if (ssDCval >= 45) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (Cs < 625) {
return 1;
}
else if (Cs >= 625) {
return 4;
}
}
}
}
else if (negBalance >= 54) {
return 8;
}
}
else if (Rs >= 860) {
if (refDCval < 62) {
if (Cs < 553) {
if (Rs < 876) {
return 2;
}
else if (Rs >= 876) {
if (Rs < 1125) {
return 4;
}
else if (Rs >= 1125) {
return 1;
}
}
}
else if (Cs >= 553) {
if (Cs < 722) {
if (Rs < 1548) {
if (Cs < 721) {
return 8;
}
else if (Cs >= 721) {
if (Rs < 1201) {
return 4;
}
else if (Rs >= 1201) {
return 8;
}
}
}
else if (Rs >= 1548) {
return 4;
}
}
else if (Cs >= 722) {
if (Rs < 890) {
return 8;
}
else if (Rs >= 890) {
return 1;
}
}
}
}
else if (refDCval >= 62) {
if (tcor < 86) {
if (MVDiff < 39) {
return 2;
}
else if (MVDiff >= 39) {
return 8;
}
}
else if (tcor >= 86) {
if (Rs < 1081) {
return 1;
}
else if (Rs >= 1081) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 73) {
if (diffAFD < -1) {
if (negBalance < 50) {
if (Rs < 712) {
if (Rs < 517) {
return 1;
}
else if (Rs >= 517) {
return 8;
}
}
else if (Rs >= 712) {
return 2;
}
}
else if (negBalance >= 50) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (SC < 1887) {
return 8;
}
else if (SC >= 1887) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
}
else if (diffAFD >= -1) {
if (diffAFD < 1) {
if (RsCsDiff < 31) {
if (negBalance < 46) {
if (MVDiff < 5) {
if (Rs < 616) {
return 1;
}
else if (Rs >= 616) {
return 2;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (negBalance >= 46) {
if (diffMVdiffVal < 7) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 673) {
return 2;
}
else if (Rs >= 673) {
return 8;
}
}
}
else if (diffMVdiffVal >= 7) {
return 2;
}
}
}
else if (RsCsDiff >= 31) {
if (negBalance < 48) {
if (Rs < 1335) {
return 2;
}
else if (Rs >= 1335) {
return 8;
}
}
else if (negBalance >= 48) {
return 4;
}
}
}
else if (diffAFD >= 1) {
if (MVDiff < 3) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
return 1;
}
}
else if (MVDiff >= 3) {
if (diffAFD < 7) {
return 8;
}
else if (diffAFD >= 7) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 1746) {
if (Rs < 2114) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
if (Rs < 2001) {
return 2;
}
else if (Rs >= 2001) {
return 1;
}
}
}
else if (Rs >= 2114) {
return 2;
}
}
}
else if (Cs >= 724) {
if (MVDiff < 12) {
if (ssDCval < 73) {
if (ssDCval < 68) {
if (Rs < 1194) {
if (SC < 2089) {
if (ssDCval < 52) {
if (MVDiff < 6) {
if (posBalance < 59) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
else if (posBalance >= 59) {
if (Rs < 983) {
if (Rs < 838) {
return 8;
}
else if (Rs >= 838) {
return 1;
}
}
else if (Rs >= 983) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (Rs < 667) {
return 4;
}
else if (Rs >= 667) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
return 4;
}
}
}
}
else if (ssDCval >= 52) {
if (mcTcor < 88) {
return 4;
}
else if (mcTcor >= 88) {
if (SC < 2069) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (SC >= 2069) {
return 4;
}
}
}
}
else if (SC >= 2089) {
if (RsDiff < 4) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (RsDiff >= 4) {
return 4;
}
}
}
else if (Rs >= 1194) {
if (TSC < 9) {
if (diffMVdiffVal < -5) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
return 4;
}
}
else if (diffMVdiffVal >= -5) {
if (Cs < 1095) {
return 8;
}
else if (Cs >= 1095) {
if (refDCval < 63) {
if (Rs < 1266) {
return 4;
}
else if (Rs >= 1266) {
if (Cs < 1097) {
return 1;
}
else if (Cs >= 1097) {
return 8;
}
}
}
else if (refDCval >= 63) {
if (Rs < 1481) {
return 8;
}
else if (Rs >= 1481) {
return 1;
}
}
}
}
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (ssDCval >= 68) {
if (mcTcor < 95) {
if (MVDiff < 6) {
if (Rs < 1565) {
return 8;
}
else if (Rs >= 1565) {
return 1;
}
}
else if (MVDiff >= 6) {
if (Cs < 775) {
return 8;
}
else if (Cs >= 775) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
}
}
else if (ssDCval >= 73) {
if (mcTcor < 95) {
if (Cs < 1316) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
if (mcTcor < 85) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (Rs < 1619) {
return 2;
}
else if (Rs >= 1619) {
return 8;
}
}
}
}
else if (mcTcor >= 85) {
return 8;
}
}
}
else if (Cs >= 1316) {
return 4;
}
}
else if (mcTcor >= 95) {
if (diffMVdiffVal < -3) {
if (Rs < 1604) {
if (negBalance < 51) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (negBalance >= 51) {
return 8;
}
}
else if (Rs >= 1604) {
if (Cs < 976) {
return 4;
}
else if (Cs >= 976) {
return 8;
}
}
}
else if (diffMVdiffVal >= -3) {
if (ssDCval < 96) {
if (SC < 2171) {
return 8;
}
else if (SC >= 2171) {
if (gchDC < 1) {
if (Cs < 1176) {
return 8;
}
else if (Cs >= 1176) {
return 1;
}
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (ssDCval >= 96) {
return 8;
}
}
}
}
}
else if (MVDiff >= 12) {
if (refDCval < 61) {
if (CsDiff < 5) {
if (MVDiff < 14) {
if (SC < 1655) {
if (Rs < 733) {
return 4;
}
else if (Rs >= 733) {
return 1;
}
}
else if (SC >= 1655) {
if (Rs < 1498) {
return 8;
}
else if (Rs >= 1498) {
return 1;
}
}
}
else if (MVDiff >= 14) {
if (diffTSC < -7) {
return 4;
}
else if (diffTSC >= -7) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 1042) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (Cs < 858) {
return 8;
}
else if (Cs >= 858) {
return 4;
}
}
}
else if (Rs >= 1042) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (refDCval >= 61) {
if (mcTcor < 87) {
if (SC < 2629) {
if (tcor < 68) {
if (Rs < 941) {
return 4;
}
else if (Rs >= 941) {
return 1;
}
}
else if (tcor >= 68) {
if (Cs < 1070) {
if (ssDCval < 101) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (negBalance < 50) {
if (Rs < 1184) {
return 8;
}
else if (Rs >= 1184) {
return 2;
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (Cs >= 1070) {
return 1;
}
}
}
else if (SC >= 2629) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (mcTcor >= 87) {
if (ssDCval < 95) {
if (MVDiff < 35) {
if (refDCval < 92) {
if (diffRsCsdiff < 27) {
if (RsCsDiff < 33) {
if (Rs < 1210) {
if (ssDCval < 80) {
return 4;
}
else if (ssDCval >= 80) {
return 8;
}
}
else if (Rs >= 1210) {
if (Rs < 1886) {
return 8;
}
else if (Rs >= 1886) {
return 2;
}
}
}
else if (RsCsDiff >= 33) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (Cs < 1074) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 4;
}
}
else if (Cs >= 1074) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 27) {
if (Rs < 986) {
return 2;
}
else if (Rs >= 986) {
return 1;
}
}
}
else if (refDCval >= 92) {
return 2;
}
}
else if (MVDiff >= 35) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 1;
}
}
}
else if (ssDCval >= 95) {
if (Rs < 1052) {
if (Cs < 800) {
return 8;
}
else if (Cs >= 800) {
return 4;
}
}
else if (Rs >= 1052) {
if (MVDiff < 19) {
if (diffRsCsdiff < -18) {
return 4;
}
else if (diffRsCsdiff >= -18) {
return 8;
}
}
else if (MVDiff >= 19) {
if (ssDCval < 136) {
if (Rs < 1210) {
return 4;
}
else if (Rs >= 1210) {
return 8;
}
}
else if (ssDCval >= 136) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (mcTcor < 85) {
if (ssDCval < 42) {
if (Rs < 720) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
if (Cs < 548) {
return 8;
}
else if (Cs >= 548) {
return 2;
}
}
}
else if (Rs >= 720) {
return 8;
}
}
else if (ssDCval >= 42) {
if (tcor < 74) {
if (mcTcor < 75) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 906) {
return 1;
}
else if (Rs >= 906) {
return 8;
}
}
}
}
else if (mcTcor >= 75) {
if (negBalance < 48) {
if (mu_mv_mag_sq < 97) {
if (Rs < 1017) {
if (RsCsDiff < 23) {
return 8;
}
else if (RsCsDiff >= 23) {
if (SC < 1653) {
return 1;
}
else if (SC >= 1653) {
return 2;
}
}
}
else if (Rs >= 1017) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 899) {
return 1;
}
else if (Cs >= 899) {
return 4;
}
}
}
else if (negBalance >= 48) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (SC < 1869) {
return 4;
}
else if (SC >= 1869) {
return 8;
}
}
}
}
}
else if (tcor >= 74) {
if (AFD < 6) {
if (ssDCval < 47) {
if (Cs < 935) {
return 1;
}
else if (Cs >= 935) {
return 8;
}
}
else if (ssDCval >= 47) {
return 8;
}
}
else if (AFD >= 6) {
if (SC < 1671) {
if (Rs < 854) {
if (MVDiff < 4) {
if (Rs < 680) {
return 8;
}
else if (Rs >= 680) {
return 4;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (Rs >= 854) {
if (Cs < 652) {
if (Rs < 1080) {
return 8;
}
else if (Rs >= 1080) {
return 2;
}
}
else if (Cs >= 652) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
else if (SC >= 1671) {
if (mu_mv_mag_sq < 37) {
if (Rs < 1327) {
return 8;
}
else if (Rs >= 1327) {
if (Rs < 1675) {
return 1;
}
else if (Rs >= 1675) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (tcor < 76) {
if (Cs < 1163) {
return 8;
}
else if (Cs >= 1163) {
if (Rs < 1195) {
return 8;
}
else if (Rs >= 1195) {
return 1;
}
}
}
else if (tcor >= 76) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 85) {
if (ssDCval < 78) {
if (negBalance < 44) {
if (AFD < 8) {
if (Rs < 917) {
if (refDCval < 57) {
if (Rs < 514) {
if (diffMVdiffVal < -8) {
return 8;
}
else if (diffMVdiffVal >= -8) {
if (negBalance < 22) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (Rs < 364) {
return 2;
}
else if (Rs >= 364) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (ssDCval < 46) {
if (Rs < 310) {
return 1;
}
else if (Rs >= 310) {
return 4;
}
}
else if (ssDCval >= 46) {
return 1;
}
}
}
}
else if (Rs >= 514) {
if (refDCval < 37) {
if (Cs < 815) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
return 8;
}
}
else if (Cs >= 815) {
return 1;
}
}
else if (refDCval >= 37) {
if (negBalance < 20) {
if (Rs < 566) {
return 4;
}
else if (Rs >= 566) {
return 1;
}
}
else if (negBalance >= 20) {
if (diffTSC < 1) {
if (negBalance < 35) {
if (tcor < 96) {
if (AFD < 3) {
if (Cs < 859) {
return 4;
}
else if (Cs >= 859) {
return 1;
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (tcor >= 96) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
if (Rs < 681) {
return 1;
}
else if (Rs >= 681) {
return 8;
}
}
}
}
else if (negBalance >= 35) {
if (diffRsCsdiff < 1) {
if (posBalance < 57) {
if (Rs < 768) {
if (Cs < 651) {
return 8;
}
else if (Cs >= 651) {
return 4;
}
}
else if (Rs >= 768) {
return 8;
}
}
else if (posBalance >= 57) {
return 4;
}
}
else if (diffRsCsdiff >= 1) {
if (Cs < 912) {
return 4;
}
else if (Cs >= 912) {
return 2;
}
}
}
}
else if (diffTSC >= 1) {
if (tcor < 93) {
if (ssDCval < 55) {
if (SC < 1276) {
return 4;
}
else if (SC >= 1276) {
return 8;
}
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (tcor >= 93) {
if (Rs < 670) {
return 1;
}
else if (Rs >= 670) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 57) {
if (mu_mv_mag_sq < 119) {
if (RsCsDiff < 15) {
if (AFD < 4) {
if (SC < 1153) {
if (ssDCval < 60) {
return 2;
}
else if (ssDCval >= 60) {
if (Rs < 278) {
return 4;
}
else if (Rs >= 278) {
if (Rs < 408) {
return 8;
}
else if (Rs >= 408) {
return 1;
}
}
}
}
else if (SC >= 1153) {
if (Rs < 824) {
return 4;
}
else if (Rs >= 824) {
return 8;
}
}
}
else if (AFD >= 4) {
if (Rs < 875) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (SC < 1373) {
if (mu_mv_mag_sq < 78) {
if (negBalance < 36) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (negBalance >= 36) {
return 8;
}
}
else if (mu_mv_mag_sq >= 78) {
return 8;
}
}
else if (SC >= 1373) {
if (ssDCval < 75) {
if (Rs < 608) {
return 1;
}
else if (Rs >= 608) {
if (mu_mv_mag_sq < 89) {
if (ssDCval < 69) {
if (negBalance < 38) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (diffMVdiffVal < 1) {
if (Rs < 817) {
return 8;
}
else if (Rs >= 817) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (mu_mv_mag_sq >= 89) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 730) {
return 8;
}
else if (Rs >= 730) {
return 4;
}
}
}
}
}
else if (ssDCval >= 75) {
return 2;
}
}
}
}
else if (Rs >= 875) {
if (refDCval < 63) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (refDCval >= 63) {
return 4;
}
}
}
}
else if (RsCsDiff >= 15) {
if (mu_mv_mag_sq < 82) {
if (Rs < 770) {
return 1;
}
else if (Rs >= 770) {
if (Rs < 871) {
return 4;
}
else if (Rs >= 871) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 82) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
}
else if (Rs >= 917) {
if (RsCsDiff < 9) {
if (ssDCval < 63) {
if (Cs < 530) {
return 2;
}
else if (Cs >= 530) {
if (Rs < 1249) {
if (negBalance < 23) {
if (Cs < 727) {
return 8;
}
else if (Cs >= 727) {
return 1;
}
}
else if (negBalance >= 23) {
if (SC < 2366) {
if (mcTcor < 95) {
if (gchDC < 2) {
if (diffRsCsdiff < 5) {
return 8;
}
else if (diffRsCsdiff >= 5) {
if (Rs < 1093) {
return 8;
}
else if (Rs >= 1093) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (Rs < 1073) {
return 1;
}
else if (Rs >= 1073) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (posBalance < 45) {
if (ssDCval < 56) {
if (Rs < 1151) {
return 1;
}
else if (Rs >= 1151) {
return 8;
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (posBalance >= 45) {
if (ssDCval < 60) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 1;
}
}
else if (ssDCval >= 60) {
if (Cs < 871) {
return 8;
}
else if (Cs >= 871) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
}
}
else if (SC >= 2366) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 1;
}
}
}
}
else if (Rs >= 1249) {
if (Cs < 718) {
return 4;
}
else if (Cs >= 718) {
if (ssDCval < 48) {
if (Cs < 861) {
return 2;
}
else if (Cs >= 861) {
return 8;
}
}
else if (ssDCval >= 48) {
return 8;
}
}
}
}
}
else if (ssDCval >= 63) {
if (Rs < 1056) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Rs >= 1056) {
if (AFD < 4) {
if (Rs < 1150) {
return 4;
}
else if (Rs >= 1150) {
if (Rs < 1498) {
return 1;
}
else if (Rs >= 1498) {
return 8;
}
}
}
else if (AFD >= 4) {
if (MVDiff < 19) {
if (mu_mv_mag_sq < 93) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
if (mu_mv_mag_sq < 87) {
return 8;
}
else if (mu_mv_mag_sq >= 87) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 93) {
return 8;
}
}
else if (MVDiff >= 19) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 9) {
if (Cs < 912) {
if (posBalance < 48) {
if (Cs < 757) {
if (tcor < 89) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (Rs < 986) {
return 2;
}
else if (Rs >= 986) {
return 1;
}
}
}
else if (tcor >= 89) {
return 1;
}
}
else if (Cs >= 757) {
if (MVDiff < 3) {
if (diffMVdiffVal < 1) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Rs < 1117) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
else if (Rs >= 1117) {
if (Cs < 844) {
if (Rs < 2172) {
return 1;
}
else if (Rs >= 2172) {
return 2;
}
}
else if (Cs >= 844) {
if (Rs < 1172) {
return 8;
}
else if (Rs >= 1172) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 48) {
if (Rs < 988) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
return 4;
}
}
else if (Rs >= 988) {
if (mu_mv_mag_sq < 117) {
if (refDCval < 54) {
if (Rs < 1259) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 114) {
return 8;
}
else if (mu_mv_mag_sq >= 114) {
if (Rs < 1136) {
return 8;
}
else if (Rs >= 1136) {
return 1;
}
}
}
}
else if (Rs >= 1259) {
if (ssDCval < 47) {
return 2;
}
else if (ssDCval >= 47) {
return 8;
}
}
}
else if (refDCval >= 54) {
if (RsCsDiff < 15) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
if (ssDCval < 72) {
if (Rs < 1027) {
return 1;
}
else if (Rs >= 1027) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 8;
}
}
}
else if (ssDCval >= 72) {
return 1;
}
}
}
else if (RsCsDiff >= 15) {
if (AFD < 7) {
if (tcor < 90) {
if (Cs < 719) {
return 1;
}
else if (Cs >= 719) {
return 8;
}
}
else if (tcor >= 90) {
return 1;
}
}
else if (AFD >= 7) {
if (Rs < 1621) {
if (Cs < 682) {
return 8;
}
else if (Cs >= 682) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 1;
}
}
}
else if (Rs >= 1621) {
if (Rs < 1853) {
return 4;
}
else if (Rs >= 1853) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (tcor < 87) {
return 1;
}
else if (tcor >= 87) {
if (Cs < 801) {
return 1;
}
else if (Cs >= 801) {
return 8;
}
}
}
}
}
}
else if (Cs >= 912) {
if (ssDCval < 64) {
if (SC < 2396) {
if (RsCsDiff < 27) {
if (mcTcor < 94) {
if (ssDCval < 58) {
if (Cs < 929) {
if (Rs < 1260) {
return 8;
}
else if (Rs >= 1260) {
return 1;
}
}
else if (Cs >= 929) {
return 8;
}
}
else if (ssDCval >= 58) {
if (Rs < 1142) {
return 4;
}
else if (Rs >= 1142) {
return 1;
}
}
}
else if (mcTcor >= 94) {
if (diffRsCsdiff < -3) {
return 4;
}
else if (diffRsCsdiff >= -3) {
if (SC < 2191) {
if (SC < 2062) {
return 8;
}
else if (SC >= 2062) {
return 4;
}
}
else if (SC >= 2191) {
return 8;
}
}
}
}
else if (RsCsDiff >= 27) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
return 4;
}
}
}
else if (SC >= 2396) {
if (tcor < 89) {
if (posBalance < 57) {
if (Rs < 1349) {
return 1;
}
else if (Rs >= 1349) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
if (Rs < 1424) {
return 1;
}
else if (Rs >= 1424) {
return 8;
}
}
}
}
else if (tcor >= 89) {
if (posBalance < 49) {
if (diffRsCsdiff < -6) {
return 4;
}
else if (diffRsCsdiff >= -6) {
if (Cs < 1123) {
if (RsDiff < 4) {
if (Cs < 1072) {
return 8;
}
else if (Cs >= 1072) {
return 1;
}
}
else if (RsDiff >= 4) {
if (negBalance < 40) {
if (Rs < 1944) {
return 4;
}
else if (Rs >= 1944) {
return 4;
}
}
else if (negBalance >= 40) {
return 8;
}
}
}
else if (Cs >= 1123) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (Cs < 1013) {
return 1;
}
else if (Cs >= 1013) {
if (Rs < 1272) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (Rs >= 1272) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (Cs < 1072) {
if (mu_mv_mag_sq < 69) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 69) {
if (Rs < 1602) {
return 8;
}
else if (Rs >= 1602) {
if (Rs < 1613) {
return 2;
}
else if (Rs >= 1613) {
return 8;
}
}
}
}
else if (Cs >= 1072) {
if (SC < 2734) {
if (SC < 2359) {
if (mu_mv_mag_sq < 64) {
if (Rs < 1078) {
return 8;
}
else if (Rs >= 1078) {
return 2;
}
}
else if (mu_mv_mag_sq >= 64) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 4;
}
}
}
else if (SC >= 2359) {
if (diffRsCsdiff < 11) {
if (MVDiff < 1) {
if (Rs < 1324) {
return 1;
}
else if (Rs >= 1324) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (MVDiff >= 1) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 116) {
if (Rs < 1304) {
if (Rs < 1279) {
return 8;
}
else if (Rs >= 1279) {
return 4;
}
}
else if (Rs >= 1304) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
return 4;
}
}
else if (MVDiff >= 4) {
if (diffMVdiffVal < 5) {
return 4;
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 11) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
return 1;
}
}
}
}
else if (SC >= 2734) {
if (mu_mv_mag_sq < 113) {
if (Rs < 1947) {
if (tcor < 93) {
if (Rs < 1665) {
return 1;
}
else if (Rs >= 1665) {
return 8;
}
}
else if (tcor >= 93) {
return 1;
}
}
else if (Rs >= 1947) {
return 8;
}
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 1964) {
return 8;
}
else if (Rs >= 1964) {
return 2;
}
}
}
}
}
}
}
}
}
else if (AFD >= 8) {
if (negBalance < 37) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
if (ssDCval < 60) {
if (posBalance < 59) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 1729) {
if (Rs < 976) {
if (Rs < 481) {
return 2;
}
else if (Rs >= 481) {
if (diffRsCsdiff < -11) {
return 1;
}
else if (diffRsCsdiff >= -11) {
return 4;
}
}
}
else if (Rs >= 976) {
return 2;
}
}
else if (Rs >= 1729) {
if (Rs < 1936) {
return 8;
}
else if (Rs >= 1936) {
return 4;
}
}
}
}
else if (posBalance >= 59) {
if (mu_mv_mag_sq < 48) {
if (Rs < 1262) {
return 8;
}
else if (Rs >= 1262) {
return 2;
}
}
else if (mu_mv_mag_sq >= 48) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (RsCsDiff < 43) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
if (Cs < 1051) {
return 8;
}
else if (Cs >= 1051) {
return 1;
}
}
}
else if (RsCsDiff >= 43) {
if (negBalance < 31) {
if (Rs < 1971) {
return 8;
}
else if (Rs >= 1971) {
return 1;
}
}
else if (negBalance >= 31) {
return 2;
}
}
}
}
}
else if (negBalance >= 37) {
if (Cs < 959) {
if (mu_mv_mag_sq < 68) {
if (Rs < 1864) {
if (ssDCval < 58) {
if (SC < 1657) {
if (Rs < 707) {
return 1;
}
else if (Rs >= 707) {
return 8;
}
}
else if (SC >= 1657) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (ssDCval >= 58) {
if (mu_mv_mag_sq < 38) {
return 8;
}
else if (mu_mv_mag_sq >= 38) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (Rs < 1256) {
return 1;
}
else if (Rs >= 1256) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1864) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (diffMVdiffVal < 3) {
if (RsCsDiff < 55) {
return 4;
}
else if (RsCsDiff >= 55) {
return 2;
}
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
}
else if (Cs >= 959) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (posBalance < 62) {
if (Cs < 1309) {
if (Rs < 1293) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (Rs >= 1293) {
if (Rs < 1951) {
return 8;
}
else if (Rs >= 1951) {
if (Cs < 1018) {
return 4;
}
else if (Cs >= 1018) {
if (Rs < 2113) {
return 1;
}
else if (Rs >= 2113) {
return 8;
}
}
}
}
}
else if (Cs >= 1309) {
if (Rs < 1051) {
return 2;
}
else if (Rs >= 1051) {
return 1;
}
}
}
else if (posBalance >= 62) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 1602) {
if (MVDiff < 5) {
if (Rs < 737) {
return 4;
}
else if (Rs >= 737) {
if (tcor < 96) {
if (mu_mv_mag_sq < 87) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (TSC < 6) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
else if (TSC >= 6) {
if (RsCsDiff < 46) {
if (Rs < 1014) {
return 4;
}
else if (Rs >= 1014) {
return 8;
}
}
else if (RsCsDiff >= 46) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (Cs < 1267) {
return 8;
}
else if (Cs >= 1267) {
if (Cs < 1280) {
return 1;
}
else if (Cs >= 1280) {
return 8;
}
}
}
}
else if (tcor >= 96) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
}
else if (MVDiff >= 5) {
if (TSC < 8) {
if (refDCval < 57) {
if (Rs < 806) {
if (Cs < 648) {
return 8;
}
else if (Cs >= 648) {
return 4;
}
}
else if (Rs >= 806) {
if (Cs < 557) {
return 2;
}
else if (Cs >= 557) {
if (Rs < 977) {
if (Cs < 753) {
return 8;
}
else if (Cs >= 753) {
if (Cs < 818) {
return 1;
}
else if (Cs >= 818) {
return 8;
}
}
}
else if (Rs >= 977) {
if (RsCsDiff < 38) {
return 8;
}
else if (RsCsDiff >= 38) {
if (Rs < 1165) {
return 4;
}
else if (Rs >= 1165) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 57) {
if (Rs < 850) {
if (Rs < 483) {
return 1;
}
else if (Rs >= 483) {
if (SC < 1938) {
if (AFD < 11) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
return 1;
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (SC >= 1938) {
return 1;
}
}
}
else if (Rs >= 850) {
if (Cs < 1204) {
if (RsDiff < 4) {
if (Cs < 1015) {
if (SC < 1738) {
if (mu_mv_mag_sq < 45) {
return 8;
}
else if (mu_mv_mag_sq >= 45) {
return 4;
}
}
else if (SC >= 1738) {
return 8;
}
}
else if (Cs >= 1015) {
return 4;
}
}
else if (RsDiff >= 4) {
if (AFD < 10) {
if (RsCsDiff < 38) {
if (Cs < 559) {
if (Rs < 1107) {
return 1;
}
else if (Rs >= 1107) {
return 8;
}
}
else if (Cs >= 559) {
return 4;
}
}
else if (RsCsDiff >= 38) {
return 8;
}
}
else if (AFD >= 10) {
if (Rs < 1096) {
return 4;
}
else if (Rs >= 1096) {
return 8;
}
}
}
}
else if (Cs >= 1204) {
return 1;
}
}
}
}
else if (TSC >= 8) {
if (Cs < 1229) {
return 8;
}
else if (Cs >= 1229) {
return 1;
}
}
}
}
else if (Rs >= 1602) {
if (SC < 3415) {
if (mu_mv_mag_sq < 91) {
if (Cs < 1025) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (mu_mv_mag_sq < 46) {
if (MVDiff < 34) {
return 4;
}
else if (MVDiff >= 34) {
return 8;
}
}
else if (mu_mv_mag_sq >= 46) {
if (refDCval < 71) {
return 2;
}
else if (refDCval >= 71) {
if (diffMVdiffVal < -13) {
return 2;
}
else if (diffMVdiffVal >= -13) {
return 8;
}
}
}
}
}
else if (Cs >= 1025) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
if (MVDiff < 28) {
if (mu_mv_mag_sq < 86) {
return 8;
}
else if (mu_mv_mag_sq >= 86) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
}
else if (MVDiff >= 28) {
if (Rs < 1944) {
return 8;
}
else if (Rs >= 1944) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (refDCval < 68) {
if (Cs < 1261) {
return 2;
}
else if (Cs >= 1261) {
return 8;
}
}
else if (refDCval >= 68) {
if (Rs < 1784) {
return 8;
}
else if (Rs >= 1784) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (Rs < 1949) {
return 2;
}
else if (Rs >= 1949) {
return 8;
}
}
}
}
}
}
else if (SC >= 3415) {
return 1;
}
}
}
}
else if (ssDCval >= 78) {
if (Rs < 365) {
if (negBalance < 46) {
if (Rs < 334) {
return 2;
}
else if (Rs >= 334) {
return 8;
}
}
else if (negBalance >= 46) {
if (Rs < 327) {
return 1;
}
else if (Rs >= 327) {
return 4;
}
}
}
else if (Rs >= 365) {
if (ssDCval < 114) {
if (mcTcor < 94) {
if (RsDiff < 6) {
if (tcor < 86) {
if (mcTcor < 92) {
if (posBalance < 43) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
else if (posBalance >= 43) {
if (SC < 2763) {
if (tcor < 74) {
return 2;
}
else if (tcor >= 74) {
return 8;
}
}
else if (SC >= 2763) {
if (Rs < 1655) {
return 2;
}
else if (Rs >= 1655) {
if (Rs < 1732) {
return 4;
}
else if (Rs >= 1732) {
return 8;
}
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 882) {
return 2;
}
else if (Rs >= 882) {
return 8;
}
}
}
else if (tcor >= 86) {
if (RsCsDiff < 4) {
return 1;
}
else if (RsCsDiff >= 4) {
if (MVDiff < 2) {
if (Rs < 645) {
return 2;
}
else if (Rs >= 645) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
else if (RsDiff >= 6) {
if (diffRsCsdiff < -6) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 1;
}
}
else if (diffRsCsdiff >= -6) {
if (RsCsDiff < 49) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
else if (RsCsDiff >= 49) {
if (mu_mv_mag_sq < 45) {
return 2;
}
else if (mu_mv_mag_sq >= 45) {
if (posBalance < 66) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 1696) {
return 4;
}
else if (Rs >= 1696) {
return 8;
}
}
}
else if (posBalance >= 66) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (MVDiff < 3) {
if (negBalance < 46) {
if (Rs < 824) {
if (posBalance < 53) {
if (SC < 1566) {
if (Cs < 602) {
return 8;
}
else if (Cs >= 602) {
return 2;
}
}
else if (SC >= 1566) {
return 1;
}
}
else if (posBalance >= 53) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Rs < 679) {
if (diffMVdiffVal < -3) {
if (Rs < 625) {
return 2;
}
else if (Rs >= 625) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (Rs >= 679) {
if (Rs < 700) {
return 2;
}
else if (Rs >= 700) {
return 4;
}
}
}
}
}
else if (Rs >= 824) {
if (SC < 2980) {
if (Cs < 653) {
if (Rs < 1728) {
return 8;
}
else if (Rs >= 1728) {
return 1;
}
}
else if (Cs >= 653) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (negBalance < 43) {
return 8;
}
else if (negBalance >= 43) {
if (Cs < 926) {
return 8;
}
else if (Cs >= 926) {
return 1;
}
}
}
}
}
else if (SC >= 2980) {
if (ssDCval < 85) {
if (Rs < 1899) {
return 1;
}
else if (Rs >= 1899) {
return 8;
}
}
else if (ssDCval >= 85) {
if (ssDCval < 96) {
return 2;
}
else if (ssDCval >= 96) {
return 8;
}
}
}
}
}
else if (negBalance >= 46) {
if (Rs < 417) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 4;
}
}
else if (Rs >= 417) {
if (posBalance < 56) {
if (mu_mv_mag_sq < 82) {
if (Rs < 1390) {
return 8;
}
else if (Rs >= 1390) {
return 1;
}
}
else if (mu_mv_mag_sq >= 82) {
if (Rs < 495) {
return 1;
}
else if (Rs >= 495) {
if (diffMVdiffVal < -1) {
if (Rs < 658) {
return 8;
}
else if (Rs >= 658) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
}
else if (posBalance >= 56) {
if (diffAFD < 3) {
return 1;
}
else if (diffAFD >= 3) {
return 8;
}
}
}
}
}
else if (MVDiff >= 3) {
if (diffRsCsdiff < 23) {
if (RsCsDiff < 48) {
if (posBalance < 51) {
if (diffRsCsdiff < -6) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (RsCsDiff < 11) {
if (Rs < 1899) {
return 4;
}
else if (Rs >= 1899) {
return 8;
}
}
else if (RsCsDiff >= 11) {
return 8;
}
}
}
else if (diffRsCsdiff >= -6) {
if (ssDCval < 92) {
if (tcor < 91) {
if (ssDCval < 87) {
return 2;
}
else if (ssDCval >= 87) {
return 1;
}
}
else if (tcor >= 91) {
if (diffMVdiffVal < 0) {
return 8;
}
else if (diffMVdiffVal >= 0) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (ssDCval >= 92) {
return 8;
}
}
}
else if (posBalance >= 51) {
if (Cs < 1144) {
if (RsCsDiff < 4) {
if (Rs < 447) {
return 1;
}
else if (Rs >= 447) {
return 8;
}
}
else if (RsCsDiff >= 4) {
if (Rs < 723) {
if (TSC < 4) {
if (Rs < 441) {
return 8;
}
else if (Rs >= 441) {
return 2;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (Rs >= 723) {
if (refDCval < 94) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (mcTcor < 96) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
else if (mcTcor >= 96) {
if (SC < 1594) {
return 8;
}
else if (SC >= 1594) {
return 1;
}
}
}
}
else if (refDCval >= 94) {
return 8;
}
}
}
}
else if (Cs >= 1144) {
if (Cs < 1156) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
else if (Cs >= 1156) {
return 8;
}
}
}
}
else if (RsCsDiff >= 48) {
if (mu_mv_mag_sq < 74) {
if (diffRsCsdiff < 1) {
if (negBalance < 45) {
if (Rs < 1502) {
return 8;
}
else if (Rs >= 1502) {
return 4;
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 1236) {
return 8;
}
else if (Cs >= 1236) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 23) {
if (Cs < 853) {
return 1;
}
else if (Cs >= 853) {
if (Rs < 1859) {
return 2;
}
else if (Rs >= 1859) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 114) {
if (SC < 2425) {
if (MVDiff < 23) {
if (mcTcor < 93) {
if (TSC < 5) {
if (SC < 1564) {
return 8;
}
else if (SC >= 1564) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (TSC >= 5) {
if (Cs < 664) {
return 4;
}
else if (Cs >= 664) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (posBalance < 62) {
if (RsCsDiff < 23) {
if (mu_mv_mag_sq < 88) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (Rs < 1326) {
return 4;
}
else if (Rs >= 1326) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 88) {
if (Rs < 1324) {
if (MVDiff < 5) {
if (Cs < 1317) {
return 8;
}
else if (Cs >= 1317) {
if (Rs < 953) {
return 8;
}
else if (Rs >= 953) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 126) {
return 4;
}
else if (ssDCval >= 126) {
return 8;
}
}
}
else if (Rs >= 1324) {
return 4;
}
}
}
else if (RsCsDiff >= 23) {
if (Rs < 1099) {
if (mu_mv_mag_sq < 43) {
return 1;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 840) {
return 2;
}
else if (Rs >= 840) {
return 8;
}
}
}
else if (Rs >= 1099) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (Rs < 808) {
if (Cs < 659) {
return 2;
}
else if (Cs >= 659) {
return 8;
}
}
else if (Rs >= 808) {
return 1;
}
}
}
}
else if (MVDiff >= 23) {
if (ssDCval < 135) {
return 8;
}
else if (ssDCval >= 135) {
if (ssDCval < 142) {
return 4;
}
else if (ssDCval >= 142) {
if (Rs < 1031) {
return 1;
}
else if (Rs >= 1031) {
return 8;
}
}
}
}
}
else if (SC >= 2425) {
if (negBalance < 20) {
return 2;
}
else if (negBalance >= 20) {
if (diffAFD < 5) {
if (Rs < 2356) {
if (posBalance < 68) {
if (Cs < 800) {
return 1;
}
else if (Cs >= 800) {
if (diffRsCsdiff < 27) {
return 8;
}
else if (diffRsCsdiff >= 27) {
if (Cs < 976) {
return 2;
}
else if (Cs >= 976) {
return 8;
}
}
}
}
else if (posBalance >= 68) {
if (ssDCval < 129) {
return 8;
}
else if (ssDCval >= 129) {
if (Rs < 1723) {
return 4;
}
else if (Rs >= 1723) {
return 1;
}
}
}
}
else if (Rs >= 2356) {
if (AFD < 6) {
if (Rs < 2379) {
return 1;
}
else if (Rs >= 2379) {
return 8;
}
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (diffAFD >= 5) {
return 4;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 156) {
if (Rs < 862) {
if (Cs < 910) {
if (Rs < 662) {
return 8;
}
else if (Rs >= 662) {
return 2;
}
}
else if (Cs >= 910) {
return 4;
}
}
else if (Rs >= 862) {
if (Rs < 1955) {
return 2;
}
else if (Rs >= 1955) {
return 8;
}
}
}
}
}
else if (AFD >= 12) {
if (tcor < 83) {
if (MVDiff < 15) {
if (ssDCval < 88) {
if (ssDCval < 57) {
if (TSC < 8) {
if (mcTcor < 94) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (Rs < 844) {
return 4;
}
else if (Rs >= 844) {
if (Rs < 1682) {
if (Cs < 1061) {
return 8;
}
else if (Cs >= 1061) {
if (Rs < 1257) {
return 4;
}
else if (Rs >= 1257) {
return 8;
}
}
}
else if (Rs >= 1682) {
if (Cs < 1008) {
return 4;
}
else if (Cs >= 1008) {
return 2;
}
}
}
}
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (TSC >= 8) {
if (RsDiff < 11) {
if (mu_mv_mag_sq < 3) {
if (Cs < 968) {
return 8;
}
else if (Cs >= 968) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1977) {
return 8;
}
else if (Rs >= 1977) {
return 4;
}
}
}
else if (RsDiff >= 11) {
return 2;
}
}
}
else if (ssDCval >= 57) {
if (mcTcor < 92) {
if (posBalance < 56) {
if (mu_mv_mag_sq < 55) {
if (negBalance < 61) {
if (negBalance < 56) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
if (Rs < 1979) {
return 8;
}
else if (Rs >= 1979) {
return 2;
}
}
}
else if (negBalance >= 56) {
return 2;
}
}
else if (negBalance >= 61) {
return 8;
}
}
else if (mu_mv_mag_sq >= 55) {
return 4;
}
}
else if (posBalance >= 56) {
if (mcTcor < 75) {
if (Rs < 1054) {
return 4;
}
else if (Rs >= 1054) {
return 8;
}
}
else if (mcTcor >= 75) {
if (negBalance < 51) {
if (RsCsDiff < 51) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (RsCsDiff < 37) {
return 1;
}
else if (RsCsDiff >= 37) {
return 2;
}
}
}
else if (RsCsDiff >= 51) {
if (negBalance < 40) {
if (SC < 2194) {
return 8;
}
else if (SC >= 2194) {
if (Rs < 1613) {
return 1;
}
else if (Rs >= 1613) {
return 4;
}
}
}
else if (negBalance >= 40) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (AFD < 22) {
if (posBalance < 59) {
if (diffMVdiffVal < 2) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (Rs < 1623) {
return 2;
}
else if (Rs >= 1623) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
if (negBalance < 56) {
return 2;
}
else if (negBalance >= 56) {
return 8;
}
}
}
else if (posBalance >= 59) {
return 8;
}
}
else if (AFD >= 22) {
if (Rs < 1710) {
return 4;
}
else if (Rs >= 1710) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 2130) {
if (refDCval < 74) {
if (TSC < 6) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (Rs < 219) {
return 4;
}
else if (Rs >= 219) {
return 1;
}
}
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 19) {
if (Cs < 1221) {
return 8;
}
else if (Cs >= 1221) {
if (Rs < 1446) {
return 1;
}
else if (Rs >= 1446) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 19) {
if (Rs < 1498) {
return 1;
}
else if (Rs >= 1498) {
return 2;
}
}
}
}
else if (refDCval >= 74) {
if (AFD < 16) {
if (Rs < 1593) {
if (negBalance < 62) {
return 8;
}
else if (negBalance >= 62) {
return 2;
}
}
else if (Rs >= 1593) {
if (Cs < 867) {
return 1;
}
else if (Cs >= 867) {
return 8;
}
}
}
else if (AFD >= 16) {
if (RsDiff < 6) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
return 4;
}
}
else if (RsDiff >= 6) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 901) {
if (Cs < 805) {
return 4;
}
else if (Cs >= 805) {
return 2;
}
}
else if (Rs >= 901) {
if (negBalance < 56) {
if (diffAFD < -3) {
if (Rs < 1562) {
return 1;
}
else if (Rs >= 1562) {
return 4;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (negBalance >= 56) {
if (Rs < 1233) {
return 2;
}
else if (Rs >= 1233) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 2130) {
return 2;
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 2) {
if (AFD < 15) {
if (tcor < 78) {
if (RsDiff < 7) {
if (Rs < 672) {
return 1;
}
else if (Rs >= 672) {
return 8;
}
}
else if (RsDiff >= 7) {
if (Rs < 1631) {
return 4;
}
else if (Rs >= 1631) {
if (Rs < 1670) {
return 8;
}
else if (Rs >= 1670) {
return 2;
}
}
}
}
else if (tcor >= 78) {
if (ssDCval < 93) {
return 4;
}
else if (ssDCval >= 93) {
if (Rs < 1413) {
return 1;
}
else if (Rs >= 1413) {
return 2;
}
}
}
}
else if (AFD >= 15) {
if (RsCsDiff < 56) {
if (Rs < 672) {
return 2;
}
else if (Rs >= 672) {
return 4;
}
}
else if (RsCsDiff >= 56) {
if (Rs < 2108) {
return 8;
}
else if (Rs >= 2108) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsDiff < 8) {
if (SC < 1224) {
if (Rs < 555) {
return 2;
}
else if (Rs >= 555) {
return 8;
}
}
else if (SC >= 1224) {
if (RsCsDiff < 19) {
if (Rs < 893) {
return 1;
}
else if (Rs >= 893) {
if (Rs < 976) {
return 4;
}
else if (Rs >= 976) {
return 8;
}
}
}
else if (RsCsDiff >= 19) {
if (refDCval < 107) {
if (ssDCval < 101) {
if (AFD < 20) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
return 8;
}
}
}
else if (AFD >= 20) {
return 4;
}
}
else if (ssDCval >= 101) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
}
else if (refDCval >= 107) {
if (tcor < 73) {
if (Rs < 701) {
return 2;
}
else if (Rs >= 701) {
if (Rs < 958) {
return 8;
}
else if (Rs >= 958) {
return 1;
}
}
}
else if (tcor >= 73) {
if (Rs < 646) {
return 1;
}
else if (Rs >= 646) {
if (gchDC < 6) {
return 8;
}
else if (gchDC >= 6) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (tcor < 64) {
if (mu_mv_mag_sq < 7) {
return 8;
}
else if (mu_mv_mag_sq >= 7) {
return 4;
}
}
else if (tcor >= 64) {
if (ssDCval < 109) {
if (mu_mv_mag_sq < 13) {
if (Rs < 1696) {
if (Rs < 1296) {
return 1;
}
else if (Rs >= 1296) {
return 8;
}
}
else if (Rs >= 1696) {
return 2;
}
}
else if (mu_mv_mag_sq >= 13) {
return 4;
}
}
else if (ssDCval >= 109) {
if (Cs < 840) {
if (Rs < 1314) {
return 8;
}
else if (Rs >= 1314) {
return 2;
}
}
else if (Cs >= 840) {
if (Rs < 1603) {
return 4;
}
else if (Rs >= 1603) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 15) {
if (mcTcor < 84) {
if (tcor < 58) {
if (mu_mv_mag_sq < 189) {
if (TSC < 10) {
if (negBalance < 52) {
if (RsDiff < 7) {
if (SC < 1692) {
if (posBalance < 60) {
if (Cs < 837) {
return 2;
}
else if (Cs >= 837) {
return 4;
}
}
else if (posBalance >= 60) {
if (SC < 1217) {
return 4;
}
else if (SC >= 1217) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (posBalance < 68) {
return 8;
}
else if (posBalance >= 68) {
if (Rs < 722) {
return 4;
}
else if (Rs >= 722) {
return 8;
}
}
}
}
}
}
else if (SC >= 1692) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
if (Rs < 937) {
if (Rs < 907) {
return 1;
}
else if (Rs >= 907) {
return 2;
}
}
else if (Rs >= 937) {
return 4;
}
}
}
}
else if (RsDiff >= 7) {
if (negBalance < 36) {
if (Cs < 925) {
return 8;
}
else if (Cs >= 925) {
return 1;
}
}
else if (negBalance >= 36) {
if (TSC < 9) {
if (CsDiff < 9) {
if (Cs < 961) {
return 8;
}
else if (Cs >= 961) {
if (Rs < 1114) {
return 2;
}
else if (Rs >= 1114) {
return 8;
}
}
}
else if (CsDiff >= 9) {
if (MVDiff < 57) {
return 4;
}
else if (MVDiff >= 57) {
return 8;
}
}
}
else if (TSC >= 9) {
if (MVDiff < 33) {
return 2;
}
else if (MVDiff >= 33) {
return 1;
}
}
}
}
}
else if (negBalance >= 52) {
if (ssDCval < 114) {
if (refDCval < 41) {
return 8;
}
else if (refDCval >= 41) {
if (AFD < 21) {
if (gchDC < 6) {
if (diffAFD < 2) {
if (diffMVdiffVal < -48) {
return 1;
}
else if (diffMVdiffVal >= -48) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
return 4;
}
}
}
}
else if (diffAFD >= 2) {
if (refDCval < 70) {
if (TSC < 9) {
if (Rs < 438) {
return 4;
}
else if (Rs >= 438) {
return 8;
}
}
else if (TSC >= 9) {
return 4;
}
}
else if (refDCval >= 70) {
return 4;
}
}
}
else if (gchDC >= 6) {
return 8;
}
}
else if (AFD >= 21) {
if (Rs < 896) {
if (Rs < 822) {
return 8;
}
else if (Rs >= 822) {
return 4;
}
}
else if (Rs >= 896) {
return 1;
}
}
}
}
else if (ssDCval >= 114) {
if (AFD < 16) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 1;
}
}
else if (AFD >= 16) {
if (Rs < 613) {
return 2;
}
else if (Rs >= 613) {
return 8;
}
}
}
}
}
else if (TSC >= 10) {
if (refDCval < 84) {
if (diffTSC < -3) {
return 1;
}
else if (diffTSC >= -3) {
if (MVDiff < 56) {
if (refDCval < 67) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (mcTcor < 61) {
return 4;
}
else if (mcTcor >= 61) {
if (Rs < 1571) {
return 1;
}
else if (Rs >= 1571) {
return 8;
}
}
}
}
else if (refDCval >= 67) {
if (ssDCval < 79) {
if (Cs < 1106) {
if (Rs < 2002) {
if (diffRsCsdiff < 57) {
return 4;
}
else if (diffRsCsdiff >= 57) {
return 2;
}
}
else if (Rs >= 2002) {
return 1;
}
}
else if (Cs >= 1106) {
return 2;
}
}
else if (ssDCval >= 79) {
return 8;
}
}
}
else if (MVDiff >= 56) {
if (Cs < 556) {
return 1;
}
else if (Cs >= 556) {
if (posBalance < 94) {
if (mcTcor < 82) {
if (Cs < 1153) {
if (refDCval < 40) {
return 2;
}
else if (refDCval >= 40) {
return 8;
}
}
else if (Cs >= 1153) {
if (Rs < 1211) {
return 4;
}
else if (Rs >= 1211) {
return 8;
}
}
}
else if (mcTcor >= 82) {
return 4;
}
}
else if (posBalance >= 94) {
if (Cs < 949) {
return 4;
}
else if (Cs >= 949) {
if (Rs < 810) {
return 2;
}
else if (Rs >= 810) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 84) {
if (CsDiff < 5) {
if (Cs < 620) {
if (MVDiff < 41) {
return 8;
}
else if (MVDiff >= 41) {
if (Cs < 538) {
return 2;
}
else if (Cs >= 538) {
return 4;
}
}
}
else if (Cs >= 620) {
if (mcTcor < 76) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
if (diffMVdiffVal < -34) {
return 1;
}
else if (diffMVdiffVal >= -34) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
else if (mcTcor >= 76) {
if (Rs < 2105) {
return 4;
}
else if (Rs >= 2105) {
return 2;
}
}
}
}
else if (CsDiff >= 5) {
if (gchDC < 2) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 4;
}
}
else if (gchDC >= 2) {
if (RsDiff < 10) {
if (mcTcor < 61) {
if (RsDiff < 5) {
if (Rs < 331) {
return 2;
}
else if (Rs >= 331) {
return 8;
}
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (mcTcor >= 61) {
if (diffMVdiffVal < 81) {
if (Cs < 695) {
return 4;
}
else if (Cs >= 695) {
return 2;
}
}
else if (diffMVdiffVal >= 81) {
return 1;
}
}
}
else if (RsDiff >= 10) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (AFD < 29) {
if (ssDCval < 85) {
return 1;
}
else if (ssDCval >= 85) {
return 8;
}
}
else if (AFD >= 29) {
if (Cs < 837) {
return 8;
}
else if (Cs >= 837) {
return 4;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 189) {
return 1;
}
}
else if (tcor >= 58) {
if (mcTcor < 76) {
if (diffAFD < -6) {
if (Rs < 697) {
return 4;
}
else if (Rs >= 697) {
return 1;
}
}
else if (diffAFD >= -6) {
return 8;
}
}
else if (mcTcor >= 76) {
if (tcor < 74) {
if (ssDCval < 146) {
if (refDCval < 49) {
if (mcTcor < 83) {
return 2;
}
else if (mcTcor >= 83) {
if (Rs < 1019) {
return 2;
}
else if (Rs >= 1019) {
return 8;
}
}
}
else if (refDCval >= 49) {
if (Cs < 690) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (Rs < 1441) {
return 2;
}
else if (Rs >= 1441) {
return 1;
}
}
}
else if (Cs >= 690) {
if (diffAFD < 2) {
if (tcor < 63) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Rs < 402) {
return 8;
}
else if (Rs >= 402) {
if (mu_mv_mag_sq < 39) {
return 4;
}
else if (mu_mv_mag_sq >= 39) {
return 1;
}
}
}
}
}
else if (tcor >= 63) {
if (ssDCval < 98) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
if (tcor < 73) {
return 8;
}
else if (tcor >= 73) {
if (Rs < 1396) {
return 4;
}
else if (Rs >= 1396) {
return 8;
}
}
}
}
else if (ssDCval >= 98) {
if (SC < 2836) {
return 4;
}
else if (SC >= 2836) {
return 8;
}
}
}
}
else if (diffAFD >= 2) {
if (diffRsCsdiff < 83) {
if (diffAFD < 5) {
if (RsCsDiff < 161) {
if (Cs < 1205) {
return 2;
}
else if (Cs >= 1205) {
return 4;
}
}
else if (RsCsDiff >= 161) {
if (SC < 2919) {
return 2;
}
else if (SC >= 2919) {
return 8;
}
}
}
else if (diffAFD >= 5) {
if (MVDiff < 42) {
return 1;
}
else if (MVDiff >= 42) {
return 8;
}
}
}
else if (diffRsCsdiff >= 83) {
return 4;
}
}
}
}
}
else if (ssDCval >= 146) {
return 2;
}
}
else if (tcor >= 74) {
if (negBalance < 33) {
if (Rs < 750) {
return 8;
}
else if (Rs >= 750) {
if (Rs < 1488) {
return 2;
}
else if (Rs >= 1488) {
return 1;
}
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
}
}
else if (mcTcor >= 84) {
if (negBalance < 44) {
if (Rs < 1490) {
if (MVDiff < 19) {
if (Cs < 1059) {
return 8;
}
else if (Cs >= 1059) {
if (Rs < 1253) {
return 1;
}
else if (Rs >= 1253) {
return 8;
}
}
}
else if (MVDiff >= 19) {
if (Cs < 1224) {
if (SC < 2006) {
if (refDCval < 65) {
if (refDCval < 54) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (diffTSC < 4) {
if (Rs < 680) {
return 8;
}
else if (Rs >= 680) {
return 2;
}
}
else if (diffTSC >= 4) {
return 8;
}
}
}
else if (refDCval >= 54) {
if (posBalance < 66) {
return 1;
}
else if (posBalance >= 66) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
else if (refDCval >= 65) {
if (mu_mv_mag_sq < 7) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Cs < 667) {
return 8;
}
else if (Cs >= 667) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (gchDC < 6) {
if (ssDCval < 105) {
return 2;
}
else if (ssDCval >= 105) {
if (Rs < 674) {
return 2;
}
else if (Rs >= 674) {
return 8;
}
}
}
else if (gchDC >= 6) {
if (ssDCval < 82) {
return 2;
}
else if (ssDCval >= 82) {
return 1;
}
}
}
}
}
else if (SC >= 2006) {
if (Rs < 1459) {
if (diffAFD < 1) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
return 8;
}
}
else if (diffAFD >= 1) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (Rs >= 1459) {
return 1;
}
}
}
else if (Cs >= 1224) {
if (Rs < 478) {
return 8;
}
else if (Rs >= 478) {
return 4;
}
}
}
}
else if (Rs >= 1490) {
if (mcTcor < 96) {
if (ssDCval < 75) {
if (diffAFD < 7) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (MVDiff < 29) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 4;
}
}
else if (MVDiff >= 29) {
if (Rs < 1526) {
return 8;
}
else if (Rs >= 1526) {
return 2;
}
}
}
}
else if (diffAFD >= 7) {
if (Rs < 2175) {
return 1;
}
else if (Rs >= 2175) {
return 4;
}
}
}
else if (ssDCval >= 75) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
if (Cs < 836) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
return 2;
}
}
}
}
else if (Cs >= 836) {
if (SC < 2970) {
if (posBalance < 54) {
if (Rs < 1606) {
return 2;
}
else if (Rs >= 1606) {
return 8;
}
}
else if (posBalance >= 54) {
return 2;
}
}
else if (SC >= 2970) {
if (mcTcor < 94) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 2;
}
}
else if (mcTcor >= 94) {
if (Rs < 2232) {
return 2;
}
else if (Rs >= 2232) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 2117) {
return 2;
}
else if (Rs >= 2117) {
return 4;
}
}
}
}
else if (negBalance >= 44) {
if (ssDCval < 83) {
if (Rs < 1319) {
if (mu_mv_mag_sq < 3) {
if (Cs < 919) {
if (SC < 1149) {
return 2;
}
else if (SC >= 1149) {
if (gchDC < 2) {
if (ssDCval < 60) {
if (Rs < 793) {
return 8;
}
else if (Rs >= 793) {
return 4;
}
}
else if (ssDCval >= 60) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Cs < 829) {
return 4;
}
else if (Cs >= 829) {
return 1;
}
}
}
}
else if (gchDC >= 2) {
if (Cs < 874) {
if (diffMVdiffVal < 2) {
if (Cs < 740) {
if (Cs < 689) {
if (Cs < 578) {
return 4;
}
else if (Cs >= 578) {
return 2;
}
}
else if (Cs >= 689) {
return 4;
}
}
else if (Cs >= 740) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
if (mcTcor < 94) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
}
else if (Cs >= 874) {
return 8;
}
}
}
}
else if (Cs >= 919) {
if (negBalance < 60) {
if (tcor < 77) {
if (negBalance < 49) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 4;
}
}
else if (negBalance >= 49) {
return 2;
}
}
else if (tcor >= 77) {
if (diffTSC < 1) {
if (TSC < 11) {
if (MVDiff < 20) {
return 1;
}
else if (MVDiff >= 20) {
return 2;
}
}
else if (TSC >= 11) {
return 1;
}
}
else if (diffTSC >= 1) {
if (diffMVdiffVal < 1) {
return 1;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 871) {
return 4;
}
else if (Rs >= 871) {
return 2;
}
}
}
}
}
else if (negBalance >= 60) {
if (diffRsCsdiff < 28) {
if (MVDiff < 50) {
return 4;
}
else if (MVDiff >= 50) {
if (Cs < 1014) {
return 4;
}
else if (Cs >= 1014) {
return 2;
}
}
}
else if (diffRsCsdiff >= 28) {
if (Rs < 953) {
return 1;
}
else if (Rs >= 953) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (refDCval < 60) {
if (Rs < 871) {
if (MVDiff < 43) {
if (negBalance < 52) {
return 4;
}
else if (negBalance >= 52) {
return 8;
}
}
else if (MVDiff >= 43) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 368) {
return 8;
}
else if (Rs >= 368) {
return 1;
}
}
}
}
else if (Rs >= 871) {
if (MVDiff < 30) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (MVDiff >= 30) {
if (Rs < 892) {
return 2;
}
else if (Rs >= 892) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
if (Cs < 860) {
return 8;
}
else if (Cs >= 860) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 60) {
if (tcor < 79) {
if (diffMVdiffVal < -15) {
if (gchDC < 3) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (mcTcor < 90) {
return 2;
}
else if (mcTcor >= 90) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (CsDiff < 10) {
return 1;
}
else if (CsDiff >= 10) {
return 8;
}
}
}
else if (diffMVdiffVal >= -15) {
if (RsCsDiff < 38) {
if (AFD < 15) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (AFD >= 15) {
if (mcTcor < 91) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (Rs < 343) {
return 2;
}
else if (Rs >= 343) {
return 1;
}
}
}
else if (mcTcor >= 91) {
return 2;
}
}
}
else if (RsCsDiff >= 38) {
if (RsDiff < 12) {
if (negBalance < 56) {
if (gchDC < 3) {
if (MVDiff < 47) {
if (Cs < 1179) {
if (mu_mv_mag_sq < 13) {
return 4;
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 1278) {
return 2;
}
else if (Rs >= 1278) {
return 4;
}
}
}
else if (Cs >= 1179) {
return 2;
}
}
else if (MVDiff >= 47) {
if (refDCval < 74) {
if (Rs < 774) {
return 4;
}
else if (Rs >= 774) {
return 1;
}
}
else if (refDCval >= 74) {
return 4;
}
}
}
else if (gchDC >= 3) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (RsCsDiff < 138) {
return 2;
}
else if (RsCsDiff >= 138) {
return 1;
}
}
}
}
else if (negBalance >= 56) {
if (MVDiff < 24) {
if (diffAFD < 2) {
if (Rs < 835) {
if (Rs < 405) {
return 2;
}
else if (Rs >= 405) {
return 8;
}
}
else if (Rs >= 835) {
return 4;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (MVDiff >= 24) {
if (diffTSC < -2) {
if (mu_mv_mag_sq < 12) {
return 8;
}
else if (mu_mv_mag_sq >= 12) {
if (Rs < 846) {
return 2;
}
else if (Rs >= 846) {
return 4;
}
}
}
else if (diffTSC >= -2) {
if (diffMVdiffVal < 16) {
return 4;
}
else if (diffMVdiffVal >= 16) {
if (Rs < 1144) {
if (Rs < 727) {
if (Cs < 997) {
return 8;
}
else if (Cs >= 997) {
return 4;
}
}
else if (Rs >= 727) {
return 4;
}
}
else if (Rs >= 1144) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 12) {
if (SC < 2217) {
return 2;
}
else if (SC >= 2217) {
return 4;
}
}
}
}
}
else if (tcor >= 79) {
if (mcTcor < 91) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
else if (mcTcor >= 91) {
if (TSC < 7) {
if (Rs < 816) {
return 4;
}
else if (Rs >= 816) {
return 8;
}
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1319) {
if (AFD < 19) {
if (Cs < 630) {
if (TSC < 8) {
if (Rs < 2038) {
return 8;
}
else if (Rs >= 2038) {
return 2;
}
}
else if (TSC >= 8) {
if (Rs < 1585) {
return 4;
}
else if (Rs >= 1585) {
return 2;
}
}
}
else if (Cs >= 630) {
if (MVDiff < 23) {
if (posBalance < 50) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (posBalance >= 50) {
if (refDCval < 76) {
if (RsDiff < 10) {
if (TSC < 8) {
if (Rs < 1717) {
return 8;
}
else if (Rs >= 1717) {
return 4;
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (RsDiff >= 10) {
if (Cs < 1113) {
return 2;
}
else if (Cs >= 1113) {
return 8;
}
}
}
else if (refDCval >= 76) {
if (Rs < 1447) {
return 2;
}
else if (Rs >= 1447) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
}
}
else if (MVDiff >= 23) {
if (ssDCval < 64) {
if (CsDiff < 6) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
if (MVDiff < 45) {
return 4;
}
else if (MVDiff >= 45) {
if (Rs < 1413) {
return 4;
}
else if (Rs >= 1413) {
return 1;
}
}
}
}
else if (CsDiff >= 6) {
if (Cs < 1022) {
return 2;
}
else if (Cs >= 1022) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
}
}
else if (ssDCval >= 64) {
if (diffRsCsdiff < -25) {
if (RsDiff < 9) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
else if (RsDiff >= 9) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 4;
}
}
}
else if (diffRsCsdiff >= -25) {
if (mcTcor < 92) {
if (posBalance < 70) {
return 8;
}
else if (posBalance >= 70) {
return 2;
}
}
else if (mcTcor >= 92) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
return 4;
}
}
}
}
}
}
}
else if (AFD >= 19) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (RsDiff < 14) {
if (AFD < 30) {
if (mu_mv_mag_sq < 7) {
if (negBalance < 58) {
if (mu_mv_mag_sq < 5) {
if (diffAFD < 4) {
return 4;
}
else if (diffAFD >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 1737) {
return 2;
}
else if (Rs >= 1737) {
return 8;
}
}
}
else if (negBalance >= 58) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (mu_mv_mag_sq < 17) {
if (Cs < 1079) {
if (posBalance < 57) {
if (Rs < 1673) {
if (Rs < 1594) {
return 8;
}
else if (Rs >= 1594) {
return 2;
}
}
else if (Rs >= 1673) {
return 1;
}
}
else if (posBalance >= 57) {
if (Cs < 761) {
return 4;
}
else if (Cs >= 761) {
return 8;
}
}
}
else if (Cs >= 1079) {
if (mu_mv_mag_sq < 11) {
if (SC < 2502) {
return 8;
}
else if (SC >= 2502) {
return 4;
}
}
else if (mu_mv_mag_sq >= 11) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (diffMVdiffVal < -5) {
return 1;
}
else if (diffMVdiffVal >= -5) {
return 4;
}
}
}
}
else if (AFD >= 30) {
if (Cs < 1127) {
return 2;
}
else if (Cs >= 1127) {
return 4;
}
}
}
else if (RsDiff >= 14) {
if (RsCsDiff < 248) {
return 8;
}
else if (RsCsDiff >= 248) {
if (AFD < 33) {
return 2;
}
else if (AFD >= 33) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 83) {
if (MVDiff < 31) {
if (ssDCval < 139) {
if (posBalance < 55) {
if (Cs < 1157) {
if (mcTcor < 95) {
if (diffTSC < -2) {
return 8;
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (mcTcor >= 95) {
if (AFD < 21) {
if (Cs < 826) {
return 8;
}
else if (Cs >= 826) {
return 1;
}
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (Cs >= 1157) {
if (negBalance < 58) {
if (Rs < 1629) {
return 4;
}
else if (Rs >= 1629) {
return 2;
}
}
else if (negBalance >= 58) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (negBalance < 51) {
if (RsCsDiff < 87) {
if (mu_mv_mag_sq < 7) {
if (Rs < 976) {
return 4;
}
else if (Rs >= 976) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 676) {
return 1;
}
else if (Cs >= 676) {
return 8;
}
}
}
else if (RsCsDiff >= 87) {
if (Cs < 925) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
return 1;
}
}
else if (Cs >= 925) {
if (tcor < 78) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 4;
}
}
else if (tcor >= 78) {
return 2;
}
}
}
}
else if (negBalance >= 51) {
if (negBalance < 59) {
if (CsDiff < 7) {
if (Cs < 698) {
if (tcor < 80) {
if (MVDiff < 24) {
if (diffTSC < -1) {
if (Rs < 648) {
return 4;
}
else if (Rs >= 648) {
return 8;
}
}
else if (diffTSC >= -1) {
if (Cs < 575) {
return 8;
}
else if (Cs >= 575) {
return 4;
}
}
}
else if (MVDiff >= 24) {
return 8;
}
}
else if (tcor >= 80) {
return 1;
}
}
else if (Cs >= 698) {
if (diffMVdiffVal < 2) {
if (diffRsCsdiff < -15) {
if (Rs < 538) {
return 2;
}
else if (Rs >= 538) {
return 8;
}
}
else if (diffRsCsdiff >= -15) {
if (AFD < 22) {
if (negBalance < 54) {
if (Cs < 1112) {
return 1;
}
else if (Cs >= 1112) {
return 4;
}
}
else if (negBalance >= 54) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (AFD >= 22) {
return 1;
}
}
}
else if (diffMVdiffVal >= 2) {
if (ssDCval < 108) {
if (mcTcor < 89) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
return 4;
}
}
else if (mcTcor >= 89) {
return 8;
}
}
else if (ssDCval >= 108) {
return 4;
}
}
}
}
else if (CsDiff >= 7) {
if (posBalance < 60) {
if (Cs < 1159) {
return 1;
}
else if (Cs >= 1159) {
if (Rs < 1816) {
return 2;
}
else if (Rs >= 1816) {
return 4;
}
}
}
else if (posBalance >= 60) {
if (Cs < 1273) {
if (diffMVdiffVal < 6) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (Cs >= 1273) {
if (RsCsDiff < 166) {
if (Cs < 1316) {
return 8;
}
else if (Cs >= 1316) {
return 1;
}
}
else if (RsCsDiff >= 166) {
return 4;
}
}
}
}
}
else if (negBalance >= 59) {
if (RsDiff < 9) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
if (Cs < 1122) {
return 2;
}
else if (Cs >= 1122) {
if (Rs < 1386) {
return 4;
}
else if (Rs >= 1386) {
if (Rs < 1774) {
return 1;
}
else if (Rs >= 1774) {
return 2;
}
}
}
}
}
else if (RsDiff >= 9) {
if (AFD < 32) {
if (Rs < 1997) {
if (diffAFD < -2) {
if (Cs < 819) {
return 4;
}
else if (Cs >= 819) {
return 8;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Rs >= 1997) {
return 1;
}
}
else if (AFD >= 32) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 139) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
return 8;
}
}
}
}
else if (MVDiff >= 31) {
if (Rs < 1041) {
if (Rs < 518) {
if (gchDC < 5) {
if (ssDCval < 97) {
if (Cs < 575) {
return 4;
}
else if (Cs >= 575) {
if (Rs < 301) {
return 8;
}
else if (Rs >= 301) {
return 1;
}
}
}
else if (ssDCval >= 97) {
return 2;
}
}
else if (gchDC >= 5) {
if (ssDCval < 85) {
return 4;
}
else if (ssDCval >= 85) {
return 1;
}
}
}
else if (Rs >= 518) {
if (Cs < 602) {
return 8;
}
else if (Cs >= 602) {
if (AFD < 22) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
if (Rs < 749) {
return 1;
}
else if (Rs >= 749) {
return 8;
}
}
}
else if (AFD >= 22) {
if (diffMVdiffVal < -45) {
if (Cs < 798) {
return 2;
}
else if (Cs >= 798) {
if (Rs < 860) {
return 4;
}
else if (Rs >= 860) {
return 1;
}
}
}
else if (diffMVdiffVal >= -45) {
return 2;
}
}
}
}
}
else if (Rs >= 1041) {
if (AFD < 30) {
if (diffMVdiffVal < 14) {
if (MVDiff < 52) {
if (CsDiff < 9) {
if (tcor < 68) {
if (RsCsDiff < 148) {
if (Rs < 1199) {
return 4;
}
else if (Rs >= 1199) {
return 1;
}
}
else if (RsCsDiff >= 148) {
if (Cs < 1003) {
return 4;
}
else if (Cs >= 1003) {
if (Rs < 2029) {
return 8;
}
else if (Rs >= 2029) {
return 2;
}
}
}
}
else if (tcor >= 68) {
if (Cs < 898) {
if (RsCsDiff < 141) {
if (gchDC < 2) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
return 8;
}
}
else if (gchDC >= 2) {
if (Rs < 1993) {
return 8;
}
else if (Rs >= 1993) {
return 1;
}
}
}
else if (RsCsDiff >= 141) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
return 1;
}
}
}
else if (Cs >= 898) {
return 1;
}
}
}
else if (CsDiff >= 9) {
return 8;
}
}
else if (MVDiff >= 52) {
if (refDCval < 132) {
if (mcTcor < 94) {
if (tcor < 64) {
if (Cs < 1044) {
return 1;
}
else if (Cs >= 1044) {
return 8;
}
}
else if (tcor >= 64) {
if (SC < 3383) {
return 8;
}
else if (SC >= 3383) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 1587) {
return 4;
}
else if (Rs >= 1587) {
return 1;
}
}
}
else if (refDCval >= 132) {
if (Rs < 1675) {
return 4;
}
else if (Rs >= 1675) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 14) {
if (ssDCval < 100) {
if (AFD < 20) {
if (Rs < 2017) {
return 2;
}
else if (Rs >= 2017) {
return 8;
}
}
else if (AFD >= 20) {
return 4;
}
}
else if (ssDCval >= 100) {
if (Rs < 1800) {
return 8;
}
else if (Rs >= 1800) {
if (Cs < 845) {
return 8;
}
else if (Cs >= 845) {
return 2;
}
}
}
}
}
else if (AFD >= 30) {
if (mcTcor < 96) {
if (MVDiff < 157) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 1;
}
}
else if (MVDiff >= 157) {
return 2;
}
}
else if (mcTcor >= 96) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (Cs < 755) {
return 8;
}
else if (Cs >= 755) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (tcor >= 83) {
if (Cs < 872) {
if (RsDiff < 6) {
if (mu_mv_mag_sq < 3) {
if (negBalance < 51) {
if (Cs < 675) {
if (Rs < 522) {
return 2;
}
else if (Rs >= 522) {
return 1;
}
}
else if (Cs >= 675) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
}
else if (negBalance >= 51) {
if (TSC < 7) {
if (tcor < 87) {
return 2;
}
else if (tcor >= 87) {
return 1;
}
}
else if (TSC >= 7) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffMVdiffVal < 19) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (diffTSC < 2) {
if (Cs < 752) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 4;
}
}
else if (Cs >= 752) {
if (Rs < 922) {
return 4;
}
else if (Rs >= 922) {
if (Cs < 867) {
return 8;
}
else if (Cs >= 867) {
return 4;
}
}
}
}
else if (diffTSC >= 2) {
if (Rs < 1138) {
return 1;
}
else if (Rs >= 1138) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 19) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (refDCval < 76) {
if (RsDiff < 8) {
if (diffRsCsdiff < -4) {
if (Rs < 1508) {
return 4;
}
else if (Rs >= 1508) {
return 2;
}
}
else if (diffRsCsdiff >= -4) {
if (Cs < 736) {
return 1;
}
else if (Cs >= 736) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (diffRsCsdiff < -323) {
return 4;
}
else if (diffRsCsdiff >= -323) {
return 2;
}
}
}
else if (refDCval >= 76) {
if (ssDCval < 106) {
if (RsDiff < 10) {
if (MVDiff < 39) {
if (diffAFD < -5) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
return 1;
}
}
else if (diffAFD >= -5) {
if (Cs < 822) {
if (diffAFD < -2) {
if (Cs < 567) {
return 1;
}
else if (Cs >= 567) {
return 2;
}
}
else if (diffAFD >= -2) {
if (refDCval < 102) {
if (Rs < 855) {
return 8;
}
else if (Rs >= 855) {
return 1;
}
}
else if (refDCval >= 102) {
return 2;
}
}
}
else if (Cs >= 822) {
if (Rs < 1684) {
return 8;
}
else if (Rs >= 1684) {
return 4;
}
}
}
}
else if (MVDiff >= 39) {
if (AFD < 16) {
if (Rs < 2046) {
return 2;
}
else if (Rs >= 2046) {
return 1;
}
}
else if (AFD >= 16) {
if (Cs < 825) {
return 8;
}
else if (Cs >= 825) {
return 2;
}
}
}
}
else if (RsDiff >= 10) {
if (diffRsCsdiff < -58) {
return 4;
}
else if (diffRsCsdiff >= -58) {
return 2;
}
}
}
else if (ssDCval >= 106) {
if (mu_mv_mag_sq < 3) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (ssDCval < 165) {
return 1;
}
else if (ssDCval >= 165) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 46) {
if (Rs < 2027) {
if (Cs < 626) {
return 1;
}
else if (Cs >= 626) {
return 2;
}
}
else if (Rs >= 2027) {
return 1;
}
}
else if (negBalance >= 46) {
if (posBalance < 64) {
if (Rs < 1537) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 1;
}
}
else if (Rs >= 1537) {
return 1;
}
}
else if (posBalance >= 64) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 872) {
if (ssDCval < 115) {
if (Rs < 990) {
if (tcor < 85) {
if (gchDC < 1) {
if (Cs < 1030) {
return 4;
}
else if (Cs >= 1030) {
return 2;
}
}
else if (gchDC >= 1) {
if (RsCsDiff < 49) {
if (Rs < 611) {
return 2;
}
else if (Rs >= 611) {
return 4;
}
}
else if (RsCsDiff >= 49) {
return 1;
}
}
}
else if (tcor >= 85) {
if (diffMVdiffVal < -6) {
if (AFD < 19) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
else if (AFD >= 19) {
return 4;
}
}
else if (diffMVdiffVal >= -6) {
if (negBalance < 50) {
if (Cs < 1278) {
return 1;
}
else if (Cs >= 1278) {
return 8;
}
}
else if (negBalance >= 50) {
if (Rs < 466) {
return 1;
}
else if (Rs >= 466) {
return 8;
}
}
}
}
}
else if (Rs >= 990) {
if (RsCsDiff < 87) {
if (TSC < 9) {
if (tcor < 90) {
if (negBalance < 42) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (negBalance >= 42) {
if (negBalance < 54) {
if (mu_mv_mag_sq < 2) {
if (SC < 2840) {
return 4;
}
else if (SC >= 2840) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (mu_mv_mag_sq < 50) {
if (Cs < 933) {
return 4;
}
else if (Cs >= 933) {
return 8;
}
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 1375) {
return 8;
}
else if (Rs >= 1375) {
return 4;
}
}
}
}
else if (negBalance >= 54) {
if (Rs < 1254) {
return 2;
}
else if (Rs >= 1254) {
return 8;
}
}
}
}
else if (tcor >= 90) {
if (Cs < 919) {
if (Rs < 1053) {
return 8;
}
else if (Rs >= 1053) {
if (Rs < 1645) {
return 1;
}
else if (Rs >= 1645) {
return 2;
}
}
}
else if (Cs >= 919) {
return 8;
}
}
}
else if (TSC >= 9) {
if (CsDiff < 4) {
if (Rs < 1637) {
return 8;
}
else if (Rs >= 1637) {
if (Cs < 1006) {
return 4;
}
else if (Cs >= 1006) {
return 1;
}
}
}
else if (CsDiff >= 4) {
if (diffTSC < 2) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
return 2;
}
}
}
else if (diffTSC >= 2) {
return 4;
}
}
}
}
else if (RsCsDiff >= 87) {
if (Cs < 1031) {
if (MVDiff < 16) {
if (Rs < 1963) {
return 8;
}
else if (Rs >= 1963) {
return 2;
}
}
else if (MVDiff >= 16) {
if (Rs < 2020) {
return 1;
}
else if (Rs >= 2020) {
return 4;
}
}
}
else if (Cs >= 1031) {
if (posBalance < 66) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (ssDCval < 113) {
if (Cs < 1319) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (Cs >= 1319) {
if (Cs < 1326) {
return 1;
}
else if (Cs >= 1326) {
return 8;
}
}
}
else if (ssDCval >= 113) {
return 1;
}
}
}
else if (posBalance >= 66) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 115) {
if (MVDiff < 10) {
if (gchDC < 1) {
if (Rs < 1484) {
return 4;
}
else if (Rs >= 1484) {
return 8;
}
}
else if (gchDC >= 1) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
if (tcor < 85) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
else if (tcor >= 85) {
return 2;
}
}
}
}
else if (MVDiff >= 10) {
if (posBalance < 56) {
if (AFD < 17) {
if (SC < 2275) {
if (Rs < 1032) {
return 8;
}
else if (Rs >= 1032) {
return 2;
}
}
else if (SC >= 2275) {
return 8;
}
}
else if (AFD >= 17) {
return 1;
}
}
else if (posBalance >= 56) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (diffTSC < 2) {
if (Rs < 2163) {
if (CsDiff < 6) {
if (Rs < 1532) {
return 1;
}
else if (Rs >= 1532) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
}
else if (CsDiff >= 6) {
return 1;
}
}
else if (Rs >= 2163) {
return 2;
}
}
else if (diffTSC >= 2) {
if (diffRsCsdiff < -4) {
if (Rs < 1792) {
return 2;
}
else if (Rs >= 1792) {
return 1;
}
}
else if (diffRsCsdiff >= -4) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (SC < 2155) {
if (Rs < 964) {
if (AFD < 1) {
if (Rs < 779) {
return 1;
}
else if (Rs >= 779) {
return 2;
}
}
else if (AFD >= 1) {
if (MVDiff < 1) {
if (negBalance < 26) {
if (SC < 1654) {
if (Rs < 830) {
if (Rs < 366) {
if (Cs < 768) {
return 1;
}
else if (Cs >= 768) {
return 2;
}
}
else if (Rs >= 366) {
if (ssDCval < 72) {
if (ssDCval < 49) {
if (Rs < 647) {
if (negBalance < 20) {
return 4;
}
else if (negBalance >= 20) {
return 1;
}
}
else if (Rs >= 647) {
return 8;
}
}
else if (ssDCval >= 49) {
if (Cs < 657) {
if (ssDCval < 64) {
if (Rs < 531) {
return 2;
}
else if (Rs >= 531) {
return 8;
}
}
else if (ssDCval >= 64) {
return 2;
}
}
else if (Cs >= 657) {
if (SC < 1244) {
return 4;
}
else if (SC >= 1244) {
if (tcor < 99) {
if (AFD < 2) {
if (Cs < 732) {
return 8;
}
else if (Cs >= 732) {
return 4;
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (tcor >= 99) {
if (Rs < 709) {
return 8;
}
else if (Rs >= 709) {
if (Rs < 715) {
return 4;
}
else if (Rs >= 715) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 72) {
if (ssDCval < 79) {
if (Rs < 627) {
return 8;
}
else if (Rs >= 627) {
return 1;
}
}
else if (ssDCval >= 79) {
return 2;
}
}
}
}
else if (Rs >= 830) {
if (ssDCval < 64) {
if (SC < 1496) {
return 8;
}
else if (SC >= 1496) {
return 1;
}
}
else if (ssDCval >= 64) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
}
else if (SC >= 1654) {
if (RsDiff < 1) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
if (SC < 1680) {
if (SC < 1674) {
return 8;
}
else if (SC >= 1674) {
return 4;
}
}
else if (SC >= 1680) {
return 1;
}
}
}
else if (RsDiff >= 1) {
if (tcor < 97) {
if (Rs < 940) {
return 4;
}
else if (Rs >= 940) {
return 8;
}
}
else if (tcor >= 97) {
if (Cs < 763) {
return 1;
}
else if (Cs >= 763) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 26) {
if (ssDCval < 81) {
if (ssDCval < 59) {
if (refDCval < 51) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (Rs < 917) {
if (ssDCval < 43) {
return 4;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (Rs >= 917) {
return 8;
}
}
}
else if (refDCval >= 51) {
if (Rs < 660) {
if (Cs < 612) {
return 8;
}
else if (Cs >= 612) {
if (tcor < 97) {
if (Rs < 512) {
return 1;
}
else if (Rs >= 512) {
return 4;
}
}
else if (tcor >= 97) {
return 4;
}
}
}
else if (Rs >= 660) {
if (SC < 1534) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (SC >= 1534) {
return 8;
}
}
}
}
else if (ssDCval >= 59) {
if (Rs < 621) {
return 8;
}
else if (Rs >= 621) {
if (Rs < 804) {
if (ssDCval < 67) {
if (mu_mv_mag_sq < 111) {
if (Rs < 672) {
return 1;
}
else if (Rs >= 672) {
return 8;
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 899) {
return 4;
}
else if (Cs >= 899) {
return 1;
}
}
}
else if (ssDCval >= 67) {
if (AFD < 2) {
if (Rs < 745) {
return 8;
}
else if (Rs >= 745) {
return 1;
}
}
else if (AFD >= 2) {
if (Cs < 619) {
if (Rs < 695) {
return 1;
}
else if (Rs >= 695) {
if (Rs < 768) {
return 4;
}
else if (Rs >= 768) {
return 2;
}
}
}
else if (Cs >= 619) {
return 1;
}
}
}
}
else if (Rs >= 804) {
if (Cs < 590) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 590) {
if (Cs < 1143) {
if (tcor < 99) {
if (mu_mv_mag_sq < 121) {
return 8;
}
else if (mu_mv_mag_sq >= 121) {
return 4;
}
}
else if (tcor >= 99) {
if (Cs < 875) {
return 4;
}
else if (Cs >= 875) {
if (Cs < 932) {
return 1;
}
else if (Cs >= 932) {
return 8;
}
}
}
}
else if (Cs >= 1143) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (Rs < 644) {
return 2;
}
else if (Rs >= 644) {
return 4;
}
}
}
}
else if (MVDiff >= 1) {
if (refDCval < 62) {
if (Rs < 805) {
if (refDCval < 37) {
if (Rs < 507) {
return 4;
}
else if (Rs >= 507) {
return 2;
}
}
else if (refDCval >= 37) {
if (negBalance < 26) {
if (mu_mv_mag_sq < 78) {
if (Rs < 770) {
if (Cs < 908) {
return 4;
}
else if (Cs >= 908) {
return 1;
}
}
else if (Rs >= 770) {
return 1;
}
}
else if (mu_mv_mag_sq >= 78) {
if (SC < 1402) {
if (SC < 1013) {
return 1;
}
else if (SC >= 1013) {
if (Rs < 533) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 2;
}
}
else if (Rs >= 533) {
return 4;
}
}
}
else if (SC >= 1402) {
return 1;
}
}
}
else if (negBalance >= 26) {
if (AFD < 8) {
if (MVDiff < 2) {
if (ssDCval < 51) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 4;
}
}
else if (ssDCval >= 51) {
if (Cs < 1050) {
if (Rs < 387) {
return 4;
}
else if (Rs >= 387) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
}
else if (Cs >= 1050) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (MVDiff < 11) {
if (tcor < 92) {
if (Rs < 708) {
return 8;
}
else if (Rs >= 708) {
return 2;
}
}
else if (tcor >= 92) {
if (SC < 1106) {
if (AFD < 4) {
if (Rs < 395) {
if (Rs < 211) {
return 4;
}
else if (Rs >= 211) {
return 1;
}
}
else if (Rs >= 395) {
if (Rs < 523) {
return 2;
}
else if (Rs >= 523) {
return 4;
}
}
}
else if (AFD >= 4) {
if (Cs < 678) {
return 4;
}
else if (Cs >= 678) {
return 8;
}
}
}
else if (SC >= 1106) {
if (posBalance < 34) {
if (Rs < 618) {
return 1;
}
else if (Rs >= 618) {
return 8;
}
}
else if (posBalance >= 34) {
if (diffMVdiffVal < -5) {
return 8;
}
else if (diffMVdiffVal >= -5) {
if (Rs < 762) {
return 4;
}
else if (Rs >= 762) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 11) {
if (ssDCval < 55) {
if (ssDCval < 53) {
if (AFD < 7) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
return 8;
}
}
else if (AFD >= 7) {
if (Rs < 750) {
return 1;
}
else if (Rs >= 750) {
return 4;
}
}
}
else if (ssDCval >= 53) {
return 8;
}
}
else if (ssDCval >= 55) {
return 1;
}
}
}
}
else if (AFD >= 8) {
if (SC < 1433) {
if (RsCsDiff < 23) {
if (SC < 1355) {
return 1;
}
else if (SC >= 1355) {
return 4;
}
}
else if (RsCsDiff >= 23) {
if (Cs < 609) {
return 2;
}
else if (Cs >= 609) {
return 4;
}
}
}
else if (SC >= 1433) {
if (Rs < 741) {
return 2;
}
else if (Rs >= 741) {
return 8;
}
}
}
}
}
}
else if (Rs >= 805) {
if (ssDCval < 46) {
if (Cs < 566) {
return 1;
}
else if (Cs >= 566) {
return 8;
}
}
else if (ssDCval >= 46) {
if (ssDCval < 60) {
if (posBalance < 33) {
return 4;
}
else if (posBalance >= 33) {
if (Cs < 771) {
return 1;
}
else if (Cs >= 771) {
if (negBalance < 26) {
return 4;
}
else if (negBalance >= 26) {
if (Cs < 1029) {
return 1;
}
else if (Cs >= 1029) {
if (Rs < 882) {
return 1;
}
else if (Rs >= 882) {
if (Rs < 925) {
return 8;
}
else if (Rs >= 925) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 60) {
if (AFD < 7) {
if (Cs < 735) {
return 1;
}
else if (Cs >= 735) {
return 4;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
}
}
else if (refDCval >= 62) {
if (mu_mv_mag_sq < 115) {
if (Cs < 821) {
if (refDCval < 84) {
if (Rs < 438) {
if (MVDiff < 6) {
if (Rs < 162) {
return 1;
}
else if (Rs >= 162) {
return 4;
}
}
else if (MVDiff >= 6) {
if (SC < 989) {
if (mu_mv_mag_sq < 40) {
return 1;
}
else if (mu_mv_mag_sq >= 40) {
if (Cs < 544) {
return 4;
}
else if (Cs >= 544) {
if (Rs < 316) {
return 8;
}
else if (Rs >= 316) {
return 1;
}
}
}
}
else if (SC >= 989) {
if (Cs < 747) {
return 2;
}
else if (Cs >= 747) {
return 1;
}
}
}
}
else if (Rs >= 438) {
if (Rs < 724) {
if (mu_mv_mag_sq < 3) {
if (negBalance < 43) {
if (diffMVdiffVal < -6) {
if (Rs < 565) {
return 1;
}
else if (Rs >= 565) {
return 2;
}
}
else if (diffMVdiffVal >= -6) {
if (Cs < 792) {
return 4;
}
else if (Cs >= 792) {
return 2;
}
}
}
else if (negBalance >= 43) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
if (mu_mv_mag_sq < 99) {
if (SC < 1330) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
return 1;
}
}
}
else if (SC >= 1330) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (Rs < 690) {
if (Rs < 640) {
return 8;
}
else if (Rs >= 640) {
if (Rs < 672) {
return 1;
}
else if (Rs >= 672) {
return 8;
}
}
}
else if (Rs >= 690) {
if (Rs < 709) {
return 4;
}
else if (Rs >= 709) {
return 1;
}
}
}
}
}
}
else if (Rs >= 724) {
if (mcTcor < 98) {
if (ssDCval < 68) {
if (SC < 1451) {
if (SC < 1420) {
return 8;
}
else if (SC >= 1420) {
return 1;
}
}
else if (SC >= 1451) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
}
else if (ssDCval >= 68) {
if (Rs < 915) {
if (diffTSC < -2) {
return 8;
}
else if (diffTSC >= -2) {
return 4;
}
}
else if (Rs >= 915) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (posBalance < 42) {
if (diffMVdiffVal < 1) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (posBalance >= 42) {
if (Rs < 766) {
return 4;
}
else if (Rs >= 766) {
if (Cs < 546) {
return 8;
}
else if (Cs >= 546) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 84) {
if (Rs < 646) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (Cs < 673) {
if (Cs < 598) {
return 8;
}
else if (Cs >= 598) {
return 2;
}
}
else if (Cs >= 673) {
if (AFD < 3) {
if (Rs < 590) {
return 2;
}
else if (Rs >= 590) {
return 1;
}
}
else if (AFD >= 3) {
return 1;
}
}
}
}
else if (Rs >= 646) {
if (Rs < 799) {
return 1;
}
else if (Rs >= 799) {
if (Rs < 915) {
return 4;
}
else if (Rs >= 915) {
return 8;
}
}
}
}
}
else if (Cs >= 821) {
if (ssDCval < 85) {
if (mu_mv_mag_sq < 81) {
if (Rs < 511) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (Cs < 1235) {
return 2;
}
else if (Cs >= 1235) {
return 1;
}
}
}
else if (Rs >= 511) {
if (posBalance < 60) {
if (RsDiff < 2) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 652) {
return 2;
}
else if (Rs >= 652) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (gchDC < 2) {
if (SC < 1566) {
if (Rs < 691) {
return 2;
}
else if (Rs >= 691) {
return 1;
}
}
else if (SC >= 1566) {
if (Cs < 1277) {
return 2;
}
else if (Cs >= 1277) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (Rs < 845) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 4;
}
}
else if (Rs >= 845) {
return 8;
}
}
}
}
else if (posBalance >= 60) {
if (SC < 1849) {
return 4;
}
else if (SC >= 1849) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (Rs < 757) {
if (negBalance < 33) {
if (Cs < 925) {
if (Rs < 738) {
return 1;
}
else if (Rs >= 738) {
return 2;
}
}
else if (Cs >= 925) {
return 2;
}
}
else if (negBalance >= 33) {
return 8;
}
}
else if (Rs >= 757) {
if (tcor < 96) {
if (gchDC < 1) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (tcor >= 96) {
if (negBalance < 36) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (negBalance >= 36) {
if (SC < 1992) {
if (Rs < 857) {
return 8;
}
else if (Rs >= 857) {
return 4;
}
}
else if (SC >= 1992) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (Cs < 763) {
if (Rs < 759) {
if (Rs < 453) {
if (Rs < 206) {
return 1;
}
else if (Rs >= 206) {
return 4;
}
}
else if (Rs >= 453) {
return 8;
}
}
else if (Rs >= 759) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (Cs >= 763) {
if (negBalance < 25) {
if (Rs < 736) {
return 1;
}
else if (Rs >= 736) {
return 4;
}
}
else if (negBalance >= 25) {
if (Rs < 776) {
return 2;
}
else if (Rs >= 776) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 964) {
if (mu_mv_mag_sq < 127) {
if (refDCval < 69) {
if (ssDCval < 51) {
if (Rs < 1183) {
if (Cs < 593) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
else if (Cs >= 593) {
return 8;
}
}
else if (Rs >= 1183) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (Rs < 1230) {
if (Rs < 1204) {
return 4;
}
else if (Rs >= 1204) {
return 8;
}
}
else if (Rs >= 1230) {
return 1;
}
}
}
}
else if (ssDCval >= 51) {
if (posBalance < 55) {
if (SC < 1856) {
if (refDCval < 65) {
if (SC < 1568) {
return 4;
}
else if (SC >= 1568) {
if (SC < 1854) {
return 1;
}
else if (SC >= 1854) {
return 2;
}
}
}
else if (refDCval >= 65) {
if (posBalance < 42) {
if (SC < 1612) {
return 2;
}
else if (SC >= 1612) {
return 1;
}
}
else if (posBalance >= 42) {
if (mcTcor < 99) {
if (Cs < 671) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (Cs >= 671) {
if (Rs < 1034) {
return 8;
}
else if (Rs >= 1034) {
return 4;
}
}
}
else if (mcTcor >= 99) {
if (Cs < 750) {
return 2;
}
else if (Cs >= 750) {
return 1;
}
}
}
}
}
else if (SC >= 1856) {
if (RsDiff < 4) {
if (negBalance < 41) {
if (ssDCval < 56) {
if (Cs < 855) {
if (Rs < 1363) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (Rs < 1349) {
return 1;
}
else if (Rs >= 1349) {
if (Cs < 678) {
return 1;
}
else if (Cs >= 678) {
return 8;
}
}
}
}
}
else if (Rs >= 1363) {
return 1;
}
}
else if (Cs >= 855) {
if (Rs < 976) {
return 4;
}
else if (Rs >= 976) {
return 1;
}
}
}
else if (ssDCval >= 56) {
if (Cs < 1092) {
if (MVDiff < 11) {
if (diffRsCsdiff < -4) {
if (mu_mv_mag_sq < 105) {
return 1;
}
else if (mu_mv_mag_sq >= 105) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
return 1;
}
}
else if (MVDiff >= 11) {
if (Rs < 1262) {
return 8;
}
else if (Rs >= 1262) {
return 1;
}
}
}
else if (Cs >= 1092) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Cs < 1133) {
return 8;
}
else if (Cs >= 1133) {
return 4;
}
}
}
}
}
else if (negBalance >= 41) {
if (SC < 2063) {
if (Rs < 1009) {
return 1;
}
else if (Rs >= 1009) {
return 4;
}
}
else if (SC >= 2063) {
return 8;
}
}
}
else if (RsDiff >= 4) {
if (ssDCval < 65) {
if (SC < 1902) {
return 2;
}
else if (SC >= 1902) {
return 1;
}
}
else if (ssDCval >= 65) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
if (CsDiff < 4) {
if (mu_mv_mag_sq < 81) {
return 8;
}
else if (mu_mv_mag_sq >= 81) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (Rs < 1216) {
return 4;
}
else if (Rs >= 1216) {
return 2;
}
}
}
}
}
else if (refDCval >= 69) {
if (Cs < 781) {
if (refDCval < 72) {
if (Rs < 1152) {
if (mu_mv_mag_sq < 79) {
if (Cs < 555) {
return 4;
}
else if (Cs >= 555) {
return 1;
}
}
else if (mu_mv_mag_sq >= 79) {
return 2;
}
}
else if (Rs >= 1152) {
return 8;
}
}
else if (refDCval >= 72) {
if (Rs < 1046) {
return 8;
}
else if (Rs >= 1046) {
if (diffRsCsdiff < -11) {
return 8;
}
else if (diffRsCsdiff >= -11) {
if (Rs < 1492) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (mu_mv_mag_sq < 39) {
return 1;
}
else if (mu_mv_mag_sq >= 39) {
return 8;
}
}
}
else if (Rs >= 1492) {
if (Cs < 569) {
return 1;
}
else if (Cs >= 569) {
return 8;
}
}
}
}
}
}
else if (Cs >= 781) {
if (mu_mv_mag_sq < 114) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (Rs < 1062) {
if (Rs < 1023) {
return 8;
}
else if (Rs >= 1023) {
return 1;
}
}
else if (Rs >= 1062) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (posBalance < 45) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (posBalance >= 45) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 1154) {
if (posBalance < 41) {
if (Cs < 589) {
return 2;
}
else if (Cs >= 589) {
return 8;
}
}
else if (posBalance >= 41) {
if (Rs < 1099) {
if (Rs < 1040) {
return 8;
}
else if (Rs >= 1040) {
return 2;
}
}
else if (Rs >= 1099) {
return 4;
}
}
}
else if (Rs >= 1154) {
if (SC < 1938) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 8;
}
}
else if (SC >= 1938) {
if (ssDCval < 74) {
if (Cs < 893) {
return 8;
}
else if (Cs >= 893) {
return 1;
}
}
else if (ssDCval >= 74) {
return 1;
}
}
}
}
}
}
else if (SC >= 2155) {
if (Rs < 1815) {
if (Cs < 1091) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 85) {
if (Cs < 871) {
if (ssDCval < 46) {
return 4;
}
else if (ssDCval >= 46) {
if (diffRsCsdiff < -4) {
if (ssDCval < 67) {
if (diffMVdiffVal < -1) {
if (Rs < 1661) {
return 8;
}
else if (Rs >= 1661) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (ssDCval >= 67) {
return 1;
}
}
else if (diffRsCsdiff >= -4) {
if (TSC < 6) {
if (Rs < 1384) {
if (Rs < 1367) {
return 1;
}
else if (Rs >= 1367) {
return 4;
}
}
else if (Rs >= 1384) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (Cs >= 871) {
if (posBalance < 45) {
if (ssDCval < 82) {
return 1;
}
else if (ssDCval >= 82) {
return 8;
}
}
else if (posBalance >= 45) {
if (AFD < 11) {
if (diffRsCsdiff < 10) {
return 8;
}
else if (diffRsCsdiff >= 10) {
if (Rs < 1646) {
return 2;
}
else if (Rs >= 1646) {
return 8;
}
}
}
else if (AFD >= 11) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (ssDCval < 49) {
if (MVDiff < 3) {
if (Cs < 1056) {
if (Cs < 1053) {
if (Cs < 1017) {
return 8;
}
else if (Cs >= 1017) {
if (Cs < 1027) {
return 4;
}
else if (Cs >= 1027) {
return 4;
}
}
}
else if (Cs >= 1053) {
return 8;
}
}
else if (Cs >= 1056) {
return 4;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (ssDCval >= 49) {
if (posBalance < 44) {
if (Rs < 1506) {
if (Rs < 1168) {
return 1;
}
else if (Rs >= 1168) {
if (Cs < 731) {
return 1;
}
else if (Cs >= 731) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 1324) {
return 1;
}
else if (Rs >= 1324) {
return 8;
}
}
}
}
}
else if (Rs >= 1506) {
if (negBalance < 26) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Rs < 1679) {
return 8;
}
else if (Rs >= 1679) {
return 4;
}
}
}
else if (negBalance >= 26) {
if (Rs < 1570) {
return 1;
}
else if (Rs >= 1570) {
if (Rs < 1793) {
return 8;
}
else if (Rs >= 1793) {
return 4;
}
}
}
}
}
else if (posBalance >= 44) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
if (Rs < 1344) {
if (SC < 2317) {
if (ssDCval < 66) {
if (Rs < 1260) {
return 1;
}
else if (Rs >= 1260) {
return 8;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (SC >= 2317) {
return 4;
}
}
else if (Rs >= 1344) {
if (SC < 2686) {
return 8;
}
else if (SC >= 2686) {
if (Rs < 1702) {
return 2;
}
else if (Rs >= 1702) {
return 8;
}
}
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (refDCval < 69) {
if (ssDCval < 59) {
if (MVDiff < 4) {
if (ssDCval < 49) {
if (posBalance < 34) {
if (tcor < 100) {
return 8;
}
else if (tcor >= 100) {
return 4;
}
}
else if (posBalance >= 34) {
if (negBalance < 29) {
if (Cs < 1050) {
if (Rs < 1650) {
return 4;
}
else if (Rs >= 1650) {
return 4;
}
}
else if (Cs >= 1050) {
return 4;
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (ssDCval >= 59) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 128) {
if (tcor < 98) {
if (Cs < 794) {
return 1;
}
else if (Cs >= 794) {
return 8;
}
}
else if (tcor >= 98) {
return 1;
}
}
else if (mu_mv_mag_sq >= 128) {
if (diffRsCsdiff < -1) {
if (Cs < 922) {
return 8;
}
else if (Cs >= 922) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (Cs < 904) {
if (Rs < 1565) {
return 8;
}
else if (Rs >= 1565) {
return 1;
}
}
else if (Cs >= 904) {
return 1;
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 1670) {
return 4;
}
else if (Rs >= 1670) {
return 2;
}
}
}
}
else if (refDCval >= 69) {
if (posBalance < 42) {
if (RsDiff < 1) {
if (Rs < 1757) {
return 8;
}
else if (Rs >= 1757) {
return 1;
}
}
else if (RsDiff >= 1) {
if (Cs < 613) {
return 4;
}
else if (Cs >= 613) {
return 1;
}
}
}
else if (posBalance >= 42) {
if (RsDiff < 3) {
if (Rs < 1149) {
return 4;
}
else if (Rs >= 1149) {
return 8;
}
}
else if (RsDiff >= 3) {
return 1;
}
}
}
}
}
else if (Cs >= 1091) {
if (RsDiff < 1) {
if (MVDiff < 8) {
if (ssDCval < 55) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
return 4;
}
}
else if (ssDCval >= 55) {
if (diffMVdiffVal < -54) {
return 8;
}
else if (diffMVdiffVal >= -54) {
return 1;
}
}
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (RsDiff >= 1) {
if (AFD < 3) {
if (refDCval < 52) {
if (Rs < 1651) {
if (negBalance < 23) {
if (negBalance < 20) {
return 8;
}
else if (negBalance >= 20) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (negBalance >= 23) {
return 8;
}
}
else if (Rs >= 1651) {
if (Cs < 1176) {
if (posBalance < 33) {
if (Cs < 1105) {
return 8;
}
else if (Cs >= 1105) {
if (Cs < 1157) {
if (SC < 2827) {
return 8;
}
else if (SC >= 2827) {
if (Rs < 1786) {
if (Rs < 1732) {
return 4;
}
else if (Rs >= 1732) {
if (Rs < 1766) {
if (Rs < 1752) {
return 4;
}
else if (Rs >= 1752) {
return 4;
}
}
else if (Rs >= 1766) {
return 4;
}
}
}
else if (Rs >= 1786) {
return 8;
}
}
}
else if (Cs >= 1157) {
return 8;
}
}
}
else if (posBalance >= 33) {
if (refDCval < 47) {
return 4;
}
else if (refDCval >= 47) {
if (Rs < 1677) {
return 4;
}
else if (Rs >= 1677) {
if (refDCval < 48) {
if (diffRsCsdiff < 2) {
if (Cs < 1113) {
if (Cs < 1108) {
if (Rs < 1756) {
if (Cs < 1099) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Rs < 1746) {
return 4;
}
else if (Rs >= 1746) {
return 4;
}
}
}
else if (Cs >= 1099) {
return 4;
}
}
else if (Rs >= 1756) {
return 4;
}
}
else if (Cs >= 1108) {
if (Rs < 1712) {
return 8;
}
else if (Rs >= 1712) {
return 8;
}
}
}
else if (Cs >= 1113) {
if (SC < 2915) {
if (Cs < 1160) {
if (Cs < 1155) {
if (posBalance < 37) {
if (mu_mv_mag_sq < 128) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (negBalance < 19) {
if (Rs < 1753) {
if (Rs < 1712) {
return 4;
}
else if (Rs >= 1712) {
return 4;
}
}
else if (Rs >= 1753) {
return 4;
}
}
else if (negBalance >= 19) {
if (Rs < 1727) {
return 4;
}
else if (Rs >= 1727) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
else if (posBalance >= 37) {
if (posBalance < 38) {
if (Rs < 1697) {
return 4;
}
else if (Rs >= 1697) {
return 8;
}
}
else if (posBalance >= 38) {
if (posBalance < 40) {
if (Rs < 1739) {
return 4;
}
else if (Rs >= 1739) {
return 4;
}
}
else if (posBalance >= 40) {
if (Rs < 1737) {
return 4;
}
else if (Rs >= 1737) {
return 4;
}
}
}
}
}
else if (Cs >= 1155) {
return 4;
}
}
else if (Cs >= 1160) {
return 8;
}
}
else if (SC >= 2915) {
if (Rs < 1786) {
if (AFD < 2) {
if (Rs < 1779) {
if (Rs < 1768) {
return 8;
}
else if (Rs >= 1768) {
return 4;
}
}
else if (Rs >= 1779) {
return 8;
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 1786) {
if (Rs < 1800) {
return 4;
}
else if (Rs >= 1800) {
if (Rs < 1811) {
return 4;
}
else if (Rs >= 1811) {
return 4;
}
}
}
}
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 1702) {
return 8;
}
else if (Rs >= 1702) {
return 8;
}
}
}
else if (refDCval >= 48) {
return 4;
}
}
}
}
}
else if (Cs >= 1176) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
return 8;
}
}
}
}
else if (refDCval >= 52) {
if (MVDiff < 7) {
if (refDCval < 70) {
if (negBalance < 32) {
if (mu_mv_mag_sq < 121) {
return 1;
}
else if (mu_mv_mag_sq >= 121) {
if (SC < 2555) {
return 4;
}
else if (SC >= 2555) {
if (Cs < 1244) {
return 1;
}
else if (Cs >= 1244) {
return 4;
}
}
}
}
else if (negBalance >= 32) {
return 1;
}
}
else if (refDCval >= 70) {
if (Rs < 1480) {
if (Cs < 1103) {
return 8;
}
else if (Cs >= 1103) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
else if (Rs >= 1480) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (AFD >= 3) {
if (Rs < 1394) {
if (mu_mv_mag_sq < 113) {
if (RsDiff < 2) {
if (Rs < 968) {
if (Rs < 961) {
return 2;
}
else if (Rs >= 961) {
return 4;
}
}
else if (Rs >= 968) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 4;
}
}
}
else if (RsDiff >= 2) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 66) {
if (Rs < 1038) {
return 2;
}
else if (Rs >= 1038) {
return 8;
}
}
else if (mu_mv_mag_sq >= 66) {
if (posBalance < 54) {
if (Rs < 1082) {
return 8;
}
else if (Rs >= 1082) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (diffRsCsdiff < 6) {
return 4;
}
else if (diffRsCsdiff >= 6) {
if (Rs < 1144) {
return 4;
}
else if (Rs >= 1144) {
return 8;
}
}
}
}
}
else if (posBalance >= 54) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (RsCsDiff < 16) {
return 4;
}
else if (RsCsDiff >= 16) {
if (Cs < 1273) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 2;
}
}
else if (Cs >= 1273) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 937) {
return 2;
}
else if (Rs >= 937) {
if (ssDCval < 62) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 114) {
return 4;
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (ssDCval >= 62) {
if (diffRsCsdiff < 6) {
if (Cs < 1106) {
if (Rs < 1275) {
return 8;
}
else if (Rs >= 1275) {
return 4;
}
}
else if (Cs >= 1106) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
return 1;
}
}
}
else if (diffRsCsdiff >= 6) {
return 1;
}
}
}
}
}
else if (Rs >= 1394) {
if (SC < 2992) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (SC >= 2992) {
if (Rs < 1788) {
return 2;
}
else if (Rs >= 1788) {
if (Rs < 1804) {
return 1;
}
else if (Rs >= 1804) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1815) {
if (TSC < 4) {
if (ssDCval < 82) {
if (ssDCval < 56) {
if (Cs < 855) {
if (mu_mv_mag_sq < 95) {
if (Cs < 531) {
return 1;
}
else if (Cs >= 531) {
return 4;
}
}
else if (mu_mv_mag_sq >= 95) {
return 8;
}
}
else if (Cs >= 855) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
if (gchDC < 1) {
if (SC < 2996) {
return 4;
}
else if (SC >= 2996) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 56) {
if (refDCval < 71) {
if (SC < 3208) {
if (posBalance < 36) {
if (ssDCval < 57) {
if (Rs < 1962) {
return 1;
}
else if (Rs >= 1962) {
return 4;
}
}
else if (ssDCval >= 57) {
if (Cs < 1094) {
if (Rs < 1869) {
if (refDCval < 61) {
return 4;
}
else if (refDCval >= 61) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 4;
}
}
}
else if (Rs >= 1869) {
return 4;
}
}
else if (Cs >= 1094) {
if (posBalance < 32) {
if (Rs < 1849) {
return 2;
}
else if (Rs >= 1849) {
return 4;
}
}
else if (posBalance >= 32) {
if (Rs < 1832) {
return 4;
}
else if (Rs >= 1832) {
return 1;
}
}
}
}
}
else if (posBalance >= 36) {
if (MVDiff < 2) {
if (Rs < 1843) {
if (Rs < 1831) {
return 1;
}
else if (Rs >= 1831) {
return 4;
}
}
else if (Rs >= 1843) {
return 1;
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 100) {
return 4;
}
else if (mu_mv_mag_sq >= 100) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
return 4;
}
}
}
}
}
else if (SC >= 3208) {
if (AFD < 1) {
if (Cs < 1329) {
return 2;
}
else if (Cs >= 1329) {
return 8;
}
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (refDCval >= 71) {
if (posBalance < 43) {
if (SC < 3076) {
return 1;
}
else if (SC >= 3076) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (Cs < 845) {
return 1;
}
else if (Cs >= 845) {
return 8;
}
}
}
}
}
else if (ssDCval >= 82) {
if (MVDiff < 2) {
if (Rs < 2186) {
if (Cs < 939) {
return 4;
}
else if (Cs >= 939) {
return 8;
}
}
else if (Rs >= 2186) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Cs < 541) {
return 8;
}
else if (Cs >= 541) {
return 1;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 932) {
if (ssDCval < 71) {
if (ssDCval < 61) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
else if (ssDCval >= 61) {
if (RsDiff < 11) {
return 2;
}
else if (RsDiff >= 11) {
return 4;
}
}
}
else if (ssDCval >= 71) {
if (RsCsDiff < 56) {
if (mu_mv_mag_sq < 86) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 2;
}
}
}
else if (RsCsDiff >= 56) {
if (SC < 2755) {
return 8;
}
else if (SC >= 2755) {
return 2;
}
}
}
}
else if (Cs >= 932) {
if (Cs < 1034) {
if (tcor < 97) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Rs < 2373) {
if (Rs < 2258) {
return 4;
}
else if (Rs >= 2258) {
return 4;
}
}
else if (Rs >= 2373) {
return 2;
}
}
}
else if (tcor >= 97) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (Cs >= 1034) {
if (tcor < 93) {
if (Rs < 2087) {
return 1;
}
else if (Rs >= 2087) {
return 4;
}
}
else if (tcor >= 93) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 103) {
if (SC < 1296) {
if (MVDiff < 4) {
if (Rs < 413) {
if (mcTcor < 98) {
if (Rs < 281) {
return 2;
}
else if (Rs >= 281) {
return 1;
}
}
else if (mcTcor >= 98) {
return 4;
}
}
else if (Rs >= 413) {
if (Rs < 542) {
if (SC < 1048) {
if (mu_mv_mag_sq < 87) {
if (SC < 1045) {
return 1;
}
else if (SC >= 1045) {
return 8;
}
}
else if (mu_mv_mag_sq >= 87) {
return 8;
}
}
else if (SC >= 1048) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (Cs < 563) {
if (Rs < 519) {
return 1;
}
else if (Rs >= 519) {
return 2;
}
}
else if (Cs >= 563) {
return 1;
}
}
}
}
else if (Rs >= 542) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (mu_mv_mag_sq < 85) {
if (ssDCval < 114) {
return 1;
}
else if (ssDCval >= 114) {
return 2;
}
}
else if (mu_mv_mag_sq >= 85) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 115) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (tcor < 95) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (Rs < 560) {
return 1;
}
else if (Rs >= 560) {
return 4;
}
}
}
else if (tcor >= 95) {
if (negBalance < 50) {
if (TSC < 5) {
if (Rs < 380) {
return 4;
}
else if (Rs >= 380) {
return 2;
}
}
else if (TSC >= 5) {
if (tcor < 102) {
return 8;
}
else if (tcor >= 102) {
return 2;
}
}
}
else if (negBalance >= 50) {
if (Rs < 603) {
return 8;
}
else if (Rs >= 603) {
if (Rs < 625) {
return 1;
}
else if (Rs >= 625) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 115) {
if (refDCval < 121) {
if (posBalance < 38) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 2;
}
}
else if (posBalance >= 38) {
if (SC < 981) {
return 4;
}
else if (SC >= 981) {
return 2;
}
}
}
else if (refDCval >= 121) {
if (refDCval < 150) {
if (CsDiff < 2) {
if (negBalance < 30) {
if (Rs < 508) {
return 2;
}
else if (Rs >= 508) {
return 1;
}
}
else if (negBalance >= 30) {
if (AFD < 2) {
if (Rs < 457) {
return 2;
}
else if (Rs >= 457) {
return 4;
}
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 81) {
return 2;
}
else if (mu_mv_mag_sq >= 81) {
if (Rs < 327) {
return 2;
}
else if (Rs >= 327) {
return 8;
}
}
}
}
else if (refDCval >= 150) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
}
}
else if (SC >= 1296) {
if (RsCsDiff < 81) {
if (refDCval < 112) {
if (TSC < 5) {
if (mu_mv_mag_sq < 93) {
if (Rs < 607) {
return 2;
}
else if (Rs >= 607) {
if (Cs < 1168) {
if (refDCval < 86) {
return 2;
}
else if (refDCval >= 86) {
if (TSC < 3) {
if (Rs < 1887) {
return 1;
}
else if (Rs >= 1887) {
return 8;
}
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (Cs >= 1168) {
if (Rs < 1337) {
if (Rs < 742) {
return 1;
}
else if (Rs >= 742) {
return 4;
}
}
else if (Rs >= 1337) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (ssDCval < 107) {
if (ssDCval < 90) {
if (AFD < 4) {
if (Rs < 1251) {
return 8;
}
else if (Rs >= 1251) {
return 1;
}
}
else if (AFD >= 4) {
return 4;
}
}
else if (ssDCval >= 90) {
return 8;
}
}
else if (ssDCval >= 107) {
return 1;
}
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 7) {
if (CsDiff < 4) {
if (TSC < 7) {
if (Rs < 790) {
return 4;
}
else if (Rs >= 790) {
if (diffMVdiffVal < -5) {
return 2;
}
else if (diffMVdiffVal >= -5) {
return 1;
}
}
}
else if (TSC >= 7) {
if (Rs < 1222) {
return 8;
}
else if (Rs >= 1222) {
return 1;
}
}
}
else if (CsDiff >= 4) {
if (MVDiff < 15) {
if (TSC < 7) {
if (RsCsDiff < 57) {
if (posBalance < 70) {
return 4;
}
else if (posBalance >= 70) {
return 2;
}
}
else if (RsCsDiff >= 57) {
if (Cs < 1133) {
if (Rs < 1516) {
return 1;
}
else if (Rs >= 1516) {
return 4;
}
}
else if (Cs >= 1133) {
return 8;
}
}
}
else if (TSC >= 7) {
if (Rs < 2070) {
return 8;
}
else if (Rs >= 2070) {
if (Rs < 2310) {
return 1;
}
else if (Rs >= 2310) {
return 4;
}
}
}
}
else if (MVDiff >= 15) {
if (posBalance < 55) {
if (Rs < 1525) {
return 1;
}
else if (Rs >= 1525) {
return 4;
}
}
else if (posBalance >= 55) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (MVDiff < 2) {
if (Rs < 1461) {
return 4;
}
else if (Rs >= 1461) {
if (Rs < 2098) {
return 1;
}
else if (Rs >= 2098) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (tcor < 93) {
if (posBalance < 53) {
if (ssDCval < 96) {
if (Rs < 1560) {
return 8;
}
else if (Rs >= 1560) {
return 1;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (posBalance >= 53) {
if (Rs < 659) {
return 2;
}
else if (Rs >= 659) {
return 8;
}
}
}
else if (tcor >= 93) {
if (diffTSC < 2) {
if (MVDiff < 4) {
if (tcor < 96) {
if (Rs < 1095) {
return 4;
}
else if (Rs >= 1095) {
return 1;
}
}
else if (tcor >= 96) {
return 8;
}
}
else if (MVDiff >= 4) {
if (negBalance < 53) {
return 8;
}
else if (negBalance >= 53) {
if (Cs < 804) {
return 1;
}
else if (Cs >= 804) {
return 8;
}
}
}
}
else if (diffTSC >= 2) {
if (Rs < 2059) {
return 1;
}
else if (Rs >= 2059) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 112) {
if (MVDiff < 6) {
if (SC < 2612) {
if (ssDCval < 145) {
if (Cs < 593) {
return 1;
}
else if (Cs >= 593) {
if (TSC < 3) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
return 8;
}
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 2) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (refDCval < 118) {
if (AFD < 9) {
if (Rs < 526) {
return 1;
}
else if (Rs >= 526) {
return 8;
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (refDCval >= 118) {
if (SC < 1408) {
return 2;
}
else if (SC >= 1408) {
if (Rs < 1712) {
return 8;
}
else if (Rs >= 1712) {
if (Rs < 1718) {
return 4;
}
else if (Rs >= 1718) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 145) {
if (Rs < 675) {
return 1;
}
else if (Rs >= 675) {
if (Rs < 1033) {
return 4;
}
else if (Rs >= 1033) {
return 8;
}
}
}
}
else if (SC >= 2612) {
if (SC < 3001) {
if (RsCsDiff < 15) {
return 4;
}
else if (RsCsDiff >= 15) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
return 2;
}
}
}
else if (SC >= 3001) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (SC < 3107) {
return 2;
}
else if (SC >= 3107) {
return 8;
}
}
}
}
}
else if (MVDiff >= 6) {
if (refDCval < 161) {
if (mu_mv_mag_sq < 3) {
if (AFD < 9) {
if (MVDiff < 27) {
return 8;
}
else if (MVDiff >= 27) {
return 2;
}
}
else if (AFD >= 9) {
if (ssDCval < 123) {
if (posBalance < 53) {
if (Rs < 1514) {
return 1;
}
else if (Rs >= 1514) {
return 8;
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (ssDCval >= 123) {
if (Cs < 914) {
return 1;
}
else if (Cs >= 914) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 2235) {
if (Cs < 740) {
if (mu_mv_mag_sq < 56) {
if (TSC < 6) {
if (diffMVdiffVal < 10) {
if (Cs < 559) {
if (Rs < 807) {
return 8;
}
else if (Rs >= 807) {
return 2;
}
}
else if (Cs >= 559) {
return 8;
}
}
else if (diffMVdiffVal >= 10) {
return 2;
}
}
else if (TSC >= 6) {
if (diffMVdiffVal < -1) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 1138) {
return 8;
}
else if (Rs >= 1138) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 56) {
return 1;
}
}
else if (Cs >= 740) {
if (SC < 1663) {
if (ssDCval < 139) {
return 8;
}
else if (ssDCval >= 139) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (SC >= 1663) {
if (Cs < 1313) {
return 8;
}
else if (Cs >= 1313) {
if (Rs < 1352) {
return 8;
}
else if (Rs >= 1352) {
return 2;
}
}
}
}
}
else if (Rs >= 2235) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
else if (refDCval >= 161) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 81) {
if (Cs < 792) {
if (Rs < 1463) {
if (tcor < 79) {
if (Rs < 864) {
return 1;
}
else if (Rs >= 864) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 4;
}
}
}
else if (tcor >= 79) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (Rs >= 1463) {
if (tcor < 85) {
if (AFD < 21) {
if (Cs < 549) {
return 1;
}
else if (Cs >= 549) {
return 8;
}
}
else if (AFD >= 21) {
if (Rs < 2249) {
return 1;
}
else if (Rs >= 2249) {
return 2;
}
}
}
else if (tcor >= 85) {
if (mu_mv_mag_sq < 36) {
return 1;
}
else if (mu_mv_mag_sq >= 36) {
if (Rs < 2119) {
return 2;
}
else if (Rs >= 2119) {
return 1;
}
}
}
}
}
else if (Cs >= 792) {
if (refDCval < 108) {
if (Cs < 877) {
if (MVDiff < 32) {
return 4;
}
else if (MVDiff >= 32) {
return 2;
}
}
else if (Cs >= 877) {
if (Rs < 1360) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 8;
}
}
else if (Rs >= 1360) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
if (Rs < 1660) {
return 4;
}
else if (Rs >= 1660) {
if (Rs < 1739) {
return 1;
}
else if (Rs >= 1739) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 108) {
if (tcor < 93) {
if (ssDCval < 138) {
return 2;
}
else if (ssDCval >= 138) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
}
else if (tcor >= 93) {
if (Cs < 1313) {
return 8;
}
else if (Cs >= 1313) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (ssDCval < 133) {
if (Rs < 342) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 342) {
if (Cs < 1234) {
if (diffAFD < -2) {
if (SC < 2733) {
if (Cs < 765) {
if (RsDiff < 2) {
if (Rs < 454) {
return 8;
}
else if (Rs >= 454) {
return 1;
}
}
else if (RsDiff >= 2) {
if (Rs < 1319) {
return 8;
}
else if (Rs >= 1319) {
return 4;
}
}
}
else if (Cs >= 765) {
if (MVDiff < 3) {
if (AFD < 4) {
if (Cs < 849) {
return 4;
}
else if (Cs >= 849) {
return 1;
}
}
else if (AFD >= 4) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (SC >= 2733) {
if (Rs < 1893) {
return 2;
}
else if (Rs >= 1893) {
return 8;
}
}
}
else if (diffAFD >= -2) {
if (refDCval < 97) {
if (negBalance < 22) {
if (SC < 2388) {
if (Rs < 858) {
return 8;
}
else if (Rs >= 858) {
return 4;
}
}
else if (SC >= 2388) {
return 8;
}
}
else if (negBalance >= 22) {
if (Cs < 841) {
if (ssDCval < 92) {
if (SC < 2892) {
return 8;
}
else if (SC >= 2892) {
if (Rs < 2179) {
return 4;
}
else if (Rs >= 2179) {
if (Rs < 2187) {
return 8;
}
else if (Rs >= 2187) {
return 4;
}
}
}
}
else if (ssDCval >= 92) {
if (Rs < 1264) {
return 8;
}
else if (Rs >= 1264) {
return 4;
}
}
}
else if (Cs >= 841) {
if (refDCval < 91) {
return 8;
}
else if (refDCval >= 91) {
if (ssDCval < 95) {
if (Rs < 1915) {
if (SC < 1687) {
return 1;
}
else if (SC >= 1687) {
return 4;
}
}
else if (Rs >= 1915) {
return 8;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
else if (refDCval >= 97) {
if (Rs < 1676) {
if (mu_mv_mag_sq < 108) {
if (AFD < 4) {
if (Cs < 671) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
return 8;
}
}
else if (Cs >= 671) {
if (Rs < 1070) {
return 4;
}
else if (Rs >= 1070) {
return 8;
}
}
}
else if (AFD >= 4) {
if (Rs < 464) {
return 1;
}
else if (Rs >= 464) {
if (Cs < 1212) {
return 8;
}
else if (Cs >= 1212) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 451) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
if (Rs < 369) {
if (ssDCval < 109) {
return 4;
}
else if (ssDCval >= 109) {
return 8;
}
}
else if (Rs >= 369) {
return 8;
}
}
}
else if (Rs >= 451) {
if (AFD < 18) {
if (Cs < 1219) {
return 8;
}
else if (Cs >= 1219) {
if (Rs < 600) {
return 1;
}
else if (Rs >= 600) {
return 8;
}
}
}
else if (AFD >= 18) {
return 1;
}
}
}
}
else if (Rs >= 1676) {
if (ssDCval < 123) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 983) {
return 1;
}
else if (Cs >= 983) {
return 8;
}
}
}
else if (ssDCval >= 123) {
if (Rs < 1813) {
return 2;
}
else if (Rs >= 1813) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1234) {
if (RsDiff < 1) {
if (Rs < 667) {
return 1;
}
else if (Rs >= 667) {
if (Rs < 907) {
return 8;
}
else if (Rs >= 907) {
return 4;
}
}
}
else if (RsDiff >= 1) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1147) {
if (Cs < 858) {
return 2;
}
else if (Cs >= 858) {
return 8;
}
}
else if (Rs >= 1147) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2388) {
if (MVDiff < 14) {
if (ssDCval < 109) {
if (posBalance < 48) {
if (RsCsDiff < 15) {
if (Cs < 1037) {
if (ssDCval < 86) {
if (MVDiff < 3) {
if (Cs < 758) {
return 1;
}
else if (Cs >= 758) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 2;
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 61) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (ssDCval >= 61) {
return 8;
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 91) {
if (Cs < 434) {
return 8;
}
else if (Cs >= 434) {
return 4;
}
}
else if (ssDCval >= 91) {
if (negBalance < 33) {
if (Rs < 2581) {
if (Cs < 734) {
return 8;
}
else if (Cs >= 734) {
return 4;
}
}
else if (Rs >= 2581) {
return 8;
}
}
else if (negBalance >= 33) {
return 2;
}
}
}
}
else if (Cs >= 1037) {
if (SC < 4573) {
if (Rs < 3108) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
if (negBalance < 53) {
if (Rs < 2795) {
return 8;
}
else if (Rs >= 2795) {
return 1;
}
}
else if (negBalance >= 53) {
if (Rs < 2511) {
return 8;
}
else if (Rs >= 2511) {
if (Rs < 2568) {
return 4;
}
else if (Rs >= 2568) {
return 4;
}
}
}
}
}
else if (Rs >= 3108) {
if (Rs < 3267) {
return 2;
}
else if (Rs >= 3267) {
return 8;
}
}
}
else if (SC >= 4573) {
if (CsDiff < 2) {
if (SC < 4977) {
if (Cs < 1328) {
if (tcor < 102) {
if (Rs < 3417) {
if (Rs < 3413) {
return 8;
}
else if (Rs >= 3413) {
return 4;
}
}
else if (Rs >= 3417) {
return 8;
}
}
else if (tcor >= 102) {
return 4;
}
}
else if (Cs >= 1328) {
return 4;
}
}
else if (SC >= 4977) {
if (SC < 5433) {
if (Cs < 1274) {
return 4;
}
else if (Cs >= 1274) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (diffTSC < -1) {
if (Rs < 4009) {
return 8;
}
else if (Rs >= 4009) {
return 4;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
else if (SC >= 5433) {
if (Rs < 4584) {
return 8;
}
else if (Rs >= 4584) {
return 2;
}
}
}
}
else if (CsDiff >= 2) {
if (ssDCval < 71) {
if (Cs < 1223) {
return 1;
}
else if (Cs >= 1223) {
return 8;
}
}
else if (ssDCval >= 71) {
if (Rs < 4424) {
return 4;
}
else if (Rs >= 4424) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 15) {
if (refDCval < 90) {
if (Rs < 3655) {
if (refDCval < 66) {
if (refDCval < 60) {
if (MVDiff < 8) {
if (AFD < 6) {
if (mu_mv_mag_sq < 115) {
return 4;
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
else if (AFD >= 6) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
else if (refDCval >= 60) {
if (negBalance < 43) {
if (MVDiff < 3) {
if (refDCval < 64) {
return 2;
}
else if (refDCval >= 64) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (SC < 3811) {
return 2;
}
else if (SC >= 3811) {
return 4;
}
}
}
}
else if (MVDiff >= 3) {
if (Cs < 1246) {
if (diffMVdiffVal < -3) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (Rs < 3163) {
return 2;
}
else if (Rs >= 3163) {
return 4;
}
}
}
else if (diffMVdiffVal >= -3) {
if (Rs < 3327) {
if (SC < 3326) {
return 2;
}
else if (SC >= 3326) {
return 4;
}
}
else if (Rs >= 3327) {
return 2;
}
}
}
else if (Cs >= 1246) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 2;
}
}
}
}
else if (negBalance >= 43) {
if (posBalance < 34) {
return 1;
}
else if (posBalance >= 34) {
return 2;
}
}
}
}
else if (refDCval >= 66) {
if (negBalance < 51) {
if (mu_mv_mag_sq < 85) {
if (SC < 3762) {
if (mu_mv_mag_sq < 7) {
if (TSC < 6) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (TSC >= 6) {
if (negBalance < 42) {
if (Rs < 2536) {
return 2;
}
else if (Rs >= 2536) {
return 1;
}
}
else if (negBalance >= 42) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (Rs < 2567) {
if (refDCval < 73) {
return 2;
}
else if (refDCval >= 73) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (Rs >= 2567) {
if (Cs < 497) {
return 8;
}
else if (Cs >= 497) {
if (Cs < 819) {
if (Rs < 2783) {
return 4;
}
else if (Rs >= 2783) {
if (Rs < 2911) {
return 8;
}
else if (Rs >= 2911) {
return 8;
}
}
}
else if (Cs >= 819) {
return 2;
}
}
}
}
}
else if (SC >= 3762) {
if (posBalance < 45) {
if (refDCval < 82) {
if (mu_mv_mag_sq < 75) {
if (RsDiff < 9) {
if (Rs < 2691) {
return 2;
}
else if (Rs >= 2691) {
return 4;
}
}
else if (RsDiff >= 9) {
if (SC < 4544) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
if (tcor < 93) {
if (negBalance < 41) {
return 4;
}
else if (negBalance >= 41) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
return 4;
}
}
}
else if (tcor >= 93) {
return 4;
}
}
}
else if (SC >= 4544) {
if (diffAFD < 2) {
if (Rs < 3495) {
return 2;
}
else if (Rs >= 3495) {
return 4;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (Cs < 1163) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (Cs < 900) {
return 4;
}
else if (Cs >= 900) {
return 1;
}
}
}
else if (Cs >= 1163) {
return 2;
}
}
}
else if (refDCval >= 82) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (posBalance >= 45) {
if (negBalance < 41) {
if (Rs < 2940) {
if (Rs < 2896) {
return 1;
}
else if (Rs >= 2896) {
return 2;
}
}
else if (Rs >= 2940) {
if (diffRsCsdiff < 6) {
if (Cs < 940) {
return 4;
}
else if (Cs >= 940) {
if (Cs < 1248) {
return 1;
}
else if (Cs >= 1248) {
return 4;
}
}
}
else if (diffRsCsdiff >= 6) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
}
else if (negBalance >= 41) {
if (Rs < 2601) {
return 1;
}
else if (Rs >= 2601) {
if (Rs < 3505) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
return 2;
}
}
else if (Rs >= 3505) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (posBalance < 44) {
if (CsDiff < 4) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (CsDiff >= 4) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (posBalance >= 44) {
return 2;
}
}
}
else if (negBalance >= 51) {
return 2;
}
}
}
else if (Rs >= 3655) {
if (Cs < 1139) {
if (refDCval < 79) {
if (diffAFD < -2) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Rs < 3928) {
return 1;
}
else if (Rs >= 3928) {
return 4;
}
}
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (refDCval >= 79) {
return 2;
}
}
else if (Cs >= 1139) {
if (negBalance < 29) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (negBalance >= 29) {
if (RsDiff < 8) {
if (Cs < 1330) {
return 4;
}
else if (Cs >= 1330) {
return 2;
}
}
else if (RsDiff >= 8) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 90) {
if (Cs < 802) {
if (Rs < 3094) {
if (negBalance < 44) {
if (Rs < 2730) {
return 1;
}
else if (Rs >= 2730) {
return 8;
}
}
else if (negBalance >= 44) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (Rs >= 3094) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
}
else if (Cs >= 802) {
if (ssDCval < 103) {
if (negBalance < 33) {
if (mu_mv_mag_sq < 81) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (mu_mv_mag_sq < 48) {
return 8;
}
else if (mu_mv_mag_sq >= 48) {
if (diffTSC < 1) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (Cs < 1263) {
if (Cs < 1231) {
if (Cs < 1201) {
if (Rs < 3390) {
return 4;
}
else if (Rs >= 3390) {
if (Rs < 3812) {
if (Rs < 3604) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (Rs >= 3604) {
return 8;
}
}
else if (Rs >= 3812) {
if (Cs < 1050) {
return 4;
}
else if (Cs >= 1050) {
if (negBalance < 32) {
if (mu_mv_mag_sq < 53) {
if (Cs < 1094) {
return 4;
}
else if (Cs >= 1094) {
if (Rs < 3947) {
return 4;
}
else if (Rs >= 3947) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 53) {
return 4;
}
}
else if (negBalance >= 32) {
return 4;
}
}
}
}
}
else if (Cs >= 1201) {
return 8;
}
}
else if (Cs >= 1231) {
return 4;
}
}
else if (Cs >= 1263) {
if (Rs < 4153) {
return 8;
}
else if (Rs >= 4153) {
return 8;
}
}
}
}
else if (diffTSC >= 1) {
if (Rs < 3847) {
return 8;
}
else if (Rs >= 3847) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (RsCsDiff < 29) {
return 4;
}
else if (RsCsDiff >= 29) {
if (Rs < 4744) {
return 2;
}
else if (Rs >= 4744) {
return 8;
}
}
}
}
else if (negBalance >= 33) {
if (refDCval < 92) {
if (AFD < 13) {
if (SC < 5319) {
return 8;
}
else if (SC >= 5319) {
if (SC < 5356) {
return 4;
}
else if (SC >= 5356) {
return 4;
}
}
}
else if (AFD >= 13) {
return 2;
}
}
else if (refDCval >= 92) {
if (RsCsDiff < 31) {
if (Rs < 2718) {
return 1;
}
else if (Rs >= 2718) {
return 8;
}
}
else if (RsCsDiff >= 31) {
if (Rs < 2602) {
return 1;
}
else if (Rs >= 2602) {
return 2;
}
}
}
}
}
else if (ssDCval >= 103) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (CsDiff < 7) {
if (Rs < 4109) {
return 1;
}
else if (Rs >= 4109) {
return 2;
}
}
else if (CsDiff >= 7) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (refDCval < 60) {
if (RsDiff < 5) {
if (Cs < 855) {
return 1;
}
else if (Cs >= 855) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (RsDiff >= 5) {
if (Rs < 7110) {
if (Rs < 2534) {
return 2;
}
else if (Rs >= 2534) {
return 4;
}
}
else if (Rs >= 7110) {
return 2;
}
}
}
else if (refDCval >= 60) {
if (CsDiff < 3) {
if (Rs < 3904) {
if (Cs < 379) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (posBalance < 59) {
if (Rs < 2871) {
return 1;
}
else if (Rs >= 2871) {
return 4;
}
}
else if (posBalance >= 59) {
return 2;
}
}
}
}
else if (Cs >= 379) {
if (negBalance < 19) {
if (Rs < 2808) {
if (Rs < 2651) {
return 4;
}
else if (Rs >= 2651) {
return 4;
}
}
else if (Rs >= 2808) {
return 8;
}
}
else if (negBalance >= 19) {
if (diffAFD < -2) {
if (diffTSC < 1) {
if (diffMVdiffVal < -5) {
if (Rs < 2618) {
return 2;
}
else if (Rs >= 2618) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
return 2;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (diffAFD >= -2) {
if (Cs < 672) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
return 2;
}
}
else if (Cs >= 672) {
if (refDCval < 63) {
if (Rs < 2464) {
return 2;
}
else if (Rs >= 2464) {
return 4;
}
}
else if (refDCval >= 63) {
if (Rs < 3656) {
return 2;
}
else if (Rs >= 3656) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 3904) {
return 4;
}
}
else if (CsDiff >= 3) {
if (negBalance < 38) {
if (posBalance < 55) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 74) {
if (Rs < 3354) {
if (ssDCval < 79) {
if (posBalance < 50) {
if (Rs < 2974) {
return 2;
}
else if (Rs >= 2974) {
return 4;
}
}
else if (posBalance >= 50) {
return 2;
}
}
else if (ssDCval >= 79) {
return 1;
}
}
else if (Rs >= 3354) {
return 4;
}
}
else if (mu_mv_mag_sq >= 74) {
return 2;
}
}
else if (MVDiff >= 5) {
if (SC < 3354) {
if (mu_mv_mag_sq < 2) {
return 4;
}
else if (mu_mv_mag_sq >= 2) {
return 1;
}
}
else if (SC >= 3354) {
if (Rs < 2507) {
return 2;
}
else if (Rs >= 2507) {
if (AFD < 10) {
if (Rs < 3548) {
return 2;
}
else if (Rs >= 3548) {
if (Rs < 4187) {
return 4;
}
else if (Rs >= 4187) {
return 4;
}
}
}
else if (AFD >= 10) {
return 4;
}
}
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 63) {
if (Cs < 914) {
return 2;
}
else if (Cs >= 914) {
return 4;
}
}
else if (ssDCval >= 63) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (mcTcor < 98) {
if (tcor < 85) {
return 1;
}
else if (tcor >= 85) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
if (Cs < 1091) {
return 2;
}
else if (Cs >= 1091) {
return 1;
}
}
}
}
else if (mcTcor >= 98) {
if (Cs < 1149) {
return 1;
}
else if (Cs >= 1149) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 38) {
if (refDCval < 64) {
if (Cs < 1013) {
if (Rs < 2790) {
return 2;
}
else if (Rs >= 2790) {
if (RsDiff < 12) {
if (Rs < 2797) {
return 4;
}
else if (Rs >= 2797) {
return 2;
}
}
else if (RsDiff >= 12) {
return 4;
}
}
}
else if (Cs >= 1013) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
else if (refDCval >= 64) {
if (mcTcor < 99) {
if (tcor < 76) {
if (Cs < 802) {
return 2;
}
else if (Cs >= 802) {
if (tcor < 71) {
if (Rs < 3412) {
return 2;
}
else if (Rs >= 3412) {
return 1;
}
}
else if (tcor >= 71) {
return 4;
}
}
}
else if (tcor >= 76) {
if (RsCsDiff < 90) {
if (tcor < 89) {
if (posBalance < 55) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (posBalance >= 55) {
if (Rs < 2714) {
return 2;
}
else if (Rs >= 2714) {
return 8;
}
}
}
else if (tcor >= 89) {
if (RsCsDiff < 38) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
return 2;
}
}
else if (RsCsDiff >= 38) {
return 2;
}
}
}
else if (RsCsDiff >= 90) {
return 2;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 2560) {
return 2;
}
else if (Rs >= 2560) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsDiff < 11) {
if (Cs < 1018) {
if (tcor < 96) {
if (RsCsDiff < 63) {
if (mu_mv_mag_sq < 29) {
return 8;
}
else if (mu_mv_mag_sq >= 29) {
if (tcor < 92) {
if (Rs < 3111) {
return 4;
}
else if (Rs >= 3111) {
return 4;
}
}
else if (tcor >= 92) {
if (SC < 3162) {
return 4;
}
else if (SC >= 3162) {
return 8;
}
}
}
}
else if (RsCsDiff >= 63) {
if (Cs < 768) {
return 2;
}
else if (Cs >= 768) {
if (ssDCval < 100) {
if (Rs < 3082) {
return 4;
}
else if (Rs >= 3082) {
return 2;
}
}
else if (ssDCval >= 100) {
return 8;
}
}
}
}
else if (tcor >= 96) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 2;
}
}
}
else if (Cs >= 1018) {
if (posBalance < 63) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (Rs < 3905) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
if (Rs < 3426) {
return 2;
}
else if (Rs >= 3426) {
return 8;
}
}
}
else if (Rs >= 3905) {
if (Rs < 4165) {
return 2;
}
else if (Rs >= 4165) {
return 4;
}
}
}
}
else if (posBalance >= 63) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
return 4;
}
}
}
}
else if (RsDiff >= 11) {
if (negBalance < 52) {
if (RsDiff < 13) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 4;
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (negBalance >= 52) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (Cs < 823) {
return 2;
}
else if (Cs >= 823) {
if (ssDCval < 102) {
if (Cs < 1044) {
return 4;
}
else if (Cs >= 1044) {
return 8;
}
}
else if (ssDCval >= 102) {
if (RsDiff < 16) {
return 4;
}
else if (RsDiff >= 16) {
if (Rs < 3227) {
return 2;
}
else if (Rs >= 3227) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 3404) {
if (mcTcor < 97) {
if (RsDiff < 9) {
if (tcor < 92) {
if (gchDC < 5) {
return 8;
}
else if (gchDC >= 5) {
return 2;
}
}
else if (tcor >= 92) {
if (RsCsDiff < 36) {
if (diffRsCsdiff < -12) {
if (Rs < 2568) {
if (Cs < 1286) {
return 1;
}
else if (Cs >= 1286) {
return 8;
}
}
else if (Rs >= 2568) {
if (Rs < 2947) {
return 8;
}
else if (Rs >= 2947) {
return 2;
}
}
}
else if (diffRsCsdiff >= -12) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
else if (RsCsDiff >= 36) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (Rs < 2968) {
return 1;
}
else if (Rs >= 2968) {
return 8;
}
}
}
}
}
else if (RsDiff >= 9) {
if (Cs < 734) {
if (ssDCval < 117) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (ssDCval >= 117) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
else if (Cs >= 734) {
if (negBalance < 55) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (Rs < 2885) {
return 1;
}
else if (Rs >= 2885) {
return 2;
}
}
}
else if (negBalance >= 55) {
if (Rs < 2532) {
return 8;
}
else if (Rs >= 2532) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 5) {
if (TSCindex < 3) {
if (Rs < 2688) {
return 1;
}
else if (Rs >= 2688) {
if (SC < 4163) {
return 8;
}
else if (SC >= 4163) {
return 2;
}
}
}
else if (TSCindex >= 3) {
if (Cs < 536) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 2;
}
}
else if (Cs >= 536) {
if (Rs < 2482) {
return 8;
}
else if (Rs >= 2482) {
return 2;
}
}
}
}
else if (MVDiff >= 5) {
if (TSC < 6) {
if (CsDiff < 3) {
if (Rs < 2511) {
return 1;
}
else if (Rs >= 2511) {
if (TSC < 4) {
if (Rs < 2594) {
return 2;
}
else if (Rs >= 2594) {
if (Rs < 3007) {
return 1;
}
else if (Rs >= 3007) {
return 8;
}
}
}
else if (TSC >= 4) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (SC < 3572) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (SC >= 3572) {
if (posBalance < 88) {
return 8;
}
else if (posBalance >= 88) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (SC < 3381) {
if (posBalance < 50) {
if (Rs < 2741) {
return 2;
}
else if (Rs >= 2741) {
return 1;
}
}
else if (posBalance >= 50) {
return 1;
}
}
else if (SC >= 3381) {
if (diffMVdiffVal < -13) {
if (Rs < 2461) {
return 8;
}
else if (Rs >= 2461) {
return 1;
}
}
else if (diffMVdiffVal >= -13) {
if (posBalance < 63) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
else if (posBalance >= 63) {
if (Rs < 3123) {
if (Cs < 792) {
return 1;
}
else if (Cs >= 792) {
return 4;
}
}
else if (Rs >= 3123) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 3404) {
if (mcTcor < 98) {
if (refDCval < 134) {
if (TSC < 6) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (posBalance < 66) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (posBalance >= 66) {
return 2;
}
}
}
else if (TSC >= 6) {
if (Cs < 1227) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
if (RsDiff < 20) {
if (Rs < 4571) {
if (Rs < 3872) {
return 2;
}
else if (Rs >= 3872) {
return 4;
}
}
else if (Rs >= 4571) {
return 1;
}
}
else if (RsDiff >= 20) {
return 2;
}
}
}
else if (Cs >= 1227) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 2;
}
}
}
}
else if (refDCval >= 134) {
if (mcTcor < 94) {
if (RsDiff < 20) {
if (posBalance < 55) {
if (SC < 5137) {
if (Rs < 3872) {
return 8;
}
else if (Rs >= 3872) {
return 2;
}
}
else if (SC >= 5137) {
return 1;
}
}
else if (posBalance >= 55) {
return 4;
}
}
else if (RsDiff >= 20) {
return 2;
}
}
else if (mcTcor >= 94) {
if (posBalance < 46) {
if (diffTSC < -3) {
return 1;
}
else if (diffTSC >= -3) {
if (SC < 4413) {
return 8;
}
else if (SC >= 4413) {
return 2;
}
}
}
else if (posBalance >= 46) {
if (negBalance < 33) {
return 2;
}
else if (negBalance >= 33) {
if (RsDiff < 19) {
if (SC < 4487) {
return 2;
}
else if (SC >= 4487) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
if (MVDiff < 5) {
if (Cs < 947) {
return 2;
}
else if (Cs >= 947) {
return 1;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
}
}
else if (RsDiff >= 19) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (ssDCval < 142) {
if (Cs < 617) {
if (posBalance < 43) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (posBalance >= 43) {
return 2;
}
}
else if (Cs >= 617) {
if (diffMVdiffVal < -39) {
if (Rs < 5477) {
return 1;
}
else if (Rs >= 5477) {
return 2;
}
}
else if (diffMVdiffVal >= -39) {
return 2;
}
}
}
else if (ssDCval >= 142) {
if (Rs < 3963) {
if (SC < 4786) {
return 2;
}
else if (SC >= 4786) {
if (posBalance < 39) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
return 2;
}
}
else if (posBalance >= 39) {
if (Cs < 1098) {
return 2;
}
else if (Cs >= 1098) {
if (Rs < 3668) {
if (Rs < 3658) {
return 1;
}
else if (Rs >= 3658) {
return 2;
}
}
else if (Rs >= 3668) {
return 1;
}
}
}
}
}
else if (Rs >= 3963) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 14) {
if (mcTcor < 69) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
if (gchDC < 7) {
if (diffMVdiffVal < -8) {
return 2;
}
else if (diffMVdiffVal >= -8) {
return 4;
}
}
else if (gchDC >= 7) {
return 8;
}
}
}
else if (mcTcor >= 69) {
if (RsCsDiff < 128) {
if (ssDCval < 80) {
if (RsDiff < 7) {
if (posBalance < 45) {
if (diffMVdiffVal < -1) {
if (MVDiff < 21) {
if (Rs < 2540) {
return 4;
}
else if (Rs >= 2540) {
return 8;
}
}
else if (MVDiff >= 21) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (posBalance >= 45) {
if (Rs < 2600) {
return 8;
}
else if (Rs >= 2600) {
return 4;
}
}
}
else if (RsDiff >= 7) {
if (Rs < 2738) {
if (refDCval < 57) {
return 4;
}
else if (refDCval >= 57) {
if (ssDCval < 71) {
if (mu_mv_mag_sq < 58) {
return 2;
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 2539) {
return 8;
}
else if (Rs >= 2539) {
return 1;
}
}
}
else if (ssDCval >= 71) {
if (AFD < 14) {
if (diffRsCsdiff < -55) {
return 4;
}
else if (diffRsCsdiff >= -55) {
if (mu_mv_mag_sq < 37) {
return 1;
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 2565) {
return 8;
}
else if (Rs >= 2565) {
return 2;
}
}
}
}
else if (AFD >= 14) {
return 2;
}
}
}
}
else if (Rs >= 2738) {
if (Cs < 809) {
return 4;
}
else if (Cs >= 809) {
if (mcTcor < 94) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (Rs < 3046) {
return 4;
}
else if (Rs >= 3046) {
return 2;
}
}
}
else if (mcTcor >= 94) {
return 2;
}
}
}
}
}
else if (ssDCval >= 80) {
if (Rs < 4205) {
if (SC < 3563) {
if (tcor < 81) {
if (AFD < 18) {
if (Rs < 2626) {
return 8;
}
else if (Rs >= 2626) {
if (diffAFD < -3) {
if (Rs < 2917) {
return 2;
}
else if (Rs >= 2917) {
return 8;
}
}
else if (diffAFD >= -3) {
return 4;
}
}
}
else if (AFD >= 18) {
if (Cs < 221) {
return 1;
}
else if (Cs >= 221) {
return 8;
}
}
}
else if (tcor >= 81) {
if (posBalance < 63) {
if (RsCsDiff < 108) {
if (refDCval < 90) {
return 1;
}
else if (refDCval >= 90) {
if (AFD < 9) {
if (Cs < 341) {
return 8;
}
else if (Cs >= 341) {
if (Cs < 374) {
return 1;
}
else if (Cs >= 374) {
return 8;
}
}
}
else if (AFD >= 9) {
if (ssDCval < 121) {
if (AFD < 13) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
if (negBalance < 56) {
return 2;
}
else if (negBalance >= 56) {
if (Rs < 2531) {
return 8;
}
else if (Rs >= 2531) {
if (Rs < 2813) {
return 1;
}
else if (Rs >= 2813) {
return 4;
}
}
}
}
}
else if (AFD >= 13) {
if (SC < 3163) {
if (SC < 3025) {
if (SC < 2771) {
return 1;
}
else if (SC >= 2771) {
if (negBalance < 60) {
return 8;
}
else if (negBalance >= 60) {
if (Cs < 436) {
return 8;
}
else if (Cs >= 436) {
return 1;
}
}
}
}
else if (SC >= 3025) {
return 1;
}
}
else if (SC >= 3163) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
else if (ssDCval >= 121) {
return 1;
}
}
}
}
else if (RsCsDiff >= 108) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (diffRsCsdiff < 20) {
if (Rs < 2635) {
return 1;
}
else if (Rs >= 2635) {
if (Rs < 2781) {
return 2;
}
else if (Rs >= 2781) {
return 1;
}
}
}
else if (diffRsCsdiff >= 20) {
if (Rs < 2555) {
return 2;
}
else if (Rs >= 2555) {
return 1;
}
}
}
}
}
else if (SC >= 3563) {
if (Cs < 1184) {
if (mu_mv_mag_sq < 60) {
if (CsDiff < 3) {
if (Cs < 439) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
else if (Cs >= 439) {
if (AFD < 20) {
if (refDCval < 119) {
return 8;
}
else if (refDCval >= 119) {
if (MVDiff < 21) {
return 2;
}
else if (MVDiff >= 21) {
return 8;
}
}
}
else if (AFD >= 20) {
if (Rs < 3177) {
return 4;
}
else if (Rs >= 3177) {
return 2;
}
}
}
}
else if (CsDiff >= 3) {
if (Cs < 1006) {
if (Cs < 864) {
if (posBalance < 69) {
if (Rs < 3150) {
if (Cs < 812) {
return 1;
}
else if (Cs >= 812) {
return 8;
}
}
else if (Rs >= 3150) {
if (ssDCval < 130) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 8;
}
}
else if (ssDCval >= 130) {
return 4;
}
}
}
else if (posBalance >= 69) {
if (Cs < 826) {
return 2;
}
else if (Cs >= 826) {
return 4;
}
}
}
else if (Cs >= 864) {
if (diffRsCsdiff < -15) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (diffRsCsdiff >= -15) {
if (ssDCval < 124) {
return 1;
}
else if (ssDCval >= 124) {
return 2;
}
}
}
}
else if (Cs >= 1006) {
if (diffAFD < -5) {
if (Rs < 2831) {
return 4;
}
else if (Rs >= 2831) {
return 2;
}
}
else if (diffAFD >= -5) {
if (diffTSC < 2) {
if (diffMVdiffVal < -34) {
return 1;
}
else if (diffMVdiffVal >= -34) {
if (diffAFD < -2) {
if (Cs < 1054) {
return 2;
}
else if (Cs >= 1054) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (diffTSC >= 2) {
if (Rs < 2594) {
return 2;
}
else if (Rs >= 2594) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 60) {
if (ssDCval < 117) {
if (Rs < 3344) {
return 8;
}
else if (Rs >= 3344) {
return 4;
}
}
else if (ssDCval >= 117) {
return 2;
}
}
}
else if (Cs >= 1184) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (diffRsCsdiff < -51) {
if (ssDCval < 110) {
if (Rs < 2808) {
return 8;
}
else if (Rs >= 2808) {
return 4;
}
}
else if (ssDCval >= 110) {
return 1;
}
}
else if (diffRsCsdiff >= -51) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
if (negBalance < 68) {
return 8;
}
else if (negBalance >= 68) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 4205) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (Rs < 4521) {
return 2;
}
else if (Rs >= 4521) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 128) {
if (refDCval < 135) {
if (ssDCval < 112) {
if (refDCval < 72) {
if (Rs < 2907) {
if (TSC < 8) {
if (diffRsCsdiff < 27) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (diffRsCsdiff >= 27) {
if (Rs < 2514) {
if (Rs < 2442) {
return 1;
}
else if (Rs >= 2442) {
return 8;
}
}
else if (Rs >= 2514) {
return 4;
}
}
}
else if (TSC >= 8) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
if (TSC < 10) {
if (Cs < 111) {
return 1;
}
else if (Cs >= 111) {
return 4;
}
}
else if (TSC >= 10) {
if (RsDiff < 19) {
return 1;
}
else if (RsDiff >= 19) {
return 2;
}
}
}
}
}
else if (Rs >= 2907) {
if (posBalance < 66) {
if (posBalance < 33) {
if (Rs < 4972) {
if (Rs < 3544) {
return 2;
}
else if (Rs >= 3544) {
return 4;
}
}
else if (Rs >= 4972) {
return 8;
}
}
else if (posBalance >= 33) {
if (Rs < 3097) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 4;
}
}
}
else if (Rs >= 3097) {
if (SC < 3475) {
return 1;
}
else if (SC >= 3475) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
return 1;
}
}
}
}
}
else if (posBalance >= 66) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
}
}
else if (refDCval >= 72) {
if (mcTcor < 92) {
if (RsCsDiff < 357) {
if (MVDiff < 32) {
if (TSC < 12) {
if (CsDiff < 7) {
if (diffMVdiffVal < -11) {
if (SC < 3325) {
return 1;
}
else if (SC >= 3325) {
return 4;
}
}
else if (diffMVdiffVal >= -11) {
if (tcor < 75) {
return 2;
}
else if (tcor >= 75) {
if (negBalance < 52) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
return 2;
}
}
else if (negBalance >= 52) {
if (Cs < 1002) {
return 4;
}
else if (Cs >= 1002) {
return 2;
}
}
}
}
}
else if (CsDiff >= 7) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
if (RsCsDiff < 248) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 4;
}
}
else if (RsCsDiff >= 248) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
return 2;
}
}
}
}
}
else if (TSC >= 12) {
if (AFD < 19) {
if (Cs < 1029) {
return 1;
}
else if (Cs >= 1029) {
return 8;
}
}
else if (AFD >= 19) {
if (Cs < 726) {
return 2;
}
else if (Cs >= 726) {
return 4;
}
}
}
}
else if (MVDiff >= 32) {
if (mcTcor < 82) {
if (RsDiff < 15) {
if (Rs < 2823) {
if (Cs < 801) {
return 8;
}
else if (Cs >= 801) {
return 2;
}
}
else if (Rs >= 2823) {
return 4;
}
}
else if (RsDiff >= 15) {
return 4;
}
}
else if (mcTcor >= 82) {
if (CsDiff < 3) {
if (MVDiff < 51) {
if (posBalance < 69) {
return 8;
}
else if (posBalance >= 69) {
return 1;
}
}
else if (MVDiff >= 51) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (AFD < 25) {
if (SC < 3261) {
if (Cs < 689) {
return 8;
}
else if (Cs >= 689) {
return 1;
}
}
else if (SC >= 3261) {
if (Cs < 1216) {
return 8;
}
else if (Cs >= 1216) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 2;
}
}
}
}
else if (AFD >= 25) {
if (Rs < 2755) {
if (SC < 3531) {
return 1;
}
else if (SC >= 3531) {
if (Cs < 1191) {
return 2;
}
else if (Cs >= 1191) {
return 1;
}
}
}
else if (Rs >= 2755) {
if (Rs < 2970) {
return 4;
}
else if (Rs >= 2970) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 357) {
if (CsDiff < 6) {
if (Cs < 1034) {
if (tcor < 80) {
return 2;
}
else if (tcor >= 80) {
if (Rs < 3990) {
return 2;
}
else if (Rs >= 3990) {
return 4;
}
}
}
else if (Cs >= 1034) {
if (Cs < 1258) {
return 4;
}
else if (Cs >= 1258) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (AFD < 26) {
if (tcor < 69) {
if (Rs < 3591) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (Rs >= 3591) {
return 2;
}
}
else if (tcor >= 69) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
if (Rs < 4296) {
return 8;
}
else if (Rs >= 4296) {
return 2;
}
}
}
}
else if (AFD >= 26) {
if (diffTSC < 2) {
return 2;
}
else if (diffTSC >= 2) {
return 1;
}
}
}
}
}
else if (mcTcor >= 92) {
if (gchDC < 5) {
if (RsCsDiff < 248) {
if (ssDCval < 91) {
if (Rs < 3752) {
if (TSC < 9) {
if (negBalance < 44) {
if (posBalance < 59) {
if (Cs < 1093) {
return 1;
}
else if (Cs >= 1093) {
return 4;
}
}
else if (posBalance >= 59) {
return 2;
}
}
else if (negBalance >= 44) {
if (MVDiff < 42) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 2;
}
}
else if (MVDiff >= 42) {
if (Rs < 2916) {
return 1;
}
else if (Rs >= 2916) {
return 4;
}
}
}
}
else if (TSC >= 9) {
return 2;
}
}
else if (Rs >= 3752) {
if (Cs < 1106) {
return 4;
}
else if (Cs >= 1106) {
return 2;
}
}
}
else if (ssDCval >= 91) {
if (diffRsCsdiff < -26) {
if (posBalance < 69) {
if (diffTSC < -4) {
if (Rs < 4863) {
return 4;
}
else if (Rs >= 4863) {
return 2;
}
}
else if (diffTSC >= -4) {
if (Cs < 1249) {
return 2;
}
else if (Cs >= 1249) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (posBalance >= 69) {
if (Rs < 3378) {
return 4;
}
else if (Rs >= 3378) {
if (Rs < 3651) {
return 2;
}
else if (Rs >= 3651) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= -26) {
if (AFD < 20) {
if (CsDiff < 4) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Cs < 360) {
return 4;
}
else if (Cs >= 360) {
if (RsDiff < 12) {
return 2;
}
else if (RsDiff >= 12) {
return 8;
}
}
}
}
else if (CsDiff >= 4) {
if (AFD < 18) {
if (diffMVdiffVal < -12) {
if (Rs < 4076) {
return 1;
}
else if (Rs >= 4076) {
return 8;
}
}
else if (diffMVdiffVal >= -12) {
if (diffTSC < 3) {
return 2;
}
else if (diffTSC >= 3) {
return 8;
}
}
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (AFD >= 20) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 248) {
if (Rs < 4190) {
if (refDCval < 111) {
if (diffMVdiffVal < -23) {
if (diffMVdiffVal < -40) {
return 2;
}
else if (diffMVdiffVal >= -40) {
if (Rs < 3286) {
return 1;
}
else if (Rs >= 3286) {
return 2;
}
}
}
else if (diffMVdiffVal >= -23) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
if (refDCval < 109) {
return 2;
}
else if (refDCval >= 109) {
if (RsDiff < 17) {
return 1;
}
else if (RsDiff >= 17) {
return 2;
}
}
}
}
}
else if (refDCval >= 111) {
if (Rs < 3500) {
return 1;
}
else if (Rs >= 3500) {
return 2;
}
}
}
else if (Rs >= 4190) {
if (ssDCval < 87) {
if (refDCval < 80) {
return 4;
}
else if (refDCval >= 80) {
if (Rs < 6858) {
return 1;
}
else if (Rs >= 6858) {
return 2;
}
}
}
else if (ssDCval >= 87) {
if (Cs < 654) {
return 2;
}
else if (Cs >= 654) {
if (Cs < 1188) {
if (Rs < 5756) {
return 8;
}
else if (Rs >= 5756) {
return 2;
}
}
else if (Cs >= 1188) {
return 2;
}
}
}
}
}
}
else if (gchDC >= 5) {
if (tcor < 92) {
if (tcor < 76) {
if (MVDiff < 36) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
return 1;
}
}
else if (MVDiff >= 36) {
return 8;
}
}
else if (tcor >= 76) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (Cs < 249) {
return 2;
}
else if (Cs >= 249) {
if (refDCval < 88) {
return 2;
}
else if (refDCval >= 88) {
return 1;
}
}
}
}
}
else if (tcor >= 92) {
return 8;
}
}
}
}
}
else if (ssDCval >= 112) {
if (Rs < 3083) {
if (mcTcor < 95) {
if (RsDiff < 18) {
if (AFD < 17) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 4;
}
}
else if (AFD >= 17) {
if (RsCsDiff < 203) {
if (ssDCval < 127) {
return 8;
}
else if (ssDCval >= 127) {
return 1;
}
}
else if (RsCsDiff >= 203) {
if (tcor < 61) {
if (MVDiff < 43) {
return 8;
}
else if (MVDiff >= 43) {
if (Cs < 949) {
return 1;
}
else if (Cs >= 949) {
return 4;
}
}
}
else if (tcor >= 61) {
if (MVDiff < 23) {
if (Cs < 803) {
return 2;
}
else if (Cs >= 803) {
return 1;
}
}
else if (MVDiff >= 23) {
if (diffRsCsdiff < 26) {
if (gchDC < 10) {
return 1;
}
else if (gchDC >= 10) {
return 2;
}
}
else if (diffRsCsdiff >= 26) {
if (refDCval < 122) {
if (diffTSC < -1) {
if (Cs < 910) {
return 8;
}
else if (Cs >= 910) {
return 1;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (refDCval >= 122) {
if (Rs < 2605) {
return 2;
}
else if (Rs >= 2605) {
return 8;
}
}
}
}
}
}
}
}
else if (RsDiff >= 18) {
if (MVDiff < 54) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 1;
}
}
else if (MVDiff >= 54) {
if (ssDCval < 133) {
return 2;
}
else if (ssDCval >= 133) {
return 1;
}
}
}
}
else if (mcTcor >= 95) {
if (CsDiff < 3) {
if (SC < 3007) {
if (Cs < 158) {
return 2;
}
else if (Cs >= 158) {
return 1;
}
}
else if (SC >= 3007) {
if (SC < 3105) {
return 8;
}
else if (SC >= 3105) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (Cs < 1181) {
return 1;
}
else if (Cs >= 1181) {
if (Rs < 2547) {
return 1;
}
else if (Rs >= 2547) {
return 8;
}
}
}
}
}
else if (Rs >= 3083) {
if (Cs < 507) {
if (Rs < 3418) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (MVDiff < 50) {
if (Rs < 3253) {
return 1;
}
else if (Rs >= 3253) {
return 4;
}
}
else if (MVDiff >= 50) {
return 8;
}
}
}
else if (Rs >= 3418) {
if (diffMVdiffVal < 24) {
return 2;
}
else if (diffMVdiffVal >= 24) {
if (Rs < 3988) {
return 4;
}
else if (Rs >= 3988) {
return 2;
}
}
}
}
else if (Cs >= 507) {
if (MVDiff < 30) {
if (ssDCval < 132) {
if (CsDiff < 5) {
if (mu_mv_mag_sq < 2) {
if (negBalance < 44) {
if (Rs < 4692) {
return 1;
}
else if (Rs >= 4692) {
return 2;
}
}
else if (negBalance >= 44) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
return 2;
}
}
else if (CsDiff >= 5) {
if (RsCsDiff < 439) {
if (diffRsCsdiff < -115) {
if (ssDCval < 119) {
return 2;
}
else if (ssDCval >= 119) {
return 1;
}
}
else if (diffRsCsdiff >= -115) {
if (negBalance < 49) {
if (ssDCval < 126) {
return 2;
}
else if (ssDCval >= 126) {
if (Rs < 4129) {
return 8;
}
else if (Rs >= 4129) {
return 1;
}
}
}
else if (negBalance >= 49) {
return 8;
}
}
}
else if (RsCsDiff >= 439) {
if (ssDCval < 125) {
if (SC < 4262) {
return 1;
}
else if (SC >= 4262) {
return 2;
}
}
else if (ssDCval >= 125) {
if (ssDCval < 128) {
if (Rs < 5590) {
return 2;
}
else if (Rs >= 5590) {
return 8;
}
}
else if (ssDCval >= 128) {
return 2;
}
}
}
}
}
else if (ssDCval >= 132) {
if (ssDCval < 136) {
if (RsDiff < 22) {
if (Rs < 6839) {
return 1;
}
else if (Rs >= 6839) {
return 2;
}
}
else if (RsDiff >= 22) {
if (diffMVdiffVal < -10) {
return 8;
}
else if (diffMVdiffVal >= -10) {
return 2;
}
}
}
else if (ssDCval >= 136) {
return 2;
}
}
}
else if (MVDiff >= 30) {
if (RsCsDiff < 365) {
if (refDCval < 126) {
if (posBalance < 44) {
if (Rs < 4407) {
return 1;
}
else if (Rs >= 4407) {
return 2;
}
}
else if (posBalance >= 44) {
if (MVDiff < 35) {
if (Rs < 3360) {
return 2;
}
else if (Rs >= 3360) {
return 8;
}
}
else if (MVDiff >= 35) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
}
else if (refDCval >= 126) {
if (RsCsDiff < 213) {
return 8;
}
else if (RsCsDiff >= 213) {
return 1;
}
}
}
else if (RsCsDiff >= 365) {
if (CsDiff < 10) {
if (tcor < 81) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 121) {
if (Cs < 1327) {
if (mcTcor < 95) {
return 2;
}
else if (mcTcor >= 95) {
if (MVDiff < 31) {
return 2;
}
else if (MVDiff >= 31) {
return 8;
}
}
}
else if (Cs >= 1327) {
return 1;
}
}
else if (ssDCval >= 121) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Rs < 4849) {
return 2;
}
else if (Rs >= 4849) {
if (posBalance < 60) {
if (Rs < 6326) {
return 2;
}
else if (Rs >= 6326) {
return 8;
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (gchDC < 1) {
if (Cs < 769) {
return 2;
}
else if (Cs >= 769) {
if (Rs < 3592) {
return 2;
}
else if (Rs >= 3592) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (RsDiff < 27) {
return 2;
}
else if (RsDiff >= 27) {
if (Rs < 5093) {
return 8;
}
else if (Rs >= 5093) {
return 2;
}
}
}
}
}
else if (tcor >= 81) {
if (CsDiff < 6) {
if (refDCval < 112) {
return 1;
}
else if (refDCval >= 112) {
return 2;
}
}
else if (CsDiff >= 6) {
if (Cs < 1226) {
return 1;
}
else if (Cs >= 1226) {
return 8;
}
}
}
}
else if (CsDiff >= 10) {
if (TSC < 23) {
return 1;
}
else if (TSC >= 23) {
return 2;
}
}
}
}
}
}
}
}
else if (refDCval >= 135) {
if (posBalance < 59) {
if (Rs < 3385) {
if (Cs < 1120) {
if (ssDCval < 131) {
if (Rs < 2580) {
return 2;
}
else if (Rs >= 2580) {
return 1;
}
}
else if (ssDCval >= 131) {
if (MVDiff < 117) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (MVDiff >= 117) {
if (ssDCval < 153) {
return 1;
}
else if (ssDCval >= 153) {
return 2;
}
}
}
}
else if (Cs >= 1120) {
if (SC < 3807) {
return 2;
}
else if (SC >= 3807) {
if (Rs < 2785) {
return 8;
}
else if (Rs >= 2785) {
return 1;
}
}
}
}
else if (Rs >= 3385) {
if (CsDiff < 4) {
if (tcor < 93) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
return 4;
}
}
else if (tcor >= 93) {
return 1;
}
}
else if (CsDiff >= 4) {
if (TSC < 11) {
if (Rs < 4534) {
if (MVDiff < 75) {
return 1;
}
else if (MVDiff >= 75) {
return 8;
}
}
else if (Rs >= 4534) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
if (Cs < 1275) {
if (Cs < 1212) {
return 2;
}
else if (Cs >= 1212) {
if (Rs < 6171) {
return 8;
}
else if (Rs >= 6171) {
return 2;
}
}
}
else if (Cs >= 1275) {
return 1;
}
}
}
}
else if (TSC >= 11) {
if (mu_mv_mag_sq < 35) {
if (Cs < 748) {
if (Rs < 3749) {
return 1;
}
else if (Rs >= 3749) {
return 2;
}
}
else if (Cs >= 748) {
if (SC < 7126) {
if (TSC < 15) {
return 1;
}
else if (TSC >= 15) {
if (TSC < 16) {
return 2;
}
else if (TSC >= 16) {
return 1;
}
}
}
else if (SC >= 7126) {
if (Cs < 1025) {
return 1;
}
else if (Cs >= 1025) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 35) {
return 2;
}
}
}
}
}
else if (posBalance >= 59) {
if (gchDC < 2) {
if (mcTcor < 94) {
if (AFD < 20) {
if (posBalance < 68) {
return 4;
}
else if (posBalance >= 68) {
return 1;
}
}
else if (AFD >= 20) {
if (Rs < 4059) {
if (Rs < 2486) {
return 8;
}
else if (Rs >= 2486) {
return 1;
}
}
else if (Rs >= 4059) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 4821) {
return 1;
}
else if (Rs >= 4821) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (MVDiff < 18) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (MVDiff >= 18) {
if (MVDiff < 86) {
if (ssDCval < 140) {
if (Rs < 3060) {
return 4;
}
else if (Rs >= 3060) {
if (Rs < 3494) {
return 1;
}
else if (Rs >= 3494) {
return 8;
}
}
}
else if (ssDCval >= 140) {
return 1;
}
}
else if (MVDiff >= 86) {
if (Rs < 2741) {
return 1;
}
else if (Rs >= 2741) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1333) {
if (mu_mv_mag_sq < 107) {
if (RsCsDiff < 101) {
if (ssDCval < 105) {
if (negBalance < 38) {
if (SC < 10689) {
if (mu_mv_mag_sq < 46) {
if (Rs < 810) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (Rs >= 810) {
if (Rs < 3776) {
if (Rs < 1289) {
if (Rs < 1005) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Rs < 928) {
return 2;
}
else if (Rs >= 928) {
return 1;
}
}
}
else if (Rs >= 1005) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 2;
}
}
}
else if (Rs >= 1289) {
if (CsDiff < 6) {
if (negBalance < 33) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
else if (negBalance >= 33) {
if (RsCsDiff < 70) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (posBalance < 40) {
return 2;
}
else if (posBalance >= 40) {
return 8;
}
}
}
else if (RsCsDiff >= 70) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (AFD < 11) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 2944) {
return 8;
}
else if (Rs >= 2944) {
return 2;
}
}
}
else if (AFD >= 11) {
if (Rs < 1944) {
if (Rs < 1329) {
return 1;
}
else if (Rs >= 1329) {
return 8;
}
}
else if (Rs >= 1944) {
if (Cs < 1623) {
return 4;
}
else if (Cs >= 1623) {
return 1;
}
}
}
}
}
}
else if (Rs >= 3776) {
if (ssDCval < 81) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
if (Rs < 4322) {
return 2;
}
else if (Rs >= 4322) {
return 8;
}
}
}
else if (ssDCval >= 81) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (Rs < 4195) {
return 2;
}
else if (Rs >= 4195) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (ssDCval < 87) {
if (posBalance < 46) {
if (MVDiff < 7) {
if (Rs < 3557) {
if (RsCsDiff < 57) {
if (tcor < 96) {
if (RsDiff < 4) {
if (ssDCval < 67) {
if (Rs < 1443) {
if (Rs < 1026) {
return 4;
}
else if (Rs >= 1026) {
return 8;
}
}
else if (Rs >= 1443) {
return 2;
}
}
else if (ssDCval >= 67) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (RsDiff >= 4) {
if (Rs < 2867) {
return 8;
}
else if (Rs >= 2867) {
return 4;
}
}
}
else if (tcor >= 96) {
if (Rs < 3014) {
if (MVDiff < 2) {
if (Rs < 2555) {
return 8;
}
else if (Rs >= 2555) {
return 1;
}
}
else if (MVDiff >= 2) {
if (diffRsCsdiff < 8) {
if (Rs < 335) {
return 4;
}
else if (Rs >= 335) {
return 1;
}
}
else if (diffRsCsdiff >= 8) {
return 8;
}
}
}
else if (Rs >= 3014) {
if (Rs < 3428) {
return 4;
}
else if (Rs >= 3428) {
return 2;
}
}
}
}
else if (RsCsDiff >= 57) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 3053) {
return 1;
}
else if (Rs >= 3053) {
return 2;
}
}
}
}
}
else if (Rs >= 3557) {
if (CsDiff < 6) {
if (SC < 5775) {
if (ssDCval < 80) {
return 4;
}
else if (ssDCval >= 80) {
return 2;
}
}
else if (SC >= 5775) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Rs < 5394) {
if (AFD < 8) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 4;
}
}
else if (AFD >= 8) {
return 4;
}
}
else if (Rs >= 5394) {
return 4;
}
}
}
}
else if (CsDiff >= 6) {
if (Rs < 4139) {
return 4;
}
else if (Rs >= 4139) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 55) {
if (AFD < 8) {
if (Rs < 692) {
return 1;
}
else if (Rs >= 692) {
if (mu_mv_mag_sq < 68) {
if (Rs < 3717) {
return 8;
}
else if (Rs >= 3717) {
return 4;
}
}
else if (mu_mv_mag_sq >= 68) {
return 8;
}
}
}
else if (AFD >= 8) {
if (Cs < 1576) {
return 2;
}
else if (Cs >= 1576) {
return 4;
}
}
}
else if (ssDCval >= 55) {
if (TSC < 5) {
if (Rs < 1125) {
if (Rs < 384) {
return 4;
}
else if (Rs >= 384) {
if (Rs < 859) {
return 1;
}
else if (Rs >= 859) {
return 8;
}
}
}
else if (Rs >= 1125) {
if (MVDiff < 8) {
if (Rs < 2129) {
return 2;
}
else if (Rs >= 2129) {
return 4;
}
}
else if (MVDiff >= 8) {
return 4;
}
}
}
else if (TSC >= 5) {
if (Cs < 1841) {
if (ssDCval < 77) {
if (Rs < 1724) {
return 4;
}
else if (Rs >= 1724) {
return 2;
}
}
else if (ssDCval >= 77) {
if (Cs < 1660) {
return 8;
}
else if (Cs >= 1660) {
return 1;
}
}
}
else if (Cs >= 1841) {
return 4;
}
}
}
}
}
else if (posBalance >= 46) {
if (Rs < 2861) {
if (mu_mv_mag_sq < 86) {
if (Rs < 1083) {
if (ssDCval < 58) {
return 2;
}
else if (ssDCval >= 58) {
if (TSC < 5) {
if (MVDiff < 4) {
if (Rs < 1012) {
return 4;
}
else if (Rs >= 1012) {
return 2;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (Rs >= 1083) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 58) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
if (Cs < 1597) {
return 4;
}
else if (Cs >= 1597) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (diffRsCsdiff < 15) {
if (diffTSC < -1) {
if (Rs < 2029) {
return 8;
}
else if (Rs >= 2029) {
return 4;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (diffRsCsdiff >= 15) {
if (Rs < 2051) {
if (Rs < 1327) {
return 4;
}
else if (Rs >= 1327) {
return 8;
}
}
else if (Rs >= 2051) {
return 4;
}
}
}
}
else if (mcTcor >= 95) {
if (AFD < 10) {
if (TSC < 7) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 8;
}
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (SC < 4584) {
if (refDCval < 46) {
if (Rs < 1491) {
return 1;
}
else if (Rs >= 1491) {
return 4;
}
}
else if (refDCval >= 46) {
if (Cs < 1690) {
if (negBalance < 19) {
return 4;
}
else if (negBalance >= 19) {
if (Rs < 2346) {
if (Rs < 1572) {
if (SC < 2827) {
return 8;
}
else if (SC >= 2827) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (Rs >= 1572) {
return 8;
}
}
else if (Rs >= 2346) {
if (Cs < 1609) {
return 4;
}
else if (Cs >= 1609) {
return 8;
}
}
}
}
else if (Cs >= 1690) {
if (mu_mv_mag_sq < 89) {
if (Rs < 1400) {
return 1;
}
else if (Rs >= 1400) {
return 8;
}
}
else if (mu_mv_mag_sq >= 89) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
}
else if (SC >= 4584) {
if (mu_mv_mag_sq < 98) {
return 1;
}
else if (mu_mv_mag_sq >= 98) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 4;
}
}
}
}
}
else if (Rs >= 2861) {
if (Cs < 2068) {
if (refDCval < 79) {
if (tcor < 91) {
if (Cs < 1572) {
return 2;
}
else if (Cs >= 1572) {
return 8;
}
}
else if (tcor >= 91) {
return 2;
}
}
else if (refDCval >= 79) {
if (Rs < 3377) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (Rs < 3049) {
return 2;
}
else if (Rs >= 3049) {
return 1;
}
}
}
else if (Rs >= 3377) {
return 2;
}
}
}
else if (Cs >= 2068) {
if (CsDiff < 6) {
if (Rs < 5559) {
return 8;
}
else if (Rs >= 5559) {
return 2;
}
}
else if (CsDiff >= 6) {
if (Rs < 2969) {
return 8;
}
else if (Rs >= 2969) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (MVDiff < 7) {
if (posBalance < 41) {
if (RsCsDiff < 38) {
if (negBalance < 28) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
return 4;
}
}
else if (negBalance >= 28) {
if (AFD < 5) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 8;
}
}
else if (AFD >= 5) {
if (Rs < 3083) {
return 1;
}
else if (Rs >= 3083) {
return 4;
}
}
}
}
else if (RsCsDiff >= 38) {
if (Rs < 3862) {
if (Rs < 3083) {
return 1;
}
else if (Rs >= 3083) {
return 2;
}
}
else if (Rs >= 3862) {
if (Cs < 1677) {
if (Rs < 4480) {
return 4;
}
else if (Rs >= 4480) {
return 8;
}
}
else if (Cs >= 1677) {
return 8;
}
}
}
}
else if (posBalance >= 41) {
if (Cs < 2498) {
if (Rs < 3425) {
if (Cs < 1519) {
if (MVDiff < 4) {
if (ssDCval < 96) {
if (Rs < 1811) {
return 8;
}
else if (Rs >= 1811) {
return 2;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (MVDiff >= 4) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 8;
}
}
}
else if (Cs >= 1519) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (Rs >= 3425) {
if (posBalance < 44) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
if (Cs < 1520) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 2;
}
}
else if (Cs >= 1520) {
return 8;
}
}
}
else if (posBalance >= 44) {
if (diffRsCsdiff < 5) {
return 2;
}
else if (diffRsCsdiff >= 5) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Cs < 1355) {
return 4;
}
else if (Cs >= 1355) {
return 2;
}
}
}
}
}
}
else if (Cs >= 2498) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (posBalance < 30) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (posBalance >= 30) {
if (TSC < 7) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (diffRsCsdiff < -25) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (Rs < 3977) {
return 8;
}
else if (Rs >= 3977) {
return 1;
}
}
}
else if (diffRsCsdiff >= -25) {
return 8;
}
}
}
else if (TSC >= 7) {
if (Rs < 3943) {
if (MVDiff < 9) {
if (Rs < 2889) {
return 4;
}
else if (Rs >= 2889) {
return 8;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (Rs >= 3943) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 10689) {
if (Rs < 5733) {
return 8;
}
else if (Rs >= 5733) {
return 2;
}
}
}
else if (negBalance >= 38) {
if (Rs < 1479) {
if (RsCsDiff < 51) {
if (ssDCval < 83) {
if (TSC < 10) {
if (MVDiff < 4) {
if (ssDCval < 81) {
if (negBalance < 43) {
if (tcor < 91) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (diffRsCsdiff < 12) {
return 4;
}
else if (diffRsCsdiff >= 12) {
return 1;
}
}
}
else if (tcor >= 91) {
if (mcTcor < 97) {
if (diffRsCsdiff < 6) {
return 8;
}
else if (diffRsCsdiff >= 6) {
if (Cs < 1508) {
return 4;
}
else if (Cs >= 1508) {
return 8;
}
}
}
else if (mcTcor >= 97) {
return 4;
}
}
}
else if (negBalance >= 43) {
return 8;
}
}
else if (ssDCval >= 81) {
return 1;
}
}
else if (MVDiff >= 4) {
if (Rs < 1303) {
if (mcTcor < 94) {
if (negBalance < 55) {
if (Rs < 669) {
return 2;
}
else if (Rs >= 669) {
if (Cs < 1705) {
return 4;
}
else if (Cs >= 1705) {
return 8;
}
}
}
else if (negBalance >= 55) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 1113) {
if (refDCval < 59) {
return 2;
}
else if (refDCval >= 59) {
if (MVDiff < 15) {
if (posBalance < 51) {
if (Cs < 2227) {
return 1;
}
else if (Cs >= 2227) {
return 8;
}
}
else if (posBalance >= 51) {
if (mcTcor < 97) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Rs < 503) {
return 1;
}
else if (Rs >= 503) {
return 2;
}
}
}
else if (mcTcor >= 97) {
return 2;
}
}
}
else if (MVDiff >= 15) {
return 8;
}
}
}
else if (Rs >= 1113) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
else if (Rs >= 1303) {
if (posBalance < 51) {
if (Cs < 1535) {
return 4;
}
else if (Cs >= 1535) {
if (Rs < 1324) {
return 1;
}
else if (Rs >= 1324) {
return 8;
}
}
}
else if (posBalance >= 51) {
return 1;
}
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (ssDCval >= 83) {
if (mu_mv_mag_sq < 3) {
if (Cs < 1400) {
return 4;
}
else if (Cs >= 1400) {
if (RsDiff < 5) {
if (refDCval < 103) {
return 8;
}
else if (refDCval >= 103) {
return 4;
}
}
else if (RsDiff >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 100) {
if (Cs < 1814) {
return 8;
}
else if (Cs >= 1814) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
return 8;
}
}
}
else if (ssDCval >= 100) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 51) {
if (tcor < 88) {
if (ssDCval < 56) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
if (TSC < 8) {
if (Rs < 1425) {
return 2;
}
else if (Rs >= 1425) {
if (Rs < 1469) {
return 8;
}
else if (Rs >= 1469) {
return 1;
}
}
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (CsDiff < 7) {
if (ssDCval < 74) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
if (Rs < 1111) {
return 2;
}
else if (Rs >= 1111) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (TSC < 7) {
if (Rs < 1030) {
return 4;
}
else if (Rs >= 1030) {
return 8;
}
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (CsDiff >= 7) {
if (MVDiff < 21) {
if (negBalance < 56) {
if (mcTcor < 90) {
if (Cs < 2028) {
return 4;
}
else if (Cs >= 2028) {
return 2;
}
}
else if (mcTcor >= 90) {
if (posBalance < 60) {
if (Rs < 632) {
return 4;
}
else if (Rs >= 632) {
return 8;
}
}
else if (posBalance >= 60) {
if (Cs < 1628) {
if (Cs < 1488) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (Cs >= 1488) {
return 8;
}
}
else if (Cs >= 1628) {
if (Cs < 1834) {
return 4;
}
else if (Cs >= 1834) {
return 8;
}
}
}
}
}
else if (negBalance >= 56) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (MVDiff >= 21) {
if (diffAFD < -4) {
if (Cs < 1379) {
return 1;
}
else if (Cs >= 1379) {
return 8;
}
}
else if (diffAFD >= -4) {
if (mcTcor < 87) {
if (Rs < 1224) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 4;
}
}
else if (Rs >= 1224) {
return 4;
}
}
else if (mcTcor >= 87) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 88) {
if (TSC < 5) {
if (SC < 2871) {
return 1;
}
else if (SC >= 2871) {
return 8;
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 60) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (AFD < 12) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
if (Rs < 769) {
return 2;
}
else if (Rs >= 769) {
return 8;
}
}
}
else if (AFD >= 12) {
if (Rs < 1230) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (Rs >= 1230) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 60) {
if (Cs < 1617) {
return 4;
}
else if (Cs >= 1617) {
if (Rs < 1369) {
return 8;
}
else if (Rs >= 1369) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 1479) {
if (CsDiff < 7) {
if (mcTcor < 93) {
if (posBalance < 70) {
if (ssDCval < 90) {
if (refDCval < 80) {
if (RsDiff < 9) {
if (Cs < 1439) {
if (MVDiff < 17) {
if (Rs < 3054) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (Rs < 2119) {
return 2;
}
else if (Rs >= 2119) {
return 8;
}
}
}
else if (Rs >= 3054) {
return 4;
}
}
else if (MVDiff >= 17) {
if (mu_mv_mag_sq < 3) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Rs < 1948) {
return 1;
}
else if (Rs >= 1948) {
return 8;
}
}
}
}
}
else if (Cs >= 1439) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (mu_mv_mag_sq < 91) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 91) {
if (mu_mv_mag_sq < 99) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 1980) {
return 4;
}
else if (Rs >= 1980) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 99) {
return 8;
}
}
}
}
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (refDCval >= 80) {
if (negBalance < 56) {
if (Cs < 2736) {
if (mu_mv_mag_sq < 77) {
if (Cs < 1485) {
if (RsCsDiff < 87) {
if (Cs < 1481) {
return 8;
}
else if (Cs >= 1481) {
return 2;
}
}
else if (RsCsDiff >= 87) {
return 2;
}
}
else if (Cs >= 1485) {
return 8;
}
}
else if (mu_mv_mag_sq >= 77) {
if (MVDiff < 7) {
if (mu_mv_mag_sq < 105) {
return 8;
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 2587) {
return 8;
}
else if (Rs >= 2587) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (Rs < 2194) {
return 8;
}
else if (Rs >= 2194) {
return 4;
}
}
}
}
else if (Cs >= 2736) {
return 4;
}
}
else if (negBalance >= 56) {
if (posBalance < 62) {
if (refDCval < 81) {
return 4;
}
else if (refDCval >= 81) {
if (Rs < 1586) {
if (Rs < 1537) {
return 8;
}
else if (Rs >= 1537) {
return 4;
}
}
else if (Rs >= 1586) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (Rs < 1847) {
return 4;
}
else if (Rs >= 1847) {
return 8;
}
}
}
}
}
else if (ssDCval >= 90) {
if (RsCsDiff < 99) {
if (posBalance < 64) {
if (Cs < 1346) {
if (Rs < 1700) {
return 2;
}
else if (Rs >= 1700) {
return 8;
}
}
else if (Cs >= 1346) {
return 8;
}
}
else if (posBalance >= 64) {
return 2;
}
}
else if (RsCsDiff >= 99) {
if (Rs < 2081) {
if (Rs < 1861) {
return 8;
}
else if (Rs >= 1861) {
return 2;
}
}
else if (Rs >= 2081) {
return 8;
}
}
}
}
else if (posBalance >= 70) {
return 4;
}
}
else if (mcTcor >= 93) {
if (Cs < 3415) {
if (SC < 4855) {
if (refDCval < 89) {
if (tcor < 88) {
if (TSC < 9) {
if (posBalance < 53) {
if (Rs < 2364) {
return 8;
}
else if (Rs >= 2364) {
return 2;
}
}
else if (posBalance >= 53) {
if (ssDCval < 84) {
if (Rs < 1585) {
return 2;
}
else if (Rs >= 1585) {
if (Cs < 1745) {
return 4;
}
else if (Cs >= 1745) {
return 1;
}
}
}
else if (ssDCval >= 84) {
if (Cs < 1381) {
return 8;
}
else if (Cs >= 1381) {
return 2;
}
}
}
}
else if (TSC >= 9) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
}
else if (tcor >= 88) {
if (Cs < 1488) {
if (Rs < 2709) {
if (MVDiff < 5) {
if (refDCval < 82) {
if (refDCval < 75) {
if (Rs < 2072) {
return 8;
}
else if (Rs >= 2072) {
return 2;
}
}
else if (refDCval >= 75) {
if (diffRsCsdiff < -9) {
if (Rs < 1984) {
if (Rs < 1889) {
return 4;
}
else if (Rs >= 1889) {
return 8;
}
}
else if (Rs >= 1984) {
if (Cs < 1467) {
return 1;
}
else if (Cs >= 1467) {
return 8;
}
}
}
else if (diffRsCsdiff >= -9) {
if (Rs < 1895) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Cs < 1402) {
return 1;
}
else if (Cs >= 1402) {
return 8;
}
}
}
else if (Rs >= 1895) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (MVDiff < 3) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 82) {
if (tcor < 95) {
if (mu_mv_mag_sq < 93) {
if (negBalance < 46) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (negBalance >= 46) {
return 1;
}
}
else if (mu_mv_mag_sq >= 93) {
return 8;
}
}
else if (tcor >= 95) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (refDCval < 83) {
return 8;
}
else if (refDCval >= 83) {
if (mu_mv_mag_sq < 53) {
return 8;
}
else if (mu_mv_mag_sq >= 53) {
if (TSC < 9) {
if (Cs < 1411) {
return 1;
}
else if (Cs >= 1411) {
return 2;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
else if (Rs >= 2709) {
return 2;
}
}
else if (Cs >= 1488) {
if (refDCval < 80) {
if (TSC < 7) {
if (tcor < 93) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (mu_mv_mag_sq < 86) {
return 8;
}
else if (mu_mv_mag_sq >= 86) {
return 4;
}
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (TSC >= 7) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
if (Cs < 1932) {
return 8;
}
else if (Cs >= 1932) {
return 4;
}
}
}
}
}
else if (refDCval >= 80) {
if (Cs < 2278) {
if (Rs < 2924) {
if (MVDiff < 5) {
if (TSC < 7) {
if (SC < 3820) {
if (Rs < 1886) {
return 8;
}
else if (Rs >= 1886) {
if (Rs < 2189) {
return 2;
}
else if (Rs >= 2189) {
return 1;
}
}
}
else if (SC >= 3820) {
return 8;
}
}
else if (TSC >= 7) {
if (negBalance < 47) {
if (Cs < 1747) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (Cs >= 1747) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 8;
}
}
}
else if (negBalance >= 47) {
return 8;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (Rs >= 2924) {
return 4;
}
}
else if (Cs >= 2278) {
if (tcor < 93) {
if (Rs < 2262) {
return 1;
}
else if (Rs >= 2262) {
return 8;
}
}
else if (tcor >= 93) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 89) {
if (Rs < 2176) {
if (posBalance < 58) {
if (negBalance < 46) {
if (posBalance < 49) {
if (AFD < 11) {
if (Cs < 1873) {
return 8;
}
else if (Cs >= 1873) {
return 1;
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (posBalance >= 49) {
if (negBalance < 39) {
if (Rs < 2068) {
return 2;
}
else if (Rs >= 2068) {
return 8;
}
}
else if (negBalance >= 39) {
return 8;
}
}
}
else if (negBalance >= 46) {
if (TSC < 8) {
if (posBalance < 56) {
if (posBalance < 7) {
if (Rs < 2142) {
return 8;
}
else if (Rs >= 2142) {
return 1;
}
}
else if (posBalance >= 7) {
return 2;
}
}
else if (posBalance >= 56) {
if (Cs < 1416) {
return 1;
}
else if (Cs >= 1416) {
return 8;
}
}
}
else if (TSC >= 8) {
if (refDCval < 101) {
return 8;
}
else if (refDCval >= 101) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (refDCval < 112) {
return 2;
}
else if (refDCval >= 112) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 58) {
if (diffRsCsdiff < -4) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (diffRsCsdiff >= -4) {
if (MVDiff < 7) {
if (Rs < 1920) {
return 2;
}
else if (Rs >= 1920) {
return 8;
}
}
else if (MVDiff >= 7) {
if (Rs < 1657) {
return 8;
}
else if (Rs >= 1657) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2176) {
if (Cs < 1909) {
if (ssDCval < 102) {
if (Cs < 1833) {
return 8;
}
else if (Cs >= 1833) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
return 4;
}
}
}
else if (ssDCval >= 102) {
if (mu_mv_mag_sq < 58) {
if (Rs < 3032) {
if (Rs < 2433) {
return 8;
}
else if (Rs >= 2433) {
return 4;
}
}
else if (Rs >= 3032) {
return 8;
}
}
else if (mu_mv_mag_sq >= 58) {
return 8;
}
}
}
else if (Cs >= 1909) {
if (Cs < 2057) {
if (mcTcor < 95) {
if (Rs < 2226) {
return 8;
}
else if (Rs >= 2226) {
return 1;
}
}
else if (mcTcor >= 95) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (Cs >= 2057) {
return 8;
}
}
}
}
}
else if (SC >= 4855) {
if (Cs < 1626) {
if (ssDCval < 96) {
if (RsCsDiff < 77) {
if (posBalance < 47) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 1;
}
}
else if (posBalance >= 47) {
if (Rs < 3732) {
return 2;
}
else if (Rs >= 3732) {
return 1;
}
}
}
else if (RsCsDiff >= 77) {
return 2;
}
}
else if (ssDCval >= 96) {
return 8;
}
}
else if (Cs >= 1626) {
if (Rs < 4593) {
if (mu_mv_mag_sq < 54) {
return 8;
}
else if (mu_mv_mag_sq >= 54) {
if (CsDiff < 4) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (Cs < 2927) {
return 4;
}
else if (Cs >= 2927) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (Rs < 2492) {
return 4;
}
else if (Rs >= 2492) {
if (mu_mv_mag_sq < 84) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (diffRsCsdiff < -17) {
if (Rs < 3226) {
return 4;
}
else if (Rs >= 3226) {
return 1;
}
}
else if (diffRsCsdiff >= -17) {
if (AFD < 10) {
if (Cs < 1934) {
return 4;
}
else if (Cs >= 1934) {
if (Rs < 2650) {
return 4;
}
else if (Rs >= 2650) {
return 8;
}
}
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (negBalance < 48) {
if (Cs < 2051) {
if (Rs < 3524) {
return 8;
}
else if (Rs >= 3524) {
return 2;
}
}
else if (Cs >= 2051) {
if (ssDCval < 88) {
if (ssDCval < 73) {
return 4;
}
else if (ssDCval >= 73) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Cs < 2757) {
return 8;
}
else if (Cs >= 2757) {
return 4;
}
}
}
}
else if (ssDCval >= 88) {
return 4;
}
}
}
else if (negBalance >= 48) {
if (Cs < 1974) {
return 2;
}
else if (Cs >= 1974) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 4593) {
if (Cs < 1740) {
return 1;
}
else if (Cs >= 1740) {
return 4;
}
}
}
}
}
else if (Cs >= 3415) {
return 8;
}
}
}
else if (CsDiff >= 7) {
if (Cs < 2908) {
if (Rs < 3950) {
if (MVDiff < 43) {
if (RsDiff < 6) {
if (Cs < 2295) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
else if (Cs >= 2295) {
if (mcTcor < 94) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 1690) {
return 2;
}
else if (Rs >= 1690) {
return 1;
}
}
}
}
else if (RsDiff >= 6) {
if (ssDCval < 52) {
if (negBalance < 53) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
}
}
else if (negBalance >= 53) {
return 4;
}
}
else if (ssDCval >= 52) {
if (mcTcor < 94) {
if (AFD < 16) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (diffRsCsdiff < 31) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Rs < 2465) {
return 4;
}
else if (Rs >= 2465) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 31) {
return 4;
}
}
}
else if (AFD >= 16) {
if (Cs < 1832) {
return 8;
}
else if (Cs >= 1832) {
if (Rs < 2150) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Rs >= 2150) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (diffMVdiffVal < -2) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
if (AFD < 14) {
if (Rs < 2593) {
return 8;
}
else if (Rs >= 2593) {
if (Cs < 2666) {
return 4;
}
else if (Cs >= 2666) {
return 8;
}
}
}
else if (AFD >= 14) {
if (Cs < 1904) {
return 4;
}
else if (Cs >= 1904) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 2467) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (negBalance < 54) {
if (Rs < 2009) {
return 1;
}
else if (Rs >= 2009) {
return 4;
}
}
else if (negBalance >= 54) {
return 8;
}
}
}
else if (Cs >= 2467) {
if (Rs < 2538) {
return 1;
}
else if (Rs >= 2538) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 43) {
if (AFD < 16) {
if (Rs < 2030) {
return 2;
}
else if (Rs >= 2030) {
return 8;
}
}
else if (AFD >= 16) {
return 1;
}
}
}
else if (Rs >= 3950) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
return 1;
}
}
}
else if (Cs >= 2908) {
if (Rs < 3007) {
return 4;
}
else if (Rs >= 3007) {
if (negBalance < 47) {
if (ssDCval < 86) {
return 1;
}
else if (ssDCval >= 86) {
if (Rs < 4589) {
return 4;
}
else if (Rs >= 4589) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (Rs < 3180) {
if (Cs < 2973) {
return 8;
}
else if (Cs >= 2973) {
return 1;
}
}
else if (Rs >= 3180) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 105) {
if (refDCval < 163) {
if (Rs < 4696) {
if (mu_mv_mag_sq < 38) {
if (Rs < 2017) {
if (ssDCval < 128) {
if (mcTcor < 97) {
if (AFD < 21) {
if (RsCsDiff < 67) {
if (AFD < 13) {
if (tcor < 82) {
if (Rs < 1205) {
return 4;
}
else if (Rs >= 1205) {
if (Rs < 1457) {
return 1;
}
else if (Rs >= 1457) {
return 8;
}
}
}
else if (tcor >= 82) {
if (ssDCval < 125) {
return 8;
}
else if (ssDCval >= 125) {
return 1;
}
}
}
else if (AFD >= 13) {
if (gchDC < 4) {
if (Cs < 1810) {
if (ssDCval < 124) {
return 4;
}
else if (ssDCval >= 124) {
return 8;
}
}
else if (Cs >= 1810) {
return 8;
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (RsCsDiff >= 67) {
if (negBalance < 45) {
if (diffRsCsdiff < -25) {
if (Rs < 1538) {
return 1;
}
else if (Rs >= 1538) {
return 4;
}
}
else if (diffRsCsdiff >= -25) {
return 8;
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (AFD >= 21) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
if (Rs < 663) {
return 2;
}
else if (Rs >= 663) {
return 8;
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 113) {
if (RsCsDiff < 47) {
return 4;
}
else if (RsCsDiff >= 47) {
return 2;
}
}
else if (refDCval >= 113) {
if (Rs < 1394) {
return 4;
}
else if (Rs >= 1394) {
return 8;
}
}
}
}
else if (ssDCval >= 128) {
if (ssDCval < 151) {
if (TSC < 12) {
if (posBalance < 51) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
return 8;
}
}
else if (posBalance >= 51) {
if (mu_mv_mag_sq < 30) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 131) {
if (Rs < 1335) {
return 8;
}
else if (Rs >= 1335) {
return 1;
}
}
else if (ssDCval >= 131) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (mu_mv_mag_sq < 24) {
return 1;
}
else if (mu_mv_mag_sq >= 24) {
if (Rs < 1407) {
return 1;
}
else if (Rs >= 1407) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 30) {
return 8;
}
}
}
else if (TSC >= 12) {
return 4;
}
}
else if (ssDCval >= 151) {
if (RsCsDiff < 99) {
if (Rs < 1826) {
return 2;
}
else if (Rs >= 1826) {
return 8;
}
}
else if (RsCsDiff >= 99) {
return 8;
}
}
}
}
else if (Rs >= 2017) {
if (ssDCval < 110) {
if (ssDCval < 107) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (ssDCval >= 107) {
if (AFD < 10) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
else if (AFD >= 10) {
if (Cs < 1645) {
return 1;
}
else if (Cs >= 1645) {
return 8;
}
}
}
}
else if (ssDCval >= 110) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 1488) {
if (Cs < 1509) {
if (refDCval < 111) {
if (SC < 2597) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
else if (SC >= 2597) {
return 8;
}
}
else if (refDCval >= 111) {
if (Cs < 1347) {
if (Rs < 1080) {
if (Rs < 850) {
return 8;
}
else if (Rs >= 850) {
return 4;
}
}
else if (Rs >= 1080) {
return 8;
}
}
else if (Cs >= 1347) {
return 8;
}
}
}
else if (Cs >= 1509) {
if (refDCval < 128) {
if (negBalance < 38) {
if (posBalance < 61) {
if (Rs < 1383) {
return 4;
}
else if (Rs >= 1383) {
return 8;
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (refDCval >= 128) {
if (Rs < 989) {
return 8;
}
else if (Rs >= 989) {
if (ssDCval < 134) {
return 4;
}
else if (ssDCval >= 134) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1488) {
if (refDCval < 109) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
if (tcor < 89) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
else if (tcor >= 89) {
if (mu_mv_mag_sq < 60) {
if (Rs < 3233) {
if (SC < 4383) {
if (ssDCval < 107) {
return 8;
}
else if (ssDCval >= 107) {
return 4;
}
}
else if (SC >= 4383) {
return 4;
}
}
else if (Rs >= 3233) {
return 8;
}
}
else if (mu_mv_mag_sq >= 60) {
if (negBalance < 47) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (diffRsCsdiff < 23) {
return 8;
}
else if (diffRsCsdiff >= 23) {
if (Rs < 3250) {
return 4;
}
else if (Rs >= 3250) {
return 8;
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 3083) {
return 8;
}
else if (Rs >= 3083) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 109) {
if (posBalance < 56) {
if (refDCval < 139) {
if (ssDCval < 135) {
if (tcor < 106) {
if (Cs < 1835) {
if (tcor < 93) {
if (negBalance < 32) {
if (SC < 3681) {
return 8;
}
else if (SC >= 3681) {
return 1;
}
}
else if (negBalance >= 32) {
if (TSC < 9) {
if (Rs < 1749) {
if (Rs < 1739) {
return 8;
}
else if (Rs >= 1739) {
return 1;
}
}
else if (Rs >= 1749) {
return 8;
}
}
else if (TSC >= 9) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (tcor >= 93) {
if (CsDiff < 2) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 2;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (Cs >= 1835) {
return 8;
}
}
else if (tcor >= 106) {
if (Rs < 1813) {
return 8;
}
else if (Rs >= 1813) {
return 2;
}
}
}
else if (ssDCval >= 135) {
if (Rs < 2614) {
if (SC < 3597) {
return 8;
}
else if (SC >= 3597) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
return 8;
}
}
}
}
else if (Rs >= 2614) {
return 8;
}
}
}
else if (refDCval >= 139) {
if (posBalance < 37) {
if (Rs < 1720) {
return 2;
}
else if (Rs >= 1720) {
if (Rs < 2540) {
return 8;
}
else if (Rs >= 2540) {
return 4;
}
}
}
else if (posBalance >= 37) {
return 8;
}
}
}
else if (posBalance >= 56) {
if (SC < 5438) {
if (MVDiff < 7) {
if (TSC < 12) {
if (Rs < 1876) {
if (Cs < 1759) {
return 8;
}
else if (Cs >= 1759) {
return 1;
}
}
else if (Rs >= 1876) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
if (negBalance < 51) {
if (Rs < 2242) {
return 8;
}
else if (Rs >= 2242) {
return 1;
}
}
else if (negBalance >= 51) {
return 8;
}
}
}
}
else if (TSC >= 12) {
return 1;
}
}
else if (MVDiff >= 7) {
if (diffRsCsdiff < -40) {
if (ssDCval < 147) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (ssDCval >= 147) {
return 4;
}
}
else if (diffRsCsdiff >= -40) {
if (diffMVdiffVal < -13) {
if (diffMVdiffVal < -15) {
return 8;
}
else if (diffMVdiffVal >= -15) {
return 4;
}
}
else if (diffMVdiffVal >= -13) {
return 8;
}
}
}
}
else if (SC >= 5438) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
if (mu_mv_mag_sq < 78) {
return 8;
}
else if (mu_mv_mag_sq >= 78) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 4696) {
if (Cs < 1794) {
return 2;
}
else if (Cs >= 1794) {
if (Cs < 3675) {
if (Rs < 5979) {
return 1;
}
else if (Rs >= 5979) {
return 4;
}
}
else if (Cs >= 3675) {
return 8;
}
}
}
}
else if (refDCval >= 163) {
if (Rs < 2266) {
if (Rs < 1053) {
return 1;
}
else if (Rs >= 1053) {
return 2;
}
}
else if (Rs >= 2266) {
return 8;
}
}
}
}
else if (RsCsDiff >= 101) {
if (tcor < 88) {
if (Cs < 1755) {
if (RsDiff < 15) {
if (negBalance < 46) {
if (mcTcor < 89) {
if (refDCval < 60) {
if (refDCval < 46) {
if (Rs < 2002) {
if (Rs < 1018) {
return 1;
}
else if (Rs >= 1018) {
return 8;
}
}
else if (Rs >= 2002) {
return 1;
}
}
else if (refDCval >= 46) {
if (CsDiff < 11) {
if (ssDCval < 59) {
if (Rs < 3146) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 1;
}
}
else if (Rs >= 3146) {
return 2;
}
}
else if (ssDCval >= 59) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (Rs < 2505) {
return 1;
}
else if (Rs >= 2505) {
return 4;
}
}
}
}
else if (CsDiff >= 11) {
return 4;
}
}
}
else if (refDCval >= 60) {
if (posBalance < 48) {
if (Rs < 2260) {
if (Rs < 1991) {
return 8;
}
else if (Rs >= 1991) {
return 1;
}
}
else if (Rs >= 2260) {
return 2;
}
}
else if (posBalance >= 48) {
if (mu_mv_mag_sq < 19) {
if (RsCsDiff < 142) {
return 4;
}
else if (RsCsDiff >= 142) {
if (tcor < 59) {
return 8;
}
else if (tcor >= 59) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 19) {
return 8;
}
}
}
}
else if (mcTcor >= 89) {
if (ssDCval < 66) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
return 1;
}
}
}
else if (ssDCval >= 66) {
if (RsCsDiff < 121) {
if (RsDiff < 8) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (RsDiff >= 8) {
if (diffRsCsdiff < -31) {
if (Rs < 2411) {
return 2;
}
else if (Rs >= 2411) {
if (Rs < 2980) {
return 1;
}
else if (Rs >= 2980) {
return 4;
}
}
}
else if (diffRsCsdiff >= -31) {
if (SC < 4798) {
return 4;
}
else if (SC >= 4798) {
return 8;
}
}
}
}
else if (RsCsDiff >= 121) {
if (refDCval < 98) {
if (Rs < 2306) {
if (tcor < 82) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 2;
}
}
else if (tcor >= 82) {
if (posBalance < 71) {
return 8;
}
else if (posBalance >= 71) {
return 2;
}
}
}
else if (Rs >= 2306) {
if (posBalance < 42) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 8;
}
}
else if (posBalance >= 42) {
if (ssDCval < 93) {
if (TSC < 11) {
if (RsCsDiff < 141) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (RsCsDiff >= 141) {
return 2;
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (ssDCval >= 93) {
if (negBalance < 45) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (negBalance >= 45) {
return 4;
}
}
}
}
}
else if (refDCval >= 98) {
if (diffAFD < 1) {
if (gchDC < 4) {
if (TSC < 7) {
if (Rs < 2130) {
return 8;
}
else if (Rs >= 2130) {
return 2;
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (gchDC >= 4) {
if (Rs < 3223) {
return 2;
}
else if (Rs >= 3223) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 46) {
if (Rs < 1781) {
if (refDCval < 70) {
if (mu_mv_mag_sq < 7) {
if (Cs < 1748) {
if (mu_mv_mag_sq < 5) {
if (mcTcor < 76) {
return 4;
}
else if (mcTcor >= 76) {
if (Rs < 1123) {
return 4;
}
else if (Rs >= 1123) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 1456) {
return 2;
}
else if (Rs >= 1456) {
return 4;
}
}
}
else if (Cs >= 1748) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffAFD < -1) {
if (Rs < 1420) {
if (Rs < 1351) {
return 4;
}
else if (Rs >= 1351) {
return 1;
}
}
else if (Rs >= 1420) {
return 4;
}
}
else if (diffAFD >= -1) {
if (Cs < 1685) {
if (RsCsDiff < 193) {
if (mcTcor < 57) {
return 4;
}
else if (mcTcor >= 57) {
return 8;
}
}
else if (RsCsDiff >= 193) {
if (Rs < 1063) {
return 8;
}
else if (Rs >= 1063) {
return 4;
}
}
}
else if (Cs >= 1685) {
return 4;
}
}
}
}
else if (refDCval >= 70) {
if (AFD < 19) {
if (refDCval < 83) {
if (AFD < 15) {
if (Rs < 1604) {
return 8;
}
else if (Rs >= 1604) {
return 1;
}
}
else if (AFD >= 15) {
if (negBalance < 58) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (negBalance >= 58) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
}
else if (refDCval >= 83) {
if (MVDiff < 41) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Cs < 1413) {
return 8;
}
else if (Cs >= 1413) {
return 2;
}
}
}
else if (MVDiff >= 41) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 8;
}
}
}
}
else if (AFD >= 19) {
if (negBalance < 54) {
if (diffRsCsdiff < 31) {
if (ssDCval < 72) {
if (Rs < 1644) {
return 1;
}
else if (Rs >= 1644) {
return 2;
}
}
else if (ssDCval >= 72) {
if (diffRsCsdiff < -57) {
if (Rs < 1437) {
return 1;
}
else if (Rs >= 1437) {
return 4;
}
}
else if (diffRsCsdiff >= -57) {
return 4;
}
}
}
else if (diffRsCsdiff >= 31) {
if (tcor < 75) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
return 8;
}
}
else if (tcor >= 75) {
if (Rs < 1635) {
return 4;
}
else if (Rs >= 1635) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (SC < 3077) {
if (diffAFD < -2) {
if (diffRsCsdiff < -54) {
return 4;
}
else if (diffRsCsdiff >= -54) {
if (CsDiff < 11) {
return 2;
}
else if (CsDiff >= 11) {
return 1;
}
}
}
else if (diffAFD >= -2) {
if (refDCval < 73) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
if (Rs < 1540) {
return 8;
}
else if (Rs >= 1540) {
return 2;
}
}
}
else if (refDCval >= 73) {
if (tcor < 58) {
if (RsCsDiff < 163) {
return 2;
}
else if (RsCsDiff >= 163) {
return 4;
}
}
else if (tcor >= 58) {
if (Cs < 1445) {
return 2;
}
else if (Cs >= 1445) {
if (AFD < 25) {
if (Rs < 1046) {
return 8;
}
else if (Rs >= 1046) {
return 2;
}
}
else if (AFD >= 25) {
return 8;
}
}
}
}
}
}
else if (SC >= 3077) {
if (RsCsDiff < 182) {
if (Rs < 1550) {
return 8;
}
else if (Rs >= 1550) {
if (Rs < 1678) {
return 1;
}
else if (Rs >= 1678) {
return 4;
}
}
}
else if (RsCsDiff >= 182) {
if (SC < 3172) {
return 8;
}
else if (SC >= 3172) {
if (ssDCval < 106) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (ssDCval >= 106) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1781) {
if (MVDiff < 19) {
if (refDCval < 107) {
if (TSC < 10) {
if (refDCval < 75) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (refDCval >= 75) {
if (refDCval < 86) {
if (Cs < 1594) {
if (mu_mv_mag_sq < 6) {
if (Cs < 1533) {
if (Rs < 1992) {
return 2;
}
else if (Rs >= 1992) {
return 4;
}
}
else if (Cs >= 1533) {
return 2;
}
}
else if (mu_mv_mag_sq >= 6) {
if (Cs < 1355) {
return 8;
}
else if (Cs >= 1355) {
return 2;
}
}
}
else if (Cs >= 1594) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 4;
}
}
}
else if (refDCval >= 86) {
if (posBalance < 57) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
else if (posBalance >= 57) {
if (ssDCval < 98) {
if (AFD < 14) {
if (Rs < 2279) {
return 8;
}
else if (Rs >= 2279) {
return 2;
}
}
else if (AFD >= 14) {
if (Rs < 2668) {
return 2;
}
else if (Rs >= 2668) {
return 4;
}
}
}
else if (ssDCval >= 98) {
if (Cs < 1393) {
return 4;
}
else if (Cs >= 1393) {
return 8;
}
}
}
}
}
}
else if (TSC >= 10) {
if (RsCsDiff < 141) {
return 8;
}
else if (RsCsDiff >= 141) {
if (refDCval < 85) {
if (mcTcor < 89) {
if (negBalance < 50) {
if (Rs < 2933) {
return 4;
}
else if (Rs >= 2933) {
return 8;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (mcTcor >= 89) {
if (RsCsDiff < 198) {
if (Cs < 1713) {
return 4;
}
else if (Cs >= 1713) {
return 8;
}
}
else if (RsCsDiff >= 198) {
if (Rs < 2233) {
return 4;
}
else if (Rs >= 2233) {
return 2;
}
}
}
}
else if (refDCval >= 85) {
if (refDCval < 104) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 4;
}
}
}
else if (refDCval >= 104) {
if (Rs < 2561) {
return 8;
}
else if (Rs >= 2561) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 107) {
if (RsDiff < 14) {
if (tcor < 68) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (Rs < 2718) {
return 4;
}
else if (Rs >= 2718) {
return 8;
}
}
}
else if (tcor >= 68) {
if (AFD < 19) {
if (tcor < 73) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
else if (tcor >= 73) {
return 8;
}
}
else if (AFD >= 19) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
else if (RsDiff >= 14) {
if (Rs < 3329) {
return 2;
}
else if (Rs >= 3329) {
return 4;
}
}
}
}
else if (MVDiff >= 19) {
if (TSC < 13) {
if (RsCsDiff < 197) {
if (diffRsCsdiff < 1) {
if (ssDCval < 162) {
if (AFD < 16) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (AFD >= 16) {
if (diffMVdiffVal < 15) {
if (diffRsCsdiff < -126) {
if (Cs < 1458) {
return 8;
}
else if (Cs >= 1458) {
return 1;
}
}
else if (diffRsCsdiff >= -126) {
if (MVDiff < 37) {
if (diffRsCsdiff < -33) {
return 1;
}
else if (diffRsCsdiff >= -33) {
if (ssDCval < 81) {
return 1;
}
else if (ssDCval >= 81) {
return 8;
}
}
}
else if (MVDiff >= 37) {
return 8;
}
}
}
else if (diffMVdiffVal >= 15) {
return 4;
}
}
}
else if (ssDCval >= 162) {
return 2;
}
}
else if (diffRsCsdiff >= 1) {
if (diffMVdiffVal < -16) {
return 4;
}
else if (diffMVdiffVal >= -16) {
if (SC < 3456) {
if (ssDCval < 88) {
if (Rs < 1936) {
return 4;
}
else if (Rs >= 1936) {
return 2;
}
}
else if (ssDCval >= 88) {
return 8;
}
}
else if (SC >= 3456) {
return 8;
}
}
}
}
else if (RsCsDiff >= 197) {
if (AFD < 15) {
if (tcor < 77) {
return 4;
}
else if (tcor >= 77) {
if (Rs < 2375) {
return 1;
}
else if (Rs >= 2375) {
if (Rs < 3342) {
return 8;
}
else if (Rs >= 3342) {
return 2;
}
}
}
}
else if (AFD >= 15) {
if (tcor < 62) {
if (MVDiff < 38) {
return 4;
}
else if (MVDiff >= 38) {
if (Rs < 1904) {
return 8;
}
else if (Rs >= 1904) {
return 1;
}
}
}
else if (tcor >= 62) {
if (posBalance < 47) {
if (Rs < 2749) {
return 2;
}
else if (Rs >= 2749) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 4;
}
}
}
else if (posBalance >= 47) {
if (TSC < 10) {
if (Rs < 2596) {
return 1;
}
else if (Rs >= 2596) {
if (Rs < 4422) {
return 8;
}
else if (Rs >= 4422) {
return 1;
}
}
}
else if (TSC >= 10) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (diffTSC < 1) {
if (MVDiff < 44) {
return 8;
}
else if (MVDiff >= 44) {
return 2;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
}
}
}
}
}
}
else if (TSC >= 13) {
if (ssDCval < 124) {
if (Rs < 2524) {
if (Cs < 1742) {
if (AFD < 35) {
if (MVDiff < 43) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
if (Cs < 1532) {
return 4;
}
else if (Cs >= 1532) {
return 8;
}
}
}
else if (MVDiff >= 43) {
if (CsDiff < 9) {
if (Cs < 1739) {
return 1;
}
else if (Cs >= 1739) {
return 8;
}
}
else if (CsDiff >= 9) {
return 8;
}
}
}
else if (AFD >= 35) {
if (Rs < 1866) {
return 2;
}
else if (Rs >= 1866) {
return 4;
}
}
}
else if (Cs >= 1742) {
return 1;
}
}
else if (Rs >= 2524) {
if (Rs < 2813) {
if (refDCval < 88) {
return 4;
}
else if (refDCval >= 88) {
return 8;
}
}
else if (Rs >= 2813) {
return 4;
}
}
}
else if (ssDCval >= 124) {
if (MVDiff < 50) {
return 1;
}
else if (MVDiff >= 50) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 15) {
if (MVDiff < 19) {
if (tcor < 82) {
if (RsDiff < 16) {
if (SC < 4724) {
if (Cs < 1379) {
return 2;
}
else if (Cs >= 1379) {
return 4;
}
}
else if (SC >= 4724) {
return 8;
}
}
else if (RsDiff >= 16) {
if (CsDiff < 12) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (diffTSC < 1) {
if (Rs < 4794) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (TSC < 12) {
if (AFD < 19) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
else if (AFD >= 19) {
return 2;
}
}
else if (TSC >= 12) {
if (Rs < 3275) {
return 8;
}
else if (Rs >= 3275) {
return 2;
}
}
}
}
else if (Rs >= 4794) {
if (Rs < 6069) {
return 4;
}
else if (Rs >= 6069) {
return 8;
}
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (CsDiff >= 12) {
if (TSC < 9) {
if (Rs < 2731) {
return 4;
}
else if (Rs >= 2731) {
return 2;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
else if (tcor >= 82) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
if (RsDiff < 17) {
return 4;
}
else if (RsDiff >= 17) {
return 2;
}
}
}
}
else if (MVDiff >= 19) {
if (ssDCval < 113) {
if (CsDiff < 7) {
if (SC < 5068) {
if (ssDCval < 97) {
if (Rs < 3015) {
return 8;
}
else if (Rs >= 3015) {
return 4;
}
}
else if (ssDCval >= 97) {
return 2;
}
}
else if (SC >= 5068) {
return 4;
}
}
else if (CsDiff >= 7) {
if (mcTcor < 75) {
if (Cs < 1455) {
return 1;
}
else if (Cs >= 1455) {
return 8;
}
}
else if (mcTcor >= 75) {
if (mcTcor < 92) {
if (tcor < 60) {
if (MVDiff < 67) {
if (Rs < 4080) {
if (mcTcor < 77) {
if (Cs < 1682) {
return 2;
}
else if (Cs >= 1682) {
return 8;
}
}
else if (mcTcor >= 77) {
return 2;
}
}
else if (Rs >= 4080) {
if (Rs < 4191) {
return 4;
}
else if (Rs >= 4191) {
if (Cs < 1547) {
return 2;
}
else if (Cs >= 1547) {
return 8;
}
}
}
}
else if (MVDiff >= 67) {
if (diffMVdiffVal < 17) {
return 8;
}
else if (diffMVdiffVal >= 17) {
if (Rs < 2756) {
return 4;
}
else if (Rs >= 2756) {
return 2;
}
}
}
}
else if (tcor >= 60) {
if (ssDCval < 85) {
if (posBalance < 40) {
if (Rs < 2963) {
return 8;
}
else if (Rs >= 2963) {
return 4;
}
}
else if (posBalance >= 40) {
return 2;
}
}
else if (ssDCval >= 85) {
if (Rs < 4263) {
return 8;
}
else if (Rs >= 4263) {
return 2;
}
}
}
}
else if (mcTcor >= 92) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
}
}
}
else if (ssDCval >= 113) {
if (CsDiff < 9) {
if (mu_mv_mag_sq < 12) {
if (negBalance < 62) {
if (ssDCval < 134) {
if (Cs < 1357) {
return 2;
}
else if (Cs >= 1357) {
if (posBalance < 52) {
if (Rs < 6608) {
return 8;
}
else if (Rs >= 6608) {
return 2;
}
}
else if (posBalance >= 52) {
return 8;
}
}
}
else if (ssDCval >= 134) {
if (mcTcor < 90) {
if (Cs < 1459) {
return 8;
}
else if (Cs >= 1459) {
return 4;
}
}
else if (mcTcor >= 90) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Rs < 6428) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
return 2;
}
}
else if (Rs >= 6428) {
return 8;
}
}
}
}
}
else if (negBalance >= 62) {
return 1;
}
}
else if (mu_mv_mag_sq >= 12) {
if (AFD < 27) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
if (Rs < 3194) {
return 8;
}
else if (Rs >= 3194) {
return 2;
}
}
}
else if (AFD >= 27) {
if (Cs < 1513) {
return 2;
}
else if (Cs >= 1513) {
if (Rs < 3317) {
return 2;
}
else if (Rs >= 3317) {
return 4;
}
}
}
}
}
else if (CsDiff >= 9) {
if (AFD < 29) {
if (CsDiff < 11) {
if (mu_mv_mag_sq < 8) {
if (Rs < 7760) {
return 8;
}
else if (Rs >= 7760) {
return 2;
}
}
else if (mu_mv_mag_sq >= 8) {
if (mu_mv_mag_sq < 24) {
return 1;
}
else if (mu_mv_mag_sq >= 24) {
return 2;
}
}
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (AFD >= 29) {
if (tcor < 64) {
if (diffAFD < 8) {
if (Rs < 5492) {
return 2;
}
else if (Rs >= 5492) {
return 4;
}
}
else if (diffAFD >= 8) {
return 1;
}
}
else if (tcor >= 64) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1755) {
if (AFD < 18) {
if (refDCval < 101) {
if (posBalance < 46) {
if (refDCval < 74) {
if (Rs < 3305) {
if (mcTcor < 88) {
if (CsDiff < 11) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 4;
}
}
else if (CsDiff >= 11) {
if (Rs < 1423) {
return 4;
}
else if (Rs >= 1423) {
return 1;
}
}
}
else if (mcTcor >= 88) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
return 2;
}
}
}
else if (Rs >= 3305) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
if (negBalance < 47) {
if (CsDiff < 17) {
if (Cs < 1823) {
if (Rs < 3913) {
return 4;
}
else if (Rs >= 3913) {
return 2;
}
}
else if (Cs >= 1823) {
return 4;
}
}
else if (CsDiff >= 17) {
if (Rs < 5147) {
return 8;
}
else if (Rs >= 5147) {
return 4;
}
}
}
else if (negBalance >= 47) {
if (Rs < 3999) {
return 8;
}
else if (Rs >= 3999) {
return 4;
}
}
}
}
}
else if (refDCval >= 74) {
if (TSC < 11) {
if (MVDiff < 38) {
return 8;
}
else if (MVDiff >= 38) {
return 4;
}
}
else if (TSC >= 11) {
if (Rs < 5565) {
return 2;
}
else if (Rs >= 5565) {
return 1;
}
}
}
}
else if (posBalance >= 46) {
if (mcTcor < 72) {
if (negBalance < 57) {
if (ssDCval < 58) {
if (Cs < 1802) {
if (Rs < 1892) {
return 4;
}
else if (Rs >= 1892) {
return 8;
}
}
else if (Cs >= 1802) {
if (negBalance < 56) {
return 4;
}
else if (negBalance >= 56) {
if (Rs < 1854) {
return 4;
}
else if (Rs >= 1854) {
return 2;
}
}
}
}
else if (ssDCval >= 58) {
if (Cs < 3089) {
return 8;
}
else if (Cs >= 3089) {
if (Cs < 3860) {
return 4;
}
else if (Cs >= 3860) {
return 1;
}
}
}
}
else if (negBalance >= 57) {
if (diffRsCsdiff < 1) {
if (Rs < 2234) {
return 2;
}
else if (Rs >= 2234) {
return 4;
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 1788) {
return 2;
}
else if (Rs >= 1788) {
return 4;
}
}
}
}
else if (mcTcor >= 72) {
if (mcTcor < 90) {
if (negBalance < 48) {
if (Rs < 2659) {
if (mu_mv_mag_sq < 77) {
if (RsCsDiff < 163) {
if (SC < 4299) {
return 4;
}
else if (SC >= 4299) {
if (Cs < 2712) {
return 8;
}
else if (Cs >= 2712) {
return 4;
}
}
}
else if (RsCsDiff >= 163) {
if (gchDC < 1) {
if (Rs < 1920) {
return 8;
}
else if (Rs >= 1920) {
return 4;
}
}
else if (gchDC >= 1) {
if (Rs < 1050) {
return 4;
}
else if (Rs >= 1050) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
return 4;
}
}
else if (Rs >= 2659) {
if (RsCsDiff < 175) {
if (Cs < 1798) {
return 2;
}
else if (Cs >= 1798) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 3059) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
else if (Rs >= 3059) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 175) {
if (Rs < 3423) {
if (ssDCval < 89) {
if (ssDCval < 69) {
if (Rs < 2780) {
return 2;
}
else if (Rs >= 2780) {
return 8;
}
}
else if (ssDCval >= 69) {
if (Rs < 3400) {
if (tcor < 86) {
return 4;
}
else if (tcor >= 86) {
return 8;
}
}
else if (Rs >= 3400) {
return 8;
}
}
}
else if (ssDCval >= 89) {
return 2;
}
}
else if (Rs >= 3423) {
if (MVDiff < 14) {
if (diffAFD < 2) {
if (AFD < 10) {
if (Rs < 5230) {
return 1;
}
else if (Rs >= 5230) {
return 2;
}
}
else if (AFD >= 10) {
if (CsDiff < 15) {
if (Cs < 2368) {
return 8;
}
else if (Cs >= 2368) {
return 4;
}
}
else if (CsDiff >= 15) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
return 8;
}
}
}
else if (MVDiff >= 14) {
if (ssDCval < 89) {
if (negBalance < 41) {
if (diffMVdiffVal < 4) {
if (Rs < 5361) {
return 1;
}
else if (Rs >= 5361) {
return 4;
}
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
else if (negBalance >= 41) {
if (Rs < 3697) {
if (Rs < 3442) {
return 4;
}
else if (Rs >= 3442) {
return 8;
}
}
else if (Rs >= 3697) {
return 2;
}
}
}
else if (ssDCval >= 89) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 48) {
if (Rs < 1774) {
if (AFD < 17) {
if (MVDiff < 10) {
if (Rs < 1712) {
return 8;
}
else if (Rs >= 1712) {
return 4;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (AFD >= 17) {
if (Rs < 1712) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 2;
}
}
else if (Rs >= 1712) {
return 1;
}
}
}
else if (Rs >= 1774) {
if (refDCval < 79) {
if (mcTcor < 87) {
if (SC < 9562) {
if (ssDCval < 51) {
if (Rs < 2932) {
return 4;
}
else if (Rs >= 2932) {
return 8;
}
}
else if (ssDCval >= 51) {
if (RsDiff < 18) {
if (mcTcor < 76) {
if (Rs < 2183) {
return 2;
}
else if (Rs >= 2183) {
return 8;
}
}
else if (mcTcor >= 76) {
if (negBalance < 51) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (Rs < 3067) {
return 8;
}
else if (Rs >= 3067) {
return 4;
}
}
}
else if (negBalance >= 51) {
return 8;
}
}
}
else if (RsDiff >= 18) {
return 4;
}
}
}
else if (SC >= 9562) {
return 4;
}
}
else if (mcTcor >= 87) {
if (RsDiff < 14) {
if (MVDiff < 3) {
if (Rs < 2925) {
return 1;
}
else if (Rs >= 2925) {
return 8;
}
}
else if (MVDiff >= 3) {
if (diffMVdiffVal < 4) {
if (gchDC < 2) {
if (Cs < 1898) {
if (Rs < 2895) {
return 1;
}
else if (Rs >= 2895) {
return 8;
}
}
else if (Cs >= 1898) {
return 8;
}
}
else if (gchDC >= 2) {
if (Rs < 2992) {
return 8;
}
else if (Rs >= 2992) {
return 4;
}
}
}
else if (diffMVdiffVal >= 4) {
if (refDCval < 72) {
return 8;
}
else if (refDCval >= 72) {
if (AFD < 17) {
if (diffRsCsdiff < -12) {
return 1;
}
else if (diffRsCsdiff >= -12) {
return 4;
}
}
else if (AFD >= 17) {
return 8;
}
}
}
}
}
else if (RsDiff >= 14) {
if (ssDCval < 69) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (ssDCval >= 69) {
return 2;
}
}
}
}
else if (refDCval >= 79) {
if (tcor < 85) {
if (TSC < 14) {
if (tcor < 73) {
if (Rs < 3760) {
return 8;
}
else if (Rs >= 3760) {
return 1;
}
}
else if (tcor >= 73) {
if (diffAFD < 1) {
if (refDCval < 96) {
if (Rs < 2904) {
if (RsCsDiff < 147) {
if (negBalance < 57) {
if (Cs < 2948) {
return 8;
}
else if (Cs >= 2948) {
return 4;
}
}
else if (negBalance >= 57) {
return 4;
}
}
else if (RsCsDiff >= 147) {
if (CsDiff < 11) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (CsDiff >= 11) {
return 8;
}
}
}
else if (Rs >= 2904) {
if (diffRsCsdiff < -138) {
if (Rs < 3949) {
return 1;
}
else if (Rs >= 3949) {
return 4;
}
}
else if (diffRsCsdiff >= -138) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
if (Rs < 4142) {
return 8;
}
else if (Rs >= 4142) {
return 4;
}
}
}
}
}
else if (refDCval >= 96) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (diffAFD >= 1) {
if (posBalance < 55) {
if (Rs < 2568) {
return 4;
}
else if (Rs >= 2568) {
return 8;
}
}
else if (posBalance >= 55) {
if (Cs < 1899) {
return 8;
}
else if (Cs >= 1899) {
if (mu_mv_mag_sq < 77) {
if (Rs < 4816) {
if (refDCval < 100) {
if (diffAFD < 4) {
return 4;
}
else if (diffAFD >= 4) {
return 8;
}
}
else if (refDCval >= 100) {
return 2;
}
}
else if (Rs >= 4816) {
if (Rs < 5357) {
return 8;
}
else if (Rs >= 5357) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 77) {
if (mcTcor < 84) {
return 4;
}
else if (mcTcor >= 84) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 14) {
if (SC < 7898) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
if (Rs < 1978) {
return 1;
}
else if (Rs >= 1978) {
return 4;
}
}
}
else if (SC >= 7898) {
if (mcTcor < 81) {
return 4;
}
else if (mcTcor >= 81) {
return 8;
}
}
}
}
else if (tcor >= 85) {
if (posBalance < 57) {
if (Cs < 1826) {
return 4;
}
else if (Cs >= 1826) {
if (Cs < 2897) {
return 8;
}
else if (Cs >= 2897) {
return 4;
}
}
}
else if (posBalance >= 57) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (RsCsDiff < 189) {
if (MVDiff < 4) {
if (ssDCval < 86) {
if (SC < 5774) {
if (ssDCval < 75) {
if (Cs < 2360) {
return 8;
}
else if (Cs >= 2360) {
return 1;
}
}
else if (ssDCval >= 75) {
if (refDCval < 79) {
return 2;
}
else if (refDCval >= 79) {
return 4;
}
}
}
else if (SC >= 5774) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (Rs < 3181) {
return 1;
}
else if (Rs >= 3181) {
return 2;
}
}
}
}
}
else if (ssDCval >= 86) {
if (Cs < 2865) {
return 8;
}
else if (Cs >= 2865) {
if (Rs < 4656) {
return 4;
}
else if (Rs >= 4656) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 1795) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
if (diffMVdiffVal < -2) {
if (RsCsDiff < 112) {
return 2;
}
else if (RsCsDiff >= 112) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
}
}
else if (Rs >= 1795) {
if (mu_mv_mag_sq < 70) {
if (posBalance < 55) {
if (Cs < 1804) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
else if (Cs >= 1804) {
if (diffMVdiffVal < 7) {
return 8;
}
else if (diffMVdiffVal >= 7) {
if (RsCsDiff < 172) {
if (Cs < 2007) {
return 4;
}
else if (Cs >= 2007) {
return 8;
}
}
else if (RsCsDiff >= 172) {
return 2;
}
}
}
}
else if (posBalance >= 55) {
if (SC < 5429) {
if (Rs < 3046) {
if (TSC < 11) {
if (Rs < 2001) {
return 8;
}
else if (Rs >= 2001) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
return 8;
}
}
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (Rs >= 3046) {
return 4;
}
}
else if (SC >= 5429) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (Rs < 3385) {
return 8;
}
else if (Rs >= 3385) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (refDCval < 93) {
if (RsCsDiff < 173) {
if (Cs < 3059) {
return 4;
}
else if (Cs >= 3059) {
return 1;
}
}
else if (RsCsDiff >= 173) {
return 8;
}
}
else if (refDCval >= 93) {
if (RsCsDiff < 166) {
return 8;
}
else if (RsCsDiff >= 166) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 189) {
if (Rs < 3077) {
if (RsCsDiff < 255) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 2693) {
return 8;
}
else if (Rs >= 2693) {
if (Rs < 2927) {
return 2;
}
else if (Rs >= 2927) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 255) {
if (negBalance < 55) {
if (Cs < 2608) {
return 8;
}
else if (Cs >= 2608) {
return 4;
}
}
else if (negBalance >= 55) {
return 4;
}
}
}
else if (Rs >= 3077) {
if (Rs < 4342) {
if (Cs < 3370) {
if (Cs < 2254) {
if (mu_mv_mag_sq < 43) {
if (Rs < 3263) {
return 4;
}
else if (Rs >= 3263) {
return 2;
}
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 3102) {
return 2;
}
else if (Rs >= 3102) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
}
else if (Cs >= 2254) {
if (posBalance < 60) {
if (refDCval < 76) {
return 8;
}
else if (refDCval >= 76) {
return 4;
}
}
else if (posBalance >= 60) {
if (mu_mv_mag_sq < 31) {
if (SC < 6418) {
return 2;
}
else if (SC >= 6418) {
return 8;
}
}
else if (mu_mv_mag_sq >= 31) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
return 4;
}
}
}
}
}
else if (Cs >= 3370) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (Rs >= 4342) {
if (Cs < 2705) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
return 4;
}
}
}
}
else if (Cs >= 2705) {
if (MVDiff < 14) {
if (posBalance < 48) {
if (CsDiff < 14) {
return 2;
}
else if (CsDiff >= 14) {
return 8;
}
}
else if (posBalance >= 48) {
if (CsDiff < 10) {
return 1;
}
else if (CsDiff >= 10) {
return 8;
}
}
}
else if (MVDiff >= 14) {
if (diffRsCsdiff < -28) {
return 4;
}
else if (diffRsCsdiff >= -28) {
if (Rs < 4725) {
return 2;
}
else if (Rs >= 4725) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 101) {
if (TSC < 8) {
if (mcTcor < 95) {
if (CsDiff < 11) {
if (MVDiff < 46) {
if (diffRsCsdiff < -63) {
if (Rs < 3333) {
return 1;
}
else if (Rs >= 3333) {
return 8;
}
}
else if (diffRsCsdiff >= -63) {
return 8;
}
}
else if (MVDiff >= 46) {
if (Rs < 2842) {
return 2;
}
else if (Rs >= 2842) {
return 1;
}
}
}
else if (CsDiff >= 11) {
if (posBalance < 33) {
if (Cs < 4229) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (Cs >= 4229) {
return 8;
}
}
else if (posBalance >= 33) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
if (MVDiff < 33) {
if (negBalance < 28) {
return 2;
}
else if (negBalance >= 28) {
if (diffMVdiffVal < -4) {
if (Rs < 2634) {
return 1;
}
else if (Rs >= 2634) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (MVDiff >= 33) {
if (Rs < 2025) {
return 4;
}
else if (Rs >= 2025) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (negBalance < 23) {
return 8;
}
else if (negBalance >= 23) {
if (ssDCval < 107) {
if (Rs < 4906) {
if (Rs < 2899) {
return 2;
}
else if (Rs >= 2899) {
return 8;
}
}
else if (Rs >= 4906) {
return 4;
}
}
else if (ssDCval >= 107) {
return 1;
}
}
}
}
else if (TSC >= 8) {
if (ssDCval < 123) {
if (Rs < 3549) {
if (ssDCval < 109) {
if (Rs < 2963) {
if (diffRsCsdiff < 60) {
if (RsCsDiff < 115) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
else if (RsCsDiff >= 115) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 4;
}
}
}
else if (diffRsCsdiff >= 60) {
return 4;
}
}
else if (Rs >= 2963) {
if (negBalance < 48) {
if (Rs < 3151) {
return 2;
}
else if (Rs >= 3151) {
return 8;
}
}
else if (negBalance >= 48) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 8;
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 1917) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 1;
}
}
else if (Rs >= 1917) {
return 8;
}
}
}
else if (Rs >= 3549) {
if (tcor < 85) {
if (RsDiff < 19) {
return 8;
}
else if (RsDiff >= 19) {
if (Rs < 5992) {
return 8;
}
else if (Rs >= 5992) {
return 2;
}
}
}
else if (tcor >= 85) {
if (AFD < 16) {
if (negBalance < 51) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 5228) {
return 2;
}
else if (Rs >= 5228) {
return 4;
}
}
}
else if (negBalance >= 51) {
if (mcTcor < 90) {
if (Rs < 3981) {
return 8;
}
else if (Rs >= 3981) {
return 1;
}
}
else if (mcTcor >= 90) {
if (Rs < 3942) {
return 2;
}
else if (Rs >= 3942) {
return 8;
}
}
}
}
else if (AFD >= 16) {
return 4;
}
}
}
}
else if (ssDCval >= 123) {
if (SC < 5220) {
if (negBalance < 39) {
if (mu_mv_mag_sq < 30) {
return 1;
}
else if (mu_mv_mag_sq >= 30) {
if (RsDiff < 11) {
return 2;
}
else if (RsDiff >= 11) {
return 8;
}
}
}
else if (negBalance >= 39) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (mu_mv_mag_sq < 46) {
return 8;
}
else if (mu_mv_mag_sq >= 46) {
if (Rs < 3077) {
return 4;
}
else if (Rs >= 3077) {
return 8;
}
}
}
}
}
else if (SC >= 5220) {
if (mcTcor < 82) {
if (posBalance < 54) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
if (Cs < 3301) {
if (diffRsCsdiff < 33) {
return 8;
}
else if (diffRsCsdiff >= 33) {
return 1;
}
}
else if (Cs >= 3301) {
return 1;
}
}
}
else if (posBalance >= 54) {
if (ssDCval < 134) {
if (Rs < 3292) {
return 2;
}
else if (Rs >= 3292) {
return 1;
}
}
else if (ssDCval >= 134) {
if (mu_mv_mag_sq < 42) {
return 8;
}
else if (mu_mv_mag_sq >= 42) {
if (Rs < 3896) {
return 8;
}
else if (Rs >= 3896) {
return 1;
}
}
}
}
}
else if (mcTcor >= 82) {
if (RsCsDiff < 183) {
if (TSC < 11) {
if (Cs < 2038) {
return 8;
}
else if (Cs >= 2038) {
if (diffRsCsdiff < -66) {
if (Rs < 3332) {
return 4;
}
else if (Rs >= 3332) {
return 8;
}
}
else if (diffRsCsdiff >= -66) {
return 1;
}
}
}
else if (TSC >= 11) {
if (Cs < 3019) {
return 8;
}
else if (Cs >= 3019) {
return 4;
}
}
}
else if (RsCsDiff >= 183) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 6203) {
if (CsDiff < 9) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
return 4;
}
}
else if (CsDiff >= 9) {
if (Cs < 4187) {
if (posBalance < 63) {
return 4;
}
else if (posBalance >= 63) {
return 8;
}
}
else if (Cs >= 4187) {
return 8;
}
}
}
else if (Rs >= 6203) {
if (diffAFD < -10) {
return 2;
}
else if (diffAFD >= -10) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (AFD >= 18) {
if (ssDCval < 154) {
if (refDCval < 75) {
if (MVDiff < 6) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (Rs < 3324) {
return 4;
}
else if (Rs >= 3324) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (tcor < 65) {
if (AFD < 40) {
if (CsDiff < 10) {
if (AFD < 21) {
if (Rs < 3068) {
return 4;
}
else if (Rs >= 3068) {
if (Rs < 4734) {
return 2;
}
else if (Rs >= 4734) {
return 4;
}
}
}
else if (AFD >= 21) {
if (Rs < 3743) {
return 1;
}
else if (Rs >= 3743) {
return 8;
}
}
}
else if (CsDiff >= 10) {
if (Rs < 4881) {
if (Rs < 1200) {
if (Cs < 1813) {
return 4;
}
else if (Cs >= 1813) {
if (Rs < 659) {
return 8;
}
else if (Rs >= 659) {
return 2;
}
}
}
else if (Rs >= 1200) {
if (TSC < 15) {
if (CsDiff < 12) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
else if (CsDiff >= 12) {
if (RsDiff < 22) {
if (negBalance < 50) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
return 4;
}
}
else if (negBalance >= 50) {
if (Rs < 1901) {
if (Rs < 1829) {
return 4;
}
else if (Rs >= 1829) {
return 2;
}
}
else if (Rs >= 1901) {
return 4;
}
}
}
else if (RsDiff >= 22) {
return 2;
}
}
}
else if (TSC >= 15) {
if (MVDiff < 25) {
if (Rs < 4520) {
return 8;
}
else if (Rs >= 4520) {
return 4;
}
}
else if (MVDiff >= 25) {
return 4;
}
}
}
}
else if (Rs >= 4881) {
if (Rs < 7065) {
return 2;
}
else if (Rs >= 7065) {
return 1;
}
}
}
}
else if (AFD >= 40) {
return 8;
}
}
else if (tcor >= 65) {
if (posBalance < 49) {
if (SC < 6455) {
if (CsDiff < 13) {
return 2;
}
else if (CsDiff >= 13) {
if (Rs < 2002) {
return 1;
}
else if (Rs >= 2002) {
if (Rs < 3164) {
return 4;
}
else if (Rs >= 3164) {
return 8;
}
}
}
}
else if (SC >= 6455) {
if (Cs < 5644) {
return 4;
}
else if (Cs >= 5644) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (mcTcor < 83) {
if (Rs < 2604) {
return 2;
}
else if (Rs >= 2604) {
if (Rs < 5389) {
return 8;
}
else if (Rs >= 5389) {
return 2;
}
}
}
else if (mcTcor >= 83) {
if (tcor < 77) {
if (diffMVdiffVal < -4) {
if (mu_mv_mag_sq < 50) {
if (tcor < 75) {
return 2;
}
else if (tcor >= 75) {
if (Cs < 2421) {
return 2;
}
else if (Cs >= 2421) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 50) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
if (Rs < 2710) {
if (ssDCval < 62) {
if (Rs < 1518) {
return 2;
}
else if (Rs >= 1518) {
return 4;
}
}
else if (ssDCval >= 62) {
if (diffRsCsdiff < -42) {
if (Rs < 2566) {
return 1;
}
else if (Rs >= 2566) {
return 8;
}
}
else if (diffRsCsdiff >= -42) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
}
}
else if (Rs >= 2710) {
if (CsDiff < 14) {
if (AFD < 20) {
if (Rs < 3715) {
if (Cs < 1852) {
return 2;
}
else if (Cs >= 1852) {
return 8;
}
}
else if (Rs >= 3715) {
return 2;
}
}
else if (AFD >= 20) {
return 2;
}
}
else if (CsDiff >= 14) {
if (negBalance < 49) {
if (RsDiff < 18) {
return 4;
}
else if (RsDiff >= 18) {
return 8;
}
}
else if (negBalance >= 49) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
}
}
else if (tcor >= 77) {
if (Cs < 2310) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 8;
}
}
else if (Cs >= 2310) {
if (diffAFD < -3) {
if (Rs < 3732) {
return 4;
}
else if (Rs >= 3732) {
return 2;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (mcTcor < 94) {
if (TSC < 21) {
if (RsCsDiff < 716) {
if (posBalance < 50) {
if (refDCval < 87) {
if (mu_mv_mag_sq < 52) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
return 8;
}
}
else if (mu_mv_mag_sq >= 52) {
return 4;
}
}
else if (refDCval >= 87) {
if (MVDiff < 26) {
if (mu_mv_mag_sq < 41) {
if (AFD < 25) {
if (gchDC < 2) {
if (diffRsCsdiff < -42) {
return 8;
}
else if (diffRsCsdiff >= -42) {
return 2;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (AFD >= 25) {
return 4;
}
}
else if (mu_mv_mag_sq >= 41) {
if (tcor < 76) {
if (Rs < 4200) {
return 8;
}
else if (Rs >= 4200) {
return 4;
}
}
else if (tcor >= 76) {
return 1;
}
}
}
else if (MVDiff >= 26) {
if (Rs < 888) {
return 2;
}
else if (Rs >= 888) {
if (Rs < 6040) {
return 8;
}
else if (Rs >= 6040) {
return 2;
}
}
}
}
}
else if (posBalance >= 50) {
if (AFD < 22) {
if (Rs < 4515) {
if (ssDCval < 108) {
if (negBalance < 58) {
if (SC < 3679) {
if (TSC < 11) {
if (Cs < 2162) {
return 2;
}
else if (Cs >= 2162) {
return 4;
}
}
else if (TSC >= 11) {
return 1;
}
}
else if (SC >= 3679) {
if (Cs < 3390) {
if (Rs < 2950) {
if (mcTcor < 79) {
if (CsDiff < 9) {
if (Rs < 2443) {
return 1;
}
else if (Rs >= 2443) {
return 8;
}
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (mcTcor >= 79) {
if (RsCsDiff < 223) {
if (negBalance < 50) {
if (diffRsCsdiff < -49) {
return 8;
}
else if (diffRsCsdiff >= -49) {
if (refDCval < 89) {
return 4;
}
else if (refDCval >= 89) {
return 1;
}
}
}
else if (negBalance >= 50) {
if (AFD < 19) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 8;
}
}
else if (AFD >= 19) {
if (Rs < 2646) {
if (tcor < 61) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (tcor >= 61) {
if (posBalance < 59) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (posBalance >= 59) {
return 8;
}
}
}
else if (Rs >= 2646) {
if (Cs < 1993) {
return 4;
}
else if (Cs >= 1993) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 223) {
if (ssDCval < 76) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (ssDCval >= 76) {
return 8;
}
}
}
}
else if (Rs >= 2950) {
if (refDCval < 83) {
if (SC < 5596) {
return 4;
}
else if (SC >= 5596) {
if (ssDCval < 76) {
if (Rs < 3276) {
return 8;
}
else if (Rs >= 3276) {
if (Rs < 3412) {
return 4;
}
else if (Rs >= 3412) {
return 1;
}
}
}
else if (ssDCval >= 76) {
if (diffRsCsdiff < 102) {
if (negBalance < 52) {
if (Rs < 3605) {
return 1;
}
else if (Rs >= 3605) {
return 8;
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (diffRsCsdiff >= 102) {
if (Rs < 3768) {
return 2;
}
else if (Rs >= 3768) {
return 8;
}
}
}
}
}
else if (refDCval >= 83) {
if (diffRsCsdiff < 121) {
if (diffAFD < 4) {
if (tcor < 68) {
return 8;
}
else if (tcor >= 68) {
if (diffAFD < -3) {
if (Rs < 3312) {
return 8;
}
else if (Rs >= 3312) {
return 2;
}
}
else if (diffAFD >= -3) {
if (mu_mv_mag_sq < 46) {
return 4;
}
else if (mu_mv_mag_sq >= 46) {
return 2;
}
}
}
}
else if (diffAFD >= 4) {
if (Rs < 3858) {
return 8;
}
else if (Rs >= 3858) {
return 2;
}
}
}
else if (diffRsCsdiff >= 121) {
if (MVDiff < 45) {
return 8;
}
else if (MVDiff >= 45) {
return 4;
}
}
}
}
}
else if (Cs >= 3390) {
if (ssDCval < 96) {
if (tcor < 67) {
if (Rs < 3605) {
return 4;
}
else if (Rs >= 3605) {
return 8;
}
}
else if (tcor >= 67) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
return 2;
}
}
}
else if (ssDCval >= 96) {
if (Rs < 3413) {
return 4;
}
else if (Rs >= 3413) {
return 1;
}
}
}
}
}
else if (negBalance >= 58) {
if (RsCsDiff < 328) {
if (SC < 4436) {
if (ssDCval < 88) {
return 4;
}
else if (ssDCval >= 88) {
if (SC < 4138) {
if (Rs < 1227) {
return 2;
}
else if (Rs >= 1227) {
return 1;
}
}
else if (SC >= 4138) {
return 8;
}
}
}
else if (SC >= 4436) {
if (Rs < 3717) {
return 4;
}
else if (Rs >= 3717) {
return 8;
}
}
}
else if (RsCsDiff >= 328) {
return 8;
}
}
}
else if (ssDCval >= 108) {
if (SC < 6374) {
if (negBalance < 59) {
if (TSC < 10) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
if (Rs < 2178) {
return 4;
}
else if (Rs >= 2178) {
return 2;
}
}
}
else if (TSC >= 10) {
if (AFD < 21) {
if (diffAFD < -4) {
if (Cs < 2757) {
return 8;
}
else if (Cs >= 2757) {
return 4;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (AFD >= 21) {
if (TSC < 16) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (TSC >= 16) {
return 2;
}
}
}
}
else if (negBalance >= 59) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (Cs < 2331) {
if (Rs < 1711) {
return 2;
}
else if (Rs >= 1711) {
return 1;
}
}
else if (Cs >= 2331) {
return 8;
}
}
}
}
else if (SC >= 6374) {
if (negBalance < 47) {
if (Rs < 4029) {
return 8;
}
else if (Rs >= 4029) {
return 1;
}
}
else if (negBalance >= 47) {
if (Rs < 3138) {
return 2;
}
else if (Rs >= 3138) {
return 4;
}
}
}
}
}
else if (Rs >= 4515) {
if (RsDiff < 17) {
if (tcor < 85) {
if (diffTSC < -3) {
if (Rs < 5079) {
return 1;
}
else if (Rs >= 5079) {
if (Rs < 6200) {
return 2;
}
else if (Rs >= 6200) {
return 8;
}
}
}
else if (diffTSC >= -3) {
return 8;
}
}
else if (tcor >= 85) {
return 4;
}
}
else if (RsDiff >= 17) {
if (mcTcor < 79) {
return 8;
}
else if (mcTcor >= 79) {
if (Cs < 2086) {
return 8;
}
else if (Cs >= 2086) {
if (negBalance < 53) {
if (diffRsCsdiff < -37) {
if (RsDiff < 19) {
return 4;
}
else if (RsDiff >= 19) {
return 2;
}
}
else if (diffRsCsdiff >= -37) {
return 4;
}
}
else if (negBalance >= 53) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 22) {
if (refDCval < 106) {
if (mcTcor < 87) {
if (diffAFD < 2) {
if (TSC < 16) {
if (CsDiff < 16) {
if (tcor < 57) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 1;
}
}
else if (tcor >= 57) {
if (diffRsCsdiff < -30) {
return 4;
}
else if (diffRsCsdiff >= -30) {
if (RsDiff < 16) {
if (Cs < 2282) {
return 8;
}
else if (Cs >= 2282) {
if (Cs < 2506) {
return 4;
}
else if (Cs >= 2506) {
if (Cs < 3401) {
return 1;
}
else if (Cs >= 3401) {
return 8;
}
}
}
}
else if (RsDiff >= 16) {
return 4;
}
}
}
}
else if (CsDiff >= 16) {
if (diffMVdiffVal < -13) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
else if (diffMVdiffVal >= -13) {
if (Rs < 3220) {
if (ssDCval < 84) {
return 2;
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (Rs >= 3220) {
return 2;
}
}
}
}
else if (TSC >= 16) {
if (Cs < 2406) {
return 2;
}
else if (Cs >= 2406) {
if (AFD < 23) {
if (Rs < 6005) {
return 8;
}
else if (Rs >= 6005) {
return 4;
}
}
else if (AFD >= 23) {
return 1;
}
}
}
}
else if (diffAFD >= 2) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
if (mu_mv_mag_sq < 13) {
if (RsDiff < 14) {
if (Cs < 4483) {
return 8;
}
else if (Cs >= 4483) {
return 2;
}
}
else if (RsDiff >= 14) {
if (Rs < 4036) {
return 4;
}
else if (Rs >= 4036) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 13) {
return 8;
}
}
}
}
else if (mcTcor >= 87) {
if (ssDCval < 91) {
if (mu_mv_mag_sq < 3) {
if (RsCsDiff < 229) {
if (Rs < 1655) {
return 4;
}
else if (Rs >= 1655) {
return 8;
}
}
else if (RsCsDiff >= 229) {
if (Cs < 2128) {
return 8;
}
else if (Cs >= 2128) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffAFD < -2) {
if (AFD < 27) {
return 4;
}
else if (AFD >= 27) {
return 1;
}
}
else if (diffAFD >= -2) {
if (refDCval < 90) {
if (RsDiff < 17) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
return 8;
}
}
else if (RsDiff >= 17) {
return 4;
}
}
else if (refDCval >= 90) {
if (Rs < 2204) {
return 8;
}
else if (Rs >= 2204) {
return 1;
}
}
}
}
}
else if (ssDCval >= 91) {
if (tcor < 72) {
if (Rs < 815) {
return 2;
}
else if (Rs >= 815) {
if (negBalance < 63) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
if (Rs < 1818) {
return 8;
}
else if (Rs >= 1818) {
return 4;
}
}
}
else if (negBalance >= 63) {
return 1;
}
}
}
else if (tcor >= 72) {
if (Rs < 2224) {
if (RsCsDiff < 279) {
return 1;
}
else if (RsCsDiff >= 279) {
return 4;
}
}
else if (Rs >= 2224) {
if (Cs < 4660) {
return 8;
}
else if (Cs >= 4660) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 106) {
if (RsCsDiff < 205) {
if (RsDiff < 10) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
return 2;
}
}
else if (RsDiff >= 10) {
return 1;
}
}
else if (RsCsDiff >= 205) {
if (refDCval < 117) {
if (RsCsDiff < 280) {
if (posBalance < 62) {
if (Rs < 1686) {
return 4;
}
else if (Rs >= 1686) {
return 1;
}
}
else if (posBalance >= 62) {
return 8;
}
}
else if (RsCsDiff >= 280) {
if (Rs < 3564) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (negBalance < 60) {
if (Rs < 1723) {
return 8;
}
else if (Rs >= 1723) {
return 2;
}
}
else if (negBalance >= 60) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (Cs < 2554) {
return 8;
}
else if (Cs >= 2554) {
return 4;
}
}
}
}
}
else if (Rs >= 3564) {
if (AFD < 26) {
if (Rs < 5563) {
return 1;
}
else if (Rs >= 5563) {
return 4;
}
}
else if (AFD >= 26) {
if (Rs < 4038) {
return 8;
}
else if (Rs >= 4038) {
return 2;
}
}
}
}
}
else if (refDCval >= 117) {
if (ssDCval < 134) {
if (CsDiff < 15) {
if (diffMVdiffVal < -12) {
return 2;
}
else if (diffMVdiffVal >= -12) {
if (diffAFD < 12) {
if (mcTcor < 91) {
if (gchDC < 6) {
if (Cs < 4058) {
return 8;
}
else if (Cs >= 4058) {
return 1;
}
}
else if (gchDC >= 6) {
return 1;
}
}
else if (mcTcor >= 91) {
if (Rs < 3737) {
return 1;
}
else if (Rs >= 3737) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
else if (diffAFD >= 12) {
if (Rs < 3650) {
return 1;
}
else if (Rs >= 3650) {
return 4;
}
}
}
}
else if (CsDiff >= 15) {
if (RsDiff < 15) {
if (Rs < 3166) {
return 8;
}
else if (Rs >= 3166) {
return 1;
}
}
else if (RsDiff >= 15) {
return 4;
}
}
}
else if (ssDCval >= 134) {
if (negBalance < 47) {
if (Cs < 2139) {
return 4;
}
else if (Cs >= 2139) {
return 8;
}
}
else if (negBalance >= 47) {
if (RsCsDiff < 307) {
if (Rs < 3027) {
return 8;
}
else if (Rs >= 3027) {
return 2;
}
}
else if (RsCsDiff >= 307) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 716) {
if (gchDC < 2) {
if (Rs < 5636) {
if (TSC < 17) {
return 4;
}
else if (TSC >= 17) {
if (Cs < 4255) {
if (Rs < 3998) {
return 1;
}
else if (Rs >= 3998) {
return 2;
}
}
else if (Cs >= 4255) {
return 8;
}
}
}
else if (Rs >= 5636) {
if (refDCval < 149) {
if (Cs < 6537) {
if (AFD < 19) {
if (Rs < 7006) {
return 2;
}
else if (Rs >= 7006) {
return 4;
}
}
else if (AFD >= 19) {
if (mu_mv_mag_sq < 98) {
return 4;
}
else if (mu_mv_mag_sq >= 98) {
if (RsDiff < 22) {
return 8;
}
else if (RsDiff >= 22) {
return 4;
}
}
}
}
else if (Cs >= 6537) {
return 8;
}
}
else if (refDCval >= 149) {
if (Rs < 7304) {
return 2;
}
else if (Rs >= 7304) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (refDCval < 86) {
if (RsDiff < 26) {
if (Rs < 4663) {
return 1;
}
else if (Rs >= 4663) {
return 4;
}
}
else if (RsDiff >= 26) {
return 2;
}
}
else if (refDCval >= 86) {
if (mu_mv_mag_sq < 37) {
if (diffRsCsdiff < 226) {
if (refDCval < 95) {
return 1;
}
else if (refDCval >= 95) {
if (diffTSC < -1) {
if (negBalance < 58) {
if (Rs < 7025) {
return 2;
}
else if (Rs >= 7025) {
return 4;
}
}
else if (negBalance >= 58) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (diffRsCsdiff >= 226) {
if (Rs < 2796) {
return 2;
}
else if (Rs >= 2796) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (negBalance < 64) {
if (tcor < 74) {
return 8;
}
else if (tcor >= 74) {
if (Rs < 5504) {
return 8;
}
else if (Rs >= 5504) {
return 2;
}
}
}
else if (negBalance >= 64) {
return 1;
}
}
}
}
}
}
else if (TSC >= 21) {
if (RsDiff < 25) {
return 8;
}
else if (RsDiff >= 25) {
if (AFD < 44) {
if (Cs < 2595) {
return 1;
}
else if (Cs >= 2595) {
return 4;
}
}
else if (AFD >= 44) {
if (Cs < 3720) {
return 2;
}
else if (Cs >= 3720) {
return 8;
}
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 5737) {
if (Rs < 4152) {
if (refDCval < 110) {
if (posBalance < 58) {
if (RsDiff < 10) {
if (Rs < 1006) {
return 2;
}
else if (Rs >= 1006) {
return 1;
}
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (posBalance >= 58) {
if (Rs < 2821) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
if (mu_mv_mag_sq < 3) {
if (SC < 4185) {
if (Cs < 2186) {
return 4;
}
else if (Cs >= 2186) {
return 1;
}
}
else if (SC >= 4185) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
if (Rs < 1407) {
return 4;
}
else if (Rs >= 1407) {
if (Rs < 2401) {
return 2;
}
else if (Rs >= 2401) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2821) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
if (AFD < 28) {
return 8;
}
else if (AFD >= 28) {
return 1;
}
}
}
}
}
else if (refDCval >= 110) {
if (RsCsDiff < 555) {
if (tcor < 77) {
if (Cs < 2725) {
return 1;
}
else if (Cs >= 2725) {
return 8;
}
}
else if (tcor >= 77) {
return 8;
}
}
else if (RsCsDiff >= 555) {
return 2;
}
}
}
else if (Rs >= 4152) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
if (CsDiff < 12) {
return 4;
}
else if (CsDiff >= 12) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (SC < 6960) {
return 1;
}
else if (SC >= 6960) {
if (TSC < 9) {
if (Rs < 5053) {
return 8;
}
else if (Rs >= 5053) {
return 1;
}
}
else if (TSC >= 9) {
return 2;
}
}
}
}
}
else if (Rs >= 5737) {
if (ssDCval < 126) {
if (Rs < 7690) {
return 8;
}
else if (Rs >= 7690) {
return 1;
}
}
else if (ssDCval >= 126) {
return 8;
}
}
}
}
}
else if (ssDCval >= 154) {
if (refDCval < 168) {
if (posBalance < 53) {
if (refDCval < 156) {
return 8;
}
else if (refDCval >= 156) {
if (Rs < 4500) {
return 4;
}
else if (Rs >= 4500) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (Rs < 3335) {
if (Rs < 2190) {
return 2;
}
else if (Rs >= 2190) {
return 8;
}
}
else if (Rs >= 3335) {
if (posBalance < 55) {
if (Rs < 5774) {
return 8;
}
else if (Rs >= 5774) {
return 4;
}
}
else if (posBalance >= 55) {
return 4;
}
}
}
}
else if (refDCval >= 168) {
if (mu_mv_mag_sq < 41) {
if (Rs < 3506) {
return 1;
}
else if (Rs >= 3506) {
return 2;
}
}
else if (mu_mv_mag_sq >= 41) {
if (Cs < 2760) {
return 4;
}
else if (Cs >= 2760) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 88) {
if (RsDiff < 9) {
if (Rs < 3687) {
if (ssDCval < 69) {
if (Rs < 1576) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 2;
}
}
else if (Rs >= 1576) {
if (mu_mv_mag_sq < 98) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 3369) {
return 1;
}
else if (Rs >= 3369) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (Rs < 2598) {
return 4;
}
else if (Rs >= 2598) {
return 8;
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 42) {
if (MVDiff < 5) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (Cs < 1961) {
return 2;
}
else if (Cs >= 1961) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (Rs < 1558) {
if (tcor < 92) {
if (refDCval < 103) {
return 2;
}
else if (refDCval >= 103) {
if (Rs < 1450) {
return 4;
}
else if (Rs >= 1450) {
return 2;
}
}
}
else if (tcor >= 92) {
if (Rs < 1285) {
return 1;
}
else if (Rs >= 1285) {
return 8;
}
}
}
else if (Rs >= 1558) {
if (MVDiff < 10) {
if (ssDCval < 108) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
return 2;
}
}
else if (ssDCval >= 108) {
if (Rs < 2027) {
return 2;
}
else if (Rs >= 2027) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (Cs < 1369) {
if (Rs < 1792) {
return 8;
}
else if (Rs >= 1792) {
return 1;
}
}
else if (Cs >= 1369) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (Rs < 2826) {
if (mu_mv_mag_sq < 82) {
if (Cs < 2607) {
if (mcTcor < 97) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 2601) {
return 8;
}
else if (Rs >= 2601) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (Rs < 1951) {
return 4;
}
else if (Rs >= 1951) {
return 1;
}
}
}
else if (Cs >= 2607) {
if (mcTcor < 93) {
if (Rs < 2602) {
return 4;
}
else if (Rs >= 2602) {
if (Cs < 3038) {
return 8;
}
else if (Cs >= 3038) {
return 4;
}
}
}
else if (mcTcor >= 93) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (diffTSC < -4) {
return 4;
}
else if (diffTSC >= -4) {
return 8;
}
}
}
else if (Rs >= 2826) {
if (ssDCval < 104) {
if (mcTcor < 94) {
if (ssDCval < 81) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
return 4;
}
}
}
else if (ssDCval >= 81) {
if (mu_mv_mag_sq < 71) {
return 8;
}
else if (mu_mv_mag_sq >= 71) {
if (Cs < 3970) {
return 4;
}
else if (Cs >= 3970) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (tcor < 91) {
return 1;
}
else if (tcor >= 91) {
return 4;
}
}
}
else if (ssDCval >= 104) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 3687) {
if (mcTcor < 92) {
if (mu_mv_mag_sq < 79) {
if (Rs < 4552) {
return 1;
}
else if (Rs >= 4552) {
return 8;
}
}
else if (mu_mv_mag_sq >= 79) {
return 1;
}
}
else if (mcTcor >= 92) {
if (Rs < 4384) {
if (MVDiff < 2) {
if (CsDiff < 9) {
return 4;
}
else if (CsDiff >= 9) {
if (Rs < 4075) {
return 8;
}
else if (Rs >= 4075) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (Cs < 2195) {
if (Rs < 3810) {
return 8;
}
else if (Rs >= 3810) {
return 1;
}
}
else if (Cs >= 2195) {
if (Rs < 4364) {
return 8;
}
else if (Rs >= 4364) {
return 4;
}
}
}
}
else if (Rs >= 4384) {
if (Cs < 3890) {
if (MVDiff < 9) {
if (CsDiff < 8) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 1;
}
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (MVDiff >= 9) {
if (Rs < 5155) {
return 8;
}
else if (Rs >= 5155) {
return 1;
}
}
}
else if (Cs >= 3890) {
if (Rs < 5680) {
return 8;
}
else if (Rs >= 5680) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (Cs < 2291) {
if (Rs < 3169) {
if (mu_mv_mag_sq < 57) {
if (ssDCval < 91) {
if (Cs < 1490) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (Cs >= 1490) {
if (mcTcor < 97) {
if (Cs < 2279) {
return 8;
}
else if (Cs >= 2279) {
return 4;
}
}
else if (mcTcor >= 97) {
return 2;
}
}
}
else if (ssDCval >= 91) {
if (MVDiff < 55) {
if (MVDiff < 2) {
if (Rs < 2775) {
return 8;
}
else if (Rs >= 2775) {
return 4;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (MVDiff >= 55) {
if (Cs < 1485) {
return 2;
}
else if (Cs >= 1485) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (Cs < 1478) {
if (Cs < 1343) {
return 8;
}
else if (Cs >= 1343) {
return 2;
}
}
else if (Cs >= 1478) {
if (Rs < 2151) {
if (Rs < 2004) {
if (Rs < 1914) {
return 8;
}
else if (Rs >= 1914) {
return 2;
}
}
else if (Rs >= 2004) {
if (Rs < 2080) {
return 4;
}
else if (Rs >= 2080) {
return 1;
}
}
}
else if (Rs >= 2151) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (tcor < 91) {
if (gchDC < 1) {
if (Rs < 3113) {
return 4;
}
else if (Rs >= 3113) {
return 8;
}
}
else if (gchDC >= 1) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
return 4;
}
}
}
else if (tcor >= 91) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 3169) {
if (MVDiff < 15) {
if (mcTcor < 92) {
if (mu_mv_mag_sq < 77) {
if (Cs < 1893) {
return 2;
}
else if (Cs >= 1893) {
return 8;
}
}
else if (mu_mv_mag_sq >= 77) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
if (SC < 5573) {
if (Rs < 3621) {
return 8;
}
else if (Rs >= 3621) {
return 2;
}
}
else if (SC >= 5573) {
return 1;
}
}
}
}
else if (mcTcor >= 92) {
if (refDCval < 62) {
if (Rs < 9139) {
return 4;
}
else if (Rs >= 9139) {
return 2;
}
}
else if (refDCval >= 62) {
if (ssDCval < 114) {
if (RsCsDiff < 121) {
if (refDCval < 93) {
return 2;
}
else if (refDCval >= 93) {
if (MVDiff < 5) {
if (diffMVdiffVal < 1) {
if (Rs < 3531) {
return 4;
}
else if (Rs >= 3531) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (RsCsDiff >= 121) {
if (tcor < 89) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 13) {
if (Cs < 1746) {
if (Cs < 1368) {
return 4;
}
else if (Cs >= 1368) {
return 2;
}
}
else if (Cs >= 1746) {
if (Rs < 4544) {
return 8;
}
else if (Rs >= 4544) {
return 4;
}
}
}
else if (RsDiff >= 13) {
return 2;
}
}
}
else if (tcor >= 89) {
if (refDCval < 73) {
if (negBalance < 43) {
return 2;
}
else if (negBalance >= 43) {
return 4;
}
}
else if (refDCval >= 73) {
if (RsDiff < 10) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
else if (RsDiff >= 10) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 114) {
if (Rs < 5789) {
if (Rs < 5494) {
if (refDCval < 132) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 2;
}
}
else if (refDCval >= 132) {
if (SC < 6174) {
return 1;
}
else if (SC >= 6174) {
return 2;
}
}
}
else if (Rs >= 5494) {
return 4;
}
}
else if (Rs >= 5789) {
if (ssDCval < 118) {
return 8;
}
else if (ssDCval >= 118) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 15) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (MVDiff < 23) {
if (ssDCval < 109) {
if (Cs < 2049) {
if (mcTcor < 93) {
if (refDCval < 82) {
return 2;
}
else if (refDCval >= 82) {
return 8;
}
}
else if (mcTcor >= 93) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 2;
}
}
}
else if (Cs >= 2049) {
if (negBalance < 29) {
if (Rs < 5192) {
return 2;
}
else if (Rs >= 5192) {
return 4;
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
else if (ssDCval >= 109) {
return 8;
}
}
else if (MVDiff >= 23) {
if (Cs < 1682) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
else if (Cs >= 1682) {
if (Cs < 1698) {
return 2;
}
else if (Cs >= 1698) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Cs < 2044) {
return 2;
}
else if (Cs >= 2044) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Cs >= 2291) {
if (posBalance < 51) {
if (ssDCval < 73) {
if (RsDiff < 12) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (Rs < 3852) {
return 1;
}
else if (Rs >= 3852) {
return 8;
}
}
}
else if (RsDiff >= 12) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 4;
}
}
}
else if (ssDCval >= 73) {
if (refDCval < 107) {
if (mu_mv_mag_sq < 68) {
if (mcTcor < 95) {
if (Rs < 2712) {
return 4;
}
else if (Rs >= 2712) {
if (Rs < 4992) {
return 8;
}
else if (Rs >= 4992) {
if (Rs < 5655) {
return 2;
}
else if (Rs >= 5655) {
return 4;
}
}
}
}
else if (mcTcor >= 95) {
if (Cs < 3552) {
return 2;
}
else if (Cs >= 3552) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (diffRsCsdiff < -142) {
if (Rs < 3022) {
return 2;
}
else if (Rs >= 3022) {
return 4;
}
}
else if (diffRsCsdiff >= -142) {
return 2;
}
}
}
else if (refDCval >= 107) {
if (ssDCval < 121) {
if (SC < 8609) {
if (Cs < 3321) {
return 8;
}
else if (Cs >= 3321) {
return 2;
}
}
else if (SC >= 8609) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
if (Rs < 5281) {
return 1;
}
else if (Rs >= 5281) {
return 4;
}
}
}
}
else if (ssDCval >= 121) {
if (MVDiff < 12) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
return 2;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
}
}
}
else if (posBalance >= 51) {
if (ssDCval < 126) {
if (MVDiff < 10) {
if (Cs < 3379) {
if (Cs < 2684) {
if (Rs < 3459) {
return 8;
}
else if (Rs >= 3459) {
if (SC < 6519) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
return 2;
}
}
}
else if (SC >= 6519) {
if (TSC < 10) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
if (Rs < 5423) {
return 8;
}
else if (Rs >= 5423) {
return 4;
}
}
}
else if (TSC >= 10) {
return 1;
}
}
}
}
else if (Cs >= 2684) {
if (Rs < 4122) {
if (Rs < 3020) {
if (RsCsDiff < 172) {
return 4;
}
else if (RsCsDiff >= 172) {
return 8;
}
}
else if (Rs >= 3020) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
return 4;
}
}
}
else if (Rs >= 4122) {
if (Rs < 5279) {
if (mu_mv_mag_sq < 54) {
if (Rs < 4439) {
return 4;
}
else if (Rs >= 4439) {
return 2;
}
}
else if (mu_mv_mag_sq >= 54) {
if (MVDiff < 6) {
if (Cs < 3335) {
return 8;
}
else if (Cs >= 3335) {
if (Rs < 4781) {
return 4;
}
else if (Rs >= 4781) {
return 8;
}
}
}
else if (MVDiff >= 6) {
return 1;
}
}
}
else if (Rs >= 5279) {
return 4;
}
}
}
}
else if (Cs >= 3379) {
if (mu_mv_mag_sq < 73) {
if (Cs < 4599) {
return 4;
}
else if (Cs >= 4599) {
return 8;
}
}
else if (mu_mv_mag_sq >= 73) {
if (RsCsDiff < 172) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
else if (RsCsDiff >= 172) {
return 8;
}
}
}
}
else if (MVDiff >= 10) {
if (Rs < 6441) {
return 8;
}
else if (Rs >= 6441) {
if (Rs < 7952) {
return 2;
}
else if (Rs >= 7952) {
return 1;
}
}
}
}
else if (ssDCval >= 126) {
if (mcTcor < 92) {
if (ssDCval < 140) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
else if (ssDCval >= 140) {
if (Rs < 6057) {
return 8;
}
else if (Rs >= 6057) {
return 4;
}
}
}
else if (mcTcor >= 92) {
if (RsCsDiff < 147) {
if (Rs < 4687) {
return 8;
}
else if (Rs >= 4687) {
return 1;
}
}
else if (RsCsDiff >= 147) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Cs < 1641) {
if (Rs < 3093) {
if (Rs < 1455) {
if (ssDCval < 122) {
if (negBalance < 2) {
if (Rs < 1192) {
return 2;
}
else if (Rs >= 1192) {
return 8;
}
}
else if (negBalance >= 2) {
if (refDCval < 63) {
if (AFD < 3) {
if (mu_mv_mag_sq < 128) {
if (negBalance < 24) {
if (diffRsCsdiff < -6) {
if (Rs < 1249) {
return 4;
}
else if (Rs >= 1249) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (Rs < 753) {
return 1;
}
else if (Rs >= 753) {
return 8;
}
}
}
}
else if (negBalance >= 24) {
if (Rs < 1111) {
if (MVDiff < 1) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 8;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (Rs >= 1111) {
if (SC < 2922) {
return 4;
}
else if (SC >= 2922) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1482) {
if (Cs < 1343) {
return 4;
}
else if (Cs >= 1343) {
return 8;
}
}
else if (Cs >= 1482) {
return 1;
}
}
}
else if (AFD >= 3) {
if (refDCval < 57) {
if (refDCval < 54) {
if (mu_mv_mag_sq < 111) {
if (Rs < 1424) {
return 4;
}
else if (Rs >= 1424) {
return 8;
}
}
else if (mu_mv_mag_sq >= 111) {
if (posBalance < 51) {
if (mu_mv_mag_sq < 122) {
return 4;
}
else if (mu_mv_mag_sq >= 122) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 4;
}
}
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (refDCval >= 54) {
return 4;
}
}
else if (refDCval >= 57) {
return 8;
}
}
}
else if (refDCval >= 63) {
if (MVDiff < 2) {
if (SC < 1941) {
return 1;
}
else if (SC >= 1941) {
if (negBalance < 37) {
if (Cs < 1352) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Cs >= 1352) {
return 8;
}
}
else if (negBalance >= 37) {
if (Rs < 946) {
return 1;
}
else if (Rs >= 946) {
if (ssDCval < 83) {
if (Rs < 1416) {
return 4;
}
else if (Rs >= 1416) {
return 8;
}
}
else if (ssDCval >= 83) {
return 8;
}
}
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 107) {
if (Cs < 1498) {
if (diffAFD < 1) {
if (mu_mv_mag_sq < 112) {
if (Rs < 1019) {
if (MVDiff < 79) {
return 4;
}
else if (MVDiff >= 79) {
return 8;
}
}
else if (Rs >= 1019) {
return 8;
}
}
else if (mu_mv_mag_sq >= 112) {
return 4;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (Cs >= 1498) {
return 8;
}
}
else if (ssDCval >= 107) {
return 1;
}
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 2) {
if (posBalance < 28) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (posBalance >= 28) {
if (Rs < 1382) {
return 4;
}
else if (Rs >= 1382) {
return 8;
}
}
}
else if (AFD >= 2) {
if (posBalance < 47) {
if (ssDCval < 133) {
return 8;
}
else if (ssDCval >= 133) {
if (Cs < 1391) {
return 8;
}
else if (Cs >= 1391) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (Rs < 958) {
return 8;
}
else if (Rs >= 958) {
if (ssDCval < 130) {
return 4;
}
else if (ssDCval >= 130) {
if (Rs < 1112) {
return 8;
}
else if (Rs >= 1112) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1455) {
if (ssDCval < 153) {
if (ssDCval < 86) {
if (mcTcor < 97) {
if (posBalance < 44) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (AFD < 3) {
if (Rs < 1761) {
return 8;
}
else if (Rs >= 1761) {
return 1;
}
}
else if (AFD >= 3) {
if (ssDCval < 45) {
return 4;
}
else if (ssDCval >= 45) {
if (Rs < 1992) {
return 4;
}
else if (Rs >= 1992) {
if (Rs < 2518) {
return 8;
}
else if (Rs >= 2518) {
if (Rs < 2756) {
return 4;
}
else if (Rs >= 2756) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 44) {
if (Rs < 2813) {
return 8;
}
else if (Rs >= 2813) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 18) {
if (SC < 4469) {
if (SC < 3298) {
if (Cs < 1441) {
return 8;
}
else if (Cs >= 1441) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
return 1;
}
}
}
else if (SC >= 3298) {
if (Cs < 1350) {
if (Rs < 2342) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 1341) {
return 8;
}
else if (Cs >= 1341) {
if (Rs < 2316) {
return 2;
}
else if (Rs >= 2316) {
return 8;
}
}
}
}
else if (Rs >= 2342) {
return 2;
}
}
else if (Cs >= 1350) {
return 8;
}
}
}
else if (SC >= 4469) {
return 1;
}
}
else if (negBalance >= 18) {
if (Cs < 1364) {
if (Rs < 2443) {
return 8;
}
else if (Rs >= 2443) {
return 2;
}
}
else if (Cs >= 1364) {
if (Rs < 1501) {
if (mu_mv_mag_sq < 123) {
if (Rs < 1489) {
return 1;
}
else if (Rs >= 1489) {
return 4;
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1498) {
return 8;
}
else if (Rs >= 1498) {
return 1;
}
}
}
else if (Rs >= 1501) {
if (ssDCval < 75) {
if (Rs < 1894) {
if (Cs < 1370) {
return 8;
}
else if (Cs >= 1370) {
return 4;
}
}
else if (Rs >= 1894) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 8;
}
}
}
else if (ssDCval >= 75) {
if (Rs < 2469) {
if (mu_mv_mag_sq < 124) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 4;
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (Rs >= 2469) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (refDCval < 117) {
if (MVDiff < 122) {
if (TSC < 11) {
if (ssDCval < 92) {
if (mu_mv_mag_sq < 110) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 110) {
if (diffAFD < -2) {
if (Cs < 1498) {
return 2;
}
else if (Cs >= 1498) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (ssDCval >= 92) {
return 8;
}
}
else if (TSC >= 11) {
return 1;
}
}
else if (MVDiff >= 122) {
return 4;
}
}
else if (refDCval >= 117) {
if (CsDiff < 2) {
if (diffAFD < -19) {
if (diffAFD < -22) {
return 8;
}
else if (diffAFD >= -22) {
if (Rs < 2408) {
return 4;
}
else if (Rs >= 2408) {
return 1;
}
}
}
else if (diffAFD >= -19) {
return 8;
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 125) {
if (TSC < 6) {
if (MVDiff < 113) {
return 8;
}
else if (MVDiff >= 113) {
return 1;
}
}
else if (TSC >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 2257) {
return 1;
}
else if (Rs >= 2257) {
if (Rs < 2826) {
return 8;
}
else if (Rs >= 2826) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 153) {
if (ssDCval < 155) {
return 2;
}
else if (ssDCval >= 155) {
return 8;
}
}
}
}
else if (Rs >= 3093) {
if (refDCval < 77) {
if (mu_mv_mag_sq < 125) {
if (SC < 4912) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (SC >= 4912) {
if (ssDCval < 55) {
if (Rs < 4011) {
return 4;
}
else if (Rs >= 4011) {
return 8;
}
}
else if (ssDCval >= 55) {
if (SC < 5100) {
if (ssDCval < 67) {
if (Rs < 3469) {
return 4;
}
else if (Rs >= 3469) {
if (Rs < 3571) {
return 8;
}
else if (Rs >= 3571) {
return 8;
}
}
}
else if (ssDCval >= 67) {
return 1;
}
}
else if (SC >= 5100) {
if (ssDCval < 74) {
if (Cs < 1337) {
return 1;
}
else if (Cs >= 1337) {
return 4;
}
}
else if (ssDCval >= 74) {
if (Rs < 4276) {
return 2;
}
else if (Rs >= 4276) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1570) {
if (negBalance < 25) {
if (Rs < 3658) {
if (Rs < 3270) {
return 8;
}
else if (Rs >= 3270) {
if (Rs < 3544) {
return 1;
}
else if (Rs >= 3544) {
return 4;
}
}
}
else if (Rs >= 3658) {
return 8;
}
}
else if (negBalance >= 25) {
return 4;
}
}
else if (Cs >= 1570) {
if (Rs < 3238) {
return 8;
}
else if (Rs >= 3238) {
return 4;
}
}
}
}
else if (refDCval >= 77) {
if (ssDCval < 113) {
if (posBalance < 43) {
if (negBalance < 28) {
if (Rs < 4511) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
if (Cs < 1559) {
return 4;
}
else if (Cs >= 1559) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
if (Rs < 3268) {
return 8;
}
else if (Rs >= 3268) {
return 4;
}
}
}
}
}
else if (Rs >= 4511) {
return 2;
}
}
else if (negBalance >= 28) {
if (ssDCval < 97) {
if (Rs < 3905) {
if (Rs < 3483) {
return 2;
}
else if (Rs >= 3483) {
return 4;
}
}
else if (Rs >= 3905) {
return 8;
}
}
else if (ssDCval >= 97) {
return 4;
}
}
}
else if (posBalance >= 43) {
if (refDCval < 92) {
if (Rs < 3125) {
return 4;
}
else if (Rs >= 3125) {
return 2;
}
}
else if (refDCval >= 92) {
if (SC < 6050) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (refDCval < 94) {
return 8;
}
else if (refDCval >= 94) {
return 4;
}
}
}
else if (SC >= 6050) {
if (Rs < 4944) {
return 1;
}
else if (Rs >= 4944) {
return 8;
}
}
}
}
}
else if (ssDCval >= 113) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (diffTSC < -2) {
if (Rs < 3363) {
return 2;
}
else if (Rs >= 3363) {
return 1;
}
}
else if (diffTSC >= -2) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1641) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (refDCval < 65) {
if (Rs < 8850) {
return 8;
}
else if (Rs >= 8850) {
return 2;
}
}
else if (refDCval >= 65) {
if (diffRsCsdiff < -5) {
if (ssDCval < 72) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
return 4;
}
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
if (ssDCval < 98) {
if (Cs < 1925) {
return 8;
}
else if (Cs >= 1925) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 3333) {
if (posBalance < 36) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 3280) {
if (SC < 3676) {
if (Cs < 1695) {
return 4;
}
else if (Cs >= 1695) {
return 8;
}
}
else if (SC >= 3676) {
if (negBalance < 16) {
if (Cs < 1664) {
if (Rs < 3160) {
return 1;
}
else if (Rs >= 3160) {
return 4;
}
}
else if (Cs >= 1664) {
return 1;
}
}
else if (negBalance >= 16) {
if (SC < 3955) {
return 8;
}
else if (SC >= 3955) {
if (Rs < 2780) {
return 1;
}
else if (Rs >= 2780) {
return 8;
}
}
}
}
}
else if (Rs >= 3280) {
return 4;
}
}
}
else if (posBalance >= 36) {
if (Rs < 1647) {
if (ssDCval < 57) {
if (Rs < 1073) {
if (Cs < 1692) {
if (Rs < 880) {
return 1;
}
else if (Rs >= 880) {
return 4;
}
}
else if (Cs >= 1692) {
return 8;
}
}
else if (Rs >= 1073) {
return 4;
}
}
else if (ssDCval >= 57) {
if (AFD < 3) {
if (Cs < 1729) {
return 8;
}
else if (Cs >= 1729) {
if (Rs < 1166) {
return 1;
}
else if (Rs >= 1166) {
return 8;
}
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (Rs >= 1647) {
if (negBalance < 19) {
if (Cs < 2083) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (Cs >= 2083) {
return 4;
}
}
else if (negBalance >= 19) {
if (refDCval < 66) {
if (Cs < 1863) {
if (ssDCval < 65) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (Rs < 2939) {
return 8;
}
else if (Rs >= 2939) {
return 8;
}
}
}
else if (ssDCval >= 65) {
if (Rs < 2278) {
return 8;
}
else if (Rs >= 2278) {
return 1;
}
}
}
else if (Cs >= 1863) {
return 8;
}
}
else if (refDCval >= 66) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
return 1;
}
}
}
}
}
}
else if (Cs >= 3333) {
return 8;
}
}
else if (ssDCval >= 67) {
if (mu_mv_mag_sq < 120) {
if (posBalance < 36) {
if (ssDCval < 92) {
if (negBalance < 6) {
return 1;
}
else if (negBalance >= 6) {
if (Cs < 2172) {
if (Rs < 2364) {
if (Rs < 2031) {
return 8;
}
else if (Rs >= 2031) {
return 1;
}
}
else if (Rs >= 2364) {
return 8;
}
}
else if (Cs >= 2172) {
if (Rs < 4138) {
return 1;
}
else if (Rs >= 4138) {
return 8;
}
}
}
}
else if (ssDCval >= 92) {
if (diffMVdiffVal < -1) {
if (gchDC < 1) {
if (Rs < 2301) {
return 1;
}
else if (Rs >= 2301) {
return 4;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 32) {
return 2;
}
else if (posBalance >= 32) {
if (Rs < 2299) {
return 2;
}
else if (Rs >= 2299) {
return 4;
}
}
}
}
}
else if (posBalance >= 36) {
if (negBalance < 9) {
if (Rs < 2599) {
if (Rs < 2338) {
return 2;
}
else if (Rs >= 2338) {
return 4;
}
}
else if (Rs >= 2599) {
return 8;
}
}
else if (negBalance >= 9) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (SC < 7107) {
if (negBalance < 33) {
if (diffRsCsdiff < -210) {
if (Rs < 3479) {
return 8;
}
else if (Rs >= 3479) {
return 4;
}
}
else if (diffRsCsdiff >= -210) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 40) {
if (Cs < 2149) {
if (Rs < 3019) {
return 8;
}
else if (Rs >= 3019) {
return 2;
}
}
else if (Cs >= 2149) {
return 8;
}
}
else if (posBalance >= 40) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1644) {
if (Rs < 2186) {
return 4;
}
else if (Rs >= 2186) {
return 8;
}
}
else if (Cs >= 1644) {
if (negBalance < 1) {
if (Rs < 2441) {
return 4;
}
else if (Rs >= 2441) {
return 8;
}
}
else if (negBalance >= 1) {
return 8;
}
}
}
}
}
else if (negBalance >= 33) {
if (Rs < 2131) {
if (Rs < 1464) {
return 1;
}
else if (Rs >= 1464) {
return 4;
}
}
else if (Rs >= 2131) {
return 8;
}
}
}
else if (SC >= 7107) {
if (SC < 7935) {
if (Rs < 3296) {
return 8;
}
else if (Rs >= 3296) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 4;
}
}
}
else if (SC >= 7935) {
if (diffRsCsdiff < -14) {
if (Rs < 4396) {
return 1;
}
else if (Rs >= 4396) {
return 8;
}
}
else if (diffRsCsdiff >= -14) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6642) {
if (SC < 8254) {
if (Rs < 3665) {
if (AFD < 10) {
if (tcor < 99) {
if (Cs < 3523) {
if (SC < 4257) {
if (SC < 2882) {
if (Rs < 1035) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 1;
}
}
else if (Rs >= 1035) {
return 4;
}
}
else if (SC >= 2882) {
if (negBalance < 62) {
if (diffRsCsdiff < -31) {
if (Cs < 1870) {
return 8;
}
else if (Cs >= 1870) {
if (Rs < 1955) {
return 2;
}
else if (Rs >= 1955) {
return 4;
}
}
}
else if (diffRsCsdiff >= -31) {
if (Rs < 2525) {
return 8;
}
else if (Rs >= 2525) {
if (Cs < 1671) {
return 4;
}
else if (Cs >= 1671) {
return 8;
}
}
}
}
else if (negBalance >= 62) {
if (Rs < 2004) {
return 8;
}
else if (Rs >= 2004) {
return 4;
}
}
}
}
else if (SC >= 4257) {
if (Cs < 1884) {
if (tcor < 97) {
if (Rs < 3100) {
return 8;
}
else if (Rs >= 3100) {
if (Rs < 3462) {
if (Rs < 3351) {
return 2;
}
else if (Rs >= 3351) {
return 4;
}
}
else if (Rs >= 3462) {
if (Rs < 3572) {
return 8;
}
else if (Rs >= 3572) {
return 8;
}
}
}
}
else if (tcor >= 97) {
if (mu_mv_mag_sq < 122) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 2611) {
return 8;
}
else if (Rs >= 2611) {
if (negBalance < 33) {
if (RsCsDiff < 12) {
if (Rs < 2970) {
if (Cs < 1741) {
if (Rs < 2782) {
return 4;
}
else if (Rs >= 2782) {
return 4;
}
}
else if (Cs >= 1741) {
return 8;
}
}
else if (Rs >= 2970) {
if (Cs < 1863) {
if (negBalance < 24) {
if (Rs < 3209) {
if (Rs < 3026) {
return 8;
}
else if (Rs >= 3026) {
if (Rs < 3108) {
return 4;
}
else if (Rs >= 3108) {
return 4;
}
}
}
else if (Rs >= 3209) {
return 8;
}
}
else if (negBalance >= 24) {
if (Rs < 3035) {
return 4;
}
else if (Rs >= 3035) {
if (Rs < 3110) {
if (Rs < 3080) {
return 4;
}
else if (Rs >= 3080) {
return 4;
}
}
else if (Rs >= 3110) {
return 4;
}
}
}
}
else if (Cs >= 1863) {
return 4;
}
}
}
else if (RsCsDiff >= 12) {
return 4;
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
}
}
else if (Cs >= 1884) {
if (Rs < 2142) {
if (ssDCval < 72) {
if (SC < 4314) {
return 8;
}
else if (SC >= 4314) {
return 4;
}
}
else if (ssDCval >= 72) {
if (SC < 4267) {
return 4;
}
else if (SC >= 4267) {
return 8;
}
}
}
else if (Rs >= 2142) {
if (diffAFD < 2) {
if (diffTSC < -1) {
if (AFD < 9) {
if (mu_mv_mag_sq < 116) {
if (Cs < 2706) {
if (Cs < 2012) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 4;
}
}
else if (Cs >= 2012) {
if (Rs < 3337) {
return 8;
}
else if (Rs >= 3337) {
return 4;
}
}
}
else if (Cs >= 2706) {
return 4;
}
}
else if (mu_mv_mag_sq >= 116) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (diffMVdiffVal < -3) {
if (Rs < 2476) {
return 4;
}
else if (Rs >= 2476) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (Rs < 3555) {
if (Cs < 1933) {
if (Cs < 1913) {
return 8;
}
else if (Cs >= 1913) {
if (Rs < 2551) {
return 8;
}
else if (Rs >= 2551) {
return 4;
}
}
}
else if (Cs >= 1933) {
if (diffMVdiffVal < 2) {
if (mu_mv_mag_sq < 120) {
if (AFD < 5) {
if (Rs < 3029) {
return 4;
}
else if (Rs >= 3029) {
return 8;
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 120) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (Rs < 3203) {
return 8;
}
else if (Rs >= 3203) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 2042) {
return 8;
}
else if (Cs >= 2042) {
return 4;
}
}
}
}
else if (Rs >= 3555) {
if (Rs < 3582) {
return 4;
}
else if (Rs >= 3582) {
return 8;
}
}
}
}
}
}
else if (AFD >= 9) {
if (Cs < 2514) {
return 8;
}
else if (Cs >= 2514) {
return 4;
}
}
}
else if (diffTSC >= -1) {
if (Rs < 2864) {
if (Cs < 2452) {
if (gchDC < 2) {
if (posBalance < 50) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
if (diffRsCsdiff < 7) {
if (Cs < 1931) {
if (Rs < 2698) {
return 4;
}
else if (Rs >= 2698) {
return 8;
}
}
else if (Cs >= 1931) {
return 8;
}
}
else if (diffRsCsdiff >= 7) {
return 4;
}
}
}
else if (posBalance >= 50) {
return 8;
}
}
else if (gchDC >= 2) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (Cs >= 2452) {
if (Rs < 2456) {
return 4;
}
else if (Rs >= 2456) {
return 8;
}
}
}
else if (Rs >= 2864) {
if (Cs < 2096) {
if (ssDCval < 73) {
if (Cs < 1938) {
if (negBalance < 24) {
if (Cs < 1906) {
return 8;
}
else if (Cs >= 1906) {
return 8;
}
}
else if (negBalance >= 24) {
if (Rs < 2968) {
if (Rs < 2932) {
return 4;
}
else if (Rs >= 2932) {
return 4;
}
}
else if (Rs >= 2968) {
return 4;
}
}
}
else if (Cs >= 1938) {
return 4;
}
}
else if (ssDCval >= 73) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (Rs < 3121) {
return 8;
}
else if (Rs >= 3121) {
return 2;
}
}
}
}
else if (Cs >= 2096) {
if (CsDiff < 6) {
if (ssDCval < 81) {
if (AFD < 2) {
if (Rs < 3207) {
return 8;
}
else if (Rs >= 3207) {
return 1;
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (ssDCval >= 81) {
if (SC < 6780) {
if (Cs < 2621) {
return 8;
}
else if (Cs >= 2621) {
if (MVDiff < 2) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (SC >= 6780) {
return 4;
}
}
}
else if (CsDiff >= 6) {
if (Rs < 3268) {
if (negBalance < 52) {
if (Rs < 3154) {
return 8;
}
else if (Rs >= 3154) {
if (refDCval < 98) {
return 8;
}
else if (refDCval >= 98) {
return 4;
}
}
}
else if (negBalance >= 52) {
if (Cs < 3032) {
return 4;
}
else if (Cs >= 3032) {
return 8;
}
}
}
else if (Rs >= 3268) {
if (mu_mv_mag_sq < 126) {
if (refDCval < 97) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
return 4;
}
}
else if (refDCval >= 97) {
return 8;
}
}
else if (mu_mv_mag_sq >= 126) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 4;
}
}
}
}
}
}
}
}
else if (diffAFD >= 2) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (Rs < 2827) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (Rs >= 2827) {
if (negBalance < 33) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 3523) {
if (RsDiff < 5) {
if (ssDCval < 87) {
if (Rs < 2954) {
return 4;
}
else if (Rs >= 2954) {
return 8;
}
}
else if (ssDCval >= 87) {
return 4;
}
}
else if (RsDiff >= 5) {
if (ssDCval < 80) {
return 4;
}
else if (ssDCval >= 80) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
if (Rs < 3296) {
return 4;
}
else if (Rs >= 3296) {
return 8;
}
}
}
}
}
}
else if (tcor >= 99) {
if (Cs < 2260) {
if (SC < 5106) {
if (Rs < 2588) {
if (AFD < 6) {
if (Rs < 2471) {
return 8;
}
else if (Rs >= 2471) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (Rs >= 2588) {
if (posBalance < 40) {
if (Rs < 3000) {
return 8;
}
else if (Rs >= 3000) {
return 4;
}
}
else if (posBalance >= 40) {
if (Cs < 1947) {
return 4;
}
else if (Cs >= 1947) {
if (Rs < 2805) {
return 8;
}
else if (Rs >= 2805) {
return 4;
}
}
}
}
}
else if (SC >= 5106) {
if (negBalance < 22) {
if (Cs < 1830) {
return 4;
}
else if (Cs >= 1830) {
return 8;
}
}
else if (negBalance >= 22) {
return 2;
}
}
}
else if (Cs >= 2260) {
if (diffRsCsdiff < -16) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
return 4;
}
}
else if (diffRsCsdiff >= -16) {
return 8;
}
}
}
}
else if (AFD >= 10) {
if (Cs < 2301) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 2923) {
return 4;
}
else if (Rs >= 2923) {
return 8;
}
}
}
}
else if (Cs >= 2301) {
if (Cs < 3165) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
if (Rs < 3005) {
return 8;
}
else if (Rs >= 3005) {
return 4;
}
}
}
}
else if (Cs >= 3165) {
if (negBalance < 54) {
if (ssDCval < 93) {
if (refDCval < 89) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Rs < 1831) {
return 2;
}
else if (Rs >= 1831) {
return 4;
}
}
}
else if (refDCval >= 89) {
if (Rs < 3326) {
return 8;
}
else if (Rs >= 3326) {
return 4;
}
}
}
else if (ssDCval >= 93) {
return 8;
}
}
else if (negBalance >= 54) {
if (Rs < 3450) {
return 4;
}
else if (Rs >= 3450) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3665) {
if (refDCval < 81) {
if (refDCval < 60) {
if (Cs < 3283) {
if (diffRsCsdiff < -8) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (posBalance < 37) {
if (posBalance < 30) {
if (Cs < 2430) {
return 4;
}
else if (Cs >= 2430) {
return 4;
}
}
else if (posBalance >= 30) {
if (diffRsCsdiff < -16) {
return 8;
}
else if (diffRsCsdiff >= -16) {
if (tcor < 94) {
if (Rs < 4760) {
if (Rs < 4556) {
return 4;
}
else if (Rs >= 4556) {
return 4;
}
}
else if (Rs >= 4760) {
return 8;
}
}
else if (tcor >= 94) {
return 4;
}
}
}
}
else if (posBalance >= 37) {
if (Rs < 4586) {
return 4;
}
else if (Rs >= 4586) {
if (Rs < 4870) {
return 4;
}
else if (Rs >= 4870) {
return 4;
}
}
}
}
}
else if (diffRsCsdiff >= -8) {
if (Rs < 4116) {
return 8;
}
else if (Rs >= 4116) {
if (posBalance < 26) {
if (MVDiff < 2) {
if (Rs < 4845) {
if (Rs < 4540) {
if (Rs < 4287) {
return 4;
}
else if (Rs >= 4287) {
return 4;
}
}
else if (Rs >= 4540) {
return 4;
}
}
else if (Rs >= 4845) {
return 4;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (posBalance >= 26) {
if (ssDCval < 56) {
if (AFD < 6) {
if (posBalance < 50) {
if (Rs < 5144) {
if (diffRsCsdiff < 15) {
if (RsDiff < 4) {
if (CsDiff < 4) {
if (diffTSC < 1) {
if (Cs < 2753) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (refDCval < 54) {
return 4;
}
else if (refDCval >= 54) {
if (Rs < 4662) {
return 8;
}
else if (Rs >= 4662) {
return 8;
}
}
}
}
else if (Cs >= 2753) {
return 8;
}
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (RsDiff >= 4) {
if (tcor < 92) {
if (Rs < 4610) {
return 8;
}
else if (Rs >= 4610) {
if (Rs < 4686) {
return 8;
}
else if (Rs >= 4686) {
return 4;
}
}
}
else if (tcor >= 92) {
if (posBalance < 31) {
if (MVDiff < 3) {
if (Rs < 4746) {
if (MVDiff < 2) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (Rs >= 4746) {
return 8;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (posBalance >= 31) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Cs < 2393) {
if (Rs < 4297) {
return 4;
}
else if (Rs >= 4297) {
return 4;
}
}
else if (Cs >= 2393) {
if (posBalance < 37) {
if (Rs < 4897) {
return 4;
}
else if (Rs >= 4897) {
return 4;
}
}
else if (posBalance >= 37) {
return 4;
}
}
}
}
}
}
}
else if (diffRsCsdiff >= 15) {
if (posBalance < 40) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (Rs < 4546) {
return 8;
}
else if (Rs >= 4546) {
if (Rs < 5096) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Rs >= 5096) {
return 8;
}
}
}
}
else if (posBalance >= 40) {
return 8;
}
}
}
else if (Rs >= 5144) {
return 8;
}
}
else if (posBalance >= 50) {
return 4;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (ssDCval >= 56) {
return 4;
}
}
}
}
}
else if (Cs >= 3283) {
return 8;
}
}
else if (refDCval >= 60) {
if (SC < 5904) {
if (Rs < 4203) {
return 2;
}
else if (Rs >= 4203) {
return 4;
}
}
else if (SC >= 5904) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (Cs < 2529) {
if (posBalance < 44) {
if (negBalance < 29) {
if (Cs < 2314) {
if (ssDCval < 99) {
if (Rs < 4967) {
if (tcor < 97) {
if (Cs < 1799) {
return 4;
}
else if (Cs >= 1799) {
return 8;
}
}
else if (tcor >= 97) {
if (Rs < 4706) {
if (Cs < 2119) {
if (Cs < 1769) {
return 4;
}
else if (Cs >= 1769) {
return 2;
}
}
else if (Cs >= 2119) {
return 8;
}
}
else if (Rs >= 4706) {
return 8;
}
}
}
else if (Rs >= 4967) {
return 4;
}
}
else if (ssDCval >= 99) {
return 4;
}
}
else if (Cs >= 2314) {
return 8;
}
}
else if (negBalance >= 29) {
if (ssDCval < 91) {
if (ssDCval < 84) {
if (Rs < 5098) {
return 8;
}
else if (Rs >= 5098) {
return 4;
}
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (ssDCval >= 91) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (posBalance >= 44) {
if (AFD < 8) {
if (MVDiff < 1) {
if (Cs < 1995) {
return 2;
}
else if (Cs >= 1995) {
return 8;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (Cs >= 2529) {
if (refDCval < 98) {
if (ssDCval < 90) {
if (Cs < 3353) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Rs < 4015) {
return 4;
}
else if (Rs >= 4015) {
return 8;
}
}
}
else if (Cs >= 3353) {
return 8;
}
}
else if (ssDCval >= 90) {
if (negBalance < 33) {
if (Cs < 3208) {
if (SC < 7333) {
return 8;
}
else if (SC >= 7333) {
return 4;
}
}
else if (Cs >= 3208) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 3838) {
return 4;
}
else if (Rs >= 3838) {
return 8;
}
}
}
}
else if (negBalance >= 33) {
return 4;
}
}
}
else if (refDCval >= 98) {
return 8;
}
}
}
}
}
else if (SC >= 8254) {
if (MVDiff < 2) {
if (negBalance < 29) {
if (Rs < 4586) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (Rs >= 4586) {
return 8;
}
}
else if (negBalance >= 29) {
if (Cs < 3669) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (Cs >= 3669) {
if (mu_mv_mag_sq < 109) {
if (Rs < 5855) {
return 8;
}
else if (Rs >= 5855) {
return 4;
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (posBalance < 48) {
if (mu_mv_mag_sq < 117) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 5374) {
return 8;
}
else if (Rs >= 5374) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
else if (posBalance >= 48) {
return 8;
}
}
}
}
else if (Rs >= 6642) {
if (Rs < 8880) {
if (ssDCval < 95) {
return 4;
}
else if (ssDCval >= 95) {
return 8;
}
}
else if (Rs >= 8880) {
return 2;
}
}
}
}
}
else if (ssDCval >= 102) {
if (mcTcor < 96) {
if (ssDCval < 137) {
if (refDCval < 105) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (Cs < 3066) {
if (Rs < 4923) {
if (Rs < 3017) {
return 8;
}
else if (Rs >= 3017) {
return 1;
}
}
else if (Rs >= 4923) {
return 2;
}
}
else if (Cs >= 3066) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (posBalance < 63) {
if (refDCval < 121) {
if (negBalance < 55) {
if (diffTSC < 3) {
if (Rs < 1488) {
return 2;
}
else if (Rs >= 1488) {
return 8;
}
}
else if (diffTSC >= 3) {
if (refDCval < 108) {
if (Rs < 4261) {
return 4;
}
else if (Rs >= 4261) {
return 2;
}
}
else if (refDCval >= 108) {
return 8;
}
}
}
else if (negBalance >= 55) {
if (Rs < 4307) {
if (Rs < 2970) {
if (Rs < 1326) {
return 4;
}
else if (Rs >= 1326) {
return 8;
}
}
else if (Rs >= 2970) {
return 2;
}
}
else if (Rs >= 4307) {
return 8;
}
}
}
else if (refDCval >= 121) {
if (Rs < 4077) {
if (AFD < 8) {
if (diffTSC < 3) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Cs < 3183) {
return 8;
}
else if (Cs >= 3183) {
return 1;
}
}
}
}
else if (diffTSC >= 3) {
return 1;
}
}
else if (AFD >= 8) {
if (Rs < 3382) {
return 1;
}
else if (Rs >= 3382) {
if (Rs < 3889) {
return 4;
}
else if (Rs >= 3889) {
return 2;
}
}
}
}
else if (Rs >= 4077) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Rs < 2307) {
return 8;
}
else if (Rs >= 2307) {
return 1;
}
}
}
}
}
else if (ssDCval >= 137) {
if (posBalance < 52) {
if (ssDCval < 200) {
return 8;
}
else if (ssDCval >= 200) {
return 4;
}
}
else if (posBalance >= 52) {
if (CsDiff < 11) {
if (Rs < 2185) {
return 8;
}
else if (Rs >= 2185) {
return 1;
}
}
else if (CsDiff >= 11) {
if (Rs < 8232) {
return 8;
}
else if (Rs >= 8232) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1909) {
if (ssDCval < 115) {
if (posBalance < 54) {
if (Cs < 2356) {
return 8;
}
else if (Cs >= 2356) {
return 2;
}
}
else if (posBalance >= 54) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
if (Rs < 1662) {
return 2;
}
else if (Rs >= 1662) {
return 8;
}
}
}
}
else if (ssDCval >= 115) {
if (Rs < 926) {
return 1;
}
else if (Rs >= 926) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (mcTcor < 98) {
if (ssDCval < 133) {
return 4;
}
else if (ssDCval >= 133) {
return 8;
}
}
else if (mcTcor >= 98) {
return 8;
}
}
}
}
}
else if (Rs >= 1909) {
if (ssDCval < 117) {
if (mcTcor < 98) {
if (SC < 6617) {
if (Rs < 3854) {
if (diffRsCsdiff < 18) {
return 8;
}
else if (diffRsCsdiff >= 18) {
if (RsCsDiff < 37) {
return 4;
}
else if (RsCsDiff >= 37) {
return 8;
}
}
}
else if (Rs >= 3854) {
if (Rs < 3969) {
return 4;
}
else if (Rs >= 3969) {
return 8;
}
}
}
else if (SC >= 6617) {
if (Cs < 2405) {
if (Rs < 4815) {
return 1;
}
else if (Rs >= 4815) {
return 2;
}
}
else if (Cs >= 2405) {
if (SC < 6624) {
return 2;
}
else if (SC >= 6624) {
if (diffRsCsdiff < -170) {
if (Rs < 4350) {
return 8;
}
else if (Rs >= 4350) {
return 2;
}
}
else if (diffRsCsdiff >= -170) {
return 8;
}
}
}
}
}
else if (mcTcor >= 98) {
if (diffAFD < -26) {
return 4;
}
else if (diffAFD >= -26) {
if (diffTSC < -2) {
if (Cs < 2014) {
if (Rs < 3151) {
if (Rs < 2113) {
return 2;
}
else if (Rs >= 2113) {
return 8;
}
}
else if (Rs >= 3151) {
return 4;
}
}
else if (Cs >= 2014) {
return 8;
}
}
else if (diffTSC >= -2) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (Rs < 2491) {
return 1;
}
else if (Rs >= 2491) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 117) {
if (diffTSC < -7) {
if (Cs < 5888) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (Cs >= 5888) {
return 1;
}
}
else if (diffTSC >= -7) {
if (SC < 10185) {
if (posBalance < 35) {
if (mu_mv_mag_sq < 125) {
if (SC < 5738) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (MVDiff < 4) {
if (Cs < 2938) {
return 4;
}
else if (Cs >= 2938) {
return 8;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 128) {
return 8;
}
else if (ssDCval >= 128) {
return 4;
}
}
}
}
else if (SC >= 5738) {
if (SC < 7786) {
return 2;
}
else if (SC >= 7786) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 3891) {
if (refDCval < 138) {
if (posBalance < 2) {
return 1;
}
else if (posBalance >= 2) {
if (ssDCval < 136) {
return 8;
}
else if (ssDCval >= 136) {
if (Rs < 2519) {
return 8;
}
else if (Rs >= 2519) {
return 1;
}
}
}
}
else if (refDCval >= 138) {
return 8;
}
}
else if (Cs >= 3891) {
if (Cs < 4202) {
return 1;
}
else if (Cs >= 4202) {
if (Rs < 3857) {
return 4;
}
else if (Rs >= 3857) {
return 8;
}
}
}
}
}
else if (posBalance >= 35) {
if (Rs < 6423) {
if (Rs < 4557) {
return 8;
}
else if (Rs >= 4557) {
if (Cs < 1973) {
return 2;
}
else if (Cs >= 1973) {
if (ssDCval < 143) {
return 8;
}
else if (ssDCval >= 143) {
if (Cs < 3556) {
return 1;
}
else if (Cs >= 3556) {
return 8;
}
}
}
}
}
else if (Rs >= 6423) {
if (Rs < 6942) {
return 4;
}
else if (Rs >= 6942) {
if (Rs < 7269) {
return 8;
}
else if (Rs >= 7269) {
return 2;
}
}
}
}
}
else if (SC >= 10185) {
return 8;
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect6(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1313) {
if (Rs < 2326) {
if (Cs < 495) {
if (negBalance < 14) {
if (MVDiff < 7) {
if (SC < 778) {
if (ssDCval < 48) {
if (SC < 384) {
if (negBalance < 8) {
if (Rs < 2) {
if (ssDCval < 22) {
if (ssDCval < 10) {
return 4;
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (ssDCval >= 22) {
if (ssDCval < 38) {
return 4;
}
else if (ssDCval >= 38) {
return 2;
}
}
}
else if (Rs >= 2) {
if (Cs < 185) {
return 8;
}
else if (Cs >= 185) {
if (Rs < 119) {
return 1;
}
else if (Rs >= 119) {
if (Rs < 150) {
return 4;
}
else if (Rs >= 150) {
return 8;
}
}
}
}
}
else if (negBalance >= 8) {
if (Rs < 102) {
if (ssDCval < 40) {
if (ssDCval < 24) {
if (Rs < 66) {
return 8;
}
else if (Rs >= 66) {
return 1;
}
}
else if (ssDCval >= 24) {
if (posBalance < 26) {
if (ssDCval < 31) {
return 4;
}
else if (ssDCval >= 31) {
if (Rs < 97) {
return 1;
}
else if (Rs >= 97) {
return 8;
}
}
}
else if (posBalance >= 26) {
if (mu_mv_mag_sq < 107) {
if (Rs < 67) {
if (ssDCval < 25) {
return 8;
}
else if (ssDCval >= 25) {
return 1;
}
}
else if (Rs >= 67) {
return 8;
}
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 81) {
return 4;
}
else if (Rs >= 81) {
return 8;
}
}
}
}
}
else if (ssDCval >= 40) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (Rs >= 102) {
if (TSC < 1) {
if (ssDCval < 34) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
else if (ssDCval >= 34) {
if (Rs < 125) {
return 1;
}
else if (Rs >= 125) {
return 4;
}
}
}
else if (TSC >= 1) {
if (mu_mv_mag_sq < 96) {
if (Cs < 116) {
return 8;
}
else if (Cs >= 116) {
if (Rs < 113) {
return 4;
}
else if (Rs >= 113) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 96) {
return 1;
}
}
}
}
}
else if (SC >= 384) {
if (Rs < 206) {
if (mu_mv_mag_sq < 111) {
if (Rs < 153) {
return 1;
}
else if (Rs >= 153) {
if (Rs < 200) {
return 4;
}
else if (Rs >= 200) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
return 1;
}
}
}
else if (Rs >= 206) {
if (tcor < 102) {
if (negBalance < 8) {
if (Cs < 325) {
if (diffMVdiffVal < -6) {
if (Rs < 342) {
return 8;
}
else if (Rs >= 342) {
return 4;
}
}
else if (diffMVdiffVal >= -6) {
return 4;
}
}
else if (Cs >= 325) {
if (negBalance < 3) {
return 8;
}
else if (negBalance >= 3) {
return 4;
}
}
}
else if (negBalance >= 8) {
if (ssDCval < 32) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 32) {
if (ssDCval < 40) {
return 4;
}
else if (ssDCval >= 40) {
if (Cs < 189) {
return 4;
}
else if (Cs >= 189) {
return 1;
}
}
}
}
}
else if (tcor >= 102) {
return 8;
}
}
}
}
else if (ssDCval >= 48) {
if (ssDCval < 84) {
if (Cs < 7) {
return 2;
}
else if (Cs >= 7) {
if (ssDCval < 67) {
if (Cs < 281) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (tcor < 99) {
if (negBalance < 13) {
return 8;
}
else if (negBalance >= 13) {
if (Rs < 97) {
return 8;
}
else if (Rs >= 97) {
return 1;
}
}
}
else if (tcor >= 99) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 4;
}
}
}
}
else if (Cs >= 281) {
if (Rs < 252) {
return 4;
}
else if (Rs >= 252) {
if (Rs < 341) {
return 1;
}
else if (Rs >= 341) {
return 8;
}
}
}
}
else if (ssDCval >= 67) {
if (Cs < 119) {
if (Rs < 360) {
return 4;
}
else if (Rs >= 360) {
return 8;
}
}
else if (Cs >= 119) {
if (Cs < 243) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
if (Rs < 385) {
return 8;
}
else if (Rs >= 385) {
return 2;
}
}
}
else if (Cs >= 243) {
if (SC < 644) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
return 4;
}
}
else if (SC >= 644) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
if (Rs < 241) {
return 4;
}
else if (Rs >= 241) {
if (Rs < 367) {
if (Rs < 357) {
return 8;
}
else if (Rs >= 357) {
return 2;
}
}
else if (Rs >= 367) {
if (Rs < 426) {
return 4;
}
else if (Rs >= 426) {
return 1;
}
}
}
}
}
}
}
}
else if (SC >= 778) {
if (mu_mv_mag_sq < 128) {
if (Cs < 268) {
if (ssDCval < 39) {
if (Rs < 554) {
return 8;
}
else if (Rs >= 554) {
return 4;
}
}
else if (ssDCval >= 39) {
if (negBalance < 12) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (negBalance >= 12) {
if (mu_mv_mag_sq < 125) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (Rs < 1254) {
return 8;
}
else if (Rs >= 1254) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1060) {
return 8;
}
else if (Rs >= 1060) {
return 2;
}
}
}
}
}
else if (Cs >= 268) {
if (Rs < 476) {
if (RsCsDiff < 1) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (Rs < 434) {
return 1;
}
else if (Rs >= 434) {
return 4;
}
}
}
else if (RsCsDiff >= 1) {
return 1;
}
}
else if (Rs >= 476) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Cs < 328) {
if (Rs < 946) {
return 8;
}
else if (Rs >= 946) {
return 4;
}
}
else if (Cs >= 328) {
if (mu_mv_mag_sq < 127) {
if (negBalance < 10) {
if (Rs < 906) {
return 8;
}
else if (Rs >= 906) {
return 1;
}
}
else if (negBalance >= 10) {
if (ssDCval < 57) {
if (Cs < 336) {
if (Rs < 861) {
return 8;
}
else if (Rs >= 861) {
return 1;
}
}
else if (Cs >= 336) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (Rs < 519) {
return 8;
}
else if (Rs >= 519) {
return 1;
}
}
}
else if (ssDCval >= 67) {
if (Rs < 548) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (diffMVdiffVal < 1) {
return 1;
}
else if (diffMVdiffVal >= 1) {
if (Cs < 371) {
return 1;
}
else if (Cs >= 371) {
return 8;
}
}
}
}
else if (Rs >= 548) {
if (Rs < 1816) {
return 8;
}
else if (Rs >= 1816) {
if (Rs < 1975) {
return 1;
}
else if (Rs >= 1975) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 135) {
if (Cs < 317) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (ssDCval < 87) {
if (Cs < 261) {
return 2;
}
else if (Cs >= 261) {
return 1;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
}
else if (Cs >= 317) {
if (Cs < 478) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (Rs < 560) {
return 8;
}
else if (Rs >= 560) {
return 4;
}
}
}
}
}
else if (Cs >= 478) {
if (Rs < 1382) {
return 8;
}
else if (Rs >= 1382) {
if (Rs < 1402) {
return 2;
}
else if (Rs >= 1402) {
return 8;
}
}
}
}
}
else if (ssDCval >= 135) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (tcor < 93) {
if (ssDCval < 19) {
if (ssDCval < 15) {
return 4;
}
else if (ssDCval >= 15) {
if (MVDiff < 1076) {
return 8;
}
else if (MVDiff >= 1076) {
return 4;
}
}
}
else if (ssDCval >= 19) {
if (negBalance < 9) {
if (AFD < 8) {
if (posBalance < 93) {
if (mu_mv_mag_sq < 48) {
if (AFD < 4) {
if (Rs < 58) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 2;
}
}
else if (Rs >= 58) {
if (Cs < 221) {
return 1;
}
else if (Cs >= 221) {
return 2;
}
}
}
else if (AFD >= 4) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (posBalance < 56) {
if (ssDCval < 21) {
return 1;
}
else if (ssDCval >= 21) {
return 4;
}
}
else if (posBalance >= 56) {
if (negBalance < 1) {
if (posBalance < 72) {
if (Rs < 3) {
return 4;
}
else if (Rs >= 3) {
return 2;
}
}
else if (posBalance >= 72) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (Rs < 344) {
return 2;
}
else if (Rs >= 344) {
return 8;
}
}
}
}
else if (negBalance >= 1) {
return 8;
}
}
}
}
else if (posBalance >= 93) {
if (Rs < 52) {
if (ssDCval < 47) {
if (Cs < 29) {
return 8;
}
else if (Cs >= 29) {
return 2;
}
}
else if (ssDCval >= 47) {
return 2;
}
}
else if (Rs >= 52) {
if (ssDCval < 90) {
if (mu_mv_mag_sq < 84) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (diffAFD < 4) {
return 4;
}
else if (diffAFD >= 4) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 84) {
return 8;
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
else if (AFD >= 8) {
if (MVDiff < 102) {
if (Rs < 81) {
return 2;
}
else if (Rs >= 81) {
if (ssDCval < 43) {
return 4;
}
else if (ssDCval >= 43) {
return 8;
}
}
}
else if (MVDiff >= 102) {
if (mu_mv_mag_sq < 4) {
if (Rs < 358) {
return 8;
}
else if (Rs >= 358) {
return 2;
}
}
else if (mu_mv_mag_sq >= 4) {
if (mu_mv_mag_sq < 95) {
if (Cs < 308) {
return 1;
}
else if (Cs >= 308) {
if (Rs < 305) {
return 2;
}
else if (Rs >= 305) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 95) {
return 8;
}
}
}
}
}
else if (negBalance >= 9) {
if (ssDCval < 109) {
if (refDCval < 49) {
if (diffMVdiffVal < -53) {
if (Rs < 103) {
return 4;
}
else if (Rs >= 103) {
return 2;
}
}
else if (diffMVdiffVal >= -53) {
if (diffRsCsdiff < 1) {
if (tcor < 72) {
if (AFD < 5) {
if (Cs < 65) {
if (Rs < 6) {
return 8;
}
else if (Rs >= 6) {
return 1;
}
}
else if (Cs >= 65) {
return 8;
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (tcor >= 72) {
if (refDCval < 31) {
return 1;
}
else if (refDCval >= 31) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 1) {
if (AFD < 3) {
if (Rs < 200) {
return 1;
}
else if (Rs >= 200) {
return 4;
}
}
else if (AFD >= 3) {
if (mu_mv_mag_sq < 19) {
if (Rs < 438) {
return 2;
}
else if (Rs >= 438) {
return 1;
}
}
else if (mu_mv_mag_sq >= 19) {
return 1;
}
}
}
}
}
else if (refDCval >= 49) {
if (diffRsCsdiff < 7) {
if (ssDCval < 92) {
if (posBalance < 110) {
return 8;
}
else if (posBalance >= 110) {
return 4;
}
}
else if (ssDCval >= 92) {
return 1;
}
}
else if (diffRsCsdiff >= 7) {
if (Cs < 148) {
return 2;
}
else if (Cs >= 148) {
return 1;
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 172) {
return 8;
}
else if (Rs >= 172) {
return 2;
}
}
}
}
}
else if (tcor >= 93) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
if (posBalance < 15) {
return 8;
}
else if (posBalance >= 15) {
if (Rs < 16) {
return 8;
}
else if (Rs >= 16) {
return 1;
}
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (ssDCval < 31) {
if (MVDiff < 18) {
if (Rs < 162) {
if (negBalance < 9) {
if (ssDCval < 26) {
if (posBalance < 24) {
return 4;
}
else if (posBalance >= 24) {
return 1;
}
}
else if (ssDCval >= 26) {
return 8;
}
}
else if (negBalance >= 9) {
return 1;
}
}
else if (Rs >= 162) {
return 4;
}
}
else if (MVDiff >= 18) {
if (posBalance < 50) {
if (mu_mv_mag_sq < 72) {
return 4;
}
else if (mu_mv_mag_sq >= 72) {
if (Rs < 31) {
return 4;
}
else if (Rs >= 31) {
return 1;
}
}
}
else if (posBalance >= 50) {
if (Rs < 7) {
return 8;
}
else if (Rs >= 7) {
if (Rs < 64) {
return 1;
}
else if (Rs >= 64) {
return 4;
}
}
}
}
}
else if (ssDCval >= 31) {
if (SC < 937) {
if (negBalance < 6) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (Rs < 21) {
if (MVDiff < 445) {
return 2;
}
else if (MVDiff >= 445) {
return 1;
}
}
else if (Rs >= 21) {
if (Cs < 294) {
if (gchDC < 5) {
if (Rs < 244) {
if (mcTcor < 99) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (mcTcor >= 99) {
return 8;
}
}
else if (Rs >= 244) {
return 4;
}
}
else if (gchDC >= 5) {
if (Rs < 129) {
return 2;
}
else if (Rs >= 129) {
return 8;
}
}
}
else if (Cs >= 294) {
if (MVDiff < 59) {
if (SC < 871) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 8;
}
}
else if (SC >= 871) {
if (posBalance < 15) {
return 4;
}
else if (posBalance >= 15) {
return 8;
}
}
}
else if (MVDiff >= 59) {
if (Cs < 321) {
return 1;
}
else if (Cs >= 321) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 6) {
if (ssDCval < 88) {
if (Rs < 276) {
if (MVDiff < 45) {
if (tcor < 97) {
if (Rs < 133) {
return 8;
}
else if (Rs >= 133) {
return 1;
}
}
else if (tcor >= 97) {
if (mu_mv_mag_sq < 113) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (mu_mv_mag_sq < 98) {
return 1;
}
else if (mu_mv_mag_sq >= 98) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
}
else if (MVDiff >= 45) {
return 4;
}
}
else if (Rs >= 276) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 105) {
return 4;
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 461) {
return 8;
}
else if (Rs >= 461) {
if (Rs < 506) {
return 4;
}
else if (Rs >= 506) {
return 1;
}
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (ssDCval >= 88) {
if (ssDCval < 124) {
return 4;
}
else if (ssDCval >= 124) {
if (mu_mv_mag_sq < 91) {
if (MVDiff < 8) {
if (Rs < 448) {
return 4;
}
else if (Rs >= 448) {
if (Rs < 513) {
return 1;
}
else if (Rs >= 513) {
return 8;
}
}
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (mu_mv_mag_sq >= 91) {
if (Rs < 327) {
return 8;
}
else if (Rs >= 327) {
return 4;
}
}
}
}
}
}
else if (SC >= 937) {
if (diffAFD < -8) {
if (ssDCval < 144) {
return 2;
}
else if (ssDCval >= 144) {
if (Rs < 712) {
return 8;
}
else if (Rs >= 712) {
return 4;
}
}
}
else if (diffAFD >= -8) {
if (Rs < 802) {
if (Cs < 399) {
return 1;
}
else if (Cs >= 399) {
return 8;
}
}
else if (Rs >= 802) {
if (Cs < 47) {
return 8;
}
else if (Cs >= 47) {
if (posBalance < 7) {
return 8;
}
else if (posBalance >= 7) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (tcor < 65) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (Cs < 332) {
if (negBalance < 45) {
if (posBalance < 65) {
if (CsDiff < 2) {
if (refDCval < 29) {
if (diffMVdiffVal < 26) {
return 1;
}
else if (diffMVdiffVal >= 26) {
if (Rs < 66) {
return 1;
}
else if (Rs >= 66) {
return 2;
}
}
}
else if (refDCval >= 29) {
if (Rs < 109) {
if (SC < 119) {
return 1;
}
else if (SC >= 119) {
return 4;
}
}
else if (Rs >= 109) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 5) {
return 8;
}
else if (mu_mv_mag_sq >= 5) {
return 1;
}
}
}
else if (posBalance >= 65) {
if (Cs < 95) {
return 1;
}
else if (Cs >= 95) {
return 8;
}
}
}
else if (negBalance >= 45) {
if (Rs < 528) {
if (MVDiff < 377) {
return 8;
}
else if (MVDiff >= 377) {
return 4;
}
}
else if (Rs >= 528) {
return 1;
}
}
}
else if (Cs >= 332) {
if (RsCsDiff < 16) {
if (Rs < 472) {
if (Cs < 387) {
return 1;
}
else if (Cs >= 387) {
return 2;
}
}
else if (Rs >= 472) {
return 4;
}
}
else if (RsCsDiff >= 16) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (Cs < 425) {
if (Rs < 313) {
return 2;
}
else if (Rs >= 313) {
return 8;
}
}
else if (Cs >= 425) {
return 1;
}
}
}
}
}
}
else if (tcor >= 65) {
if (posBalance < 27) {
if (MVDiff < 5) {
if (posBalance < 15) {
return 8;
}
else if (posBalance >= 15) {
if (refDCval < 28) {
return 1;
}
else if (refDCval >= 28) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (posBalance < 10) {
if (mu_mv_mag_sq < 31) {
if (mu_mv_mag_sq < 21) {
if (mu_mv_mag_sq < 4) {
if (Rs < 84) {
return 8;
}
else if (Rs >= 84) {
return 1;
}
}
else if (mu_mv_mag_sq >= 4) {
if (mu_mv_mag_sq < 18) {
return 4;
}
else if (mu_mv_mag_sq >= 18) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
return 8;
}
}
else if (mu_mv_mag_sq >= 31) {
if (ssDCval < 21) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (ssDCval >= 21) {
return 4;
}
}
}
else if (posBalance >= 10) {
if (ssDCval < 31) {
if (mu_mv_mag_sq < 79) {
if (TSC < 1) {
return 8;
}
else if (TSC >= 1) {
if (MVDiff < 7) {
if (Rs < 84) {
return 4;
}
else if (Rs >= 84) {
return 1;
}
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (MVDiff < 18) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (MVDiff >= 18) {
return 4;
}
}
}
else if (ssDCval >= 31) {
if (mcTcor < 98) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (mcTcor >= 98) {
if (Rs < 37) {
return 4;
}
else if (Rs >= 37) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 27) {
if (posBalance < 45) {
if (refDCval < 30) {
if (Cs < 359) {
if (MVDiff < 58) {
if (mu_mv_mag_sq < 42) {
if (Rs < 96) {
return 1;
}
else if (Rs >= 96) {
if (gchDC < 1) {
if (Cs < 34) {
return 4;
}
else if (Cs >= 34) {
return 1;
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
if (tcor < 71) {
return 1;
}
else if (tcor >= 71) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (tcor < 74) {
if (MVDiff < 23) {
if (Rs < 287) {
return 1;
}
else if (Rs >= 287) {
return 4;
}
}
else if (MVDiff >= 23) {
if (Rs < 313) {
return 8;
}
else if (Rs >= 313) {
return 4;
}
}
}
else if (tcor >= 74) {
if (ssDCval < 28) {
return 1;
}
else if (ssDCval >= 28) {
if (MVDiff < 34) {
if (diffMVdiffVal < -10) {
if (Rs < 164) {
return 1;
}
else if (Rs >= 164) {
return 4;
}
}
else if (diffMVdiffVal >= -10) {
if (ssDCval < 30) {
return 1;
}
else if (ssDCval >= 30) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
return 4;
}
}
}
}
else if (MVDiff >= 34) {
if (Rs < 178) {
return 1;
}
else if (Rs >= 178) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 58) {
if (Rs < 68) {
return 2;
}
else if (Rs >= 68) {
return 8;
}
}
}
else if (Cs >= 359) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (Rs < 849) {
if (Rs < 291) {
return 4;
}
else if (Rs >= 291) {
return 1;
}
}
else if (Rs >= 849) {
return 8;
}
}
}
}
else if (refDCval >= 30) {
if (Cs < 260) {
if (SC < 504) {
if (mcTcor < 97) {
if (AFD < 4) {
if (Cs < 93) {
if (AFD < 3) {
if (Rs < 97) {
return 4;
}
else if (Rs >= 97) {
return 8;
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 93) {
if (mu_mv_mag_sq < 63) {
if (Rs < 205) {
return 8;
}
else if (Rs >= 205) {
return 1;
}
}
else if (mu_mv_mag_sq >= 63) {
if (Rs < 255) {
if (mu_mv_mag_sq < 90) {
if (refDCval < 31) {
if (Rs < 223) {
return 4;
}
else if (Rs >= 223) {
return 1;
}
}
else if (refDCval >= 31) {
return 8;
}
}
else if (mu_mv_mag_sq >= 90) {
return 1;
}
}
else if (Rs >= 255) {
return 8;
}
}
}
}
else if (AFD >= 4) {
if (negBalance < 30) {
if (Rs < 280) {
if (tcor < 74) {
return 1;
}
else if (tcor >= 74) {
if (Cs < 188) {
if (AFD < 6) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 188) {
return 1;
}
}
}
else if (Rs >= 280) {
return 2;
}
}
else if (negBalance >= 30) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 31) {
if (Cs < 88) {
return 1;
}
else if (Cs >= 88) {
if (Rs < 191) {
return 8;
}
else if (Rs >= 191) {
return 2;
}
}
}
else if (posBalance >= 31) {
if (mu_mv_mag_sq < 17) {
if (Rs < 76) {
return 1;
}
else if (Rs >= 76) {
return 4;
}
}
else if (mu_mv_mag_sq >= 17) {
return 1;
}
}
}
}
else if (SC >= 504) {
if (mcTcor < 91) {
if (Rs < 426) {
return 8;
}
else if (Rs >= 426) {
return 1;
}
}
else if (mcTcor >= 91) {
if (TSC < 3) {
if (Rs < 300) {
return 1;
}
else if (Rs >= 300) {
return 4;
}
}
else if (TSC >= 3) {
if (SC < 587) {
return 4;
}
else if (SC >= 587) {
return 1;
}
}
}
}
}
else if (Cs >= 260) {
if (MVDiff < 21) {
return 1;
}
else if (MVDiff >= 21) {
if (Rs < 439) {
return 4;
}
else if (Rs >= 439) {
return 8;
}
}
}
}
}
else if (posBalance >= 45) {
if (posBalance < 69) {
if (negBalance < 16) {
if (Rs < 175) {
return 1;
}
else if (Rs >= 175) {
return 8;
}
}
else if (negBalance >= 16) {
if (CsDiff < 3) {
if (MVDiff < 46) {
if (refDCval < 33) {
return 1;
}
else if (refDCval >= 33) {
if (SC < 404) {
if (Rs < 151) {
return 1;
}
else if (Rs >= 151) {
return 2;
}
}
else if (SC >= 404) {
return 1;
}
}
}
else if (MVDiff >= 46) {
if (negBalance < 29) {
if (Cs < 194) {
if (Rs < 80) {
return 1;
}
else if (Rs >= 80) {
return 8;
}
}
else if (Cs >= 194) {
return 2;
}
}
else if (negBalance >= 29) {
return 1;
}
}
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (posBalance >= 69) {
if (TSC < 2) {
return 2;
}
else if (TSC >= 2) {
if (Rs < 41) {
return 4;
}
else if (Rs >= 41) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 307) {
if (ssDCval < 49) {
if (mu_mv_mag_sq < 50) {
if (mcTcor < 90) {
if (posBalance < 50) {
if (tcor < 65) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (Cs < 25) {
return 1;
}
else if (Cs >= 25) {
return 8;
}
}
}
else if (tcor >= 65) {
if (Cs < 163) {
if (mu_mv_mag_sq < 13) {
if (Rs < 25) {
return 8;
}
else if (Rs >= 25) {
return 1;
}
}
else if (mu_mv_mag_sq >= 13) {
if (mu_mv_mag_sq < 35) {
return 2;
}
else if (mu_mv_mag_sq >= 35) {
if (Rs < 173) {
return 8;
}
else if (Rs >= 173) {
return 1;
}
}
}
}
else if (Cs >= 163) {
if (Cs < 347) {
if (Rs < 87) {
return 1;
}
else if (Rs >= 87) {
return 8;
}
}
else if (Cs >= 347) {
return 1;
}
}
}
}
else if (posBalance >= 50) {
if (tcor < 67) {
if (posBalance < 64) {
if (tcor < 2) {
return 1;
}
else if (tcor >= 2) {
if (mu_mv_mag_sq < 14) {
if (negBalance < 44) {
if (Rs < 77) {
return 1;
}
else if (Rs >= 77) {
return 4;
}
}
else if (negBalance >= 44) {
if (refDCval < 44) {
if (mcTcor < 72) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 8;
}
}
else if (mcTcor >= 72) {
return 8;
}
}
else if (refDCval >= 44) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
}
else if (posBalance >= 64) {
if (ssDCval < 41) {
if (Rs < 190) {
return 8;
}
else if (Rs >= 190) {
return 4;
}
}
else if (ssDCval >= 41) {
return 2;
}
}
}
else if (tcor >= 67) {
if (Cs < 55) {
return 1;
}
else if (Cs >= 55) {
if (posBalance < 56) {
if (TSC < 4) {
if (AFD < 5) {
if (Rs < 75) {
return 1;
}
else if (Rs >= 75) {
return 8;
}
}
else if (AFD >= 5) {
return 2;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (posBalance >= 56) {
if (diffRsCsdiff < -6) {
return 2;
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (posBalance < 8) {
if (negBalance < 110) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 2;
}
}
else if (negBalance >= 110) {
if (Rs < 22) {
return 4;
}
else if (Rs >= 22) {
return 8;
}
}
}
else if (posBalance >= 8) {
if (mcTcor < 97) {
if (Rs < 282) {
if (ssDCval < 36) {
if (Cs < 242) {
if (SC < 111) {
return 2;
}
else if (SC >= 111) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
if (Rs < 197) {
return 2;
}
else if (Rs >= 197) {
return 1;
}
}
}
}
else if (Cs >= 242) {
return 8;
}
}
else if (ssDCval >= 36) {
if (posBalance < 42) {
if (Cs < 217) {
if (Cs < 39) {
return 2;
}
else if (Cs >= 39) {
if (Rs < 246) {
return 4;
}
else if (Rs >= 246) {
return 2;
}
}
}
else if (Cs >= 217) {
if (Rs < 214) {
return 1;
}
else if (Rs >= 214) {
return 4;
}
}
}
else if (posBalance >= 42) {
if (RsCsDiff < 3) {
if (negBalance < 48) {
if (Rs < 57) {
if (Cs < 32) {
return 2;
}
else if (Cs >= 32) {
return 1;
}
}
else if (Rs >= 57) {
if (AFD < 9) {
if (Cs < 384) {
if (Cs < 107) {
if (MVDiff < 29) {
if (refDCval < 41) {
if (Rs < 64) {
return 1;
}
else if (Rs >= 64) {
return 2;
}
}
else if (refDCval >= 41) {
return 1;
}
}
else if (MVDiff >= 29) {
return 2;
}
}
else if (Cs >= 107) {
if (Rs < 108) {
return 4;
}
else if (Rs >= 108) {
if (Cs < 117) {
return 4;
}
else if (Cs >= 117) {
return 2;
}
}
}
}
else if (Cs >= 384) {
return 1;
}
}
else if (AFD >= 9) {
if (Rs < 95) {
return 4;
}
else if (Rs >= 95) {
return 1;
}
}
}
}
else if (negBalance >= 48) {
if (Rs < 33) {
if (Rs < 21) {
return 8;
}
else if (Rs >= 21) {
return 2;
}
}
else if (Rs >= 33) {
if (mcTcor < 93) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
return 2;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
}
}
else if (RsCsDiff >= 3) {
if (mcTcor < 95) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (diffRsCsdiff < -3) {
return 8;
}
else if (diffRsCsdiff >= -3) {
return 1;
}
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
}
else if (Rs >= 282) {
if (Cs < 175) {
return 1;
}
else if (Cs >= 175) {
if (Cs < 326) {
return 4;
}
else if (Cs >= 326) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (diffMVdiffVal < 11) {
if (Rs < 5) {
return 8;
}
else if (Rs >= 5) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (SC < 266) {
if (Rs < 46) {
return 1;
}
else if (Rs >= 46) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (SC >= 266) {
if (Rs < 269) {
return 1;
}
else if (Rs >= 269) {
if (Rs < 289) {
return 4;
}
else if (Rs >= 289) {
return 1;
}
}
}
}
}
}
else if (diffMVdiffVal >= 11) {
if (diffMVdiffVal < 15) {
return 4;
}
else if (diffMVdiffVal >= 15) {
if (Cs < 82) {
return 1;
}
else if (Cs >= 82) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (SC < 514) {
if (negBalance < 27) {
if (mu_mv_mag_sq < 104) {
if (posBalance < 30) {
if (refDCval < 39) {
if (mu_mv_mag_sq < 89) {
if (Rs < 165) {
return 1;
}
else if (Rs >= 165) {
if (Rs < 251) {
return 8;
}
else if (Rs >= 251) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (posBalance < 18) {
return 1;
}
else if (posBalance >= 18) {
return 4;
}
}
}
else if (refDCval >= 39) {
return 1;
}
}
else if (posBalance >= 30) {
if (mcTcor < 97) {
if (tcor < 74) {
return 1;
}
else if (tcor >= 74) {
if (CsDiff < 1) {
if (mu_mv_mag_sq < 95) {
return 1;
}
else if (mu_mv_mag_sq >= 95) {
if (Rs < 87) {
return 8;
}
else if (Rs >= 87) {
return 4;
}
}
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 96) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 186) {
return 1;
}
else if (Rs >= 186) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (SC < 370) {
return 1;
}
else if (SC >= 370) {
if (Cs < 235) {
return 4;
}
else if (Cs >= 235) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 115) {
if (ssDCval < 40) {
if (negBalance < 17) {
return 4;
}
else if (negBalance >= 17) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
return 8;
}
}
}
else if (ssDCval >= 40) {
return 8;
}
}
else if (Rs >= 115) {
if (posBalance < 31) {
if (Rs < 229) {
return 1;
}
else if (Rs >= 229) {
return 4;
}
}
else if (posBalance >= 31) {
if (tcor < 90) {
if (Rs < 192) {
return 1;
}
else if (Rs >= 192) {
if (Rs < 207) {
return 8;
}
else if (Rs >= 207) {
return 4;
}
}
}
else if (tcor >= 90) {
if (ssDCval < 35) {
return 1;
}
else if (ssDCval >= 35) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 27) {
if (gchDC < 1) {
if (MVDiff < 37) {
if (Cs < 303) {
if (Cs < 36) {
if (Rs < 146) {
return 2;
}
else if (Rs >= 146) {
return 1;
}
}
else if (Cs >= 36) {
if (negBalance < 44) {
if (Rs < 73) {
if (Rs < 52) {
return 8;
}
else if (Rs >= 52) {
return 1;
}
}
else if (Rs >= 73) {
return 8;
}
}
else if (negBalance >= 44) {
if (Rs < 133) {
return 8;
}
else if (Rs >= 133) {
return 2;
}
}
}
}
else if (Cs >= 303) {
return 1;
}
}
else if (MVDiff >= 37) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 95) {
if (Rs < 272) {
if (TSC < 3) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (mu_mv_mag_sq < 70) {
return 1;
}
else if (mu_mv_mag_sq >= 70) {
return 8;
}
}
}
else if (TSC >= 3) {
if (diffTSC < 2) {
if (diffRsCsdiff < -2) {
return 1;
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (diffTSC >= 2) {
return 1;
}
}
}
else if (Rs >= 272) {
if (tcor < 97) {
return 4;
}
else if (tcor >= 97) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (gchDC < 2) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 64) {
return 8;
}
else if (Rs >= 64) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (Rs < 34) {
return 8;
}
else if (Rs >= 34) {
return 4;
}
}
}
}
else if (negBalance >= 44) {
if (mu_mv_mag_sq < 88) {
if (tcor < 117) {
if (Rs < 183) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (AFD < 19) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 104) {
return 8;
}
else if (Rs >= 104) {
return 2;
}
}
}
}
else if (AFD >= 19) {
if (Rs < 62) {
return 2;
}
else if (Rs >= 62) {
return 4;
}
}
}
}
else if (Rs >= 183) {
if (Cs < 101) {
return 1;
}
else if (Cs >= 101) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
else if (tcor >= 117) {
return 4;
}
}
else if (mu_mv_mag_sq >= 88) {
if (mcTcor < 81) {
if (Rs < 26) {
return 2;
}
else if (Rs >= 26) {
return 8;
}
}
else if (mcTcor >= 81) {
if (Rs < 14) {
return 2;
}
else if (Rs >= 14) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 514) {
if (negBalance < 36) {
if (SC < 537) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (SC >= 537) {
return 1;
}
}
else if (negBalance >= 36) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 228) {
return 2;
}
else if (Rs >= 228) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 49) {
if (TSCindex < 2) {
if (Cs < 250) {
if (mu_mv_mag_sq < 106) {
if (TSCindex < 1) {
return 4;
}
else if (TSCindex >= 1) {
if (Cs < 140) {
if (posBalance < 38) {
if (Rs < 20) {
return 2;
}
else if (Rs >= 20) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (Rs < 117) {
return 2;
}
else if (Rs >= 117) {
return 4;
}
}
}
}
else if (posBalance >= 38) {
if (MVDiff < 13) {
if (Rs < 131) {
if (Cs < 128) {
if (negBalance < 24) {
if (Rs < 80) {
return 1;
}
else if (Rs >= 80) {
return 2;
}
}
else if (negBalance >= 24) {
return 1;
}
}
else if (Cs >= 128) {
return 4;
}
}
else if (Rs >= 131) {
return 1;
}
}
else if (MVDiff >= 13) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 1;
}
}
}
}
else if (Cs >= 140) {
if (Rs < 145) {
if (Rs < 106) {
return 4;
}
else if (Rs >= 106) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
return 2;
}
}
}
else if (Rs >= 145) {
if (ssDCval < 65) {
if (refDCval < 57) {
return 4;
}
else if (refDCval >= 57) {
if (Cs < 195) {
return 1;
}
else if (Cs >= 195) {
return 4;
}
}
}
else if (ssDCval >= 65) {
if (Rs < 275) {
return 2;
}
else if (Rs >= 275) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (ssDCval < 65) {
if (tcor < 101) {
if (ssDCval < 54) {
if (Rs < 100) {
return 8;
}
else if (Rs >= 100) {
return 1;
}
}
else if (ssDCval >= 54) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Cs < 189) {
if (diffAFD < 1) {
if (Cs < 128) {
return 8;
}
else if (Cs >= 128) {
return 1;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (Cs >= 189) {
return 4;
}
}
}
}
else if (tcor >= 101) {
if (Rs < 39) {
return 2;
}
else if (Rs >= 39) {
return 8;
}
}
}
else if (ssDCval >= 65) {
if (ssDCval < 83) {
if (SC < 221) {
return 4;
}
else if (SC >= 221) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
if (Rs < 232) {
return 2;
}
else if (Rs >= 232) {
return 8;
}
}
}
}
else if (ssDCval >= 83) {
if (ssDCval < 94) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (ssDCval >= 94) {
if (Rs < 139) {
return 4;
}
else if (Rs >= 139) {
return 8;
}
}
}
}
}
}
else if (Cs >= 250) {
if (SC < 564) {
if (mu_mv_mag_sq < 29) {
if (Rs < 90) {
return 1;
}
else if (Rs >= 90) {
return 2;
}
}
else if (mu_mv_mag_sq >= 29) {
if (diffMVdiffVal < -2) {
if (MVDiff < 5) {
if (Rs < 194) {
return 2;
}
else if (Rs >= 194) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 156) {
if (ssDCval < 68) {
if (Rs < 145) {
return 2;
}
else if (Rs >= 145) {
return 4;
}
}
else if (ssDCval >= 68) {
return 4;
}
}
else if (Rs >= 156) {
if (ssDCval < 71) {
if (refDCval < 58) {
return 1;
}
else if (refDCval >= 58) {
if (Cs < 379) {
return 4;
}
else if (Cs >= 379) {
return 1;
}
}
}
else if (ssDCval >= 71) {
if (Cs < 352) {
return 1;
}
else if (Cs >= 352) {
return 4;
}
}
}
}
}
}
else if (SC >= 564) {
if (ssDCval < 78) {
if (ssDCval < 66) {
if (negBalance < 18) {
return 1;
}
else if (negBalance >= 18) {
if (Cs < 491) {
return 4;
}
else if (Cs >= 491) {
return 1;
}
}
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (ssDCval >= 78) {
return 4;
}
}
}
}
else if (TSCindex >= 2) {
if (refDCval < 72) {
if (Cs < 215) {
if (AFD < 4) {
if (MVDiff < 2) {
if (refDCval < 64) {
if (tcor < 96) {
if (mu_mv_mag_sq < 62) {
return 1;
}
else if (mu_mv_mag_sq >= 62) {
if (MVDiff < 1) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (MVDiff >= 1) {
if (Rs < 79) {
return 2;
}
else if (Rs >= 79) {
return 1;
}
}
}
}
else if (tcor >= 96) {
if (Rs < 78) {
return 8;
}
else if (Rs >= 78) {
return 1;
}
}
}
else if (refDCval >= 64) {
if (SC < 290) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (SC >= 290) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
}
}
else if (MVDiff >= 2) {
if (posBalance < 40) {
if (ssDCval < 60) {
if (MVDiff < 5) {
if (Rs < 101) {
return 2;
}
else if (Rs >= 101) {
return 4;
}
}
else if (MVDiff >= 5) {
if (refDCval < 61) {
return 8;
}
else if (refDCval >= 61) {
return 2;
}
}
}
else if (ssDCval >= 60) {
if (mu_mv_mag_sq < 113) {
if (Rs < 241) {
return 1;
}
else if (Rs >= 241) {
return 8;
}
}
else if (mu_mv_mag_sq >= 113) {
return 2;
}
}
}
else if (posBalance >= 40) {
if (Cs < 192) {
if (Rs < 213) {
if (ssDCval < 67) {
if (negBalance < 22) {
if (mu_mv_mag_sq < 99) {
if (MVDiff < 24) {
return 1;
}
else if (MVDiff >= 24) {
return 8;
}
}
else if (mu_mv_mag_sq >= 99) {
if (Cs < 166) {
return 4;
}
else if (Cs >= 166) {
return 1;
}
}
}
else if (negBalance >= 22) {
if (Cs < 31) {
return 2;
}
else if (Cs >= 31) {
if (Cs < 188) {
if (tcor < 94) {
if (Rs < 187) {
return 1;
}
else if (Rs >= 187) {
return 4;
}
}
else if (tcor >= 94) {
return 1;
}
}
else if (Cs >= 188) {
if (Rs < 138) {
return 4;
}
else if (Rs >= 138) {
return 1;
}
}
}
}
}
else if (ssDCval >= 67) {
if (Cs < 152) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
if (Rs < 125) {
return 2;
}
else if (Rs >= 125) {
return 1;
}
}
}
}
else if (Cs >= 152) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
else if (Rs >= 213) {
if (posBalance < 51) {
if (TSC < 2) {
if (Rs < 262) {
return 4;
}
else if (Rs >= 262) {
return 2;
}
}
else if (TSC >= 2) {
return 1;
}
}
else if (posBalance >= 51) {
return 4;
}
}
}
else if (Cs >= 192) {
if (Cs < 208) {
if (ssDCval < 67) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (mu_mv_mag_sq < 86) {
return 4;
}
else if (mu_mv_mag_sq >= 86) {
return 1;
}
}
}
else if (ssDCval >= 67) {
return 1;
}
}
else if (Cs >= 208) {
return 1;
}
}
}
}
}
else if (AFD >= 4) {
if (TSC < 5) {
if (mu_mv_mag_sq < 21) {
if (Cs < 59) {
if (SC < 83) {
if (MVDiff < 47) {
return 1;
}
else if (MVDiff >= 47) {
if (Rs < 9) {
return 4;
}
else if (Rs >= 9) {
return 1;
}
}
}
else if (SC >= 83) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 2;
}
}
}
else if (Cs >= 59) {
if (negBalance < 36) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
return 1;
}
}
else if (negBalance >= 36) {
if (refDCval < 57) {
if (mcTcor < 84) {
if (Rs < 67) {
return 4;
}
else if (Rs >= 67) {
return 8;
}
}
else if (mcTcor >= 84) {
if (ssDCval < 50) {
return 2;
}
else if (ssDCval >= 50) {
if (diffMVdiffVal < 17) {
return 1;
}
else if (diffMVdiffVal >= 17) {
return 4;
}
}
}
}
else if (refDCval >= 57) {
if (tcor < 94) {
if (Rs < 6) {
return 1;
}
else if (Rs >= 6) {
if (MVDiff < 10) {
if (Rs < 182) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
return 1;
}
}
else if (Rs >= 182) {
return 4;
}
}
else if (MVDiff >= 10) {
if (diffAFD < -1) {
if (Rs < 94) {
return 4;
}
else if (Rs >= 94) {
return 1;
}
}
else if (diffAFD >= -1) {
if (negBalance < 78) {
return 4;
}
else if (negBalance >= 78) {
return 1;
}
}
}
}
}
else if (tcor >= 94) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (posBalance < 47) {
if (mu_mv_mag_sq < 41) {
return 2;
}
else if (mu_mv_mag_sq >= 41) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (AFD < 8) {
if (Cs < 100) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (Cs >= 100) {
return 8;
}
}
else if (AFD >= 8) {
if (Rs < 177) {
return 4;
}
else if (Rs >= 177) {
return 1;
}
}
}
}
}
else if (posBalance >= 47) {
if (refDCval < 65) {
if (Rs < 64) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (Rs >= 64) {
if (mu_mv_mag_sq < 31) {
if (tcor < 88) {
if (Cs < 187) {
return 4;
}
else if (Cs >= 187) {
return 1;
}
}
else if (tcor >= 88) {
if (Rs < 121) {
return 8;
}
else if (Rs >= 121) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 31) {
if (RsDiff < 2) {
if (negBalance < 52) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
return 4;
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (RsDiff >= 2) {
return 2;
}
}
}
}
else if (refDCval >= 65) {
if (AFD < 6) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (AFD >= 6) {
if (negBalance < 36) {
if (Rs < 126) {
return 1;
}
else if (Rs >= 126) {
return 8;
}
}
else if (negBalance >= 36) {
if (Rs < 65) {
return 8;
}
else if (Rs >= 65) {
return 4;
}
}
}
}
}
}
}
else if (TSC >= 5) {
if (AFD < 13) {
if (Cs < 103) {
if (Rs < 115) {
return 2;
}
else if (Rs >= 115) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (Cs >= 103) {
if (ssDCval < 60) {
return 2;
}
else if (ssDCval >= 60) {
if (Cs < 181) {
return 2;
}
else if (Cs >= 181) {
if (Rs < 182) {
return 1;
}
else if (Rs >= 182) {
return 8;
}
}
}
}
}
else if (AFD >= 13) {
if (posBalance < 62) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 31) {
if (diffMVdiffVal < 28) {
return 1;
}
else if (diffMVdiffVal >= 28) {
return 4;
}
}
else if (mu_mv_mag_sq >= 31) {
if (Cs < 165) {
return 4;
}
else if (Cs >= 165) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (diffAFD < 5) {
return 2;
}
else if (diffAFD >= 5) {
if (Rs < 116) {
return 4;
}
else if (Rs >= 116) {
return 1;
}
}
}
}
else if (posBalance >= 62) {
if (diffTSC < -1) {
if (mcTcor < 89) {
if (posBalance < 74) {
return 1;
}
else if (posBalance >= 74) {
return 8;
}
}
else if (mcTcor >= 89) {
return 8;
}
}
else if (diffTSC >= -1) {
if (Rs < 63) {
if (Rs < 51) {
return 1;
}
else if (Rs >= 51) {
return 8;
}
}
else if (Rs >= 63) {
if (negBalance < 54) {
if (posBalance < 103) {
if (diffAFD < -6) {
if (Rs < 104) {
return 2;
}
else if (Rs >= 104) {
return 1;
}
}
else if (diffAFD >= -6) {
return 1;
}
}
else if (posBalance >= 103) {
return 2;
}
}
else if (negBalance >= 54) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
if (Rs < 181) {
return 8;
}
else if (Rs >= 181) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 215) {
if (MVDiff < 4) {
if (Rs < 178) {
if (TSCindex < 3) {
if (MVDiff < 3) {
if (Cs < 338) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 149) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
return 1;
}
}
else if (Rs >= 149) {
return 4;
}
}
}
else if (Cs >= 338) {
if (Cs < 344) {
return 1;
}
else if (Cs >= 344) {
if (Rs < 168) {
return 4;
}
else if (Rs >= 168) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (negBalance < 23) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
return 4;
}
}
else if (negBalance >= 23) {
return 4;
}
}
}
else if (TSCindex >= 3) {
if (Rs < 135) {
return 4;
}
else if (Rs >= 135) {
if (negBalance < 30) {
if (Cs < 331) {
if (Rs < 142) {
if (Rs < 139) {
return 4;
}
else if (Rs >= 139) {
return 1;
}
}
else if (Rs >= 142) {
return 4;
}
}
else if (Cs >= 331) {
return 1;
}
}
else if (negBalance >= 30) {
if (mu_mv_mag_sq < 40) {
return 4;
}
else if (mu_mv_mag_sq >= 40) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 92) {
return 1;
}
else if (mu_mv_mag_sq >= 92) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (mcTcor >= 98) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 178) {
if (ssDCval < 67) {
if (MVDiff < 1) {
if (ssDCval < 63) {
if (diffMVdiffVal < -1) {
if (Cs < 437) {
return 4;
}
else if (Cs >= 437) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (ssDCval < 53) {
if (Cs < 403) {
return 4;
}
else if (Cs >= 403) {
return 1;
}
}
else if (ssDCval >= 53) {
return 1;
}
}
}
else if (ssDCval >= 63) {
return 4;
}
}
else if (MVDiff >= 1) {
if (ssDCval < 63) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (negBalance < 18) {
return 1;
}
else if (negBalance >= 18) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 4;
}
}
}
}
}
else if (ssDCval >= 63) {
if (negBalance < 32) {
if (Cs < 306) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
else if (Cs >= 306) {
return 4;
}
}
else if (negBalance >= 32) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (refDCval < 69) {
if (Cs < 285) {
return 1;
}
else if (Cs >= 285) {
return 4;
}
}
else if (refDCval >= 69) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 106) {
if (mu_mv_mag_sq < 80) {
return 1;
}
else if (mu_mv_mag_sq >= 80) {
return 4;
}
}
else if (mu_mv_mag_sq >= 106) {
return 1;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
}
}
}
else if (MVDiff >= 4) {
if (TSC < 4) {
if (Cs < 298) {
if (refDCval < 62) {
if (refDCval < 53) {
if (mcTcor < 97) {
if (Rs < 142) {
return 1;
}
else if (Rs >= 142) {
return 2;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (refDCval >= 53) {
if (posBalance < 38) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
if (Rs < 99) {
return 2;
}
else if (Rs >= 99) {
return 1;
}
}
}
else if (posBalance >= 38) {
if (Rs < 234) {
if (negBalance < 43) {
return 4;
}
else if (negBalance >= 43) {
return 2;
}
}
else if (Rs >= 234) {
if (Cs < 273) {
return 8;
}
else if (Cs >= 273) {
return 2;
}
}
}
}
}
else if (refDCval >= 62) {
if (AFD < 5) {
if (ssDCval < 63) {
if (mu_mv_mag_sq < 60) {
return 1;
}
else if (mu_mv_mag_sq >= 60) {
return 4;
}
}
else if (ssDCval >= 63) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
return 4;
}
}
}
else if (AFD >= 5) {
if (Rs < 85) {
return 1;
}
else if (Rs >= 85) {
if (Cs < 294) {
if (negBalance < 35) {
return 2;
}
else if (negBalance >= 35) {
return 4;
}
}
else if (Cs >= 294) {
return 1;
}
}
}
}
}
else if (Cs >= 298) {
if (AFD < 10) {
if (Rs < 263) {
if (TSC < 2) {
if (MVDiff < 9) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (Rs < 113) {
return 4;
}
else if (Rs >= 113) {
return 2;
}
}
}
else if (MVDiff >= 9) {
return 1;
}
}
else if (TSC >= 2) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 65) {
if (Rs < 192) {
if (Cs < 346) {
return 1;
}
else if (Cs >= 346) {
return 4;
}
}
else if (Rs >= 192) {
return 1;
}
}
else if (ssDCval >= 65) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
if (negBalance < 39) {
return 4;
}
else if (negBalance >= 39) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Cs < 330) {
return 4;
}
else if (Cs >= 330) {
return 8;
}
}
}
}
}
}
else if (Rs >= 263) {
if (tcor < 90) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
return 2;
}
}
else if (tcor >= 90) {
if (ssDCval < 64) {
if (Cs < 367) {
return 2;
}
else if (Cs >= 367) {
if (Rs < 269) {
return 2;
}
else if (Rs >= 269) {
return 4;
}
}
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
else if (AFD >= 10) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 4;
}
}
}
}
else if (TSC >= 4) {
if (tcor < 63) {
if (refDCval < 66) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
if (MVDiff < 86) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (MVDiff >= 86) {
return 8;
}
}
}
else if (refDCval >= 66) {
if (mcTcor < 75) {
return 4;
}
else if (mcTcor >= 75) {
return 2;
}
}
}
else if (tcor >= 63) {
if (mu_mv_mag_sq < 37) {
if (diffRsCsdiff < -22) {
return 8;
}
else if (diffRsCsdiff >= -22) {
if (diffAFD < -2) {
if (Cs < 309) {
return 2;
}
else if (Cs >= 309) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 2;
}
}
}
else if (diffAFD >= -2) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Cs < 274) {
if (Rs < 215) {
return 1;
}
else if (Rs >= 215) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
else if (Cs >= 274) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (posBalance < 43) {
if (Rs < 117) {
return 2;
}
else if (Rs >= 117) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (mu_mv_mag_sq < 54) {
return 2;
}
else if (mu_mv_mag_sq >= 54) {
if (SC < 527) {
if (Rs < 123) {
return 8;
}
else if (Rs >= 123) {
return 4;
}
}
else if (SC >= 527) {
if (Rs < 219) {
return 2;
}
else if (Rs >= 219) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 72) {
if (ssDCval < 112) {
if (Cs < 125) {
if (Rs < 89) {
if (MVDiff < 88) {
if (Cs < 99) {
if (negBalance < 69) {
if (Rs < 84) {
return 1;
}
else if (Rs >= 84) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
return 2;
}
}
}
else if (negBalance >= 69) {
if (Rs < 41) {
return 1;
}
else if (Rs >= 41) {
return 8;
}
}
}
else if (Cs >= 99) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (Rs < 76) {
return 4;
}
else if (Rs >= 76) {
if (Cs < 101) {
return 2;
}
else if (Cs >= 101) {
return 1;
}
}
}
}
}
else if (MVDiff >= 88) {
if (posBalance < 51) {
if (ssDCval < 71) {
return 4;
}
else if (ssDCval >= 71) {
return 1;
}
}
else if (posBalance >= 51) {
if (gchDC < 9) {
if (Cs < 104) {
if (MVDiff < 123) {
return 2;
}
else if (MVDiff >= 123) {
return 4;
}
}
else if (Cs >= 104) {
return 1;
}
}
else if (gchDC >= 9) {
if (MVDiff < 310) {
if (Rs < 64) {
return 8;
}
else if (Rs >= 64) {
return 4;
}
}
else if (MVDiff >= 310) {
return 1;
}
}
}
}
}
else if (Rs >= 89) {
if (MVDiff < 4) {
if (ssDCval < 88) {
if (refDCval < 74) {
if (Cs < 121) {
return 8;
}
else if (Cs >= 121) {
return 2;
}
}
else if (refDCval >= 74) {
return 8;
}
}
else if (ssDCval >= 88) {
if (Rs < 173) {
return 1;
}
else if (Rs >= 173) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (Rs < 155) {
if (AFD < 7) {
if (ssDCval < 104) {
if (refDCval < 75) {
return 2;
}
else if (refDCval >= 75) {
if (posBalance < 63) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
if (Cs < 95) {
if (Cs < 85) {
if (ssDCval < 99) {
return 1;
}
else if (ssDCval >= 99) {
return 2;
}
}
else if (Cs >= 85) {
return 2;
}
}
else if (Cs >= 95) {
return 1;
}
}
}
else if (posBalance >= 63) {
if (Cs < 87) {
return 2;
}
else if (Cs >= 87) {
return 1;
}
}
}
}
else if (ssDCval >= 104) {
return 2;
}
}
else if (AFD >= 7) {
if (mcTcor < 74) {
if (Rs < 95) {
return 8;
}
else if (Rs >= 95) {
return 4;
}
}
else if (mcTcor >= 74) {
if (posBalance < 64) {
if (ssDCval < 92) {
if (tcor < 53) {
if (diffMVdiffVal < -15) {
return 2;
}
else if (diffMVdiffVal >= -15) {
if (RsDiff < 2) {
if (MVDiff < 34) {
return 8;
}
else if (MVDiff >= 34) {
if (tcor < 25) {
return 2;
}
else if (tcor >= 25) {
if (ssDCval < 89) {
if (Cs < 29) {
return 2;
}
else if (Cs >= 29) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (ssDCval >= 89) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 138) {
return 8;
}
else if (Rs >= 138) {
return 2;
}
}
}
}
else if (tcor >= 53) {
return 8;
}
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (posBalance >= 64) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
if (Rs < 100) {
return 8;
}
else if (Rs >= 100) {
return 2;
}
}
}
}
}
}
else if (Rs >= 155) {
if (Cs < 58) {
if (tcor < 60) {
if (Rs < 219) {
if (Rs < 163) {
return 1;
}
else if (Rs >= 163) {
if (Rs < 193) {
return 4;
}
else if (Rs >= 193) {
return 8;
}
}
}
else if (Rs >= 219) {
return 2;
}
}
else if (tcor >= 60) {
if (Cs < 3) {
return 8;
}
else if (Cs >= 3) {
return 1;
}
}
}
else if (Cs >= 58) {
if (mu_mv_mag_sq < 76) {
if (RsDiff < 2) {
if (MVDiff < 10) {
if (Cs < 96) {
return 1;
}
else if (Cs >= 96) {
return 4;
}
}
else if (MVDiff >= 10) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
if (posBalance < 68) {
return 8;
}
else if (posBalance >= 68) {
return 1;
}
}
}
}
else if (RsDiff >= 2) {
if (SC < 277) {
if (SC < 252) {
return 2;
}
else if (SC >= 252) {
return 4;
}
}
else if (SC >= 277) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
return 1;
}
}
}
}
}
}
else if (Cs >= 125) {
if (refDCval < 75) {
if (Cs < 211) {
if (TSCindex < 3) {
if (mu_mv_mag_sq < 66) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (mu_mv_mag_sq >= 66) {
if (mcTcor < 98) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (mcTcor >= 98) {
if (Rs < 169) {
if (diffAFD < -1) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
return 2;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (Rs >= 169) {
if (posBalance < 46) {
if (Rs < 261) {
return 2;
}
else if (Rs >= 261) {
return 1;
}
}
else if (posBalance >= 46) {
if (Cs < 161) {
return 2;
}
else if (Cs >= 161) {
return 1;
}
}
}
}
}
}
else if (TSCindex >= 3) {
if (refDCval < 74) {
if (mu_mv_mag_sq < 111) {
if (TSC < 6) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (mu_mv_mag_sq >= 111) {
return 2;
}
}
else if (refDCval >= 74) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
if (Cs < 163) {
return 4;
}
else if (Cs >= 163) {
return 1;
}
}
}
}
}
else if (Cs >= 211) {
if (Rs < 286) {
if (mu_mv_mag_sq < 39) {
if (RsDiff < 1) {
if (Rs < 214) {
return 4;
}
else if (Rs >= 214) {
return 1;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 39) {
if (negBalance < 24) {
if (SC < 394) {
if (Rs < 140) {
return 1;
}
else if (Rs >= 140) {
return 2;
}
}
else if (SC >= 394) {
if (Rs < 99) {
return 2;
}
else if (Rs >= 99) {
return 1;
}
}
}
else if (negBalance >= 24) {
return 1;
}
}
}
else if (Rs >= 286) {
if (Rs < 300) {
return 8;
}
else if (Rs >= 300) {
return 2;
}
}
}
}
else if (refDCval >= 75) {
if (Rs < 177) {
if (tcor < 91) {
if (mu_mv_mag_sq < 26) {
if (Cs < 161) {
if (Rs < 103) {
return 2;
}
else if (Rs >= 103) {
return 1;
}
}
else if (Cs >= 161) {
if (refDCval < 107) {
if (mcTcor < 86) {
if (Rs < 139) {
if (Rs < 102) {
return 2;
}
else if (Rs >= 102) {
return 8;
}
}
else if (Rs >= 139) {
return 4;
}
}
else if (mcTcor >= 86) {
if (ssDCval < 104) {
if (MVDiff < 4) {
if (Cs < 218) {
return 8;
}
else if (Cs >= 218) {
return 2;
}
}
else if (MVDiff >= 4) {
if (tcor < 59) {
if (Rs < 139) {
return 8;
}
else if (Rs >= 139) {
return 2;
}
}
else if (tcor >= 59) {
return 2;
}
}
}
else if (ssDCval >= 104) {
return 8;
}
}
}
else if (refDCval >= 107) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (posBalance < 65) {
if (refDCval < 96) {
return 8;
}
else if (refDCval >= 96) {
if (gchDC < 7) {
return 1;
}
else if (gchDC >= 7) {
return 8;
}
}
}
else if (posBalance >= 65) {
if (ssDCval < 93) {
if (Rs < 84) {
return 1;
}
else if (Rs >= 84) {
return 2;
}
}
else if (ssDCval >= 93) {
return 1;
}
}
}
}
}
else if (tcor >= 91) {
if (Cs < 212) {
if (MVDiff < 28) {
if (Rs < 59) {
return 4;
}
else if (Rs >= 59) {
if (AFD < 6) {
if (ssDCval < 96) {
if (Rs < 174) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (Rs < 106) {
return 4;
}
else if (Rs >= 106) {
return 1;
}
}
}
}
else if (Rs >= 174) {
return 2;
}
}
else if (ssDCval >= 96) {
if (Rs < 90) {
if (Cs < 197) {
return 4;
}
else if (Cs >= 197) {
return 1;
}
}
else if (Rs >= 90) {
return 2;
}
}
}
else if (AFD >= 6) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
return 2;
}
}
}
}
else if (MVDiff >= 28) {
if (ssDCval < 97) {
return 2;
}
else if (ssDCval >= 97) {
if (Rs < 145) {
return 1;
}
else if (Rs >= 145) {
return 2;
}
}
}
}
else if (Cs >= 212) {
if (AFD < 9) {
if (posBalance < 50) {
if (Rs < 148) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (Rs >= 148) {
return 4;
}
}
else if (posBalance >= 50) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 62) {
return 4;
}
else if (mu_mv_mag_sq >= 62) {
return 1;
}
}
}
}
else if (AFD >= 9) {
return 2;
}
}
}
}
else if (Rs >= 177) {
if (refDCval < 103) {
if (mcTcor < 96) {
if (AFD < 14) {
if (Cs < 362) {
if (negBalance < 41) {
if (Cs < 281) {
return 1;
}
else if (Cs >= 281) {
if (Rs < 242) {
return 8;
}
else if (Rs >= 242) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (Rs < 283) {
if (mcTcor < 86) {
if (SC < 576) {
return 2;
}
else if (SC >= 576) {
if (Cs < 353) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (Cs >= 353) {
return 2;
}
}
}
else if (mcTcor >= 86) {
if (Cs < 289) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 181) {
return 2;
}
else if (Rs >= 181) {
return 8;
}
}
}
else if (Cs >= 289) {
if (tcor < 83) {
return 8;
}
else if (tcor >= 83) {
if (Cs < 347) {
return 2;
}
else if (Cs >= 347) {
return 8;
}
}
}
}
}
else if (Rs >= 283) {
if (Rs < 304) {
return 8;
}
else if (Rs >= 304) {
return 2;
}
}
}
}
else if (Cs >= 362) {
return 8;
}
}
else if (AFD >= 14) {
if (diffAFD < 2) {
if (refDCval < 82) {
if (SC < 464) {
return 2;
}
else if (SC >= 464) {
return 4;
}
}
else if (refDCval >= 82) {
return 2;
}
}
else if (diffAFD >= 2) {
if (diffAFD < 6) {
if (diffTSC < 1) {
if (diffMVdiffVal < 37) {
return 2;
}
else if (diffMVdiffVal >= 37) {
if (Rs < 254) {
return 1;
}
else if (Rs >= 254) {
return 4;
}
}
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (diffAFD >= 6) {
if (Rs < 209) {
return 1;
}
else if (Rs >= 209) {
return 8;
}
}
}
}
}
else if (mcTcor >= 96) {
if (Cs < 281) {
if (Rs < 261) {
if (mu_mv_mag_sq < 2) {
if (SC < 457) {
if (ssDCval < 77) {
return 1;
}
else if (ssDCval >= 77) {
if (Rs < 187) {
return 1;
}
else if (Rs >= 187) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
return 2;
}
}
}
}
else if (SC >= 457) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (mu_mv_mag_sq < 113) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (SC < 424) {
if (Rs < 191) {
return 4;
}
else if (Rs >= 191) {
return 2;
}
}
else if (SC >= 424) {
if (Rs < 241) {
return 8;
}
else if (Rs >= 241) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 220) {
return 2;
}
else if (Rs >= 220) {
return 8;
}
}
}
}
else if (Rs >= 261) {
if (ssDCval < 90) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (Cs < 186) {
return 8;
}
else if (Cs >= 186) {
return 1;
}
}
}
else if (ssDCval >= 90) {
if (Cs < 182) {
return 4;
}
else if (Cs >= 182) {
return 2;
}
}
}
}
else if (Cs >= 281) {
if (ssDCval < 89) {
if (Rs < 207) {
if (ssDCval < 79) {
if (negBalance < 45) {
if (mu_mv_mag_sq < 86) {
return 1;
}
else if (mu_mv_mag_sq >= 86) {
return 8;
}
}
else if (negBalance >= 45) {
return 4;
}
}
else if (ssDCval >= 79) {
return 4;
}
}
else if (Rs >= 207) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
}
else if (ssDCval >= 89) {
if (Rs < 240) {
if (Cs < 354) {
if (Rs < 218) {
return 1;
}
else if (Rs >= 218) {
return 8;
}
}
else if (Cs >= 354) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
}
else if (Rs >= 240) {
return 1;
}
}
}
}
}
else if (refDCval >= 103) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
if (AFD < 10) {
if (Cs < 178) {
if (Rs < 256) {
return 8;
}
else if (Rs >= 256) {
return 1;
}
}
else if (Cs >= 178) {
return 2;
}
}
else if (AFD >= 10) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (AFD < 3) {
if (ssDCval < 137) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 66) {
return 4;
}
else if (Rs >= 66) {
return 2;
}
}
}
else if (ssDCval >= 137) {
if (Rs < 236) {
return 2;
}
else if (Rs >= 236) {
return 8;
}
}
}
else if (AFD >= 3) {
if (refDCval < 158) {
if (posBalance < 33) {
if (Rs < 60) {
return 1;
}
else if (Rs >= 60) {
return 8;
}
}
else if (posBalance >= 33) {
if (SC < 526) {
if (AFD < 10) {
if (mcTcor < 85) {
if (Rs < 269) {
return 1;
}
else if (Rs >= 269) {
return 8;
}
}
else if (mcTcor >= 85) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (SC < 506) {
return 2;
}
else if (SC >= 506) {
if (Rs < 162) {
return 2;
}
else if (Rs >= 162) {
return 8;
}
}
}
}
}
else if (AFD >= 10) {
if (Cs < 15) {
return 8;
}
else if (Cs >= 15) {
if (Cs < 186) {
if (refDCval < 120) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 2;
}
}
else if (refDCval >= 120) {
if (ssDCval < 158) {
return 2;
}
else if (ssDCval >= 158) {
return 1;
}
}
}
else if (Cs >= 186) {
if (Cs < 192) {
return 4;
}
else if (Cs >= 192) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
if (SC < 480) {
return 8;
}
else if (SC >= 480) {
return 2;
}
}
}
}
}
}
}
else if (SC >= 526) {
if (Cs < 376) {
if (negBalance < 55) {
if (ssDCval < 152) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
return 2;
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (negBalance >= 55) {
return 4;
}
}
else if (Cs >= 376) {
if (ssDCval < 116) {
if (Rs < 295) {
return 8;
}
else if (Rs >= 295) {
return 1;
}
}
else if (ssDCval >= 116) {
if (mcTcor < 95) {
if (Rs < 293) {
return 2;
}
else if (Rs >= 293) {
return 8;
}
}
else if (mcTcor >= 95) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 158) {
if (Cs < 399) {
if (refDCval < 179) {
if (ssDCval < 168) {
if (Rs < 4) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 8;
}
}
else if (Rs >= 4) {
if (Rs < 236) {
return 2;
}
else if (Rs >= 236) {
return 1;
}
}
}
else if (ssDCval >= 168) {
return 1;
}
}
else if (refDCval >= 179) {
if (MVDiff < 269) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
else if (MVDiff >= 269) {
return 1;
}
}
}
else if (Cs >= 399) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 307) {
if (RsDiff < 4) {
if (ssDCval < 73) {
if (mcTcor < 87) {
if (AFD < 8) {
if (negBalance < 33) {
if (diffTSC < 1) {
if (Cs < 372) {
if (Cs < 153) {
return 2;
}
else if (Cs >= 153) {
return 8;
}
}
else if (Cs >= 372) {
return 2;
}
}
else if (diffTSC >= 1) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 8;
}
}
}
else if (negBalance >= 33) {
if (diffTSC < 1) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
return 8;
}
}
else if (diffTSC >= 1) {
if (Rs < 631) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
else if (Rs >= 631) {
return 1;
}
}
}
}
else if (AFD >= 8) {
if (Cs < 197) {
if (negBalance < 51) {
return 1;
}
else if (negBalance >= 51) {
if (mcTcor < 75) {
if (Rs < 561) {
return 1;
}
else if (Rs >= 561) {
return 2;
}
}
else if (mcTcor >= 75) {
return 8;
}
}
}
else if (Cs >= 197) {
if (refDCval < 41) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
if (Cs < 344) {
if (Cs < 278) {
return 8;
}
else if (Cs >= 278) {
return 1;
}
}
else if (Cs >= 344) {
if (Rs < 567) {
return 2;
}
else if (Rs >= 567) {
return 8;
}
}
}
}
else if (refDCval >= 41) {
if (MVDiff < 43) {
if (TSC < 6) {
if (Cs < 483) {
return 4;
}
else if (Cs >= 483) {
return 8;
}
}
else if (TSC >= 6) {
if (ssDCval < 60) {
if (Cs < 222) {
return 4;
}
else if (Cs >= 222) {
return 8;
}
}
else if (ssDCval >= 60) {
return 4;
}
}
}
else if (MVDiff >= 43) {
if (diffTSC < 2) {
if (posBalance < 56) {
if (Cs < 369) {
return 2;
}
else if (Cs >= 369) {
if (tcor < 64) {
return 8;
}
else if (tcor >= 64) {
return 1;
}
}
}
else if (posBalance >= 56) {
if (MVDiff < 64) {
return 4;
}
else if (MVDiff >= 64) {
if (diffRsCsdiff < 6) {
return 2;
}
else if (diffRsCsdiff >= 6) {
return 4;
}
}
}
}
else if (diffTSC >= 2) {
if (Rs < 431) {
return 8;
}
else if (Rs >= 431) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 87) {
if (Rs < 1272) {
if (posBalance < 26) {
if (posBalance < 2) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
else if (posBalance >= 2) {
if (Rs < 844) {
if (SC < 480) {
return 1;
}
else if (SC >= 480) {
if (SC < 846) {
if (negBalance < 37) {
if (SC < 748) {
return 2;
}
else if (SC >= 748) {
return 4;
}
}
else if (negBalance >= 37) {
if (Cs < 351) {
return 8;
}
else if (Cs >= 351) {
return 4;
}
}
}
else if (SC >= 846) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Cs < 248) {
return 8;
}
else if (Cs >= 248) {
return 1;
}
}
}
}
}
else if (Rs >= 844) {
if (Cs < 405) {
return 1;
}
else if (Cs >= 405) {
if (Rs < 963) {
return 8;
}
else if (Rs >= 963) {
return 2;
}
}
}
}
}
else if (posBalance >= 26) {
if (ssDCval < 68) {
if (mu_mv_mag_sq < 107) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Cs < 133) {
if (mu_mv_mag_sq < 54) {
if (tcor < 85) {
if (RsDiff < 3) {
if (ssDCval < 43) {
return 1;
}
else if (ssDCval >= 43) {
if (Cs < 66) {
return 1;
}
else if (Cs >= 66) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 1;
}
}
}
}
else if (RsDiff >= 3) {
if (Cs < 127) {
return 1;
}
else if (Cs >= 127) {
return 8;
}
}
}
else if (tcor >= 85) {
if (negBalance < 21) {
return 2;
}
else if (negBalance >= 21) {
if (Rs < 317) {
return 4;
}
else if (Rs >= 317) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (TSC < 3) {
if (ssDCval < 63) {
if (Cs < 128) {
return 1;
}
else if (Cs >= 128) {
return 2;
}
}
else if (ssDCval >= 63) {
return 4;
}
}
else if (TSC >= 3) {
if (Cs < 87) {
if (Rs < 412) {
return 1;
}
else if (Rs >= 412) {
return 8;
}
}
else if (Cs >= 87) {
if (Rs < 840) {
return 1;
}
else if (Rs >= 840) {
return 2;
}
}
}
}
}
else if (Cs >= 133) {
if (MVDiff < 2) {
if (negBalance < 40) {
if (Rs < 710) {
if (Rs < 483) {
if (Cs < 290) {
if (Rs < 350) {
return 1;
}
else if (Rs >= 350) {
return 4;
}
}
else if (Cs >= 290) {
return 1;
}
}
else if (Rs >= 483) {
if (negBalance < 18) {
return 8;
}
else if (negBalance >= 18) {
if (refDCval < 65) {
if (TSC < 3) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
if (Rs < 598) {
return 8;
}
else if (Rs >= 598) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
return 1;
}
}
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (refDCval >= 65) {
return 1;
}
}
}
}
else if (Rs >= 710) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 96) {
return 1;
}
else if (mu_mv_mag_sq >= 96) {
return 2;
}
}
}
}
else if (negBalance >= 40) {
if (TSC < 3) {
if (Rs < 380) {
return 4;
}
else if (Rs >= 380) {
return 1;
}
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 45) {
if (negBalance < 32) {
if (posBalance < 32) {
if (ssDCval < 34) {
if (Rs < 379) {
return 1;
}
else if (Rs >= 379) {
return 4;
}
}
else if (ssDCval >= 34) {
if (diffTSC < -1) {
if (Rs < 437) {
return 4;
}
else if (Rs >= 437) {
return 1;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
}
else if (posBalance >= 32) {
if (AFD < 3) {
if (Cs < 360) {
return 4;
}
else if (Cs >= 360) {
return 2;
}
}
else if (AFD >= 3) {
if (mu_mv_mag_sq < 68) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (tcor < 88) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (tcor >= 88) {
if (Rs < 331) {
return 1;
}
else if (Rs >= 331) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 68) {
return 1;
}
}
}
}
else if (negBalance >= 32) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Cs < 466) {
if (mu_mv_mag_sq < 3) {
if (mcTcor < 92) {
if (ssDCval < 41) {
if (Cs < 237) {
return 1;
}
else if (Cs >= 237) {
return 8;
}
}
else if (ssDCval >= 41) {
if (Rs < 658) {
return 1;
}
else if (Rs >= 658) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (tcor < 88) {
if (Rs < 516) {
return 4;
}
else if (Rs >= 516) {
if (Rs < 785) {
return 8;
}
else if (Rs >= 785) {
return 1;
}
}
}
else if (tcor >= 88) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 5) {
if (diffRsCsdiff < -1) {
return 1;
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
else if (AFD >= 5) {
if (mcTcor < 89) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Rs < 472) {
return 2;
}
else if (Rs >= 472) {
return 8;
}
}
}
else if (mcTcor >= 89) {
return 1;
}
}
}
}
else if (Cs >= 466) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
}
else if (ssDCval >= 45) {
if (Rs < 699) {
if (ssDCval < 57) {
if (negBalance < 45) {
if (TSCindex < 3) {
if (Cs < 388) {
if (RsCsDiff < 1) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (RsCsDiff >= 1) {
if (mu_mv_mag_sq < 102) {
return 4;
}
else if (mu_mv_mag_sq >= 102) {
return 1;
}
}
}
else if (Cs >= 388) {
if (Rs < 515) {
return 8;
}
else if (Rs >= 515) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (RsDiff < 3) {
if (MVDiff < 3) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (MVDiff >= 3) {
if (CsDiff < 1) {
if (posBalance < 53) {
if (tcor < 96) {
if (Cs < 230) {
return 4;
}
else if (Cs >= 230) {
if (Rs < 399) {
return 4;
}
else if (Rs >= 399) {
return 1;
}
}
}
else if (tcor >= 96) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (posBalance >= 53) {
return 1;
}
}
else if (CsDiff >= 1) {
if (SC < 730) {
if (Cs < 155) {
if (Cs < 136) {
return 4;
}
else if (Cs >= 136) {
if (Rs < 371) {
return 2;
}
else if (Rs >= 371) {
return 1;
}
}
}
else if (Cs >= 155) {
return 4;
}
}
else if (SC >= 730) {
if (SC < 968) {
if (posBalance < 50) {
if (negBalance < 35) {
if (AFD < 6) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 525) {
return 4;
}
else if (Rs >= 525) {
return 1;
}
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (negBalance >= 35) {
return 1;
}
}
else if (posBalance >= 50) {
if (Rs < 507) {
if (gchDC < 2) {
if (ssDCval < 55) {
if (ssDCval < 52) {
if (Cs < 385) {
if (Cs < 306) {
return 4;
}
else if (Cs >= 306) {
return 1;
}
}
else if (Cs >= 385) {
return 4;
}
}
else if (ssDCval >= 52) {
return 1;
}
}
else if (ssDCval >= 55) {
if (Cs < 410) {
return 4;
}
else if (Cs >= 410) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (Rs < 389) {
return 2;
}
else if (Rs >= 389) {
return 4;
}
}
}
else if (Rs >= 507) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
return 8;
}
}
}
}
else if (SC >= 968) {
return 4;
}
}
}
}
}
else if (RsDiff >= 3) {
if (Rs < 637) {
if (ssDCval < 49) {
if (Cs < 276) {
return 1;
}
else if (Cs >= 276) {
return 8;
}
}
else if (ssDCval >= 49) {
return 1;
}
}
else if (Rs >= 637) {
if (Rs < 676) {
return 2;
}
else if (Rs >= 676) {
return 8;
}
}
}
}
}
else if (negBalance >= 45) {
if (SC < 920) {
if (mcTcor < 93) {
if (Rs < 363) {
return 4;
}
else if (Rs >= 363) {
if (MVDiff < 47) {
return 8;
}
else if (MVDiff >= 47) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (Cs < 244) {
return 1;
}
else if (Cs >= 244) {
if (gchDC < 3) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 4;
}
}
else if (gchDC >= 3) {
if (Rs < 315) {
return 2;
}
else if (Rs >= 315) {
return 1;
}
}
}
}
}
else if (SC >= 920) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (negBalance < 24) {
if (Rs < 402) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
return 4;
}
}
else if (Rs >= 402) {
if (mu_mv_mag_sq < 71) {
return 4;
}
else if (mu_mv_mag_sq >= 71) {
if (mu_mv_mag_sq < 102) {
return 8;
}
else if (mu_mv_mag_sq >= 102) {
return 1;
}
}
}
}
else if (negBalance >= 24) {
if (refDCval < 62) {
if (SC < 675) {
if (SC < 534) {
return 4;
}
else if (SC >= 534) {
return 2;
}
}
else if (SC >= 675) {
if (mu_mv_mag_sq < 5) {
if (negBalance < 41) {
if (posBalance < 63) {
if (SC < 738) {
return 2;
}
else if (SC >= 738) {
return 4;
}
}
else if (posBalance >= 63) {
return 2;
}
}
else if (negBalance >= 41) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (Rs < 377) {
return 2;
}
else if (Rs >= 377) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (posBalance < 46) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (Cs < 360) {
return 8;
}
else if (Cs >= 360) {
return 2;
}
}
}
}
else if (posBalance >= 46) {
if (SC < 796) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
return 1;
}
}
else if (SC >= 796) {
if (posBalance < 57) {
if (MVDiff < 5) {
if (Rs < 629) {
return 1;
}
else if (Rs >= 629) {
return 2;
}
}
else if (MVDiff >= 5) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (posBalance >= 57) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 62) {
if (tcor < 87) {
if (negBalance < 45) {
if (Rs < 386) {
return 1;
}
else if (Rs >= 386) {
if (Rs < 517) {
return 2;
}
else if (Rs >= 517) {
return 4;
}
}
}
else if (negBalance >= 45) {
if (posBalance < 52) {
if (Rs < 355) {
return 4;
}
else if (Rs >= 355) {
return 2;
}
}
else if (posBalance >= 52) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
return 4;
}
}
}
}
else if (tcor >= 87) {
if (mu_mv_mag_sq < 69) {
if (TSC < 4) {
if (mu_mv_mag_sq < 40) {
if (posBalance < 64) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (posBalance >= 64) {
return 2;
}
}
else if (mu_mv_mag_sq >= 40) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (TSC >= 4) {
if (RsCsDiff < 4) {
if (Rs < 406) {
return 2;
}
else if (Rs >= 406) {
return 4;
}
}
else if (RsCsDiff >= 4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (Cs < 275) {
if (Rs < 446) {
return 4;
}
else if (Rs >= 446) {
return 8;
}
}
else if (Cs >= 275) {
if (mu_mv_mag_sq < 94) {
if (Rs < 593) {
return 8;
}
else if (Rs >= 593) {
if (Rs < 621) {
return 1;
}
else if (Rs >= 621) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (Rs < 639) {
return 1;
}
else if (Rs >= 639) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Rs >= 699) {
if (mcTcor < 97) {
if (ssDCval < 55) {
if (Cs < 432) {
if (TSC < 6) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (Cs >= 432) {
if (Rs < 942) {
return 8;
}
else if (Rs >= 942) {
return 2;
}
}
}
else if (ssDCval >= 55) {
if (mu_mv_mag_sq < 63) {
if (AFD < 7) {
if (SC < 1076) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (SC >= 1076) {
return 1;
}
}
else if (AFD >= 7) {
if (posBalance < 45) {
if (Rs < 1101) {
return 4;
}
else if (Rs >= 1101) {
return 1;
}
}
else if (posBalance >= 45) {
if (Cs < 342) {
return 1;
}
else if (Cs >= 342) {
if (Rs < 805) {
return 4;
}
else if (Rs >= 805) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (Rs < 775) {
return 1;
}
else if (Rs >= 775) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 55) {
if (tcor < 97) {
if (ssDCval < 55) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 1;
}
}
else if (ssDCval >= 55) {
if (diffRsCsdiff < 4) {
return 1;
}
else if (diffRsCsdiff >= 4) {
if (Cs < 269) {
return 4;
}
else if (Cs >= 269) {
return 1;
}
}
}
}
else if (tcor >= 97) {
if (Cs < 164) {
return 1;
}
else if (Cs >= 164) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (RsDiff < 3) {
if (diffMVdiffVal < -9) {
if (Rs < 937) {
return 1;
}
else if (Rs >= 937) {
return 2;
}
}
else if (diffMVdiffVal >= -9) {
return 1;
}
}
else if (RsDiff >= 3) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Cs < 336) {
if (ssDCval < 52) {
if (Rs < 344) {
return 1;
}
else if (Rs >= 344) {
return 8;
}
}
else if (ssDCval >= 52) {
if (posBalance < 41) {
if (Rs < 707) {
if (Rs < 492) {
return 4;
}
else if (Rs >= 492) {
return 8;
}
}
else if (Rs >= 707) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
}
else if (posBalance >= 41) {
if (TSCindex < 3) {
if (ssDCval < 62) {
if (SC < 1250) {
return 2;
}
else if (SC >= 1250) {
return 1;
}
}
else if (ssDCval >= 62) {
if (Rs < 591) {
return 4;
}
else if (Rs >= 591) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 2;
}
}
}
}
}
else if (Cs >= 336) {
if (mcTcor < 97) {
if (SC < 806) {
if (MVDiff < 1) {
if (Rs < 402) {
return 8;
}
else if (Rs >= 402) {
return 4;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (SC >= 806) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 45) {
if (AFD < 3) {
if (Cs < 453) {
if (ssDCval < 67) {
if (posBalance < 32) {
if (Rs < 690) {
return 8;
}
else if (Rs >= 690) {
return 1;
}
}
else if (posBalance >= 32) {
return 1;
}
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (Cs >= 453) {
if (diffAFD < -1) {
if (Rs < 578) {
return 8;
}
else if (Rs >= 578) {
return 4;
}
}
else if (diffAFD >= -1) {
if (posBalance < 43) {
if (Rs < 452) {
return 8;
}
else if (Rs >= 452) {
if (Rs < 533) {
return 1;
}
else if (Rs >= 533) {
return 8;
}
}
}
else if (posBalance >= 43) {
return 1;
}
}
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (posBalance >= 45) {
return 1;
}
}
}
}
}
else if (ssDCval >= 68) {
if (mu_mv_mag_sq < 84) {
if (SC < 1238) {
if (mu_mv_mag_sq < 9) {
if (MVDiff < 7) {
if (Cs < 330) {
return 4;
}
else if (Cs >= 330) {
if (RsCsDiff < 7) {
return 1;
}
else if (RsCsDiff >= 7) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (Cs < 365) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 4;
}
}
}
else if (Cs >= 365) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (tcor < 98) {
if (Cs < 436) {
if (diffRsCsdiff < -2) {
if (Rs < 477) {
if (Rs < 399) {
return 8;
}
else if (Rs >= 399) {
return 2;
}
}
else if (Rs >= 477) {
if (Rs < 743) {
return 1;
}
else if (Rs >= 743) {
return 4;
}
}
}
else if (diffRsCsdiff >= -2) {
if (diffAFD < 1) {
if (posBalance < 45) {
if (Rs < 633) {
return 2;
}
else if (Rs >= 633) {
return 1;
}
}
else if (posBalance >= 45) {
return 1;
}
}
else if (diffAFD >= 1) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
else if (Cs >= 436) {
if (Cs < 467) {
return 2;
}
else if (Cs >= 467) {
return 4;
}
}
}
else if (tcor >= 98) {
if (tcor < 100) {
if (Cs < 175) {
return 1;
}
else if (Cs >= 175) {
return 4;
}
}
else if (tcor >= 100) {
if (refDCval < 72) {
return 8;
}
else if (refDCval >= 72) {
return 1;
}
}
}
}
}
else if (SC >= 1238) {
if (diffMVdiffVal < 5) {
return 1;
}
else if (diffMVdiffVal >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (mu_mv_mag_sq < 112) {
if (Cs < 458) {
if (mcTcor < 97) {
if (Rs < 582) {
return 2;
}
else if (Rs >= 582) {
return 4;
}
}
else if (mcTcor >= 97) {
if (Rs < 377) {
if (Cs < 345) {
if (Rs < 347) {
return 8;
}
else if (Rs >= 347) {
return 4;
}
}
else if (Cs >= 345) {
return 2;
}
}
else if (Rs >= 377) {
if (tcor < 97) {
if (Rs < 499) {
return 1;
}
else if (Rs >= 499) {
return 8;
}
}
else if (tcor >= 97) {
if (MVDiff < 1) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
return 1;
}
}
else if (MVDiff >= 1) {
if (refDCval < 69) {
if (Rs < 596) {
return 8;
}
else if (Rs >= 596) {
return 2;
}
}
else if (refDCval >= 69) {
return 8;
}
}
}
}
}
}
else if (Cs >= 458) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (negBalance < 19) {
if (ssDCval < 70) {
if (Cs < 334) {
return 4;
}
else if (Cs >= 334) {
if (Rs < 799) {
return 1;
}
else if (Rs >= 799) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (Rs < 831) {
return 4;
}
else if (Rs >= 831) {
return 8;
}
}
}
else if (negBalance >= 19) {
if (Cs < 320) {
return 4;
}
else if (Cs >= 320) {
if (tcor < 99) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
return 2;
}
}
else if (tcor >= 99) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 40) {
if (Rs < 579) {
return 2;
}
else if (Rs >= 579) {
return 4;
}
}
else if (posBalance >= 40) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1272) {
if (SC < 1597) {
if (RsCsDiff < 2) {
if (Rs < 1292) {
return 1;
}
else if (Rs >= 1292) {
return 8;
}
}
else if (RsCsDiff >= 2) {
if (SC < 1427) {
return 2;
}
else if (SC >= 1427) {
if (Cs < 293) {
return 1;
}
else if (Cs >= 293) {
return 8;
}
}
}
}
else if (SC >= 1597) {
if (Rs < 1718) {
if (mu_mv_mag_sq < 126) {
if (mu_mv_mag_sq < 71) {
if (negBalance < 24) {
if (Cs < 387) {
if (Cs < 267) {
return 1;
}
else if (Cs >= 267) {
return 4;
}
}
else if (Cs >= 387) {
return 2;
}
}
else if (negBalance >= 24) {
if (AFD < 9) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (negBalance < 20) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 124) {
return 1;
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 1449) {
if (Rs < 1349) {
return 1;
}
else if (Rs >= 1349) {
return 2;
}
}
else if (Rs >= 1449) {
return 1;
}
}
}
else if (posBalance >= 49) {
if (Cs < 356) {
return 2;
}
else if (Cs >= 356) {
return 8;
}
}
}
else if (negBalance >= 20) {
if (tcor < 98) {
if (SC < 1981) {
return 1;
}
else if (SC >= 1981) {
return 2;
}
}
else if (tcor >= 98) {
if (TSC < 2) {
if (Rs < 1340) {
return 8;
}
else if (Rs >= 1340) {
return 1;
}
}
else if (TSC >= 2) {
if (diffRsCsdiff < -2) {
return 1;
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (ssDCval < 62) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
return 8;
}
}
else if (ssDCval >= 62) {
if (Rs < 1446) {
if (Rs < 1392) {
return 8;
}
else if (Rs >= 1392) {
return 2;
}
}
else if (Rs >= 1446) {
return 1;
}
}
}
}
else if (Rs >= 1718) {
if (Cs < 434) {
return 2;
}
else if (Cs >= 434) {
if (Cs < 445) {
return 4;
}
else if (Cs >= 445) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 73) {
if (Rs < 1618) {
if (Cs < 141) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 91) {
if (mcTcor < 99) {
if (posBalance < 51) {
if (ssDCval < 106) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (SC < 404) {
return 1;
}
else if (SC >= 404) {
if (Cs < 27) {
return 1;
}
else if (Cs >= 27) {
return 8;
}
}
}
}
else if (ssDCval >= 106) {
return 8;
}
}
else if (posBalance >= 51) {
if (Cs < 85) {
if (Cs < 33) {
return 8;
}
else if (Cs >= 33) {
return 1;
}
}
else if (Cs >= 85) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (Cs < 63) {
return 8;
}
else if (Cs >= 63) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (ssDCval < 77) {
if (Rs < 910) {
return 8;
}
else if (Rs >= 910) {
return 1;
}
}
else if (ssDCval >= 77) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (tcor < 58) {
if (Cs < 21) {
return 4;
}
else if (Cs >= 21) {
if (AFD < 17) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (tcor >= 58) {
if (ssDCval < 109) {
if (CsDiff < 2) {
if (diffAFD < 4) {
return 1;
}
else if (diffAFD >= 4) {
if (Rs < 913) {
return 2;
}
else if (Rs >= 913) {
return 1;
}
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (ssDCval >= 109) {
if (Cs < 42) {
return 8;
}
else if (Cs >= 42) {
return 1;
}
}
}
}
}
else if (Cs >= 141) {
if (ssDCval < 81) {
if (AFD < 4) {
if (MVDiff < 1) {
if (Cs < 214) {
return 8;
}
else if (Cs >= 214) {
if (Cs < 347) {
return 1;
}
else if (Cs >= 347) {
if (Rs < 474) {
return 8;
}
else if (Rs >= 474) {
if (Rs < 739) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 8;
}
}
else if (Rs >= 739) {
return 8;
}
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 29) {
if (SC < 949) {
return 8;
}
else if (SC >= 949) {
if (tcor < 98) {
if (SC < 977) {
return 4;
}
else if (SC >= 977) {
return 1;
}
}
else if (tcor >= 98) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
return 8;
}
}
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 63) {
if (Cs < 299) {
if (mu_mv_mag_sq < 29) {
if (posBalance < 59) {
if (posBalance < 51) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (posBalance >= 51) {
return 4;
}
}
else if (posBalance >= 59) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (refDCval < 77) {
return 1;
}
else if (refDCval >= 77) {
if (Cs < 251) {
if (Rs < 473) {
return 8;
}
else if (Rs >= 473) {
return 1;
}
}
else if (Cs >= 251) {
return 8;
}
}
}
}
else if (Cs >= 299) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
if (Rs < 1067) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
if (MVDiff < 58) {
return 8;
}
else if (MVDiff >= 58) {
return 1;
}
}
}
else if (Rs >= 1067) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (SC < 677) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
else if (SC >= 677) {
if (RsCsDiff < 4) {
if (Cs < 234) {
return 4;
}
else if (Cs >= 234) {
return 8;
}
}
else if (RsCsDiff >= 4) {
if (Rs < 1020) {
return 4;
}
else if (Rs >= 1020) {
if (Cs < 413) {
return 8;
}
else if (Cs >= 413) {
if (Rs < 1474) {
return 1;
}
else if (Rs >= 1474) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (mcTcor < 96) {
if (MVDiff < 46) {
if (mu_mv_mag_sq < 2) {
if (negBalance < 49) {
if (Cs < 468) {
return 1;
}
else if (Cs >= 468) {
return 2;
}
}
else if (negBalance >= 49) {
if (TSC < 6) {
if (gchDC < 1) {
if (mcTcor < 94) {
return 2;
}
else if (mcTcor >= 94) {
if (Cs < 485) {
return 1;
}
else if (Cs >= 485) {
return 2;
}
}
}
else if (gchDC >= 1) {
if (Rs < 377) {
if (Cs < 356) {
return 2;
}
else if (Cs >= 356) {
return 1;
}
}
else if (Rs >= 377) {
return 8;
}
}
}
else if (TSC >= 6) {
if (Cs < 196) {
if (Rs < 350) {
return 8;
}
else if (Rs >= 350) {
if (Rs < 462) {
return 2;
}
else if (Rs >= 462) {
return 1;
}
}
}
else if (Cs >= 196) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (AFD < 12) {
if (Cs < 317) {
if (ssDCval < 96) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 2;
}
}
else if (ssDCval >= 96) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
return 2;
}
}
}
else if (Cs >= 317) {
if (diffAFD < -3) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (diffAFD >= -3) {
if (refDCval < 88) {
if (Cs < 359) {
return 1;
}
else if (Cs >= 359) {
if (Rs < 542) {
return 8;
}
else if (Rs >= 542) {
return 1;
}
}
}
else if (refDCval >= 88) {
return 8;
}
}
}
}
else if (AFD >= 12) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (negBalance < 64) {
return 8;
}
else if (negBalance >= 64) {
return 2;
}
}
}
}
}
else if (MVDiff >= 46) {
if (mcTcor < 84) {
return 2;
}
else if (mcTcor >= 84) {
if (diffRsCsdiff < 4) {
if (MVDiff < 82) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (MVDiff >= 82) {
if (TSC < 6) {
if (Rs < 1087) {
return 8;
}
else if (Rs >= 1087) {
return 1;
}
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
if (SC < 938) {
if (mu_mv_mag_sq < 112) {
if (Cs < 210) {
if (Rs < 371) {
if (Cs < 178) {
return 1;
}
else if (Cs >= 178) {
return 4;
}
}
else if (Rs >= 371) {
return 1;
}
}
else if (Cs >= 210) {
if (Cs < 379) {
if (MVDiff < 17) {
if (ssDCval < 84) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 4;
}
}
else if (ssDCval >= 84) {
if (RsCsDiff < 4) {
if (negBalance < 16) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 8;
}
}
else if (negBalance >= 16) {
if (diffRsCsdiff < -2) {
return 1;
}
else if (diffRsCsdiff >= -2) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
else if (RsCsDiff >= 4) {
if (Rs < 449) {
return 1;
}
else if (Rs >= 449) {
if (SC < 854) {
return 8;
}
else if (SC >= 854) {
return 1;
}
}
}
}
}
else if (MVDiff >= 17) {
if (MVDiff < 97) {
if (tcor < 94) {
if (Rs < 430) {
if (Cs < 367) {
return 2;
}
else if (Cs >= 367) {
return 8;
}
}
else if (Rs >= 430) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (tcor >= 94) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (Cs < 276) {
if (Rs < 406) {
return 8;
}
else if (Rs >= 406) {
return 4;
}
}
else if (Cs >= 276) {
return 2;
}
}
}
}
else if (MVDiff >= 97) {
if (Cs < 232) {
return 2;
}
else if (Cs >= 232) {
return 4;
}
}
}
}
else if (Cs >= 379) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
if (AFD < 5) {
if (ssDCval < 113) {
return 1;
}
else if (ssDCval >= 113) {
if (ssDCval < 124) {
return 2;
}
else if (ssDCval >= 124) {
if (posBalance < 40) {
return 2;
}
else if (posBalance >= 40) {
return 4;
}
}
}
}
else if (AFD >= 5) {
if (mu_mv_mag_sq < 8) {
if (TSC < 5) {
if (Cs < 490) {
return 1;
}
else if (Cs >= 490) {
return 2;
}
}
else if (TSC >= 5) {
if (Rs < 392) {
return 4;
}
else if (Rs >= 392) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 8) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (diffAFD < 2) {
if (ssDCval < 170) {
if (negBalance < 38) {
if (ssDCval < 126) {
return 2;
}
else if (ssDCval >= 126) {
return 8;
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (ssDCval >= 170) {
return 4;
}
}
else if (diffAFD >= 2) {
if (Rs < 372) {
return 1;
}
else if (Rs >= 372) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (ssDCval < 94) {
if (SC < 512) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (SC >= 512) {
return 2;
}
}
else if (ssDCval >= 94) {
if (Rs < 451) {
return 8;
}
else if (Rs >= 451) {
return 1;
}
}
}
}
else if (SC >= 938) {
if (refDCval < 85) {
if (mu_mv_mag_sq < 120) {
if (ssDCval < 82) {
if (Rs < 773) {
if (Rs < 737) {
return 1;
}
else if (Rs >= 737) {
return 4;
}
}
else if (Rs >= 773) {
if (Cs < 268) {
return 1;
}
else if (Cs >= 268) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
else if (ssDCval >= 82) {
if (MVDiff < 29) {
return 1;
}
else if (MVDiff >= 29) {
if (Rs < 628) {
return 1;
}
else if (Rs >= 628) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (Cs < 251) {
if (Rs < 878) {
return 1;
}
else if (Rs >= 878) {
return 2;
}
}
else if (Cs >= 251) {
if (refDCval < 82) {
return 2;
}
else if (refDCval >= 82) {
return 1;
}
}
}
}
else if (refDCval >= 85) {
if (ssDCval < 115) {
if (Rs < 777) {
if (posBalance < 50) {
if (Rs < 584) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Rs >= 584) {
if (MVDiff < 4) {
if (Cs < 472) {
return 4;
}
else if (Cs >= 472) {
return 8;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (Cs < 378) {
return 8;
}
else if (Cs >= 378) {
return 1;
}
}
}
else if (Rs >= 777) {
if (refDCval < 96) {
if (Rs < 1100) {
if (Cs < 254) {
return 1;
}
else if (Cs >= 254) {
return 8;
}
}
else if (Rs >= 1100) {
if (Rs < 1524) {
return 1;
}
else if (Rs >= 1524) {
return 8;
}
}
}
else if (refDCval >= 96) {
return 8;
}
}
}
else if (ssDCval >= 115) {
if (ssDCval < 136) {
if (posBalance < 41) {
return 2;
}
else if (posBalance >= 41) {
if (Rs < 1313) {
if (ssDCval < 121) {
if (Rs < 553) {
return 2;
}
else if (Rs >= 553) {
return 8;
}
}
else if (ssDCval >= 121) {
if (MVDiff < 5) {
if (Cs < 485) {
return 8;
}
else if (Cs >= 485) {
return 1;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
}
else if (Rs >= 1313) {
return 2;
}
}
}
else if (ssDCval >= 136) {
if (Rs < 547) {
if (Cs < 459) {
return 8;
}
else if (Cs >= 459) {
return 1;
}
}
else if (Rs >= 547) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 1618) {
if (refDCval < 92) {
if (Cs < 324) {
return 1;
}
else if (Cs >= 324) {
if (posBalance < 55) {
if (RsDiff < 2) {
if (Cs < 478) {
return 1;
}
else if (Cs >= 478) {
if (Rs < 2002) {
return 4;
}
else if (Rs >= 2002) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (AFD < 5) {
if (diffRsCsdiff < -4) {
if (Rs < 1961) {
return 1;
}
else if (Rs >= 1961) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
return 8;
}
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 1;
}
}
}
}
else if (refDCval >= 92) {
if (AFD < 7) {
if (mu_mv_mag_sq < 83) {
if (diffRsCsdiff < -6) {
if (Rs < 1885) {
return 2;
}
else if (Rs >= 1885) {
return 4;
}
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 1779) {
if (Rs < 1689) {
return 8;
}
else if (Rs >= 1689) {
return 4;
}
}
else if (Rs >= 1779) {
return 1;
}
}
}
else if (AFD >= 7) {
if (mu_mv_mag_sq < 10) {
return 8;
}
else if (mu_mv_mag_sq >= 10) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (Cs < 299) {
if (mu_mv_mag_sq < 39) {
if (MVDiff < 14) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
return 2;
}
}
else if (MVDiff >= 14) {
if (posBalance < 53) {
if (Cs < 218) {
return 1;
}
else if (Cs >= 218) {
if (AFD < 20) {
if (mu_mv_mag_sq < 31) {
if (Rs < 639) {
return 2;
}
else if (Rs >= 639) {
return 1;
}
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
else if (AFD >= 20) {
if (Cs < 221) {
return 2;
}
else if (Cs >= 221) {
return 1;
}
}
}
}
else if (posBalance >= 53) {
if (negBalance < 46) {
if (tcor < 32) {
if (MVDiff < 264) {
return 8;
}
else if (MVDiff >= 264) {
return 1;
}
}
else if (tcor >= 32) {
return 1;
}
}
else if (negBalance >= 46) {
if (MVDiff < 37) {
if (tcor < 65) {
if (ssDCval < 100) {
if (mcTcor < 87) {
if (Rs < 1151) {
return 8;
}
else if (Rs >= 1151) {
return 2;
}
}
else if (mcTcor >= 87) {
if (MVDiff < 32) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (Rs < 673) {
return 1;
}
else if (Rs >= 673) {
return 2;
}
}
}
else if (MVDiff >= 32) {
return 2;
}
}
}
else if (ssDCval >= 100) {
if (Rs < 1176) {
return 1;
}
else if (Rs >= 1176) {
return 8;
}
}
}
else if (tcor >= 65) {
if (Cs < 182) {
return 1;
}
else if (Cs >= 182) {
if (RsDiff < 8) {
if (Rs < 491) {
return 8;
}
else if (Rs >= 491) {
return 1;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
else if (MVDiff >= 37) {
if (tcor < 69) {
if (ssDCval < 125) {
if (ssDCval < 61) {
if (Rs < 693) {
return 4;
}
else if (Rs >= 693) {
return 1;
}
}
else if (ssDCval >= 61) {
if (refDCval < 95) {
if (diffMVdiffVal < -36) {
if (Rs < 1341) {
return 1;
}
else if (Rs >= 1341) {
return 4;
}
}
else if (diffMVdiffVal >= -36) {
if (Cs < 277) {
if (SC < 564) {
return 1;
}
else if (SC >= 564) {
return 2;
}
}
else if (Cs >= 277) {
return 1;
}
}
}
else if (refDCval >= 95) {
if (mcTcor < 92) {
if (Cs < 120) {
return 8;
}
else if (Cs >= 120) {
return 2;
}
}
else if (mcTcor >= 92) {
return 1;
}
}
}
}
else if (ssDCval >= 125) {
if (Cs < 289) {
return 1;
}
else if (Cs >= 289) {
return 2;
}
}
}
else if (tcor >= 69) {
if (MVDiff < 216) {
if (Rs < 634) {
return 4;
}
else if (Rs >= 634) {
if (ssDCval < 136) {
return 1;
}
else if (ssDCval >= 136) {
return 4;
}
}
}
else if (MVDiff >= 216) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (tcor < 37) {
if (ssDCval < 120) {
if (AFD < 26) {
return 4;
}
else if (AFD >= 26) {
return 8;
}
}
else if (ssDCval >= 120) {
return 1;
}
}
else if (tcor >= 37) {
if (diffAFD < 9) {
if (ssDCval < 51) {
if (Rs < 522) {
if (Rs < 486) {
return 4;
}
else if (Rs >= 486) {
return 8;
}
}
else if (Rs >= 522) {
return 1;
}
}
else if (ssDCval >= 51) {
if (negBalance < 77) {
if (RsDiff < 11) {
if (mcTcor < 81) {
return 8;
}
else if (mcTcor >= 81) {
if (ssDCval < 63) {
if (Cs < 91) {
return 2;
}
else if (Cs >= 91) {
return 1;
}
}
else if (ssDCval >= 63) {
if (Rs < 1330) {
return 2;
}
else if (Rs >= 1330) {
return 1;
}
}
}
}
else if (RsDiff >= 11) {
if (diffTSC < -3) {
return 1;
}
else if (diffTSC >= -3) {
return 2;
}
}
}
else if (negBalance >= 77) {
return 1;
}
}
}
else if (diffAFD >= 9) {
return 1;
}
}
}
}
else if (Cs >= 299) {
if (gchDC < 6) {
if (RsCsDiff < 115) {
if (MVDiff < 22) {
if (posBalance < 56) {
if (diffRsCsdiff < 2) {
if (Rs < 910) {
return 8;
}
else if (Rs >= 910) {
return 1;
}
}
else if (diffRsCsdiff >= 2) {
if (Cs < 459) {
return 1;
}
else if (Cs >= 459) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
return 2;
}
}
}
}
else if (posBalance >= 56) {
if (Rs < 618) {
if (ssDCval < 48) {
return 8;
}
else if (ssDCval >= 48) {
if (tcor < 61) {
return 4;
}
else if (tcor >= 61) {
return 1;
}
}
}
else if (Rs >= 618) {
if (ssDCval < 99) {
return 8;
}
else if (ssDCval >= 99) {
if (ssDCval < 114) {
return 4;
}
else if (ssDCval >= 114) {
return 8;
}
}
}
}
}
else if (MVDiff >= 22) {
if (ssDCval < 59) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
if (RsCsDiff < 67) {
if (ssDCval < 56) {
if (refDCval < 48) {
if (negBalance < 52) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
return 8;
}
}
else if (negBalance >= 52) {
if (RsDiff < 5) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
else if (RsDiff >= 5) {
return 8;
}
}
}
else if (refDCval >= 48) {
if (mcTcor < 94) {
return 2;
}
else if (mcTcor >= 94) {
return 4;
}
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (RsCsDiff >= 67) {
if (mu_mv_mag_sq < 39) {
return 1;
}
else if (mu_mv_mag_sq >= 39) {
return 8;
}
}
}
}
else if (ssDCval >= 59) {
if (ssDCval < 79) {
if (Rs < 1152) {
if (negBalance < 55) {
if (gchDC < 5) {
if (diffAFD < -2) {
if (Rs < 527) {
return 1;
}
else if (Rs >= 527) {
if (Rs < 865) {
return 4;
}
else if (Rs >= 865) {
return 2;
}
}
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (gchDC >= 5) {
if (Rs < 573) {
return 8;
}
else if (Rs >= 573) {
return 4;
}
}
}
else if (negBalance >= 55) {
if (Rs < 874) {
return 4;
}
else if (Rs >= 874) {
if (Rs < 903) {
return 8;
}
else if (Rs >= 903) {
if (Rs < 936) {
return 1;
}
else if (Rs >= 936) {
return 4;
}
}
}
}
}
else if (Rs >= 1152) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 4;
}
}
}
else if (ssDCval >= 79) {
if (RsCsDiff < 74) {
if (tcor < 40) {
if (TSC < 6) {
if (MVDiff < 28) {
return 4;
}
else if (MVDiff >= 28) {
return 1;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (tcor >= 40) {
if (Cs < 451) {
if (tcor < 68) {
if (posBalance < 59) {
if (Rs < 533) {
if (SC < 885) {
return 2;
}
else if (SC >= 885) {
return 8;
}
}
else if (Rs >= 533) {
return 4;
}
}
else if (posBalance >= 59) {
if (Cs < 445) {
return 2;
}
else if (Cs >= 445) {
return 1;
}
}
}
else if (tcor >= 68) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (Cs >= 451) {
if (Rs < 1308) {
if (Rs < 761) {
return 2;
}
else if (Rs >= 761) {
return 8;
}
}
else if (Rs >= 1308) {
return 4;
}
}
}
}
else if (RsCsDiff >= 74) {
if (ssDCval < 96) {
if (Cs < 489) {
return 1;
}
else if (Cs >= 489) {
return 2;
}
}
else if (ssDCval >= 96) {
if (AFD < 21) {
if (Rs < 1044) {
return 1;
}
else if (Rs >= 1044) {
return 4;
}
}
else if (AFD >= 21) {
return 2;
}
}
}
}
}
}
}
else if (RsCsDiff >= 115) {
if (ssDCval < 116) {
if (posBalance < 53) {
if (diffMVdiffVal < -4) {
if (Cs < 399) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
else if (Cs >= 399) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
else if (posBalance >= 53) {
if (mcTcor < 87) {
if (RsDiff < 17) {
return 4;
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (mcTcor >= 87) {
if (ssDCval < 71) {
if (Cs < 407) {
return 4;
}
else if (Cs >= 407) {
if (Rs < 1109) {
return 8;
}
else if (Rs >= 1109) {
if (Rs < 1826) {
return 1;
}
else if (Rs >= 1826) {
return 2;
}
}
}
}
else if (ssDCval >= 71) {
if (negBalance < 63) {
return 2;
}
else if (negBalance >= 63) {
if (Rs < 1883) {
return 2;
}
else if (Rs >= 1883) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 116) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (CsDiff < 5) {
if (Cs < 319) {
return 8;
}
else if (Cs >= 319) {
return 1;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
else if (gchDC >= 6) {
if (TSC < 10) {
if (mu_mv_mag_sq < 83) {
return 1;
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 707) {
return 1;
}
else if (Rs >= 707) {
return 2;
}
}
}
else if (TSC >= 10) {
if (diffAFD < 1) {
if (ssDCval < 100) {
if (AFD < 41) {
return 2;
}
else if (AFD >= 41) {
return 8;
}
}
else if (ssDCval >= 100) {
if (AFD < 35) {
return 8;
}
else if (AFD >= 35) {
return 1;
}
}
}
else if (diffAFD >= 1) {
if (diffRsCsdiff < 1) {
if (posBalance < 81) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Rs < 807) {
return 1;
}
else if (Rs >= 807) {
return 8;
}
}
}
else if (posBalance >= 81) {
if (Cs < 462) {
return 4;
}
else if (Cs >= 462) {
return 2;
}
}
}
else if (diffRsCsdiff >= 1) {
if (AFD < 42) {
if (diffMVdiffVal < 70) {
return 1;
}
else if (diffMVdiffVal >= 70) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (AFD >= 42) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 86) {
if (RsCsDiff < 71) {
if (ssDCval < 139) {
if (Cs < 290) {
if (refDCval < 86) {
if (Rs < 1444) {
if (MVDiff < 58) {
if (Cs < 205) {
return 1;
}
else if (Cs >= 205) {
if (negBalance < 27) {
return 2;
}
else if (negBalance >= 27) {
if (ssDCval < 76) {
if (Rs < 569) {
if (Rs < 483) {
return 1;
}
else if (Rs >= 483) {
return 4;
}
}
else if (Rs >= 569) {
return 1;
}
}
else if (ssDCval >= 76) {
if (Rs < 973) {
return 4;
}
else if (Rs >= 973) {
return 1;
}
}
}
}
}
else if (MVDiff >= 58) {
if (Cs < 189) {
if (Cs < 15) {
return 2;
}
else if (Cs >= 15) {
if (mcTcor < 101) {
return 1;
}
else if (mcTcor >= 101) {
return 2;
}
}
}
else if (Cs >= 189) {
if (Rs < 888) {
return 8;
}
else if (Rs >= 888) {
return 1;
}
}
}
}
else if (Rs >= 1444) {
if (refDCval < 57) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
return 2;
}
}
else if (refDCval >= 57) {
if (Rs < 1983) {
return 1;
}
else if (Rs >= 1983) {
if (Rs < 2023) {
return 2;
}
else if (Rs >= 2023) {
return 1;
}
}
}
}
}
else if (refDCval >= 86) {
if (MVDiff < 34) {
if (Cs < 243) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (CsDiff < 3) {
if (ssDCval < 97) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (tcor < 89) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (Cs < 108) {
return 1;
}
else if (Cs >= 108) {
return 2;
}
}
}
else if (tcor >= 89) {
return 1;
}
}
}
else if (ssDCval >= 97) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
return 1;
}
}
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (Cs >= 243) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (MVDiff >= 34) {
if (SC < 801) {
if (MVDiff < 156) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
else if (MVDiff >= 156) {
if (Rs < 336) {
return 1;
}
else if (Rs >= 336) {
return 2;
}
}
}
else if (SC >= 801) {
if (diffRsCsdiff < -82) {
if (SC < 2268) {
if (Rs < 2027) {
return 1;
}
else if (Rs >= 2027) {
return 2;
}
}
else if (SC >= 2268) {
return 8;
}
}
else if (diffRsCsdiff >= -82) {
return 1;
}
}
}
}
}
else if (Cs >= 290) {
if (Rs < 836) {
if (diffAFD < -1) {
if (tcor < 82) {
if (Cs < 433) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 4;
}
}
else if (Cs >= 433) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (Rs < 599) {
return 4;
}
else if (Rs >= 599) {
return 8;
}
}
}
}
else if (tcor >= 82) {
if (gchDC < 2) {
if (mcTcor < 94) {
if (Rs < 677) {
return 8;
}
else if (Rs >= 677) {
return 4;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (gchDC >= 2) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (Rs < 570) {
return 2;
}
else if (Rs >= 570) {
return 1;
}
}
}
}
}
else if (diffAFD >= -1) {
if (mu_mv_mag_sq < 69) {
if (mu_mv_mag_sq < 21) {
if (diffRsCsdiff < 19) {
if (Cs < 318) {
return 4;
}
else if (Cs >= 318) {
if (posBalance < 45) {
if (Rs < 616) {
return 4;
}
else if (Rs >= 616) {
if (Rs < 682) {
return 1;
}
else if (Rs >= 682) {
return 8;
}
}
}
else if (posBalance >= 45) {
return 1;
}
}
}
else if (diffRsCsdiff >= 19) {
return 8;
}
}
else if (mu_mv_mag_sq >= 21) {
if (Rs < 441) {
return 1;
}
else if (Rs >= 441) {
if (Cs < 315) {
return 1;
}
else if (Cs >= 315) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 69) {
return 4;
}
}
}
else if (Rs >= 836) {
if (AFD < 7) {
if (ssDCval < 126) {
if (ssDCval < 95) {
if (negBalance < 30) {
if (RsDiff < 6) {
if (diffMVdiffVal < -3) {
if (Cs < 460) {
return 1;
}
else if (Cs >= 460) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
return 4;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (negBalance >= 30) {
if (posBalance < 37) {
if (Rs < 1416) {
return 4;
}
else if (Rs >= 1416) {
return 2;
}
}
else if (posBalance >= 37) {
if (MVDiff < 4) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 2;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
}
}
else if (ssDCval >= 95) {
if (Cs < 436) {
return 4;
}
else if (Cs >= 436) {
return 1;
}
}
}
else if (ssDCval >= 126) {
return 2;
}
}
else if (AFD >= 7) {
if (ssDCval < 66) {
if (mcTcor < 98) {
if (Rs < 1831) {
if (ssDCval < 60) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 44) {
return 2;
}
else if (mu_mv_mag_sq >= 44) {
return 1;
}
}
else if (negBalance >= 37) {
return 1;
}
}
else if (ssDCval >= 60) {
if (AFD < 11) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (AFD >= 11) {
if (Rs < 995) {
return 4;
}
else if (Rs >= 995) {
return 1;
}
}
}
}
else if (Rs >= 1831) {
if (Rs < 1937) {
return 4;
}
else if (Rs >= 1937) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (ssDCval < 60) {
if (Rs < 1556) {
return 2;
}
else if (Rs >= 1556) {
return 4;
}
}
else if (ssDCval >= 60) {
if (Rs < 1063) {
return 1;
}
else if (Rs >= 1063) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
if (refDCval < 91) {
if (mcTcor < 94) {
if (negBalance < 57) {
if (Cs < 439) {
return 1;
}
else if (Cs >= 439) {
if (diffRsCsdiff < 6) {
if (Cs < 485) {
return 4;
}
else if (Cs >= 485) {
return 1;
}
}
else if (diffRsCsdiff >= 6) {
return 1;
}
}
}
else if (negBalance >= 57) {
return 2;
}
}
else if (mcTcor >= 94) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
if (Rs < 1412) {
return 2;
}
else if (Rs >= 1412) {
return 1;
}
}
}
}
else if (refDCval >= 91) {
if (TSC < 4) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (TSC >= 4) {
if (mcTcor < 96) {
if (AFD < 12) {
if (RsDiff < 5) {
if (Cs < 380) {
return 2;
}
else if (Cs >= 380) {
return 8;
}
}
else if (RsDiff >= 5) {
if (Rs < 1925) {
return 2;
}
else if (Rs >= 1925) {
return 1;
}
}
}
else if (AFD >= 12) {
if (TSC < 9) {
if (negBalance < 55) {
return 1;
}
else if (negBalance >= 55) {
if (Rs < 1867) {
return 2;
}
else if (Rs >= 1867) {
return 1;
}
}
}
else if (TSC >= 9) {
if (Rs < 2248) {
return 8;
}
else if (Rs >= 2248) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 961) {
return 8;
}
else if (Rs >= 961) {
if (Cs < 461) {
return 1;
}
else if (Cs >= 461) {
if (Cs < 473) {
return 2;
}
else if (Cs >= 473) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 139) {
if (refDCval < 146) {
if (gchDC < 2) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
if (Rs < 1109) {
return 4;
}
else if (Rs >= 1109) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (Rs < 660) {
return 2;
}
else if (Rs >= 660) {
return 1;
}
}
}
else if (refDCval >= 146) {
if (ssDCval < 168) {
if (Rs < 779) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
else if (Rs >= 779) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
return 1;
}
}
}
else if (ssDCval >= 168) {
if (Rs < 826) {
return 8;
}
else if (Rs >= 826) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 71) {
if (ssDCval < 111) {
if (Rs < 2063) {
if (negBalance < 52) {
if (mcTcor < 99) {
if (diffAFD < 2) {
if (tcor < 90) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
if (diffMVdiffVal < -15) {
return 2;
}
else if (diffMVdiffVal >= -15) {
if (ssDCval < 66) {
return 2;
}
else if (ssDCval >= 66) {
if (diffAFD < -1) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (diffMVdiffVal < 15) {
return 2;
}
else if (diffMVdiffVal >= 15) {
return 1;
}
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
}
else if (tcor >= 90) {
return 2;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (mcTcor >= 99) {
if (Cs < 333) {
return 2;
}
else if (Cs >= 333) {
return 1;
}
}
}
else if (negBalance >= 52) {
if (MVDiff < 30) {
if (RsCsDiff < 107) {
if (mcTcor < 94) {
return 2;
}
else if (mcTcor >= 94) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (RsCsDiff >= 107) {
return 2;
}
}
else if (MVDiff >= 30) {
if (RsDiff < 13) {
if (Cs < 187) {
return 1;
}
else if (Cs >= 187) {
if (mcTcor < 96) {
if (Cs < 457) {
if (Cs < 333) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (Rs < 1685) {
return 1;
}
else if (Rs >= 1685) {
return 8;
}
}
}
}
else if (Cs >= 333) {
return 1;
}
}
else if (Cs >= 457) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Cs < 311) {
return 2;
}
else if (Cs >= 311) {
if (TSC < 14) {
if (RsCsDiff < 142) {
return 1;
}
else if (RsCsDiff >= 142) {
return 2;
}
}
else if (TSC >= 14) {
return 8;
}
}
}
}
}
else if (RsDiff >= 13) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 1;
}
}
}
}
}
else if (Rs >= 2063) {
if (refDCval < 81) {
if (MVDiff < 53) {
if (negBalance < 42) {
if (Rs < 2223) {
return 2;
}
else if (Rs >= 2223) {
return 1;
}
}
else if (negBalance >= 42) {
return 2;
}
}
else if (MVDiff >= 53) {
if (Rs < 2230) {
return 4;
}
else if (Rs >= 2230) {
return 1;
}
}
}
else if (refDCval >= 81) {
if (RsCsDiff < 128) {
if (CsDiff < 2) {
if (diffTSC < 1) {
if (Rs < 2204) {
return 2;
}
else if (Rs >= 2204) {
return 1;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (CsDiff >= 2) {
if (Rs < 2307) {
return 1;
}
else if (Rs >= 2307) {
return 2;
}
}
}
else if (RsCsDiff >= 128) {
if (mcTcor < 93) {
if (Rs < 2233) {
return 8;
}
else if (Rs >= 2233) {
return 4;
}
}
else if (mcTcor >= 93) {
if (mu_mv_mag_sq < 23) {
if (Rs < 2195) {
if (Rs < 2100) {
return 1;
}
else if (Rs >= 2100) {
return 2;
}
}
else if (Rs >= 2195) {
return 1;
}
}
else if (mu_mv_mag_sq >= 23) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 111) {
if (CsDiff < 2) {
if (mu_mv_mag_sq < 22) {
if (Rs < 2298) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 1;
}
}
else if (Rs >= 2298) {
return 8;
}
}
else if (mu_mv_mag_sq >= 22) {
if (Cs < 133) {
return 4;
}
else if (Cs >= 133) {
return 2;
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 8) {
if (diffMVdiffVal < 31) {
return 1;
}
else if (diffMVdiffVal >= 31) {
if (diffMVdiffVal < 33) {
if (MVDiff < 96) {
return 1;
}
else if (MVDiff >= 96) {
return 8;
}
}
else if (diffMVdiffVal >= 33) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 8) {
if (Cs < 331) {
return 1;
}
else if (Cs >= 331) {
if (gchDC < 1) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (gchDC >= 1) {
if (posBalance < 66) {
if (CsDiff < 3) {
if (Rs < 2204) {
return 8;
}
else if (Rs >= 2204) {
return 1;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (posBalance >= 66) {
if (Rs < 1271) {
return 2;
}
else if (Rs >= 1271) {
if (diffMVdiffVal < 12) {
return 1;
}
else if (diffMVdiffVal >= 12) {
if (MVDiff < 47) {
return 2;
}
else if (MVDiff >= 47) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (AFD < 8) {
if (Cs < 379) {
if (Rs < 1485) {
return 1;
}
else if (Rs >= 1485) {
return 4;
}
}
else if (Cs >= 379) {
return 8;
}
}
else if (AFD >= 8) {
if (Rs < 732) {
return 1;
}
else if (Rs >= 732) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Cs >= 495) {
if (negBalance < 13) {
if (Rs < 697) {
if (ssDCval < 57) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 32) {
if (Rs < 508) {
return 4;
}
else if (Rs >= 508) {
return 8;
}
}
else if (ssDCval >= 32) {
if (Rs < 245) {
if (Rs < 211) {
return 4;
}
else if (Rs >= 211) {
return 2;
}
}
else if (Rs >= 245) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (Rs < 258) {
return 2;
}
else if (Rs >= 258) {
return 1;
}
}
}
}
else if (ssDCval >= 57) {
if (RsCsDiff < 1) {
if (SC < 1366) {
if (Rs < 310) {
return 4;
}
else if (Rs >= 310) {
if (diffRsCsdiff < -1) {
if (Cs < 515) {
if (Rs < 573) {
return 1;
}
else if (Rs >= 573) {
return 4;
}
}
else if (Cs >= 515) {
if (Rs < 467) {
return 1;
}
else if (Rs >= 467) {
return 8;
}
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
else if (SC >= 1366) {
if (Cs < 796) {
if (Rs < 648) {
return 2;
}
else if (Rs >= 648) {
return 1;
}
}
else if (Cs >= 796) {
return 8;
}
}
}
else if (RsCsDiff >= 1) {
if (SC < 1058) {
if (refDCval < 150) {
if (posBalance < 125) {
if (diffRsCsdiff < -8) {
return 4;
}
else if (diffRsCsdiff >= -8) {
return 1;
}
}
else if (posBalance >= 125) {
return 8;
}
}
else if (refDCval >= 150) {
return 2;
}
}
else if (SC >= 1058) {
if (TSCindex < 2) {
if (Cs < 607) {
return 8;
}
else if (Cs >= 607) {
return 1;
}
}
else if (TSCindex >= 2) {
if (refDCval < 65) {
return 4;
}
else if (refDCval >= 65) {
return 8;
}
}
}
}
}
}
else if (Rs >= 697) {
if (diffTSC < -2) {
if (Rs < 1596) {
if (Cs < 628) {
if (Rs < 1157) {
if (diffAFD < -11) {
return 8;
}
else if (diffAFD >= -11) {
return 2;
}
}
else if (Rs >= 1157) {
if (Rs < 1355) {
return 1;
}
else if (Rs >= 1355) {
return 8;
}
}
}
else if (Cs >= 628) {
if (diffAFD < -9) {
if (ssDCval < 139) {
if (ssDCval < 133) {
if (diffAFD < -21) {
if (ssDCval < 67) {
if (Rs < 1166) {
return 8;
}
else if (Rs >= 1166) {
return 1;
}
}
else if (ssDCval >= 67) {
if (Rs < 820) {
return 1;
}
else if (Rs >= 820) {
return 4;
}
}
}
else if (diffAFD >= -21) {
if (MVDiff < 10) {
if (Rs < 1235) {
return 1;
}
else if (Rs >= 1235) {
return 8;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
else if (ssDCval >= 133) {
return 4;
}
}
else if (ssDCval >= 139) {
if (diffAFD < -22) {
if (Rs < 1273) {
return 8;
}
else if (Rs >= 1273) {
return 4;
}
}
else if (diffAFD >= -22) {
return 8;
}
}
}
else if (diffAFD >= -9) {
if (ssDCval < 88) {
if (Rs < 1065) {
return 4;
}
else if (Rs >= 1065) {
return 8;
}
}
else if (ssDCval >= 88) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (TSCindex < 1) {
if (diffRsCsdiff < -31) {
if (posBalance < 5) {
return 8;
}
else if (posBalance >= 5) {
return 4;
}
}
else if (diffRsCsdiff >= -31) {
return 4;
}
}
else if (TSCindex >= 1) {
if (Rs < 1195) {
return 8;
}
else if (Rs >= 1195) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 1596) {
if (Cs < 1278) {
return 8;
}
else if (Cs >= 1278) {
if (Rs < 1935) {
return 4;
}
else if (Rs >= 1935) {
return 1;
}
}
}
}
else if (diffTSC >= -2) {
if (Rs < 1784) {
if (ssDCval < 152) {
if (Rs < 752) {
if (diffRsCsdiff < -1) {
if (tcor < 98) {
if (Rs < 708) {
return 1;
}
else if (Rs >= 708) {
return 8;
}
}
else if (tcor >= 98) {
if (mu_mv_mag_sq < 126) {
return 2;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 704) {
return 1;
}
else if (Rs >= 704) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -1) {
if (ssDCval < 135) {
if (mu_mv_mag_sq < 124) {
if (posBalance < 27) {
return 1;
}
else if (posBalance >= 27) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 718) {
if (Rs < 704) {
return 2;
}
else if (Rs >= 704) {
return 1;
}
}
else if (Rs >= 718) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (ssDCval >= 135) {
return 1;
}
}
}
else if (Rs >= 752) {
if (SC < 1792) {
if (tcor < 101) {
if (SC < 1492) {
if (ssDCval < 68) {
if (tcor < 99) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (tcor >= 99) {
return 2;
}
}
else if (ssDCval >= 68) {
if (ssDCval < 135) {
return 8;
}
else if (ssDCval >= 135) {
return 1;
}
}
}
else if (SC >= 1492) {
return 8;
}
}
else if (tcor >= 101) {
if (Rs < 972) {
return 8;
}
else if (Rs >= 972) {
return 4;
}
}
}
else if (SC >= 1792) {
if (posBalance < 30) {
if (diffMVdiffVal < -3) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (Cs < 1022) {
if (ssDCval < 129) {
return 1;
}
else if (ssDCval >= 129) {
if (Rs < 1136) {
return 4;
}
else if (Rs >= 1136) {
return 8;
}
}
}
else if (Cs >= 1022) {
if (Rs < 910) {
return 2;
}
else if (Rs >= 910) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -3) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (MVDiff < 8) {
if (Cs < 661) {
return 1;
}
else if (Cs >= 661) {
if (Cs < 1136) {
return 8;
}
else if (Cs >= 1136) {
if (Rs < 1393) {
return 1;
}
else if (Rs >= 1393) {
return 8;
}
}
}
}
else if (MVDiff >= 8) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 679) {
if (Rs < 1624) {
return 8;
}
else if (Rs >= 1624) {
return 4;
}
}
else if (Cs >= 679) {
return 8;
}
}
}
}
else if (posBalance >= 30) {
if (posBalance < 37) {
if (SC < 1936) {
if (Rs < 1394) {
return 1;
}
else if (Rs >= 1394) {
return 2;
}
}
else if (SC >= 1936) {
if (ssDCval < 58) {
if (Rs < 1544) {
return 8;
}
else if (Rs >= 1544) {
return 1;
}
}
else if (ssDCval >= 58) {
if (Cs < 1002) {
return 8;
}
else if (Cs >= 1002) {
return 1;
}
}
}
}
else if (posBalance >= 37) {
if (Cs < 1041) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
if (refDCval < 95) {
return 8;
}
else if (refDCval >= 95) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
return 8;
}
}
}
}
else if (Cs >= 1041) {
if (ssDCval < 112) {
if (ssDCval < 50) {
if (Rs < 1311) {
return 8;
}
else if (Rs >= 1311) {
return 4;
}
}
else if (ssDCval >= 50) {
if (Cs < 1248) {
return 1;
}
else if (Cs >= 1248) {
return 8;
}
}
}
else if (ssDCval >= 112) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 152) {
if (Rs < 836) {
return 8;
}
else if (Rs >= 836) {
return 2;
}
}
}
else if (Rs >= 1784) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (ssDCval < 79) {
if (Cs < 1000) {
if (Cs < 627) {
return 2;
}
else if (Cs >= 627) {
return 1;
}
}
else if (Cs >= 1000) {
if (ssDCval < 71) {
if (Rs < 2247) {
return 4;
}
else if (Rs >= 2247) {
if (Rs < 2266) {
return 8;
}
else if (Rs >= 2266) {
return 2;
}
}
}
else if (ssDCval >= 71) {
return 1;
}
}
}
else if (ssDCval >= 79) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (Rs < 2081) {
return 1;
}
else if (Rs >= 2081) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 45) {
if (ssDCval < 47) {
if (mu_mv_mag_sq < 52) {
if (Rs < 1706) {
if (CsDiff < 3) {
if (Cs < 773) {
if (Cs < 523) {
if (Rs < 598) {
return 1;
}
else if (Rs >= 598) {
return 2;
}
}
else if (Cs >= 523) {
return 4;
}
}
else if (Cs >= 773) {
return 8;
}
}
else if (CsDiff >= 3) {
if (gchDC < 3) {
if (mcTcor < 94) {
if (Rs < 354) {
if (Rs < 345) {
return 8;
}
else if (Rs >= 345) {
return 1;
}
}
else if (Rs >= 354) {
return 8;
}
}
else if (mcTcor >= 94) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 603) {
return 1;
}
else if (Rs >= 603) {
if (Rs < 754) {
return 4;
}
else if (Rs >= 754) {
return 2;
}
}
}
}
}
else if (gchDC >= 3) {
if (gchDC < 5) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (gchDC >= 5) {
return 8;
}
}
}
}
else if (Rs >= 1706) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (Rs < 1572) {
if (ssDCval < 29) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
return 4;
}
}
else if (ssDCval >= 29) {
if (posBalance < 35) {
if (TSCindex < 3) {
return 2;
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 107) {
return 1;
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 795) {
return 8;
}
else if (Rs >= 795) {
return 1;
}
}
}
}
else if (posBalance >= 35) {
if (Cs < 651) {
if (TSC < 4) {
if (Rs < 789) {
return 2;
}
else if (Rs >= 789) {
return 1;
}
}
else if (TSC >= 4) {
if (Rs < 1134) {
return 8;
}
else if (Rs >= 1134) {
return 4;
}
}
}
else if (Cs >= 651) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Cs < 976) {
if (tcor < 96) {
if (AFD < 6) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 4;
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (tcor >= 96) {
return 8;
}
}
else if (Cs >= 976) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 1572) {
if (MVDiff < 7) {
if (refDCval < 46) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
return 4;
}
}
else if (refDCval >= 46) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (Cs < 837) {
return 4;
}
else if (Cs >= 837) {
return 1;
}
}
}
}
}
else if (ssDCval >= 47) {
if (mu_mv_mag_sq < 86) {
if (negBalance < 64) {
if (Cs < 890) {
if (Rs < 808) {
if (ssDCval < 78) {
if (MVDiff < 12) {
if (CsDiff < 3) {
if (Rs < 462) {
return 2;
}
else if (Rs >= 462) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (CsDiff >= 3) {
if (Cs < 527) {
return 8;
}
else if (Cs >= 527) {
if (RsCsDiff < 29) {
return 2;
}
else if (RsCsDiff >= 29) {
return 8;
}
}
}
}
else if (MVDiff >= 12) {
if (AFD < 6) {
if (Rs < 582) {
return 4;
}
else if (Rs >= 582) {
return 8;
}
}
else if (AFD >= 6) {
if (SC < 1330) {
return 1;
}
else if (SC >= 1330) {
if (Rs < 782) {
return 8;
}
else if (Rs >= 782) {
return 1;
}
}
}
}
}
else if (ssDCval >= 78) {
if (Cs < 646) {
if (tcor < 92) {
if (Rs < 460) {
return 2;
}
else if (Rs >= 460) {
return 8;
}
}
else if (tcor >= 92) {
return 2;
}
}
else if (Cs >= 646) {
return 2;
}
}
}
else if (Rs >= 808) {
if (MVDiff < 13) {
if (tcor < 84) {
if (ssDCval < 76) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (Rs < 2013) {
return 2;
}
else if (Rs >= 2013) {
return 4;
}
}
}
else if (ssDCval >= 76) {
if (Rs < 1793) {
return 8;
}
else if (Rs >= 1793) {
return 1;
}
}
}
else if (tcor >= 84) {
if (ssDCval < 114) {
if (posBalance < 44) {
if (Rs < 2113) {
if (mcTcor < 94) {
if (Rs < 1070) {
return 8;
}
else if (Rs >= 1070) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (mcTcor >= 94) {
if (diffMVdiffVal < 4) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
if (MVDiff < 3) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (diffMVdiffVal >= 4) {
if (AFD < 7) {
if (Rs < 1512) {
return 8;
}
else if (Rs >= 1512) {
return 2;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (Rs >= 2113) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
if (Rs < 2218) {
return 4;
}
else if (Rs >= 2218) {
return 2;
}
}
}
}
else if (posBalance >= 44) {
if (negBalance < 39) {
if (Rs < 1902) {
return 1;
}
else if (Rs >= 1902) {
return 2;
}
}
else if (negBalance >= 39) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Rs < 1542) {
return 4;
}
else if (Rs >= 1542) {
return 1;
}
}
}
}
}
else if (ssDCval >= 114) {
if (Rs < 2104) {
return 4;
}
else if (Rs >= 2104) {
return 1;
}
}
}
}
else if (MVDiff >= 13) {
if (AFD < 6) {
if (Rs < 946) {
return 1;
}
else if (Rs >= 946) {
return 4;
}
}
else if (AFD >= 6) {
if (AFD < 10) {
if (Rs < 1359) {
if (Cs < 609) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (Cs >= 609) {
if (Cs < 860) {
return 8;
}
else if (Cs >= 860) {
return 2;
}
}
}
else if (Rs >= 1359) {
return 1;
}
}
else if (AFD >= 10) {
if (diffAFD < -1) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (diffAFD >= -1) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
if (ssDCval < 58) {
return 2;
}
else if (ssDCval >= 58) {
if (Rs < 1203) {
return 2;
}
else if (Rs >= 1203) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Cs >= 890) {
if (tcor < 89) {
if (mcTcor < 89) {
if (mu_mv_mag_sq < 13) {
return 8;
}
else if (mu_mv_mag_sq >= 13) {
if (RsCsDiff < 83) {
if (CsDiff < 5) {
if (Cs < 955) {
return 8;
}
else if (Cs >= 955) {
return 4;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (RsCsDiff >= 83) {
if (Rs < 1447) {
return 1;
}
else if (Rs >= 1447) {
return 2;
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1511) {
if (RsCsDiff < 23) {
if (Rs < 1001) {
return 2;
}
else if (Rs >= 1001) {
if (Cs < 1101) {
return 8;
}
else if (Cs >= 1101) {
return 2;
}
}
}
else if (RsCsDiff >= 23) {
return 2;
}
}
else if (Rs >= 1511) {
if (Rs < 1770) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 1;
}
}
else if (Rs >= 1770) {
if (MVDiff < 10) {
if (Rs < 1980) {
if (Rs < 1905) {
return 8;
}
else if (Rs >= 1905) {
return 2;
}
}
else if (Rs >= 1980) {
if (negBalance < 44) {
return 4;
}
else if (negBalance >= 44) {
return 8;
}
}
}
else if (MVDiff >= 10) {
return 2;
}
}
}
}
}
else if (tcor >= 89) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (ssDCval < 72) {
if (Cs < 1017) {
if (TSC < 5) {
if (tcor < 94) {
if (Rs < 1282) {
return 8;
}
else if (Rs >= 1282) {
return 1;
}
}
else if (tcor >= 94) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
return 4;
}
}
}
else if (TSC >= 5) {
if (tcor < 92) {
if (Rs < 2025) {
return 4;
}
else if (Rs >= 2025) {
return 2;
}
}
else if (tcor >= 92) {
return 2;
}
}
}
else if (Cs >= 1017) {
if (RsDiff < 3) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 2;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
}
else if (ssDCval >= 72) {
if (ssDCval < 81) {
return 1;
}
else if (ssDCval >= 81) {
if (MVDiff < 25) {
return 8;
}
else if (MVDiff >= 25) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 64) {
if (gchDC < 3) {
if (mu_mv_mag_sq < 43) {
if (Cs < 834) {
if (ssDCval < 114) {
return 1;
}
else if (ssDCval >= 114) {
return 2;
}
}
else if (Cs >= 834) {
if (Rs < 1527) {
return 4;
}
else if (Rs >= 1527) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 43) {
return 8;
}
}
else if (gchDC >= 3) {
if (tcor < 84) {
if (ssDCval < 88) {
return 8;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (tcor >= 84) {
if (ssDCval < 129) {
return 8;
}
else if (ssDCval >= 129) {
if (Rs < 1419) {
return 2;
}
else if (Rs >= 1419) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (refDCval < 56) {
if (TSC < 3) {
if (Rs < 589) {
if (Rs < 442) {
return 2;
}
else if (Rs >= 442) {
return 1;
}
}
else if (Rs >= 589) {
if (negBalance < 30) {
if (Cs < 1003) {
if (Rs < 1215) {
return 4;
}
else if (Rs >= 1215) {
return 8;
}
}
else if (Cs >= 1003) {
if (Rs < 1128) {
return 1;
}
else if (Rs >= 1128) {
return 4;
}
}
}
else if (negBalance >= 30) {
return 4;
}
}
}
else if (TSC >= 3) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (refDCval < 52) {
if (posBalance < 42) {
if (Rs < 1437) {
return 1;
}
else if (Rs >= 1437) {
return 4;
}
}
else if (posBalance >= 42) {
return 8;
}
}
else if (refDCval >= 52) {
return 1;
}
}
}
}
else if (refDCval >= 56) {
if (Rs < 768) {
if (Cs < 655) {
if (ssDCval < 75) {
if (Cs < 540) {
return 2;
}
else if (Cs >= 540) {
return 1;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
else if (Cs >= 655) {
if (negBalance < 32) {
if (mu_mv_mag_sq < 107) {
return 4;
}
else if (mu_mv_mag_sq >= 107) {
return 2;
}
}
else if (negBalance >= 32) {
if (tcor < 94) {
if (Rs < 438) {
return 1;
}
else if (Rs >= 438) {
return 2;
}
}
else if (tcor >= 94) {
return 8;
}
}
}
}
else if (Rs >= 768) {
if (Rs < 1750) {
if (tcor < 74) {
return 4;
}
else if (tcor >= 74) {
if (negBalance < 14) {
if (posBalance < 23) {
return 8;
}
else if (posBalance >= 23) {
return 1;
}
}
else if (negBalance >= 14) {
if (MVDiff < 4) {
if (Cs < 1179) {
return 8;
}
else if (Cs >= 1179) {
if (Rs < 1157) {
return 4;
}
else if (Rs >= 1157) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
else if (MVDiff >= 4) {
if (AFD < 4) {
if (Rs < 1383) {
return 1;
}
else if (Rs >= 1383) {
return 8;
}
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
else if (Rs >= 1750) {
if (ssDCval < 73) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 2;
}
}
}
else if (ssDCval >= 73) {
if (Cs < 1032) {
if (negBalance < 33) {
return 4;
}
else if (negBalance >= 33) {
if (Cs < 815) {
if (Rs < 2271) {
if (Rs < 2255) {
return 4;
}
else if (Rs >= 2255) {
return 4;
}
}
else if (Rs >= 2271) {
return 8;
}
}
else if (Cs >= 815) {
return 8;
}
}
}
else if (Cs >= 1032) {
return 8;
}
}
}
}
}
}
}
}
else if (posBalance >= 45) {
if (mu_mv_mag_sq < 41) {
if (negBalance < 45) {
if (ssDCval < 54) {
if (Rs < 883) {
if (negBalance < 32) {
if (refDCval < 47) {
if (Cs < 974) {
if (posBalance < 63) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 4;
}
}
else if (posBalance >= 63) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (Cs < 854) {
return 1;
}
else if (Cs >= 854) {
return 8;
}
}
}
}
else if (Cs >= 974) {
return 1;
}
}
else if (refDCval >= 47) {
if (Rs < 393) {
return 1;
}
else if (Rs >= 393) {
return 4;
}
}
}
else if (negBalance >= 32) {
if (RsCsDiff < 11) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
return 4;
}
}
else if (RsCsDiff >= 11) {
if (ssDCval < 44) {
if (MVDiff < 10) {
if (Cs < 780) {
return 8;
}
else if (Cs >= 780) {
return 2;
}
}
else if (MVDiff >= 10) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (ssDCval >= 44) {
if (negBalance < 43) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (posBalance < 53) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
else if (posBalance >= 53) {
return 8;
}
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
}
else if (Rs >= 883) {
if (Cs < 1138) {
if (Cs < 611) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (Cs >= 611) {
if (SC < 3098) {
if (Cs < 668) {
if (Cs < 662) {
return 8;
}
else if (Cs >= 662) {
return 2;
}
}
else if (Cs >= 668) {
return 8;
}
}
else if (SC >= 3098) {
if (Rs < 2150) {
return 4;
}
else if (Rs >= 2150) {
return 8;
}
}
}
}
else if (Cs >= 1138) {
if (ssDCval < 48) {
return 8;
}
else if (ssDCval >= 48) {
return 4;
}
}
}
}
else if (ssDCval >= 54) {
if (Cs < 707) {
if (ssDCval < 107) {
if (TSC < 4) {
if (Rs < 747) {
return 2;
}
else if (Rs >= 747) {
return 8;
}
}
else if (TSC >= 4) {
if (SC < 2808) {
if (Rs < 939) {
if (Cs < 591) {
if (ssDCval < 69) {
if (ssDCval < 56) {
if (Rs < 773) {
return 8;
}
else if (Rs >= 773) {
return 4;
}
}
else if (ssDCval >= 56) {
if (Rs < 922) {
return 1;
}
else if (Rs >= 922) {
return 4;
}
}
}
else if (ssDCval >= 69) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (Cs >= 591) {
if (RsCsDiff < 18) {
if (Cs < 667) {
if (Cs < 606) {
return 4;
}
else if (Cs >= 606) {
return 2;
}
}
else if (Cs >= 667) {
return 8;
}
}
else if (RsCsDiff >= 18) {
if (TSC < 6) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (TSC >= 6) {
return 2;
}
}
}
}
else if (Rs >= 939) {
if (ssDCval < 67) {
if (RsDiff < 4) {
if (SC < 1630) {
return 1;
}
else if (SC >= 1630) {
if (Rs < 1133) {
return 4;
}
else if (Rs >= 1133) {
return 1;
}
}
}
else if (RsDiff >= 4) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
return 2;
}
}
}
else if (ssDCval >= 67) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 1665) {
return 8;
}
else if (Rs >= 1665) {
return 2;
}
}
}
}
}
}
else if (SC >= 2808) {
return 2;
}
}
}
else if (ssDCval >= 107) {
if (diffMVdiffVal < -5) {
return 2;
}
else if (diffMVdiffVal >= -5) {
if (MVDiff < 27) {
return 8;
}
else if (MVDiff >= 27) {
if (Rs < 968) {
if (Rs < 746) {
return 8;
}
else if (Rs >= 746) {
return 2;
}
}
else if (Rs >= 968) {
if (Rs < 1252) {
return 4;
}
else if (Rs >= 1252) {
return 1;
}
}
}
}
}
}
else if (Cs >= 707) {
if (negBalance < 33) {
if (tcor < 90) {
if (refDCval < 137) {
if (Rs < 1270) {
return 1;
}
else if (Rs >= 1270) {
if (AFD < 10) {
if (Cs < 922) {
return 1;
}
else if (Cs >= 922) {
return 8;
}
}
else if (AFD >= 10) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
}
else if (refDCval >= 137) {
return 2;
}
}
else if (tcor >= 90) {
if (Rs < 1016) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (Rs >= 1016) {
return 2;
}
}
}
else if (negBalance >= 33) {
if (mu_mv_mag_sq < 18) {
if (mcTcor < 93) {
if (ssDCval < 75) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (negBalance < 36) {
if (Rs < 851) {
return 8;
}
else if (Rs >= 851) {
return 1;
}
}
else if (negBalance >= 36) {
if (tcor < 81) {
if (diffMVdiffVal < 4) {
if (Rs < 1102) {
return 1;
}
else if (Rs >= 1102) {
return 8;
}
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
else if (tcor >= 81) {
return 4;
}
}
}
}
else if (ssDCval >= 75) {
if (ssDCval < 125) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (ssDCval >= 125) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
if (Rs < 1161) {
if (Rs < 1082) {
return 4;
}
else if (Rs >= 1082) {
return 8;
}
}
else if (Rs >= 1161) {
return 2;
}
}
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 93) {
if (MVDiff < 6) {
if (ssDCval < 60) {
if (Cs < 1203) {
return 8;
}
else if (Cs >= 1203) {
return 4;
}
}
else if (ssDCval >= 60) {
if (Rs < 1117) {
if (Rs < 781) {
return 1;
}
else if (Rs >= 781) {
if (Cs < 1237) {
return 8;
}
else if (Cs >= 1237) {
return 2;
}
}
}
else if (Rs >= 1117) {
if (Rs < 1863) {
return 1;
}
else if (Rs >= 1863) {
return 2;
}
}
}
}
else if (MVDiff >= 6) {
if (posBalance < 52) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
if (Rs < 1974) {
return 1;
}
else if (Rs >= 1974) {
return 8;
}
}
}
else if (posBalance >= 52) {
if (RsCsDiff < 22) {
if (Cs < 771) {
return 2;
}
else if (Cs >= 771) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (Rs < 617) {
return 1;
}
else if (Rs >= 617) {
return 4;
}
}
}
}
else if (RsCsDiff >= 22) {
if (SC < 3149) {
if (AFD < 12) {
if (posBalance < 55) {
if (AFD < 11) {
if (RsCsDiff < 39) {
return 4;
}
else if (RsCsDiff >= 39) {
return 2;
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (posBalance >= 55) {
return 2;
}
}
else if (AFD >= 12) {
return 4;
}
}
else if (SC >= 3149) {
return 4;
}
}
}
}
}
else if (ssDCval >= 93) {
if (refDCval < 135) {
if (AFD < 12) {
if (diffRsCsdiff < -41) {
return 2;
}
else if (diffRsCsdiff >= -41) {
return 8;
}
}
else if (AFD >= 12) {
if (Rs < 1747) {
return 8;
}
else if (Rs >= 1747) {
return 2;
}
}
}
else if (refDCval >= 135) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (ssDCval < 138) {
if (RsDiff < 7) {
if (tcor < 78) {
if (AFD < 11) {
if (Rs < 1261) {
return 1;
}
else if (Rs >= 1261) {
return 8;
}
}
else if (AFD >= 11) {
return 8;
}
}
else if (tcor >= 78) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (TSC < 6) {
if (Rs < 834) {
return 4;
}
else if (Rs >= 834) {
return 8;
}
}
else if (TSC >= 6) {
if (ssDCval < 63) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (Rs < 1397) {
return 2;
}
else if (Rs >= 1397) {
return 8;
}
}
}
else if (ssDCval >= 63) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (ssDCval < 66) {
if (Rs < 1539) {
return 1;
}
else if (Rs >= 1539) {
return 4;
}
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (ssDCval >= 138) {
if (SC < 2406) {
return 4;
}
else if (SC >= 2406) {
if (Rs < 1430) {
return 2;
}
else if (Rs >= 1430) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 415) {
if (RsCsDiff < 32) {
return 1;
}
else if (RsCsDiff >= 32) {
if (Cs < 1079) {
if (Rs < 376) {
return 2;
}
else if (Rs >= 376) {
return 4;
}
}
else if (Cs >= 1079) {
return 1;
}
}
}
else if (Rs >= 415) {
if (MVDiff < 5) {
if (ssDCval < 89) {
if (ssDCval < 56) {
if (ssDCval < 49) {
if (Rs < 836) {
if (ssDCval < 44) {
return 4;
}
else if (ssDCval >= 44) {
return 1;
}
}
else if (Rs >= 836) {
if (negBalance < 52) {
if (Cs < 1224) {
return 8;
}
else if (Cs >= 1224) {
return 4;
}
}
else if (negBalance >= 52) {
return 1;
}
}
}
else if (ssDCval >= 49) {
if (tcor < 74) {
if (refDCval < 52) {
return 8;
}
else if (refDCval >= 52) {
return 4;
}
}
else if (tcor >= 74) {
if (Cs < 1033) {
return 8;
}
else if (Cs >= 1033) {
return 4;
}
}
}
}
else if (ssDCval >= 56) {
if (Rs < 1926) {
if (gchDC < 2) {
if (Cs < 518) {
return 1;
}
else if (Cs >= 518) {
return 8;
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (Rs >= 1926) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (Cs < 912) {
return 2;
}
else if (Cs >= 912) {
return 1;
}
}
}
}
}
else if (ssDCval >= 89) {
if (refDCval < 109) {
if (CsDiff < 4) {
if (diffMVdiffVal < -3) {
if (Rs < 1179) {
return 8;
}
else if (Rs >= 1179) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (refDCval >= 109) {
if (tcor < 88) {
if (Rs < 722) {
return 1;
}
else if (Rs >= 722) {
return 8;
}
}
else if (tcor >= 88) {
return 2;
}
}
}
}
else if (MVDiff >= 5) {
if (Rs < 2199) {
if (ssDCval < 135) {
if (AFD < 11) {
if (refDCval < 92) {
if (ssDCval < 53) {
if (Rs < 637) {
if (tcor < 65) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
return 8;
}
}
else if (tcor >= 65) {
if (tcor < 81) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (tcor >= 81) {
return 8;
}
}
}
else if (Rs >= 637) {
if (Cs < 765) {
if (ssDCval < 45) {
if (mcTcor < 89) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (mcTcor >= 89) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (Rs < 739) {
return 4;
}
else if (Rs >= 739) {
return 2;
}
}
}
}
else if (ssDCval >= 45) {
if (negBalance < 54) {
if (refDCval < 47) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 1;
}
}
else if (refDCval >= 47) {
if (Cs < 613) {
if (diffAFD < -1) {
if (Rs < 793) {
return 8;
}
else if (Rs >= 793) {
return 2;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (Cs >= 613) {
if (Rs < 909) {
if (Rs < 765) {
return 8;
}
else if (Rs >= 765) {
return 1;
}
}
else if (Rs >= 909) {
return 8;
}
}
}
}
else if (negBalance >= 54) {
return 8;
}
}
}
else if (Cs >= 765) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Rs < 1079) {
return 4;
}
else if (Rs >= 1079) {
return 8;
}
}
}
}
}
else if (ssDCval >= 53) {
if (ssDCval < 63) {
if (mcTcor < 90) {
if (SC < 2518) {
if (MVDiff < 13) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
return 4;
}
}
else if (MVDiff >= 13) {
if (RsCsDiff < 35) {
return 8;
}
else if (RsCsDiff >= 35) {
return 4;
}
}
}
else if (SC >= 2518) {
return 8;
}
}
else if (mcTcor >= 90) {
if (diffAFD < -3) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
else if (diffAFD >= -3) {
if (Cs < 1221) {
if (Cs < 523) {
return 4;
}
else if (Cs >= 523) {
if (refDCval < 54) {
if (Cs < 977) {
return 8;
}
else if (Cs >= 977) {
return 4;
}
}
else if (refDCval >= 54) {
return 8;
}
}
}
else if (Cs >= 1221) {
if (Rs < 1501) {
return 4;
}
else if (Rs >= 1501) {
return 8;
}
}
}
}
}
else if (ssDCval >= 63) {
if (Rs < 943) {
if (diffMVdiffVal < 8) {
if (ssDCval < 71) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
if (Rs < 823) {
return 4;
}
else if (Rs >= 823) {
return 8;
}
}
}
else if (ssDCval >= 71) {
if (diffAFD < 5) {
if (Cs < 509) {
return 2;
}
else if (Cs >= 509) {
return 8;
}
}
else if (diffAFD >= 5) {
return 1;
}
}
}
else if (diffMVdiffVal >= 8) {
return 2;
}
}
else if (Rs >= 943) {
if (mcTcor < 92) {
if (posBalance < 50) {
return 4;
}
else if (posBalance >= 50) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Cs < 1279) {
return 2;
}
else if (Cs >= 1279) {
return 8;
}
}
}
}
else if (mcTcor >= 92) {
if (mcTcor < 96) {
if (diffMVdiffVal < -8) {
if (Rs < 1417) {
return 4;
}
else if (Rs >= 1417) {
return 2;
}
}
else if (diffMVdiffVal >= -8) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (TSC < 7) {
if (RsDiff < 5) {
if (Rs < 951) {
return 2;
}
else if (Rs >= 951) {
return 4;
}
}
else if (RsDiff >= 5) {
if (Rs < 2054) {
return 1;
}
else if (Rs >= 2054) {
return 4;
}
}
}
else if (TSC >= 7) {
if (Rs < 1887) {
return 8;
}
else if (Rs >= 1887) {
return 2;
}
}
}
}
}
}
}
}
else if (refDCval >= 92) {
if (tcor < 71) {
if (ssDCval < 121) {
if (posBalance < 60) {
return 2;
}
else if (posBalance >= 60) {
if (diffRsCsdiff < -7) {
return 2;
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
}
else if (ssDCval >= 121) {
return 1;
}
}
else if (tcor >= 71) {
if (ssDCval < 94) {
if (Cs < 971) {
return 2;
}
else if (Cs >= 971) {
if (Rs < 1295) {
return 1;
}
else if (Rs >= 1295) {
return 8;
}
}
}
else if (ssDCval >= 94) {
if (SC < 2229) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (Rs < 1127) {
return 2;
}
else if (Rs >= 1127) {
return 1;
}
}
}
}
else if (SC >= 2229) {
if (ssDCval < 99) {
if (SC < 2341) {
return 2;
}
else if (SC >= 2341) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 1593) {
return 8;
}
else if (Rs >= 1593) {
return 2;
}
}
}
}
else if (ssDCval >= 99) {
return 8;
}
}
}
}
}
}
else if (AFD >= 11) {
if (Rs < 942) {
if (Cs < 628) {
if (TSC < 6) {
if (Cs < 550) {
return 4;
}
else if (Cs >= 550) {
return 2;
}
}
else if (TSC >= 6) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (diffTSC < 1) {
if (diffRsCsdiff < 13) {
return 8;
}
else if (diffRsCsdiff >= 13) {
if (Rs < 590) {
return 4;
}
else if (Rs >= 590) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Cs < 539) {
return 2;
}
else if (Cs >= 539) {
return 4;
}
}
}
}
}
}
else if (Cs >= 628) {
if (TSC < 7) {
if (CsDiff < 7) {
if (Rs < 843) {
if (gchDC < 1) {
if (MVDiff < 26) {
if (Rs < 638) {
return 2;
}
else if (Rs >= 638) {
return 1;
}
}
else if (MVDiff >= 26) {
if (Rs < 751) {
return 8;
}
else if (Rs >= 751) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 669) {
if (Rs < 478) {
return 4;
}
else if (Rs >= 478) {
return 1;
}
}
else if (Rs >= 669) {
return 2;
}
}
}
}
else if (Rs >= 843) {
if (ssDCval < 57) {
if (Rs < 885) {
return 2;
}
else if (Rs >= 885) {
return 1;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (TSC >= 7) {
if (negBalance < 50) {
if (posBalance < 61) {
if (Rs < 882) {
return 1;
}
else if (Rs >= 882) {
return 4;
}
}
else if (posBalance >= 61) {
return 4;
}
}
else if (negBalance >= 50) {
if (Rs < 893) {
if (TSC < 9) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (Rs >= 893) {
if (Rs < 917) {
return 8;
}
else if (Rs >= 917) {
return 4;
}
}
}
}
}
}
else if (Rs >= 942) {
if (RsCsDiff < 38) {
if (ssDCval < 113) {
if (Cs < 563) {
if (Rs < 1026) {
return 8;
}
else if (Rs >= 1026) {
if (Rs < 1181) {
return 4;
}
else if (Rs >= 1181) {
return 2;
}
}
}
else if (Cs >= 563) {
return 8;
}
}
else if (ssDCval >= 113) {
return 4;
}
}
else if (RsCsDiff >= 38) {
if (Cs < 661) {
if (ssDCval < 74) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
else if (ssDCval >= 74) {
if (Rs < 1758) {
if (ssDCval < 129) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 1;
}
}
else if (ssDCval >= 129) {
return 4;
}
}
else if (Rs >= 1758) {
return 2;
}
}
}
else if (Cs >= 661) {
if (tcor < 89) {
if (tcor < 80) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (gchDC < 3) {
if (CsDiff < 9) {
if (RsCsDiff < 85) {
return 8;
}
else if (RsCsDiff >= 85) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (gchDC >= 3) {
return 2;
}
}
}
else if (tcor >= 80) {
if (SC < 2222) {
if (diffMVdiffVal < -1) {
if (Cs < 730) {
return 1;
}
else if (Cs >= 730) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (ssDCval < 110) {
return 1;
}
else if (ssDCval >= 110) {
if (Rs < 1255) {
return 4;
}
else if (Rs >= 1255) {
return 8;
}
}
}
}
}
else if (SC >= 2222) {
if (refDCval < 59) {
return 4;
}
else if (refDCval >= 59) {
if (posBalance < 51) {
if (Rs < 1568) {
return 8;
}
else if (Rs >= 1568) {
return 4;
}
}
else if (posBalance >= 51) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (MVDiff < 20) {
if (tcor < 81) {
return 4;
}
else if (tcor >= 81) {
if (ssDCval < 71) {
if (Rs < 1660) {
return 8;
}
else if (Rs >= 1660) {
return 1;
}
}
else if (ssDCval >= 71) {
if (Cs < 788) {
return 4;
}
else if (Cs >= 788) {
if (MVDiff < 6) {
if (Rs < 1608) {
return 8;
}
else if (Rs >= 1608) {
return 4;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
}
}
}
else if (MVDiff >= 20) {
if (Cs < 897) {
return 1;
}
else if (Cs >= 897) {
return 8;
}
}
}
}
}
}
}
}
else if (tcor >= 89) {
if (Rs < 1205) {
if (Rs < 1154) {
return 4;
}
else if (Rs >= 1154) {
return 1;
}
}
else if (Rs >= 1205) {
if (RsCsDiff < 87) {
if (ssDCval < 125) {
return 8;
}
else if (ssDCval >= 125) {
if (Rs < 1352) {
return 8;
}
else if (Rs >= 1352) {
return 2;
}
}
}
else if (RsCsDiff >= 87) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 135) {
if (ssDCval < 142) {
if (MVDiff < 19) {
if (Cs < 790) {
return 2;
}
else if (Cs >= 790) {
return 8;
}
}
else if (MVDiff >= 19) {
if (mcTcor < 96) {
return 4;
}
else if (mcTcor >= 96) {
if (Rs < 1222) {
return 1;
}
else if (Rs >= 1222) {
return 4;
}
}
}
}
else if (ssDCval >= 142) {
if (negBalance < 50) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (negBalance >= 50) {
if (Rs < 689) {
return 2;
}
else if (Rs >= 689) {
return 1;
}
}
}
}
}
else if (Rs >= 2199) {
if (Cs < 858) {
return 2;
}
else if (Cs >= 858) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (Rs < 1576) {
if (TSC < 7) {
if (ssDCval < 47) {
if (mu_mv_mag_sq < 105) {
if (ssDCval < 39) {
if (Rs < 774) {
if (mu_mv_mag_sq < 59) {
if (Rs < 509) {
return 2;
}
else if (Rs >= 509) {
return 8;
}
}
else if (mu_mv_mag_sq >= 59) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (Rs >= 774) {
return 8;
}
}
else if (ssDCval >= 39) {
if (negBalance < 34) {
if (SC < 1786) {
if (refDCval < 39) {
if (Rs < 645) {
return 8;
}
else if (Rs >= 645) {
return 2;
}
}
else if (refDCval >= 39) {
return 1;
}
}
else if (SC >= 1786) {
if (Cs < 1174) {
return 8;
}
else if (Cs >= 1174) {
return 2;
}
}
}
else if (negBalance >= 34) {
if (Rs < 756) {
if (TSC < 6) {
if (Rs < 489) {
return 8;
}
else if (Rs >= 489) {
return 4;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (Rs >= 756) {
if (diffRsCsdiff < -8) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (Rs < 1016) {
return 2;
}
else if (Rs >= 1016) {
return 8;
}
}
}
else if (diffRsCsdiff >= -8) {
if (RsDiff < 5) {
if (tcor < 97) {
if (negBalance < 35) {
if (Rs < 1149) {
return 8;
}
else if (Rs >= 1149) {
return 2;
}
}
else if (negBalance >= 35) {
return 8;
}
}
else if (tcor >= 97) {
return 4;
}
}
else if (RsDiff >= 5) {
if (mcTcor < 82) {
if (SC < 1800) {
return 1;
}
else if (SC >= 1800) {
return 8;
}
}
else if (mcTcor >= 82) {
if (refDCval < 44) {
if (Rs < 1240) {
return 8;
}
else if (Rs >= 1240) {
return 2;
}
}
else if (refDCval >= 44) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (SC < 2199) {
if (Rs < 843) {
return 1;
}
else if (Rs >= 843) {
if (Rs < 1090) {
if (tcor < 81) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
else if (tcor >= 81) {
return 8;
}
}
else if (Rs >= 1090) {
if (ssDCval < 45) {
if (tcor < 81) {
return 1;
}
else if (tcor >= 81) {
return 8;
}
}
else if (ssDCval >= 45) {
return 1;
}
}
}
}
else if (SC >= 2199) {
if (refDCval < 45) {
return 8;
}
else if (refDCval >= 45) {
if (Cs < 1106) {
return 1;
}
else if (Cs >= 1106) {
return 8;
}
}
}
}
}
else if (ssDCval >= 47) {
if (ssDCval < 158) {
if (refDCval < 75) {
if (Rs < 512) {
if (Cs < 823) {
if (MVDiff < 4) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (Rs < 489) {
return 1;
}
else if (Rs >= 489) {
return 2;
}
}
}
else if (MVDiff >= 4) {
if (mcTcor < 93) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (mcTcor >= 93) {
if (mu_mv_mag_sq < 82) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 82) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
}
}
}
else if (Cs >= 823) {
if (TSC < 3) {
if (Rs < 431) {
return 2;
}
else if (Rs >= 431) {
return 1;
}
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (Rs >= 512) {
if (MVDiff < 3) {
if (negBalance < 45) {
if (refDCval < 64) {
if (mu_mv_mag_sq < 118) {
if (Rs < 717) {
if (Cs < 860) {
if (mu_mv_mag_sq < 98) {
return 8;
}
else if (mu_mv_mag_sq >= 98) {
if (negBalance < 39) {
return 4;
}
else if (negBalance >= 39) {
return 8;
}
}
}
else if (Cs >= 860) {
return 1;
}
}
else if (Rs >= 717) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (mcTcor < 96) {
if (diffMVdiffVal < 2) {
if (diffRsCsdiff < 4) {
if (RsCsDiff < 22) {
return 8;
}
else if (RsCsDiff >= 22) {
if (Rs < 1174) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (Rs < 942) {
return 8;
}
else if (Rs >= 942) {
return 4;
}
}
}
else if (Rs >= 1174) {
return 8;
}
}
}
else if (diffRsCsdiff >= 4) {
if (mu_mv_mag_sq < 70) {
return 4;
}
else if (mu_mv_mag_sq >= 70) {
if (Cs < 741) {
if (tcor < 94) {
if (Rs < 1302) {
return 4;
}
else if (Rs >= 1302) {
return 4;
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (Cs >= 741) {
if (Rs < 1064) {
if (Rs < 1022) {
return 8;
}
else if (Rs >= 1022) {
return 4;
}
}
else if (Rs >= 1064) {
return 8;
}
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 728) {
return 8;
}
else if (Cs >= 728) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (Cs < 861) {
if (mu_mv_mag_sq < 85) {
return 1;
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 923) {
return 4;
}
else if (Rs >= 923) {
return 8;
}
}
}
else if (Cs >= 861) {
if (mu_mv_mag_sq < 68) {
return 4;
}
else if (mu_mv_mag_sq >= 68) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (Cs < 1084) {
return 8;
}
else if (Cs >= 1084) {
if (RsDiff < 2) {
if (Rs < 979) {
return 1;
}
else if (Rs >= 979) {
return 8;
}
}
else if (RsDiff >= 2) {
if (Rs < 1286) {
return 4;
}
else if (Rs >= 1286) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (Rs < 663) {
if (Cs < 801) {
return 8;
}
else if (Cs >= 801) {
return 4;
}
}
else if (Rs >= 663) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
if (Rs < 1057) {
return 4;
}
else if (Rs >= 1057) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 64) {
if (RsDiff < 4) {
if (Cs < 897) {
if (RsDiff < 2) {
if (Cs < 699) {
if (Rs < 890) {
return 2;
}
else if (Rs >= 890) {
return 1;
}
}
else if (Cs >= 699) {
return 4;
}
}
else if (RsDiff >= 2) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 4;
}
}
}
else if (Cs >= 897) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 115) {
if (Rs < 1032) {
if (mu_mv_mag_sq < 84) {
if (Rs < 940) {
return 2;
}
else if (Rs >= 940) {
return 8;
}
}
else if (mu_mv_mag_sq >= 84) {
return 4;
}
}
else if (Rs >= 1032) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 1426) {
return 4;
}
else if (Rs >= 1426) {
return 8;
}
}
}
else if (negBalance >= 37) {
if (AFD < 5) {
if (negBalance < 39) {
if (Cs < 1224) {
return 8;
}
else if (Cs >= 1224) {
return 1;
}
}
else if (negBalance >= 39) {
return 1;
}
}
else if (AFD >= 5) {
if (gchDC < 1) {
if (ssDCval < 73) {
if (mu_mv_mag_sq < 70) {
return 1;
}
else if (mu_mv_mag_sq >= 70) {
return 8;
}
}
else if (ssDCval >= 73) {
return 4;
}
}
else if (gchDC >= 1) {
if (Cs < 1106) {
return 1;
}
else if (Cs >= 1106) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
return 1;
}
}
}
}
else if (negBalance >= 45) {
if (diffRsCsdiff < -12) {
if (Cs < 1224) {
return 8;
}
else if (Cs >= 1224) {
return 1;
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (RsCsDiff < 7) {
if (Rs < 842) {
if (ssDCval < 57) {
if (posBalance < 55) {
if (Cs < 574) {
return 1;
}
else if (Cs >= 574) {
return 4;
}
}
else if (posBalance >= 55) {
if (Rs < 734) {
return 1;
}
else if (Rs >= 734) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (mu_mv_mag_sq < 81) {
if (Rs < 665) {
return 1;
}
else if (Rs >= 665) {
if (Rs < 776) {
return 8;
}
else if (Rs >= 776) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (Cs < 681) {
if (Rs < 659) {
return 8;
}
else if (Rs >= 659) {
if (Rs < 729) {
return 2;
}
else if (Rs >= 729) {
return 4;
}
}
}
else if (Cs >= 681) {
return 8;
}
}
}
}
else if (Rs >= 842) {
if (Cs < 1073) {
if (negBalance < 27) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (negBalance >= 27) {
return 8;
}
}
else if (Cs >= 1073) {
return 1;
}
}
}
else if (RsCsDiff >= 7) {
if (negBalance < 40) {
if (tcor < 94) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Rs < 844) {
if (refDCval < 55) {
if (diffRsCsdiff < 8) {
return 4;
}
else if (diffRsCsdiff >= 8) {
return 8;
}
}
else if (refDCval >= 55) {
if (MVDiff < 8) {
if (Rs < 625) {
if (Cs < 897) {
return 8;
}
else if (Cs >= 897) {
return 1;
}
}
else if (Rs >= 625) {
if (Rs < 729) {
return 2;
}
else if (Rs >= 729) {
return 4;
}
}
}
else if (MVDiff >= 8) {
return 1;
}
}
}
else if (Rs >= 844) {
if (ssDCval < 60) {
if (Cs < 933) {
if (SC < 1870) {
if (diffMVdiffVal < 1) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (SC >= 1870) {
if (refDCval < 58) {
if (diffMVdiffVal < -5) {
return 8;
}
else if (diffMVdiffVal >= -5) {
if (Rs < 1428) {
return 2;
}
else if (Rs >= 1428) {
return 1;
}
}
}
else if (refDCval >= 58) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 2;
}
}
}
}
else if (Cs >= 933) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (negBalance < 36) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (diffAFD < -1) {
if (Cs < 746) {
return 8;
}
else if (Cs >= 746) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (negBalance >= 36) {
if (posBalance < 55) {
if (diffMVdiffVal < 2) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Cs < 1127) {
return 8;
}
else if (Cs >= 1127) {
return 1;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 919) {
return 4;
}
else if (Cs >= 919) {
if (Rs < 1124) {
return 2;
}
else if (Rs >= 1124) {
return 1;
}
}
}
}
else if (posBalance >= 55) {
if (mcTcor < 95) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 94) {
if (Rs < 919) {
if (Cs < 577) {
return 4;
}
else if (Cs >= 577) {
return 2;
}
}
else if (Rs >= 919) {
if (diffAFD < -1) {
if (Rs < 1081) {
return 8;
}
else if (Rs >= 1081) {
if (SC < 2320) {
return 2;
}
else if (SC >= 2320) {
return 4;
}
}
}
else if (diffAFD >= -1) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
if (diffMVdiffVal < -4) {
if (Rs < 1172) {
return 8;
}
else if (Rs >= 1172) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 40) {
if (ssDCval < 60) {
if (Rs < 1199) {
if (Cs < 814) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (TSC < 6) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 4;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (Cs >= 814) {
return 4;
}
}
else if (Rs >= 1199) {
return 8;
}
}
else if (ssDCval >= 60) {
if (Rs < 853) {
if (SC < 1677) {
return 8;
}
else if (SC >= 1677) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 1;
}
}
}
else if (Rs >= 853) {
if (Rs < 1248) {
if (Cs < 620) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Rs < 958) {
return 4;
}
else if (Rs >= 958) {
return 1;
}
}
}
else if (Cs >= 620) {
if (Cs < 1119) {
if (Cs < 640) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
return 8;
}
}
else if (Cs >= 640) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (Cs >= 1119) {
if (AFD < 8) {
if (Rs < 1238) {
return 4;
}
else if (Rs >= 1238) {
return 1;
}
}
else if (AFD >= 8) {
return 1;
}
}
}
}
else if (Rs >= 1248) {
if (CsDiff < 3) {
if (MVDiff < 6) {
if (Rs < 1344) {
return 1;
}
else if (Rs >= 1344) {
return 4;
}
}
else if (MVDiff >= 6) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (diffTSC < -1) {
if (Rs < 1382) {
return 8;
}
else if (Rs >= 1382) {
return 4;
}
}
else if (diffTSC >= -1) {
if (Cs < 915) {
if (Rs < 1325) {
return 4;
}
else if (Rs >= 1325) {
return 1;
}
}
else if (Cs >= 915) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (negBalance < 48) {
if (SC < 1685) {
if (negBalance < 42) {
if (posBalance < 54) {
if (ssDCval < 112) {
if (Rs < 481) {
return 8;
}
else if (Rs >= 481) {
if (tcor < 94) {
return 2;
}
else if (tcor >= 94) {
if (Cs < 716) {
if (Rs < 647) {
return 8;
}
else if (Rs >= 647) {
return 4;
}
}
else if (Cs >= 716) {
return 2;
}
}
}
}
else if (ssDCval >= 112) {
return 8;
}
}
else if (posBalance >= 54) {
if (TSC < 5) {
if (diffTSC < 1) {
if (Cs < 793) {
if (SC < 1268) {
if (negBalance < 32) {
if (ssDCval < 138) {
return 2;
}
else if (ssDCval >= 138) {
return 8;
}
}
else if (negBalance >= 32) {
return 8;
}
}
else if (SC >= 1268) {
return 2;
}
}
else if (Cs >= 793) {
return 8;
}
}
else if (diffTSC >= 1) {
if (Rs < 611) {
return 8;
}
else if (Rs >= 611) {
return 1;
}
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (negBalance >= 42) {
if (Cs < 712) {
if (CsDiff < 2) {
if (Rs < 616) {
return 8;
}
else if (Rs >= 616) {
if (Rs < 836) {
return 1;
}
else if (Rs >= 836) {
return 8;
}
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (Cs >= 712) {
if (CsDiff < 3) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Rs < 690) {
return 8;
}
else if (Rs >= 690) {
return 1;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 857) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
else if (Rs >= 857) {
return 1;
}
}
}
}
}
else if (SC >= 1685) {
if (refDCval < 130) {
if (Rs < 941) {
if (refDCval < 94) {
if (Cs < 961) {
return 1;
}
else if (Cs >= 961) {
if (RsDiff < 3) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
}
else if (RsDiff >= 3) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
}
}
else if (refDCval >= 94) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (Rs >= 941) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
}
else if (refDCval >= 130) {
if (Cs < 1034) {
if (AFD < 6) {
if (posBalance < 68) {
return 4;
}
else if (posBalance >= 68) {
return 1;
}
}
else if (AFD >= 6) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
return 4;
}
}
}
else if (Cs >= 1034) {
if (Rs < 1563) {
return 8;
}
else if (Rs >= 1563) {
return 4;
}
}
}
}
}
else if (negBalance >= 48) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Cs < 767) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (Rs < 1192) {
if (ssDCval < 129) {
return 1;
}
else if (ssDCval >= 129) {
return 8;
}
}
else if (Rs >= 1192) {
if (Rs < 1291) {
return 8;
}
else if (Rs >= 1291) {
return 4;
}
}
}
}
else if (Cs >= 767) {
if (Rs < 415) {
return 4;
}
else if (Rs >= 415) {
if (ssDCval < 135) {
if (Cs < 828) {
if (Rs < 985) {
return 1;
}
else if (Rs >= 985) {
return 8;
}
}
else if (Cs >= 828) {
return 8;
}
}
else if (ssDCval >= 135) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 158) {
if (Rs < 875) {
return 4;
}
else if (Rs >= 875) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (gchDC < 1) {
if (Rs < 1185) {
if (ssDCval < 70) {
if (ssDCval < 51) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 860) {
return 8;
}
else if (Rs >= 860) {
if (Rs < 983) {
return 1;
}
else if (Rs >= 983) {
return 8;
}
}
}
}
else if (ssDCval >= 51) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (AFD < 10) {
if (Cs < 525) {
return 8;
}
else if (Cs >= 525) {
return 4;
}
}
else if (AFD >= 10) {
if (ssDCval < 65) {
if (Rs < 755) {
return 4;
}
else if (Rs >= 755) {
return 8;
}
}
else if (ssDCval >= 65) {
return 1;
}
}
}
}
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (Rs >= 1185) {
if (Cs < 1225) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (Rs < 1459) {
return 2;
}
else if (Rs >= 1459) {
return 8;
}
}
}
else if (Cs >= 1225) {
if (mcTcor < 86) {
if (Rs < 1447) {
return 1;
}
else if (Rs >= 1447) {
return 8;
}
}
else if (mcTcor >= 86) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
if (refDCval < 79) {
if (Rs < 1125) {
if (refDCval < 51) {
if (Rs < 701) {
return 2;
}
else if (Rs >= 701) {
return 8;
}
}
else if (refDCval >= 51) {
if (Rs < 773) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
else if (Rs >= 773) {
if (mu_mv_mag_sq < 99) {
if (RsDiff < 5) {
if (diffRsCsdiff < -13) {
return 8;
}
else if (diffRsCsdiff >= -13) {
if (ssDCval < 54) {
if (Rs < 894) {
return 4;
}
else if (Rs >= 894) {
return 8;
}
}
else if (ssDCval >= 54) {
return 4;
}
}
}
else if (RsDiff >= 5) {
if (Rs < 1007) {
return 4;
}
else if (Rs >= 1007) {
if (Rs < 1066) {
return 1;
}
else if (Rs >= 1066) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
return 8;
}
}
}
}
else if (Rs >= 1125) {
if (posBalance < 68) {
if (Cs < 1154) {
return 8;
}
else if (Cs >= 1154) {
if (mcTcor < 91) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (mcTcor >= 91) {
if (Cs < 1236) {
return 4;
}
else if (Cs >= 1236) {
return 8;
}
}
}
}
else if (posBalance >= 68) {
return 4;
}
}
}
else if (refDCval >= 79) {
if (mu_mv_mag_sq < 101) {
if (TSC < 10) {
if (refDCval < 155) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (refDCval >= 155) {
if (Rs < 1071) {
return 2;
}
else if (Rs >= 1071) {
return 8;
}
}
}
else if (TSC >= 10) {
if (Rs < 1148) {
return 8;
}
else if (Rs >= 1148) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 1321) {
return 4;
}
else if (Rs >= 1321) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1576) {
if (refDCval < 72) {
if (refDCval < 62) {
if (Cs < 969) {
if (mu_mv_mag_sq < 55) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
return 2;
}
}
else if (mu_mv_mag_sq >= 55) {
if (posBalance < 55) {
if (mcTcor < 93) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Rs < 1664) {
return 2;
}
else if (Rs >= 1664) {
if (Rs < 1854) {
return 4;
}
else if (Rs >= 1854) {
return 4;
}
}
}
}
else if (mcTcor >= 93) {
if (Cs < 604) {
return 2;
}
else if (Cs >= 604) {
if (SC < 2677) {
return 1;
}
else if (SC >= 2677) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
if (Cs < 547) {
return 1;
}
else if (Cs >= 547) {
return 8;
}
}
}
}
else if (Cs >= 969) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
if (posBalance < 49) {
if (diffRsCsdiff < 4) {
return 8;
}
else if (diffRsCsdiff >= 4) {
if (Rs < 1771) {
return 8;
}
else if (Rs >= 1771) {
return 4;
}
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
else if (refDCval >= 62) {
if (Rs < 1650) {
if (TSC < 8) {
if (Cs < 793) {
return 2;
}
else if (Cs >= 793) {
return 8;
}
}
else if (TSC >= 8) {
if (Rs < 1582) {
return 8;
}
else if (Rs >= 1582) {
return 2;
}
}
}
else if (Rs >= 1650) {
if (CsDiff < 5) {
if (posBalance < 53) {
if (Cs < 935) {
if (Rs < 1692) {
return 8;
}
else if (Rs >= 1692) {
return 2;
}
}
else if (Cs >= 935) {
if (Rs < 2119) {
return 8;
}
else if (Rs >= 2119) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (SC < 3269) {
return 2;
}
else if (SC >= 3269) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 2131) {
return 8;
}
else if (Rs >= 2131) {
return 2;
}
}
}
}
}
else if (refDCval >= 72) {
if (refDCval < 90) {
if (Cs < 786) {
return 1;
}
else if (Cs >= 786) {
if (SC < 2980) {
if (negBalance < 30) {
if (Cs < 1005) {
return 1;
}
else if (Cs >= 1005) {
return 8;
}
}
else if (negBalance >= 30) {
if (diffRsCsdiff < 18) {
return 8;
}
else if (diffRsCsdiff >= 18) {
return 1;
}
}
}
else if (SC >= 2980) {
if (RsDiff < 5) {
if (Cs < 1295) {
if (ssDCval < 84) {
if (Rs < 2059) {
return 1;
}
else if (Rs >= 2059) {
return 8;
}
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (Cs >= 1295) {
return 8;
}
}
else if (RsDiff >= 5) {
if (ssDCval < 87) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Rs < 2276) {
if (SC < 3374) {
return 8;
}
else if (SC >= 3374) {
if (Rs < 2165) {
return 1;
}
else if (Rs >= 2165) {
return 8;
}
}
}
else if (Rs >= 2276) {
return 2;
}
}
}
else if (ssDCval >= 87) {
if (Cs < 1228) {
if (Rs < 2177) {
return 4;
}
else if (Rs >= 2177) {
return 1;
}
}
else if (Cs >= 1228) {
if (Rs < 1863) {
return 8;
}
else if (Rs >= 1863) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 90) {
if (negBalance < 49) {
if (ssDCval < 99) {
if (Cs < 1086) {
return 8;
}
else if (Cs >= 1086) {
if (TSC < 8) {
if (Rs < 2006) {
return 2;
}
else if (Rs >= 2006) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (ssDCval >= 99) {
if (mu_mv_mag_sq < 124) {
if (RsCsDiff < 87) {
if (diffRsCsdiff < 15) {
return 8;
}
else if (diffRsCsdiff >= 15) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (RsCsDiff >= 87) {
if (ssDCval < 109) {
return 4;
}
else if (ssDCval >= 109) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (negBalance >= 49) {
if (MVDiff < 5) {
if (tcor < 90) {
if (Rs < 1746) {
return 4;
}
else if (Rs >= 1746) {
return 2;
}
}
else if (tcor >= 90) {
return 1;
}
}
else if (MVDiff >= 5) {
if (Cs < 1215) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 1841) {
if (Rs < 1693) {
return 8;
}
else if (Rs >= 1693) {
return 1;
}
}
else if (Rs >= 1841) {
return 2;
}
}
}
else if (Cs >= 1215) {
if (Cs < 1230) {
return 4;
}
else if (Cs >= 1230) {
if (Rs < 1885) {
return 2;
}
else if (Rs >= 1885) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (refDCval < 114) {
if (negBalance < 45) {
if (posBalance < 49) {
if (diffAFD < 3) {
if (mu_mv_mag_sq < 32) {
if (SC < 2504) {
return 1;
}
else if (SC >= 2504) {
if (Rs < 2251) {
return 8;
}
else if (Rs >= 2251) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (mu_mv_mag_sq < 43) {
return 2;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 2116) {
return 1;
}
else if (Rs >= 2116) {
return 2;
}
}
}
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (posBalance >= 49) {
if (Rs < 1423) {
if (Rs < 451) {
if (Cs < 549) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (Cs >= 549) {
if (posBalance < 72) {
return 2;
}
else if (posBalance >= 72) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
return 4;
}
}
}
}
else if (Rs >= 451) {
if (negBalance < 21) {
if (Rs < 1129) {
return 8;
}
else if (Rs >= 1129) {
return 4;
}
}
else if (negBalance >= 21) {
if (MVDiff < 51) {
if (AFD < 16) {
if (ssDCval < 66) {
if (diffMVdiffVal < -10) {
if (Rs < 818) {
if (Rs < 541) {
return 8;
}
else if (Rs >= 541) {
return 4;
}
}
else if (Rs >= 818) {
return 2;
}
}
else if (diffMVdiffVal >= -10) {
if (Rs < 1334) {
return 1;
}
else if (Rs >= 1334) {
return 8;
}
}
}
else if (ssDCval >= 66) {
if (refDCval < 73) {
if (SC < 1701) {
if (Rs < 986) {
return 2;
}
else if (Rs >= 986) {
return 1;
}
}
else if (SC >= 1701) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 2;
}
}
}
else if (refDCval >= 73) {
if (SC < 2115) {
if (ssDCval < 83) {
if (Rs < 970) {
return 8;
}
else if (Rs >= 970) {
return 4;
}
}
else if (ssDCval >= 83) {
return 2;
}
}
else if (SC >= 2115) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
}
else if (AFD >= 16) {
if (mu_mv_mag_sq < 5) {
return 4;
}
else if (mu_mv_mag_sq >= 5) {
if (ssDCval < 113) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (CsDiff < 11) {
return 1;
}
else if (CsDiff >= 11) {
return 4;
}
}
}
else if (ssDCval >= 113) {
return 8;
}
}
}
}
else if (MVDiff >= 51) {
if (TSC < 8) {
if (Rs < 637) {
return 4;
}
else if (Rs >= 637) {
return 1;
}
}
else if (TSC >= 8) {
if (AFD < 25) {
if (mu_mv_mag_sq < 18) {
return 8;
}
else if (mu_mv_mag_sq >= 18) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (AFD >= 25) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
if (Rs < 1046) {
return 1;
}
else if (Rs >= 1046) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 1423) {
if (Cs < 804) {
if (diffAFD < 4) {
if (tcor < 76) {
if (Rs < 1986) {
if (Cs < 711) {
return 2;
}
else if (Cs >= 711) {
return 8;
}
}
else if (Rs >= 1986) {
return 4;
}
}
else if (tcor >= 76) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 2;
}
}
}
else if (diffAFD >= 4) {
if (Rs < 2090) {
if (AFD < 40) {
return 1;
}
else if (AFD >= 40) {
return 8;
}
}
else if (Rs >= 2090) {
return 2;
}
}
}
else if (Cs >= 804) {
if (RsCsDiff < 99) {
if (mcTcor < 94) {
if (SC < 2519) {
return 2;
}
else if (SC >= 2519) {
if (posBalance < 85) {
return 8;
}
else if (posBalance >= 85) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (refDCval < 100) {
return 1;
}
else if (refDCval >= 100) {
return 8;
}
}
}
else if (RsCsDiff >= 99) {
if (tcor < 65) {
if (Rs < 1783) {
if (Rs < 1706) {
return 8;
}
else if (Rs >= 1706) {
if (Rs < 1737) {
return 2;
}
else if (Rs >= 1737) {
return 4;
}
}
}
else if (Rs >= 1783) {
return 8;
}
}
else if (tcor >= 65) {
if (Cs < 1217) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (diffMVdiffVal < 5) {
if (mu_mv_mag_sq < 19) {
if (Rs < 1554) {
return 8;
}
else if (Rs >= 1554) {
return 2;
}
}
else if (mu_mv_mag_sq >= 19) {
return 8;
}
}
else if (diffMVdiffVal >= 5) {
if (Rs < 1614) {
return 8;
}
else if (Rs >= 1614) {
return 2;
}
}
}
}
else if (Cs >= 1217) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (MVDiff < 15) {
if (ssDCval < 90) {
if (tcor < 71) {
if (Rs < 1100) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (ssDCval < 85) {
if (diffRsCsdiff < 12) {
return 8;
}
else if (diffRsCsdiff >= 12) {
if (Cs < 748) {
return 8;
}
else if (Cs >= 748) {
return 4;
}
}
}
else if (ssDCval >= 85) {
if (Rs < 714) {
return 8;
}
else if (Rs >= 714) {
return 4;
}
}
}
}
else if (Rs >= 1100) {
if (posBalance < 67) {
if (ssDCval < 57) {
if (TSC < 9) {
if (Cs < 1268) {
return 4;
}
else if (Cs >= 1268) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (ssDCval >= 57) {
if (Cs < 1284) {
return 8;
}
else if (Cs >= 1284) {
if (Rs < 1428) {
return 4;
}
else if (Rs >= 1428) {
return 8;
}
}
}
}
else if (posBalance >= 67) {
if (Rs < 1775) {
return 4;
}
else if (Rs >= 1775) {
return 2;
}
}
}
}
else if (tcor >= 71) {
if (Cs < 675) {
if (posBalance < 60) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (posBalance >= 60) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
return 4;
}
}
}
}
else if (Cs >= 675) {
if (refDCval < 75) {
if (mu_mv_mag_sq < 2) {
if (negBalance < 50) {
if (Rs < 1020) {
return 1;
}
else if (Rs >= 1020) {
return 2;
}
}
else if (negBalance >= 50) {
if (MVDiff < 13) {
if (negBalance < 53) {
if (Rs < 1240) {
return 8;
}
else if (Rs >= 1240) {
return 4;
}
}
else if (negBalance >= 53) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 8;
}
}
}
else if (MVDiff >= 13) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1935) {
if (Rs < 804) {
if (Rs < 778) {
return 8;
}
else if (Rs >= 778) {
return 1;
}
}
else if (Rs >= 804) {
return 8;
}
}
else if (Rs >= 1935) {
return 2;
}
}
}
else if (refDCval >= 75) {
if (Cs < 1014) {
if (TSC < 7) {
if (Rs < 1498) {
if (Rs < 1063) {
return 8;
}
else if (Rs >= 1063) {
return 2;
}
}
else if (Rs >= 1498) {
return 1;
}
}
else if (TSC >= 7) {
if (negBalance < 48) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (negBalance >= 48) {
if (AFD < 18) {
if (Rs < 819) {
return 4;
}
else if (Rs >= 819) {
return 2;
}
}
else if (AFD >= 18) {
return 4;
}
}
}
}
else if (Cs >= 1014) {
if (mu_mv_mag_sq < 3) {
if (SC < 3148) {
return 4;
}
else if (SC >= 3148) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1217) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
else if (Rs >= 1217) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 90) {
if (negBalance < 58) {
if (AFD < 16) {
if (MVDiff < 8) {
if (diffAFD < 2) {
if (SC < 2081) {
if (Rs < 750) {
return 4;
}
else if (Rs >= 750) {
return 8;
}
}
else if (SC >= 2081) {
if (Rs < 1615) {
if (Cs < 886) {
return 8;
}
else if (Cs >= 886) {
return 4;
}
}
else if (Rs >= 1615) {
return 8;
}
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (MVDiff >= 8) {
if (posBalance < 59) {
if (RsDiff < 8) {
if (Rs < 1927) {
return 8;
}
else if (Rs >= 1927) {
return 4;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (posBalance >= 59) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
else if (AFD >= 16) {
if (Rs < 1101) {
if (Cs < 785) {
return 2;
}
else if (Cs >= 785) {
if (Cs < 1065) {
if (Cs < 817) {
return 4;
}
else if (Cs >= 817) {
return 8;
}
}
else if (Cs >= 1065) {
return 4;
}
}
}
else if (Rs >= 1101) {
if (tcor < 86) {
return 4;
}
else if (tcor >= 86) {
if (Rs < 1729) {
return 8;
}
else if (Rs >= 1729) {
return 1;
}
}
}
}
}
else if (negBalance >= 58) {
if (posBalance < 56) {
if (Rs < 1394) {
return 8;
}
else if (Rs >= 1394) {
return 2;
}
}
else if (posBalance >= 56) {
if (Rs < 841) {
return 1;
}
else if (Rs >= 841) {
if (mcTcor < 79) {
return 8;
}
else if (mcTcor >= 79) {
if (tcor < 75) {
return 4;
}
else if (tcor >= 75) {
if (Rs < 1247) {
return 4;
}
else if (Rs >= 1247) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 15) {
if (refDCval < 83) {
if (Rs < 1477) {
if (ssDCval < 71) {
if (ssDCval < 44) {
if (Rs < 431) {
if (Cs < 556) {
return 1;
}
else if (Cs >= 556) {
return 2;
}
}
else if (Rs >= 431) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (Rs < 732) {
return 8;
}
else if (Rs >= 732) {
return 2;
}
}
}
}
else if (ssDCval >= 44) {
if (Rs < 1239) {
if (MVDiff < 70) {
if (tcor < 51) {
if (AFD < 21) {
if (tcor < 50) {
return 4;
}
else if (tcor >= 50) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 1;
}
}
}
else if (AFD >= 21) {
if (Cs < 836) {
if (SC < 1643) {
return 8;
}
else if (SC >= 1643) {
return 4;
}
}
else if (Cs >= 836) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
if (Rs < 992) {
return 8;
}
else if (Rs >= 992) {
return 4;
}
}
}
}
}
else if (tcor >= 51) {
if (Cs < 526) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (Cs >= 526) {
if (Cs < 994) {
if (negBalance < 52) {
if (SC < 1191) {
if (Rs < 463) {
if (Rs < 328) {
return 1;
}
else if (Rs >= 328) {
return 4;
}
}
else if (Rs >= 463) {
return 2;
}
}
else if (SC >= 1191) {
if (AFD < 19) {
if (MVDiff < 32) {
if (Cs < 979) {
if (gchDC < 1) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (Cs >= 979) {
return 1;
}
}
else if (MVDiff >= 32) {
if (tcor < 78) {
if (AFD < 17) {
if (refDCval < 56) {
if (Rs < 833) {
return 8;
}
else if (Rs >= 833) {
return 4;
}
}
else if (refDCval >= 56) {
return 1;
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (tcor >= 78) {
return 4;
}
}
}
else if (AFD >= 19) {
if (Rs < 1062) {
if (Rs < 1033) {
return 1;
}
else if (Rs >= 1033) {
return 2;
}
}
else if (Rs >= 1062) {
return 8;
}
}
}
}
else if (negBalance >= 52) {
if (ssDCval < 61) {
if (RsDiff < 5) {
if (Rs < 235) {
return 1;
}
else if (Rs >= 235) {
return 8;
}
}
else if (RsDiff >= 5) {
if (tcor < 80) {
if (mu_mv_mag_sq < 9) {
if (Rs < 1221) {
return 4;
}
else if (Rs >= 1221) {
return 1;
}
}
else if (mu_mv_mag_sq >= 9) {
if (Rs < 671) {
return 4;
}
else if (Rs >= 671) {
return 8;
}
}
}
else if (tcor >= 80) {
if (Cs < 734) {
return 2;
}
else if (Cs >= 734) {
return 8;
}
}
}
}
else if (ssDCval >= 61) {
if (Cs < 842) {
if (mu_mv_mag_sq < 9) {
if (RsDiff < 6) {
if (refDCval < 66) {
if (Rs < 738) {
return 1;
}
else if (Rs >= 738) {
return 4;
}
}
else if (refDCval >= 66) {
return 4;
}
}
else if (RsDiff >= 6) {
if (Rs < 1049) {
return 1;
}
else if (Rs >= 1049) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (diffRsCsdiff < 4) {
if (CsDiff < 6) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (diffRsCsdiff >= 4) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
else if (Cs >= 842) {
if (Rs < 970) {
return 1;
}
else if (Rs >= 970) {
return 8;
}
}
}
}
}
else if (Cs >= 994) {
if (RsCsDiff < 57) {
if (Rs < 913) {
if (Cs < 1018) {
return 8;
}
else if (Cs >= 1018) {
return 2;
}
}
else if (Rs >= 913) {
return 8;
}
}
else if (RsCsDiff >= 57) {
if (diffMVdiffVal < 15) {
if (AFD < 17) {
if (tcor < 57) {
return 4;
}
else if (tcor >= 57) {
return 2;
}
}
else if (AFD >= 17) {
if (RsDiff < 7) {
if (Cs < 1243) {
if (RsDiff < 6) {
if (Rs < 678) {
return 8;
}
else if (Rs >= 678) {
return 4;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (Cs >= 1243) {
return 2;
}
}
else if (RsDiff >= 7) {
if (Cs < 1277) {
return 1;
}
else if (Cs >= 1277) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 15) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 70) {
if (mcTcor < 91) {
if (SC < 2123) {
if (refDCval < 70) {
if (diffMVdiffVal < 19) {
if (mu_mv_mag_sq < 2) {
if (Rs < 584) {
return 2;
}
else if (Rs >= 584) {
if (Rs < 652) {
return 1;
}
else if (Rs >= 652) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
else if (diffMVdiffVal >= 19) {
if (diffAFD < 5) {
if (diffAFD < 4) {
if (TSC < 9) {
if (Rs < 1028) {
return 4;
}
else if (Rs >= 1028) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (diffAFD >= 4) {
return 4;
}
}
else if (diffAFD >= 5) {
return 8;
}
}
}
else if (refDCval >= 70) {
if (AFD < 25) {
return 4;
}
else if (AFD >= 25) {
if (Rs < 610) {
return 2;
}
else if (Rs >= 610) {
return 1;
}
}
}
}
else if (SC >= 2123) {
if (mu_mv_mag_sq < 20) {
return 8;
}
else if (mu_mv_mag_sq >= 20) {
return 1;
}
}
}
else if (mcTcor >= 91) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
return 4;
}
}
}
}
else if (Rs >= 1239) {
if (RsCsDiff < 57) {
if (diffAFD < 8) {
return 8;
}
else if (diffAFD >= 8) {
return 1;
}
}
else if (RsCsDiff >= 57) {
if (SC < 2333) {
if (Cs < 788) {
if (CsDiff < 6) {
if (Rs < 1355) {
return 4;
}
else if (Rs >= 1355) {
return 8;
}
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (Cs >= 788) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 4;
}
}
}
}
else if (SC >= 2333) {
if (RsDiff < 9) {
if (diffRsCsdiff < 14) {
return 8;
}
else if (diffRsCsdiff >= 14) {
if (Cs < 1089) {
return 1;
}
else if (Cs >= 1089) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
if (Cs < 1233) {
return 2;
}
else if (Cs >= 1233) {
return 4;
}
}
}
}
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (CsDiff < 5) {
if (tcor < 84) {
if (Cs < 782) {
if (Rs < 700) {
if (Cs < 646) {
return 4;
}
else if (Cs >= 646) {
return 1;
}
}
else if (Rs >= 700) {
return 4;
}
}
else if (Cs >= 782) {
if (AFD < 16) {
if (MVDiff < 20) {
return 1;
}
else if (MVDiff >= 20) {
return 8;
}
}
else if (AFD >= 16) {
if (Rs < 1181) {
return 2;
}
else if (Rs >= 1181) {
return 4;
}
}
}
}
else if (tcor >= 84) {
if (Rs < 756) {
return 2;
}
else if (Rs >= 756) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (Cs < 1033) {
if (TSC < 12) {
if (diffMVdiffVal < -21) {
if (MVDiff < 80) {
if (Cs < 870) {
return 1;
}
else if (Cs >= 870) {
return 4;
}
}
else if (MVDiff >= 80) {
return 8;
}
}
else if (diffMVdiffVal >= -21) {
if (mcTcor < 95) {
if (diffRsCsdiff < -36) {
if (Cs < 988) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 4;
}
}
else if (Cs >= 988) {
return 1;
}
}
else if (diffRsCsdiff >= -36) {
if (diffMVdiffVal < 3) {
if (diffAFD < -2) {
if (diffMVdiffVal < -4) {
return 2;
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
else if (diffAFD >= -2) {
if (SC < 1356) {
return 4;
}
else if (SC >= 1356) {
if (Cs < 907) {
return 2;
}
else if (Cs >= 907) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= 3) {
if (diffMVdiffVal < 11) {
if (diffAFD < 2) {
if (Rs < 640) {
return 4;
}
else if (Rs >= 640) {
return 8;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (diffMVdiffVal >= 11) {
if (Rs < 696) {
return 8;
}
else if (Rs >= 696) {
return 2;
}
}
}
}
}
else if (mcTcor >= 95) {
if (MVDiff < 34) {
return 2;
}
else if (MVDiff >= 34) {
if (diffMVdiffVal < 8) {
return 8;
}
else if (diffMVdiffVal >= 8) {
return 1;
}
}
}
}
}
else if (TSC >= 12) {
if (Rs < 1204) {
return 8;
}
else if (Rs >= 1204) {
return 2;
}
}
}
else if (Cs >= 1033) {
if (mu_mv_mag_sq < 5) {
if (AFD < 23) {
if (Rs < 1422) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (Rs < 988) {
return 4;
}
else if (Rs >= 988) {
return 2;
}
}
}
else if (Rs >= 1422) {
if (Rs < 1465) {
return 8;
}
else if (Rs >= 1465) {
return 4;
}
}
}
else if (AFD >= 23) {
return 4;
}
}
else if (mu_mv_mag_sq >= 5) {
if (tcor < 67) {
if (diffTSC < -1) {
if (Rs < 1413) {
return 2;
}
else if (Rs >= 1413) {
return 4;
}
}
else if (diffTSC >= -1) {
if (Rs < 876) {
return 4;
}
else if (Rs >= 876) {
return 8;
}
}
}
else if (tcor >= 67) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1477) {
if (RsDiff < 10) {
if (posBalance < 46) {
if (posBalance < 38) {
if (Rs < 1554) {
return 8;
}
else if (Rs >= 1554) {
return 1;
}
}
else if (posBalance >= 38) {
return 2;
}
}
else if (posBalance >= 46) {
if (AFD < 23) {
if (diffMVdiffVal < 8) {
if (ssDCval < 59) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 8;
}
}
else if (ssDCval >= 59) {
if (diffRsCsdiff < -150) {
return 2;
}
else if (diffRsCsdiff >= -150) {
if (mu_mv_mag_sq < 54) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (Rs < 1879) {
return 1;
}
else if (Rs >= 1879) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 54) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 8) {
if (SC < 2564) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1611) {
return 8;
}
else if (Rs >= 1611) {
return 4;
}
}
}
else if (SC >= 2564) {
return 8;
}
}
}
else if (AFD >= 23) {
if (Rs < 1559) {
return 1;
}
else if (Rs >= 1559) {
return 4;
}
}
}
}
else if (RsDiff >= 10) {
if (AFD < 18) {
if (Cs < 1104) {
if (MVDiff < 31) {
return 2;
}
else if (MVDiff >= 31) {
if (tcor < 74) {
return 8;
}
else if (tcor >= 74) {
if (Rs < 2100) {
return 2;
}
else if (Rs >= 2100) {
return 1;
}
}
}
}
else if (Cs >= 1104) {
return 8;
}
}
else if (AFD >= 18) {
if (AFD < 27) {
if (mcTcor < 94) {
if (negBalance < 62) {
if (RsDiff < 15) {
if (Cs < 1053) {
if (refDCval < 76) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
return 4;
}
}
else if (refDCval >= 76) {
return 8;
}
}
else if (Cs >= 1053) {
if (diffRsCsdiff < 57) {
if (tcor < 76) {
return 4;
}
else if (tcor >= 76) {
if (Rs < 1947) {
return 1;
}
else if (Rs >= 1947) {
return 4;
}
}
}
else if (diffRsCsdiff >= 57) {
return 8;
}
}
}
else if (RsDiff >= 15) {
return 2;
}
}
else if (negBalance >= 62) {
if (Rs < 1923) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 8;
}
}
else if (Rs >= 1923) {
if (Rs < 1965) {
return 1;
}
else if (Rs >= 1965) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (AFD >= 27) {
if (Cs < 512) {
return 8;
}
else if (Cs >= 512) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 83) {
if (tcor < 62) {
if (ssDCval < 102) {
if (TSC < 9) {
if (Rs < 862) {
if (Rs < 350) {
return 4;
}
else if (Rs >= 350) {
return 1;
}
}
else if (Rs >= 862) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
else if (TSC >= 9) {
if (RsCsDiff < 124) {
if (mcTcor < 77) {
if (diffTSC < 2) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (diffTSC >= 2) {
return 8;
}
}
else if (mcTcor >= 77) {
if (diffAFD < -3) {
if (Rs < 465) {
return 4;
}
else if (Rs >= 465) {
return 8;
}
}
else if (diffAFD >= -3) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Cs < 727) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (Cs >= 727) {
if (posBalance < 65) {
return 4;
}
else if (posBalance >= 65) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 124) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (refDCval < 91) {
if (posBalance < 49) {
if (TSC < 17) {
return 1;
}
else if (TSC >= 17) {
return 4;
}
}
else if (posBalance >= 49) {
if (CsDiff < 15) {
return 4;
}
else if (CsDiff >= 15) {
return 8;
}
}
}
else if (refDCval >= 91) {
if (Rs < 683) {
return 1;
}
else if (Rs >= 683) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (diffTSC < -2) {
if (Rs < 1222) {
return 1;
}
else if (Rs >= 1222) {
return 2;
}
}
else if (diffTSC >= -2) {
if (RsCsDiff < 53) {
if (Rs < 340) {
return 4;
}
else if (Rs >= 340) {
return 1;
}
}
else if (RsCsDiff >= 53) {
if (CsDiff < 9) {
return 4;
}
else if (CsDiff >= 9) {
return 8;
}
}
}
}
}
else if (tcor >= 62) {
if (RsDiff < 12) {
if (posBalance < 52) {
if (Rs < 548) {
if (MVDiff < 93) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (MVDiff >= 93) {
return 1;
}
}
else if (Rs >= 548) {
if (MVDiff < 60) {
if (RsCsDiff < 89) {
if (diffMVdiffVal < -21) {
return 2;
}
else if (diffMVdiffVal >= -21) {
return 8;
}
}
else if (RsCsDiff >= 89) {
if (diffRsCsdiff < -18) {
return 8;
}
else if (diffRsCsdiff >= -18) {
if (Cs < 706) {
return 1;
}
else if (Cs >= 706) {
return 2;
}
}
}
}
else if (MVDiff >= 60) {
if (ssDCval < 73) {
return 4;
}
else if (ssDCval >= 73) {
return 8;
}
}
}
}
else if (posBalance >= 52) {
if (tcor < 78) {
if (MVDiff < 86) {
if (AFD < 27) {
if (ssDCval < 98) {
if (AFD < 19) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (RsCsDiff < 111) {
return 8;
}
else if (RsCsDiff >= 111) {
return 2;
}
}
}
else if (AFD >= 19) {
if (diffMVdiffVal < 9) {
if (RsCsDiff < 121) {
if (ssDCval < 93) {
if (Rs < 733) {
return 2;
}
else if (Rs >= 733) {
return 1;
}
}
else if (ssDCval >= 93) {
if (tcor < 67) {
return 1;
}
else if (tcor >= 67) {
if (MVDiff < 19) {
return 1;
}
else if (MVDiff >= 19) {
return 8;
}
}
}
}
else if (RsCsDiff >= 121) {
if (Cs < 756) {
return 4;
}
else if (Cs >= 756) {
if (AFD < 25) {
if (diffRsCsdiff < -35) {
return 1;
}
else if (diffRsCsdiff >= -35) {
return 8;
}
}
else if (AFD >= 25) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 9) {
if (Cs < 1178) {
if (MVDiff < 75) {
return 4;
}
else if (MVDiff >= 75) {
return 1;
}
}
else if (Cs >= 1178) {
return 2;
}
}
}
}
else if (ssDCval >= 98) {
if (AFD < 18) {
if (Rs < 1992) {
if (MVDiff < 35) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 1466) {
return 8;
}
else if (Rs >= 1466) {
return 4;
}
}
}
else if (MVDiff >= 35) {
return 1;
}
}
else if (Rs >= 1992) {
return 8;
}
}
else if (AFD >= 18) {
if (posBalance < 62) {
if (mcTcor < 95) {
if (diffTSC < 2) {
return 2;
}
else if (diffTSC >= 2) {
return 8;
}
}
else if (mcTcor >= 95) {
return 4;
}
}
else if (posBalance >= 62) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Rs < 966) {
return 1;
}
else if (Rs >= 966) {
return 4;
}
}
}
}
}
}
else if (AFD >= 27) {
if (posBalance < 65) {
return 4;
}
else if (posBalance >= 65) {
return 1;
}
}
}
else if (MVDiff >= 86) {
if (Rs < 877) {
if (Rs < 320) {
return 4;
}
else if (Rs >= 320) {
return 2;
}
}
else if (Rs >= 877) {
return 8;
}
}
}
else if (tcor >= 78) {
if (mcTcor < 92) {
if (diffRsCsdiff < -29) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 1;
}
}
}
else if (diffRsCsdiff >= -29) {
if (mu_mv_mag_sq < 19) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (Rs < 895) {
return 8;
}
else if (Rs >= 895) {
if (Cs < 1180) {
return 1;
}
else if (Cs >= 1180) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 19) {
if (Cs < 973) {
return 4;
}
else if (Cs >= 973) {
return 8;
}
}
}
}
else if (mcTcor >= 92) {
if (TSC < 9) {
if (mcTcor < 96) {
if (Cs < 571) {
return 1;
}
else if (Cs >= 571) {
if (diffRsCsdiff < -20) {
if (Cs < 598) {
return 2;
}
else if (Cs >= 598) {
return 4;
}
}
else if (diffRsCsdiff >= -20) {
if (RsDiff < 6) {
if (Rs < 672) {
if (Rs < 604) {
return 2;
}
else if (Rs >= 604) {
return 4;
}
}
else if (Rs >= 672) {
return 1;
}
}
else if (RsDiff >= 6) {
if (AFD < 14) {
if (Cs < 1001) {
return 8;
}
else if (Cs >= 1001) {
return 2;
}
}
else if (AFD >= 14) {
return 2;
}
}
}
}
}
else if (mcTcor >= 96) {
if (TSC < 8) {
if (Cs < 957) {
return 1;
}
else if (Cs >= 957) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (TSC >= 9) {
if (diffRsCsdiff < -47) {
if (Rs < 1093) {
return 8;
}
else if (Rs >= 1093) {
return 4;
}
}
else if (diffRsCsdiff >= -47) {
if (diffAFD < 2) {
if (refDCval < 85) {
if (Rs < 1111) {
return 8;
}
else if (Rs >= 1111) {
if (Rs < 1276) {
return 2;
}
else if (Rs >= 1276) {
return 1;
}
}
}
else if (refDCval >= 85) {
if (posBalance < 70) {
return 8;
}
else if (posBalance >= 70) {
return 2;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 1144) {
return 1;
}
else if (Rs >= 1144) {
return 2;
}
}
}
}
}
}
}
}
else if (RsDiff >= 12) {
if (mu_mv_mag_sq < 7) {
if (negBalance < 62) {
return 2;
}
else if (negBalance >= 62) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (refDCval < 90) {
if (Rs < 1528) {
return 2;
}
else if (Rs >= 1528) {
return 4;
}
}
else if (refDCval >= 90) {
if (Rs < 1759) {
if (AFD < 30) {
return 4;
}
else if (AFD >= 30) {
return 2;
}
}
else if (Rs >= 1759) {
if (AFD < 22) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (Rs < 1977) {
return 1;
}
else if (Rs >= 1977) {
return 4;
}
}
}
else if (AFD >= 22) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
if (Rs < 1919) {
return 1;
}
else if (Rs >= 1919) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (MVDiff < 24) {
if (negBalance < 59) {
if (CsDiff < 6) {
if (RsDiff < 8) {
if (Cs < 1277) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (Rs < 539) {
return 4;
}
else if (Rs >= 539) {
return 8;
}
}
}
}
else if (Cs >= 1277) {
return 1;
}
}
else if (RsDiff >= 8) {
if (Rs < 1547) {
return 4;
}
else if (Rs >= 1547) {
if (AFD < 29) {
if (Rs < 1655) {
return 8;
}
else if (Rs >= 1655) {
return 1;
}
}
else if (AFD >= 29) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (RsDiff < 5) {
if (Rs < 538) {
return 2;
}
else if (Rs >= 538) {
return 1;
}
}
else if (RsDiff >= 5) {
if (RsDiff < 9) {
if (posBalance < 57) {
if (Cs < 1187) {
if (Cs < 1068) {
return 8;
}
else if (Cs >= 1068) {
return 1;
}
}
else if (Cs >= 1187) {
return 8;
}
}
else if (posBalance >= 57) {
if (diffTSC < -1) {
if (posBalance < 71) {
if (Rs < 1042) {
return 2;
}
else if (Rs >= 1042) {
if (Rs < 1287) {
return 8;
}
else if (Rs >= 1287) {
return 4;
}
}
}
else if (posBalance >= 71) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (AFD < 16) {
if (Rs < 1810) {
return 8;
}
else if (Rs >= 1810) {
return 1;
}
}
else if (AFD >= 16) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (TSC < 9) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 59) {
if (Cs < 1025) {
if (SC < 2859) {
return 2;
}
else if (SC >= 2859) {
return 1;
}
}
else if (Cs >= 1025) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
else if (MVDiff >= 24) {
if (SC < 1846) {
if (MVDiff < 102) {
if (mcTcor < 80) {
if (Rs < 387) {
return 1;
}
else if (Rs >= 387) {
return 8;
}
}
else if (mcTcor >= 80) {
if (diffMVdiffVal < 5) {
if (SC < 1394) {
return 2;
}
else if (SC >= 1394) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 775) {
return 2;
}
else if (Rs >= 775) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
}
else if (MVDiff >= 102) {
if (diffMVdiffVal < 80) {
if (MVDiff < 222) {
return 1;
}
else if (MVDiff >= 222) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 1;
}
}
}
else if (diffMVdiffVal >= 80) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
if (Cs < 500) {
return 4;
}
else if (Cs >= 500) {
return 2;
}
}
}
}
}
else if (SC >= 1846) {
if (refDCval < 134) {
if (SC < 2662) {
if (posBalance < 58) {
if (diffTSC < -3) {
return 2;
}
else if (diffTSC >= -3) {
if (posBalance < 42) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (posBalance >= 42) {
return 1;
}
}
}
else if (posBalance >= 58) {
if (negBalance < 55) {
if (mcTcor < 94) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (diffMVdiffVal < 3) {
if (Cs < 716) {
return 8;
}
else if (Cs >= 716) {
return 1;
}
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (negBalance >= 55) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (Rs < 1644) {
return 4;
}
else if (Rs >= 1644) {
return 2;
}
}
}
}
}
else if (SC >= 2662) {
if (CsDiff < 5) {
if (ssDCval < 122) {
if (Rs < 2265) {
return 8;
}
else if (Rs >= 2265) {
return 1;
}
}
else if (ssDCval >= 122) {
return 1;
}
}
else if (CsDiff >= 5) {
if (mcTcor < 95) {
if (negBalance < 70) {
return 1;
}
else if (negBalance >= 70) {
return 4;
}
}
else if (mcTcor >= 95) {
if (Rs < 1967) {
return 1;
}
else if (Rs >= 1967) {
return 2;
}
}
}
}
}
else if (refDCval >= 134) {
if (mcTcor < 94) {
if (diffMVdiffVal < 74) {
if (MVDiff < 78) {
if (Cs < 1083) {
if (tcor < 74) {
if (Cs < 825) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
else if (Cs >= 825) {
return 4;
}
}
else if (tcor >= 74) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
}
else if (Cs >= 1083) {
if (Rs < 1585) {
return 1;
}
else if (Rs >= 1585) {
return 8;
}
}
}
else if (MVDiff >= 78) {
if (diffMVdiffVal < 26) {
if (RsCsDiff < 108) {
if (Rs < 924) {
return 1;
}
else if (Rs >= 924) {
return 8;
}
}
else if (RsCsDiff >= 108) {
return 1;
}
}
else if (diffMVdiffVal >= 26) {
if (AFD < 35) {
return 8;
}
else if (AFD >= 35) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 74) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (Cs < 854) {
if (Rs < 1583) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
if (Cs < 603) {
return 2;
}
else if (Cs >= 603) {
return 1;
}
}
}
else if (Rs >= 1583) {
return 1;
}
}
else if (Cs >= 854) {
if (RsCsDiff < 116) {
if (Rs < 1235) {
return 1;
}
else if (Rs >= 1235) {
return 8;
}
}
else if (RsCsDiff >= 116) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 88) {
if (SC < 2098) {
if (refDCval < 63) {
if (Rs < 547) {
if (SC < 1061) {
if (MVDiff < 3) {
if (ssDCval < 59) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 327) {
return 1;
}
else if (Rs >= 327) {
return 8;
}
}
}
else if (ssDCval >= 59) {
if (Rs < 268) {
return 4;
}
else if (Rs >= 268) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (AFD < 7) {
if (SC < 905) {
return 4;
}
else if (SC >= 905) {
if (mu_mv_mag_sq < 58) {
if (Rs < 294) {
return 1;
}
else if (Rs >= 294) {
return 4;
}
}
else if (mu_mv_mag_sq >= 58) {
return 1;
}
}
}
else if (AFD >= 7) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (Rs < 311) {
return 8;
}
else if (Rs >= 311) {
return 2;
}
}
}
}
}
else if (SC >= 1061) {
if (posBalance < 44) {
if (MVDiff < 4) {
if (ssDCval < 51) {
if (posBalance < 33) {
return 2;
}
else if (posBalance >= 33) {
if (Cs < 995) {
return 4;
}
else if (Cs >= 995) {
if (Cs < 1015) {
return 1;
}
else if (Cs >= 1015) {
return 4;
}
}
}
}
else if (ssDCval >= 51) {
if (Rs < 521) {
return 2;
}
else if (Rs >= 521) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 4;
}
}
}
else if (posBalance >= 44) {
if (mu_mv_mag_sq < 113) {
if (ssDCval < 55) {
if (Rs < 389) {
return 2;
}
else if (Rs >= 389) {
return 4;
}
}
else if (ssDCval >= 55) {
if (Cs < 788) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (Cs < 621) {
return 2;
}
else if (Cs >= 621) {
return 8;
}
}
}
else if (Cs >= 788) {
if (SC < 1285) {
return 4;
}
else if (SC >= 1285) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
return 4;
}
}
}
}
else if (Rs >= 547) {
if (MVDiff < 2) {
if (SC < 1571) {
if (refDCval < 54) {
if (negBalance < 32) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (Cs < 550) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 4;
}
}
else if (Cs >= 550) {
if (SC < 1379) {
return 2;
}
else if (SC >= 1379) {
return 8;
}
}
}
}
else if (negBalance >= 32) {
return 4;
}
}
else if (refDCval >= 54) {
return 8;
}
}
else if (SC >= 1571) {
if (ssDCval < 49) {
if (Rs < 1175) {
return 8;
}
else if (Rs >= 1175) {
if (Rs < 1345) {
if (Rs < 1326) {
return 4;
}
else if (Rs >= 1326) {
return 8;
}
}
else if (Rs >= 1345) {
return 8;
}
}
}
else if (ssDCval >= 49) {
if (Rs < 633) {
if (Cs < 1067) {
return 1;
}
else if (Cs >= 1067) {
return 4;
}
}
else if (Rs >= 633) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 51) {
if (Rs < 896) {
if (mu_mv_mag_sq < 125) {
return 1;
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
else if (Rs >= 896) {
if (Rs < 1155) {
return 8;
}
else if (Rs >= 1155) {
if (Rs < 1251) {
return 4;
}
else if (Rs >= 1251) {
return 1;
}
}
}
}
else if (ssDCval >= 51) {
if (Rs < 826) {
if (Rs < 723) {
return 1;
}
else if (Rs >= 723) {
if (negBalance < 28) {
if (SC < 1618) {
return 8;
}
else if (SC >= 1618) {
return 4;
}
}
else if (negBalance >= 28) {
return 8;
}
}
}
else if (Rs >= 826) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (SC < 2020) {
return 1;
}
else if (SC >= 2020) {
if (AFD < 4) {
if (negBalance < 14) {
return 8;
}
else if (negBalance >= 14) {
return 1;
}
}
else if (AFD >= 4) {
if (SC < 2035) {
return 8;
}
else if (SC >= 2035) {
if (SC < 2062) {
return 1;
}
else if (SC >= 2062) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 924) {
if (negBalance < 16) {
if (mu_mv_mag_sq < 128) {
if (Cs < 810) {
return 1;
}
else if (Cs >= 810) {
return 8;
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (negBalance >= 16) {
if (refDCval < 62) {
return 8;
}
else if (refDCval >= 62) {
return 2;
}
}
}
else if (Cs >= 924) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (Rs < 902) {
if (negBalance < 17) {
if (Cs < 796) {
return 2;
}
else if (Cs >= 796) {
return 1;
}
}
else if (negBalance >= 17) {
if (Cs < 716) {
if (SC < 1457) {
if (MVDiff < 13) {
if (negBalance < 37) {
if (mcTcor < 98) {
if (diffRsCsdiff < 4) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (diffRsCsdiff >= 4) {
return 1;
}
}
else if (mcTcor >= 98) {
if (Cs < 527) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
return 4;
}
}
else if (Cs >= 527) {
return 4;
}
}
}
else if (negBalance >= 37) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (MVDiff >= 13) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (SC >= 1457) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (Cs >= 716) {
if (AFD < 9) {
if (TSC < 3) {
if (Cs < 850) {
if (Cs < 780) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
return 4;
}
}
else if (Cs >= 780) {
if (refDCval < 54) {
return 8;
}
else if (refDCval >= 54) {
return 4;
}
}
}
else if (Cs >= 850) {
if (Rs < 767) {
return 1;
}
else if (Rs >= 767) {
return 8;
}
}
}
else if (TSC >= 3) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
return 8;
}
}
}
}
else if (AFD >= 9) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
return 8;
}
}
}
}
}
else if (Rs >= 902) {
if (ssDCval < 52) {
if (posBalance < 43) {
if (TSC < 2) {
if (Cs < 920) {
return 8;
}
else if (Cs >= 920) {
if (Rs < 956) {
return 8;
}
else if (Rs >= 956) {
return 1;
}
}
}
else if (TSC >= 2) {
if (posBalance < 36) {
if (SC < 1877) {
if (Cs < 585) {
return 2;
}
else if (Cs >= 585) {
return 8;
}
}
else if (SC >= 1877) {
return 1;
}
}
else if (posBalance >= 36) {
return 1;
}
}
}
else if (posBalance >= 43) {
if (SC < 1943) {
if (Cs < 767) {
return 1;
}
else if (Cs >= 767) {
return 8;
}
}
else if (SC >= 1943) {
return 4;
}
}
}
else if (ssDCval >= 52) {
if (TSC < 5) {
if (MVDiff < 13) {
if (tcor < 98) {
if (diffRsCsdiff < -4) {
if (Cs < 805) {
return 1;
}
else if (Cs >= 805) {
if (Rs < 972) {
return 4;
}
else if (Rs >= 972) {
return 1;
}
}
}
else if (diffRsCsdiff >= -4) {
return 1;
}
}
else if (tcor >= 98) {
if (SC < 2007) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (SC >= 2007) {
if (SC < 2074) {
return 8;
}
else if (SC >= 2074) {
return 1;
}
}
}
}
else if (MVDiff >= 13) {
if (ssDCval < 61) {
if (Cs < 778) {
return 1;
}
else if (Cs >= 778) {
return 8;
}
}
else if (ssDCval >= 61) {
if (Rs < 1122) {
return 4;
}
else if (Rs >= 1122) {
return 2;
}
}
}
}
else if (TSC >= 5) {
if (CsDiff < 4) {
if (Rs < 980) {
return 1;
}
else if (Rs >= 980) {
return 8;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
}
}
}
}
else if (refDCval >= 63) {
if (Rs < 1086) {
if (Rs < 367) {
if (AFD < 13) {
if (mu_mv_mag_sq < 88) {
if (diffRsCsdiff < -21) {
return 2;
}
else if (diffRsCsdiff >= -21) {
return 1;
}
}
else if (mu_mv_mag_sq >= 88) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (Cs < 693) {
return 1;
}
else if (Cs >= 693) {
return 8;
}
}
}
}
else if (AFD >= 13) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Rs < 164) {
return 1;
}
else if (Rs >= 164) {
return 8;
}
}
}
}
else if (Rs >= 367) {
if (RsCsDiff < 2) {
if (Cs < 704) {
if (TSC < 1) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
return 8;
}
}
else if (TSC >= 1) {
if (Cs < 583) {
if (negBalance < 23) {
if (Rs < 700) {
return 1;
}
else if (Rs >= 700) {
return 2;
}
}
else if (negBalance >= 23) {
if (Rs < 712) {
if (Rs < 546) {
return 2;
}
else if (Rs >= 546) {
return 1;
}
}
else if (Rs >= 712) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
if (Cs < 532) {
return 8;
}
else if (Cs >= 532) {
return 2;
}
}
}
}
}
}
else if (Cs >= 583) {
if (Rs < 555) {
if (Rs < 433) {
return 4;
}
else if (Rs >= 433) {
if (Rs < 483) {
return 2;
}
else if (Rs >= 483) {
return 8;
}
}
}
else if (Rs >= 555) {
if (mcTcor < 99) {
if (Rs < 945) {
if (Rs < 749) {
return 1;
}
else if (Rs >= 749) {
return 8;
}
}
else if (Rs >= 945) {
return 4;
}
}
else if (mcTcor >= 99) {
if (posBalance < 38) {
if (Rs < 904) {
return 2;
}
else if (Rs >= 904) {
return 8;
}
}
else if (posBalance >= 38) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (Cs < 697) {
return 2;
}
else if (Cs >= 697) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 704) {
if (ssDCval < 72) {
if (Rs < 891) {
if (Rs < 718) {
if (Cs < 764) {
if (mu_mv_mag_sq < 123) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
else if (Cs >= 764) {
return 8;
}
}
else if (Rs >= 718) {
if (Cs < 939) {
return 4;
}
else if (Cs >= 939) {
return 8;
}
}
}
else if (Rs >= 891) {
if (SC < 1816) {
return 2;
}
else if (SC >= 1816) {
if (Cs < 1008) {
return 1;
}
else if (Cs >= 1008) {
return 8;
}
}
}
}
else if (ssDCval >= 72) {
if (Rs < 647) {
if (ssDCval < 84) {
if (Rs < 523) {
return 8;
}
else if (Rs >= 523) {
return 1;
}
}
else if (ssDCval >= 84) {
if (Cs < 716) {
if (Rs < 560) {
return 2;
}
else if (Rs >= 560) {
return 1;
}
}
else if (Cs >= 716) {
return 2;
}
}
}
else if (Rs >= 647) {
if (TSCindex < 3) {
if (Rs < 658) {
if (Rs < 653) {
return 1;
}
else if (Rs >= 653) {
return 2;
}
}
else if (Rs >= 658) {
return 1;
}
}
else if (TSCindex >= 3) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 2) {
if (refDCval < 75) {
if (diffRsCsdiff < 1) {
if (Rs < 461) {
if (AFD < 8) {
if (Rs < 441) {
return 4;
}
else if (Rs >= 441) {
return 2;
}
}
else if (AFD >= 8) {
if (Cs < 624) {
return 1;
}
else if (Cs >= 624) {
return 2;
}
}
}
else if (Rs >= 461) {
if (AFD < 4) {
if (Cs < 661) {
if (ssDCval < 68) {
if (Rs < 613) {
return 8;
}
else if (Rs >= 613) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
}
else if (ssDCval >= 68) {
if (Cs < 538) {
return 8;
}
else if (Cs >= 538) {
return 2;
}
}
}
else if (Cs >= 661) {
if (Rs < 665) {
return 8;
}
else if (Rs >= 665) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
if (Rs < 804) {
if (Cs < 965) {
return 1;
}
else if (Cs >= 965) {
return 2;
}
}
else if (Rs >= 804) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (Rs < 842) {
return 4;
}
else if (Rs >= 842) {
return 1;
}
}
}
}
}
}
}
else if (AFD >= 4) {
if (RsCsDiff < 7) {
if (ssDCval < 67) {
if (SC < 1248) {
return 4;
}
else if (SC >= 1248) {
if (Rs < 1035) {
if (Cs < 624) {
if (Rs < 676) {
return 1;
}
else if (Rs >= 676) {
return 8;
}
}
else if (Cs >= 624) {
return 8;
}
}
else if (Rs >= 1035) {
return 1;
}
}
}
else if (ssDCval >= 67) {
if (posBalance < 80) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
if (Rs < 732) {
if (Rs < 620) {
return 8;
}
else if (Rs >= 620) {
return 1;
}
}
else if (Rs >= 732) {
return 4;
}
}
}
else if (posBalance >= 80) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (SC < 1815) {
if (tcor < 94) {
if (diffAFD < -1) {
if (Rs < 658) {
return 1;
}
else if (Rs >= 658) {
return 4;
}
}
else if (diffAFD >= -1) {
if (negBalance < 37) {
return 1;
}
else if (negBalance >= 37) {
if (Rs < 574) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (Rs >= 574) {
return 2;
}
}
}
}
else if (tcor >= 94) {
if (diffAFD < -2) {
if (Cs < 593) {
return 1;
}
else if (Cs >= 593) {
return 8;
}
}
else if (diffAFD >= -2) {
if (RsCsDiff < 16) {
return 4;
}
else if (RsCsDiff >= 16) {
if (Rs < 638) {
return 4;
}
else if (Rs >= 638) {
return 2;
}
}
}
}
}
else if (SC >= 1815) {
if (ssDCval < 66) {
if (Rs < 826) {
return 2;
}
else if (Rs >= 826) {
return 4;
}
}
else if (ssDCval >= 66) {
if (mcTcor < 98) {
if (SC < 2072) {
if (negBalance < 31) {
if (Rs < 930) {
return 2;
}
else if (Rs >= 930) {
return 8;
}
}
else if (negBalance >= 31) {
return 8;
}
}
else if (SC >= 2072) {
return 2;
}
}
else if (mcTcor >= 98) {
return 2;
}
}
}
}
}
}
}
else if (diffRsCsdiff >= 1) {
if (refDCval < 73) {
if (posBalance < 44) {
if (TSC < 3) {
if (Cs < 574) {
return 1;
}
else if (Cs >= 574) {
return 4;
}
}
else if (TSC >= 3) {
if (posBalance < 39) {
if (diffMVdiffVal < -2) {
return 2;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 424) {
return 4;
}
else if (Rs >= 424) {
return 8;
}
}
}
else if (posBalance >= 39) {
return 1;
}
}
}
else if (posBalance >= 44) {
if (tcor < 98) {
if (Cs < 993) {
if (diffMVdiffVal < -1) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
if (CsDiff < 3) {
if (Cs < 787) {
return 1;
}
else if (Cs >= 787) {
return 4;
}
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
if (TSC < 4) {
if (SC < 1488) {
return 8;
}
else if (SC >= 1488) {
return 1;
}
}
else if (TSC >= 4) {
if (Cs < 917) {
return 8;
}
else if (Cs >= 917) {
if (Rs < 755) {
return 2;
}
else if (Rs >= 755) {
return 8;
}
}
}
}
}
}
else if (Cs >= 993) {
if (CsDiff < 4) {
if (posBalance < 56) {
if (Rs < 846) {
return 1;
}
else if (Rs >= 846) {
return 4;
}
}
else if (posBalance >= 56) {
return 2;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (tcor >= 98) {
if (CsDiff < 2) {
if (Cs < 750) {
if (Rs < 711) {
if (Rs < 578) {
return 2;
}
else if (Rs >= 578) {
return 1;
}
}
else if (Rs >= 711) {
return 2;
}
}
else if (Cs >= 750) {
if (Rs < 714) {
return 8;
}
else if (Rs >= 714) {
return 1;
}
}
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
else if (refDCval >= 73) {
if (Rs < 599) {
return 8;
}
else if (Rs >= 599) {
return 1;
}
}
}
}
else if (refDCval >= 75) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 77) {
if (SC < 1358) {
if (Cs < 526) {
return 1;
}
else if (Cs >= 526) {
return 8;
}
}
else if (SC >= 1358) {
if (tcor < 90) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Cs < 851) {
return 4;
}
else if (Cs >= 851) {
return 1;
}
}
}
else if (tcor >= 90) {
if (MVDiff < 2) {
if (Rs < 863) {
return 4;
}
else if (Rs >= 863) {
return 8;
}
}
else if (MVDiff >= 2) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
if (mu_mv_mag_sq < 116) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (mu_mv_mag_sq >= 116) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
return 4;
}
}
}
}
}
else if (ssDCval >= 84) {
if (Cs < 893) {
if (TSC < 4) {
if (Cs < 716) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
if (Rs < 532) {
return 2;
}
else if (Rs >= 532) {
return 1;
}
}
}
else if (Cs >= 716) {
return 2;
}
}
else if (TSC >= 4) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Rs < 1048) {
return 8;
}
else if (Rs >= 1048) {
return 1;
}
}
}
}
else if (Cs >= 893) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1086) {
if (mu_mv_mag_sq < 128) {
if (refDCval < 69) {
if (posBalance < 44) {
if (Cs < 503) {
if (negBalance < 20) {
return 2;
}
else if (negBalance >= 20) {
if (Rs < 1464) {
return 8;
}
else if (Rs >= 1464) {
return 1;
}
}
}
else if (Cs >= 503) {
return 1;
}
}
else if (posBalance >= 44) {
if (CsDiff < 1) {
if (Cs < 682) {
return 2;
}
else if (Cs >= 682) {
return 4;
}
}
else if (CsDiff >= 1) {
if (MVDiff < 8) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (SC < 1760) {
return 1;
}
else if (SC >= 1760) {
if (Rs < 1233) {
return 2;
}
else if (Rs >= 1233) {
return 8;
}
}
}
}
else if (MVDiff >= 8) {
if (Rs < 1117) {
return 2;
}
else if (Rs >= 1117) {
return 8;
}
}
}
}
}
else if (refDCval >= 69) {
if (refDCval < 72) {
if (mu_mv_mag_sq < 53) {
if (SC < 1944) {
return 1;
}
else if (SC >= 1944) {
return 4;
}
}
else if (mu_mv_mag_sq >= 53) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (SC < 1753) {
return 8;
}
else if (SC >= 1753) {
if (Rs < 1215) {
return 2;
}
else if (Rs >= 1215) {
return 1;
}
}
}
}
}
else if (refDCval >= 72) {
if (Cs < 765) {
if (Rs < 1305) {
return 1;
}
else if (Rs >= 1305) {
if (CsDiff < 1) {
if (Cs < 645) {
return 8;
}
else if (Cs >= 645) {
return 1;
}
}
else if (CsDiff >= 1) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 8;
}
}
}
}
else if (Cs >= 765) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (Rs < 1163) {
return 2;
}
else if (Rs >= 1163) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 698) {
return 8;
}
else if (Cs >= 698) {
if (Rs < 1173) {
return 4;
}
else if (Rs >= 1173) {
return 1;
}
}
}
}
}
}
else if (SC >= 2098) {
if (Rs < 1761) {
if (Cs < 1081) {
if (mu_mv_mag_sq < 122) {
if (SC < 2453) {
if (TSC < 5) {
if (Rs < 1272) {
if (ssDCval < 67) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (MVDiff < 7) {
if (Rs < 1263) {
return 1;
}
else if (Rs >= 1263) {
return 8;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (Rs >= 1272) {
if (Rs < 1691) {
if (refDCval < 64) {
if (refDCval < 54) {
if (ssDCval < 43) {
return 4;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (refDCval >= 54) {
if (Cs < 756) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 756) {
if (diffRsCsdiff < -4) {
return 4;
}
else if (diffRsCsdiff >= -4) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (Rs < 1324) {
return 8;
}
else if (Rs >= 1324) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 64) {
if (refDCval < 69) {
return 1;
}
else if (refDCval >= 69) {
if (mu_mv_mag_sq < 78) {
return 1;
}
else if (mu_mv_mag_sq >= 78) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (diffAFD < -1) {
if (Rs < 1399) {
return 8;
}
else if (Rs >= 1399) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1691) {
if (Rs < 1700) {
return 2;
}
else if (Rs >= 1700) {
if (mu_mv_mag_sq < 114) {
return 1;
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
}
}
}
else if (TSC >= 5) {
if (AFD < 9) {
if (negBalance < 68) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (negBalance >= 68) {
return 4;
}
}
else if (AFD >= 9) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 2;
}
}
}
}
else if (SC >= 2453) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (Cs < 852) {
return 4;
}
else if (Cs >= 852) {
if (tcor < 95) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (Rs < 1596) {
return 2;
}
else if (Rs >= 1596) {
return 1;
}
}
}
else if (tcor >= 95) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (TSCindex < 2) {
if (SC < 2214) {
return 8;
}
else if (SC >= 2214) {
if (RsDiff < 1) {
if (ssDCval < 68) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (ssDCval >= 68) {
return 8;
}
}
else if (RsDiff >= 1) {
if (Rs < 1331) {
return 1;
}
else if (Rs >= 1331) {
return 8;
}
}
}
}
else if (TSCindex >= 2) {
if (ssDCval < 49) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (ssDCval < 48) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (diffRsCsdiff < 2) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
}
}
else if (ssDCval >= 49) {
if (Rs < 1244) {
if (posBalance < 39) {
if (Cs < 884) {
return 8;
}
else if (Cs >= 884) {
return 1;
}
}
else if (posBalance >= 39) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (Rs < 1237) {
return 4;
}
else if (Rs >= 1237) {
return 8;
}
}
}
}
else if (Rs >= 1244) {
if (Rs < 1753) {
if (Rs < 1335) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (negBalance < 20) {
if (Cs < 877) {
return 8;
}
else if (Cs >= 877) {
return 1;
}
}
else if (negBalance >= 20) {
return 8;
}
}
}
else if (Rs >= 1335) {
return 8;
}
}
else if (Rs >= 1753) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1081) {
if (RsDiff < 1) {
if (mu_mv_mag_sq < 104) {
if (Cs < 1139) {
return 4;
}
else if (Cs >= 1139) {
return 8;
}
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 1436) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
if (Rs < 1132) {
return 4;
}
else if (Rs >= 1132) {
return 8;
}
}
}
else if (Rs >= 1436) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
return 4;
}
}
}
}
else if (RsDiff >= 1) {
if (tcor < 98) {
if (Rs < 990) {
if (SC < 2182) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (Rs < 928) {
return 2;
}
else if (Rs >= 928) {
return 4;
}
}
}
else if (SC >= 2182) {
return 2;
}
}
else if (Rs >= 990) {
if (ssDCval < 72) {
if (RsDiff < 2) {
if (Rs < 1521) {
if (Rs < 1166) {
return 8;
}
else if (Rs >= 1166) {
if (mu_mv_mag_sq < 113) {
return 8;
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 1365) {
return 1;
}
else if (Rs >= 1365) {
return 8;
}
}
}
}
else if (Rs >= 1521) {
return 4;
}
}
else if (RsDiff >= 2) {
if (Rs < 1357) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (refDCval < 67) {
if (Rs < 1286) {
if (CsDiff < 2) {
if (Rs < 1126) {
return 4;
}
else if (Rs >= 1126) {
return 1;
}
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (Rs >= 1286) {
if (Rs < 1309) {
return 8;
}
else if (Rs >= 1309) {
return 4;
}
}
}
else if (refDCval >= 67) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
}
else if (Rs >= 1357) {
if (MVDiff < 3) {
if (ssDCval < 49) {
if (Cs < 1134) {
return 4;
}
else if (Cs >= 1134) {
if (Rs < 1726) {
return 8;
}
else if (Rs >= 1726) {
return 4;
}
}
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Rs < 1557) {
return 2;
}
else if (Rs >= 1557) {
return 1;
}
}
}
}
}
else if (ssDCval >= 72) {
if (mu_mv_mag_sq < 66) {
if (Cs < 1178) {
return 8;
}
else if (Cs >= 1178) {
return 2;
}
}
else if (mu_mv_mag_sq >= 66) {
return 8;
}
}
}
}
else if (tcor >= 98) {
if (SC < 2657) {
if (MVDiff < 6) {
if (refDCval < 70) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 55) {
if (Rs < 1154) {
return 8;
}
else if (Rs >= 1154) {
return 4;
}
}
else if (ssDCval >= 55) {
if (diffRsCsdiff < -12) {
return 4;
}
else if (diffRsCsdiff >= -12) {
if (Rs < 1374) {
return 1;
}
else if (Rs >= 1374) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
}
else if (refDCval >= 70) {
if (Cs < 1122) {
if (diffAFD < 1) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (Cs >= 1122) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
return 4;
}
}
}
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (SC >= 2657) {
if (AFD < 3) {
if (SC < 2917) {
if (posBalance < 46) {
if (ssDCval < 49) {
if (negBalance < 18) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 1742) {
return 4;
}
else if (Rs >= 1742) {
return 4;
}
}
}
else if (negBalance >= 18) {
if (SC < 2901) {
if (Rs < 1743) {
if (SC < 2865) {
if (SC < 2770) {
if (Rs < 1524) {
return 8;
}
else if (Rs >= 1524) {
if (Rs < 1646) {
return 4;
}
else if (Rs >= 1646) {
if (Rs < 1663) {
return 8;
}
else if (Rs >= 1663) {
return 8;
}
}
}
}
else if (SC >= 2770) {
if (Cs < 1112) {
if (AFD < 2) {
if (Rs < 1737) {
if (Rs < 1711) {
return 4;
}
else if (Rs >= 1711) {
return 4;
}
}
else if (Rs >= 1737) {
return 4;
}
}
else if (AFD >= 2) {
return 4;
}
}
else if (Cs >= 1112) {
if (negBalance < 22) {
if (CsDiff < 2) {
if (Rs < 1729) {
if (Cs < 1113) {
return 8;
}
else if (Cs >= 1113) {
if (Rs < 1716) {
if (Cs < 1164) {
if (Rs < 1689) {
if (Cs < 1157) {
if (Rs < 1681) {
return 4;
}
else if (Rs >= 1681) {
return 4;
}
}
else if (Cs >= 1157) {
return 4;
}
}
else if (Rs >= 1689) {
return 4;
}
}
else if (Cs >= 1164) {
return 4;
}
}
else if (Rs >= 1716) {
return 4;
}
}
}
else if (Rs >= 1729) {
return 8;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (negBalance >= 22) {
if (Rs < 1638) {
return 4;
}
else if (Rs >= 1638) {
return 4;
}
}
}
}
}
else if (SC >= 2865) {
return 4;
}
}
else if (Rs >= 1743) {
if (Cs < 1095) {
return 8;
}
else if (Cs >= 1095) {
if (Rs < 1756) {
if (Cs < 1114) {
return 8;
}
else if (Cs >= 1114) {
return 8;
}
}
else if (Rs >= 1756) {
return 4;
}
}
}
}
else if (SC >= 2901) {
return 8;
}
}
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (posBalance >= 46) {
return 4;
}
}
else if (SC >= 2917) {
return 4;
}
}
else if (AFD >= 3) {
if (Cs < 1250) {
return 8;
}
else if (Cs >= 1250) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1761) {
if (TSC < 4) {
if (refDCval < 82) {
if (ssDCval < 53) {
if (mu_mv_mag_sq < 95) {
return 4;
}
else if (mu_mv_mag_sq >= 95) {
if (Cs < 1129) {
if (Cs < 935) {
return 8;
}
else if (Cs >= 935) {
if (posBalance < 34) {
if (Rs < 1767) {
return 4;
}
else if (Rs >= 1767) {
if (Cs < 1096) {
if (Rs < 1809) {
if (Rs < 1790) {
if (Rs < 1776) {
return 4;
}
else if (Rs >= 1776) {
return 4;
}
}
else if (Rs >= 1790) {
return 4;
}
}
else if (Rs >= 1809) {
return 4;
}
}
else if (Cs >= 1096) {
if (Rs < 1810) {
return 8;
}
else if (Rs >= 1810) {
return 4;
}
}
}
}
else if (posBalance >= 34) {
if (Cs < 1052) {
return 8;
}
else if (Cs >= 1052) {
if (SC < 2875) {
if (Rs < 1772) {
return 8;
}
else if (Rs >= 1772) {
if (Rs < 1794) {
return 4;
}
else if (Rs >= 1794) {
return 8;
}
}
}
else if (SC >= 2875) {
if (Rs < 1814) {
if (Rs < 1788) {
return 8;
}
else if (Rs >= 1788) {
return 8;
}
}
else if (Rs >= 1814) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1129) {
if (posBalance < 38) {
if (SC < 2920) {
if (SC < 2913) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 4;
}
}
else if (SC >= 2913) {
return 4;
}
}
else if (SC >= 2920) {
return 4;
}
}
else if (posBalance >= 38) {
if (Rs < 1969) {
if (Cs < 1185) {
if (Cs < 1150) {
return 4;
}
else if (Cs >= 1150) {
if (Rs < 1795) {
return 8;
}
else if (Rs >= 1795) {
return 4;
}
}
}
else if (Cs >= 1185) {
return 4;
}
}
else if (Rs >= 1969) {
return 8;
}
}
}
}
}
else if (ssDCval >= 53) {
if (refDCval < 71) {
if (MVDiff < 1) {
if (SC < 2865) {
if (Cs < 990) {
return 1;
}
else if (Cs >= 990) {
return 4;
}
}
else if (SC >= 2865) {
if (tcor < 99) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 1878) {
return 1;
}
else if (Rs >= 1878) {
return 4;
}
}
}
else if (tcor >= 99) {
if (Rs < 1870) {
if (Cs < 1045) {
return 4;
}
else if (Cs >= 1045) {
if (Cs < 1103) {
return 1;
}
else if (Cs >= 1103) {
return 4;
}
}
}
else if (Rs >= 1870) {
if (Rs < 2062) {
if (diffRsCsdiff < 1) {
return 4;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 1941) {
return 1;
}
else if (Rs >= 1941) {
return 4;
}
}
}
else if (Rs >= 2062) {
if (Rs < 2187) {
return 1;
}
else if (Rs >= 2187) {
if (Rs < 2282) {
return 8;
}
else if (Rs >= 2282) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (mcTcor < 99) {
if (gchDC < 1) {
if (Rs < 1875) {
if (Rs < 1773) {
return 8;
}
else if (Rs >= 1773) {
if (Rs < 1810) {
return 1;
}
else if (Rs >= 1810) {
return 2;
}
}
}
else if (Rs >= 1875) {
return 1;
}
}
else if (gchDC >= 1) {
if (negBalance < 28) {
if (Cs < 1160) {
return 4;
}
else if (Cs >= 1160) {
return 2;
}
}
else if (negBalance >= 28) {
if (Cs < 890) {
return 1;
}
else if (Cs >= 890) {
return 8;
}
}
}
}
else if (mcTcor >= 99) {
if (mu_mv_mag_sq < 108) {
return 4;
}
else if (mu_mv_mag_sq >= 108) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 4;
}
}
}
}
}
else if (refDCval >= 71) {
if (posBalance < 43) {
if (Rs < 2096) {
return 1;
}
else if (Rs >= 2096) {
return 8;
}
}
else if (posBalance >= 43) {
if (SC < 2514) {
if (Rs < 1956) {
return 1;
}
else if (Rs >= 1956) {
return 8;
}
}
else if (SC >= 2514) {
if (diffRsCsdiff < -10) {
return 1;
}
else if (diffRsCsdiff >= -10) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 82) {
if (mu_mv_mag_sq < 100) {
if (RsDiff < 3) {
if (Rs < 2071) {
return 4;
}
else if (Rs >= 2071) {
return 8;
}
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 900) {
return 4;
}
else if (Cs >= 900) {
return 8;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 966) {
if (ssDCval < 71) {
if (refDCval < 61) {
return 1;
}
else if (refDCval >= 61) {
if (Rs < 1955) {
if (Rs < 1825) {
return 8;
}
else if (Rs >= 1825) {
return 4;
}
}
else if (Rs >= 1955) {
if (refDCval < 71) {
return 2;
}
else if (refDCval >= 71) {
if (Rs < 2064) {
return 2;
}
else if (Rs >= 2064) {
return 1;
}
}
}
}
}
else if (ssDCval >= 71) {
if (mu_mv_mag_sq < 93) {
if (MVDiff < 22) {
return 1;
}
else if (MVDiff >= 22) {
if (SC < 2626) {
if (Rs < 1973) {
return 2;
}
else if (Rs >= 1973) {
return 4;
}
}
else if (SC >= 2626) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (Rs < 2276) {
if (Rs < 2055) {
return 4;
}
else if (Rs >= 2055) {
if (Rs < 2256) {
return 8;
}
else if (Rs >= 2256) {
return 8;
}
}
}
else if (Rs >= 2276) {
return 2;
}
}
}
}
else if (Cs >= 966) {
if (negBalance < 45) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (AFD < 16) {
if (Cs < 1003) {
return 4;
}
else if (Cs >= 1003) {
if (Cs < 1295) {
return 1;
}
else if (Cs >= 1295) {
return 4;
}
}
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (negBalance >= 45) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 88) {
if (mu_mv_mag_sq < 98) {
if (SC < 1411) {
if (negBalance < 70) {
if (Rs < 422) {
if (refDCval < 115) {
if (ssDCval < 104) {
if (TSC < 6) {
if (ssDCval < 96) {
if (Rs < 342) {
return 1;
}
else if (Rs >= 342) {
return 8;
}
}
else if (ssDCval >= 96) {
return 2;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (ssDCval >= 104) {
if (Cs < 702) {
if (Rs < 372) {
return 1;
}
else if (Rs >= 372) {
return 2;
}
}
else if (Cs >= 702) {
return 4;
}
}
}
else if (refDCval >= 115) {
if (ssDCval < 159) {
if (posBalance < 37) {
return 4;
}
else if (posBalance >= 37) {
if (diffRsCsdiff < -2) {
if (Cs < 566) {
return 8;
}
else if (Cs >= 566) {
return 2;
}
}
else if (diffRsCsdiff >= -2) {
return 2;
}
}
}
else if (ssDCval >= 159) {
return 4;
}
}
}
else if (Rs >= 422) {
if (TSC < 4) {
if (refDCval < 126) {
if (ssDCval < 114) {
if (Cs < 704) {
if (ssDCval < 97) {
if (diffAFD < 1) {
if (Rs < 519) {
return 2;
}
else if (Rs >= 519) {
return 8;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (ssDCval >= 97) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
}
else if (Cs >= 704) {
if (Rs < 486) {
return 8;
}
else if (Rs >= 486) {
return 2;
}
}
}
else if (ssDCval >= 114) {
if (posBalance < 38) {
if (tcor < 99) {
if (Rs < 487) {
return 4;
}
else if (Rs >= 487) {
return 2;
}
}
else if (tcor >= 99) {
return 1;
}
}
else if (posBalance >= 38) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
if (negBalance < 18) {
return 1;
}
else if (negBalance >= 18) {
return 2;
}
}
}
}
}
else if (refDCval >= 126) {
if (diffTSC < -1) {
if (Rs < 575) {
return 8;
}
else if (Rs >= 575) {
return 4;
}
}
else if (diffTSC >= -1) {
if (Rs < 642) {
return 1;
}
else if (Rs >= 642) {
if (Rs < 717) {
return 4;
}
else if (Rs >= 717) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 48) {
if (Cs < 547) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
if (RsDiff < 5) {
if (Rs < 804) {
return 4;
}
else if (Rs >= 804) {
return 8;
}
}
else if (RsDiff >= 5) {
return 8;
}
}
}
else if (Cs >= 547) {
if (MVDiff < 23) {
if (TSC < 5) {
if (Rs < 529) {
return 2;
}
else if (Rs >= 529) {
return 1;
}
}
else if (TSC >= 5) {
return 4;
}
}
else if (MVDiff >= 23) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (mcTcor < 99) {
if (posBalance < 54) {
if (Rs < 544) {
return 8;
}
else if (Rs >= 544) {
if (Cs < 517) {
return 8;
}
else if (Cs >= 517) {
return 2;
}
}
}
else if (posBalance >= 54) {
if (RsDiff < 3) {
if (negBalance < 31) {
if (Rs < 607) {
return 2;
}
else if (Rs >= 607) {
return 8;
}
}
else if (negBalance >= 31) {
return 8;
}
}
else if (RsDiff >= 3) {
return 2;
}
}
}
else if (mcTcor >= 99) {
return 2;
}
}
}
}
}
else if (negBalance >= 70) {
if (negBalance < 87) {
if (MVDiff < 32) {
if (Rs < 600) {
return 8;
}
else if (Rs >= 600) {
return 1;
}
}
else if (MVDiff >= 32) {
return 2;
}
}
else if (negBalance >= 87) {
return 8;
}
}
}
else if (SC >= 1411) {
if (Cs < 577) {
if (RsDiff < 6) {
if (negBalance < 39) {
if (posBalance < 67) {
return 1;
}
else if (posBalance >= 67) {
return 2;
}
}
else if (negBalance >= 39) {
if (ssDCval < 95) {
return 1;
}
else if (ssDCval >= 95) {
if (Rs < 1220) {
return 2;
}
else if (Rs >= 1220) {
return 8;
}
}
}
}
else if (RsDiff >= 6) {
if (mu_mv_mag_sq < 41) {
return 1;
}
else if (mu_mv_mag_sq >= 41) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 2;
}
}
}
}
else if (Cs >= 577) {
if (refDCval < 101) {
if (TSC < 5) {
if (Rs < 1453) {
if (negBalance < 39) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (negBalance >= 39) {
if (Rs < 1121) {
return 1;
}
else if (Rs >= 1121) {
return 8;
}
}
}
else if (Rs >= 1453) {
return 1;
}
}
else if (TSC >= 5) {
if (MVDiff < 4) {
if (Rs < 1217) {
if (Rs < 930) {
return 4;
}
else if (Rs >= 930) {
return 8;
}
}
else if (Rs >= 1217) {
return 1;
}
}
else if (MVDiff >= 4) {
if (tcor < 94) {
if (posBalance < 59) {
if (ssDCval < 96) {
if (RsDiff < 6) {
if (tcor < 91) {
return 1;
}
else if (tcor >= 91) {
return 8;
}
}
else if (RsDiff >= 6) {
if (gchDC < 3) {
if (Rs < 1605) {
if (Cs < 816) {
return 4;
}
else if (Cs >= 816) {
return 2;
}
}
else if (Rs >= 1605) {
return 4;
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (ssDCval >= 96) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
return 4;
}
}
}
else if (posBalance >= 59) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (tcor >= 94) {
if (RsCsDiff < 33) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (RsCsDiff >= 33) {
if (ssDCval < 98) {
if (diffMVdiffVal < 1) {
if (Rs < 1555) {
return 4;
}
else if (Rs >= 1555) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
return 8;
}
}
}
}
else if (ssDCval >= 98) {
if (Rs < 2123) {
if (Rs < 1498) {
return 2;
}
else if (Rs >= 1498) {
return 8;
}
}
else if (Rs >= 2123) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 101) {
if (MVDiff < 5) {
if (Rs < 658) {
if (Rs < 541) {
return 8;
}
else if (Rs >= 541) {
return 1;
}
}
else if (Rs >= 658) {
if (ssDCval < 122) {
if (tcor < 92) {
if (Rs < 1139) {
return 4;
}
else if (Rs >= 1139) {
return 8;
}
}
else if (tcor >= 92) {
if (Cs < 1277) {
if (negBalance < 23) {
return 1;
}
else if (negBalance >= 23) {
return 8;
}
}
else if (Cs >= 1277) {
return 4;
}
}
}
else if (ssDCval >= 122) {
if (Cs < 784) {
if (Cs < 601) {
return 1;
}
else if (Cs >= 601) {
return 8;
}
}
else if (Cs >= 784) {
if (RsCsDiff < 29) {
if (ssDCval < 136) {
if (SC < 2628) {
return 4;
}
else if (SC >= 2628) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
return 2;
}
}
}
else if (ssDCval >= 136) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (RsCsDiff >= 29) {
return 2;
}
}
}
}
}
else if (MVDiff >= 5) {
if (RsDiff < 7) {
if (ssDCval < 161) {
if (tcor < 94) {
if (SC < 1916) {
if (tcor < 83) {
if (Rs < 763) {
return 1;
}
else if (Rs >= 763) {
return 8;
}
}
else if (tcor >= 83) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 1;
}
}
}
}
}
else if (SC >= 1916) {
return 8;
}
}
else if (tcor >= 94) {
if (posBalance < 68) {
if (negBalance < 29) {
if (Rs < 1566) {
return 8;
}
else if (Rs >= 1566) {
return 1;
}
}
else if (negBalance >= 29) {
if (diffRsCsdiff < -15) {
if (tcor < 95) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (tcor >= 95) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 8;
}
}
}
else if (diffRsCsdiff >= -15) {
if (SC < 1550) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (SC >= 1550) {
return 8;
}
}
}
}
else if (posBalance >= 68) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
}
}
else if (ssDCval >= 161) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (RsDiff >= 7) {
if (SC < 2007) {
if (RsCsDiff < 73) {
return 8;
}
else if (RsCsDiff >= 73) {
return 2;
}
}
else if (SC >= 2007) {
if (Cs < 834) {
if (Rs < 1804) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
return 1;
}
}
else if (Rs >= 1804) {
if (mu_mv_mag_sq < 37) {
if (CsDiff < 6) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
return 1;
}
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (mu_mv_mag_sq >= 37) {
return 2;
}
}
}
else if (Cs >= 834) {
if (CsDiff < 7) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (CsDiff >= 7) {
if (Rs < 1967) {
return 2;
}
else if (Rs >= 1967) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (ssDCval < 133) {
if (Cs < 1232) {
if (Rs < 343) {
if (ssDCval < 124) {
return 4;
}
else if (ssDCval >= 124) {
return 2;
}
}
else if (Rs >= 343) {
if (Cs < 978) {
if (posBalance < 46) {
if (mu_mv_mag_sq < 106) {
if (tcor < 99) {
if (Cs < 888) {
return 4;
}
else if (Cs >= 888) {
return 8;
}
}
else if (tcor >= 99) {
if (Cs < 641) {
return 1;
}
else if (Cs >= 641) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 1089) {
if (Rs < 371) {
if (Rs < 368) {
return 8;
}
else if (Rs >= 368) {
return 4;
}
}
else if (Rs >= 371) {
if (diffMVdiffVal < -1) {
if (ssDCval < 98) {
return 1;
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (Rs >= 1089) {
if (SC < 2458) {
if (negBalance < 33) {
if (ssDCval < 111) {
return 4;
}
else if (ssDCval >= 111) {
return 1;
}
}
else if (negBalance >= 33) {
if (Rs < 1499) {
return 8;
}
else if (Rs >= 1499) {
if (Rs < 1726) {
return 4;
}
else if (Rs >= 1726) {
return 1;
}
}
}
}
else if (SC >= 2458) {
if (Rs < 2155) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
return 2;
}
}
else if (Rs >= 2155) {
if (Cs < 764) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (Cs >= 764) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 46) {
if (mcTcor < 99) {
if (refDCval < 115) {
if (refDCval < 114) {
if (SC < 1201) {
if (negBalance < 33) {
if (Cs < 527) {
return 1;
}
else if (Cs >= 527) {
return 8;
}
}
else if (negBalance >= 33) {
return 1;
}
}
else if (SC >= 1201) {
if (refDCval < 95) {
if (negBalance < 40) {
if (diffTSC < -1) {
if (Rs < 1678) {
return 4;
}
else if (Rs >= 1678) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (negBalance >= 40) {
if (Rs < 749) {
return 4;
}
else if (Rs >= 749) {
return 1;
}
}
}
else if (refDCval >= 95) {
if (Rs < 349) {
return 4;
}
else if (Rs >= 349) {
if (ssDCval < 112) {
if (negBalance < 36) {
if (Cs < 691) {
return 1;
}
else if (Cs >= 691) {
if (Cs < 957) {
return 8;
}
else if (Cs >= 957) {
if (Rs < 568) {
return 1;
}
else if (Rs >= 568) {
return 8;
}
}
}
}
else if (negBalance >= 36) {
return 8;
}
}
else if (ssDCval >= 112) {
if (Rs < 1305) {
return 1;
}
else if (Rs >= 1305) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 114) {
return 1;
}
}
else if (refDCval >= 115) {
if (Rs < 1802) {
return 8;
}
else if (Rs >= 1802) {
return 1;
}
}
}
else if (mcTcor >= 99) {
if (refDCval < 110) {
if (ssDCval < 94) {
if (Cs < 607) {
return 1;
}
else if (Cs >= 607) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (ssDCval >= 94) {
if (ssDCval < 100) {
if (Rs < 971) {
return 1;
}
else if (Rs >= 971) {
return 4;
}
}
else if (ssDCval >= 100) {
if (SC < 1304) {
return 4;
}
else if (SC >= 1304) {
return 8;
}
}
}
}
else if (refDCval >= 110) {
if (mu_mv_mag_sq < 110) {
if (TSCindex < 3) {
if (Rs < 614) {
return 2;
}
else if (Rs >= 614) {
return 1;
}
}
else if (TSCindex >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 1061) {
return 8;
}
else if (Rs >= 1061) {
if (Cs < 847) {
return 1;
}
else if (Cs >= 847) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 978) {
if (ssDCval < 121) {
if (ssDCval < 94) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (Rs < 1240) {
return 4;
}
else if (Rs >= 1240) {
return 2;
}
}
}
else if (ssDCval >= 94) {
if (mu_mv_mag_sq < 100) {
if (Rs < 956) {
return 1;
}
else if (Rs >= 956) {
return 8;
}
}
else if (mu_mv_mag_sq >= 100) {
return 8;
}
}
}
else if (ssDCval >= 121) {
if (Rs < 1401) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 1401) {
return 2;
}
}
}
}
}
else if (Cs >= 1232) {
if (Rs < 667) {
if (ssDCval < 120) {
return 1;
}
else if (ssDCval >= 120) {
return 8;
}
}
else if (Rs >= 667) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1147) {
if (Cs < 854) {
return 2;
}
else if (Cs >= 854) {
return 8;
}
}
else if (Rs >= 1147) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2326) {
if (MVDiff < 18) {
if (ssDCval < 109) {
if (posBalance < 48) {
if (RsDiff < 7) {
if (Rs < 3660) {
if (Cs < 1021) {
if (SC < 3238) {
if (MVDiff < 6) {
if (diffMVdiffVal < -5) {
return 8;
}
else if (diffMVdiffVal >= -5) {
if (refDCval < 49) {
return 4;
}
else if (refDCval >= 49) {
if (posBalance < 39) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Rs < 2408) {
return 1;
}
else if (Rs >= 2408) {
if (Rs < 2483) {
return 4;
}
else if (Rs >= 2483) {
return 8;
}
}
}
}
else if (posBalance >= 39) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (RsCsDiff < 18) {
return 8;
}
else if (RsCsDiff >= 18) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (Rs < 2501) {
return 1;
}
else if (Rs >= 2501) {
return 8;
}
}
}
}
}
else if (SC >= 3238) {
if (negBalance < 28) {
if (Rs < 2606) {
return 4;
}
else if (Rs >= 2606) {
if (Rs < 2718) {
return 1;
}
else if (Rs >= 2718) {
if (ssDCval < 102) {
if (Cs < 833) {
return 4;
}
else if (Cs >= 833) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (ssDCval < 101) {
if (Cs < 955) {
if (Rs < 3078) {
return 8;
}
else if (Rs >= 3078) {
return 4;
}
}
else if (Cs >= 955) {
return 4;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
}
}
else if (ssDCval >= 102) {
return 8;
}
}
}
}
else if (negBalance >= 28) {
if (tcor < 97) {
if (ssDCval < 74) {
if (SC < 3843) {
if (SC < 3416) {
if (diffRsCsdiff < -5) {
if (Rs < 2471) {
return 4;
}
else if (Rs >= 2471) {
return 2;
}
}
else if (diffRsCsdiff >= -5) {
return 8;
}
}
else if (SC >= 3416) {
return 2;
}
}
else if (SC >= 3843) {
if (diffRsCsdiff < -12) {
return 2;
}
else if (diffRsCsdiff >= -12) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
if (Cs < 734) {
if (Cs < 551) {
return 8;
}
else if (Cs >= 551) {
return 1;
}
}
else if (Cs >= 734) {
if (diffMVdiffVal < 2) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (MVDiff < 8) {
if (CsDiff < 4) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Rs < 2923) {
return 4;
}
else if (Rs >= 2923) {
return 4;
}
}
}
else if (CsDiff >= 4) {
if (Rs < 2899) {
return 8;
}
else if (Rs >= 2899) {
return 4;
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 2726) {
return 4;
}
else if (Rs >= 2726) {
return 2;
}
}
}
}
}
}
else if (tcor >= 97) {
if (SC < 3424) {
if (Rs < 2853) {
return 1;
}
else if (Rs >= 2853) {
return 8;
}
}
else if (SC >= 3424) {
if (mcTcor < 99) {
return 2;
}
else if (mcTcor >= 99) {
if (Rs < 3097) {
return 8;
}
else if (Rs >= 3097) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 1021) {
if (RsDiff < 2) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Rs < 2981) {
return 8;
}
else if (Rs >= 2981) {
return 2;
}
}
}
}
else if (RsDiff >= 2) {
if (refDCval < 103) {
if (ssDCval < 63) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 77) {
return 2;
}
else if (mu_mv_mag_sq >= 77) {
if (mu_mv_mag_sq < 103) {
return 4;
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 3328) {
return 8;
}
else if (Rs >= 3328) {
return 4;
}
}
}
}
else if (ssDCval >= 56) {
if (diffRsCsdiff < -10) {
if (Cs < 1173) {
return 4;
}
else if (Cs >= 1173) {
return 8;
}
}
else if (diffRsCsdiff >= -10) {
return 8;
}
}
}
else if (ssDCval >= 63) {
if (ssDCval < 90) {
if (mu_mv_mag_sq < 105) {
if (Rs < 2877) {
if (mu_mv_mag_sq < 92) {
return 1;
}
else if (mu_mv_mag_sq >= 92) {
if (Rs < 2511) {
return 8;
}
else if (Rs >= 2511) {
return 4;
}
}
}
else if (Rs >= 2877) {
if (TSC < 5) {
if (SC < 4380) {
return 2;
}
else if (SC >= 4380) {
if (Rs < 3344) {
return 4;
}
else if (Rs >= 3344) {
return 1;
}
}
}
else if (TSC >= 5) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (RsDiff < 4) {
if (Rs < 2734) {
if (Rs < 2568) {
return 8;
}
else if (Rs >= 2568) {
return 4;
}
}
else if (Rs >= 2734) {
return 1;
}
}
else if (RsDiff >= 4) {
if (posBalance < 40) {
if (RsCsDiff < 31) {
return 2;
}
else if (RsCsDiff >= 31) {
if (Rs < 3406) {
return 4;
}
else if (Rs >= 3406) {
return 1;
}
}
}
else if (posBalance >= 40) {
return 2;
}
}
}
}
else if (ssDCval >= 90) {
if (ssDCval < 102) {
if (TSCindex < 3) {
return 2;
}
else if (TSCindex >= 3) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Rs < 2985) {
return 4;
}
else if (Rs >= 2985) {
return 2;
}
}
}
}
else if (ssDCval >= 102) {
if (Rs < 3617) {
return 4;
}
else if (Rs >= 3617) {
return 8;
}
}
}
}
}
else if (refDCval >= 103) {
return 8;
}
}
}
}
else if (Rs >= 3660) {
if (Cs < 1280) {
if (refDCval < 93) {
if (ssDCval < 68) {
if (ssDCval < 63) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (ssDCval >= 68) {
if (Cs < 1265) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 4141) {
return 4;
}
else if (Rs >= 4141) {
return 8;
}
}
}
else if (Cs >= 1265) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (Rs < 4089) {
return 4;
}
else if (Rs >= 4089) {
return 4;
}
}
}
}
}
else if (refDCval >= 93) {
if (Rs < 4008) {
if (MVDiff < 3) {
if (Cs < 931) {
return 2;
}
else if (Cs >= 931) {
return 8;
}
}
else if (MVDiff >= 3) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (Rs >= 4008) {
return 2;
}
}
}
else if (Cs >= 1280) {
if (ssDCval < 84) {
if (Rs < 3892) {
return 4;
}
else if (Rs >= 3892) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
}
else if (ssDCval >= 84) {
if (Cs < 1286) {
return 4;
}
else if (Cs >= 1286) {
return 2;
}
}
}
}
}
else if (RsDiff >= 7) {
if (negBalance < 51) {
if (ssDCval < 86) {
if (refDCval < 65) {
if (AFD < 17) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (refDCval < 57) {
return 4;
}
else if (refDCval >= 57) {
if (Cs < 1141) {
if (tcor < 93) {
if (Rs < 2430) {
if (Rs < 2420) {
return 4;
}
else if (Rs >= 2420) {
return 2;
}
}
else if (Rs >= 2430) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (tcor >= 93) {
return 2;
}
}
else if (Cs >= 1141) {
if (CsDiff < 5) {
if (Cs < 1183) {
return 8;
}
else if (Cs >= 1183) {
return 4;
}
}
else if (CsDiff >= 5) {
if (AFD < 11) {
if (Rs < 3477) {
return 4;
}
else if (Rs >= 3477) {
return 2;
}
}
else if (AFD >= 11) {
return 2;
}
}
}
}
}
}
else if (AFD >= 17) {
return 2;
}
}
else if (refDCval >= 65) {
if (Rs < 3662) {
if (Cs < 1055) {
if (Rs < 3019) {
if (TSC < 6) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (negBalance < 43) {
if (gchDC < 1) {
if (Rs < 2603) {
return 1;
}
else if (Rs >= 2603) {
return 2;
}
}
else if (gchDC >= 1) {
return 2;
}
}
else if (negBalance >= 43) {
return 1;
}
}
}
else if (TSC >= 6) {
if (RsCsDiff < 189) {
if (tcor < 87) {
if (Cs < 906) {
return 1;
}
else if (Cs >= 906) {
return 2;
}
}
else if (tcor >= 87) {
if (posBalance < 45) {
if (Rs < 2843) {
return 4;
}
else if (Rs >= 2843) {
return 1;
}
}
else if (posBalance >= 45) {
return 2;
}
}
}
else if (RsCsDiff >= 189) {
return 4;
}
}
}
else if (Rs >= 3019) {
if (refDCval < 73) {
if (negBalance < 40) {
if (SC < 4138) {
return 1;
}
else if (SC >= 4138) {
return 4;
}
}
else if (negBalance >= 40) {
return 4;
}
}
else if (refDCval >= 73) {
if (tcor < 88) {
if (Rs < 3477) {
return 4;
}
else if (Rs >= 3477) {
return 1;
}
}
else if (tcor >= 88) {
return 1;
}
}
}
}
else if (Cs >= 1055) {
if (diffRsCsdiff < 44) {
if (tcor < 92) {
if (ssDCval < 72) {
if (mu_mv_mag_sq < 42) {
return 2;
}
else if (mu_mv_mag_sq >= 42) {
if (ssDCval < 69) {
if (Rs < 3134) {
return 4;
}
else if (Rs >= 3134) {
return 2;
}
}
else if (ssDCval >= 69) {
return 2;
}
}
}
else if (ssDCval >= 72) {
if (diffTSC < 1) {
if (Rs < 3441) {
return 4;
}
else if (Rs >= 3441) {
if (Rs < 3495) {
return 1;
}
else if (Rs >= 3495) {
return 4;
}
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (tcor >= 92) {
return 2;
}
}
else if (diffRsCsdiff >= 44) {
if (ssDCval < 74) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (Rs < 2937) {
return 2;
}
else if (Rs >= 2937) {
return 4;
}
}
}
else if (ssDCval >= 74) {
return 4;
}
}
}
}
else if (Rs >= 3662) {
if (AFD < 12) {
if (mu_mv_mag_sq < 78) {
return 4;
}
else if (mu_mv_mag_sq >= 78) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
return 4;
}
}
}
else if (AFD >= 12) {
return 1;
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 93) {
if (MVDiff < 6) {
if (Rs < 3672) {
return 8;
}
else if (Rs >= 3672) {
if (Rs < 4215) {
return 4;
}
else if (Rs >= 4215) {
if (Cs < 1190) {
return 4;
}
else if (Cs >= 1190) {
return 4;
}
}
}
}
else if (MVDiff >= 6) {
if (Rs < 4124) {
if (mu_mv_mag_sq < 53) {
if (Rs < 4032) {
return 8;
}
else if (Rs >= 4032) {
return 8;
}
}
else if (mu_mv_mag_sq >= 53) {
if (mu_mv_mag_sq < 54) {
return 4;
}
else if (mu_mv_mag_sq >= 54) {
if (mcTcor < 95) {
if (Rs < 3940) {
return 8;
}
else if (Rs >= 3940) {
return 8;
}
}
else if (mcTcor >= 95) {
if (Rs < 3882) {
return 4;
}
else if (Rs >= 3882) {
return 8;
}
}
}
}
}
else if (Rs >= 4124) {
return 8;
}
}
}
else if (ssDCval >= 93) {
if (Rs < 3679) {
if (negBalance < 43) {
if (diffRsCsdiff < 14) {
return 8;
}
else if (diffRsCsdiff >= 14) {
return 4;
}
}
else if (negBalance >= 43) {
if (Rs < 3143) {
return 2;
}
else if (Rs >= 3143) {
return 8;
}
}
}
else if (Rs >= 3679) {
return 2;
}
}
}
}
else if (negBalance >= 51) {
if (mcTcor < 86) {
if (Cs < 792) {
return 2;
}
else if (Cs >= 792) {
return 8;
}
}
else if (mcTcor >= 86) {
return 2;
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Rs < 3656) {
if (Cs < 246) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 2;
}
}
else if (Cs >= 246) {
if (negBalance < 19) {
if (Cs < 630) {
return 1;
}
else if (Cs >= 630) {
if (RsDiff < 4) {
if (Cs < 1010) {
return 8;
}
else if (Cs >= 1010) {
return 8;
}
}
else if (RsDiff >= 4) {
if (Rs < 2626) {
return 4;
}
else if (Rs >= 2626) {
return 4;
}
}
}
}
else if (negBalance >= 19) {
if (Rs < 2838) {
if (ssDCval < 81) {
if (Cs < 1116) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (diffAFD < -2) {
if (Rs < 2416) {
return 1;
}
else if (Rs >= 2416) {
return 2;
}
}
else if (diffAFD >= -2) {
return 2;
}
}
}
else if (Cs >= 1116) {
return 8;
}
}
else if (ssDCval >= 81) {
if (Cs < 457) {
return 1;
}
else if (Cs >= 457) {
return 2;
}
}
}
else if (Rs >= 2838) {
if (SC < 3298) {
if (SC < 3171) {
return 2;
}
else if (SC >= 3171) {
return 8;
}
}
else if (SC >= 3298) {
if (mcTcor < 97) {
if (ssDCval < 69) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (mcTcor >= 97) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3656) {
if (Cs < 1066) {
return 4;
}
else if (Cs >= 1066) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 60) {
if (Rs < 2511) {
return 8;
}
else if (Rs >= 2511) {
if (Rs < 7026) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
if (Rs < 2871) {
return 2;
}
else if (Rs >= 2871) {
return 4;
}
}
}
else if (Rs >= 7026) {
return 2;
}
}
}
else if (refDCval >= 60) {
if (negBalance < 38) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 53) {
if (RsCsDiff < 70) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (Rs < 2584) {
return 1;
}
else if (Rs >= 2584) {
if (Rs < 3461) {
return 4;
}
else if (Rs >= 3461) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 70) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
if (Cs < 908) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 4;
}
}
else if (Cs >= 908) {
if (RsCsDiff < 175) {
return 2;
}
else if (RsCsDiff >= 175) {
if (Rs < 3223) {
return 4;
}
else if (Rs >= 3223) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (posBalance < 49) {
if (Rs < 3170) {
if (Rs < 2749) {
return 1;
}
else if (Rs >= 2749) {
return 2;
}
}
else if (Rs >= 3170) {
return 4;
}
}
else if (posBalance >= 49) {
return 2;
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 63) {
if (Cs < 937) {
return 2;
}
else if (Cs >= 937) {
if (Cs < 1154) {
return 4;
}
else if (Cs >= 1154) {
return 2;
}
}
}
else if (ssDCval >= 63) {
if (SC < 3134) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
return 1;
}
}
else if (SC >= 3134) {
if (RsDiff < 5) {
if (Rs < 3949) {
return 8;
}
else if (Rs >= 3949) {
return 2;
}
}
else if (RsDiff >= 5) {
if (gchDC < 4) {
if (mcTcor < 88) {
return 1;
}
else if (mcTcor >= 88) {
return 2;
}
}
else if (gchDC >= 4) {
if (Cs < 1091) {
return 2;
}
else if (Cs >= 1091) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 38) {
if (tcor < 77) {
if (RsDiff < 15) {
if (SC < 4099) {
if (SC < 3476) {
return 2;
}
else if (SC >= 3476) {
return 4;
}
}
else if (SC >= 4099) {
return 1;
}
}
else if (RsDiff >= 15) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
return 2;
}
}
}
}
else if (tcor >= 77) {
if (diffRsCsdiff < -190) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 2;
}
}
else if (diffRsCsdiff >= -190) {
if (RsDiff < 6) {
if (ssDCval < 65) {
if (Cs < 959) {
return 2;
}
else if (Cs >= 959) {
return 4;
}
}
else if (ssDCval >= 65) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (RsDiff >= 6) {
if (Rs < 2534) {
if (refDCval < 77) {
if (SC < 3536) {
return 2;
}
else if (SC >= 3536) {
if (Cs < 1156) {
return 4;
}
else if (Cs >= 1156) {
return 2;
}
}
}
else if (refDCval >= 77) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (MVDiff < 17) {
if (diffRsCsdiff < -43) {
return 1;
}
else if (diffRsCsdiff >= -43) {
if (Rs < 2335) {
return 4;
}
else if (Rs >= 2335) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 2;
}
}
}
}
else if (MVDiff >= 17) {
return 4;
}
}
}
}
else if (Rs >= 2534) {
if (refDCval < 65) {
if (SC < 3850) {
return 2;
}
else if (SC >= 3850) {
if (Rs < 2961) {
return 4;
}
else if (Rs >= 2961) {
if (RsDiff < 11) {
return 2;
}
else if (RsDiff >= 11) {
return 4;
}
}
}
}
else if (refDCval >= 65) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsCsDiff < 121) {
if (Cs < 998) {
if (tcor < 95) {
if (RsDiff < 8) {
if (mu_mv_mag_sq < 49) {
return 8;
}
else if (mu_mv_mag_sq >= 49) {
if (Rs < 3111) {
if (Rs < 2590) {
return 8;
}
else if (Rs >= 2590) {
return 4;
}
}
else if (Rs >= 3111) {
return 4;
}
}
}
else if (RsDiff >= 8) {
if (Cs < 515) {
return 2;
}
else if (Cs >= 515) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
}
}
else if (tcor >= 95) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
}
else if (Cs >= 998) {
if (diffAFD < -4) {
if (refDCval < 97) {
if (Rs < 2973) {
return 8;
}
else if (Rs >= 2973) {
return 2;
}
}
else if (refDCval >= 97) {
return 4;
}
}
else if (diffAFD >= -4) {
if (negBalance < 31) {
if (Rs < 2891) {
return 1;
}
else if (Rs >= 2891) {
if (Cs < 1186) {
if (Rs < 4019) {
return 8;
}
else if (Rs >= 4019) {
return 8;
}
}
else if (Cs >= 1186) {
return 4;
}
}
}
else if (negBalance >= 31) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (RsDiff < 9) {
if (Rs < 2498) {
return 4;
}
else if (Rs >= 2498) {
return 8;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 121) {
if (negBalance < 49) {
if (mu_mv_mag_sq < 51) {
return 2;
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 3261) {
return 8;
}
else if (Rs >= 3261) {
return 2;
}
}
}
else if (negBalance >= 49) {
if (Cs < 837) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 1;
}
}
else if (Cs >= 837) {
if (tcor < 81) {
if (Rs < 3115) {
if (CsDiff < 5) {
if (Rs < 2656) {
return 4;
}
else if (Rs >= 2656) {
if (Rs < 2681) {
return 1;
}
else if (Rs >= 2681) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (Rs >= 3115) {
if (CsDiff < 6) {
if (Rs < 3295) {
return 4;
}
else if (Rs >= 3295) {
return 1;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (tcor >= 81) {
if (RsDiff < 13) {
if (Rs < 2585) {
return 8;
}
else if (Rs >= 2585) {
return 4;
}
}
else if (RsDiff >= 13) {
if (Cs < 1220) {
return 2;
}
else if (Cs >= 1220) {
return 4;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 2981) {
if (Cs < 839) {
if (refDCval < 121) {
if (TSC < 6) {
if (tcor < 94) {
if (Cs < 242) {
return 2;
}
else if (Cs >= 242) {
return 8;
}
}
else if (tcor >= 94) {
if (Cs < 679) {
return 1;
}
else if (Cs >= 679) {
return 2;
}
}
}
else if (TSC >= 6) {
if (AFD < 15) {
if (mu_mv_mag_sq < 31) {
return 1;
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
else if (AFD >= 15) {
if (Cs < 719) {
return 2;
}
else if (Cs >= 719) {
return 4;
}
}
}
}
else if (refDCval >= 121) {
if (Cs < 704) {
if (MVDiff < 4) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
else if (MVDiff >= 4) {
if (RsDiff < 18) {
if (mu_mv_mag_sq < 32) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (mu_mv_mag_sq >= 32) {
if (AFD < 13) {
if (posBalance < 68) {
return 1;
}
else if (posBalance >= 68) {
return 2;
}
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (RsDiff >= 18) {
return 2;
}
}
}
else if (Cs >= 704) {
return 2;
}
}
}
else if (Cs >= 839) {
if (mcTcor < 97) {
if (diffRsCsdiff < -37) {
if (TSC < 8) {
if (Rs < 2627) {
return 2;
}
else if (Rs >= 2627) {
return 8;
}
}
else if (TSC >= 8) {
return 1;
}
}
else if (diffRsCsdiff >= -37) {
if (mcTcor < 95) {
if (ssDCval < 114) {
return 4;
}
else if (ssDCval >= 114) {
return 8;
}
}
else if (mcTcor >= 95) {
if (diffMVdiffVal < 2) {
if (diffMVdiffVal < -3) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (mu_mv_mag_sq < 34) {
return 1;
}
else if (mu_mv_mag_sq >= 34) {
return 8;
}
}
}
else if (diffMVdiffVal >= -3) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 1185) {
return 8;
}
else if (Cs >= 1185) {
return 1;
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
if (Rs < 2564) {
if (Rs < 2380) {
return 8;
}
else if (Rs >= 2380) {
return 1;
}
}
else if (Rs >= 2564) {
if (Rs < 2967) {
return 2;
}
else if (Rs >= 2967) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2981) {
if (MVDiff < 7) {
if (Cs < 619) {
if (RsDiff < 7) {
if (Rs < 3894) {
return 8;
}
else if (Rs >= 3894) {
return 2;
}
}
else if (RsDiff >= 7) {
if (diffMVdiffVal < -59) {
return 8;
}
else if (diffMVdiffVal >= -59) {
return 2;
}
}
}
else if (Cs >= 619) {
if (ssDCval < 142) {
if (tcor < 92) {
if (ssDCval < 133) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 7) {
if (Cs < 1053) {
return 4;
}
else if (Cs >= 1053) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
return 2;
}
}
}
else if (tcor >= 92) {
if (diffMVdiffVal < -12) {
if (Rs < 4354) {
return 1;
}
else if (Rs >= 4354) {
return 2;
}
}
else if (diffMVdiffVal >= -12) {
return 2;
}
}
}
else if (ssDCval >= 142) {
if (Rs < 3894) {
if (SC < 4865) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
else if (SC >= 4865) {
if (posBalance < 29) {
return 2;
}
else if (posBalance >= 29) {
if (Rs < 3725) {
if (MVDiff < 1) {
if (Cs < 1209) {
return 1;
}
else if (Cs >= 1209) {
return 2;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (Rs >= 3725) {
return 1;
}
}
}
}
else if (Rs >= 3894) {
if (TSC < 10) {
if (diffTSC < -2) {
if (Rs < 4042) {
return 1;
}
else if (Rs >= 4042) {
return 2;
}
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (TSC >= 10) {
return 1;
}
}
}
}
}
else if (MVDiff >= 7) {
if (mcTcor < 94) {
if (refDCval < 133) {
return 2;
}
else if (refDCval >= 133) {
if (AFD < 18) {
if (ssDCval < 135) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 1;
}
}
else if (ssDCval >= 135) {
return 4;
}
}
else if (AFD >= 18) {
if (Rs < 3776) {
return 1;
}
else if (Rs >= 3776) {
return 2;
}
}
}
}
else if (mcTcor >= 94) {
if (RsDiff < 9) {
if (CsDiff < 4) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 50) {
if (Rs < 3557) {
return 8;
}
else if (Rs >= 3557) {
if (Rs < 3600) {
return 1;
}
else if (Rs >= 3600) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (diffRsCsdiff < -113) {
return 1;
}
else if (diffRsCsdiff >= -113) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (SC < 4347) {
return 8;
}
else if (SC >= 4347) {
if (mu_mv_mag_sq < 74) {
if (ssDCval < 114) {
return 2;
}
else if (ssDCval >= 114) {
if (Rs < 4406) {
return 8;
}
else if (Rs >= 4406) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 74) {
return 2;
}
}
}
}
else if (CsDiff >= 4) {
if (ssDCval < 114) {
return 1;
}
else if (ssDCval >= 114) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 1;
}
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 134) {
if (diffMVdiffVal < -47) {
return 1;
}
else if (diffMVdiffVal >= -47) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (MVDiff < 14) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (RsDiff < 20) {
if (CsDiff < 5) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (Rs < 4035) {
return 1;
}
else if (Rs >= 4035) {
return 2;
}
}
}
else if (CsDiff >= 5) {
if (RsDiff < 19) {
return 4;
}
else if (RsDiff >= 19) {
if (Rs < 4114) {
return 2;
}
else if (Rs >= 4114) {
return 1;
}
}
}
}
else if (RsDiff >= 20) {
return 2;
}
}
}
else if (MVDiff >= 14) {
if (TSC < 8) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (TSC >= 8) {
return 2;
}
}
}
}
}
else if (ssDCval >= 134) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (TSC < 8) {
if (Rs < 4531) {
return 2;
}
else if (Rs >= 4531) {
return 1;
}
}
else if (TSC >= 8) {
if (negBalance < 54) {
if (mu_mv_mag_sq < 54) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Cs < 842) {
if (SC < 3763) {
return 1;
}
else if (SC >= 3763) {
if (Rs < 5248) {
return 2;
}
else if (Rs >= 5248) {
return 1;
}
}
}
else if (Cs >= 842) {
if (Rs < 6300) {
return 1;
}
else if (Rs >= 6300) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (negBalance >= 54) {
if (Rs < 4474) {
return 4;
}
else if (Rs >= 4474) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 18) {
if (mcTcor < 68) {
if (MVDiff < 44) {
return 1;
}
else if (MVDiff >= 44) {
if (AFD < 21) {
if (Rs < 2768) {
return 4;
}
else if (Rs >= 2768) {
return 1;
}
}
else if (AFD >= 21) {
if (gchDC < 7) {
return 2;
}
else if (gchDC >= 7) {
return 8;
}
}
}
}
else if (mcTcor >= 68) {
if (refDCval < 131) {
if (ssDCval < 82) {
if (Rs < 3333) {
if (AFD < 13) {
if (ssDCval < 61) {
if (posBalance < 51) {
if (Rs < 3307) {
return 4;
}
else if (Rs >= 3307) {
return 8;
}
}
else if (posBalance >= 51) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
return 2;
}
}
}
else if (ssDCval >= 61) {
if (Rs < 2708) {
if (Cs < 1154) {
if (Cs < 355) {
return 8;
}
else if (Cs >= 355) {
if (diffRsCsdiff < -435) {
return 4;
}
else if (diffRsCsdiff >= -435) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 8;
}
}
}
}
else if (Cs >= 1154) {
return 8;
}
}
else if (Rs >= 2708) {
if (Cs < 918) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (Cs >= 918) {
if (RsDiff < 13) {
return 2;
}
else if (RsDiff >= 13) {
return 1;
}
}
}
}
}
else if (AFD >= 13) {
if (TSC < 12) {
if (Cs < 176) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 4;
}
}
else if (Cs >= 176) {
if (RsDiff < 10) {
if (refDCval < 63) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
else if (refDCval >= 63) {
if (diffTSC < 1) {
if (AFD < 14) {
if (MVDiff < 35) {
return 1;
}
else if (MVDiff >= 35) {
return 4;
}
}
else if (AFD >= 14) {
return 4;
}
}
else if (diffTSC >= 1) {
if (CsDiff < 4) {
if (Rs < 2392) {
return 1;
}
else if (Rs >= 2392) {
if (Rs < 2726) {
return 2;
}
else if (Rs >= 2726) {
return 4;
}
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (RsDiff >= 10) {
if (diffRsCsdiff < 192) {
if (mcTcor < 97) {
if (posBalance < 54) {
if (negBalance < 51) {
if (Rs < 2512) {
return 2;
}
else if (Rs >= 2512) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (ssDCval < 69) {
if (Rs < 3117) {
return 2;
}
else if (Rs >= 3117) {
return 4;
}
}
else if (ssDCval >= 69) {
return 4;
}
}
}
}
else if (negBalance >= 51) {
return 2;
}
}
else if (posBalance >= 54) {
if (SC < 2616) {
return 4;
}
else if (SC >= 2616) {
if (negBalance < 60) {
return 2;
}
else if (negBalance >= 60) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 310) {
if (RsDiff < 18) {
return 2;
}
else if (RsDiff >= 18) {
return 1;
}
}
else if (Cs >= 310) {
return 4;
}
}
}
else if (diffRsCsdiff >= 192) {
return 4;
}
}
}
}
else if (TSC >= 12) {
if (RsDiff < 19) {
if (gchDC < 8) {
return 1;
}
else if (gchDC >= 8) {
return 4;
}
}
else if (RsDiff >= 19) {
if (Cs < 594) {
return 2;
}
else if (Cs >= 594) {
return 8;
}
}
}
}
}
else if (Rs >= 3333) {
if (posBalance < 61) {
if (Cs < 893) {
if (Rs < 5944) {
return 4;
}
else if (Rs >= 5944) {
return 8;
}
}
else if (Cs >= 893) {
if (Rs < 4130) {
if (Cs < 925) {
return 1;
}
else if (Cs >= 925) {
return 2;
}
}
else if (Rs >= 4130) {
return 4;
}
}
}
else if (posBalance >= 61) {
if (ssDCval < 79) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 2;
}
}
else if (ssDCval >= 79) {
return 1;
}
}
}
}
else if (ssDCval >= 82) {
if (RsDiff < 15) {
if (SC < 3339) {
if (refDCval < 92) {
if (gchDC < 2) {
if (MVDiff < 23) {
if (Cs < 594) {
if (Rs < 2441) {
return 2;
}
else if (Rs >= 2441) {
return 1;
}
}
else if (Cs >= 594) {
return 4;
}
}
else if (MVDiff >= 23) {
if (Rs < 2631) {
return 1;
}
else if (Rs >= 2631) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (Rs < 2343) {
return 8;
}
else if (Rs >= 2343) {
if (mu_mv_mag_sq < 24) {
if (AFD < 20) {
if (Rs < 2561) {
return 2;
}
else if (Rs >= 2561) {
return 1;
}
}
else if (AFD >= 20) {
return 1;
}
}
else if (mu_mv_mag_sq >= 24) {
return 1;
}
}
}
}
else if (refDCval >= 92) {
if (diffAFD < -3) {
if (refDCval < 107) {
if (Rs < 2409) {
return 2;
}
else if (Rs >= 2409) {
return 8;
}
}
else if (refDCval >= 107) {
if (mu_mv_mag_sq < 27) {
if (gchDC < 1) {
if (Rs < 2655) {
return 1;
}
else if (Rs >= 2655) {
return 2;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 27) {
if (Cs < 417) {
return 1;
}
else if (Cs >= 417) {
return 2;
}
}
}
}
else if (diffAFD >= -3) {
if (diffAFD < 5) {
if (tcor < 84) {
if (diffRsCsdiff < 20) {
return 8;
}
else if (diffRsCsdiff >= 20) {
if (MVDiff < 38) {
if (Rs < 2620) {
return 1;
}
else if (Rs >= 2620) {
return 2;
}
}
else if (MVDiff >= 38) {
if (diffMVdiffVal < -8) {
return 1;
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
}
}
else if (tcor >= 84) {
if (ssDCval < 102) {
if (RsDiff < 13) {
if (TSC < 11) {
if (Cs < 446) {
return 8;
}
else if (Cs >= 446) {
if (Rs < 2723) {
return 1;
}
else if (Rs >= 2723) {
return 8;
}
}
}
else if (TSC >= 11) {
return 2;
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (ssDCval >= 102) {
if (Cs < 7) {
return 8;
}
else if (Cs >= 7) {
if (ssDCval < 121) {
if (Rs < 2601) {
if (negBalance < 60) {
return 1;
}
else if (negBalance >= 60) {
if (MVDiff < 38) {
return 8;
}
else if (MVDiff >= 38) {
if (Cs < 357) {
return 8;
}
else if (Cs >= 357) {
return 1;
}
}
}
}
else if (Rs >= 2601) {
if (Cs < 172) {
return 1;
}
else if (Cs >= 172) {
return 8;
}
}
}
else if (ssDCval >= 121) {
return 1;
}
}
}
}
}
else if (diffAFD >= 5) {
return 1;
}
}
}
}
else if (SC >= 3339) {
if (Rs < 4942) {
if (AFD < 24) {
if (MVDiff < 38) {
if (RsDiff < 9) {
if (posBalance < 63) {
if (SC < 3591) {
if (Rs < 2565) {
return 8;
}
else if (Rs >= 2565) {
return 1;
}
}
else if (SC >= 3591) {
if (Rs < 4001) {
return 8;
}
else if (Rs >= 4001) {
return 2;
}
}
}
else if (posBalance >= 63) {
if (posBalance < 67) {
return 1;
}
else if (posBalance >= 67) {
if (Cs < 1268) {
return 2;
}
else if (Cs >= 1268) {
return 8;
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 96) {
if (CsDiff < 7) {
if (Rs < 3632) {
if (refDCval < 95) {
return 2;
}
else if (refDCval >= 95) {
if (Rs < 2570) {
return 2;
}
else if (Rs >= 2570) {
return 4;
}
}
}
else if (Rs >= 3632) {
return 8;
}
}
else if (CsDiff >= 7) {
if (MVDiff < 27) {
return 1;
}
else if (MVDiff >= 27) {
if (Rs < 2770) {
return 8;
}
else if (Rs >= 2770) {
return 2;
}
}
}
}
else if (refDCval >= 96) {
if (Rs < 2955) {
if (refDCval < 118) {
if (RsCsDiff < 170) {
if (diffRsCsdiff < -29) {
if (Rs < 2775) {
return 8;
}
else if (Rs >= 2775) {
return 2;
}
}
else if (diffRsCsdiff >= -29) {
if (SC < 3800) {
return 1;
}
else if (SC >= 3800) {
return 2;
}
}
}
else if (RsCsDiff >= 170) {
if (Cs < 1178) {
if (MVDiff < 30) {
return 2;
}
else if (MVDiff >= 30) {
return 8;
}
}
else if (Cs >= 1178) {
return 8;
}
}
}
else if (refDCval >= 118) {
if (Rs < 2394) {
return 8;
}
else if (Rs >= 2394) {
return 1;
}
}
}
else if (Rs >= 2955) {
if (diffRsCsdiff < -81) {
if (Cs < 1186) {
if (diffAFD < -5) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
else if (diffAFD >= -5) {
return 2;
}
}
else if (Cs >= 1186) {
return 8;
}
}
else if (diffRsCsdiff >= -81) {
if (MVDiff < 23) {
if (gchDC < 3) {
if (Cs < 948) {
return 8;
}
else if (Cs >= 948) {
return 1;
}
}
else if (gchDC >= 3) {
return 2;
}
}
else if (MVDiff >= 23) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 38) {
if (TSC < 8) {
if (RsCsDiff < 107) {
if (RsDiff < 9) {
if (ssDCval < 105) {
return 4;
}
else if (ssDCval >= 105) {
return 8;
}
}
else if (RsDiff >= 9) {
return 1;
}
}
else if (RsCsDiff >= 107) {
return 2;
}
}
else if (TSC >= 8) {
if (diffRsCsdiff < -208) {
if (Rs < 3546) {
return 1;
}
else if (Rs >= 3546) {
return 2;
}
}
else if (diffRsCsdiff >= -208) {
if (posBalance < 87) {
if (Rs < 2395) {
return 2;
}
else if (Rs >= 2395) {
if (TSC < 10) {
if (RsCsDiff < 107) {
return 8;
}
else if (RsCsDiff >= 107) {
if (gchDC < 3) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
return 8;
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (posBalance >= 87) {
if (Rs < 2747) {
return 2;
}
else if (Rs >= 2747) {
return 1;
}
}
}
}
}
}
else if (AFD >= 24) {
if (SC < 3716) {
if (ssDCval < 92) {
return 1;
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (SC >= 3716) {
if (MVDiff < 69) {
return 1;
}
else if (MVDiff >= 69) {
if (Cs < 1278) {
return 4;
}
else if (Cs >= 1278) {
if (Rs < 3638) {
return 2;
}
else if (Rs >= 3638) {
return 8;
}
}
}
}
}
}
else if (Rs >= 4942) {
if (Cs < 1175) {
return 2;
}
else if (Cs >= 1175) {
return 8;
}
}
}
}
else if (RsDiff >= 15) {
if (mcTcor < 90) {
if (RsCsDiff < 372) {
if (mcTcor < 83) {
if (Cs < 237) {
return 2;
}
else if (Cs >= 237) {
if (ssDCval < 114) {
if (mcTcor < 70) {
return 1;
}
else if (mcTcor >= 70) {
return 4;
}
}
else if (ssDCval >= 114) {
return 8;
}
}
}
else if (mcTcor >= 83) {
if (tcor < 73) {
if (mu_mv_mag_sq < 9) {
return 2;
}
else if (mu_mv_mag_sq >= 9) {
if (Cs < 591) {
return 2;
}
else if (Cs >= 591) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
return 2;
}
}
}
}
else if (tcor >= 73) {
if (ssDCval < 109) {
if (Rs < 3069) {
return 4;
}
else if (Rs >= 3069) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 4;
}
}
}
else if (ssDCval >= 109) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Rs < 3073) {
return 1;
}
else if (Rs >= 3073) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 372) {
if (TSC < 17) {
if (TSC < 12) {
if (Cs < 950) {
return 2;
}
else if (Cs >= 950) {
if (Rs < 3588) {
return 8;
}
else if (Rs >= 3588) {
return 2;
}
}
}
else if (TSC >= 12) {
if (negBalance < 58) {
if (MVDiff < 151) {
if (Cs < 1172) {
return 2;
}
else if (Cs >= 1172) {
if (MVDiff < 50) {
return 2;
}
else if (MVDiff >= 50) {
return 4;
}
}
}
else if (MVDiff >= 151) {
return 1;
}
}
else if (negBalance >= 58) {
return 2;
}
}
}
else if (TSC >= 17) {
if (Rs < 3092) {
return 1;
}
else if (Rs >= 3092) {
return 4;
}
}
}
}
else if (mcTcor >= 90) {
if (refDCval < 113) {
if (gchDC < 6) {
if (mu_mv_mag_sq < 7) {
if (RsDiff < 21) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
if (diffRsCsdiff < -16) {
if (Rs < 2997) {
return 4;
}
else if (Rs >= 2997) {
return 2;
}
}
else if (diffRsCsdiff >= -16) {
if (refDCval < 92) {
return 2;
}
else if (refDCval >= 92) {
if (posBalance < 67) {
if (MVDiff < 51) {
return 8;
}
else if (MVDiff >= 51) {
return 2;
}
}
else if (posBalance >= 67) {
return 2;
}
}
}
}
}
else if (RsDiff >= 21) {
if (Cs < 747) {
return 2;
}
else if (Cs >= 747) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffMVdiffVal < -9) {
if (diffAFD < 4) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 1;
}
}
else if (diffAFD >= 4) {
return 1;
}
}
else if (diffMVdiffVal >= -9) {
if (mcTcor < 91) {
if (Cs < 686) {
return 2;
}
else if (Cs >= 686) {
return 4;
}
}
else if (mcTcor >= 91) {
return 2;
}
}
}
}
else if (gchDC >= 6) {
if (Cs < 255) {
return 2;
}
else if (Cs >= 255) {
if (negBalance < 47) {
if (posBalance < 78) {
return 8;
}
else if (posBalance >= 78) {
if (mcTcor < 94) {
if (Rs < 3344) {
return 2;
}
else if (Rs >= 3344) {
return 8;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
}
else if (negBalance >= 47) {
return 1;
}
}
}
}
else if (refDCval >= 113) {
if (Rs < 3205) {
if (Cs < 245) {
return 2;
}
else if (Cs >= 245) {
if (MVDiff < 24) {
if (RsDiff < 16) {
if (Rs < 3070) {
return 1;
}
else if (Rs >= 3070) {
return 8;
}
}
else if (RsDiff >= 16) {
return 2;
}
}
else if (MVDiff >= 24) {
if (SC < 3713) {
return 1;
}
else if (SC >= 3713) {
if (refDCval < 127) {
if (Cs < 717) {
return 8;
}
else if (Cs >= 717) {
return 1;
}
}
else if (refDCval >= 127) {
return 2;
}
}
}
}
}
else if (Rs >= 3205) {
if (diffRsCsdiff < -169) {
return 2;
}
else if (diffRsCsdiff >= -169) {
if (Cs < 621) {
if (Rs < 3405) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
if (Rs < 3314) {
return 8;
}
else if (Rs >= 3314) {
return 1;
}
}
}
else if (Rs >= 3405) {
return 2;
}
}
else if (Cs >= 621) {
if (tcor < 82) {
if (TSC < 14) {
if (negBalance < 65) {
if (RsCsDiff < 797) {
if (tcor < 69) {
if (Cs < 878) {
return 2;
}
else if (Cs >= 878) {
if (Rs < 4054) {
return 8;
}
else if (Rs >= 4054) {
if (Rs < 4389) {
return 2;
}
else if (Rs >= 4389) {
if (Rs < 4683) {
return 8;
}
else if (Rs >= 4683) {
return 2;
}
}
}
}
}
else if (tcor >= 69) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 4507) {
return 8;
}
else if (Rs >= 4507) {
if (Rs < 5421) {
return 2;
}
else if (Rs >= 5421) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 797) {
if (mcTcor < 95) {
if (negBalance < 56) {
return 2;
}
else if (negBalance >= 56) {
if (Rs < 5974) {
return 2;
}
else if (Rs >= 5974) {
return 8;
}
}
}
else if (mcTcor >= 95) {
return 8;
}
}
}
else if (negBalance >= 65) {
if (Rs < 3313) {
return 1;
}
else if (Rs >= 3313) {
return 2;
}
}
}
else if (TSC >= 14) {
return 1;
}
}
else if (tcor >= 82) {
if (Rs < 5042) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (MVDiff < 26) {
if (Rs < 3565) {
return 8;
}
else if (Rs >= 3565) {
return 2;
}
}
else if (MVDiff >= 26) {
if (AFD < 15) {
if (Rs < 4817) {
return 2;
}
else if (Rs >= 4817) {
return 8;
}
}
else if (AFD >= 15) {
return 1;
}
}
}
}
else if (Rs >= 5042) {
if (negBalance < 53) {
return 2;
}
else if (negBalance >= 53) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 131) {
if (mcTcor < 92) {
if (AFD < 19) {
if (ssDCval < 151) {
if (Cs < 1261) {
return 4;
}
else if (Cs >= 1261) {
return 8;
}
}
else if (ssDCval >= 151) {
if (Rs < 4561) {
return 8;
}
else if (Rs >= 4561) {
return 2;
}
}
}
else if (AFD >= 19) {
if (CsDiff < 5) {
if (RsDiff < 19) {
return 8;
}
else if (RsDiff >= 19) {
return 2;
}
}
else if (CsDiff >= 5) {
if (ssDCval < 141) {
if (Rs < 2964) {
return 8;
}
else if (Rs >= 2964) {
return 2;
}
}
else if (ssDCval >= 141) {
if (MVDiff < 81) {
return 1;
}
else if (MVDiff >= 81) {
return 2;
}
}
}
}
}
else if (mcTcor >= 92) {
if (SC < 3688) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (CsDiff < 4) {
if (Rs < 2587) {
return 1;
}
else if (Rs >= 2587) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
if (Rs < 2851) {
return 2;
}
else if (Rs >= 2851) {
if (Cs < 480) {
return 1;
}
else if (Cs >= 480) {
return 2;
}
}
}
}
}
else if (CsDiff >= 4) {
return 1;
}
}
}
else if (SC >= 3688) {
if (CsDiff < 4) {
if (mcTcor < 94) {
if (Cs < 531) {
if (Rs < 4720) {
return 1;
}
else if (Rs >= 4720) {
return 2;
}
}
else if (Cs >= 531) {
if (SC < 3994) {
return 8;
}
else if (SC >= 3994) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (gchDC < 1) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (gchDC >= 1) {
if (Cs < 1041) {
if (ssDCval < 153) {
if (ssDCval < 133) {
if (Rs < 3910) {
return 2;
}
else if (Rs >= 3910) {
return 1;
}
}
else if (ssDCval >= 133) {
return 2;
}
}
else if (ssDCval >= 153) {
return 1;
}
}
else if (Cs >= 1041) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 1;
}
}
}
}
}
else if (CsDiff >= 4) {
if (mu_mv_mag_sq < 9) {
if (tcor < 82) {
if (CsDiff < 7) {
if (Cs < 613) {
return 1;
}
else if (Cs >= 613) {
return 8;
}
}
else if (CsDiff >= 7) {
if (Rs < 4714) {
return 1;
}
else if (Rs >= 4714) {
return 8;
}
}
}
else if (tcor >= 82) {
if (Rs < 4342) {
if (Rs < 2794) {
return 8;
}
else if (Rs >= 2794) {
if (diffMVdiffVal < 40) {
return 1;
}
else if (diffMVdiffVal >= 40) {
return 8;
}
}
}
else if (Rs >= 4342) {
if (gchDC < 2) {
if (Cs < 1207) {
return 2;
}
else if (Cs >= 1207) {
if (Rs < 6171) {
return 8;
}
else if (Rs >= 6171) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (diffRsCsdiff < 94) {
return 1;
}
else if (diffRsCsdiff >= 94) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (MVDiff < 59) {
if (Cs < 435) {
return 2;
}
else if (Cs >= 435) {
if (AFD < 9) {
if (Rs < 5204) {
return 8;
}
else if (Rs >= 5204) {
return 2;
}
}
else if (AFD >= 9) {
if (ssDCval < 133) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 2;
}
}
else if (ssDCval >= 133) {
return 1;
}
}
}
}
else if (MVDiff >= 59) {
if (Cs < 1274) {
return 2;
}
else if (Cs >= 1274) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1313) {
if (mu_mv_mag_sq < 107) {
if (RsCsDiff < 101) {
if (refDCval < 90) {
if (posBalance < 45) {
if (mu_mv_mag_sq < 29) {
if (RsDiff < 5) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (Rs < 858) {
return 1;
}
else if (Rs >= 858) {
if (Rs < 1874) {
return 4;
}
else if (Rs >= 1874) {
return 8;
}
}
}
}
else if (RsDiff >= 5) {
if (Rs < 3234) {
if (Rs < 2453) {
return 8;
}
else if (Rs >= 2453) {
return 2;
}
}
else if (Rs >= 3234) {
if (Cs < 1635) {
return 4;
}
else if (Cs >= 1635) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (Rs < 8905) {
if (ssDCval < 64) {
if (Rs < 5248) {
if (posBalance < 31) {
if (Rs < 1048) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
return 8;
}
}
else if (Rs >= 1048) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
}
else if (posBalance >= 31) {
if (mu_mv_mag_sq < 96) {
if (posBalance < 43) {
if (diffAFD < -1) {
if (diffMVdiffVal < -1) {
if (Rs < 3301) {
return 1;
}
else if (Rs >= 3301) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
if (negBalance < 32) {
if (Cs < 1347) {
return 8;
}
else if (Cs >= 1347) {
return 2;
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (diffRsCsdiff < 21) {
if (RsDiff < 6) {
if (MVDiff < 2) {
if (Rs < 2133) {
return 8;
}
else if (Rs >= 2133) {
return 2;
}
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (diffRsCsdiff >= 21) {
if (Rs < 1020) {
return 1;
}
else if (Rs >= 1020) {
return 8;
}
}
}
}
else if (posBalance >= 43) {
if (diffAFD < -2) {
if (Rs < 848) {
return 1;
}
else if (Rs >= 848) {
return 2;
}
}
else if (diffAFD >= -2) {
if (RsCsDiff < 43) {
return 8;
}
else if (RsCsDiff >= 43) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 796) {
return 1;
}
else if (Rs >= 796) {
if (Rs < 2811) {
if (Rs < 1091) {
if (Rs < 1001) {
return 8;
}
else if (Rs >= 1001) {
return 4;
}
}
else if (Rs >= 1091) {
return 8;
}
}
else if (Rs >= 2811) {
return 4;
}
}
}
}
}
else if (Rs >= 5248) {
return 4;
}
}
else if (ssDCval >= 64) {
if (refDCval < 69) {
if (MVDiff < 7) {
if (Rs < 3954) {
if (posBalance < 42) {
if (Rs < 2709) {
return 8;
}
else if (Rs >= 2709) {
return 1;
}
}
else if (posBalance >= 42) {
return 1;
}
}
else if (Rs >= 3954) {
return 4;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (refDCval >= 69) {
if (Rs < 3505) {
if (Cs < 1504) {
if (Rs < 2068) {
return 8;
}
else if (Rs >= 2068) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (Rs < 2539) {
return 4;
}
else if (Rs >= 2539) {
return 2;
}
}
}
}
else if (Cs >= 1504) {
if (RsCsDiff < 77) {
if (mu_mv_mag_sq < 98) {
if (Rs < 947) {
return 2;
}
else if (Rs >= 947) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (Rs < 2768) {
if (Cs < 1760) {
return 8;
}
else if (Cs >= 1760) {
return 1;
}
}
else if (Rs >= 2768) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (negBalance < 34) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
if (RsCsDiff < 60) {
return 8;
}
else if (RsCsDiff >= 60) {
return 1;
}
}
}
else if (negBalance >= 34) {
return 4;
}
}
}
else if (RsCsDiff >= 77) {
if (MVDiff < 5) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 1;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (Rs >= 3505) {
if (SC < 5840) {
if (RsCsDiff < 69) {
if (diffAFD < -1) {
if (negBalance < 37) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (negBalance >= 37) {
if (Rs < 3847) {
return 2;
}
else if (Rs >= 3847) {
return 1;
}
}
}
else if (diffAFD >= -1) {
if (tcor < 89) {
if (Rs < 3562) {
return 1;
}
else if (Rs >= 3562) {
return 8;
}
}
else if (tcor >= 89) {
return 4;
}
}
}
else if (RsCsDiff >= 69) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
if (Cs < 1779) {
return 4;
}
else if (Cs >= 1779) {
return 8;
}
}
}
}
else if (SC >= 5840) {
if (TSC < 7) {
if (SC < 7108) {
if (TSC < 6) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
else if (TSC >= 6) {
if (Rs < 4512) {
return 4;
}
else if (Rs >= 4512) {
return 8;
}
}
}
else if (SC >= 7108) {
if (Cs < 4123) {
return 4;
}
else if (Cs >= 4123) {
if (Rs < 5589) {
return 8;
}
else if (Rs >= 5589) {
return 2;
}
}
}
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
}
else if (Rs >= 8905) {
return 2;
}
}
}
else if (posBalance >= 45) {
if (Rs < 1486) {
if (mcTcor < 98) {
if (AFD < 10) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 87) {
if (TSC < 6) {
if (Cs < 1605) {
if (negBalance < 37) {
if (tcor < 87) {
if (Rs < 1050) {
return 4;
}
else if (Rs >= 1050) {
return 8;
}
}
else if (tcor >= 87) {
return 2;
}
}
else if (negBalance >= 37) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
}
else if (Cs >= 1605) {
if (Rs < 1421) {
return 8;
}
else if (Rs >= 1421) {
return 1;
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 87) {
if (Cs < 1343) {
if (MVDiff < 3) {
if (Cs < 1329) {
return 8;
}
else if (Cs >= 1329) {
return 4;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (Cs >= 1343) {
if (Rs < 838) {
if (Rs < 827) {
return 8;
}
else if (Rs >= 827) {
return 1;
}
}
else if (Rs >= 838) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 1052) {
if (negBalance < 36) {
if (mu_mv_mag_sq < 78) {
if (SC < 2285) {
if (Rs < 877) {
return 2;
}
else if (Rs >= 877) {
return 4;
}
}
else if (SC >= 2285) {
return 1;
}
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 714) {
return 1;
}
else if (Rs >= 714) {
return 8;
}
}
}
else if (negBalance >= 36) {
if (tcor < 93) {
if (SC < 2132) {
return 2;
}
else if (SC >= 2132) {
return 8;
}
}
else if (tcor >= 93) {
if (Cs < 1333) {
return 8;
}
else if (Cs >= 1333) {
return 2;
}
}
}
}
else if (Rs >= 1052) {
if (Rs < 1303) {
if (SC < 2773) {
if (ssDCval < 66) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (ssDCval >= 66) {
if (Rs < 1165) {
return 1;
}
else if (Rs >= 1165) {
return 8;
}
}
}
else if (SC >= 2773) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (Cs < 2198) {
return 1;
}
else if (Cs >= 2198) {
return 8;
}
}
}
}
else if (Rs >= 1303) {
if (ssDCval < 69) {
if (ssDCval < 48) {
if (Cs < 1473) {
return 8;
}
else if (Cs >= 1473) {
return 4;
}
}
else if (ssDCval >= 48) {
if (gchDC < 2) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (ssDCval >= 69) {
if (Rs < 1463) {
return 8;
}
else if (Rs >= 1463) {
return 4;
}
}
}
}
}
}
else if (AFD >= 10) {
if (TSC < 12) {
if (ssDCval < 75) {
if (refDCval < 59) {
if (MVDiff < 20) {
if (RsCsDiff < 63) {
if (tcor < 79) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
else if (tcor >= 79) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (Rs < 1093) {
return 2;
}
else if (Rs >= 1093) {
return 4;
}
}
}
}
else if (RsCsDiff >= 63) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 4;
}
}
}
else if (MVDiff >= 20) {
if (Rs < 1319) {
if (refDCval < 55) {
return 2;
}
else if (refDCval >= 55) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
return 4;
}
}
}
else if (Rs >= 1319) {
if (Rs < 1399) {
return 4;
}
else if (Rs >= 1399) {
return 8;
}
}
}
}
else if (refDCval >= 59) {
if (mu_mv_mag_sq < 3) {
if (RsCsDiff < 77) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
if (MVDiff < 14) {
if (SC < 2168) {
return 8;
}
else if (SC >= 2168) {
return 2;
}
}
else if (MVDiff >= 14) {
return 8;
}
}
}
else if (RsCsDiff >= 77) {
if (diffAFD < 1) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (Rs < 1246) {
return 2;
}
else if (Rs >= 1246) {
if (Rs < 1428) {
return 1;
}
else if (Rs >= 1428) {
return 8;
}
}
}
}
else if (diffAFD >= 1) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 2437) {
if (AFD < 13) {
if (mu_mv_mag_sq < 52) {
return 1;
}
else if (mu_mv_mag_sq >= 52) {
return 8;
}
}
else if (AFD >= 13) {
if (Rs < 710) {
return 2;
}
else if (Rs >= 710) {
return 8;
}
}
}
else if (SC >= 2437) {
if (diffRsCsdiff < -24) {
return 4;
}
else if (diffRsCsdiff >= -24) {
if (SC < 2992) {
return 8;
}
else if (SC >= 2992) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (mu_mv_mag_sq < 3) {
if (RsDiff < 4) {
if (Cs < 1471) {
if (Cs < 1452) {
return 4;
}
else if (Cs >= 1452) {
return 1;
}
}
else if (Cs >= 1471) {
if (Cs < 1555) {
return 8;
}
else if (Cs >= 1555) {
return 4;
}
}
}
else if (RsDiff >= 4) {
if (posBalance < 60) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (posBalance >= 60) {
if (AFD < 13) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (AFD >= 13) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (MVDiff < 21) {
if (ssDCval < 81) {
if (Rs < 1464) {
return 4;
}
else if (Rs >= 1464) {
return 8;
}
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (MVDiff >= 21) {
if (Rs < 1361) {
return 4;
}
else if (Rs >= 1361) {
return 8;
}
}
}
}
}
else if (TSC >= 12) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Rs < 1470) {
return 4;
}
else if (Rs >= 1470) {
return 8;
}
}
}
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 94) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (TSC < 5) {
if (Cs < 1668) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
if (Rs < 666) {
return 1;
}
else if (Rs >= 666) {
return 2;
}
}
}
else if (Cs >= 1668) {
return 1;
}
}
else if (TSC >= 5) {
if (Rs < 632) {
return 8;
}
else if (Rs >= 632) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (diffRsCsdiff < -4) {
return 1;
}
else if (diffRsCsdiff >= -4) {
if (Cs < 1463) {
if (SC < 2164) {
return 1;
}
else if (SC >= 2164) {
return 8;
}
}
else if (Cs >= 1463) {
if (Rs < 1317) {
return 8;
}
else if (Rs >= 1317) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1486) {
if (negBalance < 40) {
if (Rs < 2792) {
if (Cs < 2481) {
if (Rs < 2590) {
if (TSC < 9) {
if (mu_mv_mag_sq < 87) {
if (tcor < 90) {
if (Rs < 1545) {
if (Rs < 1509) {
return 4;
}
else if (Rs >= 1509) {
return 1;
}
}
else if (Rs >= 1545) {
if (negBalance < 35) {
if (SC < 3728) {
return 8;
}
else if (SC >= 3728) {
if (Rs < 2515) {
return 4;
}
else if (Rs >= 2515) {
return 8;
}
}
}
else if (negBalance >= 35) {
if (diffRsCsdiff < -36) {
if (Rs < 2166) {
return 4;
}
else if (Rs >= 2166) {
return 8;
}
}
else if (diffRsCsdiff >= -36) {
return 8;
}
}
}
}
else if (tcor >= 90) {
if (refDCval < 77) {
return 8;
}
else if (refDCval >= 77) {
if (ssDCval < 84) {
if (Cs < 1714) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Cs < 1372) {
return 1;
}
else if (Cs >= 1372) {
return 8;
}
}
}
else if (Cs >= 1714) {
return 1;
}
}
else if (ssDCval >= 84) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (Cs < 1580) {
if (mu_mv_mag_sq < 104) {
if (Rs < 2290) {
return 8;
}
else if (Rs >= 2290) {
if (Rs < 2455) {
return 4;
}
else if (Rs >= 2455) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 1703) {
return 4;
}
else if (Rs >= 1703) {
if (Rs < 1880) {
return 1;
}
else if (Rs >= 1880) {
return 8;
}
}
}
}
else if (Cs >= 1580) {
return 8;
}
}
}
else if (TSC >= 9) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (Rs >= 2590) {
if (ssDCval < 76) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (ssDCval >= 76) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (ssDCval < 83) {
return 1;
}
else if (ssDCval >= 83) {
return 8;
}
}
}
}
}
else if (Cs >= 2481) {
return 4;
}
}
else if (Rs >= 2792) {
if (Cs < 1878) {
if (Rs < 3427) {
if (Cs < 1602) {
if (diffRsCsdiff < -13) {
return 2;
}
else if (diffRsCsdiff >= -13) {
if (mu_mv_mag_sq < 92) {
if (RsCsDiff < 82) {
return 4;
}
else if (RsCsDiff >= 82) {
return 2;
}
}
else if (mu_mv_mag_sq >= 92) {
return 2;
}
}
}
else if (Cs >= 1602) {
if (refDCval < 79) {
return 8;
}
else if (refDCval >= 79) {
return 4;
}
}
}
else if (Rs >= 3427) {
if (Cs < 1321) {
if (Cs < 1316) {
return 2;
}
else if (Cs >= 1316) {
return 4;
}
}
else if (Cs >= 1321) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 4;
}
}
}
}
else if (Cs >= 1878) {
if (RsCsDiff < 69) {
if (Cs < 1941) {
if (Rs < 4271) {
return 1;
}
else if (Rs >= 4271) {
return 4;
}
}
else if (Cs >= 1941) {
if (ssDCval < 54) {
if (Rs < 3301) {
return 8;
}
else if (Rs >= 3301) {
return 4;
}
}
else if (ssDCval >= 54) {
if (posBalance < 49) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
else if (RsCsDiff >= 69) {
if (Rs < 3758) {
if (Rs < 2972) {
return 8;
}
else if (Rs >= 2972) {
return 1;
}
}
else if (Rs >= 3758) {
if (ssDCval < 71) {
return 4;
}
else if (ssDCval >= 71) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 40) {
if (refDCval < 74) {
if (Cs < 3026) {
if (Rs < 3493) {
if (tcor < 91) {
if (RsCsDiff < 83) {
if (posBalance < 68) {
if (mcTcor < 96) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (Cs < 1634) {
if (Rs < 1746) {
return 4;
}
else if (Rs >= 1746) {
return 1;
}
}
else if (Cs >= 1634) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
return 4;
}
}
else if (posBalance >= 68) {
return 4;
}
}
else if (RsCsDiff >= 83) {
if (tcor < 55) {
if (Rs < 1603) {
return 4;
}
else if (Rs >= 1603) {
return 1;
}
}
else if (tcor >= 55) {
if (mcTcor < 94) {
if (Cs < 1592) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
if (Rs < 2149) {
return 1;
}
else if (Rs >= 2149) {
return 8;
}
}
}
}
else if (Cs >= 1592) {
if (Rs < 1955) {
if (AFD < 12) {
if (Cs < 1703) {
if (Rs < 1700) {
return 4;
}
else if (Rs >= 1700) {
return 8;
}
}
else if (Cs >= 1703) {
return 4;
}
}
else if (AFD >= 12) {
if (Cs < 1632) {
return 4;
}
else if (Cs >= 1632) {
return 8;
}
}
}
else if (Rs >= 1955) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
}
}
}
else if (tcor >= 91) {
if (Cs < 1522) {
if (MVDiff < 5) {
if (Rs < 1830) {
return 8;
}
else if (Rs >= 1830) {
return 2;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (Cs >= 1522) {
return 8;
}
}
}
else if (Rs >= 3493) {
if (Cs < 1636) {
return 2;
}
else if (Cs >= 1636) {
return 4;
}
}
}
else if (Cs >= 3026) {
return 8;
}
}
else if (refDCval >= 74) {
if (Cs < 2855) {
if (mcTcor < 94) {
if (AFD < 16) {
if (ssDCval < 84) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (mu_mv_mag_sq < 93) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
if (mcTcor < 93) {
if (mu_mv_mag_sq < 71) {
return 8;
}
else if (mu_mv_mag_sq >= 71) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Cs < 1920) {
return 4;
}
else if (Cs >= 1920) {
return 8;
}
}
}
}
else if (mcTcor >= 93) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
if (Rs < 1813) {
return 4;
}
else if (Rs >= 1813) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (Rs < 2241) {
return 8;
}
else if (Rs >= 2241) {
return 4;
}
}
}
}
else if (ssDCval >= 84) {
if (negBalance < 45) {
if (Cs < 1651) {
return 2;
}
else if (Cs >= 1651) {
return 8;
}
}
else if (negBalance >= 45) {
if (CsDiff < 8) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (CsDiff >= 8) {
if (Rs < 2212) {
return 1;
}
else if (Rs >= 2212) {
return 8;
}
}
}
}
}
else if (AFD >= 16) {
if (Rs < 1667) {
if (Rs < 1497) {
return 8;
}
else if (Rs >= 1497) {
return 4;
}
}
else if (Rs >= 1667) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (tcor < 82) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 8;
}
}
else if (tcor >= 82) {
return 8;
}
}
}
}
}
else if (mcTcor >= 94) {
if (Cs < 1551) {
if (RsDiff < 8) {
if (MVDiff < 7) {
if (refDCval < 80) {
if (AFD < 11) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
if (CsDiff < 4) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 8;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (refDCval >= 80) {
if (TSC < 7) {
if (Cs < 1436) {
if (Cs < 1344) {
return 2;
}
else if (Cs >= 1344) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
else if (Cs >= 1436) {
if (Rs < 2391) {
return 2;
}
else if (Rs >= 2391) {
return 8;
}
}
}
else if (TSC >= 7) {
if (mu_mv_mag_sq < 94) {
if (mu_mv_mag_sq < 51) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 51) {
return 1;
}
}
else if (mu_mv_mag_sq >= 94) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 1733) {
if (AFD < 14) {
if (Cs < 1386) {
return 4;
}
else if (Cs >= 1386) {
return 2;
}
}
else if (AFD >= 14) {
return 8;
}
}
else if (Rs >= 1733) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (Rs < 2445) {
if (SC < 3678) {
if (Rs < 2179) {
return 2;
}
else if (Rs >= 2179) {
return 8;
}
}
else if (SC >= 3678) {
return 4;
}
}
else if (Rs >= 2445) {
return 2;
}
}
}
else if (Cs >= 1551) {
if (Rs < 2115) {
if (SC < 4092) {
if (refDCval < 78) {
return 4;
}
else if (refDCval >= 78) {
if (tcor < 88) {
if (Rs < 1583) {
return 2;
}
else if (Rs >= 1583) {
return 4;
}
}
else if (tcor >= 88) {
if (ssDCval < 89) {
if (Rs < 2052) {
return 8;
}
else if (Rs >= 2052) {
if (Rs < 2071) {
return 4;
}
else if (Rs >= 2071) {
return 8;
}
}
}
else if (ssDCval >= 89) {
return 2;
}
}
}
}
else if (SC >= 4092) {
if (Cs < 2151) {
return 8;
}
else if (Cs >= 2151) {
return 1;
}
}
}
else if (Rs >= 2115) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (ssDCval < 84) {
if (Rs < 3255) {
if (Rs < 2334) {
if (tcor < 92) {
if (Cs < 2030) {
return 4;
}
else if (Cs >= 2030) {
return 8;
}
}
else if (tcor >= 92) {
if (Cs < 2501) {
return 8;
}
else if (Cs >= 2501) {
return 4;
}
}
}
else if (Rs >= 2334) {
return 8;
}
}
else if (Rs >= 3255) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (ssDCval >= 84) {
if (posBalance < 58) {
if (diffAFD < -1) {
if (Cs < 1742) {
if (Rs < 2484) {
return 8;
}
else if (Rs >= 2484) {
return 1;
}
}
else if (Cs >= 1742) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (posBalance >= 58) {
if (refDCval < 86) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (refDCval >= 86) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 2855) {
if (Rs < 2944) {
return 4;
}
else if (Rs >= 2944) {
if (AFD < 10) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (AFD >= 10) {
if (Rs < 4240) {
if (Cs < 2928) {
return 8;
}
else if (Cs >= 2928) {
return 1;
}
}
else if (Rs >= 4240) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 90) {
if (refDCval < 163) {
if (Rs < 4344) {
if (ssDCval < 110) {
if (mcTcor < 96) {
if (Rs < 2149) {
if (Rs < 1416) {
if (Cs < 1990) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Cs < 1602) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
if (AFD < 29) {
return 4;
}
else if (AFD >= 29) {
if (Rs < 1310) {
return 2;
}
else if (Rs >= 1310) {
return 8;
}
}
}
}
}
else if (Cs >= 1602) {
if (negBalance < 55) {
return 4;
}
else if (negBalance >= 55) {
return 8;
}
}
}
}
else if (Cs >= 1990) {
if (AFD < 13) {
if (negBalance < 46) {
if (Rs < 1368) {
return 4;
}
else if (Rs >= 1368) {
return 2;
}
}
else if (negBalance >= 46) {
return 8;
}
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (Rs >= 1416) {
if (AFD < 17) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (RsDiff < 7) {
if (AFD < 11) {
if (mu_mv_mag_sq < 19) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
return 8;
}
}
else if (mu_mv_mag_sq >= 19) {
if (negBalance < 57) {
if (diffRsCsdiff < -21) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (Rs < 1802) {
return 1;
}
else if (Rs >= 1802) {
return 2;
}
}
}
else if (diffRsCsdiff >= -21) {
if (diffRsCsdiff < 13) {
return 8;
}
else if (diffRsCsdiff >= 13) {
if (Rs < 2027) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Rs >= 2027) {
return 2;
}
}
}
}
else if (negBalance >= 57) {
if (Cs < 1525) {
return 2;
}
else if (Cs >= 1525) {
return 8;
}
}
}
}
else if (AFD >= 11) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Rs < 1599) {
return 8;
}
else if (Rs >= 1599) {
return 4;
}
}
}
}
else if (RsDiff >= 7) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (RsDiff < 8) {
if (mcTcor < 94) {
if (Rs < 2143) {
return 8;
}
else if (Rs >= 2143) {
return 2;
}
}
else if (mcTcor >= 94) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
}
else if (AFD >= 17) {
if (TSC < 13) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
return 8;
}
}
}
else if (TSC >= 13) {
return 1;
}
}
}
}
else if (Rs >= 2149) {
if (Cs < 2428) {
if (tcor < 90) {
if (diffTSC < -4) {
return 4;
}
else if (diffTSC >= -4) {
if (MVDiff < 67) {
if (diffMVdiffVal < -14) {
if (Rs < 3133) {
return 4;
}
else if (Rs >= 3133) {
return 8;
}
}
else if (diffMVdiffVal >= -14) {
if (Rs < 2225) {
if (Rs < 2221) {
return 8;
}
else if (Rs >= 2221) {
return 2;
}
}
else if (Rs >= 2225) {
return 8;
}
}
}
else if (MVDiff >= 67) {
return 4;
}
}
}
else if (tcor >= 90) {
if (RsCsDiff < 63) {
if (TSC < 3) {
if (Rs < 2319) {
return 1;
}
else if (Rs >= 2319) {
return 4;
}
}
else if (TSC >= 3) {
if (negBalance < 55) {
if (refDCval < 107) {
if (Rs < 4131) {
if (mu_mv_mag_sq < 40) {
if (mu_mv_mag_sq < 20) {
return 8;
}
else if (mu_mv_mag_sq >= 20) {
return 4;
}
}
else if (mu_mv_mag_sq >= 40) {
return 8;
}
}
else if (Rs >= 4131) {
return 4;
}
}
else if (refDCval >= 107) {
if (Cs < 1854) {
if (posBalance < 55) {
if (MVDiff < 7) {
if (Rs < 3189) {
return 8;
}
else if (Rs >= 3189) {
return 4;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (Cs >= 1854) {
return 8;
}
}
}
else if (negBalance >= 55) {
if (Rs < 2685) {
return 8;
}
else if (Rs >= 2685) {
return 4;
}
}
}
}
else if (RsCsDiff >= 63) {
if (mcTcor < 93) {
if (Cs < 1566) {
return 1;
}
else if (Cs >= 1566) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 1;
}
}
}
else if (mcTcor >= 93) {
if (RsDiff < 8) {
if (negBalance < 33) {
if (Rs < 2736) {
return 8;
}
else if (Rs >= 2736) {
return 1;
}
}
else if (negBalance >= 33) {
return 8;
}
}
else if (RsDiff >= 8) {
if (mu_mv_mag_sq < 97) {
if (Cs < 1377) {
return 4;
}
else if (Cs >= 1377) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (Rs < 3652) {
return 4;
}
else if (Rs >= 3652) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 2428) {
if (MVDiff < 7) {
if (ssDCval < 107) {
if (Cs < 2490) {
if (Rs < 3436) {
return 4;
}
else if (Rs >= 3436) {
return 8;
}
}
else if (Cs >= 2490) {
return 4;
}
}
else if (ssDCval >= 107) {
return 8;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (RsCsDiff < 41) {
if (MVDiff < 4) {
if (Cs < 1576) {
if (ssDCval < 96) {
if (ssDCval < 93) {
if (Rs < 4176) {
return 8;
}
else if (Rs >= 4176) {
return 4;
}
}
else if (ssDCval >= 93) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (ssDCval >= 96) {
if (TSC < 6) {
if (ssDCval < 105) {
return 4;
}
else if (ssDCval >= 105) {
if (Rs < 2272) {
if (Rs < 681) {
return 1;
}
else if (Rs >= 681) {
return 4;
}
}
else if (Rs >= 2272) {
return 8;
}
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (Cs >= 1576) {
if (mcTcor < 98) {
if (gchDC < 1) {
if (Rs < 2376) {
if (Rs < 1879) {
return 2;
}
else if (Rs >= 1879) {
return 8;
}
}
else if (Rs >= 2376) {
if (Cs < 1629) {
if (Cs < 1611) {
return 4;
}
else if (Cs >= 1611) {
return 8;
}
}
else if (Cs >= 1629) {
return 4;
}
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (mcTcor >= 98) {
if (RsDiff < 1) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (CsDiff < 1) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 2285) {
if (Rs < 2188) {
return 1;
}
else if (Rs >= 2188) {
return 2;
}
}
else if (Rs >= 2285) {
return 8;
}
}
}
else if (CsDiff >= 1) {
if (mcTcor < 97) {
if (Rs < 2870) {
if (Rs < 2420) {
if (Cs < 1401) {
if (Rs < 1459) {
return 4;
}
else if (Rs >= 1459) {
return 2;
}
}
else if (Cs >= 1401) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (Rs >= 2420) {
if (Rs < 2565) {
return 1;
}
else if (Rs >= 2565) {
return 4;
}
}
}
else if (Rs >= 2870) {
if (posBalance < 39) {
if (Cs < 2934) {
return 4;
}
else if (Cs >= 2934) {
return 8;
}
}
else if (posBalance >= 39) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 23) {
if (ssDCval < 95) {
return 1;
}
else if (ssDCval >= 95) {
if (CsDiff < 2) {
if (Rs < 2158) {
return 1;
}
else if (Rs >= 2158) {
return 8;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (posBalance >= 23) {
if (ssDCval < 107) {
return 8;
}
else if (ssDCval >= 107) {
if (RsCsDiff < 32) {
if (Cs < 3124) {
return 4;
}
else if (Cs >= 3124) {
return 8;
}
}
else if (RsCsDiff >= 32) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 41) {
if (Rs < 3697) {
if (posBalance < 52) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (MVDiff < 3) {
if (refDCval < 96) {
return 1;
}
else if (refDCval >= 96) {
if (Rs < 2927) {
return 8;
}
else if (Rs >= 2927) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (refDCval < 107) {
if (TSC < 8) {
if (SC < 5317) {
return 8;
}
else if (SC >= 5317) {
return 2;
}
}
else if (TSC >= 8) {
if (SC < 3725) {
return 8;
}
else if (SC >= 3725) {
return 1;
}
}
}
else if (refDCval >= 107) {
if (posBalance < 42) {
if (Rs < 1501) {
return 4;
}
else if (Rs >= 1501) {
return 8;
}
}
else if (posBalance >= 42) {
if (Rs < 2258) {
return 2;
}
else if (Rs >= 2258) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 52) {
if (mu_mv_mag_sq < 7) {
if (gchDC < 2) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (Cs < 1795) {
if (Cs < 1351) {
return 2;
}
else if (Cs >= 1351) {
return 8;
}
}
else if (Cs >= 1795) {
if (Rs < 1831) {
return 8;
}
else if (Rs >= 1831) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 1919) {
if (negBalance < 39) {
if (diffRsCsdiff < -5) {
if (Cs < 1468) {
return 4;
}
else if (Cs >= 1468) {
return 2;
}
}
else if (diffRsCsdiff >= -5) {
return 8;
}
}
else if (negBalance >= 39) {
if (SC < 3600) {
if (Rs < 1785) {
return 8;
}
else if (Rs >= 1785) {
return 1;
}
}
else if (SC >= 3600) {
return 8;
}
}
}
else if (Cs >= 1919) {
if (Rs < 2717) {
if (Rs < 1974) {
return 2;
}
else if (Rs >= 1974) {
return 8;
}
}
else if (Rs >= 2717) {
return 4;
}
}
}
}
}
else if (Rs >= 3697) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
if (SC < 6009) {
if (MVDiff < 12) {
if (ssDCval < 101) {
if (Rs < 4161) {
return 2;
}
else if (Rs >= 4161) {
if (Rs < 4224) {
return 8;
}
else if (Rs >= 4224) {
return 2;
}
}
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (SC >= 6009) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 110) {
if (Rs < 1489) {
if (mu_mv_mag_sq < 36) {
if (ssDCval < 145) {
if (AFD < 18) {
if (Rs < 1347) {
return 8;
}
else if (Rs >= 1347) {
if (tcor < 89) {
if (Cs < 1603) {
return 1;
}
else if (Cs >= 1603) {
return 4;
}
}
else if (tcor >= 89) {
return 8;
}
}
}
else if (AFD >= 18) {
if (Cs < 1487) {
if (Rs < 867) {
return 8;
}
else if (Rs >= 867) {
return 1;
}
}
else if (Cs >= 1487) {
if (Rs < 1044) {
return 2;
}
else if (Rs >= 1044) {
return 4;
}
}
}
}
else if (ssDCval >= 145) {
return 2;
}
}
else if (mu_mv_mag_sq >= 36) {
if (Rs < 1262) {
if (Rs < 830) {
if (SC < 1997) {
return 1;
}
else if (SC >= 1997) {
return 8;
}
}
else if (Rs >= 830) {
if (CsDiff < 6) {
if (mcTcor < 96) {
if (diffRsCsdiff < 1) {
if (ssDCval < 116) {
return 4;
}
else if (ssDCval >= 116) {
if (Rs < 980) {
return 8;
}
else if (Rs >= 980) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (negBalance < 38) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
return 4;
}
}
else if (negBalance >= 38) {
return 4;
}
}
}
}
}
else if (diffRsCsdiff >= 1) {
if (SC < 2564) {
return 8;
}
else if (SC >= 2564) {
if (SC < 2667) {
return 4;
}
else if (SC >= 2667) {
if (negBalance < 41) {
return 8;
}
else if (negBalance >= 41) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (Rs < 1196) {
return 8;
}
else if (Rs >= 1196) {
if (Cs < 1727) {
return 8;
}
else if (Cs >= 1727) {
return 4;
}
}
}
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (Rs >= 1262) {
if (SC < 3104) {
if (diffMVdiffVal < -14) {
return 1;
}
else if (diffMVdiffVal >= -14) {
if (Cs < 1315) {
return 2;
}
else if (Cs >= 1315) {
return 8;
}
}
}
else if (SC >= 3104) {
if (RsDiff < 6) {
if (Rs < 1436) {
return 8;
}
else if (Rs >= 1436) {
return 1;
}
}
else if (RsDiff >= 6) {
return 4;
}
}
}
}
}
else if (Rs >= 1489) {
if (AFD < 14) {
if (ssDCval < 140) {
if (ssDCval < 123) {
if (TSC < 11) {
if (Rs < 4303) {
if (refDCval < 121) {
if (mu_mv_mag_sq < 3) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (Rs < 2586) {
return 2;
}
else if (Rs >= 2586) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (refDCval >= 121) {
if (mcTcor < 99) {
if (tcor < 81) {
if (Rs < 2494) {
return 2;
}
else if (Rs >= 2494) {
return 8;
}
}
else if (tcor >= 81) {
return 8;
}
}
else if (mcTcor >= 99) {
if (Cs < 1829) {
return 2;
}
else if (Cs >= 1829) {
return 8;
}
}
}
}
else if (Rs >= 4303) {
if (Rs < 4310) {
return 1;
}
else if (Rs >= 4310) {
return 8;
}
}
}
else if (TSC >= 11) {
if (Cs < 1784) {
return 1;
}
else if (Cs >= 1784) {
return 8;
}
}
}
else if (ssDCval >= 123) {
if (mcTcor < 90) {
if (Cs < 2063) {
if (tcor < 81) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (tcor >= 81) {
if (refDCval < 126) {
return 8;
}
else if (refDCval >= 126) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
if (Rs < 2299) {
return 1;
}
else if (Rs >= 2299) {
return 8;
}
}
}
}
}
else if (Cs >= 2063) {
return 8;
}
}
else if (mcTcor >= 90) {
if (TSC < 8) {
if (SC < 3717) {
if (SC < 3649) {
if (tcor < 82) {
return 1;
}
else if (tcor >= 82) {
return 8;
}
}
else if (SC >= 3649) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 1;
}
}
}
else if (SC >= 3717) {
if (Cs < 1360) {
if (ssDCval < 125) {
return 2;
}
else if (ssDCval >= 125) {
return 8;
}
}
else if (Cs >= 1360) {
if (mu_mv_mag_sq < 97) {
return 8;
}
else if (mu_mv_mag_sq >= 97) {
if (diffTSC < -1) {
if (MVDiff < 6) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
}
else if (TSC >= 8) {
if (mu_mv_mag_sq < 101) {
if (refDCval < 128) {
if (Cs < 1842) {
if (Rs < 2044) {
if (Rs < 1776) {
return 2;
}
else if (Rs >= 1776) {
return 8;
}
}
else if (Rs >= 2044) {
return 1;
}
}
else if (Cs >= 1842) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (refDCval >= 128) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 101) {
return 1;
}
}
}
}
}
else if (ssDCval >= 140) {
if (mu_mv_mag_sq < 46) {
if (Rs < 1758) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (Rs >= 1758) {
return 8;
}
}
else if (mu_mv_mag_sq >= 46) {
if (SC < 5884) {
return 8;
}
else if (SC >= 5884) {
return 1;
}
}
}
}
else if (AFD >= 14) {
if (mu_mv_mag_sq < 7) {
if (posBalance < 63) {
if (negBalance < 45) {
if (Rs < 1757) {
return 4;
}
else if (Rs >= 1757) {
return 1;
}
}
else if (negBalance >= 45) {
if (Cs < 1411) {
return 2;
}
else if (Cs >= 1411) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (Cs < 1779) {
return 4;
}
else if (Cs >= 1779) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (diffRsCsdiff < -35) {
if (gchDC < 1) {
if (Rs < 2332) {
return 8;
}
else if (Rs >= 2332) {
return 2;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (diffRsCsdiff >= -35) {
if (Rs < 1607) {
return 4;
}
else if (Rs >= 1607) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 4344) {
if (tcor < 95) {
if (mu_mv_mag_sq < 21) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 21) {
if (CsDiff < 7) {
if (mu_mv_mag_sq < 53) {
if (Cs < 2762) {
return 1;
}
else if (Cs >= 2762) {
return 8;
}
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 4381) {
return 4;
}
else if (Rs >= 4381) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
}
}
else if (CsDiff >= 7) {
return 1;
}
}
}
else if (tcor >= 95) {
if (ssDCval < 117) {
if (negBalance < 32) {
if (Cs < 1877) {
return 2;
}
else if (Cs >= 1877) {
return 8;
}
}
else if (negBalance >= 32) {
return 8;
}
}
else if (ssDCval >= 117) {
if (Cs < 1850) {
return 2;
}
else if (Cs >= 1850) {
if (Rs < 5961) {
return 8;
}
else if (Rs >= 5961) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 163) {
if (Rs < 2266) {
if (Rs < 1062) {
return 1;
}
else if (Rs >= 1062) {
return 2;
}
}
else if (Rs >= 2266) {
return 8;
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1937) {
if (Rs < 3319) {
if (ssDCval < 85) {
if (negBalance < 47) {
if (Rs < 2071) {
if (mu_mv_mag_sq < 12) {
if (mcTcor < 81) {
if (refDCval < 48) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
else if (refDCval >= 48) {
if (Cs < 1862) {
return 4;
}
else if (Cs >= 1862) {
return 8;
}
}
}
else if (mcTcor >= 81) {
if (AFD < 18) {
if (SC < 3281) {
return 1;
}
else if (SC >= 3281) {
return 8;
}
}
else if (AFD >= 18) {
if (Rs < 498) {
return 4;
}
else if (Rs >= 498) {
if (Rs < 768) {
return 2;
}
else if (Rs >= 768) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 12) {
if (ssDCval < 54) {
if (ssDCval < 48) {
if (Rs < 2009) {
return 8;
}
else if (Rs >= 2009) {
return 1;
}
}
else if (ssDCval >= 48) {
if (mcTcor < 83) {
return 4;
}
else if (mcTcor >= 83) {
return 1;
}
}
}
else if (ssDCval >= 54) {
if (AFD < 20) {
if (RsCsDiff < 115) {
if (Cs < 1404) {
return 8;
}
else if (Cs >= 1404) {
return 1;
}
}
else if (RsCsDiff >= 115) {
if (diffMVdiffVal < -3) {
if (diffMVdiffVal < -9) {
return 8;
}
else if (diffMVdiffVal >= -9) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
else if (AFD >= 20) {
if (diffAFD < 4) {
if (Rs < 1769) {
return 4;
}
else if (Rs >= 1769) {
return 2;
}
}
else if (diffAFD >= 4) {
return 8;
}
}
}
}
}
else if (Rs >= 2071) {
if (posBalance < 63) {
if (mcTcor < 84) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 1;
}
}
else if (mcTcor >= 84) {
if (negBalance < 37) {
if (CsDiff < 8) {
if (Cs < 1382) {
return 4;
}
else if (Cs >= 1382) {
return 2;
}
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (negBalance >= 37) {
if (posBalance < 44) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
if (Cs < 1441) {
return 4;
}
else if (Cs >= 1441) {
return 2;
}
}
}
else if (posBalance >= 44) {
if (diffMVdiffVal < -12) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
else if (diffMVdiffVal >= -12) {
if (ssDCval < 41) {
return 8;
}
else if (ssDCval >= 41) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 63) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 8;
}
}
}
}
else if (negBalance >= 47) {
if (mu_mv_mag_sq < 9) {
if (RsDiff < 14) {
if (negBalance < 62) {
if (Rs < 1683) {
if (refDCval < 70) {
if (mu_mv_mag_sq < 5) {
if (diffRsCsdiff < -16) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
if (Rs < 1454) {
return 2;
}
else if (Rs >= 1454) {
return 8;
}
}
}
else if (diffRsCsdiff >= -16) {
if (diffMVdiffVal < -7) {
if (Rs < 1503) {
return 4;
}
else if (Rs >= 1503) {
return 1;
}
}
else if (diffMVdiffVal >= -7) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 1141) {
if (Rs < 913) {
return 8;
}
else if (Rs >= 913) {
return 4;
}
}
else if (Rs >= 1141) {
return 2;
}
}
}
else if (refDCval >= 70) {
if (RsCsDiff < 156) {
if (negBalance < 53) {
if (Cs < 1426) {
return 2;
}
else if (Cs >= 1426) {
return 4;
}
}
else if (negBalance >= 53) {
if (negBalance < 58) {
if (diffRsCsdiff < 14) {
if (Rs < 1628) {
return 1;
}
else if (Rs >= 1628) {
return 8;
}
}
else if (diffRsCsdiff >= 14) {
if (Rs < 800) {
return 8;
}
else if (Rs >= 800) {
return 2;
}
}
}
else if (negBalance >= 58) {
return 2;
}
}
}
else if (RsCsDiff >= 156) {
if (mcTcor < 87) {
if (Rs < 1420) {
return 4;
}
else if (Rs >= 1420) {
return 2;
}
}
else if (mcTcor >= 87) {
if (Cs < 1529) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
return 2;
}
}
else if (Cs >= 1529) {
return 8;
}
}
}
}
}
else if (Rs >= 1683) {
if (ssDCval < 70) {
if (MVDiff < 19) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 4;
}
}
else if (MVDiff >= 19) {
if (Cs < 1763) {
if (Rs < 2004) {
return 4;
}
else if (Rs >= 2004) {
return 8;
}
}
else if (Cs >= 1763) {
return 1;
}
}
}
else if (ssDCval >= 70) {
if (RsCsDiff < 129) {
if (Rs < 1790) {
return 4;
}
else if (Rs >= 1790) {
if (TSC < 9) {
if (Cs < 1482) {
return 8;
}
else if (Cs >= 1482) {
return 4;
}
}
else if (TSC >= 9) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
}
}
else if (RsCsDiff >= 129) {
if (ssDCval < 74) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (Rs < 1744) {
return 8;
}
else if (Rs >= 1744) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
if (tcor < 76) {
return 2;
}
else if (tcor >= 76) {
if (Rs < 2113) {
return 8;
}
else if (Rs >= 2113) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 74) {
if (diffAFD < 4) {
if (Rs < 1782) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (Rs < 1754) {
return 4;
}
else if (Rs >= 1754) {
return 1;
}
}
}
else if (Rs >= 1782) {
if (Cs < 1862) {
if (TSC < 12) {
if (RsCsDiff < 201) {
return 4;
}
else if (RsCsDiff >= 201) {
if (Rs < 2350) {
return 8;
}
else if (Rs >= 2350) {
return 4;
}
}
}
else if (TSC >= 12) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 4;
}
}
}
}
else if (Cs >= 1862) {
if (Rs < 2473) {
return 8;
}
else if (Rs >= 2473) {
return 4;
}
}
}
}
else if (diffAFD >= 4) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 62) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
if (MVDiff < 59) {
return 2;
}
else if (MVDiff >= 59) {
if (Rs < 1269) {
return 8;
}
else if (Rs >= 1269) {
return 4;
}
}
}
}
}
else if (RsDiff >= 14) {
if (refDCval < 77) {
return 2;
}
else if (refDCval >= 77) {
if (MVDiff < 114) {
if (Rs < 2948) {
return 8;
}
else if (Rs >= 2948) {
return 2;
}
}
else if (MVDiff >= 114) {
if (Rs < 2111) {
return 4;
}
else if (Rs >= 2111) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (Rs < 1861) {
if (mu_mv_mag_sq < 21) {
if (gchDC < 3) {
if (AFD < 24) {
if (diffAFD < -1) {
if (Rs < 1466) {
return 8;
}
else if (Rs >= 1466) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (AFD >= 24) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
}
else if (gchDC >= 3) {
if (Rs < 1269) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
return 8;
}
}
else if (Rs >= 1269) {
if (Cs < 1474) {
return 8;
}
else if (Cs >= 1474) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (TSC < 10) {
if (RsDiff < 9) {
if (ssDCval < 76) {
if (mcTcor < 90) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
return 4;
}
}
else if (mcTcor >= 90) {
return 8;
}
}
else if (ssDCval >= 76) {
return 4;
}
}
else if (RsDiff >= 9) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (Rs < 1580) {
return 8;
}
else if (Rs >= 1580) {
return 2;
}
}
}
}
else if (TSC >= 10) {
if (Rs < 1743) {
if (diffMVdiffVal < -16) {
if (Cs < 1508) {
return 1;
}
else if (Cs >= 1508) {
return 8;
}
}
else if (diffMVdiffVal >= -16) {
if (Cs < 1340) {
return 8;
}
else if (Cs >= 1340) {
return 4;
}
}
}
else if (Rs >= 1743) {
if (Cs < 1828) {
return 1;
}
else if (Cs >= 1828) {
return 4;
}
}
}
}
}
else if (Rs >= 1861) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (RsCsDiff < 443) {
if (ssDCval < 80) {
if (mcTcor < 67) {
return 4;
}
else if (mcTcor >= 67) {
if (negBalance < 50) {
if (AFD < 16) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 4;
}
}
else if (AFD >= 16) {
if (Rs < 2259) {
return 1;
}
else if (Rs >= 2259) {
return 2;
}
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
else if (ssDCval >= 80) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
if (gchDC < 3) {
if (SC < 3928) {
if (SC < 3409) {
return 4;
}
else if (SC >= 3409) {
return 2;
}
}
else if (SC >= 3928) {
if (SC < 4631) {
if (Cs < 1736) {
return 4;
}
else if (Cs >= 1736) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 1;
}
}
}
else if (SC >= 4631) {
return 4;
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
else if (RsCsDiff >= 443) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (AFD < 19) {
if (Cs < 1582) {
if (TSC < 9) {
if (refDCval < 100) {
if (CsDiff < 8) {
if (refDCval < 94) {
return 2;
}
else if (refDCval >= 94) {
if (Rs < 2524) {
if (Cs < 1559) {
return 2;
}
else if (Cs >= 1559) {
return 8;
}
}
else if (Rs >= 2524) {
if (Rs < 3063) {
return 4;
}
else if (Rs >= 3063) {
return 2;
}
}
}
}
else if (CsDiff >= 8) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (Cs < 1564) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
return 2;
}
}
else if (Cs >= 1564) {
return 8;
}
}
}
}
}
else if (refDCval >= 100) {
if (RsCsDiff < 189) {
if (MVDiff < 15) {
if (negBalance < 50) {
if (ssDCval < 116) {
if (Rs < 1136) {
return 8;
}
else if (Rs >= 1136) {
return 4;
}
}
else if (ssDCval >= 116) {
if (Rs < 3146) {
return 8;
}
else if (Rs >= 3146) {
return 4;
}
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (MVDiff >= 15) {
if (Rs < 1936) {
if (gchDC < 2) {
if (Cs < 1540) {
return 8;
}
else if (Cs >= 1540) {
return 2;
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (Rs >= 1936) {
return 8;
}
}
}
else if (RsCsDiff >= 189) {
if (AFD < 14) {
if (RsDiff < 13) {
if (Rs < 2495) {
return 4;
}
else if (Rs >= 2495) {
if (Rs < 2931) {
return 8;
}
else if (Rs >= 2931) {
return 2;
}
}
}
else if (RsDiff >= 13) {
return 4;
}
}
else if (AFD >= 14) {
if (Rs < 1665) {
return 1;
}
else if (Rs >= 1665) {
return 2;
}
}
}
}
}
else if (TSC >= 9) {
if (TSC < 13) {
if (SC < 4318) {
if (tcor < 74) {
if (TSC < 10) {
if (ssDCval < 100) {
if (Rs < 1905) {
return 8;
}
else if (Rs >= 1905) {
return 2;
}
}
else if (ssDCval >= 100) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 2) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (Cs < 1377) {
if (TSC < 11) {
if (ssDCval < 87) {
return 1;
}
else if (ssDCval >= 87) {
return 8;
}
}
else if (TSC >= 11) {
return 1;
}
}
else if (Cs >= 1377) {
if (Rs < 1549) {
return 4;
}
else if (Rs >= 1549) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 1365) {
if (CsDiff < 9) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 2337) {
return 1;
}
else if (Rs >= 2337) {
return 2;
}
}
}
else if (CsDiff >= 9) {
return 1;
}
}
else if (Cs >= 1365) {
if (mcTcor < 84) {
if (Rs < 2243) {
return 8;
}
else if (Rs >= 2243) {
return 1;
}
}
else if (mcTcor >= 84) {
return 8;
}
}
}
}
}
else if (SC >= 4318) {
if (ssDCval < 107) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
else if (ssDCval >= 107) {
return 8;
}
}
}
else if (TSC >= 13) {
if (Rs < 1975) {
return 8;
}
else if (Rs >= 1975) {
return 1;
}
}
}
}
else if (Cs >= 1582) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2995) {
if (Rs < 1637) {
if (RsDiff < 8) {
if (Cs < 1656) {
if (Rs < 1202) {
return 2;
}
else if (Rs >= 1202) {
return 4;
}
}
else if (Cs >= 1656) {
return 8;
}
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (Rs >= 1637) {
if (SC < 4399) {
if (AFD < 11) {
if (Rs < 2126) {
return 8;
}
else if (Rs >= 2126) {
return 2;
}
}
else if (AFD >= 11) {
return 8;
}
}
else if (SC >= 4399) {
if (refDCval < 97) {
return 2;
}
else if (refDCval >= 97) {
if (diffMVdiffVal < -5) {
if (Rs < 2773) {
return 2;
}
else if (Rs >= 2773) {
if (Rs < 2954) {
return 4;
}
else if (Rs >= 2954) {
return 8;
}
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
}
}
}
else if (Rs >= 2995) {
if (mcTcor < 90) {
if (Cs < 1801) {
return 4;
}
else if (Cs >= 1801) {
return 8;
}
}
else if (mcTcor >= 90) {
if (Cs < 1620) {
return 4;
}
else if (Cs >= 1620) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 79) {
if (refDCval < 96) {
return 8;
}
else if (refDCval >= 96) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
return 1;
}
}
}
else if (mcTcor >= 79) {
if (posBalance < 48) {
if (RsCsDiff < 141) {
if (TSC < 7) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (RsCsDiff >= 141) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (mcTcor < 97) {
if (diffAFD < -1) {
if (Rs < 2504) {
return 8;
}
else if (Rs >= 2504) {
if (ssDCval < 99) {
if (Rs < 2680) {
return 8;
}
else if (Rs >= 2680) {
return 2;
}
}
else if (ssDCval >= 99) {
if (Rs < 2533) {
return 2;
}
else if (Rs >= 2533) {
return 8;
}
}
}
}
else if (diffAFD >= -1) {
if (negBalance < 46) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (MVDiff < 11) {
if (Cs < 1819) {
return 4;
}
else if (Cs >= 1819) {
return 8;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (Rs < 2232) {
return 2;
}
else if (Rs >= 2232) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 19) {
if (TSC < 10) {
if (mu_mv_mag_sq < 10) {
if (refDCval < 96) {
if (mcTcor < 94) {
if (Rs < 1739) {
return 4;
}
else if (Rs >= 1739) {
return 8;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (refDCval >= 96) {
if (Rs < 1729) {
return 4;
}
else if (Rs >= 1729) {
if (RsCsDiff < 260) {
return 1;
}
else if (RsCsDiff >= 260) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (Cs < 1396) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 2;
}
}
else if (Cs >= 1396) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
if (Rs < 2808) {
return 4;
}
else if (Rs >= 2808) {
return 2;
}
}
}
}
}
else if (TSC >= 10) {
if (refDCval < 121) {
if (mcTcor < 94) {
if (tcor < 56) {
if (mcTcor < 80) {
if (AFD < 33) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 8;
}
}
else if (AFD >= 33) {
if (Rs < 1944) {
return 4;
}
else if (Rs >= 1944) {
return 1;
}
}
}
else if (mcTcor >= 80) {
if (AFD < 26) {
if (ssDCval < 113) {
return 4;
}
else if (ssDCval >= 113) {
if (Rs < 1636) {
return 2;
}
else if (Rs >= 1636) {
return 8;
}
}
}
else if (AFD >= 26) {
if (RsDiff < 14) {
if (Rs < 2302) {
if (Rs < 913) {
return 2;
}
else if (Rs >= 913) {
return 8;
}
}
else if (Rs >= 2302) {
return 1;
}
}
else if (RsDiff >= 14) {
if (Cs < 1391) {
return 1;
}
else if (Cs >= 1391) {
return 2;
}
}
}
}
}
else if (tcor >= 56) {
if (Rs < 1832) {
if (posBalance < 66) {
if (SC < 3147) {
if (Cs < 1347) {
if (Rs < 1316) {
return 2;
}
else if (Rs >= 1316) {
return 1;
}
}
else if (Cs >= 1347) {
return 8;
}
}
else if (SC >= 3147) {
if (Rs < 1793) {
return 1;
}
else if (Rs >= 1793) {
return 4;
}
}
}
else if (posBalance >= 66) {
return 4;
}
}
else if (Rs >= 1832) {
if (Rs < 2975) {
if (tcor < 59) {
if (Rs < 2096) {
return 8;
}
else if (Rs >= 2096) {
return 4;
}
}
else if (tcor >= 59) {
if (Cs < 1365) {
if (Rs < 2639) {
if (Rs < 2415) {
return 4;
}
else if (Rs >= 2415) {
return 1;
}
}
else if (Rs >= 2639) {
return 2;
}
}
else if (Cs >= 1365) {
if (mcTcor < 77) {
if (Rs < 2400) {
return 1;
}
else if (Rs >= 2400) {
return 8;
}
}
else if (mcTcor >= 77) {
if (negBalance < 49) {
if (MVDiff < 25) {
return 2;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (negBalance >= 49) {
return 8;
}
}
}
}
}
else if (Rs >= 2975) {
if (mcTcor < 83) {
return 8;
}
else if (mcTcor >= 83) {
if (RsDiff < 17) {
return 4;
}
else if (RsDiff >= 17) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (negBalance < 50) {
if (Rs < 2786) {
return 1;
}
else if (Rs >= 2786) {
return 8;
}
}
else if (negBalance >= 50) {
if (Cs < 1365) {
if (Rs < 2175) {
return 8;
}
else if (Rs >= 2175) {
return 1;
}
}
else if (Cs >= 1365) {
if (Rs < 2557) {
return 2;
}
else if (Rs >= 2557) {
return 8;
}
}
}
}
}
else if (refDCval >= 121) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
if (refDCval < 151) {
if (mcTcor < 88) {
if (MVDiff < 40) {
if (Rs < 1677) {
return 2;
}
else if (Rs >= 1677) {
if (refDCval < 141) {
return 1;
}
else if (refDCval >= 141) {
return 2;
}
}
}
else if (MVDiff >= 40) {
if (CsDiff < 10) {
return 1;
}
else if (CsDiff >= 10) {
if (AFD < 31) {
return 8;
}
else if (AFD >= 31) {
return 1;
}
}
}
}
else if (mcTcor >= 88) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
if (Cs < 1439) {
if (Cs < 1382) {
if (Cs < 1316) {
return 8;
}
else if (Cs >= 1316) {
return 1;
}
}
else if (Cs >= 1382) {
return 2;
}
}
else if (Cs >= 1439) {
if (diffMVdiffVal < -1) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
}
}
else if (refDCval >= 151) {
if (Rs < 2788) {
return 8;
}
else if (Rs >= 2788) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 3319) {
if (tcor < 86) {
if (RsDiff < 12) {
if (posBalance < 49) {
if (ssDCval < 88) {
if (Rs < 3487) {
if (Cs < 1446) {
return 4;
}
else if (Cs >= 1446) {
return 2;
}
}
else if (Rs >= 3487) {
return 4;
}
}
else if (ssDCval >= 88) {
if (ssDCval < 113) {
return 8;
}
else if (ssDCval >= 113) {
if (Rs < 4203) {
return 4;
}
else if (Rs >= 4203) {
return 1;
}
}
}
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 79) {
if (tcor < 66) {
if (Rs < 3560) {
return 8;
}
else if (Rs >= 3560) {
return 1;
}
}
else if (tcor >= 66) {
return 8;
}
}
else if (mu_mv_mag_sq >= 79) {
if (Rs < 3610) {
return 8;
}
else if (Rs >= 3610) {
return 4;
}
}
}
}
else if (RsDiff >= 12) {
if (MVDiff < 23) {
if (tcor < 83) {
if (RsDiff < 16) {
if (refDCval < 138) {
if (refDCval < 96) {
if (negBalance < 55) {
if (diffRsCsdiff < 6) {
if (Rs < 3680) {
return 2;
}
else if (Rs >= 3680) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 4;
}
}
}
else if (negBalance >= 55) {
return 8;
}
}
else if (refDCval >= 96) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
if (Rs < 4842) {
return 4;
}
else if (Rs >= 4842) {
return 1;
}
}
}
}
else if (refDCval >= 138) {
return 4;
}
}
else if (RsDiff >= 16) {
if (mu_mv_mag_sq < 13) {
if (ssDCval < 101) {
if (Rs < 3444) {
if (Rs < 3365) {
return 2;
}
else if (Rs >= 3365) {
return 8;
}
}
else if (Rs >= 3444) {
return 2;
}
}
else if (ssDCval >= 101) {
if (gchDC < 2) {
if (Cs < 1337) {
return 2;
}
else if (Cs >= 1337) {
return 8;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (Cs < 1616) {
if (RsCsDiff < 383) {
if (diffMVdiffVal < -2) {
if (Cs < 1487) {
return 4;
}
else if (Cs >= 1487) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
else if (RsCsDiff >= 383) {
return 2;
}
}
else if (Cs >= 1616) {
if (RsCsDiff < 462) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 8;
}
}
else if (RsCsDiff >= 462) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
if (Rs < 4307) {
return 2;
}
else if (Rs >= 4307) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 83) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (Cs < 1379) {
if (Cs < 1378) {
return 2;
}
else if (Cs >= 1378) {
return 4;
}
}
else if (Cs >= 1379) {
return 2;
}
}
}
}
else if (MVDiff >= 23) {
if (mu_mv_mag_sq < 10) {
if (CsDiff < 8) {
if (ssDCval < 115) {
return 2;
}
else if (ssDCval >= 115) {
if (TSC < 16) {
if (RsDiff < 14) {
if (Rs < 4755) {
return 8;
}
else if (Rs >= 4755) {
return 1;
}
}
else if (RsDiff >= 14) {
return 8;
}
}
else if (TSC >= 16) {
return 4;
}
}
}
else if (CsDiff >= 8) {
if (posBalance < 55) {
if (TSC < 13) {
if (MVDiff < 32) {
return 1;
}
else if (MVDiff >= 32) {
return 8;
}
}
else if (TSC >= 13) {
return 1;
}
}
else if (posBalance >= 55) {
if (ssDCval < 118) {
return 2;
}
else if (ssDCval >= 118) {
if (CsDiff < 10) {
if (Cs < 1710) {
if (RsDiff < 20) {
if (Rs < 3660) {
return 4;
}
else if (Rs >= 3660) {
if (Rs < 4516) {
return 1;
}
else if (Rs >= 4516) {
return 8;
}
}
}
else if (RsDiff >= 20) {
return 2;
}
}
else if (Cs >= 1710) {
return 8;
}
}
else if (CsDiff >= 10) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (CsDiff < 8) {
if (MVDiff < 26) {
if (Rs < 4932) {
if (Rs < 4471) {
return 8;
}
else if (Rs >= 4471) {
return 2;
}
}
else if (Rs >= 4932) {
return 1;
}
}
else if (MVDiff >= 26) {
if (tcor < 41) {
return 2;
}
else if (tcor >= 41) {
if (mu_mv_mag_sq < 93) {
if (diffRsCsdiff < -174) {
if (Rs < 4104) {
return 2;
}
else if (Rs >= 4104) {
return 4;
}
}
else if (diffRsCsdiff >= -174) {
return 4;
}
}
else if (mu_mv_mag_sq >= 93) {
return 1;
}
}
}
}
else if (CsDiff >= 8) {
if (mu_mv_mag_sq < 26) {
if (posBalance < 59) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (ssDCval < 128) {
return 1;
}
else if (ssDCval >= 128) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
else if (posBalance >= 59) {
if (RsDiff < 21) {
if (Rs < 3560) {
return 4;
}
else if (Rs >= 3560) {
return 1;
}
}
else if (RsDiff >= 21) {
if (Rs < 3520) {
return 8;
}
else if (Rs >= 3520) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (SC < 5667) {
if (TSC < 15) {
if (diffRsCsdiff < 4) {
return 8;
}
else if (diffRsCsdiff >= 4) {
if (MVDiff < 37) {
return 2;
}
else if (MVDiff >= 37) {
return 8;
}
}
}
else if (TSC >= 15) {
if (MVDiff < 69) {
return 2;
}
else if (MVDiff >= 69) {
return 1;
}
}
}
else if (SC >= 5667) {
if (diffRsCsdiff < 33) {
return 4;
}
else if (diffRsCsdiff >= 33) {
if (gchDC < 2) {
if (Rs < 5286) {
return 8;
}
else if (Rs >= 5286) {
return 4;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
}
}
}
}
}
else if (tcor >= 86) {
if (RsDiff < 9) {
if (AFD < 12) {
if (Rs < 3599) {
if (Rs < 3513) {
return 8;
}
else if (Rs >= 3513) {
return 2;
}
}
else if (Rs >= 3599) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 1;
}
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (RsDiff >= 9) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (MVDiff < 17) {
if (refDCval < 100) {
if (posBalance < 44) {
if (refDCval < 83) {
if (AFD < 11) {
if (MVDiff < 6) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 2;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (AFD >= 11) {
return 2;
}
}
else if (refDCval >= 83) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 2;
}
}
}
else if (posBalance >= 44) {
if (refDCval < 73) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (Rs < 4871) {
if (Rs < 3561) {
return 8;
}
else if (Rs >= 3561) {
return 2;
}
}
else if (Rs >= 4871) {
return 4;
}
}
}
else if (refDCval >= 73) {
return 2;
}
}
}
else if (refDCval >= 100) {
if (RsCsDiff < 183) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 84) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
if (Rs < 3987) {
return 2;
}
else if (Rs >= 3987) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (Rs < 3652) {
return 4;
}
else if (Rs >= 3652) {
return 1;
}
}
}
else if (mcTcor >= 95) {
return 2;
}
}
else if (RsCsDiff >= 183) {
if (Rs < 3601) {
return 8;
}
else if (Rs >= 3601) {
if (negBalance < 64) {
return 2;
}
else if (negBalance >= 64) {
if (SC < 6234) {
return 1;
}
else if (SC >= 6234) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 17) {
if (ssDCval < 110) {
if (mcTcor < 93) {
if (mu_mv_mag_sq < 26) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 1;
}
}
else if (mu_mv_mag_sq >= 26) {
return 8;
}
}
else if (mcTcor >= 93) {
if (Cs < 1413) {
return 1;
}
else if (Cs >= 1413) {
return 2;
}
}
}
else if (ssDCval >= 110) {
if (refDCval < 143) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (Rs < 3835) {
return 8;
}
else if (Rs >= 3835) {
return 1;
}
}
}
else if (refDCval >= 143) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 1937) {
if (AFD < 12) {
if (refDCval < 100) {
if (Rs < 4005) {
if (negBalance < 41) {
if (refDCval < 86) {
if (TSC < 8) {
if (MVDiff < 6) {
if (Cs < 2057) {
return 4;
}
else if (Cs >= 2057) {
return 1;
}
}
else if (MVDiff >= 6) {
if (Rs < 1854) {
return 1;
}
else if (Rs >= 1854) {
if (Cs < 1971) {
return 1;
}
else if (Cs >= 1971) {
return 4;
}
}
}
}
else if (TSC >= 8) {
if (posBalance < 60) {
if (RsCsDiff < 115) {
return 8;
}
else if (RsCsDiff >= 115) {
return 4;
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
else if (refDCval >= 86) {
if (SC < 6134) {
return 8;
}
else if (SC >= 6134) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (Cs < 2659) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
if (Rs < 2887) {
return 8;
}
else if (Rs >= 2887) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (Rs < 3250) {
return 4;
}
else if (Rs >= 3250) {
if (Cs < 2420) {
return 8;
}
else if (Cs >= 2420) {
return 4;
}
}
}
}
}
}
else if (Cs >= 2659) {
if (refDCval < 83) {
if (Rs < 2624) {
if (Rs < 1384) {
return 8;
}
else if (Rs >= 1384) {
return 4;
}
}
else if (Rs >= 2624) {
if (tcor < 86) {
return 4;
}
else if (tcor >= 86) {
if (negBalance < 43) {
if (Rs < 3085) {
return 8;
}
else if (Rs >= 3085) {
if (Rs < 3524) {
return 1;
}
else if (Rs >= 3524) {
return 4;
}
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
else if (refDCval >= 83) {
if (Cs < 4036) {
if (Rs < 2782) {
if (Rs < 2531) {
return 4;
}
else if (Rs >= 2531) {
return 8;
}
}
else if (Rs >= 2782) {
return 4;
}
}
else if (Cs >= 4036) {
return 8;
}
}
}
}
}
else if (Rs >= 4005) {
if (MVDiff < 8) {
if (Cs < 2418) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
else if (Cs >= 2418) {
if (Rs < 5871) {
if (diffTSC < -2) {
if (Cs < 2886) {
return 2;
}
else if (Cs >= 2886) {
return 8;
}
}
else if (diffTSC >= -2) {
if (Cs < 3594) {
if (negBalance < 52) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (Cs >= 3594) {
return 8;
}
}
}
else if (Rs >= 5871) {
return 2;
}
}
}
else if (MVDiff >= 8) {
if (Rs < 5504) {
if (Cs < 2056) {
if (ssDCval < 89) {
if (Cs < 2022) {
return 1;
}
else if (Cs >= 2022) {
return 4;
}
}
else if (ssDCval >= 89) {
if (Rs < 4503) {
return 8;
}
else if (Rs >= 4503) {
return 2;
}
}
}
else if (Cs >= 2056) {
if (negBalance < 29) {
if (Rs < 4738) {
return 8;
}
else if (Rs >= 4738) {
return 2;
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
else if (Rs >= 5504) {
if (negBalance < 26) {
if (Rs < 6365) {
return 4;
}
else if (Rs >= 6365) {
return 8;
}
}
else if (negBalance >= 26) {
if (RsDiff < 11) {
if (Rs < 5660) {
return 8;
}
else if (Rs >= 5660) {
return 1;
}
}
else if (RsDiff >= 11) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 100) {
if (negBalance < 46) {
if (Rs < 5749) {
if (negBalance < 34) {
if (ssDCval < 114) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (Cs < 1995) {
return 8;
}
else if (Cs >= 1995) {
return 2;
}
}
}
else if (ssDCval >= 114) {
if (mu_mv_mag_sq < 62) {
if (tcor < 78) {
if (Cs < 3256) {
return 2;
}
else if (Cs >= 3256) {
if (Rs < 3476) {
return 1;
}
else if (Rs >= 3476) {
return 8;
}
}
}
else if (tcor >= 78) {
return 4;
}
}
else if (mu_mv_mag_sq >= 62) {
if (refDCval < 196) {
return 8;
}
else if (refDCval >= 196) {
return 1;
}
}
}
}
else if (negBalance >= 34) {
if (ssDCval < 107) {
if (MVDiff < 4) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (MVDiff >= 4) {
if (diffTSC < -3) {
return 2;
}
else if (diffTSC >= -3) {
return 8;
}
}
}
else if (ssDCval >= 107) {
if (MVDiff < 48) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 105) {
return 8;
}
else if (mu_mv_mag_sq >= 105) {
if (RsCsDiff < 137) {
return 1;
}
else if (RsCsDiff >= 137) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (SC < 6794) {
return 8;
}
else if (SC >= 6794) {
return 4;
}
}
}
else if (MVDiff >= 48) {
return 1;
}
}
}
}
else if (Rs >= 5749) {
if (Cs < 3747) {
if (Cs < 2626) {
if (Rs < 5972) {
return 1;
}
else if (Rs >= 5972) {
return 2;
}
}
else if (Cs >= 2626) {
if (mu_mv_mag_sq < 77) {
return 1;
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 6068) {
return 8;
}
else if (Rs >= 6068) {
return 4;
}
}
}
}
else if (Cs >= 3747) {
if (ssDCval < 141) {
if (Cs < 5449) {
if (Rs < 8171) {
return 4;
}
else if (Rs >= 8171) {
return 1;
}
}
else if (Cs >= 5449) {
return 8;
}
}
else if (ssDCval >= 141) {
return 8;
}
}
}
}
else if (negBalance >= 46) {
if (tcor < 93) {
if (mu_mv_mag_sq < 84) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (diffAFD < 1) {
if (Rs < 3101) {
return 1;
}
else if (Rs >= 3101) {
return 8;
}
}
else if (diffAFD >= 1) {
if (Rs < 3951) {
return 2;
}
else if (Rs >= 3951) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (Rs < 3534) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (Rs >= 3534) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (mu_mv_mag_sq < 92) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
}
}
}
else if (tcor >= 93) {
if (refDCval < 103) {
return 4;
}
else if (refDCval >= 103) {
return 8;
}
}
}
}
}
else if (AFD >= 12) {
if (AFD < 20) {
if (negBalance < 47) {
if (refDCval < 71) {
if (Rs < 3792) {
if (TSC < 8) {
if (SC < 5282) {
if (Rs < 2958) {
return 2;
}
else if (Rs >= 2958) {
return 1;
}
}
else if (SC >= 5282) {
return 1;
}
}
else if (TSC >= 8) {
if (mu_mv_mag_sq < 56) {
if (MVDiff < 26) {
if (RsDiff < 17) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 2;
}
}
else if (RsDiff >= 17) {
return 4;
}
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (mu_mv_mag_sq >= 56) {
if (CsDiff < 10) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
else if (CsDiff >= 10) {
if (Cs < 3577) {
return 8;
}
else if (Cs >= 3577) {
return 1;
}
}
}
}
}
else if (Rs >= 3792) {
if (SC < 9673) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
return 8;
}
}
}
else if (SC >= 9673) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
return 4;
}
}
}
}
else if (refDCval >= 71) {
if (Rs < 5480) {
if (refDCval < 100) {
if (RsCsDiff < 167) {
if (Rs < 1827) {
if (AFD < 17) {
if (Rs < 1032) {
return 2;
}
else if (Rs >= 1032) {
return 8;
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (Rs >= 1827) {
if (mcTcor < 93) {
if (tcor < 78) {
if (Rs < 2694) {
return 1;
}
else if (Rs >= 2694) {
return 4;
}
}
else if (tcor >= 78) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
if (Rs < 2939) {
return 4;
}
else if (Rs >= 2939) {
return 8;
}
}
}
}
else if (mcTcor >= 93) {
if (TSC < 10) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
return 8;
}
}
}
else if (TSC >= 10) {
return 1;
}
}
}
}
else if (RsCsDiff >= 167) {
if (Cs < 2570) {
if (Rs < 2960) {
if (Rs < 1854) {
if (Cs < 2118) {
return 2;
}
else if (Cs >= 2118) {
return 8;
}
}
else if (Rs >= 1854) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
return 8;
}
}
}
else if (Rs >= 2960) {
if (gchDC < 3) {
if (mcTcor < 89) {
if (Rs < 3482) {
return 2;
}
else if (Rs >= 3482) {
return 8;
}
}
else if (mcTcor >= 89) {
if (refDCval < 80) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
return 2;
}
}
else if (refDCval >= 80) {
if (negBalance < 37) {
if (Rs < 4672) {
return 1;
}
else if (Rs >= 4672) {
return 2;
}
}
else if (negBalance >= 37) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (Cs >= 2570) {
if (negBalance < 41) {
if (posBalance < 48) {
if (diffRsCsdiff < 5) {
if (RsCsDiff < 639) {
return 4;
}
else if (RsCsDiff >= 639) {
return 8;
}
}
else if (diffRsCsdiff >= 5) {
return 8;
}
}
else if (posBalance >= 48) {
if (SC < 8654) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (Rs < 4820) {
return 8;
}
else if (Rs >= 4820) {
return 1;
}
}
}
else if (SC >= 8654) {
return 8;
}
}
}
else if (negBalance >= 41) {
if (posBalance < 57) {
if (Cs < 3593) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (Rs < 5378) {
return 2;
}
else if (Rs >= 5378) {
return 8;
}
}
}
else if (Cs >= 3593) {
return 8;
}
}
else if (posBalance >= 57) {
if (Rs < 4704) {
if (CsDiff < 10) {
if (Rs < 3196) {
return 4;
}
else if (Rs >= 3196) {
return 2;
}
}
else if (CsDiff >= 10) {
return 4;
}
}
else if (Rs >= 4704) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 100) {
if (Rs < 2211) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
if (Cs < 2104) {
return 2;
}
else if (Cs >= 2104) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (negBalance < 45) {
if (Rs < 1917) {
return 1;
}
else if (Rs >= 1917) {
return 4;
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
}
else if (Rs >= 2211) {
if (refDCval < 124) {
if (mcTcor < 93) {
if (posBalance < 63) {
if (diffAFD < -10) {
return 4;
}
else if (diffAFD >= -10) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 4;
}
}
}
else if (posBalance >= 63) {
if (ssDCval < 107) {
return 2;
}
else if (ssDCval >= 107) {
if (Cs < 2927) {
return 8;
}
else if (Cs >= 2927) {
return 1;
}
}
}
}
else if (mcTcor >= 93) {
if (Rs < 5297) {
if (negBalance < 42) {
if (mu_mv_mag_sq < 50) {
return 2;
}
else if (mu_mv_mag_sq >= 50) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
return 8;
}
}
}
else if (negBalance >= 42) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
else if (Rs >= 5297) {
return 4;
}
}
}
else if (refDCval >= 124) {
if (mcTcor < 83) {
if (mcTcor < 69) {
return 2;
}
else if (mcTcor >= 69) {
if (tcor < 68) {
if (RsDiff < 14) {
if (Rs < 2598) {
return 1;
}
else if (Rs >= 2598) {
if (diffMVdiffVal < -24) {
return 1;
}
else if (diffMVdiffVal >= -24) {
return 8;
}
}
}
else if (RsDiff >= 14) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 4039) {
return 8;
}
else if (Rs >= 4039) {
return 1;
}
}
}
}
else if (tcor >= 68) {
return 1;
}
}
}
else if (mcTcor >= 83) {
if (ssDCval < 135) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (diffMVdiffVal < 3) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
}
else if (ssDCval >= 135) {
if (gchDC < 12) {
return 8;
}
else if (gchDC >= 12) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 5480) {
if (MVDiff < 6) {
if (RsDiff < 17) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (Rs < 6037) {
return 4;
}
else if (Rs >= 6037) {
return 8;
}
}
}
else if (RsDiff >= 17) {
if (Cs < 4481) {
return 2;
}
else if (Cs >= 4481) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (TSC < 13) {
if (Rs < 8341) {
if (Cs < 2830) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
else if (Cs >= 2830) {
if (diffTSC < -1) {
if (SC < 10012) {
return 4;
}
else if (SC >= 10012) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (Rs >= 8341) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
return 8;
}
}
}
else if (TSC >= 13) {
if (RsDiff < 22) {
return 1;
}
else if (RsDiff >= 22) {
if (Rs < 7864) {
return 2;
}
else if (Rs >= 7864) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 47) {
if (mcTcor < 76) {
if (ssDCval < 66) {
if (negBalance < 56) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (negBalance >= 56) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
}
else if (ssDCval >= 66) {
if (Rs < 3065) {
if (TSC < 14) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 1834) {
return 2;
}
else if (Rs >= 1834) {
return 4;
}
}
}
else if (TSC >= 14) {
return 4;
}
}
else if (Rs >= 3065) {
return 8;
}
}
}
else if (mcTcor >= 76) {
if (TSC < 13) {
if (MVDiff < 4) {
if (refDCval < 87) {
if (mu_mv_mag_sq < 76) {
if (Rs < 3315) {
if (AFD < 16) {
if (SC < 5863) {
if (refDCval < 80) {
if (Rs < 2389) {
return 8;
}
else if (Rs >= 2389) {
return 2;
}
}
else if (refDCval >= 80) {
return 4;
}
}
else if (SC >= 5863) {
return 1;
}
}
else if (AFD >= 16) {
if (Cs < 3818) {
return 1;
}
else if (Cs >= 3818) {
return 8;
}
}
}
else if (Rs >= 3315) {
if (Cs < 3340) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
return 4;
}
}
else if (Cs >= 3340) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (Rs < 2834) {
return 4;
}
else if (Rs >= 2834) {
return 8;
}
}
}
else if (refDCval >= 87) {
if (Cs < 2764) {
if (RsDiff < 11) {
if (diffMVdiffVal < -1) {
if (Cs < 2592) {
return 8;
}
else if (Cs >= 2592) {
if (Rs < 2713) {
return 8;
}
else if (Rs >= 2713) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (mcTcor < 93) {
if (ssDCval < 99) {
return 8;
}
else if (ssDCval >= 99) {
if (ssDCval < 128) {
if (Rs < 4501) {
return 1;
}
else if (Rs >= 4501) {
return 8;
}
}
else if (ssDCval >= 128) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (diffTSC < -1) {
if (Rs < 3803) {
return 4;
}
else if (Rs >= 3803) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
else if (RsDiff >= 11) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 4;
}
}
}
}
}
else if (Cs >= 2764) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (tcor < 81) {
if (Cs < 2876) {
return 4;
}
else if (Cs >= 2876) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (tcor >= 81) {
if (SC < 7561) {
return 4;
}
else if (SC >= 7561) {
if (tcor < 90) {
if (RsCsDiff < 194) {
return 8;
}
else if (RsCsDiff >= 194) {
if (Rs < 7128) {
return 4;
}
else if (Rs >= 7128) {
return 8;
}
}
}
else if (tcor >= 90) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (RsDiff < 15) {
if (Rs < 4357) {
if (SC < 3599) {
if (Cs < 1949) {
return 2;
}
else if (Cs >= 1949) {
return 4;
}
}
else if (SC >= 3599) {
if (Rs < 3015) {
if (mcTcor < 94) {
if (tcor < 64) {
if (ssDCval < 92) {
if (Rs < 2295) {
return 4;
}
else if (Rs >= 2295) {
if (Cs < 3011) {
return 8;
}
else if (Cs >= 3011) {
return 4;
}
}
}
else if (ssDCval >= 92) {
return 8;
}
}
else if (tcor >= 64) {
if (MVDiff < 13) {
if (tcor < 74) {
if (Cs < 3416) {
if (Cs < 2927) {
return 8;
}
else if (Cs >= 2927) {
if (Cs < 2952) {
return 1;
}
else if (Cs >= 2952) {
return 8;
}
}
}
else if (Cs >= 3416) {
return 4;
}
}
else if (tcor >= 74) {
if (refDCval < 86) {
if (SC < 4108) {
return 4;
}
else if (SC >= 4108) {
if (posBalance < 56) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (posBalance >= 56) {
if (diffRsCsdiff < 46) {
if (RsDiff < 8) {
if (Cs < 2763) {
return 8;
}
else if (Cs >= 2763) {
return 4;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (diffRsCsdiff >= 46) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
}
else if (refDCval >= 86) {
if (ssDCval < 90) {
if (mu_mv_mag_sq < 20) {
return 4;
}
else if (mu_mv_mag_sq >= 20) {
if (Cs < 2799) {
return 8;
}
else if (Cs >= 2799) {
return 4;
}
}
}
else if (ssDCval >= 90) {
if (SC < 3996) {
return 4;
}
else if (SC >= 3996) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (AFD < 16) {
if (SC < 4936) {
return 8;
}
else if (SC >= 4936) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
}
else if (AFD >= 16) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 13) {
if (CsDiff < 11) {
if (Cs < 1984) {
if (Cs < 1978) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
else if (Cs >= 1978) {
if (Rs < 2392) {
return 1;
}
else if (Rs >= 2392) {
return 4;
}
}
}
else if (Cs >= 1984) {
if (Rs < 2779) {
return 8;
}
else if (Rs >= 2779) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
if (Rs < 2781) {
return 2;
}
else if (Rs >= 2781) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
}
}
}
}
else if (CsDiff >= 11) {
if (ssDCval < 78) {
if (Cs < 2332) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 2;
}
}
else if (Cs >= 2332) {
if (Rs < 1693) {
if (Rs < 1301) {
return 1;
}
else if (Rs >= 1301) {
return 4;
}
}
else if (Rs >= 1693) {
return 8;
}
}
}
else if (ssDCval >= 78) {
if (Cs < 4786) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2480) {
return 8;
}
else if (Rs >= 2480) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (Cs >= 4786) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (ssDCval < 104) {
if (AFD < 16) {
if (Rs < 2755) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 2;
}
}
else if (Rs >= 2755) {
return 2;
}
}
else if (AFD >= 16) {
if (TSC < 10) {
if (negBalance < 53) {
return 8;
}
else if (negBalance >= 53) {
if (Rs < 2387) {
return 2;
}
else if (Rs >= 2387) {
return 4;
}
}
}
else if (TSC >= 10) {
return 2;
}
}
}
else if (ssDCval >= 104) {
if (Rs < 1503) {
return 1;
}
else if (Rs >= 1503) {
if (Cs < 2777) {
return 8;
}
else if (Cs >= 2777) {
return 4;
}
}
}
}
}
else if (Rs >= 3015) {
if (RsCsDiff < 160) {
if (Cs < 2703) {
if (ssDCval < 105) {
return 8;
}
else if (ssDCval >= 105) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Rs < 3126) {
return 1;
}
else if (Rs >= 3126) {
return 4;
}
}
}
}
else if (Cs >= 2703) {
if (mcTcor < 92) {
if (Rs < 3966) {
if (Cs < 2829) {
return 4;
}
else if (Cs >= 2829) {
return 8;
}
}
else if (Rs >= 3966) {
return 1;
}
}
else if (mcTcor >= 92) {
return 4;
}
}
}
else if (RsCsDiff >= 160) {
if (SC < 6444) {
if (RsDiff < 12) {
if (mu_mv_mag_sq < 57) {
if (Rs < 3153) {
if (MVDiff < 12) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (Rs >= 3153) {
if (AFD < 17) {
if (RsCsDiff < 212) {
return 8;
}
else if (RsCsDiff >= 212) {
if (Rs < 3165) {
return 8;
}
else if (Rs >= 3165) {
return 4;
}
}
}
else if (AFD >= 17) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (CsDiff < 9) {
if (Rs < 3224) {
return 8;
}
else if (Rs >= 3224) {
return 2;
}
}
else if (CsDiff >= 9) {
return 4;
}
}
}
else if (RsDiff >= 12) {
if (negBalance < 50) {
if (Cs < 2474) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
else if (Cs >= 2474) {
if (Cs < 2589) {
return 2;
}
else if (Cs >= 2589) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (posBalance < 50) {
if (Rs < 3311) {
return 4;
}
else if (Rs >= 3311) {
return 8;
}
}
else if (posBalance >= 50) {
if (Rs < 3377) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (diffAFD < 2) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (diffAFD >= 2) {
return 2;
}
}
}
else if (Rs >= 3377) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (Cs < 2016) {
return 2;
}
else if (Cs >= 2016) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 6444) {
if (tcor < 84) {
if (refDCval < 136) {
if (Rs < 4024) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (diffRsCsdiff < -45) {
if (SC < 6501) {
return 1;
}
else if (SC >= 6501) {
return 4;
}
}
else if (diffRsCsdiff >= -45) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (MVDiff < 9) {
if (Cs < 3272) {
return 1;
}
else if (Cs >= 3272) {
return 8;
}
}
else if (MVDiff >= 9) {
return 1;
}
}
}
}
}
else if (Rs >= 4024) {
return 8;
}
}
else if (refDCval >= 136) {
return 1;
}
}
else if (tcor >= 84) {
if (ssDCval < 89) {
if (Rs < 3945) {
if (Rs < 3772) {
return 1;
}
else if (Rs >= 3772) {
return 2;
}
}
else if (Rs >= 3945) {
return 8;
}
}
else if (ssDCval >= 89) {
if (Cs < 2553) {
if (Rs < 4312) {
return 1;
}
else if (Rs >= 4312) {
return 4;
}
}
else if (Cs >= 2553) {
if (RsCsDiff < 172) {
if (Rs < 4135) {
return 4;
}
else if (Rs >= 4135) {
return 8;
}
}
else if (RsCsDiff >= 172) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Rs >= 4357) {
if (diffAFD < -6) {
if (Rs < 4679) {
return 2;
}
else if (Rs >= 4679) {
if (Rs < 5627) {
return 4;
}
else if (Rs >= 5627) {
return 8;
}
}
}
else if (diffAFD >= -6) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (mcTcor < 94) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
if (Rs < 4510) {
return 4;
}
else if (Rs >= 4510) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (negBalance < 57) {
if (Cs < 2071) {
return 1;
}
else if (Cs >= 2071) {
return 4;
}
}
else if (negBalance >= 57) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 15) {
if (tcor < 80) {
if (CsDiff < 13) {
if (mu_mv_mag_sq < 48) {
return 4;
}
else if (mu_mv_mag_sq >= 48) {
return 8;
}
}
else if (CsDiff >= 13) {
if (ssDCval < 73) {
if (Rs < 4152) {
return 4;
}
else if (Rs >= 4152) {
return 8;
}
}
else if (ssDCval >= 73) {
if (RsDiff < 20) {
if (diffAFD < 2) {
if (refDCval < 141) {
return 8;
}
else if (refDCval >= 141) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (RsDiff >= 20) {
return 4;
}
}
}
}
else if (tcor >= 80) {
if (negBalance < 49) {
if (Rs < 7403) {
return 2;
}
else if (Rs >= 7403) {
return 1;
}
}
else if (negBalance >= 49) {
if (gchDC < 1) {
if (RsCsDiff < 351) {
return 4;
}
else if (RsCsDiff >= 351) {
return 2;
}
}
else if (gchDC >= 1) {
if (posBalance < 60) {
if (diffRsCsdiff < 175) {
if (Cs < 2814) {
return 8;
}
else if (Cs >= 2814) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 8;
}
}
}
else if (diffRsCsdiff >= 175) {
return 2;
}
}
else if (posBalance >= 60) {
return 2;
}
}
}
}
}
}
}
else if (TSC >= 13) {
if (Rs < 5805) {
if (Cs < 3337) {
if (ssDCval < 84) {
if (diffAFD < -3) {
if (Rs < 5045) {
return 4;
}
else if (Rs >= 5045) {
return 2;
}
}
else if (diffAFD >= -3) {
if (tcor < 65) {
return 4;
}
else if (tcor >= 65) {
if (diffRsCsdiff < -49) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
return 2;
}
}
else if (diffRsCsdiff >= -49) {
if (Rs < 3498) {
return 8;
}
else if (Rs >= 3498) {
if (mu_mv_mag_sq < 15) {
return 1;
}
else if (mu_mv_mag_sq >= 15) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (Cs < 2716) {
if (ssDCval < 88) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (ssDCval >= 88) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
if (Rs < 2715) {
return 1;
}
else if (Rs >= 2715) {
return 8;
}
}
}
}
}
}
else if (Cs >= 2716) {
if (tcor < 85) {
if (Rs < 4640) {
return 4;
}
else if (Rs >= 4640) {
return 8;
}
}
else if (tcor >= 85) {
if (ssDCval < 94) {
if (Rs < 2734) {
return 4;
}
else if (Rs >= 2734) {
return 8;
}
}
else if (ssDCval >= 94) {
return 4;
}
}
}
}
}
else if (Cs >= 3337) {
if (ssDCval < 113) {
if (diffAFD < -2) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 4318) {
return 4;
}
else if (Rs >= 4318) {
return 8;
}
}
}
else if (diffAFD >= -2) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
if (Rs < 4916) {
return 4;
}
else if (Rs >= 4916) {
return 8;
}
}
}
}
else if (ssDCval >= 113) {
if (Rs < 3936) {
return 8;
}
else if (Rs >= 3936) {
return 4;
}
}
}
}
else if (Rs >= 5805) {
if (Rs < 6265) {
return 2;
}
else if (Rs >= 6265) {
return 4;
}
}
}
}
}
}
else if (AFD >= 20) {
if (mu_mv_mag_sq < 29) {
if (ssDCval < 152) {
if (ssDCval < 111) {
if (Rs < 5746) {
if (negBalance < 54) {
if (ssDCval < 72) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (RsCsDiff < 547) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
return 4;
}
}
else if (RsCsDiff >= 547) {
if (Rs < 2842) {
return 4;
}
else if (Rs >= 2842) {
if (Cs < 3508) {
return 2;
}
else if (Cs >= 3508) {
return 4;
}
}
}
}
}
else if (ssDCval >= 72) {
if (refDCval < 99) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
if (Cs < 2265) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (mu_mv_mag_sq < 23) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
if (Rs < 2864) {
return 8;
}
else if (Rs >= 2864) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 23) {
return 8;
}
}
}
}
else if (Cs >= 2265) {
if (Rs < 1531) {
if (CsDiff < 15) {
return 1;
}
else if (CsDiff >= 15) {
return 2;
}
}
else if (Rs >= 1531) {
if (mcTcor < 87) {
if (negBalance < 53) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (AFD < 28) {
return 8;
}
else if (AFD >= 28) {
if (Cs < 2900) {
return 1;
}
else if (Cs >= 2900) {
return 8;
}
}
}
}
else if (negBalance >= 53) {
if (MVDiff < 13) {
if (Rs < 3268) {
return 1;
}
else if (Rs >= 3268) {
return 4;
}
}
else if (MVDiff >= 13) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 2;
}
}
}
}
else if (mcTcor >= 87) {
return 8;
}
}
}
}
}
else if (refDCval >= 99) {
if (MVDiff < 30) {
if (RsDiff < 11) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
if (Rs < 1541) {
if (Rs < 1089) {
return 2;
}
else if (Rs >= 1089) {
return 1;
}
}
else if (Rs >= 1541) {
return 8;
}
}
}
else if (RsDiff >= 11) {
if (Rs < 3683) {
return 2;
}
else if (Rs >= 3683) {
if (Cs < 2279) {
return 8;
}
else if (Cs >= 2279) {
if (ssDCval < 101) {
return 2;
}
else if (ssDCval >= 101) {
return 4;
}
}
}
}
}
else if (MVDiff >= 30) {
return 8;
}
}
}
}
else if (negBalance >= 54) {
if (RsDiff < 20) {
if (RsDiff < 10) {
if (tcor < 75) {
if (Rs < 1320) {
if (Cs < 2312) {
return 4;
}
else if (Cs >= 2312) {
if (Cs < 4744) {
return 2;
}
else if (Cs >= 4744) {
return 8;
}
}
}
else if (Rs >= 1320) {
if (Rs < 2503) {
if (Cs < 2353) {
if (Rs < 1651) {
if (Rs < 1387) {
return 1;
}
else if (Rs >= 1387) {
return 4;
}
}
else if (Rs >= 1651) {
return 8;
}
}
else if (Cs >= 2353) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (diffRsCsdiff < 36) {
if (refDCval < 101) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (refDCval >= 101) {
return 8;
}
}
else if (diffRsCsdiff >= 36) {
if (Cs < 4563) {
return 8;
}
else if (Cs >= 4563) {
return 4;
}
}
}
}
}
else if (Rs >= 2503) {
return 8;
}
}
}
else if (tcor >= 75) {
if (tcor < 78) {
if (Cs < 2079) {
return 1;
}
else if (Cs >= 2079) {
return 8;
}
}
else if (tcor >= 78) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (Cs < 4082) {
return 1;
}
else if (Cs >= 4082) {
return 4;
}
}
}
}
}
else if (RsDiff >= 10) {
if (MVDiff < 10) {
if (SC < 5512) {
if (ssDCval < 97) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
return 1;
}
}
else if (ssDCval >= 97) {
return 2;
}
}
else if (SC >= 5512) {
if (Rs < 4577) {
if (negBalance < 56) {
if (Rs < 3517) {
return 8;
}
else if (Rs >= 3517) {
if (Cs < 3770) {
return 4;
}
else if (Cs >= 3770) {
return 8;
}
}
}
else if (negBalance >= 56) {
return 4;
}
}
else if (Rs >= 4577) {
return 8;
}
}
}
else if (MVDiff >= 10) {
if (negBalance < 59) {
if (diffRsCsdiff < -103) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 4;
}
}
else if (diffRsCsdiff >= -103) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (RsCsDiff < 315) {
if (tcor < 78) {
if (ssDCval < 88) {
if (tcor < 57) {
if (Cs < 2232) {
return 2;
}
else if (Cs >= 2232) {
return 4;
}
}
else if (tcor >= 57) {
if (SC < 5554) {
if (Cs < 2233) {
if (Rs < 3220) {
return 8;
}
else if (Rs >= 3220) {
return 4;
}
}
else if (Cs >= 2233) {
if (MVDiff < 21) {
return 1;
}
else if (MVDiff >= 21) {
return 8;
}
}
}
else if (SC >= 5554) {
return 8;
}
}
}
else if (ssDCval >= 88) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
if (Cs < 3505) {
return 1;
}
else if (Cs >= 3505) {
return 8;
}
}
}
}
else if (tcor >= 78) {
return 8;
}
}
else if (RsCsDiff >= 315) {
if (CsDiff < 12) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
return 4;
}
}
else if (CsDiff >= 12) {
if (diffTSC < 2) {
if (posBalance < 68) {
return 8;
}
else if (posBalance >= 68) {
return 1;
}
}
else if (diffTSC >= 2) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 59) {
if (CsDiff < 17) {
if (gchDC < 5) {
if (mcTcor < 93) {
if (mcTcor < 78) {
return 4;
}
else if (mcTcor >= 78) {
if (Cs < 1952) {
return 4;
}
else if (Cs >= 1952) {
if (Cs < 2893) {
if (AFD < 21) {
if (Rs < 2659) {
return 1;
}
else if (Rs >= 2659) {
return 4;
}
}
else if (AFD >= 21) {
return 1;
}
}
else if (Cs >= 2893) {
return 4;
}
}
}
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (gchDC >= 5) {
if (Rs < 2333) {
return 8;
}
else if (Rs >= 2333) {
return 2;
}
}
}
else if (CsDiff >= 17) {
if (RsCsDiff < 484) {
return 2;
}
else if (RsCsDiff >= 484) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 20) {
if (posBalance < 61) {
if (mu_mv_mag_sq < 20) {
if (ssDCval < 104) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 2;
}
}
else if (ssDCval >= 104) {
if (Rs < 4903) {
return 1;
}
else if (Rs >= 4903) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 20) {
if (Cs < 2218) {
return 1;
}
else if (Cs >= 2218) {
if (Rs < 4237) {
return 8;
}
else if (Rs >= 4237) {
return 4;
}
}
}
}
else if (posBalance >= 61) {
return 1;
}
}
}
}
else if (Rs >= 5746) {
if (MVDiff < 19) {
if (RsDiff < 17) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
else if (RsDiff >= 17) {
if (RsDiff < 27) {
return 4;
}
else if (RsDiff >= 27) {
if (Rs < 7341) {
return 8;
}
else if (Rs >= 7341) {
return 1;
}
}
}
}
else if (MVDiff >= 19) {
return 1;
}
}
}
else if (ssDCval >= 111) {
if (RsCsDiff < 945) {
if (RsCsDiff < 201) {
if (ssDCval < 120) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 2;
}
}
else if (ssDCval >= 120) {
if (Cs < 2085) {
return 1;
}
else if (Cs >= 2085) {
return 2;
}
}
}
else if (RsCsDiff >= 201) {
if (negBalance < 57) {
if (Rs < 6016) {
if (mcTcor < 86) {
if (Rs < 3697) {
if (RsCsDiff < 836) {
return 8;
}
else if (RsCsDiff >= 836) {
return 4;
}
}
else if (Rs >= 3697) {
if (Rs < 4107) {
return 4;
}
else if (Rs >= 4107) {
return 8;
}
}
}
else if (mcTcor >= 86) {
if (Cs < 2343) {
if (Rs < 3934) {
if (Rs < 2550) {
if (Rs < 1864) {
return 8;
}
else if (Rs >= 1864) {
return 1;
}
}
else if (Rs >= 2550) {
return 2;
}
}
else if (Rs >= 3934) {
return 1;
}
}
else if (Cs >= 2343) {
if (RsCsDiff < 379) {
return 8;
}
else if (RsCsDiff >= 379) {
if (TSC < 15) {
if (RsCsDiff < 423) {
return 4;
}
else if (RsCsDiff >= 423) {
if (mu_mv_mag_sq < 22) {
return 2;
}
else if (mu_mv_mag_sq >= 22) {
if (Rs < 3585) {
return 4;
}
else if (Rs >= 3585) {
return 2;
}
}
}
}
else if (TSC >= 15) {
return 8;
}
}
}
}
}
else if (Rs >= 6016) {
return 8;
}
}
else if (negBalance >= 57) {
if (TSC < 15) {
if (posBalance < 58) {
if (Cs < 2942) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 8;
}
}
else if (Cs >= 2942) {
if (Rs < 3140) {
return 2;
}
else if (Rs >= 3140) {
return 4;
}
}
}
else if (posBalance >= 58) {
if (negBalance < 61) {
if (tcor < 76) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
return 1;
}
}
else if (tcor >= 76) {
return 8;
}
}
else if (negBalance >= 61) {
return 1;
}
}
}
else if (TSC >= 15) {
if (refDCval < 135) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 2891) {
return 8;
}
else if (Rs >= 2891) {
return 1;
}
}
}
else if (refDCval >= 135) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 945) {
if (mcTcor < 94) {
if (AFD < 34) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
if (Rs < 7055) {
return 2;
}
else if (Rs >= 7055) {
return 8;
}
}
}
else if (AFD >= 34) {
if (Cs < 4529) {
if (Rs < 8028) {
return 2;
}
else if (Rs >= 8028) {
return 4;
}
}
else if (Cs >= 4529) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 5738) {
return 2;
}
else if (Rs >= 5738) {
return 1;
}
}
}
}
}
else if (ssDCval >= 152) {
if (ssDCval < 168) {
if (Rs < 3471) {
return 8;
}
else if (Rs >= 3471) {
if (MVDiff < 38) {
return 4;
}
else if (MVDiff >= 38) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
return 4;
}
}
}
}
else if (ssDCval >= 168) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (Cs < 3729) {
if (mu_mv_mag_sq < 54) {
if (negBalance < 48) {
if (gchDC < 5) {
if (MVDiff < 35) {
if (CsDiff < 14) {
return 2;
}
else if (CsDiff >= 14) {
if (Rs < 3510) {
return 8;
}
else if (Rs >= 3510) {
return 4;
}
}
}
else if (MVDiff >= 35) {
if (Rs < 3671) {
return 8;
}
else if (Rs >= 3671) {
return 4;
}
}
}
else if (gchDC >= 5) {
if (refDCval < 69) {
return 4;
}
else if (refDCval >= 69) {
if (Rs < 2287) {
if (Rs < 1412) {
return 1;
}
else if (Rs >= 1412) {
return 2;
}
}
else if (Rs >= 2287) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
if (Cs < 2718) {
if (refDCval < 97) {
if (AFD < 27) {
if (diffMVdiffVal < 8) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
if (mcTcor < 83) {
return 4;
}
else if (mcTcor >= 83) {
return 8;
}
}
}
else if (diffMVdiffVal >= 8) {
return 4;
}
}
else if (AFD >= 27) {
if (Rs < 2557) {
return 4;
}
else if (Rs >= 2557) {
return 1;
}
}
}
else if (refDCval >= 97) {
if (Rs < 5687) {
if (AFD < 35) {
if (mu_mv_mag_sq < 31) {
if (Rs < 2838) {
return 2;
}
else if (Rs >= 2838) {
return 8;
}
}
else if (mu_mv_mag_sq >= 31) {
return 8;
}
}
else if (AFD >= 35) {
return 1;
}
}
else if (Rs >= 5687) {
return 1;
}
}
}
else if (Cs >= 2718) {
if (ssDCval < 105) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Cs < 3639) {
if (Rs < 1807) {
return 8;
}
else if (Rs >= 1807) {
return 4;
}
}
else if (Cs >= 3639) {
return 8;
}
}
}
else if (ssDCval >= 105) {
if (negBalance < 57) {
return 4;
}
else if (negBalance >= 57) {
if (Cs < 2770) {
return 4;
}
else if (Cs >= 2770) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 4233) {
if (MVDiff < 16) {
if (Cs < 2202) {
return 8;
}
else if (Cs >= 2202) {
return 4;
}
}
else if (MVDiff >= 16) {
if (ssDCval < 88) {
if (mu_mv_mag_sq < 63) {
return 8;
}
else if (mu_mv_mag_sq >= 63) {
return 4;
}
}
else if (ssDCval >= 88) {
if (Cs < 3084) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
if (Cs < 2898) {
return 8;
}
else if (Cs >= 2898) {
return 1;
}
}
}
else if (Cs >= 3084) {
return 8;
}
}
}
}
else if (Rs >= 4233) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 4;
}
}
}
}
else if (Cs >= 3729) {
if (negBalance < 64) {
if (mcTcor < 84) {
if (negBalance < 50) {
if (Cs < 4995) {
if (Cs < 4385) {
return 1;
}
else if (Cs >= 4385) {
return 2;
}
}
else if (Cs >= 4995) {
return 8;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (mcTcor >= 84) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (RsDiff < 18) {
if (Rs < 6259) {
return 8;
}
else if (Rs >= 6259) {
return 1;
}
}
else if (RsDiff >= 18) {
if (AFD < 26) {
return 4;
}
else if (AFD >= 26) {
if (Rs < 5783) {
return 8;
}
else if (Rs >= 5783) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 64) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Cs < 1641) {
if (Rs < 3092) {
if (Rs < 1342) {
if (Rs < 1115) {
if (AFD < 1) {
if (Rs < 599) {
return 4;
}
else if (Rs >= 599) {
return 2;
}
}
else if (AFD >= 1) {
if (ssDCval < 44) {
return 1;
}
else if (ssDCval >= 44) {
if (ssDCval < 116) {
if (ssDCval < 76) {
if (ssDCval < 63) {
if (ssDCval < 53) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (ssDCval >= 53) {
if (SC < 2404) {
return 1;
}
else if (SC >= 2404) {
if (TSC < 2) {
if (Rs < 983) {
return 8;
}
else if (Rs >= 983) {
return 1;
}
}
else if (TSC >= 2) {
if (ssDCval < 57) {
if (Rs < 930) {
return 8;
}
else if (Rs >= 930) {
return 4;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
}
}
else if (ssDCval >= 63) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (Rs < 875) {
return 8;
}
else if (Rs >= 875) {
return 2;
}
}
}
}
else if (ssDCval >= 76) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (ssDCval >= 116) {
if (Rs < 966) {
return 8;
}
else if (Rs >= 966) {
if (ssDCval < 130) {
return 4;
}
else if (ssDCval >= 130) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1115) {
if (mu_mv_mag_sq < 122) {
if (posBalance < 39) {
if (Cs < 1441) {
return 1;
}
else if (Cs >= 1441) {
return 8;
}
}
else if (posBalance >= 39) {
if (Cs < 1471) {
if (MVDiff < 3) {
if (ssDCval < 57) {
if (Rs < 1141) {
return 8;
}
else if (Rs >= 1141) {
return 4;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Rs < 1240) {
return 4;
}
else if (Rs >= 1240) {
return 8;
}
}
}
else if (Cs >= 1471) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (Cs < 1587) {
if (SC < 2657) {
return 8;
}
else if (SC >= 2657) {
return 4;
}
}
else if (Cs >= 1587) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (posBalance < 28) {
if (Cs < 1348) {
return 1;
}
else if (Cs >= 1348) {
return 8;
}
}
else if (posBalance >= 28) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (Rs < 1197) {
return 8;
}
else if (Rs >= 1197) {
if (ssDCval < 94) {
return 4;
}
else if (ssDCval >= 94) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1342) {
if (ssDCval < 86) {
if (mcTcor < 98) {
if (posBalance < 45) {
if (mu_mv_mag_sq < 123) {
if (posBalance < 30) {
if (Cs < 1396) {
return 8;
}
else if (Cs >= 1396) {
if (Rs < 2129) {
return 1;
}
else if (Rs >= 2129) {
return 8;
}
}
}
else if (posBalance >= 30) {
if (Cs < 1510) {
if (Rs < 1601) {
return 8;
}
else if (Rs >= 1601) {
return 4;
}
}
else if (Cs >= 1510) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Cs < 1356) {
if (Cs < 1346) {
if (Rs < 2062) {
return 8;
}
else if (Rs >= 2062) {
return 8;
}
}
else if (Cs >= 1346) {
return 4;
}
}
else if (Cs >= 1356) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 8;
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2791) {
if (tcor < 80) {
return 4;
}
else if (tcor >= 80) {
if (negBalance < 21) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (negBalance >= 21) {
if (Rs < 1418) {
if (Rs < 1410) {
return 8;
}
else if (Rs >= 1410) {
return 4;
}
}
else if (Rs >= 1418) {
return 8;
}
}
}
}
else if (Rs >= 2791) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 19) {
if (Cs < 1455) {
if (Cs < 1350) {
if (Rs < 1453) {
return 1;
}
else if (Rs >= 1453) {
if (Rs < 2278) {
return 8;
}
else if (Rs >= 2278) {
if (mu_mv_mag_sq < 125) {
if (Rs < 2295) {
return 8;
}
else if (Rs >= 2295) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1317) {
return 2;
}
else if (Cs >= 1317) {
if (Rs < 2348) {
if (Rs < 2301) {
if (Cs < 1342) {
if (SC < 3619) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
if (posBalance < 32) {
return 2;
}
else if (posBalance >= 32) {
return 8;
}
}
}
else if (SC >= 3619) {
return 8;
}
}
else if (Cs >= 1342) {
if (Rs < 2294) {
return 8;
}
else if (Rs >= 2294) {
return 2;
}
}
}
else if (Rs >= 2301) {
return 8;
}
}
else if (Rs >= 2348) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1350) {
return 8;
}
}
else if (Cs >= 1455) {
if (ssDCval < 61) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
return 1;
}
}
else if (ssDCval >= 61) {
return 8;
}
}
}
else if (negBalance >= 19) {
if (posBalance < 41) {
if (Rs < 1800) {
if (Rs < 1513) {
if (Cs < 1456) {
return 4;
}
else if (Cs >= 1456) {
return 1;
}
}
else if (Rs >= 1513) {
return 4;
}
}
else if (Rs >= 1800) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (posBalance >= 41) {
if (ssDCval < 80) {
if (Rs < 1818) {
return 4;
}
else if (Rs >= 1818) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
return 4;
}
}
}
else if (ssDCval >= 80) {
if (Rs < 2471) {
if (Cs < 1436) {
return 4;
}
else if (Cs >= 1436) {
return 8;
}
}
else if (Rs >= 2471) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (Rs < 1978) {
if (mu_mv_mag_sq < 119) {
if (negBalance < 5) {
return 1;
}
else if (negBalance >= 5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 119) {
if (CsDiff < 2) {
if (RsDiff < 1) {
if (ssDCval < 126) {
if (diffMVdiffVal < -2) {
if (Rs < 1834) {
return 8;
}
else if (Rs >= 1834) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
else if (ssDCval >= 126) {
return 8;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (CsDiff >= 2) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
if (ssDCval < 120) {
return 1;
}
else if (ssDCval >= 120) {
if (Rs < 1386) {
return 1;
}
else if (Rs >= 1386) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1978) {
if (Cs < 1498) {
if (negBalance < 17) {
return 8;
}
else if (negBalance >= 17) {
if (Cs < 1410) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 110) {
return 2;
}
else if (mu_mv_mag_sq >= 110) {
if (Cs < 1401) {
return 8;
}
else if (Cs >= 1401) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (posBalance >= 54) {
return 8;
}
}
else if (Cs >= 1410) {
if (Rs < 3048) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (SC < 3524) {
return 2;
}
else if (SC >= 3524) {
if (Cs < 1496) {
return 8;
}
else if (Cs >= 1496) {
return 2;
}
}
}
}
else if (Rs >= 3048) {
if (Cs < 1466) {
return 2;
}
else if (Cs >= 1466) {
return 8;
}
}
}
}
}
else if (Cs >= 1498) {
return 8;
}
}
}
else if (ssDCval >= 140) {
if (MVDiff < 3) {
if (Rs < 2008) {
return 2;
}
else if (Rs >= 2008) {
return 1;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
}
}
else if (Rs >= 3092) {
if (ssDCval < 114) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 126) {
if (Rs < 3580) {
if (RsDiff < 5) {
if (Cs < 1529) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
if (Cs < 1358) {
return 4;
}
else if (Cs >= 1358) {
return 8;
}
}
}
else if (Cs >= 1529) {
return 4;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (Rs >= 3580) {
if (AFD < 4) {
if (posBalance < 33) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 33) {
if (RsDiff < 3) {
if (negBalance < 11) {
return 4;
}
else if (negBalance >= 11) {
if (Cs < 1584) {
if (mu_mv_mag_sq < 115) {
return 1;
}
else if (mu_mv_mag_sq >= 115) {
if (Cs < 1324) {
if (Rs < 3674) {
return 8;
}
else if (Rs >= 3674) {
return 4;
}
}
else if (Cs >= 1324) {
return 8;
}
}
}
else if (Cs >= 1584) {
return 4;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 4259) {
return 4;
}
else if (Rs >= 4259) {
if (Rs < 4994) {
return 8;
}
else if (Rs >= 4994) {
return 4;
}
}
}
}
}
else if (AFD >= 4) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1577) {
if (CsDiff < 2) {
if (Rs < 3659) {
if (SC < 4983) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (SC >= 4983) {
return 4;
}
}
else if (Rs >= 3659) {
return 8;
}
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (Cs >= 1577) {
if (Rs < 3265) {
return 8;
}
else if (Rs >= 3265) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (posBalance < 44) {
if (Rs < 4083) {
if (Rs < 3225) {
return 8;
}
else if (Rs >= 3225) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
else if (Rs >= 4083) {
if (ssDCval < 96) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 8;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (posBalance >= 44) {
if (mu_mv_mag_sq < 123) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
else if (mu_mv_mag_sq >= 123) {
if (ssDCval < 83) {
return 2;
}
else if (ssDCval >= 83) {
if (CsDiff < 7) {
if (Rs < 4598) {
return 4;
}
else if (Rs >= 4598) {
return 2;
}
}
else if (CsDiff >= 7) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 114) {
if (diffRsCsdiff < -19) {
if (SC < 4893) {
if (Rs < 3127) {
return 1;
}
else if (Rs >= 3127) {
return 2;
}
}
else if (SC >= 4893) {
if (Cs < 1551) {
return 4;
}
else if (Cs >= 1551) {
return 1;
}
}
}
else if (diffRsCsdiff >= -19) {
return 2;
}
}
}
}
else if (Cs >= 1641) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (refDCval < 65) {
if (Rs < 8850) {
if (tcor < 92) {
if (Cs < 2088) {
return 8;
}
else if (Cs >= 2088) {
return 4;
}
}
else if (tcor >= 92) {
return 8;
}
}
else if (Rs >= 8850) {
return 2;
}
}
else if (refDCval >= 65) {
if (diffTSC < -1) {
if (diffRsCsdiff < -1) {
if (ssDCval < 69) {
if (Rs < 3548) {
return 4;
}
else if (Rs >= 3548) {
return 8;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
return 4;
}
}
else if (diffTSC >= -1) {
if (Cs < 1945) {
return 8;
}
else if (Cs >= 1945) {
if (Rs < 4350) {
return 4;
}
else if (Rs >= 4350) {
return 8;
}
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 3314) {
if (negBalance < 19) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 3537) {
if (posBalance < 36) {
if (Rs < 1820) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
return 8;
}
}
else if (Rs >= 1820) {
return 1;
}
}
else if (posBalance >= 36) {
if (TSCindex < 2) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (negBalance < 15) {
if (Rs < 1564) {
return 4;
}
else if (Rs >= 1564) {
return 1;
}
}
else if (negBalance >= 15) {
if (Cs < 1690) {
return 8;
}
else if (Cs >= 1690) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 8;
}
}
}
}
}
else if (TSCindex >= 2) {
if (Cs < 1786) {
if (Rs < 2422) {
return 8;
}
else if (Rs >= 2422) {
return 1;
}
}
else if (Cs >= 1786) {
return 1;
}
}
}
}
else if (Rs >= 3537) {
return 4;
}
}
}
else if (negBalance >= 19) {
if (Rs < 1659) {
if (ssDCval < 57) {
if (Rs < 1049) {
if (Rs < 791) {
return 1;
}
else if (Rs >= 791) {
return 8;
}
}
else if (Rs >= 1049) {
return 4;
}
}
else if (ssDCval >= 57) {
if (negBalance < 27) {
if (mcTcor < 98) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (Rs >= 1659) {
if (refDCval < 66) {
if (diffRsCsdiff < 3) {
if (posBalance < 34) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
return 1;
}
}
else if (posBalance >= 34) {
if (negBalance < 21) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 2189) {
return 8;
}
else if (Rs >= 2189) {
return 1;
}
}
}
else if (negBalance >= 21) {
return 8;
}
}
}
else if (diffRsCsdiff >= 3) {
if (Rs < 2420) {
return 8;
}
else if (Rs >= 2420) {
return 1;
}
}
}
else if (refDCval >= 66) {
if (Rs < 3459) {
return 1;
}
else if (Rs >= 3459) {
return 8;
}
}
}
}
}
else if (Cs >= 3314) {
return 8;
}
}
else if (ssDCval >= 67) {
if (Cs < 2289) {
if (mu_mv_mag_sq < 123) {
if (Rs < 2069) {
return 8;
}
else if (Rs >= 2069) {
if (Rs < 2386) {
if (ssDCval < 92) {
return 1;
}
else if (ssDCval >= 92) {
if (Rs < 2326) {
if (TSC < 1) {
return 2;
}
else if (TSC >= 1) {
return 1;
}
}
else if (Rs >= 2326) {
if (TSCindex < 1) {
if (Cs < 2115) {
return 4;
}
else if (Cs >= 2115) {
return 2;
}
}
else if (TSCindex >= 1) {
return 4;
}
}
}
}
else if (Rs >= 2386) {
if (ssDCval < 90) {
if (Cs < 2167) {
return 2;
}
else if (Cs >= 2167) {
if (Rs < 3781) {
return 1;
}
else if (Rs >= 3781) {
return 4;
}
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1076) {
if (Rs < 623) {
return 8;
}
else if (Rs >= 623) {
return 1;
}
}
else if (Rs >= 1076) {
if (Rs < 3840) {
if (Cs < 1644) {
return 4;
}
else if (Cs >= 1644) {
if (MVDiff < 4) {
if (diffRsCsdiff < -40) {
if (Rs < 2622) {
return 4;
}
else if (Rs >= 2622) {
return 8;
}
}
else if (diffRsCsdiff >= -40) {
if (mu_mv_mag_sq < 125) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
if (negBalance < 1) {
if (Rs < 2441) {
return 4;
}
else if (Rs >= 2441) {
return 8;
}
}
else if (negBalance >= 1) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 2140) {
return 4;
}
else if (Rs >= 2140) {
return 8;
}
}
}
}
else if (Rs >= 3840) {
return 4;
}
}
}
}
else if (Cs >= 2289) {
if (negBalance < 23) {
if (diffRsCsdiff < -12) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
else if (negBalance >= 23) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (Rs < 3980) {
if (Rs < 2502) {
return 4;
}
else if (Rs >= 2502) {
return 8;
}
}
else if (Rs >= 3980) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6692) {
if (SC < 8254) {
if (Rs < 3695) {
if (negBalance < 54) {
if (tcor < 99) {
if (AFD < 10) {
if (SC < 6942) {
if (mu_mv_mag_sq < 118) {
if (negBalance < 45) {
if (Cs < 2708) {
if (refDCval < 88) {
if (Rs < 2617) {
if (negBalance < 26) {
if (mcTcor < 96) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 1048) {
return 8;
}
else if (Rs >= 1048) {
return 4;
}
}
}
else if (mcTcor >= 96) {
return 8;
}
}
else if (negBalance >= 26) {
if (Cs < 2544) {
if (gchDC < 2) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 8;
}
}
else if (gchDC >= 2) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
return 8;
}
}
}
}
else if (Cs >= 2544) {
return 4;
}
}
}
else if (Rs >= 2617) {
if (mcTcor < 97) {
if (diffRsCsdiff < -22) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 3469) {
return 8;
}
else if (Rs >= 3469) {
return 4;
}
}
}
else if (diffRsCsdiff >= -22) {
if (diffRsCsdiff < 27) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 3271) {
if (diffRsCsdiff < 6) {
return 8;
}
else if (diffRsCsdiff >= 6) {
if (refDCval < 78) {
return 4;
}
else if (refDCval >= 78) {
return 8;
}
}
}
else if (Rs >= 3271) {
if (diffRsCsdiff < 12) {
return 4;
}
else if (diffRsCsdiff >= 12) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 27) {
if (Rs < 3256) {
if (Rs < 3023) {
return 4;
}
else if (Rs >= 3023) {
return 8;
}
}
else if (Rs >= 3256) {
return 2;
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 1663) {
return 2;
}
else if (Cs >= 1663) {
if (Rs < 3479) {
return 4;
}
else if (Rs >= 3479) {
return 4;
}
}
}
}
}
else if (refDCval >= 88) {
if (Rs < 3665) {
return 8;
}
else if (Rs >= 3665) {
return 2;
}
}
}
else if (Cs >= 2708) {
if (ssDCval < 81) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
else if (ssDCval >= 81) {
return 4;
}
}
}
else if (negBalance >= 45) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (SC < 4191) {
if (Rs < 1021) {
return 1;
}
else if (Rs >= 1021) {
return 8;
}
}
else if (SC >= 4191) {
if (Cs < 1898) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Rs < 2531) {
return 8;
}
else if (Rs >= 2531) {
if (Cs < 1712) {
if (negBalance < 43) {
return 4;
}
else if (negBalance >= 43) {
return 4;
}
}
else if (Cs >= 1712) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (AFD < 4) {
if (mu_mv_mag_sq < 123) {
return 4;
}
else if (mu_mv_mag_sq >= 123) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
if (diffRsCsdiff < 6) {
if (RsCsDiff < 11) {
if (ssDCval < 65) {
if (mcTcor < 98) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
return 4;
}
}
else if (mcTcor >= 98) {
if (Rs < 2964) {
return 4;
}
else if (Rs >= 2964) {
if (Rs < 2998) {
return 8;
}
else if (Rs >= 2998) {
if (Rs < 3043) {
return 4;
}
else if (Rs >= 3043) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 65) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (Rs < 2964) {
return 4;
}
else if (Rs >= 2964) {
return 4;
}
}
}
}
else if (RsCsDiff >= 11) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
return 8;
}
}
}
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1898) {
if (Rs < 2145) {
if (refDCval < 71) {
return 4;
}
else if (refDCval >= 71) {
if (Rs < 1958) {
return 4;
}
else if (Rs >= 1958) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
else if (Rs >= 2145) {
if (AFD < 9) {
if (ssDCval < 66) {
if (refDCval < 64) {
if (Cs < 1978) {
if (Rs < 2691) {
return 8;
}
else if (Rs >= 2691) {
return 4;
}
}
else if (Cs >= 1978) {
if (AFD < 2) {
if (Rs < 3346) {
return 1;
}
else if (Rs >= 3346) {
return 8;
}
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (refDCval >= 64) {
if (RsDiff < 3) {
if (Rs < 2968) {
return 8;
}
else if (Rs >= 2968) {
if (Cs < 1933) {
return 4;
}
else if (Cs >= 1933) {
if (Rs < 3044) {
return 8;
}
else if (Rs >= 3044) {
return 8;
}
}
}
}
else if (RsDiff >= 3) {
return 4;
}
}
}
else if (ssDCval >= 66) {
if (AFD < 6) {
if (refDCval < 81) {
return 8;
}
else if (refDCval >= 81) {
if (SC < 6868) {
if (Rs < 2706) {
if (Cs < 2540) {
if (Rs < 2702) {
return 8;
}
else if (Rs >= 2702) {
return 4;
}
}
else if (Cs >= 2540) {
return 4;
}
}
else if (Rs >= 2706) {
return 8;
}
}
else if (SC >= 6868) {
return 4;
}
}
}
else if (AFD >= 6) {
if (Rs < 3170) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (diffRsCsdiff < -19) {
return 4;
}
else if (diffRsCsdiff >= -19) {
return 8;
}
}
}
else if (Rs >= 3170) {
if (negBalance < 49) {
if (diffRsCsdiff < -21) {
return 4;
}
else if (diffRsCsdiff >= -21) {
if (diffTSC < -1) {
if (Rs < 3307) {
return 8;
}
else if (Rs >= 3307) {
return 4;
}
}
else if (diffTSC >= -1) {
if (refDCval < 67) {
return 4;
}
else if (refDCval >= 67) {
return 8;
}
}
}
}
else if (negBalance >= 49) {
return 4;
}
}
}
}
}
else if (AFD >= 9) {
if (Rs < 3182) {
if (Rs < 2206) {
return 4;
}
else if (Rs >= 2206) {
if (refDCval < 98) {
return 8;
}
else if (refDCval >= 98) {
if (Rs < 3044) {
return 8;
}
else if (Rs >= 3044) {
return 4;
}
}
}
}
else if (Rs >= 3182) {
if (Cs < 3345) {
return 4;
}
else if (Cs >= 3345) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 6942) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
return 4;
}
}
else if (mu_mv_mag_sq >= 126) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (Cs < 4322) {
return 8;
}
else if (Cs >= 4322) {
return 4;
}
}
}
}
}
}
else if (AFD >= 10) {
if (Cs < 2388) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
return 8;
}
}
else if (Cs >= 2388) {
if (Rs < 2638) {
return 4;
}
else if (Rs >= 2638) {
if (Rs < 2951) {
return 8;
}
else if (Rs >= 2951) {
if (Cs < 3253) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
if (SC < 6323) {
return 8;
}
else if (SC >= 6323) {
return 4;
}
}
}
}
else if (Cs >= 3253) {
if (ssDCval < 90) {
if (Rs < 3567) {
return 4;
}
else if (Rs >= 3567) {
return 8;
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 99) {
if (Cs < 2284) {
if (SC < 5127) {
if (Rs < 2489) {
return 8;
}
else if (Rs >= 2489) {
if (posBalance < 40) {
if (Rs < 3005) {
return 8;
}
else if (Rs >= 3005) {
return 4;
}
}
else if (posBalance >= 40) {
if (Cs < 1913) {
return 4;
}
else if (Cs >= 1913) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 8;
}
}
}
}
}
else if (SC >= 5127) {
if (negBalance < 22) {
return 8;
}
else if (negBalance >= 22) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (Cs >= 2284) {
if (diffAFD < -2) {
if (Rs < 3260) {
return 8;
}
else if (Rs >= 3260) {
return 4;
}
}
else if (diffAFD >= -2) {
if (SC < 7221) {
return 8;
}
else if (SC >= 7221) {
return 4;
}
}
}
}
}
else if (negBalance >= 54) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
if (Rs < 3035) {
if (tcor < 105) {
return 4;
}
else if (tcor >= 105) {
return 8;
}
}
else if (Rs >= 3035) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 3409) {
return 4;
}
else if (Rs >= 3409) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3695) {
if (refDCval < 81) {
if (refDCval < 60) {
if (Cs < 3141) {
if (negBalance < 8) {
return 8;
}
else if (negBalance >= 8) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (negBalance < 16) {
if (AFD < 4) {
if (AFD < 3) {
if (Rs < 4220) {
return 4;
}
else if (Rs >= 4220) {
if (Rs < 5098) {
if (Cs < 2729) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (Rs < 5061) {
return 4;
}
else if (Rs >= 5061) {
return 8;
}
}
}
else if (Cs >= 2729) {
return 8;
}
}
else if (Rs >= 5098) {
return 8;
}
}
}
else if (AFD >= 3) {
if (ssDCval < 54) {
if (Rs < 4355) {
if (diffRsCsdiff < -12) {
if (Rs < 4299) {
return 4;
}
else if (Rs >= 4299) {
return 4;
}
}
else if (diffRsCsdiff >= -12) {
return 4;
}
}
else if (Rs >= 4355) {
if (Rs < 4387) {
return 8;
}
else if (Rs >= 4387) {
return 4;
}
}
}
else if (ssDCval >= 54) {
if (Rs < 4726) {
return 4;
}
else if (Rs >= 4726) {
if (posBalance < 35) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (posBalance >= 35) {
if (RsCsDiff < 33) {
return 4;
}
else if (RsCsDiff >= 33) {
return 4;
}
}
}
}
}
}
else if (AFD >= 4) {
if (diffMVdiffVal < 1) {
if (Cs < 2138) {
return 8;
}
else if (Cs >= 2138) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
}
else if (negBalance >= 16) {
if (TSC < 3) {
if (mu_mv_mag_sq < 112) {
return 8;
}
else if (mu_mv_mag_sq >= 112) {
if (CsDiff < 4) {
if (Cs < 2753) {
if (RsDiff < 5) {
if (refDCval < 54) {
return 8;
}
else if (refDCval >= 54) {
return 4;
}
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (Cs >= 2753) {
return 4;
}
}
else if (CsDiff >= 4) {
if (MVDiff < 3) {
if (Cs < 2573) {
if (negBalance < 17) {
return 8;
}
else if (negBalance >= 17) {
if (Rs < 4519) {
return 4;
}
else if (Rs >= 4519) {
if (Cs < 2502) {
return 8;
}
else if (Cs >= 2502) {
return 8;
}
}
}
}
else if (Cs >= 2573) {
return 8;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
}
}
else if (TSC >= 3) {
if (mcTcor < 96) {
if (diffRsCsdiff < 2) {
if (Cs < 2491) {
if (Rs < 4503) {
return 8;
}
else if (Rs >= 4503) {
if (Rs < 4624) {
return 8;
}
else if (Rs >= 4624) {
return 8;
}
}
}
else if (Cs >= 2491) {
return 4;
}
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Rs < 4674) {
if (RsCsDiff < 73) {
if (posBalance < 32) {
if (diffRsCsdiff < 2) {
if (Cs < 2307) {
return 4;
}
else if (Cs >= 2307) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
else if (posBalance >= 32) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 4;
}
}
}
else if (RsCsDiff >= 73) {
return 4;
}
}
else if (Rs >= 4674) {
return 4;
}
}
}
}
}
}
}
}
else if (Cs >= 3141) {
return 8;
}
}
else if (refDCval >= 60) {
if (SC < 5930) {
if (SC < 5844) {
return 2;
}
else if (SC >= 5844) {
return 4;
}
}
else if (SC >= 5930) {
if (SC < 6049) {
if (Rs < 3980) {
return 8;
}
else if (Rs >= 3980) {
return 2;
}
}
else if (SC >= 6049) {
return 8;
}
}
}
}
else if (refDCval >= 81) {
if (Cs < 2735) {
if (CsDiff < 8) {
if (negBalance < 29) {
if (Cs < 2307) {
if (ssDCval < 99) {
if (SC < 7072) {
if (SC < 6616) {
if (tcor < 97) {
if (Cs < 2089) {
return 8;
}
else if (Cs >= 2089) {
return 2;
}
}
else if (tcor >= 97) {
if (Cs < 1728) {
return 4;
}
else if (Cs >= 1728) {
return 2;
}
}
}
else if (SC >= 6616) {
return 8;
}
}
else if (SC >= 7072) {
return 4;
}
}
else if (ssDCval >= 99) {
return 4;
}
}
else if (Cs >= 2307) {
if (Rs < 4978) {
return 8;
}
else if (Rs >= 4978) {
return 4;
}
}
}
else if (negBalance >= 29) {
if (posBalance < 44) {
if (ssDCval < 91) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
if (Rs < 3713) {
return 2;
}
else if (Rs >= 3713) {
return 8;
}
}
}
else if (ssDCval >= 91) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (posBalance >= 44) {
if (Cs < 2431) {
return 2;
}
else if (Cs >= 2431) {
return 4;
}
}
}
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (Cs >= 2735) {
if (AFD < 11) {
if (ssDCval < 98) {
if (RsDiff < 5) {
if (mu_mv_mag_sq < 128) {
if (SC < 7425) {
return 8;
}
else if (SC >= 7425) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (SC < 7065) {
return 8;
}
else if (SC >= 7065) {
return 4;
}
}
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
else if (SC >= 8254) {
if (MVDiff < 3) {
if (AFD < 2) {
if (Rs < 4763) {
return 1;
}
else if (Rs >= 4763) {
return 8;
}
}
else if (AFD >= 2) {
if (Cs < 3666) {
if (SC < 8819) {
if (Rs < 4801) {
return 8;
}
else if (Rs >= 4801) {
return 4;
}
}
else if (SC >= 8819) {
return 8;
}
}
else if (Cs >= 3666) {
if (mu_mv_mag_sq < 112) {
if (Rs < 6015) {
if (Rs < 5721) {
return 8;
}
else if (Rs >= 5721) {
if (Rs < 5881) {
return 2;
}
else if (Rs >= 5881) {
return 8;
}
}
}
else if (Rs >= 6015) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (posBalance < 44) {
if (diffAFD < -1) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (posBalance >= 44) {
return 8;
}
}
}
}
}
else if (MVDiff >= 3) {
if (Rs < 5813) {
return 8;
}
else if (Rs >= 5813) {
if (ssDCval < 82) {
if (Cs < 3885) {
return 8;
}
else if (Cs >= 3885) {
return 2;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
}
}
}
else if (Rs >= 6692) {
if (ssDCval < 68) {
if (Rs < 8744) {
return 8;
}
else if (Rs >= 8744) {
return 2;
}
}
else if (ssDCval >= 68) {
if (ssDCval < 95) {
return 4;
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (mcTcor < 96) {
if (ssDCval < 137) {
if (refDCval < 105) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Cs < 3040) {
if (Rs < 4923) {
if (Rs < 3017) {
return 8;
}
else if (Rs >= 3017) {
return 1;
}
}
else if (Rs >= 4923) {
return 2;
}
}
else if (Cs >= 3040) {
if (Rs < 1866) {
return 2;
}
else if (Rs >= 1866) {
return 8;
}
}
}
}
else if (refDCval >= 105) {
if (gchDC < 2) {
if (negBalance < 54) {
if (posBalance < 63) {
if (diffTSC < 3) {
if (refDCval < 123) {
return 8;
}
else if (refDCval >= 123) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
return 8;
}
}
}
}
else if (diffTSC >= 3) {
if (ssDCval < 108) {
if (Rs < 4186) {
return 4;
}
else if (Rs >= 4186) {
return 2;
}
}
else if (ssDCval >= 108) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (CsDiff < 5) {
if (Rs < 2319) {
return 8;
}
else if (Rs >= 2319) {
return 1;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (negBalance >= 54) {
if (Rs < 4134) {
if (Rs < 2808) {
return 8;
}
else if (Rs >= 2808) {
return 2;
}
}
else if (Rs >= 4134) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (Cs < 2255) {
if (Rs < 2055) {
return 4;
}
else if (Rs >= 2055) {
return 1;
}
}
else if (Cs >= 2255) {
if (ssDCval < 112) {
return 8;
}
else if (ssDCval >= 112) {
if (Rs < 681) {
return 8;
}
else if (Rs >= 681) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 137) {
if (posBalance < 52) {
if (ssDCval < 200) {
if (Rs < 1316) {
return 1;
}
else if (Rs >= 1316) {
if (diffAFD < 6) {
return 8;
}
else if (diffAFD >= 6) {
return 1;
}
}
}
else if (ssDCval >= 200) {
return 4;
}
}
else if (posBalance >= 52) {
if (Cs < 4095) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 4095) {
if (Rs < 8232) {
return 8;
}
else if (Rs >= 8232) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1909) {
if (ssDCval < 115) {
if (Cs < 1957) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 1614) {
return 2;
}
else if (Rs >= 1614) {
return 8;
}
}
}
else if (Cs >= 1957) {
if (diffRsCsdiff < -6) {
return 8;
}
else if (diffRsCsdiff >= -6) {
return 2;
}
}
}
else if (ssDCval >= 115) {
if (posBalance < 55) {
if (Rs < 926) {
return 1;
}
else if (Rs >= 926) {
return 8;
}
}
else if (posBalance >= 55) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
return 8;
}
}
}
}
else if (Rs >= 1909) {
if (refDCval < 116) {
if (refDCval < 105) {
if (Rs < 4958) {
if (Cs < 1996) {
if (AFD < 5) {
if (Rs < 3027) {
return 2;
}
else if (Rs >= 3027) {
return 4;
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (Cs >= 1996) {
if (diffRsCsdiff < 10) {
return 8;
}
else if (diffRsCsdiff >= 10) {
if (Cs < 3043) {
if (Rs < 3649) {
return 8;
}
else if (Rs >= 3649) {
return 1;
}
}
else if (Cs >= 3043) {
return 8;
}
}
}
}
else if (Rs >= 4958) {
if (Cs < 2902) {
return 2;
}
else if (Cs >= 2902) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (diffRsCsdiff < -70) {
if (ssDCval < 108) {
if (negBalance < 16) {
return 8;
}
else if (negBalance >= 16) {
if (Rs < 4034) {
return 4;
}
else if (Rs >= 4034) {
return 2;
}
}
}
else if (ssDCval >= 108) {
return 8;
}
}
else if (diffRsCsdiff >= -70) {
if (negBalance < 55) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (Rs < 2491) {
return 1;
}
else if (Rs >= 2491) {
return 8;
}
}
}
else if (negBalance >= 55) {
return 2;
}
}
}
}
else if (refDCval >= 116) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 33) {
if (ssDCval < 146) {
if (Rs < 2770) {
if (Rs < 2597) {
return 8;
}
else if (Rs >= 2597) {
return 4;
}
}
else if (Rs >= 2770) {
if (Cs < 3966) {
return 2;
}
else if (Cs >= 3966) {
return 8;
}
}
}
else if (ssDCval >= 146) {
return 8;
}
}
else if (posBalance >= 33) {
if (Rs < 6429) {
if (Cs < 1675) {
if (Rs < 3979) {
return 8;
}
else if (Rs >= 3979) {
return 2;
}
}
else if (Cs >= 1675) {
return 8;
}
}
else if (Rs >= 6429) {
if (Cs < 4415) {
if (Rs < 7121) {
return 4;
}
else if (Rs >= 7121) {
return 2;
}
}
else if (Cs >= 4415) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 4787) {
if (Cs < 3939) {
if (AFD < 6) {
if (diffTSC < -6) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
if (Rs < 2386) {
return 8;
}
else if (Rs >= 2386) {
return 1;
}
}
}
else if (diffTSC >= -6) {
if (SC < 3950) {
if (Rs < 2118) {
return 8;
}
else if (Rs >= 2118) {
return 1;
}
}
else if (SC >= 3950) {
return 8;
}
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 3939) {
if (ssDCval < 137) {
return 1;
}
else if (ssDCval >= 137) {
if (Rs < 3609) {
return 4;
}
else if (Rs >= 3609) {
return 8;
}
}
}
}
else if (Rs >= 4787) {
if (Cs < 1966) {
return 2;
}
else if (Cs >= 1966) {
return 8;
}
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect7(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1325) {
if (Rs < 2388) {
if (Cs < 533) {
if (negBalance < 14) {
if (MVDiff < 7) {
if (SC < 793) {
if (refDCval < 47) {
if (SC < 279) {
if (negBalance < 8) {
if (Rs < 2) {
if (ssDCval < 22) {
if (ssDCval < 9) {
return 4;
}
else if (ssDCval >= 9) {
return 8;
}
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (Rs >= 2) {
return 8;
}
}
else if (negBalance >= 8) {
if (Rs < 95) {
if (ssDCval < 39) {
if (refDCval < 25) {
if (mu_mv_mag_sq < 96) {
if (Cs < 32) {
return 1;
}
else if (Cs >= 32) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 2) {
return 4;
}
else if (Rs >= 2) {
return 8;
}
}
}
else if (refDCval >= 25) {
if (Rs < 86) {
if (posBalance < 35) {
return 4;
}
else if (posBalance >= 35) {
if (mu_mv_mag_sq < 104) {
return 1;
}
else if (mu_mv_mag_sq >= 104) {
return 4;
}
}
}
else if (Rs >= 86) {
if (Cs < 104) {
return 4;
}
else if (Cs >= 104) {
return 8;
}
}
}
}
else if (ssDCval >= 39) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
else if (Rs >= 95) {
if (ssDCval < 38) {
if (posBalance < 36) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
return 8;
}
}
else if (posBalance >= 36) {
if (Rs < 115) {
return 8;
}
else if (Rs >= 115) {
return 1;
}
}
}
else if (ssDCval >= 38) {
return 1;
}
}
}
}
else if (SC >= 279) {
if (Rs < 232) {
if (ssDCval < 36) {
if (mu_mv_mag_sq < 99) {
if (Rs < 199) {
return 4;
}
else if (Rs >= 199) {
return 2;
}
}
else if (mu_mv_mag_sq >= 99) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (diffMVdiffVal < -2) {
if (Cs < 220) {
return 1;
}
else if (Cs >= 220) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
}
else if (ssDCval >= 36) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (Cs < 233) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
return 1;
}
}
else if (Cs >= 233) {
return 1;
}
}
}
}
else if (Rs >= 232) {
if (ssDCval < 22) {
if (diffRsCsdiff < -3) {
return 4;
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
else if (ssDCval >= 22) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
if (mu_mv_mag_sq < 118) {
if (ssDCval < 40) {
if (Cs < 254) {
return 1;
}
else if (Cs >= 254) {
return 4;
}
}
else if (ssDCval >= 40) {
if (Cs < 188) {
return 4;
}
else if (Cs >= 188) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (Cs < 470) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 470) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 47) {
if (ssDCval < 82) {
if (Rs < 17) {
return 2;
}
else if (Rs >= 17) {
if (ssDCval < 67) {
if (Cs < 255) {
if (ssDCval < 58) {
if (refDCval < 54) {
if (SC < 404) {
return 8;
}
else if (SC >= 404) {
return 4;
}
}
else if (refDCval >= 54) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
return 4;
}
}
}
else if (ssDCval >= 58) {
if (Rs < 190) {
if (Rs < 170) {
if (Rs < 125) {
return 2;
}
else if (Rs >= 125) {
return 8;
}
}
else if (Rs >= 170) {
return 1;
}
}
else if (Rs >= 190) {
return 8;
}
}
}
else if (Cs >= 255) {
if (Rs < 251) {
return 4;
}
else if (Rs >= 251) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
if (Rs < 379) {
return 4;
}
else if (Rs >= 379) {
return 8;
}
}
}
}
}
else if (ssDCval >= 67) {
if (Cs < 118) {
if (Rs < 144) {
return 4;
}
else if (Rs >= 144) {
return 8;
}
}
else if (Cs >= 118) {
if (Rs < 340) {
return 2;
}
else if (Rs >= 340) {
if (Cs < 265) {
return 4;
}
else if (Cs >= 265) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
if (TSC < 2) {
if (Rs < 463) {
if (posBalance < 12) {
return 2;
}
else if (posBalance >= 12) {
return 4;
}
}
else if (Rs >= 463) {
return 1;
}
}
else if (TSC >= 2) {
if (Rs < 389) {
return 8;
}
else if (Rs >= 389) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 793) {
if (mu_mv_mag_sq < 128) {
if (Cs < 268) {
if (ssDCval < 39) {
if (Rs < 554) {
return 8;
}
else if (Rs >= 554) {
return 4;
}
}
else if (ssDCval >= 39) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (mu_mv_mag_sq < 125) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (Rs < 1301) {
return 8;
}
else if (Rs >= 1301) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1020) {
return 8;
}
else if (Rs >= 1020) {
return 2;
}
}
}
}
}
else if (Cs >= 268) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Cs < 480) {
if (RsCsDiff < 2) {
if (Rs < 894) {
if (diffRsCsdiff < -2) {
return 1;
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (Rs >= 894) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
if (Rs < 1157) {
return 1;
}
else if (Rs >= 1157) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
if (Rs < 942) {
return 1;
}
else if (Rs >= 942) {
return 4;
}
}
}
else if (Cs >= 480) {
if (ssDCval < 55) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
if (Rs < 588) {
return 1;
}
else if (Rs >= 588) {
return 2;
}
}
}
else if (ssDCval >= 55) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (Rs < 551) {
if (posBalance < 27) {
return 1;
}
else if (posBalance >= 27) {
return 8;
}
}
else if (Rs >= 551) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
if (Rs < 1471) {
return 2;
}
else if (Rs >= 1471) {
return 1;
}
}
}
}
}
else if (ssDCval >= 67) {
if (ssDCval < 141) {
if (posBalance < 39) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
if (Rs < 1044) {
return 8;
}
else if (Rs >= 1044) {
return 2;
}
}
}
else if (posBalance >= 39) {
if (negBalance < 7) {
return 8;
}
else if (negBalance >= 7) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (ssDCval >= 141) {
if (Rs < 494) {
return 1;
}
else if (Rs >= 494) {
if (Rs < 547) {
if (MVDiff < 6) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
return 1;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (Rs >= 547) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (ssDCval < 84) {
if (Cs < 359) {
if (Cs < 256) {
return 2;
}
else if (Cs >= 256) {
return 1;
}
}
else if (Cs >= 359) {
if (Rs < 1186) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 725) {
return 4;
}
else if (Rs >= 725) {
return 8;
}
}
}
else if (Rs >= 1186) {
return 2;
}
}
}
else if (ssDCval >= 84) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (SC < 1090) {
if (Rs < 575) {
return 8;
}
else if (Rs >= 575) {
return 4;
}
}
else if (SC >= 1090) {
return 8;
}
}
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (tcor < 94) {
if (negBalance < 5) {
if (posBalance < 93) {
if (diffMVdiffVal < 160) {
if (mu_mv_mag_sq < 39) {
if (Cs < 65) {
if (posBalance < 85) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 2;
}
}
else if (posBalance >= 85) {
return 8;
}
}
else if (Cs >= 65) {
return 2;
}
}
else if (mu_mv_mag_sq >= 39) {
if (refDCval < 50) {
if (diffMVdiffVal < -18) {
if (diffMVdiffVal < -22) {
return 8;
}
else if (diffMVdiffVal >= -22) {
if (posBalance < 76) {
return 4;
}
else if (posBalance >= 76) {
return 8;
}
}
}
else if (diffMVdiffVal >= -18) {
return 8;
}
}
else if (refDCval >= 50) {
if (Rs < 49) {
return 2;
}
else if (Rs >= 49) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 160) {
return 4;
}
}
else if (posBalance >= 93) {
if (mcTcor < 67) {
if (posBalance < 124) {
if (Rs < 13) {
return 8;
}
else if (Rs >= 13) {
return 2;
}
}
else if (posBalance >= 124) {
if (Cs < 231) {
if (SC < 1) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (SC >= 1) {
return 1;
}
}
else if (Cs >= 231) {
return 4;
}
}
}
else if (mcTcor >= 67) {
if (ssDCval < 103) {
if (diffMVdiffVal < 4) {
if (tcor < 70) {
if (mcTcor < 73) {
return 8;
}
else if (mcTcor >= 73) {
return 2;
}
}
else if (tcor >= 70) {
if (Cs < 3) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
return 2;
}
}
else if (Cs >= 3) {
if (mu_mv_mag_sq < 23) {
return 4;
}
else if (mu_mv_mag_sq >= 23) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= 4) {
if (tcor < 84) {
return 2;
}
else if (tcor >= 84) {
if (mcTcor < 90) {
if (Rs < 101) {
return 8;
}
else if (Rs >= 101) {
return 4;
}
}
else if (mcTcor >= 90) {
return 4;
}
}
}
}
else if (ssDCval >= 103) {
return 8;
}
}
}
}
else if (negBalance >= 5) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (ssDCval < 36) {
if (Rs < 9) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 9) {
if (tcor < 78) {
if (posBalance < 65) {
if (tcor < 61) {
if (Rs < 228) {
return 1;
}
else if (Rs >= 228) {
return 8;
}
}
else if (tcor >= 61) {
return 1;
}
}
else if (posBalance >= 65) {
if (Rs < 146) {
return 8;
}
else if (Rs >= 146) {
if (Rs < 307) {
return 1;
}
else if (Rs >= 307) {
return 4;
}
}
}
}
else if (tcor >= 78) {
if (Cs < 207) {
if (MVDiff < 187) {
return 1;
}
else if (MVDiff >= 187) {
return 4;
}
}
else if (Cs >= 207) {
if (tcor < 91) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (tcor >= 91) {
return 4;
}
}
}
}
}
else if (ssDCval >= 36) {
if (refDCval < 146) {
if (MVDiff < 13) {
if (posBalance < 95) {
return 8;
}
else if (posBalance >= 95) {
if (Cs < 356) {
return 2;
}
else if (Cs >= 356) {
return 4;
}
}
}
else if (MVDiff >= 13) {
if (posBalance < 104) {
if (mcTcor < 70) {
if (Rs < 436) {
return 2;
}
else if (Rs >= 436) {
return 1;
}
}
else if (mcTcor >= 70) {
if (AFD < 6) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
return 1;
}
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (posBalance >= 104) {
if (tcor < 20) {
if (Rs < 254) {
return 1;
}
else if (Rs >= 254) {
return 2;
}
}
else if (tcor >= 20) {
if (CsDiff < 1) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (Rs < 245) {
return 8;
}
else if (Rs >= 245) {
return 2;
}
}
}
else if (CsDiff >= 1) {
return 8;
}
}
}
}
}
else if (refDCval >= 146) {
return 2;
}
}
}
}
}
else if (tcor >= 94) {
if (negBalance < 6) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
if (posBalance < 14) {
return 8;
}
else if (posBalance >= 14) {
if (Rs < 19) {
return 8;
}
else if (Rs >= 19) {
return 1;
}
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (posBalance < 5) {
if (Rs < 778) {
return 4;
}
else if (Rs >= 778) {
return 8;
}
}
else if (posBalance >= 5) {
if (diffRsCsdiff < -7) {
if (diffAFD < -8) {
if (ssDCval < 147) {
if (Rs < 392) {
return 8;
}
else if (Rs >= 392) {
return 2;
}
}
else if (ssDCval >= 147) {
if (Rs < 391) {
if (Rs < 100) {
return 2;
}
else if (Rs >= 100) {
return 1;
}
}
else if (Rs >= 391) {
return 8;
}
}
}
else if (diffAFD >= -8) {
return 1;
}
}
else if (diffRsCsdiff >= -7) {
if (diffRsCsdiff < -1) {
if (MVDiff < 41) {
if (diffAFD < -4) {
if (Cs < 270) {
return 8;
}
else if (Cs >= 270) {
return 2;
}
}
else if (diffAFD >= -4) {
return 4;
}
}
else if (MVDiff >= 41) {
return 4;
}
}
else if (diffRsCsdiff >= -1) {
if (mcTcor < 98) {
if (Cs < 13) {
return 8;
}
else if (Cs >= 13) {
if (Cs < 174) {
return 4;
}
else if (Cs >= 174) {
if (ssDCval < 38) {
return 4;
}
else if (ssDCval >= 38) {
return 8;
}
}
}
}
else if (mcTcor >= 98) {
if (ssDCval < 27) {
return 4;
}
else if (ssDCval >= 27) {
if (Rs < 19) {
return 2;
}
else if (Rs >= 19) {
if (ssDCval < 151) {
if (MVDiff < 19) {
if (mcTcor < 100) {
return 8;
}
else if (mcTcor >= 100) {
if (Cs < 177) {
return 8;
}
else if (Cs >= 177) {
return 1;
}
}
}
else if (MVDiff >= 19) {
if (gchDC < 3) {
if (posBalance < 40) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
if (Rs < 151) {
return 8;
}
else if (Rs >= 151) {
return 4;
}
}
}
else if (posBalance >= 40) {
return 4;
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (ssDCval >= 151) {
if (Rs < 159) {
return 2;
}
else if (Rs >= 159) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (negBalance >= 6) {
if (SC < 874) {
if (Cs < 118) {
if (TSCindex < 1) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
if (diffAFD < -1) {
if (Rs < 137) {
if (Rs < 17) {
return 2;
}
else if (Rs >= 17) {
return 1;
}
}
else if (Rs >= 137) {
return 8;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (TSCindex >= 1) {
if (ssDCval < 40) {
if (Cs < 107) {
return 1;
}
else if (Cs >= 107) {
return 8;
}
}
else if (ssDCval >= 40) {
if (ssDCval < 77) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (ssDCval >= 77) {
return 1;
}
}
}
}
else if (Cs >= 118) {
if (posBalance < 31) {
if (refDCval < 40) {
if (Rs < 166) {
return 8;
}
else if (Rs >= 166) {
return 4;
}
}
else if (refDCval >= 40) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (ssDCval < 161) {
return 1;
}
else if (ssDCval >= 161) {
return 4;
}
}
}
}
else if (posBalance >= 31) {
if (TSC < 2) {
if (mu_mv_mag_sq < 72) {
return 4;
}
else if (mu_mv_mag_sq >= 72) {
if (ssDCval < 82) {
if (Rs < 235) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
else if (Rs >= 235) {
return 4;
}
}
else if (ssDCval >= 82) {
return 4;
}
}
}
else if (TSC >= 2) {
if (Rs < 455) {
return 1;
}
else if (Rs >= 455) {
return 8;
}
}
}
}
}
else if (SC >= 874) {
if (Cs < 399) {
if (Cs < 48) {
return 8;
}
else if (Cs >= 48) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (Cs >= 399) {
if (ssDCval < 48) {
if (Rs < 527) {
return 4;
}
else if (Rs >= 527) {
return 1;
}
}
else if (ssDCval >= 48) {
if (Rs < 1453) {
return 8;
}
else if (Rs >= 1453) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (mcTcor < 82) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (ssDCval < 23) {
if (mu_mv_mag_sq < 46) {
if (negBalance < 53) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (mu_mv_mag_sq >= 46) {
return 4;
}
}
else if (ssDCval >= 23) {
if (Cs < 385) {
if (Rs < 516) {
if (negBalance < 43) {
if (Cs < 72) {
if (posBalance < 87) {
return 1;
}
else if (posBalance >= 87) {
return 8;
}
}
else if (Cs >= 72) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Cs < 81) {
return 2;
}
else if (Cs >= 81) {
if (mu_mv_mag_sq < 5) {
return 8;
}
else if (mu_mv_mag_sq >= 5) {
if (refDCval < 30) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (mcTcor < 69) {
return 1;
}
else if (mcTcor >= 69) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (refDCval >= 30) {
if (Cs < 261) {
if (Rs < 155) {
return 4;
}
else if (Rs >= 155) {
return 8;
}
}
else if (Cs >= 261) {
if (Rs < 207) {
return 2;
}
else if (Rs >= 207) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 43) {
if (Rs < 6) {
if (ssDCval < 24) {
return 8;
}
else if (ssDCval >= 24) {
return 4;
}
}
else if (Rs >= 6) {
if (posBalance < 29) {
return 2;
}
else if (posBalance >= 29) {
return 8;
}
}
}
}
else if (Rs >= 516) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (Cs >= 385) {
if (RsDiff < 4) {
if (MVDiff < 25) {
if (Cs < 481) {
return 8;
}
else if (Cs >= 481) {
return 1;
}
}
else if (MVDiff >= 25) {
return 2;
}
}
else if (RsDiff >= 4) {
return 8;
}
}
}
}
}
else if (mcTcor >= 82) {
if (posBalance < 27) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 113) {
if (negBalance < 22) {
if (ssDCval < 32) {
return 4;
}
else if (ssDCval >= 32) {
if (Cs < 112) {
return 8;
}
else if (Cs >= 112) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (posBalance < 23) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (posBalance >= 23) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
else if (MVDiff >= 6) {
if (Cs < 4) {
if (negBalance < 79) {
if (ssDCval < 20) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (ssDCval >= 20) {
return 4;
}
}
else if (negBalance >= 79) {
return 8;
}
}
else if (Cs >= 4) {
if (ssDCval < 31) {
if (negBalance < 61) {
return 1;
}
else if (negBalance >= 61) {
if (Rs < 96) {
return 8;
}
else if (Rs >= 96) {
if (Rs < 295) {
return 1;
}
else if (Rs >= 295) {
return 4;
}
}
}
}
else if (ssDCval >= 31) {
if (Rs < 248) {
return 8;
}
else if (Rs >= 248) {
if (Rs < 462) {
if (Rs < 382) {
return 1;
}
else if (Rs >= 382) {
return 2;
}
}
else if (Rs >= 462) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 27) {
if (refDCval < 30) {
if (posBalance < 42) {
if (Cs < 331) {
if (mu_mv_mag_sq < 42) {
if (mcTcor < 87) {
if (ssDCval < 24) {
return 1;
}
else if (ssDCval >= 24) {
return 4;
}
}
else if (mcTcor >= 87) {
if (ssDCval < 29) {
if (Cs < 263) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (Cs >= 263) {
return 4;
}
}
else if (ssDCval >= 29) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (ssDCval < 30) {
return 1;
}
else if (ssDCval >= 30) {
if (Rs < 208) {
return 4;
}
else if (Rs >= 208) {
return 1;
}
}
}
}
else if (Cs >= 331) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
else if (posBalance >= 42) {
if (posBalance < 71) {
if (MVDiff < 58) {
if (Cs < 88) {
return 1;
}
else if (Cs >= 88) {
if (SC < 414) {
if (AFD < 5) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (negBalance < 16) {
return 8;
}
else if (negBalance >= 16) {
if (SC < 390) {
return 1;
}
else if (SC >= 390) {
return 2;
}
}
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (SC >= 414) {
return 1;
}
}
}
else if (MVDiff >= 58) {
if (diffAFD < -4) {
if (Rs < 97) {
return 1;
}
else if (Rs >= 97) {
return 2;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
}
else if (posBalance >= 71) {
return 2;
}
}
}
else if (refDCval >= 30) {
if (RsDiff < 2) {
if (Cs < 253) {
if (Cs < 83) {
if (Cs < 10) {
if (Rs < 81) {
return 2;
}
else if (Rs >= 81) {
return 1;
}
}
else if (Cs >= 10) {
if (posBalance < 28) {
if (Cs < 67) {
if (Rs < 179) {
return 1;
}
else if (Rs >= 179) {
return 2;
}
}
else if (Cs >= 67) {
return 4;
}
}
else if (posBalance >= 28) {
if (tcor < 87) {
if (tcor < 84) {
if (Rs < 143) {
if (Cs < 21) {
return 4;
}
else if (Cs >= 21) {
return 1;
}
}
else if (Rs >= 143) {
if (Rs < 277) {
return 4;
}
else if (Rs >= 277) {
return 1;
}
}
}
else if (tcor >= 84) {
return 4;
}
}
else if (tcor >= 87) {
return 1;
}
}
}
}
else if (Cs >= 83) {
if (mcTcor < 95) {
if (SC < 295) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (gchDC < 2) {
if (Cs < 86) {
return 1;
}
else if (Cs >= 86) {
return 8;
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (SC >= 295) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (MVDiff < 12) {
if (AFD < 5) {
if (Rs < 445) {
return 4;
}
else if (Rs >= 445) {
return 1;
}
}
else if (AFD >= 5) {
if (SC < 357) {
return 1;
}
else if (SC >= 357) {
return 2;
}
}
}
else if (MVDiff >= 12) {
if (tcor < 75) {
return 4;
}
else if (tcor >= 75) {
if (ssDCval < 32) {
if (Rs < 191) {
return 8;
}
else if (Rs >= 191) {
return 1;
}
}
else if (ssDCval >= 32) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (Rs < 108) {
if (AFD < 3) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (Rs >= 108) {
if (Rs < 317) {
if (mcTcor < 97) {
if (negBalance < 22) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
return 1;
}
}
else if (negBalance >= 22) {
if (MVDiff < 11) {
if (Rs < 179) {
return 1;
}
else if (Rs >= 179) {
return 4;
}
}
else if (MVDiff >= 11) {
return 1;
}
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (Rs >= 317) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
return 1;
}
}
}
}
}
}
else if (Cs >= 253) {
if (MVDiff < 22) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (MVDiff >= 22) {
return 4;
}
}
}
else if (RsDiff >= 2) {
if (posBalance < 32) {
if (Cs < 305) {
return 4;
}
else if (Cs >= 305) {
return 1;
}
}
else if (posBalance >= 32) {
if (diffAFD < -5) {
if (Rs < 1166) {
return 8;
}
else if (Rs >= 1166) {
return 1;
}
}
else if (diffAFD >= -5) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 300) {
if (mcTcor < 97) {
if (refDCval < 41) {
if (gchDC < 1) {
if (RsCsDiff < 1) {
if (TSCindex < 3) {
if (mu_mv_mag_sq < 114) {
if (AFD < 3) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
return 8;
}
}
else if (AFD >= 3) {
if (AFD < 5) {
if (Cs < 41) {
return 2;
}
else if (Cs >= 41) {
return 1;
}
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (TSCindex >= 3) {
if (Cs < 30) {
return 2;
}
else if (Cs >= 30) {
if (mu_mv_mag_sq < 12) {
if (Cs < 153) {
return 8;
}
else if (Cs >= 153) {
return 2;
}
}
else if (mu_mv_mag_sq >= 12) {
return 8;
}
}
}
}
else if (RsCsDiff >= 1) {
if (Cs < 73) {
return 1;
}
else if (Cs >= 73) {
if (negBalance < 31) {
if (mu_mv_mag_sq < 88) {
if (Rs < 100) {
return 1;
}
else if (Rs >= 100) {
return 8;
}
}
else if (mu_mv_mag_sq >= 88) {
if (posBalance < 29) {
return 4;
}
else if (posBalance >= 29) {
return 1;
}
}
}
else if (negBalance >= 31) {
if (mcTcor < 89) {
if (diffMVdiffVal < 19) {
if (mu_mv_mag_sq < 3) {
if (Rs < 191) {
return 2;
}
else if (Rs >= 191) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (diffMVdiffVal >= 19) {
return 4;
}
}
else if (mcTcor >= 89) {
if (mu_mv_mag_sq < 69) {
if (AFD < 7) {
if (Rs < 134) {
if (Rs < 116) {
return 1;
}
else if (Rs >= 116) {
return 8;
}
}
else if (Rs >= 134) {
return 2;
}
}
else if (AFD >= 7) {
if (SC < 376) {
return 1;
}
else if (SC >= 376) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 69) {
return 8;
}
}
}
}
}
}
else if (gchDC >= 1) {
if (negBalance < 35) {
if (mu_mv_mag_sq < 33) {
if (diffMVdiffVal < -6) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
return 2;
}
}
else if (diffMVdiffVal >= -6) {
if (ssDCval < 39) {
return 1;
}
else if (ssDCval >= 39) {
if (Rs < 38) {
return 1;
}
else if (Rs >= 38) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (SC < 448) {
if (mu_mv_mag_sq < 95) {
if (ssDCval < 38) {
if (Cs < 179) {
if (Cs < 108) {
return 1;
}
else if (Cs >= 108) {
return 4;
}
}
else if (Cs >= 179) {
return 8;
}
}
else if (ssDCval >= 38) {
if (Rs < 215) {
return 1;
}
else if (Rs >= 215) {
if (Rs < 251) {
if (Cs < 139) {
return 1;
}
else if (Cs >= 139) {
return 8;
}
}
else if (Rs >= 251) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
return 8;
}
}
}
else if (SC >= 448) {
return 1;
}
}
}
else if (negBalance >= 35) {
if (AFD < 9) {
if (TSC < 3) {
if (mcTcor < 96) {
if (Cs < 204) {
return 1;
}
else if (Cs >= 204) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Rs < 120) {
return 8;
}
else if (Rs >= 120) {
return 2;
}
}
}
else if (TSC >= 3) {
if (tcor < 82) {
if (TSC < 4) {
if (AFD < 6) {
if (Rs < 128) {
return 8;
}
else if (Rs >= 128) {
return 2;
}
}
else if (AFD >= 6) {
if (Rs < 76) {
return 2;
}
else if (Rs >= 76) {
return 1;
}
}
}
else if (TSC >= 4) {
if (Rs < 114) {
if (Rs < 86) {
return 2;
}
else if (Rs >= 86) {
return 1;
}
}
else if (Rs >= 114) {
return 8;
}
}
}
else if (tcor >= 82) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
else if (AFD >= 9) {
if (posBalance < 53) {
if (TSC < 6) {
if (refDCval < 35) {
return 4;
}
else if (refDCval >= 35) {
return 1;
}
}
else if (TSC >= 6) {
if (Rs < 284) {
return 8;
}
else if (Rs >= 284) {
return 1;
}
}
}
else if (posBalance >= 53) {
if (negBalance < 47) {
if (posBalance < 72) {
if (Rs < 224) {
return 4;
}
else if (Rs >= 224) {
return 1;
}
}
else if (posBalance >= 72) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (negBalance >= 47) {
if (Cs < 275) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 1;
}
}
else if (Cs >= 275) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 41) {
if (refDCval < 74) {
if (mu_mv_mag_sq < 36) {
if (TSC < 4) {
if (Cs < 52) {
if (SC < 69) {
if (Rs < 8) {
return 4;
}
else if (Rs >= 8) {
return 1;
}
}
else if (SC >= 69) {
if (Rs < 156) {
if (Cs < 4) {
return 8;
}
else if (Cs >= 4) {
if (SC < 78) {
return 1;
}
else if (SC >= 78) {
return 2;
}
}
}
else if (Rs >= 156) {
return 1;
}
}
}
else if (Cs >= 52) {
if (Rs < 203) {
if (posBalance < 71) {
if (AFD < 6) {
if (MVDiff < 14) {
if (Cs < 356) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
return 1;
}
}
else if (Cs >= 356) {
return 4;
}
}
else if (MVDiff >= 14) {
if (tcor < 86) {
if (Rs < 19) {
return 4;
}
else if (Rs >= 19) {
return 1;
}
}
else if (tcor >= 86) {
if (refDCval < 64) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 4;
}
}
else if (refDCval >= 64) {
return 1;
}
}
}
}
else if (AFD >= 6) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
if (AFD < 9) {
if (tcor < 74) {
return 4;
}
else if (tcor >= 74) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
return 4;
}
}
}
}
else if (AFD >= 9) {
return 1;
}
}
}
}
else if (posBalance >= 71) {
return 2;
}
}
else if (Rs >= 203) {
if (gchDC < 1) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 1;
}
}
else if (gchDC >= 1) {
if (AFD < 7) {
if (MVDiff < 5) {
if (Rs < 278) {
return 8;
}
else if (Rs >= 278) {
return 4;
}
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (AFD >= 7) {
if (mu_mv_mag_sq < 17) {
if (ssDCval < 42) {
return 2;
}
else if (ssDCval >= 42) {
return 1;
}
}
else if (mu_mv_mag_sq >= 17) {
if (Rs < 258) {
return 4;
}
else if (Rs >= 258) {
return 1;
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (tcor < 75) {
if (refDCval < 64) {
if (ssDCval < 57) {
if (TSC < 6) {
if (MVDiff < 59) {
if (Cs < 334) {
if (Cs < 73) {
return 1;
}
else if (Cs >= 73) {
return 4;
}
}
else if (Cs >= 334) {
return 1;
}
}
else if (MVDiff >= 59) {
if (Cs < 164) {
return 2;
}
else if (Cs >= 164) {
return 1;
}
}
}
else if (TSC >= 6) {
if (SC < 430) {
if (AFD < 30) {
if (MVDiff < 87) {
return 4;
}
else if (MVDiff >= 87) {
if (AFD < 28) {
return 1;
}
else if (AFD >= 28) {
return 4;
}
}
}
else if (AFD >= 30) {
return 8;
}
}
else if (SC >= 430) {
if (diffRsCsdiff < 5) {
return 8;
}
else if (diffRsCsdiff >= 5) {
if (Rs < 239) {
return 8;
}
else if (Rs >= 239) {
return 4;
}
}
}
}
}
else if (ssDCval >= 57) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 8;
}
}
}
else if (refDCval >= 64) {
if (mcTcor < 76) {
if (MVDiff < 147) {
return 4;
}
else if (MVDiff >= 147) {
if (Rs < 197) {
return 1;
}
else if (Rs >= 197) {
if (Rs < 246) {
return 8;
}
else if (Rs >= 246) {
return 2;
}
}
}
}
else if (mcTcor >= 76) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (diffRsCsdiff < -20) {
return 8;
}
else if (diffRsCsdiff >= -20) {
return 1;
}
}
}
}
}
}
else if (tcor >= 75) {
if (MVDiff < 47) {
if (MVDiff < 13) {
if (Rs < 234) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 2;
}
}
else if (Rs >= 234) {
return 8;
}
}
else if (MVDiff >= 13) {
if (diffAFD < -2) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 1;
}
}
else if (diffAFD >= -2) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
if (tcor < 79) {
if (Cs < 185) {
return 2;
}
else if (Cs >= 185) {
if (Rs < 162) {
return 2;
}
else if (Rs >= 162) {
return 1;
}
}
}
else if (tcor >= 79) {
return 1;
}
}
}
}
}
else if (MVDiff >= 47) {
if (TSC < 5) {
if (SC < 240) {
return 8;
}
else if (SC >= 240) {
return 1;
}
}
else if (TSC >= 5) {
if (Cs < 286) {
if (Cs < 22) {
if (Rs < 26) {
return 8;
}
else if (Rs >= 26) {
return 1;
}
}
else if (Cs >= 22) {
if (MVDiff < 377) {
if (diffAFD < -4) {
if (Rs < 201) {
return 1;
}
else if (Rs >= 201) {
return 2;
}
}
else if (diffAFD >= -4) {
if (negBalance < 43) {
if (refDCval < 53) {
return 1;
}
else if (refDCval >= 53) {
return 2;
}
}
else if (negBalance >= 43) {
return 2;
}
}
}
else if (MVDiff >= 377) {
return 1;
}
}
}
else if (Cs >= 286) {
if (Rs < 222) {
return 8;
}
else if (Rs >= 222) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 36) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (posBalance < 47) {
if (ssDCval < 60) {
if (refDCval < 45) {
if (mu_mv_mag_sq < 113) {
return 1;
}
else if (mu_mv_mag_sq >= 113) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (refDCval >= 45) {
if (Cs < 295) {
if (Cs < 51) {
if (ssDCval < 53) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (AFD < 28) {
return 4;
}
else if (AFD >= 28) {
return 1;
}
}
}
else if (ssDCval >= 53) {
return 2;
}
}
else if (Cs >= 51) {
if (Rs < 145) {
if (Cs < 229) {
return 8;
}
else if (Cs >= 229) {
return 2;
}
}
else if (Rs >= 145) {
if (posBalance < 24) {
return 1;
}
else if (posBalance >= 24) {
if (refDCval < 56) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 261) {
return 1;
}
else if (Rs >= 261) {
return 4;
}
}
}
else if (refDCval >= 56) {
return 8;
}
}
}
}
}
else if (Cs >= 295) {
if (mu_mv_mag_sq < 58) {
if (Rs < 219) {
return 4;
}
else if (Rs >= 219) {
return 2;
}
}
else if (mu_mv_mag_sq >= 58) {
return 1;
}
}
}
}
else if (ssDCval >= 60) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
if (Cs < 345) {
if (Rs < 133) {
return 8;
}
else if (Rs >= 133) {
return 1;
}
}
else if (Cs >= 345) {
if (Rs < 199) {
return 8;
}
else if (Rs >= 199) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 47) {
if (Rs < 205) {
if (diffMVdiffVal < -37) {
if (tcor < 22) {
if (Rs < 119) {
return 2;
}
else if (Rs >= 119) {
return 4;
}
}
else if (tcor >= 22) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
if (Rs < 100) {
return 2;
}
else if (Rs >= 100) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -37) {
if (MVDiff < 68) {
if (negBalance < 27) {
if (refDCval < 53) {
return 8;
}
else if (refDCval >= 53) {
if (Cs < 128) {
if (SC < 220) {
return 1;
}
else if (SC >= 220) {
return 8;
}
}
else if (Cs >= 128) {
return 1;
}
}
}
else if (negBalance >= 27) {
if (mcTcor < 94) {
if (tcor < 89) {
if (MVDiff < 54) {
if (Cs < 135) {
return 8;
}
else if (Cs >= 135) {
return 4;
}
}
else if (MVDiff >= 54) {
return 4;
}
}
else if (tcor >= 89) {
if (Rs < 71) {
return 1;
}
else if (Rs >= 71) {
if (Rs < 169) {
return 8;
}
else if (Rs >= 169) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 191) {
if (mu_mv_mag_sq < 46) {
if (negBalance < 41) {
if (Cs < 257) {
return 4;
}
else if (Cs >= 257) {
return 1;
}
}
else if (negBalance >= 41) {
return 1;
}
}
else if (mu_mv_mag_sq >= 46) {
if (AFD < 3) {
if (Rs < 132) {
return 1;
}
else if (Rs >= 132) {
return 4;
}
}
else if (AFD >= 3) {
if (diffTSC < -1) {
if (Rs < 128) {
return 4;
}
else if (Rs >= 128) {
return 1;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
}
}
else if (Rs >= 191) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 1;
}
}
}
}
}
else if (MVDiff >= 68) {
if (negBalance < 53) {
if (Rs < 43) {
return 4;
}
else if (Rs >= 43) {
if (AFD < 30) {
return 1;
}
else if (AFD >= 30) {
if (Rs < 125) {
return 8;
}
else if (Rs >= 125) {
return 1;
}
}
}
}
else if (negBalance >= 53) {
if (tcor < 27) {
if (Rs < 135) {
return 2;
}
else if (Rs >= 135) {
return 1;
}
}
else if (tcor >= 27) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
return 8;
}
}
}
}
}
}
else if (Rs >= 205) {
if (AFD < 6) {
if (negBalance < 38) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (negBalance >= 38) {
if (Rs < 281) {
return 1;
}
else if (Rs >= 281) {
return 4;
}
}
}
else if (AFD >= 6) {
if (SC < 433) {
if (Rs < 288) {
return 4;
}
else if (Rs >= 288) {
return 8;
}
}
else if (SC >= 433) {
if (Cs < 423) {
if (Rs < 278) {
return 2;
}
else if (Rs >= 278) {
if (Rs < 281) {
return 4;
}
else if (Rs >= 281) {
return 2;
}
}
}
else if (Cs >= 423) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 74) {
if (mcTcor < 72) {
if (AFD < 22) {
if (posBalance < 31) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (posBalance >= 31) {
return 4;
}
}
else if (AFD >= 22) {
if (Cs < 101) {
if (AFD < 39) {
return 1;
}
else if (AFD >= 39) {
return 4;
}
}
else if (Cs >= 101) {
if (AFD < 55) {
if (Rs < 252) {
return 8;
}
else if (Rs >= 252) {
return 1;
}
}
else if (AFD >= 55) {
return 1;
}
}
}
}
else if (mcTcor >= 72) {
if (MVDiff < 21) {
if (Rs < 114) {
if (MVDiff < 14) {
if (Rs < 96) {
if (posBalance < 64) {
return 4;
}
else if (posBalance >= 64) {
return 1;
}
}
else if (Rs >= 96) {
return 1;
}
}
else if (MVDiff >= 14) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
return 2;
}
}
}
else if (Rs >= 114) {
if (diffAFD < 1) {
if (posBalance < 51) {
if (diffRsCsdiff < 1) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
return 8;
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 288) {
return 4;
}
else if (Rs >= 288) {
return 8;
}
}
}
else if (posBalance >= 51) {
if (SC < 359) {
if (TSC < 3) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (TSC >= 3) {
if (TSC < 5) {
if (Rs < 248) {
return 1;
}
else if (Rs >= 248) {
return 8;
}
}
else if (TSC >= 5) {
if (Rs < 214) {
return 1;
}
else if (Rs >= 214) {
return 8;
}
}
}
}
else if (SC >= 359) {
if (negBalance < 47) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (SC < 472) {
return 1;
}
else if (SC >= 472) {
if (negBalance < 31) {
return 2;
}
else if (negBalance >= 31) {
if (Cs < 426) {
if (SC < 494) {
return 2;
}
else if (SC >= 494) {
return 8;
}
}
else if (Cs >= 426) {
return 2;
}
}
}
}
}
else if (negBalance >= 47) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (AFD < 9) {
if (Cs < 361) {
if (MVDiff < 10) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (Cs < 225) {
return 2;
}
else if (Cs >= 225) {
return 8;
}
}
}
}
else if (MVDiff >= 10) {
if (Cs < 131) {
return 8;
}
else if (Cs >= 131) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
return 8;
}
}
}
}
else if (Cs >= 361) {
return 8;
}
}
else if (AFD >= 9) {
if (Cs < 243) {
return 8;
}
else if (Cs >= 243) {
return 1;
}
}
}
}
}
}
}
else if (diffAFD >= 1) {
if (mu_mv_mag_sq < 67) {
if (MVDiff < 13) {
if (Rs < 129) {
return 4;
}
else if (Rs >= 129) {
return 1;
}
}
else if (MVDiff >= 13) {
if (Rs < 181) {
return 1;
}
else if (Rs >= 181) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 268) {
return 8;
}
else if (Rs >= 268) {
return 1;
}
}
}
}
}
else if (MVDiff >= 21) {
if (ssDCval < 129) {
if (refDCval < 116) {
if (mu_mv_mag_sq < 26) {
if (Cs < 151) {
if (Cs < 52) {
return 1;
}
else if (Cs >= 52) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (negBalance < 44) {
if (Rs < 37) {
return 2;
}
else if (Rs >= 37) {
return 1;
}
}
else if (negBalance >= 44) {
if (AFD < 11) {
if (Cs < 59) {
return 2;
}
else if (Cs >= 59) {
if (Rs < 175) {
return 8;
}
else if (Rs >= 175) {
return 2;
}
}
}
else if (AFD >= 11) {
return 2;
}
}
}
}
}
else if (Cs >= 151) {
if (refDCval < 80) {
if (diffAFD < 1) {
if (SC < 395) {
return 1;
}
else if (SC >= 395) {
return 4;
}
}
else if (diffAFD >= 1) {
if (diffRsCsdiff < -3) {
if (Rs < 198) {
return 1;
}
else if (Rs >= 198) {
return 8;
}
}
else if (diffRsCsdiff >= -3) {
return 2;
}
}
}
else if (refDCval >= 80) {
if (ssDCval < 111) {
if (MVDiff < 60) {
if (AFD < 19) {
if (Rs < 283) {
return 2;
}
else if (Rs >= 283) {
return 1;
}
}
else if (AFD >= 19) {
return 1;
}
}
else if (MVDiff >= 60) {
return 2;
}
}
else if (ssDCval >= 111) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (Cs < 98) {
if (tcor < 79) {
if (diffRsCsdiff < -2) {
if (Cs < 67) {
return 8;
}
else if (Cs >= 67) {
return 4;
}
}
else if (diffRsCsdiff >= -2) {
if (diffMVdiffVal < 11) {
if (diffMVdiffVal < -19) {
return 8;
}
else if (diffMVdiffVal >= -19) {
if (mcTcor < 88) {
if (Cs < 20) {
if (Rs < 140) {
if (Rs < 123) {
return 2;
}
else if (Rs >= 123) {
return 8;
}
}
else if (Rs >= 140) {
return 2;
}
}
else if (Cs >= 20) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
if (Rs < 164) {
return 2;
}
else if (Rs >= 164) {
return 8;
}
}
}
}
else if (mcTcor >= 88) {
return 2;
}
}
}
else if (diffMVdiffVal >= 11) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 2;
}
}
}
}
else if (tcor >= 79) {
if (Cs < 55) {
return 1;
}
else if (Cs >= 55) {
if (Rs < 197) {
return 4;
}
else if (Rs >= 197) {
return 8;
}
}
}
}
else if (Cs >= 98) {
if (gchDC < 3) {
if (Rs < 281) {
return 1;
}
else if (Rs >= 281) {
return 8;
}
}
else if (gchDC >= 3) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (Rs < 124) {
if (Cs < 135) {
return 8;
}
else if (Cs >= 135) {
return 1;
}
}
else if (Rs >= 124) {
if (refDCval < 100) {
return 2;
}
else if (refDCval >= 100) {
if (Rs < 177) {
return 1;
}
else if (Rs >= 177) {
if (Rs < 209) {
return 2;
}
else if (Rs >= 209) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 116) {
if (negBalance < 64) {
if (Cs < 321) {
return 2;
}
else if (Cs >= 321) {
if (Rs < 207) {
return 8;
}
else if (Rs >= 207) {
return 2;
}
}
}
else if (negBalance >= 64) {
if (AFD < 10) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
return 2;
}
}
else if (AFD >= 10) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
return 1;
}
}
}
}
}
else if (ssDCval >= 129) {
if (ssDCval < 161) {
if (Rs < 1) {
return 1;
}
else if (Rs >= 1) {
if (refDCval < 131) {
return 4;
}
else if (refDCval >= 131) {
return 2;
}
}
}
else if (ssDCval >= 161) {
if (MVDiff < 64) {
if (Rs < 60) {
return 8;
}
else if (Rs >= 60) {
return 2;
}
}
else if (MVDiff >= 64) {
if (mcTcor < 93) {
return 1;
}
else if (mcTcor >= 93) {
if (Rs < 105) {
return 2;
}
else if (Rs >= 105) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 66) {
if (posBalance < 26) {
if (SC < 6) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
else if (SC >= 6) {
if (refDCval < 39) {
if (posBalance < 12) {
if (mcTcor < 107) {
return 8;
}
else if (mcTcor >= 107) {
return 4;
}
}
else if (posBalance >= 12) {
if (mu_mv_mag_sq < 93) {
return 1;
}
else if (mu_mv_mag_sq >= 93) {
if (Cs < 320) {
return 4;
}
else if (Cs >= 320) {
return 1;
}
}
}
}
else if (refDCval >= 39) {
if (SC < 381) {
if (ssDCval < 62) {
if (posBalance < 2) {
if (Rs < 77) {
return 2;
}
else if (Rs >= 77) {
return 4;
}
}
else if (posBalance >= 2) {
if (mcTcor < 105) {
return 8;
}
else if (mcTcor >= 105) {
if (Rs < 135) {
return 4;
}
else if (Rs >= 135) {
return 8;
}
}
}
}
else if (ssDCval >= 62) {
if (Rs < 94) {
return 4;
}
else if (Rs >= 94) {
return 1;
}
}
}
else if (SC >= 381) {
if (ssDCval < 42) {
return 8;
}
else if (ssDCval >= 42) {
return 4;
}
}
}
}
}
else if (posBalance >= 26) {
if (TSCindex < 2) {
if (MVDiff < 6) {
if (posBalance < 38) {
if (Rs < 145) {
if (Cs < 94) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (Cs >= 94) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 111) {
return 1;
}
else if (Rs >= 111) {
return 4;
}
}
}
}
else if (Rs >= 145) {
if (mu_mv_mag_sq < 127) {
if (mu_mv_mag_sq < 103) {
return 4;
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 147) {
return 4;
}
else if (Rs >= 147) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 4;
}
}
}
else if (posBalance >= 38) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Cs < 188) {
if (Rs < 146) {
if (MVDiff < 2) {
if (Rs < 105) {
return 8;
}
else if (Rs >= 105) {
return 1;
}
}
else if (MVDiff >= 2) {
if (Rs < 109) {
return 1;
}
else if (Rs >= 109) {
return 4;
}
}
}
else if (Rs >= 146) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 151) {
return 4;
}
else if (Rs >= 151) {
return 1;
}
}
}
}
else if (Cs >= 188) {
if (tcor < 99) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (tcor >= 99) {
if (Rs < 254) {
return 4;
}
else if (Rs >= 254) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (Rs < 120) {
if (Cs < 222) {
return 1;
}
else if (Cs >= 222) {
if (Rs < 64) {
return 1;
}
else if (Rs >= 64) {
return 4;
}
}
}
else if (Rs >= 120) {
if (Rs < 156) {
if (Cs < 121) {
return 8;
}
else if (Cs >= 121) {
return 4;
}
}
else if (Rs >= 156) {
if (Rs < 242) {
return 1;
}
else if (Rs >= 242) {
return 2;
}
}
}
}
}
else if (TSCindex >= 2) {
if (CsDiff < 1) {
if (refDCval < 39) {
if (Cs < 191) {
if (TSCindex < 3) {
if (Rs < 89) {
return 4;
}
else if (Rs >= 89) {
return 1;
}
}
else if (TSCindex >= 3) {
if (Rs < 216) {
return 4;
}
else if (Rs >= 216) {
return 1;
}
}
}
else if (Cs >= 191) {
if (Rs < 246) {
return 8;
}
else if (Rs >= 246) {
return 1;
}
}
}
else if (refDCval >= 39) {
if (SC < 351) {
if (tcor < 100) {
if (MVDiff < 1) {
if (mu_mv_mag_sq < 86) {
if (Rs < 113) {
return 1;
}
else if (Rs >= 113) {
return 2;
}
}
else if (mu_mv_mag_sq >= 86) {
return 1;
}
}
else if (MVDiff >= 1) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 60) {
if (Cs < 219) {
if (tcor < 93) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (Cs < 80) {
return 1;
}
else if (Cs >= 80) {
if (Cs < 146) {
return 4;
}
else if (Cs >= 146) {
return 1;
}
}
}
}
else if (tcor >= 93) {
return 1;
}
}
else if (Cs >= 219) {
return 4;
}
}
else if (mu_mv_mag_sq >= 60) {
if (MVDiff < 15) {
if (posBalance < 52) {
if (Rs < 106) {
return 1;
}
else if (Rs >= 106) {
if (refDCval < 57) {
if (Rs < 149) {
return 4;
}
else if (Rs >= 149) {
return 1;
}
}
else if (refDCval >= 57) {
if (tcor < 95) {
if (Rs < 130) {
return 1;
}
else if (Rs >= 130) {
return 4;
}
}
else if (tcor >= 95) {
if (mu_mv_mag_sq < 75) {
return 2;
}
else if (mu_mv_mag_sq >= 75) {
return 1;
}
}
}
}
}
else if (posBalance >= 52) {
return 1;
}
}
else if (MVDiff >= 15) {
if (negBalance < 44) {
return 4;
}
else if (negBalance >= 44) {
if (Rs < 81) {
return 8;
}
else if (Rs >= 81) {
return 2;
}
}
}
}
}
else if (mcTcor >= 99) {
if (tcor < 96) {
if (Cs < 100) {
return 1;
}
else if (Cs >= 100) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Rs < 21) {
return 1;
}
else if (Rs >= 21) {
return 4;
}
}
}
}
else if (tcor >= 96) {
return 1;
}
}
}
}
else if (tcor >= 100) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
return 8;
}
}
}
else if (SC >= 351) {
if (MVDiff < 1) {
if (ssDCval < 63) {
if (posBalance < 46) {
if (mu_mv_mag_sq < 127) {
if (mu_mv_mag_sq < 117) {
if (mu_mv_mag_sq < 115) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 115) {
return 4;
}
}
else if (mu_mv_mag_sq >= 117) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 254) {
return 4;
}
else if (Rs >= 254) {
return 1;
}
}
}
else if (posBalance >= 46) {
return 1;
}
}
else if (ssDCval >= 63) {
return 4;
}
}
else if (MVDiff >= 1) {
if (refDCval < 63) {
if (ssDCval < 57) {
if (ssDCval < 53) {
if (tcor < 102) {
return 1;
}
else if (tcor >= 102) {
if (Rs < 222) {
return 8;
}
else if (Rs >= 222) {
return 1;
}
}
}
else if (ssDCval >= 53) {
if (Cs < 197) {
return 1;
}
else if (Cs >= 197) {
return 4;
}
}
}
else if (ssDCval >= 57) {
if (Cs < 215) {
if (refDCval < 61) {
return 1;
}
else if (refDCval >= 61) {
return 4;
}
}
else if (Cs >= 215) {
if (Rs < 283) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
return 4;
}
}
}
else if (Rs >= 283) {
if (Rs < 290) {
return 1;
}
else if (Rs >= 290) {
return 2;
}
}
}
}
}
else if (refDCval >= 63) {
if (negBalance < 28) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (negBalance >= 28) {
if (tcor < 98) {
if (Rs < 164) {
return 4;
}
else if (Rs >= 164) {
if (Cs < 163) {
return 4;
}
else if (Cs >= 163) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (diffAFD < -1) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
}
else if (tcor >= 98) {
return 1;
}
}
}
}
}
}
}
else if (CsDiff >= 1) {
if (refDCval < 52) {
if (SC < 322) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (MVDiff < 121) {
return 2;
}
else if (MVDiff >= 121) {
return 1;
}
}
}
}
else if (SC >= 322) {
if (mcTcor < 98) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 17) {
if (Cs < 325) {
return 2;
}
else if (Cs >= 325) {
return 1;
}
}
else if (mu_mv_mag_sq >= 17) {
if (MVDiff < 15) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 1;
}
}
else if (MVDiff >= 15) {
return 4;
}
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 19) {
return 4;
}
else if (negBalance >= 19) {
if (Cs < 185) {
if (Rs < 246) {
return 1;
}
else if (Rs >= 246) {
return 8;
}
}
else if (Cs >= 185) {
return 1;
}
}
}
}
}
else if (refDCval >= 52) {
if (TSC < 4) {
if (Rs < 236) {
if (mu_mv_mag_sq < 37) {
if (TSCindex < 3) {
if (ssDCval < 60) {
if (Rs < 182) {
return 4;
}
else if (Rs >= 182) {
return 1;
}
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (TSCindex >= 3) {
if (mcTcor < 99) {
if (ssDCval < 53) {
if (Cs < 429) {
return 4;
}
else if (Cs >= 429) {
return 1;
}
}
else if (ssDCval >= 53) {
return 4;
}
}
else if (mcTcor >= 99) {
if (MVDiff < 24) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 4;
}
}
else if (MVDiff >= 24) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (MVDiff < 1) {
if (Rs < 197) {
return 4;
}
else if (Rs >= 197) {
return 1;
}
}
else if (MVDiff >= 1) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
if (Cs < 274) {
return 4;
}
else if (Cs >= 274) {
return 2;
}
}
}
}
}
else if (Rs >= 236) {
if (MVDiff < 6) {
if (Rs < 249) {
if (Cs < 335) {
return 4;
}
else if (Cs >= 335) {
return 8;
}
}
else if (Rs >= 249) {
if (Rs < 296) {
return 1;
}
else if (Rs >= 296) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (refDCval < 58) {
return 2;
}
else if (refDCval >= 58) {
if (Cs < 468) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (Rs < 278) {
return 1;
}
else if (Rs >= 278) {
return 4;
}
}
}
else if (Cs >= 468) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (Cs < 347) {
if (MVDiff < 85) {
if (Rs < 252) {
return 2;
}
else if (Rs >= 252) {
return 1;
}
}
else if (MVDiff >= 85) {
return 1;
}
}
else if (Cs >= 347) {
if (Rs < 239) {
return 8;
}
else if (Rs >= 239) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 66) {
if (Cs < 238) {
if (Rs < 85) {
if (TSCindex < 2) {
if (ssDCval < 71) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
return 2;
}
}
else if (ssDCval >= 71) {
return 4;
}
}
else if (TSCindex >= 2) {
if (ssDCval < 93) {
if (RsCsDiff < 1) {
if (Rs < 71) {
return 1;
}
else if (Rs >= 71) {
if (ssDCval < 76) {
if (Cs < 129) {
return 4;
}
else if (Cs >= 129) {
return 2;
}
}
else if (ssDCval >= 76) {
return 1;
}
}
}
else if (RsCsDiff >= 1) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (tcor < 89) {
return 1;
}
else if (tcor >= 89) {
return 4;
}
}
}
}
else if (ssDCval >= 93) {
if (AFD < 3) {
if (MVDiff < 27) {
if (Rs < 82) {
return 4;
}
else if (Rs >= 82) {
return 2;
}
}
else if (MVDiff >= 27) {
if (Cs < 98) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
return 4;
}
}
else if (Cs >= 98) {
return 2;
}
}
}
else if (AFD >= 3) {
if (negBalance < 29) {
if (mcTcor < 100) {
return 1;
}
else if (mcTcor >= 100) {
if (Rs < 33) {
return 8;
}
else if (Rs >= 33) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
}
}
else if (negBalance >= 29) {
if (negBalance < 43) {
if (ssDCval < 134) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
return 4;
}
}
else if (ssDCval >= 134) {
return 2;
}
}
else if (negBalance >= 43) {
if (posBalance < 49) {
if (SC < 76) {
return 1;
}
else if (SC >= 76) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (posBalance >= 49) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 85) {
if (mu_mv_mag_sq < 102) {
if (tcor < 101) {
if (Cs < 87) {
if (posBalance < 48) {
if (Cs < 40) {
return 1;
}
else if (Cs >= 40) {
if (Rs < 138) {
return 1;
}
else if (Rs >= 138) {
if (Cs < 76) {
return 8;
}
else if (Cs >= 76) {
return 2;
}
}
}
}
else if (posBalance >= 48) {
if (posBalance < 59) {
if (Cs < 13) {
return 2;
}
else if (Cs >= 13) {
return 1;
}
}
else if (posBalance >= 59) {
if (refDCval < 93) {
return 1;
}
else if (refDCval >= 93) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
return 1;
}
}
}
}
}
else if (Cs >= 87) {
if (AFD < 6) {
if (ssDCval < 82) {
if (AFD < 3) {
if (ssDCval < 75) {
if (TSCindex < 3) {
if (mu_mv_mag_sq < 60) {
return 1;
}
else if (mu_mv_mag_sq >= 60) {
if (posBalance < 28) {
return 1;
}
else if (posBalance >= 28) {
return 2;
}
}
}
else if (TSCindex >= 3) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 165) {
return 2;
}
else if (Rs >= 165) {
return 8;
}
}
}
}
else if (ssDCval >= 75) {
if (Rs < 162) {
if (Cs < 102) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (Cs >= 102) {
return 1;
}
}
else if (Rs >= 162) {
if (Cs < 163) {
return 8;
}
else if (Cs >= 163) {
return 2;
}
}
}
}
else if (AFD >= 3) {
if (ssDCval < 76) {
if (MVDiff < 17) {
if (negBalance < 26) {
return 4;
}
else if (negBalance >= 26) {
return 1;
}
}
else if (MVDiff >= 17) {
if (Rs < 127) {
return 4;
}
else if (Rs >= 127) {
return 8;
}
}
}
else if (ssDCval >= 76) {
if (posBalance < 50) {
if (Rs < 165) {
return 4;
}
else if (Rs >= 165) {
return 1;
}
}
else if (posBalance >= 50) {
if (Cs < 183) {
return 1;
}
else if (Cs >= 183) {
return 2;
}
}
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 106) {
if (Cs < 166) {
if (MVDiff < 7) {
if (Rs < 142) {
return 1;
}
else if (Rs >= 142) {
return 2;
}
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (Cs >= 166) {
if (ssDCval < 93) {
return 1;
}
else if (ssDCval >= 93) {
if (SC < 302) {
return 1;
}
else if (SC >= 302) {
return 4;
}
}
}
}
else if (ssDCval >= 106) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 2;
}
}
}
}
else if (AFD >= 6) {
if (refDCval < 77) {
if (negBalance < 35) {
if (Rs < 119) {
return 1;
}
else if (Rs >= 119) {
return 8;
}
}
else if (negBalance >= 35) {
if (posBalance < 51) {
if (refDCval < 72) {
if (Rs < 218) {
return 8;
}
else if (Rs >= 218) {
return 4;
}
}
else if (refDCval >= 72) {
return 1;
}
}
else if (posBalance >= 51) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (Rs < 112) {
return 4;
}
else if (Rs >= 112) {
return 1;
}
}
}
}
}
else if (refDCval >= 77) {
if (negBalance < 27) {
if (Rs < 250) {
return 4;
}
else if (Rs >= 250) {
if (Rs < 275) {
return 1;
}
else if (Rs >= 275) {
return 2;
}
}
}
else if (negBalance >= 27) {
if (MVDiff < 22) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (refDCval < 94) {
return 1;
}
else if (refDCval >= 94) {
if (AFD < 12) {
if (Cs < 104) {
return 8;
}
else if (Cs >= 104) {
return 2;
}
}
else if (AFD >= 12) {
return 1;
}
}
}
}
else if (MVDiff >= 22) {
if (mu_mv_mag_sq < 7) {
if (Rs < 165) {
return 2;
}
else if (Rs >= 165) {
return 4;
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 195) {
if (MVDiff < 89) {
return 2;
}
else if (MVDiff >= 89) {
return 1;
}
}
else if (Cs >= 195) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
return 2;
}
}
}
}
}
}
}
}
}
else if (tcor >= 101) {
return 8;
}
}
else if (mu_mv_mag_sq >= 102) {
if (Cs < 125) {
if (ssDCval < 86) {
if (posBalance < 42) {
if (Cs < 113) {
return 8;
}
else if (Cs >= 113) {
return 2;
}
}
else if (posBalance >= 42) {
return 8;
}
}
else if (ssDCval >= 86) {
if (Rs < 280) {
return 1;
}
else if (Rs >= 280) {
return 2;
}
}
}
else if (Cs >= 125) {
if (negBalance < 23) {
if (ssDCval < 81) {
if (mcTcor < 98) {
if (Rs < 193) {
return 8;
}
else if (Rs >= 193) {
return 2;
}
}
else if (mcTcor >= 98) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
if (Rs < 233) {
if (mu_mv_mag_sq < 123) {
return 2;
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 180) {
return 2;
}
else if (Rs >= 180) {
return 1;
}
}
}
else if (Rs >= 233) {
return 8;
}
}
}
}
else if (ssDCval >= 81) {
if (Rs < 214) {
if (Rs < 137) {
return 4;
}
else if (Rs >= 137) {
return 2;
}
}
else if (Rs >= 214) {
if (Rs < 267) {
return 8;
}
else if (Rs >= 267) {
return 1;
}
}
}
}
else if (negBalance >= 23) {
if (Rs < 178) {
if (MVDiff < 4) {
if (Cs < 133) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
return 8;
}
}
else if (Cs >= 133) {
return 2;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (Rs >= 178) {
if (MVDiff < 1) {
if (Cs < 218) {
return 2;
}
else if (Cs >= 218) {
if (Rs < 216) {
return 1;
}
else if (Rs >= 216) {
return 2;
}
}
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 108) {
if (Rs < 194) {
return 1;
}
else if (Rs >= 194) {
if (negBalance < 29) {
if (Rs < 236) {
return 2;
}
else if (Rs >= 236) {
return 1;
}
}
else if (negBalance >= 29) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 108) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 238) {
if (ssDCval < 78) {
if (refDCval < 72) {
if (refDCval < 67) {
if (AFD < 14) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (Rs < 279) {
return 1;
}
else if (Rs >= 279) {
return 4;
}
}
}
else if (AFD >= 14) {
return 2;
}
}
else if (refDCval >= 67) {
if (TSC < 2) {
if (Cs < 365) {
if (mu_mv_mag_sq < 103) {
if (Rs < 117) {
return 4;
}
else if (Rs >= 117) {
return 2;
}
}
else if (mu_mv_mag_sq >= 103) {
if (negBalance < 22) {
return 4;
}
else if (negBalance >= 22) {
if (negBalance < 24) {
if (Cs < 263) {
return 1;
}
else if (Cs >= 263) {
if (Cs < 337) {
return 4;
}
else if (Cs >= 337) {
if (Rs < 184) {
if (Rs < 168) {
return 4;
}
else if (Rs >= 168) {
return 1;
}
}
else if (Rs >= 184) {
return 4;
}
}
}
}
else if (negBalance >= 24) {
return 4;
}
}
}
}
else if (Cs >= 365) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Rs < 249) {
return 4;
}
else if (Rs >= 249) {
return 1;
}
}
}
}
else if (TSC >= 2) {
if (mu_mv_mag_sq < 48) {
if (MVDiff < 4) {
if (Rs < 185) {
return 4;
}
else if (Rs >= 185) {
return 1;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (mu_mv_mag_sq >= 48) {
if (Rs < 121) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
return 4;
}
}
else if (Rs >= 121) {
if (mu_mv_mag_sq < 95) {
if (Rs < 177) {
if (ssDCval < 68) {
if (Rs < 164) {
if (Rs < 156) {
return 4;
}
else if (Rs >= 156) {
return 8;
}
}
else if (Rs >= 164) {
return 1;
}
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (Rs >= 177) {
if (Rs < 228) {
if (TSC < 4) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 4;
}
}
else if (TSC >= 4) {
return 2;
}
}
else if (Rs >= 228) {
if (Rs < 279) {
return 1;
}
else if (Rs >= 279) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (Cs < 339) {
if (Rs < 142) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (Rs >= 142) {
if (mu_mv_mag_sq < 102) {
if (Cs < 319) {
return 4;
}
else if (Cs >= 319) {
return 1;
}
}
else if (mu_mv_mag_sq >= 102) {
return 4;
}
}
}
else if (Cs >= 339) {
if (ssDCval < 69) {
if (Rs < 180) {
return 1;
}
else if (Rs >= 180) {
return 4;
}
}
else if (ssDCval >= 69) {
if (Cs < 359) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
else if (Cs >= 359) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 72) {
if (negBalance < 34) {
if (refDCval < 76) {
return 1;
}
else if (refDCval >= 76) {
if (Cs < 370) {
if (Rs < 196) {
return 4;
}
else if (Rs >= 196) {
return 2;
}
}
else if (Cs >= 370) {
return 1;
}
}
}
else if (negBalance >= 34) {
if (ssDCval < 76) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (posBalance < 50) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (posBalance >= 50) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
}
else if (ssDCval >= 76) {
if (TSC < 3) {
if (Cs < 427) {
return 4;
}
else if (Cs >= 427) {
return 1;
}
}
else if (TSC >= 3) {
if (Cs < 279) {
return 2;
}
else if (Cs >= 279) {
if (Rs < 155) {
return 4;
}
else if (Rs >= 155) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 78) {
if (AFD < 4) {
if (ssDCval < 135) {
if (Rs < 198) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
return 1;
}
}
}
else if (Rs >= 198) {
if (SC < 514) {
return 2;
}
else if (SC >= 514) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
}
}
else if (ssDCval >= 135) {
return 2;
}
}
else if (AFD >= 4) {
if (tcor < 79) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (tcor >= 79) {
if (CsDiff < 1) {
if (Rs < 274) {
if (Rs < 248) {
return 2;
}
else if (Rs >= 248) {
return 4;
}
}
else if (Rs >= 274) {
return 8;
}
}
else if (CsDiff >= 1) {
if (ssDCval < 110) {
if (TSC < 3) {
if (Rs < 41) {
return 2;
}
else if (Rs >= 41) {
return 1;
}
}
else if (TSC >= 3) {
if (RsDiff < 2) {
if (diffMVdiffVal < -23) {
if (Rs < 156) {
return 2;
}
else if (Rs >= 156) {
return 1;
}
}
else if (diffMVdiffVal >= -23) {
return 2;
}
}
else if (RsDiff >= 2) {
if (mcTcor < 99) {
if (Cs < 440) {
return 1;
}
else if (Cs >= 440) {
return 8;
}
}
else if (mcTcor >= 99) {
return 2;
}
}
}
}
else if (ssDCval >= 110) {
if (Rs < 289) {
return 2;
}
else if (Rs >= 289) {
if (Rs < 294) {
return 8;
}
else if (Rs >= 294) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 300) {
if (RsDiff < 4) {
if (ssDCval < 75) {
if (mcTcor < 89) {
if (AFD < 8) {
if (negBalance < 33) {
if (posBalance < 61) {
if (refDCval < 59) {
if (posBalance < 46) {
if (mcTcor < 84) {
return 2;
}
else if (mcTcor >= 84) {
return 1;
}
}
else if (posBalance >= 46) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
return 1;
}
}
}
else if (refDCval >= 59) {
if (refDCval < 60) {
return 2;
}
else if (refDCval >= 60) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (Rs < 423) {
return 4;
}
else if (Rs >= 423) {
return 8;
}
}
}
else if (negBalance >= 33) {
if (Cs < 160) {
if (Cs < 55) {
return 8;
}
else if (Cs >= 55) {
return 1;
}
}
else if (Cs >= 160) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (refDCval < 37) {
if (Rs < 752) {
if (tcor < 70) {
if (Rs < 408) {
return 2;
}
else if (Rs >= 408) {
return 8;
}
}
else if (tcor >= 70) {
return 8;
}
}
else if (Rs >= 752) {
return 1;
}
}
else if (refDCval >= 37) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
if (Rs < 531) {
return 4;
}
else if (Rs >= 531) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 8) {
if (Cs < 93) {
return 1;
}
else if (Cs >= 93) {
if (refDCval < 50) {
if (MVDiff < 45) {
if (TSC < 6) {
if (CsDiff < 3) {
if (posBalance < 51) {
if (SC < 659) {
return 2;
}
else if (SC >= 659) {
return 1;
}
}
else if (posBalance >= 51) {
if (negBalance < 51) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Rs < 405) {
if (Rs < 366) {
return 8;
}
else if (Rs >= 366) {
return 2;
}
}
else if (Rs >= 405) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (Rs < 541) {
return 4;
}
else if (Rs >= 541) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
if (diffRsCsdiff < -7) {
return 2;
}
else if (diffRsCsdiff >= -7) {
return 4;
}
}
}
else if (TSC >= 6) {
if (refDCval < 46) {
return 8;
}
else if (refDCval >= 46) {
if (Rs < 361) {
return 4;
}
else if (Rs >= 361) {
return 8;
}
}
}
}
else if (MVDiff >= 45) {
if (AFD < 13) {
if (Cs < 382) {
if (diffRsCsdiff < 3) {
return 2;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 315) {
return 4;
}
else if (Rs >= 315) {
return 1;
}
}
}
else if (Cs >= 382) {
return 4;
}
}
else if (AFD >= 13) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
else if (refDCval >= 50) {
if (MVDiff < 25) {
if (tcor < 81) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (tcor >= 81) {
return 8;
}
}
else if (MVDiff >= 25) {
if (MVDiff < 36) {
if (Cs < 401) {
return 8;
}
else if (Cs >= 401) {
return 1;
}
}
else if (MVDiff >= 36) {
if (negBalance < 47) {
if (Rs < 339) {
return 8;
}
else if (Rs >= 339) {
if (Rs < 594) {
return 1;
}
else if (Rs >= 594) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (MVDiff < 61) {
return 4;
}
else if (MVDiff >= 61) {
if (MVDiff < 106) {
if (Rs < 371) {
if (Cs < 362) {
if (Rs < 318) {
return 2;
}
else if (Rs >= 318) {
return 4;
}
}
else if (Cs >= 362) {
return 8;
}
}
else if (Rs >= 371) {
return 2;
}
}
else if (MVDiff >= 106) {
return 4;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1292) {
if (tcor < 101) {
if (MVDiff < 2) {
if (mcTcor < 97) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
if (mu_mv_mag_sq < 125) {
if (RsDiff < 2) {
if (Cs < 356) {
if (TSC < 3) {
if (Cs < 247) {
if (Rs < 521) {
return 4;
}
else if (Rs >= 521) {
return 2;
}
}
else if (Cs >= 247) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (TSC >= 3) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (Cs >= 356) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (ssDCval < 72) {
if (Cs < 423) {
if (Rs < 344) {
return 2;
}
else if (Rs >= 344) {
return 1;
}
}
else if (Cs >= 423) {
if (ssDCval < 66) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 1;
}
}
else if (ssDCval >= 66) {
return 1;
}
}
}
else if (ssDCval >= 72) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
return 8;
}
}
}
}
}
else if (RsDiff >= 2) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 69) {
if (Rs < 440) {
if (Cs < 312) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
return 4;
}
}
else if (Cs >= 312) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 2;
}
}
}
else if (Rs >= 440) {
if (posBalance < 43) {
if (Rs < 819) {
if (SC < 874) {
if (negBalance < 25) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (negBalance >= 25) {
return 4;
}
}
else if (SC >= 874) {
if (refDCval < 64) {
if (mu_mv_mag_sq < 124) {
if (SC < 1008) {
if (TSC < 3) {
if (diffMVdiffVal < -1) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (TSC >= 3) {
return 4;
}
}
else if (SC >= 1008) {
if (ssDCval < 57) {
if (mu_mv_mag_sq < 115) {
return 1;
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 593) {
return 1;
}
else if (Rs >= 593) {
return 8;
}
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (refDCval >= 64) {
return 1;
}
}
}
else if (Rs >= 819) {
if (Cs < 432) {
return 1;
}
else if (Cs >= 432) {
if (Cs < 514) {
return 8;
}
else if (Cs >= 514) {
return 1;
}
}
}
}
else if (posBalance >= 43) {
if (Rs < 630) {
if (ssDCval < 67) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Cs < 468) {
if (posBalance < 44) {
if (Cs < 423) {
return 8;
}
else if (Cs >= 423) {
return 1;
}
}
else if (posBalance >= 44) {
return 1;
}
}
else if (Cs >= 468) {
if (ssDCval < 59) {
if (Rs < 576) {
return 1;
}
else if (Rs >= 576) {
return 4;
}
}
else if (ssDCval >= 59) {
if (Rs < 589) {
return 8;
}
else if (Rs >= 589) {
return 1;
}
}
}
}
}
else if (ssDCval >= 67) {
if (Cs < 375) {
if (Rs < 562) {
return 8;
}
else if (Rs >= 562) {
return 2;
}
}
else if (Cs >= 375) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
return 2;
}
}
}
}
else if (Rs >= 630) {
if (mu_mv_mag_sq < 99) {
if (Cs < 361) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (Cs >= 361) {
return 8;
}
}
else if (mu_mv_mag_sq >= 99) {
if (Rs < 1042) {
if (Cs < 503) {
if (Cs < 227) {
return 1;
}
else if (Cs >= 227) {
return 2;
}
}
else if (Cs >= 503) {
return 1;
}
}
else if (Rs >= 1042) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 69) {
if (Rs < 755) {
if (SC < 955) {
if (mu_mv_mag_sq < 97) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 97) {
if (posBalance < 44) {
if (Cs < 368) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (Cs >= 368) {
return 1;
}
}
else if (posBalance >= 44) {
if (Rs < 631) {
return 2;
}
else if (Rs >= 631) {
return 4;
}
}
}
}
else if (SC >= 955) {
if (Rs < 609) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
if (SC < 1026) {
if (Rs < 540) {
return 1;
}
else if (Rs >= 540) {
return 4;
}
}
else if (SC >= 1026) {
return 1;
}
}
}
else if (Rs >= 609) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (tcor < 99) {
if (SC < 1081) {
return 4;
}
else if (SC >= 1081) {
return 8;
}
}
else if (tcor >= 99) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 4;
}
}
}
}
}
}
else if (Rs >= 755) {
if (Cs < 276) {
return 1;
}
else if (Cs >= 276) {
if (negBalance < 26) {
return 8;
}
else if (negBalance >= 26) {
if (Rs < 779) {
return 8;
}
else if (Rs >= 779) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 138) {
if (mu_mv_mag_sq < 56) {
if (MVDiff < 13) {
if (Rs < 890) {
if (refDCval < 69) {
if (diffMVdiffVal < 4) {
if (Cs < 131) {
return 1;
}
else if (Cs >= 131) {
if (Cs < 134) {
return 4;
}
else if (Cs >= 134) {
return 1;
}
}
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
else if (refDCval >= 69) {
if (Rs < 581) {
return 4;
}
else if (Rs >= 581) {
return 1;
}
}
}
else if (Rs >= 890) {
if (Cs < 108) {
return 1;
}
else if (Cs >= 108) {
return 2;
}
}
}
else if (MVDiff >= 13) {
if (tcor < 79) {
if (RsDiff < 3) {
if (Rs < 367) {
return 8;
}
else if (Rs >= 367) {
return 1;
}
}
else if (RsDiff >= 3) {
if (Cs < 121) {
return 1;
}
else if (Cs >= 121) {
return 8;
}
}
}
else if (tcor >= 79) {
if (mu_mv_mag_sq < 48) {
if (SC < 336) {
return 8;
}
else if (SC >= 336) {
if (tcor < 83) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (tcor >= 83) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (MVDiff < 17) {
if (Rs < 390) {
return 4;
}
else if (Rs >= 390) {
return 8;
}
}
else if (MVDiff >= 17) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (TSC < 3) {
if (refDCval < 63) {
if (Rs < 307) {
return 2;
}
else if (Rs >= 307) {
return 1;
}
}
else if (refDCval >= 63) {
return 4;
}
}
else if (TSC >= 3) {
if (Cs < 84) {
return 8;
}
else if (Cs >= 84) {
if (Rs < 429) {
return 2;
}
else if (Rs >= 429) {
return 1;
}
}
}
}
}
else if (Cs >= 138) {
if (mu_mv_mag_sq < 87) {
if (Rs < 401) {
if (negBalance < 34) {
if (Rs < 317) {
if (ssDCval < 58) {
if (MVDiff < 26) {
return 1;
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (ssDCval >= 58) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (Rs < 302) {
return 1;
}
else if (Rs >= 302) {
return 4;
}
}
}
}
else if (Rs >= 317) {
if (AFD < 4) {
if (ssDCval < 35) {
return 1;
}
else if (ssDCval >= 35) {
if (mu_mv_mag_sq < 84) {
return 4;
}
else if (mu_mv_mag_sq >= 84) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
}
}
else if (AFD >= 4) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (posBalance < 75) {
if (ssDCval < 68) {
if (tcor < 96) {
if (refDCval < 44) {
if (posBalance < 54) {
if (Rs < 382) {
return 4;
}
else if (Rs >= 382) {
return 1;
}
}
else if (posBalance >= 54) {
return 1;
}
}
else if (refDCval >= 44) {
if (RsDiff < 2) {
if (diffRsCsdiff < -4) {
return 1;
}
else if (diffRsCsdiff >= -4) {
return 4;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (tcor >= 96) {
if (Rs < 343) {
return 4;
}
else if (Rs >= 343) {
return 1;
}
}
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (posBalance >= 75) {
return 8;
}
}
}
}
}
else if (negBalance >= 34) {
if (mcTcor < 96) {
if (tcor < 83) {
if (negBalance < 38) {
if (CsDiff < 2) {
if (Cs < 274) {
return 2;
}
else if (Cs >= 274) {
return 1;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (negBalance >= 38) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (ssDCval < 44) {
return 1;
}
else if (ssDCval >= 44) {
if (MVDiff < 45) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
else if (MVDiff >= 45) {
return 1;
}
}
}
}
}
else if (tcor >= 83) {
if (ssDCval < 57) {
if (MVDiff < 43) {
if (Cs < 450) {
if (ssDCval < 48) {
if (ssDCval < 43) {
return 1;
}
else if (ssDCval >= 43) {
return 4;
}
}
else if (ssDCval >= 48) {
return 1;
}
}
else if (Cs >= 450) {
return 8;
}
}
else if (MVDiff >= 43) {
return 2;
}
}
else if (ssDCval >= 57) {
if (Cs < 231) {
return 4;
}
else if (Cs >= 231) {
return 2;
}
}
}
}
else if (mcTcor >= 96) {
if (gchDC < 2) {
if (ssDCval < 54) {
if (negBalance < 43) {
if (negBalance < 35) {
if (Rs < 359) {
return 1;
}
else if (Rs >= 359) {
return 4;
}
}
else if (negBalance >= 35) {
return 1;
}
}
else if (negBalance >= 43) {
if (Rs < 302) {
return 4;
}
else if (Rs >= 302) {
return 2;
}
}
}
else if (ssDCval >= 54) {
if (Rs < 320) {
if (MVDiff < 7) {
if (Cs < 303) {
if (Cs < 187) {
return 4;
}
else if (Cs >= 187) {
return 1;
}
}
else if (Cs >= 303) {
return 4;
}
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (Rs >= 320) {
if (TSC < 4) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (TSC >= 4) {
if (Rs < 356) {
return 1;
}
else if (Rs >= 356) {
return 2;
}
}
}
}
}
else if (gchDC >= 2) {
if (Cs < 421) {
if (Rs < 325) {
return 2;
}
else if (Rs >= 325) {
if (Rs < 339) {
return 4;
}
else if (Rs >= 339) {
return 8;
}
}
}
else if (Cs >= 421) {
return 1;
}
}
}
}
}
else if (Rs >= 401) {
if (refDCval < 61) {
if (Cs < 421) {
if (posBalance < 34) {
if (refDCval < 56) {
if (RsCsDiff < 4) {
if (Cs < 274) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (Cs >= 274) {
return 2;
}
}
else if (RsCsDiff >= 4) {
return 2;
}
}
else if (refDCval >= 56) {
if (Rs < 671) {
return 8;
}
else if (Rs >= 671) {
return 1;
}
}
}
else if (posBalance >= 34) {
if (ssDCval < 57) {
if (Rs < 566) {
if (ssDCval < 42) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (ssDCval >= 42) {
if (Cs < 227) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (mu_mv_mag_sq < 58) {
return 1;
}
else if (mu_mv_mag_sq >= 58) {
return 4;
}
}
}
else if (Cs >= 227) {
if (mcTcor < 99) {
if (tcor < 79) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Cs < 360) {
return 2;
}
else if (Cs >= 360) {
return 1;
}
}
}
else if (tcor >= 79) {
if (negBalance < 38) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (Cs < 362) {
if (AFD < 4) {
if (Cs < 238) {
return 1;
}
else if (Cs >= 238) {
return 4;
}
}
else if (AFD >= 4) {
if (Cs < 267) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 1;
}
}
else if (Cs >= 267) {
return 1;
}
}
}
else if (Cs >= 362) {
return 4;
}
}
}
else if (negBalance >= 38) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
else if (mcTcor >= 99) {
if (Cs < 367) {
if (Rs < 488) {
return 4;
}
else if (Rs >= 488) {
return 1;
}
}
else if (Cs >= 367) {
return 8;
}
}
}
}
}
else if (Rs >= 566) {
if (posBalance < 57) {
if (mcTcor < 99) {
if (refDCval < 54) {
return 1;
}
else if (refDCval >= 54) {
if (Rs < 740) {
if (Rs < 617) {
return 1;
}
else if (Rs >= 617) {
if (Cs < 231) {
return 2;
}
else if (Cs >= 231) {
return 8;
}
}
}
else if (Rs >= 740) {
return 1;
}
}
}
else if (mcTcor >= 99) {
if (SC < 1071) {
if (Rs < 643) {
return 1;
}
else if (Rs >= 643) {
return 4;
}
}
else if (SC >= 1071) {
return 2;
}
}
}
else if (posBalance >= 57) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
return 1;
}
}
}
}
else if (ssDCval >= 57) {
if (negBalance < 29) {
if (negBalance < 19) {
return 4;
}
else if (negBalance >= 19) {
return 1;
}
}
else if (negBalance >= 29) {
if (diffMVdiffVal < -5) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 4;
}
}
else if (diffMVdiffVal >= -5) {
if (tcor < 93) {
if (diffRsCsdiff < -10) {
if (SC < 1366) {
return 1;
}
else if (SC >= 1366) {
return 8;
}
}
else if (diffRsCsdiff >= -10) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
}
else if (tcor >= 93) {
if (Cs < 391) {
return 1;
}
else if (Cs >= 391) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 421) {
if (MVDiff < 10) {
if (negBalance < 43) {
if (SC < 972) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 2;
}
}
else if (SC >= 972) {
if (negBalance < 26) {
if (Rs < 687) {
return 1;
}
else if (Rs >= 687) {
return 2;
}
}
else if (negBalance >= 26) {
if (posBalance < 37) {
if (Rs < 707) {
return 8;
}
else if (Rs >= 707) {
return 2;
}
}
else if (posBalance >= 37) {
if (Rs < 507) {
if (SC < 982) {
return 1;
}
else if (SC >= 982) {
return 2;
}
}
else if (Rs >= 507) {
if (Cs < 499) {
if (posBalance < 42) {
if (Rs < 681) {
return 4;
}
else if (Rs >= 681) {
return 2;
}
}
else if (posBalance >= 42) {
if (Rs < 529) {
return 2;
}
else if (Rs >= 529) {
return 4;
}
}
}
else if (Cs >= 499) {
if (mcTcor < 94) {
return 2;
}
else if (mcTcor >= 94) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 43) {
if (tcor < 83) {
if (Rs < 558) {
return 8;
}
else if (Rs >= 558) {
return 1;
}
}
else if (tcor >= 83) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 2;
}
}
}
}
else if (MVDiff >= 10) {
if (negBalance < 32) {
if (ssDCval < 39) {
return 2;
}
else if (ssDCval >= 39) {
if (diffAFD < -1) {
if (negBalance < 27) {
if (Rs < 619) {
return 4;
}
else if (Rs >= 619) {
return 2;
}
}
else if (negBalance >= 27) {
if (SC < 883) {
return 2;
}
else if (SC >= 883) {
return 1;
}
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (negBalance >= 32) {
if (Rs < 505) {
if (MVDiff < 46) {
if (diffMVdiffVal < -1132) {
return 2;
}
else if (diffMVdiffVal >= -1132) {
return 1;
}
}
else if (MVDiff >= 46) {
return 2;
}
}
else if (Rs >= 505) {
if (ssDCval < 55) {
if (tcor < 80) {
if (diffMVdiffVal < 1) {
if (Rs < 532) {
return 8;
}
else if (Rs >= 532) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
else if (tcor >= 80) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 4;
}
}
}
else if (ssDCval >= 55) {
if (Rs < 723) {
return 1;
}
else if (Rs >= 723) {
if (Cs < 516) {
return 4;
}
else if (Cs >= 516) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 61) {
if (mu_mv_mag_sq < 36) {
if (MVDiff < 23) {
if (diffMVdiffVal < -8) {
return 1;
}
else if (diffMVdiffVal >= -8) {
if (mu_mv_mag_sq < 3) {
if (tcor < 91) {
if (posBalance < 68) {
return 4;
}
else if (posBalance >= 68) {
return 2;
}
}
else if (tcor >= 91) {
if (negBalance < 39) {
if (refDCval < 72) {
return 4;
}
else if (refDCval >= 72) {
if (Rs < 524) {
return 1;
}
else if (Rs >= 524) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 537) {
return 4;
}
else if (Rs >= 537) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
}
else if (MVDiff >= 23) {
if (ssDCval < 69) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (SC < 709) {
return 4;
}
else if (SC >= 709) {
if (mcTcor < 95) {
return 2;
}
else if (mcTcor >= 95) {
if (SC < 923) {
return 2;
}
else if (SC >= 923) {
return 1;
}
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 575) {
return 8;
}
else if (Rs >= 575) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 36) {
if (SC < 825) {
if (MVDiff < 14) {
if (MVDiff < 5) {
if (Cs < 320) {
return 1;
}
else if (Cs >= 320) {
return 4;
}
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (MVDiff >= 14) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (SC >= 825) {
if (negBalance < 22) {
if (Cs < 289) {
return 4;
}
else if (Cs >= 289) {
if (Rs < 1029) {
return 1;
}
else if (Rs >= 1029) {
return 2;
}
}
}
else if (negBalance >= 22) {
if (mcTcor < 97) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (tcor < 89) {
if (Rs < 679) {
return 2;
}
else if (Rs >= 679) {
return 4;
}
}
else if (tcor >= 89) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 532) {
return 2;
}
else if (Rs >= 532) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (diffRsCsdiff < 3) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
if (Rs < 573) {
return 1;
}
else if (Rs >= 573) {
return 4;
}
}
}
else if (diffRsCsdiff >= 3) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (refDCval < 71) {
if (refDCval < 68) {
if (TSC < 1) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (TSC >= 1) {
if (refDCval < 63) {
if (negBalance < 28) {
if (ssDCval < 56) {
if (Rs < 421) {
return 1;
}
else if (Rs >= 421) {
if (Rs < 424) {
return 8;
}
else if (Rs >= 424) {
if (mu_mv_mag_sq < 96) {
if (RsDiff < 1) {
if (Rs < 522) {
return 4;
}
else if (Rs >= 522) {
return 1;
}
}
else if (RsDiff >= 1) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (Cs < 491) {
return 4;
}
else if (Cs >= 491) {
if (Rs < 640) {
return 1;
}
else if (Rs >= 640) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (diffAFD < -1) {
if (Cs < 426) {
return 1;
}
else if (Cs >= 426) {
return 4;
}
}
else if (diffAFD >= -1) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 56) {
if (SC < 971) {
if (posBalance < 50) {
if (Cs < 473) {
return 8;
}
else if (Cs >= 473) {
return 4;
}
}
else if (posBalance >= 50) {
if (Rs < 390) {
return 2;
}
else if (Rs >= 390) {
return 1;
}
}
}
else if (SC >= 971) {
if (Rs < 463) {
return 2;
}
else if (Rs >= 463) {
if (mcTcor < 99) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
}
}
else if (negBalance >= 28) {
if (mcTcor < 97) {
if (diffRsCsdiff < -6) {
return 2;
}
else if (diffRsCsdiff >= -6) {
if (tcor < 88) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Rs < 696) {
return 4;
}
else if (Rs >= 696) {
return 8;
}
}
}
else if (tcor >= 88) {
if (ssDCval < 63) {
if (MVDiff < 3) {
if (Rs < 364) {
return 8;
}
else if (Rs >= 364) {
return 1;
}
}
else if (MVDiff >= 3) {
if (diffMVdiffVal < -5) {
if (Rs < 680) {
return 1;
}
else if (Rs >= 680) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
}
else if (ssDCval >= 63) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 54) {
if (negBalance < 32) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
else if (negBalance >= 32) {
return 8;
}
}
else if (refDCval >= 54) {
if (Rs < 345) {
return 2;
}
else if (Rs >= 345) {
return 1;
}
}
}
}
}
else if (refDCval >= 63) {
if (Rs < 389) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 320) {
return 1;
}
else if (Rs >= 320) {
return 4;
}
}
}
else if (Rs >= 389) {
if (TSC < 3) {
if (SC < 1142) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (posBalance < 33) {
if (Rs < 560) {
return 1;
}
else if (Rs >= 560) {
return 8;
}
}
else if (posBalance >= 33) {
return 8;
}
}
}
else if (SC >= 1142) {
return 1;
}
}
else if (TSC >= 3) {
return 4;
}
}
}
}
}
else if (refDCval >= 68) {
if (Rs < 533) {
if (Rs < 480) {
if (Rs < 423) {
return 4;
}
else if (Rs >= 423) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 2;
}
}
}
else if (Rs >= 480) {
return 1;
}
}
else if (Rs >= 533) {
if (mu_mv_mag_sq < 98) {
if (Rs < 875) {
return 1;
}
else if (Rs >= 875) {
return 8;
}
}
else if (mu_mv_mag_sq >= 98) {
return 2;
}
}
}
}
else if (refDCval >= 71) {
if (negBalance < 25) {
if (MVDiff < 3) {
if (Cs < 376) {
return 1;
}
else if (Cs >= 376) {
return 4;
}
}
else if (MVDiff >= 3) {
if (Rs < 800) {
return 8;
}
else if (Rs >= 800) {
return 1;
}
}
}
else if (negBalance >= 25) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 101) {
if (mu_mv_mag_sq < 100) {
if (posBalance < 21) {
if (posBalance < 2) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
else if (posBalance >= 2) {
if (Cs < 471) {
if (mcTcor < 104) {
if (Cs < 5) {
return 4;
}
else if (Cs >= 5) {
return 8;
}
}
else if (mcTcor >= 104) {
if (Cs < 296) {
return 8;
}
else if (Cs >= 296) {
return 4;
}
}
}
else if (Cs >= 471) {
return 4;
}
}
}
else if (posBalance >= 21) {
if (mu_mv_mag_sq < 9) {
return 4;
}
else if (mu_mv_mag_sq >= 9) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (mu_mv_mag_sq < 68) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 68) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
return 8;
}
}
}
else if (Rs >= 1292) {
if (SC < 1588) {
if (mu_mv_mag_sq < 65) {
if (Cs < 142) {
return 2;
}
else if (Cs >= 142) {
return 1;
}
}
else if (mu_mv_mag_sq >= 65) {
return 8;
}
}
else if (SC >= 1588) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 63) {
if (mu_mv_mag_sq < 67) {
if (negBalance < 29) {
if (posBalance < 37) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (posBalance >= 37) {
if (SC < 1942) {
if (mcTcor < 97) {
return 2;
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (SC >= 1942) {
return 2;
}
}
}
else if (negBalance >= 29) {
if (Rs < 2204) {
return 1;
}
else if (Rs >= 2204) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 67) {
if (TSCindex < 2) {
if (ssDCval < 64) {
if (ssDCval < 59) {
if (negBalance < 16) {
return 1;
}
else if (negBalance >= 16) {
return 8;
}
}
else if (ssDCval >= 59) {
return 2;
}
}
else if (ssDCval >= 64) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (posBalance < 34) {
return 1;
}
else if (posBalance >= 34) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
}
else if (TSCindex >= 2) {
if (Rs < 1799) {
if (mu_mv_mag_sq < 121) {
if (Cs < 463) {
if (tcor < 98) {
if (negBalance < 19) {
return 8;
}
else if (negBalance >= 19) {
if (Rs < 1636) {
return 1;
}
else if (Rs >= 1636) {
if (Rs < 1727) {
return 8;
}
else if (Rs >= 1727) {
return 1;
}
}
}
}
else if (tcor >= 98) {
if (refDCval < 73) {
if (MVDiff < 5) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (refDCval >= 73) {
if (mu_mv_mag_sq < 75) {
return 8;
}
else if (mu_mv_mag_sq >= 75) {
return 1;
}
}
}
}
else if (Cs >= 463) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (negBalance < 18) {
return 2;
}
else if (negBalance >= 18) {
return 8;
}
}
}
else if (Rs >= 1799) {
if (Cs < 399) {
return 2;
}
else if (Cs >= 399) {
return 1;
}
}
}
}
}
else if (posBalance >= 63) {
return 2;
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (Rs < 1395) {
return 8;
}
else if (Rs >= 1395) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (Cs < 141) {
if (refDCval < 107) {
if (posBalance < 11) {
return 8;
}
else if (posBalance >= 11) {
if (Rs < 1129) {
if (refDCval < 96) {
if (RsDiff < 2) {
if (Cs < 77) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 77) {
return 8;
}
}
else if (RsDiff >= 2) {
if (tcor < 65) {
if (Cs < 88) {
if (Rs < 325) {
return 8;
}
else if (Rs >= 325) {
return 2;
}
}
else if (Cs >= 88) {
return 8;
}
}
else if (tcor >= 65) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (ssDCval < 81) {
if (Rs < 407) {
return 2;
}
else if (Rs >= 407) {
return 1;
}
}
else if (ssDCval >= 81) {
return 8;
}
}
}
}
}
else if (refDCval >= 96) {
return 1;
}
}
else if (Rs >= 1129) {
return 1;
}
}
}
else if (refDCval >= 107) {
if (MVDiff < 20) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (MVDiff >= 20) {
if (AFD < 18) {
if (Cs < 34) {
if (Rs < 1503) {
return 8;
}
else if (Rs >= 1503) {
return 1;
}
}
else if (Cs >= 34) {
if (Rs < 334) {
return 8;
}
else if (Rs >= 334) {
return 1;
}
}
}
else if (AFD >= 18) {
return 2;
}
}
}
}
else if (Cs >= 141) {
if (SC < 1916) {
if (SC < 895) {
if (mcTcor < 96) {
if (MVDiff < 24) {
if (TSC < 7) {
if (mu_mv_mag_sq < 21) {
if (MVDiff < 5) {
if (posBalance < 57) {
if (Rs < 548) {
return 8;
}
else if (Rs >= 548) {
return 4;
}
}
else if (posBalance >= 57) {
return 1;
}
}
else if (MVDiff >= 5) {
if (TSC < 6) {
if (AFD < 12) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
return 2;
}
}
else if (AFD >= 12) {
return 1;
}
}
else if (TSC >= 6) {
if (Rs < 352) {
if (diffMVdiffVal < -19) {
return 1;
}
else if (diffMVdiffVal >= -19) {
return 8;
}
}
else if (Rs >= 352) {
if (Cs < 357) {
return 2;
}
else if (Cs >= 357) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (Cs < 180) {
return 1;
}
else if (Cs >= 180) {
if (Rs < 353) {
if (mu_mv_mag_sq < 69) {
return 8;
}
else if (mu_mv_mag_sq >= 69) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 2;
}
}
}
else if (Rs >= 353) {
if (Cs < 389) {
if (ssDCval < 118) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
return 8;
}
}
else if (ssDCval >= 118) {
return 8;
}
}
else if (Cs >= 389) {
return 8;
}
}
}
}
}
else if (TSC >= 7) {
if (Rs < 485) {
return 8;
}
else if (Rs >= 485) {
return 4;
}
}
}
else if (MVDiff >= 24) {
if (tcor < 73) {
if (ssDCval < 81) {
if (AFD < 23) {
return 4;
}
else if (AFD >= 23) {
return 8;
}
}
else if (ssDCval >= 81) {
if (AFD < 29) {
return 2;
}
else if (AFD >= 29) {
return 4;
}
}
}
else if (tcor >= 73) {
if (ssDCval < 123) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (negBalance < 55) {
if (Cs < 391) {
return 1;
}
else if (Cs >= 391) {
return 8;
}
}
else if (negBalance >= 55) {
if (Cs < 454) {
return 8;
}
else if (Cs >= 454) {
return 1;
}
}
}
}
}
else if (ssDCval >= 123) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 117) {
if (Cs < 182) {
if (ssDCval < 84) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
if (Rs < 663) {
return 1;
}
else if (Rs >= 663) {
return 4;
}
}
}
else if (ssDCval >= 84) {
if (tcor < 89) {
if (Rs < 388) {
return 4;
}
else if (Rs >= 388) {
return 8;
}
}
else if (tcor >= 89) {
return 1;
}
}
}
else if (Cs >= 182) {
if (tcor < 100) {
if (Rs < 389) {
if (mcTcor < 98) {
if (MVDiff < 20) {
if (RsDiff < 2) {
if (ssDCval < 134) {
if (TSC < 4) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 2;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (ssDCval >= 134) {
return 8;
}
}
else if (RsDiff >= 2) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (SC < 758) {
if (Rs < 379) {
return 1;
}
else if (Rs >= 379) {
return 8;
}
}
else if (SC >= 758) {
return 8;
}
}
}
}
else if (MVDiff >= 20) {
if (Cs < 465) {
return 2;
}
else if (Cs >= 465) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (ssDCval < 117) {
if (Cs < 309) {
if (mu_mv_mag_sq < 65) {
if (Rs < 320) {
return 1;
}
else if (Rs >= 320) {
return 4;
}
}
else if (mu_mv_mag_sq >= 65) {
return 2;
}
}
else if (Cs >= 309) {
if (Rs < 373) {
return 1;
}
else if (Rs >= 373) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (tcor < 97) {
if (diffMVdiffVal < 41) {
return 2;
}
else if (diffMVdiffVal >= 41) {
return 4;
}
}
else if (tcor >= 97) {
if (Rs < 378) {
if (ssDCval < 123) {
if (Rs < 342) {
return 4;
}
else if (Rs >= 342) {
return 2;
}
}
else if (ssDCval >= 123) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (Rs >= 378) {
return 1;
}
}
}
}
}
else if (Rs >= 389) {
if (Cs < 311) {
if (AFD < 4) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Rs < 491) {
return 8;
}
else if (Rs >= 491) {
return 2;
}
}
}
else if (AFD >= 4) {
if (RsCsDiff < 5) {
return 4;
}
else if (RsCsDiff >= 5) {
if (gchDC < 2) {
if (Rs < 447) {
return 4;
}
else if (Rs >= 447) {
if (SC < 709) {
return 2;
}
else if (SC >= 709) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
else if (gchDC >= 2) {
if (CsDiff < 2) {
if (Rs < 588) {
return 8;
}
else if (Rs >= 588) {
return 1;
}
}
else if (CsDiff >= 2) {
if (Cs < 279) {
return 2;
}
else if (Cs >= 279) {
return 1;
}
}
}
}
}
}
else if (Cs >= 311) {
if (MVDiff < 4) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
return 1;
}
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 66) {
if (posBalance < 56) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 5) {
return 1;
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 398) {
return 8;
}
else if (Rs >= 398) {
return 2;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 494) {
return 1;
}
else if (Rs >= 494) {
return 4;
}
}
}
else if (posBalance >= 56) {
if (negBalance < 35) {
return 2;
}
else if (negBalance >= 35) {
if (Cs < 328) {
return 1;
}
else if (Cs >= 328) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (Cs < 394) {
return 4;
}
else if (Cs >= 394) {
if (Cs < 447) {
return 2;
}
else if (Cs >= 447) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 100) {
if (Cs < 405) {
return 8;
}
else if (Cs >= 405) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (Cs < 354) {
if (SC < 509) {
if (Cs < 189) {
return 2;
}
else if (Cs >= 189) {
return 1;
}
}
else if (SC >= 509) {
if (Cs < 209) {
if (Rs < 657) {
return 2;
}
else if (Rs >= 657) {
return 1;
}
}
else if (Cs >= 209) {
if (Rs < 557) {
if (Rs < 394) {
return 2;
}
else if (Rs >= 394) {
return 1;
}
}
else if (Rs >= 557) {
return 2;
}
}
}
}
else if (Cs >= 354) {
return 8;
}
}
}
}
else if (SC >= 895) {
if (CsDiff < 2) {
if (negBalance < 32) {
if (refDCval < 94) {
if (mu_mv_mag_sq < 121) {
if (refDCval < 82) {
if (SC < 1247) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Rs < 682) {
if (Rs < 496) {
return 8;
}
else if (Rs >= 496) {
return 1;
}
}
else if (Rs >= 682) {
if (Rs < 719) {
return 4;
}
else if (Rs >= 719) {
return 8;
}
}
}
}
else if (SC >= 1247) {
if (Cs < 331) {
return 1;
}
else if (Cs >= 331) {
return 8;
}
}
}
else if (refDCval >= 82) {
if (MVDiff < 19) {
if (Rs < 454) {
return 8;
}
else if (Rs >= 454) {
return 1;
}
}
else if (MVDiff >= 19) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (Rs < 1001) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
if (Cs < 358) {
return 1;
}
else if (Cs >= 358) {
return 8;
}
}
}
else if (Rs >= 1001) {
return 2;
}
}
}
else if (refDCval >= 94) {
if (MVDiff < 3) {
if (Cs < 466) {
if (refDCval < 98) {
if (Rs < 574) {
return 8;
}
else if (Rs >= 574) {
return 4;
}
}
else if (refDCval >= 98) {
return 8;
}
}
else if (Cs >= 466) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Cs < 507) {
if (TSC < 3) {
if (mcTcor < 99) {
if (negBalance < 20) {
return 1;
}
else if (negBalance >= 20) {
if (Rs < 957) {
return 2;
}
else if (Rs >= 957) {
return 8;
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (TSC >= 3) {
if (Cs < 495) {
return 8;
}
else if (Cs >= 495) {
return 1;
}
}
}
else if (Cs >= 507) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
if (Rs < 609) {
return 2;
}
else if (Rs >= 609) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 32) {
if (CsDiff < 1) {
if (Rs < 1679) {
if (Rs < 560) {
if (Cs < 483) {
return 1;
}
else if (Cs >= 483) {
if (Rs < 501) {
return 8;
}
else if (Rs >= 501) {
return 2;
}
}
}
else if (Rs >= 560) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 751) {
return 8;
}
else if (Rs >= 751) {
return 1;
}
}
}
}
else if (Rs >= 1679) {
return 1;
}
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 101) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (Cs < 485) {
if (Cs < 403) {
if (refDCval < 96) {
if (mu_mv_mag_sq < 47) {
if (SC < 961) {
if (SC < 936) {
return 1;
}
else if (SC >= 936) {
return 4;
}
}
else if (SC >= 961) {
return 1;
}
}
else if (mu_mv_mag_sq >= 47) {
if (Rs < 557) {
return 1;
}
else if (Rs >= 557) {
return 8;
}
}
}
else if (refDCval >= 96) {
if (TSC < 5) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
return 8;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (Cs >= 403) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (Rs < 1012) {
return 1;
}
else if (Rs >= 1012) {
if (Cs < 428) {
return 8;
}
else if (Cs >= 428) {
return 1;
}
}
}
}
}
else if (Cs >= 485) {
if (mu_mv_mag_sq < 32) {
if (Rs < 463) {
return 8;
}
else if (Rs >= 463) {
if (tcor < 87) {
return 2;
}
else if (tcor >= 87) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (MVDiff < 11) {
if (ssDCval < 118) {
if (Rs < 650) {
return 2;
}
else if (Rs >= 650) {
return 1;
}
}
else if (ssDCval >= 118) {
return 8;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 614) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
return 8;
}
}
else if (Rs >= 614) {
return 8;
}
}
}
}
}
else if (CsDiff >= 2) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 30) {
if (Rs < 647) {
if (MVDiff < 6) {
if (AFD < 9) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
return 1;
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (MVDiff >= 6) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (posBalance < 64) {
if (diffAFD < -1) {
if (Cs < 499) {
return 2;
}
else if (Cs >= 499) {
return 8;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (posBalance >= 64) {
return 8;
}
}
}
}
else if (Rs >= 647) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (Rs < 907) {
return 4;
}
else if (Rs >= 907) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 30) {
return 8;
}
}
else if (mcTcor >= 95) {
if (TSC < 6) {
if (diffRsCsdiff < -11) {
if (Rs < 660) {
return 8;
}
else if (Rs >= 660) {
return 4;
}
}
else if (diffRsCsdiff >= -11) {
if (negBalance < 60) {
if (AFD < 6) {
if (MVDiff < 4) {
if (Rs < 998) {
return 1;
}
else if (Rs >= 998) {
return 8;
}
}
else if (MVDiff >= 4) {
if (Cs < 525) {
return 2;
}
else if (Cs >= 525) {
return 8;
}
}
}
else if (AFD >= 6) {
if (mu_mv_mag_sq < 55) {
if (Cs < 490) {
return 1;
}
else if (Cs >= 490) {
if (diffMVdiffVal < -5) {
return 2;
}
else if (diffMVdiffVal >= -5) {
if (MVDiff < 2) {
if (Rs < 491) {
return 1;
}
else if (Rs >= 491) {
return 2;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 774) {
return 1;
}
else if (Rs >= 774) {
return 8;
}
}
}
}
}
else if (negBalance >= 60) {
if (Cs < 420) {
if (Rs < 830) {
return 4;
}
else if (Rs >= 830) {
return 1;
}
}
else if (Cs >= 420) {
if (SC < 1003) {
return 1;
}
else if (SC >= 1003) {
return 2;
}
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
if (diffRsCsdiff < 6) {
if (diffMVdiffVal < -4) {
if (Rs < 514) {
return 8;
}
else if (Rs >= 514) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (diffRsCsdiff >= 6) {
return 2;
}
}
}
}
}
}
}
else if (SC >= 1916) {
if (Cs < 290) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
return 8;
}
}
else if (Cs >= 290) {
if (Rs < 2046) {
if (mu_mv_mag_sq < 86) {
if (tcor < 98) {
if (ssDCval < 98) {
if (AFD < 5) {
if (Cs < 365) {
return 1;
}
else if (Cs >= 365) {
return 8;
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (ssDCval >= 98) {
if (Rs < 1611) {
return 2;
}
else if (Rs >= 1611) {
return 8;
}
}
}
else if (tcor >= 98) {
if (mu_mv_mag_sq < 80) {
return 8;
}
else if (mu_mv_mag_sq >= 80) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (AFD < 3) {
if (negBalance < 16) {
return 8;
}
else if (negBalance >= 16) {
return 1;
}
}
else if (AFD >= 3) {
if (Rs < 1708) {
if (Cs < 440) {
return 4;
}
else if (Cs >= 440) {
return 1;
}
}
else if (Rs >= 1708) {
if (MVDiff < 2) {
if (Rs < 1924) {
return 4;
}
else if (Rs >= 1924) {
return 1;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
}
else if (Rs >= 2046) {
if (posBalance < 53) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (Cs < 395) {
if (RsCsDiff < 56) {
if (posBalance < 58) {
if (CsDiff < 3) {
if (MVDiff < 13) {
if (Rs < 583) {
return 1;
}
else if (Rs >= 583) {
return 2;
}
}
else if (MVDiff >= 13) {
if (Rs < 658) {
if (SC < 734) {
if (mu_mv_mag_sq < 93) {
return 1;
}
else if (mu_mv_mag_sq >= 93) {
return 2;
}
}
else if (SC >= 734) {
if (MVDiff < 48) {
return 8;
}
else if (MVDiff >= 48) {
if (Rs < 610) {
return 1;
}
else if (Rs >= 610) {
if (Rs < 635) {
return 2;
}
else if (Rs >= 635) {
return 8;
}
}
}
}
}
else if (Rs >= 658) {
return 1;
}
}
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 7) {
if (diffRsCsdiff < -4) {
if (Rs < 489) {
return 4;
}
else if (Rs >= 489) {
return 2;
}
}
else if (diffRsCsdiff >= -4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (posBalance < 55) {
if (tcor < 47) {
return 1;
}
else if (tcor >= 47) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (Rs < 612) {
if (Rs < 506) {
return 2;
}
else if (Rs >= 506) {
return 4;
}
}
else if (Rs >= 612) {
return 1;
}
}
}
}
}
else if (posBalance >= 58) {
if (AFD < 22) {
if (mu_mv_mag_sq < 52) {
if (CsDiff < 3) {
if (ssDCval < 80) {
if (SC < 507) {
return 8;
}
else if (SC >= 507) {
return 1;
}
}
else if (ssDCval >= 80) {
if (ssDCval < 105) {
if (MVDiff < 39) {
if (tcor < 53) {
return 8;
}
else if (tcor >= 53) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
return 2;
}
}
}
else if (MVDiff >= 39) {
if (Rs < 1148) {
return 2;
}
else if (Rs >= 1148) {
return 8;
}
}
}
else if (ssDCval >= 105) {
if (Cs < 93) {
if (Rs < 802) {
return 8;
}
else if (Rs >= 802) {
return 4;
}
}
else if (Cs >= 93) {
if (Rs < 1168) {
return 1;
}
else if (Rs >= 1168) {
return 2;
}
}
}
}
}
else if (CsDiff >= 3) {
if (ssDCval < 50) {
if (Cs < 354) {
return 8;
}
else if (Cs >= 354) {
return 2;
}
}
else if (ssDCval >= 50) {
if (AFD < 21) {
if (mu_mv_mag_sq < 22) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (mu_mv_mag_sq >= 22) {
if (Rs < 499) {
return 1;
}
else if (Rs >= 499) {
return 2;
}
}
}
else if (AFD >= 21) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (tcor < 46) {
return 4;
}
else if (tcor >= 46) {
if (Rs < 689) {
return 1;
}
else if (Rs >= 689) {
return 2;
}
}
}
}
else if (AFD >= 22) {
if (Cs < 190) {
if (ssDCval < 87) {
return 1;
}
else if (ssDCval >= 87) {
return 2;
}
}
else if (Cs >= 190) {
if (Cs < 364) {
if (diffMVdiffVal < 17) {
if (mcTcor < 81) {
return 8;
}
else if (mcTcor >= 81) {
return 1;
}
}
else if (diffMVdiffVal >= 17) {
if (Cs < 353) {
return 8;
}
else if (Cs >= 353) {
return 1;
}
}
}
else if (Cs >= 364) {
if (mcTcor < 90) {
return 2;
}
else if (mcTcor >= 90) {
if (Rs < 325) {
return 4;
}
else if (Rs >= 325) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 56) {
if (MVDiff < 34) {
if (Rs < 1392) {
if (negBalance < 49) {
if (posBalance < 69) {
return 1;
}
else if (posBalance >= 69) {
return 2;
}
}
else if (negBalance >= 49) {
if (Cs < 126) {
return 1;
}
else if (Cs >= 126) {
if (RsDiff < 12) {
return 2;
}
else if (RsDiff >= 12) {
return 1;
}
}
}
}
else if (Rs >= 1392) {
if (ssDCval < 115) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
if (negBalance < 49) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
if (posBalance < 53) {
if (MVDiff < 24) {
return 1;
}
else if (MVDiff >= 24) {
return 2;
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
else if (negBalance >= 49) {
return 2;
}
}
}
else if (ssDCval >= 115) {
if (Rs < 1782) {
return 1;
}
else if (Rs >= 1782) {
return 8;
}
}
}
}
else if (MVDiff >= 34) {
if (diffAFD < 6) {
if (posBalance < 53) {
if (ssDCval < 48) {
if (Cs < 162) {
return 1;
}
else if (Cs >= 162) {
return 2;
}
}
else if (ssDCval >= 48) {
if (mu_mv_mag_sq < 65) {
if (SC < 2487) {
if (mu_mv_mag_sq < 15) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 2;
}
}
else if (mu_mv_mag_sq >= 15) {
return 1;
}
}
else if (SC >= 2487) {
return 8;
}
}
else if (mu_mv_mag_sq >= 65) {
if (Rs < 1565) {
return 2;
}
else if (Rs >= 1565) {
return 1;
}
}
}
}
else if (posBalance >= 53) {
if (tcor < 59) {
if (tcor < 27) {
if (mu_mv_mag_sq < 18) {
return 2;
}
else if (mu_mv_mag_sq >= 18) {
if (diffAFD < -2) {
if (Cs < 291) {
return 4;
}
else if (Cs >= 291) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Rs < 1683) {
return 1;
}
else if (Rs >= 1683) {
return 4;
}
}
}
}
else if (tcor >= 27) {
if (posBalance < 63) {
if (ssDCval < 116) {
if (Rs < 409) {
return 8;
}
else if (Rs >= 409) {
if (Cs < 188) {
return 2;
}
else if (Cs >= 188) {
if (SC < 1737) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
return 4;
}
}
else if (SC >= 1737) {
return 4;
}
}
}
}
else if (ssDCval >= 116) {
return 8;
}
}
else if (posBalance >= 63) {
return 2;
}
}
}
else if (tcor >= 59) {
if (Cs < 142) {
if (gchDC < 7) {
return 2;
}
else if (gchDC >= 7) {
if (Rs < 420) {
return 2;
}
else if (Rs >= 420) {
return 1;
}
}
}
else if (Cs >= 142) {
if (AFD < 26) {
if (ssDCval < 76) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
return 4;
}
}
else if (ssDCval >= 76) {
if (diffTSC < 1) {
if (negBalance < 56) {
if (Rs < 2145) {
return 1;
}
else if (Rs >= 2145) {
return 2;
}
}
else if (negBalance >= 56) {
return 2;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (AFD >= 26) {
return 1;
}
}
}
}
}
else if (diffAFD >= 6) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
if (RsDiff < 21) {
if (mcTcor < 55) {
return 8;
}
else if (mcTcor >= 55) {
return 1;
}
}
else if (RsDiff >= 21) {
return 2;
}
}
}
}
}
}
else if (Cs >= 395) {
if (negBalance < 39) {
if (refDCval < 139) {
if (Cs < 464) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
if (posBalance < 71) {
return 1;
}
else if (posBalance >= 71) {
if (Rs < 790) {
return 8;
}
else if (Rs >= 790) {
return 1;
}
}
}
}
else if (Cs >= 464) {
if (tcor < 60) {
if (Rs < 528) {
return 2;
}
else if (Rs >= 528) {
return 1;
}
}
else if (tcor >= 60) {
if (mu_mv_mag_sq < 64) {
return 2;
}
else if (mu_mv_mag_sq >= 64) {
return 1;
}
}
}
}
else if (refDCval >= 139) {
return 8;
}
}
else if (negBalance >= 39) {
if (RsCsDiff < 203) {
if (MVDiff < 30) {
if (mu_mv_mag_sq < 5) {
if (Cs < 525) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (Cs >= 525) {
if (Rs < 736) {
return 4;
}
else if (Rs >= 736) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (tcor < 55) {
if (AFD < 23) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
if (Rs < 467) {
return 4;
}
else if (Rs >= 467) {
return 8;
}
}
}
else if (AFD >= 23) {
if (AFD < 31) {
return 4;
}
else if (AFD >= 31) {
return 2;
}
}
}
else if (tcor >= 55) {
if (negBalance < 56) {
if (AFD < 11) {
if (SC < 1550) {
if (Cs < 468) {
return 8;
}
else if (Cs >= 468) {
return 2;
}
}
else if (SC >= 1550) {
return 8;
}
}
else if (AFD >= 11) {
if (Rs < 1179) {
if (Cs < 445) {
return 8;
}
else if (Cs >= 445) {
return 1;
}
}
else if (Rs >= 1179) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
else if (negBalance >= 56) {
if (MVDiff < 27) {
if (Rs < 868) {
if (Rs < 534) {
return 4;
}
else if (Rs >= 534) {
return 1;
}
}
else if (Rs >= 868) {
return 4;
}
}
else if (MVDiff >= 27) {
return 8;
}
}
}
}
}
else if (MVDiff >= 30) {
if (mcTcor < 82) {
if (ssDCval < 67) {
if (AFD < 15) {
if (RsCsDiff < 29) {
return 4;
}
else if (RsCsDiff >= 29) {
return 8;
}
}
else if (AFD >= 15) {
if (MVDiff < 50) {
return 2;
}
else if (MVDiff >= 50) {
if (Rs < 508) {
return 8;
}
else if (Rs >= 508) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (RsCsDiff < 117) {
if (mu_mv_mag_sq < 22) {
if (AFD < 33) {
if (Rs < 1449) {
return 1;
}
else if (Rs >= 1449) {
return 4;
}
}
else if (AFD >= 33) {
if (Rs < 768) {
return 4;
}
else if (Rs >= 768) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 22) {
return 4;
}
}
else if (RsCsDiff >= 117) {
if (refDCval < 111) {
if (gchDC < 5) {
return 4;
}
else if (gchDC >= 5) {
return 1;
}
}
else if (refDCval >= 111) {
if (Rs < 2025) {
return 8;
}
else if (Rs >= 2025) {
return 1;
}
}
}
}
}
else if (mcTcor >= 82) {
if (diffMVdiffVal < 26) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (diffRsCsdiff < -64) {
return 4;
}
else if (diffRsCsdiff >= -64) {
if (refDCval < 75) {
if (tcor < 70) {
if (mcTcor < 87) {
if (mu_mv_mag_sq < 26) {
if (MVDiff < 67) {
return 2;
}
else if (MVDiff >= 67) {
return 8;
}
}
else if (mu_mv_mag_sq >= 26) {
return 4;
}
}
else if (mcTcor >= 87) {
if (negBalance < 45) {
if (Rs < 572) {
return 2;
}
else if (Rs >= 572) {
return 1;
}
}
else if (negBalance >= 45) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
}
else if (tcor >= 70) {
if (Cs < 415) {
return 4;
}
else if (Cs >= 415) {
return 2;
}
}
}
else if (refDCval >= 75) {
if (Rs < 2071) {
if (diffMVdiffVal < 10) {
if (MVDiff < 73) {
if (diffMVdiffVal < -26) {
if (Rs < 1469) {
return 2;
}
else if (Rs >= 1469) {
return 1;
}
}
else if (diffMVdiffVal >= -26) {
return 1;
}
}
else if (MVDiff >= 73) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (diffMVdiffVal >= 10) {
return 2;
}
}
else if (Rs >= 2071) {
return 8;
}
}
}
}
}
}
else if (diffMVdiffVal >= 26) {
if (diffMVdiffVal < 35) {
return 1;
}
else if (diffMVdiffVal >= 35) {
if (Cs < 411) {
return 1;
}
else if (Cs >= 411) {
if (Cs < 522) {
if (AFD < 40) {
return 8;
}
else if (AFD >= 40) {
return 1;
}
}
else if (Cs >= 522) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 203) {
if (gchDC < 4) {
if (diffRsCsdiff < 3) {
if (Cs < 432) {
return 8;
}
else if (Cs >= 432) {
if (tcor < 70) {
return 2;
}
else if (tcor >= 70) {
return 1;
}
}
}
else if (diffRsCsdiff >= 3) {
if (tcor < 27) {
if (Rs < 1837) {
return 4;
}
else if (Rs >= 1837) {
return 2;
}
}
else if (tcor >= 27) {
return 2;
}
}
}
else if (gchDC >= 4) {
if (MVDiff < 236) {
return 1;
}
else if (MVDiff >= 236) {
return 2;
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 81) {
if (RsCsDiff < 71) {
if (Cs < 294) {
if (ssDCval < 139) {
if (ssDCval < 76) {
if (negBalance < 32) {
if (Rs < 1403) {
if (mu_mv_mag_sq < 43) {
return 1;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 1060) {
return 2;
}
else if (Rs >= 1060) {
return 1;
}
}
}
else if (Rs >= 1403) {
return 2;
}
}
else if (negBalance >= 32) {
if (MVDiff < 93) {
if (SC < 1802) {
return 1;
}
else if (SC >= 1802) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (MVDiff >= 93) {
if (Rs < 729) {
if (Cs < 111) {
return 1;
}
else if (Cs >= 111) {
return 8;
}
}
else if (Rs >= 729) {
return 1;
}
}
}
}
else if (ssDCval >= 76) {
if (Cs < 256) {
if (MVDiff < 23) {
if (Rs < 1185) {
if (Cs < 161) {
return 1;
}
else if (Cs >= 161) {
if (Cs < 175) {
return 2;
}
else if (Cs >= 175) {
return 1;
}
}
}
else if (Rs >= 1185) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (diffRsCsdiff < 9) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 1;
}
}
}
else if (diffRsCsdiff >= 9) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (Rs < 1314) {
return 1;
}
else if (Rs >= 1314) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 23) {
if (posBalance < 94) {
if (diffRsCsdiff < -11) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
if (tcor < 91) {
if (MVDiff < 28) {
return 1;
}
else if (MVDiff >= 28) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
if (Rs < 1287) {
return 2;
}
else if (Rs >= 1287) {
return 8;
}
}
}
}
}
else if (tcor >= 91) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= -11) {
if (Cs < 142) {
return 1;
}
else if (Cs >= 142) {
if (Rs < 718) {
return 8;
}
else if (Rs >= 718) {
if (diffRsCsdiff < 34) {
if (Rs < 1034) {
if (Rs < 975) {
return 1;
}
else if (Rs >= 975) {
return 2;
}
}
else if (Rs >= 1034) {
return 1;
}
}
else if (diffRsCsdiff >= 34) {
if (Rs < 1318) {
return 1;
}
else if (Rs >= 1318) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 94) {
if (Cs < 146) {
return 2;
}
else if (Cs >= 146) {
return 1;
}
}
}
}
else if (Cs >= 256) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (RsDiff < 6) {
if (diffAFD < 1) {
if (Rs < 1778) {
return 1;
}
else if (Rs >= 1778) {
return 4;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (RsDiff >= 6) {
return 1;
}
}
}
}
}
else if (ssDCval >= 139) {
if (gchDC < 2) {
if (posBalance < 52) {
if (SC < 1141) {
return 2;
}
else if (SC >= 1141) {
return 1;
}
}
else if (posBalance >= 52) {
return 4;
}
}
else if (gchDC >= 2) {
if (ssDCval < 174) {
return 1;
}
else if (ssDCval >= 174) {
return 2;
}
}
}
}
else if (Cs >= 294) {
if (Rs < 871) {
if (posBalance < 51) {
if (ssDCval < 70) {
if (negBalance < 29) {
return 2;
}
else if (negBalance >= 29) {
if (Rs < 821) {
if (ssDCval < 62) {
if (Rs < 605) {
return 8;
}
else if (Rs >= 605) {
return 1;
}
}
else if (ssDCval >= 62) {
return 8;
}
}
else if (Rs >= 821) {
return 2;
}
}
}
else if (ssDCval >= 70) {
if (negBalance < 62) {
if (Rs < 475) {
return 2;
}
else if (Rs >= 475) {
return 4;
}
}
else if (negBalance >= 62) {
if (Cs < 457) {
return 2;
}
else if (Cs >= 457) {
if (diffRsCsdiff < 5) {
return 8;
}
else if (diffRsCsdiff >= 5) {
return 2;
}
}
}
}
}
else if (posBalance >= 51) {
if (refDCval < 87) {
if (ssDCval < 64) {
if (tcor < 85) {
if (Cs < 307) {
return 8;
}
else if (Cs >= 307) {
if (Rs < 766) {
return 4;
}
else if (Rs >= 766) {
return 2;
}
}
}
else if (tcor >= 85) {
if (Cs < 428) {
return 1;
}
else if (Cs >= 428) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (mcTcor < 98) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
if (Rs < 654) {
return 1;
}
else if (Rs >= 654) {
return 4;
}
}
}
else if (mcTcor >= 98) {
return 4;
}
}
}
else if (refDCval >= 87) {
if (MVDiff < 20) {
if (Cs < 428) {
return 4;
}
else if (Cs >= 428) {
return 1;
}
}
else if (MVDiff >= 20) {
if (Rs < 495) {
if (Rs < 430) {
return 8;
}
else if (Rs >= 430) {
return 2;
}
}
else if (Rs >= 495) {
if (negBalance < 62) {
return 8;
}
else if (negBalance >= 62) {
return 1;
}
}
}
}
}
}
else if (Rs >= 871) {
if (AFD < 7) {
if (MVDiff < 7) {
if (negBalance < 30) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
return 2;
}
}
else if (negBalance >= 30) {
if (ssDCval < 71) {
if (Cs < 395) {
return 1;
}
else if (Cs >= 395) {
return 2;
}
}
else if (ssDCval >= 71) {
if (Rs < 1646) {
return 2;
}
else if (Rs >= 1646) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 55) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
else if (ssDCval >= 55) {
if (tcor < 97) {
if (posBalance < 63) {
if (ssDCval < 96) {
if (Rs < 1041) {
return 4;
}
else if (Rs >= 1041) {
return 1;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (posBalance >= 63) {
if (Rs < 2190) {
return 2;
}
else if (Rs >= 2190) {
return 1;
}
}
}
else if (tcor >= 97) {
return 4;
}
}
}
}
else if (AFD >= 7) {
if (TSC < 4) {
if (ssDCval < 92) {
if (ssDCval < 60) {
if (Rs < 1472) {
if (Rs < 1019) {
return 2;
}
else if (Rs >= 1019) {
return 1;
}
}
else if (Rs >= 1472) {
return 4;
}
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (ssDCval >= 92) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
return 2;
}
}
}
else if (TSC >= 4) {
if (mcTcor < 98) {
if (Rs < 1259) {
if (ssDCval < 135) {
if (MVDiff < 43) {
if (negBalance < 37) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (negBalance >= 37) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1196) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
return 1;
}
}
else if (Rs >= 1196) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (mu_mv_mag_sq < 3) {
if (Cs < 443) {
return 1;
}
else if (Cs >= 443) {
if (Rs < 1089) {
return 4;
}
else if (Rs >= 1089) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 10) {
if (Rs < 1229) {
if (SC < 1488) {
return 1;
}
else if (SC >= 1488) {
return 8;
}
}
else if (Rs >= 1229) {
return 4;
}
}
else if (AFD >= 10) {
if (Cs < 525) {
return 1;
}
else if (Cs >= 525) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 43) {
if (MVDiff < 54) {
return 4;
}
else if (MVDiff >= 54) {
return 2;
}
}
}
else if (ssDCval >= 135) {
return 4;
}
}
else if (Rs >= 1259) {
if (posBalance < 54) {
if (ssDCval < 62) {
if (Rs < 2018) {
if (diffAFD < 1) {
if (SC < 1956) {
if (Rs < 1294) {
return 2;
}
else if (Rs >= 1294) {
return 1;
}
}
else if (SC >= 1956) {
return 2;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (Rs >= 2018) {
return 4;
}
}
else if (ssDCval >= 62) {
if (MVDiff < 2) {
if (Rs < 1673) {
return 4;
}
else if (Rs >= 1673) {
return 1;
}
}
else if (MVDiff >= 2) {
if (Rs < 2335) {
return 1;
}
else if (Rs >= 2335) {
if (Rs < 2361) {
return 4;
}
else if (Rs >= 2361) {
return 1;
}
}
}
}
}
else if (posBalance >= 54) {
if (mcTcor < 97) {
if (SC < 1872) {
return 1;
}
else if (SC >= 1872) {
if (Rs < 1491) {
if (ssDCval < 92) {
if (Rs < 1386) {
return 2;
}
else if (Rs >= 1386) {
return 4;
}
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (Rs >= 1491) {
if (tcor < 89) {
if (ssDCval < 110) {
if (Rs < 1584) {
return 2;
}
else if (Rs >= 1584) {
return 1;
}
}
else if (ssDCval >= 110) {
return 8;
}
}
else if (tcor >= 89) {
if (TSC < 7) {
if (Cs < 443) {
if (Rs < 1645) {
return 2;
}
else if (Rs >= 1645) {
return 1;
}
}
else if (Cs >= 443) {
return 2;
}
}
else if (TSC >= 7) {
if (Cs < 518) {
return 1;
}
else if (Cs >= 518) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 97) {
return 1;
}
}
}
}
else if (mcTcor >= 98) {
if (ssDCval < 71) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
else if (ssDCval >= 71) {
if (Cs < 509) {
return 1;
}
else if (Cs >= 509) {
if (Rs < 1388) {
return 8;
}
else if (Rs >= 1388) {
return 1;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 71) {
if (ssDCval < 111) {
if (MVDiff < 38) {
if (negBalance < 48) {
if (posBalance < 54) {
if (Rs < 2095) {
return 1;
}
else if (Rs >= 2095) {
if (Rs < 2348) {
if (SC < 2781) {
return 2;
}
else if (SC >= 2781) {
return 1;
}
}
else if (Rs >= 2348) {
if (Rs < 2358) {
return 4;
}
else if (Rs >= 2358) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
if (refDCval < 79) {
if (mcTcor < 90) {
return 1;
}
else if (mcTcor >= 90) {
return 2;
}
}
else if (refDCval >= 79) {
if (Cs < 393) {
if (diffAFD < -1) {
if (Rs < 2024) {
return 2;
}
else if (Rs >= 2024) {
return 1;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (Cs >= 393) {
if (ssDCval < 110) {
return 1;
}
else if (ssDCval >= 110) {
return 2;
}
}
}
}
}
else if (negBalance >= 48) {
if (diffMVdiffVal < -13) {
if (negBalance < 53) {
if (ssDCval < 103) {
return 2;
}
else if (ssDCval >= 103) {
return 1;
}
}
else if (negBalance >= 53) {
return 1;
}
}
else if (diffMVdiffVal >= -13) {
if (mcTcor < 88) {
return 4;
}
else if (mcTcor >= 88) {
if (posBalance < 63) {
if (diffRsCsdiff < 20) {
return 2;
}
else if (diffRsCsdiff >= 20) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
if (SC < 2168) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 1;
}
}
else if (SC >= 2168) {
return 2;
}
}
}
}
else if (posBalance >= 63) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 38) {
if (mcTcor < 88) {
if (Rs < 1875) {
return 1;
}
else if (Rs >= 1875) {
return 4;
}
}
else if (mcTcor >= 88) {
if (TSC < 6) {
if (Cs < 282) {
return 2;
}
else if (Cs >= 282) {
return 1;
}
}
else if (TSC >= 6) {
if (mcTcor < 91) {
if (MVDiff < 60) {
return 8;
}
else if (MVDiff >= 60) {
if (diffRsCsdiff < -40) {
if (Rs < 1220) {
return 8;
}
else if (Rs >= 1220) {
return 2;
}
}
else if (diffRsCsdiff >= -40) {
return 1;
}
}
}
else if (mcTcor >= 91) {
if (TSC < 14) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 83) {
if (Cs < 179) {
return 1;
}
else if (Cs >= 179) {
if (AFD < 30) {
return 2;
}
else if (AFD >= 30) {
return 8;
}
}
}
else if (ssDCval >= 83) {
if (mcTcor < 96) {
if (diffMVdiffVal < -8) {
return 8;
}
else if (diffMVdiffVal >= -8) {
if (Rs < 2268) {
return 1;
}
else if (Rs >= 2268) {
return 8;
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (mu_mv_mag_sq < 27) {
if (Rs < 739) {
return 2;
}
else if (Rs >= 739) {
return 1;
}
}
else if (mu_mv_mag_sq >= 27) {
return 2;
}
}
}
}
else if (TSC >= 14) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 111) {
if (AFD < 13) {
if (Rs < 1254) {
return 4;
}
else if (Rs >= 1254) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
}
else if (AFD >= 13) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 4) {
if (MVDiff < 69) {
return 1;
}
else if (MVDiff >= 69) {
if (RsCsDiff < 128) {
return 8;
}
else if (RsCsDiff >= 128) {
if (Rs < 1203) {
return 8;
}
else if (Rs >= 1203) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 4) {
if (Cs < 371) {
return 1;
}
else if (Cs >= 371) {
if (negBalance < 52) {
if (posBalance < 82) {
return 2;
}
else if (posBalance >= 82) {
return 1;
}
}
else if (negBalance >= 52) {
return 1;
}
}
}
}
else if (mcTcor >= 95) {
if (CsDiff < 3) {
if (Rs < 2067) {
if (ssDCval < 177) {
return 1;
}
else if (ssDCval >= 177) {
return 2;
}
}
else if (Rs >= 2067) {
if (SC < 2498) {
if (Cs < 267) {
return 1;
}
else if (Cs >= 267) {
return 2;
}
}
else if (SC >= 2498) {
if (ssDCval < 122) {
if (Cs < 352) {
return 1;
}
else if (Cs >= 352) {
return 8;
}
}
else if (ssDCval >= 122) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
if (TSC < 7) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (TSC >= 7) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (AFD < 9) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (RsCsDiff < 19) {
if (Rs < 1584) {
return 4;
}
else if (Rs >= 1584) {
if (Cs < 488) {
return 8;
}
else if (Cs >= 488) {
return 1;
}
}
}
else if (RsCsDiff >= 19) {
return 8;
}
}
}
else if (AFD >= 9) {
if (Rs < 662) {
return 1;
}
else if (Rs >= 662) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Cs >= 533) {
if (negBalance < 13) {
if (Rs < 701) {
if (ssDCval < 56) {
if (tcor < 99) {
if (AFD < 2) {
if (Rs < 336) {
if (Rs < 271) {
return 2;
}
else if (Rs >= 271) {
return 1;
}
}
else if (Rs >= 336) {
return 8;
}
}
else if (AFD >= 2) {
if (tcor < 48) {
if (Rs < 412) {
return 8;
}
else if (Rs >= 412) {
return 2;
}
}
else if (tcor >= 48) {
return 1;
}
}
}
else if (tcor >= 99) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (ssDCval >= 56) {
if (Rs < 662) {
if (Rs < 515) {
if (posBalance < 50) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
}
else if (posBalance >= 50) {
if (refDCval < 127) {
return 1;
}
else if (refDCval >= 127) {
if (Rs < 476) {
return 4;
}
else if (Rs >= 476) {
return 8;
}
}
}
}
else if (Rs >= 515) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
}
else if (Rs >= 662) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (posBalance < 31) {
return 1;
}
else if (posBalance >= 31) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
return 8;
}
}
}
}
}
}
else if (Rs >= 701) {
if (diffAFD < -3) {
if (SC < 2456) {
if (diffRsCsdiff < -47) {
if (diffAFD < -16) {
if (posBalance < 1) {
if (Rs < 820) {
return 1;
}
else if (Rs >= 820) {
return 4;
}
}
else if (posBalance >= 1) {
if (Cs < 880) {
return 8;
}
else if (Cs >= 880) {
return 1;
}
}
}
else if (diffAFD >= -16) {
if (Cs < 1062) {
return 8;
}
else if (Cs >= 1062) {
if (ssDCval < 140) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
return 1;
}
}
else if (ssDCval >= 140) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -47) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
if (Cs < 715) {
if (Rs < 936) {
return 2;
}
else if (Rs >= 936) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
return 1;
}
}
}
else if (Cs >= 715) {
if (ssDCval < 73) {
if (Rs < 1244) {
return 8;
}
else if (Rs >= 1244) {
return 1;
}
}
else if (ssDCval >= 73) {
if (Cs < 1060) {
if (diffMVdiffVal < -44) {
return 8;
}
else if (diffMVdiffVal >= -44) {
if (diffTSC < -6) {
return 8;
}
else if (diffTSC >= -6) {
if (Cs < 1014) {
return 4;
}
else if (Cs >= 1014) {
if (Rs < 1131) {
return 8;
}
else if (Rs >= 1131) {
return 4;
}
}
}
}
}
else if (Cs >= 1060) {
return 1;
}
}
}
}
}
}
else if (SC >= 2456) {
if (MVDiff < 16) {
if (diffRsCsdiff < -213) {
return 1;
}
else if (diffRsCsdiff >= -213) {
return 8;
}
}
else if (MVDiff >= 16) {
if (diffMVdiffVal < 1) {
if (ssDCval < 128) {
return 8;
}
else if (ssDCval >= 128) {
if (diffAFD < -10) {
if (diffMVdiffVal < -15) {
return 4;
}
else if (diffMVdiffVal >= -15) {
if (Cs < 1076) {
return 4;
}
else if (Cs >= 1076) {
return 8;
}
}
}
else if (diffAFD >= -10) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
if (diffRsCsdiff < -161) {
return 1;
}
else if (diffRsCsdiff >= -161) {
return 8;
}
}
}
}
}
else if (diffAFD >= -3) {
if (Rs < 1840) {
if (Rs < 967) {
if (posBalance < 12) {
if (ssDCval < 118) {
return 8;
}
else if (ssDCval >= 118) {
if (SC < 1594) {
return 8;
}
else if (SC >= 1594) {
return 2;
}
}
}
else if (posBalance >= 12) {
if (Cs < 1170) {
if (CsDiff < 3) {
if (ssDCval < 71) {
if (SC < 1474) {
if (RsDiff < 1) {
if (Rs < 882) {
return 2;
}
else if (Rs >= 882) {
return 8;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (SC >= 1474) {
return 8;
}
}
else if (ssDCval >= 71) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (MVDiff < 1) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 758) {
return 8;
}
else if (Rs >= 758) {
return 1;
}
}
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (Cs >= 1170) {
if (Rs < 872) {
return 2;
}
else if (Rs >= 872) {
return 8;
}
}
}
}
else if (Rs >= 967) {
if (Rs < 1037) {
if (ssDCval < 152) {
return 8;
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (Rs >= 1037) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (refDCval < 46) {
if (Cs < 1161) {
return 1;
}
else if (Cs >= 1161) {
return 4;
}
}
else if (refDCval >= 46) {
if (posBalance < 28) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
if (ssDCval < 137) {
if (negBalance < 1) {
if (Rs < 1465) {
return 1;
}
else if (Rs >= 1465) {
return 8;
}
}
else if (negBalance >= 1) {
if (ssDCval < 96) {
return 4;
}
else if (ssDCval >= 96) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
}
else if (ssDCval >= 137) {
if (diffMVdiffVal < -15) {
if (Rs < 1102) {
return 8;
}
else if (Rs >= 1102) {
return 4;
}
}
else if (diffMVdiffVal >= -15) {
return 8;
}
}
}
}
else if (posBalance >= 28) {
if (SC < 1801) {
return 8;
}
else if (SC >= 1801) {
if (ssDCval < 53) {
if (Cs < 620) {
return 1;
}
else if (Cs >= 620) {
return 8;
}
}
else if (ssDCval >= 53) {
if (ssDCval < 75) {
if (Cs < 996) {
if (Cs < 676) {
return 1;
}
else if (Cs >= 676) {
return 8;
}
}
else if (Cs >= 996) {
return 1;
}
}
else if (ssDCval >= 75) {
if (Rs < 1055) {
return 1;
}
else if (Rs >= 1055) {
if (SC < 1997) {
if (Cs < 710) {
return 8;
}
else if (Cs >= 710) {
return 4;
}
}
else if (SC >= 1997) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 1840) {
if (Rs < 2053) {
if (Cs < 1001) {
if (negBalance < 5) {
return 8;
}
else if (negBalance >= 5) {
if (SC < 2947) {
return 1;
}
else if (SC >= 2947) {
if (Rs < 2006) {
return 8;
}
else if (Rs >= 2006) {
return 4;
}
}
}
}
else if (Cs >= 1001) {
if (refDCval < 68) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 4;
}
}
else if (refDCval >= 68) {
if (Rs < 1911) {
return 1;
}
else if (Rs >= 1911) {
return 8;
}
}
}
}
else if (Rs >= 2053) {
if (negBalance < 11) {
if (mu_mv_mag_sq < 94) {
return 4;
}
else if (mu_mv_mag_sq >= 94) {
return 8;
}
}
else if (negBalance >= 11) {
if (AFD < 2) {
if (posBalance < 29) {
return 8;
}
else if (posBalance >= 29) {
if (Cs < 1317) {
return 2;
}
else if (Cs >= 1317) {
if (posBalance < 30) {
return 2;
}
else if (posBalance >= 30) {
return 8;
}
}
}
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 14) {
if (posBalance < 45) {
if (refDCval < 46) {
if (MVDiff < 9) {
if (Rs < 1482) {
if (ssDCval < 31) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (ssDCval >= 31) {
if (refDCval < 37) {
if (Rs < 689) {
return 1;
}
else if (Rs >= 689) {
return 2;
}
}
else if (refDCval >= 37) {
if (MVDiff < 2) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Cs < 534) {
return 2;
}
else if (Cs >= 534) {
return 8;
}
}
}
}
}
else if (Rs >= 1482) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
else if (MVDiff >= 9) {
if (tcor < 62) {
return 8;
}
else if (tcor >= 62) {
if (negBalance < 32) {
if (ssDCval < 27) {
return 4;
}
else if (ssDCval >= 27) {
if (ssDCval < 32) {
if (Rs < 842) {
return 1;
}
else if (Rs >= 842) {
return 8;
}
}
else if (ssDCval >= 32) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Rs < 1140) {
return 4;
}
else if (Rs >= 1140) {
return 1;
}
}
}
}
}
else if (negBalance >= 32) {
if (ssDCval < 44) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
else if (ssDCval >= 44) {
if (Rs < 1595) {
if (Rs < 1368) {
return 8;
}
else if (Rs >= 1368) {
return 2;
}
}
else if (Rs >= 1595) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 46) {
if (mu_mv_mag_sq < 73) {
if (Cs < 946) {
if (negBalance < 67) {
if (Cs < 762) {
if (SC < 1480) {
if (AFD < 9) {
if (mcTcor < 95) {
if (diffMVdiffVal < -1) {
if (Rs < 507) {
return 2;
}
else if (Rs >= 507) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (mcTcor >= 95) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (RsCsDiff < 4) {
return 8;
}
else if (RsCsDiff >= 4) {
return 4;
}
}
}
}
else if (AFD >= 9) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (MVDiff < 49) {
return 2;
}
else if (MVDiff >= 49) {
return 1;
}
}
}
}
else if (SC >= 1480) {
if (ssDCval < 69) {
if (negBalance < 39) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (diffAFD < -1) {
if (gchDC < 1) {
if (Rs < 1515) {
return 2;
}
else if (Rs >= 1515) {
if (Rs < 1902) {
return 1;
}
else if (Rs >= 1902) {
return 4;
}
}
}
else if (gchDC >= 1) {
return 2;
}
}
else if (diffAFD >= -1) {
if (diffMVdiffVal < 3) {
if (Rs < 2086) {
return 1;
}
else if (Rs >= 2086) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 1328) {
if (SC < 1564) {
return 2;
}
else if (SC >= 1564) {
return 8;
}
}
else if (Rs >= 1328) {
if (TSC < 7) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 2;
}
}
else if (TSC >= 7) {
return 2;
}
}
}
}
else if (ssDCval >= 69) {
if (tcor < 84) {
if (Rs < 1902) {
return 4;
}
else if (Rs >= 1902) {
return 1;
}
}
else if (tcor >= 84) {
if (ssDCval < 104) {
return 1;
}
else if (ssDCval >= 104) {
if (Cs < 700) {
return 1;
}
else if (Cs >= 700) {
return 8;
}
}
}
}
}
}
else if (Cs >= 762) {
if (Rs < 841) {
if (ssDCval < 54) {
if (Rs < 537) {
return 8;
}
else if (Rs >= 537) {
return 4;
}
}
else if (ssDCval >= 54) {
if (Rs < 471) {
return 4;
}
else if (Rs >= 471) {
return 2;
}
}
}
else if (Rs >= 841) {
if (mcTcor < 91) {
if (Rs < 1851) {
return 8;
}
else if (Rs >= 1851) {
if (Cs < 796) {
return 1;
}
else if (Cs >= 796) {
return 4;
}
}
}
else if (mcTcor >= 91) {
if (MVDiff < 5) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (RsDiff < 6) {
if (Cs < 834) {
if (Rs < 1191) {
return 1;
}
else if (Rs >= 1191) {
return 8;
}
}
else if (Cs >= 834) {
return 1;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (MVDiff >= 5) {
if (posBalance < 37) {
if (ssDCval < 127) {
return 1;
}
else if (ssDCval >= 127) {
return 2;
}
}
else if (posBalance >= 37) {
if (tcor < 90) {
if (negBalance < 39) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (negBalance >= 39) {
if (AFD < 12) {
if (Rs < 937) {
return 2;
}
else if (Rs >= 937) {
return 8;
}
}
else if (AFD >= 12) {
if (Rs < 1458) {
return 1;
}
else if (Rs >= 1458) {
return 4;
}
}
}
}
else if (tcor >= 90) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (negBalance < 46) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
if (Rs < 1558) {
return 4;
}
else if (Rs >= 1558) {
return 1;
}
}
}
else if (negBalance >= 46) {
if (Rs < 2139) {
if (Rs < 1554) {
return 4;
}
else if (Rs >= 1554) {
return 8;
}
}
else if (Rs >= 2139) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (negBalance >= 67) {
if (mu_mv_mag_sq < 8) {
if (tcor < 97) {
if (Rs < 588) {
return 8;
}
else if (Rs >= 588) {
return 2;
}
}
else if (tcor >= 97) {
return 8;
}
}
else if (mu_mv_mag_sq >= 8) {
if (tcor < 85) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (tcor >= 85) {
return 8;
}
}
}
}
else if (Cs >= 946) {
if (ssDCval < 154) {
if (refDCval < 74) {
if (tcor < 89) {
if (negBalance < 40) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (Cs < 975) {
return 8;
}
else if (Cs >= 975) {
return 2;
}
}
}
}
else if (negBalance >= 40) {
if (negBalance < 43) {
if (Rs < 1720) {
return 8;
}
else if (Rs >= 1720) {
return 4;
}
}
else if (negBalance >= 43) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
else if (tcor >= 89) {
if (RsCsDiff < 16) {
if (Rs < 959) {
return 2;
}
else if (Rs >= 959) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (posBalance < 41) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
if (Rs < 1788) {
if (Rs < 879) {
return 4;
}
else if (Rs >= 879) {
return 8;
}
}
else if (Rs >= 1788) {
return 4;
}
}
}
else if (posBalance >= 41) {
return 4;
}
}
}
}
else if (refDCval >= 74) {
if (negBalance < 31) {
if (Rs < 1181) {
return 1;
}
else if (Rs >= 1181) {
return 8;
}
}
else if (negBalance >= 31) {
if (Cs < 1015) {
if (Cs < 963) {
return 8;
}
else if (Cs >= 963) {
if (Rs < 1762) {
return 4;
}
else if (Rs >= 1762) {
return 1;
}
}
}
else if (Cs >= 1015) {
if (diffMVdiffVal < -9) {
if (diffRsCsdiff < -5) {
return 2;
}
else if (diffRsCsdiff >= -5) {
return 8;
}
}
else if (diffMVdiffVal >= -9) {
return 8;
}
}
}
}
}
else if (ssDCval >= 154) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 73) {
if (refDCval < 56) {
if (RsDiff < 2) {
if (ssDCval < 52) {
if (negBalance < 28) {
if (negBalance < 22) {
if (Rs < 726) {
return 2;
}
else if (Rs >= 726) {
return 8;
}
}
else if (negBalance >= 22) {
return 1;
}
}
else if (negBalance >= 28) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (TSC < 3) {
if (posBalance < 36) {
if (Cs < 987) {
return 8;
}
else if (Cs >= 987) {
return 4;
}
}
else if (posBalance >= 36) {
if (Rs < 2066) {
return 4;
}
else if (Rs >= 2066) {
return 8;
}
}
}
else if (TSC >= 3) {
if (refDCval < 48) {
if (mu_mv_mag_sq < 104) {
return 4;
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 1738) {
return 4;
}
else if (Rs >= 1738) {
return 8;
}
}
}
else if (refDCval >= 48) {
if (negBalance < 28) {
if (Cs < 909) {
return 1;
}
else if (Cs >= 909) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (negBalance >= 28) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Cs < 650) {
return 1;
}
else if (Cs >= 650) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 56) {
if (mu_mv_mag_sq < 91) {
if (posBalance < 34) {
if (Rs < 2232) {
if (tcor < 65) {
return 1;
}
else if (tcor >= 65) {
return 8;
}
}
else if (Rs >= 2232) {
return 4;
}
}
else if (posBalance >= 34) {
if (negBalance < 29) {
if (ssDCval < 65) {
if (Cs < 665) {
return 1;
}
else if (Cs >= 665) {
return 2;
}
}
else if (ssDCval >= 65) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (negBalance >= 29) {
if (ssDCval < 71) {
if (SC < 2281) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 811) {
if (Rs < 541) {
return 8;
}
else if (Rs >= 541) {
return 2;
}
}
else if (Rs >= 811) {
if (Rs < 1224) {
return 4;
}
else if (Rs >= 1224) {
return 8;
}
}
}
}
else if (SC >= 2281) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (Rs < 2004) {
return 1;
}
else if (Rs >= 2004) {
return 2;
}
}
}
}
else if (ssDCval >= 71) {
if (Cs < 992) {
if (tcor < 94) {
if (negBalance < 55) {
if (diffAFD < -3) {
if (Rs < 731) {
return 2;
}
else if (Rs >= 731) {
return 8;
}
}
else if (diffAFD >= -3) {
return 2;
}
}
else if (negBalance >= 55) {
return 8;
}
}
else if (tcor >= 94) {
if (Rs < 1095) {
return 2;
}
else if (Rs >= 1095) {
return 1;
}
}
}
else if (Cs >= 992) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (Rs < 516) {
if (ssDCval < 87) {
return 1;
}
else if (ssDCval >= 87) {
return 4;
}
}
else if (Rs >= 516) {
if (CsDiff < 5) {
if (tcor < 91) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 1799) {
if (Rs < 923) {
if (Rs < 714) {
return 8;
}
else if (Rs >= 714) {
return 2;
}
}
else if (Rs >= 923) {
return 8;
}
}
else if (Rs >= 1799) {
return 2;
}
}
}
else if (tcor >= 91) {
if (Rs < 1713) {
if (Rs < 857) {
if (SC < 1724) {
if (posBalance < 40) {
if (Cs < 672) {
return 8;
}
else if (Cs >= 672) {
if (Rs < 714) {
return 2;
}
else if (Rs >= 714) {
return 1;
}
}
}
else if (posBalance >= 40) {
if (mu_mv_mag_sq < 94) {
return 4;
}
else if (mu_mv_mag_sq >= 94) {
if (TSC < 3) {
if (Rs < 646) {
return 8;
}
else if (Rs >= 646) {
return 4;
}
}
else if (TSC >= 3) {
return 8;
}
}
}
}
else if (SC >= 1724) {
return 4;
}
}
else if (Rs >= 857) {
if (negBalance < 23) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
if (Rs < 1325) {
return 1;
}
else if (Rs >= 1325) {
return 8;
}
}
}
}
else if (negBalance >= 23) {
if (diffTSC < -1) {
if (Rs < 1035) {
if (Rs < 988) {
return 8;
}
else if (Rs >= 988) {
return 4;
}
}
else if (Rs >= 1035) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
else if (Rs >= 1713) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (ssDCval < 84) {
if (Rs < 1824) {
return 1;
}
else if (Rs >= 1824) {
if (MVDiff < 5) {
if (Rs < 2262) {
return 4;
}
else if (Rs >= 2262) {
return 4;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (ssDCval >= 84) {
return 8;
}
}
}
}
}
else if (CsDiff >= 5) {
return 1;
}
}
}
}
}
}
}
else if (posBalance >= 45) {
if (mu_mv_mag_sq < 32) {
if (negBalance < 44) {
if (ssDCval < 63) {
if (refDCval < 47) {
if (mu_mv_mag_sq < 23) {
if (Rs < 825) {
if (negBalance < 36) {
if (Cs < 1044) {
if (ssDCval < 37) {
return 8;
}
else if (ssDCval >= 37) {
if (Rs < 568) {
return 8;
}
else if (Rs >= 568) {
if (Rs < 643) {
return 2;
}
else if (Rs >= 643) {
if (Rs < 664) {
return 1;
}
else if (Rs >= 664) {
return 4;
}
}
}
}
}
else if (Cs >= 1044) {
return 1;
}
}
else if (negBalance >= 36) {
if (RsCsDiff < 11) {
if (Rs < 565) {
return 2;
}
else if (Rs >= 565) {
return 8;
}
}
else if (RsCsDiff >= 11) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
if (diffMVdiffVal < -3) {
if (Rs < 609) {
return 8;
}
else if (Rs >= 609) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
}
}
}
else if (Rs >= 825) {
if (MVDiff < 18) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 8;
}
}
else if (MVDiff >= 18) {
if (Cs < 719) {
if (negBalance < 38) {
if (Rs < 891) {
return 8;
}
else if (Rs >= 891) {
return 1;
}
}
else if (negBalance >= 38) {
return 2;
}
}
else if (Cs >= 719) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (negBalance < 35) {
if (mcTcor < 73) {
if (Rs < 772) {
return 2;
}
else if (Rs >= 772) {
return 8;
}
}
else if (mcTcor >= 73) {
return 1;
}
}
else if (negBalance >= 35) {
return 2;
}
}
}
else if (refDCval >= 47) {
if (negBalance < 32) {
if (Rs < 1109) {
return 4;
}
else if (Rs >= 1109) {
if (Rs < 1611) {
return 1;
}
else if (Rs >= 1611) {
return 8;
}
}
}
else if (negBalance >= 32) {
if (Cs < 732) {
if (CsDiff < 3) {
if (Cs < 583) {
return 1;
}
else if (Cs >= 583) {
return 8;
}
}
else if (CsDiff >= 3) {
if (diffMVdiffVal < 3) {
if (mu_mv_mag_sq < 12) {
return 2;
}
else if (mu_mv_mag_sq >= 12) {
if (Rs < 688) {
return 4;
}
else if (Rs >= 688) {
return 1;
}
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1460) {
return 1;
}
else if (Rs >= 1460) {
return 4;
}
}
}
}
else if (Cs >= 732) {
if (Rs < 614) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (Rs < 454) {
return 1;
}
else if (Rs >= 454) {
return 8;
}
}
}
else if (Rs >= 614) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 11) {
if (Rs < 1125) {
if (Rs < 840) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (Rs >= 840) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (Rs >= 1125) {
if (Cs < 854) {
return 2;
}
else if (Cs >= 854) {
return 8;
}
}
}
else if (MVDiff >= 11) {
if (RsDiff < 4) {
if (Cs < 758) {
return 4;
}
else if (Cs >= 758) {
return 8;
}
}
else if (RsDiff >= 4) {
if (gchDC < 2) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (Rs < 1671) {
return 1;
}
else if (Rs >= 1671) {
return 4;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 813) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
else if (Rs >= 813) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 63) {
if (RsDiff < 2) {
if (SC < 1068) {
if (Rs < 406) {
return 4;
}
else if (Rs >= 406) {
return 8;
}
}
else if (SC >= 1068) {
return 8;
}
}
else if (RsDiff >= 2) {
if (negBalance < 35) {
if (MVDiff < 18) {
if (SC < 2586) {
if (ssDCval < 147) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (ssDCval >= 147) {
return 2;
}
}
else if (SC >= 2586) {
return 2;
}
}
else if (MVDiff >= 18) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Rs < 1955) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
return 8;
}
}
else if (Rs >= 1955) {
return 1;
}
}
}
}
else if (negBalance >= 35) {
if (ssDCval < 87) {
if (Rs < 1190) {
if (Cs < 865) {
if (diffMVdiffVal < 3) {
if (Cs < 838) {
if (AFD < 8) {
if (ssDCval < 73) {
return 4;
}
else if (ssDCval >= 73) {
if (Rs < 650) {
return 2;
}
else if (Rs >= 650) {
return 1;
}
}
}
else if (AFD >= 8) {
if (Rs < 892) {
if (Rs < 849) {
return 2;
}
else if (Rs >= 849) {
return 4;
}
}
else if (Rs >= 892) {
return 1;
}
}
}
else if (Cs >= 838) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
if (ssDCval < 68) {
if (Rs < 643) {
return 4;
}
else if (Rs >= 643) {
return 2;
}
}
else if (ssDCval >= 68) {
return 8;
}
}
}
else if (Cs >= 865) {
if (Cs < 1209) {
if (SC < 1907) {
return 2;
}
else if (SC >= 1907) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (Cs >= 1209) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
if (Rs < 582) {
return 1;
}
else if (Rs >= 582) {
return 4;
}
}
}
}
}
else if (Rs >= 1190) {
if (Rs < 1886) {
if (MVDiff < 47) {
return 1;
}
else if (MVDiff >= 47) {
return 8;
}
}
else if (Rs >= 1886) {
if (refDCval < 73) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (refDCval >= 73) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
if (Rs < 2196) {
return 8;
}
else if (Rs >= 2196) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (SC < 1512) {
return 2;
}
else if (SC >= 1512) {
if (ssDCval < 137) {
if (ssDCval < 104) {
if (Cs < 838) {
if (Rs < 1730) {
return 1;
}
else if (Rs >= 1730) {
return 2;
}
}
else if (Cs >= 838) {
if (posBalance < 57) {
if (Cs < 1176) {
return 4;
}
else if (Cs >= 1176) {
return 8;
}
}
else if (posBalance >= 57) {
if (CsDiff < 5) {
if (Rs < 1648) {
return 8;
}
else if (Rs >= 1648) {
return 2;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
else if (ssDCval >= 104) {
if (posBalance < 72) {
if (negBalance < 38) {
if (Rs < 1413) {
return 1;
}
else if (Rs >= 1413) {
return 8;
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (posBalance >= 72) {
return 2;
}
}
}
else if (ssDCval >= 137) {
if (ssDCval < 149) {
return 4;
}
else if (ssDCval >= 149) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 388) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
else if (Rs >= 388) {
if (RsCsDiff < 51) {
if (Rs < 927) {
if (AFD < 11) {
if (MVDiff < 6) {
if (refDCval < 49) {
if (ssDCval < 45) {
if (Rs < 829) {
return 4;
}
else if (Rs >= 829) {
return 8;
}
}
else if (ssDCval >= 45) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
return 8;
}
}
}
}
else if (refDCval >= 49) {
if (tcor < 77) {
if (Rs < 602) {
return 8;
}
else if (Rs >= 602) {
return 4;
}
}
else if (tcor >= 77) {
if (refDCval < 85) {
return 8;
}
else if (refDCval >= 85) {
if (tcor < 88) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
return 8;
}
}
else if (tcor >= 88) {
return 1;
}
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 141) {
if (negBalance < 48) {
if (AFD < 9) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 761) {
return 8;
}
else if (Rs >= 761) {
return 1;
}
}
}
else if (AFD >= 9) {
if (Rs < 881) {
if (Cs < 586) {
if (Rs < 541) {
return 8;
}
else if (Rs >= 541) {
if (Rs < 670) {
return 2;
}
else if (Rs >= 670) {
return 1;
}
}
}
else if (Cs >= 586) {
return 2;
}
}
else if (Rs >= 881) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (ssDCval < 49) {
if (Rs < 633) {
if (tcor < 64) {
return 8;
}
else if (tcor >= 64) {
return 4;
}
}
else if (Rs >= 633) {
if (TSC < 7) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
return 1;
}
}
else if (TSC >= 7) {
if (Cs < 708) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (Rs < 731) {
return 8;
}
else if (Rs >= 731) {
return 1;
}
}
}
else if (Cs >= 708) {
return 8;
}
}
}
}
else if (ssDCval >= 49) {
if (RsCsDiff < 29) {
if (diffMVdiffVal < 8) {
return 8;
}
else if (diffMVdiffVal >= 8) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
return 8;
}
}
}
else if (RsCsDiff >= 29) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 4;
}
}
}
}
}
else if (ssDCval >= 141) {
return 2;
}
}
}
else if (AFD >= 11) {
if (tcor < 69) {
if (TSC < 9) {
if (ssDCval < 66) {
if (ssDCval < 48) {
if (posBalance < 61) {
if (Rs < 634) {
return 2;
}
else if (Rs >= 634) {
return 8;
}
}
else if (posBalance >= 61) {
if (Rs < 906) {
return 4;
}
else if (Rs >= 906) {
return 8;
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (tcor >= 69) {
if (ssDCval < 101) {
if (negBalance < 52) {
if (ssDCval < 58) {
if (Rs < 511) {
return 2;
}
else if (Rs >= 511) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
}
else if (ssDCval >= 58) {
if (ssDCval < 72) {
if (Rs < 879) {
return 1;
}
else if (Rs >= 879) {
return 8;
}
}
else if (ssDCval >= 72) {
if (MVDiff < 29) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 2;
}
}
else if (MVDiff >= 29) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (diffMVdiffVal < -17) {
if (Rs < 766) {
return 2;
}
else if (Rs >= 766) {
if (Rs < 823) {
return 8;
}
else if (Rs >= 823) {
return 4;
}
}
}
else if (diffMVdiffVal >= -17) {
if (Cs < 772) {
return 8;
}
else if (Cs >= 772) {
if (Rs < 884) {
return 2;
}
else if (Rs >= 884) {
return 8;
}
}
}
}
}
else if (ssDCval >= 101) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Rs < 690) {
return 4;
}
else if (Rs >= 690) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 927) {
if (mcTcor < 89) {
if (Rs < 1361) {
if (ssDCval < 120) {
if (Cs < 878) {
if (diffAFD < -1) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Cs >= 878) {
if (Rs < 1122) {
if (SC < 2088) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (SC < 1858) {
if (Rs < 933) {
return 8;
}
else if (Rs >= 933) {
return 1;
}
}
else if (SC >= 1858) {
return 8;
}
}
}
else if (SC >= 2088) {
if (Rs < 1115) {
return 1;
}
else if (Rs >= 1115) {
return 4;
}
}
}
else if (Rs >= 1122) {
return 8;
}
}
}
else if (ssDCval >= 120) {
return 1;
}
}
else if (Rs >= 1361) {
if (AFD < 10) {
if (MVDiff < 11) {
if (Rs < 1421) {
return 2;
}
else if (Rs >= 1421) {
return 8;
}
}
else if (MVDiff >= 11) {
if (Rs < 1461) {
if (Rs < 1378) {
return 2;
}
else if (Rs >= 1378) {
return 8;
}
}
else if (Rs >= 1461) {
return 2;
}
}
}
else if (AFD >= 10) {
if (posBalance < 58) {
if (diffMVdiffVal < 1) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
else if (mcTcor >= 89) {
if (RsDiff < 5) {
if (ssDCval < 112) {
if (mcTcor < 93) {
if (Cs < 1163) {
if (tcor < 71) {
if (Cs < 979) {
return 8;
}
else if (Cs >= 979) {
return 1;
}
}
else if (tcor >= 71) {
return 8;
}
}
else if (Cs >= 1163) {
if (ssDCval < 59) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (ssDCval >= 59) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (RsCsDiff < 19) {
if (diffMVdiffVal < 9) {
return 8;
}
else if (diffMVdiffVal >= 9) {
return 4;
}
}
else if (RsCsDiff >= 19) {
if (Rs < 1003) {
if (Cs < 1192) {
return 4;
}
else if (Cs >= 1192) {
if (Rs < 957) {
return 2;
}
else if (Rs >= 957) {
return 1;
}
}
}
else if (Rs >= 1003) {
if (AFD < 9) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (AFD >= 9) {
if (posBalance < 55) {
if (tcor < 93) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (posBalance >= 55) {
if (Cs < 570) {
return 2;
}
else if (Cs >= 570) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (AFD < 10) {
if (MVDiff < 30) {
if (ssDCval < 150) {
if (mu_mv_mag_sq < 28) {
return 8;
}
else if (mu_mv_mag_sq >= 28) {
if (Rs < 1097) {
return 4;
}
else if (Rs >= 1097) {
return 8;
}
}
}
else if (ssDCval >= 150) {
return 2;
}
}
else if (MVDiff >= 30) {
return 4;
}
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (RsDiff >= 5) {
if (Cs < 564) {
if (tcor < 84) {
return 2;
}
else if (tcor >= 84) {
return 1;
}
}
else if (Cs >= 564) {
if (MVDiff < 12) {
if (AFD < 13) {
if (tcor < 90) {
if (Cs < 696) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
if (Rs < 1366) {
return 8;
}
else if (Rs >= 1366) {
return 2;
}
}
}
else if (Cs >= 696) {
if (posBalance < 55) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Rs < 1512) {
return 4;
}
else if (Rs >= 1512) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (Rs < 949) {
return 4;
}
else if (Rs >= 949) {
return 8;
}
}
}
}
else if (tcor >= 90) {
if (RsDiff < 6) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (Rs < 1234) {
return 1;
}
else if (Rs >= 1234) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1538) {
return 8;
}
else if (Rs >= 1538) {
return 1;
}
}
}
}
}
else if (RsDiff >= 6) {
return 1;
}
}
}
else if (AFD >= 13) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (Rs < 1330) {
return 2;
}
else if (Rs >= 1330) {
return 8;
}
}
}
}
else if (MVDiff >= 12) {
if (Cs < 1085) {
if (Cs < 901) {
if (Rs < 1564) {
if (AFD < 12) {
if (diffMVdiffVal < -1) {
if (Cs < 738) {
return 4;
}
else if (Cs >= 738) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (AFD >= 12) {
if (Cs < 887) {
if (Rs < 1049) {
if (Rs < 986) {
return 2;
}
else if (Rs >= 986) {
return 1;
}
}
else if (Rs >= 1049) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (Cs >= 887) {
return 2;
}
}
}
else if (Rs >= 1564) {
if (Rs < 1812) {
return 4;
}
else if (Rs >= 1812) {
return 2;
}
}
}
else if (Cs >= 901) {
if (ssDCval < 95) {
if (Rs < 1343) {
return 4;
}
else if (Rs >= 1343) {
return 8;
}
}
else if (ssDCval >= 95) {
return 4;
}
}
}
else if (Cs >= 1085) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 51) {
if (Cs < 849) {
if (ssDCval < 82) {
if (Rs < 2054) {
if (tcor < 75) {
if (Rs < 578) {
if (Rs < 417) {
return 4;
}
else if (Rs >= 417) {
return 1;
}
}
else if (Rs >= 578) {
if (diffTSC < -1) {
if (Rs < 1307) {
return 4;
}
else if (Rs >= 1307) {
return 8;
}
}
else if (diffTSC >= -1) {
if (negBalance < 48) {
if (Rs < 1113) {
return 2;
}
else if (Rs >= 1113) {
return 8;
}
}
else if (negBalance >= 48) {
return 8;
}
}
}
}
else if (tcor >= 75) {
if (negBalance < 53) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (AFD < 11) {
if (Rs < 1295) {
return 4;
}
else if (Rs >= 1295) {
return 1;
}
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (negBalance >= 53) {
if (TSC < 8) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (TSC >= 8) {
if (Rs < 1413) {
return 4;
}
else if (Rs >= 1413) {
return 2;
}
}
}
}
}
else if (Rs >= 2054) {
return 2;
}
}
else if (ssDCval >= 82) {
if (AFD < 10) {
if (Cs < 821) {
return 8;
}
else if (Cs >= 821) {
return 4;
}
}
else if (AFD >= 10) {
if (RsDiff < 11) {
if (mcTcor < 92) {
if (mu_mv_mag_sq < 18) {
return 1;
}
else if (mu_mv_mag_sq >= 18) {
return 4;
}
}
else if (mcTcor >= 92) {
if (tcor < 89) {
if (Rs < 2019) {
return 2;
}
else if (Rs >= 2019) {
return 4;
}
}
else if (tcor >= 89) {
return 1;
}
}
}
else if (RsDiff >= 11) {
if (Rs < 2268) {
return 4;
}
else if (Rs >= 2268) {
return 8;
}
}
}
}
}
else if (Cs >= 849) {
if (Rs < 1001) {
if (ssDCval < 45) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 4;
}
}
else if (ssDCval >= 45) {
if (tcor < 65) {
return 4;
}
else if (tcor >= 65) {
if (diffRsCsdiff < 6) {
if (negBalance < 47) {
if (Rs < 815) {
return 8;
}
else if (Rs >= 815) {
return 1;
}
}
else if (negBalance >= 47) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
if (diffAFD < -1) {
if (Rs < 858) {
return 8;
}
else if (Rs >= 858) {
return 1;
}
}
else if (diffAFD >= -1) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1001) {
if (ssDCval < 81) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1412) {
if (MVDiff < 7) {
if (SC < 2325) {
return 8;
}
else if (SC >= 2325) {
return 4;
}
}
else if (MVDiff >= 7) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Cs < 1227) {
return 4;
}
else if (Cs >= 1227) {
return 2;
}
}
}
}
}
else if (Rs >= 1412) {
if (posBalance < 53) {
if (Rs < 1828) {
if (Rs < 1750) {
return 1;
}
else if (Rs >= 1750) {
return 8;
}
}
else if (Rs >= 1828) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (Cs < 903) {
if (refDCval < 59) {
return 4;
}
else if (refDCval >= 59) {
return 8;
}
}
else if (Cs >= 903) {
if (diffAFD < 2) {
if (posBalance < 56) {
if (Rs < 1821) {
if (Rs < 1634) {
return 8;
}
else if (Rs >= 1634) {
return 1;
}
}
else if (Rs >= 1821) {
return 8;
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (diffAFD >= 2) {
if (Rs < 1773) {
return 8;
}
else if (Rs >= 1773) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (tcor < 86) {
if (Rs < 1055) {
return 2;
}
else if (Rs >= 1055) {
return 8;
}
}
else if (tcor >= 86) {
if (ssDCval < 69) {
if (Rs < 1834) {
return 8;
}
else if (Rs >= 1834) {
return 2;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
else if (ssDCval >= 81) {
if (MVDiff < 25) {
if (ssDCval < 93) {
if (AFD < 13) {
if (tcor < 90) {
if (gchDC < 1) {
if (Cs < 1247) {
return 4;
}
else if (Cs >= 1247) {
return 2;
}
}
else if (gchDC >= 1) {
if (Rs < 2105) {
return 2;
}
else if (Rs >= 2105) {
return 4;
}
}
}
else if (tcor >= 90) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (AFD >= 13) {
if (TSC < 8) {
if (Rs < 1319) {
return 8;
}
else if (Rs >= 1319) {
return 4;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (ssDCval >= 93) {
if (diffMVdiffVal < -11) {
if (Rs < 1971) {
return 2;
}
else if (Rs >= 1971) {
return 4;
}
}
else if (diffMVdiffVal >= -11) {
if (posBalance < 49) {
if (Rs < 1187) {
return 8;
}
else if (Rs >= 1187) {
return 2;
}
}
else if (posBalance >= 49) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
if (RsCsDiff < 112) {
if (ssDCval < 125) {
return 8;
}
else if (ssDCval >= 125) {
if (Rs < 1786) {
return 2;
}
else if (Rs >= 1786) {
return 8;
}
}
}
else if (RsCsDiff >= 112) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 25) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (Rs < 2102) {
return 8;
}
else if (Rs >= 2102) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (Rs < 1351) {
if (refDCval < 47) {
if (mu_mv_mag_sq < 101) {
if (Rs < 690) {
if (mcTcor < 90) {
if (MVDiff < 33) {
if (Rs < 526) {
return 1;
}
else if (Rs >= 526) {
return 8;
}
}
else if (MVDiff >= 33) {
return 2;
}
}
else if (mcTcor >= 90) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
return 4;
}
}
}
else if (Rs >= 690) {
if (Cs < 673) {
if (mcTcor < 94) {
if (diffTSC < 1) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (diffTSC >= 1) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (negBalance < 41) {
return 8;
}
else if (negBalance >= 41) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (Cs >= 673) {
if (Cs < 1178) {
if (Cs < 750) {
if (MVDiff < 6) {
if (Cs < 719) {
return 8;
}
else if (Cs >= 719) {
if (Rs < 1039) {
return 4;
}
else if (Rs >= 1039) {
return 2;
}
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (Cs >= 750) {
return 8;
}
}
else if (Cs >= 1178) {
if (tcor < 70) {
if (Rs < 1073) {
return 2;
}
else if (Rs >= 1073) {
return 4;
}
}
else if (tcor >= 70) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 851) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
else if (Rs >= 851) {
if (ssDCval < 45) {
if (tcor < 80) {
if (Cs < 869) {
return 1;
}
else if (Cs >= 869) {
return 8;
}
}
else if (tcor >= 80) {
return 8;
}
}
else if (ssDCval >= 45) {
if (tcor < 87) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
}
else if (tcor >= 87) {
if (Cs < 801) {
if (Rs < 1076) {
return 8;
}
else if (Rs >= 1076) {
return 1;
}
}
else if (Cs >= 801) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 47) {
if (ssDCval < 78) {
if (Rs < 530) {
if (Cs < 830) {
if (RsDiff < 2) {
if (Rs < 319) {
if (Cs < 546) {
return 4;
}
else if (Cs >= 546) {
return 1;
}
}
else if (Rs >= 319) {
if (AFD < 3) {
if (Rs < 485) {
return 1;
}
else if (Rs >= 485) {
return 4;
}
}
else if (AFD >= 3) {
if (negBalance < 21) {
return 4;
}
else if (negBalance >= 21) {
if (diffRsCsdiff < -4) {
return 4;
}
else if (diffRsCsdiff >= -4) {
if (MVDiff < 1) {
if (Rs < 418) {
return 8;
}
else if (Rs >= 418) {
return 2;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (tcor < 84) {
if (Cs < 729) {
return 1;
}
else if (Cs >= 729) {
return 8;
}
}
else if (tcor >= 84) {
return 8;
}
}
}
}
else if (Cs >= 830) {
if (ssDCval < 53) {
if (Rs < 387) {
return 2;
}
else if (Rs >= 387) {
if (Rs < 435) {
return 8;
}
else if (Rs >= 435) {
return 1;
}
}
}
else if (ssDCval >= 53) {
return 1;
}
}
}
else if (Rs >= 530) {
if (Cs < 1098) {
if (SC < 1700) {
if (tcor < 83) {
if (mcTcor < 78) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (Rs < 922) {
return 2;
}
else if (Rs >= 922) {
return 4;
}
}
}
else if (mcTcor >= 78) {
if (Rs < 1038) {
if (Rs < 568) {
return 1;
}
else if (Rs >= 568) {
if (tcor < 80) {
if (ssDCval < 63) {
if (RsCsDiff < 29) {
return 4;
}
else if (RsCsDiff >= 29) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 1;
}
}
}
else if (ssDCval >= 63) {
return 8;
}
}
else if (tcor >= 80) {
if (Cs < 676) {
if (MVDiff < 23) {
if (refDCval < 62) {
return 8;
}
else if (refDCval >= 62) {
return 4;
}
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (Cs >= 676) {
return 4;
}
}
}
}
else if (Rs >= 1038) {
return 8;
}
}
}
else if (tcor >= 83) {
if (negBalance < 27) {
if (negBalance < 20) {
return 1;
}
else if (negBalance >= 20) {
if (SC < 1639) {
if (Cs < 546) {
return 1;
}
else if (Cs >= 546) {
return 4;
}
}
else if (SC >= 1639) {
if (Rs < 819) {
return 8;
}
else if (Rs >= 819) {
return 1;
}
}
}
}
else if (negBalance >= 27) {
if (ssDCval < 67) {
if (refDCval < 60) {
if (ssDCval < 55) {
if (Rs < 859) {
if (Rs < 690) {
if (diffRsCsdiff < 2) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
else if (Rs >= 690) {
if (Cs < 695) {
if (Cs < 661) {
return 8;
}
else if (Cs >= 661) {
return 1;
}
}
else if (Cs >= 695) {
if (MVDiff < 4) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
else if (Rs >= 859) {
if (Cs < 736) {
return 8;
}
else if (Cs >= 736) {
return 1;
}
}
}
else if (ssDCval >= 55) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Cs < 695) {
return 1;
}
else if (Cs >= 695) {
return 8;
}
}
}
}
else if (refDCval >= 60) {
if (mu_mv_mag_sq < 111) {
if (Rs < 875) {
if (posBalance < 69) {
return 8;
}
else if (posBalance >= 69) {
return 2;
}
}
else if (Rs >= 875) {
if (SC < 1603) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
else if (SC >= 1603) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 111) {
return 4;
}
}
}
else if (ssDCval >= 67) {
if (Rs < 631) {
if (Cs < 588) {
return 1;
}
else if (Cs >= 588) {
return 8;
}
}
else if (Rs >= 631) {
if (mu_mv_mag_sq < 98) {
if (mu_mv_mag_sq < 51) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
return 4;
}
}
else if (mu_mv_mag_sq >= 51) {
if (SC < 1330) {
return 4;
}
else if (SC >= 1330) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 98) {
return 2;
}
}
}
}
}
}
else if (SC >= 1700) {
if (negBalance < 44) {
if (TSC < 5) {
if (tcor < 75) {
return 4;
}
else if (tcor >= 75) {
if (negBalance < 26) {
if (RsDiff < 3) {
if (Cs < 1000) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Cs >= 1000) {
return 1;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (negBalance >= 26) {
if (gchDC < 2) {
if (mcTcor < 96) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (SC < 1915) {
if (Rs < 1068) {
return 4;
}
else if (Rs >= 1068) {
return 1;
}
}
else if (SC >= 1915) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
if (ssDCval < 61) {
if (RsDiff < 3) {
if (negBalance < 27) {
return 1;
}
else if (negBalance >= 27) {
return 8;
}
}
else if (RsDiff >= 3) {
if (Rs < 1046) {
return 1;
}
else if (Rs >= 1046) {
return 4;
}
}
}
else if (ssDCval >= 61) {
if (mu_mv_mag_sq < 77) {
return 8;
}
else if (mu_mv_mag_sq >= 77) {
if (SC < 1726) {
return 8;
}
else if (SC >= 1726) {
return 1;
}
}
}
}
}
}
else if (gchDC >= 2) {
if (Rs < 1129) {
return 1;
}
else if (Rs >= 1129) {
return 8;
}
}
}
}
}
else if (TSC >= 5) {
if (ssDCval < 55) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (SC < 1980) {
return 8;
}
else if (SC >= 1980) {
return 4;
}
}
}
else if (ssDCval >= 55) {
if (posBalance < 65) {
if (ssDCval < 63) {
if (negBalance < 33) {
if (Rs < 1261) {
return 1;
}
else if (Rs >= 1261) {
return 2;
}
}
else if (negBalance >= 33) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
}
else if (ssDCval >= 63) {
if (posBalance < 50) {
if (Rs < 969) {
return 2;
}
else if (Rs >= 969) {
if (Cs < 883) {
return 4;
}
else if (Cs >= 883) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (CsDiff < 4) {
if (Cs < 1052) {
if (posBalance < 60) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (posBalance >= 60) {
return 8;
}
}
else if (Cs >= 1052) {
if (Rs < 1206) {
return 8;
}
else if (Rs >= 1206) {
return 4;
}
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (posBalance >= 65) {
if (Rs < 1156) {
return 4;
}
else if (Rs >= 1156) {
return 8;
}
}
}
}
}
else if (negBalance >= 44) {
if (posBalance < 57) {
if (refDCval < 57) {
if (Rs < 1007) {
return 1;
}
else if (Rs >= 1007) {
return 8;
}
}
else if (refDCval >= 57) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (AFD < 10) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 931) {
return 1;
}
else if (Rs >= 931) {
return 4;
}
}
}
else if (AFD >= 10) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
else if (posBalance >= 57) {
if (negBalance < 47) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 4;
}
}
}
else if (negBalance >= 47) {
return 8;
}
}
}
}
}
else if (Cs >= 1098) {
if (ssDCval < 50) {
if (Rs < 978) {
return 4;
}
else if (Rs >= 978) {
return 8;
}
}
else if (ssDCval >= 50) {
if (diffTSC < 2) {
if (ssDCval < 73) {
if (ssDCval < 60) {
if (negBalance < 36) {
if (Rs < 847) {
return 4;
}
else if (Rs >= 847) {
if (Cs < 1213) {
return 8;
}
else if (Cs >= 1213) {
if (Rs < 1285) {
return 4;
}
else if (Rs >= 1285) {
return 8;
}
}
}
}
else if (negBalance >= 36) {
if (AFD < 10) {
if (Cs < 1099) {
return 8;
}
else if (Cs >= 1099) {
return 4;
}
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (SC < 2543) {
if (refDCval < 61) {
if (Rs < 1293) {
if (Cs < 1209) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
else if (Cs >= 1209) {
if (Rs < 687) {
return 1;
}
else if (Rs >= 687) {
return 8;
}
}
}
else if (Rs >= 1293) {
return 8;
}
}
else if (refDCval >= 61) {
if (Rs < 861) {
if (mu_mv_mag_sq < 81) {
return 1;
}
else if (mu_mv_mag_sq >= 81) {
return 8;
}
}
else if (Rs >= 861) {
if (Cs < 1173) {
if (Cs < 1138) {
if (tcor < 96) {
if (MVDiff < 2) {
if (negBalance < 34) {
if (Rs < 1015) {
return 8;
}
else if (Rs >= 1015) {
return 4;
}
}
else if (negBalance >= 34) {
return 8;
}
}
else if (MVDiff >= 2) {
if (diffRsCsdiff < 3) {
if (diffRsCsdiff < -3) {
if (Rs < 1191) {
return 4;
}
else if (Rs >= 1191) {
return 8;
}
}
else if (diffRsCsdiff >= -3) {
return 4;
}
}
else if (diffRsCsdiff >= 3) {
if (Rs < 872) {
return 4;
}
else if (Rs >= 872) {
return 8;
}
}
}
}
else if (tcor >= 96) {
return 4;
}
}
else if (Cs >= 1138) {
if (diffRsCsdiff < -11) {
return 4;
}
else if (diffRsCsdiff >= -11) {
return 8;
}
}
}
else if (Cs >= 1173) {
if (AFD < 9) {
if (negBalance < 38) {
if (Rs < 1024) {
return 4;
}
else if (Rs >= 1024) {
if (mu_mv_mag_sq < 109) {
return 8;
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 1292) {
return 4;
}
else if (Rs >= 1292) {
return 8;
}
}
}
}
else if (negBalance >= 38) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (Rs < 895) {
return 2;
}
else if (Rs >= 895) {
return 4;
}
}
}
}
else if (AFD >= 9) {
if (ssDCval < 64) {
if (Rs < 1282) {
return 1;
}
else if (Rs >= 1282) {
return 8;
}
}
else if (ssDCval >= 64) {
return 8;
}
}
}
}
}
}
else if (SC >= 2543) {
return 1;
}
}
}
else if (ssDCval >= 73) {
return 4;
}
}
else if (diffTSC >= 2) {
if (Rs < 809) {
return 1;
}
else if (Rs >= 809) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 78) {
if (CsDiff < 3) {
if (Rs < 421) {
if (TSC < 6) {
if (negBalance < 39) {
if (Rs < 344) {
return 2;
}
else if (Rs >= 344) {
return 8;
}
}
else if (negBalance >= 39) {
return 1;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (Rs >= 421) {
if (refDCval < 130) {
if (negBalance < 37) {
if (Cs < 899) {
if (TSC < 5) {
if (ssDCval < 107) {
if (Rs < 452) {
return 8;
}
else if (Rs >= 452) {
return 2;
}
}
else if (ssDCval >= 107) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (Cs >= 899) {
return 8;
}
}
else if (negBalance >= 37) {
if (Cs < 661) {
return 8;
}
else if (Cs >= 661) {
if (Cs < 1143) {
if (Rs < 472) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
return 8;
}
}
else if (Rs >= 472) {
if (tcor < 78) {
return 1;
}
else if (tcor >= 78) {
if (Rs < 751) {
if (Rs < 678) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
if (Rs < 499) {
return 1;
}
else if (Rs >= 499) {
return 8;
}
}
}
}
else if (Rs >= 678) {
if (refDCval < 87) {
return 8;
}
else if (refDCval >= 87) {
return 1;
}
}
}
else if (Rs >= 751) {
return 8;
}
}
}
}
else if (Cs >= 1143) {
return 4;
}
}
}
}
else if (refDCval >= 130) {
if (Rs < 915) {
return 8;
}
else if (Rs >= 915) {
if (ssDCval < 152) {
if (Cs < 1177) {
return 4;
}
else if (Cs >= 1177) {
return 8;
}
}
else if (ssDCval >= 152) {
return 2;
}
}
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 138) {
if (SC < 1543) {
if (tcor < 86) {
if (mcTcor < 92) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 459) {
return 8;
}
else if (Rs >= 459) {
return 2;
}
}
}
else if (mcTcor >= 92) {
return 2;
}
}
else if (tcor >= 86) {
return 8;
}
}
else if (SC >= 1543) {
if (ssDCval < 94) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 742) {
if (Rs < 698) {
return 1;
}
else if (Rs >= 698) {
return 4;
}
}
else if (Rs >= 742) {
if (diffMVdiffVal < 1) {
if (Rs < 1129) {
return 1;
}
else if (Rs >= 1129) {
if (Rs < 1243) {
return 4;
}
else if (Rs >= 1243) {
return 8;
}
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 94) {
if (negBalance < 27) {
if (SC < 1947) {
return 8;
}
else if (SC >= 1947) {
return 4;
}
}
else if (negBalance >= 27) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (CsDiff < 4) {
if (Rs < 1056) {
return 8;
}
else if (Rs >= 1056) {
return 4;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 138) {
if (refDCval < 141) {
if (mu_mv_mag_sq < 53) {
return 4;
}
else if (mu_mv_mag_sq >= 53) {
return 8;
}
}
else if (refDCval >= 141) {
if (mcTcor < 96) {
if (ssDCval < 161) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
return 8;
}
}
else if (ssDCval >= 161) {
return 2;
}
}
else if (mcTcor >= 96) {
if (Rs < 801) {
return 4;
}
else if (Rs >= 801) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 1351) {
if (Rs < 1692) {
if (ssDCval < 159) {
if (TSC < 6) {
if (ssDCval < 67) {
if (SC < 2490) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (ssDCval < 49) {
if (Rs < 1490) {
return 2;
}
else if (Rs >= 1490) {
return 4;
}
}
else if (ssDCval >= 49) {
if (ssDCval < 56) {
if (Rs < 1423) {
return 8;
}
else if (Rs >= 1423) {
if (MVDiff < 60) {
return 1;
}
else if (MVDiff >= 60) {
return 8;
}
}
}
else if (ssDCval >= 56) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (ssDCval < 61) {
if (SC < 2483) {
return 8;
}
else if (SC >= 2483) {
return 4;
}
}
else if (ssDCval >= 61) {
return 4;
}
}
}
}
}
}
else if (SC >= 2490) {
if (diffMVdiffVal < -4) {
if (Rs < 1567) {
return 8;
}
else if (Rs >= 1567) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (ssDCval >= 67) {
if (ssDCval < 73) {
if (tcor < 94) {
if (negBalance < 37) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
return 4;
}
}
else if (negBalance >= 37) {
if (Cs < 989) {
if (Rs < 1481) {
return 1;
}
else if (Rs >= 1481) {
if (Rs < 1554) {
return 4;
}
else if (Rs >= 1554) {
return 2;
}
}
}
else if (Cs >= 989) {
if (Rs < 1389) {
return 8;
}
else if (Rs >= 1389) {
return 4;
}
}
}
}
else if (tcor >= 94) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
return 4;
}
}
}
else if (ssDCval >= 73) {
if (ssDCval < 130) {
if (diffRsCsdiff < -7) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
return 8;
}
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
else if (ssDCval >= 130) {
if (CsDiff < 3) {
if (tcor < 95) {
if (Rs < 1609) {
return 4;
}
else if (Rs >= 1609) {
return 1;
}
}
else if (tcor >= 95) {
if (Rs < 1425) {
if (Rs < 1397) {
return 1;
}
else if (Rs >= 1397) {
return 4;
}
}
else if (Rs >= 1425) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 1419) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
return 4;
}
}
else if (Rs >= 1419) {
return 8;
}
}
}
}
}
}
else if (TSC >= 6) {
if (RsCsDiff < 33) {
if (Cs < 1316) {
if (TSC < 7) {
if (mu_mv_mag_sq < 61) {
if (AFD < 8) {
if (Cs < 936) {
if (Rs < 1497) {
return 1;
}
else if (Rs >= 1497) {
return 8;
}
}
else if (Cs >= 936) {
return 4;
}
}
else if (AFD >= 8) {
return 8;
}
}
else if (mu_mv_mag_sq >= 61) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (Cs >= 1316) {
return 1;
}
}
else if (RsCsDiff >= 33) {
if (ssDCval < 57) {
if (diffAFD < 3) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
return 8;
}
}
else if (diffAFD >= 3) {
return 2;
}
}
else if (ssDCval >= 57) {
if (refDCval < 100) {
if (posBalance < 56) {
if (SC < 2471) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (SC < 2373) {
if (Rs < 1381) {
return 8;
}
else if (Rs >= 1381) {
return 2;
}
}
else if (SC >= 2373) {
return 8;
}
}
}
else if (SC >= 2471) {
if (mu_mv_mag_sq < 91) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (diffMVdiffVal < 3) {
if (Rs < 1561) {
return 1;
}
else if (Rs >= 1561) {
return 2;
}
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (Rs < 1591) {
return 8;
}
else if (Rs >= 1591) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 2;
}
}
}
}
}
else if (posBalance >= 56) {
if (refDCval < 97) {
if (ssDCval < 62) {
if (Rs < 1437) {
return 1;
}
else if (Rs >= 1437) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
else if (ssDCval >= 62) {
if (Cs < 1272) {
if (Rs < 1668) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Rs < 1510) {
return 1;
}
else if (Rs >= 1510) {
return 8;
}
}
}
else if (Rs >= 1668) {
if (Cs < 1233) {
return 4;
}
else if (Cs >= 1233) {
return 8;
}
}
}
else if (Cs >= 1272) {
if (RsCsDiff < 46) {
return 4;
}
else if (RsCsDiff >= 46) {
return 8;
}
}
}
}
else if (refDCval >= 97) {
return 2;
}
}
}
else if (refDCval >= 100) {
if (MVDiff < 3) {
if (Cs < 890) {
return 1;
}
else if (Cs >= 890) {
return 8;
}
}
else if (MVDiff >= 3) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 159) {
return 2;
}
}
else if (Rs >= 1692) {
if (refDCval < 105) {
if (RsDiff < 7) {
if (ssDCval < 62) {
if (Cs < 699) {
if (Rs < 1854) {
return 2;
}
else if (Rs >= 1854) {
return 1;
}
}
else if (Cs >= 699) {
return 8;
}
}
else if (ssDCval >= 62) {
if (ssDCval < 70) {
if (mu_mv_mag_sq < 42) {
if (Cs < 994) {
return 1;
}
else if (Cs >= 994) {
return 4;
}
}
else if (mu_mv_mag_sq >= 42) {
if (CsDiff < 5) {
if (RsDiff < 5) {
if (tcor < 93) {
if (Rs < 2112) {
return 2;
}
else if (Rs >= 2112) {
return 8;
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (RsDiff >= 5) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
return 8;
}
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (refDCval < 91) {
if (Cs < 780) {
if (Rs < 1728) {
return 8;
}
else if (Rs >= 1728) {
return 1;
}
}
else if (Cs >= 780) {
if (mu_mv_mag_sq < 107) {
if (negBalance < 48) {
if (diffMVdiffVal < 2) {
if (ssDCval < 84) {
if (tcor < 93) {
if (Cs < 1277) {
return 1;
}
else if (Cs >= 1277) {
if (Rs < 2087) {
return 8;
}
else if (Rs >= 2087) {
return 1;
}
}
}
else if (tcor >= 93) {
return 1;
}
}
else if (ssDCval >= 84) {
if (Rs < 1845) {
return 8;
}
else if (Rs >= 1845) {
if (Rs < 2159) {
return 2;
}
else if (Rs >= 2159) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (ssDCval < 76) {
if (Rs < 2101) {
return 1;
}
else if (Rs >= 2101) {
return 4;
}
}
else if (ssDCval >= 76) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (SC < 3398) {
if (TSC < 7) {
if (diffRsCsdiff < 6) {
return 8;
}
else if (diffRsCsdiff >= 6) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (SC >= 3398) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 107) {
return 8;
}
}
}
else if (refDCval >= 91) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (TSC < 8) {
if (mu_mv_mag_sq < 54) {
if (Cs < 1192) {
return 4;
}
else if (Cs >= 1192) {
return 2;
}
}
else if (mu_mv_mag_sq >= 54) {
if (ssDCval < 99) {
if (Cs < 951) {
return 8;
}
else if (Cs >= 951) {
return 2;
}
}
else if (ssDCval >= 99) {
return 8;
}
}
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (TSC < 6) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (TSC >= 6) {
if (refDCval < 55) {
if (Rs < 1896) {
return 8;
}
else if (Rs >= 1896) {
return 4;
}
}
else if (refDCval >= 55) {
if (Cs < 1088) {
if (Rs < 2126) {
if (negBalance < 45) {
if (SC < 2876) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
else if (SC >= 2876) {
if (Rs < 1923) {
return 8;
}
else if (Rs >= 1923) {
if (Rs < 2035) {
return 4;
}
else if (Rs >= 2035) {
return 1;
}
}
}
}
else if (negBalance >= 45) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (Rs >= 2126) {
return 2;
}
}
else if (Cs >= 1088) {
if (ssDCval < 85) {
if (Rs < 2150) {
return 8;
}
else if (Rs >= 2150) {
if (Rs < 2255) {
if (Rs < 2164) {
return 2;
}
else if (Rs >= 2164) {
return 8;
}
}
else if (Rs >= 2255) {
return 4;
}
}
}
else if (ssDCval >= 85) {
if (TSC < 8) {
if (mcTcor < 93) {
return 2;
}
else if (mcTcor >= 93) {
return 4;
}
}
else if (TSC >= 8) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 105) {
if (SC < 2768) {
if (mu_mv_mag_sq < 108) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
return 2;
}
}
else if (mu_mv_mag_sq >= 108) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (SC >= 2768) {
if (TSC < 10) {
if (negBalance < 56) {
if (diffRsCsdiff < 38) {
if (SC < 3580) {
if (mu_mv_mag_sq < 118) {
return 8;
}
else if (mu_mv_mag_sq >= 118) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 1;
}
}
}
else if (SC >= 3580) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
return 1;
}
}
}
else if (diffRsCsdiff >= 38) {
return 2;
}
}
else if (negBalance >= 56) {
return 2;
}
}
else if (TSC >= 10) {
return 1;
}
}
}
}
}
}
}
}
else if (AFD >= 14) {
if (refDCval < 114) {
if (MVDiff < 10) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 81) {
if (mcTcor < 93) {
if (RsCsDiff < 80) {
return 8;
}
else if (RsCsDiff >= 80) {
if (Rs < 1202) {
return 8;
}
else if (Rs >= 1202) {
return 4;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 2061) {
return 1;
}
else if (Rs >= 2061) {
return 2;
}
}
}
else if (ssDCval >= 81) {
if (AFD < 15) {
if (diffTSC < 1) {
if (Cs < 1300) {
return 8;
}
else if (Cs >= 1300) {
return 1;
}
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (AFD >= 15) {
if (CsDiff < 6) {
if (ssDCval < 85) {
if (Rs < 1699) {
return 8;
}
else if (Rs >= 1699) {
return 4;
}
}
else if (ssDCval >= 85) {
return 4;
}
}
else if (CsDiff >= 6) {
if (Rs < 1933) {
return 8;
}
else if (Rs >= 1933) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsDiff < 10) {
if (Rs < 382) {
return 2;
}
else if (Rs >= 382) {
if (ssDCval < 102) {
if (mcTcor < 94) {
if (ssDCval < 60) {
if (Rs < 1191) {
return 4;
}
else if (Rs >= 1191) {
return 8;
}
}
else if (ssDCval >= 60) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 874) {
if (mu_mv_mag_sq < 15) {
return 8;
}
else if (mu_mv_mag_sq >= 15) {
return 1;
}
}
else if (Rs >= 874) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (diffMVdiffVal < -5) {
return 8;
}
else if (diffMVdiffVal >= -5) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (Rs < 719) {
return 1;
}
else if (Rs >= 719) {
return 4;
}
}
}
}
else if (RsDiff >= 10) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
if (Rs < 1572) {
return 1;
}
else if (Rs >= 1572) {
return 2;
}
}
}
}
}
else if (MVDiff >= 10) {
if (RsDiff < 12) {
if (mcTcor < 81) {
if (TSC < 10) {
if (ssDCval < 43) {
if (Rs < 614) {
return 2;
}
else if (Rs >= 614) {
return 8;
}
}
else if (ssDCval >= 43) {
if (tcor < 57) {
if (gchDC < 5) {
if (MVDiff < 154) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 584) {
return 2;
}
else if (Rs >= 584) {
return 4;
}
}
}
}
else if (MVDiff >= 154) {
return 1;
}
}
else if (gchDC >= 5) {
if (Cs < 831) {
if (Rs < 783) {
return 2;
}
else if (Rs >= 783) {
return 1;
}
}
else if (Cs >= 831) {
if (Cs < 1074) {
return 8;
}
else if (Cs >= 1074) {
return 4;
}
}
}
}
else if (tcor >= 57) {
if (posBalance < 56) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 4;
}
}
else if (posBalance >= 56) {
if (Rs < 1371) {
return 4;
}
else if (Rs >= 1371) {
return 8;
}
}
}
}
}
else if (TSC >= 10) {
if (RsCsDiff < 108) {
if (Rs < 680) {
if (ssDCval < 67) {
if (MVDiff < 96) {
return 8;
}
else if (MVDiff >= 96) {
if (refDCval < 42) {
if (Rs < 449) {
return 2;
}
else if (Rs >= 449) {
return 8;
}
}
else if (refDCval >= 42) {
return 1;
}
}
}
else if (ssDCval >= 67) {
if (Cs < 1055) {
if (AFD < 79) {
return 4;
}
else if (AFD >= 79) {
return 2;
}
}
else if (Cs >= 1055) {
return 8;
}
}
}
else if (Rs >= 680) {
if (refDCval < 98) {
if (AFD < 25) {
if (CsDiff < 6) {
if (MVDiff < 12) {
if (Rs < 1681) {
return 4;
}
else if (Rs >= 1681) {
return 8;
}
}
else if (MVDiff >= 12) {
if (Cs < 555) {
if (Rs < 1144) {
return 8;
}
else if (Rs >= 1144) {
return 1;
}
}
else if (Cs >= 555) {
return 8;
}
}
}
else if (CsDiff >= 6) {
if (Rs < 935) {
return 8;
}
else if (Rs >= 935) {
if (Rs < 1159) {
if (Cs < 996) {
return 8;
}
else if (Cs >= 996) {
return 4;
}
}
else if (Rs >= 1159) {
return 8;
}
}
}
}
else if (AFD >= 25) {
if (Cs < 826) {
return 8;
}
else if (Cs >= 826) {
if (Cs < 947) {
return 4;
}
else if (Cs >= 947) {
return 2;
}
}
}
}
else if (refDCval >= 98) {
return 4;
}
}
}
else if (RsCsDiff >= 108) {
if (negBalance < 55) {
if (Rs < 1417) {
if (mu_mv_mag_sq < 7) {
if (diffRsCsdiff < -8) {
return 1;
}
else if (diffRsCsdiff >= -8) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
if (Rs < 1146) {
return 8;
}
else if (Rs >= 1146) {
return 1;
}
}
}
else if (Rs >= 1417) {
return 8;
}
}
else if (negBalance >= 55) {
if (mu_mv_mag_sq < 6) {
if (Cs < 984) {
return 8;
}
else if (Cs >= 984) {
return 4;
}
}
else if (mu_mv_mag_sq >= 6) {
if (CsDiff < 12) {
if (gchDC < 25) {
return 1;
}
else if (gchDC >= 25) {
return 8;
}
}
else if (CsDiff >= 12) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 81) {
if (Rs < 1339) {
if (Rs < 1104) {
if (Rs < 253) {
if (AFD < 25) {
return 4;
}
else if (AFD >= 25) {
if (Rs < 205) {
return 1;
}
else if (Rs >= 205) {
return 2;
}
}
}
else if (Rs >= 253) {
if (MVDiff < 13) {
if (posBalance < 64) {
if (negBalance < 55) {
if (Rs < 757) {
return 4;
}
else if (Rs >= 757) {
return 8;
}
}
else if (negBalance >= 55) {
if (negBalance < 57) {
return 4;
}
else if (negBalance >= 57) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
}
}
else if (posBalance >= 64) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
}
else if (MVDiff >= 13) {
if (negBalance < 95) {
if (ssDCval < 62) {
if (mu_mv_mag_sq < 3) {
if (CsDiff < 5) {
if (Rs < 769) {
return 1;
}
else if (Rs >= 769) {
return 8;
}
}
else if (CsDiff >= 5) {
if (diffRsCsdiff < 19) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
if (Rs < 1065) {
return 4;
}
else if (Rs >= 1065) {
return 2;
}
}
}
else if (diffRsCsdiff >= 19) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (posBalance < 51) {
if (RsDiff < 5) {
if (Cs < 902) {
return 8;
}
else if (Cs >= 902) {
return 2;
}
}
else if (RsDiff >= 5) {
if (Rs < 817) {
return 4;
}
else if (Rs >= 817) {
return 1;
}
}
}
else if (posBalance >= 51) {
if (negBalance < 52) {
if (SC < 939) {
return 8;
}
else if (SC >= 939) {
if (diffMVdiffVal < -24) {
return 8;
}
else if (diffMVdiffVal >= -24) {
if (refDCval < 49) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 8;
}
}
else if (refDCval >= 49) {
if (diffMVdiffVal < -3) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 4;
}
}
}
}
}
else if (negBalance >= 52) {
if (diffMVdiffVal < 1) {
if (SC < 1690) {
return 8;
}
else if (SC >= 1690) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
if (MVDiff < 29) {
return 4;
}
else if (MVDiff >= 29) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 62) {
if (refDCval < 88) {
if (RsCsDiff < 146) {
if (refDCval < 86) {
if (mcTcor < 96) {
if (tcor < 84) {
if (Rs < 1067) {
if (Rs < 1033) {
if (posBalance < 50) {
if (RsCsDiff < 68) {
if (Rs < 336) {
return 8;
}
else if (Rs >= 336) {
if (Rs < 793) {
return 1;
}
else if (Rs >= 793) {
return 4;
}
}
}
else if (RsCsDiff >= 68) {
return 4;
}
}
else if (posBalance >= 50) {
if (Cs < 1039) {
if (mu_mv_mag_sq < 10) {
if (diffMVdiffVal < -16) {
if (Cs < 819) {
return 1;
}
else if (Cs >= 819) {
if (negBalance < 53) {
if (Rs < 806) {
return 4;
}
else if (Rs >= 806) {
return 1;
}
}
else if (negBalance >= 53) {
return 2;
}
}
}
else if (diffMVdiffVal >= -16) {
if (Cs < 978) {
if (posBalance < 70) {
if (RsDiff < 8) {
if (diffMVdiffVal < -3) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (AFD < 15) {
if (Rs < 663) {
return 8;
}
else if (Rs >= 663) {
return 1;
}
}
else if (AFD >= 15) {
if (refDCval < 74) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (negBalance < 55) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
else if (negBalance >= 55) {
return 1;
}
}
}
else if (refDCval >= 74) {
return 4;
}
}
}
}
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (posBalance >= 70) {
return 8;
}
}
else if (Cs >= 978) {
if (diffTSC < 1) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
if (Cs < 988) {
return 1;
}
else if (Cs >= 988) {
return 2;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 897) {
return 4;
}
else if (Rs >= 897) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (diffMVdiffVal < 33) {
if (MVDiff < 47) {
if (diffMVdiffVal < 4) {
if (Rs < 716) {
if (AFD < 16) {
if (Rs < 532) {
return 4;
}
else if (Rs >= 532) {
return 2;
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (Rs >= 716) {
if (RsCsDiff < 77) {
if (SC < 1689) {
return 1;
}
else if (SC >= 1689) {
return 2;
}
}
else if (RsCsDiff >= 77) {
return 2;
}
}
}
else if (diffMVdiffVal >= 4) {
return 1;
}
}
else if (MVDiff >= 47) {
if (ssDCval < 74) {
if (diffAFD < 4) {
return 4;
}
else if (diffAFD >= 4) {
return 2;
}
}
else if (ssDCval >= 74) {
return 2;
}
}
}
else if (diffMVdiffVal >= 33) {
if (Cs < 649) {
return 8;
}
else if (Cs >= 649) {
return 1;
}
}
}
}
else if (Cs >= 1039) {
if (refDCval < 75) {
if (Rs < 1006) {
return 2;
}
else if (Rs >= 1006) {
return 1;
}
}
else if (refDCval >= 75) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 2;
}
}
}
}
}
else if (Rs >= 1033) {
if (Rs < 1064) {
if (refDCval < 82) {
return 2;
}
else if (refDCval >= 82) {
return 1;
}
}
else if (Rs >= 1064) {
return 4;
}
}
}
else if (Rs >= 1067) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
if (Rs < 1100) {
return 4;
}
else if (Rs >= 1100) {
return 8;
}
}
}
}
else if (tcor >= 84) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (Rs < 829) {
return 4;
}
else if (Rs >= 829) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (posBalance < 61) {
if (MVDiff < 26) {
return 2;
}
else if (MVDiff >= 26) {
if (Cs < 922) {
if (Rs < 660) {
return 1;
}
else if (Rs >= 660) {
if (RsCsDiff < 138) {
return 8;
}
else if (RsCsDiff >= 138) {
return 1;
}
}
}
else if (Cs >= 922) {
return 2;
}
}
}
else if (posBalance >= 61) {
return 1;
}
}
}
else if (refDCval >= 86) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (RsCsDiff >= 146) {
if (tcor < 43) {
if (Rs < 611) {
return 4;
}
else if (Rs >= 611) {
return 1;
}
}
else if (tcor >= 43) {
if (refDCval < 77) {
return 8;
}
else if (refDCval >= 77) {
return 4;
}
}
}
}
else if (refDCval >= 88) {
if (MVDiff < 36) {
if (Cs < 706) {
if (Rs < 998) {
if (refDCval < 111) {
return 8;
}
else if (refDCval >= 111) {
return 4;
}
}
else if (Rs >= 998) {
return 4;
}
}
else if (Cs >= 706) {
if (tcor < 64) {
if (Rs < 617) {
return 2;
}
else if (Rs >= 617) {
return 1;
}
}
else if (tcor >= 64) {
if (tcor < 78) {
if (posBalance < 60) {
if (SC < 1786) {
if (Rs < 694) {
return 2;
}
else if (Rs >= 694) {
return 8;
}
}
else if (SC >= 1786) {
return 4;
}
}
else if (posBalance >= 60) {
return 8;
}
}
else if (tcor >= 78) {
if (negBalance < 53) {
if (Rs < 952) {
if (Rs < 810) {
return 1;
}
else if (Rs >= 810) {
return 2;
}
}
else if (Rs >= 952) {
return 1;
}
}
else if (negBalance >= 53) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 36) {
if (CsDiff < 8) {
if (gchDC < 8) {
if (RsCsDiff < 94) {
if (negBalance < 65) {
return 2;
}
else if (negBalance >= 65) {
if (Rs < 595) {
return 4;
}
else if (Rs >= 595) {
if (Rs < 852) {
return 8;
}
else if (Rs >= 852) {
return 2;
}
}
}
}
else if (RsCsDiff >= 94) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
if (Rs < 900) {
return 1;
}
else if (Rs >= 900) {
return 4;
}
}
}
}
else if (gchDC >= 8) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (diffAFD < -2) {
if (Rs < 515) {
return 8;
}
else if (Rs >= 515) {
return 1;
}
}
else if (diffAFD >= -2) {
if (Rs < 536) {
return 1;
}
else if (Rs >= 536) {
return 2;
}
}
}
}
}
else if (CsDiff >= 8) {
if (Cs < 987) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
return 1;
}
}
else if (Cs >= 987) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
if (Rs < 822) {
return 4;
}
else if (Rs >= 822) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 95) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 2;
}
}
}
}
}
else if (Rs >= 1104) {
if (Cs < 732) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 57) {
if (Rs < 1182) {
return 8;
}
else if (Rs >= 1182) {
return 2;
}
}
else if (posBalance >= 57) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (mu_mv_mag_sq < 11) {
if (ssDCval < 78) {
return 4;
}
else if (ssDCval >= 78) {
return 1;
}
}
else if (mu_mv_mag_sq >= 11) {
return 4;
}
}
}
else if (Cs >= 732) {
if (refDCval < 67) {
if (diffRsCsdiff < 18) {
if (mu_mv_mag_sq < 13) {
if (diffMVdiffVal < 5) {
if (Rs < 1145) {
return 1;
}
else if (Rs >= 1145) {
return 4;
}
}
else if (diffMVdiffVal >= 5) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (MVDiff < 74) {
return 8;
}
else if (MVDiff >= 74) {
if (Rs < 1221) {
return 1;
}
else if (Rs >= 1221) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 18) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
if (Rs < 1193) {
return 4;
}
else if (Rs >= 1193) {
return 8;
}
}
}
}
else if (refDCval >= 67) {
if (Cs < 1092) {
if (mcTcor < 91) {
if (tcor < 74) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
if (refDCval < 80) {
return 2;
}
else if (refDCval >= 80) {
return 4;
}
}
}
else if (tcor >= 74) {
if (SC < 2239) {
return 8;
}
else if (SC >= 2239) {
return 2;
}
}
}
else if (mcTcor >= 91) {
if (posBalance < 50) {
if (Cs < 1075) {
return 2;
}
else if (Cs >= 1075) {
return 8;
}
}
else if (posBalance >= 50) {
if (AFD < 24) {
if (refDCval < 84) {
if (CsDiff < 6) {
if (refDCval < 76) {
if (Cs < 851) {
return 4;
}
else if (Cs >= 851) {
return 8;
}
}
else if (refDCval >= 76) {
return 4;
}
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (refDCval >= 84) {
if (RsDiff < 8) {
if (Rs < 1203) {
if (Rs < 1138) {
return 4;
}
else if (Rs >= 1138) {
return 1;
}
}
else if (Rs >= 1203) {
return 8;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
}
else if (AFD >= 24) {
if (ssDCval < 83) {
return 2;
}
else if (ssDCval >= 83) {
return 4;
}
}
}
}
}
else if (Cs >= 1092) {
if (tcor < 73) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (tcor >= 73) {
if (refDCval < 82) {
if (refDCval < 73) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
if (Rs < 1261) {
return 4;
}
else if (Rs >= 1261) {
return 8;
}
}
}
else if (refDCval >= 73) {
return 4;
}
}
else if (refDCval >= 82) {
if (Rs < 1185) {
return 2;
}
else if (Rs >= 1185) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1339) {
if (negBalance < 45) {
if (diffMVdiffVal < -16) {
if (SC < 2916) {
return 1;
}
else if (SC >= 2916) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (diffMVdiffVal >= -16) {
if (Cs < 1064) {
if (tcor < 61) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Rs < 1717) {
return 1;
}
else if (Rs >= 1717) {
return 2;
}
}
}
else if (tcor >= 61) {
return 2;
}
}
else if (Cs >= 1064) {
if (posBalance < 72) {
if (AFD < 19) {
if (TSC < 7) {
if (Rs < 2274) {
return 1;
}
else if (Rs >= 2274) {
return 4;
}
}
else if (TSC >= 7) {
if (diffRsCsdiff < -31) {
return 1;
}
else if (diffRsCsdiff >= -31) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
return 8;
}
}
}
}
else if (AFD >= 19) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
return 4;
}
}
}
else if (posBalance >= 72) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 8;
}
}
}
}
}
else if (negBalance >= 45) {
if (AFD < 19) {
if (ssDCval < 96) {
if (Cs < 1036) {
if (MVDiff < 28) {
if (RsDiff < 9) {
if (mcTcor < 88) {
if (AFD < 15) {
if (CsDiff < 5) {
if (Rs < 1484) {
return 8;
}
else if (Rs >= 1484) {
return 1;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (AFD >= 15) {
return 8;
}
}
else if (mcTcor >= 88) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (diffRsCsdiff < -6) {
if (Rs < 1943) {
return 8;
}
else if (Rs >= 1943) {
return 2;
}
}
else if (diffRsCsdiff >= -6) {
if (ssDCval < 75) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (ssDCval >= 75) {
return 4;
}
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 80) {
return 2;
}
else if (refDCval >= 80) {
return 4;
}
}
}
else if (MVDiff >= 28) {
if (diffAFD < -3) {
if (mu_mv_mag_sq < 2) {
return 2;
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1886) {
if (Cs < 595) {
return 1;
}
else if (Cs >= 595) {
return 4;
}
}
else if (Rs >= 1886) {
return 8;
}
}
}
else if (diffAFD >= -3) {
if (MVDiff < 45) {
if (ssDCval < 65) {
if (Rs < 1591) {
return 4;
}
else if (Rs >= 1591) {
return 2;
}
}
else if (ssDCval >= 65) {
if (Cs < 706) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 1680) {
return 4;
}
else if (Rs >= 1680) {
return 2;
}
}
}
else if (Cs >= 706) {
return 8;
}
}
}
else if (MVDiff >= 45) {
return 8;
}
}
}
}
else if (Cs >= 1036) {
if (tcor < 80) {
if (refDCval < 77) {
if (negBalance < 51) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
if (Rs < 1819) {
return 4;
}
else if (Rs >= 1819) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 1;
}
}
}
else if (refDCval >= 77) {
if (Cs < 1236) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
if (Rs < 1815) {
return 1;
}
else if (Rs >= 1815) {
return 4;
}
}
}
else if (Cs >= 1236) {
if (Rs < 1444) {
return 2;
}
else if (Rs >= 1444) {
return 8;
}
}
}
}
else if (tcor >= 80) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 96) {
if (mcTcor < 92) {
if (tcor < 71) {
return 4;
}
else if (tcor >= 71) {
if (TSC < 10) {
if (Rs < 1653) {
return 8;
}
else if (Rs >= 1653) {
return 2;
}
}
else if (TSC >= 10) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (tcor < 84) {
if (Rs < 1995) {
return 2;
}
else if (Rs >= 1995) {
return 4;
}
}
else if (tcor >= 84) {
return 8;
}
}
}
}
else if (AFD >= 19) {
if (RsCsDiff < 103) {
if (MVDiff < 31) {
if (Rs < 1442) {
return 8;
}
else if (Rs >= 1442) {
return 2;
}
}
else if (MVDiff >= 31) {
if (Rs < 1795) {
return 1;
}
else if (Rs >= 1795) {
if (Rs < 2262) {
return 8;
}
else if (Rs >= 2262) {
return 2;
}
}
}
}
else if (RsCsDiff >= 103) {
if (SC < 3384) {
if (mu_mv_mag_sq < 33) {
if (Rs < 1457) {
if (diffAFD < 1) {
if (Rs < 1381) {
return 2;
}
else if (Rs >= 1381) {
if (Rs < 1445) {
return 4;
}
else if (Rs >= 1445) {
return 2;
}
}
}
else if (diffAFD >= 1) {
if (gchDC < 4) {
if (Rs < 1437) {
return 4;
}
else if (Rs >= 1437) {
return 8;
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (Rs >= 1457) {
if (AFD < 23) {
if (TSC < 12) {
if (mu_mv_mag_sq < 5) {
if (Cs < 854) {
if (Rs < 1736) {
return 2;
}
else if (Rs >= 1736) {
return 8;
}
}
else if (Cs >= 854) {
return 4;
}
}
else if (mu_mv_mag_sq >= 5) {
if (mcTcor < 93) {
if (ssDCval < 98) {
if (refDCval < 72) {
if (Rs < 1523) {
return 8;
}
else if (Rs >= 1523) {
if (Rs < 1638) {
return 2;
}
else if (Rs >= 1638) {
return 4;
}
}
}
else if (refDCval >= 72) {
if (Cs < 874) {
if (Rs < 1686) {
return 4;
}
else if (Rs >= 1686) {
return 8;
}
}
else if (Cs >= 874) {
if (posBalance < 68) {
return 8;
}
else if (posBalance >= 68) {
return 4;
}
}
}
}
else if (ssDCval >= 98) {
if (Rs < 1776) {
return 4;
}
else if (Rs >= 1776) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (mu_mv_mag_sq < 10) {
return 8;
}
else if (mu_mv_mag_sq >= 10) {
return 1;
}
}
}
}
else if (TSC >= 12) {
if (SC < 2685) {
return 4;
}
else if (SC >= 2685) {
return 1;
}
}
}
else if (AFD >= 23) {
if (Cs < 937) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
if (MVDiff < 63) {
return 1;
}
else if (MVDiff >= 63) {
return 8;
}
}
}
else if (Cs >= 937) {
if (mcTcor < 89) {
if (Cs < 1166) {
return 8;
}
else if (Cs >= 1166) {
return 4;
}
}
else if (mcTcor >= 89) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (ssDCval < 90) {
return 2;
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (SC >= 3384) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
return 2;
}
}
}
}
}
}
}
}
else if (RsDiff >= 12) {
if (refDCval < 86) {
if (AFD < 22) {
if (Cs < 1029) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
if (Rs < 1818) {
return 8;
}
else if (Rs >= 1818) {
return 2;
}
}
}
else if (Cs >= 1029) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
if (Cs < 1070) {
if (Rs < 2149) {
return 4;
}
else if (Rs >= 2149) {
return 1;
}
}
else if (Cs >= 1070) {
if (Rs < 1406) {
return 2;
}
else if (Rs >= 1406) {
return 8;
}
}
}
}
}
else if (AFD >= 22) {
if (RsCsDiff < 201) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 1;
}
}
else if (RsCsDiff >= 201) {
if (AFD < 25) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
return 2;
}
}
else if (AFD >= 25) {
if (diffTSC < 3) {
if (Rs < 1952) {
if (RsDiff < 15) {
return 4;
}
else if (RsDiff >= 15) {
return 2;
}
}
else if (Rs >= 1952) {
return 2;
}
}
else if (diffTSC >= 3) {
return 2;
}
}
}
}
}
else if (refDCval >= 86) {
if (gchDC < 2) {
if (mcTcor < 65) {
return 1;
}
else if (mcTcor >= 65) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (ssDCval < 102) {
if (ssDCval < 92) {
return 4;
}
else if (ssDCval >= 92) {
if (MVDiff < 35) {
return 1;
}
else if (MVDiff >= 35) {
return 2;
}
}
}
else if (ssDCval >= 102) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (Cs < 677) {
if (diffRsCsdiff < 65) {
return 4;
}
else if (diffRsCsdiff >= 65) {
return 1;
}
}
else if (Cs >= 677) {
if (mcTcor < 94) {
if (Cs < 1083) {
return 8;
}
else if (Cs >= 1083) {
if (SC < 3409) {
return 2;
}
else if (SC >= 3409) {
if (Rs < 2247) {
return 8;
}
else if (Rs >= 2247) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 1722) {
if (AFD < 34) {
return 4;
}
else if (AFD >= 34) {
return 2;
}
}
else if (Rs >= 1722) {
if (Cs < 1190) {
return 2;
}
else if (Cs >= 1190) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (MVDiff < 13) {
if (tcor < 67) {
if (Rs < 805) {
return 2;
}
else if (Rs >= 805) {
return 4;
}
}
else if (tcor >= 67) {
if (diffAFD < 5) {
if (refDCval < 122) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (Rs < 1056) {
return 8;
}
else if (Rs >= 1056) {
return 2;
}
}
}
else if (refDCval >= 122) {
return 8;
}
}
else if (diffAFD >= 5) {
return 2;
}
}
}
else if (MVDiff >= 13) {
if (SC < 2240) {
if (MVDiff < 29) {
if (tcor < 73) {
if (gchDC < 6) {
return 2;
}
else if (gchDC >= 6) {
return 8;
}
}
else if (tcor >= 73) {
if (Rs < 541) {
return 4;
}
else if (Rs >= 541) {
if (diffMVdiffVal < 11) {
if (ssDCval < 167) {
return 8;
}
else if (ssDCval >= 167) {
return 2;
}
}
else if (diffMVdiffVal >= 11) {
return 2;
}
}
}
}
else if (MVDiff >= 29) {
if (diffRsCsdiff < -20) {
if (diffRsCsdiff < -41) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (Cs < 805) {
return 8;
}
else if (Cs >= 805) {
return 1;
}
}
}
else if (diffRsCsdiff >= -41) {
if (AFD < 30) {
return 8;
}
else if (AFD >= 30) {
if (Rs < 1089) {
return 4;
}
else if (Rs >= 1089) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= -20) {
if (posBalance < 40) {
if (Cs < 558) {
return 2;
}
else if (Cs >= 558) {
if (MVDiff < 253) {
return 8;
}
else if (MVDiff >= 253) {
return 2;
}
}
}
else if (posBalance >= 40) {
if (SC < 1749) {
if (MVDiff < 167) {
return 2;
}
else if (MVDiff >= 167) {
return 1;
}
}
else if (SC >= 1749) {
if (AFD < 30) {
if (ssDCval < 151) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 8;
}
}
}
else if (ssDCval >= 151) {
if (Rs < 865) {
return 8;
}
else if (Rs >= 865) {
return 2;
}
}
}
else if (AFD >= 30) {
return 2;
}
}
}
}
}
}
else if (SC >= 2240) {
if (AFD < 35) {
if (negBalance < 45) {
if (Cs < 852) {
return 1;
}
else if (Cs >= 852) {
if (tcor < 53) {
return 1;
}
else if (tcor >= 53) {
if (Rs < 2230) {
return 2;
}
else if (Rs >= 2230) {
return 1;
}
}
}
}
else if (negBalance >= 45) {
if (mu_mv_mag_sq < 16) {
if (Rs < 1635) {
if (AFD < 19) {
if (Rs < 1566) {
return 8;
}
else if (Rs >= 1566) {
if (Rs < 1619) {
return 1;
}
else if (Rs >= 1619) {
return 8;
}
}
}
else if (AFD >= 19) {
if (ssDCval < 128) {
return 8;
}
else if (ssDCval >= 128) {
if (negBalance < 59) {
return 4;
}
else if (negBalance >= 59) {
return 2;
}
}
}
}
else if (Rs >= 1635) {
if (tcor < 63) {
if (Cs < 717) {
return 8;
}
else if (Cs >= 717) {
return 4;
}
}
else if (tcor >= 63) {
if (diffTSC < -4) {
return 8;
}
else if (diffTSC >= -4) {
if (ssDCval < 151) {
if (SC < 2275) {
return 8;
}
else if (SC >= 2275) {
return 1;
}
}
else if (ssDCval >= 151) {
if (Rs < 2218) {
return 8;
}
else if (Rs >= 2218) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (mu_mv_mag_sq < 179) {
if (ssDCval < 165) {
if (negBalance < 55) {
if (RsCsDiff < 195) {
if (AFD < 15) {
if (Rs < 1310) {
return 1;
}
else if (Rs >= 1310) {
return 8;
}
}
else if (AFD >= 15) {
if (mu_mv_mag_sq < 40) {
return 1;
}
else if (mu_mv_mag_sq >= 40) {
if (Rs < 1690) {
return 8;
}
else if (Rs >= 1690) {
return 1;
}
}
}
}
else if (RsCsDiff >= 195) {
return 8;
}
}
else if (negBalance >= 55) {
if (mcTcor < 90) {
if (Rs < 1550) {
if (Rs < 1328) {
return 8;
}
else if (Rs >= 1328) {
return 4;
}
}
else if (Rs >= 1550) {
return 8;
}
}
else if (mcTcor >= 90) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (diffTSC < -1) {
if (Rs < 1987) {
return 8;
}
else if (Rs >= 1987) {
return 1;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
}
}
}
else if (ssDCval >= 165) {
return 2;
}
}
else if (mu_mv_mag_sq >= 179) {
return 1;
}
}
}
}
else if (AFD >= 35) {
if (SC < 2360) {
return 4;
}
else if (SC >= 2360) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (SC < 2156) {
if (Rs < 1002) {
if (refDCval < 63) {
if (AFD < 1) {
if (Rs < 849) {
return 1;
}
else if (Rs >= 849) {
return 8;
}
}
else if (AFD >= 1) {
if (Rs < 829) {
if (Rs < 548) {
if (SC < 1090) {
if (ssDCval < 49) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (ssDCval >= 49) {
if (MVDiff < 4) {
if (Rs < 268) {
return 4;
}
else if (Rs >= 268) {
return 8;
}
}
else if (MVDiff >= 4) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (diffRsCsdiff < 1) {
if (Cs < 687) {
return 1;
}
else if (Cs >= 687) {
return 2;
}
}
else if (diffRsCsdiff >= 1) {
if (RsCsDiff < 11) {
return 4;
}
else if (RsCsDiff >= 11) {
return 1;
}
}
}
}
}
}
else if (SC >= 1090) {
if (negBalance < 21) {
if (Rs < 423) {
return 2;
}
else if (Rs >= 423) {
if (Rs < 519) {
return 4;
}
else if (Rs >= 519) {
return 2;
}
}
}
else if (negBalance >= 21) {
if (tcor < 91) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
if (Rs < 436) {
return 4;
}
else if (Rs >= 436) {
return 2;
}
}
}
else if (tcor >= 91) {
if (posBalance < 44) {
if (negBalance < 30) {
if (Rs < 523) {
if (SC < 1460) {
return 2;
}
else if (SC >= 1460) {
return 1;
}
}
else if (Rs >= 523) {
return 4;
}
}
else if (negBalance >= 30) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
}
else if (posBalance >= 44) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
}
}
else if (Rs >= 548) {
if (MVDiff < 2) {
if (Cs < 931) {
if (tcor < 97) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (gchDC < 1) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (tcor >= 97) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (RsDiff < 2) {
if (ssDCval < 48) {
if (Rs < 719) {
return 4;
}
else if (Rs >= 719) {
return 2;
}
}
else if (ssDCval >= 48) {
return 8;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (Cs >= 931) {
if (Rs < 616) {
if (Cs < 1067) {
return 1;
}
else if (Cs >= 1067) {
return 4;
}
}
else if (Rs >= 616) {
if (negBalance < 34) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 1;
}
}
else if (negBalance >= 34) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 41) {
if (Rs < 653) {
return 1;
}
else if (Rs >= 653) {
return 2;
}
}
else if (ssDCval >= 41) {
if (AFD < 9) {
if (ssDCval < 54) {
if (MVDiff < 3) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
else if (MVDiff >= 3) {
if (RsDiff < 2) {
if (negBalance < 29) {
if (Rs < 650) {
return 4;
}
else if (Rs >= 650) {
if (Rs < 754) {
return 1;
}
else if (Rs >= 754) {
return 8;
}
}
}
else if (negBalance >= 29) {
if (Cs < 869) {
return 8;
}
else if (Cs >= 869) {
return 4;
}
}
}
else if (RsDiff >= 2) {
if (diffMVdiffVal < -2) {
if (Cs < 649) {
return 1;
}
else if (Cs >= 649) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
}
}
else if (ssDCval >= 54) {
if (negBalance < 23) {
if (Rs < 657) {
return 4;
}
else if (Rs >= 657) {
return 1;
}
}
else if (negBalance >= 23) {
if (MVDiff < 11) {
if (diffAFD < -2) {
if (Rs < 774) {
return 8;
}
else if (Rs >= 774) {
return 4;
}
}
else if (diffAFD >= -2) {
if (posBalance < 54) {
if (negBalance < 25) {
if (Rs < 737) {
return 4;
}
else if (Rs >= 737) {
return 1;
}
}
else if (negBalance >= 25) {
if (mu_mv_mag_sq < 92) {
return 4;
}
else if (mu_mv_mag_sq >= 92) {
if (Rs < 792) {
return 8;
}
else if (Rs >= 792) {
return 4;
}
}
}
}
else if (posBalance >= 54) {
if (Rs < 679) {
return 1;
}
else if (Rs >= 679) {
return 8;
}
}
}
}
else if (MVDiff >= 11) {
if (ssDCval < 61) {
if (Rs < 668) {
return 1;
}
else if (Rs >= 668) {
return 8;
}
}
else if (ssDCval >= 61) {
return 4;
}
}
}
}
}
else if (AFD >= 9) {
if (Cs < 666) {
return 1;
}
else if (Cs >= 666) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 686) {
return 8;
}
else if (Rs >= 686) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 829) {
if (ssDCval < 49) {
if (Cs < 559) {
return 1;
}
else if (Cs >= 559) {
return 8;
}
}
else if (ssDCval >= 49) {
if (negBalance < 45) {
if (mu_mv_mag_sq < 97) {
if (Cs < 1067) {
if (MVDiff < 9) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Rs < 948) {
return 4;
}
else if (Rs >= 948) {
return 1;
}
}
}
}
else if (MVDiff >= 9) {
if (Rs < 889) {
return 8;
}
else if (Rs >= 889) {
return 1;
}
}
}
else if (Cs >= 1067) {
return 4;
}
}
else if (mu_mv_mag_sq >= 97) {
if (mcTcor < 98) {
if (Cs < 1070) {
if (Rs < 983) {
return 8;
}
else if (Rs >= 983) {
return 1;
}
}
else if (Cs >= 1070) {
return 1;
}
}
else if (mcTcor >= 98) {
if (diffRsCsdiff < -5) {
return 8;
}
else if (diffRsCsdiff >= -5) {
if (negBalance < 14) {
return 8;
}
else if (negBalance >= 14) {
if (ssDCval < 52) {
if (Cs < 1109) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (Cs >= 1109) {
return 1;
}
}
else if (ssDCval >= 52) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 45) {
return 4;
}
}
}
}
}
else if (refDCval >= 63) {
if (CsDiff < 1) {
if (Cs < 704) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (ssDCval < 72) {
if (Rs < 715) {
if (Rs < 527) {
return 4;
}
else if (Rs >= 527) {
return 8;
}
}
else if (Rs >= 715) {
if (AFD < 1) {
if (Rs < 926) {
return 2;
}
else if (Rs >= 926) {
return 8;
}
}
else if (AFD >= 1) {
return 2;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 703) {
if (Rs < 374) {
return 1;
}
else if (Rs >= 374) {
return 2;
}
}
else if (Rs >= 703) {
return 8;
}
}
}
}
else if (Cs >= 704) {
if (ssDCval < 70) {
if (Rs < 903) {
if (Rs < 713) {
if (SC < 1248) {
return 4;
}
else if (SC >= 1248) {
return 8;
}
}
else if (Rs >= 713) {
if (mu_mv_mag_sq < 112) {
return 8;
}
else if (mu_mv_mag_sq >= 112) {
if (tcor < 100) {
return 4;
}
else if (tcor >= 100) {
return 8;
}
}
}
}
else if (Rs >= 903) {
if (Cs < 804) {
return 2;
}
else if (Cs >= 804) {
if (Cs < 1021) {
return 1;
}
else if (Cs >= 1021) {
return 8;
}
}
}
}
else if (ssDCval >= 70) {
if (Rs < 643) {
return 2;
}
else if (Rs >= 643) {
return 1;
}
}
}
}
else if (CsDiff >= 1) {
if (ssDCval < 85) {
if (CsDiff < 4) {
if (ssDCval < 70) {
if (negBalance < 27) {
if (tcor < 98) {
if (negBalance < 26) {
return 4;
}
else if (negBalance >= 26) {
if (CsDiff < 2) {
if (Rs < 606) {
return 1;
}
else if (Rs >= 606) {
return 8;
}
}
else if (CsDiff >= 2) {
return 4;
}
}
}
else if (tcor >= 98) {
if (Rs < 574) {
if (Rs < 406) {
return 4;
}
else if (Rs >= 406) {
return 8;
}
}
else if (Rs >= 574) {
if (Rs < 952) {
return 2;
}
else if (Rs >= 952) {
return 1;
}
}
}
}
else if (negBalance >= 27) {
if (Cs < 582) {
if (MVDiff < 3) {
if (posBalance < 40) {
if (Rs < 776) {
return 4;
}
else if (Rs >= 776) {
return 1;
}
}
else if (posBalance >= 40) {
return 2;
}
}
else if (MVDiff >= 3) {
if (negBalance < 36) {
return 4;
}
else if (negBalance >= 36) {
if (Rs < 536) {
if (Cs < 575) {
return 1;
}
else if (Cs >= 575) {
return 4;
}
}
else if (Rs >= 536) {
return 4;
}
}
}
}
else if (Cs >= 582) {
if (RsCsDiff < 7) {
if (negBalance < 33) {
if (Rs < 667) {
if (Cs < 634) {
return 1;
}
else if (Cs >= 634) {
if (Rs < 474) {
return 4;
}
else if (Rs >= 474) {
return 8;
}
}
}
else if (Rs >= 667) {
if (mcTcor < 98) {
if (Cs < 720) {
return 1;
}
else if (Cs >= 720) {
return 4;
}
}
else if (mcTcor >= 98) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
}
}
else if (negBalance >= 33) {
if (mu_mv_mag_sq < 106) {
if (mcTcor < 99) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (Rs < 733) {
if (Rs < 587) {
return 8;
}
else if (Rs >= 587) {
return 1;
}
}
else if (Rs >= 733) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 798) {
return 1;
}
else if (Rs >= 798) {
return 2;
}
}
}
}
else if (RsCsDiff >= 7) {
if (tcor < 94) {
if (negBalance < 41) {
return 2;
}
else if (negBalance >= 41) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 1;
}
}
}
else if (tcor >= 94) {
if (ssDCval < 69) {
if (Rs < 912) {
return 4;
}
else if (Rs >= 912) {
if (Rs < 934) {
return 2;
}
else if (Rs >= 934) {
if (Rs < 957) {
return 4;
}
else if (Rs >= 957) {
return 1;
}
}
}
}
else if (ssDCval >= 69) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 892) {
return 2;
}
else if (Rs >= 892) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 70) {
if (ssDCval < 71) {
if (tcor < 97) {
if (RsCsDiff < 11) {
if (Cs < 1080) {
return 4;
}
else if (Cs >= 1080) {
return 1;
}
}
else if (RsCsDiff >= 11) {
if (Rs < 550) {
return 1;
}
else if (Rs >= 550) {
return 2;
}
}
}
else if (tcor >= 97) {
return 2;
}
}
else if (ssDCval >= 71) {
if (Rs < 843) {
if (Cs < 690) {
if (diffMVdiffVal < 2) {
if (Rs < 723) {
if (Rs < 411) {
return 4;
}
else if (Rs >= 411) {
return 8;
}
}
else if (Rs >= 723) {
if (Rs < 800) {
return 4;
}
else if (Rs >= 800) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 332) {
return 1;
}
else if (Rs >= 332) {
if (Rs < 662) {
return 4;
}
else if (Rs >= 662) {
return 2;
}
}
}
}
else if (Cs >= 690) {
if (ssDCval < 75) {
if (Rs < 783) {
if (Cs < 1101) {
if (Cs < 715) {
return 8;
}
else if (Cs >= 715) {
return 1;
}
}
else if (Cs >= 1101) {
return 8;
}
}
else if (Rs >= 783) {
return 4;
}
}
else if (ssDCval >= 75) {
if (Rs < 760) {
if (Rs < 563) {
return 1;
}
else if (Rs >= 563) {
if (negBalance < 36) {
if (ssDCval < 83) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
else if (ssDCval >= 83) {
return 2;
}
}
else if (negBalance >= 36) {
if (AFD < 7) {
if (MVDiff < 3) {
if (Rs < 685) {
return 8;
}
else if (Rs >= 685) {
return 4;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (AFD >= 7) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
}
else if (Rs >= 760) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
else if (Rs >= 843) {
if (Rs < 985) {
if (tcor < 96) {
if (Rs < 902) {
return 4;
}
else if (Rs >= 902) {
if (Rs < 925) {
return 8;
}
else if (Rs >= 925) {
return 2;
}
}
}
else if (tcor >= 96) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
if (Rs < 978) {
return 8;
}
else if (Rs >= 978) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
return 4;
}
}
}
}
}
else if (Rs >= 985) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
return 4;
}
}
}
}
}
}
else if (CsDiff >= 4) {
if (Cs < 704) {
if (diffMVdiffVal < -2) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 354) {
return 2;
}
else if (Rs >= 354) {
return 8;
}
}
}
else if (Cs >= 704) {
if (MVDiff < 5) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
else if (MVDiff >= 5) {
if (Rs < 506) {
if (Cs < 1242) {
if (posBalance < 48) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (posBalance >= 48) {
return 2;
}
}
else if (Cs >= 1242) {
return 1;
}
}
else if (Rs >= 506) {
if (negBalance < 49) {
if (RsCsDiff < 23) {
if (Rs < 740) {
return 2;
}
else if (Rs >= 740) {
return 4;
}
}
else if (RsCsDiff >= 23) {
return 2;
}
}
else if (negBalance >= 49) {
if (Rs < 581) {
return 2;
}
else if (Rs >= 581) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (TSC < 4) {
if (Rs < 723) {
if (MVDiff < 4) {
if (tcor < 99) {
return 2;
}
else if (tcor >= 99) {
if (Cs < 796) {
return 1;
}
else if (Cs >= 796) {
return 2;
}
}
}
else if (MVDiff >= 4) {
return 2;
}
}
else if (Rs >= 723) {
return 4;
}
}
else if (TSC >= 4) {
if (SC < 1361) {
if (Rs < 609) {
return 1;
}
else if (Rs >= 609) {
return 2;
}
}
else if (SC >= 1361) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1002) {
if (mu_mv_mag_sq < 124) {
if (ssDCval < 51) {
if (AFD < 5) {
if (Rs < 1271) {
return 8;
}
else if (Rs >= 1271) {
if (Rs < 1335) {
return 8;
}
else if (Rs >= 1335) {
return 8;
}
}
}
else if (AFD >= 5) {
if (Rs < 1370) {
return 1;
}
else if (Rs >= 1370) {
return 4;
}
}
}
else if (ssDCval >= 51) {
if (RsDiff < 3) {
if (ssDCval < 70) {
if (MVDiff < 11) {
if (SC < 1783) {
if (AFD < 3) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (SC >= 1783) {
if (posBalance < 79) {
if (negBalance < 14) {
if (Rs < 1171) {
return 1;
}
else if (Rs >= 1171) {
return 8;
}
}
else if (negBalance >= 14) {
if (Cs < 1088) {
if (diffAFD < -2) {
if (Rs < 1104) {
return 8;
}
else if (Rs >= 1104) {
return 1;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (Cs >= 1088) {
if (Cs < 1112) {
return 8;
}
else if (Cs >= 1112) {
return 1;
}
}
}
}
else if (posBalance >= 79) {
return 8;
}
}
}
else if (MVDiff >= 11) {
if (posBalance < 44) {
if (Cs < 907) {
return 1;
}
else if (Cs >= 907) {
return 4;
}
}
else if (posBalance >= 44) {
if (Rs < 1005) {
return 1;
}
else if (Rs >= 1005) {
return 8;
}
}
}
}
else if (ssDCval >= 70) {
if (Rs < 1082) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
if (Cs < 985) {
return 8;
}
else if (Cs >= 985) {
if (Rs < 1023) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
return 4;
}
}
else if (Rs >= 1023) {
return 4;
}
}
}
}
else if (Rs >= 1082) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Cs < 590) {
if (Rs < 1277) {
return 1;
}
else if (Rs >= 1277) {
return 8;
}
}
else if (Cs >= 590) {
return 1;
}
}
}
}
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 59) {
if (MVDiff < 12) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (ssDCval < 73) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (ssDCval >= 73) {
if (Rs < 1142) {
return 8;
}
else if (Rs >= 1142) {
return 1;
}
}
}
}
else if (MVDiff >= 12) {
if (mu_mv_mag_sq < 3) {
if (SC < 1828) {
return 1;
}
else if (SC >= 1828) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (SC < 1755) {
return 4;
}
else if (SC >= 1755) {
if (SC < 1878) {
return 8;
}
else if (SC >= 1878) {
if (Rs < 1100) {
return 1;
}
else if (Rs >= 1100) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (CsDiff < 3) {
if (MVDiff < 6) {
if (diffRsCsdiff < 3) {
if (Rs < 1168) {
return 2;
}
else if (Rs >= 1168) {
return 8;
}
}
else if (diffRsCsdiff >= 3) {
if (Rs < 1058) {
return 8;
}
else if (Rs >= 1058) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (CsDiff >= 3) {
if (MVDiff < 11) {
if (negBalance < 61) {
return 8;
}
else if (negBalance >= 61) {
return 4;
}
}
else if (MVDiff >= 11) {
if (Rs < 1062) {
return 4;
}
else if (Rs >= 1062) {
if (Rs < 1139) {
return 2;
}
else if (Rs >= 1139) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 1132) {
if (posBalance < 41) {
if (Cs < 584) {
return 2;
}
else if (Cs >= 584) {
return 8;
}
}
else if (posBalance >= 41) {
if (ssDCval < 59) {
if (Rs < 1088) {
return 8;
}
else if (Rs >= 1088) {
return 1;
}
}
else if (ssDCval >= 59) {
if (negBalance < 22) {
if (SC < 1819) {
return 2;
}
else if (SC >= 1819) {
return 4;
}
}
else if (negBalance >= 22) {
return 4;
}
}
}
}
else if (Rs >= 1132) {
if (Rs < 1395) {
if (Cs < 736) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
if (Cs < 707) {
if (SC < 1952) {
if (CsDiff < 1) {
if (Cs < 687) {
return 8;
}
else if (Cs >= 687) {
return 1;
}
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (SC >= 1952) {
return 1;
}
}
else if (Cs >= 707) {
if (Rs < 1246) {
return 4;
}
else if (Rs >= 1246) {
return 4;
}
}
}
}
else if (Cs >= 736) {
if (Cs < 887) {
return 8;
}
else if (Cs >= 887) {
if (Rs < 1244) {
return 1;
}
else if (Rs >= 1244) {
return 8;
}
}
}
}
else if (Rs >= 1395) {
if (Cs < 670) {
return 8;
}
else if (Cs >= 670) {
return 1;
}
}
}
}
}
}
else if (SC >= 2156) {
if (Rs < 1787) {
if (Cs < 1092) {
if (mu_mv_mag_sq < 122) {
if (SC < 2426) {
if (Cs < 983) {
if (tcor < 99) {
if (mu_mv_mag_sq < 92) {
if (negBalance < 38) {
if (Cs < 616) {
if (Cs < 601) {
return 1;
}
else if (Cs >= 601) {
return 2;
}
}
else if (Cs >= 616) {
return 1;
}
}
else if (negBalance >= 38) {
if (posBalance < 48) {
if (Rs < 1449) {
return 4;
}
else if (Rs >= 1449) {
return 1;
}
}
else if (posBalance >= 48) {
if (Cs < 715) {
return 1;
}
else if (Cs >= 715) {
if (Rs < 1325) {
return 1;
}
else if (Rs >= 1325) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (refDCval < 72) {
if (Cs < 763) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
return 8;
}
}
else if (Cs >= 763) {
if (Cs < 951) {
return 8;
}
else if (Cs >= 951) {
return 1;
}
}
}
else if (refDCval >= 72) {
if (posBalance < 53) {
if (Rs < 1755) {
return 8;
}
else if (Rs >= 1755) {
return 4;
}
}
else if (posBalance >= 53) {
return 1;
}
}
}
}
else if (tcor >= 99) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 8;
}
}
}
else if (Cs >= 983) {
if (negBalance < 46) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (negBalance >= 46) {
return 4;
}
}
}
else if (SC >= 2426) {
if (negBalance < 20) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
return 1;
}
}
else if (negBalance >= 20) {
if (mu_mv_mag_sq < 42) {
if (ssDCval < 60) {
if (Rs < 1673) {
return 8;
}
else if (Rs >= 1673) {
return 4;
}
}
else if (ssDCval >= 60) {
if (SC < 2444) {
return 2;
}
else if (SC >= 2444) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (diffAFD < -2) {
if (Rs < 1565) {
return 8;
}
else if (Rs >= 1565) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 4;
}
}
}
else if (diffAFD >= -2) {
if (Cs < 810) {
if (Rs < 1771) {
return 4;
}
else if (Rs >= 1771) {
return 1;
}
}
else if (Cs >= 810) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (Cs < 1023) {
return 8;
}
else if (Cs >= 1023) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (ssDCval < 71) {
if (RsDiff < 1) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (RsDiff >= 1) {
if (ssDCval < 49) {
if (RsDiff < 2) {
if (posBalance < 31) {
return 4;
}
else if (posBalance >= 31) {
if (posBalance < 37) {
if (SC < 2799) {
if (Rs < 1724) {
return 8;
}
else if (Rs >= 1724) {
if (Rs < 1731) {
return 4;
}
else if (Rs >= 1731) {
return 8;
}
}
}
else if (SC >= 2799) {
if (negBalance < 19) {
return 8;
}
else if (negBalance >= 19) {
return 8;
}
}
}
else if (posBalance >= 37) {
if (negBalance < 17) {
return 4;
}
else if (negBalance >= 17) {
return 4;
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 1653) {
if (Rs < 1478) {
return 4;
}
else if (Rs >= 1478) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
return 4;
}
}
}
else if (Rs >= 1653) {
return 4;
}
}
}
else if (ssDCval >= 49) {
if (Rs < 1307) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 4;
}
}
else if (Rs >= 1307) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 1340) {
return 1;
}
else if (Rs >= 1340) {
if (Rs < 1581) {
return 8;
}
else if (Rs >= 1581) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (Cs >= 1092) {
if (RsDiff < 1) {
if (mu_mv_mag_sq < 103) {
if (Rs < 1033) {
return 8;
}
else if (Rs >= 1033) {
return 4;
}
}
else if (mu_mv_mag_sq >= 103) {
if (ssDCval < 55) {
if (Rs < 1353) {
return 4;
}
else if (Rs >= 1353) {
return 8;
}
}
else if (ssDCval >= 55) {
if (Rs < 1432) {
return 1;
}
else if (Rs >= 1432) {
return 4;
}
}
}
}
else if (RsDiff >= 1) {
if (AFD < 3) {
if (ssDCval < 52) {
if (Rs < 1623) {
if (Cs < 1314) {
if (Cs < 1137) {
return 4;
}
else if (Cs >= 1137) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 1104) {
return 8;
}
else if (Rs >= 1104) {
return 4;
}
}
}
}
else if (Cs >= 1314) {
if (Rs < 1389) {
return 4;
}
else if (Rs >= 1389) {
return 8;
}
}
}
else if (Rs >= 1623) {
if (refDCval < 47) {
return 4;
}
else if (refDCval >= 47) {
if (SC < 2954) {
if (tcor < 99) {
if (Rs < 1782) {
if (Cs < 1192) {
if (negBalance < 21) {
if (diffRsCsdiff < 2) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (Rs < 1696) {
if (Rs < 1683) {
if (Rs < 1679) {
return 4;
}
else if (Rs >= 1679) {
return 4;
}
}
else if (Rs >= 1683) {
return 8;
}
}
else if (Rs >= 1696) {
if (Rs < 1720) {
return 4;
}
else if (Rs >= 1720) {
if (Cs < 1167) {
if (SC < 2862) {
if (Rs < 1745) {
return 4;
}
else if (Rs >= 1745) {
return 4;
}
}
else if (SC >= 2862) {
return 4;
}
}
else if (Cs >= 1167) {
return 8;
}
}
}
}
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
else if (negBalance >= 21) {
if (posBalance < 32) {
if (Rs < 1740) {
return 4;
}
else if (Rs >= 1740) {
return 4;
}
}
else if (posBalance >= 32) {
return 4;
}
}
}
else if (Cs >= 1192) {
return 8;
}
}
else if (Rs >= 1782) {
return 8;
}
}
else if (tcor >= 99) {
if (Rs < 1747) {
if (RsDiff < 2) {
if (Rs < 1683) {
if (Rs < 1662) {
return 4;
}
else if (Rs >= 1662) {
return 4;
}
}
else if (Rs >= 1683) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Cs < 1151) {
if (Cs < 1104) {
return 8;
}
else if (Cs >= 1104) {
return 4;
}
}
else if (Cs >= 1151) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (Rs >= 1747) {
if (Rs < 1759) {
return 8;
}
else if (Rs >= 1759) {
if (Rs < 1781) {
if (posBalance < 29) {
return 8;
}
else if (posBalance >= 29) {
if (Cs < 1138) {
if (SC < 2904) {
if (Cs < 1105) {
if (Rs < 1771) {
return 8;
}
else if (Rs >= 1771) {
return 4;
}
}
else if (Cs >= 1105) {
if (Rs < 1762) {
return 8;
}
else if (Rs >= 1762) {
return 8;
}
}
}
else if (SC >= 2904) {
return 8;
}
}
else if (Cs >= 1138) {
return 4;
}
}
}
else if (Rs >= 1781) {
return 8;
}
}
}
}
}
else if (SC >= 2954) {
return 4;
}
}
}
}
else if (ssDCval >= 52) {
if (ssDCval < 62) {
if (SC < 2497) {
return 1;
}
else if (SC >= 2497) {
return 4;
}
}
else if (ssDCval >= 62) {
if (negBalance < 35) {
if (negBalance < 29) {
if (Cs < 1104) {
return 8;
}
else if (Cs >= 1104) {
return 4;
}
}
else if (negBalance >= 29) {
if (tcor < 98) {
if (Rs < 1331) {
return 1;
}
else if (Rs >= 1331) {
return 8;
}
}
else if (tcor >= 98) {
if (ssDCval < 69) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (Rs < 1249) {
return 4;
}
else if (Rs >= 1249) {
return 1;
}
}
}
else if (ssDCval >= 69) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
return 4;
}
}
}
}
}
else if (negBalance >= 35) {
return 1;
}
}
}
}
else if (AFD >= 3) {
if (SC < 2708) {
if (mu_mv_mag_sq < 113) {
if (RsCsDiff < 7) {
if (Rs < 1039) {
if (SC < 2170) {
if (Rs < 964) {
return 2;
}
else if (Rs >= 964) {
return 8;
}
}
else if (SC >= 2170) {
return 4;
}
}
else if (Rs >= 1039) {
return 8;
}
}
else if (RsCsDiff >= 7) {
if (MVDiff < 3) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
if (mu_mv_mag_sq < 95) {
if (Rs < 982) {
return 4;
}
else if (Rs >= 982) {
return 8;
}
}
else if (mu_mv_mag_sq >= 95) {
if (Cs < 1260) {
if (Rs < 1070) {
return 8;
}
else if (Rs >= 1070) {
if (Cs < 1234) {
return 4;
}
else if (Cs >= 1234) {
return 8;
}
}
}
else if (Cs >= 1260) {
return 4;
}
}
}
}
else if (MVDiff >= 3) {
if (negBalance < 40) {
if (mu_mv_mag_sq < 56) {
return 2;
}
else if (mu_mv_mag_sq >= 56) {
if (ssDCval < 73) {
if (Cs < 1280) {
return 4;
}
else if (Cs >= 1280) {
return 8;
}
}
else if (ssDCval >= 73) {
if (Rs < 1311) {
return 2;
}
else if (Rs >= 1311) {
return 8;
}
}
}
}
else if (negBalance >= 40) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (negBalance < 34) {
if (ssDCval < 58) {
if (Rs < 1112) {
return 8;
}
else if (Rs >= 1112) {
return 4;
}
}
else if (ssDCval >= 58) {
if (posBalance < 50) {
if (Rs < 999) {
return 4;
}
else if (Rs >= 999) {
return 8;
}
}
else if (posBalance >= 50) {
return 4;
}
}
}
else if (negBalance >= 34) {
if (ssDCval < 70) {
if (diffRsCsdiff < -17) {
return 4;
}
else if (diffRsCsdiff >= -17) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Cs < 1239) {
return 8;
}
else if (Cs >= 1239) {
return 4;
}
}
}
}
}
}
else if (SC >= 2708) {
if (posBalance < 36) {
if (Rs < 1751) {
return 1;
}
else if (Rs >= 1751) {
return 2;
}
}
else if (posBalance >= 36) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1787) {
if (TSC < 4) {
if (refDCval < 83) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 109) {
return 4;
}
else if (mu_mv_mag_sq >= 109) {
if (Cs < 1134) {
if (Rs < 2019) {
if (ssDCval < 48) {
if (negBalance < 17) {
return 8;
}
else if (negBalance >= 17) {
if (Cs < 1040) {
return 4;
}
else if (Cs >= 1040) {
if (posBalance < 33) {
if (Rs < 1806) {
return 8;
}
else if (Rs >= 1806) {
return 8;
}
}
else if (posBalance >= 33) {
return 4;
}
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (Rs >= 2019) {
return 8;
}
}
else if (Cs >= 1134) {
if (Rs < 2177) {
return 4;
}
else if (Rs >= 2177) {
return 8;
}
}
}
}
else if (ssDCval >= 56) {
if (refDCval < 71) {
if (Cs < 1187) {
if (posBalance < 36) {
if (Rs < 1932) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
if (Rs < 1869) {
if (mu_mv_mag_sq < 124) {
return 4;
}
else if (mu_mv_mag_sq >= 124) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
if (mu_mv_mag_sq < 125) {
if (Rs < 1845) {
return 1;
}
else if (Rs >= 1845) {
return 4;
}
}
else if (mu_mv_mag_sq >= 125) {
return 1;
}
}
}
}
else if (Rs >= 1869) {
if (Cs < 1085) {
return 4;
}
else if (Cs >= 1085) {
return 1;
}
}
}
}
else if (Rs >= 1932) {
if (SC < 3101) {
return 4;
}
else if (SC >= 3101) {
if (Rs < 2117) {
return 1;
}
else if (Rs >= 2117) {
return 8;
}
}
}
}
else if (posBalance >= 36) {
if (Cs < 1000) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
if (Rs < 2185) {
return 4;
}
else if (Rs >= 2185) {
return 1;
}
}
}
else if (Cs >= 1000) {
if (mu_mv_mag_sq < 98) {
return 4;
}
else if (mu_mv_mag_sq >= 98) {
if (refDCval < 57) {
return 1;
}
else if (refDCval >= 57) {
if (negBalance < 20) {
if (Rs < 1831) {
return 1;
}
else if (Rs >= 1831) {
return 4;
}
}
else if (negBalance >= 20) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1187) {
if (mu_mv_mag_sq < 124) {
return 2;
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 2279) {
return 2;
}
else if (Rs >= 2279) {
return 8;
}
}
}
}
else if (refDCval >= 71) {
if (posBalance < 47) {
if (SC < 3043) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (Rs < 1917) {
return 8;
}
else if (Rs >= 1917) {
return 1;
}
}
}
else if (SC >= 3043) {
if (Cs < 980) {
return 1;
}
else if (Cs >= 980) {
return 8;
}
}
}
else if (posBalance >= 47) {
if (Cs < 861) {
if (Rs < 1883) {
return 1;
}
else if (Rs >= 1883) {
return 4;
}
}
else if (Cs >= 861) {
return 8;
}
}
}
}
}
else if (refDCval >= 83) {
if (mu_mv_mag_sq < 98) {
if (AFD < 6) {
if (Rs < 2048) {
return 4;
}
else if (Rs >= 2048) {
return 8;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 98) {
if (Cs < 939) {
return 4;
}
else if (Cs >= 939) {
return 8;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 932) {
if (ssDCval < 71) {
if (RsDiff < 4) {
if (Rs < 1807) {
return 8;
}
else if (Rs >= 1807) {
return 1;
}
}
else if (RsDiff >= 4) {
if (MVDiff < 1) {
if (Rs < 2069) {
return 2;
}
else if (Rs >= 2069) {
return 1;
}
}
else if (MVDiff >= 1) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 4;
}
}
}
}
else if (ssDCval >= 71) {
if (mu_mv_mag_sq < 93) {
if (RsCsDiff < 56) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 2;
}
}
else if (RsCsDiff >= 56) {
if (MVDiff < 35) {
return 2;
}
else if (MVDiff >= 35) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
}
}
else if (Cs >= 932) {
if (Cs < 1002) {
if (Rs < 2160) {
if (Cs < 965) {
if (Cs < 938) {
return 4;
}
else if (Cs >= 938) {
return 1;
}
}
else if (Cs >= 965) {
return 4;
}
}
else if (Rs >= 2160) {
if (Rs < 2345) {
return 1;
}
else if (Rs >= 2345) {
return 8;
}
}
}
else if (Cs >= 1002) {
if (diffRsCsdiff < -6) {
if (negBalance < 44) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (negBalance >= 44) {
if (Cs < 1316) {
return 8;
}
else if (Cs >= 1316) {
return 4;
}
}
}
else if (diffRsCsdiff >= -6) {
if (Rs < 2278) {
return 8;
}
else if (Rs >= 2278) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 100) {
if (SC < 1428) {
if (negBalance < 69) {
if (MVDiff < 4) {
if (Cs < 721) {
if (ssDCval < 115) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
else if (ssDCval >= 115) {
if (Rs < 632) {
if (Rs < 513) {
if (Rs < 346) {
return 2;
}
else if (Rs >= 346) {
if (Cs < 558) {
return 8;
}
else if (Cs >= 558) {
return 1;
}
}
}
else if (Rs >= 513) {
if (mcTcor < 98) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (mcTcor >= 98) {
return 2;
}
}
}
else if (Rs >= 632) {
if (Cs < 570) {
if (Rs < 710) {
return 2;
}
else if (Rs >= 710) {
return 8;
}
}
else if (Cs >= 570) {
return 4;
}
}
}
}
else if (Cs >= 721) {
if (ssDCval < 99) {
if (Rs < 507) {
return 8;
}
else if (Rs >= 507) {
return 2;
}
}
else if (ssDCval >= 99) {
if (Rs < 463) {
return 4;
}
else if (Rs >= 463) {
return 1;
}
}
}
}
else if (MVDiff >= 4) {
if (TSC < 4) {
if (Rs < 598) {
if (AFD < 3) {
if (posBalance < 35) {
return 4;
}
else if (posBalance >= 35) {
return 2;
}
}
else if (AFD >= 3) {
if (mu_mv_mag_sq < 89) {
if (posBalance < 57) {
if (ssDCval < 144) {
if (posBalance < 39) {
if (Cs < 626) {
return 8;
}
else if (Cs >= 626) {
return 2;
}
}
else if (posBalance >= 39) {
return 2;
}
}
else if (ssDCval >= 144) {
return 8;
}
}
else if (posBalance >= 57) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 89) {
return 1;
}
}
}
else if (Rs >= 598) {
if (MVDiff < 5) {
if (Rs < 634) {
return 2;
}
else if (Rs >= 634) {
return 8;
}
}
else if (MVDiff >= 5) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 630) {
return 1;
}
else if (Rs >= 630) {
if (Rs < 659) {
return 4;
}
else if (Rs >= 659) {
return 2;
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 59) {
if (tcor < 80) {
if (Rs < 309) {
return 8;
}
else if (Rs >= 309) {
return 1;
}
}
else if (tcor >= 80) {
if (RsCsDiff < 7) {
if (Rs < 536) {
return 8;
}
else if (Rs >= 536) {
return 2;
}
}
else if (RsCsDiff >= 7) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (MVDiff < 16) {
if (Rs < 739) {
return 4;
}
else if (Rs >= 739) {
return 2;
}
}
else if (MVDiff >= 16) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
return 2;
}
}
}
}
}
else if (negBalance >= 69) {
if (Rs < 603) {
return 8;
}
else if (Rs >= 603) {
if (Rs < 609) {
return 1;
}
else if (Rs >= 609) {
return 8;
}
}
}
}
else if (SC >= 1428) {
if (RsCsDiff < 70) {
if (MVDiff < 4) {
if (Rs < 661) {
if (Rs < 489) {
return 8;
}
else if (Rs >= 489) {
if (Cs < 866) {
return 8;
}
else if (Cs >= 866) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Rs < 621) {
return 8;
}
else if (Rs >= 621) {
return 1;
}
}
}
}
}
else if (Rs >= 661) {
if (ssDCval < 93) {
if (MVDiff < 2) {
if (Rs < 2029) {
return 4;
}
else if (Rs >= 2029) {
return 1;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 93) {
if (ssDCval < 122) {
if (Rs < 1198) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (refDCval < 121) {
return 4;
}
else if (refDCval >= 121) {
return 8;
}
}
}
else if (Rs >= 1198) {
return 8;
}
}
else if (ssDCval >= 122) {
if (SC < 2534) {
if (Cs < 641) {
return 1;
}
else if (Cs >= 641) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 8;
}
}
}
else if (SC >= 2534) {
if (RsCsDiff < 15) {
return 4;
}
else if (RsCsDiff >= 15) {
if (Rs < 2247) {
return 2;
}
else if (Rs >= 2247) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (negBalance < 30) {
if (ssDCval < 107) {
return 1;
}
else if (ssDCval >= 107) {
if (Rs < 1551) {
return 8;
}
else if (Rs >= 1551) {
if (Rs < 2155) {
return 4;
}
else if (Rs >= 2155) {
if (Rs < 2344) {
return 8;
}
else if (Rs >= 2344) {
return 1;
}
}
}
}
}
else if (negBalance >= 30) {
if (mu_mv_mag_sq < 28) {
if (ssDCval < 103) {
if (TSC < 7) {
if (negBalance < 54) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
if (MVDiff < 14) {
if (Cs < 1128) {
return 4;
}
else if (Cs >= 1128) {
return 2;
}
}
else if (MVDiff >= 14) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
return 1;
}
}
else if (TSC >= 7) {
if (ssDCval < 99) {
return 8;
}
else if (ssDCval >= 99) {
if (Rs < 2114) {
return 1;
}
else if (Rs >= 2114) {
return 4;
}
}
}
}
else if (ssDCval >= 103) {
if (ssDCval < 154) {
if (SC < 1963) {
if (tcor < 95) {
if (tcor < 82) {
if (Rs < 763) {
return 1;
}
else if (Rs >= 763) {
return 8;
}
}
else if (tcor >= 82) {
if (tcor < 94) {
if (SC < 1533) {
return 4;
}
else if (SC >= 1533) {
return 1;
}
}
else if (tcor >= 94) {
if (Rs < 979) {
return 1;
}
else if (Rs >= 979) {
return 8;
}
}
}
}
else if (tcor >= 95) {
return 8;
}
}
else if (SC >= 1963) {
if (diffAFD < -2) {
if (Cs < 994) {
return 1;
}
else if (Cs >= 994) {
if (Rs < 1691) {
return 8;
}
else if (Rs >= 1691) {
return 2;
}
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (ssDCval >= 154) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 28) {
if (ssDCval < 158) {
if (Cs < 741) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
if (diffRsCsdiff < -11) {
return 1;
}
else if (diffRsCsdiff >= -11) {
if (ssDCval < 117) {
if (Rs < 2009) {
return 8;
}
else if (Rs >= 2009) {
return 2;
}
}
else if (ssDCval >= 117) {
return 2;
}
}
}
}
else if (Cs >= 741) {
if (refDCval < 111) {
if (TSC < 5) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Cs < 785) {
return 8;
}
else if (Cs >= 785) {
return 1;
}
}
}
else if (TSC >= 5) {
if (tcor < 91) {
if (Rs < 1487) {
return 4;
}
else if (Rs >= 1487) {
if (Rs < 1877) {
return 1;
}
else if (Rs >= 1877) {
return 8;
}
}
}
else if (tcor >= 91) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
if (Rs < 1485) {
return 8;
}
else if (Rs >= 1485) {
return 1;
}
}
}
}
}
else if (refDCval >= 111) {
if (Rs < 2359) {
if (Cs < 801) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (Rs < 995) {
return 1;
}
else if (Rs >= 995) {
return 8;
}
}
}
else if (Cs >= 801) {
return 8;
}
}
else if (Rs >= 2359) {
return 1;
}
}
}
}
else if (ssDCval >= 158) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 70) {
if (Cs < 779) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 1463) {
if (negBalance < 52) {
if (Rs < 1400) {
return 2;
}
else if (Rs >= 1400) {
return 8;
}
}
else if (negBalance >= 52) {
if (Cs < 575) {
return 1;
}
else if (Cs >= 575) {
if (Rs < 1138) {
return 8;
}
else if (Rs >= 1138) {
return 4;
}
}
}
}
else if (Rs >= 1463) {
if (mu_mv_mag_sq < 34) {
if (ssDCval < 107) {
if (Rs < 1757) {
return 1;
}
else if (Rs >= 1757) {
return 8;
}
}
else if (ssDCval >= 107) {
return 1;
}
}
else if (mu_mv_mag_sq >= 34) {
if (Rs < 2039) {
return 1;
}
else if (Rs >= 2039) {
return 2;
}
}
}
}
}
else if (Cs >= 779) {
if (ssDCval < 101) {
if (Cs < 959) {
if (refDCval < 88) {
if (Rs < 1106) {
return 8;
}
else if (Rs >= 1106) {
return 1;
}
}
else if (refDCval >= 88) {
if (Rs < 899) {
return 2;
}
else if (Rs >= 899) {
if (Cs < 905) {
return 4;
}
else if (Cs >= 905) {
if (Rs < 1624) {
return 4;
}
else if (Rs >= 1624) {
return 2;
}
}
}
}
}
else if (Cs >= 959) {
if (Rs < 1130) {
return 1;
}
else if (Rs >= 1130) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 2;
}
}
}
}
else if (ssDCval >= 101) {
if (TSC < 8) {
if (SC < 2897) {
if (RsCsDiff < 80) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (Rs < 1719) {
return 8;
}
else if (Rs >= 1719) {
return 2;
}
}
}
else if (RsCsDiff >= 80) {
return 2;
}
}
else if (SC >= 2897) {
return 8;
}
}
else if (TSC >= 8) {
if (mu_mv_mag_sq < 6) {
if (Rs < 1620) {
return 2;
}
else if (Rs >= 1620) {
return 1;
}
}
else if (mu_mv_mag_sq >= 6) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (ssDCval < 133) {
if (Rs < 374) {
if (ssDCval < 130) {
if (posBalance < 49) {
if (negBalance < 41) {
return 4;
}
else if (negBalance >= 41) {
return 8;
}
}
else if (posBalance >= 49) {
if (ssDCval < 109) {
return 4;
}
else if (ssDCval >= 109) {
return 8;
}
}
}
else if (ssDCval >= 130) {
return 2;
}
}
else if (Rs >= 374) {
if (Rs < 631) {
if (Cs < 1203) {
if (mu_mv_mag_sq < 107) {
if (ssDCval < 117) {
if (ssDCval < 111) {
if (tcor < 98) {
if (Rs < 504) {
return 1;
}
else if (Rs >= 504) {
return 8;
}
}
else if (tcor >= 98) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (Rs < 411) {
return 4;
}
else if (Rs >= 411) {
if (Rs < 487) {
return 8;
}
else if (Rs >= 487) {
return 1;
}
}
}
}
}
else if (ssDCval >= 111) {
return 1;
}
}
else if (ssDCval >= 117) {
if (AFD < 3) {
if (refDCval < 117) {
return 1;
}
else if (refDCval >= 117) {
return 4;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (SC < 1221) {
if (posBalance < 52) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Rs < 480) {
return 8;
}
else if (Rs >= 480) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (RsCsDiff < 4) {
return 1;
}
else if (RsCsDiff >= 4) {
return 8;
}
}
}
else if (SC >= 1221) {
return 8;
}
}
}
else if (Cs >= 1203) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (Rs >= 631) {
if (refDCval < 97) {
if (SC < 2645) {
if (RsCsDiff < 7) {
if (mu_mv_mag_sq < 126) {
if (Rs < 922) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
else if (Rs >= 922) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (ssDCval < 94) {
return 4;
}
else if (ssDCval >= 94) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (Rs < 831) {
return 4;
}
else if (Rs >= 831) {
if (Rs < 1501) {
return 8;
}
else if (Rs >= 1501) {
return 4;
}
}
}
}
else if (SC >= 2645) {
if (Rs < 2145) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (Cs < 631) {
return 1;
}
else if (Cs >= 631) {
return 8;
}
}
}
else if (Rs >= 2145) {
if (Rs < 2174) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (Rs >= 2174) {
return 4;
}
}
}
}
else if (refDCval >= 97) {
if (ssDCval < 121) {
if (refDCval < 115) {
if (ssDCval < 109) {
if (SC < 1362) {
if (Cs < 618) {
return 8;
}
else if (Cs >= 618) {
return 1;
}
}
else if (SC >= 1362) {
return 8;
}
}
else if (ssDCval >= 109) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (mu_mv_mag_sq < 125) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 109) {
return 1;
}
else if (mu_mv_mag_sq >= 109) {
if (Cs < 728) {
return 1;
}
else if (Cs >= 728) {
return 8;
}
}
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1613) {
return 8;
}
else if (Rs >= 1613) {
return 1;
}
}
}
}
}
else if (refDCval >= 115) {
if (diffRsCsdiff < -46) {
return 2;
}
else if (diffRsCsdiff >= -46) {
return 8;
}
}
}
else if (ssDCval >= 121) {
if (Cs < 1153) {
if (Cs < 796) {
if (Rs < 979) {
return 8;
}
else if (Rs >= 979) {
return 1;
}
}
else if (Cs >= 796) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
if (Rs < 1293) {
return 8;
}
else if (Rs >= 1293) {
return 2;
}
}
}
}
else if (Cs >= 1153) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (Rs < 1534) {
return 8;
}
else if (Rs >= 1534) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 145) {
return 1;
}
else if (ssDCval >= 145) {
if (Rs < 977) {
return 8;
}
else if (Rs >= 977) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2388) {
if (MVDiff < 17) {
if (refDCval < 109) {
if (posBalance < 48) {
if (TSCindex < 2) {
if (Cs < 730) {
if (Rs < 2564) {
return 1;
}
else if (Rs >= 2564) {
return 4;
}
}
else if (Cs >= 730) {
if (SC < 4958) {
return 8;
}
else if (SC >= 4958) {
if (SC < 5282) {
if (SC < 5254) {
return 4;
}
else if (SC >= 5254) {
if (Rs < 4001) {
return 8;
}
else if (Rs >= 4001) {
return 4;
}
}
}
else if (SC >= 5282) {
if (Rs < 4576) {
if (Cs < 1247) {
return 4;
}
else if (Cs >= 1247) {
return 8;
}
}
else if (Rs >= 4576) {
return 2;
}
}
}
}
}
else if (TSCindex >= 2) {
if (Rs < 3876) {
if (RsDiff < 7) {
if (ssDCval < 104) {
if (RsCsDiff < 7) {
if (ssDCval < 92) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (diffRsCsdiff < -7) {
if (posBalance < 39) {
if (Rs < 2616) {
return 8;
}
else if (Rs >= 2616) {
return 8;
}
}
else if (posBalance >= 39) {
return 1;
}
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
}
else if (ssDCval >= 92) {
if (Rs < 3342) {
return 2;
}
else if (Rs >= 3342) {
return 4;
}
}
}
else if (RsCsDiff >= 7) {
if (Cs < 437) {
if (Rs < 2499) {
return 1;
}
else if (Rs >= 2499) {
return 8;
}
}
else if (Cs >= 437) {
if (negBalance < 32) {
if (ssDCval < 90) {
if (mu_mv_mag_sq < 100) {
if (MVDiff < 1) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (MVDiff >= 1) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 3140) {
if (Cs < 750) {
return 8;
}
else if (Cs >= 750) {
return 2;
}
}
else if (Rs >= 3140) {
if (ssDCval < 68) {
if (TSC < 3) {
if (Rs < 3541) {
return 8;
}
else if (Rs >= 3541) {
return 4;
}
}
else if (TSC >= 3) {
return 1;
}
}
else if (ssDCval >= 68) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 4;
}
}
}
}
}
else if (ssDCval >= 90) {
if (mu_mv_mag_sq < 85) {
if (Rs < 2832) {
return 1;
}
else if (Rs >= 2832) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (posBalance < 44) {
if (Cs < 1081) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (Rs < 3423) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 4;
}
}
else if (Rs >= 3423) {
if (diffMVdiffVal < -1) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 3453) {
return 8;
}
else if (Rs >= 3453) {
return 8;
}
}
}
}
}
else if (Cs >= 1081) {
return 4;
}
}
else if (posBalance >= 44) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (Cs < 962) {
return 2;
}
else if (Cs >= 962) {
return 4;
}
}
}
}
else if (negBalance >= 32) {
if (Rs < 3164) {
if (ssDCval < 74) {
if (ssDCval < 60) {
if (negBalance < 35) {
if (Rs < 2567) {
return 2;
}
else if (Rs >= 2567) {
return 4;
}
}
else if (negBalance >= 35) {
return 8;
}
}
else if (ssDCval >= 60) {
if (tcor < 93) {
if (negBalance < 35) {
if (Rs < 2942) {
return 8;
}
else if (Rs >= 2942) {
return 2;
}
}
else if (negBalance >= 35) {
if (Cs < 1043) {
return 4;
}
else if (Cs >= 1043) {
return 8;
}
}
}
else if (tcor >= 93) {
if (posBalance < 41) {
if (Rs < 2690) {
if (Rs < 2621) {
return 4;
}
else if (Rs >= 2621) {
return 8;
}
}
else if (Rs >= 2690) {
if (Cs < 1032) {
if (Cs < 758) {
return 8;
}
else if (Cs >= 758) {
return 2;
}
}
else if (Cs >= 1032) {
if (Rs < 2730) {
return 2;
}
else if (Rs >= 2730) {
return 1;
}
}
}
}
else if (posBalance >= 41) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 74) {
if (Rs < 2996) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (ssDCval < 83) {
if (SC < 3981) {
return 1;
}
else if (SC >= 3981) {
return 4;
}
}
else if (ssDCval >= 83) {
if (Cs < 728) {
if (Rs < 2441) {
return 8;
}
else if (Rs >= 2441) {
return 1;
}
}
else if (Cs >= 728) {
return 8;
}
}
}
}
else if (Rs >= 2996) {
if (AFD < 8) {
if (Rs < 3106) {
return 8;
}
else if (Rs >= 3106) {
return 2;
}
}
else if (AFD >= 8) {
if (Rs < 3124) {
return 4;
}
else if (Rs >= 3124) {
return 1;
}
}
}
}
}
else if (Rs >= 3164) {
if (refDCval < 79) {
if (Cs < 1318) {
return 4;
}
else if (Cs >= 1318) {
return 2;
}
}
else if (refDCval >= 79) {
if (Rs < 3312) {
if (Rs < 3188) {
return 4;
}
else if (Rs >= 3188) {
if (Rs < 3215) {
return 2;
}
else if (Rs >= 3215) {
return 8;
}
}
}
else if (Rs >= 3312) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 104) {
if (CsDiff < 6) {
if (Rs < 3463) {
return 8;
}
else if (Rs >= 3463) {
return 1;
}
}
else if (CsDiff >= 6) {
return 1;
}
}
}
else if (RsDiff >= 7) {
if (negBalance < 51) {
if (ssDCval < 86) {
if (ssDCval < 68) {
if (posBalance < 37) {
if (ssDCval < 63) {
if (Rs < 2931) {
return 2;
}
else if (Rs >= 2931) {
if (Rs < 3454) {
return 8;
}
else if (Rs >= 3454) {
return 4;
}
}
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (posBalance >= 37) {
if (MVDiff < 3) {
if (mcTcor < 96) {
if (Rs < 3187) {
return 4;
}
else if (Rs >= 3187) {
return 1;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (MVDiff >= 3) {
if (CsDiff < 6) {
if (mu_mv_mag_sq < 3) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
if (TSC < 8) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 45) {
if (gchDC < 2) {
if (diffAFD < -2) {
if (RsDiff < 12) {
return 2;
}
else if (RsDiff >= 12) {
return 4;
}
}
else if (diffAFD >= -2) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (Cs < 830) {
if (Cs < 684) {
return 8;
}
else if (Cs >= 684) {
return 4;
}
}
else if (Cs >= 830) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 2;
}
}
}
}
else if (negBalance >= 45) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (diffMVdiffVal < 3) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
}
}
}
else if (ssDCval >= 68) {
if (mu_mv_mag_sq < 79) {
if (negBalance < 44) {
if (diffMVdiffVal < -2) {
if (Rs < 2974) {
if (Rs < 2707) {
return 2;
}
else if (Rs >= 2707) {
return 4;
}
}
else if (Rs >= 2974) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
if (Rs < 3205) {
return 1;
}
else if (Rs >= 3205) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= -2) {
if (SC < 3760) {
if (Cs < 987) {
return 1;
}
else if (Cs >= 987) {
return 4;
}
}
else if (SC >= 3760) {
if (mu_mv_mag_sq < 77) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (Rs < 3362) {
return 4;
}
else if (Rs >= 3362) {
if (Cs < 958) {
return 4;
}
else if (Cs >= 958) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
if (Cs < 901) {
return 4;
}
else if (Cs >= 901) {
return 1;
}
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 2758) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (Rs < 2429) {
return 2;
}
else if (Rs >= 2429) {
return 4;
}
}
}
else if (Rs >= 2758) {
if (ssDCval < 71) {
if (Rs < 3264) {
return 4;
}
else if (Rs >= 3264) {
return 1;
}
}
else if (ssDCval >= 71) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (gchDC < 1) {
if (posBalance < 45) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (Rs < 3000) {
return 2;
}
else if (Rs >= 3000) {
return 4;
}
}
}
else if (posBalance >= 45) {
return 2;
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
else if (ssDCval >= 86) {
if (Cs < 922) {
if (Rs < 2868) {
return 8;
}
else if (Rs >= 2868) {
return 2;
}
}
else if (Cs >= 922) {
if (CsDiff < 5) {
if (posBalance < 43) {
if (Rs < 3790) {
return 4;
}
else if (Rs >= 3790) {
return 4;
}
}
else if (posBalance >= 43) {
if (MVDiff < 9) {
if (Rs < 3379) {
return 8;
}
else if (Rs >= 3379) {
return 8;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
else if (negBalance >= 51) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Rs < 2820) {
return 1;
}
else if (Rs >= 2820) {
return 2;
}
}
}
}
}
else if (Rs >= 3876) {
if (negBalance < 46) {
if (Cs < 1085) {
if (Rs < 5608) {
return 4;
}
else if (Rs >= 5608) {
return 2;
}
}
else if (Cs >= 1085) {
if (SC < 5298) {
if (AFD < 9) {
if (Cs < 1293) {
return 4;
}
else if (Cs >= 1293) {
return 8;
}
}
else if (AFD >= 9) {
if (Cs < 1094) {
return 8;
}
else if (Cs >= 1094) {
if (Rs < 4051) {
return 4;
}
else if (Rs >= 4051) {
return 4;
}
}
}
}
else if (SC >= 5298) {
if (tcor < 99) {
if (ssDCval < 90) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (ssDCval >= 90) {
if (posBalance < 41) {
if (Rs < 4127) {
return 8;
}
else if (Rs >= 4127) {
return 8;
}
}
else if (posBalance >= 41) {
if (Rs < 4215) {
if (MVDiff < 7) {
if (Cs < 1249) {
return 4;
}
else if (Cs >= 1249) {
return 8;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (Rs >= 4215) {
if (Cs < 1201) {
return 4;
}
else if (Cs >= 1201) {
return 4;
}
}
}
}
}
else if (tcor >= 99) {
if (Cs < 1275) {
if (Rs < 4347) {
return 4;
}
else if (Rs >= 4347) {
return 2;
}
}
else if (Cs >= 1275) {
return 8;
}
}
}
}
}
else if (negBalance >= 46) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
return 2;
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Rs < 3623) {
if (Cs < 246) {
if (Rs < 2450) {
return 2;
}
else if (Rs >= 2450) {
return 1;
}
}
else if (Cs >= 246) {
if (negBalance < 21) {
if (SC < 3216) {
return 1;
}
else if (SC >= 3216) {
if (negBalance < 15) {
if (Rs < 2902) {
return 4;
}
else if (Rs >= 2902) {
return 4;
}
}
else if (negBalance >= 15) {
return 8;
}
}
}
else if (negBalance >= 21) {
if (Cs < 661) {
if (ssDCval < 81) {
if (SC < 3399) {
if (TSC < 5) {
if (Rs < 2473) {
if (Rs < 2404) {
return 2;
}
else if (Rs >= 2404) {
return 1;
}
}
else if (Rs >= 2473) {
return 4;
}
}
else if (TSC >= 5) {
return 2;
}
}
else if (SC >= 3399) {
return 4;
}
}
else if (ssDCval >= 81) {
if (SC < 3171) {
if (SC < 2854) {
return 1;
}
else if (SC >= 2854) {
return 2;
}
}
else if (SC >= 3171) {
return 8;
}
}
}
else if (Cs >= 661) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
if (Rs < 2416) {
if (Rs < 2413) {
return 2;
}
else if (Rs >= 2413) {
return 1;
}
}
else if (Rs >= 2416) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 3623) {
if (posBalance < 65) {
return 4;
}
else if (posBalance >= 65) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (ssDCval < 63) {
if (SC < 3658) {
if (Cs < 1002) {
return 2;
}
else if (Cs >= 1002) {
if (Rs < 2431) {
return 8;
}
else if (Rs >= 2431) {
return 4;
}
}
}
else if (SC >= 3658) {
if (Rs < 7640) {
if (refDCval < 60) {
return 4;
}
else if (refDCval >= 60) {
if (diffAFD < 2) {
if (SC < 4231) {
if (Cs < 978) {
if (Rs < 2957) {
return 2;
}
else if (Rs >= 2957) {
return 4;
}
}
else if (Cs >= 978) {
return 4;
}
}
else if (SC >= 4231) {
return 2;
}
}
else if (diffAFD >= 2) {
if (Rs < 3708) {
return 2;
}
else if (Rs >= 3708) {
return 4;
}
}
}
}
else if (Rs >= 7640) {
return 2;
}
}
}
else if (ssDCval >= 63) {
if (tcor < 80) {
if (SC < 3475) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 4;
}
}
else if (SC >= 3475) {
if (refDCval < 74) {
if (posBalance < 56) {
if (Cs < 1052) {
return 4;
}
else if (Cs >= 1052) {
if (Rs < 2833) {
return 2;
}
else if (Rs >= 2833) {
return 4;
}
}
}
else if (posBalance >= 56) {
if (Rs < 3300) {
return 2;
}
else if (Rs >= 3300) {
return 4;
}
}
}
else if (refDCval >= 74) {
if (RsDiff < 15) {
if (Rs < 2911) {
return 4;
}
else if (Rs >= 2911) {
return 1;
}
}
else if (RsDiff >= 15) {
if (Cs < 1158) {
return 2;
}
else if (Cs >= 1158) {
return 8;
}
}
}
}
}
else if (tcor >= 80) {
if (negBalance < 37) {
if (posBalance < 55) {
if (tcor < 94) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (Cs < 673) {
return 4;
}
else if (Cs >= 673) {
return 1;
}
}
}
else if (tcor >= 94) {
return 2;
}
}
else if (posBalance >= 55) {
if (posBalance < 58) {
if (tcor < 89) {
return 1;
}
else if (tcor >= 89) {
return 2;
}
}
else if (posBalance >= 58) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
else if (negBalance >= 37) {
if (mcTcor < 99) {
if (RsDiff < 6) {
if (CsDiff < 4) {
if (Rs < 2455) {
return 1;
}
else if (Rs >= 2455) {
return 2;
}
}
else if (CsDiff >= 4) {
if (Rs < 3261) {
return 4;
}
else if (Rs >= 3261) {
return 2;
}
}
}
else if (RsDiff >= 6) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (negBalance < 41) {
if (posBalance < 52) {
if (Cs < 1122) {
return 4;
}
else if (Cs >= 1122) {
return 1;
}
}
else if (posBalance >= 52) {
return 2;
}
}
else if (negBalance >= 41) {
if (mcTcor < 91) {
if (Rs < 2595) {
return 4;
}
else if (Rs >= 2595) {
return 2;
}
}
else if (mcTcor >= 91) {
return 2;
}
}
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsCsDiff < 142) {
if (tcor < 96) {
if (diffAFD < -2) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 4;
}
}
else if (diffAFD >= -2) {
if (negBalance < 30) {
if (SC < 4318) {
if (mu_mv_mag_sq < 64) {
return 8;
}
else if (mu_mv_mag_sq >= 64) {
if (gchDC < 3) {
if (Rs < 2831) {
return 8;
}
else if (Rs >= 2831) {
return 8;
}
}
else if (gchDC >= 3) {
return 4;
}
}
}
else if (SC >= 4318) {
if (posBalance < 74) {
if (Rs < 3866) {
return 4;
}
else if (Rs >= 3866) {
return 4;
}
}
else if (posBalance >= 74) {
return 2;
}
}
}
else if (negBalance >= 30) {
if (RsCsDiff < 101) {
if (Rs < 2436) {
if (Rs < 2416) {
return 8;
}
else if (Rs >= 2416) {
return 4;
}
}
else if (Rs >= 2436) {
return 8;
}
}
else if (RsCsDiff >= 101) {
if (TSC < 8) {
if (Cs < 1211) {
return 2;
}
else if (Cs >= 1211) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
else if (tcor >= 96) {
if (posBalance < 57) {
if (Rs < 3180) {
if (SC < 3039) {
return 2;
}
else if (SC >= 3039) {
return 8;
}
}
else if (Rs >= 3180) {
return 2;
}
}
else if (posBalance >= 57) {
return 2;
}
}
}
else if (RsCsDiff >= 142) {
if (negBalance < 52) {
if (RsDiff < 13) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
return 4;
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (negBalance >= 52) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (Cs < 798) {
return 2;
}
else if (Cs >= 798) {
if (refDCval < 102) {
if (ssDCval < 96) {
return 4;
}
else if (ssDCval >= 96) {
return 8;
}
}
else if (refDCval >= 102) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 3359) {
if (mcTcor < 97) {
if (RsDiff < 9) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (RsDiff < 7) {
if (ssDCval < 136) {
return 8;
}
else if (ssDCval >= 136) {
if (ssDCval < 137) {
return 1;
}
else if (ssDCval >= 137) {
if (Rs < 3025) {
return 8;
}
else if (Rs >= 3025) {
return 2;
}
}
}
}
else if (RsDiff >= 7) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
}
}
else if (RsDiff >= 9) {
if (negBalance < 54) {
if (Cs < 980) {
if (ssDCval < 114) {
return 2;
}
else if (ssDCval >= 114) {
if (gchDC < 4) {
if (RsCsDiff < 246) {
return 1;
}
else if (RsCsDiff >= 246) {
if (Rs < 3226) {
return 2;
}
else if (Rs >= 3226) {
return 1;
}
}
}
else if (gchDC >= 4) {
return 2;
}
}
}
else if (Cs >= 980) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (negBalance >= 54) {
if (SC < 3860) {
if (Rs < 2476) {
return 8;
}
else if (Rs >= 2476) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
return 2;
}
}
}
else if (SC >= 3860) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 7) {
if (Rs < 2690) {
if (RsDiff < 5) {
if (Rs < 2458) {
return 8;
}
else if (Rs >= 2458) {
return 1;
}
}
else if (RsDiff >= 5) {
if (ssDCval < 126) {
return 1;
}
else if (ssDCval >= 126) {
return 2;
}
}
}
else if (Rs >= 2690) {
if (Cs < 530) {
if (Rs < 3090) {
return 2;
}
else if (Rs >= 3090) {
return 8;
}
}
else if (Cs >= 530) {
return 2;
}
}
}
else if (MVDiff >= 7) {
if (TSC < 6) {
if (mu_mv_mag_sq < 85) {
if (SC < 3154) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 8;
}
}
else if (SC >= 3154) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
if (Rs < 3061) {
return 8;
}
else if (Rs >= 3061) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (diffRsCsdiff < -30) {
return 1;
}
else if (diffRsCsdiff >= -30) {
if (Rs < 2789) {
return 1;
}
else if (Rs >= 2789) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (SC < 3320) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 2;
}
}
else if (SC >= 3320) {
if (CsDiff < 4) {
if (AFD < 9) {
if (Rs < 3112) {
return 1;
}
else if (Rs >= 3112) {
return 8;
}
}
else if (AFD >= 9) {
return 2;
}
}
else if (CsDiff >= 4) {
if (ssDCval < 118) {
if (Cs < 799) {
return 2;
}
else if (Cs >= 799) {
if (Cs < 1032) {
return 4;
}
else if (Cs >= 1032) {
return 8;
}
}
}
else if (ssDCval >= 118) {
if (Cs < 963) {
return 1;
}
else if (Cs >= 963) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 3359) {
if (mcTcor < 94) {
if (ssDCval < 135) {
if (RsDiff < 11) {
if (Rs < 3667) {
return 8;
}
else if (Rs >= 3667) {
return 1;
}
}
else if (RsDiff >= 11) {
if (ssDCval < 134) {
return 2;
}
else if (ssDCval >= 134) {
return 1;
}
}
}
else if (ssDCval >= 135) {
if (Rs < 5179) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 2;
}
}
else if (Rs >= 5179) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (CsDiff < 4) {
if (refDCval < 142) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (RsDiff < 9) {
if (tcor < 96) {
if (diffRsCsdiff < 5) {
if (posBalance < 59) {
if (RsDiff < 5) {
if (Rs < 3519) {
return 2;
}
else if (Rs >= 3519) {
return 1;
}
}
else if (RsDiff >= 5) {
if (Rs < 4080) {
return 8;
}
else if (Rs >= 4080) {
return 1;
}
}
}
else if (posBalance >= 59) {
return 2;
}
}
else if (diffRsCsdiff >= 5) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
return 8;
}
}
}
else if (tcor >= 96) {
return 2;
}
}
else if (RsDiff >= 9) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
if (Rs < 5585) {
return 2;
}
else if (Rs >= 5585) {
return 1;
}
}
}
}
}
else if (refDCval >= 142) {
if (Rs < 5834) {
if (Rs < 3963) {
if (Cs < 1176) {
return 2;
}
else if (Cs >= 1176) {
if (negBalance < 33) {
if (Rs < 3554) {
return 2;
}
else if (Rs >= 3554) {
return 1;
}
}
else if (negBalance >= 33) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
if (Rs < 3599) {
return 8;
}
else if (Rs >= 3599) {
return 1;
}
}
}
}
}
else if (Rs >= 3963) {
return 2;
}
}
else if (Rs >= 5834) {
if (Cs < 836) {
return 1;
}
else if (Cs >= 836) {
return 2;
}
}
}
}
else if (CsDiff >= 4) {
if (Rs < 3981) {
if (refDCval < 138) {
if (posBalance < 41) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (posBalance >= 41) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
if (diffMVdiffVal < -40) {
return 1;
}
else if (diffMVdiffVal >= -40) {
if (ssDCval < 109) {
return 4;
}
else if (ssDCval >= 109) {
return 2;
}
}
}
}
}
else if (refDCval >= 138) {
if (Cs < 985) {
if (ssDCval < 148) {
return 2;
}
else if (ssDCval >= 148) {
return 1;
}
}
else if (Cs >= 985) {
return 1;
}
}
}
else if (Rs >= 3981) {
if (ssDCval < 132) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
if (Cs < 1195) {
return 2;
}
else if (Cs >= 1195) {
return 4;
}
}
}
}
else if (ssDCval >= 132) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (RsDiff < 20) {
if (SC < 5406) {
return 2;
}
else if (SC >= 5406) {
if (Rs < 5095) {
return 1;
}
else if (Rs >= 5095) {
if (Cs < 932) {
return 1;
}
else if (Cs >= 932) {
return 2;
}
}
}
}
else if (RsDiff >= 20) {
return 2;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 17) {
if (mcTcor < 72) {
if (ssDCval < 105) {
if (gchDC < 7) {
return 1;
}
else if (gchDC >= 7) {
return 8;
}
}
else if (ssDCval >= 105) {
if (AFD < 28) {
return 4;
}
else if (AFD >= 28) {
return 2;
}
}
}
else if (mcTcor >= 72) {
if (refDCval < 132) {
if (refDCval < 79) {
if (TSC < 8) {
if (Rs < 2786) {
if (RsCsDiff < 148) {
if (AFD < 12) {
if (diffMVdiffVal < -18) {
return 1;
}
else if (diffMVdiffVal >= -18) {
if (Rs < 2666) {
if (refDCval < 58) {
if (Cs < 1232) {
return 4;
}
else if (Cs >= 1232) {
return 8;
}
}
else if (refDCval >= 58) {
return 8;
}
}
else if (Rs >= 2666) {
return 4;
}
}
}
else if (AFD >= 12) {
if (Cs < 1074) {
if (RsDiff < 8) {
if (Rs < 2585) {
return 8;
}
else if (Rs >= 2585) {
return 2;
}
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (Cs >= 1074) {
return 2;
}
}
}
else if (RsCsDiff >= 148) {
if (Cs < 203) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (Cs >= 203) {
if (TSC < 7) {
if (Rs < 2516) {
return 2;
}
else if (Rs >= 2516) {
return 4;
}
}
else if (TSC >= 7) {
return 2;
}
}
}
}
else if (Rs >= 2786) {
if (Cs < 982) {
if (gchDC < 3) {
if (Cs < 144) {
if (Rs < 3387) {
return 1;
}
else if (Rs >= 3387) {
return 4;
}
}
else if (Cs >= 144) {
if (tcor < 97) {
return 4;
}
else if (tcor >= 97) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (Cs < 534) {
return 4;
}
else if (Cs >= 534) {
return 8;
}
}
}
else if (Cs >= 982) {
if (posBalance < 36) {
if (Cs < 1173) {
return 8;
}
else if (Cs >= 1173) {
return 1;
}
}
else if (posBalance >= 36) {
if (diffRsCsdiff < 10) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 2;
}
}
else if (diffRsCsdiff >= 10) {
return 4;
}
}
}
}
}
else if (TSC >= 8) {
if (Rs < 3692) {
if (mcTcor < 82) {
if (mu_mv_mag_sq < 24) {
if (diffAFD < -2) {
if (Rs < 2609) {
return 4;
}
else if (Rs >= 2609) {
return 8;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 24) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
else if (mcTcor >= 82) {
if (AFD < 12) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 1;
}
}
else if (AFD >= 12) {
if (Cs < 1210) {
if (CsDiff < 3) {
if (Cs < 134) {
if (Rs < 3039) {
return 1;
}
else if (Rs >= 3039) {
return 2;
}
}
else if (Cs >= 134) {
return 4;
}
}
else if (CsDiff >= 3) {
if (Rs < 2761) {
if (RsCsDiff < 174) {
if (mcTcor < 86) {
return 1;
}
else if (mcTcor >= 86) {
if (posBalance < 51) {
if (Cs < 823) {
return 1;
}
else if (Cs >= 823) {
return 2;
}
}
else if (posBalance >= 51) {
return 2;
}
}
}
else if (RsCsDiff >= 174) {
if (RsDiff < 22) {
return 2;
}
else if (RsDiff >= 22) {
return 8;
}
}
}
else if (Rs >= 2761) {
if (posBalance < 54) {
if (diffRsCsdiff < 36) {
if (TSC < 9) {
if (negBalance < 41) {
if (Rs < 3306) {
return 2;
}
else if (Rs >= 3306) {
return 4;
}
}
else if (negBalance >= 41) {
return 4;
}
}
else if (TSC >= 9) {
return 2;
}
}
else if (diffRsCsdiff >= 36) {
return 4;
}
}
else if (posBalance >= 54) {
return 2;
}
}
}
}
else if (Cs >= 1210) {
return 4;
}
}
}
}
else if (Rs >= 3692) {
if (Rs < 5986) {
if (Cs < 886) {
return 4;
}
else if (Cs >= 886) {
return 2;
}
}
else if (Rs >= 5986) {
if (Rs < 7362) {
return 8;
}
else if (Rs >= 7362) {
return 2;
}
}
}
}
}
else if (refDCval >= 79) {
if (RsCsDiff < 248) {
if (SC < 3438) {
if (ssDCval < 91) {
if (RsCsDiff < 151) {
if (TSC < 8) {
if (Rs < 2591) {
return 1;
}
else if (Rs >= 2591) {
if (Rs < 2963) {
return 2;
}
else if (Rs >= 2963) {
return 1;
}
}
}
else if (TSC >= 8) {
if (Cs < 790) {
if (SC < 2912) {
return 1;
}
else if (SC >= 2912) {
return 2;
}
}
else if (Cs >= 790) {
return 8;
}
}
}
else if (RsCsDiff >= 151) {
if (AFD < 15) {
if (RsDiff < 12) {
return 1;
}
else if (RsDiff >= 12) {
return 4;
}
}
else if (AFD >= 15) {
if (tcor < 74) {
return 1;
}
else if (tcor >= 74) {
if (Cs < 724) {
return 2;
}
else if (Cs >= 724) {
return 4;
}
}
}
}
}
else if (ssDCval >= 91) {
if (gchDC < 4) {
if (MVDiff < 38) {
if (posBalance < 54) {
if (Rs < 2905) {
if (Cs < 510) {
return 8;
}
else if (Cs >= 510) {
return 1;
}
}
else if (Rs >= 2905) {
return 2;
}
}
else if (posBalance >= 54) {
if (tcor < 77) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (tcor >= 77) {
if (MVDiff < 23) {
if (Rs < 2629) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 2;
}
}
else if (Rs >= 2629) {
if (RsDiff < 15) {
return 2;
}
else if (RsDiff >= 15) {
return 1;
}
}
}
else if (MVDiff >= 23) {
if (Cs < 670) {
if (RsDiff < 7) {
if (Rs < 2502) {
return 1;
}
else if (Rs >= 2502) {
return 8;
}
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (Cs >= 670) {
return 2;
}
}
}
}
}
else if (MVDiff >= 38) {
if (diffAFD < -4) {
if (posBalance < 42) {
if (Rs < 2619) {
return 2;
}
else if (Rs >= 2619) {
return 8;
}
}
else if (posBalance >= 42) {
return 1;
}
}
else if (diffAFD >= -4) {
if (diffAFD < 3) {
if (posBalance < 47) {
if (Rs < 2654) {
if (Rs < 2530) {
return 8;
}
else if (Rs >= 2530) {
return 2;
}
}
else if (Rs >= 2654) {
return 1;
}
}
else if (posBalance >= 47) {
if (SC < 3133) {
if (mu_mv_mag_sq < 5) {
return 8;
}
else if (mu_mv_mag_sq >= 5) {
if (ssDCval < 119) {
if (mcTcor < 92) {
if (Rs < 2405) {
return 8;
}
else if (Rs >= 2405) {
return 1;
}
}
else if (mcTcor >= 92) {
return 8;
}
}
else if (ssDCval >= 119) {
return 1;
}
}
}
else if (SC >= 3133) {
return 8;
}
}
}
else if (diffAFD >= 3) {
if (Rs < 2742) {
return 1;
}
else if (Rs >= 2742) {
return 8;
}
}
}
}
}
else if (gchDC >= 4) {
if (mu_mv_mag_sq < 37) {
if (ssDCval < 115) {
if (diffTSC < -2) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
else if (diffTSC >= -2) {
if (Rs < 2701) {
return 1;
}
else if (Rs >= 2701) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
return 8;
}
}
}
}
else if (ssDCval >= 115) {
return 1;
}
}
else if (mu_mv_mag_sq >= 37) {
if (Cs < 352) {
if (Rs < 2704) {
return 1;
}
else if (Rs >= 2704) {
return 8;
}
}
else if (Cs >= 352) {
return 2;
}
}
}
}
}
else if (SC >= 3438) {
if (Rs < 5234) {
if (MVDiff < 39) {
if (RsDiff < 8) {
if (Cs < 974) {
if (SC < 3557) {
return 1;
}
else if (SC >= 3557) {
if (Cs < 871) {
if (Cs < 768) {
return 8;
}
else if (Cs >= 768) {
return 4;
}
}
else if (Cs >= 871) {
return 1;
}
}
}
else if (Cs >= 974) {
return 8;
}
}
else if (RsDiff >= 8) {
if (ssDCval < 95) {
if (AFD < 18) {
if (CsDiff < 8) {
if (RsCsDiff < 124) {
if (MVDiff < 30) {
return 2;
}
else if (MVDiff >= 30) {
return 8;
}
}
else if (RsCsDiff >= 124) {
return 2;
}
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (AFD >= 18) {
if (SC < 3681) {
return 2;
}
else if (SC >= 3681) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
else if (ssDCval >= 95) {
if (negBalance < 38) {
if (refDCval < 114) {
if (Cs < 811) {
return 1;
}
else if (Cs >= 811) {
if (Cs < 1150) {
return 2;
}
else if (Cs >= 1150) {
return 8;
}
}
}
else if (refDCval >= 114) {
return 2;
}
}
else if (negBalance >= 38) {
if (Rs < 3153) {
if (MVDiff < 26) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
if (RsDiff < 14) {
if (Rs < 2624) {
return 1;
}
else if (Rs >= 2624) {
return 2;
}
}
else if (RsDiff >= 14) {
if (Rs < 2739) {
return 4;
}
else if (Rs >= 2739) {
return 8;
}
}
}
}
else if (MVDiff >= 26) {
if (Rs < 2785) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 2;
}
}
else if (Rs >= 2785) {
if (SC < 4217) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 8;
}
}
else if (SC >= 4217) {
return 2;
}
}
}
}
else if (Rs >= 3153) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (diffRsCsdiff < -212) {
return 2;
}
else if (diffRsCsdiff >= -212) {
if (mcTcor < 97) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (Rs < 3787) {
return 2;
}
else if (Rs >= 3787) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (Rs < 3266) {
return 8;
}
else if (Rs >= 3266) {
return 2;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 39) {
if (AFD < 24) {
if (TSC < 8) {
if (Rs < 3316) {
if (diffTSC < -2) {
return 8;
}
else if (diffTSC >= -2) {
if (Rs < 2985) {
return 1;
}
else if (Rs >= 2985) {
return 2;
}
}
}
else if (Rs >= 3316) {
if (MVDiff < 88) {
return 2;
}
else if (MVDiff >= 88) {
return 4;
}
}
}
else if (TSC >= 8) {
if (Cs < 447) {
return 4;
}
else if (Cs >= 447) {
if (diffAFD < -4) {
if (RsCsDiff < 86) {
return 8;
}
else if (RsCsDiff >= 86) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
if (Cs < 897) {
return 1;
}
else if (Cs >= 897) {
return 8;
}
}
}
}
else if (diffAFD >= -4) {
if (AFD < 13) {
if (diffMVdiffVal < -13) {
return 1;
}
else if (diffMVdiffVal >= -13) {
return 8;
}
}
else if (AFD >= 13) {
if (RsCsDiff < 225) {
return 8;
}
else if (RsCsDiff >= 225) {
if (SC < 3925) {
return 4;
}
else if (SC >= 3925) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 24) {
if (mcTcor < 84) {
if (Rs < 2795) {
return 2;
}
else if (Rs >= 2795) {
return 4;
}
}
else if (mcTcor >= 84) {
if (diffRsCsdiff < -16) {
if (negBalance < 59) {
return 8;
}
else if (negBalance >= 59) {
if (Rs < 2449) {
return 1;
}
else if (Rs >= 2449) {
return 2;
}
}
}
else if (diffRsCsdiff >= -16) {
return 1;
}
}
}
}
}
else if (Rs >= 5234) {
return 2;
}
}
}
else if (RsCsDiff >= 248) {
if (TSC < 11) {
if (Cs < 622) {
if (mcTcor < 96) {
if (RsCsDiff < 291) {
if (TSC < 10) {
if (Rs < 3843) {
return 2;
}
else if (Rs >= 3843) {
if (Rs < 5270) {
return 8;
}
else if (Rs >= 5270) {
return 2;
}
}
}
else if (TSC >= 10) {
if (Rs < 2645) {
return 2;
}
else if (Rs >= 2645) {
return 4;
}
}
}
else if (RsCsDiff >= 291) {
return 2;
}
}
else if (mcTcor >= 96) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (Rs < 3676) {
if (refDCval < 114) {
if (diffMVdiffVal < -10) {
return 1;
}
else if (diffMVdiffVal >= -10) {
if (refDCval < 113) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
return 2;
}
}
else if (refDCval >= 113) {
return 8;
}
}
}
else if (refDCval >= 114) {
return 1;
}
}
else if (Rs >= 3676) {
return 2;
}
}
}
}
else if (Cs >= 622) {
if (tcor < 77) {
if (ssDCval < 90) {
if (Rs < 3574) {
return 2;
}
else if (Rs >= 3574) {
return 1;
}
}
else if (ssDCval >= 90) {
if (RsCsDiff < 431) {
if (mu_mv_mag_sq < 5) {
if (Rs < 2606) {
return 8;
}
else if (Rs >= 2606) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (CsDiff < 9) {
if (negBalance < 59) {
return 8;
}
else if (negBalance >= 59) {
return 2;
}
}
else if (CsDiff >= 9) {
return 1;
}
}
}
else if (RsCsDiff >= 431) {
if (mu_mv_mag_sq < 33) {
if (Rs < 3431) {
return 8;
}
else if (Rs >= 3431) {
if (ssDCval < 120) {
return 2;
}
else if (ssDCval >= 120) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (MVDiff < 36) {
return 8;
}
else if (MVDiff >= 36) {
if (AFD < 30) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 2;
}
}
else if (AFD >= 30) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 33) {
return 2;
}
}
}
}
else if (tcor >= 77) {
if (ssDCval < 113) {
if (CsDiff < 5) {
if (Rs < 3801) {
if (MVDiff < 71) {
return 2;
}
else if (MVDiff >= 71) {
return 1;
}
}
else if (Rs >= 3801) {
if (Rs < 5431) {
return 8;
}
else if (Rs >= 5431) {
return 2;
}
}
}
else if (CsDiff >= 5) {
if (mu_mv_mag_sq < 23) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
if (mu_mv_mag_sq < 5) {
return 1;
}
else if (mu_mv_mag_sq >= 5) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (Cs < 892) {
return 2;
}
else if (Cs >= 892) {
return 1;
}
}
}
}
else if (ssDCval >= 113) {
if (Rs < 3833) {
if (Cs < 717) {
return 8;
}
else if (Cs >= 717) {
return 1;
}
}
else if (Rs >= 3833) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
if (Cs < 1206) {
if (Rs < 4656) {
return 8;
}
else if (Rs >= 4656) {
if (Rs < 5354) {
return 2;
}
else if (Rs >= 5354) {
if (MVDiff < 32) {
return 8;
}
else if (MVDiff >= 32) {
return 2;
}
}
}
}
else if (Cs >= 1206) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 8;
}
}
}
}
}
}
}
}
else if (TSC >= 11) {
if (RsCsDiff < 408) {
if (mcTcor < 91) {
if (tcor < 54) {
return 4;
}
else if (tcor >= 54) {
if (MVDiff < 27) {
if (AFD < 19) {
if (Cs < 1065) {
return 2;
}
else if (Cs >= 1065) {
return 8;
}
}
else if (AFD >= 19) {
return 4;
}
}
else if (MVDiff >= 27) {
if (Rs < 2714) {
if (RsDiff < 17) {
return 1;
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (Rs >= 2714) {
if (MVDiff < 63) {
if (AFD < 30) {
return 8;
}
else if (AFD >= 30) {
return 2;
}
}
else if (MVDiff >= 63) {
return 4;
}
}
}
}
}
else if (mcTcor >= 91) {
if (MVDiff < 32) {
if (ssDCval < 125) {
return 2;
}
else if (ssDCval >= 125) {
return 8;
}
}
else if (MVDiff >= 32) {
if (negBalance < 47) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 1;
}
}
else if (negBalance >= 47) {
if (CsDiff < 5) {
if (Cs < 503) {
return 1;
}
else if (Cs >= 503) {
if (Rs < 2647) {
return 1;
}
else if (Rs >= 2647) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
if (Cs < 1231) {
return 1;
}
else if (Cs >= 1231) {
if (Rs < 3291) {
return 2;
}
else if (Rs >= 3291) {
return 1;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 408) {
if (CsDiff < 9) {
if (Cs < 1206) {
if (tcor < 77) {
if (ssDCval < 127) {
if (posBalance < 69) {
return 2;
}
else if (posBalance >= 69) {
if (Rs < 5166) {
return 8;
}
else if (Rs >= 5166) {
return 2;
}
}
}
else if (ssDCval >= 127) {
return 8;
}
}
else if (tcor >= 77) {
if (Rs < 5024) {
if (ssDCval < 111) {
if (Rs < 4496) {
return 2;
}
else if (Rs >= 4496) {
return 1;
}
}
else if (ssDCval >= 111) {
return 1;
}
}
else if (Rs >= 5024) {
return 2;
}
}
}
else if (Cs >= 1206) {
if (Cs < 1222) {
if (AFD < 32) {
return 8;
}
else if (AFD >= 32) {
return 4;
}
}
else if (Cs >= 1222) {
if (Rs < 5269) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (Rs >= 5269) {
return 1;
}
}
}
}
else if (CsDiff >= 9) {
if (diffMVdiffVal < 12) {
if (Rs < 4674) {
if (diffMVdiffVal < -23) {
if (Rs < 3622) {
return 1;
}
else if (Rs >= 3622) {
return 2;
}
}
else if (diffMVdiffVal >= -23) {
return 2;
}
}
else if (Rs >= 4674) {
return 1;
}
}
else if (diffMVdiffVal >= 12) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 132) {
if (mcTcor < 92) {
if (refDCval < 147) {
if (AFD < 20) {
if (tcor < 62) {
if (Cs < 1046) {
return 1;
}
else if (Cs >= 1046) {
return 8;
}
}
else if (tcor >= 62) {
if (Rs < 2438) {
return 1;
}
else if (Rs >= 2438) {
return 4;
}
}
}
else if (AFD >= 20) {
return 2;
}
}
else if (refDCval >= 147) {
if (posBalance < 55) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 1;
}
}
else if (posBalance >= 55) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 3279) {
if (TSC < 7) {
if (SC < 2634) {
return 1;
}
else if (SC >= 2634) {
return 8;
}
}
else if (TSC >= 7) {
if (tcor < 81) {
if (diffRsCsdiff < 37) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (refDCval < 143) {
if (Rs < 3006) {
return 1;
}
else if (Rs >= 3006) {
return 8;
}
}
else if (refDCval >= 143) {
if (Rs < 3258) {
return 2;
}
else if (Rs >= 3258) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 37) {
return 1;
}
}
else if (tcor >= 81) {
if (mcTcor < 95) {
if (MVDiff < 81) {
if (Cs < 1064) {
return 1;
}
else if (Cs >= 1064) {
return 2;
}
}
else if (MVDiff >= 81) {
return 8;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
else if (Rs >= 3279) {
if (CsDiff < 4) {
if (mcTcor < 93) {
if (Cs < 596) {
if (Rs < 4291) {
return 1;
}
else if (Rs >= 4291) {
return 2;
}
}
else if (Cs >= 596) {
return 4;
}
}
else if (mcTcor >= 93) {
if (SC < 3967) {
if (SC < 3820) {
return 1;
}
else if (SC >= 3820) {
return 8;
}
}
else if (SC >= 3967) {
if (Cs < 764) {
return 2;
}
else if (Cs >= 764) {
if (Cs < 787) {
if (Rs < 3909) {
return 4;
}
else if (Rs >= 3909) {
return 1;
}
}
else if (Cs >= 787) {
return 2;
}
}
}
}
}
else if (CsDiff >= 4) {
if (tcor < 74) {
if (mu_mv_mag_sq < 22) {
return 8;
}
else if (mu_mv_mag_sq >= 22) {
return 2;
}
}
else if (tcor >= 74) {
if (SC < 7243) {
if (negBalance < 52) {
if (negBalance < 36) {
if (Rs < 4686) {
return 1;
}
else if (Rs >= 4686) {
return 2;
}
}
else if (negBalance >= 36) {
return 1;
}
}
else if (negBalance >= 52) {
if (Cs < 996) {
if (ssDCval < 153) {
return 2;
}
else if (ssDCval >= 153) {
return 1;
}
}
else if (Cs >= 996) {
if (Rs < 5557) {
return 1;
}
else if (Rs >= 5557) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
if (Rs < 5655) {
return 8;
}
else if (Rs >= 5655) {
return 1;
}
}
}
}
}
}
else if (SC >= 7243) {
if (Cs < 1218) {
return 2;
}
else if (Cs >= 1218) {
if (Cs < 1238) {
return 8;
}
else if (Cs >= 1238) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1325) {
if (mu_mv_mag_sq < 107) {
if (RsCsDiff < 101) {
if (ssDCval < 90) {
if (posBalance < 45) {
if (Rs < 8916) {
if (Rs < 849) {
if (Rs < 312) {
return 4;
}
else if (Rs >= 312) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (Rs >= 849) {
if (negBalance < 32) {
if (ssDCval < 64) {
if (Rs < 4294) {
if (ssDCval < 55) {
if (AFD < 8) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (AFD >= 8) {
return 2;
}
}
else if (ssDCval >= 55) {
if (RsCsDiff < 16) {
if (Cs < 1519) {
return 2;
}
else if (Cs >= 1519) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (Rs < 1369) {
return 4;
}
else if (Rs >= 1369) {
return 2;
}
}
}
}
else if (Rs >= 4294) {
if (Cs < 3939) {
return 4;
}
else if (Cs >= 3939) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (ssDCval < 69) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (ssDCval >= 69) {
if (Rs < 3540) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (tcor < 90) {
if (Cs < 3139) {
return 1;
}
else if (Cs >= 3139) {
return 4;
}
}
else if (tcor >= 90) {
if (diffAFD < -2) {
if (Rs < 1861) {
return 8;
}
else if (Rs >= 1861) {
return 1;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
else if (Rs >= 3540) {
if (RsCsDiff < 57) {
if (Cs < 1751) {
if (Rs < 4342) {
return 4;
}
else if (Rs >= 4342) {
return 8;
}
}
else if (Cs >= 1751) {
return 4;
}
}
else if (RsCsDiff >= 57) {
if (Rs < 5707) {
return 8;
}
else if (Rs >= 5707) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 32) {
if (Rs < 3591) {
if (MVDiff < 8) {
if (CsDiff < 6) {
if (Rs < 3262) {
if (AFD < 8) {
if (ssDCval < 84) {
if (MVDiff < 5) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
if (diffRsCsdiff < -19) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (diffRsCsdiff >= -19) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (ssDCval >= 84) {
if (Rs < 2900) {
return 4;
}
else if (Rs >= 2900) {
return 8;
}
}
}
else if (AFD >= 8) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
if (posBalance < 42) {
if (Rs < 2850) {
return 1;
}
else if (Rs >= 2850) {
return 2;
}
}
else if (posBalance >= 42) {
return 8;
}
}
}
}
else if (Rs >= 3262) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
if (Rs < 3388) {
return 4;
}
else if (Rs >= 3388) {
if (Rs < 3551) {
return 2;
}
else if (Rs >= 3551) {
return 1;
}
}
}
}
}
else if (CsDiff >= 6) {
if (negBalance < 38) {
if (Rs < 1590) {
return 8;
}
else if (Rs >= 1590) {
return 1;
}
}
else if (negBalance >= 38) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (ssDCval < 71) {
if (posBalance < 29) {
return 8;
}
else if (posBalance >= 29) {
if (mu_mv_mag_sq < 74) {
if (Rs < 1920) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
if (Rs < 1655) {
return 4;
}
else if (Rs >= 1655) {
return 2;
}
}
}
else if (Rs >= 1920) {
return 8;
}
}
else if (mu_mv_mag_sq >= 74) {
if (diffTSC < -2) {
return 8;
}
else if (diffTSC >= -2) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (diffRsCsdiff < 4) {
return 4;
}
else if (diffRsCsdiff >= 4) {
if (Rs < 1355) {
return 4;
}
else if (Rs >= 1355) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (SC < 2990) {
if (Rs < 1140) {
return 2;
}
else if (Rs >= 1140) {
if (Rs < 1386) {
return 4;
}
else if (Rs >= 1386) {
return 8;
}
}
}
else if (SC >= 2990) {
if (RsDiff < 4) {
if (Rs < 2334) {
return 1;
}
else if (Rs >= 2334) {
return 8;
}
}
else if (RsDiff >= 4) {
return 8;
}
}
}
}
}
else if (Rs >= 3591) {
if (mu_mv_mag_sq < 101) {
if (CsDiff < 6) {
if (negBalance < 44) {
if (AFD < 7) {
if (Cs < 1571) {
return 4;
}
else if (Cs >= 1571) {
if (Rs < 5124) {
return 8;
}
else if (Rs >= 5124) {
return 4;
}
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (negBalance >= 44) {
if (Rs < 3806) {
return 2;
}
else if (Rs >= 3806) {
if (Rs < 4508) {
return 1;
}
else if (Rs >= 4508) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (ssDCval < 79) {
return 4;
}
else if (ssDCval >= 79) {
if (Rs < 4829) {
return 8;
}
else if (Rs >= 4829) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
}
else if (Rs >= 8916) {
return 2;
}
}
else if (posBalance >= 45) {
if (negBalance < 37) {
if (Rs < 2949) {
if (mu_mv_mag_sq < 78) {
if (TSC < 6) {
if (Cs < 1541) {
if (MVDiff < 7) {
if (tcor < 90) {
if (Rs < 1223) {
return 4;
}
else if (Rs >= 1223) {
return 8;
}
}
else if (tcor >= 90) {
if (Rs < 1483) {
return 2;
}
else if (Rs >= 1483) {
return 8;
}
}
}
else if (MVDiff >= 7) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (Cs >= 1541) {
if (MVDiff < 109) {
return 1;
}
else if (MVDiff >= 109) {
return 8;
}
}
}
else if (TSC >= 6) {
if (Cs < 2284) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1608) {
if (Rs < 1312) {
return 8;
}
else if (Rs >= 1312) {
return 4;
}
}
else if (Rs >= 1608) {
return 8;
}
}
}
else if (Cs >= 2284) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (SC < 4787) {
if (Cs < 1337) {
return 2;
}
else if (Cs >= 1337) {
if (MVDiff < 6) {
if (Rs < 2405) {
if (ssDCval < 44) {
return 1;
}
else if (ssDCval >= 44) {
if (ssDCval < 54) {
if (Cs < 1535) {
return 8;
}
else if (Cs >= 1535) {
return 4;
}
}
else if (ssDCval >= 54) {
return 8;
}
}
}
else if (Rs >= 2405) {
if (Cs < 1690) {
return 4;
}
else if (Cs >= 1690) {
if (Rs < 2664) {
if (Rs < 2407) {
return 4;
}
else if (Rs >= 2407) {
return 8;
}
}
else if (Rs >= 2664) {
if (Cs < 1961) {
return 1;
}
else if (Cs >= 1961) {
return 4;
}
}
}
}
}
else if (MVDiff >= 6) {
if (mcTcor < 98) {
if (AFD < 5) {
if (Rs < 1027) {
if (Rs < 665) {
return 1;
}
else if (Rs >= 665) {
return 8;
}
}
else if (Rs >= 1027) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (AFD >= 5) {
if (mu_mv_mag_sq < 90) {
if (RsCsDiff < 16) {
if (Cs < 1643) {
if (Rs < 1150) {
return 2;
}
else if (Rs >= 1150) {
return 8;
}
}
else if (Cs >= 1643) {
return 1;
}
}
else if (RsCsDiff >= 16) {
if (Cs < 1511) {
return 4;
}
else if (Cs >= 1511) {
if (diffRsCsdiff < -25) {
return 4;
}
else if (diffRsCsdiff >= -25) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 96) {
return 1;
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 707) {
return 1;
}
else if (Rs >= 707) {
return 8;
}
}
}
}
}
}
else if (SC >= 4787) {
if (RsCsDiff < 46) {
return 4;
}
else if (RsCsDiff >= 46) {
if (Rs < 2940) {
return 1;
}
else if (Rs >= 2940) {
return 8;
}
}
}
}
}
else if (Rs >= 2949) {
if (Cs < 2175) {
if (Rs < 3422) {
if (Cs < 1493) {
if (diffRsCsdiff < 8) {
return 2;
}
else if (diffRsCsdiff >= 8) {
if (Rs < 3199) {
return 2;
}
else if (Rs >= 3199) {
return 4;
}
}
}
else if (Cs >= 1493) {
if (diffAFD < -3) {
if (Cs < 1643) {
return 2;
}
else if (Cs >= 1643) {
return 8;
}
}
else if (diffAFD >= -3) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
return 4;
}
}
}
}
else if (Rs >= 3422) {
return 2;
}
}
else if (Cs >= 2175) {
if (CsDiff < 6) {
if (ssDCval < 56) {
if (negBalance < 29) {
if (Rs < 3792) {
return 8;
}
else if (Rs >= 3792) {
return 8;
}
}
else if (negBalance >= 29) {
return 4;
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (CsDiff >= 6) {
if (Rs < 4827) {
return 1;
}
else if (Rs >= 4827) {
return 2;
}
}
}
}
}
else if (negBalance >= 37) {
if (Rs < 1459) {
if (ssDCval < 56) {
if (MVDiff < 19) {
if (AFD < 8) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Cs < 1619) {
if (Rs < 1137) {
return 2;
}
else if (Rs >= 1137) {
return 4;
}
}
else if (Cs >= 1619) {
return 8;
}
}
}
else if (AFD >= 8) {
return 4;
}
}
else if (MVDiff >= 19) {
if (RsCsDiff < 91) {
return 2;
}
else if (RsCsDiff >= 91) {
if (SC < 2765) {
return 4;
}
else if (SC >= 2765) {
return 8;
}
}
}
}
else if (ssDCval >= 56) {
if (posBalance < 59) {
if (RsCsDiff < 33) {
if (ssDCval < 84) {
if (SC < 2440) {
if (Rs < 529) {
return 1;
}
else if (Rs >= 529) {
if (Rs < 875) {
if (Rs < 675) {
return 2;
}
else if (Rs >= 675) {
return 8;
}
}
else if (Rs >= 875) {
if (Rs < 922) {
return 2;
}
else if (Rs >= 922) {
return 4;
}
}
}
}
else if (SC >= 2440) {
if (refDCval < 61) {
if (Rs < 1144) {
return 4;
}
else if (Rs >= 1144) {
return 8;
}
}
else if (refDCval >= 61) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 99) {
return 1;
}
else if (mu_mv_mag_sq >= 99) {
if (Cs < 1386) {
return 4;
}
else if (Cs >= 1386) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (diffRsCsdiff < -15) {
return 4;
}
else if (diffRsCsdiff >= -15) {
return 8;
}
}
}
else if (RsCsDiff >= 33) {
if (mcTcor < 96) {
if (mu_mv_mag_sq < 2) {
if (SC < 2730) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
if (Cs < 1358) {
return 8;
}
else if (Cs >= 1358) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
}
}
else if (SC >= 2730) {
if (Cs < 1380) {
return 1;
}
else if (Cs >= 1380) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (MVDiff < 12) {
if (Cs < 1383) {
if (RsCsDiff < 49) {
if (Rs < 1217) {
return 1;
}
else if (Rs >= 1217) {
return 8;
}
}
else if (RsCsDiff >= 49) {
return 8;
}
}
else if (Cs >= 1383) {
return 8;
}
}
else if (MVDiff >= 12) {
if (posBalance < 53) {
if (Cs < 1387) {
return 4;
}
else if (Cs >= 1387) {
return 8;
}
}
else if (posBalance >= 53) {
if (SC < 2379) {
return 8;
}
else if (SC >= 2379) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (refDCval < 75) {
if (Cs < 1720) {
return 2;
}
else if (Cs >= 1720) {
if (Cs < 2144) {
return 8;
}
else if (Cs >= 2144) {
return 2;
}
}
}
else if (refDCval >= 75) {
return 8;
}
}
}
}
}
else if (posBalance >= 59) {
if (RsCsDiff < 73) {
if (refDCval < 76) {
if (diffRsCsdiff < 27) {
return 8;
}
else if (diffRsCsdiff >= 27) {
return 2;
}
}
else if (refDCval >= 76) {
if (mu_mv_mag_sq < 3) {
if (gchDC < 2) {
if (TSC < 7) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 1648) {
return 8;
}
else if (Cs >= 1648) {
return 4;
}
}
}
}
else if (RsCsDiff >= 73) {
if (Cs < 1392) {
if (Rs < 1219) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (Rs >= 1219) {
return 8;
}
}
else if (Cs >= 1392) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1459) {
if (refDCval < 74) {
if (Cs < 3026) {
if (SC < 5637) {
if (TSC < 5) {
if (RsDiff < 6) {
if (Rs < 1966) {
return 4;
}
else if (Rs >= 1966) {
return 8;
}
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (TSC >= 5) {
if (posBalance < 69) {
if (RsCsDiff < 87) {
if (mcTcor < 94) {
if (ssDCval < 57) {
if (AFD < 13) {
if (negBalance < 38) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (AFD >= 13) {
return 4;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (mcTcor >= 94) {
if (diffAFD < 3) {
if (Rs < 2586) {
if (diffRsCsdiff < -27) {
if (Rs < 1901) {
return 8;
}
else if (Rs >= 1901) {
return 4;
}
}
else if (diffRsCsdiff >= -27) {
return 8;
}
}
else if (Rs >= 2586) {
if (negBalance < 41) {
if (ssDCval < 65) {
if (Rs < 2843) {
return 4;
}
else if (Rs >= 2843) {
return 8;
}
}
else if (ssDCval >= 65) {
return 2;
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
else if (diffAFD >= 3) {
return 4;
}
}
}
else if (RsCsDiff >= 87) {
if (tcor < 57) {
if (Rs < 1656) {
return 4;
}
else if (Rs >= 1656) {
return 1;
}
}
else if (tcor >= 57) {
if (TSC < 8) {
if (Cs < 1617) {
if (Cs < 1388) {
if (Rs < 2780) {
return 4;
}
else if (Rs >= 2780) {
return 2;
}
}
else if (Cs >= 1388) {
return 8;
}
}
else if (Cs >= 1617) {
return 4;
}
}
else if (TSC >= 8) {
if (mcTcor < 77) {
if (Rs < 1568) {
return 4;
}
else if (Rs >= 1568) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
}
}
else if (mcTcor >= 77) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
if (Rs < 2045) {
return 1;
}
else if (Rs >= 2045) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 69) {
return 4;
}
}
}
else if (SC >= 5637) {
return 4;
}
}
else if (Cs >= 3026) {
return 8;
}
}
else if (refDCval >= 74) {
if (mcTcor < 94) {
if (Cs < 2918) {
if (Rs < 1754) {
if (mu_mv_mag_sq < 6) {
if (AFD < 16) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
return 8;
}
}
else if (AFD >= 16) {
return 4;
}
}
else if (mu_mv_mag_sq >= 6) {
return 8;
}
}
else if (Rs >= 1754) {
if (RsDiff < 7) {
if (mu_mv_mag_sq < 97) {
if (MVDiff < 21) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Rs < 2077) {
if (Rs < 1880) {
return 8;
}
else if (Rs >= 1880) {
return 1;
}
}
else if (Rs >= 2077) {
return 8;
}
}
}
else if (MVDiff >= 21) {
if (Rs < 1955) {
return 1;
}
else if (Rs >= 1955) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (RsDiff >= 7) {
if (TSC < 10) {
if (SC < 3682) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (AFD < 14) {
if (SC < 3512) {
return 8;
}
else if (SC >= 3512) {
return 2;
}
}
else if (AFD >= 14) {
return 4;
}
}
}
else if (SC >= 3682) {
if (RsDiff < 8) {
if (AFD < 17) {
if (AFD < 9) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Rs < 2295) {
return 8;
}
else if (Rs >= 2295) {
return 4;
}
}
}
else if (AFD >= 9) {
return 8;
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (RsDiff >= 8) {
if (Rs < 2617) {
if (Rs < 2431) {
return 8;
}
else if (Rs >= 2431) {
return 4;
}
}
else if (Rs >= 2617) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
}
}
}
else if (TSC >= 10) {
if (AFD < 19) {
if (AFD < 12) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 4;
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (AFD >= 19) {
if (Rs < 2409) {
return 8;
}
else if (Rs >= 2409) {
return 4;
}
}
}
}
}
}
else if (Cs >= 2918) {
if (Rs < 3000) {
return 4;
}
else if (Rs >= 3000) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (RsDiff < 8) {
if (AFD < 8) {
if (Cs < 1490) {
if (SC < 3483) {
if (SC < 3244) {
return 1;
}
else if (SC >= 3244) {
return 8;
}
}
else if (SC >= 3483) {
if (ssDCval < 85) {
if (RsCsDiff < 38) {
return 1;
}
else if (RsCsDiff >= 38) {
if (Rs < 2087) {
return 1;
}
else if (Rs >= 2087) {
return 2;
}
}
}
else if (ssDCval >= 85) {
if (Rs < 3077) {
return 4;
}
else if (Rs >= 3077) {
if (Rs < 3738) {
return 1;
}
else if (Rs >= 3738) {
return 2;
}
}
}
}
}
else if (Cs >= 1490) {
if (SC < 4764) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Cs < 1772) {
return 1;
}
else if (Cs >= 1772) {
return 8;
}
}
}
else if (SC >= 4764) {
if (Rs < 2782) {
if (Cs < 2571) {
return 8;
}
else if (Cs >= 2571) {
return 4;
}
}
else if (Rs >= 2782) {
if (Cs < 1815) {
if (Rs < 3286) {
return 4;
}
else if (Rs >= 3286) {
return 2;
}
}
else if (Cs >= 1815) {
return 8;
}
}
}
}
}
else if (AFD >= 8) {
if (MVDiff < 5) {
if (refDCval < 78) {
if (Rs < 2195) {
if (Rs < 2127) {
if (Cs < 1519) {
return 8;
}
else if (Cs >= 1519) {
return 4;
}
}
else if (Rs >= 2127) {
return 8;
}
}
else if (Rs >= 2195) {
if (Rs < 3592) {
return 1;
}
else if (Rs >= 3592) {
return 4;
}
}
}
else if (refDCval >= 78) {
if (SC < 4585) {
if (ssDCval < 80) {
if (Rs < 1928) {
return 4;
}
else if (Rs >= 1928) {
return 8;
}
}
else if (ssDCval >= 80) {
if (mu_mv_mag_sq < 91) {
if (Rs < 2408) {
if (AFD < 9) {
if (SC < 3844) {
if (ssDCval < 84) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (SC >= 3844) {
return 8;
}
}
else if (AFD >= 9) {
if (Rs < 1647) {
return 8;
}
else if (Rs >= 1647) {
return 1;
}
}
}
else if (Rs >= 2408) {
if (diffMVdiffVal < -1) {
if (Rs < 2924) {
return 8;
}
else if (Rs >= 2924) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (Cs < 2192) {
return 8;
}
else if (Cs >= 2192) {
return 1;
}
}
}
}
else if (SC >= 4585) {
if (RsCsDiff < 80) {
if (Cs < 2836) {
return 8;
}
else if (Cs >= 2836) {
if (Rs < 2870) {
return 4;
}
else if (Rs >= 2870) {
return 8;
}
}
}
else if (RsCsDiff >= 80) {
if (posBalance < 57) {
if (Rs < 3248) {
return 1;
}
else if (Rs >= 3248) {
if (Rs < 3445) {
return 4;
}
else if (Rs >= 3445) {
return 2;
}
}
}
else if (posBalance >= 57) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 5) {
if (tcor < 88) {
if (negBalance < 56) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
else if (negBalance >= 56) {
return 2;
}
}
else if (tcor >= 88) {
if (Rs < 1997) {
if (mu_mv_mag_sq < 45) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 45) {
if (Rs < 1793) {
return 4;
}
else if (Rs >= 1793) {
return 1;
}
}
}
else if (Rs >= 1997) {
if (TSC < 7) {
if (Rs < 2467) {
if (Rs < 2113) {
return 4;
}
else if (Rs >= 2113) {
return 8;
}
}
else if (Rs >= 2467) {
return 1;
}
}
else if (TSC >= 7) {
if (Rs < 3259) {
if (CsDiff < 5) {
if (tcor < 92) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (tcor >= 92) {
return 8;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (Rs >= 3259) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (Rs < 3810) {
return 4;
}
else if (Rs >= 3810) {
return 2;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (TSC < 8) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 2674) {
return 4;
}
else if (Rs >= 2674) {
if (Cs < 2302) {
return 2;
}
else if (Cs >= 2302) {
return 4;
}
}
}
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 90) {
if (Rs < 4426) {
if (refDCval < 163) {
if (refDCval < 110) {
if (Rs < 1441) {
if (Cs < 1957) {
if (SC < 2286) {
if (MVDiff < 33) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 4;
}
}
else if (MVDiff >= 33) {
return 8;
}
}
else if (SC >= 2286) {
if (mcTcor < 92) {
if (RsCsDiff < 63) {
if (diffRsCsdiff < 15) {
return 4;
}
else if (diffRsCsdiff >= 15) {
return 8;
}
}
else if (RsCsDiff >= 63) {
if (Rs < 1422) {
return 8;
}
else if (Rs >= 1422) {
return 4;
}
}
}
else if (mcTcor >= 92) {
if (RsCsDiff < 93) {
if (Cs < 1337) {
return 4;
}
else if (Cs >= 1337) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 4;
}
}
}
}
else if (RsCsDiff >= 93) {
if (gchDC < 2) {
if (Rs < 1038) {
return 4;
}
else if (Rs >= 1038) {
return 8;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
}
else if (Cs >= 1957) {
if (tcor < 78) {
if (ssDCval < 104) {
return 2;
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (tcor >= 78) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (Rs < 639) {
return 8;
}
else if (Rs >= 639) {
return 4;
}
}
}
}
}
else if (Rs >= 1441) {
if (SC < 4102) {
if (TSC < 11) {
if (Rs < 2224) {
if (mcTcor < 95) {
if (Rs < 2037) {
if (diffAFD < -3) {
if (SC < 3409) {
return 8;
}
else if (SC >= 3409) {
if (diffTSC < -1) {
if (Cs < 1591) {
return 2;
}
else if (Cs >= 1591) {
return 8;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
}
else if (diffAFD >= -3) {
if (Cs < 1346) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (Cs >= 1346) {
if (diffMVdiffVal < -4) {
if (Cs < 1441) {
if (Rs < 1834) {
return 8;
}
else if (Rs >= 1834) {
return 2;
}
}
else if (Cs >= 1441) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
}
else if (Rs >= 2037) {
if (mu_mv_mag_sq < 37) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (mu_mv_mag_sq >= 37) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
if (posBalance < 62) {
if (posBalance < 53) {
if (Rs < 2107) {
return 2;
}
else if (Rs >= 2107) {
return 8;
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (posBalance >= 62) {
if (Rs < 2078) {
return 8;
}
else if (Rs >= 2078) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (posBalance < 58) {
if (AFD < 14) {
if (diffRsCsdiff < -36) {
if (Cs < 1580) {
return 2;
}
else if (Cs >= 1580) {
return 1;
}
}
else if (diffRsCsdiff >= -36) {
if (AFD < 12) {
if (SC < 4060) {
if (Cs < 1351) {
return 2;
}
else if (Cs >= 1351) {
if (Rs < 1923) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 103) {
return 8;
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 1875) {
return 8;
}
else if (Rs >= 1875) {
return 2;
}
}
}
else if (mcTcor >= 98) {
return 2;
}
}
else if (Rs >= 1923) {
return 8;
}
}
}
else if (SC >= 4060) {
if (Rs < 2116) {
return 1;
}
else if (Rs >= 2116) {
return 2;
}
}
}
else if (AFD >= 12) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 49) {
if (Rs < 1917) {
return 2;
}
else if (Rs >= 1917) {
return 8;
}
}
else if (mu_mv_mag_sq >= 49) {
return 1;
}
}
}
}
}
else if (AFD >= 14) {
if (Cs < 1373) {
return 1;
}
else if (Cs >= 1373) {
return 2;
}
}
}
else if (posBalance >= 58) {
if (diffRsCsdiff < -5) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 1872) {
return 2;
}
else if (Rs >= 1872) {
return 1;
}
}
}
else if (diffRsCsdiff >= -5) {
if (Cs < 1844) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (posBalance < 79) {
if (Rs < 1593) {
return 8;
}
else if (Rs >= 1593) {
return 2;
}
}
else if (posBalance >= 79) {
return 8;
}
}
}
else if (Cs >= 1844) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
if (Rs < 1764) {
return 8;
}
else if (Rs >= 1764) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 2224) {
if (posBalance < 44) {
if (Cs < 1713) {
return 1;
}
else if (Cs >= 1713) {
return 8;
}
}
else if (posBalance >= 44) {
if (Rs < 2471) {
if (diffMVdiffVal < -14) {
if (Rs < 2328) {
return 2;
}
else if (Rs >= 2328) {
return 8;
}
}
else if (diffMVdiffVal >= -14) {
return 8;
}
}
else if (Rs >= 2471) {
if (negBalance < 43) {
if (Cs < 1442) {
return 4;
}
else if (Cs >= 1442) {
return 1;
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
}
else if (TSC >= 11) {
if (TSC < 13) {
if (ssDCval < 100) {
return 8;
}
else if (ssDCval >= 100) {
if (Rs < 1750) {
return 4;
}
else if (Rs >= 1750) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
}
}
else if (TSC >= 13) {
if (Rs < 1626) {
return 8;
}
else if (Rs >= 1626) {
return 1;
}
}
}
}
else if (SC >= 4102) {
if (mcTcor < 96) {
if (Cs < 2698) {
if (negBalance < 56) {
if (CsDiff < 1) {
if (Rs < 2319) {
return 1;
}
else if (Rs >= 2319) {
return 4;
}
}
else if (CsDiff >= 1) {
if (RsCsDiff < 99) {
if (refDCval < 107) {
if (Rs < 1894) {
return 4;
}
else if (Rs >= 1894) {
if (diffAFD < -6) {
if (Rs < 3445) {
return 4;
}
else if (Rs >= 3445) {
return 8;
}
}
else if (diffAFD >= -6) {
if (refDCval < 92) {
if (RsCsDiff < 73) {
return 8;
}
else if (RsCsDiff >= 73) {
if (mcTcor < 94) {
if (Rs < 4155) {
return 8;
}
else if (Rs >= 4155) {
return 4;
}
}
else if (mcTcor >= 94) {
if (Rs < 3430) {
return 4;
}
else if (Rs >= 3430) {
return 4;
}
}
}
}
else if (refDCval >= 92) {
if (mcTcor < 95) {
if (Rs < 4127) {
return 8;
}
else if (Rs >= 4127) {
if (Rs < 4137) {
return 2;
}
else if (Rs >= 4137) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (Rs < 3117) {
if (mu_mv_mag_sq < 50) {
return 4;
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 3094) {
return 8;
}
else if (Rs >= 3094) {
return 4;
}
}
}
else if (Rs >= 3117) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 107) {
if (tcor < 94) {
if (ssDCval < 107) {
if (Rs < 3445) {
return 4;
}
else if (Rs >= 3445) {
return 8;
}
}
else if (ssDCval >= 107) {
if (TSC < 7) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
return 8;
}
}
else if (TSC >= 7) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
else if (tcor >= 94) {
return 4;
}
}
}
else if (RsCsDiff >= 99) {
if (Cs < 1579) {
return 4;
}
else if (Cs >= 1579) {
if (Rs < 3264) {
return 8;
}
else if (Rs >= 3264) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (SC < 5431) {
return 1;
}
else if (SC >= 5431) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 56) {
if (Cs < 1734) {
return 8;
}
else if (Cs >= 1734) {
return 4;
}
}
}
else if (Cs >= 2698) {
if (ssDCval < 107) {
if (Rs < 4249) {
return 4;
}
else if (Rs >= 4249) {
return 8;
}
}
else if (ssDCval >= 107) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (RsCsDiff < 31) {
if (CsDiff < 1) {
if (posBalance < 11) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
return 1;
}
}
else if (posBalance >= 11) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
return 4;
}
}
}
}
else if (CsDiff >= 1) {
if (RsCsDiff < 18) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
return 8;
}
}
else if (RsCsDiff >= 18) {
if (MVDiff < 4) {
if (mcTcor < 98) {
if (Cs < 1576) {
return 4;
}
else if (Cs >= 1576) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (mcTcor >= 98) {
if (Cs < 1477) {
return 4;
}
else if (Cs >= 1477) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (SC < 4508) {
return 4;
}
else if (SC >= 4508) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 31) {
if (Rs < 3472) {
if (tcor < 93) {
if (RsCsDiff < 63) {
return 4;
}
else if (RsCsDiff >= 63) {
return 8;
}
}
else if (tcor >= 93) {
if (MVDiff < 4) {
if (mcTcor < 97) {
if (diffRsCsdiff < 21) {
return 4;
}
else if (diffRsCsdiff >= 21) {
if (Rs < 2457) {
return 8;
}
else if (Rs >= 2457) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (Cs < 1626) {
return 4;
}
else if (Cs >= 1626) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 3) {
if (AFD < 9) {
if (Cs < 1445) {
return 4;
}
else if (Cs >= 1445) {
return 8;
}
}
else if (AFD >= 9) {
if (Rs < 2562) {
return 2;
}
else if (Rs >= 2562) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
else if (Rs >= 3472) {
if (SC < 5949) {
if (RsCsDiff < 38) {
if (ssDCval < 103) {
return 4;
}
else if (ssDCval >= 103) {
return 8;
}
}
else if (RsCsDiff >= 38) {
if (ssDCval < 97) {
if (Rs < 4171) {
return 2;
}
else if (Rs >= 4171) {
return 8;
}
}
else if (ssDCval >= 97) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
else if (SC >= 5949) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (Rs < 3933) {
return 4;
}
else if (Rs >= 3933) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 110) {
if (Rs < 1488) {
if (mu_mv_mag_sq < 36) {
if (ssDCval < 144) {
if (RsCsDiff < 94) {
if (CsDiff < 4) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
return 1;
}
}
else if (CsDiff >= 4) {
if (refDCval < 142) {
if (Rs < 1484) {
return 8;
}
else if (Rs >= 1484) {
return 4;
}
}
else if (refDCval >= 142) {
return 4;
}
}
}
else if (RsCsDiff >= 94) {
if (diffRsCsdiff < 14) {
if (Cs < 1685) {
return 4;
}
else if (Cs >= 1685) {
return 2;
}
}
else if (diffRsCsdiff >= 14) {
if (Cs < 1602) {
return 1;
}
else if (Cs >= 1602) {
return 8;
}
}
}
}
else if (ssDCval >= 144) {
return 2;
}
}
else if (mu_mv_mag_sq >= 36) {
if (Rs < 1243) {
if (Rs < 972) {
if (ssDCval < 115) {
if (Rs < 750) {
return 1;
}
else if (Rs >= 750) {
return 4;
}
}
else if (ssDCval >= 115) {
return 8;
}
}
else if (Rs >= 972) {
if (CsDiff < 6) {
if (refDCval < 130) {
if (negBalance < 35) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 8;
}
}
else if (negBalance >= 35) {
return 8;
}
}
else if (refDCval >= 130) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
if (Cs < 1830) {
if (MVDiff < 19) {
if (mu_mv_mag_sq < 106) {
return 4;
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 1038) {
return 4;
}
else if (Rs >= 1038) {
return 8;
}
}
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (Cs >= 1830) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (Rs >= 1243) {
if (Cs < 1744) {
if (RsDiff < 3) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
if (Rs < 1337) {
return 4;
}
else if (Rs >= 1337) {
return 1;
}
}
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (Cs >= 1744) {
if (negBalance < 39) {
return 8;
}
else if (negBalance >= 39) {
return 4;
}
}
}
}
}
else if (Rs >= 1488) {
if (refDCval < 140) {
if (ssDCval < 124) {
if (AFD < 13) {
if (Rs < 4225) {
if (diffAFD < 3) {
if (Cs < 1376) {
if (refDCval < 122) {
return 8;
}
else if (refDCval >= 122) {
if (Rs < 3846) {
return 2;
}
else if (Rs >= 3846) {
return 8;
}
}
}
else if (Cs >= 1376) {
if (diffAFD < -2) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (diffAFD >= 3) {
if (RsDiff < 6) {
if (tcor < 104) {
return 8;
}
else if (tcor >= 104) {
return 2;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (Rs >= 4225) {
if (Cs < 1497) {
return 1;
}
else if (Cs >= 1497) {
return 8;
}
}
}
else if (AFD >= 13) {
if (SC < 3453) {
if (TSC < 10) {
if (Cs < 1467) {
return 2;
}
else if (Cs >= 1467) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (SC >= 3453) {
if (Cs < 1616) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 1;
}
}
else if (Cs >= 1616) {
if (Rs < 1546) {
return 1;
}
else if (Rs >= 1546) {
return 8;
}
}
}
}
}
else if (ssDCval >= 124) {
if (mcTcor < 92) {
if (CsDiff < 5) {
if (SC < 3776) {
if (Rs < 1550) {
if (Rs < 1503) {
return 1;
}
else if (Rs >= 1503) {
return 8;
}
}
else if (Rs >= 1550) {
return 1;
}
}
else if (SC >= 3776) {
return 8;
}
}
else if (CsDiff >= 5) {
if (Rs < 3622) {
if (posBalance < 68) {
if (mcTcor < 85) {
if (ssDCval < 135) {
return 8;
}
else if (ssDCval >= 135) {
return 1;
}
}
else if (mcTcor >= 85) {
return 8;
}
}
else if (posBalance >= 68) {
if (Rs < 1601) {
return 1;
}
else if (Rs >= 1601) {
if (Rs < 2161) {
return 4;
}
else if (Rs >= 2161) {
return 8;
}
}
}
}
else if (Rs >= 3622) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 3117) {
if (refDCval < 128) {
if (RsDiff < 7) {
if (posBalance < 50) {
if (Cs < 1869) {
if (Rs < 2123) {
return 2;
}
else if (Rs >= 2123) {
return 1;
}
}
else if (Cs >= 1869) {
return 8;
}
}
else if (posBalance >= 50) {
return 8;
}
}
else if (RsDiff >= 7) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 1;
}
}
}
else if (refDCval >= 128) {
if (mu_mv_mag_sq < 90) {
if (Cs < 2596) {
return 8;
}
else if (Cs >= 2596) {
return 4;
}
}
else if (mu_mv_mag_sq >= 90) {
if (refDCval < 136) {
if (Rs < 2564) {
return 1;
}
else if (Rs >= 2564) {
return 8;
}
}
else if (refDCval >= 136) {
return 8;
}
}
}
}
else if (Rs >= 3117) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 2;
}
}
}
}
}
else if (refDCval >= 140) {
if (negBalance < 58) {
if (posBalance < 68) {
if (SC < 5884) {
if (diffTSC < -1) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2150) {
return 2;
}
else if (Rs >= 2150) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 2;
}
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (SC >= 5884) {
return 1;
}
}
else if (posBalance >= 68) {
if (Rs < 1925) {
return 4;
}
else if (Rs >= 1925) {
return 8;
}
}
}
else if (negBalance >= 58) {
if (Rs < 1704) {
return 2;
}
else if (Rs >= 1704) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 163) {
if (Rs < 2306) {
if (Rs < 1098) {
return 1;
}
else if (Rs >= 1098) {
return 2;
}
}
else if (Rs >= 2306) {
return 8;
}
}
}
else if (Rs >= 4426) {
if (tcor < 95) {
if (mu_mv_mag_sq < 23) {
if (SC < 6176) {
return 1;
}
else if (SC >= 6176) {
return 8;
}
}
else if (mu_mv_mag_sq >= 23) {
if (ssDCval < 118) {
if (Cs < 2835) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (Cs >= 2835) {
return 8;
}
}
else if (ssDCval >= 118) {
if (Rs < 4526) {
return 1;
}
else if (Rs >= 4526) {
return 8;
}
}
}
}
else if (tcor >= 95) {
if (Cs < 1590) {
return 2;
}
else if (Cs >= 1590) {
if (Rs < 6227) {
if (Cs < 1854) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
return 2;
}
}
}
else if (Cs >= 1854) {
return 8;
}
}
else if (Rs >= 6227) {
if (Rs < 7227) {
return 4;
}
else if (Rs >= 7227) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1939) {
if (Rs < 3320) {
if (mcTcor < 88) {
if (Rs < 1683) {
if (ssDCval < 92) {
if (posBalance < 55) {
if (negBalance < 32) {
if (Rs < 1622) {
return 4;
}
else if (Rs >= 1622) {
return 8;
}
}
else if (negBalance >= 32) {
if (diffAFD < 1) {
if (SC < 3079) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (SC >= 3079) {
if (Rs < 1568) {
return 4;
}
else if (Rs >= 1568) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 4;
}
}
}
}
else if (posBalance >= 55) {
if (mu_mv_mag_sq < 9) {
if (mcTcor < 74) {
if (Rs < 878) {
return 8;
}
else if (Rs >= 878) {
return 4;
}
}
else if (mcTcor >= 74) {
if (CsDiff < 10) {
if (SC < 3133) {
if (Rs < 1154) {
if (Cs < 1670) {
return 4;
}
else if (Cs >= 1670) {
return 8;
}
}
else if (Rs >= 1154) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 4;
}
}
}
else if (SC >= 3133) {
if (Cs < 1672) {
return 1;
}
else if (Cs >= 1672) {
return 4;
}
}
}
else if (CsDiff >= 10) {
if (Rs < 1196) {
if (Rs < 1014) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 2;
}
}
else if (Rs >= 1014) {
return 2;
}
}
else if (Rs >= 1196) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (negBalance < 46) {
if (RsCsDiff < 155) {
return 8;
}
else if (RsCsDiff >= 155) {
if (Rs < 1301) {
return 4;
}
else if (Rs >= 1301) {
return 1;
}
}
}
else if (negBalance >= 46) {
if (Cs < 1482) {
if (Rs < 1441) {
if (Cs < 1469) {
return 4;
}
else if (Cs >= 1469) {
return 8;
}
}
else if (Rs >= 1441) {
return 8;
}
}
else if (Cs >= 1482) {
return 4;
}
}
}
}
}
else if (ssDCval >= 92) {
if (negBalance < 59) {
if (SC < 3115) {
if (MVDiff < 376) {
return 8;
}
else if (MVDiff >= 376) {
return 1;
}
}
else if (SC >= 3115) {
if (TSC < 14) {
if (SC < 3138) {
return 2;
}
else if (SC >= 3138) {
return 8;
}
}
else if (TSC >= 14) {
if (Rs < 1582) {
return 4;
}
else if (Rs >= 1582) {
return 1;
}
}
}
}
else if (negBalance >= 59) {
if (Cs < 1477) {
return 8;
}
else if (Cs >= 1477) {
return 2;
}
}
}
}
else if (Rs >= 1683) {
if (ssDCval < 113) {
if (RsDiff < 16) {
if (negBalance < 48) {
if (ssDCval < 59) {
if (Cs < 1749) {
if (refDCval < 39) {
return 8;
}
else if (refDCval >= 39) {
if (RsDiff < 9) {
if (mu_mv_mag_sq < 44) {
return 1;
}
else if (mu_mv_mag_sq >= 44) {
return 8;
}
}
else if (RsDiff >= 9) {
return 1;
}
}
}
else if (Cs >= 1749) {
if (tcor < 79) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 4;
}
}
else if (tcor >= 79) {
if (Rs < 2865) {
return 1;
}
else if (Rs >= 2865) {
return 8;
}
}
}
}
else if (ssDCval >= 59) {
if (posBalance < 63) {
if (TSC < 11) {
if (Cs < 1357) {
return 2;
}
else if (Cs >= 1357) {
return 8;
}
}
else if (TSC >= 11) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
if (Rs < 2664) {
return 2;
}
else if (Rs >= 2664) {
return 4;
}
}
}
}
else if (posBalance >= 63) {
if (Rs < 1864) {
return 4;
}
else if (Rs >= 1864) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
if (Rs < 2254) {
if (Cs < 1679) {
if (ssDCval < 76) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
return 2;
}
}
else if (ssDCval >= 76) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (MVDiff < 98) {
return 8;
}
else if (MVDiff >= 98) {
return 4;
}
}
}
}
else if (Cs >= 1679) {
if (AFD < 17) {
if (refDCval < 50) {
return 4;
}
else if (refDCval >= 50) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (Rs < 2122) {
return 8;
}
else if (Rs >= 2122) {
return 4;
}
}
}
}
else if (AFD >= 17) {
if (Rs < 2171) {
if (AFD < 24) {
if (SC < 4029) {
return 1;
}
else if (SC >= 4029) {
return 8;
}
}
else if (AFD >= 24) {
if (Rs < 1874) {
return 4;
}
else if (Rs >= 1874) {
return 8;
}
}
}
else if (Rs >= 2171) {
return 8;
}
}
}
}
else if (Rs >= 2254) {
if (mu_mv_mag_sq < 27) {
if (TSC < 14) {
if (RsCsDiff < 141) {
return 8;
}
else if (RsCsDiff >= 141) {
if (AFD < 24) {
if (refDCval < 109) {
if (negBalance < 52) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 2;
}
}
else if (negBalance >= 52) {
if (Rs < 3226) {
if (CsDiff < 11) {
return 4;
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (Rs >= 3226) {
return 8;
}
}
}
else if (refDCval >= 109) {
if (Rs < 2611) {
return 4;
}
else if (Rs >= 2611) {
return 8;
}
}
}
else if (AFD >= 24) {
if (RsDiff < 13) {
if (Rs < 2443) {
return 4;
}
else if (Rs >= 2443) {
return 1;
}
}
else if (RsDiff >= 13) {
return 2;
}
}
}
}
else if (TSC >= 14) {
if (Rs < 3104) {
return 8;
}
else if (Rs >= 3104) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (CsDiff < 7) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
else if (CsDiff >= 7) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
if (RsCsDiff < 283) {
return 4;
}
else if (RsCsDiff >= 283) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 16) {
if (mcTcor < 74) {
if (Cs < 1772) {
return 1;
}
else if (Cs >= 1772) {
return 8;
}
}
else if (mcTcor >= 74) {
if (mu_mv_mag_sq < 17) {
if (refDCval < 103) {
return 2;
}
else if (refDCval >= 103) {
if (Rs < 3150) {
return 4;
}
else if (Rs >= 3150) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (diffAFD < -1) {
if (Rs < 2247) {
return 2;
}
else if (Rs >= 2247) {
return 4;
}
}
else if (diffAFD >= -1) {
if (ssDCval < 59) {
if (Rs < 2528) {
return 4;
}
else if (Rs >= 2528) {
return 2;
}
}
else if (ssDCval >= 59) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (TSC < 13) {
if (Rs < 2247) {
return 8;
}
else if (Rs >= 2247) {
if (AFD < 17) {
if (tcor < 73) {
if (AFD < 15) {
if (Rs < 2652) {
return 1;
}
else if (Rs >= 2652) {
return 4;
}
}
else if (AFD >= 15) {
return 8;
}
}
else if (tcor >= 73) {
return 8;
}
}
else if (AFD >= 17) {
if (diffRsCsdiff < -10) {
return 8;
}
else if (diffRsCsdiff >= -10) {
return 1;
}
}
}
}
else if (TSC >= 13) {
if (MVDiff < 40) {
return 1;
}
else if (MVDiff >= 40) {
if (AFD < 33) {
if (MVDiff < 50) {
if (RsDiff < 12) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
return 1;
}
}
else if (RsDiff >= 12) {
return 8;
}
}
else if (MVDiff >= 50) {
return 8;
}
}
else if (AFD >= 33) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 88) {
if (negBalance < 48) {
if (ssDCval < 100) {
if (TSC < 6) {
if (Rs < 419) {
return 4;
}
else if (Rs >= 419) {
return 1;
}
}
else if (TSC >= 6) {
if (Rs < 2716) {
if (tcor < 81) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 15) {
if (Cs < 1360) {
return 8;
}
else if (Cs >= 1360) {
if (SC < 4143) {
if (AFD < 27) {
return 2;
}
else if (AFD >= 27) {
return 1;
}
}
else if (SC >= 4143) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (Rs < 2436) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (Cs < 1478) {
return 2;
}
else if (Cs >= 1478) {
return 4;
}
}
}
else if (Rs >= 2436) {
if (Cs < 1716) {
return 2;
}
else if (Cs >= 1716) {
return 4;
}
}
}
}
}
else if (tcor >= 81) {
if (posBalance < 55) {
if (Cs < 1415) {
if (RsCsDiff < 124) {
return 1;
}
else if (RsCsDiff >= 124) {
return 8;
}
}
else if (Cs >= 1415) {
if (CsDiff < 7) {
return 1;
}
else if (CsDiff >= 7) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (TSC < 8) {
if (refDCval < 86) {
if (Cs < 1698) {
if (Rs < 1536) {
return 1;
}
else if (Rs >= 1536) {
return 8;
}
}
else if (Cs >= 1698) {
return 4;
}
}
else if (refDCval >= 86) {
return 2;
}
}
else if (TSC >= 8) {
if (SC < 3718) {
if (Cs < 1451) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
else if (Cs >= 1451) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 2;
}
}
}
else if (SC >= 3718) {
if (mu_mv_mag_sq < 3) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2716) {
if (RsDiff < 10) {
if (Cs < 1706) {
if (negBalance < 44) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (SC < 4762) {
return 4;
}
else if (SC >= 4762) {
return 2;
}
}
}
else if (negBalance >= 44) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
else if (Cs >= 1706) {
return 1;
}
}
else if (RsDiff >= 10) {
if (ssDCval < 55) {
if (Rs < 2977) {
return 8;
}
else if (Rs >= 2977) {
return 1;
}
}
else if (ssDCval >= 55) {
if (diffMVdiffVal < -16) {
if (Rs < 2779) {
return 1;
}
else if (Rs >= 2779) {
return 8;
}
}
else if (diffMVdiffVal >= -16) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 100) {
if (Rs < 1424) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (TSC < 7) {
if (SC < 2576) {
return 2;
}
else if (SC >= 2576) {
return 8;
}
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (Rs >= 1424) {
if (tcor < 80) {
if (gchDC < 2) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
else if (gchDC >= 2) {
if (ssDCval < 118) {
if (Cs < 1493) {
return 1;
}
else if (Cs >= 1493) {
return 8;
}
}
else if (ssDCval >= 118) {
return 2;
}
}
}
else if (tcor >= 80) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 48) {
if (AFD < 20) {
if (mu_mv_mag_sq < 5) {
if (Rs < 3013) {
if (ssDCval < 87) {
if (mcTcor < 96) {
if (diffAFD < -1) {
if (MVDiff < 21) {
if (negBalance < 55) {
if (Cs < 1497) {
return 8;
}
else if (Cs >= 1497) {
return 4;
}
}
else if (negBalance >= 55) {
if (Cs < 1802) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 2;
}
}
else if (Cs >= 1802) {
return 8;
}
}
}
else if (MVDiff >= 21) {
return 1;
}
}
else if (diffAFD >= -1) {
if (TSC < 12) {
if (negBalance < 50) {
if (Cs < 1688) {
return 1;
}
else if (Cs >= 1688) {
if (Cs < 1877) {
return 8;
}
else if (Cs >= 1877) {
return 4;
}
}
}
else if (negBalance >= 50) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (tcor < 86) {
if (Cs < 1565) {
if (diffRsCsdiff < 14) {
return 8;
}
else if (diffRsCsdiff >= 14) {
return 4;
}
}
else if (Cs >= 1565) {
return 4;
}
}
else if (tcor >= 86) {
return 8;
}
}
}
}
else if (TSC >= 12) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 80) {
return 1;
}
else if (refDCval >= 80) {
return 2;
}
}
}
else if (ssDCval >= 87) {
if (posBalance < 65) {
if (posBalance < 52) {
if (AFD < 15) {
if (Rs < 2900) {
return 2;
}
else if (Rs >= 2900) {
return 8;
}
}
else if (AFD >= 15) {
return 4;
}
}
else if (posBalance >= 52) {
if (refDCval < 104) {
if (mcTcor < 97) {
if (TSC < 8) {
if (Rs < 1732) {
return 8;
}
else if (Rs >= 1732) {
return 2;
}
}
else if (TSC >= 8) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
else if (mcTcor >= 97) {
return 2;
}
}
else if (refDCval >= 104) {
if (MVDiff < 29) {
if (Rs < 2918) {
return 8;
}
else if (Rs >= 2918) {
if (Rs < 2967) {
return 4;
}
else if (Rs >= 2967) {
return 8;
}
}
}
else if (MVDiff >= 29) {
if (Rs < 2447) {
return 1;
}
else if (Rs >= 2447) {
return 8;
}
}
}
}
}
else if (posBalance >= 65) {
return 4;
}
}
}
else if (Rs >= 3013) {
if (Rs < 3189) {
if (negBalance < 55) {
return 4;
}
else if (negBalance >= 55) {
if (Rs < 3165) {
return 8;
}
else if (Rs >= 3165) {
return 4;
}
}
}
else if (Rs >= 3189) {
if (tcor < 79) {
if (Cs < 1370) {
return 1;
}
else if (Cs >= 1370) {
return 4;
}
}
else if (tcor >= 79) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
if (Rs < 3274) {
return 2;
}
else if (Rs >= 3274) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (refDCval < 76) {
if (TSC < 11) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (Rs < 1969) {
if (Rs < 1044) {
return 8;
}
else if (Rs >= 1044) {
return 1;
}
}
else if (Rs >= 1969) {
return 2;
}
}
}
else if (TSC >= 11) {
if (TSC < 13) {
return 1;
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (refDCval >= 76) {
if (diffMVdiffVal < -5) {
if (SC < 3709) {
if (diffRsCsdiff < 9) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (diffRsCsdiff >= 9) {
if (Rs < 1204) {
return 4;
}
else if (Rs >= 1204) {
return 1;
}
}
}
else if (SC >= 3709) {
if (Cs < 1673) {
if (negBalance < 53) {
if (ssDCval < 148) {
return 4;
}
else if (ssDCval >= 148) {
return 8;
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (Cs >= 1673) {
return 8;
}
}
}
else if (diffMVdiffVal >= -5) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (ssDCval < 87) {
if (TSC < 9) {
if (Rs < 1679) {
if (SC < 2727) {
return 4;
}
else if (SC >= 2727) {
return 8;
}
}
else if (Rs >= 1679) {
return 2;
}
}
else if (TSC >= 9) {
if (gchDC < 2) {
if (Cs < 1509) {
return 2;
}
else if (Cs >= 1509) {
if (negBalance < 54) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (negBalance >= 54) {
return 4;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (ssDCval >= 87) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (MVDiff < 57) {
if (mcTcor < 98) {
if (tcor < 77) {
if (MVDiff < 11) {
if (Rs < 1812) {
return 8;
}
else if (Rs >= 1812) {
return 2;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (tcor >= 77) {
return 8;
}
}
else if (mcTcor >= 98) {
return 2;
}
}
else if (MVDiff >= 57) {
return 2;
}
}
}
}
}
}
}
}
else if (AFD >= 20) {
if (RsDiff < 15) {
if (SC < 3436) {
if (MVDiff < 25) {
if (Rs < 714) {
return 2;
}
else if (Rs >= 714) {
if (AFD < 25) {
if (ssDCval < 80) {
if (SC < 3128) {
return 8;
}
else if (SC >= 3128) {
if (Rs < 1824) {
return 1;
}
else if (Rs >= 1824) {
return 2;
}
}
}
else if (ssDCval >= 80) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
if (diffAFD < 3) {
if (tcor < 80) {
if (Rs < 1730) {
if (Rs < 1599) {
return 4;
}
else if (Rs >= 1599) {
return 1;
}
}
else if (Rs >= 1730) {
return 8;
}
}
else if (tcor >= 80) {
return 1;
}
}
else if (diffAFD >= 3) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
return 2;
}
}
}
}
}
else if (AFD >= 25) {
if (RsCsDiff < 181) {
return 8;
}
else if (RsCsDiff >= 181) {
return 4;
}
}
}
}
else if (MVDiff >= 25) {
if (diffMVdiffVal < -15) {
return 2;
}
else if (diffMVdiffVal >= -15) {
if (diffTSC < -2) {
if (negBalance < 59) {
return 2;
}
else if (negBalance >= 59) {
return 4;
}
}
else if (diffTSC >= -2) {
if (diffMVdiffVal < 20) {
return 4;
}
else if (diffMVdiffVal >= 20) {
if (Cs < 1373) {
return 4;
}
else if (Cs >= 1373) {
return 8;
}
}
}
}
}
}
else if (SC >= 3436) {
if (TSC < 11) {
if (ssDCval < 99) {
if (CsDiff < 9) {
if (Rs < 1976) {
return 2;
}
else if (Rs >= 1976) {
return 8;
}
}
else if (CsDiff >= 9) {
if (ssDCval < 78) {
if (Cs < 1726) {
if (Rs < 2677) {
return 4;
}
else if (Rs >= 2677) {
return 8;
}
}
else if (Cs >= 1726) {
return 2;
}
}
else if (ssDCval >= 78) {
return 4;
}
}
}
else if (ssDCval >= 99) {
if (refDCval < 135) {
if (Rs < 2027) {
return 2;
}
else if (Rs >= 2027) {
return 1;
}
}
else if (refDCval >= 135) {
if (Rs < 2771) {
return 2;
}
else if (Rs >= 2771) {
return 8;
}
}
}
}
else if (TSC >= 11) {
if (refDCval < 76) {
if (Rs < 2333) {
return 4;
}
else if (Rs >= 2333) {
return 2;
}
}
else if (refDCval >= 76) {
if (Rs < 2857) {
if (tcor < 66) {
if (Rs < 2268) {
if (Rs < 2063) {
return 2;
}
else if (Rs >= 2063) {
return 4;
}
}
else if (Rs >= 2268) {
return 8;
}
}
else if (tcor >= 66) {
if (RsDiff < 9) {
if (refDCval < 121) {
return 8;
}
else if (refDCval >= 121) {
return 1;
}
}
else if (RsDiff >= 9) {
if (ssDCval < 75) {
if (Rs < 1977) {
return 8;
}
else if (Rs >= 1977) {
return 4;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
}
}
else if (Rs >= 2857) {
if (ssDCval < 127) {
return 4;
}
else if (ssDCval >= 127) {
if (Rs < 2960) {
return 2;
}
else if (Rs >= 2960) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 15) {
if (tcor < 74) {
if (Cs < 1726) {
return 2;
}
else if (Cs >= 1726) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 4;
}
}
}
}
else if (tcor >= 74) {
if (Cs < 1418) {
return 1;
}
else if (Cs >= 1418) {
if (SC < 4292) {
if (AFD < 27) {
return 2;
}
else if (AFD >= 27) {
return 8;
}
}
else if (SC >= 4292) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 3320) {
if (tcor < 87) {
if (RsDiff < 13) {
if (posBalance < 50) {
if (ssDCval < 94) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
if (Rs < 3538) {
return 8;
}
else if (Rs >= 3538) {
return 4;
}
}
}
}
else if (ssDCval >= 94) {
if (ssDCval < 137) {
if (Rs < 4293) {
return 8;
}
else if (Rs >= 4293) {
return 1;
}
}
else if (ssDCval >= 137) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (ssDCval < 85) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (diffAFD < -1) {
if (Rs < 3459) {
return 4;
}
else if (Rs >= 3459) {
return 2;
}
}
else if (diffAFD >= -1) {
if (Cs < 1890) {
return 8;
}
else if (Cs >= 1890) {
return 4;
}
}
}
}
else if (ssDCval >= 85) {
if (tcor < 67) {
if (Rs < 3522) {
return 8;
}
else if (Rs >= 3522) {
return 1;
}
}
else if (tcor >= 67) {
return 8;
}
}
}
}
else if (RsDiff >= 13) {
if (CsDiff < 6) {
if (mu_mv_mag_sq < 7) {
if (Cs < 1387) {
if (Rs < 3989) {
return 4;
}
else if (Rs >= 3989) {
return 2;
}
}
else if (Cs >= 1387) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
return 4;
}
}
else if (CsDiff >= 6) {
if (MVDiff < 10) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
if (mcTcor < 94) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 2;
}
}
else if (mcTcor >= 94) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (refDCval < 119) {
if (RsDiff < 17) {
if (posBalance < 50) {
if (Cs < 1568) {
return 2;
}
else if (Cs >= 1568) {
return 1;
}
}
else if (posBalance >= 50) {
if (AFD < 24) {
if (negBalance < 46) {
if (MVDiff < 27) {
return 2;
}
else if (MVDiff >= 27) {
return 8;
}
}
else if (negBalance >= 46) {
if (ssDCval < 93) {
if (Rs < 3549) {
return 4;
}
else if (Rs >= 3549) {
if (Rs < 3754) {
return 2;
}
else if (Rs >= 3754) {
return 8;
}
}
}
else if (ssDCval >= 93) {
return 8;
}
}
}
else if (AFD >= 24) {
return 1;
}
}
}
else if (RsDiff >= 17) {
if (mcTcor < 95) {
if (MVDiff < 49) {
if (MVDiff < 18) {
if (mu_mv_mag_sq < 48) {
if (SC < 5248) {
if (Rs < 3747) {
return 2;
}
else if (Rs >= 3747) {
return 8;
}
}
else if (SC >= 5248) {
return 2;
}
}
else if (mu_mv_mag_sq >= 48) {
return 8;
}
}
else if (MVDiff >= 18) {
if (ssDCval < 103) {
if (mu_mv_mag_sq < 13) {
return 2;
}
else if (mu_mv_mag_sq >= 13) {
if (CsDiff < 10) {
if (negBalance < 49) {
if (Rs < 4484) {
return 2;
}
else if (Rs >= 4484) {
return 4;
}
}
else if (negBalance >= 49) {
return 4;
}
}
else if (CsDiff >= 10) {
return 2;
}
}
}
else if (ssDCval >= 103) {
if (Rs < 5645) {
return 2;
}
else if (Rs >= 5645) {
return 1;
}
}
}
}
else if (MVDiff >= 49) {
if (MVDiff < 67) {
if (AFD < 30) {
if (Rs < 7156) {
return 8;
}
else if (Rs >= 7156) {
return 2;
}
}
else if (AFD >= 30) {
return 2;
}
}
else if (MVDiff >= 67) {
if (Rs < 3952) {
return 8;
}
else if (Rs >= 3952) {
return 1;
}
}
}
}
else if (mcTcor >= 95) {
if (AFD < 31) {
return 8;
}
else if (AFD >= 31) {
return 1;
}
}
}
}
else if (refDCval >= 119) {
if (negBalance < 53) {
if (refDCval < 137) {
if (RsDiff < 24) {
if (SC < 5065) {
return 2;
}
else if (SC >= 5065) {
if (posBalance < 54) {
if (tcor < 82) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
else if (tcor >= 82) {
return 8;
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (RsDiff >= 24) {
if (Cs < 1796) {
if (Rs < 4425) {
return 1;
}
else if (Rs >= 4425) {
return 2;
}
}
else if (Cs >= 1796) {
return 4;
}
}
}
else if (refDCval >= 137) {
if (gchDC < 3) {
if (SC < 5019) {
return 2;
}
else if (SC >= 5019) {
if (tcor < 77) {
if (ssDCval < 138) {
return 8;
}
else if (ssDCval >= 138) {
return 4;
}
}
else if (tcor >= 77) {
if (Rs < 5941) {
return 2;
}
else if (Rs >= 5941) {
return 8;
}
}
}
}
else if (gchDC >= 3) {
if (Rs < 4711) {
if (Rs < 3670) {
return 4;
}
else if (Rs >= 3670) {
return 1;
}
}
else if (Rs >= 4711) {
if (Cs < 1676) {
return 8;
}
else if (Cs >= 1676) {
if (Rs < 5516) {
return 2;
}
else if (Rs >= 5516) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 53) {
if (CsDiff < 9) {
if (SC < 5448) {
return 2;
}
else if (SC >= 5448) {
if (TSC < 13) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
if (Rs < 6081) {
return 1;
}
else if (Rs >= 6081) {
return 8;
}
}
}
else if (TSC >= 13) {
return 2;
}
}
}
else if (CsDiff >= 9) {
if (negBalance < 57) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 4201) {
return 1;
}
else if (Rs >= 4201) {
return 2;
}
}
}
else if (negBalance >= 57) {
return 1;
}
}
}
}
}
}
}
}
else if (tcor >= 87) {
if (RsDiff < 9) {
if (Rs < 3623) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (Rs >= 3623) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (MVDiff < 15) {
if (refDCval < 101) {
if (refDCval < 73) {
if (diffRsCsdiff < 73) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (diffRsCsdiff >= 73) {
if (RsCsDiff < 242) {
return 4;
}
else if (RsCsDiff >= 242) {
if (Rs < 6937) {
return 8;
}
else if (Rs >= 6937) {
return 2;
}
}
}
}
else if (refDCval >= 73) {
if (tcor < 89) {
if (AFD < 11) {
if (Rs < 3927) {
return 4;
}
else if (Rs >= 3927) {
return 1;
}
}
else if (AFD >= 11) {
if (Cs < 1920) {
return 2;
}
else if (Cs >= 1920) {
return 8;
}
}
}
else if (tcor >= 89) {
if (posBalance < 39) {
if (Rs < 4362) {
return 2;
}
else if (Rs >= 4362) {
return 8;
}
}
else if (posBalance >= 39) {
return 2;
}
}
}
}
else if (refDCval >= 101) {
if (RsDiff < 11) {
if (MVDiff < 7) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (Rs < 3928) {
return 4;
}
else if (Rs >= 3928) {
return 1;
}
}
}
else if (MVDiff >= 7) {
if (Cs < 1903) {
return 8;
}
else if (Cs >= 1903) {
return 1;
}
}
}
else if (RsDiff >= 11) {
if (posBalance < 60) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (RsCsDiff < 268) {
if (Rs < 4492) {
return 2;
}
else if (Rs >= 4492) {
return 1;
}
}
else if (RsCsDiff >= 268) {
return 2;
}
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
}
else if (MVDiff >= 15) {
if (ssDCval < 110) {
if (Rs < 3833) {
if (Cs < 1611) {
if (Rs < 3350) {
return 8;
}
else if (Rs >= 3350) {
return 2;
}
}
else if (Cs >= 1611) {
return 8;
}
}
else if (Rs >= 3833) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (ssDCval >= 110) {
if (Rs < 4920) {
return 8;
}
else if (Rs >= 4920) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 1939) {
if (AFD < 12) {
if (refDCval < 100) {
if (Rs < 4163) {
if (ssDCval < 71) {
if (TSC < 9) {
if (MVDiff < 4) {
if (Cs < 2853) {
return 1;
}
else if (Cs >= 2853) {
if (Cs < 3509) {
return 4;
}
else if (Cs >= 3509) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (Rs < 1699) {
return 1;
}
else if (Rs >= 1699) {
if (mu_mv_mag_sq < 96) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (mu_mv_mag_sq >= 96) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
}
}
}
}
else if (TSC >= 9) {
if (Cs < 2736) {
if (Cs < 2216) {
if (Rs < 2751) {
return 4;
}
else if (Rs >= 2751) {
if (Rs < 3164) {
return 1;
}
else if (Rs >= 3164) {
return 8;
}
}
}
else if (Cs >= 2216) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
return 4;
}
}
}
else if (Cs >= 2736) {
return 4;
}
}
}
else if (ssDCval >= 71) {
if (Cs < 2639) {
if (RsDiff < 11) {
if (diffAFD < -1) {
if (ssDCval < 92) {
if (Cs < 2405) {
return 4;
}
else if (Cs >= 2405) {
return 8;
}
}
else if (ssDCval >= 92) {
if (Rs < 4068) {
return 8;
}
else if (Rs >= 4068) {
return 2;
}
}
}
else if (diffAFD >= -1) {
if (SC < 5613) {
return 8;
}
else if (SC >= 5613) {
if (CsDiff < 10) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
else if (CsDiff >= 10) {
return 4;
}
}
}
}
else if (RsDiff >= 11) {
if (Rs < 3797) {
if (Rs < 3760) {
return 8;
}
else if (Rs >= 3760) {
return 4;
}
}
else if (Rs >= 3797) {
return 2;
}
}
}
else if (Cs >= 2639) {
if (ssDCval < 81) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (ssDCval >= 81) {
if (Rs < 3284) {
if (TSC < 10) {
if (Cs < 3464) {
return 4;
}
else if (Cs >= 3464) {
return 2;
}
}
else if (TSC >= 10) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (Cs < 3701) {
return 4;
}
else if (Cs >= 3701) {
return 8;
}
}
}
}
else if (Rs >= 3284) {
return 4;
}
}
}
}
}
else if (Rs >= 4163) {
if (MVDiff < 8) {
if (ssDCval < 73) {
if (tcor < 90) {
if (Cs < 4198) {
return 4;
}
else if (Cs >= 4198) {
return 8;
}
}
else if (tcor >= 90) {
if (Rs < 5821) {
return 8;
}
else if (Rs >= 5821) {
return 2;
}
}
}
else if (ssDCval >= 73) {
if (RsDiff < 9) {
if (Rs < 4945) {
if (Rs < 4431) {
return 4;
}
else if (Rs >= 4431) {
return 8;
}
}
else if (Rs >= 4945) {
if (Rs < 5946) {
return 1;
}
else if (Rs >= 5946) {
return 2;
}
}
}
else if (RsDiff >= 9) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (mcTcor < 92) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (mcTcor >= 92) {
return 2;
}
}
}
}
}
else if (MVDiff >= 8) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (MVDiff < 13) {
if (tcor < 91) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 5084) {
return 8;
}
else if (Rs >= 5084) {
return 2;
}
}
}
else if (tcor >= 91) {
return 2;
}
}
else if (MVDiff >= 13) {
return 8;
}
}
}
}
}
else if (refDCval >= 100) {
if (negBalance < 46) {
if (posBalance < 42) {
if (ssDCval < 112) {
if (SC < 9612) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
return 2;
}
}
else if (SC >= 9612) {
return 1;
}
}
else if (ssDCval >= 112) {
if (ssDCval < 131) {
if (SC < 6342) {
if (Cs < 2597) {
return 8;
}
else if (Cs >= 2597) {
return 4;
}
}
else if (SC >= 6342) {
if (negBalance < 22) {
if (Rs < 3892) {
return 4;
}
else if (Rs >= 3892) {
return 8;
}
}
else if (negBalance >= 22) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
return 2;
}
}
}
}
else if (ssDCval >= 131) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
else if (posBalance >= 42) {
if (Rs < 6279) {
if (RsDiff < 8) {
if (Rs < 4418) {
return 8;
}
else if (Rs >= 4418) {
return 1;
}
}
else if (RsDiff >= 8) {
if (MVDiff < 38) {
if (MVDiff < 1) {
if (Cs < 4332) {
return 4;
}
else if (Cs >= 4332) {
return 8;
}
}
else if (MVDiff >= 1) {
if (negBalance < 31) {
if (SC < 6112) {
return 8;
}
else if (SC >= 6112) {
if (Rs < 3061) {
if (Rs < 2544) {
return 1;
}
else if (Rs >= 2544) {
return 4;
}
}
else if (Rs >= 3061) {
return 2;
}
}
}
else if (negBalance >= 31) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (CsDiff < 8) {
if (Rs < 3706) {
if (Rs < 3631) {
return 2;
}
else if (Rs >= 3631) {
return 4;
}
}
else if (Rs >= 3706) {
return 8;
}
}
else if (CsDiff >= 8) {
return 8;
}
}
}
}
}
else if (MVDiff >= 38) {
return 2;
}
}
}
else if (Rs >= 6279) {
if (Rs < 8524) {
if (Cs < 3009) {
return 2;
}
else if (Cs >= 3009) {
if (Rs < 7255) {
return 4;
}
else if (Rs >= 7255) {
return 8;
}
}
}
else if (Rs >= 8524) {
return 1;
}
}
}
}
else if (negBalance >= 46) {
if (TSC < 9) {
if (posBalance < 47) {
if (Rs < 7276) {
return 2;
}
else if (Rs >= 7276) {
return 1;
}
}
else if (posBalance >= 47) {
return 8;
}
}
else if (TSC >= 9) {
if (mu_mv_mag_sq < 83) {
if (refDCval < 102) {
if (Rs < 4249) {
return 2;
}
else if (Rs >= 4249) {
return 1;
}
}
else if (refDCval >= 102) {
if (RsDiff < 8) {
if (Cs < 2116) {
return 1;
}
else if (Cs >= 2116) {
if (Cs < 2591) {
return 8;
}
else if (Cs >= 2591) {
return 1;
}
}
}
else if (RsDiff >= 8) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (Cs < 3458) {
if (posBalance < 54) {
if (Cs < 3048) {
return 8;
}
else if (Cs >= 3048) {
return 4;
}
}
else if (posBalance >= 54) {
return 1;
}
}
else if (Cs >= 3458) {
return 8;
}
}
}
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 24) {
if (ssDCval < 154) {
if (AFD < 22) {
if (ssDCval < 95) {
if (TSC < 10) {
if (MVDiff < 6) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
if (Cs < 2912) {
if (AFD < 15) {
if (SC < 5242) {
if (Rs < 2597) {
return 8;
}
else if (Rs >= 2597) {
return 4;
}
}
else if (SC >= 5242) {
if (CsDiff < 12) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
if (Rs < 3125) {
return 2;
}
else if (Rs >= 3125) {
return 4;
}
}
}
else if (CsDiff >= 12) {
return 4;
}
}
}
else if (AFD >= 15) {
if (Rs < 4231) {
if (RsDiff < 11) {
if (RsDiff < 9) {
if (Rs < 2323) {
return 4;
}
else if (Rs >= 2323) {
return 8;
}
}
else if (RsDiff >= 9) {
if (Rs < 3094) {
return 2;
}
else if (Rs >= 3094) {
return 4;
}
}
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (Rs >= 4231) {
return 8;
}
}
}
else if (Cs >= 2912) {
if (mcTcor < 94) {
if (diffRsCsdiff < -18) {
if (diffRsCsdiff < -27) {
return 4;
}
else if (diffRsCsdiff >= -27) {
if (Rs < 3486) {
return 8;
}
else if (Rs >= 3486) {
return 4;
}
}
}
else if (diffRsCsdiff >= -18) {
if (Rs < 3887) {
return 2;
}
else if (Rs >= 3887) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 2618) {
return 4;
}
else if (Rs >= 2618) {
return 1;
}
}
}
}
}
else if (MVDiff >= 6) {
if (RsCsDiff < 189) {
if (ssDCval < 59) {
return 2;
}
else if (ssDCval >= 59) {
if (SC < 5130) {
if (diffRsCsdiff < 13) {
if (tcor < 85) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
if (Rs < 2377) {
return 8;
}
else if (Rs >= 2377) {
return 4;
}
}
}
}
else if (tcor >= 85) {
return 8;
}
}
else if (diffRsCsdiff >= 13) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
else if (SC >= 5130) {
return 8;
}
}
}
else if (RsCsDiff >= 189) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (Rs < 3857) {
if (posBalance < 58) {
if (negBalance < 45) {
if (Rs < 2800) {
return 8;
}
else if (Rs >= 2800) {
return 2;
}
}
else if (negBalance >= 45) {
return 4;
}
}
else if (posBalance >= 58) {
if (Rs < 1264) {
return 8;
}
else if (Rs >= 1264) {
return 2;
}
}
}
else if (Rs >= 3857) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 10) {
if (tcor < 64) {
if (AFD < 17) {
if (Rs < 2221) {
return 2;
}
else if (Rs >= 2221) {
return 8;
}
}
else if (AFD >= 17) {
if (gchDC < 4) {
if (AFD < 19) {
if (negBalance < 57) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
if (Cs < 3216) {
return 8;
}
else if (Cs >= 3216) {
return 4;
}
}
}
else if (negBalance >= 57) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
}
else if (AFD >= 19) {
return 4;
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (tcor >= 64) {
if (ssDCval < 83) {
if (ssDCval < 75) {
if (RsCsDiff < 195) {
if (AFD < 18) {
if (Cs < 2868) {
return 8;
}
else if (Cs >= 2868) {
return 1;
}
}
else if (AFD >= 18) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 8;
}
}
}
else if (RsCsDiff >= 195) {
if (ssDCval < 67) {
if (MVDiff < 16) {
if (CsDiff < 13) {
return 1;
}
else if (CsDiff >= 13) {
return 4;
}
}
else if (MVDiff >= 16) {
if (Rs < 1935) {
return 4;
}
else if (Rs >= 1935) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
}
}
else if (ssDCval >= 67) {
if (AFD < 17) {
if (Rs < 4733) {
return 4;
}
else if (Rs >= 4733) {
return 8;
}
}
else if (AFD >= 17) {
if (Cs < 3318) {
if (AFD < 20) {
if (Cs < 2273) {
if (Cs < 2135) {
return 8;
}
else if (Cs >= 2135) {
return 2;
}
}
else if (Cs >= 2273) {
if (ssDCval < 72) {
if (Rs < 3538) {
return 1;
}
else if (Rs >= 3538) {
return 2;
}
}
else if (ssDCval >= 72) {
return 8;
}
}
}
else if (AFD >= 20) {
return 2;
}
}
else if (Cs >= 3318) {
if (RsCsDiff < 546) {
if (Rs < 3792) {
return 4;
}
else if (Rs >= 3792) {
return 2;
}
}
else if (RsCsDiff >= 546) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (Rs < 3055) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 20) {
if (Cs < 1986) {
if (Rs < 2626) {
return 8;
}
else if (Rs >= 2626) {
return 4;
}
}
else if (Cs >= 1986) {
return 8;
}
}
else if (mu_mv_mag_sq >= 20) {
return 1;
}
}
}
else if (Rs >= 3055) {
if (Cs < 2284) {
return 4;
}
else if (Cs >= 2284) {
if (posBalance < 55) {
if (Rs < 3738) {
return 4;
}
else if (Rs >= 3738) {
return 2;
}
}
else if (posBalance >= 55) {
if (MVDiff < 8) {
if (AFD < 19) {
if (Rs < 3371) {
return 2;
}
else if (Rs >= 3371) {
return 1;
}
}
else if (AFD >= 19) {
return 8;
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 83) {
if (Cs < 3876) {
if (mcTcor < 90) {
if (AFD < 17) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (Cs < 2193) {
return 4;
}
else if (Cs >= 2193) {
return 8;
}
}
}
}
else if (AFD >= 17) {
if (tcor < 78) {
if (CsDiff < 12) {
return 4;
}
else if (CsDiff >= 12) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 4;
}
}
}
else if (tcor >= 78) {
if (Rs < 3094) {
return 1;
}
else if (Rs >= 3094) {
return 8;
}
}
}
}
else if (mcTcor >= 90) {
if (Rs < 3345) {
if (posBalance < 58) {
if (Cs < 2487) {
if (Rs < 2351) {
return 8;
}
else if (Rs >= 2351) {
if (CsDiff < 13) {
return 2;
}
else if (CsDiff >= 13) {
return 8;
}
}
}
else if (Cs >= 2487) {
return 4;
}
}
else if (posBalance >= 58) {
if (Rs < 1799) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
else if (Rs >= 1799) {
if (negBalance < 52) {
if (Cs < 2145) {
if (Rs < 2506) {
return 8;
}
else if (Rs >= 2506) {
return 2;
}
}
else if (Cs >= 2145) {
return 4;
}
}
else if (negBalance >= 52) {
return 8;
}
}
}
}
else if (Rs >= 3345) {
if (TSC < 12) {
if (Cs < 3396) {
if (Rs < 4034) {
return 4;
}
else if (Rs >= 4034) {
return 2;
}
}
else if (Cs >= 3396) {
return 2;
}
}
else if (TSC >= 12) {
if (Rs < 3790) {
return 8;
}
else if (Rs >= 3790) {
return 2;
}
}
}
}
}
else if (Cs >= 3876) {
if (Rs < 6146) {
return 8;
}
else if (Rs >= 6146) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 95) {
if (negBalance < 48) {
if (RsCsDiff < 201) {
if (negBalance < 41) {
if (Rs < 2154) {
return 1;
}
else if (Rs >= 2154) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
if (Rs < 3080) {
return 2;
}
else if (Rs >= 3080) {
return 1;
}
}
}
}
else if (negBalance >= 41) {
if (SC < 3969) {
if (Cs < 2166) {
return 8;
}
else if (Cs >= 2166) {
return 1;
}
}
else if (SC >= 3969) {
return 8;
}
}
}
else if (RsCsDiff >= 201) {
if (Rs < 7958) {
if (tcor < 70) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
if (mu_mv_mag_sq < 21) {
return 1;
}
else if (mu_mv_mag_sq >= 21) {
if (Rs < 1761) {
return 8;
}
else if (Rs >= 1761) {
return 2;
}
}
}
}
else if (tcor >= 70) {
if (Cs < 2795) {
if (ssDCval < 117) {
if (mu_mv_mag_sq < 6) {
if (refDCval < 97) {
return 1;
}
else if (refDCval >= 97) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
if (Rs < 2179) {
return 1;
}
else if (Rs >= 2179) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 6) {
if (Rs < 2836) {
return 2;
}
else if (Rs >= 2836) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (TSC < 10) {
if (Rs < 6453) {
return 4;
}
else if (Rs >= 6453) {
return 2;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (Cs >= 2795) {
if (RsCsDiff < 454) {
if (TSC < 10) {
if (mcTcor < 83) {
return 8;
}
else if (mcTcor >= 83) {
return 4;
}
}
else if (TSC >= 10) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
if (Cs < 3042) {
return 1;
}
else if (Cs >= 3042) {
return 4;
}
}
}
}
else if (RsCsDiff >= 454) {
return 2;
}
}
}
}
else if (Rs >= 7958) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (AFD < 20) {
if (diffRsCsdiff < 136) {
if (Rs < 3820) {
if (SC < 5557) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
if (refDCval < 102) {
if (AFD < 15) {
if (Rs < 2082) {
return 8;
}
else if (Rs >= 2082) {
return 2;
}
}
else if (AFD >= 15) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (refDCval >= 102) {
if (AFD < 19) {
if (MVDiff < 5) {
if (Rs < 2703) {
return 8;
}
else if (Rs >= 2703) {
return 4;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (AFD >= 19) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
}
else if (SC >= 5557) {
if (ssDCval < 108) {
if (Cs < 2193) {
return 8;
}
else if (Cs >= 2193) {
return 4;
}
}
else if (ssDCval >= 108) {
return 8;
}
}
}
else if (Rs >= 3820) {
if (mcTcor < 96) {
if (RsDiff < 16) {
if (diffAFD < -3) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (diffAFD >= -3) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 4955) {
return 8;
}
else if (Rs >= 4955) {
return 4;
}
}
}
}
else if (RsDiff >= 16) {
if (negBalance < 55) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 2;
}
}
else if (negBalance >= 55) {
return 8;
}
}
}
else if (mcTcor >= 96) {
return 4;
}
}
}
else if (diffRsCsdiff >= 136) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
if (Rs < 4651) {
return 8;
}
else if (Rs >= 4651) {
if (Rs < 5912) {
return 2;
}
else if (Rs >= 5912) {
return 4;
}
}
}
}
}
else if (AFD >= 20) {
if (tcor < 84) {
if (Cs < 2443) {
if (CsDiff < 12) {
if (tcor < 75) {
return 1;
}
else if (tcor >= 75) {
if (Rs < 2381) {
return 1;
}
else if (Rs >= 2381) {
return 8;
}
}
}
else if (CsDiff >= 12) {
return 8;
}
}
else if (Cs >= 2443) {
if (negBalance < 51) {
if (Rs < 4977) {
return 4;
}
else if (Rs >= 4977) {
return 2;
}
}
else if (negBalance >= 51) {
if (ssDCval < 109) {
if (MVDiff < 7) {
if (refDCval < 97) {
return 8;
}
else if (refDCval >= 97) {
return 4;
}
}
else if (MVDiff >= 7) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 4079) {
return 4;
}
else if (Rs >= 4079) {
return 8;
}
}
}
}
else if (ssDCval >= 109) {
if (Cs < 2450) {
return 4;
}
else if (Cs >= 2450) {
if (diffAFD < -1) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 2;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
}
else if (tcor >= 84) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
return 2;
}
}
}
}
}
}
else if (AFD >= 22) {
if (Rs < 5914) {
if (posBalance < 55) {
if (diffRsCsdiff < 1) {
if (ssDCval < 107) {
return 2;
}
else if (ssDCval >= 107) {
if (refDCval < 112) {
return 1;
}
else if (refDCval >= 112) {
return 2;
}
}
}
else if (diffRsCsdiff >= 1) {
if (RsCsDiff < 335) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (tcor < 69) {
return 8;
}
else if (tcor >= 69) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 1;
}
}
}
}
else if (RsCsDiff >= 335) {
if (Cs < 2040) {
return 1;
}
else if (Cs >= 2040) {
if (tcor < 57) {
if (AFD < 31) {
return 4;
}
else if (AFD >= 31) {
return 2;
}
}
else if (tcor >= 57) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (ssDCval < 144) {
return 2;
}
else if (ssDCval >= 144) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 55) {
if (negBalance < 54) {
if (Rs < 973) {
return 2;
}
else if (Rs >= 973) {
if (RsCsDiff < 263) {
if (Rs < 1672) {
if (Rs < 1279) {
return 1;
}
else if (Rs >= 1279) {
return 4;
}
}
else if (Rs >= 1672) {
if (diffRsCsdiff < -18) {
if (Rs < 2151) {
return 8;
}
else if (Rs >= 2151) {
if (Rs < 3421) {
return 4;
}
else if (Rs >= 3421) {
return 8;
}
}
}
else if (diffRsCsdiff >= -18) {
return 8;
}
}
}
else if (RsCsDiff >= 263) {
if (gchDC < 5) {
if (CsDiff < 19) {
if (RsCsDiff < 353) {
if (AFD < 24) {
if (diffAFD < 1) {
if (Cs < 2210) {
return 2;
}
else if (Cs >= 2210) {
return 1;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (AFD >= 24) {
return 2;
}
}
else if (RsCsDiff >= 353) {
if (Cs < 2220) {
if (Rs < 2854) {
return 4;
}
else if (Rs >= 2854) {
return 2;
}
}
else if (Cs >= 2220) {
if (diffTSC < 5) {
if (diffRsCsdiff < -222) {
return 4;
}
else if (diffRsCsdiff >= -222) {
return 2;
}
}
else if (diffTSC >= 5) {
return 4;
}
}
}
}
else if (CsDiff >= 19) {
if (Rs < 1759) {
return 8;
}
else if (Rs >= 1759) {
if (MVDiff < 26) {
return 4;
}
else if (MVDiff >= 26) {
return 2;
}
}
}
}
else if (gchDC >= 5) {
if (TSC < 14) {
if (SC < 4361) {
return 4;
}
else if (SC >= 4361) {
return 2;
}
}
else if (TSC >= 14) {
if (mu_mv_mag_sq < 20) {
if (Rs < 5267) {
return 8;
}
else if (Rs >= 5267) {
return 1;
}
}
else if (mu_mv_mag_sq >= 20) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 54) {
if (ssDCval < 109) {
if (Cs < 3053) {
if (TSC < 16) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (mcTcor < 95) {
if (refDCval < 77) {
if (tcor < 57) {
if (Rs < 1643) {
return 4;
}
else if (Rs >= 1643) {
return 1;
}
}
else if (tcor >= 57) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 2116) {
if (Rs < 2173) {
return 4;
}
else if (Rs >= 2173) {
return 8;
}
}
else if (Cs >= 2116) {
return 4;
}
}
}
}
else if (refDCval >= 77) {
if (negBalance < 58) {
if (refDCval < 85) {
if (diffTSC < -1) {
if (Rs < 1354) {
return 8;
}
else if (Rs >= 1354) {
return 1;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (refDCval >= 85) {
if (mcTcor < 89) {
if (Cs < 2567) {
return 4;
}
else if (Cs >= 2567) {
if (Cs < 2761) {
return 1;
}
else if (Cs >= 2761) {
return 4;
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1985) {
return 8;
}
else if (Rs >= 1985) {
return 1;
}
}
}
}
else if (negBalance >= 58) {
if (tcor < 63) {
if (AFD < 35) {
return 1;
}
else if (AFD >= 35) {
return 8;
}
}
else if (tcor >= 63) {
if (Rs < 1827) {
return 8;
}
else if (Rs >= 1827) {
if (TSC < 14) {
if (Rs < 1923) {
return 1;
}
else if (Rs >= 1923) {
return 8;
}
}
else if (TSC >= 14) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 95) {
return 2;
}
}
}
else if (TSC >= 16) {
if (Cs < 2636) {
if (Rs < 4554) {
return 2;
}
else if (Rs >= 4554) {
return 1;
}
}
else if (Cs >= 2636) {
return 1;
}
}
}
else if (Cs >= 3053) {
if (refDCval < 91) {
if (Rs < 1655) {
return 4;
}
else if (Rs >= 1655) {
return 8;
}
}
else if (refDCval >= 91) {
if (RsDiff < 21) {
return 4;
}
else if (RsDiff >= 21) {
return 2;
}
}
}
}
else if (ssDCval >= 109) {
if (RsCsDiff < 233) {
if (Rs < 2674) {
if (MVDiff < 46) {
return 2;
}
else if (MVDiff >= 46) {
return 8;
}
}
else if (Rs >= 2674) {
return 1;
}
}
else if (RsCsDiff >= 233) {
if (TSC < 16) {
if (Cs < 2130) {
return 8;
}
else if (Cs >= 2130) {
if (SC < 5250) {
if (mu_mv_mag_sq < 8) {
return 2;
}
else if (mu_mv_mag_sq >= 8) {
return 1;
}
}
else if (SC >= 5250) {
if (RsDiff < 16) {
if (tcor < 77) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 1;
}
}
else if (tcor >= 77) {
return 8;
}
}
else if (RsDiff >= 16) {
if (Cs < 3354) {
return 8;
}
else if (Cs >= 3354) {
return 4;
}
}
}
}
}
else if (TSC >= 16) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Cs < 3061) {
return 8;
}
else if (Cs >= 3061) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 5914) {
if (negBalance < 51) {
if (Rs < 6204) {
return 8;
}
else if (Rs >= 6204) {
if (AFD < 33) {
return 1;
}
else if (AFD >= 33) {
return 2;
}
}
}
else if (negBalance >= 51) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (AFD < 41) {
if (AFD < 24) {
if (CsDiff < 18) {
return 8;
}
else if (CsDiff >= 18) {
return 4;
}
}
else if (AFD >= 24) {
return 4;
}
}
else if (AFD >= 41) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 154) {
if (ssDCval < 168) {
if (Rs < 3335) {
return 8;
}
else if (Rs >= 3335) {
if (posBalance < 55) {
if (ssDCval < 156) {
return 8;
}
else if (ssDCval >= 156) {
return 4;
}
}
else if (posBalance >= 55) {
return 4;
}
}
}
else if (ssDCval >= 168) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (negBalance < 49) {
if (refDCval < 93) {
if (posBalance < 49) {
if (refDCval < 78) {
if (Rs < 4500) {
if (mu_mv_mag_sq < 57) {
if (ssDCval < 65) {
if (Rs < 2108) {
return 4;
}
else if (Rs >= 2108) {
return 1;
}
}
else if (ssDCval >= 65) {
if (Rs < 2304) {
return 8;
}
else if (Rs >= 2304) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (negBalance < 41) {
if (Rs < 4398) {
return 4;
}
else if (Rs >= 4398) {
return 8;
}
}
else if (negBalance >= 41) {
return 2;
}
}
}
else if (Rs >= 4500) {
if (gchDC < 4) {
if (TSC < 17) {
if (Cs < 4029) {
return 4;
}
else if (Cs >= 4029) {
if (tcor < 81) {
return 4;
}
else if (tcor >= 81) {
return 8;
}
}
}
else if (TSC >= 17) {
return 2;
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (refDCval >= 78) {
if (TSC < 11) {
if (diffTSC < -1) {
if (Rs < 6623) {
return 2;
}
else if (Rs >= 6623) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (TSC >= 11) {
if (SC < 8171) {
return 2;
}
else if (SC >= 8171) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 1;
}
}
}
}
}
else if (posBalance >= 49) {
if (ssDCval < 56) {
if (mcTcor < 74) {
return 4;
}
else if (mcTcor >= 74) {
return 2;
}
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 53) {
if (TSC < 10) {
if (Rs < 1512) {
return 4;
}
else if (Rs >= 1512) {
return 8;
}
}
else if (TSC >= 10) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (diffMVdiffVal < -5) {
if (Cs < 1979) {
return 8;
}
else if (Cs >= 1979) {
return 2;
}
}
else if (diffMVdiffVal >= -5) {
if (tcor < 86) {
if (CsDiff < 13) {
return 8;
}
else if (CsDiff >= 13) {
if (Cs < 2834) {
if (Rs < 4463) {
return 4;
}
else if (Rs >= 4463) {
return 8;
}
}
else if (Cs >= 2834) {
return 8;
}
}
}
else if (tcor >= 86) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 3812) {
if (mcTcor < 93) {
if (tcor < 67) {
if (Rs < 3161) {
return 1;
}
else if (Rs >= 3161) {
return 8;
}
}
else if (tcor >= 67) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (refDCval < 66) {
return 8;
}
else if (refDCval >= 66) {
if (SC < 5602) {
if (tcor < 85) {
if (SC < 5349) {
if (Rs < 2130) {
return 8;
}
else if (Rs >= 2130) {
return 4;
}
}
else if (SC >= 5349) {
return 8;
}
}
else if (tcor >= 85) {
return 8;
}
}
else if (SC >= 5602) {
if (Rs < 3656) {
return 4;
}
else if (Rs >= 3656) {
if (Rs < 3710) {
if (Cs < 2530) {
return 8;
}
else if (Cs >= 2530) {
return 4;
}
}
else if (Rs >= 3710) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (Cs < 2187) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 1;
}
}
else if (Cs >= 2187) {
return 1;
}
}
}
else if (Rs >= 3812) {
if (Cs < 4488) {
if (gchDC < 2) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (SC < 6913) {
return 4;
}
else if (SC >= 6913) {
if (RsDiff < 17) {
return 2;
}
else if (RsDiff >= 17) {
if (Rs < 4887) {
return 4;
}
else if (Rs >= 4887) {
return 8;
}
}
}
}
}
else if (gchDC >= 2) {
if (AFD < 20) {
if (diffAFD < -2) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 4;
}
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 72) {
return 8;
}
else if (mu_mv_mag_sq >= 72) {
return 2;
}
}
}
else if (AFD >= 20) {
return 4;
}
}
}
else if (Cs >= 4488) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 93) {
if (MVDiff < 10) {
if (ssDCval < 102) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (RsCsDiff < 157) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (RsCsDiff >= 157) {
if (Cs < 3002) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
else if (Cs >= 3002) {
if (Rs < 4725) {
return 4;
}
else if (Rs >= 4725) {
return 2;
}
}
}
}
}
else if (ssDCval >= 102) {
if (AFD < 15) {
if (mcTcor < 95) {
if (diffAFD < -3) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
return 8;
}
}
else if (diffAFD >= -3) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
if (TSC < 12) {
if (diffRsCsdiff < -82) {
if (Rs < 7100) {
return 4;
}
else if (Rs >= 7100) {
return 8;
}
}
else if (diffRsCsdiff >= -82) {
return 8;
}
}
else if (TSC >= 12) {
return 4;
}
}
}
}
else if (mcTcor >= 95) {
if (SC < 9301) {
if (SC < 8138) {
return 8;
}
else if (SC >= 8138) {
return 2;
}
}
else if (SC >= 9301) {
return 4;
}
}
}
else if (AFD >= 15) {
if (refDCval < 126) {
if (Rs < 5334) {
return 8;
}
else if (Rs >= 5334) {
return 4;
}
}
else if (refDCval >= 126) {
if (ssDCval < 159) {
return 4;
}
else if (ssDCval >= 159) {
return 2;
}
}
}
}
}
else if (MVDiff >= 10) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (RsCsDiff < 397) {
if (refDCval < 116) {
if (Cs < 3574) {
if (Rs < 3222) {
if (Rs < 2952) {
return 8;
}
else if (Rs >= 2952) {
if (Cs < 2776) {
return 2;
}
else if (Cs >= 2776) {
return 8;
}
}
}
else if (Rs >= 3222) {
return 8;
}
}
else if (Cs >= 3574) {
if (mcTcor < 81) {
return 8;
}
else if (mcTcor >= 81) {
if (AFD < 16) {
if (Rs < 3543) {
return 4;
}
else if (Rs >= 3543) {
return 8;
}
}
else if (AFD >= 16) {
if (Rs < 5876) {
return 1;
}
else if (Rs >= 5876) {
return 8;
}
}
}
}
}
else if (refDCval >= 116) {
if (SC < 7437) {
if (mu_mv_mag_sq < 34) {
if (RsCsDiff < 186) {
return 1;
}
else if (RsCsDiff >= 186) {
if (Rs < 2455) {
return 8;
}
else if (Rs >= 2455) {
if (Rs < 2905) {
return 2;
}
else if (Rs >= 2905) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (mcTcor < 88) {
if (diffAFD < 5) {
if (SC < 7304) {
return 8;
}
else if (SC >= 7304) {
if (Rs < 3986) {
return 8;
}
else if (Rs >= 3986) {
return 1;
}
}
}
else if (diffAFD >= 5) {
return 1;
}
}
else if (mcTcor >= 88) {
if (posBalance < 53) {
if (tcor < 89) {
if (Cs < 2097) {
if (Rs < 2257) {
return 8;
}
else if (Rs >= 2257) {
return 2;
}
}
else if (Cs >= 2097) {
return 4;
}
}
else if (tcor >= 89) {
return 8;
}
}
else if (posBalance >= 53) {
if (Cs < 2522) {
return 8;
}
else if (Cs >= 2522) {
if (Rs < 3369) {
return 4;
}
else if (Rs >= 3369) {
return 8;
}
}
}
}
}
}
else if (SC >= 7437) {
if (tcor < 85) {
if (Rs < 3896) {
return 8;
}
else if (Rs >= 3896) {
return 1;
}
}
else if (tcor >= 85) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (Rs < 5104) {
return 2;
}
else if (Rs >= 5104) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 397) {
if (CsDiff < 12) {
return 4;
}
else if (CsDiff >= 12) {
if (MVDiff < 26) {
if (refDCval < 121) {
if (Rs < 5622) {
if (MVDiff < 22) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 4;
}
}
else if (MVDiff >= 22) {
return 1;
}
}
else if (Rs >= 5622) {
return 1;
}
}
else if (refDCval >= 121) {
if (RsCsDiff < 467) {
return 1;
}
else if (RsCsDiff >= 467) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
}
}
else if (MVDiff >= 26) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (Rs < 3387) {
return 2;
}
else if (Rs >= 3387) {
return 1;
}
}
}
}
}
}
}
}
}
else if (negBalance >= 49) {
if (Cs < 3308) {
if (mu_mv_mag_sq < 60) {
if (tcor < 70) {
if (refDCval < 64) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
if (TSC < 15) {
return 2;
}
else if (TSC >= 15) {
return 4;
}
}
}
else if (refDCval >= 64) {
if (MVDiff < 52) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (negBalance < 54) {
if (mu_mv_mag_sq < 25) {
return 4;
}
else if (mu_mv_mag_sq >= 25) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
if (SC < 7742) {
return 8;
}
else if (SC >= 7742) {
return 4;
}
}
}
}
else if (negBalance >= 54) {
if (ssDCval < 108) {
if (MVDiff < 28) {
return 4;
}
else if (MVDiff >= 28) {
if (TSC < 14) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (TSC >= 14) {
return 1;
}
}
}
else if (ssDCval >= 108) {
if (Cs < 2701) {
return 8;
}
else if (Cs >= 2701) {
if (Rs < 5216) {
return 2;
}
else if (Rs >= 5216) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 52) {
if (RsDiff < 31) {
if (mu_mv_mag_sq < 27) {
if (Rs < 2795) {
if (Rs < 1830) {
return 4;
}
else if (Rs >= 1830) {
return 8;
}
}
else if (Rs >= 2795) {
return 1;
}
}
else if (mu_mv_mag_sq >= 27) {
if (tcor < 66) {
return 8;
}
else if (tcor >= 66) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 2;
}
}
}
}
else if (RsDiff >= 31) {
return 1;
}
}
}
}
else if (tcor >= 70) {
if (mcTcor < 93) {
if (RsDiff < 17) {
if (TSC < 16) {
if (AFD < 25) {
if (SC < 5129) {
if (mu_mv_mag_sq < 26) {
if (Rs < 2172) {
return 1;
}
else if (Rs >= 2172) {
return 8;
}
}
else if (mu_mv_mag_sq >= 26) {
return 8;
}
}
else if (SC >= 5129) {
if (MVDiff < 5) {
if (CsDiff < 11) {
if (Cs < 2526) {
if (Rs < 3524) {
return 1;
}
else if (Rs >= 3524) {
return 8;
}
}
else if (Cs >= 2526) {
return 4;
}
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (MVDiff >= 5) {
if (AFD < 14) {
if (Cs < 2833) {
return 8;
}
else if (Cs >= 2833) {
return 1;
}
}
else if (AFD >= 14) {
if (mu_mv_mag_sq < 58) {
if (diffTSC < -2) {
if (SC < 5785) {
return 4;
}
else if (SC >= 5785) {
return 8;
}
}
else if (diffTSC >= -2) {
if (AFD < 20) {
if (refDCval < 54) {
return 4;
}
else if (refDCval >= 54) {
return 8;
}
}
else if (AFD >= 20) {
if (Cs < 2365) {
return 8;
}
else if (Cs >= 2365) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 4;
}
}
}
}
}
}
else if (AFD >= 25) {
return 4;
}
}
else if (TSC >= 16) {
return 4;
}
}
else if (RsDiff >= 17) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
if (Cs < 2691) {
if (mu_mv_mag_sq < 50) {
return 8;
}
else if (mu_mv_mag_sq >= 50) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
return 8;
}
}
}
}
else if (Cs >= 2691) {
if (Cs < 2808) {
return 4;
}
else if (Cs >= 2808) {
if (Rs < 4743) {
return 8;
}
else if (Rs >= 4743) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (TSC < 11) {
if (diffTSC < -1) {
if (Rs < 4572) {
return 2;
}
else if (Rs >= 4572) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (TSC >= 11) {
if (negBalance < 56) {
if (Rs < 4733) {
return 2;
}
else if (Rs >= 4733) {
return 8;
}
}
else if (negBalance >= 56) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 60) {
if (ssDCval < 92) {
if (SC < 6513) {
if (ssDCval < 90) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
if (ssDCval < 70) {
if (gchDC < 7) {
return 8;
}
else if (gchDC >= 7) {
return 4;
}
}
else if (ssDCval >= 70) {
if (RsCsDiff < 137) {
return 8;
}
else if (RsCsDiff >= 137) {
if (tcor < 70) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (tcor >= 70) {
if (Cs < 2720) {
if (MVDiff < 26) {
if (Rs < 2812) {
if (Rs < 2476) {
return 4;
}
else if (Rs >= 2476) {
return 8;
}
}
else if (Rs >= 2812) {
return 4;
}
}
else if (MVDiff >= 26) {
return 8;
}
}
else if (Cs >= 2720) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 90) {
if (Rs < 2175) {
return 8;
}
else if (Rs >= 2175) {
return 4;
}
}
}
else if (SC >= 6513) {
if (posBalance < 57) {
if (Rs < 4174) {
return 1;
}
else if (Rs >= 4174) {
if (Rs < 5129) {
return 8;
}
else if (Rs >= 5129) {
return 4;
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (ssDCval >= 92) {
if (Cs < 2660) {
if (AFD < 14) {
if (CsDiff < 9) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
if (mu_mv_mag_sq < 95) {
return 8;
}
else if (mu_mv_mag_sq >= 95) {
if (Rs < 3816) {
return 8;
}
else if (Rs >= 3816) {
return 1;
}
}
}
}
else if (CsDiff >= 9) {
return 1;
}
}
else if (AFD >= 14) {
if (diffAFD < 2) {
if (RsCsDiff < 194) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 4;
}
}
}
else if (RsCsDiff >= 194) {
return 4;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (Cs >= 2660) {
if (SC < 6772) {
return 4;
}
else if (SC >= 6772) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
}
}
}
}
else if (Cs >= 3308) {
if (refDCval < 115) {
if (negBalance < 64) {
if (Rs < 3162) {
if (mcTcor < 81) {
return 4;
}
else if (mcTcor >= 81) {
if (posBalance < 55) {
if (Rs < 1538) {
return 2;
}
else if (Rs >= 1538) {
return 4;
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (Rs >= 3162) {
if (refDCval < 55) {
if (Rs < 4970) {
return 4;
}
else if (Rs >= 4970) {
return 8;
}
}
else if (refDCval >= 55) {
if (mcTcor < 90) {
if (diffRsCsdiff < 229) {
if (Rs < 6468) {
if (diffMVdiffVal < 11) {
if (negBalance < 50) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 4;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (diffMVdiffVal >= 11) {
if (Rs < 3706) {
return 8;
}
else if (Rs >= 3706) {
return 1;
}
}
}
else if (Rs >= 6468) {
if (ssDCval < 104) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (ssDCval >= 104) {
return 8;
}
}
}
else if (diffRsCsdiff >= 229) {
return 4;
}
}
else if (mcTcor >= 90) {
if (Rs < 4078) {
if (Rs < 3526) {
return 8;
}
else if (Rs >= 3526) {
return 2;
}
}
else if (Rs >= 4078) {
if (ssDCval < 105) {
return 8;
}
else if (ssDCval >= 105) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 64) {
return 4;
}
}
else if (refDCval >= 115) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
if (CsDiff < 15) {
return 8;
}
else if (CsDiff >= 15) {
if (Rs < 5352) {
if (Rs < 4563) {
return 1;
}
else if (Rs >= 4563) {
return 2;
}
}
else if (Rs >= 5352) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Cs < 1650) {
if (Rs < 3094) {
if (Rs < 1290) {
if (Rs < 1127) {
if (AFD < 1) {
if (Rs < 599) {
return 4;
}
else if (Rs >= 599) {
return 2;
}
}
else if (AFD >= 1) {
if (refDCval < 63) {
if (SC < 2394) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (SC >= 2394) {
if (TSC < 2) {
if (Rs < 1051) {
return 8;
}
else if (Rs >= 1051) {
return 1;
}
}
else if (TSC >= 2) {
if (mcTcor < 96) {
if (Rs < 1073) {
if (Rs < 968) {
return 8;
}
else if (Rs >= 968) {
return 4;
}
}
else if (Rs >= 1073) {
return 8;
}
}
else if (mcTcor >= 96) {
return 8;
}
}
}
}
else if (refDCval >= 63) {
if (refDCval < 76) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (Rs < 1030) {
return 2;
}
else if (Rs >= 1030) {
return 4;
}
}
}
else if (refDCval >= 76) {
if (ssDCval < 114) {
if (MVDiff < 2) {
if (Rs < 953) {
return 1;
}
else if (Rs >= 953) {
return 8;
}
}
else if (MVDiff >= 2) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (ssDCval >= 114) {
if (Rs < 1015) {
if (tcor < 95) {
if (mcTcor < 94) {
if (Rs < 597) {
return 2;
}
else if (Rs >= 597) {
return 8;
}
}
else if (mcTcor >= 94) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (tcor >= 95) {
return 8;
}
}
else if (Rs >= 1015) {
if (Cs < 1582) {
return 4;
}
else if (Cs >= 1582) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1127) {
if (posBalance < 41) {
if (posBalance < 10) {
return 8;
}
else if (posBalance >= 10) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (posBalance >= 41) {
if (mu_mv_mag_sq < 120) {
if (MVDiff < 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Rs < 1257) {
return 4;
}
else if (Rs >= 1257) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1290) {
if (ssDCval < 86) {
if (mcTcor < 98) {
if (Rs < 2681) {
if (Rs < 1402) {
if (tcor < 96) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (tcor >= 96) {
if (Cs < 1330) {
return 1;
}
else if (Cs >= 1330) {
return 4;
}
}
}
else if (Rs >= 1402) {
if (negBalance < 21) {
if (negBalance < 17) {
if (MVDiff < 42) {
return 8;
}
else if (MVDiff >= 42) {
return 1;
}
}
else if (negBalance >= 17) {
if (Rs < 1565) {
return 8;
}
else if (Rs >= 1565) {
if (mu_mv_mag_sq < 127) {
return 4;
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
}
}
else if (negBalance >= 21) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
return 4;
}
}
}
}
else if (Rs >= 2681) {
if (Rs < 2987) {
if (Rs < 2799) {
return 4;
}
else if (Rs >= 2799) {
if (Rs < 2839) {
return 4;
}
else if (Rs >= 2839) {
return 8;
}
}
}
else if (Rs >= 2987) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
return 2;
}
}
}
}
else if (mcTcor >= 98) {
if (posBalance < 41) {
if (SC < 4493) {
if (Rs < 1465) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Rs < 1408) {
return 8;
}
else if (Rs >= 1408) {
return 4;
}
}
}
}
else if (Rs >= 1465) {
if (AFD < 5) {
if (Cs < 1350) {
if (Rs < 2339) {
if (Cs < 1329) {
if (Cs < 1327) {
return 8;
}
else if (Cs >= 1327) {
if (Rs < 2235) {
return 8;
}
else if (Rs >= 2235) {
return 2;
}
}
}
else if (Cs >= 1329) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
return 2;
}
}
}
}
else if (Rs >= 2339) {
if (posBalance < 35) {
return 2;
}
else if (posBalance >= 35) {
return 8;
}
}
}
else if (Cs >= 1350) {
if (SC < 3284) {
if (Cs < 1441) {
return 8;
}
else if (Cs >= 1441) {
if (diffRsCsdiff < -1) {
return 8;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
}
else if (SC >= 3284) {
return 8;
}
}
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (SC >= 4493) {
return 1;
}
}
else if (posBalance >= 41) {
if (ssDCval < 78) {
if (posBalance < 44) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
return 4;
}
}
else if (posBalance >= 44) {
if (posBalance < 60) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
return 4;
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
else if (ssDCval >= 78) {
if (Rs < 2483) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
else if (Rs >= 2483) {
return 4;
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (RsCsDiff < 7) {
if (MVDiff < 122) {
if (Rs < 1362) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 1362) {
if (Rs < 2988) {
return 8;
}
else if (Rs >= 2988) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
return 2;
}
}
}
}
else if (MVDiff >= 122) {
return 4;
}
}
else if (RsCsDiff >= 7) {
if (refDCval < 117) {
if (TSC < 11) {
if (ssDCval < 94) {
if (mu_mv_mag_sq < 114) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (TSC >= 11) {
return 1;
}
}
else if (refDCval >= 117) {
if (mu_mv_mag_sq < 125) {
if (negBalance < 5) {
return 1;
}
else if (negBalance >= 5) {
if (Rs < 2889) {
return 8;
}
else if (Rs >= 2889) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (SC < 3469) {
return 1;
}
else if (SC >= 3469) {
return 8;
}
}
}
}
}
else if (ssDCval >= 140) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
if (diffAFD < -1) {
if (Rs < 2471) {
return 8;
}
else if (Rs >= 2471) {
return 2;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3094) {
if (ssDCval < 114) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 126) {
if (SC < 4912) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
else if (SC >= 4912) {
if (AFD < 3) {
if (Cs < 1553) {
if (posBalance < 33) {
if (ssDCval < 58) {
if (Rs < 4134) {
return 4;
}
else if (Rs >= 4134) {
return 8;
}
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (posBalance >= 33) {
if (mu_mv_mag_sq < 119) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
}
else if (Cs >= 1553) {
return 4;
}
}
else if (AFD >= 3) {
if (RsDiff < 6) {
if (Rs < 4325) {
if (Rs < 3543) {
return 8;
}
else if (Rs >= 3543) {
return 4;
}
}
else if (Rs >= 4325) {
return 8;
}
}
else if (RsDiff >= 6) {
if (Rs < 4186) {
return 2;
}
else if (Rs >= 4186) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1578) {
if (Rs < 4073) {
if (mu_mv_mag_sq < 128) {
if (negBalance < 16) {
if (Rs < 3532) {
return 1;
}
else if (Rs >= 3532) {
return 4;
}
}
else if (negBalance >= 16) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (Rs >= 4073) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (Cs >= 1578) {
if (Rs < 3333) {
return 8;
}
else if (Rs >= 3333) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (posBalance < 45) {
if (negBalance < 28) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (Cs < 1559) {
return 4;
}
else if (Cs >= 1559) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
if (Rs < 3326) {
return 8;
}
else if (Rs >= 3326) {
return 4;
}
}
}
}
}
else if (negBalance >= 28) {
if (ssDCval < 97) {
if (Rs < 3838) {
if (Cs < 1626) {
return 2;
}
else if (Cs >= 1626) {
return 4;
}
}
else if (Rs >= 3838) {
return 8;
}
}
else if (ssDCval >= 97) {
return 4;
}
}
}
else if (posBalance >= 45) {
if (mcTcor < 86) {
if (Rs < 4805) {
return 4;
}
else if (Rs >= 4805) {
return 8;
}
}
else if (mcTcor >= 86) {
if (Rs < 3115) {
return 4;
}
else if (Rs >= 3115) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (Rs < 3947) {
return 2;
}
else if (Rs >= 3947) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 114) {
if (tcor < 95) {
if (Rs < 3256) {
return 1;
}
else if (Rs >= 3256) {
return 4;
}
}
else if (tcor >= 95) {
return 2;
}
}
}
}
else if (Cs >= 1650) {
if (ssDCval < 102) {
if (SC < 4286) {
if (refDCval < 67) {
if (MVDiff < 2) {
if (RsCsDiff < 2) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Cs < 1809) {
if (Rs < 1345) {
return 1;
}
else if (Rs >= 1345) {
return 8;
}
}
else if (Cs >= 1809) {
return 1;
}
}
}
else if (RsCsDiff >= 2) {
if (refDCval < 63) {
if (ssDCval < 53) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
if (Rs < 1719) {
return 4;
}
else if (Rs >= 1719) {
return 1;
}
}
}
else if (ssDCval >= 53) {
if (posBalance < 32) {
if (Rs < 2129) {
return 1;
}
else if (Rs >= 2129) {
return 8;
}
}
else if (posBalance >= 32) {
return 8;
}
}
}
else if (refDCval >= 63) {
if (AFD < 3) {
if (Rs < 1150) {
return 1;
}
else if (Rs >= 1150) {
if (refDCval < 66) {
if (SC < 4151) {
return 8;
}
else if (SC >= 4151) {
if (Rs < 2439) {
return 1;
}
else if (Rs >= 2439) {
return 8;
}
}
}
else if (refDCval >= 66) {
return 1;
}
}
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (refDCval < 54) {
if (Cs < 2145) {
return 8;
}
else if (Cs >= 2145) {
return 4;
}
}
else if (refDCval >= 54) {
if (TSC < 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (Rs < 1020) {
return 8;
}
else if (Rs >= 1020) {
return 1;
}
}
}
else if (TSC >= 3) {
if (MVDiff < 10) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
}
}
}
else if (refDCval >= 67) {
if (SC < 2832) {
if (Rs < 622) {
return 8;
}
else if (Rs >= 622) {
return 1;
}
}
else if (SC >= 2832) {
if (SC < 4241) {
if (Rs < 2478) {
if (negBalance < 67) {
return 8;
}
else if (negBalance >= 67) {
return 4;
}
}
else if (Rs >= 2478) {
if (SC < 4164) {
return 4;
}
else if (SC >= 4164) {
return 8;
}
}
}
else if (SC >= 4241) {
if (tcor < 98) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (tcor >= 98) {
if (Cs < 2002) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
return 4;
}
}
else if (Cs >= 2002) {
if (Cs < 2028) {
return 2;
}
else if (Cs >= 2028) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 4286) {
if (Cs < 2197) {
if (Rs < 3328) {
if (Cs < 1777) {
if (negBalance < 15) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (Rs < 2808) {
return 8;
}
else if (Rs >= 2808) {
return 1;
}
}
}
else if (negBalance >= 15) {
if (refDCval < 80) {
if (MVDiff < 1) {
if (mu_mv_mag_sq < 126) {
return 4;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 2938) {
return 8;
}
else if (Rs >= 2938) {
if (Cs < 1775) {
return 4;
}
else if (Cs >= 1775) {
return 4;
}
}
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (refDCval >= 80) {
if (RsCsDiff < 38) {
return 4;
}
else if (RsCsDiff >= 38) {
return 8;
}
}
}
}
else if (Cs >= 1777) {
if (ssDCval < 87) {
if (posBalance < 22) {
return 8;
}
else if (posBalance >= 22) {
if (Rs < 2616) {
if (tcor < 99) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (tcor >= 99) {
if (RsDiff < 2) {
if (Rs < 2582) {
return 1;
}
else if (Rs >= 2582) {
return 8;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (Rs >= 2616) {
if (mu_mv_mag_sq < 126) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (ssDCval < 69) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (SC < 4731) {
return 8;
}
else if (SC >= 4731) {
return 4;
}
}
}
else if (ssDCval >= 69) {
if (Cs < 1819) {
if (Cs < 1786) {
return 8;
}
else if (Cs >= 1786) {
return 4;
}
}
else if (Cs >= 1819) {
if (Rs < 2706) {
if (ssDCval < 83) {
if (Cs < 2054) {
return 4;
}
else if (Cs >= 2054) {
return 8;
}
}
else if (ssDCval >= 83) {
return 8;
}
}
else if (Rs >= 2706) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (tcor < 99) {
if (Rs < 2945) {
return 8;
}
else if (Rs >= 2945) {
if (Cs < 2128) {
if (diffRsCsdiff < -4) {
return 4;
}
else if (diffRsCsdiff >= -4) {
if (CsDiff < 3) {
if (posBalance < 42) {
if (Cs < 1885) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (negBalance < 22) {
return 4;
}
else if (negBalance >= 22) {
if (Cs < 1860) {
if (Rs < 2992) {
return 4;
}
else if (Rs >= 2992) {
return 4;
}
}
else if (Cs >= 1860) {
if (Rs < 3053) {
return 8;
}
else if (Rs >= 3053) {
return 4;
}
}
}
}
}
else if (Cs >= 1885) {
return 8;
}
}
else if (posBalance >= 42) {
return 4;
}
}
else if (CsDiff >= 3) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
return 4;
}
}
}
}
else if (Cs >= 2128) {
return 8;
}
}
}
else if (tcor >= 99) {
if (Cs < 1824) {
if (Rs < 2848) {
return 8;
}
else if (Rs >= 2848) {
if (Rs < 3108) {
return 4;
}
else if (Rs >= 3108) {
return 4;
}
}
}
else if (Cs >= 1824) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (Rs < 2373) {
if (AFD < 2) {
if (ssDCval < 92) {
return 1;
}
else if (ssDCval >= 92) {
if (mu_mv_mag_sq < 114) {
if (Rs < 2372) {
return 4;
}
else if (Rs >= 2372) {
return 2;
}
}
else if (mu_mv_mag_sq >= 114) {
if (mu_mv_mag_sq < 119) {
if (diffMVdiffVal < -2) {
if (Rs < 2326) {
return 2;
}
else if (Rs >= 2326) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 2309) {
return 2;
}
else if (Rs >= 2309) {
return 4;
}
}
}
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 2373) {
if (ssDCval < 89) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
if (Rs < 2974) {
return 8;
}
else if (Rs >= 2974) {
return 2;
}
}
}
else if (ssDCval >= 89) {
if (Rs < 2391) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
else if (Rs >= 2391) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3328) {
if (posBalance < 41) {
if (Rs < 3835) {
if (posBalance < 22) {
if (Rs < 3337) {
return 4;
}
else if (Rs >= 3337) {
return 8;
}
}
else if (posBalance >= 22) {
return 2;
}
}
else if (Rs >= 3835) {
if (CsDiff < 3) {
if (ssDCval < 67) {
if (Rs < 7357) {
if (Rs < 5065) {
return 8;
}
else if (Rs >= 5065) {
return 4;
}
}
else if (Rs >= 7357) {
return 2;
}
}
else if (ssDCval >= 67) {
if (gchDC < 2) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
return 2;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (ssDCval < 96) {
if (ssDCval < 81) {
if (negBalance < 24) {
if (Rs < 4410) {
if (AFD < 4) {
if (mu_mv_mag_sq < 114) {
return 8;
}
else if (mu_mv_mag_sq >= 114) {
if (mu_mv_mag_sq < 122) {
if (Rs < 4289) {
if (Rs < 4223) {
return 4;
}
else if (Rs >= 4223) {
return 4;
}
}
else if (Rs >= 4289) {
if (posBalance < 28) {
if (Rs < 4339) {
return 4;
}
else if (Rs >= 4339) {
return 4;
}
}
else if (posBalance >= 28) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 4245) {
return 8;
}
else if (Rs >= 4245) {
if (Rs < 4337) {
return 8;
}
else if (Rs >= 4337) {
return 8;
}
}
}
}
}
else if (AFD >= 4) {
return 8;
}
}
else if (Rs >= 4410) {
if (diffRsCsdiff < 9) {
return 4;
}
else if (diffRsCsdiff >= 9) {
if (Rs < 4531) {
return 4;
}
else if (Rs >= 4531) {
return 4;
}
}
}
}
else if (negBalance >= 24) {
return 4;
}
}
else if (ssDCval >= 81) {
if (MVDiff < 1) {
if (gchDC < 1) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 4604) {
return 2;
}
else if (Rs >= 4604) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (Rs < 4729) {
return 2;
}
else if (Rs >= 4729) {
return 8;
}
}
}
else if (MVDiff >= 1) {
return 2;
}
}
}
else if (ssDCval >= 96) {
if (Cs < 1920) {
return 4;
}
else if (Cs >= 1920) {
return 2;
}
}
}
}
}
else if (posBalance >= 41) {
if (ssDCval < 74) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 117) {
if (Cs < 2014) {
if (diffTSC < 1) {
if (Rs < 4036) {
return 8;
}
else if (Rs >= 4036) {
return 8;
}
}
else if (diffTSC >= 1) {
if (Rs < 5417) {
return 4;
}
else if (Rs >= 5417) {
return 4;
}
}
}
else if (Cs >= 2014) {
return 8;
}
}
else if (mu_mv_mag_sq >= 117) {
if (negBalance < 12) {
return 4;
}
else if (negBalance >= 12) {
if (Rs < 5873) {
if (Rs < 4504) {
return 4;
}
else if (Rs >= 4504) {
return 8;
}
}
else if (Rs >= 5873) {
return 4;
}
}
}
}
}
else if (ssDCval >= 74) {
if (posBalance < 44) {
if (ssDCval < 91) {
if (Rs < 3569) {
return 4;
}
else if (Rs >= 3569) {
return 2;
}
}
else if (ssDCval >= 91) {
if (negBalance < 27) {
if (Rs < 4477) {
if (Rs < 4408) {
return 4;
}
else if (Rs >= 4408) {
return 8;
}
}
else if (Rs >= 4477) {
return 2;
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (posBalance >= 44) {
if (RsDiff < 6) {
if (Rs < 3955) {
return 4;
}
else if (Rs >= 3955) {
return 8;
}
}
else if (RsDiff >= 6) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 2197) {
if (SC < 8344) {
if (Cs < 2819) {
if (mu_mv_mag_sq < 120) {
if (ssDCval < 101) {
if (SC < 5940) {
if (Rs < 2362) {
if (MVDiff < 2) {
if (Cs < 2448) {
return 8;
}
else if (Cs >= 2448) {
return 4;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Rs >= 2362) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Cs < 2475) {
if (negBalance < 62) {
return 8;
}
else if (negBalance >= 62) {
if (Rs < 2650) {
return 8;
}
else if (Rs >= 2650) {
return 4;
}
}
}
else if (Cs >= 2475) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (RsDiff < 8) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 3084) {
return 8;
}
else if (Rs >= 3084) {
return 1;
}
}
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
}
}
else if (SC >= 5940) {
if (negBalance < 30) {
if (ssDCval < 61) {
if (mu_mv_mag_sq < 113) {
if (diffMVdiffVal < 1) {
if (Cs < 2389) {
if (Cs < 2331) {
return 4;
}
else if (Cs >= 2331) {
if (Rs < 4564) {
return 4;
}
else if (Rs >= 4564) {
return 4;
}
}
}
else if (Cs >= 2389) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 2252) {
return 8;
}
else if (Cs >= 2252) {
if (Rs < 4673) {
return 4;
}
else if (Rs >= 4673) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (TSC < 4) {
if (Cs < 2703) {
if (negBalance < 19) {
if (negBalance < 18) {
if (posBalance < 47) {
if (mu_mv_mag_sq < 119) {
if (Rs < 5112) {
if (RsCsDiff < 33) {
if (negBalance < 13) {
return 4;
}
else if (negBalance >= 13) {
if (Rs < 5089) {
return 8;
}
else if (Rs >= 5089) {
return 4;
}
}
}
else if (RsCsDiff >= 33) {
if (posBalance < 37) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 4;
}
}
else if (posBalance >= 37) {
return 8;
}
}
}
else if (Rs >= 5112) {
return 8;
}
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 4655) {
return 4;
}
else if (Rs >= 4655) {
return 4;
}
}
}
else if (posBalance >= 47) {
if (Rs < 4886) {
return 4;
}
else if (Rs >= 4886) {
return 4;
}
}
}
else if (negBalance >= 18) {
return 4;
}
}
else if (negBalance >= 19) {
if (negBalance < 22) {
if (Cs < 2550) {
return 8;
}
else if (Cs >= 2550) {
return 8;
}
}
else if (negBalance >= 22) {
if (Cs < 2548) {
return 4;
}
else if (Cs >= 2548) {
return 8;
}
}
}
}
else if (Cs >= 2703) {
return 8;
}
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (ssDCval >= 61) {
if (Cs < 2331) {
return 4;
}
else if (Cs >= 2331) {
if (negBalance < 24) {
return 8;
}
else if (negBalance >= 24) {
if (Rs < 4873) {
return 8;
}
else if (Rs >= 4873) {
return 4;
}
}
}
}
}
else if (negBalance >= 30) {
if (Cs < 2609) {
if (mu_mv_mag_sq < 116) {
if (Rs < 5306) {
return 2;
}
else if (Rs >= 5306) {
return 4;
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
else if (Cs >= 2609) {
if (Rs < 4428) {
return 4;
}
else if (Rs >= 4428) {
return 8;
}
}
}
}
}
else if (ssDCval >= 101) {
if (Rs < 3836) {
return 8;
}
else if (Rs >= 3836) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (ssDCval < 47) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (ssDCval >= 47) {
if (Rs < 2143) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
if (Rs < 2010) {
return 4;
}
else if (Rs >= 2010) {
return 8;
}
}
}
else if (Rs >= 2143) {
if (Rs < 4973) {
if (diffRsCsdiff < -13) {
if (negBalance < 49) {
if (SC < 6352) {
if (Rs < 2232) {
if (Rs < 2204) {
return 8;
}
else if (Rs >= 2204) {
return 4;
}
}
else if (Rs >= 2232) {
return 8;
}
}
else if (SC >= 6352) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
return 4;
}
}
}
else if (negBalance >= 49) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
return 4;
}
}
}
else if (diffRsCsdiff >= -13) {
if (AFD < 7) {
if (Cs < 2283) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
return 8;
}
}
}
}
else if (Cs >= 2283) {
return 8;
}
}
else if (AFD >= 7) {
if (Rs < 3798) {
if (ssDCval < 99) {
return 8;
}
else if (ssDCval >= 99) {
if (Rs < 2947) {
return 8;
}
else if (Rs >= 2947) {
return 4;
}
}
}
else if (Rs >= 3798) {
if (Rs < 4170) {
return 4;
}
else if (Rs >= 4170) {
return 1;
}
}
}
}
}
else if (Rs >= 4973) {
if (Cs < 2474) {
return 4;
}
else if (Cs >= 2474) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Rs < 5120) {
if (Cs < 2729) {
return 8;
}
else if (Cs >= 2729) {
return 4;
}
}
else if (Rs >= 5120) {
return 4;
}
}
}
}
}
}
}
}
else if (Cs >= 2819) {
if (ssDCval < 85) {
if (negBalance < 3) {
return 4;
}
else if (negBalance >= 3) {
if (SC < 5462) {
if (Rs < 1325) {
return 8;
}
else if (Rs >= 1325) {
return 4;
}
}
else if (SC >= 5462) {
if (Rs < 2949) {
if (AFD < 9) {
if (Cs < 3508) {
if (Cs < 3124) {
if (Rs < 2660) {
return 4;
}
else if (Rs >= 2660) {
return 8;
}
}
else if (Cs >= 3124) {
return 8;
}
}
else if (Cs >= 3508) {
return 4;
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (Rs >= 2949) {
if (posBalance < 62) {
if (posBalance < 20) {
if (Rs < 3737) {
return 8;
}
else if (Rs >= 3737) {
return 4;
}
}
else if (posBalance >= 20) {
if (diffTSC < -2) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
else if (diffTSC >= -2) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (Rs < 3131) {
return 4;
}
else if (Rs >= 3131) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (ssDCval < 98) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Cs < 3082) {
if (AFD < 6) {
if (Rs < 4037) {
return 8;
}
else if (Rs >= 4037) {
return 4;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (Cs >= 3082) {
if (Rs < 3329) {
if (negBalance < 53) {
if (mu_mv_mag_sq < 116) {
if (Cs < 4241) {
return 4;
}
else if (Cs >= 4241) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
else if (negBalance >= 53) {
return 4;
}
}
else if (Rs >= 3329) {
if (AFD < 13) {
if (RsCsDiff < 38) {
if (mu_mv_mag_sq < 128) {
if (Rs < 3793) {
return 4;
}
else if (Rs >= 3793) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (Rs < 4196) {
return 8;
}
else if (Rs >= 4196) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 3125) {
return 8;
}
else if (Cs >= 3125) {
return 4;
}
}
}
else if (RsCsDiff >= 38) {
if (Cs < 3519) {
return 4;
}
else if (Cs >= 3519) {
if (Cs < 4177) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 8;
}
}
else if (Cs >= 4177) {
return 4;
}
}
}
}
else if (AFD >= 13) {
return 8;
}
}
}
}
}
else if (ssDCval >= 98) {
if (negBalance < 52) {
if (Cs < 4017) {
if (RsCsDiff < 121) {
return 8;
}
else if (RsCsDiff >= 121) {
if (Cs < 3182) {
return 4;
}
else if (Cs >= 3182) {
return 8;
}
}
}
else if (Cs >= 4017) {
return 4;
}
}
else if (negBalance >= 52) {
return 4;
}
}
}
}
}
else if (SC >= 8344) {
if (Rs < 6651) {
if (MVDiff < 3) {
if (ssDCval < 75) {
if (mu_mv_mag_sq < 111) {
return 2;
}
else if (mu_mv_mag_sq >= 111) {
if (Rs < 4552) {
if (negBalance < 29) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (negBalance >= 29) {
return 8;
}
}
else if (Rs >= 4552) {
return 8;
}
}
}
else if (ssDCval >= 75) {
if (Cs < 3666) {
if (Rs < 5347) {
return 4;
}
else if (Rs >= 5347) {
return 8;
}
}
else if (Cs >= 3666) {
if (mu_mv_mag_sq < 109) {
if (Rs < 6073) {
return 8;
}
else if (Rs >= 6073) {
return 4;
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (Rs < 5813) {
return 8;
}
else if (Rs >= 5813) {
if (Cs < 3980) {
return 8;
}
else if (Cs >= 3980) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
return 8;
}
}
}
}
}
else if (Rs >= 6651) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (negBalance < 37) {
if (ssDCval < 117) {
if (Rs < 1609) {
if (Rs < 1440) {
return 8;
}
else if (Rs >= 1440) {
return 2;
}
}
else if (Rs >= 1609) {
if (diffRsCsdiff < -7) {
if (Cs < 1706) {
if (Rs < 2283) {
return 8;
}
else if (Rs >= 2283) {
return 4;
}
}
else if (Cs >= 1706) {
if (mcTcor < 101) {
if (posBalance < 56) {
if (Cs < 1886) {
if (Rs < 3227) {
return 8;
}
else if (Rs >= 3227) {
return 4;
}
}
else if (Cs >= 1886) {
return 8;
}
}
else if (posBalance >= 56) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (mcTcor >= 101) {
return 2;
}
}
}
else if (diffRsCsdiff >= -7) {
if (refDCval < 103) {
if (Cs < 2018) {
return 4;
}
else if (Cs >= 2018) {
return 8;
}
}
else if (refDCval >= 103) {
return 8;
}
}
}
}
else if (ssDCval >= 117) {
if (ssDCval < 200) {
if (diffRsCsdiff < -1161) {
if (ssDCval < 153) {
return 1;
}
else if (ssDCval >= 153) {
return 8;
}
}
else if (diffRsCsdiff >= -1161) {
if (ssDCval < 118) {
if (CsDiff < 1) {
if (Cs < 2965) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (Cs >= 2965) {
return 2;
}
}
else if (CsDiff >= 1) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 2;
}
}
}
else if (ssDCval >= 118) {
if (Rs < 1258) {
if (ssDCval < 150) {
if (Rs < 1176) {
if (Cs < 3155) {
return 8;
}
else if (Cs >= 3155) {
return 2;
}
}
else if (Rs >= 1176) {
if (Cs < 1798) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (Cs >= 1798) {
return 8;
}
}
}
else if (ssDCval >= 150) {
return 1;
}
}
else if (Rs >= 1258) {
if (Rs < 4952) {
if (Cs < 3939) {
if (tcor < 100) {
if (negBalance < 6) {
if (mu_mv_mag_sq < 128) {
if (Rs < 2208) {
return 1;
}
else if (Rs >= 2208) {
if (Cs < 2295) {
return 8;
}
else if (Cs >= 2295) {
if (Rs < 3441) {
return 1;
}
else if (Rs >= 3441) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (negBalance >= 6) {
if (diffAFD < 6) {
return 8;
}
else if (diffAFD >= 6) {
return 1;
}
}
}
else if (tcor >= 100) {
if (negBalance < 23) {
if (SC < 3944) {
if (Rs < 2057) {
return 8;
}
else if (Rs >= 2057) {
return 1;
}
}
else if (SC >= 3944) {
return 8;
}
}
else if (negBalance >= 23) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 8;
}
}
}
}
else if (Cs >= 3939) {
if (Rs < 3868) {
return 1;
}
else if (Rs >= 3868) {
return 8;
}
}
}
else if (Rs >= 4952) {
if (Cs < 2805) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
return 8;
}
}
else if (Cs >= 2805) {
if (Cs < 3628) {
if (Rs < 6095) {
return 8;
}
else if (Rs >= 6095) {
return 4;
}
}
else if (Cs >= 3628) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 200) {
if (Rs < 2635) {
return 8;
}
else if (Rs >= 2635) {
return 4;
}
}
}
}
else if (negBalance >= 37) {
if (ssDCval < 137) {
if (RsDiff < 2) {
if (ssDCval < 120) {
if (posBalance < 53) {
if (ssDCval < 113) {
return 8;
}
else if (ssDCval >= 113) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (Rs < 2939) {
return 4;
}
else if (Rs >= 2939) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
return 2;
}
}
else if (ssDCval >= 120) {
if (Rs < 2587) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
else if (Rs >= 2587) {
if (Rs < 2703) {
return 4;
}
else if (Rs >= 2703) {
return 2;
}
}
}
}
else if (RsDiff >= 2) {
if (refDCval < 105) {
if (Cs < 2957) {
if (CsDiff < 6) {
if (refDCval < 104) {
if (Rs < 1777) {
return 2;
}
else if (Rs >= 1777) {
return 8;
}
}
else if (refDCval >= 104) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (Rs < 4932) {
return 1;
}
else if (Rs >= 4932) {
return 2;
}
}
}
else if (Cs >= 2957) {
return 8;
}
}
else if (refDCval >= 105) {
if (posBalance < 63) {
if (diffAFD < -2) {
if (Rs < 3979) {
if (Cs < 2358) {
if (ssDCval < 108) {
if (Rs < 2736) {
return 8;
}
else if (Rs >= 2736) {
return 4;
}
}
else if (ssDCval >= 108) {
return 8;
}
}
else if (Cs >= 2358) {
if (ssDCval < 120) {
return 2;
}
else if (ssDCval >= 120) {
if (Rs < 3382) {
return 1;
}
else if (Rs >= 3382) {
return 4;
}
}
}
}
else if (Rs >= 3979) {
return 8;
}
}
else if (diffAFD >= -2) {
if (gchDC < 1) {
if (diffAFD < 1) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Rs < 4280) {
if (Rs < 2472) {
return 8;
}
else if (Rs >= 2472) {
return 2;
}
}
else if (Rs >= 4280) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (negBalance < 46) {
if (diffRsCsdiff < 18) {
if (Rs < 1562) {
return 4;
}
else if (Rs >= 1562) {
return 8;
}
}
else if (diffRsCsdiff >= 18) {
if (Cs < 1977) {
if (Rs < 2867) {
return 8;
}
else if (Rs >= 2867) {
return 4;
}
}
else if (Cs >= 1977) {
if (ssDCval < 119) {
return 8;
}
else if (ssDCval >= 119) {
return 4;
}
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (tcor < 87) {
if (mcTcor < 86) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
return 1;
}
}
else if (mcTcor >= 86) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (tcor >= 87) {
if (tcor < 99) {
if (diffTSC < 3) {
if (diffMVdiffVal < -2) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (diffTSC >= 3) {
if (Rs < 4748) {
return 8;
}
else if (Rs >= 4748) {
return 2;
}
}
}
else if (tcor >= 99) {
if (Rs < 4528) {
return 2;
}
else if (Rs >= 4528) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 63) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Cs < 1985) {
return 1;
}
else if (Cs >= 1985) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 137) {
if (mcTcor < 96) {
if (Rs < 2759) {
return 8;
}
else if (Rs >= 2759) {
if (Cs < 4097) {
return 1;
}
else if (Cs >= 4097) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 144) {
if (mu_mv_mag_sq < 125) {
if (Rs < 5820) {
return 8;
}
else if (Rs >= 5820) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 3581) {
return 8;
}
else if (Rs >= 3581) {
return 1;
}
}
}
else if (refDCval >= 144) {
return 8;
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect8(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1308) {
if (Rs < 2380) {
if (Cs < 535) {
if (negBalance < 14) {
if (MVDiff < 7) {
if (SC < 783) {
if (ssDCval < 67) {
if (Cs < 5) {
if (ssDCval < 45) {
if (ssDCval < 22) {
if (ssDCval < 9) {
return 4;
}
else if (ssDCval >= 9) {
return 8;
}
}
else if (ssDCval >= 22) {
if (Rs < 45) {
return 4;
}
else if (Rs >= 45) {
return 8;
}
}
}
else if (ssDCval >= 45) {
return 2;
}
}
else if (Cs >= 5) {
if (Cs < 256) {
if (AFD < 1) {
if (Rs < 256) {
if (negBalance < 8) {
if (Cs < 186) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 2;
}
}
else if (Cs >= 186) {
if (Rs < 198) {
return 4;
}
else if (Rs >= 198) {
return 8;
}
}
}
else if (negBalance >= 8) {
if (SC < 59) {
if (ssDCval < 23) {
return 8;
}
else if (ssDCval >= 23) {
return 4;
}
}
else if (SC >= 59) {
if (Cs < 135) {
if (ssDCval < 40) {
if (posBalance < 35) {
if (negBalance < 12) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
return 1;
}
}
else if (negBalance >= 12) {
return 4;
}
}
else if (posBalance >= 35) {
return 1;
}
}
else if (ssDCval >= 40) {
return 8;
}
}
else if (Cs >= 135) {
if (ssDCval < 33) {
return 8;
}
else if (ssDCval >= 33) {
if (Rs < 186) {
if (Rs < 122) {
return 1;
}
else if (Rs >= 122) {
if (Rs < 128) {
if (Cs < 169) {
return 4;
}
else if (Cs >= 169) {
return 8;
}
}
else if (Rs >= 128) {
return 4;
}
}
}
else if (Rs >= 186) {
return 8;
}
}
}
}
}
}
else if (Rs >= 256) {
if (Cs < 157) {
if (tcor < 106) {
if (Cs < 101) {
return 8;
}
else if (Cs >= 101) {
if (Rs < 432) {
return 4;
}
else if (Rs >= 432) {
return 8;
}
}
}
else if (tcor >= 106) {
return 8;
}
}
else if (Cs >= 157) {
return 8;
}
}
}
else if (AFD >= 1) {
if (ssDCval < 44) {
if (ssDCval < 38) {
if (ssDCval < 28) {
if (posBalance < 49) {
if (Rs < 44) {
return 4;
}
else if (Rs >= 44) {
return 1;
}
}
else if (posBalance >= 49) {
if (Rs < 261) {
return 8;
}
else if (Rs >= 261) {
return 4;
}
}
}
else if (ssDCval >= 28) {
if (Rs < 218) {
if (refDCval < 32) {
if (Rs < 110) {
return 4;
}
else if (Rs >= 110) {
return 1;
}
}
else if (refDCval >= 32) {
if (ssDCval < 34) {
return 8;
}
else if (ssDCval >= 34) {
return 4;
}
}
}
else if (Rs >= 218) {
return 8;
}
}
}
else if (ssDCval >= 38) {
if (Cs < 73) {
return 8;
}
else if (Cs >= 73) {
return 1;
}
}
}
else if (ssDCval >= 44) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 107) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (mu_mv_mag_sq >= 107) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (diffAFD < 1) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
return 4;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
}
}
else if (posBalance >= 53) {
return 8;
}
}
}
}
else if (Cs >= 256) {
if (Rs < 205) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (ssDCval < 36) {
if (Cs < 273) {
return 4;
}
else if (Cs >= 273) {
return 8;
}
}
else if (ssDCval >= 36) {
return 1;
}
}
}
else if (Rs >= 205) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 40) {
if (Rs < 397) {
return 4;
}
else if (Rs >= 397) {
return 1;
}
}
else if (ssDCval >= 40) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 374) {
if (gchDC < 1) {
if (ssDCval < 23) {
return 8;
}
else if (ssDCval >= 23) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Rs >= 374) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (posBalance < 28) {
if (Cs < 276) {
return 1;
}
else if (Cs >= 276) {
if (Rs < 367) {
return 2;
}
else if (Rs >= 367) {
if (Rs < 463) {
return 4;
}
else if (Rs >= 463) {
return 1;
}
}
}
}
else if (posBalance >= 28) {
if (ssDCval < 94) {
if (Rs < 290) {
if (Cs < 130) {
if (Rs < 103) {
return 4;
}
else if (Rs >= 103) {
if (Rs < 171) {
return 2;
}
else if (Rs >= 171) {
return 8;
}
}
}
else if (Cs >= 130) {
if (Cs < 258) {
return 2;
}
else if (Cs >= 258) {
return 4;
}
}
}
else if (Rs >= 290) {
if (mu_mv_mag_sq < 121) {
if (SC < 644) {
return 4;
}
else if (SC >= 644) {
return 8;
}
}
else if (mu_mv_mag_sq >= 121) {
if (SC < 514) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (SC >= 514) {
return 2;
}
}
}
}
else if (ssDCval >= 94) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (Rs < 366) {
return 8;
}
else if (Rs >= 366) {
return 1;
}
}
}
}
}
}
else if (SC >= 783) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 33) {
if (Cs < 286) {
return 4;
}
else if (Cs >= 286) {
if (Cs < 336) {
return 8;
}
else if (Cs >= 336) {
return 1;
}
}
}
else if (ssDCval >= 33) {
if (Cs < 471) {
if (SC < 988) {
if (ssDCval < 127) {
if (RsCsDiff < 2) {
if (Rs < 515) {
return 8;
}
else if (Rs >= 515) {
if (Cs < 349) {
return 8;
}
else if (Cs >= 349) {
if (Cs < 398) {
return 1;
}
else if (Cs >= 398) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 127) {
if (Rs < 483) {
if (Cs < 374) {
return 4;
}
else if (Cs >= 374) {
return 1;
}
}
else if (Rs >= 483) {
if (Rs < 547) {
if (Rs < 516) {
return 8;
}
else if (Rs >= 516) {
if (Cs < 401) {
return 1;
}
else if (Cs >= 401) {
return 8;
}
}
}
else if (Rs >= 547) {
return 8;
}
}
}
}
else if (SC >= 988) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (RsCsDiff < 2) {
if (Rs < 890) {
return 8;
}
else if (Rs >= 890) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
if (Rs < 1093) {
return 1;
}
else if (Rs >= 1093) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
if (Rs < 824) {
if (Rs < 763) {
return 2;
}
else if (Rs >= 763) {
return 1;
}
}
else if (Rs >= 824) {
return 4;
}
}
}
else if (ssDCval >= 60) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Rs < 1439) {
if (Rs < 1144) {
return 8;
}
else if (Rs >= 1144) {
return 2;
}
}
else if (Rs >= 1439) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (TSCindex < 2) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Cs < 204) {
return 8;
}
else if (Cs >= 204) {
if (Cs < 304) {
return 2;
}
else if (Cs >= 304) {
return 8;
}
}
}
}
else if (TSCindex >= 2) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (MVDiff < 4) {
if (Rs < 1624) {
return 8;
}
else if (Rs >= 1624) {
return 2;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
}
}
}
}
else if (Cs >= 471) {
if (Rs < 1381) {
if (ssDCval < 55) {
if (Cs < 484) {
if (Rs < 894) {
return 8;
}
else if (Rs >= 894) {
return 1;
}
}
else if (Cs >= 484) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
if (Rs < 654) {
return 1;
}
else if (Rs >= 654) {
return 2;
}
}
}
}
else if (ssDCval >= 55) {
if (Rs < 491) {
return 1;
}
else if (Rs >= 491) {
return 8;
}
}
}
else if (Rs >= 1381) {
if (Rs < 2273) {
return 2;
}
else if (Rs >= 2273) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (refDCval < 60) {
return 8;
}
else if (refDCval >= 60) {
if (negBalance < 11) {
if (SC < 1101) {
if (Rs < 659) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
return 1;
}
}
else if (Rs >= 659) {
return 1;
}
}
else if (SC >= 1101) {
if (ssDCval < 66) {
if (Rs < 1115) {
return 8;
}
else if (Rs >= 1115) {
return 2;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
}
else if (negBalance >= 11) {
if (Rs < 594) {
return 8;
}
else if (Rs >= 594) {
if (ssDCval < 98) {
if (Rs < 872) {
if (Cs < 346) {
if (Rs < 718) {
return 2;
}
else if (Rs >= 718) {
return 1;
}
}
else if (Cs >= 346) {
return 4;
}
}
else if (Rs >= 872) {
if (Rs < 1077) {
return 8;
}
else if (Rs >= 1077) {
return 2;
}
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (tcor < 94) {
if (negBalance < 4) {
if (TSC < 9) {
if (refDCval < 15) {
if (Rs < 148) {
return 4;
}
else if (Rs >= 148) {
return 1;
}
}
else if (refDCval >= 15) {
if (Rs < 41) {
if (ssDCval < 50) {
if (posBalance < 72) {
if (TSCindex < 1) {
if (posBalance < 51) {
if (negBalance < 1) {
if (Rs < 1) {
return 4;
}
else if (Rs >= 1) {
if (Rs < 7) {
return 8;
}
else if (Rs >= 7) {
return 4;
}
}
}
else if (negBalance >= 1) {
return 8;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (TSCindex >= 1) {
return 4;
}
}
else if (posBalance >= 72) {
if (mu_mv_mag_sq < 9) {
if (Rs < 6) {
return 2;
}
else if (Rs >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 9) {
return 8;
}
}
}
else if (ssDCval >= 50) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
}
else if (Rs >= 41) {
if (RsCsDiff < 7) {
if (Cs < 159) {
if (Rs < 227) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (Cs < 60) {
return 4;
}
else if (Cs >= 60) {
return 8;
}
}
}
}
else if (Rs >= 227) {
if (Rs < 1396) {
if (diffRsCsdiff < -1) {
return 8;
}
else if (diffRsCsdiff >= -1) {
return 4;
}
}
else if (Rs >= 1396) {
return 8;
}
}
}
else if (Cs >= 159) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (Cs < 138) {
return 1;
}
else if (Cs >= 138) {
return 4;
}
}
}
}
}
else if (TSC >= 9) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (negBalance >= 4) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (mu_mv_mag_sq < 41) {
if (mcTcor < 96) {
if (AFD < 12) {
if (AFD < 3) {
if (CsDiff < 1) {
if (Rs < 14) {
return 2;
}
else if (Rs >= 14) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 2;
}
}
}
else if (CsDiff >= 1) {
return 2;
}
}
else if (AFD >= 3) {
if (Rs < 156) {
if (ssDCval < 46) {
if (Rs < 67) {
return 8;
}
else if (Rs >= 67) {
return 2;
}
}
else if (ssDCval >= 46) {
return 4;
}
}
else if (Rs >= 156) {
if (MVDiff < 29) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
return 2;
}
}
else if (MVDiff >= 29) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
else if (AFD >= 12) {
if (SC < 1217) {
if (SC < 658) {
return 1;
}
else if (SC >= 658) {
return 2;
}
}
else if (SC >= 1217) {
return 1;
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (mu_mv_mag_sq >= 41) {
if (refDCval < 42) {
if (tcor < 89) {
if (posBalance < 67) {
if (Cs < 183) {
return 1;
}
else if (Cs >= 183) {
if (Rs < 172) {
return 8;
}
else if (Rs >= 172) {
if (ssDCval < 21) {
return 8;
}
else if (ssDCval >= 21) {
return 1;
}
}
}
}
else if (posBalance >= 67) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (tcor >= 89) {
if (MVDiff < 29) {
if (posBalance < 30) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (posBalance >= 30) {
return 1;
}
}
else if (MVDiff >= 29) {
if (Rs < 63) {
return 1;
}
else if (Rs >= 63) {
return 4;
}
}
}
}
else if (refDCval >= 42) {
if (MVDiff < 37) {
if (posBalance < 108) {
return 8;
}
else if (posBalance >= 108) {
return 4;
}
}
else if (MVDiff >= 37) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (Cs < 359) {
return 1;
}
else if (Cs >= 359) {
return 8;
}
}
}
}
}
}
}
}
else if (tcor >= 94) {
if (mu_mv_mag_sq < 108) {
if (SC < 941) {
if (Cs < 50) {
if (Cs < 8) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
if (posBalance < 60) {
if (diffAFD < -1) {
if (Rs < 420) {
return 2;
}
else if (Rs >= 420) {
return 4;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (posBalance >= 60) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (AFD < 20) {
if (Rs < 91) {
return 1;
}
else if (Rs >= 91) {
return 4;
}
}
else if (AFD >= 20) {
return 2;
}
}
}
}
}
else if (Cs >= 8) {
if (posBalance < 107) {
if (Rs < 358) {
return 1;
}
else if (Rs >= 358) {
return 8;
}
}
else if (posBalance >= 107) {
if (Rs < 450) {
return 4;
}
else if (Rs >= 450) {
return 1;
}
}
}
}
else if (Cs >= 50) {
if (refDCval < 40) {
if (posBalance < 20) {
if (Rs < 142) {
if (Rs < 90) {
return 1;
}
else if (Rs >= 90) {
return 8;
}
}
else if (Rs >= 142) {
return 4;
}
}
else if (posBalance >= 20) {
return 4;
}
}
else if (refDCval >= 40) {
if (posBalance < 31) {
if (ssDCval < 161) {
if (diffRsCsdiff < -1) {
return 8;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (ssDCval >= 161) {
return 4;
}
}
else if (posBalance >= 31) {
if (posBalance < 72) {
if (ssDCval < 82) {
if (Rs < 263) {
if (diffMVdiffVal < 4) {
return 1;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 101) {
return 8;
}
else if (Rs >= 101) {
return 2;
}
}
}
else if (Rs >= 263) {
return 4;
}
}
else if (ssDCval >= 82) {
if (Rs < 419) {
return 4;
}
else if (Rs >= 419) {
return 1;
}
}
}
else if (posBalance >= 72) {
if (Cs < 99) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 99) {
return 8;
}
}
}
}
}
}
else if (SC >= 941) {
if (ssDCval < 99) {
if (SC < 1399) {
if (Rs < 707) {
return 1;
}
else if (Rs >= 707) {
return 8;
}
}
else if (SC >= 1399) {
return 1;
}
}
else if (ssDCval >= 99) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (ssDCval < 125) {
if (diffRsCsdiff < -12) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 1;
}
}
else if (diffRsCsdiff >= -12) {
if (diffMVdiffVal < -1) {
if (posBalance < 2) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (posBalance >= 2) {
if (Rs < 80) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Rs >= 80) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 194) {
return 8;
}
else if (Rs >= 194) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (MVDiff < 1029) {
if (AFD < 1) {
if (MVDiff < 132) {
return 8;
}
else if (MVDiff >= 132) {
if (MVDiff < 179) {
return 4;
}
else if (MVDiff >= 179) {
return 8;
}
}
}
else if (AFD >= 1) {
if (Rs < 111) {
return 8;
}
else if (Rs >= 111) {
return 4;
}
}
}
else if (MVDiff >= 1029) {
return 4;
}
}
}
}
else if (ssDCval >= 125) {
if (ssDCval < 130) {
if (diffMVdiffVal < -61) {
if (Rs < 378) {
return 8;
}
else if (Rs >= 378) {
return 4;
}
}
else if (diffMVdiffVal >= -61) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 4;
}
}
}
else if (ssDCval >= 130) {
if (Rs < 881) {
if (MVDiff < 37) {
return 8;
}
else if (MVDiff >= 37) {
if (ssDCval < 169) {
return 4;
}
else if (ssDCval >= 169) {
return 8;
}
}
}
else if (Rs >= 881) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 35) {
if (mcTcor < 82) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (tcor < 60) {
if (Cs < 389) {
if (negBalance < 45) {
if (diffTSC < 1) {
if (Cs < 308) {
if (posBalance < 68) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 68) {
return 8;
}
}
else if (Cs >= 308) {
if (Rs < 413) {
return 2;
}
else if (Rs >= 413) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (mu_mv_mag_sq < 218) {
return 1;
}
else if (mu_mv_mag_sq >= 218) {
return 4;
}
}
}
else if (negBalance >= 45) {
if (SC < 594) {
if (MVDiff < 377) {
return 8;
}
else if (MVDiff >= 377) {
return 4;
}
}
else if (SC >= 594) {
return 1;
}
}
}
else if (Cs >= 389) {
if (Rs < 563) {
return 2;
}
else if (Rs >= 563) {
return 8;
}
}
}
else if (tcor >= 60) {
if (ssDCval < 23) {
if (MVDiff < 287) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
return 4;
}
}
else if (MVDiff >= 287) {
return 8;
}
}
else if (ssDCval >= 23) {
if (mcTcor < 72) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 840) {
return 4;
}
else if (Rs >= 840) {
return 1;
}
}
}
else if (mcTcor >= 72) {
if (Cs < 206) {
if (MVDiff < 68) {
return 1;
}
else if (MVDiff >= 68) {
if (Rs < 166) {
return 1;
}
else if (Rs >= 166) {
return 8;
}
}
}
else if (Cs >= 206) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
if (Rs < 208) {
return 2;
}
else if (Rs >= 208) {
if (MVDiff < 36) {
return 1;
}
else if (MVDiff >= 36) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 82) {
if (posBalance < 27) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 110) {
if (refDCval < 28) {
if (Rs < 13) {
return 8;
}
else if (Rs >= 13) {
return 1;
}
}
else if (refDCval >= 28) {
if (Cs < 258) {
if (ssDCval < 32) {
return 4;
}
else if (ssDCval >= 32) {
if (Rs < 93) {
return 8;
}
else if (Rs >= 93) {
return 4;
}
}
}
else if (Cs >= 258) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 110) {
return 8;
}
}
else if (MVDiff >= 6) {
if (posBalance < 20) {
if (SC < 37) {
if (SC < 9) {
if (negBalance < 79) {
if (refDCval < 24) {
if (MVDiff < 247) {
return 8;
}
else if (MVDiff >= 247) {
return 4;
}
}
else if (refDCval >= 24) {
return 4;
}
}
else if (negBalance >= 79) {
return 8;
}
}
else if (SC >= 9) {
return 8;
}
}
else if (SC >= 37) {
if (mcTcor < 97) {
return 1;
}
else if (mcTcor >= 97) {
if (Rs < 18) {
return 1;
}
else if (Rs >= 18) {
return 4;
}
}
}
}
else if (posBalance >= 20) {
if (ssDCval < 31) {
return 1;
}
else if (ssDCval >= 31) {
if (refDCval < 34) {
if (Rs < 49) {
return 4;
}
else if (Rs >= 49) {
return 8;
}
}
else if (refDCval >= 34) {
return 1;
}
}
}
}
}
else if (posBalance >= 27) {
if (Cs < 71) {
if (posBalance < 69) {
if (Rs < 105) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
else if (Rs >= 105) {
if (SC < 192) {
if (mu_mv_mag_sq < 36) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
return 1;
}
}
else if (mu_mv_mag_sq >= 36) {
return 1;
}
}
else if (SC >= 192) {
if (posBalance < 28) {
if (Rs < 219) {
return 1;
}
else if (Rs >= 219) {
return 2;
}
}
else if (posBalance >= 28) {
return 1;
}
}
}
}
else if (posBalance >= 69) {
if (Rs < 5) {
return 4;
}
else if (Rs >= 5) {
return 2;
}
}
}
else if (Cs >= 71) {
if (Rs < 176) {
if (ssDCval < 32) {
if (RsCsDiff < 2) {
if (refDCval < 31) {
return 1;
}
else if (refDCval >= 31) {
if (posBalance < 35) {
if (Cs < 93) {
return 1;
}
else if (Cs >= 93) {
if (Rs < 138) {
return 8;
}
else if (Rs >= 138) {
return 1;
}
}
}
else if (posBalance >= 35) {
return 1;
}
}
}
else if (RsCsDiff >= 2) {
if (MVDiff < 46) {
if (negBalance < 41) {
if (posBalance < 32) {
if (Rs < 107) {
return 1;
}
else if (Rs >= 107) {
return 8;
}
}
else if (posBalance >= 32) {
if (Rs < 167) {
return 1;
}
else if (Rs >= 167) {
if (Rs < 171) {
return 8;
}
else if (Rs >= 171) {
return 1;
}
}
}
}
else if (negBalance >= 41) {
return 8;
}
}
else if (MVDiff >= 46) {
if (Cs < 194) {
return 8;
}
else if (Cs >= 194) {
return 2;
}
}
}
}
else if (ssDCval >= 32) {
if (negBalance < 42) {
if (Cs < 261) {
if (mu_mv_mag_sq < 114) {
if (negBalance < 32) {
if (MVDiff < 3) {
if (Rs < 112) {
return 8;
}
else if (Rs >= 112) {
return 1;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (negBalance >= 32) {
if (ssDCval < 33) {
return 1;
}
else if (ssDCval >= 33) {
if (Cs < 138) {
return 1;
}
else if (Cs >= 138) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 1;
}
}
else if (Cs >= 261) {
return 1;
}
}
else if (negBalance >= 42) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (Rs < 71) {
return 4;
}
else if (Rs >= 71) {
return 8;
}
}
}
}
}
else if (Rs >= 176) {
if (posBalance < 39) {
if (ssDCval < 30) {
if (Cs < 373) {
if (MVDiff < 23) {
if (mcTcor < 85) {
if (Rs < 379) {
return 1;
}
else if (Rs >= 379) {
return 4;
}
}
else if (mcTcor >= 85) {
return 1;
}
}
else if (MVDiff >= 23) {
if (refDCval < 28) {
return 1;
}
else if (refDCval >= 28) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
}
}
else if (Cs >= 373) {
return 4;
}
}
else if (ssDCval >= 30) {
if (Cs < 260) {
if (AFD < 5) {
if (MVDiff < 4) {
if (Rs < 355) {
return 8;
}
else if (Rs >= 355) {
if (Rs < 497) {
return 4;
}
else if (Rs >= 497) {
return 1;
}
}
}
else if (MVDiff >= 4) {
if (MVDiff < 12) {
if (Cs < 86) {
return 1;
}
else if (Cs >= 86) {
if (MVDiff < 5) {
if (Rs < 461) {
return 8;
}
else if (Rs >= 461) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
}
else if (MVDiff >= 12) {
if (Rs < 274) {
return 1;
}
else if (Rs >= 274) {
if (Rs < 530) {
return 8;
}
else if (Rs >= 530) {
return 4;
}
}
}
}
}
else if (AFD >= 5) {
if (Rs < 323) {
return 1;
}
else if (Rs >= 323) {
if (Rs < 527) {
return 2;
}
else if (Rs >= 527) {
return 4;
}
}
}
}
else if (Cs >= 260) {
return 1;
}
}
}
else if (posBalance >= 39) {
if (negBalance < 33) {
if (MVDiff < 35) {
if (ssDCval < 34) {
if (Cs < 202) {
if (MVDiff < 5) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
return 8;
}
}
else if (MVDiff >= 5) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (Cs >= 202) {
return 1;
}
}
else if (ssDCval >= 34) {
if (diffMVdiffVal < 4) {
if (TSC < 3) {
if (Rs < 236) {
return 4;
}
else if (Rs >= 236) {
return 1;
}
}
else if (TSC >= 3) {
if (Rs < 244) {
return 1;
}
else if (Rs >= 244) {
return 8;
}
}
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
}
else if (MVDiff >= 35) {
if (ssDCval < 32) {
return 8;
}
else if (ssDCval >= 32) {
return 1;
}
}
}
else if (negBalance >= 33) {
if (Rs < 320) {
if (Rs < 259) {
return 1;
}
else if (Rs >= 259) {
return 8;
}
}
else if (Rs >= 320) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 35) {
if (Rs < 300) {
if (ssDCval < 51) {
if (AFD < 4) {
if (Cs < 191) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 75) {
if (negBalance < 47) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
return 8;
}
}
else if (negBalance >= 47) {
if (Rs < 9) {
return 2;
}
else if (Rs >= 9) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (ssDCval < 39) {
if (Rs < 159) {
if (ssDCval < 34) {
return 1;
}
else if (ssDCval >= 34) {
return 4;
}
}
else if (Rs >= 159) {
return 1;
}
}
else if (ssDCval >= 39) {
if (SC < 191) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (Cs < 85) {
return 4;
}
else if (Cs >= 85) {
return 1;
}
}
}
else if (SC >= 191) {
if (negBalance < 19) {
if (Rs < 250) {
return 1;
}
else if (Rs >= 250) {
return 4;
}
}
else if (negBalance >= 19) {
if (TSCindex < 1) {
return 4;
}
else if (TSCindex >= 1) {
if (posBalance < 37) {
if (Rs < 122) {
return 8;
}
else if (Rs >= 122) {
if (Cs < 129) {
return 8;
}
else if (Cs >= 129) {
return 4;
}
}
}
else if (posBalance >= 37) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 49) {
if (Rs < 71) {
if (Cs < 63) {
if (Rs < 69) {
return 4;
}
else if (Rs >= 69) {
return 8;
}
}
else if (Cs >= 63) {
return 1;
}
}
else if (Rs >= 71) {
if (mcTcor < 97) {
if (posBalance < 69) {
return 8;
}
else if (posBalance >= 69) {
return 4;
}
}
else if (mcTcor >= 97) {
if (Cs < 113) {
return 1;
}
else if (Cs >= 113) {
return 4;
}
}
}
}
}
else if (Cs >= 191) {
if (ssDCval < 40) {
if (tcor < 91) {
if (negBalance < 17) {
return 4;
}
else if (negBalance >= 17) {
return 1;
}
}
else if (tcor >= 91) {
if (Rs < 107) {
return 1;
}
else if (Rs >= 107) {
if (SC < 562) {
if (Cs < 213) {
if (Rs < 185) {
return 4;
}
else if (Rs >= 185) {
return 2;
}
}
else if (Cs >= 213) {
return 8;
}
}
else if (SC >= 562) {
if (SC < 638) {
return 1;
}
else if (SC >= 638) {
return 8;
}
}
}
}
}
else if (ssDCval >= 40) {
if (refDCval < 51) {
return 1;
}
else if (refDCval >= 51) {
if (Rs < 211) {
return 4;
}
else if (Rs >= 211) {
return 1;
}
}
}
}
}
else if (AFD >= 4) {
if (mcTcor < 93) {
if (tcor < 65) {
if (posBalance < 50) {
if (SC < 123) {
if (MVDiff < 146) {
if (SC < 96) {
return 2;
}
else if (SC >= 96) {
return 4;
}
}
else if (MVDiff >= 146) {
return 1;
}
}
else if (SC >= 123) {
if (diffTSC < 6) {
if (Rs < 45) {
return 8;
}
else if (Rs >= 45) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (Rs < 157) {
return 1;
}
else if (Rs >= 157) {
return 2;
}
}
}
}
else if (diffTSC >= 6) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (tcor < 18) {
if (Rs < 117) {
return 4;
}
else if (Rs >= 117) {
return 1;
}
}
else if (tcor >= 18) {
if (negBalance < 52) {
if (posBalance < 62) {
if (diffMVdiffVal < 6) {
if (refDCval < 38) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 1;
}
}
else if (refDCval >= 38) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
else if (posBalance >= 62) {
if (posBalance < 74) {
if (refDCval < 39) {
if (Cs < 193) {
return 8;
}
else if (Cs >= 193) {
if (Rs < 269) {
return 4;
}
else if (Rs >= 269) {
return 8;
}
}
}
else if (refDCval >= 39) {
return 4;
}
}
else if (posBalance >= 74) {
if (mcTcor < 63) {
if (Rs < 190) {
return 1;
}
else if (Rs >= 190) {
return 2;
}
}
else if (mcTcor >= 63) {
return 8;
}
}
}
}
else if (negBalance >= 52) {
if (gchDC < 4) {
if (ssDCval < 42) {
if (Cs < 266) {
return 8;
}
else if (Cs >= 266) {
return 4;
}
}
else if (ssDCval >= 42) {
if (Rs < 158) {
return 2;
}
else if (Rs >= 158) {
return 8;
}
}
}
else if (gchDC >= 4) {
return 4;
}
}
}
}
}
else if (tcor >= 65) {
if (refDCval < 38) {
if (Cs < 128) {
if (mu_mv_mag_sq < 5) {
return 8;
}
else if (mu_mv_mag_sq >= 5) {
if (SC < 182) {
if (Rs < 56) {
return 1;
}
else if (Rs >= 56) {
return 2;
}
}
else if (SC >= 182) {
return 1;
}
}
}
else if (Cs >= 128) {
if (tcor < 67) {
return 2;
}
else if (tcor >= 67) {
if (posBalance < 53) {
if (SC < 360) {
if (Rs < 104) {
return 1;
}
else if (Rs >= 104) {
return 2;
}
}
else if (SC >= 360) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
if (Rs < 250) {
return 2;
}
else if (Rs >= 250) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (Rs < 191) {
return 2;
}
else if (Rs >= 191) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 38) {
if (mcTcor < 86) {
if (SC < 280) {
if (Rs < 129) {
if (Rs < 115) {
return 2;
}
else if (Rs >= 115) {
return 4;
}
}
else if (Rs >= 129) {
return 1;
}
}
else if (SC >= 280) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 134) {
return 8;
}
else if (Rs >= 134) {
if (Rs < 176) {
return 4;
}
else if (Rs >= 176) {
return 1;
}
}
}
}
}
else if (mcTcor >= 86) {
if (mu_mv_mag_sq < 35) {
if (tcor < 74) {
if (gchDC < 4) {
if (mu_mv_mag_sq < 9) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 1;
}
}
else if (mu_mv_mag_sq >= 9) {
return 1;
}
}
else if (gchDC >= 4) {
return 2;
}
}
else if (tcor >= 74) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
if (Cs < 22) {
return 1;
}
else if (Cs >= 22) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (negBalance < 38) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
if (Rs < 249) {
return 4;
}
else if (Rs >= 249) {
return 2;
}
}
}
else if (negBalance >= 38) {
if (Cs < 315) {
if (diffMVdiffVal < -10) {
return 1;
}
else if (diffMVdiffVal >= -10) {
return 8;
}
}
else if (Cs >= 315) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (posBalance < 24) {
if (negBalance < 109) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 1;
}
}
else if (negBalance >= 109) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (posBalance >= 24) {
if (refDCval < 36) {
if (Cs < 170) {
if (Rs < 85) {
return 2;
}
else if (Rs >= 85) {
if (Cs < 95) {
return 1;
}
else if (Cs >= 95) {
return 4;
}
}
}
else if (Cs >= 170) {
if (Cs < 286) {
return 8;
}
else if (Cs >= 286) {
return 1;
}
}
}
else if (refDCval >= 36) {
if (posBalance < 69) {
if (Cs < 117) {
if (tcor < 78) {
if (Rs < 125) {
if (Cs < 8) {
return 2;
}
else if (Cs >= 8) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Rs < 68) {
return 1;
}
else if (Rs >= 68) {
return 2;
}
}
}
}
else if (Rs >= 125) {
return 1;
}
}
else if (tcor >= 78) {
if (ssDCval < 37) {
if (refDCval < 37) {
return 1;
}
else if (refDCval >= 37) {
if (Rs < 176) {
return 4;
}
else if (Rs >= 176) {
return 1;
}
}
}
else if (ssDCval >= 37) {
return 1;
}
}
}
else if (Cs >= 117) {
if (AFD < 8) {
if (MVDiff < 21) {
if (Cs < 390) {
if (posBalance < 52) {
if (diffRsCsdiff < 1) {
if (diffMVdiffVal < 2) {
if (refDCval < 47) {
if (diffRsCsdiff < -1) {
return 1;
}
else if (diffRsCsdiff >= -1) {
if (diffAFD < -1) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (refDCval >= 47) {
return 4;
}
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
else if (diffRsCsdiff >= 1) {
if (RsCsDiff < 5) {
return 1;
}
else if (RsCsDiff >= 5) {
return 4;
}
}
}
else if (posBalance >= 52) {
if (SC < 466) {
return 1;
}
else if (SC >= 466) {
return 2;
}
}
}
else if (Cs >= 390) {
if (posBalance < 44) {
if (Rs < 216) {
return 1;
}
else if (Rs >= 216) {
return 2;
}
}
else if (posBalance >= 44) {
return 1;
}
}
}
else if (MVDiff >= 21) {
if (ssDCval < 44) {
if (Rs < 104) {
return 4;
}
else if (Rs >= 104) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 1;
}
}
}
else if (ssDCval >= 44) {
return 4;
}
}
}
else if (AFD >= 8) {
if (Rs < 294) {
if (RsCsDiff < 37) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (Cs < 197) {
return 4;
}
else if (Cs >= 197) {
return 1;
}
}
}
}
else if (RsCsDiff >= 37) {
if (Rs < 238) {
return 4;
}
else if (Rs >= 238) {
return 8;
}
}
}
else if (Rs >= 294) {
return 4;
}
}
}
}
else if (posBalance >= 69) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 51) {
if (refDCval < 70) {
if (posBalance < 33) {
if (refDCval < 59) {
if (Cs < 264) {
if (SC < 128) {
return 2;
}
else if (SC >= 128) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (Rs < 130) {
return 4;
}
else if (Rs >= 130) {
return 8;
}
}
}
}
else if (Cs >= 264) {
if (Rs < 243) {
return 4;
}
else if (Rs >= 243) {
return 1;
}
}
}
else if (refDCval >= 59) {
if (tcor < 98) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (Rs < 212) {
if (diffMVdiffVal < 13) {
return 1;
}
else if (diffMVdiffVal >= 13) {
return 2;
}
}
else if (Rs >= 212) {
return 8;
}
}
}
else if (tcor >= 98) {
if (Rs < 91) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 2;
}
}
else if (Rs >= 91) {
if (MVDiff < 4) {
if (Rs < 155) {
return 2;
}
else if (Rs >= 155) {
return 4;
}
}
else if (MVDiff >= 4) {
if (Rs < 268) {
return 8;
}
else if (Rs >= 268) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 33) {
if (TSCindex < 2) {
if (ssDCval < 64) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Cs < 119) {
if (Rs < 89) {
return 1;
}
else if (Rs >= 89) {
return 8;
}
}
else if (Cs >= 119) {
if (tcor < 99) {
if (MVDiff < 3) {
if (Cs < 374) {
return 1;
}
else if (Cs >= 374) {
return 4;
}
}
else if (MVDiff >= 3) {
if (Rs < 168) {
return 4;
}
else if (Rs >= 168) {
return 1;
}
}
}
else if (tcor >= 99) {
if (Cs < 190) {
if (ssDCval < 57) {
if (refDCval < 54) {
return 1;
}
else if (refDCval >= 54) {
return 4;
}
}
else if (ssDCval >= 57) {
if (Cs < 181) {
if (Rs < 143) {
return 1;
}
else if (Rs >= 143) {
return 4;
}
}
else if (Cs >= 181) {
return 1;
}
}
}
else if (Cs >= 190) {
if (Rs < 255) {
return 4;
}
else if (Rs >= 255) {
if (Rs < 281) {
return 1;
}
else if (Rs >= 281) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (Cs < 241) {
return 2;
}
else if (Cs >= 241) {
if (MVDiff < 2) {
if (Cs < 378) {
return 4;
}
else if (Cs >= 378) {
return 1;
}
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
else if (TSCindex >= 2) {
if (RsCsDiff < 1) {
if (Cs < 186) {
if (MVDiff < 2) {
if (ssDCval < 63) {
if (tcor < 95) {
if (Rs < 87) {
return 2;
}
else if (Rs >= 87) {
if (Rs < 141) {
return 8;
}
else if (Rs >= 141) {
return 1;
}
}
}
else if (tcor >= 95) {
return 1;
}
}
else if (ssDCval >= 63) {
if (Cs < 112) {
return 1;
}
else if (Cs >= 112) {
if (Rs < 198) {
return 2;
}
else if (Rs >= 198) {
return 4;
}
}
}
}
else if (MVDiff >= 2) {
if (mcTcor < 95) {
if (Rs < 51) {
if (MVDiff < 49) {
return 1;
}
else if (MVDiff >= 49) {
return 4;
}
}
else if (Rs >= 51) {
if (Cs < 87) {
return 2;
}
else if (Cs >= 87) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (Cs < 57) {
return 1;
}
else if (Cs >= 57) {
if (MVDiff < 15) {
if (negBalance < 22) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (Rs < 184) {
return 1;
}
else if (Rs >= 184) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (Rs < 207) {
if (gchDC < 2) {
if (posBalance < 40) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (Rs < 145) {
if (Rs < 101) {
return 1;
}
else if (Rs >= 101) {
return 4;
}
}
else if (Rs >= 145) {
return 1;
}
}
}
else if (posBalance >= 40) {
if (tcor < 94) {
if (SC < 323) {
if (Cs < 135) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (Cs >= 135) {
return 1;
}
}
else if (SC >= 323) {
return 4;
}
}
else if (tcor >= 94) {
if (posBalance < 45) {
if (MVDiff < 4) {
if (Rs < 146) {
return 4;
}
else if (Rs >= 146) {
return 1;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (posBalance >= 45) {
return 1;
}
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (Rs >= 207) {
return 4;
}
}
}
else if (MVDiff >= 15) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (AFD < 9) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
return 4;
}
}
else if (AFD >= 9) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 186) {
if (negBalance < 31) {
if (ssDCval < 60) {
if (MVDiff < 1) {
if (Cs < 291) {
return 1;
}
else if (Cs >= 291) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 53) {
if (Rs < 233) {
return 1;
}
else if (Rs >= 233) {
return 4;
}
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 243) {
return 4;
}
else if (Rs >= 243) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 160) {
return 4;
}
else if (Rs >= 160) {
return 1;
}
}
}
}
else if (ssDCval >= 60) {
if (mu_mv_mag_sq < 93) {
if (ssDCval < 64) {
if (Rs < 261) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
return 1;
}
}
else if (Rs >= 261) {
return 2;
}
}
else if (ssDCval >= 64) {
if (Cs < 288) {
return 1;
}
else if (Cs >= 288) {
if (MVDiff < 2) {
if (Rs < 183) {
return 4;
}
else if (Rs >= 183) {
return 1;
}
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (Rs < 170) {
if (posBalance < 46) {
if (Rs < 127) {
if (Rs < 112) {
return 2;
}
else if (Rs >= 112) {
if (Cs < 274) {
return 4;
}
else if (Cs >= 274) {
return 2;
}
}
}
else if (Rs >= 127) {
return 4;
}
}
else if (posBalance >= 46) {
if (Rs < 143) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (negBalance < 20) {
return 4;
}
else if (negBalance >= 20) {
return 1;
}
}
}
else if (Rs >= 143) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
return 1;
}
}
}
}
else if (Rs >= 170) {
if (ssDCval < 69) {
if (Cs < 285) {
if (MVDiff < 1) {
if (Rs < 277) {
return 1;
}
else if (Rs >= 277) {
return 4;
}
}
else if (MVDiff >= 1) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
if (posBalance < 50) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 243) {
return 1;
}
else if (Rs >= 243) {
return 4;
}
}
}
else if (posBalance >= 50) {
return 1;
}
}
}
}
else if (Cs >= 285) {
if (Rs < 176) {
if (Cs < 333) {
return 4;
}
else if (Cs >= 333) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (Rs < 172) {
return 1;
}
else if (Rs >= 172) {
return 4;
}
}
}
}
else if (Rs >= 176) {
return 4;
}
}
}
else if (ssDCval >= 69) {
if (Cs < 348) {
return 4;
}
else if (Cs >= 348) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 31) {
if (tcor < 97) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (posBalance < 51) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
else if (posBalance >= 51) {
if (AFD < 4) {
if (SC < 424) {
return 1;
}
else if (SC >= 424) {
return 4;
}
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (tcor >= 97) {
if (ssDCval < 63) {
if (ssDCval < 61) {
if (mu_mv_mag_sq < 90) {
if (Rs < 192) {
if (Rs < 114) {
return 2;
}
else if (Rs >= 114) {
return 1;
}
}
else if (Rs >= 192) {
return 4;
}
}
else if (mu_mv_mag_sq >= 90) {
return 1;
}
}
else if (ssDCval >= 61) {
if (Rs < 264) {
return 4;
}
else if (Rs >= 264) {
return 1;
}
}
}
else if (ssDCval >= 63) {
if (MVDiff < 1) {
if (Cs < 348) {
return 4;
}
else if (Cs >= 348) {
return 1;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (TSC < 4) {
if (MVDiff < 3) {
if (Rs < 224) {
if (mu_mv_mag_sq < 23) {
return 4;
}
else if (mu_mv_mag_sq >= 23) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
if (Rs < 183) {
if (Cs < 338) {
if (posBalance < 53) {
if (Cs < 213) {
return 1;
}
else if (Cs >= 213) {
return 4;
}
}
else if (posBalance >= 53) {
return 1;
}
}
else if (Cs >= 338) {
return 1;
}
}
else if (Rs >= 183) {
if (Cs < 385) {
return 4;
}
else if (Cs >= 385) {
if (Cs < 413) {
return 1;
}
else if (Cs >= 413) {
return 4;
}
}
}
}
}
}
else if (Rs >= 224) {
if (diffMVdiffVal < -1) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 37) {
return 1;
}
else if (Rs >= 37) {
if (gchDC < 1) {
if (Cs < 177) {
if (posBalance < 51) {
if (MVDiff < 16) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
else if (MVDiff >= 16) {
return 4;
}
}
else if (posBalance >= 51) {
if (Cs < 49) {
return 2;
}
else if (Cs >= 49) {
return 4;
}
}
}
else if (Cs >= 177) {
if (Rs < 266) {
if (ssDCval < 54) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (Rs < 122) {
return 2;
}
else if (Rs >= 122) {
if (mu_mv_mag_sq < 77) {
return 4;
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 161) {
return 4;
}
else if (Rs >= 161) {
return 1;
}
}
}
}
}
else if (ssDCval >= 54) {
if (posBalance < 59) {
if (Cs < 236) {
if (SC < 468) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 1;
}
}
else if (SC >= 468) {
return 1;
}
}
else if (Cs >= 236) {
return 4;
}
}
else if (posBalance >= 59) {
if (Rs < 221) {
return 1;
}
else if (Rs >= 221) {
return 4;
}
}
}
}
else if (Rs >= 266) {
if (ssDCval < 66) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
if (Cs < 276) {
if (Rs < 280) {
return 1;
}
else if (Rs >= 280) {
return 4;
}
}
else if (Cs >= 276) {
return 1;
}
}
}
else if (ssDCval >= 66) {
return 4;
}
}
}
}
else if (gchDC >= 1) {
if (mcTcor < 90) {
if (Rs < 176) {
return 2;
}
else if (Rs >= 176) {
return 8;
}
}
else if (mcTcor >= 90) {
if (Rs < 271) {
if (mcTcor < 99) {
if (posBalance < 69) {
if (diffAFD < 3) {
if (MVDiff < 6) {
if (diffTSC < -1) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (diffTSC >= -1) {
if (Rs < 261) {
return 4;
}
else if (Rs >= 261) {
return 1;
}
}
}
else if (MVDiff >= 6) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
if (negBalance < 58) {
if (mcTcor < 97) {
return 4;
}
else if (mcTcor >= 97) {
if (refDCval < 63) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
return 4;
}
}
else if (refDCval >= 63) {
if (TSC < 3) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (TSC >= 3) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
}
else if (negBalance >= 58) {
return 1;
}
}
}
}
else if (diffAFD >= 3) {
return 2;
}
}
else if (posBalance >= 69) {
if (Rs < 154) {
return 1;
}
else if (Rs >= 154) {
return 2;
}
}
}
else if (mcTcor >= 99) {
if (mu_mv_mag_sq < 27) {
if (Cs < 222) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (diffRsCsdiff < 1) {
return 4;
}
else if (diffRsCsdiff >= 1) {
return 1;
}
}
}
else if (Cs >= 222) {
return 1;
}
}
else if (mu_mv_mag_sq >= 27) {
if (ssDCval < 65) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (ssDCval >= 65) {
return 8;
}
}
}
}
else if (Rs >= 271) {
if (Rs < 291) {
if (posBalance < 56) {
if (tcor < 92) {
if (Cs < 167) {
return 4;
}
else if (Cs >= 167) {
return 2;
}
}
else if (tcor >= 92) {
return 4;
}
}
else if (posBalance >= 56) {
return 2;
}
}
else if (Rs >= 291) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 1;
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (tcor < 77) {
if (refDCval < 64) {
if (diffAFD < 6) {
if (diffAFD < 2) {
if (ssDCval < 58) {
if (mu_mv_mag_sq < 22) {
if (Cs < 169) {
return 1;
}
else if (Cs >= 169) {
return 2;
}
}
else if (mu_mv_mag_sq >= 22) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (ssDCval >= 58) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
return 2;
}
}
}
else if (diffAFD >= 2) {
if (negBalance < 23) {
if (Rs < 137) {
return 8;
}
else if (Rs >= 137) {
return 2;
}
}
else if (negBalance >= 23) {
if (Cs < 186) {
if (diffRsCsdiff < 4) {
return 4;
}
else if (diffRsCsdiff >= 4) {
return 1;
}
}
else if (Cs >= 186) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (SC < 576) {
return 2;
}
else if (SC >= 576) {
return 4;
}
}
}
}
}
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (refDCval >= 64) {
if (mcTcor < 74) {
return 4;
}
else if (mcTcor >= 74) {
if (mu_mv_mag_sq < 15) {
if (posBalance < 57) {
if (Cs < 194) {
return 2;
}
else if (Cs >= 194) {
return 1;
}
}
else if (posBalance >= 57) {
return 2;
}
}
else if (mu_mv_mag_sq >= 15) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (Rs < 130) {
return 1;
}
else if (Rs >= 130) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 77) {
if (MVDiff < 60) {
if (mu_mv_mag_sq < 24) {
if (mcTcor < 97) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (posBalance < 61) {
if (Rs < 253) {
return 8;
}
else if (Rs >= 253) {
return 1;
}
}
else if (posBalance >= 61) {
if (Rs < 138) {
return 1;
}
else if (Rs >= 138) {
return 2;
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 250) {
return 2;
}
else if (Rs >= 250) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (SC < 529) {
if (AFD < 8) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (AFD >= 8) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 8;
}
}
}
else if (SC >= 529) {
if (Rs < 239) {
if (Rs < 185) {
return 2;
}
else if (Rs >= 185) {
return 8;
}
}
else if (Rs >= 239) {
return 1;
}
}
}
}
else if (MVDiff >= 60) {
if (mcTcor < 97) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (diffRsCsdiff < -6) {
if (Rs < 155) {
return 2;
}
else if (Rs >= 155) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 104) {
if (negBalance < 22) {
if (ssDCval < 81) {
if (SC < 494) {
if (Cs < 106) {
if (Rs < 174) {
return 1;
}
else if (Rs >= 174) {
return 8;
}
}
else if (Cs >= 106) {
if (Rs < 65) {
return 1;
}
else if (Rs >= 65) {
if (Cs < 262) {
return 2;
}
else if (Cs >= 262) {
if (Rs < 105) {
return 2;
}
else if (Rs >= 105) {
return 1;
}
}
}
}
}
else if (SC >= 494) {
return 1;
}
}
else if (ssDCval >= 81) {
if (Rs < 170) {
if (MVDiff < 34) {
if (ssDCval < 141) {
if (MVDiff < 5) {
if (Rs < 122) {
return 1;
}
else if (Rs >= 122) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (ssDCval >= 141) {
return 2;
}
}
else if (MVDiff >= 34) {
return 1;
}
}
else if (Rs >= 170) {
if (Cs < 142) {
if (Rs < 216) {
return 8;
}
else if (Rs >= 216) {
return 1;
}
}
else if (Cs >= 142) {
if (Cs < 305) {
if (MVDiff < 2) {
if (Rs < 257) {
return 4;
}
else if (Rs >= 257) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Rs < 270) {
return 2;
}
else if (Rs >= 270) {
if (Rs < 287) {
return 1;
}
else if (Rs >= 287) {
return 2;
}
}
}
}
else if (Cs >= 305) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 22) {
if (ssDCval < 113) {
if (mcTcor < 97) {
if (mcTcor < 72) {
if (gchDC < 6) {
if (refDCval < 80) {
return 8;
}
else if (refDCval >= 80) {
return 4;
}
}
else if (gchDC >= 6) {
if (tcor < 22) {
return 1;
}
else if (tcor >= 22) {
if (Rs < 21) {
return 1;
}
else if (Rs >= 21) {
return 4;
}
}
}
}
else if (mcTcor >= 72) {
if (negBalance < 44) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (TSC < 6) {
if (Cs < 155) {
if (diffMVdiffVal < -17) {
if (Rs < 181) {
return 1;
}
else if (Rs >= 181) {
return 8;
}
}
else if (diffMVdiffVal >= -17) {
return 1;
}
}
else if (Cs >= 155) {
if (Rs < 90) {
if (Rs < 45) {
return 2;
}
else if (Rs >= 45) {
return 4;
}
}
else if (Rs >= 90) {
if (Rs < 289) {
return 1;
}
else if (Rs >= 289) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (MVDiff < 62) {
return 8;
}
else if (MVDiff >= 62) {
if (diffAFD < -1) {
if (Rs < 144) {
return 2;
}
else if (Rs >= 144) {
return 8;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
}
else if (negBalance >= 44) {
if (ssDCval < 91) {
if (ssDCval < 86) {
if (AFD < 10) {
if (mu_mv_mag_sq < 18) {
if (Rs < 139) {
return 4;
}
else if (Rs >= 139) {
return 1;
}
}
else if (mu_mv_mag_sq >= 18) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (AFD >= 10) {
if (ssDCval < 69) {
if (Rs < 208) {
if (ssDCval < 67) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
if (refDCval < 71) {
return 1;
}
else if (refDCval >= 71) {
return 8;
}
}
}
else if (ssDCval >= 67) {
return 2;
}
}
else if (Rs >= 208) {
return 4;
}
}
else if (ssDCval >= 69) {
if (negBalance < 49) {
if (Rs < 129) {
return 2;
}
else if (Rs >= 129) {
return 8;
}
}
else if (negBalance >= 49) {
if (refDCval < 73) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (refDCval >= 73) {
if (Cs < 41) {
if (Rs < 85) {
return 1;
}
else if (Rs >= 85) {
return 8;
}
}
else if (Cs >= 41) {
if (AFD < 64) {
if (MVDiff < 65) {
if (gchDC < 4) {
if (Rs < 96) {
return 2;
}
else if (Rs >= 96) {
return 1;
}
}
else if (gchDC >= 4) {
return 2;
}
}
else if (MVDiff >= 65) {
if (tcor < 49) {
if (Rs < 242) {
if (MVDiff < 97) {
return 2;
}
else if (MVDiff >= 97) {
return 1;
}
}
else if (Rs >= 242) {
return 2;
}
}
else if (tcor >= 49) {
return 2;
}
}
}
else if (AFD >= 64) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (TSC < 6) {
if (Cs < 361) {
if (tcor < 79) {
if (Cs < 317) {
if (tcor < 67) {
if (ssDCval < 87) {
if (Cs < 120) {
return 2;
}
else if (Cs >= 120) {
return 8;
}
}
else if (ssDCval >= 87) {
if (Rs < 103) {
return 8;
}
else if (Rs >= 103) {
if (Rs < 117) {
return 2;
}
else if (Rs >= 117) {
if (Rs < 134) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 119) {
return 8;
}
else if (Rs >= 119) {
return 2;
}
}
}
else if (Rs >= 134) {
if (ssDCval < 88) {
return 2;
}
else if (ssDCval >= 88) {
if (Rs < 136) {
return 2;
}
else if (Rs >= 136) {
if (tcor < 57) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (Cs < 12) {
return 2;
}
else if (Cs >= 12) {
if (Rs < 140) {
return 8;
}
else if (Rs >= 140) {
if (refDCval < 89) {
return 2;
}
else if (refDCval >= 89) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 57) {
return 2;
}
}
}
}
}
}
}
}
else if (tcor >= 67) {
if (tcor < 75) {
return 2;
}
else if (tcor >= 75) {
if (Rs < 198) {
return 2;
}
else if (Rs >= 198) {
return 8;
}
}
}
}
else if (Cs >= 317) {
return 8;
}
}
else if (tcor >= 79) {
if (SC < 586) {
return 2;
}
else if (SC >= 586) {
if (Cs < 352) {
return 8;
}
else if (Cs >= 352) {
return 2;
}
}
}
}
else if (Cs >= 361) {
return 8;
}
}
else if (TSC >= 6) {
if (Cs < 187) {
return 1;
}
else if (Cs >= 187) {
return 2;
}
}
}
}
else if (ssDCval >= 91) {
if (MVDiff < 22) {
if (mu_mv_mag_sq < 3) {
if (refDCval < 95) {
if (Cs < 365) {
return 2;
}
else if (Cs >= 365) {
return 1;
}
}
else if (refDCval >= 95) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (MVDiff >= 22) {
if (Cs < 397) {
if (mcTcor < 94) {
if (Rs < 286) {
return 1;
}
else if (Rs >= 286) {
return 2;
}
}
else if (mcTcor >= 94) {
if (negBalance < 46) {
return 1;
}
else if (negBalance >= 46) {
if (diffMVdiffVal < 5) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (Cs < 234) {
return 2;
}
else if (Cs >= 234) {
return 1;
}
}
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
}
}
else if (Cs >= 397) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 231) {
if (Cs < 84) {
if (posBalance < 58) {
if (MVDiff < 5) {
if (Rs < 134) {
return 1;
}
else if (Rs >= 134) {
return 8;
}
}
else if (MVDiff >= 5) {
if (ssDCval < 108) {
return 1;
}
else if (ssDCval >= 108) {
return 2;
}
}
}
else if (posBalance >= 58) {
if (AFD < 10) {
if (refDCval < 95) {
if (MVDiff < 125) {
return 1;
}
else if (MVDiff >= 125) {
return 4;
}
}
else if (refDCval >= 95) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
}
else if (AFD >= 10) {
return 2;
}
}
}
else if (Cs >= 84) {
if (Rs < 87) {
if (TSC < 2) {
if (MVDiff < 16) {
if (ssDCval < 85) {
if (Rs < 73) {
return 1;
}
else if (Rs >= 73) {
return 2;
}
}
else if (ssDCval >= 85) {
return 4;
}
}
else if (MVDiff >= 16) {
return 2;
}
}
else if (TSC >= 2) {
if (Rs < 80) {
if (tcor < 98) {
if (gchDC < 13) {
return 1;
}
else if (gchDC >= 13) {
return 8;
}
}
else if (tcor >= 98) {
if (Cs < 98) {
return 1;
}
else if (Cs >= 98) {
return 2;
}
}
}
else if (Rs >= 80) {
if (mu_mv_mag_sq < 18) {
if (Cs < 140) {
return 4;
}
else if (Cs >= 140) {
return 1;
}
}
else if (mu_mv_mag_sq >= 18) {
return 1;
}
}
}
}
else if (Rs >= 87) {
if (ssDCval < 106) {
if (AFD < 6) {
if (TSCindex < 3) {
if (ssDCval < 75) {
if (mu_mv_mag_sq < 62) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 140) {
return 1;
}
else if (Rs >= 140) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 62) {
if (Rs < 241) {
return 2;
}
else if (Rs >= 241) {
return 1;
}
}
}
else if (ssDCval >= 75) {
if (SC < 310) {
if (Cs < 98) {
if (Rs < 112) {
return 1;
}
else if (Rs >= 112) {
return 2;
}
}
else if (Cs >= 98) {
return 1;
}
}
else if (SC >= 310) {
if (Cs < 146) {
return 1;
}
else if (Cs >= 146) {
if (mcTcor < 99) {
if (SC < 318) {
return 1;
}
else if (SC >= 318) {
return 4;
}
}
else if (mcTcor >= 99) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
return 2;
}
}
}
}
}
}
else if (TSCindex >= 3) {
if (MVDiff < 4) {
if (negBalance < 29) {
if (negBalance < 27) {
if (Cs < 175) {
if (Rs < 262) {
return 8;
}
else if (Rs >= 262) {
return 4;
}
}
else if (Cs >= 175) {
return 1;
}
}
else if (negBalance >= 27) {
if (Rs < 199) {
return 2;
}
else if (Rs >= 199) {
return 8;
}
}
}
else if (negBalance >= 29) {
if (TSC < 3) {
if (Cs < 224) {
return 1;
}
else if (Cs >= 224) {
return 2;
}
}
else if (TSC >= 3) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
return 2;
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 267) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (Rs < 169) {
return 4;
}
else if (Rs >= 169) {
return 2;
}
}
}
}
else if (Rs >= 267) {
if (Rs < 280) {
return 4;
}
else if (Rs >= 280) {
if (Rs < 290) {
return 1;
}
else if (Rs >= 290) {
return 8;
}
}
}
}
}
}
else if (AFD >= 6) {
if (mu_mv_mag_sq < 43) {
if (tcor < 93) {
if (posBalance < 62) {
if (RsCsDiff < 7) {
if (posBalance < 54) {
if (ssDCval < 92) {
if (Cs < 181) {
return 1;
}
else if (Cs >= 181) {
return 2;
}
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (posBalance >= 54) {
if (diffAFD < 1) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 1;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (RsCsDiff >= 7) {
return 4;
}
}
else if (posBalance >= 62) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
if (Rs < 162) {
return 1;
}
else if (Rs >= 162) {
return 2;
}
}
}
}
else if (tcor >= 93) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (negBalance < 31) {
if (Rs < 154) {
return 1;
}
else if (Rs >= 154) {
return 4;
}
}
else if (negBalance >= 31) {
if (diffRsCsdiff < -1) {
return 8;
}
else if (diffRsCsdiff >= -1) {
return 2;
}
}
}
}
}
else if (ssDCval >= 106) {
if (Rs < 193) {
return 2;
}
else if (Rs >= 193) {
return 8;
}
}
}
}
}
else if (Cs >= 231) {
if (ssDCval < 76) {
if (Rs < 128) {
if (ssDCval < 72) {
if (Cs < 247) {
if (Rs < 118) {
return 4;
}
else if (Rs >= 118) {
return 1;
}
}
else if (Cs >= 247) {
return 4;
}
}
else if (ssDCval >= 72) {
if (Cs < 334) {
return 1;
}
else if (Cs >= 334) {
return 4;
}
}
}
else if (Rs >= 128) {
if (negBalance < 42) {
if (Rs < 142) {
if (tcor < 97) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (tcor >= 97) {
return 1;
}
}
else if (Rs >= 142) {
return 1;
}
}
else if (negBalance >= 42) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
else if (ssDCval >= 76) {
if (TSC < 3) {
if (mu_mv_mag_sq < 95) {
if (SC < 577) {
if (refDCval < 98) {
return 4;
}
else if (refDCval >= 98) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 41) {
return 2;
}
else if (Rs >= 41) {
return 1;
}
}
}
}
else if (SC >= 577) {
return 1;
}
}
else if (mu_mv_mag_sq >= 95) {
if (Cs < 269) {
return 2;
}
else if (Cs >= 269) {
if (Rs < 150) {
return 1;
}
else if (Rs >= 150) {
return 4;
}
}
}
}
else if (TSC >= 3) {
if (tcor < 91) {
if (tcor < 75) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (tcor >= 75) {
if (mu_mv_mag_sq < 21) {
if (Rs < 277) {
if (Cs < 309) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
else if (Cs >= 309) {
return 2;
}
}
else if (Rs >= 277) {
return 1;
}
}
else if (mu_mv_mag_sq >= 21) {
return 1;
}
}
}
else if (tcor >= 91) {
if (tcor < 93) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (Rs < 154) {
return 2;
}
else if (Rs >= 154) {
return 8;
}
}
}
else if (tcor >= 93) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (MVDiff < 74) {
if (AFD < 3) {
if (ssDCval < 136) {
if (diffMVdiffVal < -6) {
if (Rs < 146) {
if (Rs < 72) {
return 4;
}
else if (Rs >= 72) {
return 2;
}
}
else if (Rs >= 146) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
return 4;
}
}
else if (ssDCval >= 136) {
return 2;
}
}
else if (AFD >= 3) {
if (Rs < 164) {
if (Rs < 1) {
if (Cs < 1) {
return 8;
}
else if (Cs >= 1) {
return 1;
}
}
else if (Rs >= 1) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
else if (Rs >= 164) {
if (mu_mv_mag_sq < 70) {
if (refDCval < 144) {
if (AFD < 7) {
if (Rs < 180) {
return 2;
}
else if (Rs >= 180) {
return 8;
}
}
else if (AFD >= 7) {
if (tcor < 77) {
if (Cs < 255) {
return 4;
}
else if (Cs >= 255) {
return 8;
}
}
else if (tcor >= 77) {
if (refDCval < 117) {
return 8;
}
else if (refDCval >= 117) {
return 2;
}
}
}
}
else if (refDCval >= 144) {
if (diffMVdiffVal < 13) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
else if (diffMVdiffVal >= 13) {
if (Rs < 221) {
return 2;
}
else if (Rs >= 221) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 70) {
return 2;
}
}
}
}
else if (MVDiff >= 74) {
if (tcor < 97) {
if (Rs < 6) {
if (ssDCval < 216) {
if (diffRsCsdiff < -1) {
return 2;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (ssDCval >= 216) {
return 2;
}
}
else if (Rs >= 6) {
if (RsCsDiff < 3) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 13) {
return 8;
}
else if (Cs >= 13) {
return 2;
}
}
}
else if (RsCsDiff >= 3) {
if (diffAFD < -16) {
return 2;
}
else if (diffAFD >= -16) {
if (Cs < 26) {
return 8;
}
else if (Cs >= 26) {
if (tcor < 77) {
if (Rs < 146) {
return 8;
}
else if (Rs >= 146) {
return 1;
}
}
else if (tcor >= 77) {
return 1;
}
}
}
}
}
}
else if (tcor >= 97) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 130) {
if (refDCval < 74) {
return 4;
}
else if (refDCval >= 74) {
if (Rs < 122) {
if (ssDCval < 102) {
if (Rs < 80) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (Rs >= 80) {
return 1;
}
}
else if (ssDCval >= 102) {
if (Rs < 32) {
return 1;
}
else if (Rs >= 32) {
return 8;
}
}
}
else if (Rs >= 122) {
if (Cs < 240) {
return 8;
}
else if (Cs >= 240) {
return 4;
}
}
}
}
else if (Rs >= 130) {
if (negBalance < 24) {
if (ssDCval < 81) {
if (Cs < 127) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (Cs >= 127) {
if (Cs < 265) {
if (negBalance < 21) {
return 2;
}
else if (negBalance >= 21) {
if (Cs < 209) {
return 2;
}
else if (Cs >= 209) {
if (SC < 415) {
if (Rs < 184) {
return 2;
}
else if (Rs >= 184) {
return 1;
}
}
else if (SC >= 415) {
return 2;
}
}
}
}
else if (Cs >= 265) {
return 1;
}
}
}
else if (ssDCval >= 81) {
if (mu_mv_mag_sq < 121) {
if (Cs < 279) {
if (Rs < 199) {
return 2;
}
else if (Rs >= 199) {
return 8;
}
}
else if (Cs >= 279) {
return 4;
}
}
else if (mu_mv_mag_sq >= 121) {
return 1;
}
}
}
else if (negBalance >= 24) {
if (SC < 378) {
if (mu_mv_mag_sq < 110) {
if (Cs < 142) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
else if (Cs >= 142) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 110) {
if (SC < 299) {
if (tcor < 97) {
if (Cs < 45) {
return 8;
}
else if (Cs >= 45) {
return 2;
}
}
else if (tcor >= 97) {
return 8;
}
}
else if (SC >= 299) {
return 2;
}
}
}
else if (SC >= 378) {
if (ssDCval < 78) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (ssDCval >= 78) {
if (Cs < 286) {
return 2;
}
else if (Cs >= 286) {
if (Rs < 229) {
return 4;
}
else if (Rs >= 229) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 300) {
if (RsDiff < 4) {
if (refDCval < 74) {
if (mcTcor < 90) {
if (Cs < 155) {
if (mu_mv_mag_sq < 54) {
if (diffAFD < 5) {
if (Rs < 697) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
if (Rs < 444) {
return 8;
}
else if (Rs >= 444) {
return 1;
}
}
}
else if (Rs >= 697) {
return 2;
}
}
else if (diffAFD >= 5) {
if (Rs < 405) {
return 2;
}
else if (Rs >= 405) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (negBalance < 30) {
return 2;
}
else if (negBalance >= 30) {
return 8;
}
}
}
else if (Cs >= 155) {
if (negBalance < 35) {
if (refDCval < 41) {
if (tcor < 74) {
if (posBalance < 68) {
return 1;
}
else if (posBalance >= 68) {
return 8;
}
}
else if (tcor >= 74) {
if (negBalance < 33) {
if (Rs < 582) {
return 1;
}
else if (Rs >= 582) {
return 2;
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
else if (refDCval >= 41) {
if (posBalance < 48) {
if (refDCval < 60) {
if (TSC < 4) {
if (Cs < 269) {
return 4;
}
else if (Cs >= 269) {
return 1;
}
}
else if (TSC >= 4) {
return 2;
}
}
else if (refDCval >= 60) {
return 8;
}
}
else if (posBalance >= 48) {
if (diffTSC < 2) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (Rs < 403) {
return 4;
}
else if (Rs >= 403) {
return 8;
}
}
}
else if (diffTSC >= 2) {
return 1;
}
}
}
}
else if (negBalance >= 35) {
if (AFD < 8) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (ssDCval < 58) {
if (Cs < 529) {
if (Rs < 921) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 359) {
return 2;
}
else if (Rs >= 359) {
return 8;
}
}
}
}
else if (Rs >= 921) {
if (Rs < 967) {
return 1;
}
else if (Rs >= 967) {
return 8;
}
}
}
else if (Cs >= 529) {
return 4;
}
}
else if (ssDCval >= 58) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Cs < 263) {
return 2;
}
else if (Cs >= 263) {
return 1;
}
}
}
}
}
else if (AFD >= 8) {
if (MVDiff < 60) {
if (refDCval < 54) {
if (tcor < 58) {
if (Cs < 271) {
if (tcor < 52) {
return 8;
}
else if (tcor >= 52) {
return 4;
}
}
else if (Cs >= 271) {
return 4;
}
}
else if (tcor >= 58) {
if (Cs < 262) {
if (SC < 540) {
return 4;
}
else if (SC >= 540) {
if (Rs < 453) {
if (Rs < 367) {
return 8;
}
else if (Rs >= 367) {
return 2;
}
}
else if (Rs >= 453) {
return 1;
}
}
}
else if (Cs >= 262) {
if (diffRsCsdiff < -11) {
if (Rs < 578) {
return 2;
}
else if (Rs >= 578) {
return 4;
}
}
else if (diffRsCsdiff >= -11) {
if (MVDiff < 52) {
return 8;
}
else if (MVDiff >= 52) {
return 4;
}
}
}
}
}
else if (refDCval >= 54) {
if (diffRsCsdiff < 6) {
return 4;
}
else if (diffRsCsdiff >= 6) {
if (Cs < 491) {
return 8;
}
else if (Cs >= 491) {
return 1;
}
}
}
}
else if (MVDiff >= 60) {
if (gchDC < 9) {
if (Cs < 398) {
if (SC < 587) {
return 8;
}
else if (SC >= 587) {
return 2;
}
}
else if (Cs >= 398) {
if (diffAFD < 1) {
if (TSC < 7) {
if (Rs < 563) {
return 2;
}
else if (Rs >= 563) {
return 8;
}
}
else if (TSC >= 7) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
else if (gchDC >= 9) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (Rs < 1280) {
if (tcor < 101) {
if (TSC < 1) {
if (mcTcor < 99) {
return 4;
}
else if (mcTcor >= 99) {
if (Rs < 1017) {
if (Cs < 124) {
return 4;
}
else if (Cs >= 124) {
if (Rs < 794) {
return 8;
}
else if (Rs >= 794) {
if (Cs < 358) {
return 1;
}
else if (Cs >= 358) {
return 8;
}
}
}
}
else if (Rs >= 1017) {
return 1;
}
}
}
else if (TSC >= 1) {
if (MVDiff < 2) {
if (ssDCval < 68) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 114) {
if (posBalance < 54) {
if (diffRsCsdiff < 2) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (Cs < 527) {
return 1;
}
else if (Cs >= 527) {
return 8;
}
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 906) {
if (Cs < 334) {
return 2;
}
else if (Cs >= 334) {
return 8;
}
}
else if (Rs >= 906) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (Cs < 303) {
return 4;
}
else if (Cs >= 303) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (mcTcor >= 97) {
if (Rs < 480) {
if (TSC < 4) {
if (Cs < 271) {
return 4;
}
else if (Cs >= 271) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
return 2;
}
}
}
else if (TSC >= 4) {
if (Rs < 371) {
return 8;
}
else if (Rs >= 371) {
return 4;
}
}
}
else if (Rs >= 480) {
if (Rs < 909) {
if (Rs < 845) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (negBalance < 40) {
if (posBalance < 46) {
if (Cs < 289) {
return 1;
}
else if (Cs >= 289) {
if (ssDCval < 53) {
if (Cs < 311) {
return 8;
}
else if (Cs >= 311) {
return 1;
}
}
else if (ssDCval >= 53) {
if (Rs < 558) {
if (SC < 947) {
if (ssDCval < 63) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (ssDCval >= 63) {
return 8;
}
}
else if (SC >= 947) {
return 8;
}
}
else if (Rs >= 558) {
if (ssDCval < 59) {
if (Rs < 691) {
return 8;
}
else if (Rs >= 691) {
return 1;
}
}
else if (ssDCval >= 59) {
if (SC < 1200) {
return 1;
}
else if (SC >= 1200) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 46) {
if (Rs < 490) {
return 2;
}
else if (Rs >= 490) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
return 1;
}
}
}
}
else if (negBalance >= 40) {
if (Cs < 517) {
return 8;
}
else if (Cs >= 517) {
return 4;
}
}
}
}
else if (Rs >= 845) {
if (Cs < 325) {
if (Cs < 207) {
return 1;
}
else if (Cs >= 207) {
return 2;
}
}
else if (Cs >= 325) {
if (Cs < 449) {
return 1;
}
else if (Cs >= 449) {
if (Rs < 876) {
return 4;
}
else if (Rs >= 876) {
return 8;
}
}
}
}
}
else if (Rs >= 909) {
if (Rs < 1271) {
return 1;
}
else if (Rs >= 1271) {
if (Rs < 1275) {
return 8;
}
else if (Rs >= 1275) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 68) {
if (Rs < 771) {
if (refDCval < 69) {
if (MVDiff < 1) {
if (mu_mv_mag_sq < 117) {
if (Rs < 492) {
if (Rs < 443) {
return 2;
}
else if (Rs >= 443) {
return 1;
}
}
else if (Rs >= 492) {
return 8;
}
}
else if (mu_mv_mag_sq >= 117) {
return 1;
}
}
else if (MVDiff >= 1) {
if (SC < 902) {
if (Rs < 452) {
return 2;
}
else if (Rs >= 452) {
if (Rs < 485) {
return 1;
}
else if (Rs >= 485) {
return 8;
}
}
}
else if (SC >= 902) {
return 2;
}
}
}
else if (refDCval >= 69) {
if (Rs < 606) {
if (SC < 897) {
if (RsDiff < 1) {
if (SC < 889) {
return 2;
}
else if (SC >= 889) {
if (Rs < 521) {
return 1;
}
else if (Rs >= 521) {
return 2;
}
}
}
else if (RsDiff >= 1) {
if (posBalance < 53) {
if (Rs < 377) {
return 4;
}
else if (Rs >= 377) {
return 1;
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
else if (SC >= 897) {
if (Cs < 377) {
return 2;
}
else if (Cs >= 377) {
if (RsDiff < 2) {
if (Cs < 404) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 4;
}
}
else if (Cs >= 404) {
return 1;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (Rs >= 606) {
if (diffMVdiffVal < -1) {
if (tcor < 99) {
if (Rs < 684) {
if (Rs < 676) {
return 4;
}
else if (Rs >= 676) {
return 8;
}
}
else if (Rs >= 684) {
return 4;
}
}
else if (tcor >= 99) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (SC < 1062) {
return 4;
}
else if (SC >= 1062) {
if (ssDCval < 71) {
if (Rs < 646) {
return 1;
}
else if (Rs >= 646) {
return 2;
}
}
else if (ssDCval >= 71) {
return 4;
}
}
}
}
}
}
else if (Rs >= 771) {
if (Cs < 280) {
if (Rs < 1007) {
return 4;
}
else if (Rs >= 1007) {
return 1;
}
}
else if (Cs >= 280) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
if (Cs < 507) {
return 1;
}
else if (Cs >= 507) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 64) {
if (Cs < 139) {
if (mu_mv_mag_sq < 49) {
if (Cs < 85) {
if (tcor < 79) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (tcor >= 79) {
return 1;
}
}
else if (Cs >= 85) {
if (refDCval < 69) {
if (SC < 522) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (SC >= 522) {
if (SC < 1032) {
if (tcor < 83) {
if (Rs < 521) {
return 1;
}
else if (Rs >= 521) {
if (Rs < 627) {
return 8;
}
else if (Rs >= 627) {
return 1;
}
}
}
else if (tcor >= 83) {
return 1;
}
}
else if (SC >= 1032) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 1;
}
}
}
}
else if (refDCval >= 69) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (Rs < 524) {
return 8;
}
else if (Rs >= 524) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (mcTcor < 97) {
if (ssDCval < 37) {
return 1;
}
else if (ssDCval >= 37) {
if (Cs < 15) {
return 1;
}
else if (Cs >= 15) {
return 8;
}
}
}
else if (mcTcor >= 97) {
return 1;
}
}
}
else if (Cs >= 139) {
if (Rs < 743) {
if (RsCsDiff < 7) {
if (posBalance < 62) {
if (negBalance < 42) {
if (posBalance < 34) {
if (tcor < 96) {
return 2;
}
else if (tcor >= 96) {
return 4;
}
}
else if (posBalance >= 34) {
if (SC < 1063) {
if (TSC < 4) {
if (AFD < 4) {
if (SC < 835) {
if (Cs < 164) {
if (Rs < 574) {
return 1;
}
else if (Rs >= 574) {
return 4;
}
}
else if (Cs >= 164) {
return 4;
}
}
else if (SC >= 835) {
if (diffAFD < -1) {
if (Rs < 436) {
return 4;
}
else if (Rs >= 436) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (AFD >= 4) {
if (RsCsDiff < 2) {
if (tcor < 93) {
if (Cs < 238) {
return 4;
}
else if (Cs >= 238) {
return 1;
}
}
else if (tcor >= 93) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (RsCsDiff >= 2) {
if (negBalance < 34) {
if (Rs < 588) {
if (Cs < 480) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (Rs < 307) {
return 2;
}
else if (Rs >= 307) {
if (ssDCval < 37) {
return 1;
}
else if (ssDCval >= 37) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
if (ssDCval < 71) {
if (Rs < 548) {
return 4;
}
else if (Rs >= 548) {
if (Rs < 560) {
return 1;
}
else if (Rs >= 560) {
return 4;
}
}
}
else if (ssDCval >= 71) {
return 1;
}
}
}
}
}
}
else if (Cs >= 480) {
return 1;
}
}
else if (Rs >= 588) {
return 1;
}
}
else if (negBalance >= 34) {
if (ssDCval < 50) {
if (SC < 981) {
if (Rs < 398) {
if (Rs < 369) {
return 1;
}
else if (Rs >= 369) {
return 4;
}
}
else if (Rs >= 398) {
return 1;
}
}
else if (SC >= 981) {
return 4;
}
}
else if (ssDCval >= 50) {
if (Cs < 199) {
return 1;
}
else if (Cs >= 199) {
if (Rs < 341) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
return 1;
}
}
else if (Rs >= 341) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (Cs < 261) {
if (Cs < 204) {
return 4;
}
else if (Cs >= 204) {
return 8;
}
}
else if (Cs >= 261) {
if (negBalance < 40) {
if (refDCval < 43) {
if (Rs < 437) {
return 1;
}
else if (Rs >= 437) {
return 4;
}
}
else if (refDCval >= 43) {
return 1;
}
}
else if (negBalance >= 40) {
if (AFD < 8) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 2;
}
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
else if (SC >= 1063) {
if (Cs < 440) {
return 2;
}
else if (Cs >= 440) {
return 4;
}
}
}
}
else if (negBalance >= 42) {
if (refDCval < 62) {
if (refDCval < 59) {
if (MVDiff < 8) {
if (Rs < 475) {
return 4;
}
else if (Rs >= 475) {
return 8;
}
}
else if (MVDiff >= 8) {
if (MVDiff < 35) {
if (Rs < 372) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
if (Rs < 309) {
return 8;
}
else if (Rs >= 309) {
return 1;
}
}
}
else if (Rs >= 372) {
return 1;
}
}
else if (MVDiff >= 35) {
if (diffAFD < -2) {
if (Rs < 366) {
return 1;
}
else if (Rs >= 366) {
return 4;
}
}
else if (diffAFD >= -2) {
return 2;
}
}
}
}
else if (refDCval >= 59) {
return 2;
}
}
else if (refDCval >= 62) {
if (mcTcor < 98) {
if (Rs < 342) {
return 1;
}
else if (Rs >= 342) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 1;
}
}
}
else if (mcTcor >= 98) {
return 1;
}
}
}
}
else if (posBalance >= 62) {
if (Cs < 384) {
if (diffTSC < 1) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (MVDiff < 21) {
return 2;
}
else if (MVDiff >= 21) {
return 1;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 416) {
return 2;
}
else if (Rs >= 416) {
return 8;
}
}
}
else if (Cs >= 384) {
return 2;
}
}
}
else if (RsCsDiff >= 7) {
if (refDCval < 58) {
if (Cs < 436) {
if (Rs < 465) {
if (diffMVdiffVal < -9) {
if (Rs < 442) {
return 8;
}
else if (Rs >= 442) {
return 2;
}
}
else if (diffMVdiffVal >= -9) {
if (ssDCval < 46) {
if (Cs < 271) {
return 2;
}
else if (Cs >= 271) {
return 1;
}
}
else if (ssDCval >= 46) {
if (refDCval < 52) {
return 4;
}
else if (refDCval >= 52) {
if (Cs < 383) {
return 1;
}
else if (Cs >= 383) {
return 2;
}
}
}
}
}
else if (Rs >= 465) {
if (mcTcor < 91) {
if (Rs < 535) {
if (Cs < 420) {
return 1;
}
else if (Cs >= 420) {
return 2;
}
}
else if (Rs >= 535) {
return 8;
}
}
else if (mcTcor >= 91) {
if (diffMVdiffVal < 12) {
if (posBalance < 34) {
return 2;
}
else if (posBalance >= 34) {
if (mu_mv_mag_sq < 59) {
return 1;
}
else if (mu_mv_mag_sq >= 59) {
return 4;
}
}
}
else if (diffMVdiffVal >= 12) {
return 4;
}
}
}
}
else if (Cs >= 436) {
if (negBalance < 43) {
if (MVDiff < 17) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 4;
}
}
else if (MVDiff >= 17) {
return 1;
}
}
else if (negBalance >= 43) {
if (tcor < 77) {
return 1;
}
else if (tcor >= 77) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 1;
}
}
}
}
}
else if (refDCval >= 58) {
if (ssDCval < 70) {
if (MVDiff < 10) {
if (mcTcor < 98) {
if (Cs < 501) {
return 2;
}
else if (Cs >= 501) {
return 4;
}
}
else if (mcTcor >= 98) {
return 4;
}
}
else if (MVDiff >= 10) {
if (mcTcor < 96) {
if (Rs < 567) {
if (Rs < 430) {
if (tcor < 80) {
return 1;
}
else if (tcor >= 80) {
return 2;
}
}
else if (Rs >= 430) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (MVDiff < 54) {
return 4;
}
else if (MVDiff >= 54) {
return 2;
}
}
}
}
else if (Rs >= 567) {
if (Cs < 328) {
return 1;
}
else if (Cs >= 328) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (Cs < 414) {
return 4;
}
else if (Cs >= 414) {
return 1;
}
}
}
}
else if (ssDCval >= 70) {
if (Rs < 529) {
if (Rs < 361) {
if (Rs < 324) {
if (Rs < 312) {
return 1;
}
else if (Rs >= 312) {
return 8;
}
}
else if (Rs >= 324) {
return 4;
}
}
else if (Rs >= 361) {
return 8;
}
}
else if (Rs >= 529) {
return 4;
}
}
}
}
}
else if (Rs >= 743) {
if (ssDCval < 58) {
if (mcTcor < 99) {
if (posBalance < 58) {
if (Cs < 466) {
if (refDCval < 59) {
return 1;
}
else if (refDCval >= 59) {
return 8;
}
}
else if (Cs >= 466) {
if (Rs < 1079) {
return 2;
}
else if (Rs >= 1079) {
return 1;
}
}
}
else if (posBalance >= 58) {
if (Cs < 513) {
return 8;
}
else if (Cs >= 513) {
return 4;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 762) {
return 1;
}
else if (Rs >= 762) {
return 2;
}
}
}
else if (ssDCval >= 58) {
if (mu_mv_mag_sq < 46) {
if (tcor < 97) {
if (mcTcor < 96) {
if (negBalance < 37) {
return 1;
}
else if (negBalance >= 37) {
if (Rs < 1109) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
else if (Rs >= 1109) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 43) {
return 1;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 1001) {
return 1;
}
else if (Rs >= 1001) {
return 4;
}
}
}
}
else if (tcor >= 97) {
if (Cs < 162) {
return 1;
}
else if (Cs >= 162) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (RsDiff < 3) {
if (negBalance < 18) {
return 2;
}
else if (negBalance >= 18) {
return 1;
}
}
else if (RsDiff >= 3) {
if (Rs < 829) {
return 1;
}
else if (Rs >= 829) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (negBalance < 33) {
if (tcor < 99) {
if (ssDCval < 50) {
if (refDCval < 43) {
if (posBalance < 35) {
if (Rs < 530) {
if (ssDCval < 38) {
return 4;
}
else if (ssDCval >= 38) {
return 1;
}
}
else if (Rs >= 530) {
return 2;
}
}
else if (posBalance >= 35) {
if (AFD < 3) {
if (RsCsDiff < 1) {
return 8;
}
else if (RsCsDiff >= 1) {
if (Rs < 638) {
return 1;
}
else if (Rs >= 638) {
return 2;
}
}
}
else if (AFD >= 3) {
if (Cs < 515) {
return 1;
}
else if (Cs >= 515) {
return 2;
}
}
}
}
else if (refDCval >= 43) {
if (mu_mv_mag_sq < 80) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
return 4;
}
}
else if (mu_mv_mag_sq >= 80) {
if (AFD < 4) {
if (SC < 784) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (SC >= 784) {
if (ssDCval < 46) {
return 4;
}
else if (ssDCval >= 46) {
return 1;
}
}
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (ssDCval >= 50) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 105) {
if (Rs < 518) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (Cs < 479) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (Cs >= 479) {
return 8;
}
}
}
}
else if (Rs >= 518) {
if (Cs < 500) {
if (negBalance < 17) {
if (Rs < 955) {
if (Rs < 827) {
return 4;
}
else if (Rs >= 827) {
return 8;
}
}
else if (Rs >= 955) {
return 1;
}
}
else if (negBalance >= 17) {
if (SC < 888) {
return 2;
}
else if (SC >= 888) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (Cs >= 500) {
if (Rs < 930) {
return 2;
}
else if (Rs >= 930) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (tcor < 94) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
else if (tcor >= 94) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 41) {
if (Rs < 674) {
return 8;
}
else if (Rs >= 674) {
return 1;
}
}
else if (posBalance >= 41) {
if (ssDCval < 56) {
if (SC < 814) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 4;
}
}
else if (SC >= 814) {
if (Cs < 495) {
if (MVDiff < 15) {
if (mu_mv_mag_sq < 93) {
return 8;
}
else if (mu_mv_mag_sq >= 93) {
return 1;
}
}
else if (MVDiff >= 15) {
return 1;
}
}
else if (Cs >= 495) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
else if (ssDCval >= 56) {
if (negBalance < 23) {
if (SC < 977) {
if (Cs < 265) {
if (SC < 867) {
return 4;
}
else if (SC >= 867) {
return 1;
}
}
else if (Cs >= 265) {
if (tcor < 98) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (tcor >= 98) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
return 8;
}
}
}
}
}
else if (SC >= 977) {
if (Cs < 379) {
return 1;
}
else if (Cs >= 379) {
return 2;
}
}
}
else if (negBalance >= 23) {
if (RsCsDiff < 4) {
if (tcor < 98) {
if (Rs < 414) {
if (mu_mv_mag_sq < 97) {
if (Rs < 323) {
return 1;
}
else if (Rs >= 323) {
return 8;
}
}
else if (mu_mv_mag_sq >= 97) {
return 1;
}
}
else if (Rs >= 414) {
return 1;
}
}
else if (tcor >= 98) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 89) {
if (Rs < 567) {
return 8;
}
else if (Rs >= 567) {
return 1;
}
}
else if (mu_mv_mag_sq >= 89) {
if (Rs < 312) {
return 4;
}
else if (Rs >= 312) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 4) {
if (mu_mv_mag_sq < 71) {
return 1;
}
else if (mu_mv_mag_sq >= 71) {
if (SC < 1124) {
return 2;
}
else if (SC >= 1124) {
return 4;
}
}
}
}
}
}
}
}
}
else if (tcor >= 99) {
if (mu_mv_mag_sq < 112) {
if (SC < 679) {
if (ssDCval < 62) {
if (Rs < 338) {
return 2;
}
else if (Rs >= 338) {
return 1;
}
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (SC >= 679) {
if (MVDiff < 4) {
if (SC < 951) {
return 8;
}
else if (SC >= 951) {
if (mu_mv_mag_sq < 105) {
if (Rs < 949) {
return 2;
}
else if (Rs >= 949) {
return 8;
}
}
else if (mu_mv_mag_sq >= 105) {
return 1;
}
}
}
else if (MVDiff >= 4) {
if (SC < 777) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (SC >= 777) {
if (ssDCval < 66) {
if (AFD < 2) {
if (Rs < 722) {
return 8;
}
else if (Rs >= 722) {
return 4;
}
}
else if (AFD >= 2) {
return 4;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (negBalance < 20) {
if (Rs < 648) {
return 4;
}
else if (Rs >= 648) {
if (Rs < 665) {
return 1;
}
else if (Rs >= 665) {
return 8;
}
}
}
else if (negBalance >= 20) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 2;
}
}
}
}
}
else if (negBalance >= 33) {
if (Rs < 546) {
if (MVDiff < 5) {
if (tcor < 94) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (tcor >= 94) {
return 1;
}
}
else if (MVDiff >= 5) {
if (ssDCval < 44) {
if (diffRsCsdiff < -3) {
return 1;
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
else if (ssDCval >= 44) {
if (posBalance < 40) {
if (Rs < 431) {
return 4;
}
else if (Rs >= 431) {
return 8;
}
}
else if (posBalance >= 40) {
if (Rs < 379) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (MVDiff < 10) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
}
else if (Rs >= 379) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Cs < 476) {
if (Rs < 515) {
if (Cs < 124) {
return 1;
}
else if (Cs >= 124) {
return 2;
}
}
else if (Rs >= 515) {
return 1;
}
}
else if (Cs >= 476) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 546) {
if (RsDiff < 2) {
if (posBalance < 43) {
if (Rs < 797) {
return 4;
}
else if (Rs >= 797) {
if (Cs < 529) {
return 1;
}
else if (Cs >= 529) {
return 4;
}
}
}
else if (posBalance >= 43) {
if (Rs < 635) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (Rs < 611) {
return 4;
}
else if (Rs >= 611) {
return 1;
}
}
}
else if (Rs >= 635) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (diffRsCsdiff < -1) {
if (diffTSC < -1) {
if (SC < 1091) {
return 1;
}
else if (SC >= 1091) {
return 4;
}
}
else if (diffTSC >= -1) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
return 1;
}
}
}
else if (diffRsCsdiff >= -1) {
if (mu_mv_mag_sq < 67) {
if (Cs < 379) {
return 1;
}
else if (Cs >= 379) {
return 8;
}
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 761) {
return 4;
}
else if (Rs >= 761) {
if (Rs < 898) {
return 8;
}
else if (Rs >= 898) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (tcor >= 101) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
if (posBalance < 32) {
if (Cs < 13) {
return 4;
}
else if (Cs >= 13) {
if (Rs < 408) {
if (Rs < 395) {
if (Rs < 328) {
return 8;
}
else if (Rs >= 328) {
if (posBalance < 13) {
return 8;
}
else if (posBalance >= 13) {
return 1;
}
}
}
else if (Rs >= 395) {
return 4;
}
}
else if (Rs >= 408) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
if (Cs < 476) {
return 8;
}
else if (Cs >= 476) {
if (Cs < 482) {
return 4;
}
else if (Cs >= 482) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 32) {
if (Rs < 585) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (Rs < 511) {
return 8;
}
else if (Rs >= 511) {
return 1;
}
}
}
else if (Rs >= 585) {
if (Rs < 730) {
return 4;
}
else if (Rs >= 730) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1280) {
if (SC < 1625) {
if (mu_mv_mag_sq < 59) {
if (ssDCval < 51) {
return 2;
}
else if (ssDCval >= 51) {
return 1;
}
}
else if (mu_mv_mag_sq >= 59) {
if (Cs < 223) {
return 8;
}
else if (Cs >= 223) {
if (Cs < 288) {
return 1;
}
else if (Cs >= 288) {
return 8;
}
}
}
}
else if (SC >= 1625) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1696) {
if (ssDCval < 64) {
if (refDCval < 56) {
if (posBalance < 70) {
return 1;
}
else if (posBalance >= 70) {
return 2;
}
}
else if (refDCval >= 56) {
if (Cs < 420) {
if (Rs < 1621) {
return 1;
}
else if (Rs >= 1621) {
return 2;
}
}
else if (Cs >= 420) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (negBalance < 25) {
if (Rs < 1479) {
return 2;
}
else if (Rs >= 1479) {
if (Rs < 1534) {
return 1;
}
else if (Rs >= 1534) {
return 2;
}
}
}
else if (negBalance >= 25) {
return 8;
}
}
}
}
}
else if (ssDCval >= 64) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 59) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 35) {
return 1;
}
else if (mu_mv_mag_sq >= 35) {
if (Cs < 493) {
return 2;
}
else if (Cs >= 493) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (ssDCval >= 69) {
if (AFD < 4) {
if (Cs < 427) {
return 8;
}
else if (Cs >= 427) {
if (Cs < 514) {
return 1;
}
else if (Cs >= 514) {
return 8;
}
}
}
else if (AFD >= 4) {
if (Rs < 1306) {
return 4;
}
else if (Rs >= 1306) {
return 1;
}
}
}
}
}
else if (Rs >= 1696) {
if (Cs < 434) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 434) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (Rs < 1408) {
return 8;
}
else if (Rs >= 1408) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 74) {
if (Cs < 283) {
if (refDCval < 107) {
if (mu_mv_mag_sq < 114) {
if (SC < 1286) {
if (ssDCval < 102) {
if (AFD < 5) {
if (Cs < 185) {
if (Cs < 3) {
return 1;
}
else if (Cs >= 3) {
return 8;
}
}
else if (Cs >= 185) {
if (ssDCval < 80) {
if (RsCsDiff < 2) {
if (Rs < 363) {
return 1;
}
else if (Rs >= 363) {
return 8;
}
}
else if (RsCsDiff >= 2) {
if (Cs < 210) {
return 8;
}
else if (Cs >= 210) {
return 4;
}
}
}
else if (ssDCval >= 80) {
if (Rs < 590) {
return 2;
}
else if (Rs >= 590) {
if (mu_mv_mag_sq < 39) {
return 4;
}
else if (mu_mv_mag_sq >= 39) {
return 1;
}
}
}
}
}
else if (AFD >= 5) {
if (Cs < 85) {
if (tcor < 62) {
if (Rs < 326) {
return 8;
}
else if (Rs >= 326) {
return 2;
}
}
else if (tcor >= 62) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 2;
}
}
}
else if (Cs >= 85) {
if (Rs < 470) {
if (MVDiff < 5) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
return 1;
}
}
else if (MVDiff >= 5) {
if (SC < 577) {
if (Rs < 351) {
if (Cs < 87) {
return 2;
}
else if (Cs >= 87) {
return 8;
}
}
else if (Rs >= 351) {
if (ssDCval < 91) {
if (Cs < 106) {
return 8;
}
else if (Cs >= 106) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (ssDCval >= 91) {
if (SC < 466) {
return 8;
}
else if (SC >= 466) {
return 2;
}
}
}
}
else if (SC >= 577) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (TSC < 7) {
if (SC < 649) {
return 1;
}
else if (SC >= 649) {
return 4;
}
}
else if (TSC >= 7) {
if (Cs < 266) {
return 2;
}
else if (Cs >= 266) {
return 8;
}
}
}
}
}
}
else if (Rs >= 470) {
if (Cs < 152) {
return 1;
}
else if (Cs >= 152) {
if (posBalance < 61) {
if (Rs < 799) {
if (ssDCval < 88) {
if (diffRsCsdiff < -4) {
if (Rs < 658) {
return 1;
}
else if (Rs >= 658) {
return 4;
}
}
else if (diffRsCsdiff >= -4) {
return 4;
}
}
else if (ssDCval >= 88) {
return 1;
}
}
else if (Rs >= 799) {
if (SC < 1244) {
return 1;
}
else if (SC >= 1244) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (posBalance < 69) {
return 1;
}
else if (posBalance >= 69) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (Cs < 182) {
return 1;
}
else if (Cs >= 182) {
if (Rs < 531) {
if (Rs < 379) {
return 4;
}
else if (Rs >= 379) {
return 8;
}
}
else if (Rs >= 531) {
return 1;
}
}
}
}
else if (SC >= 1286) {
if (refDCval < 102) {
if (diffMVdiffVal < -153) {
return 8;
}
else if (diffMVdiffVal >= -153) {
if (Rs < 1098) {
if (Rs < 1088) {
return 1;
}
else if (Rs >= 1088) {
return 8;
}
}
else if (Rs >= 1098) {
return 1;
}
}
}
else if (refDCval >= 102) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (Rs < 1942) {
return 1;
}
else if (Rs >= 1942) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (ssDCval < 79) {
if (Rs < 1108) {
return 8;
}
else if (Rs >= 1108) {
return 1;
}
}
else if (ssDCval >= 79) {
if (ssDCval < 94) {
if (SC < 509) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (SC >= 509) {
if (Cs < 251) {
if (Rs < 313) {
if (Rs < 311) {
return 2;
}
else if (Rs >= 311) {
return 1;
}
}
else if (Rs >= 313) {
return 2;
}
}
else if (Cs >= 251) {
if (refDCval < 82) {
return 2;
}
else if (refDCval >= 82) {
return 1;
}
}
}
}
else if (ssDCval >= 94) {
if (Rs < 391) {
return 4;
}
else if (Rs >= 391) {
return 1;
}
}
}
}
}
else if (refDCval >= 107) {
if (RsCsDiff < 7) {
if (diffMVdiffVal < 22) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
return 1;
}
}
else if (diffMVdiffVal >= 22) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (Cs < 121) {
return 8;
}
else if (Cs >= 121) {
return 4;
}
}
}
}
else if (RsCsDiff >= 7) {
if (Rs < 446) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Cs < 65) {
return 1;
}
else if (Cs >= 65) {
if (SC < 658) {
return 2;
}
else if (SC >= 658) {
return 1;
}
}
}
}
else if (Rs >= 446) {
if (AFD < 7) {
if (Rs < 932) {
return 8;
}
else if (Rs >= 932) {
return 4;
}
}
else if (AFD >= 7) {
if (SC < 515) {
return 8;
}
else if (SC >= 515) {
return 1;
}
}
}
}
}
}
else if (Cs >= 283) {
if (refDCval < 114) {
if (Rs < 1205) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 13) {
if (Cs < 460) {
return 1;
}
else if (Cs >= 460) {
if (RsCsDiff < 7) {
return 1;
}
else if (RsCsDiff >= 7) {
if (AFD < 9) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (AFD >= 9) {
if (Cs < 465) {
return 8;
}
else if (Cs >= 465) {
return 1;
}
}
}
}
}
else if (MVDiff >= 13) {
if (negBalance < 46) {
return 1;
}
else if (negBalance >= 46) {
if (TSC < 9) {
if (AFD < 7) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
return 2;
}
}
else if (AFD >= 7) {
return 8;
}
}
else if (TSC >= 9) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 582) {
if (mu_mv_mag_sq < 74) {
if (TSC < 4) {
if (TSC < 3) {
if (Rs < 484) {
return 4;
}
else if (Rs >= 484) {
return 8;
}
}
else if (TSC >= 3) {
if (posBalance < 52) {
if (Rs < 552) {
return 1;
}
else if (Rs >= 552) {
return 2;
}
}
else if (posBalance >= 52) {
return 2;
}
}
}
else if (TSC >= 4) {
if (AFD < 10) {
if (SC < 1079) {
if (mcTcor < 84) {
if (Rs < 333) {
return 2;
}
else if (Rs >= 333) {
return 8;
}
}
else if (mcTcor >= 84) {
if (ssDCval < 81) {
if (Rs < 343) {
return 2;
}
else if (Rs >= 343) {
return 4;
}
}
else if (ssDCval >= 81) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 8;
}
}
}
}
else if (SC >= 1079) {
return 1;
}
}
else if (AFD >= 10) {
if (TSC < 6) {
if (gchDC < 2) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 390) {
return 8;
}
else if (Rs >= 390) {
return 4;
}
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (TSC >= 6) {
if (MVDiff < 35) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 8;
}
}
else if (MVDiff >= 35) {
if (AFD < 23) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
if (Rs < 534) {
return 2;
}
else if (Rs >= 534) {
return 8;
}
}
}
}
else if (AFD >= 23) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 74) {
if (refDCval < 94) {
if (Rs < 436) {
return 8;
}
else if (Rs >= 436) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 1;
}
}
}
}
else if (refDCval >= 94) {
if (Rs < 481) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (Rs >= 481) {
return 8;
}
}
}
}
else if (Rs >= 582) {
if (posBalance < 48) {
if (SC < 1093) {
if (ssDCval < 90) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 90) {
if (Cs < 465) {
return 4;
}
else if (Cs >= 465) {
return 8;
}
}
}
else if (SC >= 1093) {
if (Rs < 1004) {
if (refDCval < 75) {
return 4;
}
else if (refDCval >= 75) {
if (ssDCval < 81) {
if (Rs < 703) {
return 1;
}
else if (Rs >= 703) {
return 8;
}
}
else if (ssDCval >= 81) {
if (SC < 1100) {
if (Rs < 630) {
return 8;
}
else if (Rs >= 630) {
return 4;
}
}
else if (SC >= 1100) {
return 8;
}
}
}
}
else if (Rs >= 1004) {
if (Rs < 1134) {
return 2;
}
else if (Rs >= 1134) {
if (Rs < 1168) {
return 1;
}
else if (Rs >= 1168) {
return 8;
}
}
}
}
}
else if (posBalance >= 48) {
if (negBalance < 54) {
if (Cs < 440) {
if (diffRsCsdiff < 6) {
if (CsDiff < 2) {
if (negBalance < 32) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (negBalance >= 32) {
return 8;
}
}
else if (CsDiff >= 2) {
if (Rs < 670) {
return 1;
}
else if (Rs >= 670) {
return 8;
}
}
}
else if (diffRsCsdiff >= 6) {
return 1;
}
}
else if (Cs >= 440) {
if (MVDiff < 3) {
if (tcor < 89) {
return 2;
}
else if (tcor >= 89) {
if (SC < 1154) {
return 1;
}
else if (SC >= 1154) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (refDCval < 107) {
return 8;
}
else if (refDCval >= 107) {
return 1;
}
}
}
}
else if (negBalance >= 54) {
if (mcTcor < 92) {
if (Rs < 677) {
return 2;
}
else if (Rs >= 677) {
return 8;
}
}
else if (mcTcor >= 92) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1205) {
if (ssDCval < 96) {
if (mu_mv_mag_sq < 101) {
if (tcor < 97) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (Rs < 1345) {
return 8;
}
else if (Rs >= 1345) {
if (Rs < 2031) {
return 1;
}
else if (Rs >= 2031) {
return 2;
}
}
}
}
else if (tcor >= 97) {
if (RsCsDiff < 5) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (Cs < 339) {
return 1;
}
else if (Cs >= 339) {
return 8;
}
}
}
else if (RsCsDiff >= 5) {
if (posBalance < 59) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 69) {
if (tcor < 98) {
if (Rs < 1549) {
return 8;
}
else if (Rs >= 1549) {
return 1;
}
}
else if (tcor >= 98) {
if (Rs < 2055) {
return 8;
}
else if (Rs >= 2055) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 69) {
return 1;
}
}
}
else if (posBalance >= 59) {
if (Rs < 2004) {
return 1;
}
else if (Rs >= 2004) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (Cs < 497) {
if (Rs < 2078) {
return 1;
}
else if (Rs >= 2078) {
if (Rs < 2140) {
return 8;
}
else if (Rs >= 2140) {
return 2;
}
}
}
else if (Cs >= 497) {
return 4;
}
}
}
else if (ssDCval >= 96) {
if (posBalance < 30) {
return 1;
}
else if (posBalance >= 30) {
return 8;
}
}
}
}
else if (refDCval >= 114) {
if (mcTcor < 99) {
if (ssDCval < 160) {
if (MVDiff < 3) {
if (Rs < 1470) {
if (Rs < 329) {
return 2;
}
else if (Rs >= 329) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
else if (Rs >= 1470) {
if (Rs < 1839) {
return 4;
}
else if (Rs >= 1839) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (posBalance < 61) {
if (Rs < 605) {
if (ssDCval < 130) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (MVDiff < 109) {
if (mcTcor < 96) {
if (Rs < 373) {
return 2;
}
else if (Rs >= 373) {
return 8;
}
}
else if (mcTcor >= 96) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (TSC < 4) {
if (Rs < 390) {
return 8;
}
else if (Rs >= 390) {
return 1;
}
}
else if (TSC >= 4) {
return 2;
}
}
}
}
else if (MVDiff >= 109) {
if (Cs < 438) {
return 4;
}
else if (Cs >= 438) {
if (Rs < 459) {
return 2;
}
else if (Rs >= 459) {
return 1;
}
}
}
}
}
else if (ssDCval >= 130) {
if (posBalance < 56) {
if (Rs < 519) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
else if (Rs >= 519) {
return 1;
}
}
else if (posBalance >= 56) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (Rs < 316) {
return 2;
}
else if (Rs >= 316) {
return 8;
}
}
}
}
}
else if (Rs >= 605) {
if (negBalance < 48) {
if (ssDCval < 135) {
if (AFD < 6) {
if (Rs < 1097) {
return 8;
}
else if (Rs >= 1097) {
return 2;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (ssDCval >= 135) {
return 4;
}
}
else if (negBalance >= 48) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
return 2;
}
}
}
}
else if (posBalance >= 61) {
if (posBalance < 78) {
if (mu_mv_mag_sq < 11) {
if (ssDCval < 139) {
return 4;
}
else if (ssDCval >= 139) {
return 2;
}
}
else if (mu_mv_mag_sq >= 11) {
if (Rs < 717) {
return 8;
}
else if (Rs >= 717) {
return 1;
}
}
}
else if (posBalance >= 78) {
return 2;
}
}
}
}
else if (ssDCval >= 160) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (negBalance < 39) {
if (Rs < 377) {
return 2;
}
else if (Rs >= 377) {
return 8;
}
}
else if (negBalance >= 39) {
return 1;
}
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 122) {
if (SC < 990) {
if (posBalance < 62) {
if (negBalance < 46) {
if (gchDC < 1) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (negBalance >= 46) {
if (Rs < 451) {
return 2;
}
else if (Rs >= 451) {
if (Rs < 504) {
return 1;
}
else if (Rs >= 504) {
return 8;
}
}
}
}
else if (posBalance >= 62) {
if (Rs < 408) {
return 4;
}
else if (Rs >= 408) {
return 2;
}
}
}
else if (SC >= 990) {
return 2;
}
}
else if (ssDCval >= 122) {
if (ssDCval < 155) {
if (mu_mv_mag_sq < 116) {
return 1;
}
else if (mu_mv_mag_sq >= 116) {
if (Rs < 707) {
return 8;
}
else if (Rs >= 707) {
return 1;
}
}
}
else if (ssDCval >= 155) {
return 4;
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (posBalance < 53) {
if (MVDiff < 47) {
if (RsCsDiff < 111) {
if (Rs < 679) {
if (AFD < 13) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
if (Rs < 541) {
return 1;
}
else if (Rs >= 541) {
return 2;
}
}
}
else if (AFD >= 13) {
if (refDCval < 44) {
return 1;
}
else if (refDCval >= 44) {
return 2;
}
}
}
else if (Rs >= 679) {
if (negBalance < 48) {
if (Cs < 512) {
if (RsCsDiff < 28) {
if (Rs < 1078) {
return 1;
}
else if (Rs >= 1078) {
return 2;
}
}
else if (RsCsDiff >= 28) {
if (posBalance < 45) {
if (Rs < 1492) {
return 1;
}
else if (Rs >= 1492) {
return 2;
}
}
else if (posBalance >= 45) {
return 1;
}
}
}
else if (Cs >= 512) {
return 2;
}
}
else if (negBalance >= 48) {
if (MVDiff < 30) {
if (Rs < 1142) {
return 8;
}
else if (Rs >= 1142) {
if (Cs < 129) {
return 1;
}
else if (Cs >= 129) {
return 2;
}
}
}
else if (MVDiff >= 30) {
if (Rs < 969) {
return 8;
}
else if (Rs >= 969) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 111) {
if (diffAFD < 4) {
if (tcor < 53) {
return 1;
}
else if (tcor >= 53) {
if (diffMVdiffVal < -4) {
if (SC < 2448) {
if (Rs < 1637) {
return 1;
}
else if (Rs >= 1637) {
return 2;
}
}
else if (SC >= 2448) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
}
else if (diffAFD >= 4) {
if (Rs < 1334) {
return 2;
}
else if (Rs >= 1334) {
return 1;
}
}
}
}
else if (MVDiff >= 47) {
if (Cs < 214) {
if (RsDiff < 26) {
if (ssDCval < 34) {
return 8;
}
else if (ssDCval >= 34) {
return 1;
}
}
else if (RsDiff >= 26) {
return 2;
}
}
else if (Cs >= 214) {
if (Rs < 910) {
if (negBalance < 70) {
if (ssDCval < 49) {
if (MVDiff < 61) {
return 2;
}
else if (MVDiff >= 61) {
return 1;
}
}
else if (ssDCval >= 49) {
if (MVDiff < 64) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (MVDiff >= 64) {
if (RsCsDiff < 90) {
if (AFD < 47) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
return 1;
}
}
else if (AFD >= 47) {
return 1;
}
}
else if (RsCsDiff >= 90) {
return 4;
}
}
}
}
else if (negBalance >= 70) {
if (diffMVdiffVal < 39) {
if (SC < 1151) {
return 2;
}
else if (SC >= 1151) {
return 1;
}
}
else if (diffMVdiffVal >= 39) {
if (diffAFD < 15) {
return 1;
}
else if (diffAFD >= 15) {
return 4;
}
}
}
}
else if (Rs >= 910) {
if (gchDC < 2) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 1;
}
}
else if (gchDC >= 2) {
if (diffAFD < -7) {
return 2;
}
else if (diffAFD >= -7) {
if (mu_mv_mag_sq < 7) {
if (diffMVdiffVal < 30) {
return 1;
}
else if (diffMVdiffVal >= 30) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
return 1;
}
}
}
}
}
}
}
else if (posBalance >= 53) {
if (diffAFD < 3) {
if (Cs < 392) {
if (tcor < 44) {
if (AFD < 25) {
if (TSC < 7) {
if (ssDCval < 115) {
if (tcor < 39) {
return 4;
}
else if (tcor >= 39) {
if (Rs < 411) {
return 1;
}
else if (Rs >= 411) {
return 2;
}
}
}
else if (ssDCval >= 115) {
return 1;
}
}
else if (TSC >= 7) {
if (MVDiff < 120) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (MVDiff < 67) {
if (Rs < 658) {
return 1;
}
else if (Rs >= 658) {
return 4;
}
}
else if (MVDiff >= 67) {
return 8;
}
}
}
else if (MVDiff >= 120) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 4;
}
}
}
}
else if (AFD >= 25) {
if (Cs < 198) {
return 2;
}
else if (Cs >= 198) {
if (Rs < 1204) {
if (Cs < 333) {
if (Rs < 579) {
return 8;
}
else if (Rs >= 579) {
return 1;
}
}
else if (Cs >= 333) {
return 2;
}
}
else if (Rs >= 1204) {
return 4;
}
}
}
}
else if (tcor >= 44) {
if (mu_mv_mag_sq < 43) {
if (ssDCval < 76) {
if (negBalance < 53) {
if (mcTcor < 86) {
return 1;
}
else if (mcTcor >= 86) {
if (AFD < 15) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 1;
}
}
else if (AFD >= 15) {
if (diffMVdiffVal < -20) {
return 2;
}
else if (diffMVdiffVal >= -20) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
if (Cs < 160) {
return 1;
}
else if (Cs >= 160) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 53) {
if (AFD < 15) {
if (Rs < 590) {
return 8;
}
else if (Rs >= 590) {
return 2;
}
}
else if (AFD >= 15) {
if (MVDiff < 69) {
return 4;
}
else if (MVDiff >= 69) {
return 2;
}
}
}
}
else if (ssDCval >= 76) {
if (ssDCval < 132) {
if (tcor < 68) {
if (mcTcor < 92) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
if (CsDiff < 2) {
if (diffMVdiffVal < 7) {
if (Cs < 100) {
if (Rs < 486) {
return 8;
}
else if (Rs >= 486) {
if (Rs < 641) {
return 2;
}
else if (Rs >= 641) {
return 8;
}
}
}
else if (Cs >= 100) {
return 2;
}
}
else if (diffMVdiffVal >= 7) {
return 8;
}
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 94) {
if (MVDiff < 113) {
return 2;
}
else if (MVDiff >= 113) {
return 1;
}
}
else if (ssDCval >= 94) {
if (diffMVdiffVal < -16) {
return 2;
}
else if (diffMVdiffVal >= -16) {
return 1;
}
}
}
}
else if (tcor >= 68) {
if (refDCval < 90) {
return 1;
}
else if (refDCval >= 90) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
}
}
else if (ssDCval >= 132) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
if (TSC < 12) {
return 1;
}
else if (TSC >= 12) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (refDCval < 59) {
if (MVDiff < 125) {
if (Rs < 514) {
return 8;
}
else if (Rs >= 514) {
return 1;
}
}
else if (MVDiff >= 125) {
return 2;
}
}
else if (refDCval >= 59) {
if (negBalance < 37) {
return 1;
}
else if (negBalance >= 37) {
if (diffMVdiffVal < 13) {
return 2;
}
else if (diffMVdiffVal >= 13) {
if (ssDCval < 112) {
return 2;
}
else if (ssDCval >= 112) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 392) {
if (mcTcor < 90) {
if (tcor < 61) {
if (AFD < 17) {
if (negBalance < 52) {
if (ssDCval < 63) {
if (Cs < 450) {
return 1;
}
else if (Cs >= 450) {
if (Rs < 883) {
return 8;
}
else if (Rs >= 883) {
return 1;
}
}
}
else if (ssDCval >= 63) {
return 2;
}
}
else if (negBalance >= 52) {
if (AFD < 15) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (AFD >= 15) {
if (Rs < 1207) {
if (Cs < 513) {
return 4;
}
else if (Cs >= 513) {
return 2;
}
}
else if (Rs >= 1207) {
if (Rs < 1799) {
return 2;
}
else if (Rs >= 1799) {
return 1;
}
}
}
}
}
else if (AFD >= 17) {
if (AFD < 21) {
if (Rs < 754) {
if (Rs < 526) {
return 4;
}
else if (Rs >= 526) {
return 2;
}
}
else if (Rs >= 754) {
return 8;
}
}
else if (AFD >= 21) {
if (diffRsCsdiff < -4) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
if (diffRsCsdiff < -27) {
return 2;
}
else if (diffRsCsdiff >= -27) {
return 8;
}
}
}
else if (diffRsCsdiff >= -4) {
return 4;
}
}
}
}
else if (tcor >= 61) {
if (mu_mv_mag_sq < 10) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (gchDC < 1) {
if (Rs < 1263) {
return 2;
}
else if (Rs >= 1263) {
return 8;
}
}
else if (gchDC >= 1) {
if (Rs < 612) {
return 4;
}
else if (Rs >= 612) {
return 8;
}
}
}
}
}
else if (mcTcor >= 90) {
if (CsDiff < 4) {
if (posBalance < 64) {
if (Rs < 667) {
return 1;
}
else if (Rs >= 667) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
return 8;
}
}
}
else if (posBalance >= 64) {
if (MVDiff < 36) {
return 1;
}
else if (MVDiff >= 36) {
return 2;
}
}
}
else if (CsDiff >= 4) {
if (RsDiff < 11) {
if (tcor < 69) {
if (Cs < 497) {
if (Rs < 365) {
if (Cs < 449) {
return 2;
}
else if (Cs >= 449) {
return 8;
}
}
else if (Rs >= 365) {
if (CsDiff < 6) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (CsDiff >= 6) {
return 2;
}
}
}
else if (Cs >= 497) {
if (Cs < 525) {
return 4;
}
else if (Cs >= 525) {
return 1;
}
}
}
else if (tcor >= 69) {
if (Cs < 434) {
return 4;
}
else if (Cs >= 434) {
return 8;
}
}
}
else if (RsDiff >= 11) {
if (Rs < 1726) {
if (MVDiff < 84) {
return 2;
}
else if (MVDiff >= 84) {
return 4;
}
}
else if (Rs >= 1726) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (MVDiff < 93) {
return 1;
}
else if (MVDiff >= 93) {
if (Rs < 1836) {
return 8;
}
else if (Rs >= 1836) {
return 2;
}
}
}
}
}
}
}
}
}
else if (diffAFD >= 3) {
if (RsCsDiff < 222) {
if (Rs < 883) {
if (negBalance < 33) {
return 1;
}
else if (negBalance >= 33) {
if (diffRsCsdiff < 25) {
if (posBalance < 77) {
if (posBalance < 57) {
if (diffAFD < 5) {
return 2;
}
else if (diffAFD >= 5) {
if (Rs < 410) {
return 4;
}
else if (Rs >= 410) {
return 1;
}
}
}
else if (posBalance >= 57) {
if (diffRsCsdiff < 14) {
if (tcor < 64) {
if (Rs < 416) {
if (Cs < 366) {
return 8;
}
else if (Cs >= 366) {
return 1;
}
}
else if (Rs >= 416) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
if (diffAFD < 5) {
if (Rs < 757) {
return 8;
}
else if (Rs >= 757) {
return 2;
}
}
else if (diffAFD >= 5) {
return 2;
}
}
}
}
else if (tcor >= 64) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
return 1;
}
}
}
else if (diffRsCsdiff >= 14) {
return 1;
}
}
}
else if (posBalance >= 77) {
if (Rs < 633) {
return 2;
}
else if (Rs >= 633) {
return 4;
}
}
}
else if (diffRsCsdiff >= 25) {
if (mu_mv_mag_sq < 23) {
if (Rs < 859) {
return 1;
}
else if (Rs >= 859) {
if (Cs < 218) {
return 1;
}
else if (Cs >= 218) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 23) {
return 8;
}
}
}
}
else if (Rs >= 883) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
if (negBalance < 54) {
if (mu_mv_mag_sq < 8) {
if (ssDCval < 101) {
return 4;
}
else if (ssDCval >= 101) {
return 1;
}
}
else if (mu_mv_mag_sq >= 8) {
return 1;
}
}
else if (negBalance >= 54) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
if (tcor < 57) {
if (Rs < 1102) {
if (RsCsDiff < 59) {
return 2;
}
else if (RsCsDiff >= 59) {
return 1;
}
}
else if (Rs >= 1102) {
return 2;
}
}
else if (tcor >= 57) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 222) {
if (AFD < 31) {
return 2;
}
else if (AFD >= 31) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
return 1;
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 81) {
if (RsDiff < 9) {
if (Cs < 454) {
if (ssDCval < 139) {
if (SC < 2295) {
if (refDCval < 85) {
if (CsDiff < 2) {
if (mcTcor < 99) {
if (diffMVdiffVal < 13) {
if (posBalance < 33) {
if (Rs < 569) {
return 2;
}
else if (Rs >= 569) {
return 1;
}
}
else if (posBalance >= 33) {
if (Rs < 1981) {
if (MVDiff < 11) {
if (tcor < 83) {
if (Rs < 1148) {
return 1;
}
else if (Rs >= 1148) {
return 4;
}
}
else if (tcor >= 83) {
return 1;
}
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (Rs >= 1981) {
if (Rs < 2039) {
return 2;
}
else if (Rs >= 2039) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 13) {
if (diffAFD < -10) {
return 8;
}
else if (diffAFD >= -10) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 61) {
return 1;
}
else if (mu_mv_mag_sq >= 61) {
return 8;
}
}
else if (mcTcor >= 98) {
return 8;
}
}
}
}
else if (mcTcor >= 99) {
if (diffMVdiffVal < -3) {
if (Rs < 788) {
return 4;
}
else if (Rs >= 788) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
if (mu_mv_mag_sq < 28) {
return 1;
}
else if (mu_mv_mag_sq >= 28) {
if (Rs < 1122) {
return 1;
}
else if (Rs >= 1122) {
return 2;
}
}
}
}
}
else if (CsDiff >= 2) {
if (Rs < 826) {
if (diffRsCsdiff < 10) {
if (diffRsCsdiff < -17) {
if (tcor < 81) {
if (Rs < 623) {
return 4;
}
else if (Rs >= 623) {
return 1;
}
}
else if (tcor >= 81) {
if (Rs < 316) {
return 1;
}
else if (Rs >= 316) {
return 8;
}
}
}
else if (diffRsCsdiff >= -17) {
if (Cs < 339) {
if (Rs < 492) {
return 2;
}
else if (Rs >= 492) {
if (Cs < 314) {
return 1;
}
else if (Cs >= 314) {
if (Rs < 646) {
return 4;
}
else if (Rs >= 646) {
return 1;
}
}
}
}
else if (Cs >= 339) {
if (diffRsCsdiff < -1) {
if (Rs < 748) {
return 1;
}
else if (Rs >= 748) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
if (refDCval < 69) {
return 8;
}
else if (refDCval >= 69) {
return 1;
}
}
}
}
}
else if (diffRsCsdiff >= 10) {
return 8;
}
}
else if (Rs >= 826) {
if (posBalance < 59) {
if (posBalance < 42) {
if (mcTcor < 98) {
if (diffRsCsdiff < -52) {
return 2;
}
else if (diffRsCsdiff >= -52) {
if (Rs < 938) {
if (Rs < 919) {
return 1;
}
else if (Rs >= 919) {
return 2;
}
}
else if (Rs >= 938) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 1309) {
return 8;
}
else if (Rs >= 1309) {
return 2;
}
}
}
else if (posBalance >= 42) {
if (mcTcor < 89) {
if (Cs < 391) {
return 1;
}
else if (Cs >= 391) {
return 2;
}
}
else if (mcTcor >= 89) {
if (diffRsCsdiff < -17) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
return 1;
}
}
else if (diffRsCsdiff >= -17) {
return 1;
}
}
}
}
else if (posBalance >= 59) {
if (Cs < 417) {
return 1;
}
else if (Cs >= 417) {
return 4;
}
}
}
}
}
else if (refDCval >= 85) {
if (MVDiff < 12) {
if (Cs < 272) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
if (ssDCval < 93) {
return 1;
}
else if (ssDCval >= 93) {
return 2;
}
}
}
else if (Cs >= 272) {
if (Cs < 376) {
if (Rs < 1864) {
return 8;
}
else if (Rs >= 1864) {
return 1;
}
}
else if (Cs >= 376) {
return 1;
}
}
}
else if (MVDiff >= 12) {
if (Rs < 589) {
if (diffRsCsdiff < -35) {
return 2;
}
else if (diffRsCsdiff >= -35) {
if (MVDiff < 61) {
return 1;
}
else if (MVDiff >= 61) {
if (MVDiff < 156) {
return 8;
}
else if (MVDiff >= 156) {
if (Rs < 450) {
return 1;
}
else if (Rs >= 450) {
return 2;
}
}
}
}
}
else if (Rs >= 589) {
if (ssDCval < 125) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (MVDiff < 84) {
if (Cs < 163) {
if (diffAFD < -1) {
if (MVDiff < 21) {
return 2;
}
else if (MVDiff >= 21) {
if (diffRsCsdiff < -55) {
return 2;
}
else if (diffRsCsdiff >= -55) {
if (Rs < 922) {
return 2;
}
else if (Rs >= 922) {
return 1;
}
}
}
}
else if (diffAFD >= -1) {
if (refDCval < 119) {
if (diffRsCsdiff < -6) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 1;
}
}
else if (diffRsCsdiff >= -6) {
return 1;
}
}
else if (refDCval >= 119) {
return 2;
}
}
}
else if (Cs >= 163) {
if (tcor < 89) {
if (SC < 1541) {
if (negBalance < 60) {
if (Rs < 970) {
return 1;
}
else if (Rs >= 970) {
if (Rs < 1168) {
return 2;
}
else if (Rs >= 1168) {
return 1;
}
}
}
else if (negBalance >= 60) {
if (Cs < 446) {
return 8;
}
else if (Cs >= 446) {
return 2;
}
}
}
else if (SC >= 1541) {
if (Cs < 441) {
if (MVDiff < 62) {
if (MVDiff < 15) {
if (Rs < 1753) {
return 8;
}
else if (Rs >= 1753) {
return 1;
}
}
else if (MVDiff >= 15) {
return 1;
}
}
else if (MVDiff >= 62) {
if (Rs < 1597) {
return 1;
}
else if (Rs >= 1597) {
return 8;
}
}
}
else if (Cs >= 441) {
return 2;
}
}
}
else if (tcor >= 89) {
return 1;
}
}
}
else if (MVDiff >= 84) {
return 1;
}
}
}
else if (ssDCval >= 125) {
return 1;
}
}
}
}
}
else if (SC >= 2295) {
if (mu_mv_mag_sq < 42) {
if (ssDCval < 66) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
else if (ssDCval >= 66) {
if (SC < 2731) {
if (diffMVdiffVal < -46) {
return 8;
}
else if (diffMVdiffVal >= -46) {
return 1;
}
}
else if (SC >= 2731) {
if (Cs < 402) {
return 1;
}
else if (Cs >= 402) {
if (Rs < 2345) {
if (Rs < 2321) {
return 2;
}
else if (Rs >= 2321) {
return 4;
}
}
else if (Rs >= 2345) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (posBalance < 52) {
if (Cs < 371) {
return 1;
}
else if (Cs >= 371) {
return 8;
}
}
else if (posBalance >= 52) {
if (Rs < 2320) {
return 4;
}
else if (Rs >= 2320) {
return 1;
}
}
}
}
}
else if (ssDCval >= 139) {
if (gchDC < 2) {
if (ssDCval < 146) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (Rs < 856) {
return 4;
}
else if (Rs >= 856) {
return 1;
}
}
}
else if (ssDCval >= 146) {
if (SC < 1141) {
if (Cs < 295) {
return 2;
}
else if (Cs >= 295) {
return 8;
}
}
else if (SC >= 1141) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 172) {
if (Rs < 612) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 1;
}
}
else if (Rs >= 612) {
if (Cs < 429) {
return 1;
}
else if (Cs >= 429) {
if (Rs < 1739) {
return 8;
}
else if (Rs >= 1739) {
return 1;
}
}
}
}
else if (ssDCval >= 172) {
if (Rs < 786) {
return 8;
}
else if (Rs >= 786) {
return 2;
}
}
}
}
}
else if (Cs >= 454) {
if (Rs < 1023) {
if (diffAFD < 4) {
if (posBalance < 49) {
if (tcor < 90) {
if (MVDiff < 7) {
if (Rs < 694) {
return 8;
}
else if (Rs >= 694) {
return 2;
}
}
else if (MVDiff >= 7) {
return 2;
}
}
else if (tcor >= 90) {
if (Rs < 738) {
return 8;
}
else if (Rs >= 738) {
return 4;
}
}
}
else if (posBalance >= 49) {
if (SC < 961) {
return 2;
}
else if (SC >= 961) {
if (ssDCval < 117) {
if (MVDiff < 37) {
if (MVDiff < 19) {
if (MVDiff < 7) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
return 1;
}
}
else if (MVDiff >= 7) {
if (ssDCval < 66) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
if (Rs < 768) {
return 1;
}
else if (Rs >= 768) {
return 4;
}
}
}
else if (ssDCval >= 66) {
return 1;
}
}
}
else if (MVDiff >= 19) {
return 4;
}
}
else if (MVDiff >= 37) {
if (SC < 1048) {
return 1;
}
else if (SC >= 1048) {
return 8;
}
}
}
else if (ssDCval >= 117) {
return 8;
}
}
}
}
else if (diffAFD >= 4) {
return 1;
}
}
else if (Rs >= 1023) {
if (mcTcor < 93) {
if (ssDCval < 90) {
if (diffMVdiffVal < 4) {
if (tcor < 85) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (Rs < 1238) {
return 4;
}
else if (Rs >= 1238) {
if (Rs < 1673) {
return 2;
}
else if (Rs >= 1673) {
return 1;
}
}
}
}
else if (tcor >= 85) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (Rs < 1647) {
return 8;
}
else if (Rs >= 1647) {
return 2;
}
}
}
}
}
else if (diffMVdiffVal >= 4) {
if (MVDiff < 45) {
if (Rs < 1087) {
return 4;
}
else if (Rs >= 1087) {
return 1;
}
}
else if (MVDiff >= 45) {
return 4;
}
}
}
else if (ssDCval >= 90) {
if (Rs < 1060) {
return 2;
}
else if (Rs >= 1060) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (refDCval < 70) {
if (mcTcor < 98) {
if (Cs < 523) {
if (Rs < 2004) {
if (tcor < 82) {
return 2;
}
else if (tcor >= 82) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (diffMVdiffVal < 8) {
if (Rs < 1809) {
return 1;
}
else if (Rs >= 1809) {
if (Rs < 1947) {
return 2;
}
else if (Rs >= 1947) {
return 1;
}
}
}
else if (diffMVdiffVal >= 8) {
return 2;
}
}
}
}
else if (Rs >= 2004) {
return 2;
}
}
else if (Cs >= 523) {
return 8;
}
}
else if (mcTcor >= 98) {
if (Rs < 2184) {
return 4;
}
else if (Rs >= 2184) {
return 1;
}
}
}
else if (refDCval >= 70) {
if (refDCval < 109) {
if (ssDCval < 91) {
return 1;
}
else if (ssDCval >= 91) {
if (CsDiff < 4) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (refDCval >= 109) {
if (Rs < 1703) {
if (TSC < 7) {
if (Cs < 475) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
if (SC < 1960) {
return 1;
}
else if (SC >= 1960) {
return 2;
}
}
}
else if (Cs >= 475) {
return 2;
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (Rs >= 1703) {
if (Rs < 2319) {
return 1;
}
else if (Rs >= 2319) {
return 8;
}
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 112) {
if (MVDiff < 38) {
if (diffMVdiffVal < -20) {
if (negBalance < 44) {
if (Rs < 2055) {
return 2;
}
else if (Rs >= 2055) {
return 4;
}
}
else if (negBalance >= 44) {
return 1;
}
}
else if (diffMVdiffVal >= -20) {
if (negBalance < 41) {
if (posBalance < 59) {
if (Cs < 159) {
return 2;
}
else if (Cs >= 159) {
return 1;
}
}
else if (posBalance >= 59) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (negBalance >= 41) {
if (refDCval < 103) {
if (tcor < 91) {
if (RsCsDiff < 101) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 2;
}
}
}
else if (RsCsDiff >= 101) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
if (TSC < 8) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
return 2;
}
}
else if (TSC >= 8) {
return 2;
}
}
}
}
else if (tcor >= 91) {
if (Rs < 2236) {
return 1;
}
else if (Rs >= 2236) {
return 2;
}
}
}
else if (refDCval >= 103) {
if (AFD < 19) {
if (SC < 2337) {
return 2;
}
else if (SC >= 2337) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 2;
}
}
}
else if (AFD >= 19) {
return 1;
}
}
}
}
}
else if (MVDiff >= 38) {
if (mcTcor < 96) {
if (posBalance < 72) {
if (Cs < 287) {
if (diffMVdiffVal < -8) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 1;
}
}
else if (diffMVdiffVal >= -8) {
return 1;
}
}
else if (Cs >= 287) {
if (mu_mv_mag_sq < 7) {
if (TSC < 10) {
if (negBalance < 62) {
if (diffMVdiffVal < 13) {
return 8;
}
else if (diffMVdiffVal >= 13) {
return 1;
}
}
else if (negBalance >= 62) {
return 8;
}
}
else if (TSC >= 10) {
if (Rs < 1921) {
return 1;
}
else if (Rs >= 1921) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (MVDiff < 53) {
if (Cs < 322) {
return 1;
}
else if (Cs >= 322) {
return 8;
}
}
else if (MVDiff >= 53) {
if (negBalance < 43) {
return 4;
}
else if (negBalance >= 43) {
return 1;
}
}
}
}
}
else if (posBalance >= 72) {
return 2;
}
}
else if (mcTcor >= 96) {
if (TSC < 7) {
if (Rs < 2226) {
if (MVDiff < 41) {
return 1;
}
else if (MVDiff >= 41) {
return 2;
}
}
else if (Rs >= 2226) {
return 1;
}
}
else if (TSC >= 7) {
if (MVDiff < 69) {
return 1;
}
else if (MVDiff >= 69) {
if (ssDCval < 90) {
return 2;
}
else if (ssDCval >= 90) {
if (Rs < 1880) {
return 1;
}
else if (Rs >= 1880) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (CsDiff < 2) {
if (mcTcor < 95) {
if (Rs < 1275) {
return 4;
}
else if (Rs >= 1275) {
return 1;
}
}
else if (mcTcor >= 95) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1748) {
return 1;
}
else if (Rs >= 1748) {
return 8;
}
}
}
}
else if (CsDiff >= 2) {
if (MVDiff < 61) {
if (AFD < 34) {
if (TSC < 6) {
if (Rs < 1906) {
return 2;
}
else if (Rs >= 1906) {
return 1;
}
}
else if (TSC >= 6) {
return 1;
}
}
else if (AFD >= 34) {
if (Rs < 1461) {
return 1;
}
else if (Rs >= 1461) {
return 2;
}
}
}
else if (MVDiff >= 61) {
if (AFD < 23) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (mcTcor < 97) {
if (Cs < 417) {
if (Rs < 2150) {
return 8;
}
else if (Rs >= 2150) {
return 1;
}
}
else if (Cs >= 417) {
return 8;
}
}
else if (mcTcor >= 97) {
if (Rs < 2121) {
return 1;
}
else if (Rs >= 2121) {
return 8;
}
}
}
}
else if (AFD >= 23) {
if (AFD < 36) {
if (mu_mv_mag_sq < 66) {
return 1;
}
else if (mu_mv_mag_sq >= 66) {
return 2;
}
}
else if (AFD >= 36) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (AFD < 9) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
if (negBalance < 28) {
if (Rs < 1352) {
return 4;
}
else if (Rs >= 1352) {
if (Rs < 1915) {
return 8;
}
else if (Rs >= 1915) {
return 2;
}
}
}
else if (negBalance >= 28) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
return 4;
}
}
}
}
else if (AFD >= 9) {
if (Rs < 736) {
return 1;
}
else if (Rs >= 736) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 535) {
if (negBalance < 13) {
if (Rs < 707) {
if (ssDCval < 56) {
if (AFD < 1) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (AFD >= 1) {
if (negBalance < 6) {
return 8;
}
else if (negBalance >= 6) {
if (refDCval < 32) {
if (Rs < 570) {
return 2;
}
else if (Rs >= 570) {
return 8;
}
}
else if (refDCval >= 32) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
return 2;
}
}
}
}
}
else if (ssDCval >= 56) {
if (RsCsDiff < 1) {
if (Cs < 693) {
if (Rs < 467) {
if (Rs < 455) {
return 8;
}
else if (Rs >= 455) {
return 1;
}
}
else if (Rs >= 467) {
return 8;
}
}
else if (Cs >= 693) {
if (Rs < 659) {
if (SC < 1174) {
return 4;
}
else if (SC >= 1174) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
if (Rs < 573) {
return 8;
}
else if (Rs >= 573) {
return 2;
}
}
}
}
else if (Rs >= 659) {
if (Rs < 706) {
return 1;
}
else if (Rs >= 706) {
return 8;
}
}
}
}
else if (RsCsDiff >= 1) {
if (tcor < 95) {
if (ssDCval < 147) {
return 8;
}
else if (ssDCval >= 147) {
if (Rs < 444) {
return 1;
}
else if (Rs >= 444) {
if (Rs < 452) {
return 4;
}
else if (Rs >= 452) {
return 2;
}
}
}
}
else if (tcor >= 95) {
if (CsDiff < 1) {
if (Rs < 584) {
return 1;
}
else if (Rs >= 584) {
return 8;
}
}
else if (CsDiff >= 1) {
return 1;
}
}
}
}
}
else if (Rs >= 707) {
if (diffAFD < -3) {
if (Rs < 1607) {
if (diffRsCsdiff < -46) {
if (diffMVdiffVal < -36) {
return 4;
}
else if (diffMVdiffVal >= -36) {
if (diffMVdiffVal < 38) {
if (MVDiff < 25) {
if (Rs < 1094) {
return 1;
}
else if (Rs >= 1094) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (MVDiff >= 25) {
if (Rs < 1297) {
if (MVDiff < 29) {
return 4;
}
else if (MVDiff >= 29) {
return 8;
}
}
else if (Rs >= 1297) {
if (MVDiff < 80) {
return 4;
}
else if (MVDiff >= 80) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 38) {
return 1;
}
}
}
else if (diffRsCsdiff >= -46) {
if (tcor < 99) {
if (mcTcor < 80) {
return 1;
}
else if (mcTcor >= 80) {
return 8;
}
}
else if (tcor >= 99) {
if (Cs < 643) {
if (Rs < 1038) {
return 2;
}
else if (Rs >= 1038) {
return 1;
}
}
else if (Cs >= 643) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (Rs < 983) {
if (Rs < 863) {
return 8;
}
else if (Rs >= 863) {
if (Rs < 958) {
return 4;
}
else if (Rs >= 958) {
return 1;
}
}
}
else if (Rs >= 983) {
if (diffTSC < -6) {
return 8;
}
else if (diffTSC >= -6) {
if (Cs < 1173) {
return 4;
}
else if (Cs >= 1173) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1607) {
if (Cs < 1259) {
return 8;
}
else if (Cs >= 1259) {
if (Rs < 1736) {
return 8;
}
else if (Rs >= 1736) {
if (Rs < 1877) {
return 4;
}
else if (Rs >= 1877) {
return 1;
}
}
}
}
}
else if (diffAFD >= -3) {
if (Rs < 1841) {
if (Rs < 1053) {
if (posBalance < 12) {
if (ssDCval < 152) {
if (Rs < 806) {
return 2;
}
else if (Rs >= 806) {
return 8;
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (posBalance >= 12) {
if (Cs < 1200) {
if (TSC < 1) {
if (ssDCval < 38) {
if (Rs < 825) {
return 2;
}
else if (Rs >= 825) {
return 8;
}
}
else if (ssDCval >= 38) {
if (mu_mv_mag_sq < 112) {
if (Rs < 844) {
return 1;
}
else if (Rs >= 844) {
return 8;
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
}
else if (TSC >= 1) {
if (CsDiff < 3) {
if (SC < 1465) {
if (ssDCval < 71) {
if (tcor < 99) {
if (mu_mv_mag_sq < 112) {
if (Rs < 750) {
return 1;
}
else if (Rs >= 750) {
return 2;
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
else if (tcor >= 99) {
return 2;
}
}
else if (ssDCval >= 71) {
if (posBalance < 27) {
if (posBalance < 24) {
return 8;
}
else if (posBalance >= 24) {
return 1;
}
}
else if (posBalance >= 27) {
return 8;
}
}
}
else if (SC >= 1465) {
if (Cs < 1120) {
return 8;
}
else if (Cs >= 1120) {
return 1;
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 119) {
if (Rs < 979) {
return 8;
}
else if (Rs >= 979) {
return 4;
}
}
else if (refDCval >= 119) {
return 1;
}
}
}
}
else if (Cs >= 1200) {
if (Rs < 837) {
return 2;
}
else if (Rs >= 837) {
return 8;
}
}
}
}
else if (Rs >= 1053) {
if (negBalance < 8) {
if (mcTcor < 95) {
if (Cs < 1178) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 1604) {
return 1;
}
else if (Rs >= 1604) {
return 8;
}
}
}
else if (Cs >= 1178) {
if (Rs < 1321) {
return 1;
}
else if (Rs >= 1321) {
if (Rs < 1528) {
return 4;
}
else if (Rs >= 1528) {
return 8;
}
}
}
}
else if (mcTcor >= 95) {
if (Cs < 617) {
return 4;
}
else if (Cs >= 617) {
if (diffMVdiffVal < -14) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (diffMVdiffVal >= -14) {
if (RsDiff < 2) {
if (Cs < 1214) {
return 8;
}
else if (Cs >= 1214) {
if (Rs < 1186) {
return 1;
}
else if (Rs >= 1186) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 8) {
if (MVDiff < 3) {
if (Rs < 1740) {
if (ssDCval < 64) {
if (ssDCval < 57) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 1490) {
return 1;
}
else if (Rs >= 1490) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (Rs >= 1740) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 76) {
if (posBalance < 30) {
if (SC < 1797) {
return 8;
}
else if (SC >= 1797) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (mu_mv_mag_sq < 105) {
return 1;
}
else if (mu_mv_mag_sq >= 105) {
if (Cs < 672) {
return 1;
}
else if (Cs >= 672) {
return 8;
}
}
}
}
}
else if (posBalance >= 30) {
return 1;
}
}
else if (ssDCval >= 76) {
if (Rs < 1364) {
return 4;
}
else if (Rs >= 1364) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1841) {
if (Rs < 2021) {
if (ssDCval < 81) {
if (Cs < 1000) {
return 1;
}
else if (Cs >= 1000) {
if (ssDCval < 69) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 4;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (Rs >= 2021) {
if (posBalance < 30) {
if (mu_mv_mag_sq < 97) {
return 4;
}
else if (mu_mv_mag_sq >= 97) {
return 8;
}
}
else if (posBalance >= 30) {
if (RsCsDiff < 1) {
if (Rs < 2261) {
return 8;
}
else if (Rs >= 2261) {
return 2;
}
}
else if (RsCsDiff >= 1) {
if (diffMVdiffVal < -1) {
if (Rs < 2038) {
return 2;
}
else if (Rs >= 2038) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 1;
}
}
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 45) {
if (ssDCval < 46) {
if (MVDiff < 9) {
if (Rs < 1467) {
if (ssDCval < 29) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (ssDCval >= 29) {
if (ssDCval < 36) {
if (Rs < 936) {
return 1;
}
else if (Rs >= 936) {
return 8;
}
}
else if (ssDCval >= 36) {
if (TSC < 4) {
if (Rs < 439) {
return 4;
}
else if (Rs >= 439) {
if (Rs < 1305) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (Rs >= 1305) {
if (Cs < 840) {
return 4;
}
else if (Cs >= 840) {
return 8;
}
}
}
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (Rs >= 1467) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (MVDiff >= 9) {
if (tcor < 62) {
return 8;
}
else if (tcor >= 62) {
if (refDCval < 28) {
return 4;
}
else if (refDCval >= 28) {
if (negBalance < 32) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
if (mcTcor < 84) {
if (Rs < 574) {
return 1;
}
else if (Rs >= 574) {
return 8;
}
}
else if (mcTcor >= 84) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
else if (negBalance >= 32) {
if (diffAFD < -1) {
if (CsDiff < 5) {
if (Cs < 825) {
return 4;
}
else if (Cs >= 825) {
if (Rs < 1350) {
return 8;
}
else if (Rs >= 1350) {
return 2;
}
}
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 46) {
if (mu_mv_mag_sq < 86) {
if (negBalance < 64) {
if (Cs < 763) {
if (ssDCval < 132) {
if (Rs < 1052) {
if (TSC < 5) {
if (ssDCval < 78) {
if (gchDC < 1) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (gchDC >= 1) {
if (Rs < 326) {
return 4;
}
else if (Rs >= 326) {
if (Cs < 601) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Cs >= 601) {
return 1;
}
}
}
}
else if (ssDCval >= 78) {
return 2;
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 40) {
return 2;
}
else if (mu_mv_mag_sq >= 40) {
if (Rs < 702) {
return 2;
}
else if (Rs >= 702) {
return 8;
}
}
}
}
else if (Rs >= 1052) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 49) {
if (mcTcor < 94) {
if (ssDCval < 58) {
return 2;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (mcTcor >= 94) {
if (Rs < 1282) {
return 1;
}
else if (Rs >= 1282) {
if (Rs < 2073) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 2;
}
}
else if (Rs >= 2073) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 49) {
return 1;
}
}
else if (ssDCval >= 69) {
if (tcor < 84) {
if (Rs < 1902) {
return 4;
}
else if (Rs >= 1902) {
return 1;
}
}
else if (tcor >= 84) {
if (ssDCval < 104) {
return 1;
}
else if (ssDCval >= 104) {
if (Cs < 700) {
return 1;
}
else if (Cs >= 700) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 132) {
return 8;
}
}
else if (Cs >= 763) {
if (tcor < 89) {
if (Rs < 1889) {
if (RsCsDiff < 57) {
if (diffRsCsdiff < -26) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (diffRsCsdiff >= -26) {
if (AFD < 8) {
if (ssDCval < 63) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 8;
}
}
else if (ssDCval >= 63) {
if (mu_mv_mag_sq < 58) {
return 2;
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 957) {
return 2;
}
else if (Rs >= 957) {
return 8;
}
}
}
}
else if (AFD >= 8) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 57) {
if (MVDiff < 29) {
if (CsDiff < 5) {
if (Rs < 1715) {
return 1;
}
else if (Rs >= 1715) {
return 2;
}
}
else if (CsDiff >= 5) {
if (negBalance < 24) {
if (Rs < 1179) {
return 1;
}
else if (Rs >= 1179) {
return 2;
}
}
else if (negBalance >= 24) {
if (diffMVdiffVal < -10) {
return 2;
}
else if (diffMVdiffVal >= -10) {
return 8;
}
}
}
}
else if (MVDiff >= 29) {
return 2;
}
}
}
else if (Rs >= 1889) {
if (Cs < 1043) {
if (negBalance < 44) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (negBalance >= 44) {
if (Rs < 2100) {
return 4;
}
else if (Rs >= 2100) {
return 8;
}
}
}
else if (Cs >= 1043) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
return 2;
}
}
}
}
else if (tcor >= 89) {
if (Cs < 990) {
if (RsCsDiff < 4) {
return 8;
}
else if (RsCsDiff >= 4) {
if (ssDCval < 104) {
if (Rs < 803) {
if (tcor < 94) {
return 2;
}
else if (tcor >= 94) {
return 4;
}
}
else if (Rs >= 803) {
if (ssDCval < 61) {
if (MVDiff < 13) {
return 1;
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (ssDCval >= 61) {
if (ssDCval < 70) {
if (diffRsCsdiff < -6) {
return 2;
}
else if (diffRsCsdiff >= -6) {
return 4;
}
}
else if (ssDCval >= 70) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
}
}
else if (ssDCval >= 104) {
if (ssDCval < 148) {
return 4;
}
else if (ssDCval >= 148) {
return 2;
}
}
}
}
else if (Cs >= 990) {
if (ssDCval < 81) {
if (mu_mv_mag_sq < 63) {
if (diffRsCsdiff < -3) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (diffRsCsdiff >= -3) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 63) {
return 4;
}
}
else if (ssDCval >= 81) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 64) {
if (tcor < 84) {
if (ssDCval < 88) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (ssDCval >= 88) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
}
else if (tcor >= 84) {
if (gchDC < 2) {
if (Rs < 879) {
return 8;
}
else if (Rs >= 879) {
return 4;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (ssDCval < 56) {
if (Rs < 613) {
if (Rs < 356) {
return 2;
}
else if (Rs >= 356) {
return 1;
}
}
else if (Rs >= 613) {
if (ssDCval < 52) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
if (posBalance < 43) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 117) {
if (Rs < 2088) {
return 4;
}
else if (Rs >= 2088) {
return 4;
}
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
else if (mcTcor >= 95) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
return 8;
}
}
}
else if (posBalance >= 43) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (tcor < 85) {
if (Rs < 1408) {
return 4;
}
else if (Rs >= 1408) {
return 8;
}
}
else if (tcor >= 85) {
if (mcTcor < 94) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (mcTcor >= 94) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Rs < 1196) {
return 1;
}
else if (Rs >= 1196) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 56) {
if (Rs < 675) {
if (refDCval < 98) {
if (RsCsDiff < 4) {
return 1;
}
else if (RsCsDiff >= 4) {
return 2;
}
}
else if (refDCval >= 98) {
if (AFD < 4) {
if (Cs < 719) {
return 8;
}
else if (Cs >= 719) {
return 4;
}
}
else if (AFD >= 4) {
if (Rs < 637) {
return 8;
}
else if (Rs >= 637) {
return 4;
}
}
}
}
else if (Rs >= 675) {
if (RsCsDiff < 43) {
if (mu_mv_mag_sq < 91) {
if (tcor < 95) {
if (MVDiff < 5) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
if (Cs < 726) {
return 4;
}
else if (Cs >= 726) {
return 4;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (tcor >= 95) {
if (Rs < 1802) {
if (Rs < 1251) {
return 2;
}
else if (Rs >= 1251) {
return 1;
}
}
else if (Rs >= 1802) {
if (Cs < 972) {
return 4;
}
else if (Cs >= 972) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (TSC < 3) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 96) {
return 4;
}
else if (ssDCval >= 96) {
if (Rs < 1060) {
return 1;
}
else if (Rs >= 1060) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
else if (TSC >= 3) {
if (Rs < 1713) {
if (Cs < 1227) {
if (mu_mv_mag_sq < 94) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 94) {
return 8;
}
}
else if (Cs >= 1227) {
if (Rs < 1154) {
return 4;
}
else if (Rs >= 1154) {
return 8;
}
}
}
else if (Rs >= 1713) {
if (refDCval < 71) {
return 1;
}
else if (refDCval >= 71) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 43) {
if (Rs < 1306) {
return 1;
}
else if (Rs >= 1306) {
if (Rs < 1968) {
return 8;
}
else if (Rs >= 1968) {
return 2;
}
}
}
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 407) {
if (ssDCval < 96) {
if (ssDCval < 82) {
if (Cs < 773) {
if (ssDCval < 60) {
if (SC < 947) {
return 1;
}
else if (SC >= 947) {
if (ssDCval < 55) {
if (Cs < 589) {
return 8;
}
else if (Cs >= 589) {
return 2;
}
}
else if (ssDCval >= 55) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 1;
}
}
}
}
else if (ssDCval >= 60) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Rs < 326) {
return 2;
}
else if (Rs >= 326) {
return 4;
}
}
}
}
else if (Cs >= 773) {
if (Rs < 381) {
if (ssDCval < 52) {
if (Rs < 351) {
return 1;
}
else if (Rs >= 351) {
return 2;
}
}
else if (ssDCval >= 52) {
return 1;
}
}
else if (Rs >= 381) {
if (Cs < 830) {
return 4;
}
else if (Cs >= 830) {
if (Rs < 383) {
return 4;
}
else if (Rs >= 383) {
return 1;
}
}
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
else if (ssDCval >= 96) {
if (Cs < 719) {
if (Rs < 372) {
if (diffMVdiffVal < 6) {
return 2;
}
else if (diffMVdiffVal >= 6) {
return 8;
}
}
else if (Rs >= 372) {
return 1;
}
}
else if (Cs >= 719) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (Rs >= 407) {
if (mu_mv_mag_sq < 41) {
if (negBalance < 44) {
if (refDCval < 64) {
if (tcor < 90) {
if (Rs < 845) {
if (Cs < 662) {
if (AFD < 11) {
if (Rs < 520) {
if (Rs < 454) {
return 2;
}
else if (Rs >= 454) {
return 4;
}
}
else if (Rs >= 520) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (Cs >= 662) {
if (Cs < 1162) {
if (diffMVdiffVal < 8) {
if (ssDCval < 43) {
if (Cs < 789) {
if (diffMVdiffVal < -1) {
if (Rs < 790) {
return 8;
}
else if (Rs >= 790) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (Cs >= 789) {
return 2;
}
}
else if (ssDCval >= 43) {
if (refDCval < 50) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
else if (refDCval >= 50) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
if (diffMVdiffVal < 0) {
return 4;
}
else if (diffMVdiffVal >= 0) {
if (Rs < 553) {
return 2;
}
else if (Rs >= 553) {
return 8;
}
}
}
}
}
}
else if (diffMVdiffVal >= 8) {
if (Cs < 1017) {
return 8;
}
else if (Cs >= 1017) {
return 4;
}
}
}
else if (Cs >= 1162) {
if (MVDiff < 15) {
if (Rs < 767) {
return 8;
}
else if (Rs >= 767) {
return 2;
}
}
else if (MVDiff >= 15) {
return 1;
}
}
}
}
else if (Rs >= 845) {
if (Cs < 648) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 1964) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (Rs >= 1964) {
return 4;
}
}
}
else if (Cs >= 648) {
if (ssDCval < 54) {
if (Cs < 1028) {
return 8;
}
else if (Cs >= 1028) {
if (posBalance < 51) {
if (Cs < 1120) {
return 4;
}
else if (Cs >= 1120) {
return 8;
}
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (ssDCval >= 54) {
if (negBalance < 36) {
if (Rs < 984) {
return 1;
}
else if (Rs >= 984) {
if (Cs < 1113) {
if (tcor < 74) {
return 2;
}
else if (tcor >= 74) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 1618) {
return 2;
}
else if (Rs >= 1618) {
return 8;
}
}
}
}
else if (Cs >= 1113) {
return 1;
}
}
}
else if (negBalance >= 36) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 1210) {
if (tcor < 80) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
else if (tcor >= 80) {
if (Rs < 1177) {
return 4;
}
else if (Rs >= 1177) {
return 8;
}
}
}
else if (Rs >= 1210) {
if (Cs < 878) {
if (Rs < 1429) {
return 8;
}
else if (Rs >= 1429) {
return 4;
}
}
else if (Cs >= 878) {
return 8;
}
}
}
}
}
}
}
}
else if (tcor >= 90) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (SC < 2462) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
if (ssDCval < 59) {
if (TSC < 5) {
if (Cs < 1082) {
return 8;
}
else if (Cs >= 1082) {
return 4;
}
}
else if (TSC >= 5) {
return 2;
}
}
else if (ssDCval >= 59) {
return 4;
}
}
}
else if (SC >= 2462) {
if (RsDiff < 6) {
if (Cs < 819) {
return 1;
}
else if (Cs >= 819) {
return 8;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
else if (refDCval >= 64) {
if (posBalance < 59) {
if (Rs < 757) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
if (Rs < 641) {
if (Rs < 449) {
return 8;
}
else if (Rs >= 449) {
return 4;
}
}
else if (Rs >= 641) {
return 1;
}
}
}
else if (Rs >= 757) {
if (Cs < 856) {
if (diffMVdiffVal < 3) {
if (RsDiff < 7) {
if (AFD < 6) {
if (Rs < 845) {
return 1;
}
else if (Rs >= 845) {
return 8;
}
}
else if (AFD >= 6) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (RsDiff >= 7) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
if (ssDCval < 93) {
return 1;
}
else if (ssDCval >= 93) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1387) {
if (ssDCval < 152) {
return 4;
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (Rs >= 1387) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 4;
}
}
}
}
else if (Cs >= 856) {
if (Rs < 1420) {
if (ssDCval < 137) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Cs < 1256) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (SC < 2500) {
return 1;
}
else if (SC >= 2500) {
return 2;
}
}
}
else if (Cs >= 1256) {
return 2;
}
}
}
else if (ssDCval >= 137) {
if (Cs < 923) {
return 8;
}
else if (Cs >= 923) {
return 2;
}
}
}
else if (Rs >= 1420) {
if (RsDiff < 7) {
if (Rs < 2217) {
if (ssDCval < 87) {
if (Rs < 1646) {
return 8;
}
else if (Rs >= 1646) {
return 4;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
else if (Rs >= 2217) {
if (Cs < 1139) {
return 4;
}
else if (Cs >= 1139) {
return 8;
}
}
}
else if (RsDiff >= 7) {
if (MVDiff < 14) {
if (SC < 3111) {
if (Rs < 2191) {
return 8;
}
else if (Rs >= 2191) {
return 2;
}
}
else if (SC >= 3111) {
return 2;
}
}
else if (MVDiff >= 14) {
if (Cs < 1245) {
return 1;
}
else if (Cs >= 1245) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 59) {
if (refDCval < 137) {
if (refDCval < 103) {
if (negBalance < 35) {
if (posBalance < 67) {
return 1;
}
else if (posBalance >= 67) {
if (Rs < 1762) {
return 8;
}
else if (Rs >= 1762) {
return 1;
}
}
}
else if (negBalance >= 35) {
if (diffMVdiffVal < -1) {
if (SC < 2568) {
if (Rs < 1181) {
if (ssDCval < 70) {
if (Rs < 772) {
return 2;
}
else if (Rs >= 772) {
return 1;
}
}
else if (ssDCval >= 70) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
else if (Rs >= 1181) {
if (Rs < 1302) {
return 8;
}
else if (Rs >= 1302) {
return 1;
}
}
}
else if (SC >= 2568) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (MVDiff < 4) {
if (Rs < 1101) {
return 8;
}
else if (Rs >= 1101) {
return 1;
}
}
else if (MVDiff >= 4) {
if (tcor < 88) {
if (Cs < 1149) {
if (Rs < 847) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
else if (Rs >= 847) {
return 8;
}
}
else if (Cs >= 1149) {
if (Rs < 1451) {
return 4;
}
else if (Rs >= 1451) {
return 2;
}
}
}
else if (tcor >= 88) {
if (Cs < 549) {
return 4;
}
else if (Cs >= 549) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 103) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 14) {
if (Rs < 1558) {
return 8;
}
else if (Rs >= 1558) {
return 2;
}
}
else if (mu_mv_mag_sq >= 14) {
return 8;
}
}
else if (negBalance >= 37) {
return 8;
}
}
}
else if (refDCval >= 137) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (Rs < 752) {
return 8;
}
else if (Rs >= 752) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 44) {
if (RsDiff < 8) {
if (Rs < 957) {
if (CsDiff < 5) {
if (diffAFD < -2) {
if (mcTcor < 93) {
if (RsCsDiff < 29) {
if (AFD < 10) {
if (ssDCval < 42) {
return 8;
}
else if (ssDCval >= 42) {
return 4;
}
}
else if (AFD >= 10) {
return 8;
}
}
else if (RsCsDiff >= 29) {
if (Rs < 822) {
return 1;
}
else if (Rs >= 822) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (diffAFD < -3) {
if (Rs < 459) {
return 2;
}
else if (Rs >= 459) {
return 1;
}
}
else if (diffAFD >= -3) {
if (gchDC < 1) {
if (Rs < 636) {
return 1;
}
else if (Rs >= 636) {
return 4;
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
else if (diffAFD >= -2) {
if (refDCval < 49) {
if (ssDCval < 45) {
if (Rs < 816) {
if (CsDiff < 3) {
if (MVDiff < 32) {
return 8;
}
else if (MVDiff >= 32) {
return 2;
}
}
else if (CsDiff >= 3) {
if (ssDCval < 41) {
if (refDCval < 39) {
return 8;
}
else if (refDCval >= 39) {
return 2;
}
}
else if (ssDCval >= 41) {
if (Rs < 716) {
return 4;
}
else if (Rs >= 716) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
else if (Rs >= 816) {
return 8;
}
}
else if (ssDCval >= 45) {
if (negBalance < 54) {
if (mcTcor < 92) {
if (posBalance < 61) {
if (posBalance < 59) {
if (Cs < 716) {
return 1;
}
else if (Cs >= 716) {
return 8;
}
}
else if (posBalance >= 59) {
return 1;
}
}
else if (posBalance >= 61) {
if (Cs < 843) {
return 8;
}
else if (Cs >= 843) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (Rs < 530) {
return 2;
}
else if (Rs >= 530) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
}
}
else if (negBalance >= 54) {
if (Cs < 660) {
if (Cs < 614) {
return 8;
}
else if (Cs >= 614) {
return 1;
}
}
else if (Cs >= 660) {
return 8;
}
}
}
}
else if (refDCval >= 49) {
if (RsDiff < 4) {
if (mcTcor < 95) {
if (RsDiff < 3) {
if (ssDCval < 49) {
return 2;
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (RsDiff >= 3) {
if (ssDCval < 117) {
if (ssDCval < 58) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
return 4;
}
}
}
else if (ssDCval >= 58) {
if (negBalance < 50) {
if (CsDiff < 4) {
if (CsDiff < 3) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (negBalance >= 50) {
if (MVDiff < 4) {
if (Rs < 728) {
return 1;
}
else if (Rs >= 728) {
if (Rs < 872) {
return 8;
}
else if (Rs >= 872) {
return 2;
}
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
else if (ssDCval >= 117) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 151) {
if (ssDCval < 85) {
if (negBalance < 47) {
if (SC < 1277) {
return 8;
}
else if (SC >= 1277) {
return 1;
}
}
else if (negBalance >= 47) {
return 8;
}
}
else if (ssDCval >= 85) {
if (negBalance < 49) {
if (Rs < 595) {
return 8;
}
else if (Rs >= 595) {
return 4;
}
}
else if (negBalance >= 49) {
return 1;
}
}
}
else if (ssDCval >= 151) {
return 2;
}
}
}
else if (RsDiff >= 4) {
if (ssDCval < 51) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
return 1;
}
}
else if (ssDCval >= 51) {
if (Cs < 681) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
return 8;
}
}
else if (Cs >= 681) {
if (posBalance < 55) {
if (Cs < 784) {
return 2;
}
else if (Cs >= 784) {
if (Rs < 834) {
return 8;
}
else if (Rs >= 834) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (Rs < 944) {
return 4;
}
else if (Rs >= 944) {
return 2;
}
}
}
}
}
}
}
}
else if (CsDiff >= 5) {
if (negBalance < 50) {
if (diffAFD < -1) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
if (Rs < 914) {
return 4;
}
else if (Rs >= 914) {
return 2;
}
}
}
else if (diffAFD >= -1) {
if (ssDCval < 56) {
if (Rs < 474) {
return 8;
}
else if (Rs >= 474) {
return 2;
}
}
else if (ssDCval >= 56) {
if (Rs < 890) {
return 1;
}
else if (Rs >= 890) {
return 2;
}
}
}
}
else if (negBalance >= 50) {
if (mcTcor < 84) {
if (RsDiff < 6) {
if (AFD < 11) {
if (Rs < 689) {
return 4;
}
else if (Rs >= 689) {
return 8;
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (mcTcor >= 84) {
if (RsCsDiff < 46) {
if (Cs < 1075) {
return 1;
}
else if (Cs >= 1075) {
return 4;
}
}
else if (RsCsDiff >= 46) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 4;
}
}
}
}
}
}
else if (Rs >= 957) {
if (tcor < 80) {
if (mcTcor < 84) {
if (ssDCval < 122) {
if (Rs < 1116) {
if (Cs < 907) {
return 8;
}
else if (Cs >= 907) {
if (mcTcor < 81) {
return 8;
}
else if (mcTcor >= 81) {
return 1;
}
}
}
else if (Rs >= 1116) {
return 8;
}
}
else if (ssDCval >= 122) {
return 1;
}
}
else if (mcTcor >= 84) {
if (MVDiff < 7) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (Cs < 1243) {
return 8;
}
else if (Cs >= 1243) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
return 4;
}
}
}
}
else if (MVDiff >= 7) {
if (refDCval < 70) {
if (Cs < 1241) {
if (diffRsCsdiff < -17) {
if (Rs < 1195) {
return 1;
}
else if (Rs >= 1195) {
return 8;
}
}
else if (diffRsCsdiff >= -17) {
return 8;
}
}
else if (Cs >= 1241) {
if (Rs < 1186) {
return 4;
}
else if (Rs >= 1186) {
return 1;
}
}
}
else if (refDCval >= 70) {
if (Rs < 1157) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (Rs >= 1157) {
if (ssDCval < 99) {
if (Cs < 947) {
return 2;
}
else if (Cs >= 947) {
if (Rs < 1564) {
return 8;
}
else if (Rs >= 1564) {
if (Cs < 1177) {
return 2;
}
else if (Cs >= 1177) {
return 8;
}
}
}
}
else if (ssDCval >= 99) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 80) {
if (MVDiff < 5) {
if (CsDiff < 5) {
if (diffAFD < -1) {
if (Cs < 658) {
if (Rs < 1385) {
return 8;
}
else if (Rs >= 1385) {
return 2;
}
}
else if (Cs >= 658) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (Rs < 1469) {
return 8;
}
else if (Rs >= 1469) {
if (Rs < 2088) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (Rs >= 2088) {
return 8;
}
}
}
}
}
else if (diffAFD >= -1) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (Cs < 1041) {
return 1;
}
else if (Cs >= 1041) {
return 8;
}
}
}
}
}
else if (CsDiff >= 5) {
if (Rs < 1515) {
return 4;
}
else if (Rs >= 1515) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
}
else if (MVDiff >= 5) {
if (AFD < 10) {
if (ssDCval < 136) {
if (ssDCval < 98) {
if (ssDCval < 63) {
if (RsCsDiff < 23) {
return 8;
}
else if (RsCsDiff >= 23) {
if (AFD < 8) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (Cs < 810) {
if (Rs < 1363) {
return 2;
}
else if (Rs >= 1363) {
return 4;
}
}
else if (Cs >= 810) {
return 8;
}
}
}
else if (AFD >= 8) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
if (ssDCval < 49) {
if (Rs < 1048) {
return 8;
}
else if (Rs >= 1048) {
return 4;
}
}
else if (ssDCval >= 49) {
return 8;
}
}
}
}
}
else if (ssDCval >= 63) {
if (ssDCval < 80) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (ssDCval >= 80) {
if (RsDiff < 5) {
if (diffMVdiffVal < -5) {
return 2;
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (RsDiff >= 5) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 2;
}
}
}
}
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (ssDCval >= 136) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
return 4;
}
}
}
else if (AFD >= 10) {
if (Cs < 678) {
if (mu_mv_mag_sq < 7) {
if (MVDiff < 8) {
if (Cs < 661) {
return 2;
}
else if (Cs >= 661) {
return 1;
}
}
else if (MVDiff >= 8) {
if (CsDiff < 3) {
if (Rs < 1744) {
return 8;
}
else if (Rs >= 1744) {
return 2;
}
}
else if (CsDiff >= 3) {
if (MVDiff < 26) {
return 1;
}
else if (MVDiff >= 26) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
else if (Cs >= 678) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 18) {
if (Cs < 977) {
if (Rs < 2285) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (Rs < 1316) {
return 4;
}
else if (Rs >= 1316) {
return 8;
}
}
}
else if (Rs >= 2285) {
return 2;
}
}
else if (Cs >= 977) {
if (ssDCval < 58) {
if (Cs < 1051) {
return 8;
}
else if (Cs >= 1051) {
return 4;
}
}
else if (ssDCval >= 58) {
if (posBalance < 56) {
if (TSC < 9) {
if (Rs < 2019) {
if (negBalance < 48) {
if (Rs < 1371) {
return 4;
}
else if (Rs >= 1371) {
return 2;
}
}
else if (negBalance >= 48) {
return 1;
}
}
else if (Rs >= 2019) {
return 4;
}
}
else if (TSC >= 9) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (posBalance >= 56) {
if (tcor < 81) {
return 4;
}
else if (tcor >= 81) {
if (Rs < 1404) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 4;
}
}
}
else if (Rs >= 1404) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 18) {
if (CsDiff < 5) {
if (Rs < 1188) {
return 4;
}
else if (Rs >= 1188) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (Cs < 1082) {
if (Rs < 1166) {
return 2;
}
else if (Rs >= 1166) {
return 1;
}
}
else if (Cs >= 1082) {
if (Rs < 1936) {
return 8;
}
else if (Rs >= 1936) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 1204) {
if (RsCsDiff < 46) {
if (Rs < 1616) {
return 8;
}
else if (Rs >= 1616) {
if (Rs < 1633) {
return 1;
}
else if (Rs >= 1633) {
return 8;
}
}
}
else if (RsCsDiff >= 46) {
if (ssDCval < 93) {
if (SC < 2312) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (SC >= 2312) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Cs < 833) {
return 8;
}
else if (Cs >= 833) {
return 4;
}
}
}
}
else if (ssDCval >= 93) {
return 8;
}
}
}
else if (Cs >= 1204) {
if (Cs < 1236) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (Cs >= 1236) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (mcTcor < 93) {
if (posBalance < 51) {
if (ssDCval < 70) {
return 4;
}
else if (ssDCval >= 70) {
if (Rs < 1876) {
return 8;
}
else if (Rs >= 1876) {
return 2;
}
}
}
else if (posBalance >= 51) {
if (Cs < 595) {
if (Rs < 1603) {
return 2;
}
else if (Rs >= 1603) {
return 4;
}
}
else if (Cs >= 595) {
if (Rs < 2202) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (Cs < 879) {
if (Rs < 1696) {
return 4;
}
else if (Rs >= 1696) {
return 1;
}
}
else if (Cs >= 879) {
return 8;
}
}
}
else if (Rs >= 2202) {
return 4;
}
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 120) {
if (AFD < 10) {
if (Rs < 2165) {
return 1;
}
else if (Rs >= 2165) {
return 2;
}
}
else if (AFD >= 10) {
return 2;
}
}
else if (ssDCval >= 120) {
if (Rs < 1668) {
return 8;
}
else if (Rs >= 1668) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (Rs < 1596) {
if (TSC < 7) {
if (ssDCval < 75) {
if (ssDCval < 52) {
if (mu_mv_mag_sq < 117) {
if (mcTcor < 95) {
if (refDCval < 40) {
return 8;
}
else if (refDCval >= 40) {
if (TSC < 5) {
if (CsDiff < 4) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (posBalance < 56) {
if (TSC < 4) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (posBalance >= 56) {
if (Rs < 1214) {
return 8;
}
else if (Rs >= 1214) {
return 2;
}
}
}
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (TSC >= 5) {
if (negBalance < 47) {
if (RsCsDiff < 33) {
if (tcor < 68) {
if (Rs < 797) {
return 2;
}
else if (Rs >= 797) {
return 8;
}
}
else if (tcor >= 68) {
return 8;
}
}
else if (RsCsDiff >= 33) {
if (diffMVdiffVal < -1) {
if (ssDCval < 47) {
if (refDCval < 42) {
if (Rs < 1033) {
return 1;
}
else if (Rs >= 1033) {
return 8;
}
}
else if (refDCval >= 42) {
return 2;
}
}
else if (ssDCval >= 47) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (tcor < 82) {
return 8;
}
else if (tcor >= 82) {
if (refDCval < 49) {
return 2;
}
else if (refDCval >= 49) {
return 8;
}
}
}
}
}
else if (negBalance >= 47) {
if (Cs < 831) {
if (Cs < 675) {
return 8;
}
else if (Cs >= 675) {
return 1;
}
}
else if (Cs >= 831) {
return 8;
}
}
}
}
}
else if (mcTcor >= 95) {
if (SC < 1599) {
if (tcor < 93) {
if (Cs < 572) {
return 2;
}
else if (Cs >= 572) {
return 1;
}
}
else if (tcor >= 93) {
return 4;
}
}
else if (SC >= 1599) {
if (CsDiff < 3) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
else if (CsDiff >= 3) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 1105) {
if (Cs < 1021) {
return 4;
}
else if (Cs >= 1021) {
return 8;
}
}
else if (Cs >= 1105) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (SC < 2245) {
if (ssDCval < 47) {
if (Rs < 1110) {
if (Rs < 867) {
return 1;
}
else if (Rs >= 867) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
return 8;
}
}
}
else if (Rs >= 1110) {
if (RsCsDiff < 37) {
return 1;
}
else if (RsCsDiff >= 37) {
return 8;
}
}
}
else if (ssDCval >= 47) {
return 8;
}
}
else if (SC >= 2245) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (Rs < 1328) {
if (RsCsDiff < 15) {
if (Rs < 996) {
if (mu_mv_mag_sq < 99) {
if (ssDCval < 55) {
if (tcor < 94) {
if (refDCval < 55) {
return 4;
}
else if (refDCval >= 55) {
if (Rs < 652) {
return 1;
}
else if (Rs >= 652) {
return 8;
}
}
}
else if (tcor >= 94) {
if (diffRsCsdiff < -5) {
return 4;
}
else if (diffRsCsdiff >= -5) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
}
}
else if (ssDCval >= 55) {
if (RsCsDiff < 7) {
if (Rs < 518) {
if (Rs < 426) {
return 8;
}
else if (Rs >= 426) {
if (Rs < 489) {
return 1;
}
else if (Rs >= 489) {
return 2;
}
}
}
else if (Rs >= 518) {
if (gchDC < 2) {
if (Cs < 1288) {
return 8;
}
else if (Cs >= 1288) {
return 1;
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (RsCsDiff >= 7) {
if (Rs < 877) {
if (negBalance < 36) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
if (Rs < 834) {
if (AFD < 5) {
if (Cs < 780) {
if (Rs < 797) {
return 4;
}
else if (Rs >= 797) {
return 8;
}
}
else if (Cs >= 780) {
return 1;
}
}
else if (AFD >= 5) {
if (Rs < 524) {
return 8;
}
else if (Rs >= 524) {
return 1;
}
}
}
else if (Rs >= 834) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (mu_mv_mag_sq < 78) {
if (mu_mv_mag_sq < 60) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
return 1;
}
}
else if (mu_mv_mag_sq >= 60) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 78) {
return 8;
}
}
}
else if (Rs >= 877) {
if (negBalance < 47) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 958) {
return 8;
}
else if (Rs >= 958) {
return 1;
}
}
}
else if (negBalance >= 47) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (Cs < 568) {
return 8;
}
else if (Cs >= 568) {
if (negBalance < 42) {
if (posBalance < 53) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (refDCval < 66) {
if (Rs < 602) {
if (Cs < 696) {
return 8;
}
else if (Cs >= 696) {
return 4;
}
}
else if (Rs >= 602) {
return 8;
}
}
else if (refDCval >= 66) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 8;
}
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
}
else if (Rs >= 996) {
if (mu_mv_mag_sq < 123) {
if (posBalance < 56) {
if (negBalance < 45) {
if (ssDCval < 59) {
if (ssDCval < 53) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (ssDCval >= 53) {
if (TSC < 5) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (ssDCval >= 59) {
if (AFD < 4) {
if (Rs < 1141) {
return 4;
}
else if (Rs >= 1141) {
return 1;
}
}
else if (AFD >= 4) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
if (mu_mv_mag_sq < 117) {
if (negBalance < 40) {
if (Rs < 1249) {
if (negBalance < 35) {
if (CsDiff < 2) {
if (Cs < 748) {
return 1;
}
else if (Cs >= 748) {
return 8;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (negBalance >= 35) {
return 1;
}
}
else if (Rs >= 1249) {
return 8;
}
}
else if (negBalance >= 40) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (Cs < 1116) {
return 8;
}
else if (Cs >= 1116) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 1241) {
return 4;
}
else if (Rs >= 1241) {
return 8;
}
}
}
else if (posBalance >= 56) {
if (mu_mv_mag_sq < 51) {
if (ssDCval < 57) {
if (Rs < 1127) {
return 4;
}
else if (Rs >= 1127) {
return 1;
}
}
else if (ssDCval >= 57) {
if (Cs < 577) {
return 4;
}
else if (Cs >= 577) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 51) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (negBalance < 37) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 4;
}
}
else if (negBalance >= 37) {
return 1;
}
}
}
}
else if (RsCsDiff >= 15) {
if (Cs < 1038) {
if (Cs < 907) {
if (tcor < 96) {
if (negBalance < 45) {
if (AFD < 5) {
if (Rs < 1175) {
return 4;
}
else if (Rs >= 1175) {
return 1;
}
}
else if (AFD >= 5) {
if (Rs < 1185) {
if (MVDiff < 14) {
if (negBalance < 30) {
return 2;
}
else if (negBalance >= 30) {
if (AFD < 9) {
if (mcTcor < 92) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (Cs < 728) {
if (Cs < 610) {
return 8;
}
else if (Cs >= 610) {
return 1;
}
}
else if (Cs >= 728) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (Rs < 1167) {
return 1;
}
else if (Rs >= 1167) {
return 4;
}
}
}
else if (AFD >= 9) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 4;
}
}
}
}
else if (MVDiff >= 14) {
return 1;
}
}
else if (Rs >= 1185) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
if (Cs < 699) {
return 1;
}
else if (Cs >= 699) {
return 2;
}
}
}
}
}
else if (negBalance >= 45) {
if (CsDiff < 4) {
if (Cs < 628) {
return 8;
}
else if (Cs >= 628) {
return 4;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (tcor >= 96) {
return 8;
}
}
else if (Cs >= 907) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
if (ssDCval < 59) {
if (Cs < 933) {
return 8;
}
else if (Cs >= 933) {
return 4;
}
}
else if (ssDCval >= 59) {
return 8;
}
}
}
}
else if (Cs >= 1038) {
if (negBalance < 43) {
if (TSC < 5) {
if (negBalance < 38) {
if (Rs < 646) {
return 1;
}
else if (Rs >= 646) {
if (posBalance < 56) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
return 8;
}
}
else if (posBalance >= 56) {
if (Cs < 1097) {
return 4;
}
else if (Cs >= 1097) {
return 8;
}
}
}
}
else if (negBalance >= 38) {
return 1;
}
}
else if (TSC >= 5) {
if (mcTcor < 90) {
return 1;
}
else if (mcTcor >= 90) {
if (ssDCval < 55) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (ssDCval >= 55) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 2;
}
}
}
}
}
else if (negBalance >= 43) {
if (Cs < 1211) {
if (Rs < 1163) {
if (mu_mv_mag_sq < 86) {
if (Rs < 869) {
return 1;
}
else if (Rs >= 869) {
return 4;
}
}
else if (mu_mv_mag_sq >= 86) {
return 8;
}
}
else if (Rs >= 1163) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 4;
}
}
}
else if (Cs >= 1211) {
return 1;
}
}
}
}
}
else if (Rs >= 1328) {
if (mu_mv_mag_sq < 65) {
if (ssDCval < 67) {
if (posBalance < 46) {
if (Rs < 1454) {
return 4;
}
else if (Rs >= 1454) {
return 2;
}
}
else if (posBalance >= 46) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
return 8;
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 44) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 4;
}
}
else if (negBalance >= 44) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 65) {
if (posBalance < 61) {
if (diffRsCsdiff < 6) {
if (Cs < 1299) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
if (posBalance < 56) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
return 4;
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
else if (Cs >= 1299) {
if (Rs < 1467) {
return 1;
}
else if (Rs >= 1467) {
return 8;
}
}
}
else if (diffRsCsdiff >= 6) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (Cs < 1023) {
return 4;
}
else if (Cs >= 1023) {
return 8;
}
}
}
}
}
else if (posBalance >= 61) {
if (Cs < 1134) {
return 4;
}
else if (Cs >= 1134) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (SC < 1775) {
if (MVDiff < 6) {
if (refDCval < 89) {
if (negBalance < 31) {
if (ssDCval < 87) {
if (Rs < 480) {
return 8;
}
else if (Rs >= 480) {
return 2;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
else if (negBalance >= 31) {
if (Rs < 783) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (Rs >= 783) {
return 2;
}
}
}
else if (refDCval >= 89) {
if (mu_mv_mag_sq < 107) {
if (Cs < 973) {
if (MVDiff < 4) {
if (RsDiff < 2) {
if (Rs < 415) {
if (Cs < 742) {
return 1;
}
else if (Cs >= 742) {
return 4;
}
}
else if (Rs >= 415) {
if (posBalance < 50) {
if (Rs < 671) {
return 2;
}
else if (Rs >= 671) {
return 1;
}
}
else if (posBalance >= 50) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 77) {
return 2;
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 552) {
return 8;
}
else if (Rs >= 552) {
return 1;
}
}
}
}
else if (MVDiff >= 4) {
if (negBalance < 35) {
if (Rs < 880) {
return 2;
}
else if (Rs >= 880) {
return 8;
}
}
else if (negBalance >= 35) {
if (diffMVdiffVal < 3) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
}
else if (Cs >= 973) {
return 8;
}
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 1071) {
return 8;
}
else if (Rs >= 1071) {
return 4;
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 155) {
if (AFD < 9) {
if (TSC < 4) {
if (Rs < 508) {
return 8;
}
else if (Rs >= 508) {
return 2;
}
}
else if (TSC >= 4) {
if (posBalance < 49) {
if (Rs < 708) {
return 8;
}
else if (Rs >= 708) {
return 1;
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
else if (AFD >= 9) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (ssDCval >= 155) {
if (Rs < 850) {
return 4;
}
else if (Rs >= 850) {
return 2;
}
}
}
}
else if (SC >= 1775) {
if (ssDCval < 159) {
if (refDCval < 130) {
if (Rs < 952) {
if (ssDCval < 92) {
if (Cs < 1010) {
return 8;
}
else if (Cs >= 1010) {
if (posBalance < 56) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
return 4;
}
}
else if (posBalance >= 56) {
return 4;
}
}
}
else if (ssDCval >= 92) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (Cs < 942) {
return 1;
}
else if (Cs >= 942) {
if (SC < 2087) {
return 8;
}
else if (SC >= 2087) {
if (Rs < 848) {
return 8;
}
else if (Rs >= 848) {
return 4;
}
}
}
}
}
}
else if (Rs >= 952) {
if (ssDCval < 124) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 1206) {
return 8;
}
else if (Rs >= 1206) {
return 4;
}
}
}
else if (ssDCval >= 124) {
return 1;
}
}
}
else if (refDCval >= 130) {
if (Cs < 1033) {
if (AFD < 6) {
if (posBalance < 66) {
if (Cs < 636) {
return 8;
}
else if (Cs >= 636) {
return 4;
}
}
else if (posBalance >= 66) {
return 1;
}
}
else if (AFD >= 6) {
if (MVDiff < 24) {
if (refDCval < 131) {
return 4;
}
else if (refDCval >= 131) {
if (diffMVdiffVal < -8) {
return 4;
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
}
else if (MVDiff >= 24) {
return 4;
}
}
}
else if (Cs >= 1033) {
return 8;
}
}
}
else if (ssDCval >= 159) {
if (Rs < 959) {
return 4;
}
else if (Rs >= 959) {
return 2;
}
}
}
}
}
else if (TSC >= 7) {
if (negBalance < 44) {
if (ssDCval < 73) {
if (ssDCval < 58) {
if (Rs < 867) {
if (Rs < 737) {
return 2;
}
else if (Rs >= 737) {
return 4;
}
}
else if (Rs >= 867) {
return 8;
}
}
else if (ssDCval >= 58) {
if (RsCsDiff < 43) {
if (Rs < 947) {
if (Rs < 677) {
return 8;
}
else if (Rs >= 677) {
return 1;
}
}
else if (Rs >= 947) {
if (Rs < 1243) {
return 4;
}
else if (Rs >= 1243) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (Rs < 1372) {
return 8;
}
else if (Rs >= 1372) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 43) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 1;
}
}
}
}
else if (ssDCval >= 73) {
if (AFD < 12) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 1018) {
return 8;
}
else if (Rs >= 1018) {
return 2;
}
}
}
else if (AFD >= 12) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
if (Rs < 1060) {
return 1;
}
else if (Rs >= 1060) {
return 2;
}
}
}
}
}
else if (negBalance >= 44) {
if (tcor < 79) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (SC < 1793) {
if (ssDCval < 49) {
if (Rs < 866) {
return 8;
}
else if (Rs >= 866) {
if (Rs < 986) {
return 1;
}
else if (Rs >= 986) {
return 8;
}
}
}
else if (ssDCval >= 49) {
if (Cs < 718) {
if (mcTcor < 84) {
return 8;
}
else if (mcTcor >= 84) {
return 4;
}
}
else if (Cs >= 718) {
return 4;
}
}
}
else if (SC >= 1793) {
return 8;
}
}
}
else if (tcor >= 79) {
if (posBalance < 57) {
if (mcTcor < 96) {
if (Rs < 1315) {
if (Rs < 1282) {
if (RsDiff < 6) {
if (refDCval < 155) {
return 8;
}
else if (refDCval >= 155) {
return 2;
}
}
else if (RsDiff >= 6) {
if (Rs < 1108) {
return 1;
}
else if (Rs >= 1108) {
return 8;
}
}
}
else if (Rs >= 1282) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (Rs >= 1315) {
if (Rs < 1580) {
return 8;
}
else if (Rs >= 1580) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1431) {
return 8;
}
else if (Rs >= 1431) {
return 4;
}
}
}
else if (posBalance >= 57) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
if (MVDiff < 84) {
if (RsCsDiff < 51) {
return 8;
}
else if (RsCsDiff >= 51) {
if (Rs < 1381) {
return 8;
}
else if (Rs >= 1381) {
return 2;
}
}
}
else if (MVDiff >= 84) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1596) {
if (ssDCval < 72) {
if (refDCval < 62) {
if (Cs < 900) {
if (RsCsDiff < 39) {
if (MVDiff < 7) {
if (Cs < 703) {
return 1;
}
else if (Cs >= 703) {
return 8;
}
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (RsCsDiff >= 39) {
if (posBalance < 55) {
if (mcTcor < 91) {
if (Cs < 685) {
return 2;
}
else if (Cs >= 685) {
return 4;
}
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (posBalance >= 55) {
if (Rs < 1941) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
else if (Rs >= 1941) {
if (Rs < 2184) {
return 4;
}
else if (Rs >= 2184) {
return 2;
}
}
}
}
}
else if (Cs >= 900) {
if (diffAFD < -5) {
if (Rs < 1688) {
return 1;
}
else if (Rs >= 1688) {
return 2;
}
}
else if (diffAFD >= -5) {
if (MVDiff < 6) {
if (diffMVdiffVal < -5) {
if (Rs < 1652) {
return 8;
}
else if (Rs >= 1652) {
return 2;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (MVDiff >= 6) {
if (Rs < 2095) {
if (posBalance < 46) {
if (Rs < 1759) {
return 8;
}
else if (Rs >= 1759) {
return 4;
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (Rs >= 2095) {
return 4;
}
}
}
}
}
else if (refDCval >= 62) {
if (Cs < 1111) {
if (Rs < 1904) {
if (refDCval < 66) {
if (Rs < 1616) {
return 8;
}
else if (Rs >= 1616) {
return 2;
}
}
else if (refDCval >= 66) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (negBalance < 47) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
else if (negBalance >= 47) {
return 2;
}
}
}
}
else if (Rs >= 1904) {
if (SC < 3287) {
return 2;
}
else if (SC >= 3287) {
return 4;
}
}
}
else if (Cs >= 1111) {
if (TSC < 8) {
if (diffMVdiffVal < -14) {
return 1;
}
else if (diffMVdiffVal >= -14) {
return 8;
}
}
else if (TSC >= 8) {
if (Rs < 1915) {
return 2;
}
else if (Rs >= 1915) {
return 8;
}
}
}
}
}
else if (ssDCval >= 72) {
if (refDCval < 79) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (mu_mv_mag_sq < 114) {
if (TSC < 7) {
if (Cs < 1290) {
return 1;
}
else if (Cs >= 1290) {
return 8;
}
}
else if (TSC >= 7) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
}
else if (refDCval >= 79) {
if (RsCsDiff < 33) {
if (diffAFD < 1) {
if (Rs < 1749) {
return 8;
}
else if (Rs >= 1749) {
if (Rs < 1768) {
if (RsCsDiff < 17) {
if (Rs < 1762) {
return 8;
}
else if (Rs >= 1762) {
return 1;
}
}
else if (RsCsDiff >= 17) {
return 1;
}
}
else if (Rs >= 1768) {
if (Cs < 695) {
if (Rs < 2286) {
return 1;
}
else if (Rs >= 2286) {
return 8;
}
}
else if (Cs >= 695) {
if (SC < 3579) {
return 8;
}
else if (SC >= 3579) {
if (Cs < 1256) {
return 1;
}
else if (Cs >= 1256) {
return 8;
}
}
}
}
}
}
else if (diffAFD >= 1) {
if (diffMVdiffVal < -1) {
if (Cs < 1142) {
return 8;
}
else if (Cs >= 1142) {
if (Cs < 1242) {
return 2;
}
else if (Cs >= 1242) {
return 8;
}
}
}
else if (diffMVdiffVal >= -1) {
if (Cs < 673) {
return 1;
}
else if (Cs >= 673) {
if (Cs < 1284) {
if (posBalance < 69) {
if (SC < 3596) {
if (negBalance < 26) {
return 2;
}
else if (negBalance >= 26) {
return 8;
}
}
else if (SC >= 3596) {
return 1;
}
}
else if (posBalance >= 69) {
if (Rs < 2112) {
return 1;
}
else if (Rs >= 2112) {
return 8;
}
}
}
else if (Cs >= 1284) {
if (Rs < 1900) {
return 1;
}
else if (Rs >= 1900) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 33) {
if (mu_mv_mag_sq < 112) {
if (ssDCval < 110) {
if (Cs < 1197) {
if (CsDiff < 4) {
if (Rs < 1792) {
if (Rs < 1614) {
return 1;
}
else if (Rs >= 1614) {
return 8;
}
}
else if (Rs >= 1792) {
if (Rs < 2238) {
return 1;
}
else if (Rs >= 2238) {
return 2;
}
}
}
else if (CsDiff >= 4) {
if (mcTcor < 94) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 4;
}
}
else if (mcTcor >= 94) {
if (ssDCval < 86) {
if (Rs < 2248) {
return 8;
}
else if (Rs >= 2248) {
return 2;
}
}
else if (ssDCval >= 86) {
if (mu_mv_mag_sq < 76) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 2145) {
return 4;
}
else if (Rs >= 2145) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (Rs < 2076) {
return 8;
}
else if (Rs >= 2076) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1197) {
if (TSC < 8) {
if (ssDCval < 87) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 1895) {
return 4;
}
else if (Rs >= 1895) {
return 1;
}
}
}
else if (ssDCval >= 87) {
if (MVDiff < 4) {
if (Cs < 1280) {
return 8;
}
else if (Cs >= 1280) {
return 2;
}
}
else if (MVDiff >= 4) {
return 2;
}
}
}
else if (TSC >= 8) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
}
}
else if (ssDCval >= 110) {
if (SC < 2563) {
if (Rs < 1625) {
return 8;
}
else if (Rs >= 1625) {
return 1;
}
}
else if (SC >= 2563) {
if (diffRsCsdiff < 27) {
return 8;
}
else if (diffRsCsdiff >= 27) {
if (Rs < 1858) {
return 2;
}
else if (Rs >= 1858) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
return 1;
}
}
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (refDCval < 114) {
if (Rs < 1903) {
if (MVDiff < 15) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1424) {
if (refDCval < 88) {
if (ssDCval < 57) {
if (diffRsCsdiff < -35) {
return 8;
}
else if (diffRsCsdiff >= -35) {
return 4;
}
}
else if (ssDCval >= 57) {
if (mcTcor < 92) {
if (RsCsDiff < 57) {
return 8;
}
else if (RsCsDiff >= 57) {
if (Rs < 1324) {
return 2;
}
else if (Rs >= 1324) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 74) {
if (diffMVdiffVal < 0) {
return 1;
}
else if (diffMVdiffVal >= 0) {
if (Rs < 1243) {
return 8;
}
else if (Rs >= 1243) {
return 1;
}
}
}
else if (ssDCval >= 74) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
}
else if (refDCval >= 88) {
if (negBalance < 55) {
if (SC < 2206) {
return 8;
}
else if (SC >= 2206) {
return 4;
}
}
else if (negBalance >= 55) {
return 4;
}
}
}
else if (Rs >= 1424) {
if (TSC < 8) {
if (CsDiff < 5) {
if (Rs < 1778) {
return 8;
}
else if (Rs >= 1778) {
return 1;
}
}
else if (CsDiff >= 5) {
if (ssDCval < 93) {
return 4;
}
else if (ssDCval >= 93) {
return 2;
}
}
}
else if (TSC >= 8) {
if (AFD < 17) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Cs < 1129) {
return 1;
}
else if (Cs >= 1129) {
return 8;
}
}
}
else if (AFD >= 17) {
if (SC < 2796) {
if (Rs < 1453) {
return 8;
}
else if (Rs >= 1453) {
return 4;
}
}
else if (SC >= 2796) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (AFD < 23) {
if (Rs < 384) {
if (AFD < 16) {
if (Rs < 219) {
return 4;
}
else if (Rs >= 219) {
return 1;
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (Rs >= 384) {
if (posBalance < 59) {
if (tcor < 67) {
if (Rs < 1316) {
return 4;
}
else if (Rs >= 1316) {
return 2;
}
}
else if (tcor >= 67) {
if (mu_mv_mag_sq < 7) {
if (SC < 1832) {
return 1;
}
else if (SC >= 1832) {
if (Rs < 1663) {
if (ssDCval < 83) {
if (Cs < 925) {
if (Rs < 1374) {
return 8;
}
else if (Rs >= 1374) {
return 2;
}
}
else if (Cs >= 925) {
return 8;
}
}
else if (ssDCval >= 83) {
if (Rs < 1331) {
return 2;
}
else if (Rs >= 1331) {
return 4;
}
}
}
else if (Rs >= 1663) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (negBalance < 52) {
if (AFD < 14) {
if (diffMVdiffVal < 4) {
return 1;
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
else if (AFD >= 14) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
return 8;
}
}
}
else if (negBalance >= 52) {
if (Rs < 850) {
if (Rs < 744) {
return 8;
}
else if (Rs >= 744) {
return 1;
}
}
else if (Rs >= 850) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
if (Rs < 1348) {
return 4;
}
else if (Rs >= 1348) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 59) {
if (Rs < 1119) {
if (ssDCval < 101) {
if (refDCval < 59) {
return 4;
}
else if (refDCval >= 59) {
if (gchDC < 2) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (gchDC >= 2) {
if (Rs < 972) {
return 4;
}
else if (Rs >= 972) {
return 1;
}
}
}
}
else if (ssDCval >= 101) {
if (SC < 1566) {
return 1;
}
else if (SC >= 1566) {
return 4;
}
}
}
else if (Rs >= 1119) {
if (diffAFD < -1) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Cs < 974) {
return 4;
}
else if (Cs >= 974) {
if (diffAFD < -3) {
if (Rs < 1562) {
return 1;
}
else if (Rs >= 1562) {
return 4;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
else if (AFD >= 23) {
if (Rs < 831) {
return 1;
}
else if (Rs >= 831) {
return 4;
}
}
}
}
else if (MVDiff >= 15) {
if (tcor < 86) {
if (mcTcor < 88) {
if (AFD < 27) {
if (Rs < 593) {
if (TSC < 8) {
if (RsCsDiff < 26) {
return 2;
}
else if (RsCsDiff >= 26) {
if (SC < 965) {
if (Rs < 334) {
return 4;
}
else if (Rs >= 334) {
return 8;
}
}
else if (SC >= 965) {
return 4;
}
}
}
else if (TSC >= 8) {
if (ssDCval < 52) {
if (refDCval < 47) {
return 2;
}
else if (refDCval >= 47) {
return 8;
}
}
else if (ssDCval >= 52) {
if (MVDiff < 62) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 1;
}
}
}
else if (MVDiff >= 62) {
if (MVDiff < 83) {
if (Rs < 486) {
if (Cs < 562) {
return 2;
}
else if (Cs >= 562) {
return 8;
}
}
else if (Rs >= 486) {
return 2;
}
}
else if (MVDiff >= 83) {
if (Cs < 991) {
return 4;
}
else if (Cs >= 991) {
return 8;
}
}
}
}
}
}
else if (Rs >= 593) {
if (gchDC < 5) {
if (negBalance < 44) {
if (Rs < 1334) {
if (CsDiff < 10) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (CsDiff >= 10) {
return 4;
}
}
else if (Rs >= 1334) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (negBalance >= 44) {
if (refDCval < 95) {
if (refDCval < 66) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 71) {
if (refDCval < 61) {
if (Cs < 660) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 2;
}
}
else if (Cs >= 660) {
if (TSC < 8) {
if (MVDiff < 28) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 4;
}
}
else if (MVDiff >= 28) {
return 4;
}
}
else if (TSC >= 8) {
if (SC < 1625) {
return 1;
}
else if (SC >= 1625) {
return 4;
}
}
}
}
else if (refDCval >= 61) {
if (MVDiff < 19) {
if (Rs < 920) {
return 4;
}
else if (Rs >= 920) {
return 8;
}
}
else if (MVDiff >= 19) {
return 2;
}
}
}
else if (MVDiff >= 71) {
if (Rs < 652) {
return 1;
}
else if (Rs >= 652) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mu_mv_mag_sq < 24) {
if (ssDCval < 58) {
if (AFD < 14) {
if (mu_mv_mag_sq < 13) {
return 2;
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 726) {
return 4;
}
else if (Rs >= 726) {
return 8;
}
}
}
else if (AFD >= 14) {
if (Rs < 1711) {
if (mu_mv_mag_sq < 7) {
if (Rs < 722) {
return 4;
}
else if (Rs >= 722) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
else if (Rs >= 1711) {
return 2;
}
}
}
else if (ssDCval >= 58) {
if (RsCsDiff < 108) {
if (AFD < 19) {
if (Rs < 806) {
return 1;
}
else if (Rs >= 806) {
if (Rs < 909) {
return 4;
}
else if (Rs >= 909) {
if (Rs < 1186) {
return 2;
}
else if (Rs >= 1186) {
return 8;
}
}
}
}
else if (AFD >= 19) {
return 1;
}
}
else if (RsCsDiff >= 108) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (refDCval < 50) {
return 8;
}
else if (refDCval >= 50) {
if (Cs < 939) {
if (mcTcor < 64) {
return 8;
}
else if (mcTcor >= 64) {
return 4;
}
}
else if (Cs >= 939) {
if (Rs < 1310) {
return 1;
}
else if (Rs >= 1310) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 66) {
if (AFD < 19) {
if (Rs < 689) {
return 4;
}
else if (Rs >= 689) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (Rs < 1473) {
if (Rs < 811) {
return 8;
}
else if (Rs >= 811) {
return 2;
}
}
else if (Rs >= 1473) {
return 8;
}
}
}
}
else if (AFD >= 19) {
if (RsDiff < 10) {
if (tcor < 48) {
if (Cs < 1055) {
return 8;
}
else if (Cs >= 1055) {
return 4;
}
}
else if (tcor >= 48) {
if (refDCval < 68) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
else if (refDCval >= 68) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
if (tcor < 52) {
return 2;
}
else if (tcor >= 52) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 10) {
if (diffMVdiffVal < 9) {
return 4;
}
else if (diffMVdiffVal >= 9) {
return 8;
}
}
}
}
}
else if (refDCval >= 95) {
if (tcor < 75) {
if (AFD < 23) {
return 4;
}
else if (AFD >= 23) {
if (Rs < 1792) {
return 1;
}
else if (Rs >= 1792) {
return 8;
}
}
}
else if (tcor >= 75) {
return 1;
}
}
}
}
else if (gchDC >= 5) {
if (Cs < 728) {
if (RsDiff < 6) {
if (Cs < 716) {
return 8;
}
else if (Cs >= 716) {
return 4;
}
}
else if (RsDiff >= 6) {
if (diffAFD < 1) {
if (Rs < 941) {
return 8;
}
else if (Rs >= 941) {
return 4;
}
}
else if (diffAFD >= 1) {
if (Rs < 1687) {
return 1;
}
else if (Rs >= 1687) {
return 8;
}
}
}
}
else if (Cs >= 728) {
if (MVDiff < 33) {
if (Cs < 1138) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
else if (Cs >= 1138) {
if (Rs < 1632) {
return 1;
}
else if (Rs >= 1632) {
return 2;
}
}
}
else if (MVDiff >= 33) {
if (SC < 1726) {
if (Rs < 709) {
return 2;
}
else if (Rs >= 709) {
return 8;
}
}
else if (SC >= 1726) {
return 8;
}
}
}
}
}
}
else if (AFD >= 27) {
if (Rs < 773) {
if (diffRsCsdiff < -42) {
return 8;
}
else if (diffRsCsdiff >= -42) {
if (diffMVdiffVal < -18) {
return 8;
}
else if (diffMVdiffVal >= -18) {
if (negBalance < 63) {
if (Cs < 648) {
return 4;
}
else if (Cs >= 648) {
return 2;
}
}
else if (negBalance >= 63) {
return 4;
}
}
}
}
else if (Rs >= 773) {
if (Rs < 1103) {
if (RsCsDiff < 102) {
if (AFD < 35) {
return 4;
}
else if (AFD >= 35) {
return 2;
}
}
else if (RsCsDiff >= 102) {
if (tcor < 27) {
if (Rs < 960) {
return 2;
}
else if (Rs >= 960) {
return 4;
}
}
else if (tcor >= 27) {
return 1;
}
}
}
else if (Rs >= 1103) {
if (mu_mv_mag_sq < 31) {
if (posBalance < 75) {
if (Cs < 668) {
return 1;
}
else if (Cs >= 668) {
return 2;
}
}
else if (posBalance >= 75) {
if (MVDiff < 131) {
if (Rs < 1430) {
return 2;
}
else if (Rs >= 1430) {
return 4;
}
}
else if (MVDiff >= 131) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 31) {
return 4;
}
}
}
}
}
else if (mcTcor >= 88) {
if (posBalance < 46) {
if (negBalance < 58) {
if (diffMVdiffVal < 14) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (diffMVdiffVal >= 14) {
return 1;
}
}
else if (negBalance >= 58) {
if (RsCsDiff < 107) {
if (Rs < 485) {
if (Rs < 409) {
return 4;
}
else if (Rs >= 409) {
return 1;
}
}
else if (Rs >= 485) {
if (SC < 2224) {
if (Cs < 1003) {
return 8;
}
else if (Cs >= 1003) {
return 1;
}
}
else if (SC >= 2224) {
return 4;
}
}
}
else if (RsCsDiff >= 107) {
if (diffRsCsdiff < 49) {
return 2;
}
else if (diffRsCsdiff >= 49) {
return 1;
}
}
}
}
else if (posBalance >= 46) {
if (AFD < 26) {
if (mu_mv_mag_sq < 3) {
if (TSC < 9) {
if (refDCval < 103) {
if (refDCval < 67) {
if (diffRsCsdiff < -5) {
if (refDCval < 57) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 4;
}
}
else if (refDCval >= 57) {
if (Cs < 691) {
return 1;
}
else if (Cs >= 691) {
return 8;
}
}
}
else if (diffRsCsdiff >= -5) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 924) {
return 1;
}
else if (Rs >= 924) {
return 4;
}
}
}
}
}
else if (refDCval >= 67) {
if (SC < 1707) {
if (diffRsCsdiff < 5) {
if (Rs < 773) {
return 2;
}
else if (Rs >= 773) {
return 4;
}
}
else if (diffRsCsdiff >= 5) {
if (SC < 1564) {
if (Rs < 738) {
return 8;
}
else if (Rs >= 738) {
return 4;
}
}
else if (SC >= 1564) {
return 1;
}
}
}
else if (SC >= 1707) {
if (diffTSC < -1) {
if (Rs < 1771) {
if (Cs < 703) {
return 8;
}
else if (Cs >= 703) {
return 2;
}
}
else if (Rs >= 1771) {
return 8;
}
}
else if (diffTSC >= -1) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 2;
}
}
}
}
}
else if (refDCval >= 103) {
if (Rs < 1555) {
return 4;
}
else if (Rs >= 1555) {
return 2;
}
}
}
else if (TSC >= 9) {
if (ssDCval < 70) {
if (posBalance < 54) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (posBalance >= 54) {
if (Cs < 868) {
if (Rs < 1022) {
return 8;
}
else if (Rs >= 1022) {
return 4;
}
}
else if (Cs >= 868) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (Rs < 1603) {
if (gchDC < 7) {
if (CsDiff < 6) {
if (negBalance < 51) {
if (Cs < 775) {
return 8;
}
else if (Cs >= 775) {
return 2;
}
}
else if (negBalance >= 51) {
if (MVDiff < 31) {
if (Cs < 1107) {
return 2;
}
else if (Cs >= 1107) {
return 8;
}
}
else if (MVDiff >= 31) {
return 1;
}
}
}
else if (CsDiff >= 6) {
if (negBalance < 50) {
if (Rs < 1349) {
return 1;
}
else if (Rs >= 1349) {
return 2;
}
}
else if (negBalance >= 50) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
if (Rs < 845) {
return 2;
}
else if (Rs >= 845) {
return 8;
}
}
}
}
}
else if (gchDC >= 7) {
return 4;
}
}
else if (Rs >= 1603) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (refDCval < 82) {
if (refDCval < 60) {
if (gchDC < 8) {
if (refDCval < 59) {
if (ssDCval < 48) {
if (Rs < 843) {
return 1;
}
else if (Rs >= 843) {
return 4;
}
}
else if (ssDCval >= 48) {
if (SC < 1187) {
if (Cs < 546) {
return 8;
}
else if (Cs >= 546) {
return 4;
}
}
else if (SC >= 1187) {
if (Rs < 1630) {
if (mu_mv_mag_sq < 7) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
else if (Rs >= 1630) {
return 2;
}
}
}
}
else if (refDCval >= 59) {
if (Rs < 1305) {
return 1;
}
else if (Rs >= 1305) {
if (Rs < 1474) {
return 4;
}
else if (Rs >= 1474) {
return 8;
}
}
}
}
else if (gchDC >= 8) {
return 2;
}
}
else if (refDCval >= 60) {
if (Rs < 1466) {
if (refDCval < 75) {
if (refDCval < 71) {
if (AFD < 16) {
if (Rs < 1064) {
if (gchDC < 3) {
if (Rs < 631) {
return 2;
}
else if (Rs >= 631) {
return 4;
}
}
else if (gchDC >= 3) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
return 1;
}
}
}
else if (Rs >= 1064) {
if (Cs < 780) {
return 2;
}
else if (Cs >= 780) {
return 8;
}
}
}
else if (AFD >= 16) {
if (refDCval < 70) {
if (diffRsCsdiff < -30) {
if (RsDiff < 9) {
if (Cs < 664) {
return 1;
}
else if (Cs >= 664) {
return 8;
}
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (diffRsCsdiff >= -30) {
if (Cs < 584) {
if (Rs < 572) {
return 8;
}
else if (Rs >= 572) {
return 4;
}
}
else if (Cs >= 584) {
if (AFD < 17) {
if (refDCval < 68) {
return 4;
}
else if (refDCval >= 68) {
if (Rs < 743) {
return 2;
}
else if (Rs >= 743) {
return 8;
}
}
}
else if (AFD >= 17) {
return 4;
}
}
}
}
else if (refDCval >= 70) {
if (Rs < 1232) {
return 8;
}
else if (Rs >= 1232) {
return 4;
}
}
}
}
else if (refDCval >= 71) {
if (Rs < 764) {
if (CsDiff < 8) {
if (tcor < 76) {
return 2;
}
else if (tcor >= 76) {
if (Rs < 685) {
return 1;
}
else if (Rs >= 685) {
return 4;
}
}
}
else if (CsDiff >= 8) {
return 8;
}
}
else if (Rs >= 764) {
if (diffMVdiffVal < -7) {
if (Cs < 1011) {
return 1;
}
else if (Cs >= 1011) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
if (RsDiff < 10) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (RsDiff >= 10) {
return 8;
}
}
}
}
}
else if (refDCval >= 75) {
if (SC < 1717) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 8;
}
}
else if (SC >= 1717) {
if (MVDiff < 18) {
if (SC < 2292) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
return 4;
}
}
else if (SC >= 2292) {
return 2;
}
}
else if (MVDiff >= 18) {
return 4;
}
}
}
}
else if (Rs >= 1466) {
if (diffMVdiffVal < -3) {
if (AFD < 19) {
if (diffRsCsdiff < 3) {
if (MVDiff < 56) {
if (Rs < 1498) {
return 4;
}
else if (Rs >= 1498) {
return 8;
}
}
else if (MVDiff >= 56) {
if (Rs < 1686) {
return 1;
}
else if (Rs >= 1686) {
return 4;
}
}
}
else if (diffRsCsdiff >= 3) {
if (Rs < 1877) {
return 2;
}
else if (Rs >= 1877) {
return 4;
}
}
}
else if (AFD >= 19) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (diffMVdiffVal >= -3) {
if (tcor < 63) {
if (Cs < 853) {
return 8;
}
else if (Cs >= 853) {
return 2;
}
}
else if (tcor >= 63) {
if (mcTcor < 93) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 4;
}
}
else if (mcTcor >= 93) {
if (Rs < 1823) {
return 1;
}
else if (Rs >= 1823) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 82) {
if (negBalance < 50) {
if (Rs < 660) {
return 2;
}
else if (Rs >= 660) {
if (RsDiff < 10) {
if (MVDiff < 19) {
return 1;
}
else if (MVDiff >= 19) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
if (Cs < 784) {
return 2;
}
else if (Cs >= 784) {
return 8;
}
}
}
}
else if (RsDiff >= 10) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
}
}
else if (negBalance >= 50) {
if (mu_mv_mag_sq < 24) {
if (refDCval < 88) {
if (tcor < 72) {
if (Rs < 1639) {
return 4;
}
else if (Rs >= 1639) {
if (Rs < 1688) {
return 8;
}
else if (Rs >= 1688) {
return 1;
}
}
}
else if (tcor >= 72) {
if (tcor < 81) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
}
else if (tcor >= 81) {
if (Rs < 1531) {
return 8;
}
else if (Rs >= 1531) {
return 4;
}
}
}
}
else if (refDCval >= 88) {
if (mu_mv_mag_sq < 17) {
if (TSC < 10) {
if (SC < 2545) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
if (Rs < 1043) {
return 2;
}
else if (Rs >= 1043) {
return 8;
}
}
}
else if (SC >= 2545) {
if (Cs < 1176) {
return 1;
}
else if (Cs >= 1176) {
if (Rs < 1383) {
return 2;
}
else if (Rs >= 1383) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (SC < 2701) {
return 1;
}
else if (SC >= 2701) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (diffAFD < 2) {
if (refDCval < 101) {
return 1;
}
else if (refDCval >= 101) {
return 2;
}
}
else if (diffAFD >= 2) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (gchDC < 2) {
if (posBalance < 60) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 4;
}
}
}
else if (posBalance >= 60) {
if (Rs < 1147) {
return 8;
}
else if (Rs >= 1147) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (Cs < 1054) {
if (refDCval < 102) {
return 8;
}
else if (refDCval >= 102) {
if (Rs < 1532) {
return 2;
}
else if (Rs >= 1532) {
return 8;
}
}
}
else if (Cs >= 1054) {
if (Rs < 1622) {
return 4;
}
else if (Rs >= 1622) {
return 8;
}
}
}
}
}
}
}
}
else if (AFD >= 26) {
if (MVDiff < 113) {
if (RsCsDiff < 133) {
if (mcTcor < 94) {
if (ssDCval < 88) {
if (Cs < 967) {
if (Rs < 1202) {
return 4;
}
else if (Rs >= 1202) {
return 2;
}
}
else if (Cs >= 967) {
return 2;
}
}
else if (ssDCval >= 88) {
if (Rs < 1060) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
if (MVDiff < 101) {
return 2;
}
else if (MVDiff >= 101) {
return 1;
}
}
}
else if (Rs >= 1060) {
return 1;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 718) {
return 1;
}
else if (Rs >= 718) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
return 2;
}
}
}
}
else if (RsCsDiff >= 133) {
if (mu_mv_mag_sq < 7) {
if (MVDiff < 67) {
return 2;
}
else if (MVDiff >= 67) {
return 4;
}
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 32) {
return 4;
}
else if (AFD >= 32) {
return 2;
}
}
}
}
else if (MVDiff >= 113) {
if (RsDiff < 7) {
if (Rs < 610) {
if (Rs < 426) {
return 2;
}
else if (Rs >= 426) {
return 1;
}
}
else if (Rs >= 610) {
return 4;
}
}
else if (RsDiff >= 7) {
if (posBalance < 65) {
if (Rs < 821) {
return 8;
}
else if (Rs >= 821) {
return 2;
}
}
else if (posBalance >= 65) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 86) {
if (Cs < 924) {
if (ssDCval < 93) {
if (mu_mv_mag_sq < 5) {
if (posBalance < 61) {
if (Cs < 640) {
return 1;
}
else if (Cs >= 640) {
if (SC < 2510) {
return 8;
}
else if (SC >= 2510) {
return 1;
}
}
}
else if (posBalance >= 61) {
return 4;
}
}
else if (mu_mv_mag_sq >= 5) {
if (diffMVdiffVal < -25) {
if (Rs < 622) {
return 1;
}
else if (Rs >= 622) {
return 2;
}
}
else if (diffMVdiffVal >= -25) {
return 8;
}
}
}
else if (ssDCval >= 93) {
if (mcTcor < 93) {
return 1;
}
else if (mcTcor >= 93) {
if (refDCval < 98) {
return 1;
}
else if (refDCval >= 98) {
return 2;
}
}
}
}
else if (Cs >= 924) {
if (TSC < 9) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (AFD < 15) {
if (Rs < 1478) {
return 2;
}
else if (Rs >= 1478) {
return 8;
}
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
else if (Rs >= 1903) {
if (refDCval < 72) {
if (Cs < 1062) {
if (tcor < 50) {
if (Rs < 1971) {
return 4;
}
else if (Rs >= 1971) {
return 1;
}
}
else if (tcor >= 50) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (TSC < 12) {
if (RsDiff < 8) {
if (Rs < 2141) {
return 2;
}
else if (Rs >= 2141) {
return 4;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (TSC >= 12) {
return 4;
}
}
}
}
else if (Cs >= 1062) {
if (mcTcor < 89) {
if (negBalance < 43) {
return 2;
}
else if (negBalance >= 43) {
return 8;
}
}
else if (mcTcor >= 89) {
if (gchDC < 4) {
return 1;
}
else if (gchDC >= 4) {
if (Cs < 1260) {
return 4;
}
else if (Cs >= 1260) {
return 8;
}
}
}
}
}
else if (refDCval >= 72) {
if (mcTcor < 68) {
if (Cs < 803) {
return 1;
}
else if (Cs >= 803) {
return 8;
}
}
else if (mcTcor >= 68) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (gchDC < 3) {
if (mcTcor < 90) {
if (RsDiff < 12) {
if (Cs < 1059) {
if (MVDiff < 26) {
if (Cs < 678) {
return 2;
}
else if (Cs >= 678) {
return 1;
}
}
else if (MVDiff >= 26) {
return 8;
}
}
else if (Cs >= 1059) {
if (AFD < 19) {
if (diffRsCsdiff < -4) {
return 8;
}
else if (diffRsCsdiff >= -4) {
return 4;
}
}
else if (AFD >= 19) {
if (Cs < 1228) {
return 2;
}
else if (Cs >= 1228) {
return 4;
}
}
}
}
else if (RsDiff >= 12) {
if (mu_mv_mag_sq < 18) {
if (SC < 3122) {
if (RsDiff < 15) {
if (MVDiff < 32) {
if (Rs < 2152) {
return 8;
}
else if (Rs >= 2152) {
return 1;
}
}
else if (MVDiff >= 32) {
return 1;
}
}
else if (RsDiff >= 15) {
return 4;
}
}
else if (SC >= 3122) {
if (AFD < 26) {
return 4;
}
else if (AFD >= 26) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (AFD < 32) {
return 4;
}
else if (AFD >= 32) {
return 8;
}
}
}
}
else if (mcTcor >= 90) {
if (SC < 2723) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
return 1;
}
}
else if (SC >= 2723) {
if (negBalance < 57) {
if (tcor < 88) {
if (mcTcor < 93) {
if (ssDCval < 76) {
if (Rs < 2115) {
return 8;
}
else if (Rs >= 2115) {
return 2;
}
}
else if (ssDCval >= 76) {
if (tcor < 76) {
if (mcTcor < 91) {
if (Rs < 1915) {
return 2;
}
else if (Rs >= 1915) {
return 4;
}
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (tcor >= 76) {
if (negBalance < 53) {
if (CsDiff < 5) {
if (Rs < 2106) {
return 2;
}
else if (Rs >= 2106) {
return 4;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (negBalance >= 53) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
}
}
}
else if (mcTcor >= 93) {
if (tcor < 85) {
if (Cs < 1032) {
if (MVDiff < 23) {
return 2;
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (Cs >= 1032) {
if (tcor < 69) {
return 2;
}
else if (tcor >= 69) {
return 4;
}
}
}
else if (tcor >= 85) {
return 2;
}
}
}
else if (tcor >= 88) {
if (Cs < 889) {
return 4;
}
else if (Cs >= 889) {
return 8;
}
}
}
else if (negBalance >= 57) {
if (SC < 2753) {
return 8;
}
else if (SC >= 2753) {
if (Rs < 1925) {
if (Rs < 1906) {
return 2;
}
else if (Rs >= 1906) {
return 1;
}
}
else if (Rs >= 1925) {
return 2;
}
}
}
}
}
}
else if (gchDC >= 3) {
if (MVDiff < 39) {
if (SC < 3275) {
if (mu_mv_mag_sq < 39) {
if (RsCsDiff < 131) {
if (Rs < 2148) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
return 1;
}
}
else if (Rs >= 2148) {
return 8;
}
}
else if (RsCsDiff >= 131) {
if (diffRsCsdiff < 97) {
return 2;
}
else if (diffRsCsdiff >= 97) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (Rs < 2072) {
return 8;
}
else if (Rs >= 2072) {
return 1;
}
}
}
else if (SC >= 3275) {
if (AFD < 20) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
else if (AFD >= 20) {
if (MVDiff < 28) {
return 4;
}
else if (MVDiff >= 28) {
return 2;
}
}
}
}
else if (MVDiff >= 39) {
if (Cs < 1058) {
if (Cs < 761) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
if (SC < 2690) {
return 8;
}
else if (SC >= 2690) {
return 2;
}
}
}
else if (Cs >= 761) {
return 8;
}
}
else if (Cs >= 1058) {
if (Rs < 2294) {
if (Cs < 1167) {
return 4;
}
else if (Cs >= 1167) {
return 8;
}
}
else if (Rs >= 2294) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (MVDiff < 17) {
if (Rs < 551) {
return 2;
}
else if (Rs >= 551) {
if (diffAFD < 4) {
if (tcor < 72) {
if (mcTcor < 91) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 2;
}
}
else if (mcTcor >= 91) {
if (RsCsDiff < 287) {
return 8;
}
else if (RsCsDiff >= 287) {
return 2;
}
}
}
else if (tcor >= 72) {
if (tcor < 86) {
if (ssDCval < 169) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
return 8;
}
}
else if (ssDCval >= 169) {
return 2;
}
}
else if (tcor >= 86) {
if (MVDiff < 12) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 1308) {
return 2;
}
else if (Rs >= 1308) {
return 4;
}
}
}
else if (MVDiff >= 12) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (Rs < 1431) {
return 2;
}
else if (Rs >= 1431) {
return 8;
}
}
}
}
}
}
else if (diffAFD >= 4) {
if (ssDCval < 129) {
if (Rs < 1880) {
return 8;
}
else if (Rs >= 1880) {
return 1;
}
}
else if (ssDCval >= 129) {
return 2;
}
}
}
}
else if (MVDiff >= 17) {
if (mu_mv_mag_sq < 200) {
if (tcor < 79) {
if (mcTcor < 73) {
if (diffMVdiffVal < 26) {
return 1;
}
else if (diffMVdiffVal >= 26) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
}
else if (mcTcor >= 73) {
if (SC < 1509) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (RsDiff < 7) {
if (MVDiff < 151) {
return 2;
}
else if (MVDiff >= 151) {
return 1;
}
}
else if (RsDiff >= 7) {
return 8;
}
}
}
else if (SC >= 1509) {
if (AFD < 32) {
if (negBalance < 47) {
if (RsDiff < 10) {
if (diffRsCsdiff < 57) {
return 8;
}
else if (diffRsCsdiff >= 57) {
return 1;
}
}
else if (RsDiff >= 10) {
if (Rs < 1297) {
return 8;
}
else if (Rs >= 1297) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (negBalance < 52) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
if (Cs < 606) {
return 1;
}
else if (Cs >= 606) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (refDCval < 131) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 2;
}
}
else if (refDCval >= 131) {
if (refDCval < 134) {
if (Rs < 1796) {
return 4;
}
else if (Rs >= 1796) {
return 1;
}
}
else if (refDCval >= 134) {
if (mu_mv_mag_sq < 10) {
if (diffRsCsdiff < 10) {
if (Rs < 1095) {
return 2;
}
else if (Rs >= 1095) {
return 8;
}
}
else if (diffRsCsdiff >= 10) {
if (SC < 2332) {
return 1;
}
else if (SC >= 2332) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 32) {
if (refDCval < 134) {
return 2;
}
else if (refDCval >= 134) {
if (mu_mv_mag_sq < 10) {
if (Rs < 975) {
return 4;
}
else if (Rs >= 975) {
return 2;
}
}
else if (mu_mv_mag_sq >= 10) {
return 1;
}
}
}
}
}
}
else if (tcor >= 79) {
if (SC < 2186) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (refDCval < 145) {
if (Rs < 1506) {
return 8;
}
else if (Rs >= 1506) {
return 1;
}
}
else if (refDCval >= 145) {
if (Cs < 1083) {
return 2;
}
else if (Cs >= 1083) {
if (Rs < 982) {
return 1;
}
else if (Rs >= 982) {
return 8;
}
}
}
}
}
else if (SC >= 2186) {
if (ssDCval < 150) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
if (RsDiff < 9) {
if (TSC < 10) {
if (Rs < 1294) {
return 1;
}
else if (Rs >= 1294) {
return 8;
}
}
else if (TSC >= 10) {
if (MVDiff < 58) {
return 1;
}
else if (MVDiff >= 58) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (TSC < 8) {
if (Cs < 921) {
return 1;
}
else if (Cs >= 921) {
return 2;
}
}
else if (TSC >= 8) {
if (mu_mv_mag_sq < 22) {
return 1;
}
else if (mu_mv_mag_sq >= 22) {
if (Rs < 1877) {
return 8;
}
else if (Rs >= 1877) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 150) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 200) {
return 1;
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (SC < 2170) {
if (Rs < 964) {
if (refDCval < 64) {
if (MVDiff < 1) {
if (Rs < 528) {
if (SC < 1073) {
if (Rs < 394) {
return 1;
}
else if (Rs >= 394) {
return 8;
}
}
else if (SC >= 1073) {
if (Rs < 388) {
if (Cs < 921) {
if (Rs < 325) {
return 4;
}
else if (Rs >= 325) {
return 1;
}
}
else if (Cs >= 921) {
return 2;
}
}
else if (Rs >= 388) {
if (Rs < 520) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (Rs >= 520) {
return 2;
}
}
}
}
else if (Rs >= 528) {
if (SC < 1656) {
if (ssDCval < 57) {
if (mu_mv_mag_sq < 120) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (refDCval < 54) {
return 4;
}
else if (refDCval >= 54) {
if (Rs < 858) {
return 8;
}
else if (Rs >= 858) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 556) {
return 4;
}
else if (Rs >= 556) {
return 8;
}
}
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (SC >= 1656) {
if (Rs < 648) {
if (Cs < 1067) {
return 1;
}
else if (Cs >= 1067) {
return 4;
}
}
else if (Rs >= 648) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (diffRsCsdiff < -5) {
return 8;
}
else if (diffRsCsdiff >= -5) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (gchDC < 1) {
if (mu_mv_mag_sq < 123) {
return 1;
}
else if (mu_mv_mag_sq >= 123) {
if (Cs < 1090) {
return 8;
}
else if (Cs >= 1090) {
return 1;
}
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 17) {
if (ssDCval < 43) {
return 2;
}
else if (ssDCval >= 43) {
if (Rs < 291) {
return 4;
}
else if (Rs >= 291) {
return 1;
}
}
}
else if (negBalance >= 17) {
if (Rs < 780) {
if (SC < 1007) {
if (Cs < 548) {
return 4;
}
else if (Cs >= 548) {
if (posBalance < 40) {
if (Rs < 349) {
return 1;
}
else if (Rs >= 349) {
return 4;
}
}
else if (posBalance >= 40) {
if (AFD < 15) {
if (posBalance < 46) {
if (Rs < 309) {
return 1;
}
else if (Rs >= 309) {
return 8;
}
}
else if (posBalance >= 46) {
return 1;
}
}
else if (AFD >= 15) {
return 8;
}
}
}
}
else if (SC >= 1007) {
if (AFD < 7) {
if (negBalance < 29) {
if (Rs < 533) {
if (ssDCval < 57) {
if (Rs < 396) {
return 2;
}
else if (Rs >= 396) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (ssDCval >= 57) {
if (Rs < 406) {
return 8;
}
else if (Rs >= 406) {
return 2;
}
}
}
else if (Rs >= 533) {
if (refDCval < 51) {
if (Rs < 724) {
return 1;
}
else if (Rs >= 724) {
return 4;
}
}
else if (refDCval >= 51) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (Cs < 1144) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (Rs < 578) {
return 8;
}
else if (Rs >= 578) {
return 4;
}
}
}
else if (Cs >= 1144) {
return 1;
}
}
}
}
}
else if (negBalance >= 29) {
if (diffMVdiffVal < -6) {
if (Rs < 606) {
return 1;
}
else if (Rs >= 606) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
if (CsDiff < 2) {
if (posBalance < 43) {
if (diffRsCsdiff < -5) {
return 2;
}
else if (diffRsCsdiff >= -5) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
return 4;
}
}
}
else if (posBalance >= 43) {
if (ssDCval < 62) {
if (TSC < 3) {
if (Rs < 578) {
return 4;
}
else if (Rs >= 578) {
return 8;
}
}
else if (TSC >= 3) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
return 4;
}
}
}
else if (ssDCval >= 62) {
if (Cs < 900) {
return 8;
}
else if (Cs >= 900) {
return 2;
}
}
}
}
else if (CsDiff >= 2) {
if (CsDiff < 3) {
if (Rs < 470) {
return 4;
}
else if (Rs >= 470) {
return 8;
}
}
else if (CsDiff >= 3) {
if (SC < 1554) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 8;
}
}
else if (SC >= 1554) {
if (Rs < 733) {
return 2;
}
else if (Rs >= 733) {
return 4;
}
}
}
}
}
}
}
else if (AFD >= 7) {
if (SC < 1505) {
if (SC < 1220) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
else if (SC >= 1220) {
if (CsDiff < 3) {
if (gchDC < 1) {
if (Rs < 664) {
return 4;
}
else if (Rs >= 664) {
return 8;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (CsDiff >= 3) {
if (tcor < 93) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (Rs < 744) {
return 1;
}
else if (Rs >= 744) {
return 4;
}
}
}
}
else if (tcor >= 93) {
return 4;
}
}
}
}
else if (SC >= 1505) {
if (Cs < 886) {
return 8;
}
else if (Cs >= 886) {
return 2;
}
}
}
}
}
else if (Rs >= 780) {
if (ssDCval < 51) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (ssDCval < 41) {
return 8;
}
else if (ssDCval >= 41) {
if (Cs < 1114) {
return 4;
}
else if (Cs >= 1114) {
return 1;
}
}
}
}
else if (ssDCval >= 51) {
if (negBalance < 41) {
if (RsCsDiff < 7) {
return 1;
}
else if (RsCsDiff >= 7) {
if (Cs < 805) {
if (MVDiff < 4) {
if (Rs < 868) {
return 4;
}
else if (Rs >= 868) {
return 1;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (Cs >= 805) {
return 4;
}
}
}
else if (negBalance >= 41) {
if (TSC < 3) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (TSC >= 3) {
if (Cs < 1027) {
return 8;
}
else if (Cs >= 1027) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 64) {
if (Rs < 815) {
if (refDCval < 82) {
if (RsCsDiff < 12) {
if (Rs < 728) {
if (SC < 1155) {
if (Rs < 337) {
return 1;
}
else if (Rs >= 337) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 8;
}
}
}
else if (SC >= 1155) {
if (RsDiff < 1) {
if (ssDCval < 65) {
if (Rs < 677) {
return 8;
}
else if (Rs >= 677) {
return 4;
}
}
else if (ssDCval >= 65) {
if (Cs < 674) {
if (Rs < 599) {
return 8;
}
else if (Rs >= 599) {
return 1;
}
}
else if (Cs >= 674) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (Rs < 645) {
if (negBalance < 26) {
if (Cs < 994) {
return 4;
}
else if (Cs >= 994) {
return 8;
}
}
else if (negBalance >= 26) {
if (diffMVdiffVal < -4) {
if (Cs < 845) {
return 8;
}
else if (Cs >= 845) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (Rs >= 645) {
if (negBalance < 42) {
if (ssDCval < 65) {
if (Rs < 686) {
return 8;
}
else if (Rs >= 686) {
return 4;
}
}
else if (ssDCval >= 65) {
if (AFD < 7) {
if (AFD < 3) {
if (Rs < 718) {
return 8;
}
else if (Rs >= 718) {
return 1;
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (negBalance >= 42) {
if (Rs < 677) {
return 2;
}
else if (Rs >= 677) {
return 8;
}
}
}
}
}
}
else if (Rs >= 728) {
if (Cs < 632) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 4;
}
}
else if (Cs >= 632) {
if (MVDiff < 1) {
if (ssDCval < 66) {
if (Rs < 777) {
return 4;
}
else if (Rs >= 777) {
return 8;
}
}
else if (ssDCval >= 66) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (Rs < 767) {
return 1;
}
else if (Rs >= 767) {
return 4;
}
}
}
}
else if (MVDiff >= 1) {
if (posBalance < 57) {
if (Cs < 1026) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Cs >= 1026) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 12) {
if (posBalance < 60) {
if (Cs < 797) {
if (MVDiff < 7) {
if (Cs < 743) {
return 2;
}
else if (Cs >= 743) {
if (Rs < 756) {
return 1;
}
else if (Rs >= 756) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (Cs >= 797) {
if (Rs < 511) {
if (Cs < 1242) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (Cs >= 1242) {
return 1;
}
}
else if (Rs >= 511) {
return 2;
}
}
}
else if (posBalance >= 60) {
if (Rs < 552) {
return 1;
}
else if (Rs >= 552) {
if (Cs < 1116) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
return 1;
}
}
else if (Cs >= 1116) {
return 2;
}
}
}
}
}
else if (refDCval >= 82) {
if (TSC < 3) {
if (Rs < 652) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
if (Cs < 743) {
if (SC < 1272) {
return 2;
}
else if (SC >= 1272) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
else if (Cs >= 743) {
return 2;
}
}
}
else if (Rs >= 652) {
if (Cs < 863) {
return 1;
}
else if (Cs >= 863) {
if (Rs < 691) {
return 2;
}
else if (Rs >= 691) {
return 4;
}
}
}
}
else if (TSC >= 3) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (TSC < 5) {
if (Rs < 433) {
return 8;
}
else if (Rs >= 433) {
if (negBalance < 43) {
if (posBalance < 43) {
if (Rs < 595) {
return 1;
}
else if (Rs >= 595) {
return 2;
}
}
else if (posBalance >= 43) {
return 2;
}
}
else if (negBalance >= 43) {
if (Rs < 663) {
return 4;
}
else if (Rs >= 663) {
return 1;
}
}
}
}
else if (TSC >= 5) {
if (Cs < 662) {
return 8;
}
else if (Cs >= 662) {
return 4;
}
}
}
}
}
}
else if (Rs >= 815) {
if (Cs < 761) {
if (mu_mv_mag_sq < 106) {
if (tcor < 93) {
if (Rs < 840) {
return 2;
}
else if (Rs >= 840) {
return 4;
}
}
else if (tcor >= 93) {
if (tcor < 95) {
return 1;
}
else if (tcor >= 95) {
if (negBalance < 31) {
if (Rs < 913) {
if (Rs < 870) {
return 2;
}
else if (Rs >= 870) {
return 1;
}
}
else if (Rs >= 913) {
return 8;
}
}
else if (negBalance >= 31) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (Cs >= 761) {
if (Cs < 1098) {
if (Rs < 874) {
if (SC < 1818) {
return 4;
}
else if (SC >= 1818) {
if (gchDC < 1) {
if (SC < 1837) {
if (Rs < 854) {
return 2;
}
else if (Rs >= 854) {
return 8;
}
}
else if (SC >= 1837) {
if (SC < 1863) {
return 4;
}
else if (SC >= 1863) {
return 1;
}
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (Rs >= 874) {
if (ssDCval < 66) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
else if (ssDCval >= 66) {
if (MVDiff < 3) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (RsDiff < 2) {
if (Rs < 957) {
return 4;
}
else if (Rs >= 957) {
return 8;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
else if (Cs >= 1098) {
if (Cs < 1273) {
if (ssDCval < 75) {
if (Cs < 1187) {
if (Rs < 934) {
return 2;
}
else if (Rs >= 934) {
return 4;
}
}
else if (Cs >= 1187) {
return 2;
}
}
else if (ssDCval >= 75) {
return 4;
}
}
else if (Cs >= 1273) {
return 1;
}
}
}
}
}
}
else if (Rs >= 964) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 51) {
if (mu_mv_mag_sq < 77) {
if (Rs < 1272) {
return 1;
}
else if (Rs >= 1272) {
return 4;
}
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 1165) {
return 8;
}
else if (Rs >= 1165) {
if (AFD < 5) {
if (diffRsCsdiff < 2) {
if (Rs < 1204) {
return 4;
}
else if (Rs >= 1204) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (ssDCval >= 51) {
if (ssDCval < 69) {
if (posBalance < 55) {
if (SC < 1898) {
if (mu_mv_mag_sq < 120) {
if (MVDiff < 13) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (negBalance < 28) {
if (Cs < 724) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 724) {
return 8;
}
}
else if (negBalance >= 28) {
if (Rs < 1202) {
return 1;
}
else if (Rs >= 1202) {
return 8;
}
}
}
}
else if (MVDiff >= 13) {
if (SC < 1809) {
if (Cs < 548) {
return 4;
}
else if (Cs >= 548) {
return 1;
}
}
else if (SC >= 1809) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (posBalance < 43) {
if (Rs < 1074) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (Rs >= 1074) {
return 1;
}
}
else if (posBalance >= 43) {
if (tcor < 99) {
if (Cs < 739) {
return 4;
}
else if (Cs >= 739) {
return 1;
}
}
else if (tcor >= 99) {
return 2;
}
}
}
}
else if (SC >= 1898) {
if (Cs < 1088) {
if (negBalance < 41) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (Cs < 711) {
return 1;
}
else if (Cs >= 711) {
if (Cs < 919) {
if (negBalance < 14) {
return 8;
}
else if (negBalance >= 14) {
if (mcTcor < 98) {
if (ssDCval < 63) {
if (mu_mv_mag_sq < 121) {
if (SC < 2135) {
return 1;
}
else if (SC >= 2135) {
return 8;
}
}
else if (mu_mv_mag_sq >= 121) {
return 8;
}
}
else if (ssDCval >= 63) {
return 8;
}
}
else if (mcTcor >= 98) {
if (Rs < 1241) {
return 1;
}
else if (Rs >= 1241) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
return 1;
}
}
}
}
}
else if (Cs >= 919) {
return 1;
}
}
}
}
else if (negBalance >= 41) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 1009) {
return 1;
}
else if (Rs >= 1009) {
return 4;
}
}
}
}
else if (Cs >= 1088) {
if (ssDCval < 62) {
if (Rs < 978) {
return 4;
}
else if (Rs >= 978) {
return 1;
}
}
else if (ssDCval >= 62) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
else if (posBalance >= 55) {
if (mu_mv_mag_sq < 79) {
if (diffAFD < 1) {
if (diffMVdiffVal < -4) {
return 2;
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 79) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
return 8;
}
}
}
}
else if (ssDCval >= 69) {
if (Cs < 700) {
if (ssDCval < 72) {
if (Rs < 1353) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Rs >= 1353) {
if (Cs < 654) {
return 8;
}
else if (Cs >= 654) {
return 1;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 1058) {
return 8;
}
else if (Rs >= 1058) {
if (SC < 2108) {
if (posBalance < 23) {
return 8;
}
else if (posBalance >= 23) {
return 1;
}
}
else if (SC >= 2108) {
if (mu_mv_mag_sq < 118) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
}
}
}
else if (Cs >= 700) {
if (AFD < 4) {
if (MVDiff < 2) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
if (Cs < 934) {
if (Rs < 1177) {
return 8;
}
else if (Rs >= 1177) {
return 1;
}
}
else if (Cs >= 934) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
return 4;
}
}
}
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 4) {
if (Rs < 1051) {
return 8;
}
else if (Rs >= 1051) {
return 2;
}
}
else if (mu_mv_mag_sq >= 4) {
if (negBalance < 68) {
if (Rs < 1286) {
return 8;
}
else if (Rs >= 1286) {
return 4;
}
}
else if (negBalance >= 68) {
if (Cs < 808) {
return 8;
}
else if (Cs >= 808) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 686) {
if (ssDCval < 74) {
if (posBalance < 37) {
if (Rs < 1309) {
return 1;
}
else if (Rs >= 1309) {
return 8;
}
}
else if (posBalance >= 37) {
return 8;
}
}
else if (ssDCval >= 74) {
return 1;
}
}
else if (Cs >= 686) {
if (Cs < 757) {
if (Rs < 1152) {
if (SC < 1837) {
if (Cs < 716) {
return 8;
}
else if (Cs >= 716) {
return 2;
}
}
else if (SC >= 1837) {
return 4;
}
}
else if (Rs >= 1152) {
return 1;
}
}
else if (Cs >= 757) {
if (SC < 2133) {
return 8;
}
else if (SC >= 2133) {
if (Rs < 1251) {
return 1;
}
else if (Rs >= 1251) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 2170) {
if (Rs < 1815) {
if (Cs < 1063) {
if (mu_mv_mag_sq < 122) {
if (Cs < 936) {
if (SC < 2553) {
if (Rs < 1328) {
return 1;
}
else if (Rs >= 1328) {
if (Cs < 808) {
if (mu_mv_mag_sq < 116) {
if (TSC < 5) {
if (ssDCval < 56) {
if (Rs < 1539) {
return 1;
}
else if (Rs >= 1539) {
return 4;
}
}
else if (ssDCval >= 56) {
if (ssDCval < 85) {
return 1;
}
else if (ssDCval >= 85) {
if (Rs < 1803) {
return 4;
}
else if (Rs >= 1803) {
return 1;
}
}
}
}
else if (TSC >= 5) {
if (MVDiff < 18) {
if (Rs < 1664) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (Rs >= 1664) {
return 1;
}
}
else if (MVDiff >= 18) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 116) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 8;
}
}
}
else if (Cs >= 808) {
if (mu_mv_mag_sq < 65) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
if (Rs < 1397) {
return 4;
}
else if (Rs >= 1397) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 65) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
if (Cs < 933) {
return 8;
}
else if (Cs >= 933) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 2553) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (Cs >= 936) {
if (TSCindex < 3) {
if (Rs < 1774) {
return 1;
}
else if (Rs >= 1774) {
return 4;
}
}
else if (TSCindex >= 3) {
if (ssDCval < 52) {
if (Rs < 1284) {
return 4;
}
else if (Rs >= 1284) {
return 1;
}
}
else if (ssDCval >= 52) {
if (posBalance < 56) {
if (Rs < 1781) {
if (Rs < 1141) {
return 1;
}
else if (Rs >= 1141) {
return 8;
}
}
else if (Rs >= 1781) {
return 4;
}
}
else if (posBalance >= 56) {
if (MVDiff < 2) {
if (Rs < 1427) {
return 8;
}
else if (Rs >= 1427) {
return 2;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (ssDCval < 71) {
if (tcor < 99) {
if (Rs < 1569) {
if (Rs < 1242) {
return 4;
}
else if (Rs >= 1242) {
return 8;
}
}
else if (Rs >= 1569) {
if (Cs < 979) {
return 1;
}
else if (Cs >= 979) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
if (diffAFD < 1) {
if (Rs < 1805) {
return 4;
}
else if (Rs >= 1805) {
return 4;
}
}
else if (diffAFD >= 1) {
if (Rs < 1730) {
return 4;
}
else if (Rs >= 1730) {
return 8;
}
}
}
}
}
}
else if (tcor >= 99) {
if (ssDCval < 61) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (Cs < 1050) {
if (gchDC < 1) {
if (Rs < 1742) {
if (Rs < 1503) {
return 4;
}
else if (Rs >= 1503) {
return 4;
}
}
else if (Rs >= 1742) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Cs >= 1050) {
return 8;
}
}
}
}
else if (ssDCval >= 61) {
if (posBalance < 33) {
if (SC < 2268) {
return 8;
}
else if (SC >= 2268) {
return 1;
}
}
else if (posBalance >= 33) {
return 1;
}
}
}
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (Cs >= 1063) {
if (RsDiff < 1) {
if (MVDiff < 12) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (MVDiff >= 12) {
return 4;
}
}
else if (RsDiff >= 1) {
if (TSC < 3) {
if (ssDCval < 52) {
if (SC < 2952) {
if (Rs < 1632) {
if (MVDiff < 4) {
if (Cs < 1161) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (Cs >= 1161) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (Rs >= 1632) {
if (Rs < 1753) {
if (SC < 2757) {
return 8;
}
else if (SC >= 2757) {
if (ssDCval < 50) {
if (Cs < 1155) {
if (SC < 2768) {
return 4;
}
else if (SC >= 2768) {
if (Rs < 1689) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1133) {
if (Rs < 1667) {
return 4;
}
else if (Rs >= 1667) {
if (Rs < 1679) {
return 4;
}
else if (Rs >= 1679) {
if (Rs < 1683) {
return 4;
}
else if (Rs >= 1683) {
return 4;
}
}
}
}
else if (Cs >= 1133) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 1680) {
if (Rs < 1656) {
return 8;
}
else if (Rs >= 1656) {
return 8;
}
}
else if (Rs >= 1680) {
return 8;
}
}
}
}
}
else if (Rs >= 1689) {
if (diffRsCsdiff < 1) {
if (Cs < 1139) {
if (Cs < 1065) {
return 4;
}
else if (Cs >= 1065) {
if (SC < 2849) {
if (negBalance < 17) {
return 4;
}
else if (negBalance >= 17) {
if (Cs < 1079) {
return 4;
}
else if (Cs >= 1079) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (Rs < 1714) {
return 4;
}
else if (Rs >= 1714) {
return 4;
}
}
}
}
}
else if (SC >= 2849) {
if (Rs < 1747) {
return 4;
}
else if (Rs >= 1747) {
return 4;
}
}
}
}
else if (Cs >= 1139) {
return 4;
}
}
else if (diffRsCsdiff >= 1) {
return 4;
}
}
}
}
else if (Cs >= 1155) {
return 4;
}
}
else if (ssDCval >= 50) {
return 8;
}
}
}
else if (Rs >= 1753) {
if (TSCindex < 3) {
if (Rs < 1784) {
if (Rs < 1761) {
return 8;
}
else if (Rs >= 1761) {
if (Cs < 1143) {
if (Cs < 1140) {
if (SC < 2913) {
if (Rs < 1774) {
if (Cs < 1105) {
return 8;
}
else if (Cs >= 1105) {
return 4;
}
}
else if (Rs >= 1774) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1778) {
return 8;
}
else if (Rs >= 1778) {
return 8;
}
}
}
}
else if (SC >= 2913) {
if (Rs < 1779) {
return 4;
}
else if (Rs >= 1779) {
return 4;
}
}
}
else if (Cs >= 1140) {
return 4;
}
}
else if (Cs >= 1143) {
if (Cs < 1153) {
return 8;
}
else if (Cs >= 1153) {
if (Rs < 1770) {
return 8;
}
else if (Rs >= 1770) {
return 8;
}
}
}
}
}
else if (Rs >= 1784) {
return 8;
}
}
else if (TSCindex >= 3) {
if (Cs < 1128) {
return 4;
}
else if (Cs >= 1128) {
return 4;
}
}
}
}
}
else if (SC >= 2952) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (MVDiff < 7) {
if (tcor < 99) {
if (RsDiff < 2) {
if (diffTSC < -1) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (diffTSC >= -1) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (tcor < 98) {
if (Cs < 1200) {
return 1;
}
else if (Cs >= 1200) {
return 2;
}
}
else if (tcor >= 98) {
return 4;
}
}
}
else if (tcor >= 99) {
if (negBalance < 35) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (ssDCval < 69) {
if (SC < 2535) {
return 4;
}
else if (SC >= 2535) {
if (Rs < 1397) {
return 1;
}
else if (Rs >= 1397) {
return 4;
}
}
}
else if (ssDCval >= 69) {
return 4;
}
}
}
else if (negBalance >= 35) {
if (Rs < 1178) {
return 4;
}
else if (Rs >= 1178) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (TSC >= 3) {
if (Rs < 1357) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 113) {
return 8;
}
else if (mu_mv_mag_sq >= 113) {
if (ssDCval < 76) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
return 1;
}
}
else if (ssDCval >= 76) {
return 4;
}
}
}
else if (RsDiff >= 2) {
if (Cs < 1211) {
if (mu_mv_mag_sq < 119) {
if (Rs < 1158) {
if (Rs < 1066) {
return 8;
}
else if (Rs >= 1066) {
return 4;
}
}
else if (Rs >= 1158) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1207) {
return 8;
}
else if (Rs >= 1207) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 1029) {
return 8;
}
else if (Rs >= 1029) {
if (Cs < 1195) {
return 4;
}
else if (Cs >= 1195) {
return 1;
}
}
}
}
else if (Cs >= 1211) {
if (SC < 2308) {
if (Cs < 1229) {
return 4;
}
else if (Cs >= 1229) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
return 4;
}
}
}
else if (SC >= 2308) {
if (ssDCval < 54) {
if (Rs < 1180) {
return 8;
}
else if (Rs >= 1180) {
return 4;
}
}
else if (ssDCval >= 54) {
return 4;
}
}
}
}
}
else if (Rs >= 1357) {
if (RsDiff < 4) {
if (diffMVdiffVal < -5) {
return 1;
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (RsDiff >= 4) {
if (Cs < 1137) {
return 1;
}
else if (Cs >= 1137) {
if (Cs < 1181) {
return 8;
}
else if (Cs >= 1181) {
if (Rs < 1437) {
return 8;
}
else if (Rs >= 1437) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1815) {
if (TSC < 4) {
if (ssDCval < 82) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 103) {
return 4;
}
else if (mu_mv_mag_sq >= 103) {
if (Cs < 930) {
return 8;
}
else if (Cs >= 930) {
if (Rs < 1826) {
if (Rs < 1825) {
if (posBalance < 35) {
if (Rs < 1821) {
return 4;
}
else if (Rs >= 1821) {
return 4;
}
}
else if (posBalance >= 35) {
return 4;
}
}
else if (Rs >= 1825) {
return 8;
}
}
else if (Rs >= 1826) {
return 4;
}
}
}
}
else if (ssDCval >= 56) {
if (ssDCval < 70) {
if (mcTcor < 99) {
if (gchDC < 1) {
if (SC < 3071) {
return 1;
}
else if (SC >= 3071) {
if (Rs < 2019) {
if (Rs < 1923) {
return 2;
}
else if (Rs >= 1923) {
return 4;
}
}
else if (Rs >= 2019) {
return 1;
}
}
}
else if (gchDC >= 1) {
if (Cs < 910) {
return 1;
}
else if (Cs >= 910) {
if (Cs < 1074) {
return 4;
}
else if (Cs >= 1074) {
if (Rs < 2040) {
return 1;
}
else if (Rs >= 2040) {
return 8;
}
}
}
}
}
else if (mcTcor >= 99) {
if (SC < 2994) {
if (mu_mv_mag_sq < 105) {
return 4;
}
else if (mu_mv_mag_sq >= 105) {
if (Cs < 938) {
return 1;
}
else if (Cs >= 938) {
if (Cs < 1046) {
if (ssDCval < 57) {
if (Rs < 1947) {
return 1;
}
else if (Rs >= 1947) {
return 4;
}
}
else if (ssDCval >= 57) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
return 1;
}
}
}
else if (Cs >= 1046) {
return 1;
}
}
}
}
else if (SC >= 2994) {
if (Rs < 2142) {
return 4;
}
else if (Rs >= 2142) {
if (Rs < 2282) {
return 8;
}
else if (Rs >= 2282) {
return 2;
}
}
}
}
}
else if (ssDCval >= 70) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (Cs < 815) {
return 1;
}
else if (Cs >= 815) {
return 8;
}
}
}
}
}
else if (ssDCval >= 82) {
if (MVDiff < 2) {
if (Rs < 2195) {
return 4;
}
else if (Rs >= 2195) {
return 8;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (TSC >= 4) {
if (Cs < 922) {
if (ssDCval < 71) {
if (posBalance < 49) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Cs < 849) {
return 1;
}
else if (Cs >= 849) {
return 4;
}
}
}
else if (posBalance >= 49) {
return 2;
}
}
else if (ssDCval >= 71) {
if (TSC < 5) {
if (mu_mv_mag_sq < 93) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 93) {
if (Rs < 2013) {
return 4;
}
else if (Rs >= 2013) {
return 2;
}
}
}
else if (TSC >= 5) {
if (tcor < 96) {
if (posBalance < 53) {
if (RsCsDiff < 56) {
return 1;
}
else if (RsCsDiff >= 56) {
return 2;
}
}
else if (posBalance >= 53) {
return 2;
}
}
else if (tcor >= 96) {
if (refDCval < 82) {
if (Rs < 2073) {
return 8;
}
else if (Rs >= 2073) {
return 8;
}
}
else if (refDCval >= 82) {
if (Rs < 2032) {
return 2;
}
else if (Rs >= 2032) {
return 1;
}
}
}
}
}
}
else if (Cs >= 922) {
if (negBalance < 44) {
if (diffAFD < -2) {
if (Rs < 1870) {
return 8;
}
else if (Rs >= 1870) {
return 1;
}
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 107) {
if (Rs < 2260) {
return 4;
}
else if (Rs >= 2260) {
return 1;
}
}
else if (mu_mv_mag_sq >= 107) {
return 8;
}
}
}
else if (negBalance >= 44) {
if (Rs < 2275) {
return 8;
}
else if (Rs >= 2275) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 98) {
if (SC < 1396) {
if (MVDiff < 6) {
if (Rs < 420) {
if (mcTcor < 98) {
if (Rs < 268) {
return 2;
}
else if (Rs >= 268) {
if (Rs < 389) {
return 1;
}
else if (Rs >= 389) {
return 4;
}
}
}
else if (mcTcor >= 98) {
return 4;
}
}
else if (Rs >= 420) {
if (negBalance < 23) {
if (SC < 1216) {
return 1;
}
else if (SC >= 1216) {
return 4;
}
}
else if (negBalance >= 23) {
if (mcTcor < 98) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
if (Rs < 445) {
return 1;
}
else if (Rs >= 445) {
if (Rs < 741) {
return 8;
}
else if (Rs >= 741) {
return 1;
}
}
}
}
else if (mcTcor >= 98) {
if (posBalance < 34) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (posBalance >= 34) {
if (tcor < 91) {
if (Rs < 584) {
return 1;
}
else if (Rs >= 584) {
return 4;
}
}
else if (tcor >= 91) {
if (Rs < 516) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (Rs < 492) {
return 8;
}
else if (Rs >= 492) {
return 2;
}
}
}
else if (Rs >= 516) {
if (Rs < 723) {
return 2;
}
else if (Rs >= 723) {
return 1;
}
}
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (Rs < 427) {
if (Cs < 549) {
if (Rs < 327) {
return 2;
}
else if (Rs >= 327) {
if (Rs < 381) {
return 8;
}
else if (Rs >= 381) {
return 4;
}
}
}
else if (Cs >= 549) {
return 2;
}
}
else if (Rs >= 427) {
if (negBalance < 30) {
if (Rs < 569) {
return 2;
}
else if (Rs >= 569) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (negBalance >= 30) {
if (tcor < 94) {
if (Rs < 457) {
return 1;
}
else if (Rs >= 457) {
if (mu_mv_mag_sq < 3) {
if (Rs < 600) {
return 2;
}
else if (Rs >= 600) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 545) {
return 8;
}
else if (Cs >= 545) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
}
}
else if (tcor >= 94) {
if (Cs < 608) {
return 2;
}
else if (Cs >= 608) {
if (TSC < 4) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (TSC >= 4) {
if (ssDCval < 152) {
return 8;
}
else if (ssDCval >= 152) {
return 4;
}
}
}
}
}
}
}
}
}
else if (SC >= 1396) {
if (RsCsDiff < 81) {
if (MVDiff < 5) {
if (ssDCval < 122) {
if (Rs < 1166) {
if (RsDiff < 2) {
if (MVDiff < 2) {
if (Rs < 646) {
return 8;
}
else if (Rs >= 646) {
return 4;
}
}
else if (MVDiff >= 2) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (Cs < 820) {
if (ssDCval < 115) {
if (Rs < 875) {
return 4;
}
else if (Rs >= 875) {
return 1;
}
}
else if (ssDCval >= 115) {
return 8;
}
}
else if (Cs >= 820) {
if (Rs < 609) {
return 8;
}
else if (Rs >= 609) {
return 4;
}
}
}
}
else if (Rs >= 1166) {
if (ssDCval < 94) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Cs < 773) {
return 1;
}
else if (Cs >= 773) {
return 8;
}
}
}
else if (ssDCval >= 94) {
if (negBalance < 23) {
return 1;
}
else if (negBalance >= 23) {
return 8;
}
}
}
}
else if (ssDCval >= 122) {
if (SC < 2520) {
if (Cs < 967) {
if (RsDiff < 6) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 823) {
return 1;
}
else if (Rs >= 823) {
return 8;
}
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (Cs >= 967) {
return 4;
}
}
else if (SC >= 2520) {
if (CsDiff < 3) {
if (Rs < 2051) {
return 4;
}
else if (Rs >= 2051) {
return 8;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 158) {
if (refDCval < 102) {
if (TSC < 5) {
if (Cs < 541) {
return 8;
}
else if (Cs >= 541) {
if (SC < 1892) {
return 2;
}
else if (SC >= 1892) {
return 1;
}
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 43) {
if (TSC < 7) {
if (negBalance < 52) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 2;
}
}
}
else if (negBalance >= 52) {
return 1;
}
}
else if (TSC >= 7) {
if (AFD < 9) {
if (Rs < 2310) {
return 1;
}
else if (Rs >= 2310) {
return 4;
}
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (negBalance < 31) {
if (Rs < 1472) {
return 8;
}
else if (Rs >= 1472) {
return 1;
}
}
else if (negBalance >= 31) {
if (diffRsCsdiff < -18) {
if (Rs < 1631) {
return 2;
}
else if (Rs >= 1631) {
return 1;
}
}
else if (diffRsCsdiff >= -18) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 102) {
if (mu_mv_mag_sq < 5) {
if (ssDCval < 141) {
if (negBalance < 49) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 1288) {
return 2;
}
else if (Rs >= 1288) {
if (Rs < 2074) {
return 8;
}
else if (Rs >= 2074) {
return 1;
}
}
}
}
else if (negBalance >= 49) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 1727) {
return 2;
}
else if (Rs >= 1727) {
return 1;
}
}
}
}
}
else if (ssDCval >= 141) {
return 1;
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 2196) {
if (SC < 1564) {
if (diffMVdiffVal < 2) {
if (Rs < 509) {
return 8;
}
else if (Rs >= 509) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
return 8;
}
}
else if (SC >= 1564) {
if (Cs < 564) {
if (Rs < 1757) {
return 2;
}
else if (Rs >= 1757) {
return 1;
}
}
else if (Cs >= 564) {
if (negBalance < 31) {
if (ssDCval < 111) {
return 1;
}
else if (ssDCval >= 111) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
return 8;
}
}
}
else if (negBalance >= 31) {
if (diffRsCsdiff < -10) {
if (SC < 1877) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 957) {
return 1;
}
else if (Rs >= 957) {
return 4;
}
}
}
else if (SC >= 1877) {
if (diffRsCsdiff < -77) {
return 1;
}
else if (diffRsCsdiff >= -77) {
return 8;
}
}
}
else if (diffRsCsdiff >= -10) {
if (diffRsCsdiff < 10) {
return 8;
}
else if (diffRsCsdiff >= 10) {
if (SC < 1805) {
return 1;
}
else if (SC >= 1805) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 2196) {
if (Cs < 999) {
return 2;
}
else if (Cs >= 999) {
return 8;
}
}
}
}
}
else if (ssDCval >= 158) {
if (posBalance < 60) {
return 2;
}
else if (posBalance >= 60) {
return 4;
}
}
}
}
else if (RsCsDiff >= 81) {
if (Cs < 792) {
if (Rs < 1463) {
if (tcor < 77) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Rs < 883) {
return 1;
}
else if (Rs >= 883) {
return 4;
}
}
}
else if (tcor >= 77) {
return 2;
}
}
else if (Rs >= 1463) {
if (TSC < 5) {
if (Rs < 2025) {
return 8;
}
else if (Rs >= 2025) {
return 1;
}
}
else if (TSC >= 5) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (mu_mv_mag_sq < 46) {
return 1;
}
else if (mu_mv_mag_sq >= 46) {
return 2;
}
}
}
}
}
else if (Cs >= 792) {
if (SC < 2750) {
if (ssDCval < 100) {
if (MVDiff < 13) {
return 1;
}
else if (MVDiff >= 13) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (Rs < 839) {
return 1;
}
else if (Rs >= 839) {
return 4;
}
}
}
}
else if (ssDCval >= 100) {
if (mu_mv_mag_sq < 18) {
return 2;
}
else if (mu_mv_mag_sq >= 18) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
else if (SC >= 2750) {
if (TSC < 6) {
if (Rs < 2009) {
return 2;
}
else if (Rs >= 2009) {
return 4;
}
}
else if (TSC >= 6) {
if (diffRsCsdiff < 28) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
return 2;
}
}
else if (diffRsCsdiff >= 28) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (ssDCval < 133) {
if (Cs < 1215) {
if (Rs < 341) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 341) {
if (mu_mv_mag_sq < 105) {
if (Rs < 994) {
if (TSC < 4) {
if (refDCval < 117) {
if (refDCval < 92) {
return 8;
}
else if (refDCval >= 92) {
if (Rs < 582) {
if (Rs < 402) {
return 4;
}
else if (Rs >= 402) {
return 1;
}
}
else if (Rs >= 582) {
if (Rs < 622) {
return 2;
}
else if (Rs >= 622) {
return 1;
}
}
}
}
else if (refDCval >= 117) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (TSC >= 4) {
if (SC < 1633) {
return 8;
}
else if (SC >= 1633) {
return 4;
}
}
}
else if (Rs >= 994) {
if (refDCval < 89) {
if (Rs < 1628) {
return 8;
}
else if (Rs >= 1628) {
if (Rs < 2127) {
return 4;
}
else if (Rs >= 2127) {
return 1;
}
}
}
else if (refDCval >= 89) {
if (ssDCval < 125) {
return 8;
}
else if (ssDCval >= 125) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (refDCval < 97) {
if (Cs < 872) {
if (mu_mv_mag_sq < 126) {
if (Rs < 1043) {
if (negBalance < 18) {
return 1;
}
else if (negBalance >= 18) {
return 8;
}
}
else if (Rs >= 1043) {
if (ssDCval < 93) {
if (Rs < 2143) {
return 8;
}
else if (Rs >= 2143) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
}
else if (ssDCval >= 93) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (SC < 2895) {
if (posBalance < 24) {
return 4;
}
else if (posBalance >= 24) {
return 8;
}
}
else if (SC >= 2895) {
if (Cs < 751) {
return 8;
}
else if (Cs >= 751) {
return 4;
}
}
}
}
else if (Cs >= 872) {
if (mcTcor < 98) {
if (diffMVdiffVal < -1) {
if (Rs < 1247) {
return 1;
}
else if (Rs >= 1247) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (mcTcor >= 98) {
return 8;
}
}
}
else if (refDCval >= 97) {
if (Rs < 1676) {
if (Cs < 761) {
if (posBalance < 54) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (negBalance < 26) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (Rs < 739) {
return 1;
}
else if (Rs >= 739) {
return 8;
}
}
}
else if (negBalance >= 26) {
return 8;
}
}
}
else if (posBalance >= 54) {
if (Rs < 528) {
return 1;
}
else if (Rs >= 528) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (Cs >= 761) {
if (mu_mv_mag_sq < 107) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 107) {
return 8;
}
}
}
else if (Rs >= 1676) {
if (SC < 2749) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (SC >= 2749) {
if (ssDCval < 125) {
if (negBalance < 16) {
return 2;
}
else if (negBalance >= 16) {
return 8;
}
}
else if (ssDCval >= 125) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 1215) {
if (Rs < 663) {
if (ssDCval < 120) {
return 1;
}
else if (ssDCval >= 120) {
return 8;
}
}
else if (Rs >= 663) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 145) {
return 1;
}
else if (ssDCval >= 145) {
if (Rs < 996) {
return 8;
}
else if (Rs >= 996) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2380) {
if (MVDiff < 17) {
if (refDCval < 109) {
if (posBalance < 48) {
if (RsDiff < 6) {
if (Rs < 3765) {
if (Cs < 1027) {
if (Cs < 669) {
if (MVDiff < 3) {
if (Rs < 3042) {
return 1;
}
else if (Rs >= 3042) {
return 2;
}
}
else if (MVDiff >= 3) {
if (Rs < 2481) {
return 1;
}
else if (Rs >= 2481) {
if (Rs < 3159) {
return 8;
}
else if (Rs >= 3159) {
return 2;
}
}
}
}
else if (Cs >= 669) {
if (negBalance < 26) {
if (posBalance < 11) {
return 4;
}
else if (posBalance >= 11) {
if (CsDiff < 2) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (CsDiff >= 2) {
if (Rs < 2962) {
return 4;
}
else if (Rs >= 2962) {
if (Rs < 3375) {
return 8;
}
else if (Rs >= 3375) {
return 8;
}
}
}
}
}
else if (negBalance >= 26) {
if (tcor < 96) {
if (ssDCval < 78) {
if (Rs < 2789) {
return 2;
}
else if (Rs >= 2789) {
return 4;
}
}
else if (ssDCval >= 78) {
if (mcTcor < 97) {
if (posBalance < 40) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Cs < 923) {
return 4;
}
else if (Cs >= 923) {
return 4;
}
}
}
else if (posBalance >= 40) {
return 8;
}
}
else if (mcTcor >= 97) {
if (refDCval < 107) {
return 1;
}
else if (refDCval >= 107) {
if (Rs < 3074) {
return 8;
}
else if (Rs >= 3074) {
return 1;
}
}
}
}
}
else if (tcor >= 96) {
if (RsCsDiff < 32) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 26) {
return 4;
}
else if (mu_mv_mag_sq >= 26) {
return 2;
}
}
}
else if (RsCsDiff >= 32) {
if (Rs < 3070) {
return 1;
}
else if (Rs >= 3070) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1027) {
if (RsCsDiff < 5) {
if (negBalance < 21) {
if (Cs < 1291) {
return 8;
}
else if (Cs >= 1291) {
return 4;
}
}
else if (negBalance >= 21) {
if (Cs < 1113) {
if (Rs < 2930) {
return 8;
}
else if (Rs >= 2930) {
return 2;
}
}
else if (Cs >= 1113) {
return 8;
}
}
}
else if (RsCsDiff >= 5) {
if (refDCval < 103) {
if (ssDCval < 102) {
if (diffTSC < -1) {
if (RsDiff < 4) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
else if (RsDiff >= 4) {
if (posBalance < 45) {
if (Rs < 2427) {
return 4;
}
else if (Rs >= 2427) {
return 2;
}
}
else if (posBalance >= 45) {
if (Rs < 2697) {
return 8;
}
else if (Rs >= 2697) {
return 1;
}
}
}
}
else if (diffTSC >= -1) {
if (RsDiff < 4) {
if (Cs < 1291) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Rs < 3133) {
return 8;
}
else if (Rs >= 3133) {
return 2;
}
}
}
else if (Cs >= 1291) {
return 4;
}
}
else if (RsDiff >= 4) {
if (posBalance < 45) {
if (SC < 4466) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
else if (SC >= 4466) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 3495) {
return 1;
}
else if (Rs >= 3495) {
return 4;
}
}
}
}
else if (posBalance >= 45) {
if (ssDCval < 63) {
if (Rs < 2888) {
return 8;
}
else if (Rs >= 2888) {
return 4;
}
}
else if (ssDCval >= 63) {
if (Rs < 2900) {
if (Cs < 1176) {
return 2;
}
else if (Cs >= 1176) {
return 1;
}
}
else if (Rs >= 2900) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 102) {
return 4;
}
}
else if (refDCval >= 103) {
return 8;
}
}
}
}
else if (Rs >= 3765) {
if (Cs < 1259) {
if (ssDCval < 68) {
if (Cs < 1061) {
return 4;
}
else if (Cs >= 1061) {
return 1;
}
}
else if (ssDCval >= 68) {
return 4;
}
}
else if (Cs >= 1259) {
if (ssDCval < 83) {
if (mu_mv_mag_sq < 127) {
if (Rs < 4062) {
if (diffRsCsdiff < 1) {
return 4;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 3896) {
return 4;
}
else if (Rs >= 3896) {
return 8;
}
}
}
else if (Rs >= 4062) {
return 8;
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
else if (ssDCval >= 83) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 4;
}
}
}
}
}
else if (RsDiff >= 6) {
if (negBalance < 51) {
if (ssDCval < 84) {
if (Rs < 3859) {
if (refDCval < 66) {
if (MVDiff < 3) {
if (posBalance < 37) {
if (Rs < 3296) {
return 8;
}
else if (Rs >= 3296) {
return 1;
}
}
else if (posBalance >= 37) {
if (refDCval < 65) {
return 2;
}
else if (refDCval >= 65) {
if (Rs < 2993) {
return 2;
}
else if (Rs >= 2993) {
return 4;
}
}
}
}
else if (MVDiff >= 3) {
if (CsDiff < 6) {
if (refDCval < 59) {
if (Cs < 1281) {
return 4;
}
else if (Cs >= 1281) {
return 8;
}
}
else if (refDCval >= 59) {
if (posBalance < 36) {
return 1;
}
else if (posBalance >= 36) {
if (mu_mv_mag_sq < 46) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
if (AFD < 17) {
if (RsDiff < 12) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
else if (RsDiff >= 12) {
return 4;
}
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (SC < 3409) {
return 2;
}
else if (SC >= 3409) {
return 4;
}
}
}
}
}
else if (CsDiff >= 6) {
if (Cs < 963) {
return 4;
}
else if (Cs >= 963) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 2;
}
}
}
}
}
else if (refDCval >= 66) {
if (mu_mv_mag_sq < 82) {
if (SC < 4015) {
if (mcTcor < 97) {
if (RsCsDiff < 142) {
if (Cs < 953) {
if (ssDCval < 69) {
if (Rs < 2639) {
return 2;
}
else if (Rs >= 2639) {
return 8;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (Cs >= 953) {
if (posBalance < 40) {
if (Rs < 2669) {
return 2;
}
else if (Rs >= 2669) {
return 1;
}
}
else if (posBalance >= 40) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
if (Rs < 2887) {
return 4;
}
else if (Rs >= 2887) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 142) {
if (TSC < 8) {
if (Rs < 3103) {
return 4;
}
else if (Rs >= 3103) {
return 1;
}
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (TSC < 6) {
if (Rs < 2909) {
return 1;
}
else if (Rs >= 2909) {
if (ssDCval < 69) {
if (Rs < 2996) {
return 4;
}
else if (Rs >= 2996) {
return 8;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
else if (TSC >= 6) {
return 4;
}
}
}
else if (SC >= 4015) {
if (Cs < 1062) {
if (diffAFD < -2) {
if (Rs < 3406) {
return 4;
}
else if (Rs >= 3406) {
return 1;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Cs >= 1062) {
if (diffMVdiffVal < 3) {
if (posBalance < 46) {
if (ssDCval < 67) {
return 2;
}
else if (ssDCval >= 67) {
if (diffRsCsdiff < 39) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 4;
}
}
else if (diffRsCsdiff >= 39) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 4;
}
}
}
}
else if (posBalance >= 46) {
if (Cs < 1214) {
return 2;
}
else if (Cs >= 1214) {
if (Rs < 3052) {
return 1;
}
else if (Rs >= 3052) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 3089) {
return 4;
}
else if (Rs >= 3089) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 2;
}
}
}
}
else if (Rs >= 3859) {
if (mu_mv_mag_sq < 108) {
return 4;
}
else if (mu_mv_mag_sq >= 108) {
return 1;
}
}
}
else if (ssDCval >= 84) {
if (Cs < 922) {
if (Rs < 2868) {
if (Rs < 2479) {
return 1;
}
else if (Rs >= 2479) {
if (Rs < 2636) {
return 4;
}
else if (Rs >= 2636) {
return 8;
}
}
}
else if (Rs >= 2868) {
if (diffMVdiffVal < -9) {
return 4;
}
else if (diffMVdiffVal >= -9) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
return 2;
}
}
}
}
else if (Cs >= 922) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
if (ssDCval < 99) {
if (posBalance < 41) {
if (diffRsCsdiff < -11) {
if (Rs < 3416) {
return 8;
}
else if (Rs >= 3416) {
return 8;
}
}
else if (diffRsCsdiff >= -11) {
if (RsCsDiff < 70) {
return 4;
}
else if (RsCsDiff >= 70) {
if (RsCsDiff < 85) {
if (Rs < 4218) {
return 4;
}
else if (Rs >= 4218) {
return 4;
}
}
else if (RsCsDiff >= 85) {
return 4;
}
}
}
}
else if (posBalance >= 41) {
if (ssDCval < 91) {
if (Rs < 4119) {
return 8;
}
else if (Rs >= 4119) {
if (Cs < 1143) {
return 4;
}
else if (Cs >= 1143) {
if (Rs < 4139) {
return 4;
}
else if (Rs >= 4139) {
if (Rs < 4197) {
return 4;
}
else if (Rs >= 4197) {
if (Rs < 4286) {
return 4;
}
else if (Rs >= 4286) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (TSC < 7) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (mu_mv_mag_sq < 56) {
if (diffRsCsdiff < 17) {
if (refDCval < 94) {
if (Cs < 1162) {
if (Rs < 4119) {
if (Cs < 1052) {
return 4;
}
else if (Cs >= 1052) {
return 8;
}
}
else if (Rs >= 4119) {
return 4;
}
}
else if (Cs >= 1162) {
return 8;
}
}
else if (refDCval >= 94) {
if (Rs < 3445) {
return 8;
}
else if (Rs >= 3445) {
return 8;
}
}
}
else if (diffRsCsdiff >= 17) {
return 4;
}
}
else if (mu_mv_mag_sq >= 56) {
return 8;
}
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
else if (ssDCval >= 99) {
return 8;
}
}
}
}
}
else if (negBalance >= 51) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Rs < 2820) {
return 1;
}
else if (Rs >= 2820) {
return 2;
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Rs < 3401) {
if (refDCval < 81) {
if (negBalance < 21) {
if (Rs < 2604) {
return 1;
}
else if (Rs >= 2604) {
if (Cs < 1031) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 2750) {
return 4;
}
else if (Rs >= 2750) {
return 8;
}
}
}
else if (Cs >= 1031) {
return 8;
}
}
}
else if (negBalance >= 21) {
if (gchDC < 2) {
if (ssDCval < 60) {
if (Rs < 2657) {
return 1;
}
else if (Rs >= 2657) {
return 4;
}
}
else if (ssDCval >= 60) {
if (refDCval < 66) {
if (Cs < 639) {
if (Rs < 2508) {
return 2;
}
else if (Rs >= 2508) {
return 4;
}
}
else if (Cs >= 639) {
if (Cs < 1059) {
return 2;
}
else if (Cs >= 1059) {
return 4;
}
}
}
else if (refDCval >= 66) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (TSC < 6) {
if (Rs < 2760) {
return 1;
}
else if (Rs >= 2760) {
return 8;
}
}
else if (TSC >= 6) {
return 2;
}
}
}
}
else if (refDCval >= 81) {
if (posBalance < 51) {
if (Rs < 2656) {
return 1;
}
else if (Rs >= 2656) {
return 8;
}
}
else if (posBalance >= 51) {
if (Rs < 2631) {
if (Cs < 463) {
return 1;
}
else if (Cs >= 463) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (Rs >= 2631) {
return 1;
}
}
}
}
else if (Rs >= 3401) {
if (posBalance < 66) {
return 4;
}
else if (posBalance >= 66) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 64) {
if (SC < 3474) {
if (mu_mv_mag_sq < 100) {
return 2;
}
else if (mu_mv_mag_sq >= 100) {
return 4;
}
}
else if (SC >= 3474) {
if (mcTcor < 95) {
if (posBalance < 60) {
if (Rs < 2483) {
return 8;
}
else if (Rs >= 2483) {
if (diffRsCsdiff < 60) {
return 4;
}
else if (diffRsCsdiff >= 60) {
if (Rs < 3181) {
return 2;
}
else if (Rs >= 3181) {
return 4;
}
}
}
}
else if (posBalance >= 60) {
return 2;
}
}
else if (mcTcor >= 95) {
if (AFD < 10) {
if (mu_mv_mag_sq < 96) {
return 2;
}
else if (mu_mv_mag_sq >= 96) {
if (Cs < 978) {
return 2;
}
else if (Cs >= 978) {
return 4;
}
}
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (refDCval >= 64) {
if (negBalance < 38) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 81) {
if (diffTSC < -2) {
if (Rs < 2784) {
return 2;
}
else if (Rs >= 2784) {
return 1;
}
}
else if (diffTSC >= -2) {
if (posBalance < 50) {
if (Rs < 2807) {
return 1;
}
else if (Rs >= 2807) {
if (Cs < 884) {
return 1;
}
else if (Cs >= 884) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (Rs < 2968) {
return 1;
}
else if (Rs >= 2968) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
return 2;
}
}
else if (posBalance >= 53) {
if (AFD < 18) {
if (RsDiff < 5) {
if (Rs < 2793) {
return 1;
}
else if (Rs >= 2793) {
return 4;
}
}
else if (RsDiff >= 5) {
if (Cs < 727) {
if (posBalance < 58) {
if (Rs < 2918) {
return 1;
}
else if (Rs >= 2918) {
return 4;
}
}
else if (posBalance >= 58) {
if (Rs < 2467) {
return 1;
}
else if (Rs >= 2467) {
return 2;
}
}
}
else if (Cs >= 727) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (tcor < 89) {
return 1;
}
else if (tcor >= 89) {
return 2;
}
}
}
}
}
else if (AFD >= 18) {
if (Cs < 425) {
return 1;
}
else if (Cs >= 425) {
return 4;
}
}
}
}
else if (negBalance >= 38) {
if (diffRsCsdiff < -191) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 2;
}
}
else if (diffRsCsdiff >= -191) {
if (Rs < 2487) {
if (SC < 3536) {
if (refDCval < 85) {
return 2;
}
else if (refDCval >= 85) {
if (Cs < 599) {
return 1;
}
else if (Cs >= 599) {
if (Cs < 690) {
return 2;
}
else if (Cs >= 690) {
return 4;
}
}
}
}
else if (SC >= 3536) {
if (Rs < 2454) {
return 4;
}
else if (Rs >= 2454) {
return 2;
}
}
}
else if (Rs >= 2487) {
if (tcor < 76) {
if (negBalance < 44) {
if (Rs < 3271) {
if (Rs < 2999) {
return 2;
}
else if (Rs >= 2999) {
return 4;
}
}
else if (Rs >= 3271) {
return 1;
}
}
else if (negBalance >= 44) {
return 2;
}
}
else if (tcor >= 76) {
if (TSC < 5) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
if (Cs < 761) {
return 1;
}
else if (Cs >= 761) {
return 4;
}
}
}
else if (TSC >= 5) {
if (ssDCval < 68) {
if (mu_mv_mag_sq < 93) {
if (Rs < 3183) {
return 2;
}
else if (Rs >= 3183) {
return 4;
}
}
else if (mu_mv_mag_sq >= 93) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (ssDCval >= 68) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsDiff < 11) {
if (Cs < 998) {
if (tcor < 96) {
if (RsCsDiff < 63) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
if (Cs < 531) {
if (Rs < 2669) {
return 4;
}
else if (Rs >= 2669) {
return 4;
}
}
else if (Cs >= 531) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 3271) {
return 8;
}
else if (Rs >= 3271) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 63) {
if (gchDC < 4) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (Rs < 3151) {
return 4;
}
else if (Rs >= 3151) {
return 8;
}
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (tcor >= 96) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
return 2;
}
}
}
else if (Cs >= 998) {
if (diffAFD < -2) {
if (tcor < 90) {
return 4;
}
else if (tcor >= 90) {
if (Rs < 2973) {
return 8;
}
else if (Rs >= 2973) {
return 2;
}
}
}
else if (diffAFD >= -2) {
if (Rs < 3885) {
return 8;
}
else if (Rs >= 3885) {
if (Rs < 4053) {
return 2;
}
else if (Rs >= 4053) {
if (Rs < 4258) {
return 8;
}
else if (Rs >= 4258) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 11) {
if (Cs < 837) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
if (Rs < 2780) {
return 1;
}
else if (Rs >= 2780) {
return 2;
}
}
}
else if (Cs >= 837) {
if (negBalance < 50) {
if (RsCsDiff < 175) {
if (diffMVdiffVal < -1) {
if (Rs < 3712) {
return 4;
}
else if (Rs >= 3712) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (RsCsDiff >= 175) {
return 2;
}
}
else if (negBalance >= 50) {
if (posBalance < 58) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (posBalance >= 58) {
if (Rs < 3326) {
if (mcTcor < 85) {
if (Rs < 2929) {
return 8;
}
else if (Rs >= 2929) {
return 1;
}
}
else if (mcTcor >= 85) {
if (diffRsCsdiff < -9) {
if (RsDiff < 14) {
return 4;
}
else if (RsDiff >= 14) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
return 2;
}
}
}
else if (diffRsCsdiff >= -9) {
return 4;
}
}
}
else if (Rs >= 3326) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
return 4;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 3348) {
if (mcTcor < 97) {
if (RsDiff < 9) {
if (mcTcor < 95) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
return 8;
}
}
else if (mcTcor >= 95) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 89) {
if (MVDiff < 11) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (mu_mv_mag_sq < 84) {
return 1;
}
else if (mu_mv_mag_sq >= 84) {
return 8;
}
}
}
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (mu_mv_mag_sq >= 89) {
if (Rs < 2697) {
return 1;
}
else if (Rs >= 2697) {
return 2;
}
}
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 117) {
if (Cs < 624) {
return 2;
}
else if (Cs >= 624) {
if (Rs < 2505) {
return 4;
}
else if (Rs >= 2505) {
if (Rs < 2963) {
return 1;
}
else if (Rs >= 2963) {
return 8;
}
}
}
}
else if (ssDCval >= 117) {
if (Rs < 2866) {
if (mcTcor < 95) {
if (Rs < 2845) {
return 8;
}
else if (Rs >= 2845) {
return 1;
}
}
else if (mcTcor >= 95) {
if (Cs < 1185) {
return 1;
}
else if (Cs >= 1185) {
return 8;
}
}
}
else if (Rs >= 2866) {
if (mu_mv_mag_sq < 32) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 1;
}
}
else if (mu_mv_mag_sq >= 32) {
return 2;
}
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 7) {
if (Rs < 2694) {
if (ssDCval < 126) {
return 1;
}
else if (ssDCval >= 126) {
if (Cs < 1159) {
return 2;
}
else if (Cs >= 1159) {
return 1;
}
}
}
else if (Rs >= 2694) {
if (Cs < 530) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 2;
}
}
else if (Cs >= 530) {
return 2;
}
}
}
else if (MVDiff >= 7) {
if (TSC < 6) {
if (mu_mv_mag_sq < 85) {
if (SC < 3480) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
return 1;
}
}
else if (SC >= 3480) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (RsDiff < 4) {
if (Rs < 3032) {
return 1;
}
else if (Rs >= 3032) {
return 2;
}
}
else if (RsDiff >= 4) {
return 2;
}
}
}
else if (TSC >= 6) {
if (diffMVdiffVal < -3) {
if (mu_mv_mag_sq < 51) {
return 1;
}
else if (mu_mv_mag_sq >= 51) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
return 2;
}
}
}
else if (diffMVdiffVal >= -3) {
if (refDCval < 117) {
if (Cs < 799) {
return 2;
}
else if (Cs >= 799) {
if (Rs < 2730) {
return 8;
}
else if (Rs >= 2730) {
return 4;
}
}
}
else if (refDCval >= 117) {
if (Cs < 767) {
if (ssDCval < 130) {
return 2;
}
else if (ssDCval >= 130) {
if (Rs < 2609) {
return 2;
}
else if (Rs >= 2609) {
return 1;
}
}
}
else if (Cs >= 767) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 3348) {
if (MVDiff < 7) {
if (refDCval < 142) {
if (tcor < 89) {
if (RsDiff < 16) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (Rs < 3785) {
return 4;
}
else if (Rs >= 3785) {
return 1;
}
}
}
else if (RsDiff >= 16) {
return 2;
}
}
else if (tcor >= 89) {
if (diffMVdiffVal < -3) {
if (Rs < 3416) {
return 8;
}
else if (Rs >= 3416) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (refDCval < 135) {
return 2;
}
else if (refDCval >= 135) {
if (Rs < 5442) {
return 1;
}
else if (Rs >= 5442) {
return 2;
}
}
}
}
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
}
else if (refDCval >= 142) {
if (Rs < 3834) {
if (Rs < 3664) {
return 2;
}
else if (Rs >= 3664) {
if (Cs < 1107) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (Cs >= 1107) {
if (posBalance < 34) {
if (Rs < 3816) {
return 2;
}
else if (Rs >= 3816) {
return 1;
}
}
else if (posBalance >= 34) {
return 1;
}
}
}
}
else if (Rs >= 3834) {
if (TSC < 10) {
if (diffTSC < -2) {
if (Rs < 4037) {
return 1;
}
else if (Rs >= 4037) {
return 2;
}
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (TSC >= 10) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (mcTcor < 94) {
if (ssDCval < 135) {
if (refDCval < 133) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 2;
}
}
else if (refDCval >= 133) {
if (Rs < 3753) {
return 8;
}
else if (Rs >= 3753) {
return 1;
}
}
}
else if (ssDCval >= 135) {
if (RsDiff < 18) {
return 4;
}
else if (RsDiff >= 18) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
return 4;
}
}
}
}
else if (mcTcor >= 94) {
if (TSC < 6) {
if (diffAFD < -2) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (diffAFD >= -2) {
if (refDCval < 112) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (refDCval >= 112) {
if (RsDiff < 15) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 2;
}
}
else if (RsDiff >= 15) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 134) {
if (mu_mv_mag_sq < 2) {
if (negBalance < 44) {
if (Rs < 5362) {
return 1;
}
else if (Rs >= 5362) {
return 2;
}
}
else if (negBalance >= 44) {
if (ssDCval < 132) {
return 2;
}
else if (ssDCval >= 132) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 1143) {
return 2;
}
else if (Cs >= 1143) {
if (MVDiff < 14) {
if (negBalance < 55) {
return 2;
}
else if (negBalance >= 55) {
if (Rs < 4466) {
return 4;
}
else if (Rs >= 4466) {
return 2;
}
}
}
else if (MVDiff >= 14) {
return 8;
}
}
}
}
else if (ssDCval >= 134) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (mcTcor < 97) {
if (negBalance < 55) {
if (RsDiff < 20) {
if (diffRsCsdiff < -71) {
return 2;
}
else if (diffRsCsdiff >= -71) {
if (ssDCval < 156) {
return 1;
}
else if (ssDCval >= 156) {
return 2;
}
}
}
else if (RsDiff >= 20) {
return 2;
}
}
else if (negBalance >= 55) {
return 4;
}
}
else if (mcTcor >= 97) {
if (Cs < 1054) {
if (Rs < 5819) {
return 2;
}
else if (Rs >= 5819) {
if (Rs < 6435) {
return 1;
}
else if (Rs >= 6435) {
return 2;
}
}
}
else if (Cs >= 1054) {
if (gchDC < 3) {
if (Rs < 3746) {
return 2;
}
else if (Rs >= 3746) {
return 1;
}
}
else if (gchDC >= 3) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 17) {
if (mcTcor < 68) {
if (MVDiff < 51) {
return 1;
}
else if (MVDiff >= 51) {
if (gchDC < 7) {
if (Rs < 2426) {
return 2;
}
else if (Rs >= 2426) {
return 4;
}
}
else if (gchDC >= 7) {
return 8;
}
}
}
else if (mcTcor >= 68) {
if (ssDCval < 111) {
if (RsCsDiff < 194) {
if (refDCval < 92) {
if (Rs < 2758) {
if (AFD < 12) {
if (ssDCval < 78) {
if (TSC < 8) {
if (diffMVdiffVal < -16) {
return 1;
}
else if (diffMVdiffVal >= -16) {
if (Rs < 2388) {
return 1;
}
else if (Rs >= 2388) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
return 8;
}
}
}
}
else if (TSC >= 8) {
if (Rs < 2618) {
return 1;
}
else if (Rs >= 2618) {
return 2;
}
}
}
else if (ssDCval >= 78) {
if (Rs < 2481) {
return 1;
}
else if (Rs >= 2481) {
if (RsDiff < 10) {
if (Cs < 304) {
return 2;
}
else if (Cs >= 304) {
return 1;
}
}
else if (RsDiff >= 10) {
return 2;
}
}
}
}
else if (AFD >= 12) {
if (mcTcor < 87) {
if (Cs < 948) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 1;
}
}
else if (Cs >= 948) {
if (refDCval < 77) {
if (Rs < 2418) {
return 8;
}
else if (Rs >= 2418) {
return 1;
}
}
else if (refDCval >= 77) {
return 8;
}
}
}
else if (mcTcor >= 87) {
if (refDCval < 81) {
if (refDCval < 56) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (refDCval >= 56) {
if (TSC < 7) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
return 4;
}
}
else if (TSC >= 7) {
if (tcor < 75) {
if (Cs < 912) {
return 1;
}
else if (Cs >= 912) {
return 2;
}
}
else if (tcor >= 75) {
return 2;
}
}
}
}
else if (refDCval >= 81) {
if (Cs < 659) {
return 1;
}
else if (Cs >= 659) {
if (MVDiff < 36) {
if (ssDCval < 88) {
return 2;
}
else if (ssDCval >= 88) {
if (Rs < 2636) {
return 4;
}
else if (Rs >= 2636) {
return 2;
}
}
}
else if (MVDiff >= 36) {
if (Rs < 2584) {
return 8;
}
else if (Rs >= 2584) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 2758) {
if (Cs < 617) {
if (ssDCval < 81) {
if (posBalance < 42) {
if (Rs < 3326) {
return 8;
}
else if (Rs >= 3326) {
return 4;
}
}
else if (posBalance >= 42) {
return 4;
}
}
else if (ssDCval >= 81) {
if (Cs < 302) {
if (Rs < 3124) {
return 4;
}
else if (Rs >= 3124) {
return 1;
}
}
else if (Cs >= 302) {
if (Rs < 3233) {
return 8;
}
else if (Rs >= 3233) {
return 2;
}
}
}
}
else if (Cs >= 617) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (negBalance < 46) {
if (mu_mv_mag_sq < 41) {
if (diffRsCsdiff < -5) {
if (diffRsCsdiff < -46) {
return 4;
}
else if (diffRsCsdiff >= -46) {
if (Rs < 2820) {
return 4;
}
else if (Rs >= 2820) {
return 2;
}
}
}
else if (diffRsCsdiff >= -5) {
return 4;
}
}
else if (mu_mv_mag_sq >= 41) {
if (AFD < 10) {
if (Rs < 2883) {
return 1;
}
else if (Rs >= 2883) {
return 4;
}
}
else if (AFD >= 10) {
if (Rs < 2770) {
return 4;
}
else if (Rs >= 2770) {
return 2;
}
}
}
}
else if (negBalance >= 46) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 92) {
if (AFD < 23) {
if (diffRsCsdiff < -52) {
if (diffRsCsdiff < -194) {
return 2;
}
else if (diffRsCsdiff >= -194) {
if (Cs < 1152) {
if (MVDiff < 39) {
if (SC < 3727) {
return 4;
}
else if (SC >= 3727) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
if (Rs < 3316) {
return 2;
}
else if (Rs >= 3316) {
if (Rs < 5483) {
return 8;
}
else if (Rs >= 5483) {
return 2;
}
}
}
}
}
else if (MVDiff >= 39) {
if (diffMVdiffVal < -62) {
return 2;
}
else if (diffMVdiffVal >= -62) {
if (Cs < 375) {
return 2;
}
else if (Cs >= 375) {
return 8;
}
}
}
}
else if (Cs >= 1152) {
return 8;
}
}
}
else if (diffRsCsdiff >= -52) {
if (mu_mv_mag_sq < 65) {
if (posBalance < 58) {
if (diffRsCsdiff < -50) {
return 1;
}
else if (diffRsCsdiff >= -50) {
if (Cs < 377) {
if (Rs < 2718) {
if (Rs < 2612) {
return 1;
}
else if (Rs >= 2612) {
return 2;
}
}
else if (Rs >= 2718) {
if (Rs < 2877) {
return 8;
}
else if (Rs >= 2877) {
return 4;
}
}
}
else if (Cs >= 377) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
return 8;
}
}
}
}
else if (posBalance >= 58) {
if (MVDiff < 44) {
if (diffMVdiffVal < 4) {
if (Cs < 1073) {
if (Cs < 514) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
if (Rs < 2582) {
return 1;
}
else if (Rs >= 2582) {
return 8;
}
}
}
else if (Cs >= 514) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
return 1;
}
}
}
else if (Cs >= 1073) {
return 8;
}
}
else if (diffMVdiffVal >= 4) {
if (diffMVdiffVal < 8) {
return 8;
}
else if (diffMVdiffVal >= 8) {
if (Rs < 2886) {
return 2;
}
else if (Rs >= 2886) {
if (Rs < 5257) {
return 8;
}
else if (Rs >= 5257) {
return 2;
}
}
}
}
}
else if (MVDiff >= 44) {
if (AFD < 13) {
if (Rs < 5039) {
if (Rs < 3662) {
return 1;
}
else if (Rs >= 3662) {
return 8;
}
}
else if (Rs >= 5039) {
return 2;
}
}
else if (AFD >= 13) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 65) {
return 4;
}
}
}
else if (AFD >= 23) {
if (Rs < 2808) {
if (Rs < 2388) {
return 1;
}
else if (Rs >= 2388) {
return 2;
}
}
else if (Rs >= 2808) {
if (Cs < 796) {
return 1;
}
else if (Cs >= 796) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 194) {
if (refDCval < 74) {
if (Rs < 3256) {
if (diffRsCsdiff < 124) {
if (diffMVdiffVal < -16) {
if (mu_mv_mag_sq < 6) {
return 4;
}
else if (mu_mv_mag_sq >= 6) {
if (AFD < 32) {
return 1;
}
else if (AFD >= 32) {
return 2;
}
}
}
else if (diffMVdiffVal >= -16) {
if (Rs < 2984) {
return 2;
}
else if (Rs >= 2984) {
if (negBalance < 43) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 2;
}
}
else if (negBalance >= 43) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
return 1;
}
}
}
}
}
else if (diffRsCsdiff >= 124) {
if (mcTcor < 93) {
if (Cs < 268) {
return 1;
}
else if (Cs >= 268) {
if (Rs < 2885) {
return 8;
}
else if (Rs >= 2885) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 2506) {
return 1;
}
else if (Rs >= 2506) {
return 4;
}
}
}
}
else if (Rs >= 3256) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
if (Rs < 4789) {
return 1;
}
else if (Rs >= 4789) {
return 8;
}
}
}
}
else if (refDCval >= 74) {
if (mcTcor < 90) {
if (RsCsDiff < 359) {
if (TSC < 12) {
if (SC < 3782) {
if (MVDiff < 42) {
if (Rs < 2426) {
return 4;
}
else if (Rs >= 2426) {
return 2;
}
}
else if (MVDiff >= 42) {
return 4;
}
}
else if (SC >= 3782) {
if (Cs < 685) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 2;
}
}
else if (Cs >= 685) {
if (Rs < 4217) {
return 8;
}
else if (Rs >= 4217) {
return 2;
}
}
}
}
else if (TSC >= 12) {
if (posBalance < 65) {
if (Rs < 2655) {
if (AFD < 26) {
return 4;
}
else if (AFD >= 26) {
return 2;
}
}
else if (Rs >= 2655) {
if (mcTcor < 70) {
return 1;
}
else if (mcTcor >= 70) {
return 4;
}
}
}
else if (posBalance >= 65) {
if (MVDiff < 68) {
return 1;
}
else if (MVDiff >= 68) {
return 8;
}
}
}
}
else if (RsCsDiff >= 359) {
if (Cs < 995) {
if (TSC < 18) {
return 2;
}
else if (TSC >= 18) {
return 1;
}
}
else if (Cs >= 995) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
if (MVDiff < 77) {
return 2;
}
else if (MVDiff >= 77) {
if (diffRsCsdiff < -142) {
return 2;
}
else if (diffRsCsdiff >= -142) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (gchDC < 6) {
if (RsDiff < 15) {
if (MVDiff < 55) {
if (refDCval < 106) {
if (diffRsCsdiff < 18) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
if (MVDiff < 42) {
return 2;
}
else if (MVDiff >= 42) {
return 8;
}
}
}
else if (diffRsCsdiff >= 18) {
if (diffRsCsdiff < 42) {
if (TSC < 9) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (diffRsCsdiff >= 42) {
if (MVDiff < 29) {
return 2;
}
else if (MVDiff >= 29) {
return 8;
}
}
}
}
else if (refDCval >= 106) {
if (mu_mv_mag_sq < 16) {
if (Rs < 2652) {
return 1;
}
else if (Rs >= 2652) {
return 4;
}
}
else if (mu_mv_mag_sq >= 16) {
if (Rs < 6106) {
return 8;
}
else if (Rs >= 6106) {
return 2;
}
}
}
}
else if (MVDiff >= 55) {
if (diffAFD < 0) {
return 4;
}
else if (diffAFD >= 0) {
return 2;
}
}
}
else if (RsDiff >= 15) {
if (ssDCval < 82) {
if (CsDiff < 4) {
if (Rs < 3698) {
return 2;
}
else if (Rs >= 3698) {
return 4;
}
}
else if (CsDiff >= 4) {
if (Rs < 4490) {
return 2;
}
else if (Rs >= 4490) {
return 1;
}
}
}
else if (ssDCval >= 82) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 57) {
if (Rs < 4481) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
return 2;
}
}
else if (Rs >= 4481) {
if (RsDiff < 18) {
return 8;
}
else if (RsDiff >= 18) {
return 2;
}
}
}
else if (MVDiff >= 57) {
if (AFD < 21) {
if (refDCval < 109) {
return 2;
}
else if (refDCval >= 109) {
return 1;
}
}
else if (AFD >= 21) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (mcTcor < 91) {
if (Rs < 2758) {
if (Rs < 2507) {
return 2;
}
else if (Rs >= 2507) {
return 4;
}
}
else if (Rs >= 2758) {
return 2;
}
}
else if (mcTcor >= 91) {
if (Rs < 2398) {
if (Rs < 2388) {
return 2;
}
else if (Rs >= 2388) {
return 1;
}
}
else if (Rs >= 2398) {
if (refDCval < 81) {
if (Rs < 3732) {
return 2;
}
else if (Rs >= 3732) {
return 1;
}
}
else if (refDCval >= 81) {
if (diffRsCsdiff < 134) {
return 2;
}
else if (diffRsCsdiff >= 134) {
if (diffRsCsdiff < 140) {
return 8;
}
else if (diffRsCsdiff >= 140) {
return 2;
}
}
}
}
}
}
}
}
}
else if (gchDC >= 6) {
if (diffAFD < 4) {
return 1;
}
else if (diffAFD >= 4) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 111) {
if (mcTcor < 92) {
if (RsCsDiff < 387) {
if (refDCval < 134) {
if (mcTcor < 78) {
if (Cs < 919) {
if (Rs < 2449) {
return 8;
}
else if (Rs >= 2449) {
return 1;
}
}
else if (Cs >= 919) {
return 4;
}
}
else if (mcTcor >= 78) {
if (diffRsCsdiff < 105) {
if (MVDiff < 29) {
if (Cs < 855) {
return 2;
}
else if (Cs >= 855) {
if (Rs < 3765) {
return 8;
}
else if (Rs >= 3765) {
return 1;
}
}
}
else if (MVDiff >= 29) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
if (diffRsCsdiff < -69) {
if (Rs < 2978) {
return 1;
}
else if (Rs >= 2978) {
return 8;
}
}
else if (diffRsCsdiff >= -69) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 105) {
if (AFD < 28) {
return 2;
}
else if (AFD >= 28) {
return 1;
}
}
}
}
else if (refDCval >= 134) {
if (refDCval < 147) {
if (Cs < 1248) {
return 4;
}
else if (Cs >= 1248) {
return 8;
}
}
else if (refDCval >= 147) {
if (CsDiff < 5) {
if (AFD < 31) {
return 8;
}
else if (AFD >= 31) {
return 2;
}
}
else if (CsDiff >= 5) {
if (Rs < 3959) {
return 1;
}
else if (Rs >= 3959) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 387) {
if (CsDiff < 7) {
if (tcor < 81) {
if (ssDCval < 159) {
return 2;
}
else if (ssDCval >= 159) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
return 1;
}
}
}
else if (tcor >= 81) {
if (Rs < 4697) {
return 4;
}
else if (Rs >= 4697) {
return 2;
}
}
}
else if (CsDiff >= 7) {
if (tcor < 68) {
if (refDCval < 119) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
else if (refDCval >= 119) {
if (AFD < 33) {
return 2;
}
else if (AFD >= 33) {
return 1;
}
}
}
else if (tcor >= 68) {
if (Rs < 4923) {
return 1;
}
else if (Rs >= 4923) {
return 8;
}
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 3266) {
if (RsDiff < 9) {
if (diffAFD < -2) {
if (Rs < 2815) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
return 1;
}
}
else if (Rs >= 2815) {
return 8;
}
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < 12) {
return 8;
}
else if (diffRsCsdiff >= 12) {
if (Cs < 424) {
return 1;
}
else if (Cs >= 424) {
return 8;
}
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 121) {
if (MVDiff < 24) {
if (mu_mv_mag_sq < 2) {
if (tcor < 79) {
return 2;
}
else if (tcor >= 79) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (diffAFD < 2) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
}
else if (MVDiff >= 24) {
if (RsCsDiff < 209) {
if (posBalance < 64) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 8;
}
}
else if (posBalance >= 64) {
if (mu_mv_mag_sq < 24) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 8;
}
}
else if (mu_mv_mag_sq >= 24) {
return 8;
}
}
}
else if (RsCsDiff >= 209) {
if (refDCval < 114) {
if (Cs < 703) {
if (SC < 3143) {
return 2;
}
else if (SC >= 3143) {
return 8;
}
}
else if (Cs >= 703) {
return 1;
}
}
else if (refDCval >= 114) {
return 1;
}
}
}
}
else if (ssDCval >= 121) {
if (Cs < 1062) {
if (CsDiff < 3) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
if (Rs < 2383) {
return 1;
}
else if (Rs >= 2383) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (AFD < 32) {
if (TSC < 13) {
if (MVDiff < 29) {
if (gchDC < 4) {
return 1;
}
else if (gchDC >= 4) {
if (RsCsDiff < 182) {
if (Rs < 2526) {
return 1;
}
else if (Rs >= 2526) {
if (Cs < 463) {
return 1;
}
else if (Cs >= 463) {
return 2;
}
}
}
else if (RsCsDiff >= 182) {
return 1;
}
}
}
else if (MVDiff >= 29) {
return 1;
}
}
else if (TSC >= 13) {
if (Cs < 716) {
return 8;
}
else if (Cs >= 716) {
return 1;
}
}
}
else if (AFD >= 32) {
if (diffRsCsdiff < 110) {
return 2;
}
else if (diffRsCsdiff >= 110) {
return 1;
}
}
}
}
else if (Cs >= 1062) {
if (refDCval < 137) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 1;
}
}
else if (refDCval >= 137) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3266) {
if (CsDiff < 4) {
if (Rs < 4518) {
if (negBalance < 47) {
if (MVDiff < 61) {
return 2;
}
else if (MVDiff >= 61) {
if (Rs < 3495) {
return 8;
}
else if (Rs >= 3495) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (ssDCval < 132) {
if (Cs < 471) {
if (Rs < 3886) {
return 4;
}
else if (Rs >= 3886) {
return 2;
}
}
else if (Cs >= 471) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (Rs < 3268) {
return 2;
}
else if (Rs >= 3268) {
return 8;
}
}
}
}
else if (ssDCval >= 132) {
if (posBalance < 56) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (posBalance >= 56) {
return 4;
}
}
}
}
else if (Rs >= 4518) {
return 2;
}
}
else if (CsDiff >= 4) {
if (tcor < 80) {
if (CsDiff < 7) {
if (AFD < 30) {
if (RsDiff < 28) {
if (Cs < 833) {
if (Rs < 3670) {
return 8;
}
else if (Rs >= 3670) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
if (Rs < 4051) {
return 1;
}
else if (Rs >= 4051) {
return 8;
}
}
}
}
else if (Cs >= 833) {
return 8;
}
}
else if (RsDiff >= 28) {
return 2;
}
}
else if (AFD >= 30) {
if (ssDCval < 126) {
if (MVDiff < 43) {
if (Rs < 4054) {
if (Rs < 3813) {
return 2;
}
else if (Rs >= 3813) {
return 8;
}
}
else if (Rs >= 4054) {
return 2;
}
}
else if (MVDiff >= 43) {
return 8;
}
}
else if (ssDCval >= 126) {
if (MVDiff < 36) {
return 8;
}
else if (MVDiff >= 36) {
return 2;
}
}
}
}
else if (CsDiff >= 7) {
if (ssDCval < 134) {
if (diffTSC < -2) {
if (Rs < 4275) {
return 1;
}
else if (Rs >= 4275) {
return 8;
}
}
else if (diffTSC >= -2) {
if (Rs < 6326) {
if (AFD < 39) {
return 2;
}
else if (AFD >= 39) {
return 1;
}
}
else if (Rs >= 6326) {
return 8;
}
}
}
else if (ssDCval >= 134) {
return 1;
}
}
}
else if (tcor >= 80) {
if (RsDiff < 25) {
if (ssDCval < 143) {
if (Cs < 785) {
if (RsDiff < 16) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
if (Cs < 747) {
return 1;
}
else if (Cs >= 747) {
return 8;
}
}
}
else if (RsDiff >= 16) {
return 2;
}
}
else if (Cs >= 785) {
if (tcor < 92) {
if (posBalance < 48) {
if (diffRsCsdiff < -162) {
return 2;
}
else if (diffRsCsdiff >= -162) {
if (posBalance < 46) {
if (Rs < 5181) {
return 1;
}
else if (Rs >= 5181) {
return 2;
}
}
else if (posBalance >= 46) {
return 1;
}
}
}
else if (posBalance >= 48) {
if (CsDiff < 6) {
if (Cs < 921) {
if (Cs < 890) {
return 1;
}
else if (Cs >= 890) {
return 2;
}
}
else if (Cs >= 921) {
if (MVDiff < 43) {
return 8;
}
else if (MVDiff >= 43) {
return 1;
}
}
}
else if (CsDiff >= 6) {
if (Rs < 5799) {
return 1;
}
else if (Rs >= 5799) {
return 2;
}
}
}
}
else if (tcor >= 92) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 2;
}
}
}
}
else if (ssDCval >= 143) {
return 1;
}
}
else if (RsDiff >= 25) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1308) {
if (mu_mv_mag_sq < 111) {
if (RsCsDiff < 101) {
if (refDCval < 105) {
if (negBalance < 38) {
if (SC < 10477) {
if (Rs < 741) {
if (RsDiff < 1) {
if (Rs < 316) {
return 4;
}
else if (Rs >= 316) {
return 8;
}
}
else if (RsDiff >= 1) {
if (Cs < 2451) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (Cs >= 2451) {
return 2;
}
}
}
else if (Rs >= 741) {
if (Cs < 1975) {
if (Rs < 3192) {
if (tcor < 97) {
if (mu_mv_mag_sq < 82) {
if (Rs < 1502) {
if (refDCval < 66) {
if (mcTcor < 92) {
if (MVDiff < 15) {
if (diffAFD < 2) {
if (Cs < 1587) {
return 8;
}
else if (Cs >= 1587) {
return 4;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (MVDiff >= 15) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Cs < 1364) {
return 2;
}
else if (Cs >= 1364) {
return 8;
}
}
}
}
else if (mcTcor >= 92) {
if (refDCval < 55) {
return 2;
}
else if (refDCval >= 55) {
if (CsDiff < 5) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (CsDiff >= 5) {
if (Cs < 1709) {
if (refDCval < 63) {
return 4;
}
else if (refDCval >= 63) {
return 2;
}
}
else if (Cs >= 1709) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 66) {
if (posBalance < 59) {
if (diffRsCsdiff < -22) {
return 2;
}
else if (diffRsCsdiff >= -22) {
return 4;
}
}
else if (posBalance >= 59) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (Rs >= 1502) {
if (posBalance < 43) {
if (ssDCval < 72) {
if (ssDCval < 46) {
if (Rs < 2157) {
return 8;
}
else if (Rs >= 2157) {
return 1;
}
}
else if (ssDCval >= 46) {
return 2;
}
}
else if (ssDCval >= 72) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (RsDiff < 7) {
if (mu_mv_mag_sq < 81) {
if (MVDiff < 9) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 2134) {
return 2;
}
else if (Rs >= 2134) {
return 8;
}
}
}
else if (MVDiff >= 9) {
if (tcor < 95) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (mu_mv_mag_sq < 77) {
return 8;
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 2451) {
return 1;
}
else if (Rs >= 2451) {
return 8;
}
}
}
}
else if (tcor >= 95) {
if (MVDiff < 22) {
return 1;
}
else if (MVDiff >= 22) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (Cs < 1366) {
return 2;
}
else if (Cs >= 1366) {
if (Rs < 2050) {
return 8;
}
else if (Rs >= 2050) {
return 4;
}
}
}
}
else if (RsDiff >= 7) {
if (Cs < 1538) {
if (mcTcor < 90) {
if (Rs < 1732) {
return 1;
}
else if (Rs >= 1732) {
return 8;
}
}
else if (mcTcor >= 90) {
if (TSC < 7) {
if (Cs < 1511) {
return 2;
}
else if (Cs >= 1511) {
return 4;
}
}
else if (TSC >= 7) {
if (Cs < 1406) {
return 4;
}
else if (Cs >= 1406) {
return 8;
}
}
}
}
else if (Cs >= 1538) {
if (posBalance < 48) {
if (Rs < 2346) {
return 4;
}
else if (Rs >= 2346) {
return 1;
}
}
else if (posBalance >= 48) {
if (SC < 4799) {
return 8;
}
else if (SC >= 4799) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (posBalance < 38) {
if (ssDCval < 58) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (ssDCval >= 58) {
if (ssDCval < 67) {
return 2;
}
else if (ssDCval >= 67) {
return 8;
}
}
}
else if (posBalance >= 38) {
if (Cs < 1679) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (mcTcor < 95) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1747) {
return 4;
}
else if (Rs >= 1747) {
if (Rs < 2170) {
return 2;
}
else if (Rs >= 2170) {
return 8;
}
}
}
}
else if (mcTcor >= 95) {
if (diffRsCsdiff < -22) {
if (SC < 4544) {
return 2;
}
else if (SC >= 4544) {
if (Rs < 3112) {
return 4;
}
else if (Rs >= 3112) {
return 2;
}
}
}
else if (diffRsCsdiff >= -22) {
if (Rs < 2926) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (ssDCval < 58) {
if (Rs < 1693) {
return 4;
}
else if (Rs >= 1693) {
return 8;
}
}
else if (ssDCval >= 58) {
if (mu_mv_mag_sq < 93) {
if (Rs < 2339) {
if (gchDC < 1) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 2;
}
}
else if (gchDC >= 1) {
if (Rs < 1728) {
return 4;
}
else if (Rs >= 1728) {
return 2;
}
}
}
else if (Rs >= 2339) {
return 4;
}
}
else if (mu_mv_mag_sq >= 93) {
return 8;
}
}
}
}
else if (Rs >= 2926) {
return 2;
}
}
}
}
}
else if (Cs >= 1679) {
if (posBalance < 48) {
if (Cs < 1870) {
if (negBalance < 29) {
if (MVDiff < 8) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (MVDiff >= 8) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (negBalance >= 29) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
}
else if (Cs >= 1870) {
if (Rs < 2421) {
return 2;
}
else if (Rs >= 2421) {
return 1;
}
}
}
else if (posBalance >= 48) {
if (diffRsCsdiff < 44) {
if (TSC < 3) {
if (Rs < 1099) {
return 8;
}
else if (Rs >= 1099) {
return 4;
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (diffRsCsdiff >= 44) {
if (Rs < 2341) {
return 8;
}
else if (Rs >= 2341) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 97) {
if (Rs < 1773) {
if (ssDCval < 63) {
if (MVDiff < 6) {
if (Rs < 1119) {
return 8;
}
else if (Rs >= 1119) {
return 4;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (ssDCval >= 63) {
if (mu_mv_mag_sq < 101) {
return 4;
}
else if (mu_mv_mag_sq >= 101) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
}
}
else if (Rs >= 1773) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
if (SC < 4495) {
if (Rs < 2018) {
return 4;
}
else if (Rs >= 2018) {
if (Cs < 1607) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (Cs >= 1607) {
return 8;
}
}
}
else if (SC >= 4495) {
if (Rs < 3146) {
if (Cs < 1937) {
return 4;
}
else if (Cs >= 1937) {
return 1;
}
}
else if (Rs >= 3146) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3192) {
if (posBalance < 43) {
if (refDCval < 74) {
return 4;
}
else if (refDCval >= 74) {
if (Rs < 4226) {
if (MVDiff < 6) {
if (negBalance < 32) {
if (Rs < 4116) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (Rs >= 4116) {
if (Cs < 1618) {
return 8;
}
else if (Cs >= 1618) {
return 4;
}
}
}
else if (negBalance >= 32) {
if (Rs < 3873) {
if (CsDiff < 5) {
if (Cs < 1570) {
return 2;
}
else if (Cs >= 1570) {
if (Rs < 3439) {
return 4;
}
else if (Rs >= 3439) {
return 8;
}
}
}
else if (CsDiff >= 5) {
return 4;
}
}
else if (Rs >= 3873) {
if (Cs < 1658) {
return 4;
}
else if (Cs >= 1658) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 93) {
return 4;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
else if (Rs >= 4226) {
if (MVDiff < 3) {
if (CsDiff < 4) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Cs < 1845) {
if (TSC < 4) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (tcor < 97) {
return 4;
}
else if (tcor >= 97) {
return 2;
}
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (Cs >= 1845) {
if (Rs < 4819) {
return 2;
}
else if (Rs >= 4819) {
return 8;
}
}
}
}
else if (CsDiff >= 4) {
if (TSC < 7) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 4634) {
return 2;
}
else if (Rs >= 4634) {
return 8;
}
}
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 4836) {
return 4;
}
else if (Rs >= 4836) {
return 1;
}
}
}
}
}
else if (posBalance >= 43) {
if (refDCval < 99) {
if (RsCsDiff < 77) {
if (mu_mv_mag_sq < 66) {
if (mu_mv_mag_sq < 45) {
if (Cs < 1343) {
return 2;
}
else if (Cs >= 1343) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
if (Rs < 4220) {
return 8;
}
else if (Rs >= 4220) {
if (Rs < 4314) {
return 4;
}
else if (Rs >= 4314) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (diffAFD < -1) {
if (CsDiff < 6) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 2;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (diffAFD >= -1) {
if (ssDCval < 87) {
if (RsDiff < 7) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (diffRsCsdiff < -8) {
return 4;
}
else if (diffRsCsdiff >= -8) {
if (Rs < 3486) {
return 4;
}
else if (Rs >= 3486) {
return 8;
}
}
}
}
else if (RsDiff >= 7) {
if (ssDCval < 84) {
return 2;
}
else if (ssDCval >= 84) {
return 4;
}
}
}
else if (ssDCval >= 87) {
if (AFD < 11) {
if (SC < 5946) {
return 2;
}
else if (SC >= 5946) {
if (Rs < 4458) {
return 8;
}
else if (Rs >= 4458) {
return 2;
}
}
}
else if (AFD >= 11) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (MVDiff < 11) {
if (tcor < 90) {
if (Rs < 3969) {
return 4;
}
else if (Rs >= 3969) {
return 1;
}
}
else if (tcor >= 90) {
return 2;
}
}
else if (MVDiff >= 11) {
if (Rs < 3298) {
return 8;
}
else if (Rs >= 3298) {
return 1;
}
}
}
}
else if (refDCval >= 99) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
}
else if (Cs >= 1975) {
if (ssDCval < 87) {
if (posBalance < 45) {
if (refDCval < 56) {
if (Cs < 2577) {
if (Rs < 4906) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (ssDCval < 55) {
if (RsDiff < 8) {
if (Rs < 4768) {
if (Cs < 2307) {
return 4;
}
else if (Cs >= 2307) {
if (Rs < 4731) {
return 4;
}
else if (Rs >= 4731) {
return 4;
}
}
}
else if (Rs >= 4768) {
return 8;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (ssDCval >= 55) {
return 8;
}
}
}
else if (Rs >= 4906) {
return 4;
}
}
else if (Cs >= 2577) {
return 8;
}
}
else if (refDCval >= 56) {
if (refDCval < 68) {
if (Rs < 1485) {
return 8;
}
else if (Rs >= 1485) {
return 4;
}
}
else if (refDCval >= 68) {
if (SC < 6196) {
if (Cs < 2779) {
return 1;
}
else if (Cs >= 2779) {
if (Rs < 1771) {
return 8;
}
else if (Rs >= 1771) {
return 4;
}
}
}
else if (SC >= 6196) {
return 4;
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 1174) {
if (SC < 3049) {
return 1;
}
else if (SC >= 3049) {
return 8;
}
}
else if (Rs >= 1174) {
if (Rs < 1460) {
if (MVDiff < 4) {
if (Cs < 2173) {
return 8;
}
else if (Cs >= 2173) {
return 1;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (Rs >= 1460) {
if (posBalance < 48) {
if (Rs < 2494) {
return 8;
}
else if (Rs >= 2494) {
return 1;
}
}
else if (posBalance >= 48) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
if (Rs < 2873) {
if (SC < 4683) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 2323) {
return 4;
}
else if (Rs >= 2323) {
return 8;
}
}
}
else if (SC >= 4683) {
return 4;
}
}
else if (Rs >= 2873) {
if (ssDCval < 58) {
if (mu_mv_mag_sq < 107) {
return 4;
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 3774) {
return 8;
}
else if (Rs >= 3774) {
return 4;
}
}
}
else if (ssDCval >= 58) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (mu_mv_mag_sq < 96) {
if (Rs < 2368) {
if (MVDiff < 9) {
if (Rs < 1783) {
return 2;
}
else if (Rs >= 1783) {
return 4;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (Rs >= 2368) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
if (Cs < 2735) {
if (Rs < 2397) {
if (SC < 4456) {
if (posBalance < 37) {
if (Rs < 2338) {
return 1;
}
else if (Rs >= 2338) {
if (Rs < 2349) {
return 4;
}
else if (Rs >= 2349) {
return 1;
}
}
}
else if (posBalance >= 37) {
if (Rs < 2279) {
return 8;
}
else if (Rs >= 2279) {
return 2;
}
}
}
else if (SC >= 4456) {
if (posBalance < 14) {
return 8;
}
else if (posBalance >= 14) {
return 4;
}
}
}
else if (Rs >= 2397) {
if (RsDiff < 7) {
if (posBalance < 14) {
return 1;
}
else if (posBalance >= 14) {
return 8;
}
}
else if (RsDiff >= 7) {
if (Rs < 3884) {
return 8;
}
else if (Rs >= 3884) {
return 2;
}
}
}
}
else if (Cs >= 2735) {
if (Cs < 3288) {
return 4;
}
else if (Cs >= 3288) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 10477) {
if (Rs < 5682) {
return 8;
}
else if (Rs >= 5682) {
return 2;
}
}
}
else if (negBalance >= 38) {
if (Rs < 1454) {
if (RsCsDiff < 48) {
if (refDCval < 84) {
if (MVDiff < 4) {
if (Rs < 923) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 1;
}
}
else if (Rs >= 923) {
if (mu_mv_mag_sq < 72) {
return 4;
}
else if (mu_mv_mag_sq >= 72) {
if (negBalance < 42) {
if (SC < 2655) {
if (Rs < 1265) {
return 4;
}
else if (Rs >= 1265) {
return 1;
}
}
else if (SC >= 2655) {
if (mu_mv_mag_sq < 107) {
if (mu_mv_mag_sq < 86) {
if (Cs < 1671) {
return 4;
}
else if (Cs >= 1671) {
return 8;
}
}
else if (mu_mv_mag_sq >= 86) {
return 8;
}
}
else if (mu_mv_mag_sq >= 107) {
return 4;
}
}
}
else if (negBalance >= 42) {
if (Rs < 1434) {
return 8;
}
else if (Rs >= 1434) {
return 1;
}
}
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 77) {
if (ssDCval < 59) {
if (Rs < 944) {
if (mcTcor < 89) {
if (Rs < 860) {
return 4;
}
else if (Rs >= 860) {
return 8;
}
}
else if (mcTcor >= 89) {
return 2;
}
}
else if (Rs >= 944) {
if (negBalance < 47) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (negBalance >= 47) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
return 8;
}
}
}
}
else if (ssDCval >= 59) {
if (Rs < 1148) {
if (diffRsCsdiff < 3) {
if (TSC < 6) {
if (mcTcor < 98) {
if (tcor < 92) {
if (Rs < 1003) {
return 8;
}
else if (Rs >= 1003) {
return 2;
}
}
else if (tcor >= 92) {
if (Cs < 1609) {
return 1;
}
else if (Cs >= 1609) {
return 8;
}
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (TSC >= 6) {
if (posBalance < 56) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
else if (diffRsCsdiff >= 3) {
if (posBalance < 38) {
return 2;
}
else if (posBalance >= 38) {
return 1;
}
}
}
else if (Rs >= 1148) {
if (AFD < 7) {
if (Cs < 1670) {
return 4;
}
else if (Cs >= 1670) {
return 1;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (ssDCval >= 77) {
if (Rs < 753) {
return 8;
}
else if (Rs >= 753) {
if (negBalance < 44) {
if (Rs < 1178) {
return 8;
}
else if (Rs >= 1178) {
return 4;
}
}
else if (negBalance >= 44) {
return 4;
}
}
}
}
}
else if (refDCval >= 84) {
if (refDCval < 103) {
if (mu_mv_mag_sq < 3) {
if (Cs < 1385) {
if (posBalance < 66) {
return 4;
}
else if (posBalance >= 66) {
return 2;
}
}
else if (Cs >= 1385) {
if (Rs < 1209) {
return 8;
}
else if (Rs >= 1209) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 2230) {
return 8;
}
else if (Cs >= 2230) {
return 2;
}
}
}
else if (refDCval >= 103) {
if (Cs < 1556) {
return 4;
}
else if (Cs >= 1556) {
return 8;
}
}
}
}
else if (RsCsDiff >= 48) {
if (ssDCval < 55) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
if (RsCsDiff < 91) {
if (Rs < 1369) {
if (Cs < 1336) {
return 4;
}
else if (Cs >= 1336) {
return 2;
}
}
else if (Rs >= 1369) {
return 8;
}
}
else if (RsCsDiff >= 91) {
return 4;
}
}
}
else if (ssDCval >= 55) {
if (ssDCval < 67) {
if (RsCsDiff < 73) {
if (diffRsCsdiff < 15) {
return 8;
}
else if (diffRsCsdiff >= 15) {
if (diffRsCsdiff < 21) {
return 4;
}
else if (diffRsCsdiff >= 21) {
if (SC < 2521) {
return 2;
}
else if (SC >= 2521) {
return 8;
}
}
}
}
else if (RsCsDiff >= 73) {
if (tcor < 74) {
return 4;
}
else if (tcor >= 74) {
if (mcTcor < 96) {
if (refDCval < 59) {
return 4;
}
else if (refDCval >= 59) {
if (Rs < 1393) {
return 8;
}
else if (Rs >= 1393) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
}
}
}
else if (ssDCval >= 67) {
if (tcor < 88) {
if (TSC < 7) {
if (refDCval < 67) {
return 4;
}
else if (refDCval >= 67) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (Rs < 881) {
return 8;
}
else if (Rs >= 881) {
if (Rs < 1185) {
return 4;
}
else if (Rs >= 1185) {
return 2;
}
}
}
}
}
else if (TSC >= 7) {
if (negBalance < 52) {
if (Cs < 1698) {
if (AFD < 17) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (diffAFD < -1) {
if (Cs < 1534) {
return 8;
}
else if (Cs >= 1534) {
return 4;
}
}
else if (diffAFD >= -1) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
return 2;
}
}
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (Cs >= 1698) {
if (Cs < 2135) {
return 4;
}
else if (Cs >= 2135) {
if (Rs < 1176) {
return 2;
}
else if (Rs >= 1176) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (Cs < 1958) {
if (ssDCval < 91) {
if (Cs < 1591) {
if (Cs < 1392) {
if (AFD < 19) {
if (tcor < 76) {
return 4;
}
else if (tcor >= 76) {
if (diffRsCsdiff < -15) {
if (Rs < 1204) {
if (Rs < 1063) {
return 4;
}
else if (Rs >= 1063) {
return 8;
}
}
else if (Rs >= 1204) {
return 2;
}
}
else if (diffRsCsdiff >= -15) {
return 8;
}
}
}
else if (AFD >= 19) {
return 2;
}
}
else if (Cs >= 1392) {
if (Cs < 1453) {
return 4;
}
else if (Cs >= 1453) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (Rs < 1288) {
return 2;
}
else if (Rs >= 1288) {
return 8;
}
}
}
}
}
else if (Cs >= 1591) {
return 8;
}
}
else if (ssDCval >= 91) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
if (Rs < 1222) {
return 8;
}
else if (Rs >= 1222) {
return 4;
}
}
}
}
else if (Cs >= 1958) {
return 2;
}
}
}
}
else if (tcor >= 88) {
if (MVDiff < 14) {
if (mcTcor < 96) {
if (Rs < 1303) {
return 8;
}
else if (Rs >= 1303) {
return 2;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (MVDiff >= 14) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Cs < 2007) {
return 4;
}
else if (Cs >= 2007) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1454) {
if (CsDiff < 7) {
if (mcTcor < 93) {
if (posBalance < 70) {
if (refDCval < 90) {
if (ssDCval < 87) {
if (mu_mv_mag_sq < 91) {
if (RsCsDiff < 67) {
return 8;
}
else if (RsCsDiff >= 67) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Cs < 1439) {
if (negBalance < 49) {
if (MVDiff < 13) {
if (Rs < 2911) {
return 8;
}
else if (Rs >= 2911) {
return 4;
}
}
else if (MVDiff >= 13) {
if (Cs < 1431) {
return 1;
}
else if (Cs >= 1431) {
return 4;
}
}
}
else if (negBalance >= 49) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Cs < 1420) {
if (Rs < 1687) {
return 4;
}
else if (Rs >= 1687) {
return 8;
}
}
else if (Cs >= 1420) {
return 1;
}
}
}
}
else if (Cs >= 1439) {
if (MVDiff < 9) {
if (Rs < 1561) {
return 4;
}
else if (Rs >= 1561) {
if (Cs < 1849) {
return 8;
}
else if (Cs >= 1849) {
if (diffRsCsdiff < -9) {
if (AFD < 12) {
if (Rs < 3005) {
return 4;
}
else if (Rs >= 3005) {
return 8;
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (diffRsCsdiff >= -9) {
return 8;
}
}
}
}
else if (MVDiff >= 9) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (refDCval < 75) {
if (tcor < 80) {
if (ssDCval < 53) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Cs < 1455) {
return 8;
}
else if (Cs >= 1455) {
return 4;
}
}
}
else if (ssDCval >= 53) {
return 8;
}
}
else if (tcor >= 80) {
if (mu_mv_mag_sq < 92) {
if (Rs < 2119) {
return 2;
}
else if (Rs >= 2119) {
return 8;
}
}
else if (mu_mv_mag_sq >= 92) {
return 8;
}
}
}
else if (refDCval >= 75) {
if (ssDCval < 78) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 2043) {
return 8;
}
else if (Rs >= 2043) {
return 4;
}
}
}
else if (ssDCval >= 78) {
return 8;
}
}
}
}
else if (ssDCval >= 87) {
if (negBalance < 57) {
if (posBalance < 53) {
if (Rs < 2130) {
return 8;
}
else if (Rs >= 2130) {
return 2;
}
}
else if (posBalance >= 53) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
if (Rs < 1746) {
return 8;
}
else if (Rs >= 1746) {
return 2;
}
}
}
}
else if (negBalance >= 57) {
if (Rs < 1764) {
return 4;
}
else if (Rs >= 1764) {
return 8;
}
}
}
}
else if (refDCval >= 90) {
if (Rs < 2227) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (RsCsDiff < 57) {
return 8;
}
else if (RsCsDiff >= 57) {
if (AFD < 12) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (Rs >= 2227) {
return 8;
}
}
}
else if (posBalance >= 70) {
if (Rs < 2463) {
return 4;
}
else if (Rs >= 2463) {
return 1;
}
}
}
else if (mcTcor >= 93) {
if (RsCsDiff < 41) {
if (SC < 6522) {
if (SC < 5082) {
if (gchDC < 2) {
if (SC < 3635) {
if (AFD < 10) {
if (refDCval < 83) {
if (Rs < 1462) {
return 1;
}
else if (Rs >= 1462) {
if (Rs < 2084) {
if (ssDCval < 82) {
if (Cs < 1324) {
if (tcor < 96) {
if (Cs < 1322) {
return 8;
}
else if (Cs >= 1322) {
return 1;
}
}
else if (tcor >= 96) {
return 1;
}
}
else if (Cs >= 1324) {
if (diffRsCsdiff < -16) {
if (Rs < 1938) {
return 8;
}
else if (Rs >= 1938) {
return 1;
}
}
else if (diffRsCsdiff >= -16) {
return 8;
}
}
}
else if (ssDCval >= 82) {
return 1;
}
}
else if (Rs >= 2084) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (refDCval >= 83) {
if (mu_mv_mag_sq < 100) {
return 8;
}
else if (mu_mv_mag_sq >= 100) {
return 2;
}
}
}
else if (AFD >= 10) {
return 4;
}
}
else if (SC >= 3635) {
if (negBalance < 52) {
if (Rs < 3574) {
if (diffRsCsdiff < -26) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Rs < 2105) {
return 2;
}
else if (Rs >= 2105) {
if (Rs < 2858) {
return 8;
}
else if (Rs >= 2858) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= -26) {
if (Rs < 2027) {
if (Rs < 2013) {
return 8;
}
else if (Rs >= 2013) {
if (Cs < 1689) {
return 8;
}
else if (Cs >= 1689) {
return 4;
}
}
}
else if (Rs >= 2027) {
return 8;
}
}
}
else if (Rs >= 3574) {
return 2;
}
}
else if (negBalance >= 52) {
if (TSC < 6) {
if (Cs < 1447) {
return 1;
}
else if (Cs >= 1447) {
return 8;
}
}
else if (TSC >= 6) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (tcor < 99) {
if (diffTSC < -1) {
if (Cs < 1971) {
return 2;
}
else if (Cs >= 1971) {
return 1;
}
}
else if (diffTSC >= -1) {
if (mcTcor < 95) {
if (Rs < 1955) {
return 8;
}
else if (Rs >= 1955) {
return 1;
}
}
else if (mcTcor >= 95) {
return 4;
}
}
}
else if (tcor >= 99) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
}
else if (SC >= 5082) {
if (Rs < 2832) {
if (Cs < 2584) {
return 8;
}
else if (Cs >= 2584) {
return 4;
}
}
else if (Rs >= 2832) {
if (Rs < 3867) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Rs >= 3867) {
return 4;
}
}
}
}
else if (SC >= 6522) {
return 8;
}
}
else if (RsCsDiff >= 41) {
if (Cs < 3438) {
if (SC < 4842) {
if (TSC < 9) {
if (mu_mv_mag_sq < 9) {
if (ssDCval < 80) {
if (posBalance < 57) {
if (Rs < 2647) {
return 8;
}
else if (Rs >= 2647) {
return 2;
}
}
else if (posBalance >= 57) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (ssDCval >= 80) {
if (Rs < 2069) {
if (tcor < 94) {
if (SC < 3166) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 8;
}
}
else if (SC >= 3166) {
if (Rs < 1930) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 2;
}
}
else if (Rs >= 1930) {
if (Rs < 1957) {
return 1;
}
else if (Rs >= 1957) {
return 2;
}
}
}
}
else if (tcor >= 94) {
return 2;
}
}
else if (Rs >= 2069) {
if (TSC < 7) {
if (RsCsDiff < 70) {
if (Cs < 1358) {
return 1;
}
else if (Cs >= 1358) {
return 8;
}
}
else if (RsCsDiff >= 70) {
if (Rs < 2325) {
return 8;
}
else if (Rs >= 2325) {
return 2;
}
}
}
else if (TSC >= 7) {
if (SC < 3627) {
if (Cs < 1363) {
return 8;
}
else if (Cs >= 1363) {
return 1;
}
}
else if (SC >= 3627) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (refDCval < 89) {
if (Rs < 2896) {
if (ssDCval < 80) {
if (Cs < 1325) {
if (MVDiff < 7) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (MVDiff >= 7) {
if (Cs < 1316) {
return 8;
}
else if (Cs >= 1316) {
return 4;
}
}
}
else if (Cs >= 1325) {
if (Cs < 1600) {
if (mu_mv_mag_sq < 104) {
return 8;
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 2081) {
return 8;
}
else if (Rs >= 2081) {
return 2;
}
}
}
else if (Cs >= 1600) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 93) {
if (RsDiff < 6) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (Rs < 2304) {
return 4;
}
else if (Rs >= 2304) {
return 8;
}
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 93) {
if (Cs < 1813) {
return 8;
}
else if (Cs >= 1813) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 80) {
if (mu_mv_mag_sq < 91) {
if (MVDiff < 5) {
if (Cs < 1672) {
if (posBalance < 54) {
if (refDCval < 82) {
return 8;
}
else if (refDCval >= 82) {
if (Rs < 2021) {
return 1;
}
else if (Rs >= 2021) {
return 2;
}
}
}
else if (posBalance >= 54) {
if (TSC < 7) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (TSC >= 7) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (Rs < 2255) {
return 1;
}
else if (Rs >= 2255) {
return 8;
}
}
}
}
}
else if (Cs >= 1672) {
if (Rs < 2159) {
return 1;
}
else if (Rs >= 2159) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
}
else if (MVDiff >= 5) {
if (tcor < 89) {
if (Cs < 1745) {
return 4;
}
else if (Cs >= 1745) {
return 1;
}
}
else if (tcor >= 89) {
if (Cs < 1482) {
if (mu_mv_mag_sq < 58) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
return 8;
}
}
else if (mu_mv_mag_sq >= 58) {
if (diffRsCsdiff < -5) {
if (Rs < 1877) {
return 4;
}
else if (Rs >= 1877) {
return 1;
}
}
else if (diffRsCsdiff >= -5) {
return 2;
}
}
}
else if (Cs >= 1482) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (AFD < 7) {
if (Rs < 2303) {
return 2;
}
else if (Rs >= 2303) {
return 8;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
}
else if (Rs >= 2896) {
if (ssDCval < 85) {
if (Cs < 1587) {
return 2;
}
else if (Cs >= 1587) {
return 8;
}
}
else if (ssDCval >= 85) {
return 4;
}
}
}
else if (refDCval >= 89) {
if (Rs < 2085) {
if (diffAFD < 1) {
if (Cs < 2215) {
if (mcTcor < 98) {
if (negBalance < 40) {
if (Cs < 1449) {
return 8;
}
else if (Cs >= 1449) {
return 2;
}
}
else if (negBalance >= 40) {
if (Cs < 1356) {
return 1;
}
else if (Cs >= 1356) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
}
else if (mcTcor >= 98) {
return 2;
}
}
else if (Cs >= 2215) {
return 2;
}
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (Rs >= 2085) {
if (posBalance < 47) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
return 4;
}
}
else if (posBalance >= 47) {
if (Cs < 1929) {
if (diffRsCsdiff < -53) {
if (Cs < 1388) {
return 8;
}
else if (Cs >= 1388) {
return 4;
}
}
else if (diffRsCsdiff >= -53) {
if (posBalance < 49) {
if (Cs < 1430) {
return 2;
}
else if (Cs >= 1430) {
if (negBalance < 43) {
return 8;
}
else if (negBalance >= 43) {
return 4;
}
}
}
else if (posBalance >= 49) {
if (AFD < 15) {
if (refDCval < 92) {
if (Rs < 2549) {
return 8;
}
else if (Rs >= 2549) {
return 4;
}
}
else if (refDCval >= 92) {
return 8;
}
}
else if (AFD >= 15) {
return 4;
}
}
}
}
else if (Cs >= 1929) {
return 4;
}
}
}
}
}
}
else if (TSC >= 9) {
if (Rs < 2012) {
if (AFD < 15) {
if (refDCval < 101) {
if (SC < 3813) {
return 8;
}
else if (SC >= 3813) {
return 4;
}
}
else if (refDCval >= 101) {
return 2;
}
}
else if (AFD >= 15) {
if (Cs < 1462) {
if (Rs < 1831) {
return 8;
}
else if (Rs >= 1831) {
return 1;
}
}
else if (Cs >= 1462) {
if (Cs < 1577) {
return 2;
}
else if (Cs >= 1577) {
return 4;
}
}
}
}
else if (Rs >= 2012) {
return 8;
}
}
}
else if (SC >= 4842) {
if (Cs < 1625) {
if (tcor < 92) {
if (ssDCval < 80) {
return 4;
}
else if (ssDCval >= 80) {
if (ssDCval < 96) {
return 1;
}
else if (ssDCval >= 96) {
return 8;
}
}
}
else if (tcor >= 92) {
if (SC < 5582) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (SC >= 5582) {
return 1;
}
}
}
else if (Cs >= 1625) {
if (mu_mv_mag_sq < 31) {
return 8;
}
else if (mu_mv_mag_sq >= 31) {
if (Rs < 4334) {
if (Cs < 2991) {
if (negBalance < 55) {
if (refDCval < 79) {
if (Rs < 3508) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (Rs >= 3508) {
if (negBalance < 39) {
return 2;
}
else if (negBalance >= 39) {
return 4;
}
}
}
else if (refDCval >= 79) {
if (Rs < 3317) {
if (refDCval < 87) {
if (negBalance < 43) {
if (SC < 4899) {
return 8;
}
else if (SC >= 4899) {
return 4;
}
}
else if (negBalance >= 43) {
return 8;
}
}
else if (refDCval >= 87) {
if (Rs < 3091) {
return 4;
}
else if (Rs >= 3091) {
return 8;
}
}
}
else if (Rs >= 3317) {
if (diffMVdiffVal < -4) {
if (Rs < 3840) {
return 2;
}
else if (Rs >= 3840) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
if (posBalance < 47) {
if (Rs < 3971) {
return 8;
}
else if (Rs >= 3971) {
return 1;
}
}
else if (posBalance >= 47) {
return 8;
}
}
}
}
}
else if (negBalance >= 55) {
if (Rs < 3621) {
return 4;
}
else if (Rs >= 3621) {
return 2;
}
}
}
else if (Cs >= 2991) {
if (refDCval < 80) {
if (Rs < 3470) {
return 1;
}
else if (Rs >= 3470) {
return 8;
}
}
else if (refDCval >= 80) {
if (Rs < 2975) {
return 4;
}
else if (Rs >= 2975) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
return 4;
}
}
}
}
}
else if (Rs >= 4334) {
if (mu_mv_mag_sq < 96) {
return 4;
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 4527) {
return 2;
}
else if (Rs >= 4527) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 3438) {
return 8;
}
}
}
}
else if (CsDiff >= 7) {
if (Rs < 3009) {
if (Cs < 2908) {
if (Cs < 2304) {
if (AFD < 17) {
if (ssDCval < 53) {
if (TSC < 9) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (tcor < 54) {
return 1;
}
else if (tcor >= 54) {
return 8;
}
}
}
else if (TSC >= 9) {
return 4;
}
}
else if (ssDCval >= 53) {
if (diffTSC < -2) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
return 4;
}
}
else if (diffTSC >= -2) {
if (SC < 3611) {
if (SC < 3535) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (Rs < 1750) {
return 8;
}
else if (Rs >= 1750) {
return 1;
}
}
}
else if (SC >= 3535) {
if (AFD < 14) {
if (Rs < 2070) {
return 2;
}
else if (Rs >= 2070) {
return 1;
}
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (SC >= 3611) {
if (Rs < 2871) {
if (Rs < 1825) {
if (Rs < 1689) {
return 8;
}
else if (Rs >= 1689) {
return 4;
}
}
else if (Rs >= 1825) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
if (Rs < 2087) {
return 8;
}
else if (Rs >= 2087) {
return 4;
}
}
}
}
else if (Rs >= 2871) {
if (Rs < 2932) {
return 4;
}
else if (Rs >= 2932) {
return 8;
}
}
}
}
}
}
else if (AFD >= 17) {
if (MVDiff < 41) {
if (posBalance < 58) {
if (Cs < 1867) {
return 4;
}
else if (Cs >= 1867) {
return 2;
}
}
else if (posBalance >= 58) {
if (refDCval < 98) {
if (Rs < 2514) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Rs < 1640) {
return 4;
}
else if (Rs >= 1640) {
return 8;
}
}
}
else if (Rs >= 2514) {
return 4;
}
}
else if (refDCval >= 98) {
if (Rs < 1684) {
return 4;
}
else if (Rs >= 1684) {
return 2;
}
}
}
}
else if (MVDiff >= 41) {
return 1;
}
}
}
else if (Cs >= 2304) {
if (mcTcor < 93) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 81) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
return 8;
}
}
else if (ssDCval >= 81) {
if (TSC < 10) {
if (Rs < 1660) {
return 2;
}
else if (Rs >= 1660) {
return 1;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
else if (Cs >= 2908) {
return 4;
}
}
else if (Rs >= 3009) {
if (mu_mv_mag_sq < 90) {
if (Rs < 3450) {
if (Cs < 2973) {
return 8;
}
else if (Cs >= 2973) {
return 1;
}
}
else if (Rs >= 3450) {
if (TSC < 10) {
if (negBalance < 48) {
if (Cs < 1869) {
return 1;
}
else if (Cs >= 1869) {
return 8;
}
}
else if (negBalance >= 48) {
return 8;
}
}
else if (TSC >= 10) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (Cs < 2275) {
return 1;
}
else if (Cs >= 2275) {
if (Cs < 3316) {
return 4;
}
else if (Cs >= 3316) {
if (Rs < 3836) {
return 1;
}
else if (Rs >= 3836) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 105) {
if (Rs < 4405) {
if (ssDCval < 162) {
if (Rs < 1247) {
if (Cs < 1498) {
if (ssDCval < 142) {
if (Rs < 658) {
if (Rs < 595) {
return 2;
}
else if (Rs >= 595) {
return 1;
}
}
else if (Rs >= 658) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (diffRsCsdiff < 30) {
if (Rs < 1239) {
if (mu_mv_mag_sq < 107) {
return 8;
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 945) {
return 8;
}
else if (Rs >= 945) {
return 4;
}
}
}
else if (Rs >= 1239) {
if (Rs < 1244) {
return 1;
}
else if (Rs >= 1244) {
return 8;
}
}
}
else if (diffRsCsdiff >= 30) {
return 1;
}
}
}
}
else if (ssDCval >= 142) {
if (CsDiff < 6) {
if (Rs < 1182) {
return 4;
}
else if (Rs >= 1182) {
return 2;
}
}
else if (CsDiff >= 6) {
if (Rs < 977) {
return 8;
}
else if (Rs >= 977) {
return 1;
}
}
}
}
else if (Cs >= 1498) {
if (CsDiff < 6) {
if (mcTcor < 98) {
if (posBalance < 56) {
if (Rs < 1027) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (Rs >= 1027) {
return 4;
}
}
else if (posBalance >= 56) {
if (SC < 2794) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (SC >= 2794) {
if (diffRsCsdiff < 5) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
return 8;
}
}
else if (diffRsCsdiff >= 5) {
if (Rs < 1224) {
return 8;
}
else if (Rs >= 1224) {
return 4;
}
}
}
}
}
else if (mcTcor >= 98) {
if (Cs < 1527) {
return 4;
}
else if (Cs >= 1527) {
return 8;
}
}
}
else if (CsDiff >= 6) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 2;
}
}
}
}
else if (Rs >= 1247) {
if (mu_mv_mag_sq < 40) {
if (Rs < 2017) {
if (negBalance < 48) {
if (refDCval < 124) {
if (mcTcor < 97) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (Rs < 1533) {
return 8;
}
else if (Rs >= 1533) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (refDCval >= 124) {
if (ssDCval < 139) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
return 8;
}
}
}
else if (ssDCval >= 139) {
if (Rs < 1603) {
return 2;
}
else if (Rs >= 1603) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
if (posBalance < 63) {
if (mcTcor < 98) {
if (SC < 3180) {
if (mu_mv_mag_sq < 12) {
if (tcor < 82) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (tcor >= 82) {
if (RsCsDiff < 67) {
if (SC < 3164) {
return 8;
}
else if (SC >= 3164) {
return 2;
}
}
else if (RsCsDiff >= 67) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 12) {
return 8;
}
}
else if (SC >= 3180) {
return 8;
}
}
else if (mcTcor >= 98) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (posBalance >= 63) {
if (Cs < 1439) {
if (Rs < 1350) {
return 1;
}
else if (Rs >= 1350) {
return 4;
}
}
else if (Cs >= 1439) {
return 2;
}
}
}
}
else if (Rs >= 2017) {
if (posBalance < 47) {
if (refDCval < 111) {
if (Rs < 3208) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
return 8;
}
}
else if (Rs >= 3208) {
return 8;
}
}
else if (refDCval >= 111) {
if (Cs < 2723) {
return 8;
}
else if (Cs >= 2723) {
return 2;
}
}
}
else if (posBalance >= 47) {
if (ssDCval < 110) {
if (posBalance < 62) {
if (RsDiff < 6) {
if (Rs < 2877) {
return 4;
}
else if (Rs >= 2877) {
return 8;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (posBalance >= 62) {
if (Cs < 1680) {
return 1;
}
else if (Cs >= 1680) {
if (Rs < 2312) {
return 4;
}
else if (Rs >= 2312) {
return 8;
}
}
}
}
else if (ssDCval >= 110) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (Rs < 2942) {
return 8;
}
else if (Rs >= 2942) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (mcTcor < 97) {
if (posBalance < 57) {
if (refDCval < 109) {
if (mcTcor < 96) {
if (diffRsCsdiff < 23) {
return 8;
}
else if (diffRsCsdiff >= 23) {
if (Cs < 2106) {
return 4;
}
else if (Cs >= 2106) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 107) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (refDCval >= 107) {
if (Cs < 1392) {
return 8;
}
else if (Cs >= 1392) {
return 4;
}
}
}
}
else if (refDCval >= 109) {
if (ssDCval < 141) {
if (refDCval < 126) {
if (TSC < 4) {
if (Rs < 2566) {
return 8;
}
else if (Rs >= 2566) {
if (Rs < 3422) {
return 4;
}
else if (Rs >= 3422) {
return 1;
}
}
}
else if (TSC >= 4) {
if (diffRsCsdiff < -82) {
if (Rs < 3689) {
return 8;
}
else if (Rs >= 3689) {
return 4;
}
}
else if (diffRsCsdiff >= -82) {
return 8;
}
}
}
else if (refDCval >= 126) {
if (SC < 4685) {
if (Cs < 1674) {
if (posBalance < 38) {
if (Rs < 2508) {
return 8;
}
else if (Rs >= 2508) {
return 4;
}
}
else if (posBalance >= 38) {
return 8;
}
}
else if (Cs >= 1674) {
if (Cs < 1845) {
if (refDCval < 131) {
return 1;
}
else if (refDCval >= 131) {
if (mu_mv_mag_sq < 58) {
if (Rs < 1779) {
return 8;
}
else if (Rs >= 1779) {
return 1;
}
}
else if (mu_mv_mag_sq >= 58) {
return 8;
}
}
}
else if (Cs >= 1845) {
return 8;
}
}
}
else if (SC >= 4685) {
if (mu_mv_mag_sq < 105) {
return 8;
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 3654) {
return 8;
}
else if (Rs >= 3654) {
return 1;
}
}
}
}
}
else if (ssDCval >= 141) {
if (SC < 5884) {
return 8;
}
else if (SC >= 5884) {
return 1;
}
}
}
}
else if (posBalance >= 57) {
if (Rs < 3655) {
if (AFD < 13) {
if (TSC < 11) {
if (refDCval < 123) {
return 8;
}
else if (refDCval >= 123) {
if (mu_mv_mag_sq < 102) {
if (ssDCval < 130) {
if (tcor < 90) {
if (SC < 3605) {
return 1;
}
else if (SC >= 3605) {
if (Rs < 3101) {
return 8;
}
else if (Rs >= 3101) {
return 1;
}
}
}
else if (tcor >= 90) {
return 8;
}
}
else if (ssDCval >= 130) {
return 8;
}
}
else if (mu_mv_mag_sq >= 102) {
return 1;
}
}
}
else if (TSC >= 11) {
if (Rs < 2105) {
return 8;
}
else if (Rs >= 2105) {
return 1;
}
}
}
else if (AFD >= 13) {
if (tcor < 89) {
if (Cs < 2208) {
return 8;
}
else if (Cs >= 2208) {
if (Rs < 2923) {
return 4;
}
else if (Rs >= 2923) {
return 8;
}
}
}
else if (tcor >= 89) {
if (Cs < 1726) {
return 1;
}
else if (Cs >= 1726) {
return 4;
}
}
}
}
else if (Rs >= 3655) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
return 8;
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 121) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (SC < 4257) {
if (Cs < 1790) {
return 8;
}
else if (Cs >= 1790) {
if (mcTcor < 100) {
if (Cs < 1846) {
return 2;
}
else if (Cs >= 1846) {
return 8;
}
}
else if (mcTcor >= 100) {
if (Rs < 2149) {
return 2;
}
else if (Rs >= 2149) {
return 4;
}
}
}
}
else if (SC >= 4257) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Cs < 3384) {
if (Rs < 3101) {
return 2;
}
else if (Rs >= 3101) {
if (Rs < 3726) {
return 8;
}
else if (Rs >= 3726) {
return 4;
}
}
}
else if (Cs >= 3384) {
return 8;
}
}
}
}
}
else if (ssDCval >= 121) {
if (Cs < 1363) {
if (Rs < 2272) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
else if (Rs >= 2272) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (MVDiff < 7) {
if (Cs < 1324) {
return 1;
}
else if (Cs >= 1324) {
return 2;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
else if (Cs >= 1363) {
if (RsCsDiff < 2) {
if (SC < 5518) {
return 4;
}
else if (SC >= 5518) {
if (Rs < 2734) {
return 8;
}
else if (Rs >= 2734) {
if (Rs < 2885) {
return 2;
}
else if (Rs >= 2885) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
if (posBalance < 24) {
if (Rs < 2750) {
return 8;
}
else if (Rs >= 2750) {
if (Rs < 2799) {
return 4;
}
else if (Rs >= 2799) {
return 2;
}
}
}
else if (posBalance >= 24) {
if (Cs < 1416) {
if (ssDCval < 135) {
return 8;
}
else if (ssDCval >= 135) {
if (Rs < 2425) {
return 8;
}
else if (Rs >= 2425) {
return 1;
}
}
}
else if (Cs >= 1416) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 162) {
if (Rs < 2266) {
if (Rs < 1064) {
return 1;
}
else if (Rs >= 1064) {
return 2;
}
}
else if (Rs >= 2266) {
return 8;
}
}
}
else if (Rs >= 4405) {
if (Cs < 1851) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (Cs < 1712) {
return 1;
}
else if (Cs >= 1712) {
return 8;
}
}
}
else if (Cs >= 1851) {
if (RsCsDiff < 98) {
if (Cs < 3621) {
if (Rs < 5518) {
if (Rs < 4414) {
return 2;
}
else if (Rs >= 4414) {
return 8;
}
}
else if (Rs >= 5518) {
return 4;
}
}
else if (Cs >= 3621) {
return 8;
}
}
else if (RsCsDiff >= 98) {
if (Cs < 1918) {
return 8;
}
else if (Cs >= 1918) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1925) {
if (Rs < 3318) {
if (AFD < 19) {
if (refDCval < 83) {
if (mcTcor < 71) {
if (AFD < 13) {
if (RsDiff < 12) {
if (Rs < 1492) {
return 4;
}
else if (Rs >= 1492) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 4;
}
}
}
else if (RsDiff >= 12) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (AFD >= 13) {
if (refDCval < 42) {
return 8;
}
else if (refDCval >= 42) {
return 4;
}
}
}
else if (mcTcor >= 71) {
if (posBalance < 49) {
if (negBalance < 52) {
if (TSC < 7) {
if (mcTcor < 94) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Cs < 1399) {
return 8;
}
else if (Cs >= 1399) {
return 1;
}
}
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (TSC >= 7) {
if (mcTcor < 85) {
if (RsCsDiff < 129) {
return 4;
}
else if (RsCsDiff >= 129) {
return 1;
}
}
else if (mcTcor >= 85) {
if (Rs < 3196) {
if (RsDiff < 10) {
if (tcor < 83) {
return 2;
}
else if (tcor >= 83) {
if (Rs < 2977) {
return 1;
}
else if (Rs >= 2977) {
return 4;
}
}
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (Rs >= 3196) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (mu_mv_mag_sq < 31) {
if (MVDiff < 53) {
if (Rs < 2204) {
return 1;
}
else if (Rs >= 2204) {
if (Rs < 2636) {
return 8;
}
else if (Rs >= 2636) {
return 2;
}
}
}
else if (MVDiff >= 53) {
return 8;
}
}
else if (mu_mv_mag_sq >= 31) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (Rs < 1793) {
if (negBalance < 55) {
if (mu_mv_mag_sq < 3) {
if (AFD < 18) {
if (MVDiff < 44) {
if (diffMVdiffVal < -30) {
return 8;
}
else if (diffMVdiffVal >= -30) {
return 1;
}
}
else if (MVDiff >= 44) {
return 4;
}
}
else if (AFD >= 18) {
if (Rs < 1132) {
return 2;
}
else if (Rs >= 1132) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (MVDiff < 13) {
if (Cs < 1734) {
return 8;
}
else if (Cs >= 1734) {
if (Rs < 1699) {
return 4;
}
else if (Rs >= 1699) {
return 8;
}
}
}
else if (MVDiff >= 13) {
if (RsCsDiff < 141) {
if (diffMVdiffVal < 1) {
if (Rs < 655) {
return 2;
}
else if (Rs >= 655) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 4;
}
}
}
else if (RsCsDiff >= 141) {
if (mcTcor < 93) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
return 2;
}
}
else if (mcTcor >= 93) {
return 8;
}
}
}
}
}
else if (negBalance >= 55) {
if (TSC < 11) {
if (tcor < 77) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
return 4;
}
}
else if (tcor >= 77) {
if (Rs < 1277) {
if (Rs < 993) {
return 8;
}
else if (Rs >= 993) {
return 1;
}
}
else if (Rs >= 1277) {
return 2;
}
}
}
else if (TSC >= 11) {
if (Rs < 1497) {
return 2;
}
else if (Rs >= 1497) {
return 4;
}
}
}
}
else if (Rs >= 1793) {
if (Rs < 2976) {
if (mu_mv_mag_sq < 3) {
if (mcTcor < 92) {
if (diffRsCsdiff < -18) {
if (RsCsDiff < 194) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (RsCsDiff >= 194) {
return 2;
}
}
else if (diffRsCsdiff >= -18) {
if (RsCsDiff < 163) {
return 8;
}
else if (RsCsDiff >= 163) {
if (RsCsDiff < 219) {
return 4;
}
else if (RsCsDiff >= 219) {
return 8;
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 2170) {
return 8;
}
else if (Rs >= 2170) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 14) {
if (Cs < 1821) {
if (RsDiff < 13) {
if (negBalance < 35) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 4;
}
}
else if (negBalance >= 35) {
return 8;
}
}
else if (RsDiff >= 13) {
if (Rs < 2832) {
return 4;
}
else if (Rs >= 2832) {
return 2;
}
}
}
else if (Cs >= 1821) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (AFD >= 14) {
if (mcTcor < 88) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
else if (mcTcor >= 88) {
if (mu_mv_mag_sq < 50) {
if (TSC < 10) {
if (diffMVdiffVal < 0) {
if (diffRsCsdiff < -21) {
return 2;
}
else if (diffRsCsdiff >= -21) {
if (Rs < 2143) {
return 1;
}
else if (Rs >= 2143) {
return 4;
}
}
}
else if (diffMVdiffVal >= 0) {
if (Rs < 2334) {
return 8;
}
else if (Rs >= 2334) {
return 2;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (mu_mv_mag_sq >= 50) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2976) {
if (Cs < 1583) {
return 2;
}
else if (Cs >= 1583) {
if (Rs < 3033) {
return 4;
}
else if (Rs >= 3033) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 83) {
if (TSC < 9) {
if (refDCval < 97) {
if (CsDiff < 8) {
if (RsCsDiff < 121) {
if (Cs < 1607) {
if (Cs < 1525) {
if (ssDCval < 85) {
return 4;
}
else if (ssDCval >= 85) {
return 2;
}
}
else if (Cs >= 1525) {
return 4;
}
}
else if (Cs >= 1607) {
if (Rs < 2743) {
return 8;
}
else if (Rs >= 2743) {
return 2;
}
}
}
else if (RsCsDiff >= 121) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
if (Rs < 2294) {
return 2;
}
else if (Rs >= 2294) {
return 4;
}
}
}
}
else if (CsDiff >= 8) {
if (mcTcor < 97) {
if (AFD < 16) {
if (Rs < 1394) {
return 1;
}
else if (Rs >= 1394) {
return 8;
}
}
else if (AFD >= 16) {
if (MVDiff < 16) {
if (CsDiff < 9) {
return 4;
}
else if (CsDiff >= 9) {
if (Rs < 1935) {
return 4;
}
else if (Rs >= 1935) {
return 2;
}
}
}
else if (MVDiff >= 16) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (Rs < 1298) {
return 2;
}
else if (Rs >= 1298) {
return 1;
}
}
}
}
else if (refDCval >= 97) {
if (MVDiff < 10) {
if (tcor < 93) {
if (Cs < 1431) {
if (Rs < 2097) {
return 8;
}
else if (Rs >= 2097) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
}
else if (Cs >= 1431) {
return 8;
}
}
else if (tcor >= 93) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (MVDiff >= 10) {
if (negBalance < 27) {
if (ssDCval < 132) {
if (Rs < 2839) {
return 1;
}
else if (Rs >= 2839) {
return 8;
}
}
else if (ssDCval >= 132) {
return 2;
}
}
else if (negBalance >= 27) {
if (MVDiff < 22) {
if (RsCsDiff < 141) {
if (Rs < 3274) {
if (negBalance < 45) {
if (diffRsCsdiff < 8) {
return 4;
}
else if (diffRsCsdiff >= 8) {
if (Rs < 1435) {
return 4;
}
else if (Rs >= 1435) {
return 8;
}
}
}
else if (negBalance >= 45) {
if (Rs < 2187) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (Rs >= 2187) {
return 8;
}
}
}
else if (Rs >= 3274) {
return 1;
}
}
else if (RsCsDiff >= 141) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1174) {
return 2;
}
else if (Rs >= 1174) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (tcor < 86) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (Rs < 1184) {
return 2;
}
else if (Rs >= 1184) {
return 8;
}
}
}
else if (tcor >= 86) {
return 4;
}
}
}
}
else if (MVDiff >= 22) {
if (Rs < 3157) {
return 8;
}
else if (Rs >= 3157) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 4;
}
}
}
}
}
}
}
else if (TSC >= 9) {
if (MVDiff < 20) {
if (Rs < 2963) {
if (TSC < 11) {
if (tcor < 75) {
if (Cs < 1854) {
if (SC < 3328) {
return 8;
}
else if (SC >= 3328) {
if (Rs < 2444) {
return 4;
}
else if (Rs >= 2444) {
if (Rs < 2639) {
return 2;
}
else if (Rs >= 2639) {
if (Rs < 2656) {
return 4;
}
else if (Rs >= 2656) {
return 8;
}
}
}
}
}
else if (Cs >= 1854) {
return 8;
}
}
else if (tcor >= 75) {
if (ssDCval < 87) {
if (mcTcor < 92) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
if (Rs < 2610) {
return 2;
}
else if (Rs >= 2610) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (Cs < 1348) {
return 1;
}
else if (Cs >= 1348) {
if (Rs < 2329) {
return 8;
}
else if (Rs >= 2329) {
if (Cs < 1486) {
return 8;
}
else if (Cs >= 1486) {
return 4;
}
}
}
}
}
else if (ssDCval >= 87) {
if (Rs < 1463) {
return 4;
}
else if (Rs >= 1463) {
if (Rs < 2845) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Rs < 2277) {
return 2;
}
else if (Rs >= 2277) {
return 8;
}
}
}
else if (Rs >= 2845) {
if (ssDCval < 113) {
return 4;
}
else if (ssDCval >= 113) {
return 8;
}
}
}
}
}
}
else if (TSC >= 11) {
if (diffRsCsdiff < -116) {
return 2;
}
else if (diffRsCsdiff >= -116) {
return 8;
}
}
}
else if (Rs >= 2963) {
if (mu_mv_mag_sq < 5) {
if (Cs < 1623) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
else if (Cs >= 1623) {
return 4;
}
}
else if (mu_mv_mag_sq >= 5) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
else if (MVDiff >= 20) {
if (SC < 4266) {
if (AFD < 17) {
if (Rs < 1367) {
return 2;
}
else if (Rs >= 1367) {
return 8;
}
}
else if (AFD >= 17) {
if (Cs < 1535) {
if (TSC < 12) {
if (negBalance < 45) {
if (Rs < 1748) {
return 8;
}
else if (Rs >= 1748) {
return 1;
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (TSC >= 12) {
if (Rs < 2636) {
return 1;
}
else if (Rs >= 2636) {
return 8;
}
}
}
else if (Cs >= 1535) {
if (mcTcor < 74) {
if (Rs < 1393) {
return 8;
}
else if (Rs >= 1393) {
return 1;
}
}
else if (mcTcor >= 74) {
return 8;
}
}
}
}
else if (SC >= 4266) {
if (ssDCval < 92) {
if (Rs < 2976) {
return 2;
}
else if (Rs >= 2976) {
return 8;
}
}
else if (ssDCval >= 92) {
if (SC < 4286) {
return 2;
}
else if (SC >= 4286) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 19) {
if (TSC < 12) {
if (RsCsDiff < 290) {
if (posBalance < 50) {
if (ssDCval < 95) {
if (Cs < 1325) {
if (Rs < 1983) {
return 4;
}
else if (Rs >= 1983) {
return 2;
}
}
else if (Cs >= 1325) {
if (RsCsDiff < 215) {
return 2;
}
else if (RsCsDiff >= 215) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
return 8;
}
}
}
}
else if (ssDCval >= 95) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (refDCval < 113) {
if (SC < 2773) {
if (posBalance < 70) {
if (refDCval < 80) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (CsDiff < 10) {
if (Rs < 814) {
return 8;
}
else if (Rs >= 814) {
return 4;
}
}
else if (CsDiff >= 10) {
return 2;
}
}
}
else if (refDCval >= 80) {
if (Rs < 1338) {
if (mu_mv_mag_sq < 27) {
return 2;
}
else if (mu_mv_mag_sq >= 27) {
if (Rs < 879) {
return 2;
}
else if (Rs >= 879) {
return 8;
}
}
}
else if (Rs >= 1338) {
return 8;
}
}
}
else if (posBalance >= 70) {
return 1;
}
}
else if (SC >= 2773) {
if (mcTcor < 94) {
if (ssDCval < 82) {
if (tcor < 48) {
if (Rs < 2009) {
return 2;
}
else if (Rs >= 2009) {
return 1;
}
}
else if (tcor >= 48) {
if (RsCsDiff < 201) {
if (refDCval < 72) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (refDCval >= 72) {
if (Cs < 1835) {
if (negBalance < 55) {
return 4;
}
else if (negBalance >= 55) {
if (RsCsDiff < 163) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 4;
}
}
else if (RsCsDiff >= 163) {
return 4;
}
}
}
else if (Cs >= 1835) {
if (Rs < 1852) {
return 8;
}
else if (Rs >= 1852) {
return 4;
}
}
}
}
else if (RsCsDiff >= 201) {
if (Rs < 2331) {
if (CsDiff < 12) {
if (Cs < 1469) {
return 1;
}
else if (Cs >= 1469) {
return 8;
}
}
else if (CsDiff >= 12) {
if (Rs < 1352) {
return 2;
}
else if (Rs >= 1352) {
return 4;
}
}
}
else if (Rs >= 2331) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 2;
}
}
}
}
}
else if (ssDCval >= 82) {
if (RsDiff < 14) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
if (diffRsCsdiff < -91) {
return 2;
}
else if (diffRsCsdiff >= -91) {
if (RsCsDiff < 183) {
if (negBalance < 53) {
if (Cs < 1586) {
if (Rs < 1665) {
return 4;
}
else if (Rs >= 1665) {
return 8;
}
}
else if (Cs >= 1586) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
return 8;
}
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (RsCsDiff >= 183) {
if (SC < 3247) {
return 8;
}
else if (SC >= 3247) {
if (Cs < 1629) {
if (Cs < 1427) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
return 8;
}
}
else if (Cs >= 1427) {
return 4;
}
}
else if (Cs >= 1629) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 14) {
if (MVDiff < 77) {
return 4;
}
else if (MVDiff >= 77) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (Cs < 1546) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
if (mcTcor < 96) {
return 4;
}
else if (mcTcor >= 96) {
if (Rs < 1790) {
return 4;
}
else if (Rs >= 1790) {
return 1;
}
}
}
}
else if (Cs >= 1546) {
if (ssDCval < 90) {
return 2;
}
else if (ssDCval >= 90) {
if (Cs < 1629) {
if (Rs < 2033) {
return 4;
}
else if (Rs >= 2033) {
return 2;
}
}
else if (Cs >= 1629) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 113) {
if (mu_mv_mag_sq < 7) {
if (negBalance < 53) {
if (Cs < 1535) {
if (Rs < 2175) {
return 4;
}
else if (Rs >= 2175) {
return 8;
}
}
else if (Cs >= 1535) {
return 2;
}
}
else if (negBalance >= 53) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
if (RsDiff < 9) {
if (Rs < 1702) {
return 2;
}
else if (Rs >= 1702) {
return 1;
}
}
else if (RsDiff >= 9) {
if (ssDCval < 159) {
if (Cs < 1310) {
return 4;
}
else if (Cs >= 1310) {
return 8;
}
}
else if (ssDCval >= 159) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 290) {
if (tcor < 49) {
if (Cs < 1677) {
return 1;
}
else if (Cs >= 1677) {
if (Rs < 1966) {
return 2;
}
else if (Rs >= 1966) {
return 4;
}
}
}
else if (tcor >= 49) {
if (MVDiff < 29) {
if (gchDC < 7) {
if (negBalance < 61) {
return 2;
}
else if (negBalance >= 61) {
return 1;
}
}
else if (gchDC >= 7) {
return 1;
}
}
else if (MVDiff >= 29) {
if (AFD < 24) {
if (refDCval < 82) {
return 2;
}
else if (refDCval >= 82) {
if (Rs < 1151) {
return 2;
}
else if (Rs >= 1151) {
return 8;
}
}
}
else if (AFD >= 24) {
if (Cs < 1339) {
return 1;
}
else if (Cs >= 1339) {
return 2;
}
}
}
}
}
}
else if (TSC >= 12) {
if (RsCsDiff < 248) {
if (ssDCval < 127) {
if (diffRsCsdiff < 20) {
if (refDCval < 95) {
if (RsCsDiff < 133) {
if (TSC < 14) {
if (Cs < 1717) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
return 8;
}
}
else if (Cs >= 1717) {
return 1;
}
}
else if (TSC >= 14) {
return 4;
}
}
else if (RsCsDiff >= 133) {
if (MVDiff < 19) {
if (Rs < 2569) {
return 8;
}
else if (Rs >= 2569) {
return 4;
}
}
else if (MVDiff >= 19) {
if (RsCsDiff < 223) {
if (diffRsCsdiff < 12) {
if (RsCsDiff < 153) {
if (Cs < 1354) {
return 8;
}
else if (Cs >= 1354) {
return 1;
}
}
else if (RsCsDiff >= 153) {
if (diffMVdiffVal < -4) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (diffRsCsdiff >= 12) {
if (SC < 2871) {
return 4;
}
else if (SC >= 2871) {
return 1;
}
}
}
else if (RsCsDiff >= 223) {
if (Rs < 2052) {
return 8;
}
else if (Rs >= 2052) {
return 4;
}
}
}
}
}
else if (refDCval >= 95) {
if (Rs < 1702) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Rs < 716) {
return 2;
}
else if (Rs >= 716) {
return 8;
}
}
}
else if (Rs >= 1702) {
if (diffMVdiffVal < -101) {
return 1;
}
else if (diffMVdiffVal >= -101) {
if (CsDiff < 6) {
if (Rs < 2830) {
return 2;
}
else if (Rs >= 2830) {
return 4;
}
}
else if (CsDiff >= 6) {
if (tcor < 59) {
return 4;
}
else if (tcor >= 59) {
return 8;
}
}
}
}
}
}
else if (diffRsCsdiff >= 20) {
if (ssDCval < 105) {
if (SC < 4528) {
if (RsCsDiff < 134) {
if (AFD < 26) {
return 4;
}
else if (AFD >= 26) {
return 8;
}
}
else if (RsCsDiff >= 134) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (refDCval < 83) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
if (CsDiff < 12) {
return 8;
}
else if (CsDiff >= 12) {
return 1;
}
}
}
else if (refDCval >= 83) {
if (Rs < 1591) {
return 4;
}
else if (Rs >= 1591) {
if (Rs < 2051) {
return 2;
}
else if (Rs >= 2051) {
return 1;
}
}
}
}
}
}
else if (SC >= 4528) {
return 4;
}
}
else if (ssDCval >= 105) {
return 8;
}
}
}
else if (ssDCval >= 127) {
if (ssDCval < 142) {
if (diffAFD < 3) {
return 1;
}
else if (diffAFD >= 3) {
if (Rs < 2601) {
return 8;
}
else if (Rs >= 2601) {
return 2;
}
}
}
else if (ssDCval >= 142) {
if (Rs < 1618) {
return 2;
}
else if (Rs >= 1618) {
return 8;
}
}
}
}
else if (RsCsDiff >= 248) {
if (AFD < 30) {
if (mu_mv_mag_sq < 18) {
if (refDCval < 77) {
if (refDCval < 70) {
return 4;
}
else if (refDCval >= 70) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
}
else if (refDCval >= 77) {
if (diffRsCsdiff < -69) {
return 2;
}
else if (diffRsCsdiff >= -69) {
if (negBalance < 58) {
if (RsDiff < 16) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 4;
}
}
else if (RsDiff >= 16) {
if (Rs < 3076) {
if (Rs < 2617) {
return 2;
}
else if (Rs >= 2617) {
if (Rs < 3002) {
return 4;
}
else if (Rs >= 3002) {
return 1;
}
}
}
else if (Rs >= 3076) {
return 8;
}
}
}
else if (negBalance >= 58) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
if (MVDiff < 34) {
return 2;
}
else if (MVDiff >= 34) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (diffAFD < -1) {
if (RsDiff < 18) {
if (Rs < 1960) {
return 4;
}
else if (Rs >= 1960) {
return 8;
}
}
else if (RsDiff >= 18) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (AFD >= 30) {
if (refDCval < 98) {
if (Rs < 2111) {
if (AFD < 36) {
return 4;
}
else if (AFD >= 36) {
if (Rs < 1305) {
return 8;
}
else if (Rs >= 1305) {
return 1;
}
}
}
else if (Rs >= 2111) {
if (CsDiff < 16) {
return 1;
}
else if (CsDiff >= 16) {
return 8;
}
}
}
else if (refDCval >= 98) {
if (refDCval < 112) {
if (CsDiff < 13) {
return 2;
}
else if (CsDiff >= 13) {
return 8;
}
}
else if (refDCval >= 112) {
if (Rs < 2148) {
if (AFD < 35) {
return 8;
}
else if (AFD >= 35) {
if (Rs < 1898) {
return 2;
}
else if (Rs >= 1898) {
return 4;
}
}
}
else if (Rs >= 2148) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 3318) {
if (tcor < 88) {
if (RsCsDiff < 170) {
if (refDCval < 87) {
if (posBalance < 48) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 4;
}
}
}
else if (posBalance >= 48) {
if (Cs < 1655) {
if (Rs < 4080) {
return 2;
}
else if (Rs >= 4080) {
return 8;
}
}
else if (Cs >= 1655) {
return 8;
}
}
}
else if (refDCval >= 87) {
if (ssDCval < 138) {
if (tcor < 73) {
if (Cs < 1655) {
return 1;
}
else if (Cs >= 1655) {
return 8;
}
}
else if (tcor >= 73) {
if (SC < 6797) {
return 8;
}
else if (SC >= 6797) {
return 1;
}
}
}
else if (ssDCval >= 138) {
return 4;
}
}
}
else if (RsCsDiff >= 170) {
if (MVDiff < 17) {
if (tcor < 81) {
if (RsDiff < 17) {
if (Rs < 3618) {
if (Cs < 1574) {
return 2;
}
else if (Cs >= 1574) {
if (Rs < 3462) {
return 4;
}
else if (Rs >= 3462) {
return 1;
}
}
}
else if (Rs >= 3618) {
if (mu_mv_mag_sq < 36) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 4;
}
}
else if (mu_mv_mag_sq >= 36) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
}
}
else if (RsDiff >= 17) {
if (SC < 6548) {
if (ssDCval < 66) {
if (Rs < 3417) {
return 2;
}
else if (Rs >= 3417) {
return 8;
}
}
else if (ssDCval >= 66) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
else if (SC >= 6548) {
if (Rs < 5198) {
return 4;
}
else if (Rs >= 5198) {
if (diffRsCsdiff < 128) {
return 2;
}
else if (diffRsCsdiff >= 128) {
return 4;
}
}
}
}
}
else if (tcor >= 81) {
if (refDCval < 60) {
if (Rs < 4160) {
return 1;
}
else if (Rs >= 4160) {
return 4;
}
}
else if (refDCval >= 60) {
if (RsDiff < 15) {
if (ssDCval < 97) {
if (diffRsCsdiff < 50) {
return 2;
}
else if (diffRsCsdiff >= 50) {
if (Cs < 1553) {
return 2;
}
else if (Cs >= 1553) {
return 4;
}
}
}
else if (ssDCval >= 97) {
if (Cs < 1369) {
return 4;
}
else if (Cs >= 1369) {
return 8;
}
}
}
else if (RsDiff >= 15) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
return 2;
}
}
}
}
}
else if (MVDiff >= 17) {
if (TSC < 14) {
if (mu_mv_mag_sq < 11) {
if (ssDCval < 100) {
return 2;
}
else if (ssDCval >= 100) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (negBalance < 58) {
if (gchDC < 2) {
if (negBalance < 43) {
if (Rs < 7303) {
return 2;
}
else if (Rs >= 7303) {
return 8;
}
}
else if (negBalance >= 43) {
if (diffTSC < -2) {
if (Rs < 6384) {
return 2;
}
else if (Rs >= 6384) {
return 8;
}
}
else if (diffTSC >= -2) {
if (SC < 4852) {
return 2;
}
else if (SC >= 4852) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (MVDiff < 25) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 1;
}
}
else if (MVDiff >= 25) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (Cs < 1710) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 2;
}
}
else if (Cs >= 1710) {
return 8;
}
}
}
}
}
else if (negBalance >= 58) {
if (MVDiff < 27) {
return 1;
}
else if (MVDiff >= 27) {
if (Rs < 4556) {
if (Rs < 3768) {
return 2;
}
else if (Rs >= 3768) {
return 1;
}
}
else if (Rs >= 4556) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (RsDiff < 17) {
if (posBalance < 49) {
if (Rs < 4419) {
return 1;
}
else if (Rs >= 4419) {
return 4;
}
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 17) {
if (Rs < 3460) {
return 4;
}
else if (Rs >= 3460) {
return 1;
}
}
else if (mu_mv_mag_sq >= 17) {
if (ssDCval < 139) {
return 8;
}
else if (ssDCval >= 139) {
return 2;
}
}
}
}
else if (RsDiff >= 17) {
if (SC < 5443) {
if (MVDiff < 41) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
return 2;
}
}
else if (MVDiff >= 41) {
if (Rs < 3910) {
return 4;
}
else if (Rs >= 3910) {
return 1;
}
}
}
else if (SC >= 5443) {
if (mu_mv_mag_sq < 18) {
return 1;
}
else if (mu_mv_mag_sq >= 18) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
}
}
}
}
}
else if (TSC >= 14) {
if (gchDC < 5) {
if (MVDiff < 67) {
if (AFD < 28) {
if (AFD < 26) {
if (diffMVdiffVal < -3) {
if (Rs < 3794) {
return 8;
}
else if (Rs >= 3794) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
else if (AFD >= 26) {
return 4;
}
}
else if (AFD >= 28) {
if (TSC < 18) {
return 2;
}
else if (TSC >= 18) {
return 1;
}
}
}
else if (MVDiff >= 67) {
if (CsDiff < 10) {
if (Cs < 1332) {
return 1;
}
else if (Cs >= 1332) {
return 4;
}
}
else if (CsDiff >= 10) {
return 1;
}
}
}
else if (gchDC >= 5) {
if (refDCval < 103) {
return 2;
}
else if (refDCval >= 103) {
return 1;
}
}
}
}
}
}
else if (tcor >= 88) {
if (RsDiff < 9) {
if (AFD < 8) {
if (Rs < 3378) {
return 8;
}
else if (Rs >= 3378) {
return 2;
}
}
else if (AFD >= 8) {
if (Rs < 3626) {
if (Rs < 3513) {
return 8;
}
else if (Rs >= 3513) {
return 2;
}
}
else if (Rs >= 3626) {
if (Cs < 1420) {
return 4;
}
else if (Cs >= 1420) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (refDCval < 100) {
if (diffMVdiffVal < -11) {
if (Rs < 4444) {
return 8;
}
else if (Rs >= 4444) {
return 2;
}
}
else if (diffMVdiffVal >= -11) {
if (gchDC < 6) {
if (ssDCval < 61) {
if (Rs < 7642) {
return 4;
}
else if (Rs >= 7642) {
return 2;
}
}
else if (ssDCval >= 61) {
return 2;
}
}
else if (gchDC >= 6) {
if (Rs < 3673) {
return 8;
}
else if (Rs >= 3673) {
return 1;
}
}
}
}
else if (refDCval >= 100) {
if (MVDiff < 23) {
if (RsDiff < 11) {
if (MVDiff < 8) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (Rs < 3455) {
return 4;
}
else if (Rs >= 3455) {
return 1;
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (RsDiff >= 11) {
if (negBalance < 49) {
if (Rs < 3636) {
return 8;
}
else if (Rs >= 3636) {
return 2;
}
}
else if (negBalance >= 49) {
if (Rs < 4442) {
return 2;
}
else if (Rs >= 4442) {
if (SC < 6739) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
return 2;
}
}
else if (SC >= 6739) {
return 2;
}
}
}
}
}
else if (MVDiff >= 23) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 1925) {
if (AFD < 12) {
if (ssDCval < 100) {
if (Rs < 4006) {
if (ssDCval < 72) {
if (TSC < 8) {
if (MVDiff < 6) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
if (Rs < 2932) {
return 4;
}
else if (Rs >= 2932) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (Rs < 2049) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
return 8;
}
}
else if (Rs >= 2049) {
if (diffAFD < -3) {
if (Rs < 3551) {
return 1;
}
else if (Rs >= 3551) {
return 8;
}
}
else if (diffAFD >= -3) {
return 4;
}
}
}
}
else if (TSC >= 8) {
if (Cs < 2744) {
if (posBalance < 52) {
if (negBalance < 45) {
if (MVDiff < 5) {
if (Rs < 3490) {
return 1;
}
else if (Rs >= 3490) {
return 8;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (negBalance >= 45) {
if (Rs < 3276) {
return 1;
}
else if (Rs >= 3276) {
return 8;
}
}
}
else if (posBalance >= 52) {
if (Cs < 2238) {
if (Cs < 1941) {
return 8;
}
else if (Cs >= 1941) {
return 4;
}
}
else if (Cs >= 2238) {
return 8;
}
}
}
else if (Cs >= 2744) {
return 4;
}
}
}
else if (ssDCval >= 72) {
if (SC < 5448) {
if (mu_mv_mag_sq < 88) {
if (posBalance < 59) {
if (mu_mv_mag_sq < 87) {
return 8;
}
else if (mu_mv_mag_sq >= 87) {
if (Rs < 2776) {
return 8;
}
else if (Rs >= 2776) {
return 4;
}
}
}
else if (posBalance >= 59) {
if (Rs < 2489) {
return 8;
}
else if (Rs >= 2489) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 88) {
return 8;
}
}
else if (SC >= 5448) {
if (ssDCval < 81) {
if (mu_mv_mag_sq < 78) {
return 4;
}
else if (mu_mv_mag_sq >= 78) {
if (negBalance < 43) {
if (Cs < 2716) {
return 4;
}
else if (Cs >= 2716) {
return 8;
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
else if (ssDCval >= 81) {
if (Cs < 2363) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 2;
}
}
else if (Cs >= 2363) {
if (Cs < 3958) {
if (SC < 5651) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (SC >= 5651) {
return 4;
}
}
else if (Cs >= 3958) {
return 8;
}
}
}
}
}
}
else if (Rs >= 4006) {
if (MVDiff < 9) {
if (refDCval < 73) {
if (Cs < 3410) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
if (Rs < 4122) {
return 1;
}
else if (Rs >= 4122) {
return 8;
}
}
}
else if (Cs >= 3410) {
if (Rs < 5673) {
return 8;
}
else if (Rs >= 5673) {
return 2;
}
}
}
else if (refDCval >= 73) {
if (MVDiff < 1) {
if (Rs < 4366) {
return 4;
}
else if (Rs >= 4366) {
return 2;
}
}
else if (MVDiff >= 1) {
if (negBalance < 43) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 2;
}
}
}
}
else if (negBalance >= 43) {
if (Cs < 2542) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (Cs >= 2542) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 9) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (Rs < 4662) {
if (Rs < 4103) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (Rs >= 4103) {
return 8;
}
}
else if (Rs >= 4662) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
if (CsDiff < 11) {
return 2;
}
else if (CsDiff >= 11) {
if (Rs < 4853) {
return 8;
}
else if (Rs >= 4853) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 100) {
if (negBalance < 46) {
if (posBalance < 43) {
if (refDCval < 118) {
if (Rs < 5239) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 3454) {
return 2;
}
else if (Rs >= 3454) {
return 8;
}
}
}
}
else if (Rs >= 5239) {
if (Cs < 2494) {
if (Rs < 5380) {
return 1;
}
else if (Rs >= 5380) {
return 2;
}
}
else if (Cs >= 2494) {
return 1;
}
}
}
else if (refDCval >= 118) {
if (ssDCval < 131) {
if (Cs < 3929) {
if (Rs < 2103) {
return 8;
}
else if (Rs >= 2103) {
return 4;
}
}
else if (Cs >= 3929) {
return 1;
}
}
else if (ssDCval >= 131) {
if (gchDC < 7) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (gchDC >= 7) {
return 1;
}
}
}
}
else if (posBalance >= 43) {
if (Rs < 6363) {
if (mcTcor < 93) {
if (negBalance < 44) {
if (diffMVdiffVal < -5) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Cs < 2430) {
return 8;
}
else if (Cs >= 2430) {
if (Rs < 4039) {
return 2;
}
else if (Rs >= 4039) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -5) {
if (mu_mv_mag_sq < 38) {
if (Rs < 2482) {
return 4;
}
else if (Rs >= 2482) {
return 2;
}
}
else if (mu_mv_mag_sq >= 38) {
return 8;
}
}
}
else if (negBalance >= 44) {
if (mu_mv_mag_sq < 94) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 94) {
if (Rs < 3899) {
return 4;
}
else if (Rs >= 3899) {
return 1;
}
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 108) {
if (Cs < 2875) {
if (Rs < 4388) {
return 8;
}
else if (Rs >= 4388) {
return 2;
}
}
else if (Cs >= 2875) {
return 4;
}
}
else if (ssDCval >= 108) {
return 8;
}
}
}
else if (Rs >= 6363) {
if (Rs < 8605) {
if (Cs < 3986) {
if (Rs < 7150) {
return 4;
}
else if (Rs >= 7150) {
return 2;
}
}
else if (Cs >= 3986) {
return 8;
}
}
else if (Rs >= 8605) {
return 1;
}
}
}
}
else if (negBalance >= 46) {
if (Cs < 3264) {
if (mu_mv_mag_sq < 74) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (mu_mv_mag_sq >= 74) {
if (MVDiff < 8) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
if (Rs < 3529) {
return 8;
}
else if (Rs >= 3529) {
return 2;
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
else if (Cs >= 3264) {
if (tcor < 92) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
return 1;
}
}
else if (tcor >= 92) {
return 8;
}
}
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 24) {
if (ssDCval < 152) {
if (AFD < 22) {
if (ssDCval < 68) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (CsDiff < 11) {
if (Cs < 2402) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 4;
}
}
else if (Cs >= 2402) {
return 8;
}
}
else if (CsDiff >= 11) {
if (RsCsDiff < 255) {
if (MVDiff < 19) {
if (Cs < 2339) {
if (Rs < 1617) {
return 2;
}
else if (Rs >= 1617) {
return 1;
}
}
else if (Cs >= 2339) {
return 4;
}
}
else if (MVDiff >= 19) {
if (Cs < 3730) {
return 2;
}
else if (Cs >= 3730) {
return 1;
}
}
}
else if (RsCsDiff >= 255) {
if (negBalance < 51) {
if (MVDiff < 29) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 9) {
return 2;
}
else if (mu_mv_mag_sq >= 9) {
return 4;
}
}
else if (posBalance >= 57) {
if (Cs < 3177) {
return 4;
}
else if (Cs >= 3177) {
return 2;
}
}
}
else if (MVDiff >= 29) {
return 4;
}
}
else if (negBalance >= 51) {
return 4;
}
}
}
}
}
else if (ssDCval >= 68) {
if (mcTcor < 94) {
if (ssDCval < 101) {
if (Rs < 4682) {
if (ssDCval < 79) {
if (CsDiff < 11) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (RsDiff < 13) {
if (posBalance < 54) {
if (Rs < 2539) {
return 8;
}
else if (Rs >= 2539) {
return 4;
}
}
else if (posBalance >= 54) {
return 8;
}
}
else if (RsDiff >= 13) {
if (Rs < 3270) {
return 4;
}
else if (Rs >= 3270) {
return 2;
}
}
}
}
else if (CsDiff >= 11) {
if (posBalance < 61) {
if (TSC < 11) {
if (mu_mv_mag_sq < 2) {
if (Rs < 3036) {
return 8;
}
else if (Rs >= 3036) {
if (Rs < 3978) {
return 4;
}
else if (Rs >= 3978) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 3823) {
return 4;
}
else if (Rs >= 3823) {
return 2;
}
}
}
else if (TSC >= 11) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
if (posBalance < 58) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (posBalance >= 58) {
if (Cs < 2053) {
return 2;
}
else if (Cs >= 2053) {
if (diffAFD < 2) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (Rs < 3664) {
return 1;
}
else if (Rs >= 3664) {
return 8;
}
}
}
else if (diffAFD >= 2) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 61) {
if (mcTcor < 84) {
if (posBalance < 62) {
if (TSC < 16) {
return 8;
}
else if (TSC >= 16) {
return 4;
}
}
else if (posBalance >= 62) {
return 4;
}
}
else if (mcTcor >= 84) {
if (RsCsDiff < 486) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
return 1;
}
}
else if (RsCsDiff >= 486) {
if (Rs < 3283) {
return 8;
}
else if (Rs >= 3283) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 79) {
if (AFD < 17) {
if (Rs < 2661) {
if (Rs < 1232) {
return 2;
}
else if (Rs >= 1232) {
if (tcor < 80) {
return 8;
}
else if (tcor >= 80) {
if (SC < 5042) {
if (tcor < 81) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
return 8;
}
}
else if (tcor >= 81) {
return 8;
}
}
else if (SC >= 5042) {
return 4;
}
}
}
}
else if (Rs >= 2661) {
if (MVDiff < 11) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
if (Cs < 3294) {
if (negBalance < 47) {
if (RsDiff < 11) {
if (Rs < 3955) {
return 4;
}
else if (Rs >= 3955) {
return 8;
}
}
else if (RsDiff >= 11) {
if (Rs < 4355) {
return 2;
}
else if (Rs >= 4355) {
return 1;
}
}
}
else if (negBalance >= 47) {
if (AFD < 16) {
if (CsDiff < 10) {
if (SC < 5342) {
return 4;
}
else if (SC >= 5342) {
if (Rs < 3790) {
return 2;
}
else if (Rs >= 3790) {
return 8;
}
}
}
else if (CsDiff >= 10) {
return 4;
}
}
else if (AFD >= 16) {
return 8;
}
}
}
else if (Cs >= 3294) {
if (Rs < 3605) {
return 8;
}
else if (Rs >= 3605) {
return 2;
}
}
}
}
else if (MVDiff >= 11) {
if (SC < 4918) {
return 2;
}
else if (SC >= 4918) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (Rs < 2989) {
return 8;
}
else if (Rs >= 2989) {
return 4;
}
}
}
}
}
}
else if (AFD >= 17) {
if (mcTcor < 80) {
if (Rs < 3469) {
return 4;
}
else if (Rs >= 3469) {
return 8;
}
}
else if (mcTcor >= 80) {
if (CsDiff < 12) {
if (Rs < 3290) {
if (mcTcor < 89) {
if (refDCval < 80) {
return 8;
}
else if (refDCval >= 80) {
if (tcor < 67) {
return 4;
}
else if (tcor >= 67) {
if (MVDiff < 6) {
if (Cs < 3282) {
return 8;
}
else if (Cs >= 3282) {
return 4;
}
}
else if (MVDiff >= 6) {
if (diffMVdiffVal < -7) {
return 8;
}
else if (diffMVdiffVal >= -7) {
return 4;
}
}
}
}
}
else if (mcTcor >= 89) {
if (negBalance < 47) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 4;
}
}
else if (negBalance >= 47) {
if (Rs < 2774) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
if (Rs < 1674) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
else if (Rs >= 1674) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
return 1;
}
}
}
}
else if (Rs >= 2774) {
if (Cs < 2219) {
return 4;
}
else if (Cs >= 2219) {
if (refDCval < 97) {
return 8;
}
else if (refDCval >= 97) {
return 4;
}
}
}
}
}
}
else if (Rs >= 3290) {
if (negBalance < 48) {
if (Rs < 4505) {
return 2;
}
else if (Rs >= 4505) {
return 4;
}
}
else if (negBalance >= 48) {
if (Rs < 4591) {
if (AFD < 18) {
if (Cs < 2959) {
return 8;
}
else if (Cs >= 2959) {
return 4;
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (Rs >= 4591) {
return 8;
}
}
}
}
else if (CsDiff >= 12) {
if (diffRsCsdiff < 15) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (SC < 7070) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
else if (SC >= 7070) {
return 4;
}
}
}
else if (diffRsCsdiff >= 15) {
if (tcor < 70) {
if (Rs < 1190) {
return 2;
}
else if (Rs >= 1190) {
return 8;
}
}
else if (tcor >= 70) {
if (ssDCval < 91) {
if (Rs < 2354) {
return 4;
}
else if (Rs >= 2354) {
return 8;
}
}
else if (ssDCval >= 91) {
if (Cs < 2519) {
return 8;
}
else if (Cs >= 2519) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Rs >= 4682) {
if (RsDiff < 16) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (Cs < 2300) {
return 2;
}
else if (Cs >= 2300) {
return 1;
}
}
}
else if (RsDiff >= 16) {
if (CsDiff < 13) {
return 2;
}
else if (CsDiff >= 13) {
if (mcTcor < 87) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
return 2;
}
}
else if (mcTcor >= 87) {
return 8;
}
}
}
}
}
else if (ssDCval >= 101) {
if (negBalance < 31) {
if (Rs < 2429) {
return 4;
}
else if (Rs >= 2429) {
return 2;
}
}
else if (negBalance >= 31) {
if (ssDCval < 105) {
if (Rs < 3781) {
if (Rs < 2924) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 4;
}
}
else if (Rs >= 2924) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 4;
}
}
}
else if (Rs >= 3781) {
return 8;
}
}
else if (ssDCval >= 105) {
if (Rs < 2718) {
if (ssDCval < 129) {
return 8;
}
else if (ssDCval >= 129) {
if (posBalance < 55) {
if (diffRsCsdiff < -21) {
if (Cs < 3869) {
return 8;
}
else if (Cs >= 3869) {
return 1;
}
}
else if (diffRsCsdiff >= -21) {
return 1;
}
}
else if (posBalance >= 55) {
if (TSC < 14) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 8;
}
}
else if (TSC >= 14) {
return 2;
}
}
}
}
else if (Rs >= 2718) {
if (tcor < 69) {
if (CsDiff < 12) {
if (Rs < 4077) {
return 1;
}
else if (Rs >= 4077) {
return 8;
}
}
else if (CsDiff >= 12) {
return 8;
}
}
else if (tcor >= 69) {
if (negBalance < 48) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (negBalance >= 48) {
if (ssDCval < 110) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
if (TSC < 13) {
if (Cs < 4318) {
return 4;
}
else if (Cs >= 4318) {
return 8;
}
}
else if (TSC >= 13) {
return 1;
}
}
}
else if (ssDCval >= 110) {
if (RsDiff < 13) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 3308) {
if (Rs < 2957) {
return 2;
}
else if (Rs >= 2957) {
return 1;
}
}
else if (Rs >= 3308) {
return 8;
}
}
}
else if (RsDiff >= 13) {
if (negBalance < 53) {
if (Rs < 3635) {
return 8;
}
else if (Rs >= 3635) {
return 4;
}
}
else if (negBalance >= 53) {
if (Cs < 2045) {
return 2;
}
else if (Cs >= 2045) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (negBalance < 43) {
if (MVDiff < 6) {
if (RsDiff < 12) {
return 4;
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (MVDiff >= 6) {
if (posBalance < 64) {
if (Cs < 2045) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (Cs >= 2045) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (MVDiff < 28) {
return 1;
}
else if (MVDiff >= 28) {
return 2;
}
}
}
}
else if (posBalance >= 64) {
if (Cs < 2761) {
return 8;
}
else if (Cs >= 2761) {
return 4;
}
}
}
}
else if (negBalance >= 43) {
if (SC < 5852) {
if (MVDiff < 14) {
if (Rs < 1518) {
return 1;
}
else if (Rs >= 1518) {
if (gchDC < 2) {
if (MVDiff < 8) {
if (MVDiff < 5) {
if (Rs < 2605) {
return 2;
}
else if (Rs >= 2605) {
return 8;
}
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (MVDiff >= 8) {
if (Rs < 2725) {
return 8;
}
else if (Rs >= 2725) {
if (Rs < 2825) {
return 2;
}
else if (Rs >= 2825) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (MVDiff >= 14) {
if (diffRsCsdiff < -18) {
if (ssDCval < 104) {
return 4;
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (diffRsCsdiff >= -18) {
return 8;
}
}
}
else if (SC >= 5852) {
if (refDCval < 85) {
return 1;
}
else if (refDCval >= 85) {
if (MVDiff < 6) {
if (RsCsDiff < 157) {
if (Rs < 3803) {
if (Rs < 3222) {
return 1;
}
else if (Rs >= 3222) {
return 4;
}
}
else if (Rs >= 3803) {
return 8;
}
}
else if (RsCsDiff >= 157) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
if (AFD < 14) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (AFD >= 14) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 4;
}
}
}
}
}
else if (MVDiff >= 6) {
if (RsCsDiff < 355) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
return 8;
}
}
else if (RsCsDiff >= 355) {
if (tcor < 86) {
return 1;
}
else if (tcor >= 86) {
return 2;
}
}
}
}
}
}
}
}
}
else if (AFD >= 22) {
if (SC < 10634) {
if (tcor < 72) {
if (MVDiff < 6) {
if (posBalance < 61) {
if (diffAFD < 4) {
return 4;
}
else if (diffAFD >= 4) {
return 2;
}
}
else if (posBalance >= 61) {
if (Cs < 3001) {
return 4;
}
else if (Cs >= 3001) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (Cs < 3053) {
if (refDCval < 107) {
if (Rs < 1532) {
if (Cs < 2189) {
return 4;
}
else if (Cs >= 2189) {
if (MVDiff < 24) {
if (Cs < 2752) {
return 8;
}
else if (Cs >= 2752) {
return 4;
}
}
else if (MVDiff >= 24) {
if (CsDiff < 19) {
return 2;
}
else if (CsDiff >= 19) {
return 8;
}
}
}
}
else if (Rs >= 1532) {
if (SC < 5740) {
if (refDCval < 89) {
if (RsDiff < 15) {
if (Rs < 2846) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
if (MVDiff < 37) {
return 1;
}
else if (MVDiff >= 37) {
if (Rs < 2367) {
return 4;
}
else if (Rs >= 2367) {
return 8;
}
}
}
}
else if (Rs >= 2846) {
return 8;
}
}
else if (RsDiff >= 15) {
if (refDCval < 53) {
return 2;
}
else if (refDCval >= 53) {
return 4;
}
}
}
else if (refDCval >= 89) {
if (Rs < 3103) {
return 8;
}
else if (Rs >= 3103) {
return 4;
}
}
}
else if (SC >= 5740) {
if (RsDiff < 22) {
if (AFD < 25) {
if (ssDCval < 93) {
return 8;
}
else if (ssDCval >= 93) {
return 4;
}
}
else if (AFD >= 25) {
if (Cs < 2937) {
return 1;
}
else if (Cs >= 2937) {
return 8;
}
}
}
else if (RsDiff >= 22) {
if (Cs < 2593) {
return 2;
}
else if (Cs >= 2593) {
if (diffMVdiffVal < 8) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 1;
}
}
else if (diffMVdiffVal >= 8) {
if (Cs < 2675) {
return 4;
}
else if (Cs >= 2675) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 107) {
if (MVDiff < 14) {
if (mcTcor < 82) {
if (Rs < 2622) {
return 2;
}
else if (Rs >= 2622) {
return 1;
}
}
else if (mcTcor >= 82) {
return 2;
}
}
else if (MVDiff >= 14) {
if (TSC < 13) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (TSC >= 13) {
if (refDCval < 133) {
return 8;
}
else if (refDCval >= 133) {
return 2;
}
}
}
}
}
else if (Cs >= 3053) {
if (refDCval < 78) {
if (ssDCval < 73) {
if (Cs < 4654) {
return 2;
}
else if (Cs >= 4654) {
return 1;
}
}
else if (ssDCval >= 73) {
if (RsCsDiff < 783) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
else if (RsCsDiff >= 783) {
return 8;
}
}
}
else if (refDCval >= 78) {
if (CsDiff < 15) {
if (Rs < 5858) {
if (Rs < 1671) {
return 4;
}
else if (Rs >= 1671) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (Rs < 4739) {
return 4;
}
else if (Rs >= 4739) {
return 8;
}
}
}
}
else if (Rs >= 5858) {
if (AFD < 29) {
return 4;
}
else if (AFD >= 29) {
return 1;
}
}
}
else if (CsDiff >= 15) {
if (SC < 8460) {
if (AFD < 31) {
if (Rs < 1283) {
return 8;
}
else if (Rs >= 1283) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
return 8;
}
}
}
else if (AFD >= 31) {
if (Cs < 3882) {
if (Cs < 3075) {
return 8;
}
else if (Cs >= 3075) {
return 2;
}
}
else if (Cs >= 3882) {
return 8;
}
}
}
else if (SC >= 8460) {
if (negBalance < 58) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 2;
}
}
else if (negBalance >= 58) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 72) {
if (negBalance < 55) {
if (mcTcor < 95) {
if (Rs < 2095) {
if (Rs < 1581) {
if (Rs < 1383) {
return 8;
}
else if (Rs >= 1383) {
return 1;
}
}
else if (Rs >= 1581) {
return 4;
}
}
else if (Rs >= 2095) {
if (TSC < 16) {
if (RsCsDiff < 373) {
if (Cs < 2180) {
return 4;
}
else if (Cs >= 2180) {
return 8;
}
}
else if (RsCsDiff >= 373) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 2;
}
}
}
}
else if (TSC >= 16) {
if (RsDiff < 18) {
return 1;
}
else if (RsDiff >= 18) {
if (Rs < 5005) {
return 8;
}
else if (Rs >= 5005) {
return 4;
}
}
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
if (Rs < 2857) {
return 8;
}
else if (Rs >= 2857) {
return 2;
}
}
}
}
}
else if (negBalance >= 55) {
if (RsDiff < 16) {
if (tcor < 79) {
if (RsCsDiff < 173) {
if (Rs < 1735) {
return 2;
}
else if (Rs >= 1735) {
return 8;
}
}
else if (RsCsDiff >= 173) {
if (Cs < 2029) {
if (Rs < 2789) {
return 2;
}
else if (Rs >= 2789) {
return 8;
}
}
else if (Cs >= 2029) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
if (Rs < 3059) {
return 2;
}
else if (Rs >= 3059) {
return 8;
}
}
}
}
}
else if (tcor >= 79) {
if (SC < 4437) {
if (Cs < 2034) {
return 1;
}
else if (Cs >= 2034) {
return 2;
}
}
else if (SC >= 4437) {
return 8;
}
}
}
else if (RsDiff >= 16) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
}
else if (SC >= 10634) {
if (TSC < 10) {
if (RsDiff < 23) {
return 8;
}
else if (RsDiff >= 23) {
return 1;
}
}
else if (TSC >= 10) {
if (AFD < 41) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (AFD >= 41) {
return 8;
}
}
}
}
}
else if (ssDCval >= 152) {
if (ssDCval < 168) {
if (Rs < 3385) {
if (Rs < 2136) {
if (Rs < 1253) {
return 1;
}
else if (Rs >= 1253) {
return 2;
}
}
else if (Rs >= 2136) {
return 8;
}
}
else if (Rs >= 3385) {
if (diffTSC < 4) {
if (MVDiff < 37) {
return 4;
}
else if (MVDiff >= 37) {
if (diffRsCsdiff < -44) {
return 8;
}
else if (diffRsCsdiff >= -44) {
return 4;
}
}
}
else if (diffTSC >= 4) {
return 8;
}
}
}
else if (ssDCval >= 168) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (Cs < 3571) {
if (refDCval < 100) {
if (posBalance < 49) {
if (refDCval < 76) {
if (SC < 6862) {
if (AFD < 16) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (Cs < 2134) {
return 4;
}
else if (Cs >= 2134) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
if (Cs < 2195) {
return 1;
}
else if (Cs >= 2195) {
return 8;
}
}
}
}
}
else if (AFD >= 16) {
if (negBalance < 47) {
if (mu_mv_mag_sq < 57) {
return 2;
}
else if (mu_mv_mag_sq >= 57) {
if (Rs < 3668) {
return 8;
}
else if (Rs >= 3668) {
return 4;
}
}
}
else if (negBalance >= 47) {
if (Rs < 4238) {
return 8;
}
else if (Rs >= 4238) {
return 4;
}
}
}
}
else if (SC >= 6862) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (TSC < 16) {
return 4;
}
else if (TSC >= 16) {
return 2;
}
}
}
}
else if (refDCval >= 76) {
if (negBalance < 50) {
if (TSC < 10) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 8;
}
}
else if (TSC >= 10) {
if (refDCval < 91) {
if (SC < 8171) {
return 2;
}
else if (SC >= 8171) {
if (Rs < 5392) {
return 1;
}
else if (Rs >= 5392) {
return 2;
}
}
}
else if (refDCval >= 91) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
else if (negBalance >= 50) {
if (Rs < 4810) {
return 8;
}
else if (Rs >= 4810) {
return 4;
}
}
}
}
else if (posBalance >= 49) {
if (negBalance < 58) {
if (mu_mv_mag_sq < 54) {
if (MVDiff < 40) {
if (RsCsDiff < 255) {
if (posBalance < 64) {
if (Rs < 1286) {
if (Rs < 900) {
return 2;
}
else if (Rs >= 900) {
if (Rs < 1035) {
return 8;
}
else if (Rs >= 1035) {
return 4;
}
}
}
else if (Rs >= 1286) {
if (mcTcor < 77) {
if (Rs < 2312) {
return 2;
}
else if (Rs >= 2312) {
return 1;
}
}
else if (mcTcor >= 77) {
if (diffRsCsdiff < -22) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (mu_mv_mag_sq < 45) {
if (SC < 4087) {
return 1;
}
else if (SC >= 4087) {
return 8;
}
}
else if (mu_mv_mag_sq >= 45) {
if (ssDCval < 86) {
return 1;
}
else if (ssDCval >= 86) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= -22) {
if (diffRsCsdiff < 90) {
return 8;
}
else if (diffRsCsdiff >= 90) {
if (Rs < 3161) {
return 8;
}
else if (Rs >= 3161) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 64) {
if (Rs < 1512) {
return 8;
}
else if (Rs >= 1512) {
return 2;
}
}
}
else if (RsCsDiff >= 255) {
if (Rs < 4795) {
if (SC < 7643) {
if (ssDCval < 68) {
if (Cs < 2290) {
return 2;
}
else if (Cs >= 2290) {
return 4;
}
}
else if (ssDCval >= 68) {
if (refDCval < 95) {
if (MVDiff < 14) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 3010) {
return 8;
}
else if (Rs >= 3010) {
return 2;
}
}
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (refDCval >= 95) {
if (Cs < 2127) {
return 8;
}
else if (Cs >= 2127) {
if (posBalance < 58) {
if (Rs < 3546) {
return 2;
}
else if (Rs >= 3546) {
return 8;
}
}
else if (posBalance >= 58) {
return 4;
}
}
}
}
}
else if (SC >= 7643) {
if (Cs < 3507) {
return 4;
}
else if (Cs >= 3507) {
return 8;
}
}
}
else if (Rs >= 4795) {
if (ssDCval < 92) {
if (RsDiff < 25) {
return 2;
}
else if (RsDiff >= 25) {
return 4;
}
}
else if (ssDCval >= 92) {
if (tcor < 83) {
return 8;
}
else if (tcor >= 83) {
if (Rs < 5223) {
return 1;
}
else if (Rs >= 5223) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 40) {
if (negBalance < 47) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
if (Rs < 3599) {
return 8;
}
else if (Rs >= 3599) {
return 4;
}
}
}
else if (negBalance >= 47) {
if (Rs < 1774) {
if (Rs < 1033) {
return 2;
}
else if (Rs >= 1033) {
return 4;
}
}
else if (Rs >= 1774) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (ssDCval < 91) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (refDCval < 72) {
if (diffTSC < -3) {
if (Rs < 2437) {
return 1;
}
else if (Rs >= 2437) {
return 4;
}
}
else if (diffTSC >= -3) {
if (negBalance < 56) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (Rs < 2946) {
return 2;
}
else if (Rs >= 2946) {
return 8;
}
}
}
}
else if (refDCval >= 72) {
if (SC < 6601) {
if (SC < 4928) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (Rs < 2314) {
return 1;
}
else if (Rs >= 2314) {
return 8;
}
}
}
else if (SC >= 4928) {
if (tcor < 86) {
if (RsDiff < 13) {
if (posBalance < 53) {
if (Rs < 3173) {
return 1;
}
else if (Rs >= 3173) {
return 8;
}
}
else if (posBalance >= 53) {
if (diffRsCsdiff < 46) {
if (mu_mv_mag_sq < 87) {
return 4;
}
else if (mu_mv_mag_sq >= 87) {
if (Rs < 3211) {
return 4;
}
else if (Rs >= 3211) {
return 8;
}
}
}
else if (diffRsCsdiff >= 46) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
}
}
else if (RsDiff >= 13) {
if (Cs < 2061) {
return 1;
}
else if (Cs >= 2061) {
return 8;
}
}
}
else if (tcor >= 86) {
if (Rs < 2969) {
return 8;
}
else if (Rs >= 2969) {
if (diffRsCsdiff < -11) {
if (Rs < 3448) {
if (Rs < 3220) {
return 1;
}
else if (Rs >= 3220) {
return 4;
}
}
else if (Rs >= 3448) {
return 8;
}
}
else if (diffRsCsdiff >= -11) {
return 4;
}
}
}
}
}
else if (SC >= 6601) {
if (posBalance < 57) {
if (RsDiff < 16) {
return 1;
}
else if (RsDiff >= 16) {
if (Rs < 4739) {
if (Rs < 4137) {
return 2;
}
else if (Rs >= 4137) {
return 4;
}
}
else if (Rs >= 4739) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (Rs < 4998) {
return 8;
}
else if (Rs >= 4998) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (Cs < 2570) {
if (negBalance < 55) {
if (RsCsDiff < 183) {
return 8;
}
else if (RsCsDiff >= 183) {
if (Rs < 3287) {
return 2;
}
else if (Rs >= 3287) {
if (Rs < 4771) {
return 8;
}
else if (Rs >= 4771) {
return 2;
}
}
}
}
else if (negBalance >= 55) {
return 4;
}
}
else if (Cs >= 2570) {
if (MVDiff < 16) {
if (SC < 6994) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (SC >= 6994) {
if (Cs < 3397) {
return 8;
}
else if (Cs >= 3397) {
return 4;
}
}
}
else if (MVDiff >= 16) {
if (Cs < 3400) {
return 8;
}
else if (Cs >= 3400) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 58) {
if (posBalance < 56) {
if (tcor < 83) {
if (CsDiff < 13) {
return 8;
}
else if (CsDiff >= 13) {
if (mu_mv_mag_sq < 43) {
return 8;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 3966) {
return 4;
}
else if (Rs >= 3966) {
return 1;
}
}
}
}
else if (tcor >= 83) {
return 2;
}
}
else if (posBalance >= 56) {
if (diffRsCsdiff < 27) {
if (mu_mv_mag_sq < 33) {
if (Rs < 1957) {
return 4;
}
else if (Rs >= 1957) {
return 1;
}
}
else if (mu_mv_mag_sq >= 33) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (Rs < 1710) {
return 4;
}
else if (Rs >= 1710) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 27) {
if (ssDCval < 95) {
if (refDCval < 69) {
return 4;
}
else if (refDCval >= 69) {
if (Rs < 2998) {
return 8;
}
else if (Rs >= 2998) {
return 1;
}
}
}
else if (ssDCval >= 95) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 100) {
if (Rs < 5759) {
if (mcTcor < 87) {
if (RsDiff < 18) {
if (refDCval < 136) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (Cs < 3173) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (RsDiff < 14) {
if (Rs < 2436) {
return 1;
}
else if (Rs >= 2436) {
return 2;
}
}
else if (RsDiff >= 14) {
return 8;
}
}
}
else if (Cs >= 3173) {
if (Rs < 4868) {
return 4;
}
else if (Rs >= 4868) {
return 8;
}
}
}
}
else if (refDCval >= 136) {
if (Rs < 3398) {
if (Rs < 886) {
return 1;
}
else if (Rs >= 886) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Rs < 2556) {
return 8;
}
else if (Rs >= 2556) {
return 2;
}
}
}
}
else if (Rs >= 3398) {
if (negBalance < 45) {
if (Rs < 3694) {
return 1;
}
else if (Rs >= 3694) {
return 8;
}
}
else if (negBalance >= 45) {
return 1;
}
}
}
}
else if (RsDiff >= 18) {
if (AFD < 36) {
if (refDCval < 138) {
if (refDCval < 107) {
return 4;
}
else if (refDCval >= 107) {
if (Cs < 3357) {
return 8;
}
else if (Cs >= 3357) {
return 2;
}
}
}
else if (refDCval >= 138) {
return 4;
}
}
else if (AFD >= 36) {
if (Rs < 4807) {
return 1;
}
else if (Rs >= 4807) {
return 4;
}
}
}
}
else if (mcTcor >= 87) {
if (Cs < 2580) {
if (refDCval < 120) {
if (CsDiff < 9) {
if (mu_mv_mag_sq < 103) {
return 8;
}
else if (mu_mv_mag_sq >= 103) {
return 1;
}
}
else if (CsDiff >= 9) {
if (mu_mv_mag_sq < 72) {
if (mu_mv_mag_sq < 30) {
if (Rs < 1298) {
return 2;
}
else if (Rs >= 1298) {
return 1;
}
}
else if (mu_mv_mag_sq >= 30) {
if (mu_mv_mag_sq < 67) {
if (AFD < 24) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (MVDiff < 23) {
return 2;
}
else if (MVDiff >= 23) {
return 8;
}
}
}
else if (AFD >= 24) {
return 2;
}
}
else if (mu_mv_mag_sq >= 67) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 72) {
return 1;
}
}
}
else if (refDCval >= 120) {
if (posBalance < 54) {
if (negBalance < 53) {
if (RsDiff < 10) {
if (AFD < 16) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (AFD >= 16) {
return 1;
}
}
else if (RsDiff >= 10) {
if (RsDiff < 26) {
if (diffMVdiffVal < -23) {
return 8;
}
else if (diffMVdiffVal >= -23) {
if (Rs < 2380) {
return 8;
}
else if (Rs >= 2380) {
return 4;
}
}
}
else if (RsDiff >= 26) {
if (Rs < 5070) {
return 2;
}
else if (Rs >= 5070) {
return 8;
}
}
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (posBalance >= 54) {
if (Rs < 5336) {
if (diffTSC < 7) {
return 8;
}
else if (diffTSC >= 7) {
return 1;
}
}
else if (Rs >= 5336) {
return 4;
}
}
}
}
else if (Cs >= 2580) {
if (Rs < 4419) {
if (tcor < 87) {
if (MVDiff < 13) {
if (TSC < 11) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (ssDCval < 122) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
else if (ssDCval >= 122) {
return 4;
}
}
}
else if (TSC >= 11) {
if (ssDCval < 136) {
return 4;
}
else if (ssDCval >= 136) {
return 8;
}
}
}
else if (MVDiff >= 13) {
if (mu_mv_mag_sq < 58) {
if (mu_mv_mag_sq < 35) {
return 4;
}
else if (mu_mv_mag_sq >= 35) {
return 8;
}
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 2971) {
return 4;
}
else if (Rs >= 2971) {
return 1;
}
}
}
}
else if (tcor >= 87) {
if (negBalance < 48) {
if (MVDiff < 6) {
if (Cs < 2961) {
return 8;
}
else if (Cs >= 2961) {
return 4;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (negBalance >= 48) {
if (Cs < 2676) {
return 1;
}
else if (Cs >= 2676) {
if (Rs < 3024) {
return 4;
}
else if (Rs >= 3024) {
return 8;
}
}
}
}
}
else if (Rs >= 4419) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
}
}
else if (Rs >= 5759) {
if (posBalance < 50) {
if (Rs < 6978) {
return 8;
}
else if (Rs >= 6978) {
if (Rs < 7328) {
return 2;
}
else if (Rs >= 7328) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (AFD < 14) {
if (Rs < 6418) {
return 8;
}
else if (Rs >= 6418) {
return 4;
}
}
else if (AFD >= 14) {
return 4;
}
}
}
}
}
else if (Cs >= 3571) {
if (negBalance < 47) {
if (negBalance < 36) {
if (mcTcor < 82) {
if (tcor < 68) {
if (Rs < 3960) {
return 1;
}
else if (Rs >= 3960) {
return 8;
}
}
else if (tcor >= 68) {
if (Rs < 6125) {
return 4;
}
else if (Rs >= 6125) {
return 2;
}
}
}
else if (mcTcor >= 82) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (AFD < 15) {
if (Rs < 5902) {
if (ssDCval < 71) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
else if (ssDCval >= 71) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
if (Rs < 4649) {
return 8;
}
else if (Rs >= 4649) {
return 4;
}
}
}
}
else if (Rs >= 5902) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (Rs < 8166) {
return 8;
}
else if (Rs >= 8166) {
return 1;
}
}
}
}
else if (AFD >= 15) {
if (mu_mv_mag_sq < 46) {
if (ssDCval < 128) {
if (SC < 8541) {
return 2;
}
else if (SC >= 8541) {
if (TSC < 15) {
if (diffRsCsdiff < -82) {
if (Rs < 7567) {
return 4;
}
else if (Rs >= 7567) {
return 1;
}
}
else if (diffRsCsdiff >= -82) {
return 8;
}
}
else if (TSC >= 15) {
if (Cs < 3707) {
return 1;
}
else if (Cs >= 3707) {
return 2;
}
}
}
}
else if (ssDCval >= 128) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
if (Rs < 4167) {
return 8;
}
else if (Rs >= 4167) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (ssDCval < 83) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 4;
}
}
else if (ssDCval >= 83) {
if (MVDiff < 8) {
if (mcTcor < 86) {
if (Rs < 5400) {
return 8;
}
else if (Rs >= 5400) {
return 2;
}
}
else if (mcTcor >= 86) {
return 4;
}
}
else if (MVDiff >= 8) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 6343) {
return 4;
}
else if (Rs >= 6343) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 47) {
if (posBalance < 54) {
if (MVDiff < 16) {
if (AFD < 18) {
if (RsCsDiff < 329) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 8;
}
}
else if (RsCsDiff >= 329) {
return 8;
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (MVDiff >= 16) {
if (CsDiff < 14) {
if (mu_mv_mag_sq < 81) {
if (Rs < 5301) {
return 8;
}
else if (Rs >= 5301) {
return 2;
}
}
else if (mu_mv_mag_sq >= 81) {
return 1;
}
}
else if (CsDiff >= 14) {
if (Cs < 4582) {
return 2;
}
else if (Cs >= 4582) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
if (Rs < 7747) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (negBalance < 52) {
if (Rs < 6488) {
if (Cs < 3866) {
if (tcor < 85) {
if (RsCsDiff < 379) {
return 8;
}
else if (RsCsDiff >= 379) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
}
else if (tcor >= 85) {
if (Rs < 4871) {
return 2;
}
else if (Rs >= 4871) {
return 4;
}
}
}
else if (Cs >= 3866) {
return 8;
}
}
else if (Rs >= 6488) {
if (Rs < 7435) {
return 4;
}
else if (Rs >= 7435) {
return 8;
}
}
}
else if (negBalance >= 52) {
if (posBalance < 57) {
if (Rs < 6249) {
if (ssDCval < 113) {
return 8;
}
else if (ssDCval >= 113) {
return 4;
}
}
else if (Rs >= 6249) {
return 4;
}
}
else if (posBalance >= 57) {
if (Rs < 3465) {
if (TSC < 15) {
return 8;
}
else if (TSC >= 15) {
return 4;
}
}
else if (Rs >= 3465) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (Rs < 4025) {
return 2;
}
else if (Rs >= 4025) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 7747) {
return 4;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1652) {
if (Rs < 3093) {
if (Rs < 1455) {
if (refDCval < 122) {
if (TSCindex < 1) {
if (Cs < 1505) {
if (Rs < 756) {
return 4;
}
else if (Rs >= 756) {
return 8;
}
}
else if (Cs >= 1505) {
return 2;
}
}
else if (TSCindex >= 1) {
if (refDCval < 63) {
if (negBalance < 26) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 44) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 1;
}
}
else if (posBalance >= 44) {
if (Rs < 802) {
return 1;
}
else if (Rs >= 802) {
if (SC < 2514) {
return 8;
}
else if (SC >= 2514) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1474) {
if (Cs < 1343) {
return 4;
}
else if (Cs >= 1343) {
return 8;
}
}
else if (Cs >= 1474) {
return 1;
}
}
}
else if (negBalance >= 26) {
if (Rs < 1251) {
if (Rs < 680) {
return 1;
}
else if (Rs >= 680) {
return 8;
}
}
else if (Rs >= 1251) {
if (tcor < 98) {
if (diffRsCsdiff < -8) {
return 4;
}
else if (diffRsCsdiff >= -8) {
return 8;
}
}
else if (tcor >= 98) {
return 4;
}
}
}
}
else if (refDCval >= 63) {
if (ssDCval < 76) {
if (Rs < 1048) {
return 8;
}
else if (Rs >= 1048) {
if (diffRsCsdiff < -4) {
return 4;
}
else if (diffRsCsdiff >= -4) {
if (tcor < 99) {
if (SC < 2695) {
if (Rs < 1149) {
return 4;
}
else if (Rs >= 1149) {
if (Rs < 1274) {
return 8;
}
else if (Rs >= 1274) {
return 1;
}
}
}
else if (SC >= 2695) {
return 8;
}
}
else if (tcor >= 99) {
return 1;
}
}
}
}
else if (ssDCval >= 76) {
if (Rs < 1144) {
if (refDCval < 83) {
return 1;
}
else if (refDCval >= 83) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 4;
}
}
}
else if (Rs >= 1144) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 122) {
if (AFD < 2) {
if (posBalance < 28) {
if (Cs < 1348) {
return 1;
}
else if (Cs >= 1348) {
return 8;
}
}
else if (posBalance >= 28) {
if (Rs < 1383) {
return 4;
}
else if (Rs >= 1383) {
return 8;
}
}
}
else if (AFD >= 2) {
if (mcTcor < 96) {
if (ssDCval < 129) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 933) {
return 8;
}
else if (Rs >= 933) {
return 4;
}
}
}
else if (ssDCval >= 129) {
if (posBalance < 45) {
if (Rs < 889) {
return 2;
}
else if (Rs >= 889) {
return 1;
}
}
else if (posBalance >= 45) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (Cs < 1467) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 1075) {
return 8;
}
else if (Rs >= 1075) {
return 4;
}
}
}
else if (Cs >= 1467) {
if (ssDCval < 129) {
if (Rs < 1274) {
return 4;
}
else if (Rs >= 1274) {
return 8;
}
}
else if (ssDCval >= 129) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1455) {
if (ssDCval < 140) {
if (ssDCval < 86) {
if (tcor < 97) {
if (Rs < 2218) {
if (RsCsDiff < 4) {
if (Rs < 1560) {
return 8;
}
else if (Rs >= 1560) {
return 4;
}
}
else if (RsCsDiff >= 4) {
return 8;
}
}
else if (Rs >= 2218) {
if (Cs < 1425) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (Cs >= 1425) {
if (Rs < 2997) {
return 8;
}
else if (Rs >= 2997) {
return 4;
}
}
}
}
else if (tcor >= 97) {
if (negBalance < 19) {
if (SC < 4462) {
if (Cs < 1344) {
if (Rs < 2273) {
return 8;
}
else if (Rs >= 2273) {
if (Cs < 1342) {
if (SC < 3595) {
return 2;
}
else if (SC >= 3595) {
if (Rs < 2287) {
return 8;
}
else if (Rs >= 2287) {
if (Rs < 2302) {
if (Rs < 2300) {
if (Rs < 2297) {
if (Rs < 2296) {
if (Rs < 2288) {
return 2;
}
else if (Rs >= 2288) {
return 8;
}
}
else if (Rs >= 2296) {
return 2;
}
}
else if (Rs >= 2297) {
return 8;
}
}
else if (Rs >= 2300) {
return 2;
}
}
else if (Rs >= 2302) {
if (mu_mv_mag_sq < 123) {
return 2;
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
}
}
}
else if (Cs >= 1342) {
return 2;
}
}
}
else if (Cs >= 1344) {
if (SC < 3319) {
if (Cs < 1441) {
return 8;
}
else if (Cs >= 1441) {
if (diffRsCsdiff < -2) {
if (Rs < 1756) {
return 8;
}
else if (Rs >= 1756) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
else if (SC >= 3319) {
return 8;
}
}
}
else if (SC >= 4462) {
return 1;
}
}
else if (negBalance >= 19) {
if (posBalance < 39) {
if (mu_mv_mag_sq < 127) {
if (Rs < 2135) {
return 1;
}
else if (Rs >= 2135) {
return 2;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 2234) {
return 4;
}
else if (Rs >= 2234) {
if (Rs < 2802) {
return 8;
}
else if (Rs >= 2802) {
return 4;
}
}
}
}
else if (posBalance >= 39) {
if (mcTcor < 98) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 2896) {
return 4;
}
else if (Rs >= 2896) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 22) {
if (Cs < 1563) {
return 4;
}
else if (Cs >= 1563) {
return 8;
}
}
else if (negBalance >= 22) {
if (Rs < 1514) {
if (Cs < 1471) {
return 4;
}
else if (Cs >= 1471) {
return 8;
}
}
else if (Rs >= 1514) {
if (Cs < 1614) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 8;
}
}
else if (Cs >= 1614) {
if (Rs < 2068) {
return 4;
}
else if (Rs >= 2068) {
if (Rs < 2463) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (Rs >= 2463) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (diffAFD < -20) {
if (Rs < 1750) {
return 1;
}
else if (Rs >= 1750) {
return 4;
}
}
else if (diffAFD >= -20) {
if (posBalance < 46) {
if (Rs < 3053) {
return 8;
}
else if (Rs >= 3053) {
if (Rs < 3057) {
return 2;
}
else if (Rs >= 3057) {
return 8;
}
}
}
else if (posBalance >= 46) {
if (ssDCval < 117) {
if (SC < 4159) {
if (diffAFD < -2) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (diffAFD >= -2) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (SC >= 4159) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (ssDCval >= 117) {
if (ssDCval < 119) {
if (mu_mv_mag_sq < 119) {
return 8;
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 2035) {
return 1;
}
else if (Rs >= 2035) {
return 8;
}
}
}
else if (ssDCval >= 119) {
if (Rs < 2819) {
return 8;
}
else if (Rs >= 2819) {
if (Rs < 2915) {
return 1;
}
else if (Rs >= 2915) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (Rs < 2092) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (Rs >= 2092) {
return 1;
}
}
}
}
else if (Rs >= 3093) {
if (ssDCval < 114) {
if (ssDCval < 84) {
if (RsCsDiff < 12) {
if (SC < 5076) {
if (ssDCval < 61) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 4;
}
}
else if (ssDCval >= 61) {
if (Rs < 3589) {
return 1;
}
else if (Rs >= 3589) {
return 8;
}
}
}
else if (SC >= 5076) {
if (Cs < 1577) {
if (mu_mv_mag_sq < 124) {
if (negBalance < 19) {
return 4;
}
else if (negBalance >= 19) {
return 8;
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (Cs >= 1577) {
return 4;
}
}
}
else if (RsCsDiff >= 12) {
if (Rs < 3748) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 2;
}
}
else if (Rs >= 3748) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
if (ssDCval < 74) {
if (Cs < 1353) {
if (Cs < 1344) {
return 4;
}
else if (Cs >= 1344) {
return 1;
}
}
else if (Cs >= 1353) {
return 4;
}
}
else if (ssDCval >= 74) {
if (Rs < 4305) {
return 2;
}
else if (Rs >= 4305) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (SC < 5719) {
if (ssDCval < 104) {
if (diffRsCsdiff < -17) {
return 2;
}
else if (diffRsCsdiff >= -17) {
return 4;
}
}
else if (ssDCval >= 104) {
if (Rs < 3253) {
return 8;
}
else if (Rs >= 3253) {
return 2;
}
}
}
else if (SC >= 5719) {
if (refDCval < 97) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 2;
}
}
else if (refDCval >= 97) {
return 4;
}
}
}
}
else if (ssDCval >= 114) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (diffTSC < -2) {
if (Rs < 3363) {
return 2;
}
else if (Rs >= 3363) {
return 1;
}
}
else if (diffTSC >= -2) {
return 4;
}
}
}
}
}
else if (Cs >= 1652) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (refDCval < 65) {
if (Rs < 8850) {
if (tcor < 92) {
if (Rs < 3269) {
return 4;
}
else if (Rs >= 3269) {
return 8;
}
}
else if (tcor >= 92) {
return 8;
}
}
else if (Rs >= 8850) {
return 2;
}
}
else if (refDCval >= 65) {
if (diffTSC < -1) {
if (Rs < 2268) {
return 4;
}
else if (Rs >= 2268) {
return 8;
}
}
else if (diffTSC >= -1) {
if (ssDCval < 87) {
return 4;
}
else if (ssDCval >= 87) {
if (Rs < 2595) {
return 4;
}
else if (Rs >= 2595) {
return 8;
}
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 3312) {
if (negBalance < 19) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (ssDCval < 61) {
if (Cs < 1886) {
if (mu_mv_mag_sq < 126) {
if (Rs < 999) {
return 8;
}
else if (Rs >= 999) {
return 4;
}
}
else if (mu_mv_mag_sq >= 126) {
if (CsDiff < 1) {
if (Rs < 4190) {
return 1;
}
else if (Rs >= 4190) {
return 4;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
}
else if (Cs >= 1886) {
return 1;
}
}
else if (ssDCval >= 61) {
if (Cs < 2091) {
if (ssDCval < 66) {
if (Rs < 1539) {
return 1;
}
else if (Rs >= 1539) {
return 8;
}
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (Cs >= 2091) {
return 4;
}
}
}
}
else if (negBalance >= 19) {
if (Rs < 1660) {
if (ssDCval < 58) {
if (Rs < 989) {
return 8;
}
else if (Rs >= 989) {
return 4;
}
}
else if (ssDCval >= 58) {
if (negBalance < 27) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (Rs >= 1660) {
if (refDCval < 66) {
if (ssDCval < 65) {
if (negBalance < 22) {
if (Rs < 2678) {
return 8;
}
else if (Rs >= 2678) {
if (Rs < 2959) {
if (Rs < 2910) {
return 4;
}
else if (Rs >= 2910) {
return 4;
}
}
else if (Rs >= 2959) {
return 8;
}
}
}
else if (negBalance >= 22) {
return 8;
}
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 2302) {
return 8;
}
else if (Rs >= 2302) {
return 1;
}
}
}
}
else if (refDCval >= 66) {
if (Rs < 3452) {
return 1;
}
else if (Rs >= 3452) {
return 8;
}
}
}
}
}
else if (Cs >= 3312) {
return 8;
}
}
else if (ssDCval >= 67) {
if (mu_mv_mag_sq < 123) {
if (TSCindex < 1) {
if (refDCval < 92) {
if (Rs < 2369) {
return 1;
}
else if (Rs >= 2369) {
if (Cs < 2144) {
return 8;
}
else if (Cs >= 2144) {
if (Rs < 2485) {
return 2;
}
else if (Rs >= 2485) {
return 1;
}
}
}
}
else if (refDCval >= 92) {
if (ssDCval < 92) {
return 1;
}
else if (ssDCval >= 92) {
if (negBalance < 7) {
return 2;
}
else if (negBalance >= 7) {
if (mu_mv_mag_sq < 116) {
if (posBalance < 14) {
return 2;
}
else if (posBalance >= 14) {
if (Rs < 2298) {
return 2;
}
else if (Rs >= 2298) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 116) {
return 4;
}
}
}
}
}
else if (TSCindex >= 1) {
if (posBalance < 40) {
if (Cs < 2314) {
if (SC < 4835) {
return 8;
}
else if (SC >= 4835) {
if (Cs < 2150) {
return 2;
}
else if (Cs >= 2150) {
if (Rs < 3795) {
return 1;
}
else if (Rs >= 3795) {
return 4;
}
}
}
}
else if (Cs >= 2314) {
return 8;
}
}
else if (posBalance >= 40) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (SC < 7112) {
if (Rs < 1104) {
if (Rs < 653) {
return 8;
}
else if (Rs >= 653) {
return 1;
}
}
else if (Rs >= 1104) {
if (diffRsCsdiff < -34) {
if (diffAFD < -7) {
return 8;
}
else if (diffAFD >= -7) {
if (ssDCval < 94) {
return 4;
}
else if (ssDCval >= 94) {
return 8;
}
}
}
else if (diffRsCsdiff >= -34) {
if (SC < 4544) {
if (SC < 4437) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
return 4;
}
}
else if (SC >= 4437) {
if (SC < 4456) {
return 4;
}
else if (SC >= 4456) {
if (Rs < 2093) {
return 4;
}
else if (Rs >= 2093) {
return 8;
}
}
}
}
else if (SC >= 4544) {
if (mu_mv_mag_sq < 124) {
if (diffTSC < -1) {
if (Rs < 3388) {
return 8;
}
else if (Rs >= 3388) {
return 2;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
}
}
}
else if (SC >= 7112) {
if (SC < 7988) {
if (Rs < 3293) {
return 8;
}
else if (Rs >= 3293) {
return 4;
}
}
else if (SC >= 7988) {
if (diffRsCsdiff < -14) {
if (Rs < 4396) {
return 1;
}
else if (Rs >= 4396) {
return 8;
}
}
else if (diffRsCsdiff >= -14) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6660) {
if (SC < 8276) {
if (refDCval < 83) {
if (negBalance < 34) {
if (mu_mv_mag_sq < 121) {
if (Cs < 1819) {
if (RsDiff < 6) {
if (negBalance < 30) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 1043) {
return 8;
}
else if (Rs >= 1043) {
return 4;
}
}
}
else if (negBalance >= 30) {
return 4;
}
}
else if (RsDiff >= 6) {
if (Rs < 3232) {
return 8;
}
else if (Rs >= 3232) {
return 2;
}
}
}
else if (Cs >= 1819) {
if (mcTcor < 98) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
if (refDCval < 77) {
if (SC < 3872) {
return 8;
}
else if (SC >= 3872) {
if (SC < 5675) {
if (Rs < 1980) {
return 8;
}
else if (Rs >= 1980) {
return 4;
}
}
else if (SC >= 5675) {
if (Rs < 4966) {
if (Rs < 4651) {
if (CsDiff < 4) {
if (Rs < 4417) {
if (posBalance < 26) {
if (Rs < 4060) {
return 4;
}
else if (Rs >= 4060) {
if (Rs < 4289) {
return 4;
}
else if (Rs >= 4289) {
return 4;
}
}
}
else if (posBalance >= 26) {
return 4;
}
}
else if (Rs >= 4417) {
if (Rs < 4538) {
return 4;
}
else if (Rs >= 4538) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (diffMVdiffVal < -1) {
if (diffRsCsdiff < 14) {
if (Rs < 4570) {
if (Cs < 2261) {
if (Rs < 4230) {
return 4;
}
else if (Rs >= 4230) {
return 4;
}
}
else if (Cs >= 2261) {
return 4;
}
}
else if (Rs >= 4570) {
return 4;
}
}
else if (diffRsCsdiff >= 14) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (negBalance < 15) {
if (MVDiff < 1) {
if (Rs < 4392) {
return 4;
}
else if (Rs >= 4392) {
return 4;
}
}
else if (MVDiff >= 1) {
if (Rs < 4342) {
return 8;
}
else if (Rs >= 4342) {
return 8;
}
}
}
else if (negBalance >= 15) {
if (Rs < 4576) {
return 8;
}
else if (Rs >= 4576) {
if (Rs < 4632) {
return 4;
}
else if (Rs >= 4632) {
return 8;
}
}
}
}
}
}
else if (Rs >= 4651) {
if (diffMVdiffVal < 1) {
if (posBalance < 41) {
if (refDCval < 54) {
return 4;
}
else if (refDCval >= 54) {
if (RsCsDiff < 38) {
if (Rs < 4702) {
return 4;
}
else if (Rs >= 4702) {
if (Rs < 4820) {
return 4;
}
else if (Rs >= 4820) {
return 4;
}
}
}
else if (RsCsDiff >= 38) {
if (Cs < 2562) {
return 4;
}
else if (Cs >= 2562) {
return 4;
}
}
}
}
else if (posBalance >= 41) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
}
else if (Rs >= 4966) {
if (SC < 7723) {
return 8;
}
else if (SC >= 7723) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 5109) {
return 4;
}
else if (Rs >= 5109) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 77) {
if (posBalance < 86) {
return 8;
}
else if (posBalance >= 86) {
return 4;
}
}
}
}
else if (mcTcor >= 98) {
if (ssDCval < 80) {
if (negBalance < 19) {
if (Rs < 3500) {
return 4;
}
else if (Rs >= 3500) {
return 8;
}
}
else if (negBalance >= 19) {
return 8;
}
}
else if (ssDCval >= 80) {
if (Rs < 4958) {
return 8;
}
else if (Rs >= 4958) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (posBalance < 17) {
return 8;
}
else if (posBalance >= 17) {
if (Cs < 2465) {
if (SC < 4382) {
if (refDCval < 66) {
return 8;
}
else if (refDCval >= 66) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (SC >= 4382) {
if (Cs < 2181) {
if (Rs < 2616) {
return 8;
}
else if (Rs >= 2616) {
if (mu_mv_mag_sq < 126) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (mu_mv_mag_sq < 124) {
if (AFD < 5) {
if (Cs < 1957) {
return 4;
}
else if (Cs >= 1957) {
if (posBalance < 39) {
if (diffRsCsdiff < 4) {
if (Rs < 4231) {
if (Rs < 4184) {
return 4;
}
else if (Rs >= 4184) {
return 4;
}
}
else if (Rs >= 4231) {
if (Cs < 2162) {
return 4;
}
else if (Cs >= 2162) {
return 4;
}
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 4384) {
return 4;
}
else if (Rs >= 4384) {
return 8;
}
}
}
else if (posBalance >= 39) {
return 8;
}
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 124) {
if (Cs < 1787) {
return 8;
}
else if (Cs >= 1787) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (posBalance < 36) {
if (Rs < 3020) {
return 8;
}
else if (Rs >= 3020) {
return 8;
}
}
else if (posBalance >= 36) {
if (tcor < 99) {
if (tcor < 98) {
if (Rs < 2892) {
return 8;
}
else if (Rs >= 2892) {
if (Cs < 2143) {
if (negBalance < 22) {
if (Rs < 2972) {
return 4;
}
else if (Rs >= 2972) {
return 4;
}
}
else if (negBalance >= 22) {
return 4;
}
}
else if (Cs >= 2143) {
return 8;
}
}
}
else if (tcor >= 98) {
if (Rs < 2982) {
if (Rs < 2933) {
if (Rs < 2885) {
return 4;
}
else if (Rs >= 2885) {
return 4;
}
}
else if (Rs >= 2933) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
return 8;
}
}
}
else if (Rs >= 2982) {
if (Rs < 3012) {
return 4;
}
else if (Rs >= 3012) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (Rs < 3112) {
return 4;
}
else if (Rs >= 3112) {
return 4;
}
}
}
}
}
}
else if (tcor >= 99) {
return 4;
}
}
}
}
}
else if (Cs >= 2181) {
if (Rs < 2293) {
return 4;
}
else if (Rs >= 2293) {
if (Rs < 4778) {
if (ssDCval < 55) {
if (Rs < 4195) {
return 4;
}
else if (Rs >= 4195) {
if (Rs < 4457) {
return 4;
}
else if (Rs >= 4457) {
return 8;
}
}
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (Rs >= 4778) {
return 4;
}
}
}
}
}
else if (Cs >= 2465) {
if (negBalance < 14) {
if (Rs < 3578) {
return 1;
}
else if (Rs >= 3578) {
if (Rs < 4480) {
return 4;
}
else if (Rs >= 4480) {
return 8;
}
}
}
else if (negBalance >= 14) {
if (Rs < 2318) {
return 4;
}
else if (Rs >= 2318) {
if (Rs < 5135) {
return 8;
}
else if (Rs >= 5135) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
}
}
}
}
}
else if (negBalance >= 34) {
if (MVDiff < 4) {
if (Rs < 3001) {
if (Cs < 2965) {
if (SC < 4347) {
return 8;
}
else if (SC >= 4347) {
if (Rs < 2228) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
else if (Rs >= 2228) {
if (mu_mv_mag_sq < 118) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
if (negBalance < 45) {
if (Rs < 2387) {
return 8;
}
else if (Rs >= 2387) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 4;
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
}
}
else if (Cs >= 2965) {
if (Rs < 2972) {
return 4;
}
else if (Rs >= 2972) {
return 8;
}
}
}
else if (Rs >= 3001) {
if (Cs < 2048) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 3236) {
return 4;
}
else if (Rs >= 3236) {
return 2;
}
}
}
else if (Cs >= 2048) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (SC < 5134) {
if (Cs < 1925) {
return 8;
}
else if (Cs >= 1925) {
return 4;
}
}
else if (SC >= 5134) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 4;
}
}
}
}
}
else if (refDCval >= 83) {
if (Cs < 2647) {
if (Rs < 3845) {
if (Cs < 1881) {
if (tcor < 98) {
if (diffRsCsdiff < 17) {
return 8;
}
else if (diffRsCsdiff >= 17) {
if (Rs < 2281) {
return 4;
}
else if (Rs >= 2281) {
if (Rs < 3048) {
return 8;
}
else if (Rs >= 3048) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (SC < 4192) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (SC >= 4192) {
if (ssDCval < 94) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (ssDCval >= 94) {
if (Rs < 3492) {
return 8;
}
else if (Rs >= 3492) {
return 2;
}
}
}
}
}
else if (Cs >= 1881) {
if (negBalance < 49) {
if (tcor < 99) {
if (Cs < 2618) {
if (diffRsCsdiff < 24) {
if (Rs < 1943) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Rs < 1720) {
return 2;
}
else if (Rs >= 1720) {
return 4;
}
}
}
else if (Rs >= 1943) {
return 8;
}
}
else if (diffRsCsdiff >= 24) {
if (Cs < 1986) {
return 2;
}
else if (Cs >= 1986) {
return 8;
}
}
}
else if (Cs >= 2618) {
if (Rs < 3399) {
return 4;
}
else if (Rs >= 3399) {
return 8;
}
}
}
else if (tcor >= 99) {
if (ssDCval < 90) {
if (Rs < 3029) {
return 8;
}
else if (Rs >= 3029) {
if (Cs < 2413) {
return 2;
}
else if (Cs >= 2413) {
return 8;
}
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (mu_mv_mag_sq < 119) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 119) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
}
else if (Rs >= 3845) {
if (negBalance < 29) {
if (tcor < 97) {
if (Cs < 1810) {
return 4;
}
else if (Cs >= 1810) {
return 8;
}
}
else if (tcor >= 97) {
if (posBalance < 41) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (Cs < 1880) {
return 4;
}
else if (Cs >= 1880) {
if (Rs < 4809) {
if (ssDCval < 99) {
return 2;
}
else if (ssDCval >= 99) {
return 4;
}
}
else if (Rs >= 4809) {
return 4;
}
}
}
}
else if (posBalance >= 41) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 4;
}
}
}
}
else if (negBalance >= 29) {
if (Cs < 2284) {
if (MVDiff < 1) {
if (ssDCval < 91) {
if (Rs < 4613) {
return 2;
}
else if (Rs >= 4613) {
return 8;
}
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (Cs >= 2284) {
return 1;
}
}
}
}
else if (Cs >= 2647) {
if (ssDCval < 98) {
if (refDCval < 89) {
if (negBalance < 38) {
if (Rs < 4285) {
return 4;
}
else if (Rs >= 4285) {
return 8;
}
}
else if (negBalance >= 38) {
if (Cs < 3267) {
if (Rs < 3257) {
if (Rs < 2606) {
return 4;
}
else if (Rs >= 2606) {
return 8;
}
}
else if (Rs >= 3257) {
return 4;
}
}
else if (Cs >= 3267) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Rs < 3617) {
return 4;
}
else if (Rs >= 3617) {
return 8;
}
}
}
}
}
else if (refDCval >= 89) {
if (mu_mv_mag_sq < 123) {
if (Cs < 4559) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 2992) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
return 4;
}
}
else if (Rs >= 2992) {
return 4;
}
}
}
else if (Cs >= 4559) {
return 8;
}
}
else if (mu_mv_mag_sq >= 123) {
if (SC < 7122) {
if (Rs < 3395) {
return 8;
}
else if (Rs >= 3395) {
if (mcTcor < 99) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Cs < 3576) {
if (RsDiff < 4) {
if (Cs < 3231) {
return 8;
}
else if (Cs >= 3231) {
return 4;
}
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (Cs >= 3576) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (Cs < 3275) {
return 8;
}
else if (Cs >= 3275) {
return 4;
}
}
}
}
else if (SC >= 7122) {
if (mcTcor < 96) {
if (ssDCval < 93) {
return 8;
}
else if (ssDCval >= 93) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (SC < 7193) {
if (Rs < 3839) {
return 4;
}
else if (Rs >= 3839) {
return 8;
}
}
else if (SC >= 7193) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 98) {
if (Rs < 3252) {
if (refDCval < 101) {
if (Cs < 2724) {
if (Rs < 3222) {
return 8;
}
else if (Rs >= 3222) {
return 4;
}
}
else if (Cs >= 2724) {
return 8;
}
}
else if (refDCval >= 101) {
return 4;
}
}
else if (Rs >= 3252) {
return 8;
}
}
}
}
}
else if (SC >= 8276) {
if (ssDCval < 78) {
if (Rs < 5974) {
if (diffAFD < 2) {
if (negBalance < 24) {
if (Rs < 4773) {
return 1;
}
else if (Rs >= 4773) {
return 8;
}
}
else if (negBalance >= 24) {
return 8;
}
}
else if (diffAFD >= 2) {
if (Rs < 4605) {
return 1;
}
else if (Rs >= 4605) {
return 8;
}
}
}
else if (Rs >= 5974) {
if (tcor < 98) {
return 2;
}
else if (tcor >= 98) {
if (Rs < 6391) {
return 8;
}
else if (Rs >= 6391) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (Cs < 3664) {
if (Rs < 5223) {
return 4;
}
else if (Rs >= 5223) {
return 8;
}
}
else if (Cs >= 3664) {
return 8;
}
}
}
}
else if (Rs >= 6660) {
if (ssDCval < 68) {
if (Rs < 8744) {
return 8;
}
else if (Rs >= 8744) {
return 2;
}
}
else if (ssDCval >= 68) {
if (ssDCval < 96) {
return 4;
}
else if (ssDCval >= 96) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (negBalance < 37) {
if (refDCval < 117) {
if (Rs < 1645) {
if (Cs < 2079) {
return 8;
}
else if (Cs >= 2079) {
return 2;
}
}
else if (Rs >= 1645) {
if (diffMVdiffVal < -124) {
return 4;
}
else if (diffMVdiffVal >= -124) {
if (refDCval < 105) {
if (Cs < 1750) {
return 4;
}
else if (Cs >= 1750) {
if (Rs < 2116) {
return 2;
}
else if (Rs >= 2116) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
if (Cs < 2807) {
return 2;
}
else if (Cs >= 2807) {
return 8;
}
}
}
}
}
else if (refDCval >= 105) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
if (Cs < 1882) {
if (Rs < 2802) {
return 8;
}
else if (Rs >= 2802) {
return 4;
}
}
else if (Cs >= 1882) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 117) {
if (ssDCval < 200) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 118) {
if (CsDiff < 1) {
if (Cs < 2966) {
if (posBalance < 14) {
return 8;
}
else if (posBalance >= 14) {
if (Cs < 2936) {
return 4;
}
else if (Cs >= 2936) {
return 8;
}
}
}
else if (Cs >= 2966) {
return 2;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (ssDCval >= 118) {
if (Rs < 4952) {
if (Rs < 1258) {
if (Rs < 1170) {
return 8;
}
else if (Rs >= 1170) {
if (ssDCval < 132) {
return 4;
}
else if (ssDCval >= 132) {
return 8;
}
}
}
else if (Rs >= 1258) {
if (diffAFD < 6) {
if (diffTSC < -3) {
if (Cs < 1900) {
return 1;
}
else if (Cs >= 1900) {
if (Cs < 3860) {
return 8;
}
else if (Cs >= 3860) {
return 1;
}
}
}
else if (diffTSC >= -3) {
return 8;
}
}
else if (diffAFD >= 6) {
return 1;
}
}
}
else if (Rs >= 4952) {
if (Cs < 2892) {
return 2;
}
else if (Cs >= 2892) {
if (Cs < 3633) {
if (Rs < 6095) {
return 8;
}
else if (Rs >= 6095) {
return 4;
}
}
else if (Cs >= 3633) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (diffRsCsdiff < -1161) {
if (Rs < 4223) {
return 1;
}
else if (Rs >= 4223) {
return 8;
}
}
else if (diffRsCsdiff >= -1161) {
if (ssDCval < 138) {
if (refDCval < 123) {
return 8;
}
else if (refDCval >= 123) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Cs < 3865) {
if (tcor < 100) {
if (ssDCval < 136) {
return 8;
}
else if (ssDCval >= 136) {
if (MVDiff < 5) {
if (Rs < 3763) {
return 1;
}
else if (Rs >= 3763) {
return 2;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (tcor >= 100) {
if (posBalance < 44) {
if (ssDCval < 127) {
if (Rs < 2057) {
return 8;
}
else if (Rs >= 2057) {
return 1;
}
}
else if (ssDCval >= 127) {
return 8;
}
}
else if (posBalance >= 44) {
return 1;
}
}
}
else if (Cs >= 3865) {
if (Rs < 6511) {
return 1;
}
else if (Rs >= 6511) {
return 8;
}
}
}
}
}
else if (ssDCval >= 138) {
return 8;
}
}
}
}
else if (ssDCval >= 200) {
if (Cs < 4080) {
return 8;
}
else if (Cs >= 4080) {
return 4;
}
}
}
}
else if (negBalance >= 37) {
if (ssDCval < 137) {
if (Rs < 1649) {
if (Cs < 1940) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (ssDCval < 106) {
return 8;
}
else if (ssDCval >= 106) {
return 2;
}
}
}
else if (Cs >= 1940) {
if (Rs < 1153) {
if (Rs < 1035) {
return 8;
}
else if (Rs >= 1035) {
return 4;
}
}
else if (Rs >= 1153) {
return 2;
}
}
}
else if (Rs >= 1649) {
if (refDCval < 105) {
if (Cs < 2792) {
if (Rs < 4919) {
if (Rs < 3463) {
return 8;
}
else if (Rs >= 3463) {
return 1;
}
}
else if (Rs >= 4919) {
return 2;
}
}
else if (Cs >= 2792) {
return 8;
}
}
else if (refDCval >= 105) {
if (posBalance < 62) {
if (RsDiff < 2) {
if (refDCval < 113) {
return 2;
}
else if (refDCval >= 113) {
if (tcor < 98) {
if (Rs < 2713) {
return 4;
}
else if (Rs >= 2713) {
return 2;
}
}
else if (tcor >= 98) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (MVDiff < 6) {
if (diffAFD < -3) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
if (mu_mv_mag_sq < 114) {
if (Rs < 3593) {
return 1;
}
else if (Rs >= 3593) {
return 8;
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
}
else if (diffAFD >= -3) {
if (negBalance < 55) {
if (diffTSC < 3) {
if (posBalance < 47) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
return 4;
}
}
else if (posBalance >= 47) {
return 8;
}
}
else if (diffTSC >= 3) {
if (Cs < 2050) {
return 4;
}
else if (Cs >= 2050) {
if (ssDCval < 105) {
return 2;
}
else if (ssDCval >= 105) {
return 8;
}
}
}
}
else if (negBalance >= 55) {
if (Rs < 4144) {
if (Rs < 2970) {
return 8;
}
else if (Rs >= 2970) {
return 2;
}
}
else if (Rs >= 4144) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 117) {
if (Rs < 5215) {
if (Rs < 3422) {
return 8;
}
else if (Rs >= 3422) {
return 2;
}
}
else if (Rs >= 5215) {
return 8;
}
}
else if (ssDCval >= 117) {
if (Rs < 3743) {
if (Rs < 3317) {
return 1;
}
else if (Rs >= 3317) {
return 4;
}
}
else if (Rs >= 3743) {
return 2;
}
}
}
}
}
else if (posBalance >= 62) {
if (Rs < 3544) {
if (negBalance < 52) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 2207) {
return 8;
}
else if (Rs >= 2207) {
return 1;
}
}
}
else if (negBalance >= 52) {
if (Rs < 2934) {
return 8;
}
else if (Rs >= 2934) {
return 2;
}
}
}
else if (Rs >= 3544) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
if (mcTcor < 96) {
if (Rs < 5967) {
return 1;
}
else if (Rs >= 5967) {
return 8;
}
}
else if (mcTcor >= 96) {
if (refDCval < 144) {
if (mu_mv_mag_sq < 125) {
if (Rs < 5663) {
return 8;
}
else if (Rs >= 5663) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 3581) {
return 8;
}
else if (Rs >= 3581) {
return 1;
}
}
}
else if (refDCval >= 144) {
return 8;
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect9(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1325) {
if (Rs < 2373) {
if (Cs < 533) {
if (negBalance < 14) {
if (mu_mv_mag_sq < 108) {
if (tcor < 95) {
if (negBalance < 9) {
if (MVDiff < 207) {
if (refDCval < 45) {
if (AFD < 1) {
if (posBalance < 28) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
if (Rs < 17) {
return 8;
}
else if (Rs >= 17) {
return 1;
}
}
}
else if (posBalance >= 28) {
return 8;
}
}
else if (AFD >= 1) {
if (mcTcor < 94) {
if (mu_mv_mag_sq < 47) {
if (posBalance < 114) {
if (Cs < 49) {
if (TSC < 1) {
return 2;
}
else if (TSC >= 1) {
return 8;
}
}
else if (Cs >= 49) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 2;
}
}
}
else if (posBalance >= 114) {
if (Cs < 340) {
if (gchDC < 5) {
return 4;
}
else if (gchDC >= 5) {
return 8;
}
}
else if (Cs >= 340) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (MVDiff < 31) {
if (diffMVdiffVal < -12) {
if (diffMVdiffVal < -21) {
if (Rs < 444) {
return 8;
}
else if (Rs >= 444) {
return 4;
}
}
else if (diffMVdiffVal >= -21) {
return 4;
}
}
else if (diffMVdiffVal >= -12) {
return 8;
}
}
else if (MVDiff >= 31) {
if (diffAFD < 4) {
if (MVDiff < 78) {
return 4;
}
else if (MVDiff >= 78) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
}
else if (diffAFD >= 4) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (refDCval < 30) {
if (Cs < 14) {
return 4;
}
else if (Cs >= 14) {
if (Cs < 283) {
return 1;
}
else if (Cs >= 283) {
return 4;
}
}
}
else if (refDCval >= 30) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
}
else if (refDCval >= 45) {
if (Rs < 32) {
return 2;
}
else if (Rs >= 32) {
if (ssDCval < 151) {
if (gchDC < 7) {
if (posBalance < 93) {
if (Cs < 25) {
return 2;
}
else if (Cs >= 25) {
return 8;
}
}
else if (posBalance >= 93) {
if (Rs < 546) {
if (Rs < 132) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (Rs >= 132) {
return 8;
}
}
else if (Rs >= 546) {
return 4;
}
}
}
else if (gchDC >= 7) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (ssDCval >= 151) {
return 2;
}
}
}
}
else if (MVDiff >= 207) {
if (AFD < 7) {
if (Rs < 49) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
else if (Rs >= 49) {
if (Rs < 212) {
return 8;
}
else if (Rs >= 212) {
if (Rs < 341) {
return 1;
}
else if (Rs >= 341) {
return 4;
}
}
}
}
else if (AFD >= 7) {
if (Cs < 318) {
if (diffTSC < -3) {
return 2;
}
else if (diffTSC >= -3) {
return 1;
}
}
else if (Cs >= 318) {
return 4;
}
}
}
}
else if (negBalance >= 9) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (refDCval < 47) {
if (refDCval < 25) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (negBalance < 11) {
return 1;
}
else if (negBalance >= 11) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 1;
}
}
}
}
else if (refDCval >= 25) {
if (posBalance < 46) {
if (Cs < 331) {
if (ssDCval < 29) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
if (Rs < 394) {
return 4;
}
else if (Rs >= 394) {
return 1;
}
}
}
else if (ssDCval >= 29) {
if (posBalance < 33) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (posBalance >= 33) {
if (Rs < 171) {
return 8;
}
else if (Rs >= 171) {
return 4;
}
}
}
}
else if (Cs >= 331) {
return 1;
}
}
else if (posBalance >= 46) {
if (Cs < 55) {
return 1;
}
else if (Cs >= 55) {
if (ssDCval < 34) {
return 8;
}
else if (ssDCval >= 34) {
if (Cs < 466) {
return 1;
}
else if (Cs >= 466) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 47) {
if (MVDiff < 25) {
if (Rs < 1221) {
if (Cs < 118) {
return 8;
}
else if (Cs >= 118) {
if (ssDCval < 120) {
return 1;
}
else if (ssDCval >= 120) {
return 8;
}
}
}
else if (Rs >= 1221) {
if (Rs < 1736) {
return 8;
}
else if (Rs >= 1736) {
return 2;
}
}
}
else if (MVDiff >= 25) {
if (diffMVdiffVal < 17) {
if (Rs < 162) {
return 4;
}
else if (Rs >= 162) {
return 2;
}
}
else if (diffMVdiffVal >= 17) {
if (Rs < 1053) {
if (mu_mv_mag_sq < 39) {
return 2;
}
else if (mu_mv_mag_sq >= 39) {
return 8;
}
}
else if (Rs >= 1053) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 95) {
if (MVDiff < 8) {
if (Cs < 183) {
if (Rs < 94) {
if (ssDCval < 67) {
if (posBalance < 23) {
if (Rs < 70) {
if (ssDCval < 21) {
return 8;
}
else if (ssDCval >= 21) {
return 4;
}
}
else if (Rs >= 70) {
return 8;
}
}
else if (posBalance >= 23) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (Cs < 55) {
return 1;
}
else if (Cs >= 55) {
if (Rs < 78) {
return 4;
}
else if (Rs >= 78) {
return 8;
}
}
}
}
}
else if (ssDCval >= 67) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
return 4;
}
}
}
else if (Rs >= 94) {
if (negBalance < 12) {
if (ssDCval < 26) {
if (Rs < 136) {
return 1;
}
else if (Rs >= 136) {
return 4;
}
}
else if (ssDCval >= 26) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
if (Rs < 562) {
return 1;
}
else if (Rs >= 562) {
return 8;
}
}
}
}
else if (negBalance >= 12) {
if (Cs < 134) {
if (Cs < 57) {
return 8;
}
else if (Cs >= 57) {
if (Rs < 114) {
return 8;
}
else if (Rs >= 114) {
return 1;
}
}
}
else if (Cs >= 134) {
return 4;
}
}
}
}
else if (Cs >= 183) {
if (SC < 852) {
if (posBalance < 22) {
return 1;
}
else if (posBalance >= 22) {
if (ssDCval < 50) {
if (Rs < 163) {
return 1;
}
else if (Rs >= 163) {
return 4;
}
}
else if (ssDCval >= 50) {
if (AFD < 3) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
else if (AFD >= 3) {
return 1;
}
}
}
}
else if (SC >= 852) {
if (Cs < 393) {
if (Cs < 333) {
if (AFD < 2) {
if (Rs < 647) {
return 4;
}
else if (Rs >= 647) {
return 8;
}
}
else if (AFD >= 2) {
return 1;
}
}
else if (Cs >= 333) {
return 1;
}
}
else if (Cs >= 393) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (Rs < 1279) {
return 1;
}
else if (Rs >= 1279) {
if (Rs < 2069) {
return 8;
}
else if (Rs >= 2069) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 8) {
if (ssDCval < 109) {
if (Cs < 109) {
if (Cs < 8) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
if (posBalance < 60) {
if (MVDiff < 9) {
if (ssDCval < 36) {
return 4;
}
else if (ssDCval >= 36) {
return 2;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
else if (Cs >= 8) {
if (SC < 117) {
return 1;
}
else if (SC >= 117) {
if (negBalance < 8) {
if (Rs < 399) {
return 4;
}
else if (Rs >= 399) {
return 8;
}
}
else if (negBalance >= 8) {
if (Rs < 69) {
return 8;
}
else if (Rs >= 69) {
return 1;
}
}
}
}
}
else if (Cs >= 109) {
if (Rs < 554) {
if (mu_mv_mag_sq < 72) {
return 4;
}
else if (mu_mv_mag_sq >= 72) {
if (diffMVdiffVal < 5) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Rs < 141) {
return 4;
}
else if (Rs >= 141) {
if (Rs < 348) {
return 1;
}
else if (Rs >= 348) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 5) {
if (negBalance < 11) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (Rs < 85) {
return 4;
}
else if (Rs >= 85) {
return 8;
}
}
}
else if (negBalance >= 11) {
return 1;
}
}
}
}
else if (Rs >= 554) {
if (Rs < 941) {
return 8;
}
else if (Rs >= 941) {
return 1;
}
}
}
}
else if (ssDCval >= 109) {
if (SC < 284) {
if (ssDCval < 190) {
return 4;
}
else if (ssDCval >= 190) {
if (Rs < 60) {
return 2;
}
else if (Rs >= 60) {
return 1;
}
}
}
else if (SC >= 284) {
if (ssDCval < 160) {
if (RsDiff < 1) {
if (mu_mv_mag_sq < 87) {
return 8;
}
else if (mu_mv_mag_sq >= 87) {
if (Cs < 405) {
return 1;
}
else if (Cs >= 405) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (Rs < 1032) {
return 1;
}
else if (Rs >= 1032) {
return 8;
}
}
}
else if (ssDCval >= 160) {
if (refDCval < 161) {
return 1;
}
else if (refDCval >= 161) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (SC < 782) {
if (refDCval < 47) {
if (Cs < 102) {
if (ssDCval < 38) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
if (ssDCval < 9) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 9) {
return 8;
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (Cs < 68) {
if (Rs < 2) {
return 4;
}
else if (Rs >= 2) {
if (negBalance < 10) {
if (tcor < 102) {
return 8;
}
else if (tcor >= 102) {
return 4;
}
}
else if (negBalance >= 10) {
return 4;
}
}
}
else if (Cs >= 68) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
}
}
else if (ssDCval >= 38) {
if (Cs < 82) {
return 8;
}
else if (Cs >= 82) {
if (Cs < 86) {
return 1;
}
else if (Cs >= 86) {
return 8;
}
}
}
}
else if (Cs >= 102) {
if (Rs < 228) {
if (refDCval < 36) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (ssDCval < 28) {
return 1;
}
else if (ssDCval >= 28) {
if (Rs < 98) {
return 4;
}
else if (Rs >= 98) {
return 8;
}
}
}
}
else if (refDCval >= 36) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (refDCval < 37) {
if (Cs < 249) {
if (Rs < 140) {
return 8;
}
else if (Rs >= 140) {
return 4;
}
}
else if (Cs >= 249) {
return 1;
}
}
else if (refDCval >= 37) {
if (Rs < 75) {
return 8;
}
else if (Rs >= 75) {
return 1;
}
}
}
}
}
else if (Rs >= 228) {
if (refDCval < 39) {
if (tcor < 100) {
if (ssDCval < 35) {
if (Rs < 426) {
if (Rs < 287) {
return 8;
}
else if (Rs >= 287) {
return 4;
}
}
else if (Rs >= 426) {
if (Rs < 479) {
return 8;
}
else if (Rs >= 479) {
return 4;
}
}
}
else if (ssDCval >= 35) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
else if (tcor >= 100) {
return 8;
}
}
else if (refDCval >= 39) {
if (ssDCval < 44) {
if (ssDCval < 41) {
return 4;
}
else if (ssDCval >= 41) {
return 1;
}
}
else if (ssDCval >= 44) {
return 4;
}
}
}
}
}
else if (refDCval >= 47) {
if (ssDCval < 84) {
if (Cs < 26) {
return 2;
}
else if (Cs >= 26) {
if (ssDCval < 66) {
if (Cs < 283) {
if (ssDCval < 56) {
if (Rs < 121) {
return 8;
}
else if (Rs >= 121) {
if (ssDCval < 54) {
if (Rs < 232) {
return 8;
}
else if (Rs >= 232) {
return 4;
}
}
else if (ssDCval >= 54) {
return 4;
}
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (Cs >= 283) {
if (Rs < 251) {
return 4;
}
else if (Rs >= 251) {
if (Rs < 341) {
return 1;
}
else if (Rs >= 341) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
if (Rs < 78) {
return 4;
}
else if (Rs >= 78) {
if (Cs < 242) {
if (Cs < 130) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
return 8;
}
}
else if (Cs >= 130) {
return 2;
}
}
else if (Cs >= 242) {
if (SC < 646) {
return 4;
}
else if (SC >= 646) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (Cs < 327) {
if (Cs < 156) {
if (Rs < 129) {
return 4;
}
else if (Rs >= 129) {
return 8;
}
}
else if (Cs >= 156) {
if (Rs < 431) {
if (ssDCval < 130) {
if (Rs < 150) {
return 4;
}
else if (Rs >= 150) {
if (MVDiff < 39) {
return 2;
}
else if (MVDiff >= 39) {
return 8;
}
}
}
else if (ssDCval >= 130) {
return 4;
}
}
else if (Rs >= 431) {
return 1;
}
}
}
else if (Cs >= 327) {
if (Rs < 381) {
return 4;
}
else if (Rs >= 381) {
if (Cs < 355) {
return 2;
}
else if (Cs >= 355) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 782) {
if (diffTSC < -3) {
if (posBalance < 3) {
return 4;
}
else if (posBalance >= 3) {
if (diffAFD < -7) {
if (ssDCval < 149) {
if (ssDCval < 101) {
if (Rs < 707) {
return 1;
}
else if (Rs >= 707) {
return 8;
}
}
else if (ssDCval >= 101) {
return 2;
}
}
else if (ssDCval >= 149) {
return 8;
}
}
else if (diffAFD >= -7) {
if (Rs < 823) {
if (negBalance < 4) {
return 4;
}
else if (negBalance >= 4) {
if (Rs < 550) {
return 2;
}
else if (Rs >= 550) {
return 8;
}
}
}
else if (Rs >= 823) {
return 1;
}
}
}
}
else if (diffTSC >= -3) {
if (mu_mv_mag_sq < 128) {
if (SC < 945) {
if (Rs < 388) {
return 1;
}
else if (Rs >= 388) {
if (ssDCval < 38) {
if (Cs < 260) {
return 4;
}
else if (Cs >= 260) {
if (Cs < 375) {
return 8;
}
else if (Cs >= 375) {
return 1;
}
}
}
else if (ssDCval >= 38) {
if (TSC < 1) {
return 8;
}
else if (TSC >= 1) {
if (mu_mv_mag_sq < 127) {
if (Rs < 514) {
return 8;
}
else if (Rs >= 514) {
if (Rs < 568) {
return 1;
}
else if (Rs >= 568) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 463) {
return 1;
}
else if (Rs >= 463) {
return 8;
}
}
}
}
}
}
else if (SC >= 945) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Cs < 480) {
if (Rs < 894) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 2;
}
}
else if (Rs >= 894) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
}
else if (Cs >= 480) {
return 1;
}
}
else if (ssDCval >= 60) {
if (Cs < 446) {
return 1;
}
else if (Cs >= 446) {
if (Rs < 1471) {
return 2;
}
else if (Rs >= 1471) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (refDCval < 141) {
if (SC < 1038) {
if (Rs < 551) {
return 8;
}
else if (Rs >= 551) {
return 4;
}
}
else if (SC >= 1038) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 1114) {
return 8;
}
else if (Rs >= 1114) {
return 2;
}
}
}
}
else if (refDCval >= 141) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (Cs < 310) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (ssDCval < 87) {
if (Cs < 258) {
return 2;
}
else if (Cs >= 258) {
return 1;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
}
else if (Cs >= 310) {
if (Cs < 480) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 770) {
if (Rs < 561) {
return 8;
}
else if (Rs >= 561) {
return 4;
}
}
else if (Rs >= 770) {
return 8;
}
}
}
else if (Cs >= 480) {
if (Rs < 1389) {
return 8;
}
else if (Rs >= 1389) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
return 2;
}
}
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (tcor < 65) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (Cs < 377) {
if (negBalance < 45) {
if (posBalance < 65) {
if (Cs < 128) {
return 1;
}
else if (Cs >= 128) {
if (MVDiff < 36) {
if (Cs < 273) {
if (ssDCval < 34) {
return 1;
}
else if (ssDCval >= 34) {
if (Rs < 164) {
return 4;
}
else if (Rs >= 164) {
return 1;
}
}
}
else if (Cs >= 273) {
if (AFD < 6) {
if (Rs < 413) {
return 2;
}
else if (Rs >= 413) {
return 8;
}
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (MVDiff >= 36) {
if (negBalance < 29) {
if (Rs < 140) {
if (Rs < 109) {
return 2;
}
else if (Rs >= 109) {
return 1;
}
}
else if (Rs >= 140) {
return 8;
}
}
else if (negBalance >= 29) {
return 1;
}
}
}
}
else if (posBalance >= 65) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (Rs < 528) {
if (AFD < 3) {
if (Rs < 24) {
return 8;
}
else if (Rs >= 24) {
return 1;
}
}
else if (AFD >= 3) {
if (MVDiff < 377) {
return 8;
}
else if (MVDiff >= 377) {
return 4;
}
}
}
else if (Rs >= 528) {
return 1;
}
}
}
else if (Cs >= 377) {
if (RsDiff < 4) {
if (ssDCval < 29) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
return 2;
}
}
else if (ssDCval >= 29) {
return 1;
}
}
else if (RsDiff >= 4) {
if (Rs < 1204) {
return 8;
}
else if (Rs >= 1204) {
return 1;
}
}
}
}
}
else if (tcor >= 65) {
if (posBalance < 27) {
if (TSCindex < 2) {
if (mu_mv_mag_sq < 111) {
if (MVDiff < 5) {
if (refDCval < 33) {
return 4;
}
else if (refDCval >= 33) {
if (Rs < 97) {
return 8;
}
else if (Rs >= 97) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (posBalance < 20) {
if (AFD < 5) {
if (refDCval < 23) {
if (Rs < 27) {
if (MVDiff < 705) {
return 8;
}
else if (MVDiff >= 705) {
return 4;
}
}
else if (Rs >= 27) {
if (Rs < 239) {
return 1;
}
else if (Rs >= 239) {
return 4;
}
}
}
else if (refDCval >= 23) {
if (negBalance < 81) {
return 4;
}
else if (negBalance >= 81) {
return 8;
}
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (posBalance >= 20) {
if (ssDCval < 30) {
return 1;
}
else if (ssDCval >= 30) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
return 8;
}
}
else if (TSCindex >= 2) {
if (negBalance < 57) {
if (negBalance < 17) {
return 8;
}
else if (negBalance >= 17) {
if (RsDiff < 2) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (ssDCval < 28) {
return 1;
}
else if (ssDCval >= 28) {
return 4;
}
}
}
else if (RsDiff >= 2) {
return 2;
}
}
}
else if (negBalance >= 57) {
if (Cs < 11) {
if (mu_mv_mag_sq < 37) {
if (Rs < 14) {
return 8;
}
else if (Rs >= 14) {
return 4;
}
}
else if (mu_mv_mag_sq >= 37) {
return 4;
}
}
else if (Cs >= 11) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
return 1;
}
}
}
}
}
else if (posBalance >= 27) {
if (posBalance < 44) {
if (ssDCval < 30) {
if (Cs < 264) {
if (mu_mv_mag_sq < 26) {
if (mu_mv_mag_sq < 5) {
if (Rs < 385) {
return 1;
}
else if (Rs >= 385) {
if (Rs < 620) {
return 8;
}
else if (Rs >= 620) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (ssDCval < 25) {
return 1;
}
else if (ssDCval >= 25) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (diffRsCsdiff < -1) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (posBalance < 42) {
if (diffMVdiffVal < -11) {
if (Rs < 256) {
return 4;
}
else if (Rs >= 256) {
return 1;
}
}
else if (diffMVdiffVal >= -11) {
return 1;
}
}
else if (posBalance >= 42) {
return 8;
}
}
}
else if (diffRsCsdiff >= -1) {
if (mu_mv_mag_sq < 41) {
if (Rs < 261) {
return 1;
}
else if (Rs >= 261) {
return 4;
}
}
else if (mu_mv_mag_sq >= 41) {
return 1;
}
}
}
}
else if (Cs >= 264) {
if (refDCval < 23) {
if (Rs < 393) {
return 8;
}
else if (Rs >= 393) {
return 4;
}
}
else if (refDCval >= 23) {
if (MVDiff < 26) {
return 1;
}
else if (MVDiff >= 26) {
return 4;
}
}
}
}
else if (ssDCval >= 30) {
if (Cs < 260) {
if (mcTcor < 96) {
if (AFD < 3) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 131) {
return 8;
}
else if (Rs >= 131) {
return 1;
}
}
}
else if (AFD >= 3) {
if (posBalance < 38) {
if (posBalance < 30) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (Rs < 281) {
if (Rs < 139) {
return 1;
}
else if (Rs >= 139) {
return 4;
}
}
else if (Rs >= 281) {
return 2;
}
}
}
else if (posBalance >= 30) {
if (MVDiff < 4) {
if (Rs < 497) {
return 8;
}
else if (Rs >= 497) {
return 1;
}
}
else if (MVDiff >= 4) {
if (Rs < 154) {
return 8;
}
else if (Rs >= 154) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
}
else if (posBalance >= 38) {
if (mcTcor < 92) {
if (Rs < 201) {
if (Rs < 112) {
return 1;
}
else if (Rs >= 112) {
return 4;
}
}
else if (Rs >= 201) {
return 8;
}
}
else if (mcTcor >= 92) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 316) {
if (mu_mv_mag_sq < 35) {
if (Rs < 74) {
return 1;
}
else if (Rs >= 74) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (Cs < 59) {
return 4;
}
else if (Cs >= 59) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (SC < 220) {
if (negBalance < 18) {
if (MVDiff < 5) {
if (Rs < 106) {
return 4;
}
else if (Rs >= 106) {
return 1;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (negBalance >= 18) {
return 1;
}
}
else if (SC >= 220) {
if (Cs < 56) {
return 2;
}
else if (Cs >= 56) {
if (Cs < 185) {
return 1;
}
else if (Cs >= 185) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 1;
}
}
}
}
}
}
else if (Rs >= 316) {
if (Cs < 166) {
return 1;
}
else if (Cs >= 166) {
return 4;
}
}
}
}
else if (Cs >= 260) {
return 1;
}
}
}
else if (posBalance >= 44) {
if (posBalance < 67) {
if (CsDiff < 3) {
if (Rs < 365) {
if (mcTcor < 91) {
if (TSC < 3) {
if (MVDiff < 46) {
if (negBalance < 19) {
if (mcTcor < 88) {
if (Rs < 159) {
return 1;
}
else if (Rs >= 159) {
return 8;
}
}
else if (mcTcor >= 88) {
return 1;
}
}
else if (negBalance >= 19) {
return 1;
}
}
else if (MVDiff >= 46) {
if (Cs < 228) {
if (Rs < 80) {
return 1;
}
else if (Rs >= 80) {
return 8;
}
}
else if (Cs >= 228) {
return 2;
}
}
}
else if (TSC >= 3) {
if (Cs < 130) {
return 1;
}
else if (Cs >= 130) {
if (Rs < 169) {
return 1;
}
else if (Rs >= 169) {
return 2;
}
}
}
}
else if (mcTcor >= 91) {
if (RsCsDiff < 5) {
return 1;
}
else if (RsCsDiff >= 5) {
if (ssDCval < 29) {
return 8;
}
else if (ssDCval >= 29) {
return 1;
}
}
}
}
else if (Rs >= 365) {
return 1;
}
}
else if (CsDiff >= 3) {
if (Rs < 537) {
return 2;
}
else if (Rs >= 537) {
return 8;
}
}
}
else if (posBalance >= 67) {
if (mu_mv_mag_sq < 14) {
return 2;
}
else if (mu_mv_mag_sq >= 14) {
if (Rs < 41) {
return 4;
}
else if (Rs >= 41) {
if (Rs < 229) {
return 1;
}
else if (Rs >= 229) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 304) {
if (ssDCval < 51) {
if (mcTcor < 93) {
if (posBalance < 48) {
if (diffAFD < 3) {
if (Rs < 9) {
return 8;
}
else if (Rs >= 9) {
if (tcor < 65) {
return 1;
}
else if (tcor >= 65) {
if (posBalance < 39) {
if (ssDCval < 47) {
if (tcor < 76) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (posBalance < 33) {
if (Cs < 143) {
return 2;
}
else if (Cs >= 143) {
return 1;
}
}
else if (posBalance >= 33) {
return 8;
}
}
}
else if (tcor >= 76) {
return 1;
}
}
else if (ssDCval >= 47) {
return 8;
}
}
else if (posBalance >= 39) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
if (AFD < 7) {
if (posBalance < 42) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
return 4;
}
}
else if (posBalance >= 42) {
if (RsDiff < 1) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (AFD >= 7) {
if (Rs < 59) {
return 4;
}
else if (Rs >= 59) {
return 1;
}
}
}
}
}
}
}
else if (diffAFD >= 3) {
if (diffMVdiffVal < 26) {
if (tcor < 68) {
if (Rs < 47) {
return 4;
}
else if (Rs >= 47) {
return 1;
}
}
else if (tcor >= 68) {
if (Rs < 118) {
return 8;
}
else if (Rs >= 118) {
return 2;
}
}
}
else if (diffMVdiffVal >= 26) {
if (MVDiff < 602) {
return 2;
}
else if (MVDiff >= 602) {
return 8;
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 39) {
if (Rs < 105) {
if (mu_mv_mag_sq < 44) {
if (MVDiff < 203) {
return 1;
}
else if (MVDiff >= 203) {
return 8;
}
}
else if (mu_mv_mag_sq >= 44) {
if (Rs < 71) {
return 8;
}
else if (Rs >= 71) {
return 4;
}
}
}
else if (Rs >= 105) {
if (tcor < 58) {
if (mu_mv_mag_sq < 30) {
if (Cs < 198) {
return 8;
}
else if (Cs >= 198) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (posBalance < 68) {
return 1;
}
else if (posBalance >= 68) {
return 8;
}
}
}
else if (tcor >= 58) {
if (RsCsDiff < 1) {
return 8;
}
else if (RsCsDiff >= 1) {
if (Cs < 115) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (Cs >= 115) {
if (TSC < 4) {
if (MVDiff < 17) {
if (RsDiff < 1) {
if (Rs < 212) {
return 8;
}
else if (Rs >= 212) {
return 2;
}
}
else if (RsDiff >= 1) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (MVDiff >= 17) {
return 8;
}
}
else if (TSC >= 4) {
if (tcor < 67) {
if (Rs < 260) {
return 2;
}
else if (Rs >= 260) {
return 8;
}
}
else if (tcor >= 67) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 39) {
if (mu_mv_mag_sq < 105) {
if (MVDiff < 15) {
if (Rs < 78) {
if (Cs < 66) {
return 8;
}
else if (Cs >= 66) {
return 1;
}
}
else if (Rs >= 78) {
if (refDCval < 45) {
if (Cs < 322) {
return 8;
}
else if (Cs >= 322) {
return 2;
}
}
else if (refDCval >= 45) {
return 4;
}
}
}
else if (MVDiff >= 15) {
if (mu_mv_mag_sq < 23) {
if (negBalance < 42) {
if (Cs < 365) {
return 2;
}
else if (Cs >= 365) {
return 8;
}
}
else if (negBalance >= 42) {
if (Cs < 288) {
if (tcor < 69) {
if (mu_mv_mag_sq < 14) {
if (MVDiff < 99) {
return 4;
}
else if (MVDiff >= 99) {
if (Rs < 42) {
return 4;
}
else if (Rs >= 42) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (Rs < 237) {
if (Rs < 191) {
return 1;
}
else if (Rs >= 191) {
return 4;
}
}
else if (Rs >= 237) {
return 8;
}
}
}
else if (tcor >= 69) {
if (ssDCval < 44) {
return 1;
}
else if (ssDCval >= 44) {
if (Rs < 233) {
return 2;
}
else if (Rs >= 233) {
return 4;
}
}
}
}
else if (Cs >= 288) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (RsDiff < 1) {
if (ssDCval < 43) {
if (negBalance < 68) {
return 8;
}
else if (negBalance >= 68) {
return 2;
}
}
else if (ssDCval >= 43) {
if (Rs < 44) {
return 2;
}
else if (Rs >= 44) {
return 1;
}
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 69) {
if (tcor < 70) {
if (tcor < 34) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
if (Rs < 154) {
return 4;
}
else if (Rs >= 154) {
return 8;
}
}
}
else if (tcor >= 34) {
if (Cs < 516) {
return 8;
}
else if (Cs >= 516) {
return 1;
}
}
}
else if (tcor >= 70) {
return 1;
}
}
else if (mu_mv_mag_sq >= 69) {
if (Cs < 220) {
return 4;
}
else if (Cs >= 220) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
return 4;
}
}
}
}
else if (mcTcor >= 93) {
if (posBalance < 25) {
if (refDCval < 42) {
if (negBalance < 37) {
if (ssDCval < 37) {
if (Rs < 225) {
return 4;
}
else if (Rs >= 225) {
return 1;
}
}
else if (ssDCval >= 37) {
return 1;
}
}
else if (negBalance >= 37) {
if (MVDiff < 73) {
if (Rs < 17) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
return 8;
}
}
else if (Rs >= 17) {
return 8;
}
}
else if (MVDiff >= 73) {
return 4;
}
}
}
else if (refDCval >= 42) {
if (SC < 376) {
if (posBalance < 4) {
if (MVDiff < 54) {
if (Rs < 10) {
return 2;
}
else if (Rs >= 10) {
return 4;
}
}
else if (MVDiff >= 54) {
return 8;
}
}
else if (posBalance >= 4) {
if (AFD < 7) {
if (mcTcor < 105) {
return 8;
}
else if (mcTcor >= 105) {
return 4;
}
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (SC >= 376) {
return 4;
}
}
}
else if (posBalance >= 25) {
if (mu_mv_mag_sq < 67) {
if (posBalance < 69) {
if (Cs < 8) {
if (negBalance < 38) {
return 2;
}
else if (negBalance >= 38) {
if (Rs < 87) {
return 8;
}
else if (Rs >= 87) {
return 1;
}
}
}
else if (Cs >= 8) {
if (mcTcor < 97) {
if (negBalance < 21) {
if (Rs < 97) {
return 1;
}
else if (Rs >= 97) {
return 8;
}
}
else if (negBalance >= 21) {
if (AFD < 8) {
if (Cs < 136) {
if (tcor < 80) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (Rs < 82) {
return 1;
}
else if (Rs >= 82) {
return 2;
}
}
}
else if (tcor >= 80) {
if (diffMVdiffVal < -22) {
return 2;
}
else if (diffMVdiffVal >= -22) {
return 1;
}
}
}
else if (Cs >= 136) {
if (Rs < 268) {
if (Rs < 148) {
if (diffMVdiffVal < -8) {
return 4;
}
else if (diffMVdiffVal >= -8) {
if (MVDiff < 18) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 4;
}
}
else if (MVDiff >= 18) {
return 1;
}
}
}
else if (Rs >= 148) {
if (negBalance < 28) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
return 4;
}
}
else if (negBalance >= 28) {
if (Cs < 359) {
return 2;
}
else if (Cs >= 359) {
if (Rs < 223) {
return 1;
}
else if (Rs >= 223) {
return 2;
}
}
}
}
}
else if (Rs >= 268) {
if (refDCval < 37) {
return 1;
}
else if (refDCval >= 37) {
return 4;
}
}
}
}
else if (AFD >= 8) {
if (diffAFD < 1) {
if (TSC < 6) {
if (SC < 686) {
return 1;
}
else if (SC >= 686) {
return 8;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (diffAFD >= 1) {
if (mu_mv_mag_sq < 23) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 4;
}
}
else if (mu_mv_mag_sq >= 23) {
return 1;
}
}
}
}
}
else if (mcTcor >= 97) {
if (tcor < 92) {
if (Cs < 74) {
return 1;
}
else if (Cs >= 74) {
if (Rs < 57) {
return 1;
}
else if (Rs >= 57) {
if (SC < 292) {
return 4;
}
else if (SC >= 292) {
if (refDCval < 45) {
return 1;
}
else if (refDCval >= 45) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
}
}
}
else if (tcor >= 92) {
if (Rs < 276) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 2;
}
}
else if (Rs >= 276) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 69) {
return 2;
}
}
else if (mu_mv_mag_sq >= 67) {
if (SC < 543) {
if (negBalance < 44) {
if (MVDiff < 5) {
if (TSCindex < 1) {
return 4;
}
else if (TSCindex >= 1) {
if (tcor < 90) {
if (Cs < 144) {
if (Rs < 221) {
return 4;
}
else if (Rs >= 221) {
return 1;
}
}
else if (Cs >= 144) {
return 1;
}
}
else if (tcor >= 90) {
if (posBalance < 37) {
if (negBalance < 16) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 1;
}
}
else if (negBalance >= 16) {
if (ssDCval < 35) {
return 1;
}
else if (ssDCval >= 35) {
if (Cs < 125) {
if (ssDCval < 43) {
return 1;
}
else if (ssDCval >= 43) {
return 8;
}
}
else if (Cs >= 125) {
if (mu_mv_mag_sq < 95) {
if (Rs < 191) {
return 4;
}
else if (Rs >= 191) {
return 8;
}
}
else if (mu_mv_mag_sq >= 95) {
return 8;
}
}
}
}
}
else if (posBalance >= 37) {
return 8;
}
}
}
}
else if (MVDiff >= 5) {
if (mcTcor < 98) {
if (RsDiff < 1) {
if (negBalance < 20) {
if (Rs < 165) {
return 1;
}
else if (Rs >= 165) {
return 8;
}
}
else if (negBalance >= 20) {
if (Rs < 265) {
if (ssDCval < 35) {
return 1;
}
else if (ssDCval >= 35) {
return 8;
}
}
else if (Rs >= 265) {
return 4;
}
}
}
else if (RsDiff >= 1) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (mcTcor >= 98) {
if (ssDCval < 35) {
return 2;
}
else if (ssDCval >= 35) {
return 1;
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 8) {
return 2;
}
else if (Rs >= 8) {
return 4;
}
}
}
else if (SC >= 543) {
return 1;
}
}
}
}
}
else if (ssDCval >= 51) {
if (ssDCval < 70) {
if (TSCindex < 2) {
if (refDCval < 62) {
if (posBalance < 33) {
if (Rs < 32) {
return 2;
}
else if (Rs >= 32) {
if (Rs < 197) {
return 8;
}
else if (Rs >= 197) {
return 4;
}
}
}
else if (posBalance >= 33) {
if (MVDiff < 7) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (tcor < 99) {
if (ssDCval < 58) {
if (Cs < 182) {
if (Rs < 137) {
return 4;
}
else if (Rs >= 137) {
return 1;
}
}
else if (Cs >= 182) {
return 4;
}
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (tcor >= 99) {
if (Cs < 190) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Cs < 186) {
return 1;
}
else if (Cs >= 186) {
if (Rs < 145) {
return 4;
}
else if (Rs >= 145) {
return 1;
}
}
}
}
else if (Cs >= 190) {
if (Rs < 259) {
return 4;
}
else if (Rs >= 259) {
return 1;
}
}
}
}
}
else if (MVDiff >= 7) {
if (MVDiff < 42) {
return 1;
}
else if (MVDiff >= 42) {
return 8;
}
}
}
}
else if (refDCval >= 62) {
if (SC < 379) {
if (ssDCval < 66) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
return 2;
}
}
}
else if (ssDCval >= 66) {
return 2;
}
}
else if (SC >= 379) {
if (mu_mv_mag_sq < 103) {
if (Rs < 182) {
return 2;
}
else if (Rs >= 182) {
return 4;
}
}
else if (mu_mv_mag_sq >= 103) {
if (Cs < 375) {
return 4;
}
else if (Cs >= 375) {
if (Cs < 446) {
return 1;
}
else if (Cs >= 446) {
return 4;
}
}
}
}
}
}
else if (TSCindex >= 2) {
if (RsCsDiff < 1) {
if (SC < 345) {
if (tcor < 100) {
if (MVDiff < 2) {
if (refDCval < 64) {
if (tcor < 96) {
if (MVDiff < 1) {
if (gchDC < 1) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (MVDiff >= 1) {
if (diffMVdiffVal < -24) {
return 8;
}
else if (diffMVdiffVal >= -24) {
return 1;
}
}
}
else if (tcor >= 96) {
return 1;
}
}
else if (refDCval >= 64) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (Rs < 82) {
return 1;
}
else if (Rs >= 82) {
return 2;
}
}
}
}
else if (MVDiff >= 2) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 6) {
if (Rs < 51) {
return 1;
}
else if (Rs >= 51) {
return 2;
}
}
else if (mu_mv_mag_sq >= 6) {
if (MVDiff < 52) {
if (Rs < 69) {
return 2;
}
else if (Rs >= 69) {
return 8;
}
}
else if (MVDiff >= 52) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (gchDC < 1) {
if (Cs < 172) {
if (SC < 244) {
return 1;
}
else if (SC >= 244) {
if (ssDCval < 66) {
if (Cs < 150) {
if (tcor < 99) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (MVDiff < 11) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (MVDiff >= 11) {
return 4;
}
}
}
else if (tcor >= 99) {
return 1;
}
}
else if (Cs >= 150) {
return 1;
}
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (Cs >= 172) {
if (Rs < 142) {
if (ssDCval < 62) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (Rs >= 142) {
if (Cs < 195) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 195) {
return 4;
}
}
}
}
else if (gchDC >= 1) {
if (tcor < 93) {
if (Cs < 32) {
return 1;
}
else if (Cs >= 32) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 4;
}
}
}
else if (tcor >= 93) {
if (MVDiff < 26) {
if (negBalance < 45) {
if (negBalance < 21) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (negBalance >= 21) {
if (Rs < 183) {
return 1;
}
else if (Rs >= 183) {
return 4;
}
}
}
else if (negBalance >= 45) {
if (Rs < 98) {
return 8;
}
else if (Rs >= 98) {
return 2;
}
}
}
else if (MVDiff >= 26) {
return 8;
}
}
}
}
}
}
else if (tcor >= 100) {
if (SC < 151) {
if (Cs < 48) {
return 2;
}
else if (Cs >= 48) {
return 4;
}
}
else if (SC >= 151) {
return 8;
}
}
}
else if (SC >= 345) {
if (negBalance < 31) {
if (refDCval < 60) {
if (MVDiff < 1) {
if (Cs < 317) {
return 1;
}
else if (Cs >= 317) {
if (SC < 625) {
if (Rs < 244) {
return 4;
}
else if (Rs >= 244) {
return 1;
}
}
else if (SC >= 625) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 189) {
return 1;
}
else if (Cs >= 189) {
return 4;
}
}
}
else if (refDCval >= 60) {
if (TSC < 2) {
if (mu_mv_mag_sq < 82) {
if (Rs < 160) {
return 2;
}
else if (Rs >= 160) {
return 4;
}
}
else if (mu_mv_mag_sq >= 82) {
if (SC < 393) {
if (mu_mv_mag_sq < 123) {
if (Rs < 197) {
return 1;
}
else if (Rs >= 197) {
return 4;
}
}
else if (mu_mv_mag_sq >= 123) {
return 4;
}
}
else if (SC >= 393) {
if (mu_mv_mag_sq < 123) {
if (negBalance < 20) {
if (Rs < 186) {
return 2;
}
else if (Rs >= 186) {
return 4;
}
}
else if (negBalance >= 20) {
return 4;
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 170) {
return 4;
}
else if (Rs >= 170) {
if (Rs < 220) {
if (Cs < 375) {
if (Rs < 174) {
return 1;
}
else if (Rs >= 174) {
return 4;
}
}
else if (Cs >= 375) {
return 1;
}
}
else if (Rs >= 220) {
return 1;
}
}
}
}
}
}
else if (TSC >= 2) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 63) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (Rs < 267) {
return 4;
}
else if (Rs >= 267) {
return 2;
}
}
}
else if (ssDCval >= 63) {
if (mu_mv_mag_sq < 100) {
if (tcor < 97) {
if (negBalance < 30) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
return 1;
}
}
else if (negBalance >= 30) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (tcor >= 97) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (Rs < 163) {
return 4;
}
else if (Rs >= 163) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 339) {
if (MVDiff < 5) {
if (Rs < 128) {
return 1;
}
else if (Rs >= 128) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
if (mu_mv_mag_sq < 104) {
return 1;
}
else if (mu_mv_mag_sq >= 104) {
return 4;
}
}
}
}
else if (MVDiff >= 5) {
if (Rs < 229) {
return 1;
}
else if (Rs >= 229) {
return 4;
}
}
}
else if (Cs >= 339) {
if (Rs < 178) {
return 1;
}
else if (Rs >= 178) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
if (Rs < 184) {
return 4;
}
else if (Rs >= 184) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
return 1;
}
}
}
}
}
else if (negBalance >= 31) {
if (mu_mv_mag_sq < 89) {
if (ssDCval < 63) {
if (Cs < 215) {
if (diffRsCsdiff < -1) {
if (SC < 353) {
return 1;
}
else if (SC >= 353) {
return 4;
}
}
else if (diffRsCsdiff >= -1) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (Cs >= 215) {
return 4;
}
}
else if (ssDCval >= 63) {
if (SC < 461) {
if (TSC < 3) {
if (Rs < 195) {
return 1;
}
else if (Rs >= 195) {
return 4;
}
}
else if (TSC >= 3) {
return 4;
}
}
else if (SC >= 461) {
if (Cs < 329) {
return 1;
}
else if (Cs >= 329) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (posBalance < 40) {
if (Rs < 185) {
return 8;
}
else if (Rs >= 185) {
return 4;
}
}
else if (posBalance >= 40) {
if (Cs < 502) {
if (posBalance < 46) {
if (Rs < 205) {
return 4;
}
else if (Rs >= 205) {
return 1;
}
}
else if (posBalance >= 46) {
return 1;
}
}
else if (Cs >= 502) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (TSC < 4) {
if (mcTcor < 91) {
if (negBalance < 19) {
return 1;
}
else if (negBalance >= 19) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 189) {
if (Rs < 136) {
return 4;
}
else if (Rs >= 136) {
return 2;
}
}
else if (Rs >= 189) {
return 8;
}
}
}
}
else if (mcTcor >= 91) {
if (RsCsDiff < 5) {
if (Rs < 38) {
return 1;
}
else if (Rs >= 38) {
if (MVDiff < 4) {
if (Cs < 195) {
return 1;
}
else if (Cs >= 195) {
if (ssDCval < 69) {
if (posBalance < 57) {
if (refDCval < 52) {
if (Cs < 390) {
return 4;
}
else if (Cs >= 390) {
return 1;
}
}
else if (refDCval >= 52) {
if (SC < 514) {
if (SC < 433) {
return 4;
}
else if (SC >= 433) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
return 4;
}
}
}
else if (SC >= 514) {
if (SC < 681) {
if (mu_mv_mag_sq < 118) {
return 4;
}
else if (mu_mv_mag_sq >= 118) {
return 1;
}
}
else if (SC >= 681) {
if (Cs < 482) {
return 1;
}
else if (Cs >= 482) {
return 4;
}
}
}
}
}
else if (posBalance >= 57) {
return 1;
}
}
else if (ssDCval >= 69) {
if (Cs < 376) {
if (Cs < 312) {
return 1;
}
else if (Cs >= 312) {
if (Cs < 350) {
return 4;
}
else if (Cs >= 350) {
if (Rs < 181) {
return 1;
}
else if (Rs >= 181) {
return 4;
}
}
}
}
else if (Cs >= 376) {
return 1;
}
}
}
}
else if (MVDiff >= 4) {
if (SC < 388) {
if (posBalance < 45) {
if (Cs < 137) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (Cs >= 137) {
return 1;
}
}
else if (posBalance >= 45) {
if (MVDiff < 9) {
if (ssDCval < 64) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 171) {
if (Cs < 101) {
return 1;
}
else if (Cs >= 101) {
return 4;
}
}
else if (Rs >= 171) {
return 1;
}
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (MVDiff >= 9) {
if (MVDiff < 60) {
if (mu_mv_mag_sq < 2) {
if (Rs < 164) {
if (Rs < 81) {
return 1;
}
else if (Rs >= 81) {
return 4;
}
}
else if (Rs >= 164) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (mu_mv_mag_sq < 93) {
return 4;
}
else if (mu_mv_mag_sq >= 93) {
return 1;
}
}
}
else if (MVDiff >= 60) {
return 1;
}
}
}
}
else if (SC >= 388) {
if (negBalance < 51) {
if (refDCval < 54) {
if (Cs < 319) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
else if (Cs >= 319) {
return 4;
}
}
else if (refDCval >= 54) {
if (tcor < 93) {
if (mu_mv_mag_sq < 47) {
if (Cs < 492) {
if (diffAFD < -1) {
if (diffTSC < -1) {
if (Rs < 263) {
return 1;
}
else if (Rs >= 263) {
return 8;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (Cs >= 492) {
return 1;
}
}
else if (mu_mv_mag_sq >= 47) {
if (posBalance < 54) {
if (Rs < 227) {
return 4;
}
else if (Rs >= 227) {
return 8;
}
}
else if (posBalance >= 54) {
return 4;
}
}
}
else if (tcor >= 93) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
if (tcor < 97) {
return 4;
}
else if (tcor >= 97) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 51) {
if (Rs < 194) {
return 8;
}
else if (Rs >= 194) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 5) {
if (AFD < 9) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (SC < 455) {
return 4;
}
else if (SC >= 455) {
if (Cs < 417) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (Cs >= 417) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
return 4;
}
}
}
}
}
else if (AFD >= 9) {
if (ssDCval < 63) {
if (refDCval < 58) {
if (Cs < 248) {
return 1;
}
else if (Cs >= 248) {
return 2;
}
}
else if (refDCval >= 58) {
return 1;
}
}
else if (ssDCval >= 63) {
if (Rs < 133) {
return 1;
}
else if (Rs >= 133) {
return 4;
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 38) {
if (MVDiff < 55) {
if (AFD < 12) {
if (MVDiff < 47) {
if (mu_mv_mag_sq < 33) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (tcor < 95) {
return 1;
}
else if (tcor >= 95) {
if (Rs < 182) {
return 1;
}
else if (Rs >= 182) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 33) {
return 2;
}
}
else if (MVDiff >= 47) {
if (Rs < 253) {
return 2;
}
else if (Rs >= 253) {
if (Rs < 264) {
return 1;
}
else if (Rs >= 264) {
return 4;
}
}
}
}
else if (AFD >= 12) {
if (Rs < 172) {
if (AFD < 15) {
if (Rs < 148) {
return 4;
}
else if (Rs >= 148) {
return 8;
}
}
else if (AFD >= 15) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (Rs >= 172) {
if (refDCval < 65) {
if (Cs < 228) {
return 4;
}
else if (Cs >= 228) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
return 1;
}
}
}
else if (refDCval >= 65) {
return 4;
}
}
}
}
else if (MVDiff >= 55) {
if (AFD < 16) {
if (tcor < 75) {
if (posBalance < 62) {
if (Rs < 123) {
return 2;
}
else if (Rs >= 123) {
return 4;
}
}
else if (posBalance >= 62) {
if (diffRsCsdiff < -4) {
if (Cs < 352) {
return 2;
}
else if (Cs >= 352) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
if (SC < 425) {
return 1;
}
else if (SC >= 425) {
return 2;
}
}
}
}
else if (tcor >= 75) {
if (negBalance < 42) {
if (mu_mv_mag_sq < 11) {
if (Rs < 279) {
if (Cs < 370) {
return 2;
}
else if (Cs >= 370) {
return 8;
}
}
else if (Rs >= 279) {
return 1;
}
}
else if (mu_mv_mag_sq >= 11) {
return 1;
}
}
else if (negBalance >= 42) {
return 2;
}
}
}
else if (AFD >= 16) {
if (RsCsDiff < 11) {
if (MVDiff < 74) {
if (Rs < 180) {
return 2;
}
else if (Rs >= 180) {
return 4;
}
}
else if (MVDiff >= 74) {
if (SC < 194) {
if (Rs < 84) {
return 1;
}
else if (Rs >= 84) {
return 2;
}
}
else if (SC >= 194) {
return 1;
}
}
}
else if (RsCsDiff >= 11) {
if (SC < 661) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 8;
}
}
else if (SC >= 661) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (AFD < 8) {
if (MVDiff < 7) {
if (Rs < 218) {
return 4;
}
else if (Rs >= 218) {
return 1;
}
}
else if (MVDiff >= 7) {
if (negBalance < 46) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (negBalance >= 46) {
if (gchDC < 5) {
return 8;
}
else if (gchDC >= 5) {
return 2;
}
}
}
}
else if (AFD >= 8) {
if (mu_mv_mag_sq < 99) {
if (gchDC < 6) {
if (Rs < 107) {
return 1;
}
else if (Rs >= 107) {
return 4;
}
}
else if (gchDC >= 6) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (Cs < 64) {
return 1;
}
else if (Cs >= 64) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (diffMVdiffVal < -98) {
return 8;
}
else if (diffMVdiffVal >= -98) {
if (Rs < 191) {
return 1;
}
else if (Rs >= 191) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 70) {
if (mu_mv_mag_sq < 105) {
if (mcTcor < 97) {
if (mcTcor < 77) {
if (posBalance < 33) {
if (Cs < 4) {
if (Rs < 1) {
return 1;
}
else if (Rs >= 1) {
return 2;
}
}
else if (Cs >= 4) {
return 1;
}
}
else if (posBalance >= 33) {
if (TSC < 15) {
if (SC < 614) {
if (ssDCval < 73) {
if (Rs < 246) {
return 1;
}
else if (Rs >= 246) {
return 4;
}
}
else if (ssDCval >= 73) {
return 4;
}
}
else if (SC >= 614) {
if (Rs < 254) {
if (Rs < 239) {
return 2;
}
else if (Rs >= 239) {
return 4;
}
}
else if (Rs >= 254) {
if (Rs < 270) {
return 1;
}
else if (Rs >= 270) {
return 8;
}
}
}
}
else if (TSC >= 15) {
if (ssDCval < 91) {
if (Rs < 217) {
return 8;
}
else if (Rs >= 217) {
return 2;
}
}
else if (ssDCval >= 91) {
return 1;
}
}
}
}
else if (mcTcor >= 77) {
if (Rs < 118) {
if (ssDCval < 129) {
if (negBalance < 43) {
if (MVDiff < 48) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (MVDiff >= 48) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (Rs < 5) {
return 1;
}
else if (Rs >= 5) {
return 2;
}
}
}
}
else if (negBalance >= 43) {
if (SC < 45) {
return 1;
}
else if (SC >= 45) {
if (tcor < 88) {
if (tcor < 62) {
if (Cs < 59) {
if (SC < 123) {
return 8;
}
else if (SC >= 123) {
if (Rs < 108) {
return 2;
}
else if (Rs >= 108) {
return 8;
}
}
}
else if (Cs >= 59) {
if (Cs < 271) {
return 1;
}
else if (Cs >= 271) {
return 2;
}
}
}
else if (tcor >= 62) {
if (RsCsDiff < 12) {
if (Cs < 114) {
if (mu_mv_mag_sq < 16) {
if (Rs < 101) {
return 1;
}
else if (Rs >= 101) {
return 2;
}
}
else if (mu_mv_mag_sq >= 16) {
return 2;
}
}
else if (Cs >= 114) {
return 2;
}
}
else if (RsCsDiff >= 12) {
return 1;
}
}
}
else if (tcor >= 88) {
if (AFD < 7) {
if (Rs < 101) {
return 4;
}
else if (Rs >= 101) {
return 1;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
}
}
else if (ssDCval >= 129) {
if (Rs < 1) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (Rs >= 1) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 2;
}
}
}
}
else if (Rs >= 118) {
if (RsDiff < 2) {
if (mcTcor < 95) {
if (refDCval < 84) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (MVDiff < 29) {
if (Rs < 282) {
return 1;
}
else if (Rs >= 282) {
return 2;
}
}
else if (MVDiff >= 29) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
return 4;
}
}
}
}
}
else if (refDCval >= 84) {
if (ssDCval < 144) {
if (SC < 160) {
return 8;
}
else if (SC >= 160) {
if (MVDiff < 28) {
if (Rs < 238) {
if (tcor < 78) {
if (Rs < 149) {
if (Rs < 125) {
return 8;
}
else if (Rs >= 125) {
if (SC < 163) {
return 8;
}
else if (SC >= 163) {
return 2;
}
}
}
else if (Rs >= 149) {
if (diffMVdiffVal < -8) {
if (Rs < 183) {
return 8;
}
else if (Rs >= 183) {
return 2;
}
}
else if (diffMVdiffVal >= -8) {
if (posBalance < 57) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
else if (tcor >= 78) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (Cs < 333) {
if (Cs < 81) {
return 8;
}
else if (Cs >= 81) {
if (diffRsCsdiff < 2) {
return 2;
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
}
else if (Cs >= 333) {
return 8;
}
}
}
}
else if (Rs >= 238) {
if (mcTcor < 84) {
if (SC < 615) {
if (TSC < 5) {
if (Rs < 254) {
return 8;
}
else if (Rs >= 254) {
return 2;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (SC >= 615) {
return 8;
}
}
else if (mcTcor >= 84) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (refDCval < 85) {
return 1;
}
else if (refDCval >= 85) {
return 8;
}
}
}
}
}
else if (MVDiff >= 28) {
if (Rs < 248) {
if (TSC < 6) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (Cs < 59) {
return 2;
}
else if (Cs >= 59) {
return 8;
}
}
}
else if (TSC >= 6) {
return 1;
}
}
else if (Rs >= 248) {
return 2;
}
}
}
}
else if (ssDCval >= 144) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
return 2;
}
}
}
}
else if (mcTcor >= 95) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
if (mu_mv_mag_sq < 7) {
if (RsDiff < 1) {
if (Rs < 133) {
return 4;
}
else if (Rs >= 133) {
return 8;
}
}
else if (RsDiff >= 1) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
if (MVDiff < 7) {
if (Rs < 283) {
return 8;
}
else if (Rs >= 283) {
return 1;
}
}
else if (MVDiff >= 7) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (negBalance < 40) {
if (MVDiff < 4) {
if (Cs < 419) {
return 8;
}
else if (Cs >= 419) {
return 2;
}
}
else if (MVDiff >= 4) {
if (Cs < 297) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
return 1;
}
}
else if (Cs >= 297) {
if (Rs < 290) {
return 8;
}
else if (Rs >= 290) {
return 2;
}
}
}
}
else if (negBalance >= 40) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (TSC < 7) {
if (mu_mv_mag_sq < 61) {
if (diffMVdiffVal < 6) {
if (MVDiff < 137) {
if (tcor < 91) {
if (tcor < 85) {
return 1;
}
else if (tcor >= 85) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (tcor >= 91) {
if (Rs < 250) {
return 2;
}
else if (Rs >= 250) {
return 1;
}
}
}
else if (MVDiff >= 137) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
if (AFD < 15) {
if (Cs < 68) {
return 1;
}
else if (Cs >= 68) {
if (Cs < 395) {
return 8;
}
else if (Cs >= 395) {
return 1;
}
}
}
else if (AFD >= 15) {
if (Rs < 201) {
return 4;
}
else if (Rs >= 201) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (mcTcor < 81) {
return 2;
}
else if (mcTcor >= 81) {
if (Cs < 12) {
if (Rs < 149) {
return 2;
}
else if (Rs >= 149) {
return 8;
}
}
else if (Cs >= 12) {
return 8;
}
}
}
}
else if (TSC >= 7) {
if (AFD < 24) {
if (tcor < 70) {
if (diffMVdiffVal < 12) {
if (MVDiff < 230) {
return 2;
}
else if (MVDiff >= 230) {
return 1;
}
}
else if (diffMVdiffVal >= 12) {
if (MVDiff < 81) {
return 8;
}
else if (MVDiff >= 81) {
if (SC < 410) {
return 8;
}
else if (SC >= 410) {
return 2;
}
}
}
}
else if (tcor >= 70) {
if (Rs < 284) {
if (ssDCval < 114) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
else if (ssDCval >= 114) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (Rs >= 284) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (Cs < 368) {
return 4;
}
else if (Cs >= 368) {
return 2;
}
}
}
}
}
else if (AFD >= 24) {
if (Cs < 79) {
return 4;
}
else if (Cs >= 79) {
if (diffRsCsdiff < -22) {
return 8;
}
else if (diffRsCsdiff >= -22) {
return 1;
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 19) {
if (refDCval < 92) {
if (refDCval < 77) {
if (refDCval < 74) {
if (Cs < 160) {
return 2;
}
else if (Cs >= 160) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (refDCval >= 74) {
return 1;
}
}
else if (refDCval >= 77) {
if (Cs < 289) {
if (Cs < 132) {
return 8;
}
else if (Cs >= 132) {
return 2;
}
}
else if (Cs >= 289) {
return 4;
}
}
}
else if (refDCval >= 92) {
if (Rs < 8) {
return 1;
}
else if (Rs >= 8) {
return 4;
}
}
}
else if (negBalance >= 19) {
if (ssDCval < 106) {
if (Cs < 258) {
if (Cs < 40) {
if (AFD < 12) {
if (refDCval < 98) {
return 1;
}
else if (refDCval >= 98) {
if (Rs < 55) {
return 1;
}
else if (Rs >= 55) {
return 2;
}
}
}
else if (AFD >= 12) {
return 2;
}
}
else if (Cs >= 40) {
if (Rs < 91) {
if (ssDCval < 94) {
if (refDCval < 83) {
if (mu_mv_mag_sq < 9) {
if (Rs < 44) {
return 1;
}
else if (Rs >= 44) {
return 4;
}
}
else if (mu_mv_mag_sq >= 9) {
if (SC < 219) {
return 1;
}
else if (SC >= 219) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 2;
}
}
}
}
else if (refDCval >= 83) {
if (Cs < 183) {
return 1;
}
else if (Cs >= 183) {
return 2;
}
}
}
else if (ssDCval >= 94) {
if (tcor < 98) {
if (Rs < 10) {
return 2;
}
else if (Rs >= 10) {
if (MVDiff < 18) {
if (Rs < 53) {
return 4;
}
else if (Rs >= 53) {
return 2;
}
}
else if (MVDiff >= 18) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (tcor >= 98) {
if (SC < 322) {
return 4;
}
else if (SC >= 322) {
return 1;
}
}
}
}
else if (Rs >= 91) {
if (AFD < 2) {
if (ssDCval < 86) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Rs < 159) {
if (SC < 227) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (SC >= 227) {
return 1;
}
}
else if (Rs >= 159) {
return 2;
}
}
}
else if (ssDCval >= 86) {
return 1;
}
}
else if (AFD >= 2) {
if (refDCval < 93) {
if (MVDiff < 4) {
if (Cs < 143) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (Cs >= 143) {
if (TSC < 3) {
if (Rs < 229) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 83) {
if (Cs < 180) {
return 8;
}
else if (Cs >= 180) {
return 2;
}
}
else if (mu_mv_mag_sq >= 83) {
return 2;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (Rs >= 229) {
if (Rs < 282) {
return 1;
}
else if (Rs >= 282) {
return 8;
}
}
}
else if (TSC >= 3) {
if (ssDCval < 75) {
if (Rs < 189) {
return 4;
}
else if (Rs >= 189) {
return 1;
}
}
else if (ssDCval >= 75) {
return 2;
}
}
}
}
else if (MVDiff >= 4) {
if (gchDC < 1) {
if (AFD < 7) {
if (diffMVdiffVal < 5) {
if (SC < 240) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 1;
}
}
else if (SC >= 240) {
return 1;
}
}
else if (diffMVdiffVal >= 5) {
if (Cs < 207) {
return 1;
}
else if (Cs >= 207) {
return 4;
}
}
}
else if (AFD >= 7) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 2;
}
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 41) {
if (gchDC < 5) {
if (Cs < 212) {
if (Rs < 101) {
if (Rs < 97) {
return 1;
}
else if (Rs >= 97) {
return 4;
}
}
else if (Rs >= 101) {
return 1;
}
}
else if (Cs >= 212) {
return 4;
}
}
else if (gchDC >= 5) {
if (Rs < 152) {
return 8;
}
else if (Rs >= 152) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (Rs < 143) {
if (Cs < 195) {
if (SC < 264) {
if (Rs < 106) {
return 1;
}
else if (Rs >= 106) {
return 2;
}
}
else if (SC >= 264) {
return 1;
}
}
else if (Cs >= 195) {
return 2;
}
}
else if (Rs >= 143) {
if (MVDiff < 18) {
if (diffAFD < 1) {
if (Cs < 190) {
return 1;
}
else if (Cs >= 190) {
return 8;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (MVDiff >= 18) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 93) {
if (SC < 306) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (SC >= 306) {
if (refDCval < 103) {
if (TSC < 5) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (Rs < 119) {
return 4;
}
else if (Rs >= 119) {
if (Rs < 219) {
return 2;
}
else if (Rs >= 219) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
}
}
}
else if (TSC >= 5) {
if (Rs < 198) {
if (Cs < 193) {
return 4;
}
else if (Cs >= 193) {
return 8;
}
}
else if (Rs >= 198) {
if (Cs < 102) {
return 8;
}
else if (Cs >= 102) {
return 2;
}
}
}
}
else if (refDCval >= 103) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 258) {
if (Cs < 413) {
if (ssDCval < 75) {
if (Rs < 136) {
if (refDCval < 72) {
if (Rs < 68) {
return 2;
}
else if (Rs >= 68) {
return 4;
}
}
else if (refDCval >= 72) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (Rs >= 136) {
if (negBalance < 43) {
if (mcTcor < 98) {
if (Rs < 158) {
return 4;
}
else if (Rs >= 158) {
return 1;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (negBalance >= 43) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 1;
}
}
}
}
else if (ssDCval >= 75) {
if (TSC < 3) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
else if (TSC >= 3) {
if (tcor < 91) {
if (diffRsCsdiff < -4) {
if (Rs < 191) {
return 2;
}
else if (Rs >= 191) {
return 4;
}
}
else if (diffRsCsdiff >= -4) {
if (diffMVdiffVal < 2) {
if (Cs < 322) {
return 1;
}
else if (Cs >= 322) {
return 8;
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 237) {
return 8;
}
else if (Rs >= 237) {
return 1;
}
}
}
}
else if (tcor >= 91) {
if (ssDCval < 79) {
if (Rs < 198) {
return 4;
}
else if (Rs >= 198) {
return 2;
}
}
else if (ssDCval >= 79) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
return 1;
}
}
}
}
}
}
else if (Cs >= 413) {
if (refDCval < 83) {
return 1;
}
else if (refDCval >= 83) {
if (Rs < 270) {
if (mu_mv_mag_sq < 8) {
return 2;
}
else if (mu_mv_mag_sq >= 8) {
if (Rs < 226) {
return 1;
}
else if (Rs >= 226) {
if (Cs < 483) {
return 8;
}
else if (Cs >= 483) {
return 2;
}
}
}
}
else if (Rs >= 270) {
return 1;
}
}
}
}
}
else if (ssDCval >= 106) {
if (AFD < 3) {
if (ssDCval < 136) {
if (TSC < 2) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (TSC >= 2) {
return 2;
}
}
else if (ssDCval >= 136) {
return 2;
}
}
else if (AFD >= 3) {
if (posBalance < 10) {
return 8;
}
else if (posBalance >= 10) {
if (ssDCval < 173) {
if (Cs < 2) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 8;
}
}
else if (Cs >= 2) {
if (RsCsDiff < 1) {
return 2;
}
else if (RsCsDiff >= 1) {
if (Rs < 80) {
if (Cs < 169) {
return 8;
}
else if (Cs >= 169) {
return 1;
}
}
else if (Rs >= 80) {
if (tcor < 80) {
return 1;
}
else if (tcor >= 80) {
if (tcor < 96) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (negBalance < 41) {
return 2;
}
else if (negBalance >= 41) {
if (Rs < 238) {
return 1;
}
else if (Rs >= 238) {
return 8;
}
}
}
}
else if (tcor >= 96) {
if (mu_mv_mag_sq < 42) {
return 2;
}
else if (mu_mv_mag_sq >= 42) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 173) {
if (Cs < 324) {
return 2;
}
else if (Cs >= 324) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Cs < 121) {
if (ssDCval < 76) {
if (Rs < 101) {
if (Cs < 26) {
return 8;
}
else if (Cs >= 26) {
return 4;
}
}
else if (Rs >= 101) {
if (negBalance < 17) {
return 2;
}
else if (negBalance >= 17) {
return 8;
}
}
}
else if (ssDCval >= 76) {
if (SC < 221) {
return 1;
}
else if (SC >= 221) {
if (Cs < 77) {
if (Cs < 37) {
return 1;
}
else if (Cs >= 37) {
return 2;
}
}
else if (Cs >= 77) {
return 8;
}
}
}
}
else if (Cs >= 121) {
if (negBalance < 25) {
if (Cs < 266) {
if (ssDCval < 81) {
if (negBalance < 23) {
if (diffAFD < -1) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (diffAFD >= -1) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
if (Rs < 232) {
return 2;
}
else if (Rs >= 232) {
return 8;
}
}
}
}
else if (negBalance >= 23) {
if (Cs < 212) {
if (Cs < 129) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
return 8;
}
}
else if (Cs >= 129) {
return 2;
}
}
else if (Cs >= 212) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
return 2;
}
}
}
}
else if (ssDCval >= 81) {
if (Rs < 204) {
return 2;
}
else if (Rs >= 204) {
if (Rs < 260) {
return 8;
}
else if (Rs >= 260) {
return 1;
}
}
}
}
else if (Cs >= 266) {
if (Cs < 300) {
return 1;
}
else if (Cs >= 300) {
return 4;
}
}
}
else if (negBalance >= 25) {
if (Rs < 178) {
if (MVDiff < 4) {
if (Rs < 140) {
return 1;
}
else if (Rs >= 140) {
return 2;
}
}
else if (MVDiff >= 4) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (SC < 303) {
return 8;
}
else if (SC >= 303) {
return 2;
}
}
}
}
else if (Rs >= 178) {
if (ssDCval < 78) {
if (tcor < 98) {
if (Cs < 263) {
return 2;
}
else if (Cs >= 263) {
return 8;
}
}
else if (tcor >= 98) {
return 1;
}
}
else if (ssDCval >= 78) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 304) {
if (RsDiff < 4) {
if (refDCval < 73) {
if (mcTcor < 89) {
if (ssDCval < 41) {
if (posBalance < 50) {
if (Cs < 307) {
if (RsCsDiff < 2) {
if (Rs < 443) {
return 4;
}
else if (Rs >= 443) {
return 8;
}
}
else if (RsCsDiff >= 2) {
if (ssDCval < 40) {
return 1;
}
else if (ssDCval >= 40) {
return 2;
}
}
}
else if (Cs >= 307) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (negBalance < 37) {
if (SC < 561) {
return 8;
}
else if (SC >= 561) {
if (posBalance < 66) {
return 1;
}
else if (posBalance >= 66) {
return 8;
}
}
}
else if (negBalance >= 37) {
if (CsDiff < 3) {
if (Cs < 83) {
return 1;
}
else if (Cs >= 83) {
if (refDCval < 37) {
if (tcor < 70) {
return 2;
}
else if (tcor >= 70) {
return 8;
}
}
else if (refDCval >= 37) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 331) {
return 2;
}
else if (Rs >= 331) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 41) {
if (mu_mv_mag_sq < 101) {
if (Cs < 155) {
if (Rs < 323) {
return 8;
}
else if (Rs >= 323) {
return 1;
}
}
else if (Cs >= 155) {
if (negBalance < 48) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (negBalance < 39) {
if (diffMVdiffVal < 2) {
if (tcor < 76) {
if (diffRsCsdiff < 7) {
return 2;
}
else if (diffRsCsdiff >= 7) {
return 8;
}
}
else if (tcor >= 76) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
if (Rs < 453) {
return 8;
}
else if (Rs >= 453) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (refDCval < 60) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (refDCval >= 60) {
return 8;
}
}
}
else if (negBalance >= 39) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (mu_mv_mag_sq < 52) {
if (Cs < 253) {
return 2;
}
else if (Cs >= 253) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (SC < 933) {
return 4;
}
else if (SC >= 933) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 48) {
if (MVDiff < 59) {
if (TSC < 6) {
if (RsCsDiff < 11) {
if (diffMVdiffVal < -1) {
if (Rs < 540) {
return 8;
}
else if (Rs >= 540) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (RsCsDiff >= 11) {
return 4;
}
}
else if (TSC >= 6) {
if (ssDCval < 55) {
if (negBalance < 50) {
if (Rs < 469) {
return 4;
}
else if (Rs >= 469) {
return 8;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (ssDCval >= 55) {
return 4;
}
}
}
else if (MVDiff >= 59) {
if (mcTcor < 81) {
if (diffAFD < 1) {
if (Rs < 494) {
return 4;
}
else if (Rs >= 494) {
return 2;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (mcTcor >= 81) {
if (Rs < 637) {
return 2;
}
else if (Rs >= 637) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (mcTcor < 83) {
return 8;
}
else if (mcTcor >= 83) {
return 4;
}
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1279) {
if (tcor < 101) {
if (refDCval < 58) {
if (mu_mv_mag_sq < 102) {
if (TSCindex < 2) {
if (posBalance < 33) {
if (Cs < 342) {
if (posBalance < 31) {
return 4;
}
else if (posBalance >= 31) {
return 1;
}
}
else if (Cs >= 342) {
return 8;
}
}
else if (posBalance >= 33) {
return 4;
}
}
else if (TSCindex >= 2) {
if (Cs < 179) {
if (negBalance < 39) {
if (TSC < 3) {
if (Rs < 393) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
return 4;
}
}
}
else if (Rs >= 393) {
if (refDCval < 38) {
if (Rs < 420) {
return 1;
}
else if (Rs >= 420) {
return 2;
}
}
else if (refDCval >= 38) {
return 1;
}
}
}
else if (TSC >= 3) {
if (posBalance < 54) {
if (AFD < 9) {
if (Cs < 88) {
return 8;
}
else if (Cs >= 88) {
if (negBalance < 35) {
if (CsDiff < 1) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (negBalance >= 35) {
return 8;
}
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (posBalance >= 54) {
return 1;
}
}
}
else if (negBalance >= 39) {
if (diffMVdiffVal < -84) {
return 8;
}
else if (diffMVdiffVal >= -84) {
if (TSC < 7) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
if (Rs < 514) {
return 1;
}
else if (Rs >= 514) {
return 8;
}
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
else if (Cs >= 179) {
if (negBalance < 38) {
if (posBalance < 34) {
if (mu_mv_mag_sq < 82) {
if (Rs < 437) {
return 4;
}
else if (Rs >= 437) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (Cs < 450) {
if (Cs < 404) {
return 1;
}
else if (Cs >= 404) {
if (CsDiff < 1) {
if (Rs < 481) {
return 8;
}
else if (Rs >= 481) {
return 1;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
}
else if (Cs >= 450) {
return 2;
}
}
}
else if (posBalance >= 34) {
if (Cs < 391) {
if (Rs < 524) {
if (gchDC < 3) {
if (refDCval < 46) {
if (mu_mv_mag_sq < 70) {
if (posBalance < 55) {
if (ssDCval < 37) {
if (Cs < 273) {
return 4;
}
else if (Cs >= 273) {
return 1;
}
}
else if (ssDCval >= 37) {
return 4;
}
}
else if (posBalance >= 55) {
if (SC < 544) {
return 8;
}
else if (SC >= 544) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (SC < 521) {
return 8;
}
else if (SC >= 521) {
return 1;
}
}
}
else if (refDCval >= 46) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (ssDCval < 56) {
if (posBalance < 59) {
if (Rs < 315) {
return 1;
}
else if (Rs >= 315) {
if (Rs < 477) {
if (MVDiff < 7) {
if (Rs < 453) {
if (Rs < 380) {
return 4;
}
else if (Rs >= 380) {
return 1;
}
}
else if (Rs >= 453) {
return 4;
}
}
else if (MVDiff >= 7) {
if (diffMVdiffVal < 7) {
return 4;
}
else if (diffMVdiffVal >= 7) {
if (Rs < 419) {
return 4;
}
else if (Rs >= 419) {
return 1;
}
}
}
}
else if (Rs >= 477) {
return 4;
}
}
}
else if (posBalance >= 59) {
return 1;
}
}
else if (ssDCval >= 56) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
return 4;
}
}
}
}
}
else if (gchDC >= 3) {
if (Rs < 356) {
return 2;
}
else if (Rs >= 356) {
return 8;
}
}
}
else if (Rs >= 524) {
if (MVDiff < 24) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (refDCval < 51) {
if (Rs < 551) {
return 8;
}
else if (Rs >= 551) {
return 1;
}
}
else if (refDCval >= 51) {
if (Rs < 837) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (Rs < 702) {
return 8;
}
else if (Rs >= 702) {
return 1;
}
}
}
else if (Rs >= 837) {
return 1;
}
}
}
}
else if (MVDiff >= 24) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
}
else if (Cs >= 391) {
if (RsCsDiff < 4) {
if (refDCval < 54) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (refDCval < 46) {
if (MVDiff < 7) {
if (Rs < 445) {
return 1;
}
else if (Rs >= 445) {
return 4;
}
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (refDCval >= 46) {
return 1;
}
}
}
else if (refDCval >= 54) {
if (mcTcor < 99) {
if (Cs < 470) {
return 1;
}
else if (Cs >= 470) {
if (Rs < 449) {
return 8;
}
else if (Rs >= 449) {
if (Rs < 510) {
return 1;
}
else if (Rs >= 510) {
return 4;
}
}
}
}
else if (mcTcor >= 99) {
if (Cs < 456) {
return 8;
}
else if (Cs >= 456) {
return 4;
}
}
}
}
else if (RsCsDiff >= 4) {
if (refDCval < 45) {
if (Cs < 485) {
if (ssDCval < 46) {
if (Rs < 434) {
return 8;
}
else if (Rs >= 434) {
return 1;
}
}
else if (ssDCval >= 46) {
return 8;
}
}
else if (Cs >= 485) {
if (Rs < 720) {
return 8;
}
else if (Rs >= 720) {
return 2;
}
}
}
else if (refDCval >= 45) {
if (MVDiff < 10) {
if (posBalance < 43) {
if (Cs < 472) {
if (SC < 1324) {
return 2;
}
else if (SC >= 1324) {
return 4;
}
}
else if (Cs >= 472) {
return 4;
}
}
else if (posBalance >= 43) {
if (diffMVdiffVal < -2) {
if (Rs < 503) {
return 1;
}
else if (Rs >= 503) {
if (Rs < 576) {
return 4;
}
else if (Rs >= 576) {
return 8;
}
}
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (mu_mv_mag_sq < 64) {
if (Rs < 604) {
if (diffMVdiffVal < 4) {
return 4;
}
else if (diffMVdiffVal >= 4) {
return 1;
}
}
else if (Rs >= 604) {
if (Cs < 479) {
return 1;
}
else if (Cs >= 479) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (AFD < 5) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 2;
}
}
else if (AFD >= 5) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 38) {
if (Rs < 404) {
if (Rs < 343) {
if (tcor < 86) {
if (ssDCval < 42) {
return 1;
}
else if (ssDCval >= 42) {
return 4;
}
}
else if (tcor >= 86) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (Rs < 325) {
return 1;
}
else if (Rs >= 325) {
return 8;
}
}
}
}
}
else if (Rs >= 343) {
if (MVDiff < 36) {
if (Rs < 345) {
return 2;
}
else if (Rs >= 345) {
return 1;
}
}
else if (MVDiff >= 36) {
return 2;
}
}
}
else if (Rs >= 404) {
if (mu_mv_mag_sq < 40) {
if (MVDiff < 45) {
if (MVDiff < 4) {
if (RsDiff < 2) {
if (Rs < 473) {
return 8;
}
else if (Rs >= 473) {
return 1;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (MVDiff >= 4) {
if (TSC < 4) {
if (mu_mv_mag_sq < 36) {
if (tcor < 65) {
return 8;
}
else if (tcor >= 65) {
return 1;
}
}
else if (mu_mv_mag_sq >= 36) {
if (Rs < 715) {
return 8;
}
else if (Rs >= 715) {
return 1;
}
}
}
else if (TSC >= 4) {
if (ssDCval < 55) {
if (Cs < 314) {
if (Rs < 581) {
return 4;
}
else if (Rs >= 581) {
return 1;
}
}
else if (Cs >= 314) {
if (diffTSC < 1) {
if (diffMVdiffVal < 7) {
return 8;
}
else if (diffMVdiffVal >= 7) {
return 1;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (ssDCval >= 55) {
if (refDCval < 56) {
return 4;
}
else if (refDCval >= 56) {
return 1;
}
}
}
}
}
else if (MVDiff >= 45) {
if (Rs < 514) {
return 2;
}
else if (Rs >= 514) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (Rs < 601) {
if (mu_mv_mag_sq < 69) {
return 4;
}
else if (mu_mv_mag_sq >= 69) {
return 8;
}
}
else if (Rs >= 601) {
if (ssDCval < 42) {
if (Rs < 899) {
return 1;
}
else if (Rs >= 899) {
return 2;
}
}
else if (ssDCval >= 42) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 114) {
if (posBalance < 52) {
if (SC < 576) {
return 8;
}
else if (SC >= 576) {
return 1;
}
}
else if (posBalance >= 52) {
if (MVDiff < 1) {
if (Rs < 417) {
return 1;
}
else if (Rs >= 417) {
if (Rs < 576) {
return 4;
}
else if (Rs >= 576) {
return 2;
}
}
}
else if (MVDiff >= 1) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 18) {
if (SC < 1125) {
if (Rs < 576) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (Rs >= 576) {
return 8;
}
}
else if (SC >= 1125) {
return 1;
}
}
else if (negBalance >= 18) {
if (ssDCval < 46) {
if (Rs < 519) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Rs >= 519) {
return 4;
}
}
else if (ssDCval >= 46) {
if (diffRsCsdiff < -3) {
return 4;
}
else if (diffRsCsdiff >= -3) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 58) {
if (Cs < 273) {
if (AFD < 3) {
if (Rs < 1021) {
if (posBalance < 37) {
if (negBalance < 19) {
if (Cs < 263) {
return 8;
}
else if (Cs >= 263) {
return 4;
}
}
else if (negBalance >= 19) {
if (negBalance < 42) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
return 4;
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
else if (posBalance >= 37) {
if (ssDCval < 71) {
if (MVDiff < 1) {
if (Rs < 764) {
return 4;
}
else if (Rs >= 764) {
return 1;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (Rs >= 1021) {
return 1;
}
}
else if (AFD >= 3) {
if (tcor < 86) {
if (ssDCval < 62) {
if (SC < 548) {
if (Rs < 334) {
return 4;
}
else if (Rs >= 334) {
return 1;
}
}
else if (SC >= 548) {
return 2;
}
}
else if (ssDCval >= 62) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (posBalance < 62) {
if (Cs < 175) {
if (Rs < 565) {
return 4;
}
else if (Rs >= 565) {
return 1;
}
}
else if (Cs >= 175) {
return 4;
}
}
else if (posBalance >= 62) {
if (Rs < 776) {
return 1;
}
else if (Rs >= 776) {
return 2;
}
}
}
}
}
else if (tcor >= 86) {
if (refDCval < 69) {
if (Cs < 196) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 781) {
return 1;
}
else if (Rs >= 781) {
return 2;
}
}
}
else if (Cs >= 196) {
if (Cs < 264) {
if (tcor < 97) {
if (Cs < 239) {
return 1;
}
else if (Cs >= 239) {
if (mu_mv_mag_sq < 42) {
return 1;
}
else if (mu_mv_mag_sq >= 42) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
return 4;
}
}
}
}
else if (tcor >= 97) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (Cs >= 264) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
return 1;
}
}
}
}
else if (refDCval >= 69) {
if (MVDiff < 8) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
else if (MVDiff >= 8) {
if (AFD < 6) {
if (Rs < 599) {
return 8;
}
else if (Rs >= 599) {
return 4;
}
}
else if (AFD >= 6) {
if (mu_mv_mag_sq < 30) {
return 1;
}
else if (mu_mv_mag_sq >= 30) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 273) {
if (mu_mv_mag_sq < 78) {
if (mcTcor < 97) {
if (ssDCval < 63) {
if (RsCsDiff < 12) {
if (MVDiff < 4) {
if (Rs < 364) {
return 2;
}
else if (Rs >= 364) {
return 1;
}
}
else if (MVDiff >= 4) {
if (TSC < 5) {
if (Rs < 831) {
if (SC < 612) {
return 4;
}
else if (SC >= 612) {
if (Rs < 477) {
return 2;
}
else if (Rs >= 477) {
if (Rs < 503) {
return 4;
}
else if (Rs >= 503) {
return 2;
}
}
}
}
else if (Rs >= 831) {
if (Rs < 936) {
return 8;
}
else if (Rs >= 936) {
return 4;
}
}
}
else if (TSC >= 5) {
return 4;
}
}
}
else if (RsCsDiff >= 12) {
if (negBalance < 51) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (negBalance >= 51) {
return 2;
}
}
}
else if (ssDCval >= 63) {
if (Cs < 309) {
if (Rs < 909) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Rs >= 909) {
return 4;
}
}
else if (Cs >= 309) {
if (mu_mv_mag_sq < 11) {
if (Cs < 451) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (posBalance < 66) {
return 4;
}
else if (posBalance >= 66) {
return 2;
}
}
}
else if (Cs >= 451) {
if (ssDCval < 69) {
if (Cs < 519) {
return 1;
}
else if (Cs >= 519) {
return 4;
}
}
else if (ssDCval >= 69) {
if (refDCval < 72) {
return 4;
}
else if (refDCval >= 72) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (TSC < 5) {
if (MVDiff < 3) {
if (Rs < 671) {
return 1;
}
else if (Rs >= 671) {
return 8;
}
}
else if (MVDiff >= 3) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (TSC >= 5) {
if (Rs < 458) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
else if (Rs >= 458) {
if (ssDCval < 66) {
return 2;
}
else if (ssDCval >= 66) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 669) {
if (TSC < 4) {
if (posBalance < 70) {
if (RsCsDiff < 4) {
if (MVDiff < 6) {
if (refDCval < 65) {
if (MVDiff < 3) {
if (Rs < 543) {
if (Cs < 419) {
return 4;
}
else if (Cs >= 419) {
return 1;
}
}
else if (Rs >= 543) {
if (Rs < 634) {
return 4;
}
else if (Rs >= 634) {
return 8;
}
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (refDCval >= 65) {
return 1;
}
}
else if (MVDiff >= 6) {
if (ssDCval < 70) {
return 4;
}
else if (ssDCval >= 70) {
return 1;
}
}
}
else if (RsCsDiff >= 4) {
if (Rs < 320) {
if (Cs < 497) {
return 4;
}
else if (Cs >= 497) {
return 1;
}
}
else if (Rs >= 320) {
return 4;
}
}
}
else if (posBalance >= 70) {
return 2;
}
}
else if (TSC >= 4) {
return 1;
}
}
else if (Rs >= 669) {
if (MVDiff < 1) {
if (Rs < 983) {
return 8;
}
else if (Rs >= 983) {
return 1;
}
}
else if (MVDiff >= 1) {
if (mcTcor < 100) {
if (negBalance < 18) {
if (Rs < 986) {
return 1;
}
else if (Rs >= 986) {
if (Rs < 1200) {
return 2;
}
else if (Rs >= 1200) {
return 4;
}
}
}
else if (negBalance >= 18) {
if (tcor < 88) {
if (Rs < 935) {
return 1;
}
else if (Rs >= 935) {
return 4;
}
}
else if (tcor >= 88) {
return 1;
}
}
}
else if (mcTcor >= 100) {
if (Cs < 317) {
return 4;
}
else if (Cs >= 317) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (ssDCval < 69) {
if (tcor < 92) {
if (Cs < 477) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Rs < 366) {
return 1;
}
else if (Rs >= 366) {
return 2;
}
}
}
else if (Cs >= 477) {
return 8;
}
}
else if (tcor >= 92) {
if (SC < 706) {
if (ssDCval < 65) {
if (posBalance < 45) {
if (Rs < 321) {
return 4;
}
else if (Rs >= 321) {
return 1;
}
}
else if (posBalance >= 45) {
if (negBalance < 22) {
return 4;
}
else if (negBalance >= 22) {
return 2;
}
}
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (SC >= 706) {
if (Rs < 956) {
if (Rs < 834) {
if (mu_mv_mag_sq < 108) {
if (negBalance < 25) {
if (Rs < 736) {
if (mu_mv_mag_sq < 98) {
return 8;
}
else if (mu_mv_mag_sq >= 98) {
if (Cs < 355) {
return 8;
}
else if (Cs >= 355) {
if (Cs < 500) {
if (SC < 1011) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 550) {
return 8;
}
else if (Rs >= 550) {
return 1;
}
}
}
else if (SC >= 1011) {
return 8;
}
}
else if (Cs >= 500) {
return 4;
}
}
}
}
else if (Rs >= 736) {
return 1;
}
}
else if (negBalance >= 25) {
if (negBalance < 39) {
if (diffTSC < -1) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (diffTSC >= -1) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 83) {
return 1;
}
else if (mu_mv_mag_sq >= 83) {
if (refDCval < 66) {
if (tcor < 96) {
if (Rs < 427) {
return 1;
}
else if (Rs >= 427) {
return 2;
}
}
else if (tcor >= 96) {
if (CsDiff < 1) {
if (Cs < 420) {
return 8;
}
else if (Cs >= 420) {
return 1;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
}
else if (refDCval >= 66) {
return 1;
}
}
}
else if (posBalance >= 49) {
if (tcor < 93) {
if (Rs < 520) {
return 8;
}
else if (Rs >= 520) {
return 1;
}
}
else if (tcor >= 93) {
return 1;
}
}
}
}
else if (negBalance >= 39) {
if (Cs < 368) {
return 1;
}
else if (Cs >= 368) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (ssDCval < 66) {
if (SC < 1011) {
return 1;
}
else if (SC >= 1011) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (ssDCval >= 66) {
if (posBalance < 43) {
if (MVDiff < 1) {
if (Rs < 530) {
return 8;
}
else if (Rs >= 530) {
return 1;
}
}
else if (MVDiff >= 1) {
if (Rs < 550) {
return 1;
}
else if (Rs >= 550) {
return 2;
}
}
}
else if (posBalance >= 43) {
if (negBalance < 18) {
return 1;
}
else if (negBalance >= 18) {
if (Rs < 417) {
return 1;
}
else if (Rs >= 417) {
if (ssDCval < 68) {
if (Rs < 513) {
return 2;
}
else if (Rs >= 513) {
return 1;
}
}
else if (ssDCval >= 68) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 834) {
if (Cs < 369) {
return 2;
}
else if (Cs >= 369) {
return 8;
}
}
}
else if (Rs >= 956) {
return 1;
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 117) {
if (tcor < 98) {
if (diffAFD < 1) {
if (refDCval < 71) {
return 1;
}
else if (refDCval >= 71) {
if (MVDiff < 2) {
if (Rs < 555) {
return 2;
}
else if (Rs >= 555) {
return 4;
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 85) {
return 1;
}
else if (mu_mv_mag_sq >= 85) {
return 8;
}
}
}
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (tcor >= 98) {
if (MVDiff < 3) {
if (SC < 1071) {
if (Cs < 338) {
return 8;
}
else if (Cs >= 338) {
return 2;
}
}
else if (SC >= 1071) {
if (Cs < 520) {
return 8;
}
else if (Cs >= 520) {
return 2;
}
}
}
else if (MVDiff >= 3) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 549) {
return 8;
}
else if (Rs >= 549) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (Rs < 601) {
if (SC < 979) {
if (Cs < 398) {
return 2;
}
else if (Cs >= 398) {
if (Rs < 432) {
return 2;
}
else if (Rs >= 432) {
return 1;
}
}
}
else if (SC >= 979) {
return 1;
}
}
else if (Rs >= 601) {
if (ssDCval < 72) {
if (tcor < 99) {
if (Rs < 657) {
return 1;
}
else if (Rs >= 657) {
return 4;
}
}
else if (tcor >= 99) {
if (Rs < 859) {
return 2;
}
else if (Rs >= 859) {
return 8;
}
}
}
else if (ssDCval >= 72) {
return 4;
}
}
}
}
}
}
}
}
else if (tcor >= 101) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
if (RsDiff < 1) {
if (Rs < 380) {
return 1;
}
else if (Rs >= 380) {
if (Rs < 823) {
return 8;
}
else if (Rs >= 823) {
return 4;
}
}
}
else if (RsDiff >= 1) {
if (negBalance < 42) {
if (Cs < 185) {
return 2;
}
else if (Cs >= 185) {
return 1;
}
}
else if (negBalance >= 42) {
if (SC < 859) {
if (Cs < 86) {
return 1;
}
else if (Cs >= 86) {
return 8;
}
}
else if (SC >= 859) {
if (SC < 1089) {
return 4;
}
else if (SC >= 1089) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1279) {
if (RsDiff < 2) {
if (SC < 1587) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 8;
}
}
}
else if (SC >= 1587) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1620) {
if (TSCindex < 2) {
if (refDCval < 64) {
if (ssDCval < 60) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (refDCval >= 64) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (TSCindex >= 2) {
if (posBalance < 72) {
if (ssDCval < 63) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
return 8;
}
}
else if (ssDCval >= 63) {
if (ssDCval < 68) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 8;
}
}
else if (ssDCval >= 68) {
if (Cs < 443) {
return 8;
}
else if (Cs >= 443) {
return 1;
}
}
}
}
else if (posBalance >= 72) {
return 2;
}
}
}
else if (Rs >= 1620) {
if (Cs < 381) {
return 2;
}
else if (Cs >= 381) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (gchDC < 2) {
if (tcor < 98) {
if (SC < 2452) {
if (tcor < 96) {
if (negBalance < 23) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (negBalance >= 23) {
if (Rs < 1289) {
return 4;
}
else if (Rs >= 1289) {
return 1;
}
}
}
else if (tcor >= 96) {
if (MVDiff < 4) {
if (Cs < 411) {
return 2;
}
else if (Cs >= 411) {
if (Rs < 1504) {
return 1;
}
else if (Rs >= 1504) {
return 8;
}
}
}
else if (MVDiff >= 4) {
return 1;
}
}
}
else if (SC >= 2452) {
return 4;
}
}
else if (tcor >= 98) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (Rs < 1928) {
return 8;
}
else if (Rs >= 1928) {
return 1;
}
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 59) {
return 2;
}
else if (ssDCval >= 59) {
if (Rs < 1789) {
return 1;
}
else if (Rs >= 1789) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 73) {
if (Rs < 1118) {
if (SC < 885) {
if (mu_mv_mag_sq < 119) {
if (Cs < 179) {
if (ssDCval < 85) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Cs < 86) {
return 1;
}
else if (Cs >= 86) {
if (Rs < 408) {
return 8;
}
else if (Rs >= 408) {
if (TSC < 5) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (TSC >= 5) {
return 1;
}
}
}
}
}
else if (ssDCval >= 85) {
if (mu_mv_mag_sq < 77) {
if (mu_mv_mag_sq < 25) {
if (ssDCval < 110) {
return 1;
}
else if (ssDCval >= 110) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (tcor < 83) {
if (Rs < 309) {
return 4;
}
else if (Rs >= 309) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (diffMVdiffVal < 8) {
return 2;
}
else if (diffMVdiffVal >= 8) {
return 8;
}
}
}
}
else if (tcor >= 83) {
if (Cs < 118) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Cs >= 118) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
return 1;
}
}
}
else if (Cs >= 179) {
if (mcTcor < 96) {
if (AFD < 12) {
if (mu_mv_mag_sq < 34) {
if (TSC < 6) {
if (negBalance < 57) {
if (Cs < 341) {
if (Rs < 581) {
if (Rs < 340) {
return 4;
}
else if (Rs >= 340) {
return 1;
}
}
else if (Rs >= 581) {
return 4;
}
}
else if (Cs >= 341) {
if (AFD < 7) {
if (Rs < 369) {
return 2;
}
else if (Rs >= 369) {
return 8;
}
}
else if (AFD >= 7) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 1;
}
}
}
}
else if (negBalance >= 57) {
if (Cs < 334) {
return 2;
}
else if (Cs >= 334) {
if (Rs < 354) {
return 8;
}
else if (Rs >= 354) {
return 4;
}
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 34) {
if (Rs < 343) {
if (gchDC < 2) {
if (diffRsCsdiff < -1) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (Rs >= 343) {
if (ssDCval < 177) {
if (Cs < 389) {
if (ssDCval < 110) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (Cs < 383) {
return 8;
}
else if (Cs >= 383) {
return 1;
}
}
}
else if (ssDCval >= 110) {
return 8;
}
}
else if (Cs >= 389) {
return 8;
}
}
else if (ssDCval >= 177) {
return 2;
}
}
}
}
else if (AFD >= 12) {
if (mcTcor < 86) {
if (Rs < 321) {
if (Rs < 313) {
return 2;
}
else if (Rs >= 313) {
return 4;
}
}
else if (Rs >= 321) {
return 2;
}
}
else if (mcTcor >= 86) {
if (refDCval < 91) {
if (mu_mv_mag_sq < 57) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 8;
}
}
else if (mu_mv_mag_sq >= 57) {
return 4;
}
}
else if (refDCval >= 91) {
if (diffRsCsdiff < -4) {
if (Rs < 339) {
return 1;
}
else if (Rs >= 339) {
return 4;
}
}
else if (diffRsCsdiff >= -4) {
if (mu_mv_mag_sq < 15) {
if (negBalance < 66) {
return 1;
}
else if (negBalance >= 66) {
return 2;
}
}
else if (mu_mv_mag_sq >= 15) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 77) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 58) {
if (negBalance < 38) {
return 4;
}
else if (negBalance >= 38) {
return 1;
}
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 400) {
return 8;
}
else if (Rs >= 400) {
return 2;
}
}
}
}
else if (refDCval >= 77) {
if (mu_mv_mag_sq < 10) {
if (MVDiff < 23) {
if (Rs < 562) {
if (posBalance < 63) {
if (Cs < 506) {
if (TSC < 6) {
if (ssDCval < 180) {
return 1;
}
else if (ssDCval >= 180) {
return 2;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (Cs >= 506) {
return 8;
}
}
else if (posBalance >= 63) {
if (Rs < 446) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 4;
}
}
else if (Rs >= 446) {
return 1;
}
}
}
else if (Rs >= 562) {
if (Rs < 634) {
return 4;
}
else if (Rs >= 634) {
return 1;
}
}
}
else if (MVDiff >= 23) {
if (posBalance < 41) {
return 2;
}
else if (posBalance >= 41) {
if (Rs < 466) {
if (diffAFD < 4) {
return 4;
}
else if (diffAFD >= 4) {
if (Rs < 377) {
return 2;
}
else if (Rs >= 377) {
return 4;
}
}
}
else if (Rs >= 466) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (MVDiff < 6) {
if (ssDCval < 125) {
if (tcor < 89) {
if (ssDCval < 101) {
return 4;
}
else if (ssDCval >= 101) {
return 1;
}
}
else if (tcor >= 89) {
if (tcor < 98) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
return 2;
}
}
else if (tcor >= 98) {
if (Cs < 448) {
if (SC < 878) {
return 2;
}
else if (SC >= 878) {
return 1;
}
}
else if (Cs >= 448) {
return 1;
}
}
}
}
else if (ssDCval >= 125) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (TSCindex < 3) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Rs < 357) {
return 4;
}
else if (Rs >= 357) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (Rs < 425) {
if (negBalance < 50) {
if (ssDCval < 133) {
if (diffAFD < -1) {
if (Rs < 405) {
return 8;
}
else if (Rs >= 405) {
return 1;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (ssDCval >= 133) {
if (Cs < 342) {
return 2;
}
else if (Cs >= 342) {
if (Cs < 476) {
return 8;
}
else if (Cs >= 476) {
return 4;
}
}
}
}
else if (negBalance >= 50) {
if (Cs < 503) {
return 8;
}
else if (Cs >= 503) {
return 1;
}
}
}
else if (Rs >= 425) {
if (SC < 840) {
if (Rs < 439) {
return 4;
}
else if (Rs >= 439) {
return 8;
}
}
else if (SC >= 840) {
if (Rs < 561) {
return 4;
}
else if (Rs >= 561) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (Cs < 349) {
if (Rs < 313) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (Rs >= 313) {
if (ssDCval < 94) {
if (ssDCval < 81) {
if (Rs < 460) {
return 8;
}
else if (Rs >= 460) {
return 1;
}
}
else if (ssDCval >= 81) {
return 2;
}
}
else if (ssDCval >= 94) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (Rs < 465) {
return 2;
}
else if (Rs >= 465) {
return 8;
}
}
}
}
}
else if (Cs >= 349) {
return 8;
}
}
}
else if (SC >= 885) {
if (mu_mv_mag_sq < 63) {
if (RsCsDiff < 2) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
return 1;
}
}
}
else if (RsCsDiff >= 2) {
if (mcTcor < 93) {
if (MVDiff < 63) {
if (refDCval < 85) {
if (mcTcor < 92) {
if (Cs < 397) {
return 2;
}
else if (Cs >= 397) {
return 8;
}
}
else if (mcTcor >= 92) {
return 1;
}
}
else if (refDCval >= 85) {
if (mu_mv_mag_sq < 2) {
if (Cs < 468) {
return 8;
}
else if (Cs >= 468) {
if (Rs < 635) {
return 2;
}
else if (Rs >= 635) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (MVDiff >= 63) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (Cs < 491) {
if (MVDiff < 14) {
if (Cs < 367) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
if (ssDCval < 95) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
else if (Cs >= 367) {
if (tcor < 85) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (tcor >= 85) {
if (Rs < 1033) {
if (RsCsDiff < 16) {
return 1;
}
else if (RsCsDiff >= 16) {
if (Rs < 475) {
return 8;
}
else if (Rs >= 475) {
return 1;
}
}
}
else if (Rs >= 1033) {
return 4;
}
}
}
}
else if (MVDiff >= 14) {
if (Cs < 145) {
return 1;
}
else if (Cs >= 145) {
if (negBalance < 51) {
if (ssDCval < 135) {
if (ssDCval < 78) {
if (Rs < 913) {
return 4;
}
else if (Rs >= 913) {
return 1;
}
}
else if (ssDCval >= 78) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (ssDCval >= 135) {
return 4;
}
}
else if (negBalance >= 51) {
if (SC < 945) {
return 4;
}
else if (SC >= 945) {
if (Rs < 610) {
return 2;
}
else if (Rs >= 610) {
return 8;
}
}
}
}
}
}
else if (Cs >= 491) {
if (posBalance < 54) {
if (Rs < 410) {
if (MVDiff < 11) {
if (Rs < 390) {
return 1;
}
else if (Rs >= 390) {
if (Rs < 399) {
return 2;
}
else if (Rs >= 399) {
return 8;
}
}
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (Rs >= 410) {
if (refDCval < 99) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
if (Rs < 523) {
return 8;
}
else if (Rs >= 523) {
return 1;
}
}
}
else if (refDCval >= 99) {
return 2;
}
}
}
else if (posBalance >= 54) {
if (mcTcor < 97) {
if (MVDiff < 39) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
if (Cs < 514) {
return 8;
}
else if (Cs >= 514) {
return 1;
}
}
}
else if (MVDiff >= 39) {
return 4;
}
}
else if (mcTcor >= 97) {
if (ssDCval < 167) {
if (AFD < 7) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 2;
}
}
else if (AFD >= 7) {
return 1;
}
}
else if (ssDCval >= 167) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (ssDCval < 96) {
if (MVDiff < 1) {
if (refDCval < 94) {
if (SC < 1253) {
if (ssDCval < 78) {
if (ssDCval < 74) {
if (Rs < 578) {
return 1;
}
else if (Rs >= 578) {
return 4;
}
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (ssDCval >= 78) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
if (mu_mv_mag_sq < 108) {
return 1;
}
else if (mu_mv_mag_sq >= 108) {
return 8;
}
}
}
}
}
else if (SC >= 1253) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (Cs < 363) {
return 2;
}
else if (Cs >= 363) {
return 8;
}
}
}
}
else if (refDCval >= 94) {
return 8;
}
}
else if (MVDiff >= 1) {
if (Rs < 617) {
if (negBalance < 41) {
if (MVDiff < 3) {
if (ssDCval < 94) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
if (Rs < 567) {
return 1;
}
else if (Rs >= 567) {
return 8;
}
}
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (MVDiff >= 3) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
return 1;
}
}
}
else if (negBalance >= 41) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
return 8;
}
}
}
else if (Rs >= 617) {
if (tcor < 98) {
if (ssDCval < 77) {
if (Rs < 774) {
return 4;
}
else if (Rs >= 774) {
if (Rs < 935) {
return 8;
}
else if (Rs >= 935) {
return 1;
}
}
}
else if (ssDCval >= 77) {
if (negBalance < 36) {
if (MVDiff < 9) {
if (Rs < 838) {
return 1;
}
else if (Rs >= 838) {
return 8;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (negBalance >= 36) {
return 8;
}
}
}
else if (tcor >= 98) {
if (negBalance < 29) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (MVDiff < 9) {
if (Cs < 235) {
return 8;
}
else if (Cs >= 235) {
return 1;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
}
}
else if (ssDCval >= 96) {
if (refDCval < 98) {
if (Cs < 465) {
return 4;
}
else if (Cs >= 465) {
return 8;
}
}
else if (refDCval >= 98) {
if (MVDiff < 3) {
if (diffMVdiffVal < -3) {
if (Rs < 477) {
return 8;
}
else if (Rs >= 477) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (MVDiff >= 3) {
if (tcor < 97) {
if (TSC < 3) {
if (Rs < 543) {
return 1;
}
else if (Rs >= 543) {
return 4;
}
}
else if (TSC >= 3) {
if (Cs < 525) {
return 8;
}
else if (Cs >= 525) {
if (Rs < 456) {
return 8;
}
else if (Rs >= 456) {
return 2;
}
}
}
}
else if (tcor >= 97) {
if (Rs < 607) {
if (posBalance < 38) {
if (Rs < 510) {
return 1;
}
else if (Rs >= 510) {
return 8;
}
}
else if (posBalance >= 38) {
if (Cs < 475) {
return 1;
}
else if (Cs >= 475) {
return 2;
}
}
}
else if (Rs >= 607) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1118) {
if (ssDCval < 97) {
if (Cs < 307) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 73) {
if (Rs < 1374) {
return 1;
}
else if (Rs >= 1374) {
return 8;
}
}
else if (ssDCval >= 73) {
if (SC < 1281) {
if (Cs < 34) {
return 1;
}
else if (Cs >= 34) {
return 8;
}
}
else if (SC >= 1281) {
if (diffMVdiffVal < -163) {
return 8;
}
else if (diffMVdiffVal >= -163) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 1197) {
return 1;
}
else if (Rs >= 1197) {
return 2;
}
}
}
else if (Cs >= 307) {
if (mu_mv_mag_sq < 101) {
if (posBalance < 55) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (negBalance < 34) {
if (RsDiff < 2) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
if (Rs < 1317) {
return 1;
}
else if (Rs >= 1317) {
return 8;
}
}
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (negBalance >= 34) {
if (Cs < 414) {
if (MVDiff < 4) {
if (Rs < 1587) {
return 8;
}
else if (Rs >= 1587) {
return 1;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (Cs >= 414) {
return 1;
}
}
}
}
}
else if (posBalance >= 55) {
if (Rs < 1343) {
return 8;
}
else if (Rs >= 1343) {
if (Rs < 1708) {
return 1;
}
else if (Rs >= 1708) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (Cs < 497) {
if (Rs < 2078) {
return 1;
}
else if (Rs >= 2078) {
if (Rs < 2123) {
return 8;
}
else if (Rs >= 2123) {
return 2;
}
}
}
else if (Cs >= 497) {
return 4;
}
}
}
}
else if (ssDCval >= 97) {
if (MVDiff < 25) {
if (ssDCval < 127) {
if (mu_mv_mag_sq < 85) {
return 8;
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 1738) {
return 8;
}
else if (Rs >= 1738) {
if (Rs < 2113) {
return 1;
}
else if (Rs >= 2113) {
return 8;
}
}
}
}
else if (ssDCval >= 127) {
if (Cs < 400) {
return 4;
}
else if (Cs >= 400) {
return 2;
}
}
}
else if (MVDiff >= 25) {
if (diffMVdiffVal < -91) {
if (Rs < 1741) {
return 8;
}
else if (Rs >= 1741) {
return 1;
}
}
else if (diffMVdiffVal >= -91) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (posBalance < 53) {
if (SC < 1996) {
if (Cs < 256) {
if (negBalance < 60) {
if (mu_mv_mag_sq < 31) {
return 1;
}
else if (mu_mv_mag_sq >= 31) {
if (tcor < 67) {
if (mcTcor < 64) {
if (Rs < 511) {
return 1;
}
else if (Rs >= 511) {
return 2;
}
}
else if (mcTcor >= 64) {
return 1;
}
}
else if (tcor >= 67) {
if (Cs < 219) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (Cs >= 219) {
return 1;
}
}
}
}
else if (negBalance >= 60) {
if (AFD < 30) {
if (tcor < 73) {
if (diffMVdiffVal < 28) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
return 2;
}
}
else if (diffMVdiffVal >= 28) {
return 1;
}
}
else if (tcor >= 73) {
return 1;
}
}
else if (AFD >= 30) {
return 1;
}
}
}
else if (Cs >= 256) {
if (SC < 1170) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (negBalance < 40) {
if (Cs < 274) {
return 8;
}
else if (Cs >= 274) {
return 1;
}
}
else if (negBalance >= 40) {
if (posBalance < 51) {
if (SC < 937) {
if (negBalance < 57) {
if (Rs < 540) {
return 8;
}
else if (Rs >= 540) {
return 1;
}
}
else if (negBalance >= 57) {
if (posBalance < 40) {
if (Rs < 376) {
return 2;
}
else if (Rs >= 376) {
return 1;
}
}
else if (posBalance >= 40) {
return 2;
}
}
}
else if (SC >= 937) {
if (Cs < 471) {
return 2;
}
else if (Cs >= 471) {
if (Cs < 508) {
return 4;
}
else if (Cs >= 508) {
return 2;
}
}
}
}
else if (posBalance >= 51) {
if (Cs < 458) {
return 8;
}
else if (Cs >= 458) {
return 1;
}
}
}
}
}
else if (SC >= 1170) {
if (mu_mv_mag_sq < 23) {
if (diffMVdiffVal < 69) {
if (Cs < 468) {
if (AFD < 33) {
return 1;
}
else if (AFD >= 33) {
return 8;
}
}
else if (Cs >= 468) {
if (Rs < 855) {
return 8;
}
else if (Rs >= 855) {
return 4;
}
}
}
else if (diffMVdiffVal >= 69) {
return 8;
}
}
else if (mu_mv_mag_sq >= 23) {
if (ssDCval < 49) {
return 2;
}
else if (ssDCval >= 49) {
if (Rs < 1305) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (Rs >= 1305) {
if (diffRsCsdiff < 28) {
if (Cs < 271) {
return 1;
}
else if (Cs >= 271) {
return 2;
}
}
else if (diffRsCsdiff >= 28) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 1996) {
if (MVDiff < 43) {
if (gchDC < 4) {
if (diffMVdiffVal < -11) {
if (Rs < 1995) {
return 2;
}
else if (Rs >= 1995) {
return 1;
}
}
else if (diffMVdiffVal >= -11) {
return 2;
}
}
else if (gchDC >= 4) {
if (Rs < 2197) {
return 1;
}
else if (Rs >= 2197) {
return 2;
}
}
}
else if (MVDiff >= 43) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
if (mu_mv_mag_sq < 7) {
return 8;
}
else if (mu_mv_mag_sq >= 7) {
if (CsDiff < 5) {
if (diffAFD < -7) {
return 2;
}
else if (diffAFD >= -7) {
return 1;
}
}
else if (CsDiff >= 5) {
if (Rs < 1896) {
return 8;
}
else if (Rs >= 1896) {
return 2;
}
}
}
}
}
}
}
else if (posBalance >= 53) {
if (diffAFD < 4) {
if (Cs < 365) {
if (tcor < 39) {
if (RsCsDiff < 43) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (SC < 687) {
return 1;
}
else if (SC >= 687) {
return 8;
}
}
}
else if (RsCsDiff >= 43) {
if (RsDiff < 12) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
if (Rs < 542) {
return 2;
}
else if (Rs >= 542) {
return 4;
}
}
}
else if (RsDiff >= 12) {
if (Cs < 187) {
if (Rs < 1099) {
return 4;
}
else if (Rs >= 1099) {
return 2;
}
}
else if (Cs >= 187) {
return 4;
}
}
}
}
else if (tcor >= 39) {
if (mu_mv_mag_sq < 43) {
if (negBalance < 60) {
if (MVDiff < 16) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (MVDiff >= 16) {
if (Cs < 294) {
if (RsCsDiff < 28) {
if (ssDCval < 99) {
if (tcor < 61) {
if (MVDiff < 32) {
return 8;
}
else if (MVDiff >= 32) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 605) {
return 8;
}
else if (Rs >= 605) {
return 2;
}
}
}
}
else if (tcor >= 61) {
return 1;
}
}
else if (ssDCval >= 99) {
return 1;
}
}
else if (RsCsDiff >= 28) {
if (diffAFD < -4) {
if (Rs < 1302) {
return 2;
}
else if (Rs >= 1302) {
return 4;
}
}
else if (diffAFD >= -4) {
if (SC < 1037) {
if (Rs < 808) {
if (Cs < 230) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (Cs >= 230) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (negBalance < 59) {
return 1;
}
else if (negBalance >= 59) {
return 2;
}
}
}
}
else if (Rs >= 808) {
return 4;
}
}
else if (SC >= 1037) {
if (refDCval < 90) {
return 1;
}
else if (refDCval >= 90) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
return 1;
}
}
}
}
}
}
else if (Cs >= 294) {
if (RsCsDiff < 26) {
if (Rs < 456) {
return 8;
}
else if (Rs >= 456) {
return 1;
}
}
else if (RsCsDiff >= 26) {
if (AFD < 24) {
if (mcTcor < 89) {
return 2;
}
else if (mcTcor >= 89) {
if (Rs < 1302) {
return 1;
}
else if (Rs >= 1302) {
return 2;
}
}
}
else if (AFD >= 24) {
if (Rs < 1546) {
return 1;
}
else if (Rs >= 1546) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 60) {
if (MVDiff < 98) {
if (tcor < 70) {
return 8;
}
else if (tcor >= 70) {
if (Rs < 911) {
return 4;
}
else if (Rs >= 911) {
return 2;
}
}
}
else if (MVDiff >= 98) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (posBalance < 56) {
if (Rs < 663) {
return 8;
}
else if (Rs >= 663) {
return 1;
}
}
else if (posBalance >= 56) {
if (gchDC < 7) {
if (RsCsDiff < 52) {
if (MVDiff < 69) {
if (TSC < 8) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 2;
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (MVDiff >= 69) {
return 2;
}
}
else if (RsCsDiff >= 52) {
return 2;
}
}
else if (gchDC >= 7) {
return 1;
}
}
}
}
}
else if (Cs >= 365) {
if (MVDiff < 12) {
if (SC < 976) {
return 1;
}
else if (SC >= 976) {
return 8;
}
}
else if (MVDiff >= 12) {
if (mcTcor < 88) {
if (refDCval < 126) {
if (ssDCval < 114) {
if (refDCval < 49) {
if (negBalance < 36) {
if (Rs < 668) {
return 1;
}
else if (Rs >= 668) {
return 2;
}
}
else if (negBalance >= 36) {
if (posBalance < 57) {
if (AFD < 18) {
if (Cs < 436) {
return 8;
}
else if (Cs >= 436) {
return 2;
}
}
else if (AFD >= 18) {
return 8;
}
}
else if (posBalance >= 57) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
}
else if (refDCval >= 49) {
if (RsCsDiff < 43) {
if (Rs < 834) {
if (SC < 987) {
return 4;
}
else if (SC >= 987) {
if (mcTcor < 83) {
if (diffMVdiffVal < -9) {
return 4;
}
else if (diffMVdiffVal >= -9) {
return 2;
}
}
else if (mcTcor >= 83) {
return 8;
}
}
}
else if (Rs >= 834) {
if (mu_mv_mag_sq < 37) {
return 8;
}
else if (mu_mv_mag_sq >= 37) {
return 4;
}
}
}
else if (RsCsDiff >= 43) {
if (AFD < 28) {
if (mcTcor < 87) {
if (Rs < 748) {
if (Rs < 741) {
return 1;
}
else if (Rs >= 741) {
return 2;
}
}
else if (Rs >= 748) {
if (diffMVdiffVal < -32) {
if (Cs < 442) {
return 2;
}
else if (Cs >= 442) {
return 4;
}
}
else if (diffMVdiffVal >= -32) {
return 4;
}
}
}
else if (mcTcor >= 87) {
if (Rs < 1000) {
return 8;
}
else if (Rs >= 1000) {
return 2;
}
}
}
else if (AFD >= 28) {
if (gchDC < 8) {
return 2;
}
else if (gchDC >= 8) {
return 8;
}
}
}
}
}
else if (ssDCval >= 114) {
return 8;
}
}
else if (refDCval >= 126) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (SC < 1789) {
return 2;
}
else if (SC >= 1789) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
}
}
}
else if (mcTcor >= 88) {
if (tcor < 66) {
if (RsDiff < 7) {
if (mu_mv_mag_sq < 18) {
if (posBalance < 76) {
if (diffRsCsdiff < 20) {
return 1;
}
else if (diffRsCsdiff >= 20) {
return 8;
}
}
else if (posBalance >= 76) {
if (Rs < 378) {
return 8;
}
else if (Rs >= 378) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
if (Rs < 676) {
return 1;
}
else if (Rs >= 676) {
return 2;
}
}
}
}
else if (RsDiff >= 7) {
if (mcTcor < 93) {
if (mu_mv_mag_sq < 33) {
if (RsCsDiff < 74) {
return 2;
}
else if (RsCsDiff >= 74) {
if (RsDiff < 17) {
return 1;
}
else if (RsDiff >= 17) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (AFD < 31) {
if (Rs < 841) {
return 8;
}
else if (Rs >= 841) {
return 4;
}
}
else if (AFD >= 31) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (Cs < 517) {
if (Rs < 1134) {
return 4;
}
else if (Rs >= 1134) {
if (Rs < 1941) {
return 8;
}
else if (Rs >= 1941) {
return 4;
}
}
}
else if (Cs >= 517) {
return 1;
}
}
}
}
else if (tcor >= 66) {
if (gchDC < 2) {
if (negBalance < 48) {
if (Rs < 830) {
return 1;
}
else if (Rs >= 830) {
return 2;
}
}
else if (negBalance >= 48) {
if (RsDiff < 13) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 8;
}
}
else if (RsDiff >= 13) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (MVDiff < 31) {
return 4;
}
else if (MVDiff >= 31) {
if (diffMVdiffVal < 1) {
return 1;
}
else if (diffMVdiffVal >= 1) {
if (diffTSC < 1) {
if (Rs < 1729) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (Rs >= 1729) {
if (Rs < 2111) {
return 2;
}
else if (Rs >= 2111) {
return 1;
}
}
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
}
}
}
}
}
else if (diffAFD >= 4) {
if (MVDiff < 36) {
if (Rs < 1032) {
return 1;
}
else if (Rs >= 1032) {
return 2;
}
}
else if (MVDiff >= 36) {
if (tcor < 64) {
if (tcor < 27) {
if (MVDiff < 184) {
return 1;
}
else if (MVDiff >= 184) {
if (negBalance < 40) {
if (Cs < 303) {
return 1;
}
else if (Cs >= 303) {
return 4;
}
}
else if (negBalance >= 40) {
return 8;
}
}
}
else if (tcor >= 27) {
if (CsDiff < 6) {
if (diffAFD < 5) {
if (Cs < 383) {
return 2;
}
else if (Cs >= 383) {
return 1;
}
}
else if (diffAFD >= 5) {
if (Cs < 363) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1986) {
return 1;
}
else if (Rs >= 1986) {
return 2;
}
}
}
else if (Cs >= 363) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (Rs < 754) {
return 1;
}
else if (Rs >= 754) {
return 8;
}
}
}
}
else if (tcor >= 64) {
if (diffTSC < 4) {
if (Rs < 416) {
return 4;
}
else if (Rs >= 416) {
if (Cs < 522) {
return 1;
}
else if (Cs >= 522) {
return 4;
}
}
}
else if (diffTSC >= 4) {
if (Cs < 278) {
return 1;
}
else if (Cs >= 278) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 81) {
if (RsCsDiff < 77) {
if (Cs < 299) {
if (ssDCval < 139) {
if (MVDiff < 18) {
if (Rs < 1273) {
if (tcor < 99) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (tcor >= 99) {
return 2;
}
}
else if (Rs >= 1273) {
if (Cs < 256) {
if (AFD < 10) {
if (ssDCval < 92) {
if (diffAFD < -1) {
if (diffRsCsdiff < -8) {
return 1;
}
else if (diffRsCsdiff >= -8) {
return 2;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (ssDCval >= 92) {
if (Rs < 1371) {
return 1;
}
else if (Rs >= 1371) {
return 2;
}
}
}
else if (AFD >= 10) {
if (AFD < 17) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
return 1;
}
}
}
else if (AFD >= 17) {
return 1;
}
}
}
else if (Cs >= 256) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
return 4;
}
}
}
}
else if (MVDiff >= 18) {
if (MVDiff < 59) {
if (Cs < 156) {
if (diffRsCsdiff < -31) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
if (Rs < 732) {
return 1;
}
else if (Rs >= 732) {
return 2;
}
}
}
else if (diffRsCsdiff >= -31) {
return 1;
}
}
else if (Cs >= 156) {
if (Rs < 1018) {
if (diffTSC < 2) {
if (Cs < 162) {
if (Rs < 911) {
return 1;
}
else if (Rs >= 911) {
return 2;
}
}
else if (Cs >= 162) {
if (MVDiff < 30) {
if (Rs < 467) {
return 8;
}
else if (Rs >= 467) {
return 1;
}
}
else if (MVDiff >= 30) {
if (posBalance < 55) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 1;
}
}
else if (posBalance >= 55) {
if (diffMVdiffVal < -11) {
return 1;
}
else if (diffMVdiffVal >= -11) {
return 4;
}
}
}
}
}
else if (diffTSC >= 2) {
if (Rs < 940) {
return 2;
}
else if (Rs >= 940) {
return 1;
}
}
}
else if (Rs >= 1018) {
if (TSC < 7) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
return 4;
}
}
else if (TSC >= 7) {
if (ssDCval < 60) {
return 2;
}
else if (ssDCval >= 60) {
if (Cs < 286) {
return 1;
}
else if (Cs >= 286) {
if (Cs < 291) {
return 2;
}
else if (Cs >= 291) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 59) {
if (Rs < 744) {
if (mcTcor < 97) {
if (MVDiff < 178) {
return 1;
}
else if (MVDiff >= 178) {
if (Cs < 171) {
return 2;
}
else if (Cs >= 171) {
return 8;
}
}
}
else if (mcTcor >= 97) {
return 8;
}
}
else if (Rs >= 744) {
if (mu_mv_mag_sq < 14) {
if (SC < 2268) {
if (MVDiff < 142) {
if (diffRsCsdiff < 34) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
return 1;
}
}
else if (diffRsCsdiff >= 34) {
return 8;
}
}
else if (MVDiff >= 142) {
if (Rs < 999) {
return 2;
}
else if (Rs >= 999) {
return 1;
}
}
}
else if (SC >= 2268) {
return 8;
}
}
else if (mu_mv_mag_sq >= 14) {
return 1;
}
}
}
}
}
else if (ssDCval >= 139) {
if (gchDC < 2) {
if (posBalance < 51) {
if (Rs < 705) {
return 2;
}
else if (Rs >= 705) {
return 1;
}
}
else if (posBalance >= 51) {
return 4;
}
}
else if (gchDC >= 2) {
if (ssDCval < 176) {
return 1;
}
else if (ssDCval >= 176) {
return 2;
}
}
}
}
else if (Cs >= 299) {
if (Rs < 1169) {
if (posBalance < 56) {
if (ssDCval < 71) {
if (mu_mv_mag_sq < 34) {
if (Cs < 517) {
if (Rs < 827) {
if (Rs < 761) {
if (AFD < 14) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
return 8;
}
}
else if (AFD >= 14) {
return 1;
}
}
else if (Rs >= 761) {
return 8;
}
}
else if (Rs >= 827) {
return 1;
}
}
else if (Cs >= 517) {
return 4;
}
}
else if (mu_mv_mag_sq >= 34) {
if (SC < 1110) {
if (Rs < 407) {
return 4;
}
else if (Rs >= 407) {
return 8;
}
}
else if (SC >= 1110) {
if (diffMVdiffVal < -3) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (Cs < 446) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (Cs >= 446) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (mcTcor < 96) {
if (Cs < 343) {
return 1;
}
else if (Cs >= 343) {
return 2;
}
}
else if (mcTcor >= 96) {
if (ssDCval < 105) {
if (MVDiff < 17) {
if (SC < 1521) {
return 4;
}
else if (SC >= 1521) {
return 1;
}
}
else if (MVDiff >= 17) {
if (Rs < 565) {
return 4;
}
else if (Rs >= 565) {
return 1;
}
}
}
else if (ssDCval >= 105) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (Cs < 511) {
return 2;
}
else if (Cs >= 511) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 56) {
if (TSC < 7) {
if (ssDCval < 111) {
if (ssDCval < 66) {
if (Cs < 442) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (Rs < 600) {
return 4;
}
else if (Rs >= 600) {
return 8;
}
}
}
else if (Cs >= 442) {
if (Rs < 590) {
return 1;
}
else if (Rs >= 590) {
return 4;
}
}
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (ssDCval >= 111) {
if (MVDiff < 18) {
if (posBalance < 57) {
if (Rs < 1076) {
return 1;
}
else if (Rs >= 1076) {
return 4;
}
}
else if (posBalance >= 57) {
return 4;
}
}
else if (MVDiff >= 18) {
if (Rs < 942) {
return 8;
}
else if (Rs >= 942) {
if (Rs < 1149) {
return 1;
}
else if (Rs >= 1149) {
return 2;
}
}
}
}
}
else if (TSC >= 7) {
if (Rs < 667) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
if (ssDCval < 150) {
return 8;
}
else if (ssDCval >= 150) {
return 1;
}
}
}
else if (Rs >= 667) {
if (ssDCval < 82) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (negBalance < 48) {
if (Rs < 857) {
return 8;
}
else if (Rs >= 857) {
return 1;
}
}
else if (negBalance >= 48) {
if (Rs < 678) {
return 1;
}
else if (Rs >= 678) {
return 4;
}
}
}
}
else if (ssDCval >= 82) {
if (AFD < 16) {
if (Rs < 780) {
return 1;
}
else if (Rs >= 780) {
return 8;
}
}
else if (AFD >= 16) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1169) {
if (TSC < 3) {
if (ssDCval < 91) {
if (Rs < 1653) {
return 4;
}
else if (Rs >= 1653) {
return 1;
}
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (TSC >= 3) {
if (TSC < 6) {
if (ssDCval < 101) {
if (ssDCval < 66) {
if (Rs < 1834) {
if (CsDiff < 3) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (CsDiff >= 3) {
if (diffMVdiffVal < 5) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
if (Rs < 1273) {
return 2;
}
else if (Rs >= 1273) {
return 1;
}
}
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
}
else if (Rs >= 1834) {
if (ssDCval < 59) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
else if (ssDCval >= 59) {
if (Rs < 1874) {
return 1;
}
else if (Rs >= 1874) {
return 2;
}
}
}
}
else if (ssDCval >= 66) {
if (posBalance < 66) {
if (diffMVdiffVal < 12) {
if (diffRsCsdiff < -76) {
if (Rs < 1967) {
return 1;
}
else if (Rs >= 1967) {
return 2;
}
}
else if (diffRsCsdiff >= -76) {
if (MVDiff < 2) {
if (Rs < 1644) {
return 4;
}
else if (Rs >= 1644) {
return 1;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (diffMVdiffVal >= 12) {
return 4;
}
}
else if (posBalance >= 66) {
return 2;
}
}
}
else if (ssDCval >= 101) {
if (Cs < 458) {
if (AFD < 10) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 4;
}
}
}
else if (AFD >= 10) {
return 1;
}
}
else if (Cs >= 458) {
if (Rs < 1993) {
if (Rs < 1482) {
return 1;
}
else if (Rs >= 1482) {
return 2;
}
}
else if (Rs >= 1993) {
if (Cs < 488) {
return 1;
}
else if (Cs >= 488) {
return 8;
}
}
}
}
}
else if (TSC >= 6) {
if (mcTcor < 96) {
if (posBalance < 54) {
if (ssDCval < 64) {
if (Rs < 1773) {
if (Cs < 423) {
return 1;
}
else if (Cs >= 423) {
return 2;
}
}
else if (Rs >= 1773) {
return 2;
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (posBalance >= 54) {
if (diffRsCsdiff < 10) {
if (posBalance < 65) {
if (posBalance < 58) {
if (negBalance < 56) {
if (Rs < 1264) {
return 8;
}
else if (Rs >= 1264) {
return 1;
}
}
else if (negBalance >= 56) {
return 2;
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (posBalance >= 65) {
return 1;
}
}
else if (diffRsCsdiff >= 10) {
return 2;
}
}
}
else if (mcTcor >= 96) {
if (ssDCval < 110) {
if (diffMVdiffVal < -12) {
return 2;
}
else if (diffMVdiffVal >= -12) {
if (AFD < 10) {
if (Rs < 1418) {
return 8;
}
else if (Rs >= 1418) {
return 1;
}
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (ssDCval >= 110) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (refDCval < 111) {
if (MVDiff < 39) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (RsCsDiff < 108) {
if (CsDiff < 2) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
return 2;
}
}
else if (CsDiff >= 2) {
if (ssDCval < 71) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
return 2;
}
}
}
else if (ssDCval >= 71) {
if (diffMVdiffVal < -3) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Rs < 2086) {
return 2;
}
else if (Rs >= 2086) {
return 4;
}
}
}
else if (diffMVdiffVal >= -3) {
if (Cs < 163) {
return 2;
}
else if (Cs >= 163) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 108) {
if (negBalance < 37) {
return 1;
}
else if (negBalance >= 37) {
if (ssDCval < 106) {
if (mu_mv_mag_sq < 19) {
return 2;
}
else if (mu_mv_mag_sq >= 19) {
if (mu_mv_mag_sq < 23) {
return 1;
}
else if (mu_mv_mag_sq >= 23) {
if (Cs < 83) {
return 1;
}
else if (Cs >= 83) {
return 2;
}
}
}
}
else if (ssDCval >= 106) {
if (mcTcor < 97) {
return 2;
}
else if (mcTcor >= 97) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 39) {
if (TSC < 7) {
if (mu_mv_mag_sq < 24) {
if (mcTcor < 98) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (mcTcor >= 98) {
if (Cs < 82) {
return 1;
}
else if (Cs >= 82) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (Cs < 392) {
return 2;
}
else if (Cs >= 392) {
return 4;
}
}
}
else if (TSC >= 7) {
if (tcor < 80) {
if (diffRsCsdiff < 19) {
if (ssDCval < 104) {
if (MVDiff < 41) {
return 2;
}
else if (MVDiff >= 41) {
return 1;
}
}
else if (ssDCval >= 104) {
return 2;
}
}
else if (diffRsCsdiff >= 19) {
if (Rs < 2035) {
if (Rs < 1685) {
if (Cs < 342) {
return 1;
}
else if (Cs >= 342) {
return 8;
}
}
else if (Rs >= 1685) {
return 8;
}
}
else if (Rs >= 2035) {
return 1;
}
}
}
else if (tcor >= 80) {
if (diffRsCsdiff < 49) {
if (Cs < 482) {
return 1;
}
else if (Cs >= 482) {
if (Rs < 1446) {
return 2;
}
else if (Rs >= 1446) {
return 1;
}
}
}
else if (diffRsCsdiff >= 49) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (Rs < 702) {
return 2;
}
else if (Rs >= 702) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 111) {
if (CsDiff < 2) {
if (SC < 1895) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (Rs < 1039) {
return 2;
}
else if (Rs >= 1039) {
return 1;
}
}
}
else if (SC >= 1895) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (CsDiff < 3) {
if (SC < 2608) {
if (mcTcor < 94) {
if (Rs < 1814) {
if (Rs < 1316) {
return 4;
}
else if (Rs >= 1316) {
return 8;
}
}
else if (Rs >= 1814) {
return 1;
}
}
else if (mcTcor >= 94) {
if (refDCval < 121) {
if (Rs < 2015) {
return 1;
}
else if (Rs >= 2015) {
if (Rs < 2150) {
if (Rs < 2110) {
return 2;
}
else if (Rs >= 2110) {
return 8;
}
}
else if (Rs >= 2150) {
return 1;
}
}
}
else if (refDCval >= 121) {
return 1;
}
}
}
else if (SC >= 2608) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (AFD < 34) {
if (diffTSC < 2) {
if (TSC < 7) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (diffTSC >= 2) {
if (TSC < 9) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (TSC >= 9) {
if (MVDiff < 349) {
return 1;
}
else if (MVDiff >= 349) {
return 2;
}
}
}
}
else if (AFD >= 34) {
if (Cs < 405) {
return 1;
}
else if (Cs >= 405) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 53) {
if (Cs < 376) {
if (RsDiff < 12) {
if (Rs < 1464) {
return 1;
}
else if (Rs >= 1464) {
return 4;
}
}
else if (RsDiff >= 12) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (Cs >= 376) {
if (Cs < 519) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (Cs >= 519) {
if (Rs < 1399) {
return 4;
}
else if (Rs >= 1399) {
if (Rs < 1781) {
return 2;
}
else if (Rs >= 1781) {
return 1;
}
}
}
}
}
else if (negBalance >= 53) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 533) {
if (negBalance < 13) {
if (Rs < 702) {
if (ssDCval < 57) {
if (refDCval < 43) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (MVDiff < 45) {
return 1;
}
else if (MVDiff >= 45) {
if (Rs < 328) {
return 8;
}
else if (Rs >= 328) {
return 2;
}
}
}
}
else if (refDCval >= 43) {
if (tcor < 98) {
if (Cs < 1027) {
if (Rs < 340) {
return 4;
}
else if (Rs >= 340) {
return 8;
}
}
else if (Cs >= 1027) {
return 1;
}
}
else if (tcor >= 98) {
return 1;
}
}
}
else if (ssDCval >= 57) {
if (SC < 1359) {
if (Rs < 519) {
if (Cs < 613) {
return 1;
}
else if (Cs >= 613) {
if (refDCval < 95) {
return 4;
}
else if (refDCval >= 95) {
return 8;
}
}
}
else if (Rs >= 519) {
if (ssDCval < 135) {
return 8;
}
else if (ssDCval >= 135) {
if (TSC < 1) {
return 8;
}
else if (TSC >= 1) {
return 1;
}
}
}
}
else if (SC >= 1359) {
if (posBalance < 28) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
else if (posBalance >= 28) {
if (Cs < 796) {
if (Rs < 656) {
return 2;
}
else if (Rs >= 656) {
if (Rs < 685) {
return 1;
}
else if (Rs >= 685) {
return 8;
}
}
}
else if (Cs >= 796) {
return 8;
}
}
}
}
}
else if (Rs >= 702) {
if (diffAFD < -3) {
if (Rs < 1598) {
if (mcTcor < 99) {
if (diffAFD < -16) {
if (ssDCval < 139) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 8;
}
}
else if (ssDCval >= 139) {
return 8;
}
}
else if (diffAFD >= -16) {
if (mcTcor < 79) {
if (Cs < 1076) {
return 1;
}
else if (Cs >= 1076) {
return 8;
}
}
else if (mcTcor >= 79) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (diffRsCsdiff < -53) {
if (diffMVdiffVal < 28) {
if (diffMVdiffVal < -27) {
if (Rs < 1354) {
return 1;
}
else if (Rs >= 1354) {
return 4;
}
}
else if (diffMVdiffVal >= -27) {
if (Cs < 1056) {
return 8;
}
else if (Cs >= 1056) {
if (Rs < 1191) {
return 1;
}
else if (Rs >= 1191) {
if (Cs < 1103) {
return 4;
}
else if (Cs >= 1103) {
return 8;
}
}
}
}
}
else if (diffMVdiffVal >= 28) {
return 1;
}
}
else if (diffRsCsdiff >= -53) {
if (Cs < 624) {
if (Rs < 1071) {
return 2;
}
else if (Rs >= 1071) {
return 1;
}
}
else if (Cs >= 624) {
if (ssDCval < 86) {
if (Rs < 1228) {
return 8;
}
else if (Rs >= 1228) {
return 1;
}
}
else if (ssDCval >= 86) {
if (diffAFD < -7) {
if (ssDCval < 132) {
if (MVDiff < 31) {
return 1;
}
else if (MVDiff >= 31) {
return 8;
}
}
else if (ssDCval >= 132) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 4;
}
}
}
else if (diffAFD >= -7) {
if (diffRsCsdiff < -29) {
if (Rs < 1262) {
return 8;
}
else if (Rs >= 1262) {
return 4;
}
}
else if (diffRsCsdiff >= -29) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1598) {
if (diffMVdiffVal < -17) {
return 4;
}
else if (diffMVdiffVal >= -17) {
return 8;
}
}
}
else if (diffAFD >= -3) {
if (Rs < 1841) {
if (Rs < 967) {
if (diffMVdiffVal < -1) {
if (posBalance < 27) {
if (diffMVdiffVal < -14) {
return 8;
}
else if (diffMVdiffVal >= -14) {
if (Cs < 611) {
return 8;
}
else if (Cs >= 611) {
if (mcTcor < 99) {
if (Rs < 824) {
return 1;
}
else if (Rs >= 824) {
return 2;
}
}
else if (mcTcor >= 99) {
return 2;
}
}
}
}
else if (posBalance >= 27) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 11) {
if (ssDCval < 147) {
return 8;
}
else if (ssDCval >= 147) {
return 2;
}
}
else if (posBalance >= 11) {
if (ssDCval < 67) {
if (SC < 1466) {
if (tcor < 99) {
if (Cs < 708) {
return 8;
}
else if (Cs >= 708) {
return 2;
}
}
else if (tcor >= 99) {
return 2;
}
}
else if (SC >= 1466) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (ssDCval >= 67) {
if (Rs < 738) {
if (ssDCval < 135) {
if (tcor < 99) {
if (Cs < 750) {
return 1;
}
else if (Cs >= 750) {
return 8;
}
}
else if (tcor >= 99) {
return 8;
}
}
else if (ssDCval >= 135) {
return 1;
}
}
else if (Rs >= 738) {
return 8;
}
}
}
}
}
else if (Rs >= 967) {
if (MVDiff < 3) {
if (ssDCval < 152) {
if (refDCval < 53) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (refDCval >= 53) {
if (diffMVdiffVal < -4) {
if (RsDiff < 1) {
if (Rs < 1126) {
return 4;
}
else if (Rs >= 1126) {
return 1;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
if (ssDCval < 64) {
if (Cs < 990) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (Cs < 787) {
return 1;
}
else if (Cs >= 787) {
return 8;
}
}
}
else if (Cs >= 990) {
return 1;
}
}
else if (ssDCval >= 64) {
if (SC < 2866) {
if (Cs < 1185) {
return 8;
}
else if (Cs >= 1185) {
if (negBalance < 2) {
return 1;
}
else if (negBalance >= 2) {
return 8;
}
}
}
else if (SC >= 2866) {
if (Rs < 1718) {
return 8;
}
else if (Rs >= 1718) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (MVDiff >= 3) {
if (refDCval < 40) {
return 8;
}
else if (refDCval >= 40) {
if (Rs < 1140) {
if (diffMVdiffVal < -4) {
if (Rs < 1044) {
if (Rs < 1011) {
return 4;
}
else if (Rs >= 1011) {
return 8;
}
}
else if (Rs >= 1044) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (Rs >= 1140) {
if (ssDCval < 76) {
if (Rs < 1693) {
return 1;
}
else if (Rs >= 1693) {
return 8;
}
}
else if (ssDCval >= 76) {
if (SC < 2202) {
return 4;
}
else if (SC >= 2202) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1841) {
if (Rs < 2053) {
if (ssDCval < 81) {
if (Cs < 1002) {
if (Cs < 712) {
if (Rs < 1926) {
return 8;
}
else if (Rs >= 1926) {
return 2;
}
}
else if (Cs >= 712) {
if (SC < 2952) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
else if (SC >= 2952) {
if (Rs < 2009) {
return 1;
}
else if (Rs >= 2009) {
return 4;
}
}
}
}
else if (Cs >= 1002) {
if (ssDCval < 69) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 4;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
else if (ssDCval >= 81) {
if (Cs < 1259) {
return 8;
}
else if (Cs >= 1259) {
return 1;
}
}
}
else if (Rs >= 2053) {
if (posBalance < 29) {
return 8;
}
else if (posBalance >= 29) {
if (TSCindex < 2) {
if (Rs < 2272) {
return 8;
}
else if (Rs >= 2272) {
if (Cs < 1318) {
return 2;
}
else if (Cs >= 1318) {
if (posBalance < 30) {
return 2;
}
else if (posBalance >= 30) {
return 8;
}
}
}
}
else if (TSCindex >= 2) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 46) {
if (ssDCval < 48) {
if (MVDiff < 9) {
if (Rs < 1471) {
if (ssDCval < 29) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (ssDCval >= 29) {
if (mcTcor < 93) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (mcTcor >= 93) {
if (posBalance < 38) {
if (posBalance < 36) {
if (Rs < 1422) {
return 8;
}
else if (Rs >= 1422) {
return 4;
}
}
else if (posBalance >= 36) {
return 4;
}
}
else if (posBalance >= 38) {
if (MVDiff < 4) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (Rs < 1438) {
return 8;
}
else if (Rs >= 1438) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (gchDC < 1) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
return 1;
}
}
else if (gchDC >= 1) {
if (Rs < 683) {
return 1;
}
else if (Rs >= 683) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1471) {
if (refDCval < 47) {
if (AFD < 10) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
return 4;
}
}
else if (AFD >= 10) {
return 1;
}
}
else if (refDCval >= 47) {
if (Rs < 1989) {
if (Cs < 835) {
return 8;
}
else if (Cs >= 835) {
return 4;
}
}
else if (Rs >= 1989) {
return 8;
}
}
}
}
else if (MVDiff >= 9) {
if (tcor < 59) {
return 8;
}
else if (tcor >= 59) {
if (negBalance < 34) {
if (AFD < 8) {
if (Cs < 970) {
if (RsCsDiff < 37) {
if (negBalance < 31) {
if (mu_mv_mag_sq < 93) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (Rs < 628) {
return 8;
}
else if (Rs >= 628) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 93) {
return 1;
}
}
else if (negBalance >= 31) {
if (Rs < 872) {
return 8;
}
else if (Rs >= 872) {
return 1;
}
}
}
else if (RsCsDiff >= 37) {
return 8;
}
}
else if (Cs >= 970) {
return 1;
}
}
else if (AFD >= 8) {
if (ssDCval < 44) {
return 1;
}
else if (ssDCval >= 44) {
if (Rs < 1816) {
return 2;
}
else if (Rs >= 1816) {
return 1;
}
}
}
}
else if (negBalance >= 34) {
if (RsDiff < 8) {
if (SC < 1239) {
return 1;
}
else if (SC >= 1239) {
return 8;
}
}
else if (RsDiff >= 8) {
if (Rs < 1877) {
return 1;
}
else if (Rs >= 1877) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 48) {
if (mu_mv_mag_sq < 86) {
if (Cs < 857) {
if (negBalance < 62) {
if (Rs < 1635) {
if (ssDCval < 80) {
if (RsCsDiff < 27) {
if (gchDC < 1) {
if (ssDCval < 56) {
if (negBalance < 36) {
if (Rs < 724) {
return 4;
}
else if (Rs >= 724) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
else if (negBalance >= 36) {
return 8;
}
}
else if (ssDCval >= 56) {
if (mcTcor < 96) {
if (Rs < 760) {
return 2;
}
else if (Rs >= 760) {
if (Rs < 1267) {
return 4;
}
else if (Rs >= 1267) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (AFD < 8) {
if (Rs < 766) {
return 4;
}
else if (Rs >= 766) {
return 1;
}
}
else if (AFD >= 8) {
return 2;
}
}
}
}
else if (gchDC >= 1) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 61) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 1049) {
return 1;
}
else if (Rs >= 1049) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (Cs < 587) {
return 4;
}
else if (Cs >= 587) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (Cs < 755) {
if (SC < 914) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (SC >= 914) {
if (RsCsDiff < 3) {
return 8;
}
else if (RsCsDiff >= 3) {
if (Rs < 1317) {
return 1;
}
else if (Rs >= 1317) {
if (Rs < 1344) {
return 2;
}
else if (Rs >= 1344) {
return 1;
}
}
}
}
}
else if (Cs >= 755) {
if (mu_mv_mag_sq < 62) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (Rs < 560) {
return 2;
}
else if (Rs >= 560) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 62) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 27) {
if (mu_mv_mag_sq < 56) {
if (mu_mv_mag_sq < 36) {
if (Cs < 810) {
if (Rs < 794) {
return 2;
}
else if (Rs >= 794) {
return 8;
}
}
else if (Cs >= 810) {
return 1;
}
}
else if (mu_mv_mag_sq >= 36) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (refDCval < 55) {
return 1;
}
else if (refDCval >= 55) {
return 8;
}
}
}
}
else if (ssDCval >= 80) {
if (ssDCval < 123) {
return 2;
}
else if (ssDCval >= 123) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
if (Rs < 1228) {
if (Rs < 761) {
return 2;
}
else if (Rs >= 761) {
return 4;
}
}
else if (Rs >= 1228) {
return 1;
}
}
}
}
}
else if (Rs >= 1635) {
if (Cs < 781) {
if (ssDCval < 58) {
if (Rs < 1999) {
if (Rs < 1789) {
return 1;
}
else if (Rs >= 1789) {
return 2;
}
}
else if (Rs >= 1999) {
return 4;
}
}
else if (ssDCval >= 58) {
if (refDCval < 72) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
else if (refDCval >= 72) {
return 1;
}
}
}
else if (Cs >= 781) {
if (ssDCval < 110) {
if (mcTcor < 92) {
if (Rs < 2009) {
return 1;
}
else if (Rs >= 2009) {
return 4;
}
}
else if (mcTcor >= 92) {
if (Cs < 800) {
return 2;
}
else if (Cs >= 800) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (ssDCval >= 110) {
return 4;
}
}
}
}
else if (negBalance >= 62) {
if (ssDCval < 93) {
if (MVDiff < 6) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (ssDCval >= 93) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
}
else if (Cs >= 857) {
if (mu_mv_mag_sq < 66) {
if (ssDCval < 153) {
if (gchDC < 3) {
if (ssDCval < 138) {
if (Rs < 976) {
if (CsDiff < 5) {
if (negBalance < 36) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 2;
}
}
else if (negBalance >= 36) {
return 2;
}
}
else if (CsDiff >= 5) {
if (Cs < 987) {
return 8;
}
else if (Cs >= 987) {
return 4;
}
}
}
else if (Rs >= 976) {
if (RsCsDiff < 27) {
if (ssDCval < 59) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (ssDCval >= 59) {
if (mu_mv_mag_sq < 35) {
if (Rs < 1667) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (Rs >= 1667) {
return 1;
}
}
else if (mu_mv_mag_sq >= 35) {
return 4;
}
}
}
else if (RsCsDiff >= 27) {
if (Rs < 1949) {
if (diffRsCsdiff < 22) {
if (refDCval < 54) {
if (Rs < 1186) {
return 1;
}
else if (Rs >= 1186) {
return 2;
}
}
else if (refDCval >= 54) {
if (negBalance < 55) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
if (Cs < 1191) {
return 8;
}
else if (Cs >= 1191) {
if (Rs < 1567) {
return 2;
}
else if (Rs >= 1567) {
return 8;
}
}
}
}
else if (negBalance >= 55) {
if (Rs < 1736) {
return 8;
}
else if (Rs >= 1736) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 22) {
if (TSC < 7) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (TSC >= 7) {
if (Rs < 1752) {
return 8;
}
else if (Rs >= 1752) {
return 2;
}
}
}
}
else if (Rs >= 1949) {
if (AFD < 9) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 1;
}
}
else if (AFD >= 9) {
if (diffRsCsdiff < -10) {
if (posBalance < 44) {
if (Rs < 2338) {
if (Rs < 2211) {
return 4;
}
else if (Rs >= 2211) {
return 2;
}
}
else if (Rs >= 2338) {
return 1;
}
}
else if (posBalance >= 44) {
return 8;
}
}
else if (diffRsCsdiff >= -10) {
if (MVDiff < 13) {
if (MVDiff < 4) {
if (Rs < 2237) {
return 2;
}
else if (Rs >= 2237) {
return 8;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (MVDiff >= 13) {
if (Rs < 2207) {
return 2;
}
else if (Rs >= 2207) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 138) {
if (Rs < 1138) {
return 4;
}
else if (Rs >= 1138) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (diffRsCsdiff < 23) {
if (posBalance < 25) {
return 2;
}
else if (posBalance >= 25) {
return 8;
}
}
else if (diffRsCsdiff >= 23) {
return 2;
}
}
}
else if (ssDCval >= 153) {
return 2;
}
}
else if (mu_mv_mag_sq >= 66) {
if (Cs < 939) {
if (ssDCval < 97) {
if (diffMVdiffVal < 2) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
if (MVDiff < 10) {
if (Rs < 1635) {
return 1;
}
else if (Rs >= 1635) {
return 8;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
}
else if (ssDCval >= 97) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (Cs >= 939) {
if (ssDCval < 139) {
if (CsDiff < 5) {
if (gchDC < 2) {
if (diffRsCsdiff < -9) {
if (Rs < 1924) {
if (Rs < 1693) {
return 2;
}
else if (Rs >= 1693) {
return 8;
}
}
else if (Rs >= 1924) {
if (SC < 3379) {
return 4;
}
else if (SC >= 3379) {
return 8;
}
}
}
else if (diffRsCsdiff >= -9) {
return 4;
}
}
else if (gchDC >= 2) {
if (Rs < 2060) {
return 8;
}
else if (Rs >= 2060) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 1949) {
return 8;
}
else if (Rs >= 1949) {
return 4;
}
}
}
else if (ssDCval >= 139) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (ssDCval < 56) {
if (TSC < 3) {
if (Rs < 505) {
return 2;
}
else if (Rs >= 505) {
if (AFD < 3) {
if (Rs < 1129) {
if (Cs < 865) {
return 4;
}
else if (Cs >= 865) {
return 1;
}
}
else if (Rs >= 1129) {
return 8;
}
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (TSC >= 3) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 110) {
return 1;
}
else if (mu_mv_mag_sq >= 110) {
if (RsDiff < 2) {
if (mcTcor < 96) {
if (Rs < 1151) {
return 8;
}
else if (Rs >= 1151) {
return 4;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (ssDCval >= 56) {
if (Rs < 854) {
if (negBalance < 29) {
if (ssDCval < 124) {
if (Cs < 639) {
return 1;
}
else if (Cs >= 639) {
if (Cs < 768) {
if (Cs < 751) {
return 4;
}
else if (Cs >= 751) {
return 2;
}
}
else if (Cs >= 768) {
return 4;
}
}
}
else if (ssDCval >= 124) {
if (Rs < 829) {
return 8;
}
else if (Rs >= 829) {
return 1;
}
}
}
else if (negBalance >= 29) {
if (negBalance < 54) {
if (mu_mv_mag_sq < 92) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
return 1;
}
}
else if (mu_mv_mag_sq >= 92) {
if (Cs < 943) {
return 8;
}
else if (Cs >= 943) {
if (Rs < 605) {
return 1;
}
else if (Rs >= 605) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (Rs < 748) {
return 8;
}
else if (Rs >= 748) {
return 4;
}
}
}
}
else if (Rs >= 854) {
if (RsDiff < 5) {
if (CsDiff < 3) {
if (MVDiff < 26) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (ssDCval < 60) {
if (mu_mv_mag_sq < 114) {
return 1;
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (ssDCval >= 60) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
return 4;
}
}
}
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (CsDiff >= 3) {
if (TSC < 4) {
if (mu_mv_mag_sq < 107) {
return 4;
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 1162) {
return 1;
}
else if (Rs >= 1162) {
return 8;
}
}
}
else if (TSC >= 4) {
if (ssDCval < 90) {
if (SC < 2290) {
return 8;
}
else if (SC >= 2290) {
if (tcor < 95) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 8;
}
}
else if (tcor >= 95) {
if (RsDiff < 4) {
if (Rs < 1591) {
return 4;
}
else if (Rs >= 1591) {
return 1;
}
}
else if (RsDiff >= 4) {
if (Rs < 2073) {
return 8;
}
else if (Rs >= 2073) {
return 4;
}
}
}
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
else if (RsDiff >= 5) {
if (Rs < 1978) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 8;
}
}
else if (Rs >= 1978) {
if (Cs < 1226) {
return 2;
}
else if (Cs >= 1226) {
return 8;
}
}
}
}
}
}
}
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 37) {
if (negBalance < 45) {
if (negBalance < 36) {
if (tcor < 88) {
if (refDCval < 137) {
if (Cs < 961) {
if (AFD < 9) {
if (Rs < 650) {
if (Rs < 588) {
return 8;
}
else if (Rs >= 588) {
return 2;
}
}
else if (Rs >= 650) {
return 8;
}
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 24) {
if (mcTcor < 92) {
if (Rs < 1738) {
return 8;
}
else if (Rs >= 1738) {
return 4;
}
}
else if (mcTcor >= 92) {
if (Cs < 564) {
return 1;
}
else if (Cs >= 564) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (mu_mv_mag_sq < 36) {
return 1;
}
else if (mu_mv_mag_sq >= 36) {
return 2;
}
}
}
}
else if (Cs >= 961) {
if (gchDC < 2) {
if (Rs < 1588) {
if (mu_mv_mag_sq < 24) {
return 1;
}
else if (mu_mv_mag_sq >= 24) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 754) {
return 1;
}
else if (Rs >= 754) {
return 8;
}
}
}
}
else if (Rs >= 1588) {
return 8;
}
}
else if (gchDC >= 2) {
if (diffRsCsdiff < 30) {
if (negBalance < 31) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (negBalance >= 31) {
return 8;
}
}
else if (diffRsCsdiff >= 30) {
return 2;
}
}
}
}
else if (refDCval >= 137) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (tcor >= 88) {
if (ssDCval < 65) {
if (diffRsCsdiff < -6) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (diffRsCsdiff >= -6) {
if (tcor < 90) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
else if (tcor >= 90) {
return 4;
}
}
}
else if (ssDCval >= 65) {
if (Cs < 880) {
if (Rs < 567) {
if (Rs < 457) {
return 8;
}
else if (Rs >= 457) {
return 4;
}
}
else if (Rs >= 567) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (Cs < 692) {
if (RsCsDiff < 16) {
return 2;
}
else if (RsCsDiff >= 16) {
return 1;
}
}
else if (Cs >= 692) {
return 2;
}
}
}
}
else if (Cs >= 880) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
return 8;
}
}
}
}
}
else if (negBalance >= 36) {
if (Rs < 825) {
if (ssDCval < 56) {
if (Cs < 821) {
if (diffAFD < -1) {
if (Cs < 556) {
return 8;
}
else if (Cs >= 556) {
return 2;
}
}
else if (diffAFD >= -1) {
if (tcor < 80) {
return 8;
}
else if (tcor >= 80) {
if (MVDiff < 17) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
return 2;
}
}
else if (MVDiff >= 17) {
return 8;
}
}
}
}
else if (Cs >= 821) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 2;
}
}
}
else if (ssDCval >= 56) {
if (refDCval < 75) {
if (MVDiff < 5) {
if (Rs < 557) {
return 4;
}
else if (Rs >= 557) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (Cs < 654) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (Cs < 606) {
return 8;
}
else if (Cs >= 606) {
return 2;
}
}
}
else if (Cs >= 654) {
if (diffRsCsdiff < -17) {
return 4;
}
else if (diffRsCsdiff >= -17) {
return 2;
}
}
}
}
else if (refDCval >= 75) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (Rs < 612) {
return 8;
}
else if (Rs >= 612) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 2;
}
}
}
}
}
}
else if (Rs >= 825) {
if (ssDCval < 61) {
if (RsCsDiff < 23) {
if (diffAFD < 5) {
if (Rs < 845) {
if (Rs < 842) {
return 8;
}
else if (Rs >= 842) {
return 2;
}
}
else if (Rs >= 845) {
return 8;
}
}
else if (diffAFD >= 5) {
return 1;
}
}
else if (RsCsDiff >= 23) {
if (refDCval < 47) {
if (SC < 1728) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
return 2;
}
}
else if (SC >= 1728) {
return 8;
}
}
else if (refDCval >= 47) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (diffMVdiffVal < -7) {
if (Rs < 1228) {
return 2;
}
else if (Rs >= 1228) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
if (SC < 1777) {
return 8;
}
else if (SC >= 1777) {
if (diffRsCsdiff < -20) {
return 8;
}
else if (diffRsCsdiff >= -20) {
if (diffRsCsdiff < 17) {
return 4;
}
else if (diffRsCsdiff >= 17) {
if (Cs < 1234) {
return 8;
}
else if (Cs >= 1234) {
return 4;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 61) {
if (ssDCval < 103) {
if (SC < 2617) {
if (CsDiff < 5) {
if (posBalance < 68) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Cs < 655) {
if (diffMVdiffVal < 4) {
if (Rs < 884) {
return 4;
}
else if (Rs >= 884) {
return 1;
}
}
else if (diffMVdiffVal >= 4) {
if (Rs < 1813) {
return 8;
}
else if (Rs >= 1813) {
return 2;
}
}
}
else if (Cs >= 655) {
if (tcor < 79) {
return 1;
}
else if (tcor >= 79) {
if (RsCsDiff < 29) {
if (Rs < 1190) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (Rs < 944) {
return 8;
}
else if (Rs >= 944) {
return 2;
}
}
}
else if (Rs >= 1190) {
if (SC < 2392) {
return 1;
}
else if (SC >= 2392) {
return 8;
}
}
}
else if (RsCsDiff >= 29) {
return 4;
}
}
}
}
}
else if (posBalance >= 68) {
return 8;
}
}
else if (CsDiff >= 5) {
if (tcor < 81) {
if (Rs < 885) {
return 4;
}
else if (Rs >= 885) {
return 8;
}
}
else if (tcor >= 81) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
else if (SC >= 2617) {
if (refDCval < 70) {
return 2;
}
else if (refDCval >= 70) {
if (posBalance < 52) {
if (Cs < 1094) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
else if (Cs >= 1094) {
return 8;
}
}
else if (posBalance >= 52) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (Rs < 2140) {
return 2;
}
else if (Rs >= 2140) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 103) {
if (refDCval < 135) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (refDCval >= 135) {
if (posBalance < 56) {
if (ssDCval < 149) {
return 8;
}
else if (ssDCval >= 149) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (posBalance >= 56) {
if (SC < 2381) {
return 4;
}
else if (SC >= 2381) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 407) {
if (ssDCval < 109) {
return 1;
}
else if (ssDCval >= 109) {
if (Rs < 282) {
return 8;
}
else if (Rs >= 282) {
return 2;
}
}
}
else if (Rs >= 407) {
if (RsCsDiff < 67) {
if (Rs < 935) {
if (CsDiff < 5) {
if (MVDiff < 8) {
if (ssDCval < 62) {
if (ssDCval < 52) {
if (ssDCval < 45) {
if (refDCval < 42) {
if (Rs < 594) {
return 4;
}
else if (Rs >= 594) {
return 8;
}
}
else if (refDCval >= 42) {
return 4;
}
}
else if (ssDCval >= 45) {
if (Cs < 742) {
return 1;
}
else if (Cs >= 742) {
if (Rs < 855) {
return 8;
}
else if (Rs >= 855) {
return 1;
}
}
}
}
else if (ssDCval >= 52) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 4;
}
}
}
else if (ssDCval >= 62) {
if (ssDCval < 96) {
if (negBalance < 46) {
if (Rs < 735) {
return 2;
}
else if (Rs >= 735) {
return 1;
}
}
else if (negBalance >= 46) {
return 8;
}
}
else if (ssDCval >= 96) {
if (refDCval < 111) {
return 1;
}
else if (refDCval >= 111) {
if (MVDiff < 4) {
if (Rs < 755) {
return 2;
}
else if (Rs >= 755) {
return 4;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
}
else if (MVDiff >= 8) {
if (ssDCval < 117) {
if (tcor < 54) {
if (Cs < 634) {
return 8;
}
else if (Cs >= 634) {
return 4;
}
}
else if (tcor >= 54) {
if (RsDiff < 5) {
if (Rs < 785) {
if (MVDiff < 26) {
if (Cs < 879) {
if (diffRsCsdiff < 10) {
if (diffMVdiffVal < 6) {
if (diffMVdiffVal < -8) {
if (Cs < 711) {
return 8;
}
else if (Cs >= 711) {
return 1;
}
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
if (Rs < 725) {
return 8;
}
else if (Rs >= 725) {
return 1;
}
}
}
else if (diffRsCsdiff >= 10) {
if (Rs < 679) {
return 8;
}
else if (Rs >= 679) {
return 1;
}
}
}
else if (Cs >= 879) {
if (Rs < 715) {
return 1;
}
else if (Rs >= 715) {
return 8;
}
}
}
else if (MVDiff >= 26) {
if (diffTSC < -1) {
if (Rs < 441) {
return 8;
}
else if (Rs >= 441) {
if (Rs < 589) {
return 1;
}
else if (Rs >= 589) {
return 4;
}
}
}
else if (diffTSC >= -1) {
if (ssDCval < 48) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (ssDCval >= 48) {
return 8;
}
}
}
}
else if (Rs >= 785) {
if (Cs < 768) {
if (tcor < 66) {
return 1;
}
else if (tcor >= 66) {
if (Rs < 851) {
if (Rs < 836) {
if (ssDCval < 52) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (ssDCval >= 52) {
return 8;
}
}
else if (Rs >= 836) {
if (Cs < 713) {
return 1;
}
else if (Cs >= 713) {
return 4;
}
}
}
else if (Rs >= 851) {
return 8;
}
}
}
else if (Cs >= 768) {
if (diffRsCsdiff < -15) {
if (Rs < 908) {
if (Rs < 844) {
return 4;
}
else if (Rs >= 844) {
return 1;
}
}
else if (Rs >= 908) {
return 8;
}
}
else if (diffRsCsdiff >= -15) {
return 8;
}
}
}
}
else if (RsDiff >= 5) {
if (ssDCval < 74) {
if (TSC < 6) {
if (Rs < 856) {
return 4;
}
else if (Rs >= 856) {
if (Rs < 915) {
return 8;
}
else if (Rs >= 915) {
return 2;
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (ssDCval >= 74) {
return 4;
}
}
}
}
else if (ssDCval >= 117) {
if (ssDCval < 132) {
return 4;
}
else if (ssDCval >= 132) {
if (SC < 1718) {
return 2;
}
else if (SC >= 1718) {
return 1;
}
}
}
}
}
else if (CsDiff >= 5) {
if (negBalance < 49) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (MVDiff < 44) {
return 2;
}
else if (MVDiff >= 44) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (gchDC < 1) {
if (mcTcor < 83) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (mcTcor >= 83) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (Cs < 1252) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
return 8;
}
}
else if (Cs >= 1252) {
return 2;
}
}
}
}
}
else if (Rs >= 935) {
if (posBalance < 54) {
if (Rs < 1476) {
if (ssDCval < 134) {
if (ssDCval < 90) {
if (mcTcor < 94) {
if (SC < 2227) {
if (posBalance < 49) {
if (Cs < 816) {
return 2;
}
else if (Cs >= 816) {
return 8;
}
}
else if (posBalance >= 49) {
return 8;
}
}
else if (SC >= 2227) {
if (Rs < 1408) {
if (Rs < 1173) {
return 4;
}
else if (Rs >= 1173) {
return 8;
}
}
else if (Rs >= 1408) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (posBalance < 49) {
if (Rs < 961) {
return 2;
}
else if (Rs >= 961) {
if (Rs < 1179) {
return 8;
}
else if (Rs >= 1179) {
return 1;
}
}
}
else if (posBalance >= 49) {
if (Rs < 1457) {
return 4;
}
else if (Rs >= 1457) {
return 8;
}
}
}
}
else if (ssDCval >= 90) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
else if (ssDCval >= 134) {
if (Cs < 1109) {
return 4;
}
else if (Cs >= 1109) {
return 8;
}
}
}
else if (Rs >= 1476) {
if (negBalance < 51) {
if (MVDiff < 9) {
if (Rs < 2061) {
return 1;
}
else if (Rs >= 2061) {
return 4;
}
}
else if (MVDiff >= 9) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (Rs < 2252) {
return 8;
}
else if (Rs >= 2252) {
return 4;
}
}
}
}
else if (negBalance >= 51) {
if (ssDCval < 89) {
if (negBalance < 58) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
if (AFD < 11) {
if (diffMVdiffVal < -3) {
if (Rs < 1740) {
return 1;
}
else if (Rs >= 1740) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
if (Rs < 2153) {
return 8;
}
else if (Rs >= 2153) {
return 1;
}
}
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (negBalance >= 58) {
return 8;
}
}
else if (ssDCval >= 89) {
if (ssDCval < 123) {
return 2;
}
else if (ssDCval >= 123) {
if (Rs < 1771) {
return 4;
}
else if (Rs >= 1771) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 54) {
if (diffRsCsdiff < 22) {
if (MVDiff < 12) {
if (refDCval < 69) {
if (Cs < 1173) {
if (ssDCval < 63) {
if (Rs < 989) {
if (Cs < 1004) {
if (Cs < 850) {
return 8;
}
else if (Cs >= 850) {
if (Rs < 958) {
return 8;
}
else if (Rs >= 958) {
return 1;
}
}
}
else if (Cs >= 1004) {
return 4;
}
}
else if (Rs >= 989) {
return 8;
}
}
else if (ssDCval >= 63) {
if (Rs < 1456) {
return 4;
}
else if (Rs >= 1456) {
return 8;
}
}
}
else if (Cs >= 1173) {
if (Rs < 1330) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
else if (Rs >= 1330) {
return 8;
}
}
}
else if (refDCval >= 69) {
if (tcor < 90) {
if (Cs < 648) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
return 8;
}
}
else if (Cs >= 648) {
if (mcTcor < 85) {
if (TSC < 7) {
if (Cs < 1103) {
return 2;
}
else if (Cs >= 1103) {
return 8;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (mcTcor >= 85) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
}
else if (tcor >= 90) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
if (diffRsCsdiff < 12) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (RsCsDiff < 49) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (RsCsDiff >= 49) {
return 8;
}
}
}
else if (diffRsCsdiff >= 12) {
return 1;
}
}
}
}
}
else if (MVDiff >= 12) {
if (refDCval < 70) {
if (CsDiff < 6) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (refDCval >= 70) {
if (ssDCval < 98) {
if (negBalance < 49) {
if (SC < 2123) {
return 4;
}
else if (SC >= 2123) {
if (Rs < 1204) {
return 2;
}
else if (Rs >= 1204) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (SC < 1688) {
if (Cs < 547) {
return 1;
}
else if (Cs >= 547) {
return 8;
}
}
else if (SC >= 1688) {
if (CsDiff < 4) {
if (diffMVdiffVal < -10) {
if (Rs < 1507) {
return 8;
}
else if (Rs >= 1507) {
return 1;
}
}
else if (diffMVdiffVal >= -10) {
if (mcTcor < 84) {
if (Rs < 1555) {
return 8;
}
else if (Rs >= 1555) {
return 2;
}
}
else if (mcTcor >= 84) {
return 2;
}
}
}
else if (CsDiff >= 4) {
if (SC < 2217) {
return 2;
}
else if (SC >= 2217) {
return 8;
}
}
}
}
}
else if (ssDCval >= 98) {
if (Cs < 841) {
if (tcor < 68) {
return 1;
}
else if (tcor >= 68) {
if (RsCsDiff < 60) {
if (refDCval < 138) {
return 8;
}
else if (refDCval >= 138) {
return 4;
}
}
else if (RsCsDiff >= 60) {
return 1;
}
}
}
else if (Cs >= 841) {
if (Rs < 1421) {
if (Cs < 1026) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (Cs >= 1026) {
if (Rs < 1209) {
return 1;
}
else if (Rs >= 1209) {
return 2;
}
}
}
else if (Rs >= 1421) {
if (Rs < 2347) {
return 8;
}
else if (Rs >= 2347) {
return 4;
}
}
}
}
}
}
}
else if (diffRsCsdiff >= 22) {
if (ssDCval < 62) {
if (Rs < 1560) {
return 4;
}
else if (Rs >= 1560) {
return 8;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 67) {
if (Rs < 2234) {
if (Cs < 906) {
if (CsDiff < 3) {
if (Rs < 1965) {
return 2;
}
else if (Rs >= 1965) {
return 1;
}
}
else if (CsDiff >= 3) {
if (diffRsCsdiff < 26) {
if (diffMVdiffVal < -3) {
if (Rs < 1500) {
return 1;
}
else if (Rs >= 1500) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (Rs < 1469) {
return 1;
}
else if (Rs >= 1469) {
if (diffAFD < -6) {
return 1;
}
else if (diffAFD >= -6) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 26) {
if (Rs < 1410) {
return 8;
}
else if (Rs >= 1410) {
return 2;
}
}
}
}
else if (Cs >= 906) {
if (ssDCval < 92) {
if (ssDCval < 64) {
if (Cs < 1202) {
if (posBalance < 62) {
if (Rs < 626) {
return 2;
}
else if (Rs >= 626) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (posBalance >= 62) {
return 4;
}
}
else if (Cs >= 1202) {
if (Rs < 1387) {
return 4;
}
else if (Rs >= 1387) {
return 1;
}
}
}
else if (ssDCval >= 64) {
if (Rs < 2106) {
return 4;
}
else if (Rs >= 2106) {
return 8;
}
}
}
else if (ssDCval >= 92) {
if (negBalance < 50) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
return 8;
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
}
else if (Rs >= 2234) {
if (Cs < 858) {
return 2;
}
else if (Cs >= 858) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (mcTcor < 87) {
if (MVDiff < 20) {
if (refDCval < 52) {
if (mu_mv_mag_sq < 101) {
if (MVDiff < 6) {
if (diffRsCsdiff < 19) {
if (diffRsCsdiff < -10) {
if (Cs < 817) {
return 2;
}
else if (Cs >= 817) {
return 8;
}
}
else if (diffRsCsdiff >= -10) {
return 8;
}
}
else if (diffRsCsdiff >= 19) {
if (Rs < 1690) {
return 2;
}
else if (Rs >= 1690) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (AFD < 10) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (tcor < 73) {
return 8;
}
else if (tcor >= 73) {
if (Rs < 835) {
return 8;
}
else if (Rs >= 835) {
return 1;
}
}
}
}
else if (AFD >= 10) {
if (Rs < 907) {
if (Rs < 689) {
return 4;
}
else if (Rs >= 689) {
return 2;
}
}
else if (Rs >= 907) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (SC < 2113) {
if (posBalance < 58) {
if (Rs < 1065) {
if (Cs < 845) {
if (Rs < 825) {
return 8;
}
else if (Rs >= 825) {
return 1;
}
}
else if (Cs >= 845) {
return 8;
}
}
else if (Rs >= 1065) {
return 1;
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (SC >= 2113) {
return 8;
}
}
}
else if (refDCval >= 52) {
if (tcor < 73) {
if (ssDCval < 98) {
if (mu_mv_mag_sq < 93) {
return 8;
}
else if (mu_mv_mag_sq >= 93) {
return 4;
}
}
else if (ssDCval >= 98) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (tcor >= 73) {
if (refDCval < 64) {
if (negBalance < 51) {
if (posBalance < 63) {
if (Cs < 1148) {
if (Rs < 1080) {
if (Cs < 601) {
return 8;
}
else if (Cs >= 601) {
return 4;
}
}
else if (Rs >= 1080) {
return 8;
}
}
else if (Cs >= 1148) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (negBalance < 47) {
return 4;
}
else if (negBalance >= 47) {
return 8;
}
}
}
else if (negBalance >= 51) {
return 8;
}
}
else if (refDCval >= 64) {
if (Cs < 1227) {
return 8;
}
else if (Cs >= 1227) {
if (mu_mv_mag_sq < 53) {
if (Rs < 1537) {
return 2;
}
else if (Rs >= 1537) {
return 4;
}
}
else if (mu_mv_mag_sq >= 53) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 20) {
if (Rs < 650) {
if (ssDCval < 51) {
if (tcor < 80) {
return 2;
}
else if (tcor >= 80) {
return 8;
}
}
else if (ssDCval >= 51) {
if (SC < 1037) {
return 8;
}
else if (SC >= 1037) {
return 1;
}
}
}
else if (Rs >= 650) {
if (RsCsDiff < 32) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
if (tcor < 79) {
if (Cs < 801) {
return 4;
}
else if (Cs >= 801) {
return 8;
}
}
else if (tcor >= 79) {
return 8;
}
}
}
else if (RsCsDiff >= 32) {
if (diffAFD < 4) {
if (Cs < 1173) {
return 8;
}
else if (Cs >= 1173) {
if (Cs < 1221) {
return 2;
}
else if (Cs >= 1221) {
return 8;
}
}
}
else if (diffAFD >= 4) {
return 1;
}
}
}
}
}
else if (mcTcor >= 87) {
if (Rs < 1576) {
if (ssDCval < 75) {
if (Rs < 839) {
if (ssDCval < 37) {
if (Cs < 607) {
return 8;
}
else if (Cs >= 607) {
return 1;
}
}
else if (ssDCval >= 37) {
if (Rs < 530) {
if (Cs < 745) {
if (RsCsDiff < 7) {
if (Rs < 377) {
if (Rs < 280) {
return 1;
}
else if (Rs >= 280) {
return 8;
}
}
else if (Rs >= 377) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
if (Rs < 402) {
return 4;
}
else if (Rs >= 402) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 7) {
if (tcor < 90) {
if (Rs < 495) {
return 1;
}
else if (Rs >= 495) {
if (Rs < 521) {
return 8;
}
else if (Rs >= 521) {
return 2;
}
}
}
else if (tcor >= 90) {
return 2;
}
}
}
else if (Cs >= 745) {
if (negBalance < 21) {
if (Cs < 925) {
return 4;
}
else if (Cs >= 925) {
return 2;
}
}
else if (negBalance >= 21) {
if (tcor < 95) {
return 1;
}
else if (tcor >= 95) {
if (Rs < 361) {
return 2;
}
else if (Rs >= 361) {
return 1;
}
}
}
}
}
else if (Rs >= 530) {
if (refDCval < 61) {
if (negBalance < 34) {
if (SC < 1456) {
if (Cs < 544) {
return 8;
}
else if (Cs >= 544) {
if (posBalance < 65) {
return 4;
}
else if (posBalance >= 65) {
return 8;
}
}
}
else if (SC >= 1456) {
if (tcor < 96) {
if (CsDiff < 4) {
if (SC < 1628) {
return 1;
}
else if (SC >= 1628) {
if (SC < 1742) {
return 4;
}
else if (SC >= 1742) {
if (Rs < 700) {
return 1;
}
else if (Rs >= 700) {
return 8;
}
}
}
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (tcor >= 96) {
return 8;
}
}
}
else if (negBalance >= 34) {
if (Cs < 616) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
else if (Cs >= 616) {
if (diffAFD < 1) {
if (Rs < 668) {
return 4;
}
else if (Rs >= 668) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (refDCval < 58) {
if (mu_mv_mag_sq < 82) {
if (SC < 1376) {
return 8;
}
else if (SC >= 1376) {
return 4;
}
}
else if (mu_mv_mag_sq >= 82) {
if (Cs < 731) {
return 4;
}
else if (Cs >= 731) {
return 8;
}
}
}
else if (refDCval >= 58) {
return 8;
}
}
}
}
else if (diffAFD >= 1) {
if (Rs < 768) {
return 4;
}
else if (Rs >= 768) {
return 1;
}
}
}
}
}
else if (refDCval >= 61) {
if (mu_mv_mag_sq < 94) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (negBalance < 42) {
if (diffMVdiffVal < -1) {
if (negBalance < 31) {
return 2;
}
else if (negBalance >= 31) {
if (Rs < 794) {
return 1;
}
else if (Rs >= 794) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (negBalance >= 42) {
if (Rs < 807) {
return 8;
}
else if (Rs >= 807) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (Rs < 830) {
return 8;
}
else if (Rs >= 830) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 839) {
if (ssDCval < 59) {
if (Cs < 932) {
if (tcor < 87) {
if (mu_mv_mag_sq < 111) {
if (TSC < 4) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (TSC >= 4) {
if (Rs < 1225) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
}
else if (Rs >= 1225) {
if (refDCval < 56) {
if (Rs < 1348) {
return 8;
}
else if (Rs >= 1348) {
return 2;
}
}
else if (refDCval >= 56) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
return 1;
}
}
else if (tcor >= 87) {
if (negBalance < 28) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (refDCval < 56) {
return 1;
}
else if (refDCval >= 56) {
return 8;
}
}
}
else if (negBalance >= 28) {
if (Rs < 1145) {
if (Cs < 547) {
if (Rs < 1023) {
return 8;
}
else if (Rs >= 1023) {
return 2;
}
}
else if (Cs >= 547) {
return 8;
}
}
else if (Rs >= 1145) {
if (refDCval < 46) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
else if (refDCval >= 46) {
if (negBalance < 39) {
if (AFD < 7) {
if (Rs < 1303) {
return 8;
}
else if (Rs >= 1303) {
return 2;
}
}
else if (AFD >= 7) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (negBalance >= 39) {
return 8;
}
}
}
}
}
}
else if (Cs >= 932) {
if (AFD < 5) {
if (tcor < 87) {
return 1;
}
else if (tcor >= 87) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (AFD >= 5) {
if (Rs < 1298) {
if (RsCsDiff < 22) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
if (Rs < 1127) {
return 4;
}
else if (Rs >= 1127) {
return 8;
}
}
}
else if (RsCsDiff >= 22) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
if (SC < 2335) {
if (MVDiff < 9) {
if (Rs < 899) {
return 8;
}
else if (Rs >= 899) {
return 4;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (SC >= 2335) {
if (mu_mv_mag_sq < 95) {
return 8;
}
else if (mu_mv_mag_sq >= 95) {
return 4;
}
}
}
}
}
else if (Rs >= 1298) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 8;
}
}
}
}
}
else if (ssDCval >= 59) {
if (negBalance < 46) {
if (RsCsDiff < 15) {
if (Rs < 996) {
if (mu_mv_mag_sq < 77) {
if (MVDiff < 13) {
if (Rs < 908) {
return 2;
}
else if (Rs >= 908) {
return 4;
}
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 851) {
return 8;
}
else if (Rs >= 851) {
if (Cs < 555) {
return 8;
}
else if (Cs >= 555) {
return 4;
}
}
}
}
else if (Rs >= 996) {
if (AFD < 4) {
if (posBalance < 50) {
if (Cs < 893) {
return 1;
}
else if (Cs >= 893) {
return 4;
}
}
else if (posBalance >= 50) {
return 1;
}
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 113) {
if (mu_mv_mag_sq < 82) {
if (SC < 2587) {
if (Cs < 610) {
if (Rs < 1113) {
return 8;
}
else if (Rs >= 1113) {
return 4;
}
}
else if (Cs >= 610) {
if (diffMVdiffVal < 8) {
if (MVDiff < 3) {
if (Cs < 917) {
return 1;
}
else if (Cs >= 917) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (diffMVdiffVal >= 8) {
return 1;
}
}
}
else if (SC >= 2587) {
if (Rs < 1523) {
return 1;
}
else if (Rs >= 1523) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (SC < 2518) {
if (negBalance < 36) {
if (Rs < 1399) {
return 8;
}
else if (Rs >= 1399) {
return 1;
}
}
else if (negBalance >= 36) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (Cs < 981) {
return 1;
}
else if (Cs >= 981) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
return 1;
}
}
}
}
}
else if (SC >= 2518) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 117) {
return 8;
}
else if (mu_mv_mag_sq >= 117) {
if (refDCval < 69) {
return 4;
}
else if (refDCval >= 69) {
if (Rs < 1296) {
return 1;
}
else if (Rs >= 1296) {
return 8;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 15) {
if (Cs < 897) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
if (CsDiff < 4) {
if (ssDCval < 62) {
if (Rs < 1349) {
return 1;
}
else if (Rs >= 1349) {
return 4;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (CsDiff >= 4) {
if (Rs < 967) {
return 8;
}
else if (Rs >= 967) {
return 4;
}
}
}
}
else if (Cs >= 897) {
if (negBalance < 44) {
if (refDCval < 63) {
if (RsCsDiff < 33) {
if (Cs < 952) {
return 4;
}
else if (Cs >= 952) {
if (Rs < 1257) {
if (Rs < 1111) {
return 8;
}
else if (Rs >= 1111) {
return 1;
}
}
else if (Rs >= 1257) {
return 8;
}
}
}
else if (RsCsDiff >= 33) {
return 4;
}
}
else if (refDCval >= 63) {
if (Cs < 982) {
return 8;
}
else if (Cs >= 982) {
if (Rs < 1260) {
if (mu_mv_mag_sq < 60) {
if (tcor < 90) {
return 4;
}
else if (tcor >= 90) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 60) {
return 4;
}
}
else if (Rs >= 1260) {
if (Cs < 1253) {
if (RsDiff < 4) {
if (Rs < 1304) {
return 1;
}
else if (Rs >= 1304) {
if (Cs < 1157) {
return 4;
}
else if (Cs >= 1157) {
if (Rs < 1355) {
return 8;
}
else if (Rs >= 1355) {
if (Rs < 1479) {
return 4;
}
else if (Rs >= 1479) {
return 1;
}
}
}
}
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (Cs >= 1253) {
if (Rs < 1394) {
return 1;
}
else if (Rs >= 1394) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 44) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (ssDCval < 73) {
if (Cs < 1062) {
if (Rs < 1188) {
return 8;
}
else if (Rs >= 1188) {
return 4;
}
}
else if (Cs >= 1062) {
return 8;
}
}
else if (ssDCval >= 73) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (posBalance < 57) {
if (MVDiff < 5) {
if (diffMVdiffVal < -1) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
if (Rs < 1486) {
return 4;
}
else if (Rs >= 1486) {
return 8;
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (MVDiff >= 5) {
if (TSC < 7) {
if (Rs < 997) {
if (Cs < 850) {
return 4;
}
else if (Cs >= 850) {
return 1;
}
}
else if (Rs >= 997) {
return 4;
}
}
else if (TSC >= 7) {
if (Rs < 1193) {
return 4;
}
else if (Rs >= 1193) {
return 8;
}
}
}
}
else if (posBalance >= 57) {
if (SC < 1686) {
return 4;
}
else if (SC >= 1686) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (refDCval < 155) {
if (RsCsDiff < 7) {
if (TSC < 6) {
if (MVDiff < 5) {
if (negBalance < 37) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
if (Rs < 344) {
return 2;
}
else if (Rs >= 344) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 629) {
return 2;
}
else if (Rs >= 629) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 37) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (Rs < 795) {
if (mu_mv_mag_sq < 112) {
return 1;
}
else if (mu_mv_mag_sq >= 112) {
if (Rs < 582) {
return 8;
}
else if (Rs >= 582) {
return 1;
}
}
}
else if (Rs >= 795) {
if (negBalance < 41) {
if (Rs < 1336) {
return 8;
}
else if (Rs >= 1336) {
return 1;
}
}
else if (negBalance >= 41) {
return 4;
}
}
}
}
}
else if (MVDiff >= 5) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 96) {
return 1;
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (RsCsDiff >= 7) {
if (SC < 1668) {
if (TSC < 5) {
if (Rs < 642) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
else if (Rs >= 642) {
if (tcor < 93) {
if (Cs < 949) {
return 2;
}
else if (Cs >= 949) {
return 8;
}
}
else if (tcor >= 93) {
return 1;
}
}
}
else if (TSC >= 5) {
if (AFD < 10) {
if (Rs < 817) {
return 8;
}
else if (Rs >= 817) {
if (Rs < 847) {
return 2;
}
else if (Rs >= 847) {
return 8;
}
}
}
else if (AFD >= 10) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (SC >= 1668) {
if (TSC < 6) {
if (MVDiff < 24) {
if (Cs < 721) {
if (Rs < 1402) {
return 4;
}
else if (Rs >= 1402) {
return 8;
}
}
else if (Cs >= 721) {
if (Rs < 999) {
if (ssDCval < 118) {
if (SC < 2008) {
if (tcor < 94) {
if (Cs < 1064) {
if (Rs < 871) {
return 2;
}
else if (Rs >= 871) {
return 8;
}
}
else if (Cs >= 1064) {
return 4;
}
}
else if (tcor >= 94) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (SC >= 2008) {
return 4;
}
}
else if (ssDCval >= 118) {
if (Rs < 995) {
return 8;
}
else if (Rs >= 995) {
return 4;
}
}
}
else if (Rs >= 999) {
if (diffMVdiffVal < -9) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (diffMVdiffVal >= -9) {
return 8;
}
}
}
}
else if (MVDiff >= 24) {
if (ssDCval < 135) {
return 8;
}
else if (ssDCval >= 135) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (AFD < 10) {
if (ssDCval < 102) {
if (gchDC < 1) {
if (SC < 2551) {
return 8;
}
else if (SC >= 2551) {
if (Rs < 1478) {
return 1;
}
else if (Rs >= 1478) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (Rs < 817) {
return 4;
}
else if (Rs >= 817) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
else if (ssDCval >= 102) {
return 8;
}
}
else if (AFD >= 10) {
if (negBalance < 39) {
if (Rs < 1397) {
return 4;
}
else if (Rs >= 1397) {
return 8;
}
}
else if (negBalance >= 39) {
if (diffRsCsdiff < 8) {
if (SC < 1860) {
return 1;
}
else if (SC >= 1860) {
return 8;
}
}
else if (diffRsCsdiff >= 8) {
if (ssDCval < 95) {
return 1;
}
else if (ssDCval >= 95) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 155) {
if (AFD < 11) {
if (Rs < 827) {
return 4;
}
else if (Rs >= 827) {
return 2;
}
}
else if (AFD >= 11) {
return 8;
}
}
}
}
else if (Rs >= 1576) {
if (refDCval < 72) {
if (ssDCval < 62) {
if (MVDiff < 10) {
if (Cs < 745) {
if (Cs < 636) {
return 1;
}
else if (Cs >= 636) {
return 2;
}
}
else if (Cs >= 745) {
if (diffAFD < -2) {
if (Rs < 1693) {
if (Rs < 1637) {
return 1;
}
else if (Rs >= 1637) {
return 8;
}
}
else if (Rs >= 1693) {
return 2;
}
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 117) {
return 8;
}
else if (mu_mv_mag_sq >= 117) {
if (tcor < 90) {
if (Rs < 1933) {
return 2;
}
else if (Rs >= 1933) {
return 4;
}
}
else if (tcor >= 90) {
return 8;
}
}
}
}
}
else if (MVDiff >= 10) {
if (mcTcor < 92) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 8;
}
}
else if (mcTcor >= 92) {
if (Rs < 2000) {
if (Rs < 1689) {
return 1;
}
else if (Rs >= 1689) {
return 8;
}
}
else if (Rs >= 2000) {
return 2;
}
}
}
}
else if (ssDCval >= 62) {
if (RsDiff < 5) {
if (SC < 2626) {
if (tcor < 93) {
if (Rs < 1696) {
return 2;
}
else if (Rs >= 1696) {
return 1;
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (SC >= 2626) {
return 8;
}
}
else if (RsDiff >= 5) {
if (CsDiff < 5) {
if (mcTcor < 92) {
if (ssDCval < 66) {
if (Cs < 866) {
return 4;
}
else if (Cs >= 866) {
return 2;
}
}
else if (ssDCval >= 66) {
if (Rs < 2080) {
return 8;
}
else if (Rs >= 2080) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
if (Rs < 1893) {
return 2;
}
else if (Rs >= 1893) {
return 8;
}
}
}
}
else if (CsDiff >= 5) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
return 8;
}
}
}
}
}
else if (refDCval >= 72) {
if (RsCsDiff < 23) {
if (diffRsCsdiff < -13) {
if (Cs < 1057) {
return 1;
}
else if (Cs >= 1057) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
return 8;
}
}
}
else if (diffRsCsdiff >= -13) {
if (refDCval < 76) {
if (Rs < 1754) {
return 8;
}
else if (Rs >= 1754) {
return 1;
}
}
else if (refDCval >= 76) {
if (mu_mv_mag_sq < 123) {
return 8;
}
else if (mu_mv_mag_sq >= 123) {
if (ssDCval < 112) {
return 8;
}
else if (ssDCval >= 112) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 23) {
if (RsCsDiff < 48) {
if (tcor < 92) {
if (RsDiff < 6) {
if (gchDC < 4) {
if (tcor < 82) {
if (Rs < 1987) {
return 2;
}
else if (Rs >= 1987) {
return 8;
}
}
else if (tcor >= 82) {
return 8;
}
}
else if (gchDC >= 4) {
return 1;
}
}
else if (RsDiff >= 6) {
if (ssDCval < 110) {
if (Rs < 1847) {
return 2;
}
else if (Rs >= 1847) {
return 1;
}
}
else if (ssDCval >= 110) {
return 8;
}
}
}
else if (tcor >= 92) {
if (refDCval < 78) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (Cs < 1303) {
return 8;
}
else if (Cs >= 1303) {
return 1;
}
}
}
else if (refDCval >= 78) {
if (posBalance < 55) {
if (Rs < 2273) {
if (AFD < 7) {
if (ssDCval < 106) {
if (ssDCval < 84) {
if (Rs < 2034) {
return 8;
}
else if (Rs >= 2034) {
return 1;
}
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (ssDCval >= 106) {
if (negBalance < 27) {
return 2;
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (AFD >= 7) {
return 8;
}
}
else if (Rs >= 2273) {
if (Cs < 1138) {
return 8;
}
else if (Cs >= 1138) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (mu_mv_mag_sq < 56) {
if (Rs < 2190) {
return 8;
}
else if (Rs >= 2190) {
return 2;
}
}
else if (mu_mv_mag_sq >= 56) {
if (RsCsDiff < 33) {
if (posBalance < 57) {
if (Cs < 1295) {
return 1;
}
else if (Cs >= 1295) {
return 8;
}
}
else if (posBalance >= 57) {
if (Cs < 1167) {
return 8;
}
else if (Cs >= 1167) {
return 2;
}
}
}
else if (RsCsDiff >= 33) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 48) {
if (ssDCval < 135) {
if (tcor < 94) {
if (diffRsCsdiff < -27) {
return 2;
}
else if (diffRsCsdiff >= -27) {
if (TSC < 7) {
if (RsCsDiff < 57) {
if (Cs < 1308) {
return 8;
}
else if (Cs >= 1308) {
return 2;
}
}
else if (RsCsDiff >= 57) {
if (Cs < 1217) {
if (Rs < 2060) {
return 8;
}
else if (Rs >= 2060) {
return 1;
}
}
else if (Cs >= 1217) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (TSC >= 7) {
if (negBalance < 48) {
if (Rs < 1783) {
if (ssDCval < 101) {
if (Cs < 1042) {
return 8;
}
else if (Cs >= 1042) {
return 2;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (Rs >= 1783) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (SC < 3067) {
return 8;
}
else if (SC >= 3067) {
return 2;
}
}
}
}
else if (negBalance >= 48) {
if (ssDCval < 106) {
return 8;
}
else if (ssDCval >= 106) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
}
}
}
}
else if (tcor >= 94) {
if (MVDiff < 6) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
else if (MVDiff >= 6) {
if (Rs < 2089) {
return 4;
}
else if (Rs >= 2089) {
return 8;
}
}
}
}
else if (ssDCval >= 135) {
return 8;
}
}
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (ssDCval < 114) {
if (mcTcor < 88) {
if (refDCval < 95) {
if (Rs < 1435) {
if (CsDiff < 6) {
if (refDCval < 88) {
if (Rs < 995) {
if (negBalance < 55) {
if (posBalance < 63) {
if (ssDCval < 57) {
if (AFD < 16) {
if (Cs < 958) {
return 4;
}
else if (Cs >= 958) {
return 2;
}
}
else if (AFD >= 16) {
if (Rs < 595) {
return 1;
}
else if (Rs >= 595) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (CsDiff < 4) {
if (Rs < 982) {
return 4;
}
else if (Rs >= 982) {
return 8;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
}
else if (posBalance >= 63) {
if (AFD < 23) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
if (Rs < 666) {
if (MVDiff < 58) {
return 4;
}
else if (MVDiff >= 58) {
if (Cs < 568) {
return 4;
}
else if (Cs >= 568) {
return 8;
}
}
}
else if (Rs >= 666) {
return 8;
}
}
}
else if (AFD >= 23) {
return 1;
}
}
}
else if (negBalance >= 55) {
if (negBalance < 63) {
if (tcor < 52) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (tcor >= 52) {
if (diffTSC < 1) {
if (diffRsCsdiff < 10) {
if (diffMVdiffVal < -6) {
return 4;
}
else if (diffMVdiffVal >= -6) {
if (Cs < 567) {
return 4;
}
else if (Cs >= 567) {
return 8;
}
}
}
else if (diffRsCsdiff >= 10) {
return 2;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (negBalance >= 63) {
if (Rs < 259) {
if (Rs < 191) {
return 4;
}
else if (Rs >= 191) {
return 2;
}
}
else if (Rs >= 259) {
return 8;
}
}
}
}
else if (Rs >= 995) {
if (ssDCval < 64) {
if (diffRsCsdiff < -5) {
if (Rs < 1123) {
return 1;
}
else if (Rs >= 1123) {
return 4;
}
}
else if (diffRsCsdiff >= -5) {
if (SC < 2356) {
if (ssDCval < 62) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (Rs < 1077) {
return 8;
}
else if (Rs >= 1077) {
return 4;
}
}
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (SC >= 2356) {
if (Rs < 1287) {
return 1;
}
else if (Rs >= 1287) {
return 4;
}
}
}
}
else if (ssDCval >= 64) {
if (RsDiff < 10) {
if (tcor < 69) {
return 8;
}
else if (tcor >= 69) {
if (RsCsDiff < 46) {
return 8;
}
else if (RsCsDiff >= 46) {
if (Cs < 988) {
if (Rs < 1207) {
return 4;
}
else if (Rs >= 1207) {
return 8;
}
}
else if (Cs >= 988) {
return 2;
}
}
}
}
else if (RsDiff >= 10) {
return 4;
}
}
}
}
else if (refDCval >= 88) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 740) {
if (Rs < 501) {
return 4;
}
else if (Rs >= 501) {
return 2;
}
}
else if (Rs >= 740) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (TSC < 12) {
if (TSC < 7) {
if (Rs < 762) {
return 4;
}
else if (Rs >= 762) {
if (Rs < 932) {
return 8;
}
else if (Rs >= 932) {
return 1;
}
}
}
else if (TSC >= 7) {
if (diffMVdiffVal < 25) {
if (refDCval < 42) {
return 8;
}
else if (refDCval >= 42) {
if (mcTcor < 81) {
if (negBalance < 51) {
if (ssDCval < 61) {
if (RsCsDiff < 74) {
return 2;
}
else if (RsCsDiff >= 74) {
if (Rs < 781) {
return 4;
}
else if (Rs >= 781) {
if (Rs < 1228) {
return 2;
}
else if (Rs >= 1228) {
return 1;
}
}
}
}
else if (ssDCval >= 61) {
if (Cs < 1100) {
return 4;
}
else if (Cs >= 1100) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (refDCval < 63) {
if (CsDiff < 9) {
return 4;
}
else if (CsDiff >= 9) {
if (Rs < 707) {
return 8;
}
else if (Rs >= 707) {
return 4;
}
}
}
else if (refDCval >= 63) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (Cs < 693) {
if (Rs < 757) {
return 1;
}
else if (Rs >= 757) {
return 4;
}
}
else if (Cs >= 693) {
return 2;
}
}
}
}
}
else if (mcTcor >= 81) {
if (ssDCval < 63) {
if (RsCsDiff < 74) {
if (ssDCval < 50) {
return 2;
}
else if (ssDCval >= 50) {
return 8;
}
}
else if (RsCsDiff >= 74) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Cs < 1193) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 4;
}
}
else if (Cs >= 1193) {
return 4;
}
}
}
}
}
else if (ssDCval >= 63) {
if (SC < 2482) {
if (MVDiff < 22) {
if (Rs < 1306) {
return 1;
}
else if (Rs >= 1306) {
return 4;
}
}
else if (MVDiff >= 22) {
if (RsCsDiff < 167) {
if (Rs < 509) {
return 8;
}
else if (Rs >= 509) {
if (diffTSC < 2) {
return 2;
}
else if (diffTSC >= 2) {
return 1;
}
}
}
else if (RsCsDiff >= 167) {
return 8;
}
}
}
else if (SC >= 2482) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
return 4;
}
}
}
}
}
}
else if (diffMVdiffVal >= 25) {
if (MVDiff < 158) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
if (Rs < 843) {
return 4;
}
else if (Rs >= 843) {
return 8;
}
}
}
else if (MVDiff >= 158) {
return 1;
}
}
}
}
else if (TSC >= 12) {
if (RsDiff < 7) {
if (CsDiff < 8) {
if (refDCval < 59) {
return 2;
}
else if (refDCval >= 59) {
if (diffAFD < -1) {
if (Rs < 963) {
return 8;
}
else if (Rs >= 963) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (CsDiff >= 8) {
return 2;
}
}
else if (RsDiff >= 7) {
if (RsCsDiff < 153) {
if (diffMVdiffVal < -9) {
return 1;
}
else if (diffMVdiffVal >= -9) {
return 4;
}
}
else if (RsCsDiff >= 153) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1435) {
if (RsCsDiff < 103) {
if (SC < 2176) {
return 4;
}
else if (SC >= 2176) {
if (Rs < 2296) {
if (negBalance < 33) {
return 1;
}
else if (negBalance >= 33) {
if (diffTSC < -1) {
if (ssDCval < 66) {
return 2;
}
else if (ssDCval >= 66) {
if (Rs < 1452) {
if (Rs < 1448) {
return 8;
}
else if (Rs >= 1448) {
return 4;
}
}
else if (Rs >= 1452) {
return 8;
}
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (Rs >= 2296) {
return 4;
}
}
}
else if (RsCsDiff >= 103) {
if (refDCval < 91) {
if (negBalance < 48) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
if (gchDC < 2) {
if (Rs < 1640) {
return 8;
}
else if (Rs >= 1640) {
return 2;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (SC < 2842) {
if (diffMVdiffVal < -6) {
if (Cs < 934) {
return 2;
}
else if (Cs >= 934) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
if (mu_mv_mag_sq < 25) {
if (Cs < 913) {
return 8;
}
else if (Cs >= 913) {
if (Cs < 1198) {
return 4;
}
else if (Cs >= 1198) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
}
}
else if (SC >= 2842) {
if (AFD < 22) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (Rs < 1731) {
return 8;
}
else if (Rs >= 1731) {
return 4;
}
}
}
else if (AFD >= 22) {
if (SC < 3180) {
if (diffRsCsdiff < 2) {
if (Rs < 2072) {
return 2;
}
else if (Rs >= 2072) {
return 4;
}
}
else if (diffRsCsdiff >= 2) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
return 2;
}
}
}
else if (SC >= 3180) {
if (Cs < 1212) {
return 1;
}
else if (Cs >= 1212) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 91) {
if (Rs < 2113) {
return 8;
}
else if (Rs >= 2113) {
return 1;
}
}
}
}
}
else if (refDCval >= 95) {
if (tcor < 74) {
if (gchDC < 2) {
if (TSC < 14) {
if (Rs < 2302) {
return 4;
}
else if (Rs >= 2302) {
if (Rs < 2345) {
return 8;
}
else if (Rs >= 2345) {
return 4;
}
}
}
else if (TSC >= 14) {
if (MVDiff < 141) {
return 1;
}
else if (MVDiff >= 141) {
if (Rs < 1381) {
return 4;
}
else if (Rs >= 1381) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
if (negBalance < 54) {
if (MVDiff < 106) {
return 8;
}
else if (MVDiff >= 106) {
return 2;
}
}
else if (negBalance >= 54) {
if (diffRsCsdiff < -36) {
return 8;
}
else if (diffRsCsdiff >= -36) {
if (MVDiff < 232) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 4;
}
}
else if (MVDiff >= 232) {
return 8;
}
}
}
}
}
}
else if (tcor >= 74) {
if (RsDiff < 7) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
return 4;
}
}
else if (RsDiff >= 7) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
if (Rs < 2339) {
return 2;
}
else if (Rs >= 2339) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 88) {
if (Rs < 1787) {
if (AFD < 26) {
if (SC < 2614) {
if (posBalance < 50) {
if (negBalance < 66) {
if (MVDiff < 45) {
if (mcTcor < 90) {
if (MVDiff < 25) {
return 2;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (mcTcor >= 90) {
if (diffMVdiffVal < -4) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (Rs < 1170) {
if (Rs < 927) {
return 4;
}
else if (Rs >= 927) {
return 2;
}
}
else if (Rs >= 1170) {
return 1;
}
}
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
}
else if (MVDiff >= 45) {
if (diffMVdiffVal < -10) {
if (Cs < 832) {
return 1;
}
else if (Cs >= 832) {
return 2;
}
}
else if (diffMVdiffVal >= -10) {
if (Rs < 616) {
return 1;
}
else if (Rs >= 616) {
if (Rs < 1579) {
return 4;
}
else if (Rs >= 1579) {
return 8;
}
}
}
}
}
else if (negBalance >= 66) {
if (Rs < 556) {
if (Rs < 370) {
return 4;
}
else if (Rs >= 370) {
return 1;
}
}
else if (Rs >= 556) {
if (diffAFD < -8) {
if (Cs < 1048) {
return 4;
}
else if (Cs >= 1048) {
return 2;
}
}
else if (diffAFD >= -8) {
if (diffRsCsdiff < 14) {
if (diffRsCsdiff < -47) {
if (Rs < 1245) {
return 4;
}
else if (Rs >= 1245) {
return 8;
}
}
else if (diffRsCsdiff >= -47) {
return 8;
}
}
else if (diffRsCsdiff >= 14) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 8) {
if (CsDiff < 6) {
if (Cs < 950) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
return 4;
}
}
else if (Cs >= 950) {
return 4;
}
}
else if (CsDiff >= 6) {
if (Rs < 1298) {
if (negBalance < 51) {
return 1;
}
else if (negBalance >= 51) {
return 8;
}
}
else if (Rs >= 1298) {
return 4;
}
}
}
else if (MVDiff >= 8) {
if (MVDiff < 17) {
if (ssDCval < 75) {
if (diffMVdiffVal < -3) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
if (Rs < 1302) {
return 2;
}
else if (Rs >= 1302) {
return 8;
}
}
}
else if (diffMVdiffVal >= -3) {
if (AFD < 14) {
if (Rs < 1360) {
return 8;
}
else if (Rs >= 1360) {
return 2;
}
}
else if (AFD >= 14) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
return 2;
}
}
}
}
else if (ssDCval >= 75) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
if (tcor < 81) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
}
else if (tcor >= 81) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 2;
}
}
}
}
}
else if (MVDiff >= 17) {
if (Cs < 972) {
if (RsCsDiff < 77) {
if (MVDiff < 42) {
if (negBalance < 37) {
if (Rs < 894) {
return 8;
}
else if (Rs >= 894) {
return 1;
}
}
else if (negBalance >= 37) {
if (Rs < 789) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (Cs < 870) {
return 4;
}
else if (Cs >= 870) {
return 1;
}
}
}
}
else if (Rs >= 789) {
if (tcor < 85) {
if (Cs < 665) {
if (gchDC < 4) {
if (MVDiff < 36) {
return 1;
}
else if (MVDiff >= 36) {
return 4;
}
}
else if (gchDC >= 4) {
return 4;
}
}
else if (Cs >= 665) {
return 4;
}
}
else if (tcor >= 85) {
return 1;
}
}
}
}
else if (MVDiff >= 42) {
if (tcor < 81) {
if (ssDCval < 79) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (Cs < 898) {
return 1;
}
else if (Cs >= 898) {
return 4;
}
}
}
else if (ssDCval >= 79) {
if (ssDCval < 88) {
return 8;
}
else if (ssDCval >= 88) {
if (Rs < 1321) {
return 2;
}
else if (Rs >= 1321) {
return 1;
}
}
}
}
else if (tcor >= 81) {
return 2;
}
}
}
else if (RsCsDiff >= 77) {
if (negBalance < 49) {
if (Rs < 901) {
return 1;
}
else if (Rs >= 901) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (AFD < 21) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (AFD >= 21) {
if (Rs < 1284) {
return 8;
}
else if (Rs >= 1284) {
if (Rs < 1568) {
return 4;
}
else if (Rs >= 1568) {
return 1;
}
}
}
}
}
}
else if (Cs >= 972) {
if (diffTSC < 1) {
if (tcor < 78) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
if (Rs < 826) {
return 2;
}
else if (Rs >= 826) {
return 4;
}
}
}
}
else if (tcor >= 78) {
if (AFD < 17) {
if (diffRsCsdiff < -5) {
if (Rs < 774) {
return 2;
}
else if (Rs >= 774) {
if (Rs < 1247) {
return 1;
}
else if (Rs >= 1247) {
return 8;
}
}
}
else if (diffRsCsdiff >= -5) {
return 2;
}
}
else if (AFD >= 17) {
return 1;
}
}
}
else if (diffTSC >= 1) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (posBalance < 63) {
if (Cs < 1054) {
return 2;
}
else if (Cs >= 1054) {
return 4;
}
}
else if (posBalance >= 63) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mu_mv_mag_sq < 65) {
if (refDCval < 88) {
if (MVDiff < 14) {
if (negBalance < 49) {
if (mu_mv_mag_sq < 44) {
return 1;
}
else if (mu_mv_mag_sq >= 44) {
return 8;
}
}
else if (negBalance >= 49) {
if (diffTSC < 1) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (posBalance < 64) {
if (negBalance < 57) {
if (SC < 2504) {
if (mu_mv_mag_sq < 19) {
return 8;
}
else if (mu_mv_mag_sq >= 19) {
if (Rs < 855) {
return 1;
}
else if (Rs >= 855) {
return 8;
}
}
}
else if (SC >= 2504) {
return 1;
}
}
else if (negBalance >= 57) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (SC < 2534) {
return 2;
}
else if (SC >= 2534) {
return 8;
}
}
}
}
else if (posBalance >= 64) {
if (tcor < 74) {
return 8;
}
else if (tcor >= 74) {
return 4;
}
}
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (MVDiff >= 14) {
if (mu_mv_mag_sq < 10) {
if (diffMVdiffVal < -19) {
if (CsDiff < 6) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (diffMVdiffVal >= -19) {
if (diffMVdiffVal < 3) {
if (AFD < 14) {
if (CsDiff < 5) {
if (Rs < 893) {
return 1;
}
else if (Rs >= 893) {
return 4;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (AFD >= 14) {
if (diffAFD < 4) {
if (Cs < 867) {
return 4;
}
else if (Cs >= 867) {
if (diffMVdiffVal < -12) {
if (Cs < 1063) {
return 1;
}
else if (Cs >= 1063) {
return 4;
}
}
else if (diffMVdiffVal >= -12) {
if (posBalance < 64) {
return 4;
}
else if (posBalance >= 64) {
if (Rs < 1218) {
return 1;
}
else if (Rs >= 1218) {
if (Rs < 1411) {
return 4;
}
else if (Rs >= 1411) {
return 8;
}
}
}
}
}
}
else if (diffAFD >= 4) {
return 1;
}
}
}
else if (diffMVdiffVal >= 3) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (MVDiff < 51) {
if (SC < 1911) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
return 8;
}
}
}
else if (SC >= 1911) {
if (RsCsDiff < 57) {
return 8;
}
else if (RsCsDiff >= 57) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (Cs < 709) {
return 8;
}
else if (Cs >= 709) {
return 4;
}
}
}
}
}
else if (MVDiff >= 51) {
if (Cs < 661) {
return 8;
}
else if (Cs >= 661) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (SC < 2042) {
if (Cs < 820) {
if (Rs < 1201) {
if (ssDCval < 66) {
if (negBalance < 46) {
if (mu_mv_mag_sq < 23) {
if (Cs < 664) {
return 8;
}
else if (Cs >= 664) {
return 4;
}
}
else if (mu_mv_mag_sq >= 23) {
return 1;
}
}
else if (negBalance >= 46) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 4;
}
}
}
}
else if (ssDCval >= 66) {
if (Rs < 567) {
if (SC < 1116) {
return 4;
}
else if (SC >= 1116) {
return 2;
}
}
else if (Rs >= 567) {
if (mu_mv_mag_sq < 42) {
if (gchDC < 6) {
if (SC < 1686) {
return 1;
}
else if (SC >= 1686) {
if (Rs < 1033) {
return 2;
}
else if (Rs >= 1033) {
return 4;
}
}
}
else if (gchDC >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 42) {
if (SC < 1431) {
return 4;
}
else if (SC >= 1431) {
return 8;
}
}
}
}
}
else if (Rs >= 1201) {
if (RsCsDiff < 93) {
return 2;
}
else if (RsCsDiff >= 93) {
return 4;
}
}
}
else if (Cs >= 820) {
if (posBalance < 72) {
if (SC < 1902) {
if (Cs < 983) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
return 2;
}
}
}
else if (Cs >= 983) {
return 8;
}
}
else if (SC >= 1902) {
return 2;
}
}
else if (posBalance >= 72) {
return 4;
}
}
}
else if (SC >= 2042) {
if (ssDCval < 79) {
if (Rs < 1554) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
if (SC < 2323) {
return 8;
}
else if (SC >= 2323) {
if (Rs < 1317) {
return 4;
}
else if (Rs >= 1317) {
return 1;
}
}
}
}
else if (Rs >= 1554) {
return 4;
}
}
else if (ssDCval >= 79) {
if (Cs < 1100) {
return 4;
}
else if (Cs >= 1100) {
if (Rs < 1261) {
return 8;
}
else if (Rs >= 1261) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 88) {
if (tcor < 62) {
if (Rs < 923) {
return 1;
}
else if (Rs >= 923) {
return 4;
}
}
else if (tcor >= 62) {
if (Cs < 820) {
if (tcor < 76) {
if (refDCval < 104) {
return 8;
}
else if (refDCval >= 104) {
return 1;
}
}
else if (tcor >= 76) {
if (negBalance < 47) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
else if (negBalance >= 47) {
if (MVDiff < 21) {
return 2;
}
else if (MVDiff >= 21) {
if (gchDC < 2) {
if (Cs < 559) {
return 8;
}
else if (Cs >= 559) {
return 1;
}
}
else if (gchDC >= 2) {
if (diffAFD < -3) {
if (Rs < 848) {
return 1;
}
else if (Rs >= 848) {
return 2;
}
}
else if (diffAFD >= -3) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
if (Cs < 685) {
return 2;
}
else if (Cs >= 685) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 820) {
if (AFD < 20) {
if (negBalance < 48) {
if (Rs < 886) {
if (Rs < 670) {
return 2;
}
else if (Rs >= 670) {
return 4;
}
}
else if (Rs >= 886) {
if (Rs < 1259) {
return 1;
}
else if (Rs >= 1259) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (Cs < 881) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (Cs >= 881) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 715) {
return 2;
}
else if (Rs >= 715) {
return 8;
}
}
}
}
}
else if (AFD >= 20) {
if (RsCsDiff < 115) {
if (Cs < 856) {
return 8;
}
else if (Cs >= 856) {
return 1;
}
}
else if (RsCsDiff >= 115) {
if (Rs < 994) {
if (Rs < 908) {
return 2;
}
else if (Rs >= 908) {
return 1;
}
}
else if (Rs >= 994) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 65) {
if (posBalance < 64) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (posBalance >= 64) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 8;
}
}
}
}
}
}
else if (SC >= 2614) {
if (AFD < 19) {
if (TSC < 9) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 92) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 4;
}
}
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffAFD < -4) {
if (Cs < 1226) {
if (Rs < 1626) {
return 1;
}
else if (Rs >= 1626) {
return 8;
}
}
else if (Cs >= 1226) {
return 4;
}
}
else if (diffAFD >= -4) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
if (TSC < 7) {
if (Cs < 1290) {
return 2;
}
else if (Cs >= 1290) {
return 8;
}
}
else if (TSC >= 7) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
return 1;
}
}
}
}
}
}
else if (TSC >= 9) {
if (Rs < 1780) {
if (Cs < 982) {
return 4;
}
else if (Cs >= 982) {
return 8;
}
}
else if (Rs >= 1780) {
return 1;
}
}
}
else if (AFD >= 19) {
if (RsDiff < 9) {
if (posBalance < 65) {
return 1;
}
else if (posBalance >= 65) {
return 8;
}
}
else if (RsDiff >= 9) {
if (MVDiff < 18) {
if (SC < 2645) {
return 8;
}
else if (SC >= 2645) {
return 4;
}
}
else if (MVDiff >= 18) {
if (mu_mv_mag_sq < 7) {
if (Rs < 1649) {
return 4;
}
else if (Rs >= 1649) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 1321) {
return 8;
}
else if (Cs >= 1321) {
return 1;
}
}
}
}
}
}
}
else if (AFD >= 26) {
if (MVDiff < 112) {
if (RsCsDiff < 148) {
if (TSC < 9) {
if (refDCval < 83) {
if (SC < 1986) {
return 1;
}
else if (SC >= 1986) {
return 2;
}
}
else if (refDCval >= 83) {
if (Rs < 911) {
if (MVDiff < 45) {
return 1;
}
else if (MVDiff >= 45) {
return 4;
}
}
else if (Rs >= 911) {
return 4;
}
}
}
else if (TSC >= 9) {
if (Rs < 1238) {
if (mcTcor < 91) {
return 2;
}
else if (mcTcor >= 91) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
if (Rs < 543) {
return 1;
}
else if (Rs >= 543) {
return 2;
}
}
}
}
else if (Rs >= 1238) {
if (AFD < 30) {
return 1;
}
else if (AFD >= 30) {
return 2;
}
}
}
}
else if (RsCsDiff >= 148) {
if (mu_mv_mag_sq < 15) {
if (diffMVdiffVal < 12) {
return 2;
}
else if (diffMVdiffVal >= 12) {
return 4;
}
}
else if (mu_mv_mag_sq >= 15) {
if (diffAFD < 4) {
return 4;
}
else if (diffAFD >= 4) {
return 2;
}
}
}
}
else if (MVDiff >= 112) {
if (posBalance < 62) {
if (Rs < 646) {
if (Rs < 459) {
return 1;
}
else if (Rs >= 459) {
return 8;
}
}
else if (Rs >= 646) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 2;
}
}
}
else if (posBalance >= 62) {
if (Rs < 746) {
return 1;
}
else if (Rs >= 746) {
return 8;
}
}
}
}
}
else if (Rs >= 1787) {
if (ssDCval < 72) {
if (Cs < 1005) {
if (tcor < 60) {
if (Rs < 2130) {
if (Rs < 1926) {
return 8;
}
else if (Rs >= 1926) {
return 4;
}
}
else if (Rs >= 2130) {
return 2;
}
}
else if (tcor >= 60) {
return 2;
}
}
else if (Cs >= 1005) {
if (posBalance < 46) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
else if (posBalance >= 46) {
if (SC < 3306) {
return 4;
}
else if (SC >= 3306) {
return 8;
}
}
}
}
else if (ssDCval >= 72) {
if (Cs < 875) {
if (MVDiff < 40) {
if (ssDCval < 86) {
if (diffMVdiffVal < -15) {
return 8;
}
else if (diffMVdiffVal >= -15) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 2;
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 101) {
if (ssDCval < 90) {
if (mu_mv_mag_sq < 11) {
if (diffRsCsdiff < -5) {
if (Rs < 2334) {
return 2;
}
else if (Rs >= 2334) {
return 4;
}
}
else if (diffRsCsdiff >= -5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 11) {
return 4;
}
}
else if (ssDCval >= 90) {
return 1;
}
}
else if (ssDCval >= 101) {
if (Cs < 801) {
return 2;
}
else if (Cs >= 801) {
return 1;
}
}
}
}
else if (MVDiff >= 40) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Rs < 2292) {
return 2;
}
else if (Rs >= 2292) {
return 8;
}
}
}
}
else if (Cs >= 875) {
if (tcor < 85) {
if (ssDCval < 92) {
if (RsCsDiff < 289) {
if (posBalance < 49) {
if (tcor < 78) {
return 2;
}
else if (tcor >= 78) {
return 8;
}
}
else if (posBalance >= 49) {
if (posBalance < 71) {
if (Rs < 2105) {
if (mcTcor < 94) {
if (MVDiff < 31) {
if (ssDCval < 79) {
if (RsCsDiff < 157) {
return 8;
}
else if (RsCsDiff >= 157) {
if (Rs < 1819) {
return 8;
}
else if (Rs >= 1819) {
return 4;
}
}
}
else if (ssDCval >= 79) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
if (AFD < 22) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 4;
}
}
else if (AFD >= 22) {
return 4;
}
}
}
}
else if (MVDiff >= 31) {
return 8;
}
}
else if (mcTcor >= 94) {
return 4;
}
}
else if (Rs >= 2105) {
if (ssDCval < 88) {
if (RsDiff < 13) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (RsDiff >= 13) {
if (Rs < 2171) {
return 4;
}
else if (Rs >= 2171) {
return 2;
}
}
}
else if (ssDCval >= 88) {
return 8;
}
}
}
else if (posBalance >= 71) {
return 2;
}
}
}
else if (RsCsDiff >= 289) {
return 2;
}
}
else if (ssDCval >= 92) {
if (Cs < 1080) {
if (Rs < 2027) {
return 1;
}
else if (Rs >= 2027) {
if (MVDiff < 33) {
return 2;
}
else if (MVDiff >= 33) {
return 8;
}
}
}
else if (Cs >= 1080) {
if (mcTcor < 92) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 1996) {
return 8;
}
else if (Rs >= 1996) {
return 4;
}
}
}
else if (mcTcor >= 92) {
return 4;
}
}
}
}
else if (tcor >= 85) {
if (SC < 2925) {
return 1;
}
else if (SC >= 2925) {
if (TSC < 10) {
if (diffMVdiffVal < 0) {
if (Rs < 1978) {
return 1;
}
else if (Rs >= 1978) {
return 4;
}
}
else if (diffMVdiffVal >= 0) {
if (Cs < 1101) {
if (Rs < 2170) {
return 4;
}
else if (Rs >= 2170) {
return 1;
}
}
else if (Cs >= 1101) {
return 8;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 114) {
if (RsDiff < 8) {
if (SC < 1649) {
if (mcTcor < 91) {
if (MVDiff < 111) {
if (Rs < 966) {
return 2;
}
else if (Rs >= 966) {
return 4;
}
}
else if (MVDiff >= 111) {
if (posBalance < 68) {
return 1;
}
else if (posBalance >= 68) {
if (Rs < 555) {
return 8;
}
else if (Rs >= 555) {
if (Rs < 749) {
return 2;
}
else if (Rs >= 749) {
return 4;
}
}
}
}
}
else if (mcTcor >= 91) {
if (AFD < 20) {
if (Rs < 498) {
if (Rs < 493) {
return 2;
}
else if (Rs >= 493) {
return 4;
}
}
else if (Rs >= 498) {
if (ssDCval < 164) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
else if (ssDCval >= 164) {
return 2;
}
}
}
else if (AFD >= 20) {
if (Rs < 874) {
return 1;
}
else if (Rs >= 874) {
return 2;
}
}
}
}
else if (SC >= 1649) {
if (MVDiff < 12) {
if (gchDC < 5) {
return 8;
}
else if (gchDC >= 5) {
return 2;
}
}
else if (MVDiff >= 12) {
if (CsDiff < 5) {
if (ssDCval < 139) {
return 8;
}
else if (ssDCval >= 139) {
if (diffMVdiffVal < 1) {
if (SC < 1928) {
return 8;
}
else if (SC >= 1928) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
if (Rs < 867) {
return 1;
}
else if (Rs >= 867) {
return 8;
}
}
}
}
else if (CsDiff >= 5) {
if (posBalance < 54) {
if (Rs < 1508) {
return 2;
}
else if (Rs >= 1508) {
return 8;
}
}
else if (posBalance >= 54) {
if (negBalance < 56) {
if (TSC < 10) {
if (RsDiff < 6) {
if (SC < 1928) {
if (Rs < 912) {
return 8;
}
else if (Rs >= 912) {
return 2;
}
}
else if (SC >= 1928) {
return 1;
}
}
else if (RsDiff >= 6) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Rs < 1357) {
return 2;
}
else if (Rs >= 1357) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (Cs < 897) {
return 8;
}
else if (Cs >= 897) {
return 1;
}
}
}
else if (negBalance >= 56) {
if (tcor < 82) {
if (Cs < 825) {
return 1;
}
else if (Cs >= 825) {
return 2;
}
}
else if (tcor >= 82) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (tcor < 69) {
if (mu_mv_mag_sq < 162) {
if (AFD < 35) {
if (gchDC < 13) {
if (tcor < 56) {
if (RsDiff < 12) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (CsDiff < 12) {
return 1;
}
else if (CsDiff >= 12) {
return 8;
}
}
}
}
else if (RsDiff >= 12) {
if (Rs < 2326) {
return 8;
}
else if (Rs >= 2326) {
return 1;
}
}
}
else if (tcor >= 56) {
if (AFD < 26) {
if (TSC < 12) {
if (TSC < 7) {
if (Rs < 1112) {
return 4;
}
else if (Rs >= 1112) {
return 8;
}
}
else if (TSC >= 7) {
if (Cs < 759) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
return 4;
}
}
else if (Cs >= 759) {
if (ssDCval < 150) {
return 4;
}
else if (ssDCval >= 150) {
return 2;
}
}
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (AFD >= 26) {
if (Cs < 811) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
else if (Cs >= 811) {
return 8;
}
}
}
}
else if (gchDC >= 13) {
if (Rs < 1170) {
return 2;
}
else if (Rs >= 1170) {
return 4;
}
}
}
else if (AFD >= 35) {
if (diffAFD < 6) {
return 2;
}
else if (diffAFD >= 6) {
if (Rs < 1447) {
return 4;
}
else if (Rs >= 1447) {
if (Rs < 1900) {
return 1;
}
else if (Rs >= 1900) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 162) {
return 1;
}
}
else if (tcor >= 69) {
if (MVDiff < 14) {
if (Cs < 832) {
if (Rs < 1314) {
return 8;
}
else if (Rs >= 1314) {
return 2;
}
}
else if (Cs >= 832) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 2;
}
}
}
else if (MVDiff >= 14) {
if (Rs < 1716) {
if (diffAFD < 4) {
if (RsCsDiff < 183) {
if (MVDiff < 45) {
if (MVDiff < 30) {
return 8;
}
else if (MVDiff >= 30) {
if (Cs < 1182) {
return 1;
}
else if (Cs >= 1182) {
return 8;
}
}
}
else if (MVDiff >= 45) {
return 8;
}
}
else if (RsCsDiff >= 183) {
if (Cs < 659) {
return 1;
}
else if (Cs >= 659) {
return 2;
}
}
}
else if (diffAFD >= 4) {
if (Cs < 828) {
return 2;
}
else if (Cs >= 828) {
return 8;
}
}
}
else if (Rs >= 1716) {
if (posBalance < 46) {
if (Rs < 1806) {
return 8;
}
else if (Rs >= 1806) {
return 2;
}
}
else if (posBalance >= 46) {
if (RsCsDiff < 170) {
if (diffRsCsdiff < 5) {
if (diffTSC < -5) {
return 8;
}
else if (diffTSC >= -5) {
return 1;
}
}
else if (diffRsCsdiff >= 5) {
if (mcTcor < 95) {
if (ssDCval < 116) {
if (Rs < 1901) {
return 2;
}
else if (Rs >= 1901) {
return 1;
}
}
else if (ssDCval >= 116) {
return 8;
}
}
else if (mcTcor >= 95) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
}
}
else if (RsCsDiff >= 170) {
if (posBalance < 64) {
if (mu_mv_mag_sq < 29) {
return 1;
}
else if (mu_mv_mag_sq >= 29) {
if (Rs < 2070) {
return 8;
}
else if (Rs >= 2070) {
return 2;
}
}
}
else if (posBalance >= 64) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
if (AFD < 34) {
return 1;
}
else if (AFD >= 34) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (SC < 2133) {
if (Rs < 964) {
if (ssDCval < 64) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (Rs < 841) {
if (Rs < 549) {
if (SC < 1090) {
if (ssDCval < 49) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (ssDCval >= 49) {
if (mcTcor < 99) {
if (diffMVdiffVal < -2) {
if (Cs < 701) {
return 1;
}
else if (Cs >= 701) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (tcor < 98) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 2;
}
}
else if (tcor >= 98) {
if (Rs < 221) {
return 4;
}
else if (Rs >= 221) {
return 1;
}
}
}
}
}
else if (mcTcor >= 99) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
else if (SC >= 1090) {
if (Rs < 484) {
if (Cs < 883) {
if (AFD < 5) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
return 4;
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 883) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
if (Rs < 320) {
if (Rs < 249) {
return 1;
}
else if (Rs >= 249) {
return 2;
}
}
else if (Rs >= 320) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 4;
}
}
}
}
}
else if (Rs >= 484) {
if (Cs < 649) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (Rs < 536) {
return 2;
}
else if (Rs >= 536) {
if (Rs < 546) {
return 8;
}
else if (Rs >= 546) {
return 4;
}
}
}
}
else if (Cs >= 649) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
}
else if (Rs >= 549) {
if (MVDiff < 1) {
if (Cs < 933) {
if (tcor < 97) {
if (ssDCval < 55) {
if (Rs < 643) {
return 1;
}
else if (Rs >= 643) {
return 4;
}
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (tcor >= 97) {
if (Cs < 535) {
return 4;
}
else if (Cs >= 535) {
return 8;
}
}
}
else if (Cs >= 933) {
if (Rs < 638) {
if (Cs < 1069) {
return 1;
}
else if (Cs >= 1069) {
return 4;
}
}
else if (Rs >= 638) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (Rs < 781) {
return 1;
}
else if (Rs >= 781) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 17) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
return 1;
}
}
else if (negBalance >= 17) {
if (negBalance < 31) {
if (TSC < 3) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (SC < 1394) {
if (Rs < 687) {
return 4;
}
else if (Rs >= 687) {
return 1;
}
}
else if (SC >= 1394) {
if (ssDCval < 62) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
return 8;
}
}
else if (ssDCval >= 62) {
return 4;
}
}
}
}
else if (TSC >= 3) {
if (MVDiff < 13) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Cs < 797) {
return 8;
}
else if (Cs >= 797) {
return 4;
}
}
}
else if (MVDiff >= 13) {
return 1;
}
}
}
else if (negBalance >= 31) {
if (SC < 1383) {
if (AFD < 7) {
if (mu_mv_mag_sq < 78) {
if (SC < 1316) {
if (SC < 1220) {
return 8;
}
else if (SC >= 1220) {
return 4;
}
}
else if (SC >= 1316) {
return 8;
}
}
else if (mu_mv_mag_sq >= 78) {
return 8;
}
}
else if (AFD >= 7) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
if (Rs < 608) {
return 1;
}
else if (Rs >= 608) {
return 4;
}
}
}
}
}
else if (SC >= 1383) {
if (Cs < 911) {
if (AFD < 10) {
if (Cs < 838) {
if (Rs < 836) {
if (Cs < 778) {
if (RsDiff < 2) {
if (Rs < 752) {
return 8;
}
else if (Rs >= 752) {
return 4;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (Cs >= 778) {
return 8;
}
}
else if (Rs >= 836) {
return 8;
}
}
else if (Cs >= 838) {
if (MVDiff < 6) {
if (Cs < 870) {
return 8;
}
else if (Cs >= 870) {
return 4;
}
}
else if (MVDiff >= 6) {
if (Rs < 763) {
return 2;
}
else if (Rs >= 763) {
return 4;
}
}
}
}
else if (AFD >= 10) {
if (Rs < 815) {
return 8;
}
else if (Rs >= 815) {
return 2;
}
}
}
else if (Cs >= 911) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 33) {
return 2;
}
else if (negBalance >= 33) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 841) {
if (ssDCval < 49) {
if (Cs < 555) {
return 1;
}
else if (Cs >= 555) {
return 8;
}
}
else if (ssDCval >= 49) {
if (SC < 1495) {
if (mu_mv_mag_sq < 117) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
else if (SC >= 1495) {
if (diffRsCsdiff < -4) {
if (diffRsCsdiff < -7) {
return 1;
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
if (negBalance < 41) {
if (TSC < 4) {
if (Cs < 561) {
return 8;
}
else if (Cs >= 561) {
if (Rs < 956) {
return 1;
}
else if (Rs >= 956) {
if (Rs < 958) {
return 8;
}
else if (Rs >= 958) {
return 1;
}
}
}
}
else if (TSC >= 4) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
if (Cs < 1047) {
return 1;
}
else if (Cs >= 1047) {
return 4;
}
}
}
}
else if (negBalance >= 41) {
if (Cs < 689) {
return 8;
}
else if (Cs >= 689) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (Rs < 828) {
if (Rs < 717) {
if (ssDCval < 83) {
if (mu_mv_mag_sq < 82) {
if (Cs < 785) {
if (ssDCval < 74) {
if (Rs < 351) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 2;
}
}
else if (Rs >= 351) {
if (ssDCval < 65) {
if (Rs < 607) {
return 1;
}
else if (Rs >= 607) {
return 8;
}
}
else if (ssDCval >= 65) {
if (AFD < 11) {
if (mu_mv_mag_sq < 77) {
return 4;
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 536) {
return 4;
}
else if (Rs >= 536) {
return 1;
}
}
}
else if (AFD >= 11) {
return 1;
}
}
}
}
else if (ssDCval >= 74) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (Cs >= 785) {
if (posBalance < 59) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (mcTcor < 100) {
if (CsDiff < 4) {
if (Rs < 536) {
return 1;
}
else if (Rs >= 536) {
if (Rs < 686) {
return 2;
}
else if (Rs >= 686) {
return 1;
}
}
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (mcTcor >= 100) {
if (Rs < 586) {
return 1;
}
else if (Rs >= 586) {
return 4;
}
}
}
}
else if (posBalance >= 59) {
if (Rs < 584) {
return 1;
}
else if (Rs >= 584) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (RsDiff < 1) {
if (ssDCval < 75) {
if (refDCval < 66) {
if (Rs < 679) {
return 8;
}
else if (Rs >= 679) {
return 4;
}
}
else if (refDCval >= 66) {
return 8;
}
}
else if (ssDCval >= 75) {
return 1;
}
}
else if (RsDiff >= 1) {
if (Rs < 646) {
if (Cs < 544) {
if (Rs < 259) {
return 1;
}
else if (Rs >= 259) {
return 2;
}
}
else if (Cs >= 544) {
if (mu_mv_mag_sq < 91) {
if (Rs < 505) {
return 4;
}
else if (Rs >= 505) {
return 1;
}
}
else if (mu_mv_mag_sq >= 91) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
else if (Rs >= 646) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Rs < 685) {
return 8;
}
else if (Rs >= 685) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 83) {
if (TSC < 4) {
if (Rs < 655) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
if (Cs < 743) {
if (SC < 1278) {
return 2;
}
else if (SC >= 1278) {
if (Cs < 699) {
return 2;
}
else if (Cs >= 699) {
return 1;
}
}
}
else if (Cs >= 743) {
return 2;
}
}
}
else if (Rs >= 655) {
if (Cs < 865) {
return 1;
}
else if (Cs >= 865) {
return 2;
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 35) {
if (Rs < 640) {
if (Rs < 507) {
return 8;
}
else if (Rs >= 507) {
return 1;
}
}
else if (Rs >= 640) {
return 2;
}
}
else if (mu_mv_mag_sq >= 35) {
if (Rs < 477) {
return 8;
}
else if (Rs >= 477) {
return 4;
}
}
}
}
}
else if (Rs >= 717) {
if (ssDCval < 83) {
if (MVDiff < 3) {
if (Cs < 633) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 633) {
if (mu_mv_mag_sq < 119) {
if (Cs < 1017) {
if (Rs < 796) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (ssDCval < 75) {
if (AFD < 5) {
if (diffMVdiffVal < -1) {
if (Rs < 761) {
return 1;
}
else if (Rs >= 761) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (AFD >= 5) {
return 4;
}
}
else if (ssDCval >= 75) {
if (Rs < 744) {
return 4;
}
else if (Rs >= 744) {
return 2;
}
}
}
}
else if (Rs >= 796) {
return 4;
}
}
else if (Cs >= 1017) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (ssDCval < 68) {
if (Cs < 727) {
return 1;
}
else if (Cs >= 727) {
return 4;
}
}
else if (ssDCval >= 68) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 78) {
if (posBalance < 60) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 4;
}
}
else if (posBalance >= 60) {
if (Cs < 1116) {
return 4;
}
else if (Cs >= 1116) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
}
}
else if (ssDCval >= 83) {
return 4;
}
}
}
else if (Rs >= 828) {
if (refDCval < 73) {
if (tcor < 98) {
if (diffAFD < 1) {
if (negBalance < 38) {
if (mu_mv_mag_sq < 90) {
if (Rs < 887) {
if (Rs < 867) {
return 4;
}
else if (Rs >= 867) {
return 1;
}
}
else if (Rs >= 887) {
if (Rs < 894) {
return 8;
}
else if (Rs >= 894) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 90) {
return 4;
}
}
else if (negBalance >= 38) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 890) {
return 1;
}
else if (Rs >= 890) {
return 4;
}
}
}
}
else if (diffAFD >= 1) {
if (posBalance < 45) {
if (Cs < 829) {
return 1;
}
else if (Cs >= 829) {
return 4;
}
}
else if (posBalance >= 45) {
if (Rs < 847) {
return 1;
}
else if (Rs >= 847) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (Cs < 791) {
return 2;
}
else if (Cs >= 791) {
if (Cs < 900) {
return 4;
}
else if (Cs >= 900) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
}
}
else if (refDCval >= 73) {
if (ssDCval < 85) {
if (tcor < 96) {
if (Cs < 921) {
if (Rs < 866) {
return 4;
}
else if (Rs >= 866) {
return 8;
}
}
else if (Cs >= 921) {
return 2;
}
}
else if (tcor >= 96) {
if (SC < 2007) {
return 8;
}
else if (SC >= 2007) {
if (Cs < 1056) {
return 4;
}
else if (Cs >= 1056) {
return 8;
}
}
}
}
else if (ssDCval >= 85) {
return 4;
}
}
}
}
}
else if (Rs >= 964) {
if (mu_mv_mag_sq < 127) {
if (refDCval < 79) {
if (refDCval < 51) {
if (MVDiff < 3) {
if (Rs < 1183) {
return 8;
}
else if (Rs >= 1183) {
if (Rs < 1259) {
return 4;
}
else if (Rs >= 1259) {
if (Rs < 1335) {
return 4;
}
else if (Rs >= 1335) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 82) {
return 1;
}
else if (mu_mv_mag_sq >= 82) {
if (Cs < 798) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
return 1;
}
}
else if (Cs >= 798) {
return 8;
}
}
}
}
else if (refDCval >= 51) {
if (SC < 1801) {
if (refDCval < 65) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (refDCval >= 65) {
if (negBalance < 27) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
if (mcTcor < 99) {
if (Cs < 667) {
return 2;
}
else if (Cs >= 667) {
return 4;
}
}
else if (mcTcor >= 99) {
return 2;
}
}
}
else if (negBalance >= 27) {
if (negBalance < 46) {
if (Cs < 653) {
return 1;
}
else if (Cs >= 653) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
return 4;
}
}
}
}
else if (negBalance >= 46) {
if (Rs < 1048) {
if (Rs < 1016) {
return 1;
}
else if (Rs >= 1016) {
return 4;
}
}
else if (Rs >= 1048) {
return 8;
}
}
}
}
}
else if (SC >= 1801) {
if (ssDCval < 68) {
if (posBalance < 55) {
if (RsDiff < 3) {
if (ssDCval < 56) {
if (ssDCval < 54) {
if (mu_mv_mag_sq < 124) {
return 1;
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 1321) {
return 8;
}
else if (Rs >= 1321) {
return 1;
}
}
}
else if (ssDCval >= 54) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (SC < 1929) {
return 8;
}
else if (SC >= 1929) {
return 1;
}
}
}
}
else if (ssDCval >= 56) {
if (Cs < 1072) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (Rs < 1233) {
return 8;
}
else if (Rs >= 1233) {
return 1;
}
}
}
else if (Cs >= 1072) {
if (Rs < 1008) {
return 8;
}
else if (Rs >= 1008) {
if (Rs < 1020) {
return 1;
}
else if (Rs >= 1020) {
return 4;
}
}
}
}
}
else if (RsDiff >= 3) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 84) {
if (SC < 2119) {
return 1;
}
else if (SC >= 2119) {
return 8;
}
}
else if (mu_mv_mag_sq >= 84) {
return 8;
}
}
else if (MVDiff >= 8) {
if (mcTcor < 98) {
if (Rs < 1347) {
return 2;
}
else if (Rs >= 1347) {
return 1;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
}
}
else if (posBalance >= 55) {
if (Cs < 779) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (Cs >= 779) {
if (TSC < 5) {
if (Cs < 888) {
if (Rs < 1236) {
return 1;
}
else if (Rs >= 1236) {
return 8;
}
}
else if (Cs >= 888) {
return 8;
}
}
else if (TSC >= 5) {
return 4;
}
}
}
}
else if (ssDCval >= 68) {
if (mu_mv_mag_sq < 60) {
if (Cs < 952) {
return 1;
}
else if (Cs >= 952) {
return 2;
}
}
else if (mu_mv_mag_sq >= 60) {
if (TSC < 4) {
if (SC < 2010) {
if (Rs < 1364) {
if (ssDCval < 77) {
return 1;
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (Rs >= 1364) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (SC >= 2010) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (SC < 2042) {
return 8;
}
else if (SC >= 2042) {
if (Rs < 1457) {
if (Rs < 1191) {
return 8;
}
else if (Rs >= 1191) {
return 1;
}
}
else if (Rs >= 1457) {
return 8;
}
}
}
}
}
else if (TSC >= 4) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 79) {
if (mu_mv_mag_sq < 112) {
if (mu_mv_mag_sq < 4) {
if (Rs < 1112) {
return 2;
}
else if (Rs >= 1112) {
return 1;
}
}
else if (mu_mv_mag_sq >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 112) {
if (Cs < 955) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
if (Rs < 982) {
return 4;
}
else if (Rs >= 982) {
return 8;
}
}
}
else if (Cs >= 955) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 1124) {
if (Cs < 597) {
return 2;
}
else if (Cs >= 597) {
return 8;
}
}
else if (Rs >= 1124) {
if (SC < 1935) {
if (posBalance < 37) {
if (Rs < 1144) {
return 8;
}
else if (Rs >= 1144) {
return 1;
}
}
else if (posBalance >= 37) {
if (Cs < 698) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (Cs >= 698) {
return 4;
}
}
}
else if (SC >= 1935) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
return 1;
}
}
}
}
}
}
else if (SC >= 2133) {
if (Rs < 1815) {
if (Cs < 1050) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 43) {
if (Cs < 776) {
if (refDCval < 85) {
if (Cs < 772) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (mcTcor < 100) {
return 1;
}
else if (mcTcor >= 100) {
if (Rs < 1672) {
return 1;
}
else if (Rs >= 1672) {
return 2;
}
}
}
}
else if (Cs >= 772) {
return 4;
}
}
else if (refDCval >= 85) {
return 4;
}
}
else if (Cs >= 776) {
if (mu_mv_mag_sq < 79) {
if (ssDCval < 45) {
return 4;
}
else if (ssDCval >= 45) {
return 1;
}
}
else if (mu_mv_mag_sq >= 79) {
if (negBalance < 23) {
if (Rs < 1794) {
if (tcor < 98) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (tcor >= 98) {
if (ssDCval < 58) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (ssDCval >= 58) {
if (posBalance < 34) {
if (Cs < 940) {
return 8;
}
else if (Cs >= 940) {
return 1;
}
}
else if (posBalance >= 34) {
return 1;
}
}
}
}
else if (Rs >= 1794) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 4;
}
}
}
else if (negBalance >= 23) {
if (negBalance < 68) {
if (Rs < 1767) {
if (mu_mv_mag_sq < 87) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 87) {
if (tcor < 102) {
return 8;
}
else if (tcor >= 102) {
return 4;
}
}
}
else if (Rs >= 1767) {
return 1;
}
}
else if (negBalance >= 68) {
return 4;
}
}
}
}
}
else if (posBalance >= 43) {
if (SC < 2522) {
if (mu_mv_mag_sq < 122) {
if (tcor < 99) {
if (Cs < 717) {
if (mu_mv_mag_sq < 117) {
return 1;
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
else if (Cs >= 717) {
if (tcor < 87) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 1;
}
}
else if (tcor >= 87) {
if (gchDC < 1) {
if (Rs < 1217) {
if (Rs < 1189) {
return 8;
}
else if (Rs >= 1189) {
return 1;
}
}
else if (Rs >= 1217) {
return 8;
}
}
else if (gchDC >= 1) {
if (RsDiff < 4) {
if (posBalance < 58) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (RsDiff >= 4) {
if (Rs < 1665) {
return 8;
}
else if (Rs >= 1665) {
return 1;
}
}
}
}
}
}
else if (tcor >= 99) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (Cs < 850) {
return 1;
}
else if (Cs >= 850) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
return 8;
}
}
else if (SC >= 2522) {
if (posBalance < 56) {
if (Rs < 1803) {
return 8;
}
else if (Rs >= 1803) {
return 1;
}
}
else if (posBalance >= 56) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (refDCval < 63) {
if (tcor < 99) {
if (Rs < 1588) {
return 8;
}
else if (Rs >= 1588) {
if (Rs < 1763) {
return 8;
}
else if (Rs >= 1763) {
return 4;
}
}
}
else if (tcor >= 99) {
if (Cs < 980) {
return 1;
}
else if (Cs >= 980) {
return 8;
}
}
}
else if (refDCval >= 63) {
return 8;
}
}
}
else if (Cs >= 1050) {
if (RsDiff < 1) {
if (mu_mv_mag_sq < 103) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (mu_mv_mag_sq >= 103) {
if (ssDCval < 52) {
if (Rs < 1583) {
return 8;
}
else if (Rs >= 1583) {
return 4;
}
}
else if (ssDCval >= 52) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
if (Rs < 1269) {
return 8;
}
else if (Rs >= 1269) {
return 1;
}
}
}
}
}
else if (RsDiff >= 1) {
if (AFD < 3) {
if (Rs < 1399) {
if (mu_mv_mag_sq < 113) {
return 4;
}
else if (mu_mv_mag_sq >= 113) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 70) {
if (ssDCval < 53) {
if (SC < 2510) {
return 8;
}
else if (SC >= 2510) {
return 4;
}
}
else if (ssDCval >= 53) {
if (SC < 2316) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 4;
}
}
else if (SC >= 2316) {
return 1;
}
}
}
else if (ssDCval >= 70) {
if (Cs < 1096) {
return 8;
}
else if (Cs >= 1096) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1003) {
return 8;
}
else if (Rs >= 1003) {
return 4;
}
}
}
}
else if (Rs >= 1399) {
if (SC < 2927) {
if (mu_mv_mag_sq < 122) {
return 8;
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 1738) {
if (ssDCval < 76) {
if (ssDCval < 48) {
if (SC < 2853) {
if (SC < 2848) {
if (Cs < 1112) {
if (Cs < 1095) {
if (Rs < 1638) {
return 4;
}
else if (Rs >= 1638) {
if (mcTcor < 99) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
return 4;
}
}
else if (mcTcor >= 99) {
return 4;
}
}
}
else if (Cs >= 1095) {
return 8;
}
}
else if (Cs >= 1112) {
if (Rs < 1531) {
return 8;
}
else if (Rs >= 1531) {
if (Cs < 1122) {
return 4;
}
else if (Cs >= 1122) {
if (tcor < 99) {
if (Rs < 1686) {
return 4;
}
else if (Rs >= 1686) {
if (Rs < 1693) {
return 8;
}
else if (Rs >= 1693) {
if (Cs < 1127) {
return 4;
}
else if (Cs >= 1127) {
return 4;
}
}
}
}
else if (tcor >= 99) {
if (Rs < 1689) {
if (Rs < 1661) {
return 4;
}
else if (Rs >= 1661) {
return 4;
}
}
else if (Rs >= 1689) {
return 8;
}
}
}
}
}
}
else if (SC >= 2848) {
return 8;
}
}
else if (SC >= 2853) {
if (diffAFD < -1) {
if (Rs < 1689) {
return 4;
}
else if (Rs >= 1689) {
return 8;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (Rs >= 1738) {
if (TSCindex < 3) {
if (Cs < 1140) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (Rs < 1803) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (posBalance < 40) {
if (SC < 2871) {
if (SC < 2864) {
if (Rs < 1741) {
return 8;
}
else if (Rs >= 1741) {
if (Rs < 1797) {
return 4;
}
else if (Rs >= 1797) {
return 4;
}
}
}
else if (SC >= 2864) {
return 4;
}
}
else if (SC >= 2871) {
if (Cs < 1095) {
return 8;
}
else if (Cs >= 1095) {
if (posBalance < 31) {
return 4;
}
else if (posBalance >= 31) {
if (SC < 2913) {
if (Cs < 1133) {
if (SC < 2885) {
if (Rs < 1782) {
return 8;
}
else if (Rs >= 1782) {
return 8;
}
}
else if (SC >= 2885) {
return 8;
}
}
else if (Cs >= 1133) {
if (Rs < 1766) {
return 4;
}
else if (Rs >= 1766) {
return 8;
}
}
}
else if (SC >= 2913) {
return 8;
}
}
}
}
}
else if (posBalance >= 40) {
return 8;
}
}
}
else if (Rs >= 1803) {
return 8;
}
}
}
else if (Cs >= 1140) {
return 8;
}
}
else if (TSCindex >= 3) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
return 4;
}
}
}
}
}
else if (SC >= 2927) {
if (Cs < 1178) {
return 8;
}
else if (Cs >= 1178) {
return 4;
}
}
}
}
else if (AFD >= 3) {
if (Rs < 1399) {
if (RsDiff < 2) {
if (ssDCval < 76) {
if (Rs < 968) {
if (Rs < 961) {
return 2;
}
else if (Rs >= 961) {
return 4;
}
}
else if (Rs >= 968) {
if (mu_mv_mag_sq < 109) {
return 8;
}
else if (mu_mv_mag_sq >= 109) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
return 1;
}
}
}
}
else if (ssDCval >= 76) {
if (negBalance < 26) {
return 8;
}
else if (negBalance >= 26) {
return 4;
}
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 78) {
if (diffAFD < 1) {
if (AFD < 11) {
if (Cs < 1131) {
return 8;
}
else if (Cs >= 1131) {
return 2;
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (diffAFD >= 1) {
if (Rs < 1285) {
return 4;
}
else if (Rs >= 1285) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (negBalance < 35) {
if (mu_mv_mag_sq < 123) {
if (ssDCval < 66) {
if (posBalance < 52) {
if (Rs < 1032) {
return 4;
}
else if (Rs >= 1032) {
if (SC < 2507) {
return 8;
}
else if (SC >= 2507) {
return 4;
}
}
}
else if (posBalance >= 52) {
return 4;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (mu_mv_mag_sq >= 123) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
return 4;
}
}
}
else if (negBalance >= 35) {
if (mu_mv_mag_sq < 123) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (Rs < 1310) {
if (Cs < 1076) {
return 8;
}
else if (Cs >= 1076) {
return 4;
}
}
else if (Rs >= 1310) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1154) {
return 4;
}
else if (Rs >= 1154) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1399) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (tcor < 96) {
if (Rs < 1661) {
if (Rs < 1528) {
return 8;
}
else if (Rs >= 1528) {
return 4;
}
}
else if (Rs >= 1661) {
return 2;
}
}
else if (tcor >= 96) {
if (diffRsCsdiff < 7) {
return 8;
}
else if (diffRsCsdiff >= 7) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 1815) {
if (TSC < 4) {
if (refDCval < 83) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 97) {
return 4;
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 930) {
return 8;
}
else if (Cs >= 930) {
if (posBalance < 37) {
if (mu_mv_mag_sq < 128) {
return 4;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1821) {
return 4;
}
else if (Rs >= 1821) {
return 4;
}
}
}
else if (posBalance >= 37) {
if (Rs < 1974) {
if (Rs < 1834) {
return 4;
}
else if (Rs >= 1834) {
return 4;
}
}
else if (Rs >= 1974) {
return 8;
}
}
}
}
}
else if (ssDCval >= 56) {
if (posBalance < 46) {
if (Cs < 1135) {
if (refDCval < 71) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 99) {
if (Cs < 892) {
return 1;
}
else if (Cs >= 892) {
if (negBalance < 16) {
return 1;
}
else if (negBalance >= 16) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (negBalance < 14) {
if (Cs < 1040) {
return 1;
}
else if (Cs >= 1040) {
return 4;
}
}
else if (negBalance >= 14) {
return 1;
}
}
}
else if (mcTcor >= 99) {
if (mu_mv_mag_sq < 108) {
if (refDCval < 58) {
return 1;
}
else if (refDCval >= 58) {
return 4;
}
}
else if (mu_mv_mag_sq >= 108) {
if (SC < 2880) {
return 1;
}
else if (SC >= 2880) {
if (Rs < 1894) {
if (SC < 2893) {
return 4;
}
else if (SC >= 2893) {
return 1;
}
}
else if (Rs >= 1894) {
if (Rs < 2071) {
return 4;
}
else if (Rs >= 2071) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 71) {
if (MVDiff < 32) {
return 1;
}
else if (MVDiff >= 32) {
return 8;
}
}
}
else if (Cs >= 1135) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
if (Rs < 1928) {
return 2;
}
else if (Rs >= 1928) {
return 8;
}
}
}
}
else if (posBalance >= 46) {
if (SC < 2787) {
if (Cs < 684) {
return 1;
}
else if (Cs >= 684) {
return 4;
}
}
else if (SC >= 2787) {
return 8;
}
}
}
}
else if (refDCval >= 83) {
if (MVDiff < 2) {
if (Rs < 2195) {
return 4;
}
else if (Rs >= 2195) {
return 8;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (TSC >= 4) {
if (Cs < 931) {
if (ssDCval < 71) {
if (posBalance < 51) {
if (CsDiff < 3) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 2;
}
}
else if (CsDiff >= 3) {
if (SC < 3102) {
return 2;
}
else if (SC >= 3102) {
return 4;
}
}
}
else if (posBalance >= 51) {
return 2;
}
}
else if (ssDCval >= 71) {
if (posBalance < 34) {
if (AFD < 9) {
if (Rs < 2038) {
return 1;
}
else if (Rs >= 2038) {
if (Rs < 2156) {
return 4;
}
else if (Rs >= 2156) {
return 4;
}
}
}
else if (AFD >= 9) {
return 2;
}
}
else if (posBalance >= 34) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (Rs < 2096) {
return 2;
}
else if (Rs >= 2096) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 931) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 84) {
if (MVDiff < 7) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (Rs < 1877) {
return 4;
}
else if (Rs >= 1877) {
return 1;
}
}
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (mu_mv_mag_sq >= 84) {
if (Rs < 2078) {
return 8;
}
else if (Rs >= 2078) {
return 4;
}
}
}
else if (negBalance >= 44) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 97) {
if (SC < 1391) {
if (Rs < 414) {
if (ssDCval < 112) {
if (MVDiff < 7) {
if (Cs < 727) {
return 1;
}
else if (Cs >= 727) {
return 4;
}
}
else if (MVDiff >= 7) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (ssDCval >= 112) {
if (TSC < 4) {
if (mu_mv_mag_sq < 95) {
return 2;
}
else if (mu_mv_mag_sq >= 95) {
return 4;
}
}
else if (TSC >= 4) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
return 2;
}
}
}
}
else if (Rs >= 414) {
if (negBalance < 25) {
if (ssDCval < 138) {
if (refDCval < 95) {
return 2;
}
else if (refDCval >= 95) {
if (ssDCval < 117) {
return 1;
}
else if (ssDCval >= 117) {
if (Rs < 599) {
if (Rs < 524) {
return 4;
}
else if (Rs >= 524) {
return 2;
}
}
else if (Rs >= 599) {
return 1;
}
}
}
}
else if (ssDCval >= 138) {
return 4;
}
}
else if (negBalance >= 25) {
if (TSC < 5) {
if (refDCval < 126) {
if (posBalance < 34) {
if (posBalance < 15) {
if (Cs < 682) {
return 8;
}
else if (Cs >= 682) {
return 1;
}
}
else if (posBalance >= 15) {
return 1;
}
}
else if (posBalance >= 34) {
if (mcTcor < 98) {
if (ssDCval < 114) {
return 1;
}
else if (ssDCval >= 114) {
if (ssDCval < 117) {
return 2;
}
else if (ssDCval >= 117) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (SC < 1056) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
else if (SC >= 1056) {
if (mu_mv_mag_sq < 55) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (mu_mv_mag_sq >= 55) {
return 2;
}
}
}
}
}
else if (refDCval >= 126) {
if (SC < 1178) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (SC >= 1178) {
if (MVDiff < 27) {
return 1;
}
else if (MVDiff >= 27) {
if (Rs < 738) {
return 8;
}
else if (Rs >= 738) {
return 2;
}
}
}
}
}
else if (TSC >= 5) {
if (tcor < 94) {
if (MVDiff < 5) {
if (Rs < 722) {
return 4;
}
else if (Rs >= 722) {
return 1;
}
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (tcor >= 94) {
if (ssDCval < 140) {
if (negBalance < 32) {
if (Rs < 622) {
return 2;
}
else if (Rs >= 622) {
return 8;
}
}
else if (negBalance >= 32) {
if (Cs < 566) {
return 2;
}
else if (Cs >= 566) {
return 8;
}
}
}
else if (ssDCval >= 140) {
if (Rs < 483) {
return 4;
}
else if (Rs >= 483) {
return 2;
}
}
}
}
}
}
}
else if (SC >= 1391) {
if (refDCval < 102) {
if (TSC < 5) {
if (Rs < 930) {
if (Rs < 753) {
if (Rs < 531) {
return 8;
}
else if (Rs >= 531) {
if (Rs < 673) {
return 2;
}
else if (Rs >= 673) {
return 1;
}
}
}
else if (Rs >= 753) {
return 4;
}
}
else if (Rs >= 930) {
if (gchDC < 3) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (gchDC >= 3) {
if (Rs < 1468) {
return 2;
}
else if (Rs >= 1468) {
return 8;
}
}
}
}
else if (TSC >= 5) {
if (tcor < 95) {
if (Rs < 1888) {
if (RsDiff < 5) {
if (Rs < 1362) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
return 4;
}
}
else if (Rs >= 1362) {
return 1;
}
}
else if (RsDiff >= 5) {
if (tcor < 70) {
return 4;
}
else if (tcor >= 70) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (diffAFD < 3) {
if (diffRsCsdiff < -9) {
if (SC < 2448) {
if (SC < 2007) {
return 8;
}
else if (SC >= 2007) {
return 1;
}
}
else if (SC >= 2448) {
return 2;
}
}
else if (diffRsCsdiff >= -9) {
return 2;
}
}
else if (diffAFD >= 3) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1888) {
if (Cs < 1087) {
return 4;
}
else if (Cs >= 1087) {
return 8;
}
}
}
else if (tcor >= 95) {
if (MVDiff < 2) {
if (Rs < 1315) {
return 4;
}
else if (Rs >= 1315) {
return 1;
}
}
else if (MVDiff >= 2) {
if (ssDCval < 96) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (MVDiff < 4) {
if (Rs < 1048) {
return 4;
}
else if (Rs >= 1048) {
return 1;
}
}
else if (MVDiff >= 4) {
if (refDCval < 88) {
return 4;
}
else if (refDCval >= 88) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 96) {
if (negBalance < 48) {
if (Cs < 1149) {
return 4;
}
else if (Cs >= 1149) {
return 2;
}
}
else if (negBalance >= 48) {
if (Rs < 854) {
return 4;
}
else if (Rs >= 854) {
if (Cs < 660) {
return 1;
}
else if (Cs >= 660) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 102) {
if (RsCsDiff < 77) {
if (MVDiff < 6) {
if (Rs < 656) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 1;
}
}
}
else if (Rs >= 656) {
if (SC < 2612) {
if (Cs < 601) {
return 1;
}
else if (Cs >= 601) {
if (negBalance < 53) {
if (gchDC < 2) {
if (Cs < 823) {
return 8;
}
else if (Cs >= 823) {
if (mu_mv_mag_sq < 83) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (Rs < 1303) {
if (Cs < 905) {
return 8;
}
else if (Cs >= 905) {
return 4;
}
}
else if (Rs >= 1303) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 83) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
else if (negBalance >= 53) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
}
}
else if (SC >= 2612) {
if (Rs < 1989) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Rs < 1606) {
if (Rs < 1472) {
return 4;
}
else if (Rs >= 1472) {
return 8;
}
}
else if (Rs >= 1606) {
return 2;
}
}
}
else if (Rs >= 1989) {
if (negBalance < 26) {
if (Rs < 2165) {
return 4;
}
else if (Rs >= 2165) {
return 2;
}
}
else if (negBalance >= 26) {
return 8;
}
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 156) {
if (mu_mv_mag_sq < 3) {
if (Cs < 858) {
if (negBalance < 49) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Rs < 1307) {
return 2;
}
else if (Rs >= 1307) {
return 1;
}
}
}
else if (negBalance >= 49) {
return 1;
}
}
else if (Cs >= 858) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 2209) {
if (SC < 1534) {
if (Cs < 744) {
if (Rs < 856) {
return 2;
}
else if (Rs >= 856) {
return 8;
}
}
else if (Cs >= 744) {
return 8;
}
}
else if (SC >= 1534) {
if (Cs < 565) {
if (Rs < 1875) {
return 2;
}
else if (Rs >= 1875) {
return 1;
}
}
else if (Cs >= 565) {
if (negBalance < 26) {
if (Rs < 1228) {
return 8;
}
else if (Rs >= 1228) {
return 1;
}
}
else if (negBalance >= 26) {
if (diffRsCsdiff < -77) {
if (Rs < 1737) {
return 2;
}
else if (Rs >= 1737) {
return 1;
}
}
else if (diffRsCsdiff >= -77) {
if (Cs < 1311) {
if (SC < 1717) {
if (SC < 1711) {
return 8;
}
else if (SC >= 1711) {
return 1;
}
}
else if (SC >= 1717) {
return 8;
}
}
else if (Cs >= 1311) {
if (Rs < 1352) {
return 8;
}
else if (Rs >= 1352) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 2209) {
if (diffAFD < 1) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 8;
}
}
else if (diffAFD >= 1) {
return 2;
}
}
}
}
else if (ssDCval >= 156) {
if (Cs < 1134) {
return 2;
}
else if (Cs >= 1134) {
return 4;
}
}
}
}
else if (RsCsDiff >= 77) {
if (Cs < 691) {
if (Rs < 1393) {
if (negBalance < 53) {
return 2;
}
else if (negBalance >= 53) {
return 1;
}
}
else if (Rs >= 1393) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (gchDC < 7) {
if (mu_mv_mag_sq < 46) {
return 1;
}
else if (mu_mv_mag_sq >= 46) {
return 2;
}
}
else if (gchDC >= 7) {
return 8;
}
}
}
}
else if (Cs >= 691) {
if (mu_mv_mag_sq < 5) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
if (RsDiff < 14) {
return 1;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (tcor < 92) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (diffRsCsdiff < -19) {
if (Rs < 1330) {
return 2;
}
else if (Rs >= 1330) {
return 8;
}
}
else if (diffRsCsdiff >= -19) {
return 2;
}
}
}
else if (tcor >= 92) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (ssDCval < 133) {
if (Cs < 1220) {
if (Rs < 374) {
if (ssDCval < 122) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 122) {
if (Cs < 946) {
return 4;
}
else if (Cs >= 946) {
return 8;
}
}
else if (mu_mv_mag_sq >= 122) {
return 8;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (ssDCval >= 122) {
if (Rs < 323) {
return 2;
}
else if (Rs >= 323) {
return 8;
}
}
}
else if (Rs >= 374) {
if (refDCval < 97) {
if (mu_mv_mag_sq < 126) {
if (Cs < 872) {
if (Rs < 1043) {
return 8;
}
else if (Rs >= 1043) {
if (Rs < 1971) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Rs >= 1971) {
if (Rs < 2149) {
if (Cs < 665) {
return 1;
}
else if (Cs >= 665) {
return 8;
}
}
else if (Rs >= 2149) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
}
}
}
else if (Cs >= 872) {
if (tcor < 98) {
if (Rs < 1023) {
if (AFD < 5) {
if (Rs < 604) {
return 8;
}
else if (Rs >= 604) {
return 1;
}
}
else if (AFD >= 5) {
return 4;
}
}
else if (Rs >= 1023) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (tcor >= 98) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 2155) {
if (posBalance < 24) {
return 4;
}
else if (posBalance >= 24) {
return 8;
}
}
else if (Rs >= 2155) {
if (Cs < 747) {
return 8;
}
else if (Cs >= 747) {
return 4;
}
}
}
}
else if (refDCval >= 97) {
if (mu_mv_mag_sq < 106) {
if (AFD < 3) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (Rs < 562) {
return 1;
}
else if (Rs >= 562) {
if (Rs < 1136) {
return 2;
}
else if (Rs >= 1136) {
return 8;
}
}
}
}
else if (AFD >= 3) {
if (Rs < 524) {
if (Cs < 947) {
if (Cs < 890) {
if (Rs < 433) {
return 4;
}
else if (Rs >= 433) {
return 1;
}
}
else if (Cs >= 890) {
return 8;
}
}
else if (Cs >= 947) {
return 1;
}
}
else if (Rs >= 524) {
if (Rs < 661) {
if (Cs < 1068) {
return 8;
}
else if (Cs >= 1068) {
return 1;
}
}
else if (Rs >= 661) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 1673) {
if (Cs < 732) {
if (MVDiff < 11) {
if (posBalance < 56) {
if (mu_mv_mag_sq < 113) {
if (posBalance < 38) {
if (Cs < 696) {
return 1;
}
else if (Cs >= 696) {
return 8;
}
}
else if (posBalance >= 38) {
return 8;
}
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
else if (posBalance >= 56) {
if (Cs < 656) {
return 8;
}
else if (Cs >= 656) {
return 1;
}
}
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (Cs >= 732) {
return 8;
}
}
else if (Rs >= 1673) {
if (Cs < 918) {
if (mu_mv_mag_sq < 113) {
if (Cs < 633) {
return 1;
}
else if (Cs >= 633) {
return 8;
}
}
else if (mu_mv_mag_sq >= 113) {
return 1;
}
}
else if (Cs >= 918) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 1220) {
if (Rs < 631) {
return 1;
}
else if (Rs >= 631) {
if (ssDCval < 95) {
return 2;
}
else if (ssDCval >= 95) {
if (Rs < 667) {
if (Rs < 653) {
return 8;
}
else if (Rs >= 653) {
return 1;
}
}
else if (Rs >= 667) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 916) {
return 8;
}
else if (Rs >= 916) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1162) {
if (Rs < 978) {
return 8;
}
else if (Rs >= 978) {
return 2;
}
}
else if (Rs >= 1162) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2373) {
if (MVDiff < 14) {
if (refDCval < 109) {
if (posBalance < 48) {
if (RsCsDiff < 10) {
if (Cs < 1037) {
if (ssDCval < 82) {
if (MVDiff < 3) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (Rs < 2744) {
return 8;
}
else if (Rs >= 2744) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
return 8;
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
if (Rs < 2504) {
return 4;
}
else if (Rs >= 2504) {
return 8;
}
}
}
}
else if (Cs >= 1037) {
if (SC < 4573) {
if (Rs < 3122) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Rs >= 3122) {
if (Rs < 3270) {
return 2;
}
else if (Rs >= 3270) {
return 8;
}
}
}
else if (SC >= 4573) {
if (SC < 5433) {
if (ssDCval < 74) {
if (ssDCval < 54) {
if (Cs < 1240) {
return 8;
}
else if (Cs >= 1240) {
return 4;
}
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (ssDCval >= 74) {
if (refDCval < 76) {
return 4;
}
else if (refDCval >= 76) {
if (RsDiff < 2) {
if (Rs < 4218) {
return 8;
}
else if (Rs >= 4218) {
return 4;
}
}
else if (RsDiff >= 2) {
if (Cs < 1285) {
return 4;
}
else if (Cs >= 1285) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Rs < 3361) {
return 8;
}
else if (Rs >= 3361) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 5433) {
if (Rs < 4576) {
return 8;
}
else if (Rs >= 4576) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 10) {
if (negBalance < 51) {
if (refDCval < 97) {
if (Rs < 3817) {
if (refDCval < 60) {
if (posBalance < 45) {
if (MVDiff < 1) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 2;
}
}
else if (MVDiff >= 1) {
if (refDCval < 58) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (MVDiff < 2) {
if (Rs < 2815) {
return 4;
}
else if (Rs >= 2815) {
return 2;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (refDCval >= 58) {
if (Rs < 3172) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (Rs >= 3172) {
return 4;
}
}
}
}
else if (posBalance >= 45) {
if (Cs < 945) {
return 4;
}
else if (Cs >= 945) {
if (mu_mv_mag_sq < 104) {
return 8;
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 2652) {
return 4;
}
else if (Rs >= 2652) {
return 4;
}
}
}
}
}
else if (refDCval >= 60) {
if (Cs < 557) {
if (Rs < 2612) {
return 1;
}
else if (Rs >= 2612) {
return 8;
}
}
else if (Cs >= 557) {
if (mu_mv_mag_sq < 80) {
if (ssDCval < 69) {
if (posBalance < 39) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 2;
}
}
else if (posBalance >= 39) {
if (CsDiff < 6) {
if (SC < 3512) {
if (Rs < 2546) {
return 2;
}
else if (Rs >= 2546) {
if (Rs < 2631) {
return 4;
}
else if (Rs >= 2631) {
return 2;
}
}
}
else if (SC >= 3512) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (negBalance < 42) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (Cs < 1072) {
return 4;
}
else if (Cs >= 1072) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
else if (negBalance >= 42) {
if (RsDiff < 14) {
if (Cs < 874) {
return 4;
}
else if (Cs >= 874) {
return 2;
}
}
else if (RsDiff >= 14) {
return 4;
}
}
}
}
}
}
else if (CsDiff >= 6) {
if (Cs < 989) {
return 4;
}
else if (Cs >= 989) {
return 2;
}
}
}
}
else if (ssDCval >= 69) {
if (refDCval < 84) {
if (SC < 4009) {
if (refDCval < 76) {
if (negBalance < 38) {
if (Cs < 994) {
return 1;
}
else if (Cs >= 994) {
return 2;
}
}
else if (negBalance >= 38) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (Rs < 2955) {
if (Cs < 941) {
return 1;
}
else if (Cs >= 941) {
return 4;
}
}
else if (Rs >= 2955) {
return 2;
}
}
}
}
else if (refDCval >= 76) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (SC >= 4009) {
if (RsDiff < 10) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (Rs < 3190) {
return 1;
}
else if (Rs >= 3190) {
if (negBalance < 37) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 1;
}
}
else if (negBalance >= 37) {
return 2;
}
}
}
}
else if (RsDiff >= 10) {
if (diffRsCsdiff < 59) {
if (posBalance < 46) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
return 4;
}
}
else if (posBalance >= 46) {
if (Rs < 3598) {
return 2;
}
else if (Rs >= 3598) {
return 1;
}
}
}
else if (diffRsCsdiff >= 59) {
return 4;
}
}
}
}
else if (refDCval >= 84) {
if (tcor < 96) {
if (mu_mv_mag_sq < 45) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 2;
}
}
else if (mu_mv_mag_sq >= 45) {
if (MVDiff < 9) {
if (ssDCval < 96) {
if (diffRsCsdiff < -10) {
if (Rs < 2719) {
return 4;
}
else if (Rs >= 2719) {
return 4;
}
}
else if (diffRsCsdiff >= -10) {
if (Rs < 2636) {
return 8;
}
else if (Rs >= 2636) {
if (diffAFD < 1) {
if (Rs < 3726) {
if (Rs < 3595) {
return 4;
}
else if (Rs >= 3595) {
return 8;
}
}
else if (Rs >= 3726) {
return 4;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
}
else if (tcor >= 96) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (posBalance < 43) {
if (SC < 4331) {
if (posBalance < 41) {
if (diffRsCsdiff < -17) {
return 1;
}
else if (diffRsCsdiff >= -17) {
return 2;
}
}
else if (posBalance >= 41) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (SC >= 4331) {
if (negBalance < 28) {
return 1;
}
else if (negBalance >= 28) {
if (ssDCval < 75) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (ssDCval >= 75) {
if (Rs < 3366) {
return 4;
}
else if (Rs >= 3366) {
return 2;
}
}
}
}
}
else if (posBalance >= 43) {
if (diffRsCsdiff < 9) {
if (RsCsDiff < 12) {
return 8;
}
else if (RsCsDiff >= 12) {
return 2;
}
}
else if (diffRsCsdiff >= 9) {
if (diffTSC < 1) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
if (Rs < 2589) {
return 1;
}
else if (Rs >= 2589) {
return 2;
}
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 3817) {
if (refDCval < 90) {
if (negBalance < 29) {
if (diffRsCsdiff < 12) {
if (Cs < 1244) {
return 4;
}
else if (Cs >= 1244) {
if (Rs < 3898) {
return 4;
}
else if (Rs >= 3898) {
return 1;
}
}
}
else if (diffRsCsdiff >= 12) {
if (Cs < 1025) {
return 4;
}
else if (Cs >= 1025) {
return 1;
}
}
}
else if (negBalance >= 29) {
return 4;
}
}
else if (refDCval >= 90) {
if (negBalance < 34) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (diffRsCsdiff < -3) {
if (posBalance < 43) {
if (Rs < 3973) {
return 4;
}
else if (Rs >= 3973) {
return 8;
}
}
else if (posBalance >= 43) {
return 8;
}
}
else if (diffRsCsdiff >= -3) {
if (MVDiff < 9) {
if (mu_mv_mag_sq < 61) {
if (posBalance < 43) {
if (Cs < 1198) {
return 4;
}
else if (Cs >= 1198) {
return 4;
}
}
else if (posBalance >= 43) {
if (Cs < 1163) {
return 8;
}
else if (Cs >= 1163) {
if (Cs < 1222) {
return 4;
}
else if (Cs >= 1222) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
return 4;
}
}
else if (MVDiff >= 9) {
if (Rs < 3993) {
return 8;
}
else if (Rs >= 3993) {
return 8;
}
}
}
}
}
else if (negBalance >= 34) {
return 2;
}
}
}
}
else if (refDCval >= 97) {
if (mu_mv_mag_sq < 100) {
if (posBalance < 36) {
if (Cs < 953) {
return 1;
}
else if (Cs >= 953) {
return 8;
}
}
else if (posBalance >= 36) {
if (Rs < 3393) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (Rs >= 3393) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (ssDCval < 101) {
if (RsDiff < 6) {
if (Rs < 3437) {
return 8;
}
else if (Rs >= 3437) {
return 8;
}
}
else if (RsDiff >= 6) {
if (Rs < 3448) {
return 8;
}
else if (Rs >= 3448) {
return 4;
}
}
}
else if (ssDCval >= 101) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 1105) {
if (Rs < 4323) {
return 2;
}
else if (Rs >= 4323) {
return 8;
}
}
else if (Cs >= 1105) {
return 4;
}
}
}
}
else if (negBalance >= 51) {
if (RsDiff < 4) {
if (Rs < 2511) {
return 4;
}
else if (Rs >= 2511) {
return 4;
}
}
else if (RsDiff >= 4) {
return 2;
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Rs < 3908) {
if (Cs < 256) {
if (Rs < 2420) {
return 2;
}
else if (Rs >= 2420) {
return 1;
}
}
else if (Cs >= 256) {
if (negBalance < 21) {
if (SC < 3216) {
return 1;
}
else if (SC >= 3216) {
if (Cs < 1121) {
if (tcor < 96) {
if (Rs < 2617) {
return 8;
}
else if (Rs >= 2617) {
return 8;
}
}
else if (tcor >= 96) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
else if (Cs >= 1121) {
return 8;
}
}
}
else if (negBalance >= 21) {
if (MVDiff < 7) {
if (refDCval < 63) {
if (Rs < 2745) {
if (Cs < 653) {
return 1;
}
else if (Cs >= 653) {
return 2;
}
}
else if (Rs >= 2745) {
return 4;
}
}
else if (refDCval >= 63) {
return 2;
}
}
else if (MVDiff >= 7) {
if (ssDCval < 80) {
if (TSC < 5) {
if (Rs < 2473) {
return 1;
}
else if (Rs >= 2473) {
return 4;
}
}
else if (TSC >= 5) {
if (Cs < 373) {
return 4;
}
else if (Cs >= 373) {
return 2;
}
}
}
else if (ssDCval >= 80) {
if (Rs < 2767) {
return 1;
}
else if (Rs >= 2767) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3908) {
return 4;
}
}
else if (CsDiff >= 3) {
if (ssDCval < 64) {
if (Cs < 967) {
if (tcor < 84) {
if (Rs < 2638) {
return 2;
}
else if (Rs >= 2638) {
return 4;
}
}
else if (tcor >= 84) {
return 2;
}
}
else if (Cs >= 967) {
if (Rs < 2961) {
if (Rs < 2431) {
return 8;
}
else if (Rs >= 2431) {
if (Cs < 987) {
if (Cs < 976) {
return 4;
}
else if (Cs >= 976) {
return 2;
}
}
else if (Cs >= 987) {
return 4;
}
}
}
else if (Rs >= 2961) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (diffRsCsdiff < -13) {
return 4;
}
else if (diffRsCsdiff >= -13) {
return 2;
}
}
}
}
}
else if (ssDCval >= 64) {
if (negBalance < 38) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 74) {
if (mcTcor < 98) {
if (Cs < 916) {
if (mcTcor < 95) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 4;
}
}
else if (mcTcor >= 95) {
return 4;
}
}
else if (Cs >= 916) {
if (ssDCval < 78) {
if (posBalance < 52) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
return 4;
}
}
else if (posBalance >= 52) {
return 2;
}
}
else if (ssDCval >= 78) {
if (Rs < 3104) {
return 1;
}
else if (Rs >= 3104) {
if (Rs < 3855) {
return 4;
}
else if (Rs >= 3855) {
if (Rs < 4187) {
return 4;
}
else if (Rs >= 4187) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (Rs < 2836) {
return 4;
}
else if (Rs >= 2836) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 74) {
return 2;
}
}
else if (posBalance >= 55) {
if (RsDiff < 5) {
if (Rs < 3930) {
return 8;
}
else if (Rs >= 3930) {
return 2;
}
}
else if (RsDiff >= 5) {
if (Cs < 635) {
if (Rs < 3053) {
return 1;
}
else if (Rs >= 3053) {
return 2;
}
}
else if (Cs >= 635) {
if (mcTcor < 99) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 4;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
}
}
else if (negBalance >= 38) {
if (diffRsCsdiff < -191) {
return 4;
}
else if (diffRsCsdiff >= -191) {
if (tcor < 77) {
if (RsDiff < 15) {
if (Rs < 3525) {
return 4;
}
else if (Rs >= 3525) {
return 1;
}
}
else if (RsDiff >= 15) {
if (negBalance < 47) {
if (SC < 3466) {
return 2;
}
else if (SC >= 3466) {
return 4;
}
}
else if (negBalance >= 47) {
return 2;
}
}
}
else if (tcor >= 77) {
if (Rs < 2615) {
if (Cs < 977) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
if (refDCval < 85) {
return 2;
}
else if (refDCval >= 85) {
return 1;
}
}
}
else if (Cs >= 977) {
if (diffMVdiffVal < 4) {
if (mcTcor < 94) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 2;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (diffMVdiffVal >= 4) {
if (Rs < 2402) {
return 8;
}
else if (Rs >= 2402) {
return 4;
}
}
}
}
else if (Rs >= 2615) {
if (RsDiff < 24) {
if (Cs < 1313) {
return 2;
}
else if (Cs >= 1313) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
else if (RsDiff >= 24) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsDiff < 10) {
if (tcor < 96) {
if (diffRsCsdiff < -45) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
if (Rs < 2754) {
return 8;
}
else if (Rs >= 2754) {
return 4;
}
}
}
else if (diffRsCsdiff >= -45) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (TSC < 6) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
else if (TSC >= 6) {
if (MVDiff < 6) {
if (Rs < 3923) {
return 8;
}
else if (Rs >= 3923) {
return 2;
}
}
else if (MVDiff >= 6) {
if (Cs < 1245) {
if (Rs < 2568) {
return 4;
}
else if (Rs >= 2568) {
if (MVDiff < 7) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 4;
}
}
else if (MVDiff >= 7) {
if (ssDCval < 99) {
return 8;
}
else if (ssDCval >= 99) {
return 8;
}
}
}
}
else if (Cs >= 1245) {
return 1;
}
}
}
}
}
}
else if (tcor >= 96) {
if (Rs < 2660) {
return 8;
}
else if (Rs >= 2660) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (Rs < 3172) {
return 8;
}
else if (Rs >= 3172) {
return 2;
}
}
}
}
}
else if (RsDiff >= 10) {
if (negBalance < 50) {
if (ssDCval < 107) {
return 2;
}
else if (ssDCval >= 107) {
if (RsDiff < 14) {
if (Rs < 3712) {
return 4;
}
else if (Rs >= 3712) {
return 8;
}
}
else if (RsDiff >= 14) {
return 2;
}
}
}
else if (negBalance >= 50) {
if (Cs < 832) {
return 2;
}
else if (Cs >= 832) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (refDCval < 102) {
if (tcor < 80) {
if (Cs < 939) {
return 4;
}
else if (Cs >= 939) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
return 2;
}
}
}
else if (tcor >= 80) {
if (Rs < 3272) {
return 4;
}
else if (Rs >= 3272) {
return 2;
}
}
}
else if (refDCval >= 102) {
if (RsCsDiff < 262) {
return 4;
}
else if (RsCsDiff >= 262) {
if (Cs < 1165) {
return 2;
}
else if (Cs >= 1165) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 3427) {
if (mcTcor < 97) {
if (RsDiff < 8) {
if (tcor < 92) {
if (Rs < 3328) {
return 8;
}
else if (Rs >= 3328) {
return 2;
}
}
else if (tcor >= 92) {
if (diffTSC < 2) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (SC < 3807) {
if (SC < 3685) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 2454) {
return 1;
}
else if (Rs >= 2454) {
return 8;
}
}
}
else if (SC >= 3685) {
return 1;
}
}
else if (SC >= 3807) {
return 8;
}
}
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (RsDiff >= 8) {
if (Cs < 734) {
if (mcTcor < 94) {
return 2;
}
else if (mcTcor >= 94) {
if (MVDiff < 13) {
if (gchDC < 5) {
return 1;
}
else if (gchDC >= 5) {
return 2;
}
}
else if (MVDiff >= 13) {
return 2;
}
}
}
else if (Cs >= 734) {
if (negBalance < 54) {
if (AFD < 14) {
if (Cs < 878) {
return 2;
}
else if (Cs >= 878) {
return 8;
}
}
else if (AFD >= 14) {
if (Cs < 1070) {
return 2;
}
else if (Cs >= 1070) {
return 1;
}
}
}
else if (negBalance >= 54) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (Rs < 2796) {
return 8;
}
else if (Rs >= 2796) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 7) {
if (TSC < 3) {
if (ssDCval < 117) {
if (Rs < 3392) {
return 8;
}
else if (Rs >= 3392) {
return 2;
}
}
else if (ssDCval >= 117) {
if (Rs < 2718) {
return 1;
}
else if (Rs >= 2718) {
return 2;
}
}
}
else if (TSC >= 3) {
if (Rs < 2506) {
if (Cs < 544) {
return 2;
}
else if (Cs >= 544) {
if (Rs < 2461) {
return 8;
}
else if (Rs >= 2461) {
return 1;
}
}
}
else if (Rs >= 2506) {
if (Cs < 530) {
if (ssDCval < 118) {
return 8;
}
else if (ssDCval >= 118) {
return 2;
}
}
else if (Cs >= 530) {
return 2;
}
}
}
}
else if (MVDiff >= 7) {
if (RsCsDiff < 80) {
if (CsDiff < 3) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (diffRsCsdiff < 1) {
if (Rs < 2491) {
return 1;
}
else if (Rs >= 2491) {
return 2;
}
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (diffRsCsdiff < -17) {
if (ssDCval < 125) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (ssDCval >= 125) {
if (Rs < 2562) {
return 1;
}
else if (Rs >= 2562) {
return 2;
}
}
}
else if (diffRsCsdiff >= -17) {
if (posBalance < 86) {
return 8;
}
else if (posBalance >= 86) {
return 2;
}
}
}
}
else if (RsCsDiff >= 80) {
if (SC < 3287) {
return 1;
}
else if (SC >= 3287) {
if (Cs < 814) {
if (refDCval < 135) {
return 2;
}
else if (refDCval >= 135) {
return 1;
}
}
else if (Cs >= 814) {
if (ssDCval < 124) {
return 4;
}
else if (ssDCval >= 124) {
if (Rs < 2979) {
return 2;
}
else if (Rs >= 2979) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 3427) {
if (MVDiff < 6) {
if (refDCval < 138) {
if (ssDCval < 111) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
if (Rs < 4638) {
return 8;
}
else if (Rs >= 4638) {
return 2;
}
}
}
else if (ssDCval >= 111) {
if (Cs < 558) {
if (Cs < 549) {
return 2;
}
else if (Cs >= 549) {
return 8;
}
}
else if (Cs >= 558) {
return 2;
}
}
}
else if (refDCval >= 138) {
if (CsDiff < 4) {
if (Rs < 3967) {
if (mu_mv_mag_sq < 109) {
return 2;
}
else if (mu_mv_mag_sq >= 109) {
if (posBalance < 40) {
if (diffRsCsdiff < -6) {
if (Rs < 3708) {
return 2;
}
else if (Rs >= 3708) {
return 1;
}
}
else if (diffRsCsdiff >= -6) {
return 2;
}
}
else if (posBalance >= 40) {
return 1;
}
}
}
else if (Rs >= 3967) {
return 2;
}
}
else if (CsDiff >= 4) {
if (Rs < 4405) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (Rs < 3809) {
return 1;
}
else if (Rs >= 3809) {
return 2;
}
}
}
else if (Rs >= 4405) {
return 2;
}
}
}
}
else if (MVDiff >= 6) {
if (negBalance < 54) {
if (refDCval < 137) {
if (RsDiff < 5) {
if (Rs < 3513) {
return 2;
}
else if (Rs >= 3513) {
return 1;
}
}
else if (RsDiff >= 5) {
if (Cs < 1005) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 2;
}
}
else if (Cs >= 1005) {
if (diffMVdiffVal < -12) {
return 1;
}
else if (diffMVdiffVal >= -12) {
if (MVDiff < 12) {
if (Cs < 1305) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (refDCval < 130) {
return 2;
}
else if (refDCval >= 130) {
return 1;
}
}
}
else if (Cs >= 1305) {
return 8;
}
}
else if (MVDiff >= 12) {
if (Rs < 3982) {
return 8;
}
else if (Rs >= 3982) {
if (Rs < 5286) {
return 4;
}
else if (Rs >= 5286) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 137) {
if (CsDiff < 3) {
if (Rs < 3570) {
return 8;
}
else if (Rs >= 3570) {
return 2;
}
}
else if (CsDiff >= 3) {
if (RsDiff < 17) {
if (SC < 4441) {
return 2;
}
else if (SC >= 4441) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
if (posBalance < 45) {
if (Rs < 3823) {
return 1;
}
else if (Rs >= 3823) {
return 2;
}
}
else if (posBalance >= 45) {
return 1;
}
}
}
}
else if (RsDiff >= 17) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (posBalance < 61) {
if (diffMVdiffVal < -23) {
if (Rs < 5026) {
return 1;
}
else if (Rs >= 5026) {
return 2;
}
}
else if (diffMVdiffVal >= -23) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (RsDiff < 21) {
return 4;
}
else if (RsDiff >= 21) {
return 2;
}
}
}
}
else if (posBalance >= 61) {
if (ssDCval < 125) {
if (Cs < 1096) {
return 2;
}
else if (Cs >= 1096) {
return 4;
}
}
else if (ssDCval >= 125) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 14) {
if (mcTcor < 68) {
if (MVDiff < 42) {
return 1;
}
else if (MVDiff >= 42) {
if (refDCval < 88) {
return 8;
}
else if (refDCval >= 88) {
if (AFD < 30) {
if (Cs < 1140) {
return 1;
}
else if (Cs >= 1140) {
return 4;
}
}
else if (AFD >= 30) {
return 2;
}
}
}
}
else if (mcTcor >= 68) {
if (RsCsDiff < 248) {
if (ssDCval < 88) {
if (Rs < 3710) {
if (RsCsDiff < 53) {
if (refDCval < 80) {
if (SC < 3306) {
if (posBalance < 32) {
return 4;
}
else if (posBalance >= 32) {
return 8;
}
}
else if (SC >= 3306) {
if (refDCval < 72) {
if (posBalance < 36) {
if (AFD < 7) {
if (Rs < 2554) {
return 4;
}
else if (Rs >= 2554) {
return 8;
}
}
else if (AFD >= 7) {
return 1;
}
}
else if (posBalance >= 36) {
return 4;
}
}
else if (refDCval >= 72) {
return 8;
}
}
}
else if (refDCval >= 80) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
return 2;
}
}
}
else if (RsCsDiff >= 53) {
if (negBalance < 47) {
if (posBalance < 56) {
if (ssDCval < 77) {
if (posBalance < 38) {
if (Cs < 1050) {
return 1;
}
else if (Cs >= 1050) {
if (Rs < 2905) {
return 4;
}
else if (Rs >= 2905) {
return 8;
}
}
}
else if (posBalance >= 38) {
if (Rs < 3244) {
if (Cs < 737) {
if (Rs < 2673) {
if (Rs < 2574) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 1;
}
}
else if (Rs >= 2574) {
return 2;
}
}
else if (Rs >= 2673) {
return 4;
}
}
else if (Cs >= 737) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
if (tcor < 84) {
if (AFD < 17) {
if (Cs < 1068) {
if (Cs < 871) {
return 2;
}
else if (Cs >= 871) {
return 4;
}
}
else if (Cs >= 1068) {
if (ssDCval < 71) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (ssDCval >= 71) {
if (Cs < 1141) {
return 2;
}
else if (Cs >= 1141) {
return 4;
}
}
}
}
else if (AFD >= 17) {
return 2;
}
}
else if (tcor >= 84) {
if (Rs < 2516) {
return 1;
}
else if (Rs >= 2516) {
return 2;
}
}
}
}
}
else if (Rs >= 3244) {
return 4;
}
}
}
else if (ssDCval >= 77) {
if (Rs < 2788) {
if (MVDiff < 16) {
if (Rs < 2542) {
return 1;
}
else if (Rs >= 2542) {
return 2;
}
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (Rs >= 2788) {
return 4;
}
}
}
else if (posBalance >= 56) {
if (AFD < 11) {
if (SC < 3586) {
return 1;
}
else if (SC >= 3586) {
return 4;
}
}
else if (AFD >= 11) {
if (RsDiff < 10) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 4;
}
}
else if (RsDiff >= 10) {
return 2;
}
}
}
}
else if (negBalance >= 47) {
if (MVDiff < 33) {
if (posBalance < 43) {
if (negBalance < 51) {
return 1;
}
else if (negBalance >= 51) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
}
else if (MVDiff >= 33) {
if (Cs < 938) {
if (Cs < 435) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (Cs >= 435) {
if (MVDiff < 65) {
return 1;
}
else if (MVDiff >= 65) {
if (Rs < 2501) {
return 8;
}
else if (Rs >= 2501) {
return 4;
}
}
}
}
else if (Cs >= 938) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
if (MVDiff < 38) {
return 1;
}
else if (MVDiff >= 38) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 3710) {
if (Cs < 1041) {
return 4;
}
else if (Cs >= 1041) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
return 4;
}
}
}
}
else if (ssDCval >= 88) {
if (mcTcor < 93) {
if (refDCval < 133) {
if (MVDiff < 32) {
if (RsCsDiff < 121) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 8;
}
}
else if (RsCsDiff >= 121) {
if (AFD < 18) {
if (Cs < 999) {
if (Cs < 269) {
return 4;
}
else if (Cs >= 269) {
return 2;
}
}
else if (Cs >= 999) {
if (Rs < 2744) {
return 8;
}
else if (Rs >= 2744) {
return 4;
}
}
}
else if (AFD >= 18) {
if (mcTcor < 92) {
if (negBalance < 57) {
if (Rs < 3470) {
return 4;
}
else if (Rs >= 3470) {
return 2;
}
}
else if (negBalance >= 57) {
if (Cs < 1050) {
return 1;
}
else if (Cs >= 1050) {
return 2;
}
}
}
else if (mcTcor >= 92) {
if (Cs < 659) {
return 1;
}
else if (Cs >= 659) {
return 2;
}
}
}
}
}
else if (MVDiff >= 32) {
if (mcTcor < 84) {
if (Cs < 1098) {
if (diffAFD < 6) {
if (MVDiff < 102) {
if (Rs < 3961) {
if (mcTcor < 83) {
return 8;
}
else if (mcTcor >= 83) {
return 2;
}
}
else if (Rs >= 3961) {
return 2;
}
}
else if (MVDiff >= 102) {
return 2;
}
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (Cs >= 1098) {
if (diffTSC < -1) {
if (Rs < 2738) {
return 8;
}
else if (Rs >= 2738) {
return 2;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
}
else if (mcTcor >= 84) {
if (RsDiff < 11) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (MVDiff < 81) {
if (Cs < 1009) {
if (Cs < 357) {
return 8;
}
else if (Cs >= 357) {
if (CsDiff < 2) {
if (Rs < 2607) {
return 1;
}
else if (Rs >= 2607) {
return 8;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
}
else if (Cs >= 1009) {
return 8;
}
}
else if (MVDiff >= 81) {
if (negBalance < 58) {
if (Rs < 3041) {
return 2;
}
else if (Rs >= 3041) {
return 4;
}
}
else if (negBalance >= 58) {
return 1;
}
}
}
}
else if (RsDiff >= 11) {
if (Rs < 2515) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 1;
}
}
else if (Rs >= 2515) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
if (Rs < 2727) {
return 2;
}
else if (Rs >= 2727) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 133) {
if (negBalance < 50) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (negBalance >= 50) {
if (MVDiff < 54) {
if (Cs < 1063) {
return 4;
}
else if (Cs >= 1063) {
if (Rs < 2964) {
if (Rs < 2809) {
return 8;
}
else if (Rs >= 2809) {
return 1;
}
}
else if (Rs >= 2964) {
return 4;
}
}
}
else if (MVDiff >= 54) {
if (Cs < 859) {
return 8;
}
else if (Cs >= 859) {
return 1;
}
}
}
}
}
else if (mcTcor >= 93) {
if (SC < 3336) {
if (ssDCval < 116) {
if (negBalance < 36) {
if (Cs < 255) {
return 2;
}
else if (Cs >= 255) {
return 1;
}
}
else if (negBalance >= 36) {
if (diffAFD < -2) {
if (refDCval < 106) {
if (Rs < 2434) {
return 2;
}
else if (Rs >= 2434) {
return 4;
}
}
else if (refDCval >= 106) {
if (mcTcor < 94) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
}
else if (diffAFD >= -2) {
if (RsCsDiff < 183) {
if (posBalance < 67) {
if (mu_mv_mag_sq < 32) {
return 8;
}
else if (mu_mv_mag_sq >= 32) {
return 1;
}
}
else if (posBalance >= 67) {
return 1;
}
}
else if (RsCsDiff >= 183) {
if (negBalance < 53) {
return 2;
}
else if (negBalance >= 53) {
if (Rs < 2641) {
return 1;
}
else if (Rs >= 2641) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 116) {
if (CsDiff < 3) {
if (posBalance < 72) {
if (diffTSC < -4) {
if (Rs < 2679) {
return 2;
}
else if (Rs >= 2679) {
return 8;
}
}
else if (diffTSC >= -4) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
if (Cs < 427) {
if (gchDC < 2) {
if (Rs < 2622) {
return 8;
}
else if (Rs >= 2622) {
return 1;
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (Cs >= 427) {
if (MVDiff < 44) {
return 2;
}
else if (MVDiff >= 44) {
return 8;
}
}
}
}
}
else if (posBalance >= 72) {
if (Cs < 187) {
return 8;
}
else if (Cs >= 187) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 2891) {
if (mu_mv_mag_sq < 32) {
return 1;
}
else if (mu_mv_mag_sq >= 32) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
return 1;
}
}
}
else if (Rs >= 2891) {
return 8;
}
}
}
}
else if (SC >= 3336) {
if (RsCsDiff < 53) {
if (SC < 4728) {
if (Cs < 1022) {
if (Cs < 862) {
if (Rs < 3133) {
if (Rs < 3107) {
return 8;
}
else if (Rs >= 3107) {
return 1;
}
}
else if (Rs >= 3133) {
return 8;
}
}
else if (Cs >= 862) {
return 1;
}
}
else if (Cs >= 1022) {
if (mu_mv_mag_sq < 3) {
if (Rs < 3216) {
return 8;
}
else if (Rs >= 3216) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (SC >= 4728) {
if (Rs < 3800) {
return 4;
}
else if (Rs >= 3800) {
return 2;
}
}
}
else if (RsCsDiff >= 53) {
if (ssDCval < 127) {
if (Rs < 5226) {
if (diffRsCsdiff < -95) {
if (refDCval < 106) {
if (diffTSC < -3) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 4;
}
}
else if (diffTSC >= -3) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (Rs < 4242) {
return 2;
}
else if (Rs >= 4242) {
return 8;
}
}
}
}
else if (refDCval >= 106) {
if (Cs < 921) {
if (Rs < 3067) {
return 1;
}
else if (Rs >= 3067) {
return 2;
}
}
else if (Cs >= 921) {
if (Cs < 1255) {
return 8;
}
else if (Cs >= 1255) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= -95) {
if (refDCval < 94) {
if (SC < 3814) {
if (SC < 3557) {
return 2;
}
else if (SC >= 3557) {
return 8;
}
}
else if (SC >= 3814) {
return 2;
}
}
else if (refDCval >= 94) {
if (CsDiff < 3) {
if (MVDiff < 19) {
if (refDCval < 112) {
return 8;
}
else if (refDCval >= 112) {
return 2;
}
}
else if (MVDiff >= 19) {
if (Cs < 381) {
return 4;
}
else if (Cs >= 381) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 2;
}
}
}
}
else if (CsDiff >= 3) {
if (Cs < 1182) {
if (MVDiff < 33) {
if (refDCval < 107) {
if (Rs < 3248) {
return 2;
}
else if (Rs >= 3248) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
return 8;
}
}
}
else if (refDCval >= 107) {
if (negBalance < 37) {
if (Rs < 3478) {
return 1;
}
else if (Rs >= 3478) {
return 2;
}
}
else if (negBalance >= 37) {
if (posBalance < 61) {
if (refDCval < 123) {
return 8;
}
else if (refDCval >= 123) {
if (Rs < 3098) {
return 1;
}
else if (Rs >= 3098) {
return 2;
}
}
}
else if (posBalance >= 61) {
if (SC < 3796) {
return 1;
}
else if (SC >= 3796) {
return 8;
}
}
}
}
}
else if (MVDiff >= 33) {
if (AFD < 13) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 2650) {
return 2;
}
else if (Rs >= 2650) {
return 1;
}
}
}
else if (AFD >= 13) {
if (diffRsCsdiff < 43) {
if (Rs < 2659) {
return 8;
}
else if (Rs >= 2659) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (diffRsCsdiff >= 43) {
return 8;
}
}
}
}
else if (Cs >= 1182) {
if (SC < 3928) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (refDCval < 122) {
return 1;
}
else if (refDCval >= 122) {
return 8;
}
}
}
else if (SC >= 3928) {
return 8;
}
}
}
}
}
}
else if (Rs >= 5226) {
return 2;
}
}
else if (ssDCval >= 127) {
if (CsDiff < 4) {
if (mcTcor < 95) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Cs < 701) {
return 8;
}
else if (Cs >= 701) {
return 2;
}
}
}
else if (mcTcor >= 95) {
if (diffRsCsdiff < -62) {
if (negBalance < 38) {
if (Rs < 3974) {
return 8;
}
else if (Rs >= 3974) {
return 2;
}
}
else if (negBalance >= 38) {
return 1;
}
}
else if (diffRsCsdiff >= -62) {
if (refDCval < 123) {
return 1;
}
else if (refDCval >= 123) {
if (posBalance < 49) {
if (Cs < 707) {
return 1;
}
else if (Cs >= 707) {
return 2;
}
}
else if (posBalance >= 49) {
return 2;
}
}
}
}
}
else if (CsDiff >= 4) {
if (posBalance < 62) {
if (Rs < 6954) {
if (posBalance < 44) {
if (refDCval < 147) {
return 2;
}
else if (refDCval >= 147) {
return 1;
}
}
else if (posBalance >= 44) {
return 1;
}
}
else if (Rs >= 6954) {
return 2;
}
}
else if (posBalance >= 62) {
if (negBalance < 45) {
if (TSC < 7) {
if (Rs < 3693) {
return 8;
}
else if (Rs >= 3693) {
return 2;
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (negBalance >= 45) {
if (Cs < 1210) {
return 8;
}
else if (Cs >= 1210) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 248) {
if (refDCval < 135) {
if (refDCval < 114) {
if (MVDiff < 35) {
if (ssDCval < 72) {
if (negBalance < 48) {
if (posBalance < 63) {
if (diffRsCsdiff < 47) {
if (Rs < 3320) {
return 2;
}
else if (Rs >= 3320) {
return 4;
}
}
else if (diffRsCsdiff >= 47) {
if (Rs < 2572) {
return 1;
}
else if (Rs >= 2572) {
return 4;
}
}
}
else if (posBalance >= 63) {
return 2;
}
}
else if (negBalance >= 48) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 1;
}
}
}
else if (ssDCval >= 72) {
if (Cs < 856) {
if (refDCval < 78) {
if (Rs < 3645) {
if (Cs < 155) {
return 1;
}
else if (Cs >= 155) {
return 2;
}
}
else if (Rs >= 3645) {
return 4;
}
}
else if (refDCval >= 78) {
if (AFD < 28) {
if (Cs < 785) {
return 2;
}
else if (Cs >= 785) {
if (Rs < 2788) {
if (Rs < 2459) {
return 8;
}
else if (Rs >= 2459) {
if (Rs < 2533) {
return 2;
}
else if (Rs >= 2533) {
return 4;
}
}
}
else if (Rs >= 2788) {
return 2;
}
}
}
else if (AFD >= 28) {
if (RsDiff < 18) {
if (Rs < 2899) {
return 8;
}
else if (Rs >= 2899) {
return 1;
}
}
else if (RsDiff >= 18) {
return 2;
}
}
}
}
else if (Cs >= 856) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2892) {
if (Rs < 2865) {
if (Rs < 2796) {
return 1;
}
else if (Rs >= 2796) {
return 2;
}
}
else if (Rs >= 2865) {
return 4;
}
}
else if (Rs >= 2892) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (tcor < 74) {
if (AFD < 25) {
if (negBalance < 57) {
if (posBalance < 48) {
if (Rs < 3280) {
return 4;
}
else if (Rs >= 3280) {
return 2;
}
}
else if (posBalance >= 48) {
return 8;
}
}
else if (negBalance >= 57) {
if (ssDCval < 100) {
if (Rs < 2976) {
return 2;
}
else if (Rs >= 2976) {
return 1;
}
}
else if (ssDCval >= 100) {
return 4;
}
}
}
else if (AFD >= 25) {
return 2;
}
}
else if (tcor >= 74) {
if (MVDiff < 24) {
if (mcTcor < 92) {
if (Cs < 869) {
return 2;
}
else if (Cs >= 869) {
return 4;
}
}
else if (mcTcor >= 92) {
if (Rs < 4798) {
return 2;
}
else if (Rs >= 4798) {
if (Rs < 5177) {
return 1;
}
else if (Rs >= 5177) {
return 4;
}
}
}
}
else if (MVDiff >= 24) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (diffMVdiffVal < 11) {
return 1;
}
else if (diffMVdiffVal >= 11) {
return 4;
}
}
}
}
}
}
}
}
else if (MVDiff >= 35) {
if (mcTcor < 83) {
if (RsCsDiff < 345) {
return 4;
}
else if (RsCsDiff >= 345) {
if (Rs < 2981) {
if (Cs < 901) {
if (Cs < 479) {
return 2;
}
else if (Cs >= 479) {
return 1;
}
}
else if (Cs >= 901) {
return 8;
}
}
else if (Rs >= 2981) {
if (Cs < 1099) {
return 2;
}
else if (Cs >= 1099) {
if (negBalance < 60) {
return 4;
}
else if (negBalance >= 60) {
return 2;
}
}
}
}
}
else if (mcTcor >= 83) {
if (refDCval < 82) {
if (MVDiff < 41) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
return 8;
}
}
else if (MVDiff >= 41) {
if (posBalance < 62) {
if (Cs < 125) {
return 1;
}
else if (Cs >= 125) {
if (Cs < 912) {
return 4;
}
else if (Cs >= 912) {
return 2;
}
}
}
else if (posBalance >= 62) {
if (RsDiff < 17) {
return 2;
}
else if (RsDiff >= 17) {
return 1;
}
}
}
}
else if (refDCval >= 82) {
if (TSC < 9) {
if (AFD < 31) {
if (MVDiff < 45) {
if (MVDiff < 43) {
return 2;
}
else if (MVDiff >= 43) {
return 8;
}
}
else if (MVDiff >= 45) {
return 2;
}
}
else if (AFD >= 31) {
return 8;
}
}
else if (TSC >= 9) {
if (negBalance < 58) {
if (refDCval < 109) {
if (mcTcor < 92) {
if (ssDCval < 91) {
if (Rs < 3150) {
return 4;
}
else if (Rs >= 3150) {
return 2;
}
}
else if (ssDCval >= 91) {
if (diffMVdiffVal < 4) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (CsDiff < 9) {
return 2;
}
else if (CsDiff >= 9) {
return 8;
}
}
}
else if (diffMVdiffVal >= 4) {
if (diffRsCsdiff < -55) {
if (MVDiff < 88) {
return 2;
}
else if (MVDiff >= 88) {
return 8;
}
}
else if (diffRsCsdiff >= -55) {
return 8;
}
}
}
}
else if (mcTcor >= 92) {
if (gchDC < 6) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (gchDC >= 6) {
if (Rs < 3942) {
return 1;
}
else if (Rs >= 3942) {
return 8;
}
}
}
}
else if (refDCval >= 109) {
if (negBalance < 50) {
if (Rs < 4276) {
return 1;
}
else if (Rs >= 4276) {
return 8;
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
else if (negBalance >= 58) {
if (diffAFD < -3) {
if (Cs < 363) {
return 2;
}
else if (Cs >= 363) {
return 1;
}
}
else if (diffAFD >= -3) {
if (Rs < 2448) {
return 1;
}
else if (Rs >= 2448) {
if (Cs < 1270) {
return 2;
}
else if (Cs >= 1270) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (Rs < 2978) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (diffAFD < 0) {
if (MVDiff < 25) {
return 2;
}
else if (MVDiff >= 25) {
if (Cs < 1182) {
return 1;
}
else if (Cs >= 1182) {
return 2;
}
}
}
else if (diffAFD >= 0) {
return 1;
}
}
}
else if (Rs >= 2978) {
if (Cs < 852) {
if (RsCsDiff < 462) {
if (CsDiff < 4) {
if (MVDiff < 45) {
return 2;
}
else if (MVDiff >= 45) {
if (RsDiff < 18) {
return 8;
}
else if (RsDiff >= 18) {
if (Rs < 3979) {
return 1;
}
else if (Rs >= 3979) {
return 2;
}
}
}
}
else if (CsDiff >= 4) {
if (mcTcor < 97) {
if (tcor < 77) {
return 2;
}
else if (tcor >= 77) {
if (tcor < 90) {
return 1;
}
else if (tcor >= 90) {
return 2;
}
}
}
else if (mcTcor >= 97) {
return 2;
}
}
}
else if (RsCsDiff >= 462) {
if (CsDiff < 8) {
if (diffAFD < 13) {
return 2;
}
else if (diffAFD >= 13) {
if (Rs < 4335) {
return 8;
}
else if (Rs >= 4335) {
return 2;
}
}
}
else if (CsDiff >= 8) {
return 1;
}
}
}
else if (Cs >= 852) {
if (CsDiff < 7) {
if (RsCsDiff < 431) {
if (AFD < 18) {
if (Cs < 985) {
return 1;
}
else if (Cs >= 985) {
return 8;
}
}
else if (AFD >= 18) {
if (Rs < 4737) {
return 8;
}
else if (Rs >= 4737) {
if (Rs < 4861) {
return 2;
}
else if (Rs >= 4861) {
return 8;
}
}
}
}
else if (RsCsDiff >= 431) {
if (ssDCval < 121) {
if (TSC < 8) {
if (MVDiff < 31) {
return 2;
}
else if (MVDiff >= 31) {
return 8;
}
}
else if (TSC >= 8) {
return 2;
}
}
else if (ssDCval >= 121) {
if (MVDiff < 25) {
return 2;
}
else if (MVDiff >= 25) {
if (tcor < 70) {
if (diffMVdiffVal < 4) {
return 2;
}
else if (diffMVdiffVal >= 4) {
if (Cs < 878) {
return 2;
}
else if (Cs >= 878) {
return 8;
}
}
}
else if (tcor >= 70) {
return 8;
}
}
}
}
}
else if (CsDiff >= 7) {
if (MVDiff < 70) {
if (tcor < 72) {
return 2;
}
else if (tcor >= 72) {
if (Cs < 1242) {
if (Rs < 4961) {
if (diffMVdiffVal < 6) {
return 1;
}
else if (diffMVdiffVal >= 6) {
if (Rs < 3306) {
return 8;
}
else if (Rs >= 3306) {
if (Rs < 3677) {
return 2;
}
else if (Rs >= 3677) {
return 1;
}
}
}
}
else if (Rs >= 4961) {
if (diffAFD < 2) {
if (Cs < 1193) {
if (Rs < 5681) {
return 2;
}
else if (Rs >= 5681) {
return 8;
}
}
else if (Cs >= 1193) {
return 1;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (Cs >= 1242) {
return 2;
}
}
}
else if (MVDiff >= 70) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 135) {
if (mcTcor < 90) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (CsDiff < 5) {
if (Rs < 4619) {
return 8;
}
else if (Rs >= 4619) {
return 2;
}
}
else if (CsDiff >= 5) {
return 1;
}
}
}
else if (mcTcor >= 90) {
if (Rs < 4922) {
if (AFD < 17) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (Rs < 4510) {
return 1;
}
else if (Rs >= 4510) {
return 2;
}
}
}
else if (AFD >= 17) {
if (MVDiff < 86) {
if (mu_mv_mag_sq < 34) {
if (CsDiff < 4) {
if (RsDiff < 18) {
return 1;
}
else if (RsDiff >= 18) {
if (AFD < 30) {
return 2;
}
else if (AFD >= 30) {
return 1;
}
}
}
else if (CsDiff >= 4) {
if (Cs < 289) {
return 2;
}
else if (Cs >= 289) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (Rs < 2473) {
return 1;
}
else if (Rs >= 2473) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 2;
}
}
}
}
else if (MVDiff >= 86) {
if (ssDCval < 152) {
return 1;
}
else if (ssDCval >= 152) {
if (AFD < 31) {
return 8;
}
else if (AFD >= 31) {
return 2;
}
}
}
}
}
else if (Rs >= 4922) {
if (Cs < 805) {
return 2;
}
else if (Cs >= 805) {
if (TSC < 10) {
if (tcor < 82) {
if (Rs < 5983) {
return 8;
}
else if (Rs >= 5983) {
return 2;
}
}
else if (tcor >= 82) {
return 2;
}
}
else if (TSC >= 10) {
if (posBalance < 49) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 8;
}
}
else if (posBalance >= 49) {
if (Rs < 6350) {
return 8;
}
else if (Rs >= 6350) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1325) {
if (mu_mv_mag_sq < 111) {
if (RsCsDiff < 101) {
if (refDCval < 105) {
if (negBalance < 39) {
if (Rs < 816) {
if (Rs < 300) {
return 4;
}
else if (Rs >= 300) {
if (ssDCval < 72) {
if (TSC < 5) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (TSC >= 5) {
if (Cs < 1876) {
if (Rs < 589) {
return 2;
}
else if (Rs >= 589) {
if (Cs < 1764) {
return 8;
}
else if (Cs >= 1764) {
return 2;
}
}
}
else if (Cs >= 1876) {
return 1;
}
}
}
else if (ssDCval >= 72) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
return 4;
}
}
}
}
else if (Rs >= 816) {
if (SC < 10165) {
if (posBalance < 44) {
if (ssDCval < 78) {
if (ssDCval < 64) {
if (Rs < 3313) {
if (ssDCval < 58) {
if (AFD < 8) {
if (diffAFD < 2) {
if (ssDCval < 51) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (ssDCval >= 51) {
if (Rs < 995) {
return 8;
}
else if (Rs >= 995) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (AFD >= 8) {
return 2;
}
}
else if (ssDCval >= 58) {
if (tcor < 96) {
return 2;
}
else if (tcor >= 96) {
return 8;
}
}
}
else if (Rs >= 3313) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 109) {
if (RsDiff < 8) {
if (diffRsCsdiff < -11) {
return 4;
}
else if (diffRsCsdiff >= -11) {
if (diffRsCsdiff < 5) {
return 8;
}
else if (diffRsCsdiff >= 5) {
return 4;
}
}
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (ssDCval >= 64) {
if (MVDiff < 7) {
if (Rs < 3514) {
if (posBalance < 40) {
if (Rs < 2686) {
return 8;
}
else if (Rs >= 2686) {
return 4;
}
}
else if (posBalance >= 40) {
if (Cs < 1511) {
return 2;
}
else if (Cs >= 1511) {
return 1;
}
}
}
else if (Rs >= 3514) {
if (posBalance < 39) {
if (Rs < 4374) {
return 4;
}
else if (Rs >= 4374) {
return 8;
}
}
else if (posBalance >= 39) {
return 4;
}
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (ssDCval >= 78) {
if (MVDiff < 7) {
if (Rs < 3054) {
if (RsCsDiff < 51) {
if (Rs < 2056) {
return 8;
}
else if (Rs >= 2056) {
if (Rs < 2400) {
if (Rs < 2351) {
if (MVDiff < 1) {
if (Rs < 2337) {
return 2;
}
else if (Rs >= 2337) {
if (Rs < 2350) {
return 4;
}
else if (Rs >= 2350) {
return 1;
}
}
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (Rs >= 2351) {
return 4;
}
}
else if (Rs >= 2400) {
if (AFD < 2) {
if (ssDCval < 87) {
return 1;
}
else if (ssDCval >= 87) {
return 8;
}
}
else if (AFD >= 2) {
if (Cs < 2559) {
return 1;
}
else if (Cs >= 2559) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 51) {
return 1;
}
}
else if (Rs >= 3054) {
if (RsCsDiff < 37) {
if (SC < 6998) {
if (SC < 6070) {
if (Cs < 1582) {
if (ssDCval < 84) {
return 2;
}
else if (ssDCval >= 84) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (Cs < 1477) {
return 4;
}
else if (Cs >= 1477) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
return 8;
}
}
}
}
}
else if (Cs >= 1582) {
if (SC < 5233) {
return 4;
}
else if (SC >= 5233) {
if (SC < 5731) {
return 8;
}
else if (SC >= 5731) {
if (RsCsDiff < 29) {
if (Rs < 4268) {
return 8;
}
else if (Rs >= 4268) {
return 4;
}
}
else if (RsCsDiff >= 29) {
return 4;
}
}
}
}
}
else if (SC >= 6070) {
if (Rs < 4632) {
if (Rs < 4381) {
return 8;
}
else if (Rs >= 4381) {
return 2;
}
}
else if (Rs >= 4632) {
return 8;
}
}
}
else if (SC >= 6998) {
return 4;
}
}
else if (RsCsDiff >= 37) {
if (Rs < 4153) {
if (mcTcor < 96) {
if (Rs < 4067) {
return 8;
}
else if (Rs >= 4067) {
return 4;
}
}
else if (mcTcor >= 96) {
if (ssDCval < 80) {
return 4;
}
else if (ssDCval >= 80) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
return 4;
}
}
}
}
else if (Rs >= 4153) {
if (TSC < 7) {
if (tcor < 97) {
if (AFD < 9) {
if (CsDiff < 4) {
if (Cs < 1856) {
return 8;
}
else if (Cs >= 1856) {
return 2;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (tcor >= 97) {
if (Rs < 4675) {
if (Rs < 4360) {
return 8;
}
else if (Rs >= 4360) {
return 2;
}
}
else if (Rs >= 4675) {
return 8;
}
}
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 2658) {
if (Cs < 1767) {
return 8;
}
else if (Cs >= 1767) {
if (negBalance < 27) {
if (Rs < 2307) {
return 2;
}
else if (Rs >= 2307) {
return 4;
}
}
else if (negBalance >= 27) {
if (Rs < 2363) {
return 1;
}
else if (Rs >= 2363) {
if (Rs < 2403) {
return 8;
}
else if (Rs >= 2403) {
return 1;
}
}
}
}
}
else if (Rs >= 2658) {
if (Rs < 4415) {
if (Rs < 4246) {
return 8;
}
else if (Rs >= 4246) {
if (Rs < 4276) {
return 4;
}
else if (Rs >= 4276) {
return 8;
}
}
}
else if (Rs >= 4415) {
return 4;
}
}
}
}
}
else if (posBalance >= 44) {
if (RsDiff < 7) {
if (Cs < 1590) {
if (ssDCval < 99) {
if (Rs < 2768) {
if (Rs < 1126) {
if (mcTcor < 96) {
if (TSC < 5) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 1;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 86) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (mu_mv_mag_sq >= 86) {
return 8;
}
}
}
else if (Rs >= 1126) {
if (ssDCval < 90) {
if (tcor < 69) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (tcor >= 69) {
if (mcTcor < 97) {
if (TSC < 8) {
if (Rs < 1392) {
if (MVDiff < 5) {
if (negBalance < 37) {
if (diffAFD < -1) {
if (Rs < 1218) {
return 8;
}
else if (Rs >= 1218) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (negBalance >= 37) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (Rs >= 1392) {
return 8;
}
}
else if (TSC >= 8) {
return 4;
}
}
else if (mcTcor >= 97) {
if (Rs < 1571) {
if (Cs < 1365) {
return 8;
}
else if (Cs >= 1365) {
return 4;
}
}
else if (Rs >= 1571) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
return 4;
}
}
}
}
}
else if (ssDCval >= 90) {
if (diffTSC < 1) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
if (Rs < 1938) {
return 2;
}
else if (Rs >= 1938) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (diffTSC >= 1) {
if (Rs < 1976) {
return 4;
}
else if (Rs >= 1976) {
return 1;
}
}
}
}
}
else if (Rs >= 2768) {
if (mu_mv_mag_sq < 36) {
return 4;
}
else if (mu_mv_mag_sq >= 36) {
if (gchDC < 2) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
if (negBalance < 26) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 2;
}
}
else if (negBalance >= 26) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (Rs < 3404) {
if (Rs < 3220) {
return 1;
}
else if (Rs >= 3220) {
return 4;
}
}
else if (Rs >= 3404) {
return 8;
}
}
}
}
}
else if (ssDCval >= 99) {
if (Rs < 961) {
return 4;
}
else if (Rs >= 961) {
return 8;
}
}
}
else if (Cs >= 1590) {
if (SC < 4787) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 80) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
if (Rs < 1184) {
return 2;
}
else if (Rs >= 1184) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (SC < 4335) {
if (MVDiff < 6) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
if (Rs < 1331) {
return 8;
}
else if (Rs >= 1331) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (Cs < 1693) {
if (mcTcor < 95) {
if (Rs < 1651) {
return 4;
}
else if (Rs >= 1651) {
return 8;
}
}
else if (mcTcor >= 95) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 4;
}
}
}
else if (Cs >= 1693) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Cs < 1832) {
return 8;
}
else if (Cs >= 1832) {
return 1;
}
}
}
}
}
else if (SC >= 4335) {
if (tcor < 95) {
if (Rs < 2236) {
return 4;
}
else if (Rs >= 2236) {
return 8;
}
}
else if (tcor >= 95) {
if (RsCsDiff < 46) {
if (gchDC < 1) {
if (mu_mv_mag_sq < 96) {
if (Rs < 2375) {
return 4;
}
else if (Rs >= 2375) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (RsCsDiff >= 46) {
return 1;
}
}
}
}
}
else if (ssDCval >= 84) {
if (mu_mv_mag_sq < 3) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 2159) {
return 2;
}
else if (Rs >= 2159) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 3059) {
if (Cs < 1627) {
if (Cs < 1624) {
return 8;
}
else if (Cs >= 1624) {
return 2;
}
}
else if (Cs >= 1627) {
return 8;
}
}
else if (Rs >= 3059) {
if (Rs < 3102) {
return 4;
}
else if (Rs >= 3102) {
return 8;
}
}
}
}
}
else if (SC >= 4787) {
if (Cs < 1822) {
if (Rs < 3478) {
return 4;
}
else if (Rs >= 3478) {
if (MVDiff < 5) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (MVDiff >= 5) {
if (Rs < 4185) {
return 8;
}
else if (Rs >= 4185) {
return 1;
}
}
}
}
else if (Cs >= 1822) {
if (RsCsDiff < 43) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 85) {
if (Rs < 5650) {
return 8;
}
else if (Rs >= 5650) {
return 4;
}
}
else if (mu_mv_mag_sq >= 85) {
if (MVDiff < 4) {
if (Cs < 3236) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
if (posBalance < 50) {
if (Cs < 2126) {
return 4;
}
else if (Cs >= 2126) {
return 8;
}
}
else if (posBalance >= 50) {
return 4;
}
}
}
else if (Cs >= 3236) {
return 8;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 43) {
if (Cs < 2498) {
if (AFD < 8) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (AFD >= 8) {
if (Rs < 3258) {
return 1;
}
else if (Rs >= 3258) {
return 8;
}
}
}
else if (Cs >= 2498) {
if (mu_mv_mag_sq < 92) {
return 1;
}
else if (mu_mv_mag_sq >= 92) {
return 4;
}
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (mcTcor < 95) {
if (diffRsCsdiff < -4) {
if (MVDiff < 8) {
if (Cs < 1634) {
return 2;
}
else if (Cs >= 1634) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (Rs < 3064) {
return 8;
}
else if (Rs >= 3064) {
return 2;
}
}
}
}
else if (MVDiff >= 8) {
if (Rs < 4329) {
return 8;
}
else if (Rs >= 4329) {
return 4;
}
}
}
else if (diffRsCsdiff >= -4) {
if (posBalance < 47) {
if (Rs < 3202) {
return 4;
}
else if (Rs >= 3202) {
return 1;
}
}
else if (posBalance >= 47) {
if (mu_mv_mag_sq < 25) {
return 1;
}
else if (mu_mv_mag_sq >= 25) {
if (Rs < 4239) {
if (MVDiff < 2) {
if (Rs < 2978) {
return 4;
}
else if (Rs >= 2978) {
return 8;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (Rs >= 4239) {
return 4;
}
}
}
}
}
else if (mcTcor >= 95) {
if (refDCval < 71) {
if (Cs < 1658) {
return 2;
}
else if (Cs >= 1658) {
return 4;
}
}
else if (refDCval >= 71) {
if (Rs < 3053) {
if (Cs < 1534) {
if (Rs < 1994) {
return 8;
}
else if (Rs >= 1994) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (Cs >= 1534) {
if (ssDCval < 89) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
}
else if (Rs >= 3053) {
if (ssDCval < 92) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (Rs < 3847) {
return 4;
}
else if (Rs >= 3847) {
return 2;
}
}
}
else if (ssDCval >= 92) {
if (tcor < 96) {
if (posBalance < 51) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (posBalance >= 51) {
return 2;
}
}
else if (tcor >= 96) {
return 2;
}
}
}
}
}
}
}
}
else if (SC >= 10165) {
if (Rs < 5646) {
return 8;
}
else if (Rs >= 5646) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 1453) {
if (RsCsDiff < 41) {
if (ssDCval < 82) {
if (MVDiff < 4) {
if (ssDCval < 72) {
if (diffMVdiffVal < -3) {
if (Rs < 1159) {
return 4;
}
else if (Rs >= 1159) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
if (negBalance < 41) {
if (refDCval < 55) {
return 8;
}
else if (refDCval >= 55) {
if (Rs < 996) {
return 8;
}
else if (Rs >= 996) {
return 4;
}
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 910) {
return 1;
}
else if (Rs >= 910) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (negBalance < 45) {
if (Cs < 1711) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (Rs < 1071) {
if (Rs < 973) {
return 8;
}
else if (Rs >= 973) {
return 1;
}
}
else if (Rs >= 1071) {
return 4;
}
}
}
else if (Cs >= 1711) {
if (Rs < 739) {
return 2;
}
else if (Rs >= 739) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (Rs < 471) {
return 1;
}
else if (Rs >= 471) {
if (posBalance < 63) {
if (diffRsCsdiff < 3) {
if (negBalance < 47) {
if (AFD < 10) {
if (Rs < 1090) {
return 8;
}
else if (Rs >= 1090) {
return 1;
}
}
else if (AFD >= 10) {
return 2;
}
}
else if (negBalance >= 47) {
if (Rs < 918) {
if (Rs < 774) {
return 8;
}
else if (Rs >= 774) {
if (Rs < 859) {
return 4;
}
else if (Rs >= 859) {
return 2;
}
}
}
else if (Rs >= 918) {
return 8;
}
}
}
else if (diffRsCsdiff >= 3) {
if (Cs < 1382) {
return 1;
}
else if (Cs >= 1382) {
return 2;
}
}
}
else if (posBalance >= 63) {
return 4;
}
}
}
}
}
else if (ssDCval >= 82) {
if (posBalance < 50) {
return 4;
}
else if (posBalance >= 50) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (MVDiff < 9) {
if (Cs < 1359) {
return 2;
}
else if (Cs >= 1359) {
return 8;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 41) {
if (TSC < 11) {
if (ssDCval < 55) {
if (MVDiff < 19) {
if (Cs < 1755) {
return 4;
}
else if (Cs >= 1755) {
return 8;
}
}
else if (MVDiff >= 19) {
if (Rs < 542) {
return 4;
}
else if (Rs >= 542) {
if (Cs < 1336) {
return 4;
}
else if (Cs >= 1336) {
return 2;
}
}
}
}
else if (ssDCval >= 55) {
if (TSC < 7) {
if (ssDCval < 69) {
if (mcTcor < 98) {
if (Cs < 1665) {
if (Rs < 1348) {
if (mu_mv_mag_sq < 2) {
return 2;
}
else if (mu_mv_mag_sq >= 2) {
return 4;
}
}
else if (Rs >= 1348) {
return 8;
}
}
else if (Cs >= 1665) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (ssDCval >= 69) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
if (Cs < 1616) {
return 2;
}
else if (Cs >= 1616) {
return 8;
}
}
}
}
else if (TSC >= 7) {
if (Cs < 1572) {
if (AFD < 17) {
if (posBalance < 61) {
if (Rs < 648) {
return 2;
}
else if (Rs >= 648) {
if (MVDiff < 21) {
if (posBalance < 55) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (MVDiff >= 21) {
if (Cs < 1362) {
if (Rs < 1358) {
return 4;
}
else if (Rs >= 1358) {
return 1;
}
}
else if (Cs >= 1362) {
return 8;
}
}
}
}
else if (posBalance >= 61) {
if (MVDiff < 8) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
return 4;
}
}
else if (MVDiff >= 8) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
if (Cs < 1534) {
return 8;
}
else if (Cs >= 1534) {
return 4;
}
}
}
}
}
else if (AFD >= 17) {
if (RsCsDiff < 99) {
if (TSC < 8) {
if (Rs < 789) {
return 4;
}
else if (Rs >= 789) {
return 8;
}
}
else if (TSC >= 8) {
return 4;
}
}
else if (RsCsDiff >= 99) {
return 2;
}
}
}
else if (Cs >= 1572) {
if (negBalance < 55) {
if (tcor < 88) {
if (Rs < 1444) {
if (ssDCval < 73) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1003) {
return 4;
}
else if (Rs >= 1003) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (ssDCval >= 73) {
if (SC < 2417) {
if (Rs < 626) {
return 4;
}
else if (Rs >= 626) {
return 8;
}
}
else if (SC >= 2417) {
return 4;
}
}
}
else if (Rs >= 1444) {
return 8;
}
}
else if (tcor >= 88) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (Rs < 697) {
return 4;
}
else if (Rs >= 697) {
return 2;
}
}
}
}
else if (negBalance >= 55) {
if (tcor < 81) {
return 2;
}
else if (tcor >= 81) {
return 8;
}
}
}
}
}
}
else if (TSC >= 11) {
if (MVDiff < 36) {
if (SC < 3182) {
if (ssDCval < 81) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (Rs < 1370) {
return 4;
}
else if (Rs >= 1370) {
return 8;
}
}
}
else if (ssDCval >= 81) {
if (Rs < 1430) {
return 8;
}
else if (Rs >= 1430) {
return 4;
}
}
}
else if (SC >= 3182) {
return 4;
}
}
else if (MVDiff >= 36) {
if (Rs < 1313) {
return 2;
}
else if (Rs >= 1313) {
return 4;
}
}
}
}
}
else if (Rs >= 1453) {
if (CsDiff < 7) {
if (mcTcor < 94) {
if (posBalance < 70) {
if (ssDCval < 91) {
if (ssDCval < 77) {
if (Cs < 1424) {
if (MVDiff < 17) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (Rs < 1954) {
return 2;
}
else if (Rs >= 1954) {
return 8;
}
}
}
else if (MVDiff >= 17) {
if (Cs < 1419) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (Cs >= 1419) {
return 1;
}
}
}
else if (Cs >= 1424) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
if (RsDiff < 8) {
if (Rs < 2021) {
if (negBalance < 51) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (SC < 3099) {
return 4;
}
else if (SC >= 3099) {
if (SC < 3535) {
return 8;
}
else if (SC >= 3535) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
return 8;
}
}
}
}
}
else if (negBalance >= 51) {
return 8;
}
}
else if (Rs >= 2021) {
return 8;
}
}
else if (RsDiff >= 8) {
if (posBalance < 55) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
}
else if (ssDCval >= 77) {
if (AFD < 16) {
if (Cs < 2431) {
if (ssDCval < 84) {
if (RsDiff < 8) {
if (diffMVdiffVal < -2) {
if (Rs < 1763) {
return 4;
}
else if (Rs >= 1763) {
if (diffRsCsdiff < -12) {
if (negBalance < 50) {
if (Rs < 1880) {
return 1;
}
else if (Rs >= 1880) {
return 8;
}
}
else if (negBalance >= 50) {
if (Rs < 2070) {
return 8;
}
else if (Rs >= 2070) {
return 4;
}
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (RsDiff >= 8) {
if (TSC < 10) {
if (Cs < 2041) {
if (Rs < 2422) {
if (Rs < 2018) {
return 4;
}
else if (Rs >= 2018) {
return 8;
}
}
else if (Rs >= 2422) {
return 4;
}
}
else if (Cs >= 2041) {
return 8;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (ssDCval >= 84) {
if (negBalance < 45) {
if (Rs < 2220) {
if (Rs < 1821) {
return 8;
}
else if (Rs >= 1821) {
return 2;
}
}
else if (Rs >= 2220) {
return 8;
}
}
else if (negBalance >= 45) {
if (Cs < 1335) {
return 2;
}
else if (Cs >= 1335) {
return 8;
}
}
}
}
else if (Cs >= 2431) {
if (SC < 5604) {
if (Rs < 2711) {
return 4;
}
else if (Rs >= 2711) {
if (Rs < 2967) {
return 8;
}
else if (Rs >= 2967) {
return 4;
}
}
}
else if (SC >= 5604) {
return 8;
}
}
}
else if (AFD >= 16) {
if (Rs < 1593) {
if (refDCval < 90) {
return 4;
}
else if (refDCval >= 90) {
return 8;
}
}
else if (Rs >= 1593) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
return 8;
}
}
}
}
}
else if (ssDCval >= 91) {
if (Rs < 2153) {
if (mcTcor < 89) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
return 1;
}
}
else if (mcTcor >= 89) {
if (RsCsDiff < 57) {
if (posBalance < 51) {
if (Rs < 1945) {
return 8;
}
else if (Rs >= 1945) {
return 2;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (RsCsDiff >= 57) {
if (MVDiff < 11) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
return 8;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
}
}
else if (Rs >= 2153) {
if (Rs < 4639) {
if (diffAFD < -4) {
if (Rs < 2603) {
return 8;
}
else if (Rs >= 2603) {
return 4;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (Rs >= 4639) {
return 1;
}
}
}
}
else if (posBalance >= 70) {
if (Rs < 2463) {
return 4;
}
else if (Rs >= 2463) {
return 1;
}
}
}
else if (mcTcor >= 94) {
if (SC < 7014) {
if (SC < 5027) {
if (RsDiff < 8) {
if (Rs < 1938) {
if (refDCval < 87) {
if (AFD < 8) {
if (Rs < 1465) {
return 1;
}
else if (Rs >= 1465) {
return 8;
}
}
else if (AFD >= 8) {
if (MVDiff < 3) {
if (mcTcor < 95) {
if (Rs < 1671) {
return 8;
}
else if (Rs >= 1671) {
if (Rs < 1933) {
return 1;
}
else if (Rs >= 1933) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 81) {
if (Cs < 1359) {
return 1;
}
else if (Cs >= 1359) {
return 4;
}
}
else if (ssDCval >= 81) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 1515) {
if (Rs < 1501) {
return 4;
}
else if (Rs >= 1501) {
return 2;
}
}
else if (Rs >= 1515) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 87) {
if (mcTcor < 96) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
if (Rs < 1936) {
if (posBalance < 49) {
if (Rs < 1756) {
return 8;
}
else if (Rs >= 1756) {
return 1;
}
}
else if (posBalance >= 49) {
return 8;
}
}
else if (Rs >= 1936) {
return 2;
}
}
}
else if (mcTcor >= 96) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Cs < 1525) {
return 8;
}
else if (Cs >= 1525) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1938) {
if (Rs < 2587) {
if (refDCval < 88) {
if (SC < 4042) {
if (MVDiff < 7) {
if (Cs < 1598) {
if (ssDCval < 82) {
if (negBalance < 48) {
if (diffRsCsdiff < -15) {
return 1;
}
else if (diffRsCsdiff >= -15) {
if (diffRsCsdiff < 5) {
return 8;
}
else if (diffRsCsdiff >= 5) {
if (Rs < 2048) {
return 8;
}
else if (Rs >= 2048) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
}
else if (negBalance >= 48) {
if (Rs < 2248) {
if (Rs < 2159) {
return 4;
}
else if (Rs >= 2159) {
return 8;
}
}
else if (Rs >= 2248) {
return 1;
}
}
}
else if (ssDCval >= 82) {
if (AFD < 8) {
if (Rs < 2206) {
if (Rs < 2151) {
return 2;
}
else if (Rs >= 2151) {
return 8;
}
}
else if (Rs >= 2206) {
return 4;
}
}
else if (AFD >= 8) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
if (diffRsCsdiff < -27) {
return 8;
}
else if (diffRsCsdiff >= -27) {
return 1;
}
}
}
}
}
else if (Cs >= 1598) {
if (AFD < 9) {
if (diffAFD < -1) {
if (Rs < 2092) {
return 4;
}
else if (Rs >= 2092) {
if (Rs < 2214) {
return 1;
}
else if (Rs >= 2214) {
return 8;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (AFD >= 9) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 8;
}
}
}
else if (SC >= 4042) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (Rs < 2134) {
return 1;
}
else if (Rs >= 2134) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 88) {
if (Cs < 1917) {
if (Rs < 2168) {
if (diffAFD < 1) {
if (mu_mv_mag_sq < 95) {
return 8;
}
else if (mu_mv_mag_sq >= 95) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
}
}
else if (Rs >= 2168) {
if (diffAFD < -12) {
return 4;
}
else if (diffAFD >= -12) {
return 8;
}
}
}
else if (Cs >= 1917) {
if (diffRsCsdiff < 5) {
if (diffAFD < -2) {
if (Rs < 2248) {
return 2;
}
else if (Rs >= 2248) {
return 8;
}
}
else if (diffAFD >= -2) {
if (Rs < 2183) {
if (Rs < 2082) {
return 8;
}
else if (Rs >= 2082) {
return 1;
}
}
else if (Rs >= 2183) {
if (diffMVdiffVal < -6) {
return 1;
}
else if (diffMVdiffVal >= -6) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 5) {
if (Rs < 2040) {
return 1;
}
else if (Rs >= 2040) {
return 8;
}
}
}
}
}
else if (Rs >= 2587) {
if (posBalance < 49) {
if (RsCsDiff < 38) {
return 8;
}
else if (RsCsDiff >= 38) {
if (mu_mv_mag_sq < 74) {
if (Rs < 3514) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (Rs >= 3514) {
return 1;
}
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 1453) {
return 2;
}
else if (Cs >= 1453) {
return 8;
}
}
}
}
else if (posBalance >= 49) {
if (negBalance < 53) {
if (Cs < 1330) {
return 2;
}
else if (Cs >= 1330) {
if (Rs < 2594) {
if (Rs < 2593) {
return 8;
}
else if (Rs >= 2593) {
return 4;
}
}
else if (Rs >= 2594) {
return 8;
}
}
}
else if (negBalance >= 53) {
if (Rs < 3025) {
return 4;
}
else if (Rs >= 3025) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (TSC < 8) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (diffRsCsdiff < 26) {
if (ssDCval < 85) {
return 4;
}
else if (ssDCval >= 85) {
return 2;
}
}
else if (diffRsCsdiff >= 26) {
return 8;
}
}
}
else if (TSC >= 8) {
if (AFD < 15) {
if (Rs < 2198) {
return 2;
}
else if (Rs >= 2198) {
return 8;
}
}
else if (AFD >= 15) {
return 4;
}
}
}
}
else if (SC >= 5027) {
if (Cs < 1620) {
if (AFD < 10) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (AFD >= 10) {
if (Rs < 4038) {
return 4;
}
else if (Rs >= 4038) {
return 1;
}
}
}
else if (Cs >= 1620) {
if (Rs < 2694) {
return 4;
}
else if (Rs >= 2694) {
if (posBalance < 46) {
if (ssDCval < 85) {
if (Rs < 3314) {
return 8;
}
else if (Rs >= 3314) {
return 4;
}
}
else if (ssDCval >= 85) {
return 2;
}
}
else if (posBalance >= 46) {
if (TSC < 8) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 101) {
if (Cs < 3228) {
return 8;
}
else if (Cs >= 3228) {
if (Rs < 3366) {
return 4;
}
else if (Rs >= 3366) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Cs < 2090) {
if (Rs < 3781) {
return 8;
}
else if (Rs >= 3781) {
return 2;
}
}
else if (Cs >= 2090) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
}
}
}
else if (diffAFD >= 2) {
if (Cs < 3106) {
return 4;
}
else if (Cs >= 3106) {
return 8;
}
}
}
else if (TSC >= 8) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (TSC < 10) {
if (Rs < 3861) {
return 4;
}
else if (Rs >= 3861) {
return 8;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 7014) {
if (Cs < 2529) {
return 4;
}
else if (Cs >= 2529) {
return 8;
}
}
}
}
else if (CsDiff >= 7) {
if (ssDCval < 101) {
if (Cs < 2304) {
if (tcor < 88) {
if (MVDiff < 42) {
if (refDCval < 50) {
if (Cs < 1492) {
return 8;
}
else if (Cs >= 1492) {
if (TSC < 9) {
if (Rs < 1805) {
return 1;
}
else if (Rs >= 1805) {
return 8;
}
}
else if (TSC >= 9) {
return 4;
}
}
}
else if (refDCval >= 50) {
if (mcTcor < 89) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
if (Rs < 1781) {
return 4;
}
else if (Rs >= 1781) {
return 8;
}
}
}
else if (mcTcor >= 89) {
if (tcor < 81) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (tcor >= 81) {
if (diffMVdiffVal < -3) {
if (Rs < 2460) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
else if (Rs >= 2460) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
if (diffRsCsdiff < 31) {
if (ssDCval < 96) {
if (diffTSC < -2) {
if (Rs < 1953) {
return 8;
}
else if (Rs >= 1953) {
return 4;
}
}
else if (diffTSC >= -2) {
return 8;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (diffRsCsdiff >= 31) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 42) {
return 1;
}
}
else if (tcor >= 88) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
}
else if (Cs >= 2304) {
if (MVDiff < 6) {
if (mcTcor < 94) {
if (Cs < 2964) {
if (Rs < 2402) {
return 4;
}
else if (Rs >= 2402) {
if (mu_mv_mag_sq < 106) {
return 8;
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 3292) {
return 8;
}
else if (Rs >= 3292) {
return 4;
}
}
}
}
else if (Cs >= 2964) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (mu_mv_mag_sq < 82) {
if (Rs < 1798) {
return 2;
}
else if (Rs >= 1798) {
return 1;
}
}
else if (mu_mv_mag_sq >= 82) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Rs < 2799) {
return 8;
}
else if (Rs >= 2799) {
return 4;
}
}
}
}
}
else if (MVDiff >= 6) {
if (Rs < 1754) {
return 1;
}
else if (Rs >= 1754) {
return 8;
}
}
}
}
else if (ssDCval >= 101) {
if (mu_mv_mag_sq < 87) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (mu_mv_mag_sq >= 87) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 105) {
if (ssDCval < 163) {
if (Rs < 4726) {
if (Rs < 1436) {
if (mu_mv_mag_sq < 46) {
if (ssDCval < 146) {
if (SC < 2771) {
if (AFD < 18) {
if (ssDCval < 106) {
return 4;
}
else if (ssDCval >= 106) {
if (ssDCval < 143) {
return 8;
}
else if (ssDCval >= 143) {
return 1;
}
}
}
else if (AFD >= 18) {
if (Cs < 1505) {
if (Rs < 867) {
return 8;
}
else if (Rs >= 867) {
return 1;
}
}
else if (Cs >= 1505) {
return 2;
}
}
}
else if (SC >= 2771) {
if (CsDiff < 4) {
if (Rs < 1408) {
return 1;
}
else if (Rs >= 1408) {
return 8;
}
}
else if (CsDiff >= 4) {
if (mcTcor < 88) {
return 4;
}
else if (mcTcor >= 88) {
if (RsDiff < 7) {
if (diffMVdiffVal < 5) {
if (diffTSC < -1) {
if (Rs < 1328) {
return 4;
}
else if (Rs >= 1328) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (diffMVdiffVal >= 5) {
return 4;
}
}
else if (RsDiff >= 7) {
return 4;
}
}
}
}
}
else if (ssDCval >= 146) {
if (tcor < 89) {
return 2;
}
else if (tcor >= 89) {
if (Rs < 1410) {
return 4;
}
else if (Rs >= 1410) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (Rs < 1239) {
if (negBalance < 45) {
if (posBalance < 50) {
if (Cs < 1828) {
return 4;
}
else if (Cs >= 1828) {
if (Rs < 960) {
return 1;
}
else if (Rs >= 960) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (Rs < 972) {
if (Rs < 645) {
return 1;
}
else if (Rs >= 645) {
return 8;
}
}
else if (Rs >= 972) {
if (RsDiff < 3) {
if (Rs < 1002) {
return 8;
}
else if (Rs >= 1002) {
if (negBalance < 38) {
if (SC < 2693) {
return 8;
}
else if (SC >= 2693) {
if (Rs < 1223) {
return 4;
}
else if (Rs >= 1223) {
return 8;
}
}
}
else if (negBalance >= 38) {
return 4;
}
}
}
else if (RsDiff >= 3) {
if (diffAFD < -1) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Cs < 1512) {
return 8;
}
else if (Cs >= 1512) {
return 4;
}
}
}
else if (diffAFD >= -1) {
if (Cs < 1383) {
return 1;
}
else if (Cs >= 1383) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (Rs >= 1239) {
if (RsDiff < 3) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (Rs < 1337) {
return 4;
}
else if (Rs >= 1337) {
if (Rs < 1394) {
return 1;
}
else if (Rs >= 1394) {
return 8;
}
}
}
}
else if (RsDiff >= 3) {
return 8;
}
}
}
}
else if (Rs >= 1436) {
if (ssDCval < 109) {
if (mu_mv_mag_sq < 43) {
if (Cs < 1795) {
if (ssDCval < 107) {
if (tcor < 92) {
if (RsCsDiff < 63) {
return 4;
}
else if (RsCsDiff >= 63) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
return 8;
}
}
}
else if (tcor >= 92) {
return 8;
}
}
else if (ssDCval >= 107) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
}
}
else if (Cs >= 1795) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Rs < 1923) {
return 2;
}
else if (Rs >= 1923) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (diffTSC < 5) {
if (diffRsCsdiff < -23) {
if (Cs < 1474) {
return 4;
}
else if (Cs >= 1474) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -23) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (Cs < 1527) {
if (diffTSC < 1) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (Rs < 3285) {
return 4;
}
else if (Rs >= 3285) {
return 8;
}
}
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (Cs >= 1527) {
return 8;
}
}
}
}
else if (diffTSC >= 5) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
}
else if (ssDCval >= 109) {
if (ssDCval < 142) {
if (tcor < 94) {
if (SC < 3760) {
if (ssDCval < 123) {
if (AFD < 13) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
else if (AFD >= 13) {
if (negBalance < 48) {
if (Cs < 1972) {
return 4;
}
else if (Cs >= 1972) {
return 1;
}
}
else if (negBalance >= 48) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
}
else if (ssDCval >= 123) {
if (CsDiff < 6) {
if (RsDiff < 6) {
if (ssDCval < 132) {
if (refDCval < 125) {
return 8;
}
else if (refDCval >= 125) {
return 1;
}
}
else if (ssDCval >= 132) {
if (mcTcor < 84) {
if (Rs < 1546) {
return 1;
}
else if (Rs >= 1546) {
return 8;
}
}
else if (mcTcor >= 84) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (tcor < 87) {
return 1;
}
else if (tcor >= 87) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 1;
}
}
}
}
else if (CsDiff >= 6) {
if (Rs < 1523) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Rs < 1510) {
return 2;
}
else if (Rs >= 1510) {
return 4;
}
}
}
else if (Rs >= 1523) {
return 8;
}
}
}
}
else if (SC >= 3760) {
if (negBalance < 48) {
if (mcTcor < 94) {
if (Cs < 2888) {
if (ssDCval < 139) {
return 8;
}
else if (ssDCval >= 139) {
if (SC < 5010) {
if (posBalance < 41) {
if (Rs < 2567) {
return 8;
}
else if (Rs >= 2567) {
return 4;
}
}
else if (posBalance >= 41) {
return 8;
}
}
else if (SC >= 5010) {
if (Rs < 3156) {
return 4;
}
else if (Rs >= 3156) {
return 1;
}
}
}
}
else if (Cs >= 2888) {
if (Cs < 3161) {
if (Rs < 4651) {
if (Cs < 2900) {
return 2;
}
else if (Cs >= 2900) {
return 8;
}
}
else if (Rs >= 4651) {
return 1;
}
}
else if (Cs >= 3161) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (refDCval < 128) {
if (mu_mv_mag_sq < 64) {
if (posBalance < 51) {
if (ssDCval < 112) {
return 4;
}
else if (ssDCval >= 112) {
return 1;
}
}
else if (posBalance >= 51) {
if (TSC < 9) {
if (Rs < 2167) {
return 4;
}
else if (Rs >= 2167) {
return 8;
}
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
return 1;
}
}
}
}
else if (refDCval >= 128) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
if (mu_mv_mag_sq < 81) {
if (tcor < 93) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (RsCsDiff < 89) {
return 1;
}
else if (RsCsDiff >= 89) {
return 8;
}
}
}
else if (tcor >= 93) {
if (Rs < 3043) {
if (Rs < 2779) {
return 2;
}
else if (Rs >= 2779) {
return 1;
}
}
else if (Rs >= 3043) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (ssDCval < 137) {
if (Rs < 3156) {
if (refDCval < 123) {
if (Cs < 1758) {
return 1;
}
else if (Cs >= 1758) {
return 8;
}
}
else if (refDCval >= 123) {
return 1;
}
}
else if (Rs >= 3156) {
return 8;
}
}
else if (ssDCval >= 137) {
return 1;
}
}
}
}
}
else if (tcor >= 94) {
if (Cs < 1360) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
if (mcTcor < 98) {
if (Rs < 2438) {
if (Rs < 1975) {
return 2;
}
else if (Rs >= 1975) {
return 1;
}
}
else if (Rs >= 2438) {
return 8;
}
}
else if (mcTcor >= 98) {
return 2;
}
}
}
else if (Cs >= 1360) {
if (posBalance < 34) {
if (negBalance < 43) {
if (Rs < 2746) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Rs < 2359) {
return 2;
}
else if (Rs >= 2359) {
return 4;
}
}
}
else if (Rs >= 2746) {
return 2;
}
}
else if (negBalance >= 43) {
return 8;
}
}
else if (posBalance >= 34) {
if (mu_mv_mag_sq < 19) {
return 2;
}
else if (mu_mv_mag_sq >= 19) {
if (mcTcor < 94) {
if (Rs < 2935) {
return 8;
}
else if (Rs >= 2935) {
return 1;
}
}
else if (mcTcor >= 94) {
if (diffTSC < 2) {
if (refDCval < 109) {
if (Cs < 1558) {
return 4;
}
else if (Cs >= 1558) {
return 8;
}
}
else if (refDCval >= 109) {
if (diffAFD < -2) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (diffTSC >= 2) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 142) {
if (mu_mv_mag_sq < 46) {
if (Rs < 1692) {
if (gchDC < 1) {
if (diffMVdiffVal < -4) {
if (Rs < 1492) {
return 2;
}
else if (Rs >= 1492) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (gchDC >= 1) {
return 2;
}
}
else if (Rs >= 1692) {
return 8;
}
}
else if (mu_mv_mag_sq >= 46) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
if (Rs < 1977) {
return 4;
}
else if (Rs >= 1977) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 4726) {
if (Cs < 2015) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (Cs >= 2015) {
if (Cs < 3969) {
if (Rs < 5583) {
return 8;
}
else if (Rs >= 5583) {
return 4;
}
}
else if (Cs >= 3969) {
return 8;
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2283) {
if (Rs < 1064) {
return 1;
}
else if (Rs >= 1064) {
return 2;
}
}
else if (Rs >= 2283) {
return 8;
}
}
}
}
else if (RsCsDiff >= 101) {
if (tcor < 89) {
if (RsDiff < 15) {
if (refDCval < 101) {
if (SC < 6508) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 58) {
if (refDCval < 59) {
if (Rs < 1542) {
return 4;
}
else if (Rs >= 1542) {
if (CsDiff < 9) {
if (RsCsDiff < 148) {
if (Rs < 1830) {
return 1;
}
else if (Rs >= 1830) {
return 4;
}
}
else if (RsCsDiff >= 148) {
if (Rs < 2513) {
return 8;
}
else if (Rs >= 2513) {
return 2;
}
}
}
else if (CsDiff >= 9) {
if (negBalance < 60) {
if (Cs < 1421) {
return 8;
}
else if (Cs >= 1421) {
return 1;
}
}
else if (negBalance >= 60) {
if (Rs < 3088) {
return 8;
}
else if (Rs >= 3088) {
return 4;
}
}
}
}
}
else if (refDCval >= 59) {
if (Rs < 2607) {
if (AFD < 19) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 82) {
return 1;
}
else if (ssDCval >= 82) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (posBalance < 44) {
if (Rs < 1869) {
return 8;
}
else if (Rs >= 1869) {
return 2;
}
}
else if (posBalance >= 44) {
return 8;
}
}
}
else if (AFD >= 19) {
if (RsDiff < 12) {
if (diffAFD < 2) {
if (diffRsCsdiff < -48) {
if (Rs < 2175) {
return 8;
}
else if (Rs >= 2175) {
return 2;
}
}
else if (diffRsCsdiff >= -48) {
return 2;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
else if (RsDiff >= 12) {
return 8;
}
}
}
else if (Rs >= 2607) {
if (TSC < 8) {
if (Cs < 1800) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
return 4;
}
}
}
else if (Cs >= 1800) {
if (Rs < 3787) {
return 1;
}
else if (Rs >= 3787) {
return 8;
}
}
}
else if (TSC >= 8) {
if (negBalance < 63) {
if (CsDiff < 7) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
return 4;
}
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (negBalance >= 63) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (ssDCval < 82) {
if (ssDCval < 49) {
if (mu_mv_mag_sq < 82) {
if (Rs < 2987) {
return 1;
}
else if (Rs >= 2987) {
return 8;
}
}
else if (mu_mv_mag_sq >= 82) {
return 8;
}
}
else if (ssDCval >= 49) {
if (MVDiff < 5) {
if (Rs < 3554) {
return 1;
}
else if (Rs >= 3554) {
return 8;
}
}
else if (MVDiff >= 5) {
if (negBalance < 45) {
if (mcTcor < 87) {
if (Cs < 1787) {
return 1;
}
else if (Cs >= 1787) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
if (Cs < 2808) {
return 1;
}
else if (Cs >= 2808) {
return 4;
}
}
}
}
else if (mcTcor >= 87) {
if (Rs < 2988) {
return 2;
}
else if (Rs >= 2988) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (Rs < 3760) {
return 4;
}
else if (Rs >= 3760) {
return 2;
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 3924) {
return 8;
}
else if (Rs >= 3924) {
return 4;
}
}
}
}
}
else if (ssDCval >= 82) {
if (refDCval < 92) {
return 8;
}
else if (refDCval >= 92) {
if (Rs < 2514) {
return 1;
}
else if (Rs >= 2514) {
return 2;
}
}
}
}
}
else if (posBalance >= 49) {
if (TSC < 15) {
if (Rs < 1791) {
if (mcTcor < 70) {
if (refDCval < 46) {
return 8;
}
else if (refDCval >= 46) {
if (SC < 3598) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (SC >= 3598) {
if (negBalance < 56) {
return 4;
}
else if (negBalance >= 56) {
return 2;
}
}
}
}
else if (mcTcor >= 70) {
if (refDCval < 60) {
if (AFD < 18) {
if (negBalance < 46) {
if (CsDiff < 14) {
if (Rs < 809) {
return 2;
}
else if (Rs >= 809) {
if (Rs < 1765) {
return 1;
}
else if (Rs >= 1765) {
return 4;
}
}
}
else if (CsDiff >= 14) {
if (Rs < 1446) {
return 4;
}
else if (Rs >= 1446) {
return 2;
}
}
}
else if (negBalance >= 46) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 26) {
return 4;
}
else if (MVDiff >= 26) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
if (Rs < 1623) {
return 8;
}
else if (Rs >= 1623) {
return 4;
}
}
}
}
}
else if (AFD >= 18) {
if (negBalance < 50) {
if (Rs < 505) {
return 4;
}
else if (Rs >= 505) {
return 2;
}
}
else if (negBalance >= 50) {
if (Cs < 1644) {
if (Cs < 1493) {
return 8;
}
else if (Cs >= 1493) {
return 2;
}
}
else if (Cs >= 1644) {
return 4;
}
}
}
}
else if (refDCval >= 60) {
if (TSC < 8) {
if (mcTcor < 97) {
if (diffMVdiffVal < -5) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (gchDC < 4) {
if (RsCsDiff < 121) {
if (Cs < 1704) {
return 8;
}
else if (Cs >= 1704) {
if (Rs < 1502) {
return 4;
}
else if (Rs >= 1502) {
return 8;
}
}
}
else if (RsCsDiff >= 121) {
return 8;
}
}
else if (gchDC >= 4) {
if (Rs < 1351) {
return 4;
}
else if (Rs >= 1351) {
return 2;
}
}
}
}
}
else if (mcTcor >= 97) {
if (SC < 2816) {
return 1;
}
else if (SC >= 2816) {
return 2;
}
}
}
else if (TSC >= 8) {
if (Rs < 1543) {
if (CsDiff < 19) {
if (posBalance < 62) {
if (Cs < 1347) {
if (Cs < 1333) {
return 2;
}
else if (Cs >= 1333) {
return 8;
}
}
else if (Cs >= 1347) {
if (negBalance < 64) {
if (diffTSC < -2) {
if (Rs < 1487) {
return 1;
}
else if (Rs >= 1487) {
return 4;
}
}
else if (diffTSC >= -2) {
if (ssDCval < 78) {
if (mcTcor < 94) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
if (MVDiff < 32) {
return 4;
}
else if (MVDiff >= 32) {
if (RsDiff < 10) {
return 2;
}
else if (RsDiff >= 10) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (RsCsDiff < 161) {
return 2;
}
else if (RsCsDiff >= 161) {
if (Rs < 987) {
return 8;
}
else if (Rs >= 987) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (RsDiff < 9) {
if (refDCval < 100) {
return 4;
}
else if (refDCval >= 100) {
return 8;
}
}
else if (RsDiff >= 9) {
return 8;
}
}
}
}
else if (negBalance >= 64) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
if (Rs < 984) {
return 8;
}
else if (Rs >= 984) {
return 1;
}
}
}
}
}
else if (posBalance >= 62) {
if (CsDiff < 15) {
if (ssDCval < 102) {
if (SC < 2840) {
if (SC < 2664) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 1129) {
return 4;
}
else if (Rs >= 1129) {
return 2;
}
}
}
else if (SC >= 2664) {
if (ssDCval < 73) {
if (negBalance < 59) {
return 4;
}
else if (negBalance >= 59) {
return 2;
}
}
else if (ssDCval >= 73) {
return 2;
}
}
}
else if (SC >= 2840) {
if (diffMVdiffVal < 5) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (MVDiff < 32) {
return 8;
}
else if (MVDiff >= 32) {
if (Rs < 1113) {
return 8;
}
else if (Rs >= 1113) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 5) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 4;
}
}
}
}
else if (ssDCval >= 102) {
return 1;
}
}
else if (CsDiff >= 15) {
if (mu_mv_mag_sq < 19) {
return 2;
}
else if (mu_mv_mag_sq >= 19) {
if (Rs < 1123) {
return 4;
}
else if (Rs >= 1123) {
return 8;
}
}
}
}
}
else if (CsDiff >= 19) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 8;
}
}
}
else if (Rs >= 1543) {
if (Cs < 1873) {
if (Cs < 1445) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (negBalance < 56) {
if (RsDiff < 11) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (RsDiff >= 11) {
return 8;
}
}
else if (negBalance >= 56) {
return 2;
}
}
}
else if (Cs >= 1445) {
if (RsCsDiff < 193) {
if (mu_mv_mag_sq < 3) {
if (RsCsDiff < 137) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (RsCsDiff >= 137) {
if (Cs < 1748) {
return 1;
}
else if (Cs >= 1748) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 55) {
if (ssDCval < 80) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (ssDCval >= 80) {
return 8;
}
}
else if (negBalance >= 55) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 1;
}
}
}
}
else if (RsCsDiff >= 193) {
if (diffAFD < 1) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 4;
}
}
else if (diffAFD >= 1) {
if (Rs < 1578) {
return 8;
}
else if (Rs >= 1578) {
return 2;
}
}
}
}
}
else if (Cs >= 1873) {
if (ssDCval < 91) {
if (CsDiff < 15) {
return 4;
}
else if (CsDiff >= 15) {
if (Rs < 1666) {
return 8;
}
else if (Rs >= 1666) {
return 1;
}
}
}
else if (ssDCval >= 91) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Cs < 2124) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (Cs >= 2124) {
if (Rs < 1743) {
return 4;
}
else if (Rs >= 1743) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 1791) {
if (Cs < 2826) {
if (Rs < 3041) {
if (AFD < 18) {
if (refDCval < 52) {
if (Cs < 1773) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (Rs < 2204) {
return 4;
}
else if (Rs >= 2204) {
return 1;
}
}
}
else if (Cs >= 1773) {
if (mcTcor < 57) {
return 2;
}
else if (mcTcor >= 57) {
if (Rs < 2478) {
if (Cs < 1840) {
if (Rs < 1958) {
return 4;
}
else if (Rs >= 1958) {
return 8;
}
}
else if (Cs >= 1840) {
return 4;
}
}
else if (Rs >= 2478) {
return 8;
}
}
}
}
else if (refDCval >= 52) {
if (TSC < 9) {
if (Cs < 1555) {
if (refDCval < 85) {
if (Rs < 2562) {
return 8;
}
else if (Rs >= 2562) {
return 2;
}
}
else if (refDCval >= 85) {
if (mu_mv_mag_sq < 37) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
return 2;
}
}
else if (mu_mv_mag_sq >= 37) {
return 4;
}
}
}
else if (Cs >= 1555) {
if (Cs < 2713) {
if (Cs < 1878) {
if (ssDCval < 95) {
if (ssDCval < 80) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (Rs < 2659) {
return 8;
}
else if (Rs >= 2659) {
return 2;
}
}
}
else if (ssDCval >= 80) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (Rs < 1987) {
return 2;
}
else if (Rs >= 1987) {
return 4;
}
}
}
}
else if (ssDCval >= 95) {
if (refDCval < 97) {
if (Rs < 2492) {
return 8;
}
else if (Rs >= 2492) {
return 2;
}
}
else if (refDCval >= 97) {
return 8;
}
}
}
else if (Cs >= 1878) {
if (mu_mv_mag_sq < 20) {
if (Cs < 2326) {
if (diffMVdiffVal < -1) {
if (Rs < 2124) {
return 8;
}
else if (Rs >= 2124) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (Cs >= 2326) {
return 8;
}
}
else if (mu_mv_mag_sq >= 20) {
if (SC < 3916) {
return 1;
}
else if (SC >= 3916) {
return 8;
}
}
}
}
else if (Cs >= 2713) {
if (Rs < 2352) {
if (Rs < 2047) {
return 8;
}
else if (Rs >= 2047) {
return 1;
}
}
else if (Rs >= 2352) {
return 2;
}
}
}
}
else if (TSC >= 9) {
if (mcTcor < 74) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
if (Rs < 2013) {
return 1;
}
else if (Rs >= 2013) {
return 8;
}
}
}
else if (mcTcor >= 74) {
if (mu_mv_mag_sq < 3) {
if (diffAFD < -1) {
if (refDCval < 73) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
else if (refDCval >= 73) {
if (RsDiff < 9) {
if (AFD < 17) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (AFD >= 17) {
if (Cs < 1957) {
return 8;
}
else if (Cs >= 1957) {
return 4;
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 92) {
if (Cs < 2420) {
return 4;
}
else if (Cs >= 2420) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (ssDCval >= 92) {
if (Rs < 2882) {
return 8;
}
else if (Rs >= 2882) {
return 2;
}
}
}
}
}
else if (diffAFD >= -1) {
if (diffAFD < 1) {
if (posBalance < 64) {
if (RsDiff < 13) {
if (AFD < 14) {
if (Rs < 2229) {
return 8;
}
else if (Rs >= 2229) {
if (Rs < 2555) {
return 4;
}
else if (Rs >= 2555) {
return 2;
}
}
}
else if (AFD >= 14) {
if (TSC < 10) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
if (Cs < 2291) {
return 4;
}
else if (Cs >= 2291) {
return 8;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (RsDiff >= 13) {
return 4;
}
}
else if (posBalance >= 64) {
return 4;
}
}
else if (diffAFD >= 1) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (tcor < 84) {
if (SC < 4169) {
return 4;
}
else if (SC >= 4169) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
if (refDCval < 88) {
return 4;
}
else if (refDCval >= 88) {
return 8;
}
}
}
}
else if (tcor >= 84) {
if (Rs < 2785) {
return 8;
}
else if (Rs >= 2785) {
if (Rs < 2817) {
return 2;
}
else if (Rs >= 2817) {
if (Rs < 2895) {
return 8;
}
else if (Rs >= 2895) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 57) {
if (refDCval < 96) {
if (Rs < 2003) {
if (Cs < 2692) {
if (RsCsDiff < 129) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (Rs < 1994) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (Rs < 1881) {
return 8;
}
else if (Rs >= 1881) {
return 1;
}
}
}
else if (Rs >= 1994) {
return 1;
}
}
}
else if (RsCsDiff >= 129) {
return 8;
}
}
else if (Cs >= 2692) {
return 4;
}
}
else if (Rs >= 2003) {
if (diffRsCsdiff < 144) {
if (mu_mv_mag_sq < 53) {
return 8;
}
else if (mu_mv_mag_sq >= 53) {
if (AFD < 16) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
if (TSC < 11) {
if (mu_mv_mag_sq < 85) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Cs < 1758) {
return 8;
}
else if (Cs >= 1758) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 85) {
return 8;
}
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (AFD >= 16) {
if (posBalance < 63) {
if (Rs < 2593) {
return 1;
}
else if (Rs >= 2593) {
return 4;
}
}
else if (posBalance >= 63) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 144) {
return 4;
}
}
}
else if (refDCval >= 96) {
if (posBalance < 55) {
if (ssDCval < 97) {
return 2;
}
else if (ssDCval >= 97) {
return 8;
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (negBalance >= 57) {
if (mu_mv_mag_sq < 51) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 2213) {
if (Rs < 2079) {
return 8;
}
else if (Rs >= 2079) {
return 2;
}
}
else if (Rs >= 2213) {
return 4;
}
}
}
}
}
}
}
}
else if (AFD >= 18) {
if (mcTcor < 95) {
if (ssDCval < 89) {
if (tcor < 76) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 77) {
if (tcor < 67) {
if (tcor < 52) {
if (Rs < 2087) {
if (Rs < 1942) {
return 2;
}
else if (Rs >= 1942) {
return 1;
}
}
else if (Rs >= 2087) {
return 8;
}
}
else if (tcor >= 52) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (negBalance < 60) {
return 4;
}
else if (negBalance >= 60) {
return 2;
}
}
}
}
else if (tcor >= 67) {
if (Cs < 2209) {
if (Rs < 1882) {
if (Rs < 1840) {
return 4;
}
else if (Rs >= 1840) {
if (Rs < 1874) {
return 1;
}
else if (Rs >= 1874) {
return 8;
}
}
}
else if (Rs >= 1882) {
return 2;
}
}
else if (Cs >= 2209) {
if (refDCval < 76) {
return 1;
}
else if (refDCval >= 76) {
return 8;
}
}
}
}
else if (ssDCval >= 77) {
if (refDCval < 80) {
if (CsDiff < 13) {
if (posBalance < 66) {
if (TSC < 12) {
if (Rs < 2265) {
return 8;
}
else if (Rs >= 2265) {
return 4;
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (posBalance >= 66) {
return 4;
}
}
else if (CsDiff >= 13) {
return 2;
}
}
else if (refDCval >= 80) {
if (Cs < 1430) {
return 8;
}
else if (Cs >= 1430) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (tcor < 70) {
if (SC < 3459) {
return 2;
}
else if (SC >= 3459) {
return 4;
}
}
else if (tcor >= 70) {
if (Rs < 2165) {
return 4;
}
else if (Rs >= 2165) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (tcor < 59) {
if (Cs < 1571) {
return 8;
}
else if (Cs >= 1571) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
return 1;
}
}
}
else if (tcor >= 59) {
if (ssDCval < 67) {
if (SC < 4446) {
return 4;
}
else if (SC >= 4446) {
return 2;
}
}
else if (ssDCval >= 67) {
if (TSC < 11) {
if (Cs < 1836) {
return 4;
}
else if (Cs >= 1836) {
return 8;
}
}
else if (TSC >= 11) {
if (mu_mv_mag_sq < 19) {
return 8;
}
else if (mu_mv_mag_sq >= 19) {
if (MVDiff < 22) {
return 1;
}
else if (MVDiff >= 22) {
if (Rs < 1882) {
return 1;
}
else if (Rs >= 1882) {
return 8;
}
}
}
}
}
}
}
}
else if (tcor >= 76) {
if (diffRsCsdiff < 1) {
if (TSC < 11) {
if (SC < 3918) {
return 4;
}
else if (SC >= 3918) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 2272) {
return 8;
}
else if (Rs >= 2272) {
return 4;
}
}
}
}
else if (TSC >= 11) {
if (ssDCval < 77) {
if (diffRsCsdiff < -2) {
if (diffMVdiffVal < -6) {
if (Rs < 2542) {
return 8;
}
else if (Rs >= 2542) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
if (Rs < 1927) {
return 8;
}
else if (Rs >= 1927) {
return 4;
}
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (ssDCval >= 77) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 1) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
return 8;
}
}
}
}
}
else if (ssDCval >= 89) {
if (TSC < 14) {
if (posBalance < 58) {
if (Cs < 2088) {
if (refDCval < 89) {
return 2;
}
else if (refDCval >= 89) {
return 8;
}
}
else if (Cs >= 2088) {
if (Rs < 2370) {
return 1;
}
else if (Rs >= 2370) {
return 2;
}
}
}
else if (posBalance >= 58) {
if (Cs < 1349) {
return 4;
}
else if (Cs >= 1349) {
if (refDCval < 88) {
if (Rs < 1932) {
return 8;
}
else if (Rs >= 1932) {
return 1;
}
}
else if (refDCval >= 88) {
return 8;
}
}
}
}
else if (TSC >= 14) {
if (gchDC < 3) {
if (Rs < 1989) {
return 4;
}
else if (Rs >= 1989) {
return 1;
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
else if (mcTcor >= 95) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (Rs < 2661) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
return 2;
}
}
else if (Rs >= 2661) {
return 1;
}
}
}
}
}
else if (Rs >= 3041) {
if (CsDiff < 9) {
if (RsDiff < 12) {
if (TSC < 8) {
if (Cs < 1787) {
if (TSC < 6) {
if (Rs < 3984) {
return 8;
}
else if (Rs >= 3984) {
return 1;
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (Cs >= 1787) {
return 8;
}
}
else if (TSC >= 8) {
if (AFD < 13) {
if (Cs < 1624) {
return 2;
}
else if (Cs >= 1624) {
if (Cs < 2664) {
return 8;
}
else if (Cs >= 2664) {
return 2;
}
}
}
else if (AFD >= 13) {
if (Rs < 3099) {
if (Rs < 3069) {
if (Rs < 3047) {
return 2;
}
else if (Rs >= 3047) {
return 8;
}
}
else if (Rs >= 3069) {
return 4;
}
}
else if (Rs >= 3099) {
if (AFD < 22) {
if (SC < 5835) {
return 8;
}
else if (SC >= 5835) {
if (SC < 5991) {
if (Rs < 3321) {
return 8;
}
else if (Rs >= 3321) {
return 4;
}
}
else if (SC >= 5991) {
return 8;
}
}
}
else if (AFD >= 22) {
return 4;
}
}
}
}
}
else if (RsDiff >= 12) {
if (Cs < 1711) {
if (refDCval < 94) {
if (Rs < 3094) {
return 8;
}
else if (Rs >= 3094) {
return 2;
}
}
else if (refDCval >= 94) {
if (posBalance < 58) {
if (Rs < 3632) {
return 4;
}
else if (Rs >= 3632) {
return 2;
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
else if (Cs >= 1711) {
if (gchDC < 2) {
if (negBalance < 48) {
if (Rs < 3196) {
return 4;
}
else if (Rs >= 3196) {
if (RsCsDiff < 233) {
return 2;
}
else if (RsCsDiff >= 233) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (ssDCval < 96) {
if (SC < 4880) {
return 2;
}
else if (SC >= 4880) {
return 4;
}
}
else if (ssDCval >= 96) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
else if (CsDiff >= 9) {
if (mcTcor < 90) {
if (CsDiff < 11) {
if (diffRsCsdiff < 25) {
if (refDCval < 71) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 2;
}
}
else if (refDCval >= 71) {
if (posBalance < 62) {
if (posBalance < 59) {
if (Cs < 2195) {
return 2;
}
else if (Cs >= 2195) {
return 8;
}
}
else if (posBalance >= 59) {
return 4;
}
}
else if (posBalance >= 62) {
if (TSC < 12) {
if (Rs < 3172) {
return 2;
}
else if (Rs >= 3172) {
return 4;
}
}
else if (TSC >= 12) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 25) {
if (Rs < 3844) {
return 8;
}
else if (Rs >= 3844) {
return 1;
}
}
}
else if (CsDiff >= 11) {
if (diffRsCsdiff < -97) {
return 8;
}
else if (diffRsCsdiff >= -97) {
if (SC < 6243) {
return 4;
}
else if (SC >= 6243) {
return 1;
}
}
}
}
else if (mcTcor >= 90) {
if (TSC < 13) {
if (MVDiff < 16) {
if (posBalance < 52) {
if (SC < 5777) {
return 8;
}
else if (SC >= 5777) {
return 2;
}
}
else if (posBalance >= 52) {
if (Rs < 3075) {
return 2;
}
else if (Rs >= 3075) {
if (diffRsCsdiff < 113) {
return 4;
}
else if (diffRsCsdiff >= 113) {
return 8;
}
}
}
}
else if (MVDiff >= 16) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 3224) {
return 4;
}
else if (Rs >= 3224) {
return 2;
}
}
}
}
else if (TSC >= 13) {
return 8;
}
}
}
}
}
else if (Cs >= 2826) {
if (MVDiff < 6) {
if (mcTcor < 94) {
if (Cs < 3121) {
if (refDCval < 81) {
if (mu_mv_mag_sq < 70) {
if (Rs < 3155) {
return 1;
}
else if (Rs >= 3155) {
return 2;
}
}
else if (mu_mv_mag_sq >= 70) {
if (Rs < 2379) {
return 4;
}
else if (Rs >= 2379) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (tcor < 74) {
if (Rs < 2936) {
return 8;
}
else if (Rs >= 2936) {
return 4;
}
}
else if (tcor >= 74) {
if (Rs < 3531) {
if (Cs < 2915) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
if (RsCsDiff < 122) {
return 4;
}
else if (RsCsDiff >= 122) {
return 8;
}
}
}
}
else if (Cs >= 2915) {
return 4;
}
}
else if (Rs >= 3531) {
return 8;
}
}
}
}
else if (Cs >= 3121) {
return 4;
}
}
else if (mcTcor >= 94) {
if (Rs < 3278) {
return 1;
}
else if (Rs >= 3278) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (negBalance < 52) {
if (RsDiff < 11) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
if (Rs < 2933) {
return 4;
}
else if (Rs >= 2933) {
return 8;
}
}
}
else if (RsDiff >= 11) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
return 1;
}
}
}
else if (negBalance >= 52) {
if (refDCval < 88) {
if (tcor < 59) {
return 4;
}
else if (tcor >= 59) {
if (TSC < 10) {
if (Rs < 2688) {
return 8;
}
else if (Rs >= 2688) {
return 4;
}
}
else if (TSC >= 10) {
if (diffTSC < -2) {
if (Rs < 2707) {
return 4;
}
else if (Rs >= 2707) {
return 8;
}
}
else if (diffTSC >= -2) {
return 8;
}
}
}
}
else if (refDCval >= 88) {
if (Rs < 1831) {
return 8;
}
else if (Rs >= 1831) {
return 4;
}
}
}
}
}
}
}
else if (TSC >= 15) {
if (mcTcor < 85) {
if (negBalance < 54) {
if (ssDCval < 125) {
if (MVDiff < 145) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
return 4;
}
}
else if (MVDiff >= 145) {
if (Rs < 1124) {
return 2;
}
else if (Rs >= 1124) {
return 4;
}
}
}
else if (ssDCval >= 125) {
return 4;
}
}
else if (negBalance >= 54) {
if (MVDiff < 376) {
if (RsCsDiff < 353) {
return 4;
}
else if (RsCsDiff >= 353) {
if (Rs < 2866) {
return 4;
}
else if (Rs >= 2866) {
return 8;
}
}
}
else if (MVDiff >= 376) {
return 1;
}
}
}
else if (mcTcor >= 85) {
if (MVDiff < 22) {
return 1;
}
else if (MVDiff >= 22) {
if (Cs < 3423) {
return 8;
}
else if (Cs >= 3423) {
return 4;
}
}
}
}
}
}
else if (SC >= 6508) {
if (posBalance < 46) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (posBalance >= 46) {
if (refDCval < 97) {
if (TSC < 13) {
if (RsCsDiff < 176) {
if (mu_mv_mag_sq < 66) {
return 8;
}
else if (mu_mv_mag_sq >= 66) {
if (mu_mv_mag_sq < 72) {
return 1;
}
else if (mu_mv_mag_sq >= 72) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
else if (RsCsDiff >= 176) {
if (tcor < 83) {
if (mu_mv_mag_sq < 2) {
if (diffRsCsdiff < -13) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
if (Rs < 4483) {
return 2;
}
else if (Rs >= 4483) {
return 8;
}
}
}
else if (diffRsCsdiff >= -13) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (refDCval < 71) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (refDCval >= 71) {
if (Cs < 3441) {
return 8;
}
else if (Cs >= 3441) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
return 8;
}
}
}
}
}
else if (tcor >= 83) {
if (mcTcor < 90) {
if (Cs < 3347) {
if (Rs < 4010) {
return 1;
}
else if (Rs >= 4010) {
if (Rs < 5055) {
return 4;
}
else if (Rs >= 5055) {
return 2;
}
}
}
else if (Cs >= 3347) {
if (Rs < 6044) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
if (Rs < 3522) {
return 8;
}
else if (Rs >= 3522) {
return 4;
}
}
}
else if (Rs >= 6044) {
return 4;
}
}
}
else if (mcTcor >= 90) {
if (Rs < 4462) {
if (refDCval < 87) {
if (refDCval < 84) {
if (Cs < 3354) {
if (Rs < 3482) {
return 1;
}
else if (Rs >= 3482) {
return 8;
}
}
else if (Cs >= 3354) {
return 4;
}
}
else if (refDCval >= 84) {
return 2;
}
}
else if (refDCval >= 87) {
return 4;
}
}
else if (Rs >= 4462) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
return 8;
}
}
}
}
}
}
else if (TSC >= 13) {
if (Rs < 5896) {
if (mu_mv_mag_sq < 2) {
if (Cs < 3797) {
if (refDCval < 90) {
if (Rs < 3967) {
return 8;
}
else if (Rs >= 3967) {
return 4;
}
}
else if (refDCval >= 90) {
return 4;
}
}
else if (Cs >= 3797) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (ssDCval < 97) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 8;
}
}
else if (ssDCval >= 97) {
if (Rs < 3431) {
return 4;
}
else if (Rs >= 3431) {
return 8;
}
}
}
}
else if (Rs >= 5896) {
return 4;
}
}
}
else if (refDCval >= 97) {
if (Rs < 4185) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
return 4;
}
}
else if (Rs >= 4185) {
if (mu_mv_mag_sq < 56) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
return 8;
}
}
else if (mu_mv_mag_sq >= 56) {
if (Rs < 4513) {
return 4;
}
else if (Rs >= 4513) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 101) {
if (AFD < 20) {
if (AFD < 12) {
if (Rs < 3762) {
if (Cs < 2961) {
if (MVDiff < 47) {
if (posBalance < 15) {
return 1;
}
else if (posBalance >= 15) {
if (SC < 6180) {
if (Rs < 1779) {
if (ssDCval < 148) {
return 2;
}
else if (ssDCval >= 148) {
return 8;
}
}
else if (Rs >= 1779) {
return 8;
}
}
else if (SC >= 6180) {
return 2;
}
}
}
else if (MVDiff >= 47) {
if (Rs < 3024) {
return 2;
}
else if (Rs >= 3024) {
return 1;
}
}
}
else if (Cs >= 2961) {
if (mu_mv_mag_sq < 62) {
if (Cs < 3929) {
return 4;
}
else if (Cs >= 3929) {
return 1;
}
}
else if (mu_mv_mag_sq >= 62) {
if (RsDiff < 11) {
if (Rs < 2923) {
return 1;
}
else if (Rs >= 2923) {
return 8;
}
}
else if (RsDiff >= 11) {
return 2;
}
}
}
}
else if (Rs >= 3762) {
if (negBalance < 48) {
if (posBalance < 34) {
if (negBalance < 24) {
return 8;
}
else if (negBalance >= 24) {
return 1;
}
}
else if (posBalance >= 34) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (SC < 6834) {
return 8;
}
else if (SC >= 6834) {
return 1;
}
}
}
}
else if (negBalance >= 48) {
if (Rs < 4755) {
return 1;
}
else if (Rs >= 4755) {
return 8;
}
}
}
}
else if (AFD >= 12) {
if (ssDCval < 123) {
if (Rs < 3549) {
if (Rs < 2936) {
if (mcTcor < 93) {
if (refDCval < 121) {
if (CsDiff < 10) {
if (ssDCval < 108) {
if (tcor < 64) {
return 4;
}
else if (tcor >= 64) {
if (SC < 4777) {
if (negBalance < 53) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (mcTcor < 84) {
return 4;
}
else if (mcTcor >= 84) {
return 8;
}
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (SC >= 4777) {
if (Cs < 2487) {
return 4;
}
else if (Cs >= 2487) {
return 8;
}
}
}
}
else if (ssDCval >= 108) {
return 8;
}
}
else if (CsDiff >= 10) {
return 8;
}
}
else if (refDCval >= 121) {
if (negBalance < 45) {
return 4;
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 1579) {
if (Cs < 1921) {
if (Rs < 1476) {
return 4;
}
else if (Rs >= 1476) {
return 8;
}
}
else if (Cs >= 1921) {
if (ssDCval < 108) {
return 2;
}
else if (ssDCval >= 108) {
return 1;
}
}
}
else if (Rs >= 1579) {
if (negBalance < 48) {
if (MVDiff < 19) {
if (Rs < 2047) {
return 1;
}
else if (Rs >= 2047) {
return 2;
}
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (negBalance >= 48) {
if (diffMVdiffVal < -9) {
return 1;
}
else if (diffMVdiffVal >= -9) {
if (Rs < 1631) {
if (Rs < 1597) {
return 8;
}
else if (Rs >= 1597) {
return 4;
}
}
else if (Rs >= 1631) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2936) {
if (refDCval < 108) {
if (negBalance < 49) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Cs < 1946) {
return 4;
}
else if (Cs >= 1946) {
return 2;
}
}
}
else if (negBalance >= 49) {
if (gchDC < 2) {
if (TSC < 13) {
if (ssDCval < 101) {
return 2;
}
else if (ssDCval >= 101) {
return 4;
}
}
else if (TSC >= 13) {
return 8;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (refDCval >= 108) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
else if (Rs >= 3549) {
if (tcor < 87) {
if (mcTcor < 95) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Rs < 3680) {
return 2;
}
else if (Rs >= 3680) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (Rs < 4775) {
return 8;
}
else if (Rs >= 4775) {
return 4;
}
}
}
else if (tcor >= 87) {
if (Rs < 5507) {
if (ssDCval < 105) {
if (Cs < 2053) {
if (Rs < 3835) {
return 2;
}
else if (Rs >= 3835) {
return 8;
}
}
else if (Cs >= 2053) {
if (Rs < 4667) {
return 1;
}
else if (Rs >= 4667) {
return 8;
}
}
}
else if (ssDCval >= 105) {
return 8;
}
}
else if (Rs >= 5507) {
return 4;
}
}
}
}
else if (ssDCval >= 123) {
if (SC < 4536) {
if (mu_mv_mag_sq < 2) {
if (tcor < 68) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
else if (tcor >= 68) {
if (TSC < 11) {
if (Cs < 2180) {
return 8;
}
else if (Cs >= 2180) {
return 1;
}
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (ssDCval < 165) {
if (mu_mv_mag_sq < 32) {
if (negBalance < 44) {
if (SC < 3193) {
return 1;
}
else if (SC >= 3193) {
if (Rs < 1377) {
return 8;
}
else if (Rs >= 1377) {
return 2;
}
}
}
else if (negBalance >= 44) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Cs < 1771) {
return 1;
}
else if (Cs >= 1771) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
return 8;
}
}
else if (ssDCval >= 165) {
return 2;
}
}
}
else if (SC >= 4536) {
if (mcTcor < 85) {
if (Rs < 4061) {
if (tcor < 63) {
if (TSC < 9) {
if (Rs < 2828) {
return 2;
}
else if (Rs >= 2828) {
return 8;
}
}
else if (TSC >= 9) {
return 1;
}
}
else if (tcor >= 63) {
if (diffMVdiffVal < 14) {
if (RsCsDiff < 379) {
return 8;
}
else if (RsCsDiff >= 379) {
if (Rs < 3917) {
return 8;
}
else if (Rs >= 3917) {
return 1;
}
}
}
else if (diffMVdiffVal >= 14) {
return 1;
}
}
}
else if (Rs >= 4061) {
if (Rs < 4397) {
if (tcor < 65) {
return 8;
}
else if (tcor >= 65) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (Rs >= 4397) {
return 8;
}
}
}
else if (mcTcor >= 85) {
if (RsCsDiff < 189) {
if (SC < 6534) {
if (posBalance < 48) {
if (Rs < 2076) {
return 8;
}
else if (Rs >= 2076) {
return 4;
}
}
else if (posBalance >= 48) {
if (Cs < 2571) {
if (negBalance < 52) {
if (negBalance < 43) {
if (Rs < 2688) {
return 4;
}
else if (Rs >= 2688) {
return 8;
}
}
else if (negBalance >= 43) {
return 8;
}
}
else if (negBalance >= 52) {
if (posBalance < 60) {
if (Rs < 2782) {
return 1;
}
else if (Rs >= 2782) {
return 8;
}
}
else if (posBalance >= 60) {
if (ssDCval < 145) {
return 1;
}
else if (ssDCval >= 145) {
if (Rs < 3536) {
return 8;
}
else if (Rs >= 3536) {
return 4;
}
}
}
}
}
else if (Cs >= 2571) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
if (Rs < 3053) {
return 4;
}
else if (Rs >= 3053) {
return 8;
}
}
}
}
}
else if (SC >= 6534) {
return 4;
}
}
else if (RsCsDiff >= 189) {
if (Rs < 4774) {
if (SC < 4994) {
if (Cs < 2052) {
if (SC < 4816) {
return 4;
}
else if (SC >= 4816) {
return 8;
}
}
else if (Cs >= 2052) {
if (Rs < 2781) {
return 8;
}
else if (Rs >= 2781) {
return 2;
}
}
}
else if (SC >= 4994) {
if (Rs < 4034) {
if (TSC < 13) {
if (Cs < 6284) {
if (negBalance < 47) {
return 4;
}
else if (negBalance >= 47) {
if (Rs < 3012) {
return 8;
}
else if (Rs >= 3012) {
if (diffRsCsdiff < -105) {
return 8;
}
else if (diffRsCsdiff >= -105) {
if (Rs < 4026) {
return 4;
}
else if (Rs >= 4026) {
return 8;
}
}
}
}
}
else if (Cs >= 6284) {
return 8;
}
}
else if (TSC >= 13) {
return 8;
}
}
else if (Rs >= 4034) {
return 4;
}
}
}
else if (Rs >= 4774) {
if (Rs < 6461) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (Rs >= 6461) {
return 2;
}
}
}
}
}
}
}
}
else if (AFD >= 20) {
if (SC < 3243) {
if (Rs < 1159) {
if (diffRsCsdiff < 26) {
return 2;
}
else if (diffRsCsdiff >= 26) {
if (Rs < 624) {
return 8;
}
else if (Rs >= 624) {
return 1;
}
}
}
else if (Rs >= 1159) {
if (Cs < 1881) {
if (ssDCval < 126) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (Rs < 1422) {
return 4;
}
else if (Rs >= 1422) {
return 2;
}
}
}
else if (ssDCval >= 126) {
if (Rs < 1263) {
return 2;
}
else if (Rs >= 1263) {
return 8;
}
}
}
else if (Cs >= 1881) {
if (Rs < 1237) {
return 2;
}
else if (Rs >= 1237) {
return 1;
}
}
}
}
else if (SC >= 3243) {
if (negBalance < 57) {
if (ssDCval < 111) {
if (RsCsDiff < 345) {
if (mcTcor < 90) {
if (MVDiff < 45) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (MVDiff >= 45) {
return 8;
}
}
else if (mcTcor >= 90) {
if (diffRsCsdiff < 24) {
if (Rs < 3058) {
if (Rs < 1682) {
return 4;
}
else if (Rs >= 1682) {
return 2;
}
}
else if (Rs >= 3058) {
return 4;
}
}
else if (diffRsCsdiff >= 24) {
if (Rs < 2336) {
return 1;
}
else if (Rs >= 2336) {
if (Rs < 2615) {
return 8;
}
else if (Rs >= 2615) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 345) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
return 2;
}
}
}
else if (ssDCval >= 111) {
if (Rs < 4422) {
if (posBalance < 61) {
if (diffRsCsdiff < -33) {
if (diffRsCsdiff < -90) {
if (Rs < 3011) {
return 8;
}
else if (Rs >= 3011) {
return 4;
}
}
else if (diffRsCsdiff >= -90) {
if (negBalance < 33) {
return 2;
}
else if (negBalance >= 33) {
return 4;
}
}
}
else if (diffRsCsdiff >= -33) {
if (SC < 6802) {
if (ssDCval < 158) {
return 8;
}
else if (ssDCval >= 158) {
if (Rs < 2749) {
return 8;
}
else if (Rs >= 2749) {
return 4;
}
}
}
else if (SC >= 6802) {
if (MVDiff < 25) {
if (Rs < 3286) {
return 8;
}
else if (Rs >= 3286) {
return 4;
}
}
else if (MVDiff >= 25) {
return 2;
}
}
}
}
else if (posBalance >= 61) {
if (negBalance < 56) {
if (diffAFD < 12) {
if (gchDC < 1) {
if (Cs < 1584) {
return 1;
}
else if (Cs >= 1584) {
return 8;
}
}
else if (gchDC >= 1) {
if (diffMVdiffVal < 31) {
return 8;
}
else if (diffMVdiffVal >= 31) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 2;
}
}
}
}
else if (diffAFD >= 12) {
return 1;
}
}
else if (negBalance >= 56) {
if (RsDiff < 11) {
return 2;
}
else if (RsDiff >= 11) {
return 8;
}
}
}
}
else if (Rs >= 4422) {
if (AFD < 23) {
return 4;
}
else if (AFD >= 23) {
return 1;
}
}
}
}
else if (negBalance >= 57) {
if (RsDiff < 10) {
if (diffMVdiffVal < 2) {
if (AFD < 21) {
if (Cs < 2658) {
return 1;
}
else if (Cs >= 2658) {
return 4;
}
}
else if (AFD >= 21) {
if (diffTSC < -2) {
return 1;
}
else if (diffTSC >= -2) {
return 2;
}
}
}
else if (diffMVdiffVal >= 2) {
if (negBalance < 61) {
return 8;
}
else if (negBalance >= 61) {
return 1;
}
}
}
else if (RsDiff >= 10) {
if (diffAFD < 2) {
if (Rs < 1848) {
if (AFD < 27) {
if (Rs < 1784) {
return 2;
}
else if (Rs >= 1784) {
return 8;
}
}
else if (AFD >= 27) {
return 8;
}
}
else if (Rs >= 1848) {
if (tcor < 59) {
if (Cs < 1621) {
return 4;
}
else if (Cs >= 1621) {
return 8;
}
}
else if (tcor >= 59) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
if (mu_mv_mag_sq < 27) {
if (tcor < 67) {
if (posBalance < 57) {
if (Rs < 3169) {
return 8;
}
else if (Rs >= 3169) {
return 4;
}
}
else if (posBalance >= 57) {
return 1;
}
}
else if (tcor >= 67) {
if (tcor < 70) {
return 8;
}
else if (tcor >= 70) {
if (CsDiff < 9) {
if (MVDiff < 41) {
return 8;
}
else if (MVDiff >= 41) {
return 1;
}
}
else if (CsDiff >= 9) {
if (tcor < 78) {
return 1;
}
else if (tcor >= 78) {
if (Rs < 2192) {
return 1;
}
else if (Rs >= 2192) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (Cs < 2508) {
return 8;
}
else if (Cs >= 2508) {
return 4;
}
}
}
}
}
}
else if (diffAFD >= 2) {
if (MVDiff < 98) {
if (Rs < 2713) {
if (Cs < 3915) {
return 8;
}
else if (Cs >= 3915) {
return 2;
}
}
else if (Rs >= 2713) {
return 2;
}
}
else if (MVDiff >= 98) {
return 4;
}
}
}
}
}
}
}
}
else if (RsDiff >= 15) {
if (Cs < 1853) {
if (MVDiff < 13) {
if (ssDCval < 63) {
if (Rs < 3373) {
if (Rs < 2584) {
return 4;
}
else if (Rs >= 2584) {
return 8;
}
}
else if (Rs >= 3373) {
return 1;
}
}
else if (ssDCval >= 63) {
if (negBalance < 53) {
if (tcor < 65) {
return 4;
}
else if (tcor >= 65) {
return 2;
}
}
else if (negBalance >= 53) {
if (RsDiff < 18) {
if (ssDCval < 99) {
return 2;
}
else if (ssDCval >= 99) {
if (Rs < 2466) {
return 2;
}
else if (Rs >= 2466) {
return 4;
}
}
}
else if (RsDiff >= 18) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
return 2;
}
}
}
}
}
else if (MVDiff >= 13) {
if (AFD < 27) {
if (ssDCval < 101) {
if (mu_mv_mag_sq < 13) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
if (diffRsCsdiff < 72) {
if (Rs < 2935) {
return 8;
}
else if (Rs >= 2935) {
return 4;
}
}
else if (diffRsCsdiff >= 72) {
if (Cs < 1800) {
return 2;
}
else if (Cs >= 1800) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 2962) {
if (Rs < 2814) {
return 8;
}
else if (Rs >= 2814) {
if (Rs < 2880) {
return 2;
}
else if (Rs >= 2880) {
return 1;
}
}
}
else if (Rs >= 2962) {
if (CsDiff < 9) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 3787) {
return 4;
}
else if (Rs >= 3787) {
return 2;
}
}
}
else if (CsDiff >= 9) {
if (Cs < 1741) {
return 2;
}
else if (Cs >= 1741) {
return 8;
}
}
}
}
}
else if (ssDCval >= 101) {
if (tcor < 64) {
if (CsDiff < 8) {
if (Cs < 1722) {
return 4;
}
else if (Cs >= 1722) {
return 8;
}
}
else if (CsDiff >= 8) {
if (diffTSC < 2) {
if (Cs < 1333) {
return 4;
}
else if (Cs >= 1333) {
return 8;
}
}
else if (diffTSC >= 2) {
if (Rs < 2409) {
return 8;
}
else if (Rs >= 2409) {
return 1;
}
}
}
}
else if (tcor >= 64) {
if (posBalance < 50) {
if (Rs < 6161) {
return 1;
}
else if (Rs >= 6161) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 48) {
if (ssDCval < 108) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 2;
}
}
else if (ssDCval >= 108) {
if (Rs < 5963) {
if (diffAFD < -8) {
return 2;
}
else if (diffAFD >= -8) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
if (Rs < 3675) {
return 8;
}
else if (Rs >= 3675) {
return 1;
}
}
}
}
}
else if (Rs >= 5963) {
if (ssDCval < 128) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Cs < 1710) {
return 2;
}
else if (Cs >= 1710) {
return 8;
}
}
}
else if (ssDCval >= 128) {
if (diffRsCsdiff < 251) {
if (RsDiff < 20) {
if (Rs < 6384) {
return 2;
}
else if (Rs >= 6384) {
return 8;
}
}
else if (RsDiff >= 20) {
return 2;
}
}
else if (diffRsCsdiff >= 251) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (Rs < 4031) {
return 2;
}
else if (Rs >= 4031) {
return 8;
}
}
}
}
}
}
else if (AFD >= 27) {
if (CsDiff < 11) {
if (tcor < 72) {
if (SC < 5438) {
if (Cs < 1370) {
if (Rs < 3391) {
return 2;
}
else if (Rs >= 3391) {
return 4;
}
}
else if (Cs >= 1370) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (diffAFD < 10) {
return 2;
}
else if (diffAFD >= 10) {
return 1;
}
}
}
}
}
else if (SC >= 5438) {
if (negBalance < 56) {
return 4;
}
else if (negBalance >= 56) {
if (posBalance < 56) {
if (Rs < 4779) {
return 4;
}
else if (Rs >= 4779) {
return 8;
}
}
else if (posBalance >= 56) {
return 2;
}
}
}
}
else if (tcor >= 72) {
if (CsDiff < 8) {
if (Rs < 4447) {
return 4;
}
else if (Rs >= 4447) {
return 8;
}
}
else if (CsDiff >= 8) {
return 1;
}
}
}
else if (CsDiff >= 11) {
if (refDCval < 112) {
if (posBalance < 89) {
if (posBalance < 59) {
if (ssDCval < 104) {
if (Cs < 1418) {
return 1;
}
else if (Cs >= 1418) {
return 8;
}
}
else if (ssDCval >= 104) {
if (Rs < 2237) {
return 2;
}
else if (Rs >= 2237) {
return 4;
}
}
}
else if (posBalance >= 59) {
if (CsDiff < 17) {
return 2;
}
else if (CsDiff >= 17) {
return 8;
}
}
}
else if (posBalance >= 89) {
return 1;
}
}
else if (refDCval >= 112) {
if (diffTSC < -3) {
if (Rs < 4348) {
return 8;
}
else if (Rs >= 4348) {
return 2;
}
}
else if (diffTSC >= -3) {
if (AFD < 29) {
if (Rs < 3168) {
return 8;
}
else if (Rs >= 3168) {
return 1;
}
}
else if (AFD >= 29) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1853) {
if (mu_mv_mag_sq < 71) {
if (mcTcor < 96) {
if (ssDCval < 152) {
if (ssDCval < 106) {
if (Rs < 6044) {
if (TSC < 14) {
if (tcor < 80) {
if (ssDCval < 89) {
if (RsDiff < 19) {
if (gchDC < 2) {
if (diffAFD < -3) {
if (Rs < 3294) {
return 2;
}
else if (Rs >= 3294) {
return 4;
}
}
else if (diffAFD >= -3) {
if (diffRsCsdiff < 71) {
if (MVDiff < 42) {
if (MVDiff < 16) {
if (ssDCval < 72) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (ssDCval >= 72) {
return 2;
}
}
else if (MVDiff >= 16) {
return 4;
}
}
else if (MVDiff >= 42) {
if (Cs < 2326) {
return 8;
}
else if (Cs >= 2326) {
return 4;
}
}
}
else if (diffRsCsdiff >= 71) {
if (posBalance < 62) {
if (Cs < 2323) {
return 4;
}
else if (Cs >= 2323) {
if (mcTcor < 87) {
return 4;
}
else if (mcTcor >= 87) {
if (MVDiff < 23) {
return 2;
}
else if (MVDiff >= 23) {
return 4;
}
}
}
}
else if (posBalance >= 62) {
if (Rs < 3314) {
return 8;
}
else if (Rs >= 3314) {
return 2;
}
}
}
}
}
else if (gchDC >= 2) {
if (Rs < 3151) {
if (Cs < 1994) {
return 8;
}
else if (Cs >= 1994) {
return 2;
}
}
else if (Rs >= 3151) {
if (Rs < 3594) {
if (Rs < 3287) {
return 8;
}
else if (Rs >= 3287) {
return 4;
}
}
else if (Rs >= 3594) {
if (Rs < 4274) {
return 8;
}
else if (Rs >= 4274) {
return 4;
}
}
}
}
}
else if (RsDiff >= 19) {
if (mu_mv_mag_sq < 53) {
if (refDCval < 72) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
return 4;
}
}
}
else if (refDCval >= 72) {
if (diffMVdiffVal < 3) {
if (SC < 8494) {
return 2;
}
else if (SC >= 8494) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 53) {
return 4;
}
}
}
else if (ssDCval >= 89) {
if (mcTcor < 84) {
return 8;
}
else if (mcTcor >= 84) {
if (AFD < 31) {
if (Cs < 3695) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (TSC < 12) {
if (AFD < 25) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (AFD >= 25) {
return 4;
}
}
else if (TSC >= 12) {
if (refDCval < 104) {
return 4;
}
else if (refDCval >= 104) {
return 8;
}
}
}
}
else if (Cs >= 3695) {
return 8;
}
}
else if (AFD >= 31) {
return 1;
}
}
}
}
else if (tcor >= 80) {
if (CsDiff < 16) {
if (negBalance < 50) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (Cs < 4108) {
if (SC < 6009) {
return 8;
}
else if (SC >= 6009) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 2;
}
}
}
else if (Cs >= 4108) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (mu_mv_mag_sq < 2) {
if (RsDiff < 16) {
return 4;
}
else if (RsDiff >= 16) {
if (Rs < 5937) {
if (Rs < 3822) {
return 8;
}
else if (Rs >= 3822) {
return 2;
}
}
else if (Rs >= 5937) {
if (Rs < 6004) {
return 8;
}
else if (Rs >= 6004) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (CsDiff < 10) {
if (Cs < 2449) {
return 4;
}
else if (Cs >= 2449) {
return 2;
}
}
else if (CsDiff >= 10) {
if (MVDiff < 77) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
else if (MVDiff >= 77) {
return 4;
}
}
}
}
}
else if (CsDiff >= 16) {
return 8;
}
}
}
else if (TSC >= 14) {
if (refDCval < 59) {
return 4;
}
else if (refDCval >= 59) {
if (posBalance < 62) {
if (diffAFD < 6) {
if (RsCsDiff < 397) {
if (mcTcor < 80) {
return 8;
}
else if (mcTcor >= 80) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
if (Rs < 3648) {
return 1;
}
else if (Rs >= 3648) {
return 4;
}
}
}
}
else if (RsCsDiff >= 397) {
if (negBalance < 61) {
if (RsCsDiff < 521) {
if (posBalance < 53) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
return 1;
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (RsCsDiff >= 521) {
if (posBalance < 45) {
if (Cs < 6681) {
return 4;
}
else if (Cs >= 6681) {
return 8;
}
}
else if (posBalance >= 45) {
if (posBalance < 61) {
if (Rs < 4080) {
if (Rs < 4021) {
return 2;
}
else if (Rs >= 4021) {
return 8;
}
}
else if (Rs >= 4080) {
return 2;
}
}
else if (posBalance >= 61) {
if (Rs < 5373) {
return 8;
}
else if (Rs >= 5373) {
return 2;
}
}
}
}
}
else if (negBalance >= 61) {
if (diffAFD < 1) {
if (tcor < 63) {
return 1;
}
else if (tcor >= 63) {
if (Rs < 2836) {
return 2;
}
else if (Rs >= 2836) {
return 8;
}
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
else if (diffAFD >= 6) {
if (gchDC < 44) {
if (Cs < 3429) {
return 8;
}
else if (Cs >= 3429) {
return 2;
}
}
else if (gchDC >= 44) {
return 4;
}
}
}
else if (posBalance >= 62) {
if (tcor < 60) {
if (Cs < 2988) {
return 1;
}
else if (Cs >= 2988) {
if (Rs < 5198) {
return 4;
}
else if (Rs >= 5198) {
return 8;
}
}
}
else if (tcor >= 60) {
if (RsDiff < 19) {
if (refDCval < 95) {
return 8;
}
else if (refDCval >= 95) {
return 4;
}
}
else if (RsDiff >= 19) {
if (mcTcor < 89) {
if (posBalance < 64) {
return 4;
}
else if (posBalance >= 64) {
if (Rs < 3322) {
return 4;
}
else if (Rs >= 3322) {
return 8;
}
}
}
else if (mcTcor >= 89) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 6044) {
if (mcTcor < 77) {
if (Rs < 7063) {
return 2;
}
else if (Rs >= 7063) {
return 1;
}
}
else if (mcTcor >= 77) {
if (mu_mv_mag_sq < 56) {
if (gchDC < 4) {
if (RsDiff < 30) {
if (RsDiff < 19) {
if (ssDCval < 95) {
return 4;
}
else if (ssDCval >= 95) {
if (Rs < 6383) {
return 4;
}
else if (Rs >= 6383) {
return 8;
}
}
}
else if (RsDiff >= 19) {
return 4;
}
}
else if (RsDiff >= 30) {
return 8;
}
}
else if (gchDC >= 4) {
if (Rs < 6561) {
return 4;
}
else if (Rs >= 6561) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (TSC < 14) {
if (tcor < 83) {
return 8;
}
else if (tcor >= 83) {
return 2;
}
}
else if (TSC >= 14) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 106) {
if (gchDC < 2) {
if (RsCsDiff < 783) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 48) {
return 2;
}
else if (mu_mv_mag_sq >= 48) {
return 8;
}
}
else if (negBalance >= 44) {
if (Rs < 6428) {
if (mcTcor < 86) {
if (ssDCval < 121) {
if (Cs < 4695) {
return 8;
}
else if (Cs >= 4695) {
return 2;
}
}
else if (ssDCval >= 121) {
if (Rs < 3497) {
return 8;
}
else if (Rs >= 3497) {
if (Cs < 4412) {
if (Cs < 1997) {
return 8;
}
else if (Cs >= 1997) {
if (Rs < 5590) {
return 4;
}
else if (Rs >= 5590) {
if (Rs < 5641) {
return 8;
}
else if (Rs >= 5641) {
return 4;
}
}
}
}
else if (Cs >= 4412) {
return 8;
}
}
}
}
else if (mcTcor >= 86) {
if (ssDCval < 114) {
if (ssDCval < 110) {
return 4;
}
else if (ssDCval >= 110) {
return 2;
}
}
else if (ssDCval >= 114) {
if (negBalance < 52) {
if (ssDCval < 125) {
return 1;
}
else if (ssDCval >= 125) {
return 4;
}
}
else if (negBalance >= 52) {
if (TSC < 11) {
if (Rs < 5340) {
return 1;
}
else if (Rs >= 5340) {
return 4;
}
}
else if (TSC >= 11) {
if (Rs < 2780) {
return 1;
}
else if (Rs >= 2780) {
if (SC < 8761) {
return 8;
}
else if (SC >= 8761) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 6428) {
return 8;
}
}
}
else if (RsCsDiff >= 783) {
if (TSC < 21) {
if (refDCval < 149) {
if (Cs < 2118) {
return 2;
}
else if (Cs >= 2118) {
if (SC < 15586) {
if (AFD < 38) {
return 4;
}
else if (AFD >= 38) {
return 8;
}
}
else if (SC >= 15586) {
return 8;
}
}
}
else if (refDCval >= 149) {
if (Rs < 7710) {
return 2;
}
else if (Rs >= 7710) {
return 4;
}
}
}
else if (TSC >= 21) {
if (Rs < 8492) {
return 8;
}
else if (Rs >= 8492) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (posBalance < 60) {
if (ssDCval < 134) {
if (mu_mv_mag_sq < 50) {
if (negBalance < 55) {
if (Rs < 3971) {
return 2;
}
else if (Rs >= 3971) {
return 1;
}
}
else if (negBalance >= 55) {
if (tcor < 58) {
if (SC < 5098) {
return 2;
}
else if (SC >= 5098) {
return 1;
}
}
else if (tcor >= 58) {
if (Cs < 3616) {
return 8;
}
else if (Cs >= 3616) {
if (Rs < 6248) {
return 2;
}
else if (Rs >= 6248) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
return 8;
}
}
else if (ssDCval >= 134) {
if (refDCval < 152) {
return 8;
}
else if (refDCval >= 152) {
return 1;
}
}
}
else if (posBalance >= 60) {
if (diffMVdiffVal < 3) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
if (mcTcor < 77) {
return 8;
}
else if (mcTcor >= 77) {
if (diffRsCsdiff < 62) {
return 8;
}
else if (diffRsCsdiff >= 62) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (Rs < 3540) {
return 8;
}
else if (Rs >= 3540) {
return 4;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 152) {
if (refDCval < 166) {
if (negBalance < 36) {
if (Rs < 2967) {
return 2;
}
else if (Rs >= 2967) {
return 8;
}
}
else if (negBalance >= 36) {
if (posBalance < 55) {
if (ssDCval < 157) {
return 8;
}
else if (ssDCval >= 157) {
if (AFD < 23) {
if (Rs < 4479) {
return 4;
}
else if (Rs >= 4479) {
return 2;
}
}
else if (AFD >= 23) {
return 4;
}
}
}
else if (posBalance >= 55) {
return 4;
}
}
}
else if (refDCval >= 166) {
if (AFD < 22) {
if (Cs < 3380) {
return 4;
}
else if (Cs >= 3380) {
return 8;
}
}
else if (AFD >= 22) {
return 2;
}
}
}
}
else if (mcTcor >= 96) {
if (ssDCval < 126) {
if (Rs < 6908) {
if (Cs < 2485) {
return 4;
}
else if (Cs >= 2485) {
return 2;
}
}
else if (Rs >= 6908) {
return 1;
}
}
else if (ssDCval >= 126) {
if (Cs < 2109) {
if (Cs < 1922) {
return 1;
}
else if (Cs >= 1922) {
return 2;
}
}
else if (Cs >= 2109) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (posBalance < 59) {
if (negBalance < 36) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
if (Rs < 6480) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
else if (Rs >= 6480) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (diffRsCsdiff < -102) {
if (diffMVdiffVal < 4) {
if (RsCsDiff < 542) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (RsCsDiff >= 542) {
if (Rs < 4760) {
return 2;
}
else if (Rs >= 4760) {
return 4;
}
}
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
else if (diffRsCsdiff >= -102) {
if (AFD < 14) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 4;
}
}
else if (AFD >= 14) {
if (Rs < 4232) {
if (Rs < 2904) {
return 4;
}
else if (Rs >= 2904) {
if (Rs < 3889) {
return 8;
}
else if (Rs >= 3889) {
return 1;
}
}
}
else if (Rs >= 4232) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Cs < 4322) {
return 4;
}
else if (Cs >= 4322) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 59) {
if (Cs < 3134) {
if (CsDiff < 12) {
if (Rs < 4748) {
return 8;
}
else if (Rs >= 4748) {
return 4;
}
}
else if (CsDiff >= 12) {
return 1;
}
}
else if (Cs >= 3134) {
if (AFD < 18) {
if (Rs < 5681) {
return 8;
}
else if (Rs >= 5681) {
if (negBalance < 48) {
if (Rs < 7014) {
return 2;
}
else if (Rs >= 7014) {
return 8;
}
}
else if (negBalance >= 48) {
return 4;
}
}
}
else if (AFD >= 18) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 89) {
if (RsDiff < 9) {
if (Rs < 3686) {
if (Rs < 2981) {
if (Rs < 1539) {
if (AFD < 12) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
return 8;
}
}
else if (AFD >= 12) {
if (refDCval < 113) {
return 2;
}
else if (refDCval >= 113) {
return 4;
}
}
}
else if (Rs >= 1539) {
if (posBalance < 52) {
if (ssDCval < 98) {
if (Cs < 2252) {
if (diffRsCsdiff < 32) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 8;
}
}
else if (diffRsCsdiff >= 32) {
return 1;
}
}
else if (Cs >= 2252) {
return 4;
}
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (posBalance >= 52) {
if (Cs < 4866) {
if (Rs < 2137) {
if (MVDiff < 9) {
if (Rs < 2047) {
if (Rs < 1871) {
return 4;
}
else if (Rs >= 1871) {
return 2;
}
}
else if (Rs >= 2047) {
return 8;
}
}
else if (MVDiff >= 9) {
if (Rs < 2130) {
return 8;
}
else if (Rs >= 2130) {
return 2;
}
}
}
else if (Rs >= 2137) {
return 8;
}
}
else if (Cs >= 4866) {
return 4;
}
}
}
}
else if (Rs >= 2981) {
if (ssDCval < 105) {
if (ssDCval < 92) {
if (negBalance < 52) {
if (ssDCval < 87) {
if (mcTcor < 95) {
if (MVDiff < 3) {
if (Cs < 2606) {
return 4;
}
else if (Cs >= 2606) {
return 1;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (mcTcor >= 95) {
if (Rs < 3366) {
return 1;
}
else if (Rs >= 3366) {
return 2;
}
}
}
else if (ssDCval >= 87) {
return 8;
}
}
else if (negBalance >= 52) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
return 4;
}
}
}
else if (ssDCval >= 92) {
return 4;
}
}
else if (ssDCval >= 105) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
return 1;
}
}
}
}
}
else if (Rs >= 3686) {
if (TSC < 9) {
if (refDCval < 115) {
if (Rs < 4430) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
else if (Rs >= 4430) {
if (Cs < 3911) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (Cs < 1604) {
return 8;
}
else if (Cs >= 1604) {
return 1;
}
}
}
else if (Cs >= 3911) {
if (Rs < 5783) {
return 8;
}
else if (Rs >= 5783) {
return 2;
}
}
}
}
else if (refDCval >= 115) {
if (CsDiff < 8) {
if (Rs < 5300) {
return 1;
}
else if (Rs >= 5300) {
return 4;
}
}
else if (CsDiff >= 8) {
return 8;
}
}
}
else if (TSC >= 9) {
if (mu_mv_mag_sq < 86) {
if (Cs < 2732) {
if (Rs < 4078) {
return 4;
}
else if (Rs >= 4078) {
return 1;
}
}
else if (Cs >= 2732) {
return 8;
}
}
else if (mu_mv_mag_sq >= 86) {
if (negBalance < 43) {
return 8;
}
else if (negBalance >= 43) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
if (Cs < 2292) {
if (Rs < 3383) {
if (refDCval < 98) {
if (CsDiff < 8) {
if (gchDC < 3) {
if (Rs < 2865) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (Rs < 2192) {
return 1;
}
else if (Rs >= 2192) {
return 8;
}
}
}
else if (Rs >= 2865) {
return 2;
}
}
else if (gchDC >= 3) {
if (Rs < 2359) {
return 4;
}
else if (Rs >= 2359) {
return 8;
}
}
}
else if (CsDiff >= 8) {
if (mcTcor < 95) {
if (SC < 5109) {
return 8;
}
else if (SC >= 5109) {
if (Rs < 3285) {
return 2;
}
else if (Rs >= 3285) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (AFD < 16) {
if (CsDiff < 9) {
if (Rs < 2344) {
return 1;
}
else if (Rs >= 2344) {
if (Rs < 3145) {
return 2;
}
else if (Rs >= 3145) {
return 4;
}
}
}
else if (CsDiff >= 9) {
return 8;
}
}
else if (AFD >= 16) {
return 2;
}
}
}
}
else if (refDCval >= 98) {
if (AFD < 9) {
if (Rs < 2214) {
return 8;
}
else if (Rs >= 2214) {
return 4;
}
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (Rs >= 3383) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (ssDCval < 113) {
if (mcTcor < 92) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (MVDiff < 14) {
if (Cs < 2077) {
if (diffMVdiffVal < 4) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
if (ssDCval < 102) {
if (posBalance < 39) {
if (Rs < 4432) {
return 2;
}
else if (Rs >= 4432) {
return 8;
}
}
else if (posBalance >= 39) {
return 2;
}
}
else if (ssDCval >= 102) {
if (mu_mv_mag_sq < 87) {
return 2;
}
else if (mu_mv_mag_sq >= 87) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 4) {
if (Cs < 1927) {
return 2;
}
else if (Cs >= 1927) {
if (Rs < 4665) {
return 8;
}
else if (Rs >= 4665) {
return 2;
}
}
}
}
else if (Cs >= 2077) {
if (ssDCval < 76) {
if (Rs < 4521) {
return 8;
}
else if (Rs >= 4521) {
return 4;
}
}
else if (ssDCval >= 76) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (Cs < 2217) {
return 1;
}
else if (Cs >= 2217) {
return 2;
}
}
}
}
}
else if (MVDiff >= 14) {
if (Cs < 1874) {
if (mcTcor < 93) {
if (Rs < 4130) {
return 1;
}
else if (Rs >= 4130) {
return 8;
}
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (Cs >= 1874) {
if (RsCsDiff < 178) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
else if (RsCsDiff >= 178) {
if (Cs < 1900) {
return 8;
}
else if (Cs >= 1900) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (ssDCval < 135) {
if (posBalance < 39) {
if (Rs < 4071) {
return 8;
}
else if (Rs >= 4071) {
return 2;
}
}
else if (posBalance >= 39) {
if (diffTSC < -4) {
if (Rs < 5738) {
return 1;
}
else if (Rs >= 5738) {
return 2;
}
}
else if (diffTSC >= -4) {
return 8;
}
}
}
else if (ssDCval >= 135) {
if (mcTcor < 94) {
if (SC < 7490) {
return 1;
}
else if (SC >= 7490) {
return 2;
}
}
else if (mcTcor >= 94) {
if (Rs < 4910) {
return 1;
}
else if (Rs >= 4910) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 2292) {
if (negBalance < 40) {
if (ssDCval < 108) {
if (ssDCval < 72) {
if (Rs < 3901) {
if (Rs < 3287) {
return 4;
}
else if (Rs >= 3287) {
return 1;
}
}
else if (Rs >= 3901) {
return 8;
}
}
else if (ssDCval >= 72) {
if (AFD < 11) {
if (MVDiff < 2) {
if (Rs < 5360) {
return 4;
}
else if (Rs >= 5360) {
return 2;
}
}
else if (MVDiff >= 2) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (AFD >= 11) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
if (Rs < 5167) {
return 1;
}
else if (Rs >= 5167) {
return 2;
}
}
}
}
}
else if (ssDCval >= 108) {
if (TSC < 7) {
if (negBalance < 24) {
return 8;
}
else if (negBalance >= 24) {
return 1;
}
}
else if (TSC >= 7) {
if (Cs < 3626) {
if (Rs < 6286) {
return 8;
}
else if (Rs >= 6286) {
return 2;
}
}
else if (Cs >= 3626) {
return 2;
}
}
}
}
else if (negBalance >= 40) {
if (ssDCval < 126) {
if (Cs < 3544) {
if (MVDiff < 6) {
if (Cs < 2664) {
if (mcTcor < 91) {
return 1;
}
else if (mcTcor >= 91) {
if (mu_mv_mag_sq < 76) {
if (negBalance < 49) {
if (Rs < 4219) {
return 8;
}
else if (Rs >= 4219) {
return 2;
}
}
else if (negBalance >= 49) {
return 4;
}
}
else if (mu_mv_mag_sq >= 76) {
if (posBalance < 59) {
if (ssDCval < 89) {
if (Rs < 3841) {
return 4;
}
else if (Rs >= 3841) {
return 8;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (posBalance >= 59) {
return 2;
}
}
}
}
else if (Cs >= 2664) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Cs < 3399) {
if (RsDiff < 17) {
return 4;
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (Cs >= 3399) {
if (mu_mv_mag_sq < 100) {
return 4;
}
else if (mu_mv_mag_sq >= 100) {
return 8;
}
}
}
}
}
else if (MVDiff >= 6) {
if (posBalance < 52) {
if (mcTcor < 95) {
if (diffMVdiffVal < 0) {
if (Cs < 2761) {
return 8;
}
else if (Cs >= 2761) {
return 4;
}
}
else if (diffMVdiffVal >= 0) {
return 4;
}
}
else if (mcTcor >= 95) {
return 2;
}
}
else if (posBalance >= 52) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 8;
}
}
}
}
else if (Cs >= 3544) {
if (mu_mv_mag_sq < 78) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
return 8;
}
}
}
else if (ssDCval >= 126) {
if (mcTcor < 92) {
if (MVDiff < 6) {
if (Rs < 7668) {
return 1;
}
else if (Rs >= 7668) {
return 4;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (mcTcor >= 92) {
if (RsCsDiff < 147) {
if (Rs < 4544) {
return 8;
}
else if (Rs >= 4544) {
return 1;
}
}
else if (RsCsDiff >= 147) {
if (Rs < 6518) {
return 8;
}
else if (Rs >= 6518) {
if (Rs < 7565) {
return 2;
}
else if (Rs >= 7565) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1645) {
if (Rs < 3101) {
if (negBalance < 17) {
if (ssDCval < 61) {
if (Cs < 1458) {
return 8;
}
else if (Cs >= 1458) {
if (ssDCval < 38) {
return 8;
}
else if (ssDCval >= 38) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (ssDCval >= 61) {
if (SC < 3643) {
if (Cs < 1516) {
if (Rs < 2287) {
if (posBalance < 14) {
if (MVDiff < 111) {
return 8;
}
else if (MVDiff >= 111) {
return 4;
}
}
else if (posBalance >= 14) {
if (Rs < 1474) {
if (posBalance < 52) {
if (diffMVdiffVal < -29) {
return 8;
}
else if (diffMVdiffVal >= -29) {
return 1;
}
}
else if (posBalance >= 52) {
return 8;
}
}
else if (Rs >= 1474) {
return 8;
}
}
}
else if (Rs >= 2287) {
if (mu_mv_mag_sq < 127) {
if (Cs < 1342) {
if (Cs < 1329) {
if (Rs < 2302) {
return 2;
}
else if (Rs >= 2302) {
return 8;
}
}
else if (Cs >= 1329) {
return 8;
}
}
else if (Cs >= 1342) {
if (Rs < 2292) {
return 8;
}
else if (Rs >= 2292) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 2;
}
}
}
else if (Cs >= 1516) {
if (AFD < 1) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (AFD >= 1) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
else if (SC >= 3643) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (negBalance < 14) {
return 8;
}
else if (negBalance >= 14) {
if (Cs < 1452) {
return 2;
}
else if (Cs >= 1452) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 17) {
if (TSC < 2) {
if (Rs < 1054) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 1;
}
}
}
else if (Rs >= 1054) {
if (Rs < 1986) {
if (posBalance < 41) {
if (Rs < 1474) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (Rs >= 1474) {
if (Cs < 1483) {
return 8;
}
else if (Cs >= 1483) {
return 1;
}
}
}
else if (posBalance >= 41) {
if (Rs < 1342) {
return 4;
}
else if (Rs >= 1342) {
if (ssDCval < 77) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 4;
}
}
else if (ssDCval >= 77) {
if (SC < 2842) {
return 4;
}
else if (SC >= 2842) {
return 8;
}
}
}
}
}
else if (Rs >= 1986) {
return 8;
}
}
}
else if (TSC >= 2) {
if (Rs < 2351) {
if (Rs < 873) {
if (negBalance < 41) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
if (mu_mv_mag_sq < 128) {
if (diffRsCsdiff < -7) {
if (Rs < 811) {
return 4;
}
else if (Rs >= 811) {
return 8;
}
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 823) {
return 1;
}
else if (Rs >= 823) {
return 8;
}
}
}
}
else if (negBalance >= 41) {
if (Cs < 1483) {
return 4;
}
else if (Cs >= 1483) {
return 1;
}
}
}
else if (Rs >= 873) {
if (Rs < 1695) {
if (ssDCval < 57) {
if (mcTcor < 98) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (Rs < 1411) {
return 4;
}
else if (Rs >= 1411) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 1227) {
return 8;
}
else if (Rs >= 1227) {
return 4;
}
}
}
else if (ssDCval >= 57) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (diffRsCsdiff < -15) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 8;
}
}
else if (diffRsCsdiff >= -15) {
if (diffRsCsdiff < 10) {
if (TSCindex < 3) {
if (SC < 2974) {
if (ssDCval < 125) {
return 8;
}
else if (ssDCval >= 125) {
return 4;
}
}
else if (SC >= 2974) {
if (Rs < 1522) {
return 1;
}
else if (Rs >= 1522) {
return 4;
}
}
}
else if (TSCindex >= 3) {
if (Rs < 1161) {
if (Rs < 955) {
if (Cs < 1334) {
if (Cs < 1332) {
return 8;
}
else if (Cs >= 1332) {
return 2;
}
}
else if (Cs >= 1334) {
return 8;
}
}
else if (Rs >= 955) {
if (CsDiff < 3) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (ssDCval < 129) {
return 4;
}
else if (ssDCval >= 129) {
return 8;
}
}
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (Rs >= 1161) {
if (negBalance < 23) {
if (Rs < 1586) {
return 8;
}
else if (Rs >= 1586) {
return 4;
}
}
else if (negBalance >= 23) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 10) {
return 1;
}
}
}
}
}
else if (Rs >= 1695) {
if (ssDCval < 117) {
if (negBalance < 23) {
if (Rs < 2325) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (Rs >= 2325) {
return 8;
}
}
else if (negBalance >= 23) {
if (Cs < 1403) {
if (Rs < 1919) {
return 8;
}
else if (Rs >= 1919) {
if (diffAFD < -1) {
if (Rs < 1920) {
return 2;
}
else if (Rs >= 1920) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
return 2;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (Cs >= 1403) {
if (posBalance < 45) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (posBalance >= 45) {
return 8;
}
}
}
}
else if (ssDCval >= 117) {
if (ssDCval < 119) {
if (Cs < 1394) {
if (Rs < 1818) {
return 1;
}
else if (Rs >= 1818) {
return 8;
}
}
else if (Cs >= 1394) {
return 1;
}
}
else if (ssDCval >= 119) {
return 8;
}
}
}
}
}
else if (Rs >= 2351) {
if (ssDCval < 86) {
if (tcor < 97) {
if (Cs < 1412) {
if (Rs < 2702) {
return 4;
}
else if (Rs >= 2702) {
return 2;
}
}
else if (Cs >= 1412) {
return 8;
}
}
else if (tcor >= 97) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 137) {
if (Rs < 2925) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (Rs >= 2925) {
if (Rs < 2997) {
return 2;
}
else if (Rs >= 2997) {
return 8;
}
}
}
else if (ssDCval >= 137) {
return 1;
}
}
}
}
}
}
else if (Rs >= 3101) {
if (refDCval < 77) {
if (mu_mv_mag_sq < 126) {
if (Rs < 4122) {
if (SC < 4898) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
else if (SC >= 4898) {
if (Cs < 1337) {
if (posBalance < 34) {
if (Rs < 3904) {
return 4;
}
else if (Rs >= 3904) {
return 1;
}
}
else if (posBalance >= 34) {
return 8;
}
}
else if (Cs >= 1337) {
if (mu_mv_mag_sq < 118) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Rs < 3988) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (Rs >= 3988) {
return 8;
}
}
}
}
}
}
else if (Rs >= 4122) {
if (Cs < 1567) {
return 8;
}
else if (Cs >= 1567) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1578) {
if (negBalance < 25) {
if (Rs < 3657) {
if (mu_mv_mag_sq < 127) {
if (Rs < 3532) {
return 1;
}
else if (Rs >= 3532) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 3649) {
return 8;
}
else if (Rs >= 3649) {
return 4;
}
}
}
else if (Rs >= 3657) {
return 8;
}
}
else if (negBalance >= 25) {
if (Cs < 1425) {
return 8;
}
else if (Cs >= 1425) {
return 4;
}
}
}
else if (Cs >= 1578) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 4;
}
}
}
}
else if (refDCval >= 77) {
if (ssDCval < 114) {
if (ssDCval < 88) {
if (mu_mv_mag_sq < 126) {
return 2;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 3379) {
return 4;
}
else if (Rs >= 3379) {
if (Rs < 4024) {
return 2;
}
else if (Rs >= 4024) {
return 8;
}
}
}
}
else if (ssDCval >= 88) {
if (Rs < 4478) {
if (Cs < 1561) {
if (Rs < 3244) {
return 2;
}
else if (Rs >= 3244) {
if (Rs < 4251) {
return 4;
}
else if (Rs >= 4251) {
if (Rs < 4264) {
return 8;
}
else if (Rs >= 4264) {
return 4;
}
}
}
}
else if (Cs >= 1561) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (Rs < 3272) {
return 8;
}
else if (Rs >= 3272) {
return 4;
}
}
}
}
else if (Rs >= 4478) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (ssDCval >= 114) {
if (diffRsCsdiff < -40) {
if (Cs < 1551) {
if (Rs < 3303) {
return 2;
}
else if (Rs >= 3303) {
return 4;
}
}
else if (Cs >= 1551) {
return 1;
}
}
else if (diffRsCsdiff >= -40) {
return 2;
}
}
}
}
}
else if (Cs >= 1645) {
if (ssDCval < 98) {
if (SC < 4361) {
if (refDCval < 67) {
if (MVDiff < 3) {
if (RsCsDiff < 7) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (RsCsDiff < 2) {
if (diffAFD < -1) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
return 8;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (RsCsDiff >= 2) {
if (ssDCval < 63) {
if (refDCval < 53) {
if (Rs < 1727) {
return 4;
}
else if (Rs >= 1727) {
return 1;
}
}
else if (refDCval >= 53) {
if (Rs < 835) {
return 1;
}
else if (Rs >= 835) {
if (Cs < 1947) {
return 8;
}
else if (Cs >= 1947) {
if (Rs < 1995) {
return 8;
}
else if (Rs >= 1995) {
return 1;
}
}
}
}
}
else if (ssDCval >= 63) {
if (SC < 4142) {
if (Rs < 1166) {
if (Rs < 961) {
return 8;
}
else if (Rs >= 961) {
return 1;
}
}
else if (Rs >= 1166) {
if (SC < 4090) {
return 8;
}
else if (SC >= 4090) {
if (Rs < 2344) {
return 1;
}
else if (Rs >= 2344) {
return 8;
}
}
}
}
else if (SC >= 4142) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 7) {
return 8;
}
}
else if (MVDiff >= 3) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (negBalance < 26) {
if (Rs < 1071) {
return 1;
}
else if (Rs >= 1071) {
return 4;
}
}
else if (negBalance >= 26) {
if (Rs < 2327) {
return 8;
}
else if (Rs >= 2327) {
return 4;
}
}
}
}
}
else if (refDCval >= 67) {
if (Rs < 2566) {
if (ssDCval < 90) {
if (SC < 2761) {
if (Rs < 623) {
return 8;
}
else if (Rs >= 623) {
return 1;
}
}
else if (SC >= 2761) {
if (gchDC < 2) {
if (diffMVdiffVal < -7) {
return 4;
}
else if (diffMVdiffVal >= -7) {
if (Cs < 1651) {
if (Rs < 2489) {
return 8;
}
else if (Rs >= 2489) {
return 4;
}
}
else if (Cs >= 1651) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (ssDCval >= 90) {
if (Cs < 1861) {
return 8;
}
else if (Cs >= 1861) {
if (Cs < 2006) {
return 4;
}
else if (Cs >= 2006) {
if (gchDC < 1) {
if (SC < 4273) {
return 2;
}
else if (SC >= 4273) {
return 8;
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
}
else if (Rs >= 2566) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 4;
}
}
}
}
else if (SC >= 4361) {
if (posBalance < 3) {
if (diffRsCsdiff < -5) {
if (RsDiff < 1) {
if (Cs < 1862) {
if (Cs < 1849) {
return 8;
}
else if (Cs >= 1849) {
return 4;
}
}
else if (Cs >= 1862) {
return 8;
}
}
else if (RsDiff >= 1) {
return 4;
}
}
else if (diffRsCsdiff >= -5) {
if (refDCval < 65) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 6330) {
return 4;
}
else if (Rs >= 6330) {
return 2;
}
}
}
else if (refDCval >= 65) {
if (Cs < 1959) {
return 8;
}
else if (Cs >= 1959) {
if (ssDCval < 93) {
return 4;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
}
}
else if (posBalance >= 3) {
if (Rs < 6587) {
if (SC < 8522) {
if (refDCval < 82) {
if (ssDCval < 66) {
if (posBalance < 43) {
if (ssDCval < 51) {
if (Cs < 2708) {
if (diffAFD < -1) {
if (Rs < 2891) {
return 8;
}
else if (Rs >= 2891) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Cs >= 2708) {
if (Rs < 2641) {
return 8;
}
else if (Rs >= 2641) {
return 1;
}
}
}
else if (ssDCval >= 51) {
if (Cs < 3306) {
if (RsCsDiff < 2) {
if (SC < 4834) {
return 1;
}
else if (SC >= 4834) {
return 4;
}
}
else if (RsCsDiff >= 2) {
if (Rs < 2911) {
return 8;
}
else if (Rs >= 2911) {
if (Cs < 2826) {
if (negBalance < 26) {
if (gchDC < 1) {
if (diffRsCsdiff < 15) {
if (Rs < 2944) {
if (Cs < 1719) {
return 8;
}
else if (Cs >= 1719) {
return 4;
}
}
else if (Rs >= 2944) {
if (Rs < 2959) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
return 8;
}
}
else if (Rs >= 2959) {
if (Rs < 2983) {
return 4;
}
else if (Rs >= 2983) {
if (Rs < 5168) {
if (Rs < 5107) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
if (negBalance < 14) {
if (MVDiff < 1) {
if (CsDiff < 4) {
if (Rs < 4591) {
return 8;
}
else if (Rs >= 4591) {
return 8;
}
}
else if (CsDiff >= 4) {
if (Rs < 4381) {
return 4;
}
else if (Rs >= 4381) {
return 4;
}
}
}
else if (MVDiff >= 1) {
if (RsDiff < 4) {
if (Rs < 4779) {
if (Rs < 4327) {
return 4;
}
else if (Rs >= 4327) {
return 4;
}
}
else if (Rs >= 4779) {
return 4;
}
}
else if (RsDiff >= 4) {
return 4;
}
}
}
else if (negBalance >= 14) {
if (mu_mv_mag_sq < 113) {
return 8;
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 4893) {
if (Cs < 2485) {
if (posBalance < 40) {
if (diffMVdiffVal < 2) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 3026) {
return 4;
}
else if (Rs >= 3026) {
if (Rs < 4439) {
if (AFD < 4) {
if (diffRsCsdiff < -6) {
if (Rs < 4358) {
return 4;
}
else if (Rs >= 4358) {
return 4;
}
}
else if (diffRsCsdiff >= -6) {
if (mu_mv_mag_sq < 118) {
return 8;
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 3059) {
return 8;
}
else if (Rs >= 3059) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (Cs < 2138) {
return 8;
}
else if (Cs >= 2138) {
return 8;
}
}
}
}
}
}
else if (AFD >= 4) {
return 8;
}
}
else if (Rs >= 4439) {
if (Rs < 4766) {
return 4;
}
else if (Rs >= 4766) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 4513) {
return 4;
}
else if (Rs >= 4513) {
return 4;
}
}
}
else if (posBalance >= 40) {
if (Rs < 3026) {
return 4;
}
else if (Rs >= 3026) {
return 8;
}
}
}
else if (Cs >= 2485) {
return 8;
}
}
else if (Rs >= 4893) {
if (Rs < 4966) {
return 4;
}
else if (Rs >= 4966) {
if (Rs < 5085) {
return 4;
}
else if (Rs >= 5085) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 5107) {
return 8;
}
}
else if (Rs >= 5168) {
return 4;
}
}
}
}
}
else if (diffRsCsdiff >= 15) {
if (Rs < 4713) {
if (Rs < 4524) {
return 4;
}
else if (Rs >= 4524) {
return 4;
}
}
else if (Rs >= 4713) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (Rs < 3031) {
return 4;
}
else if (Rs >= 3031) {
if (diffRsCsdiff < -1) {
if (mu_mv_mag_sq < 117) {
return 4;
}
else if (mu_mv_mag_sq >= 117) {
if (Rs < 4262) {
return 4;
}
else if (Rs >= 4262) {
if (Rs < 4770) {
return 4;
}
else if (Rs >= 4770) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= -1) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 4391) {
return 4;
}
else if (Rs >= 4391) {
if (Rs < 4716) {
return 4;
}
else if (Rs >= 4716) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 26) {
if (posBalance < 37) {
if (Rs < 3722) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (Rs >= 3722) {
return 4;
}
}
else if (posBalance >= 37) {
return 4;
}
}
}
else if (Cs >= 2826) {
return 8;
}
}
}
}
else if (Cs >= 3306) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (Rs < 2009) {
return 4;
}
else if (Rs >= 2009) {
if (negBalance < 19) {
if (Cs < 3287) {
if (Rs < 2923) {
return 4;
}
else if (Rs >= 2923) {
if (Cs < 2662) {
if (diffRsCsdiff < 1) {
if (Cs < 2589) {
return 8;
}
else if (Cs >= 2589) {
return 4;
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 4036) {
return 4;
}
else if (Rs >= 4036) {
if (Rs < 4243) {
return 4;
}
else if (Rs >= 4243) {
return 4;
}
}
}
}
else if (Cs >= 2662) {
return 4;
}
}
}
else if (Cs >= 3287) {
return 8;
}
}
else if (negBalance >= 19) {
if (MVDiff < 3) {
if (Rs < 2537) {
return 1;
}
else if (Rs >= 2537) {
if (negBalance < 20) {
if (Rs < 4604) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 8;
}
}
else if (Rs >= 4604) {
return 4;
}
}
else if (negBalance >= 20) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 2721) {
return 8;
}
else if (Rs >= 2721) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 66) {
if (SC < 5317) {
if (mu_mv_mag_sq < 121) {
if (posBalance < 25) {
return 8;
}
else if (posBalance >= 25) {
if (Rs < 1558) {
return 8;
}
else if (Rs >= 1558) {
if (mcTcor < 96) {
if (gchDC < 2) {
if (refDCval < 70) {
return 4;
}
else if (refDCval >= 70) {
if (SC < 4411) {
return 4;
}
else if (SC >= 4411) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (Rs < 2635) {
return 4;
}
else if (Rs >= 2635) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Rs < 2451) {
return 4;
}
else if (Rs >= 2451) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (Rs < 2160) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Rs >= 2160) {
if (Cs < 1933) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (refDCval < 80) {
return 8;
}
else if (refDCval >= 80) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
return 4;
}
}
}
}
else if (Cs >= 1933) {
if (diffMVdiffVal < -5) {
if (Rs < 2585) {
return 4;
}
else if (Rs >= 2585) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
}
}
}
else if (SC >= 5317) {
if (Cs < 2167) {
if (TSC < 4) {
if (Rs < 3553) {
return 8;
}
else if (Rs >= 3553) {
return 4;
}
}
else if (TSC >= 4) {
return 2;
}
}
else if (Cs >= 2167) {
if (Rs < 2816) {
if (Cs < 2856) {
return 8;
}
else if (Cs >= 2856) {
return 4;
}
}
else if (Rs >= 2816) {
if (Cs < 3713) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
if (Rs < 4389) {
return 8;
}
else if (Rs >= 4389) {
return 4;
}
}
}
else if (Cs >= 3713) {
if (Rs < 3568) {
return 4;
}
else if (Rs >= 3568) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 82) {
if (Cs < 2703) {
if (Rs < 4927) {
if (Cs < 1801) {
if (tcor < 98) {
if (gchDC < 1) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (gchDC >= 1) {
if (Rs < 3159) {
return 8;
}
else if (Rs >= 3159) {
return 2;
}
}
}
else if (tcor >= 98) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (Rs < 3866) {
return 4;
}
else if (Rs >= 3866) {
return 8;
}
}
}
}
else if (Cs >= 1801) {
if (mu_mv_mag_sq < 127) {
if (mcTcor < 97) {
if (Cs < 2615) {
if (ssDCval < 83) {
if (Rs < 3177) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
else if (Rs >= 3177) {
return 2;
}
}
else if (ssDCval >= 83) {
return 8;
}
}
else if (Cs >= 2615) {
if (MVDiff < 1) {
if (refDCval < 87) {
return 8;
}
else if (refDCval >= 87) {
return 4;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (Cs < 2249) {
if (Rs < 3082) {
if (Rs < 2391) {
if (AFD < 3) {
if (ssDCval < 92) {
if (Rs < 2369) {
return 1;
}
else if (Rs >= 2369) {
return 2;
}
}
else if (ssDCval >= 92) {
if (mu_mv_mag_sq < 120) {
if (diffMVdiffVal < -1) {
if (Rs < 2354) {
return 4;
}
else if (Rs >= 2354) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (mu_mv_mag_sq >= 120) {
return 4;
}
}
}
else if (AFD >= 3) {
return 8;
}
}
else if (Rs >= 2391) {
return 8;
}
}
else if (Rs >= 3082) {
if (ssDCval < 91) {
if (diffMVdiffVal < -1) {
if (Rs < 3735) {
return 8;
}
else if (Rs >= 3735) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (ssDCval >= 91) {
if (RsDiff < 6) {
if (Rs < 4512) {
return 8;
}
else if (Rs >= 4512) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
else if (Cs >= 2249) {
if (TSC < 7) {
if (Rs < 2599) {
if (Rs < 2566) {
if (Cs < 2544) {
return 8;
}
else if (Cs >= 2544) {
return 4;
}
}
else if (Rs >= 2566) {
return 1;
}
}
else if (Rs >= 2599) {
if (Cs < 2286) {
if (Cs < 2280) {
return 8;
}
else if (Cs >= 2280) {
return 2;
}
}
else if (Cs >= 2286) {
return 8;
}
}
}
else if (TSC >= 7) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (CsDiff < 3) {
if (SC < 4472) {
if (Rs < 2255) {
return 8;
}
else if (Rs >= 2255) {
return 4;
}
}
else if (SC >= 4472) {
return 8;
}
}
else if (CsDiff >= 3) {
if (Cs < 1945) {
return 4;
}
else if (Cs >= 1945) {
if (Rs < 4086) {
return 8;
}
else if (Rs >= 4086) {
return 1;
}
}
}
}
}
}
else if (Rs >= 4927) {
return 4;
}
}
else if (Cs >= 2703) {
if (Cs < 3641) {
if (ssDCval < 90) {
if (Rs < 3405) {
if (SC < 5751) {
return 4;
}
else if (SC >= 5751) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
}
else if (Rs >= 3405) {
if (Cs < 3538) {
if (Cs < 2736) {
return 8;
}
else if (Cs >= 2736) {
return 4;
}
}
else if (Cs >= 3538) {
return 8;
}
}
}
else if (ssDCval >= 90) {
if (mu_mv_mag_sq < 123) {
if (Cs < 3290) {
return 4;
}
else if (Cs >= 3290) {
if (diffRsCsdiff < 14) {
return 4;
}
else if (diffRsCsdiff >= 14) {
if (Rs < 3923) {
return 8;
}
else if (Rs >= 3923) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (SC < 7201) {
if (ssDCval < 94) {
if (SC < 6649) {
if (refDCval < 93) {
return 8;
}
else if (refDCval >= 93) {
return 4;
}
}
else if (SC >= 6649) {
if (Rs < 3434) {
return 8;
}
else if (Rs >= 3434) {
return 4;
}
}
}
else if (ssDCval >= 94) {
if (Cs < 3026) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
return 4;
}
}
else if (Cs >= 3026) {
return 8;
}
}
}
else if (SC >= 7201) {
return 4;
}
}
}
}
else if (Cs >= 3641) {
if (refDCval < 95) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 8;
}
}
}
else if (refDCval >= 95) {
if (Rs < 3545) {
return 4;
}
else if (Rs >= 3545) {
return 8;
}
}
}
}
}
}
else if (SC >= 8522) {
if (negBalance < 31) {
if (RsDiff < 6) {
if (Rs < 4456) {
if (negBalance < 27) {
return 1;
}
else if (negBalance >= 27) {
return 8;
}
}
else if (Rs >= 4456) {
return 8;
}
}
else if (RsDiff >= 6) {
if (Rs < 5490) {
if (Rs < 4668) {
return 1;
}
else if (Rs >= 4668) {
return 8;
}
}
else if (Rs >= 5490) {
return 2;
}
}
}
else if (negBalance >= 31) {
if (Cs < 3747) {
if (Rs < 5208) {
return 4;
}
else if (Rs >= 5208) {
return 8;
}
}
else if (Cs >= 3747) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 5982) {
return 8;
}
else if (Rs >= 5982) {
return 2;
}
}
}
}
}
}
else if (Rs >= 6587) {
if (RsDiff < 43) {
if (ssDCval < 93) {
return 4;
}
else if (ssDCval >= 93) {
return 8;
}
}
else if (RsDiff >= 43) {
return 8;
}
}
}
}
}
else if (ssDCval >= 98) {
if (ssDCval < 129) {
if (Rs < 1546) {
if (posBalance < 54) {
if (Cs < 2184) {
return 8;
}
else if (Cs >= 2184) {
return 2;
}
}
else if (posBalance >= 54) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 4;
}
}
}
else if (Rs >= 1546) {
if (ssDCval < 105) {
if (Cs < 1992) {
if (Rs < 3492) {
return 8;
}
else if (Rs >= 3492) {
if (AFD < 5) {
if (Cs < 1926) {
return 4;
}
else if (Cs >= 1926) {
return 2;
}
}
else if (AFD >= 5) {
if (Rs < 4416) {
return 2;
}
else if (Rs >= 4416) {
return 1;
}
}
}
}
else if (Cs >= 1992) {
if (AFD < 9) {
if (ssDCval < 103) {
if (Rs < 2065) {
return 4;
}
else if (Rs >= 2065) {
if (negBalance < 48) {
if (Cs < 4092) {
return 8;
}
else if (Cs >= 4092) {
if (Rs < 3283) {
return 4;
}
else if (Rs >= 3283) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (mcTcor < 96) {
return 4;
}
else if (mcTcor >= 96) {
if (posBalance < 34) {
return 4;
}
else if (posBalance >= 34) {
return 8;
}
}
}
}
}
else if (ssDCval >= 103) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
if (Rs < 4925) {
if (Cs < 2150) {
return 1;
}
else if (Cs >= 2150) {
if (Rs < 4740) {
return 8;
}
else if (Rs >= 4740) {
return 1;
}
}
}
else if (Rs >= 4925) {
if (Cs < 2934) {
return 2;
}
else if (Cs >= 2934) {
return 8;
}
}
}
}
}
else if (AFD >= 9) {
if (mcTcor < 93) {
if (Cs < 3006) {
if (Rs < 4947) {
return 1;
}
else if (Rs >= 4947) {
return 2;
}
}
else if (Cs >= 3006) {
if (Rs < 3037) {
return 4;
}
else if (Rs >= 3037) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (RsDiff < 8) {
if (Cs < 2496) {
return 8;
}
else if (Cs >= 2496) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
}
else if (ssDCval >= 105) {
if (diffRsCsdiff < -978) {
return 1;
}
else if (diffRsCsdiff >= -978) {
if (mu_mv_mag_sq < 122) {
if (posBalance < 21) {
if (refDCval < 117) {
return 8;
}
else if (refDCval >= 117) {
if (Rs < 2769) {
if (Cs < 2082) {
return 1;
}
else if (Cs >= 2082) {
if (negBalance < 34) {
if (Cs < 2913) {
return 4;
}
else if (Cs >= 2913) {
if (Rs < 2728) {
return 8;
}
else if (Rs >= 2728) {
if (Cs < 2939) {
return 4;
}
else if (Cs >= 2939) {
return 8;
}
}
}
}
else if (negBalance >= 34) {
return 8;
}
}
}
else if (Rs >= 2769) {
return 2;
}
}
}
else if (posBalance >= 21) {
if (diffRsCsdiff < -14) {
if (ssDCval < 121) {
if (gchDC < 1) {
if (Cs < 1664) {
return 1;
}
else if (Cs >= 1664) {
return 8;
}
}
else if (gchDC >= 1) {
if (Cs < 3322) {
if (Cs < 2845) {
return 8;
}
else if (Cs >= 2845) {
if (Rs < 4676) {
return 2;
}
else if (Rs >= 4676) {
return 8;
}
}
}
else if (Cs >= 3322) {
return 8;
}
}
}
else if (ssDCval >= 121) {
if (diffTSC < -1) {
if (Rs < 3291) {
return 1;
}
else if (Rs >= 3291) {
return 4;
}
}
else if (diffTSC >= -1) {
if (Rs < 6343) {
return 8;
}
else if (Rs >= 6343) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= -14) {
if (posBalance < 62) {
if (Rs < 6504) {
if (Cs < 1650) {
if (Rs < 3727) {
return 8;
}
else if (Rs >= 3727) {
return 2;
}
}
else if (Cs >= 1650) {
if (posBalance < 32) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (posBalance >= 32) {
return 8;
}
}
}
else if (Rs >= 6504) {
if (Rs < 6687) {
return 4;
}
else if (Rs >= 6687) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (Cs < 1858) {
if (Rs < 2399) {
return 8;
}
else if (Rs >= 2399) {
return 1;
}
}
else if (Cs >= 1858) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (Cs < 2794) {
return 8;
}
else if (Cs >= 2794) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Cs < 2264) {
if (Rs < 4942) {
if (Rs < 1908) {
if (SC < 3857) {
if (Rs < 1583) {
if (refDCval < 113) {
return 2;
}
else if (refDCval >= 113) {
return 8;
}
}
else if (Rs >= 1583) {
return 8;
}
}
else if (SC >= 3857) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (Rs >= 1908) {
if (MVDiff < 55) {
if (diffAFD < -2) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (diffRsCsdiff < -40) {
if (ssDCval < 116) {
return 4;
}
else if (ssDCval >= 116) {
return 1;
}
}
else if (diffRsCsdiff >= -40) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (MVDiff >= 55) {
if (Rs < 2790) {
return 4;
}
else if (Rs >= 2790) {
return 1;
}
}
}
}
else if (Rs >= 4942) {
if (SC < 7394) {
return 2;
}
else if (SC >= 7394) {
return 8;
}
}
}
else if (Cs >= 2264) {
if (diffAFD < -2) {
if (ssDCval < 123) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Rs < 5276) {
return 2;
}
else if (Rs >= 5276) {
return 8;
}
}
}
else if (ssDCval >= 123) {
if (ssDCval < 124) {
return 1;
}
else if (ssDCval >= 124) {
return 8;
}
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 129) {
if (mcTcor < 96) {
if (negBalance < 42) {
if (ssDCval < 200) {
if (posBalance < 64) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (ssDCval < 139) {
return 1;
}
else if (ssDCval >= 139) {
return 8;
}
}
}
else if (posBalance >= 64) {
if (Rs < 1757) {
return 8;
}
else if (Rs >= 1757) {
return 1;
}
}
}
else if (ssDCval >= 200) {
return 4;
}
}
else if (negBalance >= 42) {
if (RsDiff < 11) {
if (Rs < 2507) {
return 8;
}
else if (Rs >= 2507) {
return 1;
}
}
else if (RsDiff >= 11) {
if (Cs < 6648) {
if (Rs < 6924) {
return 2;
}
else if (Rs >= 6924) {
return 4;
}
}
else if (Cs >= 6648) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (ssDCval < 131) {
if (negBalance < 34) {
if (SC < 4322) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
return 4;
}
}
else if (SC >= 4322) {
if (Rs < 4003) {
return 1;
}
else if (Rs >= 4003) {
return 2;
}
}
}
else if (negBalance >= 34) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
else if (ssDCval >= 131) {
if (ssDCval < 144) {
if (Rs < 4620) {
if (Cs < 3887) {
if (SC < 4979) {
return 8;
}
else if (SC >= 4979) {
if (SC < 4982) {
return 1;
}
else if (SC >= 4982) {
return 8;
}
}
}
else if (Cs >= 3887) {
if (Rs < 3973) {
return 1;
}
else if (Rs >= 3973) {
return 8;
}
}
}
else if (Rs >= 4620) {
if (Cs < 3481) {
if (Rs < 4874) {
return 1;
}
else if (Rs >= 4874) {
return 2;
}
}
else if (Cs >= 3481) {
return 8;
}
}
}
else if (ssDCval >= 144) {
if (ssDCval < 200) {
return 8;
}
else if (ssDCval >= 200) {
if (Rs < 2635) {
return 8;
}
else if (Rs >= 2635) {
return 4;
}
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelectA(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1325) {
if (Rs < 2354) {
if (Cs < 505) {
if (negBalance < 14) {
if (MVDiff < 7) {
if (SC < 793) {
if (refDCval < 47) {
if (SC < 279) {
if (negBalance < 8) {
if (Rs < 2) {
if (ssDCval < 22) {
if (ssDCval < 10) {
return 4;
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (Rs >= 2) {
return 8;
}
}
else if (negBalance >= 8) {
if (Rs < 99) {
if (ssDCval < 39) {
if (ssDCval < 24) {
if (Rs < 66) {
return 8;
}
else if (Rs >= 66) {
return 1;
}
}
else if (ssDCval >= 24) {
if (posBalance < 35) {
if (Rs < 86) {
return 4;
}
else if (Rs >= 86) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
return 8;
}
}
}
else if (posBalance >= 35) {
if (TSCindex < 2) {
return 1;
}
else if (TSCindex >= 2) {
if (Rs < 29) {
return 4;
}
else if (Rs >= 29) {
return 8;
}
}
}
}
}
else if (ssDCval >= 39) {
return 8;
}
}
else if (Rs >= 99) {
if (TSC < 1) {
if (posBalance < 35) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
return 1;
}
}
else if (posBalance >= 35) {
return 1;
}
}
else if (TSC >= 1) {
if (mu_mv_mag_sq < 96) {
return 8;
}
else if (mu_mv_mag_sq >= 96) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 279) {
if (Rs < 204) {
if (refDCval < 35) {
if (ssDCval < 27) {
return 1;
}
else if (ssDCval >= 27) {
return 8;
}
}
else if (refDCval >= 35) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (Cs < 250) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 1;
}
}
else if (Cs >= 250) {
return 1;
}
}
}
}
else if (Rs >= 204) {
if (refDCval < 38) {
if (mu_mv_mag_sq < 128) {
if (MVDiff < 3) {
if (SC < 541) {
if (Rs < 243) {
return 4;
}
else if (Rs >= 243) {
return 8;
}
}
else if (SC >= 541) {
if (Rs < 439) {
return 4;
}
else if (Rs >= 439) {
return 8;
}
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 24) {
return 8;
}
else if (ssDCval >= 24) {
if (Rs < 294) {
return 8;
}
else if (Rs >= 294) {
if (Cs < 259) {
return 8;
}
else if (Cs >= 259) {
return 4;
}
}
}
}
}
else if (refDCval >= 38) {
if (SC < 514) {
return 4;
}
else if (SC >= 514) {
if (ssDCval < 40) {
if (Cs < 423) {
return 4;
}
else if (Cs >= 423) {
return 8;
}
}
else if (ssDCval >= 40) {
if (Rs < 378) {
return 1;
}
else if (Rs >= 378) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 47) {
if (ssDCval < 82) {
if (SC < 84) {
return 2;
}
else if (SC >= 84) {
if (ssDCval < 61) {
if (Cs < 283) {
if (MVDiff < 1) {
if (negBalance < 6) {
if (Rs < 272) {
return 8;
}
else if (Rs >= 272) {
return 4;
}
}
else if (negBalance >= 6) {
return 8;
}
}
else if (MVDiff >= 1) {
if (refDCval < 54) {
if (SC < 389) {
return 8;
}
else if (SC >= 389) {
return 4;
}
}
else if (refDCval >= 54) {
if (tcor < 98) {
if (SC < 401) {
return 1;
}
else if (SC >= 401) {
return 8;
}
}
else if (tcor >= 98) {
if (Cs < 120) {
return 8;
}
else if (Cs >= 120) {
return 4;
}
}
}
}
}
else if (Cs >= 283) {
if (Rs < 250) {
return 4;
}
else if (Rs >= 250) {
if (Rs < 323) {
return 1;
}
else if (Rs >= 323) {
return 8;
}
}
}
}
else if (ssDCval >= 61) {
if (Rs < 331) {
if (Rs < 78) {
return 4;
}
else if (Rs >= 78) {
if (AFD < 2) {
if (Cs < 284) {
return 2;
}
else if (Cs >= 284) {
return 4;
}
}
else if (AFD >= 2) {
if (Rs < 201) {
return 1;
}
else if (Rs >= 201) {
return 8;
}
}
}
}
else if (Rs >= 331) {
if (SC < 646) {
if (Rs < 453) {
return 4;
}
else if (Rs >= 453) {
return 8;
}
}
else if (SC >= 646) {
return 8;
}
}
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
if (Cs < 258) {
if (Rs < 159) {
return 4;
}
else if (Rs >= 159) {
return 8;
}
}
else if (Cs >= 258) {
return 4;
}
}
}
}
}
}
else if (SC >= 793) {
if (mu_mv_mag_sq < 128) {
if (Cs < 268) {
if (ssDCval < 39) {
if (Rs < 555) {
return 8;
}
else if (Rs >= 555) {
return 4;
}
}
else if (ssDCval >= 39) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (mu_mv_mag_sq < 125) {
if (Rs < 1664) {
return 8;
}
else if (Rs >= 1664) {
return 1;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1015) {
return 8;
}
else if (Rs >= 1015) {
return 2;
}
}
}
}
}
else if (Cs >= 268) {
if (Cs < 405) {
if (refDCval < 56) {
if (Rs < 573) {
return 1;
}
else if (Rs >= 573) {
if (negBalance < 13) {
return 8;
}
else if (negBalance >= 13) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
return 1;
}
}
}
}
else if (refDCval >= 56) {
if (ssDCval < 67) {
if (Rs < 552) {
if (Cs < 390) {
if (Rs < 490) {
return 8;
}
else if (Rs >= 490) {
return 1;
}
}
else if (Cs >= 390) {
return 8;
}
}
else if (Rs >= 552) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
return 1;
}
}
}
else if (ssDCval >= 67) {
if (mu_mv_mag_sq < 104) {
if (Rs < 1971) {
return 1;
}
else if (Rs >= 1971) {
return 8;
}
}
else if (mu_mv_mag_sq >= 104) {
if (Cs < 404) {
return 8;
}
else if (Cs >= 404) {
if (Rs < 441) {
return 8;
}
else if (Rs >= 441) {
return 1;
}
}
}
}
}
}
else if (Cs >= 405) {
if (Rs < 1143) {
if (RsCsDiff < 1) {
if (Cs < 419) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (Rs < 510) {
return 8;
}
else if (Rs >= 510) {
return 4;
}
}
}
else if (Cs >= 419) {
if (ssDCval < 141) {
if (Rs < 575) {
return 8;
}
else if (Rs >= 575) {
if (Rs < 581) {
return 1;
}
else if (Rs >= 581) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (Rs < 796) {
return 1;
}
else if (Rs >= 796) {
return 8;
}
}
}
}
}
else if (ssDCval >= 141) {
if (Rs < 514) {
return 1;
}
else if (Rs >= 514) {
return 8;
}
}
}
}
else if (RsCsDiff >= 1) {
if (Rs < 511) {
return 1;
}
else if (Rs >= 511) {
if (Cs < 454) {
return 8;
}
else if (Cs >= 454) {
return 1;
}
}
}
}
else if (Rs >= 1143) {
if (posBalance < 44) {
if (tcor < 100) {
if (mu_mv_mag_sq < 119) {
return 2;
}
else if (mu_mv_mag_sq >= 119) {
return 1;
}
}
else if (tcor >= 100) {
if (MVDiff < 1) {
if (Rs < 1368) {
return 1;
}
else if (Rs >= 1368) {
if (Rs < 1442) {
return 2;
}
else if (Rs >= 1442) {
return 1;
}
}
}
else if (MVDiff >= 1) {
return 2;
}
}
}
else if (posBalance >= 44) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (Rs < 1830) {
return 1;
}
else if (Rs >= 1830) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (negBalance < 12) {
if (ssDCval < 135) {
if (Cs < 315) {
if (Rs < 1041) {
return 1;
}
else if (Rs >= 1041) {
return 8;
}
}
else if (Cs >= 315) {
if (Cs < 474) {
return 8;
}
else if (Cs >= 474) {
if (Rs < 989) {
return 8;
}
else if (Rs >= 989) {
return 2;
}
}
}
}
else if (ssDCval >= 135) {
return 1;
}
}
else if (negBalance >= 12) {
if (ssDCval < 89) {
if (Rs < 909) {
if (Rs < 723) {
return 4;
}
else if (Rs >= 723) {
return 1;
}
}
else if (Rs >= 909) {
if (Rs < 1077) {
return 8;
}
else if (Rs >= 1077) {
return 2;
}
}
}
else if (ssDCval >= 89) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (Rs < 1116) {
return 4;
}
else if (Rs >= 1116) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (tcor < 94) {
if (negBalance < 6) {
if (MVDiff < 177) {
if (posBalance < 93) {
if (mu_mv_mag_sq < 104) {
if (Rs < 10) {
if (AFD < 1) {
if (tcor < 23) {
return 4;
}
else if (tcor >= 23) {
return 8;
}
}
else if (AFD >= 1) {
if (mcTcor < 69) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1) {
return 4;
}
else if (Rs >= 1) {
return 2;
}
}
}
else if (mcTcor >= 69) {
return 2;
}
}
}
else if (Rs >= 10) {
if (ssDCval < 27) {
if (Rs < 166) {
return 2;
}
else if (Rs >= 166) {
return 1;
}
}
else if (ssDCval >= 27) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (diffMVdiffVal < 1) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
else if (posBalance >= 93) {
if (gchDC < 10) {
if (posBalance < 95) {
if (Cs < 3) {
if (MVDiff < 126) {
return 4;
}
else if (MVDiff >= 126) {
return 8;
}
}
else if (Cs >= 3) {
return 8;
}
}
else if (posBalance >= 95) {
if (Cs < 53) {
if (TSC < 3) {
if (Rs < 44) {
return 2;
}
else if (Rs >= 44) {
return 4;
}
}
else if (TSC >= 3) {
if (gchDC < 6) {
return 8;
}
else if (gchDC >= 6) {
return 2;
}
}
}
else if (Cs >= 53) {
if (posBalance < 105) {
return 8;
}
else if (posBalance >= 105) {
return 4;
}
}
}
}
else if (gchDC >= 10) {
if (mu_mv_mag_sq < 15) {
if (Rs < 154) {
return 2;
}
else if (Rs >= 154) {
return 4;
}
}
else if (mu_mv_mag_sq >= 15) {
if (ssDCval < 26) {
return 1;
}
else if (ssDCval >= 26) {
return 8;
}
}
}
}
}
else if (MVDiff >= 177) {
if (AFD < 8) {
if (MVDiff < 443) {
return 8;
}
else if (MVDiff >= 443) {
if (Rs < 6) {
return 2;
}
else if (Rs >= 6) {
if (ssDCval < 11) {
return 1;
}
else if (ssDCval >= 11) {
return 4;
}
}
}
}
else if (AFD >= 8) {
if (CsDiff < 3) {
if (diffMVdiffVal < -901) {
return 8;
}
else if (diffMVdiffVal >= -901) {
return 1;
}
}
else if (CsDiff >= 3) {
return 4;
}
}
}
}
else if (negBalance >= 6) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
if (mu_mv_mag_sq < 36) {
if (mcTcor < 96) {
if (TSC < 9) {
if (mu_mv_mag_sq < 17) {
if (posBalance < 68) {
if (Rs < 210) {
return 1;
}
else if (Rs >= 210) {
return 2;
}
}
else if (posBalance >= 68) {
return 2;
}
}
else if (mu_mv_mag_sq >= 17) {
if (diffMVdiffVal < 20) {
if (Rs < 479) {
return 2;
}
else if (Rs >= 479) {
return 8;
}
}
else if (diffMVdiffVal >= 20) {
return 8;
}
}
}
else if (TSC >= 9) {
return 1;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (mu_mv_mag_sq >= 36) {
if (SC < 151) {
if (posBalance < 66) {
return 1;
}
else if (posBalance >= 66) {
if (Rs < 88) {
return 8;
}
else if (Rs >= 88) {
return 1;
}
}
}
else if (SC >= 151) {
if (mu_mv_mag_sq < 50) {
if (tcor < 88) {
if (diffMVdiffVal < -3) {
if (tcor < 70) {
if (Rs < 470) {
return 1;
}
else if (Rs >= 470) {
return 4;
}
}
else if (tcor >= 70) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (tcor >= 88) {
return 4;
}
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 130) {
return 8;
}
else if (Rs >= 130) {
if (RsCsDiff < 26) {
if (Cs < 474) {
if (ssDCval < 21) {
return 8;
}
else if (ssDCval >= 21) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
return 4;
}
}
}
else if (Cs >= 474) {
return 8;
}
}
else if (RsCsDiff >= 26) {
if (Cs < 444) {
return 8;
}
else if (Cs >= 444) {
return 4;
}
}
}
}
}
}
}
}
}
else if (tcor >= 94) {
if (negBalance < 6) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
return 8;
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (ssDCval < 27) {
return 4;
}
else if (ssDCval >= 27) {
if (diffRsCsdiff < -1) {
if (Rs < 925) {
if (posBalance < 3) {
return 4;
}
else if (posBalance >= 3) {
if (diffRsCsdiff < -11) {
if (diffMVdiffVal < -4) {
if (diffTSC < -4) {
return 2;
}
else if (diffTSC >= -4) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (diffRsCsdiff >= -11) {
if (MVDiff < 41) {
if (ssDCval < 88) {
if (Rs < 465) {
return 8;
}
else if (Rs >= 465) {
return 4;
}
}
else if (ssDCval >= 88) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 4;
}
}
}
else if (MVDiff >= 41) {
if (MVDiff < 76) {
if (Rs < 376) {
return 4;
}
else if (Rs >= 376) {
return 2;
}
}
else if (MVDiff >= 76) {
return 4;
}
}
}
}
}
else if (Rs >= 925) {
if (ssDCval < 130) {
if (Rs < 1507) {
return 2;
}
else if (Rs >= 1507) {
return 8;
}
}
else if (ssDCval >= 130) {
return 1;
}
}
}
else if (diffRsCsdiff >= -1) {
if (mcTcor < 98) {
if (AFD < 6) {
if (MVDiff < 10) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (MVDiff >= 10) {
if (Rs < 43) {
return 2;
}
else if (Rs >= 43) {
return 4;
}
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (mcTcor >= 98) {
if (mcTcor < 101) {
if (ssDCval < 130) {
if (Rs < 10) {
return 2;
}
else if (Rs >= 10) {
if (Rs < 827) {
return 8;
}
else if (Rs >= 827) {
if (Rs < 1040) {
return 4;
}
else if (Rs >= 1040) {
return 8;
}
}
}
}
else if (ssDCval >= 130) {
if (Rs < 185) {
return 8;
}
else if (Rs >= 185) {
return 4;
}
}
}
else if (mcTcor >= 101) {
if (Rs < 145) {
return 2;
}
else if (Rs >= 145) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 6) {
if (SC < 876) {
if (Cs < 111) {
if (posBalance < 8) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (posBalance >= 8) {
if (SC < 98) {
return 1;
}
else if (SC >= 98) {
if (mu_mv_mag_sq < 89) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
if (SC < 163) {
return 4;
}
else if (SC >= 163) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
return 8;
}
}
}
}
}
else if (Cs >= 111) {
if (mu_mv_mag_sq < 66) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 66) {
if (mu_mv_mag_sq < 116) {
if (Rs < 92) {
return 4;
}
else if (Rs >= 92) {
if (posBalance < 52) {
if (Cs < 235) {
if (diffMVdiffVal < 2) {
if (SC < 284) {
return 8;
}
else if (SC >= 284) {
if (ssDCval < 68) {
if (Rs < 186) {
return 1;
}
else if (Rs >= 186) {
return 4;
}
}
else if (ssDCval >= 68) {
return 4;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 156) {
return 1;
}
else if (Rs >= 156) {
return 8;
}
}
}
else if (Cs >= 235) {
if (ssDCval < 133) {
return 4;
}
else if (ssDCval >= 133) {
if (ssDCval < 161) {
return 1;
}
else if (ssDCval >= 161) {
return 4;
}
}
}
}
else if (posBalance >= 52) {
if (Rs < 472) {
return 1;
}
else if (Rs >= 472) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
}
}
else if (SC >= 876) {
if (Cs < 397) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 1;
}
}
else if (Cs >= 397) {
if (CsDiff < 1) {
if (ssDCval < 41) {
return 4;
}
else if (ssDCval >= 41) {
return 8;
}
}
else if (CsDiff >= 1) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 35) {
if (mcTcor < 84) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
if (tcor < 57) {
if (negBalance < 45) {
if (Cs < 174) {
if (posBalance < 69) {
if (negBalance < 20) {
if (Rs < 109) {
return 2;
}
else if (Rs >= 109) {
if (Rs < 121) {
return 1;
}
else if (Rs >= 121) {
return 8;
}
}
}
else if (negBalance >= 20) {
return 1;
}
}
else if (posBalance >= 69) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (Cs >= 174) {
if (negBalance < 26) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (negBalance >= 26) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (Rs < 223) {
if (Rs < 127) {
return 4;
}
else if (Rs >= 127) {
return 8;
}
}
else if (Rs >= 223) {
return 1;
}
}
}
}
}
else if (negBalance >= 45) {
if (tcor < 10) {
if (Rs < 187) {
return 4;
}
else if (Rs >= 187) {
return 1;
}
}
else if (tcor >= 10) {
if (Rs < 521) {
if (tcor < 54) {
return 8;
}
else if (tcor >= 54) {
if (Rs < 197) {
return 8;
}
else if (Rs >= 197) {
return 1;
}
}
}
else if (Rs >= 521) {
return 1;
}
}
}
}
else if (tcor >= 57) {
if (ssDCval < 23) {
if (Rs < 360) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 1;
}
}
else if (Rs >= 360) {
return 4;
}
}
else if (ssDCval >= 23) {
if (mcTcor < 71) {
if (diffTSC < -1) {
if (Rs < 22) {
return 4;
}
else if (Rs >= 22) {
return 2;
}
}
else if (diffTSC >= -1) {
if (Rs < 555) {
return 1;
}
else if (Rs >= 555) {
return 4;
}
}
}
else if (mcTcor >= 71) {
if (Cs < 209) {
if (tcor < 65) {
if (CsDiff < 2) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (tcor >= 65) {
return 1;
}
}
else if (Cs >= 209) {
if (Rs < 656) {
if (posBalance < 44) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (posBalance >= 44) {
if (mu_mv_mag_sq < 47) {
return 2;
}
else if (mu_mv_mag_sq >= 47) {
return 1;
}
}
}
else if (Rs >= 656) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 84) {
if (posBalance < 25) {
if (AFD < 2) {
if (negBalance < 22) {
if (tcor < 102) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 279) {
return 8;
}
else if (Rs >= 279) {
return 4;
}
}
}
else if (tcor >= 102) {
if (Rs < 209) {
return 8;
}
else if (Rs >= 209) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (posBalance < 15) {
if (ssDCval < 24) {
return 8;
}
else if (ssDCval >= 24) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (posBalance >= 15) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 1;
}
}
}
}
else if (AFD >= 2) {
if (negBalance < 58) {
if (mu_mv_mag_sq < 15) {
if (Rs < 68) {
return 1;
}
else if (Rs >= 68) {
return 8;
}
}
else if (mu_mv_mag_sq >= 15) {
return 1;
}
}
else if (negBalance >= 58) {
if (Rs < 19) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
if (TSCindex < 2) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (TSCindex >= 2) {
return 8;
}
}
}
else if (Rs >= 19) {
if (MVDiff < 56) {
return 1;
}
else if (MVDiff >= 56) {
return 4;
}
}
}
}
}
else if (posBalance >= 25) {
if (ssDCval < 30) {
if (MVDiff < 46) {
if (posBalance < 71) {
if (Cs < 264) {
if (mu_mv_mag_sq < 40) {
if (posBalance < 42) {
if (mu_mv_mag_sq < 7) {
return 1;
}
else if (mu_mv_mag_sq >= 7) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
if (mcTcor < 91) {
if (Rs < 82) {
return 1;
}
else if (Rs >= 82) {
return 4;
}
}
else if (mcTcor >= 91) {
return 1;
}
}
}
}
else if (posBalance >= 42) {
if (negBalance < 17) {
return 8;
}
else if (negBalance >= 17) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (TSCindex < 3) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 91) {
if (tcor < 71) {
if (mcTcor < 93) {
return 1;
}
else if (mcTcor >= 93) {
return 8;
}
}
else if (tcor >= 71) {
if (diffMVdiffVal < -10) {
if (Rs < 311) {
return 4;
}
else if (Rs >= 311) {
return 1;
}
}
else if (diffMVdiffVal >= -10) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 91) {
return 8;
}
}
}
}
else if (Cs >= 264) {
if (MVDiff < 23) {
if (ssDCval < 22) {
return 4;
}
else if (ssDCval >= 22) {
return 1;
}
}
else if (MVDiff >= 23) {
if (Rs < 211) {
return 1;
}
else if (Rs >= 211) {
return 4;
}
}
}
}
else if (posBalance >= 71) {
return 2;
}
}
else if (MVDiff >= 46) {
if (negBalance < 26) {
if (Rs < 59) {
return 1;
}
else if (Rs >= 59) {
return 8;
}
}
else if (negBalance >= 26) {
if (Cs < 221) {
return 1;
}
else if (Cs >= 221) {
return 2;
}
}
}
}
else if (ssDCval >= 30) {
if (posBalance < 41) {
if (mu_mv_mag_sq < 15) {
if (MVDiff < 8) {
if (Rs < 72) {
return 1;
}
else if (Rs >= 72) {
if (ssDCval < 32) {
return 4;
}
else if (ssDCval >= 32) {
if (Rs < 190) {
return 8;
}
else if (Rs >= 190) {
return 4;
}
}
}
}
else if (MVDiff >= 8) {
if (Rs < 123) {
return 8;
}
else if (Rs >= 123) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (Cs < 260) {
if (mcTcor < 97) {
if (AFD < 6) {
if (mcTcor < 93) {
if (Cs < 225) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (Cs >= 225) {
return 1;
}
}
else if (mcTcor >= 93) {
if (TSCindex < 3) {
if (SC < 181) {
if (Rs < 81) {
return 1;
}
else if (Rs >= 81) {
return 4;
}
}
else if (SC >= 181) {
if (posBalance < 31) {
if (Rs < 247) {
return 8;
}
else if (Rs >= 247) {
return 2;
}
}
else if (posBalance >= 31) {
if (negBalance < 16) {
return 8;
}
else if (negBalance >= 16) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
}
else if (TSCindex >= 3) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (Rs < 324) {
return 1;
}
else if (Rs >= 324) {
return 4;
}
}
}
}
}
else if (AFD >= 6) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
if (Cs < 183) {
return 4;
}
else if (Cs >= 183) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (SC < 583) {
if (posBalance < 31) {
if (negBalance < 22) {
if (Rs < 191) {
return 8;
}
else if (Rs >= 191) {
return 2;
}
}
else if (negBalance >= 22) {
return 1;
}
}
else if (posBalance >= 31) {
return 1;
}
}
else if (SC >= 583) {
return 4;
}
}
}
else if (Cs >= 260) {
if (Cs < 472) {
return 1;
}
else if (Cs >= 472) {
if (Rs < 449) {
return 4;
}
else if (Rs >= 449) {
if (Rs < 772) {
return 8;
}
else if (Rs >= 772) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 41) {
if (SC < 418) {
if (refDCval < 33) {
if (posBalance < 72) {
if (mcTcor < 88) {
if (negBalance < 41) {
if (SC < 396) {
return 1;
}
else if (SC >= 396) {
return 8;
}
}
else if (negBalance >= 41) {
return 8;
}
}
else if (mcTcor >= 88) {
if (tcor < 72) {
if (tcor < 67) {
return 1;
}
else if (tcor >= 67) {
if (Rs < 91) {
return 1;
}
else if (Rs >= 91) {
return 4;
}
}
}
else if (tcor >= 72) {
return 1;
}
}
}
else if (posBalance >= 72) {
return 2;
}
}
else if (refDCval >= 33) {
if (mcTcor < 94) {
if (AFD < 7) {
if (negBalance < 36) {
if (Cs < 235) {
return 4;
}
else if (Cs >= 235) {
return 8;
}
}
else if (negBalance >= 36) {
if (Rs < 55) {
return 1;
}
else if (Rs >= 55) {
return 2;
}
}
}
else if (AFD >= 7) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 164) {
return 4;
}
else if (Rs >= 164) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (Cs < 13) {
return 2;
}
else if (Cs >= 13) {
if (Cs < 195) {
return 1;
}
else if (Cs >= 195) {
return 8;
}
}
}
}
}
else if (SC >= 418) {
if (posBalance < 62) {
if (Cs < 501) {
if (negBalance < 19) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
return 8;
}
}
else if (negBalance >= 19) {
return 1;
}
}
else if (Cs >= 501) {
return 8;
}
}
else if (posBalance >= 62) {
if (Rs < 431) {
return 8;
}
else if (Rs >= 431) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 35) {
if (Rs < 307) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 43) {
if (refDCval < 133) {
if (Cs < 67) {
if (mcTcor < 79) {
if (posBalance < 50) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
return 8;
}
}
else if (posBalance >= 50) {
if (tcor < 10) {
return 1;
}
else if (tcor >= 10) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
return 4;
}
}
}
}
else if (mcTcor >= 79) {
if (tcor < 100) {
if (Rs < 155) {
if (Rs < 41) {
if (refDCval < 66) {
if (mu_mv_mag_sq < 15) {
if (Cs < 17) {
return 8;
}
else if (Cs >= 17) {
return 4;
}
}
else if (mu_mv_mag_sq >= 15) {
if (Rs < 11) {
return 2;
}
else if (Rs >= 11) {
return 1;
}
}
}
else if (refDCval >= 66) {
return 1;
}
}
else if (Rs >= 41) {
if (SC < 107) {
if (posBalance < 52) {
if (Cs < 44) {
return 4;
}
else if (Cs >= 44) {
return 1;
}
}
else if (posBalance >= 52) {
if (Cs < 47) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
return 1;
}
}
else if (Cs >= 47) {
return 1;
}
}
}
else if (SC >= 107) {
if (posBalance < 62) {
if (AFD < 8) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
return 1;
}
}
else if (AFD >= 8) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
if (Cs < 18) {
return 8;
}
else if (Cs >= 18) {
return 2;
}
}
}
}
else if (posBalance >= 62) {
return 1;
}
}
}
}
else if (Rs >= 155) {
if (diffMVdiffVal < -30) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 252) {
return 8;
}
else if (Rs >= 252) {
return 2;
}
}
}
else if (diffMVdiffVal >= -30) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
else if (tcor >= 100) {
return 8;
}
}
}
else if (Cs >= 67) {
if (mcTcor < 90) {
if (MVDiff < 29) {
if (TSC < 5) {
if (SC < 408) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Cs < 226) {
if (Rs < 170) {
return 4;
}
else if (Rs >= 170) {
return 2;
}
}
else if (Cs >= 226) {
return 8;
}
}
}
else if (SC >= 408) {
if (negBalance < 45) {
if (ssDCval < 66) {
if (refDCval < 36) {
return 8;
}
else if (refDCval >= 36) {
if (SC < 539) {
return 2;
}
else if (SC >= 539) {
return 8;
}
}
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (negBalance >= 45) {
if (AFD < 6) {
if (Cs < 320) {
if (tcor < 78) {
if (SC < 446) {
return 2;
}
else if (SC >= 446) {
return 8;
}
}
else if (tcor >= 78) {
return 2;
}
}
else if (Cs >= 320) {
if (Rs < 260) {
return 8;
}
else if (Rs >= 260) {
return 2;
}
}
}
else if (AFD >= 6) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (Rs < 171) {
return 8;
}
else if (Rs >= 171) {
if (Rs < 237) {
return 2;
}
else if (Rs >= 237) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 5) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (tcor < 68) {
return 4;
}
else if (tcor >= 68) {
if (Rs < 249) {
return 1;
}
else if (Rs >= 249) {
return 8;
}
}
}
}
}
else if (MVDiff >= 29) {
if (MVDiff < 57) {
if (mcTcor < 86) {
if (mu_mv_mag_sq < 3) {
if (Rs < 138) {
return 8;
}
else if (Rs >= 138) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (refDCval < 42) {
return 8;
}
else if (refDCval >= 42) {
if (Cs < 264) {
return 1;
}
else if (Cs >= 264) {
if (Rs < 214) {
return 8;
}
else if (Rs >= 214) {
return 4;
}
}
}
}
}
else if (mcTcor >= 86) {
if (diffMVdiffVal < -7) {
if (Rs < 99) {
return 1;
}
else if (Rs >= 99) {
if (Rs < 142) {
return 8;
}
else if (Rs >= 142) {
return 2;
}
}
}
else if (diffMVdiffVal >= -7) {
if (Rs < 225) {
return 1;
}
else if (Rs >= 225) {
return 4;
}
}
}
}
else if (MVDiff >= 57) {
if (CsDiff < 4) {
if (MVDiff < 113) {
if (mcTcor < 77) {
if (diffRsCsdiff < 1) {
if (Cs < 248) {
if (Rs < 26) {
return 4;
}
else if (Rs >= 26) {
return 8;
}
}
else if (Cs >= 248) {
return 4;
}
}
else if (diffRsCsdiff >= 1) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 2;
}
}
}
else if (mcTcor >= 77) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (tcor < 56) {
if (tcor < 37) {
return 2;
}
else if (tcor >= 37) {
if (gchDC < 2) {
if (Rs < 165) {
return 8;
}
else if (Rs >= 165) {
return 4;
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (tcor >= 56) {
if (Rs < 105) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 2;
}
}
else if (Rs >= 105) {
return 2;
}
}
}
}
}
else if (MVDiff >= 113) {
if (Cs < 188) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (SC < 363) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
if (SC < 328) {
return 1;
}
else if (SC >= 328) {
if (Rs < 227) {
return 8;
}
else if (Rs >= 227) {
return 1;
}
}
}
}
else if (SC >= 363) {
return 2;
}
}
}
else if (Cs >= 188) {
if (diffTSC < 3) {
if (Cs < 279) {
if (Rs < 186) {
return 4;
}
else if (Rs >= 186) {
return 1;
}
}
else if (Cs >= 279) {
return 4;
}
}
else if (diffTSC >= 3) {
if (tcor < 63) {
return 2;
}
else if (tcor >= 63) {
return 1;
}
}
}
}
}
else if (CsDiff >= 4) {
if (Rs < 217) {
if (mcTcor < 62) {
if (Rs < 136) {
return 2;
}
else if (Rs >= 136) {
return 1;
}
}
else if (mcTcor >= 62) {
return 8;
}
}
else if (Rs >= 217) {
return 2;
}
}
}
}
}
else if (mcTcor >= 90) {
if (ssDCval < 73) {
if (TSC < 5) {
if (ssDCval < 45) {
if (Rs < 160) {
if (TSC < 4) {
if (ssDCval < 43) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (tcor < 85) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
if (Rs < 92) {
return 4;
}
else if (Rs >= 92) {
return 1;
}
}
}
else if (tcor >= 85) {
return 4;
}
}
}
else if (ssDCval >= 43) {
return 2;
}
}
else if (TSC >= 4) {
if (Cs < 259) {
return 8;
}
else if (Cs >= 259) {
return 1;
}
}
}
else if (Rs >= 160) {
if (AFD < 9) {
if (Rs < 248) {
if (mu_mv_mag_sq < 35) {
return 2;
}
else if (mu_mv_mag_sq >= 35) {
return 4;
}
}
else if (Rs >= 248) {
if (Cs < 370) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 4;
}
}
else if (Cs >= 370) {
return 2;
}
}
}
else if (AFD >= 9) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 4;
}
}
}
}
else if (ssDCval >= 45) {
if (CsDiff < 1) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
if (ssDCval < 49) {
if (Rs < 74) {
return 4;
}
else if (Rs >= 74) {
return 2;
}
}
else if (ssDCval >= 49) {
return 1;
}
}
}
else if (CsDiff >= 1) {
if (TSC < 4) {
if (posBalance < 59) {
if (negBalance < 50) {
if (Rs < 299) {
if (AFD < 12) {
if (Cs < 239) {
if (gchDC < 1) {
if (Rs < 253) {
return 1;
}
else if (Rs >= 253) {
return 4;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (Cs >= 239) {
return 4;
}
}
else if (AFD >= 12) {
return 1;
}
}
else if (Rs >= 299) {
if (ssDCval < 62) {
if (Rs < 303) {
return 2;
}
else if (Rs >= 303) {
return 4;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
}
else if (negBalance >= 50) {
if (Rs < 269) {
if (Cs < 143) {
return 4;
}
else if (Cs >= 143) {
return 1;
}
}
else if (Rs >= 269) {
return 2;
}
}
}
else if (posBalance >= 59) {
if (tcor < 88) {
if (Cs < 194) {
if (Rs < 81) {
return 1;
}
else if (Rs >= 81) {
return 4;
}
}
else if (Cs >= 194) {
return 1;
}
}
else if (tcor >= 88) {
if (Rs < 182) {
return 4;
}
else if (Rs >= 182) {
return 2;
}
}
}
}
else if (TSC >= 4) {
if (tcor < 76) {
if (AFD < 14) {
if (Rs < 191) {
return 1;
}
else if (Rs >= 191) {
return 4;
}
}
else if (AFD >= 14) {
return 1;
}
}
else if (tcor >= 76) {
if (AFD < 14) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 231) {
return 2;
}
else if (Rs >= 231) {
return 1;
}
}
}
}
else if (AFD >= 14) {
if (Rs < 183) {
return 2;
}
else if (Rs >= 183) {
return 4;
}
}
}
}
}
}
}
else if (TSC >= 5) {
if (Cs < 292) {
if (MVDiff < 42) {
if (SC < 396) {
return 1;
}
else if (SC >= 396) {
if (Rs < 211) {
return 8;
}
else if (Rs >= 211) {
return 4;
}
}
}
else if (MVDiff >= 42) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (gchDC < 9) {
if (mcTcor < 94) {
return 2;
}
else if (mcTcor >= 94) {
if (Cs < 125) {
return 8;
}
else if (Cs >= 125) {
return 1;
}
}
}
else if (gchDC >= 9) {
return 1;
}
}
}
}
else if (Cs >= 292) {
if (posBalance < 50) {
if (mu_mv_mag_sq < 7) {
if (Rs < 236) {
return 4;
}
else if (Rs >= 236) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
else if (posBalance >= 50) {
if (posBalance < 73) {
return 1;
}
else if (posBalance >= 73) {
return 8;
}
}
}
}
}
else if (ssDCval >= 73) {
if (ssDCval < 78) {
if (tcor < 93) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
return 4;
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (ssDCval >= 78) {
if (Rs < 156) {
if (mu_mv_mag_sq < 28) {
if (Cs < 137) {
if (TSC < 6) {
if (Rs < 134) {
if (Rs < 37) {
return 2;
}
else if (Rs >= 37) {
return 1;
}
}
else if (Rs >= 134) {
return 2;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (Cs >= 137) {
if (diffRsCsdiff < -1) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
if (Cs < 157) {
return 1;
}
else if (Cs >= 157) {
return 2;
}
}
}
else if (diffRsCsdiff >= -1) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 28) {
if (Cs < 107) {
return 2;
}
else if (Cs >= 107) {
return 1;
}
}
}
else if (Rs >= 156) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (diffTSC < 2) {
if (diffMVdiffVal < 2) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 125) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
return 2;
}
}
else if (ssDCval >= 125) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 227) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
else if (Cs >= 227) {
if (Cs < 293) {
if (Cs < 268) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
else if (Cs >= 268) {
return 1;
}
}
else if (Cs >= 293) {
if (Rs < 243) {
return 8;
}
else if (Rs >= 243) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (CsDiff < 4) {
if (Cs < 404) {
return 8;
}
else if (Cs >= 404) {
return 1;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (diffTSC >= 2) {
if (Rs < 301) {
return 1;
}
else if (Rs >= 301) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 133) {
if (Rs < 1) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 8;
}
}
else if (Rs >= 1) {
if (ssDCval < 164) {
if (MVDiff < 22) {
if (Cs < 425) {
return 8;
}
else if (Cs >= 425) {
return 2;
}
}
else if (MVDiff >= 22) {
return 2;
}
}
else if (ssDCval >= 164) {
if (Rs < 142) {
if (MVDiff < 394) {
return 2;
}
else if (MVDiff >= 394) {
return 1;
}
}
else if (Rs >= 142) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (refDCval < 41) {
if (negBalance < 29) {
if (tcor < 89) {
if (mu_mv_mag_sq < 58) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 58) {
if (posBalance < 29) {
if (refDCval < 39) {
return 4;
}
else if (refDCval >= 39) {
return 1;
}
}
else if (posBalance >= 29) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (negBalance < 15) {
if (Rs < 156) {
return 1;
}
else if (Rs >= 156) {
return 4;
}
}
else if (negBalance >= 15) {
return 1;
}
}
}
}
}
else if (tcor >= 89) {
if (Rs < 238) {
if (diffMVdiffVal < 2) {
if (posBalance < 26) {
return 1;
}
else if (posBalance >= 26) {
if (posBalance < 36) {
if (Rs < 144) {
return 4;
}
else if (Rs >= 144) {
return 8;
}
}
else if (posBalance >= 36) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 72) {
return 4;
}
else if (Cs >= 72) {
return 1;
}
}
}
else if (Rs >= 238) {
if (Rs < 255) {
return 4;
}
else if (Rs >= 255) {
return 1;
}
}
}
}
else if (negBalance >= 29) {
if (gchDC < 1) {
if (Cs < 26) {
return 2;
}
else if (Cs >= 26) {
if (diffRsCsdiff < -1) {
if (ssDCval < 36) {
return 2;
}
else if (ssDCval >= 36) {
if (diffMVdiffVal < 14) {
return 8;
}
else if (diffMVdiffVal >= 14) {
return 4;
}
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (negBalance < 41) {
if (RsCsDiff < 1) {
if (posBalance < 30) {
return 1;
}
else if (posBalance >= 30) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
return 8;
}
}
}
else if (RsCsDiff >= 1) {
if (SC < 446) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 1;
}
}
else if (SC >= 446) {
return 1;
}
}
}
else if (negBalance >= 41) {
if (refDCval < 40) {
if (Rs < 136) {
return 8;
}
else if (Rs >= 136) {
if (SC < 362) {
if (Rs < 152) {
return 2;
}
else if (Rs >= 152) {
if (Rs < 222) {
return 4;
}
else if (Rs >= 222) {
return 1;
}
}
}
else if (SC >= 362) {
if (Cs < 309) {
return 8;
}
else if (Cs >= 309) {
return 2;
}
}
}
}
else if (refDCval >= 40) {
return 4;
}
}
}
}
}
else if (refDCval >= 41) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (posBalance < 36) {
if (MVDiff < 5) {
if (Rs < 86) {
if (Rs < 66) {
return 8;
}
else if (Rs >= 66) {
return 2;
}
}
else if (Rs >= 86) {
return 1;
}
}
else if (MVDiff >= 5) {
if (MVDiff < 54) {
if (TSC < 3) {
if (mu_mv_mag_sq < 81) {
return 1;
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 21) {
if (Rs < 103) {
return 8;
}
else if (Rs >= 103) {
return 1;
}
}
else if (negBalance >= 21) {
return 8;
}
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (MVDiff >= 54) {
if (CsDiff < 3) {
if (negBalance < 121) {
return 1;
}
else if (negBalance >= 121) {
return 8;
}
}
else if (CsDiff >= 3) {
if (MVDiff < 602) {
return 2;
}
else if (MVDiff >= 602) {
return 8;
}
}
}
}
}
else if (posBalance >= 36) {
if (mcTcor < 96) {
if (MVDiff < 31) {
if (posBalance < 38) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
return 2;
}
}
else if (posBalance >= 38) {
if (ssDCval < 86) {
if (Cs < 185) {
if (mu_mv_mag_sq < 93) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (mu_mv_mag_sq < 85) {
if (Rs < 39) {
return 8;
}
else if (Rs >= 39) {
if (negBalance < 28) {
if (Cs < 119) {
return 8;
}
else if (Cs >= 119) {
return 1;
}
}
else if (negBalance >= 28) {
if (Rs < 255) {
return 1;
}
else if (Rs >= 255) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (Rs < 68) {
return 1;
}
else if (Rs >= 68) {
return 8;
}
}
}
else if (Cs >= 185) {
if (ssDCval < 68) {
if (mcTcor < 92) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (Rs < 247) {
return 4;
}
else if (Rs >= 247) {
return 2;
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 49) {
if (Rs < 274) {
return 8;
}
else if (Rs >= 274) {
return 1;
}
}
else if (ssDCval >= 49) {
if (SC < 705) {
return 4;
}
else if (SC >= 705) {
return 1;
}
}
}
}
else if (ssDCval >= 68) {
return 8;
}
}
}
else if (ssDCval >= 86) {
if (RsDiff < 1) {
if (Cs < 215) {
if (Rs < 57) {
return 4;
}
else if (Rs >= 57) {
return 1;
}
}
else if (Cs >= 215) {
return 2;
}
}
else if (RsDiff >= 1) {
if (Cs < 358) {
if (mcTcor < 92) {
if (TSC < 5) {
if (Rs < 161) {
if (MVDiff < 28) {
if (Rs < 142) {
if (Rs < 134) {
return 8;
}
else if (Rs >= 134) {
return 2;
}
}
else if (Rs >= 142) {
return 8;
}
}
else if (MVDiff >= 28) {
return 2;
}
}
else if (Rs >= 161) {
return 2;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (mcTcor >= 92) {
if (RsCsDiff < 5) {
return 8;
}
else if (RsCsDiff >= 5) {
return 1;
}
}
}
else if (Cs >= 358) {
return 8;
}
}
}
}
}
else if (MVDiff >= 31) {
if (ssDCval < 86) {
if (MVDiff < 115) {
if (tcor < 29) {
if (Rs < 209) {
return 2;
}
else if (Rs >= 209) {
return 4;
}
}
else if (tcor >= 29) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 51) {
if (Cs < 265) {
return 1;
}
else if (Cs >= 265) {
return 8;
}
}
else if (mu_mv_mag_sq >= 51) {
if (ssDCval < 68) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
else if (ssDCval >= 68) {
if (Rs < 70) {
return 8;
}
else if (Rs >= 70) {
if (Rs < 139) {
return 1;
}
else if (Rs >= 139) {
return 4;
}
}
}
}
}
else if (CsDiff >= 3) {
return 1;
}
}
}
else if (MVDiff >= 115) {
if (MVDiff < 377) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (AFD < 34) {
if (MVDiff < 128) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (MVDiff >= 128) {
if (Rs < 301) {
return 1;
}
else if (Rs >= 301) {
return 8;
}
}
}
else if (AFD >= 34) {
if (MVDiff < 235) {
return 8;
}
else if (MVDiff >= 235) {
return 4;
}
}
}
}
else if (MVDiff >= 377) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
else if (ssDCval >= 86) {
if (mcTcor < 68) {
return 4;
}
else if (mcTcor >= 68) {
if (AFD < 29) {
if (tcor < 51) {
if (MVDiff < 49) {
if (negBalance < 53) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
return 8;
}
}
else if (negBalance >= 53) {
return 2;
}
}
else if (MVDiff >= 49) {
return 8;
}
}
else if (tcor >= 51) {
if (MVDiff < 76) {
return 1;
}
else if (MVDiff >= 76) {
return 2;
}
}
}
else if (AFD >= 29) {
if (AFD < 46) {
return 1;
}
else if (AFD >= 46) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (negBalance < 26) {
if (MVDiff < 2) {
if (tcor < 96) {
if (Rs < 82) {
return 2;
}
else if (Rs >= 82) {
if (Rs < 133) {
return 1;
}
else if (Rs >= 133) {
return 8;
}
}
}
else if (tcor >= 96) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Cs < 64) {
return 8;
}
else if (Cs >= 64) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
return 8;
}
}
}
}
else if (negBalance >= 26) {
if (ssDCval < 70) {
if (MVDiff < 2) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (posBalance < 47) {
if (ssDCval < 56) {
return 2;
}
else if (ssDCval >= 56) {
if (Rs < 156) {
return 4;
}
else if (Rs >= 156) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (Cs < 176) {
if (Rs < 129) {
return 4;
}
else if (Rs >= 129) {
return 1;
}
}
else if (Cs >= 176) {
return 4;
}
}
}
}
else if (ssDCval >= 70) {
if (Rs < 65) {
return 4;
}
else if (Rs >= 65) {
if (CsDiff < 1) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (Cs < 88) {
if (Rs < 123) {
return 8;
}
else if (Rs >= 123) {
return 1;
}
}
else if (Cs >= 88) {
return 8;
}
}
}
else if (CsDiff >= 1) {
if (Rs < 266) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (Rs >= 266) {
if (Rs < 299) {
return 2;
}
else if (Rs >= 299) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 66) {
if (posBalance < 27) {
if (Cs < 1) {
if (ssDCval < 42) {
return 8;
}
else if (ssDCval >= 42) {
return 2;
}
}
else if (Cs >= 1) {
if (SC < 217) {
if (mcTcor < 104) {
if (refDCval < 38) {
if (Rs < 64) {
if (Rs < 30) {
return 4;
}
else if (Rs >= 30) {
return 8;
}
}
else if (Rs >= 64) {
return 1;
}
}
else if (refDCval >= 38) {
return 8;
}
}
else if (mcTcor >= 104) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (SC >= 217) {
if (negBalance < 26) {
if (negBalance < 17) {
if (Rs < 218) {
return 1;
}
else if (Rs >= 218) {
return 4;
}
}
else if (negBalance >= 17) {
return 4;
}
}
else if (negBalance >= 26) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
if (mu_mv_mag_sq < 31) {
return 8;
}
else if (mu_mv_mag_sq >= 31) {
if (Rs < 269) {
return 4;
}
else if (Rs >= 269) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 27) {
if (TSCindex < 2) {
if (MVDiff < 6) {
if (Rs < 146) {
if (mu_mv_mag_sq < 123) {
if (ssDCval < 58) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 106) {
return 8;
}
else if (Rs >= 106) {
return 4;
}
}
}
else if (ssDCval >= 58) {
if (Rs < 142) {
return 1;
}
else if (Rs >= 142) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
else if (Rs >= 146) {
if (SC < 344) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 1;
}
}
}
else if (SC >= 344) {
if (Rs < 245) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 206) {
return 1;
}
else if (Rs >= 206) {
return 4;
}
}
}
else if (Rs >= 245) {
if (SC < 562) {
return 4;
}
else if (SC >= 562) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (Rs < 123) {
if (Cs < 217) {
return 1;
}
else if (Cs >= 217) {
if (Rs < 65) {
return 1;
}
else if (Rs >= 65) {
return 4;
}
}
}
else if (Rs >= 123) {
if (Cs < 121) {
return 8;
}
else if (Cs >= 121) {
return 4;
}
}
}
}
else if (TSCindex >= 2) {
if (CsDiff < 1) {
if (SC < 350) {
if (tcor < 100) {
if (MVDiff < 1) {
if (Cs < 123) {
return 1;
}
else if (Cs >= 123) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 7) {
if (Rs < 36) {
return 8;
}
else if (Rs >= 36) {
return 2;
}
}
else if (Cs >= 7) {
if (mcTcor < 99) {
if (tcor < 94) {
if (TSCindex < 3) {
if (SC < 323) {
return 1;
}
else if (SC >= 323) {
return 4;
}
}
else if (TSCindex >= 3) {
if (Cs < 126) {
if (mu_mv_mag_sq < 49) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
if (Rs < 31) {
return 4;
}
else if (Rs >= 31) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 49) {
return 4;
}
}
else if (Cs >= 126) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
return 2;
}
}
}
}
else if (tcor >= 94) {
if (SC < 286) {
if (posBalance < 41) {
if (Cs < 101) {
return 1;
}
else if (Cs >= 101) {
return 8;
}
}
else if (posBalance >= 41) {
return 1;
}
}
else if (SC >= 286) {
if (negBalance < 35) {
if (posBalance < 47) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (posBalance >= 47) {
if (Rs < 139) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
else if (Rs >= 139) {
return 1;
}
}
}
else if (negBalance >= 35) {
return 1;
}
}
}
}
else if (mcTcor >= 99) {
if (Cs < 229) {
return 1;
}
else if (Cs >= 229) {
if (Rs < 38) {
return 1;
}
else if (Rs >= 38) {
return 4;
}
}
}
}
}
}
else if (tcor >= 100) {
if (ssDCval < 47) {
if (Rs < 135) {
return 4;
}
else if (Rs >= 135) {
return 1;
}
}
else if (ssDCval >= 47) {
return 8;
}
}
}
else if (SC >= 350) {
if (Rs < 176) {
if (ssDCval < 52) {
if (ssDCval < 38) {
return 8;
}
else if (ssDCval >= 38) {
return 1;
}
}
else if (ssDCval >= 52) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
if (Rs < 169) {
if (Cs < 218) {
return 1;
}
else if (Cs >= 218) {
return 4;
}
}
else if (Rs >= 169) {
return 1;
}
}
}
}
else if (Rs >= 176) {
if (ssDCval < 64) {
if (MVDiff < 1) {
if (ssDCval < 53) {
if (Cs < 410) {
return 4;
}
else if (Cs >= 410) {
return 1;
}
}
else if (ssDCval >= 53) {
if (negBalance < 23) {
if (mu_mv_mag_sq < 127) {
if (Cs < 268) {
return 4;
}
else if (Cs >= 268) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 212) {
return 1;
}
else if (Rs >= 212) {
if (Rs < 258) {
return 4;
}
else if (Rs >= 258) {
return 1;
}
}
}
}
else if (negBalance >= 23) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 56) {
return 1;
}
else if (mu_mv_mag_sq >= 56) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 1;
}
}
}
}
else if (ssDCval >= 56) {
if (negBalance < 30) {
if (SC < 373) {
return 1;
}
else if (SC >= 373) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 2;
}
}
}
else if (negBalance >= 30) {
if (SC < 505) {
if (refDCval < 63) {
if (SC < 410) {
if (MVDiff < 4) {
if (Cs < 210) {
return 1;
}
else if (Cs >= 210) {
return 4;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (SC >= 410) {
return 4;
}
}
else if (refDCval >= 63) {
if (posBalance < 53) {
if (SC < 492) {
if (Cs < 161) {
return 4;
}
else if (Cs >= 161) {
return 1;
}
}
else if (SC >= 492) {
return 4;
}
}
else if (posBalance >= 53) {
return 4;
}
}
}
else if (SC >= 505) {
return 1;
}
}
}
}
}
else if (ssDCval >= 64) {
if (negBalance < 28) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
return 1;
}
}
}
else if (negBalance >= 28) {
return 1;
}
}
}
}
}
else if (CsDiff >= 1) {
if (refDCval < 52) {
if (MVDiff < 18) {
if (Cs < 179) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 179) {
if (negBalance < 19) {
if (Rs < 220) {
return 4;
}
else if (Rs >= 220) {
return 1;
}
}
else if (negBalance >= 19) {
return 1;
}
}
}
else if (MVDiff >= 18) {
if (Cs < 128) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (Cs >= 128) {
if (refDCval < 41) {
return 1;
}
else if (refDCval >= 41) {
if (mu_mv_mag_sq < 2) {
if (Cs < 232) {
return 1;
}
else if (Cs >= 232) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (refDCval < 48) {
return 4;
}
else if (refDCval >= 48) {
if (Rs < 232) {
return 1;
}
else if (Rs >= 232) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 52) {
if (TSC < 4) {
if (Rs < 36) {
return 1;
}
else if (Rs >= 36) {
if (Rs < 229) {
if (AFD < 6) {
if (Cs < 178) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (Cs >= 178) {
if (ssDCval < 54) {
if (MVDiff < 11) {
if (Rs < 192) {
return 4;
}
else if (Rs >= 192) {
return 1;
}
}
else if (MVDiff >= 11) {
if (Rs < 157) {
return 2;
}
else if (Rs >= 157) {
return 1;
}
}
}
else if (ssDCval >= 54) {
return 4;
}
}
}
else if (AFD >= 6) {
if (mcTcor < 99) {
if (Cs < 314) {
if (SC < 184) {
if (Rs < 83) {
return 4;
}
else if (Rs >= 83) {
return 1;
}
}
else if (SC >= 184) {
return 4;
}
}
else if (Cs >= 314) {
return 1;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
else if (Rs >= 229) {
if (Rs < 301) {
if (MVDiff < 7) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Rs < 241) {
if (Rs < 233) {
return 1;
}
else if (Rs >= 233) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (Rs < 236) {
return 1;
}
else if (Rs >= 236) {
return 8;
}
}
}
}
else if (Rs >= 241) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (Cs < 468) {
if (diffAFD < -1) {
if (refDCval < 65) {
return 4;
}
else if (refDCval >= 65) {
return 1;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (Cs >= 468) {
return 1;
}
}
}
else if (Rs >= 301) {
if (Cs < 365) {
return 2;
}
else if (Cs >= 365) {
return 4;
}
}
}
}
}
else if (TSC >= 4) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
if (negBalance < 42) {
if (CsDiff < 2) {
if (Rs < 227) {
return 8;
}
else if (Rs >= 227) {
return 2;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (negBalance >= 42) {
if (Rs < 271) {
return 2;
}
else if (Rs >= 271) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 66) {
if (Cs < 239) {
if (mu_mv_mag_sq < 100) {
if (Rs < 80) {
if (ssDCval < 94) {
if (AFD < 8) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
else if (AFD >= 8) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
if (Rs < 54) {
return 8;
}
else if (Rs >= 54) {
return 1;
}
}
}
}
else if (ssDCval >= 94) {
if (AFD < 3) {
if (MVDiff < 33) {
return 4;
}
else if (MVDiff >= 33) {
if (Cs < 98) {
return 1;
}
else if (Cs >= 98) {
return 2;
}
}
}
else if (AFD >= 3) {
if (gchDC < 9) {
if (ssDCval < 100) {
return 1;
}
else if (ssDCval >= 100) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (Cs < 114) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 1;
}
}
else if (Cs >= 114) {
if (Rs < 16) {
return 2;
}
else if (Rs >= 16) {
return 1;
}
}
}
}
}
else if (gchDC >= 9) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
}
}
}
else if (Rs >= 80) {
if (ssDCval < 106) {
if (AFD < 2) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Cs < 193) {
if (Cs < 93) {
return 2;
}
else if (Cs >= 93) {
return 1;
}
}
else if (Cs >= 193) {
if (Rs < 146) {
return 1;
}
else if (Rs >= 146) {
return 2;
}
}
}
}
else if (AFD >= 2) {
if (SC < 298) {
if (AFD < 6) {
if (refDCval < 74) {
if (tcor < 97) {
if (Rs < 124) {
return 1;
}
else if (Rs >= 124) {
return 8;
}
}
else if (tcor >= 97) {
return 2;
}
}
else if (refDCval >= 74) {
if (ssDCval < 83) {
if (MVDiff < 7) {
if (Rs < 137) {
return 1;
}
else if (Rs >= 137) {
return 8;
}
}
else if (MVDiff >= 7) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
else if (ssDCval >= 83) {
if (ssDCval < 98) {
return 1;
}
else if (ssDCval >= 98) {
if (Cs < 55) {
return 2;
}
else if (Cs >= 55) {
return 1;
}
}
}
}
}
else if (AFD >= 6) {
if (ssDCval < 90) {
if (Rs < 85) {
if (Cs < 111) {
return 1;
}
else if (Cs >= 111) {
return 4;
}
}
else if (Rs >= 85) {
if (mcTcor < 100) {
if (RsCsDiff < 5) {
if (Rs < 103) {
if (Cs < 115) {
return 1;
}
else if (Cs >= 115) {
return 2;
}
}
else if (Rs >= 103) {
if (mu_mv_mag_sq < 43) {
return 1;
}
else if (mu_mv_mag_sq >= 43) {
return 8;
}
}
}
else if (RsCsDiff >= 5) {
return 4;
}
}
else if (mcTcor >= 100) {
if (Rs < 196) {
return 8;
}
else if (Rs >= 196) {
return 2;
}
}
}
}
else if (ssDCval >= 90) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Cs < 13) {
return 2;
}
else if (Cs >= 13) {
return 1;
}
}
}
}
}
else if (SC >= 298) {
if (tcor < 91) {
if (Cs < 184) {
if (mu_mv_mag_sq < 28) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 4;
}
}
else if (mu_mv_mag_sq >= 28) {
return 4;
}
}
else if (Cs >= 184) {
if (mu_mv_mag_sq < 11) {
return 4;
}
else if (mu_mv_mag_sq >= 11) {
if (Rs < 198) {
if (Rs < 106) {
return 2;
}
else if (Rs >= 106) {
return 8;
}
}
else if (Rs >= 198) {
return 2;
}
}
}
}
else if (tcor >= 91) {
if (refDCval < 93) {
if (MVDiff < 4) {
if (Cs < 184) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (Rs < 214) {
return 2;
}
else if (Rs >= 214) {
return 8;
}
}
}
}
else if (Cs >= 184) {
if (mu_mv_mag_sq < 2) {
if (Cs < 202) {
if (Rs < 217) {
return 2;
}
else if (Rs >= 217) {
return 1;
}
}
else if (Cs >= 202) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 280) {
return 1;
}
else if (Rs >= 280) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (MVDiff < 27) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (Cs < 131) {
return 1;
}
else if (Cs >= 131) {
return 8;
}
}
}
else if (MVDiff >= 27) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
if (Cs < 65) {
return 1;
}
else if (Cs >= 65) {
return 8;
}
}
}
}
}
else if (refDCval >= 93) {
if (MVDiff < 13) {
if (Cs < 165) {
return 2;
}
else if (Cs >= 165) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (MVDiff >= 13) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (SC < 333) {
if (Rs < 177) {
return 2;
}
else if (Rs >= 177) {
return 4;
}
}
else if (SC >= 333) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 106) {
if (tcor < 76) {
return 1;
}
else if (tcor >= 76) {
if (ssDCval < 121) {
return 2;
}
else if (ssDCval >= 121) {
if (ssDCval < 137) {
if (negBalance < 39) {
if (Cs < 87) {
return 1;
}
else if (Cs >= 87) {
return 4;
}
}
else if (negBalance >= 39) {
return 8;
}
}
else if (ssDCval >= 137) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 121) {
if (Rs < 87) {
if (Rs < 62) {
return 1;
}
else if (Rs >= 62) {
return 4;
}
}
else if (Rs >= 87) {
if (ssDCval < 87) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (ssDCval >= 87) {
if (Cs < 110) {
return 1;
}
else if (Cs >= 110) {
return 8;
}
}
}
}
else if (Cs >= 121) {
if (negBalance < 25) {
if (ssDCval < 80) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
if (mcTcor < 99) {
if (refDCval < 74) {
return 2;
}
else if (refDCval >= 74) {
if (Rs < 160) {
return 2;
}
else if (Rs >= 160) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (Cs < 214) {
if (TSC < 2) {
return 2;
}
else if (TSC >= 2) {
return 1;
}
}
else if (Cs >= 214) {
if (Rs < 233) {
return 1;
}
else if (Rs >= 233) {
return 2;
}
}
}
}
}
else if (ssDCval >= 80) {
if (Rs < 227) {
if (Cs < 134) {
return 8;
}
else if (Cs >= 134) {
if (Rs < 117) {
return 4;
}
else if (Rs >= 117) {
return 2;
}
}
}
else if (Rs >= 227) {
if (Rs < 260) {
return 8;
}
else if (Rs >= 260) {
if (Rs < 305) {
return 1;
}
else if (Rs >= 305) {
return 2;
}
}
}
}
}
else if (negBalance >= 25) {
if (Rs < 178) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 113) {
return 1;
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 129) {
return 8;
}
else if (Rs >= 129) {
return 2;
}
}
}
}
else if (Rs >= 178) {
if (ssDCval < 75) {
if (mcTcor < 98) {
if (Rs < 230) {
return 2;
}
else if (Rs >= 230) {
return 4;
}
}
else if (mcTcor >= 98) {
if (MVDiff < 1) {
if (Rs < 198) {
return 2;
}
else if (Rs >= 198) {
return 1;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
}
else if (ssDCval >= 75) {
return 2;
}
}
}
}
}
}
else if (Cs >= 239) {
if (ssDCval < 76) {
if (refDCval < 72) {
if (TSC < 2) {
if (mu_mv_mag_sq < 100) {
if (Rs < 113) {
return 4;
}
else if (Rs >= 113) {
return 2;
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 343) {
return 4;
}
else if (Cs >= 343) {
if (mu_mv_mag_sq < 123) {
return 4;
}
else if (mu_mv_mag_sq >= 123) {
if (ssDCval < 69) {
if (Rs < 184) {
return 1;
}
else if (Rs >= 184) {
return 4;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
}
else if (TSC >= 2) {
if (mu_mv_mag_sq < 47) {
if (Rs < 48) {
return 1;
}
else if (Rs >= 48) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (Cs < 267) {
if (SC < 387) {
if (Rs < 128) {
return 1;
}
else if (Rs >= 128) {
return 4;
}
}
else if (SC >= 387) {
return 1;
}
}
else if (Cs >= 267) {
if (Rs < 123) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (Rs >= 123) {
if (ssDCval < 70) {
if (mu_mv_mag_sq < 110) {
if (AFD < 3) {
if (SC < 447) {
if (Rs < 145) {
return 1;
}
else if (Rs >= 145) {
return 4;
}
}
else if (SC >= 447) {
return 1;
}
}
else if (AFD >= 3) {
if (Rs < 149) {
return 1;
}
else if (Rs >= 149) {
if (TSC < 3) {
if (Cs < 373) {
return 4;
}
else if (Cs >= 373) {
if (Rs < 225) {
return 1;
}
else if (Rs >= 225) {
return 4;
}
}
}
else if (TSC >= 3) {
if (Rs < 168) {
if (Cs < 303) {
return 4;
}
else if (Cs >= 303) {
return 1;
}
}
else if (Rs >= 168) {
if (negBalance < 28) {
return 1;
}
else if (negBalance >= 28) {
if (Rs < 303) {
if (Cs < 383) {
return 4;
}
else if (Cs >= 383) {
if (Rs < 204) {
return 1;
}
else if (Rs >= 204) {
return 4;
}
}
}
else if (Rs >= 303) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 110) {
if (Cs < 381) {
if (diffMVdiffVal < 1) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
if (Cs < 375) {
return 4;
}
else if (Cs >= 375) {
if (Rs < 194) {
return 1;
}
else if (Rs >= 194) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 1) {
if (Rs < 186) {
return 1;
}
else if (Rs >= 186) {
return 4;
}
}
}
else if (Cs >= 381) {
if (Rs < 219) {
return 1;
}
else if (Rs >= 219) {
return 4;
}
}
}
}
else if (ssDCval >= 70) {
if (CsDiff < 2) {
if (mu_mv_mag_sq < 92) {
return 1;
}
else if (mu_mv_mag_sq >= 92) {
if (Rs < 217) {
return 4;
}
else if (Rs >= 217) {
return 1;
}
}
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 72) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (ssDCval < 75) {
if (Rs < 131) {
return 4;
}
else if (Rs >= 131) {
return 1;
}
}
else if (ssDCval >= 75) {
return 4;
}
}
}
}
else if (ssDCval >= 76) {
if (RsCsDiff < 2) {
if (ssDCval < 129) {
if (Rs < 188) {
if (Cs < 450) {
if (gchDC < 2) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (Cs < 311) {
return 1;
}
else if (Cs >= 311) {
return 2;
}
}
}
else if (Cs >= 450) {
return 1;
}
}
else if (Rs >= 188) {
if (Cs < 296) {
if (ssDCval < 104) {
return 2;
}
else if (ssDCval >= 104) {
return 4;
}
}
else if (Cs >= 296) {
if (Cs < 486) {
return 4;
}
else if (Cs >= 486) {
if (Rs < 231) {
return 1;
}
else if (Rs >= 231) {
return 2;
}
}
}
}
}
else if (ssDCval >= 129) {
if (Rs < 232) {
return 2;
}
else if (Rs >= 232) {
return 4;
}
}
}
else if (RsCsDiff >= 2) {
if (CsDiff < 3) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (negBalance < 45) {
if (tcor < 98) {
if (Rs < 289) {
return 2;
}
else if (Rs >= 289) {
if (Rs < 292) {
return 8;
}
else if (Rs >= 292) {
return 2;
}
}
}
else if (tcor >= 98) {
if (Rs < 190) {
return 1;
}
else if (Rs >= 190) {
return 2;
}
}
}
else if (negBalance >= 45) {
if (posBalance < 55) {
if (tcor < 93) {
if (Rs < 226) {
return 4;
}
else if (Rs >= 226) {
return 1;
}
}
else if (tcor >= 93) {
return 2;
}
}
else if (posBalance >= 55) {
return 4;
}
}
}
}
else if (CsDiff >= 3) {
if (diffTSC < -1) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
else if (diffTSC >= -1) {
if (mu_mv_mag_sq < 29) {
if (Rs < 234) {
return 2;
}
else if (Rs >= 234) {
return 8;
}
}
else if (mu_mv_mag_sq >= 29) {
if (Cs < 269) {
return 8;
}
else if (Cs >= 269) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 307) {
if (RsDiff < 4) {
if (refDCval < 73) {
if (mcTcor < 87) {
if (tcor < 73) {
if (posBalance < 54) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 26) {
if (Rs < 648) {
return 8;
}
else if (Rs >= 648) {
return 1;
}
}
else if (mu_mv_mag_sq >= 26) {
if (Rs < 411) {
return 1;
}
else if (Rs >= 411) {
if (Cs < 342) {
if (Cs < 36) {
return 1;
}
else if (Cs >= 36) {
return 2;
}
}
else if (Cs >= 342) {
return 4;
}
}
}
}
else if (ssDCval >= 56) {
if (posBalance < 33) {
return 1;
}
else if (posBalance >= 33) {
return 2;
}
}
}
else if (posBalance >= 54) {
if (AFD < 9) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (negBalance < 45) {
if (diffRsCsdiff < 2) {
return 2;
}
else if (diffRsCsdiff >= 2) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
return 1;
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (AFD >= 9) {
if (refDCval < 68) {
if (diffRsCsdiff < -12) {
return 2;
}
else if (diffRsCsdiff >= -12) {
if (diffRsCsdiff < 13) {
if (mu_mv_mag_sq < 51) {
if (mcTcor < 85) {
if (MVDiff < 49) {
if (RsCsDiff < 7) {
if (Rs < 327) {
return 8;
}
else if (Rs >= 327) {
return 4;
}
}
else if (RsCsDiff >= 7) {
return 8;
}
}
else if (MVDiff >= 49) {
if (diffAFD < 1) {
if (Rs < 476) {
return 4;
}
else if (Rs >= 476) {
return 8;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
else if (mcTcor >= 85) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 51) {
return 4;
}
}
else if (diffRsCsdiff >= 13) {
return 1;
}
}
}
else if (refDCval >= 68) {
if (SC < 857) {
return 2;
}
else if (SC >= 857) {
return 4;
}
}
}
}
}
else if (tcor >= 73) {
if (Cs < 156) {
if (Cs < 40) {
return 8;
}
else if (Cs >= 40) {
return 1;
}
}
else if (Cs >= 156) {
if (negBalance < 33) {
if (diffTSC < 1) {
if (Cs < 396) {
return 8;
}
else if (Cs >= 396) {
return 2;
}
}
else if (diffTSC >= 1) {
if (Cs < 384) {
return 4;
}
else if (Cs >= 384) {
return 1;
}
}
}
else if (negBalance >= 33) {
if (RsCsDiff < 12) {
return 8;
}
else if (RsCsDiff >= 12) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 87) {
if (Rs < 1280) {
if (tcor < 101) {
if (TSC < 1) {
if (posBalance < 34) {
if (SC < 1059) {
if (Rs < 769) {
return 8;
}
else if (Rs >= 769) {
return 4;
}
}
else if (SC >= 1059) {
if (Cs < 391) {
return 1;
}
else if (Cs >= 391) {
return 8;
}
}
}
else if (posBalance >= 34) {
return 4;
}
}
else if (TSC >= 1) {
if (refDCval < 58) {
if (mu_mv_mag_sq < 88) {
if (Cs < 126) {
if (mu_mv_mag_sq < 54) {
if (diffMVdiffVal < -27) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
return 8;
}
}
else if (diffMVdiffVal >= -27) {
if (diffTSC < -2) {
if (Rs < 749) {
return 8;
}
else if (Rs >= 749) {
return 2;
}
}
else if (diffTSC >= -2) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (posBalance < 52) {
if (Rs < 763) {
return 8;
}
else if (Rs >= 763) {
return 1;
}
}
else if (posBalance >= 52) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 603) {
if (Cs < 29) {
return 8;
}
else if (Cs >= 29) {
return 1;
}
}
else if (Rs >= 603) {
if (Cs < 86) {
return 8;
}
else if (Cs >= 86) {
return 1;
}
}
}
}
else if (Cs >= 126) {
if (Rs < 626) {
if (AFD < 3) {
if (Cs < 341) {
if (Cs < 272) {
return 4;
}
else if (Cs >= 272) {
return 1;
}
}
else if (Cs >= 341) {
return 8;
}
}
else if (AFD >= 3) {
if (ssDCval < 45) {
if (posBalance < 33) {
return 2;
}
else if (posBalance >= 33) {
if (mcTcor < 92) {
if (negBalance < 40) {
if (AFD < 6) {
if (ssDCval < 38) {
if (Rs < 375) {
return 4;
}
else if (Rs >= 375) {
return 1;
}
}
else if (ssDCval >= 38) {
return 8;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (negBalance >= 40) {
if (AFD < 11) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
return 8;
}
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (MVDiff < 6) {
if (ssDCval < 40) {
return 1;
}
else if (ssDCval >= 40) {
if (Cs < 366) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (Rs < 329) {
return 1;
}
else if (Rs >= 329) {
return 4;
}
}
}
else if (Cs >= 366) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (gchDC < 2) {
if (posBalance < 34) {
return 4;
}
else if (posBalance >= 34) {
return 1;
}
}
else if (gchDC >= 2) {
if (Rs < 489) {
return 8;
}
else if (Rs >= 489) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 45) {
if (MVDiff < 53) {
if (refDCval < 57) {
if (CsDiff < 1) {
if (tcor < 95) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (diffAFD < 2) {
if (Cs < 162) {
return 2;
}
else if (Cs >= 162) {
return 1;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
}
else if (tcor >= 95) {
if (diffRsCsdiff < -4) {
return 4;
}
else if (diffRsCsdiff >= -4) {
if (Rs < 337) {
if (Rs < 331) {
return 1;
}
else if (Rs >= 331) {
return 4;
}
}
else if (Rs >= 337) {
return 1;
}
}
}
}
else if (CsDiff >= 1) {
if (TSC < 4) {
if (AFD < 4) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (Rs < 584) {
return 8;
}
else if (Rs >= 584) {
return 1;
}
}
}
else if (AFD >= 4) {
if (SC < 657) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (AFD < 8) {
if (Rs < 346) {
return 2;
}
else if (Rs >= 346) {
return 1;
}
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (SC >= 657) {
if (SC < 914) {
if (ssDCval < 55) {
if (diffMVdiffVal < 2) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
return 1;
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 535) {
return 4;
}
else if (Rs >= 535) {
return 1;
}
}
}
else if (ssDCval >= 55) {
if (RsCsDiff < 7) {
return 4;
}
else if (RsCsDiff >= 7) {
return 1;
}
}
}
else if (SC >= 914) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
if (Rs < 570) {
return 4;
}
else if (Rs >= 570) {
return 1;
}
}
}
}
}
}
else if (TSC >= 4) {
if (refDCval < 52) {
if (diffMVdiffVal < 2) {
if (mu_mv_mag_sq < 37) {
if (Rs < 349) {
return 4;
}
else if (Rs >= 349) {
return 8;
}
}
else if (mu_mv_mag_sq >= 37) {
if (Cs < 197) {
return 8;
}
else if (Cs >= 197) {
return 4;
}
}
}
else if (diffMVdiffVal >= 2) {
if (SC < 873) {
return 4;
}
else if (SC >= 873) {
if (SC < 1067) {
return 1;
}
else if (SC >= 1067) {
return 4;
}
}
}
}
else if (refDCval >= 52) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (SC < 871) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 2;
}
}
else if (SC >= 871) {
if (Cs < 476) {
if (Cs < 372) {
return 4;
}
else if (Cs >= 372) {
return 8;
}
}
else if (Cs >= 476) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 57) {
if (CsDiff < 2) {
if (Rs < 386) {
return 2;
}
else if (Rs >= 386) {
if (Rs < 444) {
return 1;
}
else if (Rs >= 444) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (tcor < 91) {
return 1;
}
else if (tcor >= 91) {
if (Rs < 424) {
return 2;
}
else if (Rs >= 424) {
return 1;
}
}
}
}
}
else if (MVDiff >= 53) {
if (mu_mv_mag_sq < 17) {
return 2;
}
else if (mu_mv_mag_sq >= 17) {
if (Rs < 335) {
return 4;
}
else if (Rs >= 335) {
return 1;
}
}
}
}
}
}
else if (Rs >= 626) {
if (negBalance < 41) {
if (mcTcor < 99) {
if (Cs < 483) {
if (Cs < 432) {
if (MVDiff < 45) {
return 1;
}
else if (MVDiff >= 45) {
return 4;
}
}
else if (Cs >= 432) {
if (refDCval < 56) {
if (negBalance < 28) {
if (Rs < 692) {
return 1;
}
else if (Rs >= 692) {
return 2;
}
}
else if (negBalance >= 28) {
return 1;
}
}
else if (refDCval >= 56) {
return 4;
}
}
}
else if (Cs >= 483) {
if (Rs < 777) {
return 2;
}
else if (Rs >= 777) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (RsCsDiff < 4) {
if (AFD < 3) {
if (Rs < 642) {
return 8;
}
else if (Rs >= 642) {
return 4;
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (RsCsDiff >= 4) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (Cs < 279) {
return 1;
}
else if (Cs >= 279) {
if (ssDCval < 38) {
return 1;
}
else if (ssDCval >= 38) {
if (diffRsCsdiff < -6) {
if (posBalance < 57) {
if (Rs < 922) {
return 1;
}
else if (Rs >= 922) {
return 2;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 88) {
if (mcTcor < 97) {
if (negBalance < 33) {
if (AFD < 3) {
if (mu_mv_mag_sq < 116) {
if (Cs < 301) {
if (Rs < 535) {
return 8;
}
else if (Rs >= 535) {
return 2;
}
}
else if (Cs >= 301) {
return 1;
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
else if (AFD >= 3) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (Cs < 376) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (Cs >= 376) {
if (tcor < 94) {
if (Cs < 474) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (Cs >= 474) {
return 4;
}
}
else if (tcor >= 94) {
return 1;
}
}
}
}
}
else if (negBalance >= 33) {
if (tcor < 96) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 4;
}
}
else if (tcor >= 96) {
if (Rs < 460) {
return 1;
}
else if (Rs >= 460) {
return 8;
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 4) {
if (diffAFD < -2) {
if (Rs < 673) {
return 4;
}
else if (Rs >= 673) {
return 2;
}
}
else if (diffAFD >= -2) {
if (negBalance < 18) {
if (refDCval < 56) {
return 1;
}
else if (refDCval >= 56) {
if (Rs < 717) {
return 8;
}
else if (Rs >= 717) {
return 1;
}
}
}
else if (negBalance >= 18) {
if (ssDCval < 47) {
if (mcTcor < 98) {
if (Rs < 519) {
return 8;
}
else if (Rs >= 519) {
return 4;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (ssDCval >= 47) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
}
else if (MVDiff >= 4) {
if (Cs < 414) {
if (mu_mv_mag_sq < 103) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (Rs < 673) {
return 4;
}
else if (Rs >= 673) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 407) {
return 4;
}
else if (Rs >= 407) {
if (Rs < 442) {
return 8;
}
else if (Rs >= 442) {
return 1;
}
}
}
}
else if (Cs >= 414) {
return 8;
}
}
}
}
}
else if (refDCval >= 58) {
if (Rs < 1057) {
if (mu_mv_mag_sq < 81) {
if (mcTcor < 97) {
if (Cs < 135) {
if (MVDiff < 9) {
if (Rs < 723) {
return 4;
}
else if (Rs >= 723) {
return 2;
}
}
else if (MVDiff >= 9) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (Cs >= 135) {
if (SC < 1071) {
if (Cs < 433) {
if (ssDCval < 70) {
if (RsCsDiff < 9) {
if (negBalance < 39) {
if (posBalance < 56) {
if (MVDiff < 10) {
if (Cs < 417) {
return 1;
}
else if (Cs >= 417) {
return 4;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (posBalance >= 56) {
if (mu_mv_mag_sq < 38) {
if (Cs < 303) {
return 1;
}
else if (Cs >= 303) {
return 2;
}
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 366) {
return 4;
}
else if (Rs >= 366) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 461) {
if (diffRsCsdiff < -2) {
if (SC < 721) {
if (SC < 672) {
return 4;
}
else if (SC >= 672) {
return 2;
}
}
else if (SC >= 721) {
return 4;
}
}
else if (diffRsCsdiff >= -2) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
if (Cs < 358) {
return 4;
}
else if (Cs >= 358) {
return 2;
}
}
}
}
else if (Rs >= 461) {
if (Rs < 763) {
return 4;
}
else if (Rs >= 763) {
return 2;
}
}
}
}
else if (RsCsDiff >= 9) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (SC < 967) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
if (Rs < 528) {
return 2;
}
else if (Rs >= 528) {
return 1;
}
}
}
else if (SC >= 967) {
if (Rs < 732) {
return 4;
}
else if (Rs >= 732) {
return 2;
}
}
}
}
}
else if (ssDCval >= 70) {
if (Cs < 273) {
if (Rs < 485) {
return 8;
}
else if (Rs >= 485) {
return 4;
}
}
else if (Cs >= 273) {
if (Rs < 328) {
return 8;
}
else if (Rs >= 328) {
return 1;
}
}
}
}
else if (Cs >= 433) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (Cs < 495) {
if (Cs < 467) {
return 2;
}
else if (Cs >= 467) {
return 8;
}
}
else if (Cs >= 495) {
return 1;
}
}
}
}
else if (SC >= 1071) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (Cs < 404) {
if (mu_mv_mag_sq < 67) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 4;
}
}
else if (mu_mv_mag_sq >= 67) {
return 4;
}
}
else if (Cs >= 404) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 557) {
if (TSC < 4) {
if (MVDiff < 17) {
if (RsDiff < 2) {
if (negBalance < 26) {
return 4;
}
else if (negBalance >= 26) {
if (posBalance < 42) {
if (Cs < 372) {
return 4;
}
else if (Cs >= 372) {
return 8;
}
}
else if (posBalance >= 42) {
if (refDCval < 69) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (Rs < 332) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Rs >= 332) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
}
else if (refDCval >= 69) {
return 1;
}
}
}
}
else if (RsDiff >= 2) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (MVDiff >= 17) {
if (Cs < 324) {
return 8;
}
else if (Cs >= 324) {
if (Rs < 372) {
return 4;
}
else if (Rs >= 372) {
return 2;
}
}
}
}
else if (TSC >= 4) {
if (diffMVdiffVal < -23) {
return 8;
}
else if (diffMVdiffVal >= -23) {
return 1;
}
}
}
else if (Rs >= 557) {
if (tcor < 99) {
if (diffRsCsdiff < 2) {
if (mu_mv_mag_sq < 7) {
if (Cs < 179) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 4;
}
}
else if (Cs >= 179) {
if (negBalance < 41) {
return 4;
}
else if (negBalance >= 41) {
if (tcor < 95) {
if (Rs < 570) {
return 4;
}
else if (Rs >= 570) {
return 1;
}
}
else if (tcor >= 95) {
if (Rs < 698) {
return 8;
}
else if (Rs >= 698) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 500) {
return 1;
}
else if (Cs >= 500) {
return 4;
}
}
}
else if (diffRsCsdiff >= 2) {
if (refDCval < 61) {
return 2;
}
else if (refDCval >= 61) {
if (Cs < 204) {
return 1;
}
else if (Cs >= 204) {
if (Cs < 444) {
return 4;
}
else if (Cs >= 444) {
return 1;
}
}
}
}
}
else if (tcor >= 99) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (refDCval < 63) {
if (Cs < 344) {
if (posBalance < 45) {
if (MVDiff < 2) {
if (Rs < 848) {
return 1;
}
else if (Rs >= 848) {
return 2;
}
}
else if (MVDiff >= 2) {
if (Cs < 312) {
return 8;
}
else if (Cs >= 312) {
return 2;
}
}
}
else if (posBalance >= 45) {
if (Cs < 170) {
return 1;
}
else if (Cs >= 170) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (Rs < 453) {
return 1;
}
else if (Rs >= 453) {
return 4;
}
}
}
}
}
else if (Cs >= 344) {
if (mu_mv_mag_sq < 96) {
if (MVDiff < 5) {
if (tcor < 96) {
if (Rs < 579) {
return 8;
}
else if (Rs >= 579) {
if (Rs < 854) {
return 2;
}
else if (Rs >= 854) {
return 4;
}
}
}
else if (tcor >= 96) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
if (Rs < 817) {
if (Rs < 501) {
return 1;
}
else if (Rs >= 501) {
return 8;
}
}
else if (Rs >= 817) {
return 1;
}
}
}
}
else if (MVDiff >= 5) {
if (Rs < 506) {
if (Rs < 366) {
return 1;
}
else if (Rs >= 366) {
return 2;
}
}
else if (Rs >= 506) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (SC < 1018) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 447) {
return 4;
}
else if (Cs >= 447) {
return 8;
}
}
}
else if (SC >= 1018) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 1;
}
}
}
}
}
else if (refDCval >= 63) {
if (Cs < 322) {
if (Rs < 423) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 4;
}
}
}
else if (Rs >= 423) {
if (Rs < 679) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Rs < 536) {
return 2;
}
else if (Rs >= 536) {
return 4;
}
}
}
else if (Rs >= 679) {
if (Rs < 814) {
if (diffRsCsdiff < -1) {
return 1;
}
else if (diffRsCsdiff >= -1) {
return 4;
}
}
else if (Rs >= 814) {
return 1;
}
}
}
}
else if (Cs >= 322) {
if (refDCval < 72) {
if (mu_mv_mag_sq < 122) {
if (refDCval < 67) {
if (mcTcor < 99) {
if (MVDiff < 4) {
if (SC < 1247) {
return 1;
}
else if (SC >= 1247) {
return 8;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (mcTcor >= 99) {
if (mu_mv_mag_sq < 102) {
return 8;
}
else if (mu_mv_mag_sq >= 102) {
return 1;
}
}
}
else if (refDCval >= 67) {
if (posBalance < 42) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
return 1;
}
}
else if (posBalance >= 42) {
if (mu_mv_mag_sq < 105) {
if (MVDiff < 3) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (negBalance < 31) {
if (Rs < 637) {
return 1;
}
else if (Rs >= 637) {
return 2;
}
}
else if (negBalance >= 31) {
return 2;
}
}
}
else if (MVDiff >= 3) {
if (tcor < 97) {
if (Rs < 449) {
return 8;
}
else if (Rs >= 449) {
return 1;
}
}
else if (tcor >= 97) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (tcor < 99) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
if (Rs < 522) {
return 2;
}
else if (Rs >= 522) {
if (Rs < 657) {
return 1;
}
else if (Rs >= 657) {
return 8;
}
}
}
}
else if (tcor >= 99) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Cs < 388) {
return 2;
}
else if (Cs >= 388) {
if (Rs < 688) {
return 1;
}
else if (Rs >= 688) {
if (MVDiff < 1) {
if (Rs < 745) {
return 1;
}
else if (Rs >= 745) {
return 8;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
else if (refDCval >= 72) {
if (AFD < 2) {
if (Rs < 554) {
if (Rs < 430) {
return 2;
}
else if (Rs >= 430) {
return 1;
}
}
else if (Rs >= 554) {
return 4;
}
}
else if (AFD >= 2) {
if (MVDiff < 2) {
if (Rs < 546) {
return 2;
}
else if (Rs >= 546) {
return 4;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1057) {
if (ssDCval < 69) {
if (diffMVdiffVal < -6) {
if (Rs < 1147) {
return 2;
}
else if (Rs >= 1147) {
if (Rs < 1182) {
return 1;
}
else if (Rs >= 1182) {
return 4;
}
}
}
else if (diffMVdiffVal >= -6) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (Rs < 1151) {
return 1;
}
else if (Rs >= 1151) {
return 2;
}
}
}
}
else if (ssDCval >= 69) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 8;
}
}
}
}
}
}
else if (tcor >= 101) {
if (mu_mv_mag_sq < 101) {
if (posBalance < 20) {
if (MVDiff < 23) {
if (diffAFD < -1) {
if (Cs < 124) {
return 8;
}
else if (Cs >= 124) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (posBalance >= 20) {
if (Rs < 641) {
if (mu_mv_mag_sq < 94) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
if (Rs < 510) {
return 8;
}
else if (Rs >= 510) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (Rs < 437) {
return 8;
}
else if (Rs >= 437) {
return 2;
}
}
}
else if (Rs >= 641) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
return 4;
}
}
}
}
else if (Rs >= 1280) {
if (RsDiff < 2) {
if (ssDCval < 63) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (Rs < 1648) {
if (TSCindex < 2) {
if (ssDCval < 60) {
if (Rs < 1428) {
return 8;
}
else if (Rs >= 1428) {
return 1;
}
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (TSCindex >= 2) {
return 8;
}
}
else if (Rs >= 1648) {
return 2;
}
}
}
else if (ssDCval >= 63) {
if (ssDCval < 64) {
if (Cs < 125) {
return 8;
}
else if (Cs >= 125) {
if (negBalance < 19) {
return 2;
}
else if (negBalance >= 19) {
if (Rs < 1406) {
return 8;
}
else if (Rs >= 1406) {
if (refDCval < 64) {
if (diffMVdiffVal < -1) {
if (Rs < 1745) {
return 1;
}
else if (Rs >= 1745) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (refDCval >= 64) {
return 1;
}
}
}
}
}
else if (ssDCval >= 64) {
if (Rs < 1728) {
if (ssDCval < 68) {
if (Cs < 490) {
return 1;
}
else if (Cs >= 490) {
if (Rs < 1463) {
return 8;
}
else if (Rs >= 1463) {
return 1;
}
}
}
else if (ssDCval >= 68) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 1;
}
}
}
else if (Rs >= 1728) {
return 2;
}
}
}
}
else if (RsDiff >= 2) {
if (posBalance < 68) {
if (TSCindex < 3) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (TSCindex >= 3) {
if (negBalance < 34) {
if (tcor < 97) {
if (gchDC < 2) {
if (Rs < 2204) {
if (negBalance < 24) {
if (mu_mv_mag_sq < 64) {
if (Cs < 304) {
return 1;
}
else if (Cs >= 304) {
return 2;
}
}
else if (mu_mv_mag_sq >= 64) {
return 1;
}
}
else if (negBalance >= 24) {
return 1;
}
}
else if (Rs >= 2204) {
return 4;
}
}
else if (gchDC >= 2) {
if (ssDCval < 59) {
return 2;
}
else if (ssDCval >= 59) {
if (Rs < 1789) {
return 1;
}
else if (Rs >= 1789) {
return 2;
}
}
}
}
else if (tcor >= 97) {
if (posBalance < 38) {
return 2;
}
else if (posBalance >= 38) {
if (Rs < 1664) {
return 8;
}
else if (Rs >= 1664) {
return 1;
}
}
}
}
else if (negBalance >= 34) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
}
}
else if (posBalance >= 68) {
if (Cs < 174) {
return 1;
}
else if (Cs >= 174) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 73) {
if (Cs < 285) {
if (ssDCval < 107) {
if (mu_mv_mag_sq < 115) {
if (SC < 1244) {
if (ssDCval < 101) {
if (RsCsDiff < 3) {
if (RsDiff < 1) {
if (mu_mv_mag_sq < 101) {
return 8;
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 712) {
return 2;
}
else if (Rs >= 712) {
if (refDCval < 81) {
return 8;
}
else if (refDCval >= 81) {
return 1;
}
}
}
}
else if (RsDiff >= 1) {
if (Cs < 81) {
return 1;
}
else if (Cs >= 81) {
if (diffTSC < -1) {
if (Cs < 218) {
if (diffMVdiffVal < 1) {
if (Cs < 196) {
return 8;
}
else if (Cs >= 196) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
else if (Cs >= 218) {
if (Cs < 263) {
return 4;
}
else if (Cs >= 263) {
if (Rs < 472) {
return 2;
}
else if (Rs >= 472) {
return 8;
}
}
}
}
else if (diffTSC >= -1) {
if (tcor < 87) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
else if (tcor >= 87) {
if (Cs < 229) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (Cs >= 229) {
if (mu_mv_mag_sq < 47) {
return 4;
}
else if (mu_mv_mag_sq >= 47) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 3) {
if (tcor < 77) {
if (mu_mv_mag_sq < 58) {
if (Cs < 76) {
if (Rs < 342) {
return 2;
}
else if (Rs >= 342) {
return 1;
}
}
else if (Cs >= 76) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (Rs < 386) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
return 2;
}
}
else if (Rs >= 386) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 58) {
return 4;
}
}
else if (tcor >= 77) {
if (Cs < 134) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (Rs < 591) {
return 1;
}
else if (Rs >= 591) {
return 8;
}
}
}
else if (Cs >= 134) {
if (mcTcor < 97) {
if (Rs < 429) {
return 8;
}
else if (Rs >= 429) {
return 1;
}
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 61) {
if (SC < 790) {
if (refDCval < 76) {
return 4;
}
else if (refDCval >= 76) {
return 1;
}
}
else if (SC >= 790) {
if (diffMVdiffVal < 1) {
if (Cs < 151) {
return 1;
}
else if (Cs >= 151) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (Rs < 387) {
return 8;
}
else if (Rs >= 387) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 101) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Rs < 403) {
return 4;
}
else if (Rs >= 403) {
if (Cs < 217) {
return 1;
}
else if (Cs >= 217) {
return 8;
}
}
}
}
}
else if (SC >= 1244) {
if (refDCval < 102) {
if (Rs < 1000) {
return 8;
}
else if (Rs >= 1000) {
return 1;
}
}
else if (refDCval >= 102) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (Rs < 1929) {
return 1;
}
else if (Rs >= 1929) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (ssDCval < 79) {
if (Cs < 210) {
return 8;
}
else if (Cs >= 210) {
return 1;
}
}
else if (ssDCval >= 79) {
if (ssDCval < 94) {
if (SC < 509) {
if (Cs < 189) {
return 2;
}
else if (Cs >= 189) {
return 1;
}
}
else if (SC >= 509) {
if (posBalance < 46) {
if (posBalance < 33) {
return 1;
}
else if (posBalance >= 33) {
if (Rs < 313) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (Rs >= 313) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 912) {
return 1;
}
else if (Rs >= 912) {
return 2;
}
}
}
}
}
else if (posBalance >= 46) {
return 1;
}
}
}
else if (ssDCval >= 94) {
return 1;
}
}
}
}
else if (ssDCval >= 107) {
if (MVDiff < 20) {
if (diffRsCsdiff < -4) {
if (Rs < 830) {
return 2;
}
else if (Rs >= 830) {
return 4;
}
}
else if (diffRsCsdiff >= -4) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (MVDiff >= 20) {
if (Rs < 472) {
if (ssDCval < 116) {
return 8;
}
else if (ssDCval >= 116) {
if (Cs < 60) {
return 1;
}
else if (Cs >= 60) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 2;
}
}
}
}
else if (Rs >= 472) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
if (Cs < 179) {
return 1;
}
else if (Cs >= 179) {
return 4;
}
}
}
}
}
}
else if (Cs >= 285) {
if (refDCval < 115) {
if (negBalance < 32) {
if (refDCval < 95) {
if (Rs < 2044) {
if (Rs < 443) {
if (refDCval < 85) {
if (Rs < 351) {
return 2;
}
else if (Rs >= 351) {
return 8;
}
}
else if (refDCval >= 85) {
if (Cs < 343) {
return 4;
}
else if (Cs >= 343) {
return 1;
}
}
}
else if (Rs >= 443) {
if (mu_mv_mag_sq < 23) {
if (Rs < 1047) {
return 1;
}
else if (Rs >= 1047) {
return 8;
}
}
else if (mu_mv_mag_sq >= 23) {
if (refDCval < 94) {
if (mu_mv_mag_sq < 86) {
if (tcor < 98) {
if (Rs < 1964) {
return 1;
}
else if (Rs >= 1964) {
return 8;
}
}
else if (tcor >= 98) {
if (Rs < 716) {
if (Cs < 359) {
return 8;
}
else if (Cs >= 359) {
return 1;
}
}
else if (Rs >= 716) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (MVDiff < 8) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (refDCval < 74) {
if (Rs < 578) {
return 1;
}
else if (Rs >= 578) {
return 4;
}
}
else if (refDCval >= 74) {
if (posBalance < 53) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 1695) {
return 1;
}
else if (Rs >= 1695) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (Cs < 386) {
return 8;
}
else if (Cs >= 386) {
return 1;
}
}
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
else if (refDCval >= 94) {
if (Rs < 887) {
return 8;
}
else if (Rs >= 887) {
return 1;
}
}
}
}
}
else if (Rs >= 2044) {
if (Cs < 453) {
if (Cs < 422) {
return 2;
}
else if (Cs >= 422) {
return 1;
}
}
else if (Cs >= 453) {
return 8;
}
}
}
else if (refDCval >= 95) {
if (ssDCval < 99) {
if (Rs < 576) {
return 8;
}
else if (Rs >= 576) {
if (Rs < 929) {
return 4;
}
else if (Rs >= 929) {
return 1;
}
}
}
else if (ssDCval >= 99) {
if (ssDCval < 111) {
return 8;
}
else if (ssDCval >= 111) {
if (Rs < 1101) {
return 2;
}
else if (Rs >= 1101) {
return 1;
}
}
}
}
}
else if (negBalance >= 32) {
if (AFD < 4) {
if (Rs < 1786) {
if (refDCval < 81) {
if (mu_mv_mag_sq < 92) {
if (mu_mv_mag_sq < 86) {
return 8;
}
else if (mu_mv_mag_sq >= 86) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 92) {
return 8;
}
}
else if (refDCval >= 81) {
if (Rs < 688) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (Rs >= 688) {
return 8;
}
}
}
else if (Rs >= 1786) {
if (ssDCval < 95) {
if (posBalance < 51) {
if (Cs < 486) {
return 1;
}
else if (Cs >= 486) {
return 4;
}
}
else if (posBalance >= 51) {
return 2;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
else if (AFD >= 4) {
if (mcTcor < 94) {
if (negBalance < 59) {
if (ssDCval < 93) {
if (Rs < 520) {
if (Cs < 366) {
if (negBalance < 55) {
return 1;
}
else if (negBalance >= 55) {
return 8;
}
}
else if (Cs >= 366) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
}
}
else if (Rs >= 520) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
}
else if (ssDCval >= 93) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (ssDCval < 110) {
return 8;
}
else if (ssDCval >= 110) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 59) {
if (SC < 638) {
return 1;
}
else if (SC >= 638) {
if (tcor < 89) {
if (diffRsCsdiff < -19) {
return 8;
}
else if (diffRsCsdiff >= -19) {
return 2;
}
}
else if (tcor >= 89) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (mu_mv_mag_sq < 7) {
if (tcor < 98) {
if (MVDiff < 19) {
if (Cs < 366) {
if (negBalance < 61) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
if (Cs < 302) {
return 8;
}
else if (Cs >= 302) {
if (Cs < 346) {
return 1;
}
else if (Cs >= 346) {
if (Rs < 588) {
if (Cs < 361) {
return 8;
}
else if (Cs >= 361) {
return 1;
}
}
else if (Rs >= 588) {
return 8;
}
}
}
}
}
else if (negBalance >= 61) {
return 2;
}
}
else if (Cs >= 366) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
else if (MVDiff >= 19) {
if (SC < 775) {
if (Rs < 360) {
return 4;
}
else if (Rs >= 360) {
return 2;
}
}
else if (SC >= 775) {
return 8;
}
}
}
else if (tcor >= 98) {
if (Rs < 1017) {
return 1;
}
else if (Rs >= 1017) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (negBalance < 41) {
if (Rs < 714) {
if (ssDCval < 83) {
if (AFD < 5) {
if (Rs < 500) {
return 8;
}
else if (Rs >= 500) {
return 4;
}
}
else if (AFD >= 5) {
return 2;
}
}
else if (ssDCval >= 83) {
if (Cs < 376) {
return 4;
}
else if (Cs >= 376) {
return 1;
}
}
}
else if (Rs >= 714) {
if (SC < 1870) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (SC >= 1870) {
if (Rs < 1929) {
if (Cs < 314) {
return 8;
}
else if (Cs >= 314) {
return 1;
}
}
else if (Rs >= 1929) {
if (Rs < 2107) {
return 8;
}
else if (Rs >= 2107) {
return 2;
}
}
}
}
}
else if (negBalance >= 41) {
if (posBalance < 54) {
if (refDCval < 84) {
if (diffMVdiffVal < 0) {
return 4;
}
else if (diffMVdiffVal >= 0) {
if (Cs < 336) {
return 4;
}
else if (Cs >= 336) {
if (Rs < 1413) {
if (Rs < 420) {
return 8;
}
else if (Rs >= 420) {
if (Rs < 849) {
return 1;
}
else if (Rs >= 849) {
return 8;
}
}
}
else if (Rs >= 1413) {
return 1;
}
}
}
}
else if (refDCval >= 84) {
if (diffRsCsdiff < -6) {
return 4;
}
else if (diffRsCsdiff >= -6) {
if (Rs < 1697) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (Rs >= 1697) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
if (refDCval < 92) {
if (mu_mv_mag_sq < 51) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 51) {
return 8;
}
}
else if (refDCval >= 92) {
if (Cs < 447) {
if (gchDC < 1) {
if (Cs < 405) {
return 1;
}
else if (Cs >= 405) {
return 2;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Cs >= 447) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 115) {
if (mcTcor < 99) {
if (refDCval < 128) {
if (mcTcor < 98) {
if (posBalance < 58) {
if (MVDiff < 6) {
if (Rs < 343) {
return 4;
}
else if (Rs >= 343) {
return 8;
}
}
else if (MVDiff >= 6) {
if (mcTcor < 97) {
return 2;
}
else if (mcTcor >= 97) {
if (TSC < 5) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
}
else if (posBalance >= 58) {
return 4;
}
}
else if (mcTcor >= 98) {
if (Rs < 797) {
return 2;
}
else if (Rs >= 797) {
return 1;
}
}
}
else if (refDCval >= 128) {
if (AFD < 6) {
if (mu_mv_mag_sq < 12) {
if (Rs < 420) {
return 4;
}
else if (Rs >= 420) {
return 2;
}
}
else if (mu_mv_mag_sq >= 12) {
if (ssDCval < 160) {
if (Rs < 917) {
if (TSC < 2) {
if (Cs < 398) {
return 1;
}
else if (Cs >= 398) {
return 8;
}
}
else if (TSC >= 2) {
return 8;
}
}
else if (Rs >= 917) {
if (Rs < 1414) {
return 1;
}
else if (Rs >= 1414) {
return 4;
}
}
}
else if (ssDCval >= 160) {
return 4;
}
}
}
else if (AFD >= 6) {
if (diffRsCsdiff < -4) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
if (Cs < 400) {
return 4;
}
else if (Cs >= 400) {
return 8;
}
}
}
else if (diffRsCsdiff >= -4) {
if (negBalance < 63) {
if (Cs < 341) {
return 2;
}
else if (Cs >= 341) {
if (negBalance < 39) {
if (Rs < 762) {
return 2;
}
else if (Rs >= 762) {
return 1;
}
}
else if (negBalance >= 39) {
return 1;
}
}
}
else if (negBalance >= 63) {
if (Rs < 392) {
return 2;
}
else if (Rs >= 392) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (Cs < 366) {
if (Cs < 333) {
return 1;
}
else if (Cs >= 333) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
}
else if (Cs >= 366) {
if (refDCval < 119) {
if (ssDCval < 116) {
if (Rs < 516) {
return 1;
}
else if (Rs >= 516) {
return 2;
}
}
else if (ssDCval >= 116) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (refDCval >= 119) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (ssDCval < 154) {
return 1;
}
else if (ssDCval >= 154) {
return 4;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (mu_mv_mag_sq < 81) {
if (tcor < 74) {
if (posBalance < 53) {
if (SC < 1964) {
if (Cs < 347) {
if (diffMVdiffVal < 22) {
if (negBalance < 57) {
if (mu_mv_mag_sq < 31) {
return 1;
}
else if (mu_mv_mag_sq >= 31) {
if (diffTSC < -1) {
if (Cs < 173) {
return 1;
}
else if (Cs >= 173) {
if (ssDCval < 47) {
return 2;
}
else if (ssDCval >= 47) {
if (Rs < 739) {
return 8;
}
else if (Rs >= 739) {
return 1;
}
}
}
}
else if (diffTSC >= -1) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (RsCsDiff < 128) {
return 1;
}
else if (RsCsDiff >= 128) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 57) {
if (MVDiff < 44) {
return 2;
}
else if (MVDiff >= 44) {
if (RsCsDiff < 69) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
else if (RsCsDiff >= 69) {
if (Rs < 1161) {
return 2;
}
else if (Rs >= 1161) {
if (Rs < 1593) {
return 1;
}
else if (Rs >= 1593) {
return 2;
}
}
}
}
}
}
else if (diffMVdiffVal >= 22) {
if (Cs < 242) {
return 1;
}
else if (Cs >= 242) {
if (Cs < 279) {
return 8;
}
else if (Cs >= 279) {
return 1;
}
}
}
}
else if (Cs >= 347) {
if (Rs < 896) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (refDCval < 92) {
if (diffTSC < -1) {
if (Rs < 782) {
return 2;
}
else if (Rs >= 782) {
return 4;
}
}
else if (diffTSC >= -1) {
if (diffRsCsdiff < -2) {
if (Rs < 635) {
return 4;
}
else if (Rs >= 635) {
return 8;
}
}
else if (diffRsCsdiff >= -2) {
if (negBalance < 67) {
if (Cs < 360) {
return 8;
}
else if (Cs >= 360) {
return 1;
}
}
else if (negBalance >= 67) {
return 2;
}
}
}
}
else if (refDCval >= 92) {
if (tcor < 49) {
if (TSC < 20) {
return 1;
}
else if (TSC >= 20) {
return 4;
}
}
else if (tcor >= 49) {
if (refDCval < 96) {
return 1;
}
else if (refDCval >= 96) {
return 8;
}
}
}
}
}
else if (Rs >= 896) {
if (Cs < 485) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (Cs >= 485) {
return 8;
}
}
}
}
else if (SC >= 1964) {
if (MVDiff < 35) {
if (Cs < 483) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
if (Cs < 327) {
return 2;
}
else if (Cs >= 327) {
return 1;
}
}
}
else if (Cs >= 483) {
return 1;
}
}
else if (MVDiff >= 35) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
if (MVDiff < 123) {
return 1;
}
else if (MVDiff >= 123) {
if (MVDiff < 167) {
return 8;
}
else if (MVDiff >= 167) {
return 1;
}
}
}
}
}
}
}
else if (posBalance >= 53) {
if (Cs < 366) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 50) {
if (RsCsDiff < 54) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
if (RsDiff < 6) {
if (Rs < 1602) {
return 8;
}
else if (Rs >= 1602) {
return 1;
}
}
else if (RsDiff >= 6) {
return 1;
}
}
}
else if (RsCsDiff >= 54) {
if (mu_mv_mag_sq < 13) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 2151) {
return 1;
}
else if (Rs >= 2151) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
return 2;
}
}
else if (negBalance >= 44) {
if (diffAFD < 7) {
if (diffRsCsdiff < 20) {
if (mcTcor < 96) {
if (MVDiff < 39) {
if (SC < 1023) {
if (ssDCval < 105) {
if (diffAFD < 2) {
if (mcTcor < 93) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
if (Rs < 619) {
return 8;
}
else if (Rs >= 619) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 492) {
return 8;
}
else if (Rs >= 492) {
return 2;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 620) {
return 4;
}
else if (Rs >= 620) {
return 1;
}
}
}
else if (ssDCval >= 105) {
return 1;
}
}
else if (SC >= 1023) {
if (negBalance < 53) {
if (posBalance < 58) {
if (Rs < 1625) {
return 1;
}
else if (Rs >= 1625) {
return 4;
}
}
else if (posBalance >= 58) {
if (ssDCval < 114) {
return 2;
}
else if (ssDCval >= 114) {
return 1;
}
}
}
else if (negBalance >= 53) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
if (Cs < 184) {
return 8;
}
else if (Cs >= 184) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (diffMVdiffVal < 8) {
return 4;
}
else if (diffMVdiffVal >= 8) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 39) {
if (tcor < 36) {
if (mcTcor < 54) {
if (Rs < 866) {
return 8;
}
else if (Rs >= 866) {
return 2;
}
}
else if (mcTcor >= 54) {
if (tcor < 23) {
if (mu_mv_mag_sq < 33) {
return 1;
}
else if (mu_mv_mag_sq >= 33) {
if (ssDCval < 120) {
return 4;
}
else if (ssDCval >= 120) {
return 1;
}
}
}
else if (tcor >= 23) {
return 4;
}
}
}
else if (tcor >= 36) {
if (tcor < 68) {
if (gchDC < 3) {
if (ssDCval < 110) {
if (ssDCval < 80) {
if (tcor < 63) {
if (mcTcor < 81) {
return 2;
}
else if (mcTcor >= 81) {
if (Rs < 529) {
return 8;
}
else if (Rs >= 529) {
return 4;
}
}
}
else if (tcor >= 63) {
return 1;
}
}
else if (ssDCval >= 80) {
return 2;
}
}
else if (ssDCval >= 110) {
if (ssDCval < 140) {
return 8;
}
else if (ssDCval >= 140) {
if (Rs < 642) {
return 4;
}
else if (Rs >= 642) {
return 1;
}
}
}
}
else if (gchDC >= 3) {
if (MVDiff < 167) {
if (gchDC < 4) {
if (Rs < 973) {
return 2;
}
else if (Rs >= 973) {
return 1;
}
}
else if (gchDC >= 4) {
if (Cs < 312) {
return 1;
}
else if (Cs >= 312) {
return 4;
}
}
}
else if (MVDiff >= 167) {
return 2;
}
}
}
else if (tcor >= 68) {
if (MVDiff < 78) {
if (Cs < 289) {
return 4;
}
else if (Cs >= 289) {
return 1;
}
}
else if (MVDiff >= 78) {
if (Cs < 223) {
return 1;
}
else if (Cs >= 223) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (diffRsCsdiff < -101) {
if (Rs < 633) {
return 8;
}
else if (Rs >= 633) {
return 1;
}
}
else if (diffRsCsdiff >= -101) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 476) {
return 2;
}
else if (Rs >= 476) {
return 1;
}
}
}
}
}
else if (diffRsCsdiff >= 20) {
if (ssDCval < 62) {
if (Rs < 483) {
return 4;
}
else if (Rs >= 483) {
return 1;
}
}
else if (ssDCval >= 62) {
return 2;
}
}
}
else if (diffAFD >= 7) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1528) {
return 8;
}
else if (Rs >= 1528) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
return 1;
}
}
}
}
else if (Cs >= 366) {
if (mu_mv_mag_sq < 5) {
if (gchDC < 2) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (Cs < 462) {
return 8;
}
else if (Cs >= 462) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (posBalance < 64) {
if (tcor < 62) {
return 1;
}
else if (tcor >= 62) {
if (MVDiff < 53) {
if (Rs < 1316) {
return 2;
}
else if (Rs >= 1316) {
return 1;
}
}
else if (MVDiff >= 53) {
return 4;
}
}
}
else if (posBalance >= 64) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Cs < 393) {
if (mu_mv_mag_sq < 21) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (Cs < 378) {
if (Rs < 848) {
return 2;
}
else if (Rs >= 848) {
return 4;
}
}
else if (Cs >= 378) {
return 1;
}
}
}
else if (Cs >= 393) {
if (Rs < 1036) {
if (negBalance < 53) {
if (AFD < 10) {
if (Cs < 432) {
return 8;
}
else if (Cs >= 432) {
return 2;
}
}
else if (AFD >= 10) {
if (posBalance < 68) {
if (Cs < 399) {
return 2;
}
else if (Cs >= 399) {
return 1;
}
}
else if (posBalance >= 68) {
if (Rs < 788) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (Cs < 469) {
return 4;
}
else if (Cs >= 469) {
return 2;
}
}
}
else if (Rs >= 788) {
if (Cs < 443) {
return 8;
}
else if (Cs >= 443) {
return 1;
}
}
}
}
}
else if (negBalance >= 53) {
if (posBalance < 62) {
if (SC < 1186) {
if (Rs < 704) {
if (Rs < 367) {
return 2;
}
else if (Rs >= 367) {
return 8;
}
}
else if (Rs >= 704) {
return 2;
}
}
else if (SC >= 1186) {
return 8;
}
}
else if (posBalance >= 62) {
if (mcTcor < 83) {
if (diffTSC < 1) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 4;
}
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (mcTcor >= 83) {
if (RsDiff < 7) {
if (ssDCval < 118) {
return 1;
}
else if (ssDCval >= 118) {
return 8;
}
}
else if (RsDiff >= 7) {
if (Rs < 718) {
return 8;
}
else if (Rs >= 718) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1036) {
if (ssDCval < 69) {
if (refDCval < 65) {
return 1;
}
else if (refDCval >= 65) {
return 4;
}
}
else if (ssDCval >= 69) {
if (mcTcor < 78) {
if (Rs < 1938) {
return 4;
}
else if (Rs >= 1938) {
return 2;
}
}
else if (mcTcor >= 78) {
if (SC < 1681) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 1160) {
return 4;
}
else if (Rs >= 1160) {
return 2;
}
}
}
else if (SC >= 1681) {
if (tcor < 33) {
if (Rs < 1772) {
return 4;
}
else if (Rs >= 1772) {
return 1;
}
}
else if (tcor >= 33) {
if (posBalance < 64) {
if (AFD < 23) {
if (Rs < 2273) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (Rs >= 2273) {
return 2;
}
}
else if (AFD >= 23) {
return 1;
}
}
else if (posBalance >= 64) {
if (Cs < 437) {
if (MVDiff < 74) {
return 1;
}
else if (MVDiff >= 74) {
return 8;
}
}
else if (Cs >= 437) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (RsDiff < 9) {
if (refDCval < 89) {
if (Cs < 420) {
if (Rs < 953) {
if (Cs < 176) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (Rs < 716) {
return 8;
}
else if (Rs >= 716) {
return 1;
}
}
}
else if (Cs >= 176) {
if (MVDiff < 28) {
if (Rs < 640) {
if (Rs < 627) {
if (Rs < 583) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
else if (Rs >= 583) {
return 1;
}
}
else if (Rs >= 627) {
return 4;
}
}
else if (Rs >= 640) {
if (tcor < 80) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
return 8;
}
}
else if (tcor >= 80) {
return 1;
}
}
}
else if (MVDiff >= 28) {
if (posBalance < 55) {
if (mcTcor < 97) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (MVDiff < 32) {
return 8;
}
else if (MVDiff >= 32) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 308) {
return 8;
}
else if (Cs >= 308) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
}
}
else if (Rs >= 953) {
if (mcTcor < 101) {
if (refDCval < 79) {
if (CsDiff < 3) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
if (refDCval < 56) {
if (Rs < 1456) {
return 1;
}
else if (Rs >= 1456) {
return 2;
}
}
else if (refDCval >= 56) {
return 1;
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 65) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (MVDiff < 60) {
return 2;
}
else if (MVDiff >= 60) {
return 1;
}
}
}
else if (refDCval >= 65) {
return 1;
}
}
}
else if (refDCval >= 79) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (MVDiff < 22) {
if (Cs < 187) {
if (Rs < 1111) {
return 1;
}
else if (Rs >= 1111) {
return 2;
}
}
else if (Cs >= 187) {
if (Rs < 2319) {
return 1;
}
else if (Rs >= 2319) {
return 2;
}
}
}
else if (MVDiff >= 22) {
if (Rs < 1242) {
if (posBalance < 61) {
return 1;
}
else if (posBalance >= 61) {
return 2;
}
}
else if (Rs >= 1242) {
return 1;
}
}
}
}
}
else if (mcTcor >= 101) {
if (Rs < 1286) {
return 4;
}
else if (Rs >= 1286) {
return 2;
}
}
}
}
else if (Cs >= 420) {
if (Rs < 639) {
if (Cs < 487) {
if (Cs < 456) {
if (Rs < 475) {
return 2;
}
else if (Rs >= 475) {
return 8;
}
}
else if (Cs >= 456) {
return 4;
}
}
else if (Cs >= 487) {
return 8;
}
}
else if (Rs >= 639) {
if (refDCval < 65) {
if (ssDCval < 63) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (Rs < 941) {
if (Rs < 763) {
return 8;
}
else if (Rs >= 763) {
return 2;
}
}
else if (Rs >= 941) {
if (Rs < 1887) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 2;
}
}
else if (Rs >= 1887) {
if (Rs < 2156) {
return 2;
}
else if (Rs >= 2156) {
return 4;
}
}
}
}
}
else if (ssDCval >= 63) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (refDCval >= 65) {
if (diffRsCsdiff < -71) {
return 2;
}
else if (diffRsCsdiff >= -71) {
if (mcTcor < 90) {
if (Cs < 455) {
return 4;
}
else if (Cs >= 455) {
if (Rs < 861) {
return 1;
}
else if (Rs >= 861) {
return 2;
}
}
}
else if (mcTcor >= 90) {
if (tcor < 96) {
if (Cs < 426) {
if (Rs < 2195) {
return 1;
}
else if (Rs >= 2195) {
return 4;
}
}
else if (Cs >= 426) {
return 1;
}
}
else if (tcor >= 96) {
if (Rs < 1859) {
if (Rs < 1764) {
return 2;
}
else if (Rs >= 1764) {
return 4;
}
}
else if (Rs >= 1859) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 89) {
if (AFD < 10) {
if (negBalance < 38) {
if (SC < 1926) {
return 1;
}
else if (SC >= 1926) {
if (ssDCval < 110) {
if (ssDCval < 101) {
if (Cs < 278) {
return 2;
}
else if (Cs >= 278) {
return 1;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (ssDCval >= 110) {
return 2;
}
}
}
else if (negBalance >= 38) {
if (posBalance < 57) {
if (tcor < 95) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
if (Rs < 1425) {
if (Rs < 583) {
return 8;
}
else if (Rs >= 583) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
return 4;
}
}
}
else if (Rs >= 1425) {
return 2;
}
}
}
}
else if (tcor >= 95) {
if (Cs < 116) {
return 2;
}
else if (Cs >= 116) {
if (mu_mv_mag_sq < 62) {
return 1;
}
else if (mu_mv_mag_sq >= 62) {
return 4;
}
}
}
}
else if (posBalance >= 57) {
if (ssDCval < 99) {
if (Cs < 308) {
return 2;
}
else if (Cs >= 308) {
if (Rs < 1320) {
return 8;
}
else if (Rs >= 1320) {
return 1;
}
}
}
else if (ssDCval >= 99) {
if (Rs < 2269) {
return 4;
}
else if (Rs >= 2269) {
return 1;
}
}
}
}
}
else if (AFD >= 10) {
if (tcor < 91) {
if (TSC < 7) {
if (ssDCval < 139) {
if (mu_mv_mag_sq < 17) {
if (RsDiff < 5) {
if (ssDCval < 100) {
return 1;
}
else if (ssDCval >= 100) {
if (Cs < 410) {
return 8;
}
else if (Cs >= 410) {
return 4;
}
}
}
else if (RsDiff >= 5) {
if (Rs < 1315) {
if (CsDiff < 2) {
if (Cs < 108) {
return 1;
}
else if (Cs >= 108) {
return 2;
}
}
else if (CsDiff >= 2) {
if (Cs < 484) {
return 1;
}
else if (Cs >= 484) {
return 2;
}
}
}
else if (Rs >= 1315) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
if (Rs < 1532) {
return 8;
}
else if (Rs >= 1532) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (negBalance < 74) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (negBalance >= 74) {
return 2;
}
}
}
else if (ssDCval >= 139) {
if (AFD < 16) {
if (Rs < 725) {
return 8;
}
else if (Rs >= 725) {
return 4;
}
}
else if (AFD >= 16) {
if (Cs < 314) {
return 1;
}
else if (Cs >= 314) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (Rs < 721) {
if (diffMVdiffVal < 25) {
if (AFD < 14) {
if (Cs < 230) {
return 1;
}
else if (Cs >= 230) {
return 2;
}
}
else if (AFD >= 14) {
if (SC < 917) {
if (mu_mv_mag_sq < 17) {
return 1;
}
else if (mu_mv_mag_sq >= 17) {
if (Rs < 353) {
return 1;
}
else if (Rs >= 353) {
return 8;
}
}
}
else if (SC >= 917) {
return 8;
}
}
}
else if (diffMVdiffVal >= 25) {
return 2;
}
}
else if (Rs >= 721) {
if (mcTcor < 95) {
if (SC < 1939) {
if (MVDiff < 36) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (MVDiff >= 36) {
if (Cs < 318) {
return 1;
}
else if (Cs >= 318) {
if (Rs < 1230) {
return 8;
}
else if (Rs >= 1230) {
return 1;
}
}
}
}
else if (SC >= 1939) {
if (posBalance < 65) {
if (diffRsCsdiff < -70) {
if (Cs < 252) {
return 1;
}
else if (Cs >= 252) {
return 8;
}
}
else if (diffRsCsdiff >= -70) {
return 8;
}
}
else if (posBalance >= 65) {
if (Cs < 180) {
return 2;
}
else if (Cs >= 180) {
return 1;
}
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 170) {
if (diffTSC < -3) {
return 8;
}
else if (diffTSC >= -3) {
return 1;
}
}
else if (ssDCval >= 170) {
return 2;
}
}
}
}
}
else if (tcor >= 91) {
if (CsDiff < 4) {
if (AFD < 11) {
if (refDCval < 96) {
if (Rs < 1129) {
return 4;
}
else if (Rs >= 1129) {
return 2;
}
}
else if (refDCval >= 96) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (CsDiff >= 4) {
if (Cs < 462) {
return 1;
}
else if (Cs >= 462) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 111) {
if (MVDiff < 37) {
if (negBalance < 43) {
if (Cs < 186) {
return 2;
}
else if (Cs >= 186) {
if (Rs < 2101) {
if (diffMVdiffVal < -16) {
return 2;
}
else if (diffMVdiffVal >= -16) {
return 1;
}
}
else if (Rs >= 2101) {
if (ssDCval < 78) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
else if (ssDCval >= 78) {
return 1;
}
}
}
}
else if (negBalance >= 43) {
if (diffMVdiffVal < -13) {
if (Cs < 182) {
if (Cs < 118) {
return 1;
}
else if (Cs >= 118) {
return 2;
}
}
else if (Cs >= 182) {
return 1;
}
}
else if (diffMVdiffVal >= -13) {
if (SC < 2238) {
if (negBalance < 49) {
if (diffMVdiffVal < 5) {
if (Cs < 472) {
return 1;
}
else if (Cs >= 472) {
return 2;
}
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
else if (negBalance >= 49) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (diffRsCsdiff < 65) {
return 2;
}
else if (diffRsCsdiff >= 65) {
if (Rs < 1909) {
return 1;
}
else if (Rs >= 1909) {
return 2;
}
}
}
}
}
else if (SC >= 2238) {
return 2;
}
}
}
}
else if (MVDiff >= 37) {
if (Rs < 2316) {
if (negBalance < 58) {
if (tcor < 75) {
if (Rs < 1917) {
return 2;
}
else if (Rs >= 1917) {
return 4;
}
}
else if (tcor >= 75) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (RsCsDiff < 216) {
if (diffRsCsdiff < 15) {
return 1;
}
else if (diffRsCsdiff >= 15) {
if (diffAFD < 3) {
if (Cs < 75) {
return 1;
}
else if (Cs >= 75) {
return 2;
}
}
else if (diffAFD >= 3) {
return 1;
}
}
}
else if (RsCsDiff >= 216) {
if (ssDCval < 100) {
return 2;
}
else if (ssDCval >= 100) {
return 1;
}
}
}
}
}
else if (negBalance >= 58) {
if (mu_mv_mag_sq < 7) {
if (negBalance < 73) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
if (MVDiff < 67) {
return 1;
}
else if (MVDiff >= 67) {
if (Rs < 1855) {
return 8;
}
else if (Rs >= 1855) {
return 1;
}
}
}
}
else if (negBalance >= 73) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffRsCsdiff < -87) {
return 2;
}
else if (diffRsCsdiff >= -87) {
if (Rs < 764) {
return 2;
}
else if (Rs >= 764) {
return 1;
}
}
}
}
}
else if (Rs >= 2316) {
return 8;
}
}
}
else if (refDCval >= 111) {
if (CsDiff < 2) {
if (TSC < 9) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
else if (TSC >= 9) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (gchDC < 1) {
if (mcTcor < 96) {
if (AFD < 21) {
if (Cs < 299) {
return 4;
}
else if (Cs >= 299) {
return 2;
}
}
else if (AFD >= 21) {
return 1;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (gchDC >= 1) {
if (MVDiff < 61) {
if (mcTcor < 92) {
if (Rs < 1812) {
return 2;
}
else if (Rs >= 1812) {
return 1;
}
}
else if (mcTcor >= 92) {
return 1;
}
}
else if (MVDiff >= 61) {
if (AFD < 21) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (diffAFD < 2) {
if (Rs < 2166) {
return 8;
}
else if (Rs >= 2166) {
if (Rs < 2188) {
return 1;
}
else if (Rs >= 2188) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (TSC < 12) {
return 1;
}
else if (TSC >= 12) {
return 2;
}
}
}
}
else if (AFD >= 21) {
if (mu_mv_mag_sq < 66) {
return 1;
}
else if (mu_mv_mag_sq >= 66) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (SC < 1901) {
if (RsDiff < 10) {
if (mcTcor < 78) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
if (Rs < 455) {
return 8;
}
else if (Rs >= 455) {
return 1;
}
}
}
else if (mcTcor >= 78) {
if (Cs < 394) {
if (ssDCval < 84) {
if (diffTSC < -2) {
if (Rs < 524) {
return 8;
}
else if (Rs >= 524) {
return 2;
}
}
else if (diffTSC >= -2) {
if (Rs < 400) {
return 4;
}
else if (Rs >= 400) {
return 1;
}
}
}
else if (ssDCval >= 84) {
if (mu_mv_mag_sq < 238) {
if (negBalance < 54) {
if (Rs < 1086) {
if (Cs < 270) {
return 8;
}
else if (Cs >= 270) {
return 2;
}
}
else if (Rs >= 1086) {
return 4;
}
}
else if (negBalance >= 54) {
return 2;
}
}
else if (mu_mv_mag_sq >= 238) {
return 1;
}
}
}
else if (Cs >= 394) {
if (diffMVdiffVal < 7) {
if (Rs < 988) {
return 8;
}
else if (Rs >= 988) {
if (Rs < 1121) {
return 1;
}
else if (Rs >= 1121) {
return 8;
}
}
}
else if (diffMVdiffVal >= 7) {
if (Rs < 707) {
return 1;
}
else if (Rs >= 707) {
return 2;
}
}
}
}
}
else if (RsDiff >= 10) {
if (diffAFD < 12) {
if (tcor < 36) {
return 1;
}
else if (tcor >= 36) {
if (Rs < 729) {
return 1;
}
else if (Rs >= 729) {
if (Rs < 1731) {
return 2;
}
else if (Rs >= 1731) {
return 1;
}
}
}
}
else if (diffAFD >= 12) {
return 1;
}
}
}
else if (SC >= 1901) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 505) {
if (negBalance < 13) {
if (Rs < 709) {
if (ssDCval < 57) {
if (ssDCval < 32) {
if (Cs < 513) {
return 4;
}
else if (Cs >= 513) {
return 8;
}
}
else if (ssDCval >= 32) {
if (mu_mv_mag_sq < 127) {
if (Cs < 925) {
return 1;
}
else if (Cs >= 925) {
if (Cs < 1042) {
return 2;
}
else if (Cs >= 1042) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 368) {
if (Rs < 258) {
return 2;
}
else if (Rs >= 258) {
return 1;
}
}
else if (Rs >= 368) {
return 8;
}
}
}
}
else if (ssDCval >= 57) {
if (Cs < 693) {
if (RsDiff < 1) {
if (Rs < 350) {
return 4;
}
else if (Rs >= 350) {
if (Rs < 467) {
if (Rs < 464) {
return 8;
}
else if (Rs >= 464) {
return 1;
}
}
else if (Rs >= 467) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (refDCval < 150) {
if (diffAFD < 6) {
if (Cs < 638) {
return 1;
}
else if (Cs >= 638) {
if (Rs < 599) {
return 4;
}
else if (Rs >= 599) {
return 8;
}
}
}
else if (diffAFD >= 6) {
return 8;
}
}
else if (refDCval >= 150) {
return 2;
}
}
}
else if (Cs >= 693) {
if (posBalance < 32) {
return 1;
}
else if (posBalance >= 32) {
if (SC < 1430) {
if (Rs < 600) {
if (Rs < 460) {
return 4;
}
else if (Rs >= 460) {
return 8;
}
}
else if (Rs >= 600) {
if (Rs < 649) {
return 2;
}
else if (Rs >= 649) {
return 1;
}
}
}
else if (SC >= 1430) {
return 8;
}
}
}
}
}
else if (Rs >= 709) {
if (diffTSC < -2) {
if (Rs < 1590) {
if (Cs < 671) {
if (Rs < 1196) {
if (diffAFD < -11) {
if (posBalance < 3) {
if (Rs < 820) {
return 1;
}
else if (Rs >= 820) {
return 4;
}
}
else if (posBalance >= 3) {
return 8;
}
}
else if (diffAFD >= -11) {
if (Rs < 1171) {
return 2;
}
else if (Rs >= 1171) {
if (Rs < 1183) {
return 4;
}
else if (Rs >= 1183) {
return 8;
}
}
}
}
else if (Rs >= 1196) {
if (Rs < 1355) {
return 1;
}
else if (Rs >= 1355) {
return 8;
}
}
}
else if (Cs >= 671) {
if (mcTcor < 99) {
if (diffAFD < -21) {
if (Rs < 1198) {
return 8;
}
else if (Rs >= 1198) {
return 4;
}
}
else if (diffAFD >= -21) {
return 8;
}
}
else if (mcTcor >= 99) {
if (ssDCval < 85) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
else if (ssDCval >= 85) {
if (diffRsCsdiff < -57) {
if (diffAFD < -19) {
if (Rs < 1027) {
return 8;
}
else if (Rs >= 1027) {
return 4;
}
}
else if (diffAFD >= -19) {
if (Cs < 1039) {
return 8;
}
else if (Cs >= 1039) {
if (ssDCval < 129) {
return 8;
}
else if (ssDCval >= 129) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= -57) {
if (MVDiff < 6) {
if (Rs < 1207) {
return 1;
}
else if (Rs >= 1207) {
return 8;
}
}
else if (MVDiff >= 6) {
if (diffTSC < -7) {
return 8;
}
else if (diffTSC >= -7) {
if (Cs < 1014) {
if (Rs < 857) {
return 8;
}
else if (Rs >= 857) {
return 4;
}
}
else if (Cs >= 1014) {
if (Rs < 1122) {
return 8;
}
else if (Rs >= 1122) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1590) {
if (Cs < 1298) {
return 8;
}
else if (Cs >= 1298) {
if (Cs < 1314) {
if (diffMVdiffVal < -38) {
return 4;
}
else if (diffMVdiffVal >= -38) {
return 1;
}
}
else if (Cs >= 1314) {
return 8;
}
}
}
}
else if (diffTSC >= -2) {
if (Rs < 1771) {
if (Rs < 967) {
if (posBalance < 12) {
if (ssDCval < 152) {
if (Cs < 1111) {
return 8;
}
else if (Cs >= 1111) {
return 2;
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (posBalance >= 12) {
if (Cs < 1170) {
if (ssDCval < 67) {
if (SC < 1492) {
if (ssDCval < 31) {
return 8;
}
else if (ssDCval >= 31) {
if (AFD < 2) {
if (Rs < 882) {
return 2;
}
else if (Rs >= 882) {
return 8;
}
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (SC >= 1492) {
return 8;
}
}
else if (ssDCval >= 67) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (posBalance < 30) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Rs < 768) {
return 1;
}
else if (Rs >= 768) {
return 8;
}
}
}
}
else if (posBalance >= 30) {
return 8;
}
}
}
}
else if (Cs >= 1170) {
if (Rs < 837) {
return 2;
}
else if (Rs >= 837) {
return 8;
}
}
}
}
else if (Rs >= 967) {
if (Rs < 1037) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (ssDCval < 148) {
if (posBalance < 119) {
return 8;
}
else if (posBalance >= 119) {
return 4;
}
}
else if (ssDCval >= 148) {
return 2;
}
}
}
else if (Rs >= 1037) {
if (negBalance < 8) {
if (ssDCval < 107) {
if (TSCindex < 3) {
if (Rs < 1091) {
if (Rs < 1067) {
return 8;
}
else if (Rs >= 1067) {
return 4;
}
}
else if (Rs >= 1091) {
return 8;
}
}
else if (TSCindex >= 3) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 1368) {
return 8;
}
else if (Rs >= 1368) {
return 4;
}
}
}
}
}
else if (ssDCval >= 107) {
if (ssDCval < 137) {
if (posBalance < 8) {
return 1;
}
else if (posBalance >= 8) {
if (Rs < 1132) {
if (Rs < 1049) {
return 8;
}
else if (Rs >= 1049) {
return 1;
}
}
else if (Rs >= 1132) {
return 8;
}
}
}
else if (ssDCval >= 137) {
if (ssDCval < 152) {
return 8;
}
else if (ssDCval >= 152) {
return 2;
}
}
}
}
else if (negBalance >= 8) {
if (ssDCval < 66) {
if (SC < 2614) {
if (SC < 1795) {
return 8;
}
else if (SC >= 1795) {
if (Cs < 688) {
return 1;
}
else if (Cs >= 688) {
if (Cs < 959) {
if (TSC < 2) {
if (ssDCval < 64) {
if (diffMVdiffVal < -6) {
return 1;
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (TSC >= 2) {
return 1;
}
}
else if (Cs >= 959) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
return 1;
}
}
}
}
}
else if (SC >= 2614) {
return 8;
}
}
else if (ssDCval >= 66) {
if (TSCindex < 1) {
return 4;
}
else if (TSCindex >= 1) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (SC < 2099) {
return 4;
}
else if (SC >= 2099) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1771) {
if (negBalance < 5) {
if (mu_mv_mag_sq < 90) {
if (Rs < 2095) {
return 8;
}
else if (Rs >= 2095) {
return 1;
}
}
else if (mu_mv_mag_sq >= 90) {
return 8;
}
}
else if (negBalance >= 5) {
if (Rs < 2092) {
if (Cs < 986) {
if (mcTcor < 98) {
if (Cs < 671) {
return 2;
}
else if (Cs >= 671) {
return 8;
}
}
else if (mcTcor >= 98) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (Cs >= 986) {
if (refDCval < 68) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (refDCval >= 68) {
if (Rs < 1911) {
return 1;
}
else if (Rs >= 1911) {
if (Rs < 1988) {
return 8;
}
else if (Rs >= 1988) {
return 1;
}
}
}
}
}
else if (Rs >= 2092) {
if (posBalance < 32) {
if (CsDiff < 2) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 2236) {
return 8;
}
else if (Rs >= 2236) {
return 2;
}
}
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (posBalance >= 32) {
if (Rs < 2294) {
if (Rs < 2203) {
return 2;
}
else if (Rs >= 2203) {
return 8;
}
}
else if (Rs >= 2294) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 43) {
if (refDCval < 47) {
if (MVDiff < 9) {
if (RsDiff < 3) {
if (Cs < 636) {
if (Rs < 786) {
if (Rs < 699) {
return 4;
}
else if (Rs >= 699) {
return 2;
}
}
else if (Rs >= 786) {
return 1;
}
}
else if (Cs >= 636) {
if (Rs < 1423) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (Rs < 501) {
return 4;
}
else if (Rs >= 501) {
return 1;
}
}
}
else if (Rs >= 1423) {
if (Rs < 1808) {
return 4;
}
else if (Rs >= 1808) {
return 8;
}
}
}
}
else if (RsDiff >= 3) {
if (CsDiff < 5) {
if (negBalance < 32) {
if (diffMVdiffVal < -4) {
if (Rs < 926) {
if (Rs < 761) {
return 4;
}
else if (Rs >= 761) {
return 2;
}
}
else if (Rs >= 926) {
if (Rs < 1435) {
return 8;
}
else if (Rs >= 1435) {
return 1;
}
}
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
else if (negBalance >= 32) {
if (Rs < 1185) {
return 8;
}
else if (Rs >= 1185) {
return 4;
}
}
}
else if (CsDiff >= 5) {
return 1;
}
}
}
else if (MVDiff >= 9) {
if (refDCval < 27) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
return 4;
}
}
else if (refDCval >= 27) {
if (mcTcor < 88) {
if (Rs < 1748) {
return 8;
}
else if (Rs >= 1748) {
return 1;
}
}
else if (mcTcor >= 88) {
if (gchDC < 1) {
if (mu_mv_mag_sq < 81) {
if (Rs < 1999) {
return 4;
}
else if (Rs >= 1999) {
return 1;
}
}
else if (mu_mv_mag_sq >= 81) {
return 1;
}
}
else if (gchDC >= 1) {
if (Cs < 639) {
return 2;
}
else if (Cs >= 639) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 73) {
if (ssDCval < 85) {
if (Cs < 931) {
if (negBalance < 63) {
if (Rs < 1206) {
if (negBalance < 36) {
if (Rs < 804) {
if (tcor < 87) {
if (Rs < 617) {
return 1;
}
else if (Rs >= 617) {
return 2;
}
}
else if (tcor >= 87) {
return 4;
}
}
else if (Rs >= 804) {
return 1;
}
}
else if (negBalance >= 36) {
if (Cs < 725) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 1;
}
}
else if (Cs >= 725) {
if (Rs < 807) {
return 2;
}
else if (Rs >= 807) {
return 8;
}
}
}
}
else if (Rs >= 1206) {
if (ssDCval < 54) {
if (Rs < 1746) {
return 1;
}
else if (Rs >= 1746) {
return 4;
}
}
else if (ssDCval >= 54) {
if (refDCval < 58) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
else if (refDCval >= 58) {
if (posBalance < 42) {
if (negBalance < 36) {
if (gchDC < 1) {
if (Rs < 1822) {
return 2;
}
else if (Rs >= 1822) {
return 1;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (negBalance >= 36) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (posBalance >= 42) {
if (MVDiff < 7) {
if (Rs < 1791) {
return 1;
}
else if (Rs >= 1791) {
return 2;
}
}
else if (MVDiff >= 7) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 63) {
if (RsDiff < 2) {
if (Rs < 809) {
return 1;
}
else if (Rs >= 809) {
return 8;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (Cs >= 931) {
if (mcTcor < 90) {
if (Rs < 1946) {
return 8;
}
else if (Rs >= 1946) {
return 2;
}
}
else if (mcTcor >= 90) {
if (mu_mv_mag_sq < 2) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (Rs < 552) {
return 2;
}
else if (Rs >= 552) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (negBalance < 40) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (posBalance < 37) {
if (Rs < 1467) {
return 4;
}
else if (Rs >= 1467) {
return 2;
}
}
else if (posBalance >= 37) {
return 2;
}
}
}
else if (negBalance >= 40) {
if (Cs < 1218) {
return 4;
}
else if (Cs >= 1218) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (ssDCval < 153) {
if (Rs < 1635) {
if (tcor < 91) {
if (mcTcor < 95) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
if (tcor < 87) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
return 2;
}
}
else if (tcor >= 87) {
return 8;
}
}
}
else if (mcTcor >= 95) {
return 2;
}
}
else if (tcor >= 91) {
if (ssDCval < 132) {
return 8;
}
else if (ssDCval >= 132) {
return 4;
}
}
}
else if (Rs >= 1635) {
if (MVDiff < 21) {
if (SC < 2763) {
return 1;
}
else if (SC >= 2763) {
return 8;
}
}
else if (MVDiff >= 21) {
if (Rs < 1701) {
return 8;
}
else if (Rs >= 1701) {
return 2;
}
}
}
}
else if (ssDCval >= 153) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 73) {
if (ssDCval < 57) {
if (TSC < 3) {
if (Rs < 589) {
if (Rs < 442) {
return 2;
}
else if (Rs >= 442) {
return 1;
}
}
else if (Rs >= 589) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (Cs < 1007) {
return 8;
}
else if (Cs >= 1007) {
if (Rs < 1320) {
return 1;
}
else if (Rs >= 1320) {
return 4;
}
}
}
}
}
else if (TSC >= 3) {
if (mcTcor < 95) {
if (MVDiff < 19) {
if (Cs < 1042) {
if (mu_mv_mag_sq < 112) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
else if (Cs >= 1042) {
return 8;
}
}
else if (MVDiff >= 19) {
if (Rs < 1952) {
return 1;
}
else if (Rs >= 1952) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (negBalance < 37) {
if (SC < 1582) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
else if (SC >= 1582) {
if (Rs < 802) {
return 4;
}
else if (Rs >= 802) {
return 1;
}
}
}
else if (negBalance >= 37) {
return 4;
}
}
}
}
else if (ssDCval >= 57) {
if (mu_mv_mag_sq < 96) {
if (posBalance < 34) {
if (Rs < 2271) {
if (Rs < 370) {
return 2;
}
else if (Rs >= 370) {
return 8;
}
}
else if (Rs >= 2271) {
if (Rs < 2289) {
return 4;
}
else if (Rs >= 2289) {
return 8;
}
}
}
else if (posBalance >= 34) {
if (negBalance < 29) {
if (MVDiff < 2) {
if (Rs < 1274) {
return 8;
}
else if (Rs >= 1274) {
return 1;
}
}
else if (MVDiff >= 2) {
if (Cs < 1276) {
if (ssDCval < 143) {
return 4;
}
else if (ssDCval >= 143) {
return 8;
}
}
else if (Cs >= 1276) {
return 2;
}
}
}
else if (negBalance >= 29) {
if (ssDCval < 67) {
if (Rs < 963) {
if (Rs < 763) {
return 8;
}
else if (Rs >= 763) {
return 2;
}
}
else if (Rs >= 963) {
return 4;
}
}
else if (ssDCval >= 67) {
if (ssDCval < 102) {
if (negBalance < 55) {
if (tcor < 94) {
if (mu_mv_mag_sq < 77) {
return 8;
}
else if (mu_mv_mag_sq >= 77) {
return 2;
}
}
else if (tcor >= 94) {
if (Rs < 1500) {
return 8;
}
else if (Rs >= 1500) {
return 1;
}
}
}
else if (negBalance >= 55) {
return 8;
}
}
else if (ssDCval >= 102) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (CsDiff < 5) {
if (mu_mv_mag_sq < 114) {
if (ssDCval < 125) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (negBalance < 31) {
if (Cs < 969) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
if (Rs < 1989) {
return 8;
}
else if (Rs >= 1989) {
return 2;
}
}
}
}
else if (Cs >= 969) {
return 4;
}
}
else if (negBalance >= 31) {
if (Rs < 1458) {
return 8;
}
else if (Rs >= 1458) {
if (Cs < 1252) {
if (Rs < 1872) {
return 4;
}
else if (Rs >= 1872) {
if (Rs < 2262) {
return 4;
}
else if (Rs >= 2262) {
return 4;
}
}
}
else if (Cs >= 1252) {
return 8;
}
}
}
}
}
else if (ssDCval >= 125) {
return 8;
}
}
else if (mu_mv_mag_sq >= 114) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
return 1;
}
}
}
else if (CsDiff >= 5) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 43) {
if (Rs < 407) {
if (ssDCval < 96) {
if (ssDCval < 60) {
if (refDCval < 30) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (refDCval >= 30) {
if (mcTcor < 96) {
if (mcTcor < 73) {
return 2;
}
else if (mcTcor >= 73) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (Rs < 387) {
return 1;
}
else if (Rs >= 387) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 303) {
return 1;
}
else if (Rs >= 303) {
if (ssDCval < 54) {
return 2;
}
else if (ssDCval >= 54) {
if (Cs < 756) {
return 2;
}
else if (Cs >= 756) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 60) {
if (negBalance < 42) {
if (Cs < 604) {
return 4;
}
else if (Cs >= 604) {
return 8;
}
}
else if (negBalance >= 42) {
if (mu_mv_mag_sq < 44) {
return 1;
}
else if (mu_mv_mag_sq >= 44) {
if (mcTcor < 95) {
if (Cs < 908) {
return 8;
}
else if (Cs >= 908) {
return 1;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
}
else if (ssDCval >= 96) {
if (Cs < 719) {
if (ssDCval < 107) {
return 1;
}
else if (ssDCval >= 107) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 255) {
return 8;
}
else if (Rs >= 255) {
return 2;
}
}
}
}
else if (Cs >= 719) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (Rs >= 407) {
if (mu_mv_mag_sq < 44) {
if (mcTcor < 93) {
if (ssDCval < 40) {
if (CsDiff < 6) {
if (Rs < 602) {
if (Cs < 521) {
return 8;
}
else if (Cs >= 521) {
return 2;
}
}
else if (Rs >= 602) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
if (RsCsDiff < 11) {
return 8;
}
else if (RsCsDiff >= 11) {
if (mcTcor < 84) {
return 8;
}
else if (mcTcor >= 84) {
if (Cs < 649) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (Rs < 615) {
return 8;
}
else if (Rs >= 615) {
return 2;
}
}
}
else if (Cs >= 649) {
if (tcor < 85) {
return 2;
}
else if (tcor >= 85) {
if (Rs < 987) {
return 8;
}
else if (Rs >= 987) {
return 2;
}
}
}
}
}
}
}
}
else if (CsDiff >= 6) {
if (Rs < 879) {
return 1;
}
else if (Rs >= 879) {
return 8;
}
}
}
else if (ssDCval >= 40) {
if (negBalance < 34) {
if (refDCval < 50) {
if (mu_mv_mag_sq < 17) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (Rs < 1900) {
return 8;
}
else if (Rs >= 1900) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 17) {
return 1;
}
}
else if (refDCval >= 50) {
if (Rs < 1253) {
if (gchDC < 2) {
if (Cs < 869) {
return 2;
}
else if (Cs >= 869) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (ssDCval < 145) {
return 1;
}
else if (ssDCval >= 145) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (Rs >= 1253) {
if (mcTcor < 88) {
if (posBalance < 51) {
if (Cs < 938) {
return 2;
}
else if (Cs >= 938) {
return 1;
}
}
else if (posBalance >= 51) {
if (Cs < 1153) {
return 8;
}
else if (Cs >= 1153) {
return 2;
}
}
}
else if (mcTcor >= 88) {
return 2;
}
}
}
}
else if (negBalance >= 34) {
if (RsDiff < 8) {
if (Rs < 648) {
if (negBalance < 50) {
if (tcor < 72) {
if (negBalance < 48) {
if (Rs < 622) {
return 2;
}
else if (Rs >= 622) {
return 4;
}
}
else if (negBalance >= 48) {
return 4;
}
}
else if (tcor >= 72) {
if (TSC < 6) {
if (mu_mv_mag_sq < 5) {
if (refDCval < 49) {
return 8;
}
else if (refDCval >= 49) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Cs < 721) {
return 8;
}
else if (Cs >= 721) {
return 2;
}
}
}
else if (TSC >= 6) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
}
}
else if (negBalance >= 50) {
if (ssDCval < 82) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (CsDiff < 5) {
if (tcor < 50) {
return 4;
}
else if (tcor >= 50) {
return 8;
}
}
else if (CsDiff >= 5) {
return 4;
}
}
}
else if (ssDCval >= 82) {
return 8;
}
}
}
else if (Rs >= 648) {
if (Rs < 935) {
if (RsCsDiff < 48) {
if (refDCval < 49) {
if (TSC < 7) {
if (ssDCval < 45) {
if (Rs < 742) {
if (Cs < 767) {
return 4;
}
else if (Cs >= 767) {
return 2;
}
}
else if (Rs >= 742) {
return 8;
}
}
else if (ssDCval >= 45) {
if (posBalance < 60) {
if (Cs < 736) {
if (Rs < 861) {
return 1;
}
else if (Rs >= 861) {
return 8;
}
}
else if (Cs >= 736) {
return 8;
}
}
else if (posBalance >= 60) {
return 1;
}
}
}
else if (TSC >= 7) {
if (tcor < 55) {
return 4;
}
else if (tcor >= 55) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (diffRsCsdiff < -6) {
if (AFD < 10) {
if (Rs < 806) {
return 8;
}
else if (Rs >= 806) {
return 1;
}
}
else if (AFD >= 10) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
}
}
}
else if (refDCval >= 49) {
if (negBalance < 52) {
if (AFD < 9) {
if (SC < 1424) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (ssDCval < 106) {
return 2;
}
else if (ssDCval >= 106) {
return 8;
}
}
}
else if (SC >= 1424) {
return 8;
}
}
else if (AFD >= 9) {
if (MVDiff < 12) {
if (Cs < 780) {
if (Rs < 676) {
return 8;
}
else if (Rs >= 676) {
return 4;
}
}
else if (Cs >= 780) {
if (Rs < 873) {
return 8;
}
else if (Rs >= 873) {
if (Rs < 900) {
return 4;
}
else if (Rs >= 900) {
return 8;
}
}
}
}
else if (MVDiff >= 12) {
if (mu_mv_mag_sq < 11) {
if (Rs < 915) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 2;
}
}
else if (Rs >= 915) {
return 4;
}
}
else if (mu_mv_mag_sq >= 11) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
}
else if (negBalance >= 52) {
if (ssDCval < 108) {
if (mcTcor < 85) {
if (mcTcor < 74) {
return 8;
}
else if (mcTcor >= 74) {
if (ssDCval < 74) {
if (refDCval < 50) {
return 8;
}
else if (refDCval >= 50) {
return 4;
}
}
else if (ssDCval >= 74) {
return 8;
}
}
}
else if (mcTcor >= 85) {
return 8;
}
}
else if (ssDCval >= 108) {
return 4;
}
}
}
}
else if (RsCsDiff >= 48) {
if (ssDCval < 45) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
else if (ssDCval >= 45) {
if (Cs < 814) {
return 8;
}
else if (Cs >= 814) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
return 4;
}
}
}
}
}
else if (Rs >= 935) {
if (tcor < 80) {
if (ssDCval < 124) {
if (Cs < 1180) {
if (refDCval < 70) {
if (Rs < 1120) {
if (Rs < 1108) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (diffAFD < 2) {
if (Rs < 973) {
if (Rs < 956) {
return 8;
}
else if (Rs >= 956) {
return 1;
}
}
else if (Rs >= 973) {
return 8;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
}
else if (Rs >= 1108) {
if (Cs < 940) {
return 8;
}
else if (Cs >= 940) {
return 1;
}
}
}
else if (Rs >= 1120) {
return 8;
}
}
else if (refDCval >= 70) {
if (Rs < 1313) {
if (RsCsDiff < 38) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (Cs < 689) {
return 8;
}
else if (Cs >= 689) {
return 2;
}
}
}
else if (RsCsDiff >= 38) {
if (Rs < 1210) {
return 2;
}
else if (Rs >= 1210) {
return 8;
}
}
}
else if (Rs >= 1313) {
if (SC < 2400) {
if (Cs < 936) {
if (ssDCval < 107) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (ssDCval >= 107) {
return 8;
}
}
else if (Cs >= 936) {
return 8;
}
}
else if (SC >= 2400) {
if (Cs < 1151) {
return 8;
}
else if (Cs >= 1151) {
return 2;
}
}
}
}
}
else if (Cs >= 1180) {
if (negBalance < 47) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (negBalance >= 47) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 124) {
if (mu_mv_mag_sq < 2) {
if (Cs < 1307) {
return 8;
}
else if (Cs >= 1307) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1281) {
return 1;
}
else if (Rs >= 1281) {
return 4;
}
}
}
}
else if (tcor >= 80) {
if (RsCsDiff < 23) {
if (mcTcor < 92) {
if (ssDCval < 150) {
if (mu_mv_mag_sq < 43) {
return 8;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 1201) {
return 4;
}
else if (Rs >= 1201) {
return 8;
}
}
}
else if (ssDCval >= 150) {
return 2;
}
}
else if (mcTcor >= 92) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
return 4;
}
}
}
}
else if (RsCsDiff >= 23) {
if (Rs < 1185) {
if (Cs < 840) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Rs < 1097) {
return 4;
}
else if (Rs >= 1097) {
return 1;
}
}
}
else if (Cs >= 840) {
if (RsDiff < 6) {
if (MVDiff < 11) {
if (SC < 2161) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (AFD < 8) {
if (Rs < 1020) {
return 4;
}
else if (Rs >= 1020) {
return 8;
}
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (SC >= 2161) {
return 4;
}
}
else if (MVDiff >= 11) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (Rs < 1063) {
return 1;
}
else if (Rs >= 1063) {
if (Rs < 1130) {
return 2;
}
else if (Rs >= 1130) {
return 4;
}
}
}
}
}
else if (Rs >= 1185) {
if (Cs < 804) {
if (refDCval < 72) {
if (SC < 2347) {
return 8;
}
else if (SC >= 2347) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (refDCval >= 72) {
if (RsDiff < 6) {
if (Cs < 628) {
return 2;
}
else if (Cs >= 628) {
return 8;
}
}
else if (RsDiff >= 6) {
if (diffAFD < -2) {
if (Cs < 719) {
return 4;
}
else if (Cs >= 719) {
return 2;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
}
}
else if (Cs >= 804) {
if (ssDCval < 149) {
if (diffMVdiffVal < -4) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (diffMVdiffVal >= -4) {
if (TSC < 11) {
if (posBalance < 58) {
if (ssDCval < 73) {
if (AFD < 11) {
if (ssDCval < 64) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (Cs < 882) {
if (Rs < 1557) {
return 8;
}
else if (Rs >= 1557) {
return 4;
}
}
else if (Cs >= 882) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (AFD >= 11) {
if (Rs < 1558) {
return 4;
}
else if (Rs >= 1558) {
if (Rs < 1594) {
return 8;
}
else if (Rs >= 1594) {
return 1;
}
}
}
}
else if (ssDCval >= 73) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
}
else if (posBalance >= 58) {
if (RsCsDiff < 60) {
return 8;
}
else if (RsCsDiff >= 60) {
if (Rs < 1491) {
if (SC < 2561) {
return 4;
}
else if (SC >= 2561) {
return 8;
}
}
else if (Rs >= 1491) {
return 8;
}
}
}
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (ssDCval >= 149) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (Rs < 1778) {
if (tcor < 58) {
if (RsCsDiff < 129) {
return 8;
}
else if (RsCsDiff >= 129) {
return 4;
}
}
else if (tcor >= 58) {
if (tcor < 84) {
if (Rs < 1479) {
if (Rs < 1430) {
return 8;
}
else if (Rs >= 1430) {
return 2;
}
}
else if (Rs >= 1479) {
return 8;
}
}
else if (tcor >= 84) {
if (MVDiff < 20) {
if (Cs < 869) {
return 4;
}
else if (Cs >= 869) {
return 8;
}
}
else if (MVDiff >= 20) {
return 1;
}
}
}
}
else if (Rs >= 1778) {
if (ssDCval < 70) {
if (TSC < 8) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (TSC >= 8) {
return 4;
}
}
else if (ssDCval >= 70) {
if (ssDCval < 126) {
if (Rs < 2240) {
if (ssDCval < 107) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 2;
}
}
else if (ssDCval >= 107) {
return 8;
}
}
else if (Rs >= 2240) {
return 4;
}
}
else if (ssDCval >= 126) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (RsDiff < 8) {
if (Cs < 729) {
if (Rs < 1341) {
if (negBalance < 43) {
if (ssDCval < 79) {
if (MVDiff < 7) {
if (SC < 1338) {
return 4;
}
else if (SC >= 1338) {
if (diffAFD < 1) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (MVDiff >= 7) {
if (AFD < 10) {
if (tcor < 91) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
return 2;
}
}
}
else if (tcor >= 91) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 2;
}
}
}
else if (AFD >= 10) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (diffRsCsdiff < -3) {
if (Rs < 762) {
return 1;
}
else if (Rs >= 762) {
return 2;
}
}
else if (diffRsCsdiff >= -3) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 79) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (Rs < 839) {
return 8;
}
else if (Rs >= 839) {
return 1;
}
}
}
}
else if (negBalance >= 43) {
if (MVDiff < 41) {
if (negBalance < 62) {
if (CsDiff < 3) {
if (ssDCval < 79) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (diffRsCsdiff < -13) {
return 1;
}
else if (diffRsCsdiff >= -13) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 79) {
if (ssDCval < 114) {
if (refDCval < 111) {
if (tcor < 76) {
return 8;
}
else if (tcor >= 76) {
if (Cs < 544) {
if (MVDiff < 5) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 1;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (Cs >= 544) {
return 1;
}
}
}
else if (refDCval >= 111) {
return 2;
}
}
else if (ssDCval >= 114) {
if (Rs < 960) {
return 8;
}
else if (Rs >= 960) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (negBalance < 50) {
if (gchDC < 1) {
if (Cs < 579) {
if (Rs < 680) {
return 8;
}
else if (Rs >= 680) {
return 1;
}
}
else if (Cs >= 579) {
if (Rs < 1131) {
return 4;
}
else if (Rs >= 1131) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (Rs < 1098) {
return 2;
}
else if (Rs >= 1098) {
if (Rs < 1226) {
return 8;
}
else if (Rs >= 1226) {
return 1;
}
}
}
}
else if (negBalance >= 50) {
if (diffTSC < -1) {
if (Rs < 835) {
return 8;
}
else if (Rs >= 835) {
return 2;
}
}
else if (diffTSC >= -1) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 62) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
return 2;
}
}
}
else if (MVDiff >= 41) {
return 2;
}
}
}
else if (Rs >= 1341) {
if (posBalance < 58) {
if (ssDCval < 72) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
if (Rs < 1555) {
return 1;
}
else if (Rs >= 1555) {
return 4;
}
}
}
}
else if (ssDCval >= 72) {
if (MVDiff < 30) {
return 1;
}
else if (MVDiff >= 30) {
return 2;
}
}
}
else if (posBalance >= 58) {
if (Cs < 686) {
if (tcor < 92) {
if (SC < 1933) {
return 2;
}
else if (SC >= 1933) {
return 1;
}
}
else if (tcor >= 92) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (Cs >= 686) {
if (Rs < 1433) {
return 4;
}
else if (Rs >= 1433) {
return 8;
}
}
}
}
}
else if (Cs >= 729) {
if (Rs < 1052) {
if (MVDiff < 5) {
if (ssDCval < 72) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
if (diffRsCsdiff < -22) {
return 1;
}
else if (diffRsCsdiff >= -22) {
if (RsDiff < 4) {
if (Rs < 1037) {
if (Cs < 773) {
if (ssDCval < 46) {
return 4;
}
else if (ssDCval >= 46) {
return 8;
}
}
else if (Cs >= 773) {
return 8;
}
}
else if (Rs >= 1037) {
return 4;
}
}
else if (RsDiff >= 4) {
if (posBalance < 58) {
if (Rs < 1000) {
return 4;
}
else if (Rs >= 1000) {
return 2;
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
}
else if (ssDCval >= 72) {
if (Rs < 801) {
return 1;
}
else if (Rs >= 801) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
else if (MVDiff >= 5) {
if (AFD < 8) {
if (tcor < 90) {
if (MVDiff < 24) {
if (Rs < 970) {
return 2;
}
else if (Rs >= 970) {
return 4;
}
}
else if (MVDiff >= 24) {
return 8;
}
}
else if (tcor >= 90) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (diffRsCsdiff < -6) {
if (Rs < 871) {
return 2;
}
else if (Rs >= 871) {
return 4;
}
}
else if (diffRsCsdiff >= -6) {
return 4;
}
}
}
}
else if (AFD >= 8) {
if (tcor < 83) {
if (ssDCval < 42) {
return 2;
}
else if (ssDCval >= 42) {
return 8;
}
}
else if (tcor >= 83) {
if (ssDCval < 51) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (ssDCval >= 51) {
if (mu_mv_mag_sq < 15) {
if (posBalance < 59) {
if (Rs < 947) {
if (ssDCval < 71) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (Rs < 941) {
if (Rs < 890) {
return 1;
}
else if (Rs >= 890) {
return 4;
}
}
else if (Rs >= 941) {
return 2;
}
}
}
else if (ssDCval >= 71) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
return 2;
}
}
}
else if (Rs >= 947) {
if (Rs < 979) {
return 1;
}
else if (Rs >= 979) {
if (Rs < 1001) {
return 4;
}
else if (Rs >= 1001) {
return 2;
}
}
}
}
else if (posBalance >= 59) {
if (ssDCval < 84) {
if (Rs < 951) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (Rs >= 951) {
return 2;
}
}
else if (ssDCval >= 84) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (ssDCval < 71) {
if (Rs < 829) {
return 2;
}
else if (Rs >= 829) {
return 8;
}
}
else if (ssDCval >= 71) {
if (ssDCval < 156) {
return 1;
}
else if (ssDCval >= 156) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 1052) {
if (ssDCval < 136) {
if (ssDCval < 105) {
if (refDCval < 63) {
if (Cs < 1045) {
if (negBalance < 36) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (Rs < 1155) {
return 1;
}
else if (Rs >= 1155) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (diffAFD < 5) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 2;
}
}
else if (diffAFD >= 5) {
return 1;
}
}
}
else if (Cs >= 1045) {
if (tcor < 89) {
if (TSC < 5) {
if (negBalance < 32) {
if (Rs < 2102) {
return 2;
}
else if (Rs >= 2102) {
return 4;
}
}
else if (negBalance >= 32) {
return 8;
}
}
else if (TSC >= 5) {
return 4;
}
}
else if (tcor >= 89) {
if (Cs < 1052) {
return 4;
}
else if (Cs >= 1052) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 63) {
if (TSC < 8) {
if (MVDiff < 6) {
if (negBalance < 44) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (negBalance >= 44) {
if (refDCval < 73) {
return 8;
}
else if (refDCval >= 73) {
if (negBalance < 50) {
if (Rs < 1157) {
return 2;
}
else if (Rs >= 1157) {
return 8;
}
}
else if (negBalance >= 50) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (Rs < 1824) {
if (CsDiff < 5) {
if (AFD < 10) {
if (Rs < 1307) {
return 8;
}
else if (Rs >= 1307) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
if (Rs < 1702) {
return 4;
}
else if (Rs >= 1702) {
return 2;
}
}
}
}
else if (AFD >= 10) {
if (Cs < 1245) {
return 8;
}
else if (Cs >= 1245) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (posBalance < 55) {
if (Rs < 1571) {
return 4;
}
else if (Rs >= 1571) {
return 8;
}
}
else if (posBalance >= 55) {
return 2;
}
}
}
else if (Rs >= 1824) {
if (Cs < 838) {
if (Cs < 810) {
return 2;
}
else if (Cs >= 810) {
return 1;
}
}
else if (Cs >= 838) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 4;
}
}
}
}
}
else if (TSC >= 8) {
if (diffRsCsdiff < 15) {
return 8;
}
else if (diffRsCsdiff >= 15) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
}
else if (ssDCval >= 105) {
if (posBalance < 67) {
if (RsCsDiff < 60) {
return 8;
}
else if (RsCsDiff >= 60) {
if (Rs < 1532) {
if (Cs < 962) {
return 8;
}
else if (Cs >= 962) {
return 2;
}
}
else if (Rs >= 1532) {
return 8;
}
}
}
else if (posBalance >= 67) {
if (Rs < 1721) {
return 8;
}
else if (Rs >= 1721) {
return 2;
}
}
}
}
else if (ssDCval >= 136) {
if (Rs < 1478) {
if (mu_mv_mag_sq < 15) {
return 2;
}
else if (mu_mv_mag_sq >= 15) {
return 4;
}
}
else if (Rs >= 1478) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (ssDCval < 72) {
if (Rs < 1519) {
return 1;
}
else if (Rs >= 1519) {
return 2;
}
}
else if (ssDCval >= 72) {
if (Rs < 1866) {
if (tcor < 90) {
return 2;
}
else if (tcor >= 90) {
return 8;
}
}
else if (Rs >= 1866) {
if (tcor < 87) {
if (Cs < 685) {
return 1;
}
else if (Cs >= 685) {
return 8;
}
}
else if (tcor >= 87) {
if (Rs < 2315) {
return 1;
}
else if (Rs >= 2315) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (refDCval < 79) {
if (negBalance < 45) {
if (refDCval < 59) {
if (Rs < 851) {
if (refDCval < 37) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (refDCval >= 37) {
if (ssDCval < 56) {
if (refDCval < 42) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (Rs < 608) {
return 8;
}
else if (Rs >= 608) {
return 2;
}
}
}
else if (refDCval >= 42) {
if (Rs < 581) {
if (Cs < 883) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (ssDCval < 54) {
if (Cs < 628) {
return 4;
}
else if (Cs >= 628) {
return 8;
}
}
else if (ssDCval >= 54) {
if (Rs < 472) {
return 8;
}
else if (Rs >= 472) {
return 1;
}
}
}
}
else if (Cs >= 883) {
if (Rs < 412) {
return 2;
}
else if (Rs >= 412) {
return 1;
}
}
}
else if (Rs >= 581) {
if (Cs < 529) {
if (Cs < 524) {
return 8;
}
else if (Cs >= 524) {
return 1;
}
}
else if (Cs >= 529) {
if (Rs < 691) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 634) {
return 4;
}
else if (Rs >= 634) {
return 1;
}
}
}
else if (Rs >= 691) {
if (Cs < 917) {
if (mu_mv_mag_sq < 57) {
return 8;
}
else if (mu_mv_mag_sq >= 57) {
if (ssDCval < 51) {
if (Rs < 796) {
return 4;
}
else if (Rs >= 796) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (ssDCval >= 51) {
if (diffMVdiffVal < 4) {
return 4;
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
}
}
else if (Cs >= 917) {
if (Cs < 962) {
return 8;
}
else if (Cs >= 962) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 56) {
if (MVDiff < 2) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
return 8;
}
}
else if (MVDiff >= 2) {
if (mcTcor < 83) {
return 8;
}
else if (mcTcor >= 83) {
if (AFD < 5) {
if (refDCval < 58) {
if (posBalance < 54) {
if (Rs < 591) {
return 1;
}
else if (Rs >= 591) {
return 8;
}
}
else if (posBalance >= 54) {
return 4;
}
}
else if (refDCval >= 58) {
return 2;
}
}
else if (AFD >= 5) {
if (Cs < 1084) {
return 1;
}
else if (Cs >= 1084) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 851) {
if (refDCval < 39) {
return 8;
}
else if (refDCval >= 39) {
if (Cs < 929) {
if (Rs < 1503) {
if (tcor < 89) {
if (posBalance < 57) {
if (Rs < 1120) {
if (mu_mv_mag_sq < 60) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
return 8;
}
}
else if (mu_mv_mag_sq >= 60) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (tcor < 84) {
if (SC < 1919) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 1021) {
return 1;
}
else if (Rs >= 1021) {
return 8;
}
}
}
else if (SC >= 1919) {
return 1;
}
}
else if (tcor >= 84) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
}
}
else if (Rs >= 1120) {
if (mu_mv_mag_sq < 104) {
if (negBalance < 32) {
if (Cs < 540) {
return 2;
}
else if (Cs >= 540) {
return 1;
}
}
else if (negBalance >= 32) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (AFD < 8) {
if (Rs < 1147) {
return 2;
}
else if (Rs >= 1147) {
return 8;
}
}
else if (AFD >= 8) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Cs < 706) {
return 8;
}
else if (Cs >= 706) {
return 1;
}
}
}
}
else if (posBalance >= 57) {
if (ssDCval < 56) {
if (diffRsCsdiff < -17) {
if (Cs < 714) {
return 8;
}
else if (Cs >= 714) {
return 2;
}
}
else if (diffRsCsdiff >= -17) {
return 8;
}
}
else if (ssDCval >= 56) {
if (Rs < 1121) {
return 4;
}
else if (Rs >= 1121) {
return 2;
}
}
}
}
else if (tcor >= 89) {
if (mu_mv_mag_sq < 73) {
if (negBalance < 35) {
if (Cs < 680) {
return 1;
}
else if (Cs >= 680) {
if (refDCval < 56) {
if (negBalance < 28) {
return 1;
}
else if (negBalance >= 28) {
if (Rs < 1142) {
return 8;
}
else if (Rs >= 1142) {
return 1;
}
}
}
else if (refDCval >= 56) {
return 8;
}
}
}
else if (negBalance >= 35) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 73) {
if (Rs < 1338) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
return 8;
}
}
}
else if (Rs >= 1338) {
if (ssDCval < 47) {
if (Rs < 1420) {
return 4;
}
else if (Rs >= 1420) {
return 8;
}
}
else if (ssDCval >= 47) {
return 2;
}
}
}
}
}
else if (Rs >= 1503) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (RsDiff < 10) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (RsDiff >= 10) {
return 8;
}
}
}
}
else if (Cs >= 929) {
if (negBalance < 27) {
if (mu_mv_mag_sq < 106) {
if (Rs < 1609) {
return 4;
}
else if (Rs >= 1609) {
return 1;
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 1167) {
return 1;
}
else if (Rs >= 1167) {
return 8;
}
}
}
else if (negBalance >= 27) {
if (posBalance < 45) {
if (Rs < 2185) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 2;
}
}
}
else if (Rs >= 2185) {
return 4;
}
}
else if (posBalance >= 45) {
if (Rs < 1279) {
if (refDCval < 50) {
if (Rs < 1264) {
if (diffRsCsdiff < -26) {
return 4;
}
else if (diffRsCsdiff >= -26) {
return 8;
}
}
else if (Rs >= 1264) {
return 1;
}
}
else if (refDCval >= 50) {
if (RsCsDiff < 11) {
return 8;
}
else if (RsCsDiff >= 11) {
if (diffAFD < 1) {
if (Rs < 1188) {
return 4;
}
else if (Rs >= 1188) {
if (Cs < 1221) {
return 8;
}
else if (Cs >= 1221) {
if (Rs < 1220) {
return 8;
}
else if (Rs >= 1220) {
return 4;
}
}
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
else if (Rs >= 1279) {
if (diffTSC < 2) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
return 8;
}
}
else if (diffTSC >= 2) {
if (Rs < 1656) {
return 2;
}
else if (Rs >= 1656) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 59) {
if (mu_mv_mag_sq < 113) {
if (Rs < 2020) {
if (Rs < 877) {
if (negBalance < 40) {
if (Rs < 644) {
if (mu_mv_mag_sq < 97) {
if (RsCsDiff < 12) {
if (negBalance < 28) {
return 1;
}
else if (negBalance >= 28) {
if (Cs < 801) {
return 8;
}
else if (Cs >= 801) {
if (SC < 1574) {
return 1;
}
else if (SC >= 1574) {
return 8;
}
}
}
}
else if (RsCsDiff >= 12) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 629) {
return 2;
}
else if (Cs >= 629) {
return 8;
}
}
}
else if (Rs >= 644) {
if (Rs < 802) {
if (MVDiff < 2) {
if (Cs < 763) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 763) {
return 1;
}
}
else if (MVDiff >= 2) {
if (negBalance < 30) {
return 2;
}
else if (negBalance >= 30) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
else if (Rs >= 802) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
else if (negBalance >= 40) {
if (Rs < 859) {
if (posBalance < 51) {
if (Rs < 696) {
return 1;
}
else if (Rs >= 696) {
return 4;
}
}
else if (posBalance >= 51) {
if (Rs < 504) {
return 4;
}
else if (Rs >= 504) {
if (Rs < 815) {
return 8;
}
else if (Rs >= 815) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
}
else if (Rs >= 859) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (Rs < 872) {
return 8;
}
else if (Rs >= 872) {
return 1;
}
}
}
}
}
else if (Rs >= 877) {
if (Rs < 996) {
if (Cs < 1091) {
if (Cs < 744) {
if (Cs < 712) {
if (AFD < 10) {
if (RsDiff < 3) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (AFD >= 10) {
return 1;
}
}
else if (Cs >= 712) {
return 1;
}
}
else if (Cs >= 744) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (Cs >= 1091) {
return 4;
}
}
else if (Rs >= 996) {
if (RsDiff < 3) {
if (posBalance < 55) {
if (negBalance < 37) {
if (diffMVdiffVal < 3) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (diffMVdiffVal >= 3) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (negBalance >= 37) {
if (AFD < 7) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (Cs < 1149) {
return 1;
}
else if (Cs >= 1149) {
return 8;
}
}
}
else if (AFD >= 7) {
if (Cs < 942) {
return 4;
}
else if (Cs >= 942) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (RsDiff >= 3) {
if (RsCsDiff < 48) {
if (Cs < 906) {
if (mu_mv_mag_sq < 63) {
if (posBalance < 59) {
if (ssDCval < 64) {
if (Rs < 1160) {
return 1;
}
else if (Rs >= 1160) {
if (Cs < 724) {
if (Rs < 1670) {
return 2;
}
else if (Rs >= 1670) {
return 1;
}
}
else if (Cs >= 724) {
return 4;
}
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (posBalance >= 59) {
if (Rs < 1590) {
return 8;
}
else if (Rs >= 1590) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (MVDiff < 5) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (MVDiff >= 5) {
if (RsDiff < 5) {
if (Cs < 620) {
if (Rs < 1184) {
return 8;
}
else if (Rs >= 1184) {
return 4;
}
}
else if (Cs >= 620) {
if (Rs < 1696) {
if (Cs < 878) {
return 8;
}
else if (Cs >= 878) {
if (Rs < 1361) {
return 1;
}
else if (Rs >= 1361) {
return 8;
}
}
}
else if (Rs >= 1696) {
return 1;
}
}
}
else if (RsDiff >= 5) {
if (Rs < 1222) {
return 1;
}
else if (Rs >= 1222) {
return 8;
}
}
}
}
}
else if (Cs >= 906) {
if (SC < 2525) {
if (posBalance < 56) {
if (negBalance < 43) {
if (tcor < 91) {
return 1;
}
else if (tcor >= 91) {
if (negBalance < 34) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
return 8;
}
}
else if (negBalance >= 34) {
if (RsCsDiff < 16) {
if (Rs < 1307) {
return 1;
}
else if (Rs >= 1307) {
if (Rs < 1430) {
return 4;
}
else if (Rs >= 1430) {
return 8;
}
}
}
else if (RsCsDiff >= 16) {
return 4;
}
}
}
}
else if (negBalance >= 43) {
return 8;
}
}
else if (posBalance >= 56) {
if (mu_mv_mag_sq < 58) {
if (gchDC < 1) {
if (Rs < 1193) {
return 2;
}
else if (Rs >= 1193) {
return 1;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 58) {
if (RsCsDiff < 22) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (RsCsDiff >= 22) {
if (SC < 2519) {
return 4;
}
else if (SC >= 2519) {
return 8;
}
}
}
}
}
else if (SC >= 2525) {
if (Cs < 1176) {
if (MVDiff < 5) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (Cs >= 1176) {
if (negBalance < 39) {
if (Rs < 1354) {
return 4;
}
else if (Rs >= 1354) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 1407) {
return 8;
}
else if (Rs >= 1407) {
return 1;
}
}
}
}
else if (negBalance >= 39) {
if (mu_mv_mag_sq < 97) {
if (posBalance < 55) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
else if (posBalance >= 55) {
return 1;
}
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 1250) {
return 8;
}
else if (Cs >= 1250) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 48) {
if (posBalance < 52) {
if (Cs < 1103) {
return 4;
}
else if (Cs >= 1103) {
return 8;
}
}
else if (posBalance >= 52) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2020) {
if (Cs < 915) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 1;
}
}
else if (Cs >= 915) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 1329) {
if (posBalance < 49) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (posBalance >= 49) {
if (AFD < 4) {
if (negBalance < 35) {
if (Cs < 709) {
return 1;
}
else if (Cs >= 709) {
return 4;
}
}
else if (negBalance >= 35) {
if (Cs < 1257) {
return 1;
}
else if (Cs >= 1257) {
return 4;
}
}
}
else if (AFD >= 4) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (ssDCval < 68) {
if (RsDiff < 3) {
if (Rs < 1137) {
return 4;
}
else if (Rs >= 1137) {
if (Cs < 1171) {
return 8;
}
else if (Cs >= 1171) {
return 4;
}
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (ssDCval >= 68) {
if (mu_mv_mag_sq < 117) {
return 8;
}
else if (mu_mv_mag_sq >= 117) {
if (Rs < 1184) {
return 4;
}
else if (Rs >= 1184) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 1329) {
if (RsCsDiff < 23) {
if (diffRsCsdiff < 7) {
return 8;
}
else if (diffRsCsdiff >= 7) {
if (Rs < 1436) {
return 4;
}
else if (Rs >= 1436) {
return 8;
}
}
}
else if (RsCsDiff >= 23) {
if (Rs < 1609) {
return 4;
}
else if (Rs >= 1609) {
if (Rs < 2039) {
return 8;
}
else if (Rs >= 2039) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 1613) {
if (TSC < 6) {
if (tcor < 86) {
if (Rs < 744) {
if (Rs < 621) {
return 8;
}
else if (Rs >= 621) {
return 4;
}
}
else if (Rs >= 744) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 1174) {
return 2;
}
else if (Rs >= 1174) {
return 8;
}
}
}
}
else if (tcor >= 86) {
if (mu_mv_mag_sq < 71) {
if (Rs < 1488) {
if (mcTcor < 90) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (mcTcor >= 90) {
if (MVDiff < 15) {
if (SC < 1983) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (SC >= 1983) {
return 4;
}
}
else if (MVDiff >= 15) {
return 8;
}
}
}
else if (Rs >= 1488) {
return 8;
}
}
else if (mu_mv_mag_sq >= 71) {
if (mu_mv_mag_sq < 98) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 98) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (MVDiff < 3) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 1245) {
return 4;
}
else if (Rs >= 1245) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (refDCval < 53) {
if (MVDiff < 18) {
if (tcor < 75) {
if (mu_mv_mag_sq < 101) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 1394) {
return 1;
}
else if (Rs >= 1394) {
return 8;
}
}
}
else if (tcor >= 75) {
if (mu_mv_mag_sq < 55) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
return 8;
}
}
else if (mu_mv_mag_sq >= 55) {
return 8;
}
}
}
else if (MVDiff >= 18) {
if (Rs < 1310) {
return 4;
}
else if (Rs >= 1310) {
return 8;
}
}
}
else if (refDCval >= 53) {
if (Rs < 1062) {
if (tcor < 82) {
if (Cs < 757) {
return 4;
}
else if (Cs >= 757) {
if (Rs < 853) {
return 8;
}
else if (Rs >= 853) {
return 4;
}
}
}
else if (tcor >= 82) {
if (SC < 1764) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (SC >= 1764) {
return 1;
}
}
}
else if (Rs >= 1062) {
if (tcor < 89) {
if (mu_mv_mag_sq < 95) {
return 8;
}
else if (mu_mv_mag_sq >= 95) {
if (Rs < 1170) {
return 4;
}
else if (Rs >= 1170) {
return 8;
}
}
}
else if (tcor >= 89) {
if (Rs < 1505) {
if (ssDCval < 62) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (ssDCval >= 62) {
if (mu_mv_mag_sq < 71) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 71) {
return 8;
}
}
}
else if (Rs >= 1505) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1613) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (ssDCval < 72) {
if (CsDiff < 5) {
if (RsDiff < 5) {
if (mu_mv_mag_sq < 118) {
return 8;
}
else if (mu_mv_mag_sq >= 118) {
return 2;
}
}
else if (RsDiff >= 5) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (Cs < 1102) {
return 2;
}
else if (Cs >= 1102) {
return 8;
}
}
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (ssDCval >= 72) {
if (negBalance < 48) {
if (Rs < 1728) {
return 8;
}
else if (Rs >= 1728) {
return 1;
}
}
else if (negBalance >= 48) {
if (SC < 3374) {
if (diffMVdiffVal < 9) {
return 8;
}
else if (diffMVdiffVal >= 9) {
return 1;
}
}
else if (SC >= 3374) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 79) {
if (ssDCval < 158) {
if (ssDCval < 114) {
if (mcTcor < 92) {
if (AFD < 11) {
if (mu_mv_mag_sq < 47) {
if (Rs < 1526) {
return 8;
}
else if (Rs >= 1526) {
return 2;
}
}
else if (mu_mv_mag_sq >= 47) {
if (posBalance < 54) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (tcor < 83) {
return 2;
}
else if (tcor >= 83) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (AFD >= 11) {
if (MVDiff < 10) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (Rs < 1921) {
return 8;
}
else if (Rs >= 1921) {
return 1;
}
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (negBalance < 42) {
if (ssDCval < 99) {
if (refDCval < 84) {
if (Rs < 2040) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (Cs < 1267) {
return 8;
}
else if (Cs >= 1267) {
if (Rs < 1821) {
return 1;
}
else if (Rs >= 1821) {
return 8;
}
}
}
}
else if (Rs >= 2040) {
if (Rs < 2253) {
return 1;
}
else if (Rs >= 2253) {
return 2;
}
}
}
else if (refDCval >= 84) {
if (SC < 1625) {
if (Rs < 458) {
return 8;
}
else if (Rs >= 458) {
if (posBalance < 60) {
if (Cs < 560) {
return 8;
}
else if (Cs >= 560) {
return 2;
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
else if (SC >= 1625) {
if (Cs < 1132) {
if (Cs < 789) {
if (Rs < 1758) {
if (Rs < 1656) {
return 8;
}
else if (Rs >= 1656) {
return 4;
}
}
else if (Rs >= 1758) {
return 1;
}
}
else if (Cs >= 789) {
if (Rs < 1204) {
if (mu_mv_mag_sq < 69) {
if (Rs < 860) {
return 2;
}
else if (Rs >= 860) {
return 8;
}
}
else if (mu_mv_mag_sq >= 69) {
if (Rs < 706) {
return 4;
}
else if (Rs >= 706) {
return 1;
}
}
}
else if (Rs >= 1204) {
if (diffRsCsdiff < -17) {
return 2;
}
else if (diffRsCsdiff >= -17) {
return 8;
}
}
}
}
else if (Cs >= 1132) {
if (RsDiff < 4) {
if (mu_mv_mag_sq < 79) {
if (Rs < 1300) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (Rs >= 1300) {
return 8;
}
}
else if (mu_mv_mag_sq >= 79) {
return 8;
}
}
else if (RsDiff >= 4) {
if (posBalance < 52) {
if (Cs < 1185) {
return 2;
}
else if (Cs >= 1185) {
return 8;
}
}
else if (posBalance >= 52) {
if (Rs < 2239) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
}
else if (Rs >= 2239) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 99) {
if (RsCsDiff < 4) {
if (Rs < 1011) {
return 1;
}
else if (Rs >= 1011) {
return 8;
}
}
else if (RsCsDiff >= 4) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
}
}
else if (negBalance >= 42) {
if (MVDiff < 6) {
if (negBalance < 45) {
if (RsCsDiff < 4) {
if (Cs < 586) {
return 8;
}
else if (Cs >= 586) {
return 1;
}
}
else if (RsCsDiff >= 4) {
if (posBalance < 53) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 1751) {
return 8;
}
else if (Rs >= 1751) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (refDCval < 80) {
return 1;
}
else if (refDCval >= 80) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 83) {
if (diffRsCsdiff < 4) {
if (CsDiff < 2) {
if (Cs < 690) {
return 1;
}
else if (Cs >= 690) {
return 8;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 1824) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (Rs >= 1824) {
if (SC < 3299) {
return 8;
}
else if (SC >= 3299) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 449) {
return 1;
}
else if (Rs >= 449) {
if (refDCval < 81) {
if (Rs < 1735) {
return 8;
}
else if (Rs >= 1735) {
return 1;
}
}
else if (refDCval >= 81) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Rs < 630) {
return 8;
}
else if (Rs >= 630) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 57) {
if (tcor < 92) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 1;
}
}
else if (tcor >= 92) {
if (ssDCval < 99) {
if (MVDiff < 3) {
if (Rs < 420) {
return 4;
}
else if (Rs >= 420) {
return 1;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (ssDCval >= 99) {
return 1;
}
}
}
}
}
else if (MVDiff >= 6) {
if (RsCsDiff < 71) {
if (TSC < 5) {
if (Cs < 852) {
return 8;
}
else if (Cs >= 852) {
return 1;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (RsCsDiff >= 71) {
if (Cs < 1096) {
return 4;
}
else if (Cs >= 1096) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 114) {
if (tcor < 83) {
if (Cs < 731) {
if (Rs < 1837) {
return 4;
}
else if (Rs >= 1837) {
return 8;
}
}
else if (Cs >= 731) {
if (Cs < 757) {
return 1;
}
else if (Cs >= 757) {
return 8;
}
}
}
else if (tcor >= 83) {
if (MVDiff < 24) {
if (negBalance < 30) {
if (ssDCval < 120) {
if (Cs < 1120) {
if (Rs < 542) {
return 2;
}
else if (Rs >= 542) {
return 8;
}
}
else if (Cs >= 1120) {
if (Rs < 1590) {
return 4;
}
else if (Rs >= 1590) {
return 8;
}
}
}
else if (ssDCval >= 120) {
if (tcor < 86) {
if (Rs < 979) {
return 8;
}
else if (Rs >= 979) {
return 4;
}
}
else if (tcor >= 86) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 2;
}
}
}
}
else if (negBalance >= 30) {
if (diffRsCsdiff < 23) {
if (mu_mv_mag_sq < 116) {
if (diffMVdiffVal < -9) {
if (Rs < 1321) {
return 4;
}
else if (Rs >= 1321) {
return 8;
}
}
else if (diffMVdiffVal >= -9) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Rs < 1588) {
return 4;
}
else if (Rs >= 1588) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 23) {
if (Rs < 1586) {
return 8;
}
else if (Rs >= 1586) {
return 2;
}
}
}
}
else if (MVDiff >= 24) {
if (Cs < 1039) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Cs >= 1039) {
return 8;
}
}
}
}
}
else if (ssDCval >= 158) {
if (Rs < 842) {
return 4;
}
else if (Rs >= 842) {
return 2;
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (MVDiff < 10) {
if (TSC < 9) {
if (posBalance < 58) {
if (AFD < 16) {
if (mu_mv_mag_sq < 2) {
if (Cs < 852) {
return 2;
}
else if (Cs >= 852) {
if (ssDCval < 82) {
return 1;
}
else if (ssDCval >= 82) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 2087) {
if (mu_mv_mag_sq < 11) {
return 8;
}
else if (mu_mv_mag_sq >= 11) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
return 8;
}
}
}
else if (Rs >= 2087) {
if (Rs < 2165) {
return 2;
}
else if (Rs >= 2165) {
return 4;
}
}
}
}
else if (AFD >= 16) {
if (diffTSC < -1) {
if (Rs < 2238) {
return 4;
}
else if (Rs >= 2238) {
return 2;
}
}
else if (diffTSC >= -1) {
if (ssDCval < 140) {
if (Cs < 938) {
return 2;
}
else if (Cs >= 938) {
if (Rs < 995) {
return 2;
}
else if (Rs >= 995) {
return 4;
}
}
}
else if (ssDCval >= 140) {
return 8;
}
}
}
}
else if (posBalance >= 58) {
if (AFD < 19) {
if (refDCval < 58) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (Rs < 1632) {
return 4;
}
else if (Rs >= 1632) {
return 8;
}
}
}
else if (refDCval >= 58) {
if (tcor < 73) {
if (ssDCval < 105) {
return 8;
}
else if (ssDCval >= 105) {
if (Rs < 1083) {
return 1;
}
else if (Rs >= 1083) {
return 4;
}
}
}
else if (tcor >= 73) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1150) {
return 8;
}
else if (Rs >= 1150) {
if (Cs < 1074) {
return 4;
}
else if (Cs >= 1074) {
if (Rs < 1878) {
return 8;
}
else if (Rs >= 1878) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 783) {
if (Rs < 581) {
return 2;
}
else if (Rs >= 581) {
return 4;
}
}
else if (Rs >= 783) {
if (Rs < 1709) {
if (Cs < 1139) {
return 8;
}
else if (Cs >= 1139) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (Rs >= 1709) {
return 1;
}
}
}
}
}
}
else if (AFD >= 19) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
if (Rs < 1024) {
if (Rs < 762) {
return 1;
}
else if (Rs >= 762) {
return 8;
}
}
else if (Rs >= 1024) {
return 4;
}
}
}
}
}
else if (TSC >= 9) {
if (Rs < 1430) {
if (negBalance < 55) {
if (Cs < 1107) {
return 8;
}
else if (Cs >= 1107) {
return 2;
}
}
else if (negBalance >= 55) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 4;
}
}
}
}
else if (Rs >= 1430) {
if (Cs < 714) {
return 2;
}
else if (Cs >= 714) {
if (Rs < 1489) {
if (negBalance < 58) {
if (Rs < 1465) {
return 8;
}
else if (Rs >= 1465) {
return 4;
}
}
else if (negBalance >= 58) {
return 8;
}
}
else if (Rs >= 1489) {
if (negBalance < 43) {
return 4;
}
else if (negBalance >= 43) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (Rs < 2020) {
return 1;
}
else if (Rs >= 2020) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 10) {
if (refDCval < 113) {
if (AFD < 19) {
if (RsDiff < 9) {
if (SC < 2565) {
if (tcor < 52) {
if (refDCval < 45) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (refDCval >= 45) {
if (mu_mv_mag_sq < 3) {
if (AFD < 15) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (AFD >= 15) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 1044) {
return 4;
}
else if (Cs >= 1044) {
if (Rs < 975) {
return 8;
}
else if (Rs >= 975) {
return 1;
}
}
}
}
}
else if (tcor >= 52) {
if (TSC < 11) {
if (tcor < 85) {
if (negBalance < 38) {
if (RsCsDiff < 43) {
if (Rs < 716) {
return 1;
}
else if (Rs >= 716) {
if (gchDC < 4) {
return 1;
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (RsCsDiff >= 43) {
if (Rs < 1637) {
if (mu_mv_mag_sq < 42) {
return 2;
}
else if (mu_mv_mag_sq >= 42) {
if (Rs < 663) {
return 2;
}
else if (Rs >= 663) {
return 8;
}
}
}
else if (Rs >= 1637) {
return 1;
}
}
}
else if (negBalance >= 38) {
if (CsDiff < 5) {
if (ssDCval < 101) {
if (SC < 1197) {
if (refDCval < 92) {
if (mcTcor < 73) {
return 8;
}
else if (mcTcor >= 73) {
if (tcor < 71) {
if (Rs < 521) {
if (diffRsCsdiff < 3) {
return 4;
}
else if (diffRsCsdiff >= 3) {
if (SC < 1031) {
return 2;
}
else if (SC >= 1031) {
return 1;
}
}
}
else if (Rs >= 521) {
return 2;
}
}
else if (tcor >= 71) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
else if (refDCval >= 92) {
if (Rs < 513) {
return 1;
}
else if (Rs >= 513) {
return 8;
}
}
}
else if (SC >= 1197) {
if (mcTcor < 92) {
if (Rs < 1237) {
if (RsDiff < 6) {
if (Cs < 929) {
if (ssDCval < 69) {
if (negBalance < 57) {
if (mcTcor < 87) {
return 4;
}
else if (mcTcor >= 87) {
if (SC < 1619) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 4;
}
}
else if (SC >= 1619) {
return 4;
}
}
}
else if (negBalance >= 57) {
return 8;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (Cs >= 929) {
if (Cs < 1040) {
return 1;
}
else if (Cs >= 1040) {
return 4;
}
}
}
else if (RsDiff >= 6) {
if (negBalance < 48) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (negBalance >= 48) {
return 4;
}
}
}
else if (Rs >= 1237) {
if (ssDCval < 77) {
if (posBalance < 58) {
if (diffRsCsdiff < -13) {
return 8;
}
else if (diffRsCsdiff >= -13) {
if (ssDCval < 64) {
if (Rs < 1597) {
return 8;
}
else if (Rs >= 1597) {
return 2;
}
}
else if (ssDCval >= 64) {
return 2;
}
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (ssDCval >= 77) {
if (SC < 2152) {
return 4;
}
else if (SC >= 2152) {
if (Rs < 1316) {
return 4;
}
else if (Rs >= 1316) {
return 8;
}
}
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 86) {
if (TSC < 7) {
if (AFD < 14) {
if (Cs < 584) {
return 8;
}
else if (Cs >= 584) {
return 2;
}
}
else if (AFD >= 14) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
else if (TSC >= 7) {
if (Rs < 783) {
if (tcor < 78) {
return 1;
}
else if (tcor >= 78) {
if (Rs < 732) {
return 4;
}
else if (Rs >= 732) {
return 8;
}
}
}
else if (Rs >= 783) {
if (ssDCval < 83) {
if (diffMVdiffVal < 11) {
return 4;
}
else if (diffMVdiffVal >= 11) {
return 8;
}
}
else if (ssDCval >= 83) {
return 1;
}
}
}
}
else if (ssDCval >= 86) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 8;
}
}
}
}
}
else if (ssDCval >= 101) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
if (posBalance < 60) {
if (Rs < 894) {
return 2;
}
else if (Rs >= 894) {
return 4;
}
}
else if (posBalance >= 60) {
return 1;
}
}
}
}
else if (CsDiff >= 5) {
if (ssDCval < 54) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
if (MVDiff < 23) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (mu_mv_mag_sq < 9) {
return 4;
}
else if (mu_mv_mag_sq >= 9) {
return 8;
}
}
}
else if (MVDiff >= 23) {
if (RsCsDiff < 67) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (Rs < 471) {
return 8;
}
else if (Rs >= 471) {
return 2;
}
}
}
else if (RsCsDiff >= 67) {
return 4;
}
}
}
}
else if (ssDCval >= 54) {
if (mu_mv_mag_sq < 5) {
if (refDCval < 70) {
if (MVDiff < 24) {
if (Rs < 891) {
return 1;
}
else if (Rs >= 891) {
if (diffRsCsdiff < -19) {
if (Rs < 1066) {
return 2;
}
else if (Rs >= 1066) {
return 1;
}
}
else if (diffRsCsdiff >= -19) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 2;
}
}
}
}
else if (MVDiff >= 24) {
if (posBalance < 55) {
if (diffRsCsdiff < -9) {
return 1;
}
else if (diffRsCsdiff >= -9) {
return 4;
}
}
else if (posBalance >= 55) {
if (SC < 1382) {
if (RsCsDiff < 57) {
if (Cs < 723) {
return 4;
}
else if (Cs >= 723) {
return 8;
}
}
else if (RsCsDiff >= 57) {
return 1;
}
}
else if (SC >= 1382) {
if (diffAFD < -1) {
if (Cs < 832) {
return 1;
}
else if (Cs >= 832) {
return 4;
}
}
else if (diffAFD >= -1) {
if (mcTcor < 86) {
return 2;
}
else if (mcTcor >= 86) {
if (Rs < 654) {
return 2;
}
else if (Rs >= 654) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 70) {
if (ssDCval < 92) {
if (posBalance < 61) {
if (gchDC < 4) {
if (Rs < 1420) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 2;
}
}
else if (Rs >= 1420) {
return 8;
}
}
else if (gchDC >= 4) {
if (Rs < 466) {
return 4;
}
else if (Rs >= 466) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (tcor < 80) {
if (TSC < 9) {
if (SC < 2186) {
return 4;
}
else if (SC >= 2186) {
return 2;
}
}
else if (TSC >= 9) {
return 2;
}
}
else if (tcor >= 80) {
if (Rs < 1150) {
return 1;
}
else if (Rs >= 1150) {
return 4;
}
}
}
}
else if (ssDCval >= 92) {
if (Cs < 1202) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
return 1;
}
}
else if (Cs >= 1202) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Cs < 1147) {
if (MVDiff < 28) {
if (Cs < 1139) {
if (Rs < 435) {
return 2;
}
else if (Rs >= 435) {
if (diffRsCsdiff < 5) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (diffRsCsdiff >= 5) {
return 2;
}
}
}
else if (Cs >= 1139) {
return 4;
}
}
else if (MVDiff >= 28) {
if (CsDiff < 6) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Rs < 426) {
return 8;
}
else if (Rs >= 426) {
return 1;
}
}
}
else if (CsDiff >= 6) {
if (TSC < 8) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
}
else if (Cs >= 1147) {
if (refDCval < 72) {
if (Rs < 693) {
return 8;
}
else if (Rs >= 693) {
return 1;
}
}
else if (refDCval >= 72) {
if (Cs < 1323) {
return 8;
}
else if (Cs >= 1323) {
return 1;
}
}
}
}
}
}
}
}
else if (tcor >= 85) {
if (Cs < 834) {
if (mu_mv_mag_sq < 32) {
if (diffRsCsdiff < -22) {
if (Rs < 1040) {
return 8;
}
else if (Rs >= 1040) {
return 2;
}
}
else if (diffRsCsdiff >= -22) {
if (Rs < 1747) {
return 1;
}
else if (Rs >= 1747) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (Cs >= 834) {
if (diffMVdiffVal < -14) {
return 2;
}
else if (diffMVdiffVal >= -14) {
if (gchDC < 2) {
if (diffAFD < 2) {
if (CsDiff < 5) {
if (Rs < 1034) {
return 8;
}
else if (Rs >= 1034) {
return 4;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (diffAFD >= 2) {
if (Rs < 1340) {
return 4;
}
else if (Rs >= 1340) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 74) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (ssDCval >= 74) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 2;
}
}
}
}
}
}
}
}
else if (TSC >= 11) {
if (MVDiff < 15) {
if (Cs < 1248) {
return 4;
}
else if (Cs >= 1248) {
return 2;
}
}
else if (MVDiff >= 15) {
if (CsDiff < 7) {
if (diffRsCsdiff < -12) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
return 8;
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
else if (CsDiff >= 7) {
return 2;
}
}
}
}
}
else if (SC >= 2565) {
if (negBalance < 38) {
if (diffAFD < 3) {
if (Cs < 690) {
return 1;
}
else if (Cs >= 690) {
return 2;
}
}
else if (diffAFD >= 3) {
if (Rs < 1936) {
return 8;
}
else if (Rs >= 1936) {
return 4;
}
}
}
else if (negBalance >= 38) {
if (Cs < 995) {
if (tcor < 81) {
if (ssDCval < 67) {
if (Rs < 1630) {
return 8;
}
else if (Rs >= 1630) {
return 4;
}
}
else if (ssDCval >= 67) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 4;
}
}
}
else if (tcor >= 81) {
if (Rs < 1969) {
return 1;
}
else if (Rs >= 1969) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
if (Rs < 2071) {
return 4;
}
else if (Rs >= 2071) {
return 8;
}
}
}
}
}
else if (Cs >= 995) {
if (Cs < 1274) {
if (TSC < 9) {
if (Rs < 2020) {
if (mu_mv_mag_sq < 54) {
if (diffAFD < 1) {
if (mcTcor < 91) {
if (Rs < 1727) {
return 2;
}
else if (Rs >= 1727) {
return 1;
}
}
else if (mcTcor >= 91) {
if (mu_mv_mag_sq < 10) {
return 8;
}
else if (mu_mv_mag_sq >= 10) {
return 1;
}
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 1532) {
return 4;
}
else if (Rs >= 1532) {
return 1;
}
}
}
else if (Rs >= 2020) {
return 4;
}
}
else if (TSC >= 9) {
if (Rs < 1434) {
if (SC < 2609) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
else if (SC >= 2609) {
if (Rs < 1411) {
return 2;
}
else if (Rs >= 1411) {
return 1;
}
}
}
else if (Rs >= 1434) {
return 8;
}
}
}
else if (Cs >= 1274) {
if (RsCsDiff < 43) {
return 4;
}
else if (RsCsDiff >= 43) {
if (Rs < 2067) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
if (diffTSC < -1) {
if (Rs < 1533) {
if (Rs < 1410) {
if (Cs < 1291) {
return 4;
}
else if (Cs >= 1291) {
return 8;
}
}
else if (Rs >= 1410) {
return 4;
}
}
else if (Rs >= 1533) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (Rs >= 2067) {
if (Rs < 2141) {
return 2;
}
else if (Rs >= 2141) {
return 4;
}
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (mcTcor < 90) {
if (ssDCval < 88) {
if (TSC < 8) {
if (negBalance < 37) {
if (Rs < 2263) {
return 1;
}
else if (Rs >= 2263) {
return 2;
}
}
else if (negBalance >= 37) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
return 8;
}
}
}
else if (TSC >= 8) {
if (Cs < 951) {
if (RsCsDiff < 121) {
if (MVDiff < 47) {
return 4;
}
else if (MVDiff >= 47) {
return 8;
}
}
else if (RsCsDiff >= 121) {
if (negBalance < 45) {
if (mu_mv_mag_sq < 46) {
return 2;
}
else if (mu_mv_mag_sq >= 46) {
return 8;
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (Cs >= 951) {
return 8;
}
}
}
else if (ssDCval >= 88) {
if (TSC < 10) {
if (ssDCval < 102) {
return 2;
}
else if (ssDCval >= 102) {
return 4;
}
}
else if (TSC >= 10) {
if (diffMVdiffVal < 9) {
if (Cs < 1119) {
return 1;
}
else if (Cs >= 1119) {
return 4;
}
}
else if (diffMVdiffVal >= 9) {
return 4;
}
}
}
}
else if (mcTcor >= 90) {
if (Rs < 1543) {
if (AFD < 16) {
if (Cs < 1143) {
return 1;
}
else if (Cs >= 1143) {
return 2;
}
}
else if (AFD >= 16) {
return 8;
}
}
else if (Rs >= 1543) {
if (refDCval < 71) {
if (Cs < 1195) {
return 2;
}
else if (Cs >= 1195) {
return 4;
}
}
else if (refDCval >= 71) {
if (posBalance < 58) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (AFD < 15) {
if (Cs < 793) {
return 2;
}
else if (Cs >= 793) {
return 1;
}
}
else if (AFD >= 15) {
if (Rs < 2058) {
if (SC < 2350) {
return 2;
}
else if (SC >= 2350) {
return 8;
}
}
else if (Rs >= 2058) {
if (Rs < 2170) {
return 4;
}
else if (Rs >= 2170) {
if (Rs < 2218) {
return 1;
}
else if (Rs >= 2218) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 58) {
if (RsCsDiff < 105) {
if (Rs < 1735) {
return 1;
}
else if (Rs >= 1735) {
return 4;
}
}
else if (RsCsDiff >= 105) {
if (negBalance < 52) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
else if (negBalance >= 52) {
if (Rs < 1808) {
if (Rs < 1749) {
return 8;
}
else if (Rs >= 1749) {
return 1;
}
}
else if (Rs >= 1808) {
if (Rs < 1971) {
return 2;
}
else if (Rs >= 1971) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (AFD >= 19) {
if (posBalance < 52) {
if (MVDiff < 35) {
if (RsCsDiff < 70) {
return 8;
}
else if (RsCsDiff >= 70) {
if (diffAFD < 5) {
return 2;
}
else if (diffAFD >= 5) {
return 8;
}
}
}
else if (MVDiff >= 35) {
if (mu_mv_mag_sq < 5) {
if (Cs < 1021) {
if (TSC < 9) {
if (Rs < 1089) {
return 1;
}
else if (Rs >= 1089) {
return 8;
}
}
else if (TSC >= 9) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (Cs >= 1021) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
if (Rs < 981) {
return 1;
}
else if (Rs >= 981) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 1903) {
if (AFD < 22) {
if (TSC < 9) {
if (Rs < 721) {
return 4;
}
else if (Rs >= 721) {
return 2;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (AFD >= 22) {
if (MVDiff < 93) {
if (posBalance < 51) {
if (diffMVdiffVal < 0) {
if (Rs < 596) {
return 4;
}
else if (Rs >= 596) {
if (Rs < 1551) {
return 2;
}
else if (Rs >= 1551) {
return 4;
}
}
}
else if (diffMVdiffVal >= 0) {
return 2;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (MVDiff >= 93) {
if (gchDC < 11) {
if (Cs < 535) {
return 4;
}
else if (Cs >= 535) {
return 1;
}
}
else if (gchDC >= 11) {
if (Rs < 859) {
if (posBalance < 43) {
if (Cs < 890) {
return 2;
}
else if (Cs >= 890) {
return 1;
}
}
else if (posBalance >= 43) {
if (Rs < 355) {
return 8;
}
else if (Rs >= 355) {
return 1;
}
}
}
else if (Rs >= 859) {
return 8;
}
}
}
}
}
else if (Rs >= 1903) {
return 1;
}
}
}
}
else if (posBalance >= 52) {
if (RsDiff < 9) {
if (refDCval < 65) {
if (negBalance < 49) {
if (diffTSC < 2) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
if (RsCsDiff < 58) {
if (Rs < 517) {
return 4;
}
else if (Rs >= 517) {
return 1;
}
}
else if (RsCsDiff >= 58) {
return 2;
}
}
}
else if (diffTSC >= 2) {
if (Rs < 715) {
if (Rs < 571) {
return 1;
}
else if (Rs >= 571) {
return 2;
}
}
else if (Rs >= 715) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (diffAFD < 4) {
if (mcTcor < 86) {
if (Cs < 524) {
return 2;
}
else if (Cs >= 524) {
return 8;
}
}
else if (mcTcor >= 86) {
if (TSC < 9) {
if (Rs < 780) {
return 4;
}
else if (Rs >= 780) {
return 1;
}
}
else if (TSC >= 9) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 397) {
return 8;
}
else if (Rs >= 397) {
return 4;
}
}
}
}
}
else if (diffAFD >= 4) {
if (Rs < 703) {
return 2;
}
else if (Rs >= 703) {
return 1;
}
}
}
}
else if (refDCval >= 65) {
if (MVDiff < 26) {
if (Cs < 970) {
if (TSC < 10) {
if (MVDiff < 19) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 14) {
return 8;
}
else if (mu_mv_mag_sq >= 14) {
return 1;
}
}
else if (gchDC >= 2) {
return 4;
}
}
else if (MVDiff >= 19) {
if (mcTcor < 96) {
if (Rs < 506) {
return 8;
}
else if (Rs >= 506) {
return 1;
}
}
else if (mcTcor >= 96) {
return 8;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (Cs >= 970) {
if (RsDiff < 7) {
if (ssDCval < 91) {
if (Rs < 1174) {
return 4;
}
else if (Rs >= 1174) {
return 1;
}
}
else if (ssDCval >= 91) {
return 2;
}
}
else if (RsDiff >= 7) {
if (CsDiff < 8) {
if (Rs < 1177) {
return 1;
}
else if (Rs >= 1177) {
return 2;
}
}
else if (CsDiff >= 8) {
if (TSC < 9) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (TSC >= 9) {
if (Cs < 1211) {
return 8;
}
else if (Cs >= 1211) {
if (Cs < 1297) {
return 1;
}
else if (Cs >= 1297) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 26) {
if (mcTcor < 96) {
if (RsCsDiff < 129) {
if (Rs < 1214) {
if (mcTcor < 84) {
if (Rs < 1101) {
if (tcor < 63) {
if (diffMVdiffVal < 36) {
if (Cs < 1062) {
return 4;
}
else if (Cs >= 1062) {
if (Rs < 792) {
return 8;
}
else if (Rs >= 792) {
return 4;
}
}
}
else if (diffMVdiffVal >= 36) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (Cs < 569) {
return 8;
}
else if (Cs >= 569) {
return 1;
}
}
}
}
else if (tcor >= 63) {
return 2;
}
}
else if (Rs >= 1101) {
return 2;
}
}
else if (mcTcor >= 84) {
if (refDCval < 91) {
if (diffAFD < 4) {
if (MVDiff < 142) {
if (RsCsDiff < 78) {
if (ssDCval < 72) {
if (MVDiff < 59) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
return 1;
}
}
else if (MVDiff >= 59) {
return 4;
}
}
else if (ssDCval >= 72) {
if (ssDCval < 85) {
if (Cs < 583) {
return 1;
}
else if (Cs >= 583) {
if (MVDiff < 30) {
return 1;
}
else if (MVDiff >= 30) {
return 2;
}
}
}
else if (ssDCval >= 85) {
if (Rs < 1172) {
if (Rs < 178) {
return 1;
}
else if (Rs >= 178) {
return 4;
}
}
else if (Rs >= 1172) {
return 2;
}
}
}
}
else if (RsCsDiff >= 78) {
if (negBalance < 55) {
if (diffRsCsdiff < 30) {
if (diffRsCsdiff < -39) {
return 4;
}
else if (diffRsCsdiff >= -39) {
return 2;
}
}
else if (diffRsCsdiff >= 30) {
if (Rs < 533) {
return 2;
}
else if (Rs >= 533) {
return 4;
}
}
}
else if (negBalance >= 55) {
if (Cs < 1136) {
return 4;
}
else if (Cs >= 1136) {
return 2;
}
}
}
}
else if (MVDiff >= 142) {
if (RsCsDiff < 76) {
return 4;
}
else if (RsCsDiff >= 76) {
return 8;
}
}
}
else if (diffAFD >= 4) {
if (posBalance < 76) {
if (diffAFD < 11) {
return 2;
}
else if (diffAFD >= 11) {
if (Rs < 721) {
return 2;
}
else if (Rs >= 721) {
return 4;
}
}
}
else if (posBalance >= 76) {
if (Rs < 396) {
return 4;
}
else if (Rs >= 396) {
return 1;
}
}
}
}
else if (refDCval >= 91) {
if (posBalance < 67) {
return 2;
}
else if (posBalance >= 67) {
if (AFD < 25) {
if (Cs < 695) {
return 8;
}
else if (Cs >= 695) {
return 2;
}
}
else if (AFD >= 25) {
return 1;
}
}
}
}
}
else if (Rs >= 1214) {
if (Cs < 1114) {
if (Rs < 1770) {
if (ssDCval < 82) {
if (Rs < 1379) {
return 4;
}
else if (Rs >= 1379) {
return 1;
}
}
else if (ssDCval >= 82) {
return 1;
}
}
else if (Rs >= 1770) {
return 8;
}
}
else if (Cs >= 1114) {
if (Rs < 2181) {
return 4;
}
else if (Rs >= 2181) {
return 2;
}
}
}
}
else if (RsCsDiff >= 129) {
if (diffAFD < 2) {
if (CsDiff < 12) {
return 1;
}
else if (CsDiff >= 12) {
return 4;
}
}
else if (diffAFD >= 2) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 2;
}
}
}
}
else if (mcTcor >= 96) {
if (TSC < 8) {
if (Cs < 892) {
return 1;
}
else if (Cs >= 892) {
return 4;
}
}
else if (TSC >= 8) {
if (diffMVdiffVal < 14) {
return 8;
}
else if (diffMVdiffVal >= 14) {
if (Rs < 1420) {
return 1;
}
else if (Rs >= 1420) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (posBalance < 65) {
if (Rs < 1589) {
if (Rs < 858) {
if (tcor < 52) {
return 8;
}
else if (tcor >= 52) {
return 2;
}
}
else if (Rs >= 858) {
if (mu_mv_mag_sq < 17) {
if (mu_mv_mag_sq < 10) {
if (Cs < 1029) {
if (RsCsDiff < 143) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
if (Rs < 1421) {
return 1;
}
else if (Rs >= 1421) {
return 8;
}
}
}
else if (RsCsDiff >= 143) {
if (Cs < 995) {
if (Rs < 1091) {
return 8;
}
else if (Rs >= 1091) {
return 2;
}
}
else if (Cs >= 995) {
if (Rs < 1186) {
return 1;
}
else if (Rs >= 1186) {
return 8;
}
}
}
}
else if (Cs >= 1029) {
if (AFD < 27) {
return 4;
}
else if (AFD >= 27) {
if (Rs < 1432) {
return 8;
}
else if (Rs >= 1432) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (tcor < 58) {
if (Cs < 765) {
if (Rs < 1223) {
return 8;
}
else if (Rs >= 1223) {
return 2;
}
}
else if (Cs >= 765) {
return 4;
}
}
else if (tcor >= 58) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (AFD < 34) {
if (tcor < 77) {
if (TSC < 11) {
if (diffRsCsdiff < 27) {
return 4;
}
else if (diffRsCsdiff >= 27) {
if (negBalance < 58) {
return 4;
}
else if (negBalance >= 58) {
return 8;
}
}
}
else if (TSC >= 11) {
if (MVDiff < 44) {
if (Rs < 1569) {
return 4;
}
else if (Rs >= 1569) {
return 8;
}
}
else if (MVDiff >= 44) {
if (MVDiff < 58) {
return 1;
}
else if (MVDiff >= 58) {
return 2;
}
}
}
}
else if (tcor >= 77) {
return 2;
}
}
else if (AFD >= 34) {
return 2;
}
}
}
}
else if (Rs >= 1589) {
if (diffRsCsdiff < -45) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (SC < 3143) {
return 2;
}
else if (SC >= 3143) {
return 8;
}
}
}
else if (diffRsCsdiff >= -45) {
if (RsCsDiff < 167) {
if (mu_mv_mag_sq < 24) {
if (MVDiff < 43) {
if (TSC < 11) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 1;
}
}
else if (TSC >= 11) {
if (Rs < 2246) {
return 1;
}
else if (Rs >= 2246) {
return 2;
}
}
}
else if (MVDiff >= 43) {
if (Rs < 1637) {
return 1;
}
else if (Rs >= 1637) {
if (RsCsDiff < 108) {
return 4;
}
else if (RsCsDiff >= 108) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
if (Rs < 1624) {
return 2;
}
else if (Rs >= 1624) {
return 1;
}
}
}
}
else if (RsCsDiff >= 167) {
if (Rs < 2318) {
if (Cs < 832) {
if (diffAFD < 1) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 1;
}
}
else if (diffAFD >= 1) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 1;
}
}
}
else if (Cs >= 832) {
if (AFD < 28) {
if (Rs < 2053) {
if (negBalance < 57) {
if (posBalance < 58) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
return 8;
}
}
else if (posBalance >= 58) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
if (Rs < 2027) {
return 4;
}
else if (Rs >= 2027) {
return 8;
}
}
}
}
else if (negBalance >= 57) {
return 4;
}
}
else if (Rs >= 2053) {
return 4;
}
}
else if (AFD >= 28) {
if (Rs < 1836) {
return 2;
}
else if (Rs >= 1836) {
return 4;
}
}
}
}
else if (Rs >= 2318) {
return 8;
}
}
}
}
}
else if (posBalance >= 65) {
if (gchDC < 5) {
if (RsCsDiff < 217) {
if (AFD < 27) {
if (tcor < 78) {
if (negBalance < 50) {
return 4;
}
else if (negBalance >= 50) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
if (Cs < 651) {
return 8;
}
else if (Cs >= 651) {
return 4;
}
}
}
}
}
else if (tcor >= 78) {
return 2;
}
}
else if (AFD >= 27) {
if (Rs < 1216) {
return 4;
}
else if (Rs >= 1216) {
return 2;
}
}
}
else if (RsCsDiff >= 217) {
if (SC < 3290) {
if (RsCsDiff < 227) {
return 1;
}
else if (RsCsDiff >= 227) {
return 2;
}
}
else if (SC >= 3290) {
return 8;
}
}
}
else if (gchDC >= 5) {
if (mcTcor < 93) {
if (SC < 1718) {
return 1;
}
else if (SC >= 1718) {
if (ssDCval < 123) {
if (Rs < 1543) {
if (Rs < 1040) {
return 8;
}
else if (Rs >= 1040) {
if (diffRsCsdiff < 14) {
return 8;
}
else if (diffRsCsdiff >= 14) {
return 4;
}
}
}
else if (Rs >= 1543) {
return 8;
}
}
else if (ssDCval >= 123) {
return 4;
}
}
}
else if (mcTcor >= 93) {
if (posBalance < 74) {
return 2;
}
else if (posBalance >= 74) {
if (ssDCval < 105) {
return 8;
}
else if (ssDCval >= 105) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 113) {
if (Rs < 1427) {
if (AFD < 23) {
if (TSC < 11) {
if (Rs < 670) {
if (TSC < 8) {
if (mu_mv_mag_sq < 14) {
if (Rs < 490) {
return 1;
}
else if (Rs >= 490) {
return 4;
}
}
else if (mu_mv_mag_sq >= 14) {
if (Rs < 441) {
return 2;
}
else if (Rs >= 441) {
return 8;
}
}
}
else if (TSC >= 8) {
return 2;
}
}
else if (Rs >= 670) {
if (mcTcor < 95) {
if (posBalance < 91) {
if (refDCval < 116) {
return 2;
}
else if (refDCval >= 116) {
if (SC < 2092) {
if (AFD < 14) {
if (Rs < 1091) {
return 1;
}
else if (Rs >= 1091) {
if (Rs < 1161) {
return 4;
}
else if (Rs >= 1161) {
return 8;
}
}
}
else if (AFD >= 14) {
return 8;
}
}
else if (SC >= 2092) {
if (mu_mv_mag_sq < 23) {
if (Rs < 1144) {
if (Cs < 1125) {
return 8;
}
else if (Cs >= 1125) {
return 1;
}
}
else if (Rs >= 1144) {
return 8;
}
}
else if (mu_mv_mag_sq >= 23) {
if (CsDiff < 7) {
return 1;
}
else if (CsDiff >= 7) {
if (Rs < 1313) {
return 8;
}
else if (Rs >= 1313) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 91) {
return 2;
}
}
else if (mcTcor >= 95) {
if (ssDCval < 157) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (Rs < 1238) {
return 2;
}
else if (Rs >= 1238) {
return 8;
}
}
}
else if (ssDCval >= 157) {
return 2;
}
}
}
}
else if (TSC >= 11) {
if (RsDiff < 7) {
if (gchDC < 8) {
return 2;
}
else if (gchDC >= 8) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 8;
}
}
}
else if (RsDiff >= 7) {
return 4;
}
}
}
else if (AFD >= 23) {
if (RsCsDiff < 93) {
if (ssDCval < 125) {
if (Rs < 784) {
return 2;
}
else if (Rs >= 784) {
return 8;
}
}
else if (ssDCval >= 125) {
if (negBalance < 55) {
if (AFD < 34) {
return 8;
}
else if (AFD >= 34) {
return 1;
}
}
else if (negBalance >= 55) {
return 1;
}
}
}
else if (RsCsDiff >= 93) {
if (mu_mv_mag_sq < 171) {
if (SC < 2280) {
if (Rs < 975) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 1;
}
}
else if (Rs >= 975) {
return 2;
}
}
else if (SC >= 2280) {
if (Rs < 1303) {
return 2;
}
else if (Rs >= 1303) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 171) {
return 1;
}
}
}
}
else if (Rs >= 1427) {
if (RsDiff < 10) {
if (AFD < 19) {
if (refDCval < 116) {
if (Rs < 1953) {
return 1;
}
else if (Rs >= 1953) {
return 8;
}
}
else if (refDCval >= 116) {
if (Rs < 1877) {
if (mu_mv_mag_sq < 34) {
return 8;
}
else if (mu_mv_mag_sq >= 34) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 1662) {
return 1;
}
else if (Rs >= 1662) {
return 8;
}
}
}
}
else if (Rs >= 1877) {
if (tcor < 86) {
if (negBalance < 56) {
return 1;
}
else if (negBalance >= 56) {
return 8;
}
}
else if (tcor >= 86) {
return 1;
}
}
}
}
else if (AFD >= 19) {
if (TSC < 11) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (TSC >= 11) {
if (Rs < 1995) {
return 4;
}
else if (Rs >= 1995) {
return 8;
}
}
}
}
else if (RsDiff >= 10) {
if (negBalance < 50) {
if (ssDCval < 136) {
if (MVDiff < 35) {
if (Cs < 852) {
return 1;
}
else if (Cs >= 852) {
return 2;
}
}
else if (MVDiff >= 35) {
return 1;
}
}
else if (ssDCval >= 136) {
if (diffAFD < -1) {
if (Cs < 1027) {
return 1;
}
else if (Cs >= 1027) {
return 2;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
}
else if (negBalance >= 50) {
if (tcor < 78) {
if (TSC < 9) {
if (CsDiff < 7) {
if (AFD < 24) {
if (Cs < 855) {
return 4;
}
else if (Cs >= 855) {
return 1;
}
}
else if (AFD >= 24) {
return 1;
}
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (TSC >= 9) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 156) {
return 1;
}
else if (MVDiff >= 156) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffRsCsdiff < -72) {
if (Cs < 1046) {
return 2;
}
else if (Cs >= 1046) {
return 4;
}
}
else if (diffRsCsdiff >= -72) {
if (diffAFD < 3) {
if (Rs < 2006) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (Rs >= 2006) {
if (AFD < 32) {
return 8;
}
else if (AFD >= 32) {
return 1;
}
}
}
else if (diffAFD >= 3) {
if (AFD < 29) {
if (Rs < 2081) {
return 8;
}
else if (Rs >= 2081) {
return 1;
}
}
else if (AFD >= 29) {
return 1;
}
}
}
}
}
}
else if (tcor >= 78) {
if (Rs < 1812) {
if (MVDiff < 29) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (Rs < 1730) {
return 2;
}
else if (Rs >= 1730) {
return 1;
}
}
}
else if (MVDiff >= 29) {
return 1;
}
}
else if (Rs >= 1812) {
if (ssDCval < 151) {
return 1;
}
else if (ssDCval >= 151) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (SC < 2150) {
if (Rs < 1002) {
if (refDCval < 64) {
if (Rs < 543) {
if (SC < 1090) {
if (MVDiff < 3) {
if (ssDCval < 58) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (Cs < 560) {
if (Rs < 289) {
return 1;
}
else if (Rs >= 289) {
return 8;
}
}
else if (Cs >= 560) {
return 1;
}
}
}
else if (ssDCval >= 58) {
if (Rs < 277) {
return 4;
}
else if (Rs >= 277) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (AFD < 7) {
if (SC < 872) {
return 4;
}
else if (SC >= 872) {
if (mu_mv_mag_sq < 58) {
return 4;
}
else if (mu_mv_mag_sq >= 58) {
if (diffAFD < 1) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (diffAFD >= 1) {
if (Rs < 492) {
return 4;
}
else if (Rs >= 492) {
return 1;
}
}
}
}
}
else if (AFD >= 7) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (Rs < 304) {
return 8;
}
else if (Rs >= 304) {
return 2;
}
}
}
}
}
else if (SC >= 1090) {
if (negBalance < 19) {
if (Rs < 315) {
return 2;
}
else if (Rs >= 315) {
if (Cs < 625) {
return 2;
}
else if (Cs >= 625) {
return 1;
}
}
}
else if (negBalance >= 19) {
if (CsDiff < 1) {
if (posBalance < 21) {
return 8;
}
else if (posBalance >= 21) {
return 4;
}
}
else if (CsDiff >= 1) {
if (negBalance < 26) {
if (AFD < 4) {
if (Rs < 531) {
return 2;
}
else if (Rs >= 531) {
return 8;
}
}
else if (AFD >= 4) {
return 4;
}
}
else if (negBalance >= 26) {
if (ssDCval < 59) {
if (diffRsCsdiff < -6) {
return 2;
}
else if (diffRsCsdiff >= -6) {
if (ssDCval < 48) {
if (Rs < 479) {
return 1;
}
else if (Rs >= 479) {
return 4;
}
}
else if (ssDCval >= 48) {
return 4;
}
}
}
else if (ssDCval >= 59) {
if (Rs < 453) {
return 1;
}
else if (Rs >= 453) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 543) {
if (MVDiff < 1) {
if (SC < 1564) {
if (refDCval < 54) {
if (mu_mv_mag_sq < 118) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
return 4;
}
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
else if (refDCval >= 54) {
return 8;
}
}
else if (SC >= 1564) {
if (ssDCval < 49) {
if (Rs < 647) {
return 1;
}
else if (Rs >= 647) {
return 8;
}
}
else if (ssDCval >= 49) {
if (Rs < 663) {
if (Cs < 1076) {
return 1;
}
else if (Cs >= 1076) {
return 4;
}
}
else if (Rs >= 663) {
if (RsCsDiff < 2) {
if (Cs < 640) {
return 8;
}
else if (Cs >= 640) {
return 1;
}
}
else if (RsCsDiff >= 2) {
if (Cs < 795) {
if (Rs < 997) {
return 1;
}
else if (Rs >= 997) {
return 4;
}
}
else if (Cs >= 795) {
if (Cs < 1094) {
if (ssDCval < 50) {
if (Cs < 908) {
return 8;
}
else if (Cs >= 908) {
return 1;
}
}
else if (ssDCval >= 50) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 1;
}
}
}
else if (Cs >= 1094) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 17) {
if (ssDCval < 50) {
if (Rs < 835) {
return 2;
}
else if (Rs >= 835) {
return 8;
}
}
else if (ssDCval >= 50) {
return 1;
}
}
else if (negBalance >= 17) {
if (Rs < 781) {
if (negBalance < 30) {
if (CsDiff < 1) {
if (Rs < 601) {
return 4;
}
else if (Rs >= 601) {
return 1;
}
}
else if (CsDiff >= 1) {
if (ssDCval < 51) {
if (Rs < 736) {
return 1;
}
else if (Rs >= 736) {
return 4;
}
}
else if (ssDCval >= 51) {
if (RsDiff < 2) {
if (posBalance < 50) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 635) {
return 8;
}
else if (Rs >= 635) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (Cs < 1017) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Rs < 637) {
return 8;
}
else if (Rs >= 637) {
return 4;
}
}
}
else if (Cs >= 1017) {
return 2;
}
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (negBalance >= 30) {
if (RsDiff < 2) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 111) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 644) {
return 1;
}
else if (Rs >= 644) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 111) {
return 1;
}
}
else if (negBalance >= 37) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (Cs < 675) {
return 8;
}
else if (Cs >= 675) {
return 4;
}
}
}
}
else if (RsDiff >= 2) {
if (MVDiff < 15) {
if (Cs < 839) {
if (negBalance < 42) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 54) {
return 4;
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 683) {
return 4;
}
else if (Rs >= 683) {
if (Rs < 729) {
return 1;
}
else if (Rs >= 729) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
if (Rs < 617) {
return 1;
}
else if (Rs >= 617) {
if (Rs < 669) {
return 2;
}
else if (Rs >= 669) {
return 8;
}
}
}
}
else if (negBalance >= 42) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (Cs >= 839) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
return 8;
}
}
}
else if (MVDiff >= 15) {
if (Rs < 706) {
return 1;
}
else if (Rs >= 706) {
return 8;
}
}
}
}
}
else if (Rs >= 781) {
if (refDCval < 52) {
if (Cs < 566) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
else if (Cs >= 566) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 827) {
return 4;
}
else if (Rs >= 827) {
return 1;
}
}
}
}
else if (refDCval >= 52) {
if (posBalance < 30) {
return 4;
}
else if (posBalance >= 30) {
if (RsCsDiff < 7) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (Cs < 639) {
return 1;
}
else if (Cs >= 639) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (tcor < 93) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 1205) {
return 8;
}
else if (Cs >= 1205) {
return 1;
}
}
}
else if (tcor >= 93) {
if (Cs < 732) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 8;
}
}
else if (Cs >= 732) {
if (Rs < 976) {
return 4;
}
else if (Rs >= 976) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 64) {
if (Rs < 834) {
if (Rs < 717) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 78) {
if (Cs < 785) {
if (Rs < 313) {
return 1;
}
else if (Rs >= 313) {
if (ssDCval < 71) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (ssDCval >= 71) {
if (diffTSC < 1) {
if (TSC < 4) {
if (Cs < 544) {
return 1;
}
else if (Cs >= 544) {
return 4;
}
}
else if (TSC >= 4) {
if (Cs < 528) {
return 4;
}
else if (Cs >= 528) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 519) {
return 8;
}
else if (Rs >= 519) {
return 1;
}
}
}
}
}
else if (Cs >= 785) {
if (mu_mv_mag_sq < 13) {
if (Cs < 1237) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (Cs >= 1237) {
if (Rs < 607) {
return 1;
}
else if (Rs >= 607) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (mu_mv_mag_sq < 17) {
return 1;
}
else if (mu_mv_mag_sq >= 17) {
if (RsCsDiff < 7) {
return 1;
}
else if (RsCsDiff >= 7) {
if (Cs < 840) {
return 4;
}
else if (Cs >= 840) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (Cs < 676) {
if (Rs < 613) {
if (CsDiff < 1) {
if (Cs < 591) {
return 1;
}
else if (Cs >= 591) {
return 4;
}
}
else if (CsDiff >= 1) {
if (posBalance < 55) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (posBalance >= 55) {
return 2;
}
}
}
else if (Rs >= 613) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
else if (Cs >= 676) {
if (SC < 1198) {
if (Rs < 347) {
return 8;
}
else if (Rs >= 347) {
return 4;
}
}
else if (SC >= 1198) {
if (Rs < 696) {
if (tcor < 99) {
if (mu_mv_mag_sq < 105) {
if (ssDCval < 81) {
if (Cs < 985) {
return 8;
}
else if (Cs >= 985) {
return 1;
}
}
else if (ssDCval >= 81) {
return 1;
}
}
else if (mu_mv_mag_sq >= 105) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (tcor >= 99) {
return 8;
}
}
else if (Rs >= 696) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (Rs < 650) {
if (TSC < 4) {
if (Cs < 571) {
return 1;
}
else if (Cs >= 571) {
if (posBalance < 41) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
if (Rs < 584) {
return 1;
}
else if (Rs >= 584) {
return 2;
}
}
}
else if (posBalance >= 41) {
return 2;
}
}
}
else if (TSC >= 4) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 349) {
return 1;
}
else if (Rs >= 349) {
return 8;
}
}
}
}
else if (Rs >= 650) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (Rs < 667) {
return 2;
}
else if (Rs >= 667) {
return 4;
}
}
}
}
}
else if (Rs >= 717) {
if (refDCval < 83) {
if (RsCsDiff < 11) {
if (Cs < 655) {
if (AFD < 4) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 71) {
if (Rs < 736) {
return 4;
}
else if (Rs >= 736) {
return 8;
}
}
else if (mu_mv_mag_sq >= 71) {
return 4;
}
}
}
else if (Cs >= 655) {
if (ssDCval < 74) {
if (Cs < 879) {
if (Cs < 733) {
if (TSC < 2) {
if (Rs < 735) {
return 8;
}
else if (Rs >= 735) {
return 4;
}
}
else if (TSC >= 2) {
return 1;
}
}
else if (Cs >= 733) {
if (tcor < 99) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 2;
}
}
else if (tcor >= 99) {
if (Rs < 730) {
return 8;
}
else if (Rs >= 730) {
return 4;
}
}
}
}
else if (Cs >= 879) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 776) {
return 8;
}
else if (Rs >= 776) {
if (Rs < 809) {
return 2;
}
else if (Rs >= 809) {
return 1;
}
}
}
}
}
else if (ssDCval >= 74) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
}
}
else if (RsCsDiff >= 11) {
if (Cs < 925) {
if (posBalance < 51) {
if (Rs < 747) {
return 4;
}
else if (Rs >= 747) {
return 2;
}
}
else if (posBalance >= 51) {
return 4;
}
}
else if (Cs >= 925) {
return 2;
}
}
}
else if (refDCval >= 83) {
if (Cs < 832) {
if (AFD < 9) {
if (Rs < 775) {
return 2;
}
else if (Rs >= 775) {
return 8;
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (Cs >= 832) {
return 4;
}
}
}
}
else if (Rs >= 834) {
if (refDCval < 73) {
if (Cs < 745) {
if (mu_mv_mag_sq < 85) {
if (MVDiff < 23) {
if (Cs < 705) {
return 1;
}
else if (Cs >= 705) {
return 8;
}
}
else if (MVDiff >= 23) {
if (Rs < 890) {
return 8;
}
else if (Rs >= 890) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (negBalance < 14) {
if (Rs < 936) {
return 2;
}
else if (Rs >= 936) {
return 8;
}
}
else if (negBalance >= 14) {
return 2;
}
}
}
else if (Cs >= 745) {
if (RsCsDiff < 7) {
if (negBalance < 34) {
if (tcor < 96) {
if (Rs < 911) {
return 4;
}
else if (Rs >= 911) {
return 8;
}
}
else if (tcor >= 96) {
if (Cs < 859) {
return 4;
}
else if (Cs >= 859) {
return 1;
}
}
}
else if (negBalance >= 34) {
if (Rs < 890) {
return 2;
}
else if (Rs >= 890) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (CsDiff < 4) {
if (SC < 2142) {
return 4;
}
else if (SC >= 2142) {
return 2;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
else if (refDCval >= 73) {
if (tcor < 96) {
if (mu_mv_mag_sq < 5) {
if (Rs < 868) {
return 4;
}
else if (Rs >= 868) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
if (negBalance < 38) {
if (Rs < 958) {
return 2;
}
else if (Rs >= 958) {
return 8;
}
}
else if (negBalance >= 38) {
if (Cs < 531) {
return 2;
}
else if (Cs >= 531) {
return 4;
}
}
}
}
else if (tcor >= 96) {
if (ssDCval < 86) {
if (Rs < 967) {
return 8;
}
else if (Rs >= 967) {
if (mu_mv_mag_sq < 114) {
return 8;
}
else if (mu_mv_mag_sq >= 114) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (Cs < 876) {
return 8;
}
else if (Cs >= 876) {
return 4;
}
}
}
}
}
else if (ssDCval >= 86) {
if (Rs < 884) {
return 4;
}
else if (Rs >= 884) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 1002) {
if (mu_mv_mag_sq < 127) {
if (refDCval < 69) {
if (ssDCval < 51) {
if (Rs < 1233) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (Rs >= 1233) {
if (ssDCval < 47) {
if (Rs < 1420) {
if (Rs < 1335) {
return 4;
}
else if (Rs >= 1335) {
return 4;
}
}
else if (Rs >= 1420) {
return 4;
}
}
else if (ssDCval >= 47) {
return 1;
}
}
}
else if (ssDCval >= 51) {
if (posBalance < 55) {
if (SC < 1801) {
if (refDCval < 65) {
return 1;
}
else if (refDCval >= 65) {
if (AFD < 3) {
if (mcTcor < 99) {
if (Cs < 667) {
return 2;
}
else if (Cs >= 667) {
return 4;
}
}
else if (mcTcor >= 99) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
return 2;
}
}
}
else if (AFD >= 3) {
if (Rs < 1012) {
return 8;
}
else if (Rs >= 1012) {
return 1;
}
}
}
}
else if (SC >= 1801) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 103) {
if (diffTSC < -1) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (SC < 1916) {
return 8;
}
else if (SC >= 1916) {
return 1;
}
}
}
else if (diffTSC >= -1) {
if (TSC < 5) {
if (MVDiff < 13) {
return 1;
}
else if (MVDiff >= 13) {
if (Rs < 1258) {
return 8;
}
else if (Rs >= 1258) {
return 1;
}
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (posBalance < 29) {
if (Rs < 1315) {
return 8;
}
else if (Rs >= 1315) {
if (Rs < 1369) {
return 1;
}
else if (Rs >= 1369) {
return 2;
}
}
}
else if (posBalance >= 29) {
if (tcor < 91) {
if (SC < 1883) {
if (Rs < 1173) {
return 1;
}
else if (Rs >= 1173) {
return 2;
}
}
else if (SC >= 1883) {
return 8;
}
}
else if (tcor >= 91) {
if (mu_mv_mag_sq < 126) {
if (MVDiff < 23) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Cs < 708) {
return 1;
}
else if (Cs >= 708) {
if (Cs < 768) {
if (Rs < 1363) {
return 8;
}
else if (Rs >= 1363) {
return 1;
}
}
else if (Cs >= 768) {
return 1;
}
}
}
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 1123) {
return 4;
}
else if (Rs >= 1123) {
if (Cs < 782) {
return 1;
}
else if (Cs >= 782) {
return 8;
}
}
}
}
}
}
}
}
else if (posBalance >= 55) {
if (Cs < 1006) {
if (refDCval < 66) {
if (CsDiff < 4) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (Rs < 1045) {
return 8;
}
else if (Rs >= 1045) {
return 1;
}
}
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (refDCval >= 66) {
return 1;
}
}
else if (Cs >= 1006) {
return 4;
}
}
}
}
else if (refDCval >= 69) {
if (mu_mv_mag_sq < 52) {
if (Cs < 895) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (Cs >= 895) {
return 2;
}
}
else if (mu_mv_mag_sq >= 52) {
if (refDCval < 76) {
if (ssDCval < 72) {
if (Rs < 1108) {
if (Cs < 932) {
if (Cs < 792) {
return 2;
}
else if (Cs >= 792) {
return 8;
}
}
else if (Cs >= 932) {
return 1;
}
}
else if (Rs >= 1108) {
return 8;
}
}
else if (ssDCval >= 72) {
if (Cs < 770) {
if (Cs < 584) {
if (Cs < 583) {
if (Cs < 559) {
return 8;
}
else if (Cs >= 559) {
return 1;
}
}
else if (Cs >= 583) {
return 8;
}
}
else if (Cs >= 584) {
return 1;
}
}
else if (Cs >= 770) {
return 8;
}
}
}
else if (refDCval >= 76) {
if (Cs < 989) {
if (Rs < 1390) {
return 8;
}
else if (Rs >= 1390) {
return 4;
}
}
else if (Cs >= 989) {
if (RsCsDiff < 7) {
return 4;
}
else if (RsCsDiff >= 7) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 1160) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (Rs < 1033) {
return 8;
}
else if (Rs >= 1033) {
if (SC < 1827) {
return 2;
}
else if (SC >= 1827) {
return 4;
}
}
}
}
else if (Rs >= 1160) {
if (SC < 1910) {
if (posBalance < 36) {
if (Cs < 509) {
return 2;
}
else if (Cs >= 509) {
return 1;
}
}
else if (posBalance >= 36) {
if (Rs < 1299) {
return 1;
}
else if (Rs >= 1299) {
return 8;
}
}
}
else if (SC >= 1910) {
if (ssDCval < 74) {
if (Rs < 1232) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (Rs < 1201) {
return 8;
}
else if (Rs >= 1201) {
return 1;
}
}
}
else if (Rs >= 1232) {
if (Cs < 524) {
if (Rs < 1417) {
return 8;
}
else if (Rs >= 1417) {
return 2;
}
}
else if (Cs >= 524) {
return 8;
}
}
}
else if (ssDCval >= 74) {
return 1;
}
}
}
}
}
}
else if (SC >= 2150) {
if (Rs < 1803) {
if (Cs < 1059) {
if (mu_mv_mag_sq < 121) {
if (SC < 2453) {
if (AFD < 3) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (AFD >= 3) {
if (Cs < 829) {
if (mu_mv_mag_sq < 104) {
if (negBalance < 47) {
if (diffTSC < 1) {
if (diffMVdiffVal < -3) {
if (Cs < 651) {
return 2;
}
else if (Cs >= 651) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (diffTSC >= 1) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Rs < 1633) {
if (Rs < 1541) {
return 8;
}
else if (Rs >= 1541) {
return 1;
}
}
else if (Rs >= 1633) {
return 2;
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 1665) {
if (Rs < 1619) {
return 8;
}
else if (Rs >= 1619) {
return 4;
}
}
else if (Rs >= 1665) {
if (Rs < 1690) {
return 2;
}
else if (Rs >= 1690) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 1712) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (Rs >= 1712) {
return 4;
}
}
}
else if (Cs >= 829) {
if (ssDCval < 53) {
if (Cs < 983) {
return 1;
}
else if (Cs >= 983) {
return 4;
}
}
else if (ssDCval >= 53) {
if (negBalance < 68) {
if (Rs < 1473) {
if (Rs < 1141) {
if (Rs < 1133) {
return 8;
}
else if (Rs >= 1133) {
return 1;
}
}
else if (Rs >= 1141) {
return 8;
}
}
else if (Rs >= 1473) {
if (Rs < 1497) {
return 1;
}
else if (Rs >= 1497) {
return 8;
}
}
}
else if (negBalance >= 68) {
return 4;
}
}
}
}
}
else if (SC >= 2453) {
if (MVDiff < 5) {
if (negBalance < 21) {
if (Rs < 1688) {
return 1;
}
else if (Rs >= 1688) {
return 8;
}
}
else if (negBalance >= 21) {
if (diffAFD < -2) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (Cs < 859) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (Cs >= 859) {
if (SC < 2581) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (SC >= 2581) {
if (Rs < 1734) {
return 1;
}
else if (Rs >= 1734) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (Rs < 1722) {
if (Rs < 1327) {
if (negBalance < 19) {
if (Cs < 920) {
if (Rs < 1283) {
return 1;
}
else if (Rs >= 1283) {
return 8;
}
}
else if (Cs >= 920) {
return 1;
}
}
else if (negBalance >= 19) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
}
else if (Rs >= 1327) {
if (ssDCval < 49) {
if (Rs < 1547) {
return 4;
}
else if (Rs >= 1547) {
if (Rs < 1600) {
return 8;
}
else if (Rs >= 1600) {
if (Cs < 1024) {
return 4;
}
else if (Cs >= 1024) {
if (Rs < 1674) {
return 8;
}
else if (Rs >= 1674) {
return 8;
}
}
}
}
}
else if (ssDCval >= 49) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (posBalance < 38) {
if (SC < 2281) {
return 8;
}
else if (SC >= 2281) {
return 1;
}
}
else if (posBalance >= 38) {
return 8;
}
}
}
}
}
else if (Rs >= 1722) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 1;
}
}
}
}
else if (Cs >= 1059) {
if (RsDiff < 1) {
if (mu_mv_mag_sq < 106) {
if (Cs < 1136) {
return 4;
}
else if (Cs >= 1136) {
return 8;
}
}
else if (mu_mv_mag_sq >= 106) {
if (ssDCval < 55) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
return 4;
}
}
else if (ssDCval >= 55) {
return 1;
}
}
}
else if (RsDiff >= 1) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 126) {
if (refDCval < 76) {
if (ssDCval < 57) {
if (negBalance < 25) {
if (TSC < 2) {
if (Rs < 1253) {
return 1;
}
else if (Rs >= 1253) {
return 8;
}
}
else if (TSC >= 2) {
if (mcTcor < 99) {
if (Rs < 1751) {
if (SC < 2842) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (Cs < 1129) {
if (Rs < 1613) {
return 8;
}
else if (Rs >= 1613) {
if (diffRsCsdiff < 2) {
if (Rs < 1677) {
return 4;
}
else if (Rs >= 1677) {
return 4;
}
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
}
else if (Cs >= 1129) {
if (SC < 2830) {
return 4;
}
else if (SC >= 2830) {
return 4;
}
}
}
}
}
else if (SC >= 2842) {
return 8;
}
}
else if (Rs >= 1751) {
return 4;
}
}
else if (mcTcor >= 99) {
return 4;
}
}
}
else if (negBalance >= 25) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
if (Rs < 1353) {
if (Rs < 1024) {
return 8;
}
else if (Rs >= 1024) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
return 4;
}
}
}
else if (Rs >= 1353) {
return 8;
}
}
}
}
else if (ssDCval >= 57) {
if (AFD < 3) {
if (SC < 2377) {
if (Rs < 1221) {
if (Rs < 1115) {
return 4;
}
else if (Rs >= 1115) {
return 1;
}
}
else if (Rs >= 1221) {
return 8;
}
}
else if (SC >= 2377) {
return 1;
}
}
else if (AFD >= 3) {
if (diffRsCsdiff < 6) {
if (RsCsDiff < 7) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 1525) {
return 1;
}
else if (Rs >= 1525) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (Rs < 1283) {
if (refDCval < 69) {
return 4;
}
else if (refDCval >= 69) {
if (Rs < 993) {
return 2;
}
else if (Rs >= 993) {
if (Rs < 1269) {
return 8;
}
else if (Rs >= 1269) {
return 4;
}
}
}
}
else if (Rs >= 1283) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 6) {
if (Rs < 1186) {
if (Rs < 1071) {
if (Cs < 1199) {
return 8;
}
else if (Cs >= 1199) {
return 2;
}
}
else if (Rs >= 1071) {
return 4;
}
}
else if (Rs >= 1186) {
return 1;
}
}
}
}
}
else if (refDCval >= 76) {
if (Rs < 1566) {
if (Cs < 1146) {
return 8;
}
else if (Cs >= 1146) {
return 4;
}
}
else if (Rs >= 1566) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (ssDCval < 74) {
if (ssDCval < 59) {
if (SC < 2952) {
if (negBalance < 24) {
if (Cs < 1152) {
if (SC < 2894) {
if (diffRsCsdiff < 2) {
if (diffRsCsdiff < 1) {
if (posBalance < 38) {
if (gchDC < 1) {
if (posBalance < 37) {
if (Rs < 1684) {
if (Rs < 1660) {
return 8;
}
else if (Rs >= 1660) {
if (Rs < 1681) {
if (negBalance < 17) {
return 8;
}
else if (negBalance >= 17) {
if (Rs < 1679) {
if (Rs < 1677) {
return 8;
}
else if (Rs >= 1677) {
return 8;
}
}
else if (Rs >= 1679) {
return 4;
}
}
}
else if (Rs >= 1681) {
return 8;
}
}
}
else if (Rs >= 1684) {
if (Rs < 1728) {
return 4;
}
else if (Rs >= 1728) {
if (SC < 2799) {
return 4;
}
else if (SC >= 2799) {
if (SC < 2871) {
if (negBalance < 20) {
return 4;
}
else if (negBalance >= 20) {
if (SC < 2864) {
if (Rs < 1735) {
return 8;
}
else if (Rs >= 1735) {
if (SC < 2836) {
return 4;
}
else if (SC >= 2836) {
if (SC < 2857) {
return 4;
}
else if (SC >= 2857) {
return 4;
}
}
}
}
else if (SC >= 2864) {
return 4;
}
}
}
else if (SC >= 2871) {
if (SC < 2881) {
if (Rs < 1764) {
return 8;
}
else if (Rs >= 1764) {
return 8;
}
}
else if (SC >= 2881) {
if (SC < 2883) {
return 4;
}
else if (SC >= 2883) {
if (Rs < 1762) {
return 4;
}
else if (Rs >= 1762) {
if (Rs < 1788) {
return 8;
}
else if (Rs >= 1788) {
return 4;
}
}
}
}
}
}
}
}
}
else if (posBalance >= 37) {
return 8;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (posBalance >= 38) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
if (posBalance < 40) {
if (Rs < 1744) {
return 4;
}
else if (Rs >= 1744) {
return 4;
}
}
else if (posBalance >= 40) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
else if (SC >= 2894) {
if (Rs < 1778) {
return 8;
}
else if (Rs >= 1778) {
return 4;
}
}
}
else if (Cs >= 1152) {
if (tcor < 99) {
if (negBalance < 19) {
if (Cs < 1167) {
if (Rs < 1727) {
return 4;
}
else if (Rs >= 1727) {
return 4;
}
}
else if (Cs >= 1167) {
return 8;
}
}
else if (negBalance >= 19) {
if (Cs < 1192) {
return 4;
}
else if (Cs >= 1192) {
if (Rs < 1479) {
return 4;
}
else if (Rs >= 1479) {
return 8;
}
}
}
}
else if (tcor >= 99) {
return 4;
}
}
}
else if (negBalance >= 24) {
if (SC < 2877) {
return 8;
}
else if (SC >= 2877) {
if (Rs < 1767) {
return 4;
}
else if (Rs >= 1767) {
return 8;
}
}
}
}
else if (SC >= 2952) {
return 4;
}
}
else if (ssDCval >= 59) {
if (Rs < 1608) {
if (posBalance < 45) {
if (Cs < 1109) {
return 8;
}
else if (Cs >= 1109) {
return 4;
}
}
else if (posBalance >= 45) {
if (Cs < 1199) {
return 4;
}
else if (Cs >= 1199) {
if (SC < 2561) {
return 1;
}
else if (SC >= 2561) {
return 4;
}
}
}
}
else if (Rs >= 1608) {
return 8;
}
}
}
else if (ssDCval >= 74) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (Rs < 1286) {
if (mu_mv_mag_sq < 97) {
return 4;
}
else if (mu_mv_mag_sq >= 97) {
return 8;
}
}
else if (Rs >= 1286) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Rs < 1594) {
return 8;
}
else if (Rs >= 1594) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1803) {
if (TSC < 4) {
if (refDCval < 82) {
if (ssDCval < 71) {
if (ssDCval < 53) {
if (CsDiff < 2) {
if (Cs < 1069) {
return 8;
}
else if (Cs >= 1069) {
if (Cs < 1127) {
if (Rs < 1828) {
return 4;
}
else if (Rs >= 1828) {
return 8;
}
}
else if (Cs >= 1127) {
if (Rs < 2036) {
return 4;
}
else if (Rs >= 2036) {
return 8;
}
}
}
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (ssDCval >= 53) {
if (Cs < 1185) {
if (posBalance < 36) {
if (Rs < 1928) {
if (SC < 2969) {
if (Cs < 966) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (Cs >= 966) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
if (mu_mv_mag_sq < 127) {
if (Rs < 1869) {
return 1;
}
else if (Rs >= 1869) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
return 4;
}
}
}
}
else if (SC >= 2969) {
return 1;
}
}
else if (Rs >= 1928) {
if (SC < 3186) {
if (negBalance < 14) {
if (Cs < 937) {
return 1;
}
else if (Cs >= 937) {
return 4;
}
}
else if (negBalance >= 14) {
return 4;
}
}
else if (SC >= 3186) {
return 8;
}
}
}
else if (posBalance >= 36) {
if (negBalance < 14) {
return 4;
}
else if (negBalance >= 14) {
if (mu_mv_mag_sq < 99) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
return 1;
}
}
else if (mu_mv_mag_sq >= 99) {
return 1;
}
}
}
}
else if (Cs >= 1185) {
if (mu_mv_mag_sq < 125) {
return 2;
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
}
}
else if (ssDCval >= 71) {
if (posBalance < 48) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
if (Rs < 1913) {
return 1;
}
else if (Rs >= 1913) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (Cs < 834) {
if (Rs < 1927) {
return 1;
}
else if (Rs >= 1927) {
return 8;
}
}
else if (Cs >= 834) {
return 8;
}
}
}
}
else if (refDCval >= 82) {
if (mu_mv_mag_sq < 100) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
return 1;
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 902) {
if (Rs < 2223) {
return 4;
}
else if (Rs >= 2223) {
return 8;
}
}
else if (Cs >= 902) {
return 8;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 883) {
if (refDCval < 71) {
if (ssDCval < 61) {
if (Cs < 552) {
return 4;
}
else if (Cs >= 552) {
return 1;
}
}
else if (ssDCval >= 61) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (refDCval >= 71) {
if (posBalance < 33) {
if (Rs < 2038) {
if (Rs < 1939) {
return 2;
}
else if (Rs >= 1939) {
return 1;
}
}
else if (Rs >= 2038) {
if (RsCsDiff < 18) {
return 4;
}
else if (RsCsDiff >= 18) {
return 4;
}
}
}
else if (posBalance >= 33) {
if (SC < 2815) {
return 1;
}
else if (SC >= 2815) {
if (RsCsDiff < 56) {
if (mu_mv_mag_sq < 54) {
return 1;
}
else if (mu_mv_mag_sq >= 54) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (RsCsDiff >= 56) {
return 2;
}
}
}
}
}
else if (Cs >= 883) {
if (negBalance < 46) {
if (mu_mv_mag_sq < 75) {
if (SC < 3114) {
if (Cs < 948) {
return 1;
}
else if (Cs >= 948) {
return 4;
}
}
else if (SC >= 3114) {
if (Rs < 2324) {
return 1;
}
else if (Rs >= 2324) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
return 8;
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 102) {
if (SC < 1310) {
if (negBalance < 32) {
if (negBalance < 20) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (ssDCval < 120) {
return 1;
}
else if (ssDCval >= 120) {
if (Rs < 505) {
return 2;
}
else if (Rs >= 505) {
if (Cs < 519) {
return 1;
}
else if (Cs >= 519) {
return 4;
}
}
}
}
}
else if (negBalance >= 20) {
if (diffTSC < 1) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (diffAFD < -1) {
if (Rs < 554) {
return 2;
}
else if (Rs >= 554) {
return 1;
}
}
else if (diffAFD >= -1) {
if (MVDiff < 2) {
if (Rs < 541) {
return 1;
}
else if (Rs >= 541) {
return 2;
}
}
else if (MVDiff >= 2) {
if (Rs < 622) {
return 2;
}
else if (Rs >= 622) {
return 1;
}
}
}
}
}
else if (diffTSC >= 1) {
if (Rs < 458) {
if (Rs < 394) {
return 4;
}
else if (Rs >= 394) {
return 2;
}
}
else if (Rs >= 458) {
return 1;
}
}
}
}
else if (negBalance >= 32) {
if (refDCval < 115) {
if (MVDiff < 3) {
if (Cs < 740) {
return 1;
}
else if (Cs >= 740) {
if (Rs < 414) {
return 4;
}
else if (Rs >= 414) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 598) {
if (TSC < 5) {
if (Cs < 792) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 2;
}
}
else if (Cs >= 792) {
if (Rs < 365) {
return 4;
}
else if (Rs >= 365) {
return 8;
}
}
}
else if (TSC >= 5) {
if (diffAFD < -1) {
if (Rs < 453) {
return 1;
}
else if (Rs >= 453) {
if (Rs < 492) {
return 8;
}
else if (Rs >= 492) {
return 2;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (Rs >= 598) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (SC < 1217) {
return 2;
}
else if (SC >= 1217) {
return 4;
}
}
}
}
}
else if (refDCval >= 115) {
if (ssDCval < 151) {
if (posBalance < 27) {
return 8;
}
else if (posBalance >= 27) {
if (mcTcor < 98) {
if (CsDiff < 3) {
if (ssDCval < 117) {
return 2;
}
else if (ssDCval >= 117) {
return 8;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (mcTcor >= 98) {
if (ssDCval < 125) {
if (diffRsCsdiff < -7) {
if (Rs < 452) {
return 4;
}
else if (Rs >= 452) {
return 1;
}
}
else if (diffRsCsdiff >= -7) {
if (mu_mv_mag_sq < 97) {
return 2;
}
else if (mu_mv_mag_sq >= 97) {
return 4;
}
}
}
else if (ssDCval >= 125) {
if (AFD < 6) {
if (Rs < 306) {
return 2;
}
else if (Rs >= 306) {
return 8;
}
}
else if (AFD >= 6) {
return 2;
}
}
}
}
}
else if (ssDCval >= 151) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 493) {
return 4;
}
else if (Rs >= 493) {
if (Rs < 672) {
return 2;
}
else if (Rs >= 672) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 1310) {
if (RsCsDiff < 67) {
if (negBalance < 36) {
if (ssDCval < 112) {
if (MVDiff < 2) {
if (ssDCval < 105) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
}
else if (ssDCval >= 105) {
return 1;
}
}
else if (MVDiff >= 2) {
if (SC < 1356) {
if (Rs < 519) {
return 1;
}
else if (Rs >= 519) {
return 2;
}
}
else if (SC >= 1356) {
if (Rs < 1683) {
if (ssDCval < 107) {
if (mu_mv_mag_sq < 68) {
if (Rs < 1514) {
return 1;
}
else if (Rs >= 1514) {
return 4;
}
}
else if (mu_mv_mag_sq >= 68) {
return 8;
}
}
else if (ssDCval >= 107) {
return 1;
}
}
else if (Rs >= 1683) {
return 1;
}
}
}
}
else if (ssDCval >= 112) {
if (RsCsDiff < 15) {
if (Cs < 922) {
if (Cs < 623) {
return 1;
}
else if (Cs >= 623) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (Cs >= 922) {
if (AFD < 2) {
if (Rs < 854) {
return 8;
}
else if (Rs >= 854) {
return 4;
}
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (RsCsDiff >= 15) {
if (Rs < 1720) {
if (TSC < 6) {
if (Rs < 698) {
return 2;
}
else if (Rs >= 698) {
return 8;
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (Rs >= 1720) {
if (Cs < 529) {
return 1;
}
else if (Cs >= 529) {
if (SC < 3235) {
return 2;
}
else if (SC >= 3235) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 36) {
if (mu_mv_mag_sq < 5) {
if (Cs < 971) {
if (RsDiff < 6) {
if (tcor < 95) {
if (ssDCval < 115) {
if (Cs < 776) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (Rs < 935) {
return 4;
}
else if (Rs >= 935) {
return 2;
}
}
}
else if (Cs >= 776) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 4;
}
}
}
else if (ssDCval >= 115) {
if (SC < 1907) {
if (tcor < 82) {
return 8;
}
else if (tcor >= 82) {
if (Cs < 552) {
return 2;
}
else if (Cs >= 552) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Rs < 1025) {
return 1;
}
else if (Rs >= 1025) {
return 8;
}
}
}
}
}
else if (SC >= 1907) {
if (Rs < 1049) {
return 4;
}
else if (Rs >= 1049) {
if (Rs < 1846) {
return 8;
}
else if (Rs >= 1846) {
return 2;
}
}
}
}
}
else if (tcor >= 95) {
if (Rs < 582) {
return 2;
}
else if (Rs >= 582) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (CsDiff < 4) {
if (Cs < 509) {
return 2;
}
else if (Cs >= 509) {
return 1;
}
}
else if (CsDiff >= 4) {
if (Rs < 1424) {
return 2;
}
else if (Rs >= 1424) {
return 4;
}
}
}
}
else if (Cs >= 971) {
if (Rs < 1117) {
return 4;
}
else if (Rs >= 1117) {
if (Rs < 2177) {
if (Rs < 1437) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (Rs >= 1437) {
if (MVDiff < 5) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (MVDiff >= 5) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
return 1;
}
}
}
}
else if (Rs >= 2177) {
if (Cs < 1152) {
return 4;
}
else if (Cs >= 1152) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (MVDiff < 2) {
if (SC < 1900) {
if (Rs < 594) {
return 1;
}
else if (Rs >= 594) {
return 4;
}
}
else if (SC >= 1900) {
if (ssDCval < 95) {
return 1;
}
else if (ssDCval >= 95) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 156) {
if (AFD < 9) {
if (MVDiff < 6) {
if (RsCsDiff < 2) {
if (Rs < 725) {
return 1;
}
else if (Rs >= 725) {
return 4;
}
}
else if (RsCsDiff >= 2) {
if (posBalance < 48) {
if (posBalance < 47) {
if (RsDiff < 5) {
if (Rs < 881) {
if (Rs < 592) {
return 8;
}
else if (Rs >= 592) {
return 1;
}
}
else if (Rs >= 881) {
return 8;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (posBalance >= 47) {
return 1;
}
}
else if (posBalance >= 48) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (SC < 1487) {
if (SC < 1478) {
return 8;
}
else if (SC >= 1478) {
return 2;
}
}
else if (SC >= 1487) {
return 8;
}
}
}
else if (AFD >= 9) {
if (Rs < 2098) {
if (SC < 1544) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (SC >= 1544) {
if (AFD < 18) {
if (diffAFD < -2) {
if (diffMVdiffVal < -1) {
if (Rs < 1487) {
return 4;
}
else if (Rs >= 1487) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (diffAFD >= -2) {
if (posBalance < 69) {
if (diffRsCsdiff < -20) {
if (Rs < 1003) {
return 1;
}
else if (Rs >= 1003) {
return 8;
}
}
else if (diffRsCsdiff >= -20) {
return 8;
}
}
else if (posBalance >= 69) {
return 2;
}
}
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (Rs >= 2098) {
if (CsDiff < 4) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 1;
}
}
else if (CsDiff >= 4) {
if (Cs < 1020) {
return 4;
}
else if (Cs >= 1020) {
return 8;
}
}
}
}
}
else if (ssDCval >= 156) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 67) {
if (Cs < 691) {
if (Rs < 1380) {
if (TSC < 8) {
if (ssDCval < 105) {
return 8;
}
else if (ssDCval >= 105) {
return 2;
}
}
else if (TSC >= 8) {
if (Cs < 594) {
return 1;
}
else if (Cs >= 594) {
return 4;
}
}
}
else if (Rs >= 1380) {
if (TSC < 6) {
if (Rs < 1851) {
return 8;
}
else if (Rs >= 1851) {
return 1;
}
}
else if (TSC >= 6) {
if (gchDC < 8) {
if (mu_mv_mag_sq < 39) {
return 1;
}
else if (mu_mv_mag_sq >= 39) {
if (Rs < 2134) {
return 1;
}
else if (Rs >= 2134) {
return 2;
}
}
}
else if (gchDC >= 8) {
if (Cs < 628) {
return 1;
}
else if (Cs >= 628) {
return 8;
}
}
}
}
}
else if (Cs >= 691) {
if (tcor < 93) {
if (CsDiff < 6) {
if (negBalance < 45) {
if (Rs < 1919) {
return 2;
}
else if (Rs >= 1919) {
return 4;
}
}
else if (negBalance >= 45) {
if (AFD < 19) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (AFD >= 19) {
if (Rs < 1696) {
return 2;
}
else if (Rs >= 1696) {
return 1;
}
}
}
}
else if (CsDiff >= 6) {
if (ssDCval < 106) {
if (ssDCval < 96) {
if (Cs < 880) {
return 2;
}
else if (Cs >= 880) {
if (Cs < 1178) {
return 8;
}
else if (Cs >= 1178) {
return 2;
}
}
}
else if (ssDCval >= 96) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
return 1;
}
}
}
else if (ssDCval >= 106) {
if (ssDCval < 140) {
return 2;
}
else if (ssDCval >= 140) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
}
else if (tcor >= 93) {
if (Cs < 1313) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (Cs >= 1313) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (ssDCval < 133) {
if (Cs < 1227) {
if (Rs < 340) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 340) {
if (Cs < 948) {
if (Rs < 1084) {
if (mu_mv_mag_sq < 117) {
if (refDCval < 100) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (Rs < 778) {
if (posBalance < 45) {
if (Rs < 641) {
return 8;
}
else if (Rs >= 641) {
return 1;
}
}
else if (posBalance >= 45) {
return 1;
}
}
else if (Rs >= 778) {
return 8;
}
}
}
else if (refDCval >= 100) {
if (negBalance < 29) {
if (Rs < 758) {
if (Cs < 688) {
return 1;
}
else if (Cs >= 688) {
if (Rs < 559) {
return 4;
}
else if (Rs >= 559) {
return 8;
}
}
}
else if (Rs >= 758) {
return 8;
}
}
else if (negBalance >= 29) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
if (diffRsCsdiff < -1) {
if (Rs < 420) {
if (Rs < 357) {
return 8;
}
else if (Rs >= 357) {
return 4;
}
}
else if (Rs >= 420) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (posBalance < 58) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Rs < 464) {
return 1;
}
else if (Rs >= 464) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (Rs < 460) {
return 1;
}
else if (Rs >= 460) {
return 8;
}
}
}
}
else if (Rs >= 1084) {
if (ssDCval < 109) {
if (Rs < 1591) {
if (AFD < 3) {
if (mu_mv_mag_sq < 127) {
return 4;
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 669) {
return 4;
}
else if (Cs >= 669) {
return 8;
}
}
}
else if (AFD >= 3) {
if (Rs < 1582) {
return 8;
}
else if (Rs >= 1582) {
return 4;
}
}
}
else if (Rs >= 1591) {
if (Rs < 2143) {
if (RsDiff < 5) {
if (diffRsCsdiff < -15) {
return 1;
}
else if (diffRsCsdiff >= -15) {
return 8;
}
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (Rs >= 2143) {
if (refDCval < 89) {
return 4;
}
else if (refDCval >= 89) {
return 8;
}
}
}
}
else if (ssDCval >= 109) {
if (Cs < 761) {
return 1;
}
else if (Cs >= 761) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1606) {
return 8;
}
else if (Rs >= 1606) {
return 1;
}
}
}
}
}
}
else if (Cs >= 948) {
if (RsCsDiff < 9) {
if (diffRsCsdiff < -36) {
if (Rs < 1237) {
return 8;
}
else if (Rs >= 1237) {
return 2;
}
}
else if (diffRsCsdiff >= -36) {
if (mu_mv_mag_sq < 110) {
if (Cs < 970) {
return 1;
}
else if (Cs >= 970) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 110) {
return 8;
}
}
}
else if (RsCsDiff >= 9) {
if (Rs < 1715) {
if (tcor < 96) {
if (Rs < 1008) {
return 4;
}
else if (Rs >= 1008) {
return 8;
}
}
else if (tcor >= 96) {
return 8;
}
}
else if (Rs >= 1715) {
if (Rs < 1863) {
return 2;
}
else if (Rs >= 1863) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1227) {
if (Rs < 633) {
return 1;
}
else if (Rs >= 633) {
return 8;
}
}
}
else if (ssDCval >= 133) {
if (Rs < 1329) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 1162) {
return 2;
}
else if (Rs >= 1162) {
return 1;
}
}
}
else if (Rs >= 1329) {
if (Cs < 1049) {
return 1;
}
else if (Cs >= 1049) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2354) {
if (Rs < 3415) {
if (RsDiff < 7) {
if (Cs < 1086) {
if (Cs < 705) {
if (TSCindex < 1) {
if (Rs < 2658) {
return 4;
}
else if (Rs >= 2658) {
return 8;
}
}
else if (TSCindex >= 1) {
if (posBalance < 58) {
if (ssDCval < 66) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (Cs < 610) {
if (Cs < 387) {
return 8;
}
else if (Cs >= 387) {
return 4;
}
}
else if (Cs >= 610) {
if (Rs < 2452) {
return 2;
}
else if (Rs >= 2452) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
if (RsDiff < 5) {
if (CsDiff < 1) {
if (SC < 2916) {
return 8;
}
else if (SC >= 2916) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (SC < 3563) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
return 2;
}
}
else if (SC >= 3563) {
return 2;
}
}
}
}
else if (CsDiff >= 1) {
if (Rs < 2522) {
if (negBalance < 28) {
return 8;
}
else if (negBalance >= 28) {
if (Rs < 2507) {
return 1;
}
else if (Rs >= 2507) {
return 2;
}
}
}
else if (Rs >= 2522) {
return 8;
}
}
}
else if (RsDiff >= 5) {
if (diffRsCsdiff < -15) {
if (Rs < 2969) {
if (posBalance < 43) {
if (Cs < 240) {
return 1;
}
else if (Cs >= 240) {
return 2;
}
}
else if (posBalance >= 43) {
if (Rs < 2365) {
if (Rs < 2355) {
return 1;
}
else if (Rs >= 2355) {
return 2;
}
}
else if (Rs >= 2365) {
return 1;
}
}
}
else if (Rs >= 2969) {
return 4;
}
}
else if (diffRsCsdiff >= -15) {
if (mu_mv_mag_sq < 41) {
if (SC < 3183) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (SC >= 3183) {
if (mu_mv_mag_sq < 23) {
return 8;
}
else if (mu_mv_mag_sq >= 23) {
if (SC < 3557) {
return 1;
}
else if (SC >= 3557) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (Cs < 658) {
if (Cs < 532) {
if (Rs < 2566) {
return 1;
}
else if (Rs >= 2566) {
return 8;
}
}
else if (Cs >= 532) {
return 8;
}
}
else if (Cs >= 658) {
if (Rs < 2844) {
return 8;
}
else if (Rs >= 2844) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 58) {
if (tcor < 96) {
if (mu_mv_mag_sq < 49) {
if (ssDCval < 93) {
return 1;
}
else if (ssDCval >= 93) {
return 8;
}
}
else if (mu_mv_mag_sq >= 49) {
if (Rs < 2486) {
return 1;
}
else if (Rs >= 2486) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
}
else if (tcor >= 96) {
return 2;
}
}
}
}
else if (Cs >= 705) {
if (refDCval < 78) {
if (negBalance < 23) {
if (mu_mv_mag_sq < 106) {
return 4;
}
else if (mu_mv_mag_sq >= 106) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (Rs < 2672) {
return 2;
}
else if (Rs >= 2672) {
return 4;
}
}
}
}
else if (negBalance >= 23) {
if (Rs < 3003) {
if (Cs < 957) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
if (Cs < 918) {
return 1;
}
else if (Cs >= 918) {
return 2;
}
}
}
else if (Cs >= 957) {
if (Rs < 2469) {
if (Cs < 1040) {
if (Rs < 2446) {
return 4;
}
else if (Rs >= 2446) {
return 2;
}
}
else if (Cs >= 1040) {
return 8;
}
}
else if (Rs >= 2469) {
if (diffRsCsdiff < -37) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (Rs < 2821) {
return 2;
}
else if (Rs >= 2821) {
return 4;
}
}
}
else if (diffRsCsdiff >= -37) {
if (refDCval < 63) {
if (ssDCval < 62) {
if (Rs < 2851) {
return 2;
}
else if (Rs >= 2851) {
return 4;
}
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (refDCval >= 63) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 3003) {
if (refDCval < 65) {
if (Rs < 3010) {
return 4;
}
else if (Rs >= 3010) {
return 2;
}
}
else if (refDCval >= 65) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 2;
}
}
}
}
}
else if (refDCval >= 78) {
if (Rs < 2829) {
if (RsCsDiff < 27) {
if (diffRsCsdiff < -29) {
return 1;
}
else if (diffRsCsdiff >= -29) {
if (ssDCval < 120) {
if (posBalance < 22) {
return 4;
}
else if (posBalance >= 22) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
}
else if (ssDCval >= 120) {
return 2;
}
}
}
else if (RsCsDiff >= 27) {
if (TSC < 6) {
if (tcor < 96) {
if (ssDCval < 104) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 4;
}
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (tcor >= 96) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
return 2;
}
}
}
else if (TSC >= 6) {
if (ssDCval < 85) {
if (Cs < 792) {
return 4;
}
else if (Cs >= 792) {
return 2;
}
}
else if (ssDCval >= 85) {
if (diffAFD < -11) {
return 1;
}
else if (diffAFD >= -11) {
if (CsDiff < 3) {
if (Rs < 2483) {
return 8;
}
else if (Rs >= 2483) {
return 2;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2829) {
if (tcor < 96) {
if (TSC < 5) {
if (ssDCval < 117) {
return 1;
}
else if (ssDCval >= 117) {
if (Rs < 3278) {
return 8;
}
else if (Rs >= 3278) {
return 2;
}
}
}
else if (TSC >= 5) {
if (AFD < 9) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 3312) {
if (Rs < 3139) {
return 4;
}
else if (Rs >= 3139) {
return 4;
}
}
else if (Rs >= 3312) {
return 8;
}
}
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (tcor >= 96) {
if (MVDiff < 8) {
if (negBalance < 24) {
if (Cs < 969) {
if (Rs < 2931) {
return 2;
}
else if (Rs >= 2931) {
return 8;
}
}
else if (Cs >= 969) {
if (Rs < 3121) {
return 4;
}
else if (Rs >= 3121) {
return 2;
}
}
}
else if (negBalance >= 24) {
return 2;
}
}
else if (MVDiff >= 8) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 1086) {
if (TSCindex < 2) {
if (ssDCval < 128) {
if (Rs < 3158) {
return 8;
}
else if (Rs >= 3158) {
if (Rs < 3252) {
return 2;
}
else if (Rs >= 3252) {
return 8;
}
}
}
else if (ssDCval >= 128) {
return 1;
}
}
else if (TSCindex >= 2) {
if (mu_mv_mag_sq < 87) {
if (refDCval < 92) {
if (Rs < 3001) {
if (SC < 3766) {
if (RsCsDiff < 43) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
else if (RsCsDiff >= 43) {
if (Rs < 2393) {
return 8;
}
else if (Rs >= 2393) {
return 2;
}
}
}
else if (SC >= 3766) {
if (mcTcor < 95) {
if (Rs < 2713) {
return 8;
}
else if (Rs >= 2713) {
return 2;
}
}
else if (mcTcor >= 95) {
if (diffTSC < 1) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
else if (diffTSC >= 1) {
if (Rs < 2790) {
return 2;
}
else if (Rs >= 2790) {
return 4;
}
}
}
}
}
else if (Rs >= 3001) {
if (mcTcor < 97) {
return 4;
}
else if (mcTcor >= 97) {
return 2;
}
}
}
else if (refDCval >= 92) {
if (mu_mv_mag_sq < 61) {
if (Rs < 2383) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
else if (Rs >= 2383) {
return 8;
}
}
else if (mu_mv_mag_sq >= 61) {
if (MVDiff < 5) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Rs < 2678) {
return 8;
}
else if (Rs >= 2678) {
return 2;
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 135) {
return 8;
}
else if (ssDCval >= 135) {
if (Cs < 1191) {
if (Rs < 2398) {
return 4;
}
else if (Rs >= 2398) {
return 8;
}
}
else if (Cs >= 1191) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (ssDCval < 108) {
if (RsCsDiff < 7) {
if (Cs < 1120) {
if (Rs < 2873) {
return 8;
}
else if (Rs >= 2873) {
return 2;
}
}
else if (Cs >= 1120) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (diffTSC < -1) {
if (ssDCval < 69) {
if (Cs < 1240) {
return 4;
}
else if (Cs >= 1240) {
if (Rs < 2837) {
return 8;
}
else if (Rs >= 2837) {
return 2;
}
}
}
else if (ssDCval >= 69) {
if (negBalance < 39) {
return 2;
}
else if (negBalance >= 39) {
if (CsDiff < 3) {
if (Rs < 2617) {
return 8;
}
else if (Rs >= 2617) {
return 2;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
}
}
else if (diffTSC >= -1) {
if (Rs < 2975) {
if (ssDCval < 65) {
if (Cs < 1170) {
if (Rs < 2371) {
return 8;
}
else if (Rs >= 2371) {
return 4;
}
}
else if (Cs >= 1170) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (ssDCval >= 65) {
if (ssDCval < 82) {
if (RsDiff < 4) {
if (Rs < 2621) {
return 4;
}
else if (Rs >= 2621) {
return 4;
}
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
}
else if (Rs >= 2975) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
return 2;
}
}
}
}
}
else if (ssDCval >= 108) {
if (Rs < 2812) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (Cs < 1196) {
return 8;
}
else if (Cs >= 1196) {
return 1;
}
}
}
}
else if (Rs >= 2812) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (refDCval < 92) {
if (posBalance < 49) {
if (negBalance < 45) {
if (tcor < 91) {
if (CsDiff < 5) {
if (refDCval < 67) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
if (ssDCval < 66) {
if (TSC < 9) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (TSC >= 9) {
return 2;
}
}
else if (ssDCval >= 66) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
return 2;
}
}
}
}
else if (refDCval >= 67) {
if (posBalance < 42) {
if (Cs < 818) {
return 8;
}
else if (Cs >= 818) {
return 1;
}
}
else if (posBalance >= 42) {
if (ssDCval < 77) {
if (mu_mv_mag_sq < 35) {
return 4;
}
else if (mu_mv_mag_sq >= 35) {
if (Rs < 2837) {
return 2;
}
else if (Rs >= 2837) {
return 1;
}
}
}
else if (ssDCval >= 77) {
return 1;
}
}
}
}
else if (CsDiff >= 5) {
if (ssDCval < 74) {
if (Cs < 1043) {
if (diffRsCsdiff < 36) {
if (Rs < 2931) {
if (SC < 3694) {
if (gchDC < 2) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
if (Rs < 2422) {
return 4;
}
else if (Rs >= 2422) {
return 2;
}
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (SC >= 3694) {
return 2;
}
}
else if (Rs >= 2931) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (diffRsCsdiff >= 36) {
if (Rs < 3360) {
return 4;
}
else if (Rs >= 3360) {
if (Rs < 3385) {
return 1;
}
else if (Rs >= 3385) {
return 4;
}
}
}
}
else if (Cs >= 1043) {
if (tcor < 79) {
return 2;
}
else if (tcor >= 79) {
if (diffRsCsdiff < 16) {
if (Rs < 2390) {
return 1;
}
else if (Rs >= 2390) {
return 2;
}
}
else if (diffRsCsdiff >= 16) {
if (Rs < 3350) {
if (Cs < 1090) {
return 2;
}
else if (Cs >= 1090) {
return 4;
}
}
else if (Rs >= 3350) {
return 2;
}
}
}
}
}
else if (ssDCval >= 74) {
if (Rs < 2669) {
if (Rs < 2500) {
return 8;
}
else if (Rs >= 2500) {
return 1;
}
}
else if (Rs >= 2669) {
return 4;
}
}
}
}
else if (tcor >= 91) {
if (posBalance < 45) {
if (SC < 4093) {
if (Cs < 325) {
return 2;
}
else if (Cs >= 325) {
if (posBalance < 43) {
if (Cs < 881) {
return 4;
}
else if (Cs >= 881) {
return 1;
}
}
else if (posBalance >= 43) {
if (diffRsCsdiff < 56) {
if (SC < 3328) {
return 1;
}
else if (SC >= 3328) {
return 4;
}
}
else if (diffRsCsdiff >= 56) {
return 1;
}
}
}
}
else if (SC >= 4093) {
if (Cs < 1247) {
return 2;
}
else if (Cs >= 1247) {
return 8;
}
}
}
else if (posBalance >= 45) {
if (negBalance < 37) {
if (SC < 4126) {
return 4;
}
else if (SC >= 4126) {
return 1;
}
}
else if (negBalance >= 37) {
if (Rs < 3295) {
return 2;
}
else if (Rs >= 3295) {
return 4;
}
}
}
}
}
else if (negBalance >= 45) {
if (negBalance < 51) {
if (posBalance < 46) {
if (Rs < 2794) {
if (mu_mv_mag_sq < 23) {
return 1;
}
else if (mu_mv_mag_sq >= 23) {
if (Cs < 786) {
return 1;
}
else if (Cs >= 786) {
if (Rs < 2475) {
return 2;
}
else if (Rs >= 2475) {
return 8;
}
}
}
}
else if (Rs >= 2794) {
if (RsDiff < 11) {
if (Cs < 999) {
return 8;
}
else if (Cs >= 999) {
return 2;
}
}
else if (RsDiff >= 11) {
return 4;
}
}
}
else if (posBalance >= 46) {
if (AFD < 16) {
if (Rs < 2431) {
if (Rs < 2429) {
return 2;
}
else if (Rs >= 2429) {
return 4;
}
}
else if (Rs >= 2431) {
return 2;
}
}
else if (AFD >= 16) {
return 4;
}
}
}
else if (negBalance >= 51) {
if (MVDiff < 26) {
if (posBalance < 34) {
if (diffRsCsdiff < 43) {
return 2;
}
else if (diffRsCsdiff >= 43) {
return 1;
}
}
else if (posBalance >= 34) {
return 2;
}
}
else if (MVDiff >= 26) {
if (Rs < 2791) {
if (Cs < 966) {
if (Rs < 2474) {
return 1;
}
else if (Rs >= 2474) {
if (Rs < 2591) {
return 4;
}
else if (Rs >= 2591) {
return 2;
}
}
}
else if (Cs >= 966) {
if (SC < 3987) {
return 8;
}
else if (SC >= 3987) {
return 4;
}
}
}
else if (Rs >= 2791) {
return 2;
}
}
}
}
}
else if (posBalance >= 49) {
if (MVDiff < 29) {
if (refDCval < 60) {
if (RsCsDiff < 142) {
return 4;
}
else if (RsCsDiff >= 142) {
if (Cs < 1017) {
return 2;
}
else if (Cs >= 1017) {
return 4;
}
}
}
else if (refDCval >= 60) {
if (negBalance < 42) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 32) {
if (MVDiff < 5) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (MVDiff >= 5) {
if (MVDiff < 14) {
if (ssDCval < 79) {
if (Cs < 1284) {
return 4;
}
else if (Cs >= 1284) {
return 2;
}
}
else if (ssDCval >= 79) {
if (Rs < 2486) {
return 4;
}
else if (Rs >= 2486) {
return 1;
}
}
}
else if (MVDiff >= 14) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (negBalance < 35) {
if (Rs < 2779) {
return 2;
}
else if (Rs >= 2779) {
return 4;
}
}
else if (negBalance >= 35) {
if (Cs < 379) {
return 1;
}
else if (Cs >= 379) {
return 2;
}
}
}
}
else if (posBalance >= 53) {
if (Cs < 524) {
if (posBalance < 64) {
if (ssDCval < 79) {
if (Rs < 2906) {
if (Cs < 274) {
return 4;
}
else if (Cs >= 274) {
return 2;
}
}
else if (Rs >= 2906) {
return 4;
}
}
else if (ssDCval >= 79) {
return 1;
}
}
else if (posBalance >= 64) {
return 2;
}
}
else if (Cs >= 524) {
if (gchDC < 2) {
if (AFD < 18) {
if (AFD < 7) {
if (Rs < 2952) {
return 2;
}
else if (Rs >= 2952) {
return 4;
}
}
else if (AFD >= 7) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (Cs < 586) {
return 4;
}
else if (Cs >= 586) {
return 2;
}
}
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (gchDC >= 2) {
if (posBalance < 59) {
if (Cs < 869) {
if (Rs < 3262) {
return 2;
}
else if (Rs >= 3262) {
return 4;
}
}
else if (Cs >= 869) {
return 1;
}
}
else if (posBalance >= 59) {
if (TSC < 6) {
if (Rs < 2919) {
return 2;
}
else if (Rs >= 2919) {
return 1;
}
}
else if (TSC >= 6) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 42) {
if (mcTcor < 93) {
if (RsDiff < 13) {
if (Cs < 1059) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 2;
}
}
else if (Cs >= 1059) {
if (Rs < 2379) {
return 8;
}
else if (Rs >= 2379) {
return 4;
}
}
}
else if (RsDiff >= 13) {
if (ssDCval < 88) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 1;
}
}
else if (ssDCval >= 88) {
if (Rs < 2635) {
return 4;
}
else if (Rs >= 2635) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
if (CsDiff < 1) {
if (Rs < 3124) {
return 1;
}
else if (Rs >= 3124) {
return 4;
}
}
else if (CsDiff >= 1) {
if (diffRsCsdiff < -127) {
if (Rs < 2648) {
return 4;
}
else if (Rs >= 2648) {
return 2;
}
}
else if (diffRsCsdiff >= -127) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 29) {
if (RsDiff < 14) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
if (Cs < 320) {
return 4;
}
else if (Cs >= 320) {
if (diffTSC < -2) {
if (MVDiff < 51) {
return 2;
}
else if (MVDiff >= 51) {
return 8;
}
}
else if (diffTSC >= -2) {
if (AFD < 22) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (MVDiff < 50) {
if (gchDC < 2) {
if (Rs < 2934) {
return 1;
}
else if (Rs >= 2934) {
return 8;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (MVDiff >= 50) {
return 8;
}
}
}
else if (AFD >= 22) {
return 1;
}
}
}
}
}
else if (RsDiff >= 14) {
if (TSC < 12) {
if (MVDiff < 49) {
return 2;
}
else if (MVDiff >= 49) {
if (posBalance < 60) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2569) {
return 4;
}
else if (Rs >= 2569) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
else if (posBalance >= 60) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 4;
}
}
}
}
else if (TSC >= 12) {
if (RsDiff < 21) {
if (refDCval < 88) {
return 1;
}
else if (refDCval >= 88) {
if (Cs < 937) {
if (Rs < 2877) {
return 8;
}
else if (Rs >= 2877) {
return 2;
}
}
else if (Cs >= 937) {
return 4;
}
}
}
else if (RsDiff >= 21) {
if (SC < 3201) {
return 2;
}
else if (SC >= 3201) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 92) {
if (mcTcor < 70) {
if (mu_mv_mag_sq < 23) {
if (Cs < 908) {
return 2;
}
else if (Cs >= 908) {
return 4;
}
}
else if (mu_mv_mag_sq >= 23) {
return 1;
}
}
else if (mcTcor >= 70) {
if (mcTcor < 95) {
if (RsCsDiff < 121) {
if (Cs < 752) {
if (MVDiff < 28) {
if (Cs < 532) {
return 2;
}
else if (Cs >= 532) {
return 8;
}
}
else if (MVDiff >= 28) {
if (refDCval < 101) {
if (RsCsDiff < 103) {
return 1;
}
else if (RsCsDiff >= 103) {
return 8;
}
}
else if (refDCval >= 101) {
if (CsDiff < 3) {
if (SC < 2772) {
return 1;
}
else if (SC >= 2772) {
return 8;
}
}
else if (CsDiff >= 3) {
if (Rs < 2669) {
return 1;
}
else if (Rs >= 2669) {
return 4;
}
}
}
}
}
else if (Cs >= 752) {
if (RsDiff < 10) {
if (MVDiff < 36) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 8;
}
}
else if (MVDiff >= 36) {
if (Rs < 2664) {
return 2;
}
else if (Rs >= 2664) {
return 4;
}
}
}
else if (RsDiff >= 10) {
if (ssDCval < 101) {
if (Cs < 1252) {
return 8;
}
else if (Cs >= 1252) {
return 4;
}
}
else if (ssDCval >= 101) {
if (Rs < 2846) {
return 8;
}
else if (Rs >= 2846) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 121) {
if (ssDCval < 113) {
if (Cs < 868) {
if (MVDiff < 29) {
if (mu_mv_mag_sq < 45) {
if (Rs < 2382) {
if (Rs < 2364) {
return 2;
}
else if (Rs >= 2364) {
return 4;
}
}
else if (Rs >= 2382) {
return 2;
}
}
else if (mu_mv_mag_sq >= 45) {
if (Rs < 2988) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 2947) {
return 2;
}
else if (Rs >= 2947) {
return 8;
}
}
}
else if (Rs >= 2988) {
return 1;
}
}
}
else if (MVDiff >= 29) {
if (RsDiff < 17) {
if (TSC < 9) {
if (Rs < 3043) {
return 2;
}
else if (Rs >= 3043) {
return 8;
}
}
else if (TSC >= 9) {
if (AFD < 26) {
if (posBalance < 70) {
if (Cs < 417) {
if (Rs < 2491) {
return 8;
}
else if (Rs >= 2491) {
return 1;
}
}
else if (Cs >= 417) {
return 8;
}
}
else if (posBalance >= 70) {
if (Rs < 3210) {
return 1;
}
else if (Rs >= 3210) {
if (Rs < 3291) {
return 8;
}
else if (Rs >= 3291) {
return 4;
}
}
}
}
else if (AFD >= 26) {
if (Rs < 2514) {
return 1;
}
else if (Rs >= 2514) {
return 2;
}
}
}
}
else if (RsDiff >= 17) {
if (AFD < 41) {
return 2;
}
else if (AFD >= 41) {
return 1;
}
}
}
}
else if (Cs >= 868) {
if (negBalance < 50) {
if (MVDiff < 27) {
if (Cs < 1268) {
return 2;
}
else if (Cs >= 1268) {
return 8;
}
}
else if (MVDiff >= 27) {
if (Cs < 971) {
return 2;
}
else if (Cs >= 971) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (AFD < 24) {
if (tcor < 75) {
if (Rs < 2586) {
if (Rs < 2373) {
return 2;
}
else if (Rs >= 2373) {
return 4;
}
}
else if (Rs >= 2586) {
if (ssDCval < 109) {
if (TSC < 13) {
if (diffRsCsdiff < 58) {
return 8;
}
else if (diffRsCsdiff >= 58) {
return 4;
}
}
else if (TSC >= 13) {
return 4;
}
}
else if (ssDCval >= 109) {
return 4;
}
}
}
else if (tcor >= 75) {
if (mcTcor < 92) {
if (gchDC < 2) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (mcTcor >= 92) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
if (Cs < 1203) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (Cs >= 1203) {
return 8;
}
}
}
}
}
else if (AFD >= 24) {
if (mcTcor < 78) {
if (Cs < 1055) {
return 2;
}
else if (Cs >= 1055) {
if (Cs < 1259) {
return 4;
}
else if (Cs >= 1259) {
return 2;
}
}
}
else if (mcTcor >= 78) {
if (diffAFD < -2) {
if (AFD < 27) {
return 1;
}
else if (AFD >= 27) {
if (Rs < 2481) {
return 1;
}
else if (Rs >= 2481) {
return 2;
}
}
}
else if (diffAFD >= -2) {
if (Cs < 959) {
return 8;
}
else if (Cs >= 959) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (RsCsDiff < 269) {
if (refDCval < 133) {
if (negBalance < 52) {
if (SC < 3836) {
if (AFD < 25) {
return 1;
}
else if (AFD >= 25) {
return 8;
}
}
else if (SC >= 3836) {
return 8;
}
}
else if (negBalance >= 52) {
if (RsCsDiff < 237) {
if (diffRsCsdiff < -55) {
return 2;
}
else if (diffRsCsdiff >= -55) {
return 8;
}
}
else if (RsCsDiff >= 237) {
if (Cs < 1230) {
return 1;
}
else if (Cs >= 1230) {
return 2;
}
}
}
}
else if (refDCval >= 133) {
if (AFD < 19) {
if (diffAFD < -2) {
if (RsCsDiff < 131) {
return 2;
}
else if (RsCsDiff >= 131) {
return 1;
}
}
else if (diffAFD >= -2) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 4;
}
}
}
}
else if (AFD >= 19) {
if (diffMVdiffVal < 10) {
if (Rs < 2928) {
return 1;
}
else if (Rs >= 2928) {
if (Rs < 3149) {
return 8;
}
else if (Rs >= 3149) {
return 1;
}
}
}
else if (diffMVdiffVal >= 10) {
if (RsDiff < 12) {
return 2;
}
else if (RsDiff >= 12) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 269) {
if (Cs < 829) {
if (refDCval < 130) {
if (mcTcor < 93) {
return 2;
}
else if (mcTcor >= 93) {
if (TSC < 11) {
if (Rs < 2984) {
return 2;
}
else if (Rs >= 2984) {
return 1;
}
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (refDCval >= 130) {
if (Cs < 391) {
return 2;
}
else if (Cs >= 391) {
return 1;
}
}
}
else if (Cs >= 829) {
if (AFD < 27) {
if (ssDCval < 133) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 1;
}
}
else if (ssDCval >= 133) {
return 1;
}
}
else if (AFD >= 27) {
if (CsDiff < 8) {
if (Rs < 2835) {
return 1;
}
else if (Rs >= 2835) {
return 2;
}
}
else if (CsDiff >= 8) {
if (Cs < 910) {
return 2;
}
else if (Cs >= 910) {
return 1;
}
}
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (MVDiff < 7) {
if (Cs < 1108) {
if (tcor < 90) {
if (RsDiff < 13) {
if (Rs < 3014) {
return 8;
}
else if (Rs >= 3014) {
return 4;
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (tcor >= 90) {
return 2;
}
}
else if (Cs >= 1108) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (RsCsDiff < 82) {
return 8;
}
else if (RsCsDiff >= 82) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (refDCval < 111) {
if (RsCsDiff < 157) {
if (TSC < 6) {
if (Cs < 1017) {
return 2;
}
else if (Cs >= 1017) {
return 8;
}
}
else if (TSC >= 6) {
if (Rs < 2923) {
if (gchDC < 1) {
if (Rs < 2628) {
return 8;
}
else if (Rs >= 2628) {
return 4;
}
}
else if (gchDC >= 1) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
if (Cs < 1285) {
return 1;
}
else if (Cs >= 1285) {
return 8;
}
}
}
}
else if (Rs >= 2923) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 157) {
if (diffAFD < 6) {
if (gchDC < 8) {
if (Cs < 1269) {
if (diffMVdiffVal < -3) {
if (refDCval < 109) {
if (RsCsDiff < 182) {
if (Rs < 2717) {
return 2;
}
else if (Rs >= 2717) {
return 8;
}
}
else if (RsCsDiff >= 182) {
return 2;
}
}
else if (refDCval >= 109) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
else if (Cs >= 1269) {
return 8;
}
}
else if (gchDC >= 8) {
return 8;
}
}
else if (diffAFD >= 6) {
if (MVDiff < 35) {
if (Cs < 982) {
return 2;
}
else if (Cs >= 982) {
return 8;
}
}
else if (MVDiff >= 35) {
return 1;
}
}
}
}
else if (refDCval >= 111) {
if (ssDCval < 121) {
if (mu_mv_mag_sq < 15) {
if (MVDiff < 42) {
if (Rs < 3198) {
return 1;
}
else if (Rs >= 3198) {
if (Rs < 3346) {
return 2;
}
else if (Rs >= 3346) {
return 1;
}
}
}
else if (MVDiff >= 42) {
if (gchDC < 4) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (gchDC >= 4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (RsDiff < 11) {
if (diffAFD < -5) {
if (Cs < 871) {
return 2;
}
else if (Cs >= 871) {
return 8;
}
}
else if (diffAFD >= -5) {
if (diffMVdiffVal < -12) {
return 1;
}
else if (diffMVdiffVal >= -12) {
if (MVDiff < 12) {
if (Cs < 506) {
if (Cs < 281) {
return 1;
}
else if (Cs >= 281) {
return 8;
}
}
else if (Cs >= 506) {
if (Cs < 846) {
return 2;
}
else if (Cs >= 846) {
return 4;
}
}
}
else if (MVDiff >= 12) {
return 8;
}
}
}
}
else if (RsDiff >= 11) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
if (mu_mv_mag_sq < 66) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 1;
}
}
else if (mu_mv_mag_sq >= 66) {
if (Rs < 2887) {
return 2;
}
else if (Rs >= 2887) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 121) {
if (mu_mv_mag_sq < 35) {
if (CsDiff < 4) {
if (diffAFD < 3) {
if (CsDiff < 3) {
if (Rs < 3070) {
if (Cs < 124) {
return 2;
}
else if (Cs >= 124) {
return 1;
}
}
else if (Rs >= 3070) {
return 2;
}
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 32) {
return 1;
}
else if (mu_mv_mag_sq >= 32) {
if (Rs < 2543) {
return 1;
}
else if (Rs >= 2543) {
return 2;
}
}
}
}
else if (diffAFD >= 3) {
if (diffMVdiffVal < 2) {
if (Rs < 2826) {
return 1;
}
else if (Rs >= 2826) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
return 2;
}
}
}
else if (CsDiff >= 4) {
if (Cs < 1188) {
return 1;
}
else if (Cs >= 1188) {
if (ssDCval < 125) {
return 1;
}
else if (ssDCval >= 125) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (SC < 3403) {
if (refDCval < 139) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
return 1;
}
}
else if (refDCval >= 139) {
return 1;
}
}
else if (SC >= 3403) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (diffAFD < -2) {
if (diffMVdiffVal < 1) {
return 1;
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
else if (diffAFD >= -2) {
if (AFD < 40) {
return 2;
}
else if (AFD >= 40) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 3415) {
if (refDCval < 104) {
if (posBalance < 47) {
if (negBalance < 30) {
if (Cs < 1033) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 52) {
return 4;
}
else if (mu_mv_mag_sq >= 52) {
if (RsDiff < 6) {
if (mu_mv_mag_sq < 60) {
if (Rs < 3441) {
return 8;
}
else if (Rs >= 3441) {
return 8;
}
}
else if (mu_mv_mag_sq >= 60) {
if (Rs < 3439) {
return 4;
}
else if (Rs >= 3439) {
return 4;
}
}
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
if (SC < 4442) {
return 4;
}
else if (SC >= 4442) {
return 2;
}
}
}
}
}
else if (Cs >= 1033) {
if (RsCsDiff < 3) {
if (SC < 4958) {
return 8;
}
else if (SC >= 4958) {
if (SC < 5427) {
if (Cs < 1255) {
return 4;
}
else if (Cs >= 1255) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
}
}
else if (SC >= 5427) {
return 8;
}
}
}
else if (RsCsDiff >= 3) {
if (ssDCval < 69) {
if (ssDCval < 63) {
if (ssDCval < 59) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
else if (ssDCval >= 59) {
return 8;
}
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (ssDCval >= 69) {
if (SC < 5345) {
if (mu_mv_mag_sq < 82) {
if (negBalance < 29) {
if (Cs < 1067) {
return 8;
}
else if (Cs >= 1067) {
return 8;
}
}
else if (negBalance >= 29) {
if (Cs < 1117) {
if (Rs < 3785) {
return 4;
}
else if (Rs >= 3785) {
if (Cs < 1050) {
return 8;
}
else if (Cs >= 1050) {
return 4;
}
}
}
else if (Cs >= 1117) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (RsDiff < 6) {
if (AFD < 2) {
if (Rs < 3739) {
return 4;
}
else if (Rs >= 3739) {
return 8;
}
}
else if (AFD >= 2) {
return 4;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (SC >= 5345) {
if (Rs < 4548) {
return 8;
}
else if (Rs >= 4548) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 30) {
if (refDCval < 79) {
if (Rs < 3817) {
if (mu_mv_mag_sq < 87) {
if (ssDCval < 71) {
if (ssDCval < 67) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (ssDCval >= 67) {
return 4;
}
}
else if (ssDCval >= 71) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 2;
}
}
}
else if (Rs >= 3817) {
if (Rs < 3856) {
if (Rs < 3855) {
return 4;
}
else if (Rs >= 3855) {
return 2;
}
}
else if (Rs >= 3856) {
return 4;
}
}
}
else if (refDCval >= 79) {
if (Rs < 3842) {
if (mcTcor < 92) {
if (Rs < 3787) {
return 4;
}
else if (Rs >= 3787) {
return 8;
}
}
else if (mcTcor >= 92) {
if (ssDCval < 103) {
if (Rs < 3431) {
if (Rs < 3424) {
return 2;
}
else if (Rs >= 3424) {
return 1;
}
}
else if (Rs >= 3431) {
return 2;
}
}
else if (ssDCval >= 103) {
if (Rs < 3457) {
return 2;
}
else if (Rs >= 3457) {
return 8;
}
}
}
}
else if (Rs >= 3842) {
if (mu_mv_mag_sq < 41) {
return 2;
}
else if (mu_mv_mag_sq >= 41) {
if (mcTcor < 96) {
if (Rs < 4240) {
if (ssDCval < 92) {
if (negBalance < 33) {
if (gchDC < 1) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (negBalance >= 33) {
return 4;
}
}
else if (ssDCval >= 92) {
return 4;
}
}
else if (Rs >= 4240) {
return 4;
}
}
else if (mcTcor >= 96) {
if (Cs < 340) {
return 2;
}
else if (Cs >= 340) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 47) {
if (refDCval < 78) {
if (Cs < 895) {
if (posBalance < 68) {
if (Rs < 3572) {
if (TSC < 6) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 2;
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (Rs >= 3572) {
if (Cs < 800) {
return 4;
}
else if (Cs >= 800) {
if (Cs < 803) {
return 2;
}
else if (Cs >= 803) {
return 4;
}
}
}
}
else if (posBalance >= 68) {
return 2;
}
}
else if (Cs >= 895) {
if (refDCval < 64) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (Cs < 1224) {
return 4;
}
else if (Cs >= 1224) {
return 2;
}
}
}
else if (refDCval >= 64) {
if (Rs < 4780) {
return 2;
}
else if (Rs >= 4780) {
return 4;
}
}
}
}
else if (refDCval >= 78) {
if (mcTcor < 68) {
return 1;
}
else if (mcTcor >= 68) {
if (MVDiff < 19) {
if (refDCval < 100) {
if (negBalance < 35) {
if (posBalance < 50) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (posBalance >= 50) {
if (Rs < 4032) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
if (Rs < 3697) {
return 4;
}
else if (Rs >= 3697) {
return 2;
}
}
}
else if (Rs >= 4032) {
if (Rs < 4496) {
return 8;
}
else if (Rs >= 4496) {
return 2;
}
}
}
}
else if (negBalance >= 35) {
if (Cs < 379) {
if (Rs < 3790) {
return 2;
}
else if (Rs >= 3790) {
return 4;
}
}
else if (Cs >= 379) {
if (tcor < 71) {
if (Rs < 3522) {
return 2;
}
else if (Rs >= 3522) {
return 1;
}
}
else if (tcor >= 71) {
if (diffAFD < -9) {
return 1;
}
else if (diffAFD >= -9) {
return 2;
}
}
}
}
}
else if (refDCval >= 100) {
if (negBalance < 51) {
if (ssDCval < 106) {
return 2;
}
else if (ssDCval >= 106) {
return 8;
}
}
else if (negBalance >= 51) {
if (Rs < 3792) {
return 8;
}
else if (Rs >= 3792) {
return 2;
}
}
}
}
else if (MVDiff >= 19) {
if (RsCsDiff < 375) {
if (mcTcor < 93) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
if (Rs < 3663) {
return 8;
}
else if (Rs >= 3663) {
return 4;
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 83) {
if (Rs < 3792) {
return 2;
}
else if (Rs >= 3792) {
return 4;
}
}
else if (ssDCval >= 83) {
if (negBalance < 48) {
if (AFD < 25) {
if (diffTSC < -4) {
return 8;
}
else if (diffTSC >= -4) {
return 2;
}
}
else if (AFD >= 25) {
return 8;
}
}
else if (negBalance >= 48) {
if (Cs < 978) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (Cs >= 978) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 375) {
if (mcTcor < 94) {
if (mcTcor < 83) {
if (Cs < 1152) {
return 2;
}
else if (Cs >= 1152) {
if (Rs < 3630) {
return 2;
}
else if (Rs >= 3630) {
return 4;
}
}
}
else if (mcTcor >= 83) {
return 2;
}
}
else if (mcTcor >= 94) {
if (CsDiff < 6) {
if (ssDCval < 91) {
if (Rs < 4078) {
return 1;
}
else if (Rs >= 4078) {
if (Rs < 6541) {
return 4;
}
else if (Rs >= 6541) {
return 2;
}
}
}
else if (ssDCval >= 91) {
return 2;
}
}
else if (CsDiff >= 6) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 104) {
if (MVDiff < 7) {
if (CsDiff < 4) {
if (refDCval < 143) {
if (Cs < 558) {
if (Cs < 553) {
return 2;
}
else if (Cs >= 553) {
return 8;
}
}
else if (Cs >= 558) {
return 2;
}
}
else if (refDCval >= 143) {
if (Rs < 3963) {
if (mu_mv_mag_sq < 118) {
if (CsDiff < 3) {
if (Cs < 1234) {
return 2;
}
else if (Cs >= 1234) {
if (Rs < 3731) {
return 2;
}
else if (Rs >= 3731) {
return 1;
}
}
}
else if (CsDiff >= 3) {
if (Cs < 1108) {
return 2;
}
else if (Cs >= 1108) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 118) {
return 1;
}
}
else if (Rs >= 3963) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 1;
}
}
}
}
else if (CsDiff >= 4) {
if (ssDCval < 108) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (Rs < 4638) {
return 8;
}
else if (Rs >= 4638) {
return 2;
}
}
}
else if (ssDCval >= 108) {
if (refDCval < 136) {
if (Rs < 3541) {
return 8;
}
else if (Rs >= 3541) {
return 2;
}
}
else if (refDCval >= 136) {
if (Rs < 4384) {
if (Cs < 1036) {
return 2;
}
else if (Cs >= 1036) {
if (Rs < 3962) {
return 1;
}
else if (Rs >= 3962) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
else if (Rs >= 4384) {
return 2;
}
}
}
}
}
else if (MVDiff >= 7) {
if (refDCval < 135) {
if (Rs < 5192) {
if (RsDiff < 20) {
if (Cs < 1109) {
if (negBalance < 41) {
if (CsDiff < 4) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Cs < 1045) {
return 2;
}
else if (Cs >= 1045) {
if (refDCval < 129) {
return 2;
}
else if (refDCval >= 129) {
if (Rs < 3896) {
return 8;
}
else if (Rs >= 3896) {
return 1;
}
}
}
}
}
else if (CsDiff >= 4) {
if (Rs < 5148) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
return 8;
}
}
else if (Rs >= 5148) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (MVDiff < 29) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (negBalance < 54) {
if (Rs < 3427) {
return 1;
}
else if (Rs >= 3427) {
if (RsDiff < 10) {
if (Rs < 3868) {
return 2;
}
else if (Rs >= 3868) {
return 8;
}
}
else if (RsDiff >= 10) {
return 2;
}
}
}
else if (negBalance >= 54) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (gchDC < 1) {
if (Rs < 3986) {
return 2;
}
else if (Rs >= 3986) {
return 1;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
}
else if (MVDiff >= 29) {
if (CsDiff < 4) {
if (Cs < 497) {
if (mu_mv_mag_sq < 2) {
return 4;
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 3653) {
return 8;
}
else if (Rs >= 3653) {
return 2;
}
}
}
else if (Cs >= 497) {
if (ssDCval < 134) {
if (diffRsCsdiff < -198) {
return 2;
}
else if (diffRsCsdiff >= -198) {
return 8;
}
}
else if (ssDCval >= 134) {
return 2;
}
}
}
else if (CsDiff >= 4) {
if (diffRsCsdiff < 3) {
if (RsCsDiff < 313) {
if (Rs < 3650) {
return 8;
}
else if (Rs >= 3650) {
return 1;
}
}
else if (RsCsDiff >= 313) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Cs < 1071) {
if (Rs < 3500) {
return 1;
}
else if (Rs >= 3500) {
return 8;
}
}
else if (Cs >= 1071) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 3) {
if (Rs < 3630) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
else if (Rs >= 3630) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1109) {
if (ssDCval < 109) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (mu_mv_mag_sq < 11) {
if (Cs < 1253) {
return 2;
}
else if (Cs >= 1253) {
return 8;
}
}
else if (mu_mv_mag_sq >= 11) {
return 4;
}
}
}
else if (ssDCval >= 109) {
if (diffTSC < -2) {
if (Rs < 3447) {
return 8;
}
else if (Rs >= 3447) {
return 2;
}
}
else if (diffTSC >= -2) {
if (Rs < 4736) {
if (posBalance < 37) {
if (Rs < 4242) {
return 1;
}
else if (Rs >= 4242) {
return 2;
}
}
else if (posBalance >= 37) {
if (refDCval < 132) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (refDCval >= 132) {
if (Rs < 3890) {
if (Rs < 3589) {
return 2;
}
else if (Rs >= 3589) {
return 8;
}
}
else if (Rs >= 3890) {
return 1;
}
}
}
}
else if (Rs >= 4736) {
if (RsDiff < 18) {
return 1;
}
else if (RsDiff >= 18) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 20) {
if (refDCval < 119) {
if (tcor < 82) {
if (Rs < 3449) {
if (Rs < 3440) {
return 2;
}
else if (Rs >= 3440) {
return 8;
}
}
else if (Rs >= 3449) {
return 2;
}
}
else if (tcor >= 82) {
if (MVDiff < 47) {
if (posBalance < 64) {
return 2;
}
else if (posBalance >= 64) {
if (Cs < 823) {
return 1;
}
else if (Cs >= 823) {
return 2;
}
}
}
else if (MVDiff >= 47) {
return 1;
}
}
}
else if (refDCval >= 119) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (TSC < 11) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
if (diffMVdiffVal < 8) {
if (diffMVdiffVal < -21) {
if (Cs < 995) {
return 8;
}
else if (Cs >= 995) {
return 2;
}
}
else if (diffMVdiffVal >= -21) {
return 8;
}
}
else if (diffMVdiffVal >= 8) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 2;
}
}
}
}
else if (TSC >= 11) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (CsDiff < 6) {
if (SC < 4660) {
return 1;
}
else if (SC >= 4660) {
return 2;
}
}
else if (CsDiff >= 6) {
if (SC < 4393) {
return 2;
}
else if (SC >= 4393) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 5192) {
if (AFD < 26) {
if (ssDCval < 135) {
if (Cs < 1265) {
return 2;
}
else if (Cs >= 1265) {
if (Rs < 6189) {
return 2;
}
else if (Rs >= 6189) {
return 8;
}
}
}
else if (ssDCval >= 135) {
return 1;
}
}
else if (AFD >= 26) {
if (Cs < 984) {
return 2;
}
else if (Cs >= 984) {
if (TSC < 14) {
if (MVDiff < 44) {
if (diffMVdiffVal < -7) {
return 8;
}
else if (diffMVdiffVal >= -7) {
return 2;
}
}
else if (MVDiff >= 44) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 5943) {
return 8;
}
else if (Rs >= 5943) {
return 2;
}
}
}
}
else if (TSC >= 14) {
return 1;
}
}
}
}
}
else if (refDCval >= 135) {
if (posBalance < 59) {
if (CsDiff < 3) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
return 4;
}
}
else if (CsDiff >= 3) {
if (RsDiff < 22) {
if (mu_mv_mag_sq < 55) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (AFD < 21) {
if (MVDiff < 12) {
if (Rs < 6042) {
return 2;
}
else if (Rs >= 6042) {
return 1;
}
}
else if (MVDiff >= 12) {
if (RsDiff < 20) {
return 1;
}
else if (RsDiff >= 20) {
if (Rs < 4854) {
return 1;
}
else if (Rs >= 4854) {
return 2;
}
}
}
}
else if (AFD >= 21) {
if (Rs < 4627) {
if (TSC < 12) {
if (Cs < 939) {
return 2;
}
else if (Cs >= 939) {
if (Rs < 4051) {
return 8;
}
else if (Rs >= 4051) {
return 1;
}
}
}
else if (TSC >= 12) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
}
else if (Rs >= 4627) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (AFD < 11) {
if (Cs < 1077) {
return 1;
}
else if (Cs >= 1077) {
return 2;
}
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (RsDiff >= 22) {
if (Cs < 979) {
if (diffMVdiffVal < -42) {
return 1;
}
else if (diffMVdiffVal >= -42) {
return 2;
}
}
else if (Cs >= 979) {
if (CsDiff < 7) {
if (Cs < 1055) {
return 8;
}
else if (Cs >= 1055) {
return 2;
}
}
else if (CsDiff >= 7) {
if (Rs < 6308) {
return 1;
}
else if (Rs >= 6308) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 59) {
if (negBalance < 50) {
if (ssDCval < 140) {
if (refDCval < 136) {
if (Rs < 4155) {
return 8;
}
else if (Rs >= 4155) {
return 1;
}
}
else if (refDCval >= 136) {
return 2;
}
}
else if (ssDCval >= 140) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
if (Rs < 5632) {
return 1;
}
else if (Rs >= 5632) {
return 2;
}
}
}
}
else if (negBalance >= 50) {
if (CsDiff < 6) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
return 8;
}
}
else if (CsDiff >= 6) {
if (MVDiff < 26) {
return 2;
}
else if (MVDiff >= 26) {
if (Rs < 4491) {
return 1;
}
else if (Rs >= 4491) {
if (Rs < 5422) {
return 2;
}
else if (Rs >= 5422) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1325) {
if (mu_mv_mag_sq < 111) {
if (RsCsDiff < 101) {
if (ssDCval < 105) {
if (negBalance < 33) {
if (SC < 10037) {
if (Rs < 846) {
if (Rs < 300) {
return 4;
}
else if (Rs >= 300) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
if (Rs < 655) {
return 8;
}
else if (Rs >= 655) {
return 4;
}
}
}
}
}
else if (Rs >= 846) {
if (posBalance < 45) {
if (ssDCval < 73) {
if (ssDCval < 64) {
if (Rs < 3382) {
if (ssDCval < 58) {
if (AFD < 8) {
if (ssDCval < 53) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (ssDCval >= 53) {
if (Rs < 1031) {
return 8;
}
else if (Rs >= 1031) {
if (Rs < 1442) {
return 4;
}
else if (Rs >= 1442) {
return 1;
}
}
}
}
else if (AFD >= 8) {
if (Rs < 1417) {
return 4;
}
else if (Rs >= 1417) {
return 2;
}
}
}
else if (ssDCval >= 58) {
if (tcor < 96) {
if (Rs < 1358) {
return 4;
}
else if (Rs >= 1358) {
return 2;
}
}
else if (tcor >= 96) {
return 8;
}
}
}
else if (Rs >= 3382) {
if (MVDiff < 4) {
if (CsDiff < 5) {
if (Rs < 4768) {
return 8;
}
else if (Rs >= 4768) {
return 8;
}
}
else if (CsDiff >= 5) {
if (Rs < 4627) {
return 4;
}
else if (Rs >= 4627) {
return 4;
}
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (ssDCval >= 64) {
if (MVDiff < 5) {
if (Rs < 3032) {
return 8;
}
else if (Rs >= 3032) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
}
else if (ssDCval >= 73) {
if (MVDiff < 7) {
if (Rs < 3284) {
if (RsCsDiff < 1) {
if (Rs < 2402) {
if (TSCindex < 1) {
if (Rs < 2327) {
return 2;
}
else if (Rs >= 2327) {
return 1;
}
}
else if (TSCindex >= 1) {
if (Rs < 2348) {
return 1;
}
else if (Rs >= 2348) {
return 4;
}
}
}
else if (Rs >= 2402) {
if (ssDCval < 87) {
return 1;
}
else if (ssDCval >= 87) {
return 8;
}
}
}
else if (RsCsDiff >= 1) {
if (mu_mv_mag_sq < 64) {
return 1;
}
else if (mu_mv_mag_sq >= 64) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (diffAFD < -1) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
else if (Rs >= 3284) {
if (SC < 5596) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (SC >= 5596) {
if (SC < 6944) {
if (TSC < 4) {
if (Rs < 4797) {
if (Rs < 4292) {
return 8;
}
else if (Rs >= 4292) {
return 2;
}
}
else if (Rs >= 4797) {
return 8;
}
}
else if (TSC >= 4) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
if (posBalance < 38) {
if (Rs < 4486) {
return 4;
}
else if (Rs >= 4486) {
return 8;
}
}
else if (posBalance >= 38) {
return 8;
}
}
}
}
else if (SC >= 6944) {
if (SC < 6999) {
return 1;
}
else if (SC >= 6999) {
if (diffAFD < -1) {
if (Rs < 5365) {
return 4;
}
else if (Rs >= 5365) {
return 8;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (TSC < 2) {
if (ssDCval < 87) {
return 1;
}
else if (ssDCval >= 87) {
return 4;
}
}
else if (TSC >= 2) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2746) {
if (mu_mv_mag_sq < 29) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
if (Rs < 1381) {
return 4;
}
else if (Rs >= 1381) {
if (Rs < 2068) {
return 8;
}
else if (Rs >= 2068) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (Cs < 3872) {
if (SC < 2516) {
if (posBalance < 54) {
if (Rs < 941) {
return 1;
}
else if (Rs >= 941) {
if (Rs < 1059) {
return 4;
}
else if (Rs >= 1059) {
return 8;
}
}
}
else if (posBalance >= 54) {
return 2;
}
}
else if (SC >= 2516) {
if (refDCval < 58) {
if (Cs < 1693) {
if (mu_mv_mag_sq < 97) {
if (Rs < 1224) {
return 4;
}
else if (Rs >= 1224) {
return 8;
}
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 1387) {
return 8;
}
else if (Cs >= 1387) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
else if (Cs >= 1693) {
if (Rs < 1202) {
return 8;
}
else if (Rs >= 1202) {
if (Cs < 2043) {
return 4;
}
else if (Cs >= 2043) {
return 8;
}
}
}
}
else if (refDCval >= 58) {
if (TSC < 4) {
if (mu_mv_mag_sq < 78) {
return 1;
}
else if (mu_mv_mag_sq >= 78) {
return 8;
}
}
else if (TSC >= 4) {
if (Cs < 2442) {
if (Rs < 2631) {
if (mu_mv_mag_sq < 91) {
if (Cs < 1618) {
if (Rs < 2218) {
if (Cs < 1610) {
return 8;
}
else if (Cs >= 1610) {
return 4;
}
}
else if (Rs >= 2218) {
return 4;
}
}
else if (Cs >= 1618) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (Rs < 2167) {
return 1;
}
else if (Rs >= 2167) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
return 8;
}
}
else if (Rs >= 2631) {
if (Rs < 2672) {
return 4;
}
else if (Rs >= 2672) {
return 1;
}
}
}
else if (Cs >= 2442) {
if (Cs < 2565) {
return 4;
}
else if (Cs >= 2565) {
return 8;
}
}
}
}
}
}
else if (Cs >= 3872) {
return 1;
}
}
}
else if (Rs >= 2746) {
if (RsCsDiff < 37) {
if (Cs < 2100) {
if (Cs < 1472) {
if (Cs < 1336) {
return 8;
}
else if (Cs >= 1336) {
return 2;
}
}
else if (Cs >= 1472) {
if (ssDCval < 87) {
if (Cs < 1685) {
if (Rs < 3399) {
return 4;
}
else if (Rs >= 3399) {
if (Rs < 3579) {
return 4;
}
else if (Rs >= 3579) {
return 4;
}
}
}
else if (Cs >= 1685) {
return 4;
}
}
else if (ssDCval >= 87) {
if (Rs < 3335) {
return 8;
}
else if (Rs >= 3335) {
return 2;
}
}
}
}
else if (Cs >= 2100) {
if (Rs < 2898) {
return 4;
}
else if (Rs >= 2898) {
return 8;
}
}
}
else if (RsCsDiff >= 37) {
if (Cs < 1805) {
if (AFD < 9) {
if (Rs < 3362) {
if (refDCval < 83) {
return 2;
}
else if (refDCval >= 83) {
if (Cs < 1452) {
return 2;
}
else if (Cs >= 1452) {
return 4;
}
}
}
else if (Rs >= 3362) {
return 2;
}
}
else if (AFD >= 9) {
if (posBalance < 69) {
return 1;
}
else if (posBalance >= 69) {
return 2;
}
}
}
else if (Cs >= 1805) {
if (tcor < 95) {
if (CsDiff < 6) {
if (MVDiff < 18) {
if (mu_mv_mag_sq < 108) {
return 4;
}
else if (mu_mv_mag_sq >= 108) {
return 4;
}
}
else if (MVDiff >= 18) {
return 8;
}
}
else if (CsDiff >= 6) {
if (Cs < 2435) {
return 8;
}
else if (Cs >= 2435) {
return 1;
}
}
}
else if (tcor >= 95) {
if (Rs < 3390) {
if (Rs < 2937) {
return 1;
}
else if (Rs >= 2937) {
return 4;
}
}
else if (Rs >= 3390) {
return 2;
}
}
}
}
}
}
}
}
else if (SC >= 10037) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 2;
}
}
}
else if (negBalance >= 33) {
if (Rs < 1448) {
if (AFD < 10) {
if (MVDiff < 5) {
if (Cs < 1553) {
if (refDCval < 100) {
if (ssDCval < 67) {
if (RsCsDiff < 19) {
return 8;
}
else if (RsCsDiff >= 19) {
if (ssDCval < 52) {
if (Rs < 1196) {
return 4;
}
else if (Rs >= 1196) {
if (Cs < 1477) {
return 8;
}
else if (Cs >= 1477) {
return 4;
}
}
}
else if (ssDCval >= 52) {
if (TSC < 7) {
if (tcor < 84) {
if (Rs < 1036) {
return 4;
}
else if (Rs >= 1036) {
return 8;
}
}
else if (tcor >= 84) {
return 4;
}
}
else if (TSC >= 7) {
if (Rs < 1357) {
return 1;
}
else if (Rs >= 1357) {
return 8;
}
}
}
}
}
else if (ssDCval >= 67) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (Rs < 680) {
return 1;
}
else if (Rs >= 680) {
return 4;
}
}
}
}
else if (refDCval >= 100) {
return 4;
}
}
else if (Cs >= 1553) {
if (ssDCval < 67) {
if (Cs < 1622) {
return 2;
}
else if (Cs >= 1622) {
return 8;
}
}
else if (ssDCval >= 67) {
if (RsCsDiff < 23) {
return 1;
}
else if (RsCsDiff >= 23) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (Rs < 1122) {
return 1;
}
else if (Rs >= 1122) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 81) {
if (ssDCval < 59) {
if (SC < 2426) {
return 2;
}
else if (SC >= 2426) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (CsDiff < 6) {
if (tcor < 97) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (tcor >= 97) {
return 8;
}
}
else if (CsDiff >= 6) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
else if (ssDCval >= 59) {
if (Rs < 1182) {
if (TSC < 5) {
if (Cs < 2247) {
if (Cs < 1405) {
if (Rs < 764) {
return 1;
}
else if (Rs >= 764) {
return 2;
}
}
else if (Cs >= 1405) {
return 1;
}
}
else if (Cs >= 2247) {
return 8;
}
}
else if (TSC >= 5) {
if (refDCval < 65) {
return 8;
}
else if (refDCval >= 65) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
}
else if (Rs >= 1182) {
if (refDCval < 69) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
return 4;
}
}
else if (refDCval >= 69) {
return 4;
}
}
}
}
else if (ssDCval >= 81) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
return 4;
}
}
}
}
else if (AFD >= 10) {
if (TSC < 11) {
if (TSC < 7) {
if (refDCval < 80) {
if (ssDCval < 58) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (ssDCval >= 58) {
if (posBalance < 54) {
if (Rs < 711) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
return 8;
}
}
else if (Rs >= 711) {
if (Rs < 1056) {
return 2;
}
else if (Rs >= 1056) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (AFD < 13) {
if (Rs < 980) {
return 8;
}
else if (Rs >= 980) {
return 1;
}
}
else if (AFD >= 13) {
if (diffRsCsdiff < -5) {
return 4;
}
else if (diffRsCsdiff >= -5) {
if (Rs < 1124) {
return 8;
}
else if (Rs >= 1124) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 80) {
if (Rs < 1318) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 990) {
return 8;
}
else if (Rs >= 990) {
return 2;
}
}
}
else if (Rs >= 1318) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (negBalance < 56) {
if (MVDiff < 13) {
if (SC < 2731) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (AFD < 14) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (Rs < 752) {
return 2;
}
else if (Rs >= 752) {
if (Cs < 1489) {
if (SC < 2219) {
return 4;
}
else if (SC >= 2219) {
return 8;
}
}
else if (Cs >= 1489) {
return 4;
}
}
}
}
else if (AFD >= 14) {
if (AFD < 16) {
if (Cs < 1333) {
return 2;
}
else if (Cs >= 1333) {
return 8;
}
}
else if (AFD >= 16) {
if (Rs < 811) {
return 8;
}
else if (Rs >= 811) {
return 4;
}
}
}
}
}
else if (SC >= 2731) {
if (Cs < 2116) {
if (negBalance < 53) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (negBalance >= 53) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
}
}
else if (Cs >= 2116) {
if (Rs < 1399) {
return 2;
}
else if (Rs >= 1399) {
return 8;
}
}
}
}
else if (MVDiff >= 13) {
if (MVDiff < 69) {
if (tcor < 88) {
if (diffTSC < 2) {
if (CsDiff < 8) {
if (Rs < 1162) {
if (ssDCval < 76) {
if (AFD < 13) {
if (Rs < 1060) {
return 4;
}
else if (Rs >= 1060) {
return 2;
}
}
else if (AFD >= 13) {
return 2;
}
}
else if (ssDCval >= 76) {
return 4;
}
}
else if (Rs >= 1162) {
return 4;
}
}
else if (CsDiff >= 8) {
return 4;
}
}
else if (diffTSC >= 2) {
if (Rs < 1104) {
return 8;
}
else if (Rs >= 1104) {
return 4;
}
}
}
else if (tcor >= 88) {
return 8;
}
}
else if (MVDiff >= 69) {
if (Rs < 734) {
return 2;
}
else if (Rs >= 734) {
return 8;
}
}
}
}
else if (negBalance >= 56) {
if (RsCsDiff < 70) {
if (Rs < 1025) {
return 8;
}
else if (Rs >= 1025) {
if (SC < 2867) {
return 4;
}
else if (SC >= 2867) {
return 8;
}
}
}
else if (RsCsDiff >= 70) {
if (diffAFD < 2) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
}
}
else if (TSC >= 11) {
if (RsCsDiff < 80) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Rs < 1320) {
return 8;
}
else if (Rs >= 1320) {
return 4;
}
}
}
else if (RsCsDiff >= 80) {
if (refDCval < 70) {
if (mcTcor < 77) {
return 4;
}
else if (mcTcor >= 77) {
return 2;
}
}
else if (refDCval >= 70) {
if (Rs < 1346) {
return 4;
}
else if (Rs >= 1346) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1448) {
if (CsDiff < 7) {
if (mcTcor < 94) {
if (ssDCval < 90) {
if (AFD < 16) {
if (Cs < 1771) {
if (posBalance < 70) {
if (mcTcor < 91) {
if (gchDC < 9) {
if (mu_mv_mag_sq < 91) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (Rs < 2224) {
if (SC < 2979) {
if (SC < 2957) {
return 8;
}
else if (SC >= 2957) {
return 4;
}
}
else if (SC >= 2979) {
return 8;
}
}
else if (Rs >= 2224) {
return 4;
}
}
}
else if (gchDC >= 9) {
return 4;
}
}
else if (mcTcor >= 91) {
if (negBalance < 36) {
if (AFD < 8) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (AFD >= 8) {
return 2;
}
}
else if (negBalance >= 36) {
if (tcor < 85) {
if (SC < 3164) {
return 4;
}
else if (SC >= 3164) {
if (Rs < 2830) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Cs < 1676) {
if (Rs < 1904) {
return 1;
}
else if (Rs >= 1904) {
return 4;
}
}
else if (Cs >= 1676) {
return 8;
}
}
}
else if (Rs >= 2830) {
return 4;
}
}
}
else if (tcor >= 85) {
if (SC < 3591) {
if (Rs < 2114) {
if (diffAFD < 4) {
if (diffAFD < -1) {
if (Rs < 1993) {
if (diffMVdiffVal < -5) {
if (Rs < 1642) {
return 8;
}
else if (Rs >= 1642) {
return 4;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (Rs >= 1993) {
if (Rs < 2026) {
return 1;
}
else if (Rs >= 2026) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (Cs < 1342) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (Cs >= 1342) {
return 8;
}
}
}
else if (diffAFD >= 4) {
return 2;
}
}
else if (Rs >= 2114) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (SC >= 3591) {
if (diffMVdiffVal < -6) {
if (Rs < 2331) {
return 4;
}
else if (Rs >= 2331) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
if (Cs < 1369) {
if (Rs < 2409) {
return 8;
}
else if (Rs >= 2409) {
return 4;
}
}
else if (Cs >= 1369) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 70) {
if (Rs < 2088) {
return 4;
}
else if (Rs >= 2088) {
return 8;
}
}
}
else if (Cs >= 1771) {
if (Rs < 2668) {
if (Cs < 2490) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (posBalance < 56) {
if (negBalance < 52) {
if (SC < 4366) {
if (Cs < 1874) {
if (Rs < 2430) {
return 8;
}
else if (Rs >= 2430) {
return 4;
}
}
else if (Cs >= 1874) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (SC >= 4366) {
return 8;
}
}
else if (negBalance >= 52) {
return 4;
}
}
else if (posBalance >= 56) {
if (diffAFD < 2) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (negBalance < 46) {
if (Rs < 2081) {
return 4;
}
else if (Rs >= 2081) {
return 8;
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 2002) {
return 8;
}
else if (Rs >= 2002) {
return 4;
}
}
}
}
}
else if (Cs >= 2490) {
return 4;
}
}
else if (Rs >= 2668) {
if (RsDiff < 7) {
if (diffRsCsdiff < -53) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (diffRsCsdiff >= -53) {
return 8;
}
}
else if (RsDiff >= 7) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (TSC < 8) {
if (mu_mv_mag_sq < 88) {
if (Cs < 1907) {
return 1;
}
else if (Cs >= 1907) {
return 8;
}
}
else if (mu_mv_mag_sq >= 88) {
if (diffRsCsdiff < 6) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
if (Rs < 3316) {
return 8;
}
else if (Rs >= 3316) {
return 4;
}
}
}
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
}
else if (AFD >= 16) {
if (ssDCval < 78) {
if (Rs < 4887) {
return 8;
}
else if (Rs >= 4887) {
return 4;
}
}
else if (ssDCval >= 78) {
if (Rs < 1722) {
if (refDCval < 90) {
return 4;
}
else if (refDCval >= 90) {
return 8;
}
}
else if (Rs >= 1722) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
return 8;
}
}
}
}
}
else if (ssDCval >= 90) {
if (Rs < 2227) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (AFD < 12) {
if (RsCsDiff < 57) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
return 2;
}
}
else if (RsCsDiff >= 57) {
if (Cs < 1743) {
return 2;
}
else if (Cs >= 1743) {
return 8;
}
}
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (Rs >= 2227) {
if (Cs < 2766) {
if (diffRsCsdiff < -124) {
if (Rs < 3332) {
return 4;
}
else if (Rs >= 3332) {
return 8;
}
}
else if (diffRsCsdiff >= -124) {
return 8;
}
}
else if (Cs >= 2766) {
return 4;
}
}
}
}
else if (mcTcor >= 94) {
if (RsDiff < 7) {
if (SC < 7197) {
if (SC < 4691) {
if (ssDCval < 96) {
if (Rs < 2801) {
if (refDCval < 84) {
if (AFD < 8) {
if (Rs < 1480) {
if (Cs < 1635) {
return 4;
}
else if (Cs >= 1635) {
return 8;
}
}
else if (Rs >= 1480) {
if (SC < 3813) {
if (Rs < 2139) {
if (diffRsCsdiff < 6) {
return 8;
}
else if (diffRsCsdiff >= 6) {
if (Cs < 1767) {
if (Cs < 1351) {
return 1;
}
else if (Cs >= 1351) {
return 8;
}
}
else if (Cs >= 1767) {
return 1;
}
}
}
else if (Rs >= 2139) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (SC >= 3813) {
if (Rs < 2681) {
return 8;
}
else if (Rs >= 2681) {
if (RsCsDiff < 23) {
return 8;
}
else if (RsCsDiff >= 23) {
return 4;
}
}
}
}
}
else if (AFD >= 8) {
if (refDCval < 80) {
if (AFD < 9) {
if (TSC < 6) {
if (Rs < 1863) {
return 4;
}
else if (Rs >= 1863) {
return 1;
}
}
else if (TSC >= 6) {
if (Cs < 1386) {
if (Rs < 1826) {
return 8;
}
else if (Rs >= 1826) {
return 1;
}
}
else if (Cs >= 1386) {
return 8;
}
}
}
else if (AFD >= 9) {
if (gchDC < 2) {
if (Cs < 1862) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 75) {
return 8;
}
else if (mu_mv_mag_sq >= 75) {
return 4;
}
}
else if (posBalance >= 57) {
return 4;
}
}
else if (Cs >= 1862) {
return 8;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (refDCval >= 80) {
if (Rs < 2105) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (diffTSC < 1) {
if (Rs < 1778) {
return 1;
}
else if (Rs >= 1778) {
return 8;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (Rs >= 2105) {
if (SC < 3851) {
if (Cs < 1369) {
return 8;
}
else if (Cs >= 1369) {
return 1;
}
}
else if (SC >= 3851) {
if (MVDiff < 2) {
if (Rs < 2403) {
return 8;
}
else if (Rs >= 2403) {
return 1;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 84) {
if (SC < 3689) {
if (TSC < 7) {
if (mu_mv_mag_sq < 102) {
if (posBalance < 57) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (posBalance >= 57) {
if (mu_mv_mag_sq < 75) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 75) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 102) {
return 2;
}
}
else if (TSC >= 7) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Rs < 2245) {
if (Cs < 1709) {
if (Cs < 1357) {
if (Cs < 1350) {
return 8;
}
else if (Cs >= 1350) {
return 1;
}
}
else if (Cs >= 1357) {
return 8;
}
}
else if (Cs >= 1709) {
return 2;
}
}
else if (Rs >= 2245) {
if (Rs < 2255) {
return 1;
}
else if (Rs >= 2255) {
return 2;
}
}
}
}
}
else if (SC >= 3689) {
if (posBalance < 46) {
if (negBalance < 43) {
if (diffRsCsdiff < -1) {
return 1;
}
else if (diffRsCsdiff >= -1) {
if (diffAFD < 1) {
if (posBalance < 9) {
return 1;
}
else if (posBalance >= 9) {
return 8;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (negBalance >= 43) {
if (AFD < 4) {
if (Rs < 2216) {
return 2;
}
else if (Rs >= 2216) {
return 1;
}
}
else if (AFD >= 4) {
if (Rs < 2512) {
return 4;
}
else if (Rs >= 2512) {
return 8;
}
}
}
}
else if (posBalance >= 46) {
if (negBalance < 51) {
if (diffRsCsdiff < -10) {
if (TSC < 8) {
if (posBalance < 56) {
if (Rs < 2727) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (Rs >= 2727) {
return 1;
}
}
else if (posBalance >= 56) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
if (Rs < 2231) {
return 2;
}
else if (Rs >= 2231) {
return 8;
}
}
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (diffRsCsdiff >= -10) {
if (posBalance < 58) {
if (diffRsCsdiff < 19) {
return 8;
}
else if (diffRsCsdiff >= 19) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (Cs < 1505) {
return 1;
}
else if (Cs >= 1505) {
return 8;
}
}
}
}
else if (negBalance >= 51) {
if (TSC < 9) {
if (Rs < 2475) {
return 1;
}
else if (Rs >= 2475) {
return 4;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2801) {
if (Cs < 1445) {
return 2;
}
else if (Cs >= 1445) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Rs < 2811) {
return 2;
}
else if (Rs >= 2811) {
return 8;
}
}
}
}
}
else if (ssDCval >= 96) {
if (Rs < 1977) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (RsCsDiff < 29) {
return 8;
}
else if (RsCsDiff >= 29) {
return 2;
}
}
}
}
else if (Rs >= 1977) {
if (posBalance < 47) {
if (RsDiff < 5) {
if (CsDiff < 1) {
if (Rs < 2181) {
return 1;
}
else if (Rs >= 2181) {
return 4;
}
}
else if (CsDiff >= 1) {
if (diffMVdiffVal < -6) {
return 2;
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (posBalance >= 47) {
if (diffAFD < -12) {
return 4;
}
else if (diffAFD >= -12) {
return 8;
}
}
}
}
}
else if (SC >= 4691) {
if (MVDiff < 10) {
if (Cs < 1854) {
if (ssDCval < 91) {
if (AFD < 7) {
if (mu_mv_mag_sq < 99) {
if (Rs < 4409) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
return 2;
}
}
else if (Rs >= 4409) {
return 8;
}
}
else if (mu_mv_mag_sq >= 99) {
if (Rs < 3328) {
if (Rs < 3204) {
return 8;
}
else if (Rs >= 3204) {
return 4;
}
}
else if (Rs >= 3328) {
return 2;
}
}
}
else if (AFD >= 7) {
if (CsDiff < 6) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
return 1;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (ssDCval >= 91) {
if (ssDCval < 96) {
if (Rs < 4137) {
return 2;
}
else if (Rs >= 4137) {
return 8;
}
}
else if (ssDCval >= 96) {
return 8;
}
}
}
else if (Cs >= 1854) {
if (Rs < 2436) {
if (Cs < 2565) {
return 8;
}
else if (Cs >= 2565) {
return 4;
}
}
else if (Rs >= 2436) {
if (refDCval < 93) {
if (ssDCval < 69) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (ssDCval >= 69) {
if (RsCsDiff < 51) {
if (negBalance < 38) {
if (tcor < 98) {
if (Cs < 2546) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (Cs >= 2546) {
return 4;
}
}
else if (tcor >= 98) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (Rs < 3784) {
return 1;
}
else if (Rs >= 3784) {
return 8;
}
}
}
}
else if (negBalance >= 38) {
if (Rs < 2482) {
if (Rs < 2465) {
return 8;
}
else if (Rs >= 2465) {
return 4;
}
}
else if (Rs >= 2482) {
return 8;
}
}
}
else if (RsCsDiff >= 51) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (SC < 5289) {
if (refDCval < 83) {
return 1;
}
else if (refDCval >= 83) {
return 8;
}
}
else if (SC >= 5289) {
if (mcTcor < 97) {
if (negBalance < 44) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
return 4;
}
}
else if (negBalance >= 44) {
if (Rs < 2798) {
return 4;
}
else if (Rs >= 2798) {
return 8;
}
}
}
else if (mcTcor >= 97) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 93) {
if (Rs < 3834) {
return 4;
}
else if (Rs >= 3834) {
return 8;
}
}
}
}
}
else if (MVDiff >= 10) {
if (Rs < 4815) {
return 8;
}
else if (Rs >= 4815) {
return 4;
}
}
}
}
else if (SC >= 7197) {
return 8;
}
}
else if (RsDiff >= 7) {
if (AFD < 9) {
if (ssDCval < 91) {
if (Cs < 1840) {
if (Rs < 3059) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (SC < 4457) {
return 2;
}
else if (SC >= 4457) {
return 4;
}
}
}
else if (Rs >= 3059) {
if (mu_mv_mag_sq < 59) {
if (RsCsDiff < 80) {
if (Rs < 4072) {
return 4;
}
else if (Rs >= 4072) {
return 1;
}
}
else if (RsCsDiff >= 80) {
return 2;
}
}
else if (mu_mv_mag_sq >= 59) {
return 2;
}
}
}
else if (Cs >= 1840) {
if (negBalance < 39) {
if (Rs < 3381) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (Rs >= 3381) {
return 2;
}
}
else if (negBalance >= 39) {
if (negBalance < 43) {
if (Rs < 3508) {
return 8;
}
else if (Rs >= 3508) {
return 4;
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
else if (ssDCval >= 91) {
if (diffAFD < -1) {
if (Cs < 2050) {
return 2;
}
else if (Cs >= 2050) {
if (Cs < 2693) {
return 8;
}
else if (Cs >= 2693) {
return 4;
}
}
}
else if (diffAFD >= -1) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 2;
}
}
}
}
else if (AFD >= 9) {
if (negBalance < 43) {
if (CsDiff < 6) {
if (Rs < 3323) {
if (Cs < 1563) {
if (Rs < 2177) {
return 8;
}
else if (Rs >= 2177) {
return 2;
}
}
else if (Cs >= 1563) {
return 8;
}
}
else if (Rs >= 3323) {
if (ssDCval < 88) {
return 4;
}
else if (ssDCval >= 88) {
if (Cs < 1353) {
return 4;
}
else if (Cs >= 1353) {
if (Cs < 1864) {
return 2;
}
else if (Cs >= 1864) {
return 8;
}
}
}
}
}
else if (CsDiff >= 6) {
if (AFD < 13) {
if (ssDCval < 91) {
return 1;
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (AFD >= 13) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
}
else if (negBalance >= 43) {
if (tcor < 86) {
if (ssDCval < 85) {
if (Cs < 3043) {
if (Rs < 1669) {
return 8;
}
else if (Rs >= 1669) {
return 4;
}
}
else if (Cs >= 3043) {
return 8;
}
}
else if (ssDCval >= 85) {
if (Rs < 2956) {
return 2;
}
else if (Rs >= 2956) {
return 8;
}
}
}
else if (tcor >= 86) {
if (MVDiff < 1) {
if (Rs < 3428) {
return 8;
}
else if (Rs >= 3428) {
return 4;
}
}
else if (MVDiff >= 1) {
if (Rs < 1978) {
if (Rs < 1904) {
if (ssDCval < 86) {
if (Rs < 1813) {
return 8;
}
else if (Rs >= 1813) {
return 4;
}
}
else if (ssDCval >= 86) {
return 2;
}
}
else if (Rs >= 1904) {
return 1;
}
}
else if (Rs >= 1978) {
if (mcTcor < 95) {
if (diffRsCsdiff < -12) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (diffRsCsdiff < -21) {
return 8;
}
else if (diffRsCsdiff >= -21) {
return 4;
}
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
else if (mcTcor >= 95) {
if (AFD < 10) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
return 8;
}
}
}
else if (AFD >= 10) {
if (RsCsDiff < 77) {
if (diffRsCsdiff < 18) {
if (diffRsCsdiff < -39) {
return 1;
}
else if (diffRsCsdiff >= -39) {
return 8;
}
}
else if (diffRsCsdiff >= 18) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (RsCsDiff >= 77) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (CsDiff >= 7) {
if (Rs < 3044) {
if (Cs < 2708) {
if (RsDiff < 6) {
if (Rs < 1744) {
if (Rs < 1575) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (Rs >= 1575) {
if (TSC < 9) {
if (Rs < 1716) {
return 2;
}
else if (Rs >= 1716) {
return 8;
}
}
else if (TSC >= 9) {
return 4;
}
}
}
else if (Rs >= 1744) {
if (mu_mv_mag_sq < 83) {
if (Rs < 1908) {
return 8;
}
else if (Rs >= 1908) {
return 1;
}
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 1819) {
return 1;
}
else if (Rs >= 1819) {
return 8;
}
}
}
}
else if (RsDiff >= 6) {
if (AFD < 17) {
if (ssDCval < 53) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 2193) {
if (diffRsCsdiff < 7) {
return 4;
}
else if (diffRsCsdiff >= 7) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
else if (Rs >= 2193) {
return 8;
}
}
}
}
else if (ssDCval >= 53) {
if (diffAFD < -8) {
return 2;
}
else if (diffAFD >= -8) {
if (Cs < 2221) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Cs < 1860) {
return 8;
}
else if (Cs >= 1860) {
if (Rs < 1968) {
return 4;
}
else if (Rs >= 1968) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
return 8;
}
}
}
}
}
else if (Cs >= 2221) {
if (SC < 4840) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
}
else if (SC >= 4840) {
if (diffRsCsdiff < 32) {
return 8;
}
else if (diffRsCsdiff >= 32) {
return 4;
}
}
}
}
}
}
else if (AFD >= 17) {
if (TSC < 12) {
if (diffMVdiffVal < -8) {
return 8;
}
else if (diffMVdiffVal >= -8) {
if (MVDiff < 15) {
if (Rs < 1653) {
return 4;
}
else if (Rs >= 1653) {
return 8;
}
}
else if (MVDiff >= 15) {
return 4;
}
}
}
else if (TSC >= 12) {
if (MVDiff < 43) {
if (Rs < 1494) {
return 4;
}
else if (Rs >= 1494) {
return 8;
}
}
else if (MVDiff >= 43) {
return 1;
}
}
}
}
}
else if (Cs >= 2708) {
if (AFD < 12) {
if (Cs < 2908) {
return 8;
}
else if (Cs >= 2908) {
return 4;
}
}
else if (AFD >= 12) {
if (Cs < 2803) {
if (Rs < 2489) {
return 4;
}
else if (Rs >= 2489) {
return 2;
}
}
else if (Cs >= 2803) {
if (Rs < 2663) {
return 8;
}
else if (Rs >= 2663) {
return 1;
}
}
}
}
}
else if (Rs >= 3044) {
if (negBalance < 48) {
if (MVDiff < 8) {
if (SC < 6866) {
if (mcTcor < 96) {
if (posBalance < 60) {
if (diffRsCsdiff < -27) {
if (Rs < 3850) {
return 4;
}
else if (Rs >= 3850) {
return 1;
}
}
else if (diffRsCsdiff >= -27) {
return 1;
}
}
else if (posBalance >= 60) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Rs < 4823) {
return 8;
}
else if (Rs >= 4823) {
return 1;
}
}
}
else if (SC >= 6866) {
if (Cs < 3337) {
return 4;
}
else if (Cs >= 3337) {
if (Rs < 4259) {
return 1;
}
else if (Rs >= 4259) {
return 8;
}
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (negBalance >= 48) {
if (mu_mv_mag_sq < 108) {
if (TSC < 10) {
if (posBalance < 57) {
if (Cs < 3285) {
return 1;
}
else if (Cs >= 3285) {
return 8;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (TSC >= 10) {
if (Rs < 3714) {
return 8;
}
else if (Rs >= 3714) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 108) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 105) {
if (ssDCval < 163) {
if (Rs < 4579) {
if (mu_mv_mag_sq < 40) {
if (Rs < 1934) {
if (ssDCval < 150) {
if (ssDCval < 126) {
if (mcTcor < 97) {
if (TSC < 8) {
if (Rs < 1814) {
if (posBalance < 54) {
if (Rs < 1381) {
return 4;
}
else if (Rs >= 1381) {
return 8;
}
}
else if (posBalance >= 54) {
return 8;
}
}
else if (Rs >= 1814) {
return 4;
}
}
else if (TSC >= 8) {
if (CsDiff < 6) {
if (Rs < 1760) {
if (RsDiff < 5) {
if (Cs < 1769) {
return 8;
}
else if (Cs >= 1769) {
return 4;
}
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (Rs >= 1760) {
return 8;
}
}
else if (CsDiff >= 6) {
if (ssDCval < 120) {
if (negBalance < 46) {
if (Rs < 1297) {
return 4;
}
else if (Rs >= 1297) {
return 1;
}
}
else if (negBalance >= 46) {
return 8;
}
}
else if (ssDCval >= 120) {
if (SC < 2738) {
return 2;
}
else if (SC >= 2738) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 113) {
return 2;
}
else if (refDCval >= 113) {
if (MVDiff < 15) {
if (Rs < 1658) {
return 4;
}
else if (Rs >= 1658) {
return 2;
}
}
else if (MVDiff >= 15) {
return 8;
}
}
}
}
else if (ssDCval >= 126) {
if (tcor < 79) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (tcor >= 79) {
if (negBalance < 52) {
if (Cs < 1414) {
if (Cs < 1332) {
return 1;
}
else if (Cs >= 1332) {
return 8;
}
}
else if (Cs >= 1414) {
if (negBalance < 25) {
return 2;
}
else if (negBalance >= 25) {
if (mcTcor < 90) {
return 1;
}
else if (mcTcor >= 90) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (AFD < 14) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (AFD >= 14) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 52) {
if (Rs < 1230) {
return 1;
}
else if (Rs >= 1230) {
if (Cs < 1334) {
return 2;
}
else if (Cs >= 1334) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 150) {
if (Rs < 1730) {
if (negBalance < 63) {
return 2;
}
else if (negBalance >= 63) {
return 1;
}
}
else if (Rs >= 1730) {
return 8;
}
}
}
else if (Rs >= 1934) {
if (ssDCval < 110) {
if (TSC < 8) {
if (Rs < 3209) {
if (gchDC < 3) {
if (tcor < 95) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (tcor >= 95) {
return 8;
}
}
else if (gchDC >= 3) {
return 8;
}
}
else if (Rs >= 3209) {
return 8;
}
}
else if (TSC >= 8) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (ssDCval >= 110) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (Rs < 3427) {
return 8;
}
else if (Rs >= 3427) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (Rs < 1248) {
if (Rs < 1028) {
if (ssDCval < 114) {
if (Rs < 845) {
if (Rs < 725) {
return 1;
}
else if (Rs >= 725) {
return 8;
}
}
else if (Rs >= 845) {
return 4;
}
}
else if (ssDCval >= 114) {
if (Cs < 2094) {
if (Rs < 989) {
return 8;
}
else if (Rs >= 989) {
if (Rs < 998) {
return 4;
}
else if (Rs >= 998) {
return 8;
}
}
}
else if (Cs >= 2094) {
return 1;
}
}
}
else if (Rs >= 1028) {
if (RsDiff < 3) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (MVDiff < 8) {
if (diffRsCsdiff < 6) {
return 4;
}
else if (diffRsCsdiff >= 6) {
if (Rs < 1138) {
return 8;
}
else if (Rs >= 1138) {
return 4;
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 45) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (mu_mv_mag_sq >= 45) {
if (diffAFD < -1) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
else if (Rs >= 1248) {
if (posBalance < 57) {
if (tcor < 97) {
if (refDCval < 109) {
if (SC < 5148) {
if (mu_mv_mag_sq < 47) {
return 4;
}
else if (mu_mv_mag_sq >= 47) {
if (ssDCval < 107) {
if (negBalance < 29) {
if (Rs < 2742) {
return 8;
}
else if (Rs >= 2742) {
return 4;
}
}
else if (negBalance >= 29) {
return 8;
}
}
else if (ssDCval >= 107) {
if (RsDiff < 6) {
if (Rs < 2578) {
return 8;
}
else if (Rs >= 2578) {
return 4;
}
}
else if (RsDiff >= 6) {
if (SC < 5094) {
return 8;
}
else if (SC >= 5094) {
return 4;
}
}
}
}
}
else if (SC >= 5148) {
return 8;
}
}
else if (refDCval >= 109) {
if (posBalance < 38) {
if (AFD < 11) {
if (mu_mv_mag_sq < 51) {
return 1;
}
else if (mu_mv_mag_sq >= 51) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (posBalance >= 38) {
if (SC < 6993) {
if (ssDCval < 142) {
if (refDCval < 126) {
return 8;
}
else if (refDCval >= 126) {
if (Cs < 1841) {
if (Cs < 1732) {
if (mu_mv_mag_sq < 99) {
if (mcTcor < 88) {
return 1;
}
else if (mcTcor >= 88) {
if (AFD < 15) {
if (ssDCval < 125) {
if (Rs < 2828) {
return 2;
}
else if (Rs >= 2828) {
return 8;
}
}
else if (ssDCval >= 125) {
return 8;
}
}
else if (AFD >= 15) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
return 8;
}
}
}
else if (Cs >= 1732) {
if (ssDCval < 128) {
return 1;
}
else if (ssDCval >= 128) {
if (mcTcor < 92) {
if (Cs < 1753) {
return 8;
}
else if (Cs >= 1753) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 8;
}
}
}
else if (mcTcor >= 92) {
return 8;
}
}
}
}
else if (Cs >= 1841) {
return 8;
}
}
}
else if (ssDCval >= 142) {
if (SC < 5843) {
return 8;
}
else if (SC >= 5843) {
return 1;
}
}
}
else if (SC >= 6993) {
if (AFD < 10) {
if (Rs < 4153) {
return 4;
}
else if (Rs >= 4153) {
return 1;
}
}
else if (AFD >= 10) {
return 8;
}
}
}
}
}
else if (tcor >= 97) {
if (ssDCval < 121) {
if (RsDiff < 5) {
if (mcTcor < 101) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (mcTcor >= 101) {
if (Rs < 1809) {
return 8;
}
else if (Rs >= 1809) {
if (Rs < 2147) {
return 2;
}
else if (Rs >= 2147) {
return 4;
}
}
}
}
else if (RsDiff >= 5) {
if (Rs < 2115) {
return 2;
}
else if (Rs >= 2115) {
if (Rs < 3783) {
return 8;
}
else if (Rs >= 3783) {
return 4;
}
}
}
}
else if (ssDCval >= 121) {
if (gchDC < 1) {
if (Rs < 1294) {
return 4;
}
else if (Rs >= 1294) {
return 8;
}
}
else if (gchDC >= 1) {
if (Rs < 2734) {
if (mu_mv_mag_sq < 107) {
return 8;
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 2554) {
return 8;
}
else if (Rs >= 2554) {
return 4;
}
}
}
else if (Rs >= 2734) {
if (Rs < 2953) {
return 2;
}
else if (Rs >= 2953) {
if (Cs < 1653) {
return 2;
}
else if (Cs >= 1653) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 57) {
if (negBalance < 47) {
if (Rs < 1906) {
if (Rs < 1901) {
if (negBalance < 37) {
if (mcTcor < 89) {
return 1;
}
else if (mcTcor >= 89) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 8;
}
}
}
else if (negBalance >= 37) {
return 8;
}
}
else if (Rs >= 1901) {
return 4;
}
}
else if (Rs >= 1906) {
if (diffRsCsdiff < -53) {
if (diffRsCsdiff < -58) {
return 8;
}
else if (diffRsCsdiff >= -58) {
return 1;
}
}
else if (diffRsCsdiff >= -53) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (Rs < 2412) {
if (ssDCval < 149) {
return 8;
}
else if (ssDCval >= 149) {
if (Rs < 1657) {
return 4;
}
else if (Rs >= 1657) {
return 8;
}
}
}
else if (Rs >= 2412) {
if (ssDCval < 137) {
if (diffAFD < 1) {
if (ssDCval < 107) {
return 4;
}
else if (ssDCval >= 107) {
if (diffAFD < -9) {
return 1;
}
else if (diffAFD >= -9) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (RsCsDiff < 85) {
return 1;
}
else if (RsCsDiff >= 85) {
if (Rs < 2988) {
return 8;
}
else if (Rs >= 2988) {
if (Rs < 3253) {
return 1;
}
else if (Rs >= 3253) {
return 4;
}
}
}
}
}
else if (ssDCval >= 137) {
if (Rs < 2802) {
return 8;
}
else if (Rs >= 2802) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 4579) {
if (Cs < 1851) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (Cs >= 1851) {
if (Cs < 3725) {
if (Rs < 6012) {
if (SC < 6662) {
return 8;
}
else if (SC >= 6662) {
return 1;
}
}
else if (Rs >= 6012) {
return 4;
}
}
else if (Cs >= 3725) {
return 8;
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2264) {
if (Rs < 1085) {
return 1;
}
else if (Rs >= 1085) {
if (Cs < 1338) {
return 8;
}
else if (Cs >= 1338) {
return 2;
}
}
}
else if (Rs >= 2264) {
return 8;
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1937) {
if (Rs < 3191) {
if (Rs < 1688) {
if (mcTcor < 71) {
if (refDCval < 46) {
return 8;
}
else if (refDCval >= 46) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
if (Rs < 1424) {
return 8;
}
else if (Rs >= 1424) {
return 4;
}
}
}
}
else if (mcTcor >= 71) {
if (AFD < 16) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (gchDC < 2) {
if (Rs < 1395) {
if (ssDCval < 63) {
if (Rs < 1200) {
return 4;
}
else if (Rs >= 1200) {
return 1;
}
}
else if (ssDCval >= 63) {
if (diffMVdiffVal < 3) {
if (negBalance < 45) {
if (Rs < 1267) {
if (Rs < 1136) {
return 8;
}
else if (Rs >= 1136) {
return 4;
}
}
else if (Rs >= 1267) {
return 2;
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (Rs >= 1395) {
if (ssDCval < 158) {
return 8;
}
else if (ssDCval >= 158) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (RsCsDiff < 134) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
if (Rs < 1449) {
return 2;
}
else if (Rs >= 1449) {
return 4;
}
}
}
else if (RsCsDiff >= 134) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 4;
}
}
}
}
}
}
else if (AFD >= 16) {
if (mu_mv_mag_sq < 2) {
if (diffRsCsdiff < 38) {
if (SC < 2390) {
return 2;
}
else if (SC >= 2390) {
if (negBalance < 55) {
if (ssDCval < 106) {
if (Cs < 1452) {
return 1;
}
else if (Cs >= 1452) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
if (SC < 3108) {
return 4;
}
else if (SC >= 3108) {
return 1;
}
}
}
}
else if (ssDCval >= 106) {
return 4;
}
}
else if (negBalance >= 55) {
if (ssDCval < 73) {
if (diffAFD < 3) {
if (diffRsCsdiff < -16) {
return 2;
}
else if (diffRsCsdiff >= -16) {
return 4;
}
}
else if (diffAFD >= 3) {
if (Rs < 1461) {
return 2;
}
else if (Rs >= 1461) {
return 1;
}
}
}
else if (ssDCval >= 73) {
if (AFD < 22) {
if (Cs < 1927) {
return 2;
}
else if (Cs >= 1927) {
return 8;
}
}
else if (AFD >= 22) {
return 8;
}
}
}
}
}
else if (diffRsCsdiff >= 38) {
if (Cs < 1677) {
return 1;
}
else if (Cs >= 1677) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 1830) {
if (diffRsCsdiff < 14) {
if (negBalance < 53) {
if (RsCsDiff < 194) {
if (ssDCval < 124) {
return 4;
}
else if (ssDCval >= 124) {
return 8;
}
}
else if (RsCsDiff >= 194) {
return 2;
}
}
else if (negBalance >= 53) {
if (Rs < 1526) {
if (tcor < 59) {
if (ssDCval < 89) {
if (posBalance < 63) {
return 4;
}
else if (posBalance >= 63) {
return 2;
}
}
else if (ssDCval >= 89) {
if (Rs < 913) {
return 2;
}
else if (Rs >= 913) {
return 8;
}
}
}
else if (tcor >= 59) {
if (negBalance < 58) {
if (diffMVdiffVal < 8) {
return 8;
}
else if (diffMVdiffVal >= 8) {
return 4;
}
}
else if (negBalance >= 58) {
if (mu_mv_mag_sq < 15) {
if (SC < 3025) {
return 2;
}
else if (SC >= 3025) {
return 8;
}
}
else if (mu_mv_mag_sq >= 15) {
if (CsDiff < 12) {
return 4;
}
else if (CsDiff >= 12) {
return 8;
}
}
}
}
}
else if (Rs >= 1526) {
return 8;
}
}
}
else if (diffRsCsdiff >= 14) {
if (Cs < 1546) {
if (mu_mv_mag_sq < 23) {
if (tcor < 70) {
if (diffMVdiffVal < 1) {
if (Rs < 1394) {
return 8;
}
else if (Rs >= 1394) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
if (MVDiff < 20) {
if (Rs < 1536) {
return 8;
}
else if (Rs >= 1536) {
return 4;
}
}
else if (MVDiff >= 20) {
if (Cs < 1528) {
return 2;
}
else if (Cs >= 1528) {
return 8;
}
}
}
}
else if (tcor >= 70) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1448) {
return 1;
}
else if (Rs >= 1448) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (posBalance < 73) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
if (Rs < 1418) {
return 1;
}
else if (Rs >= 1418) {
return 8;
}
}
}
else if (posBalance >= 73) {
return 4;
}
}
}
else if (Cs >= 1546) {
if (tcor < 49) {
return 2;
}
else if (tcor >= 49) {
if (diffMVdiffVal < 9) {
if (tcor < 80) {
if (RsDiff < 12) {
if (gchDC < 2) {
if (Rs < 1558) {
if (Rs < 899) {
return 8;
}
else if (Rs >= 899) {
return 4;
}
}
else if (Rs >= 1558) {
return 8;
}
}
else if (gchDC >= 2) {
return 4;
}
}
else if (RsDiff >= 12) {
return 1;
}
}
else if (tcor >= 80) {
if (TSC < 10) {
if (Rs < 1569) {
return 4;
}
else if (Rs >= 1569) {
return 2;
}
}
else if (TSC >= 10) {
return 1;
}
}
}
else if (diffMVdiffVal >= 9) {
if (SC < 3339) {
return 2;
}
else if (SC >= 3339) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1830) {
if (Rs < 1324) {
return 2;
}
else if (Rs >= 1324) {
if (RsDiff < 10) {
if (Cs < 1904) {
if (Rs < 1478) {
return 4;
}
else if (Rs >= 1478) {
return 2;
}
}
else if (Cs >= 1904) {
return 8;
}
}
else if (RsDiff >= 10) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 1688) {
if (AFD < 20) {
if (posBalance < 49) {
if (AFD < 16) {
if (ssDCval < 96) {
if (Cs < 1480) {
if (Rs < 1998) {
return 8;
}
else if (Rs >= 1998) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
return 1;
}
}
}
else if (Cs >= 1480) {
if (mcTcor < 81) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (mcTcor >= 81) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
if (diffRsCsdiff < -1) {
return 8;
}
else if (diffRsCsdiff >= -1) {
if (Cs < 1768) {
return 8;
}
else if (Cs >= 1768) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 96) {
if (RsCsDiff < 163) {
if (diffRsCsdiff < 80) {
return 8;
}
else if (diffRsCsdiff >= 80) {
return 1;
}
}
else if (RsCsDiff >= 163) {
if (tcor < 91) {
if (Rs < 2387) {
return 2;
}
else if (Rs >= 2387) {
return 8;
}
}
else if (tcor >= 91) {
return 4;
}
}
}
}
else if (AFD >= 16) {
if (negBalance < 48) {
if (Rs < 2021) {
return 8;
}
else if (Rs >= 2021) {
return 2;
}
}
else if (negBalance >= 48) {
if (RsDiff < 9) {
if (Cs < 1537) {
if (Rs < 2108) {
return 8;
}
else if (Rs >= 2108) {
return 2;
}
}
else if (Cs >= 1537) {
return 1;
}
}
else if (RsDiff >= 9) {
return 8;
}
}
}
}
else if (posBalance >= 49) {
if (Rs < 2477) {
if (TSC < 9) {
if (ssDCval < 85) {
if (mu_mv_mag_sq < 21) {
if (RsCsDiff < 123) {
if (Rs < 1800) {
return 2;
}
else if (Rs >= 1800) {
return 8;
}
}
else if (RsCsDiff >= 123) {
if (SC < 3442) {
return 1;
}
else if (SC >= 3442) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (diffAFD < -4) {
if (Rs < 2114) {
return 1;
}
else if (Rs >= 2114) {
return 8;
}
}
else if (diffAFD >= -4) {
if (ssDCval < 81) {
if (AFD < 10) {
if (Rs < 1895) {
return 8;
}
else if (Rs >= 1895) {
if (Rs < 2132) {
return 4;
}
else if (Rs >= 2132) {
return 8;
}
}
}
else if (AFD >= 10) {
return 8;
}
}
else if (ssDCval >= 81) {
return 4;
}
}
}
}
else if (ssDCval >= 85) {
if (ssDCval < 100) {
if (Cs < 1529) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (Cs >= 1529) {
if (AFD < 16) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
return 2;
}
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (ssDCval >= 100) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
else if (TSC >= 9) {
if (TSC < 13) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 45) {
if (mcTcor < 92) {
if (Rs < 1795) {
if (Cs < 1806) {
return 1;
}
else if (Cs >= 1806) {
return 8;
}
}
else if (Rs >= 1795) {
if (diffAFD < -2) {
if (Rs < 2134) {
return 8;
}
else if (Rs >= 2134) {
return 4;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (Rs < 2123) {
return 8;
}
else if (Rs >= 2123) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (tcor < 68) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 1;
}
}
else if (tcor >= 68) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
else if (ssDCval >= 78) {
if (mcTcor < 95) {
if (tcor < 59) {
if (Rs < 2010) {
return 8;
}
else if (Rs >= 2010) {
return 4;
}
}
else if (tcor >= 59) {
if (MVDiff < 15) {
if (TSC < 10) {
if (ssDCval < 85) {
return 4;
}
else if (ssDCval >= 85) {
if (tcor < 75) {
if (Rs < 2026) {
return 8;
}
else if (Rs >= 2026) {
return 4;
}
}
else if (tcor >= 75) {
return 8;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (MVDiff >= 15) {
if (refDCval < 114) {
if (refDCval < 79) {
if (Rs < 1913) {
return 1;
}
else if (Rs >= 1913) {
return 8;
}
}
else if (refDCval >= 79) {
return 8;
}
}
else if (refDCval >= 114) {
if (mu_mv_mag_sq < 15) {
return 1;
}
else if (mu_mv_mag_sq >= 15) {
return 8;
}
}
}
}
}
else if (mcTcor >= 95) {
if (refDCval < 92) {
if (negBalance < 46) {
return 1;
}
else if (negBalance >= 46) {
if (Rs < 2158) {
return 2;
}
else if (Rs >= 2158) {
return 8;
}
}
}
else if (refDCval >= 92) {
if (mu_mv_mag_sq < 59) {
return 8;
}
else if (mu_mv_mag_sq >= 59) {
return 1;
}
}
}
}
}
else if (TSC >= 13) {
if (mu_mv_mag_sq < 25) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
return 1;
}
}
else if (mu_mv_mag_sq >= 25) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
if (Rs < 1852) {
return 1;
}
else if (Rs >= 1852) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2477) {
if (ssDCval < 106) {
if (negBalance < 48) {
if (mu_mv_mag_sq < 59) {
if (RsCsDiff < 121) {
if (Rs < 2719) {
return 8;
}
else if (Rs >= 2719) {
return 4;
}
}
else if (RsCsDiff >= 121) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (ssDCval < 98) {
if (Cs < 1648) {
if (refDCval < 94) {
return 2;
}
else if (refDCval >= 94) {
return 4;
}
}
else if (Cs >= 1648) {
if (tcor < 87) {
if (Rs < 2601) {
return 8;
}
else if (Rs >= 2601) {
return 2;
}
}
else if (tcor >= 87) {
return 8;
}
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (posBalance < 55) {
if (Rs < 2830) {
return 4;
}
else if (Rs >= 2830) {
return 1;
}
}
else if (posBalance >= 55) {
if (Cs < 1399) {
return 2;
}
else if (Cs >= 1399) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
if (ssDCval < 104) {
if (mu_mv_mag_sq < 22) {
if (refDCval < 97) {
if (Cs < 1603) {
if (RsCsDiff < 157) {
return 8;
}
else if (RsCsDiff >= 157) {
if (SC < 4320) {
return 2;
}
else if (SC >= 4320) {
return 4;
}
}
}
else if (Cs >= 1603) {
if (ssDCval < 67) {
if (Rs < 2784) {
return 2;
}
else if (Rs >= 2784) {
return 8;
}
}
else if (ssDCval >= 67) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (tcor < 80) {
return 4;
}
else if (tcor >= 80) {
if (posBalance < 59) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 8;
}
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 97) {
return 8;
}
}
else if (mu_mv_mag_sq >= 22) {
if (ssDCval < 96) {
if (diffMVdiffVal < 5) {
if (diffAFD < 5) {
if (posBalance < 54) {
if (Rs < 2687) {
return 4;
}
else if (Rs >= 2687) {
return 8;
}
}
else if (posBalance >= 54) {
return 8;
}
}
else if (diffAFD >= 5) {
return 4;
}
}
else if (diffMVdiffVal >= 5) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
return 8;
}
}
}
else if (ssDCval >= 96) {
if (ssDCval < 97) {
return 2;
}
else if (ssDCval >= 97) {
return 8;
}
}
}
}
else if (ssDCval >= 104) {
if (Rs < 2779) {
return 2;
}
else if (Rs >= 2779) {
return 4;
}
}
}
}
else if (ssDCval >= 106) {
if (AFD < 14) {
if (RsDiff < 13) {
if (refDCval < 111) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 4;
}
}
else if (refDCval >= 111) {
return 8;
}
}
else if (RsDiff >= 13) {
return 4;
}
}
else if (AFD >= 14) {
if (Cs < 1416) {
if (Rs < 2820) {
return 8;
}
else if (Rs >= 2820) {
if (Rs < 3068) {
return 1;
}
else if (Rs >= 3068) {
return 4;
}
}
}
else if (Cs >= 1416) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 20) {
if (TSC < 14) {
if (refDCval < 87) {
if (Rs < 2372) {
if (mcTcor < 84) {
if (Cs < 1494) {
if (Rs < 1859) {
return 2;
}
else if (Rs >= 1859) {
return 8;
}
}
else if (Cs >= 1494) {
if (Rs < 2211) {
return 1;
}
else if (Rs >= 2211) {
return 2;
}
}
}
else if (mcTcor >= 84) {
if (Cs < 1415) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (AFD < 27) {
return 2;
}
else if (AFD >= 27) {
return 4;
}
}
}
else if (Cs >= 1415) {
if (tcor < 59) {
if (ssDCval < 65) {
if (Rs < 1946) {
return 1;
}
else if (Rs >= 1946) {
return 4;
}
}
else if (ssDCval >= 65) {
return 2;
}
}
else if (tcor >= 59) {
if (mcTcor < 92) {
if (Rs < 2251) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
if (gchDC < 1) {
if (Cs < 1788) {
return 4;
}
else if (Cs >= 1788) {
return 1;
}
}
else if (gchDC >= 1) {
if (diffRsCsdiff < -9) {
if (diffRsCsdiff < -91) {
return 8;
}
else if (diffRsCsdiff >= -91) {
return 4;
}
}
else if (diffRsCsdiff >= -9) {
return 8;
}
}
}
}
else if (Rs >= 2251) {
return 8;
}
}
else if (mcTcor >= 92) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
if (Rs < 1966) {
return 2;
}
else if (Rs >= 1966) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 2372) {
if (gchDC < 7) {
if (ssDCval < 80) {
if (mcTcor < 82) {
if (Cs < 1567) {
return 2;
}
else if (Cs >= 1567) {
return 4;
}
}
else if (mcTcor >= 82) {
return 2;
}
}
else if (ssDCval >= 80) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
return 4;
}
}
}
else if (gchDC >= 7) {
return 8;
}
}
}
else if (refDCval >= 87) {
if (RsDiff < 17) {
if (mcTcor < 89) {
if (Rs < 2342) {
if (Cs < 1553) {
if (MVDiff < 38) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
if (ssDCval < 126) {
return 4;
}
else if (ssDCval >= 126) {
if (Rs < 2118) {
return 1;
}
else if (Rs >= 2118) {
return 8;
}
}
}
}
else if (MVDiff >= 38) {
return 8;
}
}
else if (Cs >= 1553) {
return 8;
}
}
else if (Rs >= 2342) {
if (negBalance < 58) {
if (diffMVdiffVal < 7) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (diffMVdiffVal >= 7) {
if (Rs < 2631) {
return 4;
}
else if (Rs >= 2631) {
return 8;
}
}
}
else if (negBalance >= 58) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (AFD < 23) {
return 4;
}
else if (AFD >= 23) {
return 1;
}
}
}
}
}
else if (mcTcor >= 89) {
if (TSC < 10) {
if (diffMVdiffVal < -5) {
if (AFD < 23) {
return 1;
}
else if (AFD >= 23) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (ssDCval < 100) {
if (Cs < 1574) {
return 2;
}
else if (Cs >= 1574) {
return 4;
}
}
else if (ssDCval >= 100) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
}
}
}
else if (TSC >= 10) {
if (Rs < 1970) {
if (CsDiff < 11) {
return 1;
}
else if (CsDiff >= 11) {
return 2;
}
}
else if (Rs >= 1970) {
if (mcTcor < 90) {
if (Rs < 2631) {
return 4;
}
else if (Rs >= 2631) {
return 2;
}
}
else if (mcTcor >= 90) {
if (MVDiff < 54) {
if (negBalance < 52) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (RsDiff < 13) {
return 2;
}
else if (RsDiff >= 13) {
if (Rs < 2527) {
return 8;
}
else if (Rs >= 2527) {
return 1;
}
}
}
}
else if (negBalance >= 52) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (MVDiff >= 54) {
if (Rs < 2826) {
return 1;
}
else if (Rs >= 2826) {
return 4;
}
}
}
}
}
}
}
else if (RsDiff >= 17) {
if (negBalance < 60) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 2;
}
}
else if (negBalance >= 60) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 2;
}
}
}
}
}
else if (TSC >= 14) {
if (ssDCval < 121) {
if (RsDiff < 16) {
if (mcTcor < 90) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
if (Rs < 1915) {
return 1;
}
else if (Rs >= 1915) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (ssDCval < 109) {
return 4;
}
else if (ssDCval >= 109) {
if (Rs < 2857) {
return 8;
}
else if (Rs >= 2857) {
return 4;
}
}
}
}
else if (RsDiff >= 16) {
if (tcor < 52) {
if (MVDiff < 109) {
return 2;
}
else if (MVDiff >= 109) {
return 1;
}
}
else if (tcor >= 52) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 1457) {
return 2;
}
else if (Cs >= 1457) {
return 8;
}
}
}
}
}
else if (ssDCval >= 121) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
if (AFD < 32) {
if (TSC < 15) {
return 1;
}
else if (TSC >= 15) {
return 8;
}
}
else if (AFD >= 32) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 3191) {
if (tcor < 85) {
if (RsDiff < 13) {
if (posBalance < 50) {
if (AFD < 14) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (posBalance < 48) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 3572) {
return 4;
}
else if (Rs >= 3572) {
return 8;
}
}
}
else if (posBalance >= 48) {
return 8;
}
}
}
else if (AFD >= 14) {
if (tcor < 81) {
return 4;
}
else if (tcor >= 81) {
return 2;
}
}
}
else if (posBalance >= 50) {
if (Rs < 3325) {
if (tcor < 76) {
if (Cs < 1375) {
return 1;
}
else if (Cs >= 1375) {
return 4;
}
}
else if (tcor >= 76) {
if (Rs < 3320) {
return 8;
}
else if (Rs >= 3320) {
return 4;
}
}
}
else if (Rs >= 3325) {
if (ssDCval < 83) {
if (Cs < 1516) {
return 2;
}
else if (Cs >= 1516) {
return 8;
}
}
else if (ssDCval >= 83) {
if (tcor < 67) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 8;
}
}
else if (tcor >= 67) {
return 8;
}
}
}
}
}
else if (RsDiff >= 13) {
if (mu_mv_mag_sq < 11) {
if (refDCval < 101) {
if (negBalance < 71) {
if (Cs < 1918) {
if (CsDiff < 12) {
return 2;
}
else if (CsDiff >= 12) {
if (Rs < 5214) {
return 8;
}
else if (Rs >= 5214) {
return 2;
}
}
}
else if (Cs >= 1918) {
return 4;
}
}
else if (negBalance >= 71) {
if (Rs < 3314) {
return 1;
}
else if (Rs >= 3314) {
return 8;
}
}
}
else if (refDCval >= 101) {
if (TSC < 13) {
if (RsCsDiff < 283) {
if (AFD < 20) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (SC < 4818) {
return 4;
}
else if (SC >= 4818) {
return 8;
}
}
}
else if (AFD >= 20) {
if (Rs < 3249) {
return 8;
}
else if (Rs >= 3249) {
return 1;
}
}
}
else if (RsCsDiff >= 283) {
if (Rs < 5118) {
if (RsCsDiff < 383) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
return 2;
}
}
else if (RsCsDiff >= 383) {
if (mcTcor < 88) {
if (Rs < 3438) {
return 1;
}
else if (Rs >= 3438) {
return 8;
}
}
else if (mcTcor >= 88) {
return 2;
}
}
}
else if (Rs >= 5118) {
if (MVDiff < 18) {
if (Rs < 6369) {
return 8;
}
else if (Rs >= 6369) {
return 2;
}
}
else if (MVDiff >= 18) {
if (AFD < 20) {
if (Rs < 6268) {
return 2;
}
else if (Rs >= 6268) {
return 8;
}
}
else if (AFD >= 20) {
return 8;
}
}
}
}
}
else if (TSC >= 13) {
if (mcTcor < 90) {
if (Cs < 1641) {
return 4;
}
else if (Cs >= 1641) {
if (Cs < 1690) {
return 2;
}
else if (Cs >= 1690) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (CsDiff < 9) {
if (Rs < 4802) {
return 2;
}
else if (Rs >= 4802) {
return 1;
}
}
else if (CsDiff >= 9) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (negBalance < 57) {
if (CsDiff < 7) {
if (Cs < 1921) {
if (Rs < 3253) {
return 1;
}
else if (Rs >= 3253) {
if (mcTcor < 85) {
if (Cs < 1665) {
return 4;
}
else if (Cs >= 1665) {
return 8;
}
}
else if (mcTcor >= 85) {
return 4;
}
}
}
else if (Cs >= 1921) {
return 8;
}
}
else if (CsDiff >= 7) {
if (posBalance < 49) {
if (mcTcor < 85) {
return 1;
}
else if (mcTcor >= 85) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 1;
}
}
}
}
else if (posBalance >= 49) {
if (posBalance < 66) {
if (AFD < 24) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
if (mu_mv_mag_sq < 17) {
return 4;
}
else if (mu_mv_mag_sq >= 17) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
else if (AFD >= 24) {
if (SC < 5517) {
if (Rs < 3354) {
return 2;
}
else if (Rs >= 3354) {
if (Rs < 3928) {
return 8;
}
else if (Rs >= 3928) {
return 1;
}
}
}
else if (SC >= 5517) {
if (RsDiff < 17) {
return 1;
}
else if (RsDiff >= 17) {
return 4;
}
}
}
}
else if (posBalance >= 66) {
if (tcor < 72) {
return 2;
}
else if (tcor >= 72) {
if (Cs < 1694) {
return 1;
}
else if (Cs >= 1694) {
if (Rs < 4040) {
return 8;
}
else if (Rs >= 4040) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 57) {
if (tcor < 60) {
if (TSC < 20) {
return 2;
}
else if (TSC >= 20) {
return 1;
}
}
else if (tcor >= 60) {
if (ssDCval < 111) {
if (Cs < 1686) {
return 4;
}
else if (Cs >= 1686) {
return 8;
}
}
else if (ssDCval >= 111) {
if (Rs < 3518) {
return 2;
}
else if (Rs >= 3518) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 85) {
if (refDCval < 101) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (mcTcor < 92) {
if (negBalance < 50) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (Cs < 1332) {
return 4;
}
else if (Cs >= 1332) {
return 2;
}
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (mcTcor >= 92) {
if (posBalance < 45) {
if (CsDiff < 8) {
if (negBalance < 40) {
if (Rs < 4375) {
if (Rs < 3395) {
return 4;
}
else if (Rs >= 3395) {
return 2;
}
}
else if (Rs >= 4375) {
if (Rs < 4525) {
return 8;
}
else if (Rs >= 4525) {
return 4;
}
}
}
else if (negBalance >= 40) {
return 2;
}
}
else if (CsDiff >= 8) {
if (Rs < 3644) {
return 2;
}
else if (Rs >= 3644) {
return 1;
}
}
}
else if (posBalance >= 45) {
return 2;
}
}
}
}
else if (refDCval >= 101) {
if (RsDiff < 11) {
if (mu_mv_mag_sq < 85) {
if (posBalance < 38) {
if (Cs < 1839) {
return 2;
}
else if (Cs >= 1839) {
return 1;
}
}
else if (posBalance >= 38) {
if (mu_mv_mag_sq < 79) {
return 8;
}
else if (mu_mv_mag_sq >= 79) {
if (Rs < 4183) {
return 8;
}
else if (Rs >= 4183) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 3455) {
return 4;
}
else if (Rs >= 3455) {
return 1;
}
}
}
else if (RsDiff >= 11) {
if (MVDiff < 23) {
if (RsDiff < 14) {
if (posBalance < 54) {
if (diffRsCsdiff < 85) {
return 2;
}
else if (diffRsCsdiff >= 85) {
return 8;
}
}
else if (posBalance >= 54) {
if (Cs < 1432) {
return 1;
}
else if (Cs >= 1432) {
return 8;
}
}
}
else if (RsDiff >= 14) {
if (RsCsDiff < 227) {
if (Rs < 4574) {
return 4;
}
else if (Rs >= 4574) {
return 2;
}
}
else if (RsCsDiff >= 227) {
return 2;
}
}
}
else if (MVDiff >= 23) {
if (AFD < 20) {
if (MVDiff < 78) {
return 8;
}
else if (MVDiff >= 78) {
return 1;
}
}
else if (AFD >= 20) {
if (Rs < 3810) {
return 8;
}
else if (Rs >= 3810) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 1937) {
if (AFD < 12) {
if (ssDCval < 100) {
if (Rs < 4005) {
if (negBalance < 41) {
if (ssDCval < 85) {
if (mcTcor < 93) {
if (Cs < 2786) {
if (posBalance < 55) {
if (MVDiff < 4) {
if (Rs < 3475) {
return 1;
}
else if (Rs >= 3475) {
return 4;
}
}
else if (MVDiff >= 4) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (Rs < 2002) {
return 1;
}
else if (Rs >= 2002) {
return 8;
}
}
}
else if (Cs >= 2786) {
if (SC < 7537) {
return 4;
}
else if (SC >= 7537) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (mu_mv_mag_sq < 97) {
if (Rs < 3684) {
return 1;
}
else if (Rs >= 3684) {
return 4;
}
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 3250) {
return 4;
}
else if (Cs >= 3250) {
return 8;
}
}
}
}
else if (ssDCval >= 85) {
if (Cs < 2333) {
return 8;
}
else if (Cs >= 2333) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (Cs < 2656) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (Rs < 2823) {
return 8;
}
else if (Rs >= 2823) {
if (tcor < 89) {
if (SC < 4952) {
if (Rs < 2837) {
return 8;
}
else if (Rs >= 2837) {
return 4;
}
}
else if (SC >= 4952) {
return 8;
}
}
else if (tcor >= 89) {
if (ssDCval < 96) {
if (diffAFD < -2) {
if (Rs < 3595) {
return 8;
}
else if (Rs >= 3595) {
return 2;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (ssDCval >= 96) {
return 8;
}
}
}
}
}
else if (Cs >= 2656) {
if (negBalance < 52) {
if (SC < 5687) {
if (Rs < 2582) {
if (Rs < 1363) {
return 8;
}
else if (Rs >= 1363) {
return 4;
}
}
else if (Rs >= 2582) {
return 8;
}
}
else if (SC >= 5687) {
if (Cs < 4178) {
if (ssDCval < 80) {
if (Rs < 2745) {
return 4;
}
else if (Rs >= 2745) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 4;
}
}
}
else if (ssDCval >= 80) {
return 4;
}
}
else if (Cs >= 4178) {
return 8;
}
}
}
else if (negBalance >= 52) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
if (Rs < 3125) {
return 4;
}
else if (Rs >= 3125) {
return 8;
}
}
}
}
}
}
else if (Rs >= 4005) {
if (MVDiff < 10) {
if (ssDCval < 73) {
if (Cs < 3410) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 4029) {
return 4;
}
else if (Rs >= 4029) {
return 1;
}
}
}
}
else if (Cs >= 3410) {
if (Rs < 5783) {
return 8;
}
else if (Rs >= 5783) {
return 2;
}
}
}
else if (ssDCval >= 73) {
if (Rs < 4229) {
if (Cs < 2672) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 2;
}
}
else if (Cs >= 2672) {
if (Rs < 4217) {
return 4;
}
else if (Rs >= 4217) {
return 8;
}
}
}
else if (Rs >= 4229) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (TSC < 9) {
if (Cs < 4642) {
return 2;
}
else if (Cs >= 4642) {
return 8;
}
}
else if (TSC >= 9) {
if (Cs < 2121) {
return 2;
}
else if (Cs >= 2121) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 10) {
if (Rs < 5106) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 8;
}
}
else if (Rs >= 5106) {
if (ssDCval < 69) {
if (Rs < 5732) {
return 8;
}
else if (Rs >= 5732) {
return 4;
}
}
else if (ssDCval >= 69) {
return 2;
}
}
}
}
}
else if (ssDCval >= 100) {
if (negBalance < 45) {
if (posBalance < 39) {
if (ssDCval < 118) {
if (SC < 9119) {
if (RsDiff < 9) {
if (Rs < 3601) {
return 8;
}
else if (Rs >= 3601) {
return 1;
}
}
else if (RsDiff >= 9) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (SC >= 9119) {
return 1;
}
}
else if (ssDCval >= 118) {
if (posBalance < 31) {
if (Rs < 1810) {
return 1;
}
else if (Rs >= 1810) {
return 8;
}
}
else if (posBalance >= 31) {
if (Cs < 3869) {
return 4;
}
else if (Cs >= 3869) {
return 1;
}
}
}
}
else if (posBalance >= 39) {
if (MVDiff < 30) {
if (Rs < 7631) {
if (Cs < 3094) {
if (Rs < 4395) {
return 8;
}
else if (Rs >= 4395) {
if (ssDCval < 110) {
if (RsDiff < 11) {
if (Rs < 4771) {
return 1;
}
else if (Rs >= 4771) {
return 8;
}
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (ssDCval >= 110) {
if (Rs < 6163) {
return 8;
}
else if (Rs >= 6163) {
return 2;
}
}
}
}
else if (Cs >= 3094) {
if (MVDiff < 21) {
if (ssDCval < 129) {
if (Rs < 6363) {
if (CsDiff < 15) {
if (diffMVdiffVal < -2) {
if (RsCsDiff < 146) {
return 4;
}
else if (RsCsDiff >= 146) {
if (Rs < 5791) {
return 8;
}
else if (Rs >= 5791) {
return 4;
}
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (CsDiff >= 15) {
return 4;
}
}
else if (Rs >= 6363) {
if (Rs < 6704) {
return 4;
}
else if (Rs >= 6704) {
return 8;
}
}
}
else if (ssDCval >= 129) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (mcTcor < 93) {
if (Rs < 5727) {
return 1;
}
else if (Rs >= 5727) {
return 8;
}
}
else if (mcTcor >= 93) {
return 8;
}
}
}
}
else if (MVDiff >= 21) {
if (Rs < 2482) {
return 4;
}
else if (Rs >= 2482) {
return 2;
}
}
}
}
else if (Rs >= 7631) {
if (Rs < 8598) {
return 2;
}
else if (Rs >= 8598) {
return 1;
}
}
}
else if (MVDiff >= 30) {
if (RsDiff < 10) {
if (Rs < 2901) {
return 8;
}
else if (Rs >= 2901) {
return 1;
}
}
else if (RsDiff >= 10) {
return 2;
}
}
}
}
else if (negBalance >= 45) {
if (tcor < 93) {
if (mu_mv_mag_sq < 83) {
if (Rs < 7725) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (Rs < 3487) {
return 8;
}
else if (Rs >= 3487) {
return 2;
}
}
}
else if (Rs >= 7725) {
return 1;
}
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 3739) {
if (diffRsCsdiff < 1) {
if (Cs < 2534) {
return 1;
}
else if (Cs >= 2534) {
return 2;
}
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
else if (Rs >= 3739) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (mu_mv_mag_sq < 92) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
}
}
}
else if (tcor >= 93) {
if (posBalance < 49) {
if (Rs < 5663) {
return 2;
}
else if (Rs >= 5663) {
return 4;
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 75) {
if (AFD < 18) {
if (RsCsDiff < 195) {
if (Rs < 1545) {
if (negBalance < 44) {
if (MVDiff < 28) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
else if (MVDiff >= 28) {
if (ssDCval < 73) {
if (Rs < 867) {
return 4;
}
else if (Rs >= 867) {
return 2;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
}
else if (negBalance >= 44) {
if (TSC < 8) {
if (CsDiff < 11) {
if (Rs < 1155) {
return 2;
}
else if (Rs >= 1155) {
return 4;
}
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (TSC >= 8) {
if (Cs < 3216) {
return 4;
}
else if (Cs >= 3216) {
return 8;
}
}
}
}
else if (Rs >= 1545) {
if (MVDiff < 3) {
if (ssDCval < 86) {
if (AFD < 15) {
if (Cs < 2629) {
return 8;
}
else if (Cs >= 2629) {
return 2;
}
}
else if (AFD >= 15) {
return 1;
}
}
else if (ssDCval >= 86) {
if (ssDCval < 105) {
if (Rs < 2864) {
if (Cs < 2839) {
return 8;
}
else if (Cs >= 2839) {
return 4;
}
}
else if (Rs >= 2864) {
return 8;
}
}
else if (ssDCval >= 105) {
if (ssDCval < 123) {
if (Rs < 2836) {
return 8;
}
else if (Rs >= 2836) {
return 4;
}
}
else if (ssDCval >= 123) {
if (Rs < 3330) {
return 1;
}
else if (Rs >= 3330) {
return 8;
}
}
}
}
}
else if (MVDiff >= 3) {
if (negBalance < 42) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 22) {
if (Rs < 4287) {
return 1;
}
else if (Rs >= 4287) {
return 2;
}
}
else if (MVDiff >= 22) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (AFD < 16) {
if (ssDCval < 143) {
if (mu_mv_mag_sq < 16) {
if (Rs < 3003) {
return 2;
}
else if (Rs >= 3003) {
return 8;
}
}
else if (mu_mv_mag_sq >= 16) {
return 8;
}
}
else if (ssDCval >= 143) {
return 1;
}
}
else if (AFD >= 16) {
if (Rs < 3226) {
return 2;
}
else if (Rs >= 3226) {
return 8;
}
}
}
}
}
else if (negBalance >= 42) {
if (posBalance < 45) {
if (Cs < 2483) {
if (Rs < 3435) {
return 8;
}
else if (Rs >= 3435) {
return 4;
}
}
else if (Cs >= 2483) {
return 1;
}
}
else if (posBalance >= 45) {
if (mu_mv_mag_sq < 57) {
if (Rs < 3404) {
if (MVDiff < 13) {
if (SC < 5140) {
if (mcTcor < 94) {
if (TSC < 12) {
if (Rs < 2635) {
if (diffRsCsdiff < 58) {
return 8;
}
else if (diffRsCsdiff >= 58) {
if (Rs < 2284) {
return 8;
}
else if (Rs >= 2284) {
return 4;
}
}
}
else if (Rs >= 2635) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
if (Rs < 2905) {
return 4;
}
else if (Rs >= 2905) {
return 8;
}
}
}
}
}
}
else if (TSC >= 12) {
if (Rs < 1964) {
return 4;
}
else if (Rs >= 1964) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (gchDC < 2) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (gchDC >= 2) {
if (RsCsDiff < 163) {
return 2;
}
else if (RsCsDiff >= 163) {
if (Rs < 2080) {
return 1;
}
else if (Rs >= 2080) {
return 8;
}
}
}
}
}
else if (SC >= 5140) {
if (MVDiff < 6) {
if (AFD < 16) {
if (MVDiff < 4) {
if (diffMVdiffVal < -1) {
if (SC < 5255) {
return 8;
}
else if (SC >= 5255) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
return 4;
}
}
}
else if (AFD >= 16) {
if (tcor < 80) {
return 4;
}
else if (tcor >= 80) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
return 1;
}
}
}
}
}
else if (MVDiff >= 6) {
if (Rs < 3026) {
return 8;
}
else if (Rs >= 3026) {
if (Rs < 3269) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (Rs >= 3269) {
return 8;
}
}
}
}
}
else if (MVDiff >= 13) {
if (ssDCval < 52) {
return 2;
}
else if (ssDCval >= 52) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
return 8;
}
}
}
}
}
else if (Rs >= 3404) {
return 8;
}
}
else if (mu_mv_mag_sq >= 57) {
if (CsDiff < 9) {
if (Cs < 3108) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (TSC < 12) {
if (posBalance < 57) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (posBalance >= 57) {
if (negBalance < 49) {
if (Rs < 4784) {
return 8;
}
else if (Rs >= 4784) {
return 4;
}
}
else if (negBalance >= 49) {
return 4;
}
}
}
else if (TSC >= 12) {
return 8;
}
}
}
else if (Cs >= 3108) {
if (Rs < 4382) {
return 1;
}
else if (Rs >= 4382) {
return 8;
}
}
}
else if (CsDiff >= 9) {
if (Rs < 2963) {
if (gchDC < 2) {
if (diffRsCsdiff < 50) {
if (TSC < 11) {
if (SC < 4892) {
return 8;
}
else if (SC >= 4892) {
if (SC < 5498) {
return 4;
}
else if (SC >= 5498) {
return 8;
}
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (diffRsCsdiff >= 50) {
return 4;
}
}
else if (gchDC >= 2) {
if (Rs < 2297) {
if (Rs < 2223) {
return 8;
}
else if (Rs >= 2223) {
return 1;
}
}
else if (Rs >= 2297) {
return 2;
}
}
}
else if (Rs >= 2963) {
if (SC < 6496) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
if (Rs < 3063) {
return 1;
}
else if (Rs >= 3063) {
return 8;
}
}
}
else if (SC >= 6496) {
return 1;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 195) {
if (tcor < 81) {
if (refDCval < 57) {
if (negBalance < 56) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (diffAFD < -4) {
if (Rs < 2778) {
if (Rs < 1850) {
return 1;
}
else if (Rs >= 1850) {
return 4;
}
}
else if (Rs >= 2778) {
return 2;
}
}
else if (diffAFD >= -4) {
return 4;
}
}
}
else if (negBalance >= 56) {
if (TSC < 11) {
if (Rs < 2029) {
return 1;
}
else if (Rs >= 2029) {
return 8;
}
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (refDCval >= 57) {
if (tcor < 70) {
if (refDCval < 132) {
if (MVDiff < 35) {
if (ssDCval < 70) {
if (Rs < 2526) {
return 4;
}
else if (Rs >= 2526) {
return 8;
}
}
else if (ssDCval >= 70) {
if (Rs < 1492) {
return 2;
}
else if (Rs >= 1492) {
return 8;
}
}
}
else if (MVDiff >= 35) {
return 2;
}
}
else if (refDCval >= 132) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
if (posBalance < 54) {
if (Rs < 4215) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (Rs >= 4215) {
return 1;
}
}
else if (posBalance >= 54) {
if (refDCval < 134) {
return 1;
}
else if (refDCval >= 134) {
return 8;
}
}
}
}
}
else if (tcor >= 70) {
if (RsCsDiff < 331) {
if (diffRsCsdiff < -131) {
if (gchDC < 1) {
if (Rs < 3781) {
return 8;
}
else if (Rs >= 3781) {
if (Rs < 3967) {
return 1;
}
else if (Rs >= 3967) {
return 2;
}
}
}
else if (gchDC >= 1) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 4;
}
}
}
else if (diffRsCsdiff >= -131) {
if (negBalance < 55) {
if (mu_mv_mag_sq < 53) {
if (ssDCval < 80) {
if (MVDiff < 12) {
if (Rs < 3164) {
if (Rs < 2039) {
return 4;
}
else if (Rs >= 2039) {
return 8;
}
}
else if (Rs >= 3164) {
return 1;
}
}
else if (MVDiff >= 12) {
if (Cs < 2034) {
return 1;
}
else if (Cs >= 2034) {
return 8;
}
}
}
else if (ssDCval >= 80) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (CsDiff < 13) {
if (RsDiff < 13) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
if (RsCsDiff < 215) {
return 8;
}
else if (RsCsDiff >= 215) {
return 4;
}
}
}
else if (RsDiff >= 13) {
if (mcTcor < 87) {
if (RsCsDiff < 255) {
if (Rs < 3098) {
return 8;
}
else if (Rs >= 3098) {
return 2;
}
}
else if (RsCsDiff >= 255) {
return 8;
}
}
else if (mcTcor >= 87) {
return 4;
}
}
}
else if (CsDiff >= 13) {
return 8;
}
}
}
else if (negBalance >= 55) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 3711) {
return 8;
}
else if (Cs >= 3711) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 331) {
if (posBalance < 46) {
if (gchDC < 2) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
if (Rs < 2163) {
return 4;
}
else if (Rs >= 2163) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 4;
}
}
}
else if (posBalance >= 46) {
if (SC < 6017) {
return 8;
}
else if (SC >= 6017) {
if (ssDCval < 138) {
if (SC < 8533) {
if (Cs < 3032) {
return 8;
}
else if (Cs >= 3032) {
if (mcTcor < 81) {
if (Rs < 3792) {
return 8;
}
else if (Rs >= 3792) {
return 2;
}
}
else if (mcTcor >= 81) {
return 4;
}
}
}
else if (SC >= 8533) {
return 8;
}
}
else if (ssDCval >= 138) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 81) {
if (Rs < 8566) {
if (MVDiff < 7) {
if (RsDiff < 15) {
if (refDCval < 87) {
if (ssDCval < 80) {
if (Rs < 4993) {
return 8;
}
else if (Rs >= 4993) {
return 4;
}
}
else if (ssDCval >= 80) {
if (Cs < 3294) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
return 4;
}
}
else if (Cs >= 3294) {
return 2;
}
}
}
else if (refDCval >= 87) {
if (TSC < 8) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 4;
}
}
else if (TSC >= 8) {
if (Cs < 2404) {
if (mu_mv_mag_sq < 60) {
if (Rs < 4965) {
return 8;
}
else if (Rs >= 4965) {
return 4;
}
}
else if (mu_mv_mag_sq >= 60) {
if (Rs < 2707) {
return 2;
}
else if (Rs >= 2707) {
return 4;
}
}
}
else if (Cs >= 2404) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (negBalance < 52) {
return 4;
}
else if (negBalance >= 52) {
if (Rs < 3748) {
return 4;
}
else if (Rs >= 3748) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (mu_mv_mag_sq < 20) {
if (Rs < 3880) {
return 8;
}
else if (Rs >= 3880) {
return 4;
}
}
else if (mu_mv_mag_sq >= 20) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 15) {
if (Cs < 2621) {
return 2;
}
else if (Cs >= 2621) {
if (negBalance < 43) {
return 2;
}
else if (negBalance >= 43) {
if (TSC < 13) {
if (RsDiff < 17) {
return 4;
}
else if (RsDiff >= 17) {
return 8;
}
}
else if (TSC >= 13) {
return 8;
}
}
}
}
}
else if (MVDiff >= 7) {
if (TSC < 14) {
if (negBalance < 45) {
if (RsCsDiff < 491) {
if (refDCval < 74) {
if (Rs < 3250) {
return 2;
}
else if (Rs >= 3250) {
return 4;
}
}
else if (refDCval >= 74) {
if (ssDCval < 130) {
if (SC < 7775) {
if (TSC < 12) {
if (mu_mv_mag_sq < 54) {
if (TSC < 10) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
if (MVDiff < 24) {
return 2;
}
else if (MVDiff >= 24) {
return 8;
}
}
}
else if (TSC >= 10) {
return 2;
}
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 3376) {
if (Rs < 3068) {
return 2;
}
else if (Rs >= 3068) {
return 4;
}
}
else if (Rs >= 3376) {
return 8;
}
}
}
else if (TSC >= 12) {
if (Cs < 2594) {
return 2;
}
else if (Cs >= 2594) {
if (Rs < 4302) {
return 8;
}
else if (Rs >= 4302) {
return 1;
}
}
}
}
else if (SC >= 7775) {
if (RsCsDiff < 311) {
return 4;
}
else if (RsCsDiff >= 311) {
return 2;
}
}
}
else if (ssDCval >= 130) {
if (RsDiff < 11) {
if (Rs < 4033) {
return 4;
}
else if (Rs >= 4033) {
return 1;
}
}
else if (RsDiff >= 11) {
return 8;
}
}
}
}
else if (RsCsDiff >= 491) {
return 8;
}
}
else if (negBalance >= 45) {
if (mu_mv_mag_sq < 41) {
if (RsCsDiff < 434) {
if (ssDCval < 101) {
if (refDCval < 79) {
if (Rs < 3275) {
return 8;
}
else if (Rs >= 3275) {
return 4;
}
}
else if (refDCval >= 79) {
if (mu_mv_mag_sq < 3) {
if (AFD < 16) {
if (Rs < 2726) {
return 4;
}
else if (Rs >= 2726) {
return 2;
}
}
else if (AFD >= 16) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 2512) {
return 4;
}
else if (Rs >= 2512) {
return 8;
}
}
}
}
else if (ssDCval >= 101) {
if (Rs < 3326) {
if (Rs < 2744) {
return 8;
}
else if (Rs >= 2744) {
return 4;
}
}
else if (Rs >= 3326) {
return 8;
}
}
}
else if (RsCsDiff >= 434) {
if (Rs < 6578) {
return 2;
}
else if (Rs >= 6578) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (tcor < 86) {
if (diffMVdiffVal < 1) {
if (posBalance < 59) {
if (SC < 8444) {
return 4;
}
else if (SC >= 8444) {
if (Rs < 6111) {
return 2;
}
else if (Rs >= 6111) {
return 4;
}
}
}
else if (posBalance >= 59) {
if (Cs < 2375) {
return 4;
}
else if (Cs >= 2375) {
return 2;
}
}
}
else if (diffMVdiffVal >= 1) {
if (posBalance < 56) {
if (diffRsCsdiff < 38) {
if (Rs < 3926) {
if (Rs < 2363) {
return 1;
}
else if (Rs >= 2363) {
return 2;
}
}
else if (Rs >= 3926) {
return 8;
}
}
else if (diffRsCsdiff >= 38) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
if (Rs < 3645) {
return 8;
}
else if (Rs >= 3645) {
return 2;
}
}
}
}
else if (posBalance >= 56) {
if (mu_mv_mag_sq < 52) {
return 8;
}
else if (mu_mv_mag_sq >= 52) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 3412) {
return 4;
}
else if (Rs >= 3412) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 86) {
if (diffRsCsdiff < 88) {
if (AFD < 16) {
if (RsDiff < 11) {
if (Rs < 3411) {
return 8;
}
else if (Rs >= 3411) {
return 4;
}
}
else if (RsDiff >= 11) {
return 8;
}
}
else if (AFD >= 16) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 8;
}
}
}
else if (diffRsCsdiff >= 88) {
if (RsDiff < 14) {
return 4;
}
else if (RsDiff >= 14) {
if (Rs < 3783) {
return 8;
}
else if (Rs >= 3783) {
return 1;
}
}
}
}
}
}
}
else if (TSC >= 14) {
if (Rs < 5273) {
if (diffRsCsdiff < -50) {
return 1;
}
else if (diffRsCsdiff >= -50) {
return 8;
}
}
else if (Rs >= 5273) {
return 1;
}
}
}
}
else if (Rs >= 8566) {
return 1;
}
}
}
}
else if (AFD >= 18) {
if (refDCval < 156) {
if (ssDCval < 68) {
if (MVDiff < 6) {
if (Rs < 2050) {
return 4;
}
else if (Rs >= 2050) {
return 1;
}
}
else if (MVDiff >= 6) {
if (RsDiff < 9) {
if (mcTcor < 77) {
return 4;
}
else if (mcTcor >= 77) {
if (diffRsCsdiff < 195) {
return 2;
}
else if (diffRsCsdiff >= 195) {
return 1;
}
}
}
else if (RsDiff >= 9) {
if (mu_mv_mag_sq < 14) {
if (AFD < 26) {
if (CsDiff < 14) {
if (Rs < 2813) {
if (diffRsCsdiff < -14) {
return 4;
}
else if (diffRsCsdiff >= -14) {
if (TSC < 12) {
if (Rs < 2051) {
return 4;
}
else if (Rs >= 2051) {
return 1;
}
}
else if (TSC >= 12) {
return 2;
}
}
}
else if (Rs >= 2813) {
return 2;
}
}
else if (CsDiff >= 14) {
if (ssDCval < 48) {
return 2;
}
else if (ssDCval >= 48) {
return 4;
}
}
}
else if (AFD >= 26) {
if (Cs < 2596) {
return 4;
}
else if (Cs >= 2596) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (diffAFD < 2) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (Cs < 3360) {
return 2;
}
else if (Cs >= 3360) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (CsDiff < 15) {
if (Cs < 2099) {
return 4;
}
else if (Cs >= 2099) {
return 8;
}
}
else if (CsDiff >= 15) {
if (tcor < 52) {
if (Rs < 3536) {
return 2;
}
else if (Rs >= 3536) {
return 8;
}
}
else if (tcor >= 52) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 68) {
if (mcTcor < 95) {
if (refDCval < 111) {
if (Rs < 6267) {
if (Cs < 3449) {
if (mcTcor < 90) {
if (negBalance < 57) {
if (mcTcor < 83) {
if (SC < 6474) {
if (mu_mv_mag_sq < 20) {
if (posBalance < 64) {
return 2;
}
else if (posBalance >= 64) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
if (Rs < 1571) {
if (Rs < 1476) {
return 2;
}
else if (Rs >= 1476) {
return 4;
}
}
else if (Rs >= 1571) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 20) {
return 8;
}
}
else if (SC >= 6474) {
if (diffRsCsdiff < 799) {
return 8;
}
else if (diffRsCsdiff >= 799) {
return 4;
}
}
}
else if (mcTcor >= 83) {
if (Rs < 3020) {
if (AFD < 23) {
if (MVDiff < 15) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
if (diffMVdiffVal < 2) {
if (TSC < 15) {
return 8;
}
else if (TSC >= 15) {
return 1;
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 2669) {
return 8;
}
else if (Rs >= 2669) {
if (Rs < 2965) {
return 4;
}
else if (Rs >= 2965) {
return 2;
}
}
}
}
}
else if (MVDiff >= 15) {
if (MVDiff < 20) {
if (Rs < 2759) {
return 1;
}
else if (Rs >= 2759) {
return 8;
}
}
else if (MVDiff >= 20) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (Rs < 1633) {
return 2;
}
else if (Rs >= 1633) {
if (Cs < 2024) {
if (Rs < 2610) {
return 8;
}
else if (Rs >= 2610) {
return 2;
}
}
else if (Cs >= 2024) {
return 8;
}
}
}
}
}
}
else if (AFD >= 23) {
if (SC < 4376) {
if (Rs < 1457) {
return 8;
}
else if (Rs >= 1457) {
return 2;
}
}
else if (SC >= 4376) {
if (Cs < 3023) {
return 1;
}
else if (Cs >= 3023) {
return 8;
}
}
}
}
else if (Rs >= 3020) {
if (TSC < 14) {
if (AFD < 19) {
if (CsDiff < 11) {
if (mu_mv_mag_sq < 29) {
return 4;
}
else if (mu_mv_mag_sq >= 29) {
if (Rs < 4156) {
return 2;
}
else if (Rs >= 4156) {
return 1;
}
}
}
else if (CsDiff >= 11) {
if (MVDiff < 13) {
if (Rs < 3401) {
return 8;
}
else if (Rs >= 3401) {
return 1;
}
}
else if (MVDiff >= 13) {
return 8;
}
}
}
else if (AFD >= 19) {
if (refDCval < 74) {
if (mu_mv_mag_sq < 50) {
return 2;
}
else if (mu_mv_mag_sq >= 50) {
return 4;
}
}
else if (refDCval >= 74) {
if (diffTSC < 1) {
if (refDCval < 77) {
if (Rs < 3991) {
if (Rs < 3344) {
return 8;
}
else if (Rs >= 3344) {
return 1;
}
}
else if (Rs >= 3991) {
return 4;
}
}
else if (refDCval >= 77) {
return 4;
}
}
else if (diffTSC >= 1) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
if (RsDiff < 12) {
if (Rs < 3186) {
return 4;
}
else if (Rs >= 3186) {
return 8;
}
}
else if (RsDiff >= 12) {
return 4;
}
}
}
}
}
}
else if (TSC >= 14) {
if (diffAFD < 1) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
if (Rs < 4088) {
return 8;
}
else if (Rs >= 4088) {
if (Rs < 4829) {
return 1;
}
else if (Rs >= 4829) {
return 4;
}
}
}
}
else if (diffAFD >= 1) {
if (tcor < 70) {
if (mcTcor < 87) {
return 4;
}
else if (mcTcor >= 87) {
if (Rs < 3663) {
return 8;
}
else if (Rs >= 3663) {
return 1;
}
}
}
else if (tcor >= 70) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 57) {
if (AFD < 23) {
if (mcTcor < 79) {
return 4;
}
else if (mcTcor >= 79) {
if (diffMVdiffVal < 5) {
if (refDCval < 74) {
if (RsDiff < 20) {
if (Rs < 3582) {
return 1;
}
else if (Rs >= 3582) {
return 8;
}
}
else if (RsDiff >= 20) {
return 2;
}
}
else if (refDCval >= 74) {
if (MVDiff < 13) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
else if (MVDiff >= 13) {
if (diffRsCsdiff < 16) {
if (ssDCval < 96) {
if (mcTcor < 88) {
return 4;
}
else if (mcTcor >= 88) {
return 8;
}
}
else if (ssDCval >= 96) {
return 8;
}
}
else if (diffRsCsdiff >= 16) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 5) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 1;
}
}
}
}
}
else if (AFD >= 23) {
if (Rs < 2098) {
if (RsDiff < 9) {
if (Rs < 917) {
return 2;
}
else if (Rs >= 917) {
if (Rs < 1440) {
return 4;
}
else if (Rs >= 1440) {
return 1;
}
}
}
else if (RsDiff >= 9) {
if (Rs < 1296) {
return 4;
}
else if (Rs >= 1296) {
if (Cs < 3197) {
return 8;
}
else if (Cs >= 3197) {
return 2;
}
}
}
}
else if (Rs >= 2098) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 31) {
if (Rs < 4320) {
if (Rs < 2744) {
return 1;
}
else if (Rs >= 2744) {
return 2;
}
}
else if (Rs >= 4320) {
return 4;
}
}
else if (mu_mv_mag_sq >= 31) {
return 8;
}
}
else if (posBalance >= 57) {
if (Cs < 2883) {
if (diffAFD < 2) {
if (RsDiff < 15) {
return 1;
}
else if (RsDiff >= 15) {
if (Rs < 3395) {
return 4;
}
else if (Rs >= 3395) {
return 1;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 4010) {
return 4;
}
else if (Rs >= 4010) {
return 2;
}
}
}
else if (Cs >= 2883) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 2) {
if (Rs < 4699) {
return 4;
}
else if (Rs >= 4699) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (AFD < 24) {
if (Rs < 2073) {
return 4;
}
else if (Rs >= 2073) {
return 8;
}
}
else if (AFD >= 24) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (negBalance < 52) {
if (Rs < 1676) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 4;
}
}
else if (Rs >= 1676) {
if (posBalance < 51) {
if (ssDCval < 73) {
if (Rs < 3951) {
return 2;
}
else if (Rs >= 3951) {
return 4;
}
}
else if (ssDCval >= 73) {
if (Rs < 2620) {
return 1;
}
else if (Rs >= 2620) {
return 2;
}
}
}
else if (posBalance >= 51) {
if (refDCval < 96) {
if (mu_mv_mag_sq < 2) {
if (Cs < 2205) {
return 4;
}
else if (Cs >= 2205) {
if (Rs < 3092) {
return 1;
}
else if (Rs >= 3092) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
else if (refDCval >= 96) {
if (RsCsDiff < 263) {
return 8;
}
else if (RsCsDiff >= 263) {
if (Rs < 2203) {
return 8;
}
else if (Rs >= 2203) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 52) {
if (mcTcor < 94) {
if (AFD < 27) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 13) {
if (negBalance < 55) {
if (Rs < 2612) {
return 8;
}
else if (Rs >= 2612) {
return 4;
}
}
else if (negBalance >= 55) {
if (ssDCval < 82) {
return 2;
}
else if (ssDCval >= 82) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
if (Rs < 2720) {
return 8;
}
else if (Rs >= 2720) {
return 4;
}
}
}
}
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (SC < 4095) {
if (SC < 3952) {
return 8;
}
else if (SC >= 3952) {
if (Rs < 1725) {
return 2;
}
else if (Rs >= 1725) {
return 1;
}
}
}
else if (SC >= 4095) {
if (tcor < 68) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
else if (tcor >= 68) {
return 8;
}
}
}
}
else if (AFD >= 27) {
if (Cs < 2571) {
return 4;
}
else if (Cs >= 2571) {
if (AFD < 33) {
return 1;
}
else if (AFD >= 33) {
if (Rs < 4389) {
return 2;
}
else if (Rs >= 4389) {
return 8;
}
}
}
}
}
else if (mcTcor >= 94) {
if (ssDCval < 82) {
if (Rs < 1670) {
return 1;
}
else if (Rs >= 1670) {
return 8;
}
}
else if (ssDCval >= 82) {
if (gchDC < 1) {
if (Rs < 2461) {
return 8;
}
else if (Rs >= 2461) {
return 4;
}
}
else if (gchDC >= 1) {
if (diffAFD < 12) {
return 2;
}
else if (diffAFD >= 12) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 3449) {
if (posBalance < 50) {
if (ssDCval < 83) {
if (CsDiff < 18) {
return 2;
}
else if (CsDiff >= 18) {
return 4;
}
}
else if (ssDCval >= 83) {
if (Cs < 5837) {
return 1;
}
else if (Cs >= 5837) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (Cs < 5263) {
if (negBalance < 58) {
if (ssDCval < 93) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (ssDCval >= 93) {
if (RsDiff < 18) {
if (mcTcor < 87) {
if (MVDiff < 7) {
if (Rs < 4178) {
return 4;
}
else if (Rs >= 4178) {
return 2;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (mcTcor >= 87) {
if (ssDCval < 110) {
if (Cs < 3855) {
return 8;
}
else if (Cs >= 3855) {
if (Cs < 4800) {
return 4;
}
else if (Cs >= 4800) {
if (TSC < 15) {
return 8;
}
else if (TSC >= 15) {
return 4;
}
}
}
}
else if (ssDCval >= 110) {
return 2;
}
}
}
else if (RsDiff >= 18) {
if (mcTcor < 84) {
return 2;
}
else if (mcTcor >= 84) {
if (Rs < 4759) {
return 2;
}
else if (Rs >= 4759) {
return 4;
}
}
}
}
}
else if (negBalance >= 58) {
if (SC < 6467) {
return 4;
}
else if (SC >= 6467) {
if (RsDiff < 20) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
return 4;
}
}
else if (RsDiff >= 20) {
return 2;
}
}
}
}
else if (Cs >= 5263) {
if (tcor < 69) {
return 2;
}
else if (tcor >= 69) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
}
}
}
else if (Rs >= 6267) {
if (negBalance < 46) {
if (SC < 10300) {
return 2;
}
else if (SC >= 10300) {
if (Cs < 4489) {
return 1;
}
else if (Cs >= 4489) {
return 8;
}
}
}
else if (negBalance >= 46) {
if (diffTSC < -2) {
if (Rs < 7027) {
if (Rs < 6598) {
return 8;
}
else if (Rs >= 6598) {
return 2;
}
}
else if (Rs >= 7027) {
return 4;
}
}
else if (diffTSC >= -2) {
return 4;
}
}
}
}
else if (refDCval >= 111) {
if (RsDiff < 24) {
if (SC < 5581) {
if (negBalance < 59) {
if (refDCval < 112) {
return 2;
}
else if (refDCval >= 112) {
if (mcTcor < 91) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (ssDCval < 139) {
if (TSC < 17) {
return 8;
}
else if (TSC >= 17) {
return 1;
}
}
else if (ssDCval >= 139) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
return 2;
}
}
}
}
else if (mcTcor >= 91) {
if (negBalance < 43) {
return 2;
}
else if (negBalance >= 43) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
}
}
else if (negBalance >= 59) {
if (posBalance < 61) {
if (mcTcor < 91) {
if (CsDiff < 10) {
return 1;
}
else if (CsDiff >= 10) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Rs < 1773) {
if (Rs < 1590) {
return 2;
}
else if (Rs >= 1590) {
return 8;
}
}
else if (Rs >= 1773) {
return 1;
}
}
}
}
else if (mcTcor >= 91) {
return 1;
}
}
else if (posBalance >= 61) {
if (CsDiff < 13) {
return 2;
}
else if (CsDiff >= 13) {
return 8;
}
}
}
}
else if (SC >= 5581) {
if (Cs < 5074) {
if (CsDiff < 15) {
if (AFD < 22) {
if (diffRsCsdiff < -62) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (diffRsCsdiff >= -62) {
if (diffRsCsdiff < 111) {
return 8;
}
else if (diffRsCsdiff >= 111) {
if (Cs < 2292) {
return 8;
}
else if (Cs >= 2292) {
return 4;
}
}
}
}
else if (AFD >= 22) {
if (RsCsDiff < 396) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 32) {
if (diffRsCsdiff < 28) {
return 1;
}
else if (diffRsCsdiff >= 28) {
if (Rs < 4960) {
return 8;
}
else if (Rs >= 4960) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 32) {
return 4;
}
}
}
else if (RsCsDiff >= 396) {
if (Cs < 2081) {
return 4;
}
else if (Cs >= 2081) {
if (Cs < 3616) {
return 8;
}
else if (Cs >= 3616) {
return 4;
}
}
}
}
}
else if (CsDiff >= 15) {
if (ssDCval < 122) {
if (SC < 6526) {
return 8;
}
else if (SC >= 6526) {
if (Cs < 3281) {
return 4;
}
else if (Cs >= 3281) {
return 2;
}
}
}
else if (ssDCval >= 122) {
if (MVDiff < 17) {
if (Cs < 3073) {
return 8;
}
else if (Cs >= 3073) {
if (posBalance < 64) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
return 4;
}
}
else if (posBalance >= 64) {
return 1;
}
}
}
else if (MVDiff >= 17) {
if (RsCsDiff < 784) {
if (Rs < 6330) {
return 8;
}
else if (Rs >= 6330) {
return 1;
}
}
else if (RsCsDiff >= 784) {
return 4;
}
}
}
}
}
else if (Cs >= 5074) {
return 8;
}
}
}
else if (RsDiff >= 24) {
if (AFD < 46) {
if (tcor < 61) {
if (RsDiff < 28) {
if (Cs < 2949) {
return 8;
}
else if (Cs >= 2949) {
if (Rs < 8912) {
return 4;
}
else if (Rs >= 8912) {
return 8;
}
}
}
else if (RsDiff >= 28) {
return 4;
}
}
else if (tcor >= 61) {
if (posBalance < 60) {
if (Cs < 3154) {
return 2;
}
else if (Cs >= 3154) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
if (TSC < 16) {
if (Rs < 6500) {
return 2;
}
else if (Rs >= 6500) {
return 8;
}
}
else if (TSC >= 16) {
return 4;
}
}
}
}
}
else if (posBalance >= 60) {
return 4;
}
}
}
else if (AFD >= 46) {
return 8;
}
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 142) {
if (Rs < 7534) {
if (mcTcor < 96) {
if (TSC < 10) {
if (diffRsCsdiff < 42) {
return 8;
}
else if (diffRsCsdiff >= 42) {
if (Rs < 3462) {
return 2;
}
else if (Rs >= 3462) {
return 1;
}
}
}
else if (TSC >= 10) {
if (negBalance < 61) {
if (negBalance < 47) {
if (ssDCval < 116) {
return 1;
}
else if (ssDCval >= 116) {
return 2;
}
}
else if (negBalance >= 47) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
return 8;
}
}
}
else if (negBalance >= 61) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (Cs < 2089) {
return 2;
}
else if (Cs >= 2089) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 12) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
return 8;
}
}
else if (TSC >= 12) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 7534) {
return 1;
}
}
else if (ssDCval >= 142) {
return 8;
}
}
}
}
else if (refDCval >= 156) {
if (negBalance < 43) {
if (Cs < 3082) {
if (Rs < 4215) {
if (Rs < 2123) {
return 1;
}
else if (Rs >= 2123) {
return 4;
}
}
else if (Rs >= 4215) {
return 2;
}
}
else if (Cs >= 3082) {
return 8;
}
}
else if (negBalance >= 43) {
if (ssDCval < 169) {
if (Rs < 3335) {
return 8;
}
else if (Rs >= 3335) {
return 4;
}
}
else if (ssDCval >= 169) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (SC < 7048) {
if (negBalance < 56) {
if (Rs < 4040) {
if (CsDiff < 16) {
if (posBalance < 64) {
if (SC < 5747) {
if (ssDCval < 65) {
if (Rs < 3360) {
return 4;
}
else if (Rs >= 3360) {
return 8;
}
}
else if (ssDCval >= 65) {
if (Rs < 3421) {
return 8;
}
else if (Rs >= 3421) {
return 4;
}
}
}
else if (SC >= 5747) {
if (refDCval < 82) {
if (negBalance < 50) {
if (SC < 6463) {
return 4;
}
else if (SC >= 6463) {
return 1;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (refDCval >= 82) {
if (ssDCval < 96) {
return 4;
}
else if (ssDCval >= 96) {
if (Cs < 2637) {
if (SC < 6076) {
return 8;
}
else if (SC >= 6076) {
return 1;
}
}
else if (Cs >= 2637) {
if (Cs < 3393) {
return 4;
}
else if (Cs >= 3393) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 64) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 1;
}
}
}
else if (CsDiff >= 16) {
return 1;
}
}
else if (Rs >= 4040) {
if (Rs < 4634) {
return 1;
}
else if (Rs >= 4634) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (Cs < 2586) {
if (negBalance < 60) {
return 8;
}
else if (negBalance >= 60) {
if (AFD < 34) {
return 4;
}
else if (AFD >= 34) {
return 1;
}
}
}
else if (Cs >= 2586) {
return 4;
}
}
}
else if (SC >= 7048) {
if (negBalance < 52) {
if (Rs < 5431) {
if (MVDiff < 8) {
if (SC < 7559) {
if (posBalance < 57) {
if (CsDiff < 10) {
if (Cs < 2150) {
return 2;
}
else if (Cs >= 2150) {
return 8;
}
}
else if (CsDiff >= 10) {
return 4;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (SC >= 7559) {
return 8;
}
}
else if (MVDiff >= 8) {
if (ssDCval < 75) {
if (diffTSC < -2) {
return 8;
}
else if (diffTSC >= -2) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
}
else if (ssDCval >= 75) {
if (AFD < 17) {
if (diffAFD < -4) {
if (Rs < 4961) {
return 4;
}
else if (Rs >= 4961) {
return 8;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (AFD >= 17) {
if (Cs < 2673) {
return 4;
}
else if (Cs >= 2673) {
return 2;
}
}
}
}
}
else if (Rs >= 5431) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
if (AFD < 14) {
if (ssDCval < 92) {
return 4;
}
else if (ssDCval >= 92) {
return 8;
}
}
else if (AFD >= 14) {
if (Rs < 6221) {
if (Rs < 5981) {
return 4;
}
else if (Rs >= 5981) {
return 2;
}
}
else if (Rs >= 6221) {
return 4;
}
}
}
}
}
else if (negBalance >= 52) {
if (posBalance < 57) {
if (Cs < 5877) {
if (tcor < 82) {
return 1;
}
else if (tcor >= 82) {
if (Cs < 3825) {
return 4;
}
else if (Cs >= 3825) {
return 8;
}
}
}
else if (Cs >= 5877) {
return 4;
}
}
else if (posBalance >= 57) {
if (RsDiff < 25) {
if (Rs < 3465) {
if (Cs < 4240) {
return 8;
}
else if (Cs >= 4240) {
return 4;
}
}
else if (Rs >= 3465) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
}
else if (RsDiff >= 25) {
return 4;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1652) {
if (Rs < 3098) {
if (Rs < 1281) {
if (Rs < 1078) {
if (AFD < 1) {
if (Rs < 599) {
return 4;
}
else if (Rs >= 599) {
return 2;
}
}
else if (AFD >= 1) {
if (refDCval < 63) {
if (SC < 2385) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
return 1;
}
}
else if (SC >= 2385) {
if (mu_mv_mag_sq < 112) {
if (Cs < 1525) {
return 1;
}
else if (Cs >= 1525) {
return 4;
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
}
else if (refDCval >= 63) {
if (refDCval < 77) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (Rs < 875) {
return 8;
}
else if (Rs >= 875) {
return 2;
}
}
}
else if (refDCval >= 77) {
if (ssDCval < 116) {
if (diffRsCsdiff < -6) {
if (Rs < 949) {
if (Rs < 851) {
return 1;
}
else if (Rs >= 851) {
return 4;
}
}
else if (Rs >= 949) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
return 1;
}
}
else if (ssDCval >= 116) {
if (refDCval < 129) {
if (Rs < 901) {
return 8;
}
else if (Rs >= 901) {
return 4;
}
}
else if (refDCval >= 129) {
if (Rs < 577) {
return 2;
}
else if (Rs >= 577) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1078) {
if (posBalance < 39) {
if (posBalance < 10) {
return 8;
}
else if (posBalance >= 10) {
return 1;
}
}
else if (posBalance >= 39) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
if (negBalance < 31) {
if (Cs < 1517) {
return 8;
}
else if (Cs >= 1517) {
return 4;
}
}
else if (negBalance >= 31) {
if (ssDCval < 64) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
return 8;
}
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1281) {
if (ssDCval < 61) {
if (TSC < 2) {
if (ssDCval < 52) {
if (mu_mv_mag_sq < 118) {
return 1;
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
else if (ssDCval >= 52) {
if (Cs < 1422) {
return 8;
}
else if (Cs >= 1422) {
return 1;
}
}
}
else if (TSC >= 2) {
if (mcTcor < 98) {
if (Rs < 2208) {
if (Rs < 1328) {
return 4;
}
else if (Rs >= 1328) {
return 8;
}
}
else if (Rs >= 2208) {
if (Rs < 2542) {
return 4;
}
else if (Rs >= 2542) {
if (Rs < 2862) {
return 8;
}
else if (Rs >= 2862) {
return 8;
}
}
}
}
else if (mcTcor >= 98) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
return 4;
}
}
}
}
else if (ssDCval >= 61) {
if (ssDCval < 140) {
if (ssDCval < 86) {
if (tcor < 98) {
if (Rs < 2724) {
if (SC < 2732) {
return 4;
}
else if (SC >= 2732) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (diffRsCsdiff < -7) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (diffRsCsdiff >= -7) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 8;
}
}
}
}
}
else if (Rs >= 2724) {
if (Cs < 1400) {
return 2;
}
else if (Cs >= 1400) {
if (Rs < 2828) {
return 4;
}
else if (Rs >= 2828) {
return 8;
}
}
}
}
else if (tcor >= 98) {
if (negBalance < 17) {
if (SC < 2841) {
return 1;
}
else if (SC >= 2841) {
if (Cs < 1350) {
if (Rs < 2342) {
if (Rs < 2302) {
if (SC < 3638) {
if (Cs < 1328) {
if (Rs < 2247) {
return 8;
}
else if (Rs >= 2247) {
return 2;
}
}
else if (Cs >= 1328) {
return 8;
}
}
else if (SC >= 3638) {
return 2;
}
}
else if (Rs >= 2302) {
return 8;
}
}
else if (Rs >= 2342) {
return 2;
}
}
else if (Cs >= 1350) {
return 8;
}
}
}
else if (negBalance >= 17) {
if (ssDCval < 85) {
if (Rs < 1516) {
if (Cs < 1458) {
return 4;
}
else if (Cs >= 1458) {
return 8;
}
}
else if (Rs >= 1516) {
if (Cs < 1617) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 8;
}
}
else if (Cs >= 1617) {
if (Cs < 1628) {
if (Rs < 2053) {
return 4;
}
else if (Rs >= 2053) {
return 8;
}
}
else if (Cs >= 1628) {
return 4;
}
}
}
}
else if (ssDCval >= 85) {
if (Rs < 2482) {
return 8;
}
else if (Rs >= 2482) {
return 4;
}
}
}
}
}
else if (ssDCval >= 86) {
if (SC < 3426) {
if (RsDiff < 1) {
if (ssDCval < 126) {
if (MVDiff < 31) {
return 4;
}
else if (MVDiff >= 31) {
if (Rs < 1885) {
return 8;
}
else if (Rs >= 1885) {
return 4;
}
}
}
else if (ssDCval >= 126) {
return 8;
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 126) {
if (negBalance < 5) {
if (Rs < 1296) {
return 8;
}
else if (Rs >= 1296) {
return 1;
}
}
else if (negBalance >= 5) {
if (diffRsCsdiff < -19) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (diffRsCsdiff >= -19) {
if (Rs < 1995) {
return 8;
}
else if (Rs >= 1995) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (ssDCval < 113) {
return 8;
}
else if (ssDCval >= 113) {
return 1;
}
}
}
}
}
else if (SC >= 3426) {
if (refDCval < 131) {
if (Cs < 1649) {
if (SC < 3514) {
if (SC < 3509) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 1957) {
return 8;
}
else if (Rs >= 1957) {
return 2;
}
}
}
else if (SC >= 3509) {
if (Rs < 1913) {
return 8;
}
else if (Rs >= 1913) {
return 1;
}
}
}
else if (SC >= 3514) {
return 8;
}
}
else if (Cs >= 1649) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
else if (refDCval >= 131) {
if (Cs < 1490) {
if (Rs < 2924) {
if (Rs < 2700) {
return 8;
}
else if (Rs >= 2700) {
return 1;
}
}
else if (Rs >= 2924) {
return 2;
}
}
else if (Cs >= 1490) {
return 8;
}
}
}
}
}
else if (ssDCval >= 140) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3098) {
if (refDCval < 77) {
if (mu_mv_mag_sq < 126) {
if (SC < 4910) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
else if (SC >= 4910) {
if (ssDCval < 55) {
if (Rs < 4042) {
return 4;
}
else if (Rs >= 4042) {
return 8;
}
}
else if (ssDCval >= 55) {
if (Cs < 1384) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (Cs >= 1384) {
if (ssDCval < 74) {
if (posBalance < 27) {
if (Rs < 4404) {
return 1;
}
else if (Rs >= 4404) {
return 4;
}
}
else if (posBalance >= 27) {
if (Cs < 1396) {
if (Rs < 3757) {
return 1;
}
else if (Rs >= 3757) {
return 4;
}
}
else if (Cs >= 1396) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (Rs < 4265) {
return 2;
}
else if (Rs >= 4265) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1577) {
if (AFD < 3) {
if (SC < 4987) {
if (mu_mv_mag_sq < 127) {
if (Rs < 3544) {
return 1;
}
else if (Rs >= 3544) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
else if (SC >= 4987) {
return 8;
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (Cs >= 1577) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 4;
}
}
}
}
else if (refDCval >= 77) {
if (ssDCval < 103) {
if (ssDCval < 84) {
if (Rs < 4121) {
return 2;
}
else if (Rs >= 4121) {
return 8;
}
}
else if (ssDCval >= 84) {
if (Rs < 4121) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (Rs < 3552) {
return 2;
}
else if (Rs >= 3552) {
return 4;
}
}
}
else if (Rs >= 4121) {
if (refDCval < 97) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 2;
}
}
else if (refDCval >= 97) {
return 4;
}
}
}
}
else if (ssDCval >= 103) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
if (Rs < 3382) {
return 2;
}
else if (Rs >= 3382) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1652) {
if (ssDCval < 102) {
if (SC < 4332) {
if (ssDCval < 67) {
if (negBalance < 25) {
if (Cs < 1790) {
if (Rs < 1655) {
if (mu_mv_mag_sq < 115) {
if (Rs < 984) {
return 8;
}
else if (Rs >= 984) {
return 4;
}
}
else if (mu_mv_mag_sq >= 115) {
if (SC < 3345) {
if (negBalance < 19) {
if (Cs < 1696) {
return 4;
}
else if (Cs >= 1696) {
return 8;
}
}
else if (negBalance >= 19) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (SC >= 3345) {
return 4;
}
}
}
else if (Rs >= 1655) {
if (ssDCval < 66) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 2473) {
return 8;
}
else if (Rs >= 2473) {
return 1;
}
}
}
else if (ssDCval >= 66) {
return 1;
}
}
}
else if (Cs >= 1790) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (diffAFD < -1) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
return 8;
}
}
else if (diffAFD >= -1) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 128) {
if (mu_mv_mag_sq < 117) {
return 8;
}
else if (mu_mv_mag_sq >= 117) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 2142) {
return 8;
}
else if (Rs >= 2142) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (Rs < 2203) {
return 8;
}
else if (Rs >= 2203) {
return 1;
}
}
}
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
}
else if (negBalance >= 25) {
if (ssDCval < 63) {
if (mu_mv_mag_sq < 119) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
else if (ssDCval >= 63) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (RsDiff < 2) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
return 1;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
}
else if (ssDCval >= 67) {
if (SC < 2757) {
if (Rs < 622) {
return 8;
}
else if (Rs >= 622) {
return 1;
}
}
else if (SC >= 2757) {
if (ssDCval < 90) {
if (Rs < 2612) {
if (diffMVdiffVal < -6) {
return 4;
}
else if (diffMVdiffVal >= -6) {
if (negBalance < 60) {
if (Rs < 2557) {
return 8;
}
else if (Rs >= 2557) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (negBalance >= 60) {
return 4;
}
}
}
else if (Rs >= 2612) {
return 4;
}
}
else if (ssDCval >= 90) {
if (Cs < 1861) {
return 8;
}
else if (Cs >= 1861) {
if (Cs < 2006) {
return 4;
}
else if (Cs >= 2006) {
if (mu_mv_mag_sq < 124) {
if (Rs < 2051) {
return 4;
}
else if (Rs >= 2051) {
return 2;
}
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 1388) {
return 2;
}
else if (Rs >= 1388) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 4332) {
if (Cs < 2197) {
if (Rs < 3554) {
if (negBalance < 16) {
if (Rs < 2397) {
if (ssDCval < 92) {
if (Rs < 2366) {
return 1;
}
else if (Rs >= 2366) {
if (Cs < 2139) {
return 8;
}
else if (Cs >= 2139) {
if (Rs < 2385) {
return 4;
}
else if (Rs >= 2385) {
return 2;
}
}
}
}
else if (ssDCval >= 92) {
if (mu_mv_mag_sq < 119) {
if (negBalance < 8) {
return 2;
}
else if (negBalance >= 8) {
if (Rs < 2357) {
if (Rs < 2326) {
return 2;
}
else if (Rs >= 2326) {
return 4;
}
}
else if (Rs >= 2357) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 119) {
return 4;
}
}
}
else if (Rs >= 2397) {
if (Cs < 1682) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (Cs >= 1682) {
if (Cs < 2188) {
if (diffRsCsdiff < -127) {
if (diffAFD < -6) {
return 8;
}
else if (diffAFD >= -6) {
return 4;
}
}
else if (diffRsCsdiff >= -127) {
return 8;
}
}
else if (Cs >= 2188) {
return 4;
}
}
}
}
else if (negBalance >= 16) {
if (Cs < 1885) {
if (Rs < 2582) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (Rs >= 2582) {
if (tcor < 97) {
if (Cs < 1696) {
return 2;
}
else if (Cs >= 1696) {
if (ssDCval < 74) {
if (Rs < 3445) {
return 4;
}
else if (Rs >= 3445) {
return 8;
}
}
else if (ssDCval >= 74) {
return 8;
}
}
}
else if (tcor >= 97) {
if (ssDCval < 80) {
if (Rs < 2933) {
return 8;
}
else if (Rs >= 2933) {
if (Cs < 1881) {
if (ssDCval < 65) {
if (tcor < 98) {
if (Rs < 2991) {
return 4;
}
else if (Rs >= 2991) {
return 4;
}
}
else if (tcor >= 98) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Rs < 2950) {
return 8;
}
else if (Rs >= 2950) {
if (tcor < 99) {
if (SC < 4886) {
return 4;
}
else if (SC >= 4886) {
if (Rs < 3118) {
return 8;
}
else if (Rs >= 3118) {
return 4;
}
}
}
else if (tcor >= 99) {
if (Rs < 3039) {
return 8;
}
else if (Rs >= 3039) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 65) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
return 4;
}
}
}
else if (Cs >= 1881) {
return 4;
}
}
}
else if (ssDCval >= 80) {
if (MVDiff < 1) {
if (Cs < 1827) {
return 4;
}
else if (Cs >= 1827) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
return 8;
}
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
else if (Cs >= 1885) {
if (Rs < 3018) {
if (ssDCval < 65) {
if (Rs < 2853) {
return 8;
}
else if (Rs >= 2853) {
return 4;
}
}
else if (ssDCval >= 65) {
if (MVDiff < 4) {
if (Cs < 2028) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
if (Cs < 2026) {
if (ssDCval < 82) {
if (ssDCval < 80) {
if (ssDCval < 66) {
if (Rs < 2989) {
if (SC < 4868) {
if (Rs < 2932) {
return 8;
}
else if (Rs >= 2932) {
return 4;
}
}
else if (SC >= 4868) {
return 8;
}
}
else if (Rs >= 2989) {
return 4;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (ssDCval >= 80) {
return 4;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
else if (Cs >= 2026) {
return 4;
}
}
}
else if (Cs >= 2028) {
return 8;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 8;
}
}
}
}
else if (Rs >= 3018) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3554) {
if (posBalance < 41) {
if (ssDCval < 69) {
if (Rs < 7016) {
if (posBalance < 19) {
return 8;
}
else if (posBalance >= 19) {
if (posBalance < 27) {
if (Rs < 4352) {
return 4;
}
else if (Rs >= 4352) {
if (Cs < 2103) {
return 8;
}
else if (Cs >= 2103) {
if (Rs < 4370) {
return 4;
}
else if (Rs >= 4370) {
return 4;
}
}
}
}
else if (posBalance >= 27) {
if (RsCsDiff < 22) {
if (Rs < 4226) {
return 8;
}
else if (Rs >= 4226) {
return 8;
}
}
else if (RsCsDiff >= 22) {
if (MVDiff < 4) {
if (Cs < 1976) {
return 4;
}
else if (Cs >= 1976) {
if (mcTcor < 96) {
return 4;
}
else if (mcTcor >= 96) {
if (Cs < 2183) {
if (RsDiff < 7) {
if (diffRsCsdiff < -12) {
return 8;
}
else if (diffRsCsdiff >= -12) {
if (Rs < 4588) {
if (negBalance < 13) {
return 8;
}
else if (negBalance >= 13) {
return 4;
}
}
else if (Rs >= 4588) {
return 4;
}
}
}
else if (RsDiff >= 7) {
return 8;
}
}
else if (Cs >= 2183) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
}
else if (Rs >= 7016) {
return 2;
}
}
else if (ssDCval >= 69) {
if (Rs < 3838) {
if (posBalance < 20) {
return 8;
}
else if (posBalance >= 20) {
return 2;
}
}
else if (Rs >= 3838) {
if (negBalance < 28) {
if (diffTSC < -3) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 2;
}
}
else if (diffTSC >= -3) {
return 4;
}
}
else if (negBalance >= 28) {
if (mu_mv_mag_sq < 118) {
if (posBalance < 35) {
return 2;
}
else if (posBalance >= 35) {
return 8;
}
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 5251) {
return 2;
}
else if (Rs >= 5251) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 41) {
if (RsDiff < 6) {
if (SC < 6488) {
if (diffRsCsdiff < 1) {
if (MVDiff < 1) {
if (Cs < 1938) {
if (posBalance < 42) {
if (Rs < 4348) {
return 8;
}
else if (Rs >= 4348) {
return 2;
}
}
else if (posBalance >= 42) {
return 8;
}
}
else if (Cs >= 1938) {
if (Rs < 4402) {
return 4;
}
else if (Rs >= 4402) {
return 2;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 1934) {
return 2;
}
else if (Cs >= 1934) {
return 8;
}
}
}
else if (diffRsCsdiff >= 1) {
if (Cs < 1949) {
if (Rs < 4413) {
return 4;
}
else if (Rs >= 4413) {
return 2;
}
}
else if (Cs >= 1949) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
}
}
else if (SC >= 6488) {
return 8;
}
}
else if (RsDiff >= 6) {
if (Cs < 2153) {
if (Rs < 5832) {
return 2;
}
else if (Rs >= 5832) {
return 4;
}
}
else if (Cs >= 2153) {
return 1;
}
}
}
}
}
else if (Cs >= 2197) {
if (SC < 8344) {
if (Cs < 2765) {
if (mu_mv_mag_sq < 122) {
if (ssDCval < 101) {
if (negBalance < 52) {
if (SC < 6257) {
if (posBalance < 64) {
if (Cs < 2595) {
if (Rs < 2226) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
else if (Rs >= 2226) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
if (ssDCval < 99) {
return 8;
}
else if (ssDCval >= 99) {
return 4;
}
}
}
}
}
else if (Cs >= 2595) {
if (ssDCval < 87) {
if (posBalance < 30) {
return 1;
}
else if (posBalance >= 30) {
return 8;
}
}
else if (ssDCval >= 87) {
if (diffRsCsdiff < 3) {
return 4;
}
else if (diffRsCsdiff >= 3) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
return 4;
}
}
}
}
}
else if (posBalance >= 64) {
if (Rs < 3163) {
return 4;
}
else if (Rs >= 3163) {
return 8;
}
}
}
else if (SC >= 6257) {
if (CsDiff < 3) {
if (Cs < 2288) {
return 4;
}
else if (Cs >= 2288) {
return 8;
}
}
else if (CsDiff >= 3) {
if (diffRsCsdiff < -38) {
return 2;
}
else if (diffRsCsdiff >= -38) {
if (posBalance < 57) {
if (Rs < 5098) {
if (ssDCval < 63) {
if (SC < 7056) {
if (posBalance < 32) {
if (Cs < 2326) {
return 8;
}
else if (Cs >= 2326) {
if (Rs < 4512) {
return 4;
}
else if (Rs >= 4512) {
return 4;
}
}
}
else if (posBalance >= 32) {
if (tcor < 94) {
if (Rs < 4442) {
return 8;
}
else if (Rs >= 4442) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Rs < 4544) {
return 4;
}
else if (Rs >= 4544) {
return 4;
}
}
}
}
else if (tcor >= 94) {
return 4;
}
}
}
else if (SC >= 7056) {
if (SC < 7222) {
if (Cs < 2529) {
return 8;
}
else if (Cs >= 2529) {
return 8;
}
}
else if (SC >= 7222) {
if (RsDiff < 5) {
if (Cs < 2577) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
if (Cs < 2557) {
if (Rs < 4826) {
return 4;
}
else if (Rs >= 4826) {
return 4;
}
}
else if (Cs >= 2557) {
return 4;
}
}
}
else if (Cs >= 2577) {
if (Cs < 2707) {
if (Rs < 4729) {
if (Rs < 4677) {
return 8;
}
else if (Rs >= 4677) {
return 8;
}
}
else if (Rs >= 4729) {
if (SC < 7417) {
return 4;
}
else if (SC >= 7417) {
return 4;
}
}
}
else if (Cs >= 2707) {
return 8;
}
}
}
else if (RsDiff >= 5) {
if (negBalance < 23) {
return 8;
}
else if (negBalance >= 23) {
return 4;
}
}
}
}
}
else if (ssDCval >= 63) {
return 8;
}
}
else if (Rs >= 5098) {
return 4;
}
}
else if (posBalance >= 57) {
return 4;
}
}
}
}
}
else if (negBalance >= 52) {
return 4;
}
}
else if (ssDCval >= 101) {
if (Rs < 3931) {
return 8;
}
else if (Rs >= 3931) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (ssDCval < 47) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (ssDCval >= 47) {
if (Rs < 2145) {
if (negBalance < 32) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (negBalance >= 32) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (Rs >= 2145) {
if (AFD < 8) {
if (Cs < 2251) {
if (Rs < 4759) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 4036) {
return 2;
}
else if (Rs >= 4036) {
return 8;
}
}
}
}
else if (Rs >= 4759) {
return 4;
}
}
else if (Cs >= 2251) {
if (ssDCval < 52) {
if (Rs < 3942) {
return 4;
}
else if (Rs >= 3942) {
return 8;
}
}
else if (ssDCval >= 52) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (diffMVdiffVal < -4) {
if (Rs < 2943) {
return 4;
}
else if (Rs >= 2943) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
if (Rs < 4822) {
return 8;
}
else if (Rs >= 4822) {
if (Rs < 5094) {
return 4;
}
else if (Rs >= 5094) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 8) {
if (negBalance < 49) {
if (Rs < 3074) {
return 8;
}
else if (Rs >= 3074) {
return 4;
}
}
else if (negBalance >= 49) {
return 4;
}
}
}
}
}
}
else if (Cs >= 2765) {
if (ssDCval < 85) {
if (negBalance < 4) {
if (Cs < 2922) {
if (Rs < 3843) {
return 1;
}
else if (Rs >= 3843) {
return 8;
}
}
else if (Cs >= 2922) {
return 4;
}
}
else if (negBalance >= 4) {
if (ssDCval < 48) {
if (SC < 6018) {
return 1;
}
else if (SC >= 6018) {
return 8;
}
}
else if (ssDCval >= 48) {
if (SC < 5504) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (SC >= 5504) {
if (negBalance < 49) {
if (posBalance < 29) {
if (Rs < 3602) {
return 8;
}
else if (Rs >= 3602) {
if (ssDCval < 68) {
if (Rs < 4553) {
return 4;
}
else if (Rs >= 4553) {
return 4;
}
}
else if (ssDCval >= 68) {
return 8;
}
}
}
else if (posBalance >= 29) {
return 8;
}
}
else if (negBalance >= 49) {
if (Rs < 3331) {
if (posBalance < 60) {
if (Rs < 2871) {
return 4;
}
else if (Rs >= 2871) {
return 8;
}
}
else if (posBalance >= 60) {
return 4;
}
}
else if (Rs >= 3331) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (ssDCval < 98) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Cs < 3071) {
if (AFD < 6) {
if (Rs < 4047) {
return 8;
}
else if (Rs >= 4047) {
return 4;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (Cs >= 3071) {
if (Rs < 3328) {
if (mcTcor < 91) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
else if (mcTcor >= 91) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 4;
}
}
}
else if (Rs >= 3328) {
if (RsCsDiff < 183) {
if (tcor < 98) {
if (Cs < 3491) {
return 4;
}
else if (Cs >= 3491) {
if (Cs < 4139) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
return 8;
}
}
else if (Cs >= 4139) {
return 4;
}
}
}
else if (tcor >= 98) {
if (ssDCval < 97) {
if (Rs < 3709) {
return 4;
}
else if (Rs >= 3709) {
return 8;
}
}
else if (ssDCval >= 97) {
if (Rs < 4131) {
return 8;
}
else if (Rs >= 4131) {
return 4;
}
}
}
}
else if (RsCsDiff >= 183) {
return 8;
}
}
}
}
}
else if (ssDCval >= 98) {
if (refDCval < 101) {
return 8;
}
else if (refDCval >= 101) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Rs < 3879) {
return 4;
}
else if (Rs >= 3879) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 8344) {
if (Rs < 6660) {
if (ssDCval < 79) {
if (Rs < 5995) {
if (Rs < 4552) {
if (negBalance < 29) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (negBalance >= 29) {
return 8;
}
}
else if (Rs >= 4552) {
return 8;
}
}
else if (Rs >= 5995) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (ssDCval >= 79) {
if (Cs < 3639) {
if (Rs < 5191) {
return 4;
}
else if (Rs >= 5191) {
return 8;
}
}
else if (Cs >= 3639) {
return 8;
}
}
}
else if (Rs >= 6660) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (ssDCval < 137) {
if (Rs < 1822) {
if (ssDCval < 115) {
if (negBalance < 47) {
if (posBalance < 54) {
if (Rs < 1416) {
return 8;
}
else if (Rs >= 1416) {
return 2;
}
}
else if (posBalance >= 54) {
return 2;
}
}
else if (negBalance >= 47) {
if (Rs < 1477) {
return 2;
}
else if (Rs >= 1477) {
return 8;
}
}
}
else if (ssDCval >= 115) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (ssDCval < 131) {
if (Rs < 1492) {
return 4;
}
else if (Rs >= 1492) {
return 8;
}
}
else if (ssDCval >= 131) {
return 8;
}
}
}
}
else if (Rs >= 1822) {
if (refDCval < 105) {
if (negBalance < 42) {
if (negBalance < 36) {
if (Cs < 1749) {
return 4;
}
else if (Cs >= 1749) {
return 8;
}
}
else if (negBalance >= 36) {
if (RsCsDiff < 29) {
if (Rs < 3741) {
return 8;
}
else if (Rs >= 3741) {
return 2;
}
}
else if (RsCsDiff >= 29) {
return 8;
}
}
}
else if (negBalance >= 42) {
if (Cs < 2865) {
if (Rs < 4934) {
if (Rs < 3556) {
return 8;
}
else if (Rs >= 3556) {
return 1;
}
}
else if (Rs >= 4934) {
return 2;
}
}
else if (Cs >= 2865) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (diffAFD < -2) {
if (diffRsCsdiff < -999) {
return 1;
}
else if (diffRsCsdiff >= -999) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (ssDCval < 120) {
if (Cs < 1913) {
if (Rs < 2444) {
return 8;
}
else if (Rs >= 2444) {
if (Rs < 2945) {
return 1;
}
else if (Rs >= 2945) {
return 4;
}
}
}
else if (Cs >= 1913) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (Rs < 5824) {
return 2;
}
else if (Rs >= 5824) {
return 8;
}
}
}
}
}
else if (ssDCval >= 120) {
if (Rs < 3297) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (Rs < 2599) {
return 8;
}
else if (Rs >= 2599) {
return 1;
}
}
}
else if (Rs >= 3297) {
if (posBalance < 50) {
if (AFD < 3) {
if (Rs < 4364) {
return 1;
}
else if (Rs >= 4364) {
return 2;
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (posBalance >= 50) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
}
}
}
else if (diffAFD >= -2) {
if (MVDiff < 1) {
if (ssDCval < 129) {
if (posBalance < 62) {
if (diffRsCsdiff < 19) {
if (Rs < 1910) {
if (AFD < 3) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
else if (AFD >= 3) {
return 8;
}
}
else if (Rs >= 1910) {
if (diffMVdiffVal < -1) {
if (negBalance < 56) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (Rs < 5256) {
return 2;
}
else if (Rs >= 5256) {
return 8;
}
}
}
else if (diffMVdiffVal >= -1) {
if (Cs < 1679) {
if (Rs < 4124) {
return 8;
}
else if (Rs >= 4124) {
return 2;
}
}
else if (Cs >= 1679) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 19) {
if (RsCsDiff < 63) {
if (ssDCval < 108) {
if (Rs < 2821) {
return 8;
}
else if (Rs >= 2821) {
return 4;
}
}
else if (ssDCval >= 108) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 4;
}
}
}
else if (RsCsDiff >= 63) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (Cs < 1789) {
if (Rs < 2409) {
return 8;
}
else if (Rs >= 2409) {
return 1;
}
}
else if (Cs >= 1789) {
return 8;
}
}
}
else if (ssDCval >= 129) {
if (Cs < 3869) {
if (Rs < 4710) {
return 8;
}
else if (Rs >= 4710) {
return 2;
}
}
else if (Cs >= 3869) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (posBalance < 24) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
if (Rs < 2769) {
if (posBalance < 13) {
if (Rs < 2373) {
return 1;
}
else if (Rs >= 2373) {
return 8;
}
}
else if (posBalance >= 13) {
if (Cs < 2933) {
if (MVDiff < 8) {
if (refDCval < 117) {
return 8;
}
else if (refDCval >= 117) {
return 4;
}
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (Cs >= 2933) {
return 8;
}
}
}
else if (Rs >= 2769) {
return 2;
}
}
}
else if (posBalance >= 24) {
if (AFD < 27) {
if (negBalance < 53) {
if (Cs < 1663) {
if (Rs < 3905) {
return 8;
}
else if (Rs >= 3905) {
return 2;
}
}
else if (Cs >= 1663) {
if (Rs < 6409) {
return 8;
}
else if (Rs >= 6409) {
if (Rs < 6544) {
return 4;
}
else if (Rs >= 6544) {
return 8;
}
}
}
}
else if (negBalance >= 53) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
return 2;
}
}
}
else if (AFD >= 27) {
if (Rs < 3475) {
return 1;
}
else if (Rs >= 3475) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 137) {
if (mcTcor < 96) {
if (negBalance < 38) {
if (ssDCval < 200) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
else if (ssDCval >= 200) {
return 4;
}
}
else if (negBalance >= 38) {
if (Rs < 7738) {
if (Rs < 2450) {
return 8;
}
else if (Rs >= 2450) {
return 1;
}
}
else if (Rs >= 7738) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (ssDCval < 201) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Cs < 3336) {
if (Cs < 2767) {
if (Rs < 4775) {
return 8;
}
else if (Rs >= 4775) {
return 2;
}
}
else if (Cs >= 2767) {
return 1;
}
}
else if (Cs >= 3336) {
return 8;
}
}
}
else if (ssDCval >= 201) {
if (Cs < 4080) {
return 8;
}
else if (Cs >= 4080) {
return 4;
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelectB(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1325) {
if (Rs < 2371) {
if (Cs < 527) {
if (negBalance < 14) {
if (mu_mv_mag_sq < 100) {
if (tcor < 94) {
if (negBalance < 5) {
if (posBalance < 93) {
if (refDCval < 50) {
if (mu_mv_mag_sq < 42) {
if (AFD < 3) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 2;
}
}
else if (AFD >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 42) {
if (posBalance < 17) {
return 4;
}
else if (posBalance >= 17) {
if (tcor < 57) {
if (MVDiff < 64) {
return 4;
}
else if (MVDiff >= 64) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (tcor >= 57) {
return 8;
}
}
}
}
else if (refDCval >= 50) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (posBalance >= 93) {
if (mcTcor < 67) {
if (MVDiff < 56) {
if (Rs < 219) {
return 2;
}
else if (Rs >= 219) {
return 4;
}
}
else if (MVDiff >= 56) {
if (Rs < 1) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (Rs >= 1) {
return 1;
}
}
}
else if (mcTcor >= 67) {
if (Cs < 3) {
if (posBalance < 96) {
return 4;
}
else if (posBalance >= 96) {
if (Rs < 997) {
return 2;
}
else if (Rs >= 997) {
return 8;
}
}
}
else if (Cs >= 3) {
if (mu_mv_mag_sq < 4) {
if (Cs < 88) {
return 8;
}
else if (Cs >= 88) {
return 2;
}
}
else if (mu_mv_mag_sq >= 4) {
if (AFD < 8) {
if (ssDCval < 65) {
if (Rs < 386) {
return 8;
}
else if (Rs >= 386) {
return 4;
}
}
else if (ssDCval >= 65) {
return 4;
}
}
else if (AFD >= 8) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 5) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (MVDiff < 3) {
if (Cs < 40) {
return 1;
}
else if (Cs >= 40) {
if (Rs < 259) {
return 4;
}
else if (Rs >= 259) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 145) {
if (posBalance < 66) {
if (SC < 122) {
return 1;
}
else if (SC >= 122) {
if (RsCsDiff < 30) {
if (RsDiff < 1) {
if (MVDiff < 27) {
if (refDCval < 29) {
return 1;
}
else if (refDCval >= 29) {
if (AFD < 3) {
if (mu_mv_mag_sq < 79) {
return 8;
}
else if (mu_mv_mag_sq >= 79) {
return 4;
}
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (MVDiff >= 27) {
return 4;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (RsCsDiff >= 30) {
return 8;
}
}
}
else if (posBalance >= 66) {
if (MVDiff < 14) {
if (Cs < 139) {
return 8;
}
else if (Cs >= 139) {
if (ssDCval < 49) {
if (posBalance < 77) {
return 8;
}
else if (posBalance >= 77) {
return 2;
}
}
else if (ssDCval >= 49) {
return 8;
}
}
}
else if (MVDiff >= 14) {
if (gchDC < 11) {
if (diffMVdiffVal < -6) {
if (posBalance < 111) {
return 2;
}
else if (posBalance >= 111) {
return 4;
}
}
else if (diffMVdiffVal >= -6) {
if (Cs < 100) {
if (refDCval < 69) {
return 2;
}
else if (refDCval >= 69) {
return 4;
}
}
else if (Cs >= 100) {
if (TSC < 6) {
if (mu_mv_mag_sq < 22) {
return 1;
}
else if (mu_mv_mag_sq >= 22) {
if (tcor < 83) {
return 1;
}
else if (tcor >= 83) {
return 8;
}
}
}
else if (TSC >= 6) {
if (ssDCval < 42) {
return 2;
}
else if (ssDCval >= 42) {
return 8;
}
}
}
}
}
else if (gchDC >= 11) {
if (Cs < 413) {
return 1;
}
else if (Cs >= 413) {
return 2;
}
}
}
}
}
else if (ssDCval >= 145) {
return 2;
}
}
}
}
}
else if (tcor >= 94) {
if (SC < 874) {
if (Cs < 162) {
if (MVDiff < 13) {
if (refDCval < 28) {
if (posBalance < 34) {
if (diffMVdiffVal < 5) {
return 4;
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
else if (posBalance >= 34) {
return 1;
}
}
else if (refDCval >= 28) {
if (posBalance < 50) {
if (posBalance < 33) {
if (Rs < 81) {
if (Rs < 61) {
return 4;
}
else if (Rs >= 61) {
return 1;
}
}
else if (Rs >= 81) {
return 8;
}
}
else if (posBalance >= 33) {
if (Cs < 82) {
return 1;
}
else if (Cs >= 82) {
return 4;
}
}
}
else if (posBalance >= 50) {
return 8;
}
}
}
else if (MVDiff >= 13) {
if (ssDCval < 22) {
if (Rs < 16) {
return 8;
}
else if (Rs >= 16) {
return 1;
}
}
else if (ssDCval >= 22) {
if (ssDCval < 26) {
return 4;
}
else if (ssDCval >= 26) {
if (ssDCval < 191) {
if (Rs < 167) {
if (posBalance < 61) {
return 1;
}
else if (posBalance >= 61) {
if (mu_mv_mag_sq < 14) {
if (Rs < 30) {
return 1;
}
else if (Rs >= 30) {
return 4;
}
}
else if (mu_mv_mag_sq >= 14) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 26) {
return 1;
}
else if (Rs >= 26) {
return 4;
}
}
}
}
}
else if (Rs >= 167) {
if (diffAFD < -2) {
if (Rs < 537) {
return 8;
}
else if (Rs >= 537) {
return 4;
}
}
else if (diffAFD >= -2) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (ssDCval >= 191) {
if (MVDiff < 238) {
if (Rs < 60) {
return 2;
}
else if (Rs >= 60) {
return 1;
}
}
else if (MVDiff >= 238) {
return 1;
}
}
}
}
}
}
else if (Cs >= 162) {
if (refDCval < 37) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (refDCval >= 37) {
if (TSCindex < 2) {
if (ssDCval < 82) {
if (Rs < 328) {
if (MVDiff < 5) {
if (Rs < 174) {
return 4;
}
else if (Rs >= 174) {
return 8;
}
}
else if (MVDiff >= 5) {
if (diffMVdiffVal < 4) {
return 1;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 117) {
return 1;
}
else if (Rs >= 117) {
return 2;
}
}
}
}
else if (Rs >= 328) {
return 4;
}
}
else if (ssDCval >= 82) {
if (Rs < 453) {
return 4;
}
else if (Rs >= 453) {
if (Cs < 329) {
return 4;
}
else if (Cs >= 329) {
return 1;
}
}
}
}
else if (TSCindex >= 2) {
if (Rs < 177) {
return 4;
}
else if (Rs >= 177) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Cs < 232) {
return 8;
}
else if (Cs >= 232) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 874) {
if (TSC < 2) {
if (SC < 953) {
if (negBalance < 11) {
return 1;
}
else if (negBalance >= 11) {
if (Rs < 460) {
return 4;
}
else if (Rs >= 460) {
return 8;
}
}
}
else if (SC >= 953) {
if (ssDCval < 101) {
if (Rs < 1159) {
return 8;
}
else if (Rs >= 1159) {
return 1;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
}
else if (TSC >= 2) {
if (ssDCval < 124) {
if (Cs < 30) {
return 8;
}
else if (Cs >= 30) {
return 1;
}
}
else if (ssDCval >= 124) {
if (Cs < 392) {
return 1;
}
else if (Cs >= 392) {
if (Rs < 1352) {
return 8;
}
else if (Rs >= 1352) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (SC < 782) {
if (refDCval < 48) {
if (SC < 279) {
if (negBalance < 8) {
if (Cs < 4) {
if (refDCval < 22) {
if (MVDiff < 1076) {
if (ssDCval < 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (refDCval >= 22) {
if (ssDCval < 38) {
return 4;
}
else if (ssDCval >= 38) {
return 2;
}
}
}
else if (Cs >= 4) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (negBalance < 3) {
return 8;
}
else if (negBalance >= 3) {
if (Rs < 113) {
return 1;
}
else if (Rs >= 113) {
return 4;
}
}
}
}
}
else if (negBalance >= 8) {
if (refDCval < 40) {
if (Rs < 86) {
if (ssDCval < 24) {
if (Rs < 63) {
return 8;
}
else if (Rs >= 63) {
return 1;
}
}
else if (ssDCval >= 24) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (Rs >= 86) {
if (posBalance < 33) {
if (ssDCval < 36) {
if (tcor < 94) {
if (Rs < 114) {
return 4;
}
else if (Rs >= 114) {
return 1;
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (ssDCval >= 36) {
return 1;
}
}
else if (posBalance >= 33) {
return 1;
}
}
}
else if (refDCval >= 40) {
if (SC < 195) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (SC >= 195) {
return 1;
}
}
}
}
else if (SC >= 279) {
if (Rs < 259) {
if (ssDCval < 45) {
if (negBalance < 3) {
return 8;
}
else if (negBalance >= 3) {
if (diffRsCsdiff < -3) {
return 8;
}
else if (diffRsCsdiff >= -3) {
if (tcor < 91) {
if (posBalance < 28) {
return 4;
}
else if (posBalance >= 28) {
if (Cs < 305) {
return 1;
}
else if (Cs >= 305) {
return 4;
}
}
}
else if (tcor >= 91) {
if (Cs < 263) {
if (AFD < 1) {
if (Rs < 128) {
return 8;
}
else if (Rs >= 128) {
return 4;
}
}
else if (AFD >= 1) {
if (Cs < 237) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Cs >= 237) {
return 8;
}
}
}
else if (Cs >= 263) {
if (tcor < 92) {
if (Rs < 237) {
return 4;
}
else if (Rs >= 237) {
return 1;
}
}
else if (tcor >= 92) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 45) {
return 8;
}
}
else if (Rs >= 259) {
if (posBalance < 12) {
if (tcor < 63) {
return 4;
}
else if (tcor >= 63) {
if (negBalance < 2) {
return 8;
}
else if (negBalance >= 2) {
if (Cs < 280) {
return 8;
}
else if (Cs >= 280) {
return 4;
}
}
}
}
else if (posBalance >= 12) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
if (posBalance < 26) {
if (Cs < 226) {
return 8;
}
else if (Cs >= 226) {
if (Rs < 419) {
return 1;
}
else if (Rs >= 419) {
return 4;
}
}
}
else if (posBalance >= 26) {
if (mu_mv_mag_sq < 109) {
if (Cs < 275) {
return 1;
}
else if (Cs >= 275) {
return 4;
}
}
else if (mu_mv_mag_sq >= 109) {
if (Cs < 402) {
return 4;
}
else if (Cs >= 402) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 48) {
if (ssDCval < 84) {
if (Cs < 23) {
return 2;
}
else if (Cs >= 23) {
if (ssDCval < 67) {
if (Cs < 255) {
if (MVDiff < 3) {
if (negBalance < 6) {
if (Rs < 203) {
return 8;
}
else if (Rs >= 203) {
return 4;
}
}
else if (negBalance >= 6) {
return 8;
}
}
else if (MVDiff >= 3) {
if (diffTSC < -1) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 104) {
return 8;
}
else if (Rs >= 104) {
return 4;
}
}
}
else if (diffTSC >= -1) {
if (Rs < 160) {
return 4;
}
else if (Rs >= 160) {
return 1;
}
}
}
}
else if (Cs >= 255) {
if (Rs < 401) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 4;
}
}
else if (Rs >= 401) {
return 8;
}
}
}
else if (ssDCval >= 67) {
if (Cs < 119) {
if (Rs < 145) {
return 4;
}
else if (Rs >= 145) {
return 8;
}
}
else if (Cs >= 119) {
if (Cs < 241) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
if (Rs < 385) {
return 8;
}
else if (Rs >= 385) {
return 2;
}
}
}
else if (Cs >= 241) {
if (ssDCval < 72) {
if (Rs < 255) {
return 2;
}
else if (Rs >= 255) {
return 4;
}
}
else if (ssDCval >= 72) {
if (Rs < 323) {
return 1;
}
else if (Rs >= 323) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (Rs < 181) {
return 4;
}
else if (Rs >= 181) {
if (Cs < 156) {
return 8;
}
else if (Cs >= 156) {
if (Cs < 355) {
if (ssDCval < 130) {
if (Rs < 450) {
return 2;
}
else if (Rs >= 450) {
return 1;
}
}
else if (ssDCval >= 130) {
if (negBalance < 9) {
return 4;
}
else if (negBalance >= 9) {
if (Rs < 407) {
return 8;
}
else if (Rs >= 407) {
return 1;
}
}
}
}
else if (Cs >= 355) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 782) {
if (diffTSC < -3) {
if (posBalance < 3) {
return 4;
}
else if (posBalance >= 3) {
if (diffAFD < -8) {
if (ssDCval < 147) {
if (ssDCval < 101) {
if (Rs < 707) {
return 1;
}
else if (Rs >= 707) {
return 8;
}
}
else if (ssDCval >= 101) {
return 2;
}
}
else if (ssDCval >= 147) {
return 8;
}
}
else if (diffAFD >= -8) {
if (Cs < 426) {
if (negBalance < 3) {
if (diffAFD < -6) {
return 2;
}
else if (diffAFD >= -6) {
return 4;
}
}
else if (negBalance >= 3) {
if (SC < 823) {
return 8;
}
else if (SC >= 823) {
return 2;
}
}
}
else if (Cs >= 426) {
if (Rs < 1385) {
return 1;
}
else if (Rs >= 1385) {
return 4;
}
}
}
}
}
else if (diffTSC >= -3) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 39) {
if (RsCsDiff < 2) {
if (posBalance < 21) {
return 8;
}
else if (posBalance >= 21) {
if (Rs < 497) {
return 1;
}
else if (Rs >= 497) {
return 4;
}
}
}
else if (RsCsDiff >= 2) {
if (Rs < 654) {
return 1;
}
else if (Rs >= 654) {
return 2;
}
}
}
else if (ssDCval >= 39) {
if (SC < 962) {
if (ssDCval < 122) {
if (Rs < 396) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
return 8;
}
}
else if (Rs >= 396) {
if (SC < 945) {
return 8;
}
else if (SC >= 945) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Rs < 528) {
return 8;
}
else if (Rs >= 528) {
return 1;
}
}
}
}
}
else if (ssDCval >= 122) {
if (posBalance < 15) {
return 8;
}
else if (posBalance >= 15) {
return 1;
}
}
}
else if (SC >= 962) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Cs < 488) {
if (Rs < 901) {
return 8;
}
else if (Rs >= 901) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
if (Rs < 1160) {
return 1;
}
else if (Rs >= 1160) {
return 8;
}
}
}
}
else if (Cs >= 488) {
return 1;
}
}
else if (ssDCval >= 60) {
if (Cs < 446) {
return 1;
}
else if (Cs >= 446) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
if (Rs < 1462) {
return 8;
}
else if (Rs >= 1462) {
return 1;
}
}
}
}
}
else if (ssDCval >= 67) {
if (SC < 1045) {
if (Rs < 497) {
return 1;
}
else if (Rs >= 497) {
if (SC < 1014) {
return 8;
}
else if (SC >= 1014) {
return 4;
}
}
}
else if (SC >= 1045) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
if (Rs < 811) {
return 8;
}
else if (Rs >= 811) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 137) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (ssDCval < 84) {
if (Cs < 356) {
if (Cs < 246) {
return 2;
}
else if (Cs >= 246) {
return 1;
}
}
else if (Cs >= 356) {
if (Rs < 1182) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 733) {
return 4;
}
else if (Rs >= 733) {
return 8;
}
}
}
else if (Rs >= 1182) {
return 2;
}
}
}
else if (ssDCval >= 84) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (SC < 1090) {
if (Rs < 575) {
return 8;
}
else if (Rs >= 575) {
return 4;
}
}
else if (SC >= 1090) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 35) {
if (ssDCval < 20) {
if (MVDiff < 594) {
if (diffMVdiffVal < -79) {
return 4;
}
else if (diffMVdiffVal >= -79) {
return 8;
}
}
else if (MVDiff >= 594) {
if (Rs < 1) {
return 4;
}
else if (Rs >= 1) {
return 1;
}
}
}
else if (ssDCval >= 20) {
if (posBalance < 26) {
if (MVDiff < 5) {
if (tcor < 102) {
if (Rs < 27) {
return 1;
}
else if (Rs >= 27) {
return 4;
}
}
else if (tcor >= 102) {
if (Cs < 1) {
return 4;
}
else if (Cs >= 1) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (negBalance < 56) {
if (AFD < 2) {
if (posBalance < 19) {
if (RsDiff < 1) {
if (posBalance < 16) {
return 4;
}
else if (posBalance >= 16) {
if (Rs < 39) {
return 1;
}
else if (Rs >= 39) {
return 4;
}
}
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (posBalance >= 19) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 1;
}
}
}
else if (AFD >= 2) {
if (mu_mv_mag_sq < 77) {
return 1;
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 345) {
return 1;
}
else if (Rs >= 345) {
return 2;
}
}
}
}
else if (negBalance >= 56) {
if (tcor < 73) {
return 2;
}
else if (tcor >= 73) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 74) {
if (diffMVdiffVal < -30) {
return 4;
}
else if (diffMVdiffVal >= -30) {
return 8;
}
}
else if (mu_mv_mag_sq >= 74) {
return 4;
}
}
else if (diffAFD >= 2) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 26) {
if (tcor < 60) {
if (Cs < 386) {
if (negBalance < 45) {
if (Cs < 111) {
if (posBalance < 85) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 85) {
return 8;
}
}
else if (Cs >= 111) {
if (mu_mv_mag_sq < 5) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
if (diffRsCsdiff < -9) {
return 8;
}
else if (diffRsCsdiff >= -9) {
if (Rs < 237) {
if (SC < 364) {
if (Cs < 114) {
return 8;
}
else if (Cs >= 114) {
return 1;
}
}
else if (SC >= 364) {
return 8;
}
}
else if (Rs >= 237) {
return 1;
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 363) {
if (diffAFD < 2) {
if (diffRsCsdiff < -3) {
if (Rs < 286) {
return 1;
}
else if (Rs >= 286) {
return 8;
}
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (Rs >= 363) {
return 1;
}
}
}
else if (Cs >= 386) {
if (Rs < 472) {
return 2;
}
else if (Rs >= 472) {
return 8;
}
}
}
else if (tcor >= 60) {
if (refDCval < 30) {
if (Cs < 359) {
if (gchDC < 1) {
if (Cs < 301) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (mcTcor < 86) {
if (Cs < 27) {
if (Rs < 172) {
return 1;
}
else if (Rs >= 172) {
return 4;
}
}
else if (Cs >= 27) {
return 1;
}
}
else if (mcTcor >= 86) {
return 1;
}
}
}
else if (Cs >= 301) {
if (MVDiff < 19) {
if (ssDCval < 23) {
return 4;
}
else if (ssDCval >= 23) {
return 1;
}
}
else if (MVDiff >= 19) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (posBalance < 43) {
if (mu_mv_mag_sq < 42) {
if (Rs < 85) {
if (Cs < 48) {
return 1;
}
else if (Cs >= 48) {
return 2;
}
}
else if (Rs >= 85) {
if (Cs < 33) {
return 4;
}
else if (Cs >= 33) {
if (SC < 397) {
return 1;
}
else if (SC >= 397) {
if (ssDCval < 24) {
return 1;
}
else if (ssDCval >= 24) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (ssDCval < 30) {
return 1;
}
else if (ssDCval >= 30) {
if (Rs < 208) {
return 4;
}
else if (Rs >= 208) {
return 1;
}
}
}
}
else if (posBalance >= 43) {
if (posBalance < 71) {
if (Cs < 87) {
return 1;
}
else if (Cs >= 87) {
if (negBalance < 30) {
if (negBalance < 26) {
if (Cs < 181) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
return 1;
}
}
else if (Cs >= 181) {
return 1;
}
}
else if (negBalance >= 26) {
return 2;
}
}
else if (negBalance >= 30) {
return 1;
}
}
}
else if (posBalance >= 71) {
return 2;
}
}
}
}
else if (Cs >= 359) {
if (mu_mv_mag_sq < 38) {
return 8;
}
else if (mu_mv_mag_sq >= 38) {
if (mu_mv_mag_sq < 60) {
return 4;
}
else if (mu_mv_mag_sq >= 60) {
if (Cs < 502) {
return 1;
}
else if (Cs >= 502) {
return 4;
}
}
}
}
}
else if (refDCval >= 30) {
if (posBalance < 41) {
if (mu_mv_mag_sq < 13) {
if (AFD < 3) {
if (Rs < 81) {
if (Rs < 74) {
return 1;
}
else if (Rs >= 74) {
return 4;
}
}
else if (Rs >= 81) {
return 8;
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (mu_mv_mag_sq >= 13) {
if (mcTcor < 98) {
if (AFD < 4) {
if (Cs < 232) {
if (Cs < 92) {
return 1;
}
else if (Cs >= 92) {
if (Cs < 189) {
if (mu_mv_mag_sq < 99) {
if (mu_mv_mag_sq < 45) {
return 1;
}
else if (mu_mv_mag_sq >= 45) {
return 8;
}
}
else if (mu_mv_mag_sq >= 99) {
if (Rs < 238) {
return 1;
}
else if (Rs >= 238) {
return 8;
}
}
}
else if (Cs >= 189) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 4;
}
}
}
}
else if (Cs >= 232) {
if (ssDCval < 32) {
if (refDCval < 32) {
return 1;
}
else if (refDCval >= 32) {
return 8;
}
}
else if (ssDCval >= 32) {
if (Cs < 305) {
return 4;
}
else if (Cs >= 305) {
return 1;
}
}
}
}
else if (AFD >= 4) {
if (posBalance < 30) {
if (Cs < 110) {
if (Cs < 49) {
return 1;
}
else if (Cs >= 49) {
return 2;
}
}
else if (Cs >= 110) {
if (Cs < 140) {
return 4;
}
else if (Cs >= 140) {
return 8;
}
}
}
else if (posBalance >= 30) {
if (Cs < 253) {
if (AFD < 6) {
if (mcTcor < 97) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
if (Rs < 434) {
return 1;
}
else if (Rs >= 434) {
return 4;
}
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 253) {
if (Rs < 296) {
return 8;
}
else if (Rs >= 296) {
return 1;
}
}
}
}
}
else if (mcTcor >= 98) {
if (Rs < 487) {
if (refDCval < 34) {
return 1;
}
else if (refDCval >= 34) {
return 2;
}
}
else if (Rs >= 487) {
if (Rs < 842) {
return 4;
}
else if (Rs >= 842) {
return 1;
}
}
}
}
}
else if (posBalance >= 41) {
if (Rs < 383) {
if (mcTcor < 90) {
if (ssDCval < 33) {
if (MVDiff < 4) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
return 8;
}
}
else if (MVDiff >= 4) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 1;
}
}
}
else if (ssDCval >= 33) {
if (Cs < 200) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (Rs < 133) {
return 4;
}
else if (Rs >= 133) {
return 8;
}
}
}
else if (Cs >= 200) {
if (Cs < 318) {
return 8;
}
else if (Cs >= 318) {
return 1;
}
}
}
}
else if (mcTcor >= 90) {
if (posBalance < 62) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
return 1;
}
}
}
else if (posBalance >= 62) {
if (AFD < 4) {
if (Rs < 71) {
return 2;
}
else if (Rs >= 71) {
return 8;
}
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (Rs >= 383) {
if (ssDCval < 29) {
return 8;
}
else if (ssDCval >= 29) {
if (Cs < 482) {
return 1;
}
else if (Cs >= 482) {
if (Rs < 1155) {
return 8;
}
else if (Rs >= 1155) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 35) {
if (Rs < 307) {
if (ssDCval < 50) {
if (mcTcor < 92) {
if (tcor < 67) {
if (posBalance < 56) {
if (negBalance < 50) {
if (tcor < 65) {
return 1;
}
else if (tcor >= 65) {
if (Rs < 210) {
return 1;
}
else if (Rs >= 210) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (MVDiff < 129) {
if (posBalance < 43) {
if (Cs < 22) {
return 2;
}
else if (Cs >= 22) {
if (Cs < 65) {
return 4;
}
else if (Cs >= 65) {
return 1;
}
}
}
else if (posBalance >= 43) {
if (refDCval < 43) {
return 8;
}
else if (refDCval >= 43) {
if (Cs < 291) {
return 4;
}
else if (Cs >= 291) {
return 8;
}
}
}
}
else if (MVDiff >= 129) {
if (TSC < 9) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
else if (TSC >= 9) {
if (gchDC < 15) {
return 8;
}
else if (gchDC >= 15) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 56) {
if (negBalance < 35) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Rs < 20) {
return 8;
}
else if (Rs >= 20) {
return 1;
}
}
}
else if (negBalance >= 35) {
if (negBalance < 50) {
if (ssDCval < 38) {
if (Rs < 219) {
return 8;
}
else if (Rs >= 219) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (ssDCval >= 38) {
if (Rs < 30) {
if (Rs < 9) {
return 8;
}
else if (Rs >= 9) {
return 2;
}
}
else if (Rs >= 30) {
if (negBalance < 37) {
return 8;
}
else if (negBalance >= 37) {
return 4;
}
}
}
}
else if (negBalance >= 50) {
if (MVDiff < 44) {
return 8;
}
else if (MVDiff >= 44) {
if (tcor < 20) {
if (Rs < 106) {
return 2;
}
else if (Rs >= 106) {
return 1;
}
}
else if (tcor >= 20) {
if (MVDiff < 72) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (MVDiff >= 72) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 4;
}
}
}
}
}
}
}
}
else if (tcor >= 67) {
if (negBalance < 31) {
if (posBalance < 62) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
if (Rs < 98) {
return 4;
}
else if (Rs >= 98) {
if (Rs < 195) {
return 1;
}
else if (Rs >= 195) {
return 8;
}
}
}
}
else if (posBalance >= 62) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (Rs < 196) {
return 2;
}
else if (Rs >= 196) {
return 8;
}
}
}
}
else if (negBalance >= 31) {
if (refDCval < 42) {
if (Cs < 300) {
if (AFD < 5) {
if (Cs < 49) {
return 2;
}
else if (Cs >= 49) {
return 8;
}
}
else if (AFD >= 5) {
if (TSC < 4) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
if (Rs < 225) {
return 2;
}
else if (Rs >= 225) {
if (Rs < 246) {
return 8;
}
else if (Rs >= 246) {
return 1;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 184) {
if (negBalance < 57) {
if (Cs < 179) {
return 1;
}
else if (Cs >= 179) {
return 2;
}
}
else if (negBalance >= 57) {
if (Cs < 114) {
return 8;
}
else if (Cs >= 114) {
return 2;
}
}
}
else if (Cs >= 184) {
return 8;
}
}
}
}
else if (Cs >= 300) {
if (AFD < 7) {
if (Rs < 177) {
return 1;
}
else if (Rs >= 177) {
return 2;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (refDCval >= 42) {
if (MVDiff < 20) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 271) {
return 1;
}
else if (Rs >= 271) {
return 4;
}
}
}
}
else if (MVDiff >= 20) {
if (refDCval < 47) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (refDCval >= 47) {
if (mcTcor < 87) {
if (Rs < 92) {
return 8;
}
else if (Rs >= 92) {
return 4;
}
}
else if (mcTcor >= 87) {
if (Cs < 221) {
return 2;
}
else if (Cs >= 221) {
return 1;
}
}
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 57) {
if (negBalance < 77) {
if (mcTcor < 97) {
if (Rs < 291) {
if (SC < 115) {
if (Rs < 36) {
if (Cs < 43) {
if (Rs < 20) {
return 8;
}
else if (Rs >= 20) {
return 1;
}
}
else if (Cs >= 43) {
return 4;
}
}
else if (Rs >= 36) {
return 2;
}
}
else if (SC >= 115) {
if (Cs < 110) {
if (posBalance < 43) {
if (Rs < 115) {
if (Rs < 91) {
return 4;
}
else if (Rs >= 91) {
return 1;
}
}
else if (Rs >= 115) {
return 2;
}
}
else if (posBalance >= 43) {
return 1;
}
}
else if (Cs >= 110) {
if (refDCval < 36) {
if (SC < 338) {
if (Rs < 107) {
return 1;
}
else if (Rs >= 107) {
return 4;
}
}
else if (SC >= 338) {
return 8;
}
}
else if (refDCval >= 36) {
if (mu_mv_mag_sq < 4) {
if (Rs < 267) {
if (Rs < 132) {
if (Rs < 102) {
return 4;
}
else if (Rs >= 102) {
return 1;
}
}
else if (Rs >= 132) {
if (ssDCval < 44) {
return 2;
}
else if (ssDCval >= 44) {
if (Rs < 153) {
return 2;
}
else if (Rs >= 153) {
return 1;
}
}
}
}
else if (Rs >= 267) {
return 4;
}
}
else if (mu_mv_mag_sq >= 4) {
if (negBalance < 35) {
if (Rs < 203) {
return 1;
}
else if (Rs >= 203) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (negBalance >= 35) {
if (mu_mv_mag_sq < 32) {
if (negBalance < 45) {
if (AFD < 9) {
if (Rs < 122) {
return 4;
}
else if (Rs >= 122) {
return 2;
}
}
else if (AFD >= 9) {
if (Cs < 233) {
return 4;
}
else if (Cs >= 233) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (mcTcor < 95) {
if (Rs < 258) {
return 2;
}
else if (Rs >= 258) {
if (Rs < 274) {
return 8;
}
else if (Rs >= 274) {
return 4;
}
}
}
else if (mcTcor >= 95) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 291) {
if (Cs < 326) {
if (MVDiff < 24) {
return 4;
}
else if (MVDiff >= 24) {
return 2;
}
}
else if (Cs >= 326) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
if (CsDiff < 2) {
if (ssDCval < 42) {
if (Rs < 103) {
if (negBalance < 30) {
if (Cs < 6) {
return 2;
}
else if (Cs >= 6) {
return 1;
}
}
else if (negBalance >= 30) {
return 4;
}
}
else if (Rs >= 103) {
if (refDCval < 42) {
return 1;
}
else if (refDCval >= 42) {
return 2;
}
}
}
else if (ssDCval >= 42) {
if (diffMVdiffVal < 28) {
return 1;
}
else if (diffMVdiffVal >= 28) {
return 2;
}
}
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
else if (negBalance >= 77) {
if (Cs < 53) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 2;
}
}
else if (Cs >= 53) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (SC < 508) {
if (mu_mv_mag_sq < 99) {
if (negBalance < 33) {
if (ssDCval < 47) {
if (refDCval < 39) {
if (negBalance < 19) {
if (Rs < 128) {
return 1;
}
else if (Rs >= 128) {
if (Rs < 196) {
return 4;
}
else if (Rs >= 196) {
return 1;
}
}
}
else if (negBalance >= 19) {
if (RsDiff < 1) {
if (Rs < 114) {
return 1;
}
else if (Rs >= 114) {
return 8;
}
}
else if (RsDiff >= 1) {
if (Rs < 194) {
return 1;
}
else if (Rs >= 194) {
if (Rs < 243) {
return 8;
}
else if (Rs >= 243) {
return 4;
}
}
}
}
}
else if (refDCval >= 39) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 125) {
return 8;
}
else if (Rs >= 125) {
return 1;
}
}
}
}
else if (ssDCval >= 47) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
if (Rs < 13) {
return 2;
}
else if (Rs >= 13) {
return 8;
}
}
}
}
else if (negBalance >= 33) {
if (mcTcor < 104) {
if (ssDCval < 35) {
if (Rs < 60) {
return 8;
}
else if (Rs >= 60) {
return 1;
}
}
else if (ssDCval >= 35) {
if (diffMVdiffVal < 56) {
if (refDCval < 41) {
return 8;
}
else if (refDCval >= 41) {
if (posBalance < 25) {
return 8;
}
else if (posBalance >= 25) {
if (Rs < 255) {
return 1;
}
else if (Rs >= 255) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 56) {
return 4;
}
}
}
else if (mcTcor >= 104) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (negBalance < 53) {
if (posBalance < 37) {
if (ssDCval < 43) {
if (ssDCval < 39) {
if (Cs < 228) {
if (Rs < 164) {
return 4;
}
else if (Rs >= 164) {
if (Rs < 183) {
return 1;
}
else if (Rs >= 183) {
return 4;
}
}
}
else if (Cs >= 228) {
if (Rs < 202) {
return 8;
}
else if (Rs >= 202) {
return 4;
}
}
}
else if (ssDCval >= 39) {
return 1;
}
}
else if (ssDCval >= 43) {
if (Rs < 180) {
return 8;
}
else if (Rs >= 180) {
return 4;
}
}
}
else if (posBalance >= 37) {
if (RsCsDiff < 2) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
return 1;
}
}
else if (RsCsDiff >= 2) {
return 1;
}
}
}
else if (negBalance >= 53) {
return 4;
}
}
}
else if (SC >= 508) {
if (negBalance < 40) {
if (mu_mv_mag_sq < 77) {
if (Rs < 229) {
if (Rs < 220) {
return 4;
}
else if (Rs >= 220) {
return 8;
}
}
else if (Rs >= 229) {
return 1;
}
}
else if (mu_mv_mag_sq >= 77) {
return 1;
}
}
else if (negBalance >= 40) {
return 8;
}
}
}
}
}
else if (ssDCval >= 50) {
if (refDCval < 70) {
if (TSCindex < 2) {
if (refDCval < 63) {
if (posBalance < 30) {
if (Rs < 22) {
return 2;
}
else if (Rs >= 22) {
return 8;
}
}
else if (posBalance >= 30) {
if (MVDiff < 7) {
if (TSC < 1) {
if (diffMVdiffVal < -26) {
return 8;
}
else if (diffMVdiffVal >= -26) {
return 4;
}
}
else if (TSC >= 1) {
if (tcor < 99) {
if (ssDCval < 58) {
if (Cs < 182) {
if (Rs < 138) {
return 4;
}
else if (Rs >= 138) {
return 1;
}
}
else if (Cs >= 182) {
return 4;
}
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (tcor >= 99) {
if (Rs < 165) {
if (refDCval < 58) {
if (MVDiff < 3) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (negBalance < 17) {
if (Rs < 147) {
return 4;
}
else if (Rs >= 147) {
return 1;
}
}
else if (negBalance >= 17) {
return 1;
}
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (refDCval >= 58) {
return 1;
}
}
else if (Rs >= 165) {
if (Rs < 245) {
return 4;
}
else if (Rs >= 245) {
return 1;
}
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 242) {
return 1;
}
else if (Rs >= 242) {
return 2;
}
}
}
}
else if (refDCval >= 63) {
if (SC < 415) {
if (mcTcor < 97) {
return 1;
}
else if (mcTcor >= 97) {
if (SC < 379) {
return 2;
}
else if (SC >= 379) {
if (SC < 410) {
return 4;
}
else if (SC >= 410) {
return 2;
}
}
}
}
else if (SC >= 415) {
if (Cs < 344) {
return 4;
}
else if (Cs >= 344) {
if (mu_mv_mag_sq < 124) {
if (Rs < 163) {
return 2;
}
else if (Rs >= 163) {
return 4;
}
}
else if (mu_mv_mag_sq >= 124) {
return 1;
}
}
}
}
}
else if (TSCindex >= 2) {
if (TSC < 4) {
if (RsCsDiff < 1) {
if (SC < 328) {
if (tcor < 101) {
if (MVDiff < 2) {
if (Cs < 118) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (Cs >= 118) {
if (ssDCval < 62) {
if (Rs < 139) {
return 8;
}
else if (Rs >= 139) {
return 1;
}
}
else if (ssDCval >= 62) {
if (posBalance < 27) {
return 1;
}
else if (posBalance >= 27) {
return 2;
}
}
}
}
else if (MVDiff >= 2) {
if (mcTcor < 95) {
if (AFD < 5) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (AFD >= 5) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (posBalance < 40) {
if (ssDCval < 60) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (ssDCval >= 60) {
if (SC < 273) {
return 1;
}
else if (SC >= 273) {
return 2;
}
}
}
else if (posBalance >= 40) {
if (MVDiff < 35) {
if (Rs < 64) {
return 1;
}
else if (Rs >= 64) {
if (MVDiff < 17) {
if (Cs < 220) {
if (negBalance < 22) {
return 4;
}
else if (negBalance >= 22) {
if (SC < 318) {
if (negBalance < 23) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
return 1;
}
}
else if (negBalance >= 23) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
if (Cs < 127) {
if (Rs < 112) {
return 1;
}
else if (Rs >= 112) {
return 4;
}
}
else if (Cs >= 127) {
return 1;
}
}
}
}
else if (SC >= 318) {
if (SC < 319) {
if (Rs < 150) {
return 4;
}
else if (Rs >= 150) {
return 1;
}
}
else if (SC >= 319) {
if (Rs < 181) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (Rs >= 181) {
return 4;
}
}
}
}
}
else if (Cs >= 220) {
return 4;
}
}
else if (MVDiff >= 17) {
return 4;
}
}
}
else if (MVDiff >= 35) {
if (AFD < 8) {
if (Rs < 75) {
return 4;
}
else if (Rs >= 75) {
return 8;
}
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
}
else if (tcor >= 101) {
if (Cs < 20) {
return 2;
}
else if (Cs >= 20) {
return 8;
}
}
}
else if (SC >= 328) {
if (negBalance < 31) {
if (refDCval < 60) {
if (MVDiff < 1) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 240) {
return 1;
}
else if (Rs >= 240) {
return 4;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 193) {
return 1;
}
else if (Cs >= 193) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
return 4;
}
}
}
}
else if (refDCval >= 60) {
if (TSC < 2) {
if (SC < 367) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (Rs < 191) {
return 1;
}
else if (Rs >= 191) {
return 4;
}
}
}
else if (SC >= 367) {
if (mu_mv_mag_sq < 80) {
if (Rs < 160) {
return 2;
}
else if (Rs >= 160) {
return 4;
}
}
else if (mu_mv_mag_sq >= 80) {
if (mu_mv_mag_sq < 123) {
if (Rs < 135) {
return 2;
}
else if (Rs >= 135) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
if (Rs < 186) {
return 4;
}
else if (Rs >= 186) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 219) {
if (Cs < 340) {
return 4;
}
else if (Cs >= 340) {
if (Rs < 206) {
return 1;
}
else if (Rs >= 206) {
return 4;
}
}
}
else if (Rs >= 219) {
return 1;
}
}
}
}
}
else if (TSC >= 2) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 63) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 63) {
if (mu_mv_mag_sq < 110) {
if (negBalance < 28) {
if (Cs < 330) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (Cs < 262) {
return 1;
}
else if (Cs >= 262) {
return 4;
}
}
}
else if (Cs >= 330) {
return 1;
}
}
else if (negBalance >= 28) {
if (Rs < 194) {
if (Rs < 159) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
return 1;
}
}
}
else if (Rs >= 159) {
return 4;
}
}
else if (Rs >= 194) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 170) {
return 4;
}
else if (Rs >= 170) {
if (ssDCval < 69) {
if (SC < 533) {
if (Cs < 336) {
if (Rs < 211) {
return 4;
}
else if (Rs >= 211) {
return 1;
}
}
else if (Cs >= 336) {
return 1;
}
}
else if (SC >= 533) {
return 4;
}
}
else if (ssDCval >= 69) {
if (Cs < 334) {
return 4;
}
else if (Cs >= 334) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 4;
}
}
}
}
}
else if (negBalance >= 31) {
if (tcor < 97) {
if (mu_mv_mag_sq < 94) {
if (ssDCval < 63) {
if (ssDCval < 57) {
if (Rs < 188) {
return 4;
}
else if (Rs >= 188) {
if (Rs < 252) {
return 1;
}
else if (Rs >= 252) {
return 2;
}
}
}
else if (ssDCval >= 57) {
return 4;
}
}
else if (ssDCval >= 63) {
if (Cs < 285) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (Cs >= 285) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (Cs < 304) {
return 1;
}
else if (Cs >= 304) {
return 4;
}
}
}
else if (tcor >= 97) {
if (refDCval < 64) {
if (ssDCval < 60) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Rs < 200) {
return 8;
}
else if (Rs >= 200) {
return 4;
}
}
}
else if (ssDCval >= 60) {
if (posBalance < 51) {
if (ssDCval < 63) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 63) {
if (Rs < 265) {
return 1;
}
else if (Rs >= 265) {
return 4;
}
}
}
else if (posBalance >= 51) {
if (Rs < 193) {
return 4;
}
else if (Rs >= 193) {
return 1;
}
}
}
}
else if (refDCval >= 64) {
if (mu_mv_mag_sq < 79) {
if (Rs < 158) {
return 4;
}
else if (Rs >= 158) {
return 1;
}
}
else if (mu_mv_mag_sq >= 79) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (mcTcor < 93) {
if (MVDiff < 6) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
else if (MVDiff >= 6) {
if (Cs < 65) {
return 2;
}
else if (Cs >= 65) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 191) {
return 4;
}
else if (Rs >= 191) {
return 8;
}
}
}
}
}
else if (mcTcor >= 93) {
if (MVDiff < 3) {
if (Cs < 436) {
if (ssDCval < 69) {
if (SC < 516) {
if (Rs < 170) {
return 1;
}
else if (Rs >= 170) {
if (Rs < 214) {
return 4;
}
else if (Rs >= 214) {
return 1;
}
}
}
else if (SC >= 516) {
if (posBalance < 56) {
if (MVDiff < 1) {
if (ssDCval < 64) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 64) {
return 4;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (posBalance >= 56) {
return 1;
}
}
}
else if (ssDCval >= 69) {
if (SC < 499) {
return 4;
}
else if (SC >= 499) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 81) {
return 1;
}
else if (mu_mv_mag_sq >= 81) {
return 4;
}
}
}
}
}
else if (Cs >= 436) {
if (Rs < 196) {
return 4;
}
else if (Rs >= 196) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 38) {
return 1;
}
else if (Rs >= 38) {
if (RsCsDiff < 5) {
if (Cs < 181) {
if (TSCindex < 3) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
else if (TSCindex >= 3) {
if (tcor < 90) {
if (posBalance < 68) {
return 4;
}
else if (posBalance >= 68) {
return 1;
}
}
else if (tcor >= 90) {
if (posBalance < 58) {
if (negBalance < 37) {
if (gchDC < 1) {
if (Rs < 209) {
if (Rs < 186) {
return 4;
}
else if (Rs >= 186) {
return 1;
}
}
else if (Rs >= 209) {
return 4;
}
}
else if (gchDC >= 1) {
if (CsDiff < 1) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 8;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
}
else if (negBalance >= 37) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (posBalance >= 58) {
return 4;
}
}
}
}
else if (Cs >= 181) {
if (Rs < 282) {
if (MVDiff < 61) {
if (SC < 584) {
if (refDCval < 52) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
return 1;
}
}
else if (refDCval >= 52) {
if (MVDiff < 11) {
if (Rs < 95) {
return 1;
}
else if (Rs >= 95) {
if (Rs < 245) {
if (SC < 374) {
if (ssDCval < 67) {
if (Cs < 238) {
return 4;
}
else if (Cs >= 238) {
return 1;
}
}
else if (ssDCval >= 67) {
return 1;
}
}
else if (SC >= 374) {
return 4;
}
}
else if (Rs >= 245) {
if (gchDC < 1) {
if (tcor < 91) {
return 1;
}
else if (tcor >= 91) {
return 4;
}
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
else if (MVDiff >= 11) {
return 4;
}
}
}
else if (SC >= 584) {
if (negBalance < 51) {
if (Cs < 402) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (SC < 616) {
return 4;
}
else if (SC >= 616) {
return 8;
}
}
}
else if (Cs >= 402) {
return 4;
}
}
else if (negBalance >= 51) {
return 8;
}
}
}
else if (MVDiff >= 61) {
if (Rs < 221) {
return 1;
}
else if (Rs >= 221) {
return 4;
}
}
}
else if (Rs >= 282) {
if (tcor < 97) {
if (AFD < 6) {
if (diffAFD < -1) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 4;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (AFD >= 6) {
if (Rs < 294) {
return 2;
}
else if (Rs >= 294) {
return 4;
}
}
}
else if (tcor >= 97) {
if (Rs < 302) {
return 1;
}
else if (Rs >= 302) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 5) {
if (MVDiff < 12) {
if (Rs < 250) {
return 4;
}
else if (Rs >= 250) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 2;
}
}
}
else if (MVDiff >= 12) {
if (gchDC < 2) {
if (AFD < 13) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
return 4;
}
}
else if (AFD >= 13) {
return 1;
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 41) {
if (tcor < 74) {
if (CsDiff < 5) {
if (diffMVdiffVal < 31) {
if (refDCval < 55) {
if (diffMVdiffVal < -10) {
if (Rs < 111) {
return 2;
}
else if (Rs >= 111) {
return 1;
}
}
else if (diffMVdiffVal >= -10) {
if (mu_mv_mag_sq < 17) {
if (Cs < 194) {
if (Rs < 241) {
return 4;
}
else if (Rs >= 241) {
return 2;
}
}
else if (Cs >= 194) {
return 1;
}
}
else if (mu_mv_mag_sq >= 17) {
return 4;
}
}
}
else if (refDCval >= 55) {
if (tcor < 58) {
if (MVDiff < 259) {
if (diffMVdiffVal < -70) {
return 2;
}
else if (diffMVdiffVal >= -70) {
if (negBalance < 61) {
return 4;
}
else if (negBalance >= 61) {
return 2;
}
}
}
else if (MVDiff >= 259) {
return 1;
}
}
else if (tcor >= 58) {
if (ssDCval < 65) {
if (Rs < 179) {
return 1;
}
else if (Rs >= 179) {
if (Cs < 249) {
return 2;
}
else if (Cs >= 249) {
return 1;
}
}
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 31) {
return 1;
}
}
else if (CsDiff >= 5) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
if (Rs < 222) {
return 1;
}
else if (Rs >= 222) {
return 4;
}
}
}
}
else if (tcor >= 74) {
if (MVDiff < 47) {
if (AFD < 14) {
if (mu_mv_mag_sq < 32) {
if (negBalance < 45) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (tcor < 91) {
return 1;
}
else if (tcor >= 91) {
if (Cs < 104) {
return 1;
}
else if (Cs >= 104) {
return 2;
}
}
}
}
else if (negBalance >= 45) {
return 1;
}
}
else if (mu_mv_mag_sq >= 32) {
return 2;
}
}
else if (AFD >= 14) {
return 2;
}
}
else if (MVDiff >= 47) {
if (Cs < 252) {
if (tcor < 93) {
if (mu_mv_mag_sq < 9) {
return 2;
}
else if (mu_mv_mag_sq >= 9) {
if (Rs < 166) {
return 1;
}
else if (Rs >= 166) {
return 2;
}
}
}
else if (tcor >= 93) {
if (Rs < 106) {
return 8;
}
else if (Rs >= 106) {
return 1;
}
}
}
else if (Cs >= 252) {
if (Rs < 235) {
return 8;
}
else if (Rs >= 235) {
if (Cs < 308) {
return 2;
}
else if (Cs >= 308) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (posBalance < 75) {
if (Rs < 205) {
if (MVDiff < 13) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (MVDiff >= 13) {
if (Cs < 112) {
if (Rs < 56) {
if (Rs < 23) {
return 4;
}
else if (Rs >= 23) {
return 8;
}
}
else if (Rs >= 56) {
if (mu_mv_mag_sq < 53) {
return 1;
}
else if (mu_mv_mag_sq >= 53) {
return 4;
}
}
}
else if (Cs >= 112) {
if (Rs < 52) {
return 1;
}
else if (Rs >= 52) {
return 8;
}
}
}
}
else if (Rs >= 205) {
if (mu_mv_mag_sq < 62) {
if (Rs < 268) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 241) {
return 4;
}
else if (Rs >= 241) {
return 1;
}
}
}
else if (Rs >= 268) {
return 2;
}
}
else if (mu_mv_mag_sq >= 62) {
if (posBalance < 70) {
if (posBalance < 38) {
if (Rs < 256) {
return 4;
}
else if (Rs >= 256) {
return 8;
}
}
else if (posBalance >= 38) {
if (Rs < 301) {
if (mu_mv_mag_sq < 75) {
if (mu_mv_mag_sq < 66) {
return 1;
}
else if (mu_mv_mag_sq >= 66) {
return 4;
}
}
else if (mu_mv_mag_sq >= 75) {
return 1;
}
}
else if (Rs >= 301) {
return 8;
}
}
}
else if (posBalance >= 70) {
return 2;
}
}
}
}
else if (posBalance >= 75) {
return 1;
}
}
}
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 100) {
if (mcTcor < 97) {
if (mcTcor < 61) {
if (gchDC < 5) {
return 4;
}
else if (gchDC >= 5) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (mcTcor >= 61) {
if (Cs < 254) {
if (ssDCval < 112) {
if (tcor < 93) {
if (negBalance < 43) {
if (Rs < 12) {
if (Cs < 85) {
return 1;
}
else if (Cs >= 85) {
return 2;
}
}
else if (Rs >= 12) {
if (Rs < 259) {
return 1;
}
else if (Rs >= 259) {
if (Rs < 272) {
return 8;
}
else if (Rs >= 272) {
return 1;
}
}
}
}
else if (negBalance >= 43) {
if (mu_mv_mag_sq < 69) {
if (negBalance < 60) {
if (Rs < 184) {
if (SC < 108) {
return 1;
}
else if (SC >= 108) {
if (tcor < 61) {
if (Cs < 11) {
return 4;
}
else if (Cs >= 11) {
if (SC < 125) {
return 8;
}
else if (SC >= 125) {
if (diffMVdiffVal < -3) {
if (Rs < 140) {
return 8;
}
else if (Rs >= 140) {
if (Rs < 152) {
return 2;
}
else if (Rs >= 152) {
return 8;
}
}
}
else if (diffMVdiffVal >= -3) {
if (diffMVdiffVal < 22) {
return 2;
}
else if (diffMVdiffVal >= 22) {
if (Cs < 77) {
return 2;
}
else if (Cs >= 77) {
return 8;
}
}
}
}
}
}
else if (tcor >= 61) {
if (mcTcor < 95) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (Rs < 144) {
return 1;
}
else if (Rs >= 144) {
return 4;
}
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 78) {
return 4;
}
else if (ssDCval >= 78) {
if (Rs < 106) {
return 2;
}
else if (Rs >= 106) {
return 1;
}
}
}
}
}
}
else if (Rs >= 184) {
if (diffTSC < -1) {
if (MVDiff < 686) {
return 8;
}
else if (MVDiff >= 686) {
return 1;
}
}
else if (diffTSC >= -1) {
if (Rs < 282) {
if (mu_mv_mag_sq < 24) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
if (ssDCval < 96) {
return 2;
}
else if (ssDCval >= 96) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 2;
}
}
}
else if (Rs >= 282) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
}
else if (negBalance >= 60) {
if (SC < 319) {
if (mu_mv_mag_sq < 43) {
return 1;
}
else if (mu_mv_mag_sq >= 43) {
if (gchDC < 4) {
return 1;
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (SC >= 319) {
if (refDCval < 99) {
if (Rs < 252) {
return 2;
}
else if (Rs >= 252) {
if (Rs < 278) {
return 8;
}
else if (Rs >= 278) {
return 1;
}
}
}
else if (refDCval >= 99) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (diffAFD < 6) {
return 4;
}
else if (diffAFD >= 6) {
if (Rs < 68) {
return 8;
}
else if (Rs >= 68) {
return 2;
}
}
}
}
}
}
else if (tcor >= 93) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (mu_mv_mag_sq < 6) {
return 1;
}
else if (mu_mv_mag_sq >= 6) {
if (Cs < 42) {
if (Rs < 92) {
return 8;
}
else if (Rs >= 92) {
return 1;
}
}
else if (Cs >= 42) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (MVDiff < 13) {
if (Rs < 156) {
return 4;
}
else if (Rs >= 156) {
return 8;
}
}
else if (MVDiff >= 13) {
if (Cs < 2) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 1;
}
}
else if (Cs >= 2) {
if (diffRsCsdiff < -3) {
if (Rs < 285) {
return 1;
}
else if (Rs >= 285) {
if (Rs < 290) {
return 8;
}
else if (Rs >= 290) {
return 2;
}
}
}
else if (diffRsCsdiff >= -3) {
if (ssDCval < 121) {
if (tcor < 79) {
return 2;
}
else if (tcor >= 79) {
if (Rs < 41) {
return 1;
}
else if (Rs >= 41) {
return 8;
}
}
}
else if (ssDCval >= 121) {
if (Cs < 248) {
return 2;
}
else if (Cs >= 248) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 254) {
if (tcor < 85) {
if (RsCsDiff < 7) {
if (Cs < 361) {
if (AFD < 5) {
if (Rs < 230) {
return 2;
}
else if (Rs >= 230) {
if (Rs < 237) {
return 8;
}
else if (Rs >= 237) {
return 2;
}
}
}
else if (AFD >= 5) {
if (diffMVdiffVal < -2) {
if (Rs < 169) {
return 8;
}
else if (Rs >= 169) {
if (diffMVdiffVal < -7) {
if (Cs < 289) {
return 2;
}
else if (Cs >= 289) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
return 2;
}
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 260) {
return 2;
}
else if (Cs >= 260) {
if (Rs < 256) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 8;
}
}
else if (Rs >= 256) {
if (posBalance < 59) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
}
}
else if (Cs >= 361) {
return 8;
}
}
else if (RsCsDiff >= 7) {
if (AFD < 16) {
if (Rs < 249) {
if (Cs < 352) {
return 2;
}
else if (Cs >= 352) {
return 8;
}
}
else if (Rs >= 249) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (AFD >= 16) {
if (negBalance < 63) {
if (mcTcor < 89) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (mcTcor >= 89) {
if (Rs < 287) {
return 1;
}
else if (Rs >= 287) {
return 2;
}
}
}
else if (negBalance >= 63) {
if (diffAFD < -4) {
if (diffAFD < -6) {
if (Rs < 286) {
return 8;
}
else if (Rs >= 286) {
return 2;
}
}
else if (diffAFD >= -6) {
return 4;
}
}
else if (diffAFD >= -4) {
return 2;
}
}
}
}
}
else if (tcor >= 85) {
if (negBalance < 31) {
if (Rs < 269) {
if (Rs < 183) {
return 4;
}
else if (Rs >= 183) {
return 1;
}
}
else if (Rs >= 269) {
return 2;
}
}
else if (negBalance >= 31) {
if (MVDiff < 18) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
else if (MVDiff >= 18) {
if (diffRsCsdiff < 2) {
if (TSC < 7) {
if (Rs < 210) {
return 8;
}
else if (Rs >= 210) {
if (Rs < 272) {
return 1;
}
else if (Rs >= 272) {
return 8;
}
}
}
else if (TSC >= 7) {
return 2;
}
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 18) {
if (ssDCval < 79) {
if (Cs < 145) {
return 2;
}
else if (Cs >= 145) {
return 1;
}
}
else if (ssDCval >= 79) {
if (Rs < 246) {
return 4;
}
else if (Rs >= 246) {
return 1;
}
}
}
else if (negBalance >= 18) {
if (refDCval < 112) {
if (Cs < 256) {
if (Rs < 91) {
if (Cs < 111) {
if (MVDiff < 74) {
if (ssDCval < 100) {
return 1;
}
else if (ssDCval >= 100) {
if (MVDiff < 25) {
return 2;
}
else if (MVDiff >= 25) {
return 1;
}
}
}
else if (MVDiff >= 74) {
if (MVDiff < 189) {
if (Rs < 24) {
return 2;
}
else if (Rs >= 24) {
if (Rs < 53) {
return 8;
}
else if (Rs >= 53) {
return 1;
}
}
}
else if (MVDiff >= 189) {
return 4;
}
}
}
else if (Cs >= 111) {
if (MVDiff < 22) {
if (MVDiff < 8) {
if (AFD < 2) {
if (Rs < 67) {
return 4;
}
else if (Rs >= 67) {
return 2;
}
}
else if (AFD >= 2) {
return 1;
}
}
else if (MVDiff >= 8) {
if (ssDCval < 74) {
if (Cs < 124) {
return 4;
}
else if (Cs >= 124) {
return 2;
}
}
else if (ssDCval >= 74) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 1;
}
}
}
}
else if (MVDiff >= 22) {
if (Rs < 23) {
return 2;
}
else if (Rs >= 23) {
if (Rs < 84) {
return 1;
}
else if (Rs >= 84) {
if (posBalance < 56) {
if (Cs < 128) {
return 1;
}
else if (Cs >= 128) {
return 4;
}
}
else if (posBalance >= 56) {
return 2;
}
}
}
}
}
}
else if (Rs >= 91) {
if (TSC < 2) {
if (refDCval < 75) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
return 2;
}
}
else if (refDCval >= 75) {
if (Cs < 138) {
if (SC < 227) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (SC >= 227) {
if (Cs < 86) {
return 8;
}
else if (Cs >= 86) {
return 1;
}
}
}
else if (Cs >= 138) {
if (Rs < 154) {
if (Cs < 141) {
return 2;
}
else if (Cs >= 141) {
return 1;
}
}
else if (Rs >= 154) {
if (Rs < 270) {
return 2;
}
else if (Rs >= 270) {
return 1;
}
}
}
}
}
else if (TSC >= 2) {
if (posBalance < 54) {
if (tcor < 96) {
if (refDCval < 91) {
if (Rs < 100) {
return 4;
}
else if (Rs >= 100) {
if (SC < 455) {
if (MVDiff < 1) {
if (Rs < 179) {
return 8;
}
else if (Rs >= 179) {
return 2;
}
}
else if (MVDiff >= 1) {
if (MVDiff < 31) {
if (mu_mv_mag_sq < 88) {
return 1;
}
else if (mu_mv_mag_sq >= 88) {
return 8;
}
}
else if (MVDiff >= 31) {
if (Rs < 113) {
return 2;
}
else if (Rs >= 113) {
if (Rs < 171) {
return 1;
}
else if (Rs >= 171) {
return 8;
}
}
}
}
}
else if (SC >= 455) {
if (Rs < 241) {
return 2;
}
else if (Rs >= 241) {
if (Rs < 275) {
return 1;
}
else if (Rs >= 275) {
return 4;
}
}
}
}
}
else if (refDCval >= 91) {
if (ssDCval < 95) {
if (Cs < 182) {
return 4;
}
else if (Cs >= 182) {
return 2;
}
}
else if (ssDCval >= 95) {
if (Rs < 105) {
return 1;
}
else if (Rs >= 105) {
return 2;
}
}
}
}
else if (tcor >= 96) {
if (Rs < 134) {
if (ssDCval < 104) {
if (ssDCval < 90) {
if (tcor < 99) {
if (Cs < 68) {
return 1;
}
else if (Cs >= 68) {
return 2;
}
}
else if (tcor >= 99) {
return 1;
}
}
else if (ssDCval >= 90) {
return 1;
}
}
else if (ssDCval >= 104) {
return 2;
}
}
else if (Rs >= 134) {
if (Cs < 180) {
if (MVDiff < 2) {
if (Rs < 201) {
return 2;
}
else if (Rs >= 201) {
return 8;
}
}
else if (MVDiff >= 2) {
if (TSC < 5) {
if (SC < 221) {
return 8;
}
else if (SC >= 221) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
if (Rs < 256) {
return 1;
}
else if (Rs >= 256) {
return 8;
}
}
}
}
else if (TSC >= 5) {
if (Rs < 235) {
return 1;
}
else if (Rs >= 235) {
return 8;
}
}
}
}
else if (Cs >= 180) {
if (Rs < 136) {
return 4;
}
else if (Rs >= 136) {
if (Cs < 251) {
return 1;
}
else if (Cs >= 251) {
if (Cs < 254) {
return 4;
}
else if (Cs >= 254) {
return 1;
}
}
}
}
}
}
}
else if (posBalance >= 54) {
if (refDCval < 72) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
else if (refDCval >= 72) {
if (MVDiff < 39) {
if (TSC < 4) {
if (negBalance < 33) {
if (Cs < 121) {
return 8;
}
else if (Cs >= 121) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (negBalance >= 33) {
if (posBalance < 60) {
if (mu_mv_mag_sq < 35) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
if (Rs < 279) {
return 2;
}
else if (Rs >= 279) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 35) {
return 2;
}
}
else if (posBalance >= 60) {
return 2;
}
}
}
else if (TSC >= 4) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 1;
}
}
}
else if (MVDiff >= 39) {
if (diffTSC < -1) {
if (Cs < 193) {
if (Rs < 159) {
if (Rs < 127) {
return 1;
}
else if (Rs >= 127) {
return 8;
}
}
else if (Rs >= 159) {
return 2;
}
}
else if (Cs >= 193) {
return 8;
}
}
else if (diffTSC >= -1) {
if (MVDiff < 107) {
return 2;
}
else if (MVDiff >= 107) {
return 4;
}
}
}
}
}
}
}
}
else if (Cs >= 256) {
if (ssDCval < 76) {
if (Rs < 114) {
return 4;
}
else if (Rs >= 114) {
if (mcTcor < 99) {
if (Rs < 154) {
if (negBalance < 34) {
if (refDCval < 72) {
return 4;
}
else if (refDCval >= 72) {
return 1;
}
}
else if (negBalance >= 34) {
return 4;
}
}
else if (Rs >= 154) {
if (negBalance < 40) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (negBalance >= 40) {
if (Rs < 221) {
return 4;
}
else if (Rs >= 221) {
return 1;
}
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
else if (ssDCval >= 76) {
if (TSC < 3) {
if (SC < 589) {
if (ssDCval < 97) {
return 4;
}
else if (ssDCval >= 97) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
else if (SC >= 589) {
if (SC < 747) {
return 1;
}
else if (SC >= 747) {
return 2;
}
}
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 5) {
if (Cs < 382) {
return 4;
}
else if (Cs >= 382) {
if (Rs < 259) {
return 2;
}
else if (Rs >= 259) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (mcTcor < 98) {
if (diffMVdiffVal < 0) {
return 1;
}
else if (diffMVdiffVal >= 0) {
if (Rs < 135) {
return 1;
}
else if (Rs >= 135) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (RsCsDiff < 18) {
if (diffMVdiffVal < -23) {
if (Rs < 120) {
return 2;
}
else if (Rs >= 120) {
return 1;
}
}
else if (diffMVdiffVal >= -23) {
if (tcor < 98) {
return 2;
}
else if (tcor >= 98) {
if (Rs < 187) {
return 1;
}
else if (Rs >= 187) {
return 2;
}
}
}
}
else if (RsCsDiff >= 18) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 112) {
if (posBalance < 10) {
return 8;
}
else if (posBalance >= 10) {
if (MVDiff < 70) {
if (MVDiff < 8) {
if (Cs < 428) {
if (ssDCval < 125) {
return 1;
}
else if (ssDCval >= 125) {
if (diffMVdiffVal < -1) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 283) {
return 8;
}
else if (Rs >= 283) {
return 2;
}
}
}
}
else if (Cs >= 428) {
if (Rs < 301) {
return 2;
}
else if (Rs >= 301) {
return 1;
}
}
}
else if (MVDiff >= 8) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
if (AFD < 5) {
if (SC < 262) {
return 4;
}
else if (SC >= 262) {
if (Rs < 241) {
return 2;
}
else if (Rs >= 241) {
return 4;
}
}
}
else if (AFD >= 5) {
if (Rs < 292) {
return 2;
}
else if (Rs >= 292) {
return 1;
}
}
}
}
}
else if (MVDiff >= 70) {
if (mcTcor < 101) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (Rs < 190) {
return 1;
}
else if (Rs >= 190) {
return 2;
}
}
}
else if (mcTcor >= 101) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 130) {
if (ssDCval < 74) {
if (Rs < 115) {
return 4;
}
else if (Rs >= 115) {
return 8;
}
}
else if (ssDCval >= 74) {
if (negBalance < 16) {
return 4;
}
else if (negBalance >= 16) {
if (tcor < 98) {
if (AFD < 30) {
if (mu_mv_mag_sq < 109) {
if (Cs < 147) {
if (Rs < 102) {
return 4;
}
else if (Rs >= 102) {
return 2;
}
}
else if (Cs >= 147) {
return 1;
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
else if (AFD >= 30) {
if (Rs < 33) {
return 4;
}
else if (Rs >= 33) {
return 1;
}
}
}
else if (tcor >= 98) {
return 1;
}
}
}
}
else if (Rs >= 130) {
if (Cs < 254) {
if (Cs < 125) {
if (ssDCval < 92) {
if (diffMVdiffVal < -4) {
return 2;
}
else if (diffMVdiffVal >= -4) {
if (tcor < 99) {
if (Rs < 141) {
if (Rs < 138) {
return 8;
}
else if (Rs >= 138) {
return 2;
}
}
else if (Rs >= 141) {
return 8;
}
}
else if (tcor >= 99) {
return 2;
}
}
}
else if (ssDCval >= 92) {
if (tcor < 94) {
return 2;
}
else if (tcor >= 94) {
if (Cs < 110) {
return 1;
}
else if (Cs >= 110) {
return 8;
}
}
}
}
else if (Cs >= 125) {
if (negBalance < 24) {
if (ssDCval < 81) {
if (posBalance < 45) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 157) {
return 8;
}
else if (Rs >= 157) {
return 2;
}
}
}
else if (posBalance >= 45) {
if (ssDCval < 73) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 73) {
if (diffAFD < -1) {
if (Rs < 165) {
return 2;
}
else if (Rs >= 165) {
return 8;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
}
}
else if (ssDCval >= 81) {
if (Rs < 219) {
return 2;
}
else if (Rs >= 219) {
if (TSCindex < 2) {
return 1;
}
else if (TSCindex >= 2) {
if (Rs < 269) {
return 8;
}
else if (Rs >= 269) {
return 2;
}
}
}
}
}
else if (negBalance >= 24) {
if (Rs < 178) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 115) {
return 1;
}
else if (mu_mv_mag_sq >= 115) {
return 2;
}
}
}
else if (Rs >= 178) {
if (ssDCval < 74) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (ssDCval >= 74) {
if (SC < 410) {
if (Rs < 190) {
return 4;
}
else if (Rs >= 190) {
return 8;
}
}
else if (SC >= 410) {
return 2;
}
}
}
}
}
}
else if (Cs >= 254) {
if (ssDCval < 78) {
if (Rs < 206) {
if (Rs < 180) {
return 4;
}
else if (Rs >= 180) {
return 8;
}
}
else if (Rs >= 206) {
return 1;
}
}
else if (ssDCval >= 78) {
if (AFD < 3) {
if (SC < 448) {
return 1;
}
else if (SC >= 448) {
return 4;
}
}
else if (AFD >= 3) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 307) {
if (RsDiff < 4) {
if (ssDCval < 75) {
if (mcTcor < 89) {
if (refDCval < 41) {
if (Cs < 157) {
if (MVDiff < 21) {
if (Rs < 624) {
return 8;
}
else if (Rs >= 624) {
return 2;
}
}
else if (MVDiff >= 21) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
return 8;
}
}
}
else if (Cs >= 157) {
if (negBalance < 41) {
if (Cs < 211) {
return 8;
}
else if (Cs >= 211) {
if (Cs < 269) {
return 1;
}
else if (Cs >= 269) {
if (diffRsCsdiff < -4) {
return 8;
}
else if (diffRsCsdiff >= -4) {
if (Cs < 312) {
return 8;
}
else if (Cs >= 312) {
if (Rs < 556) {
return 1;
}
else if (Rs >= 556) {
if (Rs < 629) {
return 8;
}
else if (Rs >= 629) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 41) {
if (tcor < 70) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 325) {
return 8;
}
else if (Rs >= 325) {
return 4;
}
}
}
}
else if (tcor >= 70) {
return 8;
}
}
}
}
else if (refDCval >= 41) {
if (mu_mv_mag_sq < 96) {
if (negBalance < 45) {
if (Cs < 248) {
if (Cs < 120) {
return 1;
}
else if (Cs >= 120) {
return 2;
}
}
else if (Cs >= 248) {
if (MVDiff < 9) {
if (CsDiff < 3) {
if (RsCsDiff < 12) {
return 8;
}
else if (RsCsDiff >= 12) {
if (Rs < 683) {
return 2;
}
else if (Rs >= 683) {
return 8;
}
}
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (MVDiff >= 9) {
if (mcTcor < 82) {
if (Cs < 288) {
return 2;
}
else if (Cs >= 288) {
return 8;
}
}
else if (mcTcor >= 82) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
if (Rs < 383) {
return 1;
}
else if (Rs >= 383) {
if (Cs < 480) {
return 4;
}
else if (Cs >= 480) {
return 1;
}
}
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (MVDiff < 63) {
if (refDCval < 53) {
if (TSC < 6) {
if (AFD < 9) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
return 1;
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (TSC >= 6) {
if (MVDiff < 52) {
return 8;
}
else if (MVDiff >= 52) {
if (Rs < 408) {
return 1;
}
else if (Rs >= 408) {
return 4;
}
}
}
}
else if (refDCval >= 53) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
if (Rs < 731) {
return 4;
}
else if (Rs >= 731) {
return 8;
}
}
}
}
else if (MVDiff >= 63) {
if (mu_mv_mag_sq < 13) {
if (diffRsCsdiff < 2) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (diffRsCsdiff >= 2) {
return 2;
}
}
else if (mu_mv_mag_sq >= 13) {
if (diffAFD < 4) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
return 4;
}
}
else if (diffAFD >= 4) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (negBalance < 39) {
if (Rs < 726) {
return 4;
}
else if (Rs >= 726) {
return 8;
}
}
else if (negBalance >= 39) {
return 8;
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1280) {
if (tcor < 101) {
if (MVDiff < 1) {
if (ssDCval < 69) {
if (tcor < 94) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Rs < 373) {
return 4;
}
else if (Rs >= 373) {
if (Cs < 214) {
return 2;
}
else if (Cs >= 214) {
return 8;
}
}
}
}
else if (tcor >= 94) {
if (Rs < 476) {
if (negBalance < 19) {
if (Rs < 411) {
if (Rs < 351) {
return 1;
}
else if (Rs >= 351) {
return 8;
}
}
else if (Rs >= 411) {
return 1;
}
}
else if (negBalance >= 19) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (Rs >= 476) {
if (Rs < 923) {
if (Rs < 843) {
if (RsCsDiff < 2) {
if (posBalance < 30) {
if (SC < 1081) {
if (Rs < 733) {
return 8;
}
else if (Rs >= 733) {
if (Rs < 748) {
return 1;
}
else if (Rs >= 748) {
return 8;
}
}
}
else if (SC >= 1081) {
return 1;
}
}
else if (posBalance >= 30) {
if (ssDCval < 68) {
if (SC < 1029) {
if (Rs < 485) {
return 8;
}
else if (Rs >= 485) {
return 1;
}
}
else if (SC >= 1029) {
if (negBalance < 23) {
if (posBalance < 31) {
return 1;
}
else if (posBalance >= 31) {
return 8;
}
}
else if (negBalance >= 23) {
if (gchDC < 1) {
if (Rs < 736) {
return 1;
}
else if (Rs >= 736) {
return 8;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 68) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (Rs < 555) {
return 8;
}
else if (Rs >= 555) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 2) {
if (diffMVdiffVal < -1) {
if (mcTcor < 97) {
return 2;
}
else if (mcTcor >= 97) {
if (Cs < 455) {
return 1;
}
else if (Cs >= 455) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (diffRsCsdiff < 2) {
if (posBalance < 44) {
if (Rs < 632) {
return 4;
}
else if (Rs >= 632) {
return 1;
}
}
else if (posBalance >= 44) {
if (diffRsCsdiff < -2) {
return 1;
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
else if (diffRsCsdiff >= 2) {
return 1;
}
}
}
}
else if (Rs >= 843) {
if (mu_mv_mag_sq < 98) {
return 1;
}
else if (mu_mv_mag_sq >= 98) {
if (Cs < 315) {
return 2;
}
else if (Cs >= 315) {
if (Cs < 421) {
return 1;
}
else if (Cs >= 421) {
return 8;
}
}
}
}
}
else if (Rs >= 923) {
return 1;
}
}
}
}
else if (ssDCval >= 69) {
if (SC < 947) {
if (Rs < 626) {
if (ssDCval < 73) {
if (posBalance < 48) {
if (Rs < 517) {
return 1;
}
else if (Rs >= 517) {
return 2;
}
}
else if (posBalance >= 48) {
return 2;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (Rs >= 626) {
return 4;
}
}
else if (SC >= 947) {
if (Rs < 841) {
if (Rs < 601) {
if (AFD < 6) {
if (Rs < 597) {
return 1;
}
else if (Rs >= 597) {
if (Rs < 599) {
return 4;
}
else if (Rs >= 599) {
return 1;
}
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (Rs >= 601) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (Rs < 617) {
return 1;
}
else if (Rs >= 617) {
return 4;
}
}
}
}
else if (Rs >= 841) {
if (Cs < 347) {
return 1;
}
else if (Cs >= 347) {
if (Rs < 1180) {
return 8;
}
else if (Rs >= 1180) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (refDCval < 71) {
if (refDCval < 58) {
if (Rs < 603) {
if (Cs < 128) {
if (Rs < 595) {
if (mu_mv_mag_sq < 49) {
return 1;
}
else if (mu_mv_mag_sq >= 49) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Rs < 324) {
return 8;
}
else if (Rs >= 324) {
return 1;
}
}
}
}
}
else if (Rs >= 595) {
return 8;
}
}
else if (Cs >= 128) {
if (ssDCval < 44) {
if (posBalance < 31) {
if (mu_mv_mag_sq < 86) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (mu_mv_mag_sq >= 86) {
if (RsDiff < 1) {
if (Rs < 459) {
return 8;
}
else if (Rs >= 459) {
return 4;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (posBalance >= 31) {
if (negBalance < 32) {
if (mu_mv_mag_sq < 114) {
if (TSC < 2) {
if (Rs < 371) {
return 4;
}
else if (Rs >= 371) {
return 1;
}
}
else if (TSC >= 2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (negBalance >= 32) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 510) {
if (diffRsCsdiff < -9) {
return 8;
}
else if (diffRsCsdiff >= -9) {
if (AFD < 6) {
if (TSC < 4) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
return 1;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (Rs >= 510) {
if (Rs < 565) {
return 8;
}
else if (Rs >= 565) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 44) {
if (tcor < 70) {
if (ssDCval < 52) {
if (Rs < 359) {
return 1;
}
else if (Rs >= 359) {
return 8;
}
}
else if (ssDCval >= 52) {
return 2;
}
}
else if (tcor >= 70) {
if (mu_mv_mag_sq < 105) {
if (TSCindex < 2) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (TSCindex >= 2) {
if (Rs < 396) {
if (tcor < 97) {
if (ssDCval < 57) {
if (negBalance < 57) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (negBalance < 18) {
return 1;
}
else if (negBalance >= 18) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (mcTcor < 96) {
if (Cs < 505) {
return 4;
}
else if (Cs >= 505) {
return 2;
}
}
else if (mcTcor >= 96) {
if (Cs < 458) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (Cs >= 458) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 57) {
if (Rs < 315) {
return 2;
}
else if (Rs >= 315) {
return 1;
}
}
}
else if (ssDCval >= 57) {
if (SC < 756) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
else if (SC >= 756) {
return 1;
}
}
}
else if (tcor >= 97) {
if (Cs < 494) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 8;
}
}
else if (Cs >= 494) {
return 4;
}
}
}
else if (Rs >= 396) {
if (negBalance < 39) {
if (negBalance < 22) {
if (ssDCval < 54) {
if (Cs < 225) {
if (Rs < 431) {
return 1;
}
else if (Rs >= 431) {
return 4;
}
}
else if (Cs >= 225) {
return 1;
}
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (negBalance >= 22) {
if (tcor < 76) {
return 8;
}
else if (tcor >= 76) {
if (ssDCval < 55) {
if (diffMVdiffVal < -1) {
if (TSC < 4) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (posBalance < 59) {
if (Cs < 452) {
return 4;
}
else if (Cs >= 452) {
if (Rs < 496) {
return 1;
}
else if (Rs >= 496) {
return 4;
}
}
}
else if (posBalance >= 59) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (Rs < 540) {
return 8;
}
else if (Rs >= 540) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (negBalance < 31) {
if (Rs < 592) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (Rs >= 592) {
return 1;
}
}
else if (negBalance >= 31) {
if (refDCval < 48) {
return 1;
}
else if (refDCval >= 48) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (Rs < 530) {
if (refDCval < 54) {
return 1;
}
else if (refDCval >= 54) {
return 4;
}
}
else if (Rs >= 530) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 55) {
if (RsCsDiff < 7) {
if (AFD < 5) {
if (Rs < 544) {
return 8;
}
else if (Rs >= 544) {
return 4;
}
}
else if (AFD >= 5) {
if (Cs < 213) {
return 1;
}
else if (Cs >= 213) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
}
}
else if (RsCsDiff >= 7) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (Rs < 513) {
return 2;
}
else if (Rs >= 513) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 39) {
if (posBalance < 55) {
if (tcor < 93) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (Cs < 235) {
return 1;
}
else if (Cs >= 235) {
return 8;
}
}
}
else if (tcor >= 93) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
if (Cs < 286) {
return 1;
}
else if (Cs >= 286) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (mcTcor < 97) {
if (Cs < 487) {
return 8;
}
else if (Cs >= 487) {
return 1;
}
}
else if (mcTcor >= 97) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Cs < 477) {
return 1;
}
else if (Cs >= 477) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 603) {
if (posBalance < 44) {
if (posBalance < 34) {
if (mu_mv_mag_sq < 77) {
if (Cs < 234) {
return 1;
}
else if (Cs >= 234) {
return 2;
}
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 737) {
return 8;
}
else if (Rs >= 737) {
if (Cs < 443) {
return 1;
}
else if (Cs >= 443) {
if (Rs < 862) {
return 8;
}
else if (Rs >= 862) {
return 4;
}
}
}
}
}
else if (posBalance >= 34) {
if (Cs < 478) {
if (Cs < 230) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 230) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 709) {
return 2;
}
else if (Rs >= 709) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
if (Rs < 851) {
return 2;
}
else if (Rs >= 851) {
return 1;
}
}
}
}
}
}
else if (Cs >= 478) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (Rs < 943) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Rs >= 943) {
return 2;
}
}
}
}
}
else if (posBalance >= 44) {
if (mu_mv_mag_sq < 58) {
if (negBalance < 44) {
if (Cs < 482) {
if (negBalance < 22) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (negBalance >= 22) {
if (mcTcor < 93) {
if (MVDiff < 16) {
if (ssDCval < 38) {
return 1;
}
else if (ssDCval >= 38) {
return 8;
}
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
}
else if (Cs >= 482) {
return 8;
}
}
else if (negBalance >= 44) {
if (Cs < 231) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 8;
}
}
else if (Cs >= 231) {
if (ssDCval < 39) {
return 1;
}
else if (ssDCval >= 39) {
if (diffRsCsdiff < -9) {
if (Rs < 860) {
return 1;
}
else if (Rs >= 860) {
return 4;
}
}
else if (diffRsCsdiff >= -9) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (negBalance < 28) {
if (TSC < 3) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
else if (TSC >= 3) {
if (Cs < 432) {
return 1;
}
else if (Cs >= 432) {
return 4;
}
}
}
else if (negBalance >= 28) {
if (diffRsCsdiff < -4) {
if (TSC < 4) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
return 2;
}
}
else if (TSC >= 4) {
if (Rs < 898) {
return 4;
}
else if (Rs >= 898) {
return 1;
}
}
}
else if (diffRsCsdiff >= -4) {
if (tcor < 86) {
return 4;
}
else if (tcor >= 86) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 58) {
if (tcor < 99) {
if (mcTcor < 97) {
if (SC < 1071) {
if (Cs < 134) {
if (mu_mv_mag_sq < 57) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (mu_mv_mag_sq >= 57) {
if (Rs < 842) {
return 4;
}
else if (Rs >= 842) {
return 2;
}
}
}
else if (Cs >= 134) {
if (refDCval < 61) {
if (SC < 709) {
if (mu_mv_mag_sq < 53) {
return 2;
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 464) {
return 4;
}
else if (Rs >= 464) {
return 2;
}
}
}
else if (SC >= 709) {
if (RsCsDiff < 12) {
if (mcTcor < 96) {
if (SC < 819) {
if (Rs < 402) {
return 2;
}
else if (Rs >= 402) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (SC >= 819) {
return 2;
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 2) {
if (Rs < 375) {
return 4;
}
else if (Rs >= 375) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 433) {
return 1;
}
else if (Rs >= 433) {
if (Rs < 704) {
return 2;
}
else if (Rs >= 704) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 12) {
return 1;
}
}
}
else if (refDCval >= 61) {
if (ssDCval < 66) {
if (ssDCval < 61) {
if (Rs < 352) {
return 1;
}
else if (Rs >= 352) {
return 2;
}
}
else if (ssDCval >= 61) {
if (Cs < 144) {
return 1;
}
else if (Cs >= 144) {
if (Cs < 421) {
if (negBalance < 41) {
if (Rs < 610) {
return 2;
}
else if (Rs >= 610) {
return 4;
}
}
else if (negBalance >= 41) {
return 4;
}
}
else if (Cs >= 421) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 2;
}
}
}
}
}
else if (ssDCval >= 66) {
if (mu_mv_mag_sq < 38) {
if (SC < 996) {
if (RsCsDiff < 9) {
if (SC < 744) {
if (Rs < 323) {
return 4;
}
else if (Rs >= 323) {
return 1;
}
}
else if (SC >= 744) {
return 4;
}
}
else if (RsCsDiff >= 9) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 8;
}
}
}
else if (SC >= 996) {
return 2;
}
}
else if (mu_mv_mag_sq >= 38) {
if (posBalance < 61) {
if (tcor < 91) {
return 1;
}
else if (tcor >= 91) {
if (Rs < 511) {
return 2;
}
else if (Rs >= 511) {
return 1;
}
}
}
else if (posBalance >= 61) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 1071) {
if (Rs < 1059) {
if (tcor < 92) {
if (SC < 1210) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
return 4;
}
}
}
else if (SC >= 1210) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 8;
}
}
}
else if (tcor >= 92) {
if (Cs < 449) {
if (mu_mv_mag_sq < 66) {
return 1;
}
else if (mu_mv_mag_sq >= 66) {
return 4;
}
}
else if (Cs >= 449) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 1;
}
}
}
}
else if (Rs >= 1059) {
if (Cs < 513) {
return 1;
}
else if (Cs >= 513) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 74) {
if (Rs < 585) {
if (TSC < 4) {
if (MVDiff < 2) {
if (Cs < 391) {
return 1;
}
else if (Cs >= 391) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 28) {
if (posBalance < 70) {
return 4;
}
else if (posBalance >= 70) {
return 2;
}
}
else if (mu_mv_mag_sq >= 28) {
if (SC < 599) {
if (Rs < 310) {
return 2;
}
else if (Rs >= 310) {
return 1;
}
}
else if (SC >= 599) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
return 1;
}
}
else if (Rs >= 585) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
if (negBalance < 22) {
if (Cs < 338) {
if (Rs < 884) {
return 1;
}
else if (Rs >= 884) {
return 4;
}
}
else if (Cs >= 338) {
return 2;
}
}
else if (negBalance >= 22) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Cs < 247) {
return 1;
}
else if (Cs >= 247) {
if (CsDiff < 2) {
if (diffRsCsdiff < -2) {
return 1;
}
else if (diffRsCsdiff >= -2) {
return 4;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 309) {
if (ssDCval < 62) {
if (Rs < 927) {
return 2;
}
else if (Rs >= 927) {
return 1;
}
}
else if (ssDCval >= 62) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (SC < 577) {
return 4;
}
else if (SC >= 577) {
return 1;
}
}
}
}
else if (Cs >= 309) {
if (posBalance < 47) {
if (Rs < 819) {
if (AFD < 4) {
if (negBalance < 24) {
if (RsCsDiff < 1) {
if (Rs < 661) {
return 1;
}
else if (Rs >= 661) {
return 8;
}
}
else if (RsCsDiff >= 1) {
return 8;
}
}
else if (negBalance >= 24) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (Rs < 678) {
return 8;
}
else if (Rs >= 678) {
return 1;
}
}
}
}
else if (AFD >= 4) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (Rs >= 819) {
return 1;
}
}
else if (posBalance >= 47) {
if (negBalance < 28) {
if (gchDC < 1) {
if (Rs < 670) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 323) {
return 2;
}
else if (Rs >= 323) {
return 4;
}
}
}
else if (Rs >= 670) {
return 2;
}
}
else if (gchDC >= 1) {
if (Rs < 550) {
if (Cs < 344) {
return 1;
}
else if (Cs >= 344) {
return 8;
}
}
else if (Rs >= 550) {
return 4;
}
}
}
else if (negBalance >= 28) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (Rs < 447) {
return 1;
}
else if (Rs >= 447) {
return 2;
}
}
}
}
}
}
}
}
else if (tcor >= 99) {
if (Cs < 269) {
if (Rs < 431) {
return 4;
}
else if (Rs >= 431) {
if (negBalance < 50) {
if (Rs < 702) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (Rs >= 702) {
return 4;
}
}
else if (negBalance >= 50) {
return 1;
}
}
}
else if (Cs >= 269) {
if (refDCval < 66) {
if (MVDiff < 3) {
if (Cs < 340) {
if (Rs < 616) {
return 4;
}
else if (Rs >= 616) {
return 2;
}
}
else if (Cs >= 340) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 350) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (Rs >= 350) {
if (AFD < 3) {
if (Rs < 413) {
return 4;
}
else if (Rs >= 413) {
return 8;
}
}
else if (AFD >= 3) {
if (Rs < 872) {
return 4;
}
else if (Rs >= 872) {
return 2;
}
}
}
}
}
else if (refDCval >= 66) {
if (Rs < 880) {
if (Rs < 549) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 549) {
return 2;
}
}
else if (Rs >= 880) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 71) {
if (mu_mv_mag_sq < 80) {
if (Rs < 449) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (refDCval < 76) {
return 8;
}
else if (refDCval >= 76) {
return 2;
}
}
}
else if (Rs >= 449) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
if (tcor < 96) {
if (MVDiff < 5) {
if (Cs < 147) {
return 4;
}
else if (Cs >= 147) {
return 1;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (tcor >= 96) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Rs < 578) {
if (refDCval < 72) {
return 4;
}
else if (refDCval >= 72) {
return 1;
}
}
else if (Rs >= 578) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (posBalance < 49) {
if (Rs < 555) {
return 8;
}
else if (Rs >= 555) {
return 4;
}
}
else if (posBalance >= 49) {
if (Rs < 951) {
return 8;
}
else if (Rs >= 951) {
return 1;
}
}
}
}
}
}
else if (tcor >= 101) {
if (RsDiff < 1) {
if (MVDiff < 19) {
if (Rs < 354) {
return 1;
}
else if (Rs >= 354) {
if (posBalance < 1) {
return 4;
}
else if (posBalance >= 1) {
return 8;
}
}
}
else if (MVDiff >= 19) {
if (Cs < 201) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (Cs >= 201) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (mcTcor < 98) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (mcTcor >= 98) {
if (negBalance < 37) {
if (Rs < 433) {
return 1;
}
else if (Rs >= 433) {
return 2;
}
}
else if (negBalance >= 37) {
if (mu_mv_mag_sq < 83) {
if (Cs < 113) {
if (Rs < 426) {
return 8;
}
else if (Rs >= 426) {
return 1;
}
}
else if (Cs >= 113) {
if (posBalance < 3) {
return 4;
}
else if (posBalance >= 3) {
if (mu_mv_mag_sq < 17) {
return 4;
}
else if (mu_mv_mag_sq >= 17) {
if (Cs < 503) {
return 8;
}
else if (Cs >= 503) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 83) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1280) {
if (SC < 1601) {
if (mu_mv_mag_sq < 55) {
if (Rs < 1290) {
return 2;
}
else if (Rs >= 1290) {
return 1;
}
}
else if (mu_mv_mag_sq >= 55) {
if (Cs < 220) {
return 8;
}
else if (Cs >= 220) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 8;
}
}
}
}
else if (SC >= 1601) {
if (mu_mv_mag_sq < 127) {
if (Rs < 1751) {
if (negBalance < 29) {
if (mu_mv_mag_sq < 54) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (Cs < 187) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (Cs >= 187) {
if (Cs < 493) {
return 2;
}
else if (Cs >= 493) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (ssDCval < 64) {
if (Rs < 1371) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
return 8;
}
}
else if (Rs >= 1371) {
if (mu_mv_mag_sq < 123) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (mu_mv_mag_sq >= 123) {
if (TSCindex < 2) {
return 2;
}
else if (TSCindex >= 2) {
return 8;
}
}
}
}
else if (ssDCval >= 64) {
if (Rs < 1435) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
if (Cs < 301) {
return 1;
}
else if (Cs >= 301) {
return 8;
}
}
}
else if (Rs >= 1435) {
return 1;
}
}
}
}
else if (negBalance >= 29) {
if (tcor < 98) {
if (TSC < 5) {
if (SC < 2039) {
return 1;
}
else if (SC >= 2039) {
if (Rs < 1635) {
return 1;
}
else if (Rs >= 1635) {
return 8;
}
}
}
else if (TSC >= 5) {
return 4;
}
}
else if (tcor >= 98) {
if (AFD < 4) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (mu_mv_mag_sq < 81) {
return 8;
}
else if (mu_mv_mag_sq >= 81) {
return 1;
}
}
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (Rs >= 1751) {
if (diffRsCsdiff < -6) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
return 1;
}
}
else if (diffRsCsdiff >= -6) {
if (negBalance < 29) {
return 2;
}
else if (negBalance >= 29) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (ssDCval < 61) {
if (Cs < 513) {
return 8;
}
else if (Cs >= 513) {
return 1;
}
}
else if (ssDCval >= 61) {
if (Rs < 1393) {
if (Cs < 472) {
return 2;
}
else if (Cs >= 472) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
return 8;
}
}
}
else if (Rs >= 1393) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
if (Rs < 1764) {
return 1;
}
else if (Rs >= 1764) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (Cs < 152) {
if (refDCval < 107) {
if (posBalance < 10) {
return 8;
}
else if (posBalance >= 10) {
if (tcor < 63) {
if (Rs < 320) {
return 4;
}
else if (Rs >= 320) {
if (Rs < 327) {
return 8;
}
else if (Rs >= 327) {
if (Cs < 96) {
if (MVDiff < 37) {
return 2;
}
else if (MVDiff >= 37) {
if (Rs < 378) {
return 8;
}
else if (Rs >= 378) {
return 2;
}
}
}
else if (Cs >= 96) {
return 8;
}
}
}
}
else if (tcor >= 63) {
if (RsDiff < 2) {
if (posBalance < 50) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (negBalance < 40) {
if (Cs < 51) {
return 8;
}
else if (Cs >= 51) {
return 1;
}
}
else if (negBalance >= 40) {
if (Rs < 574) {
if (Cs < 74) {
return 1;
}
else if (Cs >= 74) {
return 8;
}
}
else if (Rs >= 574) {
return 1;
}
}
}
}
else if (posBalance >= 50) {
if (Rs < 1326) {
if (Cs < 75) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 75) {
if (Cs < 141) {
return 8;
}
else if (Cs >= 141) {
return 1;
}
}
}
else if (Rs >= 1326) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
}
}
else if (refDCval >= 107) {
if (MVDiff < 20) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (MVDiff >= 20) {
if (negBalance < 71) {
return 1;
}
else if (negBalance >= 71) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
}
else if (Cs >= 152) {
if (SC < 1877) {
if (ssDCval < 81) {
if (AFD < 5) {
if (MVDiff < 1) {
if (ssDCval < 78) {
if (Rs < 1049) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Rs >= 1049) {
return 1;
}
}
else if (ssDCval >= 78) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
if (negBalance < 34) {
if (Rs < 1252) {
return 1;
}
else if (Rs >= 1252) {
return 8;
}
}
else if (negBalance >= 34) {
return 8;
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 33) {
if (posBalance < 41) {
if (Rs < 663) {
return 8;
}
else if (Rs >= 663) {
return 4;
}
}
else if (posBalance >= 41) {
if (Rs < 961) {
if (posBalance < 49) {
if (Rs < 609) {
return 1;
}
else if (Rs >= 609) {
return 8;
}
}
else if (posBalance >= 49) {
return 8;
}
}
else if (Rs >= 961) {
if (Rs < 1270) {
return 1;
}
else if (Rs >= 1270) {
return 8;
}
}
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
else if (AFD >= 5) {
if (tcor < 76) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
if (refDCval < 82) {
return 4;
}
else if (refDCval >= 82) {
return 8;
}
}
}
else if (tcor >= 76) {
if (mu_mv_mag_sq < 52) {
if (Cs < 231) {
if (mu_mv_mag_sq < 3) {
if (Rs < 560) {
return 1;
}
else if (Rs >= 560) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 161) {
return 4;
}
else if (Cs >= 161) {
return 1;
}
}
}
else if (Cs >= 231) {
if (Cs < 465) {
return 1;
}
else if (Cs >= 465) {
if (Rs < 654) {
return 8;
}
else if (Rs >= 654) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
return 8;
}
}
}
}
}
else if (ssDCval >= 81) {
if (TSCindex < 2) {
if (Cs < 317) {
if (posBalance < 45) {
if (Cs < 249) {
if (Rs < 316) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (Rs >= 316) {
return 2;
}
}
else if (Cs >= 249) {
if (Rs < 1034) {
return 1;
}
else if (Rs >= 1034) {
return 2;
}
}
}
else if (posBalance >= 45) {
if (Cs < 176) {
if (Cs < 168) {
return 1;
}
else if (Cs >= 168) {
return 2;
}
}
else if (Cs >= 176) {
return 1;
}
}
}
else if (Cs >= 317) {
if (Cs < 467) {
if (posBalance < 34) {
if (ssDCval < 150) {
return 8;
}
else if (ssDCval >= 150) {
return 4;
}
}
else if (posBalance >= 34) {
if (Rs < 782) {
return 4;
}
else if (Rs >= 782) {
return 1;
}
}
}
else if (Cs >= 467) {
return 8;
}
}
}
else if (TSCindex >= 2) {
if (mcTcor < 95) {
if (MVDiff < 21) {
if (mu_mv_mag_sq < 2) {
if (TSC < 6) {
if (ssDCval < 120) {
if (MVDiff < 5) {
if (ssDCval < 110) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 1;
}
}
else if (ssDCval >= 110) {
return 2;
}
}
else if (MVDiff >= 5) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
return 2;
}
}
}
else if (ssDCval >= 120) {
return 8;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (AFD < 12) {
if (RsCsDiff < 4) {
if (ssDCval < 111) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 342) {
if (Rs < 314) {
return 8;
}
else if (Rs >= 314) {
return 2;
}
}
else if (Rs >= 342) {
return 1;
}
}
}
}
else if (ssDCval >= 111) {
return 8;
}
}
else if (RsCsDiff >= 4) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 21) {
if (Rs < 965) {
return 2;
}
else if (Rs >= 965) {
return 8;
}
}
else if (mu_mv_mag_sq >= 21) {
return 8;
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (AFD >= 12) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (negBalance < 64) {
return 8;
}
else if (negBalance >= 64) {
return 2;
}
}
}
}
}
else if (MVDiff >= 21) {
if (mcTcor < 86) {
if (SC < 622) {
return 8;
}
else if (SC >= 622) {
if (AFD < 26) {
if (Rs < 562) {
return 2;
}
else if (Rs >= 562) {
return 1;
}
}
else if (AFD >= 26) {
return 4;
}
}
}
else if (mcTcor >= 86) {
if (diffTSC < -1) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
return 2;
}
}
else if (diffTSC >= -1) {
if (tcor < 77) {
if (Rs < 315) {
return 1;
}
else if (Rs >= 315) {
if (ssDCval < 114) {
return 8;
}
else if (ssDCval >= 114) {
return 4;
}
}
}
else if (tcor >= 77) {
if (ssDCval < 107) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (Rs < 483) {
return 8;
}
else if (Rs >= 483) {
return 1;
}
}
}
else if (ssDCval >= 107) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (SC < 1076) {
if (ssDCval < 116) {
if (Cs < 445) {
if (MVDiff < 4) {
if (negBalance < 29) {
if (Cs < 224) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (Rs < 423) {
return 4;
}
else if (Rs >= 423) {
return 1;
}
}
}
else if (Cs >= 224) {
if (Rs < 706) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Rs < 533) {
return 1;
}
else if (Rs >= 533) {
return 4;
}
}
}
else if (Rs >= 706) {
return 1;
}
}
}
else if (negBalance >= 29) {
if (posBalance < 51) {
if (mu_mv_mag_sq < 96) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 4;
}
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 553) {
return 1;
}
else if (Rs >= 553) {
return 4;
}
}
}
else if (posBalance >= 51) {
return 1;
}
}
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 27) {
if (MVDiff < 24) {
if (mcTcor < 97) {
if (Rs < 356) {
return 2;
}
else if (Rs >= 356) {
if (Rs < 568) {
return 8;
}
else if (Rs >= 568) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
return 4;
}
}
}
else if (MVDiff >= 24) {
if (Cs < 234) {
return 4;
}
else if (Cs >= 234) {
if (Rs < 437) {
return 2;
}
else if (Rs >= 437) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (Cs < 216) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
return 1;
}
}
else if (Cs >= 216) {
if (AFD < 15) {
if (MVDiff < 11) {
if (SC < 879) {
if (negBalance < 49) {
if (Rs < 411) {
return 1;
}
else if (Rs >= 411) {
return 2;
}
}
else if (negBalance >= 49) {
return 8;
}
}
else if (SC >= 879) {
return 1;
}
}
else if (MVDiff >= 11) {
if (Rs < 696) {
return 8;
}
else if (Rs >= 696) {
return 1;
}
}
}
else if (AFD >= 15) {
return 4;
}
}
}
}
}
else if (Cs >= 445) {
if (posBalance < 46) {
if (ssDCval < 108) {
if (mcTcor < 100) {
return 8;
}
else if (mcTcor >= 100) {
return 1;
}
}
else if (ssDCval >= 108) {
if (posBalance < 20) {
return 8;
}
else if (posBalance >= 20) {
return 1;
}
}
}
else if (posBalance >= 46) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
if (negBalance < 61) {
if (Rs < 337) {
return 8;
}
else if (Rs >= 337) {
if (posBalance < 50) {
if (ssDCval < 94) {
return 1;
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (posBalance >= 50) {
return 1;
}
}
}
else if (negBalance >= 61) {
return 2;
}
}
}
}
}
else if (ssDCval >= 116) {
if (refDCval < 129) {
if (diffTSC < -1) {
if (MVDiff < 27) {
if (Rs < 442) {
return 8;
}
else if (Rs >= 442) {
if (Rs < 587) {
return 1;
}
else if (Rs >= 587) {
return 8;
}
}
}
else if (MVDiff >= 27) {
if (Cs < 362) {
return 2;
}
else if (Cs >= 362) {
return 4;
}
}
}
else if (diffTSC >= -1) {
if (negBalance < 29) {
if (Cs < 378) {
return 2;
}
else if (Cs >= 378) {
if (Rs < 480) {
return 1;
}
else if (Rs >= 480) {
return 2;
}
}
}
else if (negBalance >= 29) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (RsDiff < 3) {
if (diffMVdiffVal < -41) {
return 4;
}
else if (diffMVdiffVal >= -41) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 2;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 442) {
return 2;
}
else if (Rs >= 442) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 129) {
if (AFD < 6) {
if (ssDCval < 161) {
if (mcTcor < 99) {
if (MVDiff < 29) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Cs < 435) {
if (Rs < 343) {
return 2;
}
else if (Rs >= 343) {
return 1;
}
}
else if (Cs >= 435) {
return 8;
}
}
}
else if (MVDiff >= 29) {
if (Cs < 316) {
return 2;
}
else if (Cs >= 316) {
if (Rs < 362) {
return 8;
}
else if (Rs >= 362) {
return 4;
}
}
}
}
else if (mcTcor >= 99) {
if (Rs < 444) {
return 1;
}
else if (Rs >= 444) {
return 8;
}
}
}
else if (ssDCval >= 161) {
return 4;
}
}
else if (AFD >= 6) {
if (MVDiff < 12) {
if (RsDiff < 3) {
if (SC < 1003) {
return 1;
}
else if (SC >= 1003) {
return 2;
}
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (MVDiff >= 12) {
if (AFD < 19) {
if (refDCval < 133) {
return 4;
}
else if (refDCval >= 133) {
return 2;
}
}
else if (AFD >= 19) {
return 1;
}
}
}
}
}
}
else if (SC >= 1076) {
if (ssDCval < 122) {
if (refDCval < 85) {
if (MVDiff < 12) {
if (mu_mv_mag_sq < 121) {
if (tcor < 90) {
return 2;
}
else if (tcor >= 90) {
if (RsDiff < 1) {
if (SC < 1175) {
return 1;
}
else if (SC >= 1175) {
return 8;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 121) {
return 2;
}
}
else if (MVDiff >= 12) {
if (RsCsDiff < 5) {
return 8;
}
else if (RsCsDiff >= 5) {
return 1;
}
}
}
else if (refDCval >= 85) {
if (RsCsDiff < 10) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (Rs < 579) {
if (Cs < 503) {
return 8;
}
else if (Cs >= 503) {
return 2;
}
}
else if (Rs >= 579) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
if (refDCval < 88) {
if (Rs < 772) {
return 1;
}
else if (Rs >= 772) {
return 8;
}
}
else if (refDCval >= 88) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 10) {
if (TSC < 7) {
if (Rs < 603) {
return 2;
}
else if (Rs >= 603) {
return 1;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
else if (ssDCval >= 122) {
if (ssDCval < 136) {
if (AFD < 9) {
if (Rs < 581) {
return 8;
}
else if (Rs >= 581) {
if (posBalance < 38) {
return 2;
}
else if (posBalance >= 38) {
return 1;
}
}
}
else if (AFD >= 9) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (ssDCval >= 136) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
}
}
}
}
else if (SC >= 1877) {
if (Cs < 293) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
if (Rs < 1969) {
return 8;
}
else if (Rs >= 1969) {
return 4;
}
}
}
else if (Cs >= 293) {
if (Rs < 2073) {
if (Cs < 504) {
if (ssDCval < 96) {
if (mu_mv_mag_sq < 86) {
if (AFD < 5) {
if (mu_mv_mag_sq < 65) {
return 8;
}
else if (mu_mv_mag_sq >= 65) {
if (Cs < 453) {
if (ssDCval < 77) {
if (Rs < 1835) {
return 1;
}
else if (Rs >= 1835) {
return 8;
}
}
else if (ssDCval >= 77) {
return 1;
}
}
else if (Cs >= 453) {
return 8;
}
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 86) {
if (Rs < 2044) {
if (TSC < 4) {
if (Cs < 486) {
return 1;
}
else if (Cs >= 486) {
if (Rs < 1679) {
return 1;
}
else if (Rs >= 1679) {
return 4;
}
}
}
else if (TSC >= 4) {
return 4;
}
}
else if (Rs >= 2044) {
return 2;
}
}
}
else if (ssDCval >= 96) {
if (ssDCval < 112) {
return 8;
}
else if (ssDCval >= 112) {
if (tcor < 97) {
return 2;
}
else if (tcor >= 97) {
if (Rs < 1813) {
return 4;
}
else if (Rs >= 1813) {
return 1;
}
}
}
}
}
else if (Cs >= 504) {
if (mu_mv_mag_sq < 101) {
return 1;
}
else if (mu_mv_mag_sq >= 101) {
return 4;
}
}
}
else if (Rs >= 2073) {
if (posBalance < 53) {
if (Cs < 382) {
return 1;
}
else if (Cs >= 382) {
return 8;
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (RsDiff < 8) {
if (negBalance < 46) {
if (refDCval < 99) {
if (diffRsCsdiff < 3) {
if (diffMVdiffVal < -61) {
return 8;
}
else if (diffMVdiffVal >= -61) {
if (AFD < 11) {
if (Rs < 621) {
return 8;
}
else if (Rs >= 621) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (AFD >= 11) {
if (diffRsCsdiff < -58) {
if (Rs < 436) {
return 8;
}
else if (Rs >= 436) {
return 2;
}
}
else if (diffRsCsdiff >= -58) {
if (posBalance < 40) {
return 2;
}
else if (posBalance >= 40) {
if (Cs < 350) {
return 1;
}
else if (Cs >= 350) {
if (SC < 1038) {
if (Rs < 317) {
return 2;
}
else if (Rs >= 317) {
return 1;
}
}
else if (SC >= 1038) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 2;
}
}
}
}
}
}
}
}
else if (diffRsCsdiff >= 3) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (RsCsDiff < 62) {
if (AFD < 39) {
return 1;
}
else if (AFD >= 39) {
if (Rs < 461) {
return 2;
}
else if (Rs >= 461) {
return 1;
}
}
}
else if (RsCsDiff >= 62) {
if (AFD < 28) {
return 2;
}
else if (AFD >= 28) {
return 1;
}
}
}
}
}
else if (refDCval >= 99) {
return 8;
}
}
else if (negBalance >= 46) {
if (MVDiff < 26) {
if (Cs < 322) {
if (mu_mv_mag_sq < 6) {
if (Cs < 257) {
return 2;
}
else if (Cs >= 257) {
if (Rs < 824) {
return 8;
}
else if (Rs >= 824) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 6) {
if (mcTcor < 85) {
if (Cs < 113) {
return 1;
}
else if (Cs >= 113) {
return 8;
}
}
else if (mcTcor >= 85) {
if (Cs < 134) {
return 8;
}
else if (Cs >= 134) {
return 1;
}
}
}
}
else if (Cs >= 322) {
if (diffMVdiffVal < -8) {
if (negBalance < 53) {
if (Rs < 711) {
return 1;
}
else if (Rs >= 711) {
return 2;
}
}
else if (negBalance >= 53) {
return 4;
}
}
else if (diffMVdiffVal >= -8) {
if (ssDCval < 100) {
if (Rs < 532) {
return 4;
}
else if (Rs >= 532) {
return 8;
}
}
else if (ssDCval >= 100) {
if (ssDCval < 113) {
if (Rs < 699) {
return 1;
}
else if (Rs >= 699) {
return 4;
}
}
else if (ssDCval >= 113) {
if (Cs < 402) {
return 2;
}
else if (Cs >= 402) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 26) {
if (AFD < 22) {
if (mu_mv_mag_sq < 71) {
if (CsDiff < 3) {
if (posBalance < 58) {
if (Rs < 512) {
if (Rs < 430) {
return 1;
}
else if (Rs >= 430) {
return 8;
}
}
else if (Rs >= 512) {
return 1;
}
}
else if (posBalance >= 58) {
if (ssDCval < 98) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (tcor < 55) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
return 2;
}
}
else if (tcor >= 55) {
if (MVDiff < 160) {
return 2;
}
else if (MVDiff >= 160) {
return 8;
}
}
}
}
else if (ssDCval >= 98) {
if (diffRsCsdiff < 2) {
if (Rs < 794) {
return 1;
}
else if (Rs >= 794) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 596) {
return 4;
}
else if (Rs >= 596) {
return 2;
}
}
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 45) {
if (diffAFD < 3) {
if (negBalance < 50) {
if (Rs < 475) {
return 8;
}
else if (Rs >= 475) {
return 2;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (diffAFD >= 3) {
if (Rs < 725) {
return 2;
}
else if (Rs >= 725) {
return 1;
}
}
}
else if (refDCval >= 45) {
if (Rs < 766) {
if (diffMVdiffVal < 38) {
if (SC < 1150) {
if (MVDiff < 50) {
if (diffRsCsdiff < -8) {
if (mu_mv_mag_sq < 21) {
return 2;
}
else if (mu_mv_mag_sq >= 21) {
return 8;
}
}
else if (diffRsCsdiff >= -8) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 8;
}
}
}
else if (MVDiff >= 50) {
if (diffMVdiffVal < 8) {
return 2;
}
else if (diffMVdiffVal >= 8) {
if (ssDCval < 64) {
if (Rs < 365) {
return 4;
}
else if (Rs >= 365) {
return 1;
}
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
else if (SC >= 1150) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
return 2;
}
}
}
else if (diffMVdiffVal >= 38) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (Rs >= 766) {
if (Cs < 445) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (Rs < 924) {
return 8;
}
else if (Rs >= 924) {
return 1;
}
}
}
else if (Cs >= 445) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
if (Rs < 475) {
return 1;
}
else if (Rs >= 475) {
return 2;
}
}
}
}
else if (AFD >= 22) {
if (mu_mv_mag_sq < 50) {
if (RsCsDiff < 63) {
if (diffTSC < 4) {
if (AFD < 26) {
if (MVDiff < 58) {
return 1;
}
else if (MVDiff >= 58) {
if (SC < 858) {
return 2;
}
else if (SC >= 858) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
return 2;
}
}
}
}
else if (AFD >= 26) {
return 1;
}
}
else if (diffTSC >= 4) {
return 8;
}
}
else if (RsCsDiff >= 63) {
if (MVDiff < 114) {
if (diffAFD < 5) {
return 2;
}
else if (diffAFD >= 5) {
if (Rs < 640) {
return 4;
}
else if (Rs >= 640) {
return 1;
}
}
}
else if (MVDiff >= 114) {
if (SC < 1016) {
return 1;
}
else if (SC >= 1016) {
if (Rs < 643) {
return 1;
}
else if (Rs >= 643) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (TSC < 14) {
if (diffRsCsdiff < -14) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
else if (diffRsCsdiff >= -14) {
if (AFD < 42) {
return 2;
}
else if (AFD >= 42) {
return 8;
}
}
}
else if (TSC >= 14) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (diffAFD < 5) {
if (posBalance < 53) {
if (SC < 1971) {
if (negBalance < 55) {
if (refDCval < 44) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
else if (refDCval >= 44) {
return 1;
}
}
else if (negBalance >= 55) {
if (diffRsCsdiff < 17) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (diffRsCsdiff >= 17) {
if (Rs < 836) {
return 8;
}
else if (Rs >= 836) {
return 2;
}
}
}
}
else if (SC >= 1971) {
if (MVDiff < 48) {
if (gchDC < 4) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 2192) {
return 2;
}
else if (Rs >= 2192) {
return 1;
}
}
}
else if (gchDC >= 4) {
return 1;
}
}
else if (MVDiff >= 48) {
if (CsDiff < 4) {
if (SC < 2445) {
return 1;
}
else if (SC >= 2445) {
return 8;
}
}
else if (CsDiff >= 4) {
if (Cs < 471) {
return 2;
}
else if (Cs >= 471) {
return 1;
}
}
}
}
}
else if (posBalance >= 53) {
if (Cs < 366) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (refDCval < 131) {
if (ssDCval < 71) {
if (posBalance < 63) {
if (MVDiff < 24) {
return 2;
}
else if (MVDiff >= 24) {
if (Cs < 160) {
return 1;
}
else if (Cs >= 160) {
return 4;
}
}
}
else if (posBalance >= 63) {
if (RsDiff < 12) {
if (Rs < 1053) {
return 2;
}
else if (Rs >= 1053) {
return 1;
}
}
else if (RsDiff >= 12) {
return 2;
}
}
}
else if (ssDCval >= 71) {
if (tcor < 25) {
return 4;
}
else if (tcor >= 25) {
if (mu_mv_mag_sq < 15) {
if (MVDiff < 36) {
if (Rs < 1689) {
return 1;
}
else if (Rs >= 1689) {
return 8;
}
}
else if (MVDiff >= 36) {
if (refDCval < 71) {
return 1;
}
else if (refDCval >= 71) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (Cs < 337) {
return 2;
}
else if (Cs >= 337) {
if (diffMVdiffVal < -4) {
if (Rs < 1498) {
return 1;
}
else if (Rs >= 1498) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 131) {
if (RsCsDiff < 67) {
return 4;
}
else if (RsCsDiff >= 67) {
if (diffAFD < 1) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
if (Rs < 1374) {
return 1;
}
else if (Rs >= 1374) {
return 8;
}
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
}
else if (Cs >= 366) {
if (Rs < 1842) {
if (negBalance < 50) {
if (MVDiff < 52) {
if (Rs < 1422) {
return 1;
}
else if (Rs >= 1422) {
return 2;
}
}
else if (MVDiff >= 52) {
return 8;
}
}
else if (negBalance >= 50) {
if (mu_mv_mag_sq < 75) {
if (posBalance < 63) {
if (CsDiff < 4) {
if (Cs < 499) {
return 8;
}
else if (Cs >= 499) {
return 4;
}
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (posBalance >= 63) {
if (mu_mv_mag_sq < 43) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 841) {
return 8;
}
else if (Rs >= 841) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (Rs < 1200) {
return 1;
}
else if (Rs >= 1200) {
return 2;
}
}
}
}
else if (Rs >= 1842) {
if (AFD < 23) {
if (TSC < 8) {
if (Rs < 2234) {
return 1;
}
else if (Rs >= 2234) {
return 2;
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (AFD >= 23) {
if (tcor < 52) {
if (Rs < 2185) {
return 4;
}
else if (Rs >= 2185) {
return 2;
}
}
else if (tcor >= 52) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Cs < 486) {
return 2;
}
else if (Cs >= 486) {
return 1;
}
}
}
}
}
}
}
}
else if (diffAFD >= 5) {
if (MVDiff < 35) {
if (TSC < 7) {
if (Rs < 1108) {
return 8;
}
else if (Rs >= 1108) {
return 1;
}
}
else if (TSC >= 7) {
if (Cs < 498) {
if (diffTSC < 4) {
return 2;
}
else if (diffTSC >= 4) {
return 4;
}
}
else if (Cs >= 498) {
return 1;
}
}
}
else if (MVDiff >= 35) {
if (tcor < 57) {
if (gchDC < 5) {
if (negBalance < 57) {
if (Rs < 2042) {
return 1;
}
else if (Rs >= 2042) {
return 2;
}
}
else if (negBalance >= 57) {
if (Rs < 733) {
return 1;
}
else if (Rs >= 733) {
return 2;
}
}
}
else if (gchDC >= 5) {
if (diffTSC < 6) {
return 1;
}
else if (diffTSC >= 6) {
return 8;
}
}
}
else if (tcor >= 57) {
if (Rs < 896) {
return 8;
}
else if (Rs >= 896) {
return 1;
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 81) {
if (RsDiff < 9) {
if (refDCval < 88) {
if (Cs < 338) {
if (Rs < 678) {
if (CsDiff < 2) {
if (posBalance < 43) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (Rs < 443) {
return 1;
}
else if (Rs >= 443) {
return 8;
}
}
}
else if (posBalance >= 43) {
return 1;
}
}
else if (CsDiff >= 2) {
if (mcTcor < 97) {
if (SC < 559) {
return 8;
}
else if (SC >= 559) {
if (posBalance < 35) {
return 2;
}
else if (posBalance >= 35) {
if (mu_mv_mag_sq < 52) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (RsCsDiff < 23) {
return 4;
}
else if (RsCsDiff >= 23) {
if (Rs < 553) {
if (Rs < 415) {
return 1;
}
else if (Rs >= 415) {
return 4;
}
}
else if (Rs >= 553) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (gchDC < 5) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (gchDC >= 5) {
return 8;
}
}
}
}
else if (Rs >= 678) {
if (refDCval < 81) {
if (mcTcor < 99) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
if (Cs < 249) {
return 1;
}
else if (Cs >= 249) {
if (Rs < 1147) {
if (Rs < 804) {
return 1;
}
else if (Rs >= 804) {
return 4;
}
}
else if (Rs >= 1147) {
return 1;
}
}
}
}
else if (mcTcor >= 99) {
if (MVDiff < 13) {
if (Rs < 1581) {
return 2;
}
else if (Rs >= 1581) {
return 1;
}
}
else if (MVDiff >= 13) {
return 1;
}
}
}
else if (refDCval >= 81) {
if (mu_mv_mag_sq < 2) {
if (Cs < 177) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
else if (Cs >= 177) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
}
}
}
}
else if (Cs >= 338) {
if (Rs < 823) {
if (AFD < 12) {
if (negBalance < 42) {
if (Rs < 613) {
if (Rs < 398) {
return 2;
}
else if (Rs >= 398) {
return 8;
}
}
else if (Rs >= 613) {
if (Cs < 512) {
return 1;
}
else if (Cs >= 512) {
return 2;
}
}
}
else if (negBalance >= 42) {
return 8;
}
}
else if (AFD >= 12) {
if (diffAFD < -3) {
if (Cs < 392) {
return 4;
}
else if (Cs >= 392) {
if (Cs < 470) {
return 2;
}
else if (Cs >= 470) {
return 8;
}
}
}
else if (diffAFD >= -3) {
if (ssDCval < 73) {
if (Rs < 703) {
return 1;
}
else if (Rs >= 703) {
if (refDCval < 60) {
return 1;
}
else if (refDCval >= 60) {
return 8;
}
}
}
else if (ssDCval >= 73) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
}
else if (Rs >= 823) {
if (refDCval < 65) {
if (SC < 2598) {
if (Cs < 462) {
if (CsDiff < 3) {
if (MVDiff < 8) {
if (ssDCval < 58) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 1;
}
}
else if (ssDCval >= 58) {
return 2;
}
}
else if (MVDiff >= 8) {
if (negBalance < 29) {
if (Cs < 411) {
return 1;
}
else if (Cs >= 411) {
if (Cs < 445) {
return 4;
}
else if (Cs >= 445) {
return 1;
}
}
}
else if (negBalance >= 29) {
if (negBalance < 51) {
return 1;
}
else if (negBalance >= 51) {
return 4;
}
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 52) {
return 2;
}
else if (refDCval >= 52) {
if (diffTSC < -1) {
if (Rs < 2000) {
return 4;
}
else if (Rs >= 2000) {
return 2;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
}
}
else if (Cs >= 462) {
if (TSC < 6) {
if (ssDCval < 55) {
if (diffMVdiffVal < 1) {
if (Rs < 1526) {
return 2;
}
else if (Rs >= 1526) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
if (Rs < 1023) {
return 2;
}
else if (Rs >= 1023) {
return 1;
}
}
}
else if (ssDCval >= 55) {
if (Cs < 519) {
return 2;
}
else if (Cs >= 519) {
return 1;
}
}
}
else if (TSC >= 6) {
if (negBalance < 44) {
if (Rs < 1061) {
return 8;
}
else if (Rs >= 1061) {
return 2;
}
}
else if (negBalance >= 44) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
return 4;
}
}
}
}
}
else if (SC >= 2598) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (refDCval >= 65) {
if (mcTcor < 89) {
if (Rs < 1173) {
return 2;
}
else if (Rs >= 1173) {
return 4;
}
}
else if (mcTcor >= 89) {
if (tcor < 98) {
if (posBalance < 55) {
if (diffRsCsdiff < -51) {
return 4;
}
else if (diffRsCsdiff >= -51) {
if (RsCsDiff < 19) {
if (Rs < 1883) {
return 4;
}
else if (Rs >= 1883) {
return 1;
}
}
else if (RsCsDiff >= 19) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (Rs < 2255) {
if (negBalance < 53) {
if (MVDiff < 40) {
return 1;
}
else if (MVDiff >= 40) {
if (MVDiff < 54) {
if (Cs < 401) {
return 2;
}
else if (Cs >= 401) {
return 4;
}
}
else if (MVDiff >= 54) {
return 1;
}
}
}
else if (negBalance >= 53) {
if (Rs < 1244) {
return 2;
}
else if (Rs >= 1244) {
return 4;
}
}
}
else if (Rs >= 2255) {
return 2;
}
}
}
else if (tcor >= 98) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
if (MVDiff < 37) {
return 1;
}
else if (MVDiff >= 37) {
return 2;
}
}
}
}
}
}
}
}
else if (refDCval >= 88) {
if (AFD < 12) {
if (negBalance < 39) {
if (TSC < 3) {
if (SC < 2037) {
if (Rs < 1283) {
return 1;
}
else if (Rs >= 1283) {
return 2;
}
}
else if (SC >= 2037) {
return 8;
}
}
else if (TSC >= 3) {
if (ssDCval < 144) {
if (Cs < 448) {
if (Cs < 157) {
return 2;
}
else if (Cs >= 157) {
if (diffRsCsdiff < 10) {
return 1;
}
else if (diffRsCsdiff >= 10) {
if (Rs < 1929) {
return 2;
}
else if (Rs >= 1929) {
return 8;
}
}
}
}
else if (Cs >= 448) {
if (ssDCval < 106) {
return 1;
}
else if (ssDCval >= 106) {
return 2;
}
}
}
else if (ssDCval >= 144) {
if (Rs < 871) {
return 8;
}
else if (Rs >= 871) {
return 1;
}
}
}
}
else if (negBalance >= 39) {
if (AFD < 8) {
if (MVDiff < 9) {
if (Rs < 1682) {
return 4;
}
else if (Rs >= 1682) {
return 1;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (AFD >= 8) {
if (ssDCval < 138) {
if (SC < 1361) {
if (posBalance < 49) {
if (Rs < 991) {
return 4;
}
else if (Rs >= 991) {
return 1;
}
}
else if (posBalance >= 49) {
return 1;
}
}
else if (SC >= 1361) {
if (RsDiff < 5) {
if (diffRsCsdiff < 3) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 1;
}
}
else if (diffRsCsdiff >= 3) {
if (Rs < 1126) {
return 2;
}
else if (Rs >= 1126) {
return 8;
}
}
}
else if (RsDiff >= 5) {
if (ssDCval < 128) {
if (tcor < 89) {
return 2;
}
else if (tcor >= 89) {
if (Rs < 1407) {
return 1;
}
else if (Rs >= 1407) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (RsCsDiff < 56) {
return 2;
}
else if (RsCsDiff >= 56) {
if (Rs < 1784) {
return 1;
}
else if (Rs >= 1784) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 128) {
return 1;
}
}
}
}
else if (ssDCval >= 138) {
if (posBalance < 50) {
if (Rs < 791) {
if (MVDiff < 39) {
return 2;
}
else if (MVDiff >= 39) {
return 8;
}
}
else if (Rs >= 791) {
return 1;
}
}
else if (posBalance >= 50) {
if (Rs < 775) {
return 8;
}
else if (Rs >= 775) {
return 4;
}
}
}
}
}
}
else if (AFD >= 12) {
if (CsDiff < 4) {
if (mcTcor < 98) {
if (refDCval < 106) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (MVDiff < 37) {
if (AFD < 15) {
if (SC < 820) {
return 1;
}
else if (SC >= 820) {
return 2;
}
}
else if (AFD >= 15) {
return 1;
}
}
else if (MVDiff >= 37) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
if (Rs < 532) {
return 2;
}
else if (Rs >= 532) {
return 8;
}
}
}
}
}
else if (refDCval >= 106) {
if (SC < 1156) {
if (ssDCval < 121) {
if (RsDiff < 5) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (ssDCval >= 121) {
if (AFD < 15) {
if (Rs < 706) {
return 2;
}
else if (Rs >= 706) {
return 4;
}
}
else if (AFD >= 15) {
if (mu_mv_mag_sq < 18) {
if (Cs < 304) {
return 1;
}
else if (Cs >= 304) {
return 2;
}
}
else if (mu_mv_mag_sq >= 18) {
if (tcor < 77) {
return 1;
}
else if (tcor >= 77) {
if (Rs < 570) {
return 8;
}
else if (Rs >= 570) {
return 2;
}
}
}
}
}
}
else if (SC >= 1156) {
if (mcTcor < 95) {
if (SC < 2021) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (SC >= 2021) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (Rs < 2142) {
return 8;
}
else if (Rs >= 2142) {
if (Cs < 393) {
return 1;
}
else if (Cs >= 393) {
return 8;
}
}
}
}
}
else if (mcTcor >= 95) {
if (Rs < 798) {
return 8;
}
else if (Rs >= 798) {
return 1;
}
}
}
}
}
else if (mcTcor >= 98) {
if (Rs < 418) {
if (Rs < 362) {
return 1;
}
else if (Rs >= 362) {
return 8;
}
}
else if (Rs >= 418) {
return 1;
}
}
}
else if (CsDiff >= 4) {
if (negBalance < 52) {
if (Cs < 392) {
return 1;
}
else if (Cs >= 392) {
return 2;
}
}
else if (negBalance >= 52) {
if (negBalance < 67) {
if (negBalance < 60) {
return 8;
}
else if (negBalance >= 60) {
if (Cs < 377) {
return 8;
}
else if (Cs >= 377) {
return 1;
}
}
}
else if (negBalance >= 67) {
if (Rs < 470) {
return 8;
}
else if (Rs >= 470) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 112) {
if (MVDiff < 38) {
if (Rs < 2058) {
if (diffRsCsdiff < 82) {
if (Cs < 298) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
if (MVDiff < 26) {
if (Rs < 2048) {
return 2;
}
else if (Rs >= 2048) {
return 1;
}
}
else if (MVDiff >= 26) {
if (Cs < 93) {
return 1;
}
else if (Cs >= 93) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
return 1;
}
}
}
}
}
}
else if (Cs >= 298) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (AFD < 21) {
if (CsDiff < 4) {
if (MVDiff < 24) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (MVDiff >= 24) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 2;
}
}
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (AFD >= 21) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 82) {
return 1;
}
}
else if (Rs >= 2058) {
if (ssDCval < 79) {
if (diffMVdiffVal < -27) {
return 4;
}
else if (diffMVdiffVal >= -27) {
return 2;
}
}
else if (ssDCval >= 79) {
if (negBalance < 48) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (RsDiff < 12) {
return 1;
}
else if (RsDiff >= 12) {
return 2;
}
}
}
else if (negBalance >= 48) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 1;
}
}
}
}
}
else if (MVDiff >= 38) {
if (refDCval < 82) {
if (Cs < 389) {
return 1;
}
else if (Cs >= 389) {
if (Cs < 467) {
return 2;
}
else if (Cs >= 467) {
if (Rs < 1989) {
return 8;
}
else if (Rs >= 1989) {
return 4;
}
}
}
}
else if (refDCval >= 82) {
if (refDCval < 103) {
if (MVDiff < 70) {
if (Cs < 181) {
return 2;
}
else if (Cs >= 181) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
if (AFD < 21) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (MVDiff < 67) {
return 1;
}
else if (MVDiff >= 67) {
return 8;
}
}
}
else if (AFD >= 21) {
if (Rs < 2165) {
return 1;
}
else if (Rs >= 2165) {
return 2;
}
}
}
}
}
else if (MVDiff >= 70) {
if (refDCval < 101) {
return 2;
}
else if (refDCval >= 101) {
return 1;
}
}
}
else if (refDCval >= 103) {
if (TSC < 15) {
return 1;
}
else if (TSC >= 15) {
return 2;
}
}
}
}
}
else if (ssDCval >= 112) {
if (mcTcor < 95) {
if (SC < 1388) {
if (Rs < 876) {
return 8;
}
else if (Rs >= 876) {
if (Rs < 919) {
return 2;
}
else if (Rs >= 919) {
return 4;
}
}
}
else if (SC >= 1388) {
if (Cs < 454) {
if (negBalance < 37) {
if (Rs < 1773) {
return 2;
}
else if (Rs >= 1773) {
return 1;
}
}
else if (negBalance >= 37) {
return 1;
}
}
else if (Cs >= 454) {
if (Rs < 1851) {
if (Rs < 1437) {
return 2;
}
else if (Rs >= 1437) {
return 8;
}
}
else if (Rs >= 1851) {
return 1;
}
}
}
}
else if (mcTcor >= 95) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (diffRsCsdiff < -251) {
if (Rs < 2106) {
return 8;
}
else if (Rs >= 2106) {
return 1;
}
}
else if (diffRsCsdiff >= -251) {
if (Cs < 407) {
return 1;
}
else if (Cs >= 407) {
if (diffAFD < 3) {
return 1;
}
else if (diffAFD >= 3) {
if (diffAFD < 4) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
if (AFD < 32) {
return 1;
}
else if (AFD >= 32) {
return 2;
}
}
}
else if (diffAFD >= 4) {
return 1;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 53) {
if (posBalance < 50) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
return 1;
}
}
else if (posBalance >= 50) {
if (Rs < 1485) {
if (ssDCval < 168) {
return 1;
}
else if (ssDCval >= 168) {
return 2;
}
}
else if (Rs >= 1485) {
if (RsCsDiff < 19) {
return 4;
}
else if (RsCsDiff >= 19) {
return 2;
}
}
}
}
else if (negBalance >= 53) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 527) {
if (negBalance < 14) {
if (Rs < 702) {
if (ssDCval < 57) {
if (ssDCval < 29) {
return 8;
}
else if (ssDCval >= 29) {
if (mu_mv_mag_sq < 127) {
if (Rs < 245) {
if (Cs < 1020) {
return 2;
}
else if (Cs >= 1020) {
return 1;
}
}
else if (Rs >= 245) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 351) {
if (Rs < 288) {
return 2;
}
else if (Rs >= 288) {
return 1;
}
}
else if (Rs >= 351) {
return 8;
}
}
}
}
else if (ssDCval >= 57) {
if (ssDCval < 85) {
if (ssDCval < 77) {
if (MVDiff < 1) {
if (SC < 1152) {
return 4;
}
else if (SC >= 1152) {
return 8;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 77) {
return 2;
}
}
else if (ssDCval >= 85) {
if (negBalance < 10) {
if (SC < 1048) {
if (Rs < 398) {
return 8;
}
else if (Rs >= 398) {
if (Rs < 429) {
return 2;
}
else if (Rs >= 429) {
return 1;
}
}
}
else if (SC >= 1048) {
return 8;
}
}
else if (negBalance >= 10) {
if (refDCval < 112) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
if (Cs < 877) {
return 1;
}
else if (Cs >= 877) {
return 8;
}
}
}
else if (refDCval >= 112) {
if (posBalance < 30) {
return 1;
}
else if (posBalance >= 30) {
if (Rs < 580) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Rs < 452) {
return 4;
}
else if (Rs >= 452) {
return 2;
}
}
}
else if (Rs >= 580) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 702) {
if (diffAFD < -3) {
if (SC < 2444) {
if (tcor < 99) {
if (diffAFD < -29) {
if (Rs < 1350) {
return 4;
}
else if (Rs >= 1350) {
return 8;
}
}
else if (diffAFD >= -29) {
if (SC < 1949) {
return 8;
}
else if (SC >= 1949) {
if (Cs < 891) {
return 1;
}
else if (Cs >= 891) {
if (diffAFD < -18) {
if (Rs < 1351) {
return 1;
}
else if (Rs >= 1351) {
return 4;
}
}
else if (diffAFD >= -18) {
return 8;
}
}
}
}
}
else if (tcor >= 99) {
if (Cs < 754) {
if (ssDCval < 135) {
if (Cs < 671) {
if (Rs < 1184) {
if (Rs < 806) {
return 1;
}
else if (Rs >= 806) {
return 2;
}
}
else if (Rs >= 1184) {
return 1;
}
}
else if (Cs >= 671) {
return 8;
}
}
else if (ssDCval >= 135) {
if (Rs < 1131) {
return 8;
}
else if (Rs >= 1131) {
return 4;
}
}
}
else if (Cs >= 754) {
if (Cs < 1058) {
if (diffRsCsdiff < -33) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
return 4;
}
}
}
else if (diffRsCsdiff >= -33) {
if (diffMVdiffVal < -44) {
return 8;
}
else if (diffMVdiffVal >= -44) {
return 4;
}
}
}
else if (Cs >= 1058) {
if (Rs < 1107) {
return 1;
}
else if (Rs >= 1107) {
if (Rs < 1286) {
return 8;
}
else if (Rs >= 1286) {
return 1;
}
}
}
}
}
}
else if (SC >= 2444) {
if (diffRsCsdiff < -160) {
if (Cs < 1316) {
if (Rs < 1574) {
return 1;
}
else if (Rs >= 1574) {
if (Cs < 1310) {
return 4;
}
else if (Cs >= 1310) {
return 1;
}
}
}
else if (Cs >= 1316) {
return 8;
}
}
else if (diffRsCsdiff >= -160) {
if (MVDiff < 18) {
if (Rs < 1221) {
return 1;
}
else if (Rs >= 1221) {
return 8;
}
}
else if (MVDiff >= 18) {
if (refDCval < 128) {
return 8;
}
else if (refDCval >= 128) {
if (ssDCval < 141) {
return 4;
}
else if (ssDCval >= 141) {
return 8;
}
}
}
}
}
}
else if (diffAFD >= -3) {
if (Rs < 1816) {
if (Rs < 965) {
if (posBalance < 12) {
if (ssDCval < 152) {
if (Cs < 1101) {
return 8;
}
else if (Cs >= 1101) {
return 2;
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (posBalance >= 12) {
if (Cs < 1140) {
if (ssDCval < 73) {
if (tcor < 98) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Cs < 673) {
return 8;
}
else if (Cs >= 673) {
return 1;
}
}
}
else if (tcor >= 98) {
if (Rs < 895) {
if (ssDCval < 30) {
return 8;
}
else if (ssDCval >= 30) {
if (Rs < 882) {
return 2;
}
else if (Rs >= 882) {
if (Cs < 555) {
return 8;
}
else if (Cs >= 555) {
return 2;
}
}
}
}
else if (Rs >= 895) {
return 8;
}
}
}
else if (ssDCval >= 73) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (SC < 1435) {
return 1;
}
else if (SC >= 1435) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
}
}
else if (Cs >= 1140) {
if (refDCval < 84) {
if (Rs < 900) {
return 1;
}
else if (Rs >= 900) {
return 8;
}
}
else if (refDCval >= 84) {
return 2;
}
}
}
}
else if (Rs >= 965) {
if (SC < 1789) {
return 8;
}
else if (SC >= 1789) {
if (negBalance < 10) {
if (diffMVdiffVal < -3) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (SC < 1935) {
if (Rs < 1137) {
return 4;
}
else if (Rs >= 1137) {
return 8;
}
}
else if (SC >= 1935) {
if (ssDCval < 124) {
if (posBalance < 16) {
return 1;
}
else if (posBalance >= 16) {
if (Cs < 1048) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
return 8;
}
}
else if (Cs >= 1048) {
return 1;
}
}
}
else if (ssDCval >= 124) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -3) {
if (tcor < 95) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 1411) {
if (gchDC < 1) {
if (Rs < 1146) {
return 2;
}
else if (Rs >= 1146) {
return 8;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (Rs >= 1411) {
return 8;
}
}
}
else if (tcor >= 95) {
return 8;
}
}
}
else if (negBalance >= 10) {
if (Rs < 1441) {
if (Cs < 1048) {
if (Cs < 660) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
return 8;
}
}
else if (Cs >= 660) {
if (MVDiff < 4) {
if (diffRsCsdiff < 8) {
if (Cs < 741) {
if (Rs < 1159) {
return 8;
}
else if (Rs >= 1159) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (Cs >= 741) {
return 8;
}
}
else if (diffRsCsdiff >= 8) {
if (Rs < 1077) {
return 1;
}
else if (Rs >= 1077) {
return 2;
}
}
}
else if (MVDiff >= 4) {
if (gchDC < 2) {
if (MVDiff < 6) {
if (Cs < 704) {
return 1;
}
else if (Cs >= 704) {
if (Cs < 955) {
return 8;
}
else if (Cs >= 955) {
return 1;
}
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (gchDC >= 2) {
if (Rs < 1284) {
return 8;
}
else if (Rs >= 1284) {
return 4;
}
}
}
}
}
else if (Cs >= 1048) {
if (AFD < 4) {
if (posBalance < 27) {
return 8;
}
else if (posBalance >= 27) {
return 1;
}
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (Rs >= 1441) {
if (Rs < 1744) {
if (ssDCval < 48) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (ssDCval >= 48) {
return 8;
}
}
else if (Rs >= 1744) {
if (Cs < 1168) {
return 1;
}
else if (Cs >= 1168) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1816) {
if (Rs < 2053) {
if (SC < 2988) {
if (posBalance < 24) {
if (negBalance < 8) {
return 8;
}
else if (negBalance >= 8) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 1;
}
}
}
else if (posBalance >= 24) {
if (RsDiff < 2) {
if (Cs < 974) {
if (Rs < 1843) {
return 8;
}
else if (Rs >= 1843) {
return 1;
}
}
else if (Cs >= 974) {
if (Rs < 1838) {
return 1;
}
else if (Rs >= 1838) {
if (SC < 2876) {
return 1;
}
else if (SC >= 2876) {
if (Cs < 1077) {
return 4;
}
else if (Cs >= 1077) {
return 1;
}
}
}
}
}
else if (RsDiff >= 2) {
if (Cs < 671) {
return 2;
}
else if (Cs >= 671) {
return 8;
}
}
}
}
else if (SC >= 2988) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
return 1;
}
}
}
}
else if (Rs >= 2053) {
if (posBalance < 32) {
if (Cs < 1322) {
return 8;
}
else if (Cs >= 1322) {
if (Rs < 2295) {
return 8;
}
else if (Rs >= 2295) {
if (Cs < 1323) {
return 2;
}
else if (Cs >= 1323) {
return 8;
}
}
}
}
else if (posBalance >= 32) {
if (MVDiff < 1) {
if (SC < 3578) {
return 8;
}
else if (SC >= 3578) {
if (Cs < 1319) {
return 2;
}
else if (Cs >= 1319) {
return 8;
}
}
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 40) {
if (ssDCval < 48) {
if (MVDiff < 10) {
if (RsDiff < 3) {
if (Rs < 1423) {
if (diffMVdiffVal < -1) {
if (Rs < 885) {
return 4;
}
else if (Rs >= 885) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 777) {
return 2;
}
else if (Rs >= 777) {
return 4;
}
}
}
}
else if (Rs >= 1423) {
return 4;
}
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 43) {
if (Rs < 1482) {
if (Cs < 685) {
return 2;
}
else if (Cs >= 685) {
return 8;
}
}
else if (Rs >= 1482) {
return 4;
}
}
else if (mu_mv_mag_sq >= 43) {
if (AFD < 3) {
if (Rs < 1887) {
return 4;
}
else if (Rs >= 1887) {
return 8;
}
}
else if (AFD >= 3) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 10) {
if (SC < 1224) {
return 4;
}
else if (SC >= 1224) {
if (Cs < 1090) {
if (diffRsCsdiff < -18) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (refDCval < 46) {
return 1;
}
else if (refDCval >= 46) {
return 4;
}
}
}
else if (diffRsCsdiff >= -18) {
if (Rs < 704) {
if (Rs < 680) {
return 8;
}
else if (Rs >= 680) {
return 4;
}
}
else if (Rs >= 704) {
if (Rs < 1941) {
return 8;
}
else if (Rs >= 1941) {
return 4;
}
}
}
}
else if (Cs >= 1090) {
return 1;
}
}
}
}
else if (ssDCval >= 48) {
if (refDCval < 94) {
if (mu_mv_mag_sq < 101) {
if (negBalance < 56) {
if (Rs < 784) {
if (ssDCval < 57) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 2;
}
}
else if (ssDCval >= 57) {
if (Cs < 719) {
return 1;
}
else if (Cs >= 719) {
return 2;
}
}
}
else if (Rs >= 784) {
if (Cs < 979) {
if (mu_mv_mag_sq < 70) {
if (diffRsCsdiff < 22) {
if (SC < 3067) {
if (diffMVdiffVal < 5) {
return 1;
}
else if (diffMVdiffVal >= 5) {
if (Cs < 601) {
return 2;
}
else if (Cs >= 601) {
return 1;
}
}
}
else if (SC >= 3067) {
return 2;
}
}
else if (diffRsCsdiff >= 22) {
return 2;
}
}
else if (mu_mv_mag_sq >= 70) {
if (refDCval < 56) {
if (TSC < 3) {
if (Rs < 1181) {
return 4;
}
else if (Rs >= 1181) {
return 8;
}
}
else if (TSC >= 3) {
if (Cs < 920) {
return 1;
}
else if (Cs >= 920) {
return 8;
}
}
}
else if (refDCval >= 56) {
if (mcTcor < 95) {
if (Rs < 1494) {
return 4;
}
else if (Rs >= 1494) {
return 2;
}
}
else if (mcTcor >= 95) {
if (Rs < 1512) {
if (Cs < 761) {
return 4;
}
else if (Cs >= 761) {
return 8;
}
}
else if (Rs >= 1512) {
if (Rs < 2222) {
return 1;
}
else if (Rs >= 2222) {
if (Rs < 2271) {
return 8;
}
else if (Rs >= 2271) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 979) {
if (diffMVdiffVal < 2) {
if (Rs < 1500) {
return 8;
}
else if (Rs >= 1500) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (diffMVdiffVal >= 2) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
return 4;
}
}
}
}
}
else if (negBalance >= 56) {
if (Rs < 2153) {
if (Rs < 369) {
return 1;
}
else if (Rs >= 369) {
return 8;
}
}
else if (Rs >= 2153) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (Cs < 739) {
if (AFD < 3) {
if (Rs < 1046) {
return 8;
}
else if (Rs >= 1046) {
return 4;
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 739) {
if (Rs < 1949) {
if (Rs < 652) {
return 2;
}
else if (Rs >= 652) {
return 8;
}
}
else if (Rs >= 1949) {
if (Cs < 848) {
if (Rs < 2140) {
return 4;
}
else if (Rs >= 2140) {
return 8;
}
}
else if (Cs >= 848) {
return 2;
}
}
}
}
}
else if (refDCval >= 94) {
if (negBalance < 30) {
if (Rs < 1691) {
if (Rs < 706) {
if (Cs < 643) {
return 8;
}
else if (Cs >= 643) {
return 4;
}
}
else if (Rs >= 706) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 1;
}
}
}
else if (Rs >= 1691) {
return 4;
}
}
else if (negBalance >= 30) {
if (tcor < 92) {
if (ssDCval < 151) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
if (MVDiff < 20) {
if (Rs < 1604) {
return 2;
}
else if (Rs >= 1604) {
return 1;
}
}
else if (MVDiff >= 20) {
if (Rs < 1638) {
return 8;
}
else if (Rs >= 1638) {
return 2;
}
}
}
}
else if (ssDCval >= 151) {
return 2;
}
}
else if (tcor >= 92) {
return 8;
}
}
}
}
}
else if (posBalance >= 40) {
if (ssDCval < 156) {
if (mcTcor < 90) {
if (refDCval < 48) {
if (RsCsDiff < 31) {
if (mu_mv_mag_sq < 117) {
if (negBalance < 49) {
if (Rs < 848) {
if (mu_mv_mag_sq < 32) {
if (refDCval < 36) {
if (Rs < 364) {
return 1;
}
else if (Rs >= 364) {
return 8;
}
}
else if (refDCval >= 36) {
if (Cs < 722) {
if (TSC < 7) {
if (MVDiff < 23) {
if (Rs < 678) {
return 8;
}
else if (Rs >= 678) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (MVDiff >= 23) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (Cs >= 722) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
if (Rs < 845) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (Rs >= 845) {
return 1;
}
}
}
}
else if (Rs >= 848) {
if (TSC < 4) {
if (Cs < 691) {
return 1;
}
else if (Cs >= 691) {
return 8;
}
}
else if (TSC >= 4) {
if (RsDiff < 5) {
if (negBalance < 37) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (Cs < 578) {
return 8;
}
else if (Cs >= 578) {
return 2;
}
}
}
else if (negBalance >= 37) {
return 8;
}
}
else if (RsDiff >= 5) {
if (Rs < 1169) {
return 2;
}
else if (Rs >= 1169) {
return 1;
}
}
}
}
}
else if (negBalance >= 49) {
if (Rs < 728) {
if (TSC < 6) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
else if (TSC >= 6) {
if (Rs < 477) {
return 4;
}
else if (Rs >= 477) {
return 8;
}
}
}
else if (Rs >= 728) {
if (posBalance < 59) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (posBalance >= 59) {
if (Rs < 935) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (Rs < 826) {
return 8;
}
else if (Rs >= 826) {
return 1;
}
}
}
else if (Rs >= 935) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 31) {
if (Rs < 712) {
if (negBalance < 32) {
if (RsCsDiff < 57) {
if (Rs < 344) {
return 1;
}
else if (Rs >= 344) {
return 8;
}
}
else if (RsCsDiff >= 57) {
return 1;
}
}
else if (negBalance >= 32) {
if (negBalance < 53) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 2;
}
}
else if (negBalance >= 53) {
return 4;
}
}
}
else if (Rs >= 712) {
if (Rs < 984) {
if (CsDiff < 5) {
if (refDCval < 45) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (TSC < 5) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 2;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (refDCval >= 45) {
if (negBalance < 54) {
if (diffMVdiffVal < -1) {
if (Rs < 977) {
return 8;
}
else if (Rs >= 977) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 818) {
return 8;
}
else if (Rs >= 818) {
if (tcor < 74) {
if (Cs < 913) {
return 1;
}
else if (Cs >= 913) {
return 8;
}
}
else if (tcor >= 74) {
return 8;
}
}
}
}
else if (negBalance >= 54) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (AFD < 11) {
if (Rs < 818) {
if (Cs < 908) {
return 8;
}
else if (Cs >= 908) {
return 2;
}
}
else if (Rs >= 818) {
return 8;
}
}
else if (AFD >= 11) {
if (negBalance < 49) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (negBalance >= 49) {
return 4;
}
}
}
}
else if (Rs >= 984) {
if (CsDiff < 4) {
if (tcor < 82) {
if (mu_mv_mag_sq < 105) {
return 8;
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 1327) {
return 1;
}
else if (Rs >= 1327) {
return 8;
}
}
}
else if (tcor >= 82) {
if (TSC < 6) {
if (Rs < 1650) {
return 2;
}
else if (Rs >= 1650) {
return 4;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
}
else if (refDCval >= 48) {
if (tcor < 72) {
if (ssDCval < 52) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
if (mcTcor < 76) {
return 8;
}
else if (mcTcor >= 76) {
if (mcTcor < 81) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (mcTcor >= 81) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
}
else if (ssDCval >= 52) {
if (ssDCval < 57) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 791) {
if (Rs < 628) {
return 1;
}
else if (Rs >= 628) {
return 4;
}
}
else if (Rs >= 791) {
return 8;
}
}
}
}
else if (ssDCval >= 57) {
if (tcor < 58) {
if (TSC < 7) {
if (Rs < 790) {
return 4;
}
else if (Rs >= 790) {
if (Rs < 1275) {
return 1;
}
else if (Rs >= 1275) {
return 8;
}
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (tcor >= 58) {
if (RsCsDiff < 73) {
if (Rs < 1497) {
if (Cs < 1031) {
if (MVDiff < 17) {
if (ssDCval < 70) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
return 4;
}
}
else if (ssDCval >= 70) {
if (Rs < 1423) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
return 4;
}
}
else if (Rs >= 1423) {
return 2;
}
}
}
else if (MVDiff >= 17) {
if (diffRsCsdiff < -3) {
if (SC < 1478) {
if (Rs < 658) {
return 1;
}
else if (Rs >= 658) {
return 2;
}
}
else if (SC >= 1478) {
return 4;
}
}
else if (diffRsCsdiff >= -3) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
else if (Cs >= 1031) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (MVDiff < 28) {
return 1;
}
else if (MVDiff >= 28) {
return 8;
}
}
}
}
else if (Rs >= 1497) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
return 8;
}
}
}
else if (RsCsDiff >= 73) {
if (refDCval < 62) {
return 8;
}
else if (refDCval >= 62) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 72) {
if (posBalance < 51) {
if (mu_mv_mag_sq < 43) {
if (SC < 2040) {
if (Rs < 586) {
return 2;
}
else if (Rs >= 586) {
if (Rs < 1200) {
return 8;
}
else if (Rs >= 1200) {
return 1;
}
}
}
else if (SC >= 2040) {
if (negBalance < 38) {
if (Rs < 1478) {
return 1;
}
else if (Rs >= 1478) {
return 2;
}
}
else if (negBalance >= 38) {
if (refDCval < 70) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Cs < 1201) {
return 4;
}
else if (Cs >= 1201) {
return 8;
}
}
}
else if (refDCval >= 70) {
if (negBalance < 58) {
if (Rs < 2330) {
return 8;
}
else if (Rs >= 2330) {
return 1;
}
}
else if (negBalance >= 58) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (tcor < 79) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (Rs < 1917) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Rs >= 1917) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
return 8;
}
}
}
}
else if (tcor >= 79) {
if (Rs < 1017) {
if (tcor < 85) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
else if (tcor >= 85) {
return 8;
}
}
else if (Rs >= 1017) {
return 8;
}
}
}
}
else if (posBalance >= 51) {
if (negBalance < 45) {
if (refDCval < 75) {
if (ssDCval < 60) {
if (TSC < 7) {
if (negBalance < 30) {
if (Cs < 720) {
return 1;
}
else if (Cs >= 720) {
return 8;
}
}
else if (negBalance >= 30) {
if (refDCval < 58) {
return 8;
}
else if (refDCval >= 58) {
if (Rs < 1305) {
return 4;
}
else if (Rs >= 1305) {
return 8;
}
}
}
}
else if (TSC >= 7) {
if (Cs < 866) {
return 4;
}
else if (Cs >= 866) {
if (Rs < 1083) {
return 4;
}
else if (Rs >= 1083) {
return 8;
}
}
}
}
else if (ssDCval >= 60) {
if (RsCsDiff < 43) {
if (diffRsCsdiff < -3) {
if (Rs < 1404) {
return 8;
}
else if (Rs >= 1404) {
return 1;
}
}
else if (diffRsCsdiff >= -3) {
if (posBalance < 62) {
if (Rs < 1016) {
return 4;
}
else if (Rs >= 1016) {
if (diffAFD < 3) {
return 1;
}
else if (diffAFD >= 3) {
return 4;
}
}
}
else if (posBalance >= 62) {
if (Rs < 870) {
if (Rs < 385) {
return 1;
}
else if (Rs >= 385) {
return 2;
}
}
else if (Rs >= 870) {
if (Rs < 1457) {
return 4;
}
else if (Rs >= 1457) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 43) {
if (tcor < 74) {
return 1;
}
else if (tcor >= 74) {
if (Rs < 2137) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
return 4;
}
}
else if (Rs >= 2137) {
return 1;
}
}
}
}
}
else if (refDCval >= 75) {
if (ssDCval < 129) {
if (RsDiff < 8) {
if (MVDiff < 29) {
return 8;
}
else if (MVDiff >= 29) {
if (Rs < 564) {
return 2;
}
else if (Rs >= 564) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (Rs < 1874) {
return 2;
}
else if (Rs >= 1874) {
if (Rs < 2081) {
return 4;
}
else if (Rs >= 2081) {
return 1;
}
}
}
}
else if (ssDCval >= 129) {
if (Cs < 746) {
return 4;
}
else if (Cs >= 746) {
if (Rs < 991) {
return 2;
}
else if (Rs >= 991) {
return 1;
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 954) {
if (ssDCval < 62) {
if (TSC < 7) {
if (mu_mv_mag_sq < 10) {
if (mcTcor < 87) {
if (Rs < 872) {
return 4;
}
else if (Rs >= 872) {
return 8;
}
}
else if (mcTcor >= 87) {
return 1;
}
}
else if (mu_mv_mag_sq >= 10) {
if (mu_mv_mag_sq < 68) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 68) {
return 4;
}
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (ssDCval >= 62) {
if (Rs < 849) {
return 8;
}
else if (Rs >= 849) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 928) {
return 4;
}
else if (Rs >= 928) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
}
}
else if (Rs >= 954) {
if (RsDiff < 7) {
if (Rs < 1593) {
if (Rs < 1159) {
if (ssDCval < 62) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (ssDCval >= 62) {
return 8;
}
}
else if (Rs >= 1159) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
if (ssDCval < 99) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (Rs < 1204) {
return 2;
}
else if (Rs >= 1204) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1434) {
return 2;
}
else if (Rs >= 1434) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
}
}
else if (ssDCval >= 99) {
return 8;
}
}
}
}
else if (Rs >= 1593) {
if (Rs < 1678) {
if (AFD < 11) {
if (negBalance < 52) {
if (Rs < 1598) {
return 2;
}
else if (Rs >= 1598) {
return 8;
}
}
else if (negBalance >= 52) {
if (RsDiff < 5) {
if (Rs < 1624) {
return 2;
}
else if (Rs >= 1624) {
return 8;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (AFD >= 11) {
if (Rs < 1652) {
if (Rs < 1619) {
return 2;
}
else if (Rs >= 1619) {
return 8;
}
}
else if (Rs >= 1652) {
return 4;
}
}
}
else if (Rs >= 1678) {
return 8;
}
}
}
else if (RsDiff >= 7) {
if (Cs < 905) {
if (tcor < 85) {
if (Rs < 1483) {
return 2;
}
else if (Rs >= 1483) {
return 8;
}
}
else if (tcor >= 85) {
return 1;
}
}
else if (Cs >= 905) {
if (Cs < 1020) {
if (Rs < 1691) {
return 8;
}
else if (Rs >= 1691) {
return 2;
}
}
else if (Cs >= 1020) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (RsDiff < 7) {
if (refDCval < 75) {
if (Rs < 511) {
if (Cs < 847) {
if (Rs < 310) {
if (Cs < 546) {
return 4;
}
else if (Cs >= 546) {
return 1;
}
}
else if (Rs >= 310) {
if (Cs < 692) {
if (MVDiff < 7) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (AFD < 5) {
if (Cs < 634) {
return 2;
}
else if (Cs >= 634) {
return 4;
}
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (negBalance < 28) {
if (Rs < 390) {
if (Rs < 348) {
return 8;
}
else if (Rs >= 348) {
return 2;
}
}
else if (Rs >= 390) {
return 4;
}
}
else if (negBalance >= 28) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
if (Rs < 347) {
return 1;
}
else if (Rs >= 347) {
if (TSC < 6) {
if (Rs < 450) {
if (Rs < 397) {
return 2;
}
else if (Rs >= 397) {
return 8;
}
}
else if (Rs >= 450) {
return 2;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
}
else if (Cs >= 692) {
if (mu_mv_mag_sq < 3) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffMVdiffVal < 3) {
if (negBalance < 23) {
if (Rs < 421) {
return 4;
}
else if (Rs >= 421) {
return 8;
}
}
else if (negBalance >= 23) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
}
}
else if (Cs >= 847) {
if (negBalance < 20) {
if (Rs < 414) {
return 2;
}
else if (Rs >= 414) {
return 1;
}
}
else if (negBalance >= 20) {
if (diffRsCsdiff < 15) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (ssDCval < 53) {
if (SC < 1460) {
return 2;
}
else if (SC >= 1460) {
return 1;
}
}
else if (ssDCval >= 53) {
return 1;
}
}
}
else if (diffRsCsdiff >= 15) {
return 2;
}
}
}
}
else if (Rs >= 511) {
if (RsCsDiff < 19) {
if (Rs < 880) {
if (negBalance < 31) {
if (ssDCval < 62) {
if (negBalance < 20) {
if (Rs < 566) {
return 4;
}
else if (Rs >= 566) {
return 1;
}
}
else if (negBalance >= 20) {
if (MVDiff < 2) {
if (Rs < 709) {
if (diffMVdiffVal < -1) {
if (Rs < 583) {
return 1;
}
else if (Rs >= 583) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (Rs >= 709) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (refDCval < 46) {
if (Cs < 552) {
return 8;
}
else if (Cs >= 552) {
return 1;
}
}
else if (refDCval >= 46) {
if (RsCsDiff < 7) {
if (Rs < 657) {
return 4;
}
else if (Rs >= 657) {
return 1;
}
}
else if (RsCsDiff >= 7) {
if (Rs < 528) {
if (Rs < 525) {
return 4;
}
else if (Rs >= 525) {
return 2;
}
}
else if (Rs >= 528) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 62) {
if (Rs < 663) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Cs < 545) {
return 1;
}
else if (Cs >= 545) {
return 8;
}
}
}
else if (Rs >= 663) {
if (MVDiff < 3) {
if (Cs < 861) {
return 4;
}
else if (Cs >= 861) {
if (Rs < 728) {
return 1;
}
else if (Rs >= 728) {
return 8;
}
}
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
else if (negBalance >= 31) {
if (Cs < 877) {
if (Cs < 644) {
if (RsDiff < 3) {
if (mu_mv_mag_sq < 3) {
if (SC < 1195) {
if (ssDCval < 42) {
if (Cs < 560) {
return 8;
}
else if (Cs >= 560) {
return 2;
}
}
else if (ssDCval >= 42) {
if (Cs < 613) {
return 4;
}
else if (Cs >= 613) {
return 8;
}
}
}
else if (SC >= 1195) {
if (negBalance < 46) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 649) {
return 1;
}
else if (Rs >= 649) {
return 2;
}
}
}
else if (negBalance >= 46) {
if (Rs < 732) {
return 8;
}
else if (Rs >= 732) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (RsDiff >= 3) {
if (diffMVdiffVal < -2) {
if (ssDCval < 47) {
return 2;
}
else if (ssDCval >= 47) {
if (Rs < 836) {
return 1;
}
else if (Rs >= 836) {
return 4;
}
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 614) {
return 4;
}
else if (Cs >= 614) {
return 8;
}
}
}
}
else if (Cs >= 644) {
if (refDCval < 56) {
if (mu_mv_mag_sq < 4) {
if (tcor < 91) {
if (Cs < 701) {
return 4;
}
else if (Cs >= 701) {
return 8;
}
}
else if (tcor >= 91) {
return 4;
}
}
else if (mu_mv_mag_sq >= 4) {
if (TSC < 4) {
if (SC < 1425) {
return 4;
}
else if (SC >= 1425) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 1;
}
}
}
else if (TSC >= 4) {
if (Cs < 663) {
return 1;
}
else if (Cs >= 663) {
if (Rs < 824) {
return 4;
}
else if (Rs >= 824) {
if (Rs < 835) {
return 8;
}
else if (Rs >= 835) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 56) {
if (mu_mv_mag_sq < 23) {
return 1;
}
else if (mu_mv_mag_sq >= 23) {
if (MVDiff < 1) {
if (Cs < 808) {
if (Cs < 707) {
return 1;
}
else if (Cs >= 707) {
return 4;
}
}
else if (Cs >= 808) {
return 8;
}
}
else if (MVDiff >= 1) {
if (negBalance < 33) {
if (Rs < 698) {
return 8;
}
else if (Rs >= 698) {
if (Rs < 824) {
return 1;
}
else if (Rs >= 824) {
return 2;
}
}
}
else if (negBalance >= 33) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 877) {
if (ssDCval < 39) {
return 1;
}
else if (ssDCval >= 39) {
if (posBalance < 50) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (CsDiff < 3) {
if (Rs < 666) {
return 4;
}
else if (Rs >= 666) {
if (Rs < 708) {
return 1;
}
else if (Rs >= 708) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (Cs < 1003) {
return 2;
}
else if (Cs >= 1003) {
if (Rs < 794) {
return 8;
}
else if (Rs >= 794) {
return 2;
}
}
}
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 81) {
if (mu_mv_mag_sq < 35) {
if (Rs < 840) {
return 8;
}
else if (Rs >= 840) {
return 4;
}
}
else if (mu_mv_mag_sq >= 35) {
return 1;
}
}
else if (mu_mv_mag_sq >= 81) {
if (Rs < 872) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
return 1;
}
}
else if (Rs >= 872) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 880) {
if (refDCval < 63) {
if (RsDiff < 3) {
if (posBalance < 45) {
if (refDCval < 56) {
if (Rs < 1366) {
if (CsDiff < 2) {
if (Cs < 620) {
if (Rs < 967) {
return 8;
}
else if (Rs >= 967) {
return 1;
}
}
else if (Cs >= 620) {
return 8;
}
}
else if (CsDiff >= 2) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
return 8;
}
}
}
else if (Rs >= 1366) {
if (ssDCval < 49) {
if (Rs < 1637) {
if (Rs < 1438) {
return 4;
}
else if (Rs >= 1438) {
return 4;
}
}
else if (Rs >= 1637) {
return 4;
}
}
else if (ssDCval >= 49) {
return 8;
}
}
}
else if (refDCval >= 56) {
if (negBalance < 24) {
if (Rs < 1282) {
return 1;
}
else if (Rs >= 1282) {
return 8;
}
}
else if (negBalance >= 24) {
return 8;
}
}
}
else if (posBalance >= 45) {
if (tcor < 97) {
if (refDCval < 60) {
if (Rs < 971) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (Rs < 969) {
return 8;
}
else if (Rs >= 969) {
return 1;
}
}
}
else if (Rs >= 971) {
if (diffRsCsdiff < -11) {
if (Cs < 1167) {
return 8;
}
else if (Cs >= 1167) {
return 4;
}
}
else if (diffRsCsdiff >= -11) {
if (Rs < 1409) {
return 8;
}
else if (Rs >= 1409) {
if (Rs < 1419) {
return 2;
}
else if (Rs >= 1419) {
return 8;
}
}
}
}
}
else if (refDCval >= 60) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (ssDCval < 61) {
if (Rs < 1113) {
return 8;
}
else if (Rs >= 1113) {
if (Rs < 1291) {
return 1;
}
else if (Rs >= 1291) {
return 8;
}
}
}
else if (ssDCval >= 61) {
if (RsCsDiff < 4) {
return 1;
}
else if (RsCsDiff >= 4) {
return 8;
}
}
}
}
}
else if (tcor >= 97) {
return 1;
}
}
}
else if (RsDiff >= 3) {
if (posBalance < 42) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Rs < 1391) {
return 1;
}
else if (Rs >= 1391) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (SC < 2308) {
if (ssDCval < 56) {
if (negBalance < 36) {
if (diffAFD < 1) {
if (Rs < 1322) {
return 8;
}
else if (Rs >= 1322) {
return 2;
}
}
else if (diffAFD >= 1) {
if (SC < 1884) {
return 1;
}
else if (SC >= 1884) {
if (SC < 1983) {
return 8;
}
else if (SC >= 1983) {
return 4;
}
}
}
}
else if (negBalance >= 36) {
return 8;
}
}
else if (ssDCval >= 56) {
if (MVDiff < 13) {
if (Cs < 892) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (negBalance < 44) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (negBalance >= 44) {
return 8;
}
}
}
else if (Cs >= 892) {
if (diffAFD < 1) {
if (negBalance < 47) {
return 4;
}
else if (negBalance >= 47) {
return 8;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (MVDiff >= 13) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
return 8;
}
}
}
}
else if (SC >= 2308) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
return 8;
}
}
}
}
}
else if (refDCval >= 63) {
if (TSC < 6) {
if (Rs < 1067) {
if (mu_mv_mag_sq < 83) {
if (Cs < 691) {
if (Cs < 600) {
return 1;
}
else if (Cs >= 600) {
return 4;
}
}
else if (Cs >= 691) {
return 8;
}
}
else if (mu_mv_mag_sq >= 83) {
return 4;
}
}
else if (Rs >= 1067) {
if (RsDiff < 3) {
if (negBalance < 38) {
if (mu_mv_mag_sq < 48) {
return 1;
}
else if (mu_mv_mag_sq >= 48) {
if (MVDiff < 19) {
if (diffRsCsdiff < -9) {
if (Rs < 1339) {
return 4;
}
else if (Rs >= 1339) {
return 8;
}
}
else if (diffRsCsdiff >= -9) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
return 1;
}
}
}
else if (MVDiff >= 19) {
return 4;
}
}
}
else if (negBalance >= 38) {
if (Cs < 1160) {
if (posBalance < 54) {
if (Cs < 865) {
return 1;
}
else if (Cs >= 865) {
if (Cs < 1005) {
return 4;
}
else if (Cs >= 1005) {
return 8;
}
}
}
else if (posBalance >= 54) {
return 8;
}
}
else if (Cs >= 1160) {
return 1;
}
}
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 105) {
if (ssDCval < 73) {
if (ssDCval < 71) {
if (posBalance < 48) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (posBalance >= 48) {
if (Cs < 797) {
if (Cs < 715) {
return 4;
}
else if (Cs >= 715) {
return 8;
}
}
else if (Cs >= 797) {
if (diffAFD < -1) {
if (ssDCval < 69) {
if (diffRsCsdiff < -4) {
return 1;
}
else if (diffRsCsdiff >= -4) {
if (Rs < 1283) {
return 4;
}
else if (Rs >= 1283) {
return 8;
}
}
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
else if (ssDCval >= 71) {
return 4;
}
}
else if (ssDCval >= 73) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 1320) {
return 4;
}
else if (Rs >= 1320) {
if (diffTSC < 1) {
if (Rs < 1669) {
return 8;
}
else if (Rs >= 1669) {
return 1;
}
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
}
}
else if (TSC >= 6) {
if (MVDiff < 4) {
if (negBalance < 44) {
if (Rs < 1532) {
return 1;
}
else if (Rs >= 1532) {
return 8;
}
}
else if (negBalance >= 44) {
return 8;
}
}
else if (MVDiff >= 4) {
if (Rs < 1331) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (Rs >= 1331) {
if (Rs < 1739) {
return 8;
}
else if (Rs >= 1739) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 19) {
if (MVDiff < 7) {
if (refDCval < 62) {
if (negBalance < 32) {
if (ssDCval < 49) {
if (Cs < 1217) {
return 4;
}
else if (Cs >= 1217) {
return 8;
}
}
else if (ssDCval >= 49) {
if (SC < 2690) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (SC < 1553) {
return 4;
}
else if (SC >= 1553) {
return 1;
}
}
}
else if (SC >= 2690) {
return 8;
}
}
}
else if (negBalance >= 32) {
if (Rs < 1326) {
if (Cs < 1030) {
if (tcor < 77) {
if (Cs < 606) {
return 8;
}
else if (Cs >= 606) {
return 4;
}
}
else if (tcor >= 77) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (TSC < 4) {
if (Rs < 1166) {
if (Rs < 1028) {
return 1;
}
else if (Rs >= 1028) {
return 2;
}
}
else if (Rs >= 1166) {
return 8;
}
}
else if (TSC >= 4) {
if (Cs < 992) {
return 8;
}
else if (Cs >= 992) {
if (Rs < 1161) {
return 8;
}
else if (Rs >= 1161) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1030) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (negBalance < 44) {
if (diffAFD < 1) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (negBalance >= 44) {
return 4;
}
}
}
}
else if (Rs >= 1326) {
if (posBalance < 46) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (posBalance >= 46) {
if (diffMVdiffVal < -5) {
return 2;
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
}
}
}
else if (refDCval >= 62) {
if (negBalance < 44) {
if (Cs < 1076) {
if (mcTcor < 96) {
if (posBalance < 52) {
if (mu_mv_mag_sq < 64) {
if (RsCsDiff < 29) {
return 4;
}
else if (RsCsDiff >= 29) {
return 1;
}
}
else if (mu_mv_mag_sq >= 64) {
return 2;
}
}
else if (posBalance >= 52) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1679) {
if (refDCval < 71) {
return 8;
}
else if (refDCval >= 71) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (Rs >= 1679) {
if (Cs < 822) {
return 2;
}
else if (Cs >= 822) {
return 1;
}
}
}
}
else if (Cs >= 1076) {
if (diffAFD < 1) {
if (AFD < 8) {
if (MVDiff < 1) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
else if (MVDiff >= 1) {
if (SC < 2144) {
return 8;
}
else if (SC >= 2144) {
return 4;
}
}
}
else if (AFD >= 8) {
return 8;
}
}
else if (diffAFD >= 1) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
if (RsCsDiff < 38) {
if (Rs < 915) {
return 2;
}
else if (Rs >= 915) {
return 4;
}
}
else if (RsCsDiff >= 38) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 44) {
if (negBalance < 50) {
if (Cs < 1038) {
if (RsDiff < 5) {
if (Rs < 1729) {
return 8;
}
else if (Rs >= 1729) {
return 2;
}
}
else if (RsDiff >= 5) {
if (diffRsCsdiff < 14) {
if (diffMVdiffVal < 3) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 1517) {
return 2;
}
else if (Rs >= 1517) {
return 8;
}
}
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
else if (diffRsCsdiff >= 14) {
return 8;
}
}
}
else if (Cs >= 1038) {
if (TSC < 8) {
if (ssDCval < 72) {
if (Rs < 847) {
return 1;
}
else if (Rs >= 847) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 1486) {
return 4;
}
else if (Rs >= 1486) {
return 8;
}
}
}
else if (TSC >= 8) {
if (Rs < 1694) {
return 8;
}
else if (Rs >= 1694) {
return 2;
}
}
}
}
else if (negBalance >= 50) {
if (tcor < 89) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Cs < 975) {
return 4;
}
else if (Cs >= 975) {
return 1;
}
}
}
else if (tcor >= 89) {
if (Rs < 1508) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
else if (Rs >= 1508) {
if (Rs < 1948) {
return 2;
}
else if (Rs >= 1948) {
if (Rs < 2049) {
return 4;
}
else if (Rs >= 2049) {
return 1;
}
}
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (negBalance < 47) {
if (mcTcor < 93) {
if (refDCval < 53) {
if (refDCval < 42) {
if (Rs < 619) {
return 1;
}
else if (Rs >= 619) {
return 2;
}
}
else if (refDCval >= 42) {
if (SC < 1653) {
if (Rs < 682) {
return 8;
}
else if (Rs >= 682) {
return 4;
}
}
else if (SC >= 1653) {
return 8;
}
}
}
else if (refDCval >= 53) {
if (SC < 1874) {
if (refDCval < 74) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (refDCval >= 74) {
return 4;
}
}
else if (SC >= 1874) {
if (CsDiff < 5) {
if (diffTSC < 1) {
if (MVDiff < 27) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (Rs < 1819) {
return 1;
}
else if (Rs >= 1819) {
return 4;
}
}
}
else if (MVDiff >= 27) {
return 8;
}
}
else if (diffTSC >= 1) {
if (AFD < 10) {
if (Cs < 785) {
return 2;
}
else if (Cs >= 785) {
return 8;
}
}
else if (AFD >= 10) {
if (Rs < 2165) {
return 4;
}
else if (Rs >= 2165) {
return 1;
}
}
}
}
else if (CsDiff >= 5) {
if (diffMVdiffVal < -4) {
return 2;
}
else if (diffMVdiffVal >= -4) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 64) {
if (AFD < 8) {
if (Cs < 973) {
if (negBalance < 34) {
if (Rs < 1020) {
if (Rs < 780) {
return 8;
}
else if (Rs >= 780) {
return 4;
}
}
else if (Rs >= 1020) {
return 1;
}
}
else if (negBalance >= 34) {
if (Cs < 833) {
return 4;
}
else if (Cs >= 833) {
return 2;
}
}
}
else if (Cs >= 973) {
if (Rs < 1989) {
return 4;
}
else if (Rs >= 1989) {
return 8;
}
}
}
else if (AFD >= 8) {
if (TSC < 4) {
if (Rs < 920) {
return 4;
}
else if (Rs >= 920) {
return 1;
}
}
else if (TSC >= 4) {
if (SC < 1426) {
if (negBalance < 45) {
if (tcor < 86) {
return 4;
}
else if (tcor >= 86) {
return 2;
}
}
else if (negBalance >= 45) {
return 1;
}
}
else if (SC >= 1426) {
if (negBalance < 46) {
if (mu_mv_mag_sq < 66) {
if (mu_mv_mag_sq < 3) {
if (Cs < 1134) {
if (diffMVdiffVal < -7) {
return 4;
}
else if (diffMVdiffVal >= -7) {
return 2;
}
}
else if (Cs >= 1134) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1929) {
if (ssDCval < 46) {
if (Rs < 950) {
return 2;
}
else if (Rs >= 950) {
return 8;
}
}
else if (ssDCval >= 46) {
return 2;
}
}
else if (Rs >= 1929) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (SC < 2557) {
return 4;
}
else if (SC >= 2557) {
return 2;
}
}
}
else if (negBalance >= 46) {
if (Cs < 546) {
return 4;
}
else if (Cs >= 546) {
if (Rs < 1288) {
return 8;
}
else if (Rs >= 1288) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (Rs < 899) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (Rs >= 899) {
if (posBalance < 64) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (diffRsCsdiff < -50) {
return 2;
}
else if (diffRsCsdiff >= -50) {
if (posBalance < 55) {
if (Rs < 2109) {
return 1;
}
else if (Rs >= 2109) {
return 4;
}
}
else if (posBalance >= 55) {
if (posBalance < 58) {
if (Rs < 2087) {
return 8;
}
else if (Rs >= 2087) {
return 1;
}
}
else if (posBalance >= 58) {
if (TSC < 6) {
if (RsCsDiff < 23) {
if (Rs < 1518) {
return 4;
}
else if (Rs >= 1518) {
return 1;
}
}
else if (RsCsDiff >= 23) {
return 2;
}
}
else if (TSC >= 6) {
if (Cs < 1268) {
return 1;
}
else if (Cs >= 1268) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 64) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 47) {
if (AFD < 11) {
if (ssDCval < 64) {
if (Rs < 1317) {
if (SC < 1973) {
if (Cs < 589) {
return 2;
}
else if (Cs >= 589) {
return 8;
}
}
else if (SC >= 1973) {
if (Cs < 1203) {
return 4;
}
else if (Cs >= 1203) {
return 1;
}
}
}
else if (Rs >= 1317) {
return 8;
}
}
else if (ssDCval >= 64) {
if (mu_mv_mag_sq < 49) {
if (Cs < 637) {
return 8;
}
else if (Cs >= 637) {
return 4;
}
}
else if (mu_mv_mag_sq >= 49) {
if (mu_mv_mag_sq < 50) {
if (Rs < 1570) {
return 4;
}
else if (Rs >= 1570) {
return 2;
}
}
else if (mu_mv_mag_sq >= 50) {
return 8;
}
}
}
}
else if (AFD >= 11) {
if (diffAFD < -1) {
if (Cs < 1289) {
if (Cs < 549) {
return 4;
}
else if (Cs >= 549) {
return 8;
}
}
else if (Cs >= 1289) {
if (Rs < 915) {
return 2;
}
else if (Rs >= 915) {
return 4;
}
}
}
else if (diffAFD >= -1) {
if (tcor < 77) {
return 2;
}
else if (tcor >= 77) {
if (posBalance < 54) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (posBalance >= 54) {
if (Rs < 2005) {
if (Rs < 1045) {
if (SC < 1572) {
return 8;
}
else if (SC >= 1572) {
return 1;
}
}
else if (Rs >= 1045) {
return 8;
}
}
else if (Rs >= 2005) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (mu_mv_mag_sq < 41) {
if (MVDiff < 12) {
if (negBalance < 45) {
if (posBalance < 51) {
if (Rs < 964) {
if (negBalance < 41) {
if (Rs < 838) {
return 2;
}
else if (Rs >= 838) {
return 1;
}
}
else if (negBalance >= 41) {
return 8;
}
}
else if (Rs >= 964) {
return 1;
}
}
else if (posBalance >= 51) {
if (negBalance < 36) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (Cs < 1000) {
return 8;
}
else if (Cs >= 1000) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (ssDCval < 104) {
if (tcor < 83) {
return 8;
}
else if (tcor >= 83) {
if (Rs < 1729) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (Rs < 999) {
return 8;
}
else if (Rs >= 999) {
return 1;
}
}
}
else if (Rs >= 1729) {
return 4;
}
}
}
else if (ssDCval >= 104) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 2;
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 422) {
return 1;
}
else if (Rs >= 422) {
if (posBalance < 53) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (MVDiff < 8) {
if (SC < 1599) {
if (refDCval < 86) {
return 1;
}
else if (refDCval >= 86) {
if (diffMVdiffVal < -2) {
if (Rs < 581) {
return 8;
}
else if (Rs >= 581) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
}
else if (SC >= 1599) {
if (Rs < 1148) {
if (Rs < 887) {
return 4;
}
else if (Rs >= 887) {
return 8;
}
}
else if (Rs >= 1148) {
return 1;
}
}
}
else if (MVDiff >= 8) {
if (Rs < 1657) {
return 8;
}
else if (Rs >= 1657) {
return 4;
}
}
}
}
else if (posBalance >= 53) {
if (diffAFD < -1) {
if (SC < 2664) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
}
else if (SC >= 2664) {
return 8;
}
}
else if (diffAFD >= -1) {
if (diffAFD < 1) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
return 8;
}
}
else if (diffAFD >= 1) {
if (mcTcor < 95) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
if (Rs < 775) {
return 1;
}
else if (Rs >= 775) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (TSC < 7) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
if (Rs < 1284) {
return 1;
}
else if (Rs >= 1284) {
return 8;
}
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 12) {
if (ssDCval < 136) {
if (AFD < 9) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (Rs < 1686) {
if (posBalance < 67) {
if (Cs < 540) {
return 2;
}
else if (Cs >= 540) {
return 8;
}
}
else if (posBalance >= 67) {
return 1;
}
}
else if (Rs >= 1686) {
if (Cs < 684) {
return 1;
}
else if (Cs >= 684) {
return 2;
}
}
}
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 3) {
if (TSC < 9) {
if (Cs < 687) {
if (posBalance < 52) {
if (Cs < 643) {
return 4;
}
else if (Cs >= 643) {
if (Rs < 1195) {
return 2;
}
else if (Rs >= 1195) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (Rs < 1412) {
return 8;
}
else if (Rs >= 1412) {
return 2;
}
}
}
else if (Cs >= 687) {
if (gchDC < 2) {
if (diffMVdiffVal < 5) {
if (diffAFD < -1) {
if (MVDiff < 15) {
if (Rs < 1005) {
return 2;
}
else if (Rs >= 1005) {
return 8;
}
}
else if (MVDiff >= 15) {
return 4;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (diffMVdiffVal >= 5) {
if (Rs < 1074) {
if (Rs < 874) {
return 2;
}
else if (Rs >= 874) {
return 4;
}
}
else if (Rs >= 1074) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (Rs < 819) {
return 1;
}
else if (Rs >= 819) {
return 2;
}
}
}
}
else if (TSC >= 9) {
if (tcor < 90) {
return 1;
}
else if (tcor >= 90) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 88) {
if (diffMVdiffVal < -3) {
if (Rs < 1251) {
return 8;
}
else if (Rs >= 1251) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 1;
}
}
}
else if (ssDCval >= 88) {
if (Rs < 565) {
if (Rs < 394) {
return 2;
}
else if (Rs >= 394) {
return 1;
}
}
else if (Rs >= 565) {
if (Rs < 1259) {
if (Rs < 790) {
return 8;
}
else if (Rs >= 790) {
return 2;
}
}
else if (Rs >= 1259) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 136) {
if (TSC < 6) {
if (SC < 2434) {
return 4;
}
else if (SC >= 2434) {
if (posBalance < 51) {
if (Rs < 1412) {
return 2;
}
else if (Rs >= 1412) {
return 4;
}
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (TSC >= 6) {
if (Rs < 1295) {
if (Cs < 695) {
return 2;
}
else if (Cs >= 695) {
return 1;
}
}
else if (Rs >= 1295) {
if (refDCval < 139) {
return 4;
}
else if (refDCval >= 139) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (Rs < 382) {
if (ssDCval < 96) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (ssDCval >= 96) {
if (Cs < 653) {
return 2;
}
else if (Cs >= 653) {
return 4;
}
}
}
else if (Rs >= 382) {
if (ssDCval < 114) {
if (Rs < 1749) {
if (RsCsDiff < 7) {
if (TSC < 6) {
if (Cs < 627) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 8;
}
}
else if (Cs >= 627) {
if (Cs < 871) {
if (mu_mv_mag_sq < 90) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
if (MVDiff < 3) {
if (ssDCval < 91) {
if (Rs < 686) {
return 8;
}
else if (Rs >= 686) {
return 4;
}
}
else if (ssDCval >= 91) {
return 1;
}
}
else if (MVDiff >= 3) {
if (Rs < 431) {
return 1;
}
else if (Rs >= 431) {
if (TSC < 5) {
if (Rs < 502) {
return 8;
}
else if (Rs >= 502) {
return 2;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (Rs < 454) {
return 1;
}
else if (Rs >= 454) {
if (Cs < 835) {
return 2;
}
else if (Cs >= 835) {
return 1;
}
}
}
}
else if (Cs >= 871) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
else if (TSC >= 6) {
if (Rs < 424) {
return 4;
}
else if (Rs >= 424) {
return 1;
}
}
}
else if (RsCsDiff >= 7) {
if (RsDiff < 5) {
if (Cs < 1024) {
if (posBalance < 54) {
if (tcor < 92) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (posBalance < 41) {
return 2;
}
else if (posBalance >= 41) {
return 8;
}
}
}
}
else if (tcor >= 92) {
if (ssDCval < 78) {
if (Cs < 745) {
if (Rs < 1156) {
return 8;
}
else if (Rs >= 1156) {
return 1;
}
}
else if (Cs >= 745) {
return 2;
}
}
else if (ssDCval >= 78) {
if (diffAFD < 1) {
if (posBalance < 46) {
if (Cs < 856) {
return 1;
}
else if (Cs >= 856) {
return 8;
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (diffAFD >= 1) {
if (MVDiff < 2) {
if (Rs < 983) {
return 1;
}
else if (Rs >= 983) {
return 8;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
}
else if (posBalance >= 54) {
if (tcor < 87) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
else if (tcor >= 87) {
if (diffAFD < -2) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
else if (Cs >= 1024) {
if (ssDCval < 78) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (ssDCval >= 78) {
if (AFD < 10) {
if (RsDiff < 3) {
if (AFD < 7) {
if (mu_mv_mag_sq < 78) {
return 4;
}
else if (mu_mv_mag_sq >= 78) {
return 8;
}
}
else if (AFD >= 7) {
if (mu_mv_mag_sq < 92) {
if (posBalance < 51) {
if (Rs < 935) {
return 8;
}
else if (Rs >= 935) {
return 4;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (mu_mv_mag_sq >= 92) {
return 4;
}
}
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (AFD >= 10) {
if (negBalance < 41) {
if (Rs < 1129) {
return 1;
}
else if (Rs >= 1129) {
return 4;
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
}
}
else if (RsDiff >= 5) {
if (ssDCval < 108) {
if (ssDCval < 97) {
if (mcTcor < 96) {
if (Rs < 1079) {
return 2;
}
else if (Rs >= 1079) {
return 8;
}
}
else if (mcTcor >= 96) {
if (ssDCval < 88) {
return 8;
}
else if (ssDCval >= 88) {
return 1;
}
}
}
else if (ssDCval >= 97) {
if (RsCsDiff < 46) {
return 8;
}
else if (RsCsDiff >= 46) {
return 2;
}
}
}
else if (ssDCval >= 108) {
return 1;
}
}
}
}
else if (Rs >= 1749) {
if (refDCval < 78) {
if (negBalance < 48) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
return 8;
}
}
else if (negBalance >= 48) {
return 8;
}
}
else if (refDCval >= 78) {
if (TSC < 10) {
if (SC < 3054) {
if (RsCsDiff < 19) {
if (Cs < 1208) {
return 8;
}
else if (Cs >= 1208) {
return 2;
}
}
else if (RsCsDiff >= 19) {
if (tcor < 91) {
if (Cs < 1156) {
return 8;
}
else if (Cs >= 1156) {
return 2;
}
}
else if (tcor >= 91) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (gchDC < 1) {
if (negBalance < 45) {
if (Rs < 1754) {
return 1;
}
else if (Rs >= 1754) {
return 8;
}
}
else if (negBalance >= 45) {
return 1;
}
}
else if (gchDC >= 1) {
if (Rs < 1754) {
return 4;
}
else if (Rs >= 1754) {
return 1;
}
}
}
}
}
}
else if (SC >= 3054) {
if (AFD < 8) {
if (negBalance < 45) {
if (mu_mv_mag_sq < 88) {
if (Cs < 1300) {
if (Cs < 1204) {
return 1;
}
else if (Cs >= 1204) {
return 8;
}
}
else if (Cs >= 1300) {
return 2;
}
}
else if (mu_mv_mag_sq >= 88) {
if (SC < 3112) {
if (Rs < 1878) {
return 8;
}
else if (Rs >= 1878) {
return 2;
}
}
else if (SC >= 3112) {
return 8;
}
}
}
else if (negBalance >= 45) {
return 2;
}
}
else if (AFD >= 8) {
if (tcor < 86) {
return 1;
}
else if (tcor >= 86) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (Rs < 1820) {
return 8;
}
else if (Rs >= 1820) {
return 1;
}
}
}
}
}
else if (ssDCval >= 114) {
if (SC < 2460) {
if (MVDiff < 24) {
if (tcor < 86) {
if (Rs < 1111) {
return 8;
}
else if (Rs >= 1111) {
return 4;
}
}
else if (tcor >= 86) {
if (refDCval < 118) {
if (Cs < 940) {
return 8;
}
else if (Cs >= 940) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (refDCval >= 118) {
if (Rs < 1007) {
return 8;
}
else if (Rs >= 1007) {
if (mu_mv_mag_sq < 116) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 116) {
return 4;
}
}
}
}
}
else if (MVDiff >= 24) {
if (Cs < 1048) {
return 4;
}
else if (Cs >= 1048) {
return 8;
}
}
}
else if (SC >= 2460) {
if (mu_mv_mag_sq < 121) {
if (Rs < 2355) {
if (negBalance < 27) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 1941) {
return 2;
}
else if (Rs >= 1941) {
if (Rs < 2232) {
return 4;
}
else if (Rs >= 2232) {
return 8;
}
}
}
}
else if (negBalance >= 27) {
if (SC < 2496) {
if (SC < 2486) {
return 8;
}
else if (SC >= 2486) {
return 4;
}
}
else if (SC >= 2496) {
return 8;
}
}
}
else if (Rs >= 2355) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Cs < 1194) {
return 8;
}
else if (Cs >= 1194) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (Cs < 878) {
if (ssDCval < 76) {
if (tcor < 86) {
if (posBalance < 42) {
if (Rs < 1815) {
return 1;
}
else if (Rs >= 1815) {
return 4;
}
}
else if (posBalance >= 42) {
if (Cs < 567) {
return 8;
}
else if (Cs >= 567) {
if (refDCval < 57) {
if (Rs < 1607) {
return 8;
}
else if (Rs >= 1607) {
if (Rs < 2152) {
return 4;
}
else if (Rs >= 2152) {
return 2;
}
}
}
else if (refDCval >= 57) {
if (Cs < 812) {
return 2;
}
else if (Cs >= 812) {
if (Rs < 2257) {
return 8;
}
else if (Rs >= 2257) {
return 2;
}
}
}
}
}
}
else if (tcor >= 86) {
if (SC < 2353) {
if (Cs < 609) {
return 1;
}
else if (Cs >= 609) {
if (Rs < 1252) {
return 8;
}
else if (Rs >= 1252) {
return 2;
}
}
}
else if (SC >= 2353) {
return 2;
}
}
}
else if (ssDCval >= 76) {
if (refDCval < 94) {
if (diffMVdiffVal < -12) {
return 4;
}
else if (diffMVdiffVal >= -12) {
if (Cs < 841) {
if (Rs < 2315) {
return 1;
}
else if (Rs >= 2315) {
return 2;
}
}
else if (Cs >= 841) {
if (Rs < 2248) {
return 4;
}
else if (Rs >= 2248) {
return 2;
}
}
}
}
else if (refDCval >= 94) {
if (diffAFD < -1) {
if (diffTSC < -1) {
if (Rs < 1533) {
return 8;
}
else if (Rs >= 1533) {
if (Rs < 1943) {
return 2;
}
else if (Rs >= 1943) {
return 1;
}
}
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (diffAFD >= -1) {
if (posBalance < 57) {
if (Cs < 628) {
return 1;
}
else if (Cs >= 628) {
return 8;
}
}
else if (posBalance >= 57) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
}
}
else if (Cs >= 878) {
if (refDCval < 74) {
if (RsDiff < 8) {
if (ssDCval < 61) {
if (Rs < 1494) {
return 4;
}
else if (Rs >= 1494) {
return 8;
}
}
else if (ssDCval >= 61) {
if (Rs < 1896) {
return 8;
}
else if (Rs >= 1896) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
else if (RsDiff >= 8) {
if (Rs < 1928) {
if (Rs < 1518) {
if (Rs < 1398) {
return 4;
}
else if (Rs >= 1398) {
return 1;
}
}
else if (Rs >= 1518) {
return 8;
}
}
else if (Rs >= 1928) {
if (mcTcor < 93) {
if (Cs < 1135) {
return 4;
}
else if (Cs >= 1135) {
return 2;
}
}
else if (mcTcor >= 93) {
return 2;
}
}
}
}
else if (refDCval >= 74) {
if (posBalance < 49) {
if (MVDiff < 21) {
if (Rs < 2052) {
return 8;
}
else if (Rs >= 2052) {
return 4;
}
}
else if (MVDiff >= 21) {
return 2;
}
}
else if (posBalance >= 49) {
if (TSC < 11) {
if (gchDC < 3) {
if (Rs < 1749) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (Rs >= 1749) {
if (Rs < 2107) {
if (ssDCval < 87) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (ssDCval >= 87) {
if (ssDCval < 120) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (ssDCval >= 120) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 2;
}
}
}
}
else if (Rs >= 2107) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (Rs < 1916) {
if (Rs < 1674) {
return 8;
}
else if (Rs >= 1674) {
return 4;
}
}
else if (Rs >= 1916) {
return 1;
}
}
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 156) {
if (AFD < 10) {
if (Rs < 778) {
return 4;
}
else if (Rs >= 778) {
return 2;
}
}
else if (AFD >= 10) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
if (Rs < 956) {
return 4;
}
else if (Rs >= 956) {
return 2;
}
}
}
}
}
}
else if (AFD >= 13) {
if (MVDiff < 16) {
if (tcor < 75) {
if (ssDCval < 88) {
if (posBalance < 55) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (diffRsCsdiff < -68) {
return 4;
}
else if (diffRsCsdiff >= -68) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 4;
}
}
}
}
else if (posBalance >= 55) {
if (refDCval < 67) {
if (SC < 2631) {
if (mu_mv_mag_sq < 9) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1213) {
return 4;
}
else if (Rs >= 1213) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (SC >= 2631) {
if (Rs < 2309) {
return 8;
}
else if (Rs >= 2309) {
return 2;
}
}
}
else if (refDCval >= 67) {
if (posBalance < 67) {
if (mcTcor < 94) {
if (Rs < 2115) {
return 8;
}
else if (Rs >= 2115) {
return 2;
}
}
else if (mcTcor >= 94) {
if (Rs < 613) {
return 2;
}
else if (Rs >= 613) {
return 4;
}
}
}
else if (posBalance >= 67) {
if (diffMVdiffVal < 0) {
return 4;
}
else if (diffMVdiffVal >= 0) {
if (Cs < 917) {
return 8;
}
else if (Cs >= 917) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (negBalance < 58) {
if (AFD < 16) {
if (mu_mv_mag_sq < 3) {
if (RsDiff < 5) {
if (Cs < 931) {
return 8;
}
else if (Cs >= 931) {
return 4;
}
}
else if (RsDiff >= 5) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1892) {
if (Rs < 750) {
return 1;
}
else if (Rs >= 750) {
return 8;
}
}
else if (Rs >= 1892) {
return 2;
}
}
}
else if (AFD >= 16) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1550) {
return 4;
}
else if (Rs >= 1550) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 8) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (Rs < 1341) {
return 4;
}
else if (Rs >= 1341) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (TSC < 10) {
if (posBalance < 60) {
if (Rs < 1484) {
return 8;
}
else if (Rs >= 1484) {
return 2;
}
}
else if (posBalance >= 60) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
else if (negBalance >= 58) {
if (Rs < 886) {
if (Rs < 600) {
return 2;
}
else if (Rs >= 600) {
if (Rs < 842) {
return 1;
}
else if (Rs >= 842) {
return 8;
}
}
}
else if (Rs >= 886) {
if (mcTcor < 81) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 4;
}
}
else if (mcTcor >= 81) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
return 4;
}
}
}
}
}
}
else if (tcor >= 75) {
if (RsDiff < 10) {
if (Rs < 528) {
return 2;
}
else if (Rs >= 528) {
if (diffAFD < -3) {
if (Rs < 1144) {
return 2;
}
else if (Rs >= 1144) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2113) {
return 8;
}
else if (Rs >= 2113) {
return 2;
}
}
}
}
else if (diffAFD >= -3) {
if (TSC < 9) {
if (SC < 2224) {
if (AFD < 18) {
if (ssDCval < 72) {
if (ssDCval < 68) {
if (diffMVdiffVal < -3) {
if (Rs < 932) {
return 1;
}
else if (Rs >= 932) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (ssDCval >= 68) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 8;
}
}
}
else if (ssDCval >= 72) {
if (diffMVdiffVal < -8) {
if (Rs < 781) {
return 8;
}
else if (Rs >= 781) {
if (Rs < 886) {
return 4;
}
else if (Rs >= 886) {
return 1;
}
}
}
else if (diffMVdiffVal >= -8) {
if (Cs < 600) {
if (RsCsDiff < 40) {
return 8;
}
else if (RsCsDiff >= 40) {
if (Rs < 1274) {
return 1;
}
else if (Rs >= 1274) {
return 2;
}
}
}
else if (Cs >= 600) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
}
else if (AFD >= 18) {
if (Rs < 781) {
if (Rs < 674) {
return 1;
}
else if (Rs >= 674) {
return 8;
}
}
else if (Rs >= 781) {
if (Rs < 893) {
return 2;
}
else if (Rs >= 893) {
return 4;
}
}
}
}
else if (SC >= 2224) {
if (mu_mv_mag_sq < 2) {
if (CsDiff < 6) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 4;
}
}
}
else if (CsDiff >= 6) {
if (Rs < 1821) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
else if (Rs >= 1821) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (mu_mv_mag_sq < 41) {
if (negBalance < 58) {
if (Rs < 1688) {
if (RsDiff < 6) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (Rs >= 1688) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (negBalance >= 58) {
return 2;
}
}
else if (mu_mv_mag_sq >= 41) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (Rs < 1516) {
return 8;
}
else if (Rs >= 1516) {
return 1;
}
}
}
}
}
}
else if (TSC >= 9) {
if (Rs < 1375) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (ssDCval < 89) {
if (AFD < 18) {
if (Rs < 1316) {
if (Cs < 1320) {
return 2;
}
else if (Cs >= 1320) {
return 1;
}
}
else if (Rs >= 1316) {
return 8;
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (ssDCval >= 89) {
if (ssDCval < 127) {
return 8;
}
else if (ssDCval >= 127) {
return 1;
}
}
}
}
}
else if (Rs >= 1375) {
if (Cs < 763) {
return 2;
}
else if (Cs >= 763) {
if (Rs < 2313) {
return 8;
}
else if (Rs >= 2313) {
if (Rs < 2337) {
if (Rs < 2335) {
return 4;
}
else if (Rs >= 2335) {
return 2;
}
}
else if (Rs >= 2337) {
return 8;
}
}
}
}
}
}
}
}
else if (RsDiff >= 10) {
if (Cs < 1040) {
if (refDCval < 83) {
return 2;
}
else if (refDCval >= 83) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
if (SC < 2642) {
return 2;
}
else if (SC >= 2642) {
if (SC < 2785) {
return 1;
}
else if (SC >= 2785) {
return 2;
}
}
}
}
}
else if (Cs >= 1040) {
if (TSC < 9) {
if (diffAFD < -2) {
if (Rs < 1882) {
return 8;
}
else if (Rs >= 1882) {
return 4;
}
}
else if (diffAFD >= -2) {
if (Cs < 1239) {
return 2;
}
else if (Cs >= 1239) {
return 8;
}
}
}
else if (TSC >= 9) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 16) {
if (refDCval < 90) {
if (mcTcor < 90) {
if (Rs < 641) {
if (TSC < 8) {
if (tcor < 72) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (RsCsDiff < 26) {
return 2;
}
else if (RsCsDiff >= 26) {
if (mcTcor < 88) {
return 4;
}
else if (mcTcor >= 88) {
if (Rs < 404) {
return 4;
}
else if (Rs >= 404) {
return 1;
}
}
}
}
}
else if (tcor >= 72) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (Rs < 301) {
return 4;
}
else if (Rs >= 301) {
return 2;
}
}
}
}
else if (TSC >= 8) {
if (tcor < 69) {
if (tcor < 60) {
if (Rs < 562) {
if (ssDCval < 70) {
if (negBalance < 53) {
if (diffRsCsdiff < -63) {
return 8;
}
else if (diffRsCsdiff >= -63) {
return 2;
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (ssDCval >= 70) {
if (AFD < 37) {
if (gchDC < 2) {
if (Rs < 467) {
return 4;
}
else if (Rs >= 467) {
return 2;
}
}
else if (gchDC >= 2) {
if (Cs < 953) {
return 4;
}
else if (Cs >= 953) {
return 8;
}
}
}
else if (AFD >= 37) {
return 8;
}
}
}
else if (Rs >= 562) {
if (MVDiff < 103) {
if (Rs < 626) {
return 4;
}
else if (Rs >= 626) {
return 8;
}
}
else if (MVDiff >= 103) {
if (Rs < 624) {
return 1;
}
else if (Rs >= 624) {
return 2;
}
}
}
}
else if (tcor >= 60) {
if (SC < 1122) {
return 1;
}
else if (SC >= 1122) {
return 4;
}
}
}
else if (tcor >= 69) {
if (MVDiff < 41) {
return 1;
}
else if (MVDiff >= 41) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
return 2;
}
}
}
}
}
else if (Rs >= 641) {
if (CsDiff < 6) {
if (Cs < 1031) {
if (mu_mv_mag_sq < 24) {
if (gchDC < 3) {
if (AFD < 22) {
if (posBalance < 59) {
if (refDCval < 60) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
if (ssDCval < 57) {
if (Rs < 953) {
return 4;
}
else if (Rs >= 953) {
return 2;
}
}
else if (ssDCval >= 57) {
return 1;
}
}
}
else if (refDCval >= 60) {
if (SC < 1542) {
return 4;
}
else if (SC >= 1542) {
if (diffRsCsdiff < -31) {
return 2;
}
else if (diffRsCsdiff >= -31) {
return 8;
}
}
}
}
else if (posBalance >= 59) {
if (RsDiff < 11) {
if (Rs < 1394) {
if (SC < 2120) {
if (diffTSC < 1) {
if (Cs < 587) {
return 8;
}
else if (Cs >= 587) {
if (MVDiff < 24) {
if (Cs < 627) {
return 4;
}
else if (Cs >= 627) {
return 8;
}
}
else if (MVDiff >= 24) {
return 4;
}
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (SC >= 2120) {
return 4;
}
}
else if (Rs >= 1394) {
return 8;
}
}
else if (RsDiff >= 11) {
if (Rs < 2343) {
return 4;
}
else if (Rs >= 2343) {
return 2;
}
}
}
}
else if (AFD >= 22) {
return 1;
}
}
else if (gchDC >= 3) {
if (Cs < 621) {
if (diffTSC < 2) {
if (Rs < 896) {
return 8;
}
else if (Rs >= 896) {
return 2;
}
}
else if (diffTSC >= 2) {
return 1;
}
}
else if (Cs >= 621) {
if (AFD < 23) {
if (MVDiff < 37) {
if (MVDiff < 35) {
return 8;
}
else if (MVDiff >= 35) {
return 2;
}
}
else if (MVDiff >= 37) {
return 8;
}
}
else if (AFD >= 23) {
if (Rs < 1048) {
return 1;
}
else if (Rs >= 1048) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (TSC < 9) {
if (negBalance < 54) {
if (MVDiff < 29) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (MVDiff >= 29) {
if (Rs < 1756) {
return 1;
}
else if (Rs >= 1756) {
return 2;
}
}
}
else if (negBalance >= 54) {
if (Rs < 754) {
return 4;
}
else if (Rs >= 754) {
return 8;
}
}
}
else if (TSC >= 9) {
if (RsDiff < 8) {
if (Cs < 743) {
return 4;
}
else if (Cs >= 743) {
return 8;
}
}
else if (RsDiff >= 8) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
return 8;
}
}
}
}
}
else if (Cs >= 1031) {
if (Cs < 1097) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 71) {
if (Rs < 968) {
return 2;
}
else if (Rs >= 968) {
return 8;
}
}
else if (ssDCval >= 71) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (MVDiff < 58) {
return 8;
}
else if (MVDiff >= 58) {
if (Rs < 1874) {
return 1;
}
else if (Rs >= 1874) {
return 4;
}
}
}
}
else if (Cs >= 1097) {
if (Rs < 905) {
if (Rs < 780) {
return 8;
}
else if (Rs >= 780) {
return 1;
}
}
else if (Rs >= 905) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (negBalance < 43) {
if (posBalance < 56) {
if (SC < 3382) {
return 2;
}
else if (SC >= 3382) {
return 1;
}
}
else if (posBalance >= 56) {
if (negBalance < 20) {
if (Rs < 1166) {
return 2;
}
else if (Rs >= 1166) {
return 4;
}
}
else if (negBalance >= 20) {
if (diffRsCsdiff < 49) {
if (tcor < 59) {
return 1;
}
else if (tcor >= 59) {
if (Rs < 1024) {
return 1;
}
else if (Rs >= 1024) {
return 8;
}
}
}
else if (diffRsCsdiff >= 49) {
if (Rs < 1891) {
return 8;
}
else if (Rs >= 1891) {
return 1;
}
}
}
}
}
else if (negBalance >= 43) {
if (mu_mv_mag_sq < 7) {
if (diffMVdiffVal < 6) {
if (MVDiff < 21) {
if (diffRsCsdiff < 30) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (diffRsCsdiff >= 30) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (MVDiff >= 21) {
if (posBalance < 49) {
if (Rs < 1380) {
return 4;
}
else if (Rs >= 1380) {
return 8;
}
}
else if (posBalance >= 49) {
if (posBalance < 69) {
if (diffMVdiffVal < -16) {
if (refDCval < 67) {
return 1;
}
else if (refDCval >= 67) {
return 2;
}
}
else if (diffMVdiffVal >= -16) {
if (Cs < 1037) {
if (tcor < 63) {
if (MVDiff < 26) {
return 2;
}
else if (MVDiff >= 26) {
if (MVDiff < 46) {
return 4;
}
else if (MVDiff >= 46) {
return 2;
}
}
}
else if (tcor >= 63) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (Cs >= 1037) {
if (MVDiff < 22) {
return 4;
}
else if (MVDiff >= 22) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
if (negBalance < 56) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 1525) {
return 2;
}
else if (Rs >= 1525) {
return 4;
}
}
}
else if (negBalance >= 56) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 69) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 6) {
if (gchDC < 3) {
if (SC < 2835) {
if (TSC < 14) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
if (tcor < 73) {
return 4;
}
else if (tcor >= 73) {
return 2;
}
}
}
else if (TSC >= 14) {
return 8;
}
}
else if (SC >= 2835) {
return 8;
}
}
else if (gchDC >= 3) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
if (SC < 1894) {
return 4;
}
else if (SC >= 1894) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 23) {
if (TSC < 8) {
if (posBalance < 61) {
if (MVDiff < 106) {
return 2;
}
else if (MVDiff >= 106) {
return 1;
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (TSC >= 8) {
if (negBalance < 50) {
if (MVDiff < 34) {
if (Rs < 1489) {
return 4;
}
else if (Rs >= 1489) {
return 2;
}
}
else if (MVDiff >= 34) {
return 8;
}
}
else if (negBalance >= 50) {
if (mu_mv_mag_sq < 53) {
if (MVDiff < 85) {
if (mcTcor < 61) {
return 4;
}
else if (mcTcor >= 61) {
if (ssDCval < 89) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (ssDCval >= 89) {
return 4;
}
}
}
else if (MVDiff >= 85) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (Cs < 960) {
return 4;
}
else if (Cs >= 960) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 1640) {
return 4;
}
else if (Rs >= 1640) {
return 1;
}
}
}
}
}
else if (AFD >= 23) {
if (ssDCval < 66) {
if (diffRsCsdiff < 22) {
if (mu_mv_mag_sq < 52) {
return 8;
}
else if (mu_mv_mag_sq >= 52) {
if (Rs < 948) {
return 8;
}
else if (Rs >= 948) {
return 1;
}
}
}
else if (diffRsCsdiff >= 22) {
if (diffAFD < 3) {
if (Rs < 1615) {
return 8;
}
else if (Rs >= 1615) {
return 4;
}
}
else if (diffAFD >= 3) {
if (Rs < 1404) {
return 4;
}
else if (Rs >= 1404) {
return 2;
}
}
}
}
else if (ssDCval >= 66) {
if (Cs < 1053) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (RsCsDiff < 215) {
if (ssDCval < 86) {
if (diffMVdiffVal < -19) {
return 2;
}
else if (diffMVdiffVal >= -19) {
return 1;
}
}
else if (ssDCval >= 86) {
return 2;
}
}
else if (RsCsDiff >= 215) {
return 4;
}
}
}
else if (Cs >= 1053) {
if (diffRsCsdiff < -45) {
if (Cs < 1099) {
return 1;
}
else if (Cs >= 1099) {
return 8;
}
}
else if (diffRsCsdiff >= -45) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (Rs < 1787) {
if (negBalance < 45) {
if (RsCsDiff < 43) {
if (gchDC < 2) {
if (Cs < 613) {
return 2;
}
else if (Cs >= 613) {
return 4;
}
}
else if (gchDC >= 2) {
if (mu_mv_mag_sq < 63) {
if (ssDCval < 66) {
if (RsDiff < 4) {
if (Cs < 660) {
return 1;
}
else if (Cs >= 660) {
return 8;
}
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (mu_mv_mag_sq >= 63) {
if (Rs < 787) {
return 4;
}
else if (Rs >= 787) {
return 8;
}
}
}
}
else if (RsCsDiff >= 43) {
if (MVDiff < 35) {
if (mu_mv_mag_sq < 29) {
if (tcor < 81) {
if (Rs < 1184) {
return 1;
}
else if (Rs >= 1184) {
if (Rs < 1344) {
return 4;
}
else if (Rs >= 1344) {
return 1;
}
}
}
else if (tcor >= 81) {
if (SC < 2346) {
if (RsCsDiff < 51) {
return 4;
}
else if (RsCsDiff >= 51) {
return 2;
}
}
else if (SC >= 2346) {
if (Cs < 1010) {
return 1;
}
else if (Cs >= 1010) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (Cs < 947) {
return 2;
}
else if (Cs >= 947) {
if (Rs < 1022) {
return 2;
}
else if (Rs >= 1022) {
if (Rs < 1244) {
return 8;
}
else if (Rs >= 1244) {
return 4;
}
}
}
}
}
else if (MVDiff >= 35) {
if (diffRsCsdiff < 7) {
return 2;
}
else if (diffRsCsdiff >= 7) {
if (TSC < 8) {
if (Cs < 551) {
return 8;
}
else if (Cs >= 551) {
return 4;
}
}
else if (TSC >= 8) {
if (negBalance < 41) {
return 2;
}
else if (negBalance >= 41) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 45) {
if (RsCsDiff < 115) {
if (mu_mv_mag_sq < 3) {
if (SC < 2417) {
if (Cs < 764) {
if (MVDiff < 54) {
if (RsDiff < 6) {
if (refDCval < 62) {
if (Rs < 575) {
return 2;
}
else if (Rs >= 575) {
return 1;
}
}
else if (refDCval >= 62) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
if (Rs < 743) {
return 1;
}
else if (Rs >= 743) {
return 4;
}
}
}
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (MVDiff >= 54) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
}
else if (Cs >= 764) {
if (refDCval < 65) {
if (ssDCval < 56) {
if (Rs < 747) {
return 2;
}
else if (Rs >= 747) {
return 4;
}
}
else if (ssDCval >= 56) {
return 1;
}
}
else if (refDCval >= 65) {
if (AFD < 24) {
if (posBalance < 65) {
if (Cs < 1148) {
return 2;
}
else if (Cs >= 1148) {
return 4;
}
}
else if (posBalance >= 65) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 1;
}
}
}
else if (AFD >= 24) {
return 4;
}
}
}
}
else if (SC >= 2417) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mu_mv_mag_sq < 11) {
if (posBalance < 67) {
if (SC < 2593) {
if (mcTcor < 92) {
if (RsCsDiff < 62) {
if (Cs < 588) {
return 4;
}
else if (Cs >= 588) {
return 8;
}
}
else if (RsCsDiff >= 62) {
if (Rs < 788) {
return 2;
}
else if (Rs >= 788) {
return 4;
}
}
}
else if (mcTcor >= 92) {
if (SC < 2126) {
if (RsCsDiff < 73) {
if (tcor < 83) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (tcor >= 83) {
return 1;
}
}
else if (RsCsDiff >= 73) {
if (tcor < 72) {
return 4;
}
else if (tcor >= 72) {
return 1;
}
}
}
else if (SC >= 2126) {
if (ssDCval < 76) {
if (Rs < 1157) {
return 4;
}
else if (Rs >= 1157) {
return 8;
}
}
else if (ssDCval >= 76) {
return 4;
}
}
}
}
else if (SC >= 2593) {
if (Rs < 1695) {
return 8;
}
else if (Rs >= 1695) {
return 1;
}
}
}
else if (posBalance >= 67) {
if (SC < 1903) {
return 8;
}
else if (SC >= 1903) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (tcor < 79) {
if (MVDiff < 32) {
if (Rs < 451) {
return 4;
}
else if (Rs >= 451) {
if (Rs < 1110) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (Rs < 583) {
return 8;
}
else if (Rs >= 583) {
return 1;
}
}
}
}
else if (Rs >= 1110) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
else if (MVDiff >= 32) {
if (Rs < 619) {
if (RsDiff < 3) {
if (Cs < 1026) {
return 1;
}
else if (Cs >= 1026) {
return 4;
}
}
else if (RsDiff >= 3) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (Rs >= 619) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 8;
}
}
}
}
else if (tcor >= 79) {
if (mu_mv_mag_sq < 54) {
if (diffMVdiffVal < -20) {
if (Rs < 863) {
return 1;
}
else if (Rs >= 863) {
return 4;
}
}
else if (diffMVdiffVal >= -20) {
if (tcor < 81) {
if (SC < 1661) {
if (Rs < 1052) {
return 4;
}
else if (Rs >= 1052) {
return 1;
}
}
else if (SC >= 1661) {
return 8;
}
}
else if (tcor >= 81) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 54) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 115) {
if (SC < 2174) {
if (refDCval < 85) {
if (SC < 1899) {
if (Cs < 731) {
if (Rs < 831) {
return 2;
}
else if (Rs >= 831) {
if (Rs < 886) {
return 4;
}
else if (Rs >= 886) {
return 1;
}
}
}
else if (Cs >= 731) {
if (MVDiff < 37) {
return 2;
}
else if (MVDiff >= 37) {
return 8;
}
}
}
else if (SC >= 1899) {
return 2;
}
}
else if (refDCval >= 85) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 4;
}
}
}
else if (SC >= 2174) {
if (AFD < 19) {
if (negBalance < 53) {
if (Cs < 944) {
return 2;
}
else if (Cs >= 944) {
return 1;
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (AFD >= 19) {
if (RsDiff < 12) {
if (diffMVdiffVal < -6) {
if (SC < 2292) {
return 4;
}
else if (SC >= 2292) {
return 2;
}
}
else if (diffMVdiffVal >= -6) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
if (posBalance < 65) {
return 4;
}
else if (posBalance >= 65) {
if (SC < 2691) {
return 8;
}
else if (SC >= 2691) {
return 4;
}
}
}
}
}
else if (RsDiff >= 12) {
if (mcTcor < 92) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 8;
}
}
else if (mcTcor >= 92) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 1787) {
if (refDCval < 70) {
return 2;
}
else if (refDCval >= 70) {
if (Cs < 827) {
if (CsDiff < 4) {
if (MVDiff < 27) {
if (ssDCval < 84) {
return 2;
}
else if (ssDCval >= 84) {
return 4;
}
}
else if (MVDiff >= 27) {
if (Rs < 2033) {
return 8;
}
else if (Rs >= 2033) {
return 1;
}
}
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (Cs >= 827) {
if (negBalance < 46) {
if (CsDiff < 7) {
if (Rs < 2008) {
return 2;
}
else if (Rs >= 2008) {
return 4;
}
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (negBalance >= 46) {
if (tcor < 80) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (RsCsDiff < 157) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
else if (RsCsDiff >= 157) {
return 4;
}
}
}
else if (tcor >= 80) {
if (CsDiff < 6) {
if (mcTcor < 93) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (Cs < 961) {
return 4;
}
else if (Cs >= 961) {
return 1;
}
}
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 90) {
if (tcor < 79) {
if (mcTcor < 72) {
if (gchDC < 9) {
if (diffMVdiffVal < 27) {
if (mcTcor < 71) {
return 1;
}
else if (mcTcor >= 71) {
return 4;
}
}
else if (diffMVdiffVal >= 27) {
if (tcor < 30) {
return 1;
}
else if (tcor >= 30) {
return 8;
}
}
}
else if (gchDC >= 9) {
if (ssDCval < 118) {
if (AFD < 31) {
return 8;
}
else if (AFD >= 31) {
return 1;
}
}
else if (ssDCval >= 118) {
return 4;
}
}
}
else if (mcTcor >= 72) {
if (Rs < 801) {
if (MVDiff < 93) {
if (TSC < 8) {
if (negBalance < 70) {
if (mu_mv_mag_sq < 9) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 1;
}
}
else if (mu_mv_mag_sq >= 9) {
return 8;
}
}
else if (negBalance >= 70) {
return 2;
}
}
else if (TSC >= 8) {
if (Rs < 681) {
return 2;
}
else if (Rs >= 681) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
}
}
else if (MVDiff >= 93) {
if (RsDiff < 4) {
if (Cs < 773) {
return 2;
}
else if (Cs >= 773) {
if (Rs < 594) {
return 4;
}
else if (Rs >= 594) {
return 8;
}
}
}
else if (RsDiff >= 4) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
else if (Rs >= 801) {
if (ssDCval < 97) {
if (mcTcor < 94) {
if (AFD < 19) {
if (SC < 3061) {
return 8;
}
else if (SC >= 3061) {
return 2;
}
}
else if (AFD >= 19) {
if (gchDC < 2) {
if (SC < 1814) {
return 8;
}
else if (SC >= 1814) {
if (CsDiff < 7) {
return 1;
}
else if (CsDiff >= 7) {
if (posBalance < 62) {
if (Rs < 1657) {
return 1;
}
else if (Rs >= 1657) {
return 8;
}
}
else if (posBalance >= 62) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (negBalance < 58) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (negBalance >= 58) {
if (negBalance < 83) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 1;
}
}
else if (negBalance >= 83) {
if (Rs < 1216) {
return 4;
}
else if (Rs >= 1216) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (AFD < 24) {
if (Rs < 1760) {
return 8;
}
else if (Rs >= 1760) {
return 4;
}
}
else if (AFD >= 24) {
if (AFD < 30) {
return 4;
}
else if (AFD >= 30) {
return 2;
}
}
}
}
else if (ssDCval >= 97) {
if (refDCval < 113) {
if (gchDC < 3) {
if (tcor < 60) {
return 4;
}
else if (tcor >= 60) {
if (diffRsCsdiff < 14) {
if (refDCval < 97) {
return 1;
}
else if (refDCval >= 97) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 1811) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 4;
}
}
else if (Rs >= 1811) {
if (Cs < 1171) {
if (Rs < 1913) {
return 1;
}
else if (Rs >= 1913) {
return 2;
}
}
else if (Cs >= 1171) {
return 4;
}
}
}
}
}
else if (diffRsCsdiff >= 14) {
if (diffMVdiffVal < 1) {
if (AFD < 23) {
if (Cs < 1017) {
return 8;
}
else if (Cs >= 1017) {
return 2;
}
}
else if (AFD >= 23) {
if (AFD < 29) {
return 4;
}
else if (AFD >= 29) {
return 2;
}
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 997) {
return 1;
}
else if (Cs >= 997) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
return 1;
}
}
}
}
}
}
else if (gchDC >= 3) {
if (Cs < 615) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (Cs >= 615) {
if (ssDCval < 104) {
if (MVDiff < 42) {
return 2;
}
else if (MVDiff >= 42) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
return 2;
}
}
}
else if (ssDCval >= 104) {
if (tcor < 56) {
if (Rs < 975) {
return 4;
}
else if (Rs >= 975) {
return 1;
}
}
else if (tcor >= 56) {
return 8;
}
}
}
}
}
else if (refDCval >= 113) {
if (mcTcor < 88) {
if (AFD < 32) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (RsDiff < 8) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
else if (RsDiff >= 8) {
if (refDCval < 134) {
if (AFD < 30) {
if (ssDCval < 129) {
if (Rs < 1546) {
return 1;
}
else if (Rs >= 1546) {
return 4;
}
}
else if (ssDCval >= 129) {
return 1;
}
}
else if (AFD >= 30) {
return 8;
}
}
else if (refDCval >= 134) {
if (Rs < 1625) {
if (Rs < 1195) {
return 8;
}
else if (Rs >= 1195) {
return 4;
}
}
else if (Rs >= 1625) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
}
}
}
}
else if (AFD >= 32) {
if (Cs < 1115) {
return 2;
}
else if (Cs >= 1115) {
return 1;
}
}
}
else if (mcTcor >= 88) {
if (refDCval < 128) {
if (diffAFD < -2) {
if (Rs < 2212) {
return 8;
}
else if (Rs >= 2212) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Cs < 696) {
return 8;
}
else if (Cs >= 696) {
return 2;
}
}
}
else if (refDCval >= 128) {
if (mu_mv_mag_sq < 58) {
if (negBalance < 57) {
if (AFD < 24) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
else if (AFD >= 24) {
if (negBalance < 53) {
if (Rs < 1894) {
return 1;
}
else if (Rs >= 1894) {
return 2;
}
}
else if (negBalance >= 53) {
if (Rs < 2272) {
return 8;
}
else if (Rs >= 2272) {
return 1;
}
}
}
}
else if (negBalance >= 57) {
if (mcTcor < 96) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
else if (mcTcor >= 96) {
if (Rs < 1528) {
return 8;
}
else if (Rs >= 1528) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 58) {
return 2;
}
}
}
}
}
}
}
}
else if (tcor >= 79) {
if (RsDiff < 9) {
if (negBalance < 38) {
if (Rs < 669) {
return 8;
}
else if (Rs >= 669) {
return 2;
}
}
else if (negBalance >= 38) {
if (tcor < 92) {
if (Rs < 1320) {
if (gchDC < 2) {
if (diffRsCsdiff < -19) {
if (Rs < 1187) {
return 8;
}
else if (Rs >= 1187) {
return 1;
}
}
else if (diffRsCsdiff >= -19) {
if (SC < 2236) {
return 1;
}
else if (SC >= 2236) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (RsCsDiff < 92) {
if (gchDC < 3) {
if (Cs < 1208) {
return 2;
}
else if (Cs >= 1208) {
return 1;
}
}
else if (gchDC >= 3) {
if (ssDCval < 159) {
if (diffMVdiffVal < -98) {
return 2;
}
else if (diffMVdiffVal >= -98) {
return 8;
}
}
else if (ssDCval >= 159) {
if (Rs < 1043) {
return 2;
}
else if (Rs >= 1043) {
return 8;
}
}
}
}
else if (RsCsDiff >= 92) {
if (Rs < 976) {
return 1;
}
else if (Rs >= 976) {
return 2;
}
}
}
}
else if (Rs >= 1320) {
if (Cs < 886) {
if (MVDiff < 62) {
if (Rs < 2074) {
if (CsDiff < 7) {
return 1;
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (Rs >= 2074) {
return 8;
}
}
else if (MVDiff >= 62) {
return 8;
}
}
else if (Cs >= 886) {
if (TSC < 11) {
if (diffTSC < -1) {
if (tcor < 83) {
if (Rs < 1954) {
return 8;
}
else if (Rs >= 1954) {
return 4;
}
}
else if (tcor >= 83) {
if (diffMVdiffVal < 4) {
return 1;
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (TSC >= 11) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
if (Rs < 1699) {
return 4;
}
else if (Rs >= 1699) {
return 8;
}
}
}
}
}
}
else if (tcor >= 92) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (negBalance < 52) {
if (CsDiff < 7) {
if (refDCval < 100) {
return 2;
}
else if (refDCval >= 100) {
if (diffRsCsdiff < -6) {
return 1;
}
else if (diffRsCsdiff >= -6) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
if (mu_mv_mag_sq < 10) {
return 1;
}
else if (mu_mv_mag_sq >= 10) {
if (Rs < 2181) {
return 2;
}
else if (Rs >= 2181) {
return 1;
}
}
}
}
}
}
else if (CsDiff >= 7) {
if (Cs < 1155) {
return 2;
}
else if (Cs >= 1155) {
if (Cs < 1280) {
return 8;
}
else if (Cs >= 1280) {
if (Rs < 1936) {
return 2;
}
else if (Rs >= 1936) {
return 1;
}
}
}
}
}
else if (negBalance >= 52) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 93) {
if (AFD < 17) {
if (Rs < 1986) {
return 1;
}
else if (Rs >= 1986) {
return 4;
}
}
else if (AFD >= 17) {
return 8;
}
}
else if (ssDCval >= 93) {
if (posBalance < 44) {
if (Rs < 2036) {
return 8;
}
else if (Rs >= 2036) {
return 1;
}
}
else if (posBalance >= 44) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffMVdiffVal < 7) {
if (Rs < 1673) {
return 8;
}
else if (Rs >= 1673) {
return 1;
}
}
else if (diffMVdiffVal >= 7) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 88) {
if (SC < 2149) {
if (Rs < 1065) {
if (ssDCval < 63) {
if (Rs < 836) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (Rs < 548) {
if (SC < 1064) {
if (mcTcor < 98) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
if (diffTSC < -1) {
if (Rs < 265) {
return 1;
}
else if (Rs >= 265) {
return 8;
}
}
else if (diffTSC >= -1) {
if (mu_mv_mag_sq < 9) {
if (Rs < 278) {
return 1;
}
else if (Rs >= 278) {
return 2;
}
}
else if (mu_mv_mag_sq >= 9) {
return 1;
}
}
}
}
else if (mcTcor >= 98) {
if (ssDCval < 55) {
if (MVDiff < 13) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (MVDiff >= 13) {
if (Rs < 453) {
return 4;
}
else if (Rs >= 453) {
return 1;
}
}
}
else if (ssDCval >= 55) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (Rs < 232) {
return 1;
}
else if (Rs >= 232) {
return 8;
}
}
}
}
}
else if (SC >= 1064) {
if (Rs < 482) {
if (negBalance < 26) {
if (AFD < 4) {
if (ssDCval < 60) {
return 2;
}
else if (ssDCval >= 60) {
return 8;
}
}
else if (AFD >= 4) {
if (Rs < 395) {
return 1;
}
else if (Rs >= 395) {
return 4;
}
}
}
else if (negBalance >= 26) {
if (tcor < 97) {
if (Cs < 693) {
return 4;
}
else if (Cs >= 693) {
return 1;
}
}
else if (tcor >= 97) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
return 4;
}
}
}
}
else if (Rs >= 482) {
if (Cs < 649) {
if (ssDCval < 55) {
if (Rs < 526) {
return 4;
}
else if (Rs >= 526) {
return 1;
}
}
else if (ssDCval >= 55) {
if (Rs < 528) {
return 2;
}
else if (Rs >= 528) {
return 8;
}
}
}
else if (Cs >= 649) {
if (tcor < 100) {
if (AFD < 8) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
return 4;
}
}
}
else if (AFD >= 8) {
return 1;
}
}
else if (tcor >= 100) {
return 8;
}
}
}
}
}
else if (Rs >= 548) {
if (MVDiff < 1) {
if (Cs < 934) {
if (negBalance < 32) {
if (tcor < 97) {
if (Rs < 610) {
return 1;
}
else if (Rs >= 610) {
return 8;
}
}
else if (tcor >= 97) {
return 8;
}
}
else if (negBalance >= 32) {
if (refDCval < 54) {
if (Cs < 838) {
return 4;
}
else if (Cs >= 838) {
return 8;
}
}
else if (refDCval >= 54) {
if (Cs < 568) {
return 4;
}
else if (Cs >= 568) {
return 8;
}
}
}
}
else if (Cs >= 934) {
if (Rs < 636) {
return 4;
}
else if (Rs >= 636) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
if (Rs < 717) {
return 1;
}
else if (Rs >= 717) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 17) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
return 1;
}
}
else if (negBalance >= 17) {
if (negBalance < 30) {
if (refDCval < 51) {
if (Rs < 771) {
return 1;
}
else if (Rs >= 771) {
return 8;
}
}
else if (refDCval >= 51) {
if (Rs < 781) {
if (negBalance < 18) {
return 8;
}
else if (negBalance >= 18) {
if (MVDiff < 15) {
if (SC < 1192) {
if (Rs < 608) {
return 8;
}
else if (Rs >= 608) {
return 4;
}
}
else if (SC >= 1192) {
return 4;
}
}
else if (MVDiff >= 15) {
return 8;
}
}
}
else if (Rs >= 781) {
if (Cs < 794) {
return 1;
}
else if (Cs >= 794) {
return 4;
}
}
}
}
else if (negBalance >= 30) {
if (MVDiff < 8) {
if (ssDCval < 59) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
}
else if (ssDCval >= 59) {
if (Cs < 1064) {
return 8;
}
else if (Cs >= 1064) {
return 1;
}
}
}
else if (MVDiff >= 8) {
if (Cs < 675) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (MVDiff < 10) {
if (Rs < 752) {
return 4;
}
else if (Rs >= 752) {
return 1;
}
}
else if (MVDiff >= 10) {
return 1;
}
}
}
}
else if (Cs >= 675) {
if (SC < 1500) {
if (Rs < 644) {
if (Rs < 582) {
return 2;
}
else if (Rs >= 582) {
return 1;
}
}
else if (Rs >= 644) {
return 4;
}
}
else if (SC >= 1500) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 836) {
if (refDCval < 51) {
if (Cs < 1139) {
if (Cs < 569) {
return 1;
}
else if (Cs >= 569) {
if (MVDiff < 10) {
if (Rs < 852) {
if (Rs < 842) {
return 8;
}
else if (Rs >= 842) {
return 1;
}
}
else if (Rs >= 852) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (MVDiff >= 10) {
if (Rs < 945) {
return 4;
}
else if (Rs >= 945) {
return 1;
}
}
}
}
else if (Cs >= 1139) {
return 1;
}
}
else if (refDCval >= 51) {
if (posBalance < 31) {
if (Rs < 841) {
return 8;
}
else if (Rs >= 841) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 4;
}
}
}
else if (posBalance >= 31) {
if (TSC < 4) {
if (mu_mv_mag_sq < 127) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (Rs < 841) {
if (Rs < 840) {
return 1;
}
else if (Rs >= 840) {
return 4;
}
}
else if (Rs >= 841) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
else if (TSC >= 4) {
if (negBalance < 27) {
if (Rs < 961) {
return 4;
}
else if (Rs >= 961) {
return 1;
}
}
else if (negBalance >= 27) {
if (Cs < 782) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (Cs >= 782) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 63) {
if (Cs < 685) {
if (mu_mv_mag_sq < 108) {
if (ssDCval < 76) {
if (Cs < 586) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (negBalance < 37) {
if (Rs < 394) {
return 4;
}
else if (Rs >= 394) {
if (mu_mv_mag_sq < 91) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 765) {
return 4;
}
else if (Rs >= 765) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 91) {
return 2;
}
}
}
else if (negBalance >= 37) {
return 1;
}
}
}
else if (Cs >= 586) {
if (TSC < 4) {
if (Rs < 633) {
if (TSC < 3) {
if (Rs < 267) {
return 1;
}
else if (Rs >= 267) {
return 4;
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (Rs >= 633) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 4;
}
}
}
else if (TSC >= 4) {
if (Rs < 514) {
if (Rs < 466) {
if (Rs < 356) {
return 2;
}
else if (Rs >= 356) {
return 8;
}
}
else if (Rs >= 466) {
return 1;
}
}
else if (Rs >= 514) {
if (MVDiff < 2) {
if (mu_mv_mag_sq < 36) {
return 8;
}
else if (mu_mv_mag_sq >= 36) {
if (Rs < 574) {
return 8;
}
else if (Rs >= 574) {
return 4;
}
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
}
else if (ssDCval >= 76) {
if (ssDCval < 85) {
if (AFD < 19) {
if (Cs < 533) {
if (Rs < 804) {
return 1;
}
else if (Rs >= 804) {
return 8;
}
}
else if (Cs >= 533) {
return 8;
}
}
else if (AFD >= 19) {
if (Rs < 847) {
return 1;
}
else if (Rs >= 847) {
return 2;
}
}
}
else if (ssDCval >= 85) {
if (Cs < 602) {
if (Rs < 914) {
return 4;
}
else if (Rs >= 914) {
return 8;
}
}
else if (Cs >= 602) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (SC < 1431) {
if (AFD < 3) {
if (mu_mv_mag_sq < 122) {
if (negBalance < 24) {
return 2;
}
else if (negBalance >= 24) {
if (mu_mv_mag_sq < 118) {
return 2;
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (AFD < 2) {
if (Rs < 816) {
return 2;
}
else if (Rs >= 816) {
return 8;
}
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (AFD >= 3) {
if (refDCval < 72) {
if (Rs < 573) {
if (Rs < 259) {
return 1;
}
else if (Rs >= 259) {
return 2;
}
}
else if (Rs >= 573) {
return 1;
}
}
else if (refDCval >= 72) {
return 8;
}
}
}
else if (SC >= 1431) {
if (SC < 1730) {
if (ssDCval < 75) {
if (Rs < 796) {
return 1;
}
else if (Rs >= 796) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 2;
}
}
}
else if (ssDCval >= 75) {
return 8;
}
}
else if (SC >= 1730) {
return 4;
}
}
}
}
else if (Cs >= 685) {
if (ssDCval < 81) {
if (AFD < 6) {
if (Cs < 782) {
if (Rs < 841) {
if (mu_mv_mag_sq < 90) {
return 4;
}
else if (mu_mv_mag_sq >= 90) {
if (Rs < 683) {
if (Rs < 470) {
return 4;
}
else if (Rs >= 470) {
if (Rs < 667) {
return 8;
}
else if (Rs >= 667) {
if (Cs < 698) {
return 1;
}
else if (Cs >= 698) {
return 8;
}
}
}
}
else if (Rs >= 683) {
if (ssDCval < 67) {
if (refDCval < 66) {
return 4;
}
else if (refDCval >= 66) {
return 8;
}
}
else if (ssDCval >= 67) {
if (Rs < 804) {
return 1;
}
else if (Rs >= 804) {
return 4;
}
}
}
}
}
else if (Rs >= 841) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (Rs < 925) {
return 8;
}
else if (Rs >= 925) {
return 4;
}
}
}
}
}
else if (Cs >= 782) {
if (Cs < 1187) {
if (mu_mv_mag_sq < 107) {
if (Cs < 869) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (mu_mv_mag_sq < 94) {
if (Rs < 747) {
return 1;
}
else if (Rs >= 747) {
return 4;
}
}
else if (mu_mv_mag_sq >= 94) {
return 8;
}
}
}
else if (Cs >= 869) {
if (Cs < 969) {
if (AFD < 5) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (Cs >= 969) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (negBalance < 34) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (negBalance >= 34) {
if (Rs < 629) {
return 1;
}
else if (Rs >= 629) {
if (SC < 1956) {
return 8;
}
else if (SC >= 1956) {
if (Rs < 931) {
return 1;
}
else if (Rs >= 931) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
if (SC < 2064) {
if (ssDCval < 68) {
if (SC < 1966) {
if (Rs < 721) {
return 8;
}
else if (Rs >= 721) {
if (Rs < 972) {
return 4;
}
else if (Rs >= 972) {
return 8;
}
}
}
else if (SC >= 1966) {
return 1;
}
}
else if (ssDCval >= 68) {
if (Rs < 825) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (Rs >= 825) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
if (Cs < 1017) {
if (mu_mv_mag_sq < 118) {
return 8;
}
else if (mu_mv_mag_sq >= 118) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
return 4;
}
}
}
else if (Cs >= 1017) {
return 4;
}
}
}
}
}
else if (SC >= 2064) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
}
else if (Cs >= 1187) {
if (Rs < 752) {
return 8;
}
else if (Rs >= 752) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
}
}
else if (AFD >= 6) {
if (Rs < 528) {
if (ssDCval < 66) {
if (Rs < 468) {
return 2;
}
else if (Rs >= 468) {
return 1;
}
}
else if (ssDCval >= 66) {
if (negBalance < 55) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
if (Rs < 387) {
return 1;
}
else if (Rs >= 387) {
return 2;
}
}
}
else if (negBalance >= 55) {
if (Cs < 1138) {
return 2;
}
else if (Cs >= 1138) {
return 1;
}
}
}
}
else if (Rs >= 528) {
if (tcor < 89) {
if (Rs < 980) {
if (Cs < 1166) {
return 8;
}
else if (Cs >= 1166) {
if (Rs < 686) {
return 8;
}
else if (Rs >= 686) {
return 2;
}
}
}
else if (Rs >= 980) {
return 1;
}
}
else if (tcor >= 89) {
if (mu_mv_mag_sq < 64) {
if (Cs < 827) {
return 4;
}
else if (Cs >= 827) {
if (Rs < 795) {
return 2;
}
else if (Rs >= 795) {
if (SC < 1914) {
if (SC < 1832) {
if (Rs < 821) {
return 2;
}
else if (Rs >= 821) {
return 8;
}
}
else if (SC >= 1832) {
return 4;
}
}
else if (SC >= 1914) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (Rs < 752) {
return 8;
}
else if (Rs >= 752) {
if (Rs < 1030) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (Rs < 754) {
return 2;
}
else if (Rs >= 754) {
return 4;
}
}
}
else if (Rs >= 1030) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (SC < 1511) {
if (refDCval < 86) {
if (mu_mv_mag_sq < 117) {
if (diffRsCsdiff < -1) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
if (Rs < 478) {
return 8;
}
else if (Rs >= 478) {
return 1;
}
}
}
else if (diffRsCsdiff >= -1) {
if (Cs < 877) {
return 1;
}
else if (Cs >= 877) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 117) {
return 2;
}
}
else if (refDCval >= 86) {
return 2;
}
}
else if (SC >= 1511) {
if (negBalance < 34) {
if (Rs < 762) {
return 2;
}
else if (Rs >= 762) {
return 4;
}
}
else if (negBalance >= 34) {
if (Rs < 1005) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
return 4;
}
}
else if (Rs >= 1005) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1065) {
if (refDCval < 69) {
if (ssDCval < 51) {
if (Cs < 755) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
if (SC < 1971) {
return 8;
}
else if (SC >= 1971) {
if (Rs < 1345) {
if (Rs < 1326) {
return 4;
}
else if (Rs >= 1326) {
return 8;
}
}
else if (Rs >= 1345) {
return 4;
}
}
}
}
else if (Cs >= 755) {
return 8;
}
}
else if (ssDCval >= 51) {
if (TSC < 5) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 46) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
if (mu_mv_mag_sq < 21) {
return 2;
}
else if (mu_mv_mag_sq >= 21) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Cs < 708) {
return 1;
}
else if (Cs >= 708) {
if (Cs < 731) {
return 8;
}
else if (Cs >= 731) {
if (diffRsCsdiff < 5) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (diffRsCsdiff >= 5) {
if (Cs < 755) {
return 1;
}
else if (Cs >= 755) {
return 8;
}
}
}
}
}
}
}
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 122) {
if (mu_mv_mag_sq < 79) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 1244) {
return 1;
}
else if (Rs >= 1244) {
if (Rs < 1345) {
return 2;
}
else if (Rs >= 1345) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Cs < 677) {
return 2;
}
else if (Cs >= 677) {
if (Rs < 1170) {
return 4;
}
else if (Rs >= 1170) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 757) {
if (Rs < 1175) {
return 4;
}
else if (Rs >= 1175) {
return 1;
}
}
else if (Cs >= 757) {
return 8;
}
}
}
else if (TSC >= 5) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
if (Rs < 1157) {
if (Cs < 642) {
return 8;
}
else if (Cs >= 642) {
if (Cs < 899) {
return 4;
}
else if (Cs >= 899) {
return 8;
}
}
}
else if (Rs >= 1157) {
return 8;
}
}
}
}
}
else if (refDCval >= 69) {
if (ssDCval < 72) {
if (Rs < 1108) {
return 2;
}
else if (Rs >= 1108) {
if (mu_mv_mag_sq < 50) {
if (Rs < 1390) {
return 4;
}
else if (Rs >= 1390) {
return 1;
}
}
else if (mu_mv_mag_sq >= 50) {
return 8;
}
}
}
else if (ssDCval >= 72) {
if (Cs < 732) {
if (SC < 2138) {
return 1;
}
else if (SC >= 2138) {
if (Rs < 1499) {
return 4;
}
else if (Rs >= 1499) {
return 8;
}
}
}
else if (Cs >= 732) {
if (mu_mv_mag_sq < 15) {
if (Rs < 1222) {
return 2;
}
else if (Rs >= 1222) {
return 1;
}
}
else if (mu_mv_mag_sq >= 15) {
return 8;
}
}
}
}
}
}
else if (SC >= 2149) {
if (Rs < 1869) {
if (Cs < 1087) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 43) {
if (ssDCval < 84) {
if (Cs < 767) {
if (ssDCval < 51) {
if (Rs < 1757) {
return 2;
}
else if (Rs >= 1757) {
return 4;
}
}
else if (ssDCval >= 51) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (Cs >= 767) {
if (negBalance < 22) {
if (ssDCval < 61) {
if (Rs < 1703) {
if (negBalance < 19) {
if (SC < 2202) {
return 8;
}
else if (SC >= 2202) {
return 1;
}
}
else if (negBalance >= 19) {
if (Rs < 1517) {
return 8;
}
else if (Rs >= 1517) {
return 4;
}
}
}
else if (Rs >= 1703) {
return 4;
}
}
else if (ssDCval >= 61) {
if (negBalance < 16) {
if (Cs < 966) {
return 8;
}
else if (Cs >= 966) {
return 1;
}
}
else if (negBalance >= 16) {
return 1;
}
}
}
else if (negBalance >= 22) {
if (mu_mv_mag_sq < 79) {
if (Rs < 1649) {
return 1;
}
else if (Rs >= 1649) {
return 4;
}
}
else if (mu_mv_mag_sq >= 79) {
if (ssDCval < 56) {
if (MVDiff < 2) {
if (Rs < 1411) {
return 8;
}
else if (Rs >= 1411) {
if (Rs < 1503) {
return 4;
}
else if (Rs >= 1503) {
return 4;
}
}
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 56) {
if (negBalance < 60) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (negBalance >= 60) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (Rs < 1537) {
return 8;
}
else if (Rs >= 1537) {
return 4;
}
}
}
else if (posBalance >= 43) {
if (Cs < 957) {
if (Rs < 1371) {
if (tcor < 99) {
if (RsCsDiff < 22) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
return 1;
}
}
else if (RsCsDiff >= 22) {
return 8;
}
}
else if (tcor >= 99) {
if (Rs < 1263) {
return 8;
}
else if (Rs >= 1263) {
return 1;
}
}
}
else if (Rs >= 1371) {
if (mu_mv_mag_sq < 105) {
if (Cs < 729) {
return 1;
}
else if (Cs >= 729) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (TSC < 6) {
if (Rs < 1403) {
return 4;
}
else if (Rs >= 1403) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 119) {
return 8;
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 1764) {
if (Rs < 1595) {
return 4;
}
else if (Rs >= 1595) {
return 8;
}
}
else if (Rs >= 1764) {
return 4;
}
}
}
}
}
}
else if (Cs >= 957) {
if (Rs < 1340) {
if (negBalance < 34) {
if (ssDCval < 68) {
if (SC < 2210) {
return 8;
}
else if (SC >= 2210) {
return 1;
}
}
else if (ssDCval >= 68) {
return 8;
}
}
else if (negBalance >= 34) {
if (mcTcor < 98) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Rs < 1256) {
return 1;
}
else if (Rs >= 1256) {
return 4;
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
return 4;
}
}
}
}
else if (Rs >= 1340) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (posBalance < 57) {
if (Rs < 1709) {
return 2;
}
else if (Rs >= 1709) {
return 8;
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1031) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (posBalance < 40) {
if (SC < 2258) {
if (Rs < 1273) {
return 1;
}
else if (Rs >= 1273) {
return 8;
}
}
else if (SC >= 2258) {
return 1;
}
}
else if (posBalance >= 40) {
return 8;
}
}
}
else if (Cs >= 1031) {
if (Rs < 1826) {
if (Cs < 1067) {
if (Rs < 1734) {
return 4;
}
else if (Rs >= 1734) {
if (negBalance < 18) {
if (Rs < 1811) {
return 8;
}
else if (Rs >= 1811) {
return 4;
}
}
else if (negBalance >= 18) {
if (Rs < 1799) {
return 4;
}
else if (Rs >= 1799) {
return 4;
}
}
}
}
else if (Cs >= 1067) {
return 8;
}
}
else if (Rs >= 1826) {
return 1;
}
}
}
}
else if (Cs >= 1087) {
if (RsDiff < 1) {
if (MVDiff < 11) {
if (ssDCval < 55) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
return 4;
}
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (RsDiff >= 1) {
if (AFD < 3) {
if (ssDCval < 55) {
if (negBalance < 25) {
if (Cs < 1155) {
if (posBalance < 40) {
if (Cs < 1153) {
if (Cs < 1135) {
if (gchDC < 1) {
if (negBalance < 22) {
if (TSCindex < 3) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Rs < 1762) {
if (Cs < 1108) {
return 8;
}
else if (Cs >= 1108) {
if (Cs < 1114) {
return 4;
}
else if (Cs >= 1114) {
if (Rs < 1683) {
return 8;
}
else if (Rs >= 1683) {
if (posBalance < 35) {
return 4;
}
else if (posBalance >= 35) {
if (Rs < 1717) {
return 8;
}
else if (Rs >= 1717) {
if (Rs < 1739) {
return 4;
}
else if (Rs >= 1739) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1762) {
if (posBalance < 34) {
return 4;
}
else if (posBalance >= 34) {
return 4;
}
}
}
}
else if (TSCindex >= 3) {
if (Rs < 1652) {
return 4;
}
else if (Rs >= 1652) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (Rs < 1641) {
return 4;
}
else if (Rs >= 1641) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (Cs >= 1135) {
if (negBalance < 20) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (Rs < 1753) {
return 4;
}
else if (Rs >= 1753) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
else if (negBalance >= 20) {
return 4;
}
}
}
else if (Cs >= 1153) {
return 8;
}
}
else if (posBalance >= 40) {
if (Rs < 1806) {
if (Cs < 1092) {
return 8;
}
else if (Cs >= 1092) {
return 8;
}
}
else if (Rs >= 1806) {
return 4;
}
}
}
else if (Cs >= 1155) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (negBalance < 17) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (Rs < 1621) {
return 8;
}
else if (Rs >= 1621) {
if (Cs < 1169) {
return 4;
}
else if (Cs >= 1169) {
return 4;
}
}
}
}
else if (negBalance >= 17) {
if (Cs < 1209) {
if (ssDCval < 49) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (Cs >= 1209) {
return 4;
}
}
}
}
}
else if (negBalance >= 25) {
return 8;
}
}
else if (ssDCval >= 55) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (refDCval < 70) {
if (mu_mv_mag_sq < 127) {
if (Rs < 1232) {
return 4;
}
else if (Rs >= 1232) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
return 4;
}
}
else if (refDCval >= 70) {
if (tcor < 99) {
if (MVDiff < 2) {
if (Rs < 1380) {
if (Cs < 1121) {
return 8;
}
else if (Cs >= 1121) {
return 4;
}
}
else if (Rs >= 1380) {
return 8;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (tcor >= 99) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (Cs < 1103) {
return 8;
}
else if (Cs >= 1103) {
return 4;
}
}
}
}
}
}
}
}
else if (AFD >= 3) {
if (Rs < 1393) {
if (RsDiff < 2) {
if (ssDCval < 81) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (Rs < 1161) {
return 8;
}
else if (Rs >= 1161) {
if (mu_mv_mag_sq < 107) {
return 8;
}
else if (mu_mv_mag_sq >= 107) {
if (Cs < 1114) {
return 8;
}
else if (Cs >= 1114) {
return 1;
}
}
}
}
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (RsDiff >= 2) {
if (Rs < 1003) {
if (Rs < 960) {
return 4;
}
else if (Rs >= 960) {
return 2;
}
}
else if (Rs >= 1003) {
if (mu_mv_mag_sq < 113) {
if (refDCval < 69) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (AFD < 6) {
if (tcor < 97) {
if (ssDCval < 52) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
else if (ssDCval >= 52) {
return 8;
}
}
else if (tcor >= 97) {
return 4;
}
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (refDCval >= 69) {
return 8;
}
}
else if (mu_mv_mag_sq >= 113) {
if (posBalance < 49) {
if (Cs < 1127) {
return 8;
}
else if (Cs >= 1127) {
return 1;
}
}
else if (posBalance >= 49) {
if (posBalance < 54) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
if (Rs < 1228) {
return 1;
}
else if (Rs >= 1228) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (Rs < 1313) {
return 8;
}
else if (Rs >= 1313) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1393) {
if (mu_mv_mag_sq < 101) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
if (posBalance < 47) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
else if (posBalance >= 47) {
if (Cs < 1092) {
return 2;
}
else if (Cs >= 1092) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1869) {
if (TSC < 4) {
if (ssDCval < 82) {
if (refDCval < 71) {
if (ssDCval < 56) {
if (CsDiff < 2) {
if (Rs < 2028) {
return 4;
}
else if (Rs >= 2028) {
return 8;
}
}
else if (CsDiff >= 2) {
if (Cs < 1241) {
return 4;
}
else if (Cs >= 1241) {
return 8;
}
}
}
else if (ssDCval >= 56) {
if (posBalance < 36) {
if (Rs < 1926) {
if (refDCval < 61) {
if (Cs < 1062) {
if (Cs < 993) {
return 1;
}
else if (Cs >= 993) {
return 4;
}
}
else if (Cs >= 1062) {
return 1;
}
}
else if (refDCval >= 61) {
return 4;
}
}
else if (Rs >= 1926) {
if (SC < 3186) {
return 4;
}
else if (SC >= 3186) {
if (Rs < 2271) {
return 8;
}
else if (Rs >= 2271) {
return 2;
}
}
}
}
else if (posBalance >= 36) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 100) {
if (Rs < 2255) {
return 4;
}
else if (Rs >= 2255) {
return 1;
}
}
else if (mu_mv_mag_sq >= 100) {
if (AFD < 2) {
if (Rs < 1985) {
return 4;
}
else if (Rs >= 1985) {
return 1;
}
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 71) {
if (posBalance < 47) {
if (SC < 3009) {
return 1;
}
else if (SC >= 3009) {
if (AFD < 3) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
return 8;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (posBalance >= 47) {
return 8;
}
}
}
else if (ssDCval >= 82) {
if (mu_mv_mag_sq < 100) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 981) {
return 4;
}
else if (Cs >= 981) {
return 8;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 972) {
if (posBalance < 53) {
if (Rs < 2015) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
return 1;
}
}
else if (Rs >= 2015) {
if (mu_mv_mag_sq < 3) {
if (RsDiff < 11) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 2241) {
return 1;
}
else if (Rs >= 2241) {
return 2;
}
}
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (posBalance < 32) {
if (Cs < 862) {
if (Rs < 2156) {
return 4;
}
else if (Rs >= 2156) {
if (Rs < 2256) {
return 4;
}
else if (Rs >= 2256) {
return 8;
}
}
}
else if (Cs >= 862) {
return 1;
}
}
else if (posBalance >= 32) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
return 1;
}
}
}
}
}
else if (posBalance >= 53) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (Cs >= 972) {
if (CsDiff < 6) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
if (diffRsCsdiff < -68) {
if (Rs < 2287) {
return 1;
}
else if (Rs >= 2287) {
return 8;
}
}
else if (diffRsCsdiff >= -68) {
return 8;
}
}
}
else if (CsDiff >= 6) {
if (SC < 3173) {
if (Rs < 2051) {
return 1;
}
else if (Rs >= 2051) {
return 4;
}
}
else if (SC >= 3173) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 88) {
if (mu_mv_mag_sq < 100) {
if (SC < 1302) {
if (MVDiff < 4) {
if (Cs < 737) {
if (Cs < 561) {
if (mu_mv_mag_sq < 85) {
if (Rs < 518) {
return 1;
}
else if (Rs >= 518) {
return 2;
}
}
else if (mu_mv_mag_sq >= 85) {
if (AFD < 3) {
if (Rs < 387) {
return 4;
}
else if (Rs >= 387) {
return 2;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (Cs >= 561) {
if (Rs < 593) {
if (mcTcor < 99) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
return 1;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (Rs >= 593) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
return 2;
}
}
}
}
else if (Cs >= 737) {
if (Rs < 420) {
return 4;
}
else if (Rs >= 420) {
return 1;
}
}
}
else if (MVDiff >= 4) {
if (posBalance < 36) {
return 8;
}
else if (posBalance >= 36) {
if (Cs < 738) {
if (negBalance < 30) {
if (AFD < 7) {
if (mu_mv_mag_sq < 62) {
return 1;
}
else if (mu_mv_mag_sq >= 62) {
if (posBalance < 51) {
if (Rs < 358) {
return 2;
}
else if (Rs >= 358) {
return 1;
}
}
else if (posBalance >= 51) {
return 2;
}
}
}
else if (AFD >= 7) {
if (Rs < 646) {
return 8;
}
else if (Rs >= 646) {
return 4;
}
}
}
else if (negBalance >= 30) {
if (posBalance < 56) {
if (MVDiff < 14) {
if (MVDiff < 10) {
if (mu_mv_mag_sq < 3) {
if (Rs < 354) {
return 2;
}
else if (Rs >= 354) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 2;
}
}
else if (MVDiff >= 10) {
if (diffRsCsdiff < -5) {
return 2;
}
else if (diffRsCsdiff >= -5) {
if (Rs < 547) {
return 8;
}
else if (Rs >= 547) {
return 2;
}
}
}
}
else if (MVDiff >= 14) {
if (Cs < 549) {
return 1;
}
else if (Cs >= 549) {
return 2;
}
}
}
else if (posBalance >= 56) {
if (ssDCval < 146) {
if (mu_mv_mag_sq < 6) {
return 4;
}
else if (mu_mv_mag_sq >= 6) {
if (negBalance < 40) {
if (ssDCval < 125) {
return 2;
}
else if (ssDCval >= 125) {
return 8;
}
}
else if (negBalance >= 40) {
return 8;
}
}
}
else if (ssDCval >= 146) {
if (Rs < 503) {
if (Rs < 434) {
return 4;
}
else if (Rs >= 434) {
return 1;
}
}
else if (Rs >= 503) {
return 2;
}
}
}
}
}
else if (Cs >= 738) {
if (MVDiff < 6) {
if (Rs < 482) {
return 4;
}
else if (Rs >= 482) {
return 2;
}
}
else if (MVDiff >= 6) {
return 2;
}
}
}
}
}
else if (SC >= 1302) {
if (refDCval < 102) {
if (TSC < 5) {
if (Rs < 639) {
if (Rs < 494) {
return 8;
}
else if (Rs >= 494) {
return 2;
}
}
else if (Rs >= 639) {
if (MVDiff < 1) {
if (Rs < 1608) {
return 8;
}
else if (Rs >= 1608) {
return 4;
}
}
else if (MVDiff >= 1) {
if (Cs < 1146) {
if (SC < 2642) {
if (mu_mv_mag_sq < 73) {
if (MVDiff < 29) {
return 1;
}
else if (MVDiff >= 29) {
return 8;
}
}
else if (mu_mv_mag_sq >= 73) {
if (Rs < 1007) {
return 1;
}
else if (Rs >= 1007) {
return 8;
}
}
}
else if (SC >= 2642) {
return 1;
}
}
else if (Cs >= 1146) {
if (Rs < 1337) {
return 4;
}
else if (Rs >= 1337) {
return 1;
}
}
}
}
}
else if (TSC >= 5) {
if (tcor < 96) {
if (TSC < 7) {
if (posBalance < 55) {
if (ssDCval < 92) {
if (Cs < 814) {
return 1;
}
else if (Cs >= 814) {
if (Cs < 895) {
return 4;
}
else if (Cs >= 895) {
if (RsCsDiff < 29) {
if (Rs < 954) {
return 4;
}
else if (Rs >= 954) {
return 8;
}
}
else if (RsCsDiff >= 29) {
return 1;
}
}
}
}
else if (ssDCval >= 92) {
if (refDCval < 101) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 1;
}
}
else if (refDCval >= 101) {
if (Rs < 2023) {
return 1;
}
else if (Rs >= 2023) {
return 4;
}
}
}
}
else if (posBalance >= 55) {
if (Cs < 1005) {
if (ssDCval < 95) {
return 1;
}
else if (ssDCval >= 95) {
return 4;
}
}
else if (Cs >= 1005) {
if (Rs < 1099) {
return 1;
}
else if (Rs >= 1099) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (AFD < 28) {
if (diffAFD < -2) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Cs < 1224) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (SC < 1944) {
return 8;
}
else if (SC >= 1944) {
return 1;
}
}
}
else if (Cs >= 1224) {
return 1;
}
}
}
else if (AFD >= 28) {
return 4;
}
}
}
else if (tcor >= 96) {
if (MVDiff < 2) {
if (Rs < 1466) {
return 4;
}
else if (Rs >= 1466) {
return 1;
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 66) {
if (negBalance < 47) {
if (Cs < 1115) {
if (Rs < 640) {
return 2;
}
else if (Rs >= 640) {
return 4;
}
}
else if (Cs >= 1115) {
return 1;
}
}
else if (negBalance >= 47) {
if (Rs < 2125) {
return 8;
}
else if (Rs >= 2125) {
if (Rs < 2195) {
return 1;
}
else if (Rs >= 2195) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 102) {
if (Cs < 637) {
if (TSC < 5) {
if (Rs < 1210) {
if (Cs < 552) {
return 2;
}
else if (Cs >= 552) {
return 1;
}
}
else if (Rs >= 1210) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 8;
}
}
}
else if (TSC >= 5) {
if (Rs < 1274) {
if (negBalance < 51) {
if (ssDCval < 154) {
return 2;
}
else if (ssDCval >= 154) {
return 8;
}
}
else if (negBalance >= 51) {
if (Cs < 573) {
return 1;
}
else if (Cs >= 573) {
return 8;
}
}
}
else if (Rs >= 1274) {
if (mu_mv_mag_sq < 33) {
return 1;
}
else if (mu_mv_mag_sq >= 33) {
if (Rs < 2039) {
if (Rs < 1770) {
return 8;
}
else if (Rs >= 1770) {
return 1;
}
}
else if (Rs >= 2039) {
return 2;
}
}
}
}
}
else if (Cs >= 637) {
if (MVDiff < 5) {
if (SC < 1894) {
if (TSC < 4) {
if (negBalance < 26) {
if (Cs < 1089) {
return 4;
}
else if (Cs >= 1089) {
return 1;
}
}
else if (negBalance >= 26) {
if (Rs < 781) {
return 1;
}
else if (Rs >= 781) {
return 8;
}
}
}
else if (TSC >= 4) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (SC >= 1894) {
if (mu_mv_mag_sq < 86) {
if (SC < 2646) {
if (posBalance < 59) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
return 8;
}
}
else if (posBalance >= 59) {
return 4;
}
}
else if (SC >= 2646) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (negBalance < 28) {
if (SC < 2723) {
if (Rs < 1653) {
return 8;
}
else if (Rs >= 1653) {
return 4;
}
}
else if (SC >= 2723) {
return 2;
}
}
else if (negBalance >= 28) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
else if (MVDiff >= 5) {
if (AFD < 9) {
if (negBalance < 29) {
if (Rs < 1521) {
return 8;
}
else if (Rs >= 1521) {
return 1;
}
}
else if (negBalance >= 29) {
if (ssDCval < 161) {
if (diffRsCsdiff < 10) {
return 8;
}
else if (diffRsCsdiff >= 10) {
if (Rs < 1064) {
return 1;
}
else if (Rs >= 1064) {
return 8;
}
}
}
else if (ssDCval >= 161) {
if (Rs < 723) {
return 4;
}
else if (Rs >= 723) {
return 2;
}
}
}
}
else if (AFD >= 9) {
if (refDCval < 149) {
if (mu_mv_mag_sq < 3) {
if (tcor < 90) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 2;
}
}
else if (tcor >= 90) {
if (negBalance < 50) {
if (RsDiff < 8) {
if (Rs < 862) {
return 4;
}
else if (Rs >= 862) {
return 8;
}
}
else if (RsDiff >= 8) {
return 1;
}
}
else if (negBalance >= 50) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 121) {
if (Rs < 2190) {
if (RsCsDiff < 73) {
return 8;
}
else if (RsCsDiff >= 73) {
if (ssDCval < 112) {
if (Cs < 1313) {
return 8;
}
else if (Cs >= 1313) {
return 2;
}
}
else if (ssDCval >= 112) {
return 2;
}
}
}
else if (Rs >= 2190) {
if (Rs < 2276) {
return 2;
}
else if (Rs >= 2276) {
return 8;
}
}
}
else if (ssDCval >= 121) {
if (ssDCval < 138) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (AFD < 19) {
if (Rs < 2178) {
if (ssDCval < 135) {
if (Cs < 752) {
if (Rs < 1849) {
return 8;
}
else if (Rs >= 1849) {
return 1;
}
}
else if (Cs >= 752) {
return 8;
}
}
else if (ssDCval >= 135) {
return 2;
}
}
else if (Rs >= 2178) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 2;
}
}
}
else if (AFD >= 19) {
if (Rs < 1939) {
return 2;
}
else if (Rs >= 1939) {
return 1;
}
}
}
}
else if (ssDCval >= 138) {
if (Rs < 922) {
return 1;
}
else if (Rs >= 922) {
return 8;
}
}
}
}
}
else if (refDCval >= 149) {
if (refDCval < 170) {
if (diffRsCsdiff < -5) {
if (posBalance < 64) {
return 2;
}
else if (posBalance >= 64) {
if (Rs < 1563) {
return 4;
}
else if (Rs >= 1563) {
return 1;
}
}
}
else if (diffRsCsdiff >= -5) {
return 1;
}
}
else if (refDCval >= 170) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (ssDCval < 133) {
if (Cs < 1227) {
if (Cs < 928) {
if (refDCval < 97) {
if (ssDCval < 93) {
if (Rs < 2145) {
if (Cs < 661) {
if (Rs < 2031) {
return 4;
}
else if (Rs >= 2031) {
return 1;
}
}
else if (Cs >= 661) {
return 8;
}
}
else if (Rs >= 2145) {
if (Cs < 745) {
return 8;
}
else if (Cs >= 745) {
if (diffRsCsdiff < 1) {
return 4;
}
else if (diffRsCsdiff >= 1) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
return 4;
}
}
}
}
}
else if (ssDCval >= 93) {
if (Rs < 940) {
if (Cs < 657) {
return 8;
}
else if (Cs >= 657) {
return 1;
}
}
else if (Rs >= 940) {
if (negBalance < 35) {
if (mu_mv_mag_sq < 126) {
return 4;
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 694) {
return 4;
}
else if (Cs >= 694) {
return 8;
}
}
}
else if (negBalance >= 35) {
return 8;
}
}
}
}
else if (refDCval >= 97) {
if (Rs < 330) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 330) {
if (Rs < 560) {
if (ssDCval < 102) {
if (Cs < 790) {
return 1;
}
else if (Cs >= 790) {
return 8;
}
}
else if (ssDCval >= 102) {
if (mu_mv_mag_sq < 105) {
if (TSC < 4) {
if (ssDCval < 117) {
if (Rs < 402) {
return 4;
}
else if (Rs >= 402) {
return 1;
}
}
else if (ssDCval >= 117) {
return 4;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 105) {
if (negBalance < 23) {
if (Rs < 485) {
return 4;
}
else if (Rs >= 485) {
return 1;
}
}
else if (negBalance >= 23) {
if (Cs < 899) {
return 8;
}
else if (Cs >= 899) {
if (Rs < 369) {
return 4;
}
else if (Rs >= 369) {
return 8;
}
}
}
}
}
}
else if (Rs >= 560) {
if (SC < 2425) {
if (diffRsCsdiff < -19) {
return 1;
}
else if (diffRsCsdiff >= -19) {
if (mu_mv_mag_sq < 117) {
if (negBalance < 28) {
if (Rs < 772) {
if (Cs < 609) {
return 8;
}
else if (Cs >= 609) {
return 1;
}
}
else if (Rs >= 772) {
return 8;
}
}
else if (negBalance >= 28) {
if (mu_mv_mag_sq < 101) {
return 2;
}
else if (mu_mv_mag_sq >= 101) {
if (Cs < 883) {
return 8;
}
else if (Cs >= 883) {
if (Rs < 660) {
return 8;
}
else if (Rs >= 660) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
}
else if (SC >= 2425) {
if (MVDiff < 1) {
if (ssDCval < 107) {
return 8;
}
else if (ssDCval >= 107) {
return 1;
}
}
else if (MVDiff >= 1) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 928) {
if (ssDCval < 121) {
if (ssDCval < 94) {
if (tcor < 98) {
if (diffMVdiffVal < -1) {
if (Rs < 1247) {
return 1;
}
else if (Rs >= 1247) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (tcor >= 98) {
return 8;
}
}
else if (ssDCval >= 94) {
if (Cs < 1220) {
if (diffRsCsdiff < -46) {
if (Rs < 1256) {
return 8;
}
else if (Rs >= 1256) {
return 2;
}
}
else if (diffRsCsdiff >= -46) {
return 8;
}
}
else if (Cs >= 1220) {
if (Rs < 1202) {
return 1;
}
else if (Rs >= 1202) {
return 8;
}
}
}
}
else if (ssDCval >= 121) {
if (Rs < 1488) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 1488) {
return 2;
}
}
}
}
else if (Cs >= 1227) {
if (SC < 1936) {
return 1;
}
else if (SC >= 1936) {
if (AFD < 2) {
if (Cs < 1283) {
return 4;
}
else if (Cs >= 1283) {
if (Rs < 1288) {
return 1;
}
else if (Rs >= 1288) {
return 8;
}
}
}
else if (AFD >= 2) {
return 8;
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 145) {
return 1;
}
else if (ssDCval >= 145) {
if (Rs < 995) {
return 8;
}
else if (Rs >= 995) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2371) {
if (MVDiff < 14) {
if (refDCval < 109) {
if (posBalance < 48) {
if (RsCsDiff < 10) {
if (Cs < 1052) {
if (ssDCval < 86) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Rs < 2375) {
return 4;
}
else if (Rs >= 2375) {
return 8;
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
if (Rs < 2596) {
return 4;
}
else if (Rs >= 2596) {
return 8;
}
}
}
}
else if (Cs >= 1052) {
if (SC < 4573) {
if (Rs < 3122) {
if (negBalance < 41) {
return 8;
}
else if (negBalance >= 41) {
return 4;
}
}
else if (Rs >= 3122) {
if (SC < 4381) {
return 2;
}
else if (SC >= 4381) {
return 8;
}
}
}
else if (SC >= 4573) {
if (SC < 5403) {
if (ssDCval < 74) {
if (ssDCval < 54) {
if (Cs < 1243) {
return 8;
}
else if (Cs >= 1243) {
return 4;
}
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (ssDCval >= 74) {
if (diffRsCsdiff < -8) {
return 8;
}
else if (diffRsCsdiff >= -8) {
if (Cs < 1297) {
return 4;
}
else if (Cs >= 1297) {
if (Rs < 3774) {
return 4;
}
else if (Rs >= 3774) {
return 8;
}
}
}
}
}
else if (SC >= 5403) {
if (Rs < 4566) {
return 8;
}
else if (Rs >= 4566) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 10) {
if (ssDCval < 99) {
if (negBalance < 51) {
if (Rs < 3877) {
if (refDCval < 66) {
if (MVDiff < 6) {
if (ssDCval < 63) {
if (mcTcor < 95) {
if (ssDCval < 46) {
if (Rs < 2634) {
return 4;
}
else if (Rs >= 2634) {
return 8;
}
}
else if (ssDCval >= 46) {
return 8;
}
}
else if (mcTcor >= 95) {
if (AFD < 6) {
if (ssDCval < 58) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
return 8;
}
}
}
else if (ssDCval >= 58) {
if (Rs < 3155) {
return 2;
}
else if (Rs >= 3155) {
return 8;
}
}
}
else if (AFD >= 6) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (Cs < 1003) {
return 2;
}
else if (Cs >= 1003) {
return 8;
}
}
}
}
}
else if (ssDCval >= 63) {
if (SC < 4332) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (tcor < 91) {
if (Cs < 1043) {
return 2;
}
else if (Cs >= 1043) {
return 4;
}
}
else if (tcor >= 91) {
return 2;
}
}
}
}
else if (SC >= 4332) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
else if (MVDiff >= 6) {
if (CsDiff < 6) {
if (Rs < 2458) {
if (posBalance < 45) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 2;
}
}
else if (posBalance >= 45) {
if (Cs < 914) {
return 2;
}
else if (Cs >= 914) {
return 8;
}
}
}
else if (Rs >= 2458) {
if (AFD < 17) {
if (negBalance < 45) {
if (diffTSC < -1) {
if (Rs < 2580) {
if (Rs < 2472) {
return 4;
}
else if (Rs >= 2472) {
return 2;
}
}
else if (Rs >= 2580) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (negBalance >= 45) {
return 1;
}
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (Cs < 989) {
return 4;
}
else if (Cs >= 989) {
return 2;
}
}
}
}
}
else if (refDCval >= 66) {
if (Rs < 2959) {
if (Cs < 452) {
if (Rs < 2612) {
return 1;
}
else if (Rs >= 2612) {
return 8;
}
}
else if (Cs >= 452) {
if (mu_mv_mag_sq < 80) {
if (CsDiff < 5) {
if (TSC < 6) {
if (ssDCval < 74) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
else if (ssDCval >= 74) {
if (Rs < 2925) {
return 1;
}
else if (Rs >= 2925) {
return 2;
}
}
}
else if (TSC >= 6) {
if (negBalance < 42) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (negBalance >= 42) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
else if (CsDiff >= 5) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
if (mcTcor < 97) {
if (Rs < 2706) {
if (Rs < 2455) {
return 2;
}
else if (Rs >= 2455) {
return 1;
}
}
else if (Rs >= 2706) {
return 2;
}
}
else if (mcTcor >= 97) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
if (MVDiff < 3) {
if (Rs < 2696) {
return 1;
}
else if (Rs >= 2696) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
}
else if (Rs >= 2959) {
if (refDCval < 77) {
if (Cs < 971) {
return 4;
}
else if (Cs >= 971) {
if (RsDiff < 8) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (SC < 4168) {
if (Rs < 3121) {
return 2;
}
else if (Rs >= 3121) {
return 1;
}
}
else if (SC >= 4168) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Rs < 3442) {
return 2;
}
else if (Rs >= 3442) {
return 4;
}
}
}
}
}
else if (RsDiff >= 8) {
if (negBalance < 44) {
if (Cs < 1124) {
if (ssDCval < 68) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 2;
}
}
else if (ssDCval >= 68) {
if (mcTcor < 97) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
if (Cs < 1007) {
return 1;
}
else if (Cs >= 1007) {
return 4;
}
}
}
else if (mcTcor >= 97) {
return 4;
}
}
}
else if (Cs >= 1124) {
if (posBalance < 42) {
if (Rs < 3528) {
return 4;
}
else if (Rs >= 3528) {
return 1;
}
}
else if (posBalance >= 42) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 3685) {
return 2;
}
else if (Rs >= 3685) {
return 1;
}
}
}
}
}
else if (refDCval >= 77) {
if (mu_mv_mag_sq < 66) {
if (tcor < 96) {
if (diffTSC < -1) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 2;
}
}
else if (diffTSC >= -1) {
if (Cs < 1140) {
if (Cs < 1077) {
if (diffAFD < -1) {
if (Rs < 3326) {
return 4;
}
else if (Rs >= 3326) {
return 4;
}
}
else if (diffAFD >= -1) {
if (gchDC < 1) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
if (Rs < 3499) {
return 8;
}
else if (Rs >= 3499) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (MVDiff < 10) {
if (Rs < 3790) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
return 4;
}
}
else if (Rs >= 3790) {
return 8;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
}
}
else if (Cs >= 1077) {
return 8;
}
}
else if (Cs >= 1140) {
return 4;
}
}
}
else if (tcor >= 96) {
if (Rs < 3228) {
return 8;
}
else if (Rs >= 3228) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (mu_mv_mag_sq < 96) {
return 2;
}
else if (mu_mv_mag_sq >= 96) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 3877) {
if (negBalance < 27) {
if (Cs < 1077) {
return 4;
}
else if (Cs >= 1077) {
if (RsDiff < 5) {
if (Rs < 4296) {
return 4;
}
else if (Rs >= 4296) {
return 2;
}
}
else if (RsDiff >= 5) {
return 1;
}
}
}
else if (negBalance >= 27) {
if (mcTcor < 96) {
if (Cs < 1117) {
return 4;
}
else if (Cs >= 1117) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (posBalance < 45) {
if (Cs < 1161) {
return 8;
}
else if (Cs >= 1161) {
return 8;
}
}
else if (posBalance >= 45) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (Rs < 4179) {
if (Rs < 4069) {
return 4;
}
else if (Rs >= 4069) {
if (Rs < 4137) {
return 4;
}
else if (Rs >= 4137) {
return 4;
}
}
}
else if (Rs >= 4179) {
if (Rs < 4268) {
return 4;
}
else if (Rs >= 4268) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
if (Cs < 598) {
return 2;
}
else if (Cs >= 598) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 51) {
if (RsDiff < 4) {
if (Rs < 2614) {
if (Rs < 2511) {
return 8;
}
else if (Rs >= 2511) {
return 4;
}
}
else if (Rs >= 2614) {
return 1;
}
}
else if (RsDiff >= 4) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Rs < 3093) {
return 1;
}
else if (Rs >= 3093) {
return 2;
}
}
}
}
}
else if (ssDCval >= 99) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 96) {
if (tcor < 92) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
else if (tcor >= 92) {
if (diffMVdiffVal < -8) {
return 8;
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (Cs < 1143) {
if (Rs < 4701) {
return 2;
}
else if (Rs >= 4701) {
return 8;
}
}
else if (Cs >= 1143) {
return 4;
}
}
}
else if (negBalance >= 44) {
return 2;
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Rs < 3813) {
if (mcTcor < 99) {
if (ssDCval < 74) {
if (Cs < 674) {
if (mu_mv_mag_sq < 21) {
if (Rs < 2931) {
return 2;
}
else if (Rs >= 2931) {
return 4;
}
}
else if (mu_mv_mag_sq >= 21) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (Cs >= 674) {
if (Cs < 1059) {
return 2;
}
else if (Cs >= 1059) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
else if (ssDCval >= 74) {
if (negBalance < 17) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
else if (negBalance >= 17) {
if (MVDiff < 7) {
if (ssDCval < 75) {
if (Rs < 2541) {
return 1;
}
else if (Rs >= 2541) {
return 2;
}
}
else if (ssDCval >= 75) {
return 2;
}
}
else if (MVDiff >= 7) {
if (posBalance < 64) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 1;
}
}
else if (posBalance >= 64) {
return 2;
}
}
}
}
}
else if (mcTcor >= 99) {
if (SC < 3184) {
return 1;
}
else if (SC >= 3184) {
if (Rs < 3054) {
return 8;
}
else if (Rs >= 3054) {
return 1;
}
}
}
}
else if (Rs >= 3813) {
return 4;
}
}
else if (CsDiff >= 3) {
if (refDCval < 65) {
if (Cs < 1009) {
if (tcor < 86) {
if (SC < 3426) {
return 2;
}
else if (SC >= 3426) {
if (Cs < 629) {
return 2;
}
else if (Cs >= 629) {
return 4;
}
}
}
else if (tcor >= 86) {
if (RsDiff < 5) {
if (Rs < 2449) {
return 4;
}
else if (Rs >= 2449) {
return 2;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (Cs >= 1009) {
if (refDCval < 59) {
return 4;
}
else if (refDCval >= 59) {
if (diffRsCsdiff < 8) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (tcor < 92) {
if (Cs < 1290) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
return 4;
}
}
else if (Cs >= 1290) {
return 2;
}
}
else if (tcor >= 92) {
if (mu_mv_mag_sq < 98) {
return 2;
}
else if (mu_mv_mag_sq >= 98) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 8) {
return 4;
}
}
}
}
else if (refDCval >= 65) {
if (negBalance < 38) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 77) {
if (tcor < 90) {
if (posBalance < 49) {
if (Rs < 2696) {
return 1;
}
else if (Rs >= 2696) {
return 2;
}
}
else if (posBalance >= 49) {
if (Rs < 3668) {
return 4;
}
else if (Rs >= 3668) {
return 2;
}
}
}
else if (tcor >= 90) {
if (Rs < 2968) {
return 1;
}
else if (Rs >= 2968) {
if (Rs < 3464) {
return 2;
}
else if (Rs >= 3464) {
if (Rs < 3825) {
return 4;
}
else if (Rs >= 3825) {
if (Rs < 4187) {
return 8;
}
else if (Rs >= 4187) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
return 2;
}
}
else if (posBalance >= 55) {
if (tcor < 85) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 2;
}
}
else if (tcor >= 85) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (TSC < 7) {
if (Rs < 2793) {
return 1;
}
else if (Rs >= 2793) {
return 8;
}
}
else if (TSC >= 7) {
return 2;
}
}
}
}
}
else if (negBalance >= 38) {
if (diffRsCsdiff < -165) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (diffRsCsdiff >= -165) {
if (Rs < 2453) {
if (SC < 3536) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 2;
}
}
else if (SC >= 3536) {
if (Rs < 2418) {
if (Rs < 2393) {
return 8;
}
else if (Rs >= 2393) {
return 2;
}
}
else if (Rs >= 2418) {
return 4;
}
}
}
else if (Rs >= 2453) {
if (tcor < 77) {
if (SC < 3911) {
return 2;
}
else if (SC >= 3911) {
if (Rs < 3636) {
return 4;
}
else if (Rs >= 3636) {
if (Rs < 4082) {
return 2;
}
else if (Rs >= 4082) {
return 1;
}
}
}
}
else if (tcor >= 77) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsDiff < 11) {
if (tcor < 96) {
if (RsDiff < 9) {
if (negBalance < 31) {
if (TSC < 6) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (TSC >= 6) {
if (Rs < 2734) {
if (Rs < 2561) {
return 1;
}
else if (Rs >= 2561) {
if (Rs < 2669) {
return 4;
}
else if (Rs >= 2669) {
return 4;
}
}
}
else if (Rs >= 2734) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
else if (negBalance >= 31) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (TSC < 9) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (Rs < 3191) {
if (Cs < 679) {
return 2;
}
else if (Cs >= 679) {
return 4;
}
}
else if (Rs >= 3191) {
return 8;
}
}
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (tcor >= 96) {
if (posBalance < 56) {
if (Rs < 3180) {
return 8;
}
else if (Rs >= 3180) {
return 2;
}
}
else if (posBalance >= 56) {
return 2;
}
}
}
else if (RsDiff >= 11) {
if (Cs < 884) {
return 2;
}
else if (Cs >= 884) {
if (negBalance < 51) {
if (ssDCval < 106) {
return 2;
}
else if (ssDCval >= 106) {
if (negBalance < 47) {
if (Cs < 1159) {
return 2;
}
else if (Cs >= 1159) {
return 8;
}
}
else if (negBalance >= 47) {
return 4;
}
}
}
else if (negBalance >= 51) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (ssDCval < 104) {
if (Rs < 3263) {
if (negBalance < 57) {
return 4;
}
else if (negBalance >= 57) {
return 8;
}
}
else if (Rs >= 3263) {
return 8;
}
}
else if (ssDCval >= 104) {
if (RsDiff < 16) {
return 4;
}
else if (RsDiff >= 16) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 3356) {
if (mcTcor < 97) {
if (MVDiff < 5) {
if (ssDCval < 144) {
if (SC < 4371) {
return 8;
}
else if (SC >= 4371) {
return 2;
}
}
else if (ssDCval >= 144) {
return 2;
}
}
else if (MVDiff >= 5) {
if (RsDiff < 7) {
if (mu_mv_mag_sq < 89) {
if (diffMVdiffVal < 4) {
if (tcor < 94) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
return 8;
}
}
else if (tcor >= 94) {
if (Cs < 1262) {
return 1;
}
else if (Cs >= 1262) {
return 8;
}
}
}
else if (diffMVdiffVal >= 4) {
return 1;
}
}
else if (mu_mv_mag_sq >= 89) {
if (Rs < 2444) {
return 1;
}
else if (Rs >= 2444) {
if (Rs < 2722) {
return 8;
}
else if (Rs >= 2722) {
return 2;
}
}
}
}
else if (RsDiff >= 7) {
if (tcor < 83) {
if (Cs < 636) {
return 2;
}
else if (Cs >= 636) {
if (tcor < 75) {
if (Rs < 2703) {
return 4;
}
else if (Rs >= 2703) {
if (Rs < 3199) {
return 1;
}
else if (Rs >= 3199) {
return 2;
}
}
}
else if (tcor >= 75) {
return 8;
}
}
}
else if (tcor >= 83) {
if (Rs < 2937) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (negBalance < 58) {
return 2;
}
else if (negBalance >= 58) {
return 1;
}
}
}
else if (Rs >= 2937) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 7) {
if (Rs < 2669) {
if (RsDiff < 5) {
if (Rs < 2458) {
return 8;
}
else if (Rs >= 2458) {
return 1;
}
}
else if (RsDiff >= 5) {
if (ssDCval < 126) {
return 1;
}
else if (ssDCval >= 126) {
return 2;
}
}
}
else if (Rs >= 2669) {
if (Cs < 521) {
if (SC < 3529) {
return 2;
}
else if (SC >= 3529) {
return 8;
}
}
else if (Cs >= 521) {
return 2;
}
}
}
else if (MVDiff >= 7) {
if (TSC < 6) {
if (mu_mv_mag_sq < 84) {
if (SC < 3480) {
if (SC < 2984) {
return 8;
}
else if (SC >= 2984) {
return 1;
}
}
else if (SC >= 3480) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (Rs < 2791) {
return 1;
}
else if (Rs >= 2791) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Cs < 1274) {
return 2;
}
else if (Cs >= 1274) {
return 8;
}
}
}
}
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 43) {
if (ssDCval < 115) {
if (posBalance < 63) {
return 2;
}
else if (posBalance >= 63) {
return 4;
}
}
else if (ssDCval >= 115) {
if (Cs < 1053) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 2;
}
}
else if (Cs >= 1053) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3356) {
if (MVDiff < 5) {
if (ssDCval < 142) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 24) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (CsDiff < 8) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
return 2;
}
}
else if (CsDiff >= 8) {
if (Rs < 4095) {
return 2;
}
else if (Rs >= 4095) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (diffRsCsdiff < 42) {
return 2;
}
else if (diffRsCsdiff >= 42) {
if (Cs < 991) {
return 2;
}
else if (Cs >= 991) {
return 1;
}
}
}
}
}
else if (ssDCval >= 142) {
if (Rs < 3963) {
if (Cs < 1136) {
return 2;
}
else if (Cs >= 1136) {
if (MVDiff < 1) {
if (Rs < 3745) {
return 2;
}
else if (Rs >= 3745) {
if (Rs < 3863) {
return 1;
}
else if (Rs >= 3863) {
return 2;
}
}
}
else if (MVDiff >= 1) {
if (posBalance < 36) {
return 2;
}
else if (posBalance >= 36) {
return 1;
}
}
}
}
else if (Rs >= 3963) {
return 2;
}
}
}
else if (MVDiff >= 5) {
if (refDCval < 135) {
if (TSC < 6) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (negBalance < 29) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (negBalance >= 29) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (Cs < 1260) {
if (refDCval < 132) {
return 2;
}
else if (refDCval >= 132) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (Rs < 5488) {
return 1;
}
else if (Rs >= 5488) {
return 2;
}
}
}
}
else if (Cs >= 1260) {
if (Rs < 4466) {
if (Rs < 3905) {
return 8;
}
else if (Rs >= 3905) {
return 4;
}
}
else if (Rs >= 4466) {
return 2;
}
}
}
}
else if (refDCval >= 135) {
if (negBalance < 54) {
if (Cs < 1008) {
if (Cs < 532) {
return 8;
}
else if (Cs >= 532) {
if (Rs < 5818) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (Rs < 3678) {
return 1;
}
else if (Rs >= 3678) {
return 2;
}
}
}
else if (Rs >= 5818) {
if (RsDiff < 14) {
return 1;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
}
else if (Cs >= 1008) {
if (CsDiff < 3) {
if (Cs < 1243) {
return 2;
}
else if (Cs >= 1243) {
return 8;
}
}
else if (CsDiff >= 3) {
if (Rs < 4897) {
if (posBalance < 40) {
return 2;
}
else if (posBalance >= 40) {
return 1;
}
}
else if (Rs >= 4897) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (Cs < 1048) {
return 4;
}
else if (Cs >= 1048) {
return 2;
}
}
}
}
}
}
}
}
else if (MVDiff >= 14) {
if (mcTcor < 69) {
if (MVDiff < 44) {
return 1;
}
else if (MVDiff >= 44) {
if (gchDC < 5) {
if (MVDiff < 57) {
if (Rs < 3645) {
return 1;
}
else if (Rs >= 3645) {
return 2;
}
}
else if (MVDiff >= 57) {
return 4;
}
}
else if (gchDC >= 5) {
return 8;
}
}
}
else if (mcTcor >= 69) {
if (RsCsDiff < 159) {
if (refDCval < 77) {
if (TSC < 8) {
if (RsCsDiff < 55) {
if (refDCval < 67) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
if (Rs < 3187) {
return 8;
}
else if (Rs >= 3187) {
return 4;
}
}
}
}
else if (refDCval >= 67) {
if (Rs < 2753) {
return 8;
}
else if (Rs >= 2753) {
return 1;
}
}
}
else if (RsCsDiff >= 55) {
if (Cs < 1062) {
if (RsCsDiff < 62) {
return 2;
}
else if (RsCsDiff >= 62) {
if (posBalance < 29) {
return 1;
}
else if (posBalance >= 29) {
return 4;
}
}
}
else if (Cs >= 1062) {
if (Rs < 2821) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (Rs < 2753) {
return 2;
}
else if (Rs >= 2753) {
return 4;
}
}
}
else if (Rs >= 2821) {
if (Rs < 7949) {
return 4;
}
else if (Rs >= 7949) {
return 2;
}
}
}
}
}
else if (TSC >= 8) {
if (ssDCval < 70) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (RsCsDiff < 128) {
if (RsCsDiff < 117) {
return 2;
}
else if (RsCsDiff >= 117) {
return 4;
}
}
else if (RsCsDiff >= 128) {
return 2;
}
}
}
else if (ssDCval >= 70) {
if (tcor < 81) {
if (Rs < 2976) {
return 1;
}
else if (Rs >= 2976) {
return 4;
}
}
else if (tcor >= 81) {
if (Cs < 1158) {
return 2;
}
else if (Cs >= 1158) {
if (Rs < 2490) {
return 8;
}
else if (Rs >= 2490) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 77) {
if (SC < 3224) {
if (Rs < 2602) {
if (tcor < 77) {
if (TSC < 15) {
return 8;
}
else if (TSC >= 15) {
return 1;
}
}
else if (tcor >= 77) {
if (tcor < 97) {
if (tcor < 85) {
if (diffMVdiffVal < -8) {
return 8;
}
else if (diffMVdiffVal >= -8) {
if (Cs < 594) {
if (diffTSC < -2) {
if (Cs < 472) {
return 8;
}
else if (Cs >= 472) {
return 1;
}
}
else if (diffTSC >= -2) {
return 1;
}
}
else if (Cs >= 594) {
return 2;
}
}
}
else if (tcor >= 85) {
if (AFD < 21) {
if (Cs < 739) {
return 1;
}
else if (Cs >= 739) {
if (Rs < 2441) {
return 1;
}
else if (Rs >= 2441) {
if (Rs < 2460) {
return 2;
}
else if (Rs >= 2460) {
return 4;
}
}
}
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (tcor >= 97) {
if (Rs < 2398) {
return 2;
}
else if (Rs >= 2398) {
return 8;
}
}
}
}
else if (Rs >= 2602) {
if (mu_mv_mag_sq < 47) {
if (Rs < 2863) {
if (ssDCval < 91) {
if (Cs < 232) {
return 2;
}
else if (Cs >= 232) {
if (Rs < 2751) {
return 4;
}
else if (Rs >= 2751) {
return 8;
}
}
}
else if (ssDCval >= 91) {
if (diffTSC < 1) {
if (posBalance < 64) {
if (MVDiff < 61) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 1;
}
}
else if (MVDiff >= 61) {
if (Cs < 229) {
return 2;
}
else if (Cs >= 229) {
return 8;
}
}
}
else if (posBalance >= 64) {
return 1;
}
}
else if (diffTSC >= 1) {
if (RsCsDiff < 151) {
return 8;
}
else if (RsCsDiff >= 151) {
return 2;
}
}
}
}
else if (Rs >= 2863) {
return 1;
}
}
else if (mu_mv_mag_sq >= 47) {
if (Rs < 2654) {
if (Rs < 2637) {
return 1;
}
else if (Rs >= 2637) {
return 8;
}
}
else if (Rs >= 2654) {
return 4;
}
}
}
}
else if (SC >= 3224) {
if (Rs < 3885) {
if (RsDiff < 9) {
if (Cs < 1091) {
if (refDCval < 101) {
if (diffMVdiffVal < -1) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Rs < 2629) {
return 2;
}
else if (Rs >= 2629) {
if (Rs < 2891) {
return 1;
}
else if (Rs >= 2891) {
return 8;
}
}
}
}
}
else if (refDCval >= 101) {
if (refDCval < 125) {
if (mu_mv_mag_sq < 68) {
if (refDCval < 104) {
if (SC < 3934) {
if (Rs < 2959) {
return 2;
}
else if (Rs >= 2959) {
return 4;
}
}
else if (SC >= 3934) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (refDCval >= 104) {
return 8;
}
}
else if (mu_mv_mag_sq >= 68) {
return 2;
}
}
else if (refDCval >= 125) {
if (diffRsCsdiff < -54) {
return 1;
}
else if (diffRsCsdiff >= -54) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
}
else if (Cs >= 1091) {
if (CsDiff < 6) {
if (tcor < 100) {
return 8;
}
else if (tcor >= 100) {
return 4;
}
}
else if (CsDiff >= 6) {
if (Rs < 2583) {
return 8;
}
else if (Rs >= 2583) {
return 1;
}
}
}
}
else if (RsDiff >= 9) {
if (MVDiff < 31) {
if (ssDCval < 98) {
if (diffTSC < -4) {
return 4;
}
else if (diffTSC >= -4) {
if (tcor < 88) {
if (Rs < 3079) {
return 4;
}
else if (Rs >= 3079) {
return 2;
}
}
else if (tcor >= 88) {
if (Cs < 223) {
return 1;
}
else if (Cs >= 223) {
return 2;
}
}
}
}
else if (ssDCval >= 98) {
if (refDCval < 119) {
if (Rs < 2970) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (mu_mv_mag_sq < 20) {
return 1;
}
else if (mu_mv_mag_sq >= 20) {
return 8;
}
}
}
else if (Rs >= 2970) {
if (Rs < 3555) {
if (Cs < 604) {
return 8;
}
else if (Cs >= 604) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
return 2;
}
}
}
else if (Rs >= 3555) {
return 8;
}
}
}
else if (refDCval >= 119) {
if (mcTcor < 94) {
if (TSC < 9) {
if (Rs < 3078) {
return 1;
}
else if (Rs >= 3078) {
return 2;
}
}
else if (TSC >= 9) {
return 4;
}
}
else if (mcTcor >= 94) {
if (mu_mv_mag_sq < 34) {
if (CsDiff < 3) {
if (Rs < 2839) {
return 1;
}
else if (Rs >= 2839) {
return 2;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 34) {
return 2;
}
}
}
}
}
else if (MVDiff >= 31) {
if (AFD < 22) {
if (AFD < 13) {
if (MVDiff < 50) {
if (posBalance < 59) {
if (Cs < 443) {
return 4;
}
else if (Cs >= 443) {
return 8;
}
}
else if (posBalance >= 59) {
if (Rs < 3158) {
return 1;
}
else if (Rs >= 3158) {
return 2;
}
}
}
else if (MVDiff >= 50) {
return 1;
}
}
else if (AFD >= 13) {
if (ssDCval < 132) {
if (Rs < 2393) {
return 2;
}
else if (Rs >= 2393) {
if (refDCval < 123) {
return 8;
}
else if (refDCval >= 123) {
if (Rs < 3146) {
return 1;
}
else if (Rs >= 3146) {
return 8;
}
}
}
}
else if (ssDCval >= 132) {
if (Rs < 3096) {
if (Rs < 2676) {
return 8;
}
else if (Rs >= 2676) {
return 4;
}
}
else if (Rs >= 3096) {
return 2;
}
}
}
}
else if (AFD >= 22) {
if (refDCval < 99) {
if (RsDiff < 11) {
if (Rs < 2974) {
return 4;
}
else if (Rs >= 2974) {
return 8;
}
}
else if (RsDiff >= 11) {
return 1;
}
}
else if (refDCval >= 99) {
if (diffRsCsdiff < 22) {
if (SC < 4166) {
if (mu_mv_mag_sq < 2) {
if (Rs < 3195) {
return 8;
}
else if (Rs >= 3195) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
return 2;
}
}
else if (SC >= 4166) {
return 4;
}
}
else if (diffRsCsdiff >= 22) {
if (Rs < 3256) {
return 8;
}
else if (Rs >= 3256) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 3885) {
if (negBalance < 40) {
if (Cs < 1011) {
return 2;
}
else if (Cs >= 1011) {
if (Cs < 1101) {
return 1;
}
else if (Cs >= 1101) {
return 2;
}
}
}
else if (negBalance >= 40) {
if (Cs < 908) {
if (Rs < 6013) {
return 4;
}
else if (Rs >= 6013) {
return 2;
}
}
else if (Cs >= 908) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (Rs < 4096) {
return 8;
}
else if (Rs >= 4096) {
return 4;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 159) {
if (ssDCval < 133) {
if (refDCval < 114) {
if (refDCval < 74) {
if (Rs < 2909) {
if (SC < 2676) {
return 1;
}
else if (SC >= 2676) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Cs < 177) {
if (Cs < 151) {
return 1;
}
else if (Cs >= 151) {
return 4;
}
}
else if (Cs >= 177) {
if (TSC < 12) {
if (diffRsCsdiff < 37) {
if (refDCval < 72) {
return 2;
}
else if (refDCval >= 72) {
if (Rs < 2412) {
return 1;
}
else if (Rs >= 2412) {
return 2;
}
}
}
else if (diffRsCsdiff >= 37) {
if (SC < 3826) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 4;
}
}
else if (SC >= 3826) {
return 4;
}
}
}
else if (TSC >= 12) {
if (Cs < 576) {
return 2;
}
else if (Cs >= 576) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2909) {
if (TSC < 10) {
if (mu_mv_mag_sq < 83) {
if (Rs < 3536) {
if (tcor < 82) {
if (Cs < 112) {
return 1;
}
else if (Cs >= 112) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (tcor >= 82) {
if (Cs < 750) {
if (Rs < 3229) {
return 4;
}
else if (Rs >= 3229) {
return 1;
}
}
else if (Cs >= 750) {
return 2;
}
}
}
else if (Rs >= 3536) {
return 4;
}
}
else if (mu_mv_mag_sq >= 83) {
return 8;
}
}
else if (TSC >= 10) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
if (Rs < 3606) {
return 1;
}
else if (Rs >= 3606) {
return 4;
}
}
}
}
}
else if (refDCval >= 74) {
if (MVDiff < 28) {
if (Cs < 1056) {
if (mu_mv_mag_sq < 18) {
if (mcTcor < 91) {
if (Cs < 681) {
if (Rs < 4929) {
return 2;
}
else if (Rs >= 4929) {
return 4;
}
}
else if (Cs >= 681) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
if (Rs < 3130) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 4;
}
}
else if (Rs >= 3130) {
return 8;
}
}
}
}
else if (mcTcor >= 91) {
if (Rs < 3778) {
if (refDCval < 106) {
return 2;
}
else if (refDCval >= 106) {
if (Rs < 2744) {
return 4;
}
else if (Rs >= 2744) {
return 2;
}
}
}
else if (Rs >= 3778) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
if (Rs < 3834) {
return 8;
}
else if (Rs >= 3834) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (Cs < 413) {
return 2;
}
else if (Cs >= 413) {
if (TSC < 6) {
if (Rs < 5334) {
return 8;
}
else if (Rs >= 5334) {
return 2;
}
}
else if (TSC >= 6) {
if (diffMVdiffVal < -12) {
return 1;
}
else if (diffMVdiffVal >= -12) {
if (posBalance < 77) {
if (Cs < 421) {
return 1;
}
else if (Cs >= 421) {
if (MVDiff < 25) {
return 2;
}
else if (MVDiff >= 25) {
if (Rs < 2874) {
return 1;
}
else if (Rs >= 2874) {
return 2;
}
}
}
}
else if (posBalance >= 77) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1056) {
if (Rs < 3299) {
if (Cs < 1116) {
return 2;
}
else if (Cs >= 1116) {
if (Cs < 1264) {
if (tcor < 84) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (RsDiff < 18) {
return 4;
}
else if (RsDiff >= 18) {
return 8;
}
}
}
else if (tcor >= 84) {
if (Rs < 2644) {
return 8;
}
else if (Rs >= 2644) {
return 2;
}
}
}
else if (Cs >= 1264) {
if (diffRsCsdiff < 106) {
return 8;
}
else if (diffRsCsdiff >= 106) {
return 2;
}
}
}
}
else if (Rs >= 3299) {
if (Cs < 1072) {
return 8;
}
else if (Cs >= 1072) {
if (tcor < 69) {
if (Rs < 3479) {
return 8;
}
else if (Rs >= 3479) {
return 2;
}
}
else if (tcor >= 69) {
if (RsCsDiff < 200) {
if (Rs < 3464) {
return 2;
}
else if (Rs >= 3464) {
return 1;
}
}
else if (RsCsDiff >= 200) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 28) {
if (RsCsDiff < 497) {
if (TSC < 8) {
if (ssDCval < 86) {
if (RsDiff < 15) {
return 4;
}
else if (RsDiff >= 15) {
return 2;
}
}
else if (ssDCval >= 86) {
if (AFD < 29) {
return 2;
}
else if (AFD >= 29) {
return 8;
}
}
}
else if (TSC >= 8) {
if (tcor < 55) {
if (Cs < 995) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
if (Rs < 2943) {
return 1;
}
else if (Rs >= 2943) {
return 8;
}
}
}
else if (Cs >= 995) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
if (Rs < 3063) {
return 1;
}
else if (Rs >= 3063) {
if (Rs < 3464) {
return 4;
}
else if (Rs >= 3464) {
return 2;
}
}
}
}
}
else if (tcor >= 55) {
if (CsDiff < 4) {
if (Cs < 388) {
if (RsDiff < 15) {
if (Cs < 353) {
return 4;
}
else if (Cs >= 353) {
return 2;
}
}
else if (RsDiff >= 15) {
return 2;
}
}
else if (Cs >= 388) {
if (diffRsCsdiff < -81) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 4;
}
}
else if (diffRsCsdiff >= -81) {
if (RsDiff < 16) {
return 8;
}
else if (RsDiff >= 16) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 8;
}
}
}
}
}
}
else if (CsDiff >= 4) {
if (ssDCval < 95) {
if (mcTcor < 90) {
if (refDCval < 88) {
if (RsDiff < 18) {
return 1;
}
else if (RsDiff >= 18) {
return 2;
}
}
else if (refDCval >= 88) {
if (Rs < 3173) {
return 4;
}
else if (Rs >= 3173) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
if (diffMVdiffVal < -10) {
if (Rs < 2712) {
return 8;
}
else if (Rs >= 2712) {
return 2;
}
}
else if (diffMVdiffVal >= -10) {
return 2;
}
}
}
}
else if (ssDCval >= 95) {
if (SC < 3925) {
if (Rs < 2544) {
if (MVDiff < 42) {
return 8;
}
else if (MVDiff >= 42) {
return 1;
}
}
else if (Rs >= 2544) {
if (ssDCval < 107) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 1;
}
}
else if (ssDCval >= 107) {
if (RsDiff < 19) {
return 1;
}
else if (RsDiff >= 19) {
if (Rs < 2754) {
return 8;
}
else if (Rs >= 2754) {
return 2;
}
}
}
}
}
else if (SC >= 3925) {
if (refDCval < 111) {
if (diffRsCsdiff < -48) {
if (diffMVdiffVal < -3) {
if (Cs < 1124) {
return 1;
}
else if (Cs >= 1124) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (ssDCval < 111) {
return 2;
}
else if (ssDCval >= 111) {
if (Rs < 4726) {
return 1;
}
else if (Rs >= 4726) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -48) {
if (AFD < 30) {
return 8;
}
else if (AFD >= 30) {
return 2;
}
}
}
else if (refDCval >= 111) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 497) {
if (refDCval < 87) {
if (RsDiff < 31) {
if (Rs < 3168) {
return 8;
}
else if (Rs >= 3168) {
return 1;
}
}
else if (RsDiff >= 31) {
return 2;
}
}
else if (refDCval >= 87) {
if (Cs < 1141) {
return 2;
}
else if (Cs >= 1141) {
if (Cs < 1151) {
return 8;
}
else if (Cs >= 1151) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (Rs < 3439) {
if (mcTcor < 93) {
if (RsCsDiff < 355) {
if (Cs < 312) {
return 2;
}
else if (Cs >= 312) {
if (diffRsCsdiff < -27) {
if (Cs < 560) {
if (Rs < 3167) {
return 2;
}
else if (Rs >= 3167) {
return 4;
}
}
else if (Cs >= 560) {
return 1;
}
}
else if (diffRsCsdiff >= -27) {
if (SC < 3290) {
if (ssDCval < 121) {
if (SC < 3059) {
return 1;
}
else if (SC >= 3059) {
return 8;
}
}
else if (ssDCval >= 121) {
return 1;
}
}
else if (SC >= 3290) {
return 8;
}
}
}
}
else if (RsCsDiff >= 355) {
if (Cs < 948) {
return 2;
}
else if (Cs >= 948) {
return 1;
}
}
}
else if (mcTcor >= 93) {
if (SC < 3498) {
if (RsDiff < 17) {
if (CsDiff < 3) {
if (Rs < 2567) {
return 1;
}
else if (Rs >= 2567) {
return 8;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (RsDiff >= 17) {
if (diffAFD < -2) {
if (Cs < 506) {
return 2;
}
else if (Cs >= 506) {
return 1;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
}
else if (SC >= 3498) {
if (SC < 3809) {
if (gchDC < 2) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
else if (gchDC >= 2) {
if (ssDCval < 124) {
return 1;
}
else if (ssDCval >= 124) {
if (negBalance < 61) {
return 2;
}
else if (negBalance >= 61) {
return 1;
}
}
}
}
else if (SC >= 3809) {
if (AFD < 23) {
if (Cs < 999) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (Cs >= 999) {
return 8;
}
}
else if (AFD >= 23) {
return 1;
}
}
}
}
}
else if (Rs >= 3439) {
if (Cs < 773) {
if (Rs < 4254) {
if (MVDiff < 31) {
if (gchDC < 7) {
return 2;
}
else if (gchDC >= 7) {
return 1;
}
}
else if (MVDiff >= 31) {
if (Cs < 566) {
if (diffMVdiffVal < 23) {
return 2;
}
else if (diffMVdiffVal >= 23) {
if (Rs < 3635) {
return 1;
}
else if (Rs >= 3635) {
return 4;
}
}
}
else if (Cs >= 566) {
return 8;
}
}
}
else if (Rs >= 4254) {
return 2;
}
}
else if (Cs >= 773) {
if (RsCsDiff < 461) {
if (mu_mv_mag_sq < 8) {
if (negBalance < 44) {
if (Rs < 4631) {
return 2;
}
else if (Rs >= 4631) {
return 1;
}
}
else if (negBalance >= 44) {
if (diffAFD < -5) {
if (Rs < 4891) {
return 1;
}
else if (Rs >= 4891) {
return 2;
}
}
else if (diffAFD >= -5) {
if (diffMVdiffVal < -18) {
if (Rs < 4072) {
return 2;
}
else if (Rs >= 4072) {
return 1;
}
}
else if (diffMVdiffVal >= -18) {
if (Rs < 3564) {
return 1;
}
else if (Rs >= 3564) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 8) {
if (Cs < 1126) {
if (ssDCval < 121) {
if (AFD < 15) {
if (Rs < 5360) {
return 8;
}
else if (Rs >= 5360) {
return 2;
}
}
else if (AFD >= 15) {
return 1;
}
}
else if (ssDCval >= 121) {
return 2;
}
}
else if (Cs >= 1126) {
return 1;
}
}
}
else if (RsCsDiff >= 461) {
if (ssDCval < 122) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (Rs < 4582) {
return 8;
}
else if (Rs >= 4582) {
return 2;
}
}
}
else if (ssDCval >= 122) {
if (TSC < 15) {
if (tcor < 81) {
if (tcor < 76) {
if (refDCval < 126) {
if (MVDiff < 46) {
return 8;
}
else if (MVDiff >= 46) {
if (Rs < 4752) {
return 8;
}
else if (Rs >= 4752) {
return 2;
}
}
}
else if (refDCval >= 126) {
if (diffMVdiffVal < 12) {
return 2;
}
else if (diffMVdiffVal >= 12) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (Rs < 6326) {
return 2;
}
else if (Rs >= 6326) {
return 8;
}
}
}
}
}
else if (tcor >= 76) {
return 8;
}
}
else if (tcor >= 81) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 1;
}
}
}
else if (TSC >= 15) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 133) {
if (mcTcor < 92) {
if (AFD < 19) {
if (MVDiff < 51) {
return 4;
}
else if (MVDiff >= 51) {
if (Rs < 4292) {
if (Rs < 2665) {
return 4;
}
else if (Rs >= 2665) {
return 1;
}
}
else if (Rs >= 4292) {
return 2;
}
}
}
else if (AFD >= 19) {
if (posBalance < 53) {
if (RsDiff < 17) {
return 1;
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (posBalance >= 53) {
if (Rs < 5069) {
if (CsDiff < 5) {
if (Rs < 2858) {
return 1;
}
else if (Rs >= 2858) {
return 8;
}
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (Rs >= 5069) {
return 2;
}
}
}
}
else if (mcTcor >= 92) {
if (CsDiff < 4) {
if (SC < 4131) {
if (mcTcor < 94) {
if (Rs < 3394) {
return 8;
}
else if (Rs >= 3394) {
return 1;
}
}
else if (mcTcor >= 94) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (Cs < 232) {
return 2;
}
else if (Cs >= 232) {
if (SC < 3597) {
return 1;
}
else if (SC >= 3597) {
if (Rs < 3223) {
return 2;
}
else if (Rs >= 3223) {
return 1;
}
}
}
}
}
}
else if (SC >= 4131) {
if (Cs < 742) {
if (RsDiff < 14) {
if (Rs < 4796) {
return 2;
}
else if (Rs >= 4796) {
return 1;
}
}
else if (RsDiff >= 14) {
if (Cs < 721) {
return 2;
}
else if (Cs >= 721) {
return 4;
}
}
}
else if (Cs >= 742) {
if (refDCval < 147) {
return 4;
}
else if (refDCval >= 147) {
return 2;
}
}
}
}
else if (CsDiff >= 4) {
if (RsCsDiff < 320) {
if (AFD < 31) {
if (Cs < 1158) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
if (Cs < 777) {
if (Rs < 3332) {
return 1;
}
else if (Rs >= 3332) {
return 4;
}
}
else if (Cs >= 777) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 8;
}
}
}
}
else if (Cs >= 1158) {
if (Rs < 4029) {
return 1;
}
else if (Rs >= 4029) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 8;
}
}
}
}
else if (AFD >= 31) {
return 2;
}
}
else if (RsCsDiff >= 320) {
if (Rs < 4818) {
if (TSC < 14) {
if (mu_mv_mag_sq < 34) {
return 1;
}
else if (mu_mv_mag_sq >= 34) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
else if (TSC >= 14) {
if (Cs < 1280) {
if (Rs < 3065) {
return 1;
}
else if (Rs >= 3065) {
return 2;
}
}
else if (Cs >= 1280) {
return 1;
}
}
}
else if (Rs >= 4818) {
if (tcor < 84) {
if (Rs < 5983) {
return 8;
}
else if (Rs >= 5983) {
if (Rs < 6075) {
return 2;
}
else if (Rs >= 6075) {
return 8;
}
}
}
else if (tcor >= 84) {
if (mu_mv_mag_sq < 2) {
if (Rs < 5677) {
return 1;
}
else if (Rs >= 5677) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1325) {
if (mu_mv_mag_sq < 107) {
if (RsCsDiff < 101) {
if (ssDCval < 109) {
if (negBalance < 39) {
if (ssDCval < 87) {
if (SC < 4446) {
if (mu_mv_mag_sq < 80) {
if (mcTcor < 96) {
if (tcor < 95) {
if (TSC < 6) {
if (AFD < 8) {
if (Rs < 1370) {
if (MVDiff < 6) {
if (Cs < 1529) {
return 4;
}
else if (Cs >= 1529) {
if (Rs < 1155) {
return 2;
}
else if (Rs >= 1155) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (Cs < 1983) {
return 1;
}
else if (Cs >= 1983) {
return 8;
}
}
}
else if (Rs >= 1370) {
if (negBalance < 32) {
if (Rs < 1673) {
return 8;
}
else if (Rs >= 1673) {
if (Rs < 2334) {
return 2;
}
else if (Rs >= 2334) {
return 1;
}
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
else if (AFD >= 8) {
if (RsCsDiff < 48) {
if (Rs < 1132) {
return 4;
}
else if (Rs >= 1132) {
return 2;
}
}
else if (RsCsDiff >= 48) {
if (Rs < 2515) {
if (Cs < 1366) {
return 2;
}
else if (Cs >= 1366) {
return 1;
}
}
else if (Rs >= 2515) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 58) {
if (Rs < 1832) {
if (Cs < 1419) {
return 8;
}
else if (Cs >= 1419) {
if (MVDiff < 30) {
return 2;
}
else if (MVDiff >= 30) {
return 8;
}
}
}
else if (Rs >= 1832) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 8;
}
}
}
else if (ssDCval >= 58) {
if (posBalance < 58) {
if (ssDCval < 71) {
if (mu_mv_mag_sq < 23) {
if (Rs < 1025) {
return 8;
}
else if (Rs >= 1025) {
if (Rs < 1916) {
return 4;
}
else if (Rs >= 1916) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 23) {
return 4;
}
}
else if (ssDCval >= 71) {
if (Rs < 2847) {
return 8;
}
else if (Rs >= 2847) {
return 4;
}
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
else if (tcor >= 95) {
if (Rs < 1887) {
return 4;
}
else if (Rs >= 1887) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 860) {
if (MVDiff < 107) {
return 1;
}
else if (MVDiff >= 107) {
return 8;
}
}
else if (Rs >= 860) {
if (ssDCval < 67) {
if (Cs < 1666) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (Cs >= 1666) {
if (Rs < 1515) {
return 1;
}
else if (Rs >= 1515) {
return 8;
}
}
}
else if (ssDCval >= 67) {
if (SC < 3579) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
if (Rs < 1643) {
return 4;
}
else if (Rs >= 1643) {
return 1;
}
}
}
else if (SC >= 3579) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (MVDiff < 8) {
if (posBalance < 42) {
if (ssDCval < 66) {
if (tcor < 96) {
return 2;
}
else if (tcor >= 96) {
if (Rs < 1424) {
if (Rs < 434) {
return 4;
}
else if (Rs >= 434) {
return 1;
}
}
else if (Rs >= 1424) {
return 8;
}
}
}
else if (ssDCval >= 66) {
if (Rs < 2171) {
return 8;
}
else if (Rs >= 2171) {
return 1;
}
}
}
else if (posBalance >= 42) {
if (ssDCval < 56) {
if (Cs < 1495) {
if (Rs < 1318) {
return 4;
}
else if (Rs >= 1318) {
return 8;
}
}
else if (Cs >= 1495) {
if (Rs < 885) {
return 1;
}
else if (Rs >= 885) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (diffRsCsdiff < -23) {
if (Cs < 1629) {
return 2;
}
else if (Cs >= 1629) {
return 8;
}
}
else if (diffRsCsdiff >= -23) {
if (Cs < 1571) {
if (Rs < 2298) {
if (Cs < 1565) {
return 8;
}
else if (Cs >= 1565) {
return 4;
}
}
else if (Rs >= 2298) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (Cs >= 1571) {
if (posBalance < 49) {
if (Rs < 2561) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (mu_mv_mag_sq < 89) {
if (Rs < 1813) {
return 1;
}
else if (Rs >= 1813) {
return 8;
}
}
else if (mu_mv_mag_sq >= 89) {
return 8;
}
}
}
}
else if (Rs >= 2561) {
return 1;
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 8) {
if (refDCval < 63) {
if (MVDiff < 9) {
if (Rs < 953) {
return 2;
}
else if (Rs >= 953) {
return 4;
}
}
else if (MVDiff >= 9) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (diffRsCsdiff < 6) {
if (diffRsCsdiff < 1) {
if (Rs < 980) {
return 8;
}
else if (Rs >= 980) {
if (Rs < 1724) {
return 4;
}
else if (Rs >= 1724) {
return 8;
}
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 1192) {
return 4;
}
else if (Rs >= 1192) {
return 1;
}
}
}
else if (diffRsCsdiff >= 6) {
return 8;
}
}
}
}
else if (refDCval >= 63) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (Cs < 1518) {
return 4;
}
else if (Cs >= 1518) {
if (mu_mv_mag_sq < 87) {
if (Cs < 2528) {
return 1;
}
else if (Cs >= 2528) {
return 8;
}
}
else if (mu_mv_mag_sq >= 87) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 4446) {
if (Cs < 1821) {
if (Rs < 7129) {
if (posBalance < 43) {
if (ssDCval < 80) {
if (diffRsCsdiff < 12) {
return 4;
}
else if (diffRsCsdiff >= 12) {
if (Rs < 3652) {
return 2;
}
else if (Rs >= 3652) {
if (Rs < 4226) {
return 4;
}
else if (Rs >= 4226) {
return 8;
}
}
}
}
else if (ssDCval >= 80) {
if (refDCval < 85) {
return 8;
}
else if (refDCval >= 85) {
if (Rs < 5134) {
return 1;
}
else if (Rs >= 5134) {
return 4;
}
}
}
}
else if (posBalance >= 43) {
if (diffRsCsdiff < -34) {
if (Rs < 2975) {
return 8;
}
else if (Rs >= 2975) {
return 2;
}
}
else if (diffRsCsdiff >= -34) {
if (RsDiff < 7) {
if (negBalance < 29) {
if (ssDCval < 81) {
return 2;
}
else if (ssDCval >= 81) {
if (Rs < 3168) {
return 1;
}
else if (Rs >= 3168) {
return 4;
}
}
}
else if (negBalance >= 29) {
return 4;
}
}
else if (RsDiff >= 7) {
if (mcTcor < 98) {
if (ssDCval < 83) {
return 2;
}
else if (ssDCval >= 83) {
if (Rs < 3388) {
return 4;
}
else if (Rs >= 3388) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
return 4;
}
}
}
}
}
}
else if (Rs >= 7129) {
return 2;
}
}
else if (Cs >= 1821) {
if (MVDiff < 5) {
if (Rs < 4791) {
if (RsCsDiff < 80) {
if (tcor < 93) {
if (Rs < 3767) {
if (Cs < 2891) {
return 8;
}
else if (Cs >= 2891) {
return 1;
}
}
else if (Rs >= 3767) {
return 4;
}
}
else if (tcor >= 93) {
if (SC < 6613) {
if (RsCsDiff < 46) {
if (RsCsDiff < 19) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Rs < 2632) {
return 1;
}
else if (Rs >= 2632) {
return 4;
}
}
}
else if (RsCsDiff >= 19) {
if (Cs < 3104) {
return 4;
}
else if (Cs >= 3104) {
return 8;
}
}
}
else if (RsCsDiff >= 46) {
if (Rs < 3405) {
return 1;
}
else if (Rs >= 3405) {
return 4;
}
}
}
else if (SC >= 6613) {
return 8;
}
}
}
else if (RsCsDiff >= 80) {
if (Rs < 4027) {
return 1;
}
else if (Rs >= 4027) {
return 2;
}
}
}
else if (Rs >= 4791) {
if (Cs < 3238) {
if (SC < 6980) {
if (Rs < 4943) {
return 4;
}
else if (Rs >= 4943) {
return 8;
}
}
else if (SC >= 6980) {
return 4;
}
}
else if (Cs >= 3238) {
if (Rs < 5872) {
return 8;
}
else if (Rs >= 5872) {
return 2;
}
}
}
}
else if (MVDiff >= 5) {
if (refDCval < 52) {
if (Rs < 3237) {
if (Rs < 1706) {
return 1;
}
else if (Rs >= 1706) {
return 8;
}
}
else if (Rs >= 3237) {
return 4;
}
}
else if (refDCval >= 52) {
if (Cs < 4148) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (Cs >= 4148) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (MVDiff < 6) {
if (RsCsDiff < 51) {
if (Rs < 3402) {
if (Cs < 2773) {
if (Cs < 1435) {
if (ssDCval < 96) {
if (Rs < 1803) {
return 8;
}
else if (Rs >= 1803) {
return 2;
}
}
else if (ssDCval >= 96) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (Cs >= 1435) {
if (posBalance < 27) {
if (Rs < 2561) {
return 4;
}
else if (Rs >= 2561) {
return 1;
}
}
else if (posBalance >= 27) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Cs < 1620) {
if (Cs < 1521) {
return 4;
}
else if (Cs >= 1521) {
return 1;
}
}
else if (Cs >= 1620) {
if (Cs < 2641) {
return 8;
}
else if (Cs >= 2641) {
return 4;
}
}
}
}
}
}
else if (Cs >= 2773) {
if (Rs < 3087) {
return 4;
}
else if (Rs >= 3087) {
return 8;
}
}
}
else if (Rs >= 3402) {
if (posBalance < 41) {
if (ssDCval < 96) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (Rs < 3888) {
return 2;
}
else if (Rs >= 3888) {
return 8;
}
}
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (posBalance >= 41) {
if (Cs < 1525) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (Cs >= 1525) {
if (negBalance < 34) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (negBalance >= 34) {
if (Cs < 1571) {
return 2;
}
else if (Cs >= 1571) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 51) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (mcTcor < 96) {
if (SC < 5579) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (SC >= 5579) {
if (SC < 5854) {
return 4;
}
else if (SC >= 5854) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 4279) {
return 2;
}
else if (Rs >= 4279) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (SC < 4533) {
if (ssDCval < 108) {
if (CsDiff < 1) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
return 4;
}
}
else if (CsDiff >= 1) {
if (negBalance < 38) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
if (Cs < 1630) {
if (Rs < 2534) {
return 2;
}
else if (Rs >= 2534) {
return 8;
}
}
else if (Cs >= 1630) {
if (Cs < 1963) {
return 8;
}
else if (Cs >= 1963) {
return 4;
}
}
}
}
}
else if (negBalance >= 38) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (Cs < 1539) {
if (Rs < 2927) {
return 4;
}
else if (Rs >= 2927) {
return 8;
}
}
else if (Cs >= 1539) {
return 8;
}
}
}
}
}
else if (ssDCval >= 108) {
return 4;
}
}
else if (SC >= 4533) {
if (Rs < 4146) {
if (AFD < 10) {
if (diffAFD < 2) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 2;
}
}
}
else if (diffAFD >= 2) {
if (CsDiff < 3) {
if (Rs < 3256) {
return 1;
}
else if (Rs >= 3256) {
return 4;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (AFD >= 10) {
if (Cs < 1925) {
return 1;
}
else if (Cs >= 1925) {
return 8;
}
}
}
else if (Rs >= 4146) {
if (TSC < 7) {
if (Rs < 4348) {
return 8;
}
else if (Rs >= 4348) {
return 1;
}
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 1449) {
if (RsCsDiff < 41) {
if (refDCval < 83) {
if (ssDCval < 58) {
if (Rs < 1125) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (SC < 2251) {
return 4;
}
else if (SC >= 2251) {
return 2;
}
}
}
else if (Rs >= 1125) {
if (diffRsCsdiff < 4) {
return 8;
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
}
else if (ssDCval >= 58) {
if (Rs < 999) {
if (mcTcor < 97) {
if (refDCval < 71) {
return 8;
}
else if (refDCval >= 71) {
if (diffAFD < 1) {
if (Rs < 567) {
return 1;
}
else if (Rs >= 567) {
if (Cs < 1629) {
return 8;
}
else if (Cs >= 1629) {
return 1;
}
}
}
else if (diffAFD >= 1) {
if (Cs < 1452) {
return 4;
}
else if (Cs >= 1452) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 706) {
if (Rs < 559) {
return 1;
}
else if (Rs >= 559) {
return 8;
}
}
else if (Rs >= 706) {
return 2;
}
}
}
}
else if (Rs >= 999) {
if (refDCval < 69) {
if (MVDiff < 3) {
if (AFD < 8) {
if (Rs < 1229) {
return 4;
}
else if (Rs >= 1229) {
return 1;
}
}
else if (AFD >= 8) {
return 8;
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 91) {
return 1;
}
else if (mu_mv_mag_sq >= 91) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (refDCval >= 69) {
return 4;
}
}
}
}
else if (refDCval >= 83) {
if (posBalance < 50) {
return 4;
}
else if (posBalance >= 50) {
if (SC < 3100) {
if (mu_mv_mag_sq < 3) {
if (Rs < 976) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
else if (Rs >= 976) {
if (Cs < 1355) {
return 2;
}
else if (Cs >= 1355) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (SC >= 3100) {
if (Rs < 1324) {
return 4;
}
else if (Rs >= 1324) {
if (Rs < 1426) {
return 8;
}
else if (Rs >= 1426) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 41) {
if (refDCval < 54) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (refDCval >= 54) {
if (ssDCval < 69) {
if (mcTcor < 98) {
if (mcTcor < 83) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Rs < 1099) {
return 8;
}
else if (Rs >= 1099) {
return 4;
}
}
}
else if (mcTcor >= 83) {
if (Rs < 1300) {
if (RsCsDiff < 51) {
if (Rs < 1071) {
return 1;
}
else if (Rs >= 1071) {
return 4;
}
}
else if (RsCsDiff >= 51) {
if (tcor < 74) {
if (Rs < 1007) {
return 4;
}
else if (Rs >= 1007) {
if (Rs < 1160) {
return 2;
}
else if (Rs >= 1160) {
return 8;
}
}
}
else if (tcor >= 74) {
if (Rs < 1221) {
if (tcor < 94) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 969) {
return 8;
}
else if (Rs >= 969) {
return 2;
}
}
}
else if (tcor >= 94) {
return 2;
}
}
else if (Rs >= 1221) {
if (Rs < 1268) {
return 4;
}
else if (Rs >= 1268) {
return 8;
}
}
}
}
}
else if (Rs >= 1300) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (ssDCval >= 69) {
if (tcor < 88) {
if (mu_mv_mag_sq < 31) {
if (TSC < 7) {
if (refDCval < 79) {
if (Rs < 908) {
return 8;
}
else if (Rs >= 908) {
return 4;
}
}
else if (refDCval >= 79) {
return 8;
}
}
else if (TSC >= 7) {
if (negBalance < 52) {
if (Cs < 1526) {
if (AFD < 16) {
if (Cs < 1368) {
return 2;
}
else if (Cs >= 1368) {
return 8;
}
}
else if (AFD >= 16) {
return 4;
}
}
else if (Cs >= 1526) {
if (diffRsCsdiff < 24) {
return 4;
}
else if (diffRsCsdiff >= 24) {
if (Cs < 1922) {
return 4;
}
else if (Cs >= 1922) {
return 2;
}
}
}
}
else if (negBalance >= 52) {
if (SC < 2952) {
if (RsCsDiff < 63) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 4;
}
}
else if (RsCsDiff >= 63) {
if (diffAFD < -1) {
if (Cs < 1393) {
return 2;
}
else if (Cs >= 1393) {
if (MVDiff < 35) {
return 4;
}
else if (MVDiff >= 35) {
return 2;
}
}
}
else if (diffAFD >= -1) {
if (MVDiff < 20) {
if (Cs < 1697) {
return 8;
}
else if (Cs >= 1697) {
return 4;
}
}
else if (MVDiff >= 20) {
return 4;
}
}
}
}
else if (SC >= 2952) {
if (Rs < 1133) {
return 2;
}
else if (Rs >= 1133) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 31) {
if (negBalance < 44) {
if (Rs < 895) {
return 2;
}
else if (Rs >= 895) {
return 4;
}
}
else if (negBalance >= 44) {
if (gchDC < 3) {
if (diffMVdiffVal < -2) {
if (Rs < 1422) {
return 8;
}
else if (Rs >= 1422) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (gchDC >= 3) {
return 4;
}
}
}
}
else if (tcor >= 88) {
if (posBalance < 51) {
if (SC < 2868) {
return 2;
}
else if (SC >= 2868) {
return 4;
}
}
else if (posBalance >= 51) {
if (RsDiff < 6) {
if (ssDCval < 80) {
if (SC < 2361) {
return 2;
}
else if (SC >= 2361) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (ssDCval >= 80) {
return 8;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 1449) {
if (SC < 7187) {
if (Cs < 2916) {
if (Rs < 4312) {
if (mcTcor < 95) {
if (Rs < 2271) {
if (Cs < 2289) {
if (AFD < 16) {
if (ssDCval < 84) {
if (TSC < 11) {
if (posBalance < 68) {
if (MVDiff < 3) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (MVDiff < 13) {
if (mcTcor < 92) {
if (CsDiff < 7) {
if (mu_mv_mag_sq < 95) {
return 8;
}
else if (mu_mv_mag_sq >= 95) {
if (tcor < 80) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (tcor >= 80) {
return 8;
}
}
}
else if (CsDiff >= 7) {
if (refDCval < 50) {
return 4;
}
else if (refDCval >= 50) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
else if (mcTcor >= 92) {
if (TSC < 9) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (diffMVdiffVal < 1) {
if (RsCsDiff < 67) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 4;
}
}
else if (RsCsDiff >= 67) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 1747) {
return 8;
}
else if (Rs >= 1747) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 1631) {
return 2;
}
else if (Cs >= 1631) {
if (Cs < 1902) {
return 8;
}
else if (Cs >= 1902) {
return 4;
}
}
}
}
}
else if (TSC >= 9) {
if (ssDCval < 79) {
if (Rs < 2112) {
return 4;
}
else if (Rs >= 2112) {
return 8;
}
}
else if (ssDCval >= 79) {
return 8;
}
}
}
}
else if (MVDiff >= 13) {
if (RsCsDiff < 67) {
return 8;
}
else if (RsCsDiff >= 67) {
if (posBalance < 56) {
if (SC < 3952) {
if (gchDC < 2) {
if (Rs < 1864) {
return 1;
}
else if (Rs >= 1864) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (SC >= 3952) {
return 4;
}
}
else if (posBalance >= 56) {
if (ssDCval < 66) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (Rs < 1667) {
return 4;
}
else if (Rs >= 1667) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 68) {
return 4;
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (ssDCval >= 84) {
if (RsDiff < 8) {
if (Rs < 2139) {
if (TSC < 12) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (AFD < 11) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (AFD >= 11) {
return 8;
}
}
}
}
else if (TSC >= 12) {
return 4;
}
}
else if (Rs >= 2139) {
if (CsDiff < 5) {
if (ssDCval < 90) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (ssDCval < 85) {
return 1;
}
else if (ssDCval >= 85) {
return 8;
}
}
}
else if (ssDCval >= 90) {
if (diffRsCsdiff < -5) {
return 8;
}
else if (diffRsCsdiff >= -5) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
return 8;
}
}
}
}
else if (CsDiff >= 5) {
if (TSC < 9) {
if (mu_mv_mag_sq < 83) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (mu_mv_mag_sq >= 83) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
else if (RsDiff >= 8) {
if (TSC < 9) {
if (Rs < 1937) {
return 8;
}
else if (Rs >= 1937) {
return 2;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
else if (AFD >= 16) {
if (ssDCval < 78) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (mu_mv_mag_sq < 3) {
if (Cs < 1693) {
return 4;
}
else if (Cs >= 1693) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (ssDCval >= 78) {
if (TSC < 13) {
if (Rs < 1665) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
return 4;
}
}
else if (Rs >= 1665) {
if (Cs < 1598) {
return 8;
}
else if (Cs >= 1598) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 2104) {
return 8;
}
else if (Rs >= 2104) {
return 4;
}
}
}
}
}
else if (TSC >= 13) {
if (Rs < 1675) {
return 8;
}
else if (Rs >= 1675) {
return 1;
}
}
}
}
}
else if (Cs >= 2289) {
if (mcTcor < 94) {
if (tcor < 88) {
if (posBalance < 61) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Rs < 1927) {
return 8;
}
else if (Rs >= 1927) {
return 1;
}
}
}
else if (posBalance >= 61) {
return 4;
}
}
else if (tcor >= 88) {
return 4;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
}
else if (Rs >= 2271) {
if (mu_mv_mag_sq < 90) {
if (mcTcor < 94) {
if (RsCsDiff < 77) {
return 8;
}
else if (RsCsDiff >= 77) {
if (posBalance < 41) {
return 2;
}
else if (posBalance >= 41) {
if (Cs < 2672) {
if (SC < 4512) {
if (SC < 4263) {
return 8;
}
else if (SC >= 4263) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (diffAFD < 2) {
if (SC < 4269) {
return 4;
}
else if (SC >= 4269) {
return 8;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
else if (SC >= 4512) {
return 8;
}
}
else if (Cs >= 2672) {
if (Rs < 3014) {
return 8;
}
else if (Rs >= 3014) {
return 4;
}
}
}
}
}
else if (mcTcor >= 94) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (negBalance < 57) {
if (SC < 4072) {
if (Cs < 1491) {
if (Rs < 2375) {
return 1;
}
else if (Rs >= 2375) {
return 4;
}
}
else if (Cs >= 1491) {
if (SC < 4065) {
return 8;
}
else if (SC >= 4065) {
return 1;
}
}
}
else if (SC >= 4072) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (MVDiff < 10) {
if (Rs < 2961) {
return 8;
}
else if (Rs >= 2961) {
return 4;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
}
else if (negBalance >= 57) {
if (Rs < 2994) {
return 8;
}
else if (Rs >= 2994) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (Rs < 3269) {
if (diffTSC < 1) {
if (diffTSC < -2) {
if (Rs < 2482) {
return 4;
}
else if (Rs >= 2482) {
return 8;
}
}
else if (diffTSC >= -2) {
if (mu_mv_mag_sq < 106) {
if (Cs < 2739) {
return 8;
}
else if (Cs >= 2739) {
return 4;
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 2884) {
return 8;
}
else if (Rs >= 2884) {
if (Rs < 3054) {
return 4;
}
else if (Rs >= 3054) {
return 8;
}
}
}
}
}
else if (diffTSC >= 1) {
if (Cs < 2444) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 4;
}
}
}
else if (Cs >= 2444) {
return 4;
}
}
}
else if (Rs >= 3269) {
if (mcTcor < 93) {
if (Rs < 3764) {
return 8;
}
else if (Rs >= 3764) {
return 1;
}
}
else if (mcTcor >= 93) {
if (Rs < 3884) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (Rs >= 3884) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (RsDiff < 8) {
if (Rs < 2581) {
if (ssDCval < 89) {
if (refDCval < 80) {
if (AFD < 8) {
if (Cs < 1526) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 1732) {
if (Rs < 1517) {
return 4;
}
else if (Rs >= 1517) {
return 8;
}
}
else if (Rs >= 1732) {
return 1;
}
}
}
else if (Cs >= 1526) {
return 8;
}
}
else if (AFD >= 8) {
if (tcor < 91) {
if (ssDCval < 74) {
if (Rs < 2007) {
return 8;
}
else if (Rs >= 2007) {
return 4;
}
}
else if (ssDCval >= 74) {
return 4;
}
}
else if (tcor >= 91) {
if (RsCsDiff < 47) {
if (Cs < 1549) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (Cs >= 1549) {
return 4;
}
}
else if (RsCsDiff >= 47) {
return 8;
}
}
}
}
else if (refDCval >= 80) {
if (MVDiff < 4) {
if (AFD < 8) {
if (Cs < 1535) {
if (ssDCval < 85) {
if (Rs < 2087) {
return 1;
}
else if (Rs >= 2087) {
return 2;
}
}
else if (ssDCval >= 85) {
return 4;
}
}
else if (Cs >= 1535) {
if (TSC < 7) {
if (Cs < 2565) {
return 8;
}
else if (Cs >= 2565) {
return 4;
}
}
else if (TSC >= 7) {
if (Rs < 2515) {
return 1;
}
else if (Rs >= 2515) {
return 8;
}
}
}
}
else if (AFD >= 8) {
if (Rs < 1631) {
return 8;
}
else if (Rs >= 1631) {
if (Rs < 2151) {
return 1;
}
else if (Rs >= 2151) {
if (refDCval < 83) {
return 8;
}
else if (refDCval >= 83) {
if (Cs < 1677) {
return 1;
}
else if (Cs >= 1677) {
if (Rs < 2466) {
return 8;
}
else if (Rs >= 2466) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (Cs < 1357) {
if (Rs < 2013) {
return 1;
}
else if (Rs >= 2013) {
return 8;
}
}
else if (Cs >= 1357) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (MVDiff < 5) {
if (Cs < 1513) {
return 1;
}
else if (Cs >= 1513) {
return 8;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 89) {
if (Rs < 1918) {
if (ssDCval < 96) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 1817) {
return 2;
}
else if (Rs >= 1817) {
return 1;
}
}
}
else if (ssDCval >= 96) {
if (refDCval < 110) {
return 2;
}
else if (refDCval >= 110) {
return 8;
}
}
}
else if (Rs >= 1918) {
if (Cs < 1892) {
if (SC < 3582) {
if (diffAFD < 1) {
if (SC < 3565) {
return 8;
}
else if (SC >= 3565) {
return 1;
}
}
else if (diffAFD >= 1) {
if (Cs < 1405) {
if (Rs < 2009) {
return 8;
}
else if (Rs >= 2009) {
return 2;
}
}
else if (Cs >= 1405) {
return 1;
}
}
}
else if (SC >= 3582) {
return 8;
}
}
else if (Cs >= 1892) {
if (mu_mv_mag_sq < 81) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 2;
}
}
else if (mu_mv_mag_sq >= 81) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Cs < 2024) {
if (SC < 4067) {
return 1;
}
else if (SC >= 4067) {
return 2;
}
}
else if (Cs >= 2024) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 2581) {
if (mu_mv_mag_sq < 85) {
if (posBalance < 47) {
if (diffRsCsdiff < -30) {
return 1;
}
else if (diffRsCsdiff >= -30) {
if (Cs < 1899) {
if (Cs < 1439) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (Cs >= 1439) {
return 4;
}
}
else if (Cs >= 1899) {
return 8;
}
}
}
else if (posBalance >= 47) {
if (Rs < 3121) {
if (ssDCval < 90) {
if (Cs < 1733) {
if (SC < 4274) {
if (Rs < 2679) {
return 8;
}
else if (Rs >= 2679) {
return 1;
}
}
else if (SC >= 4274) {
return 4;
}
}
else if (Cs >= 1733) {
if (Rs < 2962) {
return 8;
}
else if (Rs >= 2962) {
return 1;
}
}
}
else if (ssDCval >= 90) {
if (negBalance < 53) {
if (mu_mv_mag_sq < 47) {
return 4;
}
else if (mu_mv_mag_sq >= 47) {
return 8;
}
}
else if (negBalance >= 53) {
return 4;
}
}
}
else if (Rs >= 3121) {
if (ssDCval < 90) {
if (Rs < 3668) {
return 2;
}
else if (Rs >= 3668) {
return 1;
}
}
else if (ssDCval >= 90) {
if (Cs < 2408) {
return 8;
}
else if (Cs >= 2408) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (Cs < 1590) {
if (Rs < 2758) {
return 8;
}
else if (Rs >= 2758) {
return 2;
}
}
else if (Cs >= 1590) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (diffRsCsdiff < 31) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (diffRsCsdiff >= 31) {
if (Rs < 3183) {
return 1;
}
else if (Rs >= 3183) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (ssDCval < 89) {
if (Rs < 2451) {
if (Rs < 2328) {
return 4;
}
else if (Rs >= 2328) {
return 8;
}
}
else if (Rs >= 2451) {
if (Cs < 2108) {
return 2;
}
else if (Cs >= 2108) {
return 4;
}
}
}
else if (ssDCval >= 89) {
if (Cs < 2413) {
return 8;
}
else if (Cs >= 2413) {
return 4;
}
}
}
}
}
else if (Rs >= 4312) {
if (mu_mv_mag_sq < 81) {
if (ssDCval < 98) {
if (Cs < 1688) {
return 1;
}
else if (Cs >= 1688) {
return 4;
}
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (mu_mv_mag_sq >= 81) {
return 1;
}
}
}
else if (Cs >= 2916) {
if (Rs < 2961) {
if (ssDCval < 100) {
return 4;
}
else if (ssDCval >= 100) {
return 8;
}
}
else if (Rs >= 2961) {
if (ssDCval < 86) {
if (Rs < 3505) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (Rs >= 3505) {
return 8;
}
}
else if (ssDCval >= 86) {
if (Rs < 3121) {
return 8;
}
else if (Rs >= 3121) {
return 4;
}
}
}
}
}
else if (SC >= 7187) {
return 8;
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 4424) {
if (refDCval < 163) {
if (Rs < 1437) {
if (mu_mv_mag_sq < 38) {
if (ssDCval < 142) {
if (AFD < 18) {
if (SC < 3010) {
return 8;
}
else if (SC >= 3010) {
if (ssDCval < 121) {
if (Rs < 1121) {
return 8;
}
else if (Rs >= 1121) {
return 4;
}
}
else if (ssDCval >= 121) {
return 8;
}
}
}
else if (AFD >= 18) {
if (Rs < 1119) {
if (MVDiff < 36) {
return 2;
}
else if (MVDiff >= 36) {
if (Rs < 867) {
return 8;
}
else if (Rs >= 867) {
return 1;
}
}
}
else if (Rs >= 1119) {
return 4;
}
}
}
else if (ssDCval >= 142) {
if (MVDiff < 26) {
return 2;
}
else if (MVDiff >= 26) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 1213) {
if (CsDiff < 6) {
if (mcTcor < 95) {
if (SC < 2215) {
return 8;
}
else if (SC >= 2215) {
if (ssDCval < 116) {
return 4;
}
else if (ssDCval >= 116) {
if (refDCval < 130) {
return 8;
}
else if (refDCval >= 130) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
return 4;
}
}
}
}
}
else if (mcTcor >= 95) {
if (Rs < 1062) {
if (Rs < 693) {
return 1;
}
else if (Rs >= 693) {
return 8;
}
}
else if (Rs >= 1062) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
return 4;
}
}
}
}
else if (CsDiff >= 6) {
if (Rs < 1203) {
return 8;
}
else if (Rs >= 1203) {
return 1;
}
}
}
else if (Rs >= 1213) {
if (SC < 3128) {
return 8;
}
else if (SC >= 3128) {
if (Rs < 1384) {
return 8;
}
else if (Rs >= 1384) {
return 4;
}
}
}
}
}
else if (Rs >= 1437) {
if (mu_mv_mag_sq < 13) {
if (Rs < 2068) {
if (mcTcor < 90) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (Cs < 1687) {
return 8;
}
else if (Cs >= 1687) {
return 2;
}
}
}
}
else if (mcTcor >= 90) {
if (posBalance < 63) {
if (MVDiff < 7) {
if (Rs < 1474) {
return 1;
}
else if (Rs >= 1474) {
return 8;
}
}
else if (MVDiff >= 7) {
if (Cs < 2076) {
if (TSC < 8) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (TSC >= 8) {
if (Rs < 1548) {
return 4;
}
else if (Rs >= 1548) {
if (Rs < 1977) {
return 8;
}
else if (Rs >= 1977) {
return 2;
}
}
}
}
else if (Cs >= 2076) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 1;
}
}
}
}
else if (posBalance >= 63) {
if (Rs < 1919) {
return 4;
}
else if (Rs >= 1919) {
return 8;
}
}
}
}
else if (Rs >= 2068) {
if (ssDCval < 110) {
if (Cs < 1734) {
if (Rs < 2506) {
return 1;
}
else if (Rs >= 2506) {
return 4;
}
}
else if (Cs >= 1734) {
return 8;
}
}
else if (ssDCval >= 110) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (ssDCval < 142) {
if (ssDCval < 123) {
if (TSC < 11) {
if (mu_mv_mag_sq < 54) {
if (Rs < 4176) {
if (ssDCval < 110) {
if (Rs < 2745) {
return 8;
}
else if (Rs >= 2745) {
if (Rs < 3347) {
return 4;
}
else if (Rs >= 3347) {
return 8;
}
}
}
else if (ssDCval >= 110) {
if (Rs < 1559) {
if (Rs < 1548) {
return 8;
}
else if (Rs >= 1548) {
return 4;
}
}
else if (Rs >= 1559) {
return 8;
}
}
}
else if (Rs >= 4176) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (diffRsCsdiff < -3) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (RsDiff < 5) {
if (Cs < 1394) {
return 2;
}
else if (Cs >= 1394) {
return 8;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
}
else if (TSC >= 11) {
if (Cs < 1633) {
return 1;
}
else if (Cs >= 1633) {
return 8;
}
}
}
else if (ssDCval >= 123) {
if (mcTcor < 92) {
if (Rs < 3624) {
if (Cs < 1841) {
if (mcTcor < 89) {
if (RsCsDiff < 89) {
return 1;
}
else if (RsCsDiff >= 89) {
return 8;
}
}
else if (mcTcor >= 89) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 2118) {
return 1;
}
else if (Rs >= 2118) {
return 8;
}
}
}
}
else if (Cs >= 1841) {
return 8;
}
}
else if (Rs >= 3624) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 139) {
if (Rs < 3168) {
if (refDCval < 127) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (mcTcor < 96) {
if (mcTcor < 94) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (mcTcor >= 96) {
if (Rs < 1601) {
return 2;
}
else if (Rs >= 1601) {
return 8;
}
}
}
}
else if (refDCval >= 127) {
if (negBalance < 35) {
if (Cs < 1392) {
if (Cs < 1356) {
return 8;
}
else if (Cs >= 1356) {
return 1;
}
}
else if (Cs >= 1392) {
return 8;
}
}
else if (negBalance >= 35) {
if (ssDCval < 127) {
if (Rs < 2447) {
return 8;
}
else if (Rs >= 2447) {
return 1;
}
}
else if (ssDCval >= 127) {
return 8;
}
}
}
}
else if (Rs >= 3168) {
if (ssDCval < 138) {
return 8;
}
else if (ssDCval >= 138) {
return 1;
}
}
}
else if (ssDCval >= 139) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (diffRsCsdiff < -1) {
if (ssDCval < 141) {
return 4;
}
else if (ssDCval >= 141) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 142) {
if (SC < 5843) {
if (posBalance < 29) {
return 2;
}
else if (posBalance >= 29) {
if (negBalance < 22) {
return 2;
}
else if (negBalance >= 22) {
return 8;
}
}
}
else if (SC >= 5843) {
return 1;
}
}
}
}
}
else if (refDCval >= 163) {
if (Rs < 2283) {
if (Rs < 1062) {
return 1;
}
else if (Rs >= 1062) {
return 2;
}
}
else if (Rs >= 2283) {
return 8;
}
}
}
else if (Rs >= 4424) {
if (Cs < 1845) {
if (tcor < 92) {
if (Cs < 1771) {
return 1;
}
else if (Cs >= 1771) {
return 8;
}
}
else if (tcor >= 92) {
return 2;
}
}
else if (Cs >= 1845) {
if (RsCsDiff < 94) {
if (Rs < 5961) {
return 8;
}
else if (Rs >= 5961) {
if (Rs < 7227) {
return 4;
}
else if (Rs >= 7227) {
return 8;
}
}
}
else if (RsCsDiff >= 94) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1919) {
if (Rs < 3338) {
if (refDCval < 86) {
if (posBalance < 50) {
if (negBalance < 54) {
if (mcTcor < 85) {
if (Rs < 1998) {
if (Cs < 1466) {
return 8;
}
else if (Cs >= 1466) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (Rs >= 1998) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (Rs < 2114) {
return 1;
}
else if (Rs >= 2114) {
return 8;
}
}
}
}
else if (mcTcor >= 85) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (Rs < 3193) {
if (ssDCval < 83) {
if (Cs < 1907) {
if (Rs < 1796) {
if (Rs < 1346) {
return 2;
}
else if (Rs >= 1346) {
return 1;
}
}
else if (Rs >= 1796) {
return 2;
}
}
else if (Cs >= 1907) {
return 1;
}
}
else if (ssDCval >= 83) {
if (Rs < 2539) {
return 8;
}
else if (Rs >= 2539) {
return 2;
}
}
}
else if (Rs >= 3193) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 2;
}
}
}
}
}
else if (negBalance >= 54) {
if (mu_mv_mag_sq < 13) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 3261) {
return 8;
}
else if (Rs >= 3261) {
return 4;
}
}
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 9) {
if (refDCval < 65) {
if (Rs < 1795) {
if (negBalance < 49) {
if (RsDiff < 7) {
if (Rs < 710) {
return 4;
}
else if (Rs >= 710) {
return 8;
}
}
else if (RsDiff >= 7) {
if (Cs < 1679) {
return 1;
}
else if (Cs >= 1679) {
return 4;
}
}
}
else if (negBalance >= 49) {
if (tcor < 64) {
return 4;
}
else if (tcor >= 64) {
if (Cs < 1783) {
if (Rs < 1482) {
return 2;
}
else if (Rs >= 1482) {
return 1;
}
}
else if (Cs >= 1783) {
return 4;
}
}
}
}
else if (Rs >= 1795) {
if (AFD < 19) {
if (negBalance < 47) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 8;
}
}
else if (negBalance >= 47) {
if (mcTcor < 88) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
return 4;
}
}
else if (mcTcor >= 88) {
if (Rs < 2077) {
return 2;
}
else if (Rs >= 2077) {
return 4;
}
}
}
}
else if (AFD >= 19) {
if (Rs < 2230) {
return 1;
}
else if (Rs >= 2230) {
return 8;
}
}
}
}
else if (refDCval >= 65) {
if (RsCsDiff < 201) {
if (Cs < 1491) {
if (tcor < 79) {
if (Rs < 2091) {
if (tcor < 54) {
if (Rs < 1021) {
if (Rs < 648) {
return 4;
}
else if (Rs >= 648) {
return 8;
}
}
else if (Rs >= 1021) {
return 2;
}
}
else if (tcor >= 54) {
if (MVDiff < 23) {
if (diffRsCsdiff < 8) {
if (Rs < 1990) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (Rs >= 1990) {
return 2;
}
}
else if (diffRsCsdiff >= 8) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
return 2;
}
}
}
else if (MVDiff >= 23) {
if (diffMVdiffVal < 17) {
return 4;
}
else if (diffMVdiffVal >= 17) {
return 1;
}
}
}
}
else if (Rs >= 2091) {
if (AFD < 23) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (AFD >= 23) {
return 4;
}
}
}
else if (tcor >= 79) {
if (diffMVdiffVal < -10) {
return 1;
}
else if (diffMVdiffVal >= -10) {
if (TSC < 12) {
if (refDCval < 72) {
if (Rs < 2658) {
return 8;
}
else if (Rs >= 2658) {
return 2;
}
}
else if (refDCval >= 72) {
if (diffAFD < -2) {
if (Rs < 2346) {
return 8;
}
else if (Rs >= 2346) {
return 2;
}
}
else if (diffAFD >= -2) {
return 2;
}
}
}
else if (TSC >= 12) {
return 8;
}
}
}
}
else if (Cs >= 1491) {
if (negBalance < 58) {
if (RsCsDiff < 129) {
if (TSC < 11) {
if (diffRsCsdiff < 23) {
if (Cs < 1784) {
if (Cs < 1696) {
if (Cs < 1612) {
return 4;
}
else if (Cs >= 1612) {
return 8;
}
}
else if (Cs >= 1696) {
return 4;
}
}
else if (Cs >= 1784) {
return 8;
}
}
else if (diffRsCsdiff >= 23) {
return 8;
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (RsCsDiff >= 129) {
if (Rs < 1924) {
if (diffRsCsdiff < 48) {
if (Rs < 1571) {
return 4;
}
else if (Rs >= 1571) {
if (negBalance < 53) {
if (Rs < 1763) {
return 8;
}
else if (Rs >= 1763) {
return 4;
}
}
else if (negBalance >= 53) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 48) {
return 8;
}
}
else if (Rs >= 1924) {
if (RsCsDiff < 163) {
if (TSC < 11) {
if (diffAFD < 2) {
if (Rs < 2119) {
return 8;
}
else if (Rs >= 2119) {
return 4;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (RsCsDiff >= 163) {
return 4;
}
}
}
}
else if (negBalance >= 58) {
if (tcor < 74) {
if (Cs < 1855) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (Cs >= 1855) {
if (Rs < 1602) {
return 2;
}
else if (Rs >= 1602) {
return 1;
}
}
}
else if (tcor >= 74) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 201) {
if (ssDCval < 77) {
if (Rs < 2567) {
if (mcTcor < 87) {
if (negBalance < 55) {
return 4;
}
else if (negBalance >= 55) {
if (Rs < 1346) {
return 4;
}
else if (Rs >= 1346) {
return 2;
}
}
}
else if (mcTcor >= 87) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (Rs >= 2567) {
return 2;
}
}
else if (ssDCval >= 77) {
if (refDCval < 81) {
if (Cs < 1529) {
return 2;
}
else if (Cs >= 1529) {
return 8;
}
}
else if (refDCval >= 81) {
if (SC < 3394) {
return 2;
}
else if (SC >= 3394) {
if (Cs < 1374) {
if (Rs < 2751) {
return 8;
}
else if (Rs >= 2751) {
return 2;
}
}
else if (Cs >= 1374) {
if (diffMVdiffVal < 5) {
return 4;
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (Rs < 2854) {
if (Rs < 1757) {
if (diffMVdiffVal < -19) {
if (Rs < 1058) {
return 2;
}
else if (Rs >= 1058) {
return 1;
}
}
else if (diffMVdiffVal >= -19) {
if (posBalance < 63) {
if (ssDCval < 47) {
if (Rs < 1000) {
return 2;
}
else if (Rs >= 1000) {
return 8;
}
}
else if (ssDCval >= 47) {
if (diffRsCsdiff < 18) {
if (AFD < 20) {
if (diffAFD < -3) {
if (Rs < 1452) {
return 8;
}
else if (Rs >= 1452) {
return 1;
}
}
else if (diffAFD >= -3) {
return 4;
}
}
else if (AFD >= 20) {
if (ssDCval < 74) {
if (MVDiff < 74) {
return 8;
}
else if (MVDiff >= 74) {
return 4;
}
}
else if (ssDCval >= 74) {
return 4;
}
}
}
else if (diffRsCsdiff >= 18) {
if (diffMVdiffVal < -2) {
if (Rs < 1612) {
return 4;
}
else if (Rs >= 1612) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
if (negBalance < 53) {
if (RsCsDiff < 115) {
return 1;
}
else if (RsCsDiff >= 115) {
return 8;
}
}
else if (negBalance >= 53) {
if (Rs < 1416) {
return 8;
}
else if (Rs >= 1416) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 63) {
if (SC < 3205) {
if (diffTSC < 1) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (SC >= 3205) {
if (Cs < 1679) {
return 1;
}
else if (Cs >= 1679) {
return 4;
}
}
}
}
}
else if (Rs >= 1757) {
if (refDCval < 83) {
if (AFD < 27) {
if (AFD < 14) {
if (diffMVdiffVal < -2) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
if (TSC < 10) {
if (Cs < 1796) {
return 8;
}
else if (Cs >= 1796) {
if (Rs < 1986) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
else if (Rs >= 1986) {
return 8;
}
}
}
else if (TSC >= 10) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 4;
}
}
}
}
else if (AFD >= 14) {
if (mu_mv_mag_sq < 47) {
if (RsCsDiff < 215) {
if (posBalance < 70) {
return 8;
}
else if (posBalance >= 70) {
if (Rs < 2005) {
return 1;
}
else if (Rs >= 2005) {
return 8;
}
}
}
else if (RsCsDiff >= 215) {
if (MVDiff < 25) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 4;
}
}
else if (MVDiff >= 25) {
if (Rs < 2487) {
return 8;
}
else if (Rs >= 2487) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (gchDC < 2) {
if (AFD < 16) {
if (Rs < 2297) {
return 8;
}
else if (Rs >= 2297) {
return 1;
}
}
else if (AFD >= 16) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 1;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
else if (AFD >= 27) {
return 4;
}
}
else if (refDCval >= 83) {
if (mcTcor < 94) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (refDCval < 85) {
return 8;
}
else if (refDCval >= 85) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 2179) {
return 2;
}
else if (Rs >= 2179) {
return 4;
}
}
}
}
}
else if (Rs >= 2854) {
if (Cs < 1515) {
return 2;
}
else if (Cs >= 1515) {
if (AFD < 17) {
if (Rs < 3302) {
return 8;
}
else if (Rs >= 3302) {
return 2;
}
}
else if (AFD >= 17) {
if (CsDiff < 9) {
if (Rs < 2964) {
return 8;
}
else if (Rs >= 2964) {
return 4;
}
}
else if (CsDiff >= 9) {
if (Rs < 3123) {
return 8;
}
else if (Rs >= 3123) {
return 2;
}
}
}
}
}
}
}
}
else if (refDCval >= 86) {
if (AFD < 18) {
if (refDCval < 98) {
if (Rs < 2573) {
if (Cs < 1561) {
if (TSC < 9) {
if (posBalance < 46) {
if (Rs < 2023) {
return 8;
}
else if (Rs >= 2023) {
return 1;
}
}
else if (posBalance >= 46) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
return 1;
}
}
}
else if (TSC >= 9) {
if (tcor < 76) {
if (Rs < 2082) {
if (Rs < 1681) {
return 1;
}
else if (Rs >= 1681) {
return 8;
}
}
else if (Rs >= 2082) {
return 2;
}
}
else if (tcor >= 76) {
if (SC < 3285) {
if (SC < 3193) {
return 8;
}
else if (SC >= 3193) {
return 1;
}
}
else if (SC >= 3285) {
return 8;
}
}
}
}
else if (Cs >= 1561) {
if (mcTcor < 96) {
if (MVDiff < 50) {
if (Rs < 1407) {
return 2;
}
else if (Rs >= 1407) {
return 8;
}
}
else if (MVDiff >= 50) {
return 1;
}
}
else if (mcTcor >= 96) {
if (Rs < 2238) {
return 2;
}
else if (Rs >= 2238) {
if (Rs < 2496) {
return 8;
}
else if (Rs >= 2496) {
return 1;
}
}
}
}
}
else if (Rs >= 2573) {
if (negBalance < 51) {
if (SC < 4391) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
if (Rs < 2603) {
return 4;
}
else if (Rs >= 2603) {
return 2;
}
}
}
else if (SC >= 4391) {
if (gchDC < 2) {
if (refDCval < 97) {
return 2;
}
else if (refDCval >= 97) {
if (Rs < 2779) {
return 8;
}
else if (Rs >= 2779) {
return 2;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
if (Rs < 2989) {
return 2;
}
else if (Rs >= 2989) {
return 4;
}
}
}
}
}
else if (refDCval >= 98) {
if (negBalance < 26) {
if (ssDCval < 138) {
if (posBalance < 17) {
return 1;
}
else if (posBalance >= 17) {
if (Rs < 1652) {
return 2;
}
else if (Rs >= 1652) {
return 8;
}
}
}
else if (ssDCval >= 138) {
return 2;
}
}
else if (negBalance >= 26) {
if (RsCsDiff < 183) {
if (MVDiff < 17) {
if (Cs < 1604) {
if (mcTcor < 86) {
return 4;
}
else if (mcTcor >= 86) {
if (negBalance < 45) {
if (mcTcor < 96) {
if (Cs < 1489) {
return 8;
}
else if (Cs >= 1489) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (mcTcor >= 96) {
return 4;
}
}
else if (negBalance >= 45) {
if (TSC < 11) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
return 2;
}
}
else if (TSC >= 11) {
if (Rs < 2859) {
return 8;
}
else if (Rs >= 2859) {
return 4;
}
}
}
}
}
else if (Cs >= 1604) {
if (RsDiff < 11) {
if (posBalance < 54) {
if (AFD < 16) {
if (TSC < 8) {
if (Rs < 2024) {
return 8;
}
else if (Rs >= 2024) {
return 2;
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (AFD >= 16) {
if (Rs < 1977) {
return 4;
}
else if (Rs >= 1977) {
return 2;
}
}
}
else if (posBalance >= 54) {
return 8;
}
}
else if (RsDiff >= 11) {
if (Cs < 1785) {
return 4;
}
else if (Cs >= 1785) {
return 8;
}
}
}
}
else if (MVDiff >= 17) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 183) {
if (TSC < 9) {
if (negBalance < 51) {
if (SC < 4729) {
return 4;
}
else if (SC >= 4729) {
return 2;
}
}
else if (negBalance >= 51) {
if (Rs < 3059) {
return 8;
}
else if (Rs >= 3059) {
return 4;
}
}
}
else if (TSC >= 9) {
if (Cs < 1470) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
else if (Cs >= 1470) {
if (CsDiff < 7) {
if (Rs < 2994) {
return 8;
}
else if (Rs >= 2994) {
return 4;
}
}
else if (CsDiff >= 7) {
if (tcor < 62) {
return 4;
}
else if (tcor >= 62) {
return 8;
}
}
}
}
}
}
}
}
else if (AFD >= 18) {
if (TSC < 11) {
if (mu_mv_mag_sq < 17) {
if (refDCval < 100) {
if (mcTcor < 94) {
if (Rs < 2971) {
if (Rs < 1722) {
if (Rs < 1412) {
return 4;
}
else if (Rs >= 1412) {
if (Rs < 1615) {
return 2;
}
else if (Rs >= 1615) {
return 1;
}
}
}
else if (Rs >= 1722) {
if (SC < 4479) {
return 8;
}
else if (SC >= 4479) {
return 2;
}
}
}
else if (Rs >= 2971) {
return 4;
}
}
else if (mcTcor >= 94) {
if (diffRsCsdiff < -68) {
return 4;
}
else if (diffRsCsdiff >= -68) {
return 2;
}
}
}
else if (refDCval >= 100) {
if (RsCsDiff < 147) {
if (Rs < 848) {
return 2;
}
else if (Rs >= 848) {
return 4;
}
}
else if (RsCsDiff >= 147) {
if (AFD < 19) {
if (Rs < 3082) {
return 8;
}
else if (Rs >= 3082) {
return 4;
}
}
else if (AFD >= 19) {
if (negBalance < 61) {
if (MVDiff < 11) {
if (Rs < 2860) {
return 2;
}
else if (Rs >= 2860) {
if (Rs < 3054) {
return 1;
}
else if (Rs >= 3054) {
return 4;
}
}
}
else if (MVDiff >= 11) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 1;
}
}
}
else if (negBalance >= 61) {
if (Rs < 1562) {
return 8;
}
else if (Rs >= 1562) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (MVDiff < 28) {
if (tcor < 82) {
if (mu_mv_mag_sq < 26) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (Rs < 1015) {
return 1;
}
else if (Rs >= 1015) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 26) {
return 2;
}
}
else if (tcor >= 82) {
if (Rs < 2051) {
return 1;
}
else if (Rs >= 2051) {
return 8;
}
}
}
else if (MVDiff >= 28) {
if (posBalance < 52) {
if (Rs < 3057) {
return 2;
}
else if (Rs >= 3057) {
return 4;
}
}
else if (posBalance >= 52) {
if (mu_mv_mag_sq < 52) {
if (gchDC < 1) {
if (Rs < 2013) {
return 8;
}
else if (Rs >= 2013) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 52) {
return 2;
}
}
}
}
}
else if (TSC >= 11) {
if (AFD < 23) {
if (ssDCval < 124) {
if (mcTcor < 77) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
if (Rs < 1606) {
return 8;
}
else if (Rs >= 1606) {
return 4;
}
}
}
else if (mcTcor >= 77) {
if (diffRsCsdiff < -217) {
if (Cs < 1614) {
return 1;
}
else if (Cs >= 1614) {
return 8;
}
}
else if (diffRsCsdiff >= -217) {
if (Rs < 1704) {
if (RsCsDiff < 129) {
return 8;
}
else if (RsCsDiff >= 129) {
return 4;
}
}
else if (Rs >= 1704) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
if (mu_mv_mag_sq < 18) {
if (mcTcor < 90) {
if (negBalance < 56) {
return 4;
}
else if (negBalance >= 56) {
return 8;
}
}
else if (mcTcor >= 90) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (Cs < 1351) {
return 4;
}
else if (Cs >= 1351) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 124) {
if (mu_mv_mag_sq < 20) {
if (refDCval < 151) {
if (mcTcor < 89) {
if (Rs < 1805) {
return 8;
}
else if (Rs >= 1805) {
return 1;
}
}
else if (mcTcor >= 89) {
if (Rs < 2218) {
return 1;
}
else if (Rs >= 2218) {
return 8;
}
}
}
else if (refDCval >= 151) {
if (Rs < 1806) {
return 2;
}
else if (Rs >= 1806) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 20) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
}
else if (AFD >= 23) {
if (Rs < 1336) {
if (CsDiff < 13) {
return 2;
}
else if (CsDiff >= 13) {
if (TSC < 13) {
if (Rs < 1066) {
return 4;
}
else if (Rs >= 1066) {
return 2;
}
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (Rs >= 1336) {
if (ssDCval < 92) {
if (diffTSC < 1) {
if (mu_mv_mag_sq < 7) {
return 8;
}
else if (mu_mv_mag_sq >= 7) {
return 1;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (ssDCval >= 92) {
if (CsDiff < 8) {
if (RsDiff < 15) {
if (Rs < 2227) {
return 8;
}
else if (Rs >= 2227) {
if (mu_mv_mag_sq < 7) {
return 8;
}
else if (mu_mv_mag_sq >= 7) {
return 4;
}
}
}
else if (RsDiff >= 15) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
return 2;
}
}
}
else if (CsDiff >= 8) {
if (ssDCval < 115) {
if (RsCsDiff < 339) {
if (mcTcor < 94) {
if (diffRsCsdiff < -27) {
if (Cs < 1699) {
return 4;
}
else if (Cs >= 1699) {
return 8;
}
}
else if (diffRsCsdiff >= -27) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 1395) {
return 8;
}
else if (Rs >= 1395) {
return 2;
}
}
}
else if (RsCsDiff >= 339) {
if (posBalance < 59) {
if (Rs < 2390) {
return 2;
}
else if (Rs >= 2390) {
if (Cs < 1629) {
return 1;
}
else if (Cs >= 1629) {
return 8;
}
}
}
else if (posBalance >= 59) {
return 2;
}
}
}
else if (ssDCval >= 115) {
if (Cs < 1469) {
return 1;
}
else if (Cs >= 1469) {
if (diffMVdiffVal < -164) {
return 1;
}
else if (diffMVdiffVal >= -164) {
if (mu_mv_mag_sq < 25) {
return 8;
}
else if (mu_mv_mag_sq >= 25) {
if (Rs < 2734) {
return 1;
}
else if (Rs >= 2734) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 3338) {
if (tcor < 86) {
if (RsCsDiff < 175) {
if (posBalance < 50) {
if (TSC < 8) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (TSC >= 8) {
if (Rs < 4325) {
if (Rs < 3456) {
return 2;
}
else if (Rs >= 3456) {
return 8;
}
}
else if (Rs >= 4325) {
return 1;
}
}
}
else if (posBalance >= 50) {
if (Cs < 1909) {
return 8;
}
else if (Cs >= 1909) {
return 4;
}
}
}
else if (RsCsDiff >= 175) {
if (mu_mv_mag_sq < 10) {
if (refDCval < 101) {
if (mcTcor < 95) {
if (ssDCval < 100) {
return 2;
}
else if (ssDCval >= 100) {
if (Rs < 5457) {
return 8;
}
else if (Rs >= 5457) {
return 2;
}
}
}
else if (mcTcor >= 95) {
return 8;
}
}
else if (refDCval >= 101) {
if (TSC < 14) {
if (RsDiff < 19) {
if (TSC < 10) {
if (Rs < 4548) {
if (RsCsDiff < 233) {
return 8;
}
else if (RsCsDiff >= 233) {
if (Rs < 3833) {
return 4;
}
else if (Rs >= 3833) {
return 8;
}
}
}
else if (Rs >= 4548) {
return 1;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (RsDiff >= 19) {
if (Rs < 5016) {
if (Cs < 1364) {
return 8;
}
else if (Cs >= 1364) {
return 2;
}
}
else if (Rs >= 5016) {
if (RsDiff < 23) {
return 8;
}
else if (RsDiff >= 23) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
return 2;
}
}
}
}
}
else if (TSC >= 14) {
if (tcor < 65) {
return 4;
}
else if (tcor >= 65) {
if (Rs < 6534) {
return 1;
}
else if (Rs >= 6534) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (MVDiff < 13) {
if (TSC < 10) {
if (SC < 5240) {
return 2;
}
else if (SC >= 5240) {
if (diffTSC < -1) {
if (Rs < 4507) {
return 8;
}
else if (Rs >= 4507) {
return 2;
}
}
else if (diffTSC >= -1) {
if (Rs < 3616) {
return 8;
}
else if (Rs >= 3616) {
return 4;
}
}
}
}
else if (TSC >= 10) {
if (Cs < 1899) {
return 2;
}
else if (Cs >= 1899) {
return 8;
}
}
}
else if (MVDiff >= 13) {
if (CsDiff < 9) {
if (ssDCval < 106) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 4;
}
}
else if (ssDCval >= 106) {
if (refDCval < 137) {
if (Cs < 1511) {
return 2;
}
else if (Cs >= 1511) {
if (Rs < 4623) {
return 8;
}
else if (Rs >= 4623) {
return 1;
}
}
}
else if (refDCval >= 137) {
if (negBalance < 57) {
if (Cs < 1868) {
return 4;
}
else if (Cs >= 1868) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
}
else if (negBalance >= 57) {
if (Rs < 3570) {
return 2;
}
else if (Rs >= 3570) {
return 1;
}
}
}
}
}
else if (CsDiff >= 9) {
if (ssDCval < 117) {
if (RsDiff < 16) {
if (refDCval < 65) {
if (Rs < 3486) {
return 2;
}
else if (Rs >= 3486) {
return 1;
}
}
else if (refDCval >= 65) {
if (Rs < 4991) {
return 8;
}
else if (Rs >= 4991) {
return 1;
}
}
}
else if (RsDiff >= 16) {
if (Cs < 1777) {
if (MVDiff < 92) {
return 2;
}
else if (MVDiff >= 92) {
return 8;
}
}
else if (Cs >= 1777) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
if (Rs < 3762) {
if (Rs < 3408) {
return 8;
}
else if (Rs >= 3408) {
return 1;
}
}
else if (Rs >= 3762) {
return 2;
}
}
}
}
}
else if (ssDCval >= 117) {
if (Rs < 4839) {
if (Rs < 3533) {
return 2;
}
else if (Rs >= 3533) {
if (Cs < 1848) {
return 1;
}
else if (Cs >= 1848) {
if (Rs < 4366) {
return 4;
}
else if (Rs >= 4366) {
return 1;
}
}
}
}
else if (Rs >= 4839) {
if (Cs < 1496) {
return 1;
}
else if (Cs >= 1496) {
return 4;
}
}
}
}
}
}
}
}
else if (tcor >= 86) {
if (refDCval < 101) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (posBalance < 41) {
if (CsDiff < 8) {
if (negBalance < 43) {
if (Rs < 3768) {
return 4;
}
else if (Rs >= 3768) {
if (Rs < 4179) {
return 2;
}
else if (Rs >= 4179) {
return 8;
}
}
}
else if (negBalance >= 43) {
return 2;
}
}
else if (CsDiff >= 8) {
if (negBalance < 63) {
return 1;
}
else if (negBalance >= 63) {
if (Rs < 4092) {
return 8;
}
else if (Rs >= 4092) {
return 2;
}
}
}
}
else if (posBalance >= 41) {
if (mcTcor < 92) {
if (MVDiff < 19) {
if (Cs < 1856) {
return 2;
}
else if (Cs >= 1856) {
return 8;
}
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (mcTcor >= 92) {
if (refDCval < 63) {
if (Cs < 1412) {
return 4;
}
else if (Cs >= 1412) {
if (Cs < 1806) {
return 2;
}
else if (Cs >= 1806) {
return 8;
}
}
}
else if (refDCval >= 63) {
if (tcor < 88) {
if (AFD < 11) {
if (Rs < 4216) {
return 4;
}
else if (Rs >= 4216) {
return 1;
}
}
else if (AFD >= 11) {
return 2;
}
}
else if (tcor >= 88) {
if (CsDiff < 4) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 4;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 101) {
if (RsDiff < 13) {
if (mu_mv_mag_sq < 71) {
return 8;
}
else if (mu_mv_mag_sq >= 71) {
if (CsDiff < 7) {
if (Rs < 3963) {
return 8;
}
else if (Rs >= 3963) {
return 2;
}
}
else if (CsDiff >= 7) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
else if (RsDiff >= 13) {
if (MVDiff < 20) {
if (Rs < 3523) {
return 8;
}
else if (Rs >= 3523) {
return 2;
}
}
else if (MVDiff >= 20) {
if (ssDCval < 136) {
if (diffRsCsdiff < -196) {
if (Rs < 4175) {
return 1;
}
else if (Rs >= 4175) {
return 2;
}
}
else if (diffRsCsdiff >= -196) {
return 8;
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1919) {
if (AFD < 12) {
if (ssDCval < 100) {
if (Rs < 3795) {
if (negBalance < 44) {
if (Cs < 2759) {
if (posBalance < 56) {
if (mcTcor < 85) {
if (Rs < 1814) {
return 1;
}
else if (Rs >= 1814) {
return 4;
}
}
else if (mcTcor >= 85) {
if (ssDCval < 71) {
if (Rs < 3598) {
return 1;
}
else if (Rs >= 3598) {
return 8;
}
}
else if (ssDCval >= 71) {
if (mu_mv_mag_sq < 93) {
if (SC < 5528) {
if (Rs < 2584) {
return 8;
}
else if (Rs >= 2584) {
return 1;
}
}
else if (SC >= 5528) {
return 8;
}
}
else if (mu_mv_mag_sq >= 93) {
return 4;
}
}
}
}
else if (posBalance >= 56) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (Cs < 2530) {
if (Rs < 2326) {
return 1;
}
else if (Rs >= 2326) {
return 4;
}
}
else if (Cs >= 2530) {
return 8;
}
}
}
}
else if (Cs >= 2759) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
if (RsDiff < 7) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
return 1;
}
}
else if (RsDiff >= 7) {
if (AFD < 8) {
if (Cs < 3941) {
return 8;
}
else if (Cs >= 3941) {
return 4;
}
}
else if (AFD >= 8) {
return 4;
}
}
}
}
}
else if (negBalance >= 44) {
if (Cs < 2668) {
if (posBalance < 52) {
if (Cs < 2106) {
return 8;
}
else if (Cs >= 2106) {
return 4;
}
}
else if (posBalance >= 52) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (Cs < 2021) {
if (Rs < 2742) {
return 8;
}
else if (Rs >= 2742) {
return 4;
}
}
else if (Cs >= 2021) {
return 8;
}
}
}
}
else if (Cs >= 2668) {
if (negBalance < 50) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (Rs < 3071) {
if (Rs < 2242) {
return 4;
}
else if (Rs >= 2242) {
return 8;
}
}
else if (Rs >= 3071) {
return 4;
}
}
}
else if (negBalance >= 50) {
if (SC < 6101) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 8;
}
}
else if (SC >= 6101) {
return 8;
}
}
}
}
}
else if (Rs >= 3795) {
if (tcor < 91) {
if (ssDCval < 73) {
if (SC < 8895) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
if (Cs < 3789) {
return 4;
}
else if (Cs >= 3789) {
return 8;
}
}
}
else if (SC >= 8895) {
return 8;
}
}
else if (ssDCval >= 73) {
if (Rs < 4887) {
if (TSC < 10) {
if (MVDiff < 4) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 4;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (TSC >= 10) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 1;
}
}
}
else if (Rs >= 4887) {
if (Cs < 4693) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (Cs >= 4693) {
return 8;
}
}
}
}
else if (tcor >= 91) {
if (Rs < 4139) {
if (Cs < 2405) {
return 2;
}
else if (Cs >= 2405) {
return 4;
}
}
else if (Rs >= 4139) {
if (ssDCval < 72) {
if (Rs < 5613) {
return 8;
}
else if (Rs >= 5613) {
return 4;
}
}
else if (ssDCval >= 72) {
if (RsDiff < 9) {
if (Cs < 2292) {
return 1;
}
else if (Cs >= 2292) {
if (Rs < 5222) {
return 8;
}
else if (Rs >= 5222) {
return 2;
}
}
}
else if (RsDiff >= 9) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 100) {
if (negBalance < 46) {
if (negBalance < 37) {
if (ssDCval < 113) {
if (SC < 9585) {
if (mcTcor < 89) {
if (Rs < 3131) {
return 2;
}
else if (Rs >= 3131) {
return 8;
}
}
else if (mcTcor >= 89) {
return 2;
}
}
else if (SC >= 9585) {
return 1;
}
}
else if (ssDCval >= 113) {
if (mu_mv_mag_sq < 61) {
if (refDCval < 120) {
return 1;
}
else if (refDCval >= 120) {
if (mcTcor < 76) {
return 2;
}
else if (mcTcor >= 76) {
if (ssDCval < 129) {
return 4;
}
else if (ssDCval >= 129) {
if (posBalance < 51) {
if (Rs < 5147) {
return 1;
}
else if (Rs >= 5147) {
return 8;
}
}
else if (posBalance >= 51) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (mu_mv_mag_sq < 93) {
if (negBalance < 22) {
return 8;
}
else if (negBalance >= 22) {
if (negBalance < 33) {
if (ssDCval < 117) {
return 2;
}
else if (ssDCval >= 117) {
return 1;
}
}
else if (negBalance >= 33) {
if (Rs < 6099) {
return 8;
}
else if (Rs >= 6099) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
return 8;
}
}
}
}
else if (negBalance >= 37) {
if (Rs < 5897) {
if (RsDiff < 8) {
if (Rs < 4204) {
return 8;
}
else if (Rs >= 4204) {
return 1;
}
}
else if (RsDiff >= 8) {
if (refDCval < 102) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (Rs < 3339) {
return 4;
}
else if (Rs >= 3339) {
return 2;
}
}
}
else if (refDCval >= 102) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (mcTcor < 92) {
if (Rs < 2940) {
return 8;
}
else if (Rs >= 2940) {
return 4;
}
}
else if (mcTcor >= 92) {
return 8;
}
}
}
}
}
else if (Rs >= 5897) {
if (Rs < 6970) {
if (Cs < 4875) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (Cs >= 4875) {
return 8;
}
}
else if (Rs >= 6970) {
if (Rs < 8589) {
return 2;
}
else if (Rs >= 8589) {
return 1;
}
}
}
}
}
else if (negBalance >= 46) {
if (tcor < 93) {
if (mu_mv_mag_sq < 81) {
if (mcTcor < 94) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (Rs < 3161) {
return 8;
}
else if (Rs >= 3161) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 3716) {
return 1;
}
else if (Rs >= 3716) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (Rs < 3715) {
if (Rs < 3170) {
return 1;
}
else if (Rs >= 3170) {
return 8;
}
}
else if (Rs >= 3715) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
return 1;
}
}
}
}
else if (tcor >= 93) {
if (refDCval < 103) {
if (Rs < 2805) {
return 8;
}
else if (Rs >= 2805) {
return 4;
}
}
else if (refDCval >= 103) {
return 8;
}
}
}
}
}
else if (AFD >= 12) {
if (AFD < 24) {
if (mu_mv_mag_sq < 76) {
if (refDCval < 156) {
if (refDCval < 101) {
if (tcor < 59) {
if (ssDCval < 71) {
if (RsCsDiff < 271) {
if (TSC < 14) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
return 2;
}
}
else if (TSC >= 14) {
return 4;
}
}
else if (RsCsDiff >= 271) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (ssDCval >= 71) {
if (negBalance < 54) {
if (Cs < 3213) {
return 8;
}
else if (Cs >= 3213) {
return 1;
}
}
else if (negBalance >= 54) {
if (Cs < 2184) {
return 8;
}
else if (Cs >= 2184) {
return 4;
}
}
}
}
else if (tcor >= 59) {
if (posBalance < 49) {
if (ssDCval < 78) {
if (SC < 6454) {
if (AFD < 15) {
if (mu_mv_mag_sq < 66) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
return 8;
}
}
else if (mu_mv_mag_sq >= 66) {
return 4;
}
}
else if (AFD >= 15) {
if (negBalance < 55) {
if (CsDiff < 14) {
return 2;
}
else if (CsDiff >= 14) {
return 4;
}
}
else if (negBalance >= 55) {
if (ssDCval < 56) {
if (Rs < 2222) {
return 1;
}
else if (Rs >= 2222) {
return 4;
}
}
else if (ssDCval >= 56) {
return 8;
}
}
}
}
else if (SC >= 6454) {
if (tcor < 81) {
if (SC < 10198) {
if (SC < 6648) {
if (SC < 6641) {
return 4;
}
else if (SC >= 6641) {
return 2;
}
}
else if (SC >= 6648) {
return 4;
}
}
else if (SC >= 10198) {
return 2;
}
}
else if (tcor >= 81) {
if (Cs < 2336) {
return 4;
}
else if (Cs >= 2336) {
if (diffRsCsdiff < -131) {
if (Cs < 3108) {
return 4;
}
else if (Cs >= 3108) {
return 2;
}
}
else if (diffRsCsdiff >= -131) {
if (mu_mv_mag_sq < 73) {
return 8;
}
else if (mu_mv_mag_sq >= 73) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 78) {
if (refDCval < 91) {
if (Rs < 2672) {
return 8;
}
else if (Rs >= 2672) {
if (RsDiff < 19) {
return 2;
}
else if (RsDiff >= 19) {
if (TSC < 14) {
if (Rs < 6623) {
return 2;
}
else if (Rs >= 6623) {
return 8;
}
}
else if (TSC >= 14) {
return 1;
}
}
}
}
else if (refDCval >= 91) {
if (Cs < 3360) {
if (Rs < 2372) {
return 1;
}
else if (Rs >= 2372) {
return 8;
}
}
else if (Cs >= 3360) {
if (Rs < 6748) {
return 1;
}
else if (Rs >= 6748) {
return 8;
}
}
}
}
}
else if (posBalance >= 49) {
if (ssDCval < 80) {
if (Rs < 1888) {
if (ssDCval < 59) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (Cs < 4045) {
return 2;
}
else if (Cs >= 4045) {
return 1;
}
}
}
else if (ssDCval >= 59) {
if (mcTcor < 91) {
if (negBalance < 60) {
if (SC < 4112) {
return 4;
}
else if (SC >= 4112) {
if (Rs < 1867) {
if (Rs < 1455) {
return 1;
}
else if (Rs >= 1455) {
return 2;
}
}
else if (Rs >= 1867) {
return 4;
}
}
}
else if (negBalance >= 60) {
return 2;
}
}
else if (mcTcor >= 91) {
if (diffMVdiffVal < -7) {
if (mcTcor < 95) {
if (Rs < 1450) {
return 4;
}
else if (Rs >= 1450) {
return 8;
}
}
else if (mcTcor >= 95) {
return 2;
}
}
else if (diffMVdiffVal >= -7) {
if (MVDiff < 41) {
return 8;
}
else if (MVDiff >= 41) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1888) {
if (RsDiff < 14) {
if (MVDiff < 6) {
if (AFD < 17) {
if (mu_mv_mag_sq < 3) {
if (Cs < 2391) {
if (Rs < 2729) {
return 8;
}
else if (Rs >= 2729) {
return 4;
}
}
else if (Cs >= 2391) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (mu_mv_mag_sq < 24) {
return 1;
}
else if (mu_mv_mag_sq >= 24) {
if (diffMVdiffVal < 4) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
if (Rs < 3085) {
return 8;
}
else if (Rs >= 3085) {
return 1;
}
}
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
}
}
else if (AFD >= 17) {
return 1;
}
}
else if (MVDiff >= 6) {
if (mcTcor < 76) {
if (Rs < 2234) {
return 2;
}
else if (Rs >= 2234) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (Cs < 3005) {
return 8;
}
else if (Cs >= 3005) {
return 4;
}
}
}
}
else if (mcTcor >= 76) {
if (gchDC < 2) {
if (tcor < 63) {
if (Rs < 2089) {
return 4;
}
else if (Rs >= 2089) {
if (Rs < 2264) {
return 1;
}
else if (Rs >= 2264) {
return 8;
}
}
}
else if (tcor >= 63) {
if (Rs < 2892) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (Rs < 2695) {
return 2;
}
else if (Rs >= 2695) {
return 8;
}
}
}
else if (Rs >= 2892) {
if (Cs < 2596) {
if (mu_mv_mag_sq < 53) {
if (SC < 5398) {
if (Rs < 3269) {
return 4;
}
else if (Rs >= 3269) {
return 8;
}
}
else if (SC >= 5398) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 53) {
return 4;
}
}
else if (Cs >= 2596) {
if (MVDiff < 10) {
if (Cs < 2971) {
if (Rs < 4144) {
return 1;
}
else if (Rs >= 4144) {
return 8;
}
}
else if (Cs >= 2971) {
return 8;
}
}
else if (MVDiff >= 10) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
return 4;
}
}
}
}
}
}
}
else if (gchDC >= 2) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
if (tcor < 77) {
if (negBalance < 50) {
if (RsDiff < 13) {
if (SC < 4307) {
return 1;
}
else if (SC >= 4307) {
return 8;
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (negBalance >= 50) {
return 1;
}
}
else if (tcor >= 77) {
if (RsCsDiff < 212) {
return 8;
}
else if (RsCsDiff >= 212) {
if (negBalance < 51) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
return 4;
}
}
else if (negBalance >= 51) {
return 4;
}
}
}
}
}
}
}
}
else if (RsDiff >= 14) {
if (refDCval < 65) {
if (ssDCval < 49) {
return 2;
}
else if (ssDCval >= 49) {
if (tcor < 82) {
if (RsDiff < 20) {
return 4;
}
else if (RsDiff >= 20) {
if (Rs < 4079) {
return 8;
}
else if (Rs >= 4079) {
if (Rs < 4351) {
return 2;
}
else if (Rs >= 4351) {
return 4;
}
}
}
}
else if (tcor >= 82) {
return 8;
}
}
}
else if (refDCval >= 65) {
if (mu_mv_mag_sq < 2) {
if (CsDiff < 16) {
if (Rs < 2712) {
return 4;
}
else if (Rs >= 2712) {
if (Cs < 3626) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (Cs >= 3626) {
return 4;
}
}
}
else if (CsDiff >= 16) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (posBalance < 59) {
if (gchDC < 1) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (RsCsDiff < 475) {
return 8;
}
else if (RsCsDiff >= 475) {
if (Rs < 4771) {
return 4;
}
else if (Rs >= 4771) {
return 2;
}
}
}
}
else if (gchDC >= 1) {
if (gchDC < 2) {
if (CsDiff < 11) {
return 2;
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (gchDC >= 2) {
if (Cs < 3682) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (Cs >= 3682) {
if (Rs < 4825) {
return 2;
}
else if (Rs >= 4825) {
return 8;
}
}
}
}
}
else if (posBalance >= 59) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 80) {
if (mcTcor < 94) {
if (SC < 4974) {
if (refDCval < 94) {
if (Cs < 2472) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
if (TSC < 11) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
if (ssDCval < 85) {
return 4;
}
else if (ssDCval >= 85) {
return 8;
}
}
}
else if (TSC >= 11) {
if (SC < 4130) {
if (Rs < 1998) {
return 4;
}
else if (Rs >= 1998) {
return 1;
}
}
else if (SC >= 4130) {
if (negBalance < 45) {
if (Rs < 2192) {
return 4;
}
else if (Rs >= 2192) {
return 1;
}
}
else if (negBalance >= 45) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
if (Rs < 2362) {
return 1;
}
else if (Rs >= 2362) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 2472) {
if (SC < 4947) {
if (Rs < 1441) {
return 8;
}
else if (Rs >= 1441) {
return 4;
}
}
else if (SC >= 4947) {
return 8;
}
}
}
else if (refDCval >= 94) {
if (Rs < 951) {
return 2;
}
else if (Rs >= 951) {
if (RsDiff < 7) {
if (Cs < 2240) {
return 4;
}
else if (Cs >= 2240) {
return 8;
}
}
else if (RsDiff >= 7) {
return 8;
}
}
}
}
else if (SC >= 4974) {
if (SC < 8277) {
if (negBalance < 51) {
if (MVDiff < 6) {
if (SC < 7038) {
if (RsDiff < 15) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
if (Rs < 3190) {
return 4;
}
else if (Rs >= 3190) {
if (RsDiff < 12) {
return 2;
}
else if (RsDiff >= 12) {
return 4;
}
}
}
}
else if (RsDiff >= 15) {
return 2;
}
}
else if (SC >= 7038) {
if (mcTcor < 90) {
if (Rs < 3876) {
return 8;
}
else if (Rs >= 3876) {
return 4;
}
}
else if (mcTcor >= 90) {
if (Rs < 3244) {
return 8;
}
else if (Rs >= 3244) {
return 2;
}
}
}
}
else if (MVDiff >= 6) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (MVDiff < 11) {
if (refDCval < 97) {
if (CsDiff < 9) {
return 1;
}
else if (CsDiff >= 9) {
if (Rs < 2781) {
return 8;
}
else if (Rs >= 2781) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 4;
}
}
}
}
else if (refDCval >= 97) {
return 2;
}
}
else if (MVDiff >= 11) {
if (posBalance < 61) {
if (refDCval < 83) {
if (Rs < 4033) {
if (Rs < 2895) {
return 8;
}
else if (Rs >= 2895) {
return 4;
}
}
else if (Rs >= 4033) {
return 2;
}
}
else if (refDCval >= 83) {
if (RsDiff < 18) {
return 8;
}
else if (RsDiff >= 18) {
return 4;
}
}
}
else if (posBalance >= 61) {
if (mu_mv_mag_sq < 3) {
if (Cs < 2619) {
if (Rs < 2920) {
return 4;
}
else if (Rs >= 2920) {
return 2;
}
}
else if (Cs >= 2619) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 5583) {
if (Rs < 3222) {
return 2;
}
else if (Rs >= 3222) {
return 8;
}
}
else if (SC >= 5583) {
if (SC < 5705) {
return 4;
}
else if (SC >= 5705) {
if (Cs < 2598) {
return 8;
}
else if (Cs >= 2598) {
if (MVDiff < 14) {
if (Rs < 4103) {
return 8;
}
else if (Rs >= 4103) {
return 4;
}
}
else if (MVDiff >= 14) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (negBalance >= 51) {
if (refDCval < 92) {
if (AFD < 16) {
if (Cs < 3373) {
if (SC < 5085) {
return 8;
}
else if (SC >= 5085) {
if (ssDCval < 81) {
if (Rs < 2999) {
return 4;
}
else if (Rs >= 2999) {
return 2;
}
}
else if (ssDCval >= 81) {
return 4;
}
}
}
else if (Cs >= 3373) {
return 2;
}
}
else if (AFD >= 16) {
if (Cs < 2275) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 8;
}
}
}
else if (Cs >= 2275) {
if (Cs < 3715) {
if (mcTcor < 88) {
if (AFD < 22) {
if (posBalance < 56) {
if (Rs < 3849) {
return 8;
}
else if (Rs >= 3849) {
return 1;
}
}
else if (posBalance >= 56) {
if (negBalance < 57) {
if (SC < 5541) {
return 8;
}
else if (SC >= 5541) {
if (SC < 6955) {
return 4;
}
else if (SC >= 6955) {
if (SC < 7505) {
return 8;
}
else if (SC >= 7505) {
return 4;
}
}
}
}
else if (negBalance >= 57) {
return 4;
}
}
}
else if (AFD >= 22) {
return 1;
}
}
else if (mcTcor >= 88) {
if (RsCsDiff < 260) {
if (posBalance < 57) {
if (Rs < 3679) {
return 1;
}
else if (Rs >= 3679) {
return 8;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (RsCsDiff >= 260) {
if (MVDiff < 16) {
if (Rs < 2742) {
return 8;
}
else if (Rs >= 2742) {
return 4;
}
}
else if (MVDiff >= 16) {
return 8;
}
}
}
}
else if (Cs >= 3715) {
return 8;
}
}
}
}
else if (refDCval >= 92) {
if (Cs < 2427) {
if (Rs < 3359) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
return 2;
}
}
else if (Rs >= 3359) {
if (Rs < 4646) {
return 4;
}
else if (Rs >= 4646) {
return 8;
}
}
}
else if (Cs >= 2427) {
if (SC < 7574) {
if (diffAFD < 5) {
return 4;
}
else if (diffAFD >= 5) {
if (Rs < 2879) {
return 4;
}
else if (Rs >= 2879) {
return 8;
}
}
}
else if (SC >= 7574) {
if (RsCsDiff < 319) {
return 8;
}
else if (RsCsDiff >= 319) {
return 4;
}
}
}
}
}
}
else if (SC >= 8277) {
if (SC < 11043) {
if (RsDiff < 19) {
return 8;
}
else if (RsDiff >= 19) {
if (tcor < 74) {
return 8;
}
else if (tcor >= 74) {
if (Rs < 6082) {
return 2;
}
else if (Rs >= 6082) {
return 4;
}
}
}
}
else if (SC >= 11043) {
return 4;
}
}
}
}
else if (mcTcor >= 94) {
if (RsDiff < 10) {
if (Cs < 2201) {
if (gchDC < 2) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 1;
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (Cs >= 2201) {
if (diffTSC < -2) {
if (Rs < 2429) {
if (Rs < 1790) {
return 1;
}
else if (Rs >= 1790) {
return 4;
}
}
else if (Rs >= 2429) {
return 8;
}
}
else if (diffTSC >= -2) {
if (MVDiff < 10) {
if (Rs < 3302) {
return 1;
}
else if (Rs >= 3302) {
return 4;
}
}
else if (MVDiff >= 10) {
if (Rs < 1986) {
return 4;
}
else if (Rs >= 1986) {
return 8;
}
}
}
}
}
else if (RsDiff >= 10) {
if (Rs < 4269) {
if (Cs < 2049) {
if (diffAFD < 1) {
if (Rs < 2675) {
return 8;
}
else if (Rs >= 2675) {
return 2;
}
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (Cs >= 2049) {
if (mu_mv_mag_sq < 2) {
return 4;
}
else if (mu_mv_mag_sq >= 2) {
if (diffMVdiffVal < -3) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (negBalance < 55) {
if (Rs < 2431) {
if (Rs < 2356) {
return 8;
}
else if (Rs >= 2356) {
return 2;
}
}
else if (Rs >= 2431) {
return 8;
}
}
else if (negBalance >= 55) {
return 2;
}
}
}
}
}
else if (Rs >= 4269) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 101) {
if (MVDiff < 10) {
if (tcor < 88) {
if (TSC < 11) {
if (diffAFD < 6) {
if (diffAFD < -2) {
if (mu_mv_mag_sq < 3) {
if (Cs < 2627) {
return 8;
}
else if (Cs >= 2627) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 2635) {
if (Rs < 2555) {
return 2;
}
else if (Rs >= 2555) {
return 8;
}
}
else if (Rs >= 2635) {
return 4;
}
}
}
else if (diffAFD >= -2) {
if (mcTcor < 96) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
if (refDCval < 107) {
if (Cs < 2293) {
return 8;
}
else if (Cs >= 2293) {
if (SC < 5838) {
return 4;
}
else if (SC >= 5838) {
if (AFD < 16) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (AFD >= 16) {
return 4;
}
}
}
}
else if (refDCval >= 107) {
if (Cs < 1991) {
if (refDCval < 132) {
return 8;
}
else if (refDCval >= 132) {
return 4;
}
}
else if (Cs >= 1991) {
if (negBalance < 43) {
if (tcor < 83) {
return 8;
}
else if (tcor >= 83) {
if (Rs < 4109) {
return 1;
}
else if (Rs >= 4109) {
return 4;
}
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 7519) {
return 4;
}
else if (Rs >= 7519) {
return 8;
}
}
}
}
else if (diffAFD >= 6) {
if (Rs < 3554) {
return 1;
}
else if (Rs >= 3554) {
if (Rs < 4344) {
return 8;
}
else if (Rs >= 4344) {
return 2;
}
}
}
}
else if (TSC >= 11) {
if (Cs < 2597) {
if (Rs < 2640) {
if (Cs < 2168) {
if (Rs < 1968) {
return 4;
}
else if (Rs >= 1968) {
return 8;
}
}
else if (Cs >= 2168) {
return 2;
}
}
else if (Rs >= 2640) {
if (MVDiff < 6) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Cs < 2205) {
return 8;
}
else if (Cs >= 2205) {
return 1;
}
}
}
else if (MVDiff >= 6) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (posBalance < 53) {
if (Rs < 3344) {
return 4;
}
else if (Rs >= 3344) {
return 2;
}
}
else if (posBalance >= 53) {
if (Rs < 2783) {
if (Rs < 2733) {
return 8;
}
else if (Rs >= 2733) {
return 4;
}
}
else if (Rs >= 2783) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 2597) {
if (Rs < 3726) {
if (posBalance < 64) {
if (tcor < 68) {
if (Rs < 3123) {
return 4;
}
else if (Rs >= 3123) {
return 8;
}
}
else if (tcor >= 68) {
return 4;
}
}
else if (posBalance >= 64) {
return 2;
}
}
else if (Rs >= 3726) {
if (posBalance < 59) {
if (mcTcor < 87) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
if (Rs < 4965) {
return 4;
}
else if (Rs >= 4965) {
return 8;
}
}
}
else if (mcTcor >= 87) {
return 4;
}
}
else if (posBalance >= 59) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
return 4;
}
}
}
}
}
}
else if (tcor >= 88) {
if (RsCsDiff < 367) {
if (Rs < 5262) {
if (RsCsDiff < 183) {
if (mcTcor < 91) {
return 1;
}
else if (mcTcor >= 91) {
return 8;
}
}
else if (RsCsDiff >= 183) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (Rs >= 5262) {
return 4;
}
}
else if (RsCsDiff >= 367) {
return 2;
}
}
}
else if (MVDiff >= 10) {
if (Rs < 8238) {
if (SC < 4071) {
if (tcor < 77) {
if (tcor < 63) {
if (SC < 3829) {
return 1;
}
else if (SC >= 3829) {
return 2;
}
}
else if (tcor >= 63) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 2;
}
}
}
else if (tcor >= 77) {
if (Rs < 1594) {
if (ssDCval < 106) {
return 2;
}
else if (ssDCval >= 106) {
return 1;
}
}
else if (Rs >= 1594) {
if (Rs < 1816) {
return 8;
}
else if (Rs >= 1816) {
if (Rs < 1964) {
return 1;
}
else if (Rs >= 1964) {
return 8;
}
}
}
}
}
else if (SC >= 4071) {
if (ssDCval < 105) {
if (diffAFD < -3) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
else if (diffAFD >= -3) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
return 2;
}
}
}
else if (ssDCval >= 105) {
if (RsCsDiff < 242) {
if (tcor < 73) {
if (Rs < 3975) {
if (negBalance < 55) {
if (Cs < 1959) {
return 4;
}
else if (Cs >= 1959) {
return 8;
}
}
else if (negBalance >= 55) {
if (Rs < 2739) {
return 2;
}
else if (Rs >= 2739) {
return 1;
}
}
}
else if (Rs >= 3975) {
if (Rs < 4469) {
return 1;
}
else if (Rs >= 4469) {
return 8;
}
}
}
else if (tcor >= 73) {
if (negBalance < 41) {
if (posBalance < 57) {
if (Cs < 2286) {
if (Rs < 3084) {
return 2;
}
else if (Rs >= 3084) {
return 8;
}
}
else if (Cs >= 2286) {
return 1;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (negBalance >= 41) {
if (MVDiff < 12) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
return 8;
}
}
}
else if (MVDiff >= 12) {
return 8;
}
}
}
}
else if (RsCsDiff >= 242) {
if (tcor < 87) {
if (mu_mv_mag_sq < 39) {
if (tcor < 67) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
if (CsDiff < 18) {
if (CsDiff < 11) {
if (AFD < 21) {
if (Rs < 4985) {
return 8;
}
else if (Rs >= 4985) {
return 4;
}
}
else if (AFD >= 21) {
return 1;
}
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (CsDiff >= 18) {
if (Rs < 3645) {
return 1;
}
else if (Rs >= 3645) {
return 8;
}
}
}
}
else if (tcor >= 67) {
if (Rs < 3914) {
if (refDCval < 132) {
if (mcTcor < 86) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
if (Cs < 2377) {
return 8;
}
else if (Cs >= 2377) {
return 4;
}
}
}
else if (mcTcor >= 86) {
if (Cs < 2805) {
if (TSC < 13) {
if (Rs < 2020) {
return 4;
}
else if (Rs >= 2020) {
return 2;
}
}
else if (TSC >= 13) {
return 8;
}
}
else if (Cs >= 2805) {
if (Cs < 3936) {
return 4;
}
else if (Cs >= 3936) {
return 2;
}
}
}
}
else if (refDCval >= 132) {
return 8;
}
}
else if (Rs >= 3914) {
if (Rs < 5817) {
if (ssDCval < 134) {
if (CsDiff < 13) {
if (AFD < 17) {
if (Rs < 4260) {
return 1;
}
else if (Rs >= 4260) {
return 8;
}
}
else if (AFD >= 17) {
return 1;
}
}
else if (CsDiff >= 13) {
return 4;
}
}
else if (ssDCval >= 134) {
if (Rs < 5709) {
if (Cs < 1925) {
return 2;
}
else if (Cs >= 1925) {
return 4;
}
}
else if (Rs >= 5709) {
return 8;
}
}
}
else if (Rs >= 5817) {
if (Cs < 3137) {
return 8;
}
else if (Cs >= 3137) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (Cs < 3764) {
return 8;
}
else if (Cs >= 3764) {
if (CsDiff < 19) {
if (tcor < 65) {
return 8;
}
else if (tcor >= 65) {
if (refDCval < 132) {
if (Rs < 4807) {
return 4;
}
else if (Rs >= 4807) {
return 8;
}
}
else if (refDCval >= 132) {
if (Rs < 3899) {
return 8;
}
else if (Rs >= 3899) {
return 1;
}
}
}
}
else if (CsDiff >= 19) {
return 8;
}
}
}
}
else if (tcor >= 87) {
if (RsCsDiff < 389) {
return 8;
}
else if (RsCsDiff >= 389) {
if (Rs < 3182) {
return 8;
}
else if (Rs >= 3182) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 8238) {
if (ssDCval < 136) {
return 1;
}
else if (ssDCval >= 136) {
if (Rs < 9401) {
return 4;
}
else if (Rs >= 9401) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 156) {
if (posBalance < 52) {
if (Cs < 2756) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
if (Rs < 2935) {
return 1;
}
else if (Rs >= 2935) {
return 2;
}
}
}
else if (Cs >= 2756) {
if (ssDCval < 155) {
return 1;
}
else if (ssDCval >= 155) {
return 8;
}
}
}
else if (posBalance >= 52) {
if (Rs < 3272) {
return 8;
}
else if (Rs >= 3272) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (SC < 7046) {
if (Rs < 4040) {
if (refDCval < 83) {
if (tcor < 71) {
if (RsCsDiff < 308) {
return 4;
}
else if (RsCsDiff >= 308) {
if (Rs < 3085) {
return 1;
}
else if (Rs >= 3085) {
return 8;
}
}
}
else if (tcor >= 71) {
if (Cs < 2099) {
return 4;
}
else if (Cs >= 2099) {
if (negBalance < 50) {
if (SC < 5661) {
return 8;
}
else if (SC >= 5661) {
if (Rs < 3686) {
return 4;
}
else if (Rs >= 3686) {
if (Cs < 2738) {
return 8;
}
else if (Cs >= 2738) {
return 1;
}
}
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
}
else if (refDCval >= 83) {
if (Cs < 2612) {
if (refDCval < 85) {
if (Rs < 3119) {
return 8;
}
else if (Rs >= 3119) {
return 4;
}
}
else if (refDCval >= 85) {
if (diffTSC < -3) {
if (Rs < 3425) {
return 2;
}
else if (Rs >= 3425) {
return 8;
}
}
else if (diffTSC >= -3) {
return 8;
}
}
}
else if (Cs >= 2612) {
if (ssDCval < 110) {
return 4;
}
else if (ssDCval >= 110) {
if (Rs < 3459) {
return 8;
}
else if (Rs >= 3459) {
return 4;
}
}
}
}
}
else if (Rs >= 4040) {
if (Rs < 4634) {
return 1;
}
else if (Rs >= 4634) {
return 8;
}
}
}
else if (SC >= 7046) {
if (negBalance < 52) {
if (ssDCval < 80) {
if (Rs < 4616) {
if (posBalance < 52) {
if (Rs < 4315) {
return 4;
}
else if (Rs >= 4315) {
return 2;
}
}
else if (posBalance >= 52) {
return 8;
}
}
else if (Rs >= 4616) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
return 4;
}
}
}
}
else if (ssDCval >= 80) {
if (AFD < 16) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
if (SC < 7649) {
if (SC < 7631) {
return 8;
}
else if (SC >= 7631) {
return 4;
}
}
else if (SC >= 7649) {
return 8;
}
}
}
else if (AFD >= 16) {
if (diffAFD < 2) {
if (RsDiff < 18) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
}
else if (RsDiff >= 18) {
if (Rs < 8151) {
return 2;
}
else if (Rs >= 8151) {
return 4;
}
}
}
else if (diffAFD >= 2) {
if (Cs < 4531) {
return 1;
}
else if (Cs >= 4531) {
return 8;
}
}
}
}
}
else if (negBalance >= 52) {
if (posBalance < 57) {
if (Rs < 5359) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 8;
}
}
else if (Rs >= 5359) {
return 4;
}
}
else if (posBalance >= 57) {
if (Rs < 8555) {
return 8;
}
else if (Rs >= 8555) {
return 4;
}
}
}
}
}
}
else if (AFD >= 24) {
if (SC < 11496) {
if (Rs < 5702) {
if (negBalance < 56) {
if (mu_mv_mag_sq < 22) {
if (posBalance < 63) {
if (posBalance < 57) {
if (RsCsDiff < 424) {
if (Cs < 2110) {
if (Rs < 3035) {
return 4;
}
else if (Rs >= 3035) {
return 1;
}
}
else if (Cs >= 2110) {
return 8;
}
}
else if (RsCsDiff >= 424) {
if (ssDCval < 149) {
if (Cs < 3742) {
if (RsCsDiff < 1306) {
if (diffTSC < -1) {
if (Rs < 2609) {
return 2;
}
else if (Rs >= 2609) {
return 8;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (RsCsDiff >= 1306) {
return 4;
}
}
else if (Cs >= 3742) {
return 8;
}
}
else if (ssDCval >= 149) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
if (CsDiff < 18) {
if (SC < 3555) {
return 4;
}
else if (SC >= 3555) {
if (Cs < 1951) {
return 4;
}
else if (Cs >= 1951) {
if (ssDCval < 64) {
if (Rs < 1582) {
return 2;
}
else if (Rs >= 1582) {
return 4;
}
}
else if (ssDCval >= 64) {
return 2;
}
}
}
}
else if (CsDiff >= 18) {
if (Cs < 5398) {
return 4;
}
else if (Cs >= 5398) {
return 2;
}
}
}
}
}
else if (posBalance >= 63) {
if (mcTcor < 90) {
if (Rs < 3885) {
if (mu_mv_mag_sq < 2) {
if (Cs < 2216) {
return 1;
}
else if (Cs >= 2216) {
if (Rs < 2255) {
return 2;
}
else if (Rs >= 2255) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (refDCval < 80) {
if (Rs < 3546) {
return 4;
}
else if (Rs >= 3546) {
return 8;
}
}
else if (refDCval >= 80) {
if (Rs < 3499) {
if (mu_mv_mag_sq < 17) {
return 8;
}
else if (mu_mv_mag_sq >= 17) {
if (Rs < 1973) {
return 2;
}
else if (Rs >= 1973) {
return 8;
}
}
}
else if (Rs >= 3499) {
return 1;
}
}
}
}
else if (Rs >= 3885) {
if (TSC < 18) {
return 1;
}
else if (TSC >= 18) {
return 4;
}
}
}
else if (mcTcor >= 90) {
if (SC < 5562) {
if (Cs < 1937) {
return 8;
}
else if (Cs >= 1937) {
if (diffRsCsdiff < 51) {
return 2;
}
else if (diffRsCsdiff >= 51) {
if (Rs < 2433) {
return 4;
}
else if (Rs >= 2433) {
return 2;
}
}
}
}
else if (SC >= 5562) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 22) {
if (Rs < 3756) {
if (Rs < 1040) {
return 2;
}
else if (Rs >= 1040) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (Rs < 3019) {
return 8;
}
else if (Rs >= 3019) {
return 2;
}
}
}
}
else if (Rs >= 3756) {
if (mcTcor < 93) {
if (diffAFD < -1) {
if (Rs < 5352) {
return 1;
}
else if (Rs >= 5352) {
return 4;
}
}
else if (diffAFD >= -1) {
if (AFD < 25) {
return 4;
}
else if (AFD >= 25) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
if (RsCsDiff < 688) {
return 8;
}
else if (RsCsDiff >= 688) {
if (Cs < 4725) {
return 4;
}
else if (Cs >= 4725) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 146) {
return 2;
}
else if (ssDCval >= 146) {
if (Rs < 5054) {
return 4;
}
else if (Rs >= 5054) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 56) {
if (tcor < 48) {
if (Cs < 2692) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Cs < 2098) {
return 1;
}
else if (Cs >= 2098) {
if (Cs < 2495) {
if (CsDiff < 14) {
return 1;
}
else if (CsDiff >= 14) {
return 8;
}
}
else if (Cs >= 2495) {
return 1;
}
}
}
}
else if (Cs >= 2692) {
if (Rs < 1098) {
return 2;
}
else if (Rs >= 1098) {
return 8;
}
}
}
else if (tcor >= 48) {
if (ssDCval < 99) {
if (ssDCval < 80) {
if (Cs < 3513) {
if (negBalance < 58) {
if (Cs < 1946) {
return 8;
}
else if (Cs >= 1946) {
return 2;
}
}
else if (negBalance >= 58) {
if (tcor < 59) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
return 1;
}
}
else if (tcor >= 59) {
if (Rs < 956) {
return 8;
}
else if (Rs >= 956) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
if (Rs < 2319) {
return 1;
}
else if (Rs >= 2319) {
return 4;
}
}
}
}
}
}
else if (Cs >= 3513) {
if (MVDiff < 167) {
return 2;
}
else if (MVDiff >= 167) {
if (Rs < 2920) {
return 1;
}
else if (Rs >= 2920) {
return 8;
}
}
}
}
else if (ssDCval >= 80) {
if (RsDiff < 9) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 4;
}
}
else if (RsDiff >= 9) {
if (mcTcor < 83) {
if (ssDCval < 85) {
return 4;
}
else if (ssDCval >= 85) {
if (TSC < 18) {
return 2;
}
else if (TSC >= 18) {
return 1;
}
}
}
else if (mcTcor >= 83) {
if (MVDiff < 20) {
if (diffMVdiffVal < -1) {
if (Rs < 3752) {
return 4;
}
else if (Rs >= 3752) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2673) {
if (Rs < 2148) {
return 8;
}
else if (Rs >= 2148) {
return 1;
}
}
else if (Rs >= 2673) {
return 8;
}
}
}
else if (MVDiff >= 20) {
return 8;
}
}
}
}
}
else if (ssDCval >= 99) {
if (TSC < 16) {
if (RsCsDiff < 223) {
return 2;
}
else if (RsCsDiff >= 223) {
if (posBalance < 58) {
if (Cs < 3133) {
if (Cs < 2132) {
if (Rs < 3581) {
if (Rs < 3050) {
return 2;
}
else if (Rs >= 3050) {
return 8;
}
}
else if (Rs >= 3581) {
return 1;
}
}
else if (Cs >= 2132) {
return 8;
}
}
else if (Cs >= 3133) {
if (RsDiff < 23) {
return 2;
}
else if (RsDiff >= 23) {
return 4;
}
}
}
else if (posBalance >= 58) {
if (Cs < 2344) {
if (Rs < 2595) {
return 2;
}
else if (Rs >= 2595) {
if (RsDiff < 17) {
if (Rs < 3392) {
return 4;
}
else if (Rs >= 3392) {
return 1;
}
}
else if (RsDiff >= 17) {
return 8;
}
}
}
else if (Cs >= 2344) {
if (RsCsDiff < 438) {
return 1;
}
else if (RsCsDiff >= 438) {
if (Cs < 3874) {
if (AFD < 26) {
return 4;
}
else if (AFD >= 26) {
return 8;
}
}
else if (Cs >= 3874) {
return 1;
}
}
}
}
}
}
else if (TSC >= 16) {
if (MVDiff < 17) {
if (Rs < 3924) {
return 4;
}
else if (Rs >= 3924) {
return 1;
}
}
else if (MVDiff >= 17) {
if (gchDC < 5) {
if (CsDiff < 20) {
return 8;
}
else if (CsDiff >= 20) {
return 4;
}
}
else if (gchDC >= 5) {
if (Rs < 2169) {
return 8;
}
else if (Rs >= 2169) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 5702) {
if (negBalance < 46) {
if (mu_mv_mag_sq < 6) {
if (TSC < 16) {
return 8;
}
else if (TSC >= 16) {
return 1;
}
}
else if (mu_mv_mag_sq >= 6) {
if (Rs < 6322) {
if (Rs < 6089) {
return 8;
}
else if (Rs >= 6089) {
return 4;
}
}
else if (Rs >= 6322) {
return 2;
}
}
}
else if (negBalance >= 46) {
if (ssDCval < 166) {
if (diffRsCsdiff < -196) {
if (Rs < 6663) {
return 2;
}
else if (Rs >= 6663) {
return 4;
}
}
else if (diffRsCsdiff >= -196) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
return 4;
}
}
}
}
else if (ssDCval >= 166) {
return 2;
}
}
}
}
else if (SC >= 11496) {
if (mu_mv_mag_sq < 37) {
if (TSC < 10) {
if (Cs < 6135) {
return 1;
}
else if (Cs >= 6135) {
return 8;
}
}
else if (TSC >= 10) {
if (AFD < 41) {
return 4;
}
else if (AFD >= 41) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 37) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Cs < 1661) {
if (Rs < 3098) {
if (Rs < 1309) {
if (Rs < 1089) {
if (AFD < 1) {
if (Rs < 604) {
return 4;
}
else if (Rs >= 604) {
return 2;
}
}
else if (AFD >= 1) {
if (refDCval < 63) {
if (Rs < 786) {
return 1;
}
else if (Rs >= 786) {
if (ssDCval < 60) {
if (mcTcor < 96) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (mcTcor >= 96) {
return 8;
}
}
else if (ssDCval >= 60) {
if (Cs < 1508) {
return 1;
}
else if (Cs >= 1508) {
return 8;
}
}
}
}
else if (refDCval >= 63) {
if (ssDCval < 76) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (Rs < 875) {
return 8;
}
else if (Rs >= 875) {
return 2;
}
}
}
else if (ssDCval >= 76) {
if (refDCval < 83) {
return 1;
}
else if (refDCval >= 83) {
if (ssDCval < 129) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Cs < 1462) {
if (Cs < 1356) {
return 4;
}
else if (Cs >= 1356) {
if (Rs < 478) {
return 4;
}
else if (Rs >= 478) {
return 8;
}
}
}
else if (Cs >= 1462) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
else if (ssDCval >= 129) {
if (Rs < 577) {
return 2;
}
else if (Rs >= 577) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1089) {
if (mu_mv_mag_sq < 122) {
if (TSC < 2) {
if (Cs < 1441) {
return 1;
}
else if (Cs >= 1441) {
if (Rs < 1259) {
return 4;
}
else if (Rs >= 1259) {
return 8;
}
}
}
else if (TSC >= 2) {
if (MVDiff < 3) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (mu_mv_mag_sq < 116) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (Cs < 1628) {
return 8;
}
else if (Cs >= 1628) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 116) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 1240) {
return 4;
}
else if (Rs >= 1240) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1309) {
if (ssDCval < 86) {
if (tcor < 98) {
if (Rs < 2221) {
if (MVDiff < 8) {
if (Rs < 1418) {
if (ssDCval < 52) {
if (Cs < 1493) {
return 8;
}
else if (Cs >= 1493) {
return 4;
}
}
else if (ssDCval >= 52) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (Rs < 1342) {
return 1;
}
else if (Rs >= 1342) {
return 8;
}
}
}
}
else if (Rs >= 1418) {
if (refDCval < 86) {
if (RsCsDiff < 4) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 4;
}
}
else if (RsCsDiff >= 4) {
return 8;
}
}
else if (refDCval >= 86) {
return 4;
}
}
}
else if (MVDiff >= 8) {
if (Rs < 2033) {
return 1;
}
else if (Rs >= 2033) {
return 8;
}
}
}
else if (Rs >= 2221) {
if (negBalance < 15) {
return 8;
}
else if (negBalance >= 15) {
if (Cs < 1416) {
if (Rs < 2744) {
if (Rs < 2564) {
return 4;
}
else if (Rs >= 2564) {
return 4;
}
}
else if (Rs >= 2744) {
return 2;
}
}
else if (Cs >= 1416) {
if (SC < 4163) {
return 8;
}
else if (SC >= 4163) {
if (ssDCval < 78) {
if (Cs < 1589) {
return 4;
}
else if (Cs >= 1589) {
return 8;
}
}
else if (ssDCval >= 78) {
return 4;
}
}
}
}
}
}
else if (tcor >= 98) {
if (posBalance < 41) {
if (Cs < 1441) {
if (Cs < 1333) {
if (Rs < 1454) {
return 1;
}
else if (Rs >= 1454) {
if (Cs < 1329) {
if (Rs < 2242) {
return 8;
}
else if (Rs >= 2242) {
if (Rs < 2306) {
return 2;
}
else if (Rs >= 2306) {
return 8;
}
}
}
else if (Cs >= 1329) {
return 8;
}
}
}
else if (Cs >= 1333) {
if (negBalance < 29) {
if (Rs < 2343) {
return 8;
}
else if (Rs >= 2343) {
if (Rs < 2347) {
return 2;
}
else if (Rs >= 2347) {
return 8;
}
}
}
else if (negBalance >= 29) {
return 4;
}
}
}
else if (Cs >= 1441) {
if (posBalance < 26) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (posBalance >= 26) {
if (ssDCval < 61) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
else if (ssDCval >= 61) {
return 8;
}
}
}
}
else if (posBalance >= 41) {
if (negBalance < 22) {
if (Rs < 2586) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
if (Rs < 1985) {
if (Rs < 1511) {
return 8;
}
else if (Rs >= 1511) {
return 4;
}
}
else if (Rs >= 1985) {
return 8;
}
}
}
else if (Rs >= 2586) {
return 4;
}
}
else if (negBalance >= 22) {
if (ssDCval < 78) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
return 8;
}
}
else if (ssDCval >= 78) {
if (Rs < 2426) {
return 8;
}
else if (Rs >= 2426) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (diffAFD < -19) {
if (AFD < 1) {
if (diffMVdiffVal < -34) {
return 8;
}
else if (diffMVdiffVal >= -34) {
return 4;
}
}
else if (AFD >= 1) {
return 1;
}
}
else if (diffAFD >= -19) {
if (ssDCval < 117) {
if (ssDCval < 94) {
if (mu_mv_mag_sq < 120) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Rs < 1858) {
return 8;
}
else if (Rs >= 1858) {
if (Cs < 1358) {
return 8;
}
else if (Cs >= 1358) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 120) {
return 8;
}
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (ssDCval >= 117) {
if (CsDiff < 2) {
if (Rs < 1323) {
if (Cs < 1454) {
return 8;
}
else if (Cs >= 1454) {
return 4;
}
}
else if (Rs >= 1323) {
if (Rs < 2905) {
return 8;
}
else if (Rs >= 2905) {
if (Rs < 3038) {
return 2;
}
else if (Rs >= 3038) {
return 8;
}
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 120) {
if (negBalance < 5) {
return 1;
}
else if (negBalance >= 5) {
if (Rs < 2889) {
return 8;
}
else if (Rs >= 2889) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (ssDCval < 120) {
if (Rs < 2373) {
return 1;
}
else if (Rs >= 2373) {
return 8;
}
}
else if (ssDCval >= 120) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (diffRsCsdiff < -14) {
if (Rs < 2545) {
return 8;
}
else if (Rs >= 2545) {
return 2;
}
}
else if (diffRsCsdiff >= -14) {
if (ssDCval < 148) {
return 1;
}
else if (ssDCval >= 148) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3098) {
if (ssDCval < 77) {
if (mu_mv_mag_sq < 125) {
if (SC < 4919) {
if (AFD < 3) {
if (Cs < 1532) {
return 1;
}
else if (Cs >= 1532) {
return 4;
}
}
else if (AFD >= 3) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (SC >= 4919) {
if (ssDCval < 68) {
if (negBalance < 22) {
if (ssDCval < 61) {
if (Rs < 5467) {
if (diffAFD < -1) {
if (negBalance < 14) {
return 4;
}
else if (negBalance >= 14) {
return 8;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (Rs >= 5467) {
return 8;
}
}
else if (ssDCval >= 61) {
if (Rs < 3505) {
return 4;
}
else if (Rs >= 3505) {
return 1;
}
}
}
else if (negBalance >= 22) {
return 8;
}
}
else if (ssDCval >= 68) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
if (Rs < 3910) {
return 2;
}
else if (Rs >= 3910) {
if (Rs < 4355) {
return 4;
}
else if (Rs >= 4355) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1568) {
if (RsCsDiff < 12) {
if (Rs < 3678) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (Rs < 3530) {
return 1;
}
else if (Rs >= 3530) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (SC < 5015) {
return 8;
}
else if (SC >= 5015) {
return 4;
}
}
}
else if (Rs >= 3678) {
return 8;
}
}
else if (RsCsDiff >= 12) {
return 4;
}
}
else if (Cs >= 1568) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 4;
}
}
}
}
else if (ssDCval >= 77) {
if (ssDCval < 113) {
if (posBalance < 44) {
if (Rs < 4432) {
if (negBalance < 28) {
if (ssDCval < 86) {
if (Rs < 3918) {
return 2;
}
else if (Rs >= 3918) {
return 8;
}
}
else if (ssDCval >= 86) {
if (Cs < 1591) {
return 4;
}
else if (Cs >= 1591) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (Rs < 3373) {
return 8;
}
else if (Rs >= 3373) {
return 4;
}
}
}
}
}
else if (negBalance >= 28) {
if (Rs < 3982) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (Rs >= 3982) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 4;
}
}
}
}
else if (Rs >= 4432) {
if (Cs < 1557) {
if (Rs < 4507) {
return 8;
}
else if (Rs >= 4507) {
return 2;
}
}
else if (Cs >= 1557) {
return 8;
}
}
}
else if (posBalance >= 44) {
if (mu_mv_mag_sq < 217) {
if (Rs < 4011) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (Rs >= 4011) {
if (ssDCval < 101) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
if (Cs < 1444) {
return 2;
}
else if (Cs >= 1444) {
return 8;
}
}
}
else if (ssDCval >= 101) {
if (Rs < 4341) {
return 4;
}
else if (Rs >= 4341) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 217) {
return 4;
}
}
}
else if (ssDCval >= 113) {
if (diffRsCsdiff < -48) {
if (Rs < 3303) {
return 2;
}
else if (Rs >= 3303) {
if (Cs < 1551) {
return 4;
}
else if (Cs >= 1551) {
return 1;
}
}
}
else if (diffRsCsdiff >= -48) {
return 2;
}
}
}
}
}
else if (Cs >= 1661) {
if (ssDCval < 102) {
if (SC < 4285) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 1799) {
if (MVDiff < 2) {
if (ssDCval < 62) {
if (Rs < 1663) {
return 4;
}
else if (Rs >= 1663) {
return 8;
}
}
else if (ssDCval >= 62) {
if (Rs < 2383) {
if (Cs < 1747) {
return 8;
}
else if (Cs >= 1747) {
if (negBalance < 18) {
return 8;
}
else if (negBalance >= 18) {
if (AFD < 3) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
return 1;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (Rs >= 2383) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (Rs < 1027) {
return 8;
}
else if (Rs >= 1027) {
return 4;
}
}
}
else if (Cs >= 1799) {
if (diffRsCsdiff < -3) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
if (Rs < 1022) {
return 8;
}
else if (Rs >= 1022) {
return 1;
}
}
}
else if (diffRsCsdiff >= -3) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (posBalance < 36) {
if (Cs < 1829) {
return 8;
}
else if (Cs >= 1829) {
if (Rs < 1819) {
if (Rs < 1815) {
return 1;
}
else if (Rs >= 1815) {
return 8;
}
}
else if (Rs >= 1819) {
return 1;
}
}
}
else if (posBalance >= 36) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (AFD < 3) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
return 1;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 34) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (Cs < 2002) {
if (Rs < 2271) {
return 4;
}
else if (Rs >= 2271) {
return 8;
}
}
else if (Cs >= 2002) {
return 2;
}
}
}
else if (negBalance >= 34) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (Rs < 1170) {
if (MVDiff < 6) {
if (Cs < 1762) {
return 1;
}
else if (Cs >= 1762) {
return 8;
}
}
else if (MVDiff >= 6) {
if (Rs < 757) {
return 2;
}
else if (Rs >= 757) {
return 4;
}
}
}
else if (Rs >= 1170) {
if (gchDC < 3) {
if (Cs < 2390) {
if (Rs < 2470) {
return 8;
}
else if (Rs >= 2470) {
if (Cs < 1673) {
return 4;
}
else if (Cs >= 1673) {
if (Cs < 1737) {
return 8;
}
else if (Cs >= 1737) {
if (Rs < 2485) {
return 1;
}
else if (Rs >= 2485) {
return 8;
}
}
}
}
}
else if (Cs >= 2390) {
return 4;
}
}
else if (gchDC >= 3) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
return 4;
}
}
}
}
}
else if (SC >= 4285) {
if (posBalance < 3) {
if (diffRsCsdiff < -5) {
if (Cs < 1865) {
if (Cs < 1854) {
return 8;
}
else if (Cs >= 1854) {
return 4;
}
}
else if (Cs >= 1865) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
if (refDCval < 65) {
if (Rs < 6759) {
return 8;
}
else if (Rs >= 6759) {
return 2;
}
}
else if (refDCval >= 65) {
if (ssDCval < 87) {
if (Rs < 3477) {
return 8;
}
else if (Rs >= 3477) {
return 4;
}
}
else if (ssDCval >= 87) {
if (Rs < 2684) {
return 4;
}
else if (Rs >= 2684) {
return 8;
}
}
}
}
}
else if (posBalance >= 3) {
if (Rs < 6814) {
if (SC < 8416) {
if (Cs < 2259) {
if (Rs < 3108) {
if (Cs < 1934) {
if (posBalance < 23) {
return 8;
}
else if (posBalance >= 23) {
if (MVDiff < 1) {
if (RsDiff < 2) {
if (tcor < 99) {
if (Rs < 2939) {
return 4;
}
else if (Rs >= 2939) {
return 8;
}
}
else if (tcor >= 99) {
return 1;
}
}
else if (RsDiff >= 2) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (ssDCval < 74) {
if (SC < 4665) {
return 8;
}
else if (SC >= 4665) {
if (CsDiff < 3) {
if (SC < 4919) {
if (mu_mv_mag_sq < 126) {
return 4;
}
else if (mu_mv_mag_sq >= 126) {
if (gchDC < 1) {
if (Cs < 1809) {
return 8;
}
else if (Cs >= 1809) {
if (Rs < 3026) {
if (Rs < 2992) {
if (SC < 4824) {
return 4;
}
else if (SC >= 4824) {
if (Rs < 2945) {
return 8;
}
else if (Rs >= 2945) {
if (Rs < 2951) {
return 8;
}
else if (Rs >= 2951) {
return 8;
}
}
}
}
else if (Rs >= 2992) {
return 8;
}
}
else if (Rs >= 3026) {
if (Rs < 3049) {
return 4;
}
else if (Rs >= 3049) {
return 4;
}
}
}
}
else if (gchDC >= 1) {
if (diffRsCsdiff < -3) {
return 4;
}
else if (diffRsCsdiff >= -3) {
if (Rs < 3024) {
return 4;
}
else if (Rs >= 3024) {
return 4;
}
}
}
}
}
else if (SC >= 4919) {
return 4;
}
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
if (Cs < 1815) {
return 4;
}
else if (Cs >= 1815) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (Rs < 2975) {
return 8;
}
else if (Rs >= 2975) {
if (Rs < 3059) {
return 4;
}
else if (Rs >= 3059) {
return 4;
}
}
}
}
}
else if (Cs >= 1934) {
if (Rs < 2393) {
if (AFD < 2) {
if (refDCval < 92) {
if (Rs < 2370) {
return 1;
}
else if (Rs >= 2370) {
if (Rs < 2388) {
return 8;
}
else if (Rs >= 2388) {
return 2;
}
}
}
else if (refDCval >= 92) {
if (diffMVdiffVal < -1) {
if (posBalance < 21) {
return 4;
}
else if (posBalance >= 21) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (mu_mv_mag_sq < 110) {
if (Rs < 2363) {
return 4;
}
else if (Rs >= 2363) {
return 8;
}
}
else if (mu_mv_mag_sq >= 110) {
if (mu_mv_mag_sq < 118) {
return 2;
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 2312) {
return 2;
}
else if (Rs >= 2312) {
return 1;
}
}
}
}
}
}
else if (AFD >= 2) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 4;
}
}
}
else if (Rs >= 2393) {
if (ssDCval < 65) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 116) {
if (refDCval < 84) {
if (posBalance < 32) {
return 1;
}
else if (posBalance >= 32) {
if (posBalance < 53) {
if (Rs < 2985) {
return 8;
}
else if (Rs >= 2985) {
return 4;
}
}
else if (posBalance >= 53) {
if (SC < 4994) {
return 4;
}
else if (SC >= 4994) {
return 8;
}
}
}
}
else if (refDCval >= 84) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
if (Rs < 3062) {
return 8;
}
else if (Rs >= 3062) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 3082) {
return 2;
}
else if (Rs >= 3082) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 3108) {
if (negBalance < 24) {
if (Rs < 4975) {
if (refDCval < 62) {
if (Rs < 4452) {
if (RsCsDiff < 16) {
return 4;
}
else if (RsCsDiff >= 16) {
if (gchDC < 1) {
if (posBalance < 21) {
return 4;
}
else if (posBalance >= 21) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 4262) {
return 8;
}
else if (Rs >= 4262) {
if (AFD < 4) {
if (TSC < 3) {
if (Rs < 4374) {
if (Rs < 4334) {
return 8;
}
else if (Rs >= 4334) {
return 8;
}
}
else if (Rs >= 4374) {
return 8;
}
}
else if (TSC >= 3) {
return 4;
}
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
else if (gchDC >= 1) {
if (Rs < 3580) {
return 8;
}
else if (Rs >= 3580) {
if (mu_mv_mag_sq < 120) {
if (posBalance < 27) {
return 4;
}
else if (posBalance >= 27) {
if (posBalance < 77) {
if (negBalance < 12) {
if (Rs < 3963) {
return 4;
}
else if (Rs >= 3963) {
return 4;
}
}
else if (negBalance >= 12) {
return 4;
}
}
else if (posBalance >= 77) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 4222) {
return 4;
}
else if (Rs >= 4222) {
if (Rs < 4267) {
return 8;
}
else if (Rs >= 4267) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 4452) {
if (Rs < 4645) {
return 4;
}
else if (Rs >= 4645) {
return 8;
}
}
}
else if (refDCval >= 62) {
if (Cs < 1729) {
return 4;
}
else if (Cs >= 1729) {
if (ssDCval < 86) {
if (SC < 5119) {
return 4;
}
else if (SC >= 5119) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (ssDCval >= 86) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
return 2;
}
}
}
}
}
else if (Rs >= 4975) {
return 4;
}
}
else if (negBalance >= 24) {
if (ssDCval < 99) {
if (refDCval < 86) {
if (RsDiff < 6) {
if (RsDiff < 3) {
if (Rs < 4957) {
if (Rs < 3178) {
return 4;
}
else if (Rs >= 3178) {
return 8;
}
}
else if (Rs >= 4957) {
return 4;
}
}
else if (RsDiff >= 3) {
if (Cs < 2195) {
if (Cs < 1670) {
return 8;
}
else if (Cs >= 1670) {
return 4;
}
}
else if (Cs >= 2195) {
if (Rs < 4374) {
return 8;
}
else if (Rs >= 4374) {
return 4;
}
}
}
}
else if (RsDiff >= 6) {
if (mcTcor < 94) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
}
else if (refDCval >= 86) {
if (ssDCval < 91) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (ssDCval >= 91) {
if (MVDiff < 1) {
if (ssDCval < 95) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (ssDCval >= 95) {
if (Cs < 1808) {
return 4;
}
else if (Cs >= 1808) {
return 2;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 2140) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
if (ssDCval < 92) {
if (Rs < 4225) {
return 2;
}
else if (Rs >= 4225) {
return 8;
}
}
else if (ssDCval >= 92) {
return 2;
}
}
}
else if (Cs >= 2140) {
return 1;
}
}
}
}
}
else if (ssDCval >= 99) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (Rs < 4106) {
return 2;
}
else if (Rs >= 4106) {
return 1;
}
}
}
}
}
}
else if (Cs >= 2259) {
if (RsCsDiff < 69) {
if (SC < 7123) {
if (Rs < 2498) {
if (AFD < 4) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
return 4;
}
}
else if (AFD >= 4) {
if (ssDCval < 88) {
if (negBalance < 41) {
if (mu_mv_mag_sq < 108) {
if (Rs < 2393) {
return 4;
}
else if (Rs >= 2393) {
return 8;
}
}
else if (mu_mv_mag_sq >= 108) {
return 4;
}
}
else if (negBalance >= 41) {
if (refDCval < 73) {
if (Rs < 1854) {
return 4;
}
else if (Rs >= 1854) {
return 8;
}
}
else if (refDCval >= 73) {
return 4;
}
}
}
else if (ssDCval >= 88) {
if (Cs < 2658) {
return 8;
}
else if (Cs >= 2658) {
return 4;
}
}
}
}
else if (Rs >= 2498) {
if (ssDCval < 55) {
if (refDCval < 43) {
return 8;
}
else if (refDCval >= 43) {
if (Cs < 2665) {
if (Rs < 3330) {
return 8;
}
else if (Rs >= 3330) {
if (posBalance < 40) {
if (RsDiff < 6) {
if (tcor < 95) {
if (Rs < 4421) {
return 4;
}
else if (Rs >= 4421) {
if (Rs < 4527) {
return 8;
}
else if (Rs >= 4527) {
return 4;
}
}
}
else if (tcor >= 95) {
return 8;
}
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (posBalance >= 40) {
return 4;
}
}
}
else if (Cs >= 2665) {
return 1;
}
}
}
else if (ssDCval >= 55) {
if (diffRsCsdiff < -11) {
if (ssDCval < 87) {
if (Rs < 4256) {
return 8;
}
else if (Rs >= 4256) {
return 4;
}
}
else if (ssDCval >= 87) {
if (Cs < 2791) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (Cs >= 2791) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Cs < 3366) {
return 4;
}
else if (Cs >= 3366) {
return 8;
}
}
}
}
}
}
else if (diffRsCsdiff >= -11) {
if (SC < 6577) {
if (SC < 4859) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
else if (SC >= 4859) {
return 8;
}
}
else if (SC >= 6577) {
if (diffRsCsdiff < 27) {
if (mu_mv_mag_sq < 112) {
if (Rs < 3556) {
return 4;
}
else if (Rs >= 3556) {
return 8;
}
}
else if (mu_mv_mag_sq >= 112) {
if (mu_mv_mag_sq < 124) {
return 8;
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 2954) {
return 4;
}
else if (Rs >= 2954) {
if (Rs < 3462) {
return 8;
}
else if (Rs >= 3462) {
if (refDCval < 66) {
return 4;
}
else if (refDCval >= 66) {
if (Cs < 3395) {
return 8;
}
else if (Cs >= 3395) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
return 4;
}
}
}
}
}
}
}
}
else if (diffRsCsdiff >= 27) {
return 4;
}
}
}
}
}
}
else if (SC >= 7123) {
if (ssDCval < 87) {
if (negBalance < 5) {
return 4;
}
else if (negBalance >= 5) {
if (ssDCval < 56) {
if (diffRsCsdiff < -8) {
return 8;
}
else if (diffRsCsdiff >= -8) {
if (Cs < 2635) {
if (Cs < 2452) {
return 8;
}
else if (Cs >= 2452) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (mu_mv_mag_sq < 116) {
if (Rs < 4694) {
return 4;
}
else if (Rs >= 4694) {
if (Rs < 4978) {
if (Rs < 4768) {
return 4;
}
else if (Rs >= 4768) {
return 4;
}
}
else if (Rs >= 4978) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 116) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 2635) {
if (SC < 7909) {
if (RsDiff < 4) {
if (Cs < 2679) {
return 8;
}
else if (Cs >= 2679) {
if (ssDCval < 55) {
if (Rs < 4930) {
return 4;
}
else if (Rs >= 4930) {
return 4;
}
}
else if (ssDCval >= 55) {
return 4;
}
}
}
else if (RsDiff >= 4) {
if (Rs < 4847) {
return 4;
}
else if (Rs >= 4847) {
return 4;
}
}
}
else if (SC >= 7909) {
return 8;
}
}
}
}
else if (ssDCval >= 56) {
if (Cs < 2474) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
return 4;
}
}
else if (Cs >= 2474) {
return 8;
}
}
}
}
else if (ssDCval >= 87) {
if (Rs < 3258) {
return 8;
}
else if (Rs >= 3258) {
if (ssDCval < 98) {
if (mu_mv_mag_sq < 127) {
if (SC < 7205) {
if (Rs < 3962) {
return 8;
}
else if (Rs >= 3962) {
return 4;
}
}
else if (SC >= 7205) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
return 4;
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 69) {
if (Cs < 2546) {
if (ssDCval < 100) {
if (posBalance < 47) {
if (negBalance < 29) {
if (ssDCval < 55) {
if (Rs < 4607) {
return 8;
}
else if (Rs >= 4607) {
return 4;
}
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (negBalance >= 29) {
if (Rs < 3097) {
return 4;
}
else if (Rs >= 3097) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (negBalance < 39) {
if (Cs < 2382) {
return 8;
}
else if (Cs >= 2382) {
if (Rs < 3369) {
return 4;
}
else if (Rs >= 3369) {
return 2;
}
}
}
else if (negBalance >= 39) {
return 8;
}
}
}
else if (ssDCval >= 100) {
return 1;
}
}
else if (Cs >= 2546) {
if (Cs < 3228) {
if (ssDCval < 81) {
if (Rs < 2730) {
return 4;
}
else if (Rs >= 2730) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 8;
}
}
}
else if (ssDCval >= 81) {
if (Cs < 2727) {
if (Rs < 3108) {
return 8;
}
else if (Rs >= 3108) {
if (tcor < 90) {
if (Rs < 4353) {
return 1;
}
else if (Rs >= 4353) {
return 8;
}
}
else if (tcor >= 90) {
return 4;
}
}
}
else if (Cs >= 2727) {
if (Cs < 3176) {
return 4;
}
else if (Cs >= 3176) {
if (Rs < 3254) {
if (Cs < 3217) {
return 8;
}
else if (Cs >= 3217) {
return 4;
}
}
else if (Rs >= 3254) {
return 4;
}
}
}
}
}
else if (Cs >= 3228) {
if (Rs < 2683) {
return 4;
}
else if (Rs >= 2683) {
if (posBalance < 62) {
if (negBalance < 43) {
if (Cs < 3633) {
return 4;
}
else if (Cs >= 3633) {
return 8;
}
}
else if (negBalance >= 43) {
if (Cs < 3277) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (Cs < 3240) {
return 8;
}
else if (Cs >= 3240) {
return 4;
}
}
}
else if (Cs >= 3277) {
if (Cs < 4421) {
if (diffTSC < -2) {
if (Rs < 3470) {
return 4;
}
else if (Rs >= 3470) {
return 8;
}
}
else if (diffTSC >= -2) {
return 8;
}
}
else if (Cs >= 4421) {
if (ssDCval < 93) {
return 8;
}
else if (ssDCval >= 93) {
return 4;
}
}
}
}
}
else if (posBalance >= 62) {
if (Rs < 3539) {
return 4;
}
else if (Rs >= 3539) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 8416) {
if (MVDiff < 3) {
if (ssDCval < 75) {
if (mu_mv_mag_sq < 111) {
return 2;
}
else if (mu_mv_mag_sq >= 111) {
if (Rs < 4554) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 1;
}
}
}
else if (Rs >= 4554) {
if (Rs < 6017) {
return 8;
}
else if (Rs >= 6017) {
if (Rs < 6178) {
return 2;
}
else if (Rs >= 6178) {
return 8;
}
}
}
}
}
else if (ssDCval >= 75) {
if (Cs < 3666) {
if (Rs < 5353) {
return 4;
}
else if (Rs >= 5353) {
return 8;
}
}
else if (Cs >= 3666) {
if (mu_mv_mag_sq < 112) {
if (Rs < 5945) {
return 8;
}
else if (Rs >= 5945) {
return 4;
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (Rs < 5774) {
return 8;
}
else if (Rs >= 5774) {
if (ssDCval < 82) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 2;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
}
}
}
else if (Rs >= 6814) {
if (ssDCval < 95) {
return 4;
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
else if (ssDCval >= 102) {
if (tcor < 95) {
if (ssDCval < 137) {
if (refDCval < 105) {
if (negBalance < 42) {
if (Rs < 1467) {
return 2;
}
else if (Rs >= 1467) {
return 8;
}
}
else if (negBalance >= 42) {
if (Cs < 3134) {
if (Rs < 4923) {
return 1;
}
else if (Rs >= 4923) {
return 2;
}
}
else if (Cs >= 3134) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (AFD < 29) {
if (posBalance < 62) {
if (diffRsCsdiff < -46) {
if (mu_mv_mag_sq < 119) {
if (refDCval < 123) {
if (Rs < 1946) {
return 2;
}
else if (Rs >= 1946) {
return 8;
}
}
else if (refDCval >= 123) {
if (TSC < 8) {
if (Rs < 4161) {
return 1;
}
else if (Rs >= 4161) {
return 8;
}
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
else if (diffRsCsdiff >= -46) {
if (posBalance < 10) {
return 1;
}
else if (posBalance >= 10) {
if (diffMVdiffVal < -2) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 5349) {
return 2;
}
else if (Rs >= 5349) {
return 8;
}
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
}
else if (posBalance >= 62) {
if (Cs < 1894) {
if (Rs < 2420) {
return 8;
}
else if (Rs >= 2420) {
return 1;
}
}
else if (Cs >= 1894) {
if (Rs < 4588) {
if (Rs < 3265) {
return 8;
}
else if (Rs >= 3265) {
return 2;
}
}
else if (Rs >= 4588) {
return 8;
}
}
}
}
else if (AFD >= 29) {
if (Cs < 1798) {
return 1;
}
else if (Cs >= 1798) {
return 2;
}
}
}
}
else if (ssDCval >= 137) {
if (posBalance < 52) {
if (ssDCval < 200) {
if (SC < 4216) {
if (Cs < 1997) {
return 8;
}
else if (Cs >= 1997) {
return 1;
}
}
else if (SC >= 4216) {
return 8;
}
}
else if (ssDCval >= 200) {
return 4;
}
}
else if (posBalance >= 52) {
if (Cs < 4127) {
if (Rs < 2185) {
return 8;
}
else if (Rs >= 2185) {
return 1;
}
}
else if (Cs >= 4127) {
if (Rs < 7373) {
return 8;
}
else if (Rs >= 7373) {
return 4;
}
}
}
}
}
else if (tcor >= 95) {
if (Rs < 1889) {
if (posBalance < 53) {
if (Rs < 872) {
return 1;
}
else if (Rs >= 872) {
return 8;
}
}
else if (posBalance >= 53) {
if (Cs < 1957) {
if (tcor < 99) {
if (negBalance < 39) {
if (Rs < 1149) {
return 8;
}
else if (Rs >= 1149) {
if (Rs < 1285) {
return 4;
}
else if (Rs >= 1285) {
return 8;
}
}
}
else if (negBalance >= 39) {
return 8;
}
}
else if (tcor >= 99) {
if (ssDCval < 109) {
return 2;
}
else if (ssDCval >= 109) {
return 8;
}
}
}
else if (Cs >= 1957) {
return 2;
}
}
}
else if (Rs >= 1889) {
if (ssDCval < 117) {
if (TSC < 4) {
if (diffAFD < -26) {
return 4;
}
else if (diffAFD >= -26) {
if (diffRsCsdiff < -19) {
if (Cs < 2124) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
return 4;
}
}
else if (Cs >= 2124) {
return 8;
}
}
else if (diffRsCsdiff >= -19) {
return 8;
}
}
}
else if (TSC >= 4) {
if (refDCval < 105) {
if (Cs < 2826) {
if (Rs < 4291) {
if (Rs < 3700) {
return 8;
}
else if (Rs >= 3700) {
return 4;
}
}
else if (Rs >= 4291) {
if (Rs < 4835) {
return 1;
}
else if (Rs >= 4835) {
return 2;
}
}
}
else if (Cs >= 2826) {
return 8;
}
}
else if (refDCval >= 105) {
if (diffTSC < -2) {
if (ssDCval < 108) {
if (Cs < 2018) {
return 4;
}
else if (Cs >= 2018) {
if (Rs < 3804) {
return 8;
}
else if (Rs >= 3804) {
return 2;
}
}
}
else if (ssDCval >= 108) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
return 2;
}
}
}
else if (diffTSC >= -2) {
if (negBalance < 56) {
if (diffRsCsdiff < 19) {
return 8;
}
else if (diffRsCsdiff >= 19) {
if (RsCsDiff < 33) {
if (Rs < 3593) {
return 4;
}
else if (Rs >= 3593) {
return 8;
}
}
else if (RsCsDiff >= 33) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (Rs < 3040) {
return 8;
}
else if (Rs >= 3040) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 117) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 35) {
if (Rs < 2772) {
if (negBalance < 8) {
return 8;
}
else if (negBalance >= 8) {
if (MVDiff < 8) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 4;
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
else if (Rs >= 2772) {
if (ssDCval < 147) {
return 2;
}
else if (ssDCval >= 147) {
return 8;
}
}
}
else if (posBalance >= 35) {
if (Rs < 6428) {
if (mcTcor < 96) {
if (mu_mv_mag_sq < 122) {
return 8;
}
else if (mu_mv_mag_sq >= 122) {
return 1;
}
}
else if (mcTcor >= 96) {
if (diffAFD < 2) {
if (Cs < 1675) {
if (Rs < 4010) {
return 8;
}
else if (Rs >= 4010) {
return 2;
}
}
else if (Cs >= 1675) {
return 8;
}
}
else if (diffAFD >= 2) {
if (Rs < 2744) {
return 8;
}
else if (Rs >= 2744) {
return 2;
}
}
}
}
else if (Rs >= 6428) {
if (Cs < 4415) {
if (Rs < 7159) {
return 4;
}
else if (Rs >= 7159) {
return 2;
}
}
else if (Cs >= 4415) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 4837) {
if (Cs < 3936) {
if (AFD < 6) {
if (diffTSC < -6) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
if (Rs < 2386) {
return 8;
}
else if (Rs >= 2386) {
return 1;
}
}
}
else if (diffTSC >= -6) {
if (TSC < 5) {
if (Rs < 2636) {
if (Rs < 2634) {
if (posBalance < 1) {
if (Cs < 1890) {
return 1;
}
else if (Cs >= 1890) {
return 8;
}
}
else if (posBalance >= 1) {
return 8;
}
}
else if (Rs >= 2634) {
return 1;
}
}
else if (Rs >= 2636) {
return 8;
}
}
else if (TSC >= 5) {
if (Rs < 4500) {
return 8;
}
else if (Rs >= 4500) {
return 1;
}
}
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 3936) {
if (ssDCval < 137) {
return 1;
}
else if (ssDCval >= 137) {
if (Rs < 3630) {
return 4;
}
else if (Rs >= 3630) {
return 8;
}
}
}
}
else if (Rs >= 4837) {
return 8;
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelectC(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1313) {
if (Rs < 2496) {
if (Cs < 533) {
if (negBalance < 14) {
if (MVDiff < 7) {
if (SC < 790) {
if (ssDCval < 67) {
if (Cs < 1) {
if (ssDCval < 42) {
if (ssDCval < 22) {
if (ssDCval < 10) {
return 4;
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (ssDCval >= 22) {
if (Rs < 43) {
return 4;
}
else if (Rs >= 43) {
return 8;
}
}
}
else if (ssDCval >= 42) {
return 2;
}
}
else if (Cs >= 1) {
if (Cs < 256) {
if (ssDCval < 47) {
if (SC < 366) {
if (negBalance < 8) {
if (Rs < 4) {
return 4;
}
else if (Rs >= 4) {
if (Cs < 183) {
return 8;
}
else if (Cs >= 183) {
if (Rs < 114) {
return 1;
}
else if (Rs >= 114) {
if (Rs < 139) {
return 8;
}
else if (Rs >= 139) {
return 4;
}
}
}
}
}
else if (negBalance >= 8) {
if (Rs < 87) {
if (ssDCval < 39) {
if (ssDCval < 24) {
if (Rs < 61) {
return 8;
}
else if (Rs >= 61) {
return 1;
}
}
else if (ssDCval >= 24) {
if (posBalance < 31) {
return 4;
}
else if (posBalance >= 31) {
if (mu_mv_mag_sq < 105) {
if (Rs < 53) {
return 1;
}
else if (Rs >= 53) {
return 8;
}
}
else if (mu_mv_mag_sq >= 105) {
return 4;
}
}
}
}
else if (ssDCval >= 39) {
return 8;
}
}
else if (Rs >= 87) {
if (ssDCval < 38) {
if (ssDCval < 27) {
if (Cs < 87) {
return 8;
}
else if (Cs >= 87) {
return 1;
}
}
else if (ssDCval >= 27) {
if (ssDCval < 35) {
if (mcTcor < 94) {
if (Rs < 129) {
return 4;
}
else if (Rs >= 129) {
return 1;
}
}
else if (mcTcor >= 94) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (Rs < 110) {
return 8;
}
else if (Rs >= 110) {
return 1;
}
}
}
}
else if (ssDCval >= 35) {
return 4;
}
}
}
else if (ssDCval >= 38) {
return 1;
}
}
}
}
else if (SC >= 366) {
if (SC < 418) {
if (ssDCval < 30) {
return 1;
}
else if (ssDCval >= 30) {
return 4;
}
}
else if (SC >= 418) {
if (tcor < 90) {
if (SC < 459) {
return 8;
}
else if (SC >= 459) {
return 4;
}
}
else if (tcor >= 90) {
if (posBalance < 22) {
return 8;
}
else if (posBalance >= 22) {
if (Rs < 347) {
return 1;
}
else if (Rs >= 347) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 47) {
if (negBalance < 13) {
if (ssDCval < 61) {
if (ssDCval < 58) {
if (refDCval < 54) {
if (SC < 406) {
return 8;
}
else if (SC >= 406) {
return 4;
}
}
else if (refDCval >= 54) {
if (Cs < 120) {
return 8;
}
else if (Cs >= 120) {
if (Rs < 244) {
return 4;
}
else if (Rs >= 244) {
return 8;
}
}
}
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (ssDCval >= 61) {
if (Rs < 293) {
return 2;
}
else if (Rs >= 293) {
return 8;
}
}
}
else if (negBalance >= 13) {
if (AFD < 2) {
if (MVDiff < 3) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (Rs < 241) {
return 4;
}
else if (Rs >= 241) {
return 8;
}
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (Cs >= 256) {
if (Rs < 204) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
return 1;
}
}
}
else if (Rs >= 204) {
if (Rs < 393) {
if (TSCindex < 2) {
if (gchDC < 1) {
if (ssDCval < 23) {
return 8;
}
else if (ssDCval >= 23) {
if (Cs < 329) {
return 4;
}
else if (Cs >= 329) {
if (Rs < 275) {
return 1;
}
else if (Rs >= 275) {
return 4;
}
}
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (TSCindex >= 2) {
if (ssDCval < 40) {
return 4;
}
else if (ssDCval >= 40) {
if (Rs < 339) {
return 1;
}
else if (Rs >= 339) {
return 4;
}
}
}
}
else if (Rs >= 393) {
if (tcor < 97) {
if (Rs < 479) {
return 1;
}
else if (Rs >= 479) {
return 8;
}
}
else if (tcor >= 97) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (posBalance < 28) {
if (Rs < 184) {
return 1;
}
else if (Rs >= 184) {
if (Cs < 314) {
if (Rs < 304) {
return 8;
}
else if (Rs >= 304) {
if (Rs < 414) {
return 2;
}
else if (Rs >= 414) {
return 1;
}
}
}
else if (Cs >= 314) {
return 4;
}
}
}
else if (posBalance >= 28) {
if (ssDCval < 94) {
if (Cs < 130) {
if (Rs < 102) {
return 4;
}
else if (Rs >= 102) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
return 8;
}
}
}
else if (Cs >= 130) {
if (Rs < 294) {
if (Cs < 238) {
return 2;
}
else if (Cs >= 238) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
if (Rs < 209) {
return 4;
}
else if (Rs >= 209) {
return 1;
}
}
}
}
else if (Rs >= 294) {
if (Rs < 316) {
return 1;
}
else if (Rs >= 316) {
if (Rs < 355) {
return 2;
}
else if (Rs >= 355) {
if (Rs < 383) {
return 8;
}
else if (Rs >= 383) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 94) {
if (Rs < 427) {
return 4;
}
else if (Rs >= 427) {
return 1;
}
}
}
}
}
else if (SC >= 790) {
if (mu_mv_mag_sq < 128) {
if (Cs < 268) {
if (ssDCval < 39) {
if (Rs < 555) {
return 8;
}
else if (Rs >= 555) {
return 4;
}
}
else if (ssDCval >= 39) {
if (MVDiff < 5) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1060) {
return 8;
}
else if (Rs >= 1060) {
return 2;
}
}
}
}
else if (MVDiff >= 5) {
if (Rs < 1385) {
return 8;
}
else if (Rs >= 1385) {
if (Rs < 1890) {
return 1;
}
else if (Rs >= 1890) {
return 8;
}
}
}
}
}
else if (Cs >= 268) {
if (Rs < 765) {
if (CsDiff < 1) {
if (Rs < 348) {
return 1;
}
else if (Rs >= 348) {
if (Rs < 488) {
if (ssDCval < 109) {
return 8;
}
else if (ssDCval >= 109) {
return 1;
}
}
else if (Rs >= 488) {
if (ssDCval < 59) {
if (Rs < 497) {
return 1;
}
else if (Rs >= 497) {
return 8;
}
}
else if (ssDCval >= 59) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (posBalance < 38) {
if (Cs < 401) {
if (ssDCval < 152) {
return 8;
}
else if (ssDCval >= 152) {
return 1;
}
}
else if (Cs >= 401) {
if (Rs < 491) {
return 1;
}
else if (Rs >= 491) {
return 8;
}
}
}
else if (posBalance >= 38) {
if (Cs < 395) {
return 8;
}
else if (Cs >= 395) {
return 4;
}
}
}
}
}
}
}
else if (CsDiff >= 1) {
if (Rs < 514) {
return 1;
}
else if (Rs >= 514) {
if (Rs < 644) {
return 8;
}
else if (Rs >= 644) {
return 2;
}
}
}
}
else if (Rs >= 765) {
if (ssDCval < 67) {
if (ssDCval < 61) {
if (SC < 1810) {
if (gchDC < 1) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (gchDC >= 1) {
if (Rs < 810) {
return 8;
}
else if (Rs >= 810) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
if (Rs < 1191) {
return 4;
}
else if (Rs >= 1191) {
return 8;
}
}
}
}
}
else if (SC >= 1810) {
if (Rs < 1474) {
return 1;
}
else if (Rs >= 1474) {
return 8;
}
}
}
else if (ssDCval >= 61) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
if (Rs < 1471) {
return 2;
}
else if (Rs >= 1471) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (ssDCval < 79) {
if (Rs < 1454) {
return 8;
}
else if (Rs >= 1454) {
return 1;
}
}
else if (ssDCval >= 79) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Cs < 255) {
return 2;
}
else if (Cs >= 255) {
if (ssDCval < 136) {
if (ssDCval < 74) {
if (Cs < 356) {
return 1;
}
else if (Cs >= 356) {
if (Rs < 1186) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 4;
}
}
else if (Rs >= 1186) {
return 2;
}
}
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (tcor < 94) {
if (negBalance < 5) {
if (TSC < 6) {
if (Rs < 58) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 42) {
if (AFD < 3) {
if (ssDCval < 19) {
if (tcor < 57) {
return 8;
}
else if (tcor >= 57) {
return 2;
}
}
else if (ssDCval >= 19) {
return 2;
}
}
else if (AFD >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 42) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
if (refDCval < 15) {
return 4;
}
else if (refDCval >= 15) {
if (mu_mv_mag_sq < 104) {
return 8;
}
else if (mu_mv_mag_sq >= 104) {
if (ssDCval < 21) {
return 8;
}
else if (ssDCval >= 21) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 50) {
if (Cs < 9) {
return 2;
}
else if (Cs >= 9) {
return 1;
}
}
}
else if (Rs >= 58) {
if (diffMVdiffVal < -18) {
if (diffMVdiffVal < -38) {
if (Rs < 385) {
return 8;
}
else if (Rs >= 385) {
if (Rs < 1354) {
return 4;
}
else if (Rs >= 1354) {
return 8;
}
}
}
else if (diffMVdiffVal >= -38) {
return 4;
}
}
else if (diffMVdiffVal >= -18) {
if (Cs < 59) {
if (AFD < 5) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (Cs >= 59) {
if (Cs < 102) {
if (Cs < 85) {
return 8;
}
else if (Cs >= 85) {
return 4;
}
}
else if (Cs >= 102) {
return 8;
}
}
}
}
}
else if (TSC >= 6) {
if (diffRsCsdiff < 1) {
if (Cs < 2) {
return 1;
}
else if (Cs >= 2) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
else if (diffRsCsdiff >= 1) {
if (Cs < 318) {
return 1;
}
else if (Cs >= 318) {
if (AFD < 48) {
return 4;
}
else if (AFD >= 48) {
return 1;
}
}
}
}
}
else if (negBalance >= 5) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (ssDCval < 36) {
if (posBalance < 61) {
if (Cs < 95) {
return 1;
}
else if (Cs >= 95) {
if (tcor < 89) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Cs < 179) {
return 1;
}
else if (Cs >= 179) {
return 8;
}
}
}
else if (tcor >= 89) {
if (ssDCval < 24) {
return 1;
}
else if (ssDCval >= 24) {
if (MVDiff < 10) {
if (Rs < 126) {
return 8;
}
else if (Rs >= 126) {
return 1;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
}
}
}
else if (posBalance >= 61) {
if (Rs < 138) {
if (mu_mv_mag_sq < 40) {
if (ssDCval < 23) {
return 8;
}
else if (ssDCval >= 23) {
return 2;
}
}
else if (mu_mv_mag_sq >= 40) {
return 8;
}
}
else if (Rs >= 138) {
if (Rs < 363) {
return 1;
}
else if (Rs >= 363) {
return 4;
}
}
}
}
else if (ssDCval >= 36) {
if (gchDC < 11) {
if (MVDiff < 23) {
if (MVDiff < 13) {
if (posBalance < 96) {
return 8;
}
else if (posBalance >= 96) {
if (Rs < 1388) {
return 2;
}
else if (Rs >= 1388) {
return 4;
}
}
}
else if (MVDiff >= 13) {
if (gchDC < 4) {
return 1;
}
else if (gchDC >= 4) {
if (Cs < 341) {
return 8;
}
else if (Cs >= 341) {
return 1;
}
}
}
}
else if (MVDiff >= 23) {
if (posBalance < 107) {
if (negBalance < 7) {
if (Rs < 232) {
return 8;
}
else if (Rs >= 232) {
if (Rs < 326) {
return 2;
}
else if (Rs >= 326) {
return 1;
}
}
}
else if (negBalance >= 7) {
return 2;
}
}
else if (posBalance >= 107) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
if (Rs < 995) {
return 4;
}
else if (Rs >= 995) {
return 2;
}
}
}
}
}
else if (gchDC >= 11) {
if (ssDCval < 112) {
return 1;
}
else if (ssDCval >= 112) {
if (Rs < 1156) {
return 2;
}
else if (Rs >= 1156) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 94) {
if (negBalance < 6) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
if (negBalance < 4) {
return 8;
}
else if (negBalance >= 4) {
return 1;
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (ssDCval < 28) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (ssDCval >= 28) {
if (diffRsCsdiff < -2) {
if (Rs < 932) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (diffAFD < -8) {
if (posBalance < 3) {
if (Rs < 519) {
return 1;
}
else if (Rs >= 519) {
return 4;
}
}
else if (posBalance >= 3) {
if (ssDCval < 147) {
if (Rs < 873) {
return 2;
}
else if (Rs >= 873) {
return 8;
}
}
else if (ssDCval >= 147) {
if (Rs < 292) {
return 2;
}
else if (Rs >= 292) {
return 8;
}
}
}
}
else if (diffAFD >= -8) {
if (Cs < 271) {
if (Rs < 216) {
return 1;
}
else if (Rs >= 216) {
return 4;
}
}
else if (Cs >= 271) {
if (MVDiff < 41) {
if (Rs < 480) {
return 2;
}
else if (Rs >= 480) {
return 4;
}
}
else if (MVDiff >= 41) {
return 4;
}
}
}
}
}
else if (Rs >= 932) {
if (diffAFD < -8) {
return 2;
}
else if (diffAFD >= -8) {
if (Rs < 1578) {
return 1;
}
else if (Rs >= 1578) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -2) {
if (Cs < 86) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (AFD < 6) {
if (Rs < 47) {
return 2;
}
else if (Rs >= 47) {
return 4;
}
}
else if (AFD >= 6) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 54) {
return 1;
}
else if (Rs >= 54) {
return 2;
}
}
}
}
}
else if (Cs >= 86) {
if (ssDCval < 95) {
if (mu_mv_mag_sq < 107) {
if (ssDCval < 36) {
return 4;
}
else if (ssDCval >= 36) {
return 1;
}
}
else if (mu_mv_mag_sq >= 107) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 239) {
return 8;
}
else if (Rs >= 239) {
return 4;
}
}
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 6) {
if (SC < 941) {
if (Cs < 116) {
if (SC < 108) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
if (posBalance < 34) {
if (diffMVdiffVal < 11) {
return 8;
}
else if (diffMVdiffVal >= 11) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
return 4;
}
}
}
else if (posBalance >= 34) {
return 1;
}
}
}
else if (SC >= 108) {
if (ssDCval < 65) {
if (diffAFD < -1) {
if (Rs < 113) {
return 8;
}
else if (Rs >= 113) {
return 1;
}
}
else if (diffAFD >= -1) {
if (MVDiff < 11) {
if (Rs < 132) {
return 8;
}
else if (Rs >= 132) {
return 4;
}
}
else if (MVDiff >= 11) {
return 4;
}
}
}
else if (ssDCval >= 65) {
if (Cs < 33) {
return 8;
}
else if (Cs >= 33) {
return 1;
}
}
}
}
else if (Cs >= 116) {
if (ssDCval < 139) {
if (mu_mv_mag_sq < 94) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
if (ssDCval < 85) {
if (Rs < 337) {
return 1;
}
else if (Rs >= 337) {
return 8;
}
}
else if (ssDCval >= 85) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (ssDCval < 87) {
if (Rs < 214) {
if (Cs < 251) {
if (negBalance < 8) {
return 8;
}
else if (negBalance >= 8) {
if (Rs < 90) {
return 8;
}
else if (Rs >= 90) {
return 1;
}
}
}
else if (Cs >= 251) {
return 2;
}
}
else if (Rs >= 214) {
if (ssDCval < 36) {
return 4;
}
else if (ssDCval >= 36) {
if (Rs < 380) {
return 8;
}
else if (Rs >= 380) {
return 1;
}
}
}
}
else if (ssDCval >= 87) {
if (Rs < 194) {
return 4;
}
else if (Rs >= 194) {
return 8;
}
}
}
}
else if (ssDCval >= 139) {
if (ssDCval < 161) {
return 1;
}
else if (ssDCval >= 161) {
return 4;
}
}
}
}
else if (SC >= 941) {
if (Cs < 412) {
if (Cs < 42) {
return 8;
}
else if (Cs >= 42) {
return 1;
}
}
else if (Cs >= 412) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (mcTcor < 85) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
if (tcor < 59) {
if (negBalance < 45) {
if (Cs < 364) {
if (posBalance < 69) {
if (AFD < 6) {
if (Cs < 73) {
return 1;
}
else if (Cs >= 73) {
if (ssDCval < 25) {
if (Rs < 237) {
return 8;
}
else if (Rs >= 237) {
return 1;
}
}
else if (ssDCval >= 25) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 2;
}
}
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (posBalance >= 69) {
if (Rs < 241) {
return 8;
}
else if (Rs >= 241) {
return 1;
}
}
}
else if (Cs >= 364) {
if (Rs < 490) {
return 2;
}
else if (Rs >= 490) {
return 8;
}
}
}
else if (negBalance >= 45) {
if (MVDiff < 377) {
if (Rs < 963) {
return 8;
}
else if (Rs >= 963) {
return 1;
}
}
else if (MVDiff >= 377) {
return 4;
}
}
}
else if (tcor >= 59) {
if (refDCval < 23) {
if (Cs < 245) {
return 1;
}
else if (Cs >= 245) {
return 4;
}
}
else if (refDCval >= 23) {
if (AFD < 6) {
if (Cs < 208) {
if (MVDiff < 35) {
return 1;
}
else if (MVDiff >= 35) {
if (Rs < 37) {
return 4;
}
else if (Rs >= 37) {
if (Rs < 136) {
return 2;
}
else if (Rs >= 136) {
return 8;
}
}
}
}
else if (Cs >= 208) {
if (refDCval < 31) {
if (SC < 778) {
return 2;
}
else if (SC >= 778) {
if (SC < 806) {
return 8;
}
else if (SC >= 806) {
return 1;
}
}
}
else if (refDCval >= 31) {
return 8;
}
}
}
else if (AFD >= 6) {
if (MVDiff < 14) {
if (Rs < 436) {
return 8;
}
else if (Rs >= 436) {
return 1;
}
}
else if (MVDiff >= 14) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 85) {
if (posBalance < 27) {
if (MVDiff < 5) {
if (tcor < 102) {
if (Rs < 31) {
return 1;
}
else if (Rs >= 31) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
return 8;
}
}
}
else if (tcor >= 102) {
return 8;
}
}
else if (MVDiff >= 5) {
if (SC < 23) {
if (MVDiff < 28) {
return 4;
}
else if (MVDiff >= 28) {
if (TSC < 1) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (TSC >= 1) {
if (MVDiff < 219) {
return 8;
}
else if (MVDiff >= 219) {
if (MVDiff < 680) {
return 4;
}
else if (MVDiff >= 680) {
return 8;
}
}
}
}
}
else if (SC >= 23) {
if (ssDCval < 31) {
if (posBalance < 4) {
if (Rs < 43) {
return 8;
}
else if (Rs >= 43) {
return 4;
}
}
else if (posBalance >= 4) {
if (ssDCval < 29) {
if (negBalance < 61) {
return 1;
}
else if (negBalance >= 61) {
if (Rs < 93) {
return 8;
}
else if (Rs >= 93) {
return 1;
}
}
}
else if (ssDCval >= 29) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
return 1;
}
}
}
}
else if (ssDCval >= 31) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (Rs < 382) {
if (ssDCval < 32) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (ssDCval >= 32) {
return 1;
}
}
else if (Rs >= 382) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 27) {
if (refDCval < 30) {
if (Cs < 300) {
if (posBalance < 65) {
if (mu_mv_mag_sq < 40) {
if (negBalance < 29) {
if (AFD < 5) {
if (MVDiff < 11) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
if (posBalance < 42) {
if (Cs < 200) {
return 4;
}
else if (Cs >= 200) {
return 1;
}
}
else if (posBalance >= 42) {
return 1;
}
}
}
else if (MVDiff >= 11) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
return 4;
}
}
}
else if (AFD >= 5) {
if (MVDiff < 26) {
return 1;
}
else if (MVDiff >= 26) {
return 8;
}
}
}
else if (negBalance >= 29) {
return 1;
}
}
else if (mu_mv_mag_sq >= 40) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (Cs < 98) {
return 1;
}
else if (Cs >= 98) {
if (SC < 371) {
if (Rs < 175) {
return 1;
}
else if (Rs >= 175) {
return 8;
}
}
else if (SC >= 371) {
if (MVDiff < 41) {
return 1;
}
else if (MVDiff >= 41) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 65) {
if (Rs < 41) {
return 2;
}
else if (Rs >= 41) {
return 8;
}
}
}
else if (Cs >= 300) {
if (mu_mv_mag_sq < 64) {
if (Rs < 243) {
return 1;
}
else if (Rs >= 243) {
if (gchDC < 1) {
if (posBalance < 40) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (posBalance >= 40) {
return 1;
}
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 64) {
return 1;
}
}
}
else if (refDCval >= 30) {
if (posBalance < 41) {
if (Cs < 257) {
if (Rs < 360) {
if (AFD < 4) {
if (mcTcor < 97) {
if (posBalance < 28) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 28) {
if (ssDCval < 33) {
if (diffMVdiffVal < -14) {
return 1;
}
else if (diffMVdiffVal >= -14) {
if (diffMVdiffVal < 5) {
if (mu_mv_mag_sq < 99) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Rs < 194) {
return 8;
}
else if (Rs >= 194) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (Rs < 217) {
return 1;
}
else if (Rs >= 217) {
return 8;
}
}
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
}
else if (ssDCval >= 33) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 32) {
return 1;
}
else if (ssDCval >= 32) {
if (mu_mv_mag_sq < 78) {
if (Rs < 71) {
return 1;
}
else if (Rs >= 71) {
return 8;
}
}
else if (mu_mv_mag_sq >= 78) {
return 1;
}
}
}
}
else if (AFD >= 4) {
if (MVDiff < 9) {
if (AFD < 5) {
if (ssDCval < 34) {
return 4;
}
else if (ssDCval >= 34) {
return 8;
}
}
else if (AFD >= 5) {
if (Rs < 256) {
return 1;
}
else if (Rs >= 256) {
return 2;
}
}
}
else if (MVDiff >= 9) {
if (AFD < 6) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (ssDCval < 31) {
if (refDCval < 31) {
return 1;
}
else if (refDCval >= 31) {
if (Rs < 246) {
return 4;
}
else if (Rs >= 246) {
return 2;
}
}
}
else if (ssDCval >= 31) {
if (Rs < 281) {
return 4;
}
else if (Rs >= 281) {
return 1;
}
}
}
}
else if (AFD >= 6) {
return 1;
}
}
}
}
else if (Rs >= 360) {
if (tcor < 89) {
if (Rs < 682) {
return 1;
}
else if (Rs >= 682) {
return 4;
}
}
else if (tcor >= 89) {
return 4;
}
}
}
else if (Cs >= 257) {
if (negBalance < 15) {
if (Rs < 280) {
return 1;
}
else if (Rs >= 280) {
return 4;
}
}
else if (negBalance >= 15) {
return 1;
}
}
}
else if (posBalance >= 41) {
if (tcor < 62) {
if (mu_mv_mag_sq < 10) {
if (Cs < 240) {
return 8;
}
else if (Cs >= 240) {
return 1;
}
}
else if (mu_mv_mag_sq >= 10) {
if (Cs < 174) {
return 1;
}
else if (Cs >= 174) {
return 4;
}
}
}
else if (tcor >= 62) {
if (SC < 420) {
if (mcTcor < 88) {
if (Rs < 145) {
if (Cs < 27) {
return 4;
}
else if (Cs >= 27) {
return 1;
}
}
else if (Rs >= 145) {
return 2;
}
}
else if (mcTcor >= 88) {
if (gchDC < 3) {
if (posBalance < 72) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (Rs < 204) {
return 2;
}
else if (Rs >= 204) {
return 1;
}
}
}
else if (posBalance >= 72) {
return 2;
}
}
else if (gchDC >= 3) {
if (Rs < 135) {
return 1;
}
else if (Rs >= 135) {
return 4;
}
}
}
}
else if (SC >= 420) {
if (negBalance < 19) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
return 8;
}
}
else if (negBalance >= 19) {
if (ssDCval < 29) {
return 8;
}
else if (ssDCval >= 29) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 276) {
if (ssDCval < 51) {
if (mu_mv_mag_sq < 52) {
if (mcTcor < 93) {
if (Cs < 95) {
if (mu_mv_mag_sq < 21) {
if (ssDCval < 38) {
if (Cs < 83) {
if (AFD < 31) {
return 1;
}
else if (AFD >= 31) {
return 8;
}
}
else if (Cs >= 83) {
return 8;
}
}
else if (ssDCval >= 38) {
if (mu_mv_mag_sq < 5) {
if (Cs < 23) {
if (Rs < 33) {
return 4;
}
else if (Rs >= 33) {
return 2;
}
}
else if (Cs >= 23) {
return 1;
}
}
else if (mu_mv_mag_sq >= 5) {
if (MVDiff < 43) {
return 1;
}
else if (MVDiff >= 43) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (RsDiff < 1) {
if (MVDiff < 92) {
return 2;
}
else if (MVDiff >= 92) {
if (Rs < 19) {
return 8;
}
else if (Rs >= 19) {
return 1;
}
}
}
else if (RsDiff >= 1) {
if (Rs < 142) {
if (Rs < 59) {
return 1;
}
else if (Rs >= 59) {
if (Rs < 86) {
return 2;
}
else if (Rs >= 86) {
if (Rs < 123) {
return 4;
}
else if (Rs >= 123) {
return 8;
}
}
}
}
else if (Rs >= 142) {
return 1;
}
}
}
}
else if (Cs >= 95) {
if (posBalance < 54) {
if (MVDiff < 29) {
if (ssDCval < 37) {
if (Cs < 222) {
if (Rs < 139) {
return 8;
}
else if (Rs >= 139) {
return 2;
}
}
else if (Cs >= 222) {
return 8;
}
}
else if (ssDCval >= 37) {
if (Cs < 218) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
return 1;
}
}
else if (Cs >= 218) {
return 2;
}
}
}
else if (MVDiff >= 29) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
if (MVDiff < 52) {
if (diffRsCsdiff < 1) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
return 2;
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 245) {
return 1;
}
else if (Rs >= 245) {
return 8;
}
}
}
else if (MVDiff >= 52) {
if (posBalance < 40) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 1;
}
}
else if (posBalance >= 40) {
if (RsCsDiff < 14) {
return 1;
}
else if (RsCsDiff >= 14) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 54) {
if (AFD < 9) {
if (mcTcor < 89) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (Cs < 290) {
return 2;
}
else if (Cs >= 290) {
return 8;
}
}
}
else if (mcTcor >= 89) {
if (tcor < 82) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
return 8;
}
}
else if (tcor >= 82) {
return 8;
}
}
}
else if (AFD >= 9) {
if (negBalance < 47) {
if (AFD < 10) {
if (Rs < 137) {
return 4;
}
else if (Rs >= 137) {
return 2;
}
}
else if (AFD >= 10) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 164) {
return 4;
}
else if (Rs >= 164) {
return 8;
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 215) {
if (diffRsCsdiff < -2) {
if (Rs < 185) {
return 4;
}
else if (Rs >= 185) {
return 8;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (Rs >= 215) {
if (Cs < 195) {
return 8;
}
else if (Cs >= 195) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (negBalance < 75) {
if (tcor < 91) {
if (SC < 112) {
if (Rs < 32) {
if (Rs < 10) {
return 8;
}
else if (Rs >= 10) {
return 4;
}
}
else if (Rs >= 32) {
return 2;
}
}
else if (SC >= 112) {
if (TSC < 6) {
if (Rs < 86) {
if (Rs < 56) {
return 1;
}
else if (Rs >= 56) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
return 4;
}
}
}
else if (Rs >= 86) {
if (AFD < 6) {
if (posBalance < 53) {
if (Rs < 185) {
if (Rs < 102) {
return 4;
}
else if (Rs >= 102) {
return 1;
}
}
else if (Rs >= 185) {
if (MVDiff < 22) {
return 4;
}
else if (MVDiff >= 22) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (Rs < 234) {
return 2;
}
else if (Rs >= 234) {
return 1;
}
}
}
else if (AFD >= 6) {
if (diffAFD < 2) {
if (CsDiff < 4) {
if (posBalance < 35) {
if (Rs < 138) {
return 1;
}
else if (Rs >= 138) {
return 4;
}
}
else if (posBalance >= 35) {
return 1;
}
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (diffAFD >= 2) {
if (Cs < 152) {
return 2;
}
else if (Cs >= 152) {
if (Rs < 214) {
return 1;
}
else if (Rs >= 214) {
if (Rs < 232) {
return 4;
}
else if (Rs >= 232) {
return 1;
}
}
}
}
}
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (tcor >= 91) {
if (MVDiff < 34) {
if (mcTcor < 98) {
if (Cs < 140) {
return 1;
}
else if (Cs >= 140) {
if (negBalance < 36) {
if (Cs < 202) {
return 2;
}
else if (Cs >= 202) {
return 1;
}
}
else if (negBalance >= 36) {
return 2;
}
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (MVDiff >= 34) {
if (posBalance < 38) {
if (Cs < 315) {
return 2;
}
else if (Cs >= 315) {
return 4;
}
}
else if (posBalance >= 38) {
if (mu_mv_mag_sq < 35) {
if (MVDiff < 109) {
return 1;
}
else if (MVDiff >= 109) {
return 2;
}
}
else if (mu_mv_mag_sq >= 35) {
return 4;
}
}
}
}
}
else if (negBalance >= 75) {
if (Cs < 25) {
return 8;
}
else if (Cs >= 25) {
if (ssDCval < 40) {
return 4;
}
else if (ssDCval >= 40) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (SC < 547) {
if (negBalance < 27) {
if (mu_mv_mag_sq < 105) {
if (posBalance < 29) {
if (Rs < 99) {
if (ssDCval < 44) {
return 1;
}
else if (ssDCval >= 44) {
if (Rs < 13) {
return 2;
}
else if (Rs >= 13) {
return 8;
}
}
}
else if (Rs >= 99) {
if (Cs < 229) {
return 4;
}
else if (Cs >= 229) {
if (Rs < 202) {
return 1;
}
else if (Rs >= 202) {
if (Rs < 231) {
return 8;
}
else if (Rs >= 231) {
return 4;
}
}
}
}
}
else if (posBalance >= 29) {
if (SC < 493) {
if (refDCval < 47) {
if (mu_mv_mag_sq < 58) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 58) {
if (MVDiff < 2) {
if (Cs < 235) {
return 4;
}
else if (Cs >= 235) {
return 1;
}
}
else if (MVDiff >= 2) {
if (posBalance < 32) {
if (Rs < 157) {
return 1;
}
else if (Rs >= 157) {
return 4;
}
}
else if (posBalance >= 32) {
if (negBalance < 26) {
if (Cs < 269) {
return 1;
}
else if (Cs >= 269) {
if (Rs < 193) {
return 1;
}
else if (Rs >= 193) {
return 8;
}
}
}
else if (negBalance >= 26) {
if (Rs < 64) {
return 1;
}
else if (Rs >= 64) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 47) {
return 8;
}
}
else if (SC >= 493) {
if (Rs < 230) {
return 8;
}
else if (Rs >= 230) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Cs < 89) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 8;
}
}
else if (Cs >= 89) {
if (tcor < 94) {
if (TSC < 2) {
if (Rs < 249) {
return 1;
}
else if (Rs >= 249) {
return 4;
}
}
else if (TSC >= 2) {
if (Cs < 222) {
return 4;
}
else if (Cs >= 222) {
return 8;
}
}
}
else if (tcor >= 94) {
if (ssDCval < 35) {
if (Rs < 121) {
return 4;
}
else if (Rs >= 121) {
return 1;
}
}
else if (ssDCval >= 35) {
if (Cs < 192) {
if (ssDCval < 43) {
return 1;
}
else if (ssDCval >= 43) {
if (Rs < 222) {
return 8;
}
else if (Rs >= 222) {
return 4;
}
}
}
else if (Cs >= 192) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 27) {
if (refDCval < 40) {
if (posBalance < 49) {
if (Rs < 17) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
else if (Rs >= 17) {
if (tcor < 86) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 144) {
return 2;
}
else if (Rs >= 144) {
return 8;
}
}
}
else if (tcor >= 86) {
if (TSCindex < 3) {
if (Cs < 167) {
return 1;
}
else if (Cs >= 167) {
return 8;
}
}
else if (TSCindex >= 3) {
if (Rs < 240) {
return 8;
}
else if (Rs >= 240) {
return 4;
}
}
}
}
}
else if (posBalance >= 49) {
if (RsDiff < 2) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 155) {
return 8;
}
else if (Rs >= 155) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (Rs < 193) {
return 1;
}
else if (Rs >= 193) {
return 4;
}
}
}
}
else if (refDCval >= 40) {
if (AFD < 6) {
if (MVDiff < 3) {
if (refDCval < 50) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
return 4;
}
}
else if (refDCval >= 50) {
return 4;
}
}
else if (MVDiff >= 3) {
if (posBalance < 25) {
if (tcor < 98) {
if (Rs < 133) {
return 2;
}
else if (Rs >= 133) {
return 4;
}
}
else if (tcor >= 98) {
if (mcTcor < 105) {
return 8;
}
else if (mcTcor >= 105) {
return 4;
}
}
}
else if (posBalance >= 25) {
if (refDCval < 41) {
return 8;
}
else if (refDCval >= 41) {
if (mu_mv_mag_sq < 100) {
if (ssDCval < 47) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (ssDCval >= 47) {
return 8;
}
}
else if (mu_mv_mag_sq >= 100) {
return 8;
}
}
}
}
}
else if (AFD >= 6) {
if (posBalance < 35) {
if (mu_mv_mag_sq < 64) {
if (Rs < 95) {
return 4;
}
else if (Rs >= 95) {
return 8;
}
}
else if (mu_mv_mag_sq >= 64) {
return 1;
}
}
else if (posBalance >= 35) {
if (diffMVdiffVal < -44) {
if (Rs < 27) {
return 8;
}
else if (Rs >= 27) {
return 2;
}
}
else if (diffMVdiffVal >= -44) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
return 1;
}
}
}
}
}
}
}
}
else if (SC >= 547) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
if (Cs < 421) {
return 8;
}
else if (Cs >= 421) {
if (Rs < 262) {
return 1;
}
else if (Rs >= 262) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 51) {
if (refDCval < 70) {
if (TSCindex < 2) {
if (refDCval < 66) {
if (posBalance < 31) {
if (Cs < 40) {
return 2;
}
else if (Cs >= 40) {
return 8;
}
}
else if (posBalance >= 31) {
if (MVDiff < 7) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (mu_mv_mag_sq < 127) {
if (MVDiff < 3) {
if (ssDCval < 60) {
if (Cs < 374) {
if (Rs < 139) {
return 4;
}
else if (Rs >= 139) {
if (mu_mv_mag_sq < 111) {
return 1;
}
else if (mu_mv_mag_sq >= 111) {
if (Rs < 248) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (Rs >= 248) {
return 1;
}
}
}
}
else if (Cs >= 374) {
return 4;
}
}
else if (ssDCval >= 60) {
return 4;
}
}
else if (MVDiff >= 3) {
if (Rs < 85) {
return 1;
}
else if (Rs >= 85) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (Rs < 89) {
return 1;
}
else if (Rs >= 89) {
return 8;
}
}
}
}
else if (refDCval >= 66) {
if (SC < 415) {
return 2;
}
else if (SC >= 415) {
if (Cs < 378) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (Cs >= 378) {
return 1;
}
}
}
}
else if (TSCindex >= 2) {
if (TSC < 4) {
if (RsCsDiff < 1) {
if (Cs < 192) {
if (posBalance < 42) {
if (tcor < 98) {
if (Cs < 59) {
if (Rs < 39) {
return 1;
}
else if (Rs >= 39) {
return 2;
}
}
else if (Cs >= 59) {
if (diffMVdiffVal < 4) {
if (Rs < 184) {
return 1;
}
else if (Rs >= 184) {
return 4;
}
}
else if (diffMVdiffVal >= 4) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 22) {
return 4;
}
else if (Rs >= 22) {
return 1;
}
}
}
}
}
else if (tcor >= 98) {
if (refDCval < 61) {
if (SC < 285) {
if (Rs < 40) {
return 2;
}
else if (Rs >= 40) {
return 8;
}
}
else if (SC >= 285) {
if (Rs < 152) {
return 4;
}
else if (Rs >= 152) {
return 1;
}
}
}
else if (refDCval >= 61) {
if (TSCindex < 3) {
return 2;
}
else if (TSCindex >= 3) {
return 8;
}
}
}
}
else if (posBalance >= 42) {
if (MVDiff < 2) {
if (refDCval < 64) {
if (negBalance < 18) {
return 8;
}
else if (negBalance >= 18) {
return 1;
}
}
else if (refDCval >= 64) {
if (Rs < 94) {
return 1;
}
else if (Rs >= 94) {
return 2;
}
}
}
else if (MVDiff >= 2) {
if (tcor < 93) {
if (ssDCval < 59) {
if (mcTcor < 95) {
if (Cs < 89) {
return 2;
}
else if (Cs >= 89) {
return 1;
}
}
else if (mcTcor >= 95) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 4;
}
}
}
else if (ssDCval >= 59) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (AFD < 7) {
if (Cs < 18) {
return 1;
}
else if (Cs >= 18) {
return 4;
}
}
else if (AFD >= 7) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
}
else if (tcor >= 93) {
if (Rs < 209) {
if (negBalance < 22) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (Rs < 156) {
return 1;
}
else if (Rs >= 156) {
if (Rs < 160) {
return 2;
}
else if (Rs >= 160) {
return 4;
}
}
}
}
else if (negBalance >= 22) {
if (MVDiff < 6) {
if (Rs < 142) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (Rs >= 142) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
if (Cs < 154) {
if (Rs < 171) {
return 2;
}
else if (Rs >= 171) {
return 4;
}
}
else if (Cs >= 154) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
return 1;
}
}
}
else if (Rs >= 209) {
return 4;
}
}
}
}
}
else if (Cs >= 192) {
if (negBalance < 31) {
if (Rs < 169) {
if (negBalance < 29) {
if (mu_mv_mag_sq < 93) {
if (AFD < 3) {
if (mu_mv_mag_sq < 72) {
return 2;
}
else if (mu_mv_mag_sq >= 72) {
if (Rs < 145) {
return 1;
}
else if (Rs >= 145) {
return 4;
}
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (mu_mv_mag_sq >= 93) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Cs < 330) {
return 4;
}
else if (Cs >= 330) {
return 1;
}
}
}
}
else if (negBalance >= 29) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Rs < 159) {
return 1;
}
else if (Rs >= 159) {
return 4;
}
}
}
}
}
else if (Rs >= 169) {
if (MVDiff < 1) {
if (ssDCval < 61) {
if (Cs < 317) {
return 1;
}
else if (Cs >= 317) {
if (SC < 625) {
if (Rs < 246) {
return 4;
}
else if (Rs >= 246) {
return 1;
}
}
else if (SC >= 625) {
return 1;
}
}
}
else if (ssDCval >= 61) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 124) {
return 4;
}
else if (mu_mv_mag_sq >= 124) {
if (SC < 516) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (SC >= 516) {
if (Rs < 183) {
return 1;
}
else if (Rs >= 183) {
return 4;
}
}
}
}
else if (ssDCval >= 69) {
if (Cs < 349) {
return 4;
}
else if (Cs >= 349) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (refDCval < 63) {
if (Cs < 211) {
return 1;
}
else if (Cs >= 211) {
return 4;
}
}
else if (refDCval >= 63) {
if (negBalance < 28) {
if (Cs < 344) {
return 4;
}
else if (Cs >= 344) {
if (Rs < 197) {
return 1;
}
else if (Rs >= 197) {
if (Cs < 416) {
return 4;
}
else if (Cs >= 416) {
return 1;
}
}
}
}
else if (negBalance >= 28) {
if (SC < 528) {
return 1;
}
else if (SC >= 528) {
if (Rs < 210) {
return 1;
}
else if (Rs >= 210) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 31) {
if (AFD < 3) {
if (posBalance < 45) {
if (Rs < 200) {
if (Cs < 237) {
return 8;
}
else if (Cs >= 237) {
return 4;
}
}
else if (Rs >= 200) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (posBalance >= 45) {
if (tcor < 98) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
return 1;
}
}
else if (tcor >= 98) {
return 1;
}
}
}
else if (AFD >= 3) {
if (ssDCval < 63) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (SC < 386) {
if (ssDCval < 59) {
if (Rs < 113) {
return 2;
}
else if (Rs >= 113) {
return 4;
}
}
else if (ssDCval >= 59) {
return 1;
}
}
else if (SC >= 386) {
return 4;
}
}
}
else if (ssDCval >= 63) {
if (negBalance < 35) {
if (posBalance < 53) {
if (Cs < 253) {
return 1;
}
else if (Cs >= 253) {
return 4;
}
}
else if (posBalance >= 53) {
return 1;
}
}
else if (negBalance >= 35) {
if (Cs < 207) {
return 4;
}
else if (Cs >= 207) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (mcTcor < 91) {
if (refDCval < 60) {
if (diffRsCsdiff < -2) {
if (Rs < 160) {
return 2;
}
else if (Rs >= 160) {
if (Rs < 196) {
return 1;
}
else if (Rs >= 196) {
return 8;
}
}
}
else if (diffRsCsdiff >= -2) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (refDCval >= 60) {
return 1;
}
}
else if (mcTcor >= 91) {
if (Rs < 37) {
return 1;
}
else if (Rs >= 37) {
if (MVDiff < 4) {
if (Cs < 202) {
return 1;
}
else if (Cs >= 202) {
if (MVDiff < 2) {
if (Cs < 404) {
if (Rs < 179) {
if (Cs < 386) {
return 1;
}
else if (Cs >= 386) {
return 4;
}
}
else if (Rs >= 179) {
if (mu_mv_mag_sq < 115) {
if (Cs < 390) {
return 4;
}
else if (Cs >= 390) {
if (Rs < 208) {
return 1;
}
else if (Rs >= 208) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 115) {
return 1;
}
}
}
else if (Cs >= 404) {
return 1;
}
}
else if (MVDiff >= 2) {
if (Rs < 252) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (tcor < 105) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 1;
}
}
}
else if (tcor >= 105) {
return 8;
}
}
}
else if (Rs >= 252) {
return 1;
}
}
}
}
else if (MVDiff >= 4) {
if (TSCindex < 3) {
if (Cs < 318) {
if (Rs < 141) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (Rs >= 141) {
return 1;
}
}
else if (Cs >= 318) {
return 4;
}
}
else if (TSCindex >= 3) {
if (RsCsDiff < 5) {
if (Cs < 174) {
if (tcor < 93) {
if (Cs < 43) {
if (Rs < 93) {
return 1;
}
else if (Rs >= 93) {
return 2;
}
}
else if (Cs >= 43) {
if (diffRsCsdiff < 1) {
return 4;
}
else if (diffRsCsdiff >= 1) {
if (mu_mv_mag_sq < 12) {
if (Rs < 193) {
return 1;
}
else if (Rs >= 193) {
return 4;
}
}
else if (mu_mv_mag_sq >= 12) {
return 4;
}
}
}
}
else if (tcor >= 93) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
return 8;
}
}
}
}
else if (Cs >= 174) {
if (posBalance < 56) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
if (ssDCval < 55) {
if (Rs < 125) {
return 2;
}
else if (Rs >= 125) {
return 4;
}
}
else if (ssDCval >= 55) {
if (diffMVdiffVal < -5) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
return 4;
}
}
}
}
else if (posBalance >= 56) {
if (tcor < 96) {
if (Rs < 89) {
return 1;
}
else if (Rs >= 89) {
return 4;
}
}
else if (tcor >= 96) {
return 1;
}
}
}
}
else if (RsCsDiff >= 5) {
if (AFD < 9) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
return 4;
}
}
else if (AFD >= 9) {
if (tcor < 70) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
return 4;
}
}
else if (tcor >= 70) {
return 1;
}
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 43) {
if (Cs < 321) {
if (MVDiff < 107) {
if (MVDiff < 42) {
if (mu_mv_mag_sq < 32) {
return 1;
}
else if (mu_mv_mag_sq >= 32) {
if (Rs < 157) {
return 1;
}
else if (Rs >= 157) {
return 2;
}
}
}
else if (MVDiff >= 42) {
if (mu_mv_mag_sq < 14) {
if (tcor < 61) {
if (MVDiff < 64) {
if (Rs < 177) {
return 4;
}
else if (Rs >= 177) {
return 1;
}
}
else if (MVDiff >= 64) {
return 2;
}
}
else if (tcor >= 61) {
return 2;
}
}
else if (mu_mv_mag_sq >= 14) {
if (CsDiff < 1) {
if (Cs < 67) {
return 1;
}
else if (Cs >= 67) {
return 8;
}
}
else if (CsDiff >= 1) {
if (CsDiff < 2) {
if (MVDiff < 46) {
return 4;
}
else if (MVDiff >= 46) {
if (Rs < 256) {
return 2;
}
else if (Rs >= 256) {
return 4;
}
}
}
else if (CsDiff >= 2) {
if (Rs < 144) {
return 1;
}
else if (Rs >= 144) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 107) {
if (refDCval < 43) {
return 4;
}
else if (refDCval >= 43) {
if (diffRsCsdiff < -10) {
return 8;
}
else if (diffRsCsdiff >= -10) {
if (AFD < 20) {
if (MVDiff < 139) {
return 1;
}
else if (MVDiff >= 139) {
if (gchDC < 5) {
if (Rs < 255) {
return 1;
}
else if (Rs >= 255) {
return 2;
}
}
else if (gchDC >= 5) {
return 2;
}
}
}
else if (AFD >= 20) {
if (negBalance < 23) {
return 8;
}
else if (negBalance >= 23) {
return 1;
}
}
}
}
}
}
else if (Cs >= 321) {
if (refDCval < 65) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (diffRsCsdiff < 8) {
return 1;
}
else if (diffRsCsdiff >= 8) {
if (Rs < 243) {
return 8;
}
else if (Rs >= 243) {
return 4;
}
}
}
}
else if (refDCval >= 65) {
if (Cs < 461) {
return 4;
}
else if (Cs >= 461) {
if (Rs < 210) {
return 2;
}
else if (Rs >= 210) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (diffTSC < 6) {
if (RsCsDiff < 2) {
if (negBalance < 41) {
if (refDCval < 52) {
return 8;
}
else if (refDCval >= 52) {
return 4;
}
}
else if (negBalance >= 41) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
return 8;
}
}
}
else if (RsCsDiff >= 2) {
if (posBalance < 54) {
if (diffTSC < -1) {
if (SC < 556) {
return 2;
}
else if (SC >= 556) {
return 8;
}
}
else if (diffTSC >= -1) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Cs < 323) {
return 4;
}
else if (Cs >= 323) {
return 8;
}
}
}
}
}
}
else if (diffTSC >= 6) {
if (Rs < 194) {
return 1;
}
else if (Rs >= 194) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 102) {
if (negBalance < 20) {
if (ssDCval < 81) {
if (Cs < 260) {
if (Rs < 86) {
return 1;
}
else if (Rs >= 86) {
if (Cs < 112) {
return 1;
}
else if (Cs >= 112) {
return 2;
}
}
}
else if (Cs >= 260) {
return 1;
}
}
else if (ssDCval >= 81) {
if (MVDiff < 26) {
if (Rs < 161) {
return 4;
}
else if (Rs >= 161) {
if (diffMVdiffVal < -2) {
if (Rs < 223) {
return 2;
}
else if (Rs >= 223) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 269) {
return 8;
}
else if (Cs >= 269) {
return 4;
}
}
}
}
else if (MVDiff >= 26) {
if (ssDCval < 98) {
if (Rs < 40) {
return 2;
}
else if (Rs >= 40) {
return 8;
}
}
else if (ssDCval >= 98) {
return 1;
}
}
}
}
else if (negBalance >= 20) {
if (mcTcor < 97) {
if (Rs < 77) {
if (ssDCval < 131) {
if (mu_mv_mag_sq < 53) {
if (mcTcor < 63) {
return 4;
}
else if (mcTcor >= 63) {
if (Cs < 149) {
if (TSC < 5) {
if (ssDCval < 111) {
return 1;
}
else if (ssDCval >= 111) {
if (Rs < 18) {
return 1;
}
else if (Rs >= 18) {
return 2;
}
}
}
else if (TSC >= 5) {
if (posBalance < 55) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 2;
}
}
else if (posBalance >= 55) {
return 2;
}
}
}
else if (Cs >= 149) {
if (negBalance < 53) {
return 2;
}
else if (negBalance >= 53) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (Rs < 62) {
return 1;
}
else if (Rs >= 62) {
return 8;
}
}
}
}
else if (ssDCval >= 131) {
if (SC < 3) {
return 1;
}
else if (SC >= 3) {
if (MVDiff < 522) {
return 2;
}
else if (MVDiff >= 522) {
return 1;
}
}
}
}
else if (Rs >= 77) {
if (mcTcor < 72) {
if (diffRsCsdiff < 6) {
if (Rs < 93) {
return 8;
}
else if (Rs >= 93) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
if (Rs < 215) {
return 8;
}
else if (Rs >= 215) {
return 1;
}
}
}
else if (mcTcor >= 72) {
if (mu_mv_mag_sq < 37) {
if (Cs < 45) {
if (Rs < 173) {
if (Cs < 31) {
if (ssDCval < 77) {
return 1;
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (Cs >= 31) {
return 2;
}
}
else if (Rs >= 173) {
if (diffMVdiffVal < -78) {
return 8;
}
else if (diffMVdiffVal >= -78) {
return 1;
}
}
}
else if (Cs >= 45) {
if (refDCval < 86) {
if (AFD < 10) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (RsDiff < 1) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 4;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (AFD >= 10) {
if (gchDC < 6) {
if (RsCsDiff < 5) {
if (MVDiff < 145) {
return 2;
}
else if (MVDiff >= 145) {
return 1;
}
}
else if (RsCsDiff >= 5) {
if (tcor < 32) {
return 2;
}
else if (tcor >= 32) {
if (AFD < 21) {
if (Cs < 363) {
return 2;
}
else if (Cs >= 363) {
return 1;
}
}
else if (AFD >= 21) {
return 1;
}
}
}
}
else if (gchDC >= 6) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
if (Rs < 86) {
return 2;
}
else if (Rs >= 86) {
if (Cs < 316) {
return 8;
}
else if (Cs >= 316) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 86) {
if (ssDCval < 165) {
if (refDCval < 143) {
if (ssDCval < 89) {
if (diffMVdiffVal < -2) {
if (Cs < 130) {
return 8;
}
else if (Cs >= 130) {
if (Cs < 321) {
return 2;
}
else if (Cs >= 321) {
return 8;
}
}
}
else if (diffMVdiffVal >= -2) {
if (MVDiff < 45) {
if (RsCsDiff < 1) {
if (Cs < 279) {
return 2;
}
else if (Cs >= 279) {
if (tcor < 83) {
return 8;
}
else if (tcor >= 83) {
return 2;
}
}
}
else if (RsCsDiff >= 1) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (diffRsCsdiff < 1) {
if (Rs < 260) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
if (Rs < 160) {
return 8;
}
else if (Rs >= 160) {
return 2;
}
}
}
else if (Rs >= 260) {
return 2;
}
}
else if (diffRsCsdiff >= 1) {
return 2;
}
}
}
}
else if (MVDiff >= 45) {
return 2;
}
}
}
else if (ssDCval >= 89) {
if (refDCval < 114) {
if (MVDiff < 16) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (MVDiff >= 16) {
if (MVDiff < 46) {
return 1;
}
else if (MVDiff >= 46) {
if (Rs < 175) {
return 2;
}
else if (Rs >= 175) {
return 1;
}
}
}
}
else if (refDCval >= 114) {
if (Cs < 250) {
if (gchDC < 8) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
else if (gchDC >= 8) {
return 8;
}
}
else if (Cs >= 250) {
return 8;
}
}
}
}
else if (refDCval >= 143) {
return 2;
}
}
else if (ssDCval >= 165) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (ssDCval < 83) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (negBalance < 71) {
if (Cs < 107) {
if (Rs < 142) {
return 4;
}
else if (Rs >= 142) {
return 8;
}
}
else if (Cs >= 107) {
return 1;
}
}
else if (negBalance >= 71) {
return 8;
}
}
}
else if (ssDCval >= 83) {
if (negBalance < 45) {
if (ssDCval < 116) {
return 1;
}
else if (ssDCval >= 116) {
if (Rs < 216) {
return 2;
}
else if (Rs >= 216) {
return 8;
}
}
}
else if (negBalance >= 45) {
if (AFD < 17) {
if (Cs < 361) {
if (AFD < 7) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (Rs < 226) {
return 2;
}
else if (Rs >= 226) {
return 8;
}
}
}
else if (AFD >= 7) {
if (tcor < 36) {
if (diffMVdiffVal < -16) {
return 2;
}
else if (diffMVdiffVal >= -16) {
if (MVDiff < 30) {
return 8;
}
else if (MVDiff >= 30) {
if (MVDiff < 47) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 2;
}
}
}
else if (MVDiff >= 47) {
return 8;
}
}
}
}
else if (tcor >= 36) {
return 8;
}
}
}
else if (Cs >= 361) {
return 8;
}
}
else if (AFD >= 17) {
if (AFD < 22) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (AFD >= 22) {
return 1;
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 113) {
if (Cs < 250) {
if (Rs < 85) {
if (ssDCval < 97) {
if (RsCsDiff < 1) {
if (MVDiff < 38) {
if (mu_mv_mag_sq < 74) {
return 1;
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 120) {
return 1;
}
else if (Cs >= 120) {
return 2;
}
}
}
else if (MVDiff >= 38) {
if (Rs < 13) {
return 2;
}
else if (Rs >= 13) {
if (Rs < 30) {
return 1;
}
else if (Rs >= 30) {
return 4;
}
}
}
}
else if (RsCsDiff >= 1) {
if (RsCsDiff < 2) {
return 4;
}
else if (RsCsDiff >= 2) {
return 1;
}
}
}
else if (ssDCval >= 97) {
if (negBalance < 31) {
if (Cs < 69) {
return 1;
}
else if (Cs >= 69) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 8;
}
}
}
else if (negBalance >= 31) {
if (Rs < 5) {
return 2;
}
else if (Rs >= 5) {
if (mu_mv_mag_sq < 5) {
if (TSC < 3) {
if (Rs < 20) {
return 1;
}
else if (Rs >= 20) {
return 4;
}
}
else if (TSC >= 3) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
if (Rs < 61) {
return 4;
}
else if (Rs >= 61) {
if (Rs < 83) {
return 1;
}
else if (Rs >= 83) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 85) {
if (AFD < 2) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Cs < 205) {
if (SC < 222) {
return 2;
}
else if (SC >= 222) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (posBalance < 40) {
if (SC < 260) {
if (Rs < 110) {
return 4;
}
else if (Rs >= 110) {
return 2;
}
}
else if (SC >= 260) {
return 1;
}
}
else if (posBalance >= 40) {
return 1;
}
}
}
}
else if (Cs >= 205) {
return 2;
}
}
}
else if (AFD >= 2) {
if (Rs < 134) {
if (Rs < 103) {
if (ssDCval < 104) {
if (SC < 310) {
if (ssDCval < 88) {
if (posBalance < 55) {
if (tcor < 96) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
return 4;
}
}
else if (tcor >= 96) {
if (tcor < 99) {
if (Cs < 80) {
return 1;
}
else if (Cs >= 80) {
return 2;
}
}
else if (tcor >= 99) {
return 1;
}
}
}
else if (posBalance >= 55) {
return 2;
}
}
else if (ssDCval >= 88) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (Rs < 96) {
return 1;
}
else if (Rs >= 96) {
return 2;
}
}
}
}
else if (SC >= 310) {
return 4;
}
}
else if (ssDCval >= 104) {
return 2;
}
}
else if (Rs >= 103) {
if (posBalance < 62) {
if (posBalance < 48) {
if (tcor < 99) {
if (mu_mv_mag_sq < 22) {
return 1;
}
else if (mu_mv_mag_sq >= 22) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (SC < 270) {
return 2;
}
else if (SC >= 270) {
if (MVDiff < 27) {
return 1;
}
else if (MVDiff >= 27) {
return 2;
}
}
}
}
}
else if (tcor >= 99) {
return 1;
}
}
else if (posBalance >= 48) {
if (ssDCval < 74) {
if (Cs < 135) {
return 2;
}
else if (Cs >= 135) {
return 1;
}
}
else if (ssDCval >= 74) {
return 1;
}
}
}
else if (posBalance >= 62) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (Cs < 179) {
if (refDCval < 98) {
return 1;
}
else if (refDCval >= 98) {
return 2;
}
}
else if (Cs >= 179) {
return 2;
}
}
}
}
}
else if (Rs >= 134) {
if (refDCval < 91) {
if (Cs < 146) {
if (TSC < 4) {
if (tcor < 93) {
if (Cs < 50) {
return 2;
}
else if (Cs >= 50) {
return 1;
}
}
else if (tcor >= 93) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
if (Rs < 145) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 8;
}
}
else if (Rs >= 145) {
return 8;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 12) {
return 2;
}
else if (Cs >= 12) {
return 1;
}
}
}
else if (Cs >= 146) {
if (MVDiff < 4) {
if (Rs < 229) {
if (Rs < 152) {
return 4;
}
else if (Rs >= 152) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (Rs >= 229) {
if (TSC < 3) {
if (AFD < 3) {
if (Cs < 200) {
return 8;
}
else if (Cs >= 200) {
return 1;
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (TSC >= 3) {
if (SC < 490) {
return 2;
}
else if (SC >= 490) {
return 1;
}
}
}
}
else if (MVDiff >= 4) {
if (AFD < 6) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
if (Rs < 197) {
return 4;
}
else if (Rs >= 197) {
return 1;
}
}
}
else if (AFD >= 6) {
if (Rs < 233) {
if (Rs < 189) {
return 1;
}
else if (Rs >= 189) {
return 8;
}
}
else if (Rs >= 233) {
return 2;
}
}
}
}
}
else if (refDCval >= 91) {
if (MVDiff < 6) {
if (Rs < 202) {
return 2;
}
else if (Rs >= 202) {
return 4;
}
}
else if (MVDiff >= 6) {
if (Cs < 196) {
if (posBalance < 51) {
if (Cs < 145) {
return 1;
}
else if (Cs >= 145) {
return 2;
}
}
else if (posBalance >= 51) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (Rs < 191) {
return 4;
}
else if (Rs >= 191) {
return 2;
}
}
}
}
else if (Cs >= 196) {
if (Cs < 241) {
return 8;
}
else if (Cs >= 241) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 250) {
if (ssDCval < 76) {
if (Rs < 130) {
if (refDCval < 72) {
return 4;
}
else if (refDCval >= 72) {
if (Rs < 109) {
return 4;
}
else if (Rs >= 109) {
return 1;
}
}
}
else if (Rs >= 130) {
if (negBalance < 41) {
if (mu_mv_mag_sq < 99) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Rs < 181) {
return 1;
}
else if (Rs >= 181) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 99) {
return 4;
}
}
else if (negBalance >= 41) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
else if (ssDCval >= 76) {
if (TSC < 3) {
if (SC < 622) {
if (mu_mv_mag_sq < 77) {
if (SC < 531) {
return 4;
}
else if (SC >= 531) {
return 1;
}
}
else if (mu_mv_mag_sq >= 77) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
if (Cs < 398) {
return 1;
}
else if (Cs >= 398) {
return 4;
}
}
}
}
else if (SC >= 622) {
if (SC < 713) {
return 1;
}
else if (SC >= 713) {
return 2;
}
}
}
else if (TSC >= 3) {
if (RsDiff < 4) {
if (mu_mv_mag_sq < 5) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 405) {
return 4;
}
else if (Cs >= 405) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (mcTcor < 98) {
if (Rs < 135) {
return 1;
}
else if (Rs >= 135) {
if (Cs < 302) {
return 1;
}
else if (Cs >= 302) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (tcor < 90) {
if (diffMVdiffVal < 20) {
if (Cs < 437) {
return 1;
}
else if (Cs >= 437) {
return 2;
}
}
else if (diffMVdiffVal >= 20) {
return 2;
}
}
else if (tcor >= 90) {
if (tcor < 98) {
return 2;
}
else if (tcor >= 98) {
if (Rs < 187) {
return 1;
}
else if (Rs >= 187) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 4) {
return 8;
}
}
}
}
}
else if (ssDCval >= 113) {
if (AFD < 3) {
if (ssDCval < 136) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 2;
}
}
else if (ssDCval >= 136) {
return 2;
}
}
else if (AFD >= 3) {
if (tcor < 100) {
if (AFD < 11) {
if (mcTcor < 100) {
return 2;
}
else if (mcTcor >= 100) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
return 1;
}
}
}
else if (AFD >= 11) {
if (SC < 206) {
if (SC < 191) {
return 8;
}
else if (SC >= 191) {
return 1;
}
}
else if (SC >= 206) {
if (Rs < 150) {
return 2;
}
else if (Rs >= 150) {
return 4;
}
}
}
}
else if (tcor >= 100) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (gchDC < 7) {
return 1;
}
else if (gchDC >= 7) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 130) {
if (refDCval < 75) {
if (Rs < 115) {
return 4;
}
else if (Rs >= 115) {
return 8;
}
}
else if (refDCval >= 75) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (negBalance < 60) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (negBalance >= 60) {
if (AFD < 46) {
return 1;
}
else if (AFD >= 46) {
return 4;
}
}
}
}
}
else if (Rs >= 130) {
if (Cs < 267) {
if (Cs < 122) {
if (mu_mv_mag_sq < 117) {
if (Rs < 242) {
return 8;
}
else if (Rs >= 242) {
return 1;
}
}
else if (mu_mv_mag_sq >= 117) {
if (ssDCval < 91) {
if (Cs < 59) {
return 8;
}
else if (Cs >= 59) {
if (Rs < 137) {
return 8;
}
else if (Rs >= 137) {
return 2;
}
}
}
else if (ssDCval >= 91) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 2;
}
}
}
}
else if (Cs >= 122) {
if (negBalance < 22) {
if (ssDCval < 80) {
if (diffAFD < -1) {
if (refDCval < 74) {
return 2;
}
else if (refDCval >= 74) {
return 8;
}
}
else if (diffAFD >= -1) {
if (Cs < 215) {
return 2;
}
else if (Cs >= 215) {
if (Rs < 207) {
return 1;
}
else if (Rs >= 207) {
return 2;
}
}
}
}
else if (ssDCval >= 80) {
if (Cs < 141) {
return 8;
}
else if (Cs >= 141) {
if (Rs < 232) {
return 2;
}
else if (Rs >= 232) {
return 8;
}
}
}
}
else if (negBalance >= 22) {
if (Rs < 178) {
if (MVDiff < 3) {
if (Cs < 133) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (Cs >= 133) {
return 2;
}
}
else if (MVDiff >= 3) {
if (Rs < 159) {
return 1;
}
else if (Rs >= 159) {
if (Rs < 170) {
return 2;
}
else if (Rs >= 170) {
return 1;
}
}
}
}
else if (Rs >= 178) {
if (ssDCval < 74) {
if (Cs < 208) {
if (negBalance < 27) {
if (Cs < 161) {
return 1;
}
else if (Cs >= 161) {
return 2;
}
}
else if (negBalance >= 27) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
return 1;
}
}
}
else if (Cs >= 208) {
return 1;
}
}
else if (ssDCval >= 74) {
return 2;
}
}
}
}
}
else if (Cs >= 267) {
if (Cs < 339) {
if (Rs < 254) {
return 1;
}
else if (Rs >= 254) {
return 4;
}
}
else if (Cs >= 339) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (Rs < 229) {
return 8;
}
else if (Rs >= 229) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 276) {
if (RsDiff < 4) {
if (mcTcor < 89) {
if (AFD < 10) {
if (gchDC < 1) {
if (RsCsDiff < 3) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Cs < 311) {
return 8;
}
else if (Cs >= 311) {
if (Cs < 316) {
return 2;
}
else if (Cs >= 316) {
return 8;
}
}
}
}
else if (RsCsDiff >= 3) {
if (Cs < 209) {
return 1;
}
else if (Cs >= 209) {
if (ssDCval < 56) {
if (ssDCval < 52) {
if (RsCsDiff < 16) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (Rs < 629) {
return 8;
}
else if (Rs >= 629) {
return 1;
}
}
}
}
else if (RsCsDiff >= 16) {
if (Rs < 375) {
return 2;
}
else if (Rs >= 375) {
if (Rs < 486) {
return 4;
}
else if (Rs >= 486) {
return 8;
}
}
}
}
else if (ssDCval >= 52) {
return 4;
}
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 2) {
if (SC < 1019) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (SC >= 1019) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsCsDiff < 16) {
if (MVDiff < 19) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Rs < 449) {
return 8;
}
else if (Rs >= 449) {
return 4;
}
}
}
else if (MVDiff >= 19) {
if (Cs < 483) {
if (SC < 782) {
if (Rs < 322) {
return 8;
}
else if (Rs >= 322) {
return 2;
}
}
else if (SC >= 782) {
return 8;
}
}
else if (Cs >= 483) {
return 1;
}
}
}
else if (RsCsDiff >= 16) {
if (Rs < 862) {
return 2;
}
else if (Rs >= 862) {
return 8;
}
}
}
}
}
}
}
else if (gchDC >= 1) {
if (negBalance < 38) {
if (mu_mv_mag_sq < 99) {
if (diffTSC < -1) {
if (Cs < 168) {
return 2;
}
else if (Cs >= 168) {
if (Rs < 378) {
return 8;
}
else if (Rs >= 378) {
return 4;
}
}
}
else if (diffTSC >= -1) {
if (posBalance < 62) {
if (negBalance < 27) {
if (Cs < 237) {
if (Rs < 566) {
return 8;
}
else if (Rs >= 566) {
return 2;
}
}
else if (Cs >= 237) {
return 1;
}
}
else if (negBalance >= 27) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
if (Cs < 473) {
return 2;
}
else if (Cs >= 473) {
return 1;
}
}
}
}
else if (posBalance >= 62) {
if (Cs < 208) {
return 1;
}
else if (Cs >= 208) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (Rs < 338) {
return 1;
}
else if (Rs >= 338) {
return 4;
}
}
}
else if (negBalance >= 38) {
if (Cs < 169) {
if (RsDiff < 2) {
if (Rs < 365) {
return 8;
}
else if (Rs >= 365) {
return 4;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Cs >= 169) {
if (diffTSC < 1) {
if (AFD < 9) {
if (negBalance < 44) {
if (SC < 719) {
if (Cs < 185) {
return 8;
}
else if (Cs >= 185) {
return 2;
}
}
else if (SC >= 719) {
return 8;
}
}
else if (negBalance >= 44) {
if (RsCsDiff < 16) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (Cs < 314) {
return 1;
}
else if (Cs >= 314) {
return 8;
}
}
}
else if (RsCsDiff >= 16) {
return 4;
}
}
}
else if (AFD >= 9) {
if (refDCval < 39) {
return 8;
}
else if (refDCval >= 39) {
if (Cs < 461) {
return 4;
}
else if (Cs >= 461) {
return 8;
}
}
}
}
else if (diffTSC >= 1) {
if (MVDiff < 18) {
if (diffMVdiffVal < -4) {
if (Rs < 477) {
return 2;
}
else if (Rs >= 477) {
return 4;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (MVDiff >= 18) {
if (SC < 808) {
return 2;
}
else if (SC >= 808) {
return 4;
}
}
}
}
}
}
}
else if (AFD >= 10) {
if (refDCval < 76) {
if (MVDiff < 31) {
if (refDCval < 45) {
return 8;
}
else if (refDCval >= 45) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (Cs < 244) {
if (Rs < 347) {
return 4;
}
else if (Rs >= 347) {
return 8;
}
}
else if (Cs >= 244) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
return 8;
}
}
}
}
}
else if (MVDiff >= 31) {
if (diffAFD < -3) {
if (MVDiff < 49) {
return 8;
}
else if (MVDiff >= 49) {
return 1;
}
}
else if (diffAFD >= -3) {
if (SC < 549) {
if (negBalance < 49) {
if (SC < 381) {
return 2;
}
else if (SC >= 381) {
return 1;
}
}
else if (negBalance >= 49) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (SC >= 549) {
if (tcor < 61) {
if (diffTSC < 3) {
if (posBalance < 58) {
if (ssDCval < 41) {
return 8;
}
else if (ssDCval >= 41) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (posBalance >= 58) {
if (Cs < 409) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (MVDiff < 522) {
return 8;
}
else if (MVDiff >= 522) {
return 4;
}
}
}
else if (Cs >= 409) {
if (Cs < 511) {
return 4;
}
else if (Cs >= 511) {
return 8;
}
}
}
}
else if (diffTSC >= 3) {
if (Rs < 555) {
return 2;
}
else if (Rs >= 555) {
return 1;
}
}
}
else if (tcor >= 61) {
if (RsCsDiff < 9) {
return 4;
}
else if (RsCsDiff >= 9) {
if (Rs < 340) {
return 1;
}
else if (Rs >= 340) {
if (Cs < 494) {
return 2;
}
else if (Cs >= 494) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 76) {
if (MVDiff < 28) {
if (Cs < 64) {
return 1;
}
else if (Cs >= 64) {
return 8;
}
}
else if (MVDiff >= 28) {
if (tcor < 74) {
if (TSC < 11) {
if (CsDiff < 4) {
if (ssDCval < 172) {
if (negBalance < 50) {
if (Rs < 356) {
return 2;
}
else if (Rs >= 356) {
return 8;
}
}
else if (negBalance >= 50) {
return 2;
}
}
else if (ssDCval >= 172) {
return 1;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (TSC >= 11) {
if (ssDCval < 109) {
if (Rs < 359) {
return 4;
}
else if (Rs >= 359) {
return 2;
}
}
else if (ssDCval >= 109) {
return 1;
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 45) {
return 1;
}
else if (mu_mv_mag_sq >= 45) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 89) {
if (SC < 878) {
if (Cs < 148) {
if (posBalance < 22) {
if (negBalance < 39) {
if (SC < 397) {
return 1;
}
else if (SC >= 397) {
return 4;
}
}
else if (negBalance >= 39) {
if (MVDiff < 42) {
return 8;
}
else if (MVDiff >= 42) {
if (Cs < 133) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
else if (Cs >= 133) {
return 4;
}
}
}
}
else if (posBalance >= 22) {
if (mu_mv_mag_sq < 37) {
if (MVDiff < 10) {
if (CsDiff < 1) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (Rs < 449) {
return 1;
}
else if (Rs >= 449) {
return 4;
}
}
}
else if (CsDiff >= 1) {
if (Rs < 308) {
return 1;
}
else if (Rs >= 308) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (Cs < 114) {
if (ssDCval < 109) {
if (tcor < 80) {
if (Cs < 84) {
if (negBalance < 40) {
if (Rs < 322) {
return 1;
}
else if (Rs >= 322) {
return 8;
}
}
else if (negBalance >= 40) {
return 1;
}
}
else if (Cs >= 84) {
return 8;
}
}
else if (tcor >= 80) {
if (gchDC < 11) {
return 1;
}
else if (gchDC >= 11) {
return 8;
}
}
}
else if (ssDCval >= 109) {
return 8;
}
}
else if (Cs >= 114) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (MVDiff < 31) {
if (Rs < 380) {
return 4;
}
else if (Rs >= 380) {
return 1;
}
}
else if (MVDiff >= 31) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (ssDCval < 101) {
if (Rs < 771) {
if (RsCsDiff < 2) {
if (mcTcor < 96) {
if (Cs < 119) {
return 8;
}
else if (Cs >= 119) {
return 2;
}
}
else if (mcTcor >= 96) {
if (ssDCval < 72) {
if (Cs < 120) {
if (Rs < 280) {
return 2;
}
else if (Rs >= 280) {
if (Rs < 727) {
return 1;
}
else if (Rs >= 727) {
return 8;
}
}
}
else if (Cs >= 120) {
if (mcTcor < 98) {
if (Rs < 292) {
return 4;
}
else if (Rs >= 292) {
return 2;
}
}
else if (mcTcor >= 98) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
else if (ssDCval >= 72) {
if (Cs < 70) {
return 1;
}
else if (Cs >= 70) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
if (tcor < 71) {
if (posBalance < 57) {
if (Cs < 26) {
return 1;
}
else if (Cs >= 26) {
return 2;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (tcor >= 71) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (mu_mv_mag_sq < 46) {
return 1;
}
else if (mu_mv_mag_sq >= 46) {
return 2;
}
}
}
}
}
else if (Rs >= 771) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 4;
}
}
}
else if (ssDCval >= 101) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 2;
}
}
}
}
}
else if (Cs >= 148) {
if (refDCval < 71) {
if (mu_mv_mag_sq < 102) {
if (refDCval < 45) {
if (posBalance < 35) {
if (mu_mv_mag_sq < 72) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
if (Rs < 330) {
return 1;
}
else if (Rs >= 330) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 72) {
return 1;
}
}
else if (posBalance >= 35) {
if (mcTcor < 92) {
if (Cs < 311) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 50) {
if (Rs < 356) {
return 4;
}
else if (Rs >= 356) {
return 1;
}
}
else if (mu_mv_mag_sq >= 50) {
return 1;
}
}
else if (posBalance >= 55) {
if (ssDCval < 41) {
return 8;
}
else if (ssDCval >= 41) {
return 1;
}
}
}
}
else if (Cs >= 311) {
return 1;
}
}
else if (mcTcor >= 92) {
if (refDCval < 43) {
if (diffTSC < 1) {
if (mu_mv_mag_sq < 4) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 4) {
return 1;
}
}
else if (diffTSC >= 1) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (refDCval >= 43) {
if (Cs < 231) {
return 1;
}
else if (Cs >= 231) {
if (Rs < 407) {
return 4;
}
else if (Rs >= 407) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 45) {
if (TSCindex < 3) {
if (mcTcor < 99) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
if (SC < 746) {
return 4;
}
else if (SC >= 746) {
return 1;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 377) {
return 4;
}
else if (Rs >= 377) {
if (AFD < 3) {
if (Cs < 338) {
if (refDCval < 58) {
return 4;
}
else if (refDCval >= 58) {
if (Rs < 423) {
return 4;
}
else if (Rs >= 423) {
return 8;
}
}
}
else if (Cs >= 338) {
return 8;
}
}
else if (AFD >= 3) {
if (negBalance < 24) {
return 8;
}
else if (negBalance >= 24) {
return 1;
}
}
}
}
}
else if (TSCindex >= 3) {
if (Cs < 258) {
if (tcor < 89) {
if (Rs < 485) {
if (MVDiff < 63) {
if (TSC < 3) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 4;
}
}
else if (TSC >= 3) {
return 4;
}
}
else if (MVDiff >= 63) {
return 2;
}
}
else if (Rs >= 485) {
if (ssDCval < 65) {
if (Cs < 186) {
if (Rs < 674) {
return 2;
}
else if (Rs >= 674) {
return 1;
}
}
else if (Cs >= 186) {
return 1;
}
}
else if (ssDCval >= 65) {
return 4;
}
}
}
else if (tcor >= 89) {
if (Cs < 200) {
if (diffMVdiffVal < -3) {
if (Rs < 404) {
return 4;
}
else if (Rs >= 404) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
if (AFD < 5) {
if (Rs < 502) {
return 1;
}
else if (Rs >= 502) {
return 4;
}
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (Cs >= 200) {
if (mcTcor < 95) {
if (Rs < 300) {
return 4;
}
else if (Rs >= 300) {
return 1;
}
}
else if (mcTcor >= 95) {
if (posBalance < 55) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (diffAFD < -1) {
if (refDCval < 64) {
return 4;
}
else if (refDCval >= 64) {
return 2;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (refDCval < 64) {
if (diffMVdiffVal < -1) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (refDCval >= 64) {
return 4;
}
}
}
}
}
}
else if (Cs >= 258) {
if (mu_mv_mag_sq < 80) {
if (mcTcor < 97) {
if (RsCsDiff < 11) {
if (Rs < 461) {
if (ssDCval < 52) {
if (MVDiff < 8) {
if (Rs < 302) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 4;
}
}
else if (Rs >= 302) {
return 1;
}
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (ssDCval >= 52) {
if (RsCsDiff < 2) {
if (Rs < 369) {
if (Rs < 299) {
return 4;
}
else if (Rs >= 299) {
return 8;
}
}
else if (Rs >= 369) {
return 1;
}
}
else if (RsCsDiff >= 2) {
if (refDCval < 67) {
if (tcor < 82) {
if (refDCval < 58) {
if (Rs < 379) {
return 1;
}
else if (Rs >= 379) {
return 2;
}
}
else if (refDCval >= 58) {
return 4;
}
}
else if (tcor >= 82) {
if (negBalance < 47) {
if (posBalance < 58) {
if (AFD < 6) {
if (ssDCval < 65) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 2;
}
}
}
else if (ssDCval >= 65) {
if (Cs < 296) {
return 1;
}
else if (Cs >= 296) {
return 4;
}
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (posBalance >= 58) {
if (Cs < 416) {
return 2;
}
else if (Cs >= 416) {
return 1;
}
}
}
else if (negBalance >= 47) {
return 2;
}
}
}
else if (refDCval >= 67) {
if (SC < 798) {
return 1;
}
else if (SC >= 798) {
return 2;
}
}
}
}
}
else if (Rs >= 461) {
if (TSC < 4) {
if (Cs < 407) {
return 1;
}
else if (Cs >= 407) {
return 4;
}
}
else if (TSC >= 4) {
if (ssDCval < 59) {
if (Rs < 504) {
return 1;
}
else if (Rs >= 504) {
return 8;
}
}
else if (ssDCval >= 59) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 11) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 40) {
if (Rs < 348) {
if (Rs < 281) {
return 2;
}
else if (Rs >= 281) {
return 4;
}
}
else if (Rs >= 348) {
if (diffRsCsdiff < 1) {
if (SC < 767) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
return 4;
}
}
else if (SC >= 767) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (TSC < 4) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (TSC >= 4) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 1) {
if (RsDiff < 2) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
return 1;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (negBalance >= 40) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (AFD < 16) {
if (diffMVdiffVal < -4) {
if (Cs < 417) {
return 4;
}
else if (Cs >= 417) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 1;
}
}
else if (AFD >= 16) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (MVDiff < 3) {
if (ssDCval < 70) {
if (Cs < 515) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 4;
}
}
else if (Cs >= 515) {
return 8;
}
}
else if (ssDCval >= 70) {
return 2;
}
}
else if (MVDiff >= 3) {
if (Rs < 365) {
if (tcor < 98) {
if (posBalance < 47) {
if (Rs < 345) {
return 8;
}
else if (Rs >= 345) {
return 1;
}
}
else if (posBalance >= 47) {
return 1;
}
}
else if (tcor >= 98) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
else if (Rs >= 365) {
if (Cs < 385) {
if (Rs < 477) {
if (Rs < 384) {
return 2;
}
else if (Rs >= 384) {
return 8;
}
}
else if (Rs >= 477) {
if (Rs < 507) {
return 2;
}
else if (Rs >= 507) {
return 1;
}
}
}
else if (Cs >= 385) {
if (Cs < 445) {
return 2;
}
else if (Cs >= 445) {
if (Rs < 405) {
return 8;
}
else if (Rs >= 405) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (ssDCval < 60) {
if (TSC < 4) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 115) {
if (gchDC < 2) {
if (SC < 576) {
return 8;
}
else if (SC >= 576) {
return 1;
}
}
else if (gchDC >= 2) {
if (Rs < 349) {
return 1;
}
else if (Rs >= 349) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 308) {
return 4;
}
else if (Rs >= 308) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (TSCindex < 2) {
if (Rs < 361) {
return 1;
}
else if (Rs >= 361) {
return 8;
}
}
else if (TSCindex >= 2) {
if (Cs < 275) {
if (Rs < 458) {
return 1;
}
else if (Rs >= 458) {
return 8;
}
}
else if (Cs >= 275) {
return 1;
}
}
}
}
else if (TSC >= 4) {
return 4;
}
}
else if (ssDCval >= 60) {
if (SC < 762) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 2;
}
}
}
else if (SC >= 762) {
if (ssDCval < 69) {
if (negBalance < 19) {
if (Cs < 367) {
return 8;
}
else if (Cs >= 367) {
return 1;
}
}
else if (negBalance >= 19) {
return 1;
}
}
else if (ssDCval >= 69) {
if (negBalance < 32) {
if (SC < 849) {
return 1;
}
else if (SC >= 849) {
return 2;
}
}
else if (negBalance >= 32) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 71) {
if (mu_mv_mag_sq < 117) {
if (mcTcor < 96) {
if (MVDiff < 21) {
if (diffAFD < 3) {
if (diffMVdiffVal < -2) {
if (TSC < 5) {
if (Cs < 319) {
if (AFD < 7) {
if (Rs < 301) {
return 8;
}
else if (Rs >= 301) {
return 1;
}
}
else if (AFD >= 7) {
if (Rs < 306) {
return 1;
}
else if (Rs >= 306) {
return 2;
}
}
}
else if (Cs >= 319) {
if (Rs < 308) {
if (Rs < 292) {
return 8;
}
else if (Rs >= 292) {
return 2;
}
}
else if (Rs >= 308) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 347) {
return 8;
}
else if (Rs >= 347) {
return 1;
}
}
}
}
}
else if (TSC >= 5) {
if (Cs < 486) {
return 8;
}
else if (Cs >= 486) {
return 2;
}
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 557) {
if (diffRsCsdiff < 2) {
if (Cs < 230) {
if (Rs < 324) {
return 8;
}
else if (Rs >= 324) {
if (ssDCval < 126) {
return 1;
}
else if (ssDCval >= 126) {
return 8;
}
}
}
else if (Cs >= 230) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
if (AFD < 10) {
if (Rs < 372) {
return 2;
}
else if (Rs >= 372) {
return 8;
}
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (Rs >= 557) {
return 4;
}
}
}
else if (diffAFD >= 3) {
if (Rs < 444) {
return 1;
}
else if (Rs >= 444) {
return 4;
}
}
}
else if (MVDiff >= 21) {
if (diffMVdiffVal < 16) {
if (posBalance < 65) {
if (tcor < 80) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (Rs < 317) {
return 1;
}
else if (Rs >= 317) {
return 8;
}
}
}
}
else if (tcor >= 80) {
if (Cs < 350) {
if (SC < 587) {
return 2;
}
else if (SC >= 587) {
return 1;
}
}
else if (Cs >= 350) {
if (Rs < 383) {
return 1;
}
else if (Rs >= 383) {
return 8;
}
}
}
}
else if (posBalance >= 65) {
if (Cs < 350) {
return 2;
}
else if (Cs >= 350) {
return 8;
}
}
}
else if (diffMVdiffVal >= 16) {
if (AFD < 11) {
if (MVDiff < 255) {
return 8;
}
else if (MVDiff >= 255) {
return 4;
}
}
else if (AFD >= 11) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
else if (mcTcor >= 96) {
if (RsCsDiff < 2) {
if (MVDiff < 2) {
if (TSC < 2) {
return 2;
}
else if (TSC >= 2) {
if (Rs < 428) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (Rs >= 428) {
if (Rs < 542) {
return 1;
}
else if (Rs >= 542) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 180) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
else if (Cs >= 180) {
if (MVDiff < 13) {
if (SC < 756) {
if (Cs < 211) {
if (Rs < 297) {
return 4;
}
else if (Rs >= 297) {
return 1;
}
}
else if (Cs >= 211) {
if (Cs < 339) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (RsCsDiff < 1) {
return 8;
}
else if (RsCsDiff >= 1) {
return 1;
}
}
}
else if (Cs >= 339) {
if (Rs < 350) {
return 4;
}
else if (Rs >= 350) {
return 8;
}
}
}
}
else if (SC >= 756) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (Rs < 471) {
return 1;
}
else if (Rs >= 471) {
return 8;
}
}
}
}
else if (MVDiff >= 13) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
if (refDCval < 116) {
if (mu_mv_mag_sq < 26) {
if (TSC < 7) {
if (ssDCval < 83) {
if (mcTcor < 99) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (mcTcor >= 99) {
return 4;
}
}
else if (ssDCval >= 83) {
if (posBalance < 47) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 357) {
return 8;
}
else if (Rs >= 357) {
return 4;
}
}
}
else if (posBalance >= 47) {
return 1;
}
}
}
else if (TSC >= 7) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (Rs < 319) {
return 1;
}
else if (Rs >= 319) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (tcor < 88) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (Cs < 199) {
return 8;
}
else if (Cs >= 199) {
return 2;
}
}
}
else if (tcor >= 88) {
if (negBalance < 49) {
if (SC < 837) {
if (RsDiff < 2) {
if (posBalance < 51) {
if (SC < 780) {
return 8;
}
else if (SC >= 780) {
return 1;
}
}
else if (posBalance >= 51) {
return 2;
}
}
else if (RsDiff >= 2) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (Cs < 270) {
return 8;
}
else if (Cs >= 270) {
return 2;
}
}
}
}
else if (SC >= 837) {
return 1;
}
}
else if (negBalance >= 49) {
if (Cs < 486) {
return 8;
}
else if (Cs >= 486) {
return 4;
}
}
}
}
}
else if (refDCval >= 116) {
if (mcTcor < 99) {
if (ssDCval < 160) {
if (refDCval < 134) {
if (posBalance < 56) {
if (Cs < 520) {
return 2;
}
else if (Cs >= 520) {
return 4;
}
}
else if (posBalance >= 56) {
if (Rs < 391) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 2;
}
}
else if (Rs >= 391) {
if (Rs < 467) {
return 4;
}
else if (Rs >= 467) {
return 8;
}
}
}
}
else if (refDCval >= 134) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (ssDCval >= 160) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 2;
}
}
}
}
else if (mcTcor >= 99) {
if (posBalance < 54) {
if (Rs < 362) {
return 2;
}
else if (Rs >= 362) {
return 1;
}
}
else if (posBalance >= 54) {
if (ssDCval < 119) {
return 2;
}
else if (ssDCval >= 119) {
if (Cs < 448) {
return 4;
}
else if (Cs >= 448) {
return 1;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (ssDCval < 94) {
if (RsDiff < 1) {
if (SC < 509) {
if (Rs < 289) {
return 2;
}
else if (Rs >= 289) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
}
else if (SC >= 509) {
if (SC < 863) {
if (posBalance < 37) {
if (mu_mv_mag_sq < 126) {
if (Rs < 319) {
return 1;
}
else if (Rs >= 319) {
return 2;
}
}
else if (mu_mv_mag_sq >= 126) {
return 2;
}
}
else if (posBalance >= 37) {
return 2;
}
}
else if (SC >= 863) {
return 1;
}
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (ssDCval >= 94) {
if (Cs < 359) {
return 1;
}
else if (Cs >= 359) {
return 8;
}
}
}
}
}
}
else if (SC >= 878) {
if (refDCval < 94) {
if (tcor < 98) {
if (Rs < 1060) {
if (Cs < 436) {
if (negBalance < 33) {
if (TSCindex < 2) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (Rs < 504) {
return 8;
}
else if (Rs >= 504) {
return 4;
}
}
}
else if (TSCindex >= 2) {
if (RsCsDiff < 3) {
if (Cs < 288) {
if (MVDiff < 12) {
if (ssDCval < 67) {
if (mcTcor < 97) {
if (SC < 964) {
if (Rs < 791) {
return 2;
}
else if (Rs >= 791) {
return 8;
}
}
else if (SC >= 964) {
return 1;
}
}
else if (mcTcor >= 97) {
if (Rs < 1036) {
return 1;
}
else if (Rs >= 1036) {
return 4;
}
}
}
else if (ssDCval >= 67) {
return 4;
}
}
else if (MVDiff >= 12) {
return 4;
}
}
else if (Cs >= 288) {
if (RsCsDiff < 1) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (RsCsDiff >= 1) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
if (Rs < 735) {
return 1;
}
else if (Rs >= 735) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 3) {
if (MVDiff < 16) {
if (posBalance < 48) {
if (mu_mv_mag_sq < 47) {
if (posBalance < 40) {
return 2;
}
else if (posBalance >= 40) {
return 1;
}
}
else if (mu_mv_mag_sq >= 47) {
if (mu_mv_mag_sq < 84) {
return 1;
}
else if (mu_mv_mag_sq >= 84) {
if (Rs < 832) {
return 2;
}
else if (Rs >= 832) {
return 8;
}
}
}
}
else if (posBalance >= 48) {
if (Cs < 323) {
return 2;
}
else if (Cs >= 323) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 8;
}
}
}
}
else if (MVDiff >= 16) {
return 1;
}
}
}
}
else if (negBalance >= 33) {
if (mu_mv_mag_sq < 49) {
if (tcor < 81) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (ssDCval < 54) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (MVDiff < 64) {
return 1;
}
else if (MVDiff >= 64) {
return 8;
}
}
}
else if (ssDCval >= 54) {
if (Cs < 304) {
return 2;
}
else if (Cs >= 304) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
if (Rs < 651) {
return 2;
}
else if (Rs >= 651) {
return 1;
}
}
}
}
}
}
else if (tcor >= 81) {
if (ssDCval < 76) {
if (refDCval < 61) {
if (TSC < 4) {
if (diffMVdiffVal < -4) {
if (Cs < 219) {
return 1;
}
else if (Cs >= 219) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
return 1;
}
}
else if (TSC >= 4) {
if (ssDCval < 55) {
if (Rs < 697) {
return 8;
}
else if (Rs >= 697) {
return 1;
}
}
else if (ssDCval >= 55) {
return 1;
}
}
}
else if (refDCval >= 61) {
if (Cs < 315) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 958) {
if (TSC < 6) {
if (Cs < 216) {
return 1;
}
else if (Cs >= 216) {
if (RsCsDiff < 7) {
return 4;
}
else if (RsCsDiff >= 7) {
if (Cs < 223) {
return 4;
}
else if (Cs >= 223) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
}
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (Rs >= 958) {
return 4;
}
}
}
else if (Cs >= 315) {
if (AFD < 6) {
if (Rs < 604) {
return 1;
}
else if (Rs >= 604) {
return 8;
}
}
else if (AFD >= 6) {
return 4;
}
}
}
}
else if (ssDCval >= 76) {
if (MVDiff < 32) {
return 1;
}
else if (MVDiff >= 32) {
if (Rs < 851) {
return 4;
}
else if (Rs >= 851) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (negBalance < 39) {
if (Rs < 849) {
if (mu_mv_mag_sq < 101) {
if (mcTcor < 97) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (SC < 1012) {
return 1;
}
else if (SC >= 1012) {
return 2;
}
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 675) {
return 8;
}
else if (Rs >= 675) {
return 4;
}
}
}
else if (Rs >= 849) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (negBalance >= 39) {
if (AFD < 5) {
if (Rs < 487) {
return 2;
}
else if (Rs >= 487) {
return 8;
}
}
else if (AFD >= 5) {
if (mu_mv_mag_sq < 83) {
if (diffTSC < 1) {
if (Rs < 990) {
return 8;
}
else if (Rs >= 990) {
return 1;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 624) {
return 2;
}
else if (Rs >= 624) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 436) {
if (TSC < 4) {
if (RsCsDiff < 4) {
if (Rs < 531) {
if (posBalance < 55) {
if (RsCsDiff < 2) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (RsCsDiff >= 2) {
if (diffMVdiffVal < 1) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
if (mcTcor < 97) {
if (Rs < 513) {
return 2;
}
else if (Rs >= 513) {
return 8;
}
}
else if (mcTcor >= 97) {
if (Rs < 508) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Rs >= 508) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
else if (posBalance >= 55) {
return 2;
}
}
else if (Rs >= 531) {
if (SC < 1251) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 48) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 48) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
else if (SC >= 1251) {
if (mu_mv_mag_sq < 83) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
if (Rs < 815) {
return 8;
}
else if (Rs >= 815) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 1001) {
if (Rs < 774) {
return 4;
}
else if (Rs >= 774) {
return 8;
}
}
else if (Rs >= 1001) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 4) {
if (mu_mv_mag_sq < 75) {
if (Rs < 622) {
if (mcTcor < 96) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 2;
}
}
else if (mcTcor >= 96) {
if (posBalance < 67) {
if (mu_mv_mag_sq < 21) {
return 4;
}
else if (mu_mv_mag_sq >= 21) {
if (Rs < 467) {
return 2;
}
else if (Rs >= 467) {
if (Rs < 584) {
return 8;
}
else if (Rs >= 584) {
return 4;
}
}
}
}
else if (posBalance >= 67) {
return 2;
}
}
}
else if (Rs >= 622) {
if (diffTSC < -1) {
if (Rs < 829) {
return 1;
}
else if (Rs >= 829) {
return 8;
}
}
else if (diffTSC >= -1) {
if (refDCval < 36) {
if (Cs < 485) {
if (Rs < 687) {
return 1;
}
else if (Rs >= 687) {
return 2;
}
}
else if (Cs >= 485) {
return 8;
}
}
else if (refDCval >= 36) {
if (diffAFD < 3) {
if (tcor < 97) {
return 2;
}
else if (tcor >= 97) {
return 1;
}
}
else if (diffAFD >= 3) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (SC < 998) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
return 1;
}
}
else if (SC >= 998) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
return 8;
}
}
}
}
}
else if (TSC >= 4) {
if (mcTcor < 97) {
if (ssDCval < 75) {
if (ssDCval < 63) {
if (Rs < 474) {
if (SC < 914) {
return 2;
}
else if (SC >= 914) {
return 1;
}
}
else if (Rs >= 474) {
if (MVDiff < 11) {
if (SC < 1299) {
if (posBalance < 60) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
if (Rs < 585) {
return 8;
}
else if (Rs >= 585) {
if (Rs < 713) {
return 2;
}
else if (Rs >= 713) {
return 4;
}
}
}
}
else if (posBalance >= 60) {
return 4;
}
}
else if (SC >= 1299) {
return 8;
}
}
else if (MVDiff >= 11) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
if (ssDCval < 55) {
if (negBalance < 38) {
return 4;
}
else if (negBalance >= 38) {
if (tcor < 77) {
return 1;
}
else if (tcor >= 77) {
return 8;
}
}
}
else if (ssDCval >= 55) {
if (Cs < 519) {
return 4;
}
else if (Cs >= 519) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 63) {
if (Rs < 599) {
if (Rs < 496) {
return 8;
}
else if (Rs >= 496) {
return 2;
}
}
else if (Rs >= 599) {
return 4;
}
}
}
else if (ssDCval >= 75) {
if (mu_mv_mag_sq < 15) {
if (AFD < 12) {
if (Cs < 523) {
return 1;
}
else if (Cs >= 523) {
return 8;
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 15) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (tcor < 90) {
if (Rs < 513) {
return 1;
}
else if (Rs >= 513) {
return 8;
}
}
else if (tcor >= 90) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1060) {
if (posBalance < 59) {
if (Cs < 105) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (Cs >= 105) {
if (refDCval < 67) {
if (Cs < 413) {
if (SC < 1762) {
if (negBalance < 21) {
return 2;
}
else if (negBalance >= 21) {
if (Cs < 131) {
if (Rs < 1275) {
return 1;
}
else if (Rs >= 1275) {
return 2;
}
}
else if (Cs >= 131) {
return 1;
}
}
}
else if (SC >= 1762) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
return 2;
}
}
}
else if (Cs >= 413) {
if (AFD < 5) {
if (Rs < 2270) {
return 1;
}
else if (Rs >= 2270) {
return 4;
}
}
else if (AFD >= 5) {
if (RsDiff < 3) {
if (Cs < 505) {
return 1;
}
else if (Cs >= 505) {
return 2;
}
}
else if (RsDiff >= 3) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
}
else if (refDCval >= 67) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (mu_mv_mag_sq < 123) {
if (diffRsCsdiff < 4) {
return 1;
}
else if (diffRsCsdiff >= 4) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 123) {
return 2;
}
}
}
}
}
else if (posBalance >= 59) {
if (Cs < 422) {
if (MVDiff < 7) {
if (posBalance < 65) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
else if (posBalance >= 65) {
return 2;
}
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (Cs >= 422) {
if (Rs < 1621) {
if (Rs < 1138) {
if (Rs < 1112) {
return 8;
}
else if (Rs >= 1112) {
return 4;
}
}
else if (Rs >= 1138) {
return 8;
}
}
else if (Rs >= 1621) {
if (Rs < 2091) {
return 1;
}
else if (Rs >= 2091) {
return 2;
}
}
}
}
}
}
else if (tcor >= 98) {
if (AFD < 3) {
if (ssDCval < 64) {
if (Cs < 278) {
if (Rs < 1722) {
if (mu_mv_mag_sq < 98) {
if (Rs < 816) {
return 1;
}
else if (Rs >= 816) {
return 4;
}
}
else if (mu_mv_mag_sq >= 98) {
if (mu_mv_mag_sq < 116) {
if (Cs < 150) {
return 8;
}
else if (Cs >= 150) {
if (Cs < 211) {
return 1;
}
else if (Cs >= 211) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
}
else if (Rs >= 1722) {
return 2;
}
}
else if (Cs >= 278) {
if (posBalance < 42) {
if (Cs < 407) {
if (tcor < 101) {
if (Rs < 779) {
if (ssDCval < 61) {
if (Rs < 536) {
return 1;
}
else if (Rs >= 536) {
return 8;
}
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (Rs >= 779) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1308) {
return 1;
}
else if (Rs >= 1308) {
return 8;
}
}
}
}
else if (tcor >= 101) {
return 8;
}
}
else if (Cs >= 407) {
if (ssDCval < 62) {
if (Cs < 507) {
if (gchDC < 1) {
if (mu_mv_mag_sq < 128) {
if (Cs < 484) {
if (mu_mv_mag_sq < 115) {
if (Rs < 697) {
if (TSC < 2) {
if (posBalance < 28) {
return 1;
}
else if (posBalance >= 28) {
return 8;
}
}
else if (TSC >= 2) {
if (Rs < 562) {
return 1;
}
else if (Rs >= 562) {
return 2;
}
}
}
else if (Rs >= 697) {
return 1;
}
}
else if (mu_mv_mag_sq >= 115) {
if (negBalance < 15) {
if (Rs < 886) {
return 8;
}
else if (Rs >= 886) {
return 1;
}
}
else if (negBalance >= 15) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
return 1;
}
}
}
}
else if (Cs >= 484) {
if (Rs < 1401) {
return 1;
}
else if (Rs >= 1401) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (gchDC >= 1) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
return 8;
}
}
}
else if (Cs >= 507) {
if (RsCsDiff < 2) {
if (Rs < 623) {
return 8;
}
else if (Rs >= 623) {
return 1;
}
}
else if (RsCsDiff >= 2) {
return 4;
}
}
}
else if (ssDCval >= 62) {
if (TSCindex < 2) {
if (Rs < 1164) {
return 1;
}
else if (Rs >= 1164) {
return 2;
}
}
else if (TSCindex >= 2) {
if (negBalance < 21) {
if (Rs < 671) {
return 4;
}
else if (Rs >= 671) {
return 1;
}
}
else if (negBalance >= 21) {
if (Cs < 445) {
return 1;
}
else if (Cs >= 445) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 42) {
if (Rs < 658) {
if (mu_mv_mag_sq < 102) {
if (negBalance < 24) {
return 8;
}
else if (negBalance >= 24) {
return 1;
}
}
else if (mu_mv_mag_sq >= 102) {
return 1;
}
}
else if (Rs >= 658) {
if (SC < 1240) {
return 2;
}
else if (SC >= 1240) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
if (Rs < 1290) {
return 8;
}
else if (Rs >= 1290) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (Rs < 1118) {
if (mu_mv_mag_sq < 113) {
if (Cs < 471) {
if (AFD < 2) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 941) {
return 1;
}
else if (Rs >= 941) {
return 8;
}
}
}
else if (AFD >= 2) {
if (negBalance < 31) {
if (Cs < 217) {
return 8;
}
else if (Cs >= 217) {
if (MVDiff < 14) {
if (SC < 898) {
return 8;
}
else if (SC >= 898) {
return 1;
}
}
else if (MVDiff >= 14) {
return 8;
}
}
}
else if (negBalance >= 31) {
if (Rs < 554) {
return 1;
}
else if (Rs >= 554) {
return 8;
}
}
}
}
else if (Cs >= 471) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (Rs < 644) {
return 1;
}
else if (Rs >= 644) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (MVDiff < 1) {
if (Rs < 797) {
if (refDCval < 70) {
if (SC < 994) {
if (negBalance < 19) {
if (Rs < 640) {
return 1;
}
else if (Rs >= 640) {
return 4;
}
}
else if (negBalance >= 19) {
return 2;
}
}
else if (SC >= 994) {
return 1;
}
}
else if (refDCval >= 70) {
if (posBalance < 43) {
if (Rs < 741) {
if (Rs < 542) {
return 1;
}
else if (Rs >= 542) {
return 4;
}
}
else if (Rs >= 741) {
return 1;
}
}
else if (posBalance >= 43) {
if (diffMVdiffVal < -2) {
return 2;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 629) {
return 1;
}
else if (Rs >= 629) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 1;
}
}
}
}
}
}
else if (Rs >= 797) {
if (ssDCval < 81) {
if (ssDCval < 75) {
if (Cs < 362) {
return 1;
}
else if (Cs >= 362) {
return 8;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
else if (ssDCval >= 81) {
if (Cs < 251) {
return 2;
}
else if (Cs >= 251) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 71) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (Rs < 849) {
return 2;
}
else if (Rs >= 849) {
return 8;
}
}
}
else if (ssDCval >= 71) {
if (Rs < 561) {
return 1;
}
else if (Rs >= 561) {
return 4;
}
}
}
}
}
else if (Rs >= 1118) {
if (Rs < 2043) {
if (mu_mv_mag_sq < 127) {
if (Rs < 1319) {
if (refDCval < 69) {
return 1;
}
else if (refDCval >= 69) {
if (Cs < 277) {
if (Rs < 1129) {
return 8;
}
else if (Rs >= 1129) {
return 1;
}
}
else if (Cs >= 277) {
return 8;
}
}
}
else if (Rs >= 1319) {
if (Rs < 1991) {
if (posBalance < 11) {
return 4;
}
else if (posBalance >= 11) {
return 1;
}
}
else if (Rs >= 1991) {
if (Cs < 279) {
return 2;
}
else if (Cs >= 279) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 20) {
if (Cs < 276) {
return 2;
}
else if (Cs >= 276) {
if (Cs < 493) {
if (refDCval < 64) {
return 2;
}
else if (refDCval >= 64) {
return 1;
}
}
else if (Cs >= 493) {
return 8;
}
}
}
else if (negBalance >= 20) {
return 1;
}
}
}
else if (Rs >= 2043) {
if (Cs < 469) {
return 2;
}
else if (Cs >= 469) {
return 8;
}
}
}
}
}
else if (AFD >= 3) {
if (ssDCval < 71) {
if (mu_mv_mag_sq < 34) {
if (Cs < 181) {
if (Rs < 1422) {
return 1;
}
else if (Rs >= 1422) {
return 2;
}
}
else if (Cs >= 181) {
if (refDCval < 53) {
if (Rs < 573) {
return 4;
}
else if (Rs >= 573) {
return 1;
}
}
else if (refDCval >= 53) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (negBalance < 22) {
if (diffMVdiffVal < 1) {
if (Cs < 302) {
return 4;
}
else if (Cs >= 302) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (negBalance >= 22) {
if (TSC < 3) {
if (ssDCval < 65) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Cs < 251) {
return 1;
}
else if (Cs >= 251) {
return 2;
}
}
}
else if (ssDCval >= 65) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (Rs < 1164) {
return 2;
}
else if (Rs >= 1164) {
return 4;
}
}
}
}
else if (TSC >= 3) {
if (MVDiff < 10) {
if (Rs < 714) {
if (mu_mv_mag_sq < 69) {
return 4;
}
else if (mu_mv_mag_sq >= 69) {
if (ssDCval < 60) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
return 4;
}
}
else if (ssDCval >= 60) {
return 8;
}
}
}
else if (Rs >= 714) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (gchDC < 5) {
return 8;
}
else if (gchDC >= 5) {
return 4;
}
}
}
}
else if (MVDiff >= 10) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (refDCval < 81) {
if (mu_mv_mag_sq < 96) {
if (RsCsDiff < 5) {
if (Rs < 1530) {
if (Cs < 422) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
return 8;
}
}
}
else if (Cs >= 422) {
return 1;
}
}
else if (Rs >= 1530) {
return 1;
}
}
else if (RsCsDiff >= 5) {
if (mu_mv_mag_sq < 90) {
return 8;
}
else if (mu_mv_mag_sq >= 90) {
if (Rs < 1640) {
return 1;
}
else if (Rs >= 1640) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 1445) {
return 8;
}
else if (Rs >= 1445) {
return 2;
}
}
}
else if (refDCval >= 81) {
if (Cs < 507) {
if (refDCval < 84) {
if (Rs < 1521) {
return 4;
}
else if (Rs >= 1521) {
if (Cs < 486) {
return 8;
}
else if (Cs >= 486) {
return 4;
}
}
}
else if (refDCval >= 84) {
if (negBalance < 36) {
if (MVDiff < 20) {
return 1;
}
else if (MVDiff >= 20) {
return 8;
}
}
else if (negBalance >= 36) {
if (Cs < 301) {
if (Rs < 1040) {
return 8;
}
else if (Rs >= 1040) {
return 1;
}
}
else if (Cs >= 301) {
return 8;
}
}
}
}
else if (Cs >= 507) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Rs < 773) {
return 1;
}
else if (Rs >= 773) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 94) {
if (MVDiff < 2) {
if (ssDCval < 98) {
if (Cs < 455) {
if (Rs < 574) {
return 8;
}
else if (Rs >= 574) {
return 4;
}
}
else if (Cs >= 455) {
if (Rs < 626) {
return 8;
}
else if (Rs >= 626) {
return 4;
}
}
}
else if (ssDCval >= 98) {
if (AFD < 8) {
if (Rs < 1713) {
return 8;
}
else if (Rs >= 1713) {
if (Rs < 1807) {
return 4;
}
else if (Rs >= 1807) {
if (Cs < 398) {
return 8;
}
else if (Cs >= 398) {
return 1;
}
}
}
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 116) {
if (MVDiff < 60) {
if (RsCsDiff < 7) {
if (Cs < 334) {
return 8;
}
else if (Cs >= 334) {
if (TSC < 3) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
if (ssDCval < 111) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
return 1;
}
}
else if (ssDCval >= 111) {
return 1;
}
}
}
else if (TSC >= 3) {
if (Cs < 351) {
if (Rs < 640) {
return 8;
}
else if (Rs >= 640) {
return 4;
}
}
else if (Cs >= 351) {
if (ssDCval < 107) {
return 8;
}
else if (ssDCval >= 107) {
if (Rs < 588) {
return 1;
}
else if (Rs >= 588) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 7) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (negBalance < 60) {
if (AFD < 8) {
if (Rs < 2003) {
return 8;
}
else if (Rs >= 2003) {
return 4;
}
}
else if (AFD >= 8) {
if (TSC < 7) {
if (tcor < 85) {
if (Rs < 657) {
return 1;
}
else if (Rs >= 657) {
return 8;
}
}
else if (tcor >= 85) {
if (Rs < 593) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Cs < 445) {
return 8;
}
else if (Cs >= 445) {
return 2;
}
}
}
else if (Rs >= 593) {
return 1;
}
}
}
else if (TSC >= 7) {
if (Cs < 264) {
return 1;
}
else if (Cs >= 264) {
return 8;
}
}
}
}
else if (negBalance >= 60) {
if (Rs < 574) {
return 2;
}
else if (Rs >= 574) {
return 4;
}
}
}
}
}
else if (MVDiff >= 60) {
if (Cs < 289) {
return 1;
}
else if (Cs >= 289) {
if (Rs < 501) {
if (Rs < 453) {
return 1;
}
else if (Rs >= 453) {
return 2;
}
}
else if (Rs >= 501) {
return 8;
}
}
}
}
else if (ssDCval >= 116) {
if (refDCval < 118) {
if (Cs < 452) {
return 1;
}
else if (Cs >= 452) {
return 2;
}
}
else if (refDCval >= 118) {
if (posBalance < 54) {
if (tcor < 93) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
if (Rs < 677) {
return 1;
}
else if (Rs >= 677) {
return 2;
}
}
}
}
}
else if (tcor >= 93) {
if (Cs < 90) {
return 8;
}
else if (Cs >= 90) {
if (mcTcor < 99) {
if (Rs < 870) {
if (AFD < 3) {
if (Rs < 465) {
return 4;
}
else if (Rs >= 465) {
return 1;
}
}
else if (AFD >= 3) {
if (MVDiff < 31) {
return 8;
}
else if (MVDiff >= 31) {
return 4;
}
}
}
else if (Rs >= 870) {
if (Rs < 1486) {
return 1;
}
else if (Rs >= 1486) {
return 2;
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
if (Cs < 412) {
return 4;
}
else if (Cs >= 412) {
if (mu_mv_mag_sq < 65) {
if (diffMVdiffVal < -2) {
if (Rs < 441) {
return 4;
}
else if (Rs >= 441) {
if (Cs < 514) {
return 2;
}
else if (Cs >= 514) {
return 1;
}
}
}
else if (diffMVdiffVal >= -2) {
if (ssDCval < 122) {
if (Rs < 466) {
return 1;
}
else if (Rs >= 466) {
if (Rs < 585) {
return 2;
}
else if (Rs >= 585) {
return 8;
}
}
}
else if (ssDCval >= 122) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 65) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (RsCsDiff < 176) {
if (Cs < 348) {
if (refDCval < 89) {
if (Rs < 610) {
if (negBalance < 46) {
if (Cs < 302) {
return 1;
}
else if (Cs >= 302) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
}
}
else if (negBalance >= 46) {
if (Cs < 168) {
if (mu_mv_mag_sq < 88) {
return 1;
}
else if (mu_mv_mag_sq >= 88) {
if (Rs < 411) {
return 4;
}
else if (Rs >= 411) {
return 8;
}
}
}
else if (Cs >= 168) {
if (Cs < 309) {
if (negBalance < 56) {
return 4;
}
else if (negBalance >= 56) {
if (ssDCval < 57) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
else if (ssDCval >= 57) {
if (mcTcor < 92) {
if (Rs < 460) {
return 8;
}
else if (Rs >= 460) {
return 2;
}
}
else if (mcTcor >= 92) {
return 4;
}
}
}
}
else if (Cs >= 309) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
else if (Rs >= 610) {
if (mu_mv_mag_sq < 28) {
if (SC < 2140) {
if (mu_mv_mag_sq < 15) {
if (Rs < 625) {
if (Rs < 622) {
return 2;
}
else if (Rs >= 622) {
return 8;
}
}
else if (Rs >= 625) {
if (RsCsDiff < 71) {
if (SC < 770) {
return 8;
}
else if (SC >= 770) {
return 1;
}
}
else if (RsCsDiff >= 71) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
if (Rs < 713) {
return 2;
}
else if (Rs >= 713) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (tcor < 22) {
return 8;
}
else if (tcor >= 22) {
return 1;
}
}
}
else if (SC >= 2140) {
return 4;
}
}
else if (mu_mv_mag_sq >= 28) {
if (MVDiff < 126) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (RsCsDiff < 71) {
return 1;
}
else if (RsCsDiff >= 71) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Rs < 1040) {
return 1;
}
else if (Rs >= 1040) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 126) {
return 2;
}
}
}
}
else if (refDCval >= 89) {
if (mcTcor < 81) {
if (ssDCval < 118) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
if (Rs < 668) {
return 1;
}
else if (Rs >= 668) {
return 2;
}
}
}
}
else if (ssDCval >= 118) {
return 1;
}
}
else if (mcTcor >= 81) {
if (refDCval < 100) {
if (RsCsDiff < 43) {
if (posBalance < 57) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
else if (posBalance >= 57) {
if (TSC < 7) {
if (gchDC < 1) {
if (MVDiff < 32) {
return 8;
}
else if (MVDiff >= 32) {
return 2;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (TSC >= 7) {
if (Rs < 379) {
return 1;
}
else if (Rs >= 379) {
return 2;
}
}
}
}
else if (RsCsDiff >= 43) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
if (refDCval < 94) {
return 2;
}
else if (refDCval >= 94) {
if (refDCval < 96) {
return 1;
}
else if (refDCval >= 96) {
return 2;
}
}
}
}
}
else if (refDCval >= 100) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 154) {
if (SC < 1182) {
if (Rs < 852) {
return 8;
}
else if (Rs >= 852) {
return 2;
}
}
else if (SC >= 1182) {
return 8;
}
}
else if (ssDCval >= 154) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffRsCsdiff < 17) {
if (tcor < 43) {
return 1;
}
else if (tcor >= 43) {
if (CsDiff < 2) {
if (mcTcor < 93) {
if (MVDiff < 59) {
return 8;
}
else if (MVDiff >= 59) {
return 4;
}
}
else if (mcTcor >= 93) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (CsDiff >= 2) {
if (mcTcor < 87) {
if (Rs < 1203) {
return 2;
}
else if (Rs >= 1203) {
return 1;
}
}
else if (mcTcor >= 87) {
if (mu_mv_mag_sq < 58) {
if (Cs < 307) {
if (Rs < 1088) {
return 1;
}
else if (Rs >= 1088) {
return 2;
}
}
else if (Cs >= 307) {
return 8;
}
}
else if (mu_mv_mag_sq >= 58) {
return 2;
}
}
}
}
}
else if (diffRsCsdiff >= 17) {
if (ssDCval < 173) {
return 1;
}
else if (ssDCval >= 173) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 348) {
if (RsCsDiff < 36) {
if (negBalance < 48) {
if (diffMVdiffVal < 5) {
if (MVDiff < 19) {
if (Cs < 395) {
return 4;
}
else if (Cs >= 395) {
return 1;
}
}
else if (MVDiff >= 19) {
if (tcor < 61) {
if (AFD < 39) {
return 1;
}
else if (AFD >= 39) {
return 2;
}
}
else if (tcor >= 61) {
return 2;
}
}
}
else if (diffMVdiffVal >= 5) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (Rs < 650) {
if (mcTcor < 89) {
if (ssDCval < 42) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
else if (ssDCval >= 42) {
if (diffTSC < 1) {
if (Cs < 515) {
return 4;
}
else if (Cs >= 515) {
if (Rs < 605) {
return 2;
}
else if (Rs >= 605) {
return 4;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 454) {
return 8;
}
else if (Rs >= 454) {
return 2;
}
}
}
}
else if (mcTcor >= 89) {
if (tcor < 67) {
return 1;
}
else if (tcor >= 67) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 8;
}
}
}
}
else if (Rs >= 650) {
if (diffMVdiffVal < -9) {
if (Rs < 999) {
return 4;
}
else if (Rs >= 999) {
return 2;
}
}
else if (diffMVdiffVal >= -9) {
if (diffMVdiffVal < 6) {
if (Cs < 364) {
return 2;
}
else if (Cs >= 364) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
if (Rs < 720) {
return 8;
}
else if (Rs >= 720) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 36) {
if (ssDCval < 52) {
if (negBalance < 47) {
if (tcor < 59) {
if (Cs < 361) {
return 2;
}
else if (Cs >= 361) {
return 1;
}
}
else if (tcor >= 59) {
return 2;
}
}
else if (negBalance >= 47) {
if (RsDiff < 7) {
if (AFD < 15) {
if (Cs < 403) {
return 2;
}
else if (Cs >= 403) {
return 8;
}
}
else if (AFD >= 15) {
return 2;
}
}
else if (RsDiff >= 7) {
if (Rs < 959) {
return 8;
}
else if (Rs >= 959) {
return 1;
}
}
}
}
else if (ssDCval >= 52) {
if (TSC < 11) {
if (negBalance < 51) {
if (Cs < 477) {
if (RsDiff < 12) {
if (mu_mv_mag_sq < 5) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
else if (mu_mv_mag_sq >= 5) {
return 1;
}
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (Cs >= 477) {
if (CsDiff < 4) {
if (Rs < 1243) {
return 1;
}
else if (Rs >= 1243) {
return 8;
}
}
else if (CsDiff >= 4) {
if (diffRsCsdiff < -5) {
if (Rs < 524) {
return 4;
}
else if (Rs >= 524) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
return 2;
}
}
}
}
else if (negBalance >= 51) {
if (posBalance < 61) {
if (refDCval < 84) {
if (diffRsCsdiff < 4) {
if (RsCsDiff < 51) {
if (MVDiff < 41) {
return 8;
}
else if (MVDiff >= 41) {
return 4;
}
}
else if (RsCsDiff >= 51) {
return 4;
}
}
else if (diffRsCsdiff >= 4) {
if (MVDiff < 27) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 4;
}
}
else if (MVDiff >= 27) {
if (tcor < 69) {
return 8;
}
else if (tcor >= 69) {
if (Rs < 911) {
return 1;
}
else if (Rs >= 911) {
return 4;
}
}
}
}
}
else if (refDCval >= 84) {
if (diffMVdiffVal < 26) {
if (RsCsDiff < 142) {
if (negBalance < 55) {
if (Rs < 1354) {
return 8;
}
else if (Rs >= 1354) {
return 1;
}
}
else if (negBalance >= 55) {
if (Rs < 344) {
return 1;
}
else if (Rs >= 344) {
if (gchDC < 1) {
if (Rs < 1107) {
return 2;
}
else if (Rs >= 1107) {
return 4;
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
else if (RsCsDiff >= 142) {
if (Rs < 1577) {
return 4;
}
else if (Rs >= 1577) {
return 8;
}
}
}
else if (diffMVdiffVal >= 26) {
if (Rs < 591) {
return 8;
}
else if (Rs >= 591) {
return 1;
}
}
}
}
else if (posBalance >= 61) {
if (tcor < 63) {
if (diffAFD < 3) {
if (tcor < 25) {
return 2;
}
else if (tcor >= 25) {
if (ssDCval < 119) {
return 4;
}
else if (ssDCval >= 119) {
if (Rs < 1500) {
if (Cs < 511) {
return 2;
}
else if (Cs >= 511) {
return 4;
}
}
else if (Rs >= 1500) {
return 4;
}
}
}
}
else if (diffAFD >= 3) {
if (Rs < 1072) {
return 1;
}
else if (Rs >= 1072) {
return 2;
}
}
}
else if (tcor >= 63) {
if (diffMVdiffVal < 5) {
return 1;
}
else if (diffMVdiffVal >= 5) {
if (Cs < 439) {
return 2;
}
else if (Cs >= 439) {
return 4;
}
}
}
}
}
}
else if (TSC >= 11) {
if (MVDiff < 211) {
if (tcor < 57) {
if (diffAFD < 18) {
return 1;
}
else if (diffAFD >= 18) {
return 8;
}
}
else if (tcor >= 57) {
if (Cs < 375) {
return 1;
}
else if (Cs >= 375) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
if (Rs < 651) {
return 4;
}
else if (Rs >= 651) {
return 8;
}
}
}
}
}
else if (MVDiff >= 211) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (Rs < 545) {
return 2;
}
else if (Rs >= 545) {
return 4;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 176) {
if (MVDiff < 34) {
if (diffMVdiffVal < -11) {
if (Rs < 2139) {
return 2;
}
else if (Rs >= 2139) {
if (Rs < 2418) {
return 1;
}
else if (Rs >= 2418) {
return 2;
}
}
}
else if (diffMVdiffVal >= -11) {
if (Cs < 519) {
return 2;
}
else if (Cs >= 519) {
return 4;
}
}
}
else if (MVDiff >= 34) {
if (AFD < 20) {
if (Rs < 1730) {
return 2;
}
else if (Rs >= 1730) {
return 8;
}
}
else if (AFD >= 20) {
if (tcor < 47) {
if (ssDCval < 67) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 2;
}
}
else if (ssDCval >= 67) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
if (Rs < 1618) {
return 8;
}
else if (Rs >= 1618) {
return 2;
}
}
}
}
else if (tcor >= 47) {
if (mu_mv_mag_sq < 61) {
if (AFD < 24) {
if (diffAFD < 1) {
if (diffMVdiffVal < 25) {
if (SC < 2419) {
return 1;
}
else if (SC >= 2419) {
return 8;
}
}
else if (diffMVdiffVal >= 25) {
return 2;
}
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (AFD >= 24) {
if (diffTSC < 8) {
if (posBalance < 66) {
if (MVDiff < 36) {
if (Rs < 1780) {
return 4;
}
else if (Rs >= 1780) {
return 1;
}
}
else if (MVDiff >= 36) {
return 1;
}
}
else if (posBalance >= 66) {
if (tcor < 59) {
return 2;
}
else if (tcor >= 59) {
if (diffRsCsdiff < 20) {
if (Rs < 1792) {
return 8;
}
else if (Rs >= 1792) {
if (Rs < 2090) {
return 2;
}
else if (Rs >= 2090) {
return 1;
}
}
}
else if (diffRsCsdiff >= 20) {
return 1;
}
}
}
}
else if (diffTSC >= 8) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (diffAFD < 12) {
if (diffMVdiffVal < -11) {
if (Rs < 1335) {
return 2;
}
else if (Rs >= 1335) {
return 1;
}
}
else if (diffMVdiffVal >= -11) {
return 2;
}
}
else if (diffAFD >= 12) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 80) {
if (RsDiff < 9) {
if (Cs < 135) {
if (refDCval < 79) {
return 1;
}
else if (refDCval >= 79) {
if (ssDCval < 140) {
if (SC < 804) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
else if (SC >= 804) {
if (MVDiff < 54) {
if (diffRsCsdiff < -30) {
if (Cs < 117) {
return 2;
}
else if (Cs >= 117) {
return 1;
}
}
else if (diffRsCsdiff >= -30) {
if (tcor < 82) {
if (Rs < 1427) {
return 2;
}
else if (Rs >= 1427) {
return 1;
}
}
else if (tcor >= 82) {
return 1;
}
}
}
else if (MVDiff >= 54) {
return 1;
}
}
}
else if (ssDCval >= 140) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 1;
}
}
}
}
else if (Cs >= 135) {
if (Rs < 781) {
if (Cs < 262) {
if (Rs < 533) {
if (diffMVdiffVal < 9) {
return 8;
}
else if (diffMVdiffVal >= 9) {
if (Rs < 295) {
return 4;
}
else if (Rs >= 295) {
return 1;
}
}
}
else if (Rs >= 533) {
if (MVDiff < 119) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (Cs < 190) {
return 8;
}
else if (Cs >= 190) {
return 1;
}
}
}
else if (MVDiff >= 119) {
if (Cs < 184) {
return 2;
}
else if (Cs >= 184) {
return 8;
}
}
}
}
else if (Cs >= 262) {
if (diffAFD < 1) {
if (mu_mv_mag_sq < 10) {
if (mu_mv_mag_sq < 3) {
if (diffTSC < -2) {
if (Rs < 718) {
return 1;
}
else if (Rs >= 718) {
return 2;
}
}
else if (diffTSC >= -2) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 18) {
if (tcor < 78) {
return 1;
}
else if (tcor >= 78) {
return 4;
}
}
else if (AFD >= 18) {
if (Rs < 657) {
if (Rs < 543) {
return 2;
}
else if (Rs >= 543) {
return 8;
}
}
else if (Rs >= 657) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (negBalance < 47) {
if (Rs < 481) {
return 2;
}
else if (Rs >= 481) {
if (MVDiff < 26) {
if (negBalance < 30) {
if (Rs < 742) {
return 2;
}
else if (Rs >= 742) {
return 1;
}
}
else if (negBalance >= 30) {
if (mu_mv_mag_sq < 55) {
return 1;
}
else if (mu_mv_mag_sq >= 55) {
return 8;
}
}
}
else if (MVDiff >= 26) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (ssDCval < 123) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Cs < 344) {
return 1;
}
else if (Cs >= 344) {
if (refDCval < 78) {
return 1;
}
else if (refDCval >= 78) {
return 8;
}
}
}
}
else if (ssDCval >= 123) {
if (SC < 1118) {
return 2;
}
else if (SC >= 1118) {
if (Rs < 666) {
return 8;
}
else if (Rs >= 666) {
return 1;
}
}
}
}
}
}
else if (diffAFD >= 1) {
if (ssDCval < 130) {
if (Cs < 336) {
if (Rs < 666) {
return 4;
}
else if (Rs >= 666) {
return 1;
}
}
else if (Cs >= 336) {
if (MVDiff < 7) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (MVDiff >= 7) {
if (mcTcor < 94) {
if (Rs < 620) {
if (Rs < 465) {
return 2;
}
else if (Rs >= 465) {
return 4;
}
}
else if (Rs >= 620) {
if (Rs < 676) {
return 1;
}
else if (Rs >= 676) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (Cs < 519) {
return 1;
}
else if (Cs >= 519) {
return 4;
}
}
}
}
}
else if (ssDCval >= 130) {
return 2;
}
}
}
}
else if (Rs >= 781) {
if (ssDCval < 139) {
if (Cs < 269) {
if (refDCval < 79) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
return 1;
}
}
}
else if (refDCval >= 79) {
if (MVDiff < 22) {
if (mu_mv_mag_sq < 10) {
if (tcor < 88) {
return 2;
}
else if (tcor >= 88) {
if (Cs < 161) {
if (Rs < 1226) {
return 1;
}
else if (Rs >= 1226) {
return 2;
}
}
else if (Cs >= 161) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (RsCsDiff < 67) {
return 1;
}
else if (RsCsDiff >= 67) {
return 2;
}
}
}
}
else if (MVDiff >= 22) {
if (Rs < 1203) {
if (posBalance < 64) {
if (TSC < 8) {
if (diffMVdiffVal < -1) {
if (Rs < 925) {
return 4;
}
else if (Rs >= 925) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (TSC >= 8) {
return 1;
}
}
else if (posBalance >= 64) {
return 1;
}
}
else if (Rs >= 1203) {
if (tcor < 87) {
if (mu_mv_mag_sq < 5) {
if (MVDiff < 62) {
if (Cs < 252) {
return 1;
}
else if (Cs >= 252) {
return 8;
}
}
else if (MVDiff >= 62) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
return 1;
}
}
else if (tcor >= 87) {
return 1;
}
}
}
}
}
else if (Cs >= 269) {
if (Cs < 458) {
if (AFD < 7) {
if (ssDCval < 97) {
if (negBalance < 29) {
if (ssDCval < 58) {
if (Rs < 947) {
return 1;
}
else if (Rs >= 947) {
return 4;
}
}
else if (ssDCval >= 58) {
return 2;
}
}
else if (negBalance >= 29) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (ssDCval >= 97) {
if (Rs < 2320) {
return 4;
}
else if (Rs >= 2320) {
return 1;
}
}
}
else if (AFD >= 7) {
if (refDCval < 56) {
if (SC < 2337) {
if (Cs < 362) {
return 1;
}
else if (Cs >= 362) {
if (MVDiff < 43) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (MVDiff >= 43) {
return 1;
}
}
}
else if (SC >= 2337) {
return 4;
}
}
else if (refDCval >= 56) {
if (TSC < 5) {
if (ssDCval < 97) {
if (ssDCval < 66) {
if (Cs < 372) {
return 1;
}
else if (Cs >= 372) {
return 8;
}
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (ssDCval >= 97) {
if (Cs < 295) {
return 1;
}
else if (Cs >= 295) {
return 8;
}
}
}
else if (TSC >= 5) {
if (Cs < 299) {
if (RsCsDiff < 32) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
return 1;
}
}
else if (RsCsDiff >= 32) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1573) {
return 2;
}
else if (Rs >= 1573) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
return 1;
}
}
}
else if (Cs >= 299) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
if (gchDC < 2) {
if (tcor < 78) {
return 2;
}
else if (tcor >= 78) {
if (Rs < 2277) {
if (negBalance < 56) {
return 1;
}
else if (negBalance >= 56) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 2;
}
}
}
else if (Rs >= 2277) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (RsCsDiff < 39) {
if (Rs < 1545) {
if (Cs < 348) {
if (Rs < 1053) {
return 8;
}
else if (Rs >= 1053) {
return 1;
}
}
else if (Cs >= 348) {
return 1;
}
}
else if (Rs >= 1545) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
if (MVDiff < 49) {
return 1;
}
else if (MVDiff >= 49) {
return 8;
}
}
}
}
else if (RsCsDiff >= 39) {
if (RsCsDiff < 56) {
if (ssDCval < 90) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
return 2;
}
}
else if (ssDCval >= 90) {
return 1;
}
}
else if (RsCsDiff >= 56) {
if (Rs < 920) {
return 8;
}
else if (Rs >= 920) {
if (negBalance < 63) {
return 1;
}
else if (negBalance >= 63) {
if (Cs < 384) {
return 8;
}
else if (Cs >= 384) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 458) {
if (ssDCval < 123) {
if (refDCval < 67) {
if (negBalance < 45) {
if (negBalance < 28) {
if (Rs < 1457) {
return 1;
}
else if (Rs >= 1457) {
return 4;
}
}
else if (negBalance >= 28) {
if (TSC < 6) {
if (MVDiff < 7) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
else if (MVDiff >= 7) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (TSC >= 6) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 1075) {
return 8;
}
else if (Rs >= 1075) {
return 4;
}
}
}
}
}
else if (negBalance >= 45) {
if (AFD < 11) {
if (Rs < 987) {
return 2;
}
else if (Rs >= 987) {
if (Rs < 1758) {
return 8;
}
else if (Rs >= 1758) {
return 2;
}
}
}
else if (AFD >= 11) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 4;
}
}
}
}
}
else if (refDCval >= 67) {
if (mcTcor < 88) {
return 4;
}
else if (mcTcor >= 88) {
if (refDCval < 91) {
if (diffRsCsdiff < -61) {
return 2;
}
else if (diffRsCsdiff >= -61) {
if (tcor < 97) {
if (negBalance < 58) {
if (diffRsCsdiff < 38) {
return 1;
}
else if (diffRsCsdiff >= 38) {
return 2;
}
}
else if (negBalance >= 58) {
if (Rs < 1933) {
return 2;
}
else if (Rs >= 1933) {
return 1;
}
}
}
else if (tcor >= 97) {
if (CsDiff < 3) {
if (Rs < 1368) {
return 1;
}
else if (Rs >= 1368) {
return 4;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
}
}
else if (refDCval >= 91) {
if (mu_mv_mag_sq < 37) {
if (tcor < 86) {
return 2;
}
else if (tcor >= 86) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
return 8;
}
}
}
}
}
else if (ssDCval >= 123) {
if (Rs < 1706) {
if (MVDiff < 5) {
if (Rs < 1366) {
return 1;
}
else if (Rs >= 1366) {
return 2;
}
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (Rs >= 1706) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 139) {
if (gchDC < 2) {
if (ssDCval < 146) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (Rs < 1479) {
return 2;
}
else if (Rs >= 1479) {
return 1;
}
}
}
}
else if (ssDCval >= 146) {
if (MVDiff < 83) {
return 1;
}
else if (MVDiff >= 83) {
if (Rs < 1381) {
return 2;
}
else if (Rs >= 1381) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (Rs < 2178) {
return 1;
}
else if (Rs >= 2178) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 109) {
if (MVDiff < 38) {
if (AFD < 12) {
if (ssDCval < 96) {
if (posBalance < 63) {
if (gchDC < 1) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (Cs < 459) {
return 2;
}
else if (Cs >= 459) {
return 1;
}
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (posBalance >= 63) {
return 2;
}
}
else if (ssDCval >= 96) {
return 2;
}
}
else if (AFD >= 12) {
if (ssDCval < 70) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (ssDCval >= 70) {
if (negBalance < 46) {
if (MVDiff < 24) {
if (AFD < 16) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
if (Cs < 480) {
return 2;
}
else if (Cs >= 480) {
return 1;
}
}
}
else if (AFD >= 16) {
return 1;
}
}
else if (MVDiff >= 24) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
return 2;
}
}
}
else if (negBalance >= 46) {
if (AFD < 21) {
if (diffMVdiffVal < -17) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 1;
}
}
else if (diffMVdiffVal >= -17) {
return 2;
}
}
else if (AFD >= 21) {
if (refDCval < 94) {
if (RsCsDiff < 150) {
return 1;
}
else if (RsCsDiff >= 150) {
return 2;
}
}
else if (refDCval >= 94) {
if (CsDiff < 5) {
if (Cs < 150) {
return 2;
}
else if (Cs >= 150) {
return 1;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 38) {
if (mcTcor < 99) {
if (gchDC < 16) {
if (negBalance < 57) {
if (AFD < 17) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (MVDiff < 66) {
return 1;
}
else if (MVDiff >= 66) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
return 4;
}
}
}
}
else if (AFD >= 17) {
if (diffMVdiffVal < 27) {
if (diffMVdiffVal < -5) {
return 1;
}
else if (diffMVdiffVal >= -5) {
if (SC < 2733) {
return 2;
}
else if (SC >= 2733) {
return 8;
}
}
}
else if (diffMVdiffVal >= 27) {
return 1;
}
}
}
else if (negBalance >= 57) {
if (diffTSC < 2) {
if (CsDiff < 2) {
if (Cs < 216) {
if (tcor < 75) {
if (Rs < 1552) {
return 1;
}
else if (Rs >= 1552) {
return 8;
}
}
else if (tcor >= 75) {
return 1;
}
}
else if (Cs >= 216) {
return 8;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (diffTSC >= 2) {
if (Rs < 1349) {
return 2;
}
else if (Rs >= 1349) {
return 8;
}
}
}
}
else if (gchDC >= 16) {
return 8;
}
}
else if (mcTcor >= 99) {
if (Rs < 2202) {
return 2;
}
else if (Rs >= 2202) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
}
else if (refDCval >= 109) {
if (CsDiff < 2) {
if (Rs < 1418) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 1;
}
}
else if (Rs >= 1418) {
return 2;
}
}
else if (CsDiff >= 2) {
if (MVDiff < 61) {
if (AFD < 19) {
if (tcor < 82) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (tcor >= 82) {
if (mu_mv_mag_sq < 35) {
if (TSC < 6) {
if (Rs < 1749) {
return 2;
}
else if (Rs >= 1749) {
return 1;
}
}
else if (TSC >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 35) {
if (ssDCval < 129) {
return 2;
}
else if (ssDCval >= 129) {
return 1;
}
}
}
}
else if (AFD >= 19) {
return 1;
}
}
else if (MVDiff >= 61) {
if (mu_mv_mag_sq < 39) {
if (AFD < 20) {
if (TSC < 10) {
if (ssDCval < 153) {
if (diffAFD < 0) {
return 1;
}
else if (diffAFD >= 0) {
if (Rs < 2082) {
return 1;
}
else if (Rs >= 2082) {
return 8;
}
}
}
else if (ssDCval >= 153) {
return 8;
}
}
else if (TSC >= 10) {
if (gchDC < 4) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
return 8;
}
}
else if (gchDC >= 4) {
if (Rs < 2088) {
return 2;
}
else if (Rs >= 2088) {
return 1;
}
}
}
}
else if (AFD >= 20) {
if (CsDiff < 3) {
if (Rs < 1996) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 1;
}
}
else if (Rs >= 1996) {
return 8;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (Rs < 1110) {
if (Cs < 211) {
return 1;
}
else if (Cs >= 211) {
return 2;
}
}
else if (Rs >= 1110) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (AFD < 8) {
if (ssDCval < 66) {
if (Rs < 1172) {
if (Rs < 924) {
return 2;
}
else if (Rs >= 924) {
return 4;
}
}
else if (Rs >= 1172) {
return 1;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (AFD >= 8) {
if (Rs < 721) {
if (Rs < 562) {
return 1;
}
else if (Rs >= 562) {
return 8;
}
}
else if (Rs >= 721) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Cs >= 533) {
if (negBalance < 14) {
if (Rs < 702) {
if (ssDCval < 58) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 29) {
return 8;
}
else if (ssDCval >= 29) {
if (Rs < 246) {
if (Rs < 223) {
return 1;
}
else if (Rs >= 223) {
return 2;
}
}
else if (Rs >= 246) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 351) {
if (Rs < 268) {
return 2;
}
else if (Rs >= 268) {
return 1;
}
}
else if (Rs >= 351) {
return 8;
}
}
}
else if (ssDCval >= 58) {
if (CsDiff < 1) {
if (Cs < 695) {
if (Rs < 519) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 1;
}
}
else if (Rs >= 519) {
return 8;
}
}
else if (Cs >= 695) {
if (Rs < 663) {
if (negBalance < 12) {
if (SC < 1246) {
return 4;
}
else if (SC >= 1246) {
return 8;
}
}
else if (negBalance >= 12) {
if (Rs < 595) {
return 8;
}
else if (Rs >= 595) {
return 2;
}
}
}
else if (Rs >= 663) {
return 1;
}
}
}
else if (CsDiff >= 1) {
if (MVDiff < 20) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
return 1;
}
}
else if (MVDiff >= 20) {
if (refDCval < 142) {
if (SC < 1453) {
if (Rs < 539) {
if (Cs < 564) {
return 1;
}
else if (Cs >= 564) {
return 4;
}
}
else if (Rs >= 539) {
return 8;
}
}
else if (SC >= 1453) {
return 8;
}
}
else if (refDCval >= 142) {
return 2;
}
}
}
}
}
else if (Rs >= 702) {
if (diffTSC < -2) {
if (Rs < 1606) {
if (tcor < 99) {
if (diffRsCsdiff < 4) {
if (ssDCval < 126) {
if (diffRsCsdiff < -12) {
if (Cs < 1049) {
return 1;
}
else if (Cs >= 1049) {
return 8;
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
else if (ssDCval >= 126) {
if (diffAFD < -16) {
if (Rs < 1347) {
return 8;
}
else if (Rs >= 1347) {
if (ssDCval < 151) {
return 4;
}
else if (ssDCval >= 151) {
return 8;
}
}
}
else if (diffAFD >= -16) {
return 8;
}
}
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
else if (tcor >= 99) {
if (ssDCval < 86) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 86) {
if (diffRsCsdiff < -19) {
if (ssDCval < 133) {
if (diffRsCsdiff < -48) {
if (diffAFD < -12) {
if (Cs < 1253) {
return 1;
}
else if (Cs >= 1253) {
return 8;
}
}
else if (diffAFD >= -12) {
return 8;
}
}
else if (diffRsCsdiff >= -48) {
return 1;
}
}
else if (ssDCval >= 133) {
if (ssDCval < 139) {
if (Cs < 1196) {
return 4;
}
else if (Cs >= 1196) {
return 1;
}
}
else if (ssDCval >= 139) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
if (Rs < 1394) {
if (ssDCval < 140) {
return 4;
}
else if (ssDCval >= 140) {
return 8;
}
}
else if (Rs >= 1394) {
return 4;
}
}
}
}
}
else if (diffRsCsdiff >= -19) {
if (Rs < 859) {
if (Rs < 814) {
return 8;
}
else if (Rs >= 814) {
return 2;
}
}
else if (Rs >= 859) {
if (diffAFD < -10) {
return 8;
}
else if (diffAFD >= -10) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1606) {
if (Cs < 1296) {
return 8;
}
else if (Cs >= 1296) {
if (diffMVdiffVal < -38) {
return 4;
}
else if (diffMVdiffVal >= -38) {
return 1;
}
}
}
}
else if (diffTSC >= -2) {
if (Rs < 1810) {
if (refDCval < 153) {
if (MVDiff < 2) {
if (Cs < 1059) {
if (negBalance < 9) {
if (diffMVdiffVal < -4) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
if (Rs < 1444) {
return 1;
}
else if (Rs >= 1444) {
return 8;
}
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (negBalance >= 9) {
if (posBalance < 37) {
if (Rs < 1651) {
if (refDCval < 137) {
if (mu_mv_mag_sq < 126) {
if (Cs < 724) {
if (refDCval < 117) {
if (Rs < 1060) {
if (Cs < 695) {
return 8;
}
else if (Cs >= 695) {
return 1;
}
}
else if (Rs >= 1060) {
return 1;
}
}
else if (refDCval >= 117) {
return 8;
}
}
else if (Cs >= 724) {
if (Rs < 1329) {
return 8;
}
else if (Rs >= 1329) {
if (Rs < 1447) {
return 1;
}
else if (Rs >= 1447) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 1113) {
return 8;
}
else if (Rs >= 1113) {
if (Rs < 1220) {
return 1;
}
else if (Rs >= 1220) {
return 8;
}
}
}
}
else if (refDCval >= 137) {
if (Rs < 885) {
return 1;
}
else if (Rs >= 885) {
return 8;
}
}
}
else if (Rs >= 1651) {
return 1;
}
}
else if (posBalance >= 37) {
if (SC < 1464) {
if (ssDCval < 81) {
return 2;
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (SC >= 1464) {
if (SC < 2002) {
return 8;
}
else if (SC >= 2002) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1059) {
if (negBalance < 12) {
if (negBalance < 1) {
if (Rs < 1247) {
return 1;
}
else if (Rs >= 1247) {
return 8;
}
}
else if (negBalance >= 1) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
return 1;
}
}
}
}
else if (negBalance >= 12) {
if (Rs < 1441) {
return 1;
}
else if (Rs >= 1441) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (Rs < 813) {
if (tcor < 98) {
if (tcor < 83) {
return 1;
}
else if (tcor >= 83) {
if (SC < 1340) {
return 1;
}
else if (SC >= 1340) {
return 8;
}
}
}
else if (tcor >= 98) {
if (ssDCval < 121) {
if (Rs < 709) {
return 1;
}
else if (Rs >= 709) {
return 2;
}
}
else if (ssDCval >= 121) {
return 8;
}
}
}
else if (Rs >= 813) {
if (negBalance < 5) {
return 8;
}
else if (negBalance >= 5) {
if (SC < 1799) {
if (mu_mv_mag_sq < 82) {
return 1;
}
else if (mu_mv_mag_sq >= 82) {
return 8;
}
}
else if (SC >= 1799) {
if (Rs < 1551) {
if (posBalance < 40) {
if (ssDCval < 32) {
return 8;
}
else if (ssDCval >= 32) {
if (diffRsCsdiff < 1) {
if (posBalance < 27) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
else if (posBalance >= 27) {
if (mu_mv_mag_sq < 92) {
return 8;
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
}
else if (diffRsCsdiff >= 1) {
return 1;
}
}
}
else if (posBalance >= 40) {
if (Cs < 812) {
return 4;
}
else if (Cs >= 812) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
}
else if (Rs >= 1551) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 153) {
if (Rs < 836) {
return 8;
}
else if (Rs >= 836) {
return 2;
}
}
}
else if (Rs >= 1810) {
if (negBalance < 8) {
if (posBalance < 27) {
if (mcTcor < 88) {
return 4;
}
else if (mcTcor >= 88) {
return 8;
}
}
else if (posBalance >= 27) {
if (mu_mv_mag_sq < 126) {
if (mu_mv_mag_sq < 98) {
if (Rs < 2077) {
return 1;
}
else if (Rs >= 2077) {
if (Rs < 2234) {
return 8;
}
else if (Rs >= 2234) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 98) {
return 8;
}
}
else if (mu_mv_mag_sq >= 126) {
if (SC < 2974) {
if (Rs < 1949) {
return 1;
}
else if (Rs >= 1949) {
return 8;
}
}
else if (SC >= 2974) {
return 4;
}
}
}
}
else if (negBalance >= 8) {
if (Rs < 2062) {
if (Cs < 989) {
if (RsDiff < 2) {
if (SC < 2952) {
return 1;
}
else if (SC >= 2952) {
if (Rs < 2009) {
return 1;
}
else if (Rs >= 2009) {
return 4;
}
}
}
else if (RsDiff >= 2) {
if (Cs < 749) {
return 2;
}
else if (Cs >= 749) {
return 8;
}
}
}
else if (Cs >= 989) {
if (refDCval < 68) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (tcor < 99) {
if (Rs < 1894) {
return 1;
}
else if (Rs >= 1894) {
return 4;
}
}
else if (tcor >= 99) {
return 4;
}
}
}
else if (refDCval >= 68) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (Rs >= 2062) {
if (Cs < 1292) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
return 1;
}
}
else if (Cs >= 1292) {
if (Rs < 2281) {
return 8;
}
else if (Rs >= 2281) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (mcTcor < 97) {
if (RsCsDiff < 63) {
if (posBalance < 41) {
if (refDCval < 49) {
if (TSCindex < 3) {
if (Rs < 1786) {
return 8;
}
else if (Rs >= 1786) {
return 4;
}
}
else if (TSCindex >= 3) {
if (Rs < 1087) {
if (ssDCval < 29) {
return 4;
}
else if (ssDCval >= 29) {
if (negBalance < 21) {
if (Cs < 604) {
return 2;
}
else if (Cs >= 604) {
return 1;
}
}
else if (negBalance >= 21) {
if (Cs < 558) {
if (Rs < 523) {
return 4;
}
else if (Rs >= 523) {
return 2;
}
}
else if (Cs >= 558) {
if (CsDiff < 5) {
if (Rs < 665) {
if (Rs < 613) {
return 8;
}
else if (Rs >= 613) {
return 4;
}
}
else if (Rs >= 665) {
return 8;
}
}
else if (CsDiff >= 5) {
return 1;
}
}
}
}
}
else if (Rs >= 1087) {
if (ssDCval < 46) {
if (posBalance < 24) {
return 1;
}
else if (posBalance >= 24) {
return 4;
}
}
else if (ssDCval >= 46) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Rs < 1673) {
return 1;
}
else if (Rs >= 1673) {
if (Cs < 1144) {
return 4;
}
else if (Cs >= 1144) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 49) {
if (refDCval < 92) {
if (mu_mv_mag_sq < 120) {
if (negBalance < 55) {
if (Rs < 784) {
if (ssDCval < 77) {
if (TSC < 5) {
if (RsCsDiff < 5) {
if (Rs < 370) {
return 2;
}
else if (Rs >= 370) {
return 1;
}
}
else if (RsCsDiff >= 5) {
return 4;
}
}
else if (TSC >= 5) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
if (Rs < 631) {
return 1;
}
else if (Rs >= 631) {
return 8;
}
}
}
}
else if (ssDCval >= 77) {
return 2;
}
}
else if (Rs >= 784) {
if (mu_mv_mag_sq < 72) {
if (Cs < 974) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (diffMVdiffVal < 4) {
return 1;
}
else if (diffMVdiffVal >= 4) {
if (SC < 2300) {
if (Rs < 1170) {
return 1;
}
else if (Rs >= 1170) {
return 2;
}
}
else if (SC >= 2300) {
return 1;
}
}
}
}
else if (Cs >= 974) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (Cs < 1083) {
return 8;
}
else if (Cs >= 1083) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (ssDCval < 59) {
if (negBalance < 34) {
if (mcTcor < 91) {
if (Rs < 1334) {
return 4;
}
else if (Rs >= 1334) {
return 8;
}
}
else if (mcTcor >= 91) {
if (TSC < 3) {
if (Rs < 1200) {
return 1;
}
else if (Rs >= 1200) {
return 8;
}
}
else if (TSC >= 3) {
if (diffRsCsdiff < -26) {
return 8;
}
else if (diffRsCsdiff >= -26) {
return 1;
}
}
}
}
else if (negBalance >= 34) {
if (Rs < 1344) {
return 4;
}
else if (Rs >= 1344) {
return 8;
}
}
}
else if (ssDCval >= 59) {
if (MVDiff < 15) {
if (Rs < 1629) {
return 8;
}
else if (Rs >= 1629) {
if (posBalance < 33) {
if (Cs < 746) {
if (Rs < 2010) {
return 8;
}
else if (Rs >= 2010) {
if (Rs < 2271) {
return 4;
}
else if (Rs >= 2271) {
return 4;
}
}
}
else if (Cs >= 746) {
return 4;
}
}
else if (posBalance >= 33) {
if (RsCsDiff < 30) {
return 1;
}
else if (RsCsDiff >= 30) {
return 2;
}
}
}
}
else if (MVDiff >= 15) {
return 4;
}
}
}
}
}
else if (negBalance >= 55) {
if (Rs < 409) {
if (TSC < 7) {
if (Rs < 283) {
return 4;
}
else if (Rs >= 283) {
return 1;
}
}
else if (TSC >= 7) {
if (Rs < 331) {
return 8;
}
else if (Rs >= 331) {
return 2;
}
}
}
else if (Rs >= 409) {
if (RsCsDiff < 43) {
if (diffMVdiffVal < -1) {
if (ssDCval < 67) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (RsCsDiff >= 43) {
if (tcor < 99) {
if (Cs < 553) {
return 1;
}
else if (Cs >= 553) {
return 8;
}
}
else if (tcor >= 99) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 120) {
return 8;
}
}
else if (refDCval >= 92) {
if (negBalance < 32) {
if (Cs < 743) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
else if (Cs >= 743) {
if (refDCval < 139) {
if (Cs < 1143) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
return 4;
}
}
else if (Cs >= 1143) {
if (mu_mv_mag_sq < 89) {
return 4;
}
else if (mu_mv_mag_sq >= 89) {
return 8;
}
}
}
else if (refDCval >= 139) {
return 8;
}
}
}
else if (negBalance >= 32) {
if (tcor < 93) {
if (ssDCval < 147) {
if (MVDiff < 85) {
if (CsDiff < 4) {
if (diffMVdiffVal < 5) {
if (Rs < 1642) {
return 2;
}
else if (Rs >= 1642) {
return 1;
}
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
else if (CsDiff >= 4) {
if (diffRsCsdiff < 12) {
return 8;
}
else if (diffRsCsdiff >= 12) {
return 2;
}
}
}
else if (MVDiff >= 85) {
if (diffAFD < -3) {
if (Rs < 450) {
return 8;
}
else if (Rs >= 450) {
return 1;
}
}
else if (diffAFD >= -3) {
if (Rs < 695) {
return 2;
}
else if (Rs >= 695) {
return 4;
}
}
}
}
else if (ssDCval >= 147) {
return 2;
}
}
else if (tcor >= 93) {
return 8;
}
}
}
}
}
else if (posBalance >= 41) {
if (AFD < 15) {
if (mcTcor < 91) {
if (Rs < 1297) {
if (refDCval < 48) {
if (ssDCval < 45) {
if (Rs < 768) {
if (posBalance < 56) {
if (Cs < 890) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Cs < 646) {
return 2;
}
else if (Cs >= 646) {
return 8;
}
}
}
else if (Cs >= 890) {
return 1;
}
}
else if (posBalance >= 56) {
if (negBalance < 49) {
if (mu_mv_mag_sq < 36) {
if (Cs < 615) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
if (Rs < 295) {
return 1;
}
else if (Rs >= 295) {
return 2;
}
}
}
else if (Cs >= 615) {
if (Cs < 1152) {
return 2;
}
else if (Cs >= 1152) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 36) {
return 8;
}
}
else if (negBalance >= 49) {
if (mu_mv_mag_sq < 6) {
return 4;
}
else if (mu_mv_mag_sq >= 6) {
return 8;
}
}
}
}
else if (Rs >= 768) {
if (mcTcor < 88) {
if (AFD < 13) {
if (mu_mv_mag_sq < 118) {
if (SC < 1716) {
if (Cs < 819) {
if (Rs < 1074) {
return 8;
}
else if (Rs >= 1074) {
return 2;
}
}
else if (Cs >= 819) {
if (mu_mv_mag_sq < 36) {
return 2;
}
else if (mu_mv_mag_sq >= 36) {
return 8;
}
}
}
else if (SC >= 1716) {
return 8;
}
}
else if (mu_mv_mag_sq >= 118) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (AFD >= 13) {
return 2;
}
}
else if (mcTcor >= 88) {
if (SC < 1794) {
if (Rs < 1023) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (negBalance < 35) {
if (Rs < 941) {
return 2;
}
else if (Rs >= 941) {
return 8;
}
}
else if (negBalance >= 35) {
return 2;
}
}
}
else if (Rs >= 1023) {
if (Rs < 1114) {
return 1;
}
else if (Rs >= 1114) {
return 2;
}
}
}
else if (SC >= 1794) {
if (Rs < 860) {
return 2;
}
else if (Rs >= 860) {
return 8;
}
}
}
}
}
else if (ssDCval >= 45) {
if (TSC < 7) {
if (tcor < 87) {
if (posBalance < 55) {
if (Rs < 1117) {
if (negBalance < 46) {
if (Rs < 760) {
return 4;
}
else if (Rs >= 760) {
if (AFD < 8) {
if (Cs < 907) {
return 1;
}
else if (Cs >= 907) {
return 8;
}
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (negBalance >= 46) {
return 2;
}
}
else if (Rs >= 1117) {
return 1;
}
}
else if (posBalance >= 55) {
if (negBalance < 49) {
if (AFD < 7) {
if (posBalance < 57) {
if (Cs < 717) {
return 8;
}
else if (Cs >= 717) {
return 1;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (AFD >= 7) {
if (diffRsCsdiff < -10) {
if (Cs < 751) {
return 2;
}
else if (Cs >= 751) {
return 8;
}
}
else if (diffRsCsdiff >= -10) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (tcor < 60) {
if (Cs < 546) {
return 4;
}
else if (Cs >= 546) {
return 8;
}
}
else if (tcor >= 60) {
if (Cs < 913) {
if (negBalance < 54) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (negBalance >= 54) {
return 8;
}
}
else if (Cs >= 913) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
}
}
else if (tcor >= 87) {
return 8;
}
}
else if (TSC >= 7) {
if (tcor < 56) {
return 4;
}
else if (tcor >= 56) {
if (MVDiff < 14) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
else if (MVDiff >= 14) {
if (ssDCval < 46) {
if (Rs < 781) {
return 8;
}
else if (Rs >= 781) {
return 1;
}
}
else if (ssDCval >= 46) {
if (RsCsDiff < 53) {
return 8;
}
else if (RsCsDiff >= 53) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 48) {
if (ssDCval < 63) {
if (refDCval < 52) {
if (AFD < 11) {
if (CsDiff < 5) {
if (RsDiff < 3) {
if (SC < 1487) {
return 1;
}
else if (SC >= 1487) {
return 8;
}
}
else if (RsDiff >= 3) {
if (diffRsCsdiff < -13) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (diffRsCsdiff >= -13) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (diffMVdiffVal < 4) {
if (Cs < 1308) {
return 8;
}
else if (Cs >= 1308) {
return 4;
}
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
}
else if (AFD >= 11) {
if (diffRsCsdiff < -8) {
if (Rs < 740) {
if (Rs < 539) {
return 2;
}
else if (Rs >= 539) {
return 8;
}
}
else if (Rs >= 740) {
if (TSC < 10) {
if (SC < 1527) {
if (Rs < 791) {
return 4;
}
else if (Rs >= 791) {
return 1;
}
}
else if (SC >= 1527) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (diffRsCsdiff >= -8) {
if (diffMVdiffVal < 2) {
if (MVDiff < 13) {
if (Cs < 1050) {
return 8;
}
else if (Cs >= 1050) {
return 4;
}
}
else if (MVDiff >= 13) {
if (Cs < 788) {
return 4;
}
else if (Cs >= 788) {
return 1;
}
}
}
else if (diffMVdiffVal >= 2) {
if (RsCsDiff < 57) {
if (diffRsCsdiff < -1) {
if (Rs < 619) {
return 8;
}
else if (Rs >= 619) {
return 4;
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
else if (RsCsDiff >= 57) {
if (Rs < 908) {
return 8;
}
else if (Rs >= 908) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 52) {
if (tcor < 75) {
if (MVDiff < 8) {
if (Cs < 1237) {
if (Cs < 550) {
return 8;
}
else if (Cs >= 550) {
return 4;
}
}
else if (Cs >= 1237) {
return 8;
}
}
else if (MVDiff >= 8) {
if (negBalance < 51) {
if (refDCval < 55) {
if (tcor < 63) {
if (Rs < 948) {
return 1;
}
else if (Rs >= 948) {
return 8;
}
}
else if (tcor >= 63) {
return 4;
}
}
else if (refDCval >= 55) {
if (mcTcor < 80) {
if (Rs < 948) {
return 2;
}
else if (Rs >= 948) {
return 4;
}
}
else if (mcTcor >= 80) {
return 1;
}
}
}
else if (negBalance >= 51) {
if (mcTcor < 85) {
if (SC < 1080) {
return 8;
}
else if (SC >= 1080) {
return 4;
}
}
else if (mcTcor >= 85) {
if (RsCsDiff < 17) {
return 4;
}
else if (RsCsDiff >= 17) {
return 8;
}
}
}
}
}
else if (tcor >= 75) {
if (Rs < 587) {
if (Rs < 574) {
return 1;
}
else if (Rs >= 574) {
return 2;
}
}
else if (Rs >= 587) {
if (Rs < 1162) {
if (Cs < 609) {
if (posBalance < 51) {
if (Rs < 926) {
return 2;
}
else if (Rs >= 926) {
return 8;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (Cs >= 609) {
if (AFD < 10) {
if (negBalance < 34) {
if (SC < 1695) {
if (Rs < 870) {
return 1;
}
else if (Rs >= 870) {
return 2;
}
}
else if (SC >= 1695) {
return 8;
}
}
else if (negBalance >= 34) {
if (MVDiff < 19) {
if (MVDiff < 7) {
if (mu_mv_mag_sq < 72) {
return 8;
}
else if (mu_mv_mag_sq >= 72) {
if (refDCval < 56) {
return 4;
}
else if (refDCval >= 56) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
return 4;
}
}
}
}
else if (MVDiff >= 7) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
}
else if (MVDiff >= 19) {
return 8;
}
}
}
else if (AFD >= 10) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 1;
}
}
}
}
else if (Rs >= 1162) {
if (mu_mv_mag_sq < 46) {
return 8;
}
else if (mu_mv_mag_sq >= 46) {
if (diffMVdiffVal < 3) {
if (Cs < 1173) {
if (SC < 1871) {
return 1;
}
else if (SC >= 1871) {
return 8;
}
}
else if (Cs >= 1173) {
return 1;
}
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 63) {
if (ssDCval < 116) {
if (refDCval < 74) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 1134) {
if (negBalance < 38) {
if (Cs < 1098) {
if (Cs < 647) {
return 1;
}
else if (Cs >= 647) {
return 8;
}
}
else if (Cs >= 1098) {
return 1;
}
}
else if (negBalance >= 38) {
if (RsDiff < 5) {
if (diffMVdiffVal < 5) {
if (SC < 1219) {
if (Rs < 498) {
return 1;
}
else if (Rs >= 498) {
return 8;
}
}
else if (SC >= 1219) {
if (refDCval < 71) {
return 4;
}
else if (refDCval >= 71) {
if (Rs < 583) {
return 4;
}
else if (Rs >= 583) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 5) {
if (Rs < 800) {
return 2;
}
else if (Rs >= 800) {
return 4;
}
}
}
else if (RsDiff >= 5) {
if (Rs < 1101) {
return 1;
}
else if (Rs >= 1101) {
return 2;
}
}
}
}
else if (Rs >= 1134) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
return 8;
}
}
}
}
else if (refDCval >= 74) {
if (RsDiff < 6) {
if (posBalance < 45) {
if (ssDCval < 91) {
if (Cs < 859) {
return 8;
}
else if (Cs >= 859) {
return 4;
}
}
else if (ssDCval >= 91) {
return 2;
}
}
else if (posBalance >= 45) {
if (MVDiff < 19) {
if (mcTcor < 80) {
if (Rs < 871) {
return 2;
}
else if (Rs >= 871) {
return 8;
}
}
else if (mcTcor >= 80) {
return 8;
}
}
else if (MVDiff >= 19) {
if (TSC < 6) {
if (Cs < 588) {
return 8;
}
else if (Cs >= 588) {
return 2;
}
}
else if (TSC >= 6) {
if (Cs < 547) {
return 2;
}
else if (Cs >= 547) {
if (SC < 1998) {
return 8;
}
else if (SC >= 1998) {
if (Rs < 1155) {
return 4;
}
else if (Rs >= 1155) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 6) {
if (Rs < 1207) {
return 4;
}
else if (Rs >= 1207) {
return 1;
}
}
}
}
else if (ssDCval >= 116) {
if (tcor < 75) {
if (negBalance < 27) {
return 2;
}
else if (negBalance >= 27) {
if (Rs < 1262) {
return 1;
}
else if (Rs >= 1262) {
return 4;
}
}
}
else if (tcor >= 75) {
if (SC < 1984) {
if (Rs < 812) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
else if (Rs >= 812) {
if (mcTcor < 88) {
if (Rs < 1043) {
return 8;
}
else if (Rs >= 1043) {
return 2;
}
}
else if (mcTcor >= 88) {
if (diffRsCsdiff < 11) {
return 4;
}
else if (diffRsCsdiff >= 11) {
return 8;
}
}
}
}
else if (SC >= 1984) {
if (ssDCval < 156) {
if (SC < 2391) {
return 8;
}
else if (SC >= 2391) {
if (Rs < 1221) {
return 1;
}
else if (Rs >= 1221) {
return 8;
}
}
}
else if (ssDCval >= 156) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 1297) {
if (posBalance < 55) {
if (Rs < 1553) {
if (ssDCval < 158) {
if (SC < 2067) {
if (Rs < 1334) {
return 4;
}
else if (Rs >= 1334) {
return 8;
}
}
else if (SC >= 2067) {
if (Cs < 1297) {
if (negBalance < 27) {
return 2;
}
else if (negBalance >= 27) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (refDCval < 46) {
return 2;
}
else if (refDCval >= 46) {
return 8;
}
}
}
}
else if (Cs >= 1297) {
return 1;
}
}
}
else if (ssDCval >= 158) {
return 2;
}
}
else if (Rs >= 1553) {
if (AFD < 8) {
if (SC < 2571) {
return 4;
}
else if (SC >= 2571) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (AFD >= 8) {
if (Rs < 1850) {
if (mu_mv_mag_sq < 39) {
if (TSC < 7) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
return 8;
}
}
else if (TSC >= 7) {
if (Cs < 651) {
return 1;
}
else if (Cs >= 651) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (MVDiff < 14) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 2;
}
}
else if (MVDiff >= 14) {
return 8;
}
}
}
else if (Rs >= 1850) {
if (ssDCval < 92) {
return 4;
}
else if (ssDCval >= 92) {
return 8;
}
}
}
}
}
else if (posBalance >= 55) {
if (negBalance < 44) {
if (TSC < 8) {
if (TSC < 4) {
if (Rs < 1309) {
return 8;
}
else if (Rs >= 1309) {
return 4;
}
}
else if (TSC >= 4) {
if (diffRsCsdiff < 10) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (tcor < 82) {
return 8;
}
else if (tcor >= 82) {
if (Rs < 1474) {
return 1;
}
else if (Rs >= 1474) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 10) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Cs < 1150) {
return 2;
}
else if (Cs >= 1150) {
return 8;
}
}
}
}
}
else if (TSC >= 8) {
if (posBalance < 66) {
return 1;
}
else if (posBalance >= 66) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
}
}
else if (negBalance >= 44) {
if (AFD < 11) {
if (mu_mv_mag_sq < 6) {
if (CsDiff < 4) {
if (negBalance < 51) {
if (Cs < 612) {
return 2;
}
else if (Cs >= 612) {
return 8;
}
}
else if (negBalance >= 51) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (diffMVdiffVal < 7) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 1451) {
return 8;
}
else if (Rs >= 1451) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 7) {
return 2;
}
}
}
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 6) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 1538) {
return 8;
}
else if (Rs >= 1538) {
return 2;
}
}
}
}
else if (AFD >= 11) {
if (RsCsDiff < 51) {
if (negBalance < 51) {
if (MVDiff < 3) {
if (diffRsCsdiff < -8) {
if (Cs < 1004) {
return 8;
}
else if (Cs >= 1004) {
return 4;
}
}
else if (diffRsCsdiff >= -8) {
return 8;
}
}
else if (MVDiff >= 3) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
return 8;
}
}
}
}
else if (negBalance >= 51) {
if (diffRsCsdiff < -21) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
else if (diffRsCsdiff >= -21) {
return 8;
}
}
}
else if (RsCsDiff >= 51) {
if (ssDCval < 91) {
if (Cs < 576) {
return 4;
}
else if (Cs >= 576) {
return 8;
}
}
else if (ssDCval >= 91) {
if (Rs < 1568) {
return 8;
}
else if (Rs >= 1568) {
if (Rs < 2067) {
return 4;
}
else if (Rs >= 2067) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (RsCsDiff < 23) {
if (Rs < 876) {
if (refDCval < 75) {
if (Rs < 511) {
if (Cs < 781) {
if (tcor < 89) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (ssDCval >= 56) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
return 1;
}
}
else if (tcor >= 89) {
if (Rs < 450) {
if (refDCval < 65) {
if (Rs < 315) {
return 2;
}
else if (Rs >= 315) {
if (Cs < 606) {
return 2;
}
else if (Cs >= 606) {
return 8;
}
}
}
else if (refDCval >= 65) {
if (Rs < 403) {
return 4;
}
else if (Rs >= 403) {
return 8;
}
}
}
else if (Rs >= 450) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
if (Rs < 503) {
return 4;
}
else if (Rs >= 503) {
return 8;
}
}
}
}
}
else if (Cs >= 781) {
if (ssDCval < 53) {
if (ssDCval < 51) {
if (Rs < 503) {
return 1;
}
else if (Rs >= 503) {
return 8;
}
}
else if (ssDCval >= 51) {
return 2;
}
}
else if (ssDCval >= 53) {
return 1;
}
}
}
else if (Rs >= 511) {
if (negBalance < 31) {
if (ssDCval < 62) {
if (refDCval < 51) {
if (Rs < 701) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 656) {
return 1;
}
else if (Rs >= 656) {
return 4;
}
}
}
else if (Rs >= 701) {
if (RsCsDiff < 11) {
return 8;
}
else if (RsCsDiff >= 11) {
return 4;
}
}
}
else if (refDCval >= 51) {
if (Rs < 796) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
if (Rs < 705) {
return 8;
}
else if (Rs >= 705) {
return 4;
}
}
}
else if (Rs >= 796) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (ssDCval >= 62) {
if (Cs < 735) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (Cs >= 735) {
if (RsDiff < 2) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
else if (negBalance >= 31) {
if (mu_mv_mag_sq < 121) {
if (refDCval < 69) {
if (ssDCval < 67) {
if (mu_mv_mag_sq < 50) {
if (TSC < 4) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
return 2;
}
}
else if (TSC >= 4) {
if (tcor < 90) {
if (refDCval < 42) {
if (Cs < 621) {
return 2;
}
else if (Cs >= 621) {
return 4;
}
}
else if (refDCval >= 42) {
if (MVDiff < 30) {
if (TSC < 6) {
if (AFD < 7) {
if (Cs < 624) {
return 1;
}
else if (Cs >= 624) {
return 8;
}
}
else if (AFD >= 7) {
return 8;
}
}
else if (TSC >= 6) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (MVDiff >= 30) {
if (Rs < 571) {
return 4;
}
else if (Rs >= 571) {
return 1;
}
}
}
}
else if (tcor >= 90) {
if (negBalance < 36) {
return 4;
}
else if (negBalance >= 36) {
if (MVDiff < 5) {
if (SC < 1596) {
return 4;
}
else if (SC >= 1596) {
return 8;
}
}
else if (MVDiff >= 5) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (ssDCval < 56) {
if (Rs < 817) {
if (Rs < 733) {
if (posBalance < 52) {
if (Rs < 670) {
return 1;
}
else if (Rs >= 670) {
return 8;
}
}
else if (posBalance >= 52) {
return 4;
}
}
else if (Rs >= 733) {
return 4;
}
}
else if (Rs >= 817) {
return 8;
}
}
else if (ssDCval >= 56) {
if (AFD < 4) {
if (ssDCval < 59) {
if (Rs < 625) {
return 2;
}
else if (Rs >= 625) {
return 8;
}
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (AFD >= 4) {
if (tcor < 92) {
if (Cs < 741) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
return 1;
}
}
else if (Cs >= 741) {
if (Rs < 692) {
return 1;
}
else if (Rs >= 692) {
return 4;
}
}
}
else if (tcor >= 92) {
return 8;
}
}
}
}
}
else if (ssDCval >= 67) {
if (Cs < 1204) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (SC < 1401) {
return 2;
}
else if (SC >= 1401) {
return 1;
}
}
}
else if (Cs >= 1204) {
return 8;
}
}
}
else if (refDCval >= 69) {
if (mu_mv_mag_sq < 78) {
if (AFD < 10) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 2;
}
}
}
else if (AFD >= 10) {
if (Rs < 720) {
return 8;
}
else if (Rs >= 720) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (ssDCval < 45) {
return 1;
}
else if (ssDCval >= 45) {
if (Cs < 829) {
return 8;
}
else if (Cs >= 829) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 75) {
if (MVDiff < 3) {
if (Rs < 503) {
if (Cs < 757) {
return 1;
}
else if (Cs >= 757) {
if (TSC < 6) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 1;
}
}
else if (TSC >= 6) {
if (Rs < 435) {
return 4;
}
else if (Rs >= 435) {
return 1;
}
}
}
}
else if (Rs >= 503) {
if (refDCval < 98) {
if (negBalance < 43) {
if (posBalance < 55) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (posBalance >= 55) {
if (negBalance < 28) {
return 2;
}
else if (negBalance >= 28) {
if (Cs < 1060) {
return 8;
}
else if (Cs >= 1060) {
return 4;
}
}
}
}
else if (negBalance >= 43) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (Cs < 936) {
return 1;
}
else if (Cs >= 936) {
return 4;
}
}
}
}
else if (refDCval >= 98) {
if (mu_mv_mag_sq < 87) {
if (Cs < 546) {
return 2;
}
else if (Cs >= 546) {
return 1;
}
}
else if (mu_mv_mag_sq >= 87) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
}
else if (MVDiff >= 3) {
if (SC < 1710) {
if (AFD < 9) {
if (TSC < 5) {
if (refDCval < 119) {
if (Rs < 611) {
if (TSC < 4) {
if (Cs < 835) {
return 2;
}
else if (Cs >= 835) {
if (Rs < 484) {
return 8;
}
else if (Rs >= 484) {
return 4;
}
}
}
else if (TSC >= 4) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (Rs >= 611) {
if (ssDCval < 94) {
if (mu_mv_mag_sq < 75) {
return 2;
}
else if (mu_mv_mag_sq >= 75) {
return 1;
}
}
else if (ssDCval >= 94) {
return 8;
}
}
}
else if (refDCval >= 119) {
return 8;
}
}
else if (TSC >= 5) {
if (RsCsDiff < 4) {
if (Rs < 443) {
return 1;
}
else if (Rs >= 443) {
return 8;
}
}
else if (RsCsDiff >= 4) {
if (Rs < 365) {
if (Rs < 360) {
return 8;
}
else if (Rs >= 360) {
return 4;
}
}
else if (Rs >= 365) {
return 8;
}
}
}
}
else if (AFD >= 9) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (SC < 1246) {
if (Rs < 472) {
return 1;
}
else if (Rs >= 472) {
return 8;
}
}
else if (SC >= 1246) {
if (RsDiff < 3) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 4;
}
}
else if (RsDiff >= 3) {
if (Rs < 826) {
return 2;
}
else if (Rs >= 826) {
return 1;
}
}
}
}
}
}
else if (SC >= 1710) {
if (posBalance < 50) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 64) {
if (Rs < 723) {
return 8;
}
else if (Rs >= 723) {
return 4;
}
}
else if (mu_mv_mag_sq >= 64) {
return 8;
}
}
}
}
}
}
else if (Rs >= 876) {
if (refDCval < 138) {
if (RsCsDiff < 7) {
if (ssDCval < 65) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
if (CsDiff < 2) {
if (diffRsCsdiff < -3) {
if (SC < 1637) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (SC >= 1637) {
if (Cs < 1135) {
return 8;
}
else if (Cs >= 1135) {
if (Rs < 1148) {
return 8;
}
else if (Rs >= 1148) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= -3) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (Cs < 1105) {
return 8;
}
else if (Cs >= 1105) {
if (Rs < 1114) {
return 8;
}
else if (Rs >= 1114) {
return 1;
}
}
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 101) {
return 8;
}
else if (mu_mv_mag_sq >= 101) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (Rs < 1049) {
return 8;
}
else if (Rs >= 1049) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 65) {
if (Cs < 819) {
if (mu_mv_mag_sq < 75) {
if (Cs < 665) {
return 1;
}
else if (Cs >= 665) {
return 8;
}
}
else if (mu_mv_mag_sq >= 75) {
if (MVDiff < 4) {
if (Rs < 1156) {
return 4;
}
else if (Rs >= 1156) {
return 8;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (Cs >= 819) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (ssDCval < 78) {
if (negBalance < 46) {
if (refDCval < 63) {
if (MVDiff < 2) {
if (Rs < 1791) {
if (refDCval < 47) {
if (Rs < 1354) {
return 8;
}
else if (Rs >= 1354) {
if (Rs < 1438) {
return 4;
}
else if (Rs >= 1438) {
return 4;
}
}
}
else if (refDCval >= 47) {
if (Cs < 1279) {
return 8;
}
else if (Cs >= 1279) {
if (Cs < 1285) {
return 4;
}
else if (Cs >= 1285) {
return 8;
}
}
}
}
else if (Rs >= 1791) {
if (ssDCval < 53) {
if (Rs < 2141) {
return 4;
}
else if (Rs >= 2141) {
return 8;
}
}
else if (ssDCval >= 53) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (posBalance < 45) {
if (mu_mv_mag_sq < 59) {
if (mcTcor < 94) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
else if (mcTcor >= 94) {
if (Rs < 1252) {
return 1;
}
else if (Rs >= 1252) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (diffRsCsdiff < -6) {
return 4;
}
else if (diffRsCsdiff >= -6) {
if (ssDCval < 53) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (Rs < 1190) {
return 1;
}
else if (Rs >= 1190) {
return 2;
}
}
}
else if (ssDCval >= 53) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
}
else if (posBalance >= 45) {
if (ssDCval < 52) {
if (negBalance < 30) {
if (Rs < 1227) {
return 4;
}
else if (Rs >= 1227) {
return 2;
}
}
else if (negBalance >= 30) {
if (diffRsCsdiff < -18) {
return 2;
}
else if (diffRsCsdiff >= -18) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (diffTSC < -1) {
if (Cs < 803) {
if (Cs < 762) {
return 4;
}
else if (Cs >= 762) {
return 2;
}
}
else if (Cs >= 803) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
return 4;
}
}
}
else if (diffTSC >= -1) {
if (SC < 2413) {
if (MVDiff < 13) {
if (Rs < 1182) {
if (mu_mv_mag_sq < 54) {
return 1;
}
else if (mu_mv_mag_sq >= 54) {
if (diffMVdiffVal < 1) {
if (ssDCval < 62) {
if (Cs < 1196) {
return 1;
}
else if (Cs >= 1196) {
return 8;
}
}
else if (ssDCval >= 62) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 571) {
return 4;
}
else if (Cs >= 571) {
return 8;
}
}
}
}
else if (Rs >= 1182) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 8;
}
}
}
else if (MVDiff >= 13) {
if (AFD < 9) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 1101) {
return 8;
}
else if (Rs >= 1101) {
return 2;
}
}
}
else if (AFD >= 9) {
if (Rs < 1254) {
return 1;
}
else if (Rs >= 1254) {
return 2;
}
}
}
}
else if (SC >= 2413) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 63) {
if (Rs < 1354) {
if (Cs < 1104) {
if (posBalance < 47) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (posBalance >= 47) {
if (RsDiff < 3) {
if (Cs < 871) {
if (Cs < 722) {
return 8;
}
else if (Cs >= 722) {
return 1;
}
}
else if (Cs >= 871) {
return 8;
}
}
else if (RsDiff >= 3) {
if (TSC < 6) {
if (SC < 2342) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
if (tcor < 95) {
if (Rs < 923) {
return 8;
}
else if (Rs >= 923) {
return 1;
}
}
else if (tcor >= 95) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
return 8;
}
}
}
}
else if (SC >= 2342) {
return 4;
}
}
else if (TSC >= 6) {
if (Cs < 730) {
return 4;
}
else if (Cs >= 730) {
return 8;
}
}
}
}
}
else if (Cs >= 1104) {
if (SC < 2497) {
if (Rs < 1170) {
if (mu_mv_mag_sq < 66) {
return 8;
}
else if (mu_mv_mag_sq >= 66) {
return 4;
}
}
else if (Rs >= 1170) {
if (posBalance < 56) {
if (RsDiff < 3) {
if (Rs < 1265) {
return 1;
}
else if (Rs >= 1265) {
return 4;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
else if (SC >= 2497) {
if (Rs < 1315) {
return 1;
}
else if (Rs >= 1315) {
return 8;
}
}
}
}
else if (Rs >= 1354) {
if (posBalance < 49) {
if (tcor < 93) {
if (AFD < 7) {
if (Rs < 1520) {
return 8;
}
else if (Rs >= 1520) {
return 2;
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (tcor >= 93) {
if (Cs < 1180) {
return 1;
}
else if (Cs >= 1180) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 104) {
if (RsCsDiff < 16) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (Rs < 1570) {
if (Cs < 972) {
return 1;
}
else if (Cs >= 972) {
return 4;
}
}
else if (Rs >= 1570) {
if (Rs < 1716) {
return 8;
}
else if (Rs >= 1716) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (diffRsCsdiff < 7) {
return 8;
}
else if (diffRsCsdiff >= 7) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (RsDiff < 4) {
if (Cs < 1044) {
return 8;
}
else if (Cs >= 1044) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (RsDiff >= 4) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
return 4;
}
}
}
}
}
else if (ssDCval >= 78) {
if (MVDiff < 23) {
if (SC < 2102) {
if (TSC < 6) {
if (diffAFD < -1) {
if (RsDiff < 3) {
if (Rs < 940) {
return 1;
}
else if (Rs >= 940) {
return 8;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (diffAFD >= -1) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 1240) {
return 8;
}
else if (Rs >= 1240) {
return 4;
}
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (SC >= 2102) {
if (Rs < 1767) {
if (ssDCval < 138) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 1282) {
return 4;
}
else if (Rs >= 1282) {
return 8;
}
}
}
else if (ssDCval >= 138) {
return 1;
}
}
else if (Rs >= 1767) {
if (Rs < 1768) {
return 1;
}
else if (Rs >= 1768) {
if (mu_mv_mag_sq < 55) {
if (Rs < 1878) {
return 1;
}
else if (Rs >= 1878) {
return 2;
}
}
else if (mu_mv_mag_sq >= 55) {
if (SC < 3635) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 1854) {
return 2;
}
else if (Rs >= 1854) {
return 8;
}
}
}
else if (SC >= 3635) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 23) {
if (Cs < 742) {
return 1;
}
else if (Cs >= 742) {
if (Rs < 1311) {
return 4;
}
else if (Rs >= 1311) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 138) {
if (MVDiff < 23) {
if (ssDCval < 155) {
if (mu_mv_mag_sq < 45) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
else if (mu_mv_mag_sq >= 45) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
}
else if (ssDCval >= 155) {
return 2;
}
}
else if (MVDiff >= 23) {
return 4;
}
}
}
}
else if (RsCsDiff >= 23) {
if (refDCval < 74) {
if (MVDiff < 7) {
if (refDCval < 62) {
if (Rs < 1275) {
if (SC < 2005) {
if (negBalance < 33) {
if (Cs < 771) {
return 4;
}
else if (Cs >= 771) {
return 1;
}
}
else if (negBalance >= 33) {
return 8;
}
}
else if (SC >= 2005) {
if (mcTcor < 92) {
if (Rs < 1227) {
return 8;
}
else if (Rs >= 1227) {
return 1;
}
}
else if (mcTcor >= 92) {
if (AFD < 12) {
if (diffRsCsdiff < -21) {
return 8;
}
else if (diffRsCsdiff >= -21) {
return 4;
}
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (Rs >= 1275) {
if (posBalance < 45) {
if (AFD < 9) {
if (diffRsCsdiff < -31) {
return 8;
}
else if (diffRsCsdiff >= -31) {
return 4;
}
}
else if (AFD >= 9) {
return 8;
}
}
else if (posBalance >= 45) {
if (diffMVdiffVal < -5) {
if (Rs < 1652) {
if (Rs < 1472) {
return 4;
}
else if (Rs >= 1472) {
return 1;
}
}
else if (Rs >= 1652) {
return 2;
}
}
else if (diffMVdiffVal >= -5) {
if (negBalance < 30) {
if (mu_mv_mag_sq < 111) {
if (Cs < 912) {
if (Rs < 2181) {
return 1;
}
else if (Rs >= 2181) {
return 2;
}
}
else if (Cs >= 912) {
return 4;
}
}
else if (mu_mv_mag_sq >= 111) {
return 8;
}
}
else if (negBalance >= 30) {
if (RsDiff < 7) {
if (diffMVdiffVal < 3) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 4;
}
}
}
else if (RsDiff >= 7) {
if (Rs < 2022) {
return 8;
}
else if (Rs >= 2022) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 62) {
if (Rs < 1545) {
if (diffRsCsdiff < 13) {
if (tcor < 90) {
if (Rs < 1036) {
return 1;
}
else if (Rs >= 1036) {
return 8;
}
}
else if (tcor >= 90) {
if (RsCsDiff < 31) {
if (Cs < 907) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
return 1;
}
}
else if (Cs >= 907) {
if (MVDiff < 2) {
if (Cs < 1217) {
return 8;
}
else if (Cs >= 1217) {
if (Rs < 1103) {
return 4;
}
else if (Rs >= 1103) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (SC < 2393) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
}
else if (SC >= 2393) {
return 4;
}
}
}
}
else if (RsCsDiff >= 31) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (Rs < 1323) {
return 4;
}
else if (Rs >= 1323) {
return 8;
}
}
}
}
}
else if (diffRsCsdiff >= 13) {
if (Rs < 948) {
return 2;
}
else if (Rs >= 948) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
if (Rs < 1360) {
return 1;
}
else if (Rs >= 1360) {
return 8;
}
}
}
}
}
else if (Rs >= 1545) {
if (CsDiff < 5) {
if (ssDCval < 70) {
if (RsDiff < 5) {
if (AFD < 7) {
if (Rs < 2188) {
return 8;
}
else if (Rs >= 2188) {
return 4;
}
}
else if (AFD >= 7) {
return 2;
}
}
else if (RsDiff >= 5) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
return 2;
}
}
}
else if (ssDCval >= 70) {
if (AFD < 9) {
if (Cs < 820) {
return 1;
}
else if (Cs >= 820) {
return 8;
}
}
else if (AFD >= 9) {
if (Rs < 1729) {
return 2;
}
else if (Rs >= 1729) {
return 4;
}
}
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
if (negBalance < 45) {
if (gchDC < 3) {
if (Cs < 894) {
if (Rs < 2178) {
if (AFD < 8) {
if (Cs < 627) {
return 2;
}
else if (Cs >= 627) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (Cs < 735) {
return 8;
}
else if (Cs >= 735) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
if (Rs < 1286) {
return 8;
}
else if (Rs >= 1286) {
return 1;
}
}
}
}
}
}
else if (AFD >= 8) {
if (tcor < 91) {
if (mu_mv_mag_sq < 58) {
if (Cs < 626) {
if (Rs < 935) {
if (Rs < 497) {
return 8;
}
else if (Rs >= 497) {
return 1;
}
}
else if (Rs >= 935) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 2;
}
}
}
else if (Cs >= 626) {
if (Rs < 1995) {
if (refDCval < 72) {
if (Rs < 691) {
if (Rs < 632) {
return 2;
}
else if (Rs >= 632) {
return 4;
}
}
else if (Rs >= 691) {
return 2;
}
}
else if (refDCval >= 72) {
return 4;
}
}
else if (Rs >= 1995) {
if (Cs < 748) {
return 4;
}
else if (Cs >= 748) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (mu_mv_mag_sq < 71) {
return 1;
}
else if (mu_mv_mag_sq >= 71) {
if (Cs < 660) {
return 1;
}
else if (Cs >= 660) {
return 4;
}
}
}
}
else if (tcor >= 91) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
return 2;
}
}
}
}
else if (Rs >= 2178) {
return 2;
}
}
else if (Cs >= 894) {
if (tcor < 86) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (Rs < 910) {
if (Rs < 402) {
return 1;
}
else if (Rs >= 402) {
return 2;
}
}
else if (Rs >= 910) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
return 1;
}
}
}
}
else if (tcor >= 86) {
if (diffRsCsdiff < 11) {
if (Rs < 2326) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (TSC < 7) {
if (MVDiff < 9) {
if (Rs < 1322) {
return 4;
}
else if (Rs >= 1322) {
return 2;
}
}
else if (MVDiff >= 9) {
return 2;
}
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (Rs >= 2326) {
return 8;
}
}
else if (diffRsCsdiff >= 11) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
return 8;
}
}
}
}
}
else if (gchDC >= 3) {
if (negBalance < 25) {
if (Rs < 1587) {
return 2;
}
else if (Rs >= 1587) {
return 1;
}
}
else if (negBalance >= 25) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (CsDiff < 5) {
if (Cs < 1156) {
if (posBalance < 54) {
if (Cs < 801) {
if (negBalance < 52) {
if (posBalance < 53) {
if (SC < 1491) {
return 2;
}
else if (SC >= 1491) {
return 1;
}
}
else if (posBalance >= 53) {
return 4;
}
}
else if (negBalance >= 52) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (Cs >= 801) {
if (Rs < 1032) {
if (Rs < 811) {
return 1;
}
else if (Rs >= 811) {
return 8;
}
}
else if (Rs >= 1032) {
if (diffMVdiffVal < -8) {
return 2;
}
else if (diffMVdiffVal >= -8) {
return 4;
}
}
}
}
else if (posBalance >= 54) {
if (ssDCval < 66) {
if (negBalance < 46) {
if (Rs < 1217) {
return 2;
}
else if (Rs >= 1217) {
return 4;
}
}
else if (negBalance >= 46) {
if (Cs < 556) {
return 4;
}
else if (Cs >= 556) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (Cs < 590) {
return 4;
}
else if (Cs >= 590) {
return 8;
}
}
}
}
}
else if (ssDCval >= 66) {
if (TSC < 7) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (Cs < 1014) {
return 2;
}
else if (Cs >= 1014) {
return 4;
}
}
}
else if (TSC >= 7) {
if (Cs < 761) {
return 1;
}
else if (Cs >= 761) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1156) {
if (Rs < 1837) {
return 8;
}
else if (Rs >= 1837) {
if (mu_mv_mag_sq < 71) {
if (Rs < 2221) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
else if (Rs >= 2221) {
return 8;
}
}
else if (mu_mv_mag_sq >= 71) {
return 8;
}
}
}
}
else if (CsDiff >= 5) {
if (Rs < 971) {
if (ssDCval < 69) {
if (ssDCval < 54) {
return 2;
}
else if (ssDCval >= 54) {
return 1;
}
}
else if (ssDCval >= 69) {
if (Rs < 958) {
if (Rs < 587) {
return 8;
}
else if (Rs >= 587) {
return 2;
}
}
else if (Rs >= 958) {
return 1;
}
}
}
else if (Rs >= 971) {
if (tcor < 86) {
if (Rs < 1025) {
return 4;
}
else if (Rs >= 1025) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (tcor >= 86) {
if (Cs < 1186) {
return 8;
}
else if (Cs >= 1186) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 74) {
if (negBalance < 34) {
if (posBalance < 57) {
if (MVDiff < 4) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 138) {
if (diffAFD < 3) {
if (Rs < 699) {
return 2;
}
else if (Rs >= 699) {
return 1;
}
}
else if (diffAFD >= 3) {
return 8;
}
}
else if (ssDCval >= 138) {
if (Rs < 1366) {
return 2;
}
else if (Rs >= 1366) {
return 8;
}
}
}
}
else if (posBalance >= 57) {
if (tcor < 93) {
if (SC < 2661) {
return 2;
}
else if (SC >= 2661) {
return 8;
}
}
else if (tcor >= 93) {
if (Rs < 2429) {
return 2;
}
else if (Rs >= 2429) {
return 1;
}
}
}
}
else if (negBalance >= 34) {
if (ssDCval < 94) {
if (TSC < 7) {
if (Rs < 1610) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 3) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
if (Rs < 1499) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (Rs >= 1499) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (tcor < 92) {
if (Cs < 694) {
return 1;
}
else if (Cs >= 694) {
if (ssDCval < 83) {
if (negBalance < 45) {
return 4;
}
else if (negBalance >= 45) {
if (Rs < 680) {
return 1;
}
else if (Rs >= 680) {
if (Rs < 1090) {
return 2;
}
else if (Rs >= 1090) {
return 8;
}
}
}
}
else if (ssDCval >= 83) {
return 2;
}
}
}
else if (tcor >= 92) {
if (RsCsDiff < 31) {
if (diffRsCsdiff < -4) {
if (Rs < 1182) {
return 1;
}
else if (Rs >= 1182) {
return 4;
}
}
else if (diffRsCsdiff >= -4) {
return 8;
}
}
else if (RsCsDiff >= 31) {
if (Rs < 1385) {
return 4;
}
else if (Rs >= 1385) {
return 1;
}
}
}
}
}
else if (Rs >= 1610) {
if (ssDCval < 88) {
if (Cs < 1290) {
if (Cs < 950) {
return 1;
}
else if (Cs >= 950) {
if (Cs < 1152) {
return 8;
}
else if (Cs >= 1152) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (Cs >= 1290) {
return 8;
}
}
else if (ssDCval >= 88) {
if (SC < 2649) {
if (mcTcor < 95) {
if (Rs < 1762) {
return 8;
}
else if (Rs >= 1762) {
return 2;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
else if (SC >= 2649) {
if (Rs < 2129) {
if (Rs < 1841) {
if (Rs < 1790) {
return 2;
}
else if (Rs >= 1790) {
return 8;
}
}
else if (Rs >= 1841) {
return 2;
}
}
else if (Rs >= 2129) {
return 4;
}
}
}
}
}
else if (TSC >= 7) {
if (tcor < 87) {
if (negBalance < 52) {
if (Rs < 1669) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (Rs >= 1669) {
if (Rs < 1771) {
return 2;
}
else if (Rs >= 1771) {
return 1;
}
}
}
else if (negBalance >= 52) {
if (diffAFD < -2) {
if (Rs < 1011) {
return 1;
}
else if (Rs >= 1011) {
return 2;
}
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 22) {
return 8;
}
else if (mu_mv_mag_sq >= 22) {
return 4;
}
}
}
}
else if (tcor >= 87) {
if (mcTcor < 94) {
if (Cs < 600) {
return 1;
}
else if (Cs >= 600) {
if (tcor < 88) {
if (Rs < 1176) {
return 2;
}
else if (Rs >= 1176) {
return 8;
}
}
else if (tcor >= 88) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (posBalance < 55) {
if (Cs < 1133) {
if (refDCval < 91) {
if (Cs < 775) {
if (Rs < 1556) {
return 8;
}
else if (Rs >= 1556) {
return 1;
}
}
else if (Cs >= 775) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (refDCval >= 91) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
return 8;
}
}
}
else if (Cs >= 1133) {
if (Rs < 1538) {
return 4;
}
else if (Rs >= 1538) {
if (Cs < 1304) {
return 8;
}
else if (Cs >= 1304) {
return 2;
}
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 81) {
if (Rs < 1492) {
return 4;
}
else if (Rs >= 1492) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
return 1;
}
}
}
else if (ssDCval >= 81) {
if (MVDiff < 2) {
if (Rs < 1781) {
return 8;
}
else if (Rs >= 1781) {
return 1;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 94) {
if (mu_mv_mag_sq < 47) {
if (TSC < 7) {
if (refDCval < 161) {
if (refDCval < 136) {
if (posBalance < 67) {
if (Rs < 945) {
if (Rs < 645) {
return 2;
}
else if (Rs >= 645) {
if (Rs < 926) {
return 8;
}
else if (Rs >= 926) {
return 4;
}
}
}
else if (Rs >= 945) {
if (refDCval < 100) {
if (Rs < 1034) {
return 8;
}
else if (Rs >= 1034) {
return 4;
}
}
else if (refDCval >= 100) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
else if (posBalance >= 67) {
return 2;
}
}
else if (refDCval >= 136) {
if (ssDCval < 140) {
if (Rs < 1615) {
if (ssDCval < 137) {
if (Rs < 1493) {
return 4;
}
else if (Rs >= 1493) {
return 8;
}
}
else if (ssDCval >= 137) {
return 4;
}
}
else if (Rs >= 1615) {
return 8;
}
}
else if (ssDCval >= 140) {
return 8;
}
}
}
else if (refDCval >= 161) {
if (Rs < 943) {
return 8;
}
else if (Rs >= 943) {
return 2;
}
}
}
else if (TSC >= 7) {
if (Rs < 1191) {
if (tcor < 88) {
if (Cs < 1169) {
if (ssDCval < 100) {
return 8;
}
else if (ssDCval >= 100) {
return 4;
}
}
else if (Cs >= 1169) {
return 1;
}
}
else if (tcor >= 88) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 874) {
return 4;
}
else if (Rs >= 874) {
return 1;
}
}
}
}
else if (Rs >= 1191) {
if (Rs < 2263) {
if (Cs < 1231) {
if (MVDiff < 38) {
if (negBalance < 62) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
}
else if (negBalance >= 62) {
return 2;
}
}
else if (MVDiff >= 38) {
if (Rs < 1662) {
return 1;
}
else if (Rs >= 1662) {
return 2;
}
}
}
else if (Cs >= 1231) {
if (Rs < 1650) {
return 1;
}
else if (Rs >= 1650) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
}
}
else if (Rs >= 2263) {
if (Cs < 1274) {
return 4;
}
else if (Cs >= 1274) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (negBalance < 49) {
if (diffAFD < 2) {
if (Rs < 2479) {
if (refDCval < 98) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (refDCval >= 98) {
return 8;
}
}
else if (Rs >= 2479) {
return 1;
}
}
else if (diffAFD >= 2) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
if (diffRsCsdiff < 15) {
if (Cs < 903) {
return 8;
}
else if (Cs >= 903) {
return 1;
}
}
else if (diffRsCsdiff >= 15) {
return 8;
}
}
}
}
else if (negBalance >= 49) {
if (MVDiff < 5) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 1;
}
}
else if (MVDiff >= 5) {
if (Cs < 772) {
return 2;
}
else if (Cs >= 772) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (AFD >= 15) {
if (MVDiff < 16) {
if (TSC < 9) {
if (posBalance < 66) {
if (Rs < 370) {
return 2;
}
else if (Rs >= 370) {
if (TSC < 8) {
if (AFD < 19) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
if (ssDCval < 108) {
if (Rs < 938) {
return 1;
}
else if (Rs >= 938) {
return 4;
}
}
else if (ssDCval >= 108) {
return 8;
}
}
}
else if (AFD >= 19) {
return 4;
}
}
else if (TSC >= 8) {
if (RsCsDiff < 33) {
return 4;
}
else if (RsCsDiff >= 33) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 66) {
if (mu_mv_mag_sq < 7) {
return 4;
}
else if (mu_mv_mag_sq >= 7) {
return 2;
}
}
}
else if (TSC >= 9) {
if (ssDCval < 88) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
return 8;
}
}
else if (ssDCval >= 88) {
if (ssDCval < 115) {
if (Rs < 1504) {
return 4;
}
else if (Rs >= 1504) {
return 8;
}
}
else if (ssDCval >= 115) {
if (AFD < 16) {
if (Rs < 1569) {
return 1;
}
else if (Rs >= 1569) {
return 8;
}
}
else if (AFD >= 16) {
return 2;
}
}
}
}
}
else if (MVDiff >= 16) {
if (Rs < 838) {
if (refDCval < 89) {
if (tcor < 60) {
if (MVDiff < 63) {
if (mu_mv_mag_sq < 3) {
if (Rs < 718) {
return 1;
}
else if (Rs >= 718) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (MVDiff >= 63) {
if (RsCsDiff < 43) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (tcor < 56) {
return 4;
}
else if (tcor >= 56) {
return 2;
}
}
}
else if (RsCsDiff >= 43) {
if (refDCval < 43) {
return 2;
}
else if (refDCval >= 43) {
return 8;
}
}
}
}
else if (tcor >= 60) {
if (AFD < 19) {
if (Cs < 841) {
if (diffRsCsdiff < -5) {
if (SC < 1246) {
return 2;
}
else if (SC >= 1246) {
if (Cs < 630) {
return 1;
}
else if (Cs >= 630) {
return 8;
}
}
}
else if (diffRsCsdiff >= -5) {
if (mu_mv_mag_sq < 9) {
if (ssDCval < 70) {
return 4;
}
else if (ssDCval >= 70) {
if (Rs < 731) {
return 1;
}
else if (Rs >= 731) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (mu_mv_mag_sq < 33) {
if (RsCsDiff < 29) {
return 2;
}
else if (RsCsDiff >= 29) {
return 8;
}
}
else if (mu_mv_mag_sq >= 33) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
}
else if (Cs >= 841) {
if (diffAFD < 3) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
return 2;
}
}
else if (diffAFD >= 3) {
if (Rs < 712) {
return 1;
}
else if (Rs >= 712) {
return 4;
}
}
}
}
else if (AFD >= 19) {
if (tcor < 72) {
if (RsDiff < 5) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 4;
}
}
else if (RsDiff >= 5) {
if (AFD < 23) {
return 1;
}
else if (AFD >= 23) {
return 4;
}
}
}
else if (tcor >= 72) {
if (diffMVdiffVal < 2) {
if (mcTcor < 91) {
return 2;
}
else if (mcTcor >= 91) {
if (Cs < 665) {
return 4;
}
else if (Cs >= 665) {
return 1;
}
}
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
}
}
}
else if (refDCval >= 89) {
if (SC < 1515) {
if (posBalance < 67) {
if (ssDCval < 90) {
if (Rs < 554) {
return 1;
}
else if (Rs >= 554) {
return 8;
}
}
else if (ssDCval >= 90) {
return 2;
}
}
else if (posBalance >= 67) {
if (MVDiff < 64) {
return 8;
}
else if (MVDiff >= 64) {
if (ssDCval < 113) {
return 2;
}
else if (ssDCval >= 113) {
if (Cs < 806) {
return 1;
}
else if (Cs >= 806) {
return 8;
}
}
}
}
}
else if (SC >= 1515) {
if (CsDiff < 5) {
if (SC < 1774) {
return 1;
}
else if (SC >= 1774) {
if (Rs < 702) {
return 8;
}
else if (Rs >= 702) {
return 1;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 667) {
if (Rs < 623) {
return 2;
}
else if (Rs >= 623) {
return 4;
}
}
else if (Rs >= 667) {
return 8;
}
}
}
}
}
else if (Rs >= 838) {
if (mcTcor < 84) {
if (ssDCval < 66) {
if (tcor < 58) {
if (AFD < 18) {
if (ssDCval < 48) {
return 8;
}
else if (ssDCval >= 48) {
return 1;
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (tcor >= 58) {
if (Rs < 1025) {
return 2;
}
else if (Rs >= 1025) {
return 8;
}
}
}
else if (ssDCval >= 66) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
if (Rs < 1167) {
return 8;
}
else if (Rs >= 1167) {
return 2;
}
}
}
}
else if (mcTcor >= 84) {
if (Rs < 1329) {
if (Rs < 1180) {
if (AFD < 17) {
if (Cs < 746) {
if (Rs < 918) {
return 8;
}
else if (Rs >= 918) {
return 4;
}
}
else if (Cs >= 746) {
return 8;
}
}
else if (AFD >= 17) {
if (MVDiff < 24) {
if (Cs < 1038) {
return 8;
}
else if (Cs >= 1038) {
if (Rs < 1020) {
return 1;
}
else if (Rs >= 1020) {
return 4;
}
}
}
else if (MVDiff >= 24) {
if (diffRsCsdiff < 21) {
if (diffTSC < 2) {
if (RsCsDiff < 57) {
if (Rs < 929) {
return 4;
}
else if (Rs >= 929) {
if (Rs < 1091) {
return 1;
}
else if (Rs >= 1091) {
return 4;
}
}
}
else if (RsCsDiff >= 57) {
if (Rs < 998) {
return 4;
}
else if (Rs >= 998) {
return 2;
}
}
}
else if (diffTSC >= 2) {
if (Cs < 618) {
return 1;
}
else if (Cs >= 618) {
return 2;
}
}
}
else if (diffRsCsdiff >= 21) {
if (Rs < 921) {
return 8;
}
else if (Rs >= 921) {
return 1;
}
}
}
}
}
else if (Rs >= 1180) {
if (mcTcor < 92) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 2;
}
}
else if (mcTcor >= 92) {
return 4;
}
}
}
else if (Rs >= 1329) {
if (ssDCval < 128) {
if (Rs < 1996) {
if (diffRsCsdiff < -6) {
return 8;
}
else if (diffRsCsdiff >= -6) {
if (tcor < 79) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 2;
}
}
else if (tcor >= 79) {
return 8;
}
}
}
else if (Rs >= 1996) {
return 2;
}
}
else if (ssDCval >= 128) {
return 1;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 63) {
if (refDCval < 42) {
if (Rs < 1927) {
if (Rs < 702) {
return 1;
}
else if (Rs >= 702) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Cs < 685) {
return 1;
}
else if (Cs >= 685) {
return 8;
}
}
}
}
else if (Rs >= 1927) {
if (Rs < 2035) {
return 2;
}
else if (Rs >= 2035) {
return 1;
}
}
}
else if (refDCval >= 42) {
if (Cs < 820) {
if (refDCval < 85) {
if (Rs < 1985) {
if (mcTcor < 90) {
if (RsDiff < 8) {
if (diffTSC < 3) {
if (ssDCval < 52) {
if (RsCsDiff < 73) {
return 8;
}
else if (RsCsDiff >= 73) {
if (posBalance < 61) {
if (RsCsDiff < 80) {
if (Rs < 823) {
return 1;
}
else if (Rs >= 823) {
return 2;
}
}
else if (RsCsDiff >= 80) {
return 8;
}
}
else if (posBalance >= 61) {
return 4;
}
}
}
else if (ssDCval >= 52) {
if (TSC < 10) {
if (AFD < 9) {
if (Rs < 1056) {
return 8;
}
else if (Rs >= 1056) {
return 1;
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (TSC >= 10) {
if (Rs < 974) {
if (Cs < 658) {
return 4;
}
else if (Cs >= 658) {
return 1;
}
}
else if (Rs >= 974) {
return 8;
}
}
}
}
else if (diffTSC >= 3) {
if (Rs < 470) {
return 8;
}
else if (Rs >= 470) {
return 2;
}
}
}
else if (RsDiff >= 8) {
if (posBalance < 41) {
if (Cs < 681) {
return 1;
}
else if (Cs >= 681) {
return 2;
}
}
else if (posBalance >= 41) {
if (RsCsDiff < 94) {
if (Cs < 595) {
return 2;
}
else if (Cs >= 595) {
return 8;
}
}
else if (RsCsDiff >= 94) {
if (mu_mv_mag_sq < 40) {
if (mu_mv_mag_sq < 11) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 8;
}
}
else if (mu_mv_mag_sq >= 11) {
if (Rs < 1448) {
return 4;
}
else if (Rs >= 1448) {
if (TSC < 10) {
if (Rs < 1479) {
return 8;
}
else if (Rs >= 1479) {
return 2;
}
}
else if (TSC >= 10) {
if (Cs < 646) {
return 4;
}
else if (Cs >= 646) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (Cs < 764) {
return 8;
}
else if (Cs >= 764) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (SC < 2174) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (MVDiff < 87) {
if (negBalance < 55) {
if (Rs < 1082) {
if (Cs < 705) {
if (CsDiff < 6) {
if (tcor < 79) {
return 1;
}
else if (tcor >= 79) {
return 2;
}
}
else if (CsDiff >= 6) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
return 4;
}
}
}
else if (Cs >= 705) {
return 1;
}
}
else if (Rs >= 1082) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
if (Rs < 1121) {
return 8;
}
else if (Rs >= 1121) {
if (MVDiff < 21) {
if (Rs < 1235) {
return 4;
}
else if (Rs >= 1235) {
return 2;
}
}
else if (MVDiff >= 21) {
return 2;
}
}
}
}
}
else if (negBalance >= 55) {
if (diffRsCsdiff < 11) {
if (Cs < 639) {
return 1;
}
else if (Cs >= 639) {
if (Cs < 770) {
return 4;
}
else if (Cs >= 770) {
return 8;
}
}
}
else if (diffRsCsdiff >= 11) {
if (MVDiff < 24) {
return 2;
}
else if (MVDiff >= 24) {
return 8;
}
}
}
}
else if (MVDiff >= 87) {
if (diffRsCsdiff < -34) {
if (Rs < 883) {
return 2;
}
else if (Rs >= 883) {
return 8;
}
}
else if (diffRsCsdiff >= -34) {
return 2;
}
}
}
}
else if (SC >= 2174) {
if (MVDiff < 13) {
if (TSC < 7) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
else if (TSC >= 7) {
return 2;
}
}
else if (MVDiff >= 13) {
if (CsDiff < 4) {
if (ssDCval < 66) {
return 2;
}
else if (ssDCval >= 66) {
if (Rs < 1707) {
return 4;
}
else if (Rs >= 1707) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (diffRsCsdiff < -52) {
if (Rs < 1534) {
return 2;
}
else if (Rs >= 1534) {
return 4;
}
}
else if (diffRsCsdiff >= -52) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1985) {
if (mcTcor < 92) {
if (refDCval < 73) {
if (TSC < 8) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 2;
}
}
else if (TSC >= 8) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (Rs < 2293) {
return 4;
}
else if (Rs >= 2293) {
return 1;
}
}
}
}
else if (refDCval >= 73) {
if (gchDC < 4) {
if (Rs < 2060) {
return 2;
}
else if (Rs >= 2060) {
return 1;
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (TSC < 6) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 1;
}
}
}
else if (TSC >= 6) {
if (Cs < 812) {
if (Cs < 534) {
return 1;
}
else if (Cs >= 534) {
return 2;
}
}
else if (Cs >= 812) {
if (Rs < 2176) {
return 8;
}
else if (Rs >= 2176) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 85) {
if (tcor < 79) {
if (mcTcor < 65) {
return 1;
}
else if (mcTcor >= 65) {
if (posBalance < 56) {
if (AFD < 26) {
if (MVDiff < 62) {
if (tcor < 75) {
if (gchDC < 3) {
if (Rs < 922) {
return 2;
}
else if (Rs >= 922) {
return 8;
}
}
else if (gchDC >= 3) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 8;
}
}
}
else if (tcor >= 75) {
return 2;
}
}
else if (MVDiff >= 62) {
return 8;
}
}
else if (AFD >= 26) {
if (mu_mv_mag_sq < 37) {
if (MVDiff < 119) {
return 1;
}
else if (MVDiff >= 119) {
if (Rs < 1283) {
return 2;
}
else if (Rs >= 1283) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 1361) {
return 4;
}
else if (Rs >= 1361) {
return 2;
}
}
}
}
else if (posBalance >= 56) {
if (posBalance < 66) {
if (AFD < 25) {
if (mu_mv_mag_sq < 7) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (tcor < 51) {
if (Rs < 2172) {
return 4;
}
else if (Rs >= 2172) {
return 1;
}
}
else if (tcor >= 51) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 555) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (Cs >= 555) {
if (CsDiff < 6) {
if (diffMVdiffVal < 19) {
if (Rs < 2299) {
if (Cs < 561) {
if (Rs < 1977) {
return 4;
}
else if (Rs >= 1977) {
return 8;
}
}
else if (Cs >= 561) {
return 4;
}
}
else if (Rs >= 2299) {
return 1;
}
}
else if (diffMVdiffVal >= 19) {
if (Rs < 1580) {
return 1;
}
else if (Rs >= 1580) {
if (Rs < 2190) {
return 2;
}
else if (Rs >= 2190) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (Rs < 1183) {
if (AFD < 22) {
if (Rs < 974) {
return 2;
}
else if (Rs >= 974) {
return 8;
}
}
else if (AFD >= 22) {
return 1;
}
}
else if (Rs >= 1183) {
if (Rs < 1751) {
return 4;
}
else if (Rs >= 1751) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 4;
}
}
}
}
}
}
}
else if (AFD >= 25) {
if (SC < 2275) {
if (Rs < 771) {
return 2;
}
else if (Rs >= 771) {
return 4;
}
}
else if (SC >= 2275) {
if (AFD < 32) {
if (Rs < 2261) {
return 1;
}
else if (Rs >= 2261) {
if (Rs < 2368) {
return 8;
}
else if (Rs >= 2368) {
return 2;
}
}
}
else if (AFD >= 32) {
if (Rs < 1916) {
return 4;
}
else if (Rs >= 1916) {
return 2;
}
}
}
}
}
else if (posBalance >= 66) {
if (tcor < 72) {
if (MVDiff < 85) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 1;
}
}
else if (MVDiff >= 85) {
if (AFD < 30) {
return 8;
}
else if (AFD >= 30) {
if (Rs < 1472) {
return 1;
}
else if (Rs >= 1472) {
return 2;
}
}
}
}
else if (tcor >= 72) {
if (RsCsDiff < 123) {
return 8;
}
else if (RsCsDiff >= 123) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
if (Rs < 2091) {
return 8;
}
else if (Rs >= 2091) {
return 1;
}
}
}
}
}
}
}
}
else if (tcor >= 79) {
if (MVDiff < 26) {
if (refDCval < 91) {
if (posBalance < 59) {
if (ssDCval < 87) {
if (Rs < 1373) {
return 8;
}
else if (Rs >= 1373) {
return 2;
}
}
else if (ssDCval >= 87) {
return 1;
}
}
else if (posBalance >= 59) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 4;
}
}
}
else if (refDCval >= 91) {
if (CsDiff < 5) {
if (TSC < 7) {
if (Rs < 1860) {
if (Rs < 1507) {
return 8;
}
else if (Rs >= 1507) {
return 1;
}
}
else if (Rs >= 1860) {
return 2;
}
}
else if (TSC >= 7) {
if (Cs < 810) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 2;
}
}
else if (Cs >= 810) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 2154) {
return 2;
}
else if (Rs >= 2154) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
}
else if (MVDiff >= 26) {
if (TSC < 9) {
if (mcTcor < 96) {
if (negBalance < 49) {
if (Rs < 1289) {
return 8;
}
else if (Rs >= 1289) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
return 2;
}
}
}
else if (negBalance >= 49) {
if (Cs < 591) {
if (Rs < 1927) {
return 2;
}
else if (Rs >= 1927) {
return 1;
}
}
else if (Cs >= 591) {
return 8;
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (TSC >= 9) {
if (AFD < 32) {
if (mcTcor < 92) {
if (MVDiff < 59) {
return 1;
}
else if (MVDiff >= 59) {
return 2;
}
}
else if (mcTcor >= 92) {
if (Rs < 1506) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 1;
}
}
else if (Rs >= 1506) {
if (ssDCval < 147) {
if (ssDCval < 104) {
if (Rs < 2203) {
return 8;
}
else if (Rs >= 2203) {
return 1;
}
}
else if (ssDCval >= 104) {
return 1;
}
}
else if (ssDCval >= 147) {
if (Rs < 2157) {
return 8;
}
else if (Rs >= 2157) {
return 1;
}
}
}
}
}
else if (AFD >= 32) {
return 2;
}
}
}
}
}
}
else if (Cs >= 820) {
if (refDCval < 114) {
if (tcor < 62) {
if (CsDiff < 6) {
if (refDCval < 90) {
if (ssDCval < 58) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (refDCval >= 90) {
if (Rs < 2183) {
return 4;
}
else if (Rs >= 2183) {
if (Rs < 2397) {
return 2;
}
else if (Rs >= 2397) {
return 1;
}
}
}
}
else if (CsDiff >= 6) {
if (diffAFD < 1) {
if (gchDC < 5) {
if (tcor < 57) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (AFD < 28) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
if (MVDiff < 84) {
if (diffTSC < 2) {
if (diffMVdiffVal < -11) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 2;
}
}
else if (diffMVdiffVal >= -11) {
return 4;
}
}
else if (diffTSC >= 2) {
return 8;
}
}
else if (MVDiff >= 84) {
if (Rs < 992) {
return 4;
}
else if (Rs >= 992) {
if (Rs < 1335) {
return 8;
}
else if (Rs >= 1335) {
return 2;
}
}
}
}
}
else if (AFD >= 28) {
if (mcTcor < 88) {
if (Cs < 1109) {
return 1;
}
else if (Cs >= 1109) {
return 2;
}
}
else if (mcTcor >= 88) {
return 4;
}
}
}
}
else if (tcor >= 57) {
if (Cs < 1112) {
if (tcor < 61) {
if (mcTcor < 88) {
return 2;
}
else if (mcTcor >= 88) {
if (Rs < 2102) {
return 8;
}
else if (Rs >= 2102) {
return 2;
}
}
}
else if (tcor >= 61) {
if (RsCsDiff < 73) {
return 2;
}
else if (RsCsDiff >= 73) {
if (CsDiff < 9) {
return 4;
}
else if (CsDiff >= 9) {
return 2;
}
}
}
}
else if (Cs >= 1112) {
if (TSC < 13) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
if (diffRsCsdiff < -29) {
if (Rs < 1766) {
return 8;
}
else if (Rs >= 1766) {
return 4;
}
}
else if (diffRsCsdiff >= -29) {
return 4;
}
}
}
}
else if (TSC >= 13) {
if (Rs < 1834) {
return 8;
}
else if (Rs >= 1834) {
return 1;
}
}
}
}
}
else if (gchDC >= 5) {
if (mu_mv_mag_sq < 46) {
if (refDCval < 72) {
return 8;
}
else if (refDCval >= 72) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 46) {
return 1;
}
}
}
else if (diffAFD >= 1) {
if (Rs < 628) {
if (ssDCval < 63) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (Cs < 887) {
return 4;
}
else if (Cs >= 887) {
return 8;
}
}
}
else if (ssDCval >= 63) {
return 4;
}
}
else if (Rs >= 628) {
if (posBalance < 61) {
if (diffMVdiffVal < 5) {
if (diffAFD < 3) {
if (Cs < 1236) {
if (tcor < 60) {
return 2;
}
else if (tcor >= 60) {
return 4;
}
}
else if (Cs >= 1236) {
return 1;
}
}
else if (diffAFD >= 3) {
return 4;
}
}
else if (diffMVdiffVal >= 5) {
if (tcor < 55) {
if (Rs < 662) {
return 8;
}
else if (Rs >= 662) {
if (RsDiff < 12) {
return 1;
}
else if (RsDiff >= 12) {
return 4;
}
}
}
else if (tcor >= 55) {
if (Rs < 1617) {
if (Cs < 1205) {
return 2;
}
else if (Cs >= 1205) {
return 1;
}
}
else if (Rs >= 1617) {
return 8;
}
}
}
}
else if (posBalance >= 61) {
if (Rs < 1215) {
if (SC < 1947) {
if (mcTcor < 76) {
if (MVDiff < 139) {
return 8;
}
else if (MVDiff >= 139) {
if (Rs < 769) {
return 2;
}
else if (Rs >= 769) {
return 4;
}
}
}
else if (mcTcor >= 76) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (posBalance < 67) {
return 4;
}
else if (posBalance >= 67) {
if (Rs < 790) {
return 2;
}
else if (Rs >= 790) {
return 1;
}
}
}
}
}
else if (SC >= 1947) {
if (TSC < 9) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (TSC >= 9) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
else if (Rs >= 1215) {
if (refDCval < 102) {
if (CsDiff < 9) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
if (diffMVdiffVal < 7) {
return 8;
}
else if (diffMVdiffVal >= 7) {
return 2;
}
}
}
else if (CsDiff >= 9) {
if (RsDiff < 16) {
return 4;
}
else if (RsDiff >= 16) {
return 8;
}
}
}
else if (refDCval >= 102) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 62) {
if (Rs < 2083) {
if (Rs < 1310) {
if (MVDiff < 10) {
if (tcor < 71) {
if (Rs < 1185) {
return 4;
}
else if (Rs >= 1185) {
return 1;
}
}
else if (tcor >= 71) {
if (Rs < 809) {
return 1;
}
else if (Rs >= 809) {
return 8;
}
}
}
else if (MVDiff >= 10) {
if (mu_mv_mag_sq < 5) {
if (tcor < 78) {
if (SC < 2128) {
if (posBalance < 56) {
if (AFD < 17) {
if (Rs < 746) {
return 8;
}
else if (Rs >= 746) {
return 1;
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (posBalance >= 56) {
if (AFD < 21) {
if (CsDiff < 10) {
if (TSC < 9) {
if (ssDCval < 48) {
return 2;
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (TSC >= 9) {
if (Rs < 1025) {
return 2;
}
else if (Rs >= 1025) {
return 4;
}
}
}
else if (CsDiff >= 10) {
return 2;
}
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (SC >= 2128) {
if (tcor < 69) {
if (Rs < 1156) {
return 8;
}
else if (Rs >= 1156) {
if (Rs < 1233) {
return 1;
}
else if (Rs >= 1233) {
return 4;
}
}
}
else if (tcor >= 69) {
if (tcor < 76) {
return 2;
}
else if (tcor >= 76) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
}
else if (tcor >= 78) {
if (ssDCval < 60) {
if (Rs < 738) {
return 1;
}
else if (Rs >= 738) {
return 4;
}
}
else if (ssDCval >= 60) {
if (diffRsCsdiff < -8) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Rs < 1072) {
return 2;
}
else if (Rs >= 1072) {
return 1;
}
}
}
else if (diffRsCsdiff >= -8) {
if (posBalance < 66) {
if (MVDiff < 34) {
if (MVDiff < 12) {
if (diffRsCsdiff < 12) {
return 8;
}
else if (diffRsCsdiff >= 12) {
return 2;
}
}
else if (MVDiff >= 12) {
return 2;
}
}
else if (MVDiff >= 34) {
return 8;
}
}
else if (posBalance >= 66) {
if (Rs < 1186) {
return 1;
}
else if (Rs >= 1186) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (ssDCval < 84) {
if (AFD < 16) {
if (Rs < 994) {
if (TSC < 7) {
if (Cs < 1103) {
return 8;
}
else if (Cs >= 1103) {
return 2;
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (Rs >= 994) {
return 8;
}
}
else if (AFD >= 16) {
if (SC < 2048) {
if (RsDiff < 6) {
if (Rs < 341) {
return 4;
}
else if (Rs >= 341) {
return 2;
}
}
else if (RsDiff >= 6) {
if (Cs < 920) {
return 4;
}
else if (Cs >= 920) {
if (Cs < 1239) {
return 8;
}
else if (Cs >= 1239) {
return 4;
}
}
}
}
else if (SC >= 2048) {
if (Rs < 1037) {
if (SC < 2172) {
return 2;
}
else if (SC >= 2172) {
return 4;
}
}
else if (Rs >= 1037) {
if (posBalance < 50) {
if (Rs < 1141) {
return 8;
}
else if (Rs >= 1141) {
return 2;
}
}
else if (posBalance >= 50) {
if (Rs < 1308) {
return 4;
}
else if (Rs >= 1308) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (Rs < 911) {
if (SC < 1753) {
if (Cs < 874) {
return 2;
}
else if (Cs >= 874) {
return 1;
}
}
else if (SC >= 1753) {
return 2;
}
}
else if (Rs >= 911) {
if (diffMVdiffVal < 4) {
if (RsDiff < 12) {
if (Cs < 1270) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 1;
}
}
else if (Cs >= 1270) {
return 2;
}
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (diffMVdiffVal >= 4) {
if (Rs < 1015) {
return 1;
}
else if (Rs >= 1015) {
if (Cs < 989) {
return 1;
}
else if (Cs >= 989) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1310) {
if (Cs < 1002) {
if (refDCval < 80) {
if (Cs < 896) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
if (ssDCval < 63) {
if (RsDiff < 11) {
return 2;
}
else if (RsDiff >= 11) {
return 8;
}
}
else if (ssDCval >= 63) {
if (RsDiff < 10) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
return 4;
}
}
}
else if (RsDiff >= 10) {
return 2;
}
}
}
}
else if (Cs >= 896) {
if (RsDiff < 12) {
if (gchDC < 1) {
if (RsDiff < 8) {
if (Rs < 1949) {
return 8;
}
else if (Rs >= 1949) {
return 2;
}
}
else if (RsDiff >= 8) {
if (ssDCval < 70) {
if (Rs < 1466) {
return 1;
}
else if (Rs >= 1466) {
return 4;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (Rs < 2041) {
return 8;
}
else if (Rs >= 2041) {
return 4;
}
}
}
else if (RsDiff >= 12) {
if (Cs < 985) {
if (negBalance < 53) {
return 2;
}
else if (negBalance >= 53) {
return 8;
}
}
else if (Cs >= 985) {
return 4;
}
}
}
}
else if (refDCval >= 80) {
if (refDCval < 90) {
if (tcor < 73) {
if (RsDiff < 9) {
if (Rs < 1719) {
return 1;
}
else if (Rs >= 1719) {
return 2;
}
}
else if (RsDiff >= 9) {
return 8;
}
}
else if (tcor >= 73) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (Cs < 998) {
return 4;
}
else if (Cs >= 998) {
return 1;
}
}
}
}
else if (refDCval >= 90) {
if (refDCval < 101) {
if (RsDiff < 11) {
if (Cs < 884) {
return 8;
}
else if (Cs >= 884) {
if (TSC < 10) {
if (Rs < 1603) {
return 8;
}
else if (Rs >= 1603) {
return 1;
}
}
else if (TSC >= 10) {
return 1;
}
}
}
else if (RsDiff >= 11) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (Rs < 1909) {
return 8;
}
else if (Rs >= 1909) {
return 4;
}
}
}
}
else if (refDCval >= 101) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
if (Rs < 1500) {
return 4;
}
else if (Rs >= 1500) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1002) {
if (AFD < 19) {
if (TSC < 10) {
if (refDCval < 85) {
if (diffMVdiffVal < 1) {
if (Rs < 1362) {
if (Rs < 1322) {
return 8;
}
else if (Rs >= 1322) {
return 4;
}
}
else if (Rs >= 1362) {
if (CsDiff < 10) {
if (ssDCval < 84) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (MVDiff < 69) {
if (mu_mv_mag_sq < 5) {
if (ssDCval < 56) {
if (Cs < 1143) {
return 8;
}
else if (Cs >= 1143) {
return 4;
}
}
else if (ssDCval >= 56) {
if (Rs < 2030) {
return 8;
}
else if (Rs >= 2030) {
if (Rs < 2055) {
return 4;
}
else if (Rs >= 2055) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
return 8;
}
}
else if (MVDiff >= 69) {
return 4;
}
}
}
else if (ssDCval >= 84) {
return 1;
}
}
else if (CsDiff >= 10) {
return 1;
}
}
}
else if (diffMVdiffVal >= 1) {
if (mu_mv_mag_sq < 39) {
if (tcor < 84) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 1513) {
if (Rs < 1439) {
return 1;
}
else if (Rs >= 1439) {
return 2;
}
}
else if (Rs >= 1513) {
if (Rs < 1655) {
return 4;
}
else if (Rs >= 1655) {
return 8;
}
}
}
}
else if (tcor >= 84) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (Rs < 1471) {
if (Rs < 1374) {
return 4;
}
else if (Rs >= 1374) {
return 1;
}
}
else if (Rs >= 1471) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (posBalance < 47) {
if (Cs < 1195) {
return 2;
}
else if (Cs >= 1195) {
return 4;
}
}
else if (posBalance >= 47) {
if (SC < 2840) {
if (Rs < 1544) {
return 8;
}
else if (Rs >= 1544) {
return 2;
}
}
else if (SC >= 2840) {
return 8;
}
}
}
}
}
else if (refDCval >= 85) {
if (MVDiff < 11) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (tcor < 81) {
if (Rs < 1966) {
return 2;
}
else if (Rs >= 1966) {
return 4;
}
}
else if (tcor >= 81) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
else if (MVDiff >= 11) {
if (mcTcor < 93) {
if (diffRsCsdiff < 22) {
return 2;
}
else if (diffRsCsdiff >= 22) {
if (ssDCval < 99) {
return 2;
}
else if (ssDCval >= 99) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (Cs < 1209) {
if (Rs < 1833) {
return 8;
}
else if (Rs >= 1833) {
return 2;
}
}
else if (Cs >= 1209) {
if (ssDCval < 96) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
if (Rs < 1461) {
return 2;
}
else if (Rs >= 1461) {
return 8;
}
}
}
else if (ssDCval >= 96) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 10) {
if (refDCval < 103) {
return 8;
}
else if (refDCval >= 103) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
if (Cs < 1063) {
return 1;
}
else if (Cs >= 1063) {
return 4;
}
}
}
}
}
else if (AFD >= 19) {
if (ssDCval < 101) {
if (Cs < 1087) {
if (RsDiff < 9) {
if (Rs < 1771) {
return 1;
}
else if (Rs >= 1771) {
return 8;
}
}
else if (RsDiff >= 9) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
return 8;
}
}
}
else if (Cs >= 1087) {
if (RsCsDiff < 176) {
if (MVDiff < 22) {
if (refDCval < 91) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
if (Rs < 1740) {
return 1;
}
else if (Rs >= 1740) {
return 4;
}
}
}
else if (refDCval >= 91) {
return 1;
}
}
else if (MVDiff >= 22) {
if (diffRsCsdiff < 18) {
if (CsDiff < 7) {
return 1;
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (diffRsCsdiff >= 18) {
if (Rs < 1961) {
return 8;
}
else if (Rs >= 1961) {
return 1;
}
}
}
}
else if (RsCsDiff >= 176) {
if (Rs < 1744) {
if (AFD < 30) {
return 4;
}
else if (AFD >= 30) {
return 2;
}
}
else if (Rs >= 1744) {
return 2;
}
}
}
}
else if (ssDCval >= 101) {
if (Rs < 1626) {
return 8;
}
else if (Rs >= 1626) {
return 2;
}
}
}
}
}
}
else if (Rs >= 2083) {
if (MVDiff < 26) {
if (Cs < 1181) {
if (Rs < 2286) {
if (RsCsDiff < 85) {
if (diffAFD < 2) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (Rs < 2260) {
return 8;
}
else if (Rs >= 2260) {
return 2;
}
}
}
}
}
else if (diffAFD >= 2) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (RsCsDiff >= 85) {
if (SC < 3009) {
return 4;
}
else if (SC >= 3009) {
if (refDCval < 56) {
return 4;
}
else if (refDCval >= 56) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
if (AFD < 17) {
if (Rs < 2156) {
return 4;
}
else if (Rs >= 2156) {
if (Rs < 2238) {
return 8;
}
else if (Rs >= 2238) {
return 2;
}
}
}
else if (AFD >= 17) {
if (diffRsCsdiff < 9) {
return 2;
}
else if (diffRsCsdiff >= 9) {
if (Cs < 922) {
return 1;
}
else if (Cs >= 922) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2286) {
if (mcTcor < 92) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (mcTcor >= 92) {
if (RsCsDiff < 85) {
if (Cs < 976) {
return 2;
}
else if (Cs >= 976) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
else if (RsCsDiff >= 85) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
}
else if (Cs >= 1181) {
if (tcor < 87) {
if (Rs < 2138) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 2094) {
return 2;
}
else if (Rs >= 2094) {
return 1;
}
}
}
else if (Rs >= 2138) {
if (RsDiff < 11) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
return 8;
}
}
else if (RsDiff >= 11) {
if (Rs < 2348) {
return 4;
}
else if (Rs >= 2348) {
return 2;
}
}
}
}
else if (tcor >= 87) {
return 8;
}
}
}
else if (MVDiff >= 26) {
if (MVDiff < 37) {
if (mcTcor < 88) {
if (mu_mv_mag_sq < 28) {
if (RsDiff < 14) {
return 1;
}
else if (RsDiff >= 14) {
return 2;
}
}
else if (mu_mv_mag_sq >= 28) {
return 2;
}
}
else if (mcTcor >= 88) {
if (TSC < 11) {
if (MVDiff < 36) {
if (posBalance < 48) {
if (tcor < 83) {
return 2;
}
else if (tcor >= 83) {
if (Rs < 2250) {
return 4;
}
else if (Rs >= 2250) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (Cs < 919) {
return 2;
}
else if (Cs >= 919) {
return 8;
}
}
}
else if (MVDiff >= 36) {
return 1;
}
}
else if (TSC >= 11) {
return 4;
}
}
}
else if (MVDiff >= 37) {
if (SC < 3372) {
if (Rs < 2116) {
return 4;
}
else if (Rs >= 2116) {
if (RsDiff < 8) {
if (SC < 3324) {
return 4;
}
else if (SC >= 3324) {
return 8;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
}
else if (SC >= 3372) {
if (Rs < 2379) {
if (mcTcor < 86) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (mcTcor >= 86) {
return 2;
}
}
else if (Rs >= 2379) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (MVDiff < 25) {
if (AFD < 15) {
if (ssDCval < 120) {
if (Cs < 1070) {
return 8;
}
else if (Cs >= 1070) {
return 2;
}
}
else if (ssDCval >= 120) {
return 8;
}
}
else if (AFD >= 15) {
if (diffRsCsdiff < 58) {
if (gchDC < 2) {
if (posBalance < 62) {
if (Cs < 1110) {
return 1;
}
else if (Cs >= 1110) {
return 8;
}
}
else if (posBalance >= 62) {
if (tcor < 79) {
if (posBalance < 64) {
return 2;
}
else if (posBalance >= 64) {
return 8;
}
}
else if (tcor >= 79) {
if (Cs < 1089) {
return 2;
}
else if (Cs >= 1089) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (refDCval < 120) {
if (Rs < 1727) {
return 8;
}
else if (Rs >= 1727) {
return 1;
}
}
else if (refDCval >= 120) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 58) {
return 2;
}
}
}
else if (MVDiff >= 25) {
if (refDCval < 137) {
if (negBalance < 67) {
if (posBalance < 74) {
if (RsDiff < 8) {
if (Rs < 1557) {
return 8;
}
else if (Rs >= 1557) {
return 1;
}
}
else if (RsDiff >= 8) {
if (CsDiff < 13) {
if (tcor < 90) {
if (Cs < 1306) {
return 1;
}
else if (Cs >= 1306) {
if (Rs < 1585) {
return 8;
}
else if (Rs >= 1585) {
return 1;
}
}
}
else if (tcor >= 90) {
return 8;
}
}
else if (CsDiff >= 13) {
return 8;
}
}
}
else if (posBalance >= 74) {
if (Rs < 1493) {
if (Rs < 1209) {
return 8;
}
else if (Rs >= 1209) {
return 4;
}
}
else if (Rs >= 1493) {
return 2;
}
}
}
else if (negBalance >= 67) {
if (Cs < 1245) {
if (Rs < 1657) {
return 2;
}
else if (Rs >= 1657) {
return 8;
}
}
else if (Cs >= 1245) {
return 4;
}
}
}
else if (refDCval >= 137) {
if (Rs < 2311) {
if (Rs < 1115) {
if (MVDiff < 317) {
if (RsDiff < 10) {
return 2;
}
else if (RsDiff >= 10) {
if (Rs < 802) {
return 8;
}
else if (Rs >= 802) {
return 1;
}
}
}
else if (MVDiff >= 317) {
return 1;
}
}
else if (Rs >= 1115) {
if (RsCsDiff < 110) {
return 8;
}
else if (RsCsDiff >= 110) {
if (mcTcor < 88) {
if (mu_mv_mag_sq < 17) {
return 4;
}
else if (mu_mv_mag_sq >= 17) {
if (diffRsCsdiff < -16) {
return 4;
}
else if (diffRsCsdiff >= -16) {
return 8;
}
}
}
else if (mcTcor >= 88) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
else if (Rs >= 2311) {
return 1;
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (Rs < 1758) {
if (Rs < 1127) {
if (ssDCval < 63) {
if (Rs < 805) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (Rs < 539) {
if (SC < 1061) {
if (ssDCval < 56) {
if (negBalance < 42) {
if (mu_mv_mag_sq < 42) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 1;
}
}
else if (mu_mv_mag_sq >= 42) {
return 1;
}
}
else if (negBalance >= 42) {
return 8;
}
}
else if (ssDCval >= 56) {
if (tcor < 94) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (Rs < 267) {
return 1;
}
else if (Rs >= 267) {
return 2;
}
}
}
else if (tcor >= 94) {
return 4;
}
}
}
else if (SC >= 1061) {
if (AFD < 7) {
if (posBalance < 49) {
if (Cs < 894) {
if (ssDCval < 58) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
return 4;
}
}
else if (ssDCval >= 58) {
if (Cs < 634) {
return 2;
}
else if (Cs >= 634) {
if (Rs < 412) {
return 8;
}
else if (Rs >= 412) {
return 4;
}
}
}
}
else if (Cs >= 894) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (Cs < 1032) {
return 4;
}
else if (Cs >= 1032) {
return 2;
}
}
}
}
else if (posBalance >= 49) {
return 4;
}
}
else if (AFD >= 7) {
if (SC < 1100) {
return 2;
}
else if (SC >= 1100) {
return 1;
}
}
}
}
else if (Rs >= 539) {
if (MVDiff < 1) {
if (Cs < 960) {
if (tcor < 97) {
if (Rs < 647) {
if (Cs < 741) {
return 8;
}
else if (Cs >= 741) {
return 1;
}
}
else if (Rs >= 647) {
return 4;
}
}
else if (tcor >= 97) {
if (Cs < 931) {
if (Cs < 535) {
return 4;
}
else if (Cs >= 535) {
return 8;
}
}
else if (Cs >= 931) {
if (Cs < 949) {
return 1;
}
else if (Cs >= 949) {
return 8;
}
}
}
}
else if (Cs >= 960) {
if (ssDCval < 52) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
else if (ssDCval >= 52) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 19) {
if (ssDCval < 44) {
return 2;
}
else if (ssDCval >= 44) {
return 1;
}
}
else if (negBalance >= 19) {
if (negBalance < 29) {
if (TSC < 3) {
if (SC < 1380) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (SC >= 1380) {
return 1;
}
}
else if (TSC >= 3) {
if (SC < 1194) {
return 8;
}
else if (SC >= 1194) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
return 4;
}
}
}
}
else if (negBalance >= 29) {
if (MVDiff < 14) {
if (RsCsDiff < 16) {
if (refDCval < 47) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
else if (refDCval >= 47) {
if (TSC < 3) {
if (Cs < 865) {
return 8;
}
else if (Cs >= 865) {
return 1;
}
}
else if (TSC >= 3) {
if (posBalance < 53) {
if (ssDCval < 55) {
if (RsCsDiff < 11) {
return 4;
}
else if (RsCsDiff >= 11) {
if (Rs < 689) {
return 2;
}
else if (Rs >= 689) {
return 1;
}
}
}
else if (ssDCval >= 55) {
if (mu_mv_mag_sq < 73) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
return 4;
}
}
else if (mu_mv_mag_sq >= 73) {
return 8;
}
}
}
else if (posBalance >= 53) {
if (Rs < 588) {
return 4;
}
else if (Rs >= 588) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 16) {
if (AFD < 9) {
if (ssDCval < 44) {
return 2;
}
else if (ssDCval >= 44) {
return 4;
}
}
else if (AFD >= 9) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (Rs < 694) {
return 8;
}
else if (Rs >= 694) {
return 4;
}
}
}
}
}
else if (MVDiff >= 14) {
if (RsDiff < 2) {
if (Rs < 752) {
return 8;
}
else if (Rs >= 752) {
return 4;
}
}
else if (RsDiff >= 2) {
if (Cs < 963) {
return 1;
}
else if (Cs >= 963) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 805) {
if (ssDCval < 50) {
if (diffMVdiffVal < -11) {
return 4;
}
else if (diffMVdiffVal >= -11) {
if (Cs < 569) {
return 1;
}
else if (Cs >= 569) {
if (MVDiff < 11) {
if (Rs < 886) {
if (Cs < 1056) {
return 8;
}
else if (Cs >= 1056) {
return 1;
}
}
else if (Rs >= 886) {
return 8;
}
}
else if (MVDiff >= 11) {
if (Rs < 847) {
return 8;
}
else if (Rs >= 847) {
if (Rs < 945) {
return 4;
}
else if (Rs >= 945) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 50) {
if (negBalance < 45) {
if (RsCsDiff < 12) {
if (ssDCval < 53) {
if (negBalance < 24) {
if (Cs < 786) {
return 8;
}
else if (Cs >= 786) {
if (Rs < 1111) {
return 1;
}
else if (Rs >= 1111) {
return 8;
}
}
}
else if (negBalance >= 24) {
if (Rs < 917) {
if (Rs < 826) {
return 8;
}
else if (Rs >= 826) {
return 1;
}
}
else if (Rs >= 917) {
return 8;
}
}
}
else if (ssDCval >= 53) {
if (diffAFD < -2) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 127) {
if (Rs < 870) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (TSC < 4) {
if (Rs < 849) {
return 1;
}
else if (Rs >= 849) {
return 4;
}
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (Rs >= 870) {
if (mu_mv_mag_sq < 71) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
return 8;
}
}
else if (mu_mv_mag_sq >= 71) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 942) {
return 8;
}
else if (Rs >= 942) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 12) {
if (Cs < 878) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (Rs < 896) {
return 4;
}
else if (Rs >= 896) {
return 1;
}
}
}
else if (Cs >= 878) {
if (Cs < 1248) {
return 4;
}
else if (Cs >= 1248) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 863) {
return 8;
}
else if (Rs >= 863) {
return 4;
}
}
}
}
}
else if (ssDCval >= 63) {
if (SC < 2156) {
if (Cs < 722) {
if (mu_mv_mag_sq < 114) {
if (ssDCval < 85) {
if (refDCval < 78) {
if (mu_mv_mag_sq < 76) {
if (AFD < 10) {
if (negBalance < 48) {
if (ssDCval < 66) {
if (mu_mv_mag_sq < 68) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (Rs < 619) {
return 1;
}
else if (Rs >= 619) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (Rs < 554) {
return 4;
}
else if (Rs >= 554) {
return 8;
}
}
}
else if (ssDCval >= 66) {
if (MVDiff < 12) {
if (Rs < 817) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (Rs >= 817) {
return 1;
}
}
else if (MVDiff >= 12) {
if (Rs < 574) {
return 1;
}
else if (Rs >= 574) {
return 4;
}
}
}
}
else if (negBalance >= 48) {
return 8;
}
}
else if (AFD >= 10) {
if (Cs < 712) {
if (MVDiff < 59) {
return 1;
}
else if (MVDiff >= 59) {
if (Rs < 659) {
return 2;
}
else if (Rs >= 659) {
return 4;
}
}
}
else if (Cs >= 712) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (Rs < 614) {
if (Cs < 601) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
if (Cs < 550) {
if (Rs < 179) {
return 1;
}
else if (Rs >= 179) {
return 4;
}
}
else if (Cs >= 550) {
return 1;
}
}
}
else if (Cs >= 601) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (Rs >= 614) {
if (RsDiff < 2) {
if (CsDiff < 1) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
return 2;
}
}
else if (CsDiff >= 1) {
if (Cs < 567) {
return 4;
}
else if (Cs >= 567) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (AFD < 4) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (AFD >= 4) {
if (Cs < 651) {
if (Rs < 939) {
return 8;
}
else if (Rs >= 939) {
return 1;
}
}
else if (Cs >= 651) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 78) {
if (Cs < 707) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
if (Cs < 691) {
return 8;
}
else if (Cs >= 691) {
if (Rs < 822) {
return 2;
}
else if (Rs >= 822) {
return 8;
}
}
}
}
else if (Cs >= 707) {
return 1;
}
}
}
else if (ssDCval >= 85) {
if (Rs < 647) {
return 2;
}
else if (Rs >= 647) {
if (Rs < 695) {
return 1;
}
else if (Rs >= 695) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (Rs < 804) {
if (SC < 1422) {
if (mu_mv_mag_sq < 119) {
if (Cs < 599) {
return 2;
}
else if (Cs >= 599) {
if (Cs < 648) {
return 4;
}
else if (Cs >= 648) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (diffRsCsdiff < -1) {
if (Rs < 421) {
return 1;
}
else if (Rs >= 421) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
else if (SC >= 1422) {
return 1;
}
}
else if (Rs >= 804) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (AFD < 1) {
if (Rs < 924) {
return 2;
}
else if (Rs >= 924) {
return 8;
}
}
else if (AFD >= 1) {
if (negBalance < 29) {
if (Rs < 1079) {
return 2;
}
else if (Rs >= 1079) {
if (Rs < 1124) {
if (Rs < 1104) {
return 4;
}
else if (Rs >= 1104) {
return 1;
}
}
else if (Rs >= 1124) {
return 2;
}
}
}
else if (negBalance >= 29) {
return 4;
}
}
}
}
}
}
else if (Cs >= 722) {
if (ssDCval < 81) {
if (AFD < 6) {
if (ssDCval < 75) {
if (Rs < 718) {
if (Cs < 778) {
if (MVDiff < 4) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (Cs >= 778) {
if (diffRsCsdiff < 3) {
if (refDCval < 73) {
if (diffRsCsdiff < -9) {
return 2;
}
else if (diffRsCsdiff >= -9) {
return 8;
}
}
else if (refDCval >= 73) {
if (Rs < 561) {
return 8;
}
else if (Rs >= 561) {
return 1;
}
}
}
else if (diffRsCsdiff >= 3) {
if (ssDCval < 67) {
if (Rs < 645) {
return 8;
}
else if (Rs >= 645) {
return 2;
}
}
else if (ssDCval >= 67) {
return 1;
}
}
}
}
else if (Rs >= 718) {
if (AFD < 4) {
if (SC < 1769) {
if (ssDCval < 74) {
if (Rs < 899) {
if (Cs < 730) {
if (Rs < 763) {
return 8;
}
else if (Rs >= 763) {
return 1;
}
}
else if (Cs >= 730) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
return 4;
}
}
}
else if (Rs >= 899) {
return 2;
}
}
else if (ssDCval >= 74) {
return 1;
}
}
else if (SC >= 1769) {
if (negBalance < 23) {
if (posBalance < 45) {
if (Rs < 1021) {
return 8;
}
else if (Rs >= 1021) {
return 2;
}
}
else if (posBalance >= 45) {
return 4;
}
}
else if (negBalance >= 23) {
if (RsDiff < 2) {
if (diffRsCsdiff < -6) {
return 2;
}
else if (diffRsCsdiff >= -6) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 913) {
return 2;
}
else if (Rs >= 913) {
return 1;
}
}
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (AFD >= 4) {
if (negBalance < 34) {
if (diffTSC < 1) {
if (TSC < 4) {
if (MVDiff < 1) {
if (Cs < 770) {
return 1;
}
else if (Cs >= 770) {
return 8;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (TSC >= 4) {
if (Rs < 903) {
return 2;
}
else if (Rs >= 903) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (Cs < 771) {
return 4;
}
else if (Cs >= 771) {
if (Rs < 854) {
return 1;
}
else if (Rs >= 854) {
return 2;
}
}
}
}
else if (negBalance >= 34) {
if (negBalance < 38) {
if (Rs < 978) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
return 4;
}
}
else if (Rs >= 978) {
return 8;
}
}
else if (negBalance >= 38) {
if (refDCval < 71) {
return 8;
}
else if (refDCval >= 71) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (Cs < 1156) {
if (refDCval < 79) {
return 8;
}
else if (refDCval >= 79) {
if (RsDiff < 2) {
if (diffRsCsdiff < -4) {
return 8;
}
else if (diffRsCsdiff >= -4) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
return 4;
}
}
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (Cs >= 1156) {
if (Rs < 788) {
return 1;
}
else if (Rs >= 788) {
return 2;
}
}
}
}
else if (AFD >= 6) {
if (Rs < 490) {
if (mu_mv_mag_sq < 13) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (mu_mv_mag_sq >= 13) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
if (Rs < 323) {
return 2;
}
else if (Rs >= 323) {
return 4;
}
}
}
}
else if (Rs >= 490) {
if (posBalance < 58) {
if (Rs < 1098) {
if (refDCval < 82) {
if (mu_mv_mag_sq < 79) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 79) {
if (Rs < 792) {
return 8;
}
else if (Rs >= 792) {
return 1;
}
}
}
else if (refDCval >= 82) {
if (Rs < 933) {
return 4;
}
else if (Rs >= 933) {
return 8;
}
}
}
else if (Rs >= 1098) {
if (Rs < 1103) {
return 1;
}
else if (Rs >= 1103) {
if (Rs < 1114) {
return 4;
}
else if (Rs >= 1114) {
return 8;
}
}
}
}
else if (posBalance >= 58) {
if (SC < 1896) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (SC >= 1896) {
if (MVDiff < 3) {
if (SC < 2114) {
return 8;
}
else if (SC >= 2114) {
return 4;
}
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (Rs < 724) {
if (Cs < 757) {
if (Rs < 603) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 1;
}
}
else if (Rs >= 603) {
return 1;
}
}
else if (Cs >= 757) {
if (negBalance < 35) {
if (ssDCval < 84) {
if (Rs < 656) {
return 1;
}
else if (Rs >= 656) {
return 2;
}
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (negBalance >= 35) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (ssDCval < 85) {
return 1;
}
else if (ssDCval >= 85) {
return 2;
}
}
}
}
}
else if (Rs >= 724) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
if (Rs < 955) {
return 4;
}
else if (Rs >= 955) {
if (Rs < 1048) {
return 8;
}
else if (Rs >= 1048) {
return 2;
}
}
}
}
}
}
}
}
else if (SC >= 2156) {
if (tcor < 98) {
if (RsDiff < 2) {
if (Rs < 988) {
return 2;
}
else if (Rs >= 988) {
return 8;
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 4) {
if (Rs < 1036) {
return 2;
}
else if (Rs >= 1036) {
return 8;
}
}
else if (mu_mv_mag_sq >= 4) {
if (posBalance < 49) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (posBalance >= 49) {
if (negBalance < 28) {
return 2;
}
else if (negBalance >= 28) {
return 4;
}
}
}
}
}
else if (tcor >= 98) {
if (diffRsCsdiff < -3) {
if (Rs < 1053) {
return 8;
}
else if (Rs >= 1053) {
return 4;
}
}
else if (diffRsCsdiff >= -3) {
return 4;
}
}
}
}
}
else if (Rs >= 1127) {
if (Cs < 1052) {
if (mu_mv_mag_sq < 124) {
if (SC < 2391) {
if (TSC < 5) {
if (Cs < 963) {
if (refDCval < 47) {
if (Cs < 615) {
return 8;
}
else if (Cs >= 615) {
if (mu_mv_mag_sq < 122) {
return 4;
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 1394) {
return 4;
}
else if (Rs >= 1394) {
return 4;
}
}
}
}
else if (refDCval >= 47) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 79) {
if (MVDiff < 15) {
if (Rs < 1326) {
if (Rs < 1324) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 8;
}
}
else if (Rs >= 1324) {
return 2;
}
}
else if (Rs >= 1326) {
if (Cs < 845) {
return 1;
}
else if (Cs >= 845) {
if (Cs < 857) {
return 4;
}
else if (Cs >= 857) {
return 1;
}
}
}
}
else if (MVDiff >= 15) {
if (Cs < 717) {
return 4;
}
else if (Cs >= 717) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (SC < 2112) {
if (MVDiff < 8) {
if (RsCsDiff < 16) {
if (diffRsCsdiff < -4) {
if (Rs < 1177) {
return 1;
}
else if (Rs >= 1177) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
if (Cs < 612) {
if (mu_mv_mag_sq < 120) {
return 1;
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 1240) {
return 1;
}
else if (Rs >= 1240) {
return 8;
}
}
}
else if (Cs >= 612) {
return 1;
}
}
}
else if (RsCsDiff >= 16) {
return 8;
}
}
else if (MVDiff >= 8) {
if (refDCval < 51) {
return 1;
}
else if (refDCval >= 51) {
return 8;
}
}
}
else if (SC >= 2112) {
if (mu_mv_mag_sq < 106) {
if (ssDCval < 74) {
if (ssDCval < 61) {
if (Rs < 1545) {
return 8;
}
else if (Rs >= 1545) {
return 1;
}
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (mu_mv_mag_sq >= 106) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
if (Rs < 1355) {
return 8;
}
else if (Rs >= 1355) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 70) {
if (ssDCval < 59) {
if (SC < 2069) {
return 1;
}
else if (SC >= 2069) {
if (Rs < 1251) {
return 1;
}
else if (Rs >= 1251) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
else if (ssDCval >= 59) {
if (AFD < 5) {
if (SC < 1798) {
if (Rs < 1131) {
return 1;
}
else if (Rs >= 1131) {
return 2;
}
}
else if (SC >= 1798) {
return 1;
}
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (ssDCval >= 70) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
return 1;
}
}
}
}
}
else if (Cs >= 963) {
if (ssDCval < 65) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
return 8;
}
}
else if (ssDCval >= 65) {
return 8;
}
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 61) {
if (SC < 2084) {
if (Cs < 646) {
return 1;
}
else if (Cs >= 646) {
return 2;
}
}
else if (SC >= 2084) {
if (negBalance < 52) {
if (Rs < 1403) {
return 4;
}
else if (Rs >= 1403) {
return 1;
}
}
else if (negBalance >= 52) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (Rs < 1553) {
return 8;
}
else if (Rs >= 1553) {
return 1;
}
}
}
}
else if (SC >= 2391) {
if (ssDCval < 52) {
if (Cs < 926) {
return 4;
}
else if (Cs >= 926) {
return 1;
}
}
else if (ssDCval >= 52) {
if (ssDCval < 72) {
if (diffRsCsdiff < -18) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Rs < 1708) {
return 8;
}
else if (Rs >= 1708) {
return 2;
}
}
}
else if (diffRsCsdiff >= -18) {
if (Cs < 1008) {
if (Cs < 800) {
return 1;
}
else if (Cs >= 800) {
return 8;
}
}
else if (Cs >= 1008) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
return 2;
}
}
}
}
else if (ssDCval >= 72) {
if (Rs < 1536) {
return 8;
}
else if (Rs >= 1536) {
if (Cs < 1034) {
return 1;
}
else if (Cs >= 1034) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (TSCindex < 2) {
if (posBalance < 39) {
if (RsDiff < 1) {
if (MVDiff < 1) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (posBalance >= 39) {
if (Rs < 1256) {
return 4;
}
else if (Rs >= 1256) {
return 8;
}
}
}
else if (TSCindex >= 2) {
if (Rs < 1321) {
if (Cs < 750) {
if (Rs < 1165) {
return 2;
}
else if (Rs >= 1165) {
if (Rs < 1283) {
return 1;
}
else if (Rs >= 1283) {
return 8;
}
}
}
else if (Cs >= 750) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (negBalance < 19) {
if (Cs < 873) {
return 8;
}
else if (Cs >= 873) {
return 1;
}
}
else if (negBalance >= 19) {
return 8;
}
}
}
}
else if (Rs >= 1321) {
if (Rs < 1734) {
if (ssDCval < 74) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 74) {
if (Cs < 793) {
return 1;
}
else if (Cs >= 793) {
return 8;
}
}
}
else if (Rs >= 1734) {
if (Rs < 1752) {
return 4;
}
else if (Rs >= 1752) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1052) {
if (RsCsDiff < 2) {
if (SC < 2714) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
if (Rs < 1459) {
return 8;
}
else if (Rs >= 1459) {
return 1;
}
}
}
else if (SC >= 2714) {
if (Rs < 1683) {
return 4;
}
else if (Rs >= 1683) {
return 8;
}
}
}
else if (RsCsDiff >= 2) {
if (AFD < 4) {
if (mu_mv_mag_sq < 126) {
if (MVDiff < 7) {
if (ssDCval < 57) {
if (mu_mv_mag_sq < 123) {
if (Rs < 1332) {
return 4;
}
else if (Rs >= 1332) {
return 8;
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1707) {
if (negBalance < 25) {
if (negBalance < 21) {
if (CsDiff < 2) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (SC < 2842) {
if (Rs < 1691) {
return 4;
}
else if (Rs >= 1691) {
return 4;
}
}
else if (SC >= 2842) {
return 8;
}
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (negBalance >= 21) {
return 4;
}
}
else if (negBalance >= 25) {
return 8;
}
}
else if (Rs >= 1707) {
return 4;
}
}
}
else if (ssDCval >= 57) {
if (Cs < 1135) {
if (mu_mv_mag_sq < 124) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (Cs < 1103) {
return 8;
}
else if (Cs >= 1103) {
return 4;
}
}
}
else if (Cs >= 1135) {
if (Rs < 1162) {
if (Rs < 1140) {
return 1;
}
else if (Rs >= 1140) {
return 4;
}
}
else if (Rs >= 1162) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (mu_mv_mag_sq >= 126) {
if (ssDCval < 74) {
if (ssDCval < 59) {
if (SC < 2917) {
if (Cs < 1201) {
if (Cs < 1184) {
if (refDCval < 47) {
return 4;
}
else if (refDCval >= 47) {
if (Rs < 1658) {
if (Rs < 1586) {
return 4;
}
else if (Rs >= 1586) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 8;
}
}
}
else if (Rs >= 1658) {
if (SC < 2793) {
return 4;
}
else if (SC >= 2793) {
if (negBalance < 18) {
if (Rs < 1745) {
if (Rs < 1733) {
return 8;
}
else if (Rs >= 1733) {
return 4;
}
}
else if (Rs >= 1745) {
return 8;
}
}
else if (negBalance >= 18) {
if (mcTcor < 99) {
if (Rs < 1749) {
if (posBalance < 32) {
if (Rs < 1740) {
return 8;
}
else if (Rs >= 1740) {
return 4;
}
}
else if (posBalance >= 32) {
if (Rs < 1732) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (RsDiff < 2) {
if (Cs < 1090) {
return 4;
}
else if (Cs >= 1090) {
return 4;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (Rs >= 1732) {
return 4;
}
}
}
else if (Rs >= 1749) {
return 8;
}
}
else if (mcTcor >= 99) {
if (Cs < 1098) {
return 8;
}
else if (Cs >= 1098) {
if (SC < 2810) {
return 8;
}
else if (SC >= 2810) {
if (diffRsCsdiff < 2) {
if (AFD < 2) {
if (Rs < 1710) {
return 4;
}
else if (Rs >= 1710) {
return 4;
}
}
else if (AFD >= 2) {
return 4;
}
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Cs >= 1184) {
return 4;
}
}
else if (Cs >= 1201) {
return 8;
}
}
else if (SC >= 2917) {
return 4;
}
}
else if (ssDCval >= 59) {
if (posBalance < 45) {
if (Cs < 1109) {
return 8;
}
else if (Cs >= 1109) {
return 4;
}
}
else if (posBalance >= 45) {
return 4;
}
}
}
else if (ssDCval >= 74) {
return 8;
}
}
}
else if (AFD >= 4) {
if (posBalance < 44) {
if (AFD < 10) {
if (posBalance < 36) {
return 1;
}
else if (posBalance >= 36) {
return 8;
}
}
else if (AFD >= 10) {
return 2;
}
}
else if (posBalance >= 44) {
if (Rs < 1353) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (mcTcor < 98) {
if (diffTSC < 1) {
if (RsCsDiff < 11) {
if (Rs < 1230) {
if (Cs < 1274) {
return 4;
}
else if (Cs >= 1274) {
return 8;
}
}
else if (Rs >= 1230) {
return 8;
}
}
else if (RsCsDiff >= 11) {
return 4;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (mcTcor >= 98) {
return 4;
}
}
}
else if (Rs >= 1353) {
if (AFD < 10) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
if (Rs < 1682) {
return 8;
}
else if (Rs >= 1682) {
return 2;
}
}
}
else if (AFD >= 10) {
if (Rs < 1644) {
return 4;
}
else if (Rs >= 1644) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1758) {
if (TSC < 4) {
if (refDCval < 83) {
if (ssDCval < 56) {
if (Rs < 2427) {
if (mu_mv_mag_sq < 97) {
return 4;
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 853) {
return 8;
}
else if (Cs >= 853) {
if (SC < 2958) {
if (Rs < 1823) {
if (posBalance < 31) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (Rs < 1798) {
if (Rs < 1787) {
return 8;
}
else if (Rs >= 1787) {
return 8;
}
}
else if (Rs >= 1798) {
return 4;
}
}
}
else if (posBalance >= 31) {
if (Rs < 1807) {
if (Rs < 1774) {
if (Rs < 1768) {
if (Cs < 1158) {
if (posBalance < 34) {
if (Cs < 1120) {
return 4;
}
else if (Cs >= 1120) {
if (Rs < 1763) {
return 4;
}
else if (Rs >= 1763) {
return 4;
}
}
}
else if (posBalance >= 34) {
if (Rs < 1761) {
return 4;
}
else if (Rs >= 1761) {
return 8;
}
}
}
else if (Cs >= 1158) {
return 4;
}
}
else if (Rs >= 1768) {
return 8;
}
}
else if (Rs >= 1774) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Cs < 1142) {
if (SC < 2915) {
if (diffAFD < 1) {
if (Rs < 1780) {
if (negBalance < 19) {
if (Rs < 1778) {
return 4;
}
else if (Rs >= 1778) {
return 4;
}
}
else if (negBalance >= 19) {
return 4;
}
}
else if (Rs >= 1780) {
return 4;
}
}
else if (diffAFD >= 1) {
if (Rs < 1793) {
return 4;
}
else if (Rs >= 1793) {
return 4;
}
}
}
else if (SC >= 2915) {
return 8;
}
}
else if (Cs >= 1142) {
return 4;
}
}
}
}
else if (Rs >= 1807) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
if (Rs < 1815) {
return 4;
}
else if (Rs >= 1815) {
return 8;
}
}
}
}
}
else if (Rs >= 1823) {
return 4;
}
}
else if (SC >= 2958) {
if (SC < 3323) {
return 4;
}
else if (SC >= 3323) {
if (Cs < 1241) {
return 4;
}
else if (Cs >= 1241) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2427) {
if (Cs < 698) {
return 1;
}
else if (Cs >= 698) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (mcTcor < 99) {
if (Cs < 825) {
if (mu_mv_mag_sq < 123) {
return 1;
}
else if (mu_mv_mag_sq >= 123) {
return 4;
}
}
else if (Cs >= 825) {
if (posBalance < 44) {
if (RsDiff < 2) {
if (diffRsCsdiff < -2) {
if (Rs < 1982) {
return 4;
}
else if (Rs >= 1982) {
if (Rs < 2064) {
return 1;
}
else if (Rs >= 2064) {
return 8;
}
}
}
else if (diffRsCsdiff >= -2) {
if (Rs < 2149) {
return 1;
}
else if (Rs >= 2149) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 92) {
if (Cs < 1166) {
return 4;
}
else if (Cs >= 1166) {
return 2;
}
}
else if (mu_mv_mag_sq >= 92) {
if (SC < 3028) {
if (Rs < 1809) {
return 8;
}
else if (Rs >= 1809) {
return 1;
}
}
else if (SC >= 3028) {
return 2;
}
}
}
}
else if (posBalance >= 44) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 1927) {
if (RsDiff < 1) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 1850) {
return 4;
}
else if (Rs >= 1850) {
return 1;
}
}
}
else if (RsDiff >= 1) {
if (TSC < 3) {
if (SC < 2942) {
if (Cs < 993) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (Cs >= 993) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
else if (SC >= 2942) {
if (Cs < 1067) {
return 4;
}
else if (Cs >= 1067) {
return 1;
}
}
}
else if (TSC >= 3) {
if (Rs < 1872) {
return 1;
}
else if (Rs >= 1872) {
return 8;
}
}
}
}
else if (Rs >= 1927) {
if (Cs < 1142) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Cs < 934) {
if (Cs < 787) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (Cs >= 787) {
return 1;
}
}
else if (Cs >= 934) {
return 4;
}
}
}
else if (Cs >= 1142) {
if (Rs < 2185) {
return 8;
}
else if (Rs >= 2185) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 83) {
if (MVDiff < 2) {
if (Rs < 2220) {
return 4;
}
else if (Rs >= 2220) {
return 8;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (TSC >= 4) {
if (Cs < 934) {
if (ssDCval < 71) {
if (refDCval < 61) {
return 1;
}
else if (refDCval >= 61) {
if (RsDiff < 5) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (RsDiff >= 5) {
if (diffRsCsdiff < 64) {
return 2;
}
else if (diffRsCsdiff >= 64) {
return 4;
}
}
}
}
else if (ssDCval >= 71) {
if (SC < 3063) {
if (mu_mv_mag_sq < 95) {
if (MVDiff < 30) {
if (diffRsCsdiff < -45) {
return 2;
}
else if (diffRsCsdiff >= -45) {
if (mu_mv_mag_sq < 58) {
return 1;
}
else if (mu_mv_mag_sq >= 58) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
else if (MVDiff >= 30) {
if (Rs < 1891) {
return 2;
}
else if (Rs >= 1891) {
if (Rs < 2128) {
return 8;
}
else if (Rs >= 2128) {
if (Rs < 2321) {
return 1;
}
else if (Rs >= 2321) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (SC >= 3063) {
if (RsCsDiff < 43) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 1;
}
}
else if (RsCsDiff >= 43) {
if (Cs < 672) {
return 4;
}
else if (Cs >= 672) {
return 2;
}
}
}
}
}
else if (Cs >= 934) {
if (posBalance < 50) {
if (mu_mv_mag_sq < 79) {
if (ssDCval < 69) {
if (ssDCval < 60) {
if (Rs < 2322) {
return 1;
}
else if (Rs >= 2322) {
return 8;
}
}
else if (ssDCval >= 60) {
return 4;
}
}
else if (ssDCval >= 69) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (posBalance < 31) {
if (Rs < 2416) {
return 4;
}
else if (Rs >= 2416) {
return 4;
}
}
else if (posBalance >= 31) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (AFD < 13) {
if (Rs < 2190) {
return 8;
}
else if (Rs >= 2190) {
if (Rs < 2269) {
return 4;
}
else if (Rs >= 2269) {
return 8;
}
}
}
else if (AFD >= 13) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (Rs < 402) {
if (ssDCval < 117) {
if (MVDiff < 8) {
if (ssDCval < 105) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Cs < 802) {
return 1;
}
else if (Cs >= 802) {
return 4;
}
}
}
else if (ssDCval >= 105) {
if (mu_mv_mag_sq < 122) {
if (Cs < 946) {
return 4;
}
else if (Cs >= 946) {
return 8;
}
}
else if (mu_mv_mag_sq >= 122) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (negBalance < 54) {
return 2;
}
else if (negBalance >= 54) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (mu_mv_mag_sq < 105) {
if (Cs < 534) {
return 4;
}
else if (Cs >= 534) {
if (posBalance < 36) {
return 4;
}
else if (posBalance >= 36) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 323) {
return 2;
}
else if (Rs >= 323) {
return 8;
}
}
}
}
else if (Rs >= 402) {
if (mu_mv_mag_sq < 100) {
if (Cs < 681) {
if (SC < 1298) {
if (negBalance < 23) {
if (ssDCval < 136) {
return 1;
}
else if (ssDCval >= 136) {
return 4;
}
}
else if (negBalance >= 23) {
if (posBalance < 35) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (posBalance >= 35) {
if (Rs < 513) {
if (ssDCval < 152) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 82) {
if (mcTcor < 98) {
if (Cs < 552) {
return 2;
}
else if (Cs >= 552) {
return 8;
}
}
else if (mcTcor >= 98) {
if (Cs < 554) {
return 8;
}
else if (Cs >= 554) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 82) {
return 8;
}
}
}
else if (ssDCval >= 152) {
if (Rs < 475) {
return 4;
}
else if (Rs >= 475) {
return 1;
}
}
}
else if (Rs >= 513) {
if (MVDiff < 2) {
if (Rs < 593) {
if (Rs < 588) {
return 2;
}
else if (Rs >= 588) {
return 1;
}
}
else if (Rs >= 593) {
return 8;
}
}
else if (MVDiff >= 2) {
if (diffMVdiffVal < 5) {
if (negBalance < 24) {
if (Rs < 616) {
return 1;
}
else if (Rs >= 616) {
return 2;
}
}
else if (negBalance >= 24) {
return 2;
}
}
else if (diffMVdiffVal >= 5) {
if (Rs < 577) {
return 2;
}
else if (Rs >= 577) {
if (Rs < 663) {
return 8;
}
else if (Rs >= 663) {
return 4;
}
}
}
}
}
}
}
}
else if (SC >= 1298) {
if (Rs < 1789) {
if (AFD < 18) {
if (Rs < 879) {
if (CsDiff < 3) {
if (refDCval < 130) {
if (ssDCval < 93) {
return 8;
}
else if (ssDCval >= 93) {
return 1;
}
}
else if (refDCval >= 130) {
if (Rs < 665) {
return 8;
}
else if (Rs >= 665) {
return 4;
}
}
}
else if (CsDiff >= 3) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (Rs >= 879) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
if (Cs < 570) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (Cs >= 570) {
if (mu_mv_mag_sq < 3) {
if (Cs < 614) {
return 8;
}
else if (Cs >= 614) {
if (Cs < 667) {
return 1;
}
else if (Cs >= 667) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 1670) {
if (Rs < 1018) {
return 8;
}
else if (Rs >= 1018) {
return 1;
}
}
else if (SC >= 1670) {
return 8;
}
}
}
}
}
}
else if (AFD >= 18) {
if (MVDiff < 86) {
if (refDCval < 99) {
return 4;
}
else if (refDCval >= 99) {
return 1;
}
}
else if (MVDiff >= 86) {
return 2;
}
}
}
else if (Rs >= 1789) {
if (RsDiff < 7) {
if (gchDC < 1) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
return 4;
}
}
else if (gchDC >= 1) {
if (diffRsCsdiff < -10) {
if (AFD < 7) {
if (Cs < 547) {
return 2;
}
else if (Cs >= 547) {
return 1;
}
}
else if (AFD >= 7) {
return 2;
}
}
else if (diffRsCsdiff >= -10) {
if (CsDiff < 2) {
if (Rs < 2065) {
return 1;
}
else if (Rs >= 2065) {
return 2;
}
}
else if (CsDiff >= 2) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
return 8;
}
}
}
}
}
else if (RsDiff >= 7) {
return 1;
}
}
}
}
else if (Cs >= 681) {
if (mu_mv_mag_sq < 21) {
if (refDCval < 123) {
if (TSC < 7) {
if (ssDCval < 104) {
if (TSC < 5) {
if (SC < 1936) {
if (Cs < 691) {
return 1;
}
else if (Cs >= 691) {
return 2;
}
}
else if (SC >= 1936) {
return 1;
}
}
else if (TSC >= 5) {
if (posBalance < 55) {
if (MVDiff < 19) {
if (ssDCval < 101) {
return 4;
}
else if (ssDCval >= 101) {
if (Rs < 806) {
return 1;
}
else if (Rs >= 806) {
return 4;
}
}
}
else if (MVDiff >= 19) {
return 1;
}
}
else if (posBalance >= 55) {
if (Cs < 987) {
if (Cs < 880) {
return 1;
}
else if (Cs >= 880) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 4;
}
}
}
else if (Cs >= 987) {
if (Rs < 1099) {
return 1;
}
else if (Rs >= 1099) {
return 2;
}
}
}
}
}
else if (ssDCval >= 104) {
if (posBalance < 60) {
if (Rs < 1218) {
if (Rs < 899) {
if (Rs < 649) {
return 8;
}
else if (Rs >= 649) {
return 4;
}
}
else if (Rs >= 899) {
return 1;
}
}
else if (Rs >= 1218) {
if (Cs < 690) {
return 1;
}
else if (Cs >= 690) {
return 8;
}
}
}
else if (posBalance >= 60) {
if (Rs < 1573) {
return 4;
}
else if (Rs >= 1573) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (tcor < 92) {
if (posBalance < 59) {
if (posBalance < 56) {
if (Rs < 623) {
return 1;
}
else if (Rs >= 623) {
return 2;
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (posBalance >= 59) {
if (tcor < 91) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (tcor >= 91) {
return 1;
}
}
}
else if (tcor >= 92) {
if (ssDCval < 93) {
return 1;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
}
else if (refDCval >= 123) {
if (Rs < 658) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 650) {
return 2;
}
else if (Rs >= 650) {
return 1;
}
}
}
else if (Rs >= 658) {
if (tcor < 80) {
if (Rs < 2091) {
return 8;
}
else if (Rs >= 2091) {
return 1;
}
}
else if (tcor >= 80) {
if (Rs < 1949) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (Cs < 785) {
if (mcTcor < 98) {
if (Cs < 751) {
return 2;
}
else if (Cs >= 751) {
return 1;
}
}
else if (mcTcor >= 98) {
return 8;
}
}
else if (Cs >= 785) {
if (ssDCval < 124) {
if (Rs < 1518) {
return 8;
}
else if (Rs >= 1518) {
return 2;
}
}
else if (ssDCval >= 124) {
if (mu_mv_mag_sq < 19) {
if (SC < 1706) {
if (Rs < 780) {
return 2;
}
else if (Rs >= 780) {
return 1;
}
}
else if (SC >= 1706) {
return 2;
}
}
else if (mu_mv_mag_sq >= 19) {
if (Rs < 983) {
return 2;
}
else if (Rs >= 983) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1949) {
if (Cs < 989) {
if (Rs < 2152) {
return 8;
}
else if (Rs >= 2152) {
return 1;
}
}
else if (Cs >= 989) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (negBalance < 35) {
if (ssDCval < 113) {
if (SC < 1490) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 2;
}
}
}
else if (SC >= 1490) {
if (mu_mv_mag_sq < 87) {
if (Rs < 1622) {
return 4;
}
else if (Rs >= 1622) {
return 1;
}
}
else if (mu_mv_mag_sq >= 87) {
if (AFD < 5) {
if (ssDCval < 99) {
if (Rs < 1732) {
return 8;
}
else if (Rs >= 1732) {
return 1;
}
}
else if (ssDCval >= 99) {
return 1;
}
}
else if (AFD >= 5) {
if (Rs < 625) {
return 1;
}
else if (Rs >= 625) {
return 8;
}
}
}
}
}
else if (ssDCval >= 113) {
if (negBalance < 29) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
if (Rs < 2045) {
return 2;
}
else if (Rs >= 2045) {
return 8;
}
}
}
else if (negBalance >= 29) {
if (posBalance < 45) {
if (Rs < 2033) {
return 2;
}
else if (Rs >= 2033) {
return 8;
}
}
else if (posBalance >= 45) {
if (diffRsCsdiff < 11) {
return 8;
}
else if (diffRsCsdiff >= 11) {
return 1;
}
}
}
}
}
else if (negBalance >= 35) {
if (MVDiff < 4) {
if (SC < 1911) {
if (negBalance < 41) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 647) {
return 8;
}
else if (Rs >= 647) {
return 4;
}
}
}
else if (negBalance >= 41) {
return 4;
}
}
else if (SC >= 1911) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
if (Cs < 1288) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 2005) {
return 1;
}
else if (Rs >= 2005) {
return 8;
}
}
}
else if (Cs >= 1288) {
return 4;
}
}
}
}
else if (MVDiff >= 4) {
if (tcor < 83) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (tcor >= 83) {
if (ssDCval < 154) {
if (diffRsCsdiff < -19) {
if (posBalance < 58) {
if (MVDiff < 17) {
if (SC < 2963) {
return 8;
}
else if (SC >= 2963) {
return 1;
}
}
else if (MVDiff >= 17) {
return 8;
}
}
else if (posBalance >= 58) {
if (SC < 3103) {
return 2;
}
else if (SC >= 3103) {
return 8;
}
}
}
else if (diffRsCsdiff >= -19) {
if (Cs < 747) {
if (negBalance < 39) {
if (Rs < 1023) {
return 2;
}
else if (Rs >= 1023) {
return 8;
}
}
else if (negBalance >= 39) {
if (ssDCval < 147) {
return 8;
}
else if (ssDCval >= 147) {
return 2;
}
}
}
else if (Cs >= 747) {
if (posBalance < 67) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (ssDCval < 129) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 2021) {
return 1;
}
else if (Rs >= 2021) {
return 8;
}
}
}
else if (ssDCval >= 129) {
return 1;
}
}
}
else if (posBalance >= 67) {
if (Rs < 1263) {
return 4;
}
else if (Rs >= 1263) {
return 8;
}
}
}
}
}
else if (ssDCval >= 154) {
if (Rs < 1662) {
return 4;
}
else if (Rs >= 1662) {
return 1;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (ssDCval < 133) {
if (refDCval < 97) {
if (mu_mv_mag_sq < 126) {
if (Cs < 848) {
if (Rs < 946) {
if (negBalance < 18) {
return 1;
}
else if (negBalance >= 18) {
return 8;
}
}
else if (Rs >= 946) {
if (SC < 2658) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (SC >= 2658) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
if (Rs < 2029) {
return 8;
}
else if (Rs >= 2029) {
return 1;
}
}
}
}
}
}
else if (Cs >= 848) {
if (negBalance < 36) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 117) {
return 8;
}
else if (mu_mv_mag_sq >= 117) {
return 2;
}
}
else if (mcTcor >= 98) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 4;
}
}
}
else if (negBalance >= 36) {
if (Rs < 1023) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
else if (Rs >= 1023) {
if (mu_mv_mag_sq < 123) {
return 8;
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1677) {
return 4;
}
else if (Rs >= 1677) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 2165) {
if (posBalance < 24) {
return 4;
}
else if (posBalance >= 24) {
if (Rs < 2145) {
return 8;
}
else if (Rs >= 2145) {
if (Rs < 2148) {
return 4;
}
else if (Rs >= 2148) {
return 8;
}
}
}
}
else if (Rs >= 2165) {
return 4;
}
}
}
else if (refDCval >= 97) {
if (RsDiff < 1) {
if (Cs < 1186) {
if (refDCval < 122) {
if (posBalance < 24) {
if (refDCval < 98) {
return 1;
}
else if (refDCval >= 98) {
return 8;
}
}
else if (posBalance >= 24) {
return 8;
}
}
else if (refDCval >= 122) {
if (ssDCval < 125) {
return 4;
}
else if (ssDCval >= 125) {
if (Rs < 757) {
return 8;
}
else if (Rs >= 757) {
return 1;
}
}
}
}
else if (Cs >= 1186) {
return 1;
}
}
else if (RsDiff >= 1) {
if (SC < 1205) {
if (CsDiff < 2) {
if (ssDCval < 121) {
if (Rs < 582) {
return 1;
}
else if (Rs >= 582) {
return 2;
}
}
else if (ssDCval >= 121) {
return 8;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (SC >= 1205) {
if (Rs < 1671) {
if (mu_mv_mag_sq < 104) {
if (Rs < 524) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (Rs >= 524) {
if (Cs < 1212) {
return 8;
}
else if (Cs >= 1212) {
if (Rs < 976) {
return 1;
}
else if (Rs >= 976) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (SC < 1360) {
if (Cs < 771) {
if (ssDCval < 112) {
return 1;
}
else if (ssDCval >= 112) {
return 8;
}
}
else if (Cs >= 771) {
return 8;
}
}
else if (SC >= 1360) {
return 8;
}
}
}
else if (Rs >= 1671) {
if (mu_mv_mag_sq < 113) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (mu_mv_mag_sq >= 113) {
if (SC < 2971) {
if (Cs < 956) {
return 1;
}
else if (Cs >= 956) {
return 2;
}
}
else if (SC >= 2971) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 1147) {
return 2;
}
else if (Rs >= 1147) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2496) {
if (MVDiff < 14) {
if (refDCval < 109) {
if (posBalance < 48) {
if (RsCsDiff < 19) {
if (SC < 4969) {
if (mu_mv_mag_sq < 124) {
if (posBalance < 25) {
if (Rs < 3209) {
return 4;
}
else if (Rs >= 3209) {
return 2;
}
}
else if (posBalance >= 25) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (mu_mv_mag_sq < 83) {
if (Rs < 3368) {
return 8;
}
else if (Rs >= 3368) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (SC < 4483) {
if (Rs < 2611) {
if (RsCsDiff < 12) {
return 8;
}
else if (RsCsDiff >= 12) {
if (Rs < 2578) {
return 4;
}
else if (Rs >= 2578) {
return 4;
}
}
}
else if (Rs >= 2611) {
if (Cs < 1072) {
if (Rs < 2703) {
return 1;
}
else if (Rs >= 2703) {
return 2;
}
}
else if (Cs >= 1072) {
if (Cs < 1243) {
if (Rs < 2816) {
return 8;
}
else if (Rs >= 2816) {
return 4;
}
}
else if (Cs >= 1243) {
return 1;
}
}
}
}
else if (SC >= 4483) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (negBalance < 20) {
if (SC < 3564) {
return 4;
}
else if (SC >= 3564) {
if (diffRsCsdiff < 4) {
return 8;
}
else if (diffRsCsdiff >= 4) {
if (Rs < 3180) {
return 8;
}
else if (Rs >= 3180) {
return 4;
}
}
}
}
else if (negBalance >= 20) {
if (ssDCval < 92) {
if (Cs < 1049) {
return 1;
}
else if (Cs >= 1049) {
return 8;
}
}
else if (ssDCval >= 92) {
if (Rs < 3344) {
return 2;
}
else if (Rs >= 3344) {
return 4;
}
}
}
}
}
else if (SC >= 4969) {
if (Cs < 1280) {
if (Rs < 4595) {
return 4;
}
else if (Rs >= 4595) {
return 2;
}
}
else if (Cs >= 1280) {
if (mu_mv_mag_sq < 127) {
if (Rs < 3920) {
return 4;
}
else if (Rs >= 3920) {
if (negBalance < 22) {
return 8;
}
else if (negBalance >= 22) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 4564) {
return 8;
}
else if (Rs >= 4564) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 19) {
if (ssDCval < 83) {
if (negBalance < 47) {
if (Rs < 2876) {
if (refDCval < 75) {
if (MVDiff < 5) {
if (Rs < 2531) {
if (Cs < 1207) {
if (Cs < 1052) {
return 2;
}
else if (Cs >= 1052) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
else if (Cs >= 1207) {
return 8;
}
}
else if (Rs >= 2531) {
if (mcTcor < 95) {
if (Rs < 2783) {
return 4;
}
else if (Rs >= 2783) {
if (Rs < 2818) {
return 1;
}
else if (Rs >= 2818) {
return 2;
}
}
}
else if (mcTcor >= 95) {
return 2;
}
}
}
else if (MVDiff >= 5) {
if (refDCval < 58) {
return 4;
}
else if (refDCval >= 58) {
if (RsCsDiff < 69) {
if (mu_mv_mag_sq < 58) {
return 8;
}
else if (mu_mv_mag_sq >= 58) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
else if (RsCsDiff >= 69) {
if (tcor < 82) {
return 2;
}
else if (tcor >= 82) {
if (diffMVdiffVal < 2) {
if (AFD < 10) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (AFD >= 10) {
return 4;
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 1103) {
return 1;
}
else if (Cs >= 1103) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (mu_mv_mag_sq < 74) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (mu_mv_mag_sq >= 74) {
return 2;
}
}
}
else if (Rs >= 2876) {
if (Rs < 3819) {
if (ssDCval < 65) {
if (ssDCval < 57) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (ssDCval >= 57) {
if (MVDiff < 7) {
if (negBalance < 39) {
return 2;
}
else if (negBalance >= 39) {
return 4;
}
}
else if (MVDiff >= 7) {
if (diffRsCsdiff < -20) {
return 4;
}
else if (diffRsCsdiff >= -20) {
if (Rs < 3167) {
return 2;
}
else if (Rs >= 3167) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 65) {
if (diffAFD < -2) {
if (mu_mv_mag_sq < 88) {
if (mcTcor < 97) {
if (RsCsDiff < 182) {
return 1;
}
else if (RsCsDiff >= 182) {
return 2;
}
}
else if (mcTcor >= 97) {
if (Rs < 3068) {
return 1;
}
else if (Rs >= 3068) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 88) {
return 2;
}
}
else if (diffAFD >= -2) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (TSC < 7) {
if (mu_mv_mag_sq < 105) {
if (SC < 4143) {
if (MVDiff < 3) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 3053) {
return 4;
}
else if (Rs >= 3053) {
return 2;
}
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (SC >= 4143) {
return 4;
}
}
else if (mu_mv_mag_sq >= 105) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (TSC >= 7) {
if (ssDCval < 69) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (ssDCval >= 69) {
if (Cs < 1209) {
if (diffAFD < 3) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (diffAFD >= 3) {
return 4;
}
}
else if (Cs >= 1209) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 3819) {
if (mu_mv_mag_sq < 104) {
return 4;
}
else if (mu_mv_mag_sq >= 104) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 4;
}
}
}
}
}
else if (negBalance >= 47) {
if (posBalance < 40) {
if (Cs < 1072) {
if (Cs < 973) {
return 2;
}
else if (Cs >= 973) {
return 1;
}
}
else if (Cs >= 1072) {
return 2;
}
}
else if (posBalance >= 40) {
return 2;
}
}
}
else if (ssDCval >= 83) {
if (negBalance < 52) {
if (tcor < 97) {
if (ssDCval < 101) {
if (mu_mv_mag_sq < 78) {
if (mu_mv_mag_sq < 44) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 2;
}
}
else if (mu_mv_mag_sq >= 44) {
if (diffMVdiffVal < -1) {
if (Rs < 3569) {
if (Rs < 3145) {
return 8;
}
else if (Rs >= 3145) {
if (Cs < 949) {
return 8;
}
else if (Cs >= 949) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
else if (Rs >= 3569) {
if (diffMVdiffVal < -2) {
if (posBalance < 42) {
if (Rs < 4036) {
return 4;
}
else if (Rs >= 4036) {
return 4;
}
}
else if (posBalance >= 42) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 4032) {
return 8;
}
else if (Rs >= 4032) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (mu_mv_mag_sq < 51) {
return 4;
}
else if (mu_mv_mag_sq >= 51) {
if (ssDCval < 88) {
if (Rs < 3012) {
return 4;
}
else if (Rs >= 3012) {
return 2;
}
}
else if (ssDCval >= 88) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Cs < 1236) {
if (gchDC < 1) {
if (mu_mv_mag_sq < 62) {
if (negBalance < 29) {
if (Rs < 3985) {
if (diffMVdiffVal < 2) {
if (Rs < 3631) {
return 4;
}
else if (Rs >= 3631) {
return 4;
}
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
else if (Rs >= 3985) {
return 4;
}
}
else if (negBalance >= 29) {
if (Rs < 3855) {
return 4;
}
else if (Rs >= 3855) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 62) {
return 4;
}
}
else if (gchDC >= 1) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 4;
}
}
}
}
else if (Cs >= 1236) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
return 2;
}
}
else if (ssDCval >= 101) {
if (Cs < 964) {
if (SC < 4221) {
return 1;
}
else if (SC >= 4221) {
if (Rs < 3428) {
return 4;
}
else if (Rs >= 3428) {
if (Rs < 3604) {
return 8;
}
else if (Rs >= 3604) {
return 8;
}
}
}
}
else if (Cs >= 964) {
return 8;
}
}
}
else if (tcor >= 97) {
if (Cs < 1119) {
if (Rs < 2922) {
return 8;
}
else if (Rs >= 2922) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
return 2;
}
}
}
else if (Cs >= 1119) {
if (AFD < 6) {
if (Rs < 3175) {
return 8;
}
else if (Rs >= 3175) {
return 2;
}
}
else if (AFD >= 6) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
return 2;
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 90) {
if (CsDiff < 3) {
if (Cs < 674) {
if (SC < 3494) {
if (Cs < 246) {
return 1;
}
else if (Cs >= 246) {
if (AFD < 12) {
if (TSC < 4) {
if (Rs < 2567) {
return 4;
}
else if (Rs >= 2567) {
return 1;
}
}
else if (TSC >= 4) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (AFD >= 12) {
return 2;
}
}
}
else if (SC >= 3494) {
return 4;
}
}
else if (Cs >= 674) {
if (negBalance < 22) {
if (negBalance < 15) {
return 4;
}
else if (negBalance >= 15) {
return 8;
}
}
else if (negBalance >= 22) {
if (refDCval < 63) {
if (Rs < 2670) {
return 2;
}
else if (Rs >= 2670) {
return 4;
}
}
else if (refDCval >= 63) {
return 2;
}
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 64) {
if (SC < 3646) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
if (Cs < 662) {
return 2;
}
else if (Cs >= 662) {
return 4;
}
}
}
else if (SC >= 3646) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 3) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 38) {
return 4;
}
else if (negBalance >= 38) {
if (mu_mv_mag_sq < 92) {
return 2;
}
else if (mu_mv_mag_sq >= 92) {
if (Cs < 996) {
return 2;
}
else if (Cs >= 996) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 64) {
if (TSC < 6) {
if (tcor < 92) {
if (negBalance < 37) {
if (Cs < 1011) {
return 4;
}
else if (Cs >= 1011) {
if (Rs < 3207) {
return 2;
}
else if (Rs >= 3207) {
return 4;
}
}
}
else if (negBalance >= 37) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (SC < 3430) {
return 2;
}
else if (SC >= 3430) {
return 4;
}
}
}
}
else if (tcor >= 92) {
if (negBalance < 19) {
return 8;
}
else if (negBalance >= 19) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (negBalance < 38) {
if (posBalance < 52) {
if (Rs < 3033) {
return 1;
}
else if (Rs >= 3033) {
return 4;
}
}
else if (posBalance >= 52) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (negBalance >= 38) {
return 2;
}
}
}
}
}
else if (TSC >= 6) {
if (tcor < 79) {
if (RsCsDiff < 190) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
return 1;
}
}
else if (RsCsDiff >= 190) {
if (Cs < 860) {
return 2;
}
else if (Cs >= 860) {
if (Cs < 999) {
return 4;
}
else if (Cs >= 999) {
return 2;
}
}
}
}
else if (tcor >= 79) {
if (mcTcor < 99) {
if (mcTcor < 89) {
if (Rs < 3477) {
return 1;
}
else if (Rs >= 3477) {
return 2;
}
}
else if (mcTcor >= 89) {
if (RsDiff < 8) {
if (AFD < 12) {
if (ssDCval < 65) {
if (Rs < 2659) {
return 2;
}
else if (Rs >= 2659) {
return 4;
}
}
else if (ssDCval >= 65) {
return 2;
}
}
else if (AFD >= 12) {
return 1;
}
}
else if (RsDiff >= 8) {
if (negBalance < 37) {
if (posBalance < 56) {
if (Rs < 3215) {
return 4;
}
else if (Rs >= 3215) {
return 2;
}
}
else if (posBalance >= 56) {
return 2;
}
}
else if (negBalance >= 37) {
return 2;
}
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 90) {
if (RsDiff < 11) {
if (tcor < 96) {
if (negBalance < 53) {
if (RsCsDiff < 59) {
if (Cs < 775) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2858) {
return 4;
}
else if (Rs >= 2858) {
return 8;
}
}
}
}
else if (Cs >= 775) {
return 8;
}
}
else if (RsCsDiff >= 59) {
if (Cs < 1201) {
if (Cs < 947) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (Cs >= 947) {
if (gchDC < 1) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 4;
}
}
else if (gchDC >= 1) {
if (Rs < 3361) {
return 4;
}
else if (Rs >= 3361) {
return 2;
}
}
}
}
else if (Cs >= 1201) {
if (Rs < 3643) {
return 8;
}
else if (Rs >= 3643) {
return 4;
}
}
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (tcor >= 96) {
if (posBalance < 57) {
if (Cs < 949) {
if (Cs < 454) {
return 8;
}
else if (Cs >= 454) {
return 2;
}
}
else if (Cs >= 949) {
if (Rs < 3204) {
return 8;
}
else if (Rs >= 3204) {
return 2;
}
}
}
else if (posBalance >= 57) {
return 2;
}
}
}
else if (RsDiff >= 11) {
if (negBalance < 49) {
if (AFD < 10) {
if (Rs < 3387) {
return 2;
}
else if (Rs >= 3387) {
return 8;
}
}
else if (AFD >= 10) {
return 2;
}
}
else if (negBalance >= 49) {
if (Cs < 838) {
return 2;
}
else if (Cs >= 838) {
if (tcor < 80) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (tcor < 72) {
if (Rs < 3250) {
return 2;
}
else if (Rs >= 3250) {
return 8;
}
}
else if (tcor >= 72) {
return 4;
}
}
}
else if (tcor >= 80) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (RsCsDiff < 256) {
return 4;
}
else if (RsCsDiff >= 256) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 3365) {
if (mcTcor < 97) {
if (AFD < 14) {
if (tcor < 92) {
if (Cs < 520) {
return 1;
}
else if (Cs >= 520) {
return 8;
}
}
else if (tcor >= 92) {
if (MVDiff < 7) {
if (SC < 3970) {
return 8;
}
else if (SC >= 3970) {
return 2;
}
}
else if (MVDiff >= 7) {
if (Cs < 1255) {
return 1;
}
else if (Cs >= 1255) {
if (Rs < 2544) {
return 1;
}
else if (Rs >= 2544) {
return 8;
}
}
}
}
}
else if (AFD >= 14) {
if (SC < 3536) {
return 2;
}
else if (SC >= 3536) {
if (RsDiff < 16) {
if (TSC < 11) {
if (Cs < 1302) {
return 1;
}
else if (Cs >= 1302) {
return 2;
}
}
else if (TSC >= 11) {
return 4;
}
}
else if (RsDiff >= 16) {
return 2;
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 7) {
if (Cs < 536) {
if (RsDiff < 5) {
if (Rs < 2869) {
return 1;
}
else if (Rs >= 2869) {
return 8;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (Cs >= 536) {
if (Rs < 2584) {
if (Rs < 2565) {
return 2;
}
else if (Rs >= 2565) {
return 1;
}
}
else if (Rs >= 2584) {
return 2;
}
}
}
else if (MVDiff >= 7) {
if (TSC < 6) {
if (mu_mv_mag_sq < 87) {
if (Cs < 907) {
if (AFD < 12) {
if (SC < 3601) {
return 1;
}
else if (SC >= 3601) {
return 2;
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (Cs >= 907) {
return 8;
}
}
else if (mu_mv_mag_sq >= 87) {
if (refDCval < 125) {
return 2;
}
else if (refDCval >= 125) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (diffMVdiffVal < -6) {
return 1;
}
else if (diffMVdiffVal >= -6) {
if (refDCval < 136) {
if (refDCval < 113) {
if (Rs < 2781) {
return 2;
}
else if (Rs >= 2781) {
return 4;
}
}
else if (refDCval >= 113) {
if (mcTcor < 101) {
return 2;
}
else if (mcTcor >= 101) {
return 1;
}
}
}
else if (refDCval >= 136) {
if (Cs < 925) {
return 1;
}
else if (Cs >= 925) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 3365) {
if (MVDiff < 7) {
if (ssDCval < 142) {
if (diffMVdiffVal < -3) {
if (SC < 4336) {
if (Rs < 3630) {
return 2;
}
else if (Rs >= 3630) {
return 8;
}
}
else if (SC >= 4336) {
if (ssDCval < 133) {
return 2;
}
else if (ssDCval >= 133) {
if (mcTcor < 98) {
if (Rs < 4644) {
if (Rs < 3928) {
return 2;
}
else if (Rs >= 3928) {
return 1;
}
}
else if (Rs >= 4644) {
return 2;
}
}
else if (mcTcor >= 98) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -3) {
if (tcor < 92) {
if (RsDiff < 9) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (tcor >= 92) {
return 2;
}
}
}
else if (ssDCval >= 142) {
if (Rs < 3836) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 107) {
return 2;
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 3687) {
return 2;
}
else if (Rs >= 3687) {
return 1;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 3652) {
return 2;
}
else if (Rs >= 3652) {
return 1;
}
}
}
else if (Rs >= 3836) {
if (Rs < 6340) {
if (diffTSC < -2) {
if (Rs < 4037) {
return 1;
}
else if (Rs >= 4037) {
return 2;
}
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (Rs >= 6340) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (negBalance < 54) {
if (TSC < 5) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (TSC >= 5) {
if (refDCval < 132) {
if (Cs < 1218) {
return 2;
}
else if (Cs >= 1218) {
if (Rs < 4785) {
return 8;
}
else if (Rs >= 4785) {
return 2;
}
}
}
else if (refDCval >= 132) {
if (mcTcor < 97) {
if (RsCsDiff < 418) {
if (SC < 4950) {
if (Cs < 879) {
return 2;
}
else if (Cs >= 879) {
if (Cs < 1178) {
return 1;
}
else if (Cs >= 1178) {
return 8;
}
}
}
else if (SC >= 4950) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 1;
}
}
}
else if (RsCsDiff >= 418) {
return 2;
}
}
else if (mcTcor >= 97) {
if (Cs < 1121) {
return 2;
}
else if (Cs >= 1121) {
return 1;
}
}
}
}
}
else if (negBalance >= 54) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (RsDiff < 19) {
return 4;
}
else if (RsDiff >= 19) {
if (Cs < 1212) {
return 2;
}
else if (Cs >= 1212) {
return 4;
}
}
}
}
}
}
}
}
else if (MVDiff >= 14) {
if (mcTcor < 68) {
if (MVDiff < 43) {
return 1;
}
else if (MVDiff >= 43) {
if (gchDC < 5) {
if (AFD < 33) {
return 4;
}
else if (AFD >= 33) {
return 2;
}
}
else if (gchDC >= 5) {
return 8;
}
}
}
else if (mcTcor >= 68) {
if (refDCval < 132) {
if (refDCval < 78) {
if (TSC < 8) {
if (posBalance < 36) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 87) {
if (diffRsCsdiff < 2) {
return 4;
}
else if (diffRsCsdiff >= 2) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Rs < 3134) {
return 8;
}
else if (Rs >= 3134) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
return 8;
}
}
else if (negBalance >= 37) {
return 1;
}
}
else if (posBalance >= 36) {
if (Rs < 2790) {
if (TSC < 5) {
if (Cs < 518) {
return 8;
}
else if (Cs >= 518) {
return 4;
}
}
else if (TSC >= 5) {
if (refDCval < 57) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
else if (refDCval >= 57) {
if (diffRsCsdiff < 57) {
if (mcTcor < 90) {
if (Cs < 1021) {
return 4;
}
else if (Cs >= 1021) {
return 2;
}
}
else if (mcTcor >= 90) {
return 2;
}
}
else if (diffRsCsdiff >= 57) {
return 4;
}
}
}
}
else if (Rs >= 2790) {
if (Cs < 737) {
if (CsDiff < 6) {
if (Rs < 2880) {
if (Rs < 2838) {
return 4;
}
else if (Rs >= 2838) {
return 2;
}
}
else if (Rs >= 2880) {
return 4;
}
}
else if (CsDiff >= 6) {
return 1;
}
}
else if (Cs >= 737) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
if (ssDCval < 74) {
if (AFD < 13) {
if (diffRsCsdiff < 10) {
return 2;
}
else if (diffRsCsdiff >= 10) {
return 4;
}
}
else if (AFD >= 13) {
return 4;
}
}
else if (ssDCval >= 74) {
return 2;
}
}
}
}
}
}
else if (TSC >= 8) {
if (Rs < 3808) {
if (TSC < 12) {
if (AFD < 12) {
if (diffRsCsdiff < 9) {
if (Cs < 1242) {
return 2;
}
else if (Cs >= 1242) {
return 8;
}
}
else if (diffRsCsdiff >= 9) {
if (Rs < 3252) {
return 1;
}
else if (Rs >= 3252) {
return 4;
}
}
}
else if (AFD >= 12) {
if (Cs < 125) {
return 1;
}
else if (Cs >= 125) {
if (RsCsDiff < 133) {
if (ssDCval < 61) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (ssDCval >= 61) {
return 2;
}
}
else if (RsCsDiff >= 133) {
if (posBalance < 47) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
if (MVDiff < 29) {
return 4;
}
else if (MVDiff >= 29) {
if (Rs < 2968) {
return 1;
}
else if (Rs >= 2968) {
return 2;
}
}
}
}
else if (posBalance >= 47) {
if (diffRsCsdiff < 87) {
if (CsDiff < 9) {
return 2;
}
else if (CsDiff >= 9) {
return 1;
}
}
else if (diffRsCsdiff >= 87) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
if (Cs < 332) {
return 4;
}
else if (Cs >= 332) {
return 2;
}
}
}
}
}
}
}
}
}
else if (TSC >= 12) {
if (RsDiff < 21) {
return 1;
}
else if (RsDiff >= 21) {
if (Rs < 2854) {
return 8;
}
else if (Rs >= 2854) {
return 2;
}
}
}
}
else if (Rs >= 3808) {
if (Rs < 5986) {
return 4;
}
else if (Rs >= 5986) {
if (Rs < 6850) {
return 8;
}
else if (Rs >= 6850) {
return 2;
}
}
}
}
}
else if (refDCval >= 78) {
if (RsCsDiff < 248) {
if (SC < 3544) {
if (ssDCval < 112) {
if (MVDiff < 38) {
if (RsDiff < 11) {
if (diffRsCsdiff < -33) {
if (diffRsCsdiff < -51) {
return 2;
}
else if (diffRsCsdiff >= -51) {
if (Rs < 2817) {
if (Rs < 2583) {
return 1;
}
else if (Rs >= 2583) {
return 8;
}
}
else if (Rs >= 2817) {
return 4;
}
}
}
else if (diffRsCsdiff >= -33) {
if (RsCsDiff < 28) {
if (Rs < 3080) {
return 8;
}
else if (Rs >= 3080) {
return 1;
}
}
else if (RsCsDiff >= 28) {
return 1;
}
}
}
else if (RsDiff >= 11) {
if (MVDiff < 25) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
else if (MVDiff >= 25) {
if (Rs < 2604) {
if (Rs < 2536) {
return 8;
}
else if (Rs >= 2536) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
return 2;
}
}
}
else if (Rs >= 2604) {
if (Rs < 2958) {
return 2;
}
else if (Rs >= 2958) {
return 8;
}
}
}
}
}
else if (MVDiff >= 38) {
if (Cs < 305) {
if (Cs < 200) {
if (Cs < 115) {
return 1;
}
else if (Cs >= 115) {
return 2;
}
}
else if (Cs >= 200) {
return 4;
}
}
else if (Cs >= 305) {
if (Rs < 2654) {
if (mu_mv_mag_sq < 4) {
if (SC < 2985) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 8;
}
}
else if (SC >= 2985) {
if (Rs < 2639) {
return 1;
}
else if (Rs >= 2639) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 4) {
if (MVDiff < 50) {
return 8;
}
else if (MVDiff >= 50) {
return 2;
}
}
}
else if (Rs >= 2654) {
if (gchDC < 10) {
return 8;
}
else if (gchDC >= 10) {
if (Rs < 2805) {
return 2;
}
else if (Rs >= 2805) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (diffAFD < -4) {
if (mu_mv_mag_sq < 32) {
return 1;
}
else if (mu_mv_mag_sq >= 32) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
else if (diffAFD >= -4) {
if (tcor < 87) {
if (gchDC < 3) {
if (SC < 3517) {
return 8;
}
else if (SC >= 3517) {
if (Rs < 3048) {
return 1;
}
else if (Rs >= 3048) {
return 2;
}
}
}
else if (gchDC >= 3) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
return 1;
}
}
}
else if (tcor >= 87) {
if (CsDiff < 3) {
if (diffRsCsdiff < 6) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
if (diffRsCsdiff < -46) {
return 1;
}
else if (diffRsCsdiff >= -46) {
return 8;
}
}
}
else if (diffRsCsdiff >= 6) {
return 1;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
}
}
}
else if (SC >= 3544) {
if (Rs < 5211) {
if (ssDCval < 88) {
if (Rs < 3853) {
if (Rs < 2752) {
if (Cs < 1116) {
if (Rs < 2663) {
if (Cs < 1048) {
return 2;
}
else if (Cs >= 1048) {
return 1;
}
}
else if (Rs >= 2663) {
return 1;
}
}
else if (Cs >= 1116) {
return 8;
}
}
else if (Rs >= 2752) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
return 2;
}
}
}
else if (Rs >= 3853) {
return 4;
}
}
else if (ssDCval >= 88) {
if (RsCsDiff < 128) {
if (Rs < 4205) {
if (ssDCval < 105) {
if (AFD < 14) {
if (diffRsCsdiff < -28) {
if (Cs < 776) {
if (Rs < 3251) {
return 1;
}
else if (Rs >= 3251) {
if (Rs < 3684) {
return 8;
}
else if (Rs >= 3684) {
return 2;
}
}
}
else if (Cs >= 776) {
return 2;
}
}
else if (diffRsCsdiff >= -28) {
if (posBalance < 37) {
if (Rs < 3498) {
return 4;
}
else if (Rs >= 3498) {
return 4;
}
}
else if (posBalance >= 37) {
if (tcor < 91) {
if (Rs < 3232) {
return 1;
}
else if (Rs >= 3232) {
return 8;
}
}
else if (tcor >= 91) {
return 8;
}
}
}
}
else if (AFD >= 14) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 4;
}
}
}
else if (ssDCval >= 105) {
if (diffAFD < -22) {
return 1;
}
else if (diffAFD >= -22) {
if (Rs < 3550) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
if (AFD < 18) {
if (mu_mv_mag_sq < 71) {
return 8;
}
else if (mu_mv_mag_sq >= 71) {
if (Rs < 2752) {
return 8;
}
else if (Rs >= 2752) {
return 2;
}
}
}
else if (AFD >= 18) {
return 2;
}
}
}
else if (Rs >= 3550) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
if (ssDCval < 112) {
if (Cs < 876) {
return 8;
}
else if (Cs >= 876) {
return 1;
}
}
else if (ssDCval >= 112) {
return 8;
}
}
}
}
}
}
else if (Rs >= 4205) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (RsCsDiff >= 128) {
if (MVDiff < 34) {
if (Cs < 1191) {
if (posBalance < 62) {
if (diffRsCsdiff < -58) {
if (Rs < 2854) {
if (Rs < 2645) {
return 8;
}
else if (Rs >= 2645) {
return 4;
}
}
else if (Rs >= 2854) {
if (ssDCval < 126) {
if (Cs < 1164) {
return 2;
}
else if (Cs >= 1164) {
return 4;
}
}
else if (ssDCval >= 126) {
return 1;
}
}
}
else if (diffRsCsdiff >= -58) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (tcor < 82) {
if (negBalance < 59) {
return 8;
}
else if (negBalance >= 59) {
if (SC < 3726) {
return 1;
}
else if (SC >= 3726) {
return 2;
}
}
}
else if (tcor >= 82) {
if (MVDiff < 22) {
if (tcor < 95) {
return 2;
}
else if (tcor >= 95) {
return 8;
}
}
else if (MVDiff >= 22) {
if (diffMVdiffVal < 8) {
if (Rs < 3308) {
return 1;
}
else if (Rs >= 3308) {
return 8;
}
}
else if (diffMVdiffVal >= 8) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 62) {
if (refDCval < 95) {
return 2;
}
else if (refDCval >= 95) {
if (diffMVdiffVal < 6) {
if (negBalance < 38) {
return 2;
}
else if (negBalance >= 38) {
if (Rs < 3536) {
return 1;
}
else if (Rs >= 3536) {
return 2;
}
}
}
else if (diffMVdiffVal >= 6) {
return 1;
}
}
}
}
else if (Cs >= 1191) {
if (gchDC < 2) {
if (Rs < 2699) {
return 4;
}
else if (Rs >= 2699) {
return 8;
}
}
else if (gchDC >= 2) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
else if (MVDiff >= 34) {
if (diffRsCsdiff < -52) {
if (diffMVdiffVal < 13) {
if (diffRsCsdiff < -104) {
return 2;
}
else if (diffRsCsdiff >= -104) {
if (Cs < 888) {
if (Cs < 515) {
if (Rs < 3516) {
return 4;
}
else if (Rs >= 3516) {
return 2;
}
}
else if (Cs >= 515) {
return 1;
}
}
else if (Cs >= 888) {
return 8;
}
}
}
else if (diffMVdiffVal >= 13) {
if (Rs < 3155) {
return 1;
}
else if (Rs >= 3155) {
return 4;
}
}
}
else if (diffRsCsdiff >= -52) {
if (AFD < 26) {
if (Cs < 487) {
if (diffMVdiffVal < 0) {
return 4;
}
else if (diffMVdiffVal >= 0) {
if (Rs < 3770) {
return 8;
}
else if (Rs >= 3770) {
return 2;
}
}
}
else if (Cs >= 487) {
if (diffMVdiffVal < 29) {
if (negBalance < 62) {
if (diffMVdiffVal < -14) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 8;
}
}
else if (diffMVdiffVal >= -14) {
return 8;
}
}
else if (negBalance >= 62) {
if (Rs < 2997) {
return 2;
}
else if (Rs >= 2997) {
if (Rs < 3326) {
return 8;
}
else if (Rs >= 3326) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 29) {
return 2;
}
}
}
else if (AFD >= 26) {
if (mcTcor < 83) {
if (Rs < 3087) {
return 2;
}
else if (Rs >= 3087) {
return 4;
}
}
else if (mcTcor >= 83) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 5211) {
return 2;
}
}
}
else if (RsCsDiff >= 248) {
if (mcTcor < 91) {
if (RsCsDiff < 368) {
if (TSC < 10) {
if (tcor < 46) {
if (Rs < 3325) {
return 8;
}
else if (Rs >= 3325) {
return 2;
}
}
else if (tcor >= 46) {
return 2;
}
}
else if (TSC >= 10) {
if (diffRsCsdiff < -36) {
if (gchDC < 2) {
if (SC < 4388) {
return 2;
}
else if (SC >= 4388) {
return 4;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (diffRsCsdiff >= -36) {
if (tcor < 83) {
if (mcTcor < 70) {
return 1;
}
else if (mcTcor >= 70) {
if (ssDCval < 103) {
return 4;
}
else if (ssDCval >= 103) {
if (diffTSC < 3) {
if (RsCsDiff < 277) {
if (Rs < 3233) {
return 4;
}
else if (Rs >= 3233) {
return 8;
}
}
else if (RsCsDiff >= 277) {
return 4;
}
}
else if (diffTSC >= 3) {
return 8;
}
}
}
}
else if (tcor >= 83) {
return 2;
}
}
}
}
else if (RsCsDiff >= 368) {
if (Cs < 1139) {
if (Rs < 2712) {
if (AFD < 32) {
return 2;
}
else if (AFD >= 32) {
return 1;
}
}
else if (Rs >= 2712) {
if (RsDiff < 21) {
if (SC < 4213) {
return 2;
}
else if (SC >= 4213) {
if (Rs < 3335) {
return 8;
}
else if (Rs >= 3335) {
return 2;
}
}
}
else if (RsDiff >= 21) {
return 2;
}
}
}
else if (Cs >= 1139) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
if (MVDiff < 57) {
return 2;
}
else if (MVDiff >= 57) {
if (Cs < 1258) {
return 4;
}
else if (Cs >= 1258) {
if (Rs < 4700) {
return 2;
}
else if (Rs >= 4700) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (refDCval < 108) {
if (AFD < 28) {
if (Rs < 3791) {
if (diffAFD < -8) {
if (Rs < 3071) {
return 1;
}
else if (Rs >= 3071) {
return 2;
}
}
else if (diffAFD >= -8) {
return 2;
}
}
else if (Rs >= 3791) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (SC < 4788) {
if (CsDiff < 4) {
if (Rs < 4237) {
return 8;
}
else if (Rs >= 4237) {
return 4;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (SC >= 4788) {
if (tcor < 86) {
return 2;
}
else if (tcor >= 86) {
if (negBalance < 46) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
else if (negBalance >= 46) {
if (Rs < 4944) {
return 1;
}
else if (Rs >= 4944) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 28) {
if (Cs < 542) {
if (RsDiff < 20) {
return 8;
}
else if (RsDiff >= 20) {
return 2;
}
}
else if (Cs >= 542) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
}
else if (refDCval >= 108) {
if (Rs < 3483) {
if (refDCval < 109) {
if (Cs < 1049) {
if (Cs < 964) {
return 2;
}
else if (Cs >= 964) {
return 4;
}
}
else if (Cs >= 1049) {
return 8;
}
}
else if (refDCval >= 109) {
if (posBalance < 63) {
if (Cs < 630) {
if (refDCval < 125) {
return 2;
}
else if (refDCval >= 125) {
return 1;
}
}
else if (Cs >= 630) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (RsCsDiff < 285) {
return 2;
}
else if (RsCsDiff >= 285) {
if (mu_mv_mag_sq < 37) {
return 1;
}
else if (mu_mv_mag_sq >= 37) {
return 2;
}
}
}
}
}
else if (posBalance >= 63) {
if (MVDiff < 48) {
return 1;
}
else if (MVDiff >= 48) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (AFD < 29) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 8;
}
}
else if (AFD >= 29) {
return 1;
}
}
}
}
}
}
else if (Rs >= 3483) {
if (Cs < 856) {
if (Rs < 4675) {
if (diffMVdiffVal < -10) {
if (Cs < 562) {
return 2;
}
else if (Cs >= 562) {
if (Cs < 807) {
return 1;
}
else if (Cs >= 807) {
return 2;
}
}
}
else if (diffMVdiffVal >= -10) {
if (MVDiff < 34) {
return 2;
}
else if (MVDiff >= 34) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (Rs < 4080) {
return 8;
}
else if (Rs >= 4080) {
if (SC < 4955) {
return 2;
}
else if (SC >= 4955) {
return 8;
}
}
}
}
}
}
else if (Rs >= 4675) {
return 2;
}
}
else if (Cs >= 856) {
if (mu_mv_mag_sq < 9) {
if (diffRsCsdiff < 238) {
if (diffAFD < -3) {
if (Cs < 987) {
if (Rs < 4493) {
return 2;
}
else if (Rs >= 4493) {
return 8;
}
}
else if (Cs >= 987) {
return 2;
}
}
else if (diffAFD >= -3) {
if (negBalance < 60) {
if (RsCsDiff < 682) {
if (diffTSC < -1) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
if (Cs < 1144) {
return 2;
}
else if (Cs >= 1144) {
return 8;
}
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (RsCsDiff >= 682) {
if (Cs < 983) {
return 8;
}
else if (Cs >= 983) {
if (MVDiff < 35) {
return 8;
}
else if (MVDiff >= 35) {
if (AFD < 32) {
return 2;
}
else if (AFD >= 32) {
if (Rs < 5974) {
return 2;
}
else if (Rs >= 5974) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 60) {
if (Cs < 1041) {
return 8;
}
else if (Cs >= 1041) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 238) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (tcor < 81) {
if (Rs < 4785) {
return 2;
}
else if (Rs >= 4785) {
return 1;
}
}
else if (tcor >= 81) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 132) {
if (mcTcor < 93) {
if (AFD < 19) {
if (Cs < 1028) {
if (Rs < 5749) {
if (ssDCval < 133) {
return 1;
}
else if (ssDCval >= 133) {
return 4;
}
}
else if (Rs >= 5749) {
return 2;
}
}
else if (Cs >= 1028) {
if (Cs < 1151) {
return 1;
}
else if (Cs >= 1151) {
return 8;
}
}
}
else if (AFD >= 19) {
if (Rs < 5881) {
if (mu_mv_mag_sq < 26) {
if (MVDiff < 56) {
if (Rs < 4943) {
return 1;
}
else if (Rs >= 4943) {
if (Cs < 938) {
return 2;
}
else if (Cs >= 938) {
if (Rs < 5627) {
return 8;
}
else if (Rs >= 5627) {
return 1;
}
}
}
}
else if (MVDiff >= 56) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (Rs < 3637) {
return 2;
}
else if (Rs >= 3637) {
return 1;
}
}
}
else if (Rs >= 5881) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (CsDiff < 4) {
if (SC < 3804) {
if (MVDiff < 29) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
if (Rs < 2522) {
return 1;
}
else if (Rs >= 2522) {
return 2;
}
}
}
else if (MVDiff >= 29) {
return 1;
}
}
else if (SC >= 3804) {
if (mcTcor < 94) {
if (negBalance < 51) {
if (Rs < 3498) {
return 8;
}
else if (Rs >= 3498) {
return 2;
}
}
else if (negBalance >= 51) {
return 4;
}
}
else if (mcTcor >= 94) {
if (Rs < 2895) {
return 8;
}
else if (Rs >= 2895) {
if (ssDCval < 136) {
if (Cs < 583) {
return 2;
}
else if (Cs >= 583) {
if (ssDCval < 132) {
return 2;
}
else if (ssDCval >= 132) {
if (Rs < 6839) {
return 1;
}
else if (Rs >= 6839) {
return 2;
}
}
}
}
else if (ssDCval >= 136) {
return 2;
}
}
}
}
}
else if (CsDiff >= 4) {
if (Rs < 3772) {
if (Cs < 1258) {
if (tcor < 73) {
if (diffRsCsdiff < 127) {
return 2;
}
else if (diffRsCsdiff >= 127) {
return 1;
}
}
else if (tcor >= 73) {
if (MVDiff < 85) {
return 1;
}
else if (MVDiff >= 85) {
if (SC < 3621) {
return 1;
}
else if (SC >= 3621) {
return 8;
}
}
}
}
else if (Cs >= 1258) {
if (Cs < 1274) {
return 2;
}
else if (Cs >= 1274) {
return 8;
}
}
}
else if (Rs >= 3772) {
if (negBalance < 49) {
if (Rs < 6446) {
if (Rs < 3947) {
return 8;
}
else if (Rs >= 3947) {
return 1;
}
}
else if (Rs >= 6446) {
return 2;
}
}
else if (negBalance >= 49) {
if (Cs < 882) {
if (SC < 4441) {
return 1;
}
else if (SC >= 4441) {
return 2;
}
}
else if (Cs >= 882) {
if (tcor < 89) {
if (Rs < 4905) {
if (mcTcor < 97) {
if (Rs < 3826) {
return 2;
}
else if (Rs >= 3826) {
return 1;
}
}
else if (mcTcor >= 97) {
return 8;
}
}
else if (Rs >= 4905) {
if (MVDiff < 26) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (MVDiff >= 26) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 1;
}
}
}
}
else if (tcor >= 89) {
if (AFD < 16) {
if (Cs < 1065) {
return 2;
}
else if (Cs >= 1065) {
return 1;
}
}
else if (AFD >= 16) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1313) {
if (mu_mv_mag_sq < 105) {
if (RsDiff < 9) {
if (CsDiff < 8) {
if (refDCval < 90) {
if (negBalance < 38) {
if (SC < 10283) {
if (posBalance < 45) {
if (refDCval < 64) {
if (Rs < 2997) {
if (Rs < 853) {
if (Rs < 321) {
return 4;
}
else if (Rs >= 321) {
return 1;
}
}
else if (Rs >= 853) {
if (posBalance < 32) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (posBalance >= 32) {
if (MVDiff < 2) {
if (posBalance < 38) {
return 2;
}
else if (posBalance >= 38) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Rs < 1707) {
if (AFD < 8) {
if (posBalance < 37) {
return 4;
}
else if (posBalance >= 37) {
return 8;
}
}
else if (AFD >= 8) {
return 2;
}
}
else if (Rs >= 1707) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
return 4;
}
}
}
}
}
}
else if (Rs >= 2997) {
return 4;
}
}
else if (refDCval >= 64) {
if (refDCval < 69) {
if (mu_mv_mag_sq < 68) {
if (Rs < 1753) {
return 1;
}
else if (Rs >= 1753) {
return 2;
}
}
else if (mu_mv_mag_sq >= 68) {
if (MVDiff < 6) {
if (Cs < 1931) {
return 8;
}
else if (Cs >= 1931) {
return 1;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (refDCval >= 69) {
if (Rs < 3519) {
if (MVDiff < 8) {
if (Rs < 1905) {
return 8;
}
else if (Rs >= 1905) {
if (diffRsCsdiff < 8) {
return 1;
}
else if (diffRsCsdiff >= 8) {
if (Rs < 3160) {
if (Rs < 2158) {
return 8;
}
else if (Rs >= 2158) {
return 1;
}
}
else if (Rs >= 3160) {
return 2;
}
}
}
}
else if (MVDiff >= 8) {
if (Cs < 1427) {
return 4;
}
else if (Cs >= 1427) {
return 8;
}
}
}
else if (Rs >= 3519) {
if (RsCsDiff < 57) {
if (Rs < 4414) {
return 4;
}
else if (Rs >= 4414) {
if (Cs < 1727) {
return 8;
}
else if (Cs >= 1727) {
if (diffRsCsdiff < 13) {
if (mu_mv_mag_sq < 103) {
return 4;
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 5624) {
return 8;
}
else if (Rs >= 5624) {
return 4;
}
}
}
else if (diffRsCsdiff >= 13) {
return 8;
}
}
}
}
else if (RsCsDiff >= 57) {
if (posBalance < 43) {
if (ssDCval < 78) {
return 4;
}
else if (ssDCval >= 78) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (posBalance >= 43) {
if (diffRsCsdiff < 5) {
return 2;
}
else if (diffRsCsdiff >= 5) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2688) {
if (Rs < 1452) {
if (mu_mv_mag_sq < 83) {
if (TSC < 6) {
if (Cs < 1567) {
if (MVDiff < 15) {
if (AFD < 9) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 1155) {
return 2;
}
else if (Rs >= 1155) {
return 4;
}
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (MVDiff >= 15) {
return 1;
}
}
else if (Cs >= 1567) {
if (MVDiff < 105) {
return 1;
}
else if (MVDiff >= 105) {
return 8;
}
}
}
else if (TSC >= 6) {
if (gchDC < 4) {
if (Rs < 1335) {
return 8;
}
else if (Rs >= 1335) {
if (Cs < 1487) {
return 8;
}
else if (Cs >= 1487) {
return 4;
}
}
}
else if (gchDC >= 4) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (mcTcor < 98) {
if (Cs < 1637) {
if (Rs < 1316) {
if (Rs < 772) {
return 1;
}
else if (Rs >= 772) {
return 4;
}
}
else if (Rs >= 1316) {
return 8;
}
}
else if (Cs >= 1637) {
if (Rs < 1307) {
return 8;
}
else if (Rs >= 1307) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (Cs < 1993) {
if (Cs < 1331) {
return 1;
}
else if (Cs >= 1331) {
return 8;
}
}
else if (Cs >= 1993) {
return 1;
}
}
}
}
else if (Rs >= 1452) {
if (Cs < 1969) {
if (mu_mv_mag_sq < 91) {
if (mu_mv_mag_sq < 90) {
if (CsDiff < 7) {
if (posBalance < 52) {
if (ssDCval < 84) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (mu_mv_mag_sq < 70) {
return 8;
}
else if (mu_mv_mag_sq >= 70) {
return 1;
}
}
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (posBalance >= 52) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Rs < 2009) {
return 8;
}
else if (Rs >= 2009) {
return 4;
}
}
}
}
}
else if (CsDiff >= 7) {
if (Cs < 1867) {
if (Rs < 1705) {
return 1;
}
else if (Rs >= 1705) {
return 8;
}
}
else if (Cs >= 1867) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 90) {
return 4;
}
}
else if (mu_mv_mag_sq >= 91) {
if (Rs < 1574) {
return 4;
}
else if (Rs >= 1574) {
return 8;
}
}
}
else if (Cs >= 1969) {
if (Rs < 2427) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
return 8;
}
}
else if (Rs >= 2427) {
if (Cs < 2538) {
return 8;
}
else if (Cs >= 2538) {
return 4;
}
}
}
}
}
else if (Rs >= 2688) {
if (Cs < 1805) {
if (RsCsDiff < 55) {
if (diffRsCsdiff < -20) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (Rs < 3029) {
return 2;
}
else if (Rs >= 3029) {
return 4;
}
}
}
else if (diffRsCsdiff >= -20) {
if (Rs < 4111) {
if (Rs < 3003) {
if (Rs < 2906) {
return 4;
}
else if (Rs >= 2906) {
return 2;
}
}
else if (Rs >= 3003) {
return 4;
}
}
else if (Rs >= 4111) {
return 2;
}
}
}
else if (RsCsDiff >= 55) {
if (tcor < 90) {
if (Rs < 3427) {
return 8;
}
else if (Rs >= 3427) {
return 2;
}
}
else if (tcor >= 90) {
if (refDCval < 79) {
return 2;
}
else if (refDCval >= 79) {
if (SC < 5260) {
if (Cs < 1378) {
return 2;
}
else if (Cs >= 1378) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (SC >= 5260) {
return 2;
}
}
}
}
}
else if (Cs >= 1805) {
if (posBalance < 48) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (Rs < 3766) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (Rs >= 3766) {
return 4;
}
}
}
else if (posBalance >= 48) {
if (Cs < 1936) {
if (RsCsDiff < 56) {
return 4;
}
else if (RsCsDiff >= 56) {
return 2;
}
}
else if (Cs >= 1936) {
if (Rs < 2837) {
return 4;
}
else if (Rs >= 2837) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 10283) {
if (Rs < 5682) {
return 8;
}
else if (Rs >= 5682) {
return 2;
}
}
}
else if (negBalance >= 38) {
if (Cs < 3414) {
if (mcTcor < 94) {
if (Rs < 1397) {
if (negBalance < 56) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 38) {
if (diffRsCsdiff < 19) {
return 4;
}
else if (diffRsCsdiff >= 19) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (Cs < 1377) {
if (ssDCval < 65) {
if (Rs < 1376) {
return 1;
}
else if (Rs >= 1376) {
return 8;
}
}
else if (ssDCval >= 65) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (Cs >= 1377) {
if (diffRsCsdiff < 12) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Rs < 839) {
return 4;
}
else if (Rs >= 839) {
return 8;
}
}
}
else if (diffRsCsdiff >= 12) {
if (Rs < 1386) {
return 4;
}
else if (Rs >= 1386) {
return 1;
}
}
}
}
}
else if (MVDiff >= 8) {
if (ssDCval < 48) {
return 4;
}
else if (ssDCval >= 48) {
if (ssDCval < 81) {
if (diffMVdiffVal < 7) {
if (mcTcor < 83) {
return 8;
}
else if (mcTcor >= 83) {
if (tcor < 79) {
if (MVDiff < 12) {
if (Rs < 1121) {
return 2;
}
else if (Rs >= 1121) {
return 8;
}
}
else if (MVDiff >= 12) {
return 2;
}
}
else if (tcor >= 79) {
if (SC < 2561) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (SC >= 2561) {
if (Rs < 1018) {
return 8;
}
else if (Rs >= 1018) {
if (Rs < 1301) {
return 4;
}
else if (Rs >= 1301) {
return 8;
}
}
}
}
}
}
else if (diffMVdiffVal >= 7) {
return 4;
}
}
else if (ssDCval >= 81) {
if (mu_mv_mag_sq < 9) {
return 4;
}
else if (mu_mv_mag_sq >= 9) {
return 8;
}
}
}
}
}
else if (negBalance >= 56) {
if (RsCsDiff < 80) {
if (tcor < 78) {
if (Rs < 1396) {
return 8;
}
else if (Rs >= 1396) {
return 4;
}
}
else if (tcor >= 78) {
return 4;
}
}
else if (RsCsDiff >= 80) {
return 4;
}
}
}
else if (Rs >= 1397) {
if (Cs < 2929) {
if (MVDiff < 4) {
if (diffMVdiffVal < 1) {
if (posBalance < 70) {
if (RsCsDiff < 73) {
return 8;
}
else if (RsCsDiff >= 73) {
if (posBalance < 58) {
if (AFD < 12) {
if (AFD < 9) {
if (Rs < 2380) {
return 8;
}
else if (Rs >= 2380) {
return 4;
}
}
else if (AFD >= 9) {
return 8;
}
}
else if (AFD >= 12) {
if (Rs < 2264) {
return 1;
}
else if (Rs >= 2264) {
return 4;
}
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
else if (posBalance >= 70) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 2709) {
if (Cs < 1341) {
if (Rs < 1755) {
return 8;
}
else if (Rs >= 1755) {
return 2;
}
}
else if (Cs >= 1341) {
if (Cs < 1410) {
if (Cs < 1405) {
return 8;
}
else if (Cs >= 1405) {
return 1;
}
}
else if (Cs >= 1410) {
return 8;
}
}
}
else if (Cs >= 2709) {
return 2;
}
}
}
else if (MVDiff >= 4) {
if (diffRsCsdiff < -15) {
if (mu_mv_mag_sq < 76) {
if (MVDiff < 13) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
if (tcor < 84) {
if (mcTcor < 89) {
if (posBalance < 50) {
if (Rs < 1895) {
return 4;
}
else if (Rs >= 1895) {
return 8;
}
}
else if (posBalance >= 50) {
return 8;
}
}
else if (mcTcor >= 89) {
if (mu_mv_mag_sq < 3) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (tcor >= 84) {
return 8;
}
}
}
else if (MVDiff >= 13) {
if (posBalance < 56) {
if (Rs < 1949) {
if (MVDiff < 37) {
return 1;
}
else if (MVDiff >= 37) {
return 8;
}
}
else if (Rs >= 1949) {
if (Cs < 1436) {
return 2;
}
else if (Cs >= 1436) {
return 8;
}
}
}
else if (posBalance >= 56) {
if (Rs < 1440) {
return 4;
}
else if (Rs >= 1440) {
if (diffTSC < 1) {
if (diffRsCsdiff < -40) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (diffRsCsdiff >= -40) {
return 8;
}
}
else if (diffTSC >= 1) {
if (RsDiff < 8) {
if (Rs < 1627) {
return 8;
}
else if (Rs >= 1627) {
return 4;
}
}
else if (RsDiff >= 8) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (diffAFD < -1) {
if (AFD < 12) {
if (Rs < 2716) {
return 4;
}
else if (Rs >= 2716) {
if (Rs < 3582) {
return 8;
}
else if (Rs >= 3582) {
return 4;
}
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (diffAFD >= -1) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 2718) {
return 4;
}
else if (Rs >= 2718) {
return 8;
}
}
}
}
}
}
}
else if (diffRsCsdiff >= -15) {
if (AFD < 16) {
if (ssDCval < 83) {
if (RsDiff < 8) {
if (SC < 2771) {
return 1;
}
else if (SC >= 2771) {
if (diffRsCsdiff < -12) {
if (TSC < 10) {
if (negBalance < 50) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
if (Rs < 1992) {
return 4;
}
else if (Rs >= 1992) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (Cs < 1458) {
return 8;
}
else if (Cs >= 1458) {
return 4;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (diffRsCsdiff >= -12) {
if (CsDiff < 7) {
if (posBalance < 71) {
if (diffTSC < -3) {
if (Rs < 2079) {
return 8;
}
else if (Rs >= 2079) {
return 4;
}
}
else if (diffTSC >= -3) {
if (diffAFD < -3) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 4;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
}
else if (posBalance >= 71) {
return 4;
}
}
else if (CsDiff >= 7) {
if (ssDCval < 49) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
else if (ssDCval >= 49) {
if (diffRsCsdiff < 31) {
return 8;
}
else if (diffRsCsdiff >= 31) {
if (Rs < 2291) {
return 4;
}
else if (Rs >= 2291) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (diffAFD < 2) {
if (TSC < 9) {
if (ssDCval < 51) {
if (Rs < 1759) {
return 1;
}
else if (Rs >= 1759) {
return 8;
}
}
else if (ssDCval >= 51) {
if (AFD < 9) {
if (Rs < 3324) {
return 8;
}
else if (Rs >= 3324) {
return 4;
}
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (TSC >= 9) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 2176) {
return 4;
}
else if (Rs >= 2176) {
return 8;
}
}
}
}
else if (diffAFD >= 2) {
if (negBalance < 50) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
return 4;
}
}
else if (negBalance >= 50) {
return 4;
}
}
}
}
else if (ssDCval >= 83) {
if (TSC < 8) {
if (RsCsDiff < 57) {
return 8;
}
else if (RsCsDiff >= 57) {
return 2;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (AFD >= 16) {
if (refDCval < 85) {
if (negBalance < 47) {
if (Cs < 1692) {
return 4;
}
else if (Cs >= 1692) {
return 1;
}
}
else if (negBalance >= 47) {
if (MVDiff < 28) {
if (negBalance < 62) {
return 8;
}
else if (negBalance >= 62) {
return 2;
}
}
else if (MVDiff >= 28) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 4;
}
}
}
}
else if (refDCval >= 85) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
if (Cs < 1637) {
return 8;
}
else if (Cs >= 1637) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 2929) {
if (Rs < 3069) {
return 4;
}
else if (Rs >= 3069) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (RsDiff < 8) {
if (Rs < 4201) {
if (Rs < 2122) {
if (mu_mv_mag_sq < 90) {
if (Cs < 2009) {
if (ssDCval < 84) {
if (ssDCval < 81) {
if (Rs < 833) {
if (MVDiff < 6) {
if (Rs < 511) {
return 1;
}
else if (Rs >= 511) {
return 8;
}
}
else if (MVDiff >= 6) {
if (MVDiff < 21) {
if (Rs < 547) {
return 1;
}
else if (Rs >= 547) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (MVDiff >= 21) {
return 8;
}
}
}
else if (Rs >= 833) {
if (Cs < 1859) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
if (Rs < 1090) {
if (ssDCval < 67) {
if (diffTSC < 1) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 2;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 1056) {
return 1;
}
else if (Rs >= 1056) {
return 8;
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 51) {
if (Cs < 1328) {
return 4;
}
else if (Cs >= 1328) {
return 8;
}
}
else if (negBalance >= 51) {
if (posBalance < 58) {
if (Rs < 890) {
return 8;
}
else if (Rs >= 890) {
return 2;
}
}
else if (posBalance >= 58) {
return 4;
}
}
}
}
else if (Rs >= 1090) {
if (tcor < 92) {
if (RsDiff < 6) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (RsDiff >= 6) {
if (Cs < 1568) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (Cs >= 1568) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
return 4;
}
}
}
}
else if (tcor >= 92) {
if (SC < 2850) {
if (Rs < 1191) {
return 8;
}
else if (Rs >= 1191) {
return 4;
}
}
else if (SC >= 2850) {
if (Rs < 2057) {
if (diffRsCsdiff < -8) {
if (mcTcor < 96) {
if (diffTSC < 1) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
return 1;
}
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (mcTcor >= 96) {
return 4;
}
}
else if (diffRsCsdiff >= -8) {
return 8;
}
}
else if (Rs >= 2057) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1859) {
if (AFD < 7) {
if (Rs < 1453) {
return 1;
}
else if (Rs >= 1453) {
return 8;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
}
else if (ssDCval >= 81) {
if (MVDiff < 4) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
else if (MVDiff >= 4) {
if (AFD < 14) {
if (mcTcor < 95) {
if (Rs < 1468) {
return 8;
}
else if (Rs >= 1468) {
return 2;
}
}
else if (mcTcor >= 95) {
if (RsCsDiff < 80) {
return 8;
}
else if (RsCsDiff >= 80) {
if (Rs < 1892) {
return 8;
}
else if (Rs >= 1892) {
return 1;
}
}
}
}
else if (AFD >= 14) {
return 4;
}
}
}
}
else if (ssDCval >= 84) {
if (RsCsDiff < 56) {
return 8;
}
else if (RsCsDiff >= 56) {
if (diffMVdiffVal < 1) {
if (Rs < 977) {
if (Rs < 646) {
return 8;
}
else if (Rs >= 646) {
return 4;
}
}
else if (Rs >= 977) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 2;
}
}
}
}
}
else if (Cs >= 2009) {
if (diffMVdiffVal < 2) {
if (posBalance < 61) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
if (Rs < 1744) {
return 8;
}
else if (Rs >= 1744) {
return 1;
}
}
}
else if (posBalance >= 61) {
if (Rs < 1795) {
return 8;
}
else if (Rs >= 1795) {
return 1;
}
}
}
else if (diffMVdiffVal >= 2) {
if (MVDiff < 25) {
if (Rs < 1067) {
return 2;
}
else if (Rs >= 1067) {
return 1;
}
}
else if (MVDiff >= 25) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
if (ssDCval < 80) {
if (diffRsCsdiff < 6) {
return 8;
}
else if (diffRsCsdiff >= 6) {
if (posBalance < 58) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (posBalance >= 58) {
return 1;
}
}
}
else if (ssDCval >= 80) {
if (ssDCval < 82) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
return 1;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
}
}
}
else if (Rs >= 2122) {
if (Cs < 1538) {
if (MVDiff < 7) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 89) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
if (Cs < 1397) {
if (Rs < 2539) {
return 4;
}
else if (Rs >= 2539) {
return 2;
}
}
else if (Cs >= 1397) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (Rs < 2547) {
if (Rs < 2436) {
return 1;
}
else if (Rs >= 2436) {
return 8;
}
}
else if (Rs >= 2547) {
if (Rs < 3597) {
return 2;
}
else if (Rs >= 3597) {
return 4;
}
}
}
}
else if (posBalance >= 53) {
if (ssDCval < 82) {
if (TSC < 6) {
if (Rs < 2831) {
return 1;
}
else if (Rs >= 2831) {
return 2;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (ssDCval >= 82) {
if (Rs < 3066) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
else if (Rs >= 3066) {
return 2;
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 2640) {
return 8;
}
else if (Rs >= 2640) {
if (Rs < 2714) {
return 4;
}
else if (Rs >= 2714) {
return 1;
}
}
}
}
else if (Cs >= 1538) {
if (CsDiff < 4) {
if (negBalance < 40) {
if (diffRsCsdiff < -20) {
return 4;
}
else if (diffRsCsdiff >= -20) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 8;
}
}
}
else if (negBalance >= 40) {
return 8;
}
}
else if (CsDiff >= 4) {
if (SC < 5099) {
if (tcor < 86) {
return 4;
}
else if (tcor >= 86) {
if (Rs < 2954) {
if (posBalance < 45) {
if (Rs < 2707) {
return 8;
}
else if (Rs >= 2707) {
return 1;
}
}
else if (posBalance >= 45) {
if (refDCval < 81) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Cs < 2115) {
if (SC < 4191) {
if (Rs < 2131) {
return 4;
}
else if (Rs >= 2131) {
return 8;
}
}
else if (SC >= 4191) {
return 4;
}
}
else if (Cs >= 2115) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (posBalance < 60) {
if (Cs < 1746) {
if (Cs < 1744) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (Cs >= 1744) {
return 1;
}
}
else if (Cs >= 1746) {
return 8;
}
}
else if (posBalance >= 60) {
if (Cs < 1932) {
return 1;
}
else if (Cs >= 1932) {
return 8;
}
}
}
}
}
else if (Rs >= 2954) {
return 4;
}
}
}
else if (SC >= 5099) {
if (Rs < 2748) {
return 4;
}
else if (Rs >= 2748) {
if (posBalance < 56) {
if (Cs < 2742) {
if (ssDCval < 85) {
if (Rs < 3496) {
if (Cs < 2240) {
return 8;
}
else if (Cs >= 2240) {
return 4;
}
}
else if (Rs >= 3496) {
if (Cs < 1607) {
return 2;
}
else if (Cs >= 1607) {
return 4;
}
}
}
else if (ssDCval >= 85) {
if (Rs < 3874) {
return 2;
}
else if (Rs >= 3874) {
return 1;
}
}
}
else if (Cs >= 2742) {
if (Rs < 3621) {
return 1;
}
else if (Rs >= 3621) {
return 8;
}
}
}
else if (posBalance >= 56) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 3065) {
return 8;
}
else if (Rs >= 3065) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 4201) {
if (Cs < 2792) {
return 4;
}
else if (Cs >= 2792) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (AFD < 10) {
if (Cs < 2134) {
return 2;
}
else if (Cs >= 2134) {
return 4;
}
}
else if (AFD >= 10) {
if (tcor < 88) {
if (mu_mv_mag_sq < 53) {
if (Rs < 2036) {
if (Rs < 1526) {
return 4;
}
else if (Rs >= 1526) {
return 2;
}
}
else if (Rs >= 2036) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (Cs < 2222) {
return 4;
}
else if (Cs >= 2222) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
return 8;
}
}
else if (tcor >= 88) {
if (Rs < 4247) {
if (posBalance < 52) {
if (Cs < 1317) {
return 8;
}
else if (Cs >= 1317) {
return 1;
}
}
else if (posBalance >= 52) {
return 8;
}
}
else if (Rs >= 4247) {
return 4;
}
}
}
}
}
}
else if (Cs >= 3414) {
if (Rs < 3573) {
return 1;
}
else if (Rs >= 3573) {
return 8;
}
}
}
}
else if (refDCval >= 90) {
if (refDCval < 163) {
if (Rs < 4592) {
if (ssDCval < 110) {
if (Cs < 2498) {
if (CsDiff < 1) {
if (negBalance < 44) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
return 4;
}
}
else if (negBalance >= 44) {
if (Rs < 2285) {
return 1;
}
else if (Rs >= 2285) {
return 4;
}
}
}
else if (CsDiff >= 1) {
if (mcTcor < 96) {
if (Rs < 2225) {
if (Rs < 1571) {
if (mcTcor < 90) {
if (Cs < 1694) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (Cs >= 1694) {
if (Cs < 2113) {
return 4;
}
else if (Cs >= 2113) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (Cs < 1964) {
if (posBalance < 60) {
if (Cs < 1337) {
return 4;
}
else if (Cs >= 1337) {
return 8;
}
}
else if (posBalance >= 60) {
if (Rs < 1002) {
return 4;
}
else if (Rs >= 1002) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
}
}
else if (Cs >= 1964) {
if (MVDiff < 6) {
if (Cs < 2291) {
return 8;
}
else if (Cs >= 2291) {
return 2;
}
}
else if (MVDiff >= 6) {
if (Cs < 2001) {
return 4;
}
else if (Cs >= 2001) {
return 2;
}
}
}
}
}
else if (Rs >= 1571) {
if (TSC < 13) {
if (Cs < 1418) {
if (AFD < 11) {
if (AFD < 10) {
if (Rs < 1824) {
return 8;
}
else if (Rs >= 1824) {
if (Cs < 1402) {
return 2;
}
else if (Cs >= 1402) {
return 8;
}
}
}
else if (AFD >= 10) {
return 2;
}
}
else if (AFD >= 11) {
if (Rs < 1845) {
return 8;
}
else if (Rs >= 1845) {
if (mu_mv_mag_sq < 36) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (mu_mv_mag_sq >= 36) {
if (Cs < 1379) {
return 1;
}
else if (Cs >= 1379) {
return 8;
}
}
}
}
}
else if (Cs >= 1418) {
if (Cs < 2306) {
if (TSC < 9) {
if (ssDCval < 102) {
if (Rs < 1858) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
return 2;
}
}
else if (Rs >= 1858) {
if (negBalance < 49) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (negBalance >= 49) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (ssDCval >= 102) {
return 8;
}
}
else if (TSC >= 9) {
if (posBalance < 49) {
if (Rs < 1784) {
return 8;
}
else if (Rs >= 1784) {
return 1;
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
else if (Cs >= 2306) {
if (Rs < 1736) {
if (Rs < 1657) {
return 2;
}
else if (Rs >= 1657) {
return 4;
}
}
else if (Rs >= 1736) {
if (Rs < 1769) {
return 1;
}
else if (Rs >= 1769) {
return 8;
}
}
}
}
}
else if (TSC >= 13) {
if (Rs < 2198) {
return 1;
}
else if (Rs >= 2198) {
return 8;
}
}
}
}
else if (Rs >= 2225) {
if (mcTcor < 94) {
if (Rs < 4183) {
if (TSC < 11) {
if (diffRsCsdiff < 28) {
return 8;
}
else if (diffRsCsdiff >= 28) {
if (refDCval < 106) {
return 8;
}
else if (refDCval >= 106) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (TSC >= 11) {
if (ssDCval < 108) {
if (diffAFD < -1) {
if (Rs < 2534) {
return 8;
}
else if (Rs >= 2534) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (ssDCval >= 108) {
return 1;
}
}
}
else if (Rs >= 4183) {
if (mu_mv_mag_sq < 74) {
return 8;
}
else if (mu_mv_mag_sq >= 74) {
if (Rs < 4299) {
return 4;
}
else if (Rs >= 4299) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (refDCval < 100) {
if (RsDiff < 8) {
if (negBalance < 55) {
if (negBalance < 33) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (negBalance >= 33) {
return 8;
}
}
else if (negBalance >= 55) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (mu_mv_mag_sq < 82) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
if (Cs < 1377) {
return 4;
}
else if (Cs >= 1377) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 82) {
return 2;
}
}
}
else if (refDCval >= 100) {
if (Rs < 3296) {
if (SC < 4985) {
if (mu_mv_mag_sq < 55) {
if (AFD < 12) {
if (Rs < 3127) {
return 4;
}
else if (Rs >= 3127) {
return 8;
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 55) {
return 8;
}
}
else if (SC >= 4985) {
if (Rs < 2991) {
return 8;
}
else if (Rs >= 2991) {
return 4;
}
}
}
else if (Rs >= 3296) {
if (mu_mv_mag_sq < 90) {
return 8;
}
else if (mu_mv_mag_sq >= 90) {
if (Rs < 4110) {
return 8;
}
else if (Rs >= 4110) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (RsCsDiff < 41) {
if (MVDiff < 4) {
if (Cs < 1592) {
if (ssDCval < 100) {
if (TSC < 5) {
if (Rs < 4001) {
return 2;
}
else if (Rs >= 4001) {
return 8;
}
}
else if (TSC >= 5) {
if (Rs < 3313) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
else if (Rs >= 3313) {
return 4;
}
}
}
else if (ssDCval >= 100) {
if (ssDCval < 105) {
return 4;
}
else if (ssDCval >= 105) {
if (Rs < 849) {
return 1;
}
else if (Rs >= 849) {
return 8;
}
}
}
}
else if (Cs >= 1592) {
if (tcor < 96) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (tcor >= 96) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 107) {
if (posBalance < 27) {
if (RsDiff < 3) {
if (Rs < 2109) {
return 2;
}
else if (Rs >= 2109) {
return 1;
}
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (posBalance >= 27) {
if (tcor < 93) {
if (Cs < 1345) {
return 4;
}
else if (Cs >= 1345) {
return 8;
}
}
else if (tcor >= 93) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Cs < 1552) {
return 2;
}
else if (Cs >= 1552) {
return 8;
}
}
}
}
}
else if (ssDCval >= 107) {
if (negBalance < 41) {
if (Rs < 2451) {
return 8;
}
else if (Rs >= 2451) {
return 4;
}
}
else if (negBalance >= 41) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 41) {
if (mcTcor < 97) {
if (SC < 4382) {
if (diffRsCsdiff < 19) {
if (diffRsCsdiff < -35) {
return 1;
}
else if (diffRsCsdiff >= -35) {
return 8;
}
}
else if (diffRsCsdiff >= 19) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
if (Rs < 2003) {
return 2;
}
else if (Rs >= 2003) {
return 8;
}
}
}
}
else if (SC >= 4382) {
if (Rs < 3055) {
if (AFD < 10) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
return 4;
}
}
else if (AFD >= 10) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
return 8;
}
}
}
else if (Rs >= 3055) {
if (MVDiff < 8) {
if (Rs < 4318) {
if (ssDCval < 99) {
return 2;
}
else if (ssDCval >= 99) {
if (mu_mv_mag_sq < 86) {
return 8;
}
else if (mu_mv_mag_sq >= 86) {
if (Rs < 3502) {
return 4;
}
else if (Rs >= 3502) {
return 2;
}
}
}
}
else if (Rs >= 4318) {
return 8;
}
}
else if (MVDiff >= 8) {
if (Rs < 4252) {
return 8;
}
else if (Rs >= 4252) {
return 1;
}
}
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 42) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Cs < 1443) {
return 2;
}
else if (Cs >= 1443) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (Rs < 3490) {
if (Rs < 2053) {
if (tcor < 91) {
if (TSC < 7) {
if (ssDCval < 105) {
return 8;
}
else if (ssDCval >= 105) {
return 2;
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (tcor >= 91) {
if (RsCsDiff < 56) {
if (Cs < 1980) {
if (diffRsCsdiff < -6) {
return 2;
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
else if (Cs >= 1980) {
return 2;
}
}
else if (RsCsDiff >= 56) {
if (Cs < 1414) {
if (Rs < 1415) {
return 2;
}
else if (Rs >= 1415) {
if (Rs < 1803) {
return 8;
}
else if (Rs >= 1803) {
return 1;
}
}
}
else if (Cs >= 1414) {
return 2;
}
}
}
}
else if (Rs >= 2053) {
if (RsDiff < 5) {
if (Rs < 2343) {
return 8;
}
else if (Rs >= 2343) {
return 4;
}
}
else if (RsDiff >= 5) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
return 2;
}
}
}
}
else if (Rs >= 3490) {
if (Rs < 4471) {
if (mu_mv_mag_sq < 102) {
return 2;
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 4273) {
return 8;
}
else if (Rs >= 4273) {
return 2;
}
}
}
else if (Rs >= 4471) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 2498) {
if (tcor < 93) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (tcor >= 93) {
if (RsCsDiff < 46) {
return 8;
}
else if (RsCsDiff >= 46) {
if (Cs < 2854) {
return 8;
}
else if (Cs >= 2854) {
return 4;
}
}
}
}
}
else if (ssDCval >= 110) {
if (Rs < 1454) {
if (ssDCval < 147) {
if (AFD < 18) {
if (negBalance < 45) {
if (Cs < 1581) {
if (posBalance < 50) {
if (negBalance < 33) {
if (ssDCval < 111) {
if (Rs < 1041) {
return 1;
}
else if (Rs >= 1041) {
return 8;
}
}
else if (ssDCval >= 111) {
return 4;
}
}
else if (negBalance >= 33) {
return 8;
}
}
else if (posBalance >= 50) {
if (RsDiff < 1) {
if (Rs < 893) {
return 1;
}
else if (Rs >= 893) {
return 4;
}
}
else if (RsDiff >= 1) {
if (SC < 2819) {
if (Cs < 1538) {
return 8;
}
else if (Cs >= 1538) {
if (Rs < 1063) {
return 8;
}
else if (Rs >= 1063) {
return 4;
}
}
}
else if (SC >= 2819) {
if (tcor < 91) {
return 1;
}
else if (tcor >= 91) {
return 8;
}
}
}
}
}
else if (Cs >= 1581) {
if (Rs < 972) {
if (ssDCval < 114) {
return 4;
}
else if (ssDCval >= 114) {
return 8;
}
}
else if (Rs >= 972) {
if (refDCval < 133) {
if (gchDC < 2) {
if (diffAFD < -2) {
if (Rs < 1013) {
return 4;
}
else if (Rs >= 1013) {
return 8;
}
}
else if (diffAFD >= -2) {
if (MVDiff < 2) {
if (Rs < 1192) {
return 4;
}
else if (Rs >= 1192) {
return 8;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (refDCval >= 133) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 1417) {
if (ssDCval < 143) {
return 8;
}
else if (ssDCval >= 143) {
return 1;
}
}
else if (Rs >= 1417) {
return 2;
}
}
}
else if (AFD >= 18) {
if (Rs < 1245) {
if (Rs < 888) {
return 2;
}
else if (Rs >= 888) {
return 1;
}
}
else if (Rs >= 1245) {
return 4;
}
}
}
else if (ssDCval >= 147) {
if (mu_mv_mag_sq < 38) {
return 2;
}
else if (mu_mv_mag_sq >= 38) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 1405) {
return 4;
}
else if (Rs >= 1405) {
return 2;
}
}
}
}
}
else if (Rs >= 1454) {
if (refDCval < 140) {
if (refDCval < 135) {
if (Rs < 1899) {
if (ssDCval < 128) {
if (mu_mv_mag_sq < 11) {
if (tcor < 91) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (Cs < 1755) {
return 4;
}
else if (Cs >= 1755) {
if (Rs < 1524) {
return 1;
}
else if (Rs >= 1524) {
return 8;
}
}
}
}
else if (tcor >= 91) {
return 2;
}
}
else if (mu_mv_mag_sq >= 11) {
if (SC < 2915) {
if (Rs < 1525) {
return 8;
}
else if (Rs >= 1525) {
return 2;
}
}
else if (SC >= 2915) {
return 8;
}
}
}
else if (ssDCval >= 128) {
if (CsDiff < 5) {
if (refDCval < 128) {
return 8;
}
else if (refDCval >= 128) {
return 1;
}
}
else if (CsDiff >= 5) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
}
}
else if (Rs >= 1899) {
if (AFD < 11) {
if (Cs < 1375) {
if (tcor < 95) {
if (Rs < 3824) {
return 8;
}
else if (Rs >= 3824) {
return 1;
}
}
else if (tcor >= 95) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
else if (Cs >= 1375) {
if (Rs < 1952) {
if (ssDCval < 127) {
if (diffAFD < -1) {
if (Rs < 1928) {
return 8;
}
else if (Rs >= 1928) {
return 2;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (ssDCval >= 127) {
if (Rs < 1946) {
return 8;
}
else if (Rs >= 1946) {
return 1;
}
}
}
else if (Rs >= 1952) {
if (Rs < 4300) {
return 8;
}
else if (Rs >= 4300) {
if (TSC < 5) {
if (Rs < 4377) {
return 1;
}
else if (Rs >= 4377) {
return 2;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (AFD >= 11) {
if (refDCval < 128) {
if (refDCval < 123) {
if (Cs < 1616) {
if (mu_mv_mag_sq < 78) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
return 1;
}
}
else if (mu_mv_mag_sq >= 78) {
return 1;
}
}
else if (Cs >= 1616) {
if (TSC < 6) {
if (Rs < 2117) {
return 8;
}
else if (Rs >= 2117) {
return 2;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (refDCval >= 123) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
else if (refDCval >= 128) {
return 8;
}
}
}
}
else if (refDCval >= 135) {
if (mcTcor < 92) {
if (diffTSC < 1) {
if (Cs < 1568) {
return 8;
}
else if (Cs >= 1568) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
if (Rs < 2762) {
return 1;
}
else if (Rs >= 2762) {
return 8;
}
}
}
}
}
else if (diffTSC >= 1) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (mcTcor < 97) {
if (MVDiff < 29) {
return 8;
}
else if (MVDiff >= 29) {
if (Rs < 2075) {
return 8;
}
else if (Rs >= 2075) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 37) {
if (Rs < 3065) {
return 1;
}
else if (Rs >= 3065) {
return 2;
}
}
else if (negBalance >= 37) {
return 8;
}
}
}
}
}
else if (refDCval >= 140) {
if (AFD < 18) {
if (mu_mv_mag_sq < 36) {
if (Rs < 1800) {
return 2;
}
else if (Rs >= 1800) {
return 8;
}
}
else if (mu_mv_mag_sq >= 36) {
if (posBalance < 62) {
if (SC < 5803) {
if (diffMVdiffVal < -35) {
if (Rs < 2616) {
return 8;
}
else if (Rs >= 2616) {
return 2;
}
}
else if (diffMVdiffVal >= -35) {
if (negBalance < 27) {
if (Rs < 2816) {
return 8;
}
else if (Rs >= 2816) {
return 4;
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (SC >= 5803) {
return 1;
}
}
else if (posBalance >= 62) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (Rs < 2548) {
return 4;
}
else if (Rs >= 2548) {
return 8;
}
}
}
}
}
else if (AFD >= 18) {
if (TSC < 13) {
if (Rs < 3108) {
return 1;
}
else if (Rs >= 3108) {
return 8;
}
}
else if (TSC >= 13) {
return 2;
}
}
}
}
}
}
else if (Rs >= 4592) {
if (Cs < 1590) {
if (tcor < 94) {
if (Rs < 4826) {
return 1;
}
else if (Rs >= 4826) {
return 8;
}
}
else if (tcor >= 94) {
return 2;
}
}
else if (Cs >= 1590) {
if (TSC < 8) {
if (posBalance < 36) {
if (Rs < 5270) {
return 1;
}
else if (Rs >= 5270) {
if (Rs < 5609) {
return 8;
}
else if (Rs >= 5609) {
return 2;
}
}
}
else if (posBalance >= 36) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
if (Rs < 5853) {
if (ssDCval < 91) {
if (Rs < 4703) {
return 2;
}
else if (Rs >= 4703) {
return 4;
}
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (Rs >= 5853) {
return 4;
}
}
}
}
else if (TSC >= 8) {
if (AFD < 11) {
if (Rs < 4999) {
return 1;
}
else if (Rs >= 4999) {
return 8;
}
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
else if (refDCval >= 163) {
if (Rs < 2306) {
if (Rs < 1062) {
return 1;
}
else if (Rs >= 1062) {
return 2;
}
}
else if (Rs >= 2306) {
return 8;
}
}
}
}
else if (CsDiff >= 8) {
if (Rs < 3843) {
if (tcor < 64) {
if (refDCval < 116) {
if (negBalance < 48) {
if (Cs < 2729) {
if (refDCval < 50) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (Rs < 939) {
return 2;
}
else if (Rs >= 939) {
return 1;
}
}
}
else if (refDCval >= 50) {
if (Rs < 602) {
return 4;
}
else if (Rs >= 602) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (ssDCval < 60) {
if (Rs < 1015) {
return 4;
}
else if (Rs >= 1015) {
return 2;
}
}
else if (ssDCval >= 60) {
return 8;
}
}
}
}
}
else if (Cs >= 2729) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (Rs < 1380) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (diffRsCsdiff < 22) {
if (tcor < 57) {
if (Rs < 968) {
return 2;
}
else if (Rs >= 968) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
return 4;
}
}
}
else if (tcor >= 57) {
if (diffMVdiffVal < -1) {
if (Rs < 1126) {
return 8;
}
else if (Rs >= 1126) {
if (Rs < 1255) {
return 2;
}
else if (Rs >= 1255) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
else if (diffRsCsdiff >= 22) {
if (MVDiff < 54) {
if (Rs < 1184) {
return 2;
}
else if (Rs >= 1184) {
if (Cs < 4010) {
return 4;
}
else if (Cs >= 4010) {
return 8;
}
}
}
else if (MVDiff >= 54) {
return 8;
}
}
}
}
else if (Rs >= 1380) {
if (MVDiff < 46) {
if (SC < 4155) {
if (tcor < 56) {
if (Rs < 1664) {
return 4;
}
else if (Rs >= 1664) {
return 2;
}
}
else if (tcor >= 56) {
if (MVDiff < 24) {
if (diffMVdiffVal < -8) {
if (Rs < 1761) {
return 4;
}
else if (Rs >= 1761) {
return 2;
}
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
else if (MVDiff >= 24) {
if (Rs < 1463) {
return 1;
}
else if (Rs >= 1463) {
return 4;
}
}
}
}
else if (SC >= 4155) {
return 4;
}
}
else if (MVDiff >= 46) {
if (diffRsCsdiff < 8) {
if (Rs < 1830) {
return 4;
}
else if (Rs >= 1830) {
return 8;
}
}
else if (diffRsCsdiff >= 8) {
return 1;
}
}
}
}
}
else if (refDCval >= 116) {
if (Cs < 2951) {
return 2;
}
else if (Cs >= 2951) {
return 8;
}
}
}
else if (tcor >= 64) {
if (mcTcor < 94) {
if (Cs < 2795) {
if (Rs < 1406) {
if (refDCval < 94) {
if (MVDiff < 26) {
if (diffRsCsdiff < -53) {
return 1;
}
else if (diffRsCsdiff >= -53) {
if (MVDiff < 17) {
if (negBalance < 54) {
if (tcor < 81) {
if (Cs < 2259) {
return 4;
}
else if (Cs >= 2259) {
return 8;
}
}
else if (tcor >= 81) {
return 8;
}
}
else if (negBalance >= 54) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 8;
}
}
}
else if (MVDiff >= 17) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
return 4;
}
}
}
}
else if (MVDiff >= 26) {
if (Cs < 1666) {
if (tcor < 65) {
return 4;
}
else if (tcor >= 65) {
return 2;
}
}
else if (Cs >= 1666) {
if (ssDCval < 59) {
return 2;
}
else if (ssDCval >= 59) {
if (Rs < 739) {
return 8;
}
else if (Rs >= 739) {
return 4;
}
}
}
}
}
else if (refDCval >= 94) {
if (mu_mv_mag_sq < 34) {
if (MVDiff < 21) {
if (ssDCval < 118) {
if (Cs < 1834) {
return 8;
}
else if (Cs >= 1834) {
return 2;
}
}
else if (ssDCval >= 118) {
return 2;
}
}
else if (MVDiff >= 21) {
if (MVDiff < 36) {
if (Rs < 1195) {
return 1;
}
else if (Rs >= 1195) {
return 8;
}
}
else if (MVDiff >= 36) {
if (Cs < 1909) {
return 4;
}
else if (Cs >= 1909) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (Rs < 935) {
return 1;
}
else if (Rs >= 935) {
return 8;
}
}
}
}
else if (Rs >= 1406) {
if (AFD < 17) {
if (Rs < 3019) {
if (ssDCval < 52) {
if (gchDC < 1) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
return 1;
}
}
else if (gchDC >= 1) {
if (posBalance < 60) {
if (Cs < 1608) {
return 8;
}
else if (Cs >= 1608) {
return 4;
}
}
else if (posBalance >= 60) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
return 4;
}
}
}
}
else if (ssDCval >= 52) {
if (Cs < 2667) {
if (posBalance < 48) {
if (ssDCval < 93) {
return 1;
}
else if (ssDCval >= 93) {
if (Rs < 1742) {
return 2;
}
else if (Rs >= 1742) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (tcor < 80) {
if (diffAFD < -2) {
if (Rs < 2045) {
return 1;
}
else if (Rs >= 2045) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (tcor >= 80) {
if (negBalance < 57) {
if (TSC < 12) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
if (Rs < 2879) {
if (RsDiff < 7) {
if (negBalance < 52) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (negBalance >= 52) {
if (Rs < 1820) {
return 8;
}
else if (Rs >= 1820) {
return 1;
}
}
}
else if (RsDiff >= 7) {
return 8;
}
}
else if (Rs >= 2879) {
if (Cs < 2321) {
return 4;
}
else if (Cs >= 2321) {
return 8;
}
}
}
}
else if (TSC >= 12) {
if (Rs < 1940) {
if (negBalance < 55) {
return 4;
}
else if (negBalance >= 55) {
return 8;
}
}
else if (Rs >= 1940) {
return 8;
}
}
}
else if (negBalance >= 57) {
return 4;
}
}
}
}
else if (Cs >= 2667) {
if (mcTcor < 92) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 2;
}
}
}
else if (mcTcor >= 92) {
if (refDCval < 78) {
return 2;
}
else if (refDCval >= 78) {
if (Rs < 2438) {
return 1;
}
else if (Rs >= 2438) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3019) {
if (MVDiff < 9) {
if (tcor < 89) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
return 1;
}
}
else if (tcor >= 89) {
if (diffTSC < -1) {
if (Cs < 2661) {
return 1;
}
else if (Cs >= 2661) {
return 4;
}
}
else if (diffTSC >= -1) {
if (Cs < 2765) {
return 4;
}
else if (Cs >= 2765) {
return 8;
}
}
}
}
else if (MVDiff >= 9) {
if (Cs < 2594) {
return 8;
}
else if (Cs >= 2594) {
return 4;
}
}
}
}
else if (AFD >= 17) {
if (refDCval < 94) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
if (refDCval < 78) {
if (posBalance < 62) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
if (ssDCval < 74) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (Rs < 1526) {
return 1;
}
else if (Rs >= 1526) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (Rs < 1603) {
return 2;
}
else if (Rs >= 1603) {
return 1;
}
}
}
}
else if (posBalance >= 62) {
if (Cs < 1823) {
return 8;
}
else if (Cs >= 1823) {
return 4;
}
}
}
else if (refDCval >= 78) {
if (tcor < 79) {
if (Rs < 1908) {
if (Rs < 1774) {
return 4;
}
else if (Rs >= 1774) {
return 8;
}
}
else if (Rs >= 1908) {
return 4;
}
}
else if (tcor >= 79) {
return 8;
}
}
}
}
else if (refDCval >= 94) {
if (AFD < 23) {
if (ssDCval < 147) {
if (refDCval < 95) {
if (Cs < 2320) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (Cs >= 2320) {
if (Cs < 2673) {
return 8;
}
else if (Cs >= 2673) {
return 4;
}
}
}
else if (refDCval >= 95) {
if (MVDiff < 9) {
if (diffRsCsdiff < 8) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
else if (diffRsCsdiff >= 8) {
return 8;
}
}
else if (MVDiff >= 9) {
if (SC < 4869) {
return 8;
}
else if (SC >= 4869) {
return 4;
}
}
}
}
else if (ssDCval >= 147) {
return 1;
}
}
else if (AFD >= 23) {
if (Rs < 1902) {
return 2;
}
else if (Rs >= 1902) {
return 8;
}
}
}
}
}
}
else if (Cs >= 2795) {
if (Rs < 1435) {
if (ssDCval < 66) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (ssDCval >= 66) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Rs < 736) {
return 2;
}
else if (Rs >= 736) {
if (Rs < 1053) {
return 8;
}
else if (Rs >= 1053) {
return 4;
}
}
}
}
}
else if (Rs >= 1435) {
if (negBalance < 48) {
if (refDCval < 129) {
if (mu_mv_mag_sq < 103) {
if (Cs < 2898) {
if (Cs < 2876) {
return 4;
}
else if (Cs >= 2876) {
return 8;
}
}
else if (Cs >= 2898) {
return 4;
}
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 2787) {
return 4;
}
else if (Rs >= 2787) {
return 8;
}
}
}
else if (refDCval >= 129) {
return 1;
}
}
else if (negBalance >= 48) {
if (refDCval < 90) {
if (TSC < 11) {
if (Rs < 2661) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (Rs >= 2661) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (refDCval >= 90) {
if (TSC < 10) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
else if (TSC >= 10) {
if (AFD < 20) {
if (Cs < 3510) {
if (Rs < 2198) {
return 8;
}
else if (Rs >= 2198) {
return 4;
}
}
else if (Cs >= 3510) {
if (Rs < 2541) {
return 4;
}
else if (Rs >= 2541) {
return 8;
}
}
}
else if (AFD >= 20) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (TSC < 6) {
if (Cs < 2165) {
if (ssDCval < 99) {
if (Cs < 1378) {
return 8;
}
else if (Cs >= 1378) {
return 1;
}
}
else if (ssDCval >= 99) {
if (Rs < 1310) {
return 2;
}
else if (Rs >= 1310) {
return 8;
}
}
}
else if (Cs >= 2165) {
if (ssDCval < 69) {
if (Rs < 2396) {
return 1;
}
else if (Rs >= 2396) {
return 4;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
}
else if (TSC >= 6) {
if (Rs < 1921) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 107) {
if (refDCval < 101) {
if (ssDCval < 87) {
return 2;
}
else if (ssDCval >= 87) {
if (Rs < 1588) {
return 4;
}
else if (Rs >= 1588) {
return 2;
}
}
}
else if (refDCval >= 101) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (Rs < 1220) {
return 1;
}
else if (Rs >= 1220) {
return 2;
}
}
}
}
else if (ssDCval >= 107) {
if (AFD < 15) {
if (Rs < 1616) {
return 4;
}
else if (Rs >= 1616) {
return 2;
}
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (diffMVdiffVal < -3) {
if (ssDCval < 73) {
if (diffRsCsdiff < 6) {
if (Rs < 1147) {
return 2;
}
else if (Rs >= 1147) {
return 1;
}
}
else if (diffRsCsdiff >= 6) {
return 8;
}
}
else if (ssDCval >= 73) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
if (AFD < 15) {
if (Rs < 1389) {
return 4;
}
else if (Rs >= 1389) {
return 8;
}
}
else if (AFD >= 15) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= -3) {
if (mcTcor < 96) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (MVDiff < 11) {
if (Cs < 1430) {
return 2;
}
else if (Cs >= 1430) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
if (Rs < 1601) {
return 1;
}
else if (Rs >= 1601) {
if (Rs < 1871) {
return 8;
}
else if (Rs >= 1871) {
return 2;
}
}
}
}
}
else if (MVDiff >= 11) {
if (RsDiff < 6) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
if (negBalance < 54) {
if (Cs < 1468) {
return 1;
}
else if (Cs >= 1468) {
return 4;
}
}
else if (negBalance >= 54) {
return 8;
}
}
}
else if (RsDiff >= 6) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (tcor < 84) {
if (Cs < 2868) {
return 1;
}
else if (Cs >= 2868) {
return 2;
}
}
else if (tcor >= 84) {
if (Rs < 1452) {
if (Rs < 935) {
return 8;
}
else if (Rs >= 935) {
return 4;
}
}
else if (Rs >= 1452) {
if (Cs < 1690) {
return 2;
}
else if (Cs >= 1690) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1921) {
if (tcor < 92) {
if (MVDiff < 6) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
return 1;
}
}
else if (MVDiff >= 6) {
if (Rs < 2153) {
if (Cs < 2043) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
else if (Cs >= 2043) {
return 1;
}
}
else if (Rs >= 2153) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 1;
}
}
}
}
else if (tcor >= 92) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (Rs < 3342) {
return 8;
}
else if (Rs >= 3342) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 3843) {
if (AFD < 12) {
if (MVDiff < 9) {
if (ssDCval < 97) {
if (Cs < 2336) {
return 1;
}
else if (Cs >= 2336) {
if (Cs < 3130) {
return 4;
}
else if (Cs >= 3130) {
if (Rs < 5162) {
return 8;
}
else if (Rs >= 5162) {
return 2;
}
}
}
}
else if (ssDCval >= 97) {
if (mu_mv_mag_sq < 92) {
if (CsDiff < 9) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 5181) {
if (diffRsCsdiff < -25) {
return 1;
}
else if (diffRsCsdiff >= -25) {
return 8;
}
}
else if (Rs >= 5181) {
return 1;
}
}
}
else if (CsDiff >= 9) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (posBalance < 56) {
if (Cs < 2443) {
return 1;
}
else if (Cs >= 2443) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 1;
}
}
}
else if (posBalance >= 56) {
return 1;
}
}
}
}
else if (MVDiff >= 9) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 83) {
return 8;
}
else if (mu_mv_mag_sq >= 83) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
if (mcTcor < 92) {
if (RsDiff < 16) {
if (refDCval < 118) {
if (SC < 6573) {
if (refDCval < 109) {
if (negBalance < 51) {
if (posBalance < 48) {
if (mcTcor < 85) {
if (Cs < 2039) {
if (SC < 3404) {
if (Rs < 1535) {
return 4;
}
else if (Rs >= 1535) {
return 8;
}
}
else if (SC >= 3404) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 1;
}
}
}
else if (Cs >= 2039) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (mcTcor >= 85) {
if (Rs < 2949) {
if (MVDiff < 16) {
if (Cs < 1400) {
return 2;
}
else if (Cs >= 1400) {
return 1;
}
}
else if (MVDiff >= 16) {
if (diffRsCsdiff < -88) {
return 4;
}
else if (diffRsCsdiff >= -88) {
return 2;
}
}
}
else if (Rs >= 2949) {
if (refDCval < 87) {
if (mu_mv_mag_sq < 68) {
if (RsCsDiff < 153) {
return 4;
}
else if (RsCsDiff >= 153) {
if (diffRsCsdiff < 53) {
return 2;
}
else if (diffRsCsdiff >= 53) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (refDCval >= 87) {
if (Cs < 2985) {
return 8;
}
else if (Cs >= 2985) {
return 2;
}
}
}
}
}
else if (posBalance >= 48) {
if (CsDiff < 8) {
if (Cs < 1827) {
if (mcTcor < 87) {
if (refDCval < 93) {
if (Cs < 1765) {
return 8;
}
else if (Cs >= 1765) {
return 4;
}
}
else if (refDCval >= 93) {
if (Cs < 1457) {
return 4;
}
else if (Cs >= 1457) {
if (Cs < 1540) {
return 2;
}
else if (Cs >= 1540) {
return 8;
}
}
}
}
else if (mcTcor >= 87) {
if (Cs < 1396) {
if (AFD < 17) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
return 2;
}
}
else if (AFD >= 17) {
return 8;
}
}
else if (Cs >= 1396) {
if (RsDiff < 12) {
if (TSC < 9) {
if (Rs < 2845) {
return 2;
}
else if (Rs >= 2845) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (RsDiff >= 12) {
if (Rs < 3441) {
return 4;
}
else if (Rs >= 3441) {
return 2;
}
}
}
}
}
else if (Cs >= 1827) {
if (SC < 5247) {
if (MVDiff < 12) {
if (Cs < 1955) {
return 8;
}
else if (Cs >= 1955) {
return 4;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (SC >= 5247) {
if (Cs < 2586) {
return 8;
}
else if (Cs >= 2586) {
return 2;
}
}
}
}
else if (CsDiff >= 8) {
if (Rs < 1793) {
if (mcTcor < 78) {
if (Cs < 1615) {
return 8;
}
else if (Cs >= 1615) {
if (SC < 4009) {
return 4;
}
else if (SC >= 4009) {
return 8;
}
}
}
else if (mcTcor >= 78) {
if (diffAFD < -1) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
return 8;
}
}
else if (diffAFD >= -1) {
if (AFD < 20) {
if (CsDiff < 9) {
return 1;
}
else if (CsDiff >= 9) {
return 8;
}
}
else if (AFD >= 20) {
if (CsDiff < 11) {
return 4;
}
else if (CsDiff >= 11) {
return 2;
}
}
}
}
}
else if (Rs >= 1793) {
if (AFD < 12) {
if (Cs < 2206) {
if (mu_mv_mag_sq < 72) {
if (Cs < 1799) {
return 8;
}
else if (Cs >= 1799) {
if (Rs < 3323) {
return 4;
}
else if (Rs >= 3323) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 1;
}
}
}
else if (Cs >= 2206) {
if (Rs < 3894) {
return 8;
}
else if (Rs >= 3894) {
return 4;
}
}
}
else if (AFD >= 12) {
if (AFD < 18) {
if (refDCval < 79) {
if (CsDiff < 10) {
if (AFD < 17) {
if (SC < 4474) {
if (mcTcor < 87) {
if (Cs < 1932) {
return 8;
}
else if (Cs >= 1932) {
return 1;
}
}
else if (mcTcor >= 87) {
if (Rs < 2605) {
return 1;
}
else if (Rs >= 2605) {
return 4;
}
}
}
else if (SC >= 4474) {
return 8;
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (CsDiff >= 10) {
if (MVDiff < 18) {
if (gchDC < 1) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
if (Rs < 3257) {
return 8;
}
else if (Rs >= 3257) {
return 1;
}
}
}
else if (gchDC >= 1) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (diffRsCsdiff < 11) {
if (Cs < 2168) {
return 4;
}
else if (Cs >= 2168) {
return 8;
}
}
else if (diffRsCsdiff >= 11) {
return 8;
}
}
}
}
}
else if (MVDiff >= 18) {
if (Cs < 1713) {
return 8;
}
else if (Cs >= 1713) {
return 4;
}
}
}
}
else if (refDCval >= 79) {
if (ssDCval < 82) {
if (Cs < 2734) {
if (Rs < 2838) {
return 8;
}
else if (Rs >= 2838) {
return 4;
}
}
else if (Cs >= 2734) {
return 2;
}
}
else if (ssDCval >= 82) {
if (Cs < 2491) {
if (MVDiff < 12) {
if (MVDiff < 9) {
if (refDCval < 83) {
return 4;
}
else if (refDCval >= 83) {
return 8;
}
}
else if (MVDiff >= 9) {
if (refDCval < 84) {
return 8;
}
else if (refDCval >= 84) {
if (Rs < 4078) {
return 2;
}
else if (Rs >= 4078) {
return 8;
}
}
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (Cs >= 2491) {
if (RsDiff < 11) {
if (Cs < 2945) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (Cs >= 2945) {
return 4;
}
}
else if (RsDiff >= 11) {
if (CsDiff < 10) {
if (Rs < 3531) {
return 2;
}
else if (Rs >= 3531) {
return 4;
}
}
else if (CsDiff >= 10) {
return 4;
}
}
}
}
}
}
else if (AFD >= 18) {
if (MVDiff < 24) {
if (RsCsDiff < 243) {
if (Cs < 2214) {
return 4;
}
else if (Cs >= 2214) {
return 1;
}
}
else if (RsCsDiff >= 243) {
if (negBalance < 47) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
else if (negBalance >= 47) {
if (ssDCval < 75) {
if (Rs < 2808) {
return 2;
}
else if (Rs >= 2808) {
return 4;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
}
}
else if (MVDiff >= 24) {
if (ssDCval < 65) {
if (Rs < 2259) {
return 1;
}
else if (Rs >= 2259) {
return 4;
}
}
else if (ssDCval >= 65) {
if (mcTcor < 74) {
if (Rs < 2015) {
return 1;
}
else if (Rs >= 2015) {
return 8;
}
}
else if (mcTcor >= 74) {
return 8;
}
}
}
}
}
}
}
}
}
else if (negBalance >= 51) {
if (tcor < 68) {
if (mcTcor < 71) {
if (Rs < 2061) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
if (Cs < 1824) {
return 8;
}
else if (Cs >= 1824) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
}
}
else if (Rs >= 2061) {
if (Cs < 2907) {
if (RsDiff < 12) {
if (Cs < 2052) {
return 8;
}
else if (Cs >= 2052) {
if (diffMVdiffVal < 5) {
return 4;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 2189) {
return 4;
}
else if (Rs >= 2189) {
return 2;
}
}
}
}
else if (RsDiff >= 12) {
return 8;
}
}
else if (Cs >= 2907) {
return 4;
}
}
}
else if (mcTcor >= 71) {
if (AFD < 20) {
if (refDCval < 65) {
if (Cs < 1554) {
return 8;
}
else if (Cs >= 1554) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
if (posBalance < 65) {
return 4;
}
else if (posBalance >= 65) {
return 2;
}
}
}
}
else if (refDCval >= 65) {
if (posBalance < 62) {
if (Rs < 1916) {
return 4;
}
else if (Rs >= 1916) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 4;
}
}
}
else if (posBalance >= 62) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (RsCsDiff < 172) {
if (Rs < 2212) {
return 1;
}
else if (Rs >= 2212) {
return 8;
}
}
else if (RsCsDiff >= 172) {
return 4;
}
}
}
}
}
else if (AFD >= 20) {
if (CsDiff < 19) {
if (RsCsDiff < 172) {
if (Cs < 1474) {
return 2;
}
else if (Cs >= 1474) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
else if (RsCsDiff >= 172) {
if (RsCsDiff < 401) {
if (CsDiff < 10) {
if (ssDCval < 97) {
if (mu_mv_mag_sq < 5) {
if (MVDiff < 20) {
if (diffMVdiffVal < 1) {
if (Rs < 2617) {
return 4;
}
else if (Rs >= 2617) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (MVDiff >= 20) {
return 4;
}
}
else if (mu_mv_mag_sq >= 5) {
if (tcor < 65) {
if (MVDiff < 16) {
if (Rs < 2390) {
return 4;
}
else if (Rs >= 2390) {
return 8;
}
}
else if (MVDiff >= 16) {
if (CsDiff < 8) {
if (Rs < 3397) {
return 1;
}
else if (Rs >= 3397) {
return 8;
}
}
else if (CsDiff >= 8) {
return 8;
}
}
}
else if (tcor >= 65) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 8;
}
}
}
}
else if (ssDCval >= 97) {
return 4;
}
}
else if (CsDiff >= 10) {
if (negBalance < 60) {
if (ssDCval < 77) {
if (AFD < 24) {
if (Rs < 2675) {
return 1;
}
else if (Rs >= 2675) {
return 2;
}
}
else if (AFD >= 24) {
if (MVDiff < 19) {
if (Rs < 1952) {
return 8;
}
else if (Rs >= 1952) {
return 2;
}
}
else if (MVDiff >= 19) {
return 4;
}
}
}
else if (ssDCval >= 77) {
if (ssDCval < 81) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
return 2;
}
}
else if (ssDCval >= 81) {
if (mu_mv_mag_sq < 16) {
if (negBalance < 53) {
return 8;
}
else if (negBalance >= 53) {
if (Cs < 3039) {
return 4;
}
else if (Cs >= 3039) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (MVDiff < 35) {
if (Rs < 2706) {
return 2;
}
else if (Rs >= 2706) {
return 4;
}
}
else if (MVDiff >= 35) {
return 8;
}
}
}
}
}
else if (negBalance >= 60) {
if (Rs < 2174) {
if (Cs < 1516) {
return 1;
}
else if (Cs >= 1516) {
return 4;
}
}
else if (Rs >= 2174) {
if (AFD < 26) {
if (Rs < 3144) {
return 8;
}
else if (Rs >= 3144) {
return 1;
}
}
else if (AFD >= 26) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 401) {
if (Rs < 2481) {
return 2;
}
else if (Rs >= 2481) {
return 8;
}
}
}
}
else if (CsDiff >= 19) {
return 8;
}
}
}
}
else if (tcor >= 68) {
if (ssDCval < 81) {
if (AFD < 19) {
if (diffAFD < 2) {
if (Rs < 2250) {
if (tcor < 72) {
if (Cs < 2229) {
return 1;
}
else if (Cs >= 2229) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (tcor >= 72) {
if (Rs < 2242) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (Rs >= 2242) {
return 1;
}
}
}
else if (Rs >= 2250) {
if (gchDC < 2) {
if (diffAFD < -2) {
if (Rs < 2863) {
return 4;
}
else if (Rs >= 2863) {
return 8;
}
}
else if (diffAFD >= -2) {
if (RsCsDiff < 302) {
return 8;
}
else if (RsCsDiff >= 302) {
if (Rs < 3497) {
return 8;
}
else if (Rs >= 3497) {
return 1;
}
}
}
}
else if (gchDC >= 2) {
if (Cs < 2357) {
if (mu_mv_mag_sq < 40) {
if (Rs < 2515) {
return 8;
}
else if (Rs >= 2515) {
return 4;
}
}
else if (mu_mv_mag_sq >= 40) {
return 8;
}
}
else if (Cs >= 2357) {
return 4;
}
}
}
}
else if (diffAFD >= 2) {
if (SC < 5603) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
return 8;
}
}
else if (SC >= 5603) {
return 4;
}
}
}
else if (AFD >= 19) {
if (TSC < 13) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
if (mu_mv_mag_sq < 14) {
if (Cs < 2550) {
if (RsDiff < 13) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (RsDiff >= 13) {
if (Rs < 2637) {
return 8;
}
else if (Rs >= 2637) {
return 2;
}
}
}
else if (Cs >= 2550) {
if (Cs < 3011) {
return 1;
}
else if (Cs >= 3011) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (Rs < 3178) {
return 8;
}
else if (Rs >= 3178) {
return 4;
}
}
}
}
else if (TSC >= 13) {
if (negBalance < 54) {
if (Rs < 3075) {
return 2;
}
else if (Rs >= 3075) {
return 1;
}
}
else if (negBalance >= 54) {
if (Rs < 3542) {
if (Rs < 2522) {
if (Cs < 2153) {
return 8;
}
else if (Cs >= 2153) {
return 1;
}
}
else if (Rs >= 2522) {
return 8;
}
}
else if (Rs >= 3542) {
if (Rs < 3864) {
return 4;
}
else if (Rs >= 3864) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (Cs < 2614) {
if (Rs < 2573) {
if (AFD < 22) {
if (mcTcor < 89) {
if (RsCsDiff < 193) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (diffMVdiffVal < -1) {
if (SC < 4355) {
return 8;
}
else if (SC >= 4355) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (RsCsDiff >= 193) {
return 8;
}
}
else if (mcTcor >= 89) {
if (AFD < 15) {
if (Rs < 1866) {
return 8;
}
else if (Rs >= 1866) {
return 2;
}
}
else if (AFD >= 15) {
return 8;
}
}
}
else if (AFD >= 22) {
if (Cs < 1750) {
if (Rs < 1738) {
return 4;
}
else if (Rs >= 1738) {
return 8;
}
}
else if (Cs >= 1750) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
if (Rs < 2020) {
return 8;
}
else if (Rs >= 2020) {
return 4;
}
}
}
}
}
else if (Rs >= 2573) {
if (refDCval < 93) {
if (CsDiff < 12) {
if (diffRsCsdiff < -4) {
if (Cs < 2569) {
return 4;
}
else if (Cs >= 2569) {
return 1;
}
}
else if (diffRsCsdiff >= -4) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (tcor < 81) {
if (AFD < 25) {
if (diffRsCsdiff < 70) {
return 4;
}
else if (diffRsCsdiff >= 70) {
return 8;
}
}
else if (AFD >= 25) {
return 8;
}
}
else if (tcor >= 81) {
return 8;
}
}
}
}
else if (CsDiff >= 12) {
if (Rs < 2597) {
return 1;
}
else if (Rs >= 2597) {
return 8;
}
}
}
else if (refDCval >= 93) {
if (ssDCval < 104) {
if (RsCsDiff < 284) {
if (Cs < 2270) {
if (AFD < 23) {
if (RsCsDiff < 242) {
return 8;
}
else if (RsCsDiff >= 242) {
if (Cs < 2139) {
return 8;
}
else if (Cs >= 2139) {
return 2;
}
}
}
else if (AFD >= 23) {
return 4;
}
}
else if (Cs >= 2270) {
if (CsDiff < 9) {
if (Rs < 3823) {
return 8;
}
else if (Rs >= 3823) {
return 4;
}
}
else if (CsDiff >= 9) {
if (Rs < 3976) {
return 2;
}
else if (Rs >= 3976) {
return 1;
}
}
}
}
else if (RsCsDiff >= 284) {
if (mu_mv_mag_sq < 7) {
return 4;
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
}
else if (ssDCval >= 104) {
if (tcor < 78) {
if (Cs < 1516) {
return 4;
}
else if (Cs >= 1516) {
return 8;
}
}
else if (tcor >= 78) {
return 4;
}
}
}
}
}
else if (Cs >= 2614) {
if (AFD < 23) {
if (Cs < 2754) {
if (diffRsCsdiff < 22) {
if (AFD < 22) {
if (Cs < 2710) {
return 4;
}
else if (Cs >= 2710) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
else if (AFD >= 22) {
return 1;
}
}
else if (diffRsCsdiff >= 22) {
if (Rs < 2686) {
return 8;
}
else if (Rs >= 2686) {
return 4;
}
}
}
else if (Cs >= 2754) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 4;
}
}
}
}
else if (AFD >= 23) {
if (Rs < 2678) {
return 8;
}
else if (Rs >= 2678) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (AFD < 20) {
if (diffRsCsdiff < 193) {
return 8;
}
else if (diffRsCsdiff >= 193) {
return 2;
}
}
else if (AFD >= 20) {
if (Cs < 1427) {
return 4;
}
else if (Cs >= 1427) {
if (negBalance < 60) {
if (refDCval < 110) {
if (Rs < 2330) {
return 4;
}
else if (Rs >= 2330) {
return 8;
}
}
else if (refDCval >= 110) {
return 8;
}
}
else if (negBalance >= 60) {
if (Cs < 2750) {
if (refDCval < 114) {
return 1;
}
else if (refDCval >= 114) {
if (Rs < 3165) {
return 8;
}
else if (Rs >= 3165) {
return 1;
}
}
}
else if (Cs >= 2750) {
if (Rs < 2841) {
return 2;
}
else if (Rs >= 2841) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 6573) {
if (posBalance < 48) {
if (refDCval < 74) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (refDCval >= 74) {
if (tcor < 80) {
return 8;
}
else if (tcor >= 80) {
if (mu_mv_mag_sq < 67) {
if (Rs < 5068) {
return 8;
}
else if (Rs >= 5068) {
return 1;
}
}
else if (mu_mv_mag_sq >= 67) {
if (Cs < 2790) {
return 4;
}
else if (Cs >= 2790) {
return 2;
}
}
}
}
}
else if (posBalance >= 48) {
if (AFD < 21) {
if (mcTcor < 90) {
if (Rs < 6297) {
if (Rs < 3373) {
if (mcTcor < 83) {
return 4;
}
else if (mcTcor >= 83) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (Rs >= 3373) {
if (Cs < 3174) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 63) {
if (ssDCval < 101) {
if (Rs < 3853) {
return 8;
}
else if (Rs >= 3853) {
return 4;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (mu_mv_mag_sq >= 63) {
if (posBalance < 60) {
if (ssDCval < 105) {
return 1;
}
else if (ssDCval >= 105) {
return 8;
}
}
else if (posBalance >= 60) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (Cs >= 3174) {
if (gchDC < 3) {
if (negBalance < 52) {
if (SC < 8242) {
if (ssDCval < 100) {
if (posBalance < 61) {
if (mcTcor < 84) {
if (Rs < 3838) {
return 2;
}
else if (Rs >= 3838) {
return 8;
}
}
else if (mcTcor >= 84) {
return 4;
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (ssDCval >= 100) {
return 8;
}
}
else if (SC >= 8242) {
return 8;
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (gchDC >= 3) {
if (Rs < 4664) {
if (Rs < 4090) {
return 1;
}
else if (Rs >= 4090) {
return 2;
}
}
else if (Rs >= 4664) {
return 8;
}
}
}
}
}
else if (Rs >= 6297) {
if (TSC < 13) {
if (Rs < 6438) {
return 2;
}
else if (Rs >= 6438) {
return 8;
}
}
else if (TSC >= 13) {
return 4;
}
}
}
else if (mcTcor >= 90) {
if (Rs < 4081) {
if (Cs < 3734) {
if (posBalance < 60) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 8;
}
}
else if (posBalance >= 60) {
if (Rs < 3352) {
return 4;
}
else if (Rs >= 3352) {
return 8;
}
}
}
else if (Cs >= 3734) {
return 2;
}
}
else if (Rs >= 4081) {
if (diffRsCsdiff < -81) {
if (MVDiff < 23) {
if (posBalance < 56) {
if (diffAFD < -2) {
if (RsDiff < 11) {
return 2;
}
else if (RsDiff >= 11) {
return 8;
}
}
else if (diffAFD >= -2) {
return 2;
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (diffRsCsdiff >= -81) {
if (mu_mv_mag_sq < 56) {
return 8;
}
else if (mu_mv_mag_sq >= 56) {
if (Cs < 3250) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 4;
}
}
else if (Cs >= 3250) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 21) {
if (refDCval < 102) {
if (refDCval < 90) {
return 8;
}
else if (refDCval >= 90) {
if (mu_mv_mag_sq < 2) {
if (tcor < 73) {
return 4;
}
else if (tcor >= 73) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 2502) {
return 4;
}
else if (Rs >= 2502) {
return 8;
}
}
}
}
else if (refDCval >= 102) {
if (Cs < 3767) {
if (Rs < 3803) {
return 8;
}
else if (Rs >= 3803) {
return 1;
}
}
else if (Cs >= 3767) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 118) {
if (ssDCval < 143) {
if (refDCval < 137) {
if (tcor < 77) {
if (negBalance < 35) {
if (posBalance < 49) {
if (Cs < 3404) {
return 8;
}
else if (Cs >= 3404) {
if (Rs < 2264) {
return 1;
}
else if (Rs >= 2264) {
return 4;
}
}
}
else if (posBalance >= 49) {
return 2;
}
}
else if (negBalance >= 35) {
if (AFD < 16) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 1;
}
}
else if (AFD >= 16) {
if (mu_mv_mag_sq < 7) {
if (diffRsCsdiff < 43) {
if (mcTcor < 88) {
if (MVDiff < 13) {
if (Rs < 3651) {
return 8;
}
else if (Rs >= 3651) {
return 4;
}
}
else if (MVDiff >= 13) {
if (Rs < 1791) {
return 8;
}
else if (Rs >= 1791) {
if (diffRsCsdiff < -42) {
return 8;
}
else if (diffRsCsdiff >= -42) {
return 1;
}
}
}
}
else if (mcTcor >= 88) {
return 1;
}
}
else if (diffRsCsdiff >= 43) {
if (Rs < 3599) {
if (Rs < 3044) {
return 2;
}
else if (Rs >= 3044) {
return 8;
}
}
else if (Rs >= 3599) {
if (ssDCval < 130) {
return 4;
}
else if (ssDCval >= 130) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (SC < 7781) {
if (mcTcor < 68) {
if (Rs < 1502) {
return 4;
}
else if (Rs >= 1502) {
return 1;
}
}
else if (mcTcor >= 68) {
if (diffAFD < 12) {
if (diffAFD < 1) {
if (mu_mv_mag_sq < 15) {
return 1;
}
else if (mu_mv_mag_sq >= 15) {
if (Cs < 1426) {
return 1;
}
else if (Cs >= 1426) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
return 8;
}
}
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (diffAFD >= 12) {
if (Cs < 2106) {
return 1;
}
else if (Cs >= 2106) {
return 8;
}
}
}
}
else if (SC >= 7781) {
return 1;
}
}
}
}
}
else if (tcor >= 77) {
if (ssDCval < 123) {
if (Cs < 1414) {
if (Rs < 2883) {
return 1;
}
else if (Rs >= 2883) {
return 8;
}
}
else if (Cs >= 1414) {
if (Rs < 1826) {
return 4;
}
else if (Rs >= 1826) {
return 8;
}
}
}
else if (ssDCval >= 123) {
if (Cs < 2596) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (diffMVdiffVal < -6) {
return 1;
}
else if (diffMVdiffVal >= -6) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
if (Rs < 2165) {
return 8;
}
else if (Rs >= 2165) {
return 1;
}
}
}
}
}
}
else if (Cs >= 2596) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
if (Rs < 3459) {
return 8;
}
else if (Rs >= 3459) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 137) {
if (TSC < 9) {
if (CsDiff < 9) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 8;
}
}
else if (CsDiff >= 9) {
return 8;
}
}
else if (TSC >= 9) {
if (RsDiff < 13) {
if (diffAFD < 3) {
if (ssDCval < 142) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
return 1;
}
}
else if (ssDCval >= 142) {
if (Rs < 3455) {
return 8;
}
else if (Rs >= 3455) {
return 1;
}
}
}
else if (diffAFD >= 3) {
return 8;
}
}
else if (RsDiff >= 13) {
if (Cs < 3254) {
if (AFD < 25) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
else if (AFD >= 25) {
if (Rs < 2467) {
if (Rs < 2125) {
return 4;
}
else if (Rs >= 2125) {
return 8;
}
}
else if (Rs >= 2467) {
return 2;
}
}
}
else if (Cs >= 3254) {
if (Rs < 5212) {
return 1;
}
else if (Rs >= 5212) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 143) {
if (Rs < 3410) {
if (AFD < 21) {
if (ssDCval < 182) {
return 8;
}
else if (ssDCval >= 182) {
return 2;
}
}
else if (AFD >= 21) {
if (RsCsDiff < 281) {
return 2;
}
else if (RsCsDiff >= 281) {
return 8;
}
}
}
else if (Rs >= 3410) {
if (ssDCval < 153) {
return 1;
}
else if (ssDCval >= 153) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 16) {
if (Cs < 4612) {
if (refDCval < 134) {
if (Cs < 1755) {
if (MVDiff < 47) {
if (mcTcor < 79) {
if (Rs < 3346) {
return 8;
}
else if (Rs >= 3346) {
return 2;
}
}
else if (mcTcor >= 79) {
if (CsDiff < 12) {
if (mu_mv_mag_sq < 48) {
if (ssDCval < 122) {
if (RsCsDiff < 381) {
if (refDCval < 101) {
return 2;
}
else if (refDCval >= 101) {
if (Rs < 3448) {
return 4;
}
else if (Rs >= 3448) {
if (Rs < 4462) {
return 8;
}
else if (Rs >= 4462) {
return 1;
}
}
}
}
else if (RsCsDiff >= 381) {
return 2;
}
}
else if (ssDCval >= 122) {
if (TSC < 13) {
if (Rs < 4579) {
return 1;
}
else if (Rs >= 4579) {
return 8;
}
}
else if (TSC >= 13) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (AFD < 32) {
if (MVDiff < 17) {
if (Rs < 4049) {
return 1;
}
else if (Rs >= 4049) {
return 2;
}
}
else if (MVDiff >= 17) {
return 4;
}
}
else if (AFD >= 32) {
return 2;
}
}
}
else if (CsDiff >= 12) {
if (Rs < 2317) {
return 4;
}
else if (Rs >= 2317) {
return 8;
}
}
}
}
else if (MVDiff >= 47) {
if (CsDiff < 7) {
if (Rs < 5081) {
if (ssDCval < 107) {
if (Rs < 3045) {
return 2;
}
else if (Rs >= 3045) {
return 4;
}
}
else if (ssDCval >= 107) {
return 2;
}
}
else if (Rs >= 5081) {
return 8;
}
}
else if (CsDiff >= 7) {
if (ssDCval < 115) {
if (posBalance < 56) {
if (RsCsDiff < 395) {
if (Rs < 3235) {
return 2;
}
else if (Rs >= 3235) {
return 1;
}
}
else if (RsCsDiff >= 395) {
if (AFD < 39) {
return 8;
}
else if (AFD >= 39) {
return 1;
}
}
}
else if (posBalance >= 56) {
if (mu_mv_mag_sq < 6) {
return 2;
}
else if (mu_mv_mag_sq >= 6) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
return 2;
}
}
}
}
else if (ssDCval >= 115) {
return 1;
}
}
}
}
else if (Cs >= 1755) {
if (TSC < 15) {
if (tcor < 79) {
if (mcTcor < 83) {
if (refDCval < 102) {
if (mu_mv_mag_sq < 37) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
if (MVDiff < 56) {
return 4;
}
else if (MVDiff >= 56) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 3001) {
return 1;
}
else if (Rs >= 3001) {
return 8;
}
}
}
else if (refDCval >= 102) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (mcTcor >= 83) {
if (mu_mv_mag_sq < 7) {
if (RsDiff < 20) {
if (mu_mv_mag_sq < 5) {
if (RsDiff < 17) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Rs < 5545) {
return 4;
}
else if (Rs >= 5545) {
return 8;
}
}
}
else if (RsDiff >= 17) {
if (mcTcor < 90) {
if (Cs < 1968) {
return 2;
}
else if (Cs >= 1968) {
return 4;
}
}
else if (mcTcor >= 90) {
if (Rs < 4453) {
return 2;
}
else if (Rs >= 4453) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
return 2;
}
}
else if (RsDiff >= 20) {
if (Cs < 3637) {
return 2;
}
else if (Cs >= 3637) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 30) {
if (CsDiff < 15) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
if (Cs < 2205) {
if (ssDCval < 125) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
if (Rs < 2828) {
return 2;
}
else if (Rs >= 2828) {
return 8;
}
}
}
else if (ssDCval >= 125) {
if (Rs < 4522) {
return 1;
}
else if (Rs >= 4522) {
return 4;
}
}
}
else if (Cs >= 2205) {
if (mu_mv_mag_sq < 46) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
return 4;
}
}
else if (mu_mv_mag_sq >= 46) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
else if (CsDiff >= 15) {
if (RsDiff < 26) {
return 4;
}
else if (RsDiff >= 26) {
return 8;
}
}
}
else if (AFD >= 30) {
if (gchDC < 4) {
if (Cs < 2691) {
return 1;
}
else if (Cs >= 2691) {
if (Cs < 2840) {
return 2;
}
else if (Cs >= 2840) {
return 4;
}
}
}
else if (gchDC >= 4) {
return 2;
}
}
}
}
}
else if (tcor >= 79) {
if (ssDCval < 77) {
if (posBalance < 48) {
if (ssDCval < 70) {
return 4;
}
else if (ssDCval >= 70) {
if (Cs < 3610) {
return 4;
}
else if (Cs >= 3610) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (posBalance < 57) {
if (diffAFD < -8) {
return 2;
}
else if (diffAFD >= -8) {
return 8;
}
}
else if (posBalance >= 57) {
return 2;
}
}
}
else if (ssDCval >= 77) {
if (negBalance < 46) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 1;
}
}
else if (negBalance >= 46) {
if (negBalance < 52) {
if (mcTcor < 91) {
if (Cs < 1900) {
return 2;
}
else if (Cs >= 1900) {
return 8;
}
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (negBalance >= 52) {
if (mu_mv_mag_sq < 51) {
return 8;
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 5663) {
return 4;
}
else if (Rs >= 5663) {
return 2;
}
}
}
}
}
}
}
else if (TSC >= 15) {
if (diffAFD < 3) {
if (negBalance < 49) {
if (Rs < 3683) {
if (Rs < 2900) {
return 1;
}
else if (Rs >= 2900) {
return 8;
}
}
else if (Rs >= 3683) {
if (ssDCval < 87) {
if (Rs < 7447) {
return 2;
}
else if (Rs >= 7447) {
return 1;
}
}
else if (ssDCval >= 87) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 1;
}
}
}
}
else if (negBalance >= 49) {
if (AFD < 26) {
if (Rs < 5327) {
return 8;
}
else if (Rs >= 5327) {
if (negBalance < 52) {
if (TSC < 16) {
return 4;
}
else if (TSC >= 16) {
return 2;
}
}
else if (negBalance >= 52) {
if (Rs < 5802) {
return 8;
}
else if (Rs >= 5802) {
return 4;
}
}
}
}
else if (AFD >= 26) {
if (refDCval < 109) {
if (Rs < 4062) {
if (Cs < 2217) {
if (Cs < 1926) {
return 1;
}
else if (Cs >= 1926) {
return 2;
}
}
else if (Cs >= 2217) {
return 1;
}
}
else if (Rs >= 4062) {
if (posBalance < 61) {
if (Rs < 4094) {
return 8;
}
else if (Rs >= 4094) {
return 2;
}
}
else if (posBalance >= 61) {
if (Rs < 5734) {
return 1;
}
else if (Rs >= 5734) {
return 4;
}
}
}
}
else if (refDCval >= 109) {
if (TSC < 18) {
if (Rs < 4104) {
if (Rs < 3576) {
return 8;
}
else if (Rs >= 3576) {
return 4;
}
}
else if (Rs >= 4104) {
return 2;
}
}
else if (TSC >= 18) {
return 8;
}
}
}
}
}
else if (diffAFD >= 3) {
if (RsDiff < 26) {
if (Cs < 2638) {
if (AFD < 49) {
if (diffAFD < 6) {
if (Rs < 3633) {
return 4;
}
else if (Rs >= 3633) {
return 8;
}
}
else if (diffAFD >= 6) {
return 8;
}
}
else if (AFD >= 49) {
return 2;
}
}
else if (Cs >= 2638) {
if (mu_mv_mag_sq < 56) {
if (Cs < 3110) {
if (mu_mv_mag_sq < 25) {
return 8;
}
else if (mu_mv_mag_sq >= 25) {
return 4;
}
}
else if (Cs >= 3110) {
if (Rs < 5324) {
return 4;
}
else if (Rs >= 5324) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (SC < 6709) {
return 1;
}
else if (SC >= 6709) {
return 8;
}
}
}
}
else if (RsDiff >= 26) {
if (CsDiff < 17) {
if (Rs < 4377) {
return 1;
}
else if (Rs >= 4377) {
return 2;
}
}
else if (CsDiff >= 17) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 134) {
if (Rs < 3923) {
if (CsDiff < 9) {
if (TSC < 10) {
if (Rs < 3513) {
return 4;
}
else if (Rs >= 3513) {
return 8;
}
}
else if (TSC >= 10) {
return 2;
}
}
else if (CsDiff >= 9) {
if (Rs < 3601) {
return 8;
}
else if (Rs >= 3601) {
return 1;
}
}
}
else if (Rs >= 3923) {
if (ssDCval < 168) {
if (posBalance < 57) {
if (negBalance < 53) {
if (diffAFD < -3) {
if (Rs < 5242) {
return 8;
}
else if (Rs >= 5242) {
if (Rs < 7602) {
return 2;
}
else if (Rs >= 7602) {
return 4;
}
}
}
else if (diffAFD >= -3) {
if (ssDCval < 138) {
if (Rs < 4890) {
return 4;
}
else if (Rs >= 4890) {
return 8;
}
}
else if (ssDCval >= 138) {
return 4;
}
}
}
else if (negBalance >= 53) {
if (mcTcor < 84) {
if (Cs < 2229) {
return 8;
}
else if (Cs >= 2229) {
return 4;
}
}
else if (mcTcor >= 84) {
if (Cs < 1364) {
return 1;
}
else if (Cs >= 1364) {
return 8;
}
}
}
}
else if (posBalance >= 57) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (AFD < 34) {
return 4;
}
else if (AFD >= 34) {
if (tcor < 63) {
return 4;
}
else if (tcor >= 63) {
return 2;
}
}
}
}
}
else if (ssDCval >= 168) {
if (Rs < 6448) {
return 8;
}
else if (Rs >= 6448) {
return 2;
}
}
}
}
}
else if (Cs >= 4612) {
if (mu_mv_mag_sq < 69) {
if (posBalance < 53) {
if (RsDiff < 18) {
if (Rs < 6238) {
return 4;
}
else if (Rs >= 6238) {
return 1;
}
}
else if (RsDiff >= 18) {
return 8;
}
}
else if (posBalance >= 53) {
if (TSC < 22) {
if (TSC < 12) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 4;
}
}
else if (TSC >= 12) {
if (AFD < 40) {
return 4;
}
else if (AFD >= 40) {
return 8;
}
}
}
else if (TSC >= 22) {
if (Rs < 8495) {
return 8;
}
else if (Rs >= 8495) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (posBalance < 60) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (posBalance >= 60) {
if (Rs < 8673) {
return 8;
}
else if (Rs >= 8673) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 103) {
if (CsDiff < 8) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Cs < 2209) {
if (Rs < 2928) {
if (TSC < 9) {
if (Cs < 1554) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
if (Rs < 2355) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
else if (Rs >= 2355) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
return 4;
}
}
}
}
else if (Cs >= 1554) {
if (SC < 4361) {
return 4;
}
else if (SC >= 4361) {
if (Rs < 2746) {
return 8;
}
else if (Rs >= 2746) {
return 2;
}
}
}
}
else if (TSC >= 9) {
if (ssDCval < 89) {
if (Cs < 1462) {
if (Rs < 2355) {
return 1;
}
else if (Rs >= 2355) {
return 2;
}
}
else if (Cs >= 1462) {
if (diffRsCsdiff < -14) {
if (Rs < 2036) {
return 2;
}
else if (Rs >= 2036) {
return 4;
}
}
else if (diffRsCsdiff >= -14) {
if (Cs < 1494) {
return 4;
}
else if (Cs >= 1494) {
return 8;
}
}
}
}
else if (ssDCval >= 89) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
}
}
else if (Rs >= 2928) {
if (posBalance < 46) {
if (refDCval < 84) {
if (diffRsCsdiff < 18) {
if (Cs < 1852) {
if (diffRsCsdiff < -75) {
if (Rs < 3575) {
return 4;
}
else if (Rs >= 3575) {
return 2;
}
}
else if (diffRsCsdiff >= -75) {
return 2;
}
}
else if (Cs >= 1852) {
if (Rs < 4730) {
return 1;
}
else if (Rs >= 4730) {
return 4;
}
}
}
else if (diffRsCsdiff >= 18) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (refDCval >= 84) {
if (RsDiff < 10) {
if (Rs < 4210) {
if (Rs < 3126) {
return 8;
}
else if (Rs >= 3126) {
return 2;
}
}
else if (Rs >= 4210) {
return 8;
}
}
else if (RsDiff >= 10) {
return 2;
}
}
}
else if (posBalance >= 46) {
if (ssDCval < 100) {
if (Cs < 2029) {
if (negBalance < 48) {
if (TSC < 6) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
return 2;
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (negBalance >= 48) {
if (RsDiff < 11) {
if (Cs < 1435) {
return 2;
}
else if (Cs >= 1435) {
return 8;
}
}
else if (RsDiff >= 11) {
if (Rs < 5571) {
return 2;
}
else if (Rs >= 5571) {
return 4;
}
}
}
}
else if (Cs >= 2029) {
if (Rs < 4385) {
if (tcor < 91) {
if (Rs < 3636) {
if (Rs < 3432) {
return 4;
}
else if (Rs >= 3432) {
return 2;
}
}
else if (Rs >= 3636) {
if (Cs < 2035) {
return 4;
}
else if (Cs >= 2035) {
return 8;
}
}
}
else if (tcor >= 91) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
}
else if (Rs >= 4385) {
return 2;
}
}
}
else if (ssDCval >= 100) {
if (RsDiff < 14) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (Rs < 4565) {
return 2;
}
else if (Rs >= 4565) {
return 1;
}
}
}
else if (RsDiff >= 14) {
return 2;
}
}
}
}
}
else if (Cs >= 2209) {
if (negBalance < 41) {
if (ssDCval < 72) {
if (Rs < 3664) {
return 1;
}
else if (Rs >= 3664) {
return 4;
}
}
else if (ssDCval >= 72) {
if (Rs < 4938) {
if (Cs < 2916) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 2;
}
}
else if (Cs >= 2916) {
return 4;
}
}
else if (Rs >= 4938) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (diffRsCsdiff < 40) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (diffRsCsdiff >= 40) {
if (Rs < 5041) {
return 4;
}
else if (Rs >= 5041) {
return 8;
}
}
}
}
}
}
else if (CsDiff >= 8) {
if (negBalance < 46) {
if (refDCval < 72) {
if (Rs < 3500) {
if (Rs < 2971) {
if (Rs < 2079) {
return 8;
}
else if (Rs >= 2079) {
return 2;
}
}
else if (Rs >= 2971) {
if (Cs < 1712) {
return 8;
}
else if (Cs >= 1712) {
return 1;
}
}
}
else if (Rs >= 3500) {
if (SC < 8085) {
if (Rs < 3748) {
return 8;
}
else if (Rs >= 3748) {
return 4;
}
}
else if (SC >= 8085) {
return 8;
}
}
}
else if (refDCval >= 72) {
if (RsDiff < 13) {
if (AFD < 11) {
if (negBalance < 36) {
if (Cs < 2996) {
if (Rs < 3987) {
return 8;
}
else if (Rs >= 3987) {
return 2;
}
}
else if (Cs >= 2996) {
return 2;
}
}
else if (negBalance >= 36) {
if (Cs < 2274) {
return 2;
}
else if (Cs >= 2274) {
if (Rs < 5264) {
return 4;
}
else if (Rs >= 5264) {
return 2;
}
}
}
}
else if (AFD >= 11) {
if (AFD < 19) {
if (posBalance < 48) {
if (Rs < 2525) {
return 8;
}
else if (Rs >= 2525) {
return 2;
}
}
else if (posBalance >= 48) {
if (diffTSC < 2) {
if (negBalance < 37) {
return 1;
}
else if (negBalance >= 37) {
if (AFD < 18) {
if (Rs < 2942) {
return 8;
}
else if (Rs >= 2942) {
if (ssDCval < 90) {
if (Rs < 3424) {
return 4;
}
else if (Rs >= 3424) {
return 2;
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (AFD >= 19) {
if (TSC < 11) {
if (Cs < 2242) {
return 2;
}
else if (Cs >= 2242) {
return 8;
}
}
else if (TSC >= 11) {
return 8;
}
}
}
}
else if (RsDiff >= 13) {
if (CsDiff < 11) {
return 2;
}
else if (CsDiff >= 11) {
if (mu_mv_mag_sq < 51) {
if (mu_mv_mag_sq < 5) {
if (Rs < 2430) {
return 4;
}
else if (Rs >= 2430) {
if (Rs < 3379) {
return 8;
}
else if (Rs >= 3379) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 5) {
return 2;
}
}
else if (mu_mv_mag_sq >= 51) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (mu_mv_mag_sq < 5) {
if (RsDiff < 15) {
if (MVDiff < 7) {
if (ssDCval < 87) {
if (Cs < 2491) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 4;
}
}
else if (Cs >= 2491) {
if (Rs < 3181) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 1;
}
}
else if (Rs >= 3181) {
return 2;
}
}
}
else if (ssDCval >= 87) {
if (RsCsDiff < 173) {
if (Rs < 2585) {
return 2;
}
else if (Rs >= 2585) {
return 8;
}
}
else if (RsCsDiff >= 173) {
if (tcor < 81) {
if (CsDiff < 11) {
if (Rs < 3488) {
return 4;
}
else if (Rs >= 3488) {
return 8;
}
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (tcor >= 81) {
return 4;
}
}
}
}
else if (MVDiff >= 7) {
if (gchDC < 3) {
if (AFD < 21) {
if (posBalance < 58) {
if (negBalance < 58) {
if (mcTcor < 95) {
if (tcor < 86) {
if (Cs < 2459) {
return 4;
}
else if (Cs >= 2459) {
if (Cs < 2969) {
return 8;
}
else if (Cs >= 2969) {
return 4;
}
}
}
else if (tcor >= 86) {
if (Rs < 2779) {
return 4;
}
else if (Rs >= 2779) {
return 2;
}
}
}
else if (mcTcor >= 95) {
return 8;
}
}
else if (negBalance >= 58) {
if (Rs < 4304) {
return 2;
}
else if (Rs >= 4304) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
if (negBalance < 50) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
}
else if (AFD >= 21) {
if (RsCsDiff < 247) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
return 8;
}
}
else if (RsCsDiff >= 247) {
if (Rs < 1717) {
return 1;
}
else if (Rs >= 1717) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
if (Cs < 1775) {
if (Cs < 1481) {
return 1;
}
else if (Cs >= 1481) {
return 2;
}
}
else if (Cs >= 1775) {
return 4;
}
}
}
}
else if (RsDiff >= 15) {
if (Rs < 3432) {
if (Cs < 1616) {
return 1;
}
else if (Cs >= 1616) {
return 8;
}
}
else if (Rs >= 3432) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (mu_mv_mag_sq < 64) {
if (AFD < 22) {
if (diffRsCsdiff < 62) {
if (SC < 4640) {
return 8;
}
else if (SC >= 4640) {
if (Cs < 2191) {
if (RsCsDiff < 243) {
return 2;
}
else if (RsCsDiff >= 243) {
if (Rs < 3606) {
return 8;
}
else if (Rs >= 3606) {
return 1;
}
}
}
else if (Cs >= 2191) {
if (Rs < 2464) {
return 2;
}
else if (Rs >= 2464) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 62) {
if (MVDiff < 17) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (AFD >= 22) {
if (AFD < 30) {
if (TSC < 12) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (diffRsCsdiff < 85) {
if (diffMVdiffVal < -5) {
if (Cs < 1530) {
return 2;
}
else if (Cs >= 1530) {
return 4;
}
}
else if (diffMVdiffVal >= -5) {
return 4;
}
}
else if (diffRsCsdiff >= 85) {
if (Rs < 1707) {
return 8;
}
else if (Rs >= 1707) {
return 2;
}
}
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (AFD >= 30) {
if (Cs < 2481) {
return 1;
}
else if (Cs >= 2481) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (ssDCval < 84) {
if (ssDCval < 56) {
if (Rs < 3188) {
return 8;
}
else if (Rs >= 3188) {
return 4;
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (ssDCval >= 84) {
if (Cs < 2493) {
if (Rs < 2494) {
return 2;
}
else if (Rs >= 2494) {
if (Rs < 3625) {
return 8;
}
else if (Rs >= 3625) {
return 1;
}
}
}
else if (Cs >= 2493) {
if (RsCsDiff < 154) {
return 8;
}
else if (RsCsDiff >= 154) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 103) {
if (RsDiff < 14) {
if (posBalance < 51) {
if (tcor < 94) {
if (TSC < 6) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (Rs < 7236) {
return 1;
}
else if (Rs >= 7236) {
return 8;
}
}
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 48) {
if (negBalance < 51) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Cs < 10867) {
return 2;
}
else if (Cs >= 10867) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (refDCval < 127) {
if (AFD < 16) {
if (Rs < 4968) {
return 2;
}
else if (Rs >= 4968) {
return 4;
}
}
else if (AFD >= 16) {
return 1;
}
}
else if (refDCval >= 127) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (negBalance < 39) {
if (refDCval < 170) {
return 8;
}
else if (refDCval >= 170) {
if (Rs < 2876) {
return 2;
}
else if (Rs >= 2876) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (AFD < 14) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (AFD >= 14) {
return 4;
}
}
}
}
}
else if (tcor >= 94) {
if (MVDiff < 10) {
if (Rs < 3246) {
return 4;
}
else if (Rs >= 3246) {
return 2;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
else if (posBalance >= 51) {
if (Rs < 5079) {
if (AFD < 18) {
if (diffAFD < -3) {
if (MVDiff < 9) {
if (Rs < 4030) {
return 4;
}
else if (Rs >= 4030) {
return 8;
}
}
else if (MVDiff >= 9) {
if (Rs < 4198) {
if (Rs < 1723) {
return 4;
}
else if (Rs >= 1723) {
return 8;
}
}
else if (Rs >= 4198) {
return 2;
}
}
}
else if (diffAFD >= -3) {
if (MVDiff < 3) {
if (mcTcor < 93) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
return 8;
}
}
else if (mcTcor >= 93) {
if (Cs < 1323) {
return 4;
}
else if (Cs >= 1323) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (posBalance < 64) {
if (diffMVdiffVal < -12) {
return 1;
}
else if (diffMVdiffVal >= -12) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
if (Cs < 2742) {
return 8;
}
else if (Cs >= 2742) {
return 4;
}
}
}
}
else if (posBalance >= 64) {
if (mu_mv_mag_sq < 57) {
return 8;
}
else if (mu_mv_mag_sq >= 57) {
return 4;
}
}
}
}
}
else if (AFD >= 18) {
if (Rs < 1978) {
if (RsCsDiff < 201) {
if (mu_mv_mag_sq < 7) {
if (Rs < 1641) {
return 8;
}
else if (Rs >= 1641) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 4;
}
}
}
else if (RsCsDiff >= 201) {
if (mcTcor < 95) {
if (RsCsDiff < 301) {
return 1;
}
else if (RsCsDiff >= 301) {
return 4;
}
}
else if (mcTcor >= 95) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 2;
}
}
}
}
else if (Rs >= 1978) {
if (Cs < 2006) {
if (TSC < 10) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (ssDCval < 148) {
return 1;
}
else if (ssDCval >= 148) {
return 2;
}
}
}
else if (TSC >= 10) {
if (diffAFD < -2) {
if (Rs < 2818) {
if (Rs < 2715) {
return 1;
}
else if (Rs >= 2715) {
return 2;
}
}
else if (Rs >= 2818) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (Cs >= 2006) {
if (ssDCval < 161) {
if (refDCval < 110) {
if (Cs < 2515) {
return 2;
}
else if (Cs >= 2515) {
return 8;
}
}
else if (refDCval >= 110) {
if (negBalance < 59) {
return 8;
}
else if (negBalance >= 59) {
if (Rs < 2679) {
return 8;
}
else if (Rs >= 2679) {
return 4;
}
}
}
}
else if (ssDCval >= 161) {
return 4;
}
}
}
}
}
else if (Rs >= 5079) {
if (mu_mv_mag_sq < 67) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
return 8;
}
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 6189) {
if (Cs < 2221) {
return 1;
}
else if (Cs >= 2221) {
return 8;
}
}
else if (Rs >= 6189) {
if (Rs < 8068) {
return 4;
}
else if (Rs >= 8068) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 14) {
if (SC < 11538) {
if (MVDiff < 18) {
if (Cs < 2072) {
if (Rs < 5944) {
if (Cs < 1392) {
if (posBalance < 57) {
if (ssDCval < 120) {
return 2;
}
else if (ssDCval >= 120) {
return 1;
}
}
else if (posBalance >= 57) {
return 4;
}
}
else if (Cs >= 1392) {
if (diffAFD < 5) {
if (CsDiff < 10) {
if (mu_mv_mag_sq < 79) {
return 2;
}
else if (mu_mv_mag_sq >= 79) {
return 4;
}
}
else if (CsDiff >= 10) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (Cs < 1840) {
if (Rs < 4235) {
return 2;
}
else if (Rs >= 4235) {
return 4;
}
}
else if (Cs >= 1840) {
return 8;
}
}
}
}
else if (diffAFD >= 5) {
if (Cs < 1801) {
return 8;
}
else if (Cs >= 1801) {
return 4;
}
}
}
}
else if (Rs >= 5944) {
return 2;
}
}
else if (Cs >= 2072) {
if (Rs < 4025) {
if (tcor < 76) {
if (Rs < 3743) {
return 1;
}
else if (Rs >= 3743) {
return 8;
}
}
else if (tcor >= 76) {
return 8;
}
}
else if (Rs >= 4025) {
if (negBalance < 41) {
if (TSC < 7) {
if (Rs < 5017) {
return 1;
}
else if (Rs >= 5017) {
return 8;
}
}
else if (TSC >= 7) {
return 2;
}
}
else if (negBalance >= 41) {
if (RsDiff < 22) {
if (Rs < 5130) {
if (Cs < 2522) {
if (Rs < 4631) {
return 2;
}
else if (Rs >= 4631) {
return 4;
}
}
else if (Cs >= 2522) {
return 8;
}
}
else if (Rs >= 5130) {
if (Rs < 6866) {
return 4;
}
else if (Rs >= 6866) {
return 8;
}
}
}
else if (RsDiff >= 22) {
return 2;
}
}
}
}
}
else if (MVDiff >= 18) {
if (refDCval < 144) {
if (Rs < 5297) {
if (MVDiff < 57) {
if (mu_mv_mag_sq < 9) {
if (posBalance < 57) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Cs < 2808) {
return 1;
}
else if (Cs >= 2808) {
return 2;
}
}
}
}
else if (posBalance >= 57) {
if (Rs < 3672) {
return 2;
}
else if (Rs >= 3672) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (ssDCval < 133) {
if (CsDiff < 15) {
if (Rs < 5264) {
if (diffMVdiffVal < -6) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
else if (Rs >= 5264) {
return 1;
}
}
else if (CsDiff >= 15) {
if (Rs < 3436) {
return 2;
}
else if (Rs >= 3436) {
return 1;
}
}
}
else if (ssDCval >= 133) {
if (Cs < 1365) {
return 1;
}
else if (Cs >= 1365) {
return 2;
}
}
}
}
else if (MVDiff >= 57) {
return 1;
}
}
else if (Rs >= 5297) {
if (TSC < 14) {
if (refDCval < 134) {
return 8;
}
else if (refDCval >= 134) {
if (SC < 7689) {
return 2;
}
else if (SC >= 7689) {
return 8;
}
}
}
else if (TSC >= 14) {
if (Rs < 5530) {
if (Rs < 5338) {
return 4;
}
else if (Rs >= 5338) {
return 2;
}
}
else if (Rs >= 5530) {
return 1;
}
}
}
}
else if (refDCval >= 144) {
if (gchDC < 3) {
if (negBalance < 57) {
if (Rs < 6425) {
return 4;
}
else if (Rs >= 6425) {
return 2;
}
}
else if (negBalance >= 57) {
if (diffRsCsdiff < -141) {
if (Rs < 4786) {
return 2;
}
else if (Rs >= 4786) {
return 1;
}
}
else if (diffRsCsdiff >= -141) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (TSC < 12) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
if (gchDC < 4) {
if (Rs < 3416) {
return 8;
}
else if (Rs >= 3416) {
return 4;
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (TSC >= 12) {
if (Rs < 3267) {
return 8;
}
else if (Rs >= 3267) {
if (mcTcor < 93) {
if (Rs < 5180) {
return 4;
}
else if (Rs >= 5180) {
return 1;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 11538) {
if (ssDCval < 131) {
return 1;
}
else if (ssDCval >= 131) {
if (Cs < 4879) {
return 4;
}
else if (Cs >= 4879) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Cs < 1661) {
if (Rs < 3094) {
if (Rs < 1455) {
if (refDCval < 121) {
if (negBalance < 1) {
if (Rs < 736) {
return 1;
}
else if (Rs >= 736) {
return 2;
}
}
else if (negBalance >= 1) {
if (refDCval < 63) {
if (AFD < 3) {
if (mu_mv_mag_sq < 128) {
if (Cs < 1393) {
if (Rs < 1225) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 8;
}
}
else if (Rs >= 1225) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 4;
}
}
}
else if (Cs >= 1393) {
if (diffRsCsdiff < -6) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
if (negBalance < 28) {
if (Cs < 1649) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
return 1;
}
}
else if (Cs >= 1649) {
return 4;
}
}
else if (negBalance >= 28) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (posBalance < 38) {
if (Cs < 1498) {
return 8;
}
else if (Cs >= 1498) {
return 1;
}
}
else if (posBalance >= 38) {
return 4;
}
}
}
else if (AFD >= 3) {
if (SC < 2549) {
if (Rs < 809) {
return 1;
}
else if (Rs >= 809) {
return 8;
}
}
else if (SC >= 2549) {
if (refDCval < 59) {
if (ssDCval < 52) {
if (Rs < 1343) {
return 4;
}
else if (Rs >= 1343) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (ssDCval >= 52) {
if (SC < 2575) {
return 8;
}
else if (SC >= 2575) {
return 4;
}
}
}
else if (refDCval >= 59) {
return 8;
}
}
}
}
else if (refDCval >= 63) {
if (ssDCval < 76) {
if (MVDiff < 2) {
if (Cs < 1433) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Cs < 1426) {
if (diffMVdiffVal < -2) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 1333) {
if (Cs < 1329) {
return 8;
}
else if (Cs >= 1329) {
return 2;
}
}
else if (Cs >= 1333) {
return 8;
}
}
}
else if (Cs >= 1426) {
return 2;
}
}
}
else if (Cs >= 1433) {
return 8;
}
}
else if (MVDiff >= 2) {
if (MVDiff < 4) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (MVDiff >= 4) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 376) {
return 4;
}
else if (Rs >= 376) {
return 8;
}
}
}
}
}
else if (ssDCval >= 76) {
if (Rs < 944) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Rs < 734) {
return 1;
}
else if (Rs >= 734) {
return 4;
}
}
}
else if (Rs >= 944) {
return 8;
}
}
}
}
}
else if (refDCval >= 121) {
if (TSC < 2) {
if (posBalance < 28) {
if (ssDCval < 132) {
return 8;
}
else if (ssDCval >= 132) {
return 1;
}
}
else if (posBalance >= 28) {
if (Rs < 1383) {
return 4;
}
else if (Rs >= 1383) {
return 8;
}
}
}
else if (TSC >= 2) {
if (Cs < 1478) {
if (mu_mv_mag_sq < 124) {
if (tcor < 94) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 1077) {
if (Rs < 581) {
return 2;
}
else if (Rs >= 581) {
return 8;
}
}
else if (Rs >= 1077) {
return 4;
}
}
}
else if (Cs >= 1478) {
if (Rs < 951) {
return 8;
}
else if (Rs >= 951) {
if (Rs < 1288) {
if (ssDCval < 130) {
return 4;
}
else if (ssDCval >= 130) {
if (Rs < 1099) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
else if (Rs >= 1099) {
return 4;
}
}
}
else if (Rs >= 1288) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1455) {
if (ssDCval < 86) {
if (mcTcor < 97) {
if (posBalance < 44) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (Rs < 2208) {
if (AFD < 7) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (Rs >= 2208) {
if (SC < 3864) {
return 4;
}
else if (SC >= 3864) {
if (Rs < 2518) {
return 8;
}
else if (Rs >= 2518) {
return 4;
}
}
}
}
}
else if (posBalance >= 44) {
if (Rs < 2142) {
return 8;
}
else if (Rs >= 2142) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 2446) {
return 1;
}
else if (Rs >= 2446) {
return 2;
}
}
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 18) {
if (SC < 4462) {
if (SC < 3379) {
if (Cs < 1427) {
return 8;
}
else if (Cs >= 1427) {
if (ssDCval < 38) {
return 8;
}
else if (ssDCval >= 38) {
if (Cs < 1584) {
return 1;
}
else if (Cs >= 1584) {
return 8;
}
}
}
}
else if (SC >= 3379) {
if (Cs < 1350) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
if (Rs < 2297) {
return 2;
}
else if (Rs >= 2297) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 32) {
if (Rs < 2319) {
return 8;
}
else if (Rs >= 2319) {
return 2;
}
}
else if (posBalance >= 32) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 2348) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (Rs < 2305) {
return 2;
}
else if (Rs >= 2305) {
return 8;
}
}
}
}
else if (Rs >= 2348) {
return 2;
}
}
}
}
}
else if (Cs >= 1350) {
return 8;
}
}
}
else if (SC >= 4462) {
return 1;
}
}
else if (negBalance >= 18) {
if (tcor < 98) {
if (Rs < 2497) {
if (negBalance < 36) {
if (RsDiff < 4) {
if (Rs < 1911) {
if (Rs < 1577) {
if (mu_mv_mag_sq < 118) {
return 4;
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
else if (Rs >= 1577) {
return 4;
}
}
else if (Rs >= 1911) {
return 8;
}
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (negBalance >= 36) {
return 8;
}
}
else if (Rs >= 2497) {
if (Rs < 2787) {
return 4;
}
else if (Rs >= 2787) {
return 2;
}
}
}
else if (tcor >= 98) {
if (Cs < 1616) {
if (Rs < 1494) {
if (ssDCval < 77) {
if (Rs < 1468) {
return 1;
}
else if (Rs >= 1468) {
return 4;
}
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (Rs >= 1494) {
if (posBalance < 54) {
if (diffRsCsdiff < -39) {
return 2;
}
else if (diffRsCsdiff >= -39) {
if (posBalance < 46) {
if (Rs < 1925) {
return 4;
}
else if (Rs >= 1925) {
if (Rs < 2262) {
return 8;
}
else if (Rs >= 2262) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
return 4;
}
}
}
}
}
else if (posBalance >= 46) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
}
else if (Cs >= 1616) {
if (RsCsDiff < 11) {
if (Rs < 2040) {
if (Rs < 1542) {
return 1;
}
else if (Rs >= 1542) {
return 4;
}
}
else if (Rs >= 2040) {
if (Rs < 2588) {
return 8;
}
else if (Rs >= 2588) {
return 4;
}
}
}
else if (RsCsDiff >= 11) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (CsDiff < 2) {
if (diffAFD < -19) {
if (diffMVdiffVal < -34) {
return 8;
}
else if (diffMVdiffVal >= -34) {
if (Rs < 2447) {
return 4;
}
else if (Rs >= 2447) {
return 1;
}
}
}
else if (diffAFD >= -19) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (ssDCval < 131) {
return 8;
}
else if (ssDCval >= 131) {
if (Rs < 2410) {
return 8;
}
else if (Rs >= 2410) {
return 2;
}
}
}
}
}
else if (CsDiff >= 2) {
if (ssDCval < 117) {
if (refDCval < 93) {
if (diffAFD < -2) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Cs < 1404) {
if (Rs < 1999) {
return 8;
}
else if (Rs >= 1999) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
else if (Cs >= 1404) {
if (mu_mv_mag_sq < 108) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 108) {
return 8;
}
}
}
}
else if (refDCval >= 93) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 2152) {
return 8;
}
else if (Rs >= 2152) {
return 1;
}
}
}
}
else if (ssDCval >= 117) {
if (mu_mv_mag_sq < 125) {
if (Rs < 2864) {
if (diffMVdiffVal < 6) {
if (negBalance < 43) {
return 8;
}
else if (negBalance >= 43) {
if (Cs < 1396) {
return 8;
}
else if (Cs >= 1396) {
return 1;
}
}
}
else if (diffMVdiffVal >= 6) {
return 1;
}
}
else if (Rs >= 2864) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (SC < 3486) {
return 1;
}
else if (SC >= 3486) {
return 8;
}
}
}
}
}
else if (ssDCval >= 140) {
if (ssDCval < 142) {
if (Rs < 2019) {
return 8;
}
else if (Rs >= 2019) {
return 1;
}
}
else if (ssDCval >= 142) {
if (Cs < 1569) {
if (Cs < 1403) {
return 8;
}
else if (Cs >= 1403) {
return 2;
}
}
else if (Cs >= 1569) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3094) {
if (ssDCval < 113) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 125) {
if (AFD < 4) {
if (Rs < 3636) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (Rs >= 3636) {
if (mu_mv_mag_sq < 119) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 119) {
if (RsCsDiff < 15) {
if (negBalance < 21) {
if (Rs < 5521) {
if (negBalance < 14) {
return 4;
}
else if (negBalance >= 14) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
}
else if (Rs >= 5521) {
return 8;
}
}
else if (negBalance >= 21) {
if (Cs < 1512) {
return 8;
}
else if (Cs >= 1512) {
return 4;
}
}
}
else if (RsCsDiff >= 15) {
return 4;
}
}
}
}
else if (AFD >= 4) {
if (negBalance < 32) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (RsDiff < 5) {
if (Rs < 3404) {
return 4;
}
else if (Rs >= 3404) {
return 4;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (negBalance >= 32) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1570) {
if (negBalance < 25) {
if (Rs < 3678) {
if (SC < 4977) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 3642) {
return 4;
}
else if (Rs >= 3642) {
return 8;
}
}
}
else if (SC >= 4977) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 8;
}
}
}
else if (Rs >= 3678) {
return 8;
}
}
else if (negBalance >= 25) {
return 4;
}
}
else if (Cs >= 1570) {
if (Rs < 3242) {
return 8;
}
else if (Rs >= 3242) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (posBalance < 44) {
if (Rs < 4107) {
if (Rs < 3209) {
return 8;
}
else if (Rs >= 3209) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
else if (Rs >= 4107) {
if (ssDCval < 96) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 8;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (posBalance >= 44) {
if (ssDCval < 94) {
if (Rs < 3127) {
if (Rs < 3103) {
return 2;
}
else if (Rs >= 3103) {
return 4;
}
}
else if (Rs >= 3127) {
if (negBalance < 26) {
if (Rs < 3909) {
return 4;
}
else if (Rs >= 3909) {
return 2;
}
}
else if (negBalance >= 26) {
return 2;
}
}
}
else if (ssDCval >= 94) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
if (Rs < 4856) {
return 4;
}
else if (Rs >= 4856) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (diffRsCsdiff < -54) {
if (Cs < 1551) {
return 4;
}
else if (Cs >= 1551) {
return 1;
}
}
else if (diffRsCsdiff >= -54) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
return 2;
}
}
}
}
}
else if (Cs >= 1661) {
if (refDCval < 102) {
if (SC < 4359) {
if (refDCval < 67) {
if (MVDiff < 3) {
if (TSCindex < 3) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (diffRsCsdiff < -5) {
return 8;
}
else if (diffRsCsdiff >= -5) {
if (Cs < 1906) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (ssDCval < 63) {
if (Rs < 908) {
return 1;
}
else if (Rs >= 908) {
return 8;
}
}
else if (ssDCval >= 63) {
if (negBalance < 17) {
return 8;
}
else if (negBalance >= 17) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
}
else if (Cs >= 1906) {
return 1;
}
}
}
}
else if (TSCindex >= 3) {
if (refDCval < 63) {
if (ssDCval < 53) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (ssDCval >= 53) {
return 8;
}
}
else if (refDCval >= 63) {
if (RsCsDiff < 7) {
if (Cs < 1725) {
return 8;
}
else if (Cs >= 1725) {
if (Rs < 984) {
return 8;
}
else if (Rs >= 984) {
return 1;
}
}
}
else if (RsCsDiff >= 7) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (Rs < 1027) {
return 8;
}
else if (Rs >= 1027) {
if (SC < 3040) {
if (tcor < 83) {
if (Rs < 1190) {
return 8;
}
else if (Rs >= 1190) {
return 4;
}
}
else if (tcor >= 83) {
return 4;
}
}
else if (SC >= 3040) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
if (Rs < 1774) {
return 8;
}
else if (Rs >= 1774) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 67) {
if (gchDC < 2) {
if (SC < 4240) {
if (SC < 2784) {
if (Rs < 648) {
return 8;
}
else if (Rs >= 648) {
return 1;
}
}
else if (SC >= 2784) {
if (diffMVdiffVal < -6) {
return 4;
}
else if (diffMVdiffVal >= -6) {
if (Cs < 1663) {
if (Rs < 2211) {
return 8;
}
else if (Rs >= 2211) {
return 4;
}
}
else if (Cs >= 1663) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
if (Rs < 1883) {
return 2;
}
else if (Rs >= 1883) {
return 8;
}
}
}
}
}
}
else if (SC >= 4240) {
if (RsCsDiff < 1) {
if (Rs < 2399) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (Rs < 2273) {
return 4;
}
else if (Rs >= 2273) {
return 1;
}
}
}
else if (Rs >= 2399) {
return 8;
}
}
else if (RsCsDiff >= 1) {
if (mu_mv_mag_sq < 123) {
if (RsCsDiff < 46) {
if (mu_mv_mag_sq < 115) {
return 8;
}
else if (mu_mv_mag_sq >= 115) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
return 4;
}
}
}
}
else if (RsCsDiff >= 46) {
return 8;
}
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (Rs < 1627) {
return 2;
}
else if (Rs >= 1627) {
return 8;
}
}
}
}
}
else if (SC >= 4359) {
if (Cs < 2197) {
if (Rs < 3548) {
if (refDCval < 86) {
if (posBalance < 23) {
if (Rs < 3213) {
return 8;
}
else if (Rs >= 3213) {
if (Rs < 3252) {
return 4;
}
else if (Rs >= 3252) {
if (SC < 5188) {
if (Rs < 3324) {
return 8;
}
else if (Rs >= 3324) {
return 4;
}
}
else if (SC >= 5188) {
return 8;
}
}
}
}
else if (posBalance >= 23) {
if (Cs < 1935) {
if (mcTcor < 97) {
if (diffRsCsdiff < 27) {
if (ssDCval < 78) {
if (ssDCval < 62) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 8;
}
}
else if (ssDCval >= 62) {
return 4;
}
}
else if (ssDCval >= 78) {
return 8;
}
}
else if (diffRsCsdiff >= 27) {
return 2;
}
}
else if (mcTcor >= 97) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 128) {
if (Rs < 2939) {
if (Rs < 2761) {
return 8;
}
else if (Rs >= 2761) {
return 8;
}
}
else if (Rs >= 2939) {
return 4;
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
return 4;
}
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 120) {
return 4;
}
else if (mu_mv_mag_sq >= 120) {
if (MVDiff < 1) {
if (ssDCval < 76) {
if (Rs < 2927) {
return 8;
}
else if (Rs >= 2927) {
if (Rs < 2985) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
if (posBalance < 37) {
return 4;
}
else if (posBalance >= 37) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
if (Rs < 2960) {
return 4;
}
else if (Rs >= 2960) {
return 4;
}
}
}
}
}
else if (Rs >= 2985) {
if (mu_mv_mag_sq < 128) {
if (Rs < 3145) {
return 4;
}
else if (Rs >= 3145) {
return 4;
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1860) {
if (Cs < 1767) {
if (Rs < 3005) {
return 4;
}
else if (Rs >= 3005) {
if (Rs < 3024) {
return 8;
}
else if (Rs >= 3024) {
return 4;
}
}
}
else if (Cs >= 1767) {
return 8;
}
}
else if (Cs >= 1860) {
return 4;
}
}
}
}
}
else if (ssDCval >= 76) {
if (mu_mv_mag_sq < 125) {
if (mu_mv_mag_sq < 123) {
return 4;
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 2677) {
return 4;
}
else if (Rs >= 2677) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 125) {
return 4;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 1755) {
return 4;
}
else if (Cs >= 1755) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1935) {
if (refDCval < 65) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 4;
}
}
else if (refDCval >= 65) {
if (MVDiff < 3) {
if (SC < 5149) {
return 8;
}
else if (SC >= 5149) {
if (ssDCval < 71) {
return 4;
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (diffMVdiffVal < 1) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 86) {
if (Rs < 2379) {
if (negBalance < 31) {
if (ssDCval < 92) {
if (Rs < 2366) {
return 1;
}
else if (Rs >= 2366) {
return 4;
}
}
else if (ssDCval >= 92) {
if (mu_mv_mag_sq < 110) {
return 4;
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 2329) {
return 2;
}
else if (Rs >= 2329) {
if (Cs < 2117) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
return 4;
}
}
else if (Cs >= 2117) {
if (Cs < 2138) {
return 2;
}
else if (Cs >= 2138) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 31) {
return 8;
}
}
else if (Rs >= 2379) {
if (refDCval < 89) {
if (SC < 5002) {
return 8;
}
else if (SC >= 5002) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
return 2;
}
}
}
else if (refDCval >= 89) {
if (Rs < 3434) {
if (Rs < 2398) {
if (posBalance < 22) {
if (Rs < 2392) {
return 2;
}
else if (Rs >= 2392) {
return 4;
}
}
else if (posBalance >= 22) {
return 8;
}
}
else if (Rs >= 2398) {
if (negBalance < 61) {
return 8;
}
else if (negBalance >= 61) {
return 4;
}
}
}
else if (Rs >= 3434) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3548) {
if (negBalance < 29) {
if (ssDCval < 68) {
if (Rs < 7132) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (RsDiff < 7) {
if (gchDC < 1) {
if (Rs < 4662) {
if (posBalance < 42) {
if (negBalance < 23) {
if (mu_mv_mag_sq < 118) {
if (negBalance < 16) {
return 4;
}
else if (negBalance >= 16) {
return 4;
}
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 4340) {
return 4;
}
else if (Rs >= 4340) {
if (Rs < 4451) {
return 4;
}
else if (Rs >= 4451) {
return 4;
}
}
}
}
else if (negBalance >= 23) {
return 8;
}
}
else if (posBalance >= 42) {
return 4;
}
}
else if (Rs >= 4662) {
return 8;
}
}
else if (gchDC >= 1) {
if (posBalance < 43) {
if (Rs < 4262) {
return 8;
}
else if (Rs >= 4262) {
if (Rs < 4271) {
return 4;
}
else if (Rs >= 4271) {
if (Rs < 4340) {
return 8;
}
else if (Rs >= 4340) {
return 8;
}
}
}
}
else if (posBalance >= 43) {
if (MVDiff < 3) {
if (negBalance < 15) {
return 4;
}
else if (negBalance >= 15) {
return 4;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
}
}
else if (RsDiff >= 7) {
return 8;
}
}
}
else if (Rs >= 7132) {
return 2;
}
}
else if (ssDCval >= 68) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
if (Rs < 3863) {
if (posBalance < 20) {
return 8;
}
else if (posBalance >= 20) {
return 2;
}
}
else if (Rs >= 3863) {
if (posBalance < 42) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (ssDCval < 95) {
if (Rs < 4696) {
return 2;
}
else if (Rs >= 4696) {
return 8;
}
}
else if (ssDCval >= 95) {
if (Cs < 1926) {
return 4;
}
else if (Cs >= 1926) {
return 2;
}
}
}
}
else if (posBalance >= 42) {
if (ssDCval < 93) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
else if (ssDCval >= 93) {
if (SC < 6197) {
return 4;
}
else if (SC >= 6197) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 29) {
if (posBalance < 43) {
if (ssDCval < 91) {
if (Rs < 4647) {
return 2;
}
else if (Rs >= 4647) {
return 8;
}
}
else if (ssDCval >= 91) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (Rs < 3874) {
return 4;
}
else if (Rs >= 3874) {
return 1;
}
}
}
}
}
}
else if (Cs >= 2197) {
if (SC < 8345) {
if (Cs < 2818) {
if (mu_mv_mag_sq < 121) {
if (ssDCval < 101) {
if (SC < 6360) {
if (Rs < 2389) {
if (Cs < 2350) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (Cs >= 2350) {
if (MVDiff < 1) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 4;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (Rs >= 2389) {
if (Cs < 2515) {
if (posBalance < 43) {
if (mu_mv_mag_sq < 112) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 112) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
return 4;
}
}
}
else if (posBalance >= 43) {
if (posBalance < 71) {
if (mu_mv_mag_sq < 106) {
if (Cs < 2452) {
return 8;
}
else if (Cs >= 2452) {
return 4;
}
}
else if (mu_mv_mag_sq >= 106) {
return 8;
}
}
else if (posBalance >= 71) {
if (Rs < 2885) {
return 4;
}
else if (Rs >= 2885) {
return 8;
}
}
}
}
else if (Cs >= 2515) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (mu_mv_mag_sq < 112) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
return 4;
}
}
else if (mu_mv_mag_sq >= 112) {
if (Cs < 2670) {
if (diffRsCsdiff < 15) {
if (Cs < 2637) {
if (Rs < 2880) {
return 8;
}
else if (Rs >= 2880) {
return 4;
}
}
else if (Cs >= 2637) {
return 8;
}
}
else if (diffRsCsdiff >= 15) {
return 8;
}
}
else if (Cs >= 2670) {
if (RsCsDiff < 121) {
return 4;
}
else if (RsCsDiff >= 121) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 6360) {
if (RsDiff < 3) {
if (Cs < 2300) {
return 4;
}
else if (Cs >= 2300) {
return 8;
}
}
else if (RsDiff >= 3) {
if (ssDCval < 68) {
if (tcor < 95) {
if (Cs < 2506) {
if (Cs < 2349) {
if (Rs < 4414) {
return 8;
}
else if (Rs >= 4414) {
if (Cs < 2311) {
if (Cs < 2212) {
return 4;
}
else if (Cs >= 2212) {
return 4;
}
}
else if (Cs >= 2311) {
return 8;
}
}
}
else if (Cs >= 2349) {
if (Rs < 4632) {
return 4;
}
else if (Rs >= 4632) {
if (negBalance < 18) {
if (diffRsCsdiff < -5) {
return 4;
}
else if (diffRsCsdiff >= -5) {
if (Rs < 4795) {
return 4;
}
else if (Rs >= 4795) {
return 8;
}
}
}
else if (negBalance >= 18) {
if (Rs < 4656) {
return 4;
}
else if (Rs >= 4656) {
return 4;
}
}
}
}
}
else if (Cs >= 2506) {
if (posBalance < 45) {
if (Cs < 2552) {
return 8;
}
else if (Cs >= 2552) {
if (mcTcor < 96) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (mcTcor >= 96) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 4716) {
return 8;
}
else if (Rs >= 4716) {
return 4;
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 4886) {
return 4;
}
else if (Rs >= 4886) {
return 4;
}
}
}
}
else if (tcor >= 95) {
if (diffTSC < 1) {
if (posBalance < 42) {
if (Cs < 2385) {
return 4;
}
else if (Cs >= 2385) {
if (negBalance < 17) {
if (Rs < 5060) {
if (Rs < 4956) {
return 4;
}
else if (Rs >= 4956) {
return 4;
}
}
else if (Rs >= 5060) {
return 4;
}
}
else if (negBalance >= 17) {
return 4;
}
}
}
else if (posBalance >= 42) {
return 4;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
}
else if (ssDCval >= 68) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
else if (ssDCval >= 101) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (ssDCval < 47) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (ssDCval >= 47) {
if (Rs < 2145) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
if (Rs < 2055) {
return 4;
}
else if (Rs >= 2055) {
return 8;
}
}
}
else if (Rs >= 2145) {
if (Rs < 4441) {
if (TSC < 8) {
if (Cs < 2283) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (mu_mv_mag_sq < 125) {
if (Rs < 3837) {
return 2;
}
else if (Rs >= 3837) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
}
else if (Cs >= 2283) {
if (ssDCval < 52) {
if (Rs < 3950) {
return 4;
}
else if (Rs >= 3950) {
return 8;
}
}
else if (ssDCval >= 52) {
if (mu_mv_mag_sq < 124) {
if (Rs < 2184) {
return 4;
}
else if (Rs >= 2184) {
if (Cs < 2761) {
return 8;
}
else if (Cs >= 2761) {
if (Rs < 3371) {
return 8;
}
else if (Rs >= 3371) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
}
}
else if (TSC >= 8) {
if (SC < 5772) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
else if (SC >= 5772) {
return 4;
}
}
}
else if (Rs >= 4441) {
if (TSCindex < 3) {
if (posBalance < 34) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (Rs < 5059) {
return 4;
}
else if (Rs >= 5059) {
if (Rs < 5094) {
return 8;
}
else if (Rs >= 5094) {
return 4;
}
}
}
}
else if (posBalance >= 34) {
return 8;
}
}
else if (TSCindex >= 3) {
if (Cs < 2478) {
return 4;
}
else if (Cs >= 2478) {
if (Cs < 2502) {
return 1;
}
else if (Cs >= 2502) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 2818) {
if (refDCval < 85) {
if (negBalance < 3) {
if (Rs < 4273) {
return 4;
}
else if (Rs >= 4273) {
return 8;
}
}
else if (negBalance >= 3) {
if (Rs < 2968) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (ssDCval < 83) {
if (Cs < 2925) {
return 8;
}
else if (Cs >= 2925) {
return 4;
}
}
else if (ssDCval >= 83) {
return 8;
}
}
}
else if (Rs >= 2968) {
if (posBalance < 62) {
if (posBalance < 41) {
if (mu_mv_mag_sq < 127) {
if (Cs < 3121) {
return 8;
}
else if (Cs >= 3121) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
else if (posBalance >= 41) {
return 8;
}
}
else if (posBalance >= 62) {
if (ssDCval < 81) {
if (Cs < 3707) {
return 8;
}
else if (Cs >= 3707) {
return 4;
}
}
else if (ssDCval >= 81) {
return 4;
}
}
}
}
}
else if (refDCval >= 85) {
if (refDCval < 98) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Cs < 3100) {
if (AFD < 6) {
if (Rs < 4123) {
return 8;
}
else if (Rs >= 4123) {
return 4;
}
}
else if (AFD >= 6) {
if (diffTSC < 2) {
return 4;
}
else if (diffTSC >= 2) {
if (Rs < 3093) {
return 8;
}
else if (Rs >= 3093) {
return 4;
}
}
}
}
else if (Cs >= 3100) {
if (negBalance < 46) {
if (RsDiff < 3) {
if (ssDCval < 97) {
if (Rs < 3661) {
return 4;
}
else if (Rs >= 3661) {
return 8;
}
}
else if (ssDCval >= 97) {
if (Rs < 4162) {
return 8;
}
else if (Rs >= 4162) {
return 4;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 3337) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
if (Rs < 2016) {
return 2;
}
else if (Rs >= 2016) {
if (Rs < 3239) {
return 8;
}
else if (Rs >= 3239) {
return 4;
}
}
}
}
else if (Rs >= 3337) {
if (Cs < 3106) {
return 8;
}
else if (Cs >= 3106) {
if (SC < 8023) {
if (tcor < 99) {
if (Rs < 3405) {
if (Rs < 3378) {
return 4;
}
else if (Rs >= 3378) {
return 8;
}
}
else if (Rs >= 3405) {
return 4;
}
}
else if (tcor >= 99) {
if (Rs < 4167) {
return 8;
}
else if (Rs >= 4167) {
return 4;
}
}
}
else if (SC >= 8023) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (Cs < 3203) {
if (ssDCval < 88) {
return 8;
}
else if (ssDCval >= 88) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
return 8;
}
}
}
else if (Cs >= 3203) {
if (mcTcor < 85) {
if (Rs < 3769) {
return 4;
}
else if (Rs >= 3769) {
return 8;
}
}
else if (mcTcor >= 85) {
if (Cs < 4303) {
if (Cs < 3233) {
if (Rs < 3512) {
return 8;
}
else if (Rs >= 3512) {
return 4;
}
}
else if (Cs >= 3233) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (Cs >= 4303) {
if (Rs < 3313) {
return 8;
}
else if (Rs >= 3313) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 98) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
if (Rs < 3494) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
}
else if (Rs >= 3494) {
return 8;
}
}
}
}
}
}
else if (SC >= 8345) {
if (Rs < 6608) {
if (MVDiff < 3) {
if (refDCval < 75) {
if (Rs < 5956) {
if (mu_mv_mag_sq < 127) {
if (RsCsDiff < 329) {
return 8;
}
else if (RsCsDiff >= 329) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (refDCval < 72) {
return 1;
}
else if (refDCval >= 72) {
return 8;
}
}
}
}
else if (Rs >= 5956) {
if (Cs < 4296) {
return 8;
}
else if (Cs >= 4296) {
return 2;
}
}
}
else if (refDCval >= 75) {
if (Cs < 3663) {
if (Rs < 5336) {
return 4;
}
else if (Rs >= 5336) {
return 8;
}
}
else if (Cs >= 3663) {
if (ssDCval < 79) {
if (SC < 10148) {
return 8;
}
else if (SC >= 10148) {
return 4;
}
}
else if (ssDCval >= 79) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (Rs < 5813) {
return 8;
}
else if (Rs >= 5813) {
if (ssDCval < 82) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
}
}
else if (Rs >= 6608) {
if (AFD < 10) {
if (mu_mv_mag_sq < 112) {
return 2;
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
else if (AFD >= 10) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 102) {
if (posBalance < 53) {
if (ssDCval < 117) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 116) {
if (refDCval < 117) {
if (Cs < 1859) {
if (Rs < 3239) {
return 8;
}
else if (Rs >= 3239) {
return 4;
}
}
else if (Cs >= 1859) {
return 8;
}
}
else if (refDCval >= 117) {
return 4;
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
else if (negBalance >= 37) {
if (RsCsDiff < 38) {
if (mu_mv_mag_sq < 123) {
return 8;
}
else if (mu_mv_mag_sq >= 123) {
if (Cs < 2096) {
if (Rs < 2428) {
return 8;
}
else if (Rs >= 2428) {
return 4;
}
}
else if (Cs >= 2096) {
if (Rs < 5389) {
return 2;
}
else if (Rs >= 5389) {
return 8;
}
}
}
}
else if (RsCsDiff >= 38) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (ssDCval < 200) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 25) {
if (ssDCval < 146) {
if (Rs < 2774) {
if (negBalance < 9) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Rs < 2182) {
return 1;
}
else if (Rs >= 2182) {
return 8;
}
}
}
else if (negBalance >= 9) {
if (mu_mv_mag_sq < 116) {
if (Rs < 2755) {
return 8;
}
else if (Rs >= 2755) {
return 4;
}
}
else if (mu_mv_mag_sq >= 116) {
return 4;
}
}
}
else if (Rs >= 2774) {
return 2;
}
}
else if (ssDCval >= 146) {
if (Rs < 1374) {
return 1;
}
else if (Rs >= 1374) {
return 8;
}
}
}
else if (posBalance >= 25) {
if (SC < 6802) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 1;
}
}
else if (SC >= 6802) {
if (Cs < 2905) {
if (Rs < 4883) {
return 8;
}
else if (Rs >= 4883) {
return 2;
}
}
else if (Cs >= 2905) {
if (negBalance < 41) {
if (Rs < 6344) {
return 8;
}
else if (Rs >= 6344) {
if (Cs < 4271) {
return 4;
}
else if (Cs >= 4271) {
return 8;
}
}
}
else if (negBalance >= 41) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (diffRsCsdiff < -1161) {
if (Rs < 4110) {
return 1;
}
else if (Rs >= 4110) {
return 8;
}
}
else if (diffRsCsdiff >= -1161) {
if (Rs < 3903) {
if (Cs < 3927) {
if (negBalance < 24) {
if (posBalance < 3) {
if (Cs < 1807) {
if (Cs < 1779) {
return 8;
}
else if (Cs >= 1779) {
return 1;
}
}
else if (Cs >= 1807) {
return 8;
}
}
else if (posBalance >= 3) {
if (refDCval < 137) {
return 8;
}
else if (refDCval >= 137) {
if (Rs < 2576) {
return 8;
}
else if (Rs >= 2576) {
if (Rs < 2638) {
return 1;
}
else if (Rs >= 2638) {
return 8;
}
}
}
}
}
else if (negBalance >= 24) {
if (diffAFD < -5) {
if (Rs < 2354) {
return 8;
}
else if (Rs >= 2354) {
return 1;
}
}
else if (diffAFD >= -5) {
return 8;
}
}
}
else if (Cs >= 3927) {
return 1;
}
}
else if (Rs >= 3903) {
return 8;
}
}
}
}
else if (ssDCval >= 200) {
if (Rs < 2690) {
return 8;
}
else if (Rs >= 2690) {
return 4;
}
}
}
}
else if (posBalance >= 53) {
if (Rs < 1698) {
if (Cs < 1863) {
if (mu_mv_mag_sq < 128) {
if (CsDiff < 3) {
if (Rs < 1176) {
return 8;
}
else if (Rs >= 1176) {
if (Rs < 1259) {
return 4;
}
else if (Rs >= 1259) {
return 8;
}
}
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 128) {
return 2;
}
}
else if (Cs >= 1863) {
if (mcTcor < 92) {
if (Rs < 695) {
return 8;
}
else if (Rs >= 695) {
return 4;
}
}
else if (mcTcor >= 92) {
return 2;
}
}
}
else if (Rs >= 1698) {
if (ssDCval < 137) {
if (refDCval < 105) {
if (mu_mv_mag_sq < 115) {
if (Cs < 3253) {
return 1;
}
else if (Cs >= 3253) {
return 8;
}
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 4924) {
if (Cs < 2525) {
if (Rs < 3240) {
return 8;
}
else if (Rs >= 3240) {
return 1;
}
}
else if (Cs >= 2525) {
return 8;
}
}
else if (Rs >= 4924) {
return 2;
}
}
}
else if (refDCval >= 105) {
if (RsDiff < 2) {
if (ssDCval < 118) {
if (ssDCval < 110) {
return 8;
}
else if (ssDCval >= 110) {
if (Rs < 5312) {
return 2;
}
else if (Rs >= 5312) {
return 8;
}
}
}
else if (ssDCval >= 118) {
return 8;
}
}
else if (RsDiff >= 2) {
if (diffAFD < -2) {
if (gchDC < 1) {
if (negBalance < 40) {
if (Rs < 3336) {
return 8;
}
else if (Rs >= 3336) {
return 4;
}
}
else if (negBalance >= 40) {
if (Cs < 1715) {
return 1;
}
else if (Cs >= 1715) {
if (mu_mv_mag_sq < 116) {
if (ssDCval < 118) {
return 8;
}
else if (ssDCval >= 118) {
return 1;
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
if (MVDiff < 7) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
return 8;
}
}
else if (MVDiff >= 7) {
return 2;
}
}
}
}
else if (diffAFD >= -2) {
if (diffAFD < 2) {
if (TSC < 9) {
if (negBalance < 36) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (negBalance >= 36) {
return 8;
}
}
else if (TSC >= 9) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (SC < 6736) {
if (Rs < 3102) {
return 8;
}
else if (Rs >= 3102) {
if (ssDCval < 109) {
return 8;
}
else if (ssDCval >= 109) {
return 2;
}
}
}
else if (SC >= 6736) {
return 8;
}
}
}
}
else if (diffAFD >= 2) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
if (diffAFD < 4) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Cs < 2236) {
return 2;
}
else if (Cs >= 2236) {
return 8;
}
}
}
else if (diffAFD >= 4) {
if (ssDCval < 108) {
return 4;
}
else if (ssDCval >= 108) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 137) {
if (tcor < 95) {
if (ssDCval < 145) {
if (Cs < 4097) {
return 1;
}
else if (Cs >= 4097) {
return 8;
}
}
else if (ssDCval >= 145) {
if (Rs < 6133) {
return 8;
}
else if (Rs >= 6133) {
return 4;
}
}
}
else if (tcor >= 95) {
if (Cs < 3252) {
if (negBalance < 39) {
return 8;
}
else if (negBalance >= 39) {
return 1;
}
}
else if (Cs >= 3252) {
return 8;
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelectD(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1309) {
if (Rs < 2388) {
if (Cs < 533) {
if (negBalance < 14) {
if (mu_mv_mag_sq < 100) {
if (tcor < 94) {
if (negBalance < 5) {
if (TSC < 7) {
if (posBalance < 93) {
if (refDCval < 50) {
if (mu_mv_mag_sq < 39) {
if (AFD < 3) {
if (ssDCval < 18) {
if (MVDiff < 45) {
return 2;
}
else if (MVDiff >= 45) {
return 8;
}
}
else if (ssDCval >= 18) {
return 2;
}
}
else if (AFD >= 3) {
if (Cs < 65) {
return 8;
}
else if (Cs >= 65) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (posBalance < 14) {
return 4;
}
else if (posBalance >= 14) {
if (mcTcor < 61) {
if (MVDiff < 44) {
return 4;
}
else if (MVDiff >= 44) {
return 8;
}
}
else if (mcTcor >= 61) {
return 8;
}
}
}
}
else if (refDCval >= 50) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (posBalance >= 93) {
if (ssDCval < 106) {
if (mcTcor < 90) {
if (ssDCval < 14) {
if (Rs < 142) {
return 4;
}
else if (Rs >= 142) {
return 1;
}
}
else if (ssDCval >= 14) {
if (mu_mv_mag_sq < 28) {
if (MVDiff < 69) {
return 2;
}
else if (MVDiff >= 69) {
if (Rs < 52) {
return 8;
}
else if (Rs >= 52) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 28) {
if (Rs < 490) {
return 8;
}
else if (Rs >= 490) {
return 4;
}
}
}
}
else if (mcTcor >= 90) {
if (diffAFD < -1) {
if (Rs < 78) {
return 2;
}
else if (Rs >= 78) {
return 8;
}
}
else if (diffAFD >= -1) {
if (MVDiff < 112) {
if (tcor < 81) {
return 2;
}
else if (tcor >= 81) {
return 4;
}
}
else if (MVDiff >= 112) {
return 8;
}
}
}
}
else if (ssDCval >= 106) {
if (Rs < 90) {
return 1;
}
else if (Rs >= 90) {
return 8;
}
}
}
}
else if (TSC >= 7) {
if (posBalance < 124) {
if (mu_mv_mag_sq < 37) {
if (Rs < 609) {
return 8;
}
else if (Rs >= 609) {
return 2;
}
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 242) {
return 1;
}
else if (Rs >= 242) {
return 4;
}
}
}
else if (posBalance >= 124) {
return 1;
}
}
}
else if (negBalance >= 5) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (posBalance < 66) {
if (ssDCval < 73) {
if (MVDiff < 4) {
if (ssDCval < 28) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
return 8;
}
}
else if (ssDCval >= 28) {
if (Rs < 242) {
return 4;
}
else if (Rs >= 242) {
return 1;
}
}
}
else if (MVDiff >= 4) {
if (Cs < 217) {
if (Cs < 102) {
return 1;
}
else if (Cs >= 102) {
if (refDCval < 21) {
return 8;
}
else if (refDCval >= 21) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
if (Rs < 166) {
return 8;
}
else if (Rs >= 166) {
return 1;
}
}
}
}
}
else if (Cs >= 217) {
if (MVDiff < 29) {
if (negBalance < 9) {
return 4;
}
else if (negBalance >= 9) {
return 1;
}
}
else if (MVDiff >= 29) {
return 4;
}
}
}
}
else if (ssDCval >= 73) {
if (Rs < 542) {
return 2;
}
else if (Rs >= 542) {
return 8;
}
}
}
else if (posBalance >= 66) {
if (MVDiff < 16) {
if (posBalance < 75) {
return 8;
}
else if (posBalance >= 75) {
if (ssDCval < 46) {
return 2;
}
else if (ssDCval >= 46) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
}
else if (MVDiff >= 16) {
if (tcor < 88) {
if (gchDC < 12) {
if (diffMVdiffVal < 25) {
if (negBalance < 12) {
if (TSC < 6) {
if (Cs < 134) {
return 2;
}
else if (Cs >= 134) {
return 1;
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (negBalance >= 12) {
if (Cs < 96) {
return 2;
}
else if (Cs >= 96) {
return 8;
}
}
}
else if (diffMVdiffVal >= 25) {
if (tcor < 32) {
return 2;
}
else if (tcor >= 32) {
return 8;
}
}
}
else if (gchDC >= 12) {
if (diffRsCsdiff < -14) {
return 2;
}
else if (diffRsCsdiff >= -14) {
if (Rs < 78) {
return 8;
}
else if (Rs >= 78) {
return 1;
}
}
}
}
else if (tcor >= 88) {
if (Rs < 158) {
return 4;
}
else if (Rs >= 158) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 94) {
if (SC < 875) {
if (Cs < 167) {
if (refDCval < 28) {
if (ssDCval < 22) {
if (Rs < 15) {
return 8;
}
else if (Rs >= 15) {
return 1;
}
}
else if (ssDCval >= 22) {
if (negBalance < 10) {
if (refDCval < 27) {
return 4;
}
else if (refDCval >= 27) {
if (Rs < 191) {
return 8;
}
else if (Rs >= 191) {
return 4;
}
}
}
else if (negBalance >= 10) {
if (posBalance < 34) {
return 4;
}
else if (posBalance >= 34) {
return 1;
}
}
}
}
else if (refDCval >= 28) {
if (MVDiff < 7) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 4;
}
}
else if (MVDiff >= 7) {
if (negBalance < 8) {
if (ssDCval < 130) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
if (Cs < 85) {
return 4;
}
else if (Cs >= 85) {
return 8;
}
}
}
else if (ssDCval >= 130) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
return 4;
}
}
}
else if (negBalance >= 8) {
if (Cs < 124) {
if (Rs < 723) {
if (negBalance < 10) {
return 4;
}
else if (negBalance >= 10) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (Rs < 358) {
return 8;
}
else if (Rs >= 358) {
return 1;
}
}
}
}
else if (Rs >= 723) {
return 4;
}
}
else if (Cs >= 124) {
return 4;
}
}
}
}
}
else if (Cs >= 167) {
if (refDCval < 37) {
return 4;
}
else if (refDCval >= 37) {
if (gchDC < 2) {
if (SC < 850) {
if (ssDCval < 82) {
if (Rs < 264) {
if (Cs < 199) {
if (Rs < 134) {
return 1;
}
else if (Rs >= 134) {
return 4;
}
}
else if (Cs >= 199) {
return 1;
}
}
else if (Rs >= 264) {
return 4;
}
}
else if (ssDCval >= 82) {
if (diffRsCsdiff < -1) {
return 1;
}
else if (diffRsCsdiff >= -1) {
return 4;
}
}
}
else if (SC >= 850) {
return 1;
}
}
else if (gchDC >= 2) {
if (Rs < 221) {
return 1;
}
else if (Rs >= 221) {
return 8;
}
}
}
}
}
else if (SC >= 875) {
if (TSC < 2) {
if (SC < 940) {
if (Cs < 399) {
return 1;
}
else if (Cs >= 399) {
if (SC < 918) {
return 4;
}
else if (SC >= 918) {
return 8;
}
}
}
else if (SC >= 940) {
if (ssDCval < 101) {
if (Rs < 1150) {
return 8;
}
else if (Rs >= 1150) {
return 1;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
}
else if (TSC >= 2) {
if (posBalance < 85) {
if (ssDCval < 158) {
return 1;
}
else if (ssDCval >= 158) {
return 8;
}
}
else if (posBalance >= 85) {
if (Rs < 1911) {
if (Rs < 424) {
return 1;
}
else if (Rs >= 424) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (Rs >= 1911) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (SC < 779) {
if (refDCval < 48) {
if (SC < 375) {
if (negBalance < 8) {
if (Cs < 4) {
if (refDCval < 22) {
if (MVDiff < 1076) {
if (ssDCval < 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (refDCval >= 22) {
if (ssDCval < 37) {
return 4;
}
else if (ssDCval >= 37) {
return 2;
}
}
}
else if (Cs >= 4) {
if (MVDiff < 14) {
if (Cs < 186) {
return 8;
}
else if (Cs >= 186) {
if (Rs < 139) {
return 8;
}
else if (Rs >= 139) {
return 4;
}
}
}
else if (MVDiff >= 14) {
if (negBalance < 5) {
if (MVDiff < 651) {
return 8;
}
else if (MVDiff >= 651) {
return 4;
}
}
else if (negBalance >= 5) {
if (Rs < 136) {
return 1;
}
else if (Rs >= 136) {
return 4;
}
}
}
}
}
else if (negBalance >= 8) {
if (Rs < 102) {
if (ssDCval < 40) {
if (ssDCval < 24) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (ssDCval >= 24) {
if (Rs < 86) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (Rs >= 86) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (mu_mv_mag_sq < 107) {
if (Cs < 75) {
return 1;
}
else if (Cs >= 75) {
return 8;
}
}
else if (mu_mv_mag_sq >= 107) {
return 1;
}
}
}
}
}
else if (ssDCval >= 40) {
if (diffMVdiffVal < -3) {
if (Rs < 54) {
return 8;
}
else if (Rs >= 54) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
else if (Rs >= 102) {
if (TSC < 1) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
if (Rs < 123) {
return 1;
}
else if (Rs >= 123) {
return 4;
}
}
}
else if (TSC >= 1) {
return 1;
}
}
}
}
else if (SC >= 375) {
if (Rs < 229) {
if (AFD < 2) {
if (SC < 438) {
if (ssDCval < 36) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 36) {
if (Rs < 185) {
return 4;
}
else if (Rs >= 185) {
return 8;
}
}
}
else if (SC >= 438) {
if (refDCval < 36) {
return 8;
}
else if (refDCval >= 36) {
return 1;
}
}
}
else if (AFD >= 2) {
return 4;
}
}
else if (Rs >= 229) {
if (refDCval < 39) {
if (tcor < 100) {
if (MVDiff < 3) {
if (AFD < 1) {
if (Rs < 430) {
return 4;
}
else if (Rs >= 430) {
return 8;
}
}
else if (AFD >= 1) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
else if (tcor >= 100) {
return 8;
}
}
else if (refDCval >= 39) {
if (Cs < 178) {
return 4;
}
else if (Cs >= 178) {
if (ssDCval < 40) {
return 4;
}
else if (ssDCval >= 40) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 48) {
if (ssDCval < 84) {
if (Cs < 26) {
return 2;
}
else if (Cs >= 26) {
if (ssDCval < 65) {
if (Cs < 283) {
if (negBalance < 13) {
if (ssDCval < 56) {
if (Rs < 121) {
return 8;
}
else if (Rs >= 121) {
if (gchDC < 1) {
if (posBalance < 27) {
return 8;
}
else if (posBalance >= 27) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
return 8;
}
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (negBalance >= 13) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (Rs < 125) {
return 1;
}
else if (Rs >= 125) {
return 4;
}
}
}
}
else if (Cs >= 283) {
if (Rs < 252) {
return 4;
}
else if (Rs >= 252) {
if (Rs < 404) {
return 1;
}
else if (Rs >= 404) {
return 8;
}
}
}
}
else if (ssDCval >= 65) {
if (Cs < 241) {
if (Cs < 130) {
if (Rs < 100) {
return 4;
}
else if (Rs >= 100) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
return 8;
}
}
}
else if (Cs >= 130) {
return 2;
}
}
else if (Cs >= 241) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
if (Rs < 320) {
return 1;
}
else if (Rs >= 320) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (Rs < 179) {
if (ssDCval < 119) {
return 4;
}
else if (ssDCval >= 119) {
return 8;
}
}
else if (Rs >= 179) {
if (ssDCval < 130) {
if (Cs < 384) {
if (Cs < 156) {
return 8;
}
else if (Cs >= 156) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (Cs >= 384) {
return 4;
}
}
else if (ssDCval >= 130) {
if (Cs < 255) {
return 8;
}
else if (Cs >= 255) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 779) {
if (diffTSC < -3) {
if (Rs < 760) {
if (diffMVdiffVal < -12) {
if (Cs < 421) {
return 4;
}
else if (Cs >= 421) {
if (SC < 1130) {
return 2;
}
else if (SC >= 1130) {
return 4;
}
}
}
else if (diffMVdiffVal >= -12) {
if (Rs < 568) {
return 2;
}
else if (Rs >= 568) {
return 8;
}
}
}
else if (Rs >= 760) {
if (diffAFD < -8) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (diffAFD >= -8) {
if (Rs < 1385) {
return 1;
}
else if (Rs >= 1385) {
return 4;
}
}
}
}
else if (diffTSC >= -3) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 35) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
if (SC < 926) {
return 4;
}
else if (SC >= 926) {
return 1;
}
}
}
else if (ssDCval >= 35) {
if (SC < 941) {
if (ssDCval < 122) {
if (Rs < 397) {
return 1;
}
else if (Rs >= 397) {
if (Rs < 515) {
return 8;
}
else if (Rs >= 515) {
if (Rs < 593) {
if (Cs < 393) {
return 1;
}
else if (Cs >= 393) {
return 8;
}
}
else if (Rs >= 593) {
return 8;
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
}
else if (SC >= 941) {
if (ssDCval < 67) {
if (refDCval < 60) {
if (Rs < 1315) {
if (ssDCval < 41) {
if (Rs < 640) {
return 1;
}
else if (Rs >= 640) {
return 2;
}
}
else if (ssDCval >= 41) {
if (Rs < 901) {
return 8;
}
else if (Rs >= 901) {
if (Rs < 1160) {
return 1;
}
else if (Rs >= 1160) {
return 8;
}
}
}
}
else if (Rs >= 1315) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (refDCval >= 60) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
if (Rs < 1381) {
return 8;
}
else if (Rs >= 1381) {
if (Rs < 1471) {
return 2;
}
else if (Rs >= 1471) {
return 1;
}
}
}
}
}
else if (ssDCval >= 67) {
if (Rs < 487) {
return 1;
}
else if (Rs >= 487) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 1163) {
if (SC < 1041) {
return 4;
}
else if (SC >= 1041) {
return 8;
}
}
else if (Rs >= 1163) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 137) {
if (Cs < 317) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (ssDCval < 88) {
if (Rs < 1074) {
return 1;
}
else if (Rs >= 1074) {
return 2;
}
}
else if (ssDCval >= 88) {
return 8;
}
}
}
else if (Cs >= 317) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (posBalance < 38) {
if (ssDCval < 66) {
if (Rs < 1091) {
return 8;
}
else if (Rs >= 1091) {
return 2;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (posBalance >= 38) {
if (SC < 1096) {
return 4;
}
else if (SC >= 1096) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 14) {
if (ssDCval < 35) {
if (mcTcor < 83) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (ssDCval < 23) {
if (tcor < 55) {
if (Rs < 1) {
return 8;
}
else if (Rs >= 1) {
return 1;
}
}
else if (tcor >= 55) {
return 4;
}
}
else if (ssDCval >= 23) {
if (Cs < 308) {
if (Rs < 374) {
if (negBalance < 38) {
if (posBalance < 55) {
if (RsCsDiff < 7) {
if (diffMVdiffVal < 30) {
return 1;
}
else if (diffMVdiffVal >= 30) {
return 2;
}
}
else if (RsCsDiff >= 7) {
return 8;
}
}
else if (posBalance >= 55) {
if (diffMVdiffVal < -7) {
return 8;
}
else if (diffMVdiffVal >= -7) {
if (Rs < 82) {
return 1;
}
else if (Rs >= 82) {
return 2;
}
}
}
}
else if (negBalance >= 38) {
if (MVDiff < 294) {
if (Cs < 34) {
if (tcor < 50) {
if (diffAFD < -7) {
if (Rs < 58) {
return 1;
}
else if (Rs >= 58) {
return 8;
}
}
else if (diffAFD >= -7) {
return 8;
}
}
else if (tcor >= 50) {
if (Rs < 71) {
return 4;
}
else if (Rs >= 71) {
return 1;
}
}
}
else if (Cs >= 34) {
if (ssDCval < 30) {
return 2;
}
else if (ssDCval >= 30) {
return 8;
}
}
}
else if (MVDiff >= 294) {
if (Rs < 3) {
return 4;
}
else if (Rs >= 3) {
return 1;
}
}
}
}
else if (Rs >= 374) {
return 1;
}
}
else if (Cs >= 308) {
if (TSC < 3) {
if (RsDiff < 4) {
if (MVDiff < 25) {
return 8;
}
else if (MVDiff >= 25) {
return 2;
}
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (TSC >= 3) {
if (ssDCval < 29) {
return 2;
}
else if (ssDCval >= 29) {
if (Rs < 326) {
return 2;
}
else if (Rs >= 326) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 83) {
if (posBalance < 27) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 111) {
if (negBalance < 21) {
if (ssDCval < 32) {
if (ssDCval < 27) {
return 1;
}
else if (ssDCval >= 27) {
return 4;
}
}
else if (ssDCval >= 32) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
if (Rs < 215) {
return 4;
}
else if (Rs >= 215) {
return 2;
}
}
}
}
else if (negBalance >= 21) {
if (negBalance < 49) {
if (Rs < 105) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (Rs >= 105) {
return 1;
}
}
else if (negBalance >= 49) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 84) {
return 4;
}
else if (Rs >= 84) {
return 8;
}
}
}
}
else if (MVDiff >= 8) {
if (posBalance < 15) {
if (Rs < 41) {
if (negBalance < 68) {
if (ssDCval < 22) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (negBalance >= 68) {
if (MVDiff < 126) {
return 8;
}
else if (MVDiff >= 126) {
if (MVDiff < 673) {
return 4;
}
else if (MVDiff >= 673) {
return 8;
}
}
}
}
else if (Rs >= 41) {
if (posBalance < 3) {
return 4;
}
else if (posBalance >= 3) {
if (Rs < 317) {
return 1;
}
else if (Rs >= 317) {
return 4;
}
}
}
}
else if (posBalance >= 15) {
if (ssDCval < 31) {
return 1;
}
else if (ssDCval >= 31) {
if (mu_mv_mag_sq < 71) {
return 8;
}
else if (mu_mv_mag_sq >= 71) {
return 1;
}
}
}
}
}
else if (posBalance >= 27) {
if (posBalance < 42) {
if (refDCval < 30) {
if (Cs < 301) {
if (mu_mv_mag_sq < 41) {
if (mu_mv_mag_sq < 7) {
return 1;
}
else if (mu_mv_mag_sq >= 7) {
if (Rs < 94) {
if (Cs < 53) {
return 1;
}
else if (Cs >= 53) {
return 8;
}
}
else if (Rs >= 94) {
if (Cs < 56) {
return 4;
}
else if (Cs >= 56) {
if (refDCval < 27) {
return 1;
}
else if (refDCval >= 27) {
if (SC < 397) {
return 1;
}
else if (SC >= 397) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
return 1;
}
}
else if (Cs >= 301) {
if (mu_mv_mag_sq < 62) {
if (gchDC < 1) {
if (Rs < 253) {
if (Cs < 344) {
return 1;
}
else if (Cs >= 344) {
return 4;
}
}
else if (Rs >= 253) {
return 4;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 62) {
return 1;
}
}
}
else if (refDCval >= 30) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 8) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 4;
}
}
}
else if (MVDiff >= 8) {
if (Rs < 279) {
return 8;
}
else if (Rs >= 279) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 261) {
if (SC < 511) {
if (mcTcor < 98) {
if (AFD < 4) {
if (Cs < 92) {
return 1;
}
else if (Cs >= 92) {
if (MVDiff < 13) {
if (tcor < 97) {
if (ssDCval < 33) {
if (posBalance < 28) {
return 4;
}
else if (posBalance >= 28) {
return 8;
}
}
else if (ssDCval >= 33) {
return 4;
}
}
else if (tcor >= 97) {
return 1;
}
}
else if (MVDiff >= 13) {
if (ssDCval < 32) {
return 1;
}
else if (ssDCval >= 32) {
return 8;
}
}
}
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 57) {
if (tcor < 85) {
return 1;
}
else if (tcor >= 85) {
if (mu_mv_mag_sq < 41) {
return 4;
}
else if (mu_mv_mag_sq >= 41) {
if (RsDiff < 2) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (RsDiff >= 2) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (Rs < 208) {
if (Rs < 153) {
return 8;
}
else if (Rs >= 153) {
return 1;
}
}
else if (Rs >= 208) {
return 4;
}
}
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 11) {
return 4;
}
else if (mu_mv_mag_sq >= 11) {
if (posBalance < 31) {
if (Rs < 118) {
return 8;
}
else if (Rs >= 118) {
return 1;
}
}
else if (posBalance >= 31) {
return 1;
}
}
}
}
else if (SC >= 511) {
if (ssDCval < 32) {
if (Cs < 225) {
if (Rs < 459) {
return 8;
}
else if (Rs >= 459) {
return 4;
}
}
else if (Cs >= 225) {
return 1;
}
}
else if (ssDCval >= 32) {
return 4;
}
}
}
else if (Cs >= 261) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
}
}
else if (posBalance >= 42) {
if (Rs < 293) {
if (TSC < 3) {
if (posBalance < 67) {
if (RsCsDiff < 2) {
if (mcTcor < 88) {
if (Cs < 105) {
return 1;
}
else if (Cs >= 105) {
return 8;
}
}
else if (mcTcor >= 88) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (RsCsDiff >= 2) {
if (MVDiff < 24) {
return 1;
}
else if (MVDiff >= 24) {
if (Cs < 74) {
return 1;
}
else if (Cs >= 74) {
if (Cs < 250) {
return 8;
}
else if (Cs >= 250) {
if (Rs < 188) {
return 2;
}
else if (Rs >= 188) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 67) {
if (Cs < 15) {
return 2;
}
else if (Cs >= 15) {
return 4;
}
}
}
else if (TSC >= 3) {
if (CsDiff < 1) {
if (Cs < 154) {
return 1;
}
else if (Cs >= 154) {
return 2;
}
}
else if (CsDiff >= 1) {
if (AFD < 6) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
else if (AFD >= 6) {
if (Rs < 276) {
return 8;
}
else if (Rs >= 276) {
return 1;
}
}
}
}
}
else if (Rs >= 293) {
if (negBalance < 33) {
if (Cs < 487) {
if (MVDiff < 48) {
if (posBalance < 65) {
return 1;
}
else if (posBalance >= 65) {
if (Rs < 426) {
return 8;
}
else if (Rs >= 426) {
return 1;
}
}
}
else if (MVDiff >= 48) {
return 8;
}
}
else if (Cs >= 487) {
if (Rs < 511) {
return 2;
}
else if (Rs >= 511) {
return 8;
}
}
}
else if (negBalance >= 33) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 35) {
if (Rs < 241) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 50) {
if (mcTcor < 88) {
if (tcor < 68) {
if (mu_mv_mag_sq < 23) {
if (diffAFD < 2) {
if (negBalance < 56) {
if (MVDiff < 17) {
if (Rs < 191) {
return 8;
}
else if (Rs >= 191) {
return 2;
}
}
else if (MVDiff >= 17) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (Rs < 156) {
return 1;
}
else if (Rs >= 156) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (Cs < 104) {
return 1;
}
else if (Cs >= 104) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (Cs < 92) {
if (Rs < 136) {
if (Rs < 58) {
return 8;
}
else if (Rs >= 58) {
return 4;
}
}
else if (Rs >= 136) {
return 1;
}
}
else if (Cs >= 92) {
if (diffMVdiffVal < 26) {
if (Rs < 234) {
return 8;
}
else if (Rs >= 234) {
return 1;
}
}
else if (diffMVdiffVal >= 26) {
return 1;
}
}
}
}
else if (tcor >= 68) {
if (mcTcor < 84) {
return 8;
}
else if (mcTcor >= 84) {
if (Rs < 144) {
if (Cs < 303) {
return 8;
}
else if (Cs >= 303) {
return 1;
}
}
else if (Rs >= 144) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (Cs < 81) {
return 1;
}
else if (Cs >= 81) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 88) {
if (posBalance < 8) {
if (ssDCval < 36) {
return 4;
}
else if (ssDCval >= 36) {
if (Rs < 4) {
return 2;
}
else if (Rs >= 4) {
return 8;
}
}
}
else if (posBalance >= 8) {
if (tcor < 91) {
if (TSC < 4) {
if (posBalance < 40) {
if (negBalance < 58) {
if (Cs < 39) {
return 2;
}
else if (Cs >= 39) {
return 4;
}
}
else if (negBalance >= 58) {
return 1;
}
}
else if (posBalance >= 40) {
if (SC < 112) {
if (negBalance < 51) {
if (Rs < 9) {
return 8;
}
else if (Rs >= 9) {
return 2;
}
}
else if (negBalance >= 51) {
if (ssDCval < 43) {
return 1;
}
else if (ssDCval >= 43) {
return 2;
}
}
}
else if (SC >= 112) {
if (Rs < 95) {
if (Rs < 57) {
return 1;
}
else if (Rs >= 57) {
if (negBalance < 47) {
if (ssDCval < 42) {
return 4;
}
else if (ssDCval >= 42) {
if (Rs < 74) {
return 1;
}
else if (Rs >= 74) {
return 2;
}
}
}
else if (negBalance >= 47) {
return 1;
}
}
}
else if (Rs >= 95) {
if (mu_mv_mag_sq < 16) {
if (refDCval < 36) {
return 8;
}
else if (refDCval >= 36) {
if (Rs < 166) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (Cs < 86) {
return 1;
}
else if (Cs >= 86) {
return 2;
}
}
}
else if (Rs >= 166) {
if (Cs < 90) {
return 1;
}
else if (Cs >= 90) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (Rs < 125) {
return 2;
}
else if (Rs >= 125) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Cs < 263) {
return 4;
}
else if (Cs >= 263) {
return 1;
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (ssDCval < 37) {
if (Rs < 119) {
return 2;
}
else if (Rs >= 119) {
return 8;
}
}
else if (ssDCval >= 37) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
if (Rs < 143) {
if (CsDiff < 2) {
if (Rs < 135) {
return 1;
}
else if (Rs >= 135) {
return 8;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (Rs >= 143) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (Cs < 150) {
return 8;
}
else if (Cs >= 150) {
return 4;
}
}
}
}
}
}
}
}
else if (tcor >= 91) {
if (MVDiff < 102) {
if (posBalance < 71) {
if (Rs < 207) {
return 1;
}
else if (Rs >= 207) {
if (ssDCval < 40) {
return 2;
}
else if (ssDCval >= 40) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
else if (posBalance >= 71) {
return 2;
}
}
else if (MVDiff >= 102) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (negBalance < 30) {
if (mu_mv_mag_sq < 103) {
if (ssDCval < 48) {
if (posBalance < 29) {
if (Rs < 127) {
if (Rs < 93) {
return 4;
}
else if (Rs >= 93) {
return 1;
}
}
else if (Rs >= 127) {
if (ssDCval < 39) {
if (Cs < 277) {
return 4;
}
else if (Cs >= 277) {
return 8;
}
}
else if (ssDCval >= 39) {
return 1;
}
}
}
else if (posBalance >= 29) {
if (mcTcor < 97) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (ssDCval < 39) {
if (negBalance < 22) {
if (mu_mv_mag_sq < 93) {
return 8;
}
else if (mu_mv_mag_sq >= 93) {
if (Rs < 186) {
return 1;
}
else if (Rs >= 186) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (ssDCval >= 39) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
if (Rs < 60) {
return 1;
}
else if (Rs >= 60) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 1) {
if (Rs < 197) {
return 4;
}
else if (Rs >= 197) {
return 1;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
else if (ssDCval >= 48) {
if (refDCval < 50) {
return 8;
}
else if (refDCval >= 50) {
if (Rs < 108) {
return 2;
}
else if (Rs >= 108) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (tcor < 90) {
if (Cs < 133) {
return 4;
}
else if (Cs >= 133) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (tcor >= 90) {
if (negBalance < 16) {
if (TSCindex < 2) {
if (ssDCval < 39) {
if (Rs < 91) {
return 4;
}
else if (Rs >= 91) {
return 8;
}
}
else if (ssDCval >= 39) {
return 8;
}
}
else if (TSCindex >= 2) {
return 1;
}
}
else if (negBalance >= 16) {
if (posBalance < 35) {
if (ssDCval < 41) {
if (Cs < 199) {
return 1;
}
else if (Cs >= 199) {
return 8;
}
}
else if (ssDCval >= 41) {
return 8;
}
}
else if (posBalance >= 35) {
return 8;
}
}
}
}
}
else if (negBalance >= 30) {
if (gchDC < 1) {
if (Cs < 62) {
if (Rs < 68) {
return 8;
}
else if (Rs >= 68) {
if (Rs < 146) {
if (Rs < 111) {
return 4;
}
else if (Rs >= 111) {
return 2;
}
}
else if (Rs >= 146) {
return 1;
}
}
}
else if (Cs >= 62) {
if (Cs < 374) {
if (posBalance < 47) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
else if (posBalance >= 47) {
return 8;
}
}
else if (Cs >= 374) {
return 1;
}
}
}
else if (gchDC >= 1) {
if (AFD < 12) {
if (negBalance < 63) {
if (gchDC < 3) {
if (tcor < 84) {
if (Rs < 123) {
return 8;
}
else if (Rs >= 123) {
if (ssDCval < 38) {
if (Rs < 158) {
return 2;
}
else if (Rs >= 158) {
return 8;
}
}
else if (ssDCval >= 38) {
return 4;
}
}
}
else if (tcor >= 84) {
if (Cs < 440) {
return 8;
}
else if (Cs >= 440) {
return 1;
}
}
}
else if (gchDC >= 3) {
if (Rs < 96) {
return 8;
}
else if (Rs >= 96) {
return 1;
}
}
}
else if (negBalance >= 63) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
return 8;
}
}
}
else if (AFD >= 12) {
if (posBalance < 52) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Cs < 68) {
return 8;
}
else if (Cs >= 68) {
if (Rs < 63) {
return 2;
}
else if (Rs >= 63) {
return 4;
}
}
}
}
else if (posBalance >= 52) {
if (mu_mv_mag_sq < 100) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 4;
}
}
else if (mu_mv_mag_sq >= 100) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 50) {
if (TSCindex < 2) {
if (Cs < 249) {
if (refDCval < 62) {
if (posBalance < 30) {
if (Rs < 18) {
return 2;
}
else if (Rs >= 18) {
return 8;
}
}
else if (posBalance >= 30) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Cs < 135) {
if (MVDiff < 37) {
return 1;
}
else if (MVDiff >= 37) {
return 8;
}
}
else if (Cs >= 135) {
if (MVDiff < 3) {
if (Rs < 165) {
if (Rs < 146) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Rs >= 146) {
return 1;
}
}
else if (Rs >= 165) {
return 4;
}
}
else if (MVDiff >= 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (Rs < 115) {
return 4;
}
else if (Rs >= 115) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 62) {
if (ssDCval < 93) {
if (SC < 256) {
if (mu_mv_mag_sq < 117) {
if (ssDCval < 75) {
if (Rs < 78) {
if (Rs < 29) {
return 2;
}
else if (Rs >= 29) {
return 1;
}
}
else if (Rs >= 78) {
return 2;
}
}
else if (ssDCval >= 75) {
if (Cs < 100) {
return 2;
}
else if (Cs >= 100) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 117) {
return 4;
}
}
else if (SC >= 256) {
if (Rs < 234) {
if (posBalance < 25) {
return 1;
}
else if (posBalance >= 25) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 2;
}
}
}
else if (Rs >= 234) {
if (Cs < 164) {
return 8;
}
else if (Cs >= 164) {
return 2;
}
}
}
}
else if (ssDCval >= 93) {
if (Rs < 107) {
if (Rs < 8) {
return 1;
}
else if (Rs >= 8) {
return 4;
}
}
else if (Rs >= 107) {
if (posBalance < 28) {
if (Rs < 155) {
return 2;
}
else if (Rs >= 155) {
return 8;
}
}
else if (posBalance >= 28) {
return 1;
}
}
}
}
}
else if (Cs >= 249) {
if (AFD < 2) {
if (Rs < 156) {
return 4;
}
else if (Rs >= 156) {
if (ssDCval < 68) {
if (SC < 491) {
return 1;
}
else if (SC >= 491) {
return 4;
}
}
else if (ssDCval >= 68) {
if (ssDCval < 78) {
if (Rs < 183) {
if (MVDiff < 4) {
if (Cs < 373) {
return 4;
}
else if (Cs >= 373) {
return 1;
}
}
else if (MVDiff >= 4) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (Rs >= 183) {
return 1;
}
}
else if (ssDCval >= 78) {
return 4;
}
}
}
}
else if (AFD >= 2) {
if (Rs < 116) {
if (Rs < 61) {
return 1;
}
else if (Rs >= 61) {
return 4;
}
}
else if (Rs >= 116) {
return 2;
}
}
}
}
else if (TSCindex >= 2) {
if (refDCval < 72) {
if (Cs < 192) {
if (AFD < 5) {
if (refDCval < 66) {
if (posBalance < 39) {
if (negBalance < 19) {
if (Rs < 82) {
return 2;
}
else if (Rs >= 82) {
return 1;
}
}
else if (negBalance >= 19) {
if (SC < 119) {
return 2;
}
else if (SC >= 119) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 4;
}
}
}
}
else if (posBalance >= 39) {
if (mcTcor < 95) {
if (Rs < 94) {
return 2;
}
else if (Rs >= 94) {
if (Cs < 164) {
return 4;
}
else if (Cs >= 164) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (Rs < 209) {
if (SC < 231) {
return 1;
}
else if (SC >= 231) {
if (MVDiff < 1) {
if (Rs < 149) {
return 2;
}
else if (Rs >= 149) {
return 1;
}
}
else if (MVDiff >= 1) {
if (negBalance < 22) {
if (SC < 318) {
return 4;
}
else if (SC >= 318) {
return 1;
}
}
else if (negBalance >= 22) {
if (tcor < 97) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
if (MVDiff < 14) {
if (tcor < 94) {
if (mu_mv_mag_sq < 41) {
if (Rs < 125) {
return 4;
}
else if (Rs >= 125) {
return 1;
}
}
else if (mu_mv_mag_sq >= 41) {
return 4;
}
}
else if (tcor >= 94) {
return 1;
}
}
else if (MVDiff >= 14) {
if (Cs < 138) {
return 1;
}
else if (Cs >= 138) {
return 4;
}
}
}
}
else if (tcor >= 97) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 209) {
if (Cs < 188) {
return 4;
}
else if (Cs >= 188) {
return 1;
}
}
}
}
}
else if (refDCval >= 66) {
if (MVDiff < 3) {
if (Cs < 138) {
return 1;
}
else if (Cs >= 138) {
if (diffMVdiffVal < -2) {
if (mu_mv_mag_sq < 109) {
return 1;
}
else if (mu_mv_mag_sq >= 109) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
}
else if (MVDiff >= 3) {
if (Cs < 145) {
if (Rs < 68) {
return 1;
}
else if (Rs >= 68) {
if (Rs < 133) {
return 2;
}
else if (Rs >= 133) {
return 1;
}
}
}
else if (Cs >= 145) {
return 1;
}
}
}
}
else if (AFD >= 5) {
if (mu_mv_mag_sq < 21) {
if (TSC < 5) {
if (Cs < 31) {
return 1;
}
else if (Cs >= 31) {
if (Cs < 170) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (refDCval < 69) {
if (mcTcor < 96) {
if (AFD < 6) {
if (Rs < 15) {
return 4;
}
else if (Rs >= 15) {
return 1;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (mcTcor >= 96) {
if (MVDiff < 7) {
if (Rs < 173) {
return 1;
}
else if (Rs >= 173) {
return 4;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (refDCval >= 69) {
return 1;
}
}
}
else if (Cs >= 170) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 4;
}
}
}
}
else if (TSC >= 5) {
if (AFD < 16) {
if (MVDiff < 47) {
return 1;
}
else if (MVDiff >= 47) {
if (diffMVdiffVal < 13) {
return 2;
}
else if (diffMVdiffVal >= 13) {
if (tcor < 80) {
return 1;
}
else if (tcor >= 80) {
return 2;
}
}
}
}
else if (AFD >= 16) {
if (mcTcor < 48) {
if (Rs < 98) {
return 8;
}
else if (Rs >= 98) {
return 4;
}
}
else if (mcTcor >= 48) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
if (Cs < 79) {
return 1;
}
else if (Cs >= 79) {
if (Rs < 98) {
return 2;
}
else if (Rs >= 98) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (tcor < 94) {
if (mcTcor < 73) {
if (negBalance < 23) {
return 2;
}
else if (negBalance >= 23) {
if (Rs < 112) {
if (AFD < 32) {
return 4;
}
else if (AFD >= 32) {
return 1;
}
}
else if (Rs >= 112) {
return 1;
}
}
}
else if (mcTcor >= 73) {
if (tcor < 41) {
if (Cs < 115) {
return 8;
}
else if (Cs >= 115) {
return 1;
}
}
else if (tcor >= 41) {
if (negBalance < 37) {
if (ssDCval < 66) {
if (mcTcor < 91) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
else if (mcTcor >= 91) {
if (Rs < 65) {
return 8;
}
else if (Rs >= 65) {
return 4;
}
}
}
else if (ssDCval >= 66) {
if (Rs < 166) {
return 1;
}
else if (Rs >= 166) {
return 8;
}
}
}
else if (negBalance >= 37) {
if (tcor < 90) {
if (SC < 158) {
if (Rs < 77) {
return 4;
}
else if (Rs >= 77) {
return 1;
}
}
else if (SC >= 158) {
if (ssDCval < 51) {
return 2;
}
else if (ssDCval >= 51) {
return 4;
}
}
}
else if (tcor >= 90) {
return 1;
}
}
}
}
}
else if (tcor >= 94) {
if (Rs < 41) {
return 2;
}
else if (Rs >= 41) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
}
}
else if (Cs >= 192) {
if (TSC < 4) {
if (CsDiff < 1) {
if (negBalance < 30) {
if (ssDCval < 60) {
if (MVDiff < 1) {
if (Rs < 171) {
return 4;
}
else if (Rs >= 171) {
return 1;
}
}
else if (MVDiff >= 1) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
return 4;
}
}
}
else if (ssDCval >= 60) {
if (TSC < 2) {
if (Rs < 162) {
if (mu_mv_mag_sq < 83) {
if (Rs < 117) {
if (Rs < 88) {
return 4;
}
else if (Rs >= 88) {
return 1;
}
}
else if (Rs >= 117) {
return 2;
}
}
else if (mu_mv_mag_sq >= 83) {
if (mcTcor < 99) {
if (AFD < 2) {
if (refDCval < 69) {
return 4;
}
else if (refDCval >= 69) {
return 2;
}
}
else if (AFD >= 2) {
return 4;
}
}
else if (mcTcor >= 99) {
if (negBalance < 18) {
return 2;
}
else if (negBalance >= 18) {
return 4;
}
}
}
}
else if (Rs >= 162) {
if (mu_mv_mag_sq < 123) {
return 4;
}
else if (mu_mv_mag_sq >= 123) {
if (Cs < 343) {
if (Rs < 214) {
return 4;
}
else if (Rs >= 214) {
return 1;
}
}
else if (Cs >= 343) {
if (Cs < 368) {
if (Rs < 184) {
if (Rs < 168) {
return 4;
}
else if (Rs >= 168) {
return 1;
}
}
else if (Rs >= 184) {
return 4;
}
}
else if (Cs >= 368) {
return 1;
}
}
}
}
}
else if (TSC >= 2) {
if (MVDiff < 6) {
if (SC < 465) {
if (Cs < 274) {
if (posBalance < 50) {
if (Rs < 128) {
if (Rs < 120) {
return 4;
}
else if (Rs >= 120) {
return 1;
}
}
else if (Rs >= 128) {
return 4;
}
}
else if (posBalance >= 50) {
if (MVDiff < 3) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
return 1;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (Cs >= 274) {
return 4;
}
}
else if (SC >= 465) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 124) {
if (Rs < 177) {
if (Cs < 321) {
return 4;
}
else if (Cs >= 321) {
if (mu_mv_mag_sq < 113) {
return 1;
}
else if (mu_mv_mag_sq >= 113) {
if (Cs < 333) {
return 4;
}
else if (Cs >= 333) {
if (mu_mv_mag_sq < 119) {
return 4;
}
else if (mu_mv_mag_sq >= 119) {
return 1;
}
}
}
}
}
else if (Rs >= 177) {
if (Cs < 257) {
if (SC < 474) {
return 4;
}
else if (SC >= 474) {
return 1;
}
}
else if (Cs >= 257) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (Cs < 360) {
return 1;
}
else if (Cs >= 360) {
return 4;
}
}
}
else if (ssDCval >= 69) {
if (Cs < 334) {
return 4;
}
else if (Cs >= 334) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
return 1;
}
}
}
}
else if (negBalance >= 30) {
if (Rs < 117) {
if (Cs < 270) {
if (Rs < 110) {
return 2;
}
else if (Rs >= 110) {
return 8;
}
}
else if (Cs >= 270) {
return 4;
}
}
else if (Rs >= 117) {
if (ssDCval < 63) {
if (MVDiff < 1) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 108) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
return 4;
}
}
else if (mu_mv_mag_sq >= 108) {
return 1;
}
}
}
else if (ssDCval >= 63) {
if (MVDiff < 2) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Cs < 333) {
return 1;
}
else if (Cs >= 333) {
if (Rs < 163) {
return 1;
}
else if (Rs >= 163) {
return 4;
}
}
}
}
else if (MVDiff >= 2) {
if (Rs < 154) {
if (negBalance < 33) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (negBalance >= 33) {
return 1;
}
}
else if (Rs >= 154) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 4;
}
}
}
}
}
}
}
else if (CsDiff >= 1) {
if (Rs < 35) {
return 1;
}
else if (Rs >= 35) {
if (MVDiff < 2) {
if (SC < 588) {
if (Rs < 183) {
if (Cs < 398) {
if (Rs < 175) {
return 1;
}
else if (Rs >= 175) {
if (Rs < 180) {
return 4;
}
else if (Rs >= 180) {
return 1;
}
}
}
else if (Cs >= 398) {
return 4;
}
}
else if (Rs >= 183) {
if (Cs < 388) {
return 4;
}
else if (Cs >= 388) {
return 1;
}
}
}
else if (SC >= 588) {
if (Cs < 390) {
return 4;
}
else if (Cs >= 390) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (mcTcor < 94) {
if (Rs < 188) {
if (Rs < 163) {
if (Rs < 144) {
return 4;
}
else if (Rs >= 144) {
return 2;
}
}
else if (Rs >= 163) {
return 1;
}
}
else if (Rs >= 188) {
return 8;
}
}
else if (mcTcor >= 94) {
if (gchDC < 3) {
if (refDCval < 53) {
if (Cs < 316) {
if (Rs < 162) {
return 2;
}
else if (Rs >= 162) {
return 1;
}
}
else if (Cs >= 316) {
if (negBalance < 45) {
return 4;
}
else if (negBalance >= 45) {
return 1;
}
}
}
else if (refDCval >= 53) {
if (RsCsDiff < 7) {
if (diffAFD < 1) {
if (MVDiff < 53) {
if (Cs < 236) {
if (posBalance < 56) {
if (Rs < 231) {
return 4;
}
else if (Rs >= 231) {
return 1;
}
}
else if (posBalance >= 56) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (Cs >= 236) {
if (MVDiff < 3) {
if (SC < 521) {
return 1;
}
else if (SC >= 521) {
return 4;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (MVDiff >= 53) {
if (Rs < 221) {
return 1;
}
else if (Rs >= 221) {
return 4;
}
}
}
else if (diffAFD >= 1) {
if (Rs < 235) {
if (tcor < 105) {
return 4;
}
else if (tcor >= 105) {
return 8;
}
}
else if (Rs >= 235) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (negBalance < 48) {
if (Rs < 214) {
return 4;
}
else if (Rs >= 214) {
return 2;
}
}
else if (negBalance >= 48) {
return 1;
}
}
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
}
}
else if (TSC >= 4) {
if (mcTcor < 88) {
if (diffMVdiffVal < -9) {
if (Rs < 221) {
return 4;
}
else if (Rs >= 221) {
return 2;
}
}
else if (diffMVdiffVal >= -9) {
if (CsDiff < 4) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
return 2;
}
}
else if (CsDiff >= 4) {
if (MVDiff < 87) {
return 4;
}
else if (MVDiff >= 87) {
return 8;
}
}
}
}
else if (mcTcor >= 88) {
if (mu_mv_mag_sq < 24) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (posBalance < 64) {
if (Cs < 198) {
return 1;
}
else if (Cs >= 198) {
return 2;
}
}
else if (posBalance >= 64) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (negBalance < 53) {
if (MVDiff < 55) {
return 2;
}
else if (MVDiff >= 55) {
if (Rs < 192) {
return 1;
}
else if (Rs >= 192) {
return 8;
}
}
}
else if (negBalance >= 53) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 72) {
if (negBalance < 46) {
if (ssDCval < 106) {
if (Cs < 240) {
if (Rs < 112) {
if (Cs < 104) {
if (TSC < 3) {
if (posBalance < 93) {
return 1;
}
else if (posBalance >= 93) {
return 2;
}
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 5) {
if (diffMVdiffVal < 7) {
if (Rs < 57) {
return 4;
}
else if (Rs >= 57) {
return 2;
}
}
else if (diffMVdiffVal >= 7) {
if (Cs < 67) {
return 1;
}
else if (Cs >= 67) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (gchDC < 14) {
return 1;
}
else if (gchDC >= 14) {
return 8;
}
}
}
}
else if (Cs >= 104) {
if (refDCval < 96) {
if (MVDiff < 35) {
if (refDCval < 82) {
if (MVDiff < 13) {
if (Rs < 98) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
return 4;
}
}
else if (Rs >= 98) {
if (gchDC < 1) {
if (Rs < 106) {
return 1;
}
else if (Rs >= 106) {
return 2;
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
else if (MVDiff >= 13) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
else if (refDCval >= 82) {
return 1;
}
}
else if (MVDiff >= 35) {
return 2;
}
}
else if (refDCval >= 96) {
if (tcor < 97) {
if (Rs < 41) {
return 2;
}
else if (Rs >= 41) {
return 1;
}
}
else if (tcor >= 97) {
return 4;
}
}
}
}
else if (Rs >= 112) {
if (mcTcor < 98) {
if (MVDiff < 3) {
if (Cs < 149) {
if (AFD < 5) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
return 2;
}
}
else if (AFD >= 5) {
return 4;
}
}
else if (Cs >= 149) {
return 2;
}
}
else if (MVDiff >= 3) {
if (Cs < 206) {
return 1;
}
else if (Cs >= 206) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (Rs < 146) {
return 4;
}
else if (Rs >= 146) {
if (Rs < 202) {
return 8;
}
else if (Rs >= 202) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (Cs < 127) {
if (refDCval < 75) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
if (mu_mv_mag_sq < 119) {
if (Rs < 144) {
return 2;
}
else if (Rs >= 144) {
return 8;
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
}
else if (refDCval >= 75) {
if (MVDiff < 4) {
if (Rs < 150) {
return 1;
}
else if (Rs >= 150) {
return 8;
}
}
else if (MVDiff >= 4) {
if (SC < 220) {
return 2;
}
else if (SC >= 220) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
}
else if (Cs >= 127) {
if (TSCindex < 3) {
if (Cs < 210) {
if (Rs < 136) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
return 1;
}
}
else if (Rs >= 136) {
if (diffAFD < -1) {
if (mu_mv_mag_sq < 122) {
return 2;
}
else if (mu_mv_mag_sq >= 122) {
return 8;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
}
else if (Cs >= 210) {
if (ssDCval < 74) {
if (Rs < 177) {
return 2;
}
else if (Rs >= 177) {
return 1;
}
}
else if (ssDCval >= 74) {
return 2;
}
}
}
else if (TSCindex >= 3) {
if (posBalance < 59) {
if (MVDiff < 3) {
if (gchDC < 1) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (posBalance >= 59) {
if (mu_mv_mag_sq < 29) {
if (Cs < 173) {
return 1;
}
else if (Cs >= 173) {
return 2;
}
}
else if (mu_mv_mag_sq >= 29) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 240) {
if (ssDCval < 76) {
if (negBalance < 38) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (negBalance >= 38) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
if (Rs < 150) {
return 4;
}
else if (Rs >= 150) {
return 1;
}
}
}
}
else if (ssDCval >= 76) {
if (TSC < 3) {
if (Cs < 367) {
if (ssDCval < 89) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
return 4;
}
}
else if (ssDCval >= 89) {
if (mu_mv_mag_sq < 54) {
return 4;
}
else if (mu_mv_mag_sq >= 54) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (Cs >= 367) {
return 1;
}
}
else if (TSC >= 3) {
if (diffRsCsdiff < 2) {
if (mcTcor < 97) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
return 2;
}
}
else if (mcTcor >= 97) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 210) {
return 1;
}
else if (Rs >= 210) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 2) {
if (SC < 453) {
return 2;
}
else if (SC >= 453) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 106) {
if (TSC < 2) {
if (ssDCval < 136) {
if (AFD < 3) {
if (ssDCval < 107) {
if (Rs < 90) {
if (Rs < 37) {
return 1;
}
else if (Rs >= 37) {
return 4;
}
}
else if (Rs >= 90) {
return 2;
}
}
else if (ssDCval >= 107) {
if (diffTSC < -1) {
if (Rs < 76) {
return 4;
}
else if (Rs >= 76) {
return 2;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (ssDCval >= 136) {
return 2;
}
}
else if (TSC >= 2) {
if (AFD < 12) {
if (Cs < 12) {
if (MVDiff < 107) {
if (Rs < 1) {
return 8;
}
else if (Rs >= 1) {
return 4;
}
}
else if (MVDiff >= 107) {
return 1;
}
}
else if (Cs >= 12) {
if (negBalance < 36) {
if (Cs < 54) {
return 1;
}
else if (Cs >= 54) {
if (MVDiff < 10) {
if (Rs < 163) {
return 1;
}
else if (Rs >= 163) {
return 2;
}
}
else if (MVDiff >= 10) {
return 2;
}
}
}
else if (negBalance >= 36) {
return 2;
}
}
}
else if (AFD >= 12) {
if (MVDiff < 140) {
if (mu_mv_mag_sq < 38) {
return 8;
}
else if (mu_mv_mag_sq >= 38) {
return 1;
}
}
else if (MVDiff >= 140) {
if (Rs < 158) {
return 4;
}
else if (Rs >= 158) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (SC < 285) {
if (mcTcor < 60) {
if (gchDC < 9) {
return 4;
}
else if (gchDC >= 9) {
return 1;
}
}
else if (mcTcor >= 60) {
if (ssDCval < 87) {
if (gchDC < 4) {
if (Cs < 140) {
if (mu_mv_mag_sq < 45) {
return 1;
}
else if (mu_mv_mag_sq >= 45) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 126) {
return 2;
}
else if (Rs >= 126) {
return 4;
}
}
}
}
else if (Cs >= 140) {
if (Rs < 96) {
if (Rs < 82) {
return 8;
}
else if (Rs >= 82) {
return 2;
}
}
else if (Rs >= 96) {
if (Cs < 173) {
return 4;
}
else if (Cs >= 173) {
return 1;
}
}
}
}
else if (gchDC >= 4) {
if (diffTSC < -3) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Rs < 4) {
return 4;
}
else if (Rs >= 4) {
return 2;
}
}
}
else if (diffTSC >= -3) {
if (diffMVdiffVal < 37) {
if (MVDiff < 53) {
if (SC < 259) {
if (Rs < 151) {
return 8;
}
else if (Rs >= 151) {
if (Cs < 15) {
return 8;
}
else if (Cs >= 15) {
return 1;
}
}
}
else if (SC >= 259) {
return 2;
}
}
else if (MVDiff >= 53) {
if (Rs < 13) {
return 1;
}
else if (Rs >= 13) {
return 4;
}
}
}
else if (diffMVdiffVal >= 37) {
return 8;
}
}
}
}
else if (ssDCval >= 87) {
if (Rs < 80) {
if (mu_mv_mag_sq < 33) {
if (ssDCval < 108) {
if (Cs < 131) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
return 1;
}
}
else if (Cs >= 131) {
if (Rs < 20) {
return 2;
}
else if (Rs >= 20) {
return 1;
}
}
}
else if (ssDCval >= 108) {
if (Rs < 1) {
return 1;
}
else if (Rs >= 1) {
if (SC < 171) {
return 2;
}
else if (SC >= 171) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (posBalance < 59) {
if (diffAFD < -4) {
if (Rs < 4) {
return 1;
}
else if (Rs >= 4) {
return 2;
}
}
else if (diffAFD >= -4) {
return 1;
}
}
else if (posBalance >= 59) {
return 4;
}
}
}
else if (Rs >= 80) {
if (SC < 162) {
if (SC < 125) {
return 8;
}
else if (SC >= 125) {
if (MVDiff < 31) {
if (MVDiff < 21) {
return 2;
}
else if (MVDiff >= 21) {
if (Cs < 11) {
return 4;
}
else if (Cs >= 11) {
return 8;
}
}
}
else if (MVDiff >= 31) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
if (diffMVdiffVal < 19) {
return 2;
}
else if (diffMVdiffVal >= 19) {
return 8;
}
}
}
}
}
else if (SC >= 162) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
if (refDCval < 90) {
if (MVDiff < 37) {
return 8;
}
else if (MVDiff >= 37) {
if (SC < 207) {
return 4;
}
else if (SC >= 207) {
return 2;
}
}
}
else if (refDCval >= 90) {
if (Rs < 181) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
if (ssDCval < 122) {
return 1;
}
else if (ssDCval >= 122) {
return 2;
}
}
}
}
else if (Rs >= 181) {
if (refDCval < 96) {
return 1;
}
else if (refDCval >= 96) {
if (Rs < 226) {
return 8;
}
else if (Rs >= 226) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (SC >= 285) {
if (negBalance < 84) {
if (posBalance < 51) {
if (refDCval < 105) {
if (tcor < 90) {
if (Rs < 189) {
return 1;
}
else if (Rs >= 189) {
return 2;
}
}
else if (tcor >= 90) {
if (mcTcor < 97) {
return 4;
}
else if (mcTcor >= 97) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
}
}
}
else if (refDCval >= 105) {
return 2;
}
}
else if (posBalance >= 51) {
if (AFD < 13) {
if (refDCval < 78) {
if (Rs < 209) {
return 4;
}
else if (Rs >= 209) {
return 1;
}
}
else if (refDCval >= 78) {
if (SC < 586) {
if (Cs < 118) {
return 8;
}
else if (Cs >= 118) {
if (Cs < 201) {
return 2;
}
else if (Cs >= 201) {
if (Rs < 210) {
if (Cs < 291) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
if (diffMVdiffVal < -5) {
return 2;
}
else if (diffMVdiffVal >= -5) {
if (Cs < 219) {
return 8;
}
else if (Cs >= 219) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
}
else if (Cs >= 291) {
return 8;
}
}
else if (Rs >= 210) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
if (Rs < 230) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
return 2;
}
}
else if (Rs >= 230) {
return 2;
}
}
}
}
}
}
else if (SC >= 586) {
return 8;
}
}
}
else if (AFD >= 13) {
if (mu_mv_mag_sq < 32) {
if (ssDCval < 165) {
if (SC < 405) {
if (diffMVdiffVal < -3) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 4;
}
}
else if (SC >= 405) {
if (refDCval < 80) {
return 8;
}
else if (refDCval >= 80) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
}
}
else if (ssDCval >= 165) {
return 1;
}
}
else if (mu_mv_mag_sq >= 32) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (negBalance < 55) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (Rs < 112) {
return 8;
}
else if (Rs >= 112) {
return 2;
}
}
}
else if (negBalance >= 55) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 84) {
if (Rs < 135) {
return 4;
}
else if (Rs >= 135) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 241) {
if (RsDiff < 4) {
if (refDCval < 74) {
if (mcTcor < 90) {
if (AFD < 8) {
if (negBalance < 39) {
if (mu_mv_mag_sq < 100) {
if (AFD < 5) {
if (gchDC < 2) {
if (Rs < 566) {
return 8;
}
else if (Rs >= 566) {
return 1;
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (AFD >= 5) {
if (mcTcor < 87) {
if (mu_mv_mag_sq < 90) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (ssDCval < 61) {
if (Cs < 483) {
if (negBalance < 19) {
if (Rs < 419) {
return 2;
}
else if (Rs >= 419) {
return 8;
}
}
else if (negBalance >= 19) {
return 2;
}
}
else if (Cs >= 483) {
return 1;
}
}
else if (ssDCval >= 61) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 90) {
return 1;
}
}
else if (mcTcor >= 87) {
if (tcor < 77) {
if (Rs < 342) {
return 4;
}
else if (Rs >= 342) {
return 8;
}
}
else if (tcor >= 77) {
if (Rs < 794) {
return 1;
}
else if (Rs >= 794) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (ssDCval < 43) {
if (Rs < 335) {
return 1;
}
else if (Rs >= 335) {
return 8;
}
}
else if (ssDCval >= 43) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (Cs < 138) {
if (Cs < 26) {
return 8;
}
else if (Cs >= 26) {
return 1;
}
}
else if (Cs >= 138) {
if (mcTcor < 86) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (negBalance < 45) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 2;
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (mcTcor >= 86) {
if (MVDiff < 11) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (Rs < 510) {
return 4;
}
else if (Rs >= 510) {
return 8;
}
}
}
else if (MVDiff >= 11) {
if (RsDiff < 2) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (Rs < 287) {
return 2;
}
else if (Rs >= 287) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (Cs < 284) {
return 1;
}
else if (Cs >= 284) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 8) {
if (posBalance < 51) {
if (ssDCval < 49) {
if (diffTSC < -2) {
if (Rs < 397) {
return 8;
}
else if (Rs >= 397) {
if (Rs < 435) {
return 2;
}
else if (Rs >= 435) {
return 1;
}
}
}
else if (diffTSC >= -2) {
return 1;
}
}
else if (ssDCval >= 49) {
if (Rs < 427) {
return 4;
}
else if (Rs >= 427) {
return 2;
}
}
}
else if (posBalance >= 51) {
if (mu_mv_mag_sq < 51) {
if (negBalance < 47) {
if (MVDiff < 18) {
if (Cs < 181) {
return 1;
}
else if (Cs >= 181) {
return 8;
}
}
else if (MVDiff >= 18) {
if (AFD < 13) {
if (Cs < 167) {
if (Rs < 303) {
return 4;
}
else if (Rs >= 303) {
return 1;
}
}
else if (Cs >= 167) {
if (tcor < 51) {
return 8;
}
else if (tcor >= 51) {
return 2;
}
}
}
else if (AFD >= 13) {
if (mu_mv_mag_sq < 3) {
if (Rs < 422) {
return 2;
}
else if (Rs >= 422) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (MVDiff < 27) {
return 8;
}
else if (MVDiff >= 27) {
if (refDCval < 67) {
return 1;
}
else if (refDCval >= 67) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 47) {
if (ssDCval < 46) {
if (mcTcor < 65) {
return 4;
}
else if (mcTcor >= 65) {
if (TSC < 6) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (gchDC < 1) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (TSC >= 6) {
if (tcor < 33) {
return 4;
}
else if (tcor >= 33) {
return 8;
}
}
}
}
else if (ssDCval >= 46) {
if (MVDiff < 24) {
if (Rs < 715) {
return 4;
}
else if (Rs >= 715) {
return 8;
}
}
else if (MVDiff >= 24) {
if (refDCval < 66) {
if (diffRsCsdiff < 4) {
if (mu_mv_mag_sq < 2) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
if (Cs < 418) {
return 8;
}
else if (Cs >= 418) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (diffRsCsdiff < -11) {
return 4;
}
else if (diffRsCsdiff >= -11) {
return 8;
}
}
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
else if (refDCval >= 66) {
if (MVDiff < 65) {
return 4;
}
else if (MVDiff >= 65) {
if (MVDiff < 103) {
return 2;
}
else if (MVDiff >= 103) {
if (Rs < 309) {
return 1;
}
else if (Rs >= 309) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 51) {
if (gchDC < 3) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
return 4;
}
}
else if (gchDC >= 3) {
if (Rs < 374) {
return 4;
}
else if (Rs >= 374) {
if (Rs < 633) {
return 8;
}
else if (Rs >= 633) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (Rs < 1292) {
if (tcor < 101) {
if (ssDCval < 69) {
if (MVDiff < 2) {
if (tcor < 94) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
if (posBalance < 50) {
if (ssDCval < 47) {
if (Cs < 273) {
return 8;
}
else if (Cs >= 273) {
return 1;
}
}
else if (ssDCval >= 47) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
return 1;
}
}
}
else if (posBalance >= 50) {
if (Rs < 407) {
return 4;
}
else if (Rs >= 407) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
return 1;
}
}
}
}
}
else if (tcor >= 94) {
if (Rs < 476) {
if (ssDCval < 63) {
if (TSCindex < 2) {
if (mu_mv_mag_sq < 125) {
return 1;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 331) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
return 1;
}
}
else if (Rs >= 331) {
return 8;
}
}
}
else if (TSCindex >= 2) {
if (gchDC < 2) {
if (diffMVdiffVal < -2) {
if (Cs < 423) {
return 1;
}
else if (Cs >= 423) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 244) {
if (ssDCval < 53) {
if (Cs < 434) {
return 4;
}
else if (Cs >= 434) {
return 1;
}
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (Rs >= 244) {
return 1;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (ssDCval >= 63) {
if (Cs < 318) {
return 4;
}
else if (Cs >= 318) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
return 2;
}
}
}
}
else if (Rs >= 476) {
if (Rs < 916) {
if (posBalance < 42) {
if (negBalance < 43) {
if (Rs < 562) {
if (Rs < 528) {
if (SC < 865) {
return 8;
}
else if (SC >= 865) {
return 1;
}
}
else if (Rs >= 528) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
return 8;
}
}
}
else if (Rs >= 562) {
if (ssDCval < 59) {
if (Rs < 682) {
return 8;
}
else if (Rs >= 682) {
if (mu_mv_mag_sq < 126) {
return 1;
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
}
else if (ssDCval >= 59) {
if (SC < 1226) {
if (diffMVdiffVal < -1) {
if (Rs < 671) {
return 8;
}
else if (Rs >= 671) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (SC >= 1226) {
return 8;
}
}
}
}
else if (negBalance >= 43) {
return 4;
}
}
else if (posBalance >= 42) {
if (SC < 1080) {
if (ssDCval < 47) {
if (Rs < 546) {
return 8;
}
else if (Rs >= 546) {
return 4;
}
}
else if (ssDCval >= 47) {
if (refDCval < 68) {
if (mcTcor < 97) {
if (Rs < 643) {
return 8;
}
else if (Rs >= 643) {
return 2;
}
}
else if (mcTcor >= 97) {
if (negBalance < 43) {
if (negBalance < 17) {
return 8;
}
else if (negBalance >= 17) {
if (mu_mv_mag_sq < 101) {
if (Rs < 557) {
if (Rs < 527) {
return 1;
}
else if (Rs >= 527) {
return 8;
}
}
else if (Rs >= 557) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 101) {
return 1;
}
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
else if (refDCval >= 68) {
if (Rs < 577) {
if (Rs < 557) {
return 8;
}
else if (Rs >= 557) {
return 1;
}
}
else if (Rs >= 577) {
if (Cs < 430) {
return 2;
}
else if (Cs >= 430) {
return 1;
}
}
}
}
}
else if (SC >= 1080) {
if (mcTcor < 99) {
if (diffMVdiffVal < -1) {
if (mcTcor < 98) {
if (Rs < 702) {
return 2;
}
else if (Rs >= 702) {
return 8;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (AFD < 2) {
if (Rs < 764) {
return 1;
}
else if (Rs >= 764) {
return 2;
}
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 650) {
return 4;
}
else if (Rs >= 650) {
return 2;
}
}
}
}
}
else if (Rs >= 916) {
if (posBalance < 24) {
return 8;
}
else if (posBalance >= 24) {
return 1;
}
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 112) {
if (AFD < 5) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 595) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
else if (Rs >= 595) {
if (Cs < 90) {
return 8;
}
else if (Cs >= 90) {
return 1;
}
}
}
}
else if (AFD >= 5) {
if (tcor < 83) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (refDCval < 54) {
return 1;
}
else if (refDCval >= 54) {
if (Rs < 408) {
return 4;
}
else if (Rs >= 408) {
if (Rs < 640) {
return 8;
}
else if (Rs >= 640) {
return 1;
}
}
}
}
}
else if (tcor >= 83) {
return 1;
}
}
}
else if (Cs >= 112) {
if (SC < 832) {
if (ssDCval < 44) {
if (tcor < 86) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (tcor >= 86) {
if (negBalance < 41) {
if (RsCsDiff < 1) {
if (gchDC < 1) {
if (Rs < 458) {
return 8;
}
else if (Rs >= 458) {
return 4;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (RsCsDiff >= 1) {
if (mu_mv_mag_sq < 82) {
if (posBalance < 38) {
if (negBalance < 26) {
return 2;
}
else if (negBalance >= 26) {
return 4;
}
}
else if (posBalance >= 38) {
if (Rs < 394) {
if (Rs < 354) {
return 1;
}
else if (Rs >= 354) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
return 4;
}
}
}
else if (Rs >= 394) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 82) {
return 1;
}
}
}
else if (negBalance >= 41) {
if (Cs < 414) {
return 2;
}
else if (Cs >= 414) {
return 8;
}
}
}
}
else if (ssDCval >= 44) {
if (MVDiff < 56) {
if (negBalance < 31) {
if (Cs < 281) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
if (posBalance < 58) {
if (tcor < 85) {
if (Rs < 450) {
return 1;
}
else if (Rs >= 450) {
return 2;
}
}
else if (tcor >= 85) {
if (diffMVdiffVal < -3) {
if (SC < 687) {
if (Cs < 136) {
return 2;
}
else if (Cs >= 136) {
return 4;
}
}
else if (SC >= 687) {
if (Cs < 250) {
return 1;
}
else if (Cs >= 250) {
return 4;
}
}
}
else if (diffMVdiffVal >= -3) {
if (Rs < 269) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (Rs >= 269) {
return 4;
}
}
}
}
else if (posBalance >= 58) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
else if (Cs >= 281) {
if (SC < 655) {
if (SC < 575) {
if (Rs < 285) {
return 4;
}
else if (Rs >= 285) {
return 8;
}
}
else if (SC >= 575) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
return 2;
}
}
}
else if (SC >= 655) {
if (tcor < 91) {
if (Rs < 489) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
return 1;
}
}
else if (Rs >= 489) {
return 4;
}
}
else if (tcor >= 91) {
if (mu_mv_mag_sq < 81) {
if (posBalance < 60) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
if (Rs < 338) {
return 4;
}
else if (Rs >= 338) {
return 1;
}
}
}
else if (posBalance >= 60) {
if (Rs < 429) {
return 2;
}
else if (Rs >= 429) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 28) {
if (Rs < 467) {
if (Rs < 332) {
return 4;
}
else if (Rs >= 332) {
if (Rs < 462) {
return 8;
}
else if (Rs >= 462) {
return 4;
}
}
}
else if (Rs >= 467) {
return 1;
}
}
else if (negBalance >= 28) {
if (refDCval < 59) {
return 1;
}
else if (refDCval >= 59) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 31) {
if (tcor < 96) {
if (AFD < 6) {
if (negBalance < 47) {
if (SC < 787) {
if (MVDiff < 8) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (Cs < 227) {
return 1;
}
else if (Cs >= 227) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 2;
}
}
}
}
}
else if (MVDiff >= 8) {
if (mcTcor < 96) {
if (Cs < 292) {
return 4;
}
else if (Cs >= 292) {
if (SC < 777) {
return 2;
}
else if (SC >= 777) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 2;
}
}
}
}
else if (SC >= 787) {
if (tcor < 92) {
if (Rs < 452) {
return 2;
}
else if (Rs >= 452) {
return 4;
}
}
else if (tcor >= 92) {
return 1;
}
}
}
else if (negBalance >= 47) {
if (Rs < 264) {
return 4;
}
else if (Rs >= 264) {
return 2;
}
}
}
else if (AFD >= 6) {
if (Rs < 488) {
if (Cs < 408) {
if (posBalance < 61) {
if (tcor < 89) {
if (diffMVdiffVal < -5) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Rs < 289) {
return 4;
}
else if (Rs >= 289) {
return 1;
}
}
}
else if (diffMVdiffVal >= -5) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
if (mu_mv_mag_sq < 41) {
return 4;
}
else if (mu_mv_mag_sq >= 41) {
if (Cs < 221) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (Cs >= 221) {
return 4;
}
}
}
}
}
else if (tcor >= 89) {
if (mu_mv_mag_sq < 50) {
if (Cs < 346) {
if (refDCval < 51) {
return 1;
}
else if (refDCval >= 51) {
if (TSC < 4) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
return 4;
}
}
else if (TSC >= 4) {
return 1;
}
}
}
else if (Cs >= 346) {
return 4;
}
}
else if (mu_mv_mag_sq >= 50) {
return 2;
}
}
}
else if (posBalance >= 61) {
if (Cs < 256) {
if (Rs < 354) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
return 2;
}
}
else if (Rs >= 354) {
return 8;
}
}
else if (Cs >= 256) {
if (diffTSC < -2) {
return 1;
}
else if (diffTSC >= -2) {
return 2;
}
}
}
}
else if (Cs >= 408) {
if (negBalance < 40) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 292) {
return 2;
}
else if (Rs >= 292) {
return 1;
}
}
}
else if (negBalance >= 40) {
return 1;
}
}
}
else if (Rs >= 488) {
if (Cs < 239) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
return 1;
}
}
}
else if (Cs >= 239) {
if (Rs < 503) {
return 1;
}
else if (Rs >= 503) {
return 8;
}
}
}
}
}
else if (tcor >= 96) {
if (MVDiff < 10) {
if (mcTcor < 97) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
else if (mcTcor >= 97) {
if (ssDCval < 68) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (MVDiff < 8) {
if (SC < 500) {
if (SC < 493) {
return 1;
}
else if (SC >= 493) {
return 4;
}
}
else if (SC >= 500) {
return 1;
}
}
else if (MVDiff >= 8) {
if (Cs < 364) {
return 1;
}
else if (Cs >= 364) {
return 2;
}
}
}
}
else if (ssDCval >= 68) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (mcTcor < 98) {
if (Rs < 285) {
return 8;
}
else if (Rs >= 285) {
return 4;
}
}
else if (mcTcor >= 98) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 367) {
return 4;
}
else if (Rs >= 367) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 56) {
if (Cs < 340) {
return 2;
}
else if (Cs >= 340) {
if (RsDiff < 2) {
if (Rs < 266) {
return 4;
}
else if (Rs >= 266) {
return 2;
}
}
else if (RsDiff >= 2) {
if (Rs < 366) {
return 1;
}
else if (Rs >= 366) {
return 2;
}
}
}
}
}
}
else if (SC >= 832) {
if (Rs < 744) {
if (RsCsDiff < 2) {
if (TSCindex < 2) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (TSCindex >= 2) {
if (Rs < 448) {
if (mu_mv_mag_sq < 93) {
return 8;
}
else if (mu_mv_mag_sq >= 93) {
if (Rs < 392) {
return 4;
}
else if (Rs >= 392) {
return 1;
}
}
}
else if (Rs >= 448) {
if (refDCval < 68) {
if (negBalance < 36) {
if (TSCindex < 3) {
if (RsCsDiff < 1) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (RsCsDiff >= 1) {
if (mu_mv_mag_sq < 102) {
return 8;
}
else if (mu_mv_mag_sq >= 102) {
return 1;
}
}
}
else if (TSCindex >= 3) {
return 1;
}
}
else if (negBalance >= 36) {
if (Rs < 494) {
return 2;
}
else if (Rs >= 494) {
return 8;
}
}
}
else if (refDCval >= 68) {
if (SC < 1062) {
if (Rs < 627) {
return 1;
}
else if (Rs >= 627) {
return 4;
}
}
else if (SC >= 1062) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 2) {
if (refDCval < 58) {
if (negBalance < 42) {
if (MVDiff < 15) {
if (posBalance < 34) {
if (Rs < 572) {
return 8;
}
else if (Rs >= 572) {
return 2;
}
}
else if (posBalance >= 34) {
if (Cs < 345) {
if (Rs < 530) {
return 4;
}
else if (Rs >= 530) {
return 1;
}
}
else if (Cs >= 345) {
if (MVDiff < 7) {
if (TSC < 4) {
if (Rs < 439) {
if (Rs < 418) {
return 1;
}
else if (Rs >= 418) {
return 8;
}
}
else if (Rs >= 439) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (diffMVdiffVal < -3) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
if (Rs < 523) {
if (Rs < 520) {
return 4;
}
else if (Rs >= 520) {
return 8;
}
}
else if (Rs >= 523) {
return 4;
}
}
}
}
}
else if (TSC >= 4) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (MVDiff < 8) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 537) {
return 1;
}
else if (Rs >= 537) {
return 2;
}
}
}
else if (MVDiff >= 8) {
if (mcTcor < 94) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 524) {
return 4;
}
else if (Rs >= 524) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (posBalance < 52) {
if (Rs < 636) {
if (SC < 869) {
return 2;
}
else if (SC >= 869) {
return 4;
}
}
else if (Rs >= 636) {
if (Rs < 692) {
return 1;
}
else if (Rs >= 692) {
return 2;
}
}
}
else if (posBalance >= 52) {
if (Rs < 510) {
return 2;
}
else if (Rs >= 510) {
return 1;
}
}
}
}
}
}
}
}
else if (MVDiff >= 15) {
if (gchDC < 1) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (Cs < 529) {
return 1;
}
else if (Cs >= 529) {
return 4;
}
}
}
}
else if (negBalance >= 42) {
if (mu_mv_mag_sq < 36) {
if (mu_mv_mag_sq < 2) {
if (Rs < 465) {
return 2;
}
else if (Rs >= 465) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (diffRsCsdiff < -5) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 1;
}
}
else if (diffRsCsdiff >= -5) {
if (Cs < 492) {
return 1;
}
else if (Cs >= 492) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 36) {
return 8;
}
}
}
else if (refDCval >= 58) {
if (mu_mv_mag_sq < 68) {
if (mcTcor < 96) {
if (mu_mv_mag_sq < 13) {
if (ssDCval < 63) {
if (Rs < 717) {
return 2;
}
else if (Rs >= 717) {
return 4;
}
}
else if (ssDCval >= 63) {
if (Cs < 485) {
return 4;
}
else if (Cs >= 485) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (TSC < 6) {
if (SC < 933) {
return 1;
}
else if (SC >= 933) {
if (Rs < 526) {
return 2;
}
else if (Rs >= 526) {
return 4;
}
}
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (mcTcor >= 96) {
if (TSC < 4) {
if (Cs < 262) {
return 1;
}
else if (Cs >= 262) {
if (posBalance < 70) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (posBalance >= 70) {
return 2;
}
}
}
else if (TSC >= 4) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (diffMVdiffVal < 1) {
if (AFD < 5) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
return 2;
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
if (Cs < 439) {
return 4;
}
else if (Cs >= 439) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 744) {
if (Cs < 448) {
if (ssDCval < 55) {
if (mu_mv_mag_sq < 51) {
if (negBalance < 34) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
return 1;
}
}
else if (negBalance >= 34) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 51) {
if (posBalance < 52) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 939) {
if (Rs < 807) {
return 1;
}
else if (Rs >= 807) {
return 8;
}
}
else if (Rs >= 939) {
return 1;
}
}
}
else if (posBalance >= 52) {
return 8;
}
}
}
else if (ssDCval >= 55) {
if (mcTcor < 96) {
if (gchDC < 1) {
if (SC < 1066) {
if (SC < 975) {
if (Rs < 807) {
return 1;
}
else if (Rs >= 807) {
return 4;
}
}
else if (SC >= 975) {
return 2;
}
}
else if (SC >= 1066) {
return 4;
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 39) {
return 1;
}
else if (mu_mv_mag_sq >= 39) {
if (SC < 1247) {
if (SC < 1235) {
return 4;
}
else if (SC >= 1235) {
return 8;
}
}
else if (SC >= 1247) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (SC < 1475) {
if (tcor < 97) {
if (diffRsCsdiff < 4) {
return 1;
}
else if (diffRsCsdiff >= 4) {
if (Cs < 273) {
return 4;
}
else if (Cs >= 273) {
return 1;
}
}
}
else if (tcor >= 97) {
if (mu_mv_mag_sq < 2) {
return 4;
}
else if (mu_mv_mag_sq >= 2) {
if (mcTcor < 97) {
return 4;
}
else if (mcTcor >= 97) {
return 1;
}
}
}
}
else if (SC >= 1475) {
if (negBalance < 26) {
return 2;
}
else if (negBalance >= 26) {
return 1;
}
}
}
}
}
else if (Cs >= 448) {
if (posBalance < 43) {
if (RsCsDiff < 4) {
if (negBalance < 37) {
if (MVDiff < 19) {
return 1;
}
else if (MVDiff >= 19) {
return 4;
}
}
else if (negBalance >= 37) {
if (Rs < 1063) {
return 4;
}
else if (Rs >= 1063) {
return 8;
}
}
}
else if (RsCsDiff >= 4) {
return 2;
}
}
else if (posBalance >= 43) {
if (ssDCval < 56) {
if (Rs < 969) {
return 8;
}
else if (Rs >= 969) {
if (Rs < 1208) {
return 2;
}
else if (Rs >= 1208) {
return 1;
}
}
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 112) {
if (diffRsCsdiff < -10) {
if (Rs < 853) {
return 1;
}
else if (Rs >= 853) {
return 2;
}
}
else if (diffRsCsdiff >= -10) {
if (AFD < 4) {
if (Rs < 953) {
return 8;
}
else if (Rs >= 953) {
return 4;
}
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 92) {
if (Rs < 329) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
if (Rs < 278) {
return 4;
}
else if (Rs >= 278) {
if (Rs < 302) {
return 2;
}
else if (Rs >= 302) {
return 1;
}
}
}
}
else if (Rs >= 329) {
if (Rs < 516) {
if (MVDiff < 12) {
if (mu_mv_mag_sq < 49) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 1;
}
}
else if (mu_mv_mag_sq >= 49) {
if (Cs < 236) {
return 4;
}
else if (Cs >= 236) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
return 2;
}
}
}
}
else if (MVDiff >= 12) {
if (mu_mv_mag_sq < 16) {
if (Rs < 381) {
return 4;
}
else if (Rs >= 381) {
return 1;
}
}
else if (mu_mv_mag_sq >= 16) {
if (SC < 880) {
return 8;
}
else if (SC >= 880) {
if (Rs < 435) {
return 8;
}
else if (Rs >= 435) {
return 2;
}
}
}
}
}
else if (Rs >= 516) {
if (Rs < 1029) {
if (MVDiff < 8) {
if (SC < 907) {
return 4;
}
else if (SC >= 907) {
if (diffMVdiffVal < 2) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (RsCsDiff < 7) {
return 1;
}
else if (RsCsDiff >= 7) {
if (RsCsDiff < 12) {
return 4;
}
else if (RsCsDiff >= 12) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
}
else if (MVDiff >= 8) {
if (Cs < 135) {
return 1;
}
else if (Cs >= 135) {
if (MVDiff < 29) {
if (SC < 966) {
if (Cs < 258) {
return 4;
}
else if (Cs >= 258) {
return 8;
}
}
else if (SC >= 966) {
return 1;
}
}
else if (MVDiff >= 29) {
return 4;
}
}
}
}
else if (Rs >= 1029) {
if (AFD < 7) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (AFD >= 7) {
if (Rs < 1230) {
return 8;
}
else if (Rs >= 1230) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (Rs < 600) {
if (Cs < 378) {
if (mu_mv_mag_sq < 104) {
if (Cs < 225) {
return 2;
}
else if (Cs >= 225) {
return 8;
}
}
else if (mu_mv_mag_sq >= 104) {
if (negBalance < 26) {
return 2;
}
else if (negBalance >= 26) {
return 1;
}
}
}
else if (Cs >= 378) {
if (Rs < 416) {
if (Rs < 315) {
return 1;
}
else if (Rs >= 315) {
return 2;
}
}
else if (Rs >= 416) {
return 1;
}
}
}
else if (Rs >= 600) {
if (SC < 1067) {
if (MVDiff < 1) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (MVDiff >= 1) {
if (ssDCval < 71) {
if (Cs < 313) {
return 4;
}
else if (Cs >= 313) {
return 2;
}
}
else if (ssDCval >= 71) {
if (Rs < 653) {
return 4;
}
else if (Rs >= 653) {
return 8;
}
}
}
}
else if (SC >= 1067) {
if (Cs < 300) {
return 1;
}
else if (Cs >= 300) {
if (tcor < 97) {
if (Rs < 620) {
return 1;
}
else if (Rs >= 620) {
return 4;
}
}
else if (tcor >= 97) {
if (refDCval < 71) {
if (Rs < 897) {
return 2;
}
else if (Rs >= 897) {
return 8;
}
}
else if (refDCval >= 71) {
return 8;
}
}
}
}
}
}
}
}
else if (tcor >= 101) {
if (RsCsDiff < 1) {
if (Rs < 823) {
if (ssDCval < 38) {
if (Rs < 303) {
return 4;
}
else if (Rs >= 303) {
return 1;
}
}
else if (ssDCval >= 38) {
if (Cs < 12) {
return 4;
}
else if (Cs >= 12) {
return 8;
}
}
}
else if (Rs >= 823) {
return 4;
}
}
else if (RsCsDiff >= 1) {
if (negBalance < 42) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (negBalance >= 42) {
if (posBalance < 3) {
return 4;
}
else if (posBalance >= 3) {
if (tcor < 106) {
if (diffMVdiffVal < 3) {
if (Rs < 707) {
return 4;
}
else if (Rs >= 707) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (tcor >= 106) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1292) {
if (SC < 1588) {
if (mu_mv_mag_sq < 61) {
return 1;
}
else if (mu_mv_mag_sq >= 61) {
return 8;
}
}
else if (SC >= 1588) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1716) {
if (negBalance < 23) {
if (ssDCval < 64) {
if (Rs < 1365) {
if (Cs < 475) {
return 8;
}
else if (Cs >= 475) {
return 1;
}
}
else if (Rs >= 1365) {
if (TSCindex < 2) {
if (Rs < 1454) {
return 2;
}
else if (Rs >= 1454) {
return 1;
}
}
else if (TSCindex >= 2) {
if (mu_mv_mag_sq < 106) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 106) {
if (SC < 1930) {
if (Cs < 416) {
return 1;
}
else if (Cs >= 416) {
return 8;
}
}
else if (SC >= 1930) {
return 1;
}
}
}
}
}
else if (ssDCval >= 64) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (MVDiff < 8) {
if (Cs < 377) {
return 8;
}
else if (Cs >= 377) {
return 2;
}
}
else if (MVDiff >= 8) {
return 4;
}
}
}
}
else if (negBalance >= 23) {
if (tcor < 98) {
if (posBalance < 31) {
if (Rs < 1501) {
return 2;
}
else if (Rs >= 1501) {
return 1;
}
}
else if (posBalance >= 31) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
if (Rs < 1540) {
return 1;
}
else if (Rs >= 1540) {
return 8;
}
}
}
}
else if (tcor >= 98) {
if (AFD < 4) {
if (posBalance < 28) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (posBalance >= 28) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (Cs < 394) {
return 8;
}
else if (Cs >= 394) {
if (Cs < 513) {
return 1;
}
else if (Cs >= 513) {
return 8;
}
}
}
}
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (Rs >= 1716) {
if (Cs < 434) {
if (RsCsDiff < 10) {
return 2;
}
else if (RsCsDiff >= 10) {
if (Cs < 270) {
return 1;
}
else if (Cs >= 270) {
return 2;
}
}
}
else if (Cs >= 434) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (Rs < 1409) {
return 8;
}
else if (Rs >= 1409) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 74) {
if (Cs < 281) {
if (refDCval < 107) {
if (mu_mv_mag_sq < 106) {
if (SC < 1293) {
if (ssDCval < 102) {
if (mu_mv_mag_sq < 34) {
if (Cs < 85) {
if (tcor < 57) {
return 2;
}
else if (tcor >= 57) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (Cs >= 85) {
if (SC < 488) {
if (MVDiff < 8) {
if (Cs < 210) {
return 1;
}
else if (Cs >= 210) {
if (Rs < 246) {
return 1;
}
else if (Rs >= 246) {
return 2;
}
}
}
else if (MVDiff >= 8) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (ssDCval < 97) {
if (diffRsCsdiff < 2) {
if (tcor < 51) {
if (Rs < 361) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
return 2;
}
}
else if (Rs >= 361) {
return 8;
}
}
else if (tcor >= 51) {
return 2;
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 278) {
return 2;
}
else if (Rs >= 278) {
return 1;
}
}
}
else if (ssDCval >= 97) {
return 8;
}
}
}
}
else if (SC >= 488) {
if (MVDiff < 4) {
if (MVDiff < 2) {
if (Rs < 498) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
return 1;
}
}
else if (Rs >= 498) {
return 8;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (MVDiff >= 4) {
if (AFD < 10) {
if (refDCval < 76) {
if (Rs < 443) {
return 1;
}
else if (Rs >= 443) {
return 4;
}
}
else if (refDCval >= 76) {
if (MVDiff < 34) {
if (Rs < 324) {
if (Rs < 291) {
return 1;
}
else if (Rs >= 291) {
return 2;
}
}
else if (Rs >= 324) {
return 1;
}
}
else if (MVDiff >= 34) {
if (Rs < 586) {
return 1;
}
else if (Rs >= 586) {
if (Rs < 680) {
return 8;
}
else if (Rs >= 680) {
return 4;
}
}
}
}
}
else if (AFD >= 10) {
if (ssDCval < 89) {
if (Cs < 200) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
else if (Cs >= 200) {
return 1;
}
}
else if (ssDCval >= 89) {
if (diffRsCsdiff < -3) {
if (Rs < 358) {
return 4;
}
else if (Rs >= 358) {
return 2;
}
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (RsCsDiff < 2) {
if (Rs < 295) {
if (ssDCval < 91) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 251) {
return 2;
}
else if (Rs >= 251) {
if (MVDiff < 41) {
return 1;
}
else if (MVDiff >= 41) {
return 8;
}
}
}
}
else if (ssDCval >= 91) {
if (Cs < 97) {
return 8;
}
else if (Cs >= 97) {
return 4;
}
}
}
else if (Rs >= 295) {
if (mu_mv_mag_sq < 89) {
return 8;
}
else if (mu_mv_mag_sq >= 89) {
if (ssDCval < 82) {
if (Rs < 926) {
return 8;
}
else if (Rs >= 926) {
return 1;
}
}
else if (ssDCval >= 82) {
return 2;
}
}
}
}
else if (RsCsDiff >= 2) {
if (negBalance < 33) {
if (ssDCval < 76) {
if (Rs < 523) {
return 1;
}
else if (Rs >= 523) {
return 8;
}
}
else if (ssDCval >= 76) {
return 4;
}
}
else if (negBalance >= 33) {
if (tcor < 80) {
if (mu_mv_mag_sq < 68) {
if (diffMVdiffVal < 10) {
if (gchDC < 1) {
if (MVDiff < 29) {
return 8;
}
else if (MVDiff >= 29) {
if (Rs < 273) {
return 8;
}
else if (Rs >= 273) {
return 2;
}
}
}
else if (gchDC >= 1) {
return 2;
}
}
else if (diffMVdiffVal >= 10) {
return 8;
}
}
else if (mu_mv_mag_sq >= 68) {
return 4;
}
}
else if (tcor >= 80) {
if (mu_mv_mag_sq < 58) {
if (Rs < 963) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Rs >= 963) {
return 8;
}
}
else if (mu_mv_mag_sq >= 58) {
if (TSC < 6) {
if (Rs < 531) {
return 2;
}
else if (Rs >= 531) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (Cs < 239) {
return 1;
}
else if (Cs >= 239) {
return 8;
}
}
}
else if (SC >= 1293) {
if (ssDCval < 102) {
if (refDCval < 95) {
return 1;
}
else if (refDCval >= 95) {
if (Cs < 201) {
return 1;
}
else if (Cs >= 201) {
return 8;
}
}
}
else if (ssDCval >= 102) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Cs < 176) {
return 1;
}
else if (Cs >= 176) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Cs < 177) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
if (Rs < 607) {
return 2;
}
else if (Rs >= 607) {
return 1;
}
}
}
else if (Cs >= 177) {
if (posBalance < 45) {
if (SC < 508) {
if (Rs < 271) {
return 2;
}
else if (Rs >= 271) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
}
else if (SC >= 508) {
if (Cs < 251) {
if (ssDCval < 94) {
if (ssDCval < 78) {
if (Rs < 1067) {
return 8;
}
else if (Rs >= 1067) {
return 1;
}
}
else if (ssDCval >= 78) {
if (Rs < 312) {
if (Rs < 311) {
return 2;
}
else if (Rs >= 311) {
return 1;
}
}
else if (Rs >= 312) {
return 2;
}
}
}
else if (ssDCval >= 94) {
return 1;
}
}
else if (Cs >= 251) {
if (tcor < 98) {
return 2;
}
else if (tcor >= 98) {
return 1;
}
}
}
}
else if (posBalance >= 45) {
if (Cs < 211) {
if (Rs < 594) {
return 2;
}
else if (Rs >= 594) {
return 8;
}
}
else if (Cs >= 211) {
return 1;
}
}
}
}
}
else if (refDCval >= 107) {
if (AFD < 8) {
if (MVDiff < 20) {
if (diffRsCsdiff < -5) {
return 4;
}
else if (diffRsCsdiff >= -5) {
if (Rs < 291) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 278) {
return 2;
}
else if (Rs >= 278) {
return 1;
}
}
}
else if (Rs >= 291) {
return 8;
}
}
}
else if (MVDiff >= 20) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (tcor < 98) {
return 2;
}
else if (tcor >= 98) {
if (Rs < 538) {
return 4;
}
else if (Rs >= 538) {
return 1;
}
}
}
}
}
else if (AFD >= 8) {
if (ssDCval < 118) {
if (RsCsDiff < 12) {
if (gchDC < 13) {
if (mu_mv_mag_sq < 36) {
return 1;
}
else if (mu_mv_mag_sq >= 36) {
return 8;
}
}
else if (gchDC >= 13) {
return 8;
}
}
else if (RsCsDiff >= 12) {
return 4;
}
}
else if (ssDCval >= 118) {
if (Rs < 446) {
if (Cs < 252) {
if (diffRsCsdiff < -13) {
if (Rs < 306) {
return 1;
}
else if (Rs >= 306) {
return 2;
}
}
else if (diffRsCsdiff >= -13) {
return 2;
}
}
else if (Cs >= 252) {
if (Rs < 313) {
return 8;
}
else if (Rs >= 313) {
return 4;
}
}
}
else if (Rs >= 446) {
if (MVDiff < 87) {
return 1;
}
else if (MVDiff >= 87) {
return 8;
}
}
}
}
}
}
else if (Cs >= 281) {
if (mu_mv_mag_sq < 100) {
if (mcTcor < 95) {
if (MVDiff < 67) {
if (posBalance < 58) {
if (ssDCval < 112) {
if (mu_mv_mag_sq < 27) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (ssDCval < 110) {
if (diffTSC < 1) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 1;
}
}
}
else if (diffTSC >= 1) {
if (negBalance < 64) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 1;
}
}
else if (negBalance >= 64) {
return 2;
}
}
}
else if (ssDCval >= 110) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (Cs < 364) {
return 1;
}
else if (Cs >= 364) {
if (ssDCval < 94) {
if (Rs < 1230) {
return 8;
}
else if (Rs >= 1230) {
return 1;
}
}
else if (ssDCval >= 94) {
return 2;
}
}
}
}
else if (ssDCval >= 112) {
if (Rs < 391) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
if (Cs < 292) {
return 2;
}
else if (Cs >= 292) {
return 8;
}
}
}
else if (Rs >= 391) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (negBalance < 50) {
if (mu_mv_mag_sq < 41) {
if (posBalance < 62) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
if (SC < 990) {
if (mcTcor < 85) {
if (Rs < 260) {
return 8;
}
else if (Rs >= 260) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
}
else if (mcTcor >= 85) {
return 8;
}
}
else if (SC >= 990) {
return 2;
}
}
}
else if (posBalance >= 62) {
if (SC < 939) {
if (Cs < 468) {
return 8;
}
else if (Cs >= 468) {
return 2;
}
}
else if (SC >= 939) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (SC < 659) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
else if (SC >= 659) {
if (diffAFD < -3) {
if (Rs < 395) {
return 1;
}
else if (Rs >= 395) {
return 8;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
}
}
else if (negBalance >= 50) {
if (ssDCval < 137) {
if (AFD < 6) {
if (Cs < 312) {
return 2;
}
else if (Cs >= 312) {
return 8;
}
}
else if (AFD >= 6) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
return 1;
}
}
}
else if (ssDCval >= 137) {
if (Rs < 296) {
return 1;
}
else if (Rs >= 296) {
return 2;
}
}
}
}
}
else if (MVDiff >= 67) {
if (MVDiff < 140) {
if (negBalance < 81) {
if (refDCval < 78) {
return 4;
}
else if (refDCval >= 78) {
return 2;
}
}
else if (negBalance >= 81) {
if (Rs < 420) {
return 8;
}
else if (Rs >= 420) {
return 1;
}
}
}
else if (MVDiff >= 140) {
if (SC < 845) {
return 4;
}
else if (SC >= 845) {
if (AFD < 54) {
return 8;
}
else if (AFD >= 54) {
return 1;
}
}
}
}
}
else if (mcTcor >= 95) {
if (refDCval < 115) {
if (tcor < 98) {
if (mu_mv_mag_sq < 24) {
if (MVDiff < 12) {
if (Cs < 489) {
if (refDCval < 109) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
return 1;
}
}
else if (refDCval >= 109) {
if (Rs < 588) {
return 1;
}
else if (Rs >= 588) {
return 8;
}
}
}
else if (Cs >= 489) {
if (posBalance < 57) {
if (Rs < 1103) {
return 2;
}
else if (Rs >= 1103) {
return 1;
}
}
else if (posBalance >= 57) {
return 1;
}
}
}
else if (MVDiff >= 12) {
if (TSC < 6) {
if (Rs < 515) {
return 1;
}
else if (Rs >= 515) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (TSC >= 6) {
if (MVDiff < 22) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
return 1;
}
}
else if (MVDiff >= 22) {
if (SC < 727) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 2;
}
}
else if (SC >= 727) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (negBalance < 47) {
if (tcor < 94) {
if (CsDiff < 2) {
if (posBalance < 52) {
if (Cs < 388) {
return 4;
}
else if (Cs >= 388) {
return 8;
}
}
else if (posBalance >= 52) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
if (Rs < 324) {
return 8;
}
else if (Rs >= 324) {
return 2;
}
}
}
}
else if (CsDiff >= 2) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (Cs < 305) {
return 4;
}
else if (Cs >= 305) {
if (Rs < 271) {
return 1;
}
else if (Rs >= 271) {
return 2;
}
}
}
}
}
}
else if (tcor >= 94) {
if (Rs < 2190) {
if (MVDiff < 14) {
if (TSC < 3) {
if (Cs < 484) {
return 1;
}
else if (Cs >= 484) {
if (Rs < 517) {
return 1;
}
else if (Rs >= 517) {
return 8;
}
}
}
else if (TSC >= 3) {
if (ssDCval < 82) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (ssDCval >= 82) {
if (refDCval < 93) {
return 1;
}
else if (refDCval >= 93) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
if (Rs < 550) {
return 1;
}
else if (Rs >= 550) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 14) {
if (Rs < 1080) {
return 8;
}
else if (Rs >= 1080) {
if (Cs < 385) {
return 8;
}
else if (Cs >= 385) {
return 1;
}
}
}
}
else if (Rs >= 2190) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (Cs < 409) {
if (mu_mv_mag_sq < 65) {
if (Cs < 299) {
return 8;
}
else if (Cs >= 299) {
return 1;
}
}
else if (mu_mv_mag_sq >= 65) {
return 8;
}
}
else if (Cs >= 409) {
return 8;
}
}
}
}
else if (tcor >= 98) {
if (refDCval < 78) {
if (TSC < 3) {
if (Cs < 289) {
return 8;
}
else if (Cs >= 289) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
return 8;
}
}
}
else if (TSC >= 3) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
return 2;
}
}
}
else if (refDCval >= 78) {
if (ssDCval < 110) {
if (diffAFD < -1) {
if (diffRsCsdiff < -2) {
if (Cs < 399) {
return 1;
}
else if (Cs >= 399) {
if (Rs < 1045) {
return 1;
}
else if (Rs >= 1045) {
return 8;
}
}
}
else if (diffRsCsdiff >= -2) {
if (Rs < 2163) {
return 8;
}
else if (Rs >= 2163) {
return 1;
}
}
}
else if (diffAFD >= -1) {
if (diffRsCsdiff < 7) {
if (ssDCval < 79) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (Rs < 1889) {
return 8;
}
else if (Rs >= 1889) {
if (Rs < 1902) {
return 1;
}
else if (Rs >= 1902) {
return 8;
}
}
}
}
else if (ssDCval >= 79) {
if (MVDiff < 1) {
if (Rs < 1046) {
return 1;
}
else if (Rs >= 1046) {
return 8;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (diffRsCsdiff >= 7) {
if (Rs < 1006) {
return 2;
}
else if (Rs >= 1006) {
if (Rs < 1786) {
return 4;
}
else if (Rs >= 1786) {
return 8;
}
}
}
}
}
else if (ssDCval >= 110) {
return 1;
}
}
}
}
else if (refDCval >= 115) {
if (refDCval < 130) {
if (mcTcor < 99) {
if (Rs < 901) {
if (RsCsDiff < 2) {
if (mu_mv_mag_sq < 73) {
return 8;
}
else if (mu_mv_mag_sq >= 73) {
if (Rs < 433) {
return 4;
}
else if (Rs >= 433) {
return 2;
}
}
}
else if (RsCsDiff >= 2) {
if (Rs < 568) {
if (ssDCval < 113) {
if (Rs < 376) {
return 2;
}
else if (Rs >= 376) {
return 4;
}
}
else if (ssDCval >= 113) {
if (AFD < 4) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
return 2;
}
}
else if (AFD >= 4) {
if (mcTcor < 96) {
if (Rs < 348) {
return 2;
}
else if (Rs >= 348) {
return 8;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
}
}
else if (Rs >= 568) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 2;
}
}
}
}
else if (Rs >= 901) {
if (Rs < 1415) {
return 1;
}
else if (Rs >= 1415) {
return 2;
}
}
}
else if (mcTcor >= 99) {
if (posBalance < 64) {
if (ssDCval < 122) {
if (SC < 989) {
if (TSC < 3) {
if (Rs < 465) {
return 1;
}
else if (Rs >= 465) {
return 8;
}
}
else if (TSC >= 3) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (SC >= 989) {
return 2;
}
}
else if (ssDCval >= 122) {
return 1;
}
}
else if (posBalance >= 64) {
return 4;
}
}
}
else if (refDCval >= 130) {
if (Rs < 370) {
if (RsCsDiff < 7) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 69) {
if (Rs < 256) {
return 2;
}
else if (Rs >= 256) {
return 8;
}
}
else if (mu_mv_mag_sq >= 69) {
return 2;
}
}
else if (mcTcor >= 98) {
return 4;
}
}
else if (RsCsDiff >= 7) {
if (tcor < 81) {
return 1;
}
else if (tcor >= 81) {
return 2;
}
}
}
else if (Rs >= 370) {
if (Cs < 385) {
if (MVDiff < 4) {
if (Rs < 441) {
return 1;
}
else if (Rs >= 441) {
return 8;
}
}
else if (MVDiff >= 4) {
if (AFD < 9) {
if (negBalance < 16) {
return 1;
}
else if (negBalance >= 16) {
return 4;
}
}
else if (AFD >= 9) {
if (Cs < 380) {
return 1;
}
else if (Cs >= 380) {
return 2;
}
}
}
}
else if (Cs >= 385) {
if (mu_mv_mag_sq < 70) {
if (negBalance < 63) {
if (Rs < 381) {
if (Rs < 373) {
return 4;
}
else if (Rs >= 373) {
return 8;
}
}
else if (Rs >= 381) {
if (Rs < 403) {
if (mu_mv_mag_sq < 9) {
return 1;
}
else if (mu_mv_mag_sq >= 9) {
if (Cs < 437) {
return 1;
}
else if (Cs >= 437) {
return 8;
}
}
}
else if (Rs >= 403) {
return 1;
}
}
}
else if (negBalance >= 63) {
return 8;
}
}
else if (mu_mv_mag_sq >= 70) {
if (ssDCval < 158) {
return 8;
}
else if (ssDCval >= 158) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 1050) {
if (TSC < 3) {
if (ssDCval < 94) {
if (ssDCval < 81) {
if (posBalance < 48) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (ssDCval < 77) {
if (refDCval < 75) {
return 1;
}
else if (refDCval >= 75) {
return 8;
}
}
else if (ssDCval >= 77) {
return 1;
}
}
}
else if (posBalance >= 48) {
return 8;
}
}
else if (ssDCval >= 81) {
return 1;
}
}
else if (ssDCval >= 94) {
if (Cs < 442) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (Rs < 589) {
return 4;
}
else if (Rs >= 589) {
return 1;
}
}
}
}
else if (Cs >= 442) {
if (MVDiff < 3) {
if (Cs < 466) {
if (Rs < 591) {
return 8;
}
else if (Rs >= 591) {
return 4;
}
}
else if (Cs >= 466) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Rs < 509) {
return 4;
}
else if (Rs >= 509) {
return 1;
}
}
}
}
}
else if (TSC >= 3) {
if (Rs < 333) {
return 2;
}
else if (Rs >= 333) {
if (Rs < 587) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
if (ssDCval < 119) {
return 1;
}
else if (ssDCval >= 119) {
return 8;
}
}
}
else if (Rs >= 587) {
return 8;
}
}
}
}
else if (Rs >= 1050) {
if (Cs < 497) {
if (ssDCval < 93) {
if (Rs < 2070) {
if (Rs < 1115) {
return 2;
}
else if (Rs >= 1115) {
if (diffRsCsdiff < -3) {
if (Rs < 1807) {
return 1;
}
else if (Rs >= 1807) {
return 4;
}
}
else if (diffRsCsdiff >= -3) {
return 1;
}
}
}
else if (Rs >= 2070) {
if (Cs < 463) {
return 2;
}
else if (Cs >= 463) {
return 8;
}
}
}
else if (ssDCval >= 93) {
return 8;
}
}
else if (Cs >= 497) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (Cs < 387) {
if (RsCsDiff < 51) {
if (Cs < 188) {
if (tcor < 36) {
if (posBalance < 65) {
if (tcor < 28) {
return 1;
}
else if (tcor >= 28) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (Rs < 723) {
return 4;
}
else if (Rs >= 723) {
return 8;
}
}
}
}
else if (posBalance >= 65) {
if (Rs < 455) {
return 8;
}
else if (Rs >= 455) {
return 1;
}
}
}
else if (tcor >= 36) {
if (ssDCval < 84) {
if (MVDiff < 15) {
if (Rs < 583) {
return 1;
}
else if (Rs >= 583) {
return 2;
}
}
else if (MVDiff >= 15) {
if (posBalance < 100) {
return 1;
}
else if (posBalance >= 100) {
return 2;
}
}
}
else if (ssDCval >= 84) {
if (mu_mv_mag_sq < 43) {
if (refDCval < 91) {
return 2;
}
else if (refDCval >= 91) {
if (AFD < 12) {
if (Rs < 661) {
return 8;
}
else if (Rs >= 661) {
return 2;
}
}
else if (AFD >= 12) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 43) {
return 2;
}
}
}
}
else if (Cs >= 188) {
if (negBalance < 42) {
if (ssDCval < 96) {
if (Rs < 326) {
return 8;
}
else if (Rs >= 326) {
return 1;
}
}
else if (ssDCval >= 96) {
if (Rs < 484) {
return 4;
}
else if (Rs >= 484) {
return 8;
}
}
}
else if (negBalance >= 42) {
if (diffMVdiffVal < -6) {
if (mcTcor < 87) {
if (gchDC < 7) {
if (TSC < 8) {
if (Cs < 337) {
return 4;
}
else if (Cs >= 337) {
return 2;
}
}
else if (TSC >= 8) {
if (ssDCval < 112) {
return 8;
}
else if (ssDCval >= 112) {
if (Rs < 525) {
return 2;
}
else if (Rs >= 525) {
return 4;
}
}
}
}
else if (gchDC >= 7) {
if (Cs < 335) {
return 1;
}
else if (Cs >= 335) {
return 2;
}
}
}
else if (mcTcor >= 87) {
if (negBalance < 59) {
return 1;
}
else if (negBalance >= 59) {
if (Rs < 415) {
return 1;
}
else if (Rs >= 415) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -6) {
if (AFD < 16) {
if (tcor < 36) {
return 4;
}
else if (tcor >= 36) {
if (posBalance < 58) {
if (SC < 896) {
if (AFD < 13) {
if (Rs < 577) {
return 8;
}
else if (Rs >= 577) {
return 2;
}
}
else if (AFD >= 13) {
return 2;
}
}
else if (SC >= 896) {
if (Cs < 356) {
return 1;
}
else if (Cs >= 356) {
if (Rs < 565) {
return 4;
}
else if (Rs >= 565) {
return 2;
}
}
}
}
else if (posBalance >= 58) {
if (refDCval < 86) {
return 8;
}
else if (refDCval >= 86) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
}
}
else if (AFD >= 16) {
if (Cs < 360) {
if (mu_mv_mag_sq < 8) {
if (mu_mv_mag_sq < 5) {
return 8;
}
else if (mu_mv_mag_sq >= 5) {
return 4;
}
}
else if (mu_mv_mag_sq >= 8) {
if (CsDiff < 4) {
if (ssDCval < 85) {
return 1;
}
else if (ssDCval >= 85) {
if (refDCval < 107) {
return 8;
}
else if (refDCval >= 107) {
return 1;
}
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (Cs >= 360) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 51) {
if (ssDCval < 112) {
if (SC < 1972) {
if (negBalance < 51) {
if (posBalance < 62) {
if (SC < 851) {
return 2;
}
else if (SC >= 851) {
if (mu_mv_mag_sq < 15) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 4;
}
}
else if (mu_mv_mag_sq >= 15) {
if (RsCsDiff < 269) {
return 1;
}
else if (RsCsDiff >= 269) {
return 2;
}
}
}
}
else if (posBalance >= 62) {
if (diffRsCsdiff < 34) {
if (AFD < 21) {
if (Cs < 194) {
if (Rs < 896) {
return 1;
}
else if (Rs >= 896) {
return 2;
}
}
else if (Cs >= 194) {
return 1;
}
}
else if (AFD >= 21) {
return 2;
}
}
else if (diffRsCsdiff >= 34) {
if (Rs < 917) {
return 8;
}
else if (Rs >= 917) {
return 1;
}
}
}
}
else if (negBalance >= 51) {
if (posBalance < 40) {
if (AFD < 46) {
return 1;
}
else if (AFD >= 46) {
return 8;
}
}
else if (posBalance >= 40) {
if (AFD < 26) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
if (AFD < 22) {
if (tcor < 51) {
return 4;
}
else if (tcor >= 51) {
if (Rs < 619) {
return 8;
}
else if (Rs >= 619) {
return 1;
}
}
}
else if (AFD >= 22) {
if (diffTSC < -1) {
if (AFD < 25) {
if (Rs < 1490) {
return 4;
}
else if (Rs >= 1490) {
return 2;
}
}
else if (AFD >= 25) {
if (ssDCval < 85) {
return 1;
}
else if (ssDCval >= 85) {
if (Rs < 1054) {
return 2;
}
else if (Rs >= 1054) {
return 4;
}
}
}
}
else if (diffTSC >= -1) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 2;
}
}
}
}
}
else if (AFD >= 26) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
}
else if (SC >= 1972) {
if (MVDiff < 60) {
if (diffAFD < 7) {
if (SC < 2485) {
return 2;
}
else if (SC >= 2485) {
if (mcTcor < 94) {
return 2;
}
else if (mcTcor >= 94) {
if (AFD < 30) {
return 1;
}
else if (AFD >= 30) {
return 2;
}
}
}
}
else if (diffAFD >= 7) {
return 1;
}
}
else if (MVDiff >= 60) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (AFD < 35) {
return 2;
}
else if (AFD >= 35) {
return 4;
}
}
}
}
}
else if (ssDCval >= 112) {
if (mcTcor < 93) {
if (Rs < 1524) {
if (tcor < 67) {
if (negBalance < 54) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 2;
}
}
else if (negBalance >= 54) {
if (tcor < 60) {
return 2;
}
else if (tcor >= 60) {
if (Cs < 107) {
return 2;
}
else if (Cs >= 107) {
return 1;
}
}
}
}
else if (tcor >= 67) {
if (Rs < 1263) {
return 4;
}
else if (Rs >= 1263) {
return 1;
}
}
}
else if (Rs >= 1524) {
if (diffAFD < 3) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (Rs < 2005) {
return 1;
}
else if (Rs >= 2005) {
return 8;
}
}
}
else if (diffAFD >= 3) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
if (SC < 1100) {
return 2;
}
else if (SC >= 1100) {
return 1;
}
}
}
}
}
else if (Cs >= 387) {
if (posBalance < 53) {
if (AFD < 12) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
return 1;
}
}
else if (AFD >= 12) {
if (diffTSC < 4) {
if (ssDCval < 57) {
if (Cs < 417) {
if (Rs < 1618) {
return 1;
}
else if (Rs >= 1618) {
return 2;
}
}
else if (Cs >= 417) {
if (Cs < 525) {
return 2;
}
else if (Cs >= 525) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (MVDiff < 100) {
if (refDCval < 78) {
if (negBalance < 56) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (negBalance >= 56) {
return 4;
}
}
else if (refDCval >= 78) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (mu_mv_mag_sq < 23) {
return 1;
}
else if (mu_mv_mag_sq >= 23) {
return 2;
}
}
}
}
else if (MVDiff >= 100) {
if (Rs < 482) {
return 8;
}
else if (Rs >= 482) {
return 1;
}
}
}
}
else if (diffTSC >= 4) {
if (Rs < 791) {
if (Rs < 679) {
return 4;
}
else if (Rs >= 679) {
return 1;
}
}
else if (Rs >= 791) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
if (AFD < 23) {
if (ssDCval < 126) {
if (diffAFD < 2) {
if (negBalance < 46) {
if (SC < 1055) {
if (Rs < 451) {
return 1;
}
else if (Rs >= 451) {
return 8;
}
}
else if (SC >= 1055) {
if (Cs < 502) {
return 2;
}
else if (Cs >= 502) {
return 8;
}
}
}
else if (negBalance >= 46) {
if (tcor < 61) {
if (AFD < 17) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (AFD >= 17) {
if (Rs < 610) {
if (Rs < 524) {
return 4;
}
else if (Rs >= 524) {
return 1;
}
}
else if (Rs >= 610) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
else if (tcor >= 61) {
if (diffRsCsdiff < 16) {
return 8;
}
else if (diffRsCsdiff >= 16) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
}
else if (diffAFD >= 2) {
if (negBalance < 55) {
if (TSC < 7) {
if (refDCval < 64) {
if (posBalance < 67) {
return 1;
}
else if (posBalance >= 67) {
if (Rs < 568) {
return 4;
}
else if (Rs >= 568) {
if (Rs < 710) {
return 8;
}
else if (Rs >= 710) {
return 1;
}
}
}
}
else if (refDCval >= 64) {
if (Rs < 1005) {
return 2;
}
else if (Rs >= 1005) {
return 8;
}
}
}
else if (TSC >= 7) {
if (Cs < 478) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
else if (Cs >= 478) {
return 8;
}
}
}
else if (negBalance >= 55) {
if (posBalance < 63) {
return 4;
}
else if (posBalance >= 63) {
if (Rs < 754) {
return 2;
}
else if (Rs >= 754) {
return 1;
}
}
}
}
}
else if (ssDCval >= 126) {
if (Rs < 1459) {
if (RsCsDiff < 37) {
return 1;
}
else if (RsCsDiff >= 37) {
return 2;
}
}
else if (Rs >= 1459) {
if (Rs < 1991) {
return 4;
}
else if (Rs >= 1991) {
return 1;
}
}
}
}
else if (AFD >= 23) {
if (MVDiff < 38) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Cs < 520) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 2;
}
}
else if (Cs >= 520) {
if (Rs < 1679) {
return 4;
}
else if (Rs >= 1679) {
return 1;
}
}
}
}
else if (MVDiff >= 38) {
if (mu_mv_mag_sq < 62) {
if (MVDiff < 181) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (diffRsCsdiff < -58) {
return 8;
}
else if (diffRsCsdiff >= -58) {
if (AFD < 28) {
if (TSC < 10) {
if (Rs < 501) {
return 2;
}
else if (Rs >= 501) {
return 1;
}
}
else if (TSC >= 10) {
return 4;
}
}
else if (AFD >= 28) {
return 1;
}
}
}
}
else if (MVDiff >= 181) {
if (MVDiff < 287) {
return 8;
}
else if (MVDiff >= 287) {
if (Rs < 2016) {
return 2;
}
else if (Rs >= 2016) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 62) {
if (AFD < 28) {
if (Rs < 534) {
return 2;
}
else if (Rs >= 534) {
return 8;
}
}
else if (AFD >= 28) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 81) {
if (RsCsDiff < 84) {
if (ssDCval < 139) {
if (Cs < 229) {
if (refDCval < 80) {
if (mcTcor < 99) {
if (MVDiff < 187) {
return 1;
}
else if (MVDiff >= 187) {
return 8;
}
}
else if (mcTcor >= 99) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (Cs < 201) {
return 1;
}
else if (Cs >= 201) {
return 4;
}
}
}
}
else if (refDCval >= 80) {
if (RsCsDiff < 28) {
if (Rs < 765) {
if (MVDiff < 41) {
return 1;
}
else if (MVDiff >= 41) {
if (MVDiff < 156) {
return 8;
}
else if (MVDiff >= 156) {
return 2;
}
}
}
else if (Rs >= 765) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 95) {
return 1;
}
else if (ssDCval >= 95) {
if (ssDCval < 97) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
else if (ssDCval >= 97) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 7) {
return 1;
}
}
}
else if (RsCsDiff >= 28) {
if (AFD < 15) {
if (ssDCval < 118) {
if (tcor < 82) {
return 2;
}
else if (tcor >= 82) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (refDCval < 94) {
if (Cs < 117) {
return 2;
}
else if (Cs >= 117) {
return 1;
}
}
else if (refDCval >= 94) {
if (MVDiff < 63) {
if (Cs < 187) {
return 2;
}
else if (Cs >= 187) {
if (Rs < 1403) {
return 1;
}
else if (Rs >= 1403) {
return 2;
}
}
}
else if (MVDiff >= 63) {
if (Rs < 1573) {
return 8;
}
else if (Rs >= 1573) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 118) {
return 1;
}
}
else if (AFD >= 15) {
if (negBalance < 82) {
if (ssDCval < 80) {
if (Rs < 940) {
return 2;
}
else if (Rs >= 940) {
return 1;
}
}
else if (ssDCval >= 80) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
if (Rs < 1552) {
return 1;
}
else if (Rs >= 1552) {
return 8;
}
}
}
}
else if (negBalance >= 82) {
return 2;
}
}
}
}
}
else if (Cs >= 229) {
if (Rs < 624) {
if (MVDiff < 23) {
if (mcTcor < 97) {
if (SC < 861) {
return 8;
}
else if (SC >= 861) {
return 1;
}
}
else if (mcTcor >= 97) {
if (Rs < 421) {
return 8;
}
else if (Rs >= 421) {
return 4;
}
}
}
else if (MVDiff >= 23) {
if (negBalance < 47) {
if (mu_mv_mag_sq < 47) {
return 2;
}
else if (mu_mv_mag_sq >= 47) {
if (Rs < 473) {
return 1;
}
else if (Rs >= 473) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (tcor < 90) {
if (diffTSC < -1) {
if (Cs < 370) {
if (SC < 757) {
return 8;
}
else if (SC >= 757) {
return 4;
}
}
else if (Cs >= 370) {
return 8;
}
}
else if (diffTSC >= -1) {
if (diffMVdiffVal < 11) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (diffMVdiffVal >= 11) {
return 1;
}
}
}
else if (tcor >= 90) {
return 2;
}
}
}
}
else if (Rs >= 624) {
if (Cs < 447) {
if (Rs < 1932) {
if (Rs < 927) {
if (tcor < 89) {
if (Cs < 437) {
if (diffRsCsdiff < 12) {
if (negBalance < 48) {
if (Rs < 915) {
return 1;
}
else if (Rs >= 915) {
return 4;
}
}
else if (negBalance >= 48) {
if (diffRsCsdiff < 6) {
if (diffAFD < -3) {
if (Rs < 799) {
return 4;
}
else if (Rs >= 799) {
return 8;
}
}
else if (diffAFD >= -3) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 6) {
return 4;
}
}
}
else if (diffRsCsdiff >= 12) {
return 8;
}
}
else if (Cs >= 437) {
return 4;
}
}
else if (tcor >= 89) {
return 1;
}
}
else if (Rs >= 927) {
if (RsCsDiff < 26) {
if (ssDCval < 98) {
if (tcor < 98) {
if (Rs < 930) {
return 2;
}
else if (Rs >= 930) {
if (mcTcor < 90) {
if (Rs < 1025) {
return 4;
}
else if (Rs >= 1025) {
return 1;
}
}
else if (mcTcor >= 90) {
return 1;
}
}
}
else if (tcor >= 98) {
if (Rs < 1055) {
return 8;
}
else if (Rs >= 1055) {
return 4;
}
}
}
else if (ssDCval >= 98) {
if (TSC < 8) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
return 1;
}
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (RsCsDiff >= 26) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (Rs < 1259) {
if (AFD < 12) {
if (SC < 1552) {
return 1;
}
else if (SC >= 1552) {
if (Rs < 1235) {
return 2;
}
else if (Rs >= 1235) {
if (Rs < 1252) {
return 1;
}
else if (Rs >= 1252) {
return 4;
}
}
}
}
else if (AFD >= 12) {
if (Cs < 431) {
if (tcor < 86) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (tcor >= 86) {
if (Rs < 1048) {
return 1;
}
else if (Rs >= 1048) {
return 2;
}
}
}
else if (Cs >= 431) {
return 4;
}
}
}
else if (Rs >= 1259) {
if (tcor < 84) {
if (mu_mv_mag_sq < 27) {
return 1;
}
else if (mu_mv_mag_sq >= 27) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
else if (tcor >= 84) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1932) {
if (ssDCval < 94) {
if (ssDCval < 66) {
if (SC < 2508) {
return 2;
}
else if (SC >= 2508) {
return 4;
}
}
else if (ssDCval >= 66) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (diffRsCsdiff < -47) {
return 4;
}
else if (diffRsCsdiff >= -47) {
return 1;
}
}
}
}
else if (ssDCval >= 94) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (tcor < 91) {
if (diffMVdiffVal < -12) {
if (MVDiff < 57) {
return 8;
}
else if (MVDiff >= 57) {
return 1;
}
}
else if (diffMVdiffVal >= -12) {
return 8;
}
}
else if (tcor >= 91) {
return 1;
}
}
}
}
}
else if (Cs >= 447) {
if (ssDCval < 66) {
if (Rs < 1166) {
if (refDCval < 52) {
if (SC < 1605) {
return 2;
}
else if (SC >= 1605) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (refDCval >= 52) {
if (posBalance < 53) {
if (negBalance < 31) {
if (Rs < 951) {
return 2;
}
else if (Rs >= 951) {
return 4;
}
}
else if (negBalance >= 31) {
if (posBalance < 4) {
return 4;
}
else if (posBalance >= 4) {
return 8;
}
}
}
else if (posBalance >= 53) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
else if (Rs >= 1166) {
if (ssDCval < 57) {
if (negBalance < 30) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 2;
}
}
else if (negBalance >= 30) {
if (Rs < 1224) {
return 2;
}
else if (Rs >= 1224) {
return 1;
}
}
}
else if (ssDCval >= 57) {
return 2;
}
}
}
else if (ssDCval >= 66) {
if (ssDCval < 123) {
if (mu_mv_mag_sq < 3) {
if (RsCsDiff < 53) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (ssDCval < 108) {
return 1;
}
else if (ssDCval >= 108) {
return 8;
}
}
}
else if (RsCsDiff >= 53) {
if (mcTcor < 96) {
if (ssDCval < 84) {
if (Cs < 512) {
return 1;
}
else if (Cs >= 512) {
return 2;
}
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (posBalance < 57) {
if (ssDCval < 96) {
if (MVDiff < 2) {
if (Rs < 1644) {
return 4;
}
else if (Rs >= 1644) {
return 1;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 96) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 1;
}
}
}
else if (posBalance >= 57) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (AFD < 19) {
if (Rs < 779) {
return 4;
}
else if (Rs >= 779) {
return 1;
}
}
else if (AFD >= 19) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 123) {
if (AFD < 13) {
if (diffTSC < -1) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (AFD >= 13) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 139) {
if (gchDC < 2) {
if (refDCval < 146) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 4;
}
}
else if (refDCval >= 146) {
if (ssDCval < 173) {
if (MVDiff < 83) {
return 1;
}
else if (MVDiff >= 83) {
return 8;
}
}
else if (ssDCval >= 173) {
if (Cs < 295) {
return 2;
}
else if (Cs >= 295) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (RsDiff < 6) {
if (ssDCval < 164) {
if (Cs < 271) {
return 1;
}
else if (Cs >= 271) {
if (ssDCval < 154) {
return 2;
}
else if (ssDCval >= 154) {
return 1;
}
}
}
else if (ssDCval >= 164) {
if (Rs < 555) {
return 1;
}
else if (Rs >= 555) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (Cs < 125) {
return 2;
}
else if (Cs >= 125) {
if (mcTcor < 95) {
if (Rs < 1490) {
return 1;
}
else if (Rs >= 1490) {
return 8;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 84) {
if (refDCval < 109) {
if (MVDiff < 37) {
if (negBalance < 48) {
if (refDCval < 74) {
if (TSC < 6) {
if (tcor < 79) {
return 1;
}
else if (tcor >= 79) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
return 2;
}
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (refDCval >= 74) {
if (ssDCval < 91) {
if (MVDiff < 28) {
return 1;
}
else if (MVDiff >= 28) {
return 4;
}
}
else if (ssDCval >= 91) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
if (Cs < 486) {
return 2;
}
else if (Cs >= 486) {
return 1;
}
}
}
}
}
else if (negBalance >= 48) {
if (Rs < 2018) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (tcor < 86) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (Cs < 422) {
if (Rs < 1475) {
return 1;
}
else if (Rs >= 1475) {
return 2;
}
}
else if (Cs >= 422) {
return 1;
}
}
}
else if (tcor >= 86) {
return 1;
}
}
}
else if (Rs >= 2018) {
return 2;
}
}
}
else if (MVDiff >= 37) {
if (mcTcor < 96) {
if (Cs < 250) {
if (diffRsCsdiff < -133) {
return 2;
}
else if (diffRsCsdiff >= -133) {
if (SC < 2439) {
return 1;
}
else if (SC >= 2439) {
return 4;
}
}
}
else if (Cs >= 250) {
if (mu_mv_mag_sq < 7) {
if (MVDiff < 44) {
return 2;
}
else if (MVDiff >= 44) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (gchDC < 2) {
if (Rs < 2244) {
return 4;
}
else if (Rs >= 2244) {
return 8;
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (RsDiff < 13) {
if (diffMVdiffVal < -46) {
return 2;
}
else if (diffMVdiffVal >= -46) {
if (Cs < 476) {
return 1;
}
else if (Cs >= 476) {
return 2;
}
}
}
else if (RsDiff >= 13) {
return 2;
}
}
}
}
else if (refDCval >= 109) {
if (TSC < 6) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (TSC >= 6) {
if (CsDiff < 3) {
if (Rs < 2036) {
if (AFD < 13) {
if (Rs < 1550) {
return 4;
}
else if (Rs >= 1550) {
return 1;
}
}
else if (AFD >= 13) {
return 1;
}
}
else if (Rs >= 2036) {
if (MVDiff < 61) {
if (Rs < 2109) {
if (Rs < 2067) {
return 1;
}
else if (Rs >= 2067) {
return 2;
}
}
else if (Rs >= 2109) {
return 1;
}
}
else if (MVDiff >= 61) {
if (Rs < 2107) {
return 8;
}
else if (Rs >= 2107) {
if (MVDiff < 64) {
if (Cs < 352) {
return 1;
}
else if (Cs >= 352) {
return 8;
}
}
else if (MVDiff >= 64) {
return 1;
}
}
}
}
}
else if (CsDiff >= 3) {
if (AFD < 34) {
if (mcTcor < 95) {
if (RsCsDiff < 94) {
if (Rs < 1349) {
return 2;
}
else if (Rs >= 1349) {
if (Rs < 1823) {
return 8;
}
else if (Rs >= 1823) {
return 1;
}
}
}
else if (RsCsDiff >= 94) {
if (negBalance < 37) {
if (Rs < 1891) {
return 2;
}
else if (Rs >= 1891) {
return 1;
}
}
else if (negBalance >= 37) {
return 1;
}
}
}
else if (mcTcor >= 95) {
return 1;
}
}
else if (AFD >= 34) {
if (Cs < 442) {
return 1;
}
else if (Cs >= 442) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (AFD < 9) {
if (ssDCval < 59) {
if (Rs < 1112) {
return 4;
}
else if (Rs >= 1112) {
return 1;
}
}
else if (ssDCval >= 59) {
if (Cs < 510) {
return 8;
}
else if (Cs >= 510) {
return 1;
}
}
}
else if (AFD >= 9) {
if (Rs < 736) {
if (Cs < 322) {
return 1;
}
else if (Cs >= 322) {
return 8;
}
}
else if (Rs >= 736) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Cs >= 533) {
if (negBalance < 13) {
if (Rs < 699) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 117) {
if (ssDCval < 30) {
return 8;
}
else if (ssDCval >= 30) {
if (ssDCval < 56) {
if (Rs < 246) {
if (Cs < 1020) {
return 2;
}
else if (Cs >= 1020) {
return 1;
}
}
else if (Rs >= 246) {
return 1;
}
}
else if (ssDCval >= 56) {
if (Rs < 437) {
return 4;
}
else if (Rs >= 437) {
if (ssDCval < 104) {
if (mu_mv_mag_sq < 114) {
if (Rs < 674) {
return 1;
}
else if (Rs >= 674) {
return 8;
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (ssDCval >= 104) {
return 1;
}
}
}
}
}
else if (ssDCval >= 117) {
if (ssDCval < 153) {
if (Rs < 514) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Rs < 490) {
return 1;
}
else if (Rs >= 490) {
return 4;
}
}
}
else if (Rs >= 514) {
return 8;
}
}
else if (ssDCval >= 153) {
if (Rs < 448) {
return 1;
}
else if (Rs >= 448) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (posBalance < 37) {
if (posBalance < 28) {
if (ssDCval < 107) {
if (Rs < 364) {
return 1;
}
else if (Rs >= 364) {
return 8;
}
}
else if (ssDCval >= 107) {
return 1;
}
}
else if (posBalance >= 28) {
return 8;
}
}
else if (posBalance >= 37) {
if (Cs < 695) {
return 8;
}
else if (Cs >= 695) {
return 2;
}
}
}
}
}
else if (Rs >= 699) {
if (diffTSC < -2) {
if (SC < 2304) {
if (Cs < 653) {
if (SC < 1480) {
return 2;
}
else if (SC >= 1480) {
if (Rs < 1215) {
return 8;
}
else if (Rs >= 1215) {
if (Rs < 1489) {
return 1;
}
else if (Rs >= 1489) {
return 8;
}
}
}
}
else if (Cs >= 653) {
if (mu_mv_mag_sq < 125) {
if (SC < 2160) {
return 8;
}
else if (SC >= 2160) {
return 1;
}
}
else if (mu_mv_mag_sq >= 125) {
if (diffAFD < -8) {
if (MVDiff < 19) {
if (Rs < 1056) {
return 1;
}
else if (Rs >= 1056) {
return 8;
}
}
else if (MVDiff >= 19) {
if (ssDCval < 140) {
if (Cs < 960) {
return 4;
}
else if (Cs >= 960) {
return 8;
}
}
else if (ssDCval >= 140) {
if (Rs < 1328) {
return 8;
}
else if (Rs >= 1328) {
return 4;
}
}
}
}
else if (diffAFD >= -8) {
if (tcor < 99) {
if (Cs < 853) {
return 1;
}
else if (Cs >= 853) {
return 8;
}
}
else if (tcor >= 99) {
if (Cs < 1005) {
return 4;
}
else if (Cs >= 1005) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 2304) {
if (MVDiff < 16) {
if (Rs < 1116) {
return 1;
}
else if (Rs >= 1116) {
return 8;
}
}
else if (MVDiff >= 16) {
if (refDCval < 128) {
if (Rs < 1780) {
return 8;
}
else if (Rs >= 1780) {
return 4;
}
}
else if (refDCval >= 128) {
if (MVDiff < 80) {
if (ssDCval < 144) {
if (MVDiff < 18) {
if (Rs < 1538) {
return 8;
}
else if (Rs >= 1538) {
return 4;
}
}
else if (MVDiff >= 18) {
return 4;
}
}
else if (ssDCval >= 144) {
if (Rs < 1426) {
return 4;
}
else if (Rs >= 1426) {
return 8;
}
}
}
else if (MVDiff >= 80) {
if (Cs < 1004) {
return 4;
}
else if (Cs >= 1004) {
return 1;
}
}
}
}
}
}
else if (diffTSC >= -2) {
if (Rs < 1840) {
if (Rs < 967) {
if (diffMVdiffVal < -1) {
if (posBalance < 29) {
if (ssDCval < 32) {
return 8;
}
else if (ssDCval >= 32) {
if (mu_mv_mag_sq < 127) {
return 2;
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
}
else if (posBalance >= 29) {
if (SC < 1752) {
return 8;
}
else if (SC >= 1752) {
return 1;
}
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 10) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
return 2;
}
}
else if (posBalance >= 10) {
if (negBalance < 10) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (negBalance >= 10) {
if (Rs < 745) {
if (posBalance < 27) {
if (Cs < 730) {
if (posBalance < 26) {
return 1;
}
else if (posBalance >= 26) {
if (Rs < 732) {
return 1;
}
else if (Rs >= 732) {
return 8;
}
}
}
else if (Cs >= 730) {
return 2;
}
}
else if (posBalance >= 27) {
if (Rs < 713) {
return 1;
}
else if (Rs >= 713) {
return 8;
}
}
}
else if (Rs >= 745) {
if (AFD < 2) {
if (SC < 1463) {
if (ssDCval < 81) {
if (Rs < 882) {
return 2;
}
else if (Rs >= 882) {
return 8;
}
}
else if (ssDCval >= 81) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
}
else if (SC >= 1463) {
return 8;
}
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
}
else if (Rs >= 967) {
if (SC < 1804) {
return 8;
}
else if (SC >= 1804) {
if (negBalance < 10) {
if (ssDCval < 152) {
if (tcor < 96) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (MVDiff < 1) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (MVDiff >= 1) {
if (AFD < 7) {
if (Cs < 847) {
return 8;
}
else if (Cs >= 847) {
return 1;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
}
else if (tcor >= 96) {
if (diffMVdiffVal < -1) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (AFD < 1) {
if (Rs < 1124) {
return 4;
}
else if (Rs >= 1124) {
return 1;
}
}
else if (AFD >= 1) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (negBalance >= 10) {
if (ssDCval < 57) {
if (MVDiff < 4) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
if (ssDCval < 53) {
if (Cs < 682) {
return 1;
}
else if (Cs >= 682) {
return 8;
}
}
else if (ssDCval >= 53) {
return 1;
}
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (ssDCval >= 57) {
if (ssDCval < 66) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (ssDCval >= 66) {
if (SC < 2798) {
if (tcor < 86) {
if (Rs < 1075) {
return 8;
}
else if (Rs >= 1075) {
if (Rs < 1230) {
return 2;
}
else if (Rs >= 1230) {
return 4;
}
}
}
else if (tcor >= 86) {
return 8;
}
}
else if (SC >= 2798) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 1840) {
if (Rs < 2044) {
if (Cs < 986) {
if (ssDCval < 81) {
if (Cs < 656) {
if (Rs < 1926) {
return 8;
}
else if (Rs >= 1926) {
return 2;
}
}
else if (Cs >= 656) {
if (SC < 2929) {
return 1;
}
else if (SC >= 2929) {
if (Rs < 2009) {
return 1;
}
else if (Rs >= 2009) {
return 4;
}
}
}
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (Cs >= 986) {
if (refDCval < 69) {
if (Cs < 1002) {
if (Rs < 1958) {
return 1;
}
else if (Rs >= 1958) {
return 4;
}
}
else if (Cs >= 1002) {
return 4;
}
}
else if (refDCval >= 69) {
if (Rs < 1922) {
return 1;
}
else if (Rs >= 1922) {
if (Rs < 1991) {
return 8;
}
else if (Rs >= 1991) {
return 1;
}
}
}
}
}
else if (Rs >= 2044) {
if (Cs < 1289) {
if (posBalance < 38) {
if (mu_mv_mag_sq < 94) {
return 4;
}
else if (mu_mv_mag_sq >= 94) {
return 8;
}
}
else if (posBalance >= 38) {
if (Rs < 2164) {
return 1;
}
else if (Rs >= 2164) {
return 8;
}
}
}
else if (Cs >= 1289) {
if (Rs < 2227) {
return 8;
}
else if (Rs >= 2227) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 43) {
if (refDCval < 47) {
if (Rs < 1686) {
if (refDCval < 29) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
return 4;
}
}
else if (refDCval >= 29) {
if (negBalance < 21) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (Cs < 604) {
return 2;
}
else if (Cs >= 604) {
return 1;
}
}
}
else if (negBalance >= 21) {
if (mcTcor < 92) {
if (negBalance < 25) {
if (RsCsDiff < 37) {
return 2;
}
else if (RsCsDiff >= 37) {
if (Rs < 857) {
return 1;
}
else if (Rs >= 857) {
return 8;
}
}
}
else if (negBalance >= 25) {
if (diffMVdiffVal < -3) {
if (Rs < 1354) {
if (Rs < 659) {
return 1;
}
else if (Rs >= 659) {
return 8;
}
}
else if (Rs >= 1354) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (Rs < 1456) {
if (Cs < 827) {
if (Cs < 733) {
if (Rs < 1191) {
return 8;
}
else if (Rs >= 1191) {
return 4;
}
}
else if (Cs >= 733) {
return 4;
}
}
else if (Cs >= 827) {
return 8;
}
}
else if (Rs >= 1456) {
return 4;
}
}
}
}
}
else if (Rs >= 1686) {
if (MVDiff < 7) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (MVDiff >= 7) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 4;
}
}
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 66) {
if (ssDCval < 84) {
if (Cs < 891) {
if (Rs < 1366) {
if (negBalance < 36) {
if (Rs < 766) {
return 4;
}
else if (Rs >= 766) {
if (Rs < 1290) {
return 1;
}
else if (Rs >= 1290) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (Cs < 725) {
if (Cs < 636) {
if (Rs < 452) {
return 4;
}
else if (Rs >= 452) {
return 8;
}
}
else if (Cs >= 636) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (Cs >= 725) {
if (Rs < 721) {
return 2;
}
else if (Rs >= 721) {
return 8;
}
}
}
}
else if (Rs >= 1366) {
if (ssDCval < 69) {
if (gchDC < 1) {
if (Rs < 1614) {
if (Rs < 1452) {
return 8;
}
else if (Rs >= 1452) {
return 2;
}
}
else if (Rs >= 1614) {
if (Rs < 1923) {
return 1;
}
else if (Rs >= 1923) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (ssDCval < 67) {
if (tcor < 84) {
if (Rs < 1754) {
return 2;
}
else if (Rs >= 1754) {
return 1;
}
}
else if (tcor >= 84) {
return 1;
}
}
else if (ssDCval >= 67) {
return 2;
}
}
}
else if (ssDCval >= 69) {
if (SC < 3045) {
return 1;
}
else if (SC >= 3045) {
return 2;
}
}
}
}
else if (Cs >= 891) {
if (mcTcor < 93) {
if (Rs < 1952) {
return 8;
}
else if (Rs >= 1952) {
if (Cs < 1094) {
return 4;
}
else if (Cs >= 1094) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (tcor < 90) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (tcor >= 90) {
if (Cs < 979) {
if (TSC < 5) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (Rs < 931) {
return 4;
}
else if (Rs >= 931) {
return 8;
}
}
}
else if (TSC >= 5) {
if (Rs < 2285) {
return 2;
}
else if (Rs >= 2285) {
return 1;
}
}
}
else if (Cs >= 979) {
if (posBalance < 37) {
if (Cs < 1103) {
return 2;
}
else if (Cs >= 1103) {
if (Rs < 1225) {
return 4;
}
else if (Rs >= 1225) {
return 8;
}
}
}
else if (posBalance >= 37) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (posBalance < 40) {
if (ssDCval < 152) {
if (refDCval < 88) {
return 2;
}
else if (refDCval >= 88) {
if (TSC < 7) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
else if (TSC >= 7) {
if (Rs < 820) {
return 8;
}
else if (Rs >= 820) {
if (Rs < 2225) {
return 2;
}
else if (Rs >= 2225) {
return 8;
}
}
}
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (posBalance >= 40) {
if (tcor < 89) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
else if (tcor >= 89) {
if (Rs < 1521) {
return 4;
}
else if (Rs >= 1521) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (ssDCval < 58) {
if (TSC < 3) {
if (negBalance < 25) {
if (Rs < 841) {
if (Rs < 442) {
return 2;
}
else if (Rs >= 442) {
return 1;
}
}
else if (Rs >= 841) {
if (Cs < 1030) {
return 8;
}
else if (Cs >= 1030) {
return 4;
}
}
}
else if (negBalance >= 25) {
return 4;
}
}
else if (TSC >= 3) {
if (mcTcor < 95) {
if (Cs < 598) {
return 1;
}
else if (Cs >= 598) {
if (AFD < 10) {
if (ssDCval < 51) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (ssDCval >= 51) {
return 8;
}
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 49) {
if (Rs < 1078) {
return 4;
}
else if (Rs >= 1078) {
if (Rs < 1213) {
return 8;
}
else if (Rs >= 1213) {
return 1;
}
}
}
else if (ssDCval >= 49) {
if (posBalance < 31) {
return 4;
}
else if (posBalance >= 31) {
return 1;
}
}
}
}
}
else if (ssDCval >= 58) {
if (mu_mv_mag_sq < 94) {
if (posBalance < 36) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (SC < 2919) {
if (MVDiff < 10) {
if (Cs < 905) {
return 4;
}
else if (Cs >= 905) {
return 1;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (SC >= 2919) {
return 8;
}
}
}
}
else if (posBalance >= 36) {
if (negBalance < 29) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (negBalance >= 29) {
if (MVDiff < 19) {
if (Cs < 1021) {
if (mu_mv_mag_sq < 78) {
if (Rs < 1475) {
return 8;
}
else if (Rs >= 1475) {
if (Rs < 2198) {
return 1;
}
else if (Rs >= 2198) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (tcor < 92) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 2;
}
}
else if (tcor >= 92) {
if (Rs < 1308) {
return 2;
}
else if (Rs >= 1308) {
return 1;
}
}
}
}
else if (Cs >= 1021) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 8;
}
}
}
else if (MVDiff >= 19) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (RsCsDiff < 41) {
if (Rs < 691) {
if (MVDiff < 4) {
if (Rs < 618) {
return 2;
}
else if (Rs >= 618) {
return 8;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (Rs >= 691) {
if (negBalance < 14) {
if (Rs < 1136) {
return 1;
}
else if (Rs >= 1136) {
return 8;
}
}
else if (negBalance >= 14) {
if (SC < 2718) {
if (Rs < 1003) {
if (Cs < 779) {
return 8;
}
else if (Cs >= 779) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (Rs >= 1003) {
return 8;
}
}
else if (SC >= 2718) {
if (SC < 2764) {
if (Rs < 1724) {
return 4;
}
else if (Rs >= 1724) {
return 1;
}
}
else if (SC >= 2764) {
if (Cs < 810) {
if (Rs < 2276) {
return 4;
}
else if (Rs >= 2276) {
return 8;
}
}
else if (Cs >= 810) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 41) {
if (CsDiff < 5) {
if (Rs < 1795) {
return 8;
}
else if (Rs >= 1795) {
return 2;
}
}
else if (CsDiff >= 5) {
return 1;
}
}
}
}
}
}
}
else if (posBalance >= 43) {
if (mu_mv_mag_sq < 32) {
if (negBalance < 47) {
if (ssDCval < 58) {
if (refDCval < 44) {
if (negBalance < 38) {
if (mu_mv_mag_sq < 7) {
if (RsCsDiff < 57) {
if (posBalance < 72) {
return 8;
}
else if (posBalance >= 72) {
return 2;
}
}
else if (RsCsDiff >= 57) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 6) {
if (Rs < 898) {
return 2;
}
else if (Rs >= 898) {
return 8;
}
}
else if (AFD >= 6) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
}
}
else if (negBalance >= 38) {
if (Rs < 804) {
if (AFD < 9) {
if (negBalance < 43) {
if (posBalance < 60) {
return 2;
}
else if (posBalance >= 60) {
if (Rs < 644) {
return 8;
}
else if (Rs >= 644) {
return 2;
}
}
}
else if (negBalance >= 43) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
}
else if (AFD >= 9) {
return 2;
}
}
else if (Rs >= 804) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
if (SC < 2028) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
else if (SC >= 2028) {
return 8;
}
}
}
}
}
else if (refDCval >= 44) {
if (negBalance < 32) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
if (SC < 1382) {
return 1;
}
else if (SC >= 1382) {
return 8;
}
}
}
else if (negBalance >= 32) {
if (Cs < 586) {
if (Cs < 578) {
if (SC < 1212) {
if (Cs < 574) {
return 8;
}
else if (Cs >= 574) {
return 2;
}
}
else if (SC >= 1212) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
else if (Cs >= 578) {
return 2;
}
}
else if (Cs >= 586) {
if (tcor < 91) {
if (Cs < 1068) {
if (SC < 1849) {
if (CsDiff < 4) {
if (Cs < 600) {
return 4;
}
else if (Cs >= 600) {
if (Cs < 968) {
return 8;
}
else if (Cs >= 968) {
return 2;
}
}
}
else if (CsDiff >= 4) {
if (refDCval < 50) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
else if (refDCval >= 50) {
return 2;
}
}
}
else if (SC >= 1849) {
if (Cs < 684) {
return 4;
}
else if (Cs >= 684) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 4;
}
}
}
}
else if (Cs >= 1068) {
if (negBalance < 46) {
if (refDCval < 52) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
if (Rs < 1563) {
return 4;
}
else if (Rs >= 1563) {
return 8;
}
}
}
else if (refDCval >= 52) {
if (Rs < 900) {
if (Cs < 1135) {
return 2;
}
else if (Cs >= 1135) {
return 8;
}
}
else if (Rs >= 900) {
if (Rs < 2241) {
return 4;
}
else if (Rs >= 2241) {
return 8;
}
}
}
}
else if (negBalance >= 46) {
if (tcor < 75) {
if (Rs < 1122) {
return 2;
}
else if (Rs >= 1122) {
return 1;
}
}
else if (tcor >= 75) {
return 4;
}
}
}
}
else if (tcor >= 91) {
if (Rs < 1070) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (Rs >= 1070) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 58) {
if (refDCval < 101) {
if (Cs < 719) {
if (posBalance < 49) {
if (ssDCval < 72) {
if (MVDiff < 6) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (MVDiff >= 6) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (Rs < 1923) {
return 4;
}
else if (Rs >= 1923) {
return 2;
}
}
}
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (posBalance >= 49) {
if (RsDiff < 3) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (ssDCval < 80) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (Rs < 756) {
if (Rs < 647) {
return 4;
}
else if (Rs >= 647) {
return 2;
}
}
else if (Rs >= 756) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (ssDCval >= 80) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
else if (RsDiff >= 3) {
if (refDCval < 68) {
if (Rs < 1457) {
if (Cs < 651) {
if (mu_mv_mag_sq < 6) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
else if (mu_mv_mag_sq >= 6) {
if (Rs < 979) {
return 2;
}
else if (Rs >= 979) {
return 8;
}
}
}
else if (Cs >= 651) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
return 4;
}
}
}
else if (Rs >= 1457) {
return 2;
}
}
else if (refDCval >= 68) {
if (Rs < 762) {
return 2;
}
else if (Rs >= 762) {
if (mcTcor < 94) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 2;
}
}
else if (mcTcor >= 94) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1510) {
return 4;
}
else if (Rs >= 1510) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 719) {
if (Rs < 2251) {
if (posBalance < 55) {
if (Cs < 1032) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
if (SC < 1557) {
if (ssDCval < 67) {
if (Rs < 418) {
return 2;
}
else if (Rs >= 418) {
return 4;
}
}
else if (ssDCval >= 67) {
if (Rs < 553) {
return 8;
}
else if (Rs >= 553) {
return 1;
}
}
}
else if (SC >= 1557) {
if (RsDiff < 9) {
if (negBalance < 35) {
if (Rs < 1256) {
return 4;
}
else if (Rs >= 1256) {
return 8;
}
}
else if (negBalance >= 35) {
return 4;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
if (Rs < 2049) {
return 1;
}
else if (Rs >= 2049) {
return 4;
}
}
}
}
else if (Cs >= 1032) {
if (Rs < 1993) {
if (Rs < 950) {
return 2;
}
else if (Rs >= 950) {
return 8;
}
}
else if (Rs >= 1993) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
if (Cs < 1182) {
return 2;
}
else if (Cs >= 1182) {
return 8;
}
}
}
}
}
else if (posBalance >= 55) {
if (RsDiff < 3) {
if (refDCval < 57) {
return 1;
}
else if (refDCval >= 57) {
if (Cs < 768) {
if (Rs < 797) {
return 1;
}
else if (Rs >= 797) {
return 8;
}
}
else if (Cs >= 768) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (ssDCval < 70) {
if (Rs < 1048) {
if (diffAFD < 1) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Cs < 964) {
return 4;
}
else if (Cs >= 964) {
return 1;
}
}
}
else if (diffAFD >= 1) {
if (Cs < 797) {
return 1;
}
else if (Cs >= 797) {
if (Cs < 1191) {
return 8;
}
else if (Cs >= 1191) {
return 2;
}
}
}
}
else if (Rs >= 1048) {
if (posBalance < 61) {
if (Rs < 1794) {
return 1;
}
else if (Rs >= 1794) {
return 2;
}
}
else if (posBalance >= 61) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
return 8;
}
}
}
}
else if (ssDCval >= 70) {
if (posBalance < 69) {
if (RsCsDiff < 43) {
if (negBalance < 45) {
if (negBalance < 40) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 2;
}
}
else if (negBalance >= 40) {
if (Rs < 820) {
return 2;
}
else if (Rs >= 820) {
return 4;
}
}
}
else if (negBalance >= 45) {
if (Rs < 1112) {
if (Rs < 678) {
return 1;
}
else if (Rs >= 678) {
return 2;
}
}
else if (Rs >= 1112) {
return 8;
}
}
}
else if (RsCsDiff >= 43) {
if (TSC < 8) {
if (SC < 3081) {
if (tcor < 83) {
if (Rs < 1379) {
return 8;
}
else if (Rs >= 1379) {
return 2;
}
}
else if (tcor >= 83) {
return 2;
}
}
else if (SC >= 3081) {
return 1;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (posBalance >= 69) {
return 8;
}
}
}
}
}
else if (Rs >= 2251) {
if (refDCval < 71) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 2;
}
}
else if (refDCval >= 71) {
if (Rs < 2352) {
return 4;
}
else if (Rs >= 2352) {
return 1;
}
}
}
}
}
else if (refDCval >= 101) {
if (negBalance < 38) {
if (refDCval < 137) {
if (Rs < 1405) {
return 1;
}
else if (Rs >= 1405) {
if (Rs < 1558) {
return 8;
}
else if (Rs >= 1558) {
if (Rs < 2039) {
return 2;
}
else if (Rs >= 2039) {
return 1;
}
}
}
}
else if (refDCval >= 137) {
if (diffAFD < 3) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
return 4;
}
}
else if (diffAFD >= 3) {
return 8;
}
}
}
else if (negBalance >= 38) {
if (MVDiff < 29) {
if (refDCval < 118) {
if (Rs < 877) {
return 1;
}
else if (Rs >= 877) {
return 8;
}
}
else if (refDCval >= 118) {
if (tcor < 92) {
if (diffMVdiffVal < 4) {
if (tcor < 89) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
if (Rs < 1427) {
return 2;
}
else if (Rs >= 1427) {
return 8;
}
}
}
else if (tcor >= 89) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
else if (tcor >= 92) {
if (ssDCval < 137) {
return 8;
}
else if (ssDCval >= 137) {
return 4;
}
}
}
}
else if (MVDiff >= 29) {
if (Rs < 1451) {
return 4;
}
else if (Rs >= 1451) {
if (Rs < 1937) {
return 2;
}
else if (Rs >= 1937) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 407) {
if (diffAFD < -4) {
if (Rs < 383) {
if (Rs < 289) {
return 8;
}
else if (Rs >= 289) {
return 4;
}
}
else if (Rs >= 383) {
if (Rs < 392) {
return 2;
}
else if (Rs >= 392) {
return 1;
}
}
}
else if (diffAFD >= -4) {
return 1;
}
}
else if (Rs >= 407) {
if (Rs < 1001) {
if (CsDiff < 5) {
if (refDCval < 49) {
if (ssDCval < 45) {
if (Rs < 640) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
return 4;
}
}
else if (Rs >= 640) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (SC < 1430) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
else if (SC >= 1430) {
return 8;
}
}
}
}
else if (ssDCval >= 45) {
if (TSC < 7) {
if (Rs < 641) {
if (Rs < 483) {
return 2;
}
else if (Rs >= 483) {
return 8;
}
}
else if (Rs >= 641) {
if (SC < 1547) {
return 1;
}
else if (SC >= 1547) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (diffRsCsdiff < -8) {
if (Cs < 845) {
return 8;
}
else if (Cs >= 845) {
return 1;
}
}
else if (diffRsCsdiff >= -8) {
if (Rs < 940) {
return 1;
}
else if (Rs >= 940) {
return 8;
}
}
}
}
}
}
else if (TSC >= 7) {
if (ssDCval < 46) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 1;
}
}
else if (ssDCval >= 46) {
return 8;
}
}
}
}
else if (refDCval >= 49) {
if (RsDiff < 3) {
if (ssDCval < 101) {
if (gchDC < 2) {
if (tcor < 72) {
if (Rs < 628) {
return 8;
}
else if (Rs >= 628) {
return 4;
}
}
else if (tcor >= 72) {
return 8;
}
}
else if (gchDC >= 2) {
if (Rs < 463) {
return 2;
}
else if (Rs >= 463) {
return 1;
}
}
}
else if (ssDCval >= 101) {
if (ssDCval < 110) {
return 1;
}
else if (ssDCval >= 110) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (ssDCval < 52) {
if (diffRsCsdiff < -18) {
if (Cs < 755) {
if (Rs < 815) {
return 8;
}
else if (Rs >= 815) {
return 2;
}
}
else if (Cs >= 755) {
return 4;
}
}
else if (diffRsCsdiff >= -18) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
if (Cs < 724) {
return 1;
}
else if (Cs >= 724) {
return 8;
}
}
}
}
else if (ssDCval >= 52) {
if (refDCval < 77) {
if (mcTcor < 84) {
return 4;
}
else if (mcTcor >= 84) {
if (negBalance < 52) {
if (diffRsCsdiff < -3) {
return 4;
}
else if (diffRsCsdiff >= -3) {
if (Rs < 967) {
if (SC < 1471) {
return 4;
}
else if (SC >= 1471) {
return 8;
}
}
else if (Rs >= 967) {
return 4;
}
}
}
else if (negBalance >= 52) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
return 4;
}
}
}
}
else if (refDCval >= 77) {
if (Cs < 809) {
if (mcTcor < 93) {
if (mcTcor < 80) {
if (Rs < 764) {
return 2;
}
else if (Rs >= 764) {
return 8;
}
}
else if (mcTcor >= 80) {
if (RsCsDiff < 23) {
return 8;
}
else if (RsCsDiff >= 23) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
if (MVDiff < 8) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (MVDiff >= 8) {
if (Rs < 800) {
if (Cs < 707) {
return 8;
}
else if (Cs >= 707) {
return 2;
}
}
else if (Rs >= 800) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 2;
}
}
}
}
}
else if (Cs >= 809) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 4;
}
}
}
}
}
}
}
else if (CsDiff >= 5) {
if (negBalance < 50) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (negBalance >= 50) {
if (diffMVdiffVal < 8) {
if (Rs < 741) {
if (tcor < 75) {
return 4;
}
else if (tcor >= 75) {
return 1;
}
}
else if (Rs >= 741) {
if (refDCval < 76) {
if (SC < 1700) {
return 8;
}
else if (SC >= 1700) {
if (negBalance < 55) {
return 4;
}
else if (negBalance >= 55) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
else if (refDCval >= 76) {
return 8;
}
}
}
else if (diffMVdiffVal >= 8) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
else if (Rs >= 1001) {
if (posBalance < 55) {
if (Rs < 1541) {
if (AFD < 11) {
if (ssDCval < 126) {
if (MVDiff < 10) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Rs < 1149) {
return 4;
}
else if (Rs >= 1149) {
return 8;
}
}
}
}
else if (MVDiff >= 10) {
if (negBalance < 49) {
if (Rs < 1181) {
return 8;
}
else if (Rs >= 1181) {
return 4;
}
}
else if (negBalance >= 49) {
return 8;
}
}
}
else if (ssDCval >= 126) {
if (Cs < 682) {
return 1;
}
else if (Cs >= 682) {
return 4;
}
}
}
else if (AFD >= 11) {
if (Cs < 645) {
if (Rs < 1458) {
return 2;
}
else if (Rs >= 1458) {
return 1;
}
}
else if (Cs >= 645) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (AFD < 12) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
if (Rs < 1097) {
return 1;
}
else if (Rs >= 1097) {
return 8;
}
}
}
}
else if (AFD >= 12) {
return 8;
}
}
}
}
}
else if (Rs >= 1541) {
if (posBalance < 51) {
if (MVDiff < 32) {
if (MVDiff < 7) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 1785) {
return 1;
}
else if (Rs >= 1785) {
return 8;
}
}
}
else if (MVDiff >= 7) {
if (RsCsDiff < 85) {
if (Cs < 674) {
return 1;
}
else if (Cs >= 674) {
return 4;
}
}
else if (RsCsDiff >= 85) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
return 4;
}
}
}
}
else if (MVDiff >= 32) {
if (diffRsCsdiff < 30) {
return 8;
}
else if (diffRsCsdiff >= 30) {
if (Rs < 1883) {
return 2;
}
else if (Rs >= 1883) {
return 1;
}
}
}
}
else if (posBalance >= 51) {
if (Cs < 843) {
if (ssDCval < 75) {
if (Rs < 1926) {
return 8;
}
else if (Rs >= 1926) {
return 2;
}
}
else if (ssDCval >= 75) {
if (RsCsDiff < 49) {
return 2;
}
else if (RsCsDiff >= 49) {
return 1;
}
}
}
else if (Cs >= 843) {
if (ssDCval < 73) {
if (Cs < 1050) {
return 8;
}
else if (Cs >= 1050) {
if (Rs < 2008) {
return 1;
}
else if (Rs >= 2008) {
return 8;
}
}
}
else if (ssDCval >= 73) {
if (Rs < 1675) {
return 4;
}
else if (Rs >= 1675) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 55) {
if (MVDiff < 18) {
if (tcor < 89) {
if (Rs < 1297) {
if (RsCsDiff < 38) {
return 8;
}
else if (RsCsDiff >= 38) {
if (Cs < 1178) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (Cs < 909) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (Cs >= 909) {
return 8;
}
}
}
}
else if (Cs >= 1178) {
if (Rs < 1268) {
return 4;
}
else if (Rs >= 1268) {
return 8;
}
}
}
}
else if (Rs >= 1297) {
if (Cs < 671) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (Cs >= 671) {
if (RsCsDiff < 27) {
if (ssDCval < 92) {
if (Rs < 1437) {
return 8;
}
else if (Rs >= 1437) {
return 4;
}
}
else if (ssDCval >= 92) {
if (Cs < 948) {
if (ssDCval < 104) {
return 2;
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (Cs >= 948) {
return 8;
}
}
}
else if (RsCsDiff >= 27) {
if (Rs < 1328) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
return 4;
}
}
else if (Rs >= 1328) {
if (AFD < 12) {
if (mcTcor < 86) {
if (diffAFD < 2) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
if (Rs < 1747) {
return 2;
}
else if (Rs >= 1747) {
return 8;
}
}
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (mcTcor >= 86) {
return 8;
}
}
else if (AFD >= 12) {
if (RsCsDiff < 51) {
return 8;
}
else if (RsCsDiff >= 51) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (ssDCval < 88) {
return 8;
}
else if (ssDCval >= 88) {
if (Rs < 1568) {
return 8;
}
else if (Rs >= 1568) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (tcor >= 89) {
if (ssDCval < 81) {
if (Rs < 2135) {
return 8;
}
else if (Rs >= 2135) {
return 2;
}
}
else if (ssDCval >= 81) {
if (CsDiff < 4) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (Cs < 631) {
return 1;
}
else if (Cs >= 631) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (MVDiff < 14) {
if (Rs < 1224) {
if (Rs < 1201) {
return 8;
}
else if (Rs >= 1201) {
return 1;
}
}
else if (Rs >= 1224) {
return 8;
}
}
else if (MVDiff >= 14) {
return 2;
}
}
}
}
}
else if (MVDiff >= 18) {
if (ssDCval < 105) {
if (Rs < 1741) {
return 8;
}
else if (Rs >= 1741) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (Rs < 1900) {
return 1;
}
else if (Rs >= 1900) {
return 8;
}
}
}
}
else if (ssDCval >= 105) {
if (ssDCval < 110) {
return 1;
}
else if (ssDCval >= 110) {
if (SC < 1989) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (SC >= 1989) {
if (diffMVdiffVal < 6) {
return 4;
}
else if (diffMVdiffVal >= 6) {
if (ssDCval < 128) {
return 8;
}
else if (ssDCval >= 128) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (mcTcor < 86) {
if (ssDCval < 42) {
if (Rs < 724) {
if (SC < 1287) {
return 8;
}
else if (SC >= 1287) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (Rs >= 724) {
return 8;
}
}
else if (ssDCval >= 42) {
if (AFD < 6) {
if (ssDCval < 47) {
if (Rs < 1029) {
return 8;
}
else if (Rs >= 1029) {
return 1;
}
}
else if (ssDCval >= 47) {
if (tcor < 84) {
if (Rs < 896) {
return 8;
}
else if (Rs >= 896) {
return 4;
}
}
else if (tcor >= 84) {
return 8;
}
}
}
else if (AFD >= 6) {
if (tcor < 74) {
if (ssDCval < 100) {
if (tcor < 73) {
if (Rs < 659) {
if (RsCsDiff < 22) {
return 8;
}
else if (RsCsDiff >= 22) {
if (Cs < 1209) {
if (negBalance < 46) {
return 1;
}
else if (negBalance >= 46) {
return 4;
}
}
else if (Cs >= 1209) {
return 2;
}
}
}
else if (Rs >= 659) {
if (ssDCval < 44) {
if (Rs < 1126) {
return 2;
}
else if (Rs >= 1126) {
return 8;
}
}
else if (ssDCval >= 44) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (Cs < 651) {
if (Rs < 943) {
return 4;
}
else if (Rs >= 943) {
return 8;
}
}
else if (Cs >= 651) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 73) {
if (mu_mv_mag_sq < 69) {
if (Cs < 822) {
return 8;
}
else if (Cs >= 822) {
return 4;
}
}
else if (mu_mv_mag_sq >= 69) {
if (mcTcor < 76) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (mcTcor >= 76) {
return 1;
}
}
}
}
else if (ssDCval >= 100) {
if (Rs < 1226) {
return 1;
}
else if (Rs >= 1226) {
return 4;
}
}
}
else if (tcor >= 74) {
if (refDCval < 63) {
if (ssDCval < 53) {
if (Cs < 783) {
if (mu_mv_mag_sq < 101) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 1184) {
return 1;
}
else if (Rs >= 1184) {
return 8;
}
}
}
else if (Cs >= 783) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
return 4;
}
}
}
else if (ssDCval >= 53) {
if (AFD < 11) {
if (posBalance < 58) {
if (Cs < 1048) {
return 8;
}
else if (Cs >= 1048) {
if (Rs < 1766) {
return 1;
}
else if (Rs >= 1766) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (Cs < 1129) {
if (Rs < 648) {
return 8;
}
else if (Rs >= 648) {
if (Rs < 1494) {
return 4;
}
else if (Rs >= 1494) {
return 8;
}
}
}
else if (Cs >= 1129) {
if (Rs < 1138) {
return 8;
}
else if (Rs >= 1138) {
return 1;
}
}
}
}
else if (AFD >= 11) {
if (Rs < 1706) {
return 8;
}
else if (Rs >= 1706) {
return 2;
}
}
}
}
else if (refDCval >= 63) {
return 8;
}
}
}
}
}
else if (mcTcor >= 86) {
if (refDCval < 78) {
if (RsCsDiff < 19) {
if (Rs < 868) {
if (negBalance < 29) {
if (refDCval < 48) {
return 1;
}
else if (refDCval >= 48) {
if (gchDC < 2) {
if (Rs < 440) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
if (MVDiff < 7) {
if (Cs < 845) {
return 8;
}
else if (Cs >= 845) {
return 1;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (Rs >= 440) {
if (mcTcor < 96) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (Rs < 797) {
return 4;
}
else if (Rs >= 797) {
if (Rs < 824) {
return 8;
}
else if (Rs >= 824) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
if (MVDiff < 4) {
if (refDCval < 64) {
if (negBalance < 24) {
if (Cs < 579) {
return 8;
}
else if (Cs >= 579) {
return 1;
}
}
else if (negBalance >= 24) {
if (Rs < 559) {
return 4;
}
else if (Rs >= 559) {
return 8;
}
}
}
else if (refDCval >= 64) {
if (Rs < 581) {
return 1;
}
else if (Rs >= 581) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 4;
}
}
}
}
}
else if (gchDC >= 2) {
if (Cs < 772) {
return 1;
}
else if (Cs >= 772) {
return 2;
}
}
}
}
else if (negBalance >= 29) {
if (ssDCval < 37) {
if (Cs < 744) {
return 8;
}
else if (Cs >= 744) {
return 1;
}
}
else if (ssDCval >= 37) {
if (Rs < 526) {
if (Cs < 772) {
if (negBalance < 35) {
if (AFD < 6) {
if (MVDiff < 5) {
if (Rs < 336) {
return 1;
}
else if (Rs >= 336) {
return 8;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (negBalance >= 35) {
if (MVDiff < 4) {
if (Rs < 500) {
return 8;
}
else if (Rs >= 500) {
return 2;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
return 1;
}
}
}
}
else if (Cs >= 772) {
return 1;
}
}
else if (Rs >= 526) {
if (ssDCval < 56) {
if (Cs < 619) {
if (Rs < 623) {
return 1;
}
else if (Rs >= 623) {
return 8;
}
}
else if (Cs >= 619) {
if (Rs < 840) {
if (SC < 1703) {
if (ssDCval < 40) {
return 8;
}
else if (ssDCval >= 40) {
if (diffAFD < 2) {
if (Rs < 754) {
return 4;
}
else if (Rs >= 754) {
if (Cs < 773) {
if (tcor < 95) {
if (mu_mv_mag_sq < 101) {
return 8;
}
else if (mu_mv_mag_sq >= 101) {
return 4;
}
}
else if (tcor >= 95) {
return 4;
}
}
else if (Cs >= 773) {
return 4;
}
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (SC >= 1703) {
return 2;
}
}
else if (Rs >= 840) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 89) {
if (Rs < 796) {
if (posBalance < 55) {
if (TSC < 4) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
else if (TSC >= 4) {
if (negBalance < 40) {
if (Rs < 785) {
return 1;
}
else if (Rs >= 785) {
if (Cs < 635) {
return 1;
}
else if (Cs >= 635) {
return 8;
}
}
}
else if (negBalance >= 40) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
if (Cs < 757) {
if (Rs < 622) {
return 8;
}
else if (Rs >= 622) {
return 4;
}
}
else if (Cs >= 757) {
return 1;
}
}
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (Rs >= 796) {
if (Cs < 983) {
if (Rs < 835) {
return 4;
}
else if (Rs >= 835) {
return 2;
}
}
else if (Cs >= 983) {
if (Rs < 852) {
return 1;
}
else if (Rs >= 852) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (ssDCval < 68) {
if (tcor < 81) {
return 4;
}
else if (tcor >= 81) {
if (negBalance < 37) {
if (mcTcor < 95) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (mcTcor >= 95) {
return 8;
}
}
else if (negBalance >= 37) {
return 8;
}
}
}
else if (ssDCval >= 68) {
if (posBalance < 54) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (Rs < 697) {
return 8;
}
else if (Rs >= 697) {
return 4;
}
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 868) {
if (RsCsDiff < 9) {
if (negBalance < 22) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 1;
}
}
else if (negBalance >= 22) {
if (CsDiff < 2) {
if (ssDCval < 65) {
if (diffAFD < 3) {
if (Cs < 1150) {
return 8;
}
else if (Cs >= 1150) {
if (Rs < 1175) {
return 8;
}
else if (Rs >= 1175) {
return 1;
}
}
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (ssDCval >= 65) {
if (Rs < 1068) {
return 4;
}
else if (Rs >= 1068) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (MVDiff < 4) {
if (ssDCval < 66) {
if (Cs < 584) {
if (Cs < 575) {
return 8;
}
else if (Cs >= 575) {
return 1;
}
}
else if (Cs >= 584) {
if (AFD < 7) {
if (MVDiff < 3) {
if (Cs < 625) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 8;
}
}
else if (Cs >= 625) {
return 8;
}
}
else if (MVDiff >= 3) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (ssDCval >= 66) {
if (posBalance < 54) {
if (Rs < 1335) {
if (negBalance < 36) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (negBalance >= 36) {
if (Rs < 1039) {
return 4;
}
else if (Rs >= 1039) {
return 1;
}
}
}
else if (Rs >= 1335) {
return 8;
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (MVDiff < 11) {
if (negBalance < 32) {
return 1;
}
else if (negBalance >= 32) {
if (mu_mv_mag_sq < 107) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Rs < 1049) {
return 1;
}
else if (Rs >= 1049) {
if (Cs < 1131) {
return 8;
}
else if (Cs >= 1131) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 1095) {
return 8;
}
else if (Rs >= 1095) {
return 1;
}
}
}
}
else if (MVDiff >= 11) {
if (Cs < 609) {
return 1;
}
else if (Cs >= 609) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 9) {
if (TSC < 6) {
if (refDCval < 56) {
if (tcor < 88) {
if (ssDCval < 45) {
if (Cs < 622) {
return 1;
}
else if (Cs >= 622) {
return 8;
}
}
else if (ssDCval >= 45) {
if (posBalance < 57) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 1;
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (tcor >= 88) {
if (posBalance < 46) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 835) {
return 8;
}
else if (Cs >= 835) {
if (Rs < 1419) {
return 4;
}
else if (Rs >= 1419) {
return 4;
}
}
}
}
else if (posBalance >= 46) {
if (tcor < 94) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Rs < 1340) {
return 8;
}
else if (Rs >= 1340) {
return 2;
}
}
}
else if (tcor >= 94) {
if (Rs < 1108) {
if (Rs < 1000) {
return 8;
}
else if (Rs >= 1000) {
return 4;
}
}
else if (Rs >= 1108) {
return 8;
}
}
}
}
}
else if (refDCval >= 56) {
if (Rs < 1305) {
if (Rs < 1080) {
if (negBalance < 35) {
if (Rs < 994) {
return 4;
}
else if (Rs >= 994) {
if (negBalance < 26) {
return 4;
}
else if (negBalance >= 26) {
return 8;
}
}
}
else if (negBalance >= 35) {
if (RsCsDiff < 16) {
return 4;
}
else if (RsCsDiff >= 16) {
if (Cs < 1102) {
if (Cs < 656) {
return 1;
}
else if (Cs >= 656) {
return 8;
}
}
else if (Cs >= 1102) {
return 4;
}
}
}
}
else if (Rs >= 1080) {
if (MVDiff < 6) {
if (negBalance < 38) {
if (Cs < 925) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (Cs >= 925) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
return 4;
}
}
}
else if (negBalance >= 38) {
if (Cs < 1254) {
return 1;
}
else if (Cs >= 1254) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (tcor < 88) {
if (Rs < 1169) {
return 4;
}
else if (Rs >= 1169) {
return 8;
}
}
else if (tcor >= 88) {
return 2;
}
}
}
}
}
else if (Rs >= 1305) {
if (mu_mv_mag_sq < 104) {
if (SC < 2685) {
if (SC < 2289) {
if (CsDiff < 3) {
if (Cs < 680) {
return 2;
}
else if (Cs >= 680) {
return 8;
}
}
else if (CsDiff >= 3) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (SC >= 2289) {
if (ssDCval < 68) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
return 8;
}
}
else if (ssDCval >= 68) {
if (Cs < 1046) {
if (Rs < 1549) {
return 8;
}
else if (Rs >= 1549) {
return 2;
}
}
else if (Cs >= 1046) {
return 4;
}
}
}
}
else if (SC >= 2685) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 1732) {
if (Rs < 1320) {
if (Rs < 1317) {
return 8;
}
else if (Rs >= 1317) {
return 4;
}
}
else if (Rs >= 1320) {
return 8;
}
}
else if (Rs >= 1732) {
return 1;
}
}
}
}
}
else if (TSC >= 6) {
if (negBalance < 43) {
if (ssDCval < 59) {
if (SC < 2159) {
return 8;
}
else if (SC >= 2159) {
return 2;
}
}
else if (ssDCval >= 59) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (negBalance >= 43) {
if (Cs < 640) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (Cs >= 640) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 19) {
if (negBalance < 45) {
if (Cs < 957) {
if (tcor < 85) {
if (mu_mv_mag_sq < 55) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (mcTcor < 90) {
if (Rs < 1694) {
return 8;
}
else if (Rs >= 1694) {
return 4;
}
}
else if (mcTcor >= 90) {
if (Cs < 699) {
if (Cs < 546) {
return 2;
}
else if (Cs >= 546) {
return 4;
}
}
else if (Cs >= 699) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (Cs < 692) {
if (Rs < 894) {
if (Cs < 667) {
return 1;
}
else if (Cs >= 667) {
return 2;
}
}
else if (Rs >= 894) {
if (posBalance < 48) {
if (Rs < 1349) {
return 1;
}
else if (Rs >= 1349) {
return 2;
}
}
else if (posBalance >= 48) {
return 8;
}
}
}
else if (Cs >= 692) {
if (posBalance < 53) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
if (Rs < 1454) {
return 2;
}
else if (Rs >= 1454) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (Cs < 744) {
return 4;
}
else if (Cs >= 744) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Rs < 1314) {
return 1;
}
else if (Rs >= 1314) {
if (Rs < 1513) {
return 8;
}
else if (Rs >= 1513) {
return 2;
}
}
}
}
}
}
}
}
else if (tcor >= 85) {
if (mu_mv_mag_sq < 92) {
if (negBalance < 38) {
if (ssDCval < 69) {
if (negBalance < 32) {
if (mu_mv_mag_sq < 82) {
if (RsCsDiff < 59) {
return 1;
}
else if (RsCsDiff >= 59) {
return 2;
}
}
else if (mu_mv_mag_sq >= 82) {
return 2;
}
}
else if (negBalance >= 32) {
return 2;
}
}
else if (ssDCval >= 69) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (negBalance >= 38) {
if (Rs < 1868) {
if (ssDCval < 59) {
if (Cs < 854) {
if (Cs < 625) {
return 2;
}
else if (Cs >= 625) {
return 8;
}
}
else if (Cs >= 854) {
return 2;
}
}
else if (ssDCval >= 59) {
if (mu_mv_mag_sq < 42) {
if (Cs < 789) {
return 4;
}
else if (Cs >= 789) {
return 8;
}
}
else if (mu_mv_mag_sq >= 42) {
if (posBalance < 52) {
if (Rs < 1391) {
return 4;
}
else if (Rs >= 1391) {
return 2;
}
}
else if (posBalance >= 52) {
if (Rs < 1527) {
if (Rs < 918) {
return 8;
}
else if (Rs >= 918) {
return 1;
}
}
else if (Rs >= 1527) {
return 8;
}
}
}
}
}
else if (Rs >= 1868) {
if (Cs < 916) {
return 2;
}
else if (Cs >= 916) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (mu_mv_mag_sq < 116) {
if (Rs < 1350) {
return 8;
}
else if (Rs >= 1350) {
if (Rs < 1429) {
if (Rs < 1412) {
return 2;
}
else if (Rs >= 1412) {
return 8;
}
}
else if (Rs >= 1429) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 116) {
if (Rs < 1123) {
return 8;
}
else if (Rs >= 1123) {
return 1;
}
}
}
}
}
else if (Cs >= 957) {
if (refDCval < 60) {
if (SC < 2351) {
if (mcTcor < 92) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 4;
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 78) {
return 4;
}
else if (mu_mv_mag_sq >= 78) {
if (AFD < 6) {
if (MVDiff < 1) {
if (Rs < 1107) {
return 8;
}
else if (Rs >= 1107) {
return 4;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (AFD >= 6) {
return 8;
}
}
}
}
else if (SC >= 2351) {
if (MVDiff < 13) {
if (RsCsDiff < 70) {
if (Cs < 1013) {
return 1;
}
else if (Cs >= 1013) {
if (negBalance < 25) {
return 4;
}
else if (negBalance >= 25) {
if (Rs < 1275) {
if (Rs < 1250) {
return 8;
}
else if (Rs >= 1250) {
return 4;
}
}
else if (Rs >= 1275) {
return 8;
}
}
}
}
else if (RsCsDiff >= 70) {
return 4;
}
}
else if (MVDiff >= 13) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
if (Cs < 1052) {
return 1;
}
else if (Cs >= 1052) {
return 8;
}
}
}
}
}
else if (refDCval >= 60) {
if (negBalance < 36) {
if (tcor < 89) {
if (posBalance < 50) {
if (Rs < 1554) {
return 4;
}
else if (Rs >= 1554) {
return 2;
}
}
else if (posBalance >= 50) {
return 1;
}
}
else if (tcor >= 89) {
if (diffRsCsdiff < 8) {
if (Cs < 993) {
return 8;
}
else if (Cs >= 993) {
return 2;
}
}
else if (diffRsCsdiff >= 8) {
return 8;
}
}
}
else if (negBalance >= 36) {
if (ssDCval < 75) {
if (negBalance < 42) {
if (Rs < 1502) {
if (mu_mv_mag_sq < 52) {
if (Cs < 1162) {
return 4;
}
else if (Cs >= 1162) {
return 2;
}
}
else if (mu_mv_mag_sq >= 52) {
return 4;
}
}
else if (Rs >= 1502) {
if (mu_mv_mag_sq < 86) {
return 4;
}
else if (mu_mv_mag_sq >= 86) {
return 8;
}
}
}
else if (negBalance >= 42) {
if (ssDCval < 63) {
if (Rs < 802) {
return 1;
}
else if (Rs >= 802) {
return 8;
}
}
else if (ssDCval >= 63) {
if (Cs < 1257) {
if (RsCsDiff < 55) {
if (mu_mv_mag_sq < 50) {
return 8;
}
else if (mu_mv_mag_sq >= 50) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
else if (RsCsDiff >= 55) {
return 8;
}
}
else if (Cs >= 1257) {
if (Rs < 1410) {
return 1;
}
else if (Rs >= 1410) {
return 4;
}
}
}
}
}
else if (ssDCval >= 75) {
if (Rs < 1442) {
return 4;
}
else if (Rs >= 1442) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Cs < 1132) {
return 8;
}
else if (Cs >= 1132) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 1586) {
if (posBalance < 56) {
if (mu_mv_mag_sq < 55) {
if (RsCsDiff < 26) {
if (TSC < 7) {
if (MVDiff < 5) {
if (Rs < 1346) {
return 8;
}
else if (Rs >= 1346) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (RsCsDiff >= 26) {
if (Rs < 1028) {
return 1;
}
else if (Rs >= 1028) {
if (gchDC < 1) {
if (MVDiff < 7) {
if (Rs < 1164) {
return 4;
}
else if (Rs >= 1164) {
return 8;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (Rs < 1330) {
if (Cs < 1167) {
if (SC < 1730) {
if (Rs < 1046) {
return 4;
}
else if (Rs >= 1046) {
return 2;
}
}
else if (SC >= 1730) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 4;
}
}
}
else if (Cs >= 1167) {
if (Cs < 1239) {
return 4;
}
else if (Cs >= 1239) {
return 1;
}
}
}
else if (Rs >= 1330) {
return 8;
}
}
}
else if (posBalance >= 56) {
if (Rs < 1080) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (Rs >= 1080) {
if (SC < 1872) {
if (Rs < 1210) {
return 8;
}
else if (Rs >= 1210) {
return 4;
}
}
else if (SC >= 1872) {
return 8;
}
}
}
}
else if (Rs >= 1586) {
if (TSC < 8) {
if (MVDiff < 10) {
if (mu_mv_mag_sq < 104) {
if (Rs < 2344) {
return 8;
}
else if (Rs >= 2344) {
return 2;
}
}
else if (mu_mv_mag_sq >= 104) {
if (RsDiff < 5) {
if (Rs < 1738) {
return 8;
}
else if (Rs >= 1738) {
return 1;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (MVDiff >= 10) {
if (TSC < 7) {
if (Rs < 1637) {
return 4;
}
else if (Rs >= 1637) {
return 8;
}
}
else if (TSC >= 7) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
if (Rs < 1893) {
return 2;
}
else if (Rs >= 1893) {
return 1;
}
}
}
}
}
else if (TSC >= 8) {
if (ssDCval < 70) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 78) {
if (Rs < 366) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (Rs >= 366) {
if (MVDiff < 25) {
if (tcor < 87) {
if (mu_mv_mag_sq < 70) {
if (mcTcor < 90) {
if (SC < 2766) {
if (posBalance < 53) {
if (Cs < 764) {
return 8;
}
else if (Cs >= 764) {
return 2;
}
}
else if (posBalance >= 53) {
if (ssDCval < 149) {
return 8;
}
else if (ssDCval >= 149) {
return 2;
}
}
}
else if (SC >= 2766) {
if (Rs < 1679) {
return 4;
}
else if (Rs >= 1679) {
if (Rs < 1719) {
return 8;
}
else if (Rs >= 1719) {
return 2;
}
}
}
}
else if (mcTcor >= 90) {
if (negBalance < 45) {
if (Rs < 1898) {
if (mu_mv_mag_sq < 60) {
if (RsDiff < 7) {
if (Cs < 1232) {
return 2;
}
else if (Cs >= 1232) {
if (Cs < 1240) {
return 8;
}
else if (Cs >= 1240) {
return 2;
}
}
}
else if (RsDiff >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 60) {
if (AFD < 11) {
if (Rs < 611) {
return 8;
}
else if (Rs >= 611) {
if (Cs < 1136) {
return 2;
}
else if (Cs >= 1136) {
return 8;
}
}
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (Rs >= 1898) {
return 8;
}
}
else if (negBalance >= 45) {
if (ssDCval < 88) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
if (Rs < 1894) {
return 8;
}
else if (Rs >= 1894) {
return 1;
}
}
}
else if (ssDCval >= 88) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (Cs < 717) {
if (Rs < 1074) {
return 8;
}
else if (Rs >= 1074) {
return 4;
}
}
else if (Cs >= 717) {
if (SC < 3540) {
return 8;
}
else if (SC >= 3540) {
return 4;
}
}
}
}
else if (tcor >= 87) {
if (Rs < 438) {
if (refDCval < 93) {
return 8;
}
else if (refDCval >= 93) {
if (MVDiff < 9) {
if (Cs < 758) {
return 1;
}
else if (Cs >= 758) {
if (Rs < 415) {
return 4;
}
else if (Rs >= 415) {
return 1;
}
}
}
else if (MVDiff >= 9) {
return 8;
}
}
}
else if (Rs >= 438) {
if (ssDCval < 158) {
if (posBalance < 48) {
if (ssDCval < 97) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (Rs < 1850) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (Rs >= 1850) {
if (diffAFD < -1) {
if (Rs < 2111) {
return 8;
}
else if (Rs >= 2111) {
return 1;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
}
}
else if (ssDCval >= 97) {
if (negBalance < 28) {
if (Cs < 1155) {
return 8;
}
else if (Cs >= 1155) {
return 4;
}
}
else if (negBalance >= 28) {
if (diffRsCsdiff < 26) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 1396) {
return 4;
}
else if (Rs >= 1396) {
return 8;
}
}
}
else if (diffRsCsdiff >= 26) {
if (Rs < 1995) {
return 2;
}
else if (Rs >= 1995) {
return 8;
}
}
}
}
}
else if (posBalance >= 48) {
if (RsDiff < 6) {
if (diffAFD < -2) {
if (refDCval < 109) {
if (negBalance < 36) {
if (Cs < 811) {
return 1;
}
else if (Cs >= 811) {
return 8;
}
}
else if (negBalance >= 36) {
return 8;
}
}
else if (refDCval >= 109) {
if (Rs < 1431) {
return 1;
}
else if (Rs >= 1431) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (Cs < 1260) {
return 1;
}
else if (Cs >= 1260) {
return 8;
}
}
}
}
}
else if (diffAFD >= -2) {
if (negBalance < 34) {
if (RsCsDiff < 16) {
if (SC < 1473) {
if (SC < 1439) {
if (RsCsDiff < 7) {
if (MVDiff < 9) {
if (MVDiff < 2) {
if (SC < 1366) {
return 8;
}
else if (SC >= 1366) {
return 2;
}
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (RsCsDiff >= 7) {
return 8;
}
}
else if (SC >= 1439) {
return 2;
}
}
else if (SC >= 1473) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (mu_mv_mag_sq < 84) {
if (AFD < 6) {
if (Rs < 1536) {
return 4;
}
else if (Rs >= 1536) {
return 1;
}
}
else if (AFD >= 6) {
if (posBalance < 51) {
if (Cs < 1207) {
return 1;
}
else if (Cs >= 1207) {
return 8;
}
}
else if (posBalance >= 51) {
if (Cs < 828) {
return 1;
}
else if (Cs >= 828) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (negBalance >= 34) {
if (MVDiff < 3) {
if (Cs < 994) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (posBalance < 59) {
if (ssDCval < 93) {
return 1;
}
else if (ssDCval >= 93) {
if (negBalance < 43) {
if (negBalance < 37) {
if (Rs < 1499) {
return 2;
}
else if (Rs >= 1499) {
return 8;
}
}
else if (negBalance >= 37) {
return 8;
}
}
else if (negBalance >= 43) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 1;
}
}
}
}
else if (posBalance >= 59) {
if (diffMVdiffVal < 1) {
return 1;
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
}
else if (Cs >= 994) {
if (SC < 1854) {
if (Rs < 683) {
return 8;
}
else if (Rs >= 683) {
return 4;
}
}
else if (SC >= 1854) {
if (mu_mv_mag_sq < 92) {
if (ssDCval < 84) {
if (Rs < 1635) {
return 8;
}
else if (Rs >= 1635) {
return 1;
}
}
else if (ssDCval >= 84) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 1638) {
return 8;
}
else if (Rs >= 1638) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (refDCval < 138) {
if (Cs < 859) {
return 8;
}
else if (Cs >= 859) {
if (ssDCval < 94) {
if (Cs < 949) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (Cs >= 949) {
if (diffRsCsdiff < -4) {
if (posBalance < 57) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 4;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
return 8;
}
}
}
else if (ssDCval >= 94) {
return 8;
}
}
}
else if (refDCval >= 138) {
if (mu_mv_mag_sq < 47) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (Rs < 1231) {
return 1;
}
else if (Rs >= 1231) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
}
}
}
}
}
else if (RsDiff >= 6) {
if (CsDiff < 4) {
if (tcor < 91) {
if (ssDCval < 106) {
if (Rs < 1950) {
return 2;
}
else if (Rs >= 1950) {
return 1;
}
}
else if (ssDCval >= 106) {
return 8;
}
}
else if (tcor >= 91) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (refDCval < 105) {
if (tcor < 94) {
if (TSC < 8) {
if (ssDCval < 87) {
if (Cs < 1208) {
return 8;
}
else if (Cs >= 1208) {
return 4;
}
}
else if (ssDCval >= 87) {
if (Rs < 1521) {
return 8;
}
else if (Rs >= 1521) {
return 2;
}
}
}
else if (TSC >= 8) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 2;
}
}
}
else if (tcor >= 94) {
if (Cs < 1212) {
return 4;
}
else if (Cs >= 1212) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (Cs < 867) {
return 1;
}
else if (Cs >= 867) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 158) {
if (Rs < 890) {
return 4;
}
else if (Rs >= 890) {
return 2;
}
}
}
}
}
else if (MVDiff >= 25) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (Cs < 1045) {
if (refDCval < 137) {
if (ssDCval < 88) {
if (Rs < 566) {
return 2;
}
else if (Rs >= 566) {
return 8;
}
}
else if (ssDCval >= 88) {
return 8;
}
}
else if (refDCval >= 137) {
if (Rs < 1623) {
if (Rs < 902) {
return 1;
}
else if (Rs >= 902) {
return 4;
}
}
else if (Rs >= 1623) {
return 8;
}
}
}
else if (Cs >= 1045) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (ssDCval < 114) {
if (Rs < 1726) {
if (MVDiff < 18) {
if (ssDCval < 88) {
if (ssDCval < 57) {
if (Rs < 1269) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
else if (Rs >= 1269) {
if (refDCval < 54) {
if (Cs < 894) {
return 1;
}
else if (Cs >= 894) {
return 8;
}
}
else if (refDCval >= 54) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 1597) {
return 8;
}
else if (Rs >= 1597) {
if (Rs < 1672) {
return 4;
}
else if (Rs >= 1672) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 57) {
if (mcTcor < 88) {
if (CsDiff < 7) {
if (posBalance < 58) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 4;
}
}
else if (posBalance >= 58) {
if (negBalance < 47) {
if (Rs < 1271) {
return 1;
}
else if (Rs >= 1271) {
return 8;
}
}
else if (negBalance >= 47) {
return 8;
}
}
}
else if (CsDiff >= 7) {
if (Rs < 1172) {
if (Rs < 724) {
return 4;
}
else if (Rs >= 724) {
return 8;
}
}
else if (Rs >= 1172) {
if (Rs < 1220) {
return 2;
}
else if (Rs >= 1220) {
return 1;
}
}
}
}
else if (mcTcor >= 88) {
if (SC < 2519) {
if (TSC < 9) {
if (MVDiff < 14) {
if (negBalance < 50) {
if (AFD < 15) {
if (ssDCval < 82) {
return 1;
}
else if (ssDCval >= 82) {
return 2;
}
}
else if (AFD >= 15) {
return 8;
}
}
else if (negBalance >= 50) {
if (Rs < 313) {
return 2;
}
else if (Rs >= 313) {
if (RsCsDiff < 80) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
else if (RsCsDiff >= 80) {
if (Cs < 1213) {
if (Rs < 931) {
return 1;
}
else if (Rs >= 931) {
return 8;
}
}
else if (Cs >= 1213) {
return 2;
}
}
}
}
}
else if (MVDiff >= 14) {
if (Rs < 1139) {
if (SC < 1906) {
if (mcTcor < 94) {
if (Rs < 879) {
if (Rs < 539) {
return 4;
}
else if (Rs >= 539) {
return 8;
}
}
else if (Rs >= 879) {
return 2;
}
}
else if (mcTcor >= 94) {
if (tcor < 85) {
return 1;
}
else if (tcor >= 85) {
return 8;
}
}
}
else if (SC >= 1906) {
return 8;
}
}
else if (Rs >= 1139) {
if (posBalance < 60) {
if (Rs < 1385) {
return 4;
}
else if (Rs >= 1385) {
return 2;
}
}
else if (posBalance >= 60) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
}
else if (TSC >= 9) {
if (refDCval < 74) {
if (Cs < 1100) {
return 8;
}
else if (Cs >= 1100) {
return 2;
}
}
else if (refDCval >= 74) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 2;
}
}
}
}
else if (SC >= 2519) {
if (diffRsCsdiff < 6) {
return 8;
}
else if (diffRsCsdiff >= 6) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (MVDiff < 13) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (MVDiff >= 13) {
if (SC < 2774) {
return 1;
}
else if (SC >= 2774) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 2) {
if (negBalance < 57) {
if (Rs < 1101) {
return 8;
}
else if (Rs >= 1101) {
if (SC < 2678) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (SC >= 2678) {
if (Cs < 1132) {
return 2;
}
else if (Cs >= 1132) {
return 8;
}
}
}
}
else if (negBalance >= 57) {
if (Rs < 1580) {
return 4;
}
else if (Rs >= 1580) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (AFD < 22) {
if (Cs < 627) {
return 1;
}
else if (Cs >= 627) {
if (Cs < 857) {
if (Rs < 1173) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (Rs < 669) {
return 2;
}
else if (Rs >= 669) {
return 8;
}
}
}
}
else if (Rs >= 1173) {
return 4;
}
}
else if (Cs >= 857) {
if (diffAFD < -2) {
if (Rs < 1494) {
return 8;
}
else if (Rs >= 1494) {
return 4;
}
}
else if (diffAFD >= -2) {
if (gchDC < 2) {
if (negBalance < 58) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
return 8;
}
}
else if (negBalance >= 58) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (tcor < 82) {
return 8;
}
else if (tcor >= 82) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
}
}
}
}
}
else if (AFD >= 22) {
if (Rs < 678) {
return 1;
}
else if (Rs >= 678) {
return 4;
}
}
}
}
}
else if (MVDiff >= 18) {
if (tcor < 83) {
if (mcTcor < 86) {
if (RsCsDiff < 60) {
if (SC < 1655) {
if (posBalance < 52) {
if (gchDC < 6) {
return 1;
}
else if (gchDC >= 6) {
if (Rs < 287) {
if (Rs < 191) {
return 4;
}
else if (Rs >= 191) {
return 2;
}
}
else if (Rs >= 287) {
return 8;
}
}
}
else if (posBalance >= 52) {
if (TSC < 10) {
if (SC < 1238) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
if (ssDCval < 46) {
return 2;
}
else if (ssDCval >= 46) {
return 4;
}
}
}
else if (SC >= 1238) {
if (ssDCval < 62) {
if (Cs < 549) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (Cs >= 549) {
return 4;
}
}
else if (ssDCval >= 62) {
return 8;
}
}
}
else if (TSC >= 10) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Rs < 719) {
if (AFD < 37) {
return 4;
}
else if (AFD >= 37) {
return 8;
}
}
else if (Rs >= 719) {
return 8;
}
}
}
}
}
else if (SC >= 1655) {
if (diffAFD < 11) {
if (tcor < 56) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
return 8;
}
}
else if (tcor >= 56) {
return 8;
}
}
else if (diffAFD >= 11) {
if (Cs < 911) {
return 8;
}
else if (Cs >= 911) {
return 2;
}
}
}
}
else if (RsCsDiff >= 60) {
if (MVDiff < 55) {
if (diffTSC < 2) {
if (CsDiff < 8) {
if (SC < 2430) {
if (ssDCval < 46) {
if (SC < 1777) {
return 8;
}
else if (SC >= 1777) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 8;
}
}
}
else if (ssDCval >= 46) {
if (gchDC < 3) {
if (posBalance < 64) {
if (tcor < 61) {
return 4;
}
else if (tcor >= 61) {
if (Rs < 1112) {
return 2;
}
else if (Rs >= 1112) {
return 4;
}
}
}
else if (posBalance >= 64) {
if (CsDiff < 7) {
if (diffRsCsdiff < -12) {
return 4;
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
else if (CsDiff >= 7) {
return 2;
}
}
}
else if (gchDC >= 3) {
if (Rs < 887) {
if (Rs < 797) {
return 8;
}
else if (Rs >= 797) {
return 1;
}
}
else if (Rs >= 887) {
return 2;
}
}
}
}
else if (SC >= 2430) {
if (Rs < 1655) {
if (ssDCval < 51) {
return 2;
}
else if (ssDCval >= 51) {
if (diffMVdiffVal < 9) {
return 8;
}
else if (diffMVdiffVal >= 9) {
if (Rs < 1583) {
return 8;
}
else if (Rs >= 1583) {
return 2;
}
}
}
}
else if (Rs >= 1655) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
else if (CsDiff >= 8) {
if (ssDCval < 78) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
return 4;
}
}
else if (ssDCval >= 78) {
return 2;
}
}
}
else if (diffTSC >= 2) {
if (Cs < 1058) {
if (AFD < 19) {
if (MVDiff < 31) {
return 8;
}
else if (MVDiff >= 31) {
if (Cs < 874) {
return 4;
}
else if (Cs >= 874) {
return 1;
}
}
}
else if (AFD >= 19) {
if (Rs < 373) {
return 4;
}
else if (Rs >= 373) {
return 2;
}
}
}
else if (Cs >= 1058) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
if (MVDiff < 43) {
return 4;
}
else if (MVDiff >= 43) {
return 1;
}
}
}
}
}
else if (MVDiff >= 55) {
if (refDCval < 73) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
if (negBalance < 52) {
if (mu_mv_mag_sq < 29) {
return 8;
}
else if (mu_mv_mag_sq >= 29) {
if (Rs < 1180) {
return 4;
}
else if (Rs >= 1180) {
return 8;
}
}
}
else if (negBalance >= 52) {
if (diffAFD < 0) {
if (Rs < 400) {
return 8;
}
else if (Rs >= 400) {
if (Rs < 1056) {
return 1;
}
else if (Rs >= 1056) {
return 4;
}
}
}
else if (diffAFD >= 0) {
if (Rs < 822) {
return 8;
}
else if (Rs >= 822) {
if (Cs < 1101) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 1;
}
}
else if (Cs >= 1101) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 73) {
if (AFD < 35) {
if (SC < 1909) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
if (Rs < 944) {
return 1;
}
else if (Rs >= 944) {
return 2;
}
}
}
else if (SC >= 1909) {
if (tcor < 54) {
if (Rs < 942) {
return 8;
}
else if (Rs >= 942) {
return 4;
}
}
else if (tcor >= 54) {
return 8;
}
}
}
else if (AFD >= 35) {
if (AFD < 43) {
if (Rs < 1308) {
return 1;
}
else if (Rs >= 1308) {
return 2;
}
}
else if (AFD >= 43) {
if (Rs < 785) {
return 4;
}
else if (Rs >= 785) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 86) {
if (AFD < 23) {
if (tcor < 54) {
if (Rs < 806) {
if (diffMVdiffVal < 19) {
return 4;
}
else if (diffMVdiffVal >= 19) {
if (Rs < 296) {
return 4;
}
else if (Rs >= 296) {
return 1;
}
}
}
else if (Rs >= 806) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (tcor >= 54) {
if (posBalance < 54) {
if (ssDCval < 94) {
if (Cs < 1198) {
if (negBalance < 45) {
if (Cs < 757) {
return 2;
}
else if (Cs >= 757) {
return 1;
}
}
else if (negBalance >= 45) {
if (MVDiff < 51) {
if (diffMVdiffVal < 9) {
if (MVDiff < 33) {
if (tcor < 79) {
if (mcTcor < 89) {
if (Rs < 1017) {
return 8;
}
else if (Rs >= 1017) {
return 2;
}
}
else if (mcTcor >= 89) {
if (diffMVdiffVal < -25) {
return 8;
}
else if (diffMVdiffVal >= -25) {
return 2;
}
}
}
else if (tcor >= 79) {
if (posBalance < 50) {
if (Cs < 891) {
return 2;
}
else if (Cs >= 891) {
return 4;
}
}
else if (posBalance >= 50) {
return 8;
}
}
}
else if (MVDiff >= 33) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (Rs < 832) {
return 8;
}
else if (Rs >= 832) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= 9) {
if (negBalance < 60) {
if (Cs < 973) {
return 4;
}
else if (Cs >= 973) {
return 8;
}
}
else if (negBalance >= 60) {
return 8;
}
}
}
else if (MVDiff >= 51) {
if (posBalance < 47) {
if (refDCval < 74) {
if (ssDCval < 56) {
return 2;
}
else if (ssDCval >= 56) {
return 1;
}
}
else if (refDCval >= 74) {
if (Rs < 843) {
return 4;
}
else if (Rs >= 843) {
return 8;
}
}
}
else if (posBalance >= 47) {
if (refDCval < 57) {
return 1;
}
else if (refDCval >= 57) {
return 8;
}
}
}
}
}
else if (Cs >= 1198) {
return 4;
}
}
else if (ssDCval >= 94) {
return 2;
}
}
else if (posBalance >= 54) {
if (negBalance < 52) {
if (ssDCval < 85) {
if (MVDiff < 54) {
if (Rs < 1169) {
if (Cs < 986) {
if (diffTSC < 1) {
if (gchDC < 3) {
if (CsDiff < 10) {
if (Rs < 1138) {
if (Cs < 563) {
return 1;
}
else if (Cs >= 563) {
if (MVDiff < 19) {
return 1;
}
else if (MVDiff >= 19) {
if (mu_mv_mag_sq < 11) {
return 4;
}
else if (mu_mv_mag_sq >= 11) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 1;
}
}
}
}
}
else if (Rs >= 1138) {
return 1;
}
}
else if (CsDiff >= 10) {
return 2;
}
}
else if (gchDC >= 3) {
if (Rs < 730) {
return 4;
}
else if (Rs >= 730) {
return 1;
}
}
}
else if (diffTSC >= 1) {
if (MVDiff < 31) {
if (Rs < 1013) {
return 2;
}
else if (Rs >= 1013) {
return 1;
}
}
else if (MVDiff >= 31) {
if (Cs < 715) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 4;
}
}
else if (Cs >= 715) {
return 8;
}
}
}
}
else if (Cs >= 986) {
if (diffTSC < 1) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (Rs >= 1169) {
if (RsDiff < 6) {
if (mu_mv_mag_sq < 47) {
return 8;
}
else if (mu_mv_mag_sq >= 47) {
if (Rs < 1459) {
return 4;
}
else if (Rs >= 1459) {
return 2;
}
}
}
else if (RsDiff >= 6) {
if (AFD < 22) {
if (tcor < 70) {
return 8;
}
else if (tcor >= 70) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
if (Rs < 1385) {
return 4;
}
else if (Rs >= 1385) {
return 8;
}
}
}
}
else if (AFD >= 22) {
return 2;
}
}
}
}
else if (MVDiff >= 54) {
if (RsDiff < 5) {
if (Rs < 414) {
return 2;
}
else if (Rs >= 414) {
return 1;
}
}
else if (RsDiff >= 5) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
return 1;
}
}
}
}
else if (ssDCval >= 85) {
if (Cs < 908) {
if (MVDiff < 31) {
if (Cs < 686) {
return 2;
}
else if (Cs >= 686) {
return 8;
}
}
else if (MVDiff >= 31) {
if (mu_mv_mag_sq < 46) {
if (mcTcor < 94) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (Rs < 886) {
return 8;
}
else if (Rs >= 886) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (Cs < 634) {
return 1;
}
else if (Cs >= 634) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 46) {
return 8;
}
}
}
else if (Cs >= 908) {
if (ssDCval < 93) {
return 1;
}
else if (ssDCval >= 93) {
if (Rs < 969) {
return 4;
}
else if (Rs >= 969) {
return 2;
}
}
}
}
}
else if (negBalance >= 52) {
if (AFD < 19) {
if (Rs < 1468) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 96) {
if (Cs < 1029) {
if (Rs < 677) {
return 2;
}
else if (Rs >= 677) {
if (Rs < 1301) {
if (Cs < 636) {
return 4;
}
else if (Cs >= 636) {
return 8;
}
}
else if (Rs >= 1301) {
return 2;
}
}
}
else if (Cs >= 1029) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
return 2;
}
}
}
else if (ssDCval >= 96) {
if (Rs < 1139) {
return 4;
}
else if (Rs >= 1139) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 9) {
if (SC < 1282) {
if (negBalance < 61) {
return 4;
}
else if (negBalance >= 61) {
return 8;
}
}
else if (SC >= 1282) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
if (diffMVdiffVal < -22) {
return 2;
}
else if (diffMVdiffVal >= -22) {
return 8;
}
}
}
}
else if (RsDiff >= 9) {
return 4;
}
}
}
else if (Rs >= 1468) {
if (ssDCval < 93) {
return 8;
}
else if (ssDCval >= 93) {
return 1;
}
}
}
else if (AFD >= 19) {
if (RsDiff < 9) {
if (diffAFD < 1) {
if (diffRsCsdiff < -1) {
if (MVDiff < 30) {
if (gchDC < 2) {
if (MVDiff < 28) {
return 1;
}
else if (MVDiff >= 28) {
return 8;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (MVDiff >= 30) {
if (refDCval < 74) {
return 1;
}
else if (refDCval >= 74) {
if (Rs < 338) {
return 4;
}
else if (Rs >= 338) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= -1) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 1;
}
}
}
else if (diffAFD >= 1) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (diffAFD < 4) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 1008) {
return 4;
}
else if (Rs >= 1008) {
return 1;
}
}
}
else if (diffAFD >= 4) {
return 1;
}
}
}
}
else if (RsDiff >= 9) {
if (mcTcor < 92) {
if (negBalance < 55) {
if (Cs < 1088) {
return 8;
}
else if (Cs >= 1088) {
return 4;
}
}
else if (negBalance >= 55) {
return 4;
}
}
else if (mcTcor >= 92) {
if (Rs < 903) {
return 2;
}
else if (Rs >= 903) {
return 8;
}
}
}
}
}
}
}
}
else if (AFD >= 23) {
if (RsCsDiff < 148) {
if (SC < 2411) {
if (Cs < 891) {
if (ssDCval < 68) {
if (Cs < 744) {
return 8;
}
else if (Cs >= 744) {
return 4;
}
}
else if (ssDCval >= 68) {
if (tcor < 64) {
if (Cs < 767) {
if (RsCsDiff < 99) {
if (diffRsCsdiff < -14) {
if (Rs < 328) {
return 8;
}
else if (Rs >= 328) {
return 1;
}
}
else if (diffRsCsdiff >= -14) {
if (RsCsDiff < 68) {
if (AFD < 28) {
return 2;
}
else if (AFD >= 28) {
return 1;
}
}
else if (RsCsDiff >= 68) {
return 2;
}
}
}
else if (RsCsDiff >= 99) {
if (tcor < 54) {
if (Rs < 1343) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 4;
}
}
else if (Rs >= 1343) {
return 1;
}
}
else if (tcor >= 54) {
if (Cs < 580) {
return 4;
}
else if (Cs >= 580) {
return 1;
}
}
}
}
else if (Cs >= 767) {
return 2;
}
}
else if (tcor >= 64) {
if (diffRsCsdiff < -34) {
if (Rs < 1022) {
if (Rs < 716) {
return 1;
}
else if (Rs >= 716) {
return 8;
}
}
else if (Rs >= 1022) {
return 4;
}
}
else if (diffRsCsdiff >= -34) {
if (posBalance < 63) {
if (ssDCval < 96) {
if (RsDiff < 6) {
if (Rs < 464) {
return 4;
}
else if (Rs >= 464) {
return 2;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (ssDCval >= 96) {
return 2;
}
}
else if (posBalance >= 63) {
return 2;
}
}
}
}
}
else if (Cs >= 891) {
if (mu_mv_mag_sq < 5) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (RsCsDiff < 110) {
if (ssDCval < 84) {
if (diffTSC < -1) {
if (Rs < 934) {
return 2;
}
else if (Rs >= 934) {
return 1;
}
}
else if (diffTSC >= -1) {
if (AFD < 29) {
return 4;
}
else if (AFD >= 29) {
return 2;
}
}
}
else if (ssDCval >= 84) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
return 1;
}
}
}
else if (RsCsDiff >= 110) {
if (mu_mv_mag_sq < 16) {
return 8;
}
else if (mu_mv_mag_sq >= 16) {
return 1;
}
}
}
}
}
else if (SC >= 2411) {
if (Rs < 1627) {
return 4;
}
else if (Rs >= 1627) {
return 1;
}
}
}
else if (RsCsDiff >= 148) {
if (mcTcor < 92) {
if (Cs < 1166) {
if (AFD < 29) {
if (Cs < 1016) {
return 8;
}
else if (Cs >= 1016) {
if (tcor < 64) {
return 8;
}
else if (tcor >= 64) {
return 4;
}
}
}
else if (AFD >= 29) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 4;
}
}
}
else if (Cs >= 1166) {
return 4;
}
}
else if (mcTcor >= 92) {
if (gchDC < 5) {
if (AFD < 31) {
if (RsCsDiff < 223) {
return 4;
}
else if (RsCsDiff >= 223) {
return 2;
}
}
else if (AFD >= 31) {
return 2;
}
}
else if (gchDC >= 5) {
return 2;
}
}
}
}
}
}
else if (tcor >= 83) {
if (Cs < 1046) {
if (ssDCval < 92) {
if (mu_mv_mag_sq < 3) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (RsCsDiff < 97) {
if (Cs < 695) {
return 1;
}
else if (Cs >= 695) {
if (Cs < 935) {
return 4;
}
else if (Cs >= 935) {
return 8;
}
}
}
else if (RsCsDiff >= 97) {
if (Cs < 880) {
return 2;
}
else if (Cs >= 880) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1531) {
if (diffAFD < 5) {
if (Cs < 1010) {
return 8;
}
else if (Cs >= 1010) {
return 2;
}
}
else if (diffAFD >= 5) {
return 4;
}
}
else if (Rs >= 1531) {
if (RsCsDiff < 69) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
else if (RsCsDiff >= 69) {
return 4;
}
}
}
}
else if (ssDCval >= 92) {
if (negBalance < 51) {
if (Rs < 1358) {
if (Rs < 750) {
return 4;
}
else if (Rs >= 750) {
return 2;
}
}
else if (Rs >= 1358) {
return 1;
}
}
else if (negBalance >= 51) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 8;
}
}
}
}
else if (Cs >= 1046) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (Rs < 1222) {
return 2;
}
else if (Rs >= 1222) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1726) {
if (ssDCval < 72) {
if (Cs < 960) {
if (RsDiff < 9) {
if (Rs < 1953) {
return 8;
}
else if (Rs >= 1953) {
return 2;
}
}
else if (RsDiff >= 9) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 4;
}
}
}
else if (Cs >= 960) {
if (mcTcor < 87) {
if (refDCval < 53) {
return 2;
}
else if (refDCval >= 53) {
if (negBalance < 40) {
if (Rs < 2030) {
return 8;
}
else if (Rs >= 2030) {
return 2;
}
}
else if (negBalance >= 40) {
return 8;
}
}
}
else if (mcTcor >= 87) {
if (AFD < 16) {
if (Cs < 1149) {
return 1;
}
else if (Cs >= 1149) {
if (Rs < 1873) {
return 4;
}
else if (Rs >= 1873) {
if (Cs < 1232) {
return 2;
}
else if (Cs >= 1232) {
return 8;
}
}
}
}
else if (AFD >= 16) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 72) {
if (SC < 3034) {
if (gchDC < 2) {
if (Cs < 689) {
if (ssDCval < 90) {
if (ssDCval < 84) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
else if (ssDCval >= 84) {
if (Rs < 2100) {
return 2;
}
else if (Rs >= 2100) {
return 4;
}
}
}
else if (ssDCval >= 90) {
if (Cs < 680) {
return 1;
}
else if (Cs >= 680) {
return 2;
}
}
}
else if (Cs >= 689) {
if (RsDiff < 14) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (negBalance < 51) {
if (mcTcor < 95) {
if (posBalance < 57) {
if (Rs < 1894) {
return 2;
}
else if (Rs >= 1894) {
return 4;
}
}
else if (posBalance >= 57) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (SC < 2817) {
return 2;
}
else if (SC >= 2817) {
if (Rs < 1851) {
return 8;
}
else if (Rs >= 1851) {
return 1;
}
}
}
}
}
else if (mcTcor >= 95) {
if (RsCsDiff < 73) {
if (Rs < 1840) {
return 1;
}
else if (Rs >= 1840) {
return 2;
}
}
else if (RsCsDiff >= 73) {
return 4;
}
}
}
else if (negBalance >= 51) {
if (diffMVdiffVal < 6) {
if (mu_mv_mag_sq < 22) {
if (TSC < 10) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
if (SC < 2763) {
return 8;
}
else if (SC >= 2763) {
return 4;
}
}
}
else if (TSC >= 10) {
if (Cs < 1149) {
return 1;
}
else if (Cs >= 1149) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 22) {
if (RsDiff < 10) {
if (Rs < 1883) {
return 2;
}
else if (Rs >= 1883) {
return 8;
}
}
else if (RsDiff >= 10) {
if (MVDiff < 20) {
if (Cs < 823) {
return 1;
}
else if (Cs >= 823) {
if (Rs < 1781) {
return 8;
}
else if (Rs >= 1781) {
return 4;
}
}
}
else if (MVDiff >= 20) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 6) {
if (Rs < 1922) {
return 8;
}
else if (Rs >= 1922) {
return 4;
}
}
}
}
}
else if (RsDiff >= 14) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (MVDiff < 37) {
if (mcTcor < 90) {
if (Rs < 1846) {
return 8;
}
else if (Rs >= 1846) {
if (TSC < 9) {
if (Rs < 1931) {
return 2;
}
else if (Rs >= 1931) {
return 8;
}
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (mcTcor >= 90) {
if (ssDCval < 89) {
if (Cs < 994) {
return 2;
}
else if (Cs >= 994) {
return 4;
}
}
else if (ssDCval >= 89) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (Rs < 1815) {
if (Rs < 1770) {
return 1;
}
else if (Rs >= 1770) {
return 8;
}
}
else if (Rs >= 1815) {
return 2;
}
}
}
}
}
else if (MVDiff >= 37) {
if (AFD < 16) {
if (Cs < 772) {
return 2;
}
else if (Cs >= 772) {
return 8;
}
}
else if (AFD >= 16) {
if (diffRsCsdiff < 73) {
if (MVDiff < 120) {
return 8;
}
else if (MVDiff >= 120) {
if (Rs < 1970) {
return 8;
}
else if (Rs >= 1970) {
return 2;
}
}
}
else if (diffRsCsdiff >= 73) {
if (Rs < 1981) {
if (Rs < 1834) {
return 2;
}
else if (Rs >= 1834) {
return 1;
}
}
else if (Rs >= 1981) {
return 8;
}
}
}
}
}
}
else if (SC >= 3034) {
if (RsCsDiff < 129) {
if (AFD < 21) {
if (RsCsDiff < 103) {
if (mu_mv_mag_sq < 56) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
if (SC < 3279) {
return 8;
}
else if (SC >= 3279) {
if (tcor < 83) {
if (Rs < 2360) {
return 4;
}
else if (Rs >= 2360) {
return 8;
}
}
else if (tcor >= 83) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (Rs < 2073) {
return 1;
}
else if (Rs >= 2073) {
return 4;
}
}
}
else if (RsCsDiff >= 103) {
if (tcor < 74) {
return 8;
}
else if (tcor >= 74) {
if (TSC < 9) {
if (Rs < 1811) {
return 8;
}
else if (Rs >= 1811) {
return 2;
}
}
else if (TSC >= 9) {
if (Rs < 1868) {
return 2;
}
else if (Rs >= 1868) {
return 8;
}
}
}
}
}
else if (AFD >= 21) {
if (TSC < 10) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 4;
}
}
else if (TSC >= 10) {
if (Cs < 1019) {
return 8;
}
else if (Cs >= 1019) {
return 2;
}
}
}
}
else if (RsCsDiff >= 129) {
if (mcTcor < 68) {
return 1;
}
else if (mcTcor >= 68) {
if (negBalance < 50) {
if (Rs < 1989) {
if (mu_mv_mag_sq < 5) {
return 4;
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 1892) {
return 4;
}
else if (Rs >= 1892) {
return 2;
}
}
}
else if (Rs >= 1989) {
if (mu_mv_mag_sq < 7) {
if (Cs < 1162) {
return 2;
}
else if (Cs >= 1162) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 1154) {
return 8;
}
else if (Cs >= 1154) {
if (Rs < 2019) {
return 8;
}
else if (Rs >= 2019) {
return 2;
}
}
}
}
}
else if (negBalance >= 50) {
if (negBalance < 59) {
if (mcTcor < 90) {
if (SC < 3160) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
return 4;
}
}
else if (SC >= 3160) {
if (refDCval < 74) {
if (Rs < 2061) {
return 1;
}
else if (Rs >= 2061) {
return 8;
}
}
else if (refDCval >= 74) {
return 4;
}
}
}
else if (mcTcor >= 90) {
if (RsDiff < 12) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (SC < 3063) {
return 8;
}
else if (SC >= 3063) {
return 4;
}
}
}
else if (RsDiff >= 12) {
if (CsDiff < 8) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 4;
}
}
else if (CsDiff >= 8) {
return 8;
}
}
}
}
else if (negBalance >= 59) {
if (tcor < 82) {
if (diffRsCsdiff < 43) {
if (Cs < 968) {
return 1;
}
else if (Cs >= 968) {
if (SC < 3562) {
return 2;
}
else if (SC >= 3562) {
return 4;
}
}
}
else if (diffRsCsdiff >= 43) {
return 1;
}
}
else if (tcor >= 82) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 114) {
if (MVDiff < 17) {
if (tcor < 87) {
if (refDCval < 123) {
if (mcTcor < 94) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (mcTcor >= 94) {
if (RsDiff < 5) {
if (Rs < 876) {
return 2;
}
else if (Rs >= 876) {
return 4;
}
}
else if (RsDiff >= 5) {
if (Cs < 962) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (Cs >= 962) {
return 8;
}
}
}
}
else if (refDCval >= 123) {
if (negBalance < 63) {
if (diffAFD < 4) {
if (ssDCval < 169) {
if (Rs < 646) {
if (Rs < 551) {
return 2;
}
else if (Rs >= 551) {
if (Rs < 614) {
return 8;
}
else if (Rs >= 614) {
return 1;
}
}
}
else if (Rs >= 646) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
else if (ssDCval >= 169) {
return 2;
}
}
else if (diffAFD >= 4) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
return 8;
}
}
}
else if (negBalance >= 63) {
if (Rs < 1713) {
return 2;
}
else if (Rs >= 1713) {
return 8;
}
}
}
}
else if (tcor >= 87) {
if (MVDiff < 12) {
if (Rs < 1299) {
return 2;
}
else if (Rs >= 1299) {
return 8;
}
}
else if (MVDiff >= 12) {
if (Cs < 1231) {
return 1;
}
else if (Cs >= 1231) {
return 8;
}
}
}
}
else if (MVDiff >= 17) {
if (tcor < 81) {
if (tcor < 38) {
if (refDCval < 123) {
return 2;
}
else if (refDCval >= 123) {
if (TSC < 17) {
if (Rs < 382) {
return 2;
}
else if (Rs >= 382) {
return 1;
}
}
else if (TSC >= 17) {
if (Rs < 1160) {
return 2;
}
else if (Rs >= 1160) {
return 4;
}
}
}
}
else if (tcor >= 38) {
if (Rs < 674) {
if (refDCval < 121) {
if (refDCval < 113) {
if (Rs < 476) {
return 1;
}
else if (Rs >= 476) {
return 8;
}
}
else if (refDCval >= 113) {
return 4;
}
}
else if (refDCval >= 121) {
return 2;
}
}
else if (Rs >= 674) {
if (mcTcor < 88) {
if (gchDC < 13) {
if (Rs < 890) {
if (Rs < 815) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (Rs < 759) {
return 8;
}
else if (Rs >= 759) {
return 1;
}
}
}
else if (Rs >= 815) {
return 2;
}
}
else if (Rs >= 890) {
if (negBalance < 52) {
if (SC < 2201) {
return 8;
}
else if (SC >= 2201) {
if (posBalance < 68) {
if (ssDCval < 150) {
return 1;
}
else if (ssDCval >= 150) {
return 8;
}
}
else if (posBalance >= 68) {
return 8;
}
}
}
else if (negBalance >= 52) {
if (mu_mv_mag_sq < 17) {
if (tcor < 61) {
return 8;
}
else if (tcor >= 61) {
if (Cs < 566) {
return 8;
}
else if (Cs >= 566) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
return 8;
}
}
}
}
else if (gchDC >= 13) {
if (TSC < 20) {
return 4;
}
else if (TSC >= 20) {
if (Rs < 1352) {
return 2;
}
else if (Rs >= 1352) {
return 8;
}
}
}
}
else if (mcTcor >= 88) {
if (Rs < 1802) {
if (RsCsDiff < 110) {
if (negBalance < 56) {
if (negBalance < 53) {
return 8;
}
else if (negBalance >= 53) {
if (Rs < 1479) {
return 4;
}
else if (Rs >= 1479) {
return 8;
}
}
}
else if (negBalance >= 56) {
return 1;
}
}
else if (RsCsDiff >= 110) {
if (Cs < 1001) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (CsDiff < 5) {
if (Rs < 1293) {
return 2;
}
else if (Rs >= 1293) {
return 8;
}
}
else if (CsDiff >= 5) {
return 1;
}
}
}
else if (Cs >= 1001) {
if (RsCsDiff < 123) {
return 4;
}
else if (RsCsDiff >= 123) {
return 8;
}
}
}
}
else if (Rs >= 1802) {
if (mu_mv_mag_sq < 52) {
if (gchDC < 7) {
if (CsDiff < 9) {
return 1;
}
else if (CsDiff >= 9) {
return 2;
}
}
else if (gchDC >= 7) {
return 2;
}
}
else if (mu_mv_mag_sq >= 52) {
if (Rs < 1975) {
return 8;
}
else if (Rs >= 1975) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 81) {
if (RsCsDiff < 133) {
if (SC < 2179) {
if (refDCval < 138) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (refDCval >= 138) {
if (Cs < 1083) {
return 2;
}
else if (Cs >= 1083) {
return 1;
}
}
}
else if (SC >= 2179) {
if (TSC < 10) {
if (mcTcor < 95) {
if (Rs < 1334) {
return 1;
}
else if (Rs >= 1334) {
return 8;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
else if (TSC >= 10) {
if (diffRsCsdiff < 22) {
return 1;
}
else if (diffRsCsdiff >= 22) {
return 8;
}
}
}
}
else if (RsCsDiff >= 133) {
if (ssDCval < 131) {
return 1;
}
else if (ssDCval >= 131) {
if (diffMVdiffVal < 6) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 1;
}
}
else if (diffMVdiffVal >= 6) {
if (Rs < 1830) {
return 8;
}
else if (Rs >= 1830) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (SC < 2388) {
if (Rs < 1129) {
if (refDCval < 63) {
if (MVDiff < 1) {
if (Rs < 546) {
if (SC < 1109) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (Rs < 461) {
return 1;
}
else if (Rs >= 461) {
return 8;
}
}
}
else if (SC >= 1109) {
if (Rs < 388) {
if (Cs < 896) {
return 4;
}
else if (Cs >= 896) {
return 2;
}
}
else if (Rs >= 388) {
if (diffRsCsdiff < -1) {
if (Rs < 513) {
return 8;
}
else if (Rs >= 513) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
}
}
}
else if (Rs >= 546) {
if (SC < 1606) {
if (ssDCval < 57) {
if (mu_mv_mag_sq < 117) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (mu_mv_mag_sq < 92) {
return 8;
}
else if (mu_mv_mag_sq >= 92) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (Rs < 993) {
return 8;
}
else if (Rs >= 993) {
return 4;
}
}
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (SC >= 1606) {
if (ssDCval < 53) {
if (Rs < 912) {
if (Cs < 988) {
if (RsDiff < 1) {
if (Rs < 832) {
return 1;
}
else if (Rs >= 832) {
return 8;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (Cs >= 988) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (Rs >= 912) {
if (mu_mv_mag_sq < 110) {
if (Rs < 1006) {
return 8;
}
else if (Rs >= 1006) {
if (Rs < 1085) {
return 1;
}
else if (Rs >= 1085) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 110) {
return 8;
}
}
}
else if (ssDCval >= 53) {
if (Rs < 849) {
if (Rs < 738) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (Rs >= 738) {
return 8;
}
}
else if (Rs >= 849) {
return 1;
}
}
}
}
}
else if (MVDiff >= 1) {
if (AFD < 1) {
if (Rs < 849) {
return 1;
}
else if (Rs >= 849) {
return 8;
}
}
else if (AFD >= 1) {
if (Rs < 805) {
if (negBalance < 29) {
if (TSC < 3) {
if (ssDCval < 40) {
return 2;
}
else if (ssDCval >= 40) {
if (SC < 1380) {
if (posBalance < 49) {
if (MVDiff < 3) {
if (Cs < 602) {
if (Rs < 511) {
return 1;
}
else if (Rs >= 511) {
return 4;
}
}
else if (Cs >= 602) {
return 2;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (posBalance >= 49) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 2;
}
}
}
}
else if (SC >= 1380) {
if (Cs < 1009) {
return 1;
}
else if (Cs >= 1009) {
return 4;
}
}
}
}
else if (TSC >= 3) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (Rs < 625) {
return 8;
}
else if (Rs >= 625) {
return 4;
}
}
}
}
}
else if (negBalance >= 29) {
if (AFD < 9) {
if (SC < 1209) {
if (Cs < 544) {
return 4;
}
else if (Cs >= 544) {
if (diffMVdiffVal < 3) {
if (CsDiff < 1) {
if (Rs < 475) {
return 4;
}
else if (Rs >= 475) {
return 8;
}
}
else if (CsDiff >= 1) {
if (diffRsCsdiff < -2) {
if (Rs < 582) {
return 4;
}
else if (Rs >= 582) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
}
else if (SC >= 1209) {
if (refDCval < 50) {
if (Cs < 629) {
return 8;
}
else if (Cs >= 629) {
return 4;
}
}
else if (refDCval >= 50) {
if (RsCsDiff < 16) {
if (Cs < 1064) {
if (MVDiff < 10) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 99) {
if (posBalance < 51) {
if (negBalance < 41) {
return 4;
}
else if (negBalance >= 41) {
return 8;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (mu_mv_mag_sq >= 99) {
return 8;
}
}
}
else if (MVDiff >= 10) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Cs < 639) {
return 4;
}
else if (Cs >= 639) {
return 1;
}
}
}
}
else if (Cs >= 1064) {
return 1;
}
}
else if (RsCsDiff >= 16) {
if (Rs < 585) {
return 1;
}
else if (Rs >= 585) {
return 4;
}
}
}
}
}
else if (AFD >= 9) {
if (SC < 1439) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
if (Rs < 412) {
return 8;
}
else if (Rs >= 412) {
return 2;
}
}
}
else if (SC >= 1439) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (negBalance < 42) {
if (Rs < 710) {
return 2;
}
else if (Rs >= 710) {
return 4;
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
}
}
}
else if (Rs >= 805) {
if (ssDCval < 51) {
if (Cs < 1201) {
if (Cs < 622) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (Cs >= 622) {
if (Cs < 1067) {
return 8;
}
else if (Cs >= 1067) {
if (Rs < 913) {
return 1;
}
else if (Rs >= 913) {
return 8;
}
}
}
}
else if (Cs >= 1201) {
if (Cs < 1252) {
return 4;
}
else if (Cs >= 1252) {
return 8;
}
}
}
else if (ssDCval >= 51) {
if (posBalance < 45) {
if (negBalance < 45) {
if (RsCsDiff < 12) {
return 1;
}
else if (RsCsDiff >= 12) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (Rs < 1021) {
return 4;
}
else if (Rs >= 1021) {
return 1;
}
}
}
}
else if (negBalance >= 45) {
return 4;
}
}
else if (posBalance >= 45) {
if (AFD < 6) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
return 1;
}
}
else if (AFD >= 6) {
if (negBalance < 35) {
if (Rs < 931) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (Rs >= 931) {
return 1;
}
}
else if (negBalance >= 35) {
if (ssDCval < 58) {
if (Rs < 1033) {
return 1;
}
else if (Rs >= 1033) {
return 8;
}
}
else if (ssDCval >= 58) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 63) {
if (SC < 2155) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (negBalance < 28) {
if (Cs < 701) {
if (refDCval < 64) {
if (mu_mv_mag_sq < 109) {
if (Rs < 305) {
return 4;
}
else if (Rs >= 305) {
return 1;
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
else if (refDCval >= 64) {
if (ssDCval < 72) {
if (Rs < 708) {
return 4;
}
else if (Rs >= 708) {
if (SC < 1782) {
if (mu_mv_mag_sq < 89) {
return 1;
}
else if (mu_mv_mag_sq >= 89) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
return 1;
}
}
}
else if (SC >= 1782) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (Rs < 1110) {
return 4;
}
else if (Rs >= 1110) {
return 2;
}
}
}
}
}
else if (ssDCval >= 72) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
else if (Cs >= 701) {
if (ssDCval < 81) {
if (Rs < 686) {
if (SC < 1279) {
return 4;
}
else if (SC >= 1279) {
return 8;
}
}
else if (Rs >= 686) {
if (Cs < 761) {
if (Cs < 718) {
if (Rs < 1075) {
return 2;
}
else if (Rs >= 1075) {
return 1;
}
}
else if (Cs >= 718) {
if (posBalance < 41) {
if (Rs < 874) {
return 8;
}
else if (Rs >= 874) {
return 2;
}
}
else if (posBalance >= 41) {
if (diffRsCsdiff < 2) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 819) {
return 4;
}
else if (Rs >= 819) {
return 2;
}
}
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
}
}
else if (Cs >= 761) {
if (ssDCval < 75) {
if (mu_mv_mag_sq < 112) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (TSC < 4) {
if (Cs < 1044) {
return 1;
}
else if (Cs >= 1044) {
return 2;
}
}
else if (TSC >= 4) {
if (Rs < 962) {
return 4;
}
else if (Rs >= 962) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (ssDCval < 71) {
if (Rs < 1078) {
if (tcor < 100) {
return 4;
}
else if (tcor >= 100) {
return 8;
}
}
else if (Rs >= 1078) {
return 1;
}
}
else if (ssDCval >= 71) {
return 1;
}
}
}
else if (ssDCval >= 75) {
if (Rs < 830) {
return 1;
}
else if (Rs >= 830) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (Rs < 652) {
if (mu_mv_mag_sq < 91) {
if (gchDC < 1) {
if (ssDCval < 83) {
return 1;
}
else if (ssDCval >= 83) {
return 2;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 91) {
return 2;
}
}
else if (Rs >= 652) {
if (Cs < 798) {
return 1;
}
else if (Cs >= 798) {
return 2;
}
}
}
}
}
else if (negBalance >= 28) {
if (MVDiff < 2) {
if (ssDCval < 81) {
if (refDCval < 75) {
if (negBalance < 41) {
if (RsDiff < 1) {
if (Rs < 710) {
if (Cs < 675) {
if (Rs < 610) {
return 8;
}
else if (Rs >= 610) {
return 1;
}
}
else if (Cs >= 675) {
return 8;
}
}
else if (Rs >= 710) {
return 4;
}
}
else if (RsDiff >= 1) {
if (Cs < 1040) {
if (Cs < 591) {
if (tcor < 98) {
if (Rs < 535) {
return 2;
}
else if (Rs >= 535) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (tcor >= 98) {
return 2;
}
}
else if (Cs >= 591) {
if (Rs < 661) {
if (Cs < 636) {
return 1;
}
else if (Cs >= 636) {
return 8;
}
}
else if (Rs >= 661) {
if (mcTcor < 98) {
if (SC < 1598) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (SC >= 1598) {
return 4;
}
}
else if (mcTcor >= 98) {
if (diffAFD < -1) {
if (Rs < 793) {
return 1;
}
else if (Rs >= 793) {
return 4;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
}
else if (Cs >= 1040) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (Rs < 968) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
if (Rs < 598) {
return 8;
}
else if (Rs >= 598) {
return 1;
}
}
}
else if (Rs >= 968) {
return 8;
}
}
}
}
}
else if (negBalance >= 41) {
if (Cs < 768) {
if (Rs < 525) {
return 8;
}
else if (Rs >= 525) {
return 4;
}
}
else if (Cs >= 768) {
return 8;
}
}
}
else if (refDCval >= 75) {
if (RsCsDiff < 4) {
if (posBalance < 48) {
if (Rs < 813) {
return 1;
}
else if (Rs >= 813) {
return 8;
}
}
else if (posBalance >= 48) {
if (Rs < 650) {
return 8;
}
else if (Rs >= 650) {
return 4;
}
}
}
else if (RsCsDiff >= 4) {
if (mu_mv_mag_sq < 74) {
if (Rs < 644) {
return 8;
}
else if (Rs >= 644) {
return 4;
}
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 1058) {
return 8;
}
else if (Cs >= 1058) {
if (Cs < 1074) {
return 4;
}
else if (Cs >= 1074) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (posBalance < 44) {
if (Rs < 633) {
return 2;
}
else if (Rs >= 633) {
return 1;
}
}
else if (posBalance >= 44) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 14) {
if (Cs < 793) {
if (ssDCval < 72) {
if (AFD < 8) {
if (Rs < 932) {
if (Rs < 297) {
return 1;
}
else if (Rs >= 297) {
return 4;
}
}
else if (Rs >= 932) {
return 1;
}
}
else if (AFD >= 8) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 648) {
return 8;
}
else if (Rs >= 648) {
return 4;
}
}
}
else if (Cs >= 793) {
if (AFD < 17) {
if (diffRsCsdiff < -6) {
if (refDCval < 67) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (Rs < 664) {
return 1;
}
else if (Rs >= 664) {
return 8;
}
}
}
else if (refDCval >= 67) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 4;
}
}
}
else if (diffRsCsdiff >= -6) {
if (ssDCval < 65) {
if (Rs < 805) {
return 2;
}
else if (Rs >= 805) {
return 4;
}
}
else if (ssDCval >= 65) {
return 2;
}
}
}
else if (AFD >= 17) {
if (Rs < 843) {
return 4;
}
else if (Rs >= 843) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (ssDCval < 85) {
if (mu_mv_mag_sq < 87) {
if (TSC < 4) {
if (mcTcor < 98) {
if (SC < 1500) {
return 8;
}
else if (SC >= 1500) {
if (SC < 1658) {
return 2;
}
else if (SC >= 1658) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (tcor < 99) {
if (posBalance < 31) {
return 4;
}
else if (posBalance >= 31) {
return 1;
}
}
else if (tcor >= 99) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (Cs < 936) {
if (tcor < 97) {
if (mcTcor < 98) {
if (MVDiff < 15) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
return 8;
}
}
else if (MVDiff >= 15) {
if (Rs < 580) {
return 1;
}
else if (Rs >= 580) {
return 4;
}
}
}
else if (mcTcor >= 98) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (Rs < 915) {
return 4;
}
else if (Rs >= 915) {
return 1;
}
}
}
}
else if (tcor >= 97) {
return 8;
}
}
else if (Cs >= 936) {
if (Cs < 1156) {
if (Rs < 1028) {
return 2;
}
else if (Rs >= 1028) {
return 8;
}
}
else if (Cs >= 1156) {
if (Rs < 605) {
return 1;
}
else if (Rs >= 605) {
if (Cs < 1235) {
return 4;
}
else if (Cs >= 1235) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (MVDiff < 3) {
if (Cs < 773) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (Cs >= 773) {
if (mu_mv_mag_sq < 96) {
return 8;
}
else if (mu_mv_mag_sq >= 96) {
if (Cs < 1120) {
return 1;
}
else if (Cs >= 1120) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (negBalance < 31) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (negBalance >= 31) {
return 8;
}
}
}
}
else if (ssDCval >= 85) {
if (Rs < 680) {
if (mu_mv_mag_sq < 96) {
return 2;
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 581) {
return 2;
}
else if (Rs >= 581) {
return 1;
}
}
}
else if (Rs >= 680) {
if (Rs < 782) {
return 1;
}
else if (Rs >= 782) {
return 4;
}
}
}
}
}
}
}
}
else if (SC >= 2155) {
if (tcor < 98) {
if (RsDiff < 2) {
if (Rs < 1001) {
return 2;
}
else if (Rs >= 1001) {
return 8;
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 29) {
return 2;
}
else if (mu_mv_mag_sq >= 29) {
if (diffAFD < -2) {
if (Rs < 1062) {
return 8;
}
else if (Rs >= 1062) {
return 4;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
}
}
else if (tcor >= 98) {
if (mu_mv_mag_sq < 121) {
return 4;
}
else if (mu_mv_mag_sq >= 121) {
if (RsCsDiff < 4) {
return 8;
}
else if (RsCsDiff >= 4) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1129) {
if (mu_mv_mag_sq < 122) {
if (Cs < 988) {
if (TSC < 5) {
if (ssDCval < 69) {
if (tcor < 92) {
if (Cs < 749) {
if (Rs < 1412) {
return 2;
}
else if (Rs >= 1412) {
return 4;
}
}
else if (Cs >= 749) {
if (Rs < 1267) {
return 1;
}
else if (Rs >= 1267) {
return 8;
}
}
}
else if (tcor >= 92) {
if (TSC < 3) {
if (ssDCval < 59) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (SC < 2347) {
if (negBalance < 16) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (negBalance >= 16) {
if (Rs < 1344) {
return 1;
}
else if (Rs >= 1344) {
if (Rs < 1370) {
return 8;
}
else if (Rs >= 1370) {
return 1;
}
}
}
}
else if (SC >= 2347) {
return 8;
}
}
}
else if (ssDCval >= 59) {
return 1;
}
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 108) {
if (mcTcor < 100) {
if (Rs < 1268) {
if (SC < 2151) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (SC >= 2151) {
return 8;
}
}
else if (Rs >= 1268) {
return 1;
}
}
else if (mcTcor >= 100) {
if (Rs < 1452) {
return 1;
}
else if (Rs >= 1452) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (Rs < 1322) {
return 1;
}
else if (Rs >= 1322) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 72) {
return 1;
}
else if (mu_mv_mag_sq >= 72) {
if (SC < 1973) {
return 1;
}
else if (SC >= 1973) {
if (Cs < 636) {
if (Rs < 1757) {
return 8;
}
else if (Rs >= 1757) {
return 4;
}
}
else if (Cs >= 636) {
if (Rs < 1378) {
return 8;
}
else if (Rs >= 1378) {
if (Rs < 1694) {
return 1;
}
else if (Rs >= 1694) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 59) {
if (tcor < 93) {
if (negBalance < 50) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 1214) {
return 4;
}
else if (Rs >= 1214) {
return 1;
}
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (tcor >= 93) {
if (Rs < 1590) {
return 1;
}
else if (Rs >= 1590) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
if (Rs < 1519) {
return 4;
}
else if (Rs >= 1519) {
return 1;
}
}
}
}
}
else if (Cs >= 988) {
if (mu_mv_mag_sq < 105) {
if (Cs < 1176) {
if (Cs < 994) {
return 4;
}
else if (Cs >= 994) {
return 8;
}
}
else if (Cs >= 1176) {
if (Rs < 1144) {
return 8;
}
else if (Rs >= 1144) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 1395) {
if (Cs < 735) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
if (mu_mv_mag_sq < 128) {
if (Cs < 565) {
return 8;
}
else if (Cs >= 565) {
if (ssDCval < 50) {
if (Rs < 1345) {
return 8;
}
else if (Rs >= 1345) {
return 4;
}
}
else if (ssDCval >= 50) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 706) {
return 8;
}
else if (Cs >= 706) {
return 4;
}
}
}
}
else if (Cs >= 735) {
if (Cs < 1056) {
if (negBalance < 19) {
if (Cs < 959) {
if (posBalance < 36) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (posBalance >= 36) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (Cs < 873) {
if (Rs < 1232) {
return 1;
}
else if (Rs >= 1232) {
return 8;
}
}
else if (Cs >= 873) {
return 1;
}
}
}
}
else if (Cs >= 959) {
if (Rs < 1355) {
return 1;
}
else if (Rs >= 1355) {
return 8;
}
}
}
else if (negBalance >= 19) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 1224) {
return 8;
}
else if (Rs >= 1224) {
return 1;
}
}
}
}
else if (Cs >= 1056) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (ssDCval < 72) {
if (Cs < 1181) {
return 4;
}
else if (Cs >= 1181) {
return 8;
}
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
}
else if (Rs >= 1395) {
if (ssDCval < 62) {
if (Rs < 1597) {
if (Rs < 1418) {
return 8;
}
else if (Rs >= 1418) {
return 4;
}
}
else if (Rs >= 1597) {
return 1;
}
}
else if (ssDCval >= 62) {
return 8;
}
}
}
}
}
else if (SC >= 2388) {
if (TSC < 3) {
if (refDCval < 82) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 102) {
return 4;
}
else if (mu_mv_mag_sq >= 102) {
if (MVDiff < 1) {
if (Cs < 1161) {
if (TSCindex < 3) {
if (Rs < 2025) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
if (Cs < 1060) {
if (Cs < 1052) {
if (posBalance < 33) {
return 4;
}
else if (posBalance >= 33) {
if (Cs < 1037) {
if (Rs < 1746) {
return 4;
}
else if (Rs >= 1746) {
if (Cs < 1031) {
return 4;
}
else if (Cs >= 1031) {
return 4;
}
}
}
else if (Cs >= 1037) {
if (Rs < 1790) {
return 8;
}
else if (Rs >= 1790) {
return 4;
}
}
}
}
else if (Cs >= 1052) {
return 4;
}
}
else if (Cs >= 1060) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (Cs < 1156) {
if (Cs < 1151) {
if (Cs < 1147) {
if (Cs < 1143) {
if (diffAFD < 1) {
if (Cs < 1136) {
if (Cs < 1131) {
if (posBalance < 40) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 36) {
if (Rs < 1679) {
return 8;
}
else if (Rs >= 1679) {
if (Rs < 1788) {
return 4;
}
else if (Rs >= 1788) {
return 4;
}
}
}
else if (posBalance >= 36) {
return 8;
}
}
else if (mu_mv_mag_sq >= 128) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (AFD < 2) {
if (Cs < 1090) {
if (Rs < 1782) {
return 8;
}
else if (Rs >= 1782) {
return 8;
}
}
else if (Cs >= 1090) {
if (Cs < 1113) {
return 4;
}
else if (Cs >= 1113) {
if (Rs < 1704) {
return 8;
}
else if (Rs >= 1704) {
return 8;
}
}
}
}
else if (AFD >= 2) {
return 8;
}
}
}
}
else if (posBalance >= 40) {
if (Rs < 1762) {
return 4;
}
else if (Rs >= 1762) {
if (Rs < 1806) {
return 4;
}
else if (Rs >= 1806) {
return 4;
}
}
}
}
else if (Cs >= 1131) {
if (Rs < 1769) {
return 4;
}
else if (Rs >= 1769) {
if (Rs < 1779) {
return 8;
}
else if (Rs >= 1779) {
return 4;
}
}
}
}
else if (Cs >= 1136) {
if (Rs < 1714) {
return 8;
}
else if (Rs >= 1714) {
return 8;
}
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (Cs >= 1143) {
return 4;
}
}
else if (Cs >= 1147) {
return 8;
}
}
else if (Cs >= 1151) {
return 4;
}
}
else if (Cs >= 1156) {
return 8;
}
}
}
}
}
else if (Rs >= 2025) {
return 8;
}
}
else if (TSCindex >= 3) {
if (negBalance < 25) {
if (RsCsDiff < 7) {
if (SC < 2794) {
return 4;
}
else if (SC >= 2794) {
if (mu_mv_mag_sq < 128) {
if (SC < 2839) {
if (Rs < 1704) {
if (Rs < 1680) {
return 4;
}
else if (Rs >= 1680) {
return 8;
}
}
else if (Rs >= 1704) {
return 4;
}
}
else if (SC >= 2839) {
return 4;
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
if (Rs < 1773) {
return 4;
}
else if (Rs >= 1773) {
return 4;
}
}
}
}
else if (negBalance >= 25) {
if (Rs < 1593) {
return 8;
}
else if (Rs >= 1593) {
return 8;
}
}
}
}
else if (Cs >= 1161) {
if (Rs < 1688) {
if (negBalance < 20) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 1614) {
return 8;
}
else if (Rs >= 1614) {
return 4;
}
}
}
else if (negBalance >= 20) {
return 4;
}
}
else if (Rs >= 1688) {
if (Rs < 2177) {
return 4;
}
else if (Rs >= 2177) {
return 8;
}
}
}
}
else if (MVDiff >= 1) {
if (SC < 2464) {
return 1;
}
else if (SC >= 2464) {
return 8;
}
}
}
}
else if (ssDCval >= 56) {
if (Cs < 993) {
if (Rs < 1727) {
if (Cs < 863) {
if (Rs < 1688) {
return 4;
}
else if (Rs >= 1688) {
return 1;
}
}
else if (Cs >= 863) {
return 8;
}
}
else if (Rs >= 1727) {
if (AFD < 4) {
if (TSCindex < 2) {
if (Rs < 1940) {
return 1;
}
else if (Rs >= 1940) {
return 4;
}
}
else if (TSCindex >= 2) {
if (diffRsCsdiff < -2) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
if (Cs < 759) {
return 1;
}
else if (Cs >= 759) {
return 8;
}
}
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (Cs >= 993) {
if (RsDiff < 1) {
if (Rs < 1910) {
if (Cs < 1048) {
return 4;
}
else if (Cs >= 1048) {
return 1;
}
}
else if (Rs >= 1910) {
if (Rs < 2111) {
return 4;
}
else if (Rs >= 2111) {
if (Rs < 2245) {
return 8;
}
else if (Rs >= 2245) {
return 2;
}
}
}
}
else if (RsDiff >= 1) {
if (tcor < 99) {
if (mu_mv_mag_sq < 100) {
if (AFD < 4) {
if (posBalance < 28) {
return 1;
}
else if (posBalance >= 28) {
if (ssDCval < 59) {
if (Rs < 1890) {
return 4;
}
else if (Rs >= 1890) {
return 1;
}
}
else if (ssDCval >= 59) {
return 4;
}
}
}
else if (AFD >= 4) {
return 2;
}
}
else if (mu_mv_mag_sq >= 100) {
if (mu_mv_mag_sq < 126) {
if (SC < 2907) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (SC >= 2907) {
if (Cs < 1072) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
return 4;
}
}
else if (Cs >= 1072) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (tcor < 98) {
if (Rs < 1336) {
return 1;
}
else if (Rs >= 1336) {
return 8;
}
}
else if (tcor >= 98) {
return 4;
}
}
}
}
else if (tcor >= 99) {
if (SC < 3050) {
if (negBalance < 35) {
if (negBalance < 19) {
if (Cs < 1096) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (Cs >= 1096) {
return 1;
}
}
else if (negBalance >= 19) {
return 4;
}
}
else if (negBalance >= 35) {
return 1;
}
}
else if (SC >= 3050) {
if (Cs < 1104) {
return 4;
}
else if (Cs >= 1104) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 82) {
if (Rs < 2195) {
if (Cs < 981) {
return 4;
}
else if (Cs >= 981) {
return 8;
}
}
else if (Rs >= 2195) {
return 8;
}
}
}
else if (TSC >= 3) {
if (Cs < 871) {
if (Rs < 1982) {
if (mu_mv_mag_sq < 93) {
if (ssDCval < 57) {
if (Rs < 1704) {
return 8;
}
else if (Rs >= 1704) {
return 4;
}
}
else if (ssDCval >= 57) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (Rs < 1891) {
return 2;
}
else if (Rs >= 1891) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (ssDCval < 77) {
if (Cs < 812) {
return 1;
}
else if (Cs >= 812) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (ssDCval >= 77) {
return 4;
}
}
}
else if (Rs >= 1982) {
if (posBalance < 46) {
if (TSC < 6) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (Rs < 2038) {
return 1;
}
else if (Rs >= 2038) {
return 4;
}
}
}
}
else if (TSC >= 6) {
if (Cs < 839) {
return 2;
}
else if (Cs >= 839) {
return 4;
}
}
}
else if (posBalance >= 46) {
if (diffMVdiffVal < 4) {
if (posBalance < 52) {
if (ssDCval < 71) {
if (SC < 2747) {
return 1;
}
else if (SC >= 2747) {
return 2;
}
}
else if (ssDCval >= 71) {
return 1;
}
}
else if (posBalance >= 52) {
return 2;
}
}
else if (diffMVdiffVal >= 4) {
return 1;
}
}
}
}
else if (Cs >= 871) {
if (MVDiff < 6) {
if (Rs < 1357) {
if (tcor < 98) {
if (mu_mv_mag_sq < 121) {
if (mu_mv_mag_sq < 106) {
if (Rs < 1190) {
if (Rs < 1135) {
return 4;
}
else if (Rs >= 1135) {
return 8;
}
}
else if (Rs >= 1190) {
return 4;
}
}
else if (mu_mv_mag_sq >= 106) {
return 8;
}
}
else if (mu_mv_mag_sq >= 121) {
return 4;
}
}
else if (tcor >= 98) {
if (Cs < 1164) {
if (Rs < 1289) {
return 8;
}
else if (Rs >= 1289) {
return 4;
}
}
else if (Cs >= 1164) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
else if (Rs >= 1357) {
if (posBalance < 45) {
if (refDCval < 61) {
return 4;
}
else if (refDCval >= 61) {
if (mu_mv_mag_sq < 94) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
if (refDCval < 83) {
return 1;
}
else if (refDCval >= 83) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (Rs < 2215) {
return 8;
}
else if (Rs >= 2215) {
return 4;
}
}
}
}
else if (posBalance >= 45) {
if (posBalance < 53) {
if (Rs < 1953) {
return 8;
}
else if (Rs >= 1953) {
if (Rs < 2058) {
return 1;
}
else if (Rs >= 2058) {
return 8;
}
}
}
else if (posBalance >= 53) {
if (diffAFD < 1) {
if (Cs < 1053) {
if (Cs < 1031) {
return 8;
}
else if (Cs >= 1031) {
return 2;
}
}
else if (Cs >= 1053) {
return 8;
}
}
else if (diffAFD >= 1) {
if (Rs < 1553) {
return 4;
}
else if (Rs >= 1553) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (TSC < 4) {
if (SC < 3333) {
if (SC < 2469) {
return 1;
}
else if (SC >= 2469) {
return 4;
}
}
else if (SC >= 3333) {
return 8;
}
}
else if (TSC >= 4) {
if (tcor < 89) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
if (SC < 2510) {
return 8;
}
else if (SC >= 2510) {
return 2;
}
}
}
else if (tcor >= 89) {
if (Cs < 969) {
return 1;
}
else if (Cs >= 969) {
if (diffAFD < -2) {
if (Rs < 2113) {
return 1;
}
else if (Rs >= 2113) {
return 8;
}
}
else if (diffAFD >= -2) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 2081) {
return 4;
}
else if (Rs >= 2081) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 98) {
if (SC < 1302) {
if (MVDiff < 3) {
if (Cs < 742) {
if (ssDCval < 117) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
if (posBalance < 62) {
if (Rs < 541) {
return 8;
}
else if (Rs >= 541) {
return 2;
}
}
else if (posBalance >= 62) {
return 1;
}
}
}
else if (ssDCval >= 117) {
if (posBalance < 53) {
if (Rs < 604) {
return 2;
}
else if (Rs >= 604) {
return 4;
}
}
else if (posBalance >= 53) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
else if (Cs >= 742) {
if (Rs < 420) {
return 4;
}
else if (Rs >= 420) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (posBalance < 34) {
if (MVDiff < 14) {
if (CsDiff < 1) {
if (Rs < 459) {
return 4;
}
else if (Rs >= 459) {
return 8;
}
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (posBalance >= 34) {
if (TSC < 4) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
if (ssDCval < 146) {
if (negBalance < 25) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 94) {
if (Rs < 358) {
return 2;
}
else if (Rs >= 358) {
return 1;
}
}
else if (mu_mv_mag_sq >= 94) {
if (Rs < 515) {
return 4;
}
else if (Rs >= 515) {
return 2;
}
}
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (negBalance >= 25) {
if (diffTSC < -1) {
if (Cs < 793) {
return 2;
}
else if (Cs >= 793) {
if (Rs < 365) {
return 4;
}
else if (Rs >= 365) {
return 8;
}
}
}
else if (diffTSC >= -1) {
return 2;
}
}
}
else if (ssDCval >= 146) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (Rs < 536) {
return 8;
}
else if (Rs >= 536) {
return 4;
}
}
}
}
}
else if (TSC >= 4) {
if (posBalance < 50) {
if (negBalance < 58) {
return 2;
}
else if (negBalance >= 58) {
if (Cs < 594) {
return 2;
}
else if (Cs >= 594) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (Rs < 504) {
if (Rs < 443) {
if (MVDiff < 38) {
return 4;
}
else if (MVDiff >= 38) {
if (Rs < 302) {
return 8;
}
else if (Rs >= 302) {
return 1;
}
}
}
else if (Rs >= 443) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (Rs >= 504) {
if (mu_mv_mag_sq < 9) {
if (Rs < 572) {
return 2;
}
else if (Rs >= 572) {
return 4;
}
}
else if (mu_mv_mag_sq >= 9) {
if (Cs < 684) {
if (SC < 1070) {
return 8;
}
else if (SC >= 1070) {
return 2;
}
}
else if (Cs >= 684) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 1302) {
if (Cs < 648) {
if (SC < 2381) {
if (mu_mv_mag_sq < 5) {
if (CsDiff < 5) {
if (posBalance < 52) {
if (Rs < 883) {
return 1;
}
else if (Rs >= 883) {
return 8;
}
}
else if (posBalance >= 52) {
return 1;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
if (SC < 1938) {
if (tcor < 93) {
if (negBalance < 55) {
return 2;
}
else if (negBalance >= 55) {
if (Rs < 1218) {
return 8;
}
else if (Rs >= 1218) {
return 1;
}
}
}
else if (tcor >= 93) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (SC >= 1938) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 1;
}
}
}
}
else if (SC >= 2381) {
if (posBalance < 45) {
if (refDCval < 113) {
if (refDCval < 98) {
return 1;
}
else if (refDCval >= 98) {
return 8;
}
}
else if (refDCval >= 113) {
return 1;
}
}
else if (posBalance >= 45) {
if (Rs < 2313) {
return 1;
}
else if (Rs >= 2313) {
if (Rs < 2339) {
return 2;
}
else if (Rs >= 2339) {
return 1;
}
}
}
}
}
else if (Cs >= 648) {
if (refDCval < 102) {
if (CsDiff < 4) {
if (Rs < 1357) {
if (Rs < 624) {
if (Rs < 514) {
return 8;
}
else if (Rs >= 514) {
return 2;
}
}
else if (Rs >= 624) {
if (MVDiff < 5) {
if (RsDiff < 3) {
if (Rs < 662) {
return 1;
}
else if (Rs >= 662) {
return 4;
}
}
else if (RsDiff >= 3) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (SC < 1346) {
return 1;
}
else if (SC >= 1346) {
return 8;
}
}
}
}
else if (Rs >= 1357) {
if (TSC < 6) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Rs < 1844) {
return 1;
}
else if (Rs >= 1844) {
return 8;
}
}
}
else if (TSC >= 6) {
if (negBalance < 52) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 1;
}
}
else if (negBalance >= 52) {
if (Rs < 1915) {
return 2;
}
else if (Rs >= 1915) {
return 1;
}
}
}
}
}
else if (CsDiff >= 4) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 20) {
if (TSC < 7) {
if (Cs < 748) {
return 1;
}
else if (Cs >= 748) {
if (ssDCval < 102) {
return 4;
}
else if (ssDCval >= 102) {
return 1;
}
}
}
else if (TSC >= 7) {
if (Rs < 2157) {
return 8;
}
else if (Rs >= 2157) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 20) {
if (tcor < 96) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (tcor >= 96) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (posBalance < 61) {
if (diffMVdiffVal < -2) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (Rs < 1114) {
return 1;
}
else if (Rs >= 1114) {
if (diffMVdiffVal < -7) {
if (Rs < 1391) {
return 1;
}
else if (Rs >= 1391) {
return 4;
}
}
else if (diffMVdiffVal >= -7) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 102) {
if (MVDiff < 5) {
if (Rs < 667) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (Rs >= 667) {
if (Rs < 2038) {
if (SC < 2628) {
if (Rs < 867) {
if (mu_mv_mag_sq < 78) {
return 4;
}
else if (mu_mv_mag_sq >= 78) {
return 8;
}
}
else if (Rs >= 867) {
if (Cs < 948) {
if (SC < 2514) {
return 8;
}
else if (SC >= 2514) {
return 4;
}
}
else if (Cs >= 948) {
return 4;
}
}
}
else if (SC >= 2628) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
if (Rs < 1488) {
return 4;
}
else if (Rs >= 1488) {
return 2;
}
}
}
}
else if (Rs >= 2038) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (refDCval < 157) {
if (AFD < 20) {
if (RsCsDiff < 16) {
if (refDCval < 104) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
else if (refDCval >= 104) {
if (posBalance < 68) {
if (negBalance < 26) {
return 4;
}
else if (negBalance >= 26) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (SC < 1894) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
return 1;
}
}
else if (SC >= 1894) {
return 8;
}
}
}
}
else if (posBalance >= 68) {
return 4;
}
}
}
else if (RsCsDiff >= 16) {
if (Cs < 853) {
if (mu_mv_mag_sq < 3) {
if (RsCsDiff < 31) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
else if (RsCsDiff >= 31) {
if (Rs < 927) {
return 2;
}
else if (Rs >= 927) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (ssDCval < 122) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
if (Cs < 723) {
return 2;
}
else if (Cs >= 723) {
return 8;
}
}
}
else if (ssDCval >= 122) {
if (tcor < 82) {
return 8;
}
else if (tcor >= 82) {
if (negBalance < 43) {
return 2;
}
else if (negBalance >= 43) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 853) {
if (SC < 3669) {
if (AFD < 14) {
if (Cs < 1296) {
return 8;
}
else if (Cs >= 1296) {
if (Rs < 1492) {
return 2;
}
else if (Rs >= 1492) {
return 8;
}
}
}
else if (AFD >= 14) {
if (negBalance < 45) {
if (ssDCval < 148) {
return 2;
}
else if (ssDCval >= 148) {
return 8;
}
}
else if (negBalance >= 45) {
if (refDCval < 114) {
return 2;
}
else if (refDCval >= 114) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
}
else if (SC >= 3669) {
return 1;
}
}
}
}
else if (AFD >= 20) {
if (Rs < 1889) {
if (Rs < 658) {
return 1;
}
else if (Rs >= 658) {
return 2;
}
}
else if (Rs >= 1889) {
if (Cs < 834) {
return 1;
}
else if (Cs >= 834) {
return 8;
}
}
}
}
else if (refDCval >= 157) {
if (tcor < 93) {
if (Cs < 747) {
return 1;
}
else if (Cs >= 747) {
return 2;
}
}
else if (tcor >= 93) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (ssDCval < 133) {
if (Cs < 1227) {
if (Rs < 342) {
if (ssDCval < 123) {
if (mu_mv_mag_sq < 118) {
return 4;
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 342) {
if (Cs < 967) {
if (mu_mv_mag_sq < 124) {
if (posBalance < 48) {
if (refDCval < 97) {
if (Rs < 612) {
return 8;
}
else if (Rs >= 612) {
if (posBalance < 47) {
if (Cs < 880) {
if (Rs < 2049) {
return 4;
}
else if (Rs >= 2049) {
if (Rs < 2143) {
return 8;
}
else if (Rs >= 2143) {
return 4;
}
}
}
else if (Cs >= 880) {
return 8;
}
}
else if (posBalance >= 47) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
else if (refDCval >= 97) {
if (mu_mv_mag_sq < 107) {
if (tcor < 99) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 494) {
return 1;
}
else if (Rs >= 494) {
return 8;
}
}
}
else if (tcor >= 99) {
if (Cs < 630) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 476) {
return 8;
}
else if (Rs >= 476) {
return 2;
}
}
}
else if (Cs >= 630) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Cs < 714) {
if (refDCval < 104) {
if (Cs < 636) {
return 8;
}
else if (Cs >= 636) {
return 1;
}
}
else if (refDCval >= 104) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (Cs >= 714) {
if (Rs < 371) {
return 4;
}
else if (Rs >= 371) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 48) {
if (negBalance < 21) {
if (Rs < 549) {
return 1;
}
else if (Rs >= 549) {
if (Rs < 1051) {
return 8;
}
else if (Rs >= 1051) {
return 4;
}
}
}
else if (negBalance >= 21) {
if (MVDiff < 3) {
if (negBalance < 40) {
if (mu_mv_mag_sq < 103) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (Rs < 486) {
return 4;
}
else if (Rs >= 486) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (Cs < 562) {
if (Rs < 1250) {
return 8;
}
else if (Rs >= 1250) {
return 1;
}
}
else if (Cs >= 562) {
if (Rs < 367) {
return 4;
}
else if (Rs >= 367) {
if (negBalance < 39) {
if (mu_mv_mag_sq < 123) {
return 8;
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1463) {
return 8;
}
else if (Rs >= 1463) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (Rs < 503) {
return 1;
}
else if (Rs >= 503) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 40) {
if (ssDCval < 102) {
if (Rs < 831) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
return 1;
}
}
else if (Rs >= 831) {
return 8;
}
}
else if (ssDCval >= 102) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (SC < 1200) {
return 2;
}
else if (SC >= 1200) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (diffTSC < -1) {
if (Rs < 1667) {
if (Rs < 482) {
return 1;
}
else if (Rs >= 482) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
return 1;
}
}
}
else if (Rs >= 1667) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
return 1;
}
}
}
else if (diffTSC >= -1) {
if (RsDiff < 2) {
if (Rs < 2183) {
if (posBalance < 25) {
if (Rs < 796) {
return 8;
}
else if (Rs >= 796) {
return 4;
}
}
else if (posBalance >= 25) {
if (negBalance < 24) {
return 8;
}
else if (negBalance >= 24) {
if (ssDCval < 97) {
if (refDCval < 95) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 2104) {
return 8;
}
else if (Rs >= 2104) {
return 4;
}
}
}
else if (refDCval >= 95) {
if (Rs < 1671) {
return 4;
}
else if (Rs >= 1671) {
return 8;
}
}
}
else if (ssDCval >= 97) {
return 8;
}
}
}
}
else if (Rs >= 2183) {
return 4;
}
}
else if (RsDiff >= 2) {
if (Rs < 1929) {
return 1;
}
else if (Rs >= 1929) {
return 8;
}
}
}
}
}
else if (Cs >= 967) {
if (refDCval < 121) {
if (mu_mv_mag_sq < 104) {
if (tcor < 98) {
if (ssDCval < 105) {
if (Cs < 1052) {
return 4;
}
else if (Cs >= 1052) {
return 8;
}
}
else if (ssDCval >= 105) {
if (Rs < 1088) {
return 1;
}
else if (Rs >= 1088) {
return 8;
}
}
}
else if (tcor >= 98) {
return 8;
}
}
else if (mu_mv_mag_sq >= 104) {
if (Cs < 1220) {
if (diffRsCsdiff < -50) {
if (Rs < 1256) {
return 8;
}
else if (Rs >= 1256) {
return 2;
}
}
else if (diffRsCsdiff >= -50) {
return 8;
}
}
else if (Cs >= 1220) {
if (ssDCval < 113) {
if (Rs < 1176) {
return 1;
}
else if (Rs >= 1176) {
return 2;
}
}
else if (ssDCval >= 113) {
return 8;
}
}
}
}
else if (refDCval >= 121) {
if (Rs < 1489) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (Rs >= 1489) {
return 2;
}
}
}
}
}
else if (Cs >= 1227) {
if (Rs < 667) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (Rs >= 667) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1147) {
return 2;
}
else if (Rs >= 1147) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2388) {
if (MVDiff < 14) {
if (refDCval < 109) {
if (posBalance < 47) {
if (Rs < 3781) {
if (RsCsDiff < 7) {
if (Cs < 1034) {
if (posBalance < 16) {
return 4;
}
else if (posBalance >= 16) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (Cs < 1032) {
return 8;
}
else if (Cs >= 1032) {
return 1;
}
}
}
}
else if (Cs >= 1034) {
if (SC < 4985) {
if (ssDCval < 92) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (ssDCval >= 92) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
return 8;
}
}
}
else if (SC >= 4985) {
return 4;
}
}
}
else if (RsCsDiff >= 7) {
if (ssDCval < 99) {
if (negBalance < 51) {
if (Cs < 408) {
if (tcor < 94) {
if (Rs < 2910) {
return 1;
}
else if (Rs >= 2910) {
return 4;
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (Cs >= 408) {
if (refDCval < 65) {
if (MVDiff < 5) {
if (Cs < 1070) {
if (mcTcor < 95) {
if (Rs < 2907) {
return 4;
}
else if (Rs >= 2907) {
return 8;
}
}
else if (mcTcor >= 95) {
if (Cs < 738) {
return 8;
}
else if (Cs >= 738) {
return 2;
}
}
}
else if (Cs >= 1070) {
if (ssDCval < 57) {
if (Rs < 3660) {
return 4;
}
else if (Rs >= 3660) {
return 8;
}
}
else if (ssDCval >= 57) {
if (Rs < 3669) {
return 8;
}
else if (Rs >= 3669) {
if (Rs < 3738) {
return 2;
}
else if (Rs >= 3738) {
return 4;
}
}
}
}
}
else if (MVDiff >= 5) {
if (CsDiff < 6) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (ssDCval < 61) {
if (Rs < 2514) {
return 2;
}
else if (Rs >= 2514) {
if (Rs < 2570) {
return 8;
}
else if (Rs >= 2570) {
return 4;
}
}
}
else if (ssDCval >= 61) {
if (mu_mv_mag_sq < 36) {
if (diffMVdiffVal < 2) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 36) {
return 4;
}
}
}
}
else if (CsDiff >= 6) {
if (Cs < 972) {
return 4;
}
else if (Cs >= 972) {
return 2;
}
}
}
}
else if (refDCval >= 65) {
if (refDCval < 86) {
if (mu_mv_mag_sq < 88) {
if (ssDCval < 68) {
if (AFD < 13) {
if (negBalance < 42) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (negBalance >= 42) {
if (Cs < 950) {
if (Rs < 2810) {
return 1;
}
else if (Rs >= 2810) {
return 4;
}
}
else if (Cs >= 950) {
return 2;
}
}
}
else if (AFD >= 13) {
return 2;
}
}
else if (ssDCval >= 68) {
if (Rs < 3077) {
if (TSC < 6) {
if (ssDCval < 74) {
if (AFD < 8) {
if (Rs < 2644) {
return 2;
}
else if (Rs >= 2644) {
return 1;
}
}
else if (AFD >= 8) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
else if (ssDCval >= 74) {
return 1;
}
}
else if (TSC >= 6) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
if (mu_mv_mag_sq < 79) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 79) {
return 2;
}
}
}
}
else if (Rs >= 3077) {
if (tcor < 91) {
if (SC < 4603) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 1;
}
}
}
else if (SC >= 4603) {
return 1;
}
}
else if (tcor >= 91) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 88) {
if (ssDCval < 69) {
if (Rs < 3051) {
if (Cs < 840) {
return 1;
}
else if (Cs >= 840) {
return 2;
}
}
else if (Rs >= 3051) {
if (Rs < 3385) {
return 4;
}
else if (Rs >= 3385) {
return 1;
}
}
}
else if (ssDCval >= 69) {
if (diffMVdiffVal < -2) {
if (Rs < 2975) {
return 1;
}
else if (Rs >= 2975) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
}
}
else if (refDCval >= 86) {
if (tcor < 96) {
if (posBalance < 43) {
if (Rs < 2588) {
return 1;
}
else if (Rs >= 2588) {
if (Rs < 3569) {
if (diffTSC < -1) {
if (Rs < 3104) {
return 8;
}
else if (Rs >= 3104) {
return 4;
}
}
else if (diffTSC >= -1) {
if (gchDC < 2) {
if (Rs < 3460) {
return 4;
}
else if (Rs >= 3460) {
return 4;
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (Rs >= 3569) {
if (Rs < 3689) {
return 8;
}
else if (Rs >= 3689) {
return 8;
}
}
}
}
else if (posBalance >= 43) {
if (negBalance < 31) {
if (Rs < 3558) {
return 4;
}
else if (Rs >= 3558) {
return 8;
}
}
else if (negBalance >= 31) {
return 2;
}
}
}
else if (tcor >= 96) {
return 2;
}
}
}
}
}
else if (negBalance >= 51) {
if (RsDiff < 5) {
if (Rs < 2614) {
return 4;
}
else if (Rs >= 2614) {
return 1;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (ssDCval >= 99) {
if (mu_mv_mag_sq < 104) {
if (tcor < 94) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 901) {
return 2;
}
else if (Cs >= 901) {
return 8;
}
}
}
else if (tcor >= 94) {
if (Cs < 921) {
if (SC < 4085) {
if (Rs < 3365) {
return 8;
}
else if (Rs >= 3365) {
return 2;
}
}
else if (SC >= 4085) {
if (RsDiff < 6) {
if (Cs < 833) {
return 4;
}
else if (Cs >= 833) {
return 4;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (Cs >= 921) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (ssDCval < 103) {
return 4;
}
else if (ssDCval >= 103) {
if (Rs < 3162) {
return 8;
}
else if (Rs >= 3162) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3781) {
if (Cs < 1280) {
if (negBalance < 46) {
if (mcTcor < 96) {
if (ssDCval < 80) {
if (Rs < 3810) {
if (Rs < 3794) {
return 4;
}
else if (Rs >= 3794) {
return 2;
}
}
else if (Rs >= 3810) {
return 4;
}
}
else if (ssDCval >= 80) {
if (Rs < 4313) {
if (RsDiff < 8) {
if (mu_mv_mag_sq < 62) {
if (diffMVdiffVal < 2) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (Rs < 4242) {
if (Cs < 1158) {
return 4;
}
else if (Cs >= 1158) {
return 4;
}
}
else if (Rs >= 4242) {
return 4;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 3877) {
return 4;
}
else if (Rs >= 3877) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 62) {
return 8;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (Rs >= 4313) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (SC < 5464) {
if (ssDCval < 69) {
if (Cs < 1067) {
return 4;
}
else if (Cs >= 1067) {
return 1;
}
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (SC >= 5464) {
if (AFD < 5) {
if (Rs < 4450) {
return 1;
}
else if (Rs >= 4450) {
return 2;
}
}
else if (AFD >= 5) {
return 4;
}
}
}
}
else if (negBalance >= 46) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 2;
}
}
}
else if (Cs >= 1280) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 42) {
if (Rs < 4133) {
return 4;
}
else if (Rs >= 4133) {
return 1;
}
}
else if (posBalance >= 42) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (negBalance < 16) {
if (Rs < 4275) {
return 4;
}
else if (Rs >= 4275) {
return 2;
}
}
else if (negBalance >= 16) {
return 8;
}
}
}
}
}
else if (posBalance >= 47) {
if (ssDCval < 90) {
if (RsCsDiff < 43) {
if (Cs < 682) {
if (tcor < 94) {
if (Rs < 2639) {
return 1;
}
else if (Rs >= 2639) {
if (Cs < 348) {
return 8;
}
else if (Cs >= 348) {
if (Rs < 3794) {
return 8;
}
else if (Rs >= 3794) {
return 4;
}
}
}
}
else if (tcor >= 94) {
if (mu_mv_mag_sq < 69) {
return 1;
}
else if (mu_mv_mag_sq >= 69) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
if (Rs < 2492) {
return 2;
}
else if (Rs >= 2492) {
return 4;
}
}
}
}
}
else if (Cs >= 682) {
if (negBalance < 22) {
if (Cs < 1121) {
if (negBalance < 15) {
if (Rs < 2902) {
return 4;
}
else if (Rs >= 2902) {
return 4;
}
}
else if (negBalance >= 15) {
return 4;
}
}
else if (Cs >= 1121) {
return 8;
}
}
else if (negBalance >= 22) {
if (ssDCval < 59) {
if (Rs < 2444) {
return 8;
}
else if (Rs >= 2444) {
return 4;
}
}
else if (ssDCval >= 59) {
if (ssDCval < 74) {
if (ssDCval < 65) {
if (Cs < 977) {
return 2;
}
else if (Cs >= 977) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (ssDCval >= 65) {
return 2;
}
}
else if (ssDCval >= 74) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (Rs < 3312) {
return 1;
}
else if (Rs >= 3312) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 43) {
if (refDCval < 77) {
if (Rs < 3728) {
if (refDCval < 60) {
if (Rs < 2507) {
return 2;
}
else if (Rs >= 2507) {
return 4;
}
}
else if (refDCval >= 60) {
if (posBalance < 53) {
if (negBalance < 40) {
if (mu_mv_mag_sq < 77) {
if (Rs < 2524) {
return 2;
}
else if (Rs >= 2524) {
if (AFD < 13) {
if (Rs < 2989) {
if (SC < 3893) {
if (Cs < 1014) {
if (Cs < 906) {
return 1;
}
else if (Cs >= 906) {
return 2;
}
}
else if (Cs >= 1014) {
return 4;
}
}
else if (SC >= 3893) {
return 2;
}
}
else if (Rs >= 2989) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 1;
}
}
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 77) {
return 2;
}
}
else if (negBalance >= 40) {
if (AFD < 19) {
if (mcTcor < 93) {
if (Rs < 2501) {
return 4;
}
else if (Rs >= 2501) {
if (Rs < 2745) {
return 8;
}
else if (Rs >= 2745) {
return 2;
}
}
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (diffRsCsdiff < -162) {
return 4;
}
else if (diffRsCsdiff >= -162) {
if (refDCval < 62) {
if (Cs < 934) {
return 2;
}
else if (Cs >= 934) {
if (negBalance < 39) {
return 2;
}
else if (negBalance >= 39) {
return 4;
}
}
}
else if (refDCval >= 62) {
if (SC < 3710) {
return 2;
}
else if (SC >= 3710) {
if (Cs < 568) {
return 4;
}
else if (Cs >= 568) {
if (Rs < 2702) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 2;
}
}
else if (Rs >= 2702) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 3728) {
if (Cs < 957) {
return 4;
}
else if (Cs >= 957) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
}
}
else if (refDCval >= 77) {
if (TSC < 6) {
if (MVDiff < 6) {
if (tcor < 94) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 2;
}
}
}
else if (tcor >= 94) {
return 2;
}
}
else if (MVDiff >= 6) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 6) {
return 4;
}
else if (mu_mv_mag_sq >= 6) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (posBalance >= 54) {
if (RsCsDiff < 69) {
return 1;
}
else if (RsCsDiff >= 69) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (tcor < 71) {
if (Rs < 3237) {
return 2;
}
else if (Rs >= 3237) {
return 1;
}
}
else if (tcor >= 71) {
if (RsDiff < 9) {
if (AFD < 12) {
if (SC < 4912) {
return 2;
}
else if (SC >= 4912) {
if (Rs < 3706) {
return 4;
}
else if (Rs >= 3706) {
return 2;
}
}
}
else if (AFD >= 12) {
if (Rs < 2454) {
return 4;
}
else if (Rs >= 2454) {
return 1;
}
}
}
else if (RsDiff >= 9) {
if (mcTcor < 86) {
return 1;
}
else if (mcTcor >= 86) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Cs < 344) {
return 1;
}
else if (Cs >= 344) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 90) {
if (RsDiff < 11) {
if (posBalance < 63) {
if (Cs < 811) {
if (Rs < 2513) {
return 8;
}
else if (Rs >= 2513) {
if (posBalance < 52) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
else if (posBalance >= 52) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
else if (Cs >= 811) {
if (diffRsCsdiff < -45) {
if (Cs < 1202) {
return 4;
}
else if (Cs >= 1202) {
if (Rs < 3296) {
return 8;
}
else if (Rs >= 3296) {
return 2;
}
}
}
else if (diffRsCsdiff >= -45) {
if (Rs < 3601) {
if (ssDCval < 110) {
if (negBalance < 30) {
return 2;
}
else if (negBalance >= 30) {
return 8;
}
}
else if (ssDCval >= 110) {
return 2;
}
}
else if (Rs >= 3601) {
if (posBalance < 54) {
if (MVDiff < 5) {
if (Rs < 4085) {
return 4;
}
else if (Rs >= 4085) {
if (Rs < 4286) {
return 8;
}
else if (Rs >= 4286) {
return 4;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (posBalance >= 54) {
return 2;
}
}
}
}
}
else if (posBalance >= 63) {
if (tcor < 95) {
if (negBalance < 22) {
if (gchDC < 4) {
if (Cs < 496) {
return 4;
}
else if (Cs >= 496) {
if (diffRsCsdiff < -5) {
return 8;
}
else if (diffRsCsdiff >= -5) {
if (ssDCval < 101) {
if (Rs < 3627) {
return 4;
}
else if (Rs >= 3627) {
return 8;
}
}
else if (ssDCval >= 101) {
return 4;
}
}
}
}
else if (gchDC >= 4) {
return 8;
}
}
else if (negBalance >= 22) {
return 2;
}
}
else if (tcor >= 95) {
return 2;
}
}
}
else if (RsDiff >= 11) {
if (negBalance < 49) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
return 2;
}
}
else if (negBalance >= 49) {
if (Cs < 867) {
return 2;
}
else if (Cs >= 867) {
if (ssDCval < 106) {
if (tcor < 80) {
if (Rs < 3214) {
if (Cs < 1246) {
return 4;
}
else if (Cs >= 1246) {
return 2;
}
}
else if (Rs >= 3214) {
return 8;
}
}
else if (tcor >= 80) {
if (RsDiff < 12) {
return 4;
}
else if (RsDiff >= 12) {
return 2;
}
}
}
else if (ssDCval >= 106) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 2830) {
if (mcTcor < 97) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (diffAFD < -1) {
if (ssDCval < 118) {
if (Cs < 893) {
return 2;
}
else if (Cs >= 893) {
if (Rs < 2493) {
return 1;
}
else if (Rs >= 2493) {
return 4;
}
}
}
else if (ssDCval >= 118) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
return 1;
}
}
}
else if (diffAFD >= -1) {
if (diffAFD < 2) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (Rs < 2478) {
return 8;
}
else if (Rs >= 2478) {
return 1;
}
}
}
else if (diffAFD >= 2) {
if (negBalance < 37) {
if (Rs < 2621) {
return 8;
}
else if (Rs >= 2621) {
return 2;
}
}
else if (negBalance >= 37) {
return 1;
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 116) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 2;
}
}
else if (ssDCval >= 116) {
if (ssDCval < 128) {
if (Rs < 2717) {
if (refDCval < 128) {
return 1;
}
else if (refDCval >= 128) {
return 2;
}
}
else if (Rs >= 2717) {
if (Cs < 789) {
return 2;
}
else if (Cs >= 789) {
return 4;
}
}
}
else if (ssDCval >= 128) {
if (mu_mv_mag_sq < 91) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (Cs < 616) {
return 1;
}
else if (Cs >= 616) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (Rs < 2481) {
return 8;
}
else if (Rs >= 2481) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2830) {
if (MVDiff < 7) {
if (refDCval < 138) {
if (Cs < 701) {
if (RsDiff < 7) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 2;
}
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (Cs >= 701) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (refDCval < 134) {
if (Cs < 1275) {
return 2;
}
else if (Cs >= 1275) {
if (Rs < 5037) {
return 8;
}
else if (Rs >= 5037) {
return 2;
}
}
}
else if (refDCval >= 134) {
if (Cs < 1024) {
return 2;
}
else if (Cs >= 1024) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 138) {
if (CsDiff < 4) {
if (TSC < 10) {
if (Rs < 3963) {
if (mu_mv_mag_sq < 106) {
return 2;
}
else if (mu_mv_mag_sq >= 106) {
if (MVDiff < 1) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
else if (MVDiff >= 1) {
if (Cs < 1128) {
return 2;
}
else if (Cs >= 1128) {
return 1;
}
}
}
}
else if (Rs >= 3963) {
return 2;
}
}
else if (TSC >= 10) {
return 1;
}
}
else if (CsDiff >= 4) {
if (SC < 5647) {
if (Cs < 1083) {
return 2;
}
else if (Cs >= 1083) {
if (refDCval < 147) {
if (RsDiff < 6) {
if (Rs < 3918) {
return 1;
}
else if (Rs >= 3918) {
return 2;
}
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (refDCval >= 147) {
return 2;
}
}
}
else if (SC >= 5647) {
return 2;
}
}
}
}
else if (MVDiff >= 7) {
if (TSC < 6) {
if (Rs < 3937) {
if (tcor < 98) {
if (diffMVdiffVal < 5) {
if (refDCval < 133) {
if (ssDCval < 115) {
if (posBalance < 40) {
if (Rs < 2891) {
return 8;
}
else if (Rs >= 2891) {
return 1;
}
}
else if (posBalance >= 40) {
return 8;
}
}
else if (ssDCval >= 115) {
return 8;
}
}
else if (refDCval >= 133) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (SC < 4578) {
return 2;
}
else if (SC >= 4578) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
else if (tcor >= 98) {
if (ssDCval < 132) {
return 2;
}
else if (ssDCval >= 132) {
return 1;
}
}
}
else if (Rs >= 3937) {
if (Cs < 1184) {
return 2;
}
else if (Cs >= 1184) {
return 1;
}
}
}
else if (TSC >= 6) {
if (refDCval < 137) {
if (Rs < 3440) {
if (Cs < 768) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (Rs < 2936) {
return 1;
}
else if (Rs >= 2936) {
if (Rs < 3237) {
return 2;
}
else if (Rs >= 3237) {
return 1;
}
}
}
}
else if (Cs >= 768) {
if (AFD < 13) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 3392) {
return 2;
}
else if (Rs >= 3392) {
return 8;
}
}
}
else if (AFD >= 13) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 2;
}
}
}
}
else if (Rs >= 3440) {
if (negBalance < 52) {
if (ssDCval < 136) {
return 2;
}
else if (ssDCval >= 136) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
else if (negBalance >= 52) {
if (RsDiff < 21) {
if (Cs < 784) {
return 2;
}
else if (Cs >= 784) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
if (Rs < 4110) {
return 2;
}
else if (Rs >= 4110) {
return 1;
}
}
}
}
else if (RsDiff >= 21) {
return 2;
}
}
}
}
else if (refDCval >= 137) {
if (mcTcor < 95) {
if (RsDiff < 21) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
return 4;
}
}
else if (RsDiff >= 21) {
return 2;
}
}
else if (mcTcor >= 95) {
if (mu_mv_mag_sq < 38) {
if (RsDiff < 15) {
return 1;
}
else if (RsDiff >= 15) {
return 2;
}
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 3869) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 2;
}
}
else if (Rs >= 3869) {
return 2;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 14) {
if (mcTcor < 68) {
if (MVDiff < 42) {
return 1;
}
else if (MVDiff >= 42) {
if (Rs < 3207) {
if (Rs < 2940) {
return 8;
}
else if (Rs >= 2940) {
return 1;
}
}
else if (Rs >= 3207) {
if (AFD < 33) {
return 4;
}
else if (AFD >= 33) {
return 2;
}
}
}
}
else if (mcTcor >= 68) {
if (RsCsDiff < 248) {
if (ssDCval < 82) {
if (TSC < 8) {
if (Rs < 2900) {
if (RsCsDiff < 53) {
if (SC < 3178) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (SC >= 3178) {
if (tcor < 91) {
if (Rs < 2889) {
return 8;
}
else if (Rs >= 2889) {
return 2;
}
}
else if (tcor >= 91) {
return 4;
}
}
}
else if (RsCsDiff >= 53) {
if (negBalance < 47) {
if (posBalance < 51) {
if (negBalance < 38) {
if (AFD < 15) {
if (diffAFD < -1) {
if (Cs < 963) {
return 4;
}
else if (Cs >= 963) {
return 2;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (AFD >= 15) {
return 2;
}
}
else if (negBalance >= 38) {
if (diffRsCsdiff < -11) {
return 4;
}
else if (diffRsCsdiff >= -11) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
if (Rs < 2490) {
return 1;
}
else if (Rs >= 2490) {
return 4;
}
}
}
}
}
else if (posBalance >= 51) {
if (SC < 2979) {
return 4;
}
else if (SC >= 2979) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
if (Cs < 995) {
return 1;
}
else if (Cs >= 995) {
return 8;
}
}
}
}
}
else if (Rs >= 2900) {
if (tcor < 97) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (Rs < 8180) {
if (refDCval < 80) {
if (mu_mv_mag_sq < 70) {
return 4;
}
else if (mu_mv_mag_sq >= 70) {
if (Rs < 3277) {
return 2;
}
else if (Rs >= 3277) {
return 4;
}
}
}
else if (refDCval >= 80) {
return 2;
}
}
else if (Rs >= 8180) {
return 2;
}
}
}
else if (tcor >= 97) {
if (Cs < 209) {
return 1;
}
else if (Cs >= 209) {
return 8;
}
}
}
}
else if (TSC >= 8) {
if (Rs < 3700) {
if (AFD < 12) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
if (Rs < 2769) {
return 1;
}
else if (Rs >= 2769) {
return 2;
}
}
}
else if (AFD >= 12) {
if (mcTcor < 89) {
if (RsCsDiff < 131) {
if (Rs < 2560) {
return 8;
}
else if (Rs >= 2560) {
return 4;
}
}
else if (RsCsDiff >= 131) {
if (negBalance < 42) {
if (Rs < 3057) {
return 2;
}
else if (Rs >= 3057) {
return 4;
}
}
else if (negBalance >= 42) {
if (ssDCval < 74) {
if (MVDiff < 42) {
return 1;
}
else if (MVDiff >= 42) {
if (Rs < 2505) {
return 2;
}
else if (Rs >= 2505) {
return 4;
}
}
}
else if (ssDCval >= 74) {
return 2;
}
}
}
}
else if (mcTcor >= 89) {
if (posBalance < 47) {
if (mu_mv_mag_sq < 40) {
if (diffAFD < 4) {
if (Cs < 495) {
return 4;
}
else if (Cs >= 495) {
return 2;
}
}
else if (diffAFD >= 4) {
return 1;
}
}
else if (mu_mv_mag_sq >= 40) {
if (posBalance < 39) {
if (Rs < 3225) {
return 2;
}
else if (Rs >= 3225) {
return 1;
}
}
else if (posBalance >= 39) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
return 2;
}
}
}
}
else if (posBalance >= 47) {
if (Rs < 2389) {
return 4;
}
else if (Rs >= 2389) {
return 2;
}
}
}
}
}
else if (Rs >= 3700) {
return 4;
}
}
}
else if (ssDCval >= 82) {
if (mcTcor < 93) {
if (refDCval < 133) {
if (AFD < 24) {
if (MVDiff < 34) {
if (RsCsDiff < 121) {
if (AFD < 13) {
if (Rs < 2898) {
return 1;
}
else if (Rs >= 2898) {
return 2;
}
}
else if (AFD >= 13) {
return 8;
}
}
else if (RsCsDiff >= 121) {
if (TSC < 13) {
if (Rs < 2707) {
if (MVDiff < 24) {
if (negBalance < 58) {
if (Cs < 721) {
return 2;
}
else if (Cs >= 721) {
return 4;
}
}
else if (negBalance >= 58) {
if (Rs < 2469) {
return 8;
}
else if (Rs >= 2469) {
return 2;
}
}
}
else if (MVDiff >= 24) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (Rs >= 2707) {
if (negBalance < 57) {
return 2;
}
else if (negBalance >= 57) {
return 8;
}
}
}
else if (TSC >= 13) {
return 1;
}
}
}
else if (MVDiff >= 34) {
if (SC < 3055) {
if (ssDCval < 88) {
if (Rs < 2529) {
return 2;
}
else if (Rs >= 2529) {
return 4;
}
}
else if (ssDCval >= 88) {
if (RsCsDiff < 103) {
return 1;
}
else if (RsCsDiff >= 103) {
if (diffMVdiffVal < -5) {
return 8;
}
else if (diffMVdiffVal >= -5) {
if (Rs < 2620) {
return 1;
}
else if (Rs >= 2620) {
return 8;
}
}
}
}
}
else if (SC >= 3055) {
if (Cs < 292) {
if (Rs < 3598) {
return 1;
}
else if (Rs >= 3598) {
return 2;
}
}
else if (Cs >= 292) {
if (mu_mv_mag_sq < 64) {
if (diffAFD < -4) {
if (MVDiff < 62) {
if (Rs < 3195) {
if (Rs < 2921) {
return 8;
}
else if (Rs >= 2921) {
return 4;
}
}
else if (Rs >= 3195) {
return 1;
}
}
else if (MVDiff >= 62) {
return 8;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 64) {
if (Rs < 3086) {
return 1;
}
else if (Rs >= 3086) {
return 4;
}
}
}
}
}
}
else if (AFD >= 24) {
if (mcTcor < 84) {
if (Rs < 2787) {
if (refDCval < 93) {
return 1;
}
else if (refDCval >= 93) {
return 2;
}
}
else if (Rs >= 2787) {
return 4;
}
}
else if (mcTcor >= 84) {
if (RsCsDiff < 227) {
if (MVDiff < 65) {
return 1;
}
else if (MVDiff >= 65) {
if (Rs < 2630) {
return 1;
}
else if (Rs >= 2630) {
return 8;
}
}
}
else if (RsCsDiff >= 227) {
return 2;
}
}
}
}
else if (refDCval >= 133) {
if (ssDCval < 147) {
if (AFD < 19) {
if (MVDiff < 54) {
return 4;
}
else if (MVDiff >= 54) {
if (Rs < 2816) {
return 1;
}
else if (Rs >= 2816) {
return 8;
}
}
}
else if (AFD >= 19) {
if (Rs < 2750) {
return 1;
}
else if (Rs >= 2750) {
return 8;
}
}
}
else if (ssDCval >= 147) {
if (RsCsDiff < 159) {
return 8;
}
else if (RsCsDiff >= 159) {
return 1;
}
}
}
}
else if (mcTcor >= 93) {
if (SC < 3519) {
if (refDCval < 114) {
if (RsDiff < 13) {
if (refDCval < 90) {
if (diffMVdiffVal < 10) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (diffMVdiffVal >= 10) {
if (Rs < 2406) {
return 1;
}
else if (Rs >= 2406) {
return 2;
}
}
}
else if (refDCval >= 90) {
if (Rs < 2641) {
if (tcor < 80) {
return 8;
}
else if (tcor >= 80) {
if (refDCval < 105) {
if (negBalance < 55) {
if (Rs < 2490) {
return 4;
}
else if (Rs >= 2490) {
return 1;
}
}
else if (negBalance >= 55) {
if (MVDiff < 36) {
return 8;
}
else if (MVDiff >= 36) {
return 1;
}
}
}
else if (refDCval >= 105) {
return 1;
}
}
}
else if (Rs >= 2641) {
if (ssDCval < 112) {
if (Rs < 2973) {
return 8;
}
else if (Rs >= 2973) {
if (Rs < 3048) {
return 4;
}
else if (Rs >= 3048) {
return 8;
}
}
}
else if (ssDCval >= 112) {
return 2;
}
}
}
}
else if (RsDiff >= 13) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (Cs < 550) {
return 8;
}
else if (Cs >= 550) {
if (Rs < 2634) {
return 1;
}
else if (Rs >= 2634) {
return 2;
}
}
}
}
}
else if (refDCval >= 114) {
if (CsDiff < 3) {
if (diffAFD < -3) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (RsDiff < 11) {
if (Cs < 124) {
return 2;
}
else if (Cs >= 124) {
return 1;
}
}
else if (RsDiff >= 11) {
return 2;
}
}
}
else if (diffAFD >= -3) {
if (Cs < 383) {
if (Cs < 13) {
return 8;
}
else if (Cs >= 13) {
if (Cs < 264) {
return 1;
}
else if (Cs >= 264) {
if (SC < 2966) {
return 8;
}
else if (SC >= 2966) {
return 1;
}
}
}
}
else if (Cs >= 383) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 2955) {
if (Cs < 1021) {
if (mu_mv_mag_sq < 32) {
return 1;
}
else if (mu_mv_mag_sq >= 32) {
if (gchDC < 5) {
return 1;
}
else if (gchDC >= 5) {
return 2;
}
}
}
else if (Cs >= 1021) {
return 8;
}
}
else if (Rs >= 2955) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
return 1;
}
}
}
}
}
else if (SC >= 3519) {
if (RsCsDiff < 71) {
if (Rs < 3866) {
if (diffMVdiffVal < 5) {
if (Cs < 1017) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (Cs >= 1017) {
return 8;
}
}
else if (diffMVdiffVal >= 5) {
if (ssDCval < 104) {
if (Rs < 2746) {
return 2;
}
else if (Rs >= 2746) {
if (Rs < 3210) {
return 8;
}
else if (Rs >= 3210) {
if (Rs < 3514) {
return 4;
}
else if (Rs >= 3514) {
return 4;
}
}
}
}
else if (ssDCval >= 104) {
if (diffRsCsdiff < -17) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
return 8;
}
}
else if (diffRsCsdiff >= -17) {
return 8;
}
}
}
}
else if (Rs >= 3866) {
return 2;
}
}
else if (RsCsDiff >= 71) {
if (refDCval < 132) {
if (Rs < 5211) {
if (diffAFD < -5) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (refDCval < 102) {
if (RsCsDiff < 206) {
if (MVDiff < 25) {
return 4;
}
else if (MVDiff >= 25) {
if (Rs < 3193) {
return 8;
}
else if (Rs >= 3193) {
return 1;
}
}
}
else if (RsCsDiff >= 206) {
return 2;
}
}
else if (refDCval >= 102) {
return 1;
}
}
}
else if (diffAFD >= -5) {
if (Cs < 1174) {
if (negBalance < 49) {
if (Cs < 595) {
if (Cs < 508) {
if (Rs < 3299) {
return 8;
}
else if (Rs >= 3299) {
return 2;
}
}
else if (Cs >= 508) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 4;
}
}
}
else if (Cs >= 595) {
if (SC < 3833) {
if (MVDiff < 24) {
if (Cs < 1016) {
return 2;
}
else if (Cs >= 1016) {
return 8;
}
}
else if (MVDiff >= 24) {
return 1;
}
}
else if (SC >= 3833) {
return 2;
}
}
}
else if (negBalance >= 49) {
if (MVDiff < 19) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 2690) {
return 8;
}
else if (Rs >= 2690) {
return 2;
}
}
}
else if (MVDiff >= 19) {
if (negBalance < 64) {
if (Cs < 516) {
if (Rs < 3030) {
return 2;
}
else if (Rs >= 3030) {
return 1;
}
}
else if (Cs >= 516) {
if (diffMVdiffVal < -30) {
return 1;
}
else if (diffMVdiffVal >= -30) {
return 8;
}
}
}
else if (negBalance >= 64) {
if (Cs < 606) {
return 4;
}
else if (Cs >= 606) {
return 2;
}
}
}
}
}
else if (Cs >= 1174) {
if (Rs < 2699) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 2508) {
return 8;
}
else if (Rs >= 2508) {
return 1;
}
}
}
else if (Rs >= 2699) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
}
}
else if (Rs >= 5211) {
return 2;
}
}
else if (refDCval >= 132) {
if (CsDiff < 4) {
if (diffRsCsdiff < -61) {
if (diffAFD < 1) {
if (Cs < 434) {
return 8;
}
else if (Cs >= 434) {
return 1;
}
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (diffRsCsdiff >= -61) {
if (mcTcor < 94) {
if (Cs < 667) {
return 8;
}
else if (Cs >= 667) {
return 4;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
}
else if (CsDiff >= 4) {
if (tcor < 79) {
if (mu_mv_mag_sq < 15) {
return 8;
}
else if (mu_mv_mag_sq >= 15) {
if (Rs < 2727) {
return 2;
}
else if (Rs >= 2727) {
return 1;
}
}
}
else if (tcor >= 79) {
if (AFD < 10) {
if (Rs < 5046) {
return 8;
}
else if (Rs >= 5046) {
return 2;
}
}
else if (AFD >= 10) {
if (RsDiff < 11) {
return 2;
}
else if (RsDiff >= 11) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 248) {
if (refDCval < 112) {
if (ssDCval < 73) {
if (Rs < 3249) {
if (posBalance < 52) {
if (SC < 3340) {
if (Rs < 2531) {
return 1;
}
else if (Rs >= 2531) {
return 2;
}
}
else if (SC >= 3340) {
if (SC < 4117) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
return 2;
}
}
else if (SC >= 4117) {
return 2;
}
}
}
else if (posBalance >= 52) {
if (CsDiff < 3) {
if (Rs < 2700) {
return 1;
}
else if (Rs >= 2700) {
return 4;
}
}
else if (CsDiff >= 3) {
if (Rs < 3167) {
return 2;
}
else if (Rs >= 3167) {
if (Cs < 1115) {
return 1;
}
else if (Cs >= 1115) {
return 2;
}
}
}
}
}
else if (Rs >= 3249) {
if (AFD < 24) {
if (RsDiff < 16) {
if (Rs < 3554) {
return 1;
}
else if (Rs >= 3554) {
return 4;
}
}
else if (RsDiff >= 16) {
return 4;
}
}
else if (AFD >= 24) {
if (Rs < 4789) {
return 1;
}
else if (Rs >= 4789) {
return 8;
}
}
}
}
else if (ssDCval >= 73) {
if (TSC < 12) {
if (mcTcor < 90) {
if (Cs < 988) {
if (AFD < 18) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
return 4;
}
}
else if (AFD >= 18) {
if (Rs < 2407) {
return 4;
}
else if (Rs >= 2407) {
return 2;
}
}
}
else if (Cs >= 988) {
if (RsCsDiff < 301) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (RsCsDiff >= 301) {
if (Rs < 2807) {
return 2;
}
else if (Rs >= 2807) {
return 8;
}
}
}
}
else if (mcTcor >= 90) {
if (refDCval < 78) {
if (Rs < 3286) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
return 2;
}
}
else if (Rs >= 3286) {
if (Cs < 873) {
return 4;
}
else if (Cs >= 873) {
return 2;
}
}
}
else if (refDCval >= 78) {
if (MVDiff < 24) {
if (mcTcor < 92) {
if (refDCval < 91) {
return 2;
}
else if (refDCval >= 91) {
if (Cs < 652) {
return 2;
}
else if (Cs >= 652) {
return 4;
}
}
}
else if (mcTcor >= 92) {
return 2;
}
}
else if (MVDiff >= 24) {
if (TSC < 9) {
if (gchDC < 8) {
if (Cs < 792) {
return 2;
}
else if (Cs >= 792) {
if (Cs < 807) {
return 8;
}
else if (Cs >= 807) {
return 2;
}
}
}
else if (gchDC >= 8) {
if (Rs < 3151) {
if (Rs < 2707) {
return 1;
}
else if (Rs >= 2707) {
return 8;
}
}
else if (Rs >= 3151) {
return 2;
}
}
}
else if (TSC >= 9) {
if (CsDiff < 3) {
if (Cs < 267) {
return 2;
}
else if (Cs >= 267) {
return 4;
}
}
else if (CsDiff >= 3) {
if (mcTcor < 95) {
if (diffMVdiffVal < -10) {
if (Rs < 3627) {
if (Rs < 2942) {
return 8;
}
else if (Rs >= 2942) {
return 1;
}
}
else if (Rs >= 3627) {
return 2;
}
}
else if (diffMVdiffVal >= -10) {
return 2;
}
}
else if (mcTcor >= 95) {
if (refDCval < 98) {
if (Rs < 3682) {
if (Rs < 2652) {
return 1;
}
else if (Rs >= 2652) {
return 2;
}
}
else if (Rs >= 3682) {
return 1;
}
}
else if (refDCval >= 98) {
return 2;
}
}
}
}
}
}
}
}
else if (TSC >= 12) {
if (RsCsDiff < 454) {
if (mcTcor < 90) {
if (posBalance < 64) {
if (gchDC < 4) {
if (diffMVdiffVal < -38) {
if (Rs < 2985) {
return 2;
}
else if (Rs >= 2985) {
return 4;
}
}
else if (diffMVdiffVal >= -38) {
if (AFD < 20) {
if (Rs < 2931) {
return 4;
}
else if (Rs >= 2931) {
return 1;
}
}
else if (AFD >= 20) {
return 4;
}
}
}
else if (gchDC >= 4) {
if (AFD < 35) {
return 2;
}
else if (AFD >= 35) {
return 4;
}
}
}
else if (posBalance >= 64) {
if (Rs < 3697) {
return 1;
}
else if (Rs >= 3697) {
return 4;
}
}
}
else if (mcTcor >= 90) {
if (MVDiff < 35) {
return 2;
}
else if (MVDiff >= 35) {
if (SC < 3905) {
return 1;
}
else if (SC >= 3905) {
return 8;
}
}
}
}
else if (RsCsDiff >= 454) {
if (ssDCval < 92) {
if (mu_mv_mag_sq < 10) {
return 1;
}
else if (mu_mv_mag_sq >= 10) {
if (Rs < 2804) {
return 8;
}
else if (Rs >= 2804) {
return 2;
}
}
}
else if (ssDCval >= 92) {
if (TSC < 18) {
return 2;
}
else if (TSC >= 18) {
if (Rs < 3365) {
return 1;
}
else if (Rs >= 3365) {
if (Rs < 4818) {
return 4;
}
else if (Rs >= 4818) {
return 2;
}
}
}
}
}
}
}
}
else if (refDCval >= 112) {
if (RsCsDiff < 431) {
if (mcTcor < 92) {
if (AFD < 18) {
if (ssDCval < 134) {
if (SC < 3729) {
return 1;
}
else if (SC >= 3729) {
return 8;
}
}
else if (ssDCval >= 134) {
return 4;
}
}
else if (AFD >= 18) {
if (Cs < 860) {
if (tcor < 77) {
if (Rs < 3310) {
if (AFD < 36) {
return 2;
}
else if (AFD >= 36) {
return 1;
}
}
else if (Rs >= 3310) {
return 4;
}
}
else if (tcor >= 77) {
return 1;
}
}
else if (Cs >= 860) {
if (AFD < 24) {
if (MVDiff < 26) {
return 1;
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (AFD >= 24) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Rs < 3021) {
return 1;
}
else if (Rs >= 3021) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 3186) {
if (tcor < 75) {
if (diffAFD < 1) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
return 2;
}
}
else if (diffAFD >= 1) {
if (Cs < 1178) {
return 1;
}
else if (Cs >= 1178) {
return 2;
}
}
}
else if (tcor >= 75) {
if (Cs < 1218) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (Cs >= 1218) {
if (Rs < 2647) {
return 2;
}
else if (Rs >= 2647) {
return 8;
}
}
}
}
else if (Rs >= 3186) {
if (CsDiff < 5) {
if (MVDiff < 32) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (TSC < 12) {
if (RsDiff < 17) {
if (Rs < 3563) {
return 1;
}
else if (Rs >= 3563) {
if (Rs < 5226) {
return 2;
}
else if (Rs >= 5226) {
return 1;
}
}
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (TSC >= 12) {
return 1;
}
}
}
else if (MVDiff >= 32) {
if (AFD < 14) {
if (MVDiff < 38) {
return 8;
}
else if (MVDiff >= 38) {
return 2;
}
}
else if (AFD >= 14) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (Rs < 4482) {
return 1;
}
else if (Rs >= 4482) {
return 8;
}
}
}
}
}
else if (CsDiff >= 5) {
if (diffRsCsdiff < -202) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
return 1;
}
}
}
else if (diffRsCsdiff >= -202) {
if (tcor < 85) {
if (Rs < 4307) {
if (MVDiff < 36) {
return 8;
}
else if (MVDiff >= 36) {
return 1;
}
}
else if (Rs >= 4307) {
if (Cs < 1049) {
return 2;
}
else if (Cs >= 1049) {
return 8;
}
}
}
else if (tcor >= 85) {
if (Rs < 3259) {
return 8;
}
else if (Rs >= 3259) {
if (posBalance < 45) {
if (Rs < 4865) {
return 1;
}
else if (Rs >= 4865) {
return 8;
}
}
else if (posBalance >= 45) {
return 1;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 431) {
if (Cs < 791) {
if (SC < 4292) {
if (tcor < 70) {
return 2;
}
else if (tcor >= 70) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (refDCval < 113) {
return 2;
}
else if (refDCval >= 113) {
return 1;
}
}
}
}
else if (SC >= 4292) {
if (AFD < 29) {
return 2;
}
else if (AFD >= 29) {
if (posBalance < 53) {
if (Rs < 4080) {
return 8;
}
else if (Rs >= 4080) {
if (SC < 4955) {
return 2;
}
else if (SC >= 4955) {
return 8;
}
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
}
else if (Cs >= 791) {
if (ssDCval < 121) {
if (gchDC < 6) {
if (Rs < 3252) {
if (Cs < 906) {
return 2;
}
else if (Cs >= 906) {
if (Rs < 3113) {
return 1;
}
else if (Rs >= 3113) {
return 8;
}
}
}
else if (Rs >= 3252) {
if (Cs < 807) {
if (Rs < 5171) {
return 1;
}
else if (Rs >= 5171) {
return 2;
}
}
else if (Cs >= 807) {
return 2;
}
}
}
else if (gchDC >= 6) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
return 1;
}
}
}
else if (ssDCval >= 121) {
if (refDCval < 135) {
if (mu_mv_mag_sq < 15) {
if (tcor < 81) {
if (negBalance < 59) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
if (diffMVdiffVal < 30) {
if (diffTSC < -1) {
if (Cs < 1049) {
return 8;
}
else if (Cs >= 1049) {
if (Rs < 5974) {
return 2;
}
else if (Rs >= 5974) {
if (MVDiff < 43) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (MVDiff >= 43) {
return 8;
}
}
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (diffMVdiffVal >= 30) {
return 2;
}
}
}
else if (negBalance >= 59) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 1;
}
}
}
else if (tcor >= 81) {
if (Rs < 4459) {
return 1;
}
else if (Rs >= 4459) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (MVDiff < 33) {
if (posBalance < 73) {
return 2;
}
else if (posBalance >= 73) {
return 1;
}
}
else if (MVDiff >= 33) {
return 1;
}
}
}
else if (refDCval >= 135) {
if (Rs < 4979) {
if (CsDiff < 5) {
if (Cs < 966) {
return 4;
}
else if (Cs >= 966) {
return 2;
}
}
else if (CsDiff >= 5) {
if (MVDiff < 83) {
return 1;
}
else if (MVDiff >= 83) {
return 2;
}
}
}
else if (Rs >= 4979) {
if (negBalance < 67) {
return 2;
}
else if (negBalance >= 67) {
if (Rs < 5270) {
return 8;
}
else if (Rs >= 5270) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1309) {
if (mu_mv_mag_sq < 111) {
if (RsCsDiff < 101) {
if (ssDCval < 105) {
if (negBalance < 39) {
if (Rs < 741) {
if (Rs < 300) {
return 4;
}
else if (Rs >= 300) {
if (ssDCval < 71) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (Rs >= 741) {
if (SC < 10165) {
if (posBalance < 45) {
if (refDCval < 88) {
if (ssDCval < 64) {
if (Rs < 4848) {
if (mu_mv_mag_sq < 97) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
if (RsDiff < 5) {
if (Rs < 2373) {
return 2;
}
else if (Rs >= 2373) {
return 8;
}
}
else if (RsDiff >= 5) {
if (Rs < 1619) {
return 8;
}
else if (Rs >= 1619) {
if (ssDCval < 49) {
if (Rs < 2265) {
return 2;
}
else if (Rs >= 2265) {
return 1;
}
}
else if (ssDCval >= 49) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (posBalance < 28) {
if (Cs < 1520) {
if (Rs < 1690) {
return 8;
}
else if (Rs >= 1690) {
return 1;
}
}
else if (Cs >= 1520) {
return 8;
}
}
else if (posBalance >= 28) {
if (Cs < 1320) {
if (Rs < 2397) {
return 2;
}
else if (Rs >= 2397) {
return 4;
}
}
else if (Cs >= 1320) {
if (Cs < 1554) {
return 8;
}
else if (Cs >= 1554) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (negBalance < 36) {
if (ssDCval < 56) {
if (MVDiff < 8) {
if (negBalance < 14) {
return 8;
}
else if (negBalance >= 14) {
if (Rs < 2454) {
return 4;
}
else if (Rs >= 2454) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (negBalance < 16) {
return 4;
}
else if (negBalance >= 16) {
if (RsCsDiff < 70) {
return 8;
}
else if (RsCsDiff >= 70) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (negBalance >= 36) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 4848) {
return 4;
}
}
else if (ssDCval >= 64) {
if (ssDCval < 68) {
if (MVDiff < 7) {
if (Rs < 2595) {
if (Rs < 1640) {
return 1;
}
else if (Rs >= 1640) {
return 8;
}
}
else if (Rs >= 2595) {
return 2;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (ssDCval >= 68) {
if (Rs < 3252) {
if (Rs < 1905) {
return 8;
}
else if (Rs >= 1905) {
if (MVDiff < 10) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (diffRsCsdiff < 8) {
return 1;
}
else if (diffRsCsdiff >= 8) {
if (AFD < 7) {
if (Rs < 2391) {
return 8;
}
else if (Rs >= 2391) {
return 4;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (MVDiff >= 10) {
if (Cs < 2545) {
return 8;
}
else if (Cs >= 2545) {
return 4;
}
}
}
}
else if (Rs >= 3252) {
if (CsDiff < 6) {
if (SC < 5028) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (SC >= 5028) {
if (diffRsCsdiff < -31) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 2;
}
}
else if (diffRsCsdiff >= -31) {
if (SC < 5578) {
return 4;
}
else if (SC >= 5578) {
if (SC < 7108) {
if (Cs < 1926) {
if (Cs < 1804) {
if (Rs < 4333) {
return 4;
}
else if (Rs >= 4333) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (Cs >= 1804) {
return 4;
}
}
else if (Cs >= 1926) {
return 8;
}
}
else if (SC >= 7108) {
return 4;
}
}
}
}
}
else if (CsDiff >= 6) {
if (ssDCval < 79) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 1;
}
}
else if (ssDCval >= 79) {
if (Rs < 4829) {
return 8;
}
else if (Rs >= 4829) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 88) {
if (MVDiff < 9) {
if (SC < 4415) {
if (Rs < 2141) {
return 8;
}
else if (Rs >= 2141) {
return 1;
}
}
else if (SC >= 4415) {
if (SC < 6086) {
if (RsCsDiff < 38) {
if (gchDC < 2) {
if (posBalance < 43) {
if (tcor < 98) {
if (Cs < 1577) {
return 4;
}
else if (Cs >= 1577) {
if (posBalance < 40) {
if (Cs < 2149) {
return 4;
}
else if (Cs >= 2149) {
if (Rs < 2404) {
return 4;
}
else if (Rs >= 2404) {
return 8;
}
}
}
else if (posBalance >= 40) {
if (RsCsDiff < 29) {
return 8;
}
else if (RsCsDiff >= 29) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
}
else if (tcor >= 98) {
if (Rs < 2392) {
if (Rs < 2337) {
return 2;
}
else if (Rs >= 2337) {
return 4;
}
}
else if (Rs >= 2392) {
return 8;
}
}
}
else if (posBalance >= 43) {
return 8;
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (RsCsDiff >= 38) {
if (Rs < 4168) {
if (ssDCval < 98) {
if (Cs < 1340) {
return 8;
}
else if (Cs >= 1340) {
return 2;
}
}
else if (ssDCval >= 98) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (Rs >= 4168) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
else if (SC >= 6086) {
if (tcor < 97) {
if (ssDCval < 91) {
if (Cs < 1897) {
return 4;
}
else if (Cs >= 1897) {
return 2;
}
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (tcor >= 97) {
if (Rs < 4645) {
if (Rs < 3492) {
return 8;
}
else if (Rs >= 3492) {
return 2;
}
}
else if (Rs >= 4645) {
return 8;
}
}
}
}
}
else if (MVDiff >= 9) {
if (TSC < 3) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (TSC >= 3) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2913) {
if (mu_mv_mag_sq < 88) {
if (Rs < 1401) {
if (MVDiff < 4) {
if (diffMVdiffVal < -2) {
if (Cs < 1504) {
return 2;
}
else if (Cs >= 1504) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 1587) {
return 4;
}
else if (Cs >= 1587) {
if (Rs < 1053) {
return 8;
}
else if (Rs >= 1053) {
return 2;
}
}
}
}
else if (MVDiff >= 4) {
if (TSC < 6) {
if (SC < 2641) {
if (AFD < 8) {
if (ssDCval < 58) {
return 2;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (AFD >= 8) {
return 4;
}
}
else if (SC >= 2641) {
if (ssDCval < 69) {
if (Rs < 1016) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (Rs >= 1016) {
return 1;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
}
else if (TSC >= 6) {
if (Cs < 2125) {
if (Cs < 1386) {
if (Rs < 1048) {
return 2;
}
else if (Rs >= 1048) {
return 4;
}
}
else if (Cs >= 1386) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (Cs >= 2125) {
return 2;
}
}
}
}
else if (Rs >= 1401) {
if (refDCval < 86) {
if (diffRsCsdiff < 27) {
if (diffMVdiffVal < 2) {
if (mu_mv_mag_sq < 19) {
if (Cs < 1502) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
else if (Cs >= 1502) {
return 8;
}
}
else if (mu_mv_mag_sq >= 19) {
if (tcor < 93) {
if (posBalance < 46) {
if (Rs < 2328) {
return 8;
}
else if (Rs >= 2328) {
return 1;
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (tcor >= 93) {
if (TSC < 7) {
if (RsDiff < 4) {
if (RsCsDiff < 16) {
if (Rs < 1559) {
return 4;
}
else if (Rs >= 1559) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (Rs < 2180) {
return 1;
}
else if (Rs >= 2180) {
return 4;
}
}
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (TSC >= 7) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 1687) {
if (Rs < 1440) {
return 1;
}
else if (Rs >= 1440) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 4;
}
}
}
else if (Cs >= 1687) {
if (Cs < 1878) {
return 1;
}
else if (Cs >= 1878) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 27) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
else if (refDCval >= 86) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 20) {
if (Rs < 2518) {
return 8;
}
else if (Rs >= 2518) {
return 1;
}
}
else if (mu_mv_mag_sq >= 20) {
return 8;
}
}
else if (posBalance >= 57) {
if (Cs < 1631) {
if (TSC < 7) {
if (Rs < 1662) {
return 8;
}
else if (Rs >= 1662) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (Cs >= 1631) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 88) {
if (Rs < 2626) {
if (Cs < 1658) {
if (refDCval < 88) {
if (AFD < 9) {
if (tcor < 97) {
if (MVDiff < 7) {
if (diffAFD < -1) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (MVDiff >= 7) {
if (Rs < 1807) {
return 4;
}
else if (Rs >= 1807) {
return 8;
}
}
}
else if (tcor >= 97) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Cs < 1452) {
return 4;
}
else if (Cs >= 1452) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 4;
}
}
}
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (refDCval >= 88) {
if (Rs < 1052) {
return 4;
}
else if (Rs >= 1052) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
if (diffMVdiffVal < -1) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
}
}
}
else if (Cs >= 1658) {
if (SC < 5004) {
if (mcTcor < 82) {
return 4;
}
else if (mcTcor >= 82) {
if (diffAFD < -1) {
if (negBalance < 38) {
if (diffMVdiffVal < 3) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
if (Rs < 1729) {
return 1;
}
else if (Rs >= 1729) {
return 8;
}
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1666) {
return 4;
}
else if (Rs >= 1666) {
return 8;
}
}
}
else if (negBalance >= 38) {
return 4;
}
}
else if (diffAFD >= -1) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (AFD < 13) {
if (SC < 4899) {
return 8;
}
else if (SC >= 4899) {
if (Rs < 2461) {
return 4;
}
else if (Rs >= 2461) {
return 8;
}
}
}
else if (AFD >= 13) {
return 4;
}
}
}
}
}
else if (SC >= 5004) {
if (AFD < 5) {
if (Rs < 1703) {
return 1;
}
else if (Rs >= 1703) {
return 8;
}
}
else if (AFD >= 5) {
return 4;
}
}
}
}
else if (Rs >= 2626) {
if (ssDCval < 94) {
if (RsCsDiff < 23) {
if (Rs < 2756) {
return 8;
}
else if (Rs >= 2756) {
return 4;
}
}
else if (RsCsDiff >= 23) {
if (Cs < 1320) {
return 2;
}
else if (Cs >= 1320) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 4;
}
}
}
}
else if (ssDCval >= 94) {
if (Cs < 2646) {
return 8;
}
else if (Cs >= 2646) {
if (Rs < 2756) {
return 8;
}
else if (Rs >= 2756) {
return 4;
}
}
}
}
}
}
else if (Rs >= 2913) {
if (Cs < 1935) {
if (MVDiff < 6) {
if (RsCsDiff < 51) {
if (refDCval < 87) {
if (SC < 4556) {
if (Rs < 3114) {
return 2;
}
else if (Rs >= 3114) {
return 1;
}
}
else if (SC >= 4556) {
if (Cs < 1489) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Cs < 1438) {
return 4;
}
else if (Cs >= 1438) {
return 8;
}
}
}
else if (Cs >= 1489) {
return 4;
}
}
}
else if (refDCval >= 87) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (Rs < 3794) {
return 8;
}
else if (Rs >= 3794) {
return 4;
}
}
}
}
else if (RsCsDiff >= 51) {
if (tcor < 92) {
if (SC < 5181) {
return 2;
}
else if (SC >= 5181) {
if (RsDiff < 8) {
if (Rs < 4937) {
if (Rs < 3411) {
return 8;
}
else if (Rs >= 3411) {
if (Rs < 3913) {
return 4;
}
else if (Rs >= 3913) {
return 8;
}
}
}
else if (Rs >= 4937) {
return 4;
}
}
else if (RsDiff >= 8) {
if (Rs < 4773) {
return 2;
}
else if (Rs >= 4773) {
return 4;
}
}
}
}
else if (tcor >= 92) {
if (Rs < 3239) {
if (SC < 4679) {
return 2;
}
else if (SC >= 4679) {
return 4;
}
}
else if (Rs >= 3239) {
return 2;
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 99) {
if (mu_mv_mag_sq < 50) {
if (Rs < 5248) {
return 1;
}
else if (Rs >= 5248) {
return 4;
}
}
else if (mu_mv_mag_sq >= 50) {
if (tcor < 93) {
if (SC < 5876) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (SC >= 5876) {
return 2;
}
}
else if (tcor >= 93) {
if (Cs < 1326) {
return 4;
}
else if (Cs >= 1326) {
return 2;
}
}
}
}
else if (ssDCval >= 99) {
return 8;
}
}
}
else if (Cs >= 1935) {
if (RsDiff < 7) {
if (mu_mv_mag_sq < 97) {
return 8;
}
else if (mu_mv_mag_sq >= 97) {
if (Rs < 2994) {
return 4;
}
else if (Rs >= 2994) {
if (RsDiff < 5) {
if (CsDiff < 3) {
if (Rs < 3954) {
return 4;
}
else if (Rs >= 3954) {
return 8;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (RsDiff >= 5) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
}
else if (RsDiff >= 7) {
if (Rs < 4095) {
if (ssDCval < 83) {
return 1;
}
else if (ssDCval >= 83) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (Rs >= 4095) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 2;
}
}
}
}
}
}
}
else if (SC >= 10165) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 1456) {
if (AFD < 11) {
if (MVDiff < 5) {
if (TSC < 6) {
if (Cs < 1976) {
if (tcor < 96) {
if (mcTcor < 96) {
if (Rs < 1295) {
return 4;
}
else if (Rs >= 1295) {
return 8;
}
}
else if (mcTcor >= 96) {
if (SC < 2187) {
return 4;
}
else if (SC >= 2187) {
return 8;
}
}
}
else if (tcor >= 96) {
return 4;
}
}
else if (Cs >= 1976) {
return 1;
}
}
else if (TSC >= 6) {
if (RsDiff < 5) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (SC < 2401) {
return 8;
}
else if (SC >= 2401) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (CsDiff < 6) {
if (Rs < 1392) {
return 1;
}
else if (Rs >= 1392) {
return 8;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
}
}
else if (RsDiff >= 5) {
if (ssDCval < 67) {
if (Cs < 1725) {
return 4;
}
else if (Cs >= 1725) {
return 8;
}
}
else if (ssDCval >= 67) {
if (Cs < 1889) {
return 8;
}
else if (Cs >= 1889) {
return 2;
}
}
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 57) {
if (RsCsDiff < 22) {
if (Rs < 1044) {
return 2;
}
else if (Rs >= 1044) {
return 8;
}
}
else if (RsCsDiff >= 22) {
if (negBalance < 57) {
return 4;
}
else if (negBalance >= 57) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (refDCval < 78) {
if (Rs < 1250) {
if (mu_mv_mag_sq < 76) {
if (TSC < 5) {
if (Rs < 1058) {
return 1;
}
else if (Rs >= 1058) {
return 2;
}
}
else if (TSC >= 5) {
if (Cs < 1655) {
if (Rs < 686) {
if (Rs < 496) {
return 1;
}
else if (Rs >= 496) {
return 8;
}
}
else if (Rs >= 686) {
if (Rs < 997) {
return 2;
}
else if (Rs >= 997) {
return 1;
}
}
}
else if (Cs >= 1655) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (Rs < 1066) {
return 8;
}
else if (Rs >= 1066) {
return 4;
}
}
}
else if (Rs >= 1250) {
return 1;
}
}
else if (refDCval >= 78) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 1290) {
if (Rs < 799) {
return 8;
}
else if (Rs >= 799) {
return 2;
}
}
else if (Rs >= 1290) {
return 4;
}
}
}
}
}
}
else if (AFD >= 11) {
if (RsCsDiff < 77) {
if (TSC < 11) {
if (TSC < 7) {
if (SC < 1890) {
return 1;
}
else if (SC >= 1890) {
if (ssDCval < 59) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 2;
}
}
else if (ssDCval >= 59) {
if (Rs < 1016) {
if (gchDC < 1) {
if (Rs < 719) {
return 8;
}
else if (Rs >= 719) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Rs >= 1016) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (tcor < 90) {
return 4;
}
else if (tcor >= 90) {
return 2;
}
}
}
}
}
}
else if (TSC >= 7) {
if (ssDCval < 75) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (Cs < 1327) {
return 2;
}
else if (Cs >= 1327) {
if (diffMVdiffVal < -11) {
return 4;
}
else if (diffMVdiffVal >= -11) {
return 8;
}
}
}
}
else if (ssDCval >= 75) {
if (SC < 2554) {
if (MVDiff < 8) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
else if (MVDiff >= 8) {
if (refDCval < 86) {
if (MVDiff < 28) {
return 4;
}
else if (MVDiff >= 28) {
return 8;
}
}
else if (refDCval >= 86) {
if (Cs < 1656) {
return 8;
}
else if (Cs >= 1656) {
return 4;
}
}
}
}
else if (SC >= 2554) {
if (RsDiff < 6) {
if (mu_mv_mag_sq < 27) {
if (Cs < 1999) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (Rs < 1184) {
return 8;
}
else if (Rs >= 1184) {
return 4;
}
}
}
else if (Cs >= 1999) {
return 2;
}
}
else if (mu_mv_mag_sq >= 27) {
if (Rs < 1337) {
return 4;
}
else if (Rs >= 1337) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
}
}
}
}
else if (TSC >= 11) {
if (Rs < 1449) {
return 8;
}
else if (Rs >= 1449) {
if (Rs < 1451) {
return 4;
}
else if (Rs >= 1451) {
return 8;
}
}
}
}
else if (RsCsDiff >= 77) {
if (tcor < 72) {
if (posBalance < 56) {
if (Rs < 867) {
return 4;
}
else if (Rs >= 867) {
return 2;
}
}
else if (posBalance >= 56) {
if (Cs < 1392) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
return 2;
}
}
else if (Cs >= 1392) {
return 4;
}
}
}
else if (tcor >= 72) {
if (MVDiff < 23) {
if (Cs < 1420) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (Cs >= 1420) {
if (diffTSC < -1) {
if (posBalance < 64) {
return 4;
}
else if (posBalance >= 64) {
return 2;
}
}
else if (diffTSC >= -1) {
if (ssDCval < 59) {
if (Rs < 917) {
return 4;
}
else if (Rs >= 917) {
return 2;
}
}
else if (ssDCval >= 59) {
if (Cs < 2097) {
if (ssDCval < 81) {
if (Cs < 1736) {
return 8;
}
else if (Cs >= 1736) {
return 4;
}
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (Cs >= 2097) {
return 2;
}
}
}
}
}
else if (MVDiff >= 23) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
if (Cs < 1318) {
return 2;
}
else if (Cs >= 1318) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1456) {
if (CsDiff < 7) {
if (mcTcor < 94) {
if (mu_mv_mag_sq < 93) {
if (Rs < 2224) {
if (negBalance < 49) {
if (ssDCval < 84) {
if (AFD < 14) {
if (diffRsCsdiff < -17) {
if (diffMVdiffVal < 0) {
if (Cs < 1899) {
return 8;
}
else if (Cs >= 1899) {
return 4;
}
}
else if (diffMVdiffVal >= 0) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
return 1;
}
}
}
else if (diffRsCsdiff >= -17) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (Cs < 1836) {
return 8;
}
else if (Cs >= 1836) {
if (Rs < 2057) {
return 4;
}
else if (Rs >= 2057) {
return 8;
}
}
}
}
}
else if (AFD >= 14) {
if (Cs < 1438) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 1;
}
}
else if (Cs >= 1438) {
return 4;
}
}
}
else if (ssDCval >= 84) {
if (MVDiff < 14) {
if (Rs < 1834) {
if (mu_mv_mag_sq < 20) {
if (Rs < 1772) {
return 2;
}
else if (Rs >= 1772) {
return 8;
}
}
else if (mu_mv_mag_sq >= 20) {
return 8;
}
}
else if (Rs >= 1834) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (MVDiff >= 14) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (AFD < 16) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (tcor < 84) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
else if (tcor >= 84) {
if (ssDCval < 89) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
if (diffRsCsdiff < 11) {
return 8;
}
else if (diffRsCsdiff >= 11) {
return 1;
}
}
}
else if (ssDCval >= 89) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
}
else if (AFD >= 16) {
if (refDCval < 79) {
return 8;
}
else if (refDCval >= 79) {
if (posBalance < 60) {
if (Cs < 1621) {
return 8;
}
else if (Cs >= 1621) {
return 1;
}
}
else if (posBalance >= 60) {
if (Rs < 1684) {
return 4;
}
else if (Rs >= 1684) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2224) {
if (posBalance < 45) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 3808) {
return 2;
}
else if (Rs >= 3808) {
return 4;
}
}
}
else if (posBalance >= 45) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (MVDiff < 4) {
if (Cs < 2862) {
return 8;
}
else if (Cs >= 2862) {
if (Rs < 3116) {
return 4;
}
else if (Rs >= 3116) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (ssDCval < 83) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
if (Cs < 2041) {
if (Rs < 2503) {
return 8;
}
else if (Rs >= 2503) {
return 4;
}
}
else if (Cs >= 2041) {
return 8;
}
}
}
else if (ssDCval >= 83) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (Rs < 3646) {
if (posBalance < 66) {
if (Cs < 2428) {
if (MVDiff < 7) {
if (ssDCval < 47) {
if (Rs < 1611) {
return 4;
}
else if (Rs >= 1611) {
return 8;
}
}
else if (ssDCval >= 47) {
if (diffRsCsdiff < 44) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
if (RsCsDiff < 80) {
if (Rs < 3073) {
return 4;
}
else if (Rs >= 3073) {
return 8;
}
}
else if (RsCsDiff >= 80) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 44) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (tcor < 91) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
else if (tcor >= 91) {
return 4;
}
}
}
else if (Cs >= 2428) {
if (Rs < 2584) {
return 4;
}
else if (Rs >= 2584) {
return 8;
}
}
}
else if (posBalance >= 66) {
return 4;
}
}
else if (Rs >= 3646) {
if (Rs < 4453) {
return 4;
}
else if (Rs >= 4453) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (Cs < 3253) {
if (SC < 4756) {
if (ssDCval < 89) {
if (Cs < 1560) {
if (Rs < 2783) {
if (tcor < 88) {
if (CsDiff < 6) {
if (Rs < 2445) {
return 4;
}
else if (Rs >= 2445) {
return 2;
}
}
else if (CsDiff >= 6) {
if (Rs < 1593) {
return 2;
}
else if (Rs >= 1593) {
return 8;
}
}
}
else if (tcor >= 88) {
if (MVDiff < 7) {
if (refDCval < 82) {
if (gchDC < 1) {
if (mu_mv_mag_sq < 87) {
if (RsCsDiff < 57) {
if (TSC < 7) {
if (diffRsCsdiff < -4) {
return 1;
}
else if (diffRsCsdiff >= -4) {
if (Cs < 1351) {
return 1;
}
else if (Cs >= 1351) {
return 8;
}
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (RsCsDiff >= 57) {
return 8;
}
}
else if (mu_mv_mag_sq >= 87) {
if (Rs < 2355) {
return 8;
}
else if (Rs >= 2355) {
return 2;
}
}
}
else if (gchDC >= 1) {
if (MVDiff < 3) {
if (Rs < 1688) {
if (Rs < 1478) {
return 4;
}
else if (Rs >= 1478) {
return 8;
}
}
else if (Rs >= 1688) {
return 1;
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 71) {
return 8;
}
else if (mu_mv_mag_sq >= 71) {
if (Rs < 1926) {
return 4;
}
else if (Rs >= 1926) {
return 1;
}
}
}
}
}
else if (refDCval >= 82) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
if (mu_mv_mag_sq < 94) {
if (MVDiff < 4) {
if (negBalance < 41) {
return 4;
}
else if (negBalance >= 41) {
return 1;
}
}
else if (MVDiff >= 4) {
if (Rs < 2179) {
if (Cs < 1370) {
return 8;
}
else if (Cs >= 1370) {
return 2;
}
}
else if (Rs >= 2179) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 94) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
if (negBalance < 47) {
if (Cs < 1356) {
if (diffTSC < -1) {
if (Cs < 1314) {
return 2;
}
else if (Cs >= 1314) {
return 8;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (Cs >= 1356) {
return 8;
}
}
else if (negBalance >= 47) {
return 8;
}
}
}
}
else if (Rs >= 2783) {
if (Cs < 1437) {
return 2;
}
else if (Cs >= 1437) {
if (Rs < 3104) {
return 4;
}
else if (Rs >= 3104) {
return 2;
}
}
}
}
else if (Cs >= 1560) {
if (AFD < 8) {
if (diffAFD < -1) {
if (SC < 3805) {
return 1;
}
else if (SC >= 3805) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 1;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (AFD >= 8) {
if (refDCval < 78) {
if (negBalance < 51) {
if (Rs < 2127) {
return 4;
}
else if (Rs >= 2127) {
if (SC < 4269) {
return 8;
}
else if (SC >= 4269) {
return 4;
}
}
}
else if (negBalance >= 51) {
return 8;
}
}
else if (refDCval >= 78) {
if (MVDiff < 2) {
if (SC < 4087) {
if (SC < 3831) {
if (Rs < 1998) {
return 4;
}
else if (Rs >= 1998) {
return 1;
}
}
else if (SC >= 3831) {
return 8;
}
}
else if (SC >= 4087) {
if (Rs < 2204) {
return 1;
}
else if (Rs >= 2204) {
if (Rs < 2425) {
return 8;
}
else if (Rs >= 2425) {
return 1;
}
}
}
}
else if (MVDiff >= 2) {
if (AFD < 14) {
if (negBalance < 42) {
if (Rs < 2429) {
return 4;
}
else if (Rs >= 2429) {
return 1;
}
}
else if (negBalance >= 42) {
if (Cs < 1594) {
if (diffRsCsdiff < 7) {
return 8;
}
else if (diffRsCsdiff >= 7) {
if (Rs < 2178) {
return 4;
}
else if (Rs >= 2178) {
return 1;
}
}
}
else if (Cs >= 1594) {
return 8;
}
}
}
else if (AFD >= 14) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 89) {
if (Rs < 2186) {
if (Rs < 1816) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Rs < 1482) {
return 8;
}
else if (Rs >= 1482) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
if (mu_mv_mag_sq < 103) {
if (Rs < 1525) {
if (Rs < 1504) {
return 2;
}
else if (Rs >= 1504) {
return 8;
}
}
else if (Rs >= 1525) {
return 2;
}
}
else if (mu_mv_mag_sq >= 103) {
return 8;
}
}
}
}
}
else if (Rs >= 1816) {
if (Cs < 1937) {
if (Rs < 2072) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (AFD < 11) {
if (Rs < 2067) {
if (mu_mv_mag_sq < 33) {
if (Cs < 1771) {
return 8;
}
else if (Cs >= 1771) {
return 2;
}
}
else if (mu_mv_mag_sq >= 33) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 2;
}
}
}
else if (Rs >= 2067) {
return 2;
}
}
else if (AFD >= 11) {
if (MVDiff < 9) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
}
}
else if (Rs >= 2072) {
return 8;
}
}
else if (Cs >= 1937) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (posBalance < 7) {
if (Rs < 2166) {
return 1;
}
else if (Rs >= 2166) {
return 2;
}
}
else if (posBalance >= 7) {
return 2;
}
}
}
}
}
else if (Rs >= 2186) {
if (posBalance < 47) {
if (diffAFD < -1) {
if (SC < 4251) {
return 2;
}
else if (SC >= 4251) {
return 8;
}
}
else if (diffAFD >= -1) {
if (negBalance < 47) {
if (ssDCval < 93) {
return 1;
}
else if (ssDCval >= 93) {
return 4;
}
}
else if (negBalance >= 47) {
return 8;
}
}
}
else if (posBalance >= 47) {
if (MVDiff < 27) {
if (Cs < 1885) {
if (posBalance < 48) {
if (Rs < 2772) {
return 8;
}
else if (Rs >= 2772) {
return 4;
}
}
else if (posBalance >= 48) {
return 8;
}
}
else if (Cs >= 1885) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (MVDiff >= 27) {
return 4;
}
}
}
}
}
else if (SC >= 4756) {
if (Cs < 1624) {
if (ssDCval < 94) {
if (RsCsDiff < 77) {
if (mu_mv_mag_sq < 87) {
if (Rs < 3685) {
return 4;
}
else if (Rs >= 3685) {
return 1;
}
}
else if (mu_mv_mag_sq >= 87) {
return 2;
}
}
else if (RsCsDiff >= 77) {
return 2;
}
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (Cs >= 1624) {
if (TSC < 5) {
if (Rs < 2757) {
return 4;
}
else if (Rs >= 2757) {
if (Cs < 1913) {
return 2;
}
else if (Cs >= 1913) {
return 8;
}
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 57) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 8;
}
}
else if (mu_mv_mag_sq >= 57) {
if (SC < 5217) {
if (negBalance < 52) {
if (Rs < 2348) {
return 4;
}
else if (Rs >= 2348) {
if (Rs < 3346) {
return 8;
}
else if (Rs >= 3346) {
return 4;
}
}
}
else if (negBalance >= 52) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (SC >= 5217) {
if (mu_mv_mag_sq < 85) {
if (Cs < 3045) {
return 4;
}
else if (Cs >= 3045) {
return 1;
}
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 2925) {
return 4;
}
else if (Rs >= 2925) {
if (Rs < 3613) {
return 8;
}
else if (Rs >= 3613) {
if (tcor < 93) {
if (Cs < 3176) {
return 8;
}
else if (Cs >= 3176) {
return 4;
}
}
else if (tcor >= 93) {
if (Cs < 2155) {
return 2;
}
else if (Cs >= 2155) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 3253) {
if (Rs < 2978) {
return 4;
}
else if (Rs >= 2978) {
return 8;
}
}
}
}
else if (CsDiff >= 7) {
if (Rs < 3007) {
if (Cs < 2918) {
if (refDCval < 98) {
if (Cs < 2217) {
if (tcor < 81) {
if (MVDiff < 41) {
if (TSC < 11) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (ssDCval < 90) {
if (refDCval < 77) {
if (refDCval < 51) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
return 4;
}
}
else if (refDCval >= 51) {
if (Rs < 1543) {
return 4;
}
else if (Rs >= 1543) {
return 8;
}
}
}
else if (refDCval >= 77) {
return 4;
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (TSC >= 11) {
if (MVDiff < 32) {
if (Cs < 2076) {
return 8;
}
else if (Cs >= 2076) {
if (Rs < 2125) {
return 4;
}
else if (Rs >= 2125) {
return 8;
}
}
}
else if (MVDiff >= 32) {
return 4;
}
}
}
else if (MVDiff >= 41) {
return 1;
}
}
else if (tcor >= 81) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
if (Rs < 2842) {
if (diffRsCsdiff < 31) {
return 8;
}
else if (diffRsCsdiff >= 31) {
if (ssDCval < 91) {
if (ssDCval < 79) {
return 4;
}
else if (ssDCval >= 79) {
return 8;
}
}
else if (ssDCval >= 91) {
return 1;
}
}
}
else if (Rs >= 2842) {
if (Rs < 2920) {
return 4;
}
else if (Rs >= 2920) {
return 8;
}
}
}
}
}
else if (Cs >= 2217) {
if (mcTcor < 93) {
if (Cs < 2710) {
return 8;
}
else if (Cs >= 2710) {
if (Rs < 2507) {
return 4;
}
else if (Rs >= 2507) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (SC < 4877) {
if (Rs < 1786) {
return 8;
}
else if (Rs >= 1786) {
return 1;
}
}
else if (SC >= 4877) {
if (Cs < 2726) {
return 8;
}
else if (Cs >= 2726) {
if (Rs < 2837) {
return 2;
}
else if (Rs >= 2837) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 98) {
if (Rs < 2133) {
if (mcTcor < 94) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (Rs >= 2133) {
return 8;
}
}
}
else if (Cs >= 2918) {
return 4;
}
}
else if (Rs >= 3007) {
if (mu_mv_mag_sq < 57) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (Rs < 3935) {
return 1;
}
else if (Rs >= 3935) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (SC < 7056) {
if (MVDiff < 1) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (Rs < 3471) {
return 8;
}
else if (Rs >= 3471) {
return 4;
}
}
}
else if (MVDiff >= 1) {
if (SC < 5835) {
return 8;
}
else if (SC >= 5835) {
return 1;
}
}
}
else if (SC >= 7056) {
if (Cs < 3194) {
return 4;
}
else if (Cs >= 3194) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 105) {
if (ssDCval < 163) {
if (Rs < 4404) {
if (Rs < 1494) {
if (mu_mv_mag_sq < 46) {
if (ssDCval < 146) {
if (ssDCval < 110) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
if (Rs < 1268) {
return 2;
}
else if (Rs >= 1268) {
return 8;
}
}
}
else if (ssDCval >= 110) {
if (refDCval < 125) {
if (SC < 3122) {
if (diffRsCsdiff < 16) {
return 8;
}
else if (diffRsCsdiff >= 16) {
if (Cs < 1379) {
return 2;
}
else if (Cs >= 1379) {
return 8;
}
}
}
else if (SC >= 3122) {
if (posBalance < 56) {
if (Rs < 1489) {
return 4;
}
else if (Rs >= 1489) {
return 1;
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
else if (refDCval >= 125) {
if (AFD < 20) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (ssDCval < 131) {
return 8;
}
else if (ssDCval >= 131) {
if (mu_mv_mag_sq < 31) {
if (RsCsDiff < 37) {
return 8;
}
else if (RsCsDiff >= 37) {
return 1;
}
}
else if (mu_mv_mag_sq >= 31) {
return 8;
}
}
}
}
else if (AFD >= 20) {
return 4;
}
}
}
}
else if (ssDCval >= 146) {
if (mu_mv_mag_sq < 38) {
return 2;
}
else if (mu_mv_mag_sq >= 38) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (RsDiff < 3) {
if (MVDiff < 6) {
if (Cs < 1515) {
if (SC < 2886) {
if (AFD < 3) {
if (Rs < 851) {
return 1;
}
else if (Rs >= 851) {
return 4;
}
}
else if (AFD >= 3) {
if (RsCsDiff < 12) {
return 8;
}
else if (RsCsDiff >= 12) {
if (ssDCval < 119) {
return 4;
}
else if (ssDCval >= 119) {
if (Rs < 972) {
return 8;
}
else if (Rs >= 972) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
}
}
else if (SC >= 2886) {
return 1;
}
}
else if (Cs >= 1515) {
if (Rs < 1018) {
return 8;
}
else if (Rs >= 1018) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
if (Rs < 1083) {
if (AFD < 7) {
if (Cs < 1546) {
return 4;
}
else if (Cs >= 1546) {
return 8;
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (Rs >= 1083) {
return 4;
}
}
}
}
}
else if (MVDiff >= 6) {
if (refDCval < 113) {
return 4;
}
else if (refDCval >= 113) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (Cs < 1670) {
if (ssDCval < 159) {
return 8;
}
else if (ssDCval >= 159) {
return 2;
}
}
else if (Cs >= 1670) {
if (CsDiff < 7) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (gchDC < 2) {
if (ssDCval < 152) {
return 4;
}
else if (ssDCval >= 152) {
return 8;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (CsDiff >= 7) {
return 8;
}
}
}
}
}
else if (Rs >= 1494) {
if (mu_mv_mag_sq < 42) {
if (Rs < 2017) {
if (refDCval < 123) {
if (mcTcor < 97) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (CsDiff < 6) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (CsDiff >= 6) {
if (SC < 3265) {
return 4;
}
else if (SC >= 3265) {
return 8;
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 114) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (refDCval >= 114) {
return 8;
}
}
}
else if (refDCval >= 123) {
if (ssDCval < 139) {
if (negBalance < 49) {
if (RsDiff < 6) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (negBalance >= 49) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
}
else if (ssDCval >= 139) {
if (Rs < 1696) {
if (mu_mv_mag_sq < 12) {
return 2;
}
else if (mu_mv_mag_sq >= 12) {
if (Rs < 1580) {
return 1;
}
else if (Rs >= 1580) {
return 8;
}
}
}
else if (Rs >= 1696) {
return 8;
}
}
}
}
else if (Rs >= 2017) {
if (ssDCval < 110) {
if (Rs < 3124) {
if (negBalance < 51) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (Rs < 2192) {
return 8;
}
else if (Rs >= 2192) {
return 1;
}
}
}
}
else if (negBalance >= 51) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Rs < 2406) {
return 1;
}
else if (Rs >= 2406) {
return 8;
}
}
}
}
else if (Rs >= 3124) {
return 8;
}
}
else if (ssDCval >= 110) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (Rs < 3427) {
return 8;
}
else if (Rs >= 3427) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (tcor < 97) {
if (refDCval < 142) {
if (refDCval < 123) {
if (TSC < 11) {
if (TSC < 3) {
if (Rs < 3476) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
return 4;
}
}
else if (Rs >= 3476) {
return 1;
}
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 71) {
if (RsCsDiff < 43) {
if (RsCsDiff < 38) {
if (diffMVdiffVal < 7) {
return 8;
}
else if (diffMVdiffVal >= 7) {
if (Rs < 3082) {
return 4;
}
else if (Rs >= 3082) {
return 8;
}
}
}
else if (RsCsDiff >= 38) {
if (Cs < 1447) {
return 8;
}
else if (Cs >= 1447) {
if (Rs < 3292) {
return 4;
}
else if (Rs >= 3292) {
return 8;
}
}
}
}
else if (RsCsDiff >= 43) {
if (Cs < 2464) {
return 8;
}
else if (Cs >= 2464) {
if (Rs < 3647) {
return 4;
}
else if (Rs >= 3647) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 71) {
return 8;
}
}
}
else if (TSC >= 11) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (refDCval < 105) {
return 1;
}
else if (refDCval >= 105) {
if (diffRsCsdiff < -36) {
return 4;
}
else if (diffRsCsdiff >= -36) {
return 8;
}
}
}
}
}
else if (refDCval >= 123) {
if (posBalance < 55) {
if (SC < 4640) {
if (Cs < 1719) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (negBalance < 29) {
if (Rs < 2293) {
return 8;
}
else if (Rs >= 2293) {
return 4;
}
}
else if (negBalance >= 29) {
if (mu_mv_mag_sq < 107) {
return 8;
}
else if (mu_mv_mag_sq >= 107) {
return 2;
}
}
}
}
else if (Cs >= 1719) {
if (Cs < 1845) {
if (mu_mv_mag_sq < 93) {
if (Rs < 1638) {
return 8;
}
else if (Rs >= 1638) {
return 1;
}
}
else if (mu_mv_mag_sq >= 93) {
return 8;
}
}
else if (Cs >= 1845) {
return 8;
}
}
}
else if (SC >= 4640) {
return 8;
}
}
else if (posBalance >= 55) {
if (mu_mv_mag_sq < 99) {
if (mcTcor < 89) {
if (SC < 3664) {
return 1;
}
else if (SC >= 3664) {
if (Rs < 3532) {
return 8;
}
else if (Rs >= 3532) {
return 1;
}
}
}
else if (mcTcor >= 89) {
if (Cs < 2334) {
if (Cs < 1321) {
return 1;
}
else if (Cs >= 1321) {
if (RsCsDiff < 83) {
return 8;
}
else if (RsCsDiff >= 83) {
if (mu_mv_mag_sq < 88) {
if (ssDCval < 127) {
if (Cs < 1755) {
return 8;
}
else if (Cs >= 1755) {
return 1;
}
}
else if (ssDCval >= 127) {
return 8;
}
}
else if (mu_mv_mag_sq >= 88) {
return 1;
}
}
}
}
else if (Cs >= 2334) {
if (tcor < 89) {
if (Rs < 2735) {
return 4;
}
else if (Rs >= 2735) {
return 1;
}
}
else if (tcor >= 89) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (ssDCval < 136) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Rs < 3058) {
return 8;
}
else if (Rs >= 3058) {
return 4;
}
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
}
else if (refDCval >= 142) {
if (posBalance < 66) {
if (posBalance < 29) {
return 2;
}
else if (posBalance >= 29) {
return 8;
}
}
else if (posBalance >= 66) {
if (Rs < 2056) {
return 4;
}
else if (Rs >= 2056) {
return 8;
}
}
}
}
else if (tcor >= 97) {
if (Cs < 1376) {
if (Rs < 2629) {
return 8;
}
else if (Rs >= 2629) {
if (ssDCval < 120) {
if (Rs < 3661) {
return 8;
}
else if (Rs >= 3661) {
return 4;
}
}
else if (ssDCval >= 120) {
return 2;
}
}
}
else if (Cs >= 1376) {
if (posBalance < 17) {
if (negBalance < 46) {
if (posBalance < 9) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (posBalance >= 9) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
else if (posBalance >= 17) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 2058) {
return 2;
}
else if (Rs >= 2058) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 4404) {
if (Cs < 1716) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (Cs >= 1716) {
if (Cs < 3432) {
if (MVDiff < 3) {
if (Rs < 6055) {
return 2;
}
else if (Rs >= 6055) {
return 4;
}
}
else if (MVDiff >= 3) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 1;
}
}
}
else if (Cs >= 3432) {
return 8;
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2283) {
if (Rs < 1015) {
return 1;
}
else if (Rs >= 1015) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (Rs >= 2283) {
return 8;
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1897) {
if (Rs < 3320) {
if (AFD < 18) {
if (refDCval < 59) {
if (refDCval < 46) {
if (refDCval < 43) {
if (diffAFD < -9) {
return 1;
}
else if (diffAFD >= -9) {
return 8;
}
}
else if (refDCval >= 43) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
if (RsCsDiff < 226) {
return 8;
}
else if (RsCsDiff >= 226) {
return 1;
}
}
}
}
else if (refDCval >= 46) {
if (Rs < 1991) {
if (mcTcor < 82) {
if (ssDCval < 55) {
if (CsDiff < 8) {
if (Rs < 1764) {
return 4;
}
else if (Rs >= 1764) {
return 8;
}
}
else if (CsDiff >= 8) {
if (diffTSC < 2) {
return 4;
}
else if (diffTSC >= 2) {
if (Rs < 1651) {
return 1;
}
else if (Rs >= 1651) {
return 4;
}
}
}
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (mcTcor >= 82) {
if (MVDiff < 18) {
if (AFD < 10) {
if (Rs < 1690) {
return 1;
}
else if (Rs >= 1690) {
return 4;
}
}
else if (AFD >= 10) {
return 8;
}
}
else if (MVDiff >= 18) {
if (Rs < 1094) {
return 4;
}
else if (Rs >= 1094) {
return 1;
}
}
}
}
else if (Rs >= 1991) {
if (CsDiff < 9) {
if (Cs < 1397) {
return 4;
}
else if (Cs >= 1397) {
return 8;
}
}
else if (CsDiff >= 9) {
if (AFD < 13) {
if (Rs < 2995) {
return 1;
}
else if (Rs >= 2995) {
return 8;
}
}
else if (AFD >= 13) {
return 8;
}
}
}
}
}
else if (refDCval >= 59) {
if (TSC < 9) {
if (RsDiff < 9) {
if (RsDiff < 6) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (TSC < 6) {
if (mcTcor < 96) {
if (Rs < 2145) {
return 8;
}
else if (Rs >= 2145) {
return 1;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (Rs < 1548) {
if (MVDiff < 15) {
if (Cs < 1421) {
return 1;
}
else if (Cs >= 1421) {
if (SC < 2846) {
return 4;
}
else if (SC >= 2846) {
return 8;
}
}
}
else if (MVDiff >= 15) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (Cs < 1327) {
return 8;
}
else if (Cs >= 1327) {
return 2;
}
}
}
}
else if (Rs >= 1548) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 88) {
if (Rs < 2553) {
return 2;
}
else if (Rs >= 2553) {
return 4;
}
}
else if (ssDCval >= 88) {
if (Rs < 1637) {
return 4;
}
else if (Rs >= 1637) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
if (MVDiff < 22) {
if (Cs < 1654) {
return 1;
}
else if (Cs >= 1654) {
return 8;
}
}
else if (MVDiff >= 22) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 100) {
if (Rs < 1938) {
if (negBalance < 64) {
return 8;
}
else if (negBalance >= 64) {
return 1;
}
}
else if (Rs >= 1938) {
if (RsDiff < 10) {
if (Cs < 1543) {
if (MVDiff < 13) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (tcor < 85) {
return 2;
}
else if (tcor >= 85) {
return 4;
}
}
}
else if (MVDiff >= 13) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 2;
}
}
}
else if (Cs >= 1543) {
if (gchDC < 1) {
if (mcTcor < 92) {
if (Cs < 1736) {
return 8;
}
else if (Cs >= 1736) {
return 2;
}
}
else if (mcTcor >= 92) {
if (Rs < 2836) {
return 4;
}
else if (Rs >= 2836) {
return 1;
}
}
}
else if (gchDC >= 1) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
else if (RsDiff >= 10) {
if (Cs < 1559) {
if (RsCsDiff < 161) {
if (Cs < 1377) {
if (Cs < 1359) {
if (Rs < 2804) {
return 4;
}
else if (Rs >= 2804) {
if (Cs < 1327) {
return 4;
}
else if (Cs >= 1327) {
return 2;
}
}
}
else if (Cs >= 1359) {
return 4;
}
}
else if (Cs >= 1377) {
if (AFD < 16) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (Rs < 2435) {
return 2;
}
else if (Rs >= 2435) {
return 4;
}
}
}
else if (AFD >= 16) {
return 4;
}
}
}
else if (RsCsDiff >= 161) {
return 2;
}
}
else if (Cs >= 1559) {
if (Rs < 2483) {
return 8;
}
else if (Rs >= 2483) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (SC < 5005) {
return 2;
}
else if (SC >= 5005) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 100) {
if (posBalance < 49) {
if (ssDCval < 116) {
if (SC < 4170) {
return 4;
}
else if (SC >= 4170) {
return 8;
}
}
else if (ssDCval >= 116) {
if (gchDC < 1) {
if (Rs < 1528) {
return 1;
}
else if (Rs >= 1528) {
return 8;
}
}
else if (gchDC >= 1) {
if (Rs < 2823) {
return 2;
}
else if (Rs >= 2823) {
return 4;
}
}
}
}
else if (posBalance >= 49) {
if (RsCsDiff < 189) {
if (Cs < 1358) {
if (diffMVdiffVal < -4) {
if (Rs < 2581) {
return 2;
}
else if (Rs >= 2581) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
if (Cs < 1338) {
return 4;
}
else if (Cs >= 1338) {
return 8;
}
}
}
else if (Cs >= 1358) {
if (ssDCval < 166) {
if (posBalance < 83) {
return 8;
}
else if (posBalance >= 83) {
return 4;
}
}
else if (ssDCval >= 166) {
return 2;
}
}
}
else if (RsCsDiff >= 189) {
if (TSC < 8) {
if (Rs < 3056) {
return 8;
}
else if (Rs >= 3056) {
return 2;
}
}
else if (TSC >= 8) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
return 4;
}
}
}
}
}
}
}
else if (TSC >= 9) {
if (Rs < 2501) {
if (mu_mv_mag_sq < 3) {
if (Cs < 1528) {
if (tcor < 83) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 1;
}
}
else if (tcor >= 83) {
if (ssDCval < 86) {
if (Rs < 2015) {
return 1;
}
else if (Rs >= 2015) {
return 8;
}
}
else if (ssDCval >= 86) {
return 8;
}
}
}
else if (Cs >= 1528) {
if (refDCval < 83) {
if (TSC < 11) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
return 8;
}
}
else if (TSC >= 11) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 4;
}
}
}
else if (refDCval >= 83) {
if (Rs < 1579) {
return 4;
}
else if (Rs >= 1579) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 11) {
if (mcTcor < 97) {
if (mcTcor < 82) {
if (Rs < 2168) {
return 8;
}
else if (Rs >= 2168) {
return 4;
}
}
else if (mcTcor >= 82) {
if (ssDCval < 80) {
if (RsCsDiff < 110) {
return 4;
}
else if (RsCsDiff >= 110) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 2;
}
}
}
else if (ssDCval >= 80) {
if (Rs < 1367) {
if (Rs < 1240) {
return 8;
}
else if (Rs >= 1240) {
return 2;
}
}
else if (Rs >= 1367) {
if (posBalance < 50) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (posBalance >= 50) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
return 2;
}
}
else if (TSC >= 11) {
if (SC < 3551) {
if (mcTcor < 91) {
if (CsDiff < 9) {
if (SC < 3394) {
return 8;
}
else if (SC >= 3394) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2160) {
return 8;
}
else if (Rs >= 2160) {
return 1;
}
}
}
}
else if (CsDiff >= 9) {
return 1;
}
}
else if (mcTcor >= 91) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 2;
}
}
}
else if (SC >= 3551) {
if (RsCsDiff < 141) {
return 8;
}
else if (RsCsDiff >= 141) {
if (posBalance < 58) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2501) {
if (ssDCval < 107) {
if (negBalance < 47) {
if (TSC < 11) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (TSC >= 11) {
if (Rs < 3024) {
return 8;
}
else if (Rs >= 3024) {
return 1;
}
}
}
else if (negBalance >= 47) {
if (ssDCval < 102) {
if (negBalance < 50) {
if (Cs < 1698) {
if (RsDiff < 14) {
return 2;
}
else if (RsDiff >= 14) {
return 8;
}
}
else if (Cs >= 1698) {
if (AFD < 17) {
if (Cs < 1850) {
return 8;
}
else if (Cs >= 1850) {
return 4;
}
}
else if (AFD >= 17) {
return 4;
}
}
}
else if (negBalance >= 50) {
if (Rs < 2574) {
if (refDCval < 85) {
return 4;
}
else if (refDCval >= 85) {
return 8;
}
}
else if (Rs >= 2574) {
return 8;
}
}
}
else if (ssDCval >= 102) {
if (MVDiff < 16) {
if (Rs < 3226) {
return 4;
}
else if (Rs >= 3226) {
return 8;
}
}
else if (MVDiff >= 16) {
return 8;
}
}
}
}
else if (ssDCval >= 107) {
return 8;
}
}
}
}
}
else if (AFD >= 18) {
if (refDCval < 105) {
if (posBalance < 49) {
if (gchDC < 7) {
if (RsDiff < 18) {
return 2;
}
else if (RsDiff >= 18) {
if (Rs < 2282) {
return 8;
}
else if (Rs >= 2282) {
return 1;
}
}
}
else if (gchDC >= 7) {
if (Rs < 961) {
return 2;
}
else if (Rs >= 961) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (mcTcor < 95) {
if (RsDiff < 15) {
if (refDCval < 85) {
if (RsCsDiff < 213) {
if (Cs < 1429) {
if (refDCval < 73) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (MVDiff < 25) {
if (Rs < 1306) {
return 2;
}
else if (Rs >= 1306) {
return 4;
}
}
else if (MVDiff >= 25) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (AFD < 23) {
return 4;
}
else if (AFD >= 23) {
return 8;
}
}
}
}
}
else if (refDCval >= 73) {
if (RsDiff < 11) {
if (posBalance < 67) {
if (SC < 2840) {
return 2;
}
else if (SC >= 2840) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 4;
}
}
}
else if (posBalance >= 67) {
return 1;
}
}
else if (RsDiff >= 11) {
if (Rs < 2896) {
return 8;
}
else if (Rs >= 2896) {
return 4;
}
}
}
}
else if (Cs >= 1429) {
if (MVDiff < 49) {
if (SC < 3002) {
if (MVDiff < 35) {
if (Rs < 1077) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 8;
}
}
else if (Rs >= 1077) {
if (Rs < 1349) {
return 2;
}
else if (Rs >= 1349) {
if (posBalance < 63) {
return 4;
}
else if (posBalance >= 63) {
return 2;
}
}
}
}
else if (MVDiff >= 35) {
return 2;
}
}
else if (SC >= 3002) {
if (Cs < 1855) {
if (tcor < 75) {
if (mu_mv_mag_sq < 5) {
if (diffAFD < 3) {
if (Cs < 1725) {
return 4;
}
else if (Cs >= 1725) {
if (Cs < 1740) {
return 8;
}
else if (Cs >= 1740) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
return 4;
}
}
}
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 5) {
if (diffTSC < 1) {
if (Cs < 1835) {
return 4;
}
else if (Cs >= 1835) {
return 1;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (tcor >= 75) {
if (mcTcor < 90) {
if (Rs < 2393) {
return 8;
}
else if (Rs >= 2393) {
if (Rs < 2585) {
return 2;
}
else if (Rs >= 2585) {
return 4;
}
}
}
else if (mcTcor >= 90) {
if (posBalance < 56) {
if (Cs < 1625) {
return 2;
}
else if (Cs >= 1625) {
if (Rs < 2278) {
return 2;
}
else if (Rs >= 2278) {
return 4;
}
}
}
else if (posBalance >= 56) {
return 4;
}
}
}
}
else if (Cs >= 1855) {
if (refDCval < 77) {
return 1;
}
else if (refDCval >= 77) {
return 8;
}
}
}
}
else if (MVDiff >= 49) {
if (posBalance < 63) {
if (SC < 2472) {
return 8;
}
else if (SC >= 2472) {
return 4;
}
}
else if (posBalance >= 63) {
if (TSC < 16) {
return 1;
}
else if (TSC >= 16) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 213) {
if (negBalance < 50) {
if (MVDiff < 47) {
return 2;
}
else if (MVDiff >= 47) {
if (AFD < 23) {
return 1;
}
else if (AFD >= 23) {
return 4;
}
}
}
else if (negBalance >= 50) {
if (RsCsDiff < 263) {
if (refDCval < 80) {
if (Rs < 2328) {
if (Rs < 1236) {
return 2;
}
else if (Rs >= 1236) {
return 8;
}
}
else if (Rs >= 2328) {
if (diffAFD < 6) {
return 2;
}
else if (diffAFD >= 6) {
return 4;
}
}
}
else if (refDCval >= 80) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 4;
}
}
}
else if (RsCsDiff >= 263) {
if (AFD < 29) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (AFD >= 29) {
return 2;
}
}
}
}
}
else if (refDCval >= 85) {
if (diffMVdiffVal < -9) {
if (RsCsDiff < 142) {
return 4;
}
else if (RsCsDiff >= 142) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
return 1;
}
}
}
else if (diffMVdiffVal >= -9) {
if (TSC < 13) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
if (Rs < 1379) {
if (ssDCval < 102) {
return 4;
}
else if (ssDCval >= 102) {
if (Rs < 858) {
return 2;
}
else if (Rs >= 858) {
return 1;
}
}
}
else if (Rs >= 1379) {
if (mcTcor < 94) {
if (SC < 4726) {
if (mu_mv_mag_sq < 15) {
if (RsDiff < 14) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (Cs < 1555) {
return 1;
}
else if (Cs >= 1555) {
return 8;
}
}
}
else if (RsDiff >= 14) {
return 4;
}
}
else if (mu_mv_mag_sq >= 15) {
if (AFD < 20) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (AFD >= 20) {
if (Rs < 1549) {
return 8;
}
else if (Rs >= 1549) {
return 4;
}
}
}
}
else if (SC >= 4726) {
return 4;
}
}
else if (mcTcor >= 94) {
if (negBalance < 50) {
if (Rs < 2521) {
return 8;
}
else if (Rs >= 2521) {
return 1;
}
}
else if (negBalance >= 50) {
if (Cs < 1379) {
return 4;
}
else if (Cs >= 1379) {
return 2;
}
}
}
}
}
}
else if (TSC >= 13) {
if (refDCval < 95) {
if (Cs < 1634) {
if (ssDCval < 89) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (Cs >= 1634) {
if (Cs < 1669) {
return 4;
}
else if (Cs >= 1669) {
return 1;
}
}
}
else if (refDCval >= 95) {
if (Cs < 1461) {
return 2;
}
else if (Cs >= 1461) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 15) {
if (Cs < 1753) {
if (gchDC < 7) {
if (mcTcor < 75) {
return 8;
}
else if (mcTcor >= 75) {
if (negBalance < 61) {
if (Rs < 3266) {
return 2;
}
else if (Rs >= 3266) {
return 4;
}
}
else if (negBalance >= 61) {
if (Rs < 3015) {
if (Rs < 2492) {
return 4;
}
else if (Rs >= 2492) {
return 2;
}
}
else if (Rs >= 3015) {
return 8;
}
}
}
}
else if (gchDC >= 7) {
if (AFD < 31) {
return 8;
}
else if (AFD >= 31) {
return 1;
}
}
}
else if (Cs >= 1753) {
if (negBalance < 54) {
if (CsDiff < 12) {
return 2;
}
else if (CsDiff >= 12) {
return 8;
}
}
else if (negBalance >= 54) {
if (AFD < 29) {
return 4;
}
else if (AFD >= 29) {
return 8;
}
}
}
}
}
else if (mcTcor >= 95) {
if (Rs < 995) {
if (Cs < 1842) {
return 8;
}
else if (Cs >= 1842) {
return 2;
}
}
else if (Rs >= 995) {
if (Rs < 2644) {
if (mu_mv_mag_sq < 45) {
if (refDCval < 103) {
return 2;
}
else if (refDCval >= 103) {
if (Rs < 1921) {
return 4;
}
else if (Rs >= 1921) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (Rs < 1650) {
return 4;
}
else if (Rs >= 1650) {
return 1;
}
}
}
else if (Rs >= 2644) {
return 1;
}
}
}
}
}
else if (refDCval >= 105) {
if (Rs < 1733) {
if (negBalance < 56) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
if (Rs < 1490) {
return 2;
}
else if (Rs >= 1490) {
return 8;
}
}
}
}
else if (negBalance >= 56) {
if (posBalance < 54) {
if (AFD < 31) {
if (SC < 2756) {
return 8;
}
else if (SC >= 2756) {
if (Rs < 1439) {
return 4;
}
else if (Rs >= 1439) {
return 8;
}
}
}
else if (AFD >= 31) {
return 1;
}
}
else if (posBalance >= 54) {
return 2;
}
}
}
else if (Rs >= 1733) {
if (mu_mv_mag_sq < 9) {
if (refDCval < 113) {
if (mcTcor < 92) {
if (MVDiff < 16) {
if (Rs < 2119) {
return 8;
}
else if (Rs >= 2119) {
if (Cs < 1732) {
return 4;
}
else if (Cs >= 1732) {
return 8;
}
}
}
else if (MVDiff >= 16) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 2;
}
}
}
else if (mcTcor >= 92) {
if (Cs < 1518) {
return 1;
}
else if (Cs >= 1518) {
if (Rs < 2185) {
return 2;
}
else if (Rs >= 2185) {
return 8;
}
}
}
}
else if (refDCval >= 113) {
if (Rs < 2578) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
if (Rs < 2330) {
return 8;
}
else if (Rs >= 2330) {
return 1;
}
}
}
else if (Rs >= 2578) {
if (TSC < 10) {
if (Rs < 2788) {
return 8;
}
else if (Rs >= 2788) {
return 1;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (TSC < 10) {
if (mcTcor < 89) {
if (Rs < 2648) {
return 4;
}
else if (Rs >= 2648) {
return 8;
}
}
else if (mcTcor >= 89) {
return 2;
}
}
else if (TSC >= 10) {
if (RsDiff < 20) {
if (refDCval < 125) {
if (tcor < 57) {
if (ssDCval < 113) {
if (Rs < 2930) {
return 4;
}
else if (Rs >= 2930) {
return 2;
}
}
else if (ssDCval >= 113) {
return 8;
}
}
else if (tcor >= 57) {
return 8;
}
}
else if (refDCval >= 125) {
if (CsDiff < 7) {
if (Cs < 1460) {
return 2;
}
else if (Cs >= 1460) {
return 4;
}
}
else if (CsDiff >= 7) {
if (mu_mv_mag_sq < 26) {
if (RsCsDiff < 183) {
if (Cs < 1645) {
return 4;
}
else if (Cs >= 1645) {
return 1;
}
}
else if (RsCsDiff >= 183) {
if (negBalance < 61) {
if (mcTcor < 85) {
return 1;
}
else if (mcTcor >= 85) {
if (AFD < 32) {
return 8;
}
else if (AFD >= 32) {
return 1;
}
}
}
else if (negBalance >= 61) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 2;
}
}
}
}
}
else if (RsDiff >= 20) {
if (refDCval < 149) {
return 1;
}
else if (refDCval >= 149) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 3320) {
if (tcor < 85) {
if (mu_mv_mag_sq < 11) {
if (refDCval < 94) {
if (RsDiff < 13) {
if (Rs < 3476) {
if (Rs < 3434) {
return 8;
}
else if (Rs >= 3434) {
return 2;
}
}
else if (Rs >= 3476) {
return 4;
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (refDCval >= 94) {
if (TSC < 14) {
if (RsDiff < 16) {
if (tcor < 79) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (Rs < 4091) {
return 4;
}
else if (Rs >= 4091) {
return 8;
}
}
}
}
else if (tcor >= 79) {
return 8;
}
}
else if (RsDiff >= 16) {
if (MVDiff < 15) {
if (Rs < 4079) {
return 8;
}
else if (Rs >= 4079) {
return 2;
}
}
else if (MVDiff >= 15) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
if (Rs < 5090) {
if (RsCsDiff < 404) {
if (RsCsDiff < 355) {
return 2;
}
else if (RsCsDiff >= 355) {
return 8;
}
}
else if (RsCsDiff >= 404) {
return 2;
}
}
else if (Rs >= 5090) {
if (RsCsDiff < 635) {
return 8;
}
else if (RsCsDiff >= 635) {
if (Cs < 1580) {
return 8;
}
else if (Cs >= 1580) {
return 2;
}
}
}
}
}
}
}
else if (TSC >= 14) {
if (mcTcor < 90) {
if (MVDiff < 69) {
return 2;
}
else if (MVDiff >= 69) {
if (Cs < 1349) {
return 1;
}
else if (Cs >= 1349) {
return 4;
}
}
}
else if (mcTcor >= 90) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (RsDiff < 16) {
if (posBalance < 49) {
if (RsCsDiff < 229) {
if (Rs < 3639) {
if (ssDCval < 72) {
if (Cs < 1636) {
return 2;
}
else if (Cs >= 1636) {
return 8;
}
}
else if (ssDCval >= 72) {
return 4;
}
}
else if (Rs >= 3639) {
return 4;
}
}
else if (RsCsDiff >= 229) {
return 1;
}
}
else if (posBalance >= 49) {
if (AFD < 24) {
if (refDCval < 62) {
return 2;
}
else if (refDCval >= 62) {
if (ssDCval < 137) {
if (Rs < 3344) {
return 4;
}
else if (Rs >= 3344) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 4280) {
return 2;
}
else if (Rs >= 4280) {
return 8;
}
}
}
}
else if (ssDCval >= 137) {
if (Rs < 3945) {
return 2;
}
else if (Rs >= 3945) {
return 4;
}
}
}
}
else if (AFD >= 24) {
if (Rs < 4457) {
return 4;
}
else if (Rs >= 4457) {
return 1;
}
}
}
}
else if (RsDiff >= 16) {
if (negBalance < 56) {
if (SC < 5517) {
if (ssDCval < 118) {
if (mu_mv_mag_sq < 17) {
return 4;
}
else if (mu_mv_mag_sq >= 17) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (ssDCval >= 118) {
if (mu_mv_mag_sq < 22) {
return 1;
}
else if (mu_mv_mag_sq >= 22) {
return 8;
}
}
}
else if (SC >= 5517) {
if (posBalance < 67) {
if (MVDiff < 14) {
if (posBalance < 47) {
if (Cs < 1580) {
return 4;
}
else if (Cs >= 1580) {
return 2;
}
}
else if (posBalance >= 47) {
return 4;
}
}
else if (MVDiff >= 14) {
if (TSC < 15) {
return 4;
}
else if (TSC >= 15) {
if (Rs < 4665) {
return 1;
}
else if (Rs >= 4665) {
return 4;
}
}
}
}
else if (posBalance >= 67) {
if (MVDiff < 21) {
return 1;
}
else if (MVDiff >= 21) {
return 2;
}
}
}
}
else if (negBalance >= 56) {
if (posBalance < 50) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
else if (posBalance >= 50) {
if (MVDiff < 84) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 2;
}
}
else if (MVDiff >= 84) {
if (Cs < 1384) {
return 4;
}
else if (Cs >= 1384) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 85) {
if (refDCval < 101) {
if (posBalance < 41) {
if (negBalance < 37) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (negBalance >= 37) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
if (refDCval < 84) {
if (Rs < 4957) {
return 8;
}
else if (Rs >= 4957) {
return 4;
}
}
else if (refDCval >= 84) {
if (Rs < 4026) {
return 1;
}
else if (Rs >= 4026) {
return 2;
}
}
}
}
}
else if (posBalance >= 41) {
if (mcTcor < 92) {
if (RsDiff < 13) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
if (Rs < 4304) {
return 8;
}
else if (Rs >= 4304) {
return 2;
}
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (mcTcor >= 92) {
if (refDCval < 73) {
if (CsDiff < 9) {
if (Cs < 1475) {
if (Rs < 4213) {
return 2;
}
else if (Rs >= 4213) {
if (Rs < 8599) {
return 4;
}
else if (Rs >= 8599) {
return 2;
}
}
}
else if (Cs >= 1475) {
return 2;
}
}
else if (CsDiff >= 9) {
return 8;
}
}
else if (refDCval >= 73) {
if (CsDiff < 5) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
else if (refDCval >= 101) {
if (RsDiff < 11) {
if (mu_mv_mag_sq < 80) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
return 2;
}
}
else if (mu_mv_mag_sq >= 80) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 1;
}
}
}
else if (RsDiff >= 11) {
if (MVDiff < 17) {
if (SC < 5661) {
if (ssDCval < 105) {
return 2;
}
else if (ssDCval >= 105) {
if (Cs < 1432) {
return 4;
}
else if (Cs >= 1432) {
return 8;
}
}
}
else if (SC >= 5661) {
if (Cs < 1317) {
if (Rs < 5723) {
return 1;
}
else if (Rs >= 5723) {
return 2;
}
}
else if (Cs >= 1317) {
return 2;
}
}
}
else if (MVDiff >= 17) {
if (ssDCval < 114) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
if (Rs < 5487) {
return 2;
}
else if (Rs >= 5487) {
return 1;
}
}
}
else if (ssDCval >= 114) {
if (refDCval < 143) {
if (AFD < 22) {
if (MVDiff < 86) {
return 8;
}
else if (MVDiff >= 86) {
return 1;
}
}
else if (AFD >= 22) {
return 1;
}
}
else if (refDCval >= 143) {
if (Cs < 1493) {
return 1;
}
else if (Cs >= 1493) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Cs >= 1897) {
if (AFD < 12) {
if (ssDCval < 101) {
if (Rs < 4408) {
if (Cs < 2728) {
if (ssDCval < 73) {
if (negBalance < 39) {
if (MVDiff < 4) {
if (Rs < 3906) {
if (Cs < 1948) {
return 8;
}
else if (Cs >= 1948) {
return 1;
}
}
else if (Rs >= 3906) {
return 4;
}
}
else if (MVDiff >= 4) {
if (Cs < 1987) {
return 1;
}
else if (Cs >= 1987) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
}
}
else if (negBalance >= 39) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (AFD < 10) {
if (mcTcor < 87) {
if (Rs < 2359) {
return 1;
}
else if (Rs >= 2359) {
return 8;
}
}
else if (mcTcor >= 87) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (AFD >= 10) {
return 8;
}
}
}
}
else if (ssDCval >= 73) {
if (SC < 5602) {
if (TSC < 6) {
if (Rs < 2492) {
return 1;
}
else if (Rs >= 2492) {
return 4;
}
}
else if (TSC >= 6) {
if (Cs < 2052) {
if (Cs < 2042) {
if (posBalance < 62) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (Rs < 2742) {
return 8;
}
else if (Rs >= 2742) {
return 4;
}
}
}
else if (posBalance >= 62) {
return 4;
}
}
else if (Cs >= 2042) {
return 4;
}
}
else if (Cs >= 2052) {
return 8;
}
}
}
else if (SC >= 5602) {
if (tcor < 91) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 4;
}
}
else if (tcor >= 91) {
if (Cs < 2436) {
if (RsCsDiff < 115) {
if (Rs < 4034) {
return 2;
}
else if (Rs >= 4034) {
return 8;
}
}
else if (RsCsDiff >= 115) {
if (Rs < 4398) {
return 2;
}
else if (Rs >= 4398) {
return 1;
}
}
}
else if (Cs >= 2436) {
return 4;
}
}
}
}
}
else if (Cs >= 2728) {
if (negBalance < 51) {
if (Rs < 1582) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (Rs >= 1582) {
if (posBalance < 36) {
if (Rs < 3111) {
return 2;
}
else if (Rs >= 3111) {
return 8;
}
}
else if (posBalance >= 36) {
if (Cs < 4358) {
if (Cs < 2965) {
if (RsDiff < 8) {
if (AFD < 10) {
if (RsCsDiff < 124) {
return 4;
}
else if (RsCsDiff >= 124) {
return 1;
}
}
else if (AFD >= 10) {
if (Rs < 3010) {
return 8;
}
else if (Rs >= 3010) {
return 4;
}
}
}
else if (RsDiff >= 8) {
if (Cs < 2956) {
return 4;
}
else if (Cs >= 2956) {
return 8;
}
}
}
else if (Cs >= 2965) {
if (Rs < 4222) {
if (RsDiff < 7) {
if (Rs < 3415) {
return 4;
}
else if (Rs >= 3415) {
return 1;
}
}
else if (RsDiff >= 7) {
return 4;
}
}
else if (Rs >= 4222) {
return 1;
}
}
}
else if (Cs >= 4358) {
return 8;
}
}
}
}
else if (negBalance >= 51) {
if (ssDCval < 85) {
if (Rs < 2932) {
return 4;
}
else if (Rs >= 2932) {
return 8;
}
}
else if (ssDCval >= 85) {
return 4;
}
}
}
}
else if (Rs >= 4408) {
if (ssDCval < 73) {
if (Cs < 3402) {
if (RsDiff < 23) {
return 4;
}
else if (RsDiff >= 23) {
return 8;
}
}
else if (Cs >= 3402) {
if (Rs < 5854) {
return 8;
}
else if (Rs >= 5854) {
return 2;
}
}
}
else if (ssDCval >= 73) {
if (tcor < 89) {
if (Rs < 5348) {
return 8;
}
else if (Rs >= 5348) {
if (Rs < 6194) {
return 4;
}
else if (Rs >= 6194) {
return 2;
}
}
}
else if (tcor >= 89) {
if (RsDiff < 9) {
if (Rs < 5764) {
return 1;
}
else if (Rs >= 5764) {
return 2;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
}
}
}
else if (ssDCval >= 101) {
if (negBalance < 46) {
if (negBalance < 32) {
if (ssDCval < 113) {
if (SC < 9638) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
else if (SC >= 9638) {
return 1;
}
}
else if (ssDCval >= 113) {
if (mu_mv_mag_sq < 62) {
if (MVDiff < 33) {
if (negBalance < 27) {
if (Rs < 4895) {
return 4;
}
else if (Rs >= 4895) {
return 8;
}
}
else if (negBalance >= 27) {
return 1;
}
}
else if (MVDiff >= 33) {
return 2;
}
}
else if (mu_mv_mag_sq >= 62) {
if (mu_mv_mag_sq < 93) {
if (SC < 7185) {
if (Cs < 2593) {
return 8;
}
else if (Cs >= 2593) {
if (Rs < 2837) {
return 1;
}
else if (Rs >= 2837) {
return 2;
}
}
}
else if (SC >= 7185) {
return 8;
}
}
else if (mu_mv_mag_sq >= 93) {
return 8;
}
}
}
}
else if (negBalance >= 32) {
if (Rs < 6187) {
if (negBalance < 44) {
if (diffAFD < -3) {
if (Cs < 2495) {
if (Cs < 2157) {
return 1;
}
else if (Cs >= 2157) {
return 8;
}
}
else if (Cs >= 2495) {
return 2;
}
}
else if (diffAFD >= -3) {
if (ssDCval < 103) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
return 2;
}
}
else if (ssDCval >= 103) {
return 8;
}
}
}
else if (negBalance >= 44) {
if (mcTcor < 93) {
if (tcor < 89) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (tcor >= 89) {
if (Rs < 3899) {
if (Rs < 3258) {
return 8;
}
else if (Rs >= 3258) {
return 4;
}
}
else if (Rs >= 3899) {
return 1;
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 106) {
if (Cs < 2747) {
return 8;
}
else if (Cs >= 2747) {
return 4;
}
}
else if (ssDCval >= 106) {
return 8;
}
}
}
}
else if (Rs >= 6187) {
if (mu_mv_mag_sq < 77) {
if (Cs < 2745) {
return 2;
}
else if (Cs >= 2745) {
return 1;
}
}
else if (mu_mv_mag_sq >= 77) {
if (Cs < 4110) {
if (Rs < 7337) {
return 4;
}
else if (Rs >= 7337) {
return 2;
}
}
else if (Cs >= 4110) {
return 8;
}
}
}
}
}
else if (negBalance >= 46) {
if (tcor < 93) {
if (mu_mv_mag_sq < 83) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 4249) {
if (Rs < 3175) {
return 8;
}
else if (Rs >= 3175) {
return 2;
}
}
else if (Rs >= 4249) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (Cs < 3533) {
if (diffAFD < 3) {
if (diffRsCsdiff < 31) {
return 1;
}
else if (diffRsCsdiff >= 31) {
if (diffAFD < -1) {
if (Rs < 4356) {
return 1;
}
else if (Rs >= 4356) {
return 8;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (diffAFD >= 3) {
if (Rs < 3272) {
return 8;
}
else if (Rs >= 3272) {
return 4;
}
}
}
else if (Cs >= 3533) {
return 8;
}
}
}
else if (tcor >= 93) {
if (mcTcor < 94) {
if (Rs < 3929) {
return 4;
}
else if (Rs >= 3929) {
return 1;
}
}
else if (mcTcor >= 94) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 76) {
if (RsDiff < 17) {
if (ssDCval < 101) {
if (tcor < 64) {
if (refDCval < 66) {
if (mcTcor < 89) {
if (RsCsDiff < 290) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
if (TSC < 14) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
if (Cs < 3478) {
return 2;
}
else if (Cs >= 3478) {
return 1;
}
}
}
else if (TSC >= 14) {
return 4;
}
}
}
else if (RsCsDiff >= 290) {
if (ssDCval < 48) {
return 2;
}
else if (ssDCval >= 48) {
if (mcTcor < 88) {
return 4;
}
else if (mcTcor >= 88) {
return 2;
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1912) {
return 2;
}
else if (Rs >= 1912) {
return 1;
}
}
}
else if (refDCval >= 66) {
if (RsDiff < 12) {
if (gchDC < 3) {
if (SC < 5869) {
if (Cs < 1977) {
if (Rs < 2808) {
return 1;
}
else if (Rs >= 2808) {
return 8;
}
}
else if (Cs >= 1977) {
if (tcor < 45) {
if (Rs < 2047) {
return 8;
}
else if (Rs >= 2047) {
return 1;
}
}
else if (tcor >= 45) {
if (diffAFD < 4) {
if (negBalance < 54) {
if (Rs < 1509) {
return 4;
}
else if (Rs >= 1509) {
return 8;
}
}
else if (negBalance >= 54) {
if (ssDCval < 75) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 4;
}
}
else if (ssDCval >= 75) {
return 4;
}
}
}
else if (diffAFD >= 4) {
if (Rs < 1276) {
return 1;
}
else if (Rs >= 1276) {
return 2;
}
}
}
}
}
else if (SC >= 5869) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (posBalance < 68) {
if (Rs < 1223) {
return 8;
}
else if (Rs >= 1223) {
if (Rs < 2305) {
return 2;
}
else if (Rs >= 2305) {
return 1;
}
}
}
else if (posBalance >= 68) {
if (Cs < 2001) {
return 4;
}
else if (Cs >= 2001) {
return 8;
}
}
}
}
else if (RsDiff >= 12) {
if (SC < 5283) {
if (ssDCval < 82) {
if (TSC < 15) {
return 1;
}
else if (TSC >= 15) {
if (Rs < 2602) {
return 4;
}
else if (Rs >= 2602) {
return 8;
}
}
}
else if (ssDCval >= 82) {
return 8;
}
}
else if (SC >= 5283) {
if (Cs < 2976) {
return 8;
}
else if (Cs >= 2976) {
if (AFD < 32) {
return 4;
}
else if (AFD >= 32) {
return 8;
}
}
}
}
}
}
else if (tcor >= 64) {
if (negBalance < 49) {
if (posBalance < 46) {
if (ssDCval < 73) {
if (Rs < 3303) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
if (Rs < 2988) {
return 2;
}
else if (Rs >= 2988) {
return 4;
}
}
}
else if (Rs >= 3303) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (Rs < 5051) {
return 2;
}
else if (Rs >= 5051) {
return 4;
}
}
}
}
else if (ssDCval >= 73) {
if (refDCval < 94) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
if (Cs < 2127) {
return 2;
}
else if (Cs >= 2127) {
return 8;
}
}
}
else if (refDCval >= 94) {
return 1;
}
}
}
else if (posBalance >= 46) {
if (posBalance < 56) {
if (SC < 4742) {
if (Rs < 1849) {
if (Cs < 2465) {
if (Rs < 1539) {
return 8;
}
else if (Rs >= 1539) {
return 2;
}
}
else if (Cs >= 2465) {
return 4;
}
}
else if (Rs >= 1849) {
if (mu_mv_mag_sq < 62) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 8;
}
}
else if (mu_mv_mag_sq >= 62) {
if (Rs < 2083) {
return 2;
}
else if (Rs >= 2083) {
if (Rs < 2222) {
return 8;
}
else if (Rs >= 2222) {
return 1;
}
}
}
}
}
else if (SC >= 4742) {
if (refDCval < 88) {
if (RsCsDiff < 172) {
if (Cs < 2016) {
return 2;
}
else if (Cs >= 2016) {
return 8;
}
}
else if (RsCsDiff >= 172) {
if (mu_mv_mag_sq < 54) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (MVDiff < 12) {
if (Rs < 3137) {
return 8;
}
else if (Rs >= 3137) {
if (Rs < 5002) {
return 2;
}
else if (Rs >= 5002) {
return 8;
}
}
}
else if (MVDiff >= 12) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 3373) {
return 8;
}
else if (Rs >= 3373) {
if (Cs < 2956) {
return 4;
}
else if (Cs >= 2956) {
return 2;
}
}
}
}
}
else if (refDCval >= 88) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (posBalance < 53) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (MVDiff < 27) {
if (ssDCval < 88) {
return 2;
}
else if (ssDCval >= 88) {
if (ssDCval < 100) {
return 1;
}
else if (ssDCval >= 100) {
return 8;
}
}
}
else if (MVDiff >= 27) {
return 8;
}
}
}
else if (posBalance >= 53) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 56) {
if (TSC < 9) {
if (SC < 6499) {
if (tcor < 85) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (tcor >= 85) {
if (RsCsDiff < 162) {
return 1;
}
else if (RsCsDiff >= 162) {
return 2;
}
}
}
else if (SC >= 6499) {
if (Cs < 2030) {
return 2;
}
else if (Cs >= 2030) {
return 8;
}
}
}
else if (TSC >= 9) {
if (CsDiff < 9) {
if (MVDiff < 14) {
if (RsDiff < 10) {
if (gchDC < 2) {
if (Cs < 2924) {
return 8;
}
else if (Cs >= 2924) {
return 4;
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (RsDiff >= 10) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
return 2;
}
}
}
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (CsDiff >= 9) {
if (TSC < 13) {
if (refDCval < 98) {
if (mcTcor < 87) {
if (RsCsDiff < 223) {
if (tcor < 80) {
if (CsDiff < 10) {
if (Rs < 2782) {
return 8;
}
else if (Rs >= 2782) {
return 4;
}
}
else if (CsDiff >= 10) {
return 4;
}
}
else if (tcor >= 80) {
return 8;
}
}
else if (RsCsDiff >= 223) {
if (mcTcor < 85) {
if (Rs < 2503) {
return 4;
}
else if (Rs >= 2503) {
return 8;
}
}
else if (mcTcor >= 85) {
return 2;
}
}
}
else if (mcTcor >= 87) {
if (Cs < 2570) {
if (mu_mv_mag_sq < 5) {
if (Rs < 3976) {
return 4;
}
else if (Rs >= 3976) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
if (CsDiff < 12) {
if (gchDC < 1) {
if (Rs < 2701) {
return 8;
}
else if (Rs >= 2701) {
return 4;
}
}
else if (gchDC >= 1) {
if (MVDiff < 31) {
if (Rs < 4629) {
return 8;
}
else if (Rs >= 4629) {
return 2;
}
}
else if (MVDiff >= 31) {
return 4;
}
}
}
else if (CsDiff >= 12) {
if (Cs < 2075) {
return 2;
}
else if (Cs >= 2075) {
return 4;
}
}
}
}
else if (Cs >= 2570) {
if (Cs < 4199) {
if (ssDCval < 79) {
if (refDCval < 78) {
return 4;
}
else if (refDCval >= 78) {
return 1;
}
}
else if (ssDCval >= 79) {
return 4;
}
}
else if (Cs >= 4199) {
return 8;
}
}
}
}
else if (refDCval >= 98) {
if (Cs < 2710) {
return 2;
}
else if (Cs >= 2710) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 2;
}
}
}
}
else if (TSC >= 13) {
if (Rs < 2762) {
return 1;
}
else if (Rs >= 2762) {
if (CsDiff < 12) {
return 8;
}
else if (CsDiff >= 12) {
return 4;
}
}
}
}
}
}
}
}
else if (negBalance >= 49) {
if (mu_mv_mag_sq < 9) {
if (refDCval < 86) {
if (MVDiff < 4) {
if (AFD < 16) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (AFD >= 16) {
if (Rs < 3401) {
return 1;
}
else if (Rs >= 3401) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (gchDC < 2) {
if (RsCsDiff < 201) {
if (diffAFD < -3) {
if (Cs < 3862) {
return 4;
}
else if (Cs >= 3862) {
return 8;
}
}
else if (diffAFD >= -3) {
if (Rs < 1708) {
if (SC < 3601) {
return 8;
}
else if (SC >= 3601) {
return 4;
}
}
else if (Rs >= 1708) {
if (negBalance < 50) {
if (tcor < 83) {
if (Cs < 2181) {
return 4;
}
else if (Cs >= 2181) {
return 8;
}
}
else if (tcor >= 83) {
return 8;
}
}
else if (negBalance >= 50) {
if (MVDiff < 6) {
if (Rs < 2252) {
return 1;
}
else if (Rs >= 2252) {
return 8;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 201) {
if (Cs < 2241) {
if (Rs < 2972) {
if (Rs < 2525) {
return 8;
}
else if (Rs >= 2525) {
return 2;
}
}
else if (Rs >= 2972) {
return 4;
}
}
else if (Cs >= 2241) {
if (TSC < 11) {
if (posBalance < 57) {
if (Cs < 2648) {
return 8;
}
else if (Cs >= 2648) {
return 4;
}
}
else if (posBalance >= 57) {
if (tcor < 75) {
if (AFD < 23) {
return 1;
}
else if (AFD >= 23) {
return 4;
}
}
else if (tcor >= 75) {
if (Rs < 3515) {
return 8;
}
else if (Rs >= 3515) {
return 2;
}
}
}
}
else if (TSC >= 11) {
if (ssDCval < 72) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
else if (ssDCval >= 72) {
if (MVDiff < 8) {
if (Rs < 3733) {
return 1;
}
else if (Rs >= 3733) {
return 2;
}
}
else if (MVDiff >= 8) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (diffRsCsdiff < 33) {
if (Rs < 2826) {
return 8;
}
else if (Rs >= 2826) {
return 1;
}
}
else if (diffRsCsdiff >= 33) {
return 8;
}
}
}
}
}
}
}
}
else if (gchDC >= 2) {
if (MVDiff < 21) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
if (TSC < 13) {
return 1;
}
else if (TSC >= 13) {
return 4;
}
}
}
else if (MVDiff >= 21) {
if (ssDCval < 74) {
if (diffMVdiffVal < 6) {
return 2;
}
else if (diffMVdiffVal >= 6) {
if (Rs < 1638) {
return 4;
}
else if (Rs >= 1638) {
return 1;
}
}
}
else if (ssDCval >= 74) {
if (Rs < 2036) {
return 8;
}
else if (Rs >= 2036) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 86) {
if (SC < 8269) {
if (mu_mv_mag_sq < 3) {
if (Cs < 2221) {
if (mcTcor < 94) {
if (MVDiff < 14) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
if (Rs < 2458) {
return 4;
}
else if (Rs >= 2458) {
return 8;
}
}
}
else if (MVDiff >= 14) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (Cs >= 2221) {
if (RsCsDiff < 258) {
if (mcTcor < 82) {
return 4;
}
else if (mcTcor >= 82) {
if (tcor < 75) {
if (negBalance < 58) {
return 8;
}
else if (negBalance >= 58) {
if (AFD < 22) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (Cs < 3506) {
return 4;
}
else if (Cs >= 3506) {
return 8;
}
}
}
else if (AFD >= 22) {
return 8;
}
}
}
else if (tcor >= 75) {
if (negBalance < 58) {
if (Cs < 3408) {
if (Cs < 2694) {
if (ssDCval < 89) {
if (Cs < 2645) {
return 4;
}
else if (Cs >= 2645) {
return 8;
}
}
else if (ssDCval >= 89) {
if (SC < 4183) {
return 4;
}
else if (SC >= 4183) {
if (RsCsDiff < 226) {
return 8;
}
else if (RsCsDiff >= 226) {
return 2;
}
}
}
}
else if (Cs >= 2694) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
return 4;
}
}
}
else if (Cs >= 3408) {
if (Rs < 3482) {
return 8;
}
else if (Rs >= 3482) {
return 2;
}
}
}
else if (negBalance >= 58) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 258) {
if (posBalance < 62) {
if (Rs < 4628) {
if (mcTcor < 83) {
return 8;
}
else if (mcTcor >= 83) {
if (Cs < 2298) {
return 2;
}
else if (Cs >= 2298) {
if (CsDiff < 13) {
return 4;
}
else if (CsDiff >= 13) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
if (Rs < 3850) {
return 8;
}
else if (Rs >= 3850) {
return 4;
}
}
}
}
}
}
else if (Rs >= 4628) {
return 8;
}
}
else if (posBalance >= 62) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 15) {
if (RsCsDiff < 258) {
if (ssDCval < 99) {
if (negBalance < 58) {
return 8;
}
else if (negBalance >= 58) {
if (Rs < 2478) {
return 1;
}
else if (Rs >= 2478) {
return 8;
}
}
}
else if (ssDCval >= 99) {
return 2;
}
}
else if (RsCsDiff >= 258) {
if (Rs < 4692) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 2562) {
return 4;
}
else if (Rs >= 2562) {
if (Rs < 2848) {
return 8;
}
else if (Rs >= 2848) {
return 2;
}
}
}
}
else if (Rs >= 4692) {
return 8;
}
}
}
else if (TSC >= 15) {
if (Rs < 1991) {
return 8;
}
else if (Rs >= 1991) {
return 1;
}
}
}
}
else if (SC >= 8269) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (TSC < 18) {
if (ssDCval < 97) {
if (SC < 6550) {
if (Rs < 3093) {
if (SC < 3266) {
if (Rs < 794) {
return 2;
}
else if (Rs >= 794) {
return 4;
}
}
else if (SC >= 3266) {
if (mcTcor < 77) {
if (MVDiff < 12) {
if (Rs < 2358) {
return 8;
}
else if (Rs >= 2358) {
return 4;
}
}
else if (MVDiff >= 12) {
if (Cs < 2229) {
return 1;
}
else if (Cs >= 2229) {
return 2;
}
}
}
else if (mcTcor >= 77) {
if (mcTcor < 94) {
if (ssDCval < 85) {
if (SC < 4072) {
if (Cs < 2350) {
if (Rs < 1915) {
return 4;
}
else if (Rs >= 1915) {
return 1;
}
}
else if (Cs >= 2350) {
return 8;
}
}
else if (SC >= 4072) {
if (posBalance < 58) {
if (Rs < 2714) {
if (CsDiff < 12) {
return 8;
}
else if (CsDiff >= 12) {
if (Rs < 2214) {
return 1;
}
else if (Rs >= 2214) {
if (Rs < 2605) {
return 8;
}
else if (Rs >= 2605) {
return 1;
}
}
}
}
else if (Rs >= 2714) {
if (Rs < 3001) {
return 4;
}
else if (Rs >= 3001) {
return 1;
}
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
else if (ssDCval >= 85) {
if (Cs < 2877) {
if (TSC < 14) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 4;
}
}
else if (TSC >= 14) {
if (posBalance < 61) {
if (Rs < 2223) {
return 1;
}
else if (Rs >= 2223) {
return 8;
}
}
else if (posBalance >= 61) {
return 4;
}
}
}
else if (Cs >= 2877) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (diffAFD < -1) {
if (Rs < 2583) {
return 2;
}
else if (Rs >= 2583) {
return 8;
}
}
else if (diffAFD >= -1) {
if (Rs < 964) {
return 2;
}
else if (Rs >= 964) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3093) {
if (Cs < 2741) {
if (refDCval < 88) {
if (gchDC < 6) {
if (refDCval < 81) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
if (Rs < 3540) {
return 4;
}
else if (Rs >= 3540) {
if (Cs < 2472) {
return 8;
}
else if (Cs >= 2472) {
return 4;
}
}
}
}
}
else if (refDCval >= 81) {
return 4;
}
}
else if (gchDC >= 6) {
return 8;
}
}
else if (refDCval >= 88) {
if (mu_mv_mag_sq < 33) {
return 4;
}
else if (mu_mv_mag_sq >= 33) {
return 8;
}
}
}
else if (Cs >= 2741) {
return 8;
}
}
}
else if (SC >= 6550) {
if (diffMVdiffVal < -11) {
if (Rs < 5255) {
return 8;
}
else if (Rs >= 5255) {
return 2;
}
}
else if (diffMVdiffVal >= -11) {
if (posBalance < 57) {
if (mcTcor < 88) {
if (AFD < 17) {
if (Rs < 4354) {
return 1;
}
else if (Rs >= 4354) {
return 4;
}
}
else if (AFD >= 17) {
return 8;
}
}
else if (mcTcor >= 88) {
return 8;
}
}
else if (posBalance >= 57) {
if (diffRsCsdiff < 111) {
return 8;
}
else if (diffRsCsdiff >= 111) {
if (Rs < 4128) {
return 2;
}
else if (Rs >= 4128) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 97) {
if (mu_mv_mag_sq < 59) {
if (Cs < 5151) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Rs < 3879) {
return 2;
}
else if (Rs >= 3879) {
if (Rs < 4829) {
return 4;
}
else if (Rs >= 4829) {
return 8;
}
}
}
}
else if (Cs >= 5151) {
return 4;
}
}
else if (mu_mv_mag_sq >= 59) {
return 4;
}
}
}
else if (TSC >= 18) {
if (Rs < 3155) {
return 4;
}
else if (Rs >= 3155) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 101) {
if (AFD < 20) {
if (ssDCval < 119) {
if (mcTcor < 87) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (Cs < 1990) {
return 4;
}
else if (Cs >= 1990) {
return 8;
}
}
}
else if (mcTcor >= 87) {
if (RsCsDiff < 229) {
if (Rs < 3526) {
if (SC < 5577) {
if (negBalance < 43) {
if (SC < 4461) {
return 1;
}
else if (SC >= 4461) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (negBalance >= 43) {
if (MVDiff < 9) {
if (Rs < 2854) {
if (TSC < 10) {
if (RsCsDiff < 115) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (RsCsDiff >= 115) {
return 8;
}
}
else if (TSC >= 10) {
if (mcTcor < 91) {
if (Cs < 2053) {
return 8;
}
else if (Cs >= 2053) {
return 4;
}
}
else if (mcTcor >= 91) {
return 8;
}
}
}
else if (Rs >= 2854) {
if (Rs < 3248) {
return 4;
}
else if (Rs >= 3248) {
return 8;
}
}
}
else if (MVDiff >= 9) {
if (diffAFD < -5) {
if (Rs < 2195) {
return 4;
}
else if (Rs >= 2195) {
return 8;
}
}
else if (diffAFD >= -5) {
return 8;
}
}
}
}
else if (SC >= 5577) {
if (posBalance < 65) {
return 4;
}
else if (posBalance >= 65) {
return 2;
}
}
}
else if (Rs >= 3526) {
if (Rs < 5089) {
if (mu_mv_mag_sq < 74) {
return 8;
}
else if (mu_mv_mag_sq >= 74) {
if (Rs < 4055) {
return 8;
}
else if (Rs >= 4055) {
return 1;
}
}
}
else if (Rs >= 5089) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
else if (RsCsDiff >= 229) {
if (tcor < 76) {
return 8;
}
else if (tcor >= 76) {
if (MVDiff < 6) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
else if (MVDiff >= 6) {
if (refDCval < 104) {
if (Cs < 1968) {
return 8;
}
else if (Cs >= 1968) {
return 2;
}
}
else if (refDCval >= 104) {
if (AFD < 15) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
else if (AFD >= 15) {
if (MVDiff < 15) {
if (diffMVdiffVal < -3) {
if (Rs < 2132) {
return 1;
}
else if (Rs >= 2132) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (MVDiff >= 15) {
if (Rs < 2246) {
return 2;
}
else if (Rs >= 2246) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 119) {
if (mcTcor < 84) {
if (Rs < 3917) {
if (tcor < 63) {
if (Rs < 2828) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
return 2;
}
}
else if (Rs >= 2828) {
return 8;
}
}
else if (tcor >= 63) {
if (Cs < 3912) {
return 8;
}
else if (Cs >= 3912) {
return 1;
}
}
}
else if (Rs >= 3917) {
if (mu_mv_mag_sq < 15) {
return 8;
}
else if (mu_mv_mag_sq >= 15) {
if (negBalance < 45) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (Rs < 4272) {
return 8;
}
else if (Rs >= 4272) {
return 1;
}
}
}
else if (negBalance >= 45) {
return 1;
}
}
}
}
else if (mcTcor >= 84) {
if (RsCsDiff < 161) {
if (negBalance < 40) {
if (Cs < 2724) {
return 1;
}
else if (Cs >= 2724) {
return 8;
}
}
else if (negBalance >= 40) {
if (negBalance < 52) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (RsCsDiff < 123) {
return 4;
}
else if (RsCsDiff >= 123) {
return 8;
}
}
}
else if (negBalance >= 52) {
if (ssDCval < 126) {
return 1;
}
else if (ssDCval >= 126) {
return 8;
}
}
}
}
else if (RsCsDiff >= 161) {
if (SC < 5223) {
if (posBalance < 53) {
if (negBalance < 46) {
if (ssDCval < 136) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 2;
}
}
else if (ssDCval >= 136) {
return 8;
}
}
else if (negBalance >= 46) {
return 8;
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (SC >= 5223) {
if (Cs < 3651) {
if (MVDiff < 28) {
if (Cs < 2554) {
if (ssDCval < 140) {
if (Rs < 3187) {
if (Rs < 3016) {
return 1;
}
else if (Rs >= 3016) {
return 4;
}
}
else if (Rs >= 3187) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (ssDCval >= 140) {
if (negBalance < 49) {
if (Cs < 2177) {
return 4;
}
else if (Cs >= 2177) {
if (Rs < 5302) {
return 8;
}
else if (Rs >= 5302) {
return 2;
}
}
}
else if (negBalance >= 49) {
return 4;
}
}
}
else if (Cs >= 2554) {
if (mu_mv_mag_sq < 68) {
return 4;
}
else if (mu_mv_mag_sq >= 68) {
if (diffAFD < 0) {
return 4;
}
else if (diffAFD >= 0) {
return 8;
}
}
}
}
else if (MVDiff >= 28) {
if (Cs < 3390) {
if (Rs < 3390) {
return 4;
}
else if (Rs >= 3390) {
return 8;
}
}
else if (Cs >= 3390) {
return 2;
}
}
}
else if (Cs >= 3651) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
if (Rs < 3539) {
return 8;
}
else if (Rs >= 3539) {
return 1;
}
}
}
}
}
}
}
}
else if (AFD >= 20) {
if (negBalance < 57) {
if (CsDiff < 16) {
if (ssDCval < 157) {
if (ssDCval < 109) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
if (mcTcor < 93) {
if (CsDiff < 11) {
if (Rs < 3058) {
return 2;
}
else if (Rs >= 3058) {
if (Rs < 4052) {
return 4;
}
else if (Rs >= 4052) {
return 1;
}
}
}
else if (CsDiff >= 11) {
return 4;
}
}
else if (mcTcor >= 93) {
if (diffMVdiffVal < -6) {
if (Rs < 1984) {
return 1;
}
else if (Rs >= 1984) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
return 2;
}
}
}
}
else if (ssDCval >= 109) {
if (SC < 6547) {
if (diffTSC < -1) {
if (Rs < 2327) {
return 2;
}
else if (Rs >= 2327) {
if (Cs < 2152) {
if (Rs < 2897) {
return 2;
}
else if (Rs >= 2897) {
return 8;
}
}
else if (Cs >= 2152) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (RsCsDiff < 423) {
return 8;
}
else if (RsCsDiff >= 423) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 2;
}
}
}
}
else if (SC >= 6547) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
if (ssDCval < 119) {
return 1;
}
else if (ssDCval >= 119) {
return 8;
}
}
}
}
}
else if (ssDCval >= 157) {
if (Rs < 1912) {
return 1;
}
else if (Rs >= 1912) {
return 4;
}
}
}
else if (CsDiff >= 16) {
if (refDCval < 140) {
if (mcTcor < 87) {
if (Cs < 5269) {
return 8;
}
else if (Cs >= 5269) {
return 2;
}
}
else if (mcTcor >= 87) {
return 2;
}
}
else if (refDCval >= 140) {
return 8;
}
}
}
else if (negBalance >= 57) {
if (RsDiff < 11) {
if (tcor < 76) {
if (ssDCval < 117) {
if (diffMVdiffVal < 1) {
if (CsDiff < 13) {
return 1;
}
else if (CsDiff >= 13) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
if (Rs < 2332) {
return 8;
}
else if (Rs >= 2332) {
return 2;
}
}
}
else if (ssDCval >= 117) {
return 2;
}
}
else if (tcor >= 76) {
if (Rs < 2006) {
return 1;
}
else if (Rs >= 2006) {
if (Rs < 2431) {
return 8;
}
else if (Rs >= 2431) {
if (Rs < 2788) {
return 2;
}
else if (Rs >= 2788) {
return 4;
}
}
}
}
}
else if (RsDiff >= 11) {
if (Rs < 3109) {
if (MVDiff < 22) {
if (Rs < 2546) {
if (Cs < 2323) {
return 2;
}
else if (Cs >= 2323) {
return 1;
}
}
else if (Rs >= 2546) {
if (Rs < 2905) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (Rs >= 2905) {
return 8;
}
}
}
else if (MVDiff >= 22) {
return 8;
}
}
else if (Rs >= 3109) {
if (ssDCval < 134) {
if (diffAFD < 2) {
if (diffAFD < -1) {
if (tcor < 74) {
if (Cs < 3200) {
return 1;
}
else if (Cs >= 3200) {
return 4;
}
}
else if (tcor >= 74) {
if (Rs < 3237) {
return 4;
}
else if (Rs >= 3237) {
return 8;
}
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (diffAFD >= 2) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 8;
}
}
}
else if (ssDCval >= 134) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 8;
}
}
}
}
}
}
}
}
else if (RsDiff >= 17) {
if (TSC < 8) {
if (ssDCval < 127) {
if (Rs < 8171) {
if (Cs < 2632) {
return 4;
}
else if (Cs >= 2632) {
return 8;
}
}
else if (Rs >= 8171) {
return 1;
}
}
else if (ssDCval >= 127) {
if (Cs < 2476) {
return 2;
}
else if (Cs >= 2476) {
return 8;
}
}
}
else if (TSC >= 8) {
if (refDCval < 140) {
if (refDCval < 111) {
if (Cs < 4565) {
if (TSC < 15) {
if (tcor < 79) {
if (diffRsCsdiff < 149) {
if (SC < 7541) {
if (AFD < 28) {
if (MVDiff < 17) {
if (Rs < 3723) {
if (ssDCval < 107) {
if (tcor < 75) {
return 4;
}
else if (tcor >= 75) {
return 8;
}
}
else if (ssDCval >= 107) {
if (Rs < 3142) {
return 2;
}
else if (Rs >= 3142) {
return 8;
}
}
}
else if (Rs >= 3723) {
if (Rs < 4588) {
return 2;
}
else if (Rs >= 4588) {
return 4;
}
}
}
else if (MVDiff >= 17) {
if (gchDC < 2) {
if (Rs < 4653) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
return 2;
}
}
else if (Rs >= 4653) {
return 8;
}
}
else if (gchDC >= 2) {
if (Rs < 4199) {
return 8;
}
else if (Rs >= 4199) {
return 4;
}
}
}
}
else if (AFD >= 28) {
if (Cs < 3194) {
return 1;
}
else if (Cs >= 3194) {
return 2;
}
}
}
else if (SC >= 7541) {
if (Cs < 3041) {
if (Cs < 2938) {
if (SC < 7738) {
if (Rs < 4947) {
return 4;
}
else if (Rs >= 4947) {
return 2;
}
}
else if (SC >= 7738) {
return 4;
}
}
else if (Cs >= 2938) {
return 2;
}
}
else if (Cs >= 3041) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (gchDC < 1) {
if (Cs < 3688) {
return 4;
}
else if (Cs >= 3688) {
if (Rs < 6604) {
return 8;
}
else if (Rs >= 6604) {
return 4;
}
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
}
else if (diffRsCsdiff >= 149) {
if (mu_mv_mag_sq < 13) {
if (diffMVdiffVal < -9) {
return 8;
}
else if (diffMVdiffVal >= -9) {
if (Rs < 2807) {
return 4;
}
else if (Rs >= 2807) {
if (Rs < 5348) {
return 2;
}
else if (Rs >= 5348) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (CsDiff < 15) {
return 8;
}
else if (CsDiff >= 15) {
if (RsDiff < 26) {
return 4;
}
else if (RsDiff >= 26) {
if (Rs < 5469) {
return 1;
}
else if (Rs >= 5469) {
return 8;
}
}
}
}
}
}
else if (tcor >= 79) {
if (Rs < 5113) {
if (MVDiff < 23) {
if (refDCval < 81) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (Rs < 3934) {
return 8;
}
else if (Rs >= 3934) {
if (Rs < 5071) {
return 2;
}
else if (Rs >= 5071) {
return 8;
}
}
}
}
else if (refDCval >= 81) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (Rs < 4025) {
return 8;
}
else if (Rs >= 4025) {
if (Cs < 2201) {
return 2;
}
else if (Cs >= 2201) {
return 4;
}
}
}
}
}
else if (MVDiff >= 23) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
if (Rs < 3142) {
return 8;
}
else if (Rs >= 3142) {
return 4;
}
}
}
}
else if (Rs >= 5113) {
if (ssDCval < 100) {
if (ssDCval < 76) {
if (Rs < 5262) {
return 2;
}
else if (Rs >= 5262) {
return 4;
}
}
else if (ssDCval >= 76) {
return 2;
}
}
else if (ssDCval >= 100) {
if (Rs < 5844) {
return 8;
}
else if (Rs >= 5844) {
return 4;
}
}
}
}
}
else if (TSC >= 15) {
if (diffAFD < 4) {
if (Rs < 4748) {
if (AFD < 31) {
if (RsDiff < 22) {
if (CsDiff < 11) {
return 1;
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (RsDiff >= 22) {
if (Rs < 4200) {
return 1;
}
else if (Rs >= 4200) {
return 4;
}
}
}
else if (AFD >= 31) {
if (posBalance < 59) {
if (AFD < 38) {
return 2;
}
else if (AFD >= 38) {
if (Rs < 4312) {
return 1;
}
else if (Rs >= 4312) {
return 8;
}
}
}
else if (posBalance >= 59) {
return 1;
}
}
}
else if (Rs >= 4748) {
if (tcor < 78) {
if (AFD < 25) {
if (diffRsCsdiff < 116) {
return 2;
}
else if (diffRsCsdiff >= 116) {
if (tcor < 73) {
return 8;
}
else if (tcor >= 73) {
return 2;
}
}
}
else if (AFD >= 25) {
if (gchDC < 2) {
if (Rs < 5806) {
return 2;
}
else if (Rs >= 5806) {
if (Rs < 6970) {
return 4;
}
else if (Rs >= 6970) {
return 1;
}
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (tcor >= 78) {
return 1;
}
}
}
else if (diffAFD >= 4) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
if (negBalance < 38) {
if (RsDiff < 19) {
return 8;
}
else if (RsDiff >= 19) {
return 4;
}
}
else if (negBalance >= 38) {
if (Cs < 3472) {
return 8;
}
else if (Cs >= 3472) {
return 2;
}
}
}
}
}
}
else if (Cs >= 4565) {
if (posBalance < 53) {
if (SC < 9858) {
if (Rs < 3865) {
return 1;
}
else if (Rs >= 3865) {
return 4;
}
}
else if (SC >= 9858) {
return 8;
}
}
else if (posBalance >= 53) {
if (mu_mv_mag_sq < 66) {
return 4;
}
else if (mu_mv_mag_sq >= 66) {
return 8;
}
}
}
}
else if (refDCval >= 111) {
if (TSC < 18) {
if (MVDiff < 24) {
if (tcor < 88) {
if (gchDC < 2) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (SC < 5751) {
if (Rs < 2924) {
return 8;
}
else if (Rs >= 2924) {
return 2;
}
}
else if (SC >= 5751) {
if (AFD < 29) {
return 4;
}
else if (AFD >= 29) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
return 4;
}
}
}
else if (tcor >= 88) {
return 2;
}
}
else if (MVDiff >= 24) {
if (Rs < 4872) {
if (AFD < 26) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
return 8;
}
}
else if (AFD >= 26) {
if (Rs < 2624) {
return 1;
}
else if (Rs >= 2624) {
return 2;
}
}
}
else if (Rs >= 4872) {
if (MVDiff < 64) {
return 8;
}
else if (MVDiff >= 64) {
return 2;
}
}
}
}
else if (TSC >= 18) {
if (TSC < 26) {
return 8;
}
else if (TSC >= 26) {
return 4;
}
}
}
}
else if (refDCval >= 140) {
if (refDCval < 166) {
if (mcTcor < 91) {
if (Rs < 5019) {
if (CsDiff < 13) {
return 4;
}
else if (CsDiff >= 13) {
if (Rs < 2340) {
return 2;
}
else if (Rs >= 2340) {
return 8;
}
}
}
else if (Rs >= 5019) {
if (diffRsCsdiff < 365) {
return 4;
}
else if (diffRsCsdiff >= 365) {
if (Cs < 3871) {
if (Rs < 6391) {
return 8;
}
else if (Rs >= 6391) {
return 2;
}
}
else if (Cs >= 3871) {
return 4;
}
}
}
}
else if (mcTcor >= 91) {
if (ssDCval < 153) {
if (Rs < 6540) {
return 2;
}
else if (Rs >= 6540) {
return 8;
}
}
else if (ssDCval >= 153) {
if (Cs < 1992) {
return 2;
}
else if (Cs >= 1992) {
return 4;
}
}
}
}
else if (refDCval >= 166) {
if (Cs < 3720) {
return 2;
}
else if (Cs >= 3720) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (Cs < 3421) {
if (refDCval < 90) {
if (posBalance < 48) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 4;
}
}
else if (posBalance >= 48) {
if (Rs < 2625) {
if (CsDiff < 9) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (CsDiff >= 9) {
if (Rs < 1835) {
return 1;
}
else if (Rs >= 1835) {
return 4;
}
}
}
else if (Rs >= 2625) {
if (mu_mv_mag_sq < 80) {
if (refDCval < 81) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
return 8;
}
}
else if (refDCval >= 81) {
if (SC < 5665) {
return 8;
}
else if (SC >= 5665) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 80) {
return 8;
}
}
}
}
else if (refDCval >= 90) {
if (CsDiff < 9) {
if (posBalance < 58) {
if (AFD < 18) {
if (MVDiff < 3) {
if (Rs < 4288) {
return 4;
}
else if (Rs >= 4288) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (posBalance >= 58) {
if (Cs < 2546) {
if (Rs < 4735) {
if (Rs < 3723) {
return 2;
}
else if (Rs >= 3723) {
return 1;
}
}
else if (Rs >= 4735) {
return 8;
}
}
else if (Cs >= 2546) {
return 4;
}
}
}
else if (CsDiff >= 9) {
if (Rs < 3554) {
if (ssDCval < 110) {
return 4;
}
else if (ssDCval >= 110) {
return 8;
}
}
else if (Rs >= 3554) {
if (ssDCval < 122) {
if (Cs < 2768) {
return 1;
}
else if (Cs >= 2768) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (ssDCval >= 122) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
if (Rs < 3829) {
return 1;
}
else if (Rs >= 3829) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 3421) {
if (negBalance < 63) {
if (negBalance < 51) {
if (Rs < 5431) {
if (ssDCval < 70) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
else if (ssDCval >= 70) {
if (diffTSC < -2) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
return 4;
}
}
else if (diffTSC >= -2) {
if (MVDiff < 29) {
if (diffMVdiffVal < -6) {
return 4;
}
else if (diffMVdiffVal >= -6) {
if (RsCsDiff < 201) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
else if (RsCsDiff >= 201) {
return 8;
}
}
}
else if (MVDiff >= 29) {
return 1;
}
}
}
}
else if (Rs >= 5431) {
if (diffRsCsdiff < 141) {
if (ssDCval < 99) {
if (RsDiff < 22) {
return 4;
}
else if (RsDiff >= 22) {
return 8;
}
}
else if (ssDCval >= 99) {
if (Rs < 6167) {
return 2;
}
else if (Rs >= 6167) {
if (Rs < 8493) {
return 8;
}
else if (Rs >= 8493) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 141) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (Rs < 3179) {
if (Cs < 4625) {
return 4;
}
else if (Cs >= 4625) {
return 8;
}
}
else if (Rs >= 3179) {
if (MVDiff < 19) {
if (Rs < 7760) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
if (mu_mv_mag_sq < 110) {
if (Rs < 3465) {
if (Rs < 3461) {
return 8;
}
else if (Rs >= 3461) {
return 4;
}
}
else if (Rs >= 3465) {
return 8;
}
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 6235) {
return 8;
}
else if (Rs >= 6235) {
if (Rs < 7034) {
return 4;
}
else if (Rs >= 7034) {
return 8;
}
}
}
}
}
else if (Rs >= 7760) {
return 4;
}
}
else if (MVDiff >= 19) {
if (Rs < 4677) {
return 1;
}
else if (Rs >= 4677) {
return 2;
}
}
}
}
}
else if (negBalance >= 63) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1645) {
if (Rs < 3143) {
if (Rs < 1455) {
if (ssDCval < 122) {
if (negBalance < 1) {
return 2;
}
else if (negBalance >= 1) {
if (ssDCval < 63) {
if (AFD < 3) {
if (posBalance < 33) {
if (Cs < 1383) {
return 8;
}
else if (Cs >= 1383) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 1159) {
if (Rs < 838) {
return 1;
}
else if (Rs >= 838) {
return 8;
}
}
else if (Rs >= 1159) {
return 1;
}
}
}
}
else if (posBalance >= 33) {
if (SC < 2522) {
return 1;
}
else if (SC >= 2522) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
return 4;
}
}
}
}
else if (AFD >= 3) {
if (Rs < 1247) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (Rs >= 1247) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 4;
}
}
}
}
}
else if (ssDCval >= 63) {
if (Cs < 1414) {
if (RsDiff < 1) {
if (Rs < 392) {
return 4;
}
else if (Rs >= 392) {
return 1;
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 115) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 115) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (Rs < 1388) {
return 4;
}
else if (Rs >= 1388) {
if (Cs < 1379) {
return 8;
}
else if (Cs >= 1379) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1414) {
if (SC < 1934) {
return 1;
}
else if (SC >= 1934) {
if (negBalance < 41) {
if (Cs < 1431) {
if (Rs < 905) {
return 8;
}
else if (Rs >= 905) {
return 2;
}
}
else if (Cs >= 1431) {
return 8;
}
}
else if (negBalance >= 41) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (Rs < 1123) {
return 4;
}
else if (Rs >= 1123) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 2) {
if (posBalance < 28) {
if (Cs < 1358) {
return 1;
}
else if (Cs >= 1358) {
return 8;
}
}
else if (posBalance >= 28) {
if (Rs < 1362) {
return 4;
}
else if (Rs >= 1362) {
return 8;
}
}
}
else if (AFD >= 2) {
if (mcTcor < 96) {
if (ssDCval < 129) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 934) {
return 8;
}
else if (Rs >= 934) {
return 4;
}
}
}
else if (ssDCval >= 129) {
if (posBalance < 45) {
if (Rs < 889) {
return 2;
}
else if (Rs >= 889) {
return 1;
}
}
else if (posBalance >= 45) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (Cs < 1512) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (Cs >= 1512) {
if (refDCval < 128) {
if (Rs < 1271) {
return 4;
}
else if (Rs >= 1271) {
return 8;
}
}
else if (refDCval >= 128) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1455) {
if (ssDCval < 86) {
if (tcor < 97) {
if (posBalance < 44) {
if (AFD < 3) {
if (Cs < 1350) {
return 4;
}
else if (Cs >= 1350) {
if (diffMVdiffVal < -46) {
return 1;
}
else if (diffMVdiffVal >= -46) {
return 8;
}
}
}
else if (AFD >= 3) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 4;
}
}
}
else if (posBalance >= 44) {
if (Rs < 2831) {
return 8;
}
else if (Rs >= 2831) {
return 2;
}
}
}
else if (tcor >= 97) {
if (negBalance < 19) {
if (SC < 4464) {
if (Cs < 1351) {
if (Rs < 2274) {
return 8;
}
else if (Rs >= 2274) {
if (SC < 3619) {
if (Rs < 2295) {
if (Cs < 1315) {
return 2;
}
else if (Cs >= 1315) {
if (Rs < 2287) {
return 8;
}
else if (Rs >= 2287) {
if (Rs < 2290) {
return 2;
}
else if (Rs >= 2290) {
return 8;
}
}
}
}
else if (Rs >= 2295) {
return 2;
}
}
else if (SC >= 3619) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (negBalance < 13) {
if (Cs < 1330) {
if (Rs < 2306) {
return 2;
}
else if (Rs >= 2306) {
return 8;
}
}
else if (Cs >= 1330) {
return 8;
}
}
else if (negBalance >= 13) {
if (Cs < 1334) {
if (Cs < 1330) {
return 2;
}
else if (Cs >= 1330) {
return 8;
}
}
else if (Cs >= 1334) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1351) {
if (Rs < 1488) {
if (ssDCval < 48) {
return 8;
}
else if (ssDCval >= 48) {
return 1;
}
}
else if (Rs >= 1488) {
if (mu_mv_mag_sq < 116) {
return 1;
}
else if (mu_mv_mag_sq >= 116) {
if (Rs < 1854) {
if (Cs < 1443) {
return 8;
}
else if (Cs >= 1443) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (Rs >= 1854) {
return 8;
}
}
}
}
}
else if (SC >= 4464) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
return 1;
}
}
}
else if (negBalance >= 19) {
if (Cs < 1364) {
if (Rs < 2419) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 2419) {
return 2;
}
}
else if (Cs >= 1364) {
if (Rs < 1512) {
if (mu_mv_mag_sq < 127) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
if (Rs < 1474) {
return 8;
}
else if (Rs >= 1474) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
else if (Rs >= 1512) {
if (Rs < 1967) {
if (ssDCval < 78) {
if (Cs < 1370) {
return 8;
}
else if (Cs >= 1370) {
return 4;
}
}
else if (ssDCval >= 78) {
if (Rs < 1756) {
return 8;
}
else if (Rs >= 1756) {
return 4;
}
}
}
else if (Rs >= 1967) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (ssDCval < 117) {
if (diffAFD < -1) {
if (MVDiff < 103) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (tcor < 95) {
if (Rs < 2541) {
return 8;
}
else if (Rs >= 2541) {
return 1;
}
}
else if (tcor >= 95) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
else if (MVDiff >= 103) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (ssDCval >= 117) {
if (CsDiff < 2) {
if (Rs < 2919) {
if (diffAFD < -20) {
if (Cs < 1578) {
if (Rs < 1750) {
return 1;
}
else if (Rs >= 1750) {
return 4;
}
}
else if (Cs >= 1578) {
return 8;
}
}
else if (diffAFD >= -20) {
return 8;
}
}
else if (Rs >= 2919) {
if (Cs < 1511) {
return 2;
}
else if (Cs >= 1511) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 125) {
if (Rs < 1467) {
return 1;
}
else if (Rs >= 1467) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (ssDCval < 128) {
return 1;
}
else if (ssDCval >= 128) {
return 8;
}
}
}
}
}
else if (ssDCval >= 140) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3143) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 125) {
if (Rs < 3595) {
if (AFD < 3) {
if (Cs < 1515) {
return 1;
}
else if (Cs >= 1515) {
return 4;
}
}
else if (AFD >= 3) {
return 2;
}
}
else if (Rs >= 3595) {
if (ssDCval < 68) {
if (negBalance < 21) {
if (posBalance < 33) {
return 4;
}
else if (posBalance >= 33) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (negBalance >= 21) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (ssDCval >= 68) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
if (Rs < 4304) {
return 2;
}
else if (Rs >= 4304) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1568) {
if (RsCsDiff < 11) {
if (SC < 5014) {
if (negBalance < 16) {
if (Rs < 3532) {
return 1;
}
else if (Rs >= 3532) {
return 4;
}
}
else if (negBalance >= 16) {
return 8;
}
}
else if (SC >= 5014) {
return 8;
}
}
else if (RsCsDiff >= 11) {
if (Rs < 3681) {
return 8;
}
else if (Rs >= 3681) {
return 4;
}
}
}
else if (Cs >= 1568) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 110) {
if (posBalance < 44) {
if (negBalance < 28) {
if (diffAFD < -1) {
if (ssDCval < 96) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 2;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (diffAFD >= -1) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (Rs < 3800) {
return 4;
}
else if (Rs >= 3800) {
return 8;
}
}
}
}
else if (negBalance >= 28) {
if (ssDCval < 97) {
if (Rs < 4727) {
return 8;
}
else if (Rs >= 4727) {
return 2;
}
}
else if (ssDCval >= 97) {
return 4;
}
}
}
else if (posBalance >= 44) {
if (mu_mv_mag_sq < 217) {
if (Rs < 5054) {
if (Cs < 1636) {
return 2;
}
else if (Cs >= 1636) {
return 4;
}
}
else if (Rs >= 5054) {
return 8;
}
}
else if (mu_mv_mag_sq >= 217) {
return 4;
}
}
}
else if (ssDCval >= 110) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (diffAFD < -11) {
return 1;
}
else if (diffAFD >= -11) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1645) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (Rs < 3843) {
if (Cs < 2804) {
if (Rs < 3242) {
return 8;
}
else if (Rs >= 3242) {
if (Rs < 3253) {
return 4;
}
else if (Rs >= 3253) {
if (Cs < 1859) {
if (Cs < 1851) {
return 8;
}
else if (Cs >= 1851) {
return 4;
}
}
else if (Cs >= 1859) {
return 8;
}
}
}
}
else if (Cs >= 2804) {
return 4;
}
}
else if (Rs >= 3843) {
if (ssDCval < 68) {
if (Rs < 8850) {
return 8;
}
else if (Rs >= 8850) {
return 2;
}
}
else if (ssDCval >= 68) {
if (Rs < 4215) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
return 8;
}
}
else if (Rs >= 4215) {
return 8;
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (SC < 6073) {
if (posBalance < 36) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
if (negBalance < 16) {
if (Rs < 1369) {
return 8;
}
else if (Rs >= 1369) {
return 1;
}
}
else if (negBalance >= 16) {
if (Cs < 1664) {
return 4;
}
else if (Cs >= 1664) {
if (SC < 3955) {
if (Rs < 1195) {
return 1;
}
else if (Rs >= 1195) {
return 8;
}
}
else if (SC >= 3955) {
if (Rs < 2780) {
return 1;
}
else if (Rs >= 2780) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 36) {
if (Rs < 1659) {
if (ssDCval < 58) {
if (Rs < 992) {
return 8;
}
else if (Rs >= 992) {
return 4;
}
}
else if (ssDCval >= 58) {
if (AFD < 3) {
if (Cs < 1728) {
return 8;
}
else if (Cs >= 1728) {
return 1;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (Rs >= 1659) {
if (ssDCval < 66) {
if (negBalance < 19) {
if (Rs < 2634) {
if (negBalance < 14) {
return 1;
}
else if (negBalance >= 14) {
return 8;
}
}
else if (Rs >= 2634) {
if (Cs < 2083) {
return 8;
}
else if (Cs >= 2083) {
return 4;
}
}
}
else if (negBalance >= 19) {
if (posBalance < 43) {
if (SC < 4094) {
return 8;
}
else if (SC >= 4094) {
if (ssDCval < 65) {
if (negBalance < 22) {
return 4;
}
else if (negBalance >= 22) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 8;
}
}
}
else if (ssDCval >= 65) {
return 1;
}
}
}
else if (posBalance >= 43) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 2353) {
return 8;
}
else if (Rs >= 2353) {
return 1;
}
}
}
}
}
else if (ssDCval >= 66) {
return 1;
}
}
}
}
else if (SC >= 6073) {
if (Cs < 3453) {
if (negBalance < 3) {
return 4;
}
else if (negBalance >= 3) {
if (CsDiff < 1) {
if (SC < 7242) {
if (Rs < 3649) {
return 8;
}
else if (Rs >= 3649) {
return 4;
}
}
else if (SC >= 7242) {
return 8;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
}
else if (Cs >= 3453) {
return 8;
}
}
}
else if (ssDCval >= 67) {
if (mu_mv_mag_sq < 123) {
if (TSCindex < 1) {
if (refDCval < 92) {
if (negBalance < 8) {
return 1;
}
else if (negBalance >= 8) {
if (Cs < 2138) {
return 8;
}
else if (Cs >= 2138) {
return 2;
}
}
}
else if (refDCval >= 92) {
if (Rs < 2319) {
return 2;
}
else if (Rs >= 2319) {
if (Cs < 2105) {
return 4;
}
else if (Cs >= 2105) {
return 2;
}
}
}
}
else if (TSCindex >= 1) {
if (posBalance < 35) {
if (mu_mv_mag_sq < 115) {
if (Rs < 2478) {
if (Rs < 2055) {
return 8;
}
else if (Rs >= 2055) {
return 4;
}
}
else if (Rs >= 2478) {
return 1;
}
}
else if (mu_mv_mag_sq >= 115) {
if (Cs < 1932) {
if (Rs < 3092) {
return 8;
}
else if (Rs >= 3092) {
return 2;
}
}
else if (Cs >= 1932) {
return 8;
}
}
}
else if (posBalance >= 35) {
if (diffAFD < -1) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (posBalance < 43) {
if (Cs < 2076) {
return 2;
}
else if (Cs >= 2076) {
return 8;
}
}
else if (posBalance >= 43) {
return 8;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (SC < 6961) {
if (Rs < 1099) {
if (Rs < 648) {
return 8;
}
else if (Rs >= 648) {
return 1;
}
}
else if (Rs >= 1099) {
if (MVDiff < 4) {
if (SC < 4550) {
if (SC < 4410) {
return 8;
}
else if (SC >= 4410) {
if (Rs < 2093) {
return 4;
}
else if (Rs >= 2093) {
if (diffRsCsdiff < -2) {
if (Cs < 1794) {
return 4;
}
else if (Cs >= 1794) {
return 8;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
}
else if (SC >= 4550) {
if (Cs < 2052) {
if (Cs < 2037) {
return 8;
}
else if (Cs >= 2037) {
return 2;
}
}
else if (Cs >= 2052) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 79) {
return 4;
}
else if (ssDCval >= 79) {
return 8;
}
}
}
}
else if (SC >= 6961) {
if (SC < 7988) {
if (SC < 7184) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (SC >= 7184) {
return 4;
}
}
else if (SC >= 7988) {
if (refDCval < 72) {
if (Rs < 4964) {
return 1;
}
else if (Rs >= 4964) {
return 8;
}
}
else if (refDCval >= 72) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6660) {
if (SC < 8581) {
if (ssDCval < 81) {
if (refDCval < 67) {
if (posBalance < 41) {
if (Rs < 2948) {
return 8;
}
else if (Rs >= 2948) {
if (refDCval < 53) {
return 8;
}
else if (refDCval >= 53) {
if (diffAFD < 1) {
if (Rs < 2954) {
return 4;
}
else if (Rs >= 2954) {
if (SC < 7777) {
if (negBalance < 25) {
if (posBalance < 27) {
if (diffMVdiffVal < 1) {
if (AFD < 4) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (AFD >= 4) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (posBalance >= 27) {
if (diffTSC < 1) {
if (RsCsDiff < 38) {
if (diffRsCsdiff < 4) {
if (RsDiff < 5) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
if (gchDC < 1) {
if (Rs < 3110) {
if (Rs < 3026) {
if (Rs < 3003) {
return 8;
}
else if (Rs >= 3003) {
return 8;
}
}
else if (Rs >= 3026) {
return 8;
}
}
else if (Rs >= 3110) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < -5) {
if (Cs < 2637) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (Cs >= 2637) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
if (mu_mv_mag_sq < 128) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
}
}
}
else if (gchDC >= 1) {
if (posBalance < 33) {
return 4;
}
else if (posBalance >= 33) {
return 4;
}
}
}
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (diffRsCsdiff >= 4) {
if (Cs < 2024) {
if (Cs < 1872) {
return 4;
}
else if (Cs >= 1872) {
return 4;
}
}
else if (Cs >= 2024) {
if (Rs < 4311) {
return 4;
}
else if (Rs >= 4311) {
return 4;
}
}
}
}
else if (RsCsDiff >= 38) {
if (Rs < 4338) {
return 4;
}
else if (Rs >= 4338) {
if (Rs < 4936) {
if (tcor < 94) {
if (diffRsCsdiff < -17) {
return 4;
}
else if (diffRsCsdiff >= -17) {
return 4;
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (Rs >= 4936) {
return 4;
}
}
}
}
else if (diffTSC >= 1) {
if (Rs < 5009) {
return 8;
}
else if (Rs >= 5009) {
return 8;
}
}
}
}
else if (negBalance >= 25) {
if (Cs < 1839) {
if (posBalance < 36) {
if (AFD < 3) {
if (Rs < 3020) {
return 4;
}
else if (Rs >= 3020) {
return 4;
}
}
else if (AFD >= 3) {
return 8;
}
}
else if (posBalance >= 36) {
return 8;
}
}
else if (Cs >= 1839) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
return 4;
}
}
}
}
else if (SC >= 7777) {
if (tcor < 97) {
if (Rs < 5124) {
return 4;
}
else if (Rs >= 5124) {
return 4;
}
}
else if (tcor >= 97) {
return 4;
}
}
}
}
else if (diffAFD >= 1) {
if (tcor < 92) {
if (Rs < 4877) {
return 4;
}
else if (Rs >= 4877) {
return 4;
}
}
else if (tcor >= 92) {
return 4;
}
}
}
}
}
else if (posBalance >= 41) {
if (mu_mv_mag_sq < 119) {
if (MVDiff < 3) {
if (ssDCval < 59) {
if (Cs < 2851) {
if (SC < 6373) {
return 8;
}
else if (SC >= 6373) {
if (gchDC < 1) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (gchDC >= 1) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
return 4;
}
}
}
}
else if (Cs >= 2851) {
return 4;
}
}
else if (ssDCval >= 59) {
return 8;
}
}
else if (MVDiff >= 3) {
if (TSC < 6) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 2962) {
return 4;
}
else if (Rs >= 2962) {
if (Cs < 1987) {
if (Rs < 3550) {
return 8;
}
else if (Rs >= 3550) {
return 8;
}
}
else if (Cs >= 1987) {
return 4;
}
}
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (diffMVdiffVal < 1) {
if (RsCsDiff < 16) {
if (posBalance < 42) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (Rs < 2501) {
return 1;
}
else if (Rs >= 2501) {
return 8;
}
}
}
else if (posBalance >= 42) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (mu_mv_mag_sq < 126) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Rs < 4426) {
if (Rs < 4075) {
return 4;
}
else if (Rs >= 4075) {
return 4;
}
}
else if (Rs >= 4426) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 1) {
if (diffRsCsdiff < 6) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 67) {
if (posBalance < 62) {
if (diffMVdiffVal < -2) {
if (mcTcor < 96) {
if (mcTcor < 82) {
if (Rs < 3502) {
return 4;
}
else if (Rs >= 3502) {
return 8;
}
}
else if (mcTcor >= 82) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Cs < 2003) {
if (Rs < 3339) {
return 8;
}
else if (Rs >= 3339) {
return 2;
}
}
else if (Cs >= 2003) {
return 4;
}
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 1875) {
if (Rs < 2616) {
return 8;
}
else if (Rs >= 2616) {
if (tcor < 95) {
if (Rs < 3130) {
return 8;
}
else if (Rs >= 3130) {
return 4;
}
}
else if (tcor >= 95) {
if (RsCsDiff < 41) {
return 4;
}
else if (RsCsDiff >= 41) {
return 2;
}
}
}
}
else if (Cs >= 1875) {
if (gchDC < 3) {
if (diffRsCsdiff < -24) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Rs < 2699) {
return 8;
}
else if (Rs >= 2699) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= -24) {
if (Rs < 2184) {
if (Cs < 2301) {
return 8;
}
else if (Cs >= 2301) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (Rs >= 2184) {
if (diffAFD < -2) {
if (Rs < 3184) {
return 8;
}
else if (Rs >= 3184) {
return 4;
}
}
else if (diffAFD >= -2) {
if (Cs < 1906) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 4;
}
}
else if (Cs >= 1906) {
return 8;
}
}
}
}
}
else if (gchDC >= 3) {
return 4;
}
}
}
}
else if (posBalance >= 62) {
if (diffAFD < 2) {
if (posBalance < 69) {
if (Cs < 3656) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
else if (Cs >= 3656) {
return 4;
}
}
else if (posBalance >= 69) {
if (Rs < 1979) {
return 8;
}
else if (Rs >= 1979) {
return 4;
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
else if (ssDCval >= 81) {
if (Cs < 2641) {
if (Rs < 3797) {
if (Cs < 1881) {
if (tcor < 98) {
if (Rs < 3040) {
if (Rs < 1106) {
return 1;
}
else if (Rs >= 1106) {
if (negBalance < 59) {
return 8;
}
else if (negBalance >= 59) {
return 4;
}
}
}
else if (Rs >= 3040) {
return 2;
}
}
else if (tcor >= 98) {
if (SC < 4194) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (SC >= 4194) {
if (ssDCval < 86) {
return 4;
}
else if (ssDCval >= 86) {
if (Rs < 3438) {
if (Rs < 2836) {
return 4;
}
else if (Rs >= 2836) {
return 8;
}
}
else if (Rs >= 3438) {
return 2;
}
}
}
}
}
else if (Cs >= 1881) {
if (tcor < 99) {
if (negBalance < 49) {
if (Cs < 2014) {
if (Rs < 3101) {
if (diffAFD < 4) {
if (posBalance < 59) {
if (ssDCval < 82) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
return 4;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
else if (posBalance >= 59) {
return 4;
}
}
else if (diffAFD >= 4) {
return 4;
}
}
else if (Rs >= 3101) {
return 2;
}
}
else if (Cs >= 2014) {
if (TSC < 10) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
if (Rs < 2925) {
return 8;
}
else if (Rs >= 2925) {
return 4;
}
}
}
else if (TSC >= 10) {
if (RsDiff < 8) {
if (Rs < 2463) {
return 8;
}
else if (Rs >= 2463) {
return 4;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
else if (negBalance >= 49) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
else if (tcor >= 99) {
if (Cs < 2292) {
if (Rs < 3042) {
return 8;
}
else if (Rs >= 3042) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (Cs >= 2292) {
return 8;
}
}
}
}
else if (Rs >= 3797) {
if (negBalance < 29) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (posBalance < 44) {
if (ssDCval < 97) {
if (Rs < 4884) {
if (Cs < 2167) {
if (Rs < 4679) {
return 2;
}
else if (Rs >= 4679) {
return 8;
}
}
else if (Cs >= 2167) {
return 8;
}
}
else if (Rs >= 4884) {
return 4;
}
}
else if (ssDCval >= 97) {
return 4;
}
}
else if (posBalance >= 44) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (negBalance >= 29) {
if (AFD < 8) {
if (posBalance < 42) {
if (SC < 6478) {
return 8;
}
else if (SC >= 6478) {
if (Rs < 4670) {
return 2;
}
else if (Rs >= 4670) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (MVDiff < 1) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
}
else if (AFD >= 8) {
return 1;
}
}
}
}
else if (Cs >= 2641) {
if (Cs < 3639) {
if (RsDiff < 6) {
if (SC < 7322) {
if (posBalance < 44) {
if (Cs < 2777) {
return 8;
}
else if (Cs >= 2777) {
return 4;
}
}
else if (posBalance >= 44) {
if (Rs < 2629) {
return 4;
}
else if (Rs >= 2629) {
if (SC < 6486) {
return 8;
}
else if (SC >= 6486) {
if (ssDCval < 95) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
if (Cs < 3544) {
return 4;
}
else if (Cs >= 3544) {
return 8;
}
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
else if (SC >= 7322) {
return 4;
}
}
else if (RsDiff >= 6) {
if (Cs < 3348) {
if (Rs < 3453) {
if (Cs < 3166) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
if (Rs < 3108) {
return 8;
}
else if (Rs >= 3108) {
return 4;
}
}
}
else if (Cs >= 3166) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
return 4;
}
}
}
else if (Rs >= 3453) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
return 8;
}
}
}
else if (Cs >= 3348) {
if (posBalance < 56) {
if (Rs < 3506) {
return 8;
}
else if (Rs >= 3506) {
return 4;
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
}
else if (Cs >= 3639) {
if (AFD < 7) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
if (diffRsCsdiff < -18) {
return 8;
}
else if (diffRsCsdiff >= -18) {
return 4;
}
}
}
else if (AFD >= 7) {
if (Cs < 4249) {
if (SC < 6695) {
if (Rs < 2936) {
return 8;
}
else if (Rs >= 2936) {
return 4;
}
}
else if (SC >= 6695) {
return 8;
}
}
else if (Cs >= 4249) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 8581) {
if (ssDCval < 79) {
if (Rs < 5980) {
if (SC < 8675) {
if (AFD < 2) {
if (Rs < 4763) {
return 1;
}
else if (Rs >= 4763) {
return 8;
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (SC >= 8675) {
if (diffAFD < 2) {
if (SC < 8770) {
if (SC < 8759) {
return 8;
}
else if (SC >= 8759) {
return 1;
}
}
else if (SC >= 8770) {
return 8;
}
}
else if (diffAFD >= 2) {
if (Rs < 4605) {
return 1;
}
else if (Rs >= 4605) {
return 8;
}
}
}
}
else if (Rs >= 5980) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (ssDCval >= 79) {
return 8;
}
}
}
else if (Rs >= 6660) {
if (ssDCval < 95) {
if (Rs < 9152) {
return 4;
}
else if (Rs >= 9152) {
return 2;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
else if (ssDCval >= 102) {
if (ssDCval < 129) {
if (Rs < 1496) {
if (posBalance < 54) {
if (Cs < 2179) {
return 8;
}
else if (Cs >= 2179) {
return 2;
}
}
else if (posBalance >= 54) {
if (SC < 3119) {
if (Cs < 1740) {
return 4;
}
else if (Cs >= 1740) {
return 8;
}
}
else if (SC >= 3119) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 4;
}
}
}
}
else if (Rs >= 1496) {
if (refDCval < 105) {
if (negBalance < 41) {
if (Cs < 1991) {
if (Rs < 3464) {
if (Rs < 2127) {
return 2;
}
else if (Rs >= 2127) {
return 8;
}
}
else if (Rs >= 3464) {
return 4;
}
}
else if (Cs >= 1991) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
if (Cs < 2804) {
return 2;
}
else if (Cs >= 2804) {
return 8;
}
}
}
}
else if (negBalance >= 41) {
if (Cs < 3040) {
if (Rs < 4934) {
if (Rs < 3556) {
return 8;
}
else if (Rs >= 3556) {
return 1;
}
}
else if (Rs >= 4934) {
return 2;
}
}
else if (Cs >= 3040) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (diffRsCsdiff < -999) {
return 1;
}
else if (diffRsCsdiff >= -999) {
if (Rs < 1908) {
if (tcor < 99) {
if (Cs < 1651) {
return 1;
}
else if (Cs >= 1651) {
return 8;
}
}
else if (tcor >= 99) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (Cs < 1890) {
return 8;
}
else if (Cs >= 1890) {
return 2;
}
}
}
}
else if (Rs >= 1908) {
if (SC < 5699) {
if (Cs < 2842) {
if (TSC < 10) {
if (refDCval < 123) {
if (Rs < 3367) {
if (diffTSC < -10) {
if (ssDCval < 108) {
if (Cs < 2344) {
return 4;
}
else if (Cs >= 2344) {
return 8;
}
}
else if (ssDCval >= 108) {
return 8;
}
}
else if (diffTSC >= -10) {
return 8;
}
}
else if (Rs >= 3367) {
if (Cs < 2023) {
if (Cs < 1756) {
return 8;
}
else if (Cs >= 1756) {
return 4;
}
}
else if (Cs >= 2023) {
return 8;
}
}
}
else if (refDCval >= 123) {
if (diffRsCsdiff < -49) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
else if (diffRsCsdiff >= -49) {
if (posBalance < 8) {
return 1;
}
else if (posBalance >= 8) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (Rs < 2416) {
return 8;
}
else if (Rs >= 2416) {
return 1;
}
}
}
else if (Cs >= 2842) {
if (posBalance < 32) {
if (MVDiff < 6) {
if (negBalance < 8) {
return 8;
}
else if (negBalance >= 8) {
if (mu_mv_mag_sq < 116) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 116) {
return 4;
}
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (posBalance >= 32) {
return 8;
}
}
}
else if (SC >= 5699) {
if (Rs < 2807) {
if (RsDiff < 1) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (Rs < 2789) {
return 8;
}
else if (Rs >= 2789) {
return 2;
}
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (Rs >= 2807) {
if (Cs < 1885) {
return 2;
}
else if (Cs >= 1885) {
if (posBalance < 59) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
if (mu_mv_mag_sq < 122) {
if (Rs < 6345) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 3883) {
return 8;
}
else if (Rs >= 3883) {
return 4;
}
}
}
else if (Rs >= 6345) {
return 4;
}
}
else if (mu_mv_mag_sq >= 122) {
return 8;
}
}
}
else if (posBalance >= 59) {
if (SC < 7280) {
if (negBalance < 52) {
if (Rs < 4973) {
return 8;
}
else if (Rs >= 4973) {
return 2;
}
}
else if (negBalance >= 52) {
if (Rs < 4112) {
return 2;
}
else if (Rs >= 4112) {
return 8;
}
}
}
else if (SC >= 7280) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 129) {
if (mcTcor < 96) {
if (TSC < 6) {
if (ssDCval < 199) {
return 8;
}
else if (ssDCval >= 199) {
return 4;
}
}
else if (TSC >= 6) {
if (RsCsDiff < 180) {
if (posBalance < 51) {
if (Cs < 2186) {
return 1;
}
else if (Cs >= 2186) {
return 8;
}
}
else if (posBalance >= 51) {
return 1;
}
}
else if (RsCsDiff >= 180) {
if (Cs < 2167) {
return 2;
}
else if (Cs >= 2167) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 144) {
if (SC < 6378) {
if (posBalance < 59) {
if (SC < 2914) {
if (Rs < 1149) {
return 8;
}
else if (Rs >= 1149) {
return 4;
}
}
else if (SC >= 2914) {
if (SC < 4979) {
return 8;
}
else if (SC >= 4979) {
if (SC < 4982) {
return 1;
}
else if (SC >= 4982) {
return 8;
}
}
}
}
else if (posBalance >= 59) {
if (Rs < 1571) {
return 4;
}
else if (Rs >= 1571) {
return 8;
}
}
}
else if (SC >= 6378) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 46) {
if (Cs < 3891) {
return 2;
}
else if (Cs >= 3891) {
return 8;
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 4250) {
if (gchDC < 1) {
if (posBalance < 15) {
if (Rs < 3015) {
return 1;
}
else if (Rs >= 3015) {
return 8;
}
}
else if (posBalance >= 15) {
return 1;
}
}
else if (gchDC >= 1) {
if (Cs < 3860) {
return 8;
}
else if (Cs >= 3860) {
return 1;
}
}
}
else if (Cs >= 4250) {
return 8;
}
}
}
}
else if (refDCval >= 144) {
if (ssDCval < 200) {
return 8;
}
else if (ssDCval >= 200) {
if (Cs < 4080) {
return 8;
}
else if (Cs >= 4080) {
return 4;
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelectE(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1313) {
if (Rs < 2322) {
if (Cs < 534) {
if (negBalance < 14) {
if (MVDiff < 8) {
if (SC < 782) {
if (ssDCval < 67) {
if (Cs < 1) {
if (ssDCval < 42) {
if (ssDCval < 22) {
if (ssDCval < 10) {
return 4;
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (ssDCval >= 22) {
if (Rs < 15) {
return 4;
}
else if (Rs >= 15) {
return 8;
}
}
}
else if (ssDCval >= 42) {
return 2;
}
}
else if (Cs >= 1) {
if (Cs < 205) {
if (Rs < 259) {
if (negBalance < 8) {
if (Rs < 4) {
if (Rs < 1) {
return 8;
}
else if (Rs >= 1) {
return 4;
}
}
else if (Rs >= 4) {
if (ssDCval < 59) {
if (Cs < 186) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
return 4;
}
}
}
else if (Cs >= 186) {
return 4;
}
}
else if (ssDCval >= 59) {
return 2;
}
}
}
else if (negBalance >= 8) {
if (MVDiff < 1) {
if (tcor < 98) {
if (Rs < 89) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 8;
}
}
else if (Rs >= 89) {
if (Cs < 168) {
return 1;
}
else if (Cs >= 168) {
return 8;
}
}
}
else if (tcor >= 98) {
if (Cs < 177) {
return 8;
}
else if (Cs >= 177) {
return 4;
}
}
}
else if (MVDiff >= 1) {
if (refDCval < 32) {
if (posBalance < 28) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (MVDiff < 6) {
if (Rs < 109) {
return 4;
}
else if (Rs >= 109) {
return 8;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
}
else if (posBalance >= 28) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Rs < 42) {
return 4;
}
else if (Rs >= 42) {
return 8;
}
}
}
}
else if (refDCval >= 32) {
if (ssDCval < 53) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
if (ssDCval < 41) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (ssDCval >= 41) {
if (diffMVdiffVal < -4) {
if (Rs < 60) {
return 1;
}
else if (Rs >= 60) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
}
else if (ssDCval >= 53) {
if (Rs < 118) {
return 1;
}
else if (Rs >= 118) {
return 4;
}
}
}
}
}
}
else if (Rs >= 259) {
if (ssDCval < 51) {
if (tcor < 104) {
if (Cs < 159) {
if (Cs < 67) {
return 8;
}
else if (Cs >= 67) {
if (negBalance < 12) {
return 4;
}
else if (negBalance >= 12) {
if (ssDCval < 37) {
return 8;
}
else if (ssDCval >= 37) {
return 4;
}
}
}
}
else if (Cs >= 159) {
return 8;
}
}
else if (tcor >= 104) {
return 8;
}
}
else if (ssDCval >= 51) {
return 8;
}
}
}
else if (Cs >= 205) {
if (Rs < 200) {
if (ssDCval < 36) {
if (Rs < 170) {
if (Cs < 231) {
return 1;
}
else if (Cs >= 231) {
return 8;
}
}
else if (Rs >= 170) {
return 4;
}
}
else if (ssDCval >= 36) {
if (tcor < 93) {
if (Cs < 253) {
if (Rs < 170) {
return 8;
}
else if (Rs >= 170) {
return 1;
}
}
else if (Cs >= 253) {
return 4;
}
}
else if (tcor >= 93) {
if (Cs < 215) {
if (Rs < 146) {
return 1;
}
else if (Rs >= 146) {
return 4;
}
}
else if (Cs >= 215) {
return 1;
}
}
}
}
else if (Rs >= 200) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 40) {
if (posBalance < 36) {
return 4;
}
else if (posBalance >= 36) {
if (Rs < 256) {
if (Rs < 226) {
return 2;
}
else if (Rs >= 226) {
return 8;
}
}
else if (Rs >= 256) {
return 1;
}
}
}
else if (ssDCval >= 40) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
if (Rs < 430) {
return 4;
}
else if (Rs >= 430) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 367) {
if (ssDCval < 26) {
return 8;
}
else if (ssDCval >= 26) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (SC < 565) {
if (Rs < 221) {
return 8;
}
else if (Rs >= 221) {
return 4;
}
}
else if (SC >= 565) {
return 8;
}
}
}
}
else if (Rs >= 367) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (posBalance < 28) {
if (Rs < 181) {
return 1;
}
else if (Rs >= 181) {
if (ssDCval < 129) {
return 2;
}
else if (ssDCval >= 129) {
if (Rs < 467) {
return 4;
}
else if (Rs >= 467) {
return 1;
}
}
}
}
else if (posBalance >= 28) {
if (ssDCval < 94) {
if (Cs < 120) {
if (Rs < 145) {
return 4;
}
else if (Rs >= 145) {
return 8;
}
}
else if (Cs >= 120) {
if (Cs < 245) {
if (Rs < 294) {
if (Cs < 238) {
return 2;
}
else if (Cs >= 238) {
return 1;
}
}
else if (Rs >= 294) {
if (ssDCval < 88) {
if (Cs < 214) {
return 1;
}
else if (Cs >= 214) {
return 2;
}
}
else if (ssDCval >= 88) {
return 2;
}
}
}
else if (Cs >= 245) {
if (SC < 644) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
return 4;
}
}
else if (SC >= 644) {
return 8;
}
}
}
}
else if (ssDCval >= 94) {
if (Rs < 426) {
return 4;
}
else if (Rs >= 426) {
return 1;
}
}
}
}
}
else if (SC >= 782) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 33) {
if (Cs < 273) {
return 4;
}
else if (Cs >= 273) {
if (Cs < 355) {
return 8;
}
else if (Cs >= 355) {
return 1;
}
}
}
else if (ssDCval >= 33) {
if (Rs < 718) {
if (Cs < 492) {
if (ssDCval < 141) {
if (SC < 957) {
if (RsCsDiff < 2) {
if (diffRsCsdiff < -1) {
if (Rs < 540) {
return 8;
}
else if (Rs >= 540) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
else if (RsCsDiff >= 2) {
return 1;
}
}
else if (SC >= 957) {
if (Rs < 599) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
return 8;
}
}
else if (Rs >= 599) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
if (Cs < 449) {
return 4;
}
else if (Cs >= 449) {
return 8;
}
}
}
}
}
else if (ssDCval >= 141) {
if (Rs < 547) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
if (SC < 965) {
if (SC < 799) {
return 4;
}
else if (SC >= 799) {
return 1;
}
}
else if (SC >= 965) {
return 8;
}
}
}
else if (Rs >= 547) {
return 8;
}
}
}
else if (Cs >= 492) {
if (Rs < 485) {
return 1;
}
else if (Rs >= 485) {
if (ssDCval < 65) {
if (Rs < 654) {
return 1;
}
else if (Rs >= 654) {
return 2;
}
}
else if (ssDCval >= 65) {
return 8;
}
}
}
}
else if (Rs >= 718) {
if (ssDCval < 67) {
if (refDCval < 59) {
if (Cs < 379) {
if (AFD < 4) {
if (mu_mv_mag_sq < 125) {
if (Rs < 861) {
return 8;
}
else if (Rs >= 861) {
return 1;
}
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
else if (AFD >= 4) {
if (Rs < 1423) {
return 4;
}
else if (Rs >= 1423) {
return 2;
}
}
}
else if (Cs >= 379) {
if (ssDCval < 57) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
return 1;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (refDCval >= 59) {
if (Cs < 448) {
if (Cs < 199) {
return 8;
}
else if (Cs >= 199) {
return 1;
}
}
else if (Cs >= 448) {
return 2;
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 12) {
if (SC < 2220) {
return 8;
}
else if (SC >= 2220) {
return 1;
}
}
else if (negBalance >= 12) {
if (Cs < 203) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Cs < 51) {
return 8;
}
else if (Cs >= 51) {
return 1;
}
}
}
else if (Cs >= 203) {
if (ssDCval < 79) {
if (Rs < 1343) {
return 8;
}
else if (Rs >= 1343) {
return 1;
}
}
else if (ssDCval >= 79) {
if (negBalance < 13) {
if (Rs < 2072) {
return 8;
}
else if (Rs >= 2072) {
return 2;
}
}
else if (negBalance >= 13) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (Cs < 317) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (ssDCval < 87) {
if (Rs < 1074) {
return 1;
}
else if (Rs >= 1074) {
return 2;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
}
else if (Cs >= 317) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (posBalance < 38) {
if (ssDCval < 66) {
if (Rs < 1099) {
return 8;
}
else if (Rs >= 1099) {
return 2;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (posBalance >= 38) {
if (SC < 1090) {
return 4;
}
else if (SC >= 1090) {
if (ssDCval < 81) {
return 2;
}
else if (ssDCval >= 81) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
else if (MVDiff >= 8) {
if (negBalance < 6) {
if (ssDCval < 22) {
if (AFD < 1) {
if (MVDiff < 1076) {
return 8;
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (AFD >= 1) {
if (posBalance < 83) {
return 2;
}
else if (posBalance >= 83) {
if (tcor < 58) {
if (Rs < 175) {
return 8;
}
else if (Rs >= 175) {
return 1;
}
}
else if (tcor >= 58) {
return 4;
}
}
}
}
else if (ssDCval >= 22) {
if (posBalance < 5) {
if (tcor < 100) {
if (Rs < 8) {
return 4;
}
else if (Rs >= 8) {
if (Rs < 381) {
return 8;
}
else if (Rs >= 381) {
return 4;
}
}
}
else if (tcor >= 100) {
return 4;
}
}
else if (posBalance >= 5) {
if (diffRsCsdiff < -7) {
if (diffAFD < -8) {
if (ssDCval < 145) {
if (ssDCval < 91) {
if (Rs < 991) {
return 8;
}
else if (Rs >= 991) {
return 2;
}
}
else if (ssDCval >= 91) {
return 2;
}
}
else if (ssDCval >= 145) {
if (Rs < 292) {
return 2;
}
else if (Rs >= 292) {
return 8;
}
}
}
else if (diffAFD >= -8) {
if (ssDCval < 101) {
return 8;
}
else if (ssDCval >= 101) {
return 1;
}
}
}
else if (diffRsCsdiff >= -7) {
if (MVDiff < 233) {
if (negBalance < 4) {
if (refDCval < 49) {
if (refDCval < 33) {
if (Rs < 146) {
if (mu_mv_mag_sq < 77) {
if (refDCval < 22) {
return 4;
}
else if (refDCval >= 22) {
return 8;
}
}
else if (mu_mv_mag_sq >= 77) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (Rs >= 146) {
return 4;
}
}
else if (refDCval >= 33) {
if (MVDiff < 29) {
return 8;
}
else if (MVDiff >= 29) {
if (Cs < 223) {
if (Rs < 228) {
return 8;
}
else if (Rs >= 228) {
return 4;
}
}
else if (Cs >= 223) {
return 8;
}
}
}
}
else if (refDCval >= 49) {
if (SC < 159) {
if (negBalance < 2) {
return 2;
}
else if (negBalance >= 2) {
return 4;
}
}
else if (SC >= 159) {
if (tcor < 99) {
if (Cs < 101) {
if (AFD < 6) {
if (Rs < 105) {
return 8;
}
else if (Rs >= 105) {
return 4;
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 101) {
if (Rs < 1185) {
return 8;
}
else if (Rs >= 1185) {
return 4;
}
}
}
else if (tcor >= 99) {
if (MVDiff < 32) {
return 2;
}
else if (MVDiff >= 32) {
if (diffAFD < -6) {
if (ssDCval < 130) {
if (Rs < 300) {
return 8;
}
else if (Rs >= 300) {
return 2;
}
}
else if (ssDCval >= 130) {
return 4;
}
}
else if (diffAFD >= -6) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 4) {
if (RsCsDiff < 1) {
if (diffTSC < -1) {
if (diffTSC < -3) {
if (Rs < 415) {
return 4;
}
else if (Rs >= 415) {
return 2;
}
}
else if (diffTSC >= -3) {
return 8;
}
}
else if (diffTSC >= -1) {
if (Rs < 61) {
return 2;
}
else if (Rs >= 61) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
else if (RsCsDiff >= 1) {
if (Cs < 378) {
if (Cs < 96) {
if (Rs < 158) {
return 1;
}
else if (Rs >= 158) {
return 8;
}
}
else if (Cs >= 96) {
return 1;
}
}
else if (Cs >= 378) {
if (Rs < 502) {
return 2;
}
else if (Rs >= 502) {
if (Rs < 1259) {
return 8;
}
else if (Rs >= 1259) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 233) {
if (mu_mv_mag_sq < 4) {
return 8;
}
else if (mu_mv_mag_sq >= 4) {
if (AFD < 7) {
if (Rs < 197) {
return 8;
}
else if (Rs >= 197) {
return 4;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 6) {
if (tcor < 94) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (ssDCval < 36) {
if (posBalance < 68) {
if (Cs < 201) {
return 1;
}
else if (Cs >= 201) {
if (MVDiff < 29) {
if (Cs < 211) {
return 8;
}
else if (Cs >= 211) {
if (negBalance < 9) {
return 4;
}
else if (negBalance >= 9) {
return 1;
}
}
}
else if (MVDiff >= 29) {
return 4;
}
}
}
else if (posBalance >= 68) {
if (Rs < 129) {
if (TSC < 1) {
return 2;
}
else if (TSC >= 1) {
return 8;
}
}
else if (Rs >= 129) {
if (Rs < 327) {
return 1;
}
else if (Rs >= 327) {
return 4;
}
}
}
}
else if (ssDCval >= 36) {
if (MVDiff < 13) {
if (posBalance < 95) {
return 8;
}
else if (posBalance >= 95) {
if (Rs < 1388) {
return 2;
}
else if (Rs >= 1388) {
return 4;
}
}
}
else if (MVDiff >= 13) {
if (refDCval < 135) {
if (MVDiff < 57) {
if (posBalance < 94) {
if (Rs < 1261) {
return 1;
}
else if (Rs >= 1261) {
if (Rs < 1718) {
return 8;
}
else if (Rs >= 1718) {
return 1;
}
}
}
else if (posBalance >= 94) {
return 8;
}
}
else if (MVDiff >= 57) {
if (gchDC < 11) {
if (Rs < 156) {
return 4;
}
else if (Rs >= 156) {
return 2;
}
}
else if (gchDC >= 11) {
if (tcor < 76) {
if (Rs < 218) {
if (Cs < 29) {
return 1;
}
else if (Cs >= 29) {
return 8;
}
}
else if (Rs >= 218) {
return 2;
}
}
else if (tcor >= 76) {
return 1;
}
}
}
}
else if (refDCval >= 135) {
return 2;
}
}
}
}
}
else if (tcor >= 94) {
if (SC < 876) {
if (Cs < 118) {
if (ssDCval < 40) {
if (posBalance < 21) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
return 4;
}
}
else if (posBalance >= 21) {
return 1;
}
}
else if (ssDCval >= 40) {
if (MVDiff < 24) {
if (diffMVdiffVal < -10) {
return 1;
}
else if (diffMVdiffVal >= -10) {
return 8;
}
}
else if (MVDiff >= 24) {
if (Rs < 39) {
return 2;
}
else if (Rs >= 39) {
return 4;
}
}
}
}
else if (Cs >= 118) {
if (mu_mv_mag_sq < 66) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 66) {
if (RsDiff < 1) {
if (mu_mv_mag_sq < 116) {
if (ssDCval < 88) {
if (Rs < 212) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (mu_mv_mag_sq < 71) {
return 4;
}
else if (mu_mv_mag_sq >= 71) {
if (mu_mv_mag_sq < 98) {
return 1;
}
else if (mu_mv_mag_sq >= 98) {
return 2;
}
}
}
}
else if (Rs >= 212) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (ssDCval >= 88) {
if (Rs < 470) {
return 4;
}
else if (Rs >= 470) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
else if (SC >= 876) {
if (Cs < 399) {
if (ssDCval < 64) {
if (Rs < 610) {
return 4;
}
else if (Rs >= 610) {
return 8;
}
}
else if (ssDCval >= 64) {
if (Cs < 100) {
return 8;
}
else if (Cs >= 100) {
if (posBalance < 81) {
return 1;
}
else if (posBalance >= 81) {
return 8;
}
}
}
}
else if (Cs >= 399) {
if (ssDCval < 43) {
return 4;
}
else if (ssDCval >= 43) {
if (Rs < 1330) {
return 8;
}
else if (Rs >= 1330) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (mcTcor < 79) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (tcor < 60) {
if (Cs < 292) {
if (negBalance < 49) {
if (posBalance < 55) {
if (ssDCval < 37) {
return 1;
}
else if (ssDCval >= 37) {
return 4;
}
}
else if (posBalance >= 55) {
if (Rs < 162) {
return 8;
}
else if (Rs >= 162) {
if (Cs < 129) {
return 1;
}
else if (Cs >= 129) {
return 8;
}
}
}
}
else if (negBalance >= 49) {
if (MVDiff < 288) {
if (Rs < 842) {
return 8;
}
else if (Rs >= 842) {
return 1;
}
}
else if (MVDiff >= 288) {
return 4;
}
}
}
else if (Cs >= 292) {
if (RsCsDiff < 16) {
return 2;
}
else if (RsCsDiff >= 16) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
else if (tcor >= 60) {
if (TSC < 3) {
if (ssDCval < 25) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
if (ssDCval < 23) {
return 4;
}
else if (ssDCval >= 23) {
if (Rs < 27) {
return 4;
}
else if (Rs >= 27) {
return 1;
}
}
}
}
else if (ssDCval >= 25) {
if (Rs < 442) {
return 2;
}
else if (Rs >= 442) {
if (Rs < 664) {
return 8;
}
else if (Rs >= 664) {
return 4;
}
}
}
}
else if (TSC >= 3) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 2;
}
}
}
}
}
else if (mcTcor >= 79) {
if (posBalance < 27) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 115) {
if (tcor < 102) {
return 4;
}
else if (tcor >= 102) {
return 8;
}
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Rs < 10) {
if (SC < 9) {
if (negBalance < 80) {
if (ssDCval < 23) {
if (MVDiff < 247) {
return 8;
}
else if (MVDiff >= 247) {
if (refDCval < 18) {
return 8;
}
else if (refDCval >= 18) {
return 4;
}
}
}
else if (ssDCval >= 23) {
return 4;
}
}
else if (negBalance >= 80) {
return 8;
}
}
else if (SC >= 9) {
return 8;
}
}
else if (Rs >= 10) {
if (ssDCval < 31) {
if (posBalance < 21) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (posBalance < 3) {
return 4;
}
else if (posBalance >= 3) {
if (mu_mv_mag_sq < 62) {
return 1;
}
else if (mu_mv_mag_sq >= 62) {
if (Rs < 17) {
return 1;
}
else if (Rs >= 17) {
return 8;
}
}
}
}
}
else if (posBalance >= 21) {
if (negBalance < 15) {
return 4;
}
else if (negBalance >= 15) {
return 1;
}
}
}
else if (ssDCval >= 31) {
if (Rs < 382) {
if (diffMVdiffVal < -3) {
if (ssDCval < 32) {
return 8;
}
else if (ssDCval >= 32) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (Rs >= 382) {
if (Rs < 462) {
return 2;
}
else if (Rs >= 462) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 27) {
if (posBalance < 69) {
if (Rs < 72) {
return 1;
}
else if (Rs >= 72) {
if (posBalance < 45) {
if (refDCval < 29) {
if (refDCval < 23) {
if (Rs < 485) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
return 1;
}
}
else if (Rs >= 485) {
return 4;
}
}
else if (refDCval >= 23) {
if (mu_mv_mag_sq < 39) {
if (gchDC < 1) {
if (MVDiff < 21) {
if (Rs < 592) {
return 1;
}
else if (Rs >= 592) {
return 4;
}
}
else if (MVDiff >= 21) {
return 8;
}
}
else if (gchDC >= 1) {
if (posBalance < 37) {
if (Rs < 635) {
return 1;
}
else if (Rs >= 635) {
return 4;
}
}
else if (posBalance >= 37) {
if (Rs < 94) {
return 8;
}
else if (Rs >= 94) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (diffMVdiffVal < 11) {
return 1;
}
else if (diffMVdiffVal >= 11) {
if (Rs < 189) {
return 1;
}
else if (Rs >= 189) {
return 4;
}
}
}
}
}
else if (refDCval >= 29) {
if (Cs < 267) {
if (RsCsDiff < 1) {
if (mu_mv_mag_sq < 68) {
if (negBalance < 23) {
if (posBalance < 41) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 376) {
return 8;
}
else if (Rs >= 376) {
return 4;
}
}
}
else if (posBalance >= 41) {
return 1;
}
}
else if (negBalance >= 23) {
if (TSCindex < 2) {
if (Rs < 100) {
return 8;
}
else if (Rs >= 100) {
return 4;
}
}
else if (TSCindex >= 2) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (posBalance < 31) {
if (MVDiff < 4) {
if (ssDCval < 33) {
return 8;
}
else if (ssDCval >= 33) {
return 1;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (posBalance >= 31) {
return 1;
}
}
}
else if (RsCsDiff >= 1) {
if (TSC < 3) {
if (Rs < 364) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (mcTcor < 95) {
if (CsDiff < 1) {
if (negBalance < 25) {
return 4;
}
else if (negBalance >= 25) {
if (Cs < 65) {
if (Cs < 25) {
return 1;
}
else if (Cs >= 25) {
return 4;
}
}
else if (Cs >= 65) {
return 1;
}
}
}
else if (CsDiff >= 1) {
if (Rs < 185) {
return 8;
}
else if (Rs >= 185) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 345) {
return 8;
}
else if (Rs >= 345) {
return 2;
}
}
}
}
}
else if (mcTcor >= 95) {
if (AFD < 5) {
if (SC < 317) {
if (Cs < 136) {
return 4;
}
else if (Cs >= 136) {
return 8;
}
}
else if (SC >= 317) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (Rs >= 364) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
return 4;
}
}
}
else if (TSC >= 3) {
if (ssDCval < 32) {
return 1;
}
else if (ssDCval >= 32) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
if (Rs < 286) {
return 1;
}
else if (Rs >= 286) {
return 8;
}
}
}
}
}
}
else if (Cs >= 267) {
if (CsDiff < 3) {
if (TSCindex < 2) {
if (Rs < 204) {
return 1;
}
else if (Rs >= 204) {
return 4;
}
}
else if (TSCindex >= 2) {
return 1;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
else if (posBalance >= 45) {
if (tcor < 63) {
if (Rs < 174) {
return 8;
}
else if (Rs >= 174) {
if (posBalance < 52) {
if (Rs < 198) {
return 1;
}
else if (Rs >= 198) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (posBalance >= 52) {
return 1;
}
}
}
else if (tcor >= 63) {
if (negBalance < 26) {
if (mcTcor < 88) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
return 8;
}
}
else if (mcTcor >= 88) {
if (MVDiff < 27) {
return 1;
}
else if (MVDiff >= 27) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (negBalance >= 26) {
if (SC < 352) {
if (tcor < 65) {
return 8;
}
else if (tcor >= 65) {
if (SC < 347) {
return 1;
}
else if (SC >= 347) {
return 2;
}
}
}
else if (SC >= 352) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 69) {
if (AFD < 7) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 317) {
if (ssDCval < 49) {
if (mcTcor < 92) {
if (posBalance < 49) {
if (tcor < 77) {
if (gchDC < 22) {
if (posBalance < 42) {
if (TSC < 6) {
if (diffMVdiffVal < 0) {
if (RsDiff < 1) {
if (Rs < 43) {
return 4;
}
else if (Rs >= 43) {
return 2;
}
}
else if (RsDiff >= 1) {
if (Rs < 259) {
return 1;
}
else if (Rs >= 259) {
return 4;
}
}
}
else if (diffMVdiffVal >= 0) {
return 2;
}
}
else if (TSC >= 6) {
return 1;
}
}
else if (posBalance >= 42) {
return 1;
}
}
else if (gchDC >= 22) {
if (Rs < 70) {
return 1;
}
else if (Rs >= 70) {
return 8;
}
}
}
else if (tcor >= 77) {
if (TSC < 3) {
if (tcor < 82) {
if (Rs < 137) {
return 4;
}
else if (Rs >= 137) {
return 2;
}
}
else if (tcor >= 82) {
return 1;
}
}
else if (TSC >= 3) {
if (AFD < 11) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
if (mu_mv_mag_sq < 109) {
return 8;
}
else if (mu_mv_mag_sq >= 109) {
return 2;
}
}
}
else if (AFD >= 11) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
}
}
}
else if (posBalance >= 49) {
if (gchDC < 1) {
if (MVDiff < 39) {
if (SC < 508) {
if (ssDCval < 45) {
if (Cs < 111) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
return 1;
}
}
else if (Cs >= 111) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
}
}
else if (ssDCval >= 45) {
return 4;
}
}
else if (SC >= 508) {
if (mcTcor < 88) {
if (tcor < 66) {
return 2;
}
else if (tcor >= 66) {
return 8;
}
}
else if (mcTcor >= 88) {
if (Cs < 379) {
return 2;
}
else if (Cs >= 379) {
return 1;
}
}
}
}
else if (MVDiff >= 39) {
if (posBalance < 57) {
if (Rs < 268) {
return 1;
}
else if (Rs >= 268) {
return 8;
}
}
else if (posBalance >= 57) {
if (AFD < 12) {
if (mu_mv_mag_sq < 16) {
return 4;
}
else if (mu_mv_mag_sq >= 16) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
return 8;
}
}
}
else if (AFD >= 12) {
return 2;
}
}
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 103) {
if (AFD < 8) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (MVDiff < 7) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (Cs < 232) {
return 1;
}
else if (Cs >= 232) {
return 8;
}
}
}
else if (MVDiff >= 7) {
if (Cs < 172) {
if (negBalance < 34) {
if (Rs < 131) {
return 1;
}
else if (Rs >= 131) {
return 8;
}
}
else if (negBalance >= 34) {
return 8;
}
}
else if (Cs >= 172) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
}
else if (AFD >= 8) {
if (MVDiff < 49) {
if (ssDCval < 45) {
if (gchDC < 4) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (ssDCval < 41) {
if (diffRsCsdiff < -7) {
return 2;
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
else if (ssDCval >= 41) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
else if (gchDC >= 4) {
return 1;
}
}
else if (ssDCval >= 45) {
return 4;
}
}
else if (MVDiff >= 49) {
if (mcTcor < 89) {
if (posBalance < 64) {
if (diffMVdiffVal < 37) {
if (AFD < 11) {
if (diffMVdiffVal < 6) {
if (refDCval < 42) {
return 4;
}
else if (refDCval >= 42) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
return 8;
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (diffMVdiffVal >= 37) {
if (MVDiff < 265) {
return 1;
}
else if (MVDiff >= 265) {
return 8;
}
}
}
else if (posBalance >= 64) {
if (MVDiff < 83) {
if (Rs < 264) {
return 4;
}
else if (Rs >= 264) {
return 1;
}
}
else if (MVDiff >= 83) {
if (Cs < 46) {
return 1;
}
else if (Cs >= 46) {
return 2;
}
}
}
}
else if (mcTcor >= 89) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 29) {
return 8;
}
else if (Rs >= 29) {
if (Cs < 320) {
return 4;
}
else if (Cs >= 320) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 61) {
if (posBalance < 8) {
if (Cs < 68) {
if (diffMVdiffVal < 23) {
return 8;
}
else if (diffMVdiffVal >= 23) {
if (Cs < 2) {
return 8;
}
else if (Cs >= 2) {
return 4;
}
}
}
else if (Cs >= 68) {
return 4;
}
}
else if (posBalance >= 8) {
if (ssDCval < 43) {
if (AFD < 5) {
if (negBalance < 44) {
if (MVDiff < 18) {
if (Cs < 244) {
return 1;
}
else if (Cs >= 244) {
return 8;
}
}
else if (MVDiff >= 18) {
if (AFD < 4) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 2;
}
}
else if (AFD >= 4) {
if (tcor < 87) {
return 2;
}
else if (tcor >= 87) {
if (Cs < 134) {
return 1;
}
else if (Cs >= 134) {
return 2;
}
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 126) {
if (Rs < 44) {
return 8;
}
else if (Rs >= 44) {
return 1;
}
}
else if (Rs >= 126) {
return 2;
}
}
}
else if (AFD >= 5) {
if (Cs < 183) {
if (RsDiff < 1) {
if (Cs < 50) {
if (Rs < 39) {
if (Rs < 15) {
return 8;
}
else if (Rs >= 15) {
return 1;
}
}
else if (Rs >= 39) {
return 2;
}
}
else if (Cs >= 50) {
if (Rs < 131) {
if (Rs < 41) {
return 1;
}
else if (Rs >= 41) {
return 4;
}
}
else if (Rs >= 131) {
if (Rs < 207) {
return 2;
}
else if (Rs >= 207) {
return 1;
}
}
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 36) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 2) {
if (Cs < 73) {
return 1;
}
else if (Cs >= 73) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
return 1;
}
}
else if (diffAFD >= 2) {
if (Rs < 220) {
return 2;
}
else if (Rs >= 220) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 36) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
return 1;
}
}
}
}
else if (Cs >= 183) {
if (Rs < 137) {
if (gchDC < 1) {
if (Rs < 109) {
return 4;
}
else if (Rs >= 109) {
return 8;
}
}
else if (gchDC >= 1) {
if (refDCval < 40) {
return 4;
}
else if (refDCval >= 40) {
return 1;
}
}
}
else if (Rs >= 137) {
if (negBalance < 40) {
if (Rs < 199) {
return 8;
}
else if (Rs >= 199) {
return 1;
}
}
else if (negBalance >= 40) {
if (Rs < 258) {
if (Rs < 140) {
return 8;
}
else if (Rs >= 140) {
return 2;
}
}
else if (Rs >= 258) {
if (Rs < 277) {
return 8;
}
else if (Rs >= 277) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 43) {
if (Rs < 271) {
if (diffMVdiffVal < 10) {
if (posBalance < 67) {
if (MVDiff < 36) {
return 1;
}
else if (MVDiff >= 36) {
if (posBalance < 44) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 1;
}
}
else if (posBalance >= 44) {
return 1;
}
}
}
else if (posBalance >= 67) {
return 2;
}
}
else if (diffMVdiffVal >= 10) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
if (Rs < 100) {
return 2;
}
else if (Rs >= 100) {
return 4;
}
}
}
}
else if (Rs >= 271) {
if (Cs < 163) {
return 1;
}
else if (Cs >= 163) {
if (mu_mv_mag_sq < 50) {
if (ssDCval < 48) {
return 4;
}
else if (ssDCval >= 48) {
return 1;
}
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 280) {
return 2;
}
else if (Rs >= 280) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (SC < 559) {
if (mu_mv_mag_sq < 100) {
if (negBalance < 27) {
if (Cs < 269) {
if (Rs < 137) {
if (Rs < 132) {
if (ssDCval < 45) {
return 1;
}
else if (ssDCval >= 45) {
return 8;
}
}
else if (Rs >= 132) {
return 8;
}
}
else if (Rs >= 137) {
if (RsCsDiff < 1) {
if (diffMVdiffVal < -1) {
if (Rs < 151) {
return 4;
}
else if (Rs >= 151) {
if (Rs < 250) {
return 1;
}
else if (Rs >= 250) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (RsCsDiff >= 1) {
if (tcor < 76) {
return 4;
}
else if (tcor >= 76) {
return 1;
}
}
}
}
else if (Cs >= 269) {
if (Rs < 193) {
return 1;
}
else if (Rs >= 193) {
return 8;
}
}
}
else if (negBalance >= 27) {
if (mcTcor < 104) {
if (Rs < 274) {
if (posBalance < 23) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (posBalance >= 23) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (RsDiff < 1) {
if (Cs < 320) {
if (Rs < 68) {
return 1;
}
else if (Rs >= 68) {
return 8;
}
}
else if (Cs >= 320) {
return 1;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
else if (Rs >= 274) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
else if (mcTcor >= 104) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 44) {
if (Cs < 8) {
if (Rs < 8) {
return 2;
}
else if (Rs >= 8) {
if (Rs < 100) {
return 4;
}
else if (Rs >= 100) {
return 1;
}
}
}
else if (Cs >= 8) {
return 4;
}
}
else if (Cs >= 44) {
if (posBalance < 37) {
if (tcor < 101) {
if (Cs < 253) {
if (ssDCval < 44) {
if (SC < 206) {
return 4;
}
else if (SC >= 206) {
if (Cs < 166) {
return 1;
}
else if (Cs >= 166) {
if (Rs < 144) {
return 8;
}
else if (Rs >= 144) {
return 4;
}
}
}
}
else if (ssDCval >= 44) {
if (Rs < 187) {
return 8;
}
else if (Rs >= 187) {
return 4;
}
}
}
else if (Cs >= 253) {
if (Rs < 232) {
return 8;
}
else if (Rs >= 232) {
return 1;
}
}
}
else if (tcor >= 101) {
if (Rs < 114) {
return 8;
}
else if (Rs >= 114) {
if (Cs < 202) {
return 4;
}
else if (Cs >= 202) {
return 8;
}
}
}
}
else if (posBalance >= 37) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (posBalance < 44) {
if (Cs < 286) {
return 1;
}
else if (Cs >= 286) {
return 8;
}
}
else if (posBalance >= 44) {
return 8;
}
}
}
}
}
}
else if (SC >= 559) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
return 1;
}
}
}
}
}
else if (ssDCval >= 49) {
if (refDCval < 70) {
if (posBalance < 34) {
if (refDCval < 61) {
if (Cs < 294) {
if (Rs < 19) {
return 2;
}
else if (Rs >= 19) {
if (tcor < 87) {
if (negBalance < 17) {
if (Rs < 75) {
return 8;
}
else if (Rs >= 75) {
return 1;
}
}
else if (negBalance >= 17) {
if (MVDiff < 18) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
if (Rs < 252) {
return 1;
}
else if (Rs >= 252) {
return 8;
}
}
}
else if (MVDiff >= 18) {
return 1;
}
}
}
else if (tcor >= 87) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (Cs < 125) {
return 8;
}
else if (Cs >= 125) {
return 4;
}
}
}
}
}
else if (Cs >= 294) {
if (RsDiff < 2) {
if (Rs < 282) {
return 4;
}
else if (Rs >= 282) {
if (Rs < 293) {
return 1;
}
else if (Rs >= 293) {
return 4;
}
}
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (refDCval >= 61) {
if (Rs < 105) {
if (MVDiff < 72) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (MVDiff >= 72) {
return 4;
}
}
else if (Rs >= 105) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
if (MVDiff < 115) {
if (Cs < 243) {
return 1;
}
else if (Cs >= 243) {
if (Rs < 195) {
return 2;
}
else if (Rs >= 195) {
return 4;
}
}
}
else if (MVDiff >= 115) {
return 2;
}
}
}
}
}
else if (posBalance >= 34) {
if (TSCindex < 2) {
if (ssDCval < 65) {
if (Cs < 45) {
return 1;
}
else if (Cs >= 45) {
if (diffMVdiffVal < -5) {
if (Rs < 214) {
return 8;
}
else if (Rs >= 214) {
return 2;
}
}
else if (diffMVdiffVal >= -5) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Rs < 245) {
if (Cs < 189) {
if (Rs < 146) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Rs >= 146) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 1;
}
}
}
}
else if (Cs >= 189) {
return 4;
}
}
else if (Rs >= 245) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (Cs < 316) {
return 4;
}
else if (Cs >= 316) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 65) {
if (Cs < 240) {
return 2;
}
else if (Cs >= 240) {
if (Cs < 375) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (Cs >= 375) {
return 1;
}
}
}
}
else if (TSCindex >= 2) {
if (TSC < 4) {
if (RsCsDiff < 1) {
if (Cs < 185) {
if (MVDiff < 2) {
if (Cs < 129) {
if (Cs < 121) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (Cs >= 121) {
return 8;
}
}
else if (Cs >= 129) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 2;
}
}
}
else if (MVDiff >= 2) {
if (mcTcor < 95) {
if (AFD < 5) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (AFD >= 5) {
if (MVDiff < 49) {
return 1;
}
else if (MVDiff >= 49) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (diffAFD < -2) {
if (diffMVdiffVal < -3) {
if (Rs < 199) {
return 1;
}
else if (Rs >= 199) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (diffAFD >= -2) {
if (Rs < 208) {
if (MVDiff < 48) {
if (Rs < 68) {
return 1;
}
else if (Rs >= 68) {
if (ssDCval < 67) {
if (negBalance < 23) {
if (mu_mv_mag_sq < 98) {
return 1;
}
else if (mu_mv_mag_sq >= 98) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
return 1;
}
}
}
else if (negBalance >= 23) {
if (tcor < 95) {
if (mu_mv_mag_sq < 58) {
if (MVDiff < 20) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
if (posBalance < 61) {
return 1;
}
else if (posBalance >= 61) {
return 4;
}
}
}
else if (MVDiff >= 20) {
if (Rs < 118) {
return 4;
}
else if (Rs >= 118) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (SC < 346) {
return 4;
}
else if (SC >= 346) {
return 1;
}
}
}
else if (tcor >= 95) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (posBalance < 43) {
if (diffMVdiffVal < 0) {
return 4;
}
else if (diffMVdiffVal >= 0) {
return 1;
}
}
else if (posBalance >= 43) {
return 1;
}
}
}
}
}
else if (ssDCval >= 67) {
if (Rs < 105) {
return 4;
}
else if (Rs >= 105) {
return 2;
}
}
}
}
else if (MVDiff >= 48) {
return 4;
}
}
else if (Rs >= 208) {
return 4;
}
}
}
}
}
else if (Cs >= 185) {
if (negBalance < 29) {
if (ssDCval < 60) {
if (MVDiff < 1) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
if (Cs < 188) {
return 1;
}
else if (Cs >= 188) {
return 4;
}
}
}
}
else if (ssDCval >= 60) {
if (MVDiff < 5) {
if (MVDiff < 1) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (Cs < 365) {
if (Cs < 219) {
return 2;
}
else if (Cs >= 219) {
if (Cs < 339) {
if (SC < 419) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 1;
}
}
else if (SC >= 419) {
return 4;
}
}
else if (Cs >= 339) {
if (Rs < 179) {
if (Rs < 163) {
return 4;
}
else if (Rs >= 163) {
return 1;
}
}
else if (Rs >= 179) {
return 4;
}
}
}
}
else if (Cs >= 365) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 104) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
return 1;
}
}
else if (mu_mv_mag_sq >= 104) {
if (diffTSC < -1) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
}
}
}
else if (MVDiff >= 5) {
if (Cs < 221) {
return 1;
}
else if (Cs >= 221) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
if (Cs < 377) {
return 2;
}
else if (Cs >= 377) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 29) {
if (tcor < 98) {
if (Rs < 111) {
return 2;
}
else if (Rs >= 111) {
if (mu_mv_mag_sq < 85) {
if (SC < 516) {
if (AFD < 5) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (Rs < 165) {
return 4;
}
else if (Rs >= 165) {
return 1;
}
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (SC >= 516) {
return 1;
}
}
else if (mu_mv_mag_sq >= 85) {
if (SC < 503) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
if (Rs < 128) {
return 4;
}
else if (Rs >= 128) {
return 1;
}
}
}
}
else if (SC >= 503) {
if (posBalance < 54) {
if (Cs < 424) {
return 4;
}
else if (Cs >= 424) {
return 1;
}
}
else if (posBalance >= 54) {
return 1;
}
}
}
}
}
else if (tcor >= 98) {
if (posBalance < 43) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (posBalance >= 43) {
if (Cs < 497) {
if (negBalance < 32) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 98) {
if (TSCindex < 3) {
if (SC < 487) {
return 1;
}
else if (SC >= 487) {
return 4;
}
}
else if (TSCindex >= 3) {
return 4;
}
}
else if (mu_mv_mag_sq >= 98) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
return 1;
}
}
}
}
else if (posBalance >= 49) {
return 1;
}
}
else if (negBalance >= 32) {
return 1;
}
}
else if (Cs >= 497) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (MVDiff < 2) {
if (Cs < 404) {
if (Rs < 175) {
return 1;
}
else if (Rs >= 175) {
if (posBalance < 47) {
if (Cs < 354) {
return 4;
}
else if (Cs >= 354) {
if (Rs < 228) {
return 1;
}
else if (Rs >= 228) {
return 4;
}
}
}
else if (posBalance >= 47) {
if (ssDCval < 64) {
if (diffMVdiffVal < -1) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
else if (Cs >= 404) {
return 1;
}
}
else if (MVDiff >= 2) {
if (Rs < 263) {
if (Rs < 34) {
return 1;
}
else if (Rs >= 34) {
if (ssDCval < 53) {
if (SC < 328) {
return 2;
}
else if (SC >= 328) {
if (tcor < 96) {
if (Rs < 197) {
if (TSC < 3) {
if (Cs < 323) {
return 1;
}
else if (Cs >= 323) {
return 4;
}
}
else if (TSC >= 3) {
return 1;
}
}
else if (Rs >= 197) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
else if (tcor >= 96) {
if (Rs < 247) {
return 1;
}
else if (Rs >= 247) {
return 2;
}
}
}
}
else if (ssDCval >= 53) {
if (mcTcor < 93) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (Rs < 161) {
if (Rs < 114) {
return 4;
}
else if (Rs >= 114) {
return 2;
}
}
else if (Rs >= 161) {
return 1;
}
}
}
else if (mcTcor >= 93) {
if (gchDC < 2) {
if (Cs < 205) {
if (MVDiff < 11) {
if (negBalance < 41) {
if (mu_mv_mag_sq < 53) {
return 1;
}
else if (mu_mv_mag_sq >= 53) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
return 1;
}
}
}
}
else if (negBalance >= 41) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (MVDiff >= 11) {
if (posBalance < 48) {
if (negBalance < 33) {
return 1;
}
else if (negBalance >= 33) {
if (Rs < 144) {
return 4;
}
else if (Rs >= 144) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (Rs < 149) {
return 4;
}
else if (Rs >= 149) {
return 8;
}
}
}
}
}
else if (Cs >= 205) {
if (negBalance < 39) {
if (MVDiff < 52) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (AFD < 5) {
if (posBalance < 49) {
if (Cs < 393) {
return 1;
}
else if (Cs >= 393) {
return 4;
}
}
else if (posBalance >= 49) {
if (diffRsCsdiff < 1) {
return 4;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 163) {
return 4;
}
else if (Rs >= 163) {
return 1;
}
}
}
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (MVDiff >= 52) {
return 1;
}
}
else if (negBalance >= 39) {
if (mu_mv_mag_sq < 45) {
return 4;
}
else if (mu_mv_mag_sq >= 45) {
if (Cs < 503) {
return 8;
}
else if (Cs >= 503) {
return 4;
}
}
}
}
}
else if (gchDC >= 2) {
if (RsCsDiff < 5) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
if (Cs < 208) {
return 1;
}
else if (Cs >= 208) {
return 4;
}
}
}
else if (RsCsDiff >= 5) {
return 1;
}
}
}
}
}
}
else if (Rs >= 263) {
if (RsCsDiff < 2) {
if (posBalance < 57) {
if (mcTcor < 99) {
return 4;
}
else if (mcTcor >= 99) {
if (Rs < 294) {
return 4;
}
else if (Rs >= 294) {
return 2;
}
}
}
else if (posBalance >= 57) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (Rs < 270) {
return 4;
}
else if (Rs >= 270) {
return 2;
}
}
}
}
else if (RsCsDiff >= 2) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (diffRsCsdiff < -2) {
if (gchDC < 1) {
if (Rs < 287) {
return 1;
}
else if (Rs >= 287) {
return 4;
}
}
else if (gchDC >= 1) {
return 2;
}
}
else if (diffRsCsdiff >= -2) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (tcor < 89) {
if (MVDiff < 9) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (MVDiff >= 9) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (tcor >= 89) {
if (tcor < 94) {
if (MVDiff < 11) {
if (diffRsCsdiff < 2) {
return 2;
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (tcor >= 94) {
if (ssDCval < 62) {
if (Rs < 314) {
return 4;
}
else if (Rs >= 314) {
return 1;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 16) {
if (tcor < 59) {
if (RsCsDiff < 19) {
if (refDCval < 61) {
if (TSC < 15) {
return 1;
}
else if (TSC >= 15) {
return 4;
}
}
else if (refDCval >= 61) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (RsCsDiff >= 19) {
if (Rs < 197) {
return 1;
}
else if (Rs >= 197) {
return 8;
}
}
}
else if (tcor >= 59) {
if (MVDiff < 38) {
if (MVDiff < 10) {
if (Rs < 191) {
return 2;
}
else if (Rs >= 191) {
return 8;
}
}
else if (MVDiff >= 10) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (MVDiff >= 38) {
if (AFD < 15) {
if (Cs < 35) {
return 1;
}
else if (Cs >= 35) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (ssDCval < 61) {
if (Rs < 301) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
else if (Rs >= 301) {
return 1;
}
}
else if (ssDCval >= 61) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 2;
}
}
}
}
}
else if (AFD >= 15) {
if (diffMVdiffVal < 54) {
return 1;
}
else if (diffMVdiffVal >= 54) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (RsCsDiff < 1) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (RsCsDiff >= 1) {
if (Cs < 68) {
if (SC < 76) {
if (Cs < 17) {
return 8;
}
else if (Cs >= 17) {
return 2;
}
}
else if (SC >= 76) {
return 1;
}
}
else if (Cs >= 68) {
if (gchDC < 6) {
if (Cs < 416) {
if (posBalance < 52) {
if (SC < 488) {
if (Rs < 127) {
return 8;
}
else if (Rs >= 127) {
return 2;
}
}
else if (SC >= 488) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 4;
}
}
}
else if (posBalance >= 52) {
if (tcor < 88) {
if (mu_mv_mag_sq < 54) {
if (Rs < 242) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (ssDCval < 68) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
if (AFD < 10) {
if (Rs < 195) {
return 2;
}
else if (Rs >= 195) {
return 1;
}
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (ssDCval >= 68) {
if (Cs < 342) {
return 4;
}
else if (Cs >= 342) {
return 2;
}
}
}
}
else if (Rs >= 242) {
if (negBalance < 44) {
if (Cs < 181) {
return 8;
}
else if (Cs >= 181) {
return 2;
}
}
else if (negBalance >= 44) {
if (Rs < 275) {
return 4;
}
else if (Rs >= 275) {
if (RsDiff < 2) {
if (Rs < 307) {
return 2;
}
else if (Rs >= 307) {
return 4;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 296) {
return 4;
}
else if (Rs >= 296) {
return 8;
}
}
}
else if (tcor >= 88) {
if (Cs < 346) {
return 1;
}
else if (Cs >= 346) {
return 8;
}
}
}
}
else if (Cs >= 416) {
if (diffRsCsdiff < -3) {
if (Rs < 152) {
return 8;
}
else if (Rs >= 152) {
return 2;
}
}
else if (diffRsCsdiff >= -3) {
return 1;
}
}
}
else if (gchDC >= 6) {
if (ssDCval < 70) {
if (posBalance < 69) {
return 8;
}
else if (posBalance >= 69) {
if (Rs < 138) {
return 1;
}
else if (Rs >= 138) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (Cs < 219) {
return 1;
}
else if (Cs >= 219) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 109) {
if (mcTcor < 97) {
if (MVDiff < 31) {
if (TSC < 2) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 119) {
return 2;
}
else if (Rs >= 119) {
return 8;
}
}
}
else if (TSC >= 2) {
if (refDCval < 84) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
if (Rs < 289) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (Rs < 242) {
if (SC < 226) {
return 8;
}
else if (SC >= 226) {
return 2;
}
}
else if (Rs >= 242) {
return 1;
}
}
}
else if (Rs >= 289) {
if (Rs < 301) {
return 2;
}
else if (Rs >= 301) {
return 8;
}
}
}
}
}
else if (refDCval >= 84) {
if (AFD < 15) {
if (Cs < 199) {
if (negBalance < 41) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (negBalance >= 41) {
if (ssDCval < 118) {
if (refDCval < 93) {
if (posBalance < 61) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (Rs < 147) {
if (diffMVdiffVal < -15) {
return 2;
}
else if (diffMVdiffVal >= -15) {
if (Cs < 12) {
return 2;
}
else if (Cs >= 12) {
if (Cs < 127) {
return 8;
}
else if (Cs >= 127) {
return 2;
}
}
}
}
else if (Rs >= 147) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (Cs < 137) {
return 1;
}
else if (Cs >= 137) {
return 2;
}
}
}
else if (refDCval >= 93) {
if (Rs < 142) {
return 1;
}
else if (Rs >= 142) {
return 8;
}
}
}
else if (ssDCval >= 118) {
return 2;
}
}
}
else if (Cs >= 199) {
if (TSC < 6) {
if (AFD < 7) {
if (diffMVdiffVal < -5) {
if (Cs < 289) {
return 2;
}
else if (Cs >= 289) {
if (MVDiff < 17) {
if (Cs < 329) {
return 8;
}
else if (Cs >= 329) {
return 2;
}
}
else if (MVDiff >= 17) {
return 2;
}
}
}
else if (diffMVdiffVal >= -5) {
if (tcor < 95) {
if (Rs < 275) {
if (MVDiff < 16) {
if (SC < 577) {
if (MVDiff < 5) {
if (Cs < 285) {
if (Cs < 241) {
if (Rs < 168) {
return 8;
}
else if (Rs >= 168) {
return 2;
}
}
else if (Cs >= 241) {
return 8;
}
}
else if (Cs >= 285) {
return 2;
}
}
else if (MVDiff >= 5) {
if (refDCval < 88) {
if (tcor < 78) {
if (Rs < 180) {
return 2;
}
else if (Rs >= 180) {
return 8;
}
}
else if (tcor >= 78) {
return 2;
}
}
else if (refDCval >= 88) {
if (mcTcor < 90) {
return 2;
}
else if (mcTcor >= 90) {
return 8;
}
}
}
}
else if (SC >= 577) {
if (Cs < 419) {
return 8;
}
else if (Cs >= 419) {
return 2;
}
}
}
else if (MVDiff >= 16) {
if (Rs < 264) {
if (Rs < 219) {
return 2;
}
else if (Rs >= 219) {
if (Cs < 353) {
return 8;
}
else if (Cs >= 353) {
return 2;
}
}
}
else if (Rs >= 264) {
return 2;
}
}
}
else if (Rs >= 275) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
if (Cs < 312) {
return 2;
}
else if (Cs >= 312) {
return 8;
}
}
}
}
else if (tcor >= 95) {
return 2;
}
}
}
else if (AFD >= 7) {
if (ssDCval < 123) {
if (Cs < 255) {
return 2;
}
else if (Cs >= 255) {
if (Rs < 304) {
return 1;
}
else if (Rs >= 304) {
return 2;
}
}
}
else if (ssDCval >= 123) {
return 8;
}
}
}
else if (TSC >= 6) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
return 4;
}
}
}
}
else if (AFD >= 15) {
if (Cs < 428) {
return 1;
}
else if (Cs >= 428) {
return 2;
}
}
}
}
}
else if (MVDiff >= 31) {
if (mcTcor < 83) {
if (refDCval < 139) {
if (ssDCval < 102) {
if (posBalance < 49) {
if (Cs < 41) {
if (Cs < 8) {
return 2;
}
else if (Cs >= 8) {
return 1;
}
}
else if (Cs >= 41) {
if (MVDiff < 323) {
if (Rs < 218) {
return 1;
}
else if (Rs >= 218) {
return 4;
}
}
else if (MVDiff >= 323) {
return 4;
}
}
}
else if (posBalance >= 49) {
if (posBalance < 72) {
if (MVDiff < 170) {
if (Rs < 166) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
return 8;
}
}
else if (Rs >= 166) {
if (Cs < 13) {
return 4;
}
else if (Cs >= 13) {
return 2;
}
}
}
else if (MVDiff >= 170) {
if (tcor < 19) {
return 8;
}
else if (tcor >= 19) {
if (Rs < 209) {
return 4;
}
else if (Rs >= 209) {
return 1;
}
}
}
}
else if (posBalance >= 72) {
if (posBalance < 84) {
return 4;
}
else if (posBalance >= 84) {
return 1;
}
}
}
}
else if (ssDCval >= 102) {
if (Cs < 272) {
return 4;
}
else if (Cs >= 272) {
return 2;
}
}
}
else if (refDCval >= 139) {
if (mu_mv_mag_sq < 35) {
if (Rs < 214) {
if (MVDiff < 394) {
return 2;
}
else if (MVDiff >= 394) {
return 1;
}
}
else if (Rs >= 214) {
return 1;
}
}
else if (mu_mv_mag_sq >= 35) {
return 1;
}
}
}
else if (mcTcor >= 83) {
if (Cs < 11) {
if (diffAFD < -2) {
if (MVDiff < 347) {
return 8;
}
else if (MVDiff >= 347) {
return 1;
}
}
else if (diffAFD >= -2) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 2;
}
}
}
else if (Cs >= 11) {
if (ssDCval < 134) {
if (ssDCval < 91) {
if (AFD < 9) {
if (MVDiff < 33) {
return 1;
}
else if (MVDiff >= 33) {
return 8;
}
}
else if (AFD >= 9) {
if (tcor < 42) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (tcor >= 42) {
if (posBalance < 58) {
if (RsCsDiff < 10) {
if (Cs < 331) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (Cs >= 331) {
return 8;
}
}
else if (RsCsDiff >= 10) {
if (refDCval < 73) {
return 1;
}
else if (refDCval >= 73) {
if (refDCval < 82) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 4;
}
}
else if (refDCval >= 82) {
if (Rs < 302) {
return 2;
}
else if (Rs >= 302) {
return 4;
}
}
}
}
}
else if (posBalance >= 58) {
if (diffRsCsdiff < -9) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (diffRsCsdiff >= -9) {
if (diffMVdiffVal < -77) {
return 1;
}
else if (diffMVdiffVal >= -77) {
if (mcTcor < 85) {
if (Rs < 232) {
return 4;
}
else if (Rs >= 232) {
return 2;
}
}
else if (mcTcor >= 85) {
if (Cs < 56) {
return 1;
}
else if (Cs >= 56) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (refDCval < 117) {
if (Rs < 20) {
return 2;
}
else if (Rs >= 20) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (negBalance < 59) {
if (negBalance < 52) {
if (tcor < 77) {
if (tcor < 68) {
if (Cs < 392) {
return 1;
}
else if (Cs >= 392) {
return 2;
}
}
else if (tcor >= 68) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 2;
}
}
}
else if (tcor >= 77) {
return 1;
}
}
else if (negBalance >= 52) {
return 2;
}
}
else if (negBalance >= 59) {
return 1;
}
}
}
}
else if (refDCval >= 117) {
if (mu_mv_mag_sq < 3) {
if (Cs < 30) {
return 8;
}
else if (Cs >= 30) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 18) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
return 2;
}
}
else if (AFD >= 18) {
if (AFD < 29) {
return 2;
}
else if (AFD >= 29) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 134) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 18) {
if (refDCval < 81) {
if (Cs < 352) {
if (Cs < 133) {
return 8;
}
else if (Cs >= 133) {
return 2;
}
}
else if (Cs >= 352) {
return 1;
}
}
else if (refDCval >= 81) {
if (Rs < 179) {
if (Rs < 11) {
return 1;
}
else if (Rs >= 11) {
return 4;
}
}
else if (Rs >= 179) {
if (Rs < 194) {
return 2;
}
else if (Rs >= 194) {
return 4;
}
}
}
}
else if (negBalance >= 18) {
if (Cs < 124) {
if (ssDCval < 106) {
if (Rs < 142) {
if (Rs < 109) {
if (TSC < 3) {
if (Cs < 118) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
else if (Cs >= 118) {
if (Rs < 55) {
return 4;
}
else if (Rs >= 55) {
return 1;
}
}
}
else if (TSC >= 3) {
if (MVDiff < 21) {
return 2;
}
else if (MVDiff >= 21) {
if (gchDC < 10) {
return 1;
}
else if (gchDC >= 10) {
if (Rs < 44) {
return 4;
}
else if (Rs >= 44) {
return 8;
}
}
}
}
}
else if (Rs >= 109) {
if (ssDCval < 75) {
if (Rs < 134) {
return 2;
}
else if (Rs >= 134) {
return 1;
}
}
else if (ssDCval >= 75) {
if (SC < 224) {
if (SC < 185) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
if (Rs < 129) {
return 2;
}
else if (Rs >= 129) {
return 1;
}
}
}
else if (SC >= 185) {
return 2;
}
}
else if (SC >= 224) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
}
}
else if (Rs >= 142) {
if (MVDiff < 5) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (MVDiff >= 5) {
if (MVDiff < 43) {
if (MVDiff < 10) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (Rs < 200) {
return 2;
}
else if (Rs >= 200) {
return 1;
}
}
}
else if (MVDiff >= 10) {
if (Rs < 309) {
if (refDCval < 73) {
return 8;
}
else if (refDCval >= 73) {
if (Rs < 269) {
return 1;
}
else if (Rs >= 269) {
if (mu_mv_mag_sq < 43) {
if (Cs < 71) {
return 1;
}
else if (Cs >= 71) {
return 8;
}
}
else if (mu_mv_mag_sq >= 43) {
return 1;
}
}
}
}
else if (Rs >= 309) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
}
else if (MVDiff >= 43) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Cs < 49) {
return 2;
}
else if (Cs >= 49) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 106) {
if (negBalance < 28) {
if (negBalance < 26) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
else if (negBalance >= 26) {
return 1;
}
}
else if (negBalance >= 28) {
if (MVDiff < 7) {
if (Rs < 119) {
return 4;
}
else if (Rs >= 119) {
return 8;
}
}
else if (MVDiff >= 7) {
if (AFD < 22) {
if (refDCval < 106) {
return 1;
}
else if (refDCval >= 106) {
if (posBalance < 9) {
return 8;
}
else if (posBalance >= 9) {
return 2;
}
}
}
else if (AFD >= 22) {
if (Rs < 145) {
return 8;
}
else if (Rs >= 145) {
return 1;
}
}
}
}
}
}
else if (Cs >= 124) {
if (ssDCval < 129) {
if (Cs < 256) {
if (refDCval < 94) {
if (TSC < 2) {
if (Rs < 145) {
if (Cs < 152) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
return 2;
}
}
else if (Cs >= 152) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
return 2;
}
}
}
}
else if (Rs >= 145) {
return 2;
}
}
else if (TSC >= 2) {
if (negBalance < 43) {
if (MVDiff < 2) {
if (Rs < 232) {
if (Cs < 149) {
return 8;
}
else if (Cs >= 149) {
return 2;
}
}
else if (Rs >= 232) {
if (Cs < 188) {
if (Rs < 283) {
return 8;
}
else if (Rs >= 283) {
return 4;
}
}
else if (Cs >= 188) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (MVDiff < 33) {
if (Cs < 163) {
if (ssDCval < 74) {
if (Rs < 147) {
return 2;
}
else if (Rs >= 147) {
if (Cs < 150) {
return 2;
}
else if (Cs >= 150) {
return 1;
}
}
}
else if (ssDCval >= 74) {
return 1;
}
}
else if (Cs >= 163) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Cs < 203) {
if (Rs < 269) {
return 1;
}
else if (Rs >= 269) {
return 4;
}
}
else if (Cs >= 203) {
if (ssDCval < 75) {
if (Rs < 106) {
return 4;
}
else if (Rs >= 106) {
return 1;
}
}
else if (ssDCval >= 75) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 33) {
if (Rs < 120) {
return 2;
}
else if (Rs >= 120) {
if (Rs < 136) {
return 1;
}
else if (Rs >= 136) {
return 8;
}
}
}
}
}
else if (negBalance >= 43) {
if (refDCval < 77) {
if (mu_mv_mag_sq < 55) {
if (Rs < 106) {
return 4;
}
else if (Rs >= 106) {
if (refDCval < 72) {
return 4;
}
else if (refDCval >= 72) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 55) {
return 8;
}
}
else if (refDCval >= 77) {
if (Cs < 173) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 245) {
return 2;
}
else if (Rs >= 245) {
return 8;
}
}
}
else if (Cs >= 173) {
if (Rs < 270) {
return 2;
}
else if (Rs >= 270) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 94) {
if (Rs < 258) {
if (negBalance < 31) {
if (Rs < 116) {
if (Rs < 22) {
return 2;
}
else if (Rs >= 22) {
return 4;
}
}
else if (Rs >= 116) {
if (tcor < 99) {
if (SC < 335) {
if (Rs < 146) {
return 1;
}
else if (Rs >= 146) {
return 2;
}
}
else if (SC >= 335) {
if (Cs < 138) {
return 1;
}
else if (Cs >= 138) {
return 4;
}
}
}
else if (tcor >= 99) {
if (Rs < 234) {
return 2;
}
else if (Rs >= 234) {
return 1;
}
}
}
}
else if (negBalance >= 31) {
if (Rs < 129) {
if (Rs < 19) {
return 2;
}
else if (Rs >= 19) {
if (AFD < 3) {
if (Rs < 101) {
return 4;
}
else if (Rs >= 101) {
if (Rs < 118) {
return 2;
}
else if (Rs >= 118) {
return 1;
}
}
}
else if (AFD >= 3) {
if (Rs < 114) {
return 1;
}
else if (Rs >= 114) {
return 2;
}
}
}
}
else if (Rs >= 129) {
if (gchDC < 3) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (tcor < 97) {
if (negBalance < 59) {
return 4;
}
else if (negBalance >= 59) {
return 1;
}
}
else if (tcor >= 97) {
return 2;
}
}
}
else if (gchDC >= 3) {
if (negBalance < 54) {
if (Cs < 197) {
return 2;
}
else if (Cs >= 197) {
if (Rs < 185) {
return 8;
}
else if (Rs >= 185) {
return 1;
}
}
}
else if (negBalance >= 54) {
return 8;
}
}
}
}
}
else if (Rs >= 258) {
if (gchDC < 1) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
else if (Cs >= 256) {
if (RsDiff < 1) {
if (Cs < 436) {
if (ssDCval < 75) {
if (ssDCval < 72) {
if (Rs < 131) {
return 4;
}
else if (Rs >= 131) {
if (Cs < 304) {
if (mu_mv_mag_sq < 92) {
return 1;
}
else if (mu_mv_mag_sq >= 92) {
return 4;
}
}
else if (Cs >= 304) {
if (diffRsCsdiff < -1) {
if (Rs < 155) {
return 4;
}
else if (Rs >= 155) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
return 4;
}
}
}
}
else if (ssDCval >= 72) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 4;
}
}
}
else if (ssDCval >= 75) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 63) {
if (Rs < 233) {
return 4;
}
else if (Rs >= 233) {
return 8;
}
}
else if (mu_mv_mag_sq >= 63) {
if (AFD < 3) {
if (mu_mv_mag_sq < 101) {
return 4;
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 171) {
return 1;
}
else if (Rs >= 171) {
return 4;
}
}
}
else if (AFD >= 3) {
if (ssDCval < 89) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Rs < 188) {
return 8;
}
else if (Rs >= 188) {
if (Rs < 228) {
return 1;
}
else if (Rs >= 228) {
return 2;
}
}
}
}
else if (ssDCval >= 89) {
return 1;
}
}
}
}
}
}
else if (Cs >= 436) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
if (Cs < 509) {
return 2;
}
else if (Cs >= 509) {
return 4;
}
}
}
}
else if (RsDiff >= 1) {
if (CsDiff < 2) {
if (MVDiff < 16) {
if (SC < 437) {
return 4;
}
else if (SC >= 437) {
return 1;
}
}
else if (MVDiff >= 16) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 28) {
return 2;
}
else if (mu_mv_mag_sq >= 28) {
if (Rs < 264) {
return 1;
}
else if (Rs >= 264) {
return 2;
}
}
}
else if (RsDiff >= 2) {
if (diffMVdiffVal < 3) {
if (Cs < 263) {
return 4;
}
else if (Cs >= 263) {
if (negBalance < 62) {
return 1;
}
else if (negBalance >= 62) {
return 2;
}
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 243) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
else if (Rs >= 243) {
if (Rs < 272) {
return 1;
}
else if (Rs >= 272) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 129) {
if (negBalance < 39) {
return 2;
}
else if (negBalance >= 39) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (Rs < 295) {
return 2;
}
else if (Rs >= 295) {
return 1;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 131) {
if (refDCval < 74) {
if (Rs < 104) {
return 4;
}
else if (Rs >= 104) {
return 8;
}
}
else if (refDCval >= 74) {
if (tcor < 98) {
if (AFD < 28) {
if (Cs < 257) {
return 8;
}
else if (Cs >= 257) {
return 2;
}
}
else if (AFD >= 28) {
if (Rs < 22) {
return 4;
}
else if (Rs >= 22) {
return 1;
}
}
}
else if (tcor >= 98) {
return 1;
}
}
}
else if (Rs >= 131) {
if (Cs < 241) {
if (SC < 408) {
if (Cs < 125) {
if (SC < 225) {
return 1;
}
else if (SC >= 225) {
if (MVDiff < 2) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (MVDiff >= 2) {
return 2;
}
}
}
else if (Cs >= 125) {
if (MVDiff < 4) {
if (Rs < 225) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 114) {
return 2;
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (mcTcor >= 98) {
if (Cs < 127) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
return 8;
}
}
else if (Cs >= 127) {
if (diffTSC < -1) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
if (Rs < 160) {
return 2;
}
else if (Rs >= 160) {
return 8;
}
}
}
else if (diffTSC >= -1) {
return 2;
}
}
}
}
else if (Rs >= 225) {
return 8;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
return 2;
}
}
}
}
else if (SC >= 408) {
if (negBalance < 24) {
if (ssDCval < 83) {
if (Rs < 198) {
if (Cs < 220) {
return 1;
}
else if (Cs >= 220) {
return 2;
}
}
else if (Rs >= 198) {
return 2;
}
}
else if (ssDCval >= 83) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
}
else if (negBalance >= 24) {
return 1;
}
}
}
else if (Cs >= 241) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Rs < 165) {
return 4;
}
else if (Rs >= 165) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 317) {
if (RsDiff < 4) {
if (ssDCval < 73) {
if (mcTcor < 89) {
if (refDCval < 41) {
if (Cs < 155) {
if (Rs < 732) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (Rs >= 732) {
return 2;
}
}
else if (Cs >= 155) {
if (mcTcor < 71) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 4;
}
}
else if (mcTcor >= 71) {
if (negBalance < 41) {
if (SC < 621) {
return 8;
}
else if (SC >= 621) {
if (Cs < 339) {
if (posBalance < 61) {
return 1;
}
else if (posBalance >= 61) {
return 8;
}
}
else if (Cs >= 339) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (negBalance >= 41) {
if (CsDiff < 4) {
if (tcor < 70) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
return 8;
}
}
else if (tcor >= 70) {
return 8;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
}
else if (refDCval >= 41) {
if (mu_mv_mag_sq < 101) {
if (Cs < 157) {
if (diffAFD < 5) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
else if (diffAFD >= 5) {
return 8;
}
}
else if (Cs >= 157) {
if (MVDiff < 60) {
if (diffRsCsdiff < 6) {
if (CsDiff < 3) {
if (mcTcor < 77) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
else if (mcTcor >= 77) {
if (Rs < 660) {
if (posBalance < 57) {
if (Cs < 343) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (Cs >= 343) {
if (negBalance < 34) {
if (Cs < 380) {
return 4;
}
else if (Cs >= 380) {
return 2;
}
}
else if (negBalance >= 34) {
return 4;
}
}
}
else if (posBalance >= 57) {
return 4;
}
}
else if (Rs >= 660) {
if (Cs < 311) {
if (Rs < 753) {
return 1;
}
else if (Rs >= 753) {
return 2;
}
}
else if (Cs >= 311) {
return 8;
}
}
}
}
else if (CsDiff >= 3) {
if (negBalance < 47) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
return 1;
}
}
}
else if (negBalance >= 47) {
if (Cs < 281) {
return 2;
}
else if (Cs >= 281) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 6) {
if (tcor < 60) {
return 8;
}
else if (tcor >= 60) {
if (Cs < 337) {
return 8;
}
else if (Cs >= 337) {
return 1;
}
}
}
}
else if (MVDiff >= 60) {
if (mu_mv_mag_sq < 47) {
if (posBalance < 48) {
if (Cs < 509) {
return 1;
}
else if (Cs >= 509) {
return 8;
}
}
else if (posBalance >= 48) {
if (Rs < 371) {
if (Rs < 361) {
return 2;
}
else if (Rs >= 361) {
return 8;
}
}
else if (Rs >= 371) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (Cs < 297) {
return 2;
}
else if (Cs >= 297) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1305) {
if (TSC < 1) {
if (mu_mv_mag_sq < 105) {
if (posBalance < 6) {
return 8;
}
else if (posBalance >= 6) {
return 4;
}
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 789) {
return 8;
}
else if (Rs >= 789) {
if (Cs < 390) {
return 1;
}
else if (Cs >= 390) {
return 8;
}
}
}
}
else if (TSC >= 1) {
if (ssDCval < 59) {
if (mu_mv_mag_sq < 96) {
if (Cs < 128) {
if (mu_mv_mag_sq < 56) {
if (posBalance < 15) {
return 8;
}
else if (posBalance >= 15) {
if (Rs < 1006) {
if (tcor < 81) {
if (RsDiff < 3) {
if (ssDCval < 41) {
return 1;
}
else if (ssDCval >= 41) {
return 8;
}
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (tcor >= 81) {
if (Cs < 21) {
if (Rs < 397) {
return 1;
}
else if (Rs >= 397) {
return 8;
}
}
else if (Cs >= 21) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
if (refDCval < 56) {
return 8;
}
else if (refDCval >= 56) {
return 1;
}
}
}
}
}
else if (Rs >= 1006) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Cs < 86) {
return 8;
}
else if (Cs >= 86) {
if (Rs < 414) {
return 8;
}
else if (Rs >= 414) {
return 1;
}
}
}
}
}
else if (Cs >= 128) {
if (negBalance < 40) {
if (Rs < 695) {
if (AFD < 3) {
if (MVDiff < 5) {
if (refDCval < 54) {
if (mu_mv_mag_sq < 89) {
if (Rs < 583) {
return 4;
}
else if (Rs >= 583) {
return 2;
}
}
else if (mu_mv_mag_sq >= 89) {
return 1;
}
}
else if (refDCval >= 54) {
return 8;
}
}
else if (MVDiff >= 5) {
if (Rs < 525) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 405) {
return 4;
}
else if (Rs >= 405) {
return 8;
}
}
}
else if (Rs >= 525) {
if (Rs < 597) {
return 1;
}
else if (Rs >= 597) {
return 4;
}
}
}
}
else if (AFD >= 3) {
if (posBalance < 30) {
if (Cs < 319) {
return 2;
}
else if (Cs >= 319) {
return 1;
}
}
else if (posBalance >= 30) {
if (refDCval < 56) {
if (refDCval < 45) {
if (Cs < 500) {
if (mcTcor < 92) {
if (Rs < 535) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (mu_mv_mag_sq < 3) {
if (Rs < 364) {
return 1;
}
else if (Rs >= 364) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (Rs >= 535) {
return 1;
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 11) {
if (ssDCval < 41) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
return 4;
}
}
else if (ssDCval >= 41) {
return 4;
}
}
else if (mu_mv_mag_sq >= 11) {
if (posBalance < 35) {
if (Rs < 370) {
return 1;
}
else if (Rs >= 370) {
return 4;
}
}
else if (posBalance >= 35) {
return 1;
}
}
}
}
else if (Cs >= 500) {
return 8;
}
}
else if (refDCval >= 45) {
if (RsCsDiff < 5) {
if (SC < 688) {
if (tcor < 95) {
if (negBalance < 18) {
return 1;
}
else if (negBalance >= 18) {
return 4;
}
}
else if (tcor >= 95) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (SC >= 688) {
if (negBalance < 31) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (negBalance < 23) {
return 1;
}
else if (negBalance >= 23) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (SC < 851) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (Cs < 330) {
return 1;
}
else if (Cs >= 330) {
return 4;
}
}
}
else if (SC >= 851) {
return 1;
}
}
}
}
}
else if (negBalance >= 31) {
if (SC < 991) {
return 1;
}
else if (SC >= 991) {
if (Rs < 631) {
return 4;
}
else if (Rs >= 631) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 5) {
if (mcTcor < 92) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (mcTcor >= 92) {
if (Cs < 370) {
if (posBalance < 50) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (Rs < 551) {
return 8;
}
else if (Rs >= 551) {
return 2;
}
}
}
else if (posBalance >= 50) {
return 4;
}
}
else if (Cs >= 370) {
return 4;
}
}
}
}
}
else if (refDCval >= 56) {
if (mcTcor < 98) {
if (MVDiff < 4) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
return 8;
}
}
else if (MVDiff >= 4) {
if (mcTcor < 94) {
if (Cs < 160) {
return 2;
}
else if (Cs >= 160) {
return 1;
}
}
else if (mcTcor >= 94) {
if (refDCval < 57) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
else if (refDCval >= 57) {
return 2;
}
}
}
}
else if (mcTcor >= 98) {
if (MVDiff < 5) {
if (Cs < 280) {
return 2;
}
else if (Cs >= 280) {
return 8;
}
}
else if (MVDiff >= 5) {
if (Cs < 433) {
if (Rs < 471) {
if (Cs < 233) {
return 2;
}
else if (Cs >= 233) {
return 4;
}
}
else if (Rs >= 471) {
return 8;
}
}
else if (Cs >= 433) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 695) {
if (mu_mv_mag_sq < 57) {
if (posBalance < 37) {
return 2;
}
else if (posBalance >= 37) {
if (Cs < 483) {
if (mcTcor < 99) {
if (SC < 978) {
return 2;
}
else if (SC >= 978) {
return 1;
}
}
else if (mcTcor >= 99) {
return 2;
}
}
else if (Cs >= 483) {
if (Rs < 796) {
return 2;
}
else if (Rs >= 796) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (Cs < 236) {
if (Cs < 208) {
return 1;
}
else if (Cs >= 208) {
return 4;
}
}
else if (Cs >= 236) {
if (posBalance < 48) {
if (negBalance < 19) {
if (Rs < 772) {
return 2;
}
else if (Rs >= 772) {
return 1;
}
}
else if (negBalance >= 19) {
return 1;
}
}
else if (posBalance >= 48) {
if (Rs < 778) {
if (Rs < 720) {
return 1;
}
else if (Rs >= 720) {
return 2;
}
}
else if (Rs >= 778) {
if (Rs < 846) {
return 4;
}
else if (Rs >= 846) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 40) {
if (SC < 924) {
if (mu_mv_mag_sq < 49) {
if (Rs < 394) {
if (Cs < 363) {
if (refDCval < 41) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (refDCval >= 41) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (MVDiff < 92) {
if (Cs < 271) {
return 1;
}
else if (Cs >= 271) {
return 8;
}
}
else if (MVDiff >= 92) {
return 4;
}
}
}
}
else if (Cs >= 363) {
return 2;
}
}
else if (Rs >= 394) {
if (MVDiff < 67) {
if (negBalance < 41) {
if (Rs < 492) {
if (Rs < 439) {
return 1;
}
else if (Rs >= 439) {
return 4;
}
}
else if (Rs >= 492) {
return 8;
}
}
else if (negBalance >= 41) {
if (diffMVdiffVal < 2) {
if (diffRsCsdiff < -9) {
return 8;
}
else if (diffRsCsdiff >= -9) {
if (AFD < 9) {
if (TSC < 4) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
return 8;
}
}
else if (TSC >= 4) {
return 1;
}
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (diffMVdiffVal >= 2) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (negBalance < 47) {
return 4;
}
else if (negBalance >= 47) {
return 1;
}
}
}
}
}
else if (MVDiff >= 67) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (diffRsCsdiff < 3) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (diffRsCsdiff >= 3) {
if (Rs < 510) {
return 1;
}
else if (Rs >= 510) {
return 4;
}
}
}
}
}
else if (SC >= 924) {
if (refDCval < 38) {
if (Rs < 592) {
return 8;
}
else if (Rs >= 592) {
return 1;
}
}
else if (refDCval >= 38) {
if (Rs < 924) {
if (AFD < 11) {
if (ssDCval < 56) {
if (ssDCval < 38) {
return 1;
}
else if (ssDCval >= 38) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 1;
}
}
}
else if (AFD >= 11) {
if (Rs < 627) {
return 1;
}
else if (Rs >= 627) {
return 8;
}
}
}
else if (Rs >= 924) {
if (Cs < 394) {
return 1;
}
else if (Cs >= 394) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (tcor < 101) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 114) {
if (Rs < 436) {
if (SC < 576) {
return 8;
}
else if (SC >= 576) {
if (refDCval < 59) {
return 1;
}
else if (refDCval >= 59) {
return 2;
}
}
}
else if (Rs >= 436) {
if (Rs < 1047) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 493) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
else if (Rs >= 493) {
if (Rs < 966) {
return 8;
}
else if (Rs >= 966) {
return 4;
}
}
}
}
else if (Rs >= 1047) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (Cs < 293) {
if (mu_mv_mag_sq < 118) {
if (TSC < 2) {
if (Rs < 573) {
return 4;
}
else if (Rs >= 573) {
return 2;
}
}
else if (TSC >= 2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
else if (Cs >= 293) {
if (posBalance < 38) {
if (negBalance < 42) {
if (SC < 1136) {
if (Rs < 558) {
if (Rs < 453) {
return 8;
}
else if (Rs >= 453) {
return 1;
}
}
else if (Rs >= 558) {
if (Cs < 477) {
return 8;
}
else if (Cs >= 477) {
return 4;
}
}
}
else if (SC >= 1136) {
if (Rs < 1274) {
return 1;
}
else if (Rs >= 1274) {
if (Rs < 1294) {
return 8;
}
else if (Rs >= 1294) {
return 1;
}
}
}
}
else if (negBalance >= 42) {
return 4;
}
}
else if (posBalance >= 38) {
if (Rs < 558) {
return 1;
}
else if (Rs >= 558) {
if (ssDCval < 47) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
else if (ssDCval >= 47) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 101) {
if (SC < 758) {
if (Cs < 238) {
return 1;
}
else if (Cs >= 238) {
return 4;
}
}
else if (SC >= 758) {
return 8;
}
}
}
}
else if (ssDCval >= 59) {
if (Cs < 277) {
if (Rs < 548) {
if (TSC < 4) {
if (MVDiff < 8) {
if (posBalance < 25) {
if (Rs < 514) {
return 8;
}
else if (Rs >= 514) {
return 2;
}
}
else if (posBalance >= 25) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
if (Rs < 437) {
return 1;
}
else if (Rs >= 437) {
return 4;
}
}
}
}
else if (MVDiff >= 8) {
if (ssDCval < 64) {
if (Rs < 405) {
return 4;
}
else if (Rs >= 405) {
return 2;
}
}
else if (ssDCval >= 64) {
return 8;
}
}
}
else if (TSC >= 4) {
if (Cs < 123) {
return 1;
}
else if (Cs >= 123) {
if (AFD < 7) {
if (Cs < 230) {
return 2;
}
else if (Cs >= 230) {
return 1;
}
}
else if (AFD >= 7) {
if (SC < 508) {
return 8;
}
else if (SC >= 508) {
return 4;
}
}
}
}
}
else if (Rs >= 548) {
if (negBalance < 21) {
if (Rs < 681) {
return 8;
}
else if (Rs >= 681) {
if (Rs < 1092) {
if (tcor < 100) {
return 4;
}
else if (tcor >= 100) {
return 1;
}
}
else if (Rs >= 1092) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (negBalance >= 21) {
if (Cs < 195) {
if (TSC < 6) {
if (tcor < 100) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
if (Rs < 639) {
return 4;
}
else if (Rs >= 639) {
return 1;
}
}
}
else if (tcor >= 100) {
if (Cs < 91) {
return 4;
}
else if (Cs >= 91) {
if (Rs < 724) {
return 8;
}
else if (Rs >= 724) {
return 1;
}
}
}
}
else if (TSC >= 6) {
if (Rs < 774) {
return 1;
}
else if (Rs >= 774) {
return 2;
}
}
}
else if (Cs >= 195) {
if (TSC < 3) {
if (MVDiff < 11) {
if (mu_mv_mag_sq < 120) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 2;
}
}
else if (mu_mv_mag_sq >= 120) {
return 4;
}
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (TSC >= 3) {
if (ssDCval < 61) {
if (Rs < 1027) {
return 2;
}
else if (Rs >= 1027) {
return 1;
}
}
else if (ssDCval >= 61) {
if (negBalance < 46) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
else if (negBalance >= 46) {
if (Cs < 219) {
if (Rs < 724) {
return 4;
}
else if (Rs >= 724) {
return 8;
}
}
else if (Cs >= 219) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 277) {
if (mu_mv_mag_sq < 55) {
if (Rs < 516) {
if (tcor < 93) {
if (refDCval < 65) {
if (mcTcor < 98) {
if (AFD < 10) {
if (Cs < 501) {
return 2;
}
else if (Cs >= 501) {
return 4;
}
}
else if (AFD >= 10) {
if (RsDiff < 3) {
if (AFD < 12) {
if (Cs < 434) {
return 4;
}
else if (Cs >= 434) {
return 1;
}
}
else if (AFD >= 12) {
return 1;
}
}
else if (RsDiff >= 3) {
if (Cs < 476) {
return 2;
}
else if (Cs >= 476) {
return 1;
}
}
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (refDCval >= 65) {
if (Rs < 398) {
if (TSC < 4) {
if (Rs < 320) {
return 1;
}
else if (Rs >= 320) {
return 4;
}
}
else if (TSC >= 4) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (Rs >= 398) {
if (mu_mv_mag_sq < 10) {
return 4;
}
else if (mu_mv_mag_sq >= 10) {
if (Cs < 395) {
if (Rs < 453) {
return 8;
}
else if (Rs >= 453) {
return 4;
}
}
else if (Cs >= 395) {
return 2;
}
}
}
}
}
else if (tcor >= 93) {
if (refDCval < 67) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (refDCval >= 67) {
return 1;
}
}
}
else if (Rs >= 516) {
if (Rs < 794) {
if (RsDiff < 2) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (Rs < 629) {
return 1;
}
else if (Rs >= 629) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (diffMVdiffVal < -11) {
return 1;
}
else if (diffMVdiffVal >= -11) {
return 4;
}
}
}
else if (Rs >= 794) {
if (mcTcor < 96) {
if (Cs < 404) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
if (Cs < 372) {
return 4;
}
else if (Cs >= 372) {
return 8;
}
}
}
else if (Cs >= 404) {
return 4;
}
}
else if (mcTcor >= 96) {
if (Cs < 282) {
return 4;
}
else if (Cs >= 282) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (refDCval < 72) {
if (posBalance < 41) {
if (RsDiff < 3) {
if (mu_mv_mag_sq < 119) {
if (tcor < 98) {
if (mcTcor < 96) {
return 4;
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (tcor >= 98) {
if (ssDCval < 67) {
if (Rs < 675) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Rs < 517) {
if (Cs < 443) {
if (SC < 827) {
return 8;
}
else if (SC >= 827) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
}
else if (Cs >= 443) {
return 4;
}
}
else if (Rs >= 517) {
return 8;
}
}
}
else if (Rs >= 675) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
}
else if (ssDCval >= 67) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
if (Rs < 647) {
return 2;
}
else if (Rs >= 647) {
if (Rs < 901) {
return 4;
}
else if (Rs >= 901) {
return 1;
}
}
}
}
}
else if (RsDiff >= 3) {
if (Rs < 1256) {
return 2;
}
else if (Rs >= 1256) {
return 1;
}
}
}
else if (posBalance >= 41) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 106) {
if (tcor < 98) {
if (diffTSC < 1) {
if (negBalance < 36) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
if (mu_mv_mag_sq < 99) {
return 1;
}
else if (mu_mv_mag_sq >= 99) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 564) {
return 8;
}
else if (Rs >= 564) {
return 1;
}
}
}
}
}
else if (negBalance >= 36) {
if (Cs < 438) {
return 1;
}
else if (Cs >= 438) {
return 2;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 474) {
return 2;
}
else if (Rs >= 474) {
if (SC < 992) {
return 1;
}
else if (SC >= 992) {
return 8;
}
}
}
}
else if (tcor >= 98) {
if (CsDiff < 1) {
if (SC < 1024) {
if (Cs < 375) {
return 8;
}
else if (Cs >= 375) {
return 1;
}
}
else if (SC >= 1024) {
if (Cs < 332) {
return 1;
}
else if (Cs >= 332) {
if (Rs < 654) {
return 8;
}
else if (Rs >= 654) {
if (Rs < 993) {
return 2;
}
else if (Rs >= 993) {
return 8;
}
}
}
}
}
else if (CsDiff >= 1) {
if (negBalance < 24) {
return 8;
}
else if (negBalance >= 24) {
if (Rs < 710) {
return 8;
}
else if (Rs >= 710) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Cs < 359) {
if (Rs < 447) {
if (Rs < 329) {
return 4;
}
else if (Rs >= 329) {
return 1;
}
}
else if (Rs >= 447) {
if (Rs < 1047) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
return 4;
}
}
else if (Rs >= 1047) {
return 1;
}
}
}
else if (Cs >= 359) {
if (ssDCval < 66) {
if (Rs < 723) {
return 1;
}
else if (Rs >= 723) {
return 8;
}
}
else if (ssDCval >= 66) {
if (mu_mv_mag_sq < 122) {
if (posBalance < 43) {
if (Rs < 487) {
return 8;
}
else if (Rs >= 487) {
return 1;
}
}
else if (posBalance >= 43) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (CsDiff < 1) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 2;
}
}
else if (CsDiff >= 1) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (MVDiff < 1) {
if (Rs < 416) {
return 2;
}
else if (Rs >= 416) {
return 1;
}
}
else if (MVDiff >= 1) {
if (Rs < 588) {
return 1;
}
else if (Rs >= 588) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 61) {
if (Rs < 528) {
if (AFD < 6) {
if (negBalance < 20) {
return 4;
}
else if (negBalance >= 20) {
return 2;
}
}
else if (AFD >= 6) {
if (Rs < 378) {
return 4;
}
else if (Rs >= 378) {
return 1;
}
}
}
else if (Rs >= 528) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (ssDCval >= 61) {
if (tcor < 94) {
if (mcTcor < 96) {
if (posBalance < 58) {
if (posBalance < 51) {
if (Rs < 687) {
return 2;
}
else if (Rs >= 687) {
return 8;
}
}
else if (posBalance >= 51) {
return 4;
}
}
else if (posBalance >= 58) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
if (Cs < 485) {
return 2;
}
else if (Cs >= 485) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (tcor >= 94) {
if (mu_mv_mag_sq < 78) {
if (SC < 811) {
if (Rs < 397) {
return 8;
}
else if (Rs >= 397) {
return 1;
}
}
else if (SC >= 811) {
return 4;
}
}
else if (mu_mv_mag_sq >= 78) {
if (SC < 704) {
if (Cs < 334) {
return 2;
}
else if (Cs >= 334) {
return 4;
}
}
else if (SC >= 704) {
if (Rs < 1087) {
if (RsCsDiff < 1) {
if (Rs < 633) {
return 8;
}
else if (Rs >= 633) {
return 2;
}
}
else if (RsCsDiff >= 1) {
return 8;
}
}
else if (Rs >= 1087) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 72) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
if (MVDiff < 1) {
if (Rs < 812) {
return 2;
}
else if (Rs >= 812) {
return 1;
}
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 91) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 91) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1305) {
if (SC < 1604) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
}
else if (SC >= 1604) {
if (mu_mv_mag_sq < 128) {
if (mu_mv_mag_sq < 72) {
if (negBalance < 29) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (CsDiff < 2) {
if (ssDCval < 70) {
if (SC < 1644) {
if (Rs < 1435) {
return 2;
}
else if (Rs >= 1435) {
return 1;
}
}
else if (SC >= 1644) {
return 2;
}
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (CsDiff >= 2) {
if (Rs < 1806) {
return 1;
}
else if (Rs >= 1806) {
return 4;
}
}
}
}
else if (negBalance >= 29) {
if (posBalance < 63) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
else if (posBalance >= 63) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (Rs < 1645) {
if (negBalance < 20) {
if (ssDCval < 64) {
if (ssDCval < 61) {
if (refDCval < 56) {
if (Rs < 1468) {
return 1;
}
else if (Rs >= 1468) {
return 2;
}
}
else if (refDCval >= 56) {
if (Cs < 418) {
return 1;
}
else if (Cs >= 418) {
return 8;
}
}
}
else if (ssDCval >= 61) {
if (Rs < 1479) {
return 2;
}
else if (Rs >= 1479) {
return 1;
}
}
}
else if (ssDCval >= 64) {
if (refDCval < 64) {
if (Rs < 1478) {
return 2;
}
else if (Rs >= 1478) {
return 1;
}
}
else if (refDCval >= 64) {
return 1;
}
}
}
else if (negBalance >= 20) {
if (mu_mv_mag_sq < 124) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (AFD < 4) {
if (Cs < 494) {
if (refDCval < 58) {
return 2;
}
else if (refDCval >= 58) {
if (MVDiff < 6) {
if (negBalance < 24) {
if (Rs < 1454) {
return 8;
}
else if (Rs >= 1454) {
return 1;
}
}
else if (negBalance >= 24) {
return 8;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
}
else if (Cs >= 494) {
return 1;
}
}
else if (AFD >= 4) {
if (Rs < 1524) {
return 1;
}
else if (Rs >= 1524) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
}
else if (Rs >= 1645) {
if (Cs < 434) {
return 2;
}
else if (Cs >= 434) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1753) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
else if (Rs >= 1753) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 73) {
if (Cs < 285) {
if (ssDCval < 107) {
if (mu_mv_mag_sq < 114) {
if (Rs < 1613) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 83) {
if (MVDiff < 5) {
if (AFD < 4) {
if (TSC < 3) {
if (mu_mv_mag_sq < 77) {
return 8;
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 1592) {
return 8;
}
else if (Rs >= 1592) {
return 1;
}
}
}
else if (TSC >= 3) {
if (Rs < 909) {
return 4;
}
else if (Rs >= 909) {
return 1;
}
}
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 53) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
if (Rs < 393) {
return 1;
}
else if (Rs >= 393) {
if (Rs < 902) {
return 4;
}
else if (Rs >= 902) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (gchDC < 1) {
if (Rs < 727) {
return 4;
}
else if (Rs >= 727) {
return 1;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (MVDiff >= 5) {
if (tcor < 101) {
if (Rs < 842) {
if (ssDCval < 101) {
if (ssDCval < 77) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 513) {
return 2;
}
else if (Rs >= 513) {
return 8;
}
}
}
else if (ssDCval >= 77) {
if (MVDiff < 17) {
if (mu_mv_mag_sq < 36) {
return 1;
}
else if (mu_mv_mag_sq >= 36) {
if (SC < 461) {
if (Rs < 324) {
return 8;
}
else if (Rs >= 324) {
return 1;
}
}
else if (SC >= 461) {
return 8;
}
}
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (ssDCval >= 101) {
return 1;
}
}
else if (Rs >= 842) {
return 1;
}
}
else if (tcor >= 101) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (ssDCval < 80) {
if (SC < 1264) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (Rs < 769) {
if (Rs < 512) {
return 1;
}
else if (Rs >= 512) {
return 4;
}
}
else if (Rs >= 769) {
return 8;
}
}
}
else if (SC >= 1264) {
return 1;
}
}
else if (ssDCval >= 80) {
if (Cs < 183) {
return 1;
}
else if (Cs >= 183) {
if (Cs < 239) {
return 2;
}
else if (Cs >= 239) {
if (Rs < 745) {
return 8;
}
else if (Rs >= 745) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (tcor < 74) {
if (mu_mv_mag_sq < 57) {
if (mcTcor < 93) {
if (Rs < 345) {
if (ssDCval < 91) {
if (MVDiff < 29) {
return 8;
}
else if (MVDiff >= 29) {
return 2;
}
}
else if (ssDCval >= 91) {
return 2;
}
}
else if (Rs >= 345) {
return 2;
}
}
else if (mcTcor >= 93) {
return 8;
}
}
else if (mu_mv_mag_sq >= 57) {
return 4;
}
}
else if (tcor >= 74) {
if (CsDiff < 2) {
if (Cs < 221) {
if (Cs < 142) {
return 1;
}
else if (Cs >= 142) {
if (mcTcor < 99) {
if (Cs < 144) {
return 4;
}
else if (Cs >= 144) {
return 1;
}
}
else if (mcTcor >= 99) {
if (Rs < 533) {
return 1;
}
else if (Rs >= 533) {
return 4;
}
}
}
}
else if (Cs >= 221) {
if (mcTcor < 98) {
if (Rs < 354) {
return 2;
}
else if (Rs >= 354) {
return 1;
}
}
else if (mcTcor >= 98) {
if (Rs < 920) {
return 4;
}
else if (Rs >= 920) {
return 8;
}
}
}
}
else if (CsDiff >= 2) {
if (tcor < 90) {
if (Rs < 440) {
return 8;
}
else if (Rs >= 440) {
return 1;
}
}
else if (tcor >= 90) {
return 2;
}
}
}
}
}
else if (Rs >= 1613) {
if (ssDCval < 103) {
return 1;
}
else if (ssDCval >= 103) {
if (Rs < 1782) {
return 8;
}
else if (Rs >= 1782) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (ssDCval < 79) {
if (Cs < 204) {
return 8;
}
else if (Cs >= 204) {
return 1;
}
}
else if (ssDCval >= 79) {
if (posBalance < 46) {
if (negBalance < 21) {
return 2;
}
else if (negBalance >= 21) {
if (Cs < 241) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
else if (Cs >= 241) {
return 1;
}
}
}
else if (posBalance >= 46) {
if (Rs < 389) {
return 4;
}
else if (Rs >= 389) {
return 1;
}
}
}
}
}
else if (ssDCval >= 107) {
if (MVDiff < 20) {
if (diffRsCsdiff < -5) {
if (Rs < 830) {
return 2;
}
else if (Rs >= 830) {
return 4;
}
}
else if (diffRsCsdiff >= -5) {
if (Cs < 274) {
return 8;
}
else if (Cs >= 274) {
if (Rs < 403) {
return 1;
}
else if (Rs >= 403) {
return 8;
}
}
}
}
else if (MVDiff >= 20) {
if (Cs < 170) {
if (negBalance < 70) {
return 1;
}
else if (negBalance >= 70) {
if (Rs < 463) {
if (Rs < 446) {
return 2;
}
else if (Rs >= 446) {
return 1;
}
}
else if (Rs >= 463) {
return 8;
}
}
}
else if (Cs >= 170) {
if (posBalance < 60) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
return 4;
}
}
else if (posBalance >= 60) {
if (ssDCval < 116) {
return 8;
}
else if (ssDCval >= 116) {
return 2;
}
}
}
}
}
}
else if (Cs >= 285) {
if (mu_mv_mag_sq < 97) {
if (mcTcor < 95) {
if (MVDiff < 24) {
if (mcTcor < 92) {
if (mcTcor < 83) {
return 2;
}
else if (mcTcor >= 83) {
if (negBalance < 60) {
return 8;
}
else if (negBalance >= 60) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 8;
}
}
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 54) {
if (posBalance < 61) {
if (ssDCval < 112) {
if (tcor < 83) {
if (refDCval < 100) {
if (Rs < 364) {
return 1;
}
else if (Rs >= 364) {
return 8;
}
}
else if (refDCval >= 100) {
return 1;
}
}
else if (tcor >= 83) {
if (Rs < 385) {
if (Cs < 421) {
return 1;
}
else if (Cs >= 421) {
return 8;
}
}
else if (Rs >= 385) {
if (negBalance < 57) {
if (TSC < 4) {
if (Rs < 499) {
return 1;
}
else if (Rs >= 499) {
return 2;
}
}
else if (TSC >= 4) {
if (Rs < 518) {
return 2;
}
else if (Rs >= 518) {
return 1;
}
}
}
else if (negBalance >= 57) {
return 2;
}
}
}
}
else if (ssDCval >= 112) {
if (Rs < 374) {
return 2;
}
else if (Rs >= 374) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (Cs < 355) {
return 1;
}
else if (Cs >= 355) {
return 8;
}
}
}
}
else if (MVDiff >= 24) {
if (mcTcor < 87) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
if (MVDiff < 104) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
else if (MVDiff >= 104) {
return 4;
}
}
}
else if (mcTcor >= 87) {
if (RsDiff < 3) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Cs < 370) {
return 8;
}
else if (Cs >= 370) {
if (Rs < 346) {
return 2;
}
else if (Rs >= 346) {
return 4;
}
}
}
}
else if (RsDiff >= 3) {
if (tcor < 76) {
if (SC < 838) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 2;
}
}
else if (SC >= 838) {
if (ssDCval < 79) {
return 4;
}
else if (ssDCval >= 79) {
return 8;
}
}
}
else if (tcor >= 76) {
return 1;
}
}
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 116) {
if (tcor < 98) {
if (RsCsDiff < 2) {
if (Rs < 1459) {
if (Rs < 493) {
return 1;
}
else if (Rs >= 493) {
return 8;
}
}
else if (Rs >= 1459) {
return 1;
}
}
else if (RsCsDiff >= 2) {
if (refDCval < 85) {
if (mu_mv_mag_sq < 52) {
if (refDCval < 74) {
if (Rs < 1282) {
if (Cs < 521) {
return 8;
}
else if (Cs >= 521) {
return 4;
}
}
else if (Rs >= 1282) {
return 1;
}
}
else if (refDCval >= 74) {
if (Cs < 487) {
return 1;
}
else if (Cs >= 487) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (ssDCval < 83) {
if (posBalance < 49) {
if (tcor < 97) {
if (Rs < 1120) {
return 4;
}
else if (Rs >= 1120) {
return 1;
}
}
else if (tcor >= 97) {
if (Rs < 1035) {
return 1;
}
else if (Rs >= 1035) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (AFD < 10) {
if (AFD < 4) {
if (Rs < 1299) {
return 1;
}
else if (Rs >= 1299) {
return 8;
}
}
else if (AFD >= 4) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (AFD >= 10) {
if (Rs < 390) {
return 8;
}
else if (Rs >= 390) {
return 4;
}
}
}
}
else if (ssDCval >= 83) {
return 1;
}
}
}
else if (refDCval >= 85) {
if (TSC < 5) {
if (Cs < 488) {
if (CsDiff < 2) {
if (tcor < 94) {
if (negBalance < 37) {
if (Rs < 1191) {
return 4;
}
else if (Rs >= 1191) {
return 8;
}
}
else if (negBalance >= 37) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (tcor >= 94) {
if (ssDCval < 95) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (RsCsDiff < 16) {
return 1;
}
else if (RsCsDiff >= 16) {
return 4;
}
}
}
else if (Cs >= 488) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
return 1;
}
}
}
else if (TSC >= 5) {
if (gchDC < 5) {
if (mu_mv_mag_sq < 22) {
if (gchDC < 2) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
else if (gchDC >= 2) {
if (Cs < 507) {
return 8;
}
else if (Cs >= 507) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 22) {
return 8;
}
}
else if (gchDC >= 5) {
if (Cs < 346) {
return 4;
}
else if (Cs >= 346) {
return 2;
}
}
}
}
}
}
else if (tcor >= 98) {
if (MVDiff < 13) {
if (RsDiff < 2) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (ssDCval < 88) {
if (Rs < 752) {
return 1;
}
else if (Rs >= 752) {
return 8;
}
}
else if (ssDCval >= 88) {
if (gchDC < 4) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
return 2;
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 89) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
if (mu_mv_mag_sq < 71) {
if (Rs < 2102) {
return 8;
}
else if (Rs >= 2102) {
return 1;
}
}
else if (mu_mv_mag_sq >= 71) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (Cs < 449) {
if (Rs < 1736) {
return 4;
}
else if (Rs >= 1736) {
return 2;
}
}
else if (Cs >= 449) {
if (Cs < 491) {
return 1;
}
else if (Cs >= 491) {
return 8;
}
}
}
}
}
else if (MVDiff >= 13) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
}
}
else if (ssDCval >= 116) {
if (refDCval < 130) {
if (TSC < 3) {
if (ssDCval < 119) {
if (Rs < 433) {
return 1;
}
else if (Rs >= 433) {
return 2;
}
}
else if (ssDCval >= 119) {
return 1;
}
}
else if (TSC >= 3) {
if (Rs < 555) {
if (diffMVdiffVal < -41) {
return 4;
}
else if (diffMVdiffVal >= -41) {
if (AFD < 26) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (AFD >= 26) {
return 4;
}
}
}
else if (Rs >= 555) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (negBalance < 51) {
if (Rs < 1415) {
return 1;
}
else if (Rs >= 1415) {
return 2;
}
}
else if (negBalance >= 51) {
return 2;
}
}
}
}
}
else if (refDCval >= 130) {
if (AFD < 6) {
if (ssDCval < 159) {
if (TSCindex < 3) {
if (Rs < 537) {
return 1;
}
else if (Rs >= 537) {
return 8;
}
}
else if (TSCindex >= 3) {
if (Cs < 382) {
if (ssDCval < 144) {
return 4;
}
else if (ssDCval >= 144) {
return 8;
}
}
else if (Cs >= 382) {
return 8;
}
}
}
else if (ssDCval >= 159) {
return 4;
}
}
else if (AFD >= 6) {
if (Cs < 316) {
if (Rs < 377) {
return 2;
}
else if (Rs >= 377) {
return 4;
}
}
else if (Cs >= 316) {
if (TSC < 5) {
if (Rs < 575) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (Rs >= 575) {
if (Rs < 1071) {
return 4;
}
else if (Rs >= 1071) {
return 1;
}
}
}
else if (TSC >= 5) {
if (refDCval < 135) {
if (Rs < 874) {
return 8;
}
else if (Rs >= 874) {
return 1;
}
}
else if (refDCval >= 135) {
if (Cs < 526) {
return 2;
}
else if (Cs >= 526) {
return 4;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (SC < 1959) {
if (negBalance < 32) {
if (ssDCval < 94) {
if (Rs < 452) {
return 8;
}
else if (Rs >= 452) {
if (mcTcor < 99) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 1;
}
}
else if (mcTcor >= 99) {
if (Cs < 390) {
if (Rs < 580) {
return 1;
}
else if (Rs >= 580) {
if (Rs < 1120) {
return 8;
}
else if (Rs >= 1120) {
return 1;
}
}
}
else if (Cs >= 390) {
return 1;
}
}
}
}
else if (ssDCval >= 94) {
if (Cs < 452) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
return 1;
}
}
}
else if (Cs >= 452) {
if (MVDiff < 4) {
if (negBalance < 15) {
if (Rs < 635) {
return 8;
}
else if (Rs >= 635) {
return 4;
}
}
else if (negBalance >= 15) {
return 8;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
}
}
else if (negBalance >= 32) {
if (Rs < 555) {
if (Rs < 439) {
if (Cs < 396) {
return 1;
}
else if (Cs >= 396) {
return 8;
}
}
else if (Rs >= 439) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (Rs >= 555) {
if (Cs < 431) {
return 8;
}
else if (Cs >= 431) {
if (Rs < 704) {
if (Cs < 497) {
return 1;
}
else if (Cs >= 497) {
return 8;
}
}
else if (Rs >= 704) {
return 8;
}
}
}
}
}
else if (SC >= 1959) {
if (Rs < 2044) {
if (Cs < 497) {
if (ssDCval < 114) {
return 1;
}
else if (ssDCval >= 114) {
return 4;
}
}
else if (Cs >= 497) {
if (ssDCval < 106) {
return 4;
}
else if (ssDCval >= 106) {
return 1;
}
}
}
else if (Rs >= 2044) {
if (Cs < 444) {
return 2;
}
else if (Cs >= 444) {
if (Rs < 2094) {
return 1;
}
else if (Rs >= 2094) {
return 8;
}
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (Cs < 387) {
if (RsCsDiff < 51) {
if (negBalance < 43) {
if (MVDiff < 13) {
if (Rs < 861) {
return 1;
}
else if (Rs >= 861) {
return 2;
}
}
else if (MVDiff >= 13) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
return 1;
}
}
}
else if (negBalance >= 43) {
if (posBalance < 49) {
if (ssDCval < 37) {
return 2;
}
else if (ssDCval >= 37) {
if (AFD < 36) {
return 1;
}
else if (AFD >= 36) {
return 8;
}
}
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 52) {
if (Cs < 295) {
if (RsDiff < 6) {
if (tcor < 53) {
if (SC < 678) {
if (Cs < 54) {
return 8;
}
else if (Cs >= 54) {
return 1;
}
}
else if (SC >= 678) {
if (MVDiff < 37) {
return 8;
}
else if (MVDiff >= 37) {
return 2;
}
}
}
else if (tcor >= 53) {
if (diffMVdiffVal < 4) {
if (refDCval < 59) {
if (SC < 751) {
return 4;
}
else if (SC >= 751) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
else if (refDCval >= 59) {
if (CsDiff < 2) {
if (Rs < 492) {
return 8;
}
else if (Rs >= 492) {
return 2;
}
}
else if (CsDiff >= 2) {
if (diffRsCsdiff < -10) {
return 2;
}
else if (diffRsCsdiff >= -10) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 4) {
if (Cs < 133) {
return 1;
}
else if (Cs >= 133) {
return 8;
}
}
}
}
else if (RsDiff >= 6) {
if (Cs < 97) {
if (Cs < 82) {
if (Rs < 725) {
return 1;
}
else if (Rs >= 725) {
return 4;
}
}
else if (Cs >= 82) {
return 8;
}
}
else if (Cs >= 97) {
return 1;
}
}
}
else if (Cs >= 295) {
if (posBalance < 62) {
if (MVDiff < 22) {
if (Rs < 801) {
if (Rs < 670) {
return 8;
}
else if (Rs >= 670) {
return 1;
}
}
else if (Rs >= 801) {
return 8;
}
}
else if (MVDiff >= 22) {
if (Rs < 489) {
if (Cs < 347) {
if (Rs < 407) {
return 1;
}
else if (Rs >= 407) {
return 4;
}
}
else if (Cs >= 347) {
return 8;
}
}
else if (Rs >= 489) {
return 2;
}
}
}
else if (posBalance >= 62) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (mcTcor < 82) {
if (AFD < 19) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (AFD >= 19) {
if (Cs < 330) {
return 8;
}
else if (Cs >= 330) {
return 2;
}
}
}
else if (mcTcor >= 82) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 51) {
if (MVDiff < 31) {
if (Rs < 1421) {
if (ssDCval < 82) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (ssDCval >= 82) {
if (diffMVdiffVal < -14) {
return 1;
}
else if (diffMVdiffVal >= -14) {
return 2;
}
}
}
else if (Rs >= 1421) {
if (TSC < 10) {
if (SC < 2485) {
if (ssDCval < 41) {
return 1;
}
else if (ssDCval >= 41) {
return 2;
}
}
else if (SC >= 2485) {
if (diffTSC < -1) {
if (Rs < 2282) {
return 1;
}
else if (Rs >= 2282) {
return 2;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (MVDiff >= 31) {
if (gchDC < 6) {
if (posBalance < 56) {
if (mu_mv_mag_sq < 30) {
if (Rs < 409) {
return 4;
}
else if (Rs >= 409) {
if (mu_mv_mag_sq < 7) {
if (Rs < 1536) {
return 1;
}
else if (Rs >= 1536) {
if (Rs < 1808) {
return 2;
}
else if (Rs >= 1808) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 7) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (Rs < 1269) {
return 2;
}
else if (Rs >= 1269) {
if (RsCsDiff < 182) {
return 1;
}
else if (RsCsDiff >= 182) {
if (diffMVdiffVal < 1) {
return 1;
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
}
}
}
else if (posBalance >= 56) {
if (diffAFD < 5) {
if (tcor < 68) {
if (RsDiff < 12) {
if (ssDCval < 83) {
if (Rs < 639) {
if (Cs < 328) {
return 8;
}
else if (Cs >= 328) {
return 2;
}
}
else if (Rs >= 639) {
if (Rs < 676) {
return 4;
}
else if (Rs >= 676) {
if (MVDiff < 54) {
return 4;
}
else if (MVDiff >= 54) {
return 1;
}
}
}
}
else if (ssDCval >= 83) {
if (MVDiff < 51) {
if (Cs < 239) {
if (Cs < 133) {
return 8;
}
else if (Cs >= 133) {
return 1;
}
}
else if (Cs >= 239) {
return 2;
}
}
else if (MVDiff >= 51) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 2;
}
}
}
}
else if (RsDiff >= 12) {
if (tcor < 20) {
return 4;
}
else if (tcor >= 20) {
if (Cs < 229) {
return 2;
}
else if (Cs >= 229) {
if (MVDiff < 139) {
return 4;
}
else if (MVDiff >= 139) {
return 2;
}
}
}
}
}
else if (tcor >= 68) {
if (posBalance < 69) {
if (Rs < 1682) {
return 1;
}
else if (Rs >= 1682) {
return 8;
}
}
else if (posBalance >= 69) {
return 2;
}
}
}
else if (diffAFD >= 5) {
if (diffAFD < 9) {
if (AFD < 28) {
return 1;
}
else if (AFD >= 28) {
return 2;
}
}
else if (diffAFD >= 9) {
return 1;
}
}
}
}
else if (gchDC >= 6) {
if (Cs < 207) {
return 1;
}
else if (Cs >= 207) {
if (Cs < 274) {
if (gchDC < 12) {
if (AFD < 30) {
return 2;
}
else if (AFD >= 30) {
return 1;
}
}
else if (gchDC >= 12) {
return 8;
}
}
else if (Cs >= 274) {
if (Rs < 2049) {
if (diffRsCsdiff < -26) {
if (Rs < 524) {
return 1;
}
else if (Rs >= 524) {
return 4;
}
}
else if (diffRsCsdiff >= -26) {
return 1;
}
}
else if (Rs >= 2049) {
if (Rs < 2288) {
return 8;
}
else if (Rs >= 2288) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 387) {
if (negBalance < 51) {
if (negBalance < 39) {
if (mu_mv_mag_sq < 69) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (tcor < 67) {
return 1;
}
else if (tcor >= 67) {
if (ssDCval < 51) {
return 2;
}
else if (ssDCval >= 51) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (Cs < 486) {
return 8;
}
else if (Cs >= 486) {
if (Cs < 528) {
return 1;
}
else if (Cs >= 528) {
return 8;
}
}
}
}
else if (negBalance >= 39) {
if (SC < 1906) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Cs < 470) {
if (negBalance < 47) {
if (mcTcor < 86) {
if (Rs < 512) {
return 4;
}
else if (Rs >= 512) {
if (AFD < 43) {
return 8;
}
else if (AFD >= 43) {
return 4;
}
}
}
else if (mcTcor >= 86) {
if (Rs < 809) {
return 1;
}
else if (Rs >= 809) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (diffMVdiffVal < -6) {
if (Cs < 413) {
return 8;
}
else if (Cs >= 413) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
}
else if (Cs >= 470) {
if (diffMVdiffVal < -10) {
return 8;
}
else if (diffMVdiffVal >= -10) {
if (Rs < 606) {
return 2;
}
else if (Rs >= 606) {
if (posBalance < 60) {
if (Cs < 507) {
return 2;
}
else if (Cs >= 507) {
return 8;
}
}
else if (posBalance >= 60) {
if (Rs < 731) {
return 8;
}
else if (Rs >= 731) {
if (Rs < 1009) {
return 1;
}
else if (Rs >= 1009) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 1906) {
if (ssDCval < 82) {
return 2;
}
else if (ssDCval >= 82) {
if (Cs < 465) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 2;
}
}
else if (Cs >= 465) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 51) {
if (AFD < 22) {
if (mcTcor < 83) {
if (Cs < 463) {
if (MVDiff < 34) {
return 8;
}
else if (MVDiff >= 34) {
if (Cs < 430) {
return 8;
}
else if (Cs >= 430) {
return 4;
}
}
}
else if (Cs >= 463) {
if (AFD < 18) {
if (posBalance < 58) {
if (Rs < 1572) {
if (Rs < 706) {
return 4;
}
else if (Rs >= 706) {
return 8;
}
}
else if (Rs >= 1572) {
return 1;
}
}
else if (posBalance >= 58) {
return 4;
}
}
else if (AFD >= 18) {
if (Rs < 616) {
return 2;
}
else if (Rs >= 616) {
return 8;
}
}
}
}
else if (mcTcor >= 83) {
if (AFD < 14) {
if (MVDiff < 47) {
if (Rs < 1434) {
return 8;
}
else if (Rs >= 1434) {
return 4;
}
}
else if (MVDiff >= 47) {
return 2;
}
}
else if (AFD >= 14) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (TSC < 6) {
if (MVDiff < 114) {
return 1;
}
else if (MVDiff >= 114) {
return 2;
}
}
else if (TSC >= 6) {
if (posBalance < 53) {
if (Rs < 532) {
return 1;
}
else if (Rs >= 532) {
if (Rs < 1456) {
if (negBalance < 66) {
return 4;
}
else if (negBalance >= 66) {
return 8;
}
}
else if (Rs >= 1456) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (mcTcor < 90) {
if (negBalance < 58) {
return 8;
}
else if (negBalance >= 58) {
if (Cs < 512) {
return 4;
}
else if (Cs >= 512) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (mu_mv_mag_sq < 12) {
return 1;
}
else if (mu_mv_mag_sq >= 12) {
return 8;
}
}
}
}
}
}
}
}
else if (AFD >= 22) {
if (ssDCval < 72) {
if (tcor < 49) {
if (CsDiff < 6) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 2;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (tcor >= 49) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
}
}
else if (ssDCval >= 72) {
if (MVDiff < 24) {
return 4;
}
else if (MVDiff >= 24) {
if (tcor < 67) {
if (RsCsDiff < 213) {
if (TSC < 11) {
if (Cs < 411) {
return 1;
}
else if (Cs >= 411) {
if (negBalance < 58) {
if (Rs < 957) {
return 1;
}
else if (Rs >= 957) {
if (Cs < 451) {
return 2;
}
else if (Cs >= 451) {
return 4;
}
}
}
else if (negBalance >= 58) {
return 2;
}
}
}
else if (TSC >= 11) {
if (Cs < 402) {
return 8;
}
else if (Cs >= 402) {
return 1;
}
}
}
else if (RsCsDiff >= 213) {
if (AFD < 32) {
if (Rs < 1759) {
return 8;
}
else if (Rs >= 1759) {
return 4;
}
}
else if (AFD >= 32) {
return 2;
}
}
}
else if (tcor >= 67) {
if (RsDiff < 13) {
if (ssDCval < 94) {
return 1;
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (RsDiff >= 13) {
if (MVDiff < 33) {
return 2;
}
else if (MVDiff >= 33) {
return 1;
}
}
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 82) {
if (RsDiff < 9) {
if (refDCval < 90) {
if (Cs < 294) {
if (mcTcor < 101) {
if (refDCval < 86) {
if (negBalance < 37) {
if (posBalance < 43) {
if (Rs < 1880) {
return 2;
}
else if (Rs >= 1880) {
return 1;
}
}
else if (posBalance >= 43) {
if (RsDiff < 8) {
if (diffAFD < 4) {
return 1;
}
else if (diffAFD >= 4) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (RsDiff >= 8) {
if (Cs < 164) {
return 2;
}
else if (Cs >= 164) {
return 1;
}
}
}
}
else if (negBalance >= 37) {
if (diffTSC < -5) {
if (Rs < 654) {
return 8;
}
else if (Rs >= 654) {
return 1;
}
}
else if (diffTSC >= -5) {
if (AFD < 16) {
if (Cs < 206) {
return 1;
}
else if (Cs >= 206) {
if (MVDiff < 30) {
if (posBalance < 39) {
if (Cs < 242) {
return 1;
}
else if (Cs >= 242) {
return 2;
}
}
else if (posBalance >= 39) {
return 1;
}
}
else if (MVDiff >= 30) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (MVDiff < 39) {
if (Rs < 950) {
return 4;
}
else if (Rs >= 950) {
return 1;
}
}
else if (MVDiff >= 39) {
return 1;
}
}
}
}
}
else if (AFD >= 16) {
if (Rs < 806) {
if (Rs < 461) {
if (Rs < 361) {
return 8;
}
else if (Rs >= 361) {
return 1;
}
}
else if (Rs >= 461) {
return 8;
}
}
else if (Rs >= 806) {
return 1;
}
}
}
}
}
else if (refDCval >= 86) {
if (AFD < 14) {
if (tcor < 91) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (tcor >= 91) {
return 1;
}
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (mcTcor >= 101) {
if (Rs < 802) {
return 8;
}
else if (Rs >= 802) {
return 2;
}
}
}
else if (Cs >= 294) {
if (ssDCval < 73) {
if (Rs < 1161) {
if (posBalance < 55) {
if (negBalance < 44) {
if (diffMVdiffVal < 12) {
if (MVDiff < 9) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
else if (MVDiff >= 9) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
if (Cs < 390) {
return 1;
}
else if (Cs >= 390) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
}
else if (diffMVdiffVal >= 12) {
return 8;
}
}
else if (negBalance >= 44) {
if (mu_mv_mag_sq < 34) {
if (Rs < 698) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
else if (Rs >= 698) {
if (Rs < 828) {
return 8;
}
else if (Rs >= 828) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (gchDC < 3) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (gchDC >= 3) {
return 4;
}
}
}
}
else if (posBalance >= 55) {
if (diffAFD < -3) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (diffAFD >= -3) {
if (Cs < 442) {
if (Rs < 642) {
if (Rs < 521) {
return 1;
}
else if (Rs >= 521) {
return 4;
}
}
else if (Rs >= 642) {
return 1;
}
}
else if (Cs >= 442) {
if (mcTcor < 95) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 4;
}
}
else if (mcTcor >= 95) {
if (Cs < 526) {
return 1;
}
else if (Cs >= 526) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1161) {
if (Rs < 1836) {
if (Cs < 297) {
return 4;
}
else if (Cs >= 297) {
if (mu_mv_mag_sq < 36) {
if (mcTcor < 97) {
return 1;
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 36) {
if (mu_mv_mag_sq < 47) {
if (MVDiff < 9) {
if (Rs < 1386) {
return 8;
}
else if (Rs >= 1386) {
return 4;
}
}
else if (MVDiff >= 9) {
return 2;
}
}
else if (mu_mv_mag_sq >= 47) {
if (ssDCval < 63) {
if (Cs < 508) {
return 1;
}
else if (Cs >= 508) {
return 2;
}
}
else if (ssDCval >= 63) {
if (Cs < 410) {
return 4;
}
else if (Cs >= 410) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1836) {
if (posBalance < 46) {
if (refDCval < 69) {
return 4;
}
else if (refDCval >= 69) {
return 1;
}
}
else if (posBalance >= 46) {
if (negBalance < 40) {
if (Rs < 2058) {
return 2;
}
else if (Rs >= 2058) {
return 1;
}
}
else if (negBalance >= 40) {
return 2;
}
}
}
}
}
else if (ssDCval >= 73) {
if (mcTcor < 94) {
if (negBalance < 48) {
if (refDCval < 72) {
return 4;
}
else if (refDCval >= 72) {
if (Cs < 516) {
return 1;
}
else if (Cs >= 516) {
return 2;
}
}
}
else if (negBalance >= 48) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 730) {
return 8;
}
else if (Rs >= 730) {
return 2;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 640) {
if (Rs < 567) {
return 1;
}
else if (Rs >= 567) {
if (Rs < 579) {
return 4;
}
else if (Rs >= 579) {
return 8;
}
}
}
else if (Rs >= 640) {
if (posBalance < 70) {
return 1;
}
else if (posBalance >= 70) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 90) {
if (gchDC < 2) {
if (refDCval < 139) {
if (mcTcor < 98) {
if (Cs < 427) {
if (Cs < 224) {
if (tcor < 89) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (Rs < 807) {
return 1;
}
else if (Rs >= 807) {
if (Rs < 1681) {
return 2;
}
else if (Rs >= 1681) {
return 1;
}
}
}
}
else if (tcor >= 89) {
return 1;
}
}
else if (Cs >= 224) {
if (mcTcor < 94) {
if (Cs < 350) {
return 8;
}
else if (Cs >= 350) {
return 2;
}
}
else if (mcTcor >= 94) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (Rs < 735) {
if (Rs < 617) {
return 4;
}
else if (Rs >= 617) {
return 8;
}
}
else if (Rs >= 735) {
return 1;
}
}
}
}
}
else if (Cs >= 427) {
if (negBalance < 47) {
if (Rs < 1855) {
return 2;
}
else if (Rs >= 1855) {
return 1;
}
}
else if (negBalance >= 47) {
if (Rs < 1167) {
return 8;
}
else if (Rs >= 1167) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (diffRsCsdiff < -6) {
return 1;
}
else if (diffRsCsdiff >= -6) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 39) {
if (tcor < 87) {
if (Rs < 743) {
return 4;
}
else if (Rs >= 743) {
if (Rs < 1320) {
if (Rs < 951) {
return 8;
}
else if (Rs >= 951) {
return 1;
}
}
else if (Rs >= 1320) {
return 2;
}
}
}
else if (tcor >= 87) {
if (TSC < 4) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (TSC >= 4) {
if (diffRsCsdiff < 17) {
return 1;
}
else if (diffRsCsdiff >= 17) {
if (Rs < 1919) {
return 2;
}
else if (Rs >= 1919) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (Rs < 1906) {
return 1;
}
else if (Rs >= 1906) {
return 4;
}
}
}
}
else if (refDCval >= 139) {
if (posBalance < 50) {
if (Rs < 693) {
return 2;
}
else if (Rs >= 693) {
return 1;
}
}
else if (posBalance >= 50) {
if (refDCval < 146) {
if (Cs < 400) {
return 4;
}
else if (Cs >= 400) {
return 2;
}
}
else if (refDCval >= 146) {
if (Rs < 1080) {
return 8;
}
else if (Rs >= 1080) {
return 1;
}
}
}
}
}
else if (gchDC >= 2) {
if (Rs < 848) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (mcTcor < 97) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
if (MVDiff < 132) {
if (SC < 1096) {
if (Rs < 612) {
if (RsCsDiff < 37) {
if (Rs < 585) {
return 8;
}
else if (Rs >= 585) {
return 2;
}
}
else if (RsCsDiff >= 37) {
return 1;
}
}
else if (Rs >= 612) {
return 1;
}
}
else if (SC >= 1096) {
return 2;
}
}
else if (MVDiff >= 132) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (diffAFD < -4) {
if (Cs < 432) {
if (Rs < 559) {
return 8;
}
else if (Rs >= 559) {
return 1;
}
}
else if (Cs >= 432) {
return 2;
}
}
else if (diffAFD >= -4) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
return 8;
}
}
}
}
}
}
else if (Rs >= 848) {
if (gchDC < 3) {
if (posBalance < 59) {
if (SC < 1624) {
if (SC < 1529) {
return 1;
}
else if (SC >= 1529) {
if (ssDCval < 94) {
return 1;
}
else if (ssDCval >= 94) {
return 2;
}
}
}
else if (SC >= 1624) {
return 1;
}
}
else if (posBalance >= 59) {
if (ssDCval < 100) {
return 2;
}
else if (ssDCval >= 100) {
if (SC < 1996) {
return 1;
}
else if (SC >= 1996) {
if (diffRsCsdiff < -7) {
if (Rs < 1696) {
return 2;
}
else if (Rs >= 1696) {
return 1;
}
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
}
}
}
else if (gchDC >= 3) {
if (negBalance < 86) {
if (posBalance < 100) {
if (negBalance < 70) {
return 1;
}
else if (negBalance >= 70) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Rs < 1213) {
return 1;
}
else if (Rs >= 1213) {
return 8;
}
}
}
}
else if (posBalance >= 100) {
return 2;
}
}
else if (negBalance >= 86) {
if (Rs < 1297) {
return 2;
}
else if (Rs >= 1297) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 109) {
if (MVDiff < 38) {
if (AFD < 12) {
if (ssDCval < 100) {
if (MVDiff < 7) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
else if (MVDiff >= 7) {
if (Rs < 2229) {
return 1;
}
else if (Rs >= 2229) {
return 2;
}
}
}
else if (ssDCval >= 100) {
return 2;
}
}
else if (AFD >= 12) {
if (negBalance < 47) {
if (mu_mv_mag_sq < 28) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
if (tcor < 83) {
if (Cs < 248) {
return 1;
}
else if (Cs >= 248) {
return 2;
}
}
else if (tcor >= 83) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 28) {
return 2;
}
}
else if (negBalance >= 47) {
if (diffMVdiffVal < -13) {
if (negBalance < 53) {
if (Rs < 1672) {
return 1;
}
else if (Rs >= 1672) {
return 2;
}
}
else if (negBalance >= 53) {
return 1;
}
}
else if (diffMVdiffVal >= -13) {
if (AFD < 24) {
if (posBalance < 63) {
return 2;
}
else if (posBalance >= 63) {
if (Cs < 450) {
return 2;
}
else if (Cs >= 450) {
return 1;
}
}
}
else if (AFD >= 24) {
return 1;
}
}
}
}
}
else if (MVDiff >= 38) {
if (negBalance < 72) {
if (Rs < 2226) {
if (mu_mv_mag_sq < 31) {
if (Cs < 422) {
if (negBalance < 54) {
if (negBalance < 52) {
if (RsCsDiff < 216) {
return 1;
}
else if (RsCsDiff >= 216) {
return 2;
}
}
else if (negBalance >= 52) {
if (SC < 2358) {
return 2;
}
else if (SC >= 2358) {
return 1;
}
}
}
else if (negBalance >= 54) {
if (RsCsDiff < 86) {
if (Rs < 1552) {
return 1;
}
else if (Rs >= 1552) {
return 8;
}
}
else if (RsCsDiff >= 86) {
return 1;
}
}
}
else if (Cs >= 422) {
if (RsDiff < 13) {
return 2;
}
else if (RsDiff >= 13) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 31) {
if (Cs < 215) {
return 2;
}
else if (Cs >= 215) {
return 1;
}
}
}
else if (Rs >= 2226) {
if (Rs < 2283) {
return 4;
}
else if (Rs >= 2283) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
else if (negBalance >= 72) {
if (Cs < 220) {
return 2;
}
else if (Cs >= 220) {
return 8;
}
}
}
}
else if (refDCval >= 109) {
if (CsDiff < 2) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 23) {
if (Rs < 2298) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 1;
}
}
else if (Rs >= 2298) {
return 8;
}
}
else if (mu_mv_mag_sq >= 23) {
return 2;
}
}
}
else if (CsDiff >= 2) {
if (CsDiff < 3) {
if (Cs < 325) {
return 1;
}
else if (Cs >= 325) {
if (MVDiff < 61) {
if (refDCval < 117) {
if (Rs < 2015) {
return 1;
}
else if (Rs >= 2015) {
return 2;
}
}
else if (refDCval >= 117) {
return 1;
}
}
else if (MVDiff >= 61) {
if (TSC < 10) {
if (tcor < 82) {
return 8;
}
else if (tcor >= 82) {
return 1;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 61) {
if (mcTcor < 95) {
if (Rs < 1284) {
return 2;
}
else if (Rs >= 1284) {
if (posBalance < 78) {
return 1;
}
else if (posBalance >= 78) {
if (Rs < 1891) {
return 2;
}
else if (Rs >= 1891) {
return 1;
}
}
}
}
else if (mcTcor >= 95) {
if (TSC < 6) {
if (Rs < 1749) {
return 2;
}
else if (Rs >= 1749) {
return 1;
}
}
else if (TSC >= 6) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 61) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (negBalance < 53) {
if (posBalance < 51) {
if (Cs < 510) {
return 8;
}
else if (Cs >= 510) {
if (Rs < 1421) {
return 4;
}
else if (Rs >= 1421) {
return 1;
}
}
}
else if (posBalance >= 51) {
if (Rs < 1441) {
if (Cs < 54) {
return 2;
}
else if (Cs >= 54) {
return 1;
}
}
else if (Rs >= 1441) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
return 4;
}
}
}
}
else if (negBalance >= 53) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 534) {
if (negBalance < 14) {
if (Rs < 702) {
if (ssDCval < 57) {
if (tcor < 99) {
if (tcor < 81) {
if (Rs < 667) {
return 1;
}
else if (Rs >= 667) {
return 2;
}
}
else if (tcor >= 81) {
if (Rs < 336) {
if (Rs < 271) {
if (Rs < 211) {
return 4;
}
else if (Rs >= 211) {
return 2;
}
}
else if (Rs >= 271) {
return 1;
}
}
else if (Rs >= 336) {
return 8;
}
}
}
else if (tcor >= 99) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (SC < 1023) {
return 1;
}
else if (SC >= 1023) {
if (Rs < 413) {
return 2;
}
else if (Rs >= 413) {
return 8;
}
}
}
}
}
else if (ssDCval >= 57) {
if (SC < 1091) {
if (RsCsDiff < 1) {
if (Rs < 368) {
return 4;
}
else if (Rs >= 368) {
if (Rs < 467) {
return 1;
}
else if (Rs >= 467) {
return 8;
}
}
}
else if (RsCsDiff >= 1) {
if (refDCval < 130) {
return 1;
}
else if (refDCval >= 130) {
return 4;
}
}
}
else if (SC >= 1091) {
if (Rs < 657) {
if (refDCval < 68) {
if (Rs < 532) {
return 4;
}
else if (Rs >= 532) {
return 8;
}
}
else if (refDCval >= 68) {
if (Cs < 716) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (Rs < 557) {
return 1;
}
else if (Rs >= 557) {
return 8;
}
}
}
else if (Cs >= 716) {
if (Cs < 790) {
return 2;
}
else if (Cs >= 790) {
return 8;
}
}
}
}
else if (Rs >= 657) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (diffTSC < 1) {
if (Cs < 602) {
return 8;
}
else if (Cs >= 602) {
if (tcor < 100) {
return 1;
}
else if (tcor >= 100) {
return 8;
}
}
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
}
}
else if (Rs >= 702) {
if (diffTSC < -2) {
if (Rs < 1601) {
if (ssDCval < 144) {
if (ssDCval < 136) {
if (MVDiff < 8) {
if (diffTSC < -3) {
if (Cs < 883) {
return 8;
}
else if (Cs >= 883) {
return 1;
}
}
else if (diffTSC >= -3) {
return 4;
}
}
else if (MVDiff >= 8) {
if (diffRsCsdiff < -10) {
if (diffTSC < -9) {
if (diffAFD < -16) {
if (ssDCval < 83) {
if (Rs < 1166) {
return 8;
}
else if (Rs >= 1166) {
return 1;
}
}
else if (ssDCval >= 83) {
return 4;
}
}
else if (diffAFD >= -16) {
return 8;
}
}
else if (diffTSC >= -9) {
if (Cs < 571) {
return 2;
}
else if (Cs >= 571) {
if (posBalance < 3) {
if (ssDCval < 127) {
return 8;
}
else if (ssDCval >= 127) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
return 1;
}
}
}
else if (posBalance >= 3) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= -10) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
}
}
else if (ssDCval >= 136) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
if (diffRsCsdiff < -129) {
if (Rs < 1345) {
return 8;
}
else if (Rs >= 1345) {
return 4;
}
}
else if (diffRsCsdiff >= -129) {
return 4;
}
}
}
}
else if (ssDCval >= 144) {
if (diffAFD < -21) {
return 4;
}
else if (diffAFD >= -21) {
return 8;
}
}
}
else if (Rs >= 1601) {
if (Cs < 1262) {
return 8;
}
else if (Cs >= 1262) {
if (Rs < 1692) {
return 1;
}
else if (Rs >= 1692) {
return 4;
}
}
}
}
else if (diffTSC >= -2) {
if (Rs < 1810) {
if (Rs < 967) {
if (posBalance < 10) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
if (TSC < 1) {
return 2;
}
else if (TSC >= 1) {
return 8;
}
}
}
else if (posBalance >= 10) {
if (MVDiff < 2) {
if (ssDCval < 73) {
if (ssDCval < 61) {
if (Cs < 1096) {
return 8;
}
else if (Cs >= 1096) {
if (Rs < 864) {
return 1;
}
else if (Rs >= 864) {
return 8;
}
}
}
else if (ssDCval >= 61) {
return 2;
}
}
else if (ssDCval >= 73) {
if (RsCsDiff < 2) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (negBalance < 1) {
return 8;
}
else if (negBalance >= 1) {
if (Rs < 729) {
return 1;
}
else if (Rs >= 729) {
return 2;
}
}
}
}
else if (RsCsDiff >= 2) {
if (Rs < 756) {
return 1;
}
else if (Rs >= 756) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
else if (MVDiff >= 2) {
if (posBalance < 28) {
if (Rs < 850) {
if (ssDCval < 122) {
if (ssDCval < 30) {
return 8;
}
else if (ssDCval >= 30) {
return 2;
}
}
else if (ssDCval >= 122) {
return 8;
}
}
else if (Rs >= 850) {
return 8;
}
}
else if (posBalance >= 28) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (refDCval < 36) {
return 8;
}
else if (refDCval >= 36) {
if (mu_mv_mag_sq < 98) {
return 2;
}
else if (mu_mv_mag_sq >= 98) {
if (Cs < 603) {
return 2;
}
else if (Cs >= 603) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 967) {
if (SC < 1805) {
if (negBalance < 7) {
if (SC < 1676) {
return 8;
}
else if (SC >= 1676) {
if (ssDCval < 115) {
return 4;
}
else if (ssDCval >= 115) {
return 8;
}
}
}
else if (negBalance >= 7) {
return 8;
}
}
else if (SC >= 1805) {
if (negBalance < 10) {
if (refDCval < 92) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
if (tcor < 89) {
if (diffTSC < 1) {
if (Rs < 1568) {
return 4;
}
else if (Rs >= 1568) {
return 8;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (tcor >= 89) {
if (posBalance < 23) {
return 1;
}
else if (posBalance >= 23) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
else if (refDCval >= 92) {
if (ssDCval < 137) {
if (posBalance < 8) {
return 1;
}
else if (posBalance >= 8) {
if (Cs < 682) {
return 4;
}
else if (Cs >= 682) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
}
else if (ssDCval >= 137) {
if (diffMVdiffVal < -11) {
return 4;
}
else if (diffMVdiffVal >= -11) {
return 8;
}
}
}
}
else if (negBalance >= 10) {
if (ssDCval < 57) {
if (MVDiff < 6) {
if (Cs < 620) {
return 1;
}
else if (Cs >= 620) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
if (Rs < 1105) {
return 1;
}
else if (Rs >= 1105) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 1;
}
}
}
else if (ssDCval >= 57) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
if (Cs < 965) {
if (SC < 2033) {
if (SC < 1856) {
return 8;
}
else if (SC >= 1856) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
return 4;
}
}
}
else if (SC >= 2033) {
return 8;
}
}
else if (Cs >= 965) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (mu_mv_mag_sq < 120) {
return 8;
}
else if (mu_mv_mag_sq >= 120) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1620) {
return 8;
}
else if (Rs >= 1620) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 1810) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (ssDCval < 79) {
if (SC < 3008) {
if (Cs < 989) {
if (Cs < 654) {
return 2;
}
else if (Cs >= 654) {
return 1;
}
}
else if (Cs >= 989) {
if (Cs < 1066) {
if (SC < 2882) {
return 1;
}
else if (SC >= 2882) {
return 4;
}
}
else if (Cs >= 1066) {
return 1;
}
}
}
else if (SC >= 3008) {
if (Rs < 2247) {
return 4;
}
else if (Rs >= 2247) {
if (SC < 3567) {
return 8;
}
else if (SC >= 3567) {
return 2;
}
}
}
}
else if (ssDCval >= 79) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Rs < 1948) {
return 4;
}
else if (Rs >= 1948) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (mcTcor < 97) {
if (AFD < 14) {
if (posBalance < 46) {
if (ssDCval < 46) {
if (tcor < 58) {
return 8;
}
else if (tcor >= 58) {
if (negBalance < 32) {
if (RsCsDiff < 7) {
if (Rs < 565) {
if (Rs < 331) {
return 1;
}
else if (Rs >= 331) {
return 4;
}
}
else if (Rs >= 565) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
else if (RsCsDiff >= 7) {
if (Rs < 580) {
if (Rs < 550) {
return 1;
}
else if (Rs >= 550) {
return 8;
}
}
else if (Rs >= 580) {
if (Rs < 1572) {
if (ssDCval < 29) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (ssDCval >= 29) {
if (mcTcor < 91) {
if (negBalance < 26) {
if (Rs < 873) {
return 1;
}
else if (Rs >= 873) {
return 2;
}
}
else if (negBalance >= 26) {
return 8;
}
}
else if (mcTcor >= 91) {
if (negBalance < 22) {
if (Cs < 1143) {
return 1;
}
else if (Cs >= 1143) {
return 8;
}
}
else if (negBalance >= 22) {
if (Cs < 928) {
return 4;
}
else if (Cs >= 928) {
return 8;
}
}
}
}
}
else if (Rs >= 1572) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 1;
}
}
}
}
}
else if (negBalance >= 32) {
if (Rs < 1718) {
if (mcTcor < 96) {
if (Cs < 629) {
if (posBalance < 33) {
return 4;
}
else if (posBalance >= 33) {
if (Rs < 497) {
return 1;
}
else if (Rs >= 497) {
if (Cs < 613) {
return 8;
}
else if (Cs >= 613) {
return 1;
}
}
}
}
else if (Cs >= 629) {
if (AFD < 13) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (Rs < 1350) {
return 8;
}
else if (Rs >= 1350) {
return 2;
}
}
}
else if (AFD >= 13) {
if (Rs < 1335) {
return 1;
}
else if (Rs >= 1335) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 763) {
return 4;
}
else if (Rs >= 763) {
return 1;
}
}
}
else if (Rs >= 1718) {
if (Cs < 1193) {
return 4;
}
else if (Cs >= 1193) {
return 1;
}
}
}
}
}
else if (ssDCval >= 46) {
if (mu_mv_mag_sq < 94) {
if (Cs < 895) {
if (mcTcor < 93) {
if (Rs < 1512) {
if (ssDCval < 87) {
if (negBalance < 41) {
if (mu_mv_mag_sq < 39) {
if (Cs < 821) {
return 2;
}
else if (Cs >= 821) {
return 4;
}
}
else if (mu_mv_mag_sq >= 39) {
if (refDCval < 60) {
if (tcor < 85) {
if (Cs < 708) {
return 1;
}
else if (Cs >= 708) {
return 4;
}
}
else if (tcor >= 85) {
if (Rs < 1201) {
return 8;
}
else if (Rs >= 1201) {
return 1;
}
}
}
else if (refDCval >= 60) {
if (refDCval < 62) {
if (Cs < 805) {
return 2;
}
else if (Cs >= 805) {
return 8;
}
}
else if (refDCval >= 62) {
return 8;
}
}
}
}
else if (negBalance >= 41) {
if (tcor < 53) {
return 4;
}
else if (tcor >= 53) {
return 8;
}
}
}
else if (ssDCval >= 87) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
if (CsDiff < 4) {
if (ssDCval < 92) {
if (Rs < 665) {
return 8;
}
else if (Rs >= 665) {
return 2;
}
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (CsDiff >= 4) {
if (Rs < 353) {
return 2;
}
else if (Rs >= 353) {
return 8;
}
}
}
}
}
else if (Rs >= 1512) {
if (negBalance < 53) {
if (negBalance < 35) {
if (RsCsDiff < 49) {
if (Rs < 1940) {
return 2;
}
else if (Rs >= 1940) {
return 1;
}
}
else if (RsCsDiff >= 49) {
return 1;
}
}
else if (negBalance >= 35) {
if (tcor < 86) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
return 4;
}
}
else if (tcor >= 86) {
return 1;
}
}
}
else if (negBalance >= 53) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
if (posBalance < 31) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (posBalance >= 31) {
if (RsCsDiff < 3) {
if (mu_mv_mag_sq < 70) {
return 8;
}
else if (mu_mv_mag_sq >= 70) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
if (Rs < 1199) {
return 4;
}
else if (Rs >= 1199) {
return 8;
}
}
}
}
else if (RsCsDiff >= 3) {
if (Rs < 784) {
if (MVDiff < 38) {
if (ssDCval < 75) {
if (refDCval < 48) {
return 8;
}
else if (refDCval >= 48) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
if (Rs < 373) {
return 4;
}
else if (Rs >= 373) {
if (Cs < 564) {
return 1;
}
else if (Cs >= 564) {
return 2;
}
}
}
}
}
else if (ssDCval >= 75) {
if (tcor < 94) {
if (Cs < 609) {
return 8;
}
else if (Cs >= 609) {
return 2;
}
}
else if (tcor >= 94) {
if (Cs < 745) {
return 8;
}
else if (Cs >= 745) {
if (Rs < 608) {
return 4;
}
else if (Rs >= 608) {
if (Rs < 742) {
return 2;
}
else if (Rs >= 742) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 38) {
return 1;
}
}
else if (Rs >= 784) {
if (ssDCval < 130) {
if (ssDCval < 69) {
if (RsDiff < 4) {
if (negBalance < 38) {
if (MVDiff < 11) {
if (SC < 1464) {
return 2;
}
else if (SC >= 1464) {
if (CsDiff < 2) {
if (Rs < 1243) {
return 1;
}
else if (Rs >= 1243) {
return 8;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
}
else if (MVDiff >= 11) {
if (Rs < 1178) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Cs < 728) {
return 1;
}
else if (Cs >= 728) {
return 4;
}
}
}
else if (Rs >= 1178) {
if (Rs < 1287) {
return 8;
}
else if (Rs >= 1287) {
return 2;
}
}
}
}
else if (negBalance >= 38) {
if (gchDC < 1) {
if (Rs < 865) {
return 1;
}
else if (Rs >= 865) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (RsDiff >= 4) {
if (mu_mv_mag_sq < 55) {
if (Cs < 786) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
if (Rs < 1851) {
if (Rs < 1126) {
return 2;
}
else if (Rs >= 1126) {
return 1;
}
}
else if (Rs >= 1851) {
return 2;
}
}
}
else if (Cs >= 786) {
if (Rs < 1557) {
return 4;
}
else if (Rs >= 1557) {
if (Cs < 865) {
return 8;
}
else if (Cs >= 865) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (MVDiff < 9) {
if (negBalance < 32) {
if (Rs < 1862) {
return 2;
}
else if (Rs >= 1862) {
return 1;
}
}
else if (negBalance >= 32) {
return 4;
}
}
else if (MVDiff >= 9) {
return 1;
}
}
}
}
else if (ssDCval >= 69) {
if (tcor < 86) {
if (Rs < 1475) {
return 8;
}
else if (Rs >= 1475) {
return 4;
}
}
else if (tcor >= 86) {
if (Rs < 2238) {
if (Rs < 1352) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
if (Cs < 635) {
return 1;
}
else if (Cs >= 635) {
return 8;
}
}
}
else if (Rs >= 1352) {
return 1;
}
}
else if (Rs >= 2238) {
if (Rs < 2289) {
return 4;
}
else if (Rs >= 2289) {
return 2;
}
}
}
}
}
else if (ssDCval >= 130) {
if (ssDCval < 141) {
if (Rs < 2141) {
return 4;
}
else if (Rs >= 2141) {
return 8;
}
}
else if (ssDCval >= 141) {
if (Rs < 1483) {
return 2;
}
else if (Rs >= 1483) {
if (Rs < 2176) {
return 8;
}
else if (Rs >= 2176) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Cs >= 895) {
if (tcor < 89) {
if (ssDCval < 154) {
if (Rs < 1893) {
if (RsDiff < 5) {
if (negBalance < 35) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
else if (negBalance >= 35) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
if (Rs < 1195) {
return 4;
}
else if (Rs >= 1195) {
return 8;
}
}
}
}
else if (RsDiff >= 5) {
if (Rs < 1068) {
if (Rs < 1000) {
return 1;
}
else if (Rs >= 1000) {
return 4;
}
}
else if (Rs >= 1068) {
if (diffRsCsdiff < -40) {
if (Rs < 1758) {
return 8;
}
else if (Rs >= 1758) {
return 2;
}
}
else if (diffRsCsdiff >= -40) {
return 8;
}
}
}
}
else if (Rs >= 1893) {
if (MVDiff < 11) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 22) {
if (Rs < 2068) {
return 2;
}
else if (Rs >= 2068) {
if (Rs < 2161) {
return 4;
}
else if (Rs >= 2161) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 22) {
return 4;
}
}
else if (negBalance >= 44) {
return 8;
}
}
else if (MVDiff >= 11) {
if (gchDC < 3) {
if (Cs < 1250) {
if (MVDiff < 13) {
if (Rs < 2248) {
return 2;
}
else if (Rs >= 2248) {
return 8;
}
}
else if (MVDiff >= 13) {
return 2;
}
}
else if (Cs >= 1250) {
return 8;
}
}
else if (gchDC >= 3) {
if (Rs < 2251) {
return 4;
}
else if (Rs >= 2251) {
return 8;
}
}
}
}
}
else if (ssDCval >= 154) {
return 2;
}
}
else if (tcor >= 89) {
if (negBalance < 29) {
if (ssDCval < 61) {
if (ssDCval < 48) {
return 4;
}
else if (ssDCval >= 48) {
if (Cs < 1203) {
return 1;
}
else if (Cs >= 1203) {
return 2;
}
}
}
else if (ssDCval >= 61) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (negBalance >= 29) {
if (ssDCval < 92) {
if (SC < 1777) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
return 2;
}
}
else if (SC >= 1777) {
if (diffAFD < -2) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 1658) {
return 4;
}
else if (Rs >= 1658) {
return 2;
}
}
}
else if (diffAFD >= -2) {
if (Cs < 969) {
if (CsDiff < 4) {
if (AFD < 7) {
if (Rs < 1251) {
return 8;
}
else if (Rs >= 1251) {
return 1;
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (CsDiff >= 4) {
if (RsDiff < 6) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (RsDiff >= 6) {
if (Rs < 2057) {
return 2;
}
else if (Rs >= 2057) {
return 8;
}
}
}
}
else if (Cs >= 969) {
if (tcor < 93) {
if (Cs < 1171) {
return 4;
}
else if (Cs >= 1171) {
return 2;
}
}
else if (tcor >= 93) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (tcor < 94) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (tcor >= 94) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 92) {
if (mu_mv_mag_sq < 50) {
if (AFD < 8) {
if (Rs < 1152) {
return 2;
}
else if (Rs >= 1152) {
return 4;
}
}
else if (AFD >= 8) {
if (diffMVdiffVal < -10) {
return 2;
}
else if (diffMVdiffVal >= -10) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 50) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (refDCval < 56) {
if (Rs < 354) {
return 2;
}
else if (Rs >= 354) {
if (tcor < 85) {
if (Rs < 1572) {
return 4;
}
else if (Rs >= 1572) {
return 8;
}
}
else if (tcor >= 85) {
if (ssDCval < 52) {
if (posBalance < 40) {
if (tcor < 95) {
if (Cs < 1171) {
return 8;
}
else if (Cs >= 1171) {
return 4;
}
}
else if (tcor >= 95) {
return 4;
}
}
else if (posBalance >= 40) {
if (Cs < 1007) {
if (Rs < 1438) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
else if (Rs >= 1438) {
if (Rs < 2124) {
if (AFD < 5) {
if (Rs < 1746) {
return 4;
}
else if (Rs >= 1746) {
return 4;
}
}
else if (AFD >= 5) {
return 4;
}
}
else if (Rs >= 2124) {
return 8;
}
}
}
else if (Cs >= 1007) {
return 1;
}
}
}
else if (ssDCval >= 52) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 1;
}
}
}
}
}
else if (refDCval >= 56) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 114) {
if (TSC < 3) {
if (Rs < 1084) {
return 4;
}
else if (Rs >= 1084) {
return 8;
}
}
else if (TSC >= 3) {
if (SC < 2713) {
if (MVDiff < 1) {
if (Cs < 1097) {
return 8;
}
else if (Cs >= 1097) {
if (Rs < 901) {
return 2;
}
else if (Rs >= 901) {
return 4;
}
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (SC >= 2713) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
if (diffAFD < -1) {
if (ssDCval < 95) {
return 4;
}
else if (ssDCval >= 95) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
else if (MVDiff >= 4) {
if (AFD < 5) {
if (Rs < 812) {
return 4;
}
else if (Rs >= 812) {
if (diffRsCsdiff < -6) {
if (RsDiff < 7) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Rs < 1794) {
return 8;
}
else if (Rs >= 1794) {
return 4;
}
}
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
}
else if (AFD >= 5) {
if (RsCsDiff < 23) {
if (Rs < 478) {
return 1;
}
else if (Rs >= 478) {
return 8;
}
}
else if (RsCsDiff >= 23) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
return 8;
}
}
}
}
}
}
}
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 32) {
if (negBalance < 43) {
if (refDCval < 47) {
if (Rs < 825) {
if (negBalance < 30) {
if (refDCval < 34) {
if (Rs < 484) {
if (Rs < 321) {
return 1;
}
else if (Rs >= 321) {
return 8;
}
}
else if (Rs >= 484) {
return 2;
}
}
else if (refDCval >= 34) {
return 1;
}
}
else if (negBalance >= 30) {
if (negBalance < 36) {
if (Cs < 994) {
return 8;
}
else if (Cs >= 994) {
return 2;
}
}
else if (negBalance >= 36) {
if (MVDiff < 15) {
if (TSC < 5) {
if (Rs < 588) {
return 8;
}
else if (Rs >= 588) {
return 2;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (MVDiff >= 15) {
return 2;
}
}
}
}
else if (Rs >= 825) {
if (Cs < 615) {
if (Rs < 973) {
return 1;
}
else if (Rs >= 973) {
return 8;
}
}
else if (Cs >= 615) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
if (SC < 1867) {
if (mu_mv_mag_sq < 2) {
if (diffRsCsdiff < -2) {
if (Cs < 731) {
return 2;
}
else if (Cs >= 731) {
return 8;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
return 2;
}
}
else if (SC >= 1867) {
return 8;
}
}
}
}
}
else if (refDCval >= 47) {
if (SC < 2602) {
if (ssDCval < 64) {
if (tcor < 80) {
if (Rs < 632) {
if (tcor < 74) {
return 4;
}
else if (tcor >= 74) {
return 1;
}
}
else if (Rs >= 632) {
if (mcTcor < 91) {
if (Cs < 631) {
return 2;
}
else if (Cs >= 631) {
return 1;
}
}
else if (mcTcor >= 91) {
return 8;
}
}
}
else if (tcor >= 80) {
if (mcTcor < 93) {
if (Rs < 1443) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
if (Rs < 503) {
if (Rs < 425) {
return 8;
}
else if (Rs >= 425) {
return 2;
}
}
else if (Rs >= 503) {
return 8;
}
}
}
else if (Rs >= 1443) {
return 4;
}
}
else if (mcTcor >= 93) {
if (TSC < 5) {
if (Rs < 1070) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 410) {
return 2;
}
else if (Rs >= 410) {
return 4;
}
}
}
else if (Rs >= 1070) {
if (Rs < 1587) {
return 8;
}
else if (Rs >= 1587) {
return 1;
}
}
}
else if (TSC >= 5) {
if (mcTcor < 96) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
return 2;
}
}
else if (mcTcor >= 96) {
return 4;
}
}
}
}
}
else if (ssDCval >= 64) {
if (posBalance < 57) {
if (Rs < 757) {
if (tcor < 91) {
if (Rs < 426) {
return 8;
}
else if (Rs >= 426) {
return 2;
}
}
else if (tcor >= 91) {
if (ssDCval < 73) {
return 4;
}
else if (ssDCval >= 73) {
if (Rs < 538) {
return 8;
}
else if (Rs >= 538) {
return 1;
}
}
}
}
else if (Rs >= 757) {
if (refDCval < 137) {
if (RsCsDiff < 63) {
if (negBalance < 42) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
return 8;
}
}
}
else if (negBalance >= 42) {
if (Rs < 1316) {
return 4;
}
else if (Rs >= 1316) {
return 1;
}
}
}
else if (RsCsDiff >= 63) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (refDCval >= 137) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
return 2;
}
}
}
}
else if (posBalance >= 57) {
if (Rs < 1599) {
if (negBalance < 35) {
if (MVDiff < 30) {
if (Rs < 978) {
if (Rs < 762) {
if (Rs < 621) {
return 8;
}
else if (Rs >= 621) {
return 1;
}
}
else if (Rs >= 762) {
return 8;
}
}
else if (Rs >= 978) {
return 1;
}
}
else if (MVDiff >= 30) {
return 2;
}
}
else if (negBalance >= 35) {
if (posBalance < 61) {
if (negBalance < 40) {
if (Cs < 1129) {
return 8;
}
else if (Cs >= 1129) {
return 2;
}
}
else if (negBalance >= 40) {
if (Cs < 899) {
return 1;
}
else if (Cs >= 899) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (mu_mv_mag_sq < 3) {
if (diffMVdiffVal < -3) {
if (Rs < 775) {
return 2;
}
else if (Rs >= 775) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
if (AFD < 11) {
if (diffAFD < 3) {
if (Cs < 1041) {
return 8;
}
else if (Cs >= 1041) {
return 4;
}
}
else if (diffAFD >= 3) {
if (Rs < 1019) {
return 2;
}
else if (Rs >= 1019) {
return 8;
}
}
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (MVDiff < 28) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (MVDiff >= 28) {
if (Rs < 1385) {
return 4;
}
else if (Rs >= 1385) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1599) {
if (diffTSC < 2) {
return 2;
}
else if (diffTSC >= 2) {
return 1;
}
}
}
}
}
else if (SC >= 2602) {
if (RsDiff < 7) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (MVDiff < 10) {
if (Rs < 2137) {
return 2;
}
else if (Rs >= 2137) {
return 4;
}
}
else if (MVDiff >= 10) {
if (Cs < 657) {
return 2;
}
else if (Cs >= 657) {
return 8;
}
}
}
}
else if (RsDiff >= 7) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (CsDiff < 5) {
if (MVDiff < 38) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (MVDiff >= 38) {
return 8;
}
}
else if (CsDiff >= 5) {
if (negBalance < 32) {
if (Cs < 1159) {
return 8;
}
else if (Cs >= 1159) {
return 2;
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 43) {
if (RsCsDiff < 51) {
if (Rs < 927) {
if (negBalance < 47) {
if (AFD < 12) {
if (AFD < 9) {
if (tcor < 88) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 755) {
return 2;
}
else if (Rs >= 755) {
return 8;
}
}
}
else if (tcor >= 88) {
if (MVDiff < 3) {
if (Rs < 660) {
return 8;
}
else if (Rs >= 660) {
return 4;
}
}
else if (MVDiff >= 3) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 1;
}
}
}
}
}
else if (AFD >= 9) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 315) {
return 1;
}
else if (Rs >= 315) {
if (diffMVdiffVal < 1) {
if (Rs < 873) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (Rs >= 873) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 573) {
return 8;
}
else if (Cs >= 573) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 12) {
if (Rs < 714) {
return 8;
}
else if (Rs >= 714) {
if (tcor < 80) {
if (MVDiff < 36) {
return 1;
}
else if (MVDiff >= 36) {
return 4;
}
}
else if (tcor >= 80) {
return 4;
}
}
}
}
else if (negBalance >= 47) {
if (mcTcor < 93) {
if (MVDiff < 7) {
if (ssDCval < 48) {
if (posBalance < 59) {
if (TSC < 6) {
if (Rs < 730) {
return 4;
}
else if (Rs >= 730) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (posBalance >= 59) {
return 1;
}
}
else if (ssDCval >= 48) {
if (ssDCval < 65) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (ssDCval >= 65) {
if (SC < 1170) {
return 1;
}
else if (SC >= 1170) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
return 8;
}
}
}
}
}
else if (MVDiff >= 7) {
if (AFD < 11) {
if (diffRsCsdiff < -15) {
if (AFD < 10) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (AFD >= 10) {
return 8;
}
}
else if (diffRsCsdiff >= -15) {
if (ssDCval < 49) {
if (Cs < 752) {
if (Rs < 739) {
if (diffMVdiffVal < -17) {
return 4;
}
else if (diffMVdiffVal >= -17) {
return 8;
}
}
else if (Rs >= 739) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
if (Rs < 831) {
return 8;
}
else if (Rs >= 831) {
return 1;
}
}
}
}
}
else if (Cs >= 752) {
return 8;
}
}
else if (ssDCval >= 49) {
if (diffMVdiffVal < 8) {
return 8;
}
else if (diffMVdiffVal >= 8) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
}
else if (AFD >= 11) {
if (Cs < 846) {
if (tcor < 62) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
else if (tcor >= 62) {
if (negBalance < 55) {
if (posBalance < 59) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (Rs < 845) {
return 1;
}
else if (Rs >= 845) {
return 8;
}
}
}
else if (posBalance >= 59) {
if (diffRsCsdiff < 11) {
if (Rs < 406) {
return 1;
}
else if (Rs >= 406) {
return 8;
}
}
else if (diffRsCsdiff >= 11) {
if (Rs < 680) {
return 4;
}
else if (Rs >= 680) {
return 1;
}
}
}
}
else if (negBalance >= 55) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
}
}
else if (Cs >= 846) {
if (TSC < 7) {
if (mu_mv_mag_sq < 5) {
return 1;
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 709) {
return 4;
}
else if (Rs >= 709) {
return 2;
}
}
}
else if (TSC >= 7) {
if (Cs < 1077) {
return 4;
}
else if (Cs >= 1077) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (posBalance < 57) {
if (Rs < 447) {
return 1;
}
else if (Rs >= 447) {
if (RsDiff < 5) {
if (Cs < 548) {
if (Rs < 547) {
return 8;
}
else if (Rs >= 547) {
return 2;
}
}
else if (Cs >= 548) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Rs < 732) {
return 1;
}
else if (Rs >= 732) {
return 8;
}
}
}
}
else if (RsDiff >= 5) {
if (Rs < 903) {
return 2;
}
else if (Rs >= 903) {
return 4;
}
}
}
}
else if (posBalance >= 57) {
if (MVDiff < 12) {
if (ssDCval < 77) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (ssDCval >= 77) {
return 1;
}
}
else if (MVDiff >= 12) {
if (negBalance < 57) {
if (AFD < 13) {
if (Cs < 1074) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 2;
}
}
else if (Cs >= 1074) {
return 1;
}
}
else if (AFD >= 13) {
return 4;
}
}
else if (negBalance >= 57) {
return 8;
}
}
}
}
}
}
else if (Rs >= 927) {
if (tcor < 81) {
if (ssDCval < 62) {
if (negBalance < 45) {
if (Cs < 855) {
return 8;
}
else if (Cs >= 855) {
return 1;
}
}
else if (negBalance >= 45) {
if (MVDiff < 4) {
if (tcor < 71) {
if (Rs < 1115) {
return 1;
}
else if (Rs >= 1115) {
if (Cs < 1091) {
return 8;
}
else if (Cs >= 1091) {
return 4;
}
}
}
else if (tcor >= 71) {
return 8;
}
}
else if (MVDiff >= 4) {
if (refDCval < 61) {
return 8;
}
else if (refDCval >= 61) {
if (Rs < 1037) {
return 8;
}
else if (Rs >= 1037) {
return 1;
}
}
}
}
}
else if (ssDCval >= 62) {
if (tcor < 61) {
return 1;
}
else if (tcor >= 61) {
if (tcor < 76) {
if (Cs < 538) {
return 2;
}
else if (Cs >= 538) {
if (posBalance < 59) {
if (AFD < 11) {
if (Rs < 1328) {
return 8;
}
else if (Rs >= 1328) {
if (Rs < 1480) {
return 2;
}
else if (Rs >= 1480) {
if (Rs < 1666) {
if (Rs < 1580) {
return 8;
}
else if (Rs >= 1580) {
return 2;
}
}
else if (Rs >= 1666) {
return 8;
}
}
}
}
else if (AFD >= 11) {
return 8;
}
}
else if (posBalance >= 59) {
return 8;
}
}
}
else if (tcor >= 76) {
if (ssDCval < 98) {
if (Cs < 943) {
if (SC < 2248) {
if (Cs < 781) {
if (SC < 1998) {
return 2;
}
else if (SC >= 1998) {
return 8;
}
}
else if (Cs >= 781) {
return 8;
}
}
else if (SC >= 2248) {
return 2;
}
}
else if (Cs >= 943) {
if (Rs < 1693) {
return 8;
}
else if (Rs >= 1693) {
return 2;
}
}
}
else if (ssDCval >= 98) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
return 4;
}
}
}
}
}
}
else if (tcor >= 81) {
if (RsDiff < 4) {
if (ssDCval < 114) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (negBalance < 45) {
if (Cs < 910) {
return 1;
}
else if (Cs >= 910) {
return 8;
}
}
else if (negBalance >= 45) {
if (negBalance < 47) {
if (Rs < 1253) {
return 8;
}
else if (Rs >= 1253) {
return 4;
}
}
else if (negBalance >= 47) {
return 8;
}
}
}
}
else if (ssDCval >= 114) {
if (Cs < 1015) {
return 4;
}
else if (Cs >= 1015) {
return 8;
}
}
}
else if (RsDiff >= 4) {
if (Rs < 997) {
if (negBalance < 52) {
if (Cs < 1192) {
return 4;
}
else if (Cs >= 1192) {
return 2;
}
}
else if (negBalance >= 52) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (Rs < 959) {
return 2;
}
else if (Rs >= 959) {
return 4;
}
}
}
}
else if (Rs >= 997) {
if (MVDiff < 7) {
if (refDCval < 89) {
if (Cs < 1137) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (Cs >= 1137) {
if (Rs < 1298) {
return 4;
}
else if (Rs >= 1298) {
return 8;
}
}
}
else if (refDCval >= 89) {
if (ssDCval < 92) {
return 1;
}
else if (ssDCval >= 92) {
if (diffRsCsdiff < 11) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (Rs < 1271) {
return 8;
}
else if (Rs >= 1271) {
if (Rs < 1560) {
return 4;
}
else if (Rs >= 1560) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 11) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 1802) {
if (TSC < 9) {
if (tcor < 93) {
if (Cs < 564) {
if (AFD < 9) {
if (Rs < 1115) {
return 2;
}
else if (Rs >= 1115) {
return 4;
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (Cs >= 564) {
if (SC < 2031) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
if (diffMVdiffVal < -26) {
if (Rs < 1098) {
return 8;
}
else if (Rs >= 1098) {
if (Rs < 1139) {
return 2;
}
else if (Rs >= 1139) {
return 4;
}
}
}
else if (diffMVdiffVal >= -26) {
return 8;
}
}
}
else if (SC >= 2031) {
if (SC < 2489) {
if (posBalance < 56) {
if (SC < 2124) {
if (Rs < 1266) {
if (Rs < 1097) {
return 8;
}
else if (Rs >= 1097) {
return 4;
}
}
else if (Rs >= 1266) {
return 2;
}
}
else if (SC >= 2124) {
return 4;
}
}
else if (posBalance >= 56) {
if (SC < 2186) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (SC >= 2186) {
if (refDCval < 92) {
return 8;
}
else if (refDCval >= 92) {
return 2;
}
}
}
}
else if (SC >= 2489) {
if (Cs < 1015) {
if (Rs < 1616) {
return 8;
}
else if (Rs >= 1616) {
return 1;
}
}
else if (Cs >= 1015) {
return 8;
}
}
}
}
}
else if (tcor >= 93) {
if (MVDiff < 9) {
if (Cs < 958) {
return 8;
}
else if (Cs >= 958) {
return 1;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
}
else if (TSC >= 9) {
if (posBalance < 69) {
return 8;
}
else if (posBalance >= 69) {
return 2;
}
}
}
else if (Rs >= 1802) {
if (MVDiff < 14) {
if (Rs < 1929) {
return 1;
}
else if (Rs >= 1929) {
return 8;
}
}
else if (MVDiff >= 14) {
if (Rs < 2185) {
return 2;
}
else if (Rs >= 2185) {
return 4;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 51) {
if (Cs < 824) {
if (refDCval < 82) {
if (SC < 2698) {
if (negBalance < 49) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 1063) {
return 2;
}
else if (Rs >= 1063) {
return 8;
}
}
}
}
else if (negBalance >= 49) {
if (diffAFD < -3) {
if (Rs < 1707) {
return 4;
}
else if (Rs >= 1707) {
return 8;
}
}
else if (diffAFD >= -3) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (SC >= 2698) {
return 2;
}
}
else if (refDCval >= 82) {
if (ssDCval < 117) {
if (tcor < 82) {
if (Rs < 2085) {
if (Rs < 1706) {
return 4;
}
else if (Rs >= 1706) {
return 2;
}
}
else if (Rs >= 2085) {
return 8;
}
}
else if (tcor >= 82) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (ssDCval >= 117) {
if (tcor < 66) {
if (Rs < 1488) {
return 1;
}
else if (Rs >= 1488) {
return 4;
}
}
else if (tcor >= 66) {
if (Rs < 960) {
return 2;
}
else if (Rs >= 960) {
if (Rs < 1815) {
return 8;
}
else if (Rs >= 1815) {
return 1;
}
}
}
}
}
}
else if (Cs >= 824) {
if (Rs < 1319) {
if (ssDCval < 45) {
if (Rs < 1030) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 4;
}
}
else if (Rs >= 1030) {
return 8;
}
}
else if (ssDCval >= 45) {
if (ssDCval < 67) {
if (Cs < 1104) {
if (ssDCval < 51) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
return 4;
}
}
else if (ssDCval >= 51) {
return 4;
}
}
else if (Cs >= 1104) {
if (diffMVdiffVal < -4) {
if (Rs < 1145) {
return 4;
}
else if (Rs >= 1145) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
}
else if (ssDCval >= 67) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (SC < 2083) {
if (Cs < 933) {
return 2;
}
else if (Cs >= 933) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
}
else if (SC >= 2083) {
if (Cs < 869) {
return 4;
}
else if (Cs >= 869) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1319) {
if (RsDiff < 8) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 55) {
if (TSC < 6) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (Rs < 1757) {
return 8;
}
else if (Rs >= 1757) {
return 2;
}
}
}
else if (TSC >= 6) {
if (Cs < 1203) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (Cs >= 1203) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (Rs < 1691) {
if (posBalance < 59) {
if (diffRsCsdiff < 6) {
if (Rs < 1656) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
else if (Rs >= 1656) {
return 2;
}
}
else if (diffRsCsdiff >= 6) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
else if (posBalance >= 59) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 4;
}
}
}
else if (Rs >= 1691) {
if (diffRsCsdiff < 29) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
else if (diffRsCsdiff >= 29) {
if (Rs < 1817) {
return 8;
}
else if (Rs >= 1817) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 855) {
if (Rs < 1879) {
return 8;
}
else if (Rs >= 1879) {
return 4;
}
}
else if (Cs >= 855) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (diffMVdiffVal < -2) {
if (MVDiff < 26) {
if (Cs < 915) {
return 4;
}
else if (Cs >= 915) {
if (Rs < 2193) {
return 8;
}
else if (Rs >= 2193) {
return 4;
}
}
}
else if (MVDiff >= 26) {
if (Cs < 880) {
return 8;
}
else if (Cs >= 880) {
if (Rs < 2037) {
return 1;
}
else if (Rs >= 2037) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -2) {
if (MVDiff < 17) {
if (ssDCval < 65) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
else if (ssDCval >= 65) {
if (diffRsCsdiff < 26) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 2;
}
}
else if (diffRsCsdiff >= 26) {
return 8;
}
}
}
else if (MVDiff >= 17) {
if (Rs < 1729) {
return 2;
}
else if (Rs >= 1729) {
if (Cs < 1179) {
return 4;
}
else if (Cs >= 1179) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (Rs < 1633) {
if (TSC < 7) {
if (ssDCval < 47) {
if (mu_mv_mag_sq < 103) {
if (Rs < 700) {
if (mcTcor < 91) {
if (MVDiff < 37) {
return 8;
}
else if (MVDiff >= 37) {
if (Rs < 628) {
return 2;
}
else if (Rs >= 628) {
return 4;
}
}
}
else if (mcTcor >= 91) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
return 4;
}
}
}
else if (Rs >= 700) {
if (ssDCval < 40) {
return 8;
}
else if (ssDCval >= 40) {
if (TSC < 5) {
if (mu_mv_mag_sq < 65) {
if (Rs < 1018) {
if (Rs < 871) {
return 4;
}
else if (Rs >= 871) {
return 1;
}
}
else if (Rs >= 1018) {
return 8;
}
}
else if (mu_mv_mag_sq >= 65) {
if (Cs < 796) {
if (Rs < 1004) {
return 8;
}
else if (Rs >= 1004) {
return 2;
}
}
else if (Cs >= 796) {
if (mu_mv_mag_sq < 95) {
return 8;
}
else if (mu_mv_mag_sq >= 95) {
return 2;
}
}
}
}
else if (TSC >= 5) {
if (RsCsDiff < 27) {
return 8;
}
else if (RsCsDiff >= 27) {
if (diffMVdiffVal < -2) {
if (tcor < 79) {
if (SC < 1716) {
return 1;
}
else if (SC >= 1716) {
return 8;
}
}
else if (tcor >= 79) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (SC < 2245) {
if (Rs < 867) {
return 1;
}
else if (Rs >= 867) {
if (Rs < 1124) {
if (refDCval < 45) {
return 8;
}
else if (refDCval >= 45) {
if (Rs < 1067) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (Rs >= 1067) {
return 1;
}
}
}
else if (Rs >= 1124) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
return 1;
}
}
}
}
else if (SC >= 2245) {
return 8;
}
}
}
else if (ssDCval >= 47) {
if (ssDCval < 163) {
if (ssDCval < 103) {
if (ssDCval < 98) {
if (ssDCval < 78) {
if (Rs < 530) {
if (Cs < 823) {
if (Cs < 660) {
if (posBalance < 53) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
return 1;
}
}
}
else if (posBalance >= 53) {
if (Rs < 515) {
return 1;
}
else if (Rs >= 515) {
return 8;
}
}
}
else if (Cs >= 660) {
if (ssDCval < 60) {
if (TSC < 4) {
if (Rs < 351) {
return 1;
}
else if (Rs >= 351) {
if (SC < 1216) {
return 4;
}
else if (SC >= 1216) {
return 1;
}
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (ssDCval >= 60) {
return 8;
}
}
}
else if (Cs >= 823) {
if (TSC < 3) {
if (Rs < 431) {
return 2;
}
else if (Rs >= 431) {
return 1;
}
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (Rs >= 530) {
if (Rs < 962) {
if (SC < 2069) {
if (ssDCval < 63) {
if (negBalance < 32) {
if (TSC < 5) {
if (Rs < 721) {
if (posBalance < 64) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (Cs < 873) {
return 4;
}
else if (Cs >= 873) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (posBalance >= 64) {
return 8;
}
}
else if (Rs >= 721) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
else if (TSC >= 5) {
return 1;
}
}
else if (negBalance >= 32) {
if (Cs < 618) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
return 8;
}
}
}
else if (Cs >= 618) {
if (Rs < 845) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 108) {
if (ssDCval < 56) {
if (SC < 1405) {
return 8;
}
else if (SC >= 1405) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (tcor < 94) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
else if (tcor >= 94) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 108) {
return 4;
}
}
else if (posBalance >= 57) {
return 4;
}
}
else if (Rs >= 845) {
if (mu_mv_mag_sq < 78) {
if (ssDCval < 48) {
if (Rs < 901) {
return 8;
}
else if (Rs >= 901) {
return 1;
}
}
else if (ssDCval >= 48) {
return 8;
}
}
else if (mu_mv_mag_sq >= 78) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 63) {
if (mu_mv_mag_sq < 81) {
if (ssDCval < 75) {
if (diffMVdiffVal < -3) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (Rs < 845) {
if (diffTSC < -1) {
if (Rs < 618) {
return 8;
}
else if (Rs >= 618) {
return 4;
}
}
else if (diffTSC >= -1) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 1;
}
}
}
else if (Rs >= 845) {
if (Cs < 1089) {
if (Cs < 936) {
return 1;
}
else if (Cs >= 936) {
return 2;
}
}
else if (Cs >= 1089) {
return 4;
}
}
}
}
else if (ssDCval >= 75) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (Rs < 869) {
if (negBalance < 29) {
if (Rs < 694) {
return 4;
}
else if (Rs >= 694) {
return 2;
}
}
else if (negBalance >= 29) {
if (posBalance < 54) {
if (SC < 1416) {
return 8;
}
else if (SC >= 1416) {
if (Rs < 733) {
return 1;
}
else if (Rs >= 733) {
return 2;
}
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (Rs >= 869) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
}
}
}
else if (SC >= 2069) {
if (Rs < 831) {
return 8;
}
else if (Rs >= 831) {
return 4;
}
}
}
else if (Rs >= 962) {
if (RsCsDiff < 15) {
if (Cs < 1108) {
if (Cs < 610) {
if (tcor < 94) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 4;
}
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (Cs >= 610) {
if (ssDCval < 48) {
if (SC < 2058) {
if (Rs < 978) {
return 1;
}
else if (Rs >= 978) {
return 8;
}
}
else if (SC >= 2058) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (ssDCval >= 48) {
if (Cs < 1059) {
if (gchDC < 2) {
if (mcTcor < 96) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (posBalance < 54) {
if (Rs < 1182) {
if (Cs < 827) {
return 4;
}
else if (Cs >= 827) {
return 8;
}
}
else if (Rs >= 1182) {
return 1;
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1196) {
if (Cs < 849) {
return 8;
}
else if (Cs >= 849) {
if (negBalance < 38) {
if (Rs < 1183) {
return 8;
}
else if (Rs >= 1183) {
return 1;
}
}
else if (negBalance >= 38) {
return 1;
}
}
}
else if (Rs >= 1196) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (Rs < 1095) {
return 1;
}
else if (Rs >= 1095) {
return 8;
}
}
}
else if (Cs >= 1059) {
if (Rs < 1178) {
if (AFD < 5) {
if (Rs < 1086) {
return 8;
}
else if (Rs >= 1086) {
return 1;
}
}
else if (AFD >= 5) {
return 4;
}
}
else if (Rs >= 1178) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 1108) {
if (MVDiff < 4) {
if (negBalance < 39) {
if (Cs < 1125) {
if (Rs < 1216) {
return 8;
}
else if (Rs >= 1216) {
return 4;
}
}
else if (Cs >= 1125) {
if (Rs < 1003) {
return 4;
}
else if (Rs >= 1003) {
if (diffRsCsdiff < -11) {
if (Rs < 1051) {
return 8;
}
else if (Rs >= 1051) {
return 4;
}
}
else if (diffRsCsdiff >= -11) {
return 8;
}
}
}
}
else if (negBalance >= 39) {
if (tcor < 96) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (tcor >= 96) {
if (Rs < 1339) {
return 1;
}
else if (Rs >= 1339) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 1125) {
if (Rs < 1091) {
return 8;
}
else if (Rs >= 1091) {
return 4;
}
}
else if (Rs >= 1125) {
if (Rs < 1515) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (Cs < 1200) {
return 8;
}
else if (Cs >= 1200) {
return 1;
}
}
}
else if (Rs >= 1515) {
if (Cs < 1146) {
return 4;
}
else if (Cs >= 1146) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 15) {
if (Cs < 898) {
if (negBalance < 45) {
if (ssDCval < 63) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (mcTcor < 90) {
if (SC < 1871) {
if (Rs < 1153) {
return 4;
}
else if (Rs >= 1153) {
return 1;
}
}
else if (SC >= 1871) {
return 8;
}
}
else if (mcTcor >= 90) {
if (CsDiff < 3) {
if (Rs < 1469) {
return 4;
}
else if (Rs >= 1469) {
return 1;
}
}
else if (CsDiff >= 3) {
if (Cs < 634) {
return 1;
}
else if (Cs >= 634) {
return 2;
}
}
}
}
}
else if (ssDCval >= 63) {
if (mu_mv_mag_sq < 83) {
if (Rs < 1558) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (Rs >= 1558) {
if (Rs < 1594) {
return 8;
}
else if (Rs >= 1594) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 4;
}
}
}
}
else if (Cs >= 898) {
if (Rs < 1212) {
if (ssDCval < 68) {
if (Cs < 1032) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Rs < 980) {
return 4;
}
else if (Rs >= 980) {
return 8;
}
}
}
else if (Cs >= 1032) {
if (ssDCval < 49) {
if (Rs < 1060) {
return 4;
}
else if (Rs >= 1060) {
return 8;
}
}
else if (ssDCval >= 49) {
if (diffMVdiffVal < 2) {
if (negBalance < 46) {
if (negBalance < 34) {
if (Cs < 1258) {
return 4;
}
else if (Cs >= 1258) {
return 8;
}
}
else if (negBalance >= 34) {
return 4;
}
}
else if (negBalance >= 46) {
if (Rs < 1164) {
return 4;
}
else if (Rs >= 1164) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 1197) {
return 4;
}
else if (Cs >= 1197) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 68) {
if (Cs < 1203) {
return 8;
}
else if (Cs >= 1203) {
return 4;
}
}
}
else if (Rs >= 1212) {
if (ssDCval < 65) {
if (MVDiff < 5) {
if (RsDiff < 6) {
if (Cs < 904) {
return 4;
}
else if (Cs >= 904) {
if (Rs < 1275) {
if (Cs < 1221) {
return 8;
}
else if (Cs >= 1221) {
return 4;
}
}
else if (Rs >= 1275) {
if (ssDCval < 50) {
if (Rs < 1351) {
return 8;
}
else if (Rs >= 1351) {
return 2;
}
}
else if (ssDCval >= 50) {
return 8;
}
}
}
}
else if (RsDiff >= 6) {
if (Rs < 1417) {
return 4;
}
else if (Rs >= 1417) {
if (Rs < 1517) {
return 2;
}
else if (Rs >= 1517) {
return 8;
}
}
}
}
else if (MVDiff >= 5) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (mu_mv_mag_sq < 88) {
return 8;
}
else if (mu_mv_mag_sq >= 88) {
if (Rs < 1277) {
return 1;
}
else if (Rs >= 1277) {
return 4;
}
}
}
}
}
else if (ssDCval >= 65) {
if (MVDiff < 4) {
if (TSC < 6) {
if (diffAFD < 2) {
if (Rs < 1304) {
if (RsCsDiff < 29) {
return 1;
}
else if (RsCsDiff >= 29) {
return 4;
}
}
else if (Rs >= 1304) {
if (mcTcor < 95) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (mcTcor >= 95) {
if (ssDCval < 68) {
if (Cs < 1123) {
return 8;
}
else if (Cs >= 1123) {
return 4;
}
}
else if (ssDCval >= 68) {
return 8;
}
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (TSC >= 6) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 76) {
if (SC < 2740) {
return 4;
}
else if (SC >= 2740) {
if (Rs < 1519) {
return 8;
}
else if (Rs >= 1519) {
return 1;
}
}
}
else if (ssDCval >= 76) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 78) {
if (refDCval < 91) {
if (negBalance < 30) {
if (ssDCval < 88) {
return 2;
}
else if (ssDCval >= 88) {
return 8;
}
}
else if (negBalance >= 30) {
if (AFD < 9) {
if (posBalance < 50) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
else if (posBalance >= 50) {
if (MVDiff < 2) {
if (Rs < 829) {
if (Rs < 720) {
return 8;
}
else if (Rs >= 720) {
return 4;
}
}
else if (Rs >= 829) {
return 8;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (AFD >= 9) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (Rs < 1112) {
return 2;
}
else if (Rs >= 1112) {
return 1;
}
}
}
}
}
else if (refDCval >= 91) {
if (SC < 2004) {
if (tcor < 87) {
return 2;
}
else if (tcor >= 87) {
if (refDCval < 94) {
if (MVDiff < 5) {
if (Cs < 991) {
return 1;
}
else if (Cs >= 991) {
return 4;
}
}
else if (MVDiff >= 5) {
if (Rs < 805) {
return 8;
}
else if (Rs >= 805) {
return 1;
}
}
}
else if (refDCval >= 94) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
else if (SC >= 2004) {
if (RsCsDiff < 38) {
return 8;
}
else if (RsCsDiff >= 38) {
return 2;
}
}
}
}
}
else if (ssDCval >= 98) {
if (Rs < 449) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 4;
}
}
else if (Rs >= 449) {
if (Cs < 1120) {
if (SC < 1225) {
return 1;
}
else if (SC >= 1225) {
return 8;
}
}
else if (Cs >= 1120) {
return 4;
}
}
}
}
else if (ssDCval >= 103) {
if (MVDiff < 25) {
if (tcor < 82) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 1;
}
}
}
else if (tcor >= 82) {
if (Rs < 750) {
if (CsDiff < 2) {
if (Rs < 654) {
return 8;
}
else if (Rs >= 654) {
return 1;
}
}
else if (CsDiff >= 2) {
if (Cs < 616) {
if (Cs < 612) {
return 8;
}
else if (Cs >= 612) {
return 2;
}
}
else if (Cs >= 616) {
return 8;
}
}
}
else if (Rs >= 750) {
if (RsCsDiff < 19) {
if (ssDCval < 130) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
return 8;
}
}
else if (ssDCval >= 130) {
if (Rs < 1027) {
return 8;
}
else if (Rs >= 1027) {
if (SC < 2214) {
return 4;
}
else if (SC >= 2214) {
if (mu_mv_mag_sq < 51) {
return 4;
}
else if (mu_mv_mag_sq >= 51) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 19) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
}
}
else if (MVDiff >= 25) {
if (mu_mv_mag_sq < 48) {
if (CsDiff < 5) {
if (ssDCval < 134) {
return 8;
}
else if (ssDCval >= 134) {
return 4;
}
}
else if (CsDiff >= 5) {
if (Rs < 1167) {
return 1;
}
else if (Rs >= 1167) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (Rs < 708) {
return 8;
}
else if (Rs >= 708) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 842) {
return 4;
}
else if (Rs >= 842) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (negBalance < 44) {
if (ssDCval < 80) {
if (tcor < 83) {
if (RsCsDiff < 33) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (RsCsDiff >= 33) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Cs < 1020) {
return 8;
}
else if (Cs >= 1020) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
else if (tcor >= 83) {
if (mu_mv_mag_sq < 62) {
if (gchDC < 2) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 1312) {
return 4;
}
else if (Rs >= 1312) {
return 1;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 62) {
if (Rs < 1545) {
return 4;
}
else if (Rs >= 1545) {
return 8;
}
}
}
}
else if (ssDCval >= 80) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (AFD < 11) {
if (Rs < 1159) {
return 4;
}
else if (Rs >= 1159) {
return 8;
}
}
else if (AFD >= 11) {
if (Cs < 1070) {
return 2;
}
else if (Cs >= 1070) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 44) {
if (tcor < 77) {
if (Cs < 660) {
if (ssDCval < 57) {
if (Cs < 562) {
return 4;
}
else if (Cs >= 562) {
return 8;
}
}
else if (ssDCval >= 57) {
return 4;
}
}
else if (Cs >= 660) {
if (SC < 1770) {
if (ssDCval < 47) {
if (Rs < 863) {
return 8;
}
else if (Rs >= 863) {
return 1;
}
}
else if (ssDCval >= 47) {
if (Cs < 697) {
return 8;
}
else if (Cs >= 697) {
return 4;
}
}
}
else if (SC >= 1770) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
return 8;
}
}
}
}
else if (tcor >= 77) {
if (diffAFD < -2) {
if (Cs < 866) {
if (SC < 1472) {
return 4;
}
else if (SC >= 1472) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 2;
}
}
}
else if (Cs >= 866) {
if (Cs < 1215) {
return 8;
}
else if (Cs >= 1215) {
return 1;
}
}
}
else if (diffAFD >= -2) {
if (Rs < 1588) {
if (AFD < 9) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Rs < 1431) {
return 8;
}
else if (Rs >= 1431) {
return 1;
}
}
}
else if (AFD >= 9) {
if (posBalance < 59) {
if (SC < 1663) {
if (Rs < 879) {
return 8;
}
else if (Rs >= 879) {
return 4;
}
}
else if (SC >= 1663) {
if (mcTcor < 92) {
if (Cs < 1243) {
if (Rs < 1016) {
if (Rs < 1010) {
return 8;
}
else if (Rs >= 1010) {
return 1;
}
}
else if (Rs >= 1016) {
return 8;
}
}
else if (Cs >= 1243) {
if (Rs < 1326) {
return 1;
}
else if (Rs >= 1326) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 47) {
if (Rs < 1137) {
return 1;
}
else if (Rs >= 1137) {
return 4;
}
}
else if (mu_mv_mag_sq >= 47) {
if (Cs < 711) {
return 1;
}
else if (Cs >= 711) {
if (ssDCval < 83) {
if (diffRsCsdiff < -10) {
return 4;
}
else if (diffRsCsdiff >= -10) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (Rs < 1291) {
return 4;
}
else if (Rs >= 1291) {
return 8;
}
}
}
}
else if (ssDCval >= 83) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 59) {
if (mu_mv_mag_sq < 94) {
return 8;
}
else if (mu_mv_mag_sq >= 94) {
if (Cs < 950) {
return 4;
}
else if (Cs >= 950) {
return 8;
}
}
}
}
}
else if (Rs >= 1588) {
if (MVDiff < 22) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (diffMVdiffVal < 6) {
if (Rs < 1590) {
if (Cs < 955) {
return 8;
}
else if (Cs >= 955) {
return 2;
}
}
else if (Rs >= 1590) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
return 2;
}
}
}
else if (MVDiff >= 22) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1633) {
if (ssDCval < 72) {
if (ssDCval < 63) {
if (Cs < 867) {
if (mu_mv_mag_sq < 59) {
if (mu_mv_mag_sq < 44) {
if (diffTSC < 2) {
return 4;
}
else if (diffTSC >= 2) {
return 2;
}
}
else if (mu_mv_mag_sq >= 44) {
return 2;
}
}
else if (mu_mv_mag_sq >= 59) {
if (diffAFD < -1) {
if (Rs < 1829) {
return 1;
}
else if (Rs >= 1829) {
return 2;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (Cs >= 867) {
if (TSC < 9) {
if (mu_mv_mag_sq < 116) {
return 8;
}
else if (mu_mv_mag_sq >= 116) {
if (Cs < 1205) {
if (Rs < 1969) {
return 2;
}
else if (Rs >= 1969) {
return 4;
}
}
else if (Cs >= 1205) {
return 8;
}
}
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (ssDCval >= 63) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (posBalance < 51) {
if (Rs < 2134) {
if (SC < 2802) {
return 8;
}
else if (SC >= 2802) {
return 1;
}
}
else if (Rs >= 2134) {
return 2;
}
}
else if (posBalance >= 51) {
if (SC < 3215) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (SC >= 3215) {
return 8;
}
}
}
}
}
else if (ssDCval >= 72) {
if (refDCval < 81) {
if (negBalance < 48) {
if (mu_mv_mag_sq < 107) {
if (refDCval < 78) {
if (tcor < 85) {
if (Rs < 2091) {
return 8;
}
else if (Rs >= 2091) {
if (Rs < 2179) {
return 2;
}
else if (Rs >= 2179) {
return 1;
}
}
}
else if (tcor >= 85) {
return 1;
}
}
else if (refDCval >= 78) {
if (AFD < 8) {
if (SC < 3080) {
return 1;
}
else if (SC >= 3080) {
return 8;
}
}
else if (AFD >= 8) {
if (Rs < 2095) {
if (Cs < 1273) {
return 8;
}
else if (Cs >= 1273) {
return 1;
}
}
else if (Rs >= 2095) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
return 8;
}
}
else if (negBalance >= 48) {
if (SC < 3374) {
return 8;
}
else if (SC >= 3374) {
return 1;
}
}
}
else if (refDCval >= 81) {
if (RsDiff < 6) {
if (ssDCval < 94) {
if (TSC < 6) {
if (RsDiff < 4) {
if (SC < 3027) {
if (Cs < 785) {
return 4;
}
else if (Cs >= 785) {
return 8;
}
}
else if (SC >= 3027) {
if (Rs < 1799) {
return 8;
}
else if (Rs >= 1799) {
return 2;
}
}
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (TSC >= 6) {
if (diffTSC < 2) {
if (Rs < 1821) {
return 8;
}
else if (Rs >= 1821) {
if (Rs < 1826) {
return 1;
}
else if (Rs >= 1826) {
return 8;
}
}
}
else if (diffTSC >= 2) {
return 4;
}
}
}
else if (ssDCval >= 94) {
if (mu_mv_mag_sq < 116) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
if (Cs < 999) {
return 1;
}
else if (Cs >= 999) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
else if (RsDiff >= 6) {
if (CsDiff < 4) {
if (tcor < 91) {
if (Rs < 1949) {
return 2;
}
else if (Rs >= 1949) {
if (ssDCval < 101) {
if (Rs < 2145) {
return 1;
}
else if (Rs >= 2145) {
return 4;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
}
else if (tcor >= 91) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
if (Rs < 1982) {
return 8;
}
else if (Rs >= 1982) {
return 2;
}
}
}
}
else if (CsDiff >= 4) {
if (refDCval < 106) {
if (RsCsDiff < 85) {
if (TSC < 8) {
if (mcTcor < 95) {
if (tcor < 82) {
return 8;
}
else if (tcor >= 82) {
return 2;
}
}
else if (mcTcor >= 95) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (TSC >= 8) {
if (Rs < 1696) {
return 4;
}
else if (Rs >= 1696) {
return 8;
}
}
}
else if (RsCsDiff >= 85) {
return 4;
}
}
else if (refDCval >= 106) {
if (Cs < 891) {
return 1;
}
else if (Cs >= 891) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (AFD >= 14) {
if (refDCval < 116) {
if (MVDiff < 10) {
if (mu_mv_mag_sq < 3) {
if (refDCval < 86) {
if (negBalance < 54) {
if (posBalance < 64) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
if (ssDCval < 81) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
if (Rs < 1640) {
return 8;
}
else if (Rs >= 1640) {
return 2;
}
}
}
else if (ssDCval >= 81) {
return 4;
}
}
}
else if (posBalance >= 64) {
return 8;
}
}
else if (negBalance >= 54) {
return 8;
}
}
else if (refDCval >= 86) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (RsDiff < 6) {
if (Rs < 1524) {
return 4;
}
else if (Rs >= 1524) {
return 8;
}
}
else if (RsDiff >= 6) {
if (Rs < 1904) {
return 8;
}
else if (Rs >= 1904) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 93) {
if (ssDCval < 102) {
if (RsDiff < 10) {
if (ssDCval < 54) {
if (Rs < 1128) {
return 4;
}
else if (Rs >= 1128) {
return 8;
}
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (RsDiff >= 10) {
if (Rs < 1820) {
return 4;
}
else if (Rs >= 1820) {
return 8;
}
}
}
else if (ssDCval >= 102) {
return 4;
}
}
else if (mcTcor >= 93) {
if (Rs < 447) {
return 2;
}
else if (Rs >= 447) {
if (RsDiff < 9) {
if (posBalance < 65) {
if (Rs < 1026) {
return 8;
}
else if (Rs >= 1026) {
if (SC < 2451) {
return 4;
}
else if (SC >= 2451) {
if (Rs < 1865) {
return 8;
}
else if (Rs >= 1865) {
return 4;
}
}
}
}
else if (posBalance >= 65) {
return 4;
}
}
else if (RsDiff >= 9) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
if (Rs < 2019) {
return 1;
}
else if (Rs >= 2019) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 10) {
if (negBalance < 49) {
if (Rs < 1489) {
if (refDCval < 49) {
if (posBalance < 62) {
if (refDCval < 45) {
return 1;
}
else if (refDCval >= 45) {
return 2;
}
}
else if (posBalance >= 62) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Rs < 813) {
return 2;
}
else if (Rs >= 813) {
return 8;
}
}
}
}
else if (refDCval >= 49) {
if (Rs < 337) {
if (Rs < 118) {
return 1;
}
else if (Rs >= 118) {
return 4;
}
}
else if (Rs >= 337) {
if (diffMVdiffVal < -15) {
if (RsCsDiff < 115) {
if (MVDiff < 31) {
return 8;
}
else if (MVDiff >= 31) {
if (Cs < 1225) {
return 2;
}
else if (Cs >= 1225) {
return 4;
}
}
}
else if (RsCsDiff >= 115) {
if (Cs < 1070) {
if (AFD < 33) {
return 1;
}
else if (AFD >= 33) {
return 4;
}
}
else if (Cs >= 1070) {
return 8;
}
}
}
else if (diffMVdiffVal >= -15) {
if (diffTSC < 3) {
if (refDCval < 68) {
if (AFD < 16) {
if (RsDiff < 7) {
if (mu_mv_mag_sq < 35) {
if (RsDiff < 6) {
if (Rs < 714) {
return 4;
}
else if (Rs >= 714) {
if (Cs < 1164) {
return 8;
}
else if (Cs >= 1164) {
return 4;
}
}
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (mu_mv_mag_sq >= 35) {
if (Rs < 1010) {
return 1;
}
else if (Rs >= 1010) {
return 8;
}
}
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (AFD >= 16) {
if (AFD < 32) {
if (RsCsDiff < 46) {
if (TSC < 8) {
if (Rs < 644) {
return 8;
}
else if (Rs >= 644) {
return 1;
}
}
else if (TSC >= 8) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (RsCsDiff >= 46) {
return 4;
}
}
else if (AFD >= 32) {
return 8;
}
}
}
else if (refDCval >= 68) {
if (SC < 2265) {
if (RsDiff < 7) {
if (Rs < 1114) {
if (Rs < 906) {
if (RsCsDiff < 55) {
if (MVDiff < 35) {
if (Rs < 812) {
return 2;
}
else if (Rs >= 812) {
return 8;
}
}
else if (MVDiff >= 35) {
return 1;
}
}
else if (RsCsDiff >= 55) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
if (Rs < 436) {
return 1;
}
else if (Rs >= 436) {
return 8;
}
}
}
}
else if (Rs >= 906) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
if (tcor < 81) {
return 1;
}
else if (tcor >= 81) {
if (Rs < 1073) {
return 2;
}
else if (Rs >= 1073) {
return 1;
}
}
}
}
}
else if (Rs >= 1114) {
return 4;
}
}
else if (RsDiff >= 7) {
if (negBalance < 42) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
else if (negBalance >= 42) {
if (Rs < 1009) {
if (refDCval < 73) {
return 2;
}
else if (refDCval >= 73) {
return 8;
}
}
else if (Rs >= 1009) {
if (Rs < 1250) {
return 4;
}
else if (Rs >= 1250) {
return 8;
}
}
}
}
}
else if (SC >= 2265) {
if (RsDiff < 9) {
if (mu_mv_mag_sq < 38) {
if (Rs < 1460) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
else if (Rs >= 1460) {
return 4;
}
}
else if (mu_mv_mag_sq >= 38) {
return 4;
}
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
else if (diffTSC >= 3) {
if (RsCsDiff < 59) {
return 8;
}
else if (RsCsDiff >= 59) {
if (RsDiff < 9) {
if (Rs < 916) {
return 2;
}
else if (Rs >= 916) {
return 1;
}
}
else if (RsDiff >= 9) {
if (Rs < 1272) {
return 8;
}
else if (Rs >= 1272) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 1489) {
if (posBalance < 50) {
if (tcor < 82) {
if (refDCval < 41) {
return 1;
}
else if (refDCval >= 41) {
if (mu_mv_mag_sq < 31) {
if (Rs < 2251) {
return 8;
}
else if (Rs >= 2251) {
return 2;
}
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
}
else if (tcor >= 82) {
if (Rs < 1943) {
return 1;
}
else if (Rs >= 1943) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (Cs < 876) {
if (tcor < 61) {
if (mu_mv_mag_sq < 14) {
return 8;
}
else if (mu_mv_mag_sq >= 14) {
if (Rs < 1689) {
return 1;
}
else if (Rs >= 1689) {
return 2;
}
}
}
else if (tcor >= 61) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (negBalance < 33) {
return 1;
}
else if (negBalance >= 33) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
}
else if (Cs >= 876) {
if (SC < 2803) {
if (ssDCval < 70) {
if (MVDiff < 35) {
return 4;
}
else if (MVDiff >= 35) {
if (Cs < 1146) {
return 8;
}
else if (Cs >= 1146) {
return 1;
}
}
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (SC >= 2803) {
if (Rs < 1715) {
return 2;
}
else if (Rs >= 1715) {
if (RsCsDiff < 143) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 2020) {
return 8;
}
else if (Rs >= 2020) {
return 4;
}
}
}
else if (RsCsDiff >= 143) {
if (CsDiff < 8) {
if (diffAFD < -3) {
if (Rs < 2070) {
return 8;
}
else if (Rs >= 2070) {
return 4;
}
}
else if (diffAFD >= -3) {
if (Cs < 891) {
return 4;
}
else if (Cs >= 891) {
return 2;
}
}
}
else if (CsDiff >= 8) {
if (Rs < 1765) {
return 4;
}
else if (Rs >= 1765) {
return 8;
}
}
}
}
}
}
}
}
}
else if (negBalance >= 49) {
if (AFD < 23) {
if (mcTcor < 88) {
if (ssDCval < 95) {
if (Rs < 1032) {
if (MVDiff < 60) {
if (Rs < 727) {
return 4;
}
else if (Rs >= 727) {
if (ssDCval < 78) {
if (ssDCval < 57) {
if (SC < 1508) {
if (Rs < 908) {
return 8;
}
else if (Rs >= 908) {
return 2;
}
}
else if (SC >= 1508) {
if (AFD < 20) {
if (diffMVdiffVal < 5) {
return 4;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 960) {
return 4;
}
else if (Rs >= 960) {
return 2;
}
}
}
else if (AFD >= 20) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (mu_mv_mag_sq < 5) {
if (Cs < 683) {
return 4;
}
else if (Cs >= 683) {
return 1;
}
}
else if (mu_mv_mag_sq >= 5) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (Cs < 677) {
return 4;
}
else if (Cs >= 677) {
return 2;
}
}
}
}
}
else if (ssDCval >= 78) {
return 8;
}
}
}
else if (MVDiff >= 60) {
if (RsDiff < 7) {
if (Rs < 396) {
if (Rs < 390) {
return 4;
}
else if (Rs >= 390) {
return 2;
}
}
else if (Rs >= 396) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
if (RsCsDiff < 22) {
return 2;
}
else if (RsCsDiff >= 22) {
if (diffMVdiffVal < -20) {
if (Rs < 639) {
return 2;
}
else if (Rs >= 639) {
return 8;
}
}
else if (diffMVdiffVal >= -20) {
return 8;
}
}
}
}
}
else if (RsDiff >= 7) {
if (Rs < 890) {
return 1;
}
else if (Rs >= 890) {
return 4;
}
}
}
}
else if (Rs >= 1032) {
if (refDCval < 85) {
if (ssDCval < 50) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (ssDCval >= 50) {
if (AFD < 21) {
if (Cs < 1032) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
if (diffMVdiffVal < -4) {
if (Rs < 1231) {
return 4;
}
else if (Rs >= 1231) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
if (AFD < 18) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (AFD >= 18) {
if (Rs < 1087) {
return 8;
}
else if (Rs >= 1087) {
return 4;
}
}
}
}
}
else if (Cs >= 1032) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (Rs < 1380) {
return 2;
}
else if (Rs >= 1380) {
return 8;
}
}
}
}
else if (AFD >= 21) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
if (Rs < 1596) {
return 4;
}
else if (Rs >= 1596) {
if (Rs < 1948) {
return 8;
}
else if (Rs >= 1948) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 85) {
if (mcTcor < 83) {
if (Rs < 2158) {
return 8;
}
else if (Rs >= 2158) {
return 1;
}
}
else if (mcTcor >= 83) {
if (TSC < 10) {
if (Cs < 867) {
return 1;
}
else if (Cs >= 867) {
return 8;
}
}
else if (TSC >= 10) {
if (Rs < 1214) {
return 8;
}
else if (Rs >= 1214) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 95) {
if (tcor < 73) {
if (ssDCval < 114) {
if (ssDCval < 102) {
if (Rs < 1663) {
return 4;
}
else if (Rs >= 1663) {
return 2;
}
}
else if (ssDCval >= 102) {
return 4;
}
}
else if (ssDCval >= 114) {
return 8;
}
}
else if (tcor >= 73) {
return 1;
}
}
}
else if (mcTcor >= 88) {
if (tcor < 89) {
if (Rs < 255) {
return 4;
}
else if (Rs >= 255) {
if (SC < 2522) {
if (Cs < 583) {
if (ssDCval < 98) {
if (tcor < 67) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
else if (tcor >= 67) {
return 4;
}
}
else if (ssDCval >= 98) {
if (Rs < 1335) {
return 8;
}
else if (Rs >= 1335) {
return 1;
}
}
}
else if (Cs >= 583) {
if (mu_mv_mag_sq < 2) {
if (refDCval < 103) {
if (mcTcor < 91) {
if (refDCval < 68) {
if (posBalance < 57) {
if (Cs < 910) {
return 4;
}
else if (Cs >= 910) {
return 8;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (refDCval >= 68) {
if (posBalance < 38) {
if (Rs < 1048) {
return 8;
}
else if (Rs >= 1048) {
return 4;
}
}
else if (posBalance >= 38) {
if (Cs < 743) {
return 4;
}
else if (Cs >= 743) {
return 2;
}
}
}
}
else if (mcTcor >= 91) {
if (AFD < 15) {
if (TSC < 7) {
if (Rs < 1026) {
return 1;
}
else if (Rs >= 1026) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (AFD >= 15) {
if (posBalance < 61) {
if (gchDC < 2) {
if (Rs < 1116) {
if (Rs < 908) {
return 1;
}
else if (Rs >= 908) {
return 2;
}
}
else if (Rs >= 1116) {
return 4;
}
}
else if (gchDC >= 2) {
if (gchDC < 8) {
return 2;
}
else if (gchDC >= 8) {
return 4;
}
}
}
else if (posBalance >= 61) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (Cs < 1097) {
return 1;
}
else if (Cs >= 1097) {
if (Rs < 857) {
return 4;
}
else if (Rs >= 857) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 103) {
if (RsCsDiff < 51) {
return 4;
}
else if (RsCsDiff >= 51) {
if (tcor < 79) {
if (Rs < 904) {
return 4;
}
else if (Rs >= 904) {
return 8;
}
}
else if (tcor >= 79) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsDiff < 12) {
if (Rs < 698) {
if (ssDCval < 65) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Rs < 262) {
return 2;
}
else if (Rs >= 262) {
return 8;
}
}
}
else if (ssDCval >= 65) {
if (posBalance < 63) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (ssDCval < 67) {
if (Rs < 598) {
return 2;
}
else if (Rs >= 598) {
return 1;
}
}
else if (ssDCval >= 67) {
return 2;
}
}
}
else if (posBalance >= 63) {
if (Rs < 416) {
return 2;
}
else if (Rs >= 416) {
return 8;
}
}
}
}
else if (Rs >= 698) {
if (mcTcor < 94) {
if (diffAFD < 3) {
if (Cs < 1120) {
if (RsDiff < 10) {
if (tcor < 65) {
if (Rs < 1427) {
return 1;
}
else if (Rs >= 1427) {
return 4;
}
}
else if (tcor >= 65) {
if (diffRsCsdiff < -26) {
if (Cs < 938) {
return 4;
}
else if (Cs >= 938) {
return 8;
}
}
else if (diffRsCsdiff >= -26) {
if (RsDiff < 9) {
if (SC < 2048) {
if (CsDiff < 5) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (MVDiff < 18) {
if (Cs < 871) {
return 4;
}
else if (Cs >= 871) {
return 8;
}
}
else if (MVDiff >= 18) {
if (Cs < 854) {
if (Rs < 1155) {
return 1;
}
else if (Rs >= 1155) {
return 4;
}
}
else if (Cs >= 854) {
return 4;
}
}
}
}
else if (SC >= 2048) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Rs < 1049) {
return 8;
}
else if (Rs >= 1049) {
return 1;
}
}
}
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
else if (RsDiff >= 10) {
if (TSC < 8) {
if (Rs < 1518) {
return 2;
}
else if (Rs >= 1518) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (Cs >= 1120) {
if (Rs < 849) {
return 4;
}
else if (Rs >= 849) {
if (ssDCval < 72) {
if (Rs < 930) {
return 8;
}
else if (Rs >= 930) {
return 1;
}
}
else if (ssDCval >= 72) {
return 2;
}
}
}
}
else if (diffAFD >= 3) {
if (diffMVdiffVal < 9) {
return 4;
}
else if (diffMVdiffVal >= 9) {
if (Rs < 1285) {
return 1;
}
else if (Rs >= 1285) {
return 2;
}
}
}
}
else if (mcTcor >= 94) {
if (diffAFD < -3) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (diffAFD >= -3) {
if (negBalance < 53) {
if (Cs < 994) {
if (RsCsDiff < 92) {
if (Cs < 737) {
return 2;
}
else if (Cs >= 737) {
return 4;
}
}
else if (RsCsDiff >= 92) {
return 1;
}
}
else if (Cs >= 994) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
return 2;
}
}
}
else if (negBalance >= 53) {
if (refDCval < 84) {
if (TSC < 9) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (TSC >= 9) {
return 4;
}
}
else if (refDCval >= 84) {
if (tcor < 75) {
return 1;
}
else if (tcor >= 75) {
if (negBalance < 63) {
return 8;
}
else if (negBalance >= 63) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 12) {
return 2;
}
}
}
}
else if (SC >= 2522) {
if (Rs < 1981) {
if (refDCval < 82) {
if (refDCval < 66) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 1568) {
return 1;
}
else if (Rs >= 1568) {
return 8;
}
}
}
}
else if (refDCval >= 66) {
if (negBalance < 50) {
if (SC < 2982) {
return 8;
}
else if (SC >= 2982) {
return 4;
}
}
else if (negBalance >= 50) {
if (Cs < 944) {
if (SC < 2730) {
return 8;
}
else if (SC >= 2730) {
return 4;
}
}
else if (Cs >= 944) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (refDCval < 72) {
if (Rs < 1472) {
return 1;
}
else if (Rs >= 1472) {
return 2;
}
}
else if (refDCval >= 72) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 82) {
if (SC < 2988) {
if (MVDiff < 14) {
if (Cs < 1067) {
return 4;
}
else if (Cs >= 1067) {
return 8;
}
}
else if (MVDiff >= 14) {
if (TSC < 11) {
if (tcor < 79) {
if (Rs < 1812) {
if (RsCsDiff < 137) {
return 1;
}
else if (RsCsDiff >= 137) {
return 8;
}
}
else if (Rs >= 1812) {
return 8;
}
}
else if (tcor >= 79) {
if (Rs < 1394) {
return 2;
}
else if (Rs >= 1394) {
return 8;
}
}
}
else if (TSC >= 11) {
if (mu_mv_mag_sq < 23) {
return 1;
}
else if (mu_mv_mag_sq >= 23) {
if (Rs < 1883) {
return 4;
}
else if (Rs >= 1883) {
return 8;
}
}
}
}
}
else if (SC >= 2988) {
if (posBalance < 60) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 1;
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
}
else if (Rs >= 1981) {
if (SC < 2926) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 4;
}
}
else if (SC >= 2926) {
if (MVDiff < 25) {
if (Cs < 1208) {
if (TSC < 10) {
if (diffMVdiffVal < -3) {
if (Rs < 2046) {
return 4;
}
else if (Rs >= 2046) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (Cs < 1171) {
return 2;
}
else if (Cs >= 1171) {
return 4;
}
}
}
else if (TSC >= 10) {
return 4;
}
}
else if (Cs >= 1208) {
if (Rs < 2121) {
return 8;
}
else if (Rs >= 2121) {
return 2;
}
}
}
else if (MVDiff >= 25) {
if (Cs < 1058) {
return 8;
}
else if (Cs >= 1058) {
if (SC < 3156) {
return 8;
}
else if (SC >= 3156) {
return 4;
}
}
}
}
}
}
}
}
else if (tcor >= 89) {
if (mu_mv_mag_sq < 7) {
if (diffRsCsdiff < -2) {
return 1;
}
else if (diffRsCsdiff >= -2) {
if (Rs < 2234) {
if (Cs < 613) {
return 1;
}
else if (Cs >= 613) {
return 8;
}
}
else if (Rs >= 2234) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
}
}
else if (AFD >= 23) {
if (refDCval < 90) {
if (SC < 1765) {
if (MVDiff < 145) {
if (AFD < 25) {
if (Rs < 732) {
return 8;
}
else if (Rs >= 732) {
if (ssDCval < 64) {
if (Cs < 713) {
return 4;
}
else if (Cs >= 713) {
return 8;
}
}
else if (ssDCval >= 64) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
}
else if (AFD >= 25) {
if (RsCsDiff < 127) {
if (negBalance < 59) {
if (CsDiff < 7) {
if (Rs < 881) {
return 4;
}
else if (Rs >= 881) {
return 1;
}
}
else if (CsDiff >= 7) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 2;
}
}
}
else if (negBalance >= 59) {
return 4;
}
}
else if (RsCsDiff >= 127) {
if (diffTSC < -1) {
if (Rs < 746) {
return 8;
}
else if (Rs >= 746) {
return 4;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
else if (MVDiff >= 145) {
if (Rs < 232) {
return 2;
}
else if (Rs >= 232) {
return 8;
}
}
}
else if (SC >= 1765) {
if (negBalance < 72) {
if (mu_mv_mag_sq < 2) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (MVDiff < 43) {
if (tcor < 37) {
return 2;
}
else if (tcor >= 37) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (AFD < 30) {
if (RsDiff < 14) {
return 4;
}
else if (RsDiff >= 14) {
if (Rs < 1916) {
return 8;
}
else if (Rs >= 1916) {
return 4;
}
}
}
else if (AFD >= 30) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
return 2;
}
}
}
}
}
else if (MVDiff >= 43) {
if (RsDiff < 9) {
if (diffMVdiffVal < 18) {
if (Rs < 768) {
if (Rs < 695) {
return 8;
}
else if (Rs >= 695) {
return 4;
}
}
else if (Rs >= 768) {
return 1;
}
}
else if (diffMVdiffVal >= 18) {
if (Rs < 779) {
return 2;
}
else if (Rs >= 779) {
return 4;
}
}
}
else if (RsDiff >= 9) {
if (CsDiff < 10) {
if (mcTcor < 90) {
if (negBalance < 56) {
if (Rs < 1280) {
return 1;
}
else if (Rs >= 1280) {
return 2;
}
}
else if (negBalance >= 56) {
if (Rs < 1880) {
if (RsCsDiff < 131) {
return 2;
}
else if (RsCsDiff >= 131) {
return 4;
}
}
else if (Rs >= 1880) {
return 1;
}
}
}
else if (mcTcor >= 90) {
return 2;
}
}
else if (CsDiff >= 10) {
return 4;
}
}
}
}
}
else if (negBalance >= 72) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 1;
}
}
}
}
else if (refDCval >= 90) {
if (MVDiff < 51) {
if (refDCval < 100) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
if (Cs < 1007) {
if (AFD < 29) {
if (mu_mv_mag_sq < 23) {
return 1;
}
else if (mu_mv_mag_sq >= 23) {
if (Rs < 985) {
return 1;
}
else if (Rs >= 985) {
if (Rs < 1610) {
return 4;
}
else if (Rs >= 1610) {
return 1;
}
}
}
}
else if (AFD >= 29) {
if (Rs < 1417) {
return 4;
}
else if (Rs >= 1417) {
return 2;
}
}
}
else if (Cs >= 1007) {
if (MVDiff < 42) {
if (RsDiff < 10) {
if (RsDiff < 7) {
return 2;
}
else if (RsDiff >= 7) {
if (Cs < 1056) {
return 8;
}
else if (Cs >= 1056) {
return 1;
}
}
}
else if (RsDiff >= 10) {
if (Rs < 1678) {
if (Rs < 1592) {
return 4;
}
else if (Rs >= 1592) {
return 2;
}
}
else if (Rs >= 1678) {
return 8;
}
}
}
else if (MVDiff >= 42) {
return 4;
}
}
}
}
else if (refDCval >= 100) {
if (RsCsDiff < 80) {
return 2;
}
else if (RsCsDiff >= 80) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 2;
}
}
}
}
}
else if (MVDiff >= 51) {
if (diffAFD < 5) {
if (Cs < 939) {
if (mu_mv_mag_sq < 6) {
if (tcor < 67) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (AFD < 27) {
if (Rs < 1138) {
return 4;
}
else if (Rs >= 1138) {
return 8;
}
}
else if (AFD >= 27) {
return 1;
}
}
}
else if (tcor >= 67) {
if (Rs < 1681) {
if (diffRsCsdiff < 18) {
return 2;
}
else if (diffRsCsdiff >= 18) {
return 8;
}
}
else if (Rs >= 1681) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 6) {
if (gchDC < 20) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
if (SC < 1395) {
return 1;
}
else if (SC >= 1395) {
return 4;
}
}
}
else if (gchDC >= 20) {
if (Rs < 594) {
return 2;
}
else if (Rs >= 594) {
return 8;
}
}
}
}
else if (Cs >= 939) {
if (refDCval < 92) {
return 4;
}
else if (refDCval >= 92) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
return 2;
}
}
}
}
else if (diffAFD >= 5) {
if (RsDiff < 12) {
if (Rs < 338) {
return 2;
}
else if (Rs >= 338) {
return 1;
}
}
else if (RsDiff >= 12) {
return 2;
}
}
}
}
}
}
}
}
else if (refDCval >= 116) {
if (RsDiff < 9) {
if (negBalance < 58) {
if (refDCval < 133) {
if (SC < 1969) {
if (refDCval < 126) {
if (Cs < 599) {
return 4;
}
else if (Cs >= 599) {
return 2;
}
}
else if (refDCval >= 126) {
if (Rs < 680) {
return 2;
}
else if (Rs >= 680) {
return 8;
}
}
}
else if (SC >= 1969) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (Cs < 877) {
return 8;
}
else if (Cs >= 877) {
return 4;
}
}
}
}
else if (refDCval >= 133) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
if (RsDiff < 7) {
if (SC < 1892) {
if (Cs < 631) {
return 2;
}
else if (Cs >= 631) {
return 8;
}
}
else if (SC >= 1892) {
if (Rs < 1262) {
return 1;
}
else if (Rs >= 1262) {
return 2;
}
}
}
else if (RsDiff >= 7) {
if (Rs < 1071) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 1031) {
return 2;
}
else if (Rs >= 1031) {
return 1;
}
}
}
else if (Rs >= 1071) {
return 8;
}
}
}
}
}
else if (negBalance >= 58) {
if (CsDiff < 4) {
if (Rs < 536) {
return 2;
}
else if (Rs >= 536) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (MVDiff < 181) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
return 8;
}
}
else if (MVDiff >= 181) {
return 8;
}
}
}
}
else if (CsDiff >= 4) {
if (MVDiff < 24) {
if (Rs < 1245) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
else if (Rs >= 1245) {
return 4;
}
}
else if (MVDiff >= 24) {
if (diffMVdiffVal < 7) {
if (AFD < 19) {
if (Rs < 1163) {
return 2;
}
else if (Rs >= 1163) {
if (Rs < 1494) {
return 8;
}
else if (Rs >= 1494) {
return 1;
}
}
}
else if (AFD >= 19) {
return 1;
}
}
else if (diffMVdiffVal >= 7) {
if (Rs < 1457) {
if (ssDCval < 81) {
return 4;
}
else if (ssDCval >= 81) {
return 2;
}
}
else if (Rs >= 1457) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (mcTcor < 92) {
if (mu_mv_mag_sq < 179) {
if (tcor < 76) {
if (SC < 2547) {
if (RsDiff < 11) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
return 8;
}
}
else if (RsDiff >= 11) {
if (diffAFD < 3) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (diffAFD >= 3) {
if (Cs < 1143) {
return 1;
}
else if (Cs >= 1143) {
return 8;
}
}
}
}
else if (SC >= 2547) {
if (diffAFD < 14) {
if (negBalance < 52) {
if (Rs < 1504) {
return 8;
}
else if (Rs >= 1504) {
if (Rs < 2116) {
return 2;
}
else if (Rs >= 2116) {
return 4;
}
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (diffAFD >= 14) {
return 4;
}
}
}
else if (tcor >= 76) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 179) {
return 1;
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 13) {
if (diffTSC < 2) {
if (refDCval < 146) {
if (RsDiff < 10) {
if (negBalance < 56) {
if (Rs < 1362) {
return 1;
}
else if (Rs >= 1362) {
return 8;
}
}
else if (negBalance >= 56) {
return 1;
}
}
else if (RsDiff >= 10) {
if (AFD < 17) {
if (Rs < 1894) {
return 1;
}
else if (Rs >= 1894) {
return 2;
}
}
else if (AFD >= 17) {
return 1;
}
}
}
else if (refDCval >= 146) {
return 8;
}
}
else if (diffTSC >= 2) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (TSC < 9) {
if (Rs < 1154) {
return 8;
}
else if (Rs >= 1154) {
if (Cs < 553) {
return 1;
}
else if (Cs >= 553) {
return 2;
}
}
}
else if (TSC >= 9) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
if (ssDCval < 130) {
if (Rs < 1300) {
return 8;
}
else if (Rs >= 1300) {
return 1;
}
}
else if (ssDCval >= 130) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (SC < 2092) {
if (Rs < 994) {
if (ssDCval < 64) {
if (Rs < 539) {
if (SC < 1069) {
if (Cs < 547) {
if (Rs < 453) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (Rs >= 453) {
return 1;
}
}
else if (Cs >= 547) {
if (posBalance < 41) {
if (posBalance < 36) {
if (Rs < 360) {
return 1;
}
else if (Rs >= 360) {
return 4;
}
}
else if (posBalance >= 36) {
return 4;
}
}
else if (posBalance >= 41) {
if (RsCsDiff < 19) {
return 1;
}
else if (RsCsDiff >= 19) {
if (Cs < 681) {
return 8;
}
else if (Cs >= 681) {
return 2;
}
}
}
}
}
else if (SC >= 1069) {
if (AFD < 7) {
if (Rs < 388) {
if (Cs < 905) {
return 4;
}
else if (Cs >= 905) {
return 2;
}
}
else if (Rs >= 388) {
if (SC < 1150) {
if (diffAFD < 1) {
if (SC < 1139) {
if (MVDiff < 2) {
if (Rs < 531) {
return 4;
}
else if (Rs >= 531) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Rs < 492) {
return 4;
}
else if (Rs >= 492) {
return 2;
}
}
}
else if (SC >= 1139) {
return 8;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (SC >= 1150) {
return 4;
}
}
}
else if (AFD >= 7) {
if (SC < 1104) {
return 2;
}
else if (SC >= 1104) {
if (Rs < 321) {
if (Rs < 249) {
return 1;
}
else if (Rs >= 249) {
return 2;
}
}
else if (Rs >= 321) {
return 1;
}
}
}
}
}
else if (Rs >= 539) {
if (MVDiff < 1) {
if (SC < 1652) {
if (ssDCval < 57) {
if (mu_mv_mag_sq < 120) {
if (negBalance < 35) {
if (Cs < 828) {
return 1;
}
else if (Cs >= 828) {
if (Rs < 670) {
return 1;
}
else if (Rs >= 670) {
return 8;
}
}
}
else if (negBalance >= 35) {
return 4;
}
}
else if (mu_mv_mag_sq >= 120) {
return 8;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (SC >= 1652) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (Rs < 648) {
if (Cs < 1060) {
return 1;
}
else if (Cs >= 1060) {
return 4;
}
}
else if (Rs >= 648) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 124) {
if (TSC < 4) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
return 1;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 807) {
if (Cs < 1008) {
return 8;
}
else if (Cs >= 1008) {
return 1;
}
}
else if (Rs >= 807) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (Rs < 836) {
if (negBalance < 30) {
if (ssDCval < 45) {
if (refDCval < 37) {
return 2;
}
else if (refDCval >= 37) {
if (Rs < 719) {
return 1;
}
else if (Rs >= 719) {
return 8;
}
}
}
else if (ssDCval >= 45) {
if (TSC < 3) {
if (Cs < 815) {
if (mu_mv_mag_sq < 86) {
return 1;
}
else if (mu_mv_mag_sq >= 86) {
return 4;
}
}
else if (Cs >= 815) {
return 1;
}
}
else if (TSC >= 3) {
if (MVDiff < 2) {
if (RsDiff < 2) {
if (Rs < 733) {
if (Rs < 703) {
return 1;
}
else if (Rs >= 703) {
return 2;
}
}
else if (Rs >= 733) {
return 8;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
}
}
else if (negBalance >= 30) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 109) {
if (SC < 1456) {
if (Rs < 660) {
if (gchDC < 1) {
if (Cs < 719) {
return 4;
}
else if (Cs >= 719) {
return 8;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Rs >= 660) {
return 8;
}
}
else if (SC >= 1456) {
if (AFD < 5) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 1;
}
}
else if (AFD >= 5) {
if (Rs < 708) {
return 8;
}
else if (Rs >= 708) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 730) {
return 1;
}
else if (Rs >= 730) {
return 4;
}
}
}
else if (RsDiff >= 2) {
if (Cs < 795) {
if (MVDiff < 11) {
if (Rs < 762) {
return 4;
}
else if (Rs >= 762) {
if (Cs < 714) {
return 8;
}
else if (Cs >= 714) {
return 4;
}
}
}
else if (MVDiff >= 11) {
if (SC < 1356) {
return 1;
}
else if (SC >= 1356) {
if (Rs < 702) {
return 4;
}
else if (Rs >= 702) {
return 8;
}
}
}
}
else if (Cs >= 795) {
if (MVDiff < 4) {
if (Cs < 1032) {
return 8;
}
else if (Cs >= 1032) {
return 2;
}
}
else if (MVDiff >= 4) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
return 2;
}
}
}
}
}
}
else if (Rs >= 836) {
if (refDCval < 54) {
if (MVDiff < 11) {
if (tcor < 95) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (tcor >= 95) {
if (Cs < 1067) {
return 8;
}
else if (Cs >= 1067) {
if (Rs < 886) {
return 1;
}
else if (Rs >= 886) {
return 8;
}
}
}
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (refDCval >= 54) {
if (negBalance < 40) {
if (RsCsDiff < 7) {
return 1;
}
else if (RsCsDiff >= 7) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
return 1;
}
}
}
}
else if (negBalance >= 40) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (Cs < 683) {
if (negBalance < 28) {
if (MVDiff < 5) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Rs < 765) {
return 2;
}
else if (Rs >= 765) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (CsDiff < 1) {
if (Rs < 496) {
return 1;
}
else if (Rs >= 496) {
return 8;
}
}
else if (CsDiff >= 1) {
if (Cs < 679) {
return 4;
}
else if (Cs >= 679) {
return 2;
}
}
}
}
else if (negBalance >= 28) {
if (refDCval < 74) {
if (AFD < 4) {
if (Cs < 602) {
return 2;
}
else if (Cs >= 602) {
if (Rs < 442) {
return 8;
}
else if (Rs >= 442) {
return 1;
}
}
}
else if (AFD >= 4) {
if (AFD < 8) {
if (refDCval < 67) {
if (RsCsDiff < 7) {
if (MVDiff < 2) {
if (Rs < 750) {
return 1;
}
else if (Rs >= 750) {
return 8;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (RsCsDiff >= 7) {
return 4;
}
}
else if (refDCval >= 67) {
if (Rs < 265) {
return 1;
}
else if (Rs >= 265) {
return 4;
}
}
}
else if (AFD >= 8) {
if (RsCsDiff < 26) {
return 1;
}
else if (RsCsDiff >= 26) {
if (Rs < 545) {
return 2;
}
else if (Rs >= 545) {
return 8;
}
}
}
}
}
else if (refDCval >= 74) {
if (refDCval < 86) {
if (ssDCval < 85) {
if (ssDCval < 77) {
if (Cs < 585) {
if (Rs < 453) {
return 1;
}
else if (Rs >= 453) {
return 4;
}
}
else if (Cs >= 585) {
if (refDCval < 75) {
return 8;
}
else if (refDCval >= 75) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
return 4;
}
}
}
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (ssDCval >= 85) {
return 4;
}
}
else if (refDCval >= 86) {
return 2;
}
}
}
}
else if (Cs >= 683) {
if (MVDiff < 3) {
if (ssDCval < 82) {
if (Cs < 782) {
if (Rs < 854) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (refDCval < 67) {
if (Cs < 736) {
if (Rs < 562) {
return 4;
}
else if (Rs >= 562) {
return 8;
}
}
else if (Cs >= 736) {
return 4;
}
}
else if (refDCval >= 67) {
if (Rs < 627) {
if (Rs < 470) {
return 4;
}
else if (Rs >= 470) {
return 8;
}
}
else if (Rs >= 627) {
if (SC < 1521) {
if (Cs < 775) {
return 1;
}
else if (Cs >= 775) {
return 8;
}
}
else if (SC >= 1521) {
if (Rs < 783) {
return 1;
}
else if (Rs >= 783) {
return 4;
}
}
}
}
}
}
else if (Rs >= 854) {
if (Cs < 740) {
return 2;
}
else if (Cs >= 740) {
return 4;
}
}
}
else if (Cs >= 782) {
if (Rs < 679) {
return 8;
}
else if (Rs >= 679) {
if (Rs < 895) {
if (refDCval < 69) {
if (mu_mv_mag_sq < 111) {
if (Rs < 865) {
if (SC < 1962) {
return 8;
}
else if (SC >= 1962) {
return 1;
}
}
else if (Rs >= 865) {
return 1;
}
}
else if (mu_mv_mag_sq >= 111) {
if (Rs < 729) {
return 8;
}
else if (Rs >= 729) {
if (Cs < 1048) {
return 4;
}
else if (Cs >= 1048) {
return 2;
}
}
}
}
else if (refDCval >= 69) {
if (Rs < 774) {
return 1;
}
else if (Rs >= 774) {
if (mu_mv_mag_sq < 71) {
return 2;
}
else if (mu_mv_mag_sq >= 71) {
if (refDCval < 78) {
if (Cs < 1007) {
return 4;
}
else if (Cs >= 1007) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (refDCval >= 78) {
return 8;
}
}
}
}
}
else if (Rs >= 895) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 122) {
if (Rs < 928) {
if (ssDCval < 72) {
if (Cs < 939) {
return 8;
}
else if (Cs >= 939) {
return 4;
}
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (Rs >= 928) {
return 8;
}
}
else if (mu_mv_mag_sq >= 122) {
return 4;
}
}
else if (mcTcor >= 99) {
if (ssDCval < 79) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (ssDCval >= 79) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 82) {
if (negBalance < 30) {
if (Cs < 761) {
return 1;
}
else if (Cs >= 761) {
if (Rs < 663) {
return 2;
}
else if (Rs >= 663) {
if (Rs < 694) {
return 1;
}
else if (Rs >= 694) {
return 2;
}
}
}
}
else if (negBalance >= 30) {
if (Cs < 880) {
return 2;
}
else if (Cs >= 880) {
return 4;
}
}
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 79) {
if (Rs < 587) {
if (SC < 1184) {
if (Rs < 366) {
return 1;
}
else if (Rs >= 366) {
return 4;
}
}
else if (SC >= 1184) {
if (ssDCval < 86) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (RsDiff < 4) {
if (Cs < 1210) {
return 2;
}
else if (Cs >= 1210) {
if (Rs < 410) {
return 1;
}
else if (Rs >= 410) {
return 2;
}
}
}
else if (RsDiff >= 4) {
return 1;
}
}
}
else if (ssDCval >= 86) {
return 2;
}
}
}
else if (Rs >= 587) {
if (Cs < 998) {
if (posBalance < 53) {
if (Rs < 919) {
if (Cs < 747) {
return 4;
}
else if (Cs >= 747) {
return 2;
}
}
else if (Rs >= 919) {
return 1;
}
}
else if (posBalance >= 53) {
if (diffRsCsdiff < -13) {
return 8;
}
else if (diffRsCsdiff >= -13) {
if (refDCval < 75) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
return 4;
}
}
else if (refDCval >= 75) {
if (Cs < 908) {
return 4;
}
else if (Cs >= 908) {
return 2;
}
}
}
}
}
else if (Cs >= 998) {
if (refDCval < 83) {
if (Rs < 621) {
return 1;
}
else if (Rs >= 621) {
if (posBalance < 65) {
return 2;
}
else if (posBalance >= 65) {
return 4;
}
}
}
else if (refDCval >= 83) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (AFD < 3) {
if (Rs < 658) {
return 2;
}
else if (Rs >= 658) {
if (Cs < 947) {
return 1;
}
else if (Cs >= 947) {
return 2;
}
}
}
else if (AFD >= 3) {
if (negBalance < 29) {
if (posBalance < 53) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
return 1;
}
}
else if (posBalance >= 53) {
if (Cs < 791) {
return 8;
}
else if (Cs >= 791) {
return 2;
}
}
}
else if (negBalance >= 29) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 994) {
if (posBalance < 39) {
if (ssDCval < 51) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (Rs < 1288) {
return 8;
}
else if (Rs >= 1288) {
return 1;
}
}
}
else if (ssDCval >= 51) {
if (RsCsDiff < 12) {
if (mu_mv_mag_sq < 126) {
if (SC < 2072) {
return 1;
}
else if (SC >= 2072) {
if (SC < 2074) {
return 8;
}
else if (SC >= 2074) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 736) {
if (Rs < 1102) {
return 8;
}
else if (Rs >= 1102) {
return 1;
}
}
else if (Cs >= 736) {
return 8;
}
}
}
else if (RsCsDiff >= 12) {
if (SC < 2031) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (SC >= 2031) {
return 1;
}
}
}
}
else if (posBalance >= 39) {
if (tcor < 99) {
if (MVDiff < 12) {
if (refDCval < 76) {
if (TSC < 5) {
if (mu_mv_mag_sq < 119) {
if (Cs < 724) {
if (AFD < 3) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 724) {
if (RsCsDiff < 16) {
if (Cs < 1035) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (diffRsCsdiff < -5) {
if (ssDCval < 65) {
if (Cs < 751) {
return 8;
}
else if (Cs >= 751) {
return 1;
}
}
else if (ssDCval >= 65) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
return 1;
}
}
}
else if (Cs >= 1035) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (ssDCval < 52) {
if (Rs < 1267) {
return 8;
}
else if (Rs >= 1267) {
if (Rs < 1357) {
return 4;
}
else if (Rs >= 1357) {
return 4;
}
}
}
else if (ssDCval >= 52) {
if (Rs < 1321) {
if (mu_mv_mag_sq < 125) {
return 1;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1137) {
return 4;
}
else if (Rs >= 1137) {
return 1;
}
}
}
else if (Rs >= 1321) {
if (Cs < 651) {
return 8;
}
else if (Cs >= 651) {
return 1;
}
}
}
}
}
else if (TSC >= 5) {
if (RsCsDiff < 38) {
if (Rs < 1032) {
return 4;
}
else if (Rs >= 1032) {
return 8;
}
}
else if (RsCsDiff >= 38) {
return 1;
}
}
}
else if (refDCval >= 76) {
if (ssDCval < 86) {
if (Cs < 989) {
return 8;
}
else if (Cs >= 989) {
if (Rs < 1010) {
return 8;
}
else if (Rs >= 1010) {
return 4;
}
}
}
else if (ssDCval >= 86) {
return 1;
}
}
}
else if (MVDiff >= 12) {
if (RsDiff < 2) {
if (Rs < 1217) {
return 8;
}
else if (Rs >= 1217) {
return 1;
}
}
else if (RsDiff >= 2) {
if (MVDiff < 27) {
if (diffMVdiffVal < 5) {
if (refDCval < 63) {
return 2;
}
else if (refDCval >= 63) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
else if (MVDiff >= 27) {
if (Rs < 1030) {
return 8;
}
else if (Rs >= 1030) {
return 4;
}
}
}
}
}
else if (tcor >= 99) {
if (SC < 1814) {
if (ssDCval < 62) {
if (Rs < 1051) {
return 8;
}
else if (Rs >= 1051) {
return 1;
}
}
else if (ssDCval >= 62) {
if (Rs < 1143) {
return 2;
}
else if (Rs >= 1143) {
return 1;
}
}
}
else if (SC >= 1814) {
if (Rs < 1372) {
if (AFD < 3) {
if (mu_mv_mag_sq < 126) {
return 1;
}
else if (mu_mv_mag_sq >= 126) {
if (SC < 1891) {
return 4;
}
else if (SC >= 1891) {
if (Rs < 1287) {
return 8;
}
else if (Rs >= 1287) {
return 1;
}
}
}
}
else if (AFD >= 3) {
if (Cs < 943) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (Cs >= 943) {
return 4;
}
}
}
else if (Rs >= 1372) {
return 8;
}
}
}
}
}
}
else if (SC >= 2092) {
if (Rs < 1763) {
if (Cs < 1064) {
if (mu_mv_mag_sq < 121) {
if (SC < 2425) {
if (TSC < 5) {
if (Rs < 1287) {
if (ssDCval < 69) {
if (negBalance < 35) {
if (mu_mv_mag_sq < 104) {
return 8;
}
else if (mu_mv_mag_sq >= 104) {
return 1;
}
}
else if (negBalance >= 35) {
return 1;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (Rs >= 1287) {
if (ssDCval < 61) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
if (Rs < 1460) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (Rs >= 1460) {
return 1;
}
}
}
else if (ssDCval >= 61) {
if (Cs < 617) {
if (mu_mv_mag_sq < 115) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (Rs < 1721) {
return 2;
}
else if (Rs >= 1721) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 1685) {
return 8;
}
else if (Rs >= 1685) {
return 1;
}
}
}
else if (Cs >= 617) {
if (refDCval < 63) {
if (AFD < 6) {
if (posBalance < 36) {
if (Cs < 802) {
return 1;
}
else if (Cs >= 802) {
return 8;
}
}
else if (posBalance >= 36) {
return 1;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (refDCval >= 63) {
if (refDCval < 79) {
return 1;
}
else if (refDCval >= 79) {
if (Rs < 1427) {
return 8;
}
else if (Rs >= 1427) {
return 1;
}
}
}
}
}
}
}
else if (TSC >= 5) {
if (Rs < 1419) {
if (MVDiff < 4) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (Cs < 994) {
return 4;
}
else if (Cs >= 994) {
if (Rs < 1280) {
return 8;
}
else if (Rs >= 1280) {
return 4;
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 1129) {
return 2;
}
else if (Rs >= 1129) {
return 8;
}
}
}
else if (Rs >= 1419) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (Cs < 704) {
return 4;
}
else if (Cs >= 704) {
return 1;
}
}
}
}
}
else if (SC >= 2425) {
if (Cs < 852) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (Rs < 1741) {
return 2;
}
else if (Rs >= 1741) {
return 4;
}
}
}
}
else if (Cs >= 852) {
if (posBalance < 45) {
if (ssDCval < 70) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (ssDCval >= 70) {
if (Rs < 1544) {
return 8;
}
else if (Rs >= 1544) {
return 1;
}
}
}
else if (posBalance >= 45) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 1585) {
return 8;
}
else if (Rs >= 1585) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (ssDCval < 49) {
if (Cs < 1024) {
if (gchDC < 1) {
if (Cs < 1016) {
if (mu_mv_mag_sq < 123) {
return 4;
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1577) {
return 4;
}
else if (Rs >= 1577) {
return 4;
}
}
}
else if (Cs >= 1016) {
return 8;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Cs >= 1024) {
return 4;
}
}
else if (ssDCval >= 49) {
if (ssDCval < 69) {
if (mcTcor < 99) {
if (SC < 2600) {
if (mu_mv_mag_sq < 126) {
if (Rs < 1274) {
if (Cs < 873) {
return 8;
}
else if (Cs >= 873) {
return 1;
}
}
else if (Rs >= 1274) {
return 8;
}
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
else if (SC >= 2600) {
return 1;
}
}
else if (mcTcor >= 99) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
return 1;
}
}
}
}
}
else if (ssDCval >= 69) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
if (Cs < 770) {
return 1;
}
else if (Cs >= 770) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1064) {
if (RsDiff < 1) {
if (ssDCval < 64) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
return 1;
}
}
else if (ssDCval >= 64) {
if (posBalance < 48) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 4;
}
}
}
else if (posBalance >= 48) {
if (Cs < 1168) {
return 4;
}
else if (Cs >= 1168) {
return 1;
}
}
}
}
else if (RsDiff >= 1) {
if (TSC < 3) {
if (ssDCval < 52) {
if (negBalance < 25) {
if (TSC < 2) {
if (SC < 2919) {
if (mu_mv_mag_sq < 128) {
if (negBalance < 16) {
return 8;
}
else if (negBalance >= 16) {
if (Rs < 1704) {
return 4;
}
else if (Rs >= 1704) {
if (Cs < 1120) {
if (Rs < 1734) {
return 4;
}
else if (Rs >= 1734) {
return 8;
}
}
else if (Cs >= 1120) {
if (Cs < 1138) {
return 8;
}
else if (Cs >= 1138) {
if (Rs < 1753) {
return 8;
}
else if (Rs >= 1753) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (SC < 2843) {
if (Rs < 1683) {
return 4;
}
else if (Rs >= 1683) {
return 4;
}
}
else if (SC >= 2843) {
if (Rs < 1698) {
return 4;
}
else if (Rs >= 1698) {
if (Rs < 1729) {
if (Rs < 1713) {
return 4;
}
else if (Rs >= 1713) {
return 8;
}
}
else if (Rs >= 1729) {
return 8;
}
}
}
}
}
else if (SC >= 2919) {
return 4;
}
}
else if (TSC >= 2) {
if (Cs < 1294) {
if (Cs < 1098) {
if (negBalance < 22) {
if (Cs < 1095) {
return 8;
}
else if (Cs >= 1095) {
if (Rs < 1702) {
return 4;
}
else if (Rs >= 1702) {
return 4;
}
}
}
else if (negBalance >= 22) {
return 4;
}
}
else if (Cs >= 1098) {
if (Cs < 1151) {
if (Cs < 1129) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (Cs >= 1129) {
if (posBalance < 35) {
return 4;
}
else if (posBalance >= 35) {
if (Cs < 1134) {
if (Rs < 1733) {
return 8;
}
else if (Rs >= 1733) {
return 4;
}
}
else if (Cs >= 1134) {
if (CsDiff < 2) {
if (Rs < 1699) {
if (Rs < 1679) {
return 4;
}
else if (Rs >= 1679) {
return 4;
}
}
else if (Rs >= 1699) {
if (Rs < 1731) {
return 4;
}
else if (Rs >= 1731) {
return 4;
}
}
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
}
else if (Cs >= 1151) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
if (Rs < 1479) {
return 4;
}
else if (Rs >= 1479) {
return 8;
}
}
}
}
}
else if (Cs >= 1294) {
return 8;
}
}
}
else if (negBalance >= 25) {
return 8;
}
}
else if (ssDCval >= 52) {
if (ssDCval < 62) {
if (SC < 2511) {
return 1;
}
else if (SC >= 2511) {
return 4;
}
}
else if (ssDCval >= 62) {
if (tcor < 98) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (tcor >= 98) {
if (negBalance < 37) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (Cs < 1191) {
if (negBalance < 24) {
if (Rs < 1687) {
return 8;
}
else if (Rs >= 1687) {
return 1;
}
}
else if (negBalance >= 24) {
if (Cs < 1079) {
return 8;
}
else if (Cs >= 1079) {
if (posBalance < 45) {
if (Rs < 1149) {
return 8;
}
else if (Rs >= 1149) {
return 4;
}
}
else if (posBalance >= 45) {
return 4;
}
}
}
}
else if (Cs >= 1191) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
return 4;
}
}
}
}
else if (negBalance >= 37) {
return 1;
}
}
}
}
}
else if (TSC >= 3) {
if (Rs < 1432) {
if (RsDiff < 2) {
if (tcor < 98) {
if (Rs < 966) {
return 2;
}
else if (Rs >= 966) {
if (Rs < 1221) {
return 8;
}
else if (Rs >= 1221) {
return 1;
}
}
}
else if (tcor >= 98) {
if (Cs < 1260) {
if (Rs < 940) {
return 2;
}
else if (Rs >= 940) {
return 4;
}
}
else if (Cs >= 1260) {
if (Cs < 1305) {
return 1;
}
else if (Cs >= 1305) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (mu_mv_mag_sq < 66) {
if (RsDiff < 4) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (RsDiff >= 4) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Rs < 1114) {
return 2;
}
else if (Rs >= 1114) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 1286) {
if (Cs < 1124) {
return 8;
}
else if (Cs >= 1124) {
if (AFD < 5) {
if (diffRsCsdiff < 6) {
if (diffTSC < -2) {
return 8;
}
else if (diffTSC >= -2) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
return 8;
}
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (Rs >= 1286) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1432) {
if (negBalance < 37) {
return 8;
}
else if (negBalance >= 37) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 1603) {
return 4;
}
else if (Rs >= 1603) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 1763) {
if (TSC < 4) {
if (refDCval < 83) {
if (ssDCval < 56) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (Cs < 835) {
if (SC < 2886) {
return 4;
}
else if (SC >= 2886) {
return 8;
}
}
else if (Cs >= 835) {
if (SC < 2920) {
if (refDCval < 48) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (Rs < 1780) {
if (Cs < 1116) {
return 4;
}
else if (Cs >= 1116) {
if (Cs < 1137) {
return 4;
}
else if (Cs >= 1137) {
return 4;
}
}
}
else if (Rs >= 1780) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (Rs < 1802) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Cs < 1116) {
if (Rs < 1793) {
return 4;
}
else if (Rs >= 1793) {
return 8;
}
}
else if (Cs >= 1116) {
return 8;
}
}
}
else if (Rs >= 1802) {
if (negBalance < 18) {
if (Rs < 1826) {
if (Rs < 1813) {
return 4;
}
else if (Rs >= 1813) {
return 8;
}
}
else if (Rs >= 1826) {
return 4;
}
}
else if (negBalance >= 18) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 48) {
return 4;
}
}
else if (SC >= 2920) {
if (posBalance < 39) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
return 4;
}
}
else if (posBalance >= 39) {
if (Rs < 2172) {
return 8;
}
else if (Rs >= 2172) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 56) {
if (TSC < 3) {
if (ssDCval < 80) {
if (ssDCval < 72) {
if (mu_mv_mag_sq < 96) {
if (SC < 3084) {
return 4;
}
else if (SC >= 3084) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
if (SC < 2862) {
if (Cs < 999) {
return 1;
}
else if (Cs >= 999) {
return 4;
}
}
else if (SC >= 2862) {
if (tcor < 99) {
if (gchDC < 1) {
if (Rs < 1933) {
return 1;
}
else if (Rs >= 1933) {
if (mu_mv_mag_sq < 111) {
return 4;
}
else if (mu_mv_mag_sq >= 111) {
return 1;
}
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (tcor >= 99) {
if (Rs < 2062) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1904) {
return 1;
}
else if (Rs >= 1904) {
return 4;
}
}
}
else if (Rs >= 2062) {
return 1;
}
}
}
}
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (ssDCval >= 80) {
if (Cs < 875) {
return 4;
}
else if (Cs >= 875) {
return 8;
}
}
}
else if (TSC >= 3) {
if (Cs < 840) {
if (Rs < 2235) {
return 1;
}
else if (Rs >= 2235) {
if (Rs < 2274) {
return 8;
}
else if (Rs >= 2274) {
return 1;
}
}
}
else if (Cs >= 840) {
if (ssDCval < 70) {
if (MVDiff < 1) {
if (Rs < 1819) {
return 8;
}
else if (Rs >= 1819) {
return 1;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
}
}
else if (refDCval >= 83) {
if (mu_mv_mag_sq < 100) {
if (SC < 2568) {
return 4;
}
else if (SC >= 2568) {
if (Rs < 2196) {
return 1;
}
else if (Rs >= 2196) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 100) {
return 4;
}
}
}
else if (TSC >= 4) {
if (Cs < 878) {
if (refDCval < 71) {
if (negBalance < 31) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 8;
}
}
else if (negBalance >= 31) {
if (posBalance < 48) {
if (Rs < 2097) {
if (Rs < 1886) {
return 8;
}
else if (Rs >= 1886) {
return 1;
}
}
else if (Rs >= 2097) {
if (Rs < 2240) {
return 2;
}
else if (Rs >= 2240) {
return 4;
}
}
}
else if (posBalance >= 48) {
return 2;
}
}
}
else if (refDCval >= 71) {
if (mu_mv_mag_sq < 93) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
if (AFD < 10) {
if (Rs < 2115) {
return 4;
}
else if (Rs >= 2115) {
return 1;
}
}
else if (AFD >= 10) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
if (Rs < 2055) {
return 4;
}
else if (Rs >= 2055) {
if (Rs < 2256) {
return 4;
}
else if (Rs >= 2256) {
return 8;
}
}
}
}
}
}
else if (Cs >= 878) {
if (Rs < 2208) {
if (Cs < 1050) {
if (mu_mv_mag_sq < 70) {
if (gchDC < 1) {
if (Rs < 1905) {
return 8;
}
else if (Rs >= 1905) {
return 1;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 70) {
return 8;
}
}
else if (Cs >= 1050) {
if (diffTSC < -2) {
return 1;
}
else if (diffTSC >= -2) {
return 8;
}
}
}
else if (Rs >= 2208) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 100) {
if (SC < 1296) {
if (MVDiff < 4) {
if (Rs < 415) {
if (ssDCval < 104) {
if (Rs < 384) {
return 1;
}
else if (Rs >= 384) {
return 8;
}
}
else if (ssDCval >= 104) {
if (Rs < 246) {
return 2;
}
else if (Rs >= 246) {
return 4;
}
}
}
else if (Rs >= 415) {
if (TSC < 4) {
if (Rs < 518) {
return 1;
}
else if (Rs >= 518) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
if (Rs < 631) {
return 2;
}
else if (Rs >= 631) {
return 4;
}
}
}
}
else if (TSC >= 4) {
if (MVDiff < 3) {
if (ssDCval < 115) {
if (Cs < 555) {
return 8;
}
else if (Cs >= 555) {
return 1;
}
}
else if (ssDCval >= 115) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Rs < 527) {
return 8;
}
else if (Rs >= 527) {
return 2;
}
}
}
}
}
else if (MVDiff >= 4) {
if (negBalance < 70) {
if (TSC < 4) {
if (Rs < 414) {
if (MVDiff < 6) {
if (Rs < 251) {
return 2;
}
else if (Rs >= 251) {
return 4;
}
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (Rs >= 414) {
if (negBalance < 25) {
if (ssDCval < 112) {
return 1;
}
else if (ssDCval >= 112) {
if (Cs < 585) {
if (Rs < 637) {
return 2;
}
else if (Rs >= 637) {
return 1;
}
}
else if (Cs >= 585) {
return 4;
}
}
}
else if (negBalance >= 25) {
if (ssDCval < 127) {
if (diffMVdiffVal < 4) {
return 2;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 556) {
return 2;
}
else if (Rs >= 556) {
if (Rs < 602) {
return 1;
}
else if (Rs >= 602) {
return 8;
}
}
}
}
else if (ssDCval >= 127) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (ssDCval < 163) {
if (ssDCval < 100) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
else if (ssDCval >= 100) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (diffRsCsdiff < 14) {
return 2;
}
else if (diffRsCsdiff >= 14) {
return 4;
}
}
}
}
else if (ssDCval >= 163) {
if (Cs < 597) {
return 2;
}
else if (Cs >= 597) {
return 4;
}
}
}
}
}
else if (negBalance >= 70) {
if (Rs < 603) {
return 8;
}
else if (Rs >= 603) {
return 1;
}
}
}
}
else if (SC >= 1296) {
if (MVDiff < 4) {
if (Rs < 667) {
if (ssDCval < 95) {
if (Rs < 516) {
return 8;
}
else if (Rs >= 516) {
return 2;
}
}
else if (ssDCval >= 95) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 640) {
return 8;
}
else if (Rs >= 640) {
return 2;
}
}
}
}
else if (Rs >= 667) {
if (Cs < 891) {
if (ssDCval < 95) {
if (Rs < 1274) {
return 8;
}
else if (Rs >= 1274) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (ssDCval >= 95) {
if (Cs < 641) {
if (Rs < 1988) {
return 1;
}
else if (Rs >= 1988) {
return 8;
}
}
else if (Cs >= 641) {
if (negBalance < 26) {
return 4;
}
else if (negBalance >= 26) {
if (Rs < 801) {
if (Rs < 677) {
return 8;
}
else if (Rs >= 677) {
return 4;
}
}
else if (Rs >= 801) {
return 8;
}
}
}
}
}
else if (Cs >= 891) {
if (Rs < 1382) {
if (Cs < 1199) {
return 4;
}
else if (Cs >= 1199) {
if (Cs < 1263) {
return 8;
}
else if (Cs >= 1263) {
return 4;
}
}
}
else if (Rs >= 1382) {
if (Rs < 1973) {
if (Rs < 1597) {
if (Rs < 1492) {
return 1;
}
else if (Rs >= 1492) {
return 8;
}
}
else if (Rs >= 1597) {
return 2;
}
}
else if (Rs >= 1973) {
return 8;
}
}
}
}
}
else if (MVDiff >= 4) {
if (RsDiff < 8) {
if (negBalance < 31) {
if (posBalance < 60) {
if (ssDCval < 130) {
if (posBalance < 42) {
if (Rs < 1560) {
return 8;
}
else if (Rs >= 1560) {
return 1;
}
}
else if (posBalance >= 42) {
return 1;
}
}
else if (ssDCval >= 130) {
if (Rs < 2155) {
return 4;
}
else if (Rs >= 2155) {
return 8;
}
}
}
else if (posBalance >= 60) {
if (Cs < 801) {
return 2;
}
else if (Cs >= 801) {
if (mu_mv_mag_sq < 65) {
if (Rs < 1151) {
return 2;
}
else if (Rs >= 1151) {
return 4;
}
}
else if (mu_mv_mag_sq >= 65) {
return 8;
}
}
}
}
else if (negBalance >= 31) {
if (mu_mv_mag_sq < 21) {
if (Rs < 2039) {
if (Rs < 885) {
if (negBalance < 70) {
if (AFD < 13) {
if (SC < 1662) {
return 2;
}
else if (SC >= 1662) {
if (Cs < 1048) {
if (Rs < 769) {
return 4;
}
else if (Rs >= 769) {
return 1;
}
}
else if (Cs >= 1048) {
return 2;
}
}
}
else if (AFD >= 13) {
if (negBalance < 43) {
return 2;
}
else if (negBalance >= 43) {
if (diffAFD < -2) {
if (Rs < 762) {
return 2;
}
else if (Rs >= 762) {
return 4;
}
}
else if (diffAFD >= -2) {
if (Rs < 827) {
return 1;
}
else if (Rs >= 827) {
return 8;
}
}
}
}
}
else if (negBalance >= 70) {
return 8;
}
}
else if (Rs >= 885) {
if (negBalance < 51) {
if (refDCval < 99) {
if (TSC < 6) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (Rs < 1084) {
return 1;
}
else if (Rs >= 1084) {
return 2;
}
}
}
else if (TSC >= 6) {
if (tcor < 93) {
if (Cs < 1048) {
return 1;
}
else if (Cs >= 1048) {
return 4;
}
}
else if (tcor >= 93) {
return 8;
}
}
}
else if (refDCval >= 99) {
if (diffAFD < -2) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
if (Rs < 1246) {
return 8;
}
else if (Rs >= 1246) {
return 2;
}
}
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < 25) {
if (SC < 1569) {
return 1;
}
else if (SC >= 1569) {
if (MVDiff < 28) {
return 8;
}
else if (MVDiff >= 28) {
return 1;
}
}
}
else if (diffRsCsdiff >= 25) {
if (Rs < 1434) {
return 4;
}
else if (Rs >= 1434) {
return 2;
}
}
}
}
}
else if (negBalance >= 51) {
if (ssDCval < 120) {
if (TSC < 7) {
if (RsCsDiff < 80) {
if (SC < 2862) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
else if (SC >= 2862) {
return 8;
}
}
else if (RsCsDiff >= 80) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (ssDCval >= 120) {
if (diffTSC < -3) {
return 2;
}
else if (diffTSC >= -3) {
return 8;
}
}
}
}
}
else if (Rs >= 2039) {
if (ssDCval < 113) {
return 4;
}
else if (ssDCval >= 113) {
if (Cs < 976) {
return 1;
}
else if (Cs >= 976) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (ssDCval < 160) {
if (negBalance < 37) {
if (mcTcor < 99) {
if (Cs < 669) {
if (Rs < 1833) {
return 2;
}
else if (Rs >= 1833) {
return 1;
}
}
else if (Cs >= 669) {
if (mu_mv_mag_sq < 38) {
if (Rs < 1338) {
return 4;
}
else if (Rs >= 1338) {
if (Rs < 1729) {
return 8;
}
else if (Rs >= 1729) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 38) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 1;
}
}
}
else if (negBalance >= 37) {
if (diffRsCsdiff < -20) {
if (Rs < 1354) {
if (SC < 1821) {
return 1;
}
else if (SC >= 1821) {
return 8;
}
}
else if (Rs >= 1354) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
}
else if (diffRsCsdiff >= -20) {
if (Cs < 857) {
if (RsDiff < 6) {
if (Rs < 1010) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
if (ssDCval < 134) {
if (Cs < 680) {
if (Cs < 673) {
return 8;
}
else if (Cs >= 673) {
return 1;
}
}
else if (Cs >= 680) {
return 8;
}
}
else if (ssDCval >= 134) {
if (Rs < 865) {
return 2;
}
else if (Rs >= 865) {
return 1;
}
}
}
}
else if (Rs >= 1010) {
return 8;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (Cs >= 857) {
return 8;
}
}
}
}
else if (ssDCval >= 160) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
}
else if (RsDiff >= 8) {
if (Cs < 797) {
if (Rs < 1777) {
if (tcor < 87) {
if (diffAFD < 5) {
if (CsDiff < 7) {
if (MVDiff < 186) {
return 8;
}
else if (MVDiff >= 186) {
return 2;
}
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (diffAFD >= 5) {
return 1;
}
}
else if (tcor >= 87) {
if (refDCval < 139) {
return 1;
}
else if (refDCval >= 139) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 2;
}
}
}
}
else if (Rs >= 1777) {
if (mu_mv_mag_sq < 32) {
return 1;
}
else if (mu_mv_mag_sq >= 32) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 2;
}
}
}
}
else if (Cs >= 797) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
if (Rs < 1947) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
if (negBalance < 37) {
return 8;
}
else if (negBalance >= 37) {
return 2;
}
}
}
}
}
else if (Rs >= 1947) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (ssDCval < 133) {
if (Cs < 1215) {
if (Rs < 342) {
if (ssDCval < 123) {
if (mu_mv_mag_sq < 117) {
return 4;
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 342) {
if (Cs < 962) {
if (refDCval < 97) {
if (mu_mv_mag_sq < 126) {
if (Rs < 1043) {
if (Rs < 685) {
return 8;
}
else if (Rs >= 685) {
if (Cs < 731) {
return 8;
}
else if (Cs >= 731) {
return 1;
}
}
}
else if (Rs >= 1043) {
if (ssDCval < 93) {
if (refDCval < 89) {
if (Rs < 2143) {
if (Cs < 685) {
return 4;
}
else if (Cs >= 685) {
return 8;
}
}
else if (Rs >= 2143) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
}
else if (refDCval >= 89) {
if (Cs < 665) {
return 1;
}
else if (Cs >= 665) {
return 8;
}
}
}
else if (ssDCval >= 93) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 2155) {
return 8;
}
else if (Rs >= 2155) {
if (Cs < 747) {
return 8;
}
else if (Cs >= 747) {
return 4;
}
}
}
}
else if (refDCval >= 97) {
if (mu_mv_mag_sq < 103) {
if (ssDCval < 106) {
return 8;
}
else if (ssDCval >= 106) {
if (refDCval < 117) {
if (Rs < 402) {
return 4;
}
else if (Rs >= 402) {
return 1;
}
}
else if (refDCval >= 117) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 1668) {
if (SC < 1360) {
if (Cs < 782) {
if (Cs < 584) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (Rs < 451) {
return 8;
}
else if (Rs >= 451) {
return 1;
}
}
}
else if (Cs >= 584) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
if (negBalance < 31) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (negBalance >= 31) {
return 8;
}
}
}
}
else if (Cs >= 782) {
if (AFD < 3) {
if (Rs < 468) {
return 4;
}
else if (Rs >= 468) {
return 8;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (SC >= 1360) {
if (diffRsCsdiff < -19) {
if (Cs < 808) {
return 1;
}
else if (Cs >= 808) {
return 8;
}
}
else if (diffRsCsdiff >= -19) {
return 8;
}
}
}
else if (Rs >= 1668) {
if (MVDiff < 1) {
if (Rs < 2009) {
return 1;
}
else if (Rs >= 2009) {
return 8;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
else if (Cs >= 962) {
if (ssDCval < 121) {
if (diffRsCsdiff < -11) {
if (Rs < 1722) {
return 8;
}
else if (Rs >= 1722) {
if (Rs < 1902) {
return 2;
}
else if (Rs >= 1902) {
return 8;
}
}
}
else if (diffRsCsdiff >= -11) {
if (tcor < 95) {
if (Rs < 1055) {
return 4;
}
else if (Rs >= 1055) {
return 8;
}
}
else if (tcor >= 95) {
if (Cs < 996) {
if (mu_mv_mag_sq < 109) {
if (MVDiff < 3) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
else if (Cs >= 996) {
return 8;
}
}
}
}
else if (ssDCval >= 121) {
if (Rs < 1401) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 1401) {
return 2;
}
}
}
}
}
else if (Cs >= 1215) {
if (Rs < 631) {
return 1;
}
else if (Rs >= 631) {
if (ssDCval < 95) {
return 2;
}
else if (ssDCval >= 95) {
if (Cs < 1233) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 1228) {
return 1;
}
else if (Rs >= 1228) {
return 2;
}
}
}
else if (Cs >= 1233) {
return 8;
}
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1147) {
if (Cs < 858) {
return 2;
}
else if (Cs >= 858) {
return 8;
}
}
else if (Rs >= 1147) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2322) {
if (Rs < 3279) {
if (ssDCval < 90) {
if (RsDiff < 5) {
if (mu_mv_mag_sq < 125) {
if (Cs < 669) {
if (Cs < 386) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 1;
}
}
else if (Cs >= 386) {
if (posBalance < 55) {
if (Rs < 2659) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 1;
}
}
}
else if (Rs >= 2659) {
return 8;
}
}
else if (posBalance >= 55) {
if (Rs < 2410) {
return 2;
}
else if (Rs >= 2410) {
if (Rs < 2516) {
return 1;
}
else if (Rs >= 2516) {
if (Rs < 2617) {
return 8;
}
else if (Rs >= 2617) {
return 8;
}
}
}
}
}
}
else if (Cs >= 669) {
if (MVDiff < 5) {
if (diffRsCsdiff < -9) {
if (ssDCval < 75) {
if (mu_mv_mag_sq < 90) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (Cs < 1120) {
return 4;
}
else if (Cs >= 1120) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
}
else if (ssDCval >= 75) {
return 1;
}
}
else if (diffRsCsdiff >= -9) {
if (mcTcor < 98) {
if (Rs < 2364) {
if (Cs < 1152) {
return 8;
}
else if (Cs >= 1152) {
return 4;
}
}
else if (Rs >= 2364) {
if (Cs < 1183) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
if (Rs < 2808) {
return 4;
}
else if (Rs >= 2808) {
return 4;
}
}
}
}
else if (Cs >= 1183) {
if (TSC < 4) {
if (Cs < 1231) {
return 4;
}
else if (Cs >= 1231) {
return 4;
}
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (mcTcor >= 98) {
if (SC < 4125) {
if (Rs < 2689) {
if (posBalance < 39) {
if (Rs < 2342) {
return 2;
}
else if (Rs >= 2342) {
return 8;
}
}
else if (posBalance >= 39) {
return 1;
}
}
else if (Rs >= 2689) {
return 2;
}
}
else if (SC >= 4125) {
if (negBalance < 27) {
if (Rs < 3195) {
if (Rs < 3073) {
return 4;
}
else if (Rs >= 3073) {
return 4;
}
}
else if (Rs >= 3195) {
return 8;
}
}
else if (negBalance >= 27) {
return 4;
}
}
}
}
}
else if (MVDiff >= 5) {
if (Rs < 2459) {
if (refDCval < 51) {
return 4;
}
else if (refDCval >= 51) {
if (SC < 3128) {
return 4;
}
else if (SC >= 3128) {
return 8;
}
}
}
else if (Rs >= 2459) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (SC < 3628) {
if (SC < 3604) {
return 8;
}
else if (SC >= 3604) {
if (Rs < 2566) {
return 2;
}
else if (Rs >= 2566) {
return 4;
}
}
}
else if (SC >= 3628) {
return 8;
}
}
}
else if (RsDiff >= 5) {
if (posBalance < 49) {
if (negBalance < 51) {
if (ssDCval < 69) {
if (RsDiff < 7) {
if (tcor < 91) {
if (SC < 3870) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (SC >= 3870) {
return 4;
}
}
else if (tcor >= 91) {
if (Rs < 3026) {
if (MVDiff < 3) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (TSC < 5) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (TSC >= 5) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (Rs >= 3026) {
return 4;
}
}
}
else if (RsDiff >= 7) {
if (posBalance < 40) {
if (refDCval < 57) {
if (Cs < 1274) {
return 4;
}
else if (Cs >= 1274) {
return 8;
}
}
else if (refDCval >= 57) {
if (TSC < 7) {
if (Rs < 2943) {
return 1;
}
else if (Rs >= 2943) {
return 8;
}
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (posBalance >= 40) {
if (mu_mv_mag_sq < 53) {
if (Cs < 854) {
if (Rs < 2900) {
if (negBalance < 38) {
return 4;
}
else if (negBalance >= 38) {
if (refDCval < 54) {
return 4;
}
else if (refDCval >= 54) {
if (diffRsCsdiff < 66) {
if (mu_mv_mag_sq < 41) {
return 2;
}
else if (mu_mv_mag_sq >= 41) {
if (Rs < 2598) {
return 4;
}
else if (Rs >= 2598) {
return 2;
}
}
}
else if (diffRsCsdiff >= 66) {
return 4;
}
}
}
}
else if (Rs >= 2900) {
return 4;
}
}
else if (Cs >= 854) {
if (AFD < 16) {
if (RsCsDiff < 87) {
return 2;
}
else if (RsCsDiff >= 87) {
if (Cs < 956) {
return 4;
}
else if (Cs >= 956) {
if (CsDiff < 5) {
if (Rs < 2943) {
return 4;
}
else if (Rs >= 2943) {
return 2;
}
}
else if (CsDiff >= 5) {
if (ssDCval < 68) {
if (refDCval < 69) {
return 2;
}
else if (refDCval >= 69) {
return 4;
}
}
else if (ssDCval >= 68) {
return 4;
}
}
}
}
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (Rs < 3217) {
return 4;
}
else if (Rs >= 3217) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 67) {
if (Rs < 2874) {
if (RsDiff < 8) {
if (ssDCval < 74) {
if (Rs < 2497) {
return 4;
}
else if (Rs >= 2497) {
return 8;
}
}
else if (ssDCval >= 74) {
if (posBalance < 35) {
return 2;
}
else if (posBalance >= 35) {
if (diffMVdiffVal < 8) {
return 1;
}
else if (diffMVdiffVal >= 8) {
if (Rs < 2575) {
return 1;
}
else if (Rs >= 2575) {
return 4;
}
}
}
}
}
else if (RsDiff >= 8) {
if (MVDiff < 26) {
if (gchDC < 2) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (MVDiff < 23) {
if (mu_mv_mag_sq < 45) {
return 4;
}
else if (mu_mv_mag_sq >= 45) {
return 2;
}
}
else if (MVDiff >= 23) {
if (Rs < 2656) {
return 1;
}
else if (Rs >= 2656) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
if (tcor < 87) {
if (Rs < 2546) {
return 1;
}
else if (Rs >= 2546) {
return 2;
}
}
else if (tcor >= 87) {
return 2;
}
}
}
}
else if (MVDiff >= 26) {
if (ssDCval < 78) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 1;
}
}
else if (ssDCval >= 78) {
if (Cs < 769) {
return 1;
}
else if (Cs >= 769) {
return 8;
}
}
}
}
}
else if (Rs >= 2874) {
if (posBalance < 47) {
if (CsDiff < 5) {
if (posBalance < 42) {
if (Cs < 1036) {
if (Rs < 3024) {
return 8;
}
else if (Rs >= 3024) {
return 8;
}
}
else if (Cs >= 1036) {
return 1;
}
}
else if (posBalance >= 42) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Rs < 3000) {
if (Rs < 2954) {
return 4;
}
else if (Rs >= 2954) {
return 8;
}
}
else if (Rs >= 3000) {
return 2;
}
}
}
}
else if (CsDiff >= 5) {
return 4;
}
}
else if (posBalance >= 47) {
if (negBalance < 38) {
if (Rs < 3062) {
return 8;
}
else if (Rs >= 3062) {
return 1;
}
}
else if (negBalance >= 38) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 67) {
if (negBalance < 47) {
if (diffAFD < -3) {
if (Rs < 3059) {
return 1;
}
else if (Rs >= 3059) {
return 2;
}
}
else if (diffAFD >= -3) {
return 2;
}
}
else if (negBalance >= 47) {
if (Rs < 2741) {
if (Rs < 2408) {
return 8;
}
else if (Rs >= 2408) {
return 8;
}
}
else if (Rs >= 2741) {
return 2;
}
}
}
}
}
else if (negBalance >= 51) {
if (ssDCval < 58) {
if (Rs < 2481) {
return 8;
}
else if (Rs >= 2481) {
return 4;
}
}
else if (ssDCval >= 58) {
if (mcTcor < 89) {
if (Cs < 651) {
return 2;
}
else if (Cs >= 651) {
return 8;
}
}
else if (mcTcor >= 89) {
if (refDCval < 92) {
return 2;
}
else if (refDCval >= 92) {
if (Rs < 2566) {
return 8;
}
else if (Rs >= 2566) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 49) {
if (MVDiff < 29) {
if (refDCval < 60) {
if (AFD < 15) {
if (Cs < 768) {
return 2;
}
else if (Cs >= 768) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
if (diffRsCsdiff < 10) {
if (Rs < 2526) {
return 4;
}
else if (Rs >= 2526) {
return 2;
}
}
else if (diffRsCsdiff >= 10) {
return 4;
}
}
}
}
else if (AFD >= 15) {
return 2;
}
}
else if (refDCval >= 60) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (negBalance < 39) {
if (AFD < 9) {
if (MVDiff < 1) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 2;
}
}
else if (MVDiff >= 1) {
if (Rs < 3261) {
if (diffMVdiffVal < -5) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
else if (diffMVdiffVal >= -5) {
return 2;
}
}
else if (Rs >= 3261) {
return 4;
}
}
}
else if (AFD >= 9) {
if (posBalance < 59) {
if (ssDCval < 74) {
if (diffMVdiffVal < 2) {
if (tcor < 85) {
if (MVDiff < 16) {
if (refDCval < 62) {
return 2;
}
else if (refDCval >= 62) {
return 4;
}
}
else if (MVDiff >= 16) {
return 2;
}
}
else if (tcor >= 85) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
else if (ssDCval >= 74) {
if (TSC < 6) {
if (mcTcor < 98) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (mcTcor >= 98) {
if (Cs < 391) {
return 8;
}
else if (Cs >= 391) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
return 4;
}
}
}
}
else if (TSC >= 6) {
if (tcor < 89) {
if (Rs < 3005) {
return 1;
}
else if (Rs >= 3005) {
return 4;
}
}
else if (tcor >= 89) {
if (RsDiff < 9) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
}
}
}
else if (posBalance >= 59) {
if (negBalance < 27) {
if (Rs < 2802) {
return 1;
}
else if (Rs >= 2802) {
return 2;
}
}
else if (negBalance >= 27) {
if (CsDiff < 3) {
if (Rs < 2931) {
return 2;
}
else if (Rs >= 2931) {
return 4;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
}
}
}
else if (negBalance >= 39) {
if (Cs < 1101) {
if (MVDiff < 11) {
if (mu_mv_mag_sq < 94) {
return 2;
}
else if (mu_mv_mag_sq >= 94) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
if (Rs < 2608) {
if (Cs < 975) {
return 2;
}
else if (Cs >= 975) {
return 4;
}
}
else if (Rs >= 2608) {
return 2;
}
}
}
}
else if (MVDiff >= 11) {
if (CsDiff < 4) {
if (diffRsCsdiff < -45) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (MVDiff < 25) {
return 2;
}
else if (MVDiff >= 25) {
return 4;
}
}
}
else if (diffRsCsdiff >= -45) {
if (SC < 3530) {
if (mcTcor < 90) {
if (Rs < 2452) {
return 2;
}
else if (Rs >= 2452) {
return 4;
}
}
else if (mcTcor >= 90) {
return 2;
}
}
else if (SC >= 3530) {
return 4;
}
}
}
else if (CsDiff >= 4) {
if (posBalance < 52) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
return 2;
}
}
else if (posBalance >= 52) {
if (negBalance < 62) {
if (mcTcor < 91) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
if (Rs < 2577) {
return 4;
}
else if (Rs >= 2577) {
return 2;
}
}
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (negBalance >= 62) {
return 1;
}
}
}
}
}
else if (Cs >= 1101) {
if (ssDCval < 65) {
if (Rs < 3056) {
return 4;
}
else if (Rs >= 3056) {
return 2;
}
}
else if (ssDCval >= 65) {
if (Rs < 2467) {
if (diffMVdiffVal < 4) {
if (diffRsCsdiff < -5) {
return 2;
}
else if (diffRsCsdiff >= -5) {
if (Cs < 1248) {
return 4;
}
else if (Cs >= 1248) {
return 2;
}
}
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
else if (Rs >= 2467) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
if (Cs < 1209) {
return 1;
}
else if (Cs >= 1209) {
return 2;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 29) {
if (Rs < 2638) {
if (Cs < 987) {
if (mcTcor < 77) {
return 8;
}
else if (mcTcor >= 77) {
if (diffRsCsdiff < -70) {
if (SC < 2906) {
return 4;
}
else if (SC >= 2906) {
return 2;
}
}
else if (diffRsCsdiff >= -70) {
if (mu_mv_mag_sq < 33) {
if (diffAFD < -4) {
if (Rs < 2365) {
return 8;
}
else if (Rs >= 2365) {
return 2;
}
}
else if (diffAFD >= -4) {
if (RsDiff < 18) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
if (Rs < 2477) {
return 1;
}
else if (Rs >= 2477) {
return 4;
}
}
}
else if (RsDiff >= 18) {
if (Rs < 2468) {
return 1;
}
else if (Rs >= 2468) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 33) {
return 2;
}
}
}
}
else if (Cs >= 987) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
if (AFD < 27) {
if (mu_mv_mag_sq < 52) {
return 8;
}
else if (mu_mv_mag_sq >= 52) {
return 1;
}
}
else if (AFD >= 27) {
return 2;
}
}
}
}
else if (Rs >= 2638) {
if (CsDiff < 3) {
if (Cs < 606) {
return 4;
}
else if (Cs >= 606) {
return 1;
}
}
else if (CsDiff >= 3) {
if (mcTcor < 82) {
if (RsDiff < 20) {
return 1;
}
else if (RsDiff >= 20) {
if (Rs < 2854) {
return 8;
}
else if (Rs >= 2854) {
return 2;
}
}
}
else if (mcTcor >= 82) {
if (mcTcor < 96) {
if (AFD < 26) {
if (CsDiff < 9) {
return 2;
}
else if (CsDiff >= 9) {
return 1;
}
}
else if (AFD >= 26) {
return 1;
}
}
else if (mcTcor >= 96) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 90) {
if (RsDiff < 9) {
if (mcTcor < 97) {
if (SC < 3018) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (CsDiff < 3) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 8;
}
}
else if (CsDiff >= 3) {
if (Cs < 499) {
return 1;
}
else if (Cs >= 499) {
return 2;
}
}
}
}
else if (SC >= 3018) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (refDCval < 107) {
if (TSC < 6) {
if (Rs < 2420) {
return 4;
}
else if (Rs >= 2420) {
return 8;
}
}
else if (TSC >= 6) {
if (TSC < 8) {
if (SC < 4195) {
if (Cs < 496) {
if (Rs < 2734) {
return 8;
}
else if (Rs >= 2734) {
if (Rs < 2875) {
return 4;
}
else if (Rs >= 2875) {
return 4;
}
}
}
else if (Cs >= 496) {
if (diffRsCsdiff < -94) {
return 1;
}
else if (diffRsCsdiff >= -94) {
return 4;
}
}
}
else if (SC >= 4195) {
return 8;
}
}
else if (TSC >= 8) {
if (SC < 3915) {
return 8;
}
else if (SC >= 3915) {
return 1;
}
}
}
}
else if (refDCval >= 107) {
if (Rs < 2950) {
if (ssDCval < 134) {
return 8;
}
else if (ssDCval >= 134) {
if (ssDCval < 137) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (ssDCval >= 137) {
if (Cs < 1077) {
return 1;
}
else if (Cs >= 1077) {
return 8;
}
}
}
}
else if (Rs >= 2950) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 121) {
if (Rs < 3108) {
if (Cs < 974) {
if (posBalance < 56) {
if (posBalance < 37) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (posBalance >= 37) {
if (SC < 3518) {
if (mu_mv_mag_sq < 103) {
return 8;
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 2734) {
return 2;
}
else if (Rs >= 2734) {
return 8;
}
}
}
else if (SC >= 3518) {
if (negBalance < 26) {
return 8;
}
else if (negBalance >= 26) {
if (SC < 3581) {
return 1;
}
else if (SC >= 3581) {
return 2;
}
}
}
}
}
else if (posBalance >= 56) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 1;
}
}
}
}
else if (Cs >= 974) {
if (Rs < 2325) {
return 4;
}
else if (Rs >= 2325) {
return 8;
}
}
}
else if (Rs >= 3108) {
if (MVDiff < 8) {
if (negBalance < 24) {
if (Cs < 979) {
return 8;
}
else if (Cs >= 979) {
return 2;
}
}
else if (negBalance >= 24) {
return 2;
}
}
else if (MVDiff >= 8) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
if (Rs < 3215) {
return 8;
}
else if (Rs >= 3215) {
return 2;
}
}
}
}
}
else if (refDCval >= 121) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (TSC < 6) {
if (CsDiff < 1) {
if (Cs < 13) {
return 8;
}
else if (Cs >= 13) {
return 1;
}
}
else if (CsDiff >= 1) {
if (Cs < 1063) {
if (SC < 3319) {
if (Rs < 2750) {
return 1;
}
else if (Rs >= 2750) {
return 8;
}
}
else if (SC >= 3319) {
return 2;
}
}
else if (Cs >= 1063) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (diffRsCsdiff < -34) {
return 1;
}
else if (diffRsCsdiff >= -34) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (mcTcor < 68) {
if (MVDiff < 68) {
return 1;
}
else if (MVDiff >= 68) {
if (Rs < 2693) {
return 4;
}
else if (Rs >= 2693) {
return 8;
}
}
}
else if (mcTcor >= 68) {
if (refDCval < 114) {
if (RsCsDiff < 207) {
if (MVDiff < 32) {
if (Cs < 1083) {
if (Cs < 880) {
if (MVDiff < 16) {
if (mu_mv_mag_sq < 44) {
return 2;
}
else if (mu_mv_mag_sq >= 44) {
if (Rs < 2978) {
return 8;
}
else if (Rs >= 2978) {
return 2;
}
}
}
else if (MVDiff >= 16) {
if (diffMVdiffVal < -9) {
if (Cs < 732) {
return 1;
}
else if (Cs >= 732) {
return 2;
}
}
else if (diffMVdiffVal >= -9) {
if (RsDiff < 11) {
if (RsCsDiff < 105) {
return 2;
}
else if (RsCsDiff >= 105) {
return 1;
}
}
else if (RsDiff >= 11) {
if (Rs < 2519) {
return 4;
}
else if (Rs >= 2519) {
if (diffRsCsdiff < 60) {
return 2;
}
else if (diffRsCsdiff >= 60) {
return 4;
}
}
}
}
}
}
else if (Cs >= 880) {
if (MVDiff < 12) {
if (Rs < 3066) {
return 4;
}
else if (Rs >= 3066) {
return 2;
}
}
else if (MVDiff >= 12) {
if (mcTcor < 89) {
if (Rs < 2657) {
return 1;
}
else if (Rs >= 2657) {
return 8;
}
}
else if (mcTcor >= 89) {
return 2;
}
}
}
}
else if (Cs >= 1083) {
if (diffRsCsdiff < -20) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
if (Rs < 2918) {
return 8;
}
else if (Rs >= 2918) {
return 2;
}
}
}
else if (diffRsCsdiff >= -20) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (mu_mv_mag_sq < 2) {
if (Rs < 2576) {
return 1;
}
else if (Rs >= 2576) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 32) {
if (Rs < 2663) {
if (TSC < 13) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
if (Rs < 2544) {
if (SC < 2831) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
else if (SC >= 2831) {
return 8;
}
}
else if (Rs >= 2544) {
if (diffRsCsdiff < 20) {
if (Rs < 2597) {
return 2;
}
else if (Rs >= 2597) {
return 8;
}
}
else if (diffRsCsdiff >= 20) {
if (mu_mv_mag_sq < 19) {
return 1;
}
else if (mu_mv_mag_sq >= 19) {
if (Rs < 2622) {
return 8;
}
else if (Rs >= 2622) {
return 1;
}
}
}
}
}
}
else if (TSC >= 13) {
if (refDCval < 97) {
return 1;
}
else if (refDCval >= 97) {
return 2;
}
}
}
else if (Rs >= 2663) {
if (Cs < 386) {
if (Rs < 3139) {
if (Cs < 231) {
return 1;
}
else if (Cs >= 231) {
return 2;
}
}
else if (Rs >= 3139) {
return 4;
}
}
else if (Cs >= 386) {
if (posBalance < 70) {
if (RsCsDiff < 121) {
if (Cs < 656) {
return 8;
}
else if (Cs >= 656) {
if (Rs < 2882) {
return 4;
}
else if (Rs >= 2882) {
return 1;
}
}
}
else if (RsCsDiff >= 121) {
return 8;
}
}
else if (posBalance >= 70) {
if (Rs < 3208) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
else if (Rs >= 3208) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 207) {
if (TSC < 10) {
if (CsDiff < 7) {
if (diffAFD < 4) {
if (refDCval < 112) {
return 2;
}
else if (refDCval >= 112) {
if (Cs < 728) {
return 2;
}
else if (Cs >= 728) {
if (Rs < 3013) {
return 1;
}
else if (Rs >= 3013) {
return 8;
}
}
}
}
else if (diffAFD >= 4) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 1;
}
}
}
else if (CsDiff >= 7) {
if (tcor < 78) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
if (ssDCval < 106) {
return 2;
}
else if (ssDCval >= 106) {
return 8;
}
}
}
else if (tcor >= 78) {
return 2;
}
}
}
else if (TSC >= 10) {
if (mcTcor < 93) {
if (AFD < 25) {
if (Cs < 680) {
if (TSC < 12) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 2;
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (Cs >= 680) {
if (negBalance < 50) {
if (Cs < 978) {
return 8;
}
else if (Cs >= 978) {
if (MVDiff < 30) {
return 2;
}
else if (MVDiff >= 30) {
return 1;
}
}
}
else if (negBalance >= 50) {
if (SC < 4354) {
if (AFD < 17) {
if (Rs < 2764) {
return 8;
}
else if (Rs >= 2764) {
return 2;
}
}
else if (AFD >= 17) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
if (negBalance < 55) {
if (tcor < 73) {
return 8;
}
else if (tcor >= 73) {
return 4;
}
}
else if (negBalance >= 55) {
if (negBalance < 60) {
return 4;
}
else if (negBalance >= 60) {
if (Rs < 2514) {
return 4;
}
else if (Rs >= 2514) {
return 2;
}
}
}
}
}
}
else if (SC >= 4354) {
return 8;
}
}
}
}
else if (AFD >= 25) {
if (diffTSC < 4) {
if (mu_mv_mag_sq < 378) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (mu_mv_mag_sq >= 378) {
return 4;
}
}
else if (diffTSC >= 4) {
return 1;
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 108) {
if (Rs < 2457) {
return 8;
}
else if (Rs >= 2457) {
if (MVDiff < 131) {
return 2;
}
else if (MVDiff >= 131) {
return 1;
}
}
}
else if (ssDCval >= 108) {
if (Cs < 687) {
return 8;
}
else if (Cs >= 687) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 114) {
if (MVDiff < 7) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 2;
}
}
else if (MVDiff >= 7) {
if (mcTcor < 90) {
if (ssDCval < 131) {
if (MVDiff < 27) {
return 8;
}
else if (MVDiff >= 27) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
if (Cs < 963) {
return 2;
}
else if (Cs >= 963) {
return 1;
}
}
}
}
else if (ssDCval >= 131) {
if (diffRsCsdiff < 42) {
if (tcor < 65) {
if (Rs < 2824) {
return 4;
}
else if (Rs >= 2824) {
return 1;
}
}
else if (tcor >= 65) {
return 4;
}
}
else if (diffRsCsdiff >= 42) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
if (posBalance < 55) {
if (Rs < 2759) {
return 8;
}
else if (Rs >= 2759) {
return 2;
}
}
else if (posBalance >= 55) {
return 1;
}
}
}
}
}
else if (mcTcor >= 90) {
if (SC < 3393) {
if (CsDiff < 2) {
if (Cs < 292) {
return 2;
}
else if (Cs >= 292) {
return 8;
}
}
else if (CsDiff >= 2) {
if (mcTcor < 93) {
if (Rs < 2620) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 1;
}
}
else if (Rs >= 2620) {
if (Cs < 502) {
return 8;
}
else if (Cs >= 502) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 170) {
if (MVDiff < 23) {
if (AFD < 26) {
if (Rs < 2539) {
return 1;
}
else if (Rs >= 2539) {
if (gchDC < 4) {
if (MVDiff < 21) {
return 1;
}
else if (MVDiff >= 21) {
return 2;
}
}
else if (gchDC >= 4) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (RsDiff < 16) {
return 1;
}
else if (RsDiff >= 16) {
return 2;
}
}
}
}
}
else if (AFD >= 26) {
if (Rs < 2438) {
return 8;
}
else if (Rs >= 2438) {
return 2;
}
}
}
else if (MVDiff >= 23) {
if (ssDCval < 114) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 1;
}
}
else if (ssDCval >= 114) {
return 1;
}
}
}
else if (ssDCval >= 170) {
if (Rs < 2416) {
return 1;
}
else if (Rs >= 2416) {
return 2;
}
}
}
}
}
else if (SC >= 3393) {
if (mu_mv_mag_sq < 46) {
if (AFD < 24) {
if (MVDiff < 41) {
if (Cs < 974) {
if (Cs < 558) {
if (diffAFD < 3) {
return 1;
}
else if (diffAFD >= 3) {
return 2;
}
}
else if (Cs >= 558) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 1;
}
}
}
else if (Cs >= 974) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Cs < 1263) {
return 1;
}
else if (Cs >= 1263) {
return 8;
}
}
}
}
else if (MVDiff >= 41) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
return 8;
}
}
}
}
else if (AFD >= 24) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 13) {
if (Rs < 3201) {
if (MVDiff < 67) {
return 1;
}
else if (MVDiff >= 67) {
if (Rs < 3033) {
return 1;
}
else if (Rs >= 3033) {
return 8;
}
}
}
else if (Rs >= 3201) {
if (Rs < 3239) {
return 2;
}
else if (Rs >= 3239) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (diffRsCsdiff < 153) {
if (negBalance < 68) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (negBalance >= 68) {
return 1;
}
}
else if (diffRsCsdiff >= 153) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (refDCval < 120) {
if (AFD < 33) {
if (Rs < 2932) {
return 8;
}
else if (Rs >= 2932) {
return 2;
}
}
else if (AFD >= 33) {
return 1;
}
}
else if (refDCval >= 120) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 3279) {
if (ssDCval < 104) {
if (posBalance < 48) {
if (negBalance < 31) {
if (Cs < 1026) {
if (ssDCval < 92) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Cs < 941) {
if (Rs < 5821) {
return 4;
}
else if (Rs >= 5821) {
return 8;
}
}
else if (Cs >= 941) {
if (Rs < 3775) {
return 2;
}
else if (Rs >= 3775) {
return 4;
}
}
}
}
else if (ssDCval >= 92) {
if (Cs < 787) {
return 2;
}
else if (Cs >= 787) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (ssDCval < 97) {
if (Rs < 3565) {
return 8;
}
else if (Rs >= 3565) {
return 8;
}
}
else if (ssDCval >= 97) {
if (mu_mv_mag_sq < 58) {
if (Rs < 3358) {
return 4;
}
else if (Rs >= 3358) {
return 4;
}
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 3475) {
if (Rs < 3425) {
return 8;
}
else if (Rs >= 3425) {
if (Rs < 3439) {
if (Rs < 3428) {
return 8;
}
else if (Rs >= 3428) {
return 8;
}
}
else if (Rs >= 3439) {
return 4;
}
}
}
else if (Rs >= 3475) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 1026) {
if (RsCsDiff < 7) {
if (SC < 4917) {
return 8;
}
else if (SC >= 4917) {
if (Cs < 1274) {
return 4;
}
else if (Cs >= 1274) {
if (Rs < 3892) {
return 4;
}
else if (Rs >= 3892) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 4566) {
return 8;
}
else if (Rs >= 4566) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 7) {
if (ssDCval < 69) {
if (ssDCval < 64) {
if (Cs < 1074) {
return 2;
}
else if (Cs >= 1074) {
if (ssDCval < 59) {
if (Rs < 3439) {
return 8;
}
else if (Rs >= 3439) {
return 4;
}
}
else if (ssDCval >= 59) {
return 8;
}
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 68) {
if (Rs < 3708) {
return 8;
}
else if (Rs >= 3708) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (TSC < 7) {
if (Rs < 4286) {
if (Rs < 3817) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 4;
}
}
else if (Rs >= 3817) {
if (Rs < 4064) {
return 4;
}
else if (Rs >= 4064) {
return 4;
}
}
}
else if (Rs >= 4286) {
return 8;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (Rs < 3485) {
return 2;
}
else if (Rs >= 3485) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 31) {
if (refDCval < 79) {
if (Rs < 3861) {
if (negBalance < 44) {
if (Cs < 1112) {
if (ssDCval < 61) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (ssDCval >= 61) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (ssDCval < 71) {
return 4;
}
else if (ssDCval >= 71) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 3415) {
if (Rs < 3299) {
return 1;
}
else if (Rs >= 3299) {
return 4;
}
}
else if (Rs >= 3415) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1112) {
if (ssDCval < 70) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 3827) {
return 2;
}
else if (Rs >= 3827) {
if (Rs < 3847) {
return 4;
}
else if (Rs >= 3847) {
return 2;
}
}
}
}
else if (ssDCval >= 70) {
if (diffMVdiffVal < -2) {
if (AFD < 17) {
if (Rs < 3447) {
return 2;
}
else if (Rs >= 3447) {
return 1;
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 1121) {
return 2;
}
else if (Cs >= 1121) {
return 4;
}
}
}
}
}
else if (negBalance >= 44) {
return 2;
}
}
else if (Rs >= 3861) {
return 4;
}
}
else if (refDCval >= 79) {
if (SC < 4727) {
if (refDCval < 106) {
if (Cs < 1214) {
return 2;
}
else if (Cs >= 1214) {
if (Rs < 3308) {
return 4;
}
else if (Rs >= 3308) {
if (Rs < 3372) {
return 2;
}
else if (Rs >= 3372) {
return 1;
}
}
}
}
else if (refDCval >= 106) {
if (Rs < 3488) {
return 1;
}
else if (Rs >= 3488) {
return 8;
}
}
}
else if (SC >= 4727) {
if (diffRsCsdiff < -7) {
if (diffMVdiffVal < -7) {
if (Rs < 4158) {
return 1;
}
else if (Rs >= 4158) {
return 4;
}
}
else if (diffMVdiffVal >= -7) {
return 2;
}
}
else if (diffRsCsdiff >= -7) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (Cs < 1228) {
if (Cs < 949) {
return 2;
}
else if (Cs >= 949) {
if (CsDiff < 6) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (AFD < 11) {
if (posBalance < 41) {
if (Rs < 3870) {
return 4;
}
else if (Rs >= 3870) {
return 8;
}
}
else if (posBalance >= 41) {
if (Rs < 4139) {
return 4;
}
else if (Rs >= 4139) {
return 4;
}
}
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (Cs >= 1228) {
return 2;
}
}
}
}
}
}
}
else if (posBalance >= 48) {
if (refDCval < 78) {
if (Cs < 949) {
if (posBalance < 68) {
if (Rs < 3603) {
if (mcTcor < 96) {
if (MVDiff < 16) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
else if (MVDiff >= 16) {
return 4;
}
}
else if (mcTcor >= 96) {
return 4;
}
}
else if (Rs >= 3603) {
return 4;
}
}
else if (posBalance >= 68) {
return 2;
}
}
else if (Cs >= 949) {
if (MVDiff < 29) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (posBalance < 49) {
if (Cs < 1179) {
return 4;
}
else if (Cs >= 1179) {
return 2;
}
}
else if (posBalance >= 49) {
return 2;
}
}
}
else if (MVDiff >= 29) {
return 4;
}
}
}
else if (refDCval >= 78) {
if (mcTcor < 67) {
return 1;
}
else if (mcTcor >= 67) {
if (ssDCval < 90) {
if (Cs < 490) {
if (RsCsDiff < 263) {
if (Rs < 3591) {
if (Rs < 3299) {
return 2;
}
else if (Rs >= 3299) {
return 1;
}
}
else if (Rs >= 3591) {
return 4;
}
}
else if (RsCsDiff >= 263) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 2;
}
}
}
else if (Cs >= 490) {
if (tcor < 74) {
if (Cs < 842) {
return 2;
}
else if (Cs >= 842) {
if (Cs < 1206) {
return 1;
}
else if (Cs >= 1206) {
return 4;
}
}
}
else if (tcor >= 74) {
if (Rs < 4631) {
if (diffAFD < -4) {
if (negBalance < 37) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 4;
}
}
else if (negBalance >= 37) {
return 2;
}
}
else if (diffAFD >= -4) {
return 2;
}
}
else if (Rs >= 4631) {
return 1;
}
}
}
}
else if (ssDCval >= 90) {
if (mcTcor < 81) {
if (RsDiff < 22) {
return 4;
}
else if (RsDiff >= 22) {
return 2;
}
}
else if (mcTcor >= 81) {
if (RsCsDiff < 190) {
if (MVDiff < 21) {
if (RsCsDiff < 94) {
if (mcTcor < 96) {
if (negBalance < 24) {
if (Rs < 3455) {
return 8;
}
else if (Rs >= 3455) {
return 4;
}
}
else if (negBalance >= 24) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Rs < 3297) {
return 8;
}
else if (Rs >= 3297) {
return 2;
}
}
}
else if (RsCsDiff >= 94) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
if (Rs < 3438) {
return 4;
}
else if (Rs >= 3438) {
return 8;
}
}
}
}
else if (MVDiff >= 21) {
if (Rs < 5393) {
if (Rs < 3423) {
if (Rs < 3328) {
return 8;
}
else if (Rs >= 3328) {
return 1;
}
}
else if (Rs >= 3423) {
return 8;
}
}
else if (Rs >= 5393) {
return 2;
}
}
}
else if (RsCsDiff >= 190) {
if (mcTcor < 93) {
if (TSC < 11) {
if (posBalance < 70) {
if (Cs < 648) {
return 2;
}
else if (Cs >= 648) {
return 8;
}
}
else if (posBalance >= 70) {
return 2;
}
}
else if (TSC >= 11) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (gchDC < 4) {
if (diffRsCsdiff < -409) {
return 1;
}
else if (diffRsCsdiff >= -409) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (gchDC >= 4) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (Rs < 3636) {
return 8;
}
else if (Rs >= 3636) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 104) {
if (MVDiff < 7) {
if (MVDiff < 1) {
if (ssDCval < 146) {
return 2;
}
else if (ssDCval >= 146) {
if (ssDCval < 147) {
if (Rs < 3836) {
return 1;
}
else if (Rs >= 3836) {
return 2;
}
}
else if (ssDCval >= 147) {
return 2;
}
}
}
else if (MVDiff >= 1) {
if (Rs < 3870) {
if (ssDCval < 142) {
if (mcTcor < 96) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
else if (mcTcor >= 96) {
if (diffMVdiffVal < -2) {
if (TSC < 4) {
if (CsDiff < 4) {
if (Rs < 3337) {
return 2;
}
else if (Rs >= 3337) {
return 8;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (TSC >= 4) {
if (Rs < 3466) {
if (Rs < 3441) {
return 8;
}
else if (Rs >= 3441) {
return 4;
}
}
else if (Rs >= 3466) {
return 2;
}
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 1277) {
return 2;
}
else if (Cs >= 1277) {
return 8;
}
}
}
}
else if (ssDCval >= 142) {
if (Cs < 1129) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 1;
}
}
else if (Cs >= 1129) {
if (Rs < 3566) {
return 2;
}
else if (Rs >= 3566) {
return 1;
}
}
}
}
else if (Rs >= 3870) {
if (refDCval < 136) {
if (refDCval < 107) {
return 4;
}
else if (refDCval >= 107) {
return 2;
}
}
else if (refDCval >= 136) {
if (mcTcor < 98) {
if (gchDC < 2) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (Rs < 4418) {
return 1;
}
else if (Rs >= 4418) {
if (Cs < 786) {
return 1;
}
else if (Cs >= 786) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (mcTcor >= 98) {
if (diffTSC < 3) {
return 2;
}
else if (diffTSC >= 3) {
if (Rs < 3950) {
return 1;
}
else if (Rs >= 3950) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 135) {
if (Cs < 500) {
if (Rs < 3677) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (RsCsDiff < 334) {
return 4;
}
else if (RsCsDiff >= 334) {
return 2;
}
}
}
else if (Rs >= 3677) {
return 2;
}
}
else if (Cs >= 500) {
if (RsCsDiff < 492) {
if (Rs < 3896) {
if (Cs < 1157) {
if (MVDiff < 36) {
if (TSC < 6) {
if (RsDiff < 5) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (RsDiff >= 5) {
if (CsDiff < 3) {
if (Rs < 3437) {
return 8;
}
else if (Rs >= 3437) {
return 2;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (TSC >= 6) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
if (CsDiff < 4) {
if (mcTcor < 97) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 2;
}
}
}
else if (mcTcor >= 97) {
return 2;
}
}
else if (CsDiff >= 4) {
if (refDCval < 114) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (refDCval >= 114) {
if (tcor < 76) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
else if (tcor >= 76) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 36) {
if (CsDiff < 4) {
if (diffRsCsdiff < -83) {
return 4;
}
else if (diffRsCsdiff >= -83) {
return 8;
}
}
else if (CsDiff >= 4) {
if (SC < 4523) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 1;
}
}
else if (SC >= 4523) {
if (MVDiff < 50) {
return 8;
}
else if (MVDiff >= 50) {
if (Rs < 3706) {
return 2;
}
else if (Rs >= 3706) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1157) {
if (Rs < 3708) {
if (MVDiff < 9) {
if (Cs < 1233) {
return 8;
}
else if (Cs >= 1233) {
return 2;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (Rs >= 3708) {
if (Cs < 1239) {
return 4;
}
else if (Cs >= 1239) {
return 8;
}
}
}
}
else if (Rs >= 3896) {
if (mu_mv_mag_sq < 39) {
if (diffRsCsdiff < -167) {
return 2;
}
else if (diffRsCsdiff >= -167) {
if (MVDiff < 16) {
if (ssDCval < 131) {
if (Rs < 4074) {
if (Rs < 4010) {
return 8;
}
else if (Rs >= 4010) {
return 4;
}
}
else if (Rs >= 4074) {
if (Cs < 1303) {
if (RsCsDiff < 451) {
return 2;
}
else if (RsCsDiff >= 451) {
return 1;
}
}
else if (Cs >= 1303) {
return 4;
}
}
}
else if (ssDCval >= 131) {
if (Rs < 4650) {
return 2;
}
else if (Rs >= 4650) {
return 1;
}
}
}
else if (MVDiff >= 16) {
if (CsDiff < 7) {
if (negBalance < 50) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (MVDiff < 45) {
if (Cs < 675) {
return 2;
}
else if (Cs >= 675) {
return 1;
}
}
else if (MVDiff >= 45) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (refDCval < 111) {
if (Rs < 4776) {
return 2;
}
else if (Rs >= 4776) {
return 8;
}
}
else if (refDCval >= 111) {
if (RsDiff < 14) {
if (tcor < 76) {
return 1;
}
else if (tcor >= 76) {
return 8;
}
}
else if (RsDiff >= 14) {
if (Cs < 1045) {
if (RsDiff < 19) {
return 8;
}
else if (RsDiff >= 19) {
if (Cs < 946) {
return 2;
}
else if (Cs >= 946) {
if (Cs < 1041) {
return 8;
}
else if (Cs >= 1041) {
return 2;
}
}
}
}
else if (Cs >= 1045) {
return 8;
}
}
}
}
}
else if (CsDiff >= 7) {
if (Rs < 5182) {
return 1;
}
else if (Rs >= 5182) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
if (Rs < 5549) {
return 1;
}
else if (Rs >= 5549) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (ssDCval < 132) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (ssDCval >= 132) {
if (Cs < 917) {
return 2;
}
else if (Cs >= 917) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 492) {
if (refDCval < 120) {
if (AFD < 36) {
if (Rs < 3390) {
if (Rs < 3359) {
return 2;
}
else if (Rs >= 3359) {
return 8;
}
}
else if (Rs >= 3390) {
return 2;
}
}
else if (AFD >= 36) {
if (Rs < 3679) {
return 2;
}
else if (Rs >= 3679) {
return 4;
}
}
}
else if (refDCval >= 120) {
if (Cs < 838) {
if (diffAFD < 12) {
return 2;
}
else if (diffAFD >= 12) {
if (Rs < 4335) {
return 8;
}
else if (Rs >= 4335) {
return 2;
}
}
}
else if (Cs >= 838) {
if (MVDiff < 23) {
if (refDCval < 133) {
return 2;
}
else if (refDCval >= 133) {
if (Rs < 5619) {
return 2;
}
else if (Rs >= 5619) {
return 8;
}
}
}
else if (MVDiff >= 23) {
if (tcor < 72) {
if (diffRsCsdiff < 165) {
if (Rs < 6061) {
if (Rs < 4328) {
return 2;
}
else if (Rs >= 4328) {
if (Rs < 4854) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
else if (Rs >= 4854) {
return 2;
}
}
}
else if (Rs >= 6061) {
return 8;
}
}
else if (diffRsCsdiff >= 165) {
return 8;
}
}
else if (tcor >= 72) {
if (AFD < 23) {
if (Rs < 5202) {
return 1;
}
else if (Rs >= 5202) {
return 2;
}
}
else if (AFD >= 23) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (ssDCval < 128) {
return 1;
}
else if (ssDCval >= 128) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 135) {
if (mcTcor < 93) {
if (refDCval < 147) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (Cs < 1218) {
if (refDCval < 136) {
return 8;
}
else if (refDCval >= 136) {
return 2;
}
}
else if (Cs >= 1218) {
return 4;
}
}
}
else if (refDCval >= 147) {
if (Rs < 5648) {
if (MVDiff < 81) {
return 1;
}
else if (MVDiff >= 81) {
return 8;
}
}
else if (Rs >= 5648) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (CsDiff < 4) {
if (mcTcor < 95) {
if (Cs < 690) {
if (Rs < 3708) {
return 8;
}
else if (Rs >= 3708) {
return 2;
}
}
else if (Cs >= 690) {
return 4;
}
}
else if (mcTcor >= 95) {
if (SC < 6479) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (diffAFD < -5) {
if (negBalance < 42) {
if (Cs < 1146) {
return 2;
}
else if (Cs >= 1146) {
return 1;
}
}
else if (negBalance >= 42) {
return 1;
}
}
else if (diffAFD >= -5) {
return 2;
}
}
}
else if (SC >= 6479) {
if (Rs < 6473) {
if (Rs < 5810) {
return 2;
}
else if (Rs >= 5810) {
return 1;
}
}
else if (Rs >= 6473) {
return 2;
}
}
}
}
else if (CsDiff >= 4) {
if (RsDiff < 20) {
if (negBalance < 58) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 8;
}
}
else if (negBalance >= 58) {
if (TSC < 12) {
if (Cs < 1180) {
return 2;
}
else if (Cs >= 1180) {
return 1;
}
}
else if (TSC >= 12) {
return 1;
}
}
}
else if (RsDiff >= 20) {
if (Rs < 4961) {
if (tcor < 86) {
if (tcor < 73) {
return 2;
}
else if (tcor >= 73) {
return 1;
}
}
else if (tcor >= 86) {
return 2;
}
}
else if (Rs >= 4961) {
if (RsCsDiff < 431) {
if (Rs < 6171) {
return 8;
}
else if (Rs >= 6171) {
return 2;
}
}
else if (RsCsDiff >= 431) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1313) {
if (mu_mv_mag_sq < 111) {
if (RsDiff < 9) {
if (CsDiff < 8) {
if (ssDCval < 105) {
if (negBalance < 39) {
if (Rs < 825) {
if (Cs < 1494) {
if (refDCval < 63) {
return 1;
}
else if (refDCval >= 63) {
if (MVDiff < 3) {
if (ssDCval < 88) {
return 8;
}
else if (ssDCval >= 88) {
return 4;
}
}
else if (MVDiff >= 3) {
if (Cs < 1462) {
return 4;
}
else if (Cs >= 1462) {
return 1;
}
}
}
}
else if (Cs >= 1494) {
if (TSC < 5) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (TSC >= 5) {
if (Cs < 1764) {
return 8;
}
else if (Cs >= 1764) {
return 2;
}
}
}
}
else if (Rs >= 825) {
if (Rs < 6007) {
if (negBalance < 31) {
if (posBalance < 45) {
if (ssDCval < 64) {
if (ssDCval < 59) {
if (ssDCval < 50) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
else if (ssDCval >= 50) {
if (Rs < 4731) {
if (Rs < 1034) {
return 8;
}
else if (Rs >= 1034) {
if (Cs < 1935) {
if (Cs < 1544) {
return 1;
}
else if (Cs >= 1544) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (Cs >= 1935) {
if (Cs < 2364) {
if (Rs < 4580) {
if (Rs < 4488) {
return 4;
}
else if (Rs >= 4488) {
return 4;
}
}
else if (Rs >= 4580) {
return 4;
}
}
else if (Cs >= 2364) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
return 8;
}
}
}
}
}
else if (Rs >= 4731) {
return 4;
}
}
}
else if (ssDCval >= 59) {
if (Rs < 2353) {
return 2;
}
else if (Rs >= 2353) {
if (Rs < 3743) {
return 8;
}
else if (Rs >= 3743) {
return 4;
}
}
}
}
else if (ssDCval >= 64) {
if (refDCval < 71) {
if (Rs < 1052) {
return 1;
}
else if (Rs >= 1052) {
if (MVDiff < 3) {
if (Rs < 2923) {
return 8;
}
else if (Rs >= 2923) {
return 4;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (refDCval >= 71) {
if (Rs < 3291) {
if (CsDiff < 1) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
if (TSCindex < 1) {
if (Rs < 2379) {
return 1;
}
else if (Rs >= 2379) {
return 8;
}
}
else if (TSCindex >= 1) {
if (Rs < 2405) {
return 4;
}
else if (Rs >= 2405) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
}
}
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 86) {
return 1;
}
else if (mu_mv_mag_sq >= 86) {
return 8;
}
}
}
else if (Rs >= 3291) {
if (MVDiff < 4) {
if (SC < 5655) {
return 4;
}
else if (SC >= 5655) {
if (SC < 6991) {
if (ssDCval < 96) {
if (TSC < 4) {
if (Rs < 4802) {
return 2;
}
else if (Rs >= 4802) {
return 8;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (ssDCval >= 96) {
if (Rs < 3684) {
return 8;
}
else if (Rs >= 3684) {
return 4;
}
}
}
else if (SC >= 6991) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
if (Cs < 1848) {
return 4;
}
else if (Cs >= 1848) {
if (Rs < 5595) {
return 8;
}
else if (Rs >= 5595) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2844) {
if (refDCval < 57) {
if (refDCval < 55) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (negBalance < 28) {
return 8;
}
else if (negBalance >= 28) {
if (Rs < 1751) {
return 1;
}
else if (Rs >= 1751) {
return 8;
}
}
}
}
else if (refDCval >= 55) {
if (Rs < 1033) {
return 8;
}
else if (Rs >= 1033) {
if (SC < 3187) {
return 4;
}
else if (SC >= 3187) {
return 8;
}
}
}
}
else if (refDCval >= 57) {
if (mcTcor < 96) {
if (Cs < 2057) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 1483) {
return 4;
}
else if (Rs >= 1483) {
return 8;
}
}
}
else if (Cs >= 2057) {
if (negBalance < 29) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Rs < 1032) {
return 2;
}
else if (Rs >= 1032) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (ssDCval < 85) {
return 1;
}
else if (ssDCval >= 85) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 2844) {
if (Cs < 1830) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (CsDiff < 4) {
if (mu_mv_mag_sq < 105) {
return 4;
}
else if (mu_mv_mag_sq >= 105) {
if (mcTcor < 98) {
if (RsCsDiff < 43) {
return 2;
}
else if (RsCsDiff >= 43) {
return 4;
}
}
else if (mcTcor >= 98) {
if (Rs < 3486) {
return 4;
}
else if (Rs >= 3486) {
return 4;
}
}
}
}
else if (CsDiff >= 4) {
if (MVDiff < 4) {
if (Rs < 2980) {
return 4;
}
else if (Rs >= 2980) {
return 2;
}
}
else if (MVDiff >= 4) {
if (Rs < 3995) {
return 2;
}
else if (Rs >= 3995) {
if (Rs < 4641) {
return 4;
}
else if (Rs >= 4641) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1830) {
if (Rs < 3826) {
if (Rs < 2886) {
return 4;
}
else if (Rs >= 2886) {
if (Rs < 3441) {
return 8;
}
else if (Rs >= 3441) {
if (Rs < 3543) {
return 8;
}
else if (Rs >= 3543) {
return 8;
}
}
}
}
else if (Rs >= 3826) {
if (tcor < 96) {
if (Cs < 2135) {
return 4;
}
else if (Cs >= 2135) {
return 8;
}
}
else if (tcor >= 96) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 31) {
if (Cs < 1623) {
if (RsDiff < 7) {
if (Rs < 3670) {
if (Rs < 1462) {
if (mcTcor < 92) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (MVDiff < 27) {
return 8;
}
else if (MVDiff >= 27) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 82) {
if (ssDCval < 56) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (Cs < 1550) {
return 4;
}
else if (Cs >= 1550) {
return 2;
}
}
}
else if (ssDCval >= 56) {
if (posBalance < 55) {
if (TSC < 4) {
if (Rs < 1138) {
return 1;
}
else if (Rs >= 1138) {
return 4;
}
}
else if (TSC >= 4) {
return 4;
}
}
else if (posBalance >= 55) {
if (Rs < 1020) {
return 8;
}
else if (Rs >= 1020) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Cs < 1341) {
return 4;
}
else if (Cs >= 1341) {
if (SC < 2729) {
if (ssDCval < 56) {
if (Rs < 1319) {
return 4;
}
else if (Rs >= 1319) {
return 8;
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (SC >= 2729) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1462) {
if (mu_mv_mag_sq < 12) {
if (CsDiff < 6) {
if (posBalance < 52) {
if (MVDiff < 3) {
if (Rs < 2575) {
return 1;
}
else if (Rs >= 2575) {
return 2;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (posBalance >= 52) {
return 2;
}
}
else if (CsDiff >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 12) {
if (ssDCval < 99) {
if (Rs < 3192) {
if (mcTcor < 98) {
if (diffMVdiffVal < 2) {
if (Cs < 1326) {
if (diffRsCsdiff < -8) {
return 2;
}
else if (diffRsCsdiff >= -8) {
if (Rs < 2279) {
return 8;
}
else if (Rs >= 2279) {
return 4;
}
}
}
else if (Cs >= 1326) {
if (posBalance < 37) {
return 2;
}
else if (posBalance >= 37) {
if (diffAFD < 1) {
if (AFD < 6) {
if (RsCsDiff < 33) {
if (Rs < 1568) {
return 4;
}
else if (Rs >= 1568) {
return 8;
}
}
else if (RsCsDiff >= 33) {
if (Rs < 2149) {
return 8;
}
else if (Rs >= 2149) {
return 4;
}
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (diffAFD >= 1) {
if (mcTcor < 96) {
if (posBalance < 41) {
return 2;
}
else if (posBalance >= 41) {
return 8;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (CsDiff < 6) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 2;
}
}
else if (CsDiff >= 6) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (refDCval < 87) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (refDCval >= 87) {
return 8;
}
}
}
else if (Rs >= 3192) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 2;
}
}
}
else if (ssDCval >= 99) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
return 8;
}
}
}
}
}
else if (Rs >= 3670) {
if (ssDCval < 91) {
if (Rs < 4370) {
return 4;
}
else if (Rs >= 4370) {
return 8;
}
}
else if (ssDCval >= 91) {
if (Rs < 4099) {
return 2;
}
else if (Rs >= 4099) {
return 8;
}
}
}
}
else if (RsDiff >= 7) {
if (MVDiff < 5) {
if (Rs < 4081) {
return 2;
}
else if (Rs >= 4081) {
if (negBalance < 33) {
return 2;
}
else if (negBalance >= 33) {
if (Rs < 4204) {
return 4;
}
else if (Rs >= 4204) {
return 8;
}
}
}
}
else if (MVDiff >= 5) {
if (diffRsCsdiff < 2) {
if (diffMVdiffVal < 2) {
if (TSC < 6) {
if (Rs < 1954) {
return 1;
}
else if (Rs >= 1954) {
return 2;
}
}
else if (TSC >= 6) {
if (RsDiff < 8) {
if (Rs < 3080) {
return 8;
}
else if (Rs >= 3080) {
return 4;
}
}
else if (RsDiff >= 8) {
return 4;
}
}
}
else if (diffMVdiffVal >= 2) {
return 2;
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 4112) {
if (Cs < 1346) {
return 4;
}
else if (Cs >= 1346) {
return 8;
}
}
else if (Rs >= 4112) {
return 1;
}
}
}
}
}
else if (Cs >= 1623) {
if (RsCsDiff < 73) {
if (SC < 4199) {
if (posBalance < 42) {
if (ssDCval < 70) {
if (Rs < 1486) {
return 4;
}
else if (Rs >= 1486) {
return 2;
}
}
else if (ssDCval >= 70) {
if (Rs < 1920) {
return 8;
}
else if (Rs >= 1920) {
return 1;
}
}
}
else if (posBalance >= 42) {
if (mu_mv_mag_sq < 89) {
if (ssDCval < 83) {
if (RsCsDiff < 31) {
if (Cs < 1680) {
return 8;
}
else if (Cs >= 1680) {
return 1;
}
}
else if (RsCsDiff >= 31) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (posBalance < 53) {
if (Rs < 1682) {
return 4;
}
else if (Rs >= 1682) {
if (Rs < 2101) {
return 8;
}
else if (Rs >= 2101) {
return 1;
}
}
}
else if (posBalance >= 53) {
return 8;
}
}
}
}
else if (ssDCval >= 83) {
return 8;
}
}
else if (mu_mv_mag_sq >= 89) {
if (diffTSC < -2) {
if (Rs < 1635) {
return 1;
}
else if (Rs >= 1635) {
return 8;
}
}
else if (diffTSC >= -2) {
return 8;
}
}
}
}
else if (SC >= 4199) {
if (mcTcor < 96) {
if (Cs < 2652) {
if (Rs < 2256) {
return 4;
}
else if (Rs >= 2256) {
if (ssDCval < 71) {
if (Rs < 2795) {
return 8;
}
else if (Rs >= 2795) {
return 1;
}
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (Cs >= 2652) {
if (Rs < 3648) {
return 4;
}
else if (Rs >= 3648) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (ssDCval < 89) {
if (Cs < 3178) {
if (Rs < 3017) {
if (RsDiff < 5) {
if (Cs < 2401) {
if (Rs < 2680) {
return 8;
}
else if (Rs >= 2680) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (Cs >= 2401) {
return 4;
}
}
else if (RsDiff >= 5) {
if (mu_mv_mag_sq < 104) {
if (Rs < 2543) {
return 8;
}
else if (Rs >= 2543) {
return 1;
}
}
else if (mu_mv_mag_sq >= 104) {
return 4;
}
}
}
else if (Rs >= 3017) {
if (CsDiff < 5) {
if (MVDiff < 1) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (CsDiff >= 5) {
if (Rs < 4299) {
return 4;
}
else if (Rs >= 4299) {
return 8;
}
}
}
}
else if (Cs >= 3178) {
if (Rs < 2669) {
return 1;
}
else if (Rs >= 2669) {
return 8;
}
}
}
else if (ssDCval >= 89) {
if (AFD < 7) {
if (Rs < 2367) {
return 1;
}
else if (Rs >= 2367) {
if (Cs < 3001) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (ssDCval < 92) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
return 2;
}
}
else if (ssDCval >= 92) {
return 8;
}
}
}
else if (Cs >= 3001) {
return 4;
}
}
}
else if (AFD >= 7) {
if (MVDiff < 2) {
if (Rs < 3526) {
return 4;
}
else if (Rs >= 3526) {
return 2;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 73) {
if (Rs < 3500) {
if (Cs < 2286) {
if (posBalance < 54) {
if (mcTcor < 91) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
return 4;
}
}
else if (mcTcor >= 91) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 8;
}
}
}
else if (posBalance >= 54) {
if (RsCsDiff < 77) {
return 4;
}
else if (RsCsDiff >= 77) {
return 8;
}
}
}
else if (Cs >= 2286) {
return 1;
}
}
else if (Rs >= 3500) {
if (tcor < 93) {
if (ssDCval < 89) {
if (Cs < 1830) {
return 1;
}
else if (Cs >= 1830) {
return 4;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (tcor >= 93) {
if (posBalance < 43) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
else if (posBalance >= 43) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 6007) {
if (ssDCval < 59) {
if (Rs < 6772) {
return 8;
}
else if (Rs >= 6772) {
return 4;
}
}
else if (ssDCval >= 59) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 1450) {
if (RsCsDiff < 33) {
if (ssDCval < 84) {
if (ssDCval < 61) {
if (Rs < 1015) {
if (refDCval < 59) {
return 2;
}
else if (refDCval >= 59) {
return 8;
}
}
else if (Rs >= 1015) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (Cs < 1363) {
return 4;
}
else if (Cs >= 1363) {
return 8;
}
}
}
}
}
else if (ssDCval >= 61) {
if (SC < 2424) {
if (negBalance < 50) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 4;
}
}
else if (negBalance >= 50) {
return 1;
}
}
else if (SC >= 2424) {
if (AFD < 10) {
if (diffRsCsdiff < -17) {
if (Rs < 1163) {
return 2;
}
else if (Rs >= 1163) {
return 8;
}
}
else if (diffRsCsdiff >= -17) {
if (tcor < 80) {
return 8;
}
else if (tcor >= 80) {
if (Cs < 1362) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (Cs >= 1362) {
return 1;
}
}
}
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 100) {
if (posBalance < 62) {
if (diffRsCsdiff < -15) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (diffRsCsdiff >= -15) {
return 8;
}
}
else if (posBalance >= 62) {
if (Rs < 1371) {
return 2;
}
else if (Rs >= 1371) {
return 4;
}
}
}
else if (ssDCval >= 100) {
return 4;
}
}
}
else if (RsCsDiff >= 33) {
if (ssDCval < 49) {
if (TSC < 7) {
if (MVDiff < 18) {
if (Rs < 1232) {
return 4;
}
else if (Rs >= 1232) {
return 1;
}
}
else if (MVDiff >= 18) {
return 2;
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (ssDCval >= 49) {
if (mcTcor < 83) {
if (CsDiff < 7) {
if (ssDCval < 88) {
return 8;
}
else if (ssDCval >= 88) {
if (Rs < 1421) {
return 8;
}
else if (Rs >= 1421) {
return 4;
}
}
}
else if (CsDiff >= 7) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
else if (mcTcor >= 83) {
if (mcTcor < 96) {
if (posBalance < 58) {
if (ssDCval < 95) {
if (mcTcor < 90) {
if (SC < 2584) {
return 2;
}
else if (SC >= 2584) {
if (Cs < 1394) {
return 1;
}
else if (Cs >= 1394) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (SC < 2897) {
if (Cs < 1411) {
if (diffRsCsdiff < -22) {
if (SC < 2763) {
return 4;
}
else if (SC >= 2763) {
return 8;
}
}
else if (diffRsCsdiff >= -22) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 1138) {
return 1;
}
else if (Rs >= 1138) {
return 4;
}
}
}
}
else if (Cs >= 1411) {
if (tcor < 87) {
if (AFD < 16) {
if (Rs < 891) {
return 4;
}
else if (Rs >= 891) {
return 8;
}
}
else if (AFD >= 16) {
return 8;
}
}
else if (tcor >= 87) {
return 4;
}
}
}
else if (SC >= 2897) {
return 8;
}
}
}
else if (ssDCval >= 95) {
if (Rs < 1410) {
return 2;
}
else if (Rs >= 1410) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (mcTcor < 87) {
if (SC < 2759) {
if (Cs < 1383) {
if (Cs < 1324) {
return 2;
}
else if (Cs >= 1324) {
return 4;
}
}
else if (Cs >= 1383) {
return 2;
}
}
else if (SC >= 2759) {
return 8;
}
}
else if (mcTcor >= 87) {
if (posBalance < 67) {
if (Cs < 1369) {
if (Cs < 1343) {
if (RsDiff < 5) {
if (Rs < 1075) {
return 8;
}
else if (Rs >= 1075) {
return 2;
}
}
else if (RsDiff >= 5) {
if (Cs < 1322) {
return 8;
}
else if (Cs >= 1322) {
return 4;
}
}
}
else if (Cs >= 1343) {
return 8;
}
}
else if (Cs >= 1369) {
if (negBalance < 57) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
if (Cs < 1489) {
return 4;
}
else if (Cs >= 1489) {
if (mu_mv_mag_sq < 17) {
if (Cs < 1524) {
if (Rs < 928) {
return 4;
}
else if (Rs >= 928) {
return 8;
}
}
else if (Cs >= 1524) {
if (Cs < 1778) {
return 4;
}
else if (Cs >= 1778) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 17) {
return 8;
}
}
}
}
else if (negBalance >= 57) {
return 8;
}
}
}
else if (posBalance >= 67) {
if (Cs < 1628) {
return 8;
}
else if (Cs >= 1628) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (AFD < 11) {
if (ssDCval < 64) {
if (Rs < 1081) {
return 1;
}
else if (Rs >= 1081) {
return 8;
}
}
else if (ssDCval >= 64) {
if (RsDiff < 4) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (Cs < 1718) {
return 2;
}
else if (Cs >= 1718) {
return 8;
}
}
}
else if (RsDiff >= 4) {
return 2;
}
}
}
else if (AFD >= 11) {
if (RsDiff < 5) {
if (mu_mv_mag_sq < 27) {
if (ssDCval < 71) {
if (Rs < 767) {
return 2;
}
else if (Rs >= 767) {
return 8;
}
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (mu_mv_mag_sq >= 27) {
return 4;
}
}
else if (RsDiff >= 5) {
if (ssDCval < 89) {
if (Rs < 1060) {
return 2;
}
else if (Rs >= 1060) {
return 4;
}
}
else if (ssDCval >= 89) {
if (Rs < 1310) {
return 8;
}
else if (Rs >= 1310) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1450) {
if (mcTcor < 93) {
if (CsDiff < 7) {
if (negBalance < 58) {
if (refDCval < 87) {
if (RsCsDiff < 67) {
if (posBalance < 69) {
if (RsCsDiff < 63) {
return 8;
}
else if (RsCsDiff >= 63) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
return 2;
}
}
}
else if (posBalance >= 69) {
return 4;
}
}
else if (RsCsDiff >= 67) {
if (MVDiff < 17) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
if (ssDCval < 75) {
if (diffRsCsdiff < -38) {
if (Rs < 2626) {
return 4;
}
else if (Rs >= 2626) {
return 8;
}
}
else if (diffRsCsdiff >= -38) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (ssDCval >= 75) {
if (mu_mv_mag_sq < 70) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (ssDCval < 83) {
if (posBalance < 59) {
if (Cs < 1726) {
return 8;
}
else if (Cs >= 1726) {
return 4;
}
}
else if (posBalance >= 59) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
}
else if (ssDCval >= 83) {
return 8;
}
}
}
}
}
else if (MVDiff >= 17) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (mu_mv_mag_sq < 59) {
return 8;
}
else if (mu_mv_mag_sq >= 59) {
if (Rs < 1936) {
return 8;
}
else if (Rs >= 1936) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 87) {
if (Rs < 2227) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (RsDiff < 7) {
if (diffRsCsdiff < 19) {
if (Cs < 2275) {
return 8;
}
else if (Cs >= 2275) {
return 2;
}
}
else if (diffRsCsdiff >= 19) {
if (Rs < 1685) {
return 8;
}
else if (Rs >= 1685) {
return 2;
}
}
}
else if (RsDiff >= 7) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
if (Rs < 2183) {
return 8;
}
else if (Rs >= 2183) {
return 2;
}
}
}
}
}
else if (Rs >= 2227) {
return 8;
}
}
}
else if (negBalance >= 58) {
if (refDCval < 78) {
return 8;
}
else if (refDCval >= 78) {
if (posBalance < 60) {
if (SC < 3839) {
if (Cs < 1621) {
return 8;
}
else if (Cs >= 1621) {
return 1;
}
}
else if (SC >= 3839) {
return 4;
}
}
else if (posBalance >= 60) {
return 4;
}
}
}
}
else if (CsDiff >= 7) {
if (refDCval < 94) {
if (refDCval < 50) {
if (negBalance < 48) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (negBalance >= 48) {
if (Cs < 1492) {
return 8;
}
else if (Cs >= 1492) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (refDCval >= 50) {
if (Cs < 2767) {
if (AFD < 16) {
if (posBalance < 55) {
if (mcTcor < 90) {
if (AFD < 10) {
if (Rs < 1596) {
return 8;
}
else if (Rs >= 1596) {
return 4;
}
}
else if (AFD >= 10) {
return 8;
}
}
else if (mcTcor >= 90) {
return 8;
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (AFD >= 16) {
if (TSC < 10) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
if (Rs < 2158) {
return 1;
}
else if (Rs >= 2158) {
return 2;
}
}
}
else if (TSC >= 10) {
if (refDCval < 86) {
return 8;
}
else if (refDCval >= 86) {
if (posBalance < 62) {
if (Rs < 2458) {
return 8;
}
else if (Rs >= 2458) {
return 4;
}
}
else if (posBalance >= 62) {
return 4;
}
}
}
}
}
else if (Cs >= 2767) {
if (Rs < 3004) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
else if (Rs >= 3004) {
if (Cs < 2843) {
return 4;
}
else if (Cs >= 2843) {
return 8;
}
}
}
}
}
else if (refDCval >= 94) {
if (mu_mv_mag_sq < 88) {
if (negBalance < 59) {
if (RsDiff < 7) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
else if (RsDiff >= 7) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
if (Rs < 3788) {
return 8;
}
else if (Rs >= 3788) {
return 1;
}
}
}
}
else if (negBalance >= 59) {
return 1;
}
}
else if (mu_mv_mag_sq >= 88) {
if (Cs < 2662) {
return 1;
}
else if (Cs >= 2662) {
if (Rs < 4489) {
return 4;
}
else if (Rs >= 4489) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (SC < 7530) {
if (SC < 5039) {
if (tcor < 88) {
if (ssDCval < 85) {
if (diffRsCsdiff < -14) {
if (mcTcor < 94) {
if (AFD < 17) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
return 8;
}
}
else if (AFD >= 17) {
return 1;
}
}
else if (mcTcor >= 94) {
if (diffMVdiffVal < 4) {
if (diffRsCsdiff < -22) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Rs < 1979) {
return 2;
}
else if (Rs >= 1979) {
return 4;
}
}
}
else if (diffRsCsdiff >= -22) {
return 4;
}
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
}
else if (diffRsCsdiff >= -14) {
if (posBalance < 53) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 2;
}
}
else if (posBalance >= 53) {
if (diffMVdiffVal < 1) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
if (CsDiff < 7) {
if (Rs < 1909) {
return 8;
}
else if (Rs >= 1909) {
return 4;
}
}
else if (CsDiff >= 7) {
return 8;
}
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
}
}
else if (ssDCval >= 85) {
if (negBalance < 47) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 2391) {
return 8;
}
else if (Rs >= 2391) {
return 4;
}
}
}
else if (negBalance >= 47) {
return 8;
}
}
}
else if (tcor >= 88) {
if (TSC < 9) {
if (RsCsDiff < 33) {
if (gchDC < 3) {
if (SC < 3637) {
if (AFD < 10) {
if (Rs < 1465) {
if (Rs < 1461) {
return 1;
}
else if (Rs >= 1461) {
return 4;
}
}
else if (Rs >= 1465) {
if (Cs < 1395) {
if (posBalance < 49) {
if (Rs < 2143) {
return 2;
}
else if (Rs >= 2143) {
return 4;
}
}
else if (posBalance >= 49) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1395) {
return 8;
}
}
}
else if (AFD >= 10) {
return 4;
}
}
else if (SC >= 3637) {
if (Rs < 3123) {
if (gchDC < 2) {
if (AFD < 9) {
if (Cs < 1409) {
if (Cs < 1397) {
return 8;
}
else if (Cs >= 1397) {
return 4;
}
}
else if (Cs >= 1409) {
return 8;
}
}
else if (AFD >= 9) {
if (Cs < 1665) {
return 8;
}
else if (Cs >= 1665) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (Cs < 2025) {
return 2;
}
else if (Cs >= 2025) {
return 8;
}
}
}
else if (Rs >= 3123) {
return 4;
}
}
}
else if (gchDC >= 3) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (CsDiff < 2) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (CsDiff >= 2) {
if (Rs < 2116) {
return 2;
}
else if (Rs >= 2116) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 33) {
if (mcTcor < 97) {
if (Rs < 2511) {
if (mu_mv_mag_sq < 96) {
if (MVDiff < 4) {
if (ssDCval < 80) {
if (CsDiff < 5) {
if (Cs < 1538) {
return 1;
}
else if (Cs >= 1538) {
return 8;
}
}
else if (CsDiff >= 5) {
if (AFD < 11) {
if (mu_mv_mag_sq < 94) {
return 8;
}
else if (mu_mv_mag_sq >= 94) {
return 4;
}
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (ssDCval >= 80) {
if (tcor < 94) {
if (negBalance < 48) {
if (Cs < 1693) {
if (ssDCval < 88) {
if (mu_mv_mag_sq < 29) {
return 8;
}
else if (mu_mv_mag_sq >= 29) {
if (posBalance < 53) {
if (Cs < 1373) {
return 1;
}
else if (Cs >= 1373) {
return 8;
}
}
else if (posBalance >= 53) {
return 1;
}
}
}
else if (ssDCval >= 88) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (Cs >= 1693) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (Rs < 1653) {
return 8;
}
else if (Rs >= 1653) {
return 1;
}
}
}
else if (tcor >= 94) {
if (Cs < 1541) {
if (Rs < 2293) {
if (Rs < 2111) {
return 1;
}
else if (Rs >= 2111) {
return 2;
}
}
else if (Rs >= 2293) {
return 4;
}
}
else if (Cs >= 1541) {
if (MVDiff < 1) {
if (Rs < 2061) {
return 2;
}
else if (Rs >= 2061) {
return 1;
}
}
else if (MVDiff >= 1) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 2146) {
if (MVDiff < 7) {
if (refDCval < 80) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (refDCval >= 80) {
if (Cs < 1489) {
if (Rs < 1832) {
return 8;
}
else if (Rs >= 1832) {
return 2;
}
}
else if (Cs >= 1489) {
return 8;
}
}
}
else if (MVDiff >= 7) {
if (SC < 3489) {
return 8;
}
else if (SC >= 3489) {
if (mu_mv_mag_sq < 17) {
return 2;
}
else if (mu_mv_mag_sq >= 17) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 2091) {
return 2;
}
else if (Rs >= 2091) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2146) {
if (ssDCval < 96) {
if (TSC < 8) {
if (MVDiff < 6) {
if (Rs < 2328) {
return 1;
}
else if (Rs >= 2328) {
return 4;
}
}
else if (MVDiff >= 6) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 2220) {
return 8;
}
else if (Rs >= 2220) {
return 1;
}
}
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (ssDCval >= 96) {
if (Rs < 2241) {
return 1;
}
else if (Rs >= 2241) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (Cs < 1324) {
return 2;
}
else if (Cs >= 1324) {
if (diffMVdiffVal < -5) {
return 2;
}
else if (diffMVdiffVal >= -5) {
if (diffAFD < 1) {
if (Cs < 2204) {
return 8;
}
else if (Cs >= 2204) {
if (Rs < 2129) {
return 1;
}
else if (Rs >= 2129) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (diffRsCsdiff < -5) {
return 2;
}
else if (diffRsCsdiff >= -5) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2511) {
if (RsDiff < 7) {
if (mu_mv_mag_sq < 61) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (negBalance < 53) {
if (ssDCval < 69) {
if (Rs < 2867) {
return 4;
}
else if (Rs >= 2867) {
return 8;
}
}
else if (ssDCval >= 69) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
return 8;
}
}
}
else if (negBalance >= 53) {
return 4;
}
}
}
else if (RsDiff >= 7) {
if (MVDiff < 5) {
if (Cs < 1562) {
return 2;
}
else if (Cs >= 1562) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 1955) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (Rs >= 1955) {
if (SC < 4050) {
if (SC < 3355) {
return 1;
}
else if (SC >= 3355) {
return 8;
}
}
else if (SC >= 4050) {
if (Cs < 1429) {
return 2;
}
else if (Cs >= 1429) {
if (diffTSC < 1) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (diffTSC >= 1) {
if (RsCsDiff < 62) {
if (Rs < 2781) {
return 8;
}
else if (Rs >= 2781) {
return 4;
}
}
else if (RsCsDiff >= 62) {
return 2;
}
}
}
}
}
}
}
}
else if (TSC >= 9) {
if (Rs < 2074) {
if (SC < 3788) {
if (refDCval < 94) {
return 8;
}
else if (refDCval >= 94) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 1742) {
return 2;
}
else if (Rs >= 1742) {
return 1;
}
}
}
}
else if (SC >= 3788) {
if (Rs < 2025) {
if (Rs < 1753) {
return 2;
}
else if (Rs >= 1753) {
return 4;
}
}
else if (Rs >= 2025) {
return 1;
}
}
}
else if (Rs >= 2074) {
return 8;
}
}
}
}
else if (SC >= 5039) {
if (Rs < 2806) {
if (mu_mv_mag_sq < 61) {
if (Rs < 2428) {
return 8;
}
else if (Rs >= 2428) {
return 2;
}
}
else if (mu_mv_mag_sq >= 61) {
if (Cs < 2563) {
return 8;
}
else if (Cs >= 2563) {
if (diffRsCsdiff < 44) {
return 4;
}
else if (diffRsCsdiff >= 44) {
return 8;
}
}
}
}
else if (Rs >= 2806) {
if (Cs < 1863) {
if (ssDCval < 89) {
if (AFD < 9) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (ssDCval >= 89) {
if (mu_mv_mag_sq < 79) {
if (MVDiff < 25) {
return 8;
}
else if (MVDiff >= 25) {
return 1;
}
}
else if (mu_mv_mag_sq >= 79) {
return 1;
}
}
}
else if (Cs >= 1863) {
if (mu_mv_mag_sq < 54) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
return 8;
}
}
else if (mu_mv_mag_sq >= 54) {
if (RsCsDiff < 53) {
if (AFD < 9) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (RsCsDiff >= 53) {
if (negBalance < 50) {
if (Cs < 3068) {
if (AFD < 8) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
return 8;
}
}
else if (AFD >= 8) {
if (SC < 5505) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (SC >= 5505) {
return 4;
}
}
}
else if (Cs >= 3068) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (negBalance >= 50) {
if (mu_mv_mag_sq < 84) {
return 4;
}
else if (mu_mv_mag_sq >= 84) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 7530) {
return 8;
}
}
}
}
}
else if (ssDCval >= 105) {
if (ssDCval < 164) {
if (Rs < 4772) {
if (Rs < 1513) {
if (mu_mv_mag_sq < 47) {
if (ssDCval < 141) {
if (ssDCval < 111) {
if (CsDiff < 7) {
if (refDCval < 111) {
return 4;
}
else if (refDCval >= 111) {
return 8;
}
}
else if (CsDiff >= 7) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (ssDCval >= 111) {
if (AFD < 18) {
if (mcTcor < 90) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (mcTcor >= 90) {
if (Rs < 1425) {
return 8;
}
else if (Rs >= 1425) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 1500) {
return 1;
}
else if (Rs >= 1500) {
if (Rs < 1509) {
return 8;
}
else if (Rs >= 1509) {
return 2;
}
}
}
}
}
}
else if (AFD >= 18) {
if (Rs < 906) {
return 2;
}
else if (Rs >= 906) {
return 4;
}
}
}
}
else if (ssDCval >= 141) {
if (mu_mv_mag_sq < 38) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 1410) {
return 4;
}
else if (Rs >= 1410) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (refDCval < 157) {
if (Rs < 1230) {
if (Rs < 693) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Rs < 619) {
return 4;
}
else if (Rs >= 619) {
return 8;
}
}
}
else if (Rs >= 693) {
if (posBalance < 55) {
if (ssDCval < 115) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (ssDCval >= 115) {
if (Rs < 1019) {
return 8;
}
else if (Rs >= 1019) {
if (RsDiff < 3) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
return 4;
}
}
else if (RsDiff >= 3) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
if (Rs < 968) {
return 8;
}
else if (Rs >= 968) {
if (negBalance < 41) {
if (posBalance < 62) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 103) {
return 4;
}
else if (mu_mv_mag_sq >= 103) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 4;
}
}
}
}
else if (posBalance >= 62) {
return 8;
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
}
}
else if (Rs >= 1230) {
if (SC < 3328) {
if (AFD < 3) {
if (Rs < 1302) {
return 4;
}
else if (Rs >= 1302) {
return 8;
}
}
else if (AFD >= 3) {
if (Cs < 1315) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (Cs >= 1315) {
return 8;
}
}
}
else if (SC >= 3328) {
return 4;
}
}
}
else if (refDCval >= 157) {
if (Rs < 1183) {
return 1;
}
else if (Rs >= 1183) {
return 2;
}
}
}
}
else if (Rs >= 1513) {
if (Rs < 3677) {
if (mu_mv_mag_sq < 42) {
if (Rs < 2319) {
if (mu_mv_mag_sq < 7) {
if (mcTcor < 97) {
if (ssDCval < 126) {
if (Cs < 2490) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Rs < 2043) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
else if (Rs >= 2043) {
return 1;
}
}
}
else if (Cs >= 2490) {
return 2;
}
}
else if (ssDCval >= 126) {
if (ssDCval < 149) {
if (Rs < 1937) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (Rs < 1803) {
return 4;
}
else if (Rs >= 1803) {
return 1;
}
}
}
}
}
else if (Rs >= 1937) {
return 8;
}
}
else if (ssDCval >= 149) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (CsDiff < 7) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 2113) {
return 2;
}
else if (Rs >= 2113) {
return 8;
}
}
}
else if (CsDiff >= 7) {
if (Rs < 1924) {
return 8;
}
else if (Rs >= 1924) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffRsCsdiff < 19) {
if (mu_mv_mag_sq < 30) {
if (mu_mv_mag_sq < 26) {
return 8;
}
else if (mu_mv_mag_sq >= 26) {
if (Rs < 1746) {
return 4;
}
else if (Rs >= 1746) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 30) {
return 8;
}
}
else if (diffRsCsdiff >= 19) {
if (Rs < 1854) {
if (Rs < 1793) {
return 8;
}
else if (Rs >= 1793) {
return 1;
}
}
else if (Rs >= 1854) {
return 8;
}
}
}
}
else if (Rs >= 2319) {
if (ssDCval < 110) {
if (TSC < 7) {
if (AFD < 8) {
if (Rs < 2813) {
return 4;
}
else if (Rs >= 2813) {
return 8;
}
}
else if (AFD >= 8) {
return 4;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (ssDCval >= 110) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (ssDCval < 142) {
if (refDCval < 123) {
if (TSC < 11) {
if (tcor < 97) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (RsCsDiff < 43) {
if (Rs < 2593) {
return 8;
}
else if (Rs >= 2593) {
if (TSC < 6) {
if (diffTSC < -1) {
if (Cs < 2883) {
if (MVDiff < 31) {
return 4;
}
else if (MVDiff >= 31) {
return 8;
}
}
else if (Cs >= 2883) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (TSC >= 6) {
if (refDCval < 111) {
return 4;
}
else if (refDCval >= 111) {
return 8;
}
}
}
}
else if (RsCsDiff >= 43) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
return 4;
}
}
}
}
else if (tcor >= 97) {
if (posBalance < 10) {
if (mu_mv_mag_sq < 106) {
return 8;
}
else if (mu_mv_mag_sq >= 106) {
if (Cs < 2026) {
return 1;
}
else if (Cs >= 2026) {
if (Rs < 2614) {
return 2;
}
else if (Rs >= 2614) {
return 8;
}
}
}
}
else if (posBalance >= 10) {
if (RsDiff < 5) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (RsDiff >= 5) {
if (Rs < 2067) {
return 2;
}
else if (Rs >= 2067) {
return 8;
}
}
}
}
}
else if (TSC >= 11) {
if (Cs < 1768) {
if (Rs < 2312) {
if (Cs < 1419) {
return 1;
}
else if (Cs >= 1419) {
return 8;
}
}
else if (Rs >= 2312) {
return 1;
}
}
else if (Cs >= 1768) {
if (ssDCval < 108) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (ssDCval >= 108) {
return 8;
}
}
}
}
else if (refDCval >= 123) {
if (CsDiff < 2) {
if (TSC < 4) {
if (TSC < 3) {
if (Cs < 2970) {
return 8;
}
else if (Cs >= 2970) {
if (Rs < 2715) {
return 8;
}
else if (Rs >= 2715) {
return 2;
}
}
}
else if (TSC >= 3) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 2627) {
return 4;
}
else if (Rs >= 2627) {
return 8;
}
}
}
}
else if (TSC >= 4) {
return 4;
}
}
else if (CsDiff >= 2) {
if (mcTcor < 89) {
if (Cs < 1872) {
if (Cs < 1599) {
return 8;
}
else if (Cs >= 1599) {
return 1;
}
}
else if (Cs >= 1872) {
return 8;
}
}
else if (mcTcor >= 89) {
if (Cs < 1344) {
if (mu_mv_mag_sq < 81) {
return 8;
}
else if (mu_mv_mag_sq >= 81) {
return 2;
}
}
else if (Cs >= 1344) {
if (tcor < 85) {
if (ssDCval < 138) {
return 8;
}
else if (ssDCval >= 138) {
return 4;
}
}
else if (tcor >= 85) {
if (MVDiff < 1) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
return 1;
}
}
else if (MVDiff >= 1) {
if (SC < 4737) {
if (TSC < 9) {
if (mu_mv_mag_sq < 92) {
if (mcTcor < 90) {
if (Rs < 1938) {
return 1;
}
else if (Rs >= 1938) {
return 8;
}
}
else if (mcTcor >= 90) {
return 8;
}
}
else if (mu_mv_mag_sq >= 92) {
if (mcTcor < 97) {
if (diffRsCsdiff < -21) {
if (Cs < 1868) {
return 1;
}
else if (Cs >= 1868) {
return 8;
}
}
else if (diffRsCsdiff >= -21) {
return 8;
}
}
else if (mcTcor >= 97) {
if (negBalance < 32) {
return 1;
}
else if (negBalance >= 32) {
return 8;
}
}
}
}
else if (TSC >= 9) {
if (Cs < 1787) {
if (Rs < 2040) {
return 8;
}
else if (Rs >= 2040) {
return 1;
}
}
else if (Cs >= 1787) {
return 8;
}
}
}
else if (SC >= 4737) {
if (Rs < 3643) {
return 8;
}
else if (Rs >= 3643) {
if (Rs < 3649) {
return 1;
}
else if (Rs >= 3649) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 142) {
if (posBalance < 66) {
if (ssDCval < 162) {
if (posBalance < 30) {
return 2;
}
else if (posBalance >= 30) {
return 8;
}
}
else if (ssDCval >= 162) {
return 2;
}
}
else if (posBalance >= 66) {
if (Rs < 2107) {
return 4;
}
else if (Rs >= 2107) {
return 8;
}
}
}
}
}
else if (Rs >= 3677) {
if (mu_mv_mag_sq < 98) {
if (Cs < 1373) {
if (tcor < 94) {
if (Rs < 4112) {
return 8;
}
else if (Rs >= 4112) {
return 1;
}
}
else if (tcor >= 94) {
return 2;
}
}
else if (Cs >= 1373) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (ssDCval < 137) {
if (Cs < 1688) {
return 2;
}
else if (Cs >= 1688) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 4;
}
}
}
else if (ssDCval >= 137) {
if (AFD < 7) {
if (Rs < 3730) {
return 8;
}
else if (Rs >= 3730) {
return 2;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
}
}
}
else if (Rs >= 4772) {
if (Cs < 1862) {
return 2;
}
else if (Cs >= 1862) {
if (Rs < 5947) {
return 8;
}
else if (Rs >= 5947) {
if (Rs < 8356) {
return 4;
}
else if (Rs >= 8356) {
return 8;
}
}
}
}
}
else if (ssDCval >= 164) {
if (Rs < 2283) {
if (Rs < 1048) {
return 1;
}
else if (Rs >= 1048) {
return 2;
}
}
else if (Rs >= 2283) {
return 8;
}
}
}
}
else if (CsDiff >= 8) {
if (Rs < 3843) {
if (Rs < 1532) {
if (negBalance < 44) {
if (MVDiff < 18) {
if (posBalance < 55) {
if (posBalance < 48) {
if (Rs < 641) {
return 1;
}
else if (Rs >= 641) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (diffMVdiffVal < -5) {
return 2;
}
else if (diffMVdiffVal >= -5) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (AFD < 11) {
if (SC < 2794) {
return 1;
}
else if (SC >= 2794) {
return 8;
}
}
else if (AFD >= 11) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
if (Rs < 985) {
return 8;
}
else if (Rs >= 985) {
return 2;
}
}
}
}
}
else if (MVDiff >= 18) {
if (mu_mv_mag_sq < 34) {
if (TSC < 11) {
if (RsCsDiff < 99) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 4;
}
}
else if (RsCsDiff >= 99) {
if (negBalance < 43) {
if (Cs < 1518) {
return 8;
}
else if (Cs >= 1518) {
if (ssDCval < 155) {
return 2;
}
else if (ssDCval >= 155) {
return 1;
}
}
}
else if (negBalance >= 43) {
return 1;
}
}
}
else if (TSC >= 11) {
if (RsCsDiff < 189) {
if (Rs < 1280) {
return 8;
}
else if (Rs >= 1280) {
return 4;
}
}
else if (RsCsDiff >= 189) {
if (Rs < 951) {
return 2;
}
else if (Rs >= 951) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (ssDCval < 65) {
if (tcor < 63) {
if (Cs < 2657) {
return 4;
}
else if (Cs >= 2657) {
return 2;
}
}
else if (tcor >= 63) {
return 1;
}
}
else if (ssDCval >= 65) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (Rs < 944) {
return 1;
}
else if (Rs >= 944) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 44) {
if (tcor < 67) {
if (AFD < 17) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (AFD >= 17) {
if (Rs < 1434) {
if (ssDCval < 102) {
if (ssDCval < 61) {
if (Rs < 627) {
return 4;
}
else if (Rs >= 627) {
return 2;
}
}
else if (ssDCval >= 61) {
if (Rs < 1266) {
if (diffRsCsdiff < 38) {
if (Cs < 2505) {
if (AFD < 25) {
if (tcor < 63) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
if (Cs < 1385) {
return 4;
}
else if (Cs >= 1385) {
return 2;
}
}
}
else if (tcor >= 63) {
return 2;
}
}
else if (AFD >= 25) {
return 4;
}
}
else if (Cs >= 2505) {
if (Cs < 4396) {
return 2;
}
else if (Cs >= 4396) {
return 8;
}
}
}
else if (diffRsCsdiff >= 38) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
return 8;
}
}
}
else if (Rs >= 1266) {
if (Rs < 1424) {
return 4;
}
else if (Rs >= 1424) {
return 8;
}
}
}
}
else if (ssDCval >= 102) {
return 2;
}
}
else if (Rs >= 1434) {
if (diffAFD < 2) {
if (negBalance < 62) {
if (MVDiff < 21) {
if (Cs < 1951) {
return 8;
}
else if (Cs >= 1951) {
return 4;
}
}
else if (MVDiff >= 21) {
return 4;
}
}
else if (negBalance >= 62) {
if (Rs < 1468) {
return 1;
}
else if (Rs >= 1468) {
return 4;
}
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
}
else if (tcor >= 67) {
if (ssDCval < 93) {
if (mu_mv_mag_sq < 2) {
if (mcTcor < 94) {
if (TSC < 11) {
if (MVDiff < 32) {
return 4;
}
else if (MVDiff >= 32) {
if (Rs < 912) {
return 4;
}
else if (Rs >= 912) {
return 2;
}
}
}
else if (TSC >= 11) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (Cs < 1521) {
return 8;
}
else if (Cs >= 1521) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 852) {
return 8;
}
else if (Rs >= 852) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (MVDiff < 13) {
if (mcTcor < 96) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 4;
}
}
else if (mcTcor >= 96) {
if (SC < 2381) {
return 8;
}
else if (SC >= 2381) {
return 1;
}
}
}
else if (MVDiff >= 13) {
if (ssDCval < 75) {
if (AFD < 19) {
if (tcor < 72) {
return 4;
}
else if (tcor >= 72) {
if (Cs < 3433) {
return 8;
}
else if (Cs >= 3433) {
return 1;
}
}
}
else if (AFD >= 19) {
if (Rs < 1118) {
return 2;
}
else if (Rs >= 1118) {
return 8;
}
}
}
else if (ssDCval >= 75) {
if (MVDiff < 14) {
if (Cs < 1459) {
return 2;
}
else if (Cs >= 1459) {
return 8;
}
}
else if (MVDiff >= 14) {
if (AFD < 21) {
if (Cs < 1354) {
return 8;
}
else if (Cs >= 1354) {
return 4;
}
}
else if (AFD >= 21) {
if (Rs < 998) {
return 8;
}
else if (Rs >= 998) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 93) {
if (posBalance < 62) {
if (SC < 3061) {
if (Cs < 1830) {
if (diffRsCsdiff < -28) {
if (Rs < 1326) {
return 2;
}
else if (Rs >= 1326) {
return 8;
}
}
else if (diffRsCsdiff >= -28) {
if (AFD < 25) {
if (Rs < 1309) {
return 8;
}
else if (Rs >= 1309) {
if (Rs < 1371) {
return 2;
}
else if (Rs >= 1371) {
return 8;
}
}
}
else if (AFD >= 25) {
return 4;
}
}
}
else if (Cs >= 1830) {
return 2;
}
}
else if (SC >= 3061) {
if (diffRsCsdiff < -72) {
if (Rs < 1392) {
return 1;
}
else if (Rs >= 1392) {
return 4;
}
}
else if (diffRsCsdiff >= -72) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (MVDiff < 17) {
if (ssDCval < 106) {
if (mcTcor < 90) {
return 2;
}
else if (mcTcor >= 90) {
return 8;
}
}
else if (ssDCval >= 106) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 4;
}
}
}
else if (MVDiff >= 17) {
if (Rs < 1368) {
return 1;
}
else if (Rs >= 1368) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1532) {
if (Cs < 2794) {
if (AFD < 17) {
if (negBalance < 38) {
if (MVDiff < 4) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 8;
}
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 8;
}
}
}
}
else if (negBalance >= 38) {
if (ssDCval < 52) {
if (Cs < 1733) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (Cs >= 1733) {
if (Rs < 2337) {
return 4;
}
else if (Rs >= 2337) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (Rs < 2981) {
if (mu_mv_mag_sq < 77) {
if (MVDiff < 4) {
if (SC < 5145) {
if (Cs < 2314) {
return 8;
}
else if (Cs >= 2314) {
if (ssDCval < 87) {
if (Rs < 2386) {
return 1;
}
else if (Rs >= 2386) {
return 8;
}
}
else if (ssDCval >= 87) {
if (Rs < 1866) {
return 4;
}
else if (Rs >= 1866) {
return 8;
}
}
}
}
else if (SC >= 5145) {
if (Rs < 2734) {
return 2;
}
else if (Rs >= 2734) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (SC < 3999) {
if (diffAFD < -2) {
if (Rs < 2018) {
return 1;
}
else if (Rs >= 2018) {
return 8;
}
}
else if (diffAFD >= -2) {
if (mcTcor < 96) {
if (SC < 3980) {
if (mcTcor < 78) {
return 1;
}
else if (mcTcor >= 78) {
return 8;
}
}
else if (SC >= 3980) {
if (Rs < 1902) {
if (Rs < 1780) {
return 4;
}
else if (Rs >= 1780) {
return 1;
}
}
else if (Rs >= 1902) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Rs < 1762) {
return 4;
}
else if (Rs >= 1762) {
return 2;
}
}
}
}
}
else if (SC >= 3999) {
if (diffTSC < -2) {
if (Rs < 2476) {
return 8;
}
else if (Rs >= 2476) {
return 4;
}
}
else if (diffTSC >= -2) {
if (Rs < 1573) {
return 4;
}
else if (Rs >= 1573) {
if (diffAFD < 2) {
if (Rs < 2877) {
return 8;
}
else if (Rs >= 2877) {
if (Rs < 2922) {
return 4;
}
else if (Rs >= 2922) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 1882) {
return 4;
}
else if (Rs >= 1882) {
if (Rs < 2721) {
return 8;
}
else if (Rs >= 2721) {
if (Rs < 2800) {
return 4;
}
else if (Rs >= 2800) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
return 8;
}
}
else if (Rs >= 2981) {
if (ssDCval < 115) {
if (CsDiff < 9) {
if (SC < 6033) {
if (Rs < 3007) {
return 4;
}
else if (Rs >= 3007) {
return 8;
}
}
else if (SC >= 6033) {
return 4;
}
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (ssDCval >= 115) {
if (Rs < 3124) {
return 1;
}
else if (Rs >= 3124) {
return 8;
}
}
}
}
}
}
else if (AFD >= 17) {
if (refDCval < 83) {
if (tcor < 67) {
if (TSC < 16) {
return 4;
}
else if (TSC >= 16) {
return 8;
}
}
else if (tcor >= 67) {
if (refDCval < 78) {
if (RsCsDiff < 121) {
if (Rs < 1785) {
return 4;
}
else if (Rs >= 1785) {
return 1;
}
}
else if (RsCsDiff >= 121) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
if (negBalance < 59) {
return 1;
}
else if (negBalance >= 59) {
return 2;
}
}
}
}
else if (refDCval >= 78) {
if (refDCval < 80) {
return 2;
}
else if (refDCval >= 80) {
if (Rs < 1636) {
return 2;
}
else if (Rs >= 1636) {
return 4;
}
}
}
}
}
else if (refDCval >= 83) {
if (tcor < 84) {
if (mcTcor < 77) {
if (Cs < 2344) {
return 1;
}
else if (Cs >= 2344) {
return 4;
}
}
else if (mcTcor >= 77) {
if (posBalance < 54) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
if (AFD < 28) {
return 8;
}
else if (AFD >= 28) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (negBalance < 59) {
if (MVDiff < 9) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
if (posBalance < 62) {
if (negBalance < 52) {
return 4;
}
else if (negBalance >= 52) {
return 8;
}
}
else if (posBalance >= 62) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1956) {
return 8;
}
else if (Rs >= 1956) {
return 4;
}
}
}
}
}
else if (MVDiff >= 9) {
if (RsCsDiff < 157) {
if (Rs < 1603) {
if (Rs < 1582) {
return 8;
}
else if (Rs >= 1582) {
return 4;
}
}
else if (Rs >= 1603) {
return 8;
}
}
else if (RsCsDiff >= 157) {
if (ssDCval < 99) {
if (Rs < 1848) {
return 1;
}
else if (Rs >= 1848) {
return 4;
}
}
else if (ssDCval >= 99) {
return 8;
}
}
}
}
else if (negBalance >= 59) {
if (ssDCval < 119) {
if (Rs < 2102) {
return 8;
}
else if (Rs >= 2102) {
return 4;
}
}
else if (ssDCval >= 119) {
return 2;
}
}
}
}
}
else if (tcor >= 84) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (Cs < 1919) {
return 8;
}
else if (Cs >= 1919) {
return 1;
}
}
}
}
}
}
else if (Cs >= 2794) {
if (ssDCval < 103) {
if (mcTcor < 94) {
if (negBalance < 48) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 3514) {
if (refDCval < 83) {
return 4;
}
else if (refDCval >= 83) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (Rs < 2384) {
return 4;
}
else if (Rs >= 2384) {
return 8;
}
}
}
}
else if (Rs >= 3514) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (ssDCval < 81) {
if (posBalance < 54) {
if (Rs < 2823) {
return 4;
}
else if (Rs >= 2823) {
return 1;
}
}
else if (posBalance >= 54) {
if (Rs < 2420) {
if (Rs < 2070) {
return 8;
}
else if (Rs >= 2070) {
return 1;
}
}
else if (Rs >= 2420) {
return 8;
}
}
}
else if (ssDCval >= 81) {
if (negBalance < 54) {
if (mu_mv_mag_sq < 63) {
if (Cs < 5183) {
if (RsCsDiff < 138) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (RsCsDiff >= 138) {
return 8;
}
}
else if (Cs >= 5183) {
return 4;
}
}
else if (mu_mv_mag_sq >= 63) {
return 4;
}
}
else if (negBalance >= 54) {
if (refDCval < 90) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
else if (refDCval >= 90) {
return 4;
}
}
}
}
}
else if (mcTcor >= 94) {
if (ssDCval < 86) {
return 1;
}
else if (ssDCval >= 86) {
if (Cs < 3956) {
return 4;
}
else if (Cs >= 3956) {
return 8;
}
}
}
}
else if (ssDCval >= 103) {
if (AFD < 21) {
if (tcor < 91) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (AFD < 11) {
if (Cs < 6851) {
return 1;
}
else if (Cs >= 6851) {
return 8;
}
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (tcor >= 91) {
return 8;
}
}
else if (AFD >= 21) {
return 2;
}
}
}
}
}
else if (Rs >= 3843) {
if (AFD < 12) {
if (TSC < 9) {
if (ssDCval < 116) {
if (Rs < 4231) {
return 4;
}
else if (Rs >= 4231) {
if (Cs < 4019) {
if (tcor < 95) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
return 8;
}
}
else if (tcor >= 95) {
return 8;
}
}
else if (Cs >= 4019) {
if (Rs < 5971) {
return 8;
}
else if (Rs >= 5971) {
return 2;
}
}
}
}
else if (ssDCval >= 116) {
if (CsDiff < 12) {
return 8;
}
else if (CsDiff >= 12) {
return 1;
}
}
}
else if (TSC >= 9) {
if (mu_mv_mag_sq < 95) {
if (diffAFD < -1) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 1;
}
}
else if (diffAFD >= -1) {
if (mu_mv_mag_sq < 77) {
return 8;
}
else if (mu_mv_mag_sq >= 77) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 95) {
return 1;
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 83) {
return 8;
}
else if (mu_mv_mag_sq >= 83) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
if (Cs < 1921) {
if (Rs < 3338) {
if (ssDCval < 111) {
if (mcTcor < 87) {
if (RsDiff < 15) {
if (MVDiff < 16) {
if (ssDCval < 102) {
if (Rs < 1471) {
return 4;
}
else if (Rs >= 1471) {
if (Cs < 1748) {
if (ssDCval < 94) {
if (RsCsDiff < 267) {
return 8;
}
else if (RsCsDiff >= 267) {
if (Cs < 1690) {
return 8;
}
else if (Cs >= 1690) {
return 1;
}
}
}
else if (ssDCval >= 94) {
if (Rs < 2897) {
return 2;
}
else if (Rs >= 2897) {
return 8;
}
}
}
else if (Cs >= 1748) {
if (SC < 3793) {
return 4;
}
else if (SC >= 3793) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
if (Rs < 2348) {
return 2;
}
else if (Rs >= 2348) {
if (Rs < 2861) {
return 4;
}
else if (Rs >= 2861) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (SC < 4877) {
if (Rs < 2113) {
return 8;
}
else if (Rs >= 2113) {
return 4;
}
}
else if (SC >= 4877) {
return 8;
}
}
}
else if (MVDiff >= 16) {
if (gchDC < 2) {
if (ssDCval < 59) {
if (Rs < 1612) {
return 4;
}
else if (Rs >= 1612) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (Rs < 1954) {
return 4;
}
else if (Rs >= 1954) {
return 8;
}
}
}
}
else if (ssDCval >= 59) {
if (tcor < 62) {
if (negBalance < 58) {
if (RsCsDiff < 173) {
if (Rs < 1922) {
return 1;
}
else if (Rs >= 1922) {
return 8;
}
}
else if (RsCsDiff >= 173) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
return 4;
}
}
}
else if (negBalance >= 58) {
if (mu_mv_mag_sq < 6) {
return 2;
}
else if (mu_mv_mag_sq >= 6) {
if (Rs < 1964) {
return 8;
}
else if (Rs >= 1964) {
return 1;
}
}
}
}
else if (tcor >= 62) {
if (Rs < 2478) {
if (posBalance < 56) {
if (Cs < 1587) {
return 2;
}
else if (Cs >= 1587) {
return 4;
}
}
else if (posBalance >= 56) {
if (diffMVdiffVal < 9) {
return 8;
}
else if (diffMVdiffVal >= 9) {
return 4;
}
}
}
else if (Rs >= 2478) {
if (ssDCval < 88) {
if (MVDiff < 66) {
return 4;
}
else if (MVDiff >= 66) {
return 8;
}
}
else if (ssDCval >= 88) {
if (Cs < 1644) {
return 4;
}
else if (Cs >= 1644) {
return 8;
}
}
}
}
}
}
else if (gchDC >= 2) {
if (Cs < 1738) {
if (AFD < 24) {
if (SC < 3449) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (SC >= 3449) {
return 8;
}
}
else if (AFD >= 24) {
if (negBalance < 50) {
if (Rs < 1910) {
return 2;
}
else if (Rs >= 1910) {
return 4;
}
}
else if (negBalance >= 50) {
if (negBalance < 65) {
return 8;
}
else if (negBalance >= 65) {
if (Rs < 1675) {
if (Rs < 1254) {
return 1;
}
else if (Rs >= 1254) {
return 4;
}
}
else if (Rs >= 1675) {
return 8;
}
}
}
}
}
else if (Cs >= 1738) {
if (tcor < 66) {
if (MVDiff < 28) {
return 4;
}
else if (MVDiff >= 28) {
if (Rs < 1549) {
return 4;
}
else if (Rs >= 1549) {
return 1;
}
}
}
else if (tcor >= 66) {
if (Cs < 1810) {
return 8;
}
else if (Cs >= 1810) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 15) {
if (mcTcor < 75) {
if (MVDiff < 109) {
return 8;
}
else if (MVDiff >= 109) {
return 1;
}
}
else if (mcTcor >= 75) {
if (refDCval < 104) {
if (mu_mv_mag_sq < 13) {
return 2;
}
else if (mu_mv_mag_sq >= 13) {
if (gchDC < 3) {
if (Cs < 1688) {
return 2;
}
else if (Cs >= 1688) {
return 4;
}
}
else if (gchDC >= 3) {
if (Cs < 1459) {
if (Rs < 3076) {
return 2;
}
else if (Rs >= 3076) {
return 4;
}
}
else if (Cs >= 1459) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
else if (refDCval >= 104) {
if (Rs < 2553) {
return 4;
}
else if (Rs >= 2553) {
return 8;
}
}
}
}
}
else if (mcTcor >= 87) {
if (negBalance < 46) {
if (Rs < 2451) {
if (tcor < 85) {
if (ssDCval < 63) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 2;
}
}
else if (ssDCval >= 63) {
if (posBalance < 68) {
if (AFD < 20) {
if (diffTSC < -1) {
if (Rs < 2240) {
return 2;
}
else if (Rs >= 2240) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (AFD >= 20) {
if (MVDiff < 42) {
return 2;
}
else if (MVDiff >= 42) {
return 8;
}
}
}
else if (posBalance >= 68) {
if (Cs < 1678) {
return 2;
}
else if (Cs >= 1678) {
return 4;
}
}
}
}
else if (tcor >= 85) {
if (mcTcor < 95) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
return 8;
}
}
else if (mcTcor >= 95) {
if (mu_mv_mag_sq < 29) {
if (Rs < 2053) {
return 1;
}
else if (Rs >= 2053) {
if (Cs < 1535) {
return 2;
}
else if (Cs >= 1535) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 29) {
return 4;
}
}
}
}
else if (Rs >= 2451) {
if (Cs < 1798) {
if (posBalance < 62) {
if (posBalance < 41) {
if (Rs < 3144) {
if (Rs < 2569) {
return 2;
}
else if (Rs >= 2569) {
return 8;
}
}
else if (Rs >= 3144) {
return 4;
}
}
else if (posBalance >= 41) {
if (diffRsCsdiff < 32) {
return 2;
}
else if (diffRsCsdiff >= 32) {
if (RsDiff < 10) {
if (Rs < 3095) {
return 8;
}
else if (Rs >= 3095) {
return 2;
}
}
else if (RsDiff >= 10) {
if (diffRsCsdiff < 37) {
return 4;
}
else if (diffRsCsdiff >= 37) {
if (diffMVdiffVal < -30) {
return 8;
}
else if (diffMVdiffVal >= -30) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 62) {
if (Cs < 1372) {
return 4;
}
else if (Cs >= 1372) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 2495) {
if (Rs < 2465) {
return 8;
}
else if (Rs >= 2465) {
return 2;
}
}
else if (Rs >= 2495) {
return 8;
}
}
}
}
}
else if (Cs >= 1798) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
if (Rs < 2868) {
return 2;
}
else if (Rs >= 2868) {
return 8;
}
}
}
}
}
else if (negBalance >= 46) {
if (AFD < 19) {
if (Rs < 2935) {
if (mu_mv_mag_sq < 7) {
if (refDCval < 86) {
if (diffRsCsdiff < -9) {
if (Rs < 2121) {
return 2;
}
else if (Rs >= 2121) {
return 4;
}
}
else if (diffRsCsdiff >= -9) {
if (CsDiff < 11) {
if (Cs < 1468) {
if (Cs < 1461) {
if (Rs < 1995) {
return 8;
}
else if (Rs >= 1995) {
return 4;
}
}
else if (Cs >= 1461) {
return 2;
}
}
else if (Cs >= 1468) {
if (Cs < 1662) {
return 8;
}
else if (Cs >= 1662) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (diffRsCsdiff < 2) {
if (Rs < 2627) {
return 4;
}
else if (Rs >= 2627) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
}
}
}
else if (CsDiff >= 11) {
if (Rs < 2016) {
return 1;
}
else if (Rs >= 2016) {
return 2;
}
}
}
}
else if (refDCval >= 86) {
if (TSC < 10) {
if (mcTcor < 92) {
return 2;
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
}
}
else if (TSC >= 10) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (Rs < 2188) {
return 1;
}
else if (Rs >= 2188) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (mu_mv_mag_sq < 52) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
if (Cs < 1351) {
if (Rs < 2053) {
return 8;
}
else if (Rs >= 2053) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (Rs < 2206) {
return 1;
}
else if (Rs >= 2206) {
return 8;
}
}
}
}
else if (Cs >= 1351) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (tcor < 78) {
return 2;
}
else if (tcor >= 78) {
if (AFD < 10) {
if (Rs < 2687) {
return 4;
}
else if (Rs >= 2687) {
if (Rs < 2781) {
return 8;
}
else if (Rs >= 2781) {
return 2;
}
}
}
else if (AFD >= 10) {
if (mcTcor < 88) {
if (SC < 4462) {
return 4;
}
else if (SC >= 4462) {
return 8;
}
}
else if (mcTcor >= 88) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2935) {
if (ssDCval < 104) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
if (Cs < 1497) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (Cs >= 1497) {
if (negBalance < 58) {
return 4;
}
else if (negBalance >= 58) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
else if (ssDCval >= 104) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
if (Rs < 3111) {
return 4;
}
else if (Rs >= 3111) {
return 2;
}
}
}
}
}
else if (AFD >= 19) {
if (mcTcor < 95) {
if (RsDiff < 18) {
if (ssDCval < 87) {
if (diffRsCsdiff < -69) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
if (Rs < 2608) {
return 4;
}
else if (Rs >= 2608) {
return 8;
}
}
}
else if (diffRsCsdiff >= -69) {
if (tcor < 58) {
if (ssDCval < 71) {
if (Rs < 1892) {
return 8;
}
else if (Rs >= 1892) {
return 4;
}
}
else if (ssDCval >= 71) {
if (ssDCval < 81) {
return 2;
}
else if (ssDCval >= 81) {
return 8;
}
}
}
else if (tcor >= 58) {
if (Rs < 1921) {
if (posBalance < 55) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (posBalance >= 55) {
if (AFD < 24) {
if (Cs < 1367) {
return 1;
}
else if (Cs >= 1367) {
if (Cs < 1629) {
return 4;
}
else if (Cs >= 1629) {
return 1;
}
}
}
else if (AFD >= 24) {
if (mu_mv_mag_sq < 9) {
if (Cs < 1539) {
return 4;
}
else if (Cs >= 1539) {
return 8;
}
}
else if (mu_mv_mag_sq >= 9) {
return 4;
}
}
}
}
else if (Rs >= 1921) {
if (refDCval < 76) {
if (Rs < 2345) {
if (SC < 3595) {
if (Rs < 2071) {
return 2;
}
else if (Rs >= 2071) {
return 8;
}
}
else if (SC >= 3595) {
if (Rs < 2069) {
return 8;
}
else if (Rs >= 2069) {
return 4;
}
}
}
else if (Rs >= 2345) {
return 2;
}
}
else if (refDCval >= 76) {
if (mu_mv_mag_sq < 13) {
if (TSC < 13) {
if (negBalance < 51) {
if (Rs < 2375) {
return 4;
}
else if (Rs >= 2375) {
return 8;
}
}
else if (negBalance >= 51) {
return 4;
}
}
else if (TSC >= 13) {
if (RsCsDiff < 209) {
return 8;
}
else if (RsCsDiff >= 209) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 13) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (mcTcor < 93) {
if (MVDiff < 11) {
if (RsCsDiff < 190) {
return 8;
}
else if (RsCsDiff >= 190) {
if (diffRsCsdiff < -78) {
return 2;
}
else if (diffRsCsdiff >= -78) {
return 4;
}
}
}
else if (MVDiff >= 11) {
if (posBalance < 63) {
if (MVDiff < 99) {
return 8;
}
else if (MVDiff >= 99) {
return 4;
}
}
else if (posBalance >= 63) {
if (ssDCval < 92) {
if (Rs < 1851) {
return 8;
}
else if (Rs >= 1851) {
return 1;
}
}
else if (ssDCval >= 92) {
return 4;
}
}
}
}
else if (mcTcor >= 93) {
if (Rs < 1963) {
return 2;
}
else if (Rs >= 1963) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2348) {
return 2;
}
else if (Rs >= 2348) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
}
else if (RsDiff >= 18) {
if (CsDiff < 12) {
return 2;
}
else if (CsDiff >= 12) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (Cs < 1378) {
if (Rs < 1759) {
return 4;
}
else if (Rs >= 1759) {
return 1;
}
}
else if (Cs >= 1378) {
if (refDCval < 101) {
return 2;
}
else if (refDCval >= 101) {
if (TSC < 11) {
if (Cs < 1551) {
return 1;
}
else if (Cs >= 1551) {
if (Rs < 1874) {
return 4;
}
else if (Rs >= 1874) {
return 2;
}
}
}
else if (TSC >= 11) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 111) {
if (negBalance < 53) {
if (tcor < 82) {
if (mcTcor < 87) {
if (gchDC < 81) {
if (Cs < 1331) {
return 1;
}
else if (Cs >= 1331) {
return 8;
}
}
else if (gchDC >= 81) {
return 4;
}
}
else if (mcTcor >= 87) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 7) {
if (tcor < 77) {
if (MVDiff < 12) {
if (Rs < 2713) {
return 8;
}
else if (Rs >= 2713) {
return 1;
}
}
else if (MVDiff >= 12) {
return 4;
}
}
else if (tcor >= 77) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (negBalance < 38) {
if (Rs < 2247) {
return 8;
}
else if (Rs >= 2247) {
return 2;
}
}
else if (negBalance >= 38) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (negBalance < 48) {
if (Rs < 3196) {
if (diffAFD < 6) {
return 2;
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (Rs >= 3196) {
return 4;
}
}
else if (negBalance >= 48) {
if (Rs < 1900) {
return 2;
}
else if (Rs >= 1900) {
return 8;
}
}
}
}
}
else if (tcor >= 82) {
if (negBalance < 33) {
if (Rs < 2728) {
return 4;
}
else if (Rs >= 2728) {
return 8;
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
else if (negBalance >= 53) {
if (MVDiff < 21) {
if (posBalance < 61) {
if (diffMVdiffVal < -10) {
if (Cs < 1576) {
return 1;
}
else if (Cs >= 1576) {
return 8;
}
}
else if (diffMVdiffVal >= -10) {
if (diffRsCsdiff < -94) {
return 1;
}
else if (diffRsCsdiff >= -94) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (tcor < 66) {
if (CsDiff < 10) {
if (Rs < 2855) {
return 4;
}
else if (Rs >= 2855) {
return 1;
}
}
else if (CsDiff >= 10) {
return 2;
}
}
else if (tcor >= 66) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
}
else if (MVDiff >= 21) {
if (mu_mv_mag_sq < 14) {
if (CsDiff < 11) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (AFD < 21) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
return 8;
}
}
else if (AFD >= 21) {
if (negBalance < 64) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (negBalance < 58) {
return 1;
}
else if (negBalance >= 58) {
if (CsDiff < 9) {
return 1;
}
else if (CsDiff >= 9) {
return 8;
}
}
}
}
else if (negBalance >= 64) {
return 4;
}
}
}
}
else if (CsDiff >= 11) {
if (diffAFD < 17) {
return 8;
}
else if (diffAFD >= 17) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (mu_mv_mag_sq < 50) {
if (diffMVdiffVal < -99) {
return 4;
}
else if (diffMVdiffVal >= -99) {
if (AFD < 33) {
if (Cs < 1431) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
else if (Cs >= 1431) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
return 8;
}
}
}
else if (AFD >= 33) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 50) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3338) {
if (tcor < 86) {
if (RsDiff < 13) {
if (negBalance < 47) {
if (refDCval < 87) {
if (posBalance < 47) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 2;
}
}
else if (posBalance >= 47) {
if (Cs < 1661) {
return 2;
}
else if (Cs >= 1661) {
return 8;
}
}
}
else if (refDCval >= 87) {
if (ssDCval < 136) {
return 8;
}
else if (ssDCval >= 136) {
return 4;
}
}
}
else if (negBalance >= 47) {
if (Rs < 4853) {
return 8;
}
else if (Rs >= 4853) {
return 1;
}
}
}
else if (RsDiff >= 13) {
if (MVDiff < 46) {
if (mu_mv_mag_sq < 11) {
if (refDCval < 101) {
if (Rs < 3390) {
return 4;
}
else if (Rs >= 3390) {
return 2;
}
}
else if (refDCval >= 101) {
if (TSC < 14) {
if (RsDiff < 16) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
}
else if (RsDiff >= 16) {
if (SC < 6671) {
if (mcTcor < 90) {
if (Rs < 3477) {
return 2;
}
else if (Rs >= 3477) {
return 8;
}
}
else if (mcTcor >= 90) {
return 2;
}
}
else if (SC >= 6671) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
if (MVDiff < 37) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (Cs < 1710) {
if (Cs < 1456) {
if (Rs < 6268) {
return 2;
}
else if (Rs >= 6268) {
return 8;
}
}
else if (Cs >= 1456) {
if (diffRsCsdiff < 251) {
return 2;
}
else if (diffRsCsdiff >= 251) {
return 8;
}
}
}
else if (Cs >= 1710) {
return 8;
}
}
}
else if (MVDiff >= 37) {
return 8;
}
}
}
}
}
else if (TSC >= 14) {
if (Cs < 1549) {
return 1;
}
else if (Cs >= 1549) {
if (Cs < 1790) {
return 2;
}
else if (Cs >= 1790) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (tcor < 52) {
if (AFD < 30) {
if (Rs < 3677) {
return 1;
}
else if (Rs >= 3677) {
return 8;
}
}
else if (AFD >= 30) {
return 2;
}
}
else if (tcor >= 52) {
if (tcor < 71) {
if (Rs < 4208) {
if (ssDCval < 103) {
if (negBalance < 53) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (negBalance >= 53) {
return 4;
}
}
else if (ssDCval >= 103) {
return 2;
}
}
else if (Rs >= 4208) {
if (tcor < 55) {
if (Rs < 5411) {
return 8;
}
else if (Rs >= 5411) {
return 4;
}
}
else if (tcor >= 55) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 4;
}
}
}
}
else if (tcor >= 71) {
if (Cs < 1817) {
if (AFD < 25) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (MVDiff < 34) {
if (CsDiff < 7) {
if (Rs < 3829) {
return 2;
}
else if (Rs >= 3829) {
if (Rs < 4215) {
return 4;
}
else if (Rs >= 4215) {
return 8;
}
}
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (MVDiff >= 34) {
return 8;
}
}
}
else if (AFD >= 25) {
if (Rs < 3756) {
return 8;
}
else if (Rs >= 3756) {
if (Cs < 1681) {
return 1;
}
else if (Cs >= 1681) {
return 4;
}
}
}
}
else if (Cs >= 1817) {
if (Rs < 4172) {
if (TSC < 10) {
if (Rs < 3516) {
return 2;
}
else if (Rs >= 3516) {
return 1;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (Rs >= 4172) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 5576) {
return 2;
}
else if (Rs >= 5576) {
if (Rs < 6043) {
return 1;
}
else if (Rs >= 6043) {
return 4;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 46) {
if (CsDiff < 8) {
if (TSC < 14) {
if (mu_mv_mag_sq < 21) {
return 8;
}
else if (mu_mv_mag_sq >= 21) {
return 4;
}
}
else if (TSC >= 14) {
return 4;
}
}
else if (CsDiff >= 8) {
if (posBalance < 53) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
else if (posBalance >= 53) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 7043) {
if (diffRsCsdiff < -171) {
if (AFD < 41) {
return 2;
}
else if (AFD >= 41) {
return 1;
}
}
else if (diffRsCsdiff >= -171) {
return 1;
}
}
else if (Rs >= 7043) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 86) {
if (CsDiff < 3) {
if (Rs < 4866) {
return 2;
}
else if (Rs >= 4866) {
return 4;
}
}
else if (CsDiff >= 3) {
if (refDCval < 101) {
if (posBalance < 44) {
if (mcTcor < 93) {
if (Cs < 1647) {
if (Rs < 3670) {
return 4;
}
else if (Rs >= 3670) {
return 1;
}
}
else if (Cs >= 1647) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (negBalance < 35) {
if (Cs < 1556) {
return 4;
}
else if (Cs >= 1556) {
return 2;
}
}
else if (negBalance >= 35) {
if (MVDiff < 2) {
if (Rs < 4183) {
return 2;
}
else if (Rs >= 4183) {
return 8;
}
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
else if (posBalance >= 44) {
if (negBalance < 48) {
if (MVDiff < 45) {
if (tcor < 88) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
else if (tcor >= 88) {
return 2;
}
}
else if (MVDiff >= 45) {
return 8;
}
}
else if (negBalance >= 48) {
if (RsDiff < 12) {
if (Cs < 1674) {
return 2;
}
else if (Cs >= 1674) {
return 8;
}
}
else if (RsDiff >= 12) {
if (Rs < 5456) {
return 2;
}
else if (Rs >= 5456) {
return 4;
}
}
}
}
}
else if (refDCval >= 101) {
if (RsDiff < 12) {
if (tcor < 94) {
if (mu_mv_mag_sq < 85) {
if (Rs < 5686) {
return 8;
}
else if (Rs >= 5686) {
return 2;
}
}
else if (mu_mv_mag_sq >= 85) {
return 1;
}
}
else if (tcor >= 94) {
return 2;
}
}
else if (RsDiff >= 12) {
if (MVDiff < 24) {
if (posBalance < 60) {
if (negBalance < 64) {
return 2;
}
else if (negBalance >= 64) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
else if (posBalance >= 60) {
if (TSC < 9) {
if (Rs < 4286) {
return 4;
}
else if (Rs >= 4286) {
return 1;
}
}
else if (TSC >= 9) {
if (Rs < 3830) {
return 8;
}
else if (Rs >= 3830) {
return 2;
}
}
}
}
else if (MVDiff >= 24) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (Rs < 3835) {
return 8;
}
else if (Rs >= 3835) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Cs >= 1921) {
if (posBalance < 51) {
if (refDCval < 77) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
if (Rs < 5825) {
if (mu_mv_mag_sq < 55) {
if (ssDCval < 65) {
if (posBalance < 46) {
if (Rs < 3630) {
return 1;
}
else if (Rs >= 3630) {
return 4;
}
}
else if (posBalance >= 46) {
if (mcTcor < 77) {
return 4;
}
else if (mcTcor >= 77) {
if (negBalance < 59) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (negBalance >= 59) {
if (Cs < 2174) {
return 4;
}
else if (Cs >= 2174) {
return 8;
}
}
}
}
}
else if (ssDCval >= 65) {
if (CsDiff < 15) {
if (gchDC < 7) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
else if (gchDC >= 7) {
return 8;
}
}
else if (CsDiff >= 15) {
if (Cs < 3565) {
return 4;
}
else if (Cs >= 3565) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (Rs < 5473) {
return 2;
}
else if (Rs >= 5473) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (tcor < 87) {
if (diffAFD < 3) {
if (diffTSC < 2) {
if (posBalance < 49) {
if (RsDiff < 23) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
if (Rs < 3779) {
return 8;
}
else if (Rs >= 3779) {
return 4;
}
}
}
else if (RsDiff >= 23) {
return 8;
}
}
else if (posBalance >= 49) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
else if (diffTSC >= 2) {
return 2;
}
}
else if (diffAFD >= 3) {
return 8;
}
}
else if (tcor >= 87) {
if (Rs < 3354) {
if (Rs < 3210) {
return 4;
}
else if (Rs >= 3210) {
return 1;
}
}
else if (Rs >= 3354) {
if (Rs < 5048) {
if (RsCsDiff < 155) {
if (Rs < 3952) {
return 8;
}
else if (Rs >= 3952) {
return 4;
}
}
else if (RsCsDiff >= 155) {
return 8;
}
}
else if (Rs >= 5048) {
if (Rs < 5323) {
return 4;
}
else if (Rs >= 5323) {
return 8;
}
}
}
}
}
}
else if (Rs >= 5825) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (Rs < 7262) {
return 2;
}
else if (Rs >= 7262) {
return 1;
}
}
}
}
}
else if (refDCval >= 77) {
if (tcor < 89) {
if (SC < 8439) {
if (ssDCval < 120) {
if (mcTcor < 88) {
if (ssDCval < 82) {
if (TSC < 22) {
if (refDCval < 82) {
return 2;
}
else if (refDCval >= 82) {
return 8;
}
}
else if (TSC >= 22) {
return 4;
}
}
else if (ssDCval >= 82) {
if (Rs < 1921) {
return 2;
}
else if (Rs >= 1921) {
return 8;
}
}
}
else if (mcTcor >= 88) {
if (Cs < 2565) {
if (TSC < 10) {
if (Rs < 5102) {
if (Rs < 1866) {
return 2;
}
else if (Rs >= 1866) {
return 8;
}
}
else if (Rs >= 5102) {
if (Rs < 5561) {
return 1;
}
else if (Rs >= 5561) {
return 4;
}
}
}
else if (TSC >= 10) {
if (negBalance < 43) {
return 2;
}
else if (negBalance >= 43) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 1;
}
}
}
}
else if (Cs >= 2565) {
if (RsCsDiff < 297) {
if (Rs < 3997) {
return 8;
}
else if (Rs >= 3997) {
return 1;
}
}
else if (RsCsDiff >= 297) {
if (negBalance < 55) {
return 2;
}
else if (negBalance >= 55) {
if (Cs < 3354) {
if (Rs < 3595) {
return 4;
}
else if (Rs >= 3595) {
return 8;
}
}
else if (Cs >= 3354) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 120) {
if (RsCsDiff < 539) {
if (mcTcor < 84) {
if (mu_mv_mag_sq < 37) {
return 1;
}
else if (mu_mv_mag_sq >= 37) {
if (diffMVdiffVal < -6) {
if (Rs < 3534) {
return 2;
}
else if (Rs >= 3534) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
}
else if (mcTcor >= 84) {
if (SC < 4840) {
if (mu_mv_mag_sq < 41) {
if (Rs < 2364) {
return 4;
}
else if (Rs >= 2364) {
return 2;
}
}
else if (mu_mv_mag_sq >= 41) {
if (Rs < 2676) {
return 8;
}
else if (Rs >= 2676) {
return 1;
}
}
}
else if (SC >= 4840) {
if (negBalance < 53) {
if (mu_mv_mag_sq < 73) {
if (ssDCval < 181) {
return 4;
}
else if (ssDCval >= 181) {
return 8;
}
}
else if (mu_mv_mag_sq >= 73) {
return 8;
}
}
else if (negBalance >= 53) {
return 8;
}
}
}
}
else if (RsCsDiff >= 539) {
if (CsDiff < 21) {
if (SC < 8304) {
return 2;
}
else if (SC >= 8304) {
return 8;
}
}
else if (CsDiff >= 21) {
return 8;
}
}
}
}
else if (SC >= 8439) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
if (tcor < 77) {
if (SC < 8965) {
return 8;
}
else if (SC >= 8965) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
if (CsDiff < 20) {
return 2;
}
else if (CsDiff >= 20) {
if (Rs < 7413) {
return 4;
}
else if (Rs >= 7413) {
return 1;
}
}
}
}
}
else if (tcor >= 77) {
if (mu_mv_mag_sq < 74) {
if (Cs < 2975) {
return 8;
}
else if (Cs >= 2975) {
return 1;
}
}
else if (mu_mv_mag_sq >= 74) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
}
}
}
}
else if (tcor >= 89) {
if (negBalance < 44) {
if (ssDCval < 113) {
if (mu_mv_mag_sq < 50) {
if (RsDiff < 11) {
if (Rs < 4538) {
return 8;
}
else if (Rs >= 4538) {
return 2;
}
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (mu_mv_mag_sq >= 50) {
if (posBalance < 32) {
if (Cs < 3385) {
return 2;
}
else if (Cs >= 3385) {
return 1;
}
}
else if (posBalance >= 32) {
return 2;
}
}
}
else if (ssDCval >= 113) {
if (refDCval < 116) {
return 1;
}
else if (refDCval >= 116) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (ssDCval < 124) {
if (Rs < 4535) {
return 2;
}
else if (Rs >= 4535) {
if (Rs < 7622) {
return 8;
}
else if (Rs >= 7622) {
return 1;
}
}
}
else if (ssDCval >= 124) {
return 8;
}
}
}
}
}
else if (negBalance >= 44) {
if (Cs < 2657) {
if (negBalance < 50) {
if (Rs < 3319) {
return 8;
}
else if (Rs >= 3319) {
return 2;
}
}
else if (negBalance >= 50) {
if (SC < 6473) {
return 8;
}
else if (SC >= 6473) {
if (Cs < 2435) {
return 8;
}
else if (Cs >= 2435) {
return 4;
}
}
}
}
else if (Cs >= 2657) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
if (Cs < 3204) {
return 4;
}
else if (Cs >= 3204) {
return 2;
}
}
}
}
}
}
}
}
else if (posBalance >= 51) {
if (refDCval < 154) {
if (mu_mv_mag_sq < 26) {
if (AFD < 23) {
if (negBalance < 43) {
if (MVDiff < 6) {
if (Rs < 5523) {
return 2;
}
else if (Rs >= 5523) {
return 4;
}
}
else if (MVDiff >= 6) {
if (negBalance < 33) {
return 2;
}
else if (negBalance >= 33) {
if (Rs < 8221) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
if (MVDiff < 10) {
if (Rs < 4652) {
if (Rs < 3550) {
return 8;
}
else if (Rs >= 3550) {
return 2;
}
}
else if (Rs >= 4652) {
if (Rs < 5507) {
return 1;
}
else if (Rs >= 5507) {
return 4;
}
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
else if (Rs >= 8221) {
return 1;
}
}
}
}
else if (negBalance >= 43) {
if (ssDCval < 74) {
if (RsCsDiff < 201) {
if (negBalance < 57) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 8;
}
}
else if (negBalance >= 57) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1872) {
return 2;
}
else if (Rs >= 1872) {
return 8;
}
}
}
}
else if (RsCsDiff >= 201) {
if (refDCval < 63) {
if (negBalance < 57) {
if (Cs < 2064) {
return 2;
}
else if (Cs >= 2064) {
return 4;
}
}
else if (negBalance >= 57) {
return 2;
}
}
else if (refDCval >= 63) {
if (Rs < 2730) {
if (mcTcor < 83) {
if (TSC < 13) {
if (Rs < 2284) {
return 2;
}
else if (Rs >= 2284) {
return 8;
}
}
else if (TSC >= 13) {
return 4;
}
}
else if (mcTcor >= 83) {
return 1;
}
}
else if (Rs >= 2730) {
if (TSC < 15) {
if (Rs < 3631) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
if (Rs < 2751) {
return 2;
}
else if (Rs >= 2751) {
return 8;
}
}
}
else if (Rs >= 3631) {
return 2;
}
}
else if (TSC >= 15) {
return 8;
}
}
}
}
}
else if (ssDCval >= 74) {
if (MVDiff < 7) {
if (tcor < 86) {
if (Rs < 4293) {
if (ssDCval < 87) {
if (Cs < 2293) {
if (Rs < 2737) {
return 2;
}
else if (Rs >= 2737) {
return 4;
}
}
else if (Cs >= 2293) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
if (Rs < 3215) {
return 1;
}
else if (Rs >= 3215) {
return 8;
}
}
}
}
else if (ssDCval >= 87) {
if (refDCval < 91) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (diffAFD < -1) {
if (AFD < 18) {
if (Rs < 4175) {
return 4;
}
else if (Rs >= 4175) {
return 8;
}
}
else if (AFD >= 18) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (refDCval >= 91) {
if (Rs < 2372) {
return 8;
}
else if (Rs >= 2372) {
if (ssDCval < 102) {
if (Cs < 4282) {
return 4;
}
else if (Cs >= 4282) {
return 8;
}
}
else if (ssDCval >= 102) {
if (Rs < 3756) {
if (ssDCval < 108) {
return 4;
}
else if (ssDCval >= 108) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
if (Rs < 3077) {
return 1;
}
else if (Rs >= 3077) {
return 4;
}
}
}
}
else if (Rs >= 3756) {
return 8;
}
}
}
}
}
}
else if (Rs >= 4293) {
if (TSC < 15) {
if (mcTcor < 94) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (negBalance < 57) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (tcor < 82) {
return 8;
}
else if (tcor >= 82) {
return 4;
}
}
}
else if (negBalance >= 57) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
if (Rs < 4731) {
return 8;
}
else if (Rs >= 4731) {
return 4;
}
}
}
}
else if (TSC >= 15) {
if (Rs < 6313) {
return 2;
}
else if (Rs >= 6313) {
return 4;
}
}
}
}
else if (tcor >= 86) {
if (Cs < 2129) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (Cs >= 2129) {
if (Rs < 4775) {
if (TSC < 11) {
if (Cs < 2273) {
return 2;
}
else if (Cs >= 2273) {
if (Cs < 3197) {
return 4;
}
else if (Cs >= 3197) {
return 2;
}
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (Rs >= 4775) {
return 4;
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 96) {
if (refDCval < 84) {
if (tcor < 59) {
return 4;
}
else if (tcor >= 59) {
if (TSC < 13) {
if (posBalance < 65) {
if (Rs < 2766) {
return 8;
}
else if (Rs >= 2766) {
if (SC < 5390) {
if (Cs < 2330) {
return 4;
}
else if (Cs >= 2330) {
return 8;
}
}
else if (SC >= 5390) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 8;
}
}
}
}
else if (posBalance >= 65) {
return 4;
}
}
else if (TSC >= 13) {
if (MVDiff < 20) {
if (Rs < 3407) {
if (AFD < 21) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 1;
}
}
else if (AFD >= 21) {
if (TSC < 14) {
return 1;
}
else if (TSC >= 14) {
return 8;
}
}
}
else if (Rs >= 3407) {
return 1;
}
}
else if (MVDiff >= 20) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 84) {
if (mcTcor < 88) {
if (Cs < 3838) {
if (AFD < 17) {
if (Cs < 2094) {
return 4;
}
else if (Cs >= 2094) {
if (Cs < 2375) {
return 8;
}
else if (Cs >= 2375) {
return 2;
}
}
}
else if (AFD >= 17) {
if (AFD < 22) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 3021) {
return 8;
}
else if (Rs >= 3021) {
return 4;
}
}
}
else if (AFD >= 22) {
if (Rs < 3208) {
return 1;
}
else if (Rs >= 3208) {
return 4;
}
}
}
}
else if (Cs >= 3838) {
return 8;
}
}
else if (mcTcor >= 88) {
if (mcTcor < 94) {
if (mu_mv_mag_sq < 2) {
if (SC < 6693) {
if (TSC < 14) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
if (Cs < 2076) {
return 8;
}
else if (Cs >= 2076) {
if (Rs < 3164) {
return 2;
}
else if (Rs >= 3164) {
return 4;
}
}
}
}
else if (TSC >= 14) {
return 1;
}
}
else if (SC >= 6693) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (negBalance < 59) {
if (mcTcor < 89) {
if (Rs < 3674) {
return 4;
}
else if (Rs >= 3674) {
return 8;
}
}
else if (mcTcor >= 89) {
return 8;
}
}
else if (negBalance >= 59) {
if (Rs < 2582) {
return 1;
}
else if (Rs >= 2582) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (CsDiff < 12) {
if (Cs < 1959) {
return 8;
}
else if (Cs >= 1959) {
return 2;
}
}
else if (CsDiff >= 12) {
if (Rs < 2512) {
return 4;
}
else if (Rs >= 2512) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 96) {
if (RsDiff < 20) {
if (AFD < 20) {
if (RsDiff < 16) {
if (negBalance < 57) {
if (Rs < 1588) {
return 4;
}
else if (Rs >= 1588) {
if (RsDiff < 15) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (ssDCval < 109) {
if (Rs < 3018) {
return 4;
}
else if (Rs >= 3018) {
if (Rs < 3977) {
return 2;
}
else if (Rs >= 3977) {
return 8;
}
}
}
else if (ssDCval >= 109) {
return 8;
}
}
}
}
else if (RsDiff >= 15) {
if (Cs < 2307) {
return 2;
}
else if (Cs >= 2307) {
return 8;
}
}
}
}
else if (negBalance >= 57) {
if (Rs < 2782) {
return 8;
}
else if (Rs >= 2782) {
return 4;
}
}
}
else if (RsDiff >= 16) {
if (Cs < 2370) {
return 4;
}
else if (Cs >= 2370) {
return 8;
}
}
}
else if (AFD >= 20) {
if (RsDiff < 14) {
if (Rs < 2875) {
if (diffAFD < 1) {
if (TSC < 11) {
if (Rs < 2230) {
return 4;
}
else if (Rs >= 2230) {
return 1;
}
}
else if (TSC >= 11) {
return 2;
}
}
else if (diffAFD >= 1) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
return 2;
}
}
}
else if (Rs >= 2875) {
if (tcor < 74) {
if (Cs < 2664) {
return 1;
}
else if (Cs >= 2664) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 8;
}
}
}
}
else if (tcor >= 74) {
return 8;
}
}
}
else if (RsDiff >= 14) {
if (mcTcor < 95) {
if (diffRsCsdiff < 108) {
return 8;
}
else if (diffRsCsdiff >= 108) {
if (Cs < 2468) {
return 2;
}
else if (Cs >= 2468) {
return 8;
}
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
else if (RsDiff >= 20) {
if (ssDCval < 101) {
return 2;
}
else if (ssDCval >= 101) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 4;
}
}
}
}
}
}
}
}
else if (AFD >= 23) {
if (SC < 10305) {
if (negBalance < 56) {
if (posBalance < 63) {
if (AFD < 30) {
if (TSC < 16) {
if (posBalance < 57) {
if (RsCsDiff < 647) {
return 8;
}
else if (RsCsDiff >= 647) {
return 2;
}
}
else if (posBalance >= 57) {
if (MVDiff < 23) {
if (mu_mv_mag_sq < 7) {
if (RsCsDiff < 269) {
return 8;
}
else if (RsCsDiff >= 269) {
if (ssDCval < 95) {
return 4;
}
else if (ssDCval >= 95) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 4670) {
return 4;
}
else if (Cs >= 4670) {
return 2;
}
}
}
else if (MVDiff >= 23) {
if (negBalance < 54) {
return 2;
}
else if (negBalance >= 54) {
if (Rs < 4610) {
if (Rs < 2718) {
return 8;
}
else if (Rs >= 2718) {
return 4;
}
}
else if (Rs >= 4610) {
return 2;
}
}
}
}
}
else if (TSC >= 16) {
if (Cs < 3483) {
if (ssDCval < 128) {
return 1;
}
else if (ssDCval >= 128) {
return 8;
}
}
else if (Cs >= 3483) {
if (Rs < 5015) {
return 4;
}
else if (Rs >= 5015) {
return 2;
}
}
}
}
else if (AFD >= 30) {
return 2;
}
}
else if (posBalance >= 63) {
if (diffAFD < 3) {
if (RsDiff < 18) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
if (SC < 5173) {
if (AFD < 26) {
if (Cs < 2054) {
return 8;
}
else if (Cs >= 2054) {
return 2;
}
}
else if (AFD >= 26) {
if (RsDiff < 16) {
return 8;
}
else if (RsDiff >= 16) {
return 2;
}
}
}
else if (SC >= 5173) {
return 8;
}
}
}
else if (RsDiff >= 18) {
if (Rs < 4302) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 1;
}
}
else if (Rs >= 4302) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
return 4;
}
}
}
}
else if (diffAFD >= 3) {
if (diffRsCsdiff < 126) {
if (AFD < 34) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 4;
}
}
else if (AFD >= 34) {
return 8;
}
}
else if (diffRsCsdiff >= 126) {
return 8;
}
}
}
}
else if (negBalance >= 56) {
if (TSC < 12) {
if (ssDCval < 100) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
if (Rs < 3591) {
return 1;
}
else if (Rs >= 3591) {
return 8;
}
}
}
else if (ssDCval >= 100) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Cs < 2089) {
return 2;
}
else if (Cs >= 2089) {
return 1;
}
}
}
}
else if (TSC >= 12) {
if (MVDiff < 23) {
if (diffRsCsdiff < -48) {
if (negBalance < 58) {
if (Cs < 2609) {
return 8;
}
else if (Cs >= 2609) {
return 1;
}
}
else if (negBalance >= 58) {
return 8;
}
}
else if (diffRsCsdiff >= -48) {
if (diffAFD < 2) {
if (RsDiff < 10) {
if (Rs < 2611) {
return 2;
}
else if (Rs >= 2611) {
return 8;
}
}
else if (RsDiff >= 10) {
if (Cs < 2183) {
if (diffRsCsdiff < 69) {
if (Rs < 3023) {
return 1;
}
else if (Rs >= 3023) {
return 8;
}
}
else if (diffRsCsdiff >= 69) {
return 2;
}
}
else if (Cs >= 2183) {
if (Rs < 4666) {
return 1;
}
else if (Rs >= 4666) {
return 8;
}
}
}
}
else if (diffAFD >= 2) {
if (CsDiff < 15) {
return 8;
}
else if (CsDiff >= 15) {
return 4;
}
}
}
}
else if (MVDiff >= 23) {
if (refDCval < 89) {
if (tcor < 53) {
if (Rs < 1537) {
return 8;
}
else if (Rs >= 1537) {
return 1;
}
}
else if (tcor >= 53) {
if (RsCsDiff < 355) {
if (diffRsCsdiff < 13) {
return 8;
}
else if (diffRsCsdiff >= 13) {
return 4;
}
}
else if (RsCsDiff >= 355) {
return 8;
}
}
}
else if (refDCval >= 89) {
if (RsCsDiff < 693) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
return 8;
}
}
else if (RsCsDiff >= 693) {
if (negBalance < 58) {
return 4;
}
else if (negBalance >= 58) {
if (Cs < 2796) {
if (Rs < 4368) {
return 2;
}
else if (Rs >= 4368) {
return 1;
}
}
else if (Cs >= 2796) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 10305) {
if (TSC < 10) {
if (Cs < 6088) {
return 1;
}
else if (Cs >= 6088) {
return 8;
}
}
else if (TSC >= 10) {
if (AFD < 41) {
if (ssDCval < 83) {
if (Rs < 6653) {
return 8;
}
else if (Rs >= 6653) {
return 1;
}
}
else if (ssDCval >= 83) {
if (Rs < 5357) {
return 8;
}
else if (Rs >= 5357) {
if (Cs < 3708) {
return 2;
}
else if (Cs >= 3708) {
return 4;
}
}
}
}
else if (AFD >= 41) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (Cs < 3478) {
if (refDCval < 137) {
if (Cs < 2601) {
if (tcor < 91) {
if (RsCsDiff < 777) {
if (ssDCval < 53) {
if (refDCval < 52) {
return 4;
}
else if (refDCval >= 52) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (ssDCval >= 53) {
if (RsCsDiff < 183) {
if (Rs < 4091) {
if (Rs < 2800) {
if (SC < 4291) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (TSC < 13) {
return 1;
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (SC >= 4291) {
return 8;
}
}
else if (Rs >= 2800) {
if (AFD < 14) {
if (diffRsCsdiff < 1) {
if (negBalance < 52) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (Cs < 2151) {
if (Rs < 3865) {
return 4;
}
else if (Rs >= 3865) {
return 8;
}
}
else if (Cs >= 2151) {
return 8;
}
}
}
else if (negBalance >= 52) {
return 4;
}
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
else if (AFD >= 14) {
if (Cs < 2264) {
if (mu_mv_mag_sq < 58) {
if (Rs < 3192) {
return 4;
}
else if (Rs >= 3192) {
return 8;
}
}
else if (mu_mv_mag_sq >= 58) {
return 4;
}
}
else if (Cs >= 2264) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
}
}
else if (Rs >= 4091) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (mu_mv_mag_sq < 78) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
else if (mu_mv_mag_sq >= 78) {
return 1;
}
}
}
}
else if (RsCsDiff >= 183) {
if (mcTcor < 74) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
else if (mcTcor >= 74) {
if (mcTcor < 85) {
if (TSC < 14) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
return 2;
}
}
}
else if (TSC >= 14) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
if (Rs < 3028) {
return 8;
}
else if (Rs >= 3028) {
return 4;
}
}
}
}
}
else if (mcTcor >= 85) {
if (mu_mv_mag_sq < 52) {
if (ssDCval < 73) {
if (Rs < 3980) {
return 4;
}
else if (Rs >= 3980) {
return 2;
}
}
else if (ssDCval >= 73) {
if (AFD < 17) {
if (Rs < 4707) {
if (ssDCval < 127) {
return 8;
}
else if (ssDCval >= 127) {
if (Rs < 2468) {
return 8;
}
else if (Rs >= 2468) {
return 4;
}
}
}
else if (Rs >= 4707) {
return 2;
}
}
else if (AFD >= 17) {
if (diffMVdiffVal < 2) {
if (mcTcor < 91) {
if (ssDCval < 76) {
if (Rs < 2938) {
return 2;
}
else if (Rs >= 2938) {
return 4;
}
}
else if (ssDCval >= 76) {
if (Rs < 3319) {
return 8;
}
else if (Rs >= 3319) {
if (Rs < 3468) {
return 2;
}
else if (Rs >= 3468) {
return 8;
}
}
}
}
else if (mcTcor >= 91) {
if (Rs < 2903) {
if (ssDCval < 96) {
return 4;
}
else if (ssDCval >= 96) {
return 2;
}
}
else if (Rs >= 2903) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
if (mu_mv_mag_sq < 30) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 1;
}
}
else if (mu_mv_mag_sq >= 30) {
if (diffMVdiffVal < 11) {
return 8;
}
else if (diffMVdiffVal >= 11) {
if (Cs < 2196) {
return 8;
}
else if (Cs >= 2196) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (RsDiff < 11) {
if (negBalance < 50) {
if (Rs < 2840) {
return 8;
}
else if (Rs >= 2840) {
return 4;
}
}
else if (negBalance >= 50) {
return 1;
}
}
else if (RsDiff >= 11) {
if (refDCval < 88) {
if (RsDiff < 14) {
if (RsCsDiff < 217) {
return 8;
}
else if (RsCsDiff >= 217) {
return 4;
}
}
else if (RsDiff >= 14) {
return 8;
}
}
else if (refDCval >= 88) {
if (ssDCval < 104) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
if (Rs < 3724) {
return 8;
}
else if (Rs >= 3724) {
return 4;
}
}
}
else if (ssDCval >= 104) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 777) {
if (AFD < 30) {
return 4;
}
else if (AFD >= 30) {
return 1;
}
}
}
else if (tcor >= 91) {
if (ssDCval < 104) {
if (Rs < 3667) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (Rs >= 3667) {
if (Cs < 2434) {
return 2;
}
else if (Cs >= 2434) {
return 4;
}
}
}
else if (ssDCval >= 104) {
if (Rs < 5791) {
return 8;
}
else if (Rs >= 5791) {
return 2;
}
}
}
}
else if (Cs >= 2601) {
if (RsDiff < 14) {
if (SC < 6967) {
if (ssDCval < 82) {
if (negBalance < 43) {
return 4;
}
else if (negBalance >= 43) {
if (tcor < 69) {
return 4;
}
else if (tcor >= 69) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
else if (ssDCval >= 82) {
if (refDCval < 107) {
if (diffRsCsdiff < -74) {
return 8;
}
else if (diffRsCsdiff >= -74) {
if (Cs < 2678) {
if (Rs < 2989) {
return 8;
}
else if (Rs >= 2989) {
if (refDCval < 99) {
return 4;
}
else if (refDCval >= 99) {
if (Rs < 3158) {
if (Rs < 3078) {
return 4;
}
else if (Rs >= 3078) {
return 8;
}
}
else if (Rs >= 3158) {
return 2;
}
}
}
}
else if (Cs >= 2678) {
if (mu_mv_mag_sq < 69) {
if (Rs < 3489) {
return 4;
}
else if (Rs >= 3489) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 69) {
return 4;
}
}
}
}
else if (refDCval >= 107) {
if (AFD < 14) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
else if (AFD >= 14) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
return 8;
}
}
}
}
}
}
else if (SC >= 6967) {
if (Rs < 5598) {
if (negBalance < 48) {
if (ssDCval < 99) {
if (Cs < 2824) {
return 8;
}
else if (Cs >= 2824) {
return 4;
}
}
else if (ssDCval >= 99) {
if (Cs < 3318) {
return 8;
}
else if (Cs >= 3318) {
if (RsCsDiff < 243) {
if (Cs < 3441) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (Cs >= 3441) {
return 8;
}
}
else if (RsCsDiff >= 243) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
return 8;
}
}
else if (Rs >= 5598) {
if (Cs < 3324) {
return 4;
}
else if (Cs >= 3324) {
return 8;
}
}
}
}
else if (RsDiff >= 14) {
if (mu_mv_mag_sq < 53) {
if (mcTcor < 71) {
if (MVDiff < 26) {
if (Rs < 4540) {
return 4;
}
else if (Rs >= 4540) {
return 8;
}
}
else if (MVDiff >= 26) {
return 2;
}
}
else if (mcTcor >= 71) {
if (Rs < 4666) {
if (refDCval < 118) {
if (diffMVdiffVal < -4) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 4;
}
}
else if (diffMVdiffVal >= -4) {
if (CsDiff < 11) {
return 4;
}
else if (CsDiff >= 11) {
return 8;
}
}
}
else if (refDCval >= 118) {
return 4;
}
}
else if (Rs >= 4666) {
if (negBalance < 55) {
if (CsDiff < 11) {
return 4;
}
else if (CsDiff >= 11) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 2;
}
}
}
else if (negBalance >= 55) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (refDCval < 82) {
if (RsDiff < 19) {
return 8;
}
else if (RsDiff >= 19) {
return 4;
}
}
else if (refDCval >= 82) {
if (negBalance < 45) {
if (Rs < 4951) {
return 8;
}
else if (Rs >= 4951) {
if (Rs < 7455) {
return 4;
}
else if (Rs >= 7455) {
return 1;
}
}
}
else if (negBalance >= 45) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 137) {
if (Rs < 3560) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (Cs < 2701) {
if (Rs < 2556) {
return 4;
}
else if (Rs >= 2556) {
return 8;
}
}
else if (Cs >= 2701) {
return 2;
}
}
}
else if (Rs >= 3560) {
if (RsDiff < 16) {
if (negBalance < 43) {
if (Rs < 3832) {
return 1;
}
else if (Rs >= 3832) {
return 8;
}
}
else if (negBalance >= 43) {
return 1;
}
}
else if (RsDiff >= 16) {
if (ssDCval < 139) {
if (Cs < 2080) {
return 8;
}
else if (Cs >= 2080) {
return 2;
}
}
else if (ssDCval >= 139) {
return 4;
}
}
}
}
}
else if (Cs >= 3478) {
if (Rs < 5822) {
if (ssDCval < 132) {
if (Cs < 3743) {
if (diffAFD < -1) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
if (Rs < 4648) {
if (ssDCval < 83) {
if (Rs < 3484) {
return 4;
}
else if (Rs >= 3484) {
return 8;
}
}
else if (ssDCval >= 83) {
return 4;
}
}
else if (Rs >= 4648) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (RsDiff < 17) {
if (refDCval < 123) {
if (diffRsCsdiff < 36) {
return 8;
}
else if (diffRsCsdiff >= 36) {
if (Cs < 3717) {
if (Cs < 3516) {
if (Cs < 3507) {
return 8;
}
else if (Cs >= 3507) {
return 4;
}
}
else if (Cs >= 3516) {
return 8;
}
}
else if (Cs >= 3717) {
return 4;
}
}
}
else if (refDCval >= 123) {
if (Rs < 3752) {
return 8;
}
else if (Rs >= 3752) {
return 4;
}
}
}
else if (RsDiff >= 17) {
if (Cs < 3696) {
return 4;
}
else if (Cs >= 3696) {
if (Rs < 4440) {
return 8;
}
else if (Rs >= 4440) {
return 2;
}
}
}
}
}
else if (Cs >= 3743) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (Rs < 3465) {
if (mcTcor < 82) {
return 4;
}
else if (mcTcor >= 82) {
return 8;
}
}
else if (Rs >= 3465) {
if (TSC < 13) {
if (Rs < 4198) {
if (mu_mv_mag_sq < 72) {
return 2;
}
else if (mu_mv_mag_sq >= 72) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (Rs >= 4198) {
return 8;
}
}
else if (TSC >= 13) {
return 8;
}
}
}
}
}
else if (ssDCval >= 132) {
if (tcor < 92) {
if (tcor < 69) {
if (Rs < 4156) {
if (Cs < 3912) {
return 8;
}
else if (Cs >= 3912) {
return 1;
}
}
else if (Rs >= 4156) {
return 1;
}
}
else if (tcor >= 69) {
return 1;
}
}
else if (tcor >= 92) {
return 8;
}
}
}
else if (Rs >= 5822) {
if (TSC < 21) {
if (negBalance < 40) {
if (Rs < 8236) {
return 2;
}
else if (Rs >= 8236) {
return 1;
}
}
else if (negBalance >= 40) {
if (TSC < 12) {
if (SC < 10112) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
return 4;
}
}
else if (SC >= 10112) {
return 8;
}
}
else if (TSC >= 12) {
if (MVDiff < 2) {
if (ssDCval < 97) {
return 4;
}
else if (ssDCval >= 97) {
return 8;
}
}
else if (MVDiff >= 2) {
if (mcTcor < 80) {
if (Cs < 4424) {
return 2;
}
else if (Cs >= 4424) {
return 8;
}
}
else if (mcTcor >= 80) {
if (RsCsDiff < 563) {
if (Rs < 6117) {
return 2;
}
else if (Rs >= 6117) {
return 8;
}
}
else if (RsCsDiff >= 563) {
return 4;
}
}
}
}
}
}
else if (TSC >= 21) {
return 8;
}
}
}
}
}
else if (refDCval >= 154) {
if (ssDCval < 168) {
if (Cs < 2216) {
if (ssDCval < 159) {
return 8;
}
else if (ssDCval >= 159) {
return 4;
}
}
else if (Cs >= 2216) {
return 4;
}
}
else if (ssDCval >= 168) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1640) {
if (Rs < 3139) {
if (Rs < 1454) {
if (ssDCval < 122) {
if (negBalance < 1) {
return 2;
}
else if (negBalance >= 1) {
if (refDCval < 63) {
if (AFD < 3) {
if (mu_mv_mag_sq < 128) {
if (Cs < 1414) {
if (ssDCval < 55) {
if (Rs < 693) {
return 1;
}
else if (Rs >= 693) {
if (Rs < 1154) {
return 8;
}
else if (Rs >= 1154) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 4;
}
}
}
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (Cs >= 1414) {
if (negBalance < 27) {
return 1;
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (SC < 2852) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (Rs < 1162) {
return 8;
}
else if (Rs >= 1162) {
return 4;
}
}
}
else if (SC >= 2852) {
return 1;
}
}
}
else if (AFD >= 3) {
if (SC < 2569) {
return 8;
}
else if (SC >= 2569) {
if (tcor < 98) {
if (MVDiff < 3) {
if (SC < 2963) {
return 8;
}
else if (SC >= 2963) {
return 4;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (tcor >= 98) {
return 4;
}
}
}
}
else if (refDCval >= 63) {
if (ssDCval < 77) {
if (diffRsCsdiff < -4) {
return 4;
}
else if (diffRsCsdiff >= -4) {
if (diffMVdiffVal < -2) {
if (Rs < 1270) {
return 4;
}
else if (Rs >= 1270) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (posBalance < 33) {
if (Rs < 1417) {
return 8;
}
else if (Rs >= 1417) {
return 1;
}
}
else if (posBalance >= 33) {
return 8;
}
}
}
}
else if (ssDCval >= 77) {
if (Rs < 953) {
return 1;
}
else if (Rs >= 953) {
return 8;
}
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 2) {
if (posBalance < 28) {
if (Cs < 1348) {
if (Rs < 1327) {
return 1;
}
else if (Rs >= 1327) {
return 8;
}
}
else if (Cs >= 1348) {
return 8;
}
}
else if (posBalance >= 28) {
if (Rs < 1383) {
return 4;
}
else if (Rs >= 1383) {
return 8;
}
}
}
else if (AFD >= 2) {
if (mcTcor < 96) {
if (Rs < 933) {
if (Rs < 577) {
return 2;
}
else if (Rs >= 577) {
return 8;
}
}
else if (Rs >= 933) {
if (AFD < 4) {
if (Rs < 1296) {
return 8;
}
else if (Rs >= 1296) {
return 1;
}
}
else if (AFD >= 4) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 121) {
return 8;
}
else if (mu_mv_mag_sq >= 121) {
if (Rs < 1009) {
return 8;
}
else if (Rs >= 1009) {
if (Rs < 1270) {
return 4;
}
else if (Rs >= 1270) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1454) {
if (ssDCval < 154) {
if (ssDCval < 86) {
if (mcTcor < 98) {
if (Rs < 2218) {
if (gchDC < 2) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
if (Rs < 2033) {
return 1;
}
else if (Rs >= 2033) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
else if (Rs >= 2218) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
if (Cs < 1447) {
if (Rs < 2853) {
return 8;
}
else if (Rs >= 2853) {
return 2;
}
}
else if (Cs >= 1447) {
if (tcor < 97) {
if (Rs < 2752) {
return 8;
}
else if (Rs >= 2752) {
return 8;
}
}
else if (tcor >= 97) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 18) {
if (SC < 4462) {
if (Cs < 1344) {
if (SC < 3612) {
return 8;
}
else if (SC >= 3612) {
if (SC < 3616) {
return 2;
}
else if (SC >= 3616) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
if (Rs < 2297) {
return 2;
}
else if (Rs >= 2297) {
if (Cs < 1342) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (posBalance < 31) {
if (Rs < 2335) {
return 8;
}
else if (Rs >= 2335) {
return 2;
}
}
else if (posBalance >= 31) {
if (posBalance < 35) {
if (negBalance < 13) {
if (Rs < 2313) {
return 8;
}
else if (Rs >= 2313) {
return 2;
}
}
else if (negBalance >= 13) {
return 2;
}
}
else if (posBalance >= 35) {
return 8;
}
}
}
}
else if (Cs >= 1342) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1344) {
if (SC < 3298) {
if (Cs < 1441) {
return 8;
}
else if (Cs >= 1441) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (SC >= 3298) {
return 8;
}
}
}
else if (SC >= 4462) {
return 1;
}
}
else if (negBalance >= 18) {
if (Rs < 1476) {
if (Cs < 1565) {
if (Rs < 1471) {
return 8;
}
else if (Rs >= 1471) {
return 4;
}
}
else if (Cs >= 1565) {
return 1;
}
}
else if (Rs >= 1476) {
if (negBalance < 22) {
if (ssDCval < 84) {
if (Rs < 1601) {
return 8;
}
else if (Rs >= 1601) {
return 4;
}
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (negBalance >= 22) {
if (negBalance < 35) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (MVDiff < 3) {
if (Rs < 1512) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (Rs >= 1512) {
if (Cs < 1614) {
return 4;
}
else if (Cs >= 1614) {
if (Cs < 1618) {
if (Rs < 2453) {
if (Rs < 2034) {
return 4;
}
else if (Rs >= 2034) {
return 8;
}
}
else if (Rs >= 2453) {
return 4;
}
}
else if (Cs >= 1618) {
return 4;
}
}
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (negBalance >= 35) {
if (Rs < 1757) {
return 4;
}
else if (Rs >= 1757) {
if (posBalance < 36) {
return 8;
}
else if (posBalance >= 36) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (TSC < 3) {
if (diffAFD < -20) {
if (Cs < 1392) {
return 4;
}
else if (Cs >= 1392) {
if (diffAFD < -22) {
return 8;
}
else if (diffAFD >= -22) {
return 4;
}
}
}
else if (diffAFD >= -20) {
if (Rs < 3053) {
return 8;
}
else if (Rs >= 3053) {
if (Cs < 1511) {
return 2;
}
else if (Cs >= 1511) {
return 8;
}
}
}
}
else if (TSC >= 3) {
if (ssDCval < 117) {
if (diffAFD < -1) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Cs < 1509) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Cs >= 1509) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (ssDCval >= 117) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (Rs < 1661) {
return 8;
}
else if (Rs >= 1661) {
if (diffAFD < 1) {
if (refDCval < 117) {
return 8;
}
else if (refDCval >= 117) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
if (Cs < 1620) {
return 1;
}
else if (Cs >= 1620) {
if (Rs < 2406) {
return 8;
}
else if (Rs >= 2406) {
return 1;
}
}
}
}
}
else if (diffAFD >= 1) {
if (Cs < 1435) {
return 8;
}
else if (Cs >= 1435) {
if (Rs < 2985) {
return 1;
}
else if (Rs >= 2985) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 154) {
if (Rs < 1725) {
return 8;
}
else if (Rs >= 1725) {
return 2;
}
}
}
}
else if (Rs >= 3139) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 74) {
if (ssDCval < 63) {
if (negBalance < 20) {
if (Rs < 4111) {
return 4;
}
else if (Rs >= 4111) {
if (Cs < 1544) {
return 8;
}
else if (Cs >= 1544) {
return 4;
}
}
}
else if (negBalance >= 20) {
return 8;
}
}
else if (ssDCval >= 63) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (Rs < 4304) {
return 2;
}
else if (Rs >= 4304) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1568) {
if (Rs < 3660) {
if (Cs < 1326) {
return 8;
}
else if (Cs >= 1326) {
if (Rs < 3532) {
return 1;
}
else if (Rs >= 3532) {
return 4;
}
}
}
else if (Rs >= 3660) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (Cs >= 1568) {
if (Rs < 3265) {
return 8;
}
else if (Rs >= 3265) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 104) {
if (posBalance < 42) {
if (negBalance < 28) {
if (Rs < 4493) {
if (mu_mv_mag_sq < 113) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (mu_mv_mag_sq >= 113) {
return 4;
}
}
else if (Rs >= 4493) {
return 2;
}
}
else if (negBalance >= 28) {
if (Rs < 3762) {
return 4;
}
else if (Rs >= 3762) {
if (Rs < 4719) {
return 8;
}
else if (Rs >= 4719) {
return 2;
}
}
}
}
else if (posBalance >= 42) {
if (refDCval < 91) {
return 2;
}
else if (refDCval >= 91) {
if (Cs < 1534) {
return 4;
}
else if (Cs >= 1534) {
if (ssDCval < 101) {
return 8;
}
else if (ssDCval >= 101) {
if (Rs < 3872) {
return 2;
}
else if (Rs >= 3872) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 104) {
if (mcTcor < 93) {
if (SC < 4798) {
return 2;
}
else if (SC >= 4798) {
return 4;
}
}
else if (mcTcor >= 93) {
return 2;
}
}
}
}
}
else if (Cs >= 1640) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (diffRsCsdiff < -5) {
if (tcor < 67) {
return 4;
}
else if (tcor >= 67) {
if (Cs < 1862) {
if (Cs < 1853) {
return 8;
}
else if (Cs >= 1853) {
return 4;
}
}
else if (Cs >= 1862) {
return 8;
}
}
}
else if (diffRsCsdiff >= -5) {
if (refDCval < 65) {
if (ssDCval < 38) {
return 8;
}
else if (ssDCval >= 38) {
if (Rs < 6330) {
return 4;
}
else if (Rs >= 6330) {
return 2;
}
}
}
else if (refDCval >= 65) {
if (Cs < 1925) {
return 8;
}
else if (Cs >= 1925) {
if (ssDCval < 93) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (ssDCval >= 93) {
if (Rs < 3035) {
return 4;
}
else if (Rs >= 3035) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 3319) {
if (negBalance < 19) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 3541) {
if (negBalance < 16) {
if (Cs < 1829) {
if (negBalance < 14) {
if (Rs < 2557) {
return 8;
}
else if (Rs >= 2557) {
return 1;
}
}
else if (negBalance >= 14) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 8;
}
}
}
else if (Cs >= 1829) {
return 1;
}
}
else if (negBalance >= 16) {
if (TSCindex < 2) {
if (posBalance < 35) {
if (Rs < 1923) {
return 8;
}
else if (Rs >= 1923) {
return 1;
}
}
else if (posBalance >= 35) {
if (ssDCval < 64) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
return 4;
}
}
}
else if (ssDCval >= 64) {
return 8;
}
}
}
else if (TSCindex >= 2) {
if (Rs < 2374) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (Rs >= 2374) {
return 1;
}
}
}
}
else if (Rs >= 3541) {
return 4;
}
}
}
else if (negBalance >= 19) {
if (Rs < 1662) {
if (ssDCval < 57) {
if (Rs < 880) {
return 1;
}
else if (Rs >= 880) {
return 4;
}
}
else if (ssDCval >= 57) {
if (negBalance < 27) {
if (mcTcor < 98) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (mcTcor >= 98) {
if (Cs < 1746) {
return 8;
}
else if (Cs >= 1746) {
return 1;
}
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (Rs >= 1662) {
if (refDCval < 66) {
if (posBalance < 43) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
if (Rs < 2247) {
return 8;
}
else if (Rs >= 2247) {
return 1;
}
}
}
else if (posBalance >= 43) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 2203) {
return 8;
}
else if (Rs >= 2203) {
return 1;
}
}
}
}
else if (refDCval >= 66) {
if (Rs < 3450) {
return 1;
}
else if (Rs >= 3450) {
return 8;
}
}
}
}
}
else if (Cs >= 3319) {
return 8;
}
}
else if (ssDCval >= 67) {
if (Rs < 2389) {
if (ssDCval < 90) {
if (negBalance < 35) {
if (Cs < 1644) {
if (Rs < 1107) {
return 8;
}
else if (Rs >= 1107) {
return 4;
}
}
else if (Cs >= 1644) {
return 8;
}
}
else if (negBalance >= 35) {
if (Rs < 1546) {
return 1;
}
else if (Rs >= 1546) {
return 8;
}
}
}
else if (ssDCval >= 90) {
if (refDCval < 92) {
if (Rs < 2370) {
return 1;
}
else if (Rs >= 2370) {
return 8;
}
}
else if (refDCval >= 92) {
if (mu_mv_mag_sq < 119) {
if (Rs < 2320) {
return 2;
}
else if (Rs >= 2320) {
if (Cs < 2117) {
return 4;
}
else if (Cs >= 2117) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (mu_mv_mag_sq < 127) {
if (Rs < 2339) {
return 4;
}
else if (Rs >= 2339) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 2324) {
return 8;
}
else if (Rs >= 2324) {
return 4;
}
}
}
}
}
}
else if (Rs >= 2389) {
if (SC < 6966) {
if (refDCval < 89) {
if (refDCval < 84) {
if (diffRsCsdiff < -40) {
return 4;
}
else if (diffRsCsdiff >= -40) {
return 8;
}
}
else if (refDCval >= 84) {
if (mu_mv_mag_sq < 125) {
if (Cs < 2163) {
return 2;
}
else if (Cs >= 2163) {
if (Rs < 3013) {
return 1;
}
else if (Rs >= 3013) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
}
else if (refDCval >= 89) {
return 8;
}
}
else if (SC >= 6966) {
if (SC < 7503) {
if (SC < 7163) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (SC >= 7163) {
return 4;
}
}
else if (SC >= 7503) {
if (Cs < 2357) {
return 4;
}
else if (Cs >= 2357) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6769) {
if (SC < 8342) {
if (ssDCval < 81) {
if (negBalance < 34) {
if (RsCsDiff < 15) {
if (SC < 5212) {
if (Rs < 2868) {
if (ssDCval < 80) {
if (mu_mv_mag_sq < 113) {
if (Rs < 1152) {
return 8;
}
else if (Rs >= 1152) {
return 4;
}
}
else if (mu_mv_mag_sq >= 113) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
if (Cs < 1751) {
return 1;
}
else if (Cs >= 1751) {
return 8;
}
}
}
}
}
else if (ssDCval >= 80) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
return 4;
}
}
}
else if (Rs >= 2868) {
if (ssDCval < 71) {
if (Rs < 2944) {
return 4;
}
else if (Rs >= 2944) {
if (SC < 4734) {
return 8;
}
else if (SC >= 4734) {
if (negBalance < 25) {
if (Cs < 1734) {
return 4;
}
else if (Cs >= 1734) {
if (Cs < 2009) {
if (mu_mv_mag_sq < 127) {
return 4;
}
else if (mu_mv_mag_sq >= 127) {
if (CsDiff < 3) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (Rs < 2983) {
if (Rs < 2972) {
if (Rs < 2956) {
return 8;
}
else if (Rs >= 2956) {
return 4;
}
}
else if (Rs >= 2972) {
return 8;
}
}
else if (Rs >= 2983) {
if (diffRsCsdiff < -3) {
return 4;
}
else if (diffRsCsdiff >= -3) {
if (posBalance < 40) {
if (Rs < 3027) {
return 4;
}
else if (Rs >= 3027) {
if (Rs < 3142) {
if (Rs < 3084) {
return 8;
}
else if (Rs >= 3084) {
return 8;
}
}
else if (Rs >= 3142) {
return 4;
}
}
}
else if (posBalance >= 40) {
if (Rs < 3032) {
if (Rs < 2991) {
return 4;
}
else if (Rs >= 2991) {
return 4;
}
}
else if (Rs >= 3032) {
return 4;
}
}
}
}
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (Cs >= 2009) {
return 4;
}
}
}
else if (negBalance >= 25) {
if (posBalance < 36) {
return 8;
}
else if (posBalance >= 36) {
return 4;
}
}
}
}
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (SC >= 5212) {
if (refDCval < 60) {
if (Rs < 3676) {
return 8;
}
else if (Rs >= 3676) {
if (Cs < 2321) {
return 8;
}
else if (Cs >= 2321) {
return 4;
}
}
}
else if (refDCval >= 60) {
return 8;
}
}
}
else if (RsCsDiff >= 15) {
if (Rs < 3762) {
if (SC < 5325) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
if (ssDCval < 63) {
if (Rs < 1349) {
return 4;
}
else if (Rs >= 1349) {
if (Cs < 2805) {
return 8;
}
else if (Cs >= 2805) {
return 4;
}
}
}
else if (ssDCval >= 63) {
if (Rs < 1998) {
return 8;
}
else if (Rs >= 1998) {
if (refDCval < 76) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (diffRsCsdiff < 11) {
return 4;
}
else if (diffRsCsdiff >= 11) {
return 8;
}
}
}
else if (refDCval >= 76) {
if (Rs < 2863) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Rs >= 2863) {
return 8;
}
}
}
}
}
}
else if (SC >= 5325) {
return 8;
}
}
else if (Rs >= 3762) {
if (Rs < 3843) {
return 2;
}
else if (Rs >= 3843) {
if (diffMVdiffVal < 2) {
if (ssDCval < 72) {
if (SC < 7908) {
if (negBalance < 29) {
if (mcTcor < 95) {
if (Rs < 4360) {
return 8;
}
else if (Rs >= 4360) {
if (Rs < 4898) {
return 4;
}
else if (Rs >= 4898) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (tcor < 93) {
if (posBalance < 39) {
if (diffRsCsdiff < -10) {
return 8;
}
else if (diffRsCsdiff >= -10) {
if (Rs < 4768) {
return 4;
}
else if (Rs >= 4768) {
return 4;
}
}
}
else if (posBalance >= 39) {
return 8;
}
}
else if (tcor >= 93) {
if (Cs < 2585) {
if (Rs < 4820) {
if (ssDCval < 54) {
if (Rs < 4466) {
if (posBalance < 28) {
return 4;
}
else if (posBalance >= 28) {
if (mcTcor < 97) {
if (Rs < 4363) {
if (Rs < 4271) {
if (Rs < 4197) {
return 4;
}
else if (Rs >= 4197) {
return 4;
}
}
else if (Rs >= 4271) {
return 4;
}
}
else if (Rs >= 4363) {
return 4;
}
}
else if (mcTcor >= 97) {
if (posBalance < 35) {
return 4;
}
else if (posBalance >= 35) {
return 4;
}
}
}
}
else if (Rs >= 4466) {
if (Rs < 4639) {
return 4;
}
else if (Rs >= 4639) {
return 8;
}
}
}
else if (ssDCval >= 54) {
if (diffRsCsdiff < 8) {
if (negBalance < 14) {
if (MVDiff < 2) {
if (Rs < 4331) {
return 8;
}
else if (Rs >= 4331) {
return 4;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (negBalance >= 14) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
return 8;
}
}
}
else if (diffRsCsdiff >= 8) {
if (Rs < 4368) {
return 4;
}
else if (Rs >= 4368) {
if (Rs < 4421) {
return 4;
}
else if (Rs >= 4421) {
return 4;
}
}
}
}
}
else if (Rs >= 4820) {
if (negBalance < 15) {
return 4;
}
else if (negBalance >= 15) {
if (Rs < 5189) {
if (Rs < 4909) {
return 4;
}
else if (Rs >= 4909) {
if (Rs < 5025) {
return 4;
}
else if (Rs >= 5025) {
return 4;
}
}
}
else if (Rs >= 5189) {
return 4;
}
}
}
}
else if (Cs >= 2585) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (TSC < 3) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (Rs < 4677) {
return 8;
}
else if (Rs >= 4677) {
if (Rs < 4892) {
return 8;
}
else if (Rs >= 4892) {
return 4;
}
}
}
}
else if (TSC >= 3) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 29) {
return 4;
}
}
else if (SC >= 7908) {
return 8;
}
}
else if (ssDCval >= 72) {
return 4;
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 4513) {
return 4;
}
else if (Rs >= 4513) {
if (Rs < 4717) {
if (Rs < 4660) {
return 8;
}
else if (Rs >= 4660) {
return 8;
}
}
else if (Rs >= 4717) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 34) {
if (posBalance < 49) {
if (RsDiff < 8) {
if (ssDCval < 72) {
if (ssDCval < 69) {
if (Cs < 1684) {
return 4;
}
else if (Cs >= 1684) {
if (Cs < 3110) {
return 8;
}
else if (Cs >= 3110) {
if (Rs < 3081) {
return 4;
}
else if (Rs >= 3081) {
return 8;
}
}
}
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (posBalance >= 49) {
if (Cs < 3523) {
if (SC < 5742) {
if (Cs < 2281) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (Cs >= 2281) {
if (Rs < 2566) {
if (negBalance < 44) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
return 4;
}
}
else if (negBalance >= 44) {
return 8;
}
}
else if (Rs >= 2566) {
if (Cs < 2935) {
return 8;
}
else if (Cs >= 2935) {
return 4;
}
}
}
}
else if (SC >= 5742) {
return 8;
}
}
else if (Cs >= 3523) {
if (Rs < 3379) {
return 4;
}
else if (Rs >= 3379) {
return 8;
}
}
}
}
}
else if (ssDCval >= 81) {
if (Cs < 2622) {
if (Rs < 3759) {
if (tcor < 98) {
if (negBalance < 49) {
if (refDCval < 81) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 2900) {
if (Rs < 2738) {
return 8;
}
else if (Rs >= 2738) {
return 4;
}
}
else if (Rs >= 2900) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (Cs < 1859) {
if (Rs < 2795) {
return 8;
}
else if (Rs >= 2795) {
if (Rs < 3003) {
return 4;
}
else if (Rs >= 3003) {
return 2;
}
}
}
else if (Cs >= 1859) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
if (Rs < 2964) {
return 8;
}
else if (Rs >= 2964) {
return 4;
}
}
}
}
}
else if (negBalance >= 49) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (tcor >= 98) {
if (Cs < 1889) {
if (MVDiff < 1) {
if (Rs < 2437) {
if (Rs < 1118) {
return 1;
}
else if (Rs >= 1118) {
return 8;
}
}
else if (Rs >= 2437) {
if (Rs < 3485) {
return 4;
}
else if (Rs >= 3485) {
return 2;
}
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (Cs >= 1889) {
if (posBalance < 43) {
if (refDCval < 90) {
if (Rs < 3026) {
if (Cs < 2044) {
return 8;
}
else if (Cs >= 2044) {
return 4;
}
}
else if (Rs >= 3026) {
if (Cs < 2326) {
return 2;
}
else if (Cs >= 2326) {
return 8;
}
}
}
else if (refDCval >= 90) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (posBalance >= 43) {
if (ssDCval < 82) {
if (Rs < 3116) {
return 4;
}
else if (Rs >= 3116) {
return 8;
}
}
else if (ssDCval >= 82) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
}
}
else if (Rs >= 3759) {
if (RsCsDiff < 63) {
if (negBalance < 29) {
if (Cs < 2314) {
if (diffTSC < -1) {
if (tcor < 97) {
if (Cs < 1768) {
return 4;
}
else if (Cs >= 1768) {
return 8;
}
}
else if (tcor >= 97) {
if (Rs < 4309) {
return 2;
}
else if (Rs >= 4309) {
return 4;
}
}
}
else if (diffTSC >= -1) {
if (Rs < 3850) {
return 2;
}
else if (Rs >= 3850) {
if (Rs < 4471) {
return 4;
}
else if (Rs >= 4471) {
if (Rs < 4767) {
return 2;
}
else if (Rs >= 4767) {
return 4;
}
}
}
}
}
else if (Cs >= 2314) {
if (Rs < 4784) {
return 8;
}
else if (Rs >= 4784) {
return 4;
}
}
}
else if (negBalance >= 29) {
if (SC < 6374) {
return 8;
}
else if (SC >= 6374) {
if (AFD < 5) {
if (Rs < 4685) {
return 2;
}
else if (Rs >= 4685) {
return 8;
}
}
else if (AFD >= 5) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 63) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
else if (Cs >= 2622) {
if (ssDCval < 98) {
if (ssDCval < 89) {
if (Cs < 3214) {
if (negBalance < 39) {
return 4;
}
else if (negBalance >= 39) {
if (mu_mv_mag_sq < 115) {
return 4;
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
}
else if (Cs >= 3214) {
if (posBalance < 52) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (posBalance >= 52) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
if (Cs < 3218) {
if (Rs < 3645) {
return 8;
}
else if (Rs >= 3645) {
return 4;
}
}
else if (Cs >= 3218) {
return 8;
}
}
}
}
}
else if (ssDCval >= 89) {
if (Rs < 2614) {
return 4;
}
else if (Rs >= 2614) {
if (Rs < 3331) {
if (negBalance < 50) {
if (mu_mv_mag_sq < 115) {
if (Rs < 3163) {
return 8;
}
else if (Rs >= 3163) {
return 4;
}
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
else if (negBalance >= 50) {
if (Cs < 3377) {
return 4;
}
else if (Cs >= 3377) {
if (Cs < 3931) {
return 8;
}
else if (Cs >= 3931) {
return 4;
}
}
}
}
else if (Rs >= 3331) {
if (mu_mv_mag_sq < 123) {
return 4;
}
else if (mu_mv_mag_sq >= 123) {
if (SC < 7360) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (tcor < 99) {
if (Cs < 3543) {
if (posBalance < 52) {
if (Cs < 3223) {
return 8;
}
else if (Cs >= 3223) {
return 4;
}
}
else if (posBalance >= 52) {
if (Cs < 2707) {
return 8;
}
else if (Cs >= 2707) {
if (Rs < 3393) {
if (Rs < 3370) {
return 4;
}
else if (Rs >= 3370) {
return 8;
}
}
else if (Rs >= 3393) {
return 4;
}
}
}
}
else if (Cs >= 3543) {
return 8;
}
}
else if (tcor >= 99) {
return 8;
}
}
}
else if (SC >= 7360) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 98) {
if (Rs < 3310) {
if (Rs < 2998) {
return 4;
}
else if (Rs >= 2998) {
if (Cs < 3164) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (Cs >= 3164) {
return 8;
}
}
}
else if (Rs >= 3310) {
return 8;
}
}
}
}
}
else if (SC >= 8342) {
if (negBalance < 30) {
if (mu_mv_mag_sq < 116) {
if (Rs < 5853) {
return 8;
}
else if (Rs >= 5853) {
return 2;
}
}
else if (mu_mv_mag_sq >= 116) {
if (Rs < 4456) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
else if (Rs >= 4456) {
if (SC < 10849) {
return 8;
}
else if (SC >= 10849) {
if (Rs < 6059) {
return 8;
}
else if (Rs >= 6059) {
return 2;
}
}
}
}
}
else if (negBalance >= 30) {
if (Cs < 3666) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
return 4;
}
}
else if (Cs >= 3666) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 5972) {
return 8;
}
else if (Rs >= 5972) {
return 2;
}
}
}
}
}
}
else if (Rs >= 6769) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (diffRsCsdiff < 48) {
return 4;
}
else if (diffRsCsdiff >= 48) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (ssDCval < 129) {
if (Rs < 1893) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (tcor < 99) {
if (Cs < 2024) {
return 8;
}
else if (Cs >= 2024) {
if (Rs < 657) {
return 8;
}
else if (Rs >= 657) {
return 2;
}
}
}
else if (tcor >= 99) {
if (Cs < 1691) {
return 8;
}
else if (Cs >= 1691) {
return 2;
}
}
}
}
else if (Rs >= 1893) {
if (refDCval < 105) {
if (negBalance < 41) {
if (Cs < 2353) {
if (Rs < 3929) {
if (Rs < 2109) {
return 2;
}
else if (Rs >= 2109) {
return 8;
}
}
else if (Rs >= 3929) {
if (Rs < 4502) {
return 4;
}
else if (Rs >= 4502) {
return 2;
}
}
}
else if (Cs >= 2353) {
return 8;
}
}
else if (negBalance >= 41) {
if (Cs < 2865) {
if (Rs < 4923) {
if (Rs < 3556) {
return 8;
}
else if (Rs >= 3556) {
return 1;
}
}
else if (Rs >= 4923) {
return 2;
}
}
else if (Cs >= 2865) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (diffAFD < -2) {
if (diffRsCsdiff < -999) {
return 1;
}
else if (diffRsCsdiff >= -999) {
if (posBalance < 45) {
if (mu_mv_mag_sq < 127) {
if (Cs < 1848) {
if (Rs < 2071) {
return 8;
}
else if (Rs >= 2071) {
return 1;
}
}
else if (Cs >= 1848) {
if (MVDiff < 2) {
if (Rs < 3196) {
return 2;
}
else if (Rs >= 3196) {
return 8;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
else if (posBalance >= 45) {
if (tcor < 100) {
if (posBalance < 57) {
if (ssDCval < 108) {
if (Rs < 4836) {
return 4;
}
else if (Rs >= 4836) {
return 8;
}
}
else if (ssDCval >= 108) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
if (Rs < 5045) {
return 4;
}
else if (Rs >= 5045) {
return 8;
}
}
}
}
else if (posBalance >= 57) {
if (Rs < 3600) {
if (Rs < 2477) {
return 8;
}
else if (Rs >= 2477) {
if (Rs < 3340) {
return 1;
}
else if (Rs >= 3340) {
return 2;
}
}
}
else if (Rs >= 3600) {
return 8;
}
}
}
else if (tcor >= 100) {
return 1;
}
}
}
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 124) {
if (posBalance < 20) {
if (Cs < 2966) {
if (posBalance < 17) {
if (MVDiff < 2) {
if (Cs < 2861) {
return 8;
}
else if (Cs >= 2861) {
return 4;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (posBalance >= 17) {
if (Cs < 2928) {
return 4;
}
else if (Cs >= 2928) {
return 8;
}
}
}
else if (Cs >= 2966) {
if (Rs < 3610) {
return 2;
}
else if (Rs >= 3610) {
return 8;
}
}
}
else if (posBalance >= 20) {
if (Rs < 3299) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 2697) {
return 8;
}
else if (Rs >= 2697) {
return 1;
}
}
}
else if (Rs >= 3299) {
if (Cs < 1942) {
if (Rs < 4323) {
return 4;
}
else if (Rs >= 4323) {
return 2;
}
}
else if (Cs >= 1942) {
if (Rs < 3380) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
else if (Rs >= 3380) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
if (Rs < 6363) {
if (negBalance < 43) {
return 8;
}
else if (negBalance >= 43) {
return 4;
}
}
else if (Rs >= 6363) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (AFD < 28) {
if (Cs < 1679) {
if (Rs < 4132) {
return 8;
}
else if (Rs >= 4132) {
return 2;
}
}
else if (Cs >= 1679) {
if (Rs < 1910) {
if (Rs < 1906) {
return 8;
}
else if (Rs >= 1906) {
if (Cs < 2003) {
return 2;
}
else if (Cs >= 2003) {
return 8;
}
}
}
else if (Rs >= 1910) {
return 8;
}
}
}
else if (AFD >= 28) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 129) {
if (mcTcor < 96) {
if (negBalance < 42) {
if (refDCval < 199) {
if (RsCsDiff < 56) {
return 8;
}
else if (RsCsDiff >= 56) {
if (RsCsDiff < 91) {
return 1;
}
else if (RsCsDiff >= 91) {
if (Rs < 580) {
return 1;
}
else if (Rs >= 580) {
return 8;
}
}
}
}
else if (refDCval >= 199) {
return 4;
}
}
else if (negBalance >= 42) {
if (RsCsDiff < 191) {
if (Rs < 2060) {
return 8;
}
else if (Rs >= 2060) {
return 1;
}
}
else if (RsCsDiff >= 191) {
if (Rs < 4111) {
return 2;
}
else if (Rs >= 4111) {
if (Rs < 8232) {
return 8;
}
else if (Rs >= 8232) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (ssDCval < 144) {
if (SC < 6333) {
if (mcTcor < 98) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Rs < 1728) {
return 4;
}
else if (Rs >= 1728) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 2635) {
return 8;
}
else if (Rs >= 2635) {
if (negBalance < 4) {
return 8;
}
else if (negBalance >= 4) {
if (Rs < 2640) {
return 1;
}
else if (Rs >= 2640) {
if (Cs < 1694) {
return 1;
}
else if (Cs >= 1694) {
return 8;
}
}
}
}
}
}
else if (SC >= 6333) {
if (mu_mv_mag_sq < 125) {
if (Rs < 5607) {
if (Cs < 1978) {
return 2;
}
else if (Cs >= 1978) {
return 8;
}
}
else if (Rs >= 5607) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 4279) {
if (Cs < 2590) {
if (Rs < 4591) {
return 8;
}
else if (Rs >= 4591) {
return 2;
}
}
else if (Cs >= 2590) {
if (posBalance < 15) {
if (Rs < 3022) {
return 1;
}
else if (Rs >= 3022) {
return 8;
}
}
else if (posBalance >= 15) {
return 1;
}
}
}
else if (Cs >= 4279) {
return 8;
}
}
}
}
else if (ssDCval >= 144) {
if (ssDCval < 200) {
return 8;
}
else if (ssDCval >= 200) {
if (Rs < 2688) {
return 8;
}
else if (Rs >= 2688) {
return 4;
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelectF(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1309) {
if (Rs < 2501) {
if (Cs < 533) {
if (negBalance < 14) {
if (mu_mv_mag_sq < 116) {
if (posBalance < 50) {
if (MVDiff < 9) {
if (negBalance < 9) {
if (ssDCval < 56) {
if (Rs < 7) {
if (ssDCval < 36) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
return 4;
}
}
else if (ssDCval >= 36) {
return 2;
}
}
else if (Rs >= 7) {
if (Cs < 188) {
if (ssDCval < 26) {
if (Rs < 57) {
return 8;
}
else if (Rs >= 57) {
if (Rs < 112) {
return 1;
}
else if (Rs >= 112) {
return 4;
}
}
}
else if (ssDCval >= 26) {
return 8;
}
}
else if (Cs >= 188) {
if (mu_mv_mag_sq < 111) {
if (Rs < 130) {
return 1;
}
else if (Rs >= 130) {
return 4;
}
}
else if (mu_mv_mag_sq >= 111) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
else if (ssDCval >= 56) {
if (Rs < 542) {
if (posBalance < 29) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (posBalance >= 29) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (Rs < 477) {
return 1;
}
else if (Rs >= 477) {
return 8;
}
}
}
}
else if (Rs >= 542) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
return 1;
}
}
}
}
else if (negBalance >= 9) {
if (SC < 919) {
if (ssDCval < 93) {
if (ssDCval < 40) {
if (posBalance < 35) {
if (Cs < 154) {
if (Rs < 90) {
if (ssDCval < 24) {
return 1;
}
else if (ssDCval >= 24) {
if (ssDCval < 31) {
return 4;
}
else if (ssDCval >= 31) {
return 8;
}
}
}
else if (Rs >= 90) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
return 8;
}
}
}
else if (Cs >= 154) {
if (Rs < 171) {
return 1;
}
else if (Rs >= 171) {
if (Cs < 204) {
return 1;
}
else if (Cs >= 204) {
if (posBalance < 21) {
return 1;
}
else if (posBalance >= 21) {
return 4;
}
}
}
}
}
else if (posBalance >= 35) {
if (ssDCval < 32) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
return 8;
}
}
else if (ssDCval >= 32) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
return 4;
}
}
}
}
else if (ssDCval >= 40) {
if (posBalance < 42) {
if (negBalance < 10) {
return 1;
}
else if (negBalance >= 10) {
if (Cs < 205) {
if (MVDiff < 2) {
if (Cs < 203) {
return 8;
}
else if (Cs >= 203) {
return 2;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Cs >= 205) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
return 8;
}
}
}
}
else if (posBalance >= 42) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (Rs < 199) {
return 2;
}
else if (Rs >= 199) {
return 8;
}
}
}
}
}
else if (ssDCval >= 93) {
if (Rs < 467) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 8;
}
}
else if (Rs >= 467) {
return 1;
}
}
}
else if (SC >= 919) {
if (ssDCval < 43) {
if (Rs < 640) {
return 1;
}
else if (Rs >= 640) {
return 2;
}
}
else if (ssDCval >= 43) {
if (mu_mv_mag_sq < 113) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Rs < 717) {
return 1;
}
else if (Rs >= 717) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 113) {
return 1;
}
}
}
}
}
else if (MVDiff >= 9) {
if (tcor < 94) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (ssDCval < 40) {
if (SC < 520) {
if (diffMVdiffVal < -17) {
return 8;
}
else if (diffMVdiffVal >= -17) {
return 1;
}
}
else if (SC >= 520) {
if (tcor < 90) {
return 1;
}
else if (tcor >= 90) {
return 4;
}
}
}
else if (ssDCval >= 40) {
if (Rs < 428) {
return 2;
}
else if (Rs >= 428) {
return 8;
}
}
}
}
else if (tcor >= 94) {
if (ssDCval < 22) {
if (Rs < 15) {
return 8;
}
else if (Rs >= 15) {
return 1;
}
}
else if (ssDCval >= 22) {
if (SC < 943) {
if (Cs < 124) {
if (MVDiff < 25) {
if (negBalance < 6) {
return 4;
}
else if (negBalance >= 6) {
if (SC < 189) {
return 1;
}
else if (SC >= 189) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
return 4;
}
}
}
}
else if (MVDiff >= 25) {
if (ssDCval < 149) {
return 4;
}
else if (ssDCval >= 149) {
return 2;
}
}
}
else if (Cs >= 124) {
if (ssDCval < 132) {
if (mu_mv_mag_sq < 98) {
return 4;
}
else if (mu_mv_mag_sq >= 98) {
if (MVDiff < 33) {
if (diffMVdiffVal < 4) {
return 4;
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
else if (MVDiff >= 33) {
return 8;
}
}
}
else if (ssDCval >= 132) {
if (Rs < 446) {
return 1;
}
else if (Rs >= 446) {
return 4;
}
}
}
}
else if (SC >= 943) {
return 8;
}
}
}
}
}
else if (posBalance >= 50) {
if (RsCsDiff < 1) {
if (MVDiff < 37) {
if (Rs < 12) {
if (mcTcor < 87) {
if (mu_mv_mag_sq < 102) {
return 8;
}
else if (mu_mv_mag_sq >= 102) {
return 4;
}
}
else if (mcTcor >= 87) {
if (AFD < 1) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (AFD >= 1) {
return 2;
}
}
}
else if (Rs >= 12) {
if (negBalance < 6) {
if (refDCval < 55) {
if (mu_mv_mag_sq < 112) {
return 8;
}
else if (mu_mv_mag_sq >= 112) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (refDCval >= 55) {
if (Rs < 128) {
return 2;
}
else if (Rs >= 128) {
return 8;
}
}
}
else if (negBalance >= 6) {
if (MVDiff < 7) {
if (Cs < 77) {
return 8;
}
else if (Cs >= 77) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 85) {
if (posBalance < 67) {
if (negBalance < 8) {
return 8;
}
else if (negBalance >= 8) {
return 1;
}
}
else if (posBalance >= 67) {
if (posBalance < 98) {
if (mcTcor < 77) {
if (Rs < 69) {
return 8;
}
else if (Rs >= 69) {
return 2;
}
}
else if (mcTcor >= 77) {
return 8;
}
}
else if (posBalance >= 98) {
return 1;
}
}
}
else if (ssDCval >= 85) {
if (Rs < 335) {
if (Rs < 29) {
return 1;
}
else if (Rs >= 29) {
return 4;
}
}
else if (Rs >= 335) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 37) {
if (tcor < 75) {
if (diffMVdiffVal < 315) {
if (diffMVdiffVal < -71) {
return 8;
}
else if (diffMVdiffVal >= -71) {
if (Rs < 3) {
if (tcor < 56) {
return 8;
}
else if (tcor >= 56) {
if (mu_mv_mag_sq < 60) {
return 2;
}
else if (mu_mv_mag_sq >= 60) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
}
}
else if (Rs >= 3) {
return 2;
}
}
}
else if (diffMVdiffVal >= 315) {
return 1;
}
}
else if (tcor >= 75) {
if (Cs < 104) {
if (AFD < 4) {
if (posBalance < 86) {
if (Rs < 78) {
return 2;
}
else if (Rs >= 78) {
if (Rs < 497) {
return 1;
}
else if (Rs >= 497) {
return 4;
}
}
}
else if (posBalance >= 86) {
return 4;
}
}
else if (AFD >= 4) {
if (MVDiff < 102) {
if (mu_mv_mag_sq < 37) {
if (Rs < 36) {
return 2;
}
else if (Rs >= 36) {
return 4;
}
}
else if (mu_mv_mag_sq >= 37) {
return 8;
}
}
else if (MVDiff >= 102) {
return 1;
}
}
}
else if (Cs >= 104) {
if (MVDiff < 574) {
return 8;
}
else if (MVDiff >= 574) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (posBalance < 98) {
if (MVDiff < 53) {
if (Cs < 96) {
return 8;
}
else if (Cs >= 96) {
if (mcTcor < 96) {
if (diffAFD < 2) {
if (Rs < 540) {
if (Cs < 209) {
return 1;
}
else if (Cs >= 209) {
return 8;
}
}
else if (Rs >= 540) {
return 8;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (mcTcor >= 96) {
if (Rs < 2052) {
return 1;
}
else if (Rs >= 2052) {
return 2;
}
}
}
}
else if (MVDiff >= 53) {
if (gchDC < 10) {
if (MVDiff < 477) {
return 2;
}
else if (MVDiff >= 477) {
return 1;
}
}
else if (gchDC >= 10) {
return 1;
}
}
}
else if (posBalance >= 98) {
if (MVDiff < 228) {
if (RsDiff < 3) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (tcor < 76) {
if (Cs < 201) {
return 1;
}
else if (Cs >= 201) {
return 4;
}
}
else if (tcor >= 76) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (refDCval < 46) {
if (Cs < 369) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (Cs >= 369) {
return 4;
}
}
else if (refDCval >= 46) {
return 2;
}
}
}
else if (MVDiff >= 228) {
if (mu_mv_mag_sq < 4) {
return 2;
}
else if (mu_mv_mag_sq >= 4) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 116) {
if (SC < 702) {
if (refDCval < 47) {
if (Cs < 105) {
if (ssDCval < 38) {
if (refDCval < 22) {
if (MVDiff < 1076) {
if (ssDCval < 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (refDCval >= 22) {
if (Rs < 4) {
return 4;
}
else if (Rs >= 4) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (ssDCval < 28) {
if (Rs < 16) {
return 4;
}
else if (Rs >= 16) {
return 8;
}
}
else if (ssDCval >= 28) {
return 4;
}
}
}
}
}
else if (ssDCval >= 38) {
if (Cs < 98) {
if (mu_mv_mag_sq < 117) {
if (Rs < 44) {
return 8;
}
else if (Rs >= 44) {
return 1;
}
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
else if (Cs >= 98) {
return 1;
}
}
}
else if (Cs >= 105) {
if (Rs < 203) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
if (SC < 424) {
if (ssDCval < 38) {
if (Cs < 155) {
return 8;
}
else if (Cs >= 155) {
return 4;
}
}
else if (ssDCval >= 38) {
return 1;
}
}
else if (SC >= 424) {
return 1;
}
}
}
else if (Rs >= 203) {
if (ssDCval < 39) {
if (SC < 541) {
return 8;
}
else if (SC >= 541) {
if (mu_mv_mag_sq < 128) {
return 4;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 380) {
if (Rs < 280) {
return 8;
}
else if (Rs >= 280) {
if (ssDCval < 23) {
return 8;
}
else if (ssDCval >= 23) {
return 4;
}
}
}
else if (Rs >= 380) {
return 8;
}
}
}
}
else if (ssDCval >= 39) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
if (Cs < 314) {
return 1;
}
else if (Cs >= 314) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 47) {
if (Rs < 25) {
return 2;
}
else if (Rs >= 25) {
if (ssDCval < 64) {
if (Cs < 283) {
if (ssDCval < 56) {
if (Rs < 130) {
return 8;
}
else if (Rs >= 130) {
if (Rs < 238) {
if (Rs < 134) {
return 4;
}
else if (Rs >= 134) {
return 8;
}
}
else if (Rs >= 238) {
return 4;
}
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (Cs >= 283) {
if (Rs < 251) {
return 4;
}
else if (Rs >= 251) {
if (Rs < 341) {
return 1;
}
else if (Rs >= 341) {
return 8;
}
}
}
}
else if (ssDCval >= 64) {
if (ssDCval < 80) {
if (Rs < 79) {
return 4;
}
else if (Rs >= 79) {
if (Cs < 242) {
return 2;
}
else if (Cs >= 242) {
if (SC < 639) {
return 4;
}
else if (SC >= 639) {
return 8;
}
}
}
}
else if (ssDCval >= 80) {
if (ssDCval < 88) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
return 1;
}
}
else if (ssDCval >= 88) {
if (Cs < 156) {
return 8;
}
else if (Cs >= 156) {
if (ssDCval < 101) {
return 2;
}
else if (ssDCval >= 101) {
if (ssDCval < 130) {
if (Rs < 251) {
return 4;
}
else if (Rs >= 251) {
if (Rs < 327) {
return 8;
}
else if (Rs >= 327) {
return 2;
}
}
}
else if (ssDCval >= 130) {
return 4;
}
}
}
}
}
}
}
}
}
else if (SC >= 702) {
if (diffTSC < -3) {
if (SC < 1309) {
if (ssDCval < 159) {
if (diffTSC < -5) {
if (Rs < 583) {
if (Rs < 467) {
return 8;
}
else if (Rs >= 467) {
return 1;
}
}
else if (Rs >= 583) {
return 4;
}
}
else if (diffTSC >= -5) {
if (diffAFD < -6) {
return 2;
}
else if (diffAFD >= -6) {
if (Cs < 352) {
return 2;
}
else if (Cs >= 352) {
if (negBalance < 4) {
return 4;
}
else if (negBalance >= 4) {
return 2;
}
}
}
}
}
else if (ssDCval >= 159) {
return 8;
}
}
else if (SC >= 1309) {
if (diffAFD < -8) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (diffAFD >= -8) {
return 1;
}
}
}
else if (diffTSC >= -3) {
if (mu_mv_mag_sq < 128) {
if (Rs < 711) {
if (Rs < 391) {
return 1;
}
else if (Rs >= 391) {
if (RsCsDiff < 1) {
if (Rs < 474) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
return 1;
}
}
else if (Rs >= 474) {
if (Rs < 592) {
if (ssDCval < 63) {
if (Cs < 319) {
return 8;
}
else if (Cs >= 319) {
return 1;
}
}
else if (ssDCval >= 63) {
if (Rs < 477) {
if (Cs < 466) {
return 8;
}
else if (Cs >= 466) {
return 1;
}
}
else if (Rs >= 477) {
return 8;
}
}
}
else if (Rs >= 592) {
if (Rs < 638) {
if (negBalance < 6) {
return 8;
}
else if (negBalance >= 6) {
if (Cs < 421) {
if (tcor < 100) {
return 4;
}
else if (tcor >= 100) {
return 8;
}
}
else if (Cs >= 421) {
return 8;
}
}
}
else if (Rs >= 638) {
return 8;
}
}
}
}
else if (RsCsDiff >= 1) {
if (Cs < 356) {
return 8;
}
else if (Cs >= 356) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 1;
}
}
}
}
}
else if (Rs >= 711) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Rs < 1308) {
if (mu_mv_mag_sq < 118) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (Rs < 819) {
return 2;
}
else if (Rs >= 819) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
else if (Rs >= 1308) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
if (Rs < 1500) {
return 2;
}
else if (Rs >= 1500) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 1113) {
return 8;
}
else if (Rs >= 1113) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 137) {
if (refDCval < 60) {
return 8;
}
else if (refDCval >= 60) {
if (ssDCval < 84) {
if (Rs < 549) {
return 8;
}
else if (Rs >= 549) {
if (SC < 1209) {
if (Rs < 673) {
return 4;
}
else if (Rs >= 673) {
return 1;
}
}
else if (SC >= 1209) {
if (Rs < 1107) {
return 8;
}
else if (Rs >= 1107) {
return 2;
}
}
}
}
else if (ssDCval >= 84) {
return 8;
}
}
}
else if (ssDCval >= 137) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 40) {
if (mcTcor < 88) {
if (refDCval < 36) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (ssDCval < 23) {
if (Cs < 283) {
if (diffMVdiffVal < 30) {
if (diffTSC < -1) {
if (Rs < 114) {
return 8;
}
else if (Rs >= 114) {
return 1;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (diffMVdiffVal >= 30) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (Cs >= 283) {
return 4;
}
}
else if (ssDCval >= 23) {
if (tcor < 59) {
if (negBalance < 41) {
if (Rs < 568) {
if (SC < 808) {
if (AFD < 5) {
if (MVDiff < 29) {
return 1;
}
else if (MVDiff >= 29) {
if (Rs < 59) {
return 1;
}
else if (Rs >= 59) {
return 2;
}
}
}
else if (AFD >= 5) {
if (diffMVdiffVal < -7) {
if (tcor < 43) {
if (diffTSC < 5) {
return 1;
}
else if (diffTSC >= 5) {
return 8;
}
}
else if (tcor >= 43) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
return 1;
}
}
}
else if (SC >= 808) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (Rs >= 568) {
if (Cs < 386) {
return 1;
}
else if (Cs >= 386) {
return 8;
}
}
}
else if (negBalance >= 41) {
if (SC < 554) {
if (gchDC < 3) {
if (Cs < 254) {
return 8;
}
else if (Cs >= 254) {
return 4;
}
}
else if (gchDC >= 3) {
if (Rs < 218) {
return 4;
}
else if (Rs >= 218) {
return 8;
}
}
}
else if (SC >= 554) {
if (Cs < 344) {
return 1;
}
else if (Cs >= 344) {
return 8;
}
}
}
}
else if (tcor >= 59) {
if (Cs < 216) {
if (Rs < 330) {
if (Cs < 105) {
if (mcTcor < 68) {
return 2;
}
else if (mcTcor >= 68) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 204) {
return 4;
}
else if (Rs >= 204) {
return 1;
}
}
}
}
}
else if (Cs >= 105) {
if (posBalance < 46) {
if (Rs < 294) {
return 1;
}
else if (Rs >= 294) {
return 8;
}
}
else if (posBalance >= 46) {
if (Cs < 189) {
if (Cs < 178) {
return 8;
}
else if (Cs >= 178) {
if (Rs < 194) {
return 1;
}
else if (Rs >= 194) {
return 8;
}
}
}
else if (Cs >= 189) {
if (Rs < 222) {
return 2;
}
else if (Rs >= 222) {
return 8;
}
}
}
}
}
else if (Rs >= 330) {
if (posBalance < 41) {
if (Rs < 390) {
return 1;
}
else if (Rs >= 390) {
return 4;
}
}
else if (posBalance >= 41) {
return 1;
}
}
}
else if (Cs >= 216) {
if (Rs < 655) {
if (negBalance < 40) {
if (posBalance < 55) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (refDCval < 28) {
return 1;
}
else if (refDCval >= 28) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 34) {
return 2;
}
else if (ssDCval >= 34) {
return 8;
}
}
}
else if (negBalance >= 40) {
return 8;
}
}
else if (Rs >= 655) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 36) {
if (RsDiff < 4) {
if (tcor < 67) {
if (negBalance < 47) {
if (posBalance < 55) {
if (Rs < 559) {
return 1;
}
else if (Rs >= 559) {
return 2;
}
}
else if (posBalance >= 55) {
if (Cs < 179) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (refDCval < 38) {
if (Rs < 131) {
return 1;
}
else if (Rs >= 131) {
return 8;
}
}
else if (refDCval >= 38) {
return 4;
}
}
}
else if (Cs >= 179) {
if (Rs < 301) {
return 2;
}
else if (Rs >= 301) {
if (Rs < 409) {
return 1;
}
else if (Rs >= 409) {
return 8;
}
}
}
}
}
else if (negBalance >= 47) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 41) {
return 8;
}
else if (MVDiff >= 41) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 3) {
if (Rs < 41) {
return 1;
}
else if (Rs >= 41) {
return 4;
}
}
else if (TSC >= 3) {
return 8;
}
}
}
}
else if (tcor >= 67) {
if (gchDC < 2) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (MVDiff < 40) {
if (tcor < 70) {
if (Cs < 291) {
return 8;
}
else if (Cs >= 291) {
return 2;
}
}
else if (tcor >= 70) {
if (posBalance < 51) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (MVDiff >= 40) {
if (mcTcor < 83) {
return 8;
}
else if (mcTcor >= 83) {
return 1;
}
}
}
}
else if (gchDC >= 2) {
if (diffMVdiffVal < 11) {
if (Rs < 319) {
return 2;
}
else if (Rs >= 319) {
return 1;
}
}
else if (diffMVdiffVal >= 11) {
return 8;
}
}
}
}
else if (RsDiff >= 4) {
if (negBalance < 50) {
if (Rs < 510) {
if (Rs < 339) {
return 1;
}
else if (Rs >= 339) {
return 8;
}
}
else if (Rs >= 510) {
if (Cs < 466) {
return 1;
}
else if (Cs >= 466) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (Rs < 836) {
return 8;
}
else if (Rs >= 836) {
return 1;
}
}
}
}
}
else if (mcTcor >= 88) {
if (posBalance < 27) {
if (posBalance < 15) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (SC < 60) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
if (AFD < 2) {
if (ssDCval < 21) {
return 8;
}
else if (ssDCval >= 21) {
return 4;
}
}
else if (AFD >= 2) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
if (refDCval < 20) {
return 4;
}
else if (refDCval >= 20) {
return 8;
}
}
}
}
}
else if (SC >= 60) {
if (posBalance < 4) {
return 4;
}
else if (posBalance >= 4) {
if (Rs < 317) {
return 1;
}
else if (Rs >= 317) {
return 4;
}
}
}
}
}
else if (posBalance >= 15) {
if (MVDiff < 5) {
if (Cs < 225) {
if (negBalance < 22) {
if (tcor < 102) {
if (mu_mv_mag_sq < 118) {
return 4;
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
else if (tcor >= 102) {
return 8;
}
}
else if (negBalance >= 22) {
if (Rs < 21) {
return 4;
}
else if (Rs >= 21) {
return 1;
}
}
}
else if (Cs >= 225) {
return 1;
}
}
else if (MVDiff >= 5) {
if (ssDCval < 31) {
if (mu_mv_mag_sq < 79) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (Cs < 180) {
return 4;
}
else if (Cs >= 180) {
return 1;
}
}
}
else if (ssDCval >= 31) {
if (Rs < 276) {
if (refDCval < 34) {
if (Rs < 4) {
return 4;
}
else if (Rs >= 4) {
return 8;
}
}
else if (refDCval >= 34) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 8;
}
}
}
else if (Rs >= 276) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (Rs < 363) {
return 1;
}
else if (Rs >= 363) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 27) {
if (refDCval < 34) {
if (Rs < 84) {
if (posBalance < 68) {
return 1;
}
else if (posBalance >= 68) {
return 2;
}
}
else if (Rs >= 84) {
if (posBalance < 47) {
if (ssDCval < 30) {
if (MVDiff < 27) {
if (SC < 132) {
if (Rs < 92) {
return 1;
}
else if (Rs >= 92) {
return 4;
}
}
else if (SC >= 132) {
if (Cs < 405) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (Cs >= 405) {
return 4;
}
}
}
else if (MVDiff >= 27) {
if (tcor < 96) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
return 8;
}
}
else if (tcor >= 96) {
return 4;
}
}
}
else if (ssDCval >= 30) {
if (Cs < 260) {
if (SC < 507) {
if (Rs < 109) {
if (SC < 180) {
if (mu_mv_mag_sq < 42) {
if (Rs < 86) {
return 8;
}
else if (Rs >= 86) {
return 4;
}
}
else if (mu_mv_mag_sq >= 42) {
return 1;
}
}
else if (SC >= 180) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (Rs >= 109) {
if (mcTcor < 95) {
if (gchDC < 1) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 136) {
return 8;
}
else if (Rs >= 136) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
else if (gchDC >= 1) {
if (MVDiff < 8) {
if (Rs < 330) {
return 8;
}
else if (Rs >= 330) {
return 2;
}
}
else if (MVDiff >= 8) {
if (Cs < 122) {
if (Rs < 169) {
if (Cs < 59) {
return 4;
}
else if (Cs >= 59) {
return 1;
}
}
else if (Rs >= 169) {
return 4;
}
}
else if (Cs >= 122) {
return 1;
}
}
}
}
else if (mcTcor >= 95) {
if (posBalance < 31) {
if (tcor < 90) {
if (Rs < 208) {
return 8;
}
else if (Rs >= 208) {
return 2;
}
}
else if (tcor >= 90) {
if (Rs < 124) {
return 8;
}
else if (Rs >= 124) {
return 1;
}
}
}
else if (posBalance >= 31) {
if (mu_mv_mag_sq < 42) {
if (posBalance < 33) {
return 4;
}
else if (posBalance >= 33) {
if (Rs < 112) {
return 4;
}
else if (Rs >= 112) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 42) {
return 1;
}
}
}
}
}
else if (SC >= 507) {
if (posBalance < 41) {
if (MVDiff < 4) {
if (Rs < 364) {
return 8;
}
else if (Rs >= 364) {
if (Rs < 491) {
return 4;
}
else if (Rs >= 491) {
return 1;
}
}
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (posBalance >= 41) {
if (Rs < 703) {
return 1;
}
else if (Rs >= 703) {
return 8;
}
}
}
}
else if (Cs >= 260) {
if (negBalance < 15) {
return 4;
}
else if (negBalance >= 15) {
return 1;
}
}
}
}
else if (posBalance >= 47) {
if (SC < 349) {
if (tcor < 74) {
if (mu_mv_mag_sq < 9) {
if (Rs < 166) {
return 8;
}
else if (Rs >= 166) {
return 2;
}
}
else if (mu_mv_mag_sq >= 9) {
if (SC < 336) {
return 1;
}
else if (SC >= 336) {
return 8;
}
}
}
else if (tcor >= 74) {
return 1;
}
}
else if (SC >= 349) {
if (MVDiff < 44) {
return 1;
}
else if (MVDiff >= 44) {
if (Rs < 198) {
return 2;
}
else if (Rs >= 198) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 34) {
if (RsCsDiff < 3) {
if (gchDC < 1) {
if (RsDiff < 1) {
if (Rs < 115) {
if (mu_mv_mag_sq < 91) {
if (SC < 100) {
return 2;
}
else if (SC >= 100) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (Rs < 40) {
return 8;
}
else if (Rs >= 40) {
return 4;
}
}
}
else if (Rs >= 115) {
if (ssDCval < 35) {
if (Rs < 510) {
return 1;
}
else if (Rs >= 510) {
return 4;
}
}
else if (ssDCval >= 35) {
if (tcor < 83) {
if (Rs < 276) {
return 2;
}
else if (Rs >= 276) {
return 8;
}
}
else if (tcor >= 83) {
if (diffAFD < -2) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
return 1;
}
}
else if (diffAFD >= -2) {
if (Cs < 59) {
return 1;
}
else if (Cs >= 59) {
if (negBalance < 17) {
if (Cs < 245) {
return 4;
}
else if (Cs >= 245) {
return 8;
}
}
else if (negBalance >= 17) {
if (diffRsCsdiff < -1) {
if (Cs < 173) {
return 2;
}
else if (Cs >= 173) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
}
}
}
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 3) {
if (Rs < 246) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
else if (Rs >= 246) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 91) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
return 8;
}
}
else if (mcTcor >= 91) {
if (diffMVdiffVal < -1) {
if (Rs < 515) {
if (posBalance < 35) {
if (Rs < 410) {
return 4;
}
else if (Rs >= 410) {
return 1;
}
}
else if (posBalance >= 35) {
return 1;
}
}
else if (Rs >= 515) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (tcor < 92) {
if (posBalance < 46) {
if (Rs < 200) {
return 8;
}
else if (Rs >= 200) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (posBalance >= 46) {
return 1;
}
}
else if (tcor >= 92) {
if (Cs < 317) {
if (Rs < 412) {
return 2;
}
else if (Rs >= 412) {
return 8;
}
}
else if (Cs >= 317) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 1;
}
}
}
}
}
}
}
}
else if (gchDC >= 1) {
if (TSC < 4) {
if (mu_mv_mag_sq < 53) {
if (Cs < 22) {
return 2;
}
else if (Cs >= 22) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (negBalance < 42) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 39) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 1;
}
}
else if (mu_mv_mag_sq >= 39) {
if (Rs < 270) {
return 1;
}
else if (Rs >= 270) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (Rs < 177) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 4;
}
}
else if (Rs >= 177) {
return 1;
}
}
}
else if (negBalance >= 42) {
if (mu_mv_mag_sq < 6) {
return 2;
}
else if (mu_mv_mag_sq >= 6) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (mcTcor < 90) {
if (Rs < 238) {
return 4;
}
else if (Rs >= 238) {
return 8;
}
}
else if (mcTcor >= 90) {
if (ssDCval < 36) {
if (SC < 417) {
if (Rs < 77) {
return 4;
}
else if (Rs >= 77) {
return 8;
}
}
else if (SC >= 417) {
if (Rs < 368) {
return 1;
}
else if (Rs >= 368) {
return 4;
}
}
}
else if (ssDCval >= 36) {
if (negBalance < 15) {
if (Rs < 212) {
return 4;
}
else if (Rs >= 212) {
return 8;
}
}
else if (negBalance >= 15) {
if (MVDiff < 24) {
if (Rs < 579) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (Rs >= 579) {
return 8;
}
}
else if (MVDiff >= 24) {
if (Rs < 448) {
return 8;
}
else if (Rs >= 448) {
return 4;
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (Rs < 173) {
if (Rs < 90) {
return 4;
}
else if (Rs >= 90) {
return 8;
}
}
else if (Rs >= 173) {
if (Rs < 178) {
return 2;
}
else if (Rs >= 178) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 3) {
if (RsCsDiff < 9) {
if (mu_mv_mag_sq < 21) {
if (RsCsDiff < 5) {
if (Rs < 791) {
return 1;
}
else if (Rs >= 791) {
return 2;
}
}
else if (RsCsDiff >= 5) {
if (Rs < 142) {
return 4;
}
else if (Rs >= 142) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (Rs < 728) {
return 8;
}
else if (Rs >= 728) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (posBalance < 32) {
return 2;
}
else if (posBalance >= 32) {
if (MVDiff < 2) {
if (Rs < 921) {
return 8;
}
else if (Rs >= 921) {
return 1;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
else if (RsCsDiff >= 9) {
if (Cs < 437) {
return 1;
}
else if (Cs >= 437) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 40) {
if (Rs < 307) {
if (ssDCval < 70) {
if (posBalance < 31) {
if (tcor < 98) {
if (MVDiff < 36) {
if (ssDCval < 47) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
else if (ssDCval >= 47) {
if (ssDCval < 59) {
if (Cs < 279) {
return 8;
}
else if (Cs >= 279) {
if (Rs < 222) {
return 4;
}
else if (Rs >= 222) {
return 1;
}
}
}
else if (ssDCval >= 59) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
return 8;
}
}
}
}
else if (MVDiff >= 36) {
if (negBalance < 74) {
return 4;
}
else if (negBalance >= 74) {
if (diffMVdiffVal < 40) {
if (Rs < 3) {
return 2;
}
else if (Rs >= 3) {
return 1;
}
}
else if (diffMVdiffVal >= 40) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (Rs < 15) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
else if (Rs >= 15) {
if (mcTcor < 94) {
if (ssDCval < 45) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
else if (ssDCval >= 45) {
return 8;
}
}
else if (mcTcor >= 94) {
if (posBalance < 25) {
if (SC < 285) {
if (mcTcor < 103) {
return 8;
}
else if (mcTcor >= 103) {
if (Rs < 59) {
return 4;
}
else if (Rs >= 59) {
return 8;
}
}
}
else if (SC >= 285) {
if (MVDiff < 3) {
if (Cs < 152) {
return 4;
}
else if (Cs >= 152) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (posBalance >= 25) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 31) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 44) {
if (mcTcor < 85) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
if (refDCval < 64) {
if (Cs < 130) {
if (SC < 149) {
if (Cs < 90) {
return 8;
}
else if (Cs >= 90) {
return 4;
}
}
else if (SC >= 149) {
if (AFD < 22) {
if (refDCval < 41) {
return 4;
}
else if (refDCval >= 41) {
return 1;
}
}
else if (AFD >= 22) {
if (Rs < 155) {
return 4;
}
else if (Rs >= 155) {
if (Rs < 181) {
return 8;
}
else if (Rs >= 181) {
return 1;
}
}
}
}
}
else if (Cs >= 130) {
if (tcor < 69) {
if (gchDC < 1) {
if (Rs < 256) {
return 1;
}
else if (Rs >= 256) {
return 4;
}
}
else if (gchDC >= 1) {
if (MVDiff < 337) {
return 8;
}
else if (MVDiff >= 337) {
return 4;
}
}
}
else if (tcor >= 69) {
if (Cs < 290) {
return 2;
}
else if (Cs >= 290) {
return 1;
}
}
}
}
else if (refDCval >= 64) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (AFD < 54) {
return 4;
}
else if (AFD >= 54) {
return 1;
}
}
}
}
}
else if (mcTcor >= 85) {
if (TSC < 4) {
if (CsDiff < 1) {
if (Rs < 21) {
if (Rs < 9) {
if (Rs < 4) {
return 8;
}
else if (Rs >= 4) {
return 1;
}
}
else if (Rs >= 9) {
return 4;
}
}
else if (Rs >= 21) {
if (refDCval < 42) {
if (Rs < 58) {
if (Rs < 37) {
return 1;
}
else if (Rs >= 37) {
return 4;
}
}
else if (Rs >= 58) {
return 2;
}
}
else if (refDCval >= 42) {
if (SC < 114) {
if (Rs < 26) {
return 1;
}
else if (Rs >= 26) {
return 2;
}
}
else if (SC >= 114) {
if (posBalance < 66) {
if (refDCval < 59) {
if (Cs < 200) {
return 1;
}
else if (Cs >= 200) {
return 4;
}
}
else if (refDCval >= 59) {
if (Rs < 118) {
if (Rs < 103) {
return 4;
}
else if (Rs >= 103) {
return 1;
}
}
else if (Rs >= 118) {
if (SC < 268) {
return 2;
}
else if (SC >= 268) {
return 1;
}
}
}
}
else if (posBalance >= 66) {
return 2;
}
}
}
}
}
else if (CsDiff >= 1) {
if (refDCval < 52) {
if (Cs < 325) {
if (ssDCval < 43) {
if (Rs < 137) {
return 4;
}
else if (Rs >= 137) {
return 2;
}
}
else if (ssDCval >= 43) {
if (refDCval < 49) {
if (Rs < 181) {
return 1;
}
else if (Rs >= 181) {
return 4;
}
}
else if (refDCval >= 49) {
return 2;
}
}
}
else if (Cs >= 325) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
return 1;
}
}
}
else if (refDCval >= 52) {
if (RsCsDiff < 5) {
if (posBalance < 70) {
if (diffMVdiffVal < -5) {
if (MVDiff < 9) {
if (Rs < 236) {
return 1;
}
else if (Rs >= 236) {
return 4;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (diffMVdiffVal >= -5) {
return 4;
}
}
else if (posBalance >= 70) {
return 1;
}
}
else if (RsCsDiff >= 5) {
if (AFD < 10) {
if (Rs < 291) {
return 4;
}
else if (Rs >= 291) {
return 2;
}
}
else if (AFD >= 10) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (MVDiff < 58) {
if (MVDiff < 41) {
if (tcor < 83) {
if (mu_mv_mag_sq < 18) {
if (tcor < 67) {
return 4;
}
else if (tcor >= 67) {
if (Cs < 473) {
return 1;
}
else if (Cs >= 473) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (TSC < 5) {
if (RsCsDiff < 7) {
return 2;
}
else if (RsCsDiff >= 7) {
return 1;
}
}
else if (TSC >= 5) {
if (Cs < 241) {
return 4;
}
else if (Cs >= 241) {
return 1;
}
}
}
}
else if (tcor >= 83) {
return 1;
}
}
else if (MVDiff >= 41) {
if (SC < 280) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
return 8;
}
}
else if (SC >= 280) {
if (Cs < 223) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
if (Rs < 252) {
return 2;
}
else if (Rs >= 252) {
return 1;
}
}
}
else if (Cs >= 223) {
if (posBalance < 50) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
else if (posBalance >= 50) {
return 1;
}
}
}
}
}
else if (MVDiff >= 58) {
if (AFD < 15) {
if (TSC < 5) {
if (Rs < 58) {
return 2;
}
else if (Rs >= 58) {
return 1;
}
}
else if (TSC >= 5) {
if (Cs < 35) {
return 1;
}
else if (Cs >= 35) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
return 1;
}
}
}
}
else if (AFD >= 15) {
if (RsCsDiff < 30) {
if (diffAFD < 4) {
return 1;
}
else if (diffAFD >= 4) {
if (Rs < 148) {
return 1;
}
else if (Rs >= 148) {
return 2;
}
}
}
else if (RsCsDiff >= 30) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (negBalance < 27) {
if (SC < 304) {
if (tcor < 93) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
return 2;
}
}
}
}
else if (tcor >= 93) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
else if (SC >= 304) {
if (diffTSC < 1) {
if (Cs < 298) {
return 1;
}
else if (Cs >= 298) {
return 2;
}
}
else if (diffTSC >= 1) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (Rs < 259) {
return 4;
}
else if (Rs >= 259) {
return 1;
}
}
}
}
}
else if (negBalance >= 27) {
if (ssDCval < 59) {
if (gchDC < 1) {
if (ssDCval < 51) {
if (Cs < 294) {
if (Rs < 65) {
if (Rs < 35) {
return 8;
}
else if (Rs >= 35) {
return 1;
}
}
else if (Rs >= 65) {
return 8;
}
}
else if (Cs >= 294) {
return 1;
}
}
else if (ssDCval >= 51) {
if (Cs < 189) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 2;
}
}
else if (Cs >= 189) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (posBalance < 52) {
if (posBalance < 43) {
if (ssDCval < 43) {
return 1;
}
else if (ssDCval >= 43) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
return 2;
}
}
}
else if (posBalance >= 43) {
if (mcTcor < 95) {
if (Cs < 60) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (Cs >= 60) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 1;
}
}
}
else if (mcTcor >= 95) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 209) {
return 2;
}
else if (Rs >= 209) {
return 1;
}
}
}
}
}
else if (posBalance >= 52) {
if (tcor < 30) {
if (MVDiff < 163) {
return 2;
}
else if (MVDiff >= 163) {
if (TSC < 16) {
return 8;
}
else if (TSC >= 16) {
return 1;
}
}
}
else if (tcor >= 30) {
if (MVDiff < 158) {
if (Rs < 288) {
if (mcTcor < 68) {
return 1;
}
else if (mcTcor >= 68) {
return 4;
}
}
else if (Rs >= 288) {
return 8;
}
}
else if (MVDiff >= 158) {
return 8;
}
}
}
}
}
else if (ssDCval >= 59) {
if (SC < 228) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Cs < 33) {
return 1;
}
else if (Cs >= 33) {
return 4;
}
}
}
else if (SC >= 228) {
if (MVDiff < 10) {
if (negBalance < 30) {
if (Cs < 244) {
return 4;
}
else if (Cs >= 244) {
return 1;
}
}
else if (negBalance >= 30) {
if (Cs < 176) {
return 1;
}
else if (Cs >= 176) {
if (mu_mv_mag_sq < 110) {
return 4;
}
else if (mu_mv_mag_sq >= 110) {
return 1;
}
}
}
}
else if (MVDiff >= 10) {
if (Cs < 504) {
return 4;
}
else if (Cs >= 504) {
return 8;
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (TSCindex < 2) {
if (Rs < 119) {
if (mu_mv_mag_sq < 111) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Cs < 217) {
return 1;
}
else if (Cs >= 217) {
if (Rs < 65) {
return 1;
}
else if (Rs >= 65) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 139) {
return 8;
}
else if (Cs >= 139) {
return 2;
}
}
}
else if (Rs >= 119) {
if (ssDCval < 65) {
if (Rs < 245) {
if (Rs < 200) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
if (Cs < 191) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 1;
}
}
}
else if (Cs >= 191) {
if (Rs < 197) {
return 4;
}
else if (Rs >= 197) {
return 1;
}
}
}
}
else if (Rs >= 200) {
return 4;
}
}
else if (Rs >= 245) {
if (posBalance < 41) {
if (Cs < 263) {
return 4;
}
else if (Cs >= 263) {
return 1;
}
}
else if (posBalance >= 41) {
if (Rs < 296) {
return 4;
}
else if (Rs >= 296) {
return 2;
}
}
}
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 118) {
return 2;
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 183) {
return 4;
}
else if (Rs >= 183) {
if (Cs < 275) {
return 2;
}
else if (Cs >= 275) {
return 1;
}
}
}
}
}
}
else if (TSCindex >= 2) {
if (CsDiff < 1) {
if (refDCval < 67) {
if (SC < 336) {
if (MVDiff < 1) {
if (Rs < 115) {
return 1;
}
else if (Rs >= 115) {
if (Cs < 165) {
return 2;
}
else if (Cs >= 165) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 54) {
if (diffAFD < 2) {
if (tcor < 94) {
if (Cs < 90) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
return 4;
}
}
else if (Cs >= 90) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
if (TSCindex < 3) {
if (Rs < 187) {
return 1;
}
else if (Rs >= 187) {
return 4;
}
}
else if (TSCindex >= 3) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (Rs < 123) {
return 4;
}
else if (Rs >= 123) {
return 1;
}
}
}
}
}
}
else if (tcor >= 94) {
if (Cs < 229) {
if (MVDiff < 4) {
if (posBalance < 47) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
return 4;
}
}
}
else if (posBalance >= 47) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
if (Rs < 107) {
return 1;
}
else if (Rs >= 107) {
return 2;
}
}
}
}
else if (MVDiff >= 4) {
if (diffMVdiffVal < -11) {
return 4;
}
else if (diffMVdiffVal >= -11) {
return 1;
}
}
}
else if (Cs >= 229) {
return 4;
}
}
}
else if (diffAFD >= 2) {
if (TSC < 3) {
if (Rs < 92) {
return 8;
}
else if (Rs >= 92) {
return 2;
}
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (negBalance >= 54) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
if (Rs < 96) {
return 2;
}
else if (Rs >= 96) {
return 4;
}
}
}
}
}
}
else if (SC >= 336) {
if (negBalance < 31) {
if (MVDiff < 1) {
if (ssDCval < 63) {
if (diffRsCsdiff < -1) {
if (Cs < 395) {
return 4;
}
else if (Cs >= 395) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (Rs < 172) {
return 4;
}
else if (Rs >= 172) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 237) {
return 1;
}
else if (Rs >= 237) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 234) {
return 4;
}
else if (Rs >= 234) {
return 1;
}
}
}
}
}
else if (ssDCval >= 63) {
return 4;
}
}
else if (MVDiff >= 1) {
if (ssDCval < 56) {
if (diffTSC < -1) {
if (Rs < 159) {
return 4;
}
else if (Rs >= 159) {
return 1;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (ssDCval >= 56) {
if (Cs < 228) {
if (tcor < 98) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 4;
}
}
}
}
else if (tcor >= 98) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 1;
}
}
}
else if (Cs >= 228) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (mu_mv_mag_sq < 99) {
return 1;
}
else if (mu_mv_mag_sq >= 99) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 31) {
if (refDCval < 64) {
if (MVDiff < 3) {
if (diffMVdiffVal < -2) {
if (Rs < 256) {
return 4;
}
else if (Rs >= 256) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
if (posBalance < 49) {
if (SC < 484) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
return 4;
}
}
else if (SC >= 484) {
return 1;
}
}
else if (posBalance >= 49) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (mcTcor < 98) {
if (MVDiff < 4) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (mcTcor >= 98) {
if (Rs < 232) {
if (Cs < 198) {
return 1;
}
else if (Cs >= 198) {
if (MVDiff < 6) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
if (Rs < 215) {
return 1;
}
else if (Rs >= 215) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (Cs < 202) {
return 4;
}
else if (Cs >= 202) {
return 1;
}
}
}
}
else if (Rs >= 232) {
return 1;
}
}
}
}
else if (refDCval >= 64) {
if (Rs < 153) {
if (Cs < 288) {
return 4;
}
else if (Cs >= 288) {
return 1;
}
}
else if (Rs >= 153) {
return 1;
}
}
}
}
}
else if (refDCval >= 67) {
if (mu_mv_mag_sq < 91) {
if (tcor < 97) {
if (Rs < 147) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 4;
}
}
else if (Rs >= 147) {
if (refDCval < 68) {
return 4;
}
else if (refDCval >= 68) {
if (Rs < 182) {
return 1;
}
else if (Rs >= 182) {
return 8;
}
}
}
}
else if (tcor >= 97) {
if (Cs < 167) {
return 2;
}
else if (Cs >= 167) {
if (negBalance < 26) {
if (Rs < 91) {
return 1;
}
else if (Rs >= 91) {
return 2;
}
}
else if (negBalance >= 26) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (Cs < 228) {
return 2;
}
else if (Cs >= 228) {
if (Cs < 377) {
if (Rs < 182) {
if (Cs < 337) {
if (tcor < 99) {
if (Rs < 142) {
if (posBalance < 47) {
if (Rs < 138) {
return 4;
}
else if (Rs >= 138) {
return 1;
}
}
else if (posBalance >= 47) {
return 1;
}
}
else if (Rs >= 142) {
if (mu_mv_mag_sq < 104) {
if (Cs < 309) {
return 4;
}
else if (Cs >= 309) {
return 1;
}
}
else if (mu_mv_mag_sq >= 104) {
return 4;
}
}
}
else if (tcor >= 99) {
return 4;
}
}
else if (Cs >= 337) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (diffAFD < -1) {
if (Rs < 166) {
return 2;
}
else if (Rs >= 166) {
return 1;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
}
else if (Rs >= 182) {
return 4;
}
}
else if (Cs >= 377) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
}
}
else if (CsDiff >= 1) {
if (refDCval < 53) {
if (mcTcor < 98) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (diffAFD < -2) {
if (Rs < 88) {
return 2;
}
else if (Rs >= 88) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (diffAFD >= -2) {
if (MVDiff < 50) {
return 4;
}
else if (MVDiff >= 50) {
if (Rs < 125) {
return 4;
}
else if (Rs >= 125) {
return 1;
}
}
}
}
}
else if (mcTcor >= 98) {
if (Cs < 193) {
if (ssDCval < 44) {
if (Rs < 97) {
return 4;
}
else if (Rs >= 97) {
return 2;
}
}
else if (ssDCval >= 44) {
if (Rs < 246) {
return 1;
}
else if (Rs >= 246) {
return 8;
}
}
}
else if (Cs >= 193) {
if (negBalance < 21) {
return 4;
}
else if (negBalance >= 21) {
return 1;
}
}
}
}
else if (refDCval >= 53) {
if (RsDiff < 1) {
if (Rs < 38) {
return 1;
}
else if (Rs >= 38) {
if (MVDiff < 3) {
if (Rs < 174) {
return 1;
}
else if (Rs >= 174) {
if (Cs < 369) {
return 4;
}
else if (Cs >= 369) {
if (Rs < 223) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (Rs >= 223) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (mcTcor < 100) {
if (Cs < 218) {
if (AFD < 5) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (AFD >= 5) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
return 4;
}
}
}
else if (Cs >= 218) {
if (gchDC < 1) {
if (ssDCval < 54) {
if (Cs < 308) {
return 2;
}
else if (Cs >= 308) {
return 4;
}
}
else if (ssDCval >= 54) {
return 4;
}
}
else if (gchDC >= 1) {
if (MVDiff < 18) {
if (Rs < 243) {
return 4;
}
else if (Rs >= 243) {
if (Rs < 265) {
return 1;
}
else if (Rs >= 265) {
return 4;
}
}
}
else if (MVDiff >= 18) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
else if (mcTcor >= 100) {
if (Rs < 149) {
return 8;
}
else if (Rs >= 149) {
if (Rs < 200) {
return 4;
}
else if (Rs >= 200) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 1) {
if (TSCindex < 3) {
if (Rs < 154) {
return 4;
}
else if (Rs >= 154) {
return 1;
}
}
else if (TSCindex >= 3) {
if (TSC < 4) {
if (Rs < 269) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (CsDiff < 2) {
if (Rs < 120) {
return 1;
}
else if (Rs >= 120) {
if (tcor < 92) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (tcor >= 92) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (Rs < 162) {
return 8;
}
else if (Rs >= 162) {
return 4;
}
}
}
}
}
else if (CsDiff >= 2) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 4;
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 242) {
return 8;
}
else if (Rs >= 242) {
return 1;
}
}
}
else if (Rs >= 269) {
if (gchDC < 1) {
if (Cs < 189) {
return 4;
}
else if (Cs >= 189) {
return 1;
}
}
else if (gchDC >= 1) {
if (posBalance < 51) {
if (diffRsCsdiff < -4) {
return 2;
}
else if (diffRsCsdiff >= -4) {
return 4;
}
}
else if (posBalance >= 51) {
return 2;
}
}
}
}
else if (TSC >= 4) {
if (Rs < 250) {
if (Rs < 132) {
return 1;
}
else if (Rs >= 132) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
if (gchDC < 6) {
return 2;
}
else if (gchDC >= 6) {
return 8;
}
}
}
}
else if (Rs >= 250) {
if (Cs < 215) {
return 4;
}
else if (Cs >= 215) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 70) {
if (TSC < 2) {
if (refDCval < 93) {
if (Cs < 254) {
if (Rs < 78) {
if (Rs < 63) {
return 1;
}
else if (Rs >= 63) {
return 4;
}
}
else if (Rs >= 78) {
if (Cs < 125) {
if (tcor < 99) {
if (posBalance < 41) {
if (Rs < 174) {
return 2;
}
else if (Rs >= 174) {
return 8;
}
}
else if (posBalance >= 41) {
return 8;
}
}
else if (tcor >= 99) {
if (ssDCval < 75) {
if (Rs < 95) {
return 4;
}
else if (Rs >= 95) {
return 2;
}
}
else if (ssDCval >= 75) {
if (Rs < 123) {
return 2;
}
else if (Rs >= 123) {
return 1;
}
}
}
}
else if (Cs >= 125) {
if (ssDCval < 81) {
if (negBalance < 23) {
if (negBalance < 21) {
return 2;
}
else if (negBalance >= 21) {
if (mu_mv_mag_sq < 124) {
return 2;
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 203) {
if (Rs < 180) {
return 2;
}
else if (Rs >= 180) {
return 1;
}
}
else if (Rs >= 203) {
return 2;
}
}
}
}
else if (negBalance >= 23) {
if (mu_mv_mag_sq < 71) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 99) {
return 2;
}
else if (Rs >= 99) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (Cs < 215) {
if (diffMVdiffVal < 4) {
return 2;
}
else if (diffMVdiffVal >= 4) {
return 1;
}
}
else if (Cs >= 215) {
if (ssDCval < 77) {
return 1;
}
else if (ssDCval >= 77) {
return 2;
}
}
}
}
}
else if (ssDCval >= 81) {
if (ssDCval < 86) {
if (Rs < 262) {
return 8;
}
else if (Rs >= 262) {
return 1;
}
}
else if (ssDCval >= 86) {
if (Rs < 150) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
if (Rs < 119) {
return 2;
}
else if (Rs >= 119) {
return 8;
}
}
}
else if (Rs >= 150) {
return 2;
}
}
}
}
}
}
else if (Cs >= 254) {
if (Rs < 114) {
return 4;
}
else if (Rs >= 114) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
return 4;
}
}
}
}
else if (refDCval >= 93) {
if (Cs < 127) {
if (refDCval < 110) {
if (diffTSC < 1) {
if (Cs < 117) {
return 1;
}
else if (Cs >= 117) {
if (Rs < 89) {
return 4;
}
else if (Rs >= 89) {
return 1;
}
}
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (refDCval >= 110) {
if (negBalance < 48) {
if (Rs < 122) {
return 4;
}
else if (Rs >= 122) {
return 2;
}
}
else if (negBalance >= 48) {
if (Rs < 87) {
return 2;
}
else if (Rs >= 87) {
return 8;
}
}
}
}
else if (Cs >= 127) {
if (ssDCval < 136) {
if (MVDiff < 29) {
if (Rs < 185) {
if (diffTSC < -1) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (Rs >= 185) {
if (Cs < 168) {
if (Rs < 244) {
return 2;
}
else if (Rs >= 244) {
return 1;
}
}
else if (Cs >= 168) {
return 4;
}
}
}
else if (MVDiff >= 29) {
if (Rs < 120) {
return 2;
}
else if (Rs >= 120) {
return 1;
}
}
}
else if (ssDCval >= 136) {
return 2;
}
}
}
}
else if (TSC >= 2) {
if (ssDCval < 112) {
if (negBalance < 43) {
if (Cs < 239) {
if (MVDiff < 3) {
if (mcTcor < 98) {
if (Cs < 151) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 151) {
if (Rs < 228) {
return 2;
}
else if (Rs >= 228) {
if (Rs < 290) {
return 1;
}
else if (Rs >= 290) {
return 8;
}
}
}
}
else if (mcTcor >= 98) {
if (Rs < 239) {
if (posBalance < 47) {
if (Rs < 201) {
return 2;
}
else if (Rs >= 201) {
return 1;
}
}
else if (posBalance >= 47) {
if (Rs < 219) {
if (negBalance < 22) {
return 2;
}
else if (negBalance >= 22) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (negBalance < 27) {
return 1;
}
else if (negBalance >= 27) {
return 2;
}
}
}
}
else if (Rs >= 219) {
return 2;
}
}
}
else if (Rs >= 239) {
if (Cs < 181) {
if (Rs < 283) {
return 8;
}
else if (Rs >= 283) {
return 4;
}
}
else if (Cs >= 181) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (gchDC < 6) {
if (Rs < 243) {
if (ssDCval < 98) {
if (Rs < 85) {
return 1;
}
else if (Rs >= 85) {
if (Cs < 102) {
if (negBalance < 35) {
if (Rs < 199) {
return 8;
}
else if (Rs >= 199) {
return 1;
}
}
else if (negBalance >= 35) {
if (diffTSC < -1) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 1;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
}
else if (Cs >= 102) {
if (posBalance < 65) {
if (TSCindex < 3) {
if (mu_mv_mag_sq < 80) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 1;
}
}
else if (mu_mv_mag_sq >= 80) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 2;
}
}
}
else if (TSCindex >= 3) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (Cs < 188) {
return 1;
}
else if (Cs >= 188) {
return 2;
}
}
}
}
else if (posBalance >= 65) {
if (Rs < 118) {
return 2;
}
else if (Rs >= 118) {
return 1;
}
}
}
}
}
else if (ssDCval >= 98) {
if (Rs < 75) {
return 1;
}
else if (Rs >= 75) {
if (Rs < 178) {
return 2;
}
else if (Rs >= 178) {
return 8;
}
}
}
}
else if (Rs >= 243) {
if (MVDiff < 9) {
if (Cs < 189) {
return 4;
}
else if (Cs >= 189) {
return 1;
}
}
else if (MVDiff >= 9) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
if (Rs < 265) {
return 1;
}
else if (Rs >= 265) {
return 8;
}
}
}
}
}
else if (gchDC >= 6) {
if (diffTSC < 2) {
if (Rs < 98) {
return 4;
}
else if (Rs >= 98) {
if (Cs < 65) {
return 4;
}
else if (Cs >= 65) {
return 8;
}
}
}
else if (diffTSC >= 2) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
return 1;
}
}
}
}
}
else if (Cs >= 239) {
if (SC < 486) {
if (refDCval < 76) {
if (Cs < 311) {
if (Rs < 111) {
if (Rs < 69) {
return 2;
}
else if (Rs >= 69) {
return 4;
}
}
else if (Rs >= 111) {
if (Cs < 241) {
if (Rs < 173) {
return 4;
}
else if (Rs >= 173) {
return 1;
}
}
else if (Cs >= 241) {
if (mu_mv_mag_sq < 4) {
if (Rs < 152) {
return 1;
}
else if (Rs >= 152) {
return 4;
}
}
else if (mu_mv_mag_sq >= 4) {
return 1;
}
}
}
}
else if (Cs >= 311) {
if (negBalance < 27) {
if (Rs < 82) {
return 2;
}
else if (Rs >= 82) {
if (Rs < 106) {
return 4;
}
else if (Rs >= 106) {
return 1;
}
}
}
else if (negBalance >= 27) {
return 4;
}
}
}
else if (refDCval >= 76) {
if (ssDCval < 88) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Rs < 211) {
return 8;
}
else if (Rs >= 211) {
return 2;
}
}
}
else if (ssDCval >= 88) {
if (Rs < 79) {
return 2;
}
else if (Rs >= 79) {
return 1;
}
}
}
}
else if (SC >= 486) {
if (refDCval < 79) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 80) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (ssDCval < 85) {
if (Cs < 323) {
return 1;
}
else if (Cs >= 323) {
if (refDCval < 65) {
return 1;
}
else if (refDCval >= 65) {
if (Cs < 461) {
return 4;
}
else if (Cs >= 461) {
return 1;
}
}
}
}
else if (ssDCval >= 85) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 80) {
return 8;
}
}
else if (mcTcor >= 98) {
if (posBalance < 57) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Rs < 163) {
return 1;
}
else if (Rs >= 163) {
return 4;
}
}
}
else if (posBalance >= 57) {
if (Cs < 295) {
return 2;
}
else if (Cs >= 295) {
return 1;
}
}
}
}
else if (refDCval >= 79) {
if (TSC < 3) {
if (Cs < 471) {
return 1;
}
else if (Cs >= 471) {
return 2;
}
}
else if (TSC >= 3) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 43) {
if (mcTcor < 68) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
if (refDCval < 78) {
return 8;
}
else if (refDCval >= 78) {
return 4;
}
}
}
else if (mcTcor >= 68) {
if (refDCval < 78) {
if (Cs < 65) {
if (mu_mv_mag_sq < 47) {
return 1;
}
else if (mu_mv_mag_sq >= 47) {
return 8;
}
}
else if (Cs >= 65) {
if (tcor < 81) {
if (diffRsCsdiff < -26) {
if (Rs < 146) {
return 4;
}
else if (Rs >= 146) {
return 8;
}
}
else if (diffRsCsdiff >= -26) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
if (MVDiff < 206) {
return 2;
}
else if (MVDiff >= 206) {
return 1;
}
}
}
}
}
else if (tcor >= 81) {
if (MVDiff < 4) {
if (refDCval < 77) {
if (Cs < 247) {
return 1;
}
else if (Cs >= 247) {
return 8;
}
}
else if (refDCval >= 77) {
return 2;
}
}
else if (MVDiff >= 4) {
if (negBalance < 59) {
if (CsDiff < 1) {
if (Rs < 116) {
return 4;
}
else if (Rs >= 116) {
return 1;
}
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 72) {
return 4;
}
else if (mu_mv_mag_sq >= 72) {
return 8;
}
}
}
else if (negBalance >= 59) {
if (Cs < 106) {
return 1;
}
else if (Cs >= 106) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 78) {
if (gchDC < 1) {
if (mcTcor < 97) {
if (posBalance < 63) {
if (MVDiff < 30) {
if (Cs < 361) {
if (AFD < 5) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
return 8;
}
}
else if (AFD >= 5) {
if (RsCsDiff < 2) {
if (RsCsDiff < 1) {
if (Cs < 245) {
return 2;
}
else if (Cs >= 245) {
return 8;
}
}
else if (RsCsDiff >= 1) {
return 8;
}
}
else if (RsCsDiff >= 2) {
if (Rs < 267) {
if (posBalance < 60) {
if (negBalance < 49) {
if (Cs < 298) {
return 2;
}
else if (Cs >= 298) {
return 8;
}
}
else if (negBalance >= 49) {
return 2;
}
}
else if (posBalance >= 60) {
if (Cs < 120) {
return 2;
}
else if (Cs >= 120) {
return 8;
}
}
}
else if (Rs >= 267) {
if (Rs < 304) {
return 8;
}
else if (Rs >= 304) {
return 2;
}
}
}
}
}
else if (Cs >= 361) {
return 8;
}
}
else if (MVDiff >= 30) {
if (SC < 125) {
return 8;
}
else if (SC >= 125) {
if (tcor < 52) {
return 2;
}
else if (tcor >= 52) {
return 1;
}
}
}
}
else if (posBalance >= 63) {
return 1;
}
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 7) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
if (Rs < 52) {
return 2;
}
else if (Rs >= 52) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 7) {
return 1;
}
}
}
else if (gchDC >= 1) {
if (Cs < 163) {
if (tcor < 61) {
if (AFD < 20) {
if (mcTcor < 83) {
if (Cs < 13) {
return 4;
}
else if (Cs >= 13) {
return 8;
}
}
else if (mcTcor >= 83) {
if (MVDiff < 44) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (MVDiff >= 44) {
return 2;
}
}
}
else if (AFD >= 20) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 4;
}
}
}
else if (tcor >= 61) {
if (SC < 341) {
if (ssDCval < 106) {
if (posBalance < 60) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (tcor < 94) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (tcor >= 94) {
if (refDCval < 93) {
if (Rs < 147) {
return 8;
}
else if (Rs >= 147) {
return 1;
}
}
else if (refDCval >= 93) {
return 8;
}
}
}
}
else if (posBalance >= 60) {
if (mcTcor < 98) {
if (SC < 308) {
return 1;
}
else if (SC >= 308) {
return 2;
}
}
else if (mcTcor >= 98) {
if (Cs < 77) {
return 2;
}
else if (Cs >= 77) {
if (Rs < 88) {
return 4;
}
else if (Rs >= 88) {
return 1;
}
}
}
}
}
else if (ssDCval >= 106) {
if (Cs < 123) {
return 2;
}
else if (Cs >= 123) {
return 1;
}
}
}
else if (SC >= 341) {
if (mu_mv_mag_sq < 7) {
if (Rs < 257) {
if (Rs < 227) {
return 4;
}
else if (Rs >= 227) {
return 2;
}
}
else if (Rs >= 257) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
}
}
else if (Cs >= 163) {
if (mu_mv_mag_sq < 70) {
if (refDCval < 98) {
if (tcor < 83) {
if (AFD < 14) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (AFD >= 14) {
if (MVDiff < 73) {
if (diffRsCsdiff < -3) {
return 2;
}
else if (diffRsCsdiff >= -3) {
return 1;
}
}
else if (MVDiff >= 73) {
if (CsDiff < 4) {
if (Cs < 224) {
if (Rs < 182) {
return 2;
}
else if (Rs >= 182) {
return 1;
}
}
else if (Cs >= 224) {
return 2;
}
}
else if (CsDiff >= 4) {
if (Rs < 216) {
return 8;
}
else if (Rs >= 216) {
if (Rs < 264) {
return 2;
}
else if (Rs >= 264) {
return 4;
}
}
}
}
}
}
else if (tcor >= 83) {
if (SC < 699) {
return 2;
}
else if (SC >= 699) {
return 8;
}
}
}
else if (refDCval >= 98) {
if (Cs < 360) {
if (Cs < 172) {
if (Rs < 267) {
return 2;
}
else if (Rs >= 267) {
return 8;
}
}
else if (Cs >= 172) {
if (Rs < 293) {
return 1;
}
else if (Rs >= 293) {
return 2;
}
}
}
else if (Cs >= 360) {
if (mu_mv_mag_sq < 9) {
if (Rs < 226) {
return 2;
}
else if (Rs >= 226) {
return 1;
}
}
else if (mu_mv_mag_sq >= 9) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (AFD < 30) {
if (Rs < 74) {
return 1;
}
else if (Rs >= 74) {
return 8;
}
}
else if (AFD >= 30) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (Rs < 1) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
if (MVDiff < 164) {
return 8;
}
else if (MVDiff >= 164) {
return 1;
}
}
}
else if (Rs >= 1) {
if (refDCval < 145) {
if (negBalance < 74) {
if (Cs < 255) {
if (TSC < 5) {
if (AFD < 11) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 2;
}
}
else if (AFD >= 11) {
if (Rs < 224) {
return 1;
}
else if (Rs >= 224) {
return 2;
}
}
}
else if (TSC >= 5) {
if (Cs < 186) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (diffRsCsdiff < -7) {
if (Rs < 182) {
return 1;
}
else if (Rs >= 182) {
return 8;
}
}
else if (diffRsCsdiff >= -7) {
if (Rs < 40) {
return 8;
}
else if (Rs >= 40) {
return 2;
}
}
}
}
else if (Cs >= 186) {
if (Rs < 116) {
return 8;
}
else if (Rs >= 116) {
return 4;
}
}
}
}
else if (Cs >= 255) {
if (SC < 645) {
if (gchDC < 4) {
if (ssDCval < 119) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (ssDCval >= 119) {
return 8;
}
}
else if (gchDC >= 4) {
return 2;
}
}
else if (SC >= 645) {
if (ssDCval < 113) {
return 8;
}
else if (ssDCval >= 113) {
return 2;
}
}
}
}
else if (negBalance >= 74) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (refDCval >= 145) {
if (Rs < 254) {
if (MVDiff < 492) {
if (AFD < 24) {
if (MVDiff < 19) {
if (Cs < 327) {
return 1;
}
else if (Cs >= 327) {
return 2;
}
}
else if (MVDiff >= 19) {
return 2;
}
}
else if (AFD >= 24) {
return 1;
}
}
else if (MVDiff >= 492) {
return 1;
}
}
else if (Rs >= 254) {
if (AFD < 11) {
if (Rs < 282) {
return 8;
}
else if (Rs >= 282) {
return 4;
}
}
else if (AFD >= 11) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 307) {
if (RsDiff < 4) {
if (refDCval < 74) {
if (mu_mv_mag_sq < 101) {
if (TSCindex < 2) {
if (posBalance < 29) {
if (Rs < 699) {
return 8;
}
else if (Rs >= 699) {
return 4;
}
}
else if (posBalance >= 29) {
if (Rs < 818) {
if (mu_mv_mag_sq < 80) {
if (Rs < 490) {
return 8;
}
else if (Rs >= 490) {
return 4;
}
}
else if (mu_mv_mag_sq >= 80) {
return 4;
}
}
else if (Rs >= 818) {
if (Cs < 302) {
return 4;
}
else if (Cs >= 302) {
return 1;
}
}
}
}
else if (TSCindex >= 2) {
if (Cs < 127) {
if (AFD < 5) {
if (diffAFD < -1) {
if (Rs < 459) {
return 8;
}
else if (Rs >= 459) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (diffAFD >= -1) {
if (AFD < 3) {
if (Rs < 1155) {
return 4;
}
else if (Rs >= 1155) {
return 8;
}
}
else if (AFD >= 3) {
if (mcTcor < 99) {
if (TSC < 3) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (mcTcor >= 99) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
}
else if (AFD >= 5) {
if (refDCval < 69) {
if (tcor < 81) {
if (RsDiff < 3) {
if (ssDCval < 47) {
if (Rs < 839) {
return 1;
}
else if (Rs >= 839) {
return 2;
}
}
else if (ssDCval >= 47) {
return 8;
}
}
else if (RsDiff >= 3) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (diffMVdiffVal < 8) {
if (Rs < 414) {
return 2;
}
else if (Rs >= 414) {
return 1;
}
}
else if (diffMVdiffVal >= 8) {
return 8;
}
}
}
}
else if (tcor >= 81) {
if (posBalance < 15) {
return 8;
}
else if (posBalance >= 15) {
if (Cs < 114) {
return 1;
}
else if (Cs >= 114) {
if (MVDiff < 38) {
return 1;
}
else if (MVDiff >= 38) {
return 8;
}
}
}
}
}
else if (refDCval >= 69) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 1;
}
}
}
}
else if (Cs >= 127) {
if (mcTcor < 97) {
if (ssDCval < 58) {
if (negBalance < 40) {
if (SC < 697) {
if (refDCval < 46) {
if (Rs < 396) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Rs >= 396) {
if (Rs < 423) {
return 4;
}
else if (Rs >= 423) {
return 2;
}
}
}
else if (refDCval >= 46) {
if (SC < 480) {
return 2;
}
else if (SC >= 480) {
if (negBalance < 36) {
return 4;
}
else if (negBalance >= 36) {
if (SC < 662) {
return 1;
}
else if (SC >= 662) {
return 4;
}
}
}
}
}
else if (SC >= 697) {
if (Cs < 347) {
if (tcor < 78) {
if (Rs < 619) {
return 8;
}
else if (Rs >= 619) {
return 2;
}
}
else if (tcor >= 78) {
if (posBalance < 70) {
if (posBalance < 39) {
if (Cs < 295) {
return 1;
}
else if (Cs >= 295) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
}
else if (posBalance >= 39) {
return 1;
}
}
else if (posBalance >= 70) {
return 2;
}
}
}
else if (Cs >= 347) {
if (SC < 1239) {
if (mu_mv_mag_sq < 70) {
if (TSC < 9) {
if (MVDiff < 4) {
if (Rs < 553) {
return 4;
}
else if (Rs >= 553) {
return 8;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 45) {
return 1;
}
else if (ssDCval >= 45) {
if (TSC < 5) {
if (diffTSC < -1) {
if (MVDiff < 33) {
return 1;
}
else if (MVDiff >= 33) {
return 4;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (TSC >= 5) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
return 1;
}
}
}
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (mu_mv_mag_sq >= 70) {
if (Cs < 526) {
return 1;
}
else if (Cs >= 526) {
return 8;
}
}
}
else if (SC >= 1239) {
if (ssDCval < 55) {
if (tcor < 94) {
if (Cs < 468) {
return 8;
}
else if (Cs >= 468) {
return 2;
}
}
else if (tcor >= 94) {
return 1;
}
}
else if (ssDCval >= 55) {
return 4;
}
}
}
}
}
else if (negBalance >= 40) {
if (Rs < 535) {
if (TSC < 6) {
if (mcTcor < 90) {
if (posBalance < 51) {
if (posBalance < 48) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Cs < 347) {
return 8;
}
else if (Cs >= 347) {
return 1;
}
}
}
else if (posBalance >= 48) {
return 2;
}
}
else if (posBalance >= 51) {
if (mcTcor < 87) {
if (diffRsCsdiff < -16) {
return 2;
}
else if (diffRsCsdiff >= -16) {
return 4;
}
}
else if (mcTcor >= 87) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
}
}
else if (mcTcor >= 90) {
if (SC < 736) {
if (posBalance < 55) {
if (tcor < 89) {
if (ssDCval < 42) {
return 1;
}
else if (ssDCval >= 42) {
return 4;
}
}
else if (tcor >= 89) {
if (Rs < 410) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (Rs >= 410) {
return 4;
}
}
}
else if (posBalance >= 55) {
return 1;
}
}
else if (SC >= 736) {
if (AFD < 11) {
if (Rs < 393) {
return 2;
}
else if (Rs >= 393) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (AFD >= 11) {
if (MVDiff < 53) {
return 1;
}
else if (MVDiff >= 53) {
return 2;
}
}
}
}
}
else if (TSC >= 6) {
if (MVDiff < 43) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
return 4;
}
}
else if (MVDiff >= 43) {
if (mcTcor < 70) {
if (Rs < 360) {
return 8;
}
else if (Rs >= 360) {
return 4;
}
}
else if (mcTcor >= 70) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 1;
}
}
}
}
}
else if (Rs >= 535) {
if (Cs < 268) {
if (MVDiff < 69) {
return 1;
}
else if (MVDiff >= 69) {
return 2;
}
}
else if (Cs >= 268) {
if (negBalance < 52) {
if (MVDiff < 15) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (MVDiff >= 15) {
if (MVDiff < 17) {
if (Cs < 501) {
return 8;
}
else if (Cs >= 501) {
return 4;
}
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (negBalance >= 52) {
if (Cs < 496) {
return 1;
}
else if (Cs >= 496) {
if (Rs < 582) {
return 4;
}
else if (Rs >= 582) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 58) {
if (ssDCval < 61) {
if (Rs < 717) {
if (MVDiff < 16) {
if (SC < 793) {
if (SC < 775) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
if (Rs < 364) {
return 4;
}
else if (Rs >= 364) {
return 2;
}
}
}
else if (SC >= 775) {
if (mu_mv_mag_sq < 56) {
return 2;
}
else if (mu_mv_mag_sq >= 56) {
return 1;
}
}
}
else if (SC >= 793) {
if (negBalance < 29) {
if (Cs < 477) {
return 2;
}
else if (Cs >= 477) {
return 8;
}
}
else if (negBalance >= 29) {
return 2;
}
}
}
else if (MVDiff >= 16) {
if (SC < 678) {
if (Cs < 217) {
return 4;
}
else if (Cs >= 217) {
return 2;
}
}
else if (SC >= 678) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
if (Rs < 577) {
return 2;
}
else if (Rs >= 577) {
return 1;
}
}
}
}
}
else if (Rs >= 717) {
if (mu_mv_mag_sq < 64) {
if (AFD < 9) {
if (mu_mv_mag_sq < 40) {
return 1;
}
else if (mu_mv_mag_sq >= 40) {
if (Rs < 1163) {
return 8;
}
else if (Rs >= 1163) {
return 1;
}
}
}
else if (AFD >= 9) {
if (Rs < 804) {
return 4;
}
else if (Rs >= 804) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 64) {
return 4;
}
}
}
else if (ssDCval >= 61) {
if (negBalance < 41) {
if (Rs < 592) {
if (RsCsDiff < 7) {
if (mu_mv_mag_sq < 2) {
if (Cs < 326) {
return 1;
}
else if (Cs >= 326) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
if (posBalance < 74) {
if (negBalance < 34) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (negBalance >= 34) {
return 2;
}
}
else if (posBalance >= 74) {
if (Rs < 443) {
return 8;
}
else if (Rs >= 443) {
return 1;
}
}
}
}
else if (RsCsDiff >= 7) {
if (Rs < 515) {
if (ssDCval < 65) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (ssDCval >= 65) {
return 8;
}
}
else if (Rs >= 515) {
return 4;
}
}
}
else if (Rs >= 592) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 796) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
if (TSC < 4) {
if (Rs < 658) {
return 4;
}
else if (Rs >= 658) {
return 8;
}
}
else if (TSC >= 4) {
return 4;
}
}
}
else if (Rs >= 796) {
if (negBalance < 21) {
if (Rs < 1083) {
return 1;
}
else if (Rs >= 1083) {
return 2;
}
}
else if (negBalance >= 21) {
return 1;
}
}
}
}
}
else if (negBalance >= 41) {
if (negBalance < 46) {
if (posBalance < 61) {
if (diffAFD < -1) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (diffAFD >= -1) {
if (negBalance < 45) {
return 4;
}
else if (negBalance >= 45) {
return 1;
}
}
}
else if (posBalance >= 61) {
if (SC < 889) {
return 1;
}
else if (SC >= 889) {
if (Rs < 503) {
return 1;
}
else if (Rs >= 503) {
return 4;
}
}
}
}
else if (negBalance >= 46) {
if (tcor < 91) {
if (MVDiff < 63) {
if (ssDCval < 70) {
if (SC < 1007) {
if (Rs < 320) {
return 2;
}
else if (Rs >= 320) {
if (Cs < 440) {
return 4;
}
else if (Cs >= 440) {
if (Rs < 514) {
return 2;
}
else if (Rs >= 514) {
return 4;
}
}
}
}
else if (SC >= 1007) {
if (Rs < 991) {
return 2;
}
else if (Rs >= 991) {
return 4;
}
}
}
else if (ssDCval >= 70) {
if (AFD < 9) {
if (Rs < 537) {
return 1;
}
else if (Rs >= 537) {
return 4;
}
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (MVDiff >= 63) {
if (MVDiff < 97) {
return 2;
}
else if (MVDiff >= 97) {
if (Rs < 326) {
return 8;
}
else if (Rs >= 326) {
return 4;
}
}
}
}
else if (tcor >= 91) {
if (RsCsDiff < 9) {
return 2;
}
else if (RsCsDiff >= 9) {
if (Rs < 872) {
return 1;
}
else if (Rs >= 872) {
return 8;
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (tcor < 98) {
if (Rs < 611) {
if (TSC < 4) {
if (mu_mv_mag_sq < 24) {
if (RsCsDiff < 5) {
if (MVDiff < 21) {
if (Cs < 282) {
if (Rs < 317) {
return 2;
}
else if (Rs >= 317) {
return 1;
}
}
else if (Cs >= 282) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
return 1;
}
}
}
else if (MVDiff >= 21) {
return 2;
}
}
else if (RsCsDiff >= 5) {
if (posBalance < 70) {
if (refDCval < 73) {
if (ssDCval < 56) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
return 1;
}
}
else if (ssDCval >= 56) {
return 4;
}
}
else if (refDCval >= 73) {
return 1;
}
}
else if (posBalance >= 70) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (MVDiff < 4) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (Cs < 463) {
return 1;
}
else if (Cs >= 463) {
if (AFD < 4) {
if (Cs < 485) {
return 2;
}
else if (Cs >= 485) {
return 1;
}
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (refDCval < 45) {
if (posBalance < 36) {
return 2;
}
else if (posBalance >= 36) {
return 1;
}
}
else if (refDCval >= 45) {
if (ssDCval < 57) {
if (diffMVdiffVal < 5) {
if (Cs < 268) {
if (mcTcor < 99) {
return 4;
}
else if (mcTcor >= 99) {
if (Rs < 433) {
return 4;
}
else if (Rs >= 433) {
return 1;
}
}
}
else if (Cs >= 268) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= 5) {
if (Cs < 349) {
if (mu_mv_mag_sq < 52) {
return 4;
}
else if (mu_mv_mag_sq >= 52) {
return 1;
}
}
else if (Cs >= 349) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (Cs < 375) {
if (posBalance < 51) {
if (Cs < 344) {
return 8;
}
else if (Cs >= 344) {
return 1;
}
}
else if (posBalance >= 51) {
if (posBalance < 53) {
if (Cs < 292) {
return 2;
}
else if (Cs >= 292) {
return 1;
}
}
else if (posBalance >= 53) {
return 1;
}
}
}
else if (Cs >= 375) {
if (Rs < 397) {
return 4;
}
else if (Rs >= 397) {
if (Cs < 391) {
return 1;
}
else if (Cs >= 391) {
return 2;
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (MVDiff < 6) {
if (Rs < 415) {
return 8;
}
else if (Rs >= 415) {
return 2;
}
}
else if (MVDiff >= 6) {
if (Rs < 571) {
return 1;
}
else if (Rs >= 571) {
if (Cs < 311) {
return 1;
}
else if (Cs >= 311) {
return 8;
}
}
}
}
}
else if (Rs >= 611) {
if (RsDiff < 2) {
if (negBalance < 40) {
if (Rs < 1522) {
if (ssDCval < 49) {
if (Rs < 975) {
return 2;
}
else if (Rs >= 975) {
return 1;
}
}
else if (ssDCval >= 49) {
if (negBalance < 19) {
if (posBalance < 37) {
return 4;
}
else if (posBalance >= 37) {
return 1;
}
}
else if (negBalance >= 19) {
return 1;
}
}
}
else if (Rs >= 1522) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 1;
}
}
}
}
else if (negBalance >= 40) {
if (Cs < 323) {
if (SC < 880) {
return 4;
}
else if (SC >= 880) {
return 1;
}
}
else if (Cs >= 323) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (negBalance < 30) {
if (SC < 2295) {
if (tcor < 94) {
if (ssDCval < 58) {
return 2;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (tcor >= 94) {
if (MVDiff < 7) {
if (Rs < 683) {
return 4;
}
else if (Rs >= 683) {
return 2;
}
}
else if (MVDiff >= 7) {
if (Rs < 890) {
return 2;
}
else if (Rs >= 890) {
return 1;
}
}
}
}
else if (SC >= 2295) {
if (Cs < 421) {
return 2;
}
else if (Cs >= 421) {
return 4;
}
}
}
else if (negBalance >= 30) {
if (posBalance < 57) {
if (ssDCval < 55) {
if (posBalance < 41) {
if (Rs < 1638) {
return 2;
}
else if (Rs >= 1638) {
return 4;
}
}
else if (posBalance >= 41) {
return 1;
}
}
else if (ssDCval >= 55) {
if (tcor < 87) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
else if (tcor >= 87) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 726) {
return 8;
}
else if (Rs >= 726) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 57) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (Rs < 709) {
return 8;
}
else if (Rs >= 709) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 98) {
if (mu_mv_mag_sq < 32) {
if (MVDiff < 3) {
if (Rs < 374) {
return 8;
}
else if (Rs >= 374) {
if (Cs < 447) {
return 1;
}
else if (Cs >= 447) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 40) {
return 1;
}
else if (ssDCval >= 40) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (Rs < 388) {
if (TSC < 3) {
if (Cs < 210) {
return 1;
}
else if (Cs >= 210) {
if (mcTcor < 99) {
if (Rs < 317) {
return 4;
}
else if (Rs >= 317) {
return 2;
}
}
else if (mcTcor >= 99) {
return 4;
}
}
}
else if (TSC >= 3) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (Cs < 366) {
return 8;
}
else if (Cs >= 366) {
return 2;
}
}
}
}
else if (Rs >= 388) {
if (Rs < 766) {
if (ssDCval < 54) {
if (ssDCval < 47) {
if (Cs < 392) {
return 1;
}
else if (Cs >= 392) {
return 8;
}
}
else if (ssDCval >= 47) {
if (mu_mv_mag_sq < 66) {
return 1;
}
else if (mu_mv_mag_sq >= 66) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 4;
}
}
}
}
else if (ssDCval >= 54) {
if (RsCsDiff < 1) {
if (ssDCval < 68) {
if (Cs < 432) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 2;
}
}
else if (Cs >= 432) {
if (Rs < 563) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 563) {
return 1;
}
}
}
else if (ssDCval >= 68) {
if (negBalance < 23) {
if (Rs < 546) {
return 4;
}
else if (Rs >= 546) {
return 2;
}
}
else if (negBalance >= 23) {
return 8;
}
}
}
else if (RsCsDiff >= 1) {
if (diffAFD < -2) {
if (Rs < 542) {
return 2;
}
else if (Rs >= 542) {
return 1;
}
}
else if (diffAFD >= -2) {
if (RsCsDiff < 2) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
if (Cs < 255) {
return 8;
}
else if (Cs >= 255) {
return 1;
}
}
}
else if (RsCsDiff >= 2) {
if (diffMVdiffVal < 2) {
if (Rs < 429) {
return 4;
}
else if (Rs >= 429) {
if (RsDiff < 2) {
if (negBalance < 81) {
return 8;
}
else if (negBalance >= 81) {
if (Rs < 612) {
return 4;
}
else if (Rs >= 612) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 766) {
if (refDCval < 69) {
if (negBalance < 23) {
if (mu_mv_mag_sq < 58) {
return 4;
}
else if (mu_mv_mag_sq >= 58) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
}
else if (negBalance >= 23) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 1048) {
return 2;
}
else if (Rs >= 1048) {
return 1;
}
}
}
}
else if (refDCval >= 69) {
if (negBalance < 32) {
if (Cs < 452) {
return 8;
}
else if (Cs >= 452) {
return 1;
}
}
else if (negBalance >= 32) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (Cs < 192) {
if (Rs < 1724) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
if (Rs < 784) {
return 2;
}
else if (Rs >= 784) {
return 1;
}
}
}
else if (Rs >= 1724) {
return 2;
}
}
else if (Cs >= 192) {
if (refDCval < 60) {
if (Rs < 418) {
if (negBalance < 21) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (Cs < 365) {
return 4;
}
else if (Cs >= 365) {
return 8;
}
}
}
else if (negBalance >= 21) {
if (ssDCval < 53) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
else if (ssDCval >= 53) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 4;
}
}
}
}
else if (Rs >= 418) {
if (ssDCval < 56) {
if (tcor < 100) {
if (TSCindex < 3) {
if (mu_mv_mag_sq < 128) {
if (Cs < 299) {
if (Rs < 845) {
return 8;
}
else if (Rs >= 845) {
return 1;
}
}
else if (Cs >= 299) {
return 1;
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1335) {
return 8;
}
else if (Rs >= 1335) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (Rs < 869) {
if (Rs < 553) {
if (Cs < 489) {
return 1;
}
else if (Cs >= 489) {
return 4;
}
}
else if (Rs >= 553) {
return 4;
}
}
else if (Rs >= 869) {
if (Cs < 208) {
return 8;
}
else if (Cs >= 208) {
if (Rs < 1005) {
return 8;
}
else if (Rs >= 1005) {
return 1;
}
}
}
}
}
else if (tcor >= 100) {
if (SC < 1148) {
return 8;
}
else if (SC >= 1148) {
return 1;
}
}
}
else if (ssDCval >= 56) {
if (Rs < 475) {
if (Cs < 382) {
return 4;
}
else if (Cs >= 382) {
return 1;
}
}
else if (Rs >= 475) {
if (Cs < 421) {
if (MVDiff < 1) {
if (mu_mv_mag_sq < 123) {
if (Rs < 721) {
if (Rs < 538) {
return 1;
}
else if (Rs >= 538) {
return 8;
}
}
else if (Rs >= 721) {
if (Cs < 357) {
return 2;
}
else if (Cs >= 357) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
else if (MVDiff >= 1) {
if (Rs < 794) {
return 2;
}
else if (Rs >= 794) {
return 1;
}
}
}
else if (Cs >= 421) {
return 8;
}
}
}
}
}
else if (refDCval >= 60) {
if (ssDCval < 69) {
if (Cs < 484) {
if (refDCval < 61) {
if (posBalance < 41) {
if (Rs < 1005) {
return 4;
}
else if (Rs >= 1005) {
return 8;
}
}
else if (posBalance >= 41) {
return 2;
}
}
else if (refDCval >= 61) {
if (SC < 714) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (SC >= 714) {
if (negBalance < 37) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
if (posBalance < 46) {
if (ssDCval < 64) {
if (Rs < 1284) {
if (Cs < 421) {
return 1;
}
else if (Cs >= 421) {
return 8;
}
}
else if (Rs >= 1284) {
if (Cs < 481) {
return 2;
}
else if (Cs >= 481) {
return 1;
}
}
}
else if (ssDCval >= 64) {
if (Rs < 956) {
if (MVDiff < 1) {
if (mu_mv_mag_sq < 116) {
if (Rs < 554) {
return 8;
}
else if (Rs >= 554) {
return 1;
}
}
else if (mu_mv_mag_sq >= 116) {
return 1;
}
}
else if (MVDiff >= 1) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
if (Rs < 875) {
return 2;
}
else if (Rs >= 875) {
return 8;
}
}
}
}
}
else if (Rs >= 956) {
return 1;
}
}
}
else if (posBalance >= 46) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 1;
}
}
}
}
else if (negBalance >= 37) {
if (Rs < 1097) {
if (Rs < 546) {
return 1;
}
else if (Rs >= 546) {
return 4;
}
}
else if (Rs >= 1097) {
return 1;
}
}
}
}
}
else if (Cs >= 484) {
if (negBalance < 22) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
if (tcor < 100) {
return 1;
}
else if (tcor >= 100) {
return 8;
}
}
}
else if (negBalance >= 22) {
if (Rs < 512) {
if (Rs < 406) {
return 1;
}
else if (Rs >= 406) {
return 2;
}
}
else if (Rs >= 512) {
if (AFD < 5) {
if (posBalance < 26) {
return 2;
}
else if (posBalance >= 26) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (mu_mv_mag_sq < 114) {
return 1;
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
}
}
else if (AFD >= 5) {
if (Rs < 1037) {
return 4;
}
else if (Rs >= 1037) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 601) {
if (SC < 955) {
if (SC < 765) {
return 8;
}
else if (SC >= 765) {
if (posBalance < 46) {
if (Rs < 524) {
return 1;
}
else if (Rs >= 524) {
return 2;
}
}
else if (posBalance >= 46) {
return 2;
}
}
}
else if (SC >= 955) {
return 1;
}
}
else if (Rs >= 601) {
if (negBalance < 35) {
if (Rs < 895) {
if (MVDiff < 1) {
if (refDCval < 70) {
if (Rs < 655) {
if (Rs < 629) {
return 1;
}
else if (Rs >= 629) {
return 2;
}
}
else if (Rs >= 655) {
return 4;
}
}
else if (refDCval >= 70) {
return 4;
}
}
else if (MVDiff >= 1) {
if (ssDCval < 71) {
if (Cs < 311) {
return 4;
}
else if (Cs >= 311) {
return 2;
}
}
else if (ssDCval >= 71) {
if (Cs < 405) {
return 8;
}
else if (Cs >= 405) {
return 4;
}
}
}
}
else if (Rs >= 895) {
if (Rs < 1009) {
return 8;
}
else if (Rs >= 1009) {
if (Rs < 1675) {
return 1;
}
else if (Rs >= 1675) {
return 2;
}
}
}
}
else if (negBalance >= 35) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 74) {
if (Cs < 285) {
if (refDCval < 107) {
if (mu_mv_mag_sq < 104) {
if (Rs < 1610) {
if (RsCsDiff < 2) {
if (negBalance < 36) {
if (mu_mv_mag_sq < 79) {
if (posBalance < 51) {
if (MVDiff < 3) {
if (Cs < 209) {
return 8;
}
else if (Cs >= 209) {
return 1;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (posBalance >= 51) {
if (Cs < 228) {
return 8;
}
else if (Cs >= 228) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (posBalance < 57) {
if (Cs < 178) {
return 1;
}
else if (Cs >= 178) {
if (mu_mv_mag_sq < 96) {
if (ssDCval < 81) {
if (Rs < 936) {
return 8;
}
else if (Rs >= 936) {
return 1;
}
}
else if (ssDCval >= 81) {
if (Rs < 1085) {
return 2;
}
else if (Rs >= 1085) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (negBalance < 32) {
return 1;
}
else if (negBalance >= 32) {
return 8;
}
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (negBalance >= 36) {
if (SC < 1423) {
if (mu_mv_mag_sq < 24) {
if (SC < 493) {
if (Rs < 321) {
return 1;
}
else if (Rs >= 321) {
return 8;
}
}
else if (SC >= 493) {
return 1;
}
}
else if (mu_mv_mag_sq >= 24) {
if (SC < 390) {
return 1;
}
else if (SC >= 390) {
return 8;
}
}
}
else if (SC >= 1423) {
if (Cs < 112) {
return 1;
}
else if (Cs >= 112) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
if (Cs < 176) {
if (tcor < 77) {
if (Rs < 326) {
return 8;
}
else if (Rs >= 326) {
if (Cs < 42) {
return 1;
}
else if (Cs >= 42) {
return 2;
}
}
}
else if (tcor >= 77) {
if (mcTcor < 99) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (Cs < 134) {
if (posBalance < 33) {
if (Rs < 714) {
return 8;
}
else if (Rs >= 714) {
return 1;
}
}
else if (posBalance >= 33) {
return 1;
}
}
else if (Cs >= 134) {
if (Rs < 589) {
return 1;
}
else if (Rs >= 589) {
return 4;
}
}
}
}
}
else if (Cs >= 176) {
if (negBalance < 30) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (negBalance >= 30) {
if (Rs < 470) {
if (diffRsCsdiff < -2) {
if (negBalance < 54) {
return 2;
}
else if (negBalance >= 54) {
if (Cs < 249) {
return 4;
}
else if (Cs >= 249) {
return 2;
}
}
}
else if (diffRsCsdiff >= -2) {
if (Cs < 272) {
if (mu_mv_mag_sq < 36) {
return 1;
}
else if (mu_mv_mag_sq >= 36) {
return 8;
}
}
else if (Cs >= 272) {
if (Cs < 282) {
return 2;
}
else if (Cs >= 282) {
return 8;
}
}
}
}
else if (Rs >= 470) {
if (gchDC < 4) {
if (refDCval < 81) {
if (SC < 1089) {
if (gchDC < 3) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (Rs < 529) {
return 4;
}
else if (Rs >= 529) {
return 1;
}
}
}
else if (gchDC >= 3) {
return 1;
}
}
else if (SC >= 1089) {
return 1;
}
}
else if (refDCval >= 81) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (gchDC >= 4) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1610) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
if (Rs < 1942) {
return 1;
}
else if (Rs >= 1942) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (ssDCval < 80) {
if (SC < 1267) {
return 8;
}
else if (SC >= 1267) {
return 1;
}
}
else if (ssDCval >= 80) {
if (ssDCval < 94) {
if (Cs < 251) {
if (Rs < 313) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (Rs >= 313) {
if (Rs < 1535) {
if (SC < 508) {
if (Cs < 189) {
return 2;
}
else if (Cs >= 189) {
return 1;
}
}
else if (SC >= 508) {
return 2;
}
}
else if (Rs >= 1535) {
return 1;
}
}
}
else if (Cs >= 251) {
if (Rs < 1034) {
return 1;
}
else if (Rs >= 1034) {
return 2;
}
}
}
else if (ssDCval >= 94) {
return 1;
}
}
}
}
else if (refDCval >= 107) {
if (MVDiff < 23) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (Rs < 841) {
return 2;
}
else if (Rs >= 841) {
return 4;
}
}
}
else if (MVDiff >= 23) {
if (Rs < 452) {
if (ssDCval < 116) {
if (Rs < 384) {
return 8;
}
else if (Rs >= 384) {
return 4;
}
}
else if (ssDCval >= 116) {
if (diffMVdiffVal < 33) {
return 2;
}
else if (diffMVdiffVal >= 33) {
return 4;
}
}
}
else if (Rs >= 452) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
if (SC < 515) {
return 8;
}
else if (SC >= 515) {
return 1;
}
}
}
}
}
}
else if (Cs >= 285) {
if (refDCval < 115) {
if (mcTcor < 85) {
if (MVDiff < 20) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
return 2;
}
}
else if (MVDiff >= 20) {
if (Rs < 323) {
return 4;
}
else if (Rs >= 323) {
if (RsCsDiff < 4) {
return 8;
}
else if (RsCsDiff >= 4) {
return 2;
}
}
}
}
else if (mcTcor >= 85) {
if (mu_mv_mag_sq < 46) {
if (mcTcor < 92) {
if (Rs < 313) {
if (Cs < 317) {
return 8;
}
else if (Cs >= 317) {
return 1;
}
}
else if (Rs >= 313) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (diffRsCsdiff < 6) {
return 8;
}
else if (diffRsCsdiff >= 6) {
return 2;
}
}
}
}
else if (mcTcor >= 92) {
if (tcor < 98) {
if (TSC < 6) {
if (mcTcor < 96) {
if (MVDiff < 3) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (diffTSC < -1) {
if (ssDCval < 110) {
return 8;
}
else if (ssDCval >= 110) {
return 2;
}
}
else if (diffTSC >= -1) {
if (TSC < 5) {
if (Rs < 567) {
return 1;
}
else if (Rs >= 567) {
return 8;
}
}
else if (TSC >= 5) {
return 1;
}
}
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 111) {
if (CsDiff < 1) {
if (Rs < 882) {
return 1;
}
else if (Rs >= 882) {
return 4;
}
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (refDCval >= 111) {
if (Rs < 588) {
return 1;
}
else if (Rs >= 588) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (refDCval < 86) {
if (SC < 803) {
return 8;
}
else if (SC >= 803) {
return 1;
}
}
else if (refDCval >= 86) {
if (diffRsCsdiff < 4) {
if (Rs < 322) {
return 2;
}
else if (Rs >= 322) {
return 8;
}
}
else if (diffRsCsdiff >= 4) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (Rs < 1050) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (Rs >= 1050) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (Rs < 1085) {
if (ssDCval < 96) {
if (negBalance < 33) {
if (ssDCval < 94) {
if (refDCval < 78) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 485) {
return 8;
}
else if (Rs >= 485) {
return 1;
}
}
}
else if (refDCval >= 78) {
if (MVDiff < 3) {
if (Rs < 690) {
if (Rs < 440) {
return 4;
}
else if (Rs >= 440) {
return 1;
}
}
else if (Rs >= 690) {
if (Cs < 425) {
return 8;
}
else if (Cs >= 425) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (posBalance < 52) {
if (Rs < 435) {
return 4;
}
else if (Rs >= 435) {
return 1;
}
}
else if (posBalance >= 52) {
if (Rs < 356) {
return 2;
}
else if (Rs >= 356) {
return 8;
}
}
}
}
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (negBalance >= 33) {
if (tcor < 98) {
if (negBalance < 41) {
if (ssDCval < 78) {
if (Rs < 699) {
return 8;
}
else if (Rs >= 699) {
return 4;
}
}
else if (ssDCval >= 78) {
if (mu_mv_mag_sq < 83) {
if (ssDCval < 84) {
if (Cs < 291) {
return 8;
}
else if (Cs >= 291) {
return 2;
}
}
else if (ssDCval >= 84) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 83) {
return 1;
}
}
}
else if (negBalance >= 41) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (ssDCval < 83) {
if (MVDiff < 8) {
if (Rs < 490) {
return 4;
}
else if (Rs >= 490) {
return 2;
}
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (ssDCval >= 83) {
return 8;
}
}
}
}
else if (tcor >= 98) {
if (Cs < 412) {
return 8;
}
else if (Cs >= 412) {
if (Rs < 717) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 8;
}
}
}
else if (Rs >= 717) {
return 8;
}
}
}
}
}
else if (ssDCval >= 96) {
if (refDCval < 98) {
if (Cs < 457) {
return 4;
}
else if (Cs >= 457) {
if (RsCsDiff < 1) {
return 4;
}
else if (RsCsDiff >= 1) {
return 8;
}
}
}
else if (refDCval >= 98) {
if (Rs < 384) {
if (Rs < 381) {
return 1;
}
else if (Rs >= 381) {
return 2;
}
}
else if (Rs >= 384) {
return 8;
}
}
}
}
else if (Rs >= 1085) {
if (TSC < 3) {
if (Cs < 492) {
if (ssDCval < 103) {
if (Rs < 2081) {
if (Rs < 1134) {
return 2;
}
else if (Rs >= 1134) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (Rs >= 2081) {
if (Rs < 2158) {
return 8;
}
else if (Rs >= 2158) {
return 2;
}
}
}
else if (ssDCval >= 103) {
if (Rs < 1769) {
return 8;
}
else if (Rs >= 1769) {
return 1;
}
}
}
else if (Cs >= 492) {
if (mu_mv_mag_sq < 99) {
if (Rs < 1998) {
return 1;
}
else if (Rs >= 1998) {
return 8;
}
}
else if (mu_mv_mag_sq >= 99) {
return 4;
}
}
}
else if (TSC >= 3) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 112) {
if (tcor < 98) {
if (ssDCval < 81) {
if (Rs < 1419) {
return 1;
}
else if (Rs >= 1419) {
return 8;
}
}
else if (ssDCval >= 81) {
if (Rs < 1302) {
return 8;
}
else if (Rs >= 1302) {
return 1;
}
}
}
else if (tcor >= 98) {
if (Cs < 306) {
if (Rs < 1786) {
return 8;
}
else if (Rs >= 1786) {
return 1;
}
}
else if (Cs >= 306) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
return 4;
}
}
else if (posBalance >= 54) {
if (Rs < 1483) {
return 8;
}
else if (Rs >= 1483) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 115) {
if (mcTcor < 99) {
if (MVDiff < 4) {
if (ssDCval < 160) {
if (Rs < 1255) {
if (Rs < 379) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 379) {
return 8;
}
}
else if (Rs >= 1255) {
if (Rs < 1839) {
return 4;
}
else if (Rs >= 1839) {
if (Rs < 2138) {
return 1;
}
else if (Rs >= 2138) {
return 8;
}
}
}
}
else if (ssDCval >= 160) {
return 1;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 130) {
if (Rs < 492) {
if (posBalance < 62) {
if (AFD < 4) {
if (ssDCval < 121) {
return 2;
}
else if (ssDCval >= 121) {
return 4;
}
}
else if (AFD >= 4) {
if (tcor < 98) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
if (Rs < 405) {
return 2;
}
else if (Rs >= 405) {
return 8;
}
}
}
else if (tcor >= 98) {
return 1;
}
}
}
else if (posBalance >= 62) {
return 4;
}
}
else if (Rs >= 492) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (CsDiff < 2) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 8;
}
}
else if (CsDiff >= 2) {
if (gchDC < 2) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (Rs < 1452) {
return 1;
}
else if (Rs >= 1452) {
return 2;
}
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
}
else if (ssDCval >= 130) {
if (RsCsDiff < 7) {
if (ssDCval < 154) {
if (Cs < 383) {
if (Rs < 353) {
return 8;
}
else if (Rs >= 353) {
return 4;
}
}
else if (Cs >= 383) {
return 8;
}
}
else if (ssDCval >= 154) {
if (Cs < 317) {
return 2;
}
else if (Cs >= 317) {
return 4;
}
}
}
else if (RsCsDiff >= 7) {
if (Rs < 620) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (Rs < 364) {
return 1;
}
else if (Rs >= 364) {
return 2;
}
}
}
else if (Rs >= 620) {
if (Cs < 403) {
return 4;
}
else if (Cs >= 403) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 122) {
if (AFD < 7) {
if (posBalance < 50) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
return 2;
}
}
else if (posBalance >= 50) {
return 2;
}
}
else if (AFD >= 7) {
if (Rs < 411) {
return 4;
}
else if (Rs >= 411) {
return 1;
}
}
}
else if (ssDCval >= 122) {
if (ssDCval < 155) {
if (posBalance < 21) {
return 8;
}
else if (posBalance >= 21) {
if (Rs < 369) {
return 8;
}
else if (Rs >= 369) {
return 1;
}
}
}
else if (ssDCval >= 155) {
return 4;
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (mcTcor < 94) {
if (SC < 1984) {
if (Cs < 341) {
if (tcor < 76) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 7) {
if (MVDiff < 137) {
if (SC < 831) {
if (Rs < 424) {
return 1;
}
else if (Rs >= 424) {
if (Rs < 501) {
if (Rs < 463) {
return 8;
}
else if (Rs >= 463) {
return 2;
}
}
else if (Rs >= 501) {
return 8;
}
}
}
else if (SC >= 831) {
return 2;
}
}
else if (MVDiff >= 137) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
if (refDCval < 112) {
if (CsDiff < 4) {
if (ssDCval < 45) {
if (SC < 1129) {
return 1;
}
else if (SC >= 1129) {
return 2;
}
}
else if (ssDCval >= 45) {
if (negBalance < 66) {
return 1;
}
else if (negBalance >= 66) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
else if (CsDiff >= 4) {
if (Cs < 324) {
if (MVDiff < 108) {
return 8;
}
else if (MVDiff >= 108) {
return 1;
}
}
else if (Cs >= 324) {
return 2;
}
}
}
else if (refDCval >= 112) {
if (negBalance < 62) {
if (Cs < 190) {
return 4;
}
else if (Cs >= 190) {
return 8;
}
}
else if (negBalance >= 62) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 1;
}
}
}
}
}
}
}
else if (posBalance >= 57) {
if (tcor < 37) {
if (ssDCval < 118) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (Rs < 851) {
if (ssDCval < 72) {
if (Rs < 384) {
return 8;
}
else if (Rs >= 384) {
return 4;
}
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (Rs >= 851) {
if (Rs < 1191) {
return 1;
}
else if (Rs >= 1191) {
if (Cs < 100) {
return 4;
}
else if (Cs >= 100) {
return 2;
}
}
}
}
}
else if (ssDCval >= 118) {
return 1;
}
}
else if (tcor >= 37) {
if (mu_mv_mag_sq < 43) {
if (diffAFD < 1) {
if (mcTcor < 83) {
if (Cs < 313) {
return 1;
}
else if (Cs >= 313) {
return 2;
}
}
else if (mcTcor >= 83) {
if (Rs < 772) {
if (MVDiff < 41) {
if (ssDCval < 102) {
if (Cs < 294) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
return 2;
}
}
else if (Cs >= 294) {
return 4;
}
}
else if (ssDCval >= 102) {
return 1;
}
}
else if (MVDiff >= 41) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (AFD < 32) {
return 4;
}
else if (AFD >= 32) {
return 2;
}
}
}
}
else if (Rs >= 772) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (Cs < 291) {
if (Cs < 269) {
if (diffMVdiffVal < 5) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 1;
}
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
else if (Cs >= 269) {
return 1;
}
}
else if (Cs >= 291) {
return 2;
}
}
}
}
}
else if (diffAFD >= 1) {
if (tcor < 53) {
if (Cs < 148) {
return 1;
}
else if (Cs >= 148) {
return 8;
}
}
else if (tcor >= 53) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (gchDC < 3) {
if (diffTSC < 1) {
if (SC < 824) {
return 1;
}
else if (SC >= 824) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (diffTSC >= 1) {
if (Cs < 261) {
return 2;
}
else if (Cs >= 261) {
return 1;
}
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (ssDCval < 61) {
if (SC < 760) {
if (Rs < 559) {
return 4;
}
else if (Rs >= 559) {
return 2;
}
}
else if (SC >= 760) {
return 1;
}
}
else if (ssDCval >= 61) {
if (mcTcor < 81) {
if (Rs < 670) {
return 8;
}
else if (Rs >= 670) {
return 2;
}
}
else if (mcTcor >= 81) {
return 2;
}
}
}
}
}
}
else if (tcor >= 76) {
if (ssDCval < 135) {
if (AFD < 17) {
if (diffAFD < -10) {
if (Rs < 1159) {
return 8;
}
else if (Rs >= 1159) {
return 1;
}
}
else if (diffAFD >= -10) {
if (mu_mv_mag_sq < 39) {
return 1;
}
else if (mu_mv_mag_sq >= 39) {
if (mu_mv_mag_sq < 45) {
if (Rs < 907) {
return 2;
}
else if (Rs >= 907) {
return 4;
}
}
else if (mu_mv_mag_sq >= 45) {
return 1;
}
}
}
}
else if (AFD >= 17) {
if (RsCsDiff < 56) {
if (Rs < 861) {
return 2;
}
else if (Rs >= 861) {
return 1;
}
}
else if (RsCsDiff >= 56) {
if (Cs < 256) {
return 2;
}
else if (Cs >= 256) {
return 1;
}
}
}
}
else if (ssDCval >= 135) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 1;
}
}
}
}
else if (Cs >= 341) {
if (posBalance < 55) {
if (Rs < 1124) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (AFD < 14) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 64) {
if (ssDCval < 75) {
if (Cs < 512) {
return 1;
}
else if (Cs >= 512) {
return 2;
}
}
else if (ssDCval >= 75) {
if (Rs < 445) {
return 2;
}
else if (Rs >= 445) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (diffMVdiffVal < 1) {
if (Rs < 694) {
return 8;
}
else if (Rs >= 694) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
else if (negBalance >= 44) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Cs < 422) {
return 8;
}
else if (Cs >= 422) {
if (Cs < 526) {
return 2;
}
else if (Cs >= 526) {
return 8;
}
}
}
}
}
else if (AFD >= 14) {
if (RsCsDiff < 48) {
if (Cs < 413) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
else if (Cs >= 413) {
if (Cs < 485) {
return 2;
}
else if (Cs >= 485) {
return 1;
}
}
}
else if (RsCsDiff >= 48) {
if (posBalance < 54) {
if (Cs < 477) {
if (MVDiff < 219) {
return 1;
}
else if (MVDiff >= 219) {
return 8;
}
}
else if (Cs >= 477) {
return 8;
}
}
else if (posBalance >= 54) {
if (Rs < 882) {
return 1;
}
else if (Rs >= 882) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1124) {
if (Cs < 482) {
if (Rs < 1607) {
if (mu_mv_mag_sq < 7) {
if (negBalance < 61) {
return 1;
}
else if (negBalance >= 61) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
return 1;
}
}
else if (Rs >= 1607) {
return 2;
}
}
else if (Cs >= 482) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 1143) {
return 1;
}
else if (Rs >= 1143) {
return 8;
}
}
}
}
}
else if (posBalance >= 55) {
if (MVDiff < 19) {
if (Rs < 1506) {
if (mu_mv_mag_sq < 21) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (MVDiff < 12) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
else if (MVDiff >= 12) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (MVDiff < 15) {
if (Cs < 509) {
return 4;
}
else if (Cs >= 509) {
return 8;
}
}
else if (MVDiff >= 15) {
return 8;
}
}
}
else if (Rs >= 1506) {
return 2;
}
}
else if (MVDiff >= 19) {
if (negBalance < 51) {
if (tcor < 42) {
if (diffMVdiffVal < 34) {
return 1;
}
else if (diffMVdiffVal >= 34) {
return 2;
}
}
else if (tcor >= 42) {
if (MVDiff < 40) {
if (negBalance < 47) {
if (diffRsCsdiff < -14) {
if (Rs < 974) {
return 4;
}
else if (Rs >= 974) {
return 1;
}
}
else if (diffRsCsdiff >= -14) {
if (mcTcor < 85) {
if (Rs < 1000) {
return 2;
}
else if (Rs >= 1000) {
return 8;
}
}
else if (mcTcor >= 85) {
if (RsCsDiff < 26) {
return 8;
}
else if (RsCsDiff >= 26) {
return 1;
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 1123) {
if (Cs < 367) {
return 2;
}
else if (Cs >= 367) {
return 8;
}
}
else if (Rs >= 1123) {
if (Rs < 1408) {
return 2;
}
else if (Rs >= 1408) {
return 4;
}
}
}
}
else if (MVDiff >= 40) {
if (diffAFD < 1) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
else if (diffAFD >= 1) {
if (diffRsCsdiff < -9) {
return 8;
}
else if (diffRsCsdiff >= -9) {
if (RsDiff < 8) {
if (ssDCval < 100) {
return 4;
}
else if (ssDCval >= 100) {
return 2;
}
}
else if (RsDiff >= 8) {
if (Rs < 904) {
return 1;
}
else if (Rs >= 904) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 51) {
if (Rs < 1016) {
if (AFD < 18) {
if (mcTcor < 86) {
if (mu_mv_mag_sq < 9) {
if (Cs < 462) {
return 2;
}
else if (Cs >= 462) {
if (Rs < 604) {
return 4;
}
else if (Rs >= 604) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 9) {
return 4;
}
}
else if (mcTcor >= 86) {
if (negBalance < 54) {
if (Rs < 748) {
return 4;
}
else if (Rs >= 748) {
return 8;
}
}
else if (negBalance >= 54) {
return 8;
}
}
}
else if (AFD >= 18) {
if (SC < 1272) {
if (CsDiff < 4) {
if (ssDCval < 140) {
return 2;
}
else if (ssDCval >= 140) {
return 1;
}
}
else if (CsDiff >= 4) {
if (mu_mv_mag_sq < 30) {
if (refDCval < 91) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (refDCval >= 91) {
if (Rs < 558) {
return 1;
}
else if (Rs >= 558) {
if (Rs < 702) {
return 4;
}
else if (Rs >= 702) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (ssDCval < 113) {
if (Rs < 816) {
return 8;
}
else if (Rs >= 816) {
return 4;
}
}
else if (ssDCval >= 113) {
return 2;
}
}
}
}
else if (SC >= 1272) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (Rs < 817) {
return 8;
}
else if (Rs >= 817) {
return 1;
}
}
}
}
}
else if (Rs >= 1016) {
if (Cs < 466) {
if (negBalance < 56) {
if (Rs < 1348) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 2;
}
}
else if (Rs >= 1348) {
return 4;
}
}
else if (negBalance >= 56) {
if (Rs < 1490) {
return 4;
}
else if (Rs >= 1490) {
if (Rs < 1534) {
return 1;
}
else if (Rs >= 1534) {
return 2;
}
}
}
}
else if (Cs >= 466) {
if (diffAFD < 10) {
return 4;
}
else if (diffAFD >= 10) {
return 2;
}
}
}
}
}
}
}
}
else if (SC >= 1984) {
if (refDCval < 114) {
if (mu_mv_mag_sq < 79) {
if (MVDiff < 35) {
if (Rs < 1785) {
if (AFD < 23) {
if (diffMVdiffVal < 0) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
else if (diffMVdiffVal >= 0) {
if (mu_mv_mag_sq < 30) {
return 1;
}
else if (mu_mv_mag_sq >= 30) {
if (Rs < 1569) {
return 8;
}
else if (Rs >= 1569) {
if (Rs < 1636) {
return 1;
}
else if (Rs >= 1636) {
return 2;
}
}
}
}
}
else if (AFD >= 23) {
return 4;
}
}
else if (Rs >= 1785) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (AFD < 31) {
if (diffMVdiffVal < -10) {
if (diffRsCsdiff < 34) {
return 2;
}
else if (diffRsCsdiff >= 34) {
if (Cs < 394) {
return 1;
}
else if (Cs >= 394) {
return 4;
}
}
}
else if (diffMVdiffVal >= -10) {
if (RsDiff < 9) {
if (Rs < 1940) {
return 2;
}
else if (Rs >= 1940) {
return 1;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
}
else if (AFD >= 31) {
return 1;
}
}
}
}
else if (MVDiff >= 35) {
if (tcor < 75) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
if (mu_mv_mag_sq < 25) {
if (mu_mv_mag_sq < 10) {
return 2;
}
else if (mu_mv_mag_sq >= 10) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (Cs < 427) {
return 2;
}
else if (Cs >= 427) {
if (Rs < 1739) {
return 2;
}
else if (Rs >= 1739) {
return 4;
}
}
}
}
}
else if (tcor >= 75) {
if (Cs < 287) {
if (Rs < 1974) {
return 2;
}
else if (Rs >= 1974) {
return 1;
}
}
else if (Cs >= 287) {
if (MVDiff < 54) {
if (Rs < 1775) {
return 1;
}
else if (Rs >= 1775) {
return 8;
}
}
else if (MVDiff >= 54) {
if (Rs < 2414) {
return 1;
}
else if (Rs >= 2414) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 79) {
return 2;
}
}
else if (refDCval >= 114) {
if (refDCval < 123) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (diffMVdiffVal < 47) {
return 1;
}
else if (diffMVdiffVal >= 47) {
return 8;
}
}
}
else if (refDCval >= 123) {
if (CsDiff < 3) {
if (SC < 2104) {
return 1;
}
else if (SC >= 2104) {
if (MVDiff < 78) {
return 1;
}
else if (MVDiff >= 78) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (SC < 2305) {
if (Cs < 359) {
return 1;
}
else if (Cs >= 359) {
if (Cs < 465) {
return 2;
}
else if (Cs >= 465) {
return 8;
}
}
}
else if (SC >= 2305) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (RsDiff < 10) {
if (Rs < 787) {
if (Cs < 290) {
if (mu_mv_mag_sq < 18) {
if (diffRsCsdiff < 10) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (diffRsCsdiff >= 10) {
if (RsDiff < 5) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (RsDiff >= 5) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (Rs < 513) {
if (diffMVdiffVal < -6) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (Rs < 439) {
return 8;
}
else if (Rs >= 439) {
return 2;
}
}
}
else if (diffMVdiffVal >= -6) {
if (diffMVdiffVal < 13) {
if (Rs < 486) {
return 2;
}
else if (Rs >= 486) {
return 1;
}
}
else if (diffMVdiffVal >= 13) {
if (Rs < 478) {
return 1;
}
else if (Rs >= 478) {
return 8;
}
}
}
}
else if (Rs >= 513) {
if (MVDiff < 136) {
if (tcor < 61) {
if (Cs < 127) {
return 1;
}
else if (Cs >= 127) {
return 2;
}
}
else if (tcor >= 61) {
if (Rs < 763) {
return 1;
}
else if (Rs >= 763) {
if (Rs < 771) {
return 4;
}
else if (Rs >= 771) {
return 1;
}
}
}
}
else if (MVDiff >= 136) {
return 2;
}
}
}
}
else if (Cs >= 290) {
if (diffRsCsdiff < -8) {
if (ssDCval < 78) {
if (diffRsCsdiff < -12) {
if (negBalance < 56) {
return 1;
}
else if (negBalance >= 56) {
if (Rs < 528) {
return 4;
}
else if (Rs >= 528) {
return 8;
}
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
else if (ssDCval >= 78) {
if (Rs < 314) {
return 1;
}
else if (Rs >= 314) {
return 8;
}
}
}
else if (diffRsCsdiff >= -8) {
if (ssDCval < 118) {
if (posBalance < 65) {
if (mu_mv_mag_sq < 43) {
if (Rs < 698) {
if (SC < 964) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (mu_mv_mag_sq < 10) {
return 1;
}
else if (mu_mv_mag_sq >= 10) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 4;
}
}
}
}
else if (SC >= 964) {
if (Cs < 484) {
if (Rs < 687) {
return 1;
}
else if (Rs >= 687) {
if (Rs < 692) {
return 4;
}
else if (Rs >= 692) {
return 1;
}
}
}
else if (Cs >= 484) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
else if (Rs >= 698) {
if (Cs < 484) {
if (Rs < 706) {
return 2;
}
else if (Rs >= 706) {
return 8;
}
}
else if (Cs >= 484) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (posBalance >= 65) {
if (AFD < 20) {
if (Rs < 643) {
return 4;
}
else if (Rs >= 643) {
return 1;
}
}
else if (AFD >= 20) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
}
else if (ssDCval >= 118) {
if (Rs < 579) {
if (Rs < 568) {
return 2;
}
else if (Rs >= 568) {
return 1;
}
}
else if (Rs >= 579) {
if (MVDiff < 24) {
return 2;
}
else if (MVDiff >= 24) {
return 8;
}
}
}
}
}
}
else if (Rs >= 787) {
if (refDCval < 139) {
if (refDCval < 56) {
if (SC < 2324) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
if (RsCsDiff < 28) {
if (Rs < 1870) {
return 1;
}
else if (Rs >= 1870) {
return 2;
}
}
else if (RsCsDiff >= 28) {
if (mu_mv_mag_sq < 18) {
if (Rs < 1597) {
return 1;
}
else if (Rs >= 1597) {
return 4;
}
}
else if (mu_mv_mag_sq >= 18) {
if (mu_mv_mag_sq < 42) {
return 2;
}
else if (mu_mv_mag_sq >= 42) {
return 1;
}
}
}
}
}
else if (SC >= 2324) {
if (negBalance < 36) {
if (RsCsDiff < 19) {
return 1;
}
else if (RsCsDiff >= 19) {
return 4;
}
}
else if (negBalance >= 36) {
return 1;
}
}
}
else if (refDCval >= 56) {
if (MVDiff < 9) {
if (posBalance < 63) {
if (ssDCval < 67) {
if (posBalance < 45) {
if (Rs < 1345) {
return 8;
}
else if (Rs >= 1345) {
if (Rs < 1603) {
return 2;
}
else if (Rs >= 1603) {
return 1;
}
}
}
else if (posBalance >= 45) {
return 2;
}
}
else if (ssDCval >= 67) {
if (tcor < 82) {
if (Cs < 240) {
if (Rs < 1657) {
return 2;
}
else if (Rs >= 1657) {
return 1;
}
}
else if (Cs >= 240) {
if (Cs < 423) {
return 1;
}
else if (Cs >= 423) {
return 8;
}
}
}
else if (tcor >= 82) {
if (ssDCval < 126) {
if (diffAFD < -4) {
if (Rs < 2195) {
return 2;
}
else if (Rs >= 2195) {
return 1;
}
}
else if (diffAFD >= -4) {
if (negBalance < 31) {
if (Cs < 457) {
return 1;
}
else if (Cs >= 457) {
return 8;
}
}
else if (negBalance >= 31) {
return 1;
}
}
}
else if (ssDCval >= 126) {
if (Rs < 1387) {
return 1;
}
else if (Rs >= 1387) {
return 2;
}
}
}
}
}
else if (posBalance >= 63) {
if (Rs < 1248) {
return 1;
}
else if (Rs >= 1248) {
return 2;
}
}
}
else if (MVDiff >= 9) {
if (AFD < 7) {
if (Rs < 1806) {
if (Cs < 457) {
return 1;
}
else if (Cs >= 457) {
if (Rs < 1289) {
return 4;
}
else if (Rs >= 1289) {
return 2;
}
}
}
else if (Rs >= 1806) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
return 4;
}
}
}
else if (AFD >= 7) {
if (Cs < 399) {
if (Cs < 186) {
if (RsDiff < 7) {
if (Cs < 156) {
if (posBalance < 100) {
return 1;
}
else if (posBalance >= 100) {
return 2;
}
}
else if (Cs >= 156) {
if (Rs < 1041) {
if (Rs < 875) {
return 1;
}
else if (Rs >= 875) {
return 2;
}
}
else if (Rs >= 1041) {
return 1;
}
}
}
else if (RsDiff >= 7) {
if (MVDiff < 30) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
if (Cs < 125) {
if (Cs < 99) {
return 2;
}
else if (Cs >= 99) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (Cs >= 125) {
return 2;
}
}
}
}
else if (MVDiff >= 30) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (diffMVdiffVal < -18) {
if (Cs < 171) {
if (Rs < 1145) {
return 1;
}
else if (Rs >= 1145) {
return 8;
}
}
else if (Cs >= 171) {
return 2;
}
}
else if (diffMVdiffVal >= -18) {
return 1;
}
}
}
}
}
else if (Cs >= 186) {
if (Rs < 1229) {
if (negBalance < 51) {
if (diffAFD < -8) {
return 8;
}
else if (diffAFD >= -8) {
if (diffRsCsdiff < -19) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (diffRsCsdiff >= -19) {
return 1;
}
}
}
else if (negBalance >= 51) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (diffRsCsdiff < -14) {
if (Cs < 216) {
return 1;
}
else if (Cs >= 216) {
return 2;
}
}
else if (diffRsCsdiff >= -14) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 2;
}
}
}
}
}
else if (Rs >= 1229) {
if (tcor < 79) {
if (refDCval < 95) {
return 1;
}
else if (refDCval >= 95) {
if (Rs < 1552) {
return 1;
}
else if (Rs >= 1552) {
return 8;
}
}
}
else if (tcor >= 79) {
if (ssDCval < 59) {
if (Rs < 1646) {
return 1;
}
else if (Rs >= 1646) {
return 2;
}
}
else if (ssDCval >= 59) {
return 1;
}
}
}
}
}
else if (Cs >= 399) {
if (tcor < 91) {
if (SC < 1679) {
if (ssDCval < 67) {
if (AFD < 13) {
if (Rs < 932) {
return 8;
}
else if (Rs >= 932) {
return 1;
}
}
else if (AFD >= 13) {
return 4;
}
}
else if (ssDCval >= 67) {
if (Rs < 1175) {
return 1;
}
else if (Rs >= 1175) {
return 8;
}
}
}
else if (SC >= 1679) {
if (RsCsDiff < 43) {
if (diffRsCsdiff < -6) {
if (diffMVdiffVal < 6) {
return 1;
}
else if (diffMVdiffVal >= 6) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
if (Rs < 1794) {
return 2;
}
else if (Rs >= 1794) {
return 1;
}
}
}
}
else if (RsCsDiff >= 43) {
if (Rs < 2353) {
if (diffRsCsdiff < 12) {
if (diffRsCsdiff < -18) {
if (refDCval < 70) {
return 2;
}
else if (refDCval >= 70) {
return 1;
}
}
else if (diffRsCsdiff >= -18) {
if (Rs < 1325) {
return 8;
}
else if (Rs >= 1325) {
if (Rs < 2109) {
if (Cs < 519) {
return 2;
}
else if (Cs >= 519) {
return 1;
}
}
else if (Rs >= 2109) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 12) {
if (ssDCval < 72) {
if (Rs < 1910) {
return 1;
}
else if (Rs >= 1910) {
return 2;
}
}
else if (ssDCval >= 72) {
return 1;
}
}
}
else if (Rs >= 2353) {
if (Rs < 2397) {
if (Rs < 2370) {
return 4;
}
else if (Rs >= 2370) {
return 1;
}
}
else if (Rs >= 2397) {
return 8;
}
}
}
}
}
else if (tcor >= 91) {
if (AFD < 8) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (AFD >= 8) {
if (Cs < 400) {
return 2;
}
else if (Cs >= 400) {
if (AFD < 9) {
if (ssDCval < 95) {
return 1;
}
else if (ssDCval >= 95) {
return 2;
}
}
else if (AFD >= 9) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 139) {
if (TSC < 6) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (negBalance < 28) {
return 2;
}
else if (negBalance >= 28) {
return 1;
}
}
}
}
else if (TSC >= 6) {
if (refDCval < 175) {
if (Rs < 1172) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (Rs < 1124) {
return 1;
}
else if (Rs >= 1124) {
return 2;
}
}
}
else if (Rs >= 1172) {
if (MVDiff < 97) {
return 1;
}
else if (MVDiff >= 97) {
if (Rs < 1475) {
return 1;
}
else if (Rs >= 1475) {
return 8;
}
}
}
}
else if (refDCval >= 175) {
return 2;
}
}
}
}
}
else if (RsDiff >= 10) {
if (ssDCval < 106) {
if (Rs < 1926) {
if (Cs < 506) {
if (negBalance < 48) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (TSC < 8) {
if (SC < 2052) {
return 2;
}
else if (SC >= 2052) {
return 4;
}
}
else if (TSC >= 8) {
return 1;
}
}
}
}
else if (negBalance >= 48) {
if (ssDCval < 96) {
if (Cs < 99) {
if (TSC < 6) {
if (Rs < 1578) {
return 2;
}
else if (Rs >= 1578) {
return 1;
}
}
else if (TSC >= 6) {
return 1;
}
}
else if (Cs >= 99) {
if (AFD < 36) {
return 2;
}
else if (AFD >= 36) {
return 1;
}
}
}
else if (ssDCval >= 96) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (RsDiff < 16) {
return 1;
}
else if (RsDiff >= 16) {
return 2;
}
}
}
}
}
else if (Cs >= 506) {
if (tcor < 68) {
if (Rs < 1452) {
return 4;
}
else if (Rs >= 1452) {
return 8;
}
}
else if (tcor >= 68) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 2;
}
}
}
}
else if (Rs >= 1926) {
if (RsCsDiff < 151) {
if (MVDiff < 29) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
if (Rs < 2158) {
return 2;
}
else if (Rs >= 2158) {
return 1;
}
}
}
else if (MVDiff >= 29) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 1;
}
}
}
else if (RsCsDiff >= 151) {
if (posBalance < 71) {
if (ssDCval < 100) {
return 2;
}
else if (ssDCval >= 100) {
if (Cs < 147) {
return 2;
}
else if (Cs >= 147) {
return 1;
}
}
}
else if (posBalance >= 71) {
if (refDCval < 85) {
return 1;
}
else if (refDCval >= 85) {
return 2;
}
}
}
}
}
else if (ssDCval >= 106) {
if (CsDiff < 2) {
if (mu_mv_mag_sq < 26) {
if (Rs < 2298) {
if (Cs < 129) {
return 2;
}
else if (Cs >= 129) {
return 1;
}
}
else if (Rs >= 2298) {
return 8;
}
}
else if (mu_mv_mag_sq >= 26) {
return 2;
}
}
else if (CsDiff >= 2) {
if (tcor < 75) {
if (diffAFD < 9) {
if (mu_mv_mag_sq < 15) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 2;
}
}
else if (mu_mv_mag_sq >= 15) {
if (gchDC < 4) {
if (Rs < 1631) {
return 2;
}
else if (Rs >= 1631) {
return 1;
}
}
else if (gchDC >= 4) {
return 1;
}
}
}
else if (diffAFD >= 9) {
return 1;
}
}
else if (tcor >= 75) {
if (CsDiff < 3) {
if (refDCval < 123) {
if (diffTSC < -2) {
if (Rs < 731) {
return 1;
}
else if (Rs >= 731) {
return 8;
}
}
else if (diffTSC >= -2) {
if (Cs < 311) {
return 1;
}
else if (Cs >= 311) {
if (posBalance < 74) {
return 2;
}
else if (posBalance >= 74) {
return 1;
}
}
}
}
else if (refDCval >= 123) {
return 1;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 533) {
if (RsCsDiff < 73) {
if (Cs < 974) {
if (mcTcor < 95) {
if (MVDiff < 18) {
if (refDCval < 28) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
return 4;
}
}
else if (refDCval >= 28) {
if (refDCval < 153) {
if (mu_mv_mag_sq < 118) {
if (posBalance < 45) {
if (Rs < 1357) {
if (TSC < 2) {
if (negBalance < 28) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
return 4;
}
}
else if (negBalance >= 28) {
return 4;
}
}
else if (TSC >= 2) {
if (refDCval < 62) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (AFD < 6) {
if (Cs < 891) {
if (negBalance < 24) {
if (Rs < 904) {
return 1;
}
else if (Rs >= 904) {
return 8;
}
}
else if (negBalance >= 24) {
if (Rs < 727) {
if (Cs < 639) {
return 8;
}
else if (Cs >= 639) {
return 4;
}
}
else if (Rs >= 727) {
return 8;
}
}
}
else if (Cs >= 891) {
if (Rs < 421) {
return 1;
}
else if (Rs >= 421) {
return 4;
}
}
}
else if (AFD >= 6) {
if (negBalance < 34) {
if (diffMVdiffVal < 5) {
if (Rs < 871) {
return 2;
}
else if (Rs >= 871) {
return 1;
}
}
else if (diffMVdiffVal >= 5) {
if (Rs < 1157) {
return 2;
}
else if (Rs >= 1157) {
return 8;
}
}
}
else if (negBalance >= 34) {
if (diffMVdiffVal < -2) {
if (Rs < 1076) {
return 1;
}
else if (Rs >= 1076) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 815) {
if (Cs < 768) {
return 8;
}
else if (Cs >= 768) {
return 2;
}
}
else if (Rs >= 815) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 62) {
if (ssDCval < 133) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (SC < 1442) {
if (SC < 1172) {
return 2;
}
else if (SC >= 1172) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (SC >= 1442) {
if (Rs < 1033) {
return 2;
}
else if (Rs >= 1033) {
return 4;
}
}
}
}
else if (ssDCval >= 133) {
return 8;
}
}
}
}
else if (Rs >= 1357) {
if (MVDiff < 5) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
if (negBalance < 37) {
return 8;
}
else if (negBalance >= 37) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (Cs < 664) {
if (Cs < 626) {
return 1;
}
else if (Cs >= 626) {
if (Rs < 2073) {
return 4;
}
else if (Rs >= 2073) {
return 1;
}
}
}
else if (Cs >= 664) {
if (Rs < 1839) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (Rs >= 1839) {
if (ssDCval < 69) {
if (posBalance < 41) {
if (Rs < 2124) {
return 8;
}
else if (Rs >= 2124) {
return 1;
}
}
else if (posBalance >= 41) {
return 4;
}
}
else if (ssDCval >= 69) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 1001) {
if (TSC < 9) {
if (posBalance < 58) {
if (mu_mv_mag_sq < 100) {
if (refDCval < 39) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Rs < 628) {
return 2;
}
else if (Rs >= 628) {
if (diffMVdiffVal < -4) {
if (Rs < 786) {
return 2;
}
else if (Rs >= 786) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
}
else if (refDCval >= 39) {
if (negBalance < 32) {
if (refDCval < 58) {
if (mu_mv_mag_sq < 57) {
return 4;
}
else if (mu_mv_mag_sq >= 57) {
return 1;
}
}
else if (refDCval >= 58) {
if (Rs < 395) {
return 8;
}
else if (Rs >= 395) {
return 2;
}
}
}
else if (negBalance >= 32) {
if (mu_mv_mag_sq < 37) {
if (SC < 1399) {
if (tcor < 75) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
return 2;
}
}
else if (tcor >= 75) {
if (negBalance < 46) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (diffAFD < 3) {
if (ssDCval < 48) {
if (Cs < 604) {
return 2;
}
else if (Cs >= 604) {
return 8;
}
}
else if (ssDCval >= 48) {
return 2;
}
}
else if (diffAFD >= 3) {
return 1;
}
}
}
else if (negBalance >= 46) {
if (tcor < 89) {
if (Rs < 737) {
if (negBalance < 58) {
return 8;
}
else if (negBalance >= 58) {
if (Cs < 673) {
if (Rs < 647) {
return 1;
}
else if (Rs >= 647) {
return 2;
}
}
else if (Cs >= 673) {
return 8;
}
}
}
else if (Rs >= 737) {
return 1;
}
}
else if (tcor >= 89) {
if (Rs < 721) {
return 2;
}
else if (Rs >= 721) {
return 4;
}
}
}
}
}
else if (SC >= 1399) {
if (MVDiff < 4) {
if (AFD < 11) {
if (Cs < 640) {
return 1;
}
else if (Cs >= 640) {
return 8;
}
}
else if (AFD >= 11) {
if (Rs < 866) {
return 1;
}
else if (Rs >= 866) {
return 2;
}
}
}
else if (MVDiff >= 4) {
if (diffMVdiffVal < 5) {
if (Cs < 954) {
if (tcor < 93) {
if (TSC < 8) {
if (diffRsCsdiff < 7) {
return 8;
}
else if (diffRsCsdiff >= 7) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Rs < 705) {
return 8;
}
else if (Rs >= 705) {
return 2;
}
}
}
}
else if (TSC >= 8) {
if (Rs < 841) {
return 4;
}
else if (Rs >= 841) {
return 8;
}
}
}
else if (tcor >= 93) {
return 4;
}
}
else if (Cs >= 954) {
if (Rs < 833) {
return 2;
}
else if (Rs >= 833) {
return 1;
}
}
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (ssDCval < 70) {
if (MVDiff < 11) {
if (SC < 1459) {
if (SC < 1403) {
return 8;
}
else if (SC >= 1403) {
if (Rs < 877) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (Rs >= 877) {
return 8;
}
}
}
else if (SC >= 1459) {
if (MVDiff < 8) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (MVDiff >= 8) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (Rs < 955) {
return 1;
}
else if (Rs >= 955) {
return 8;
}
}
}
}
}
else if (MVDiff >= 11) {
if (mu_mv_mag_sq < 78) {
if (AFD < 8) {
if (Rs < 993) {
return 8;
}
else if (Rs >= 993) {
return 1;
}
}
else if (AFD >= 8) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 828) {
return 8;
}
else if (Rs >= 828) {
return 4;
}
}
}
}
else if (ssDCval >= 70) {
if (MVDiff < 3) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (AFD < 7) {
if (Cs < 704) {
return 1;
}
else if (Cs >= 704) {
return 2;
}
}
else if (AFD >= 7) {
if (Rs < 634) {
return 4;
}
else if (Rs >= 634) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (SC < 1574) {
return 8;
}
else if (SC >= 1574) {
if (Rs < 735) {
return 1;
}
else if (Rs >= 735) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 444) {
return 1;
}
else if (Rs >= 444) {
if (tcor < 75) {
return 4;
}
else if (tcor >= 75) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (SC < 1625) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (SC >= 1625) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
return 1;
}
}
}
}
}
}
}
else if (posBalance >= 58) {
if (MVDiff < 4) {
if (negBalance < 47) {
if (ssDCval < 74) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 1;
}
}
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (negBalance >= 47) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (ssDCval >= 50) {
if (ssDCval < 89) {
if (ssDCval < 63) {
return 4;
}
else if (ssDCval >= 63) {
return 8;
}
}
else if (ssDCval >= 89) {
if (Cs < 749) {
return 1;
}
else if (Cs >= 749) {
if (Rs < 431) {
return 4;
}
else if (Rs >= 431) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 82) {
if (AFD < 12) {
if (ssDCval < 44) {
if (negBalance < 48) {
if (CsDiff < 5) {
if (posBalance < 72) {
if (diffRsCsdiff < -7) {
return 2;
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
else if (posBalance >= 72) {
return 2;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (negBalance >= 48) {
if (diffMVdiffVal < -3) {
if (Rs < 640) {
return 4;
}
else if (Rs >= 640) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
return 4;
}
}
}
else if (ssDCval >= 44) {
if (mcTcor < 78) {
if (ssDCval < 46) {
if (Cs < 768) {
if (diffRsCsdiff < 4) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (diffRsCsdiff >= 4) {
return 8;
}
}
else if (Cs >= 768) {
return 8;
}
}
else if (ssDCval >= 46) {
return 8;
}
}
else if (mcTcor >= 78) {
if (Rs < 936) {
if (negBalance < 52) {
if (posBalance < 60) {
if (RsCsDiff < 29) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Rs < 924) {
return 4;
}
else if (Rs >= 924) {
return 8;
}
}
}
else if (RsCsDiff >= 29) {
return 8;
}
}
else if (posBalance >= 60) {
if (Cs < 578) {
if (Rs < 414) {
return 1;
}
else if (Rs >= 414) {
return 8;
}
}
else if (Cs >= 578) {
if (diffRsCsdiff < 6) {
return 1;
}
else if (diffRsCsdiff >= 6) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (ssDCval < 51) {
if (mcTcor < 79) {
if (Rs < 786) {
return 1;
}
else if (Rs >= 786) {
return 8;
}
}
else if (mcTcor >= 79) {
return 1;
}
}
else if (ssDCval >= 51) {
if (Rs < 861) {
return 4;
}
else if (Rs >= 861) {
return 8;
}
}
}
}
else if (Rs >= 936) {
if (negBalance < 46) {
if (Rs < 983) {
return 4;
}
else if (Rs >= 983) {
return 8;
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
}
}
else if (AFD >= 12) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (refDCval < 79) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Cs < 718) {
return 1;
}
else if (Cs >= 718) {
return 4;
}
}
}
else if (refDCval >= 79) {
return 2;
}
}
}
}
else if (ssDCval >= 82) {
if (negBalance < 54) {
if (AFD < 6) {
if (mu_mv_mag_sq < 79) {
return 8;
}
else if (mu_mv_mag_sq >= 79) {
return 2;
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (negBalance >= 54) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (Rs < 671) {
return 2;
}
else if (Rs >= 671) {
return 8;
}
}
}
}
}
}
}
}
}
else if (TSC >= 9) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
if (Cs < 719) {
return 8;
}
else if (Cs >= 719) {
return 4;
}
}
}
}
else if (Rs >= 1001) {
if (mcTcor < 90) {
if (ssDCval < 121) {
if (mcTcor < 84) {
return 8;
}
else if (mcTcor >= 84) {
if (posBalance < 59) {
if (MVDiff < 4) {
if (posBalance < 55) {
if (Rs < 1115) {
return 8;
}
else if (Rs >= 1115) {
if (Rs < 1274) {
return 1;
}
else if (Rs >= 1274) {
return 8;
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (MVDiff >= 4) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 1289) {
if (SC < 1683) {
if (Rs < 1088) {
if (Cs < 624) {
return 8;
}
else if (Cs >= 624) {
return 4;
}
}
else if (Rs >= 1088) {
return 2;
}
}
else if (SC >= 1683) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 1;
}
}
}
else if (Rs >= 1289) {
if (tcor < 83) {
if (negBalance < 50) {
if (Rs < 1315) {
return 2;
}
else if (Rs >= 1315) {
return 8;
}
}
else if (negBalance >= 50) {
return 2;
}
}
else if (tcor >= 83) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (Rs < 1441) {
return 2;
}
else if (Rs >= 1441) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 59) {
if (AFD < 15) {
if (posBalance < 70) {
if (diffMVdiffVal < -8) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
return 4;
}
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
else if (posBalance >= 70) {
if (Cs < 912) {
return 8;
}
else if (Cs >= 912) {
return 4;
}
}
}
else if (AFD >= 15) {
if (Rs < 1154) {
return 4;
}
else if (Rs >= 1154) {
return 8;
}
}
}
}
}
else if (ssDCval >= 121) {
if (Cs < 730) {
return 4;
}
else if (Cs >= 730) {
if (Rs < 1078) {
return 8;
}
else if (Rs >= 1078) {
return 1;
}
}
}
}
else if (mcTcor >= 90) {
if (RsCsDiff < 23) {
if (ssDCval < 125) {
if (gchDC < 1) {
if (posBalance < 47) {
if (AFD < 6) {
if (Cs < 818) {
return 8;
}
else if (Cs >= 818) {
return 4;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (posBalance >= 47) {
if (diffRsCsdiff < -17) {
return 2;
}
else if (diffRsCsdiff >= -17) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (Rs < 1158) {
return 4;
}
else if (Rs >= 1158) {
return 8;
}
}
}
}
}
else if (gchDC >= 1) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (RsCsDiff < 16) {
if (refDCval < 51) {
if (Rs < 1096) {
return 1;
}
else if (Rs >= 1096) {
if (Rs < 1502) {
return 8;
}
else if (Rs >= 1502) {
return 4;
}
}
}
else if (refDCval >= 51) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (Rs < 1096) {
return 2;
}
else if (Rs >= 1096) {
return 1;
}
}
}
}
}
else if (ssDCval >= 125) {
return 4;
}
}
else if (RsCsDiff >= 23) {
if (ssDCval < 74) {
if (ssDCval < 56) {
if (negBalance < 37) {
if (Rs < 1040) {
return 8;
}
else if (Rs >= 1040) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 2;
}
}
}
else if (negBalance >= 37) {
return 8;
}
}
else if (ssDCval >= 56) {
if (Rs < 1491) {
if (AFD < 9) {
if (negBalance < 44) {
if (Cs < 932) {
return 1;
}
else if (Cs >= 932) {
return 4;
}
}
else if (negBalance >= 44) {
if (TSC < 7) {
if (Rs < 1458) {
return 4;
}
else if (Rs >= 1458) {
return 8;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (AFD >= 9) {
if (negBalance < 38) {
if (Cs < 674) {
return 1;
}
else if (Cs >= 674) {
return 2;
}
}
else if (negBalance >= 38) {
if (refDCval < 71) {
if (diffRsCsdiff < -12) {
return 4;
}
else if (diffRsCsdiff >= -12) {
if (posBalance < 50) {
if (Rs < 1310) {
return 2;
}
else if (Rs >= 1310) {
return 1;
}
}
else if (posBalance >= 50) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 1;
}
}
}
}
else if (refDCval >= 71) {
if (diffTSC < -1) {
if (Rs < 1324) {
return 4;
}
else if (Rs >= 1324) {
return 8;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
}
}
}
else if (Rs >= 1491) {
if (mu_mv_mag_sq < 60) {
if (Cs < 720) {
return 2;
}
else if (Cs >= 720) {
if (mcTcor < 94) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (mcTcor >= 94) {
if (RsDiff < 7) {
if (Rs < 1515) {
return 4;
}
else if (Rs >= 1515) {
return 8;
}
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 60) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (Rs < 1823) {
return 8;
}
else if (Rs >= 1823) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 74) {
if (tcor < 86) {
if (ssDCval < 81) {
if (mu_mv_mag_sq < 42) {
if (SC < 2037) {
return 2;
}
else if (SC >= 2037) {
return 8;
}
}
else if (mu_mv_mag_sq >= 42) {
return 4;
}
}
else if (ssDCval >= 81) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (Rs < 1042) {
return 4;
}
else if (Rs >= 1042) {
return 8;
}
}
}
}
else if (tcor >= 86) {
if (Rs < 1580) {
if (Rs < 1157) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (Rs < 1110) {
return 2;
}
else if (Rs >= 1110) {
return 8;
}
}
}
else if (Rs >= 1157) {
return 8;
}
}
else if (Rs >= 1580) {
if (SC < 2656) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 1867) {
if (Rs < 1619) {
return 1;
}
else if (Rs >= 1619) {
return 8;
}
}
else if (Rs >= 1867) {
return 1;
}
}
}
else if (SC >= 2656) {
if (Rs < 2302) {
return 8;
}
else if (Rs >= 2302) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (ssDCval < 47) {
if (negBalance < 35) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (negBalance >= 35) {
if (posBalance < 55) {
if (AFD < 4) {
if (Rs < 858) {
return 1;
}
else if (Rs >= 858) {
return 8;
}
}
else if (AFD >= 4) {
return 1;
}
}
else if (posBalance >= 55) {
if (negBalance < 39) {
if (Cs < 852) {
return 8;
}
else if (Cs >= 852) {
return 1;
}
}
else if (negBalance >= 39) {
if (diffRsCsdiff < -7) {
return 1;
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
}
}
}
else if (ssDCval >= 47) {
if (ssDCval < 121) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 1152) {
return 4;
}
else if (Rs >= 1152) {
return 8;
}
}
}
else if (ssDCval >= 121) {
if (RsDiff < 3) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 795) {
return 8;
}
else if (Rs >= 795) {
return 1;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 1318) {
return 1;
}
else if (Rs >= 1318) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 153) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
if (Rs < 506) {
return 2;
}
else if (Rs >= 506) {
return 8;
}
}
}
}
}
else if (MVDiff >= 18) {
if (CsDiff < 4) {
if (refDCval < 138) {
if (Cs < 750) {
if (SC < 980) {
if (tcor < 57) {
if (Rs < 353) {
return 2;
}
else if (Rs >= 353) {
return 4;
}
}
else if (tcor >= 57) {
return 1;
}
}
else if (SC >= 980) {
if (ssDCval < 125) {
if (ssDCval < 84) {
if (ssDCval < 43) {
if (mu_mv_mag_sq < 53) {
if (mcTcor < 87) {
if (SC < 1297) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
else if (SC >= 1297) {
return 8;
}
}
else if (mcTcor >= 87) {
if (RsCsDiff < 16) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
else if (RsCsDiff >= 16) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (TSC < 4) {
if (Rs < 801) {
return 4;
}
else if (Rs >= 801) {
return 1;
}
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (ssDCval >= 43) {
if (MVDiff < 32) {
if (Rs < 1363) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (negBalance < 39) {
if (mu_mv_mag_sq < 49) {
if (Rs < 648) {
return 8;
}
else if (Rs >= 648) {
return 2;
}
}
else if (mu_mv_mag_sq >= 49) {
if (Rs < 894) {
return 1;
}
else if (Rs >= 894) {
return 8;
}
}
}
else if (negBalance >= 39) {
if (posBalance < 54) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
else if (posBalance >= 54) {
if (MVDiff < 28) {
if (RsDiff < 4) {
if (Rs < 452) {
return 1;
}
else if (Rs >= 452) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (MVDiff >= 28) {
if (Rs < 689) {
return 8;
}
else if (Rs >= 689) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1363) {
if (Rs < 1606) {
if (Rs < 1482) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 4;
}
}
else if (Rs >= 1482) {
return 4;
}
}
else if (Rs >= 1606) {
if (Rs < 2239) {
return 2;
}
else if (Rs >= 2239) {
return 1;
}
}
}
}
else if (MVDiff >= 32) {
if (Rs < 951) {
if (RsDiff < 3) {
if (ssDCval < 49) {
if (SC < 1130) {
return 2;
}
else if (SC >= 1130) {
if (Rs < 500) {
return 8;
}
else if (Rs >= 500) {
return 4;
}
}
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (RsDiff >= 3) {
if (refDCval < 49) {
if (TSC < 9) {
if (Rs < 546) {
return 2;
}
else if (Rs >= 546) {
return 1;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (refDCval >= 49) {
return 4;
}
}
}
else if (Rs >= 951) {
if (mu_mv_mag_sq < 54) {
return 8;
}
else if (mu_mv_mag_sq >= 54) {
return 4;
}
}
}
}
}
else if (ssDCval >= 84) {
if (RsDiff < 7) {
if (negBalance < 36) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (diffRsCsdiff < -5) {
if (diffAFD < 1) {
if (diffRsCsdiff < -11) {
return 8;
}
else if (diffRsCsdiff >= -11) {
if (gchDC < 1) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (Rs < 1239) {
return 4;
}
else if (Rs >= 1239) {
return 2;
}
}
}
else if (diffRsCsdiff >= -5) {
return 8;
}
}
}
else if (RsDiff >= 7) {
return 2;
}
}
}
else if (ssDCval >= 125) {
if (mcTcor < 80) {
if (Rs < 1244) {
return 1;
}
else if (Rs >= 1244) {
return 4;
}
}
else if (mcTcor >= 80) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
}
}
}
else if (Cs >= 750) {
if (ssDCval < 61) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Cs < 862) {
if (Rs < 1415) {
return 1;
}
else if (Rs >= 1415) {
return 4;
}
}
else if (Cs >= 862) {
return 8;
}
}
}
else if (ssDCval >= 61) {
if (gchDC < 3) {
if (mcTcor < 89) {
if (negBalance < 51) {
if (Rs < 1186) {
return 1;
}
else if (Rs >= 1186) {
if (Rs < 1204) {
return 2;
}
else if (Rs >= 1204) {
return 8;
}
}
}
else if (negBalance >= 51) {
return 8;
}
}
else if (mcTcor >= 89) {
if (RsCsDiff < 43) {
if (mu_mv_mag_sq < 11) {
return 4;
}
else if (mu_mv_mag_sq >= 11) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
}
else if (RsCsDiff >= 43) {
if (ssDCval < 113) {
if (Rs < 1553) {
return 1;
}
else if (Rs >= 1553) {
if (Rs < 2091) {
return 4;
}
else if (Rs >= 2091) {
return 2;
}
}
}
else if (ssDCval >= 113) {
return 8;
}
}
}
}
else if (gchDC >= 3) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
return 8;
}
}
}
}
}
else if (refDCval >= 138) {
if (refDCval < 143) {
if (MVDiff < 51) {
return 4;
}
else if (MVDiff >= 51) {
if (Rs < 834) {
if (Rs < 582) {
return 4;
}
else if (Rs >= 582) {
return 2;
}
}
else if (Rs >= 834) {
return 8;
}
}
}
else if (refDCval >= 143) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
}
}
else if (CsDiff >= 4) {
if (ssDCval < 49) {
if (Rs < 820) {
if (tcor < 62) {
if (negBalance < 49) {
if (mcTcor < 79) {
if (AFD < 20) {
if (Cs < 605) {
return 8;
}
else if (Cs >= 605) {
return 2;
}
}
else if (AFD >= 20) {
return 1;
}
}
else if (mcTcor >= 79) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
else if (negBalance >= 49) {
if (diffMVdiffVal < -7) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (Rs < 448) {
return 2;
}
else if (Rs >= 448) {
return 8;
}
}
}
else if (diffMVdiffVal >= -7) {
if (MVDiff < 56) {
return 8;
}
else if (MVDiff >= 56) {
return 4;
}
}
}
}
else if (tcor >= 62) {
if (Cs < 714) {
if (diffRsCsdiff < 4) {
if (tcor < 66) {
return 2;
}
else if (tcor >= 66) {
return 1;
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 602) {
return 2;
}
else if (Rs >= 602) {
return 4;
}
}
}
else if (Cs >= 714) {
return 2;
}
}
}
else if (Rs >= 820) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (CsDiff < 5) {
if (Cs < 605) {
if (Rs < 908) {
return 8;
}
else if (Rs >= 908) {
return 4;
}
}
else if (Cs >= 605) {
return 8;
}
}
else if (CsDiff >= 5) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
return 8;
}
}
}
}
}
else if (ssDCval >= 49) {
if (tcor < 50) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (mu_mv_mag_sq < 20) {
if (AFD < 30) {
if (Rs < 319) {
return 4;
}
else if (Rs >= 319) {
return 8;
}
}
else if (AFD >= 30) {
return 4;
}
}
else if (mu_mv_mag_sq >= 20) {
if (Cs < 719) {
if (TSC < 12) {
if (Rs < 482) {
return 2;
}
else if (Rs >= 482) {
return 4;
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (Cs >= 719) {
return 2;
}
}
}
}
else if (tcor >= 50) {
if (tcor < 91) {
if (MVDiff < 67) {
if (TSC < 11) {
if (SC < 1495) {
if (refDCval < 79) {
if (refDCval < 70) {
if (AFD < 16) {
if (ssDCval < 53) {
if (MVDiff < 35) {
if (Rs < 765) {
return 4;
}
else if (Rs >= 765) {
return 2;
}
}
else if (MVDiff >= 35) {
return 8;
}
}
else if (ssDCval >= 53) {
if (mcTcor < 93) {
if (diffRsCsdiff < 17) {
if (ssDCval < 57) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
return 8;
}
}
else if (ssDCval >= 57) {
if (refDCval < 66) {
return 1;
}
else if (refDCval >= 66) {
if (Rs < 477) {
return 1;
}
else if (Rs >= 477) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 17) {
return 8;
}
}
else if (mcTcor >= 93) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
else if (AFD >= 16) {
if (AFD < 17) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 638) {
return 2;
}
else if (Rs >= 638) {
return 1;
}
}
}
else if (AFD >= 17) {
if (gchDC < 5) {
return 8;
}
else if (gchDC >= 5) {
if (Rs < 499) {
return 8;
}
else if (Rs >= 499) {
return 4;
}
}
}
}
}
else if (refDCval >= 70) {
if (mcTcor < 93) {
return 2;
}
else if (mcTcor >= 93) {
return 1;
}
}
}
else if (refDCval >= 79) {
if (TSC < 9) {
if (MVDiff < 46) {
if (Cs < 615) {
if (Cs < 614) {
return 8;
}
else if (Cs >= 614) {
return 4;
}
}
else if (Cs >= 615) {
return 8;
}
}
else if (MVDiff >= 46) {
return 2;
}
}
else if (TSC >= 9) {
if (Cs < 727) {
return 2;
}
else if (Cs >= 727) {
return 8;
}
}
}
}
else if (SC >= 1495) {
if (posBalance < 67) {
if (ssDCval < 56) {
if (Rs < 974) {
if (RsCsDiff < 33) {
return 1;
}
else if (RsCsDiff >= 33) {
return 4;
}
}
else if (Rs >= 974) {
if (MVDiff < 35) {
return 2;
}
else if (MVDiff >= 35) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (Cs < 791) {
if (diffMVdiffVal < -13) {
return 8;
}
else if (diffMVdiffVal >= -13) {
if (ssDCval < 80) {
if (diffMVdiffVal < 4) {
if (mcTcor < 89) {
if (diffAFD < 1) {
if (Rs < 1828) {
return 4;
}
else if (Rs >= 1828) {
return 1;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (mcTcor >= 89) {
if (diffMVdiffVal < 0) {
if (diffMVdiffVal < -5) {
if (Rs < 1196) {
return 1;
}
else if (Rs >= 1196) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
if (Rs < 767) {
return 1;
}
else if (Rs >= 767) {
return 4;
}
}
}
else if (diffMVdiffVal >= 0) {
if (Rs < 1014) {
return 8;
}
else if (Rs >= 1014) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
else if (ssDCval >= 80) {
if (diffRsCsdiff < 9) {
return 1;
}
else if (diffRsCsdiff >= 9) {
if (Rs < 1026) {
return 4;
}
else if (Rs >= 1026) {
if (Rs < 1152) {
return 1;
}
else if (Rs >= 1152) {
return 8;
}
}
}
}
}
}
else if (Cs >= 791) {
if (mu_mv_mag_sq < 3) {
if (mcTcor < 87) {
if (Rs < 1588) {
if (Rs < 1321) {
return 8;
}
else if (Rs >= 1321) {
return 1;
}
}
else if (Rs >= 1588) {
return 2;
}
}
else if (mcTcor >= 87) {
if (SC < 1563) {
return 2;
}
else if (SC >= 1563) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Cs < 891) {
return 2;
}
else if (Cs >= 891) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 51) {
if (negBalance < 38) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (negBalance >= 38) {
if (mu_mv_mag_sq < 41) {
if (Rs < 1132) {
return 4;
}
else if (Rs >= 1132) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (Rs < 1644) {
return 4;
}
else if (Rs >= 1644) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (MVDiff < 31) {
if (mu_mv_mag_sq < 54) {
return 8;
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 830) {
return 8;
}
else if (Rs >= 830) {
return 1;
}
}
}
else if (MVDiff >= 31) {
if (Rs < 1465) {
return 8;
}
else if (Rs >= 1465) {
return 2;
}
}
}
}
}
else if (negBalance >= 51) {
if (CsDiff < 5) {
if (Rs < 1322) {
return 8;
}
else if (Rs >= 1322) {
return 4;
}
}
else if (CsDiff >= 5) {
if (mu_mv_mag_sq < 11) {
if (Rs < 1299) {
return 1;
}
else if (Rs >= 1299) {
if (Rs < 1568) {
return 8;
}
else if (Rs >= 1568) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (SC < 2111) {
return 8;
}
else if (SC >= 2111) {
return 1;
}
}
}
}
}
}
}
}
else if (posBalance >= 67) {
if (MVDiff < 54) {
if (posBalance < 98) {
return 8;
}
else if (posBalance >= 98) {
if (Rs < 939) {
return 1;
}
else if (Rs >= 939) {
return 2;
}
}
}
else if (MVDiff >= 54) {
return 2;
}
}
}
}
else if (TSC >= 11) {
if (mu_mv_mag_sq < 3) {
return 8;
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
else if (MVDiff >= 67) {
if (SC < 1674) {
if (mcTcor < 84) {
if (tcor < 55) {
if (Rs < 397) {
return 4;
}
else if (Rs >= 397) {
if (Rs < 543) {
return 8;
}
else if (Rs >= 543) {
return 1;
}
}
}
else if (tcor >= 55) {
return 4;
}
}
else if (mcTcor >= 84) {
if (diffAFD < -5) {
if (negBalance < 60) {
if (MVDiff < 80) {
if (Rs < 351) {
return 8;
}
else if (Rs >= 351) {
return 4;
}
}
else if (MVDiff >= 80) {
if (Rs < 522) {
return 1;
}
else if (Rs >= 522) {
return 2;
}
}
}
else if (negBalance >= 60) {
return 1;
}
}
else if (diffAFD >= -5) {
if (refDCval < 91) {
if (Rs < 479) {
return 4;
}
else if (Rs >= 479) {
return 2;
}
}
else if (refDCval >= 91) {
return 2;
}
}
}
}
else if (SC >= 1674) {
return 8;
}
}
}
else if (tcor >= 91) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (negBalance < 14) {
if (diffTSC < -2) {
if (mu_mv_mag_sq < 128) {
if (Rs < 537) {
return 1;
}
else if (Rs >= 537) {
if (Cs < 866) {
return 8;
}
else if (Cs >= 866) {
if (Rs < 1059) {
return 8;
}
else if (Rs >= 1059) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1662) {
if (Cs < 643) {
if (Rs < 1215) {
if (SC < 1480) {
return 2;
}
else if (SC >= 1480) {
return 8;
}
}
else if (Rs >= 1215) {
return 1;
}
}
else if (Cs >= 643) {
if (diffRsCsdiff < -51) {
if (diffTSC < -9) {
if (MVDiff < 204) {
return 4;
}
else if (MVDiff >= 204) {
return 8;
}
}
else if (diffTSC >= -9) {
return 8;
}
}
else if (diffRsCsdiff >= -51) {
if (negBalance < 8) {
if (diffAFD < -8) {
if (ssDCval < 116) {
return 8;
}
else if (ssDCval >= 116) {
return 4;
}
}
else if (diffAFD >= -8) {
return 4;
}
}
else if (negBalance >= 8) {
return 8;
}
}
}
}
else if (Rs >= 1662) {
return 8;
}
}
}
else if (diffTSC >= -2) {
if (Rs < 701) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 58) {
if (Cs < 904) {
return 1;
}
else if (Cs >= 904) {
if (Rs < 208) {
return 4;
}
else if (Rs >= 208) {
return 2;
}
}
}
else if (ssDCval >= 58) {
if (SC < 1384) {
if (Rs < 509) {
if (Rs < 441) {
return 4;
}
else if (Rs >= 441) {
return 1;
}
}
else if (Rs >= 509) {
if (ssDCval < 117) {
if (MVDiff < 1) {
if (Cs < 682) {
return 8;
}
else if (Cs >= 682) {
return 1;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 117) {
return 8;
}
}
}
else if (SC >= 1384) {
if (Cs < 814) {
return 1;
}
else if (Cs >= 814) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 756) {
if (SC < 1380) {
if (RsDiff < 1) {
if (Rs < 374) {
return 1;
}
else if (Rs >= 374) {
return 8;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (SC >= 1380) {
return 1;
}
}
else if (Cs >= 756) {
if (ssDCval < 73) {
if (Rs < 378) {
return 2;
}
else if (Rs >= 378) {
return 8;
}
}
else if (ssDCval >= 73) {
return 2;
}
}
}
}
else if (Rs >= 701) {
if (ssDCval < 147) {
if (Rs < 1128) {
if (ssDCval < 38) {
if (ssDCval < 32) {
return 8;
}
else if (ssDCval >= 32) {
return 2;
}
}
else if (ssDCval >= 38) {
if (MVDiff < 3) {
if (TSC < 1) {
if (tcor < 101) {
return 8;
}
else if (tcor >= 101) {
return 4;
}
}
else if (TSC >= 1) {
if (ssDCval < 74) {
if (Rs < 917) {
return 2;
}
else if (Rs >= 917) {
return 8;
}
}
else if (ssDCval >= 74) {
if (ssDCval < 138) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 903) {
return 1;
}
else if (Rs >= 903) {
return 8;
}
}
}
else if (ssDCval >= 138) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (RsDiff < 3) {
if (Rs < 710) {
if (Cs < 812) {
return 8;
}
else if (Cs >= 812) {
return 1;
}
}
else if (Rs >= 710) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
return 8;
}
}
}
else if (RsDiff >= 3) {
return 1;
}
}
}
}
else if (Rs >= 1128) {
if (negBalance < 6) {
if (Rs < 1134) {
if (Rs < 1131) {
return 8;
}
else if (Rs >= 1131) {
return 1;
}
}
else if (Rs >= 1134) {
if (SC < 2956) {
return 8;
}
else if (SC >= 2956) {
if (Rs < 2156) {
return 1;
}
else if (Rs >= 2156) {
return 8;
}
}
}
}
else if (negBalance >= 6) {
if (Cs < 911) {
if (Cs < 679) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (Rs < 1680) {
if (Cs < 671) {
return 8;
}
else if (Cs >= 671) {
if (Rs < 1258) {
return 8;
}
else if (Rs >= 1258) {
return 1;
}
}
}
else if (Rs >= 1680) {
if (Rs < 1855) {
return 4;
}
else if (Rs >= 1855) {
return 2;
}
}
}
}
else if (Cs >= 679) {
if (posBalance < 35) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
if (Rs < 1325) {
return 1;
}
else if (Rs >= 1325) {
return 8;
}
}
}
else if (posBalance >= 35) {
if (ssDCval < 79) {
if (Rs < 1189) {
if (Rs < 1131) {
return 4;
}
else if (Rs >= 1131) {
return 8;
}
}
else if (Rs >= 1189) {
if (AFD < 4) {
if (Rs < 2164) {
return 1;
}
else if (Rs >= 2164) {
return 8;
}
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (ssDCval >= 79) {
if (Rs < 1245) {
return 4;
}
else if (Rs >= 1245) {
return 8;
}
}
}
}
}
else if (Cs >= 911) {
if (ssDCval < 60) {
if (Rs < 1978) {
return 1;
}
else if (Rs >= 1978) {
return 4;
}
}
else if (ssDCval >= 60) {
if (Rs < 1732) {
return 8;
}
else if (Rs >= 1732) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 147) {
return 2;
}
}
}
}
else if (negBalance >= 14) {
if (SC < 1670) {
if (refDCval < 63) {
if (MVDiff < 2) {
if (Rs < 534) {
if (ssDCval < 56) {
if (posBalance < 45) {
if (Cs < 880) {
if (Cs < 759) {
if (Rs < 355) {
return 1;
}
else if (Rs >= 355) {
return 8;
}
}
else if (Cs >= 759) {
return 4;
}
}
else if (Cs >= 880) {
return 2;
}
}
else if (posBalance >= 45) {
return 1;
}
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 111) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 111) {
return 4;
}
}
}
else if (Rs >= 534) {
if (AFD < 3) {
if (RsDiff < 2) {
if (SC < 1633) {
return 8;
}
else if (SC >= 1633) {
if (mu_mv_mag_sq < 123) {
if (Rs < 954) {
return 1;
}
else if (Rs >= 954) {
return 8;
}
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (AFD >= 3) {
if (Rs < 926) {
if (negBalance < 25) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
else if (negBalance >= 25) {
if (ssDCval < 46) {
if (Rs < 819) {
return 4;
}
else if (Rs >= 819) {
return 8;
}
}
else if (ssDCval >= 46) {
if (tcor < 97) {
if (Rs < 672) {
if (Cs < 714) {
return 8;
}
else if (Cs >= 714) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Rs < 628) {
return 8;
}
else if (Rs >= 628) {
return 1;
}
}
}
}
else if (Rs >= 672) {
if (posBalance < 43) {
if (Rs < 774) {
return 8;
}
else if (Rs >= 774) {
return 4;
}
}
else if (posBalance >= 43) {
return 8;
}
}
}
else if (tcor >= 97) {
if (Rs < 860) {
return 8;
}
else if (Rs >= 860) {
return 4;
}
}
}
}
}
else if (Rs >= 926) {
if (tcor < 95) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (tcor >= 95) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (negBalance < 17) {
if (diffRsCsdiff < 1) {
if (ssDCval < 43) {
return 2;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 474) {
return 4;
}
else if (Rs >= 474) {
return 2;
}
}
}
else if (negBalance >= 17) {
if (Rs < 797) {
if (AFD < 8) {
if (negBalance < 28) {
if (TSC < 3) {
if (SC < 1419) {
if (posBalance < 50) {
if (tcor < 93) {
if (Rs < 717) {
return 8;
}
else if (Rs >= 717) {
return 4;
}
}
else if (tcor >= 93) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 97) {
return 1;
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 647) {
return 2;
}
else if (Cs >= 647) {
return 4;
}
}
}
}
else if (SC >= 1419) {
return 1;
}
}
else if (TSC >= 3) {
if (ssDCval < 51) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
return 1;
}
}
else if (ssDCval >= 51) {
return 4;
}
}
}
else if (negBalance >= 28) {
if (Rs < 602) {
if (refDCval < 58) {
if (Rs < 415) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (Rs >= 415) {
if (refDCval < 55) {
return 4;
}
else if (refDCval >= 55) {
if (Cs < 798) {
return 1;
}
else if (Cs >= 798) {
return 4;
}
}
}
}
else if (refDCval >= 58) {
if (Cs < 554) {
return 4;
}
else if (Cs >= 554) {
if (ssDCval < 59) {
if (Cs < 686) {
return 2;
}
else if (Cs >= 686) {
return 4;
}
}
else if (ssDCval >= 59) {
if (Rs < 595) {
return 8;
}
else if (Rs >= 595) {
return 1;
}
}
}
}
}
else if (Rs >= 602) {
if (SC < 1436) {
if (RsDiff < 3) {
if (Rs < 660) {
if (MVDiff < 6) {
if (Rs < 642) {
return 8;
}
else if (Rs >= 642) {
return 4;
}
}
else if (MVDiff >= 6) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 4;
}
}
}
else if (Rs >= 660) {
if (Cs < 679) {
return 8;
}
else if (Cs >= 679) {
return 4;
}
}
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (SC >= 1436) {
if (TSC < 3) {
if (ssDCval < 51) {
if (Rs < 702) {
return 4;
}
else if (Rs >= 702) {
return 1;
}
}
else if (ssDCval >= 51) {
return 8;
}
}
else if (TSC >= 3) {
if (Cs < 919) {
return 4;
}
else if (Cs >= 919) {
return 2;
}
}
}
}
}
}
else if (AFD >= 8) {
if (SC < 1334) {
if (mcTcor < 97) {
if (SC < 1104) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (SC >= 1104) {
if (RsDiff < 4) {
if (negBalance < 36) {
if (Rs < 712) {
return 1;
}
else if (Rs >= 712) {
return 2;
}
}
else if (negBalance >= 36) {
if (diffTSC < -1) {
if (Rs < 495) {
return 8;
}
else if (Rs >= 495) {
return 2;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (RsDiff >= 4) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
else if (SC >= 1334) {
if (negBalance < 42) {
if (ssDCval < 46) {
if (Rs < 682) {
return 8;
}
else if (Rs >= 682) {
return 2;
}
}
else if (ssDCval >= 46) {
return 4;
}
}
else if (negBalance >= 42) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 764) {
return 2;
}
else if (Rs >= 764) {
return 1;
}
}
}
}
}
}
else if (Rs >= 797) {
if (negBalance < 38) {
if (refDCval < 43) {
return 8;
}
else if (refDCval >= 43) {
if (mu_mv_mag_sq < 109) {
if (Cs < 716) {
if (Cs < 554) {
return 2;
}
else if (Cs >= 554) {
return 1;
}
}
else if (Cs >= 716) {
if (SC < 1610) {
return 4;
}
else if (SC >= 1610) {
if (negBalance < 19) {
return 4;
}
else if (negBalance >= 19) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
}
else if (negBalance >= 38) {
if (tcor < 87) {
if (Rs < 881) {
return 1;
}
else if (Rs >= 881) {
return 2;
}
}
else if (tcor >= 87) {
if (MVDiff < 6) {
if (diffRsCsdiff < -7) {
return 8;
}
else if (diffRsCsdiff >= -7) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 63) {
if (mu_mv_mag_sq < 117) {
if (mcTcor < 98) {
if (tcor < 93) {
if (ssDCval < 78) {
if (ssDCval < 72) {
if (posBalance < 52) {
if (ssDCval < 63) {
if (Rs < 379) {
return 4;
}
else if (Rs >= 379) {
return 8;
}
}
else if (ssDCval >= 63) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (Rs < 368) {
return 1;
}
else if (Rs >= 368) {
if (Rs < 794) {
return 4;
}
else if (Rs >= 794) {
if (SC < 1584) {
return 1;
}
else if (SC >= 1584) {
if (SC < 1616) {
return 4;
}
else if (SC >= 1616) {
return 1;
}
}
}
}
}
}
}
else if (posBalance >= 52) {
if (MVDiff < 33) {
if (diffMVdiffVal < 11) {
if (MVDiff < 6) {
if (Cs < 703) {
return 2;
}
else if (Cs >= 703) {
return 8;
}
}
else if (MVDiff >= 6) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
if (Cs < 612) {
return 1;
}
else if (Cs >= 612) {
if (Cs < 905) {
return 2;
}
else if (Cs >= 905) {
return 1;
}
}
}
}
}
else if (diffMVdiffVal >= 11) {
return 8;
}
}
else if (MVDiff >= 33) {
if (Rs < 448) {
return 2;
}
else if (Rs >= 448) {
return 4;
}
}
}
}
else if (ssDCval >= 72) {
if (diffAFD < -1) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (Rs < 719) {
return 8;
}
else if (Rs >= 719) {
return 1;
}
}
}
else if (diffAFD >= -1) {
if (Cs < 703) {
if (ssDCval < 73) {
if (Cs < 619) {
return 4;
}
else if (Cs >= 619) {
return 8;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (Cs >= 703) {
return 1;
}
}
}
}
else if (ssDCval >= 78) {
if (negBalance < 44) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (tcor < 87) {
return 2;
}
else if (tcor >= 87) {
if (mcTcor < 97) {
if (Cs < 942) {
return 8;
}
else if (Cs >= 942) {
return 2;
}
}
else if (mcTcor >= 97) {
return 2;
}
}
}
}
}
else if (negBalance >= 44) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 5) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 1;
}
}
else if (mu_mv_mag_sq >= 5) {
if (ssDCval < 117) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 4;
}
}
else if (ssDCval >= 117) {
if (Rs < 765) {
return 1;
}
else if (Rs >= 765) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
if (refDCval < 102) {
if (SC < 1502) {
if (Rs < 663) {
if (Rs < 573) {
return 1;
}
else if (Rs >= 573) {
return 2;
}
}
else if (Rs >= 663) {
return 8;
}
}
else if (SC >= 1502) {
return 4;
}
}
else if (refDCval >= 102) {
if (MVDiff < 14) {
if (ssDCval < 115) {
return 2;
}
else if (ssDCval >= 115) {
if (Cs < 766) {
return 8;
}
else if (Cs >= 766) {
if (Rs < 546) {
return 1;
}
else if (Rs >= 546) {
return 2;
}
}
}
}
else if (MVDiff >= 14) {
if (diffRsCsdiff < -17) {
if (Rs < 583) {
return 2;
}
else if (Rs >= 583) {
return 4;
}
}
else if (diffRsCsdiff >= -17) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 93) {
if (MVDiff < 6) {
if (Rs < 384) {
if (ssDCval < 96) {
if (negBalance < 37) {
if (Cs < 718) {
return 4;
}
else if (Cs >= 718) {
return 8;
}
}
else if (negBalance >= 37) {
return 1;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (Rs >= 384) {
if (TSC < 3) {
if (Cs < 743) {
if (ssDCval < 66) {
if (Cs < 700) {
return 1;
}
else if (Cs >= 700) {
return 4;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (Cs >= 743) {
if (SC < 1644) {
return 4;
}
else if (SC >= 1644) {
return 8;
}
}
}
else if (TSC >= 3) {
if (ssDCval < 76) {
if (ssDCval < 66) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (ssDCval >= 66) {
if (Rs < 775) {
if (Cs < 594) {
if (Rs < 474) {
return 2;
}
else if (Rs >= 474) {
return 4;
}
}
else if (Cs >= 594) {
if (Rs < 652) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 512) {
return 2;
}
else if (Rs >= 512) {
if (Rs < 598) {
return 8;
}
else if (Rs >= 598) {
return 2;
}
}
}
}
else if (Rs >= 652) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
return 4;
}
}
}
}
}
else if (Rs >= 775) {
return 4;
}
}
}
else if (ssDCval >= 76) {
if (negBalance < 43) {
if (Rs < 590) {
if (diffAFD < -1) {
if (SC < 1325) {
return 8;
}
else if (SC >= 1325) {
if (Rs < 511) {
return 1;
}
else if (Rs >= 511) {
if (Rs < 561) {
return 4;
}
else if (Rs >= 561) {
return 8;
}
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Rs >= 590) {
if (ssDCval < 97) {
if (CsDiff < 3) {
if (Cs < 895) {
return 2;
}
else if (Cs >= 895) {
return 1;
}
}
else if (CsDiff >= 3) {
if (Cs < 891) {
return 8;
}
else if (Cs >= 891) {
return 4;
}
}
}
else if (ssDCval >= 97) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 8;
}
}
}
}
else if (negBalance >= 43) {
if (Rs < 499) {
if (mu_mv_mag_sq < 74) {
return 4;
}
else if (mu_mv_mag_sq >= 74) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (Rs >= 499) {
if (Rs < 733) {
if (Cs < 649) {
return 2;
}
else if (Cs >= 649) {
if (SC < 1612) {
return 8;
}
else if (SC >= 1612) {
return 4;
}
}
}
else if (Rs >= 733) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (negBalance < 30) {
if (Cs < 646) {
if (Rs < 699) {
return 2;
}
else if (Rs >= 699) {
return 1;
}
}
else if (Cs >= 646) {
if (SC < 1171) {
return 8;
}
else if (SC >= 1171) {
return 4;
}
}
}
else if (negBalance >= 30) {
if (Rs < 343) {
return 2;
}
else if (Rs >= 343) {
if (Rs < 753) {
if (Cs < 552) {
if (Rs < 508) {
return 2;
}
else if (Rs >= 508) {
return 8;
}
}
else if (Cs >= 552) {
if (RsCsDiff < 23) {
return 8;
}
else if (RsCsDiff >= 23) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (Rs >= 753) {
if (tcor < 95) {
if (Rs < 876) {
return 2;
}
else if (Rs >= 876) {
return 1;
}
}
else if (tcor >= 95) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 60) {
if (MVDiff < 4) {
if (Rs < 413) {
if (ssDCval < 124) {
if (SC < 1256) {
return 4;
}
else if (SC >= 1256) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (ssDCval >= 124) {
if (Rs < 313) {
return 2;
}
else if (Rs >= 313) {
return 8;
}
}
}
else if (Rs >= 413) {
if (Cs < 769) {
if (Rs < 807) {
if (mu_mv_mag_sq < 97) {
if (Rs < 526) {
if (Rs < 432) {
return 4;
}
else if (Rs >= 432) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (Rs >= 526) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (posBalance < 34) {
return 1;
}
else if (posBalance >= 34) {
if (SC < 1355) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 598) {
return 1;
}
else if (Rs >= 598) {
return 2;
}
}
}
else if (SC >= 1355) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (ssDCval < 117) {
if (ssDCval < 82) {
if (Cs < 730) {
if (Rs < 623) {
if (Rs < 507) {
return 2;
}
else if (Rs >= 507) {
return 8;
}
}
else if (Rs >= 623) {
if (Cs < 713) {
return 1;
}
else if (Cs >= 713) {
return 8;
}
}
}
else if (Cs >= 730) {
return 4;
}
}
else if (ssDCval >= 82) {
return 1;
}
}
else if (ssDCval >= 117) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (Rs >= 807) {
if (mu_mv_mag_sq < 90) {
return 1;
}
else if (mu_mv_mag_sq >= 90) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
}
else if (Cs >= 769) {
if (TSC < 3) {
if (posBalance < 47) {
if (diffMVdiffVal < 1) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (posBalance >= 47) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
if (Rs < 498) {
return 8;
}
else if (Rs >= 498) {
return 2;
}
}
}
}
else if (TSC >= 3) {
if (ssDCval < 99) {
if (negBalance < 33) {
if (Rs < 649) {
return 2;
}
else if (Rs >= 649) {
return 1;
}
}
else if (negBalance >= 33) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (ssDCval >= 99) {
if (tcor < 98) {
if (Rs < 660) {
return 1;
}
else if (Rs >= 660) {
return 4;
}
}
else if (tcor >= 98) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 89) {
if (RsCsDiff < 19) {
if (mu_mv_mag_sq < 5) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (mu_mv_mag_sq >= 5) {
if (mu_mv_mag_sq < 69) {
return 1;
}
else if (mu_mv_mag_sq >= 69) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
else if (RsCsDiff >= 19) {
if (Rs < 285) {
return 1;
}
else if (Rs >= 285) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (Cs < 863) {
return 8;
}
else if (Cs >= 863) {
return 1;
}
}
}
}
else if (ssDCval >= 84) {
if (negBalance < 19) {
if (Rs < 456) {
return 4;
}
else if (Rs >= 456) {
return 1;
}
}
else if (negBalance >= 19) {
if (refDCval < 127) {
if (SC < 1290) {
if (TSC < 6) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
if (mu_mv_mag_sq < 100) {
if (Rs < 361) {
if (Rs < 342) {
return 2;
}
else if (Rs >= 342) {
return 4;
}
}
else if (Rs >= 361) {
return 2;
}
}
else if (mu_mv_mag_sq >= 100) {
return 4;
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (SC >= 1290) {
if (Rs < 703) {
if (mu_mv_mag_sq < 96) {
if (negBalance < 47) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 604) {
return 1;
}
else if (Rs >= 604) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (Rs < 630) {
return 2;
}
else if (Rs >= 630) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (mu_mv_mag_sq < 109) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
}
else if (Rs >= 703) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
}
}
else if (refDCval >= 127) {
if (refDCval < 153) {
if (ssDCval < 130) {
if (Rs < 480) {
return 8;
}
else if (Rs >= 480) {
return 1;
}
}
else if (ssDCval >= 130) {
if (tcor < 86) {
if (Rs < 780) {
return 1;
}
else if (Rs >= 780) {
return 8;
}
}
else if (tcor >= 86) {
if (Rs < 339) {
return 2;
}
else if (Rs >= 339) {
if (Cs < 679) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
return 8;
}
}
else if (Cs >= 679) {
return 2;
}
}
}
}
}
else if (refDCval >= 153) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 503) {
return 1;
}
else if (Rs >= 503) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 60) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Cs < 696) {
if (Rs < 599) {
if (ssDCval < 119) {
return 8;
}
else if (ssDCval >= 119) {
return 2;
}
}
else if (Rs >= 599) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
return 8;
}
}
}
else if (Cs >= 696) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (ssDCval < 88) {
if (Rs < 830) {
if (SC < 1490) {
if (ssDCval < 64) {
if (Rs < 729) {
return 4;
}
else if (Rs >= 729) {
return 8;
}
}
else if (ssDCval >= 64) {
if (posBalance < 47) {
if (negBalance < 28) {
if (diffAFD < -2) {
if (Rs < 781) {
return 1;
}
else if (Rs >= 781) {
return 8;
}
}
else if (diffAFD >= -2) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
if (SC < 1398) {
return 8;
}
else if (SC >= 1398) {
return 2;
}
}
}
}
else if (negBalance >= 28) {
if (Rs < 701) {
return 8;
}
else if (Rs >= 701) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (Rs < 684) {
return 8;
}
else if (Rs >= 684) {
return 1;
}
}
}
}
else if (SC >= 1490) {
if (mu_mv_mag_sq < 121) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Cs < 858) {
return 1;
}
else if (Cs >= 858) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 121) {
return 1;
}
}
}
else if (Rs >= 830) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (ssDCval < 75) {
if (Cs < 786) {
return 2;
}
else if (Cs >= 786) {
return 4;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
}
}
else if (ssDCval >= 88) {
if (Rs < 466) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
if (Rs < 316) {
return 2;
}
else if (Rs >= 316) {
return 8;
}
}
}
else if (Rs >= 466) {
if (Rs < 1033) {
return 8;
}
else if (Rs >= 1033) {
return 4;
}
}
}
}
}
}
else if (SC >= 1670) {
if (Rs < 1796) {
if (ssDCval < 69) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 127) {
if (negBalance < 45) {
if (ssDCval < 48) {
if (RsDiff < 3) {
if (Rs < 1194) {
return 8;
}
else if (Rs >= 1194) {
if (mu_mv_mag_sq < 118) {
return 4;
}
else if (mu_mv_mag_sq >= 118) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (Rs < 1338) {
return 8;
}
else if (Rs >= 1338) {
return 8;
}
}
}
}
}
else if (RsDiff >= 3) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (ssDCval >= 48) {
if (mcTcor < 97) {
if (RsCsDiff < 12) {
if (posBalance < 44) {
if (mu_mv_mag_sq < 123) {
if (diffMVdiffVal < -2) {
if (ssDCval < 52) {
if (Rs < 1139) {
return 8;
}
else if (Rs >= 1139) {
return 1;
}
}
else if (ssDCval >= 52) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (posBalance < 30) {
return 4;
}
else if (posBalance >= 30) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
else if (posBalance >= 44) {
if (negBalance < 27) {
if (Rs < 1285) {
return 1;
}
else if (Rs >= 1285) {
return 8;
}
}
else if (negBalance >= 27) {
if (refDCval < 67) {
return 8;
}
else if (refDCval >= 67) {
if (Rs < 1333) {
return 1;
}
else if (Rs >= 1333) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 12) {
if (tcor < 96) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Cs < 822) {
if (RsDiff < 5) {
if (MVDiff < 5) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (SC < 1831) {
return 4;
}
else if (SC >= 1831) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
}
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (Cs >= 822) {
if (AFD < 7) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
else if (AFD >= 7) {
if (TSC < 5) {
if (Rs < 957) {
return 4;
}
else if (Rs >= 957) {
return 1;
}
}
else if (TSC >= 5) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (Rs < 1129) {
return 8;
}
else if (Rs >= 1129) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 96) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (Rs < 905) {
if (Rs < 812) {
if (Rs < 740) {
return 2;
}
else if (Rs >= 740) {
return 8;
}
}
else if (Rs >= 812) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (Rs >= 905) {
if (SC < 2438) {
if (MVDiff < 15) {
if (mu_mv_mag_sq < 125) {
if (RsDiff < 4) {
if (TSC < 2) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
else if (TSC >= 2) {
if (negBalance < 41) {
if (RsDiff < 3) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 108) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (diffRsCsdiff < -8) {
if (Rs < 1046) {
return 1;
}
else if (Rs >= 1046) {
return 8;
}
}
else if (diffRsCsdiff >= -8) {
if (SC < 2361) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
return 1;
}
}
else if (SC >= 2361) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 1265) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (Rs >= 1265) {
return 8;
}
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (RsDiff >= 3) {
if (diffRsCsdiff < -4) {
if (Rs < 1550) {
return 4;
}
else if (Rs >= 1550) {
return 1;
}
}
else if (diffRsCsdiff >= -4) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
if (Rs < 1134) {
return 1;
}
else if (Rs >= 1134) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 41) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
else if (RsDiff >= 4) {
if (Cs < 802) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
else if (Cs >= 802) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1122) {
return 4;
}
else if (Rs >= 1122) {
return 8;
}
}
}
else if (MVDiff >= 15) {
if (Cs < 726) {
if (Rs < 1129) {
return 4;
}
else if (Rs >= 1129) {
return 2;
}
}
else if (Cs >= 726) {
if (Rs < 1344) {
return 8;
}
else if (Rs >= 1344) {
return 1;
}
}
}
}
else if (SC >= 2438) {
if (Rs < 1715) {
return 8;
}
else if (Rs >= 1715) {
if (Rs < 1760) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (Rs >= 1760) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 1314) {
if (negBalance < 49) {
if (RsDiff < 3) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (RsDiff >= 3) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (negBalance >= 49) {
return 8;
}
}
else if (Rs >= 1314) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 1624) {
return 8;
}
else if (Rs >= 1624) {
return 1;
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 56) {
if (SC < 2052) {
if (refDCval < 40) {
return 8;
}
else if (refDCval >= 40) {
if (Rs < 1008) {
return 1;
}
else if (Rs >= 1008) {
if (Cs < 624) {
return 1;
}
else if (Cs >= 624) {
if (Rs < 1171) {
return 8;
}
else if (Rs >= 1171) {
return 1;
}
}
}
}
}
else if (SC >= 2052) {
if (MVDiff < 7) {
if (Rs < 1478) {
return 8;
}
else if (Rs >= 1478) {
if (Rs < 1547) {
return 4;
}
else if (Rs >= 1547) {
return 4;
}
}
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (ssDCval >= 56) {
if (SC < 1883) {
if (refDCval < 66) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
return 4;
}
}
else if (refDCval >= 66) {
if (Rs < 1128) {
if (Rs < 884) {
return 4;
}
else if (Rs >= 884) {
return 2;
}
}
else if (Rs >= 1128) {
if (Cs < 720) {
return 1;
}
else if (Cs >= 720) {
return 4;
}
}
}
}
else if (SC >= 1883) {
if (gchDC < 1) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (Cs < 904) {
if (Rs < 1572) {
return 8;
}
else if (Rs >= 1572) {
return 1;
}
}
else if (Cs >= 904) {
return 1;
}
}
}
}
else if (gchDC >= 1) {
if (SC < 2411) {
if (TSC < 4) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (SC >= 2411) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1312) {
if (posBalance < 65) {
if (Rs < 891) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
else if (Rs >= 891) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 2) {
if (CsDiff < 3) {
if (Cs < 631) {
return 1;
}
else if (Cs >= 631) {
return 4;
}
}
else if (CsDiff >= 3) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
if (posBalance < 36) {
return 2;
}
else if (posBalance >= 36) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (diffMVdiffVal < -3) {
if (Rs < 1171) {
return 8;
}
else if (Rs >= 1171) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
if (ssDCval < 141) {
return 8;
}
else if (ssDCval >= 141) {
return 1;
}
}
}
}
}
else if (posBalance >= 65) {
if (Rs < 1018) {
return 8;
}
else if (Rs >= 1018) {
return 2;
}
}
}
else if (Rs >= 1312) {
if (ssDCval < 107) {
if (mcTcor < 96) {
if (negBalance < 46) {
return 1;
}
else if (negBalance >= 46) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (Rs < 1565) {
return 8;
}
else if (Rs >= 1565) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (ssDCval >= 107) {
if (SC < 2128) {
return 1;
}
else if (SC >= 2128) {
if (negBalance < 40) {
if (ssDCval < 133) {
return 2;
}
else if (ssDCval >= 133) {
return 8;
}
}
else if (negBalance >= 40) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 79) {
if (mu_mv_mag_sq < 66) {
if (posBalance < 51) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (posBalance >= 51) {
if (Cs < 736) {
if (MVDiff < 22) {
return 1;
}
else if (MVDiff >= 22) {
return 4;
}
}
else if (Cs >= 736) {
if (Rs < 1601) {
return 8;
}
else if (Rs >= 1601) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (Rs < 1305) {
if (AFD < 3) {
if (Rs < 1009) {
if (Rs < 937) {
return 4;
}
else if (Rs >= 937) {
if (Rs < 1003) {
return 2;
}
else if (Rs >= 1003) {
return 8;
}
}
}
else if (Rs >= 1009) {
if (SC < 1745) {
if (Rs < 1117) {
return 2;
}
else if (Rs >= 1117) {
return 1;
}
}
else if (SC >= 1745) {
return 1;
}
}
}
else if (AFD >= 3) {
if (negBalance < 68) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 116) {
return 8;
}
else if (mu_mv_mag_sq >= 116) {
if (Rs < 1213) {
return 1;
}
else if (Rs >= 1213) {
return 8;
}
}
}
}
else if (negBalance >= 68) {
return 4;
}
}
}
else if (Rs >= 1305) {
if (mu_mv_mag_sq < 113) {
if (Cs < 756) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 756) {
if (SC < 2232) {
return 1;
}
else if (SC >= 2232) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 1566) {
return 1;
}
else if (Rs >= 1566) {
if (Rs < 1644) {
return 2;
}
else if (Rs >= 1644) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (refDCval < 75) {
return 8;
}
else if (refDCval >= 75) {
if (Cs < 713) {
return 1;
}
else if (Cs >= 713) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 79) {
if (tcor < 99) {
if (MVDiff < 27) {
if (Rs < 917) {
if (negBalance < 36) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
else if (negBalance >= 36) {
if (Cs < 942) {
return 1;
}
else if (Cs >= 942) {
if (AFD < 9) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
if (Rs < 760) {
return 4;
}
else if (Rs >= 760) {
return 1;
}
}
}
else if (AFD >= 9) {
return 2;
}
}
}
}
else if (Rs >= 917) {
if (SC < 2551) {
if (refDCval < 95) {
if (mu_mv_mag_sq < 69) {
if (TSC < 7) {
if (Rs < 1211) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (Rs >= 1211) {
return 1;
}
}
else if (TSC >= 7) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (diffRsCsdiff < -25) {
return 4;
}
else if (diffRsCsdiff >= -25) {
if (Rs < 983) {
if (Cs < 858) {
return 8;
}
else if (Cs >= 858) {
if (Rs < 960) {
return 1;
}
else if (Rs >= 960) {
return 4;
}
}
}
else if (Rs >= 983) {
return 8;
}
}
}
}
else if (refDCval >= 95) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 111) {
if (ssDCval < 157) {
if (MVDiff < 25) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 1057) {
return 1;
}
else if (Rs >= 1057) {
return 8;
}
}
}
else if (MVDiff >= 25) {
if (Cs < 894) {
return 8;
}
else if (Cs >= 894) {
return 4;
}
}
}
else if (ssDCval >= 157) {
return 2;
}
}
else if (mu_mv_mag_sq >= 111) {
if (posBalance < 58) {
if (ssDCval < 125) {
return 8;
}
else if (ssDCval >= 125) {
return 4;
}
}
else if (posBalance >= 58) {
if (Rs < 1233) {
return 8;
}
else if (Rs >= 1233) {
if (ssDCval < 110) {
return 8;
}
else if (ssDCval >= 110) {
return 1;
}
}
}
}
}
else if (diffAFD >= 2) {
if (MVDiff < 16) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (MVDiff >= 16) {
return 1;
}
}
}
}
else if (SC >= 2551) {
if (ssDCval < 115) {
if (Cs < 834) {
if (Rs < 1770) {
return 4;
}
else if (Rs >= 1770) {
return 8;
}
}
else if (Cs >= 834) {
if (Rs < 1686) {
return 8;
}
else if (Rs >= 1686) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 1;
}
}
}
}
else if (ssDCval >= 115) {
if (posBalance < 47) {
if (Cs < 947) {
return 4;
}
else if (Cs >= 947) {
return 2;
}
}
else if (posBalance >= 47) {
return 8;
}
}
}
}
}
else if (MVDiff >= 27) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
if (Cs < 745) {
if (Rs < 1198) {
return 8;
}
else if (Rs >= 1198) {
return 1;
}
}
else if (Cs >= 745) {
if (Cs < 921) {
return 8;
}
else if (Cs >= 921) {
return 4;
}
}
}
}
}
else if (tcor >= 99) {
if (ssDCval < 100) {
if (negBalance < 33) {
if (Rs < 986) {
return 8;
}
else if (Rs >= 986) {
return 4;
}
}
else if (negBalance >= 33) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 8;
}
}
}
else if (ssDCval >= 100) {
if (ssDCval < 125) {
if (SC < 2520) {
return 8;
}
else if (SC >= 2520) {
return 1;
}
}
else if (ssDCval >= 125) {
if (Rs < 977) {
return 8;
}
else if (Rs >= 977) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 1796) {
if (TSC < 4) {
if (ssDCval < 122) {
if (ssDCval < 88) {
if (refDCval < 83) {
if (refDCval < 56) {
if (mu_mv_mag_sq < 104) {
return 4;
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 2396) {
if (Rs < 1985) {
return 4;
}
else if (Rs >= 1985) {
return 8;
}
}
else if (Rs >= 2396) {
return 1;
}
}
}
else if (refDCval >= 56) {
if (mcTcor < 99) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (MVDiff < 3) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (mcTcor >= 99) {
if (TSC < 2) {
if (SC < 2974) {
return 1;
}
else if (SC >= 2974) {
return 4;
}
}
else if (TSC >= 2) {
if (SC < 2915) {
return 4;
}
else if (SC >= 2915) {
return 1;
}
}
}
}
}
else if (refDCval >= 83) {
if (MVDiff < 2) {
if (Rs < 2195) {
return 4;
}
else if (Rs >= 2195) {
return 8;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 97) {
if (ssDCval < 100) {
return 1;
}
else if (ssDCval >= 100) {
return 8;
}
}
else if (mu_mv_mag_sq >= 97) {
if (SC < 2644) {
return 4;
}
else if (SC >= 2644) {
if (Rs < 2147) {
return 8;
}
else if (Rs >= 2147) {
if (Cs < 764) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (Cs >= 764) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 122) {
if (tcor < 97) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (Rs < 1845) {
return 2;
}
else if (Rs >= 1845) {
return 8;
}
}
}
else if (tcor >= 97) {
return 1;
}
}
}
else if (TSC >= 4) {
if (refDCval < 71) {
if (posBalance < 49) {
if (negBalance < 40) {
if (Cs < 846) {
if (mu_mv_mag_sq < 107) {
return 1;
}
else if (mu_mv_mag_sq >= 107) {
return 2;
}
}
else if (Cs >= 846) {
if (Rs < 2037) {
return 4;
}
else if (Rs >= 2037) {
return 1;
}
}
}
else if (negBalance >= 40) {
if (RsDiff < 7) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 2;
}
}
else if (RsDiff >= 7) {
return 2;
}
}
}
else if (posBalance >= 49) {
if (Rs < 1885) {
return 8;
}
else if (Rs >= 1885) {
return 2;
}
}
}
else if (refDCval >= 71) {
if (refDCval < 102) {
if (mu_mv_mag_sq < 85) {
if (RsCsDiff < 37) {
if (posBalance < 33) {
if (Cs < 737) {
if (Rs < 2241) {
return 4;
}
else if (Rs >= 2241) {
return 4;
}
}
else if (Cs >= 737) {
return 1;
}
}
else if (posBalance >= 33) {
return 1;
}
}
else if (RsCsDiff >= 37) {
if (SC < 2910) {
if (refDCval < 91) {
if (Rs < 2123) {
return 1;
}
else if (Rs >= 2123) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (refDCval >= 91) {
if (negBalance < 52) {
if (Rs < 2019) {
return 1;
}
else if (Rs >= 2019) {
return 4;
}
}
else if (negBalance >= 52) {
return 2;
}
}
}
else if (SC >= 2910) {
if (ssDCval < 83) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
return 2;
}
}
else if (ssDCval >= 83) {
if (MVDiff < 18) {
if (Cs < 778) {
return 4;
}
else if (Cs >= 778) {
return 4;
}
}
else if (MVDiff >= 18) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (posBalance < 54) {
if (SC < 2683) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (SC >= 2683) {
if (ssDCval < 77) {
if (Rs < 2222) {
return 2;
}
else if (Rs >= 2222) {
if (Rs < 2321) {
return 8;
}
else if (Rs >= 2321) {
return 4;
}
}
}
else if (ssDCval >= 77) {
return 8;
}
}
}
else if (posBalance >= 54) {
return 2;
}
}
}
else if (refDCval >= 102) {
if (Cs < 828) {
if (refDCval < 118) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
else if (refDCval >= 118) {
if (mu_mv_mag_sq < 37) {
return 1;
}
else if (mu_mv_mag_sq >= 37) {
if (gchDC < 2) {
if (Rs < 2058) {
return 2;
}
else if (Rs >= 2058) {
return 1;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
else if (Cs >= 828) {
if (tcor < 97) {
if (diffRsCsdiff < -55) {
return 2;
}
else if (diffRsCsdiff >= -55) {
return 8;
}
}
else if (tcor >= 97) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 974) {
if (posBalance < 22) {
if (Rs < 758) {
if (Rs < 729) {
return 8;
}
else if (Rs >= 729) {
return 2;
}
}
else if (Rs >= 758) {
if (MVDiff < 15) {
if (negBalance < 7) {
if (Cs < 1213) {
if (diffAFD < -9) {
if (Rs < 1098) {
return 1;
}
else if (Rs >= 1098) {
return 8;
}
}
else if (diffAFD >= -9) {
return 8;
}
}
else if (Cs >= 1213) {
if (Rs < 1224) {
if (Rs < 1044) {
return 8;
}
else if (Rs >= 1044) {
return 1;
}
}
else if (Rs >= 1224) {
if (diffRsCsdiff < -213) {
return 1;
}
else if (diffRsCsdiff >= -213) {
return 8;
}
}
}
}
else if (negBalance >= 7) {
if (ssDCval < 63) {
if (SC < 3007) {
if (Cs < 1016) {
return 1;
}
else if (Cs >= 1016) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 1216) {
return 8;
}
else if (Rs >= 1216) {
return 1;
}
}
}
}
else if (SC >= 3007) {
return 4;
}
}
else if (ssDCval >= 63) {
if (diffRsCsdiff < 11) {
return 8;
}
else if (diffRsCsdiff >= 11) {
return 4;
}
}
}
}
else if (MVDiff >= 15) {
if (ssDCval < 133) {
if (diffAFD < -14) {
return 1;
}
else if (diffAFD >= -14) {
return 8;
}
}
else if (ssDCval >= 133) {
if (refDCval < 141) {
return 4;
}
else if (refDCval >= 141) {
if (MVDiff < 26) {
if (Rs < 1166) {
return 2;
}
else if (Rs >= 1166) {
return 8;
}
}
else if (MVDiff >= 26) {
if (Rs < 1638) {
return 4;
}
else if (Rs >= 1638) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 22) {
if (TSC < 3) {
if (ssDCval < 87) {
if (RsDiff < 1) {
if (Rs < 1858) {
if (ssDCval < 52) {
if (SC < 2079) {
if (Rs < 437) {
return 2;
}
else if (Rs >= 437) {
if (ssDCval < 45) {
return 4;
}
else if (ssDCval >= 45) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
return 8;
}
}
}
}
else if (SC >= 2079) {
if (diffRsCsdiff < -2) {
if (Rs < 1416) {
return 8;
}
else if (Rs >= 1416) {
return 8;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (Rs < 683) {
if (Rs < 548) {
return 8;
}
else if (Rs >= 548) {
return 2;
}
}
else if (Rs >= 683) {
if (mu_mv_mag_sq < 100) {
if (Rs < 1028) {
return 8;
}
else if (Rs >= 1028) {
return 4;
}
}
else if (mu_mv_mag_sq >= 100) {
if (ssDCval < 80) {
if (Rs < 1443) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
if (Rs < 1010) {
return 1;
}
else if (Rs >= 1010) {
return 8;
}
}
}
else if (Rs >= 1443) {
if (Rs < 1597) {
return 4;
}
else if (Rs >= 1597) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 1;
}
}
}
}
else if (ssDCval >= 80) {
return 4;
}
}
}
}
}
else if (Rs >= 1858) {
if (Rs < 2154) {
if (ssDCval < 72) {
if (Rs < 1884) {
if (Rs < 1864) {
return 4;
}
else if (Rs >= 1864) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
}
else if (Rs >= 1884) {
return 4;
}
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (Rs >= 2154) {
return 2;
}
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 101) {
if (Rs < 732) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Rs >= 732) {
if (ssDCval < 61) {
if (RsDiff < 2) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (posBalance < 32) {
return 1;
}
else if (posBalance >= 32) {
return 4;
}
}
}
else if (RsDiff >= 2) {
if (refDCval < 61) {
return 4;
}
else if (refDCval >= 61) {
return 2;
}
}
}
else if (ssDCval >= 61) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (mu_mv_mag_sq < 124) {
if (ssDCval < 81) {
if (Rs < 1804) {
if (Rs < 917) {
if (Rs < 443) {
if (Rs < 386) {
return 2;
}
else if (Rs >= 386) {
return 8;
}
}
else if (Rs >= 443) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (Rs < 834) {
return 1;
}
else if (Rs >= 834) {
if (Rs < 883) {
return 2;
}
else if (Rs >= 883) {
return 4;
}
}
}
}
}
}
else if (Rs >= 917) {
if (ssDCval < 52) {
if (negBalance < 35) {
if (posBalance < 50) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Cs < 1037) {
return 1;
}
else if (Cs >= 1037) {
return 8;
}
}
}
else if (posBalance >= 50) {
return 4;
}
}
else if (negBalance >= 35) {
return 4;
}
}
else if (ssDCval >= 52) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 123) {
if (diffTSC < -1) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
return 1;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 123) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 1;
}
}
}
else if (ssDCval >= 69) {
if (negBalance < 20) {
if (Rs < 1302) {
return 8;
}
else if (Rs >= 1302) {
return 1;
}
}
else if (negBalance >= 20) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
if (Rs < 1015) {
return 4;
}
else if (Rs >= 1015) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1804) {
if (tcor < 99) {
if (TSCindex < 3) {
if (Rs < 1949) {
return 1;
}
else if (Rs >= 1949) {
if (Rs < 1983) {
return 4;
}
else if (Rs >= 1983) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (Rs < 2203) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (Rs >= 2203) {
return 8;
}
}
}
else if (tcor >= 99) {
return 4;
}
}
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (mu_mv_mag_sq >= 124) {
if (SC < 2581) {
if (negBalance < 26) {
if (CsDiff < 1) {
if (Rs < 1331) {
return 1;
}
else if (Rs >= 1331) {
return 8;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (negBalance >= 26) {
if (refDCval < 53) {
if (Rs < 900) {
return 1;
}
else if (Rs >= 900) {
return 8;
}
}
else if (refDCval >= 53) {
if (Cs < 1030) {
return 8;
}
else if (Cs >= 1030) {
if (tcor < 98) {
if (Rs < 1343) {
return 1;
}
else if (Rs >= 1343) {
return 8;
}
}
else if (tcor >= 98) {
if (Cs < 1202) {
return 4;
}
else if (Cs >= 1202) {
if (Rs < 1233) {
return 4;
}
else if (Rs >= 1233) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 2581) {
if (ssDCval < 75) {
if (ssDCval < 53) {
if (negBalance < 13) {
return 8;
}
else if (negBalance >= 13) {
if (Cs < 1176) {
if (ssDCval < 48) {
if (Rs < 1731) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
if (SC < 2839) {
if (Rs < 1651) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
return 4;
}
}
else if (Rs >= 1651) {
if (Rs < 1655) {
return 8;
}
else if (Rs >= 1655) {
if (SC < 2811) {
if (Rs < 1685) {
return 4;
}
else if (Rs >= 1685) {
if (Rs < 1724) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (Rs < 1706) {
return 4;
}
else if (Rs >= 1706) {
if (Rs < 1720) {
return 4;
}
else if (Rs >= 1720) {
return 4;
}
}
}
}
else if (Rs >= 1724) {
return 4;
}
}
}
else if (SC >= 2811) {
if (SC < 2814) {
return 8;
}
else if (SC >= 2814) {
if (CsDiff < 2) {
if (Rs < 1675) {
return 4;
}
else if (Rs >= 1675) {
if (Rs < 1683) {
return 8;
}
else if (Rs >= 1683) {
if (Rs < 1691) {
return 4;
}
else if (Rs >= 1691) {
return 4;
}
}
}
}
else if (CsDiff >= 2) {
return 8;
}
}
}
}
}
}
else if (SC >= 2839) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
return 4;
}
}
}
}
else if (Rs >= 1731) {
if (Rs < 1741) {
if (Rs < 1734) {
if (Rs < 1733) {
return 8;
}
else if (Rs >= 1733) {
return 4;
}
}
else if (Rs >= 1734) {
return 8;
}
}
else if (Rs >= 1741) {
if (CsDiff < 2) {
if (Cs < 1035) {
if (Rs < 1813) {
return 8;
}
else if (Rs >= 1813) {
return 8;
}
}
else if (Cs >= 1035) {
if (posBalance < 30) {
return 4;
}
else if (posBalance >= 30) {
if (refDCval < 48) {
if (Rs < 1828) {
if (SC < 2804) {
return 8;
}
else if (SC >= 2804) {
if (Rs < 1744) {
return 4;
}
else if (Rs >= 1744) {
if (SC < 2881) {
if (AFD < 2) {
if (SC < 2866) {
if (Cs < 1046) {
if (Rs < 1775) {
return 4;
}
else if (Rs >= 1775) {
return 4;
}
}
else if (Cs >= 1046) {
if (tcor < 99) {
if (Rs < 1775) {
return 4;
}
else if (Rs >= 1775) {
return 4;
}
}
else if (tcor >= 99) {
return 4;
}
}
}
else if (SC >= 2866) {
if (Rs < 1771) {
return 8;
}
else if (Rs >= 1771) {
return 4;
}
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (SC >= 2881) {
if (AFD < 2) {
if (SC < 2899) {
if (Rs < 1770) {
return 8;
}
else if (Rs >= 1770) {
if (Rs < 1788) {
return 4;
}
else if (Rs >= 1788) {
return 8;
}
}
}
else if (SC >= 2899) {
if (negBalance < 20) {
if (negBalance < 19) {
if (diffAFD < -1) {
if (Rs < 1769) {
return 4;
}
else if (Rs >= 1769) {
return 4;
}
}
else if (diffAFD >= -1) {
if (Rs < 1795) {
if (Rs < 1768) {
return 4;
}
else if (Rs >= 1768) {
return 8;
}
}
else if (Rs >= 1795) {
if (Rs < 1819) {
return 4;
}
else if (Rs >= 1819) {
return 4;
}
}
}
}
else if (negBalance >= 19) {
return 8;
}
}
else if (negBalance >= 20) {
return 4;
}
}
}
else if (AFD >= 2) {
if (SC < 2884) {
return 4;
}
else if (SC >= 2884) {
if (posBalance < 34) {
return 4;
}
else if (posBalance >= 34) {
if (Rs < 1763) {
return 4;
}
else if (Rs >= 1763) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 1828) {
return 8;
}
}
else if (refDCval >= 48) {
return 4;
}
}
}
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (Cs >= 1176) {
if (Rs < 1669) {
if (Rs < 1380) {
return 4;
}
else if (Rs >= 1380) {
return 8;
}
}
else if (Rs >= 1669) {
if (Rs < 2043) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (Rs >= 2043) {
return 8;
}
}
}
}
}
else if (ssDCval >= 53) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 4;
}
}
}
else if (ssDCval >= 75) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (MVDiff < 1) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (MVDiff >= 1) {
if (ssDCval < 112) {
if (Cs < 1176) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Cs >= 1176) {
return 1;
}
}
else if (ssDCval >= 112) {
if (RsDiff < 2) {
if (posBalance < 50) {
if (SC < 2545) {
if (negBalance < 41) {
return 4;
}
else if (negBalance >= 41) {
return 8;
}
}
else if (SC >= 2545) {
return 8;
}
}
else if (posBalance >= 50) {
return 8;
}
}
else if (RsDiff >= 2) {
if (Rs < 1145) {
return 1;
}
else if (Rs >= 1145) {
if (Rs < 1572) {
return 8;
}
else if (Rs >= 1572) {
return 2;
}
}
}
}
}
}
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 7) {
if (AFD < 15) {
if (Rs < 1028) {
if (Rs < 806) {
if (MVDiff < 4) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 749) {
return 4;
}
else if (Rs >= 749) {
return 2;
}
}
}
else if (MVDiff >= 4) {
if (mcTcor < 96) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (Cs < 1173) {
if (MVDiff < 31) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Cs < 993) {
return 1;
}
else if (Cs >= 993) {
return 2;
}
}
}
else if (MVDiff >= 31) {
return 4;
}
}
else if (Cs >= 1173) {
if (Rs < 526) {
return 1;
}
else if (Rs >= 526) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 83) {
if (mcTcor < 99) {
return 2;
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (refDCval >= 83) {
if (Rs < 764) {
return 2;
}
else if (Rs >= 764) {
return 4;
}
}
}
}
}
else if (Rs >= 806) {
if (refDCval < 63) {
if (mcTcor < 90) {
if (MVDiff < 20) {
if (Rs < 973) {
return 8;
}
else if (Rs >= 973) {
return 4;
}
}
else if (MVDiff >= 20) {
if (SC < 1989) {
return 4;
}
else if (SC >= 1989) {
return 2;
}
}
}
else if (mcTcor >= 90) {
if (Rs < 880) {
return 8;
}
else if (Rs >= 880) {
return 4;
}
}
}
else if (refDCval >= 63) {
if (posBalance < 59) {
if (tcor < 89) {
if (ssDCval < 150) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
if (Cs < 1053) {
return 8;
}
else if (Cs >= 1053) {
return 1;
}
}
}
else if (ssDCval >= 150) {
return 2;
}
}
else if (tcor >= 89) {
if (Cs < 1157) {
return 8;
}
else if (Cs >= 1157) {
if (posBalance < 55) {
if (Rs < 818) {
return 4;
}
else if (Rs >= 818) {
return 2;
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
else if (posBalance >= 59) {
if (diffRsCsdiff < -4) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
if (Cs < 989) {
return 8;
}
else if (Cs >= 989) {
return 4;
}
}
}
else if (diffRsCsdiff >= -4) {
if (Rs < 877) {
return 4;
}
else if (Rs >= 877) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (Rs < 986) {
return 4;
}
else if (Rs >= 986) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 1028) {
if (negBalance < 46) {
if (ssDCval < 60) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (Rs < 1148) {
if (Cs < 1030) {
return 8;
}
else if (Cs >= 1030) {
return 4;
}
}
else if (Rs >= 1148) {
if (negBalance < 33) {
return 4;
}
else if (negBalance >= 33) {
if (CsDiff < 6) {
if (posBalance < 39) {
return 2;
}
else if (posBalance >= 39) {
return 8;
}
}
else if (CsDiff >= 6) {
return 1;
}
}
}
}
}
else if (ssDCval >= 60) {
if (tcor < 83) {
if (ssDCval < 77) {
return 1;
}
else if (ssDCval >= 77) {
if (RsCsDiff < 46) {
if (MVDiff < 31) {
return 8;
}
else if (MVDiff >= 31) {
return 2;
}
}
else if (RsCsDiff >= 46) {
if (Rs < 1207) {
return 1;
}
else if (Rs >= 1207) {
return 8;
}
}
}
}
else if (tcor >= 83) {
if (Cs < 1098) {
if (negBalance < 40) {
if (diffAFD < 2) {
if (posBalance < 42) {
if (Rs < 1601) {
return 8;
}
else if (Rs >= 1601) {
return 4;
}
}
else if (posBalance >= 42) {
if (posBalance < 63) {
return 2;
}
else if (posBalance >= 63) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (negBalance >= 40) {
if (mcTcor < 96) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 2432) {
return 4;
}
else if (Rs >= 2432) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (ssDCval < 86) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
if (Rs < 1703) {
return 2;
}
else if (Rs >= 1703) {
return 4;
}
}
}
else if (ssDCval >= 86) {
return 4;
}
}
}
}
else if (Cs >= 1098) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (RsDiff < 5) {
if (TSC < 6) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 2;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (RsDiff >= 5) {
if (negBalance < 41) {
return 2;
}
else if (negBalance >= 41) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
if (ssDCval < 97) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (ssDCval >= 97) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (negBalance >= 46) {
if (posBalance < 50) {
if (ssDCval < 110) {
if (Rs < 1816) {
if (diffAFD < -9) {
return 4;
}
else if (diffAFD >= -9) {
return 8;
}
}
else if (Rs >= 1816) {
if (Cs < 1188) {
return 4;
}
else if (Cs >= 1188) {
return 8;
}
}
}
else if (ssDCval >= 110) {
if (Rs < 1102) {
return 8;
}
else if (Rs >= 1102) {
return 2;
}
}
}
else if (posBalance >= 50) {
if (refDCval < 58) {
if (mcTcor < 90) {
if (AFD < 13) {
if (Cs < 981) {
return 1;
}
else if (Cs >= 981) {
return 8;
}
}
else if (AFD >= 13) {
if (Rs < 1195) {
return 4;
}
else if (Rs >= 1195) {
return 2;
}
}
}
else if (mcTcor >= 90) {
return 4;
}
}
else if (refDCval >= 58) {
if (diffMVdiffVal < 6) {
if (posBalance < 55) {
if (AFD < 12) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
if (SC < 2269) {
return 1;
}
else if (SC >= 2269) {
return 8;
}
}
}
}
else if (AFD >= 12) {
if (SC < 2187) {
return 4;
}
else if (SC >= 2187) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
}
}
else if (posBalance >= 55) {
if (refDCval < 91) {
if (Rs < 1064) {
return 2;
}
else if (Rs >= 1064) {
if (diffMVdiffVal < -7) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
}
else if (refDCval >= 91) {
if (ssDCval < 110) {
if (CsDiff < 4) {
if (Cs < 1064) {
return 8;
}
else if (Cs >= 1064) {
if (Rs < 1757) {
return 2;
}
else if (Rs >= 1757) {
return 1;
}
}
}
else if (CsDiff >= 4) {
if (TSC < 9) {
if (diffAFD < 1) {
if (Rs < 1084) {
return 8;
}
else if (Rs >= 1084) {
return 4;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (TSC >= 9) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
else if (ssDCval >= 110) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= 6) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 15) {
if (ssDCval < 89) {
if (SC < 1953) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (SC >= 1953) {
if (ssDCval < 74) {
if (MVDiff < 29) {
if (diffMVdiffVal < -7) {
return 2;
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
else if (MVDiff >= 29) {
if (TSC < 11) {
if (Cs < 1252) {
return 8;
}
else if (Cs >= 1252) {
return 4;
}
}
else if (TSC >= 11) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (diffMVdiffVal < 3) {
if (mcTcor < 93) {
if (SC < 2123) {
return 4;
}
else if (SC >= 2123) {
if (ssDCval < 79) {
if (Rs < 1437) {
return 2;
}
else if (Rs >= 1437) {
return 8;
}
}
else if (ssDCval >= 79) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 932) {
return 8;
}
else if (Rs >= 932) {
return 4;
}
}
}
else if (diffMVdiffVal >= 3) {
if (mcTcor < 87) {
return 2;
}
else if (mcTcor >= 87) {
if (Rs < 1093) {
return 2;
}
else if (Rs >= 1093) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 89) {
if (RsCsDiff < 46) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 1076) {
return 8;
}
else if (Cs >= 1076) {
return 4;
}
}
}
else if (RsCsDiff >= 46) {
if (AFD < 17) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
return 2;
}
}
else if (AFD >= 17) {
if (Rs < 1222) {
return 2;
}
else if (Rs >= 1222) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (posBalance < 40) {
if (negBalance < 32) {
if (Rs < 1885) {
if (mu_mv_mag_sq < 105) {
if (negBalance < 22) {
if (Cs < 1165) {
return 1;
}
else if (Cs >= 1165) {
if (Rs < 902) {
return 4;
}
else if (Rs >= 902) {
return 8;
}
}
}
else if (negBalance >= 22) {
if (negBalance < 26) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (negBalance >= 26) {
if (mu_mv_mag_sq < 86) {
if (Rs < 1845) {
return 4;
}
else if (Rs >= 1845) {
return 8;
}
}
else if (mu_mv_mag_sq >= 86) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
return 8;
}
}
else if (Rs >= 1885) {
if (ssDCval < 139) {
if (diffRsCsdiff < -18) {
if (Rs < 2127) {
return 8;
}
else if (Rs >= 2127) {
return 1;
}
}
else if (diffRsCsdiff >= -18) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 2160) {
return 8;
}
else if (Rs >= 2160) {
return 2;
}
}
}
}
}
else if (ssDCval >= 139) {
return 8;
}
}
}
else if (negBalance >= 32) {
if (ssDCval < 146) {
if (ssDCval < 58) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
else if (ssDCval >= 58) {
if (ssDCval < 79) {
if (Rs < 948) {
return 2;
}
else if (Rs >= 948) {
if (Rs < 1590) {
return 8;
}
else if (Rs >= 1590) {
if (Rs < 1776) {
return 1;
}
else if (Rs >= 1776) {
return 8;
}
}
}
}
else if (ssDCval >= 79) {
return 8;
}
}
}
else if (ssDCval >= 146) {
return 2;
}
}
}
else if (posBalance >= 40) {
if (Rs < 1320) {
if (ssDCval < 89) {
if (Rs < 857) {
if (ssDCval < 81) {
if (Rs < 612) {
if (CsDiff < 6) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (CsDiff >= 6) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (Cs < 1008) {
return 8;
}
else if (Cs >= 1008) {
return 4;
}
}
}
}
else if (Rs >= 612) {
if (ssDCval < 57) {
if (refDCval < 45) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
else if (refDCval >= 45) {
if (Rs < 805) {
if (Cs < 1031) {
return 2;
}
else if (Cs >= 1031) {
return 4;
}
}
else if (Rs >= 805) {
if (Cs < 1084) {
return 8;
}
else if (Cs >= 1084) {
return 1;
}
}
}
}
else if (ssDCval >= 57) {
if (mu_mv_mag_sq < 77) {
if (mcTcor < 97) {
if (Rs < 839) {
if (MVDiff < 18) {
if (Rs < 668) {
return 8;
}
else if (Rs >= 668) {
return 1;
}
}
else if (MVDiff >= 18) {
return 8;
}
}
else if (Rs >= 839) {
return 2;
}
}
else if (mcTcor >= 97) {
return 2;
}
}
else if (mu_mv_mag_sq >= 77) {
if (diffRsCsdiff < -6) {
return 2;
}
else if (diffRsCsdiff >= -6) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (SC < 1906) {
if (Cs < 1041) {
return 8;
}
else if (Cs >= 1041) {
return 1;
}
}
else if (SC >= 1906) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
return 4;
}
}
}
else if (Rs >= 857) {
if (refDCval < 47) {
return 8;
}
else if (refDCval >= 47) {
if (Cs < 1094) {
if (Rs < 1100) {
if (Rs < 996) {
if (mu_mv_mag_sq < 121) {
if (MVDiff < 2) {
if (negBalance < 26) {
if (Rs < 932) {
return 4;
}
else if (Rs >= 932) {
return 8;
}
}
else if (negBalance >= 26) {
if (ssDCval < 88) {
return 8;
}
else if (ssDCval >= 88) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (TSC < 5) {
if (Rs < 879) {
return 2;
}
else if (Rs >= 879) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (TSC >= 5) {
if (Rs < 868) {
return 2;
}
else if (Rs >= 868) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
return 4;
}
}
else if (Rs >= 996) {
if (Cs < 1012) {
if (Rs < 1011) {
if (Rs < 1006) {
return 4;
}
else if (Rs >= 1006) {
return 1;
}
}
else if (Rs >= 1011) {
return 8;
}
}
else if (Cs >= 1012) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
if (diffTSC < -1) {
if (Rs < 1034) {
return 8;
}
else if (Rs >= 1034) {
return 4;
}
}
else if (diffTSC >= -1) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
if (Rs < 1019) {
return 8;
}
else if (Rs >= 1019) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1100) {
if (RsCsDiff < 4) {
if (Cs < 1039) {
return 8;
}
else if (Cs >= 1039) {
return 1;
}
}
else if (RsCsDiff >= 4) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 1292) {
return 4;
}
else if (Rs >= 1292) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (mcTcor < 95) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (mcTcor >= 95) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1094) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 113) {
return 8;
}
else if (mu_mv_mag_sq >= 113) {
if (ssDCval < 76) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
return 1;
}
}
else if (ssDCval >= 76) {
return 4;
}
}
}
else if (RsDiff >= 2) {
if (tcor < 85) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (negBalance < 50) {
if (posBalance < 54) {
if (Rs < 1276) {
return 8;
}
else if (Rs >= 1276) {
if (Rs < 1318) {
return 2;
}
else if (Rs >= 1318) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 1;
}
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
else if (tcor >= 85) {
if (Rs < 1152) {
if (posBalance < 42) {
if (Rs < 965) {
return 2;
}
else if (Rs >= 965) {
return 4;
}
}
else if (posBalance >= 42) {
if (posBalance < 68) {
if (AFD < 5) {
if (RsDiff < 3) {
if (refDCval < 63) {
if (SC < 2202) {
return 4;
}
else if (SC >= 2202) {
return 8;
}
}
else if (refDCval >= 63) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 989) {
return 4;
}
else if (Rs >= 989) {
return 8;
}
}
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (AFD >= 5) {
if (TSC < 7) {
if (posBalance < 47) {
if (Cs < 1218) {
return 8;
}
else if (Cs >= 1218) {
return 2;
}
}
else if (posBalance >= 47) {
if (SC < 2007) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (SC >= 2007) {
if (negBalance < 33) {
if (Cs < 1170) {
return 8;
}
else if (Cs >= 1170) {
if (Cs < 1254) {
return 2;
}
else if (Cs >= 1254) {
return 4;
}
}
}
else if (negBalance >= 33) {
if (ssDCval < 49) {
if (Rs < 978) {
return 4;
}
else if (Rs >= 978) {
return 8;
}
}
else if (ssDCval >= 49) {
if (AFD < 11) {
if (Cs < 1097) {
return 8;
}
else if (Cs >= 1097) {
if (negBalance < 34) {
if (Rs < 998) {
return 4;
}
else if (Rs >= 998) {
return 8;
}
}
else if (negBalance >= 34) {
return 4;
}
}
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
}
else if (TSC >= 7) {
if (negBalance < 44) {
return 4;
}
else if (negBalance >= 44) {
return 8;
}
}
}
}
else if (posBalance >= 68) {
return 2;
}
}
}
else if (Rs >= 1152) {
if (ssDCval < 59) {
if (AFD < 8) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1263) {
return 4;
}
else if (Rs >= 1263) {
return 8;
}
}
}
}
else if (AFD >= 8) {
if (SC < 2499) {
return 2;
}
else if (SC >= 2499) {
return 4;
}
}
}
else if (ssDCval >= 59) {
if (Cs < 1195) {
if (ssDCval < 61) {
if (AFD < 7) {
if (Rs < 1266) {
return 1;
}
else if (Rs >= 1266) {
return 8;
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (ssDCval >= 61) {
if (mu_mv_mag_sq < 121) {
if (diffAFD < 2) {
if (Rs < 1161) {
if (Rs < 1158) {
return 4;
}
else if (Rs >= 1158) {
return 1;
}
}
else if (Rs >= 1161) {
if (negBalance < 61) {
return 8;
}
else if (negBalance >= 61) {
return 4;
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 121) {
return 4;
}
}
}
else if (Cs >= 1195) {
if (MVDiff < 1) {
if (negBalance < 39) {
if (Rs < 1308) {
return 4;
}
else if (Rs >= 1308) {
return 1;
}
}
else if (negBalance >= 39) {
return 1;
}
}
else if (MVDiff >= 1) {
if (ssDCval < 64) {
if (Rs < 1188) {
return 8;
}
else if (Rs >= 1188) {
return 1;
}
}
else if (ssDCval >= 64) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 89) {
if (ssDCval < 143) {
if (mu_mv_mag_sq < 102) {
if (tcor < 90) {
if (mu_mv_mag_sq < 33) {
if (ssDCval < 130) {
if (MVDiff < 51) {
return 8;
}
else if (MVDiff >= 51) {
return 2;
}
}
else if (ssDCval >= 130) {
if (Cs < 1131) {
if (tcor < 47) {
return 1;
}
else if (tcor >= 47) {
return 8;
}
}
else if (Cs >= 1131) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (ssDCval < 94) {
if (SC < 2151) {
return 1;
}
else if (SC >= 2151) {
return 8;
}
}
else if (ssDCval >= 94) {
return 8;
}
}
}
else if (tcor >= 90) {
if (TSC < 4) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
if (Rs < 667) {
if (Rs < 481) {
return 8;
}
else if (Rs >= 481) {
return 1;
}
}
else if (Rs >= 667) {
return 8;
}
}
}
else if (TSC >= 4) {
if (Rs < 718) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
return 8;
}
}
}
else if (Rs >= 718) {
if (SC < 2109) {
if (refDCval < 141) {
if (TSC < 6) {
if (ssDCval < 119) {
return 4;
}
else if (ssDCval >= 119) {
if (Rs < 1008) {
return 8;
}
else if (Rs >= 1008) {
return 4;
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (refDCval >= 141) {
return 8;
}
}
else if (SC >= 2109) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (Cs < 1054) {
if (Rs < 1222) {
return 4;
}
else if (Rs >= 1222) {
if (Rs < 1293) {
return 1;
}
else if (Rs >= 1293) {
return 8;
}
}
}
else if (Cs >= 1054) {
if (Rs < 927) {
return 4;
}
else if (Rs >= 927) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (ssDCval < 94) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
return 4;
}
}
else if (ssDCval >= 94) {
return 8;
}
}
}
else if (ssDCval >= 143) {
if (ssDCval < 155) {
if (Rs < 1239) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 8;
}
}
else if (Rs >= 1239) {
return 8;
}
}
else if (ssDCval >= 155) {
if (Cs < 991) {
return 4;
}
else if (Cs >= 991) {
return 2;
}
}
}
}
}
else if (Rs >= 1320) {
if (ssDCval < 73) {
if (posBalance < 45) {
if (Rs < 1857) {
if (mu_mv_mag_sq < 66) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
if (Rs < 1335) {
return 1;
}
else if (Rs >= 1335) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (mcTcor < 96) {
if (TSC < 6) {
if (Rs < 1361) {
return 4;
}
else if (Rs >= 1361) {
return 8;
}
}
else if (TSC >= 6) {
if (Cs < 1189) {
return 4;
}
else if (Cs >= 1189) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (RsDiff < 4) {
if (RsCsDiff < 11) {
return 8;
}
else if (RsCsDiff >= 11) {
return 1;
}
}
else if (RsDiff >= 4) {
return 4;
}
}
}
}
else if (Rs >= 1857) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (posBalance >= 45) {
if (ssDCval < 65) {
if (Rs < 1656) {
if (diffRsCsdiff < -51) {
if (Cs < 1023) {
return 1;
}
else if (Cs >= 1023) {
return 8;
}
}
else if (diffRsCsdiff >= -51) {
if (posBalance < 68) {
if (Cs < 1266) {
return 8;
}
else if (Cs >= 1266) {
if (Cs < 1270) {
return 1;
}
else if (Cs >= 1270) {
return 8;
}
}
}
else if (posBalance >= 68) {
if (Rs < 1352) {
return 4;
}
else if (Rs >= 1352) {
return 8;
}
}
}
}
else if (Rs >= 1656) {
if (ssDCval < 62) {
if (diffAFD < -6) {
return 2;
}
else if (diffAFD >= -6) {
if (posBalance < 46) {
if (Rs < 2127) {
return 4;
}
else if (Rs >= 2127) {
return 8;
}
}
else if (posBalance >= 46) {
return 8;
}
}
}
else if (ssDCval >= 62) {
if (Cs < 1097) {
if (SC < 3472) {
return 2;
}
else if (SC >= 3472) {
return 4;
}
}
else if (Cs >= 1097) {
return 8;
}
}
}
}
else if (ssDCval >= 65) {
if (Rs < 1559) {
if (ssDCval < 71) {
if (mu_mv_mag_sq < 117) {
if (RsCsDiff < 16) {
if (refDCval < 69) {
return 8;
}
else if (refDCval >= 69) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 1;
}
}
}
else if (RsCsDiff >= 16) {
if (negBalance < 46) {
if (CsDiff < 4) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (negBalance >= 46) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 8;
}
}
}
else if (ssDCval >= 71) {
if (TSC < 6) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
return 1;
}
}
else if (TSC >= 6) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
else if (Rs >= 1559) {
if (ssDCval < 71) {
if (CsDiff < 5) {
if (mu_mv_mag_sq < 73) {
if (mu_mv_mag_sq < 49) {
if (Rs < 1997) {
return 1;
}
else if (Rs >= 1997) {
return 4;
}
}
else if (mu_mv_mag_sq >= 49) {
return 8;
}
}
else if (mu_mv_mag_sq >= 73) {
if (AFD < 10) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
return 8;
}
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (ssDCval >= 71) {
if (MVDiff < 16) {
if (Rs < 2117) {
return 8;
}
else if (Rs >= 2117) {
return 4;
}
}
else if (MVDiff >= 16) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 73) {
if (ssDCval < 159) {
if (SC < 2488) {
if (MVDiff < 22) {
if (ssDCval < 78) {
return 4;
}
else if (ssDCval >= 78) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 8;
}
}
}
else if (MVDiff >= 22) {
if (Rs < 1383) {
return 8;
}
else if (Rs >= 1383) {
return 4;
}
}
}
else if (SC >= 2488) {
if (CsDiff < 3) {
if (negBalance < 27) {
if (mu_mv_mag_sq < 89) {
return 1;
}
else if (mu_mv_mag_sq >= 89) {
return 8;
}
}
else if (negBalance >= 27) {
if (diffRsCsdiff < -11) {
if (Rs < 1768) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
return 8;
}
}
else if (Rs >= 1768) {
return 8;
}
}
else if (diffRsCsdiff >= -11) {
if (mcTcor < 91) {
return 1;
}
else if (mcTcor >= 91) {
return 8;
}
}
}
}
else if (CsDiff >= 3) {
if (ssDCval < 81) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (mu_mv_mag_sq < 109) {
if (Rs < 1574) {
if (Rs < 1415) {
return 1;
}
else if (Rs >= 1415) {
return 4;
}
}
else if (Rs >= 1574) {
if (TSC < 7) {
if (Cs < 1298) {
if (Rs < 1997) {
return 1;
}
else if (Rs >= 1997) {
if (Rs < 2200) {
return 8;
}
else if (Rs >= 2200) {
return 1;
}
}
}
else if (Cs >= 1298) {
if (Rs < 1978) {
return 8;
}
else if (Rs >= 1978) {
return 4;
}
}
}
else if (TSC >= 7) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Cs < 1279) {
return 8;
}
else if (Cs >= 1279) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
}
else if (ssDCval >= 81) {
if (diffAFD < -2) {
if (Rs < 1463) {
return 1;
}
else if (Rs >= 1463) {
if (Rs < 1825) {
if (gchDC < 1) {
if (mcTcor < 94) {
return 2;
}
else if (mcTcor >= 94) {
if (Rs < 1742) {
return 8;
}
else if (Rs >= 1742) {
return 2;
}
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Rs >= 1825) {
if (mcTcor < 96) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 2013) {
return 1;
}
else if (Rs >= 2013) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (Cs < 1126) {
return 2;
}
else if (Cs >= 1126) {
return 1;
}
}
}
}
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 51) {
if (diffMVdiffVal < -15) {
if (Rs < 1772) {
return 4;
}
else if (Rs >= 1772) {
return 2;
}
}
else if (diffMVdiffVal >= -15) {
if (mcTcor < 90) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (mcTcor >= 90) {
if (MVDiff < 8) {
if (Rs < 2371) {
return 8;
}
else if (Rs >= 2371) {
return 4;
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 2352) {
if (Rs < 1988) {
if (mcTcor < 96) {
if (Rs < 1889) {
if (refDCval < 100) {
if (SC < 2775) {
return 8;
}
else if (SC >= 2775) {
if (ssDCval < 88) {
return 8;
}
else if (ssDCval >= 88) {
if (TSC < 7) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
return 8;
}
}
else if (TSC >= 7) {
if (Rs < 1513) {
return 2;
}
else if (Rs >= 1513) {
return 8;
}
}
}
}
}
else if (refDCval >= 100) {
return 8;
}
}
else if (Rs >= 1889) {
if (TSC < 7) {
if (RsCsDiff < 29) {
if (Rs < 1916) {
return 2;
}
else if (Rs >= 1916) {
return 1;
}
}
else if (RsCsDiff >= 29) {
return 8;
}
}
else if (TSC >= 7) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (negBalance < 46) {
if (Rs < 1716) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 2;
}
}
else if (Rs >= 1716) {
if (SC < 2993) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 1;
}
}
else if (SC >= 2993) {
if (Cs < 1233) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
else if (Cs >= 1233) {
if (Cs < 1248) {
if (ssDCval < 106) {
return 1;
}
else if (ssDCval >= 106) {
return 8;
}
}
else if (Cs >= 1248) {
return 8;
}
}
}
}
}
else if (negBalance >= 46) {
return 1;
}
}
}
else if (Rs >= 1988) {
if (refDCval < 90) {
if (Rs < 2095) {
return 8;
}
else if (Rs >= 2095) {
return 1;
}
}
else if (refDCval >= 90) {
return 8;
}
}
}
else if (Rs >= 2352) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (Rs < 2479) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (Rs >= 2479) {
return 1;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 159) {
return 2;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 73) {
if (refDCval < 42) {
if (Rs < 968) {
if (refDCval < 31) {
return 8;
}
else if (refDCval >= 31) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (Rs >= 968) {
if (Rs < 1896) {
return 8;
}
else if (Rs >= 1896) {
if (Rs < 2035) {
return 2;
}
else if (Rs >= 2035) {
if (Rs < 2101) {
return 1;
}
else if (Rs >= 2101) {
return 8;
}
}
}
}
}
else if (refDCval >= 42) {
if (refDCval < 114) {
if (Cs < 831) {
if (tcor < 79) {
if (tcor < 58) {
if (TSC < 9) {
if (ssDCval < 49) {
if (AFD < 22) {
if (Rs < 1395) {
return 1;
}
else if (Rs >= 1395) {
return 2;
}
}
else if (AFD >= 22) {
if (Rs < 880) {
return 8;
}
else if (Rs >= 880) {
return 4;
}
}
}
else if (ssDCval >= 49) {
if (CsDiff < 5) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (Cs < 812) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
else if (Cs >= 812) {
return 4;
}
}
}
else if (CsDiff >= 5) {
return 4;
}
}
}
else if (TSC >= 9) {
if (AFD < 18) {
if (Rs < 1859) {
return 4;
}
else if (Rs >= 1859) {
return 1;
}
}
else if (AFD >= 18) {
if (RsDiff < 10) {
if (mu_mv_mag_sq < 14) {
if (mu_mv_mag_sq < 4) {
if (Rs < 1136) {
return 2;
}
else if (Rs >= 1136) {
return 1;
}
}
else if (mu_mv_mag_sq >= 4) {
return 1;
}
}
else if (mu_mv_mag_sq >= 14) {
if (ssDCval < 83) {
if (TSC < 11) {
if (Rs < 488) {
return 8;
}
else if (Rs >= 488) {
return 2;
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (ssDCval >= 83) {
if (Rs < 1351) {
return 2;
}
else if (Rs >= 1351) {
return 1;
}
}
}
}
else if (RsDiff >= 10) {
if (gchDC < 7) {
if (Rs < 2190) {
if (refDCval < 78) {
if (MVDiff < 39) {
return 2;
}
else if (MVDiff >= 39) {
if (Cs < 703) {
return 4;
}
else if (Cs >= 703) {
return 2;
}
}
}
else if (refDCval >= 78) {
return 4;
}
}
else if (Rs >= 2190) {
if (gchDC < 2) {
if (Cs < 769) {
return 1;
}
else if (Cs >= 769) {
return 4;
}
}
else if (gchDC >= 2) {
if (Rs < 2306) {
return 8;
}
else if (Rs >= 2306) {
return 2;
}
}
}
}
else if (gchDC >= 7) {
if (Cs < 646) {
return 1;
}
else if (Cs >= 646) {
return 8;
}
}
}
}
}
}
else if (tcor >= 58) {
if (posBalance < 47) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (refDCval < 85) {
return 2;
}
else if (refDCval >= 85) {
if (Cs < 752) {
return 1;
}
else if (Cs >= 752) {
return 8;
}
}
}
}
else if (posBalance >= 47) {
if (RsCsDiff < 233) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (MVDiff < 29) {
if (Rs < 1177) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 8;
}
}
else if (Rs >= 1177) {
if (Rs < 1804) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
if (diffMVdiffVal < -2) {
if (Cs < 690) {
return 4;
}
else if (Cs >= 690) {
if (Rs < 1559) {
return 2;
}
else if (Rs >= 1559) {
return 4;
}
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 1400) {
return 8;
}
else if (Rs >= 1400) {
return 4;
}
}
}
}
else if (Rs >= 1804) {
if (ssDCval < 75) {
if (Cs < 699) {
return 2;
}
else if (Cs >= 699) {
return 4;
}
}
else if (ssDCval >= 75) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 1;
}
}
}
}
}
else if (MVDiff >= 29) {
if (mcTcor < 92) {
if (SC < 2501) {
if (ssDCval < 71) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (ssDCval >= 71) {
if (Cs < 743) {
if (tcor < 74) {
if (diffAFD < -2) {
if (Rs < 908) {
return 4;
}
else if (Rs >= 908) {
if (Rs < 1029) {
return 8;
}
else if (Rs >= 1029) {
return 1;
}
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (tcor >= 74) {
if (mcTcor < 86) {
return 4;
}
else if (mcTcor >= 86) {
if (Cs < 621) {
return 2;
}
else if (Cs >= 621) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 2;
}
}
}
}
}
else if (Cs >= 743) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 1;
}
}
}
}
else if (SC >= 2501) {
if (SC < 3122) {
if (tcor < 59) {
return 1;
}
else if (tcor >= 59) {
return 8;
}
}
else if (SC >= 3122) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 90) {
if (diffMVdiffVal < 15) {
if (TSC < 9) {
if (Rs < 606) {
return 4;
}
else if (Rs >= 606) {
return 1;
}
}
else if (TSC >= 9) {
return 4;
}
}
else if (diffMVdiffVal >= 15) {
if (refDCval < 82) {
return 2;
}
else if (refDCval >= 82) {
if (Cs < 795) {
return 1;
}
else if (Cs >= 795) {
return 2;
}
}
}
}
else if (ssDCval >= 90) {
if (RsCsDiff < 147) {
if (refDCval < 103) {
return 8;
}
else if (refDCval >= 103) {
return 2;
}
}
else if (RsCsDiff >= 147) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
if (Rs < 1330) {
return 2;
}
else if (Rs >= 1330) {
return 1;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 233) {
if (diffRsCsdiff < 89) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (Cs < 615) {
return 2;
}
else if (Cs >= 615) {
return 4;
}
}
}
else if (diffRsCsdiff >= 89) {
return 4;
}
}
}
}
}
else if (tcor >= 79) {
if (refDCval < 77) {
if (Rs < 2104) {
if (AFD < 10) {
if (Rs < 1269) {
return 8;
}
else if (Rs >= 1269) {
return 1;
}
}
else if (AFD >= 10) {
if (MVDiff < 31) {
if (SC < 2783) {
if (posBalance < 49) {
if (diffRsCsdiff < 12) {
return 2;
}
else if (diffRsCsdiff >= 12) {
if (Rs < 1633) {
return 2;
}
else if (Rs >= 1633) {
return 1;
}
}
}
else if (posBalance >= 49) {
return 2;
}
}
else if (SC >= 2783) {
if (Rs < 2090) {
return 4;
}
else if (Rs >= 2090) {
return 1;
}
}
}
else if (MVDiff >= 31) {
if (RsCsDiff < 107) {
return 8;
}
else if (RsCsDiff >= 107) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
else if (Rs >= 2104) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
if (Rs < 2330) {
return 2;
}
else if (Rs >= 2330) {
return 4;
}
}
}
}
else if (refDCval >= 77) {
if (negBalance < 57) {
if (RsCsDiff < 137) {
if (MVDiff < 25) {
if (diffRsCsdiff < -37) {
return 4;
}
else if (diffRsCsdiff >= -37) {
if (SC < 2843) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
return 1;
}
}
else if (SC >= 2843) {
if (mu_mv_mag_sq < 17) {
if (SC < 3108) {
return 4;
}
else if (SC >= 3108) {
return 2;
}
}
else if (mu_mv_mag_sq >= 17) {
if (Rs < 2187) {
return 8;
}
else if (Rs >= 2187) {
return 1;
}
}
}
}
}
else if (MVDiff >= 25) {
if (MVDiff < 69) {
if (negBalance < 47) {
if (Rs < 2010) {
return 2;
}
else if (Rs >= 2010) {
if (Rs < 2238) {
return 1;
}
else if (Rs >= 2238) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (RsCsDiff < 74) {
return 2;
}
else if (RsCsDiff >= 74) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
if (Rs < 810) {
return 1;
}
else if (Rs >= 810) {
return 8;
}
}
}
}
}
else if (MVDiff >= 69) {
return 8;
}
}
}
else if (RsCsDiff >= 137) {
if (negBalance < 48) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 1;
}
}
else if (negBalance >= 48) {
if (ssDCval < 97) {
if (refDCval < 87) {
return 2;
}
else if (refDCval >= 87) {
return 4;
}
}
else if (ssDCval >= 97) {
return 1;
}
}
}
}
else if (negBalance >= 57) {
if (Cs < 691) {
if (Rs < 1429) {
return 1;
}
else if (Rs >= 1429) {
return 8;
}
}
else if (Cs >= 691) {
if (RsCsDiff < 123) {
if (Rs < 710) {
return 2;
}
else if (Rs >= 710) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (RsCsDiff >= 123) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 831) {
if (RsDiff < 12) {
if (Rs < 1240) {
if (RsDiff < 7) {
if (tcor < 68) {
if (ssDCval < 92) {
if (tcor < 57) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (AFD < 25) {
if (Rs < 369) {
return 4;
}
else if (Rs >= 369) {
return 8;
}
}
else if (AFD >= 25) {
return 4;
}
}
}
else if (tcor >= 57) {
if (mcTcor < 81) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 2;
}
}
else if (mcTcor >= 81) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (Rs < 407) {
return 4;
}
else if (Rs >= 407) {
return 8;
}
}
}
}
}
else if (ssDCval >= 92) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (tcor >= 68) {
if (diffMVdiffVal < -4) {
if (diffAFD < -2) {
if (Rs < 484) {
return 1;
}
else if (Rs >= 484) {
if (Rs < 702) {
return 2;
}
else if (Rs >= 702) {
return 8;
}
}
}
else if (diffAFD >= -2) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
if (Rs < 507) {
return 4;
}
else if (Rs >= 507) {
if (Rs < 980) {
return 8;
}
else if (Rs >= 980) {
return 2;
}
}
}
}
}
else if (diffMVdiffVal >= -4) {
if (MVDiff < 21) {
if (diffAFD < 1) {
if (RsDiff < 6) {
if (Cs < 1066) {
return 8;
}
else if (Cs >= 1066) {
return 2;
}
}
else if (RsDiff >= 6) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (MVDiff >= 21) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 369) {
return 1;
}
else if (Rs >= 369) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (ssDCval < 111) {
if (MVDiff < 62) {
if (mu_mv_mag_sq < 30) {
if (diffRsCsdiff < 14) {
if (tcor < 75) {
if (negBalance < 64) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (Cs < 1013) {
return 4;
}
else if (Cs >= 1013) {
if (diffRsCsdiff < -8) {
return 2;
}
else if (diffRsCsdiff >= -8) {
if (Rs < 986) {
return 2;
}
else if (Rs >= 986) {
return 4;
}
}
}
}
}
else if (negBalance >= 64) {
return 1;
}
}
else if (tcor >= 75) {
if (MVDiff < 16) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
return 2;
}
}
else if (MVDiff >= 16) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
if (Cs < 1030) {
return 4;
}
else if (Cs >= 1030) {
return 1;
}
}
}
}
}
else if (diffRsCsdiff >= 14) {
if (MVDiff < 33) {
if (negBalance < 51) {
if (Cs < 946) {
return 4;
}
else if (Cs >= 946) {
if (mu_mv_mag_sq < 18) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
return 8;
}
}
else if (mu_mv_mag_sq >= 18) {
return 8;
}
}
}
else if (negBalance >= 51) {
if (SC < 2286) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 1127) {
return 4;
}
else if (Rs >= 1127) {
return 8;
}
}
}
else if (SC >= 2286) {
return 2;
}
}
}
else if (MVDiff >= 33) {
if (MVDiff < 47) {
return 4;
}
else if (MVDiff >= 47) {
if (Rs < 1078) {
if (Rs < 954) {
return 8;
}
else if (Rs >= 954) {
return 2;
}
}
else if (Rs >= 1078) {
if (Rs < 1171) {
return 4;
}
else if (Rs >= 1171) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (TSC < 6) {
if (Rs < 1069) {
return 4;
}
else if (Rs >= 1069) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (MVDiff >= 62) {
if (negBalance < 68) {
if (diffRsCsdiff < 41) {
if (tcor < 50) {
return 8;
}
else if (tcor >= 50) {
if (mu_mv_mag_sq < 5) {
return 8;
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 1082) {
return 1;
}
else if (Rs >= 1082) {
if (Rs < 1141) {
return 8;
}
else if (Rs >= 1141) {
return 2;
}
}
}
}
}
else if (diffRsCsdiff >= 41) {
if (Rs < 975) {
return 2;
}
else if (Rs >= 975) {
return 4;
}
}
}
else if (negBalance >= 68) {
if (Rs < 1176) {
return 1;
}
else if (Rs >= 1176) {
return 8;
}
}
}
}
else if (ssDCval >= 111) {
if (Cs < 1229) {
return 2;
}
else if (Cs >= 1229) {
return 8;
}
}
}
}
else if (Rs >= 1240) {
if (negBalance < 46) {
if (mcTcor < 87) {
if (posBalance < 64) {
if (Rs < 1526) {
return 1;
}
else if (Rs >= 1526) {
if (tcor < 80) {
if (ssDCval < 91) {
if (Cs < 907) {
return 4;
}
else if (Cs >= 907) {
if (Rs < 1660) {
return 2;
}
else if (Rs >= 1660) {
if (posBalance < 43) {
if (Cs < 1037) {
return 8;
}
else if (Cs >= 1037) {
return 2;
}
}
else if (posBalance >= 43) {
return 8;
}
}
}
}
else if (ssDCval >= 91) {
return 4;
}
}
else if (tcor >= 80) {
if (Rs < 2025) {
return 8;
}
else if (Rs >= 2025) {
return 1;
}
}
}
}
else if (posBalance >= 64) {
if (Cs < 1226) {
return 8;
}
else if (Cs >= 1226) {
return 1;
}
}
}
else if (mcTcor >= 87) {
if (refDCval < 56) {
if (diffAFD < -2) {
if (Cs < 906) {
return 2;
}
else if (Cs >= 906) {
if (Rs < 1748) {
return 1;
}
else if (Rs >= 1748) {
return 8;
}
}
}
else if (diffAFD >= -2) {
if (MVDiff < 29) {
return 4;
}
else if (MVDiff >= 29) {
return 2;
}
}
}
else if (refDCval >= 56) {
if (Rs < 2254) {
if (refDCval < 102) {
if (Cs < 1175) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (RsCsDiff < 102) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (RsCsDiff < 80) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (RsCsDiff >= 80) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 8;
}
}
}
}
else if (RsCsDiff >= 102) {
if (Cs < 841) {
return 8;
}
else if (Cs >= 841) {
if (diffRsCsdiff < -115) {
return 4;
}
else if (diffRsCsdiff >= -115) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
return 4;
}
}
}
}
}
}
}
}
else if (Cs >= 1175) {
if (MVDiff < 21) {
if (mu_mv_mag_sq < 5) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Rs < 2020) {
return 2;
}
else if (Rs >= 2020) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (MVDiff >= 21) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 8;
}
}
}
}
else if (refDCval >= 102) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (Rs < 1388) {
return 2;
}
else if (Rs >= 1388) {
return 8;
}
}
}
}
}
else if (Rs >= 2254) {
if (refDCval < 72) {
if (posBalance < 32) {
return 1;
}
else if (posBalance >= 32) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (diffRsCsdiff < -118) {
return 4;
}
else if (diffRsCsdiff >= -118) {
return 2;
}
}
}
}
else if (refDCval >= 72) {
if (diffMVdiffVal < 1) {
if (Rs < 2423) {
return 2;
}
else if (Rs >= 2423) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 975) {
return 1;
}
else if (Cs >= 975) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 46) {
if (AFD < 23) {
if (Rs < 1663) {
if (MVDiff < 25) {
if (Cs < 894) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
if (Cs < 861) {
if (Rs < 1475) {
if (Rs < 1411) {
return 1;
}
else if (Rs >= 1411) {
return 8;
}
}
else if (Rs >= 1475) {
return 4;
}
}
else if (Cs >= 861) {
return 8;
}
}
}
else if (Cs >= 894) {
if (TSC < 8) {
if (mu_mv_mag_sq < 3) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
if (SC < 2591) {
if (Rs < 1335) {
return 8;
}
else if (Rs >= 1335) {
return 1;
}
}
else if (SC >= 2591) {
return 2;
}
}
}
}
else if (TSC >= 8) {
if (diffAFD < -2) {
if (posBalance < 63) {
if (Rs < 1632) {
return 8;
}
else if (Rs >= 1632) {
return 4;
}
}
else if (posBalance >= 63) {
return 1;
}
}
else if (diffAFD >= -2) {
if (AFD < 21) {
if (mu_mv_mag_sq < 43) {
return 8;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 1521) {
return 8;
}
else if (Rs >= 1521) {
return 1;
}
}
}
else if (AFD >= 21) {
if (Cs < 1159) {
return 8;
}
else if (Cs >= 1159) {
if (Rs < 1631) {
return 1;
}
else if (Rs >= 1631) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 25) {
if (refDCval < 72) {
if (mu_mv_mag_sq < 13) {
return 4;
}
else if (mu_mv_mag_sq >= 13) {
if (Cs < 881) {
return 4;
}
else if (Cs >= 881) {
return 8;
}
}
}
else if (refDCval >= 72) {
if (Cs < 970) {
if (Cs < 848) {
return 8;
}
else if (Cs >= 848) {
return 1;
}
}
else if (Cs >= 970) {
if (Rs < 1655) {
if (negBalance < 51) {
if (Rs < 1459) {
return 4;
}
else if (Rs >= 1459) {
return 8;
}
}
else if (negBalance >= 51) {
return 8;
}
}
else if (Rs >= 1655) {
return 4;
}
}
}
}
}
else if (Rs >= 1663) {
if (Cs < 1203) {
if (gchDC < 4) {
if (negBalance < 51) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 70) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
else if (ssDCval >= 70) {
if (Rs < 2453) {
return 4;
}
else if (Rs >= 2453) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (Rs < 1940) {
return 8;
}
else if (Rs >= 1940) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 8;
}
}
}
}
else if (posBalance >= 54) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (mu_mv_mag_sq < 17) {
if (mcTcor < 92) {
return 2;
}
else if (mcTcor >= 92) {
if (Rs < 2143) {
return 8;
}
else if (Rs >= 2143) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 17) {
return 8;
}
}
}
}
else if (negBalance >= 51) {
if (negBalance < 58) {
if (TSC < 11) {
if (AFD < 21) {
if (Rs < 2053) {
if (MVDiff < 21) {
if (Rs < 1879) {
if (SC < 2731) {
if (Rs < 1688) {
return 1;
}
else if (Rs >= 1688) {
return 2;
}
}
else if (SC >= 2731) {
return 8;
}
}
else if (Rs >= 1879) {
if (tcor < 80) {
return 4;
}
else if (tcor >= 80) {
if (Rs < 1966) {
if (Rs < 1934) {
return 8;
}
else if (Rs >= 1934) {
return 2;
}
}
else if (Rs >= 1966) {
return 1;
}
}
}
}
else if (MVDiff >= 21) {
if (Rs < 1737) {
return 4;
}
else if (Rs >= 1737) {
if (diffRsCsdiff < 6) {
if (Rs < 1795) {
return 1;
}
else if (Rs >= 1795) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
return 1;
}
}
}
}
else if (Rs >= 2053) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
}
}
}
else if (AFD >= 21) {
return 8;
}
}
else if (TSC >= 11) {
if (ssDCval < 89) {
if (Rs < 2146) {
if (Rs < 1978) {
return 8;
}
else if (Rs >= 1978) {
return 1;
}
}
else if (Rs >= 2146) {
return 4;
}
}
else if (ssDCval >= 89) {
if (Cs < 1070) {
return 1;
}
else if (Cs >= 1070) {
return 2;
}
}
}
}
else if (negBalance >= 58) {
if (ssDCval < 87) {
if (AFD < 19) {
if (Rs < 2357) {
return 8;
}
else if (Rs >= 2357) {
return 2;
}
}
else if (AFD >= 19) {
return 1;
}
}
else if (ssDCval >= 87) {
if (Cs < 1096) {
if (Rs < 1837) {
if (Rs < 1725) {
return 4;
}
else if (Rs >= 1725) {
return 1;
}
}
else if (Rs >= 1837) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 2;
}
}
}
else if (Cs >= 1096) {
if (Rs < 2105) {
return 2;
}
else if (Rs >= 2105) {
return 4;
}
}
}
}
}
}
else if (gchDC >= 4) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
return 8;
}
}
}
else if (Cs >= 1203) {
if (diffMVdiffVal < 13) {
if (SC < 3616) {
if (RsCsDiff < 194) {
if (diffRsCsdiff < 41) {
if (diffTSC < 1) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
if (TSC < 7) {
if (Rs < 1890) {
return 8;
}
else if (Rs >= 1890) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (mu_mv_mag_sq < 42) {
return 8;
}
else if (mu_mv_mag_sq >= 42) {
if (Cs < 1258) {
return 1;
}
else if (Cs >= 1258) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 41) {
return 4;
}
}
else if (RsCsDiff >= 194) {
return 2;
}
}
else if (SC >= 3616) {
if (Rs < 2369) {
return 8;
}
else if (Rs >= 2369) {
return 4;
}
}
}
else if (diffMVdiffVal >= 13) {
if (Rs < 2148) {
return 4;
}
else if (Rs >= 2148) {
return 2;
}
}
}
}
}
else if (AFD >= 23) {
if (MVDiff < 34) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 86) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
else if (mcTcor >= 86) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 4;
}
}
}
}
else if (MVDiff >= 34) {
if (gchDC < 4) {
if (RsCsDiff < 137) {
if (Cs < 1068) {
return 1;
}
else if (Cs >= 1068) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 1825) {
return 8;
}
else if (Rs >= 1825) {
return 1;
}
}
}
}
else if (RsCsDiff >= 137) {
if (Cs < 1159) {
return 2;
}
else if (Cs >= 1159) {
if (tcor < 55) {
return 2;
}
else if (tcor >= 55) {
return 4;
}
}
}
}
else if (gchDC >= 4) {
if (SC < 2407) {
return 2;
}
else if (SC >= 2407) {
if (Rs < 2080) {
return 8;
}
else if (Rs >= 2080) {
return 4;
}
}
}
}
}
}
}
}
else if (RsDiff >= 12) {
if (refDCval < 109) {
if (refDCval < 92) {
if (ssDCval < 64) {
if (AFD < 14) {
if (Rs < 1803) {
return 1;
}
else if (Rs >= 1803) {
if (Rs < 2210) {
return 8;
}
else if (Rs >= 2210) {
return 4;
}
}
}
else if (AFD >= 14) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (gchDC < 3) {
if (Rs < 2267) {
if (negBalance < 60) {
if (TSC < 11) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (tcor < 65) {
if (Cs < 919) {
return 8;
}
else if (Cs >= 919) {
return 4;
}
}
else if (tcor >= 65) {
return 8;
}
}
}
else if (TSC >= 11) {
if (MVDiff < 38) {
return 4;
}
else if (MVDiff >= 38) {
return 2;
}
}
}
else if (negBalance >= 60) {
return 2;
}
}
else if (Rs >= 2267) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (TSC < 10) {
if (SC < 3393) {
return 8;
}
else if (SC >= 3393) {
return 2;
}
}
else if (TSC >= 10) {
if (MVDiff < 103) {
if (RsCsDiff < 201) {
return 4;
}
else if (RsCsDiff >= 201) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (negBalance < 66) {
return 2;
}
else if (negBalance >= 66) {
return 4;
}
}
}
}
else if (MVDiff >= 103) {
if (refDCval < 71) {
return 4;
}
else if (refDCval >= 71) {
if (Rs < 1857) {
return 8;
}
else if (Rs >= 1857) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 92) {
if (mu_mv_mag_sq < 7) {
if (TSC < 16) {
if (SC < 2771) {
if (Rs < 1378) {
return 2;
}
else if (Rs >= 1378) {
return 8;
}
}
else if (SC >= 2771) {
return 2;
}
}
else if (TSC >= 16) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffMVdiffVal < 2) {
if (ssDCval < 103) {
return 1;
}
else if (ssDCval >= 103) {
return 4;
}
}
else if (diffMVdiffVal >= 2) {
if (negBalance < 58) {
return 8;
}
else if (negBalance >= 58) {
if (Cs < 1043) {
return 2;
}
else if (Cs >= 1043) {
if (Rs < 1545) {
return 4;
}
else if (Rs >= 1545) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (diffRsCsdiff < 66) {
return 4;
}
else if (diffRsCsdiff >= 66) {
if (Rs < 1639) {
return 2;
}
else if (Rs >= 1639) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 114) {
if (RsDiff < 9) {
if (AFD < 18) {
if (diffRsCsdiff < -16) {
if (negBalance < 47) {
if (tcor < 88) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
else if (tcor >= 88) {
if (diffRsCsdiff < -50) {
return 1;
}
else if (diffRsCsdiff >= -50) {
if (Rs < 1660) {
return 8;
}
else if (Rs >= 1660) {
return 2;
}
}
}
}
else if (negBalance >= 47) {
if (MVDiff < 25) {
return 8;
}
else if (MVDiff >= 25) {
if (Rs < 1078) {
return 8;
}
else if (Rs >= 1078) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= -16) {
if (mcTcor < 93) {
if (Rs < 1147) {
return 1;
}
else if (Rs >= 1147) {
if (TSC < 11) {
if (Cs < 600) {
return 4;
}
else if (Cs >= 600) {
return 8;
}
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 1220) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
if (RsCsDiff < 92) {
return 2;
}
else if (RsCsDiff >= 92) {
return 8;
}
}
}
}
}
else if (Rs >= 1220) {
if (Rs < 1900) {
return 8;
}
else if (Rs >= 1900) {
if (Cs < 1018) {
if (Rs < 1998) {
return 2;
}
else if (Rs >= 1998) {
return 1;
}
}
else if (Cs >= 1018) {
return 8;
}
}
}
}
}
}
else if (AFD >= 18) {
if (posBalance < 54) {
if (diffMVdiffVal < -20) {
if (Rs < 779) {
return 2;
}
else if (Rs >= 779) {
return 1;
}
}
else if (diffMVdiffVal >= -20) {
if (Rs < 649) {
return 8;
}
else if (Rs >= 649) {
if (Cs < 589) {
return 8;
}
else if (Cs >= 589) {
if (Cs < 1050) {
return 2;
}
else if (Cs >= 1050) {
return 8;
}
}
}
}
}
else if (posBalance >= 54) {
if (RsDiff < 6) {
if (Cs < 923) {
return 1;
}
else if (Cs >= 923) {
return 2;
}
}
else if (RsDiff >= 6) {
if (tcor < 85) {
if (diffRsCsdiff < 6) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
return 8;
}
}
else if (tcor >= 85) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
if (tcor < 79) {
if (mu_mv_mag_sq < 174) {
if (diffRsCsdiff < 106) {
if (mcTcor < 90) {
if (AFD < 34) {
if (ssDCval < 139) {
if (AFD < 16) {
if (Rs < 1309) {
return 1;
}
else if (Rs >= 1309) {
return 4;
}
}
else if (AFD >= 16) {
if (Rs < 1935) {
if (AFD < 29) {
if (Cs < 591) {
return 8;
}
else if (Cs >= 591) {
return 4;
}
}
else if (AFD >= 29) {
if (Rs < 1570) {
return 8;
}
else if (Rs >= 1570) {
return 1;
}
}
}
else if (Rs >= 1935) {
return 8;
}
}
}
else if (ssDCval >= 139) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
if (Rs < 1573) {
return 4;
}
else if (Rs >= 1573) {
return 8;
}
}
}
}
else if (AFD >= 34) {
if (TSC < 21) {
return 2;
}
else if (TSC >= 21) {
return 4;
}
}
}
else if (mcTcor >= 90) {
if (MVDiff < 17) {
if (posBalance < 55) {
if (Rs < 1441) {
return 8;
}
else if (Rs >= 1441) {
return 2;
}
}
else if (posBalance >= 55) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (MVDiff >= 17) {
if (RsCsDiff < 201) {
if (Rs < 1715) {
return 8;
}
else if (Rs >= 1715) {
if (SC < 3239) {
if (MVDiff < 87) {
return 1;
}
else if (MVDiff >= 87) {
return 8;
}
}
else if (SC >= 3239) {
return 2;
}
}
}
else if (RsCsDiff >= 201) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Rs < 2485) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (Rs >= 2485) {
return 1;
}
}
}
}
}
}
else if (diffRsCsdiff >= 106) {
if (posBalance < 42) {
if (Cs < 740) {
return 1;
}
else if (Cs >= 740) {
return 8;
}
}
else if (posBalance >= 42) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 174) {
return 1;
}
}
else if (tcor >= 79) {
if (Rs < 1862) {
if (MVDiff < 12) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
if (Rs < 1766) {
return 8;
}
else if (Rs >= 1766) {
return 2;
}
}
}
}
else if (MVDiff >= 12) {
if (refDCval < 132) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (mu_mv_mag_sq < 24) {
return 1;
}
else if (mu_mv_mag_sq >= 24) {
return 8;
}
}
}
else if (refDCval >= 132) {
if (diffRsCsdiff < -44) {
return 1;
}
else if (diffRsCsdiff >= -44) {
if (tcor < 88) {
if (Cs < 1281) {
if (gchDC < 8) {
return 2;
}
else if (gchDC >= 8) {
return 1;
}
}
else if (Cs >= 1281) {
return 8;
}
}
else if (tcor >= 88) {
if (CsDiff < 5) {
if (Rs < 1299) {
return 2;
}
else if (Rs >= 1299) {
return 1;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1862) {
if (Cs < 920) {
if (MVDiff < 84) {
return 1;
}
else if (MVDiff >= 84) {
if (Rs < 2068) {
return 1;
}
else if (Rs >= 2068) {
return 8;
}
}
}
else if (Cs >= 920) {
if (gchDC < 3) {
if (negBalance < 47) {
if (Cs < 1137) {
return 2;
}
else if (Cs >= 1137) {
return 8;
}
}
else if (negBalance >= 47) {
if (Cs < 951) {
return 8;
}
else if (Cs >= 951) {
return 1;
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2501) {
if (MVDiff < 16) {
if (refDCval < 109) {
if (posBalance < 48) {
if (Rs < 3822) {
if (RsCsDiff < 19) {
if (posBalance < 3) {
if (Cs < 847) {
return 4;
}
else if (Cs >= 847) {
if (negBalance < 3) {
return 8;
}
else if (negBalance >= 3) {
return 4;
}
}
}
else if (posBalance >= 3) {
if (Rs < 3124) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (diffRsCsdiff < -26) {
if (Cs < 991) {
return 2;
}
else if (Cs >= 991) {
return 1;
}
}
else if (diffRsCsdiff >= -26) {
if (ssDCval < 74) {
if (Rs < 2751) {
if (Cs < 716) {
return 8;
}
else if (Cs >= 716) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (Rs < 2542) {
return 8;
}
else if (Rs >= 2542) {
return 8;
}
}
}
}
else if (Rs >= 2751) {
return 2;
}
}
else if (ssDCval >= 74) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
else if (Rs >= 3124) {
if (SC < 4399) {
if (ssDCval < 98) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 2;
}
}
else if (ssDCval >= 98) {
if (Rs < 3363) {
return 8;
}
else if (Rs >= 3363) {
return 1;
}
}
}
else if (SC >= 4399) {
if (Cs < 1227) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (Rs < 3512) {
if (Rs < 3381) {
return 2;
}
else if (Rs >= 3381) {
return 4;
}
}
else if (Rs >= 3512) {
return 8;
}
}
}
else if (Cs >= 1227) {
if (negBalance < 27) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 3636) {
return 8;
}
else if (Rs >= 3636) {
return 4;
}
}
}
else if (negBalance >= 27) {
if (Rs < 3258) {
return 4;
}
else if (Rs >= 3258) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 19) {
if (ssDCval < 93) {
if (negBalance < 51) {
if (refDCval < 65) {
if (mu_mv_mag_sq < 70) {
if (Cs < 998) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (AFD < 17) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 4;
}
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (Cs >= 998) {
if (Rs < 2586) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
else if (Rs >= 2586) {
if (mu_mv_mag_sq < 54) {
if (Cs < 1145) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Cs >= 1145) {
return 2;
}
}
else if (mu_mv_mag_sq >= 54) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (MVDiff < 3) {
if (Rs < 2578) {
return 8;
}
else if (Rs >= 2578) {
if (Rs < 2783) {
return 2;
}
else if (Rs >= 2783) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (Cs < 1288) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 8;
}
}
else if (Cs >= 1288) {
return 8;
}
}
}
}
else if (refDCval >= 65) {
if (SC < 3926) {
if (mu_mv_mag_sq < 80) {
if (TSC < 6) {
if (mu_mv_mag_sq < 7) {
if (Rs < 2966) {
return 1;
}
else if (Rs >= 2966) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 684) {
if (Rs < 2641) {
return 4;
}
else if (Rs >= 2641) {
return 8;
}
}
else if (Cs >= 684) {
if (Cs < 1135) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
else if (Cs >= 1135) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (Rs < 2830) {
if (negBalance < 38) {
if (Cs < 953) {
return 1;
}
else if (Cs >= 953) {
return 2;
}
}
else if (negBalance >= 38) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 4;
}
}
}
}
else if (Rs >= 2830) {
if (Cs < 920) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
return 4;
}
}
else if (Cs >= 920) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (Cs < 1189) {
return 2;
}
else if (Cs >= 1189) {
return 1;
}
}
}
else if (SC >= 3926) {
if (refDCval < 80) {
if (Cs < 1067) {
if (SC < 4768) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (Rs < 3177) {
if (mcTcor < 97) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 2;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (Rs >= 3177) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 1;
}
}
}
}
else if (SC >= 4768) {
return 1;
}
}
else if (Cs >= 1067) {
if (posBalance < 44) {
if (posBalance < 41) {
if (Cs < 1251) {
if (Rs < 3387) {
if (Rs < 3112) {
return 1;
}
else if (Rs >= 3112) {
return 4;
}
}
else if (Rs >= 3387) {
return 1;
}
}
else if (Cs >= 1251) {
if (Rs < 3705) {
return 2;
}
else if (Rs >= 3705) {
return 1;
}
}
}
else if (posBalance >= 41) {
if (diffAFD < -2) {
if (Rs < 2982) {
return 1;
}
else if (Rs >= 2982) {
return 2;
}
}
else if (diffAFD >= -2) {
if (RsCsDiff < 175) {
if (diffRsCsdiff < 13) {
return 4;
}
else if (diffRsCsdiff >= 13) {
if (Cs < 1187) {
return 4;
}
else if (Cs >= 1187) {
return 2;
}
}
}
else if (RsCsDiff >= 175) {
return 2;
}
}
}
}
else if (posBalance >= 44) {
if (ssDCval < 73) {
if (Rs < 3588) {
if (diffAFD < 2) {
if (Rs < 3458) {
return 2;
}
else if (Rs >= 3458) {
if (Rs < 3487) {
return 1;
}
else if (Rs >= 3487) {
return 2;
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (Rs >= 3588) {
if (Rs < 3714) {
return 4;
}
else if (Rs >= 3714) {
return 1;
}
}
}
else if (ssDCval >= 73) {
if (mu_mv_mag_sq < 91) {
if (Cs < 1120) {
return 1;
}
else if (Cs >= 1120) {
if (Rs < 2984) {
if (Rs < 2771) {
return 4;
}
else if (Rs >= 2771) {
return 1;
}
}
else if (Rs >= 2984) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 91) {
return 2;
}
}
}
}
}
else if (refDCval >= 80) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (ssDCval < 85) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (ssDCval >= 85) {
if (CsDiff < 5) {
if (Rs < 3817) {
if (refDCval < 93) {
if (Rs < 3500) {
return 4;
}
else if (Rs >= 3500) {
return 4;
}
}
else if (refDCval >= 93) {
return 4;
}
}
else if (Rs >= 3817) {
return 8;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 51) {
return 2;
}
}
else if (ssDCval >= 93) {
if (negBalance < 43) {
if (mu_mv_mag_sq < 80) {
if (Cs < 973) {
if (TSC < 4) {
if (Cs < 481) {
return 8;
}
else if (Cs >= 481) {
return 1;
}
}
else if (TSC >= 4) {
if (Rs < 3377) {
return 8;
}
else if (Rs >= 3377) {
if (ssDCval < 102) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (Rs < 3435) {
return 4;
}
else if (Rs >= 3435) {
return 4;
}
}
}
}
else if (ssDCval >= 102) {
return 8;
}
}
}
}
else if (Cs >= 973) {
return 8;
}
}
else if (mu_mv_mag_sq >= 80) {
if (Rs < 3326) {
return 8;
}
else if (Rs >= 3326) {
return 2;
}
}
}
else if (negBalance >= 43) {
if (Cs < 1183) {
return 2;
}
else if (Cs >= 1183) {
return 8;
}
}
}
}
}
else if (Rs >= 3822) {
if (Cs < 1130) {
if (refDCval < 91) {
return 4;
}
else if (refDCval >= 91) {
if (Cs < 949) {
return 2;
}
else if (Cs >= 949) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
if (Rs < 4062) {
if (Rs < 3967) {
return 4;
}
else if (Rs >= 3967) {
return 4;
}
}
else if (Rs >= 4062) {
return 8;
}
}
}
}
}
else if (Cs >= 1130) {
if (mu_mv_mag_sq < 31) {
return 2;
}
else if (mu_mv_mag_sq >= 31) {
if (SC < 5342) {
if (Cs < 1275) {
if (mu_mv_mag_sq < 57) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (Rs < 4089) {
if (Rs < 3981) {
return 4;
}
else if (Rs >= 3981) {
return 8;
}
}
else if (Rs >= 4089) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (ssDCval < 69) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
else if (ssDCval >= 69) {
return 4;
}
}
}
else if (Cs >= 1275) {
if (mu_mv_mag_sq < 127) {
if (Cs < 1291) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (Cs >= 1291) {
if (Cs < 1303) {
return 4;
}
else if (Cs >= 1303) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
}
else if (SC >= 5342) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
if (mu_mv_mag_sq < 63) {
if (Cs < 1186) {
return 8;
}
else if (Cs >= 1186) {
if (SC < 5525) {
if (posBalance < 45) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 8;
}
}
else if (posBalance >= 45) {
return 4;
}
}
else if (SC >= 5525) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (Rs < 4492) {
return 8;
}
else if (Rs >= 4492) {
return 2;
}
}
}
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Cs < 659) {
if (SC < 3473) {
if (AFD < 9) {
if (Rs < 2760) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Rs >= 2760) {
return 1;
}
}
else if (AFD >= 9) {
if (Rs < 2830) {
if (Rs < 2567) {
return 1;
}
else if (Rs >= 2567) {
return 2;
}
}
else if (Rs >= 2830) {
return 8;
}
}
}
else if (SC >= 3473) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (Cs >= 659) {
if (negBalance < 23) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 2750) {
return 8;
}
else if (Rs >= 2750) {
if (Rs < 2934) {
return 4;
}
else if (Rs >= 2934) {
return 4;
}
}
}
}
else if (negBalance >= 23) {
if (gchDC < 2) {
if (Rs < 2545) {
if (Rs < 2535) {
return 2;
}
else if (Rs >= 2535) {
return 4;
}
}
else if (Rs >= 2545) {
return 2;
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 64) {
if (SC < 3474) {
return 2;
}
else if (SC >= 3474) {
if (tcor < 92) {
if (refDCval < 62) {
return 4;
}
else if (refDCval >= 62) {
if (RsDiff < 11) {
if (Rs < 2605) {
return 4;
}
else if (Rs >= 2605) {
return 2;
}
}
else if (RsDiff >= 11) {
return 4;
}
}
}
else if (tcor >= 92) {
if (mu_mv_mag_sq < 94) {
if (Cs < 1269) {
if (diffRsCsdiff < 57) {
return 2;
}
else if (diffRsCsdiff >= 57) {
return 4;
}
}
else if (Cs >= 1269) {
return 4;
}
}
else if (mu_mv_mag_sq >= 94) {
if (Cs < 975) {
return 2;
}
else if (Cs >= 975) {
return 4;
}
}
}
}
}
else if (refDCval >= 64) {
if (diffRsCsdiff < -285) {
return 4;
}
else if (diffRsCsdiff >= -285) {
if (negBalance < 39) {
if (posBalance < 52) {
if (mu_mv_mag_sq < 83) {
if (Rs < 2761) {
if (MVDiff < 10) {
if (Cs < 977) {
return 4;
}
else if (Cs >= 977) {
return 1;
}
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (Rs >= 2761) {
if (negBalance < 38) {
if (Rs < 3955) {
return 4;
}
else if (Rs >= 3955) {
return 8;
}
}
else if (negBalance >= 38) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 83) {
return 2;
}
}
else if (posBalance >= 52) {
if (mcTcor < 98) {
if (tcor < 79) {
if (TSC < 9) {
if (Rs < 2968) {
return 4;
}
else if (Rs >= 2968) {
return 1;
}
}
else if (TSC >= 9) {
return 2;
}
}
else if (tcor >= 79) {
return 2;
}
}
else if (mcTcor >= 98) {
if (diffTSC < 1) {
if (negBalance < 34) {
if (Rs < 3323) {
return 1;
}
else if (Rs >= 3323) {
return 2;
}
}
else if (negBalance >= 34) {
return 2;
}
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
else if (negBalance >= 39) {
if (tcor < 77) {
if (mu_mv_mag_sq < 5) {
if (Rs < 3510) {
return 4;
}
else if (Rs >= 3510) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 3911) {
return 2;
}
else if (Rs >= 3911) {
return 1;
}
}
}
else if (tcor >= 77) {
if (RsDiff < 6) {
if (Rs < 2738) {
return 4;
}
else if (Rs >= 2738) {
return 2;
}
}
else if (RsDiff >= 6) {
if (mcTcor < 99) {
return 2;
}
else if (mcTcor >= 99) {
if (Rs < 2560) {
return 2;
}
else if (Rs >= 2560) {
return 1;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsDiff < 11) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
if (Cs < 1018) {
if (tcor < 96) {
if (RsDiff < 7) {
if (gchDC < 3) {
if (diffMVdiffVal < 1) {
if (Rs < 3187) {
return 8;
}
else if (Rs >= 3187) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (Rs < 3147) {
return 4;
}
else if (Rs >= 3147) {
return 4;
}
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (tcor >= 96) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 2;
}
}
}
else if (Cs >= 1018) {
if (diffRsCsdiff < -28) {
if (Rs < 3198) {
return 4;
}
else if (Rs >= 3198) {
return 2;
}
}
else if (diffRsCsdiff >= -28) {
if (Rs < 3955) {
if (refDCval < 108) {
return 8;
}
else if (refDCval >= 108) {
if (Rs < 3539) {
return 2;
}
else if (Rs >= 3539) {
return 8;
}
}
}
else if (Rs >= 3955) {
return 4;
}
}
}
}
}
else if (RsDiff >= 11) {
if (negBalance < 50) {
if (RsDiff < 13) {
if (ssDCval < 106) {
if (Rs < 2663) {
return 4;
}
else if (Rs >= 2663) {
return 2;
}
}
else if (ssDCval >= 106) {
if (Rs < 3712) {
return 4;
}
else if (Rs >= 3712) {
return 8;
}
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (negBalance >= 50) {
if (Cs < 876) {
if (diffTSC < 3) {
return 2;
}
else if (diffTSC >= 3) {
if (Rs < 2904) {
return 4;
}
else if (Rs >= 2904) {
return 2;
}
}
}
else if (Cs >= 876) {
if (ssDCval < 103) {
if (SC < 4557) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (tcor < 64) {
if (Rs < 2892) {
return 2;
}
else if (Rs >= 2892) {
return 1;
}
}
else if (tcor >= 64) {
return 4;
}
}
}
else if (SC >= 4557) {
if (Rs < 3900) {
return 8;
}
else if (Rs >= 3900) {
return 2;
}
}
}
else if (ssDCval >= 103) {
if (RsCsDiff < 268) {
return 4;
}
else if (RsCsDiff >= 268) {
if (Cs < 1245) {
return 2;
}
else if (Cs >= 1245) {
return 4;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 3397) {
if (mcTcor < 97) {
if (RsDiff < 9) {
if (gchDC < 3) {
if (posBalance < 63) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (Rs < 2590) {
if (Cs < 1090) {
return 8;
}
else if (Cs >= 1090) {
return 1;
}
}
else if (Rs >= 2590) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (Rs < 3000) {
return 1;
}
else if (Rs >= 3000) {
return 2;
}
}
}
else if (gchDC >= 3) {
if (Rs < 3377) {
return 2;
}
else if (Rs >= 3377) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 117) {
if (diffAFD < 0) {
return 2;
}
else if (diffAFD >= 0) {
if (Rs < 2912) {
return 1;
}
else if (Rs >= 2912) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (SC < 4036) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2916) {
return 8;
}
else if (Rs >= 2916) {
if (Cs < 842) {
return 2;
}
else if (Cs >= 842) {
return 1;
}
}
}
}
else if (SC >= 4036) {
if (Rs < 3243) {
return 1;
}
else if (Rs >= 3243) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 7) {
if (Cs < 528) {
if (TSC < 3) {
if (Rs < 2869) {
return 1;
}
else if (Rs >= 2869) {
return 8;
}
}
else if (TSC >= 3) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
return 2;
}
}
}
else if (Cs >= 528) {
return 2;
}
}
else if (MVDiff >= 7) {
if (TSC < 6) {
if (SC < 3245) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (SC >= 3245) {
if (diffTSC < -1) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (TSC < 4) {
if (Rs < 3168) {
return 1;
}
else if (Rs >= 3168) {
return 8;
}
}
else if (TSC >= 4) {
if (Rs < 2926) {
return 8;
}
else if (Rs >= 2926) {
return 2;
}
}
}
}
else if (diffTSC >= -1) {
if (ssDCval < 129) {
return 8;
}
else if (ssDCval >= 129) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (SC < 3298) {
return 1;
}
else if (SC >= 3298) {
if (diffMVdiffVal < -2) {
if (Cs < 825) {
return 2;
}
else if (Cs >= 825) {
if (Rs < 3031) {
return 1;
}
else if (Rs >= 3031) {
return 4;
}
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3397) {
if (MVDiff < 5) {
if (ssDCval < 142) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
if (diffMVdiffVal < -3) {
if (SC < 4246) {
return 8;
}
else if (SC >= 4246) {
if (diffRsCsdiff < 40) {
return 2;
}
else if (diffRsCsdiff >= 40) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
}
else if (ssDCval >= 142) {
if (Rs < 3963) {
if (Cs < 1179) {
return 2;
}
else if (Cs >= 1179) {
if (posBalance < 30) {
return 2;
}
else if (posBalance >= 30) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
return 2;
}
}
}
}
else if (Rs >= 3963) {
return 2;
}
}
}
else if (MVDiff >= 5) {
if (mcTcor < 93) {
if (ssDCval < 135) {
if (RsDiff < 12) {
return 1;
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (ssDCval >= 135) {
if (Cs < 1027) {
return 4;
}
else if (Cs >= 1027) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 3931) {
if (TSC < 7) {
if (CsDiff < 3) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Cs < 1278) {
return 2;
}
else if (Cs >= 1278) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (ssDCval < 138) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
return 2;
}
}
else if (ssDCval >= 138) {
if (Rs < 3797) {
return 1;
}
else if (Rs >= 3797) {
return 2;
}
}
}
}
}
else if (TSC >= 7) {
if (Cs < 915) {
return 2;
}
else if (Cs >= 915) {
if (ssDCval < 138) {
if (Cs < 1024) {
return 1;
}
else if (Cs >= 1024) {
return 2;
}
}
else if (ssDCval >= 138) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
}
}
else if (Rs >= 3931) {
if (ssDCval < 132) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
if (Rs < 4209) {
return 4;
}
else if (Rs >= 4209) {
return 2;
}
}
}
}
else if (ssDCval >= 132) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (RsDiff < 20) {
if (mcTcor < 97) {
if (SC < 5163) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
else if (SC >= 5163) {
if (ssDCval < 149) {
if (Cs < 1263) {
return 1;
}
else if (Cs >= 1263) {
return 2;
}
}
else if (ssDCval >= 149) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (CsDiff < 5) {
if (diffMVdiffVal < -11) {
return 1;
}
else if (diffMVdiffVal >= -11) {
return 2;
}
}
else if (CsDiff >= 5) {
return 1;
}
}
}
else if (RsDiff >= 20) {
return 2;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 16) {
if (mcTcor < 70) {
if (MVDiff < 51) {
if (Rs < 4055) {
return 1;
}
else if (Rs >= 4055) {
return 2;
}
}
else if (MVDiff >= 51) {
if (gchDC < 5) {
if (Cs < 995) {
return 2;
}
else if (Cs >= 995) {
return 4;
}
}
else if (gchDC >= 5) {
return 8;
}
}
}
else if (mcTcor >= 70) {
if (ssDCval < 135) {
if (ssDCval < 82) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 80) {
if (SC < 3318) {
if (Cs < 251) {
return 4;
}
else if (Cs >= 251) {
if (Rs < 2538) {
return 4;
}
else if (Rs >= 2538) {
return 8;
}
}
}
else if (SC >= 3318) {
if (Cs < 688) {
return 4;
}
else if (Cs >= 688) {
if (Rs < 3914) {
return 2;
}
else if (Rs >= 3914) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (Rs < 3052) {
return 1;
}
else if (Rs >= 3052) {
if (Rs < 4310) {
return 8;
}
else if (Rs >= 4310) {
return 2;
}
}
}
}
else if (CsDiff >= 3) {
if (TSC < 8) {
if (posBalance < 48) {
if (Rs < 2860) {
if (tcor < 88) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (Rs < 2682) {
return 8;
}
else if (Rs >= 2682) {
if (Rs < 2802) {
return 4;
}
else if (Rs >= 2802) {
return 2;
}
}
}
}
else if (tcor >= 88) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 1;
}
}
}
else if (Rs >= 2860) {
if (mcTcor < 92) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
if (Rs < 3466) {
return 4;
}
else if (Rs >= 3466) {
return 1;
}
}
}
else if (mcTcor >= 92) {
return 4;
}
}
}
else if (posBalance >= 48) {
if (posBalance < 56) {
if (Cs < 751) {
return 4;
}
else if (Cs >= 751) {
if (mcTcor < 94) {
if (RsDiff < 13) {
if (Cs < 1092) {
return 4;
}
else if (Cs >= 1092) {
if (Rs < 2863) {
return 2;
}
else if (Rs >= 2863) {
return 4;
}
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
}
else if (posBalance >= 56) {
return 2;
}
}
}
else if (TSC >= 8) {
if (Rs < 2808) {
if (AFD < 13) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
if (Rs < 2766) {
return 1;
}
else if (Rs >= 2766) {
return 2;
}
}
}
else if (AFD >= 13) {
if (mcTcor < 83) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (mcTcor >= 83) {
if (Cs < 1266) {
if (diffRsCsdiff < 107) {
return 2;
}
else if (diffRsCsdiff >= 107) {
if (Cs < 840) {
return 1;
}
else if (Cs >= 840) {
return 2;
}
}
}
else if (Cs >= 1266) {
if (Rs < 2710) {
return 8;
}
else if (Rs >= 2710) {
return 2;
}
}
}
}
}
else if (Rs >= 2808) {
if (TSC < 12) {
if (tcor < 83) {
if (SC < 3669) {
return 2;
}
else if (SC >= 3669) {
if (diffRsCsdiff < -32) {
if (Cs < 696) {
return 4;
}
else if (Cs >= 696) {
return 2;
}
}
else if (diffRsCsdiff >= -32) {
if (CsDiff < 7) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (tcor < 78) {
if (negBalance < 55) {
return 4;
}
else if (negBalance >= 55) {
if (Rs < 3560) {
return 2;
}
else if (Rs >= 3560) {
return 4;
}
}
}
else if (tcor >= 78) {
return 2;
}
}
}
else if (CsDiff >= 7) {
if (posBalance < 51) {
if (Cs < 1181) {
return 4;
}
else if (Cs >= 1181) {
return 2;
}
}
else if (posBalance >= 51) {
return 2;
}
}
}
}
}
else if (tcor >= 83) {
if (Rs < 4316) {
return 2;
}
else if (Rs >= 4316) {
return 1;
}
}
}
else if (TSC >= 12) {
return 1;
}
}
}
}
}
else if (ssDCval >= 82) {
if (RsCsDiff < 194) {
if (Rs < 4149) {
if (diffAFD < -4) {
if (SC < 3829) {
if (MVDiff < 26) {
if (CsDiff < 4) {
if (Rs < 2548) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 2;
}
}
else if (Rs >= 2548) {
return 2;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (MVDiff >= 26) {
if (ssDCval < 111) {
if (negBalance < 48) {
if (Rs < 2897) {
return 8;
}
else if (Rs >= 2897) {
return 2;
}
}
else if (negBalance >= 48) {
return 1;
}
}
else if (ssDCval >= 111) {
return 1;
}
}
}
else if (SC >= 3829) {
if (posBalance < 64) {
if (gchDC < 1) {
if (Rs < 2888) {
return 8;
}
else if (Rs >= 2888) {
return 4;
}
}
else if (gchDC >= 1) {
if (SC < 3955) {
if (Rs < 3337) {
return 8;
}
else if (Rs >= 3337) {
return 2;
}
}
else if (SC >= 3955) {
return 2;
}
}
}
else if (posBalance >= 64) {
if (Rs < 3643) {
if (SC < 3968) {
return 8;
}
else if (SC >= 3968) {
return 1;
}
}
else if (Rs >= 3643) {
return 8;
}
}
}
}
else if (diffAFD >= -4) {
if (Rs < 2663) {
if (RsCsDiff < 128) {
if (gchDC < 8) {
if (posBalance < 58) {
if (Rs < 2526) {
if (Rs < 2509) {
return 8;
}
else if (Rs >= 2509) {
return 1;
}
}
else if (Rs >= 2526) {
return 8;
}
}
else if (posBalance >= 58) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
if (diffMVdiffVal < 12) {
return 1;
}
else if (diffMVdiffVal >= 12) {
if (ssDCval < 112) {
return 2;
}
else if (ssDCval >= 112) {
return 8;
}
}
}
}
}
else if (gchDC >= 8) {
if (Rs < 2635) {
return 1;
}
else if (Rs >= 2635) {
return 2;
}
}
}
else if (RsCsDiff >= 128) {
if (tcor < 84) {
if (diffAFD < 5) {
if (diffTSC < 1) {
if (Rs < 2640) {
return 8;
}
else if (Rs >= 2640) {
return 2;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (diffAFD >= 5) {
return 1;
}
}
else if (tcor >= 84) {
if (MVDiff < 55) {
return 1;
}
else if (MVDiff >= 55) {
return 8;
}
}
}
}
else if (Rs >= 2663) {
if (refDCval < 132) {
if (refDCval < 89) {
if (mcTcor < 87) {
if (Cs < 751) {
return 4;
}
else if (Cs >= 751) {
return 8;
}
}
else if (mcTcor >= 87) {
if (diffRsCsdiff < -8) {
return 2;
}
else if (diffRsCsdiff >= -8) {
if (SC < 3757) {
return 1;
}
else if (SC >= 3757) {
return 2;
}
}
}
}
else if (refDCval >= 89) {
if (mcTcor < 80) {
if (tcor < 57) {
return 4;
}
else if (tcor >= 57) {
if (Rs < 2805) {
return 2;
}
else if (Rs >= 2805) {
return 1;
}
}
}
else if (mcTcor >= 80) {
if (MVDiff < 46) {
if (diffAFD < -1) {
if (posBalance < 61) {
if (ssDCval < 94) {
if (MVDiff < 36) {
return 2;
}
else if (MVDiff >= 36) {
return 8;
}
}
else if (ssDCval >= 94) {
if (negBalance < 36) {
if (Rs < 3612) {
return 4;
}
else if (Rs >= 3612) {
return 2;
}
}
else if (negBalance >= 36) {
if (negBalance < 59) {
if (Rs < 2897) {
return 1;
}
else if (Rs >= 2897) {
return 8;
}
}
else if (negBalance >= 59) {
if (Cs < 540) {
return 4;
}
else if (Cs >= 540) {
return 8;
}
}
}
}
}
else if (posBalance >= 61) {
if (Rs < 3241) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (Rs >= 3241) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (mcTcor < 98) {
if (diffRsCsdiff < -32) {
if (SC < 3898) {
return 2;
}
else if (SC >= 3898) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 2;
}
}
}
else if (diffRsCsdiff >= -32) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 3299) {
if (Rs < 2970) {
return 8;
}
else if (Rs >= 2970) {
return 1;
}
}
else if (Rs >= 3299) {
return 2;
}
}
}
}
else if (MVDiff >= 46) {
if (Rs < 3649) {
if (Cs < 1273) {
if (mcTcor < 100) {
if (Cs < 212) {
if (Rs < 2722) {
return 8;
}
else if (Rs >= 2722) {
return 1;
}
}
else if (Cs >= 212) {
return 8;
}
}
else if (mcTcor >= 100) {
return 1;
}
}
else if (Cs >= 1273) {
return 4;
}
}
else if (Rs >= 3649) {
if (MVDiff < 65) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (MVDiff >= 65) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 132) {
if (Rs < 3537) {
return 4;
}
else if (Rs >= 3537) {
return 1;
}
}
}
}
}
else if (Rs >= 4149) {
if (negBalance < 44) {
if (Cs < 779) {
return 2;
}
else if (Cs >= 779) {
if (diffRsCsdiff < -63) {
if (Rs < 5990) {
return 1;
}
else if (Rs >= 5990) {
return 2;
}
}
else if (diffRsCsdiff >= -63) {
return 2;
}
}
}
else if (negBalance >= 44) {
if (Rs < 5570) {
return 8;
}
else if (Rs >= 5570) {
if (Rs < 6134) {
return 1;
}
else if (Rs >= 6134) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 194) {
if (refDCval < 110) {
if (mcTcor < 92) {
if (Cs < 916) {
if (CsDiff < 7) {
if (RsDiff < 19) {
if (Cs < 476) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (Cs < 419) {
return 2;
}
else if (Cs >= 419) {
return 8;
}
}
}
else if (Cs >= 476) {
if (Rs < 2577) {
if (Cs < 776) {
return 1;
}
else if (Cs >= 776) {
return 4;
}
}
else if (Rs >= 2577) {
return 2;
}
}
}
else if (RsDiff >= 19) {
return 2;
}
}
else if (CsDiff >= 7) {
if (AFD < 26) {
if (Rs < 2968) {
return 8;
}
else if (Rs >= 2968) {
return 2;
}
}
else if (AFD >= 26) {
return 1;
}
}
}
else if (Cs >= 916) {
if (TSC < 12) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 56) {
return 2;
}
else if (MVDiff >= 56) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 3076) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
if (Rs < 2839) {
return 2;
}
else if (Rs >= 2839) {
return 8;
}
}
}
else if (Rs >= 3076) {
if (diffAFD < -9) {
return 2;
}
else if (diffAFD >= -9) {
return 8;
}
}
}
}
else if (TSC >= 12) {
if (CsDiff < 7) {
if (refDCval < 100) {
if (ssDCval < 104) {
return 4;
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (refDCval >= 100) {
if (MVDiff < 104) {
return 2;
}
else if (MVDiff >= 104) {
return 4;
}
}
}
else if (CsDiff >= 7) {
if (Rs < 2551) {
return 4;
}
else if (Rs >= 2551) {
return 2;
}
}
}
}
}
else if (mcTcor >= 92) {
if (gchDC < 6) {
if (diffRsCsdiff < -196) {
if (TSC < 11) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 3498) {
return 2;
}
else if (Rs >= 3498) {
return 1;
}
}
}
else if (TSC >= 11) {
return 1;
}
}
else if (diffRsCsdiff >= -196) {
if (Cs < 1271) {
if (AFD < 28) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
if (refDCval < 101) {
if (Rs < 3814) {
return 2;
}
else if (Rs >= 3814) {
if (Rs < 3881) {
return 8;
}
else if (Rs >= 3881) {
return 2;
}
}
}
else if (refDCval >= 101) {
if (RsCsDiff < 351) {
if (diffRsCsdiff < -19) {
return 2;
}
else if (diffRsCsdiff >= -19) {
if (diffMVdiffVal < 8) {
if (SC < 3675) {
if (ssDCval < 104) {
return 2;
}
else if (ssDCval >= 104) {
if (Rs < 2962) {
return 8;
}
else if (Rs >= 2962) {
return 4;
}
}
}
else if (SC >= 3675) {
return 8;
}
}
else if (diffMVdiffVal >= 8) {
return 2;
}
}
}
else if (RsCsDiff >= 351) {
return 2;
}
}
}
}
else if (AFD >= 28) {
if (Cs < 542) {
return 2;
}
else if (Cs >= 542) {
return 1;
}
}
}
else if (Cs >= 1271) {
return 8;
}
}
}
else if (gchDC >= 6) {
if (tcor < 79) {
if (Cs < 255) {
return 2;
}
else if (Cs >= 255) {
if (RsDiff < 23) {
return 8;
}
else if (RsDiff >= 23) {
return 1;
}
}
}
else if (tcor >= 79) {
return 1;
}
}
}
}
else if (refDCval >= 110) {
if (Rs < 3478) {
if (mcTcor < 91) {
if (posBalance < 62) {
if (negBalance < 59) {
if (Rs < 2722) {
return 2;
}
else if (Rs >= 2722) {
return 8;
}
}
else if (negBalance >= 59) {
if (Cs < 989) {
return 2;
}
else if (Cs >= 989) {
return 1;
}
}
}
else if (posBalance >= 62) {
if (Cs < 1080) {
if (tcor < 77) {
if (AFD < 37) {
return 2;
}
else if (AFD >= 37) {
return 1;
}
}
else if (tcor >= 77) {
if (Rs < 3073) {
return 1;
}
else if (Rs >= 3073) {
return 4;
}
}
}
else if (Cs >= 1080) {
return 4;
}
}
}
else if (mcTcor >= 91) {
if (CsDiff < 3) {
if (Cs < 258) {
return 2;
}
else if (Cs >= 258) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
if (Rs < 3047) {
return 2;
}
else if (Rs >= 3047) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
if (MVDiff < 42) {
if (tcor < 76) {
if (Cs < 930) {
return 2;
}
else if (Cs >= 930) {
if (Rs < 3369) {
return 1;
}
else if (Rs >= 3369) {
return 8;
}
}
}
else if (tcor >= 76) {
if (posBalance < 59) {
if (AFD < 20) {
if (Cs < 607) {
return 2;
}
else if (Cs >= 607) {
return 8;
}
}
else if (AFD >= 20) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (Rs < 3261) {
return 1;
}
else if (Rs >= 3261) {
return 2;
}
}
}
}
else if (posBalance >= 59) {
return 1;
}
}
}
else if (MVDiff >= 42) {
if (CsDiff < 4) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 1;
}
}
else if (CsDiff >= 4) {
if (refDCval < 111) {
return 8;
}
else if (refDCval >= 111) {
if (Rs < 3263) {
return 1;
}
else if (Rs >= 3263) {
if (Rs < 3311) {
return 8;
}
else if (Rs >= 3311) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 3478) {
if (Cs < 847) {
if (Rs < 4913) {
if (Cs < 584) {
if (refDCval < 132) {
return 2;
}
else if (refDCval >= 132) {
return 1;
}
}
else if (Cs >= 584) {
if (MVDiff < 36) {
if (mcTcor < 95) {
if (RsDiff < 16) {
return 8;
}
else if (RsDiff >= 16) {
if (MVDiff < 34) {
return 2;
}
else if (MVDiff >= 34) {
if (Rs < 4335) {
return 8;
}
else if (Rs >= 4335) {
return 2;
}
}
}
}
else if (mcTcor >= 95) {
if (AFD < 23) {
return 1;
}
else if (AFD >= 23) {
return 2;
}
}
}
else if (MVDiff >= 36) {
if (Rs < 4494) {
return 8;
}
else if (Rs >= 4494) {
return 2;
}
}
}
}
else if (Rs >= 4913) {
return 2;
}
}
else if (Cs >= 847) {
if (mu_mv_mag_sq < 15) {
if (refDCval < 118) {
if (MVDiff < 33) {
return 2;
}
else if (MVDiff >= 33) {
if (SC < 5873) {
if (SC < 5455) {
return 1;
}
else if (SC >= 5455) {
return 8;
}
}
else if (SC >= 5873) {
return 2;
}
}
}
else if (refDCval >= 118) {
if (tcor < 85) {
if (diffRsCsdiff < 346) {
if (RsDiff < 22) {
if (diffMVdiffVal < -21) {
if (Rs < 4113) {
return 2;
}
else if (Rs >= 4113) {
return 8;
}
}
else if (diffMVdiffVal >= -21) {
return 8;
}
}
else if (RsDiff >= 22) {
if (diffMVdiffVal < 13) {
if (diffMVdiffVal < -10) {
return 8;
}
else if (diffMVdiffVal >= -10) {
if (Cs < 1063) {
return 2;
}
else if (Cs >= 1063) {
if (Rs < 5983) {
return 8;
}
else if (Rs >= 5983) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 13) {
if (Rs < 4001) {
return 1;
}
else if (Rs >= 4001) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 346) {
return 1;
}
}
else if (tcor >= 85) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (mu_mv_mag_sq < 35) {
if (tcor < 79) {
if (AFD < 25) {
return 1;
}
else if (AFD >= 25) {
return 2;
}
}
else if (tcor >= 79) {
return 1;
}
}
else if (mu_mv_mag_sq >= 35) {
if (diffMVdiffVal < -46) {
return 1;
}
else if (diffMVdiffVal >= -46) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 135) {
if (mcTcor < 93) {
if (AFD < 19) {
if (ssDCval < 150) {
if (Cs < 1276) {
return 4;
}
else if (Cs >= 1276) {
return 8;
}
}
else if (ssDCval >= 150) {
if (Cs < 1014) {
if (Rs < 4850) {
return 1;
}
else if (Rs >= 4850) {
return 2;
}
}
else if (Cs >= 1014) {
return 8;
}
}
}
else if (AFD >= 19) {
if (posBalance < 59) {
if (MVDiff < 56) {
if (Rs < 4738) {
return 1;
}
else if (Rs >= 4738) {
return 2;
}
}
else if (MVDiff >= 56) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
return 1;
}
}
}
else if (posBalance >= 59) {
if (MVDiff < 67) {
if (SC < 4639) {
return 1;
}
else if (SC >= 4639) {
if (Rs < 4534) {
return 4;
}
else if (Rs >= 4534) {
return 8;
}
}
}
else if (MVDiff >= 67) {
if (ssDCval < 161) {
return 8;
}
else if (ssDCval >= 161) {
return 1;
}
}
}
}
}
else if (mcTcor >= 93) {
if (RsDiff < 24) {
if (negBalance < 36) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
if (RsDiff < 20) {
return 1;
}
else if (RsDiff >= 20) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (mcTcor < 94) {
if (TSC < 11) {
if (Cs < 1007) {
return 4;
}
else if (Cs >= 1007) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 2;
}
}
}
else if (TSC >= 11) {
if (MVDiff < 58) {
if (Cs < 813) {
if (Rs < 4000) {
return 1;
}
else if (Rs >= 4000) {
return 2;
}
}
else if (Cs >= 813) {
return 1;
}
}
else if (MVDiff >= 58) {
if (Rs < 2728) {
return 1;
}
else if (Rs >= 2728) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (diffAFD < -1) {
if (MVDiff < 100) {
if (SC < 7024) {
return 1;
}
else if (SC >= 7024) {
if (Rs < 6129) {
return 2;
}
else if (Rs >= 6129) {
return 1;
}
}
}
else if (MVDiff >= 100) {
if (Cs < 827) {
return 8;
}
else if (Cs >= 827) {
return 1;
}
}
}
else if (diffAFD >= -1) {
if (CsDiff < 4) {
if (Cs < 716) {
if (diffRsCsdiff < 37) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 1;
}
}
else if (diffRsCsdiff >= 37) {
return 1;
}
}
else if (Cs >= 716) {
if (Rs < 2854) {
return 8;
}
else if (Rs >= 2854) {
return 2;
}
}
}
else if (CsDiff >= 4) {
if (RsCsDiff < 317) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
if (Rs < 3484) {
return 1;
}
else if (Rs >= 3484) {
return 8;
}
}
}
else if (RsCsDiff >= 317) {
if (diffAFD < 4) {
if (ssDCval < 136) {
return 1;
}
else if (ssDCval >= 136) {
return 2;
}
}
else if (diffAFD >= 4) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 24) {
if (Cs < 1013) {
return 2;
}
else if (Cs >= 1013) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Cs >= 1309) {
if (mu_mv_mag_sq < 111) {
if (RsCsDiff < 101) {
if (ssDCval < 105) {
if (negBalance < 39) {
if (Rs < 816) {
if (Cs < 1505) {
if (refDCval < 63) {
return 1;
}
else if (refDCval >= 63) {
if (MVDiff < 2) {
if (Rs < 792) {
return 8;
}
else if (Rs >= 792) {
return 1;
}
}
else if (MVDiff >= 2) {
if (Cs < 1357) {
if (Rs < 592) {
return 1;
}
else if (Rs >= 592) {
return 8;
}
}
else if (Cs >= 1357) {
if (MVDiff < 84) {
return 4;
}
else if (MVDiff >= 84) {
return 8;
}
}
}
}
}
else if (Cs >= 1505) {
if (TSC < 5) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (TSC >= 5) {
if (Cs < 1764) {
return 8;
}
else if (Cs >= 1764) {
if (Rs < 698) {
return 2;
}
else if (Rs >= 698) {
return 1;
}
}
}
}
}
else if (Rs >= 816) {
if (SC < 10037) {
if (posBalance < 43) {
if (ssDCval < 75) {
if (ssDCval < 64) {
if (Rs < 2997) {
if (ssDCval < 58) {
if (AFD < 8) {
if (ssDCval < 55) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (MVDiff < 16) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (MVDiff >= 16) {
return 2;
}
}
}
else if (ssDCval >= 55) {
if (Rs < 991) {
return 8;
}
else if (Rs >= 991) {
if (Cs < 1869) {
return 4;
}
else if (Cs >= 1869) {
return 8;
}
}
}
}
else if (AFD >= 8) {
return 2;
}
}
else if (ssDCval >= 58) {
if (tcor < 96) {
if (Rs < 1733) {
return 8;
}
else if (Rs >= 1733) {
return 2;
}
}
else if (tcor >= 96) {
return 8;
}
}
}
else if (Rs >= 2997) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (Rs < 4708) {
if (SC < 6832) {
return 8;
}
else if (SC >= 6832) {
if (Rs < 4516) {
return 4;
}
else if (Rs >= 4516) {
return 4;
}
}
}
else if (Rs >= 4708) {
return 4;
}
}
}
}
else if (ssDCval >= 64) {
if (MVDiff < 7) {
if (Rs < 3351) {
if (Rs < 2595) {
return 8;
}
else if (Rs >= 2595) {
if (Rs < 2875) {
return 2;
}
else if (Rs >= 2875) {
return 1;
}
}
}
else if (Rs >= 3351) {
if (Rs < 4456) {
return 4;
}
else if (Rs >= 4456) {
if (Rs < 5959) {
return 8;
}
else if (Rs >= 5959) {
return 4;
}
}
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (ssDCval >= 75) {
if (MVDiff < 9) {
if (Rs < 2977) {
if (negBalance < 30) {
if (MVDiff < 3) {
if (ssDCval < 92) {
if (Rs < 2476) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 2382) {
return 1;
}
else if (Rs >= 2382) {
return 8;
}
}
}
else if (Rs >= 2476) {
return 1;
}
}
else if (ssDCval >= 92) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (negBalance >= 30) {
if (Rs < 1909) {
return 8;
}
else if (Rs >= 1909) {
if (SC < 4487) {
return 1;
}
else if (SC >= 4487) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
else if (Rs >= 2977) {
if (RsCsDiff < 38) {
if (MVDiff < 6) {
if (ssDCval < 96) {
if (ssDCval < 91) {
if (refDCval < 84) {
if (Cs < 2009) {
if (Rs < 4340) {
return 4;
}
else if (Rs >= 4340) {
return 8;
}
}
else if (Cs >= 2009) {
return 4;
}
}
else if (refDCval >= 84) {
if (Rs < 3697) {
if (Rs < 3312) {
return 2;
}
else if (Rs >= 3312) {
return 8;
}
}
else if (Rs >= 3697) {
return 4;
}
}
}
else if (ssDCval >= 91) {
if (TSC < 4) {
if (Rs < 4292) {
return 8;
}
else if (Rs >= 4292) {
return 2;
}
}
else if (TSC >= 4) {
if (Rs < 3986) {
return 2;
}
else if (Rs >= 3986) {
return 8;
}
}
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (MVDiff >= 6) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (RsCsDiff >= 38) {
if (AFD < 9) {
if (SC < 5459) {
if (SC < 5370) {
if (Rs < 3432) {
if (Rs < 3327) {
return 8;
}
else if (Rs >= 3327) {
return 4;
}
}
else if (Rs >= 3432) {
return 2;
}
}
else if (SC >= 5370) {
return 4;
}
}
else if (SC >= 5459) {
if (TSC < 7) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Cs < 1805) {
return 8;
}
else if (Cs >= 1805) {
return 2;
}
}
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (AFD >= 9) {
if (Rs < 3823) {
return 8;
}
else if (Rs >= 3823) {
if (Rs < 4786) {
return 4;
}
else if (Rs >= 4786) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 9) {
if (posBalance < 27) {
if (TSC < 2) {
if (Rs < 2275) {
return 2;
}
else if (Rs >= 2275) {
return 4;
}
}
else if (TSC >= 2) {
if (Cs < 1763) {
return 8;
}
else if (Cs >= 1763) {
return 1;
}
}
}
else if (posBalance >= 27) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 2505) {
return 8;
}
else if (Rs >= 2505) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 43) {
if (Rs < 2835) {
if (mu_mv_mag_sq < 85) {
if (Rs < 1512) {
if (posBalance < 59) {
if (SC < 2536) {
if (tcor < 89) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
else if (tcor >= 89) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
}
else if (SC >= 2536) {
if (Rs < 1099) {
if (TSC < 5) {
if (mu_mv_mag_sq < 81) {
return 1;
}
else if (mu_mv_mag_sq >= 81) {
if (Rs < 916) {
return 1;
}
else if (Rs >= 916) {
return 8;
}
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (Rs >= 1099) {
if (Cs < 2144) {
if (Rs < 1317) {
if (Cs < 1851) {
if (Rs < 1159) {
if (Rs < 1114) {
return 4;
}
else if (Rs >= 1114) {
return 1;
}
}
else if (Rs >= 1159) {
return 4;
}
}
else if (Cs >= 1851) {
return 1;
}
}
else if (Rs >= 1317) {
if (SC < 2897) {
if (SC < 2703) {
return 4;
}
else if (SC >= 2703) {
return 8;
}
}
else if (SC >= 2897) {
return 4;
}
}
}
else if (Cs >= 2144) {
return 1;
}
}
}
}
else if (posBalance >= 59) {
if (ssDCval < 58) {
if (Rs < 1307) {
if (Rs < 942) {
return 2;
}
else if (Rs >= 942) {
if (Rs < 1261) {
return 8;
}
else if (Rs >= 1261) {
return 4;
}
}
}
else if (Rs >= 1307) {
return 1;
}
}
else if (ssDCval >= 58) {
if (Cs < 2125) {
return 8;
}
else if (Cs >= 2125) {
return 2;
}
}
}
}
else if (Rs >= 1512) {
if (refDCval < 85) {
if (posBalance < 44) {
if (Cs < 1671) {
if (Rs < 2128) {
return 4;
}
else if (Rs >= 2128) {
return 2;
}
}
else if (Cs >= 1671) {
return 1;
}
}
else if (posBalance >= 44) {
if (RsDiff < 7) {
if (SC < 3610) {
return 8;
}
else if (SC >= 3610) {
if (Rs < 2011) {
return 1;
}
else if (Rs >= 2011) {
if (posBalance < 47) {
if (Rs < 2262) {
return 8;
}
else if (Rs >= 2262) {
return 1;
}
}
else if (posBalance >= 47) {
return 8;
}
}
}
}
else if (RsDiff >= 7) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
if (MVDiff < 10) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 85) {
if (posBalance < 57) {
if (negBalance < 31) {
if (Rs < 2150) {
return 8;
}
else if (Rs >= 2150) {
return 1;
}
}
else if (negBalance >= 31) {
return 8;
}
}
else if (posBalance >= 57) {
if (Cs < 1631) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
else if (Cs >= 1631) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (Cs < 1642) {
if (refDCval < 87) {
if (MVDiff < 8) {
if (posBalance < 53) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (refDCval < 57) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (refDCval >= 57) {
if (MVDiff < 1) {
if (Rs < 1898) {
return 8;
}
else if (Rs >= 1898) {
if (Cs < 1592) {
return 4;
}
else if (Cs >= 1592) {
return 8;
}
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
else if (MVDiff >= 8) {
if (diffMVdiffVal < 5) {
if (Rs < 983) {
return 8;
}
else if (Rs >= 983) {
return 4;
}
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
}
else if (refDCval >= 87) {
if (Rs < 927) {
return 4;
}
else if (Rs >= 927) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (Rs < 1434) {
return 8;
}
else if (Rs >= 1434) {
if (Rs < 2258) {
return 2;
}
else if (Rs >= 2258) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1642) {
if (Cs < 2456) {
if (mu_mv_mag_sq < 109) {
if (Rs < 2694) {
if (refDCval < 78) {
if (mcTcor < 98) {
if (AFD < 4) {
if (Rs < 1079) {
return 8;
}
else if (Rs >= 1079) {
return 4;
}
}
else if (AFD >= 4) {
if (refDCval < 45) {
return 4;
}
else if (refDCval >= 45) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (AFD < 5) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
return 1;
}
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (refDCval >= 78) {
return 8;
}
}
else if (Rs >= 2694) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (Cs < 1952) {
return 1;
}
else if (Cs >= 1952) {
if (Rs < 2712) {
return 4;
}
else if (Rs >= 2712) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (Cs < 1938) {
if (Rs < 1100) {
return 8;
}
else if (Rs >= 1100) {
if (ssDCval < 84) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
return 8;
}
}
else if (ssDCval >= 84) {
return 8;
}
}
}
else if (Cs >= 1938) {
return 8;
}
}
}
else if (Cs >= 2456) {
if (RsDiff < 3) {
if (Cs < 3908) {
return 8;
}
else if (Cs >= 3908) {
return 1;
}
}
else if (RsDiff >= 3) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (Rs < 1934) {
return 1;
}
else if (Rs >= 1934) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 2835) {
if (Cs < 1933) {
if (MVDiff < 5) {
if (diffRsCsdiff < -34) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 2;
}
}
else if (diffRsCsdiff >= -34) {
if (refDCval < 87) {
if (RsCsDiff < 51) {
if (mcTcor < 97) {
if (Rs < 3202) {
return 2;
}
else if (Rs >= 3202) {
return 4;
}
}
else if (mcTcor >= 97) {
return 4;
}
}
else if (RsCsDiff >= 51) {
if (ssDCval < 71) {
if (Rs < 3151) {
return 2;
}
else if (Rs >= 3151) {
if (mu_mv_mag_sq < 108) {
return 4;
}
else if (mu_mv_mag_sq >= 108) {
return 4;
}
}
}
else if (ssDCval >= 71) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
if (Rs < 3239) {
return 4;
}
else if (Rs >= 3239) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 87) {
if (negBalance < 34) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (diffRsCsdiff < 5) {
if (mu_mv_mag_sq < 98) {
if (Cs < 1335) {
return 8;
}
else if (Cs >= 1335) {
return 2;
}
}
else if (mu_mv_mag_sq >= 98) {
return 8;
}
}
else if (diffRsCsdiff >= 5) {
if (negBalance < 30) {
if (Rs < 3846) {
return 4;
}
else if (Rs >= 3846) {
return 4;
}
}
else if (negBalance >= 30) {
return 8;
}
}
}
}
else if (negBalance >= 34) {
return 2;
}
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 87) {
if (diffRsCsdiff < -30) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
else if (diffRsCsdiff >= -30) {
return 4;
}
}
else if (ssDCval >= 87) {
if (RsCsDiff < 69) {
if (SC < 5829) {
if (Cs < 1313) {
return 8;
}
else if (Cs >= 1313) {
return 8;
}
}
else if (SC >= 5829) {
return 1;
}
}
else if (RsCsDiff >= 69) {
if (mu_mv_mag_sq < 53) {
return 1;
}
else if (mu_mv_mag_sq >= 53) {
if (Cs < 1562) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (Cs >= 1562) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1933) {
if (CsDiff < 6) {
if (mu_mv_mag_sq < 98) {
if (ssDCval < 71) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 3506) {
return 1;
}
else if (Rs >= 3506) {
return 4;
}
}
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (mu_mv_mag_sq >= 98) {
if (Rs < 3005) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 1;
}
}
else if (Rs >= 3005) {
if (Rs < 3406) {
return 8;
}
else if (Rs >= 3406) {
if (Rs < 3895) {
return 4;
}
else if (Rs >= 3895) {
if (Rs < 4225) {
return 2;
}
else if (Rs >= 4225) {
if (ssDCval < 58) {
if (Rs < 4525) {
return 4;
}
else if (Rs >= 4525) {
return 4;
}
}
else if (ssDCval >= 58) {
return 8;
}
}
}
}
}
}
}
else if (CsDiff >= 6) {
if (ssDCval < 84) {
if (AFD < 7) {
if (Cs < 3182) {
return 8;
}
else if (Cs >= 3182) {
return 4;
}
}
else if (AFD >= 7) {
return 1;
}
}
else if (ssDCval >= 84) {
if (TSC < 9) {
if (Rs < 3014) {
return 4;
}
else if (Rs >= 3014) {
return 8;
}
}
else if (TSC >= 9) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 10037) {
if (Rs < 5646) {
return 8;
}
else if (Rs >= 5646) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 1454) {
if (RsCsDiff < 51) {
if (refDCval < 83) {
if (TSC < 10) {
if (MVDiff < 4) {
if (Rs < 923) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 1;
}
}
else if (Rs >= 923) {
if (Rs < 1295) {
if (ssDCval < 55) {
if (CsDiff < 5) {
if (Cs < 1681) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 4;
}
}
else if (Cs >= 1681) {
return 8;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (Rs >= 1295) {
if (diffMVdiffVal < -3) {
if (Rs < 1348) {
return 1;
}
else if (Rs >= 1348) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
}
else if (MVDiff >= 4) {
if (posBalance < 59) {
if (ssDCval < 59) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (RsCsDiff < 32) {
if (Rs < 1040) {
return 2;
}
else if (Rs >= 1040) {
return 4;
}
}
else if (RsCsDiff >= 32) {
return 4;
}
}
}
else if (ssDCval >= 59) {
if (Rs < 1069) {
if (TSC < 7) {
if (tcor < 95) {
if (MVDiff < 20) {
if (mu_mv_mag_sq < 51) {
return 1;
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 885) {
return 8;
}
else if (Rs >= 885) {
return 1;
}
}
}
else if (MVDiff >= 20) {
return 8;
}
}
else if (tcor >= 95) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (Rs >= 1069) {
if (AFD < 10) {
if (Rs < 1304) {
if (Cs < 1517) {
return 4;
}
else if (Cs >= 1517) {
return 1;
}
}
else if (Rs >= 1304) {
return 1;
}
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (posBalance >= 59) {
if (negBalance < 55) {
if (Rs < 939) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
return 4;
}
}
else if (Rs >= 939) {
if (diffRsCsdiff < -5) {
if (Rs < 1257) {
return 2;
}
else if (Rs >= 1257) {
return 4;
}
}
else if (diffRsCsdiff >= -5) {
return 4;
}
}
}
else if (negBalance >= 55) {
return 8;
}
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (refDCval >= 83) {
if (ssDCval < 100) {
if (posBalance < 60) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (Cs < 1374) {
return 4;
}
else if (Cs >= 1374) {
return 8;
}
}
}
else if (posBalance >= 60) {
if (posBalance < 68) {
if (AFD < 15) {
if (Cs < 1676) {
return 8;
}
else if (Cs >= 1676) {
if (Rs < 1061) {
return 8;
}
else if (Rs >= 1061) {
return 4;
}
}
}
else if (AFD >= 15) {
return 4;
}
}
else if (posBalance >= 68) {
return 2;
}
}
}
else if (ssDCval >= 100) {
if (diffAFD < -1) {
if (Rs < 976) {
return 4;
}
else if (Rs >= 976) {
return 8;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
}
else if (RsCsDiff >= 51) {
if (tcor < 88) {
if (ssDCval < 56) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (negBalance < 56) {
if (Cs < 1612) {
if (AFD < 16) {
if (Rs < 874) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
else if (Rs >= 874) {
if (diffRsCsdiff < 28) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
if (TSC < 10) {
if (SC < 2317) {
return 2;
}
else if (SC >= 2317) {
return 8;
}
}
else if (TSC >= 10) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 28) {
return 2;
}
}
}
else if (AFD >= 16) {
if (MVDiff < 53) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
return 8;
}
}
else if (MVDiff >= 53) {
return 8;
}
}
}
else if (Cs >= 1612) {
if (TSC < 6) {
if (Rs < 821) {
return 1;
}
else if (Rs >= 821) {
return 8;
}
}
else if (TSC >= 6) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (diffAFD < 2) {
if (diffRsCsdiff < -42) {
return 8;
}
else if (diffRsCsdiff >= -42) {
if (TSC < 8) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (RsCsDiff < 87) {
return 8;
}
else if (RsCsDiff >= 87) {
if (posBalance < 60) {
if (Cs < 1832) {
return 8;
}
else if (Cs >= 1832) {
return 4;
}
}
else if (posBalance >= 60) {
return 4;
}
}
}
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 1255) {
return 2;
}
else if (Rs >= 1255) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 56) {
if (MVDiff < 28) {
if (TSC < 9) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (SC < 2610) {
return 4;
}
else if (SC >= 2610) {
return 2;
}
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (MVDiff >= 28) {
if (mcTcor < 80) {
if (Cs < 1473) {
if (Cs < 1358) {
return 8;
}
else if (Cs >= 1358) {
return 4;
}
}
else if (Cs >= 1473) {
return 8;
}
}
else if (mcTcor >= 80) {
return 2;
}
}
}
}
}
else if (tcor >= 88) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 2) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1220) {
if (tcor < 94) {
if (Cs < 1357) {
return 2;
}
else if (Cs >= 1357) {
return 8;
}
}
else if (tcor >= 94) {
return 2;
}
}
else if (Rs >= 1220) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
else if (mcTcor >= 98) {
if (Rs < 1312) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
else if (Rs >= 1312) {
return 2;
}
}
}
}
}
else if (Rs >= 1454) {
if (CsDiff < 7) {
if (mcTcor < 94) {
if (negBalance < 48) {
if (RsCsDiff < 67) {
if (mu_mv_mag_sq < 95) {
if (posBalance < 64) {
if (Cs < 1420) {
if (Cs < 1402) {
return 8;
}
else if (Cs >= 1402) {
return 1;
}
}
else if (Cs >= 1420) {
return 8;
}
}
else if (posBalance >= 64) {
if (Rs < 2207) {
if (Rs < 1914) {
return 4;
}
else if (Rs >= 1914) {
return 2;
}
}
else if (Rs >= 2207) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (Cs < 2706) {
if (TSC < 9) {
if (mcTcor < 83) {
return 4;
}
else if (mcTcor >= 83) {
return 8;
}
}
else if (TSC >= 9) {
return 4;
}
}
else if (Cs >= 2706) {
return 4;
}
}
}
else if (RsCsDiff >= 67) {
if (Rs < 2225) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 42) {
if (AFD < 15) {
if (Cs < 1533) {
return 8;
}
else if (Cs >= 1533) {
return 1;
}
}
else if (AFD >= 15) {
return 4;
}
}
else if (mu_mv_mag_sq >= 42) {
if (Cs < 1839) {
return 8;
}
else if (Cs >= 1839) {
if (Rs < 1916) {
return 4;
}
else if (Rs >= 1916) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 8;
}
}
}
}
}
else if (ssDCval >= 84) {
if (AFD < 12) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 2;
}
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (Rs >= 2225) {
if (SC < 5485) {
if (diffRsCsdiff < -48) {
if (Rs < 2560) {
return 8;
}
else if (Rs >= 2560) {
return 4;
}
}
else if (diffRsCsdiff >= -48) {
if (posBalance < 68) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
else if (posBalance >= 68) {
return 4;
}
}
}
else if (SC >= 5485) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (TSC < 8) {
if (Rs < 3572) {
return 4;
}
else if (Rs >= 3572) {
return 8;
}
}
else if (TSC >= 8) {
if (Rs < 4453) {
return 4;
}
else if (Rs >= 4453) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 48) {
if (negBalance < 58) {
if (posBalance < 56) {
if (mu_mv_mag_sq < 74) {
if (diffRsCsdiff < -12) {
if (Cs < 1469) {
if (Rs < 2035) {
return 1;
}
else if (Rs >= 2035) {
return 2;
}
}
else if (Cs >= 1469) {
if (ssDCval < 99) {
if (mu_mv_mag_sq < 23) {
if (Rs < 2324) {
return 8;
}
else if (Rs >= 2324) {
return 4;
}
}
else if (mu_mv_mag_sq >= 23) {
return 8;
}
}
else if (ssDCval >= 99) {
return 2;
}
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 1738) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 2;
}
}
else if (Cs >= 1738) {
if (Cs < 1862) {
return 4;
}
else if (Cs >= 1862) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 4;
}
}
}
}
}
else if (posBalance >= 56) {
if (gchDC < 1) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Rs < 2171) {
return 2;
}
else if (Rs >= 2171) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (diffRsCsdiff < -40) {
if (Rs < 2148) {
return 4;
}
else if (Rs >= 2148) {
return 8;
}
}
else if (diffRsCsdiff >= -40) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 58) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
if (refDCval < 90) {
return 4;
}
else if (refDCval >= 90) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (SC < 6973) {
if (SC < 4689) {
if (refDCval < 89) {
if (Rs < 2866) {
if (AFD < 8) {
if (Cs < 1535) {
if (SC < 3572) {
if (Cs < 1314) {
if (Rs < 1986) {
return 1;
}
else if (Rs >= 1986) {
return 2;
}
}
else if (Cs >= 1314) {
if (ssDCval < 80) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 1872) {
return 8;
}
else if (Rs >= 1872) {
return 1;
}
}
}
else if (ssDCval >= 80) {
return 1;
}
}
}
else if (SC >= 3572) {
if (ssDCval < 85) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
return 1;
}
}
else if (ssDCval >= 85) {
if (Rs < 2206) {
return 8;
}
else if (Rs >= 2206) {
return 4;
}
}
}
}
else if (Cs >= 1535) {
if (Rs < 1481) {
return 1;
}
else if (Rs >= 1481) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Rs < 2396) {
return 8;
}
else if (Rs >= 2396) {
return 1;
}
}
}
}
}
else if (AFD >= 8) {
if (MVDiff < 4) {
if (refDCval < 80) {
if (Cs < 1513) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (Cs < 1320) {
return 1;
}
else if (Cs >= 1320) {
return 8;
}
}
}
else if (Cs >= 1513) {
if (AFD < 10) {
if (RsCsDiff < 46) {
return 4;
}
else if (RsCsDiff >= 46) {
return 8;
}
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (refDCval >= 80) {
if (negBalance < 51) {
if (Rs < 2603) {
if (Rs < 1631) {
return 8;
}
else if (Rs >= 1631) {
if (AFD < 9) {
if (Rs < 2111) {
if (mcTcor < 95) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
else if (Rs >= 2111) {
if (Rs < 2422) {
if (Cs < 1459) {
return 1;
}
else if (Cs >= 1459) {
return 8;
}
}
else if (Rs >= 2422) {
return 1;
}
}
}
else if (AFD >= 9) {
if (Cs < 1349) {
return 8;
}
else if (Cs >= 1349) {
return 1;
}
}
}
}
else if (Rs >= 2603) {
return 8;
}
}
else if (negBalance >= 51) {
if (Rs < 2186) {
return 1;
}
else if (Rs >= 2186) {
return 4;
}
}
}
}
else if (MVDiff >= 4) {
if (refDCval < 81) {
if (Cs < 1593) {
if (diffRsCsdiff < 12) {
if (SC < 3004) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
else if (SC >= 3004) {
return 8;
}
}
else if (diffRsCsdiff >= 12) {
return 4;
}
}
else if (Cs >= 1593) {
if (Rs < 2296) {
return 4;
}
else if (Rs >= 2296) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (diffMVdiffVal < -8) {
if (Cs < 1490) {
return 2;
}
else if (Cs >= 1490) {
if (Rs < 2325) {
return 8;
}
else if (Rs >= 2325) {
return 4;
}
}
}
else if (diffMVdiffVal >= -8) {
if (MVDiff < 9) {
if (Cs < 1569) {
if (refDCval < 83) {
return 8;
}
else if (refDCval >= 83) {
if (negBalance < 50) {
if (TSC < 7) {
if (Rs < 2685) {
return 2;
}
else if (Rs >= 2685) {
return 1;
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
else if (Cs >= 1569) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
return 8;
}
}
}
else if (MVDiff >= 9) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2866) {
if (Cs < 1437) {
return 2;
}
else if (Cs >= 1437) {
if (Rs < 3092) {
return 4;
}
else if (Rs >= 3092) {
return 2;
}
}
}
}
else if (refDCval >= 89) {
if (Rs < 2186) {
if (posBalance < 61) {
if (Cs < 1862) {
if (mu_mv_mag_sq < 109) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Rs < 1920) {
if (Cs < 1560) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 8;
}
}
else if (Cs >= 1560) {
return 2;
}
}
else if (Rs >= 1920) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 1894) {
return 8;
}
else if (Rs >= 1894) {
return 2;
}
}
}
else if (Cs >= 1862) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (MVDiff < 16) {
if (mu_mv_mag_sq < 100) {
return 2;
}
else if (mu_mv_mag_sq >= 100) {
return 1;
}
}
else if (MVDiff >= 16) {
return 8;
}
}
}
}
else if (posBalance >= 61) {
if (Rs < 1503) {
return 8;
}
else if (Rs >= 1503) {
return 2;
}
}
}
else if (Rs >= 2186) {
if (refDCval < 102) {
if (Cs < 1897) {
if (mu_mv_mag_sq < 81) {
if (diffRsCsdiff < -42) {
if (Rs < 2321) {
return 1;
}
else if (Rs >= 2321) {
return 8;
}
}
else if (diffRsCsdiff >= -42) {
return 8;
}
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
if (RsCsDiff < 46) {
return 8;
}
else if (RsCsDiff >= 46) {
return 4;
}
}
}
}
else if (Cs >= 1897) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 2319) {
return 1;
}
else if (Rs >= 2319) {
return 8;
}
}
}
}
else if (refDCval >= 102) {
if (Rs < 2811) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (Rs >= 2811) {
return 8;
}
}
}
}
}
else if (SC >= 4689) {
if (Cs < 1624) {
if (tcor < 92) {
if (ssDCval < 96) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (ssDCval >= 96) {
return 8;
}
}
else if (tcor >= 92) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
else if (Cs >= 1624) {
if (CsDiff < 4) {
if (Rs < 4398) {
if (negBalance < 52) {
if (Rs < 2444) {
return 4;
}
else if (Rs >= 2444) {
return 8;
}
}
else if (negBalance >= 52) {
if (Rs < 2575) {
return 8;
}
else if (Rs >= 2575) {
return 4;
}
}
}
else if (Rs >= 4398) {
return 4;
}
}
else if (CsDiff >= 4) {
if (mu_mv_mag_sq < 57) {
return 8;
}
else if (mu_mv_mag_sq >= 57) {
if (SC < 5084) {
if (CsDiff < 5) {
if (diffRsCsdiff < -17) {
return 8;
}
else if (diffRsCsdiff >= -17) {
return 4;
}
}
else if (CsDiff >= 5) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (Rs < 2762) {
return 8;
}
else if (Rs >= 2762) {
return 4;
}
}
}
}
else if (SC >= 5084) {
if (posBalance < 54) {
if (mcTcor < 97) {
return 4;
}
else if (mcTcor >= 97) {
if (Cs < 2315) {
return 2;
}
else if (Cs >= 2315) {
if (Rs < 3278) {
return 4;
}
else if (Rs >= 3278) {
return 8;
}
}
}
}
else if (posBalance >= 54) {
if (ssDCval < 81) {
if (posBalance < 56) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
else if (posBalance >= 56) {
if (Rs < 2721) {
return 4;
}
else if (Rs >= 2721) {
return 8;
}
}
}
else if (ssDCval >= 81) {
if (Rs < 3831) {
if (TSC < 10) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (Rs >= 3831) {
if (Rs < 3982) {
return 8;
}
else if (Rs >= 3982) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (SC >= 6973) {
if (Cs < 2489) {
return 4;
}
else if (Cs >= 2489) {
return 8;
}
}
}
}
else if (CsDiff >= 7) {
if (Rs < 3195) {
if (Cs < 2964) {
if (refDCval < 98) {
if (RsDiff < 6) {
if (Cs < 2298) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 1;
}
}
else if (Cs >= 2298) {
if (diffAFD < -1) {
if (Rs < 2063) {
return 8;
}
else if (Rs >= 2063) {
return 4;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (RsDiff >= 6) {
if (refDCval < 50) {
if (posBalance < 59) {
if (refDCval < 48) {
return 8;
}
else if (refDCval >= 48) {
if (Rs < 1836) {
return 1;
}
else if (Rs >= 1836) {
return 4;
}
}
}
else if (posBalance >= 59) {
return 4;
}
}
else if (refDCval >= 50) {
if (AFD < 17) {
if (mcTcor < 94) {
if (Rs < 1917) {
if (negBalance < 45) {
if (Cs < 1617) {
return 8;
}
else if (Cs >= 1617) {
return 4;
}
}
else if (negBalance >= 45) {
if (diffAFD < -1) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (Rs >= 1917) {
if (Rs < 1937) {
return 2;
}
else if (Rs >= 1937) {
if (SC < 5475) {
return 8;
}
else if (SC >= 5475) {
if (Rs < 2788) {
return 2;
}
else if (Rs >= 2788) {
return 8;
}
}
}
}
}
else if (mcTcor >= 94) {
if (AFD < 12) {
if (diffRsCsdiff < 36) {
return 8;
}
else if (diffRsCsdiff >= 36) {
if (Rs < 2616) {
return 8;
}
else if (Rs >= 2616) {
return 4;
}
}
}
else if (AFD >= 12) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
if (negBalance < 48) {
if (MVDiff < 12) {
if (Cs < 1463) {
return 1;
}
else if (Cs >= 1463) {
return 4;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (negBalance >= 48) {
return 8;
}
}
}
}
}
else if (AFD >= 17) {
if (MVDiff < 42) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (posBalance < 62) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 4;
}
}
else if (posBalance >= 62) {
return 4;
}
}
}
else if (MVDiff >= 42) {
return 1;
}
}
}
}
}
else if (refDCval >= 98) {
if (mcTcor < 96) {
if (AFD < 16) {
if (refDCval < 99) {
if (Cs < 2145) {
if (Rs < 2295) {
return 2;
}
else if (Rs >= 2295) {
return 8;
}
}
else if (Cs >= 2145) {
return 4;
}
}
else if (refDCval >= 99) {
return 8;
}
}
else if (AFD >= 16) {
return 4;
}
}
else if (mcTcor >= 96) {
if (Rs < 2136) {
return 2;
}
else if (Rs >= 2136) {
return 8;
}
}
}
}
else if (Cs >= 2964) {
return 4;
}
}
else if (Rs >= 3195) {
if (negBalance < 48) {
if (SC < 6828) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
if (SC < 5846) {
return 4;
}
else if (SC >= 5846) {
return 1;
}
}
}
else if (SC >= 6828) {
if (Cs < 3337) {
return 4;
}
else if (Cs >= 3337) {
if (Rs < 3969) {
return 1;
}
else if (Rs >= 3969) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
if (mu_mv_mag_sq < 94) {
return 8;
}
else if (mu_mv_mag_sq >= 94) {
if (Cs < 3212) {
return 1;
}
else if (Cs >= 3212) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 105) {
if (ssDCval < 163) {
if (Rs < 4405) {
if (Rs < 1436) {
if (mu_mv_mag_sq < 46) {
if (refDCval < 140) {
if (mcTcor < 87) {
if (Rs < 962) {
return 2;
}
else if (Rs >= 962) {
return 4;
}
}
else if (mcTcor >= 87) {
if (Rs < 1339) {
if (ssDCval < 110) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 8;
}
}
else if (ssDCval >= 110) {
return 8;
}
}
else if (Rs >= 1339) {
if (RsDiff < 7) {
if (diffTSC < 1) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
return 8;
}
}
else if (diffTSC >= 1) {
if (Rs < 1356) {
return 2;
}
else if (Rs >= 1356) {
return 1;
}
}
}
else if (RsDiff >= 7) {
return 4;
}
}
}
}
else if (refDCval >= 140) {
if (mu_mv_mag_sq < 34) {
if (Rs < 1273) {
if (CsDiff < 7) {
if (Rs < 1224) {
return 2;
}
else if (Rs >= 1224) {
return 4;
}
}
else if (CsDiff >= 7) {
return 1;
}
}
else if (Rs >= 1273) {
return 2;
}
}
else if (mu_mv_mag_sq >= 34) {
if (Rs < 1410) {
return 4;
}
else if (Rs >= 1410) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (Cs < 1497) {
if (Rs < 697) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
else if (Rs >= 697) {
if (ssDCval < 107) {
if (Rs < 1017) {
return 4;
}
else if (Rs >= 1017) {
return 8;
}
}
else if (ssDCval >= 107) {
if (tcor < 75) {
return 1;
}
else if (tcor >= 75) {
if (Rs < 1076) {
if (Rs < 1018) {
if (diffRsCsdiff < -14) {
if (Rs < 833) {
return 8;
}
else if (Rs >= 833) {
return 4;
}
}
else if (diffRsCsdiff >= -14) {
return 8;
}
}
else if (Rs >= 1018) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (Rs >= 1076) {
return 8;
}
}
}
}
}
else if (Cs >= 1497) {
if (negBalance < 45) {
if (MVDiff < 19) {
if (mcTcor < 94) {
if (Rs < 1296) {
if (ssDCval < 135) {
return 4;
}
else if (ssDCval >= 135) {
return 8;
}
}
else if (Rs >= 1296) {
if (Rs < 1372) {
return 8;
}
else if (Rs >= 1372) {
return 1;
}
}
}
else if (mcTcor >= 94) {
if (mu_mv_mag_sq < 100) {
if (SC < 3031) {
return 8;
}
else if (SC >= 3031) {
if (Rs < 1099) {
return 1;
}
else if (Rs >= 1099) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 1059) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
else if (Rs >= 1059) {
if (ssDCval < 118) {
return 8;
}
else if (ssDCval >= 118) {
return 4;
}
}
}
}
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
}
else if (Rs >= 1436) {
if (mu_mv_mag_sq < 40) {
if (Rs < 2017) {
if (ssDCval < 125) {
if (mcTcor < 97) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (CsDiff < 6) {
if (MVDiff < 52) {
return 4;
}
else if (MVDiff >= 52) {
return 8;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (SC < 3018) {
return 4;
}
else if (SC >= 3018) {
return 2;
}
}
}
}
else if (ssDCval >= 125) {
if (ssDCval < 139) {
if (negBalance < 48) {
if (RsDiff < 5) {
if (Rs < 1481) {
return 1;
}
else if (Rs >= 1481) {
return 8;
}
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (negBalance >= 48) {
if (Cs < 1334) {
return 2;
}
else if (Cs >= 1334) {
return 8;
}
}
}
else if (ssDCval >= 139) {
if (Rs < 1697) {
if (Cs < 1440) {
return 1;
}
else if (Cs >= 1440) {
return 2;
}
}
else if (Rs >= 1697) {
if (ssDCval < 152) {
return 8;
}
else if (ssDCval >= 152) {
return 2;
}
}
}
}
}
else if (Rs >= 2017) {
if (ssDCval < 110) {
if (RsCsDiff < 55) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
}
else if (RsCsDiff >= 55) {
if (posBalance < 62) {
if (MVDiff < 19) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
return 8;
}
}
else if (MVDiff >= 19) {
if (Rs < 3124) {
return 4;
}
else if (Rs >= 3124) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (Rs < 2166) {
return 4;
}
else if (Rs >= 2166) {
return 1;
}
}
}
}
else if (ssDCval >= 110) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (negBalance < 49) {
if (posBalance < 33) {
if (Rs < 2768) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (diffMVdiffVal < 4) {
return 4;
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
}
else if (Rs >= 2768) {
if (ssDCval < 120) {
if (Rs < 3952) {
return 8;
}
else if (Rs >= 3952) {
return 1;
}
}
else if (ssDCval >= 120) {
return 2;
}
}
}
else if (posBalance >= 33) {
if (Cs < 1363) {
if (Rs < 2842) {
if (negBalance < 29) {
if (Rs < 2289) {
return 8;
}
else if (Rs >= 2289) {
return 1;
}
}
else if (negBalance >= 29) {
return 8;
}
}
else if (Rs >= 2842) {
if (mu_mv_mag_sq < 84) {
if (Rs < 3776) {
return 8;
}
else if (Rs >= 3776) {
return 4;
}
}
else if (mu_mv_mag_sq >= 84) {
return 2;
}
}
}
else if (Cs >= 1363) {
if (mu_mv_mag_sq < 64) {
if (refDCval < 139) {
if (TSC < 9) {
if (RsCsDiff < 43) {
if (refDCval < 109) {
if (refDCval < 107) {
return 8;
}
else if (refDCval >= 107) {
if (Cs < 1896) {
return 4;
}
else if (Cs >= 1896) {
return 8;
}
}
}
else if (refDCval >= 109) {
return 8;
}
}
else if (RsCsDiff >= 43) {
if (Rs < 1994) {
if (SC < 3643) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (SC >= 3643) {
if (Cs < 1936) {
return 1;
}
else if (Cs >= 1936) {
return 8;
}
}
}
else if (Rs >= 1994) {
if (Rs < 4257) {
if (mu_mv_mag_sq < 63) {
return 8;
}
else if (mu_mv_mag_sq >= 63) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (Rs >= 4257) {
return 1;
}
}
}
}
else if (TSC >= 9) {
if (mcTcor < 93) {
if (Cs < 1513) {
return 1;
}
else if (Cs >= 1513) {
return 8;
}
}
else if (mcTcor >= 93) {
if (Cs < 1952) {
return 1;
}
else if (Cs >= 1952) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 4;
}
}
}
}
}
else if (refDCval >= 139) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
if (Rs < 2819) {
return 4;
}
else if (Rs >= 2819) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (negBalance < 46) {
if (SC < 3932) {
if (Rs < 2451) {
if (mcTcor < 87) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (mcTcor >= 87) {
if (diffTSC < -2) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (Rs < 2059) {
return 1;
}
else if (Rs >= 2059) {
return 8;
}
}
}
else if (diffTSC >= -2) {
return 8;
}
}
}
else if (Rs >= 2451) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (SC >= 3932) {
if (Cs < 1509) {
if (AFD < 11) {
if (diffRsCsdiff < 19) {
return 8;
}
else if (diffRsCsdiff >= 19) {
if (Rs < 2723) {
return 8;
}
else if (Rs >= 2723) {
return 4;
}
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (Cs >= 1509) {
return 8;
}
}
}
else if (negBalance >= 46) {
if (diffAFD < -4) {
if (Rs < 3197) {
return 1;
}
else if (Rs >= 3197) {
return 4;
}
}
else if (diffAFD >= -4) {
if (diffAFD < 4) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
if (Cs < 2524) {
if (Rs < 2815) {
return 8;
}
else if (Rs >= 2815) {
return 1;
}
}
else if (Cs >= 2524) {
return 8;
}
}
}
}
else if (diffAFD >= 4) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 49) {
if (posBalance < 56) {
if (Rs < 1529) {
return 2;
}
else if (Rs >= 1529) {
if (RsDiff < 9) {
if (negBalance < 50) {
if (posBalance < 46) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (RsDiff >= 9) {
return 4;
}
}
}
else if (posBalance >= 56) {
if (tcor < 88) {
if (Cs < 2221) {
return 8;
}
else if (Cs >= 2221) {
return 4;
}
}
else if (tcor >= 88) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (ssDCval < 122) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
return 4;
}
}
else if (ssDCval >= 122) {
if (mu_mv_mag_sq < 81) {
if (Cs < 1730) {
return 2;
}
else if (Cs >= 1730) {
return 8;
}
}
else if (mu_mv_mag_sq >= 81) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 4405) {
if (Cs < 1798) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (Cs >= 1798) {
if (Cs < 3462) {
if (CsDiff < 7) {
if (MVDiff < 3) {
if (Rs < 6055) {
return 2;
}
else if (Rs >= 6055) {
return 4;
}
}
else if (MVDiff >= 3) {
if (Rs < 4912) {
return 8;
}
else if (Rs >= 4912) {
return 1;
}
}
}
else if (CsDiff >= 7) {
return 1;
}
}
else if (Cs >= 3462) {
return 8;
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2283) {
if (Rs < 1049) {
return 1;
}
else if (Rs >= 1049) {
return 2;
}
}
else if (Rs >= 2283) {
return 8;
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1921) {
if (Rs < 3333) {
if (refDCval < 86) {
if (negBalance < 48) {
if (mcTcor < 86) {
if (posBalance < 55) {
if (TSC < 7) {
if (MVDiff < 40) {
return 8;
}
else if (MVDiff >= 40) {
return 1;
}
}
else if (TSC >= 7) {
if (Rs < 1998) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (Rs >= 1998) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 1;
}
}
}
}
else if (posBalance >= 55) {
if (Rs < 1558) {
if (mcTcor < 54) {
return 8;
}
else if (mcTcor >= 54) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 4;
}
}
}
else if (Rs >= 1558) {
if (mu_mv_mag_sq < 20) {
if (Rs < 2383) {
return 2;
}
else if (Rs >= 2383) {
return 8;
}
}
else if (mu_mv_mag_sq >= 20) {
return 8;
}
}
}
}
else if (mcTcor >= 86) {
if (TSC < 6) {
if (diffMVdiffVal < -9) {
if (Rs < 637) {
return 2;
}
else if (Rs >= 637) {
return 8;
}
}
else if (diffMVdiffVal >= -9) {
if (Rs < 448) {
return 4;
}
else if (Rs >= 448) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (TSC >= 6) {
if (RsDiff < 10) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2114) {
if (Rs < 768) {
return 2;
}
else if (Rs >= 768) {
if (Rs < 1585) {
return 1;
}
else if (Rs >= 1585) {
return 2;
}
}
}
else if (Rs >= 2114) {
if (Rs < 3171) {
return 4;
}
else if (Rs >= 3171) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 2005) {
if (negBalance < 47) {
if (refDCval < 64) {
return 4;
}
else if (refDCval >= 64) {
return 8;
}
}
else if (negBalance >= 47) {
return 4;
}
}
else if (Rs >= 2005) {
if (AFD < 14) {
if (Cs < 1576) {
if (negBalance < 47) {
if (diffMVdiffVal < 1) {
if (Rs < 2633) {
return 2;
}
else if (Rs >= 2633) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
else if (negBalance >= 47) {
return 8;
}
}
else if (Cs >= 1576) {
return 8;
}
}
else if (AFD >= 14) {
if (RsCsDiff < 137) {
if (Rs < 2193) {
return 2;
}
else if (Rs >= 2193) {
return 4;
}
}
else if (RsCsDiff >= 137) {
return 2;
}
}
}
}
}
else if (RsDiff >= 10) {
if (Cs < 1440) {
if (Rs < 1763) {
if (Rs < 1743) {
return 4;
}
else if (Rs >= 1743) {
return 1;
}
}
else if (Rs >= 1763) {
if (TSC < 8) {
if (Rs < 3189) {
return 2;
}
else if (Rs >= 3189) {
if (Rs < 3248) {
return 4;
}
else if (Rs >= 3248) {
return 2;
}
}
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (Cs >= 1440) {
if (tcor < 79) {
if (AFD < 16) {
if (Rs < 2910) {
return 8;
}
else if (Rs >= 2910) {
return 1;
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (tcor >= 79) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (Rs < 3307) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
return 8;
}
}
else if (Rs >= 3307) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 48) {
if (mu_mv_mag_sq < 14) {
if (Rs < 2780) {
if (Cs < 1656) {
if (mcTcor < 93) {
if (MVDiff < 18) {
if (MVDiff < 10) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (Rs < 1956) {
return 8;
}
else if (Rs >= 1956) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (posBalance < 62) {
if (Rs < 2223) {
if (diffAFD < 3) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (diffAFD < -1) {
if (Cs < 1618) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 4;
}
}
else if (Cs >= 1618) {
return 4;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (diffAFD >= 3) {
return 8;
}
}
else if (Rs >= 2223) {
return 8;
}
}
else if (posBalance >= 62) {
if (diffRsCsdiff < 18) {
return 8;
}
else if (diffRsCsdiff >= 18) {
if (Rs < 1643) {
return 4;
}
else if (Rs >= 1643) {
return 8;
}
}
}
}
}
else if (MVDiff >= 18) {
if (mcTcor < 70) {
if (Rs < 1938) {
return 4;
}
else if (Rs >= 1938) {
return 1;
}
}
else if (mcTcor >= 70) {
if (tcor < 54) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 8;
}
}
else if (tcor >= 54) {
if (Cs < 1572) {
if (RsCsDiff < 194) {
if (Rs < 1899) {
if (refDCval < 64) {
return 8;
}
else if (refDCval >= 64) {
if (diffRsCsdiff < 51) {
return 4;
}
else if (diffRsCsdiff >= 51) {
return 2;
}
}
}
else if (Rs >= 1899) {
if (tcor < 70) {
if (Rs < 2235) {
return 2;
}
else if (Rs >= 2235) {
return 4;
}
}
else if (tcor >= 70) {
return 8;
}
}
}
else if (RsCsDiff >= 194) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
return 4;
}
}
}
else if (Cs >= 1572) {
if (Cs < 1626) {
if (posBalance < 66) {
return 2;
}
else if (posBalance >= 66) {
if (Rs < 1834) {
return 1;
}
else if (Rs >= 1834) {
return 4;
}
}
}
else if (Cs >= 1626) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (Rs < 1701) {
if (diffRsCsdiff < 10) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (Cs < 1478) {
return 1;
}
else if (Cs >= 1478) {
return 2;
}
}
}
else if (diffRsCsdiff >= 10) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
}
else if (Rs >= 1701) {
if (diffTSC < 1) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
}
else if (Cs >= 1656) {
if (tcor < 59) {
if (mcTcor < 70) {
if (Rs < 1962) {
return 4;
}
else if (Rs >= 1962) {
return 8;
}
}
else if (mcTcor >= 70) {
if (MVDiff < 21) {
if (Rs < 1603) {
return 4;
}
else if (Rs >= 1603) {
return 2;
}
}
else if (MVDiff >= 21) {
if (Rs < 2350) {
return 1;
}
else if (Rs >= 2350) {
return 4;
}
}
}
}
else if (tcor >= 59) {
if (tcor < 84) {
if (mu_mv_mag_sq < 3) {
if (diffAFD < 3) {
if (gchDC < 2) {
if (Rs < 2738) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
if (AFD < 26) {
return 4;
}
else if (AFD >= 26) {
return 8;
}
}
}
}
else if (Rs >= 2738) {
return 8;
}
}
else if (gchDC >= 2) {
if (posBalance < 55) {
if (AFD < 19) {
if (Rs < 1544) {
return 2;
}
else if (Rs >= 1544) {
return 1;
}
}
else if (AFD >= 19) {
return 8;
}
}
else if (posBalance >= 55) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
if (negBalance < 54) {
return 4;
}
else if (negBalance >= 54) {
if (Cs < 1861) {
return 2;
}
else if (Cs >= 1861) {
return 1;
}
}
}
}
}
}
else if (diffAFD >= 3) {
if (Rs < 1784) {
return 1;
}
else if (Rs >= 1784) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 1871) {
if (Rs < 1884) {
if (tcor < 77) {
return 4;
}
else if (tcor >= 77) {
return 8;
}
}
else if (Rs >= 1884) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
return 8;
}
}
}
else if (Cs >= 1871) {
return 8;
}
}
}
else if (tcor >= 84) {
if (Rs < 1922) {
return 1;
}
else if (Rs >= 1922) {
return 8;
}
}
}
}
}
else if (Rs >= 2780) {
if (refDCval < 81) {
if (mcTcor < 93) {
return 2;
}
else if (mcTcor >= 93) {
if (Rs < 2947) {
return 2;
}
else if (Rs >= 2947) {
return 4;
}
}
}
else if (refDCval >= 81) {
if (Rs < 3103) {
return 4;
}
else if (Rs >= 3103) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (AFD < 17) {
if (tcor < 66) {
if (negBalance < 53) {
if (Rs < 967) {
return 4;
}
else if (Rs >= 967) {
return 8;
}
}
else if (negBalance >= 53) {
return 4;
}
}
else if (tcor >= 66) {
if (ssDCval < 77) {
if (Rs < 1640) {
if (Rs < 1525) {
return 8;
}
else if (Rs >= 1525) {
return 1;
}
}
else if (Rs >= 1640) {
return 8;
}
}
else if (ssDCval >= 77) {
if (diffRsCsdiff < -15) {
if (SC < 4270) {
return 4;
}
else if (SC >= 4270) {
if (Cs < 1885) {
return 8;
}
else if (Cs >= 1885) {
return 4;
}
}
}
else if (diffRsCsdiff >= -15) {
if (Rs < 3123) {
if (tcor < 82) {
if (Rs < 2576) {
return 4;
}
else if (Rs >= 2576) {
return 8;
}
}
else if (tcor >= 82) {
return 8;
}
}
else if (Rs >= 3123) {
return 2;
}
}
}
}
}
else if (AFD >= 17) {
if (diffMVdiffVal < -11) {
if (mcTcor < 89) {
return 1;
}
else if (mcTcor >= 89) {
if (Rs < 1888) {
return 4;
}
else if (Rs >= 1888) {
return 2;
}
}
}
else if (diffMVdiffVal >= -11) {
if (RsDiff < 15) {
if (negBalance < 53) {
if (AFD < 20) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
if (Rs < 1849) {
return 1;
}
else if (Rs >= 1849) {
if (SC < 3858) {
return 2;
}
else if (SC >= 3858) {
return 4;
}
}
}
}
else if (AFD >= 20) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 4;
}
}
}
else if (negBalance >= 53) {
if (ssDCval < 82) {
if (diffMVdiffVal < -9) {
return 4;
}
else if (diffMVdiffVal >= -9) {
if (Rs < 1607) {
if (gchDC < 3) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
return 4;
}
}
else if (gchDC >= 3) {
return 8;
}
}
else if (Rs >= 1607) {
return 8;
}
}
}
else if (ssDCval >= 82) {
if (Rs < 2964) {
return 8;
}
else if (Rs >= 2964) {
return 4;
}
}
}
}
else if (RsDiff >= 15) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 86) {
if (AFD < 19) {
if (Cs < 1616) {
if (TSC < 9) {
if (refDCval < 98) {
if (posBalance < 49) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
else if (posBalance >= 49) {
if (MVDiff < 17) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
if (Rs < 2618) {
return 2;
}
else if (Rs >= 2618) {
if (Rs < 3189) {
return 4;
}
else if (Rs >= 3189) {
return 2;
}
}
}
}
else if (MVDiff >= 17) {
if (Rs < 2068) {
if (Rs < 1571) {
return 1;
}
else if (Rs >= 1571) {
return 8;
}
}
else if (Rs >= 2068) {
return 2;
}
}
}
}
else if (refDCval >= 98) {
if (negBalance < 27) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
else if (negBalance >= 27) {
if (RsCsDiff < 189) {
if (Rs < 2292) {
if (mcTcor < 95) {
if (AFD < 17) {
if (ssDCval < 99) {
return 2;
}
else if (ssDCval >= 99) {
return 8;
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (mcTcor >= 95) {
if (MVDiff < 15) {
if (SC < 3577) {
return 4;
}
else if (SC >= 3577) {
return 8;
}
}
else if (MVDiff >= 15) {
if (posBalance < 65) {
return 2;
}
else if (posBalance >= 65) {
return 4;
}
}
}
}
else if (Rs >= 2292) {
if (tcor < 68) {
return 1;
}
else if (tcor >= 68) {
return 8;
}
}
}
else if (RsCsDiff >= 189) {
if (ssDCval < 108) {
return 2;
}
else if (ssDCval >= 108) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (Rs < 2114) {
return 1;
}
else if (Rs >= 2114) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 9) {
if (SC < 4020) {
if (TSC < 13) {
if (Cs < 1370) {
if (tcor < 73) {
return 2;
}
else if (tcor >= 73) {
if (TSC < 12) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
if (ssDCval < 87) {
return 1;
}
else if (ssDCval >= 87) {
return 8;
}
}
}
else if (TSC >= 12) {
if (Rs < 1761) {
return 8;
}
else if (Rs >= 1761) {
return 1;
}
}
}
}
else if (Cs >= 1370) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (diffAFD < -4) {
if (Rs < 1967) {
return 1;
}
else if (Rs >= 1967) {
return 8;
}
}
else if (diffAFD >= -4) {
if (RsCsDiff < 205) {
return 8;
}
else if (RsCsDiff >= 205) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 1;
}
}
}
}
}
}
else if (TSC >= 13) {
if (Cs < 1535) {
return 1;
}
else if (Cs >= 1535) {
return 8;
}
}
}
else if (SC >= 4020) {
if (posBalance < 56) {
if (negBalance < 59) {
if (Rs < 3044) {
if (RsDiff < 12) {
return 2;
}
else if (RsDiff >= 12) {
if (Rs < 2882) {
return 8;
}
else if (Rs >= 2882) {
return 2;
}
}
}
else if (Rs >= 3044) {
return 8;
}
}
else if (negBalance >= 59) {
return 8;
}
}
else if (posBalance >= 56) {
if (SC < 4339) {
return 8;
}
else if (SC >= 4339) {
if (posBalance < 61) {
if (Rs < 3310) {
return 4;
}
else if (Rs >= 3310) {
return 8;
}
}
else if (posBalance >= 61) {
if (Rs < 3129) {
return 8;
}
else if (Rs >= 3129) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 1616) {
if (Rs < 2620) {
if (mcTcor < 97) {
if (negBalance < 41) {
if (posBalance < 58) {
if (AFD < 15) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 1;
}
}
else if (AFD >= 15) {
if (Rs < 2197) {
return 8;
}
else if (Rs >= 2197) {
return 2;
}
}
}
else if (posBalance >= 58) {
if (Rs < 1175) {
return 8;
}
else if (Rs >= 1175) {
return 4;
}
}
}
else if (negBalance >= 41) {
if (mcTcor < 79) {
if (Rs < 1893) {
return 1;
}
else if (Rs >= 1893) {
if (tcor < 59) {
return 4;
}
else if (tcor >= 59) {
return 8;
}
}
}
else if (mcTcor >= 79) {
if (SC < 3229) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (SC >= 3229) {
if (refDCval < 87) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (refDCval >= 87) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
if (AFD < 16) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (Rs >= 2620) {
if (refDCval < 104) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (Cs < 1795) {
return 2;
}
else if (Cs >= 1795) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
else if (refDCval >= 104) {
if (refDCval < 106) {
if (Rs < 3198) {
return 4;
}
else if (Rs >= 3198) {
return 8;
}
}
else if (refDCval >= 106) {
if (Cs < 1623) {
return 4;
}
else if (Cs >= 1623) {
if (MVDiff < 6) {
if (ssDCval < 124) {
return 4;
}
else if (ssDCval >= 124) {
return 8;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 19) {
if (TSC < 10) {
if (mu_mv_mag_sq < 16) {
if (Rs < 1765) {
if (mcTcor < 96) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (Rs >= 1765) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
if (negBalance < 62) {
return 1;
}
else if (negBalance >= 62) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (Rs < 2047) {
if (diffAFD < 1) {
if (SC < 2329) {
return 1;
}
else if (SC >= 2329) {
return 2;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (Rs >= 2047) {
return 2;
}
}
}
else if (TSC >= 10) {
if (RsDiff < 20) {
if (Rs < 1359) {
if (mu_mv_mag_sq < 23) {
if (posBalance < 67) {
return 2;
}
else if (posBalance >= 67) {
return 8;
}
}
else if (mu_mv_mag_sq >= 23) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
if (Cs < 1418) {
return 1;
}
else if (Cs >= 1418) {
return 4;
}
}
}
}
else if (Rs >= 1359) {
if (ssDCval < 126) {
if (refDCval < 92) {
if (Cs < 1676) {
if (Rs < 1613) {
return 1;
}
else if (Rs >= 1613) {
return 8;
}
}
else if (Cs >= 1676) {
if (diffRsCsdiff < 6) {
if (Rs < 2823) {
return 1;
}
else if (Rs >= 2823) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
if (Rs < 2158) {
return 1;
}
else if (Rs >= 2158) {
return 8;
}
}
}
}
else if (refDCval >= 92) {
if (tcor < 59) {
if (RsDiff < 16) {
if (Rs < 1959) {
return 8;
}
else if (Rs >= 1959) {
return 4;
}
}
else if (RsDiff >= 16) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
if (AFD < 28) {
if (Cs < 1665) {
return 2;
}
else if (Cs >= 1665) {
return 4;
}
}
else if (AFD >= 28) {
return 2;
}
}
}
}
else if (tcor >= 59) {
if (Cs < 1377) {
if (diffRsCsdiff < 28) {
if (Rs < 2751) {
return 1;
}
else if (Rs >= 2751) {
return 2;
}
}
else if (diffRsCsdiff >= 28) {
return 4;
}
}
else if (Cs >= 1377) {
if (mcTcor < 94) {
if (negBalance < 49) {
if (posBalance < 73) {
if (Rs < 2631) {
return 1;
}
else if (Rs >= 2631) {
return 2;
}
}
else if (posBalance >= 73) {
if (Rs < 2080) {
return 4;
}
else if (Rs >= 2080) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (MVDiff < 56) {
if (negBalance < 56) {
if (ssDCval < 109) {
if (Rs < 1880) {
return 4;
}
else if (Rs >= 1880) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (ssDCval >= 109) {
return 8;
}
}
else if (negBalance >= 56) {
return 8;
}
}
else if (MVDiff >= 56) {
if (RsCsDiff < 206) {
if (Rs < 2676) {
return 1;
}
else if (Rs >= 2676) {
return 4;
}
}
else if (RsCsDiff >= 206) {
if (Rs < 2099) {
return 4;
}
else if (Rs >= 2099) {
return 8;
}
}
}
}
}
else if (mcTcor >= 94) {
if (MVDiff < 46) {
if (MVDiff < 25) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 1973) {
return 1;
}
else if (Rs >= 1973) {
return 2;
}
}
}
else if (MVDiff >= 25) {
return 2;
}
}
else if (MVDiff >= 46) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 126) {
if (mu_mv_mag_sq < 5) {
if (mcTcor < 89) {
return 1;
}
else if (mcTcor >= 89) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (mu_mv_mag_sq < 27) {
if (MVDiff < 45) {
if (mcTcor < 85) {
if (Rs < 1808) {
if (Rs < 1505) {
return 4;
}
else if (Rs >= 1505) {
return 2;
}
}
else if (Rs >= 1808) {
return 1;
}
}
else if (mcTcor >= 85) {
if (diffRsCsdiff < 49) {
return 8;
}
else if (diffRsCsdiff >= 49) {
if (Rs < 3079) {
return 1;
}
else if (Rs >= 3079) {
return 8;
}
}
}
}
else if (MVDiff >= 45) {
return 8;
}
}
else if (mu_mv_mag_sq >= 27) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (Rs < 3127) {
return 2;
}
else if (Rs >= 3127) {
return 4;
}
}
}
}
}
}
}
else if (RsDiff >= 20) {
if (diffRsCsdiff < -98) {
if (AFD < 30) {
return 8;
}
else if (AFD >= 30) {
return 1;
}
}
else if (diffRsCsdiff >= -98) {
if (ssDCval < 115) {
return 2;
}
else if (ssDCval >= 115) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 3333) {
if (tcor < 86) {
if (RsDiff < 13) {
if (negBalance < 44) {
if (ssDCval < 89) {
if (mcTcor < 93) {
if (mu_mv_mag_sq < 60) {
if (Cs < 1678) {
return 2;
}
else if (Cs >= 1678) {
return 4;
}
}
else if (mu_mv_mag_sq >= 60) {
return 4;
}
}
else if (mcTcor >= 93) {
return 4;
}
}
else if (ssDCval >= 89) {
if (ssDCval < 133) {
if (Rs < 4638) {
return 8;
}
else if (Rs >= 4638) {
return 1;
}
}
else if (ssDCval >= 133) {
return 4;
}
}
}
else if (negBalance >= 44) {
if (MVDiff < 21) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 3818) {
return 2;
}
else if (Rs >= 3818) {
return 8;
}
}
}
else if (MVDiff >= 21) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
return 8;
}
}
}
}
}
else if (RsDiff >= 13) {
if (mu_mv_mag_sq < 11) {
if (refDCval < 101) {
if (Rs < 3410) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
if (Rs < 3388) {
return 4;
}
else if (Rs >= 3388) {
return 8;
}
}
}
else if (Rs >= 3410) {
return 2;
}
}
else if (refDCval >= 101) {
if (TSC < 14) {
if (RsCsDiff < 640) {
if (Rs < 5301) {
if (Cs < 1560) {
if (tcor < 84) {
if (diffMVdiffVal < -1) {
if (RsDiff < 21) {
return 8;
}
else if (RsDiff >= 21) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (TSC < 11) {
if (Cs < 1357) {
return 2;
}
else if (Cs >= 1357) {
return 8;
}
}
else if (TSC >= 11) {
return 2;
}
}
}
else if (tcor >= 84) {
return 2;
}
}
else if (Cs >= 1560) {
if (AFD < 22) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (refDCval < 103) {
return 2;
}
else if (refDCval >= 103) {
return 8;
}
}
}
else if (AFD >= 22) {
return 1;
}
}
}
else if (Rs >= 5301) {
if (ssDCval < 128) {
return 8;
}
else if (ssDCval >= 128) {
if (diffMVdiffVal < 4) {
if (SC < 7689) {
if (Rs < 5959) {
return 8;
}
else if (Rs >= 5959) {
return 2;
}
}
else if (SC >= 7689) {
return 8;
}
}
else if (diffMVdiffVal >= 4) {
if (Rs < 6328) {
return 8;
}
else if (Rs >= 6328) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 640) {
if (Rs < 7303) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
return 8;
}
}
else if (Rs >= 7303) {
return 8;
}
}
}
else if (TSC >= 14) {
if (CsDiff < 9) {
if (MVDiff < 64) {
return 2;
}
else if (MVDiff >= 64) {
return 4;
}
}
else if (CsDiff >= 9) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (RsDiff < 17) {
if (MVDiff < 10) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
if (Rs < 3674) {
return 2;
}
else if (Rs >= 3674) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (ssDCval < 139) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 1;
}
}
}
else if (ssDCval >= 139) {
if (Rs < 3959) {
return 2;
}
else if (Rs >= 3959) {
return 4;
}
}
}
}
}
else if (RsDiff >= 17) {
if (negBalance < 58) {
if (MVDiff < 15) {
if (Cs < 1899) {
if (Rs < 4790) {
return 2;
}
else if (Rs >= 4790) {
if (Rs < 5198) {
return 4;
}
else if (Rs >= 5198) {
return 2;
}
}
}
else if (Cs >= 1899) {
return 8;
}
}
else if (MVDiff >= 15) {
if (gchDC < 4) {
if (SC < 5113) {
if (Cs < 1537) {
return 2;
}
else if (Cs >= 1537) {
return 8;
}
}
else if (SC >= 5113) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
if (Cs < 1407) {
return 2;
}
else if (Cs >= 1407) {
if (tcor < 84) {
return 4;
}
else if (tcor >= 84) {
return 2;
}
}
}
}
}
else if (gchDC >= 4) {
if (tcor < 74) {
return 2;
}
else if (tcor >= 74) {
return 1;
}
}
}
}
else if (negBalance >= 58) {
if (MVDiff < 63) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 2;
}
}
else if (MVDiff >= 63) {
if (refDCval < 106) {
if (Rs < 3635) {
return 8;
}
else if (Rs >= 3635) {
return 4;
}
}
else if (refDCval >= 106) {
return 1;
}
}
}
}
}
}
}
}
else if (tcor >= 86) {
if (refDCval < 101) {
if (posBalance < 45) {
if (refDCval < 62) {
if (Rs < 3662) {
return 2;
}
else if (Rs >= 3662) {
return 4;
}
}
else if (refDCval >= 62) {
if (CsDiff < 8) {
if (negBalance < 41) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
if (ssDCval < 86) {
return 4;
}
else if (ssDCval >= 86) {
if (Rs < 4281) {
return 2;
}
else if (Rs >= 4281) {
return 8;
}
}
}
}
else if (negBalance >= 41) {
return 2;
}
}
else if (CsDiff >= 8) {
if (negBalance < 63) {
return 1;
}
else if (negBalance >= 63) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
}
else if (posBalance >= 45) {
if (refDCval < 59) {
return 8;
}
else if (refDCval >= 59) {
if (negBalance < 57) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
return 8;
}
}
}
else if (negBalance >= 57) {
if (Rs < 3530) {
return 8;
}
else if (Rs >= 3530) {
return 2;
}
}
}
}
}
else if (refDCval >= 101) {
if (mu_mv_mag_sq < 84) {
if (RsDiff < 12) {
if (mcTcor < 97) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 4551) {
return 1;
}
else if (Rs >= 4551) {
return 8;
}
}
}
else if (mcTcor >= 97) {
return 2;
}
}
else if (RsDiff >= 12) {
if (MVDiff < 17) {
if (posBalance < 60) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 4539) {
return 2;
}
else if (Rs >= 4539) {
return 1;
}
}
}
else if (posBalance >= 60) {
if (Rs < 4135) {
return 4;
}
else if (Rs >= 4135) {
return 2;
}
}
}
else if (MVDiff >= 17) {
if (refDCval < 129) {
if (ssDCval < 110) {
return 2;
}
else if (ssDCval >= 110) {
return 8;
}
}
else if (refDCval >= 129) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (Rs < 5088) {
return 1;
}
else if (Rs >= 5088) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1921) {
if (AFD < 12) {
if (ssDCval < 101) {
if (Rs < 4005) {
if (Cs < 2664) {
if (posBalance < 55) {
if (ssDCval < 85) {
if (posBalance < 47) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (diffAFD < -2) {
if (Rs < 2716) {
return 1;
}
else if (Rs >= 2716) {
return 8;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
}
else if (posBalance >= 47) {
if (AFD < 10) {
if (Rs < 2080) {
return 1;
}
else if (Rs >= 2080) {
return 4;
}
}
else if (AFD >= 10) {
if (mu_mv_mag_sq < 84) {
if (Cs < 2135) {
return 8;
}
else if (Cs >= 2135) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 84) {
return 8;
}
}
}
}
else if (ssDCval >= 85) {
if (Rs < 3773) {
return 8;
}
else if (Rs >= 3773) {
return 2;
}
}
}
else if (posBalance >= 55) {
if (diffTSC < 1) {
if (refDCval < 52) {
if (Rs < 2362) {
if (Rs < 1793) {
return 1;
}
else if (Rs >= 1793) {
return 4;
}
}
else if (Rs >= 2362) {
return 8;
}
}
else if (refDCval >= 52) {
if (negBalance < 42) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (Rs < 3400) {
return 4;
}
else if (Rs >= 3400) {
return 2;
}
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (MVDiff < 9) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (TSC < 10) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 3764) {
return 4;
}
else if (Rs >= 3764) {
return 8;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (MVDiff >= 9) {
return 8;
}
}
}
}
else if (Cs >= 2664) {
if (negBalance < 51) {
if (Rs < 1582) {
if (Rs < 1196) {
return 8;
}
else if (Rs >= 1196) {
return 1;
}
}
else if (Rs >= 1582) {
if (ssDCval < 99) {
if (TSC < 10) {
if (diffRsCsdiff < 29) {
if (SC < 7520) {
if (Cs < 2705) {
if (Rs < 3072) {
return 8;
}
else if (Rs >= 3072) {
return 4;
}
}
else if (Cs >= 2705) {
return 4;
}
}
else if (SC >= 7520) {
return 8;
}
}
else if (diffRsCsdiff >= 29) {
if (mu_mv_mag_sq < 107) {
if (mu_mv_mag_sq < 85) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (mu_mv_mag_sq >= 85) {
return 4;
}
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 3701) {
return 1;
}
else if (Rs >= 3701) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (Cs < 2729) {
return 8;
}
else if (Cs >= 2729) {
if (Cs < 4182) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (Cs >= 4182) {
return 8;
}
}
}
}
else if (ssDCval >= 99) {
return 2;
}
}
}
else if (negBalance >= 51) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (Rs >= 4005) {
if (tcor < 91) {
if (Rs < 4933) {
if (MVDiff < 7) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (TSC < 10) {
if (AFD < 11) {
if (Cs < 3930) {
return 4;
}
else if (Cs >= 3930) {
return 8;
}
}
else if (AFD >= 11) {
return 8;
}
}
else if (TSC >= 10) {
if (Rs < 4741) {
return 1;
}
else if (Rs >= 4741) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
if (SC < 6180) {
return 4;
}
else if (SC >= 6180) {
return 8;
}
}
}
else if (Rs >= 4933) {
if (ssDCval < 74) {
if (Cs < 3449) {
return 4;
}
else if (Cs >= 3449) {
return 8;
}
}
else if (ssDCval >= 74) {
return 2;
}
}
}
else if (tcor >= 91) {
if (MVDiff < 1) {
if (Rs < 4213) {
return 4;
}
else if (Rs >= 4213) {
return 2;
}
}
else if (MVDiff >= 1) {
if (ssDCval < 72) {
if (Rs < 5550) {
return 8;
}
else if (Rs >= 5550) {
return 4;
}
}
else if (ssDCval >= 72) {
if (RsDiff < 9) {
if (Rs < 5532) {
if (Cs < 2358) {
return 1;
}
else if (Cs >= 2358) {
return 8;
}
}
else if (Rs >= 5532) {
return 2;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 101) {
if (negBalance < 46) {
if (posBalance < 41) {
if (ssDCval < 113) {
if (SC < 9433) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
else if (SC >= 9433) {
return 1;
}
}
else if (ssDCval >= 113) {
if (mu_mv_mag_sq < 62) {
if (negBalance < 26) {
if (Rs < 5184) {
return 4;
}
else if (Rs >= 5184) {
return 8;
}
}
else if (negBalance >= 26) {
return 1;
}
}
else if (mu_mv_mag_sq >= 62) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (Cs < 3123) {
return 8;
}
else if (Cs >= 3123) {
if (Rs < 2727) {
return 1;
}
else if (Rs >= 2727) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 41) {
if (Rs < 6360) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 39) {
if (RsCsDiff < 172) {
if (Rs < 4457) {
return 8;
}
else if (Rs >= 4457) {
return 2;
}
}
else if (RsCsDiff >= 172) {
if (Cs < 3013) {
return 2;
}
else if (Cs >= 3013) {
if (Rs < 2214) {
return 1;
}
else if (Rs >= 2214) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (MVDiff < 26) {
if (diffAFD < -3) {
if (SC < 7285) {
return 8;
}
else if (SC >= 7285) {
return 2;
}
}
else if (diffAFD >= -3) {
if (mu_mv_mag_sq < 109) {
return 8;
}
else if (mu_mv_mag_sq >= 109) {
if (negBalance < 43) {
return 8;
}
else if (negBalance >= 43) {
if (Rs < 5040) {
return 8;
}
else if (Rs >= 5040) {
return 1;
}
}
}
}
}
else if (MVDiff >= 26) {
if (Cs < 2816) {
return 8;
}
else if (Cs >= 2816) {
return 2;
}
}
}
}
else if (negBalance >= 44) {
if (mcTcor < 93) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (ssDCval < 134) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (ssDCval >= 134) {
return 1;
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 106) {
if (Cs < 2740) {
return 8;
}
else if (Cs >= 2740) {
return 4;
}
}
else if (ssDCval >= 106) {
return 8;
}
}
}
}
else if (Rs >= 6360) {
if (SC < 10824) {
if (Rs < 6803) {
return 4;
}
else if (Rs >= 6803) {
return 2;
}
}
else if (SC >= 10824) {
if (Rs < 8517) {
return 8;
}
else if (Rs >= 8517) {
return 1;
}
}
}
}
}
else if (negBalance >= 46) {
if (tcor < 93) {
if (mu_mv_mag_sq < 83) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (diffTSC < 2) {
if (SC < 6982) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Rs < 4388) {
return 1;
}
else if (Rs >= 4388) {
return 8;
}
}
}
else if (SC >= 6982) {
return 1;
}
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 3739) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 8;
}
}
else if (Rs >= 3739) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (diffRsCsdiff < 31) {
return 1;
}
else if (diffRsCsdiff >= 31) {
if (diffAFD < -1) {
if (Rs < 4398) {
return 1;
}
else if (Rs >= 4398) {
return 8;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
}
}
else if (tcor >= 93) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
return 8;
}
}
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 76) {
if (RsDiff < 17) {
if (ssDCval < 101) {
if (tcor < 64) {
if (refDCval < 63) {
if (RsCsDiff < 267) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
return 2;
}
}
else if (RsCsDiff >= 267) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
if (negBalance < 49) {
if (Rs < 1015) {
return 4;
}
else if (Rs >= 1015) {
return 2;
}
}
else if (negBalance >= 49) {
return 4;
}
}
}
}
else if (refDCval >= 63) {
if (gchDC < 2) {
if (Rs < 2371) {
if (CsDiff < 15) {
if (negBalance < 54) {
if (Rs < 1549) {
return 4;
}
else if (Rs >= 1549) {
return 8;
}
}
else if (negBalance >= 54) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
else if (CsDiff >= 15) {
if (CsDiff < 17) {
return 2;
}
else if (CsDiff >= 17) {
if (Rs < 2036) {
return 8;
}
else if (Rs >= 2036) {
return 1;
}
}
}
}
else if (Rs >= 2371) {
if (TSC < 13) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (Cs < 3127) {
return 8;
}
else if (Cs >= 3127) {
return 4;
}
}
}
else if (TSC >= 13) {
if (tcor < 51) {
return 1;
}
else if (tcor >= 51) {
if (Rs < 3133) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
else if (Rs >= 3133) {
return 1;
}
}
}
}
}
else if (gchDC >= 2) {
if (SC < 4938) {
if (Rs < 1537) {
if (SC < 3460) {
if (Rs < 1219) {
return 4;
}
else if (Rs >= 1219) {
return 1;
}
}
else if (SC >= 3460) {
if (RsCsDiff < 517) {
return 2;
}
else if (RsCsDiff >= 517) {
return 8;
}
}
}
else if (Rs >= 1537) {
if (Cs < 1952) {
return 4;
}
else if (Cs >= 1952) {
if (TSC < 14) {
return 1;
}
else if (TSC >= 14) {
return 8;
}
}
}
}
else if (SC >= 4938) {
if (tcor < 63) {
if (tcor < 41) {
if (Rs < 2409) {
return 8;
}
else if (Rs >= 2409) {
return 1;
}
}
else if (tcor >= 41) {
return 8;
}
}
else if (tcor >= 63) {
if (CsDiff < 20) {
if (Rs < 1915) {
return 4;
}
else if (Rs >= 1915) {
return 2;
}
}
else if (CsDiff >= 20) {
return 8;
}
}
}
}
}
}
else if (tcor >= 64) {
if (negBalance < 47) {
if (posBalance < 46) {
if (ssDCval < 73) {
if (Rs < 3337) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
else if (Rs >= 3337) {
return 4;
}
}
else if (ssDCval >= 73) {
if (ssDCval < 91) {
if (Rs < 2371) {
return 8;
}
else if (Rs >= 2371) {
return 2;
}
}
else if (ssDCval >= 91) {
return 1;
}
}
}
else if (posBalance >= 46) {
if (RsCsDiff < 189) {
if (AFD < 18) {
if (RsDiff < 8) {
if (Rs < 2770) {
if (diffAFD < 1) {
if (MVDiff < 78) {
return 4;
}
else if (MVDiff >= 78) {
return 2;
}
}
else if (diffAFD >= 1) {
if (Rs < 900) {
return 4;
}
else if (Rs >= 900) {
return 8;
}
}
}
else if (Rs >= 2770) {
return 1;
}
}
else if (RsDiff >= 8) {
if (AFD < 14) {
if (diffAFD < 2) {
if (RsCsDiff < 172) {
return 8;
}
else if (RsCsDiff >= 172) {
if (Rs < 3879) {
return 4;
}
else if (Rs >= 3879) {
return 8;
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (AFD >= 14) {
if (mu_mv_mag_sq < 52) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
return 8;
}
}
else if (mu_mv_mag_sq >= 52) {
if (mu_mv_mag_sq < 54) {
if (Rs < 2680) {
return 1;
}
else if (Rs >= 2680) {
return 2;
}
}
else if (mu_mv_mag_sq >= 54) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Rs < 2943) {
return 8;
}
else if (Rs >= 2943) {
return 2;
}
}
}
}
}
}
}
else if (AFD >= 18) {
if (Rs < 1890) {
if (Rs < 1433) {
return 2;
}
else if (Rs >= 1433) {
return 4;
}
}
else if (Rs >= 1890) {
return 1;
}
}
}
else if (RsCsDiff >= 189) {
if (SC < 5114) {
if (ssDCval < 64) {
if (SC < 4775) {
return 4;
}
else if (SC >= 4775) {
return 2;
}
}
else if (ssDCval >= 64) {
if (RsCsDiff < 327) {
return 8;
}
else if (RsCsDiff >= 327) {
if (posBalance < 65) {
if (Rs < 2203) {
return 2;
}
else if (Rs >= 2203) {
return 8;
}
}
else if (posBalance >= 65) {
return 4;
}
}
}
}
else if (SC >= 5114) {
if (gchDC < 3) {
if (refDCval < 99) {
if (mu_mv_mag_sq < 52) {
if (refDCval < 94) {
if (negBalance < 44) {
if (Rs < 1362) {
return 1;
}
else if (Rs >= 1362) {
return 2;
}
}
else if (negBalance >= 44) {
if (tcor < 83) {
return 2;
}
else if (tcor >= 83) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
else if (refDCval >= 94) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (tcor < 77) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (tcor >= 77) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 3994) {
return 4;
}
else if (Rs >= 3994) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 99) {
if (Rs < 3591) {
return 2;
}
else if (Rs >= 3591) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (posBalance < 69) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
if (Rs < 3345) {
return 4;
}
else if (Rs >= 3345) {
return 8;
}
}
}
else if (posBalance >= 69) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 4544) {
if (ssDCval < 84) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 19) {
if (AFD < 16) {
if (tcor < 85) {
if (Rs < 2509) {
return 4;
}
else if (Rs >= 2509) {
return 2;
}
}
else if (tcor >= 85) {
if (Rs < 3195) {
return 1;
}
else if (Rs >= 3195) {
return 4;
}
}
}
else if (AFD >= 16) {
if (Rs < 3242) {
return 1;
}
else if (Rs >= 3242) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 19) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (Rs < 1862) {
if (mcTcor < 91) {
if (negBalance < 61) {
if (ssDCval < 59) {
if (Rs < 1462) {
return 2;
}
else if (Rs >= 1462) {
return 1;
}
}
else if (ssDCval >= 59) {
if (Cs < 3830) {
return 4;
}
else if (Cs >= 3830) {
return 8;
}
}
}
else if (negBalance >= 61) {
return 2;
}
}
else if (mcTcor >= 91) {
if (diffRsCsdiff < 16) {
if (diffRsCsdiff < -56) {
if (Rs < 1166) {
return 8;
}
else if (Rs >= 1166) {
return 1;
}
}
else if (diffRsCsdiff >= -56) {
return 2;
}
}
else if (diffRsCsdiff >= 16) {
return 8;
}
}
}
else if (Rs >= 1862) {
if (Rs < 2892) {
if (AFD < 18) {
if (mcTcor < 74) {
if (MVDiff < 11) {
if (Rs < 2533) {
return 8;
}
else if (Rs >= 2533) {
return 4;
}
}
else if (MVDiff >= 11) {
return 2;
}
}
else if (mcTcor >= 74) {
if (diffRsCsdiff < 60) {
if (refDCval < 84) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 2551) {
return 1;
}
else if (Rs >= 2551) {
return 4;
}
}
}
}
else if (refDCval >= 84) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
}
else if (diffRsCsdiff >= 60) {
if (mcTcor < 86) {
if (Cs < 2151) {
return 1;
}
else if (Cs >= 2151) {
return 8;
}
}
else if (mcTcor >= 86) {
if (Rs < 2284) {
return 8;
}
else if (Rs >= 2284) {
return 4;
}
}
}
}
}
else if (AFD >= 18) {
if (ssDCval < 73) {
if (MVDiff < 21) {
if (AFD < 23) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 1;
}
}
else if (AFD >= 23) {
return 4;
}
}
else if (MVDiff >= 21) {
if (tcor < 71) {
return 2;
}
else if (tcor >= 71) {
if (TSC < 12) {
if (Rs < 2488) {
return 1;
}
else if (Rs >= 2488) {
return 2;
}
}
else if (TSC >= 12) {
return 8;
}
}
}
}
else if (ssDCval >= 73) {
if (gchDC < 2) {
if (diffMVdiffVal < 3) {
if (mu_mv_mag_sq < 31) {
return 8;
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
else if (diffMVdiffVal >= 3) {
if (MVDiff < 20) {
return 1;
}
else if (MVDiff >= 20) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (Cs < 2187) {
if (Cs < 2086) {
return 4;
}
else if (Cs >= 2086) {
return 2;
}
}
else if (Cs >= 2187) {
if (Cs < 2398) {
return 8;
}
else if (Cs >= 2398) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2892) {
if (TSC < 13) {
if (Rs < 3732) {
if (MVDiff < 11) {
if (Cs < 2656) {
return 4;
}
else if (Cs >= 2656) {
if (SC < 6235) {
return 8;
}
else if (SC >= 6235) {
return 4;
}
}
}
else if (MVDiff >= 11) {
if (Rs < 3118) {
if (ssDCval < 74) {
if (Rs < 2911) {
return 8;
}
else if (Rs >= 2911) {
if (Rs < 2981) {
return 4;
}
else if (Rs >= 2981) {
return 2;
}
}
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (Rs >= 3118) {
if (tcor < 85) {
if (RsDiff < 16) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
if (AFD < 17) {
if (Cs < 2076) {
return 8;
}
else if (Cs >= 2076) {
return 4;
}
}
else if (AFD >= 17) {
return 8;
}
}
}
else if (RsDiff >= 16) {
return 8;
}
}
else if (tcor >= 85) {
return 8;
}
}
}
}
else if (Rs >= 3732) {
if (ssDCval < 70) {
if (mu_mv_mag_sq < 56) {
if (SC < 6653) {
return 4;
}
else if (SC >= 6653) {
if (Rs < 4379) {
return 2;
}
else if (Rs >= 4379) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 56) {
return 8;
}
}
else if (ssDCval >= 70) {
if (Rs < 3798) {
return 1;
}
else if (Rs >= 3798) {
return 2;
}
}
}
}
else if (TSC >= 13) {
if (refDCval < 77) {
if (Cs < 2782) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 3283) {
return 4;
}
else if (Rs >= 3283) {
return 1;
}
}
}
else if (Cs >= 2782) {
return 8;
}
}
else if (refDCval >= 77) {
if (SC < 7051) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
if (AFD < 21) {
if (Rs < 3970) {
return 4;
}
else if (Rs >= 3970) {
return 8;
}
}
else if (AFD >= 21) {
return 8;
}
}
}
else if (SC >= 7051) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (mcTcor < 94) {
if (AFD < 19) {
if (Cs < 2901) {
if (RsDiff < 11) {
if (MVDiff < 6) {
if (Cs < 2879) {
return 8;
}
else if (Cs >= 2879) {
if (Rs < 2950) {
return 4;
}
else if (Rs >= 2950) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (negBalance < 50) {
if (diffMVdiffVal < -10) {
return 2;
}
else if (diffMVdiffVal >= -10) {
return 8;
}
}
else if (negBalance >= 50) {
if (TSC < 12) {
if (SC < 4476) {
return 8;
}
else if (SC >= 4476) {
if (refDCval < 98) {
if (SC < 4815) {
return 4;
}
else if (SC >= 4815) {
if (Rs < 3139) {
return 8;
}
else if (Rs >= 3139) {
return 4;
}
}
}
else if (refDCval >= 98) {
return 8;
}
}
}
else if (TSC >= 12) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
return 8;
}
}
}
}
}
else if (RsDiff >= 11) {
if (SC < 6375) {
if (ssDCval < 87) {
return 4;
}
else if (ssDCval >= 87) {
if (ssDCval < 98) {
if (posBalance < 60) {
if (gchDC < 1) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (posBalance >= 60) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 4;
}
}
}
else if (ssDCval >= 98) {
if (Rs < 2622) {
return 8;
}
else if (Rs >= 2622) {
return 2;
}
}
}
}
else if (SC >= 6375) {
return 4;
}
}
}
else if (Cs >= 2901) {
if (refDCval < 87) {
if (Cs < 3290) {
return 4;
}
else if (Cs >= 3290) {
return 2;
}
}
else if (refDCval >= 87) {
if (ssDCval < 94) {
if (AFD < 15) {
if (Cs < 4200) {
return 4;
}
else if (Cs >= 4200) {
return 8;
}
}
else if (AFD >= 15) {
if (mcTcor < 81) {
return 4;
}
else if (mcTcor >= 81) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
return 8;
}
}
}
}
else if (ssDCval >= 94) {
if (SC < 5643) {
return 8;
}
else if (SC >= 5643) {
return 4;
}
}
}
}
}
else if (AFD >= 19) {
if (posBalance < 63) {
if (ssDCval < 91) {
if (Cs < 3610) {
if (TSC < 15) {
if (mcTcor < 90) {
if (RsCsDiff < 192) {
if (Rs < 2770) {
return 1;
}
else if (Rs >= 2770) {
return 8;
}
}
else if (RsCsDiff >= 192) {
return 4;
}
}
else if (mcTcor >= 90) {
if (negBalance < 57) {
if (Rs < 2373) {
return 2;
}
else if (Rs >= 2373) {
return 4;
}
}
else if (negBalance >= 57) {
return 8;
}
}
}
else if (TSC >= 15) {
return 1;
}
}
else if (Cs >= 3610) {
if (Rs < 1995) {
return 4;
}
else if (Rs >= 1995) {
return 8;
}
}
}
else if (ssDCval >= 91) {
if (Cs < 2437) {
if (Rs < 3262) {
if (SC < 4151) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 1;
}
}
else if (SC >= 4151) {
return 8;
}
}
else if (Rs >= 3262) {
return 4;
}
}
else if (Cs >= 2437) {
if (refDCval < 92) {
if (Rs < 3361) {
return 1;
}
else if (Rs >= 3361) {
return 4;
}
}
else if (refDCval >= 92) {
if (SC < 5055) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
return 8;
}
}
else if (SC >= 5055) {
return 4;
}
}
}
}
}
else if (posBalance >= 63) {
if (negBalance < 57) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
if (Rs < 2161) {
return 4;
}
else if (Rs >= 2161) {
return 8;
}
}
}
}
else if (negBalance >= 57) {
if (Rs < 2020) {
return 8;
}
else if (Rs >= 2020) {
return 4;
}
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 2995) {
if (Rs < 2492) {
if (diffMVdiffVal < -3) {
if (SC < 4327) {
if (Rs < 2179) {
return 2;
}
else if (Rs >= 2179) {
return 8;
}
}
else if (SC >= 4327) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
}
else if (Rs >= 2492) {
if (diffRsCsdiff < -9) {
if (Rs < 2791) {
return 2;
}
else if (Rs >= 2791) {
return 8;
}
}
else if (diffRsCsdiff >= -9) {
return 8;
}
}
}
else if (Rs >= 2995) {
if (Rs < 3315) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (Rs >= 3315) {
if (Rs < 3680) {
return 4;
}
else if (Rs >= 3680) {
if (CsDiff < 12) {
return 2;
}
else if (CsDiff >= 12) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 4544) {
if (RsDiff < 16) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (Rs < 4687) {
if (Rs < 4679) {
return 8;
}
else if (Rs >= 4679) {
return 4;
}
}
else if (Rs >= 4687) {
return 8;
}
}
}
else if (RsDiff >= 16) {
if (diffRsCsdiff < -115) {
return 2;
}
else if (diffRsCsdiff >= -115) {
if (Cs < 3046) {
return 2;
}
else if (Cs >= 3046) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 101) {
if (AFD < 20) {
if (mcTcor < 86) {
if (refDCval < 130) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
else if (refDCval >= 130) {
if (SC < 5880) {
if (tcor < 63) {
if (Cs < 3035) {
return 2;
}
else if (Cs >= 3035) {
return 1;
}
}
else if (tcor >= 63) {
return 8;
}
}
else if (SC >= 5880) {
if (mu_mv_mag_sq < 36) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 8;
}
}
else if (mu_mv_mag_sq >= 36) {
if (negBalance < 45) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
if (mcTcor < 79) {
return 8;
}
else if (mcTcor >= 79) {
if (Rs < 3953) {
return 1;
}
else if (Rs >= 3953) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 86) {
if (RsDiff < 10) {
if (Rs < 3372) {
if (ssDCval < 109) {
if (Rs < 2976) {
if (Rs < 1357) {
return 2;
}
else if (Rs >= 1357) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (MVDiff < 5) {
if (Rs < 2826) {
return 4;
}
else if (Rs >= 2826) {
return 8;
}
}
else if (MVDiff >= 5) {
if (Rs < 1958) {
return 4;
}
else if (Rs >= 1958) {
if (ssDCval < 107) {
return 8;
}
else if (ssDCval >= 107) {
return 4;
}
}
}
}
}
}
else if (Rs >= 2976) {
if (Cs < 3710) {
return 4;
}
else if (Cs >= 3710) {
return 8;
}
}
}
else if (ssDCval >= 109) {
if (mcTcor < 91) {
if (Rs < 3220) {
return 8;
}
else if (Rs >= 3220) {
return 1;
}
}
else if (mcTcor >= 91) {
if (Rs < 1594) {
return 1;
}
else if (Rs >= 1594) {
if (mcTcor < 96) {
if (diffMVdiffVal < -3) {
if (refDCval < 131) {
return 8;
}
else if (refDCval >= 131) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
if (mu_mv_mag_sq < 20) {
if (Cs < 2844) {
return 1;
}
else if (Cs >= 2844) {
return 2;
}
}
else if (mu_mv_mag_sq >= 20) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Cs < 1942) {
return 8;
}
else if (Cs >= 1942) {
return 1;
}
}
}
}
}
else if (mcTcor >= 96) {
return 8;
}
}
}
}
}
else if (Rs >= 3372) {
return 8;
}
}
else if (RsDiff >= 10) {
if (MVDiff < 4) {
if (Rs < 5076) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
return 4;
}
}
}
else if (Rs >= 5076) {
if (Rs < 7171) {
return 4;
}
else if (Rs >= 7171) {
return 2;
}
}
}
else if (MVDiff >= 4) {
if (refDCval < 154) {
if (negBalance < 46) {
if (MVDiff < 11) {
if (ssDCval < 108) {
return 2;
}
else if (ssDCval >= 108) {
if (AFD < 13) {
if (Rs < 4024) {
return 8;
}
else if (Rs >= 4024) {
return 2;
}
}
else if (AFD >= 13) {
if (ssDCval < 115) {
if (Rs < 4947) {
return 8;
}
else if (Rs >= 4947) {
return 4;
}
}
else if (ssDCval >= 115) {
return 4;
}
}
}
}
else if (MVDiff >= 11) {
if (RsCsDiff < 227) {
if (SC < 9199) {
return 8;
}
else if (SC >= 9199) {
return 4;
}
}
else if (RsCsDiff >= 227) {
if (mu_mv_mag_sq < 40) {
if (ssDCval < 121) {
return 2;
}
else if (ssDCval >= 121) {
if (Rs < 3566) {
return 4;
}
else if (Rs >= 3566) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (Cs < 3990) {
return 8;
}
else if (Cs >= 3990) {
if (Rs < 6802) {
return 2;
}
else if (Rs >= 6802) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (Cs < 2881) {
if (RsDiff < 14) {
if (ssDCval < 103) {
if (mu_mv_mag_sq < 58) {
if (refDCval < 101) {
return 2;
}
else if (refDCval >= 101) {
return 8;
}
}
else if (mu_mv_mag_sq >= 58) {
return 2;
}
}
else if (ssDCval >= 103) {
if (tcor < 80) {
if (Rs < 2835) {
return 8;
}
else if (Rs >= 2835) {
if (Cs < 2318) {
return 8;
}
else if (Cs >= 2318) {
return 4;
}
}
}
else if (tcor >= 80) {
if (SC < 6570) {
return 8;
}
else if (SC >= 6570) {
if (Rs < 4373) {
return 1;
}
else if (Rs >= 4373) {
if (Cs < 1975) {
return 1;
}
else if (Cs >= 1975) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 14) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
return 8;
}
}
}
else if (Cs >= 2881) {
if (Rs < 4500) {
if (posBalance < 60) {
if (Cs < 3577) {
return 4;
}
else if (Cs >= 3577) {
return 8;
}
}
else if (posBalance >= 60) {
return 8;
}
}
else if (Rs >= 4500) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
return 4;
}
}
}
}
}
else if (refDCval >= 154) {
if (TSC < 8) {
if (Cs < 3566) {
return 8;
}
else if (Cs >= 3566) {
return 1;
}
}
else if (TSC >= 8) {
return 4;
}
}
}
}
}
}
else if (AFD >= 20) {
if (negBalance < 61) {
if (Rs < 3634) {
if (refDCval < 115) {
if (CsDiff < 14) {
if (SC < 5766) {
if (MVDiff < 15) {
if (Cs < 2519) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
return 2;
}
}
else if (Cs >= 2519) {
return 4;
}
}
else if (MVDiff >= 15) {
if (SC < 3785) {
if (Rs < 1384) {
if (Rs < 961) {
return 2;
}
else if (Rs >= 961) {
return 1;
}
}
else if (Rs >= 1384) {
return 4;
}
}
else if (SC >= 3785) {
if (Rs < 3392) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
if (diffRsCsdiff < -81) {
if (Rs < 1940) {
return 8;
}
else if (Rs >= 1940) {
return 2;
}
}
else if (diffRsCsdiff >= -81) {
return 8;
}
}
}
else if (Rs >= 3392) {
return 1;
}
}
}
}
else if (SC >= 5766) {
return 4;
}
}
else if (CsDiff >= 14) {
if (SC < 6917) {
if (Cs < 2319) {
return 2;
}
else if (Cs >= 2319) {
return 8;
}
}
else if (SC >= 6917) {
return 2;
}
}
}
else if (refDCval >= 115) {
if (diffRsCsdiff < 38) {
if (tcor < 76) {
if (RsCsDiff < 206) {
if (Rs < 1002) {
return 1;
}
else if (Rs >= 1002) {
return 2;
}
}
else if (RsCsDiff >= 206) {
if (mcTcor < 91) {
if (Rs < 891) {
return 2;
}
else if (Rs >= 891) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 8;
}
}
}
else if (mcTcor >= 91) {
if (Cs < 2089) {
return 2;
}
else if (Cs >= 2089) {
return 1;
}
}
}
}
else if (tcor >= 76) {
return 8;
}
}
else if (diffRsCsdiff >= 38) {
return 8;
}
}
}
else if (Rs >= 3634) {
if (Rs < 5686) {
if (SC < 7143) {
if (negBalance < 55) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 8;
}
}
else if (negBalance >= 55) {
if (SC < 6744) {
return 8;
}
else if (SC >= 6744) {
return 4;
}
}
}
else if (SC >= 7143) {
if (AFD < 22) {
if (refDCval < 122) {
if (Rs < 4452) {
return 2;
}
else if (Rs >= 4452) {
return 1;
}
}
else if (refDCval >= 122) {
return 4;
}
}
else if (AFD >= 22) {
if (SC < 7220) {
return 4;
}
else if (SC >= 7220) {
return 1;
}
}
}
}
else if (Rs >= 5686) {
if (ssDCval < 115) {
return 4;
}
else if (ssDCval >= 115) {
return 8;
}
}
}
}
else if (negBalance >= 61) {
if (diffAFD < 2) {
if (RsCsDiff < 423) {
if (RsCsDiff < 203) {
if (CsDiff < 10) {
return 1;
}
else if (CsDiff >= 10) {
return 2;
}
}
else if (RsCsDiff >= 203) {
if (diffRsCsdiff < -48) {
if (Rs < 2574) {
return 1;
}
else if (Rs >= 2574) {
return 8;
}
}
else if (diffRsCsdiff >= -48) {
return 1;
}
}
}
else if (RsCsDiff >= 423) {
return 8;
}
}
else if (diffAFD >= 2) {
if (Cs < 3742) {
return 8;
}
else if (Cs >= 3742) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 17) {
if (TSC < 8) {
if (refDCval < 127) {
if (Rs < 7651) {
if (Cs < 2632) {
if (Rs < 1970) {
return 1;
}
else if (Rs >= 1970) {
return 4;
}
}
else if (Cs >= 2632) {
return 8;
}
}
else if (Rs >= 7651) {
return 1;
}
}
else if (refDCval >= 127) {
if (Cs < 2476) {
return 2;
}
else if (Cs >= 2476) {
return 8;
}
}
}
else if (TSC >= 8) {
if (Cs < 4431) {
if (TSC < 15) {
if (tcor < 80) {
if (RsDiff < 23) {
if (ssDCval < 140) {
if (mcTcor < 84) {
if (ssDCval < 63) {
if (Cs < 2424) {
return 2;
}
else if (Cs >= 2424) {
return 4;
}
}
else if (ssDCval >= 63) {
if (diffRsCsdiff < -125) {
if (Rs < 4648) {
return 4;
}
else if (Rs >= 4648) {
return 2;
}
}
else if (diffRsCsdiff >= -125) {
if (Rs < 2629) {
return 2;
}
else if (Rs >= 2629) {
return 8;
}
}
}
}
else if (mcTcor >= 84) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (Rs < 3085) {
if (negBalance < 56) {
if (CsDiff < 16) {
return 2;
}
else if (CsDiff >= 16) {
return 4;
}
}
else if (negBalance >= 56) {
if (Rs < 2878) {
return 4;
}
else if (Rs >= 2878) {
return 1;
}
}
}
else if (Rs >= 3085) {
if (ssDCval < 89) {
if (MVDiff < 12) {
if (mcTcor < 87) {
if (Cs < 2410) {
return 2;
}
else if (Cs >= 2410) {
return 8;
}
}
else if (mcTcor >= 87) {
return 2;
}
}
else if (MVDiff >= 12) {
if (mu_mv_mag_sq < 9) {
if (tcor < 73) {
return 2;
}
else if (tcor >= 73) {
return 8;
}
}
else if (mu_mv_mag_sq >= 9) {
if (CsDiff < 13) {
return 8;
}
else if (CsDiff >= 13) {
return 4;
}
}
}
}
else if (ssDCval >= 89) {
if (posBalance < 57) {
if (Cs < 3553) {
if (Cs < 2863) {
return 8;
}
else if (Cs >= 2863) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
}
else if (Cs >= 3553) {
return 2;
}
}
else if (posBalance >= 57) {
if (mu_mv_mag_sq < 31) {
return 4;
}
else if (mu_mv_mag_sq >= 31) {
if (RsDiff < 18) {
return 4;
}
else if (RsDiff >= 18) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (Rs < 3895) {
return 8;
}
else if (Rs >= 3895) {
return 4;
}
}
}
else if (RsDiff >= 23) {
if (tcor < 59) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
if (Rs < 4574) {
return 2;
}
else if (Rs >= 4574) {
return 8;
}
}
}
else if (tcor >= 59) {
if (SC < 6922) {
if (AFD < 29) {
return 2;
}
else if (AFD >= 29) {
return 1;
}
}
else if (SC >= 6922) {
if (AFD < 26) {
if (ssDCval < 132) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
return 2;
}
}
else if (ssDCval >= 132) {
return 2;
}
}
else if (AFD >= 26) {
if (Cs < 2043) {
return 4;
}
else if (Cs >= 2043) {
return 2;
}
}
}
}
}
}
else if (tcor >= 80) {
if (MVDiff < 23) {
if (negBalance < 45) {
if (Cs < 3341) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (Rs < 6510) {
return 8;
}
else if (Rs >= 6510) {
return 2;
}
}
}
else if (Cs >= 3341) {
return 8;
}
}
else if (negBalance >= 45) {
if (Rs < 5307) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (mu_mv_mag_sq < 58) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (mu_mv_mag_sq >= 58) {
return 4;
}
}
}
else if (Rs >= 5307) {
if (diffAFD < 3) {
if (Rs < 6611) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (Rs >= 6611) {
return 8;
}
}
else if (diffAFD >= 3) {
return 2;
}
}
}
}
else if (MVDiff >= 23) {
if (diffMVdiffVal < -27) {
if (RsDiff < 19) {
return 4;
}
else if (RsDiff >= 19) {
if (Rs < 5239) {
return 8;
}
else if (Rs >= 5239) {
return 2;
}
}
}
else if (diffMVdiffVal >= -27) {
if (RsCsDiff < 716) {
if (Cs < 3026) {
return 8;
}
else if (Cs >= 3026) {
return 1;
}
}
else if (RsCsDiff >= 716) {
if (Rs < 4605) {
return 1;
}
else if (Rs >= 4605) {
return 2;
}
}
}
}
}
}
else if (TSC >= 15) {
if (SC < 9090) {
if (diffAFD < 6) {
if (AFD < 31) {
if (negBalance < 49) {
if (RsDiff < 18) {
if (Rs < 4611) {
return 8;
}
else if (Rs >= 4611) {
return 2;
}
}
else if (RsDiff >= 18) {
if (refDCval < 84) {
return 4;
}
else if (refDCval >= 84) {
return 1;
}
}
}
else if (negBalance >= 49) {
if (diffRsCsdiff < 180) {
if (refDCval < 121) {
if (diffRsCsdiff < -134) {
if (Rs < 3412) {
return 8;
}
else if (Rs >= 3412) {
return 1;
}
}
else if (diffRsCsdiff >= -134) {
return 8;
}
}
else if (refDCval >= 121) {
if (gchDC < 2) {
if (Cs < 3077) {
return 8;
}
else if (Cs >= 3077) {
return 4;
}
}
else if (gchDC >= 2) {
if (ssDCval < 124) {
return 2;
}
else if (ssDCval >= 124) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 180) {
if (AFD < 27) {
if (CsDiff < 16) {
return 2;
}
else if (CsDiff >= 16) {
return 8;
}
}
else if (AFD >= 27) {
if (posBalance < 58) {
if (TSC < 20) {
return 2;
}
else if (TSC >= 20) {
return 4;
}
}
else if (posBalance >= 58) {
return 4;
}
}
}
}
}
else if (AFD >= 31) {
if (mu_mv_mag_sq < 6) {
if (Cs < 2184) {
return 1;
}
else if (Cs >= 2184) {
return 2;
}
}
else if (mu_mv_mag_sq >= 6) {
if (Cs < 2834) {
if (AFD < 42) {
return 1;
}
else if (AFD >= 42) {
return 8;
}
}
else if (Cs >= 2834) {
return 8;
}
}
}
}
else if (diffAFD >= 6) {
if (diffTSC < 5) {
if (Rs < 2752) {
return 8;
}
else if (Rs >= 2752) {
if (TSC < 16) {
return 8;
}
else if (TSC >= 16) {
return 4;
}
}
}
else if (diffTSC >= 5) {
if (Rs < 5365) {
if (diffRsCsdiff < 162) {
return 4;
}
else if (diffRsCsdiff >= 162) {
return 8;
}
}
else if (Rs >= 5365) {
return 4;
}
}
}
}
else if (SC >= 9090) {
if (negBalance < 47) {
if (ssDCval < 81) {
if (Rs < 6952) {
return 2;
}
else if (Rs >= 6952) {
return 1;
}
}
else if (ssDCval >= 81) {
if (RsDiff < 24) {
return 1;
}
else if (RsDiff >= 24) {
return 4;
}
}
}
else if (negBalance >= 47) {
if (Rs < 6053) {
if (Cs < 3482) {
return 8;
}
else if (Cs >= 3482) {
return 2;
}
}
else if (Rs >= 6053) {
if (mcTcor < 88) {
return 4;
}
else if (mcTcor >= 88) {
return 2;
}
}
}
}
}
}
else if (Cs >= 4431) {
if (mu_mv_mag_sq < 16) {
if (negBalance < 61) {
if (TSC < 12) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
if (Rs < 5906) {
return 4;
}
else if (Rs >= 5906) {
if (Rs < 8784) {
return 8;
}
else if (Rs >= 8784) {
return 1;
}
}
}
}
else if (TSC >= 12) {
if (SC < 10944) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
return 8;
}
}
else if (SC >= 10944) {
return 4;
}
}
}
else if (negBalance >= 61) {
if (negBalance < 66) {
return 8;
}
else if (negBalance >= 66) {
if (Rs < 4668) {
return 1;
}
else if (Rs >= 4668) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (AFD < 23) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 4;
}
}
else if (AFD >= 23) {
if (SC < 10377) {
return 2;
}
else if (SC >= 10377) {
if (Rs < 9545) {
return 8;
}
else if (Rs >= 9545) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (SC < 6864) {
if (Rs < 3989) {
if (negBalance < 55) {
if (ssDCval < 107) {
if (refDCval < 80) {
if (ssDCval < 67) {
if (RsDiff < 12) {
if (Cs < 2488) {
return 4;
}
else if (Cs >= 2488) {
if (Rs < 2662) {
return 4;
}
else if (Rs >= 2662) {
return 8;
}
}
}
else if (RsDiff >= 12) {
return 8;
}
}
else if (ssDCval >= 67) {
if (posBalance < 59) {
if (RsCsDiff < 183) {
return 8;
}
else if (RsCsDiff >= 183) {
if (SC < 6435) {
return 4;
}
else if (SC >= 6435) {
return 1;
}
}
}
else if (posBalance >= 59) {
return 8;
}
}
}
else if (refDCval >= 80) {
if (Cs < 2683) {
if (Rs < 3459) {
if (diffTSC < -3) {
return 2;
}
else if (diffTSC >= -3) {
return 8;
}
}
else if (Rs >= 3459) {
if (Cs < 2597) {
return 4;
}
else if (Cs >= 2597) {
return 8;
}
}
}
else if (Cs >= 2683) {
return 4;
}
}
}
else if (ssDCval >= 107) {
return 8;
}
}
else if (negBalance >= 55) {
if (Cs < 2552) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
if (Rs < 2832) {
return 8;
}
else if (Rs >= 2832) {
return 1;
}
}
}
else if (Cs >= 2552) {
return 4;
}
}
}
else if (Rs >= 3989) {
if (Cs < 2205) {
return 4;
}
else if (Cs >= 2205) {
return 1;
}
}
}
else if (SC >= 6864) {
if (negBalance < 52) {
if (Rs < 5300) {
if (mu_mv_mag_sq < 90) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
if (tcor < 73) {
if (Cs < 9564) {
return 1;
}
else if (Cs >= 9564) {
return 8;
}
}
else if (tcor >= 73) {
if (refDCval < 72) {
if (mu_mv_mag_sq < 79) {
return 8;
}
else if (mu_mv_mag_sq >= 79) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
return 4;
}
}
}
else if (refDCval >= 72) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
if (diffAFD < 5) {
if (TSC < 11) {
if (Rs < 4432) {
return 4;
}
else if (Rs >= 4432) {
return 8;
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (diffAFD >= 5) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (SC < 6965) {
return 1;
}
else if (SC >= 6965) {
if (CsDiff < 11) {
if (posBalance < 57) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (CsDiff >= 11) {
return 8;
}
}
}
}
else if (Rs >= 5300) {
if (negBalance < 38) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
if (Rs < 5502) {
return 8;
}
else if (Rs >= 5502) {
return 4;
}
}
}
else if (negBalance >= 38) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (TSC < 15) {
if (Rs < 7610) {
if (SC < 8483) {
if (Rs < 5833) {
return 8;
}
else if (Rs >= 5833) {
return 4;
}
}
else if (SC >= 8483) {
return 4;
}
}
else if (Rs >= 7610) {
return 8;
}
}
else if (TSC >= 15) {
if (Rs < 6588) {
return 2;
}
else if (Rs >= 6588) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 52) {
if (negBalance < 63) {
if (Rs < 3188) {
return 4;
}
else if (Rs >= 3188) {
if (ssDCval < 122) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
return 1;
}
}
else if (ssDCval >= 122) {
return 4;
}
}
}
else if (negBalance >= 63) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1651) {
if (Rs < 3128) {
if (Rs < 1454) {
if (refDCval < 122) {
if (negBalance < 1) {
if (Rs < 736) {
return 1;
}
else if (Rs >= 736) {
return 2;
}
}
else if (negBalance >= 1) {
if (refDCval < 62) {
if (AFD < 3) {
if (mu_mv_mag_sq < 128) {
if (Cs < 1393) {
if (ssDCval < 55) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 1165) {
if (Rs < 764) {
return 1;
}
else if (Rs >= 764) {
return 8;
}
}
else if (Rs >= 1165) {
return 4;
}
}
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (Cs >= 1393) {
if (Cs < 1649) {
if (negBalance < 24) {
return 1;
}
else if (negBalance >= 24) {
if (Rs < 1410) {
return 1;
}
else if (Rs >= 1410) {
return 4;
}
}
}
else if (Cs >= 1649) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1498) {
if (Cs < 1343) {
return 4;
}
else if (Cs >= 1343) {
return 8;
}
}
else if (Cs >= 1498) {
return 1;
}
}
}
else if (AFD >= 3) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (SC < 2582) {
return 8;
}
else if (SC >= 2582) {
return 4;
}
}
}
}
else if (refDCval >= 62) {
if (Cs < 1498) {
if (MVDiff < 2) {
if (Rs < 436) {
if (Cs < 1383) {
return 8;
}
else if (Cs >= 1383) {
return 1;
}
}
else if (Rs >= 436) {
if (Cs < 1434) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
if (ssDCval < 71) {
if (Rs < 1220) {
if (Rs < 1093) {
return 2;
}
else if (Rs >= 1093) {
return 8;
}
}
else if (Rs >= 1220) {
if (Rs < 1282) {
return 4;
}
else if (Rs >= 1282) {
return 1;
}
}
}
else if (ssDCval >= 71) {
if (Cs < 1313) {
return 4;
}
else if (Cs >= 1313) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
}
else if (Cs >= 1434) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (MVDiff >= 2) {
if (diffMVdiffVal < 1) {
if (mu_mv_mag_sq < 113) {
if (Rs < 694) {
return 4;
}
else if (Rs >= 694) {
return 8;
}
}
else if (mu_mv_mag_sq >= 113) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
else if (Cs >= 1498) {
if (AFD < 5) {
if (SC < 1940) {
return 1;
}
else if (SC >= 1940) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (AFD >= 5) {
if (Rs < 969) {
return 1;
}
else if (Rs >= 969) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 122) {
if (AFD < 2) {
if (diffAFD < -1) {
if (Cs < 1348) {
return 1;
}
else if (Cs >= 1348) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
return 4;
}
}
}
else if (diffAFD >= -1) {
if (mu_mv_mag_sq < 120) {
return 8;
}
else if (mu_mv_mag_sq >= 120) {
return 4;
}
}
}
else if (AFD >= 2) {
if (refDCval < 133) {
if (Rs < 1239) {
if (Rs < 964) {
return 8;
}
else if (Rs >= 964) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (ssDCval < 129) {
return 4;
}
else if (ssDCval >= 129) {
if (SC < 2660) {
return 8;
}
else if (SC >= 2660) {
return 4;
}
}
}
}
}
else if (Rs >= 1239) {
return 8;
}
}
else if (refDCval >= 133) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 719) {
return 2;
}
else if (Rs >= 719) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1454) {
if (ssDCval < 86) {
if (tcor < 97) {
if (posBalance < 44) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (posBalance < 38) {
if (Rs < 2097) {
return 1;
}
else if (Rs >= 2097) {
return 4;
}
}
else if (posBalance >= 38) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (posBalance >= 44) {
if (RsCsDiff < 4) {
if (Rs < 1560) {
return 8;
}
else if (Rs >= 1560) {
return 4;
}
}
else if (RsCsDiff >= 4) {
return 8;
}
}
}
else if (tcor >= 97) {
if (negBalance < 18) {
if (SC < 4469) {
if (Cs < 1350) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
if (Rs < 2297) {
return 2;
}
else if (Rs >= 2297) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 2342) {
if (Cs < 1319) {
return 2;
}
else if (Cs >= 1319) {
if (mu_mv_mag_sq < 127) {
if (Cs < 1323) {
return 2;
}
else if (Cs >= 1323) {
return 8;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 1329) {
return 8;
}
else if (Cs >= 1329) {
return 2;
}
}
}
}
else if (Rs >= 2342) {
return 2;
}
}
}
}
}
else if (Cs >= 1350) {
if (SC < 3319) {
if (Cs < 1441) {
return 8;
}
else if (Cs >= 1441) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
}
else if (SC >= 3319) {
return 8;
}
}
}
else if (SC >= 4469) {
if (Rs < 3090) {
return 1;
}
else if (Rs >= 3090) {
return 8;
}
}
}
else if (negBalance >= 18) {
if (posBalance < 39) {
if (mu_mv_mag_sq < 127) {
if (Rs < 2135) {
return 1;
}
else if (Rs >= 2135) {
return 2;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 1525) {
return 8;
}
else if (Rs >= 1525) {
if (Cs < 1318) {
return 4;
}
else if (Cs >= 1318) {
return 4;
}
}
}
}
else if (posBalance >= 39) {
if (mcTcor < 98) {
if (RsDiff < 4) {
if (negBalance < 21) {
return 4;
}
else if (negBalance >= 21) {
return 8;
}
}
else if (RsDiff >= 4) {
if (Rs < 2896) {
return 4;
}
else if (Rs >= 2896) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (posBalance < 57) {
if (negBalance < 22) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (negBalance >= 22) {
if (MVDiff < 2) {
if (Cs < 1614) {
if (Rs < 1514) {
if (Cs < 1471) {
return 4;
}
else if (Cs >= 1471) {
return 8;
}
}
else if (Rs >= 1514) {
return 4;
}
}
else if (Cs >= 1614) {
if (mu_mv_mag_sq < 128) {
if (Rs < 2463) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 4;
}
}
else if (Rs >= 2463) {
return 4;
}
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (RsCsDiff < 7) {
if (diffAFD < -19) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
if (ssDCval < 125) {
return 4;
}
else if (ssDCval >= 125) {
return 1;
}
}
}
else if (diffAFD >= -19) {
if (Rs < 3053) {
return 8;
}
else if (Rs >= 3053) {
if (Cs < 1466) {
return 2;
}
else if (Cs >= 1466) {
return 8;
}
}
}
}
else if (RsCsDiff >= 7) {
if (refDCval < 117) {
if (AFD < 11) {
if (diffAFD < -2) {
if (diffTSC < -2) {
if (Cs < 1374) {
return 2;
}
else if (Cs >= 1374) {
return 8;
}
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (refDCval >= 117) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (SC < 3520) {
return 1;
}
else if (SC >= 3520) {
return 8;
}
}
}
}
}
else if (ssDCval >= 140) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3128) {
if (ssDCval < 110) {
if (ssDCval < 88) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 44) {
if (Rs < 3578) {
if (Cs < 1482) {
return 1;
}
else if (Cs >= 1482) {
if (Cs < 1570) {
return 2;
}
else if (Cs >= 1570) {
return 4;
}
}
}
else if (Rs >= 3578) {
if (ssDCval < 64) {
if (negBalance < 21) {
if (Rs < 5521) {
if (mu_mv_mag_sq < 124) {
return 4;
}
else if (mu_mv_mag_sq >= 124) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (Rs >= 5521) {
return 8;
}
}
else if (negBalance >= 21) {
return 8;
}
}
else if (ssDCval >= 64) {
if (negBalance < 27) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Rs < 4411) {
return 4;
}
else if (Rs >= 4411) {
return 8;
}
}
}
else if (negBalance >= 27) {
return 4;
}
}
}
}
else if (posBalance >= 44) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1568) {
if (RsCsDiff < 12) {
if (SC < 5009) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (Rs < 3520) {
return 1;
}
else if (Rs >= 3520) {
return 8;
}
}
}
else if (SC >= 5009) {
return 8;
}
}
else if (RsCsDiff >= 12) {
if (Rs < 4348) {
return 4;
}
else if (Rs >= 4348) {
return 8;
}
}
}
else if (Cs >= 1568) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 2;
}
}
}
}
}
else if (ssDCval >= 88) {
if (SC < 5694) {
if (RsDiff < 22) {
return 4;
}
else if (RsDiff >= 22) {
return 2;
}
}
else if (SC >= 5694) {
if (refDCval < 97) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 2;
}
}
else if (refDCval >= 97) {
return 4;
}
}
}
}
else if (ssDCval >= 110) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (Rs < 3303) {
return 2;
}
else if (Rs >= 3303) {
if (Cs < 1551) {
return 4;
}
else if (Cs >= 1551) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1651) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (refDCval < 65) {
if (Rs < 8850) {
if (tcor < 92) {
if (Rs < 3319) {
return 4;
}
else if (Rs >= 3319) {
return 8;
}
}
else if (tcor >= 92) {
return 8;
}
}
else if (Rs >= 8850) {
return 2;
}
}
else if (refDCval >= 65) {
if (diffTSC < -1) {
if (Rs < 2385) {
return 4;
}
else if (Rs >= 2385) {
if (ssDCval < 73) {
if (diffAFD < -6) {
return 8;
}
else if (diffAFD >= -6) {
return 4;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (ssDCval < 87) {
if (Rs < 3477) {
return 8;
}
else if (Rs >= 3477) {
return 4;
}
}
else if (ssDCval >= 87) {
if (Rs < 2733) {
return 4;
}
else if (Rs >= 2733) {
return 8;
}
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 3319) {
if (negBalance < 19) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 3537) {
if (posBalance < 36) {
if (Rs < 1819) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
return 8;
}
}
else if (Rs >= 1819) {
if (Cs < 1824) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (Rs < 2687) {
return 8;
}
else if (Rs >= 2687) {
return 4;
}
}
}
else if (Cs >= 1824) {
return 1;
}
}
}
else if (posBalance >= 36) {
if (ssDCval < 66) {
if (Cs < 2070) {
if (mu_mv_mag_sq < 115) {
return 4;
}
else if (mu_mv_mag_sq >= 115) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (tcor < 100) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
if (Rs < 1539) {
return 1;
}
else if (Rs >= 1539) {
return 8;
}
}
}
else if (tcor >= 100) {
return 1;
}
}
}
}
else if (Cs >= 2070) {
return 4;
}
}
else if (ssDCval >= 66) {
return 1;
}
}
}
else if (Rs >= 3537) {
return 4;
}
}
}
else if (negBalance >= 19) {
if (Rs < 1663) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (negBalance < 26) {
if (Cs < 1728) {
return 8;
}
else if (Cs >= 1728) {
return 1;
}
}
else if (negBalance >= 26) {
return 8;
}
}
}
else if (Rs >= 1663) {
if (Cs < 1819) {
if (SC < 4093) {
return 8;
}
else if (SC >= 4093) {
if (ssDCval < 65) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (Rs < 2939) {
return 4;
}
else if (Rs >= 2939) {
return 4;
}
}
}
else if (ssDCval >= 65) {
return 1;
}
}
}
else if (Cs >= 1819) {
if (negBalance < 20) {
if (diffRsCsdiff < -3) {
if (Rs < 2412) {
return 8;
}
else if (Rs >= 2412) {
if (Rs < 2945) {
return 4;
}
else if (Rs >= 2945) {
return 8;
}
}
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
else if (negBalance >= 20) {
return 8;
}
}
}
}
}
else if (Cs >= 3319) {
return 8;
}
}
else if (ssDCval >= 67) {
if (mu_mv_mag_sq < 120) {
if (posBalance < 36) {
if (refDCval < 92) {
if (negBalance < 7) {
return 1;
}
else if (negBalance >= 7) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (refDCval >= 92) {
if (Rs < 2319) {
return 2;
}
else if (Rs >= 2319) {
if (Cs < 2105) {
if (ssDCval < 92) {
return 1;
}
else if (ssDCval >= 92) {
return 4;
}
}
else if (Cs >= 2105) {
if (posBalance < 23) {
return 2;
}
else if (posBalance >= 23) {
if (ssDCval < 93) {
return 4;
}
else if (ssDCval >= 93) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 36) {
return 8;
}
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 1104) {
if (Rs < 623) {
return 8;
}
else if (Rs >= 623) {
return 1;
}
}
else if (Rs >= 1104) {
if (mu_mv_mag_sq < 124) {
if (Cs < 2149) {
if (Rs < 3180) {
if (AFD < 1) {
if (Rs < 2339) {
return 4;
}
else if (Rs >= 2339) {
return 1;
}
}
else if (AFD >= 1) {
return 8;
}
}
else if (Rs >= 3180) {
return 2;
}
}
else if (Cs >= 2149) {
return 8;
}
}
else if (mu_mv_mag_sq >= 124) {
if (SC < 6961) {
if (negBalance < 35) {
if (diffRsCsdiff < -210) {
return 4;
}
else if (diffRsCsdiff >= -210) {
if (SC < 4456) {
if (SC < 4407) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
return 4;
}
}
else if (SC >= 4407) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 2640) {
return 4;
}
else if (Rs >= 2640) {
return 8;
}
}
}
}
else if (SC >= 4456) {
return 8;
}
}
}
else if (negBalance >= 35) {
if (Rs < 2152) {
return 4;
}
else if (Rs >= 2152) {
return 8;
}
}
}
else if (SC >= 6961) {
if (SC < 7786) {
if (SC < 7241) {
if (SC < 7008) {
return 4;
}
else if (SC >= 7008) {
return 8;
}
}
else if (SC >= 7241) {
return 4;
}
}
else if (SC >= 7786) {
if (diffRsCsdiff < -14) {
if (Rs < 4396) {
return 1;
}
else if (Rs >= 4396) {
if (Rs < 6283) {
return 8;
}
else if (Rs >= 6283) {
return 4;
}
}
}
else if (diffRsCsdiff >= -14) {
return 8;
}
}
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6807) {
if (SC < 8381) {
if (ssDCval < 81) {
if (posBalance < 50) {
if (ssDCval < 74) {
if (mcTcor < 98) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (SC < 4257) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (SC >= 4257) {
if (Rs < 2355) {
return 4;
}
else if (Rs >= 2355) {
if (Cs < 2755) {
if (Rs < 2908) {
return 8;
}
else if (Rs >= 2908) {
if (SC < 5443) {
if (CsDiff < 3) {
if (Cs < 1869) {
if (Cs < 1710) {
return 4;
}
else if (Cs >= 1710) {
return 4;
}
}
else if (Cs >= 1869) {
return 8;
}
}
else if (CsDiff >= 3) {
return 4;
}
}
else if (SC >= 5443) {
if (negBalance < 22) {
if (SC < 7802) {
if (refDCval < 53) {
return 4;
}
else if (refDCval >= 53) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (RsCsDiff < 43) {
if (Rs < 5104) {
if (Cs < 2042) {
if (RsCsDiff < 38) {
return 8;
}
else if (RsCsDiff >= 38) {
return 4;
}
}
else if (Cs >= 2042) {
if (Cs < 2092) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (Cs >= 2092) {
if (posBalance < 24) {
return 8;
}
else if (posBalance >= 24) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (diffRsCsdiff < 1) {
if (posBalance < 40) {
if (posBalance < 37) {
if (Cs < 2485) {
if (ssDCval < 54) {
if (posBalance < 27) {
return 4;
}
else if (posBalance >= 27) {
if (Rs < 4271) {
return 8;
}
else if (Rs >= 4271) {
if (refDCval < 54) {
return 4;
}
else if (refDCval >= 54) {
if (Rs < 4557) {
return 4;
}
else if (Rs >= 4557) {
return 4;
}
}
}
}
}
else if (ssDCval >= 54) {
if (Rs < 4438) {
return 8;
}
else if (Rs >= 4438) {
return 8;
}
}
}
else if (Cs >= 2485) {
if (Rs < 4762) {
return 4;
}
else if (Rs >= 4762) {
if (diffRsCsdiff < -9) {
return 4;
}
else if (diffRsCsdiff >= -9) {
return 4;
}
}
}
}
else if (posBalance >= 37) {
return 8;
}
}
else if (posBalance >= 40) {
return 4;
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 4638) {
if (posBalance < 32) {
return 4;
}
else if (posBalance >= 32) {
if (Rs < 4304) {
return 8;
}
else if (Rs >= 4304) {
return 4;
}
}
}
else if (Rs >= 4638) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 5104) {
return 8;
}
}
else if (RsCsDiff >= 43) {
if (Rs < 4347) {
return 4;
}
else if (Rs >= 4347) {
if (MVDiff < 2) {
if (Cs < 2370) {
if (Rs < 4473) {
return 4;
}
else if (Rs >= 4473) {
return 4;
}
}
else if (Cs >= 2370) {
return 4;
}
}
else if (MVDiff >= 2) {
if (AFD < 5) {
if (Rs < 4874) {
return 4;
}
else if (Rs >= 4874) {
return 4;
}
}
else if (AFD >= 5) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 7802) {
return 4;
}
}
else if (negBalance >= 22) {
if (Rs < 3757) {
return 8;
}
else if (Rs >= 3757) {
if (mu_mv_mag_sq < 112) {
return 8;
}
else if (mu_mv_mag_sq >= 112) {
if (diffRsCsdiff < 4) {
return 8;
}
else if (diffRsCsdiff >= 4) {
if (diffRsCsdiff < 18) {
return 8;
}
else if (diffRsCsdiff >= 18) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 2755) {
return 8;
}
}
}
}
}
else if (mcTcor >= 98) {
if (Rs < 5964) {
if (ssDCval < 66) {
if (SC < 4750) {
return 8;
}
else if (SC >= 4750) {
if (SC < 5335) {
if (Cs < 1887) {
if (Rs < 3118) {
if (tcor < 99) {
if (Rs < 2959) {
return 8;
}
else if (Rs >= 2959) {
return 4;
}
}
else if (tcor >= 99) {
return 8;
}
}
else if (Rs >= 3118) {
return 8;
}
}
else if (Cs >= 1887) {
return 4;
}
}
else if (SC >= 5335) {
if (ssDCval < 59) {
if (Cs < 2487) {
return 4;
}
else if (Cs >= 2487) {
if (Rs < 4218) {
return 8;
}
else if (Rs >= 4218) {
return 4;
}
}
}
else if (ssDCval >= 59) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (Rs >= 5964) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (Cs < 1853) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (ssDCval < 80) {
if (Rs < 3138) {
return 8;
}
else if (Rs >= 3138) {
return 2;
}
}
else if (ssDCval >= 80) {
return 4;
}
}
}
else if (Cs >= 1853) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (negBalance < 34) {
if (mu_mv_mag_sq < 124) {
if (RsDiff < 7) {
if (SC < 4200) {
return 8;
}
else if (SC >= 4200) {
if (Rs < 2398) {
return 4;
}
else if (Rs >= 2398) {
if (MVDiff < 5) {
if (Cs < 1945) {
if (Rs < 2589) {
return 8;
}
else if (Rs >= 2589) {
return 4;
}
}
else if (Cs >= 1945) {
if (ssDCval < 57) {
if (Rs < 3627) {
return 8;
}
else if (Rs >= 3627) {
if (Rs < 4317) {
return 4;
}
else if (Rs >= 4317) {
return 4;
}
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (negBalance >= 34) {
if (MVDiff < 5) {
if (diffRsCsdiff < 94) {
if (diffMVdiffVal < -4) {
if (posBalance < 61) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 2820) {
return 4;
}
else if (Rs >= 2820) {
return 8;
}
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
if (gchDC < 2) {
if (Rs < 2212) {
if (Cs < 2431) {
if (Cs < 2301) {
return 8;
}
else if (Cs >= 2301) {
if (Cs < 2342) {
return 4;
}
else if (Cs >= 2342) {
return 8;
}
}
}
else if (Cs >= 2431) {
return 4;
}
}
else if (Rs >= 2212) {
if (Cs < 3571) {
return 8;
}
else if (Cs >= 3571) {
if (Rs < 3321) {
return 4;
}
else if (Rs >= 3321) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (diffRsCsdiff >= 94) {
if (Rs < 3695) {
return 4;
}
else if (Rs >= 3695) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (Cs < 2189) {
return 4;
}
else if (Cs >= 2189) {
return 8;
}
}
}
}
}
else if (ssDCval >= 81) {
if (Cs < 2615) {
if (Rs < 3759) {
if (Cs < 1887) {
if (tcor < 97) {
if (Rs < 1107) {
return 1;
}
else if (Rs >= 1107) {
if (Rs < 3054) {
return 8;
}
else if (Rs >= 3054) {
return 2;
}
}
}
else if (tcor >= 97) {
if (MVDiff < 1) {
if (Rs < 1994) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 1336) {
return 1;
}
else if (Rs >= 1336) {
return 4;
}
}
}
else if (Rs >= 1994) {
if (posBalance < 39) {
if (Rs < 3438) {
return 8;
}
else if (Rs >= 3438) {
return 2;
}
}
else if (posBalance >= 39) {
return 4;
}
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (Cs >= 1887) {
if (tcor < 99) {
if (negBalance < 49) {
if (ssDCval < 83) {
if (TSC < 7) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 4;
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (ssDCval >= 83) {
if (Rs < 1581) {
if (Rs < 1549) {
return 8;
}
else if (Rs >= 1549) {
return 2;
}
}
else if (Rs >= 1581) {
if (Cs < 1976) {
if (Rs < 3062) {
return 8;
}
else if (Rs >= 3062) {
return 2;
}
}
else if (Cs >= 1976) {
return 8;
}
}
}
}
else if (negBalance >= 49) {
if (refDCval < 90) {
return 8;
}
else if (refDCval >= 90) {
return 4;
}
}
}
else if (tcor >= 99) {
if (ssDCval < 90) {
if (Rs < 3042) {
return 8;
}
else if (Rs >= 3042) {
if (Rs < 3730) {
return 2;
}
else if (Rs >= 3730) {
return 8;
}
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
else if (Rs >= 3759) {
if (RsCsDiff < 62) {
if (negBalance < 29) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (Rs < 3874) {
if (Cs < 2193) {
return 2;
}
else if (Cs >= 2193) {
return 8;
}
}
else if (Rs >= 3874) {
if (mu_mv_mag_sq < 118) {
if (SC < 6099) {
return 4;
}
else if (SC >= 6099) {
if (Rs < 5090) {
if (SC < 6507) {
return 2;
}
else if (SC >= 6507) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 2;
}
}
}
else if (Rs >= 5090) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 118) {
return 4;
}
}
}
}
else if (negBalance >= 29) {
if (ssDCval < 91) {
if (SC < 6378) {
return 8;
}
else if (SC >= 6378) {
return 2;
}
}
else if (ssDCval >= 91) {
return 8;
}
}
}
else if (RsCsDiff >= 62) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
else if (Cs >= 2615) {
if (Rs < 2557) {
if (Rs < 1391) {
return 2;
}
else if (Rs >= 1391) {
return 4;
}
}
else if (Rs >= 2557) {
if (Rs < 3162) {
if (refDCval < 99) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 4;
}
}
}
else if (refDCval >= 99) {
return 4;
}
}
else if (Rs >= 3162) {
if (refDCval < 98) {
if (ssDCval < 86) {
if (Cs < 3217) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (Cs >= 3217) {
return 8;
}
}
else if (ssDCval >= 86) {
if (Cs < 3651) {
if (mu_mv_mag_sq < 124) {
if (RsDiff < 11) {
if (Cs < 3592) {
return 4;
}
else if (Cs >= 3592) {
return 8;
}
}
else if (RsDiff >= 11) {
return 8;
}
}
else if (mu_mv_mag_sq >= 124) {
if (SC < 7192) {
if (ssDCval < 94) {
if (SC < 6426) {
return 8;
}
else if (SC >= 6426) {
if (Rs < 3379) {
return 8;
}
else if (Rs >= 3379) {
return 4;
}
}
}
else if (ssDCval >= 94) {
if (Cs < 2953) {
return 4;
}
else if (Cs >= 2953) {
return 8;
}
}
}
else if (SC >= 7192) {
return 4;
}
}
}
else if (Cs >= 3651) {
if (refDCval < 96) {
if (AFD < 6) {
if (Rs < 4120) {
return 4;
}
else if (Rs >= 4120) {
return 8;
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (refDCval >= 96) {
return 4;
}
}
}
}
else if (refDCval >= 98) {
if (SC < 5932) {
if (Cs < 2665) {
return 4;
}
else if (Cs >= 2665) {
return 8;
}
}
else if (SC >= 5932) {
if (mu_mv_mag_sq < 119) {
if (ssDCval < 101) {
return 8;
}
else if (ssDCval >= 101) {
return 4;
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 8381) {
if (refDCval < 75) {
if (Rs < 5948) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
if (refDCval < 72) {
if (diffRsCsdiff < -60) {
return 8;
}
else if (diffRsCsdiff >= -60) {
return 1;
}
}
else if (refDCval >= 72) {
return 8;
}
}
}
else if (Rs >= 5948) {
return 2;
}
}
else if (refDCval >= 75) {
if (Cs < 3639) {
if (Rs < 5191) {
return 4;
}
else if (Rs >= 5191) {
return 8;
}
}
else if (Cs >= 3639) {
if (mu_mv_mag_sq < 113) {
if (Rs < 6243) {
return 8;
}
else if (Rs >= 6243) {
if (Rs < 6485) {
return 2;
}
else if (Rs >= 6485) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
}
}
}
else if (Rs >= 6807) {
if (ssDCval < 68) {
if (Rs < 8744) {
return 8;
}
else if (Rs >= 8744) {
return 2;
}
}
else if (ssDCval >= 68) {
if (ssDCval < 95) {
return 4;
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (ssDCval < 129) {
if (Rs < 1893) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (tcor < 97) {
if (Rs < 1380) {
if (Rs < 999) {
if (Rs < 657) {
return 8;
}
else if (Rs >= 657) {
return 2;
}
}
else if (Rs >= 999) {
return 4;
}
}
else if (Rs >= 1380) {
return 8;
}
}
else if (tcor >= 97) {
if (Cs < 1816) {
return 8;
}
else if (Cs >= 1816) {
return 2;
}
}
}
}
else if (Rs >= 1893) {
if (refDCval < 105) {
if (negBalance < 38) {
if (Cs < 1731) {
return 4;
}
else if (Cs >= 1731) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (Cs < 2404) {
return 2;
}
else if (Cs >= 2404) {
return 8;
}
}
}
}
else if (negBalance >= 38) {
if (Cs < 2865) {
if (Rs < 4905) {
if (Rs < 3579) {
return 8;
}
else if (Rs >= 3579) {
return 1;
}
}
else if (Rs >= 4905) {
return 2;
}
}
else if (Cs >= 2865) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (diffAFD < -2) {
if (diffRsCsdiff < -921) {
return 1;
}
else if (diffRsCsdiff >= -921) {
if (ssDCval < 123) {
if (negBalance < 41) {
return 8;
}
else if (negBalance >= 41) {
if (Cs < 1766) {
if (Rs < 2945) {
return 1;
}
else if (Rs >= 2945) {
return 4;
}
}
else if (Cs >= 1766) {
if (MVDiff < 7) {
if (ssDCval < 115) {
return 8;
}
else if (ssDCval >= 115) {
if (Rs < 2899) {
return 8;
}
else if (Rs >= 2899) {
return 2;
}
}
}
else if (MVDiff >= 7) {
return 2;
}
}
}
}
else if (ssDCval >= 123) {
if (Cs < 2751) {
if (Rs < 2758) {
return 8;
}
else if (Rs >= 2758) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
}
else if (Cs >= 2751) {
if (Cs < 3368) {
return 4;
}
else if (Cs >= 3368) {
return 8;
}
}
}
}
}
else if (diffAFD >= -2) {
if (MVDiff < 1) {
if (Cs < 1680) {
if (Rs < 4147) {
return 8;
}
else if (Rs >= 4147) {
return 2;
}
}
else if (Cs >= 1680) {
if (mu_mv_mag_sq < 119) {
if (Rs < 6335) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
return 1;
}
}
else if (Rs >= 6335) {
if (Rs < 6825) {
return 4;
}
else if (Rs >= 6825) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (diffMVdiffVal < -1) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (Rs < 5349) {
return 2;
}
else if (Rs >= 5349) {
return 8;
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
}
else if (MVDiff >= 1) {
if (posBalance < 32) {
if (ssDCval < 115) {
return 8;
}
else if (ssDCval >= 115) {
if (posBalance < 15) {
if (Rs < 2774) {
return 8;
}
else if (Rs >= 2774) {
return 2;
}
}
else if (posBalance >= 15) {
return 4;
}
}
}
else if (posBalance >= 32) {
if (negBalance < 53) {
if (diffRsCsdiff < -28) {
if (RsDiff < 8) {
if (Rs < 4265) {
return 8;
}
else if (Rs >= 4265) {
return 4;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (diffRsCsdiff >= -28) {
return 8;
}
}
else if (negBalance >= 53) {
if (Rs < 4181) {
if (Rs < 3130) {
if (Rs < 2648) {
return 8;
}
else if (Rs >= 2648) {
return 1;
}
}
else if (Rs >= 3130) {
return 2;
}
}
else if (Rs >= 4181) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 129) {
if (mcTcor < 96) {
if (negBalance < 42) {
if (ssDCval < 200) {
if (negBalance < 38) {
if (diffAFD < 6) {
if (ssDCval < 130) {
if (Rs < 2478) {
return 1;
}
else if (Rs >= 2478) {
return 8;
}
}
else if (ssDCval >= 130) {
return 8;
}
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (negBalance >= 38) {
if (mu_mv_mag_sq < 122) {
return 8;
}
else if (mu_mv_mag_sq >= 122) {
return 1;
}
}
}
else if (ssDCval >= 200) {
return 4;
}
}
else if (negBalance >= 42) {
if (ssDCval < 137) {
if (Rs < 2552) {
return 8;
}
else if (Rs >= 2552) {
return 2;
}
}
else if (ssDCval >= 137) {
if (Rs < 5949) {
return 1;
}
else if (Rs >= 5949) {
if (Rs < 8158) {
return 8;
}
else if (Rs >= 8158) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 130) {
if (negBalance < 34) {
if (Rs < 1787) {
return 4;
}
else if (Rs >= 1787) {
return 1;
}
}
else if (negBalance >= 34) {
return 8;
}
}
else if (refDCval >= 130) {
if (ssDCval < 144) {
if (Rs < 4534) {
if (Rs < 2635) {
if (mu_mv_mag_sq < 113) {
if (Rs < 1753) {
return 4;
}
else if (Rs >= 1753) {
return 8;
}
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
else if (Rs >= 2635) {
if (Rs < 2636) {
return 1;
}
else if (Rs >= 2636) {
if (diffAFD < -3) {
if (Rs < 3801) {
if (Cs < 1716) {
return 1;
}
else if (Cs >= 1716) {
return 8;
}
}
else if (Rs >= 3801) {
return 1;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
}
}
else if (Rs >= 4534) {
if (Cs < 3419) {
if (Rs < 4896) {
return 1;
}
else if (Rs >= 4896) {
return 2;
}
}
else if (Cs >= 3419) {
return 8;
}
}
}
else if (ssDCval >= 144) {
if (ssDCval < 200) {
return 8;
}
else if (ssDCval >= 200) {
if (Cs < 4080) {
return 8;
}
else if (Cs >= 4080) {
return 4;
}
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect10(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1308) {
if (Rs < 2348) {
if (Cs < 533) {
if (negBalance < 14) {
if (mu_mv_mag_sq < 100) {
if (tcor < 94) {
if (negBalance < 8) {
if (gchDC < 12) {
if (posBalance < 93) {
if (RsCsDiff < 2) {
if (Rs < 6) {
if (AFD < 1) {
if (MVDiff < 22) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
return 8;
}
}
else if (MVDiff >= 22) {
return 8;
}
}
else if (AFD >= 1) {
if (mcTcor < 80) {
if (ssDCval < 24) {
return 8;
}
else if (ssDCval >= 24) {
return 4;
}
}
else if (mcTcor >= 80) {
return 2;
}
}
}
else if (Rs >= 6) {
if (AFD < 7) {
if (refDCval < 57) {
if (ssDCval < 27) {
if (Rs < 94) {
return 8;
}
else if (Rs >= 94) {
if (Rs < 201) {
return 2;
}
else if (Rs >= 201) {
return 1;
}
}
}
else if (ssDCval >= 27) {
return 8;
}
}
else if (refDCval >= 57) {
if (Rs < 128) {
return 2;
}
else if (Rs >= 128) {
return 8;
}
}
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (RsCsDiff >= 2) {
if (SC < 481) {
return 1;
}
else if (SC >= 481) {
return 2;
}
}
}
else if (posBalance >= 93) {
if (ssDCval < 31) {
if (diffRsCsdiff < -1) {
if (Rs < 464) {
return 1;
}
else if (Rs >= 464) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
if (mcTcor < 69) {
if (Rs < 9) {
return 8;
}
else if (Rs >= 9) {
return 2;
}
}
else if (mcTcor >= 69) {
return 4;
}
}
}
else if (ssDCval >= 31) {
if (SC < 17) {
return 2;
}
else if (SC >= 17) {
if (AFD < 6) {
if (tcor < 89) {
if (Rs < 436) {
return 8;
}
else if (Rs >= 436) {
return 4;
}
}
else if (tcor >= 89) {
return 4;
}
}
else if (AFD >= 6) {
if (mu_mv_mag_sq < 10) {
if (Cs < 71) {
return 8;
}
else if (Cs >= 71) {
return 2;
}
}
else if (mu_mv_mag_sq >= 10) {
return 8;
}
}
}
}
}
}
else if (gchDC >= 12) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
}
else if (negBalance >= 8) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (mu_mv_mag_sq < 38) {
if (mu_mv_mag_sq < 27) {
if (ssDCval < 109) {
if (ssDCval < 73) {
if (refDCval < 38) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
if (MVDiff < 490) {
return 2;
}
else if (MVDiff >= 490) {
if (Rs < 41) {
return 8;
}
else if (Rs >= 41) {
return 1;
}
}
}
}
else if (refDCval >= 38) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (Rs < 335) {
return 8;
}
else if (Rs >= 335) {
return 2;
}
}
}
}
else if (ssDCval >= 73) {
return 1;
}
}
else if (ssDCval >= 109) {
return 2;
}
}
else if (mu_mv_mag_sq >= 27) {
if (Rs < 521) {
return 2;
}
else if (Rs >= 521) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (refDCval < 24) {
return 1;
}
else if (refDCval >= 24) {
if (MVDiff < 3) {
if (ssDCval < 40) {
return 4;
}
else if (ssDCval >= 40) {
return 1;
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 50) {
if (tcor < 87) {
if (SC < 784) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (SC >= 784) {
if (Rs < 502) {
return 2;
}
else if (Rs >= 502) {
return 4;
}
}
}
else if (tcor >= 87) {
return 4;
}
}
else if (mu_mv_mag_sq >= 50) {
if (posBalance < 24) {
if (Rs < 397) {
return 4;
}
else if (Rs >= 397) {
return 8;
}
}
else if (posBalance >= 24) {
if (posBalance < 67) {
if (posBalance < 26) {
if (Rs < 164) {
return 8;
}
else if (Rs >= 164) {
return 1;
}
}
else if (posBalance >= 26) {
if (ssDCval < 25) {
if (Rs < 100) {
return 8;
}
else if (Rs >= 100) {
return 1;
}
}
else if (ssDCval >= 25) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Rs < 91) {
return 1;
}
else if (Rs >= 91) {
return 8;
}
}
}
}
}
else if (posBalance >= 67) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (tcor >= 94) {
if (SC < 918) {
if (Cs < 118) {
if (MVDiff < 12) {
if (ssDCval < 28) {
if (posBalance < 31) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
return 4;
}
}
}
else if (posBalance >= 31) {
return 1;
}
}
else if (ssDCval >= 28) {
return 8;
}
}
else if (MVDiff >= 12) {
if (ssDCval < 22) {
if (Rs < 15) {
return 8;
}
else if (Rs >= 15) {
return 1;
}
}
else if (ssDCval >= 22) {
if (negBalance < 11) {
if (AFD < 5) {
if (MVDiff < 16) {
if (ssDCval < 29) {
return 4;
}
else if (ssDCval >= 29) {
return 8;
}
}
else if (MVDiff >= 16) {
if (tcor < 95) {
if (Rs < 43) {
return 8;
}
else if (Rs >= 43) {
return 4;
}
}
else if (tcor >= 95) {
return 4;
}
}
}
else if (AFD >= 5) {
if (AFD < 20) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
return 8;
}
}
else if (AFD >= 20) {
return 2;
}
}
}
else if (negBalance >= 11) {
if (Cs < 7) {
return 4;
}
else if (Cs >= 7) {
if (Cs < 104) {
return 1;
}
else if (Cs >= 104) {
return 8;
}
}
}
}
}
}
else if (Cs >= 118) {
if (ssDCval < 109) {
if (mu_mv_mag_sq < 72) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 72) {
if (ssDCval < 88) {
if (Rs < 135) {
if (Cs < 185) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 1;
}
}
else if (Cs >= 185) {
return 1;
}
}
else if (Rs >= 135) {
if (ssDCval < 64) {
if (SC < 322) {
if (Rs < 170) {
return 1;
}
else if (Rs >= 170) {
return 4;
}
}
else if (SC >= 322) {
return 4;
}
}
else if (ssDCval >= 64) {
if (Rs < 347) {
return 1;
}
else if (Rs >= 347) {
return 8;
}
}
}
}
else if (ssDCval >= 88) {
return 4;
}
}
}
else if (ssDCval >= 109) {
if (ssDCval < 161) {
if (Rs < 320) {
if (Rs < 120) {
return 4;
}
else if (Rs >= 120) {
return 8;
}
}
else if (Rs >= 320) {
return 1;
}
}
else if (ssDCval >= 161) {
return 4;
}
}
}
}
else if (SC >= 918) {
if (ssDCval < 102) {
if (Cs < 51) {
return 8;
}
else if (Cs >= 51) {
if (Rs < 531) {
if (Rs < 454) {
return 4;
}
else if (Rs >= 454) {
return 8;
}
}
else if (Rs >= 531) {
return 1;
}
}
}
else if (ssDCval >= 102) {
if (TSC < 4) {
if (SC < 940) {
if (Cs < 397) {
return 1;
}
else if (Cs >= 397) {
return 8;
}
}
else if (SC >= 940) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 1;
}
}
}
else if (TSC >= 4) {
if (Rs < 1717) {
return 1;
}
else if (Rs >= 1717) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (SC < 779) {
if (refDCval < 48) {
if (Cs < 103) {
if (ssDCval < 40) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
if (ssDCval < 9) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 9) {
return 8;
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (Rs < 87) {
if (Cs < 5) {
return 4;
}
else if (Cs >= 5) {
if (refDCval < 25) {
if (Rs < 60) {
return 8;
}
else if (Rs >= 60) {
return 1;
}
}
else if (refDCval >= 25) {
if (negBalance < 6) {
return 8;
}
else if (negBalance >= 6) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
}
}
else if (Rs >= 87) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
}
}
else if (ssDCval >= 40) {
if (Rs < 8) {
return 2;
}
else if (Rs >= 8) {
return 8;
}
}
}
else if (Cs >= 103) {
if (Rs < 244) {
if (MVDiff < 2) {
if (refDCval < 36) {
if (Cs < 108) {
return 4;
}
else if (Cs >= 108) {
return 8;
}
}
else if (refDCval >= 36) {
if (tcor < 91) {
if (Rs < 241) {
return 4;
}
else if (Rs >= 241) {
return 1;
}
}
else if (tcor >= 91) {
if (posBalance < 30) {
if (ssDCval < 37) {
if (Cs < 231) {
return 4;
}
else if (Cs >= 231) {
return 1;
}
}
else if (ssDCval >= 37) {
return 1;
}
}
else if (posBalance >= 30) {
if (ssDCval < 38) {
if (Rs < 197) {
return 4;
}
else if (Rs >= 197) {
return 8;
}
}
else if (ssDCval >= 38) {
if (Rs < 98) {
return 8;
}
else if (Rs >= 98) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (negBalance < 9) {
if (ssDCval < 26) {
return 4;
}
else if (ssDCval >= 26) {
return 8;
}
}
else if (negBalance >= 9) {
if (Rs < 116) {
if (ssDCval < 27) {
return 1;
}
else if (ssDCval >= 27) {
return 8;
}
}
else if (Rs >= 116) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 4;
}
}
}
}
}
else if (Rs >= 244) {
if (posBalance < 12) {
if (tcor < 63) {
if (Rs < 437) {
return 4;
}
else if (Rs >= 437) {
return 8;
}
}
else if (tcor >= 63) {
if (mu_mv_mag_sq < 126) {
return 4;
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
}
else if (posBalance >= 12) {
if (TSCindex < 1) {
return 4;
}
else if (TSCindex >= 1) {
if (mu_mv_mag_sq < 127) {
if (Cs < 224) {
if (ssDCval < 45) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (ssDCval >= 45) {
return 4;
}
}
else if (Cs >= 224) {
if (ssDCval < 40) {
if (Rs < 414) {
return 4;
}
else if (Rs >= 414) {
return 1;
}
}
else if (ssDCval >= 40) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 48) {
if (ssDCval < 82) {
if (Cs < 22) {
return 2;
}
else if (Cs >= 22) {
if (ssDCval < 66) {
if (Cs < 283) {
if (MVDiff < 1) {
if (ssDCval < 61) {
if (negBalance < 6) {
if (Rs < 272) {
return 8;
}
else if (Rs >= 272) {
return 4;
}
}
else if (negBalance >= 6) {
return 8;
}
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (MVDiff >= 1) {
if (Rs < 113) {
return 8;
}
else if (Rs >= 113) {
if (SC < 282) {
if (negBalance < 10) {
if (Rs < 126) {
return 4;
}
else if (Rs >= 126) {
return 8;
}
}
else if (negBalance >= 10) {
return 4;
}
}
else if (SC >= 282) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 134) {
return 8;
}
else if (Rs >= 134) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 283) {
if (Rs < 252) {
return 4;
}
else if (Rs >= 252) {
if (Rs < 341) {
return 1;
}
else if (Rs >= 341) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
if (Cs < 120) {
if (Rs < 145) {
return 4;
}
else if (Rs >= 145) {
return 8;
}
}
else if (Cs >= 120) {
if (Cs < 238) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
return 8;
}
}
else if (Cs >= 238) {
if (negBalance < 11) {
if (Rs < 256) {
return 2;
}
else if (Rs >= 256) {
return 8;
}
}
else if (negBalance >= 11) {
if (Cs < 245) {
return 1;
}
else if (Cs >= 245) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (Rs < 181) {
return 4;
}
else if (Rs >= 181) {
if (ssDCval < 130) {
if (Cs < 381) {
if (Cs < 157) {
return 8;
}
else if (Cs >= 157) {
if (MVDiff < 39) {
return 2;
}
else if (MVDiff >= 39) {
return 8;
}
}
}
else if (Cs >= 381) {
return 4;
}
}
else if (ssDCval >= 130) {
if (negBalance < 9) {
if (Rs < 210) {
return 8;
}
else if (Rs >= 210) {
return 4;
}
}
else if (negBalance >= 9) {
if (Rs < 407) {
return 8;
}
else if (Rs >= 407) {
return 1;
}
}
}
}
}
}
}
}
else if (SC >= 779) {
if (diffTSC < -3) {
if (posBalance < 3) {
return 4;
}
else if (posBalance >= 3) {
if (MVDiff < 30) {
if (Rs < 924) {
if (ssDCval < 101) {
if (Rs < 601) {
return 1;
}
else if (Rs >= 601) {
return 8;
}
}
else if (ssDCval >= 101) {
return 2;
}
}
else if (Rs >= 924) {
return 1;
}
}
else if (MVDiff >= 30) {
if (diffAFD < -10) {
return 8;
}
else if (diffAFD >= -10) {
if (SC < 953) {
if (negBalance < 4) {
if (Cs < 360) {
return 2;
}
else if (Cs >= 360) {
return 4;
}
}
else if (negBalance >= 4) {
if (Rs < 550) {
return 2;
}
else if (Rs >= 550) {
return 8;
}
}
}
else if (SC >= 953) {
return 2;
}
}
}
}
}
else if (diffTSC >= -3) {
if (mu_mv_mag_sq < 128) {
if (Rs < 711) {
if (RsCsDiff < 1) {
if (ssDCval < 38) {
if (ssDCval < 24) {
return 8;
}
else if (ssDCval >= 24) {
return 4;
}
}
else if (ssDCval >= 38) {
if (Rs < 346) {
return 1;
}
else if (Rs >= 346) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
if (ssDCval < 141) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (Cs < 385) {
return 8;
}
else if (Cs >= 385) {
return 4;
}
}
}
else if (ssDCval >= 141) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (Rs < 495) {
return 1;
}
else if (Rs >= 495) {
if (ssDCval < 62) {
if (Cs < 325) {
if (Rs < 588) {
return 8;
}
else if (Rs >= 588) {
return 4;
}
}
else if (Cs >= 325) {
return 1;
}
}
else if (ssDCval >= 62) {
return 8;
}
}
}
}
else if (Rs >= 711) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Rs < 1314) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Rs < 990) {
if (Rs < 800) {
return 2;
}
else if (Rs >= 800) {
return 8;
}
}
else if (Rs >= 990) {
return 1;
}
}
}
else if (Rs >= 1314) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (Cs < 474) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Cs < 442) {
return 1;
}
else if (Cs >= 442) {
return 2;
}
}
}
else if (Cs >= 474) {
return 2;
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 137) {
if (refDCval < 60) {
return 8;
}
else if (refDCval >= 60) {
if (ssDCval < 83) {
if (Rs < 552) {
return 8;
}
else if (Rs >= 552) {
if (SC < 1209) {
if (Rs < 682) {
if (Cs < 313) {
return 2;
}
else if (Cs >= 313) {
return 4;
}
}
else if (Rs >= 682) {
return 1;
}
}
else if (SC >= 1209) {
if (Rs < 1161) {
return 8;
}
else if (Rs >= 1161) {
return 2;
}
}
}
}
else if (ssDCval >= 83) {
return 8;
}
}
}
else if (ssDCval >= 137) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (mcTcor < 82) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
if (tcor < 57) {
if (Cs < 389) {
if (negBalance < 45) {
if (gchDC < 1) {
if (ssDCval < 26) {
if (Cs < 314) {
return 1;
}
else if (Cs >= 314) {
return 8;
}
}
else if (ssDCval >= 26) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (posBalance < 80) {
if (diffRsCsdiff < -2) {
if (Rs < 226) {
return 2;
}
else if (Rs >= 226) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
else if (posBalance >= 80) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 521) {
return 8;
}
else if (Rs >= 521) {
return 1;
}
}
}
else if (Cs >= 389) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
return 8;
}
}
}
}
else if (tcor >= 57) {
if (posBalance < 38) {
if (Rs < 332) {
if (posBalance < 30) {
if (Rs < 22) {
return 4;
}
else if (Rs >= 22) {
return 2;
}
}
else if (posBalance >= 30) {
return 1;
}
}
else if (Rs >= 332) {
if (RsCsDiff < 19) {
return 4;
}
else if (RsCsDiff >= 19) {
return 8;
}
}
}
else if (posBalance >= 38) {
if (negBalance < 35) {
if (tcor < 70) {
if (ssDCval < 26) {
return 8;
}
else if (ssDCval >= 26) {
return 2;
}
}
else if (tcor >= 70) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 2;
}
}
}
else if (negBalance >= 35) {
return 1;
}
}
}
}
}
else if (mcTcor >= 82) {
if (posBalance < 25) {
if (posBalance < 15) {
if (Rs < 155) {
if (tcor < 110) {
return 8;
}
else if (tcor >= 110) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (ssDCval < 28) {
if (diffAFD < 2) {
if (gchDC < 2) {
if (Rs < 1) {
return 4;
}
else if (Rs >= 1) {
return 8;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (diffAFD >= 2) {
if (MVDiff < 680) {
return 4;
}
else if (MVDiff >= 680) {
return 8;
}
}
}
else if (ssDCval >= 28) {
if (posBalance < 5) {
return 8;
}
else if (posBalance >= 5) {
return 4;
}
}
}
}
}
else if (Rs >= 155) {
if (Rs < 217) {
return 1;
}
else if (Rs >= 217) {
return 4;
}
}
}
else if (posBalance >= 15) {
if (MVDiff < 7) {
if (tcor < 102) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 2;
}
}
else if (tcor >= 102) {
if (Rs < 44) {
return 4;
}
else if (Rs >= 44) {
if (Rs < 148) {
return 8;
}
else if (Rs >= 148) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 31) {
return 1;
}
else if (ssDCval >= 31) {
if (Rs < 86) {
return 1;
}
else if (Rs >= 86) {
return 8;
}
}
}
}
}
else if (posBalance >= 25) {
if (refDCval < 30) {
if (MVDiff < 55) {
if (Rs < 100) {
if (posBalance < 68) {
return 1;
}
else if (posBalance >= 68) {
return 2;
}
}
else if (Rs >= 100) {
if (mu_mv_mag_sq < 62) {
if (posBalance < 42) {
if (SC < 556) {
if (negBalance < 28) {
if (Cs < 31) {
if (Rs < 151) {
return 1;
}
else if (Rs >= 151) {
return 4;
}
}
else if (Cs >= 31) {
if (refDCval < 28) {
return 1;
}
else if (refDCval >= 28) {
if (mu_mv_mag_sq < 60) {
if (Cs < 254) {
return 1;
}
else if (Cs >= 254) {
return 4;
}
}
else if (mu_mv_mag_sq >= 60) {
if (Cs < 215) {
return 4;
}
else if (Cs >= 215) {
return 1;
}
}
}
}
}
else if (negBalance >= 28) {
if (tcor < 80) {
return 4;
}
else if (tcor >= 80) {
if (Rs < 123) {
return 4;
}
else if (Rs >= 123) {
return 1;
}
}
}
}
else if (SC >= 556) {
if (negBalance < 23) {
if (posBalance < 33) {
return 1;
}
else if (posBalance >= 33) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
if (MVDiff < 24) {
if (Cs < 391) {
return 1;
}
else if (Cs >= 391) {
return 4;
}
}
else if (MVDiff >= 24) {
return 4;
}
}
}
}
else if (negBalance >= 23) {
return 1;
}
}
}
else if (posBalance >= 42) {
if (negBalance < 23) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (Rs < 145) {
return 1;
}
else if (Rs >= 145) {
return 8;
}
}
}
else if (negBalance >= 23) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 62) {
if (mcTcor < 88) {
if (negBalance < 16) {
return 8;
}
else if (negBalance >= 16) {
return 1;
}
}
else if (mcTcor >= 88) {
return 1;
}
}
}
}
else if (MVDiff >= 55) {
if (diffAFD < -4) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 2;
}
}
else if (diffAFD >= -4) {
if (Rs < 47) {
return 1;
}
else if (Rs >= 47) {
return 8;
}
}
}
}
else if (refDCval >= 30) {
if (posBalance < 41) {
if (mu_mv_mag_sq < 17) {
if (MVDiff < 8) {
if (Rs < 72) {
return 1;
}
else if (Rs >= 72) {
if (ssDCval < 32) {
return 4;
}
else if (ssDCval >= 32) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (Rs < 123) {
return 8;
}
else if (Rs >= 123) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (mcTcor < 93) {
if (TSC < 3) {
if (AFD < 5) {
if (Rs < 76) {
return 1;
}
else if (Rs >= 76) {
if (tcor < 81) {
if (Rs < 399) {
return 4;
}
else if (Rs >= 399) {
return 8;
}
}
else if (tcor >= 81) {
return 8;
}
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (TSC >= 3) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 1;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 212) {
if (negBalance < 19) {
if (mu_mv_mag_sq < 78) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 78) {
return 1;
}
}
else if (negBalance >= 19) {
if (Cs < 35) {
if (Rs < 76) {
return 1;
}
else if (Rs >= 76) {
return 4;
}
}
else if (Cs >= 35) {
return 1;
}
}
}
else if (Rs >= 212) {
if (Cs < 257) {
if (Rs < 313) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 56) {
if (Rs < 235) {
return 4;
}
else if (Rs >= 235) {
return 1;
}
}
else if (mu_mv_mag_sq >= 56) {
return 4;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (Rs >= 313) {
if (Cs < 142) {
return 2;
}
else if (Cs >= 142) {
return 4;
}
}
}
else if (Cs >= 257) {
if (Rs < 296) {
return 4;
}
else if (Rs >= 296) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 41) {
if (posBalance < 67) {
if (tcor < 71) {
if (posBalance < 60) {
if (ssDCval < 34) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (ssDCval >= 34) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (posBalance >= 60) {
if (Rs < 575) {
return 8;
}
else if (Rs >= 575) {
return 1;
}
}
}
else if (tcor >= 71) {
if (CsDiff < 3) {
if (SC < 454) {
if (mcTcor < 88) {
if (TSC < 3) {
if (Rs < 166) {
return 1;
}
else if (Rs >= 166) {
return 8;
}
}
else if (TSC >= 3) {
return 2;
}
}
else if (mcTcor >= 88) {
if (tcor < 72) {
return 2;
}
else if (tcor >= 72) {
return 1;
}
}
}
else if (SC >= 454) {
return 1;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (posBalance >= 67) {
if (posBalance < 74) {
if (Rs < 62) {
return 4;
}
else if (Rs >= 62) {
return 8;
}
}
else if (posBalance >= 74) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 300) {
if (ssDCval < 51) {
if (mcTcor < 92) {
if (tcor < 68) {
if (posBalance < 50) {
if (RsCsDiff < 1) {
if (Rs < 35) {
return 1;
}
else if (Rs >= 35) {
return 4;
}
}
else if (RsCsDiff >= 1) {
if (negBalance < 62) {
return 1;
}
else if (negBalance >= 62) {
if (Cs < 164) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (tcor < 35) {
if (Rs < 175) {
return 8;
}
else if (Rs >= 175) {
return 1;
}
}
else if (tcor >= 35) {
return 2;
}
}
}
else if (Cs >= 164) {
if (Rs < 188) {
return 8;
}
else if (Rs >= 188) {
return 2;
}
}
}
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 3) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
if (diffAFD < -8) {
return 1;
}
else if (diffAFD >= -8) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (RsCsDiff < 9) {
if (negBalance < 50) {
if (refDCval < 39) {
if (tcor < 57) {
if (diffMVdiffVal < -9) {
return 8;
}
else if (diffMVdiffVal >= -9) {
if (Cs < 192) {
return 1;
}
else if (Cs >= 192) {
return 4;
}
}
}
else if (tcor >= 57) {
if (Cs < 319) {
return 8;
}
else if (Cs >= 319) {
return 1;
}
}
}
else if (refDCval >= 39) {
if (Rs < 30) {
if (Rs < 14) {
return 8;
}
else if (Rs >= 14) {
return 2;
}
}
else if (Rs >= 30) {
if (diffTSC < 2) {
return 4;
}
else if (diffTSC >= 2) {
if (Rs < 133) {
return 1;
}
else if (Rs >= 133) {
return 8;
}
}
}
}
}
else if (negBalance >= 50) {
if (AFD < 22) {
if (Cs < 272) {
if (refDCval < 44) {
return 8;
}
else if (refDCval >= 44) {
return 2;
}
}
else if (Cs >= 272) {
return 4;
}
}
else if (AFD >= 22) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 4;
}
}
}
}
else if (RsCsDiff >= 9) {
if (Rs < 286) {
if (Rs < 139) {
return 8;
}
else if (Rs >= 139) {
if (posBalance < 71) {
return 1;
}
else if (posBalance >= 71) {
return 8;
}
}
}
else if (Rs >= 286) {
if (Rs < 288) {
return 4;
}
else if (Rs >= 288) {
return 8;
}
}
}
}
}
}
else if (tcor >= 68) {
if (negBalance < 31) {
if (SC < 238) {
return 4;
}
else if (SC >= 238) {
if (posBalance < 62) {
if (Rs < 277) {
return 1;
}
else if (Rs >= 277) {
return 8;
}
}
else if (posBalance >= 62) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (Rs < 196) {
return 2;
}
else if (Rs >= 196) {
return 8;
}
}
}
}
}
else if (negBalance >= 31) {
if (MVDiff < 31) {
if (AFD < 4) {
if (Rs < 85) {
return 1;
}
else if (Rs >= 85) {
return 8;
}
}
else if (AFD >= 4) {
if (mcTcor < 87) {
if (negBalance < 35) {
return 2;
}
else if (negBalance >= 35) {
if (Rs < 183) {
if (Cs < 174) {
return 8;
}
else if (Cs >= 174) {
if (Cs < 220) {
return 2;
}
else if (Cs >= 220) {
return 8;
}
}
}
else if (Rs >= 183) {
return 8;
}
}
}
else if (mcTcor >= 87) {
if (TSC < 4) {
if (negBalance < 41) {
if (tcor < 87) {
if (Cs < 243) {
return 1;
}
else if (Cs >= 243) {
return 2;
}
}
else if (tcor >= 87) {
return 8;
}
}
else if (negBalance >= 41) {
if (tcor < 75) {
return 1;
}
else if (tcor >= 75) {
return 2;
}
}
}
else if (TSC >= 4) {
if (refDCval < 37) {
return 8;
}
else if (refDCval >= 37) {
if (mcTcor < 89) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (mcTcor >= 89) {
if (Cs < 245) {
return 1;
}
else if (Cs >= 245) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 31) {
if (Rs < 124) {
if (ssDCval < 35) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
if (Rs < 27) {
return 8;
}
else if (Rs >= 27) {
return 2;
}
}
}
else if (ssDCval >= 35) {
if (AFD < 17) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 2;
}
}
else if (AFD >= 17) {
return 4;
}
}
}
else if (Rs >= 124) {
if (SC < 440) {
if (SC < 271) {
return 1;
}
else if (SC >= 271) {
if (Rs < 275) {
return 2;
}
else if (Rs >= 275) {
return 1;
}
}
}
else if (SC >= 440) {
if (Cs < 280) {
return 8;
}
else if (Cs >= 280) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 61) {
if (negBalance < 75) {
if (posBalance < 68) {
if (tcor < 95) {
if (posBalance < 56) {
if (Cs < 43) {
if (MVDiff < 45) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 120) {
return 1;
}
else if (Rs >= 120) {
return 2;
}
}
}
else if (MVDiff >= 45) {
if (negBalance < 41) {
return 2;
}
else if (negBalance >= 41) {
if (Rs < 20) {
return 8;
}
else if (Rs >= 20) {
return 1;
}
}
}
}
else if (Cs >= 43) {
if (Rs < 171) {
if (AFD < 5) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (AFD >= 5) {
if (SC < 292) {
if (diffRsCsdiff < -1) {
return 1;
}
else if (diffRsCsdiff >= -1) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
if (Rs < 82) {
return 1;
}
else if (Rs >= 82) {
return 2;
}
}
}
}
else if (SC >= 292) {
if (CsDiff < 4) {
if (ssDCval < 37) {
return 8;
}
else if (ssDCval >= 37) {
return 1;
}
}
else if (CsDiff >= 4) {
return 4;
}
}
}
}
else if (Rs >= 171) {
if (refDCval < 45) {
if (mu_mv_mag_sq < 37) {
if (AFD < 7) {
if (mcTcor < 95) {
return 2;
}
else if (mcTcor >= 95) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (AFD >= 7) {
return 1;
}
}
else if (mu_mv_mag_sq >= 37) {
if (RsDiff < 1) {
if (Rs < 216) {
return 1;
}
else if (Rs >= 216) {
return 2;
}
}
else if (RsDiff >= 1) {
if (Cs < 186) {
return 4;
}
else if (Cs >= 186) {
if (Rs < 280) {
return 8;
}
else if (Rs >= 280) {
return 4;
}
}
}
}
}
else if (refDCval >= 45) {
if (RsCsDiff < 7) {
if (SC < 642) {
return 4;
}
else if (SC >= 642) {
if (Rs < 279) {
return 4;
}
else if (Rs >= 279) {
return 1;
}
}
}
else if (RsCsDiff >= 7) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 56) {
if (SC < 146) {
return 4;
}
else if (SC >= 146) {
if (mcTcor < 93) {
if (Rs < 187) {
return 2;
}
else if (Rs >= 187) {
return 1;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
}
}
else if (tcor >= 95) {
if (posBalance < 38) {
if (Rs < 201) {
if (Rs < 22) {
return 2;
}
else if (Rs >= 22) {
return 1;
}
}
else if (Rs >= 201) {
if (Rs < 239) {
return 4;
}
else if (Rs >= 239) {
return 2;
}
}
}
else if (posBalance >= 38) {
if (diffRsCsdiff < -2) {
if (Rs < 199) {
if (Cs < 144) {
return 2;
}
else if (Cs >= 144) {
return 1;
}
}
else if (Rs >= 199) {
return 2;
}
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
}
else if (posBalance >= 68) {
if (Rs < 279) {
return 2;
}
else if (Rs >= 279) {
return 1;
}
}
}
else if (negBalance >= 75) {
if (posBalance < 2) {
if (refDCval < 50) {
return 4;
}
else if (refDCval >= 50) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (posBalance >= 2) {
if (Cs < 87) {
return 8;
}
else if (Cs >= 87) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (Cs < 69) {
if (refDCval < 40) {
if (Rs < 189) {
return 4;
}
else if (Rs >= 189) {
return 2;
}
}
else if (refDCval >= 40) {
if (Cs < 23) {
if (tcor < 51) {
if (Rs < 114) {
return 4;
}
else if (Rs >= 114) {
return 1;
}
}
else if (tcor >= 51) {
if (MVDiff < 30) {
if (Rs < 9) {
return 2;
}
else if (Rs >= 9) {
return 8;
}
}
else if (MVDiff >= 30) {
return 2;
}
}
}
else if (Cs >= 23) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (Rs < 132) {
return 4;
}
else if (Rs >= 132) {
return 8;
}
}
}
}
}
else if (Cs >= 69) {
if (SC < 559) {
if (mu_mv_mag_sq < 102) {
if (negBalance < 27) {
if (Cs < 241) {
if (Rs < 132) {
return 1;
}
else if (Rs >= 132) {
if (refDCval < 39) {
return 4;
}
else if (refDCval >= 39) {
if (diffMVdiffVal < 2) {
return 1;
}
else if (diffMVdiffVal >= 2) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
}
else if (Cs >= 241) {
if (Rs < 193) {
return 1;
}
else if (Rs >= 193) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
else if (negBalance >= 27) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (RsDiff < 1) {
if (MVDiff < 10) {
if (Cs < 77) {
return 1;
}
else if (Cs >= 77) {
if (diffMVdiffVal < 4) {
if (ssDCval < 34) {
return 1;
}
else if (ssDCval >= 34) {
if (negBalance < 77) {
if (mu_mv_mag_sq < 71) {
if (Rs < 194) {
return 1;
}
else if (Rs >= 194) {
return 8;
}
}
else if (mu_mv_mag_sq >= 71) {
return 8;
}
}
else if (negBalance >= 77) {
return 4;
}
}
}
else if (diffMVdiffVal >= 4) {
return 1;
}
}
}
else if (MVDiff >= 10) {
if (Rs < 276) {
return 8;
}
else if (Rs >= 276) {
return 4;
}
}
}
else if (RsDiff >= 1) {
if (diffAFD < 1) {
if (Cs < 141) {
return 1;
}
else if (Cs >= 141) {
if (Cs < 178) {
return 8;
}
else if (Cs >= 178) {
if (Rs < 272) {
return 1;
}
else if (Rs >= 272) {
return 8;
}
}
}
}
else if (diffAFD >= 1) {
if (Rs < 279) {
return 4;
}
else if (Rs >= 279) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (negBalance < 19) {
if (Rs < 249) {
if (mu_mv_mag_sq < 124) {
if (mu_mv_mag_sq < 107) {
if (Rs < 155) {
return 4;
}
else if (Rs >= 155) {
return 8;
}
}
else if (mu_mv_mag_sq >= 107) {
if (ssDCval < 43) {
if (Cs < 205) {
return 1;
}
else if (Cs >= 205) {
if (Rs < 201) {
return 8;
}
else if (Rs >= 201) {
return 1;
}
}
}
else if (ssDCval >= 43) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (Rs >= 249) {
return 4;
}
}
else if (negBalance >= 19) {
if (negBalance < 59) {
if (posBalance < 21) {
if (posBalance < 14) {
return 8;
}
else if (posBalance >= 14) {
return 4;
}
}
else if (posBalance >= 21) {
return 8;
}
}
else if (negBalance >= 59) {
return 4;
}
}
}
}
else if (SC >= 559) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 51) {
if (ssDCval < 70) {
if (TSCindex < 2) {
if (refDCval < 63) {
if (posBalance < 31) {
if (Rs < 23) {
return 2;
}
else if (Rs >= 23) {
return 8;
}
}
else if (posBalance >= 31) {
if (MVDiff < 7) {
if (TSC < 1) {
if (diffMVdiffVal < -26) {
return 8;
}
else if (diffMVdiffVal >= -26) {
return 4;
}
}
else if (TSC >= 1) {
if (Rs < 245) {
if (SC < 491) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
if (Cs < 187) {
if (Rs < 147) {
if (MVDiff < 3) {
if (negBalance < 17) {
return 4;
}
else if (negBalance >= 17) {
return 1;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (Rs >= 147) {
return 1;
}
}
else if (Cs >= 187) {
if (Cs < 275) {
if (Rs < 148) {
if (Rs < 145) {
return 4;
}
else if (Rs >= 145) {
return 1;
}
}
else if (Rs >= 148) {
return 4;
}
}
else if (Cs >= 275) {
return 1;
}
}
}
}
else if (SC >= 491) {
return 4;
}
}
else if (Rs >= 245) {
if (posBalance < 41) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (posBalance >= 41) {
return 4;
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 89) {
return 1;
}
else if (Rs >= 89) {
if (Rs < 214) {
return 8;
}
else if (Rs >= 214) {
return 2;
}
}
}
}
}
else if (refDCval >= 63) {
if (SC < 415) {
if (ssDCval < 66) {
if (Rs < 150) {
if (Rs < 29) {
return 2;
}
else if (Rs >= 29) {
return 1;
}
}
else if (Rs >= 150) {
if (Rs < 180) {
return 2;
}
else if (Rs >= 180) {
return 4;
}
}
}
else if (ssDCval >= 66) {
return 2;
}
}
else if (SC >= 415) {
if (Cs < 379) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (Cs >= 379) {
return 1;
}
}
}
}
else if (TSCindex >= 2) {
if (TSC < 4) {
if (RsCsDiff < 1) {
if (SC < 338) {
if (tcor < 100) {
if (MVDiff < 3) {
if (ssDCval < 64) {
if (tcor < 96) {
if (negBalance < 35) {
if (mu_mv_mag_sq < 83) {
return 1;
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 84) {
return 2;
}
else if (Rs >= 84) {
if (Rs < 112) {
return 1;
}
else if (Rs >= 112) {
return 8;
}
}
}
}
else if (negBalance >= 35) {
return 2;
}
}
else if (tcor >= 96) {
if (negBalance < 22) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (negBalance >= 22) {
return 1;
}
}
}
else if (ssDCval >= 64) {
if (SC < 279) {
if (Cs < 138) {
return 1;
}
else if (Cs >= 138) {
return 2;
}
}
else if (SC >= 279) {
return 2;
}
}
}
else if (MVDiff >= 3) {
if (mcTcor < 95) {
if (Rs < 45) {
if (MVDiff < 86) {
return 1;
}
else if (MVDiff >= 86) {
return 4;
}
}
else if (Rs >= 45) {
if (Cs < 64) {
return 2;
}
else if (Cs >= 64) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (Cs < 193) {
if (TSCindex < 3) {
if (MVDiff < 6) {
if (SC < 302) {
return 1;
}
else if (SC >= 302) {
if (SC < 325) {
if (posBalance < 50) {
return 4;
}
else if (posBalance >= 50) {
if (Cs < 157) {
return 4;
}
else if (Cs >= 157) {
return 1;
}
}
}
else if (SC >= 325) {
return 1;
}
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (TSCindex >= 3) {
if (MVDiff < 12) {
if (SC < 317) {
return 1;
}
else if (SC >= 317) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (MVDiff >= 12) {
if (Cs < 32) {
return 1;
}
else if (Cs >= 32) {
if (refDCval < 60) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (refDCval >= 60) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 52) {
return 4;
}
else if (Rs >= 52) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 193) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
if (Rs < 43) {
return 1;
}
else if (Rs >= 43) {
if (Rs < 99) {
return 2;
}
else if (Rs >= 99) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 100) {
if (Cs < 17) {
return 2;
}
else if (Cs >= 17) {
return 8;
}
}
}
else if (SC >= 338) {
if (negBalance < 31) {
if (refDCval < 60) {
if (MVDiff < 1) {
if (Rs < 172) {
return 4;
}
else if (Rs >= 172) {
if (mu_mv_mag_sq < 127) {
if (diffRsCsdiff < -1) {
if (SC < 641) {
return 4;
}
else if (SC >= 641) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 21) {
if (Rs < 258) {
return 4;
}
else if (Rs >= 258) {
return 1;
}
}
else if (negBalance >= 21) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (Cs < 191) {
return 1;
}
else if (Cs >= 191) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
return 1;
}
}
}
}
else if (refDCval >= 60) {
if (negBalance < 28) {
if (Cs < 344) {
if (MVDiff < 3) {
if (Cs < 275) {
if (MVDiff < 1) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Cs < 247) {
return 4;
}
else if (Cs >= 247) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (SC < 361) {
return 2;
}
else if (SC >= 361) {
return 4;
}
}
}
else if (Cs >= 275) {
if (Cs < 343) {
return 4;
}
else if (Cs >= 343) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 192) {
if (Rs < 169) {
return 2;
}
else if (Rs >= 169) {
return 1;
}
}
else if (Rs >= 192) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (Rs < 222) {
return 4;
}
else if (Rs >= 222) {
return 1;
}
}
}
}
}
else if (Cs >= 344) {
if (ssDCval < 69) {
if (SC < 536) {
return 1;
}
else if (SC >= 536) {
return 4;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
else if (negBalance >= 28) {
if (mu_mv_mag_sq < 123) {
if (ssDCval < 63) {
if (Rs < 262) {
return 4;
}
else if (Rs >= 262) {
return 2;
}
}
else if (ssDCval >= 63) {
if (mu_mv_mag_sq < 111) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (SC < 443) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
if (mu_mv_mag_sq < 98) {
return 1;
}
else if (mu_mv_mag_sq >= 98) {
if (Rs < 173) {
return 4;
}
else if (Rs >= 173) {
return 1;
}
}
}
}
else if (SC >= 443) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 196) {
return 2;
}
else if (Rs >= 196) {
return 1;
}
}
}
}
}
else if (negBalance >= 31) {
if (tcor < 98) {
if (ssDCval < 63) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
return 4;
}
}
}
else if (ssDCval >= 63) {
if (refDCval < 69) {
if (diffRsCsdiff < -1) {
if (diffMVdiffVal < -1) {
if (Rs < 167) {
return 1;
}
else if (Rs >= 167) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 156) {
return 4;
}
else if (Rs >= 156) {
return 1;
}
}
}
}
else if (refDCval >= 69) {
return 4;
}
}
}
else if (tcor >= 98) {
if (posBalance < 43) {
if (Cs < 265) {
if (Rs < 174) {
return 8;
}
else if (Rs >= 174) {
return 1;
}
}
else if (Cs >= 265) {
return 4;
}
}
else if (posBalance >= 43) {
if (Cs < 482) {
if (mu_mv_mag_sq < 89) {
if (Cs < 243) {
return 4;
}
else if (Cs >= 243) {
return 1;
}
}
else if (mu_mv_mag_sq >= 89) {
return 1;
}
}
else if (Cs >= 482) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (mcTcor < 91) {
if (negBalance < 20) {
if (Rs < 255) {
return 1;
}
else if (Rs >= 255) {
return 8;
}
}
else if (negBalance >= 20) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 136) {
return 4;
}
else if (Rs >= 136) {
if (Rs < 216) {
return 2;
}
else if (Rs >= 216) {
return 8;
}
}
}
}
}
else if (mcTcor >= 91) {
if (MVDiff < 3) {
if (ssDCval < 69) {
if (Cs < 413) {
if (negBalance < 38) {
if (Rs < 170) {
return 1;
}
else if (Rs >= 170) {
if (Rs < 253) {
if (Cs < 404) {
return 4;
}
else if (Cs >= 404) {
return 1;
}
}
else if (Rs >= 253) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 4;
}
}
}
}
else if (negBalance >= 38) {
return 4;
}
}
else if (Cs >= 413) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 4;
}
}
}
else if (ssDCval >= 69) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Cs < 320) {
return 1;
}
else if (Cs >= 320) {
return 4;
}
}
}
}
else if (MVDiff >= 3) {
if (Rs < 37) {
return 1;
}
else if (Rs >= 37) {
if (RsDiff < 1) {
if (mu_mv_mag_sq < 37) {
if (refDCval < 56) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 1;
}
}
else if (refDCval >= 56) {
if (posBalance < 45) {
if (Rs < 241) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (Rs >= 241) {
return 4;
}
}
else if (posBalance >= 45) {
if (AFD < 11) {
if (MVDiff < 61) {
if (AFD < 4) {
if (Rs < 94) {
return 1;
}
else if (Rs >= 94) {
return 4;
}
}
else if (AFD >= 4) {
return 4;
}
}
else if (MVDiff >= 61) {
return 1;
}
}
else if (AFD >= 11) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (Cs < 297) {
if (Rs < 226) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
return 1;
}
}
}
else if (Rs >= 226) {
return 1;
}
}
else if (Cs >= 297) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
return 8;
}
}
}
}
else if (RsDiff >= 1) {
if (Rs < 265) {
if (TSC < 3) {
if (Rs < 211) {
if (refDCval < 56) {
return 4;
}
else if (refDCval >= 56) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (SC < 254) {
if (ssDCval < 61) {
if (Rs < 128) {
return 2;
}
else if (Rs >= 128) {
return 4;
}
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (SC >= 254) {
return 4;
}
}
}
}
else if (Rs >= 211) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (TSC >= 3) {
if (AFD < 10) {
if (mu_mv_mag_sq < 47) {
return 4;
}
else if (mu_mv_mag_sq >= 47) {
if (mu_mv_mag_sq < 57) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
return 4;
}
}
else if (mu_mv_mag_sq >= 57) {
return 4;
}
}
}
else if (AFD >= 10) {
if (gchDC < 2) {
if (ssDCval < 54) {
return 2;
}
else if (ssDCval >= 54) {
return 4;
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
else if (Rs >= 265) {
if (gchDC < 1) {
if (AFD < 6) {
if (CsDiff < 2) {
if (CsDiff < 1) {
if (Rs < 290) {
return 1;
}
else if (Rs >= 290) {
return 4;
}
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (gchDC >= 1) {
if (SC < 476) {
return 4;
}
else if (SC >= 476) {
if (AFD < 6) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 2;
}
}
else if (AFD >= 6) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 25) {
if (mcTcor < 76) {
if (diffMVdiffVal < 4) {
if (Cs < 101) {
return 2;
}
else if (Cs >= 101) {
return 4;
}
}
else if (diffMVdiffVal >= 4) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (Rs < 138) {
return 4;
}
else if (Rs >= 138) {
return 1;
}
}
}
}
else if (mcTcor >= 76) {
if (MVDiff < 47) {
if (AFD < 15) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (diffMVdiffVal < -28) {
return 2;
}
else if (diffMVdiffVal >= -28) {
return 1;
}
}
}
else if (AFD >= 15) {
return 2;
}
}
else if (MVDiff >= 47) {
if (AFD < 11) {
if (mu_mv_mag_sq < 14) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 14) {
if (Rs < 199) {
return 1;
}
else if (Rs >= 199) {
if (Rs < 280) {
return 2;
}
else if (Rs >= 280) {
return 4;
}
}
}
}
else if (AFD >= 11) {
if (diffRsCsdiff < -24) {
return 8;
}
else if (diffRsCsdiff >= -24) {
if (Rs < 187) {
if (MVDiff < 63) {
return 2;
}
else if (MVDiff >= 63) {
if (Cs < 142) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
else if (Cs >= 142) {
return 1;
}
}
}
else if (Rs >= 187) {
if (diffMVdiffVal < -12) {
return 1;
}
else if (diffMVdiffVal >= -12) {
if (posBalance < 55) {
if (Rs < 202) {
return 2;
}
else if (Rs >= 202) {
return 4;
}
}
else if (posBalance >= 55) {
if (Cs < 333) {
return 2;
}
else if (Cs >= 333) {
return 1;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (tcor < 86) {
if (AFD < 8) {
if (Cs < 267) {
return 2;
}
else if (Cs >= 267) {
return 8;
}
}
else if (AFD >= 8) {
if (RsCsDiff < 23) {
if (diffMVdiffVal < -73) {
if (Rs < 26) {
return 4;
}
else if (Rs >= 26) {
return 8;
}
}
else if (diffMVdiffVal >= -73) {
if (RsCsDiff < 1) {
if (gchDC < 14) {
return 4;
}
else if (gchDC >= 14) {
return 2;
}
}
else if (RsCsDiff >= 1) {
if (diffRsCsdiff < 7) {
if (posBalance < 62) {
if (negBalance < 65) {
if (Rs < 275) {
return 4;
}
else if (Rs >= 275) {
return 1;
}
}
else if (negBalance >= 65) {
return 1;
}
}
else if (posBalance >= 62) {
if (Rs < 45) {
return 8;
}
else if (Rs >= 45) {
if (AFD < 49) {
return 1;
}
else if (AFD >= 49) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 7) {
if (AFD < 21) {
if (Rs < 225) {
return 4;
}
else if (Rs >= 225) {
return 8;
}
}
else if (AFD >= 21) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 23) {
if (MVDiff < 313) {
return 8;
}
else if (MVDiff >= 313) {
return 4;
}
}
}
}
else if (tcor >= 86) {
if (MVDiff < 7) {
if (Rs < 207) {
return 4;
}
else if (Rs >= 207) {
return 1;
}
}
else if (MVDiff >= 7) {
if (refDCval < 53) {
return 4;
}
else if (refDCval >= 53) {
if (diffRsCsdiff < -1) {
if (ssDCval < 64) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
if (Rs < 79) {
return 8;
}
else if (Rs >= 79) {
return 2;
}
}
}
else if (ssDCval >= 64) {
if (Rs < 110) {
return 4;
}
else if (Rs >= 110) {
return 2;
}
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 70) {
if (TSCindex < 2) {
if (refDCval < 93) {
if (Cs < 263) {
if (SC < 246) {
if (Rs < 87) {
if (Rs < 59) {
return 1;
}
else if (Rs >= 59) {
return 4;
}
}
else if (Rs >= 87) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (SC < 226) {
return 2;
}
else if (SC >= 226) {
if (ssDCval < 75) {
if (Rs < 103) {
return 1;
}
else if (Rs >= 103) {
return 2;
}
}
else if (ssDCval >= 75) {
return 1;
}
}
}
}
}
else if (SC >= 246) {
if (Rs < 286) {
if (tcor < 99) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
return 8;
}
}
else if (tcor >= 99) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (Rs < 164) {
return 1;
}
else if (Rs >= 164) {
return 2;
}
}
}
}
else if (Rs >= 286) {
if (Cs < 205) {
return 1;
}
else if (Cs >= 205) {
return 2;
}
}
}
}
else if (Cs >= 263) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
return 4;
}
}
}
else if (refDCval >= 93) {
if (Cs < 140) {
if (Rs < 81) {
if (Rs < 8) {
return 1;
}
else if (Rs >= 8) {
return 4;
}
}
else if (Rs >= 81) {
return 1;
}
}
else if (Cs >= 140) {
if (MVDiff < 41) {
return 4;
}
else if (MVDiff >= 41) {
return 2;
}
}
}
}
else if (TSCindex >= 2) {
if (mcTcor < 97) {
if (Rs < 54) {
if (ssDCval < 132) {
if (mu_mv_mag_sq < 17) {
if (Cs < 118) {
return 1;
}
else if (Cs >= 118) {
return 2;
}
}
else if (mu_mv_mag_sq >= 17) {
if (ssDCval < 105) {
if (tcor < 85) {
if (diffMVdiffVal < -163) {
return 4;
}
else if (diffMVdiffVal >= -163) {
return 1;
}
}
else if (tcor >= 85) {
if (Cs < 50) {
return 1;
}
else if (Cs >= 50) {
return 2;
}
}
}
else if (ssDCval >= 105) {
if (posBalance < 48) {
if (Rs < 15) {
return 1;
}
else if (Rs >= 15) {
return 2;
}
}
else if (posBalance >= 48) {
return 4;
}
}
}
}
else if (ssDCval >= 132) {
if (SC < 12) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (SC >= 12) {
return 2;
}
}
}
else if (Rs >= 54) {
if (tcor < 87) {
if (RsCsDiff < 3) {
if (diffMVdiffVal < -17) {
if (posBalance < 54) {
if (Rs < 71) {
return 2;
}
else if (Rs >= 71) {
return 1;
}
}
else if (posBalance >= 54) {
if (Cs < 46) {
if (Rs < 100) {
return 1;
}
else if (Rs >= 100) {
return 4;
}
}
else if (Cs >= 46) {
return 2;
}
}
}
else if (diffMVdiffVal >= -17) {
if (negBalance < 47) {
if (Cs < 251) {
if (TSC < 6) {
if (MVDiff < 40) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (MVDiff >= 40) {
return 2;
}
}
else if (TSC >= 6) {
if (Rs < 177) {
if (Rs < 123) {
return 4;
}
else if (Rs >= 123) {
return 2;
}
}
else if (Rs >= 177) {
return 8;
}
}
}
else if (Cs >= 251) {
if (diffAFD < -1) {
if (Rs < 189) {
return 8;
}
else if (Rs >= 189) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (ssDCval < 104) {
if (Cs < 361) {
if (AFD < 5) {
if (posBalance < 58) {
if (Rs < 147) {
return 1;
}
else if (Rs >= 147) {
if (Rs < 222) {
return 2;
}
else if (Rs >= 222) {
return 8;
}
}
}
else if (posBalance >= 58) {
return 2;
}
}
else if (AFD >= 5) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 2) {
return 2;
}
else if (mu_mv_mag_sq >= 2) {
if (TSC < 5) {
if (SC < 124) {
return 8;
}
else if (SC >= 124) {
if (SC < 132) {
return 2;
}
else if (SC >= 132) {
if (SC < 160) {
if (MVDiff < 33) {
return 8;
}
else if (MVDiff >= 33) {
if (Rs < 137) {
return 2;
}
else if (Rs >= 137) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
return 8;
}
}
}
}
else if (SC >= 160) {
if (mu_mv_mag_sq < 14) {
if (Cs < 301) {
if (tcor < 55) {
if (MVDiff < 45) {
return 8;
}
else if (MVDiff >= 45) {
return 2;
}
}
else if (tcor >= 55) {
return 2;
}
}
else if (Cs >= 301) {
return 8;
}
}
else if (mu_mv_mag_sq >= 14) {
if (SC < 163) {
if (Rs < 146) {
return 8;
}
else if (Rs >= 146) {
return 2;
}
}
else if (SC >= 163) {
return 2;
}
}
}
}
}
}
else if (TSC >= 5) {
if (Cs < 31) {
return 1;
}
else if (Cs >= 31) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 361) {
return 8;
}
}
else if (ssDCval >= 104) {
if (tcor < 80) {
if (posBalance < 61) {
if (Rs < 152) {
return 2;
}
else if (Rs >= 152) {
return 4;
}
}
else if (posBalance >= 61) {
return 1;
}
}
else if (tcor >= 80) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 3) {
if (MVDiff < 22) {
if (ssDCval < 71) {
return 4;
}
else if (ssDCval >= 71) {
if (AFD < 15) {
if (ssDCval < 96) {
return 1;
}
else if (ssDCval >= 96) {
return 8;
}
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (MVDiff >= 22) {
if (TSC < 5) {
if (diffMVdiffVal < 11) {
if (RsCsDiff < 9) {
if (MVDiff < 26) {
return 2;
}
else if (MVDiff >= 26) {
if (mcTcor < 81) {
return 2;
}
else if (mcTcor >= 81) {
return 8;
}
}
}
else if (RsCsDiff >= 9) {
return 2;
}
}
else if (diffMVdiffVal >= 11) {
return 8;
}
}
else if (TSC >= 5) {
if (Cs < 70) {
if (Rs < 181) {
if (diffMVdiffVal < 9) {
if (Rs < 160) {
return 1;
}
else if (Rs >= 160) {
return 8;
}
}
else if (diffMVdiffVal >= 9) {
return 4;
}
}
else if (Rs >= 181) {
return 1;
}
}
else if (Cs >= 70) {
if (tcor < 66) {
if (posBalance < 66) {
if (AFD < 22) {
if (refDCval < 83) {
if (Rs < 139) {
return 1;
}
else if (Rs >= 139) {
return 2;
}
}
else if (refDCval >= 83) {
if (MVDiff < 57) {
return 2;
}
else if (MVDiff >= 57) {
if (Rs < 290) {
return 8;
}
else if (Rs >= 290) {
return 2;
}
}
}
}
else if (AFD >= 22) {
if (tcor < 1) {
return 8;
}
else if (tcor >= 1) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 129) {
return 1;
}
else if (mu_mv_mag_sq >= 129) {
if (Rs < 161) {
return 1;
}
else if (Rs >= 161) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 66) {
if (TSC < 11) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
if (diffAFD < 6) {
return 4;
}
else if (diffAFD >= 6) {
return 8;
}
}
}
else if (TSC >= 11) {
if (AFD < 37) {
return 1;
}
else if (AFD >= 37) {
if (AFD < 59) {
return 2;
}
else if (AFD >= 59) {
return 1;
}
}
}
}
}
else if (tcor >= 66) {
if (AFD < 14) {
if (Cs < 308) {
return 4;
}
else if (Cs >= 308) {
return 8;
}
}
else if (AFD >= 14) {
if (refDCval < 82) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 233) {
return 8;
}
else if (Rs >= 233) {
return 4;
}
}
}
else if (refDCval >= 82) {
if (ssDCval < 96) {
return 2;
}
else if (ssDCval >= 96) {
if (Cs < 375) {
if (Cs < 161) {
return 2;
}
else if (Cs >= 161) {
return 1;
}
}
else if (Cs >= 375) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (tcor >= 87) {
if (Cs < 73) {
if (SC < 117) {
if (Rs < 73) {
if (Rs < 63) {
return 1;
}
else if (Rs >= 63) {
return 2;
}
}
else if (Rs >= 73) {
return 8;
}
}
else if (SC >= 117) {
if (Cs < 62) {
return 1;
}
else if (Cs >= 62) {
if (Rs < 148) {
return 1;
}
else if (Rs >= 148) {
return 8;
}
}
}
}
else if (Cs >= 73) {
if (ssDCval < 129) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (Cs < 131) {
if (SC < 201) {
if (Rs < 107) {
return 1;
}
else if (Rs >= 107) {
return 8;
}
}
else if (SC >= 201) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (Cs >= 131) {
if (negBalance < 43) {
if (ssDCval < 77) {
if (MVDiff < 12) {
if (ssDCval < 74) {
if (Rs < 234) {
return 8;
}
else if (Rs >= 234) {
return 1;
}
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (ssDCval >= 77) {
if (SC < 454) {
return 1;
}
else if (SC >= 454) {
if (Rs < 284) {
return 4;
}
else if (Rs >= 284) {
return 2;
}
}
}
}
else if (negBalance >= 43) {
if (mu_mv_mag_sq < 3) {
if (Rs < 192) {
return 4;
}
else if (Rs >= 192) {
if (Cs < 248) {
return 2;
}
else if (Cs >= 248) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 107) {
return 2;
}
else if (Rs >= 107) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 129) {
if (Cs < 239) {
return 2;
}
else if (Cs >= 239) {
if (Cs < 426) {
return 8;
}
else if (Cs >= 426) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 47) {
if (mu_mv_mag_sq < 95) {
if (refDCval < 109) {
if (SC < 376) {
if (Cs < 89) {
if (Rs < 122) {
return 1;
}
else if (Rs >= 122) {
if (Cs < 40) {
return 1;
}
else if (Cs >= 40) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (Cs >= 89) {
if (refDCval < 93) {
if (mu_mv_mag_sq < 74) {
if (Cs < 200) {
if (Cs < 103) {
if (Rs < 92) {
return 1;
}
else if (Rs >= 92) {
return 2;
}
}
else if (Cs >= 103) {
return 1;
}
}
else if (Cs >= 200) {
return 2;
}
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 168) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (posBalance < 42) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
return 1;
}
}
else if (posBalance >= 42) {
return 1;
}
}
}
else if (Cs >= 168) {
return 1;
}
}
}
else if (refDCval >= 93) {
if (ssDCval < 97) {
return 2;
}
else if (ssDCval >= 97) {
if (MVDiff < 39) {
if (Rs < 73) {
return 4;
}
else if (Rs >= 73) {
return 1;
}
}
else if (MVDiff >= 39) {
return 2;
}
}
}
}
}
else if (SC >= 376) {
if (ssDCval < 80) {
if (Rs < 106) {
return 4;
}
else if (Rs >= 106) {
return 1;
}
}
else if (ssDCval >= 80) {
if (ssDCval < 101) {
if (Rs < 112) {
return 2;
}
else if (Rs >= 112) {
return 4;
}
}
else if (ssDCval >= 101) {
return 1;
}
}
}
}
else if (refDCval >= 109) {
if (tcor < 101) {
if (Cs < 16) {
return 8;
}
else if (Cs >= 16) {
if (posBalance < 29) {
return 4;
}
else if (posBalance >= 29) {
return 2;
}
}
}
else if (tcor >= 101) {
if (Rs < 88) {
return 1;
}
else if (Rs >= 88) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (Cs < 211) {
if (Cs < 127) {
if (ssDCval < 88) {
if (posBalance < 41) {
if (Cs < 117) {
return 8;
}
else if (Cs >= 117) {
return 2;
}
}
else if (posBalance >= 41) {
return 8;
}
}
else if (ssDCval >= 88) {
return 1;
}
}
else if (Cs >= 127) {
if (MVDiff < 3) {
if (mcTcor < 98) {
if (Cs < 157) {
return 8;
}
else if (Cs >= 157) {
return 2;
}
}
else if (mcTcor >= 98) {
if (ssDCval < 79) {
if (Cs < 210) {
return 2;
}
else if (Cs >= 210) {
if (Rs < 184) {
return 2;
}
else if (Rs >= 184) {
return 1;
}
}
}
else if (ssDCval >= 79) {
if (Rs < 220) {
return 2;
}
else if (Rs >= 220) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (Cs >= 211) {
if (mu_mv_mag_sq < 109) {
if (Cs < 273) {
return 2;
}
else if (Cs >= 273) {
return 4;
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 177) {
return 2;
}
else if (Rs >= 177) {
return 1;
}
}
}
}
}
else if (posBalance >= 47) {
if (Rs < 172) {
if (Cs < 219) {
if (Rs < 92) {
if (negBalance < 24) {
if (ssDCval < 100) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (ssDCval >= 100) {
return 4;
}
}
else if (negBalance >= 24) {
if (ssDCval < 123) {
if (SC < 90) {
return 1;
}
else if (SC >= 90) {
if (tcor < 98) {
if (refDCval < 83) {
if (tcor < 91) {
if (Rs < 87) {
return 1;
}
else if (Rs >= 87) {
return 2;
}
}
else if (tcor >= 91) {
return 4;
}
}
else if (refDCval >= 83) {
if (SC < 145) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (ssDCval < 93) {
return 1;
}
else if (ssDCval >= 93) {
return 2;
}
}
}
}
else if (SC >= 145) {
return 1;
}
}
}
else if (tcor >= 98) {
if (Rs < 75) {
return 4;
}
else if (Rs >= 75) {
if (mu_mv_mag_sq < 53) {
return 1;
}
else if (mu_mv_mag_sq >= 53) {
return 2;
}
}
}
}
}
else if (ssDCval >= 123) {
if (negBalance < 38) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
else if (negBalance >= 38) {
if (Rs < 2) {
return 1;
}
else if (Rs >= 2) {
return 2;
}
}
}
}
}
else if (Rs >= 92) {
if (refDCval < 75) {
if (Cs < 126) {
if (Rs < 125) {
return 2;
}
else if (Rs >= 125) {
return 8;
}
}
else if (Cs >= 126) {
if (AFD < 2) {
if (MVDiff < 1) {
if (Rs < 162) {
return 2;
}
else if (Rs >= 162) {
return 8;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (AFD >= 2) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (Rs < 98) {
return 2;
}
else if (Rs >= 98) {
return 1;
}
}
}
}
}
else if (refDCval >= 75) {
if (ssDCval < 98) {
if (TSC < 3) {
if (Cs < 202) {
return 1;
}
else if (Cs >= 202) {
if (Cs < 214) {
if (Rs < 114) {
return 2;
}
else if (Rs >= 114) {
return 1;
}
}
else if (Cs >= 214) {
return 4;
}
}
}
else if (TSC >= 3) {
if (Cs < 45) {
return 1;
}
else if (Cs >= 45) {
if (diffMVdiffVal < 4) {
if (diffMVdiffVal < -6) {
if (Rs < 113) {
return 2;
}
else if (Rs >= 113) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
}
else if (diffMVdiffVal >= 4) {
if (Rs < 141) {
return 8;
}
else if (Rs >= 141) {
if (Rs < 155) {
return 1;
}
else if (Rs >= 155) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 98) {
if (tcor < 88) {
if (gchDC < 8) {
return 1;
}
else if (gchDC >= 8) {
return 8;
}
}
else if (tcor >= 88) {
if (TSC < 2) {
if (MVDiff < 24) {
if (Rs < 102) {
return 2;
}
else if (Rs >= 102) {
return 4;
}
}
else if (MVDiff >= 24) {
return 1;
}
}
else if (TSC >= 2) {
return 2;
}
}
}
}
}
}
else if (Cs >= 219) {
if (Cs < 378) {
if (mu_mv_mag_sq < 43) {
if (MVDiff < 30) {
return 4;
}
else if (MVDiff >= 30) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (MVDiff < 5) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (negBalance < 33) {
if (ssDCval < 75) {
if (Rs < 111) {
return 4;
}
else if (Rs >= 111) {
return 1;
}
}
else if (ssDCval >= 75) {
if (CsDiff < 1) {
if (Cs < 298) {
if (ssDCval < 103) {
return 4;
}
else if (ssDCval >= 103) {
return 1;
}
}
else if (Cs >= 298) {
if (Rs < 76) {
return 2;
}
else if (Rs >= 76) {
return 1;
}
}
}
else if (CsDiff >= 1) {
return 1;
}
}
}
else if (negBalance >= 33) {
if (AFD < 6) {
if (SC < 485) {
return 4;
}
else if (SC >= 485) {
return 1;
}
}
else if (AFD >= 6) {
if (Rs < 111) {
return 1;
}
else if (Rs >= 111) {
return 8;
}
}
}
}
}
}
else if (Cs >= 378) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
else if (Rs >= 172) {
if (ssDCval < 76) {
if (Cs < 211) {
if (MVDiff < 3) {
if (negBalance < 30) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (negBalance >= 30) {
return 1;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (Cs >= 211) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
if (mu_mv_mag_sq < 18) {
return 1;
}
else if (mu_mv_mag_sq >= 18) {
if (Rs < 219) {
return 4;
}
else if (Rs >= 219) {
return 8;
}
}
}
}
}
else if (ssDCval >= 76) {
if (Cs < 115) {
if (SC < 381) {
if (Cs < 12) {
return 2;
}
else if (Cs >= 12) {
if (tcor < 92) {
if (Cs < 88) {
return 1;
}
else if (Cs >= 88) {
if (Cs < 106) {
return 4;
}
else if (Cs >= 106) {
return 1;
}
}
}
else if (tcor >= 92) {
return 1;
}
}
}
else if (SC >= 381) {
return 8;
}
}
else if (Cs >= 115) {
if (RsDiff < 2) {
if (TSC < 3) {
if (ssDCval < 96) {
if (ssDCval < 81) {
if (Rs < 178) {
return 4;
}
else if (Rs >= 178) {
if (Cs < 201) {
return 1;
}
else if (Cs >= 201) {
return 2;
}
}
}
else if (ssDCval >= 81) {
if (mu_mv_mag_sq < 104) {
return 4;
}
else if (mu_mv_mag_sq >= 104) {
return 2;
}
}
}
else if (ssDCval >= 96) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
if (Rs < 278) {
return 2;
}
else if (Rs >= 278) {
return 1;
}
}
}
}
else if (TSC >= 3) {
if (diffMVdiffVal < -16) {
return 8;
}
else if (diffMVdiffVal >= -16) {
if (mcTcor < 99) {
if (ssDCval < 77) {
if (Rs < 246) {
return 2;
}
else if (Rs >= 246) {
return 1;
}
}
else if (ssDCval >= 77) {
return 2;
}
}
else if (mcTcor >= 99) {
if (Rs < 203) {
return 8;
}
else if (Rs >= 203) {
return 2;
}
}
}
}
}
else if (RsDiff >= 2) {
if (Cs < 328) {
if (CsDiff < 3) {
if (Rs < 277) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
else if (Rs >= 277) {
return 1;
}
}
else if (CsDiff >= 3) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
else if (Cs >= 328) {
if (Rs < 284) {
return 2;
}
else if (Rs >= 284) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 300) {
if (RsDiff < 4) {
if (refDCval < 74) {
if (mcTcor < 89) {
if (refDCval < 41) {
if (tcor < 60) {
if (negBalance < 47) {
if (posBalance < 65) {
return 1;
}
else if (posBalance >= 65) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 1;
}
}
}
else if (negBalance >= 47) {
if (Cs < 268) {
return 8;
}
else if (Cs >= 268) {
return 4;
}
}
}
else if (tcor >= 60) {
if (Cs < 155) {
if (Cs < 29) {
return 8;
}
else if (Cs >= 29) {
return 1;
}
}
else if (Cs >= 155) {
if (negBalance < 30) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
if (Rs < 439) {
return 2;
}
else if (Rs >= 439) {
return 8;
}
}
}
else if (negBalance >= 30) {
if (ssDCval < 41) {
if (Rs < 774) {
if (tcor < 70) {
if (negBalance < 46) {
if (Rs < 444) {
return 2;
}
else if (Rs >= 444) {
return 8;
}
}
else if (negBalance >= 46) {
return 8;
}
}
else if (tcor >= 70) {
return 8;
}
}
else if (Rs >= 774) {
return 1;
}
}
else if (ssDCval >= 41) {
return 2;
}
}
}
}
}
else if (refDCval >= 41) {
if (mu_mv_mag_sq < 105) {
if (Cs < 199) {
if (Cs < 120) {
if (Rs < 321) {
return 8;
}
else if (Rs >= 321) {
return 1;
}
}
else if (Cs >= 120) {
if (posBalance < 70) {
if (Rs < 414) {
return 8;
}
else if (Rs >= 414) {
return 2;
}
}
else if (posBalance >= 70) {
return 1;
}
}
}
else if (Cs >= 199) {
if (posBalance < 54) {
if (ssDCval < 56) {
if (diffRsCsdiff < -25) {
return 2;
}
else if (diffRsCsdiff >= -25) {
if (negBalance < 38) {
if (refDCval < 49) {
return 1;
}
else if (refDCval >= 49) {
return 4;
}
}
else if (negBalance >= 38) {
if (diffRsCsdiff < -10) {
if (Rs < 464) {
return 1;
}
else if (Rs >= 464) {
return 4;
}
}
else if (diffRsCsdiff >= -10) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (Rs < 388) {
return 1;
}
else if (Rs >= 388) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 56) {
if (refDCval < 62) {
if (mcTcor < 88) {
return 2;
}
else if (mcTcor >= 88) {
if (Rs < 424) {
return 8;
}
else if (Rs >= 424) {
return 1;
}
}
}
else if (refDCval >= 62) {
if (Rs < 373) {
return 1;
}
else if (Rs >= 373) {
return 8;
}
}
}
}
else if (posBalance >= 54) {
if (mcTcor < 75) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Cs < 511) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 1;
}
}
else if (Cs >= 511) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (Rs < 417) {
return 4;
}
else if (Rs >= 417) {
return 8;
}
}
}
}
}
else if (mcTcor >= 75) {
if (negBalance < 51) {
if (refDCval < 61) {
if (Rs < 555) {
if (refDCval < 52) {
return 4;
}
else if (refDCval >= 52) {
if (tcor < 68) {
return 8;
}
else if (tcor >= 68) {
if (Cs < 417) {
return 4;
}
else if (Cs >= 417) {
return 2;
}
}
}
}
else if (Rs >= 555) {
return 8;
}
}
else if (refDCval >= 61) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
return 4;
}
}
}
else if (negBalance >= 51) {
if (MVDiff < 73) {
if (diffRsCsdiff < -13) {
return 2;
}
else if (diffRsCsdiff >= -13) {
return 4;
}
}
else if (MVDiff >= 73) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 8;
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1280) {
if (posBalance < 27) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
if (tcor < 101) {
if (Cs < 251) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (ssDCval < 39) {
return 2;
}
else if (ssDCval >= 39) {
return 1;
}
}
}
else if (Cs >= 251) {
if (MVDiff < 2) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (tcor >= 101) {
if (Rs < 402) {
if (posBalance < 13) {
return 8;
}
else if (posBalance >= 13) {
return 1;
}
}
else if (Rs >= 402) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Rs < 496) {
return 4;
}
else if (Rs >= 496) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 27) {
if (TSC < 1) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
if (Rs < 781) {
if (Rs < 627) {
return 4;
}
else if (Rs >= 627) {
return 8;
}
}
else if (Rs >= 781) {
return 1;
}
}
}
else if (TSC >= 1) {
if (ssDCval < 59) {
if (mu_mv_mag_sq < 98) {
if (Cs < 118) {
if (mu_mv_mag_sq < 49) {
if (tcor < 81) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (tcor >= 81) {
if (Rs < 960) {
return 1;
}
else if (Rs >= 960) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Cs < 80) {
if (Rs < 384) {
return 1;
}
else if (Rs >= 384) {
return 8;
}
}
else if (Cs >= 80) {
if (Rs < 411) {
return 8;
}
else if (Rs >= 411) {
return 1;
}
}
}
}
}
else if (Cs >= 118) {
if (negBalance < 40) {
if (Rs < 669) {
if (mu_mv_mag_sq < 71) {
if (posBalance < 57) {
if (Rs < 402) {
if (ssDCval < 57) {
if (tcor < 96) {
if (ssDCval < 38) {
if (Cs < 254) {
return 4;
}
else if (Cs >= 254) {
return 1;
}
}
else if (ssDCval >= 38) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (tcor >= 96) {
if (Cs < 204) {
return 1;
}
else if (Cs >= 204) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 4;
}
}
}
}
else if (ssDCval >= 57) {
if (Cs < 486) {
return 2;
}
else if (Cs >= 486) {
return 4;
}
}
}
else if (Rs >= 402) {
if (SC < 985) {
if (AFD < 4) {
if (Cs < 272) {
return 4;
}
else if (Cs >= 272) {
if (Cs < 376) {
return 1;
}
else if (Cs >= 376) {
return 8;
}
}
}
else if (AFD >= 4) {
if (posBalance < 36) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
else if (posBalance >= 36) {
if (posBalance < 50) {
if (gchDC < 2) {
if (diffMVdiffVal < -1) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (Rs < 477) {
return 8;
}
else if (Rs >= 477) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (gchDC >= 2) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (posBalance >= 50) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
if (Rs < 540) {
if (ssDCval < 57) {
if (negBalance < 27) {
if (Rs < 491) {
return 1;
}
else if (Rs >= 491) {
return 4;
}
}
else if (negBalance >= 27) {
if (MVDiff < 4) {
if (Rs < 494) {
return 4;
}
else if (Rs >= 494) {
return 1;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (ssDCval >= 57) {
if (Cs < 417) {
return 1;
}
else if (Cs >= 417) {
return 2;
}
}
}
else if (Rs >= 540) {
return 1;
}
}
}
}
}
}
else if (SC >= 985) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (SC < 1047) {
return 1;
}
else if (SC >= 1047) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 57) {
if (refDCval < 53) {
if (mcTcor < 93) {
if (ssDCval < 45) {
return 1;
}
else if (ssDCval >= 45) {
return 8;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
else if (refDCval >= 53) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
if (Rs < 447) {
return 2;
}
else if (Rs >= 447) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (refDCval < 54) {
if (negBalance < 33) {
if (ssDCval < 48) {
if (mu_mv_mag_sq < 77) {
if (Cs < 241) {
if (Rs < 500) {
return 1;
}
else if (Rs >= 500) {
return 2;
}
}
else if (Cs >= 241) {
return 4;
}
}
else if (mu_mv_mag_sq >= 77) {
if (ssDCval < 37) {
if (Rs < 416) {
return 1;
}
else if (Rs >= 416) {
return 2;
}
}
else if (ssDCval >= 37) {
return 1;
}
}
}
else if (ssDCval >= 48) {
if (TSC < 2) {
if (Rs < 423) {
return 2;
}
else if (Rs >= 423) {
return 1;
}
}
else if (TSC >= 2) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (Cs < 521) {
return 4;
}
else if (Cs >= 521) {
return 1;
}
}
}
}
}
else if (negBalance >= 33) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (refDCval >= 54) {
if (mcTcor < 97) {
if (posBalance < 47) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
return 1;
}
}
else if (posBalance >= 47) {
if (Cs < 263) {
return 2;
}
else if (Cs >= 263) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (Cs < 404) {
if (Rs < 413) {
return 1;
}
else if (Rs >= 413) {
return 2;
}
}
else if (Cs >= 404) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 1;
}
}
}
}
}
}
else if (Rs >= 669) {
if (Cs < 410) {
if (TSC < 2) {
if (Cs < 268) {
return 4;
}
else if (Cs >= 268) {
return 1;
}
}
else if (TSC >= 2) {
if (CsDiff < 2) {
if (diffMVdiffVal < -1) {
if (tcor < 85) {
return 2;
}
else if (tcor >= 85) {
if (mcTcor < 96) {
if (Cs < 268) {
return 2;
}
else if (Cs >= 268) {
return 1;
}
}
else if (mcTcor >= 96) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (Rs < 761) {
return 1;
}
else if (Rs >= 761) {
return 2;
}
}
}
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (Cs >= 410) {
if (MVDiff < 8) {
if (ssDCval < 55) {
if (Rs < 687) {
return 1;
}
else if (Rs >= 687) {
return 2;
}
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (MVDiff >= 8) {
if (mu_mv_mag_sq < 63) {
if (Rs < 724) {
return 8;
}
else if (Rs >= 724) {
return 1;
}
}
else if (mu_mv_mag_sq >= 63) {
if (mu_mv_mag_sq < 69) {
return 2;
}
else if (mu_mv_mag_sq >= 69) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 40) {
if (Rs < 419) {
if (Cs < 205) {
if (mu_mv_mag_sq < 49) {
return 1;
}
else if (mu_mv_mag_sq >= 49) {
if (Cs < 194) {
return 8;
}
else if (Cs >= 194) {
return 1;
}
}
}
else if (Cs >= 205) {
if (SC < 596) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 2;
}
}
else if (SC >= 596) {
if (MVDiff < 21) {
if (diffMVdiffVal < 1) {
if (refDCval < 49) {
return 1;
}
else if (refDCval >= 49) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (Rs < 362) {
return 2;
}
else if (Rs >= 362) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (MVDiff >= 21) {
if (tcor < 96) {
if (diffRsCsdiff < 4) {
if (AFD < 6) {
if (Rs < 369) {
return 2;
}
else if (Rs >= 369) {
return 4;
}
}
else if (AFD >= 6) {
return 2;
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 341) {
return 1;
}
else if (Rs >= 341) {
return 4;
}
}
}
else if (tcor >= 96) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
}
}
}
else if (Rs >= 419) {
if (AFD < 11) {
if (Cs < 235) {
return 1;
}
else if (Cs >= 235) {
if (mu_mv_mag_sq < 36) {
if (mu_mv_mag_sq < 2) {
if (posBalance < 53) {
if (Rs < 609) {
return 1;
}
else if (Rs >= 609) {
return 2;
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (negBalance < 46) {
if (SC < 878) {
return 8;
}
else if (SC >= 878) {
return 1;
}
}
else if (negBalance >= 46) {
if (Cs < 365) {
return 1;
}
else if (Cs >= 365) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 36) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Rs < 618) {
return 4;
}
else if (Rs >= 618) {
return 8;
}
}
}
}
}
else if (AFD >= 11) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
if (diffAFD < -1) {
if (Rs < 493) {
return 8;
}
else if (Rs >= 493) {
return 4;
}
}
else if (diffAFD >= -1) {
if (TSC < 6) {
if (Rs < 645) {
return 8;
}
else if (Rs >= 645) {
return 1;
}
}
else if (TSC >= 6) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 113) {
if (negBalance < 27) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (Rs < 593) {
return 1;
}
else if (Rs >= 593) {
if (Rs < 872) {
return 2;
}
else if (Rs >= 872) {
return 4;
}
}
}
}
else if (negBalance >= 27) {
if (tcor < 95) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
return 8;
}
}
else if (tcor >= 95) {
if (Rs < 347) {
return 4;
}
else if (Rs >= 347) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
else if (mcTcor >= 97) {
if (negBalance < 43) {
if (ssDCval < 49) {
if (mu_mv_mag_sq < 114) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
if (mu_mv_mag_sq < 103) {
return 1;
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 870) {
return 4;
}
else if (Rs >= 870) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (Rs < 378) {
return 1;
}
else if (Rs >= 378) {
return 8;
}
}
}
else if (ssDCval >= 49) {
if (posBalance < 42) {
if (ssDCval < 57) {
if (tcor < 101) {
if (Rs < 633) {
if (Rs < 570) {
return 1;
}
else if (Rs >= 570) {
return 8;
}
}
else if (Rs >= 633) {
if (Rs < 1271) {
return 1;
}
else if (Rs >= 1271) {
return 8;
}
}
}
else if (tcor >= 101) {
return 8;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (posBalance >= 42) {
return 1;
}
}
}
else if (negBalance >= 43) {
return 4;
}
}
}
}
else if (ssDCval >= 59) {
if (Rs < 1027) {
if (ssDCval < 72) {
if (mu_mv_mag_sq < 84) {
if (Rs < 509) {
if (ssDCval < 62) {
if (mcTcor < 95) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (diffAFD < 1) {
if (Cs < 177) {
return 1;
}
else if (Cs >= 177) {
return 2;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (mcTcor >= 95) {
if (mu_mv_mag_sq < 74) {
if (AFD < 9) {
if (negBalance < 36) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (Cs < 409) {
if (Rs < 442) {
return 2;
}
else if (Rs >= 442) {
return 1;
}
}
else if (Cs >= 409) {
return 1;
}
}
}
}
else if (negBalance >= 36) {
return 4;
}
}
else if (AFD >= 9) {
if (Cs < 404) {
if (Cs < 208) {
return 1;
}
else if (Cs >= 208) {
return 2;
}
}
else if (Cs >= 404) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 74) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
}
else if (ssDCval >= 62) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 28) {
if (posBalance < 48) {
if (Rs < 379) {
return 4;
}
else if (Rs >= 379) {
return 1;
}
}
else if (posBalance >= 48) {
return 4;
}
}
else if (mu_mv_mag_sq >= 28) {
if (Rs < 461) {
if (Cs < 301) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
else if (Cs >= 301) {
if (diffAFD < 1) {
if (mu_mv_mag_sq < 65) {
return 4;
}
else if (mu_mv_mag_sq >= 65) {
if (Cs < 424) {
return 2;
}
else if (Cs >= 424) {
return 4;
}
}
}
else if (diffAFD >= 1) {
return 2;
}
}
}
else if (Rs >= 461) {
return 1;
}
}
}
else if (MVDiff >= 8) {
if (mu_mv_mag_sq < 50) {
if (ssDCval < 69) {
if (Cs < 419) {
if (negBalance < 38) {
if (Rs < 431) {
return 1;
}
else if (Rs >= 431) {
return 4;
}
}
else if (negBalance >= 38) {
if (posBalance < 66) {
return 4;
}
else if (posBalance >= 66) {
return 2;
}
}
}
else if (Cs >= 419) {
if (Rs < 440) {
return 1;
}
else if (Rs >= 440) {
if (refDCval < 62) {
return 1;
}
else if (refDCval >= 62) {
return 2;
}
}
}
}
else if (ssDCval >= 69) {
if (negBalance < 42) {
if (Rs < 314) {
return 1;
}
else if (Rs >= 314) {
return 8;
}
}
else if (negBalance >= 42) {
if (Cs < 428) {
return 1;
}
else if (Cs >= 428) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (AFD < 11) {
if (tcor < 94) {
if (Cs < 485) {
return 2;
}
else if (Cs >= 485) {
return 8;
}
}
else if (tcor >= 94) {
if (Cs < 460) {
return 8;
}
else if (Cs >= 460) {
return 4;
}
}
}
else if (AFD >= 11) {
return 4;
}
}
}
}
}
else if (Rs >= 509) {
if (Cs < 176) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (TSC < 6) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 1;
}
}
}
else if (TSC >= 6) {
if (Rs < 781) {
return 1;
}
else if (Rs >= 781) {
return 2;
}
}
}
}
else if (Cs >= 176) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 828) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 1;
}
}
else if (Rs >= 828) {
if (CsDiff < 2) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 968) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 73) {
if (tcor < 98) {
if (diffMVdiffVal < -13) {
if (Rs < 786) {
return 8;
}
else if (Rs >= 786) {
return 1;
}
}
else if (diffMVdiffVal >= -13) {
return 1;
}
}
else if (tcor >= 98) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 73) {
if (MVDiff < 9) {
if (RsDiff < 2) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (RsDiff >= 2) {
if (Rs < 773) {
return 8;
}
else if (Rs >= 773) {
return 2;
}
}
}
else if (MVDiff >= 9) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (MVDiff < 21) {
if (diffMVdiffVal < -1) {
if (Rs < 813) {
return 4;
}
else if (Rs >= 813) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (MVDiff >= 21) {
if (Cs < 430) {
return 4;
}
else if (Cs >= 430) {
if (Rs < 527) {
return 2;
}
else if (Rs >= 527) {
return 1;
}
}
}
}
}
else if (Rs >= 968) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (Cs < 343) {
if (refDCval < 63) {
if (Cs < 283) {
if (Rs < 817) {
return 1;
}
else if (Rs >= 817) {
return 2;
}
}
else if (Cs >= 283) {
return 2;
}
}
else if (refDCval >= 63) {
if (posBalance < 37) {
if (MVDiff < 2) {
if (Rs < 338) {
return 2;
}
else if (Rs >= 338) {
if (Cs < 303) {
return 4;
}
else if (Cs >= 303) {
return 8;
}
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (posBalance >= 37) {
if (Cs < 323) {
if (negBalance < 34) {
if (Cs < 224) {
if (Cs < 113) {
return 4;
}
else if (Cs >= 113) {
return 1;
}
}
else if (Cs >= 224) {
if (Rs < 305) {
if (Cs < 257) {
return 2;
}
else if (Cs >= 257) {
return 1;
}
}
else if (Rs >= 305) {
if (MVDiff < 11) {
if (negBalance < 15) {
if (Rs < 452) {
return 4;
}
else if (Rs >= 452) {
return 8;
}
}
else if (negBalance >= 15) {
return 4;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
}
}
else if (negBalance >= 34) {
if (Rs < 405) {
return 1;
}
else if (Rs >= 405) {
return 2;
}
}
}
else if (Cs >= 323) {
if (Rs < 516) {
if (Rs < 358) {
return 4;
}
else if (Rs >= 358) {
if (Rs < 448) {
return 8;
}
else if (Rs >= 448) {
return 1;
}
}
}
else if (Rs >= 516) {
return 2;
}
}
}
}
}
else if (Cs >= 343) {
if (ssDCval < 67) {
if (TSC < 4) {
if (SC < 1003) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 102) {
if (negBalance < 25) {
if (Cs < 439) {
return 8;
}
else if (Cs >= 439) {
return 1;
}
}
else if (negBalance >= 25) {
return 1;
}
}
else if (mu_mv_mag_sq >= 102) {
return 1;
}
}
else if (MVDiff >= 4) {
if (Rs < 487) {
if (Cs < 490) {
return 2;
}
else if (Cs >= 490) {
return 4;
}
}
else if (Rs >= 487) {
return 1;
}
}
}
else if (SC >= 1003) {
if (Rs < 566) {
return 8;
}
else if (Rs >= 566) {
if (mu_mv_mag_sq < 109) {
if (Cs < 397) {
return 8;
}
else if (Cs >= 397) {
return 1;
}
}
else if (mu_mv_mag_sq >= 109) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (negBalance < 34) {
if (Rs < 467) {
return 2;
}
else if (Rs >= 467) {
return 8;
}
}
else if (negBalance >= 34) {
return 4;
}
}
}
else if (ssDCval >= 67) {
if (posBalance < 42) {
if (Rs < 859) {
if (SC < 860) {
return 8;
}
else if (SC >= 860) {
if (Cs < 363) {
if (Rs < 631) {
return 2;
}
else if (Rs >= 631) {
return 1;
}
}
else if (Cs >= 363) {
return 1;
}
}
}
else if (Rs >= 859) {
return 8;
}
}
else if (posBalance >= 42) {
if (mu_mv_mag_sq < 107) {
if (Cs < 414) {
if (CsDiff < 1) {
if (Rs < 604) {
return 8;
}
else if (Rs >= 604) {
return 2;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (Cs >= 414) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 121) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 121) {
return 1;
}
}
else if (mcTcor >= 99) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
return 2;
}
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 665) {
return 1;
}
else if (Rs >= 665) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 72) {
if (MVDiff < 2) {
if (Rs < 567) {
if (Rs < 437) {
return 2;
}
else if (Rs >= 437) {
return 1;
}
}
else if (Rs >= 567) {
return 4;
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 84) {
if (Rs < 481) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
if (Rs < 381) {
return 4;
}
else if (Rs >= 381) {
return 8;
}
}
}
else if (Rs >= 481) {
if (Cs < 100) {
return 1;
}
else if (Cs >= 100) {
if (Rs < 735) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 1;
}
}
}
else if (Rs >= 735) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (mu_mv_mag_sq < 114) {
return 8;
}
else if (mu_mv_mag_sq >= 114) {
return 4;
}
}
}
}
}
else if (Rs >= 1027) {
if (negBalance < 42) {
if (MVDiff < 2) {
if (diffMVdiffVal < -1) {
if (Rs < 1245) {
if (negBalance < 22) {
return 8;
}
else if (negBalance >= 22) {
return 1;
}
}
else if (Rs >= 1245) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (MVDiff >= 2) {
if (negBalance < 22) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (Rs < 1261) {
if (Rs < 1092) {
return 4;
}
else if (Rs >= 1092) {
return 1;
}
}
else if (Rs >= 1261) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (mcTcor < 95) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
else if (negBalance >= 42) {
if (refDCval < 68) {
return 4;
}
else if (refDCval >= 68) {
if (Rs < 1091) {
return 1;
}
else if (Rs >= 1091) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1280) {
if (SC < 1600) {
if (mu_mv_mag_sq < 59) {
if (Rs < 1291) {
return 2;
}
else if (Rs >= 1291) {
return 1;
}
}
else if (mu_mv_mag_sq >= 59) {
if (Cs < 221) {
return 8;
}
else if (Cs >= 221) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
}
}
else if (SC >= 1600) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1762) {
if (ssDCval < 64) {
if (refDCval < 56) {
if (posBalance < 65) {
return 1;
}
else if (posBalance >= 65) {
return 2;
}
}
else if (refDCval >= 56) {
if (Cs < 424) {
if (SC < 1910) {
return 1;
}
else if (SC >= 1910) {
return 2;
}
}
else if (Cs >= 424) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (TSCindex < 2) {
return 2;
}
else if (TSCindex >= 2) {
if (Rs < 1414) {
return 8;
}
else if (Rs >= 1414) {
if (Rs < 1540) {
if (Rs < 1473) {
return 2;
}
else if (Rs >= 1473) {
return 1;
}
}
else if (Rs >= 1540) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (mu_mv_mag_sq < 56) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (AFD < 9) {
if (negBalance < 26) {
if (Rs < 1589) {
return 2;
}
else if (Rs >= 1589) {
return 1;
}
}
else if (negBalance >= 26) {
return 1;
}
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (posBalance < 38) {
if (refDCval < 64) {
return 2;
}
else if (refDCval >= 64) {
return 1;
}
}
else if (posBalance >= 38) {
if (MVDiff < 4) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (Cs < 399) {
return 8;
}
else if (Cs >= 399) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
else if (MVDiff >= 4) {
return 1;
}
}
}
}
}
else if (Rs >= 1762) {
if (Cs < 430) {
if (RsCsDiff < 10) {
return 2;
}
else if (RsCsDiff >= 10) {
if (Cs < 261) {
return 1;
}
else if (Cs >= 261) {
return 2;
}
}
}
else if (Cs >= 430) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (AFD < 1) {
if (Rs < 1403) {
return 1;
}
else if (Rs >= 1403) {
return 2;
}
}
else if (AFD >= 1) {
if (Rs < 1784) {
return 8;
}
else if (Rs >= 1784) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 74) {
if (Cs < 141) {
if (posBalance < 15) {
return 8;
}
else if (posBalance >= 15) {
if (tcor < 58) {
if (Cs < 36) {
return 4;
}
else if (Cs >= 36) {
if (mcTcor < 85) {
return 2;
}
else if (mcTcor >= 85) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (diffMVdiffVal < -3) {
if (MVDiff < 29) {
return 8;
}
else if (MVDiff >= 29) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
}
}
}
else if (tcor >= 58) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 72) {
if (ssDCval < 106) {
if (Rs < 1213) {
if (diffRsCsdiff < -2) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
else if (diffRsCsdiff >= -2) {
if (SC < 391) {
return 1;
}
else if (SC >= 391) {
if (Cs < 10) {
return 1;
}
else if (Cs >= 10) {
return 8;
}
}
}
}
else if (Rs >= 1213) {
return 1;
}
}
else if (ssDCval >= 106) {
return 8;
}
}
else if (mu_mv_mag_sq >= 72) {
if (Cs < 48) {
return 8;
}
else if (Cs >= 48) {
if (posBalance < 50) {
if (diffMVdiffVal < -10) {
return 8;
}
else if (diffMVdiffVal >= -10) {
return 1;
}
}
else if (posBalance >= 50) {
if (Rs < 871) {
return 8;
}
else if (Rs >= 871) {
return 1;
}
}
}
}
}
else if (RsDiff >= 2) {
if (refDCval < 111) {
if (mu_mv_mag_sq < 44) {
return 1;
}
else if (mu_mv_mag_sq >= 44) {
if (Rs < 630) {
return 2;
}
else if (Rs >= 630) {
return 1;
}
}
}
else if (refDCval >= 111) {
if (mcTcor < 97) {
return 1;
}
else if (mcTcor >= 97) {
if (Rs < 1552) {
return 8;
}
else if (Rs >= 1552) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 141) {
if (SC < 1873) {
if (SC < 857) {
if (mcTcor < 97) {
if (MVDiff < 20) {
if (Rs < 504) {
if (TSC < 6) {
if (AFD < 9) {
if (Cs < 358) {
if (ssDCval < 121) {
if (AFD < 7) {
if (tcor < 97) {
if (Rs < 319) {
if (Rs < 310) {
return 8;
}
else if (Rs >= 310) {
return 2;
}
}
else if (Rs >= 319) {
return 1;
}
}
else if (tcor >= 97) {
return 8;
}
}
else if (AFD >= 7) {
if (Rs < 378) {
return 2;
}
else if (Rs >= 378) {
return 1;
}
}
}
else if (ssDCval >= 121) {
return 8;
}
}
else if (Cs >= 358) {
if (Rs < 315) {
if (Cs < 429) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (Cs >= 429) {
return 2;
}
}
else if (Rs >= 315) {
if (RsCsDiff < 1) {
return 1;
}
else if (RsCsDiff >= 1) {
return 8;
}
}
}
}
else if (AFD >= 9) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
}
else if (TSC >= 6) {
if (Cs < 486) {
return 8;
}
else if (Cs >= 486) {
return 2;
}
}
}
else if (Rs >= 504) {
if (Cs < 266) {
return 4;
}
else if (Cs >= 266) {
return 8;
}
}
}
else if (MVDiff >= 20) {
if (Cs < 176) {
if (Cs < 144) {
return 2;
}
else if (Cs >= 144) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (Cs >= 176) {
if (MVDiff < 39) {
if (ssDCval < 119) {
if (diffAFD < -1) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 356) {
return 1;
}
else if (Rs >= 356) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 318) {
return 8;
}
else if (Rs >= 318) {
if (Rs < 327) {
return 1;
}
else if (Rs >= 327) {
return 2;
}
}
}
}
}
else if (ssDCval >= 119) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
else if (MVDiff >= 39) {
if (MVDiff < 91) {
if (Cs < 282) {
return 4;
}
else if (Cs >= 282) {
return 2;
}
}
else if (MVDiff >= 91) {
if (SC < 675) {
return 2;
}
else if (SC >= 675) {
if (posBalance < 55) {
if (Rs < 342) {
return 8;
}
else if (Rs >= 342) {
return 2;
}
}
else if (posBalance >= 55) {
if (SC < 845) {
return 4;
}
else if (SC >= 845) {
return 8;
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 116) {
if (Cs < 182) {
if (ssDCval < 79) {
if (Rs < 456) {
return 8;
}
else if (Rs >= 456) {
return 4;
}
}
else if (ssDCval >= 79) {
return 1;
}
}
else if (Cs >= 182) {
if (ssDCval < 83) {
if (AFD < 5) {
if (Cs < 314) {
if (Rs < 348) {
return 1;
}
else if (Rs >= 348) {
if (Rs < 578) {
return 8;
}
else if (Rs >= 578) {
if (Rs < 612) {
return 2;
}
else if (Rs >= 612) {
return 8;
}
}
}
}
else if (Cs >= 314) {
if (Rs < 402) {
return 8;
}
else if (Rs >= 402) {
return 1;
}
}
}
else if (AFD >= 5) {
if (SC < 731) {
return 2;
}
else if (SC >= 731) {
if (negBalance < 38) {
return 4;
}
else if (negBalance >= 38) {
if (Rs < 415) {
return 4;
}
else if (Rs >= 415) {
return 1;
}
}
}
}
}
else if (ssDCval >= 83) {
if (negBalance < 30) {
if (Cs < 398) {
if (ssDCval < 84) {
return 2;
}
else if (ssDCval >= 84) {
return 4;
}
}
else if (Cs >= 398) {
return 1;
}
}
else if (negBalance >= 30) {
if (mu_mv_mag_sq < 12) {
if (MVDiff < 23) {
if (refDCval < 85) {
return 4;
}
else if (refDCval >= 85) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
if (Rs < 414) {
return 2;
}
else if (Rs >= 414) {
return 4;
}
}
}
}
else if (MVDiff >= 23) {
if (MVDiff < 95) {
if (diffMVdiffVal < 3) {
if (tcor < 95) {
return 2;
}
else if (tcor >= 95) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
else if (MVDiff >= 95) {
if (Cs < 232) {
return 2;
}
else if (Cs >= 232) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 12) {
if (AFD < 5) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
return 2;
}
}
}
else if (AFD >= 5) {
if (SC < 676) {
if (refDCval < 113) {
return 1;
}
else if (refDCval >= 113) {
return 2;
}
}
else if (SC >= 676) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (ssDCval < 160) {
if (ssDCval < 88) {
return 2;
}
else if (ssDCval >= 88) {
return 8;
}
}
else if (ssDCval >= 160) {
return 4;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 116) {
if (Cs < 305) {
if (ssDCval < 94) {
if (Rs < 316) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (Rs >= 316) {
return 2;
}
}
else if (ssDCval >= 94) {
return 1;
}
}
else if (Cs >= 305) {
return 8;
}
}
}
}
else if (SC >= 857) {
if (ssDCval < 81) {
if (Rs < 846) {
if (mu_mv_mag_sq < 96) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (Rs < 707) {
if (MVDiff < 15) {
if (mu_mv_mag_sq < 92) {
if (diffAFD < -2) {
if (Rs < 620) {
return 1;
}
else if (Rs >= 620) {
return 4;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 92) {
return 2;
}
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (Rs >= 707) {
if (MVDiff < 15) {
if (Cs < 457) {
return 8;
}
else if (Cs >= 457) {
return 4;
}
}
else if (MVDiff >= 15) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (negBalance < 29) {
if (SC < 891) {
return 1;
}
else if (SC >= 891) {
if (Cs < 335) {
return 8;
}
else if (Cs >= 335) {
if (Cs < 347) {
return 1;
}
else if (Cs >= 347) {
if (Rs < 499) {
return 8;
}
else if (Rs >= 499) {
if (Rs < 735) {
return 1;
}
else if (Rs >= 735) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
else if (Rs >= 846) {
if (posBalance < 44) {
if (refDCval < 75) {
return 1;
}
else if (refDCval >= 75) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (posBalance >= 44) {
if (SC < 1827) {
if (SC < 1146) {
return 8;
}
else if (SC >= 1146) {
if (Cs < 440) {
return 1;
}
else if (Cs >= 440) {
return 8;
}
}
}
else if (SC >= 1827) {
return 8;
}
}
}
}
else if (ssDCval >= 81) {
if (ssDCval < 85) {
if (Cs < 251) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
if (SC < 966) {
return 4;
}
else if (SC >= 966) {
return 1;
}
}
}
else if (Cs >= 251) {
if (MVDiff < 5) {
if (SC < 1126) {
if (Rs < 450) {
return 8;
}
else if (Rs >= 450) {
return 1;
}
}
else if (SC >= 1126) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
if (ssDCval < 82) {
if (Rs < 994) {
if (posBalance < 27) {
return 8;
}
else if (posBalance >= 27) {
return 1;
}
}
else if (Rs >= 994) {
return 2;
}
}
else if (ssDCval >= 82) {
return 1;
}
}
}
}
else if (MVDiff >= 5) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (Rs < 499) {
return 8;
}
else if (Rs >= 499) {
return 1;
}
}
}
}
}
else if (ssDCval >= 85) {
if (ssDCval < 115) {
if (posBalance < 48) {
if (Cs < 452) {
if (Rs < 755) {
if (Rs < 576) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Rs >= 576) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
return 8;
}
}
}
else if (Rs >= 755) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
}
}
else if (Cs >= 452) {
if (RsCsDiff < 7) {
if (ssDCval < 94) {
if (negBalance < 32) {
return 1;
}
else if (negBalance >= 32) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (RsCsDiff >= 7) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
return 2;
}
}
}
}
else if (posBalance >= 48) {
if (mcTcor < 93) {
if (mcTcor < 85) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
return 2;
}
}
else if (mcTcor >= 85) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (Cs < 287) {
return 1;
}
else if (Cs >= 287) {
return 8;
}
}
}
}
else if (mcTcor >= 93) {
if (Rs < 911) {
if (MVDiff < 4) {
if (RsCsDiff < 7) {
if (CsDiff < 2) {
if (Rs < 725) {
if (ssDCval < 95) {
return 1;
}
else if (ssDCval >= 95) {
if (Rs < 549) {
return 1;
}
else if (Rs >= 549) {
return 8;
}
}
}
else if (Rs >= 725) {
return 8;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (RsCsDiff >= 7) {
if (negBalance < 51) {
if (Rs < 587) {
return 2;
}
else if (Rs >= 587) {
return 1;
}
}
else if (negBalance >= 51) {
return 1;
}
}
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 34) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (RsCsDiff < 12) {
if (Rs < 450) {
return 2;
}
else if (Rs >= 450) {
return 8;
}
}
else if (RsCsDiff >= 12) {
if (mcTcor < 97) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
return 8;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (mu_mv_mag_sq < 103) {
return 8;
}
else if (mu_mv_mag_sq >= 103) {
return 1;
}
}
}
}
else if (Rs >= 911) {
if (ssDCval < 97) {
if (Rs < 1094) {
return 8;
}
else if (Rs >= 1094) {
return 1;
}
}
else if (ssDCval >= 97) {
return 8;
}
}
}
}
}
else if (ssDCval >= 115) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 81) {
if (ssDCval < 139) {
if (Rs < 900) {
if (mcTcor < 98) {
if (MVDiff < 29) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (MVDiff >= 29) {
if (Rs < 521) {
return 4;
}
else if (Rs >= 521) {
return 2;
}
}
}
else if (mcTcor >= 98) {
return 2;
}
}
else if (Rs >= 900) {
if (Cs < 361) {
return 8;
}
else if (Cs >= 361) {
return 1;
}
}
}
else if (ssDCval >= 139) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (AFD < 9) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (AFD >= 9) {
if (Rs < 425) {
return 1;
}
else if (Rs >= 425) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (posBalance < 37) {
if (Rs < 436) {
return 4;
}
else if (Rs >= 436) {
return 1;
}
}
else if (posBalance >= 37) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (ssDCval < 127) {
if (Rs < 442) {
return 4;
}
else if (Rs >= 442) {
return 2;
}
}
else if (ssDCval >= 127) {
if (Rs < 584) {
return 8;
}
else if (Rs >= 584) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (refDCval < 120) {
if (SC < 987) {
return 1;
}
else if (SC >= 987) {
return 2;
}
}
else if (refDCval >= 120) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (MVDiff < 75) {
if (ssDCval < 155) {
return 1;
}
else if (ssDCval >= 155) {
return 4;
}
}
else if (MVDiff >= 75) {
return 8;
}
}
}
}
}
}
}
}
}
else if (SC >= 1873) {
if (Cs < 290) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (Cs >= 290) {
if (mu_mv_mag_sq < 82) {
if (tcor < 98) {
if (posBalance < 53) {
if (negBalance < 30) {
if (Rs < 1821) {
return 2;
}
else if (Rs >= 1821) {
return 8;
}
}
else if (negBalance >= 30) {
return 1;
}
}
else if (posBalance >= 53) {
if (Cs < 360) {
if (Rs < 1837) {
return 8;
}
else if (Rs >= 1837) {
return 1;
}
}
else if (Cs >= 360) {
if (Cs < 479) {
return 2;
}
else if (Cs >= 479) {
return 1;
}
}
}
}
else if (tcor >= 98) {
if (Cs < 300) {
return 1;
}
else if (Cs >= 300) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (refDCval < 80) {
if (TSC < 3) {
if (Rs < 2121) {
return 1;
}
else if (Rs >= 2121) {
return 2;
}
}
else if (TSC >= 3) {
if (Cs < 417) {
return 2;
}
else if (Cs >= 417) {
if (Rs < 2105) {
return 8;
}
else if (Rs >= 2105) {
return 1;
}
}
}
}
else if (refDCval >= 80) {
if (Rs < 1512) {
return 1;
}
else if (Rs >= 1512) {
if (Rs < 1979) {
if (negBalance < 20) {
return 8;
}
else if (negBalance >= 20) {
if (Rs < 1917) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Cs < 429) {
return 1;
}
else if (Cs >= 429) {
return 4;
}
}
}
else if (Rs >= 1917) {
return 1;
}
}
}
else if (Rs >= 1979) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (Cs < 367) {
if (RsDiff < 8) {
if (mu_mv_mag_sq < 53) {
if (gchDC < 3) {
if (tcor < 65) {
if (tcor < 25) {
return 1;
}
else if (tcor >= 25) {
if (Rs < 482) {
if (CsDiff < 4) {
if (tcor < 51) {
return 1;
}
else if (tcor >= 51) {
if (Cs < 43) {
return 1;
}
else if (Cs >= 43) {
return 8;
}
}
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (Rs >= 482) {
if (tcor < 48) {
if (MVDiff < 43) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
else if (MVDiff >= 43) {
return 2;
}
}
else if (tcor >= 48) {
if (RsDiff < 6) {
if (MVDiff < 31) {
if (AFD < 14) {
if (negBalance < 49) {
if (diffMVdiffVal < -13) {
return 8;
}
else if (diffMVdiffVal >= -13) {
return 2;
}
}
else if (negBalance >= 49) {
if (Rs < 824) {
return 8;
}
else if (Rs >= 824) {
if (Rs < 840) {
return 2;
}
else if (Rs >= 840) {
return 8;
}
}
}
}
else if (AFD >= 14) {
return 1;
}
}
else if (MVDiff >= 31) {
if (ssDCval < 149) {
return 2;
}
else if (ssDCval >= 149) {
return 1;
}
}
}
else if (RsDiff >= 6) {
if (ssDCval < 92) {
return 1;
}
else if (ssDCval >= 92) {
if (Rs < 1174) {
return 8;
}
else if (Rs >= 1174) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 65) {
if (diffAFD < 5) {
if (Rs < 793) {
if (negBalance < 58) {
if (diffRsCsdiff < 2) {
if (negBalance < 52) {
if (Cs < 340) {
return 1;
}
else if (Cs >= 340) {
return 8;
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
return 1;
}
}
else if (negBalance >= 58) {
return 2;
}
}
else if (Rs >= 793) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
if (SC < 894) {
return 4;
}
else if (SC >= 894) {
return 1;
}
}
}
}
else if (diffAFD >= 5) {
if (Rs < 598) {
return 4;
}
else if (Rs >= 598) {
return 8;
}
}
}
}
else if (gchDC >= 3) {
if (mu_mv_mag_sq < 7) {
if (Cs < 208) {
if (ssDCval < 48) {
return 8;
}
else if (ssDCval >= 48) {
return 1;
}
}
else if (Cs >= 208) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (Rs < 1044) {
return 4;
}
else if (Rs >= 1044) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffRsCsdiff < 13) {
if (diffTSC < -4) {
if (Rs < 458) {
return 1;
}
else if (Rs >= 458) {
return 2;
}
}
else if (diffTSC >= -4) {
return 1;
}
}
else if (diffRsCsdiff >= 13) {
if (tcor < 59) {
return 1;
}
else if (tcor >= 59) {
if (posBalance < 56) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
else if (posBalance >= 56) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (mcTcor < 82) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
return 1;
}
}
}
else if (mcTcor >= 82) {
if (ssDCval < 80) {
if (CsDiff < 3) {
if (posBalance < 76) {
return 1;
}
else if (posBalance >= 76) {
return 2;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (ssDCval >= 80) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (tcor < 36) {
return 4;
}
else if (tcor >= 36) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (MVDiff < 31) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
if (Rs < 1200) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
return 1;
}
}
else if (Rs >= 1200) {
if (diffMVdiffVal < -7) {
if (tcor < 58) {
if (Rs < 1966) {
return 4;
}
else if (Rs >= 1966) {
return 8;
}
}
else if (tcor >= 58) {
if (Cs < 237) {
return 2;
}
else if (Cs >= 237) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -7) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
return 2;
}
}
}
}
}
else if (MVDiff >= 31) {
if (negBalance < 53) {
if (AFD < 30) {
if (posBalance < 72) {
if (MVDiff < 62) {
if (refDCval < 71) {
if (negBalance < 44) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
return 2;
}
}
else if (negBalance >= 44) {
return 1;
}
}
else if (refDCval >= 71) {
return 2;
}
}
else if (MVDiff >= 62) {
if (Cs < 341) {
return 1;
}
else if (Cs >= 341) {
return 2;
}
}
}
else if (posBalance >= 72) {
if (diffRsCsdiff < 40) {
return 2;
}
else if (diffRsCsdiff >= 40) {
if (Cs < 240) {
return 2;
}
else if (Cs >= 240) {
return 1;
}
}
}
}
else if (AFD >= 30) {
return 1;
}
}
else if (negBalance >= 53) {
if (ssDCval < 115) {
if (ssDCval < 64) {
if (Cs < 335) {
return 1;
}
else if (Cs >= 335) {
return 4;
}
}
else if (ssDCval >= 64) {
if (tcor < 37) {
if (AFD < 26) {
return 4;
}
else if (AFD >= 26) {
return 2;
}
}
else if (tcor >= 37) {
if (Cs < 149) {
if (RsDiff < 16) {
return 1;
}
else if (RsDiff >= 16) {
return 2;
}
}
else if (Cs >= 149) {
if (diffMVdiffVal < -5) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 2;
}
}
else if (diffMVdiffVal >= -5) {
return 2;
}
}
}
}
}
else if (ssDCval >= 115) {
if (diffAFD < 5) {
if (Rs < 1427) {
if (mu_mv_mag_sq < 50) {
if (Rs < 1016) {
return 4;
}
else if (Rs >= 1016) {
return 1;
}
}
else if (mu_mv_mag_sq >= 50) {
return 2;
}
}
else if (Rs >= 1427) {
if (posBalance < 55) {
if (Rs < 2028) {
return 1;
}
else if (Rs >= 2028) {
if (Rs < 2269) {
return 8;
}
else if (Rs >= 2269) {
return 1;
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (diffAFD >= 5) {
return 1;
}
}
}
}
}
}
else if (Cs >= 367) {
if (negBalance < 48) {
if (Rs < 1576) {
if (posBalance < 69) {
if (mu_mv_mag_sq < 68) {
if (Rs < 531) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 1;
}
}
else if (Rs >= 531) {
if (negBalance < 46) {
if (MVDiff < 32) {
if (RsCsDiff < 37) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (RsCsDiff >= 37) {
return 1;
}
}
else if (MVDiff >= 32) {
if (diffRsCsdiff < 16) {
if (diffTSC < -1) {
if (Cs < 445) {
return 1;
}
else if (Cs >= 445) {
return 8;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (diffRsCsdiff >= 16) {
return 1;
}
}
}
else if (negBalance >= 46) {
if (Rs < 657) {
return 2;
}
else if (Rs >= 657) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (negBalance < 30) {
if (Rs < 522) {
return 8;
}
else if (Rs >= 522) {
return 1;
}
}
else if (negBalance >= 30) {
return 8;
}
}
}
else if (posBalance >= 69) {
if (posBalance < 81) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (posBalance < 70) {
return 4;
}
else if (posBalance >= 70) {
return 8;
}
}
}
else if (posBalance >= 81) {
if (Rs < 852) {
if (Cs < 457) {
return 4;
}
else if (Cs >= 457) {
return 2;
}
}
else if (Rs >= 852) {
return 1;
}
}
}
}
else if (Rs >= 1576) {
if (ssDCval < 115) {
return 2;
}
else if (ssDCval >= 115) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (MVDiff < 44) {
if (mu_mv_mag_sq < 19) {
if (AFD < 24) {
if (Rs < 695) {
if (ssDCval < 41) {
return 8;
}
else if (ssDCval >= 41) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 4;
}
}
}
else if (Rs >= 695) {
if (ssDCval < 99) {
if (posBalance < 58) {
if (tcor < 62) {
return 2;
}
else if (tcor >= 62) {
if (Rs < 1622) {
return 8;
}
else if (Rs >= 1622) {
return 2;
}
}
}
else if (posBalance >= 58) {
if (Cs < 500) {
return 8;
}
else if (Cs >= 500) {
if (Rs < 1038) {
return 8;
}
else if (Rs >= 1038) {
return 2;
}
}
}
}
else if (ssDCval >= 99) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
if (Cs < 417) {
return 2;
}
else if (Cs >= 417) {
return 4;
}
}
}
}
}
else if (AFD >= 24) {
if (AFD < 27) {
return 1;
}
else if (AFD >= 27) {
if (Rs < 1885) {
return 4;
}
else if (Rs >= 1885) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 19) {
if (ssDCval < 112) {
if (SC < 1400) {
if (tcor < 55) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (tcor >= 55) {
if (AFD < 23) {
if (mu_mv_mag_sq < 25) {
return 2;
}
else if (mu_mv_mag_sq >= 25) {
return 8;
}
}
else if (AFD >= 23) {
return 2;
}
}
}
else if (SC >= 1400) {
return 4;
}
}
else if (ssDCval >= 112) {
if (MVDiff < 31) {
return 2;
}
else if (MVDiff >= 31) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
}
else if (MVDiff >= 44) {
if (mcTcor < 78) {
if (tcor < 39) {
if (refDCval < 60) {
if (Rs < 1476) {
return 8;
}
else if (Rs >= 1476) {
return 2;
}
}
else if (refDCval >= 60) {
if (ssDCval < 121) {
return 4;
}
else if (ssDCval >= 121) {
return 1;
}
}
}
else if (tcor >= 39) {
if (Cs < 532) {
if (refDCval < 126) {
return 8;
}
else if (refDCval >= 126) {
if (Cs < 448) {
return 8;
}
else if (Cs >= 448) {
return 4;
}
}
}
else if (Cs >= 532) {
return 2;
}
}
}
else if (mcTcor >= 78) {
if (refDCval < 77) {
if (ssDCval < 52) {
if (Cs < 416) {
return 1;
}
else if (Cs >= 416) {
if (Cs < 475) {
return 2;
}
else if (Cs >= 475) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (negBalance < 55) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 707) {
return 4;
}
else if (Rs >= 707) {
return 1;
}
}
}
else if (negBalance >= 55) {
if (negBalance < 64) {
if (AFD < 17) {
if (Rs < 683) {
return 4;
}
else if (Rs >= 683) {
return 8;
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (negBalance >= 64) {
return 8;
}
}
}
}
else if (refDCval >= 77) {
if (diffMVdiffVal < 124) {
if (negBalance < 59) {
if (mu_mv_mag_sq < 21) {
if (Rs < 748) {
return 1;
}
else if (Rs >= 748) {
if (Cs < 512) {
return 2;
}
else if (Cs >= 512) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (ssDCval < 131) {
if (ssDCval < 91) {
return 1;
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (ssDCval >= 131) {
return 1;
}
}
}
else if (negBalance >= 59) {
if (diffMVdiffVal < 10) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
else if (diffMVdiffVal >= 10) {
if (diffMVdiffVal < 39) {
if (diffTSC < -2) {
if (Cs < 478) {
return 1;
}
else if (Cs >= 478) {
return 2;
}
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (diffMVdiffVal >= 39) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 124) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 86) {
if (RsDiff < 9) {
if (Cs < 263) {
if (ssDCval < 140) {
if (refDCval < 79) {
if (negBalance < 35) {
if (diffRsCsdiff < 22) {
if (Cs < 233) {
return 1;
}
else if (Cs >= 233) {
return 2;
}
}
else if (diffRsCsdiff >= 22) {
return 2;
}
}
else if (negBalance >= 35) {
if (MVDiff < 206) {
if (tcor < 100) {
return 1;
}
else if (tcor >= 100) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (MVDiff >= 206) {
if (Rs < 911) {
return 8;
}
else if (Rs >= 911) {
return 1;
}
}
}
}
else if (refDCval >= 79) {
if (SC < 801) {
if (MVDiff < 50) {
return 1;
}
else if (MVDiff >= 50) {
if (MVDiff < 202) {
return 8;
}
else if (MVDiff >= 202) {
if (Rs < 336) {
return 1;
}
else if (Rs >= 336) {
return 2;
}
}
}
}
else if (SC >= 801) {
if (RsCsDiff < 28) {
if (mu_mv_mag_sq < 14) {
if (MVDiff < 26) {
if (tcor < 85) {
if (Rs < 837) {
return 1;
}
else if (Rs >= 837) {
return 2;
}
}
else if (tcor >= 85) {
return 1;
}
}
else if (MVDiff >= 26) {
if (diffRsCsdiff < 4) {
if (AFD < 11) {
if (Rs < 1459) {
return 1;
}
else if (Rs >= 1459) {
return 2;
}
}
else if (AFD >= 11) {
return 2;
}
}
else if (diffRsCsdiff >= 4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 14) {
return 1;
}
}
else if (RsCsDiff >= 28) {
if (refDCval < 102) {
if (MVDiff < 26) {
if (negBalance < 54) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (ssDCval < 91) {
return 1;
}
else if (ssDCval >= 91) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
return 2;
}
}
else if (MVDiff >= 26) {
if (diffRsCsdiff < -23) {
if (ssDCval < 96) {
return 2;
}
else if (ssDCval >= 96) {
return 1;
}
}
else if (diffRsCsdiff >= -23) {
if (posBalance < 90) {
return 1;
}
else if (posBalance >= 90) {
return 2;
}
}
}
}
else if (refDCval >= 102) {
if (negBalance < 82) {
return 1;
}
else if (negBalance >= 82) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (gchDC < 2) {
if (posBalance < 51) {
if (Cs < 194) {
return 2;
}
else if (Cs >= 194) {
return 1;
}
}
else if (posBalance >= 51) {
return 4;
}
}
else if (gchDC >= 2) {
if (ssDCval < 176) {
return 1;
}
else if (ssDCval >= 176) {
return 2;
}
}
}
}
else if (Cs >= 263) {
if (Rs < 826) {
if (RsCsDiff < 38) {
if (CsDiff < 3) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
if (negBalance < 52) {
if (negBalance < 49) {
if (Rs < 575) {
if (Rs < 517) {
return 8;
}
else if (Rs >= 517) {
return 2;
}
}
else if (Rs >= 575) {
if (ssDCval < 111) {
return 1;
}
else if (ssDCval >= 111) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (Cs < 301) {
return 1;
}
else if (Cs >= 301) {
return 4;
}
}
}
else if (negBalance >= 52) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
if (Cs < 487) {
if (mu_mv_mag_sq < 2) {
if (SC < 1100) {
return 1;
}
else if (SC >= 1100) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 483) {
return 2;
}
else if (Rs >= 483) {
if (diffTSC < -1) {
if (mu_mv_mag_sq < 27) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 2;
}
}
else if (mu_mv_mag_sq >= 27) {
return 8;
}
}
else if (diffTSC >= -1) {
if (diffRsCsdiff < -2) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
if (Rs < 727) {
return 1;
}
else if (Rs >= 727) {
return 8;
}
}
}
else if (diffRsCsdiff >= -2) {
if (mu_mv_mag_sq < 17) {
if (Cs < 302) {
return 1;
}
else if (Cs >= 302) {
return 4;
}
}
else if (mu_mv_mag_sq >= 17) {
if (Rs < 607) {
return 4;
}
else if (Rs >= 607) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 487) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (mu_mv_mag_sq < 9) {
return 4;
}
else if (mu_mv_mag_sq >= 9) {
if (diffRsCsdiff < 10) {
return 8;
}
else if (diffRsCsdiff >= 10) {
if (Rs < 675) {
return 8;
}
else if (Rs >= 675) {
return 2;
}
}
}
}
}
}
}
else if (RsCsDiff >= 38) {
if (ssDCval < 80) {
if (posBalance < 51) {
if (tcor < 79) {
if (Rs < 594) {
return 4;
}
else if (Rs >= 594) {
if (Rs < 757) {
return 1;
}
else if (Rs >= 757) {
return 2;
}
}
}
else if (tcor >= 79) {
return 8;
}
}
else if (posBalance >= 51) {
return 1;
}
}
else if (ssDCval >= 80) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (mu_mv_mag_sq < 51) {
return 8;
}
else if (mu_mv_mag_sq >= 51) {
return 1;
}
}
}
}
}
else if (Rs >= 826) {
if (Cs < 464) {
if (AFD < 10) {
if (ssDCval < 99) {
if (negBalance < 30) {
if (RsCsDiff < 23) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
else if (RsCsDiff >= 23) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (negBalance >= 30) {
if (tcor < 98) {
if (Cs < 383) {
return 1;
}
else if (Cs >= 383) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (AFD < 8) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (Rs < 1164) {
if (Rs < 905) {
return 1;
}
else if (Rs >= 905) {
return 2;
}
}
else if (Rs >= 1164) {
return 1;
}
}
}
else if (AFD >= 8) {
if (refDCval < 72) {
if (Rs < 1086) {
return 1;
}
else if (Rs >= 1086) {
return 4;
}
}
else if (refDCval >= 72) {
return 1;
}
}
}
}
}
else if (tcor >= 98) {
if (Rs < 1055) {
return 8;
}
else if (Rs >= 1055) {
return 4;
}
}
}
}
else if (ssDCval >= 99) {
if (negBalance < 38) {
if (TSC < 5) {
if (Cs < 427) {
return 8;
}
else if (Cs >= 427) {
return 2;
}
}
else if (TSC >= 5) {
return 1;
}
}
else if (negBalance >= 38) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
return 4;
}
}
}
}
}
else if (AFD >= 10) {
if (refDCval < 65) {
if (MVDiff < 19) {
if (tcor < 90) {
if (mu_mv_mag_sq < 31) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
else if (tcor >= 90) {
if (Rs < 1075) {
return 1;
}
else if (Rs >= 1075) {
return 8;
}
}
}
else if (MVDiff >= 19) {
if (diffRsCsdiff < -89) {
return 2;
}
else if (diffRsCsdiff >= -89) {
return 1;
}
}
}
else if (refDCval >= 65) {
if (mcTcor < 90) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
if (Rs < 1109) {
return 4;
}
else if (Rs >= 1109) {
return 1;
}
}
}
else if (mcTcor >= 90) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (SC < 1517) {
if (diffMVdiffVal < -13) {
if (diffRsCsdiff < -48) {
return 1;
}
else if (diffRsCsdiff >= -48) {
if (Rs < 974) {
return 8;
}
else if (Rs >= 974) {
return 2;
}
}
}
else if (diffMVdiffVal >= -13) {
if (SC < 1204) {
return 4;
}
else if (SC >= 1204) {
if (diffRsCsdiff < 19) {
return 1;
}
else if (diffRsCsdiff >= 19) {
return 4;
}
}
}
}
else if (SC >= 1517) {
if (RsCsDiff < 23) {
if (Rs < 1580) {
return 1;
}
else if (Rs >= 1580) {
return 8;
}
}
else if (RsCsDiff >= 23) {
if (diffMVdiffVal < -58) {
if (Rs < 1810) {
return 1;
}
else if (Rs >= 1810) {
return 8;
}
}
else if (diffMVdiffVal >= -58) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 464) {
if (refDCval < 72) {
if (negBalance < 30) {
if (MVDiff < 18) {
if (Rs < 1456) {
return 1;
}
else if (Rs >= 1456) {
return 4;
}
}
else if (MVDiff >= 18) {
return 2;
}
}
else if (negBalance >= 30) {
if (Cs < 524) {
if (MVDiff < 7) {
if (posBalance < 46) {
if (Rs < 1376) {
return 8;
}
else if (Rs >= 1376) {
if (Rs < 1879) {
return 4;
}
else if (Rs >= 1879) {
return 1;
}
}
}
else if (posBalance >= 46) {
return 2;
}
}
else if (MVDiff >= 7) {
if (MVDiff < 18) {
if (Cs < 509) {
return 1;
}
else if (Cs >= 509) {
if (RsDiff < 6) {
if (Rs < 1405) {
return 4;
}
else if (Rs >= 1405) {
return 1;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (MVDiff >= 18) {
if (TSC < 5) {
if (Rs < 1175) {
return 4;
}
else if (Rs >= 1175) {
return 1;
}
}
else if (TSC >= 5) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 4;
}
}
}
}
}
else if (Cs >= 524) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 4;
}
}
}
}
else if (refDCval >= 72) {
if (refDCval < 91) {
if (mcTcor < 93) {
if (negBalance < 57) {
return 1;
}
else if (negBalance >= 57) {
return 2;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
else if (refDCval >= 91) {
if (Rs < 1755) {
if (ssDCval < 103) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (ssDCval >= 103) {
if (diffMVdiffVal < -8) {
if (mu_mv_mag_sq < 2) {
return 2;
}
else if (mu_mv_mag_sq >= 2) {
return 1;
}
}
else if (diffMVdiffVal >= -8) {
return 2;
}
}
}
else if (Rs >= 1755) {
if (TSC < 5) {
if (Rs < 2233) {
return 2;
}
else if (Rs >= 2233) {
return 8;
}
}
else if (TSC >= 5) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
return 1;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 108) {
if (Rs < 2060) {
if (AFD < 12) {
if (posBalance < 64) {
if (TSC < 8) {
if (RsCsDiff < 82) {
return 2;
}
else if (RsCsDiff >= 82) {
return 1;
}
}
else if (TSC >= 8) {
return 2;
}
}
else if (posBalance >= 64) {
return 2;
}
}
else if (AFD >= 12) {
if (MVDiff < 39) {
if (diffMVdiffVal < -12) {
if (Cs < 244) {
if (Cs < 154) {
return 1;
}
else if (Cs >= 154) {
return 2;
}
}
else if (Cs >= 244) {
return 1;
}
}
else if (diffMVdiffVal >= -12) {
if (negBalance < 45) {
if (Cs < 491) {
if (Cs < 215) {
return 2;
}
else if (Cs >= 215) {
if (ssDCval < 56) {
return 2;
}
else if (ssDCval >= 56) {
return 1;
}
}
}
else if (Cs >= 491) {
return 2;
}
}
else if (negBalance >= 45) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
if (Rs < 1745) {
return 2;
}
else if (Rs >= 1745) {
return 1;
}
}
}
}
}
else if (MVDiff >= 39) {
if (diffMVdiffVal < 90) {
if (SC < 1704) {
if (mcTcor < 99) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
if (RsDiff < 16) {
return 1;
}
else if (RsDiff >= 16) {
return 2;
}
}
}
else if (mcTcor >= 99) {
return 2;
}
}
else if (SC >= 1704) {
if (mcTcor < 94) {
if (Cs < 207) {
return 1;
}
else if (Cs >= 207) {
if (Rs < 2035) {
if (SC < 2340) {
return 8;
}
else if (SC >= 2340) {
return 1;
}
}
else if (Rs >= 2035) {
return 1;
}
}
}
else if (mcTcor >= 94) {
if (RsCsDiff < 86) {
if (Rs < 1552) {
return 1;
}
else if (Rs >= 1552) {
return 8;
}
}
else if (RsCsDiff >= 86) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 90) {
return 2;
}
}
}
}
else if (Rs >= 2060) {
if (refDCval < 81) {
if (MVDiff < 53) {
return 2;
}
else if (MVDiff >= 53) {
if (Rs < 2230) {
return 4;
}
else if (Rs >= 2230) {
return 1;
}
}
}
else if (refDCval >= 81) {
if (negBalance < 44) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 1;
}
}
else if (negBalance >= 44) {
if (MVDiff < 36) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
return 1;
}
}
else if (MVDiff >= 36) {
if (ssDCval < 95) {
if (diffMVdiffVal < -9) {
if (Rs < 2165) {
return 1;
}
else if (Rs >= 2165) {
return 2;
}
}
else if (diffMVdiffVal >= -9) {
return 1;
}
}
else if (ssDCval >= 95) {
if (Cs < 379) {
return 8;
}
else if (Cs >= 379) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 108) {
if (CsDiff < 3) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (Rs < 2027) {
if (mcTcor < 94) {
if (diffMVdiffVal < 19) {
if (Rs < 1316) {
return 4;
}
else if (Rs >= 1316) {
return 8;
}
}
else if (diffMVdiffVal >= 19) {
return 1;
}
}
else if (mcTcor >= 94) {
if (ssDCval < 177) {
return 1;
}
else if (ssDCval >= 177) {
return 2;
}
}
}
else if (Rs >= 2027) {
if (ssDCval < 128) {
if (diffAFD < 2) {
if (posBalance < 65) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 2150) {
return 8;
}
else if (Rs >= 2150) {
return 1;
}
}
}
else if (posBalance >= 65) {
return 1;
}
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (ssDCval >= 128) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
if (AFD < 37) {
if (mu_mv_mag_sq < 67) {
if (TSC < 7) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (mu_mv_mag_sq >= 67) {
if (Cs < 358) {
return 1;
}
else if (Cs >= 358) {
return 2;
}
}
}
else if (AFD >= 37) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (negBalance < 53) {
if (RsDiff < 11) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (MVDiff < 3) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
return 4;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (negBalance >= 53) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 533) {
if (negBalance < 13) {
if (Rs < 703) {
if (ssDCval < 58) {
if (tcor < 98) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (tcor < 48) {
if (Rs < 328) {
return 8;
}
else if (Rs >= 328) {
return 2;
}
}
else if (tcor >= 48) {
return 1;
}
}
}
else if (tcor >= 98) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 594) {
return 1;
}
else if (Cs >= 594) {
if (Rs < 439) {
return 2;
}
else if (Rs >= 439) {
return 8;
}
}
}
}
}
else if (ssDCval >= 58) {
if (SC < 1393) {
if (Rs < 515) {
if (refDCval < 74) {
return 4;
}
else if (refDCval >= 74) {
if (refDCval < 123) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
return 1;
}
}
else if (refDCval >= 123) {
if (Cs < 605) {
if (Rs < 452) {
return 4;
}
else if (Rs >= 452) {
return 2;
}
}
else if (Cs >= 605) {
return 8;
}
}
}
}
else if (Rs >= 515) {
if (ssDCval < 135) {
if (SC < 1367) {
return 8;
}
else if (SC >= 1367) {
if (Rs < 636) {
return 2;
}
else if (Rs >= 636) {
if (Rs < 674) {
return 1;
}
else if (Rs >= 674) {
return 8;
}
}
}
}
else if (ssDCval >= 135) {
if (Rs < 664) {
return 2;
}
else if (Rs >= 664) {
return 1;
}
}
}
}
else if (SC >= 1393) {
if (Cs < 814) {
return 1;
}
else if (Cs >= 814) {
if (Rs < 287) {
return 4;
}
else if (Rs >= 287) {
return 8;
}
}
}
}
}
else if (Rs >= 703) {
if (diffTSC < -2) {
if (Rs < 1598) {
if (diffTSC < -6) {
if (ssDCval < 138) {
if (diffAFD < -16) {
if (diffMVdiffVal < -9) {
return 4;
}
else if (diffMVdiffVal >= -9) {
if (posBalance < 22) {
return 1;
}
else if (posBalance >= 22) {
if (Rs < 1029) {
return 8;
}
else if (Rs >= 1029) {
return 4;
}
}
}
}
else if (diffAFD >= -16) {
if (diffTSC < -7) {
return 8;
}
else if (diffTSC >= -7) {
if (MVDiff < 74) {
return 1;
}
else if (MVDiff >= 74) {
return 8;
}
}
}
}
else if (ssDCval >= 138) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (ssDCval < 147) {
return 4;
}
else if (ssDCval >= 147) {
return 8;
}
}
}
}
else if (diffTSC >= -6) {
if (tcor < 99) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Rs < 1274) {
return 4;
}
else if (Rs >= 1274) {
return 1;
}
}
}
else if (tcor >= 99) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
if (diffAFD < -7) {
if (ssDCval < 132) {
if (ssDCval < 127) {
return 8;
}
else if (ssDCval >= 127) {
if (Rs < 909) {
return 2;
}
else if (Rs >= 909) {
return 1;
}
}
}
else if (ssDCval >= 132) {
if (ssDCval < 141) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (ssDCval >= 141) {
return 8;
}
}
}
else if (diffAFD >= -7) {
return 4;
}
}
}
}
}
else if (Rs >= 1598) {
if (diffAFD < -20) {
return 4;
}
else if (diffAFD >= -20) {
return 8;
}
}
}
else if (diffTSC >= -2) {
if (SC < 2940) {
if (Rs < 1132) {
if (posBalance < 23) {
if (mu_mv_mag_sq < 126) {
if (diffMVdiffVal < -1) {
if (Rs < 992) {
return 2;
}
else if (Rs >= 992) {
if (Cs < 840) {
return 8;
}
else if (Cs >= 840) {
return 1;
}
}
}
else if (diffMVdiffVal >= -1) {
if (ssDCval < 147) {
if (Rs < 824) {
return 2;
}
else if (Rs >= 824) {
return 8;
}
}
else if (ssDCval >= 147) {
if (Rs < 828) {
return 8;
}
else if (Rs >= 828) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (posBalance < 9) {
if (ssDCval < 147) {
return 8;
}
else if (ssDCval >= 147) {
return 2;
}
}
else if (posBalance >= 9) {
if (Rs < 1116) {
return 8;
}
else if (Rs >= 1116) {
if (Rs < 1123) {
return 4;
}
else if (Rs >= 1123) {
return 8;
}
}
}
}
}
else if (posBalance >= 23) {
if (CsDiff < 3) {
if (posBalance < 29) {
if (negBalance < 11) {
if (diffRsCsdiff < -1) {
if (tcor < 100) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
if (Rs < 797) {
return 2;
}
else if (Rs >= 797) {
return 8;
}
}
}
else if (tcor >= 100) {
if (Rs < 915) {
return 2;
}
else if (Rs >= 915) {
return 4;
}
}
}
else if (diffRsCsdiff >= -1) {
if (ssDCval < 131) {
return 8;
}
else if (ssDCval >= 131) {
if (Rs < 816) {
return 1;
}
else if (Rs >= 816) {
return 8;
}
}
}
}
else if (negBalance >= 11) {
if (ssDCval < 106) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 774) {
return 2;
}
else if (Rs >= 774) {
return 8;
}
}
}
else if (ssDCval >= 106) {
if (Rs < 773) {
return 1;
}
else if (Rs >= 773) {
return 8;
}
}
}
}
else if (posBalance >= 29) {
if (Cs < 1154) {
if (diffTSC < -1) {
if (ssDCval < 73) {
if (Rs < 920) {
return 2;
}
else if (Rs >= 920) {
return 8;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (Cs >= 1154) {
if (Rs < 864) {
return 1;
}
else if (Rs >= 864) {
return 8;
}
}
}
}
else if (CsDiff >= 3) {
if (AFD < 15) {
if (posBalance < 122) {
return 1;
}
else if (posBalance >= 122) {
return 4;
}
}
else if (AFD >= 15) {
return 8;
}
}
}
}
else if (Rs >= 1132) {
if (negBalance < 8) {
if (Rs < 1134) {
return 1;
}
else if (Rs >= 1134) {
if (Cs < 1213) {
if (negBalance < 5) {
return 8;
}
else if (negBalance >= 5) {
if (TSCindex < 2) {
if (posBalance < 27) {
return 8;
}
else if (posBalance >= 27) {
if (Cs < 865) {
return 8;
}
else if (Cs >= 865) {
return 1;
}
}
}
else if (TSCindex >= 2) {
if (Rs < 1210) {
return 8;
}
else if (Rs >= 1210) {
return 1;
}
}
}
}
else if (Cs >= 1213) {
if (Rs < 1204) {
return 1;
}
else if (Rs >= 1204) {
return 8;
}
}
}
}
else if (negBalance >= 8) {
if (ssDCval < 56) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (Cs < 1152) {
if (ssDCval < 35) {
return 8;
}
else if (ssDCval >= 35) {
return 1;
}
}
else if (Cs >= 1152) {
return 8;
}
}
}
else if (ssDCval >= 56) {
if (ssDCval < 62) {
if (SC < 2867) {
return 1;
}
else if (SC >= 2867) {
if (Rs < 1874) {
return 4;
}
else if (Rs >= 1874) {
return 1;
}
}
}
else if (ssDCval >= 62) {
if (Rs < 1794) {
if (ssDCval < 66) {
if (Rs < 1333) {
return 1;
}
else if (Rs >= 1333) {
return 8;
}
}
else if (ssDCval >= 66) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1432) {
return 4;
}
else if (Rs >= 1432) {
return 8;
}
}
}
}
}
else if (Rs >= 1794) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Rs < 1945) {
return 8;
}
else if (Rs >= 1945) {
return 2;
}
}
}
}
}
}
}
}
else if (SC >= 2940) {
if (refDCval < 71) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (Rs < 2247) {
return 4;
}
else if (Rs >= 2247) {
if (SC < 3567) {
return 8;
}
else if (SC >= 3567) {
return 2;
}
}
}
}
else if (refDCval >= 71) {
if (ssDCval < 81) {
return 1;
}
else if (ssDCval >= 81) {
if (Cs < 1300) {
return 8;
}
else if (Cs >= 1300) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 45) {
if (refDCval < 47) {
if (Rs < 1739) {
if (refDCval < 29) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (refDCval >= 29) {
if (mcTcor < 92) {
if (Rs < 643) {
if (Cs < 613) {
return 8;
}
else if (Cs >= 613) {
return 1;
}
}
else if (Rs >= 643) {
if (RsCsDiff < 11) {
if (Rs < 1197) {
return 1;
}
else if (Rs >= 1197) {
return 8;
}
}
else if (RsCsDiff >= 11) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 1424) {
if (negBalance < 36) {
if (diffRsCsdiff < 4) {
if (Rs < 722) {
if (Cs < 1009) {
if (Cs < 995) {
return 4;
}
else if (Cs >= 995) {
return 8;
}
}
else if (Cs >= 1009) {
return 1;
}
}
else if (Rs >= 722) {
if (mu_mv_mag_sq < 78) {
if (Rs < 744) {
return 8;
}
else if (Rs >= 744) {
return 2;
}
}
else if (mu_mv_mag_sq >= 78) {
return 8;
}
}
}
else if (diffRsCsdiff >= 4) {
if (Cs < 538) {
return 2;
}
else if (Cs >= 538) {
return 1;
}
}
}
else if (negBalance >= 36) {
return 8;
}
}
else if (Rs >= 1424) {
if (Cs < 926) {
return 4;
}
else if (Cs >= 926) {
return 1;
}
}
}
}
}
else if (Rs >= 1739) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (Rs < 1993) {
return 4;
}
else if (Rs >= 1993) {
return 1;
}
}
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 86) {
if (Cs < 844) {
if (negBalance < 61) {
if (Rs < 1128) {
if (TSC < 5) {
if (ssDCval < 77) {
if (tcor < 92) {
if (Cs < 738) {
if (TSC < 2) {
if (Rs < 629) {
return 8;
}
else if (Rs >= 629) {
return 4;
}
}
else if (TSC >= 2) {
if (RsCsDiff < 3) {
if (Rs < 782) {
return 8;
}
else if (Rs >= 782) {
return 1;
}
}
else if (RsCsDiff >= 3) {
return 1;
}
}
}
else if (Cs >= 738) {
if (Cs < 787) {
return 4;
}
else if (Cs >= 787) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
return 1;
}
}
}
}
else if (tcor >= 92) {
if (negBalance < 35) {
if (mu_mv_mag_sq < 82) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 1049) {
return 4;
}
else if (Rs >= 1049) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 82) {
return 1;
}
}
else if (negBalance >= 35) {
if (Rs < 445) {
return 4;
}
else if (Rs >= 445) {
return 8;
}
}
}
}
else if (ssDCval >= 77) {
if (posBalance < 36) {
return 8;
}
else if (posBalance >= 36) {
return 2;
}
}
}
else if (TSC >= 5) {
if (Rs < 815) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
return 2;
}
}
else if (Rs >= 815) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
else if (Rs >= 1128) {
if (tcor < 85) {
if (negBalance < 35) {
if (Cs < 689) {
return 1;
}
else if (Cs >= 689) {
return 2;
}
}
else if (negBalance >= 35) {
if (RsCsDiff < 60) {
return 4;
}
else if (RsCsDiff >= 60) {
if (Cs < 802) {
return 1;
}
else if (Cs >= 802) {
if (Rs < 1783) {
return 8;
}
else if (Rs >= 1783) {
return 4;
}
}
}
}
}
else if (tcor >= 85) {
if (Cs < 766) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (Rs < 1415) {
if (Cs < 616) {
return 2;
}
else if (Cs >= 616) {
if (Rs < 1240) {
return 1;
}
else if (Rs >= 1240) {
return 8;
}
}
}
else if (Rs >= 1415) {
if (AFD < 12) {
if (negBalance < 28) {
return 2;
}
else if (negBalance >= 28) {
return 1;
}
}
else if (AFD >= 12) {
return 2;
}
}
}
}
else if (Cs >= 766) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
if (RsDiff < 5) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (RsDiff >= 5) {
if (Rs < 2247) {
return 1;
}
else if (Rs >= 2247) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 61) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (ssDCval < 88) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (ssDCval >= 88) {
return 2;
}
}
}
}
else if (Cs >= 844) {
if (tcor < 90) {
if (Rs < 1954) {
if (mcTcor < 89) {
if (CsDiff < 5) {
if (MVDiff < 13) {
if (Rs < 1286) {
return 4;
}
else if (Rs >= 1286) {
return 8;
}
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (mcTcor >= 89) {
if (Rs < 1528) {
if (Cs < 1107) {
if (ssDCval < 79) {
if (Cs < 1072) {
return 8;
}
else if (Cs >= 1072) {
return 4;
}
}
else if (ssDCval >= 79) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
return 8;
}
}
}
else if (Cs >= 1107) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (Rs >= 1528) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (mu_mv_mag_sq < 52) {
return 2;
}
else if (mu_mv_mag_sq >= 52) {
return 8;
}
}
}
}
}
else if (Rs >= 1954) {
if (negBalance < 44) {
if (AFD < 12) {
if (diffRsCsdiff < -21) {
if (Cs < 977) {
return 4;
}
else if (Cs >= 977) {
return 2;
}
}
else if (diffRsCsdiff >= -21) {
return 4;
}
}
else if (AFD >= 12) {
return 2;
}
}
else if (negBalance >= 44) {
if (Rs < 2067) {
return 2;
}
else if (Rs >= 2067) {
return 8;
}
}
}
}
else if (tcor >= 90) {
if (negBalance < 57) {
if (Cs < 969) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (posBalance < 43) {
if (MVDiff < 3) {
if (Rs < 679) {
return 4;
}
else if (Rs >= 679) {
return 2;
}
}
else if (MVDiff >= 3) {
if (Rs < 1084) {
return 2;
}
else if (Rs >= 1084) {
if (diffMVdiffVal < -1) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
}
}
else if (posBalance >= 43) {
if (diffAFD < -2) {
if (Rs < 1764) {
return 8;
}
else if (Rs >= 1764) {
return 2;
}
}
else if (diffAFD >= -2) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 4;
}
}
}
}
}
}
else if (Cs >= 969) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (ssDCval < 76) {
if (Cs < 986) {
return 8;
}
else if (Cs >= 986) {
return 4;
}
}
else if (ssDCval >= 76) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
if (Cs < 1050) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (Cs >= 1050) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 57) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (diffRsCsdiff < -4) {
return 8;
}
else if (diffRsCsdiff >= -4) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (Rs < 857) {
if (Rs < 426) {
if (Rs < 356) {
return 2;
}
else if (Rs >= 356) {
return 1;
}
}
else if (Rs >= 426) {
if (ssDCval < 127) {
if (ssDCval < 109) {
if (TSC < 3) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Rs < 615) {
return 2;
}
else if (Rs >= 615) {
return 4;
}
}
}
else if (TSC >= 3) {
if (Cs < 781) {
if (SC < 1542) {
return 2;
}
else if (SC >= 1542) {
return 8;
}
}
else if (Cs >= 781) {
if (Rs < 745) {
return 8;
}
else if (Rs >= 745) {
if (SC < 1637) {
if (Rs < 785) {
return 8;
}
else if (Rs >= 785) {
return 1;
}
}
else if (SC >= 1637) {
return 4;
}
}
}
}
}
else if (ssDCval >= 109) {
return 4;
}
}
else if (ssDCval >= 127) {
if (Rs < 843) {
return 8;
}
else if (Rs >= 843) {
return 1;
}
}
}
}
else if (Rs >= 857) {
if (ssDCval < 56) {
if (Cs < 603) {
if (SC < 1589) {
return 8;
}
else if (SC >= 1589) {
return 1;
}
}
else if (Cs >= 603) {
if (mcTcor < 94) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 8;
}
}
else if (mcTcor >= 94) {
if (negBalance < 34) {
if (ssDCval < 52) {
if (Cs < 1163) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 1163) {
return 4;
}
}
else if (ssDCval >= 52) {
return 1;
}
}
else if (negBalance >= 34) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
return 8;
}
}
}
}
}
else if (ssDCval >= 56) {
if (tcor < 75) {
if (Rs < 1538) {
return 4;
}
else if (Rs >= 1538) {
return 2;
}
}
else if (tcor >= 75) {
if (CsDiff < 6) {
if (Rs < 1715) {
if (diffAFD < -12) {
return 4;
}
else if (diffAFD >= -12) {
if (negBalance < 14) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (negBalance >= 14) {
if (Cs < 1179) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (AFD < 4) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 1;
}
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (Cs >= 1179) {
if (Rs < 1142) {
return 4;
}
else if (Rs >= 1142) {
return 8;
}
}
}
}
}
else if (Rs >= 1715) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (Cs < 952) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Cs < 860) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (Cs < 819) {
if (Rs < 2097) {
return 8;
}
else if (Rs >= 2097) {
return 8;
}
}
else if (Cs >= 819) {
return 4;
}
}
}
else if (Cs >= 860) {
return 2;
}
}
}
else if (Cs >= 952) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (Rs < 1826) {
return 1;
}
else if (Rs >= 1826) {
return 2;
}
}
}
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 407) {
if (ssDCval < 96) {
if (ssDCval < 82) {
if (ssDCval < 27) {
return 8;
}
else if (ssDCval >= 27) {
if (ssDCval < 60) {
if (MVDiff < 12) {
if (mu_mv_mag_sq < 91) {
return 2;
}
else if (mu_mv_mag_sq >= 91) {
if (Cs < 989) {
return 1;
}
else if (Cs >= 989) {
return 2;
}
}
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (ssDCval >= 60) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Cs < 1016) {
if (TSC < 6) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (Cs >= 1016) {
return 1;
}
}
}
}
}
else if (ssDCval >= 82) {
if (CsDiff < 3) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (ssDCval >= 96) {
if (Cs < 719) {
if (Cs < 625) {
return 2;
}
else if (Cs >= 625) {
return 1;
}
}
else if (Cs >= 719) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (Rs >= 407) {
if (mu_mv_mag_sq < 56) {
if (refDCval < 35) {
return 8;
}
else if (refDCval >= 35) {
if (negBalance < 44) {
if (AFD < 8) {
if (ssDCval < 151) {
if (negBalance < 32) {
if (tcor < 89) {
if (Rs < 689) {
if (SC < 1407) {
return 2;
}
else if (SC >= 1407) {
return 1;
}
}
else if (Rs >= 689) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
else if (tcor >= 89) {
if (Rs < 922) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (Rs < 583) {
return 8;
}
else if (Rs >= 583) {
return 1;
}
}
}
else if (Rs >= 922) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
return 1;
}
}
}
}
}
else if (negBalance >= 32) {
if (ssDCval < 137) {
if (MVDiff < 5) {
if (ssDCval < 84) {
if (refDCval < 67) {
if (AFD < 7) {
if (mu_mv_mag_sq < 52) {
return 8;
}
else if (mu_mv_mag_sq >= 52) {
return 1;
}
}
else if (AFD >= 7) {
if (Cs < 897) {
if (Rs < 1609) {
return 4;
}
else if (Rs >= 1609) {
return 2;
}
}
else if (Cs >= 897) {
if (Cs < 1029) {
return 8;
}
else if (Cs >= 1029) {
return 1;
}
}
}
}
else if (refDCval >= 67) {
if (Cs < 951) {
if (Rs < 643) {
return 4;
}
else if (Rs >= 643) {
return 1;
}
}
else if (Cs >= 951) {
if (Rs < 1276) {
return 8;
}
else if (Rs >= 1276) {
return 4;
}
}
}
}
else if (ssDCval >= 84) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (RsCsDiff < 9) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 2;
}
}
else if (RsCsDiff >= 9) {
if (ssDCval < 83) {
if (ssDCval < 70) {
if (Rs < 865) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
return 8;
}
}
}
else if (Rs >= 865) {
if (Cs < 741) {
if (ssDCval < 39) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
else if (ssDCval >= 39) {
if (SC < 1648) {
return 1;
}
else if (SC >= 1648) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (Cs >= 741) {
if (posBalance < 64) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 1894) {
return 4;
}
else if (Rs >= 1894) {
return 8;
}
}
}
else if (posBalance >= 64) {
return 4;
}
}
}
}
else if (ssDCval >= 70) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (Rs < 1303) {
return 2;
}
else if (Rs >= 1303) {
return 4;
}
}
}
}
else if (ssDCval >= 83) {
if (Cs < 646) {
return 1;
}
else if (Cs >= 646) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
if (posBalance < 57) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 4;
}
}
else if (posBalance >= 57) {
if (mu_mv_mag_sq < 53) {
return 4;
}
else if (mu_mv_mag_sq >= 53) {
return 8;
}
}
}
}
}
else if (ssDCval >= 151) {
return 2;
}
}
else if (AFD >= 8) {
if (negBalance < 36) {
if (gchDC < 2) {
if (negBalance < 34) {
if (Cs < 951) {
if (AFD < 9) {
if (Rs < 1446) {
return 4;
}
else if (Rs >= 1446) {
return 1;
}
}
else if (AFD >= 9) {
if (CsDiff < 5) {
if (Rs < 2016) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (Rs >= 2016) {
return 4;
}
}
else if (CsDiff >= 5) {
if (Rs < 1247) {
return 1;
}
else if (Rs >= 1247) {
return 2;
}
}
}
}
else if (Cs >= 951) {
if (refDCval < 137) {
if (SC < 2489) {
return 1;
}
else if (SC >= 2489) {
if (Rs < 1515) {
return 2;
}
else if (Rs >= 1515) {
return 8;
}
}
}
else if (refDCval >= 137) {
return 2;
}
}
}
else if (negBalance >= 34) {
if (RsCsDiff < 73) {
if (ssDCval < 54) {
if (Rs < 1492) {
return 8;
}
else if (Rs >= 1492) {
return 4;
}
}
else if (ssDCval >= 54) {
if (tcor < 82) {
return 1;
}
else if (tcor >= 82) {
if (Cs < 924) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Cs < 589) {
return 1;
}
else if (Cs >= 589) {
return 2;
}
}
}
else if (Cs >= 924) {
if (Rs < 725) {
return 2;
}
else if (Rs >= 725) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 73) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 115) {
if (Cs < 967) {
if (Rs < 1253) {
if (tcor < 58) {
if (Cs < 793) {
return 1;
}
else if (Cs >= 793) {
return 2;
}
}
else if (tcor >= 58) {
if (MVDiff < 118) {
return 8;
}
else if (MVDiff >= 118) {
return 1;
}
}
}
else if (Rs >= 1253) {
if (ssDCval < 64) {
if (Cs < 814) {
return 2;
}
else if (Cs >= 814) {
return 8;
}
}
else if (ssDCval >= 64) {
if (RsCsDiff < 74) {
return 1;
}
else if (RsCsDiff >= 74) {
return 8;
}
}
}
}
else if (Cs >= 967) {
if (Rs < 2206) {
if (tcor < 58) {
return 2;
}
else if (tcor >= 58) {
return 8;
}
}
else if (Rs >= 2206) {
return 2;
}
}
}
else if (ssDCval >= 115) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (Cs < 695) {
if (posBalance < 53) {
if (Rs < 716) {
return 2;
}
else if (Rs >= 716) {
if (SC < 2909) {
return 1;
}
else if (SC >= 2909) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (ssDCval < 78) {
if (RsCsDiff < 19) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
if (Rs < 1006) {
return 8;
}
else if (Rs >= 1006) {
return 1;
}
}
}
else if (RsCsDiff >= 19) {
if (SC < 1160) {
return 1;
}
else if (SC >= 1160) {
return 2;
}
}
}
else if (ssDCval >= 78) {
if (Rs < 1309) {
return 8;
}
else if (Rs >= 1309) {
if (Rs < 1638) {
return 1;
}
else if (Rs >= 1638) {
return 8;
}
}
}
}
}
else if (Cs >= 695) {
if (Rs < 838) {
if (Cs < 932) {
if (diffMVdiffVal < 4) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (mcTcor < 92) {
if (Rs < 649) {
return 2;
}
else if (Rs >= 649) {
return 8;
}
}
else if (mcTcor >= 92) {
return 2;
}
}
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
else if (Cs >= 932) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
if (Rs < 792) {
return 4;
}
else if (Rs >= 792) {
return 2;
}
}
}
}
else if (Rs >= 838) {
if (ssDCval < 53) {
if (mu_mv_mag_sq < 35) {
return 8;
}
else if (mu_mv_mag_sq >= 35) {
if (gchDC < 1) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (ssDCval >= 53) {
if (tcor < 76) {
if (diffRsCsdiff < 5) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (diffRsCsdiff >= 5) {
return 8;
}
}
else if (tcor >= 76) {
if (Cs < 875) {
if (mcTcor < 90) {
if (Rs < 1429) {
if (Rs < 1183) {
return 2;
}
else if (Rs >= 1183) {
return 8;
}
}
else if (Rs >= 1429) {
return 4;
}
}
else if (mcTcor >= 90) {
if (TSC < 6) {
if (mu_mv_mag_sq < 9) {
if (diffMVdiffVal < 0) {
return 4;
}
else if (diffMVdiffVal >= 0) {
return 1;
}
}
else if (mu_mv_mag_sq >= 9) {
return 8;
}
}
else if (TSC >= 6) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
if (AFD < 12) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (Cs < 741) {
return 1;
}
else if (Cs >= 741) {
return 8;
}
}
}
else if (AFD >= 12) {
return 2;
}
}
}
}
}
else if (Cs >= 875) {
if (Rs < 1286) {
if (posBalance < 62) {
if (ssDCval < 58) {
if (Rs < 1172) {
return 4;
}
else if (Rs >= 1172) {
return 8;
}
}
else if (ssDCval >= 58) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (TSC < 7) {
if (refDCval < 57) {
if (Rs < 1056) {
return 8;
}
else if (Rs >= 1056) {
return 1;
}
}
else if (refDCval >= 57) {
return 2;
}
}
else if (TSC >= 7) {
if (Cs < 1143) {
return 8;
}
else if (Cs >= 1143) {
return 4;
}
}
}
}
}
else if (posBalance >= 62) {
if (SC < 2132) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
else if (SC >= 2132) {
return 4;
}
}
}
else if (Rs >= 1286) {
if (diffAFD < 3) {
if (diffRsCsdiff < -1) {
if (Cs < 1060) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (Cs >= 1060) {
if (Rs < 2082) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 2;
}
}
else if (Rs >= 2082) {
return 4;
}
}
}
else if (diffRsCsdiff >= -1) {
if (SC < 2411) {
return 4;
}
else if (SC >= 2411) {
if (ssDCval < 159) {
return 8;
}
else if (ssDCval >= 159) {
return 2;
}
}
}
}
else if (diffAFD >= 3) {
if (Cs < 1078) {
return 8;
}
else if (Cs >= 1078) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (negBalance >= 44) {
if (Cs < 766) {
if (RsCsDiff < 27) {
if (Rs < 892) {
if (MVDiff < 7) {
if (ssDCval < 71) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 50) {
if (ssDCval < 44) {
if (Rs < 728) {
return 4;
}
else if (Rs >= 728) {
return 8;
}
}
else if (ssDCval >= 44) {
if (Cs < 742) {
return 1;
}
else if (Cs >= 742) {
return 8;
}
}
}
else if (ssDCval >= 50) {
if (Cs < 561) {
return 1;
}
else if (Cs >= 561) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 758) {
return 4;
}
else if (Rs >= 758) {
return 8;
}
}
}
else if (ssDCval >= 71) {
if (ssDCval < 89) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
else if (ssDCval >= 89) {
if (MVDiff < 5) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (refDCval < 111) {
return 1;
}
else if (refDCval >= 111) {
return 2;
}
}
}
else if (MVDiff >= 5) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 78) {
if (Rs < 806) {
if (diffMVdiffVal < -15) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 4;
}
}
else if (diffMVdiffVal >= -15) {
if (SC < 1226) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (MVDiff < 52) {
if (negBalance < 48) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (negBalance >= 48) {
return 8;
}
}
else if (MVDiff >= 52) {
return 2;
}
}
}
else if (SC >= 1226) {
if (Rs < 636) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
return 8;
}
}
else if (Rs >= 636) {
if (MVDiff < 12) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
}
}
}
else if (Rs >= 806) {
if (ssDCval < 51) {
if (tcor < 70) {
return 1;
}
else if (tcor >= 70) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (ssDCval >= 51) {
if (AFD < 9) {
if (Rs < 855) {
return 1;
}
else if (Rs >= 855) {
return 8;
}
}
else if (AFD >= 9) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 142) {
if (mcTcor < 80) {
if (Rs < 742) {
return 2;
}
else if (Rs >= 742) {
return 8;
}
}
else if (mcTcor >= 80) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
if (Rs < 645) {
return 2;
}
else if (Rs >= 645) {
return 8;
}
}
}
}
else if (ssDCval >= 142) {
return 2;
}
}
}
}
else if (Rs >= 892) {
if (mcTcor < 80) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
return 1;
}
}
else if (mcTcor >= 80) {
if (tcor < 91) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (diffRsCsdiff < -15) {
if (Rs < 1045) {
return 2;
}
else if (Rs >= 1045) {
return 8;
}
}
else if (diffRsCsdiff >= -15) {
return 8;
}
}
}
else if (tcor >= 91) {
if (ssDCval < 72) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (Rs < 1041) {
return 1;
}
else if (Rs >= 1041) {
return 4;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 1007) {
return 4;
}
else if (Rs >= 1007) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 27) {
if (Rs < 2142) {
if (Cs < 576) {
if (tcor < 69) {
return 4;
}
else if (tcor >= 69) {
if (mcTcor < 94) {
if (Cs < 543) {
return 1;
}
else if (Cs >= 543) {
if (Cs < 559) {
if (Cs < 545) {
return 2;
}
else if (Cs >= 545) {
return 8;
}
}
else if (Cs >= 559) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 909) {
return 4;
}
else if (Rs >= 909) {
return 1;
}
}
}
}
else if (Cs >= 576) {
if (negBalance < 55) {
if (posBalance < 52) {
if (MVDiff < 12) {
if (Rs < 1397) {
return 2;
}
else if (Rs >= 1397) {
return 1;
}
}
else if (MVDiff >= 12) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (posBalance >= 52) {
if (TSC < 7) {
if (SC < 1650) {
if (negBalance < 49) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (negBalance >= 49) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (Rs < 854) {
return 8;
}
else if (Rs >= 854) {
return 4;
}
}
}
}
else if (SC >= 1650) {
if (mcTcor < 96) {
if (Cs < 643) {
if (negBalance < 51) {
if (mu_mv_mag_sq < 5) {
return 2;
}
else if (mu_mv_mag_sq >= 5) {
return 8;
}
}
else if (negBalance >= 51) {
return 8;
}
}
else if (Cs >= 643) {
if (diffAFD < 6) {
if (Rs < 1791) {
if (Rs < 996) {
if (Rs < 963) {
return 8;
}
else if (Rs >= 963) {
return 2;
}
}
else if (Rs >= 996) {
return 8;
}
}
else if (Rs >= 1791) {
return 2;
}
}
else if (diffAFD >= 6) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1395) {
return 4;
}
else if (Rs >= 1395) {
if (Rs < 1816) {
return 1;
}
else if (Rs >= 1816) {
return 2;
}
}
}
}
}
else if (TSC >= 7) {
if (posBalance < 62) {
if (MVDiff < 18) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Rs < 1608) {
return 8;
}
else if (Rs >= 1608) {
return 1;
}
}
}
else if (MVDiff >= 18) {
if (mcTcor < 86) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (mcTcor >= 86) {
if (Cs < 659) {
return 4;
}
else if (Cs >= 659) {
if (Rs < 990) {
return 4;
}
else if (Rs >= 990) {
if (ssDCval < 128) {
return 1;
}
else if (ssDCval >= 128) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 62) {
return 4;
}
}
}
}
else if (negBalance >= 55) {
if (negBalance < 58) {
return 8;
}
else if (negBalance >= 58) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
}
else if (Rs >= 2142) {
return 2;
}
}
}
else if (Cs >= 766) {
if (posBalance < 55) {
if (AFD < 10) {
if (MVDiff < 6) {
if (posBalance < 52) {
if (Rs < 1755) {
return 8;
}
else if (Rs >= 1755) {
return 2;
}
}
else if (posBalance >= 52) {
if (MVDiff < 3) {
if (Rs < 1409) {
return 4;
}
else if (Rs >= 1409) {
if (Rs < 1996) {
return 8;
}
else if (Rs >= 1996) {
return 1;
}
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 57) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (Cs < 840) {
if (ssDCval < 136) {
if (ssDCval < 101) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (Rs < 1122) {
return 8;
}
else if (Rs >= 1122) {
return 4;
}
}
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (ssDCval >= 136) {
return 4;
}
}
else if (Cs >= 840) {
if (Cs < 1128) {
if (mu_mv_mag_sq < 52) {
if (ssDCval < 147) {
if (Rs < 1521) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
else if (Rs >= 1521) {
if (Rs < 1776) {
if (Rs < 1747) {
return 8;
}
else if (Rs >= 1747) {
return 2;
}
}
else if (Rs >= 1776) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
}
}
else if (ssDCval >= 147) {
if (Rs < 1205) {
return 2;
}
else if (Rs >= 1205) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 52) {
return 8;
}
}
else if (Cs >= 1128) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (RsDiff < 6) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
}
}
else if (AFD >= 10) {
if (MVDiff < 10) {
if (RsDiff < 8) {
if (mu_mv_mag_sq < 45) {
if (diffRsCsdiff < -22) {
return 1;
}
else if (diffRsCsdiff >= -22) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (ssDCval < 72) {
if (Rs < 1650) {
return 8;
}
else if (Rs >= 1650) {
return 1;
}
}
else if (ssDCval >= 72) {
if (diffRsCsdiff < 11) {
return 8;
}
else if (diffRsCsdiff >= 11) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (Cs < 1294) {
return 1;
}
else if (Cs >= 1294) {
return 2;
}
}
}
else if (RsDiff >= 8) {
if (Rs < 2142) {
return 2;
}
else if (Rs >= 2142) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (negBalance < 54) {
if (ssDCval < 86) {
if (mu_mv_mag_sq < 28) {
if (tcor < 88) {
if (diffMVdiffVal < -7) {
return 8;
}
else if (diffMVdiffVal >= -7) {
if (MVDiff < 21) {
if (RsDiff < 8) {
if (SC < 2574) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (SC >= 2574) {
return 1;
}
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (MVDiff >= 21) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 2;
}
}
}
}
else if (tcor >= 88) {
if (Cs < 832) {
return 1;
}
else if (Cs >= 832) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 28) {
if (Cs < 783) {
return 1;
}
else if (Cs >= 783) {
if (SC < 3044) {
if (diffMVdiffVal < 4) {
return 2;
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
else if (SC >= 3044) {
return 8;
}
}
}
}
else if (ssDCval >= 86) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
}
else if (negBalance >= 54) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 51) {
if (Rs < 1567) {
return 8;
}
else if (Rs >= 1567) {
return 2;
}
}
else if (posBalance >= 51) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 784) {
return 4;
}
else if (Rs >= 784) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 55) {
if (Rs < 1019) {
if (AFD < 11) {
if (Cs < 1141) {
if (ssDCval < 51) {
if (TSC < 6) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 878) {
return 2;
}
else if (Rs >= 878) {
return 1;
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (ssDCval >= 51) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
if (mcTcor < 95) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
return 4;
}
}
else if (mcTcor >= 95) {
if (Cs < 911) {
if (Rs < 793) {
return 2;
}
else if (Rs >= 793) {
return 1;
}
}
else if (Cs >= 911) {
return 4;
}
}
}
}
}
else if (Cs >= 1141) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 793) {
return 8;
}
else if (Rs >= 793) {
return 1;
}
}
}
}
else if (AFD >= 11) {
if (Cs < 847) {
if (Rs < 874) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 799) {
return 2;
}
else if (Rs >= 799) {
if (Rs < 843) {
return 8;
}
else if (Rs >= 843) {
return 4;
}
}
}
}
else if (Rs >= 874) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 4;
}
}
}
else if (Cs >= 847) {
if (negBalance < 49) {
if (posBalance < 62) {
if (Rs < 980) {
return 2;
}
else if (Rs >= 980) {
return 1;
}
}
else if (posBalance >= 62) {
if (Rs < 910) {
return 4;
}
else if (Rs >= 910) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (tcor < 90) {
if (TSC < 8) {
if (ssDCval < 106) {
if (SC < 1744) {
return 4;
}
else if (SC >= 1744) {
if (Cs < 1122) {
return 1;
}
else if (Cs >= 1122) {
return 4;
}
}
}
else if (ssDCval >= 106) {
return 8;
}
}
else if (TSC >= 8) {
return 4;
}
}
else if (tcor >= 90) {
if (Rs < 944) {
return 1;
}
else if (Rs >= 944) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1019) {
if (RsCsDiff < 55) {
if (refDCval < 90) {
if (mcTcor < 92) {
if (Rs < 1147) {
if (Cs < 972) {
return 8;
}
else if (Cs >= 972) {
if (mcTcor < 88) {
if (Rs < 1064) {
return 8;
}
else if (Rs >= 1064) {
return 1;
}
}
else if (mcTcor >= 88) {
return 8;
}
}
}
else if (Rs >= 1147) {
return 8;
}
}
else if (mcTcor >= 92) {
if (negBalance < 48) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (Cs < 1006) {
if (Rs < 1377) {
return 8;
}
else if (Rs >= 1377) {
if (Rs < 1526) {
return 4;
}
else if (Rs >= 1526) {
return 8;
}
}
}
else if (Cs >= 1006) {
if (refDCval < 75) {
return 4;
}
else if (refDCval >= 75) {
if (Rs < 1344) {
return 2;
}
else if (Rs >= 1344) {
return 8;
}
}
}
}
}
else if (negBalance >= 48) {
if (tcor < 73) {
return 4;
}
else if (tcor >= 73) {
return 8;
}
}
}
}
else if (refDCval >= 90) {
if (ssDCval < 99) {
if (AFD < 10) {
if (posBalance < 60) {
if (tcor < 76) {
if (Rs < 1576) {
return 8;
}
else if (Rs >= 1576) {
return 2;
}
}
else if (tcor >= 76) {
if (diffMVdiffVal < 4) {
return 2;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 1499) {
return 2;
}
else if (Rs >= 1499) {
return 8;
}
}
}
}
else if (posBalance >= 60) {
return 8;
}
}
else if (AFD >= 10) {
if (RsDiff < 6) {
if (MVDiff < 8) {
if (posBalance < 62) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
else if (posBalance >= 62) {
return 8;
}
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (RsDiff >= 6) {
if (tcor < 88) {
if (Cs < 1262) {
return 4;
}
else if (Cs >= 1262) {
return 2;
}
}
else if (tcor >= 88) {
return 8;
}
}
}
}
else if (ssDCval >= 99) {
if (MVDiff < 15) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (Cs < 1076) {
return 8;
}
else if (Cs >= 1076) {
return 1;
}
}
}
else if (MVDiff >= 15) {
if (SC < 2136) {
if (ssDCval < 111) {
return 8;
}
else if (ssDCval >= 111) {
if (Cs < 857) {
return 1;
}
else if (Cs >= 857) {
return 4;
}
}
}
else if (SC >= 2136) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 55) {
if (Cs < 921) {
if (Rs < 1469) {
return 1;
}
else if (Rs >= 1469) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (Cs >= 921) {
if (Rs < 1411) {
if (MVDiff < 8) {
if (Rs < 1097) {
return 8;
}
else if (Rs >= 1097) {
return 4;
}
}
else if (MVDiff >= 8) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
}
else if (Rs >= 1411) {
if (TSC < 11) {
if (diffAFD < 1) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
else if (diffAFD >= 1) {
if (tcor < 84) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 2;
}
}
else if (tcor >= 84) {
if (MVDiff < 2) {
if (Rs < 1799) {
return 8;
}
else if (Rs >= 1799) {
return 1;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (Rs < 1596) {
if (TSC < 6) {
if (ssDCval < 78) {
if (ssDCval < 59) {
if (Rs < 851) {
if (ssDCval < 58) {
if (ssDCval < 37) {
if (Cs < 724) {
return 8;
}
else if (Cs >= 724) {
return 1;
}
}
else if (ssDCval >= 37) {
if (Rs < 581) {
if (Cs < 885) {
if (MVDiff < 11) {
if (Cs < 625) {
return 8;
}
else if (Cs >= 625) {
return 4;
}
}
else if (MVDiff >= 11) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 1;
}
}
}
else if (Cs >= 885) {
return 1;
}
}
else if (Rs >= 581) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 40) {
return 8;
}
else if (ssDCval >= 40) {
if (diffAFD < 2) {
if (refDCval < 57) {
if (mu_mv_mag_sq < 87) {
return 4;
}
else if (mu_mv_mag_sq >= 87) {
if (ssDCval < 52) {
if (Rs < 697) {
return 1;
}
else if (Rs >= 697) {
return 8;
}
}
else if (ssDCval >= 52) {
if (Cs < 561) {
return 8;
}
else if (Cs >= 561) {
return 4;
}
}
}
}
else if (refDCval >= 57) {
if (Cs < 691) {
if (Rs < 731) {
return 4;
}
else if (Rs >= 731) {
return 1;
}
}
else if (Cs >= 691) {
return 8;
}
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
}
}
else if (ssDCval >= 58) {
if (Rs < 532) {
return 2;
}
else if (Rs >= 532) {
return 8;
}
}
}
else if (Rs >= 851) {
if (tcor < 89) {
if (ssDCval < 44) {
if (Cs < 577) {
if (Cs < 574) {
return 8;
}
else if (Cs >= 574) {
return 1;
}
}
else if (Cs >= 577) {
return 8;
}
}
else if (ssDCval >= 44) {
if (ssDCval < 48) {
if (Rs < 1121) {
if (negBalance < 47) {
if (ssDCval < 45) {
if (Rs < 1020) {
return 8;
}
else if (Rs >= 1020) {
return 1;
}
}
else if (ssDCval >= 45) {
return 8;
}
}
else if (negBalance >= 47) {
if (Rs < 1035) {
return 1;
}
else if (Rs >= 1035) {
return 2;
}
}
}
else if (Rs >= 1121) {
if (Rs < 1293) {
return 1;
}
else if (Rs >= 1293) {
return 2;
}
}
}
else if (ssDCval >= 48) {
if (TSC < 4) {
if (SC < 1598) {
return 1;
}
else if (SC >= 1598) {
return 4;
}
}
else if (TSC >= 4) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (Rs < 1146) {
return 8;
}
else if (Rs >= 1146) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 89) {
if (RsCsDiff < 12) {
if (negBalance < 29) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (SC < 2067) {
return 8;
}
else if (SC >= 2067) {
if (mu_mv_mag_sq < 125) {
return 1;
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
}
}
else if (negBalance >= 29) {
return 8;
}
}
else if (RsCsDiff >= 12) {
if (mcTcor < 95) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (negBalance < 34) {
if (ssDCval < 49) {
return 2;
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (negBalance >= 34) {
if (mu_mv_mag_sq < 69) {
return 2;
}
else if (mu_mv_mag_sq >= 69) {
return 8;
}
}
}
}
else if (mcTcor >= 95) {
if (Rs < 1144) {
if (Rs < 990) {
return 8;
}
else if (Rs >= 990) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (Rs >= 1144) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 59) {
if (Rs < 1319) {
if (MVDiff < 5) {
if (RsDiff < 3) {
if (Rs < 997) {
if (Rs < 875) {
if (AFD < 5) {
if (posBalance < 61) {
if (diffAFD < -1) {
if (Rs < 723) {
return 1;
}
else if (Rs >= 723) {
return 2;
}
}
else if (diffAFD >= -1) {
if (SC < 1135) {
return 2;
}
else if (SC >= 1135) {
return 4;
}
}
}
else if (posBalance >= 61) {
if (Rs < 696) {
if (Rs < 655) {
return 8;
}
else if (Rs >= 655) {
return 1;
}
}
else if (Rs >= 696) {
return 2;
}
}
}
else if (AFD >= 5) {
if (posBalance < 55) {
if (Rs < 790) {
return 1;
}
else if (Rs >= 790) {
return 2;
}
}
else if (posBalance >= 55) {
if (AFD < 7) {
if (Rs < 487) {
return 1;
}
else if (Rs >= 487) {
return 8;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (Rs >= 875) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (Rs >= 997) {
if (AFD < 4) {
if (diffRsCsdiff < -4) {
return 4;
}
else if (diffRsCsdiff >= -4) {
return 1;
}
}
else if (AFD >= 4) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
if (posBalance < 52) {
if (Cs < 952) {
return 4;
}
else if (Cs >= 952) {
return 8;
}
}
else if (posBalance >= 52) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 111) {
return 1;
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1160) {
return 8;
}
else if (Cs >= 1160) {
if (Rs < 1244) {
return 1;
}
else if (Rs >= 1244) {
return 4;
}
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 3) {
if (negBalance < 38) {
if (Cs < 772) {
return 1;
}
else if (Cs >= 772) {
if (negBalance < 34) {
if (Cs < 1168) {
return 8;
}
else if (Cs >= 1168) {
if (Cs < 1258) {
return 4;
}
else if (Cs >= 1258) {
return 8;
}
}
}
else if (negBalance >= 34) {
return 4;
}
}
}
else if (negBalance >= 38) {
if (RsCsDiff < 22) {
if (diffMVdiffVal < -3) {
if (Rs < 1017) {
return 4;
}
else if (Rs >= 1017) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (Cs < 1254) {
if (Cs < 725) {
return 8;
}
else if (Cs >= 725) {
return 1;
}
}
else if (Cs >= 1254) {
return 4;
}
}
}
else if (RsCsDiff >= 22) {
if (mu_mv_mag_sq < 101) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
else if (mu_mv_mag_sq >= 101) {
return 4;
}
}
}
}
}
else if (MVDiff >= 5) {
if (negBalance < 34) {
if (mu_mv_mag_sq < 80) {
if (Rs < 721) {
return 1;
}
else if (Rs >= 721) {
return 2;
}
}
else if (mu_mv_mag_sq >= 80) {
if (negBalance < 24) {
return 1;
}
else if (negBalance >= 24) {
return 8;
}
}
}
else if (negBalance >= 34) {
if (mu_mv_mag_sq < 100) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 886) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (Rs >= 886) {
if (mcTcor < 96) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (Cs < 862) {
return 1;
}
else if (Cs >= 862) {
return 8;
}
}
}
else if (mcTcor >= 96) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 614) {
return 4;
}
else if (Cs >= 614) {
return 1;
}
}
}
}
}
else if (Rs >= 1319) {
if (mu_mv_mag_sq < 71) {
if (Cs < 893) {
if (Cs < 845) {
return 8;
}
else if (Cs >= 845) {
return 1;
}
}
else if (Cs >= 893) {
if (RsDiff < 3) {
if (Rs < 1370) {
return 8;
}
else if (Rs >= 1370) {
return 1;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (RsCsDiff < 23) {
if (diffRsCsdiff < 8) {
if (Cs < 703) {
if (Rs < 1387) {
return 8;
}
else if (Rs >= 1387) {
return 1;
}
}
else if (Cs >= 703) {
return 8;
}
}
else if (diffRsCsdiff >= 8) {
if (Rs < 1326) {
return 8;
}
else if (Rs >= 1326) {
return 4;
}
}
}
else if (RsCsDiff >= 23) {
return 4;
}
}
}
}
}
else if (ssDCval >= 78) {
if (tcor < 84) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
else if (tcor >= 84) {
if (SC < 1808) {
if (negBalance < 42) {
if (Rs < 700) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (RsCsDiff < 7) {
if (Cs < 800) {
if (SC < 1230) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (Cs < 596) {
return 8;
}
else if (Cs >= 596) {
return 2;
}
}
}
else if (SC >= 1230) {
return 2;
}
}
else if (Cs >= 800) {
return 8;
}
}
else if (RsCsDiff >= 7) {
return 8;
}
}
}
else if (Rs >= 700) {
if (mu_mv_mag_sq < 75) {
if (Cs < 921) {
return 2;
}
else if (Cs >= 921) {
return 8;
}
}
else if (mu_mv_mag_sq >= 75) {
if (negBalance < 39) {
if (ssDCval < 86) {
if (Rs < 767) {
return 2;
}
else if (Rs >= 767) {
return 8;
}
}
else if (ssDCval >= 86) {
return 8;
}
}
else if (negBalance >= 39) {
if (Cs < 905) {
return 8;
}
else if (Cs >= 905) {
return 1;
}
}
}
}
}
else if (negBalance >= 42) {
if (Rs < 454) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (Rs >= 454) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (refDCval < 90) {
if (Cs < 827) {
return 8;
}
else if (Cs >= 827) {
return 1;
}
}
else if (refDCval >= 90) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
}
}
else if (SC >= 1808) {
if (Cs < 711) {
if (Rs < 1468) {
return 4;
}
else if (Rs >= 1468) {
if (Rs < 1543) {
return 1;
}
else if (Rs >= 1543) {
return 8;
}
}
}
else if (Cs >= 711) {
if (mu_mv_mag_sq < 65) {
if (Rs < 965) {
return 4;
}
else if (Rs >= 965) {
return 8;
}
}
else if (mu_mv_mag_sq >= 65) {
if (negBalance < 27) {
if (Cs < 1215) {
return 8;
}
else if (Cs >= 1215) {
return 4;
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 6) {
if (negBalance < 44) {
if (ssDCval < 76) {
if (ssDCval < 58) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (MVDiff < 2) {
if (Rs < 1295) {
return 4;
}
else if (Rs >= 1295) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Cs < 1045) {
if (Rs < 987) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (Rs >= 987) {
return 2;
}
}
else if (Cs >= 1045) {
return 8;
}
}
}
}
else if (ssDCval >= 58) {
if (Rs < 835) {
if (negBalance < 38) {
if (diffMVdiffVal < 2) {
if (Rs < 708) {
return 8;
}
else if (Rs >= 708) {
return 1;
}
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (Rs >= 835) {
if (RsCsDiff < 43) {
if (posBalance < 55) {
if (tcor < 89) {
return 1;
}
else if (tcor >= 89) {
if (Cs < 1165) {
return 8;
}
else if (Cs >= 1165) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (SC < 2514) {
return 4;
}
else if (SC >= 2514) {
if (ssDCval < 65) {
if (Rs < 1358) {
return 8;
}
else if (Rs >= 1358) {
return 1;
}
}
else if (ssDCval >= 65) {
return 4;
}
}
}
}
else if (RsCsDiff >= 43) {
if (Rs < 966) {
return 1;
}
else if (Rs >= 966) {
return 8;
}
}
}
}
}
else if (ssDCval >= 76) {
if (diffMVdiffVal < -1) {
if (Rs < 1566) {
if (mu_mv_mag_sq < 101) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 972) {
return 4;
}
else if (Rs >= 972) {
return 8;
}
}
}
else if (Rs >= 1566) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Cs < 948) {
return 1;
}
else if (Cs >= 948) {
return 8;
}
}
}
}
}
else if (negBalance >= 44) {
if (RsDiff < 2) {
if (Rs < 431) {
return 4;
}
else if (Rs >= 431) {
if (Rs < 521) {
return 1;
}
else if (Rs >= 521) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (Cs < 732) {
if (tcor < 82) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 66) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 66) {
return 4;
}
}
}
}
else if (tcor >= 82) {
if (posBalance < 56) {
if (Cs < 552) {
return 4;
}
else if (Cs >= 552) {
if (negBalance < 45) {
if (Rs < 1060) {
return 8;
}
else if (Rs >= 1060) {
return 4;
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
else if (Cs >= 732) {
if (gchDC < 1) {
if (mu_mv_mag_sq < 80) {
if (diffMVdiffVal < 1) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
if (Rs < 1318) {
if (tcor < 87) {
if (Rs < 1290) {
if (Cs < 791) {
if (Rs < 832) {
return 8;
}
else if (Rs >= 832) {
return 1;
}
}
else if (Cs >= 791) {
return 8;
}
}
else if (Rs >= 1290) {
return 1;
}
}
else if (tcor >= 87) {
return 8;
}
}
else if (Rs >= 1318) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Cs < 1014) {
return 8;
}
else if (Cs >= 1014) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
}
else if (gchDC >= 1) {
if (tcor < 77) {
if (posBalance < 58) {
if (Rs < 1065) {
return 4;
}
else if (Rs >= 1065) {
return 1;
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (tcor >= 77) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (ssDCval < 75) {
if (ssDCval < 74) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (ssDCval >= 74) {
return 4;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1596) {
if (ssDCval < 70) {
if (refDCval < 62) {
if (negBalance < 39) {
if (diffTSC < -1) {
if (Rs < 1684) {
if (Rs < 1647) {
return 1;
}
else if (Rs >= 1647) {
return 8;
}
}
else if (Rs >= 1684) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (diffTSC >= -1) {
if (mu_mv_mag_sq < 71) {
if (Rs < 1841) {
return 8;
}
else if (Rs >= 1841) {
if (Rs < 1999) {
return 4;
}
else if (Rs >= 1999) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (mu_mv_mag_sq < 117) {
return 8;
}
else if (mu_mv_mag_sq >= 117) {
if (Rs < 1927) {
return 8;
}
else if (Rs >= 1927) {
return 4;
}
}
}
}
}
else if (negBalance >= 39) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (Cs < 682) {
return 8;
}
else if (Cs >= 682) {
return 2;
}
}
}
}
else if (refDCval >= 62) {
if (tcor < 89) {
if (diffAFD < 1) {
if (Rs < 2116) {
return 8;
}
else if (Rs >= 2116) {
return 4;
}
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (tcor >= 89) {
if (mu_mv_mag_sq < 81) {
if (Cs < 914) {
return 2;
}
else if (Cs >= 914) {
return 8;
}
}
else if (mu_mv_mag_sq >= 81) {
return 2;
}
}
}
}
else if (ssDCval >= 70) {
if (refDCval < 79) {
if (mcTcor < 93) {
if (refDCval < 71) {
return 2;
}
else if (refDCval >= 71) {
return 8;
}
}
else if (mcTcor >= 93) {
if (Rs < 1745) {
if (posBalance < 56) {
if (posBalance < 49) {
if (Rs < 1684) {
return 1;
}
else if (Rs >= 1684) {
return 8;
}
}
else if (posBalance >= 49) {
return 8;
}
}
else if (posBalance >= 56) {
return 1;
}
}
else if (Rs >= 1745) {
if (RsCsDiff < 41) {
if (Cs < 1288) {
return 1;
}
else if (Cs >= 1288) {
if (Rs < 1973) {
return 1;
}
else if (Rs >= 1973) {
return 8;
}
}
}
else if (RsCsDiff >= 41) {
return 8;
}
}
}
}
else if (refDCval >= 79) {
if (negBalance < 49) {
if (RsDiff < 8) {
if (refDCval < 99) {
if (RsCsDiff < 23) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (SC < 3027) {
return 8;
}
else if (SC >= 3027) {
if (SC < 3055) {
return 2;
}
else if (SC >= 3055) {
return 8;
}
}
}
}
else if (RsCsDiff >= 23) {
if (ssDCval < 87) {
if (mu_mv_mag_sq < 97) {
if (RsCsDiff < 49) {
if (Rs < 1727) {
return 8;
}
else if (Rs >= 1727) {
if (Cs < 1295) {
return 1;
}
else if (Cs >= 1295) {
return 8;
}
}
}
else if (RsCsDiff >= 49) {
return 8;
}
}
else if (mu_mv_mag_sq >= 97) {
return 8;
}
}
else if (ssDCval >= 87) {
if (AFD < 9) {
if (Cs < 947) {
if (Rs < 2017) {
return 8;
}
else if (Rs >= 2017) {
return 1;
}
}
else if (Cs >= 947) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
return 8;
}
}
}
else if (AFD >= 9) {
if (RsCsDiff < 51) {
return 8;
}
else if (RsCsDiff >= 51) {
return 2;
}
}
}
}
}
else if (refDCval >= 99) {
if (mu_mv_mag_sq < 118) {
return 8;
}
else if (mu_mv_mag_sq >= 118) {
if (AFD < 7) {
if (ssDCval < 130) {
return 8;
}
else if (ssDCval >= 130) {
return 1;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (RsDiff >= 8) {
if (CsDiff < 5) {
if (Rs < 2177) {
return 1;
}
else if (Rs >= 2177) {
return 2;
}
}
else if (CsDiff >= 5) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
}
else if (negBalance >= 49) {
if (ssDCval < 100) {
if (TSC < 7) {
if (Rs < 1750) {
return 1;
}
else if (Rs >= 1750) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (ssDCval >= 100) {
if (posBalance < 58) {
if (diffTSC < 3) {
return 8;
}
else if (diffTSC >= 3) {
return 1;
}
}
else if (posBalance >= 58) {
if (Rs < 2039) {
return 1;
}
else if (Rs >= 2039) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (refDCval < 114) {
if (MVDiff < 10) {
if (mu_mv_mag_sq < 2) {
if (AFD < 15) {
if (TSC < 8) {
if (CsDiff < 5) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (gchDC < 2) {
if (diffAFD < -2) {
if (Rs < 1017) {
return 8;
}
else if (Rs >= 1017) {
return 4;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (CsDiff >= 5) {
if (posBalance < 58) {
if (CsDiff < 7) {
return 1;
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (posBalance >= 58) {
return 4;
}
}
}
else if (TSC >= 8) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 1768) {
return 8;
}
else if (Rs >= 1768) {
if (Rs < 1990) {
return 1;
}
else if (Rs >= 1990) {
return 2;
}
}
}
}
}
else if (AFD >= 15) {
if (ssDCval < 86) {
if (Rs < 1726) {
return 8;
}
else if (Rs >= 1726) {
return 4;
}
}
else if (ssDCval >= 86) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (Rs < 1933) {
return 8;
}
else if (Rs >= 1933) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1684) {
if (SC < 1586) {
if (MVDiff < 5) {
if (Rs < 883) {
return 1;
}
else if (Rs >= 883) {
return 4;
}
}
else if (MVDiff >= 5) {
if (Rs < 382) {
return 2;
}
else if (Rs >= 382) {
if (ssDCval < 107) {
return 8;
}
else if (ssDCval >= 107) {
return 2;
}
}
}
}
else if (SC >= 1586) {
if (mcTcor < 93) {
if (Rs < 964) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (Rs < 810) {
return 4;
}
else if (Rs >= 810) {
return 8;
}
}
}
else if (Rs >= 964) {
if (diffMVdiffVal < -4) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (posBalance < 64) {
if (Rs < 1440) {
return 8;
}
else if (Rs >= 1440) {
return 4;
}
}
else if (posBalance >= 64) {
if (Cs < 1081) {
return 4;
}
else if (Cs >= 1081) {
return 1;
}
}
}
}
}
else if (Rs >= 1684) {
if (Cs < 988) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (Cs >= 988) {
if (Rs < 1929) {
return 8;
}
else if (Rs >= 1929) {
return 4;
}
}
}
}
}
else if (MVDiff >= 10) {
if (tcor < 57) {
if (AFD < 26) {
if (Rs < 424) {
if (ssDCval < 46) {
return 2;
}
else if (ssDCval >= 46) {
if (Cs < 578) {
if (Rs < 351) {
return 2;
}
else if (Rs >= 351) {
return 4;
}
}
else if (Cs >= 578) {
return 4;
}
}
}
else if (Rs >= 424) {
if (ssDCval < 95) {
if (Rs < 1537) {
if (diffAFD < 1) {
if (AFD < 16) {
if (posBalance < 66) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 4;
}
}
else if (posBalance >= 66) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (Cs < 1058) {
return 8;
}
else if (Cs >= 1058) {
return 1;
}
}
}
}
else if (AFD >= 16) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
if (RsDiff < 6) {
if (Rs < 962) {
return 4;
}
else if (Rs >= 962) {
return 1;
}
}
else if (RsDiff >= 6) {
if (gchDC < 2) {
if (Cs < 1010) {
return 8;
}
else if (Cs >= 1010) {
return 1;
}
}
else if (gchDC >= 2) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
}
else if (diffAFD >= 1) {
if (ssDCval < 52) {
if (gchDC < 1) {
if (Cs < 785) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 8;
}
}
else if (Cs >= 785) {
if (Rs < 881) {
return 2;
}
else if (Rs >= 881) {
return 4;
}
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (ssDCval >= 52) {
if (Cs < 1021) {
if (tcor < 50) {
if (mcTcor < 78) {
if (Rs < 720) {
return 4;
}
else if (Rs >= 720) {
return 8;
}
}
else if (mcTcor >= 78) {
if (Cs < 887) {
return 4;
}
else if (Cs >= 887) {
return 8;
}
}
}
else if (tcor >= 50) {
if (RsCsDiff < 38) {
if (ssDCval < 73) {
return 4;
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (RsCsDiff >= 38) {
if (SC < 1629) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 1;
}
}
else if (SC >= 1629) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
return 1;
}
}
}
}
}
else if (Cs >= 1021) {
if (ssDCval < 65) {
if (refDCval < 47) {
return 2;
}
else if (refDCval >= 47) {
if (diffAFD < 2) {
if (Cs < 1240) {
return 1;
}
else if (Cs >= 1240) {
return 8;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (ssDCval >= 65) {
if (Rs < 1108) {
return 4;
}
else if (Rs >= 1108) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1537) {
if (posBalance < 60) {
if (Rs < 1591) {
return 1;
}
else if (Rs >= 1591) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 4;
}
}
}
else if (posBalance >= 60) {
if (refDCval < 90) {
return 8;
}
else if (refDCval >= 90) {
return 1;
}
}
}
}
else if (ssDCval >= 95) {
return 4;
}
}
}
else if (AFD >= 26) {
if (TSC < 9) {
if (mu_mv_mag_sq < 51) {
return 4;
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 701) {
return 1;
}
else if (Rs >= 701) {
return 8;
}
}
}
else if (TSC >= 9) {
if (MVDiff < 45) {
if (diffRsCsdiff < -21) {
return 2;
}
else if (diffRsCsdiff >= -21) {
if (tcor < 49) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (tcor >= 49) {
if (SC < 1945) {
return 2;
}
else if (SC >= 1945) {
return 4;
}
}
}
}
else if (MVDiff >= 45) {
if (posBalance < 53) {
if (TSC < 16) {
if (MVDiff < 195) {
return 1;
}
else if (MVDiff >= 195) {
return 8;
}
}
else if (TSC >= 16) {
return 1;
}
}
else if (posBalance >= 53) {
if (mcTcor < 85) {
if (refDCval < 102) {
if (diffAFD < 4) {
if (MVDiff < 187) {
if (Cs < 693) {
return 2;
}
else if (Cs >= 693) {
return 8;
}
}
else if (MVDiff >= 187) {
if (Rs < 1130) {
return 1;
}
else if (Rs >= 1130) {
return 2;
}
}
}
else if (diffAFD >= 4) {
if (diffAFD < 13) {
if (Rs < 2080) {
return 4;
}
else if (Rs >= 2080) {
return 2;
}
}
else if (diffAFD >= 13) {
if (Cs < 988) {
if (Rs < 1891) {
return 8;
}
else if (Rs >= 1891) {
return 1;
}
}
else if (Cs >= 988) {
return 2;
}
}
}
}
else if (refDCval >= 102) {
if (posBalance < 75) {
if (Cs < 609) {
return 1;
}
else if (Cs >= 609) {
return 2;
}
}
else if (posBalance >= 75) {
return 4;
}
}
}
else if (mcTcor >= 85) {
if (CsDiff < 6) {
if (AFD < 32) {
return 1;
}
else if (AFD >= 32) {
return 2;
}
}
else if (CsDiff >= 6) {
if (Cs < 629) {
return 8;
}
else if (Cs >= 629) {
if (RsCsDiff < 92) {
if (Rs < 743) {
return 2;
}
else if (Rs >= 743) {
return 4;
}
}
else if (RsCsDiff >= 92) {
return 2;
}
}
}
}
}
}
}
}
}
else if (tcor >= 57) {
if (Cs < 1007) {
if (negBalance < 45) {
if (posBalance < 48) {
if (RsCsDiff < 60) {
return 1;
}
else if (RsCsDiff >= 60) {
return 2;
}
}
else if (posBalance >= 48) {
if (Cs < 817) {
if (diffMVdiffVal < 17) {
if (CsDiff < 6) {
if (MVDiff < 20) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (Cs < 724) {
return 2;
}
else if (Cs >= 724) {
return 1;
}
}
}
else if (MVDiff >= 20) {
if (tcor < 71) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (tcor >= 71) {
if (tcor < 88) {
if (diffMVdiffVal < 1) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
if (ssDCval < 85) {
if (diffMVdiffVal < 5) {
if (Cs < 758) {
return 4;
}
else if (Cs >= 758) {
return 1;
}
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
else if (ssDCval >= 85) {
if (Rs < 1817) {
return 2;
}
else if (Rs >= 1817) {
return 1;
}
}
}
}
else if (tcor >= 88) {
return 1;
}
}
}
}
else if (CsDiff >= 6) {
if (tcor < 60) {
if (Rs < 459) {
return 2;
}
else if (Rs >= 459) {
return 4;
}
}
else if (tcor >= 60) {
if (ssDCval < 88) {
return 2;
}
else if (ssDCval >= 88) {
if (Rs < 1561) {
return 8;
}
else if (Rs >= 1561) {
return 2;
}
}
}
}
}
else if (diffMVdiffVal >= 17) {
if (RsCsDiff < 55) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
else if (RsCsDiff >= 55) {
if (Cs < 728) {
return 4;
}
else if (Cs >= 728) {
return 2;
}
}
}
}
else if (Cs >= 817) {
if (gchDC < 2) {
if (ssDCval < 67) {
if (Rs < 1525) {
return 8;
}
else if (Rs >= 1525) {
return 2;
}
}
else if (ssDCval >= 67) {
return 4;
}
}
else if (gchDC >= 2) {
if (RsCsDiff < 171) {
if (AFD < 25) {
if (posBalance < 67) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
else if (posBalance >= 67) {
return 8;
}
}
else if (AFD >= 25) {
return 1;
}
}
else if (RsCsDiff >= 171) {
if (Rs < 2314) {
return 2;
}
else if (Rs >= 2314) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 45) {
if (refDCval < 89) {
if (Rs < 1594) {
if (negBalance < 52) {
if (Rs < 1429) {
if (CsDiff < 6) {
if (mu_mv_mag_sq < 41) {
if (SC < 1731) {
if (SC < 1631) {
if (diffMVdiffVal < -4) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 751) {
return 8;
}
else if (Rs >= 751) {
return 4;
}
}
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
else if (SC >= 1631) {
if (ssDCval < 71) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 4;
}
}
else if (ssDCval >= 71) {
return 1;
}
}
}
else if (SC >= 1731) {
if (MVDiff < 42) {
if (tcor < 83) {
if (Cs < 583) {
return 2;
}
else if (Cs >= 583) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (tcor >= 83) {
return 8;
}
}
else if (MVDiff >= 42) {
if (Cs < 851) {
return 1;
}
else if (Cs >= 851) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (Rs < 1251) {
return 8;
}
else if (Rs >= 1251) {
return 4;
}
}
}
else if (CsDiff >= 6) {
if (diffMVdiffVal < 1) {
if (diffMVdiffVal < -15) {
if (Cs < 946) {
return 1;
}
else if (Cs >= 946) {
return 4;
}
}
else if (diffMVdiffVal >= -15) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
if (posBalance < 75) {
if (Cs < 953) {
if (MVDiff < 77) {
return 2;
}
else if (MVDiff >= 77) {
if (Cs < 653) {
return 2;
}
else if (Cs >= 653) {
return 1;
}
}
}
else if (Cs >= 953) {
if (MVDiff < 57) {
return 1;
}
else if (MVDiff >= 57) {
return 4;
}
}
}
else if (posBalance >= 75) {
return 8;
}
}
}
}
else if (Rs >= 1429) {
if (mu_mv_mag_sq < 5) {
return 8;
}
else if (mu_mv_mag_sq >= 5) {
if (Cs < 643) {
return 1;
}
else if (Cs >= 643) {
return 2;
}
}
}
}
else if (negBalance >= 52) {
if (Rs < 285) {
if (SC < 804) {
return 1;
}
else if (SC >= 804) {
return 4;
}
}
else if (Rs >= 285) {
if (RsCsDiff < 43) {
if (SC < 1520) {
if (RsCsDiff < 23) {
if (negBalance < 59) {
if (Rs < 496) {
return 8;
}
else if (Rs >= 496) {
return 4;
}
}
else if (negBalance >= 59) {
if (SC < 1116) {
return 4;
}
else if (SC >= 1116) {
return 2;
}
}
}
else if (RsCsDiff >= 23) {
if (gchDC < 2) {
if (MVDiff < 43) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (MVDiff >= 43) {
return 4;
}
}
else if (gchDC >= 2) {
if (ssDCval < 58) {
if (Rs < 557) {
return 2;
}
else if (Rs >= 557) {
return 8;
}
}
else if (ssDCval >= 58) {
return 8;
}
}
}
}
else if (SC >= 1520) {
if (Rs < 953) {
if (MVDiff < 24) {
return 1;
}
else if (MVDiff >= 24) {
return 8;
}
}
else if (Rs >= 953) {
return 8;
}
}
}
else if (RsCsDiff >= 43) {
if (Cs < 688) {
if (mu_mv_mag_sq < 17) {
if (ssDCval < 62) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (Rs < 763) {
return 1;
}
else if (Rs >= 763) {
return 2;
}
}
}
else if (ssDCval >= 62) {
if (posBalance < 56) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
if (MVDiff < 127) {
return 2;
}
else if (MVDiff >= 127) {
return 8;
}
}
}
else if (posBalance >= 56) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 17) {
return 4;
}
}
else if (Cs >= 688) {
if (refDCval < 68) {
if (Cs < 785) {
if (diffMVdiffVal < 7) {
return 8;
}
else if (diffMVdiffVal >= 7) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 8;
}
}
}
else if (Cs >= 785) {
if (diffRsCsdiff < 18) {
if (mcTcor < 92) {
if (Cs < 886) {
return 4;
}
else if (Cs >= 886) {
if (Rs < 737) {
return 4;
}
else if (Rs >= 737) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (Rs < 756) {
return 2;
}
else if (Rs >= 756) {
return 1;
}
}
}
else if (diffRsCsdiff >= 18) {
if (negBalance < 77) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (negBalance >= 77) {
if (Rs < 846) {
return 4;
}
else if (Rs >= 846) {
return 1;
}
}
}
}
}
else if (refDCval >= 68) {
if (Rs < 667) {
if (Cs < 790) {
if (Rs < 602) {
return 8;
}
else if (Rs >= 602) {
return 1;
}
}
else if (Cs >= 790) {
return 2;
}
}
else if (Rs >= 667) {
if (AFD < 27) {
if (diffRsCsdiff < -8) {
if (diffMVdiffVal < -8) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (Rs < 869) {
return 1;
}
else if (Rs >= 869) {
return 8;
}
}
}
else if (diffMVdiffVal >= -8) {
if (diffMVdiffVal < 5) {
if (MVDiff < 27) {
if (Rs < 885) {
return 8;
}
else if (Rs >= 885) {
return 4;
}
}
else if (MVDiff >= 27) {
return 8;
}
}
else if (diffMVdiffVal >= 5) {
if (Rs < 799) {
return 2;
}
else if (Rs >= 799) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= -8) {
if (Cs < 951) {
if (tcor < 62) {
if (Rs < 1060) {
return 2;
}
else if (Rs >= 1060) {
return 8;
}
}
else if (tcor >= 62) {
return 8;
}
}
else if (Cs >= 951) {
if (ssDCval < 77) {
if (Rs < 1271) {
return 4;
}
else if (Rs >= 1271) {
return 8;
}
}
else if (ssDCval >= 77) {
return 1;
}
}
}
}
else if (AFD >= 27) {
if (Rs < 986) {
return 4;
}
else if (Rs >= 986) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1594) {
if (mcTcor < 93) {
if (ssDCval < 84) {
if (Cs < 873) {
if (refDCval < 61) {
return 4;
}
else if (refDCval >= 61) {
if (gchDC < 4) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (SC < 2337) {
return 8;
}
else if (SC >= 2337) {
return 2;
}
}
}
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (Cs >= 873) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (Cs < 886) {
return 8;
}
else if (Cs >= 886) {
return 4;
}
}
}
}
}
else if (ssDCval >= 84) {
if (diffMVdiffVal < -7) {
if (Cs < 663) {
return 8;
}
else if (Cs >= 663) {
return 1;
}
}
else if (diffMVdiffVal >= -7) {
if (Rs < 2335) {
return 4;
}
else if (Rs >= 2335) {
return 8;
}
}
}
}
else if (mcTcor >= 93) {
if (Rs < 1628) {
return 4;
}
else if (Rs >= 1628) {
if (refDCval < 85) {
return 2;
}
else if (refDCval >= 85) {
if (SC < 2619) {
return 2;
}
else if (SC >= 2619) {
if (Rs < 1943) {
return 1;
}
else if (Rs >= 1943) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 89) {
if (tcor < 80) {
if (AFD < 26) {
if (gchDC < 2) {
if (RsCsDiff < 48) {
if (Cs < 784) {
if (Rs < 707) {
return 2;
}
else if (Rs >= 707) {
if (Rs < 979) {
return 1;
}
else if (Rs >= 979) {
return 8;
}
}
}
else if (Cs >= 784) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
return 4;
}
}
}
else if (RsCsDiff >= 48) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (mu_mv_mag_sq < 15) {
if (Rs < 1350) {
if (Rs < 811) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 2;
}
}
else if (Rs >= 811) {
return 8;
}
}
else if (Rs >= 1350) {
if (diffRsCsdiff < -31) {
return 8;
}
else if (diffRsCsdiff >= -31) {
if (Rs < 2212) {
return 1;
}
else if (Rs >= 2212) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (diffAFD < -1) {
if (mu_mv_mag_sq < 24) {
return 2;
}
else if (mu_mv_mag_sq >= 24) {
return 4;
}
}
else if (diffAFD >= -1) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
if (gchDC < 1) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 1;
}
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
}
}
}
else if (gchDC >= 2) {
if (diffAFD < -4) {
if (Cs < 766) {
return 2;
}
else if (Cs >= 766) {
return 1;
}
}
else if (diffAFD >= -4) {
if (MVDiff < 104) {
if (Cs < 968) {
if (diffMVdiffVal < 7) {
return 8;
}
else if (diffMVdiffVal >= 7) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
if (Rs < 1638) {
return 2;
}
else if (Rs >= 1638) {
return 1;
}
}
}
}
else if (Cs >= 968) {
return 2;
}
}
else if (MVDiff >= 104) {
return 2;
}
}
}
}
else if (AFD >= 26) {
if (CsDiff < 7) {
if (MVDiff < 35) {
return 4;
}
else if (MVDiff >= 35) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
else if (CsDiff >= 7) {
if (mu_mv_mag_sq < 27) {
if (Rs < 518) {
return 1;
}
else if (Rs >= 518) {
return 2;
}
}
else if (mu_mv_mag_sq >= 27) {
if (AFD < 31) {
return 4;
}
else if (AFD >= 31) {
return 2;
}
}
}
}
}
else if (tcor >= 80) {
if (gchDC < 4) {
if (mcTcor < 91) {
return 1;
}
else if (mcTcor >= 91) {
if (ssDCval < 99) {
if (Cs < 755) {
return 1;
}
else if (Cs >= 755) {
if (Rs < 1147) {
return 1;
}
else if (Rs >= 1147) {
if (Rs < 1876) {
return 8;
}
else if (Rs >= 1876) {
return 4;
}
}
}
}
else if (ssDCval >= 99) {
if (mcTcor < 96) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 1;
}
}
else if (mcTcor >= 96) {
if (Rs < 1487) {
return 8;
}
else if (Rs >= 1487) {
return 1;
}
}
}
}
}
else if (gchDC >= 4) {
if (diffMVdiffVal < -21) {
if (Rs < 1449) {
return 2;
}
else if (Rs >= 1449) {
return 8;
}
}
else if (diffMVdiffVal >= -21) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1007) {
if (AFD < 19) {
if (TSC < 10) {
if (negBalance < 36) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
else if (negBalance >= 36) {
if (mu_mv_mag_sq < 9) {
if (TSC < 8) {
if (tcor < 78) {
if (MVDiff < 13) {
if (Rs < 1820) {
return 8;
}
else if (Rs >= 1820) {
return 2;
}
}
else if (MVDiff >= 13) {
return 2;
}
}
else if (tcor >= 78) {
if (refDCval < 73) {
if (mu_mv_mag_sq < 2) {
if (RsCsDiff < 69) {
return 1;
}
else if (RsCsDiff >= 69) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
else if (refDCval >= 73) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (posBalance >= 55) {
if (ssDCval < 95) {
return 2;
}
else if (ssDCval >= 95) {
if (Cs < 1063) {
return 2;
}
else if (Cs >= 1063) {
return 8;
}
}
}
}
}
}
else if (TSC >= 8) {
if (Rs < 1350) {
if (ssDCval < 90) {
if (diffAFD < 2) {
if (Cs < 1081) {
if (Rs < 884) {
return 2;
}
else if (Rs >= 884) {
return 8;
}
}
else if (Cs >= 1081) {
if (diffRsCsdiff < -1) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
return 2;
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (ssDCval >= 90) {
if (Rs < 1210) {
return 8;
}
else if (Rs >= 1210) {
return 4;
}
}
}
else if (Rs >= 1350) {
if (tcor < 77) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (tcor >= 77) {
if (Rs < 2022) {
if (posBalance < 73) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 2;
}
}
else if (posBalance >= 73) {
return 2;
}
}
else if (Rs >= 2022) {
if (Cs < 1142) {
return 4;
}
else if (Cs >= 1142) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (posBalance < 57) {
if (RsCsDiff < 67) {
if (mu_mv_mag_sq < 49) {
if (ssDCval < 70) {
if (Rs < 1626) {
return 1;
}
else if (Rs >= 1626) {
return 8;
}
}
else if (ssDCval >= 70) {
if (Cs < 1106) {
return 2;
}
else if (Cs >= 1106) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 49) {
return 4;
}
}
else if (RsCsDiff >= 67) {
if (MVDiff < 31) {
if (tcor < 77) {
if (diffMVdiffVal < -7) {
return 8;
}
else if (diffMVdiffVal >= -7) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (tcor >= 77) {
if (diffMVdiffVal < 7) {
return 8;
}
else if (diffMVdiffVal >= 7) {
return 2;
}
}
}
else if (MVDiff >= 31) {
return 4;
}
}
}
else if (posBalance >= 57) {
if (mcTcor < 83) {
if (Rs < 1582) {
return 8;
}
else if (Rs >= 1582) {
return 4;
}
}
else if (mcTcor >= 83) {
if (diffRsCsdiff < 7) {
return 8;
}
else if (diffRsCsdiff >= 7) {
if (mcTcor < 93) {
if (Rs < 1124) {
return 2;
}
else if (Rs >= 1124) {
return 8;
}
}
else if (mcTcor >= 93) {
if (mcTcor < 95) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (mcTcor >= 95) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
}
}
}
}
}
else if (TSC >= 10) {
if (refDCval < 95) {
if (SC < 2420) {
if (RsCsDiff < 46) {
if (Rs < 890) {
return 1;
}
else if (Rs >= 890) {
return 8;
}
}
else if (RsCsDiff >= 46) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 8;
}
}
}
else if (SC >= 2420) {
if (MVDiff < 12) {
if (Rs < 1504) {
return 4;
}
else if (Rs >= 1504) {
return 8;
}
}
else if (MVDiff >= 12) {
if (diffRsCsdiff < 42) {
return 8;
}
else if (diffRsCsdiff >= 42) {
if (Rs < 1842) {
return 8;
}
else if (Rs >= 1842) {
return 4;
}
}
}
}
}
else if (refDCval >= 95) {
if (tcor < 76) {
if (Rs < 1822) {
return 4;
}
else if (Rs >= 1822) {
return 1;
}
}
else if (tcor >= 76) {
if (Rs < 1860) {
return 8;
}
else if (Rs >= 1860) {
return 1;
}
}
}
}
}
else if (AFD >= 19) {
if (RsCsDiff < 223) {
if (RsDiff < 9) {
if (RsCsDiff < 127) {
if (mcTcor < 87) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
return 2;
}
}
else if (mcTcor >= 87) {
if (mu_mv_mag_sq < 14) {
if (AFD < 23) {
if (diffMVdiffVal < 4) {
if (diffAFD < -2) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 2;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (diffMVdiffVal >= 4) {
if (Rs < 1350) {
return 2;
}
else if (Rs >= 1350) {
return 1;
}
}
}
else if (AFD >= 23) {
if (negBalance < 60) {
return 4;
}
else if (negBalance >= 60) {
if (Cs < 1055) {
return 4;
}
else if (Cs >= 1055) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (refDCval < 92) {
if (Rs < 1070) {
return 2;
}
else if (Rs >= 1070) {
return 8;
}
}
else if (refDCval >= 92) {
return 1;
}
}
}
}
else if (RsCsDiff >= 127) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
if (negBalance < 51) {
if (Rs < 1034) {
return 8;
}
else if (Rs >= 1034) {
return 2;
}
}
else if (negBalance >= 51) {
if (Cs < 1164) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
if (Rs < 400) {
return 1;
}
else if (Rs >= 400) {
return 8;
}
}
}
else if (Cs >= 1164) {
if (Rs < 751) {
return 4;
}
else if (Rs >= 751) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (Rs < 1968) {
if (AFD < 23) {
if (CsDiff < 8) {
if (refDCval < 73) {
return 8;
}
else if (refDCval >= 73) {
if (Cs < 1039) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (Cs >= 1039) {
return 4;
}
}
}
else if (CsDiff >= 8) {
if (diffRsCsdiff < -31) {
if (Rs < 1525) {
return 4;
}
else if (Rs >= 1525) {
return 2;
}
}
else if (diffRsCsdiff >= -31) {
return 8;
}
}
}
else if (AFD >= 23) {
if (gchDC < 9) {
if (RsCsDiff < 124) {
return 1;
}
else if (RsCsDiff >= 124) {
return 4;
}
}
else if (gchDC >= 9) {
return 8;
}
}
}
else if (Rs >= 1968) {
if (RsDiff < 11) {
if (Cs < 1168) {
return 2;
}
else if (Cs >= 1168) {
return 1;
}
}
else if (RsDiff >= 11) {
if (posBalance < 54) {
if (Rs < 2046) {
return 8;
}
else if (Rs >= 2046) {
return 2;
}
}
else if (posBalance >= 54) {
if (Cs < 1270) {
return 4;
}
else if (Cs >= 1270) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 223) {
if (refDCval < 85) {
return 2;
}
else if (refDCval >= 85) {
if (diffAFD < 4) {
return 2;
}
else if (diffAFD >= 4) {
if (RsDiff < 17) {
return 8;
}
else if (RsDiff >= 17) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (MVDiff < 12) {
if (negBalance < 59) {
if (tcor < 63) {
return 4;
}
else if (tcor >= 63) {
if (Rs < 538) {
return 2;
}
else if (Rs >= 538) {
if (diffRsCsdiff < 61) {
if (MVDiff < 5) {
if (Cs < 749) {
return 8;
}
else if (Cs >= 749) {
if (Rs < 990) {
return 1;
}
else if (Rs >= 990) {
return 4;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (diffRsCsdiff >= 61) {
return 2;
}
}
}
}
else if (negBalance >= 59) {
if (SC < 2925) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
else if (SC >= 2925) {
return 8;
}
}
}
else if (MVDiff >= 12) {
if (SC < 2242) {
if (SC < 1458) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (refDCval < 119) {
return 4;
}
else if (refDCval >= 119) {
if (TSC < 8) {
if (Rs < 641) {
return 8;
}
else if (Rs >= 641) {
return 2;
}
}
else if (TSC >= 8) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (Rs < 565) {
return 8;
}
else if (Rs >= 565) {
return 4;
}
}
}
}
}
}
else if (SC >= 1458) {
if (tcor < 31) {
if (AFD < 37) {
return 1;
}
else if (AFD >= 37) {
return 2;
}
}
else if (tcor >= 31) {
if (TSC < 10) {
if (diffTSC < -2) {
if (MVDiff < 66) {
return 2;
}
else if (MVDiff >= 66) {
if (Rs < 1166) {
return 1;
}
else if (Rs >= 1166) {
return 8;
}
}
}
else if (diffTSC >= -2) {
if (Cs < 1133) {
if (tcor < 41) {
return 1;
}
else if (tcor >= 41) {
if (AFD < 24) {
if (negBalance < 30) {
return 2;
}
else if (negBalance >= 30) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
if (SC < 2204) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (tcor < 80) {
if (Cs < 749) {
return 2;
}
else if (Cs >= 749) {
return 8;
}
}
else if (tcor >= 80) {
return 2;
}
}
}
else if (SC >= 2204) {
return 2;
}
}
}
}
else if (AFD >= 24) {
return 2;
}
}
}
else if (Cs >= 1133) {
return 1;
}
}
}
else if (TSC >= 10) {
if (negBalance < 71) {
if (RsCsDiff < 41) {
return 8;
}
else if (RsCsDiff >= 41) {
if (Rs < 1122) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (mu_mv_mag_sq < 12) {
return 2;
}
else if (mu_mv_mag_sq >= 12) {
if (Rs < 689) {
return 2;
}
else if (Rs >= 689) {
return 1;
}
}
}
}
else if (Rs >= 1122) {
if (MVDiff < 52) {
return 1;
}
else if (MVDiff >= 52) {
if (RsCsDiff < 101) {
return 8;
}
else if (RsCsDiff >= 101) {
if (Rs < 1354) {
return 4;
}
else if (Rs >= 1354) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 71) {
if (AFD < 39) {
return 8;
}
else if (AFD >= 39) {
return 1;
}
}
}
}
}
}
else if (SC >= 2242) {
if (tcor < 82) {
if (mu_mv_mag_sq < 189) {
if (mcTcor < 90) {
if (tcor < 78) {
if (AFD < 16) {
if (Cs < 755) {
return 4;
}
else if (Cs >= 755) {
return 1;
}
}
else if (AFD >= 16) {
if (Rs < 1735) {
if (mu_mv_mag_sq < 17) {
if (Rs < 1352) {
return 8;
}
else if (Rs >= 1352) {
return 4;
}
}
else if (mu_mv_mag_sq >= 17) {
if (SC < 2965) {
return 8;
}
else if (SC >= 2965) {
return 1;
}
}
}
else if (Rs >= 1735) {
if (ssDCval < 144) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
return 1;
}
}
else if (ssDCval >= 144) {
return 1;
}
}
}
}
else if (tcor >= 78) {
return 1;
}
}
else if (mcTcor >= 90) {
if (Cs < 683) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (Cs >= 683) {
if (RsCsDiff < 149) {
if (Cs < 1137) {
if (Rs < 2252) {
return 8;
}
else if (Rs >= 2252) {
return 1;
}
}
else if (Cs >= 1137) {
if (Rs < 1525) {
return 4;
}
else if (Rs >= 1525) {
return 8;
}
}
}
else if (RsCsDiff >= 149) {
if (MVDiff < 17) {
if (Cs < 829) {
return 2;
}
else if (Cs >= 829) {
return 8;
}
}
else if (MVDiff >= 17) {
if (mcTcor < 92) {
if (negBalance < 54) {
return 2;
}
else if (negBalance >= 54) {
return 8;
}
}
else if (mcTcor >= 92) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 189) {
return 1;
}
}
else if (tcor >= 82) {
if (RsDiff < 9) {
if (TSC < 10) {
if (Rs < 1268) {
return 1;
}
else if (Rs >= 1268) {
return 8;
}
}
else if (TSC >= 10) {
if (MVDiff < 43) {
return 1;
}
else if (MVDiff >= 43) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 111) {
return 8;
}
else if (ssDCval >= 111) {
if (diffMVdiffVal < 7) {
if (TSC < 8) {
if (Cs < 851) {
return 1;
}
else if (Cs >= 851) {
return 2;
}
}
else if (TSC >= 8) {
return 1;
}
}
else if (diffMVdiffVal >= 7) {
if (ssDCval < 151) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
return 1;
}
}
else if (ssDCval >= 151) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (Rs < 1746) {
if (Rs < 1134) {
if (refDCval < 63) {
if (SC < 1478) {
if (Rs < 549) {
if (Cs < 883) {
if (refDCval < 47) {
if (negBalance < 36) {
if (Rs < 522) {
return 1;
}
else if (Rs >= 522) {
return 4;
}
}
else if (negBalance >= 36) {
if (Rs < 525) {
return 8;
}
else if (Rs >= 525) {
return 4;
}
}
}
else if (refDCval >= 47) {
if (ssDCval < 58) {
if (Rs < 304) {
if (diffMVdiffVal < 2) {
if (Cs < 761) {
return 1;
}
else if (Cs >= 761) {
return 4;
}
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
else if (Rs >= 304) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (diffMVdiffVal < -5) {
if (Rs < 451) {
return 4;
}
else if (Rs >= 451) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
return 4;
}
}
}
}
else if (ssDCval >= 58) {
if (AFD < 7) {
if (posBalance < 46) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
else if (posBalance >= 46) {
if (Rs < 520) {
return 4;
}
else if (Rs >= 520) {
return 8;
}
}
}
else if (AFD >= 7) {
if (refDCval < 62) {
return 1;
}
else if (refDCval >= 62) {
return 2;
}
}
}
}
}
else if (Cs >= 883) {
if (gchDC < 1) {
if (Rs < 422) {
return 2;
}
else if (Rs >= 422) {
return 4;
}
}
else if (gchDC >= 1) {
if (tcor < 100) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
else if (tcor >= 100) {
return 4;
}
}
}
}
else if (Rs >= 549) {
if (MVDiff < 1) {
if (tcor < 97) {
if (Cs < 810) {
if (Rs < 636) {
return 8;
}
else if (Rs >= 636) {
return 4;
}
}
else if (Cs >= 810) {
return 1;
}
}
else if (tcor >= 97) {
return 8;
}
}
else if (MVDiff >= 1) {
if (negBalance < 17) {
if (Rs < 672) {
return 1;
}
else if (Rs >= 672) {
return 2;
}
}
else if (negBalance >= 17) {
if (MVDiff < 15) {
if (RsDiff < 2) {
if (negBalance < 30) {
if (SC < 1265) {
if (Rs < 578) {
return 8;
}
else if (Rs >= 578) {
return 4;
}
}
else if (SC >= 1265) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (Rs < 593) {
return 4;
}
else if (Rs >= 593) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 1;
}
}
}
}
}
else if (negBalance >= 30) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (negBalance < 39) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Rs < 644) {
return 8;
}
else if (Rs >= 644) {
return 4;
}
}
}
else if (negBalance >= 39) {
return 4;
}
}
}
}
else if (RsDiff >= 2) {
if (tcor < 93) {
if (SC < 1392) {
return 8;
}
else if (SC >= 1392) {
if (Rs < 631) {
return 2;
}
else if (Rs >= 631) {
if (Rs < 760) {
return 4;
}
else if (Rs >= 760) {
return 1;
}
}
}
}
else if (tcor >= 93) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 4;
}
}
}
}
}
else if (MVDiff >= 15) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
}
}
else if (SC >= 1478) {
if (ssDCval < 52) {
if (ssDCval < 49) {
if (SC < 1700) {
if (refDCval < 48) {
if (Rs < 832) {
if (ssDCval < 42) {
return 2;
}
else if (ssDCval >= 42) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
return 1;
}
}
}
else if (Rs >= 832) {
if (Rs < 920) {
return 8;
}
else if (Rs >= 920) {
return 1;
}
}
}
else if (refDCval >= 48) {
if (Rs < 639) {
return 1;
}
else if (Rs >= 639) {
return 8;
}
}
}
else if (SC >= 1700) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 4;
}
}
}
else if (ssDCval >= 49) {
if (Rs < 910) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (refDCval < 50) {
if (Cs < 994) {
if (Rs < 735) {
return 1;
}
else if (Rs >= 735) {
return 8;
}
}
else if (Cs >= 994) {
return 1;
}
}
else if (refDCval >= 50) {
if (refDCval < 51) {
return 1;
}
else if (refDCval >= 51) {
return 8;
}
}
}
}
else if (Rs >= 910) {
if (mu_mv_mag_sq < 114) {
if (Cs < 1028) {
if (posBalance < 33) {
if (Cs < 871) {
return 8;
}
else if (Cs >= 871) {
return 1;
}
}
else if (posBalance >= 33) {
return 1;
}
}
else if (Cs >= 1028) {
if (mu_mv_mag_sq < 102) {
return 4;
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 1101) {
return 8;
}
else if (Rs >= 1101) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
}
}
else if (ssDCval >= 52) {
if (TSCindex < 3) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (negBalance < 28) {
if (Cs < 550) {
if (Rs < 946) {
return 8;
}
else if (Rs >= 946) {
return 1;
}
}
else if (Cs >= 550) {
return 1;
}
}
else if (negBalance >= 28) {
if (Rs < 647) {
return 4;
}
else if (Rs >= 647) {
if (Rs < 831) {
return 8;
}
else if (Rs >= 831) {
return 1;
}
}
}
}
}
else if (TSCindex >= 3) {
if (Rs < 829) {
if (Rs < 686) {
if (AFD < 6) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
if (Rs < 671) {
return 1;
}
else if (Rs >= 671) {
return 4;
}
}
}
else if (AFD >= 6) {
if (SC < 1550) {
return 4;
}
else if (SC >= 1550) {
return 2;
}
}
}
else if (Rs >= 686) {
if (AFD < 5) {
if (negBalance < 27) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (Rs < 818) {
return 1;
}
else if (Rs >= 818) {
return 4;
}
}
}
else if (negBalance >= 27) {
return 8;
}
}
else if (AFD >= 5) {
if (negBalance < 39) {
return 4;
}
else if (negBalance >= 39) {
if (ssDCval < 56) {
return 2;
}
else if (ssDCval >= 56) {
return 8;
}
}
}
}
}
else if (Rs >= 829) {
if (Cs < 1106) {
if (Rs < 893) {
if (TSC < 4) {
if (Cs < 716) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 716) {
return 4;
}
}
else if (TSC >= 4) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
else if (Rs >= 893) {
if (mu_mv_mag_sq < 127) {
if (MVDiff < 8) {
if (diffRsCsdiff < -4) {
if (refDCval < 59) {
return 1;
}
else if (refDCval >= 59) {
if (Rs < 949) {
return 4;
}
else if (Rs >= 949) {
return 8;
}
}
}
else if (diffRsCsdiff >= -4) {
return 1;
}
}
else if (MVDiff >= 8) {
if (refDCval < 59) {
if (Rs < 1063) {
return 1;
}
else if (Rs >= 1063) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (refDCval >= 59) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
}
else if (Cs >= 1106) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 63) {
if (SC < 2150) {
if (ssDCval < 82) {
if (AFD < 1) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
return 8;
}
}
else if (AFD >= 1) {
if (ssDCval < 74) {
if (Rs < 753) {
if (mu_mv_mag_sq < 83) {
if (Cs < 807) {
if (MVDiff < 16) {
if (tcor < 97) {
if (diffMVdiffVal < -1) {
if (Rs < 577) {
return 1;
}
else if (Rs >= 577) {
if (Rs < 672) {
return 2;
}
else if (Rs >= 672) {
return 8;
}
}
}
else if (diffMVdiffVal >= -1) {
if (mu_mv_mag_sq < 63) {
return 4;
}
else if (mu_mv_mag_sq >= 63) {
if (Rs < 732) {
return 1;
}
else if (Rs >= 732) {
return 8;
}
}
}
}
else if (tcor >= 97) {
return 4;
}
}
else if (MVDiff >= 16) {
if (diffAFD < -5) {
if (Rs < 356) {
return 2;
}
else if (Rs >= 356) {
return 8;
}
}
else if (diffAFD >= -5) {
return 1;
}
}
}
else if (Cs >= 807) {
if (Rs < 621) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
return 2;
}
}
}
else if (Rs >= 621) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (Cs < 801) {
if (negBalance < 29) {
if (CsDiff < 1) {
if (Rs < 538) {
return 4;
}
else if (Rs >= 538) {
if (SC < 1461) {
return 8;
}
else if (SC >= 1461) {
if (Rs < 701) {
return 8;
}
else if (Rs >= 701) {
return 4;
}
}
}
}
else if (CsDiff >= 1) {
if (ssDCval < 71) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 422) {
return 1;
}
else if (Rs >= 422) {
if (Rs < 517) {
return 4;
}
else if (Rs >= 517) {
return 8;
}
}
}
}
else if (ssDCval >= 71) {
if (Rs < 607) {
return 8;
}
else if (Rs >= 607) {
return 1;
}
}
}
}
else if (negBalance >= 29) {
if (Rs < 661) {
if (Cs < 629) {
if (Rs < 553) {
return 2;
}
else if (Rs >= 553) {
return 4;
}
}
else if (Cs >= 629) {
return 8;
}
}
else if (Rs >= 661) {
if (Cs < 570) {
return 4;
}
else if (Cs >= 570) {
if (Cs < 786) {
return 1;
}
else if (Cs >= 786) {
return 8;
}
}
}
}
}
else if (Cs >= 801) {
if (RsDiff < 2) {
if (diffRsCsdiff < -3) {
if (Cs < 1016) {
return 1;
}
else if (Cs >= 1016) {
return 2;
}
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
else if (RsDiff >= 2) {
if (Rs < 688) {
return 1;
}
else if (Rs >= 688) {
return 2;
}
}
}
}
}
else if (Rs >= 753) {
if (mcTcor < 99) {
if (mcTcor < 98) {
if (AFD < 6) {
if (Rs < 905) {
if (ssDCval < 66) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (ssDCval >= 66) {
if (TSC < 5) {
if (Cs < 726) {
if (Cs < 619) {
return 4;
}
else if (Cs >= 619) {
return 1;
}
}
else if (Cs >= 726) {
return 4;
}
}
else if (TSC >= 5) {
if (Rs < 838) {
return 8;
}
else if (Rs >= 838) {
return 1;
}
}
}
}
else if (Rs >= 905) {
if (Cs < 1030) {
return 8;
}
else if (Cs >= 1030) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
}
}
}
else if (AFD >= 6) {
if (MVDiff < 5) {
if (Cs < 1104) {
return 8;
}
else if (Cs >= 1104) {
return 4;
}
}
else if (MVDiff >= 5) {
if (Rs < 898) {
if (RsDiff < 3) {
if (Rs < 802) {
return 4;
}
else if (Rs >= 802) {
return 1;
}
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (Rs >= 898) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
}
else if (mcTcor >= 98) {
if (RsDiff < 2) {
if (TSCindex < 3) {
if (mu_mv_mag_sq < 116) {
if (Cs < 796) {
return 2;
}
else if (Cs >= 796) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
return 4;
}
}
else if (TSCindex >= 3) {
if (Cs < 590) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 2;
}
}
else if (Cs >= 590) {
if (Cs < 1194) {
if (TSC < 4) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
return 8;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (Cs >= 1194) {
if (Rs < 835) {
return 1;
}
else if (Rs >= 835) {
return 2;
}
}
}
}
}
else if (RsDiff >= 2) {
if (diffRsCsdiff < 2) {
if (Cs < 1042) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 92) {
return 1;
}
else if (mu_mv_mag_sq >= 92) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 864) {
return 4;
}
else if (Rs >= 864) {
return 8;
}
}
}
else if (Cs >= 1042) {
return 2;
}
}
else if (diffRsCsdiff >= 2) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
else if (mcTcor >= 99) {
if (SC < 1808) {
if (Cs < 791) {
if (mu_mv_mag_sq < 90) {
return 1;
}
else if (mu_mv_mag_sq >= 90) {
if (ssDCval < 72) {
if (posBalance < 38) {
if (Rs < 1040) {
return 2;
}
else if (Rs >= 1040) {
return 1;
}
}
else if (posBalance >= 38) {
return 2;
}
}
else if (ssDCval >= 72) {
return 1;
}
}
}
else if (Cs >= 791) {
if (Rs < 880) {
return 4;
}
else if (Rs >= 880) {
return 1;
}
}
}
else if (SC >= 1808) {
if (negBalance < 18) {
if (Rs < 1081) {
return 8;
}
else if (Rs >= 1081) {
return 4;
}
}
else if (negBalance >= 18) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 74) {
if (mu_mv_mag_sq < 83) {
if (AFD < 10) {
if (Cs < 820) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 953) {
return 4;
}
else if (Rs >= 953) {
return 8;
}
}
}
}
else if (Cs >= 820) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 2;
}
}
}
else if (AFD >= 10) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
if (negBalance < 56) {
if (Rs < 933) {
return 4;
}
else if (Rs >= 933) {
return 2;
}
}
else if (negBalance >= 56) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (refDCval < 78) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (Rs < 820) {
if (Rs < 622) {
return 8;
}
else if (Rs >= 622) {
return 1;
}
}
else if (Rs >= 820) {
if (tcor < 96) {
if (Rs < 916) {
return 4;
}
else if (Rs >= 916) {
return 2;
}
}
else if (tcor >= 96) {
if (Rs < 1102) {
return 8;
}
else if (Rs >= 1102) {
return 1;
}
}
}
}
}
else if (refDCval >= 78) {
if (SC < 2001) {
if (mu_mv_mag_sq < 120) {
return 8;
}
else if (mu_mv_mag_sq >= 120) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (Rs < 941) {
return 8;
}
else if (Rs >= 941) {
return 4;
}
}
}
}
else if (SC >= 2001) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 82) {
if (SC < 1611) {
if (Rs < 650) {
if (mcTcor < 99) {
if (SC < 1323) {
if (Cs < 719) {
if (Rs < 509) {
return 8;
}
else if (Rs >= 509) {
return 2;
}
}
else if (Cs >= 719) {
return 1;
}
}
else if (SC >= 1323) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
if (Rs < 556) {
return 4;
}
else if (Rs >= 556) {
return 1;
}
}
}
}
else if (mcTcor >= 99) {
return 2;
}
}
else if (Rs >= 650) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (MVDiff < 5) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
else if (MVDiff >= 5) {
if (Cs < 738) {
return 8;
}
else if (Cs >= 738) {
return 1;
}
}
}
}
}
else if (SC >= 1611) {
if (negBalance < 28) {
return 2;
}
else if (negBalance >= 28) {
if (Rs < 1055) {
if (Cs < 1181) {
return 4;
}
else if (Cs >= 1181) {
return 2;
}
}
else if (Rs >= 1055) {
if (Rs < 1087) {
return 2;
}
else if (Rs >= 1087) {
return 8;
}
}
}
}
}
}
else if (SC >= 2150) {
if (tcor < 98) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 1006) {
if (Rs < 953) {
return 4;
}
else if (Rs >= 953) {
return 2;
}
}
else if (Rs >= 1006) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
}
else if (tcor >= 98) {
return 4;
}
}
}
}
else if (Rs >= 1134) {
if (Cs < 1073) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 69) {
if (SC < 2375) {
if (posBalance < 54) {
if (refDCval < 62) {
if (ssDCval < 47) {
if (Rs < 1326) {
if (Rs < 1271) {
return 8;
}
else if (Rs >= 1271) {
return 8;
}
}
else if (Rs >= 1326) {
if (Rs < 1526) {
return 4;
}
else if (Rs >= 1526) {
return 1;
}
}
}
else if (ssDCval >= 47) {
if (mu_mv_mag_sq < 82) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 1273) {
return 2;
}
else if (Rs >= 1273) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (Cs < 677) {
return 1;
}
else if (Cs >= 677) {
if (refDCval < 56) {
if (Rs < 1157) {
if (Cs < 831) {
return 8;
}
else if (Cs >= 831) {
return 1;
}
}
else if (Rs >= 1157) {
if (SC < 1921) {
return 4;
}
else if (SC >= 1921) {
return 8;
}
}
}
else if (refDCval >= 56) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (MVDiff < 3) {
if (ssDCval < 60) {
if (Rs < 1301) {
return 1;
}
else if (Rs >= 1301) {
return 8;
}
}
else if (ssDCval >= 60) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 62) {
if (mcTcor < 98) {
if (Rs < 1228) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
else if (Rs >= 1228) {
if (SC < 1906) {
return 2;
}
else if (SC >= 1906) {
if (Cs < 846) {
return 1;
}
else if (Cs >= 846) {
if (Rs < 1360) {
return 8;
}
else if (Rs >= 1360) {
return 4;
}
}
}
}
}
else if (mcTcor >= 98) {
if (Rs < 1682) {
if (Cs < 1046) {
return 1;
}
else if (Cs >= 1046) {
return 8;
}
}
else if (Rs >= 1682) {
return 2;
}
}
}
}
else if (posBalance >= 54) {
if (CsDiff < 4) {
if (SC < 2111) {
if (Cs < 777) {
return 8;
}
else if (Cs >= 777) {
return 1;
}
}
else if (SC >= 2111) {
return 8;
}
}
else if (CsDiff >= 4) {
if (Rs < 1280) {
return 4;
}
else if (Rs >= 1280) {
return 2;
}
}
}
}
else if (SC >= 2375) {
if (negBalance < 22) {
if (SC < 2655) {
return 1;
}
else if (SC >= 2655) {
if (MVDiff < 3) {
if (negBalance < 17) {
return 4;
}
else if (negBalance >= 17) {
if (Cs < 1031) {
return 8;
}
else if (Cs >= 1031) {
return 4;
}
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
else if (negBalance >= 22) {
if (MVDiff < 5) {
if (ssDCval < 50) {
if (diffRsCsdiff < 2) {
if (Rs < 1628) {
return 8;
}
else if (Rs >= 1628) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 1620) {
return 4;
}
else if (Rs >= 1620) {
return 4;
}
}
}
else if (ssDCval >= 50) {
if (Cs < 764) {
return 1;
}
else if (Cs >= 764) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (Rs < 1723) {
return 1;
}
else if (Rs >= 1723) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 56) {
if (Rs < 1177) {
if (Rs < 1159) {
return 2;
}
else if (Rs >= 1159) {
return 8;
}
}
else if (Rs >= 1177) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (ssDCval < 73) {
if (mu_mv_mag_sq < 104) {
if (Rs < 1387) {
return 8;
}
else if (Rs >= 1387) {
if (Cs < 837) {
return 1;
}
else if (Cs >= 837) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (posBalance < 33) {
return 1;
}
else if (posBalance >= 33) {
if (Rs < 1189) {
return 4;
}
else if (Rs >= 1189) {
return 8;
}
}
}
}
else if (ssDCval >= 73) {
if (RsCsDiff < 7) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
else if (RsCsDiff >= 7) {
if (negBalance < 69) {
if (Rs < 1533) {
return 8;
}
else if (Rs >= 1533) {
if (Cs < 956) {
return 1;
}
else if (Cs >= 956) {
return 8;
}
}
}
else if (negBalance >= 69) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1248) {
if (ssDCval < 67) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 1174) {
if (Cs < 679) {
return 1;
}
else if (Cs >= 679) {
return 8;
}
}
else if (Rs >= 1174) {
return 1;
}
}
}
else if (ssDCval >= 67) {
return 4;
}
}
else if (Rs >= 1248) {
if (Rs < 1662) {
if (Rs < 1335) {
if (Cs < 954) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 1261) {
return 8;
}
else if (Rs >= 1261) {
return 1;
}
}
}
else if (Cs >= 954) {
return 1;
}
}
else if (Rs >= 1335) {
return 8;
}
}
else if (Rs >= 1662) {
if (Cs < 933) {
return 1;
}
else if (Cs >= 933) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1073) {
if (RsCsDiff < 2) {
if (Rs < 1436) {
return 1;
}
else if (Rs >= 1436) {
if (Rs < 1464) {
return 4;
}
else if (Rs >= 1464) {
if (Cs < 1097) {
return 8;
}
else if (Cs >= 1097) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
if (AFD < 3) {
if (Rs < 1340) {
if (ssDCval < 62) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (ssDCval >= 62) {
if (negBalance < 34) {
if (Cs < 1103) {
if (Rs < 1254) {
return 4;
}
else if (Rs >= 1254) {
return 8;
}
}
else if (Cs >= 1103) {
if (SC < 2540) {
return 4;
}
else if (SC >= 2540) {
return 1;
}
}
}
else if (negBalance >= 34) {
return 1;
}
}
}
else if (Rs >= 1340) {
if (mu_mv_mag_sq < 122) {
if (Rs < 1659) {
return 8;
}
else if (Rs >= 1659) {
return 1;
}
}
else if (mu_mv_mag_sq >= 122) {
if (ssDCval < 76) {
if (ssDCval < 59) {
if (SC < 2869) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1183) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (posBalance < 37) {
if (negBalance < 19) {
if (Cs < 1158) {
return 8;
}
else if (Cs >= 1158) {
return 4;
}
}
else if (negBalance >= 19) {
if (Rs < 1741) {
if (ssDCval < 48) {
if (SC < 2788) {
if (SC < 2766) {
return 8;
}
else if (SC >= 2766) {
return 4;
}
}
else if (SC >= 2788) {
if (Cs < 1138) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1705) {
return 4;
}
else if (Rs >= 1705) {
return 4;
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1714) {
return 4;
}
else if (Rs >= 1714) {
return 4;
}
}
}
else if (Cs >= 1138) {
return 8;
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (Rs >= 1741) {
return 8;
}
}
}
else if (posBalance >= 37) {
if (SC < 2844) {
if (SC < 2754) {
return 8;
}
else if (SC >= 2754) {
if (Rs < 1684) {
return 4;
}
else if (Rs >= 1684) {
return 4;
}
}
}
else if (SC >= 2844) {
return 4;
}
}
}
}
else if (Cs >= 1183) {
return 8;
}
}
}
else if (SC >= 2869) {
return 4;
}
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (ssDCval >= 76) {
return 8;
}
}
}
}
else if (AFD >= 3) {
if (SC < 2698) {
if (RsCsDiff < 7) {
if (diffTSC < 1) {
if (mu_mv_mag_sq < 109) {
return 8;
}
else if (mu_mv_mag_sq >= 109) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 8;
}
}
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (RsCsDiff >= 7) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (SC < 2680) {
if (mu_mv_mag_sq < 121) {
if (Cs < 1223) {
if (tcor < 95) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
else if (tcor >= 95) {
if (Rs < 1217) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
else if (Rs >= 1217) {
return 8;
}
}
}
else if (Cs >= 1223) {
if (SC < 2435) {
return 1;
}
else if (SC >= 2435) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (diffRsCsdiff < 5) {
return 4;
}
else if (diffRsCsdiff >= 5) {
return 1;
}
}
}
else if (SC >= 2680) {
return 1;
}
}
}
}
else if (SC >= 2698) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 1746) {
if (TSC < 4) {
if (refDCval < 83) {
if (ssDCval < 56) {
if (refDCval < 47) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
else if (refDCval >= 47) {
if (refDCval < 54) {
if (CsDiff < 1) {
if (Rs < 1912) {
return 8;
}
else if (Rs >= 1912) {
return 8;
}
}
else if (CsDiff >= 1) {
if (TSCindex < 3) {
if (Cs < 1176) {
if (diffRsCsdiff < 1) {
if (negBalance < 22) {
if (tcor < 99) {
if (SC < 2845) {
return 4;
}
else if (SC >= 2845) {
if (Cs < 1169) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
if (Rs < 1780) {
if (posBalance < 41) {
if (SC < 2885) {
return 4;
}
else if (SC >= 2885) {
if (Rs < 1758) {
return 8;
}
else if (Rs >= 1758) {
return 4;
}
}
}
else if (posBalance >= 41) {
return 4;
}
}
else if (Rs >= 1780) {
if (negBalance < 18) {
if (Rs < 1790) {
return 8;
}
else if (Rs >= 1790) {
return 4;
}
}
else if (negBalance >= 18) {
return 4;
}
}
}
}
else if (Cs >= 1169) {
return 8;
}
}
}
else if (tcor >= 99) {
if (SC < 2860) {
return 8;
}
else if (SC >= 2860) {
if (negBalance < 19) {
if (Rs < 1793) {
return 4;
}
else if (Rs >= 1793) {
return 4;
}
}
else if (negBalance >= 19) {
if (Cs < 1100) {
return 4;
}
else if (Cs >= 1100) {
if (Rs < 1778) {
return 4;
}
else if (Rs >= 1778) {
if (Cs < 1140) {
return 8;
}
else if (Cs >= 1140) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 22) {
if (Rs < 1775) {
return 4;
}
else if (Rs >= 1775) {
if (Rs < 1810) {
return 4;
}
else if (Rs >= 1810) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
else if (Cs >= 1176) {
return 4;
}
}
else if (TSCindex >= 3) {
if (SC < 3312) {
return 4;
}
else if (SC >= 3312) {
return 8;
}
}
}
}
else if (refDCval >= 54) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (mcTcor < 99) {
if (posBalance < 47) {
if (Rs < 1767) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 1767) {
if (Cs < 1184) {
if (Cs < 956) {
return 1;
}
else if (Cs >= 956) {
if (RsDiff < 2) {
if (Rs < 2096) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 1997) {
return 4;
}
else if (Rs >= 1997) {
return 1;
}
}
}
else if (Rs >= 2096) {
return 8;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (Cs >= 1184) {
return 2;
}
}
}
else if (posBalance >= 47) {
if (SC < 2790) {
if (Rs < 1988) {
return 1;
}
else if (Rs >= 1988) {
return 4;
}
}
else if (SC >= 2790) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (Cs < 1093) {
if (SC < 2937) {
if (RsDiff < 3) {
if (refDCval < 82) {
if (Cs < 978) {
if (mcTcor < 100) {
return 1;
}
else if (mcTcor >= 100) {
if (Rs < 1931) {
return 1;
}
else if (Rs >= 1931) {
return 4;
}
}
}
else if (Cs >= 978) {
if (Cs < 1045) {
if (Rs < 1776) {
return 1;
}
else if (Rs >= 1776) {
return 4;
}
}
else if (Cs >= 1045) {
if (Rs < 1870) {
return 1;
}
else if (Rs >= 1870) {
return 4;
}
}
}
}
else if (refDCval >= 82) {
if (Rs < 1916) {
return 8;
}
else if (Rs >= 1916) {
return 4;
}
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (SC >= 2937) {
return 4;
}
}
else if (Cs >= 1093) {
if (Rs < 1931) {
if (negBalance < 19) {
return 1;
}
else if (negBalance >= 19) {
if (Rs < 1842) {
return 4;
}
else if (Rs >= 1842) {
if (Rs < 1885) {
return 1;
}
else if (Rs >= 1885) {
return 8;
}
}
}
}
else if (Rs >= 1931) {
if (Rs < 2173) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 2173) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 83) {
if (mu_mv_mag_sq < 100) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (Rs < 2077) {
return 4;
}
else if (Rs >= 2077) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 939) {
return 4;
}
else if (Cs >= 939) {
return 8;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 980) {
if (ssDCval < 71) {
if (posBalance < 48) {
if (SC < 2805) {
return 1;
}
else if (SC >= 2805) {
if (Cs < 846) {
return 2;
}
else if (Cs >= 846) {
if (SC < 3122) {
return 4;
}
else if (SC >= 3122) {
return 1;
}
}
}
}
else if (posBalance >= 48) {
return 2;
}
}
else if (ssDCval >= 71) {
if (mu_mv_mag_sq < 92) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (tcor < 96) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (tcor >= 96) {
if (Rs < 1827) {
return 8;
}
else if (Rs >= 1827) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (Rs < 2276) {
if (Rs < 2256) {
return 4;
}
else if (Rs >= 2256) {
return 8;
}
}
else if (Rs >= 2276) {
return 2;
}
}
}
}
else if (Cs >= 980) {
if (RsDiff < 9) {
if (negBalance < 44) {
if (tcor < 97) {
if (SC < 2970) {
return 4;
}
else if (SC >= 2970) {
return 1;
}
}
else if (tcor >= 97) {
return 8;
}
}
else if (negBalance >= 44) {
return 8;
}
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 97) {
if (SC < 1437) {
if (negBalance < 69) {
if (Rs < 417) {
if (MVDiff < 6) {
if (Rs < 308) {
if (Rs < 276) {
return 2;
}
else if (Rs >= 276) {
return 1;
}
}
else if (Rs >= 308) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 4;
}
}
}
else if (Rs >= 417) {
if (TSC < 4) {
if (ssDCval < 125) {
if (negBalance < 26) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (negBalance >= 26) {
if (posBalance < 34) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
else if (posBalance >= 34) {
if (Rs < 457) {
if (Rs < 443) {
return 2;
}
else if (Rs >= 443) {
return 1;
}
}
else if (Rs >= 457) {
return 2;
}
}
}
}
else if (ssDCval >= 125) {
if (SC < 1130) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (SC >= 1130) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
if (posBalance < 47) {
if (Rs < 720) {
return 4;
}
else if (Rs >= 720) {
return 1;
}
}
else if (posBalance >= 47) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 50) {
if (MVDiff < 4) {
if (Cs < 598) {
return 1;
}
else if (Cs >= 598) {
return 4;
}
}
else if (MVDiff >= 4) {
if (CsDiff < 4) {
if (SC < 1240) {
if (SC < 1026) {
return 8;
}
else if (SC >= 1026) {
return 2;
}
}
else if (SC >= 1240) {
return 4;
}
}
else if (CsDiff >= 4) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (posBalance < 54) {
if (Rs < 489) {
return 1;
}
else if (Rs >= 489) {
return 2;
}
}
else if (posBalance >= 54) {
if (mcTcor < 99) {
if (negBalance < 39) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (Rs < 693) {
return 8;
}
else if (Rs >= 693) {
return 2;
}
}
}
else if (negBalance >= 39) {
return 8;
}
}
else if (mcTcor >= 99) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 69) {
if (refDCval < 107) {
if (Rs < 569) {
return 8;
}
else if (Rs >= 569) {
return 1;
}
}
else if (refDCval >= 107) {
if (tcor < 85) {
return 2;
}
else if (tcor >= 85) {
return 8;
}
}
}
}
else if (SC >= 1437) {
if (RsCsDiff < 77) {
if (refDCval < 102) {
if (TSC < 5) {
if (Rs < 930) {
if (Rs < 611) {
return 8;
}
else if (Rs >= 611) {
return 4;
}
}
else if (Rs >= 930) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 1;
}
}
}
else if (TSC >= 5) {
if (tcor < 94) {
if (diffRsCsdiff < 3) {
if (diffTSC < -1) {
if (Rs < 1446) {
return 2;
}
else if (Rs >= 1446) {
return 1;
}
}
else if (diffTSC >= -1) {
if (negBalance < 52) {
if (ssDCval < 95) {
if (mu_mv_mag_sq < 5) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (mu_mv_mag_sq >= 5) {
return 1;
}
}
else if (ssDCval >= 95) {
return 4;
}
}
else if (negBalance >= 52) {
return 1;
}
}
}
else if (diffRsCsdiff >= 3) {
if (ssDCval < 96) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (tcor >= 94) {
if (MVDiff < 4) {
if (Rs < 941) {
return 4;
}
else if (Rs >= 941) {
if (Rs < 2249) {
if (Rs < 1217) {
return 8;
}
else if (Rs >= 1217) {
return 1;
}
}
else if (Rs >= 2249) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 99) {
if (refDCval < 88) {
return 4;
}
else if (refDCval >= 88) {
if (gchDC < 4) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
return 8;
}
}
else if (gchDC >= 4) {
return 1;
}
}
}
else if (ssDCval >= 99) {
if (Cs < 899) {
if (Rs < 1788) {
return 8;
}
else if (Rs >= 1788) {
return 1;
}
}
else if (Cs >= 899) {
if (Rs < 1527) {
return 2;
}
else if (Rs >= 1527) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 102) {
if (MVDiff < 5) {
if (RsCsDiff < 12) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 822) {
return 8;
}
else if (Rs >= 822) {
if (Rs < 1328) {
return 1;
}
else if (Rs >= 1328) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 12) {
if (Cs < 954) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (Rs < 1040) {
if (Rs < 673) {
if (Rs < 580) {
return 8;
}
else if (Rs >= 580) {
return 1;
}
}
else if (Rs >= 673) {
return 4;
}
}
else if (Rs >= 1040) {
return 8;
}
}
}
else if (Cs >= 954) {
if (Rs < 1610) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (Rs >= 1610) {
if (Rs < 2028) {
return 2;
}
else if (Rs >= 2028) {
return 8;
}
}
}
}
}
else if (MVDiff >= 5) {
if (refDCval < 151) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 20) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
return 1;
}
}
}
else if (MVDiff >= 20) {
if (Cs < 882) {
return 1;
}
else if (Cs >= 882) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (posBalance < 68) {
if (negBalance < 31) {
if (ssDCval < 112) {
return 1;
}
else if (ssDCval >= 112) {
if (Cs < 925) {
return 4;
}
else if (Cs >= 925) {
return 8;
}
}
}
else if (negBalance >= 31) {
if (mu_mv_mag_sq < 40) {
if (TSC < 6) {
if (Rs < 712) {
return 1;
}
else if (Rs >= 712) {
return 8;
}
}
else if (TSC >= 6) {
if (CsDiff < 4) {
if (SC < 1611) {
return 8;
}
else if (SC >= 1611) {
return 1;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Cs < 776) {
return 2;
}
else if (Cs >= 776) {
return 8;
}
}
}
}
}
else if (posBalance >= 68) {
if (MVDiff < 22) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 2;
}
}
else if (MVDiff >= 22) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (refDCval >= 151) {
if (posBalance < 60) {
return 2;
}
else if (posBalance >= 60) {
if (Rs < 1311) {
return 4;
}
else if (Rs >= 1311) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (Rs < 1777) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 79) {
return 1;
}
else if (MVDiff >= 79) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (CsDiff < 7) {
if (diffMVdiffVal < -2) {
if (Rs < 1684) {
return 2;
}
else if (Rs >= 1684) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (SC < 2511) {
return 8;
}
else if (SC >= 2511) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
}
else if (CsDiff >= 7) {
if (posBalance < 64) {
if (Cs < 800) {
return 4;
}
else if (Cs >= 800) {
if (diffMVdiffVal < 6) {
return 2;
}
else if (diffMVdiffVal >= 6) {
return 8;
}
}
}
else if (posBalance >= 64) {
return 1;
}
}
}
}
else if (Rs >= 1777) {
if (Cs < 796) {
if (mu_mv_mag_sq < 39) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 1;
}
}
else if (mu_mv_mag_sq >= 39) {
if (Cs < 624) {
return 2;
}
else if (Cs >= 624) {
return 1;
}
}
}
else if (Cs >= 796) {
if (Rs < 2008) {
return 2;
}
else if (Rs >= 2008) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
if (Rs < 2201) {
return 1;
}
else if (Rs >= 2201) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (ssDCval < 133) {
if (Cs < 1232) {
if (Rs < 340) {
if (mu_mv_mag_sq < 115) {
if (ssDCval < 124) {
return 4;
}
else if (ssDCval >= 124) {
return 2;
}
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 316) {
return 2;
}
else if (Rs >= 316) {
return 8;
}
}
}
else if (Rs >= 340) {
if (SC < 2695) {
if (refDCval < 97) {
if (SC < 1585) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
if (Rs < 775) {
return 1;
}
else if (Rs >= 775) {
return 8;
}
}
}
else if (SC >= 1585) {
if (CsDiff < 1) {
if (Rs < 1807) {
return 4;
}
else if (Rs >= 1807) {
return 8;
}
}
else if (CsDiff >= 1) {
if (Rs < 887) {
if (Cs < 970) {
return 1;
}
else if (Cs >= 970) {
return 4;
}
}
else if (Rs >= 887) {
if (Cs < 840) {
if (negBalance < 29) {
if (Rs < 2028) {
return 4;
}
else if (Rs >= 2028) {
return 1;
}
}
else if (negBalance >= 29) {
if (Cs < 777) {
if (Rs < 1839) {
return 8;
}
else if (Rs >= 1839) {
return 4;
}
}
else if (Cs >= 777) {
if (Rs < 1613) {
return 4;
}
else if (Rs >= 1613) {
return 8;
}
}
}
}
else if (Cs >= 840) {
return 8;
}
}
}
}
}
else if (refDCval >= 97) {
if (mu_mv_mag_sq < 106) {
if (Rs < 1045) {
if (AFD < 3) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (SC < 1124) {
return 1;
}
else if (SC >= 1124) {
return 2;
}
}
}
}
else if (AFD >= 3) {
if (TSC < 4) {
if (Cs < 853) {
if (Rs < 448) {
return 4;
}
else if (Rs >= 448) {
return 8;
}
}
else if (Cs >= 853) {
return 1;
}
}
else if (TSC >= 4) {
if (SC < 1139) {
return 1;
}
else if (SC >= 1139) {
return 8;
}
}
}
}
else if (Rs >= 1045) {
return 8;
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 1673) {
if (Cs < 778) {
if (diffTSC < -1) {
if (Cs < 648) {
return 8;
}
else if (Cs >= 648) {
if (posBalance < 43) {
if (Cs < 677) {
return 1;
}
else if (Cs >= 677) {
return 8;
}
}
else if (posBalance >= 43) {
if (Cs < 763) {
return 1;
}
else if (Cs >= 763) {
return 8;
}
}
}
}
else if (diffTSC >= -1) {
if (Rs < 557) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
return 8;
}
}
}
else if (Rs >= 557) {
if (ssDCval < 97) {
return 1;
}
else if (ssDCval >= 97) {
return 8;
}
}
}
}
else if (Cs >= 778) {
if (Rs < 369) {
if (Rs < 360) {
return 8;
}
else if (Rs >= 360) {
return 4;
}
}
else if (Rs >= 369) {
return 8;
}
}
}
else if (Rs >= 1673) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
else if (SC >= 2695) {
if (ssDCval < 122) {
if (Rs < 2145) {
if (diffMVdiffVal < -1) {
if (Rs < 1862) {
return 2;
}
else if (Rs >= 1862) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (Rs >= 2145) {
if (refDCval < 89) {
if (mu_mv_mag_sq < 125) {
if (Rs < 2173) {
return 4;
}
else if (Rs >= 2173) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
return 4;
}
}
else if (refDCval >= 89) {
return 8;
}
}
}
else if (ssDCval >= 122) {
return 2;
}
}
}
}
else if (Cs >= 1232) {
if (Rs < 670) {
if (ssDCval < 121) {
return 1;
}
else if (ssDCval >= 121) {
return 8;
}
}
else if (Rs >= 670) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1162) {
return 2;
}
else if (Rs >= 1162) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2348) {
if (Rs < 3556) {
if (RsDiff < 7) {
if (Cs < 1068) {
if (MVDiff < 8) {
if (Cs < 678) {
if (posBalance < 54) {
if (Rs < 2678) {
if (Cs < 371) {
if (Rs < 2633) {
return 8;
}
else if (Rs >= 2633) {
return 1;
}
}
else if (Cs >= 371) {
if (ssDCval < 102) {
if (refDCval < 49) {
if (Rs < 2523) {
return 4;
}
else if (Rs >= 2523) {
return 1;
}
}
else if (refDCval >= 49) {
if (Rs < 2579) {
return 1;
}
else if (Rs >= 2579) {
if (Cs < 582) {
if (Rs < 2666) {
return 4;
}
else if (Rs >= 2666) {
return 8;
}
}
else if (Cs >= 582) {
if (Rs < 2583) {
return 4;
}
else if (Rs >= 2583) {
return 1;
}
}
}
}
}
else if (ssDCval >= 102) {
if (Rs < 2594) {
return 8;
}
else if (Rs >= 2594) {
return 1;
}
}
}
}
else if (Rs >= 2678) {
if (posBalance < 33) {
return 2;
}
else if (posBalance >= 33) {
return 8;
}
}
}
else if (posBalance >= 54) {
if (tcor < 96) {
if (Rs < 2380) {
return 2;
}
else if (Rs >= 2380) {
if (ssDCval < 94) {
if (Cs < 546) {
return 4;
}
else if (Cs >= 546) {
if (Rs < 2617) {
return 4;
}
else if (Rs >= 2617) {
return 4;
}
}
}
else if (ssDCval >= 94) {
return 8;
}
}
}
else if (tcor >= 96) {
if (Cs < 644) {
return 2;
}
else if (Cs >= 644) {
return 1;
}
}
}
}
else if (Cs >= 678) {
if (negBalance < 26) {
if (ssDCval < 120) {
if (ssDCval < 96) {
if (negBalance < 22) {
if (ssDCval < 83) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Rs < 2563) {
return 4;
}
else if (Rs >= 2563) {
return 4;
}
}
}
else if (ssDCval >= 83) {
return 4;
}
}
else if (negBalance >= 22) {
if (Rs < 2739) {
if (Rs < 2567) {
if (Rs < 2440) {
return 2;
}
else if (Rs >= 2440) {
return 1;
}
}
else if (Rs >= 2567) {
return 8;
}
}
else if (Rs >= 2739) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 2;
}
}
}
}
else if (ssDCval >= 96) {
if (Rs < 3427) {
return 8;
}
else if (Rs >= 3427) {
if (Rs < 3535) {
return 2;
}
else if (Rs >= 3535) {
return 8;
}
}
}
}
else if (ssDCval >= 120) {
return 2;
}
}
else if (negBalance >= 26) {
if (refDCval < 74) {
if (Rs < 3085) {
if (mcTcor < 95) {
if (Rs < 2432) {
return 2;
}
else if (Rs >= 2432) {
return 4;
}
}
else if (mcTcor >= 95) {
if (refDCval < 63) {
if (mu_mv_mag_sq < 95) {
return 2;
}
else if (mu_mv_mag_sq >= 95) {
if (Cs < 928) {
return 2;
}
else if (Cs >= 928) {
return 4;
}
}
}
else if (refDCval >= 63) {
if (posBalance < 46) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (mu_mv_mag_sq < 109) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
else if (mu_mv_mag_sq >= 109) {
return 2;
}
}
}
else if (posBalance >= 46) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 1;
}
}
}
}
}
else if (Rs >= 3085) {
if (diffRsCsdiff < -17) {
return 2;
}
else if (diffRsCsdiff >= -17) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 3212) {
return 4;
}
else if (Rs >= 3212) {
return 2;
}
}
}
}
}
else if (refDCval >= 74) {
if (Rs < 2801) {
if (refDCval < 86) {
if (posBalance < 37) {
if (Cs < 1001) {
return 4;
}
else if (Cs >= 1001) {
return 8;
}
}
else if (posBalance >= 37) {
if (AFD < 8) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (AFD >= 8) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 4;
}
}
}
}
else if (refDCval >= 86) {
if (posBalance < 38) {
if (Cs < 764) {
return 2;
}
else if (Cs >= 764) {
if (Rs < 2594) {
return 4;
}
else if (Rs >= 2594) {
return 8;
}
}
}
else if (posBalance >= 38) {
return 8;
}
}
}
else if (Rs >= 2801) {
if (tcor < 96) {
if (ssDCval < 102) {
if (Cs < 964) {
if (Rs < 3476) {
return 4;
}
else if (Rs >= 3476) {
return 4;
}
}
else if (Cs >= 964) {
return 8;
}
}
else if (ssDCval >= 102) {
if (Cs < 978) {
return 1;
}
else if (Cs >= 978) {
return 2;
}
}
}
else if (tcor >= 96) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 8) {
if (negBalance < 28) {
if (ssDCval < 87) {
if (Rs < 2455) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Rs < 2425) {
return 4;
}
else if (Rs >= 2425) {
return 1;
}
}
}
else if (Rs >= 2455) {
return 4;
}
}
else if (ssDCval >= 87) {
if (posBalance < 75) {
if (Rs < 2836) {
return 1;
}
else if (Rs >= 2836) {
if (diffRsCsdiff < -13) {
return 8;
}
else if (diffRsCsdiff >= -13) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (mu_mv_mag_sq < 61) {
if (ssDCval < 101) {
if (Rs < 3376) {
return 4;
}
else if (Rs >= 3376) {
return 4;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (mu_mv_mag_sq >= 61) {
return 4;
}
}
}
}
}
else if (posBalance >= 75) {
return 2;
}
}
}
else if (negBalance >= 28) {
if (ssDCval < 69) {
if (Rs < 2843) {
if (Cs < 562) {
return 8;
}
else if (Cs >= 562) {
return 2;
}
}
else if (Rs >= 2843) {
return 4;
}
}
else if (ssDCval >= 69) {
if (Cs < 974) {
if (RsCsDiff < 23) {
if (Rs < 3017) {
return 8;
}
else if (Rs >= 3017) {
return 1;
}
}
else if (RsCsDiff >= 23) {
if (TSC < 8) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
if (posBalance < 36) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 195) {
return 1;
}
else if (Cs >= 195) {
return 2;
}
}
}
else if (posBalance >= 36) {
if (tcor < 94) {
if (Rs < 2822) {
return 1;
}
else if (Rs >= 2822) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 2987) {
return 8;
}
else if (Rs >= 2987) {
return 1;
}
}
}
}
else if (tcor >= 94) {
return 1;
}
}
}
}
else if (TSC >= 8) {
if (MVDiff < 40) {
return 8;
}
else if (MVDiff >= 40) {
if (MVDiff < 48) {
return 1;
}
else if (MVDiff >= 48) {
return 8;
}
}
}
}
}
else if (Cs >= 974) {
return 8;
}
}
}
}
}
else if (Cs >= 1068) {
if (Rs < 2859) {
if (ssDCval < 81) {
if (ssDCval < 64) {
if (mcTcor < 97) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (Rs < 2583) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Rs >= 2583) {
if (Rs < 2674) {
return 4;
}
else if (Rs >= 2674) {
return 2;
}
}
}
}
else if (mcTcor >= 97) {
return 8;
}
}
else if (ssDCval >= 64) {
if (posBalance < 47) {
if (Rs < 2637) {
if (Rs < 2517) {
return 8;
}
else if (Rs >= 2517) {
return 8;
}
}
else if (Rs >= 2637) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
else if (ssDCval >= 81) {
if (diffRsCsdiff < -16) {
if (Rs < 2830) {
if (AFD < 7) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (AFD >= 7) {
if (RsCsDiff < 57) {
return 8;
}
else if (RsCsDiff >= 57) {
return 4;
}
}
}
else if (Rs >= 2830) {
return 2;
}
}
else if (diffRsCsdiff >= -16) {
if (refDCval < 135) {
if (diffMVdiffVal < -1) {
if (ssDCval < 82) {
return 1;
}
else if (ssDCval >= 82) {
if (Cs < 1302) {
return 8;
}
else if (Cs >= 1302) {
return 1;
}
}
}
else if (diffMVdiffVal >= -1) {
if (diffTSC < -2) {
if (Rs < 2620) {
return 8;
}
else if (Rs >= 2620) {
return 2;
}
}
else if (diffTSC >= -2) {
return 8;
}
}
}
else if (refDCval >= 135) {
if (ssDCval < 137) {
if (SC < 3611) {
return 8;
}
else if (SC >= 3611) {
return 1;
}
}
else if (ssDCval >= 137) {
return 8;
}
}
}
}
}
else if (Rs >= 2859) {
if (refDCval < 108) {
if (RsDiff < 2) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
return 2;
}
}
else if (RsDiff >= 2) {
if (posBalance < 44) {
if (ssDCval < 104) {
if (RsCsDiff < 7) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
if (Rs < 3347) {
return 2;
}
else if (Rs >= 3347) {
return 4;
}
}
}
else if (RsCsDiff >= 7) {
if (diffAFD < -2) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (diffAFD >= -2) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
if (posBalance < 37) {
if (Cs < 1267) {
return 4;
}
else if (Cs >= 1267) {
return 2;
}
}
else if (posBalance >= 37) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 3400) {
return 4;
}
else if (Rs >= 3400) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (posBalance >= 44) {
if (ssDCval < 90) {
if (mcTcor < 95) {
if (Rs < 2903) {
return 2;
}
else if (Rs >= 2903) {
return 4;
}
}
else if (mcTcor >= 95) {
if (negBalance < 18) {
if (Rs < 3073) {
return 4;
}
else if (Rs >= 3073) {
return 8;
}
}
else if (negBalance >= 18) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
else if (refDCval >= 108) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (mu_mv_mag_sq < 87) {
if (SC < 3977) {
return 2;
}
else if (SC >= 3977) {
return 8;
}
}
else if (mu_mv_mag_sq >= 87) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (refDCval < 91) {
if (posBalance < 49) {
if (negBalance < 51) {
if (ssDCval < 69) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 71) {
return 2;
}
else if (mu_mv_mag_sq >= 71) {
if (posBalance < 44) {
if (Rs < 2800) {
return 4;
}
else if (Rs >= 2800) {
return 1;
}
}
else if (posBalance >= 44) {
return 2;
}
}
}
else if (MVDiff >= 5) {
if (CsDiff < 6) {
if (refDCval < 59) {
if (Cs < 1292) {
return 4;
}
else if (Cs >= 1292) {
return 8;
}
}
else if (refDCval >= 59) {
if (posBalance < 40) {
if (Rs < 2698) {
return 4;
}
else if (Rs >= 2698) {
if (Rs < 2936) {
return 1;
}
else if (Rs >= 2936) {
return 8;
}
}
}
else if (posBalance >= 40) {
if (SC < 3815) {
if (Rs < 2955) {
if (diffTSC < -1) {
if (diffRsCsdiff < 7) {
return 2;
}
else if (diffRsCsdiff >= 7) {
return 8;
}
}
else if (diffTSC >= -1) {
if (AFD < 15) {
if (ssDCval < 67) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
return 2;
}
}
else if (ssDCval >= 67) {
return 2;
}
}
else if (AFD >= 15) {
if (Cs < 558) {
return 4;
}
else if (Cs >= 558) {
return 2;
}
}
}
}
else if (Rs >= 2955) {
return 4;
}
}
else if (SC >= 3815) {
if (negBalance < 42) {
if (SC < 4446) {
if (RsDiff < 8) {
if (Rs < 3008) {
return 2;
}
else if (Rs >= 3008) {
return 4;
}
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (SC >= 4446) {
return 2;
}
}
else if (negBalance >= 42) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
}
}
else if (CsDiff >= 6) {
if (TSC < 8) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (Cs < 804) {
return 4;
}
else if (Cs >= 804) {
return 2;
}
}
}
else if (TSC >= 8) {
return 2;
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 61) {
if (SC < 3534) {
if (Cs < 283) {
if (Rs < 2834) {
return 2;
}
else if (Rs >= 2834) {
return 1;
}
}
else if (Cs >= 283) {
if (AFD < 18) {
if (diffRsCsdiff < 70) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
return 1;
}
}
else if (diffRsCsdiff >= 70) {
if (SC < 3460) {
return 4;
}
else if (SC >= 3460) {
return 1;
}
}
}
else if (AFD >= 18) {
if (Cs < 761) {
return 1;
}
else if (Cs >= 761) {
return 8;
}
}
}
}
else if (SC >= 3534) {
if (negBalance < 44) {
if (Cs < 1094) {
if (diffMVdiffVal < 1) {
if (RsDiff < 15) {
return 4;
}
else if (RsDiff >= 15) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
if (MVDiff < 18) {
if (TSC < 7) {
if (Rs < 2941) {
return 1;
}
else if (Rs >= 2941) {
return 4;
}
}
else if (TSC >= 7) {
if (Rs < 2877) {
return 4;
}
else if (Rs >= 2877) {
return 1;
}
}
}
else if (MVDiff >= 18) {
return 4;
}
}
}
else if (Cs >= 1094) {
if (ssDCval < 73) {
if (Rs < 3052) {
return 2;
}
else if (Rs >= 3052) {
return 1;
}
}
else if (ssDCval >= 73) {
if (Cs < 1107) {
return 2;
}
else if (Cs >= 1107) {
return 4;
}
}
}
}
else if (negBalance >= 44) {
if (posBalance < 46) {
if (Rs < 2955) {
return 1;
}
else if (Rs >= 2955) {
return 4;
}
}
else if (posBalance >= 46) {
if (AFD < 16) {
if (Rs < 2568) {
return 4;
}
else if (Rs >= 2568) {
return 2;
}
}
else if (AFD >= 16) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (posBalance < 39) {
if (RsDiff < 9) {
if (Rs < 3267) {
return 4;
}
else if (Rs >= 3267) {
return 4;
}
}
else if (RsDiff >= 9) {
return 1;
}
}
else if (posBalance >= 39) {
if (Cs < 895) {
if (Rs < 3063) {
return 2;
}
else if (Rs >= 3063) {
return 4;
}
}
else if (Cs >= 895) {
return 2;
}
}
}
}
}
else if (negBalance >= 51) {
if (mcTcor < 89) {
if (Cs < 1074) {
return 2;
}
else if (Cs >= 1074) {
return 8;
}
}
else if (mcTcor >= 89) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Cs < 745) {
return 2;
}
else if (Cs >= 745) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 49) {
if (MVDiff < 26) {
if (refDCval < 60) {
if (MVDiff < 18) {
if (Rs < 2454) {
return 2;
}
else if (Rs >= 2454) {
return 4;
}
}
else if (MVDiff >= 18) {
if (Rs < 2994) {
return 4;
}
else if (Rs >= 2994) {
return 2;
}
}
}
else if (refDCval >= 60) {
if (MVDiff < 11) {
if (negBalance < 36) {
if (posBalance < 53) {
if (Rs < 3107) {
return 4;
}
else if (Rs >= 3107) {
return 2;
}
}
else if (posBalance >= 53) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
return 1;
}
}
}
else if (negBalance >= 36) {
if (refDCval < 62) {
if (Rs < 2790) {
return 2;
}
else if (Rs >= 2790) {
return 4;
}
}
else if (refDCval >= 62) {
if (ssDCval < 92) {
if (diffRsCsdiff < -509) {
return 4;
}
else if (diffRsCsdiff >= -509) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
if (Cs < 1116) {
return 2;
}
else if (Cs >= 1116) {
return 1;
}
}
}
}
else if (ssDCval >= 92) {
return 8;
}
}
}
}
else if (MVDiff >= 11) {
if (negBalance < 42) {
if (posBalance < 64) {
if (Cs < 445) {
return 1;
}
else if (Cs >= 445) {
if (negBalance < 39) {
if (mu_mv_mag_sq < 25) {
return 4;
}
else if (mu_mv_mag_sq >= 25) {
if (diffMVdiffVal < -5) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
if (RsDiff < 11) {
if (diffRsCsdiff < -10) {
return 2;
}
else if (diffRsCsdiff >= -10) {
return 1;
}
}
else if (RsDiff >= 11) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
return 1;
}
}
}
}
else if (posBalance >= 64) {
if (Cs < 1301) {
return 2;
}
else if (Cs >= 1301) {
return 4;
}
}
}
else if (negBalance >= 42) {
if (tcor < 78) {
if (RsDiff < 13) {
if (SC < 3396) {
return 1;
}
else if (SC >= 3396) {
return 4;
}
}
else if (RsDiff >= 13) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
if (Cs < 728) {
return 2;
}
else if (Cs >= 728) {
if (Cs < 773) {
return 4;
}
else if (Cs >= 773) {
if (RsDiff < 20) {
return 2;
}
else if (RsDiff >= 20) {
return 4;
}
}
}
}
}
}
else if (tcor >= 78) {
if (diffAFD < -8) {
if (Rs < 2653) {
return 4;
}
else if (Rs >= 2653) {
return 2;
}
}
else if (diffAFD >= -8) {
if (diffAFD < -5) {
if (ssDCval < 87) {
return 2;
}
else if (ssDCval >= 87) {
return 4;
}
}
else if (diffAFD >= -5) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 26) {
if (CsDiff < 2) {
if (Cs < 139) {
return 2;
}
else if (Cs >= 139) {
return 4;
}
}
else if (CsDiff >= 2) {
if (RsDiff < 14) {
if (ssDCval < 70) {
if (TSC < 11) {
if (Rs < 2373) {
return 4;
}
else if (Rs >= 2373) {
return 2;
}
}
else if (TSC >= 11) {
return 1;
}
}
else if (ssDCval >= 70) {
if (mcTcor < 91) {
if (AFD < 22) {
if (Cs < 1221) {
return 8;
}
else if (Cs >= 1221) {
return 1;
}
}
else if (AFD >= 22) {
if (Rs < 2733) {
return 1;
}
else if (Rs >= 2733) {
return 8;
}
}
}
else if (mcTcor >= 91) {
if (tcor < 83) {
return 2;
}
else if (tcor >= 83) {
if (diffMVdiffVal < -11) {
return 2;
}
else if (diffMVdiffVal >= -11) {
if (diffMVdiffVal < 6) {
return 1;
}
else if (diffMVdiffVal >= 6) {
if (Rs < 2539) {
return 4;
}
else if (Rs >= 2539) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 14) {
if (TSC < 13) {
if (posBalance < 59) {
if (MVDiff < 42) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 2;
}
}
else if (MVDiff >= 42) {
if (Rs < 3241) {
return 4;
}
else if (Rs >= 3241) {
return 2;
}
}
}
else if (posBalance >= 59) {
return 2;
}
}
else if (TSC >= 13) {
if (diffMVdiffVal < 9) {
if (Cs < 932) {
return 1;
}
else if (Cs >= 932) {
return 8;
}
}
else if (diffMVdiffVal >= 9) {
if (Rs < 2796) {
return 2;
}
else if (Rs >= 2796) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 91) {
if (mcTcor < 69) {
if (MVDiff < 57) {
return 1;
}
else if (MVDiff >= 57) {
if (Rs < 2992) {
return 4;
}
else if (Rs >= 2992) {
return 8;
}
}
}
else if (mcTcor >= 69) {
if (mcTcor < 95) {
if (RsCsDiff < 207) {
if (RsDiff < 9) {
if (Cs < 477) {
if (TSC < 8) {
if (Cs < 439) {
return 8;
}
else if (Cs >= 439) {
return 2;
}
}
else if (TSC >= 8) {
return 1;
}
}
else if (Cs >= 477) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 2;
}
}
}
}
else if (RsDiff >= 9) {
if (MVDiff < 31) {
if (SC < 3526) {
if (MVDiff < 26) {
if (Rs < 2370) {
return 1;
}
else if (Rs >= 2370) {
return 2;
}
}
else if (MVDiff >= 26) {
if (Cs < 525) {
return 4;
}
else if (Cs >= 525) {
return 1;
}
}
}
else if (SC >= 3526) {
if (refDCval < 103) {
if (RsCsDiff < 131) {
return 8;
}
else if (RsCsDiff >= 131) {
if (mu_mv_mag_sq < 49) {
if (Rs < 3134) {
if (SC < 4131) {
return 4;
}
else if (SC >= 4131) {
return 2;
}
}
else if (Rs >= 3134) {
return 2;
}
}
else if (mu_mv_mag_sq >= 49) {
if (Cs < 1128) {
return 1;
}
else if (Cs >= 1128) {
return 8;
}
}
}
}
else if (refDCval >= 103) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (diffMVdiffVal < -6) {
if (Rs < 2742) {
return 4;
}
else if (Rs >= 2742) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 31) {
if (Cs < 748) {
if (refDCval < 133) {
if (SC < 3164) {
if (Rs < 2407) {
return 1;
}
else if (Rs >= 2407) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (diffAFD < -6) {
return 1;
}
else if (diffAFD >= -6) {
if (mu_mv_mag_sq < 21) {
if (MVDiff < 38) {
if (Rs < 2552) {
return 8;
}
else if (Rs >= 2552) {
return 1;
}
}
else if (MVDiff >= 38) {
return 8;
}
}
else if (mu_mv_mag_sq >= 21) {
return 1;
}
}
}
}
}
else if (SC >= 3164) {
if (Cs < 356) {
if (Rs < 3145) {
return 1;
}
else if (Rs >= 3145) {
return 4;
}
}
else if (Cs >= 356) {
return 8;
}
}
}
else if (refDCval >= 133) {
if (mu_mv_mag_sq < 19) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 8;
}
}
else if (mu_mv_mag_sq >= 19) {
return 4;
}
}
}
else if (Cs >= 748) {
if (TSC < 13) {
if (negBalance < 53) {
if (mcTcor < 94) {
if (Cs < 1205) {
return 1;
}
else if (Cs >= 1205) {
return 8;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (negBalance >= 53) {
if (Cs < 1226) {
return 8;
}
else if (Cs >= 1226) {
return 1;
}
}
}
else if (TSC >= 13) {
if (Rs < 2708) {
if (SC < 3240) {
return 1;
}
else if (SC >= 3240) {
return 2;
}
}
else if (Rs >= 2708) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 207) {
if (refDCval < 110) {
if (Cs < 750) {
if (RsCsDiff < 269) {
if (Cs < 418) {
return 2;
}
else if (Cs >= 418) {
if (mcTcor < 90) {
return 2;
}
else if (mcTcor >= 90) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
else if (RsCsDiff >= 269) {
if (Rs < 3437) {
return 2;
}
else if (Rs >= 3437) {
return 4;
}
}
}
else if (Cs >= 750) {
if (tcor < 83) {
if (AFD < 26) {
if (Rs < 3092) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (posBalance < 64) {
if (MVDiff < 66) {
return 4;
}
else if (MVDiff >= 66) {
if (Rs < 2954) {
return 4;
}
else if (Rs >= 2954) {
return 8;
}
}
}
else if (posBalance >= 64) {
if (Cs < 1157) {
return 2;
}
else if (Cs >= 1157) {
return 8;
}
}
}
}
else if (Rs >= 3092) {
if (ssDCval < 100) {
return 2;
}
else if (ssDCval >= 100) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
return 8;
}
}
}
}
else if (AFD >= 26) {
if (TSC < 17) {
if (Rs < 2658) {
return 1;
}
else if (Rs >= 2658) {
return 2;
}
}
else if (TSC >= 17) {
return 4;
}
}
}
else if (tcor >= 83) {
return 2;
}
}
}
else if (refDCval >= 110) {
if (RsCsDiff < 382) {
if (SC < 3398) {
if (TSC < 10) {
if (mu_mv_mag_sq < 19) {
if (Cs < 552) {
return 2;
}
else if (Cs >= 552) {
return 1;
}
}
else if (mu_mv_mag_sq >= 19) {
return 1;
}
}
else if (TSC >= 10) {
if (refDCval < 113) {
if (Rs < 2719) {
return 8;
}
else if (Rs >= 2719) {
return 4;
}
}
else if (refDCval >= 113) {
if (MVDiff < 24) {
if (Rs < 2611) {
return 2;
}
else if (Rs >= 2611) {
return 1;
}
}
else if (MVDiff >= 24) {
return 1;
}
}
}
}
else if (SC >= 3398) {
if (AFD < 19) {
if (SC < 3892) {
if (RsCsDiff < 242) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
return 4;
}
}
else if (RsCsDiff >= 242) {
if (Rs < 2579) {
return 8;
}
else if (Rs >= 2579) {
return 1;
}
}
}
else if (SC >= 3892) {
return 4;
}
}
else if (AFD >= 19) {
if (ssDCval < 120) {
if (CsDiff < 7) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 3194) {
return 2;
}
else if (Rs >= 3194) {
return 8;
}
}
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (ssDCval >= 120) {
if (ssDCval < 158) {
if (Cs < 1125) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (Cs >= 1125) {
if (ssDCval < 124) {
return 1;
}
else if (ssDCval >= 124) {
return 8;
}
}
}
else if (ssDCval >= 158) {
if (Rs < 3137) {
return 2;
}
else if (Rs >= 3137) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 382) {
if (diffAFD < 6) {
if (mu_mv_mag_sq < 34) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
if (tcor < 58) {
return 8;
}
else if (tcor >= 58) {
if (Cs < 766) {
if (ssDCval < 148) {
return 2;
}
else if (ssDCval >= 148) {
return 1;
}
}
else if (Cs >= 766) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 34) {
return 2;
}
}
else if (diffAFD >= 6) {
return 1;
}
}
}
}
}
else if (mcTcor >= 95) {
if (MVDiff < 13) {
if (diffMVdiffVal < -11) {
if (ssDCval < 110) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 2;
}
}
else if (ssDCval >= 110) {
if (RsDiff < 18) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
else if (RsDiff >= 18) {
return 2;
}
}
}
else if (diffMVdiffVal >= -11) {
if (Cs < 1117) {
if (Rs < 2757) {
if (mu_mv_mag_sq < 3) {
if (tcor < 88) {
if (Rs < 2579) {
if (Rs < 2461) {
return 2;
}
else if (Rs >= 2461) {
return 8;
}
}
else if (Rs >= 2579) {
return 4;
}
}
else if (tcor >= 88) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 2751) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (posBalance < 54) {
if (Rs < 2610) {
return 8;
}
else if (Rs >= 2610) {
return 1;
}
}
else if (posBalance >= 54) {
if (TSC < 4) {
if (Rs < 2563) {
return 2;
}
else if (Rs >= 2563) {
return 8;
}
}
else if (TSC >= 4) {
return 2;
}
}
}
}
else if (Rs >= 2751) {
return 4;
}
}
}
else if (Rs >= 2757) {
if (SC < 4480) {
if (Rs < 3058) {
if (Rs < 3056) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (gchDC < 2) {
if (Cs < 768) {
return 1;
}
else if (Cs >= 768) {
return 2;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (Rs >= 3056) {
return 4;
}
}
else if (Rs >= 3058) {
return 2;
}
}
else if (SC >= 4480) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
else if (Cs >= 1117) {
if (Rs < 2892) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
else if (Rs >= 2892) {
if (diffMVdiffVal < -4) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
if (Cs < 1121) {
return 8;
}
else if (Cs >= 1121) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 13) {
if (refDCval < 109) {
if (RsCsDiff < 142) {
if (diffAFD < -2) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
if (RsDiff < 10) {
if (Cs < 389) {
return 4;
}
else if (Cs >= 389) {
return 8;
}
}
else if (RsDiff >= 10) {
return 4;
}
}
}
else if (diffAFD >= -2) {
if (Cs < 1028) {
if (Rs < 2983) {
return 1;
}
else if (Rs >= 2983) {
return 8;
}
}
else if (Cs >= 1028) {
return 8;
}
}
}
else if (RsCsDiff >= 142) {
if (gchDC < 7) {
if (RsCsDiff < 183) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
return 8;
}
}
else if (RsCsDiff >= 183) {
if (SC < 2606) {
if (Rs < 2420) {
return 1;
}
else if (Rs >= 2420) {
return 2;
}
}
else if (SC >= 2606) {
return 2;
}
}
}
else if (gchDC >= 7) {
if (Rs < 2757) {
return 1;
}
else if (Rs >= 2757) {
return 8;
}
}
}
}
else if (refDCval >= 109) {
if (mu_mv_mag_sq < 26) {
if (SC < 3606) {
if (CsDiff < 3) {
if (diffAFD < -3) {
if (Cs < 124) {
return 2;
}
else if (Cs >= 124) {
return 1;
}
}
else if (diffAFD >= -3) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 2477) {
return 1;
}
else if (Rs >= 2477) {
return 2;
}
}
}
}
else if (CsDiff >= 3) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (MVDiff < 14) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 2;
}
}
else if (MVDiff >= 14) {
return 1;
}
}
}
}
else if (SC >= 3606) {
if (diffAFD < -2) {
if (CsDiff < 3) {
if (Rs < 3339) {
return 2;
}
else if (Rs >= 3339) {
return 1;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (diffAFD >= -2) {
if (AFD < 24) {
if (SC < 3836) {
if (diffRsCsdiff < 12) {
if (Rs < 3076) {
return 1;
}
else if (Rs >= 3076) {
return 2;
}
}
else if (diffRsCsdiff >= 12) {
return 8;
}
}
else if (SC >= 3836) {
return 8;
}
}
else if (AFD >= 24) {
if (Cs < 869) {
if (diffAFD < 8) {
return 2;
}
else if (diffAFD >= 8) {
return 1;
}
}
else if (Cs >= 869) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (refDCval < 120) {
if (RsDiff < 17) {
if (CsDiff < 3) {
if (negBalance < 46) {
if (Rs < 2604) {
return 1;
}
else if (Rs >= 2604) {
return 2;
}
}
else if (negBalance >= 46) {
if (Rs < 3441) {
return 8;
}
else if (Rs >= 3441) {
return 2;
}
}
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (RsDiff >= 17) {
if (Cs < 722) {
return 2;
}
else if (Cs >= 722) {
return 1;
}
}
}
else if (refDCval >= 120) {
if (MVDiff < 19) {
if (diffMVdiffVal < 1) {
if (Cs < 641) {
return 2;
}
else if (Cs >= 641) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
else if (MVDiff >= 19) {
if (tcor < 75) {
if (diffAFD < 6) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 1;
}
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (tcor >= 75) {
if (refDCval < 128) {
if (Cs < 874) {
return 2;
}
else if (Cs >= 874) {
return 1;
}
}
else if (refDCval >= 128) {
if (Cs < 1218) {
if (negBalance < 73) {
return 1;
}
else if (negBalance >= 73) {
if (Rs < 3045) {
return 1;
}
else if (Rs >= 3045) {
return 2;
}
}
}
else if (Cs >= 1218) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 3556) {
if (ssDCval < 93) {
if (posBalance < 48) {
if (Cs < 1091) {
if (TSC < 6) {
if (negBalance < 52) {
if (Rs < 5722) {
return 4;
}
else if (Rs >= 5722) {
return 8;
}
}
else if (negBalance >= 52) {
return 2;
}
}
else if (TSC >= 6) {
if (ssDCval < 75) {
if (mu_mv_mag_sq < 60) {
return 4;
}
else if (mu_mv_mag_sq >= 60) {
if (Rs < 3889) {
return 1;
}
else if (Rs >= 3889) {
return 4;
}
}
}
else if (ssDCval >= 75) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
if (Rs < 4010) {
return 8;
}
else if (Rs >= 4010) {
return 4;
}
}
}
}
}
else if (Cs >= 1091) {
if (diffRsCsdiff < -16) {
if (ssDCval < 78) {
if (SC < 5023) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (SC >= 5023) {
return 4;
}
}
else if (ssDCval >= 78) {
return 2;
}
}
else if (diffRsCsdiff >= -16) {
if (AFD < 3) {
if (refDCval < 76) {
if (SC < 4985) {
return 8;
}
else if (SC >= 4985) {
if (SC < 5244) {
return 4;
}
else if (SC >= 5244) {
if (Rs < 4029) {
return 8;
}
else if (Rs >= 4029) {
return 4;
}
}
}
}
else if (refDCval >= 76) {
return 8;
}
}
else if (AFD >= 3) {
if (ssDCval < 89) {
if (SC < 4833) {
if (Rs < 3581) {
return 8;
}
else if (Rs >= 3581) {
return 1;
}
}
else if (SC >= 4833) {
if (negBalance < 29) {
if (RsCsDiff < 27) {
return 4;
}
else if (RsCsDiff >= 27) {
if (Rs < 3712) {
return 4;
}
else if (Rs >= 3712) {
return 1;
}
}
}
else if (negBalance >= 29) {
if (Rs < 3868) {
if (RsDiff < 7) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 4;
}
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (Rs >= 3868) {
return 4;
}
}
}
}
else if (ssDCval >= 89) {
if (mu_mv_mag_sq < 63) {
if (tcor < 91) {
if (Cs < 1234) {
return 8;
}
else if (Cs >= 1234) {
return 2;
}
}
else if (tcor >= 91) {
if (MVDiff < 7) {
if (Rs < 4136) {
if (Cs < 1158) {
return 4;
}
else if (Cs >= 1158) {
return 4;
}
}
else if (Rs >= 4136) {
if (Cs < 1164) {
return 8;
}
else if (Cs >= 1164) {
if (Cs < 1186) {
return 4;
}
else if (Cs >= 1186) {
if (MVDiff < 5) {
if (Rs < 4264) {
return 8;
}
else if (Rs >= 4264) {
return 8;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 4224) {
return 4;
}
else if (Rs >= 4224) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 63) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 48) {
if (Cs < 692) {
if (ssDCval < 82) {
if (MVDiff < 81) {
if (Rs < 3612) {
return 2;
}
else if (Rs >= 3612) {
return 4;
}
}
else if (MVDiff >= 81) {
return 2;
}
}
else if (ssDCval >= 82) {
if (Cs < 593) {
return 2;
}
else if (Cs >= 593) {
if (Rs < 3605) {
return 2;
}
else if (Rs >= 3605) {
return 8;
}
}
}
}
else if (Cs >= 692) {
if (Rs < 4295) {
if (refDCval < 74) {
if (Cs < 1059) {
return 4;
}
else if (Cs >= 1059) {
return 2;
}
}
else if (refDCval >= 74) {
if (negBalance < 35) {
if (posBalance < 52) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (Rs < 3855) {
return 4;
}
else if (Rs >= 3855) {
return 4;
}
}
}
else if (posBalance >= 52) {
return 2;
}
}
else if (negBalance >= 35) {
if (posBalance < 64) {
return 2;
}
else if (posBalance >= 64) {
if (tcor < 80) {
return 1;
}
else if (tcor >= 80) {
return 2;
}
}
}
}
}
else if (Rs >= 4295) {
if (AFD < 16) {
if (SC < 5627) {
if (Rs < 4508) {
if (Rs < 4397) {
return 4;
}
else if (Rs >= 4397) {
return 4;
}
}
else if (Rs >= 4508) {
return 8;
}
}
else if (SC >= 5627) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (AFD >= 16) {
return 1;
}
}
}
}
}
else if (ssDCval >= 93) {
if (MVDiff < 7) {
if (ssDCval < 108) {
if (ssDCval < 100) {
if (negBalance < 31) {
if (Cs < 898) {
return 2;
}
else if (Cs >= 898) {
return 4;
}
}
else if (negBalance >= 31) {
return 2;
}
}
else if (ssDCval >= 100) {
if (Cs < 1021) {
if (Rs < 4759) {
return 2;
}
else if (Rs >= 4759) {
return 8;
}
}
else if (Cs >= 1021) {
if (SC < 4861) {
return 8;
}
else if (SC >= 4861) {
if (RsDiff < 15) {
return 4;
}
else if (RsDiff >= 15) {
return 8;
}
}
}
}
}
else if (ssDCval >= 108) {
if (ssDCval < 142) {
if (diffMVdiffVal < -3) {
if (refDCval < 134) {
if (SC < 4246) {
return 8;
}
else if (SC >= 4246) {
return 2;
}
}
else if (refDCval >= 134) {
if (MVDiff < 5) {
if (Rs < 3778) {
return 8;
}
else if (Rs >= 3778) {
if (Cs < 1227) {
return 2;
}
else if (Cs >= 1227) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (Rs < 4747) {
return 1;
}
else if (Rs >= 4747) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -3) {
if (tcor < 85) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (tcor >= 85) {
return 2;
}
}
}
else if (ssDCval >= 142) {
if (Rs < 3965) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 119) {
if (mcTcor < 100) {
return 2;
}
else if (mcTcor >= 100) {
return 1;
}
}
else if (mu_mv_mag_sq >= 119) {
return 1;
}
}
else if (CsDiff >= 3) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
if (Rs < 3652) {
return 2;
}
else if (Rs >= 3652) {
return 1;
}
}
}
}
else if (Rs >= 3965) {
return 2;
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 137) {
if (Cs < 793) {
if (Rs < 4205) {
if (RsCsDiff < 392) {
if (gchDC < 2) {
if (MVDiff < 21) {
if (posBalance < 37) {
if (Rs < 3750) {
return 1;
}
else if (Rs >= 3750) {
return 8;
}
}
else if (posBalance >= 37) {
return 2;
}
}
else if (MVDiff >= 21) {
if (SC < 4547) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (SC >= 4547) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (Cs < 500) {
if (SC < 4111) {
return 2;
}
else if (SC >= 4111) {
return 4;
}
}
else if (Cs >= 500) {
if (negBalance < 62) {
if (mu_mv_mag_sq < 46) {
return 8;
}
else if (mu_mv_mag_sq >= 46) {
return 2;
}
}
else if (negBalance >= 62) {
return 2;
}
}
}
}
else if (RsCsDiff >= 392) {
if (gchDC < 6) {
return 2;
}
else if (gchDC >= 6) {
if (Rs < 4020) {
return 1;
}
else if (Rs >= 4020) {
return 2;
}
}
}
}
else if (Rs >= 4205) {
if (Cs < 622) {
return 2;
}
else if (Cs >= 622) {
if (Cs < 627) {
return 8;
}
else if (Cs >= 627) {
if (ssDCval < 134) {
if (refDCval < 107) {
return 1;
}
else if (refDCval >= 107) {
return 2;
}
}
else if (ssDCval >= 134) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
}
}
else if (Cs >= 793) {
if (MVDiff < 17) {
if (Rs < 3913) {
if (RsDiff < 7) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 4;
}
}
}
}
else if (RsDiff >= 7) {
if (gchDC < 1) {
if (SC < 4595) {
if (Rs < 3617) {
return 4;
}
else if (Rs >= 3617) {
return 2;
}
}
else if (SC >= 4595) {
return 8;
}
}
else if (gchDC >= 1) {
if (Cs < 1221) {
if (ssDCval < 135) {
return 2;
}
else if (ssDCval >= 135) {
return 4;
}
}
else if (Cs >= 1221) {
if (Cs < 1235) {
return 4;
}
else if (Cs >= 1235) {
if (Rs < 3759) {
return 2;
}
else if (Rs >= 3759) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3913) {
if (refDCval < 134) {
if (Cs < 799) {
return 1;
}
else if (Cs >= 799) {
if (tcor < 67) {
if (Rs < 4193) {
return 4;
}
else if (Rs >= 4193) {
return 2;
}
}
else if (tcor >= 67) {
if (negBalance < 23) {
if (Rs < 4341) {
return 4;
}
else if (Rs >= 4341) {
return 2;
}
}
else if (negBalance >= 23) {
return 2;
}
}
}
}
else if (refDCval >= 134) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 1;
}
}
}
}
else if (MVDiff >= 17) {
if (mu_mv_mag_sq < 9) {
if (tcor < 82) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
if (TSC < 13) {
if (diffAFD < 6) {
if (Rs < 5809) {
if (ssDCval < 121) {
if (RsDiff < 22) {
if (diffMVdiffVal < -7) {
if (Cs < 1041) {
return 8;
}
else if (Cs >= 1041) {
return 2;
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
else if (RsDiff >= 22) {
return 2;
}
}
else if (ssDCval >= 121) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (MVDiff < 72) {
if (negBalance < 50) {
if (Rs < 4507) {
return 8;
}
else if (Rs >= 4507) {
if (Rs < 5071) {
return 2;
}
else if (Rs >= 5071) {
return 8;
}
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (MVDiff >= 72) {
return 2;
}
}
}
}
else if (Rs >= 5809) {
if (AFD < 30) {
return 2;
}
else if (AFD >= 30) {
if (Rs < 6220) {
return 2;
}
else if (Rs >= 6220) {
return 8;
}
}
}
}
else if (diffAFD >= 6) {
return 2;
}
}
else if (TSC >= 13) {
if (tcor < 71) {
return 2;
}
else if (tcor >= 71) {
return 1;
}
}
}
}
else if (tcor >= 82) {
if (ssDCval < 122) {
if (CsDiff < 5) {
if (MVDiff < 79) {
if (RsDiff < 18) {
return 8;
}
else if (RsDiff >= 18) {
if (Rs < 4626) {
return 1;
}
else if (Rs >= 4626) {
if (Rs < 4787) {
return 8;
}
else if (Rs >= 4787) {
return 2;
}
}
}
}
else if (MVDiff >= 79) {
return 2;
}
}
else if (CsDiff >= 5) {
if (Cs < 819) {
return 1;
}
else if (Cs >= 819) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
return 8;
}
}
}
}
else if (ssDCval >= 122) {
if (ssDCval < 136) {
if (TSC < 12) {
return 1;
}
else if (TSC >= 12) {
if (Rs < 4166) {
return 8;
}
else if (Rs >= 4166) {
if (Rs < 5602) {
return 1;
}
else if (Rs >= 5602) {
return 2;
}
}
}
}
else if (ssDCval >= 136) {
if (Cs < 902) {
return 1;
}
else if (Cs >= 902) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (AFD < 28) {
if (AFD < 16) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
if (ssDCval < 120) {
if (TSC < 7) {
if (Rs < 3949) {
return 4;
}
else if (Rs >= 3949) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (ssDCval >= 120) {
if (Rs < 4421) {
if (Cs < 888) {
return 2;
}
else if (Cs >= 888) {
return 1;
}
}
else if (Rs >= 4421) {
return 4;
}
}
}
}
else if (AFD >= 16) {
if (diffAFD < -3) {
if (SC < 5271) {
return 2;
}
else if (SC >= 5271) {
if (Rs < 4949) {
return 8;
}
else if (Rs >= 4949) {
return 1;
}
}
}
else if (diffAFD >= -3) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
if (diffAFD < 8) {
return 2;
}
else if (diffAFD >= 8) {
return 1;
}
}
}
}
}
else if (AFD >= 28) {
if (RsDiff < 19) {
return 4;
}
else if (RsDiff >= 19) {
if (tcor < 68) {
if (TSC < 19) {
return 2;
}
else if (TSC >= 19) {
return 4;
}
}
else if (tcor >= 68) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 137) {
if (mcTcor < 92) {
if (posBalance < 54) {
if (MVDiff < 51) {
return 1;
}
else if (MVDiff >= 51) {
return 2;
}
}
else if (posBalance >= 54) {
if (posBalance < 67) {
return 4;
}
else if (posBalance >= 67) {
if (Rs < 3668) {
return 8;
}
else if (Rs >= 3668) {
return 1;
}
}
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 35) {
if (RsDiff < 18) {
if (mcTcor < 94) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
if (Rs < 3932) {
return 1;
}
else if (Rs >= 3932) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (MVDiff < 52) {
if (negBalance < 56) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (negBalance >= 56) {
return 2;
}
}
else if (MVDiff >= 52) {
return 8;
}
}
}
else if (RsDiff >= 18) {
if (Rs < 4083) {
if (MVDiff < 16) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 2;
}
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (Rs >= 4083) {
if (ssDCval < 155) {
if (posBalance < 64) {
if (MVDiff < 39) {
return 2;
}
else if (MVDiff >= 39) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 8;
}
}
}
else if (posBalance >= 64) {
return 1;
}
}
else if (ssDCval >= 155) {
if (Rs < 5809) {
return 1;
}
else if (Rs >= 5809) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (Cs < 1172) {
if (CsDiff < 5) {
if (diffMVdiffVal < -27) {
return 1;
}
else if (diffMVdiffVal >= -27) {
return 2;
}
}
else if (CsDiff >= 5) {
if (Rs < 4959) {
return 1;
}
else if (Rs >= 4959) {
return 2;
}
}
}
else if (Cs >= 1172) {
if (RsDiff < 14) {
return 1;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1308) {
if (mu_mv_mag_sq < 108) {
if (RsCsDiff < 101) {
if (refDCval < 90) {
if (negBalance < 38) {
if (Rs < 8154) {
if (Rs < 796) {
if (Rs < 308) {
return 4;
}
else if (Rs >= 308) {
if (MVDiff < 3) {
if (Cs < 1328) {
return 1;
}
else if (Cs >= 1328) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Cs < 2451) {
return 1;
}
else if (Cs >= 2451) {
return 2;
}
}
}
}
else if (Rs >= 796) {
if (posBalance < 45) {
if (MVDiff < 7) {
if (Rs < 3572) {
if (RsCsDiff < 55) {
if (ssDCval < 64) {
if (ssDCval < 48) {
return 8;
}
else if (ssDCval >= 48) {
if (negBalance < 32) {
if (posBalance < 42) {
if (Rs < 2593) {
return 2;
}
else if (Rs >= 2593) {
return 4;
}
}
else if (posBalance >= 42) {
if (Rs < 1543) {
return 4;
}
else if (Rs >= 1543) {
return 8;
}
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (Rs < 3440) {
if (Cs < 1776) {
if (mu_mv_mag_sq < 18) {
return 1;
}
else if (mu_mv_mag_sq >= 18) {
return 8;
}
}
else if (Cs >= 1776) {
if (Rs < 2565) {
return 1;
}
else if (Rs >= 2565) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 2859) {
return 1;
}
else if (Rs >= 2859) {
return 4;
}
}
}
}
}
else if (Rs >= 3440) {
return 2;
}
}
}
else if (RsCsDiff >= 55) {
if (Cs < 1533) {
if (Rs < 2480) {
if (Rs < 1932) {
return 8;
}
else if (Rs >= 1932) {
return 1;
}
}
else if (Rs >= 2480) {
return 2;
}
}
else if (Cs >= 1533) {
if (Rs < 3459) {
return 1;
}
else if (Rs >= 3459) {
return 4;
}
}
}
}
else if (Rs >= 3572) {
if (SC < 5865) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
else if (SC >= 5865) {
if (CsDiff < 5) {
if (Cs < 1802) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
else if (Cs >= 1802) {
if (diffRsCsdiff < -14) {
if (Rs < 4730) {
return 4;
}
else if (Rs >= 4730) {
return 8;
}
}
else if (diffRsCsdiff >= -14) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (AFD < 9) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 5413) {
if (Cs < 2247) {
if (Rs < 4253) {
return 8;
}
else if (Rs >= 4253) {
return 4;
}
}
else if (Cs >= 2247) {
return 8;
}
}
else if (Rs >= 5413) {
return 2;
}
}
}
else if (AFD >= 9) {
return 1;
}
}
}
}
}
else if (MVDiff >= 7) {
if (refDCval < 79) {
if (ssDCval < 64) {
if (Rs < 3262) {
if (tcor < 95) {
if (posBalance < 43) {
if (posBalance < 28) {
if (Rs < 2218) {
return 1;
}
else if (Rs >= 2218) {
return 8;
}
}
else if (posBalance >= 28) {
if (Rs < 1082) {
return 8;
}
else if (Rs >= 1082) {
return 2;
}
}
}
else if (posBalance >= 43) {
return 4;
}
}
else if (tcor >= 95) {
if (Rs < 829) {
return 1;
}
else if (Rs >= 829) {
return 8;
}
}
}
else if (Rs >= 3262) {
if (Cs < 3315) {
return 4;
}
else if (Cs >= 3315) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
if (Rs < 2622) {
return 1;
}
else if (Rs >= 2622) {
return 4;
}
}
}
}
else if (refDCval >= 79) {
if (Rs < 3844) {
return 8;
}
else if (Rs >= 3844) {
return 4;
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2684) {
if (mu_mv_mag_sq < 80) {
if (SC < 2597) {
if (tcor < 90) {
if (AFD < 11) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 4;
}
}
else if (AFD >= 11) {
if (Rs < 853) {
return 8;
}
else if (Rs >= 853) {
return 2;
}
}
}
else if (tcor >= 90) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
}
else if (SC >= 2597) {
if (TSC < 6) {
if (MVDiff < 5) {
if (Rs < 1620) {
if (Rs < 1261) {
return 8;
}
else if (Rs >= 1261) {
return 4;
}
}
else if (Rs >= 1620) {
return 8;
}
}
else if (MVDiff >= 5) {
if (mu_mv_mag_sq < 64) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 1;
}
}
else if (mu_mv_mag_sq >= 64) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
if (Rs < 1227) {
return 8;
}
else if (Rs >= 1227) {
return 1;
}
}
}
}
}
}
else if (TSC >= 6) {
if (tcor < 90) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (Cs < 1877) {
return 8;
}
else if (Cs >= 1877) {
if (Rs < 2188) {
return 4;
}
else if (Rs >= 2188) {
return 8;
}
}
}
}
else if (tcor >= 90) {
if (Rs < 2500) {
if (Cs < 1374) {
return 1;
}
else if (Cs >= 1374) {
return 8;
}
}
else if (Rs >= 2500) {
if (Rs < 2611) {
return 2;
}
else if (Rs >= 2611) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (Cs < 1973) {
if (Cs < 1683) {
if (Rs < 1319) {
if (refDCval < 57) {
if (Rs < 953) {
return 8;
}
else if (Rs >= 953) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (refDCval >= 57) {
if (Cs < 1402) {
return 2;
}
else if (Cs >= 1402) {
return 8;
}
}
}
else if (Rs >= 1319) {
if (Rs < 2346) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (mu_mv_mag_sq < 107) {
return 8;
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 1764) {
return 4;
}
else if (Rs >= 1764) {
return 8;
}
}
}
}
else if (Rs >= 2346) {
if (Cs < 1529) {
return 4;
}
else if (Cs >= 1529) {
return 8;
}
}
}
}
else if (Cs >= 1683) {
if (ssDCval < 58) {
if (Rs < 1153) {
return 8;
}
else if (Rs >= 1153) {
return 4;
}
}
else if (ssDCval >= 58) {
return 8;
}
}
}
else if (Cs >= 1973) {
if (Rs < 1949) {
if (CsDiff < 4) {
if (mu_mv_mag_sq < 95) {
return 1;
}
else if (mu_mv_mag_sq >= 95) {
if (Rs < 1259) {
return 8;
}
else if (Rs >= 1259) {
return 1;
}
}
}
else if (CsDiff >= 4) {
if (Cs < 3807) {
return 8;
}
else if (Cs >= 3807) {
return 1;
}
}
}
else if (Rs >= 1949) {
if (Cs < 2481) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
else if (Cs >= 2481) {
if (Cs < 3247) {
return 4;
}
else if (Cs >= 3247) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2684) {
if (Cs < 1855) {
if (diffAFD < -1) {
if (MVDiff < 12) {
if (Rs < 3066) {
if (Rs < 2993) {
return 2;
}
else if (Rs >= 2993) {
return 4;
}
}
else if (Rs >= 3066) {
return 2;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (diffAFD >= -1) {
if (RsCsDiff < 48) {
if (SC < 4708) {
if (SC < 4545) {
return 4;
}
else if (SC >= 4545) {
if (Rs < 3062) {
return 8;
}
else if (Rs >= 3062) {
return 4;
}
}
}
else if (SC >= 4708) {
return 4;
}
}
else if (RsCsDiff >= 48) {
if (SC < 5008) {
if (Cs < 1462) {
if (Rs < 3559) {
return 2;
}
else if (Rs >= 3559) {
return 4;
}
}
else if (Cs >= 1462) {
if (Rs < 3328) {
return 4;
}
else if (Rs >= 3328) {
return 2;
}
}
}
else if (SC >= 5008) {
return 2;
}
}
}
}
else if (Cs >= 1855) {
if (CsDiff < 5) {
if (Rs < 5236) {
if (Cs < 2065) {
if (Rs < 3065) {
return 4;
}
else if (Rs >= 3065) {
if (Rs < 3817) {
return 8;
}
else if (Rs >= 3817) {
return 2;
}
}
}
else if (Cs >= 2065) {
return 8;
}
}
else if (Rs >= 5236) {
return 4;
}
}
else if (CsDiff >= 5) {
if (Rs < 3492) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
if (Rs < 2822) {
return 4;
}
else if (Rs >= 2822) {
return 8;
}
}
}
else if (Rs >= 3492) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 8154) {
return 2;
}
}
else if (negBalance >= 38) {
if (Rs < 1454) {
if (RsCsDiff < 33) {
if (MVDiff < 6) {
if (TSC < 6) {
if (Cs < 1616) {
if (RsDiff < 3) {
if (mcTcor < 96) {
return 4;
}
else if (mcTcor >= 96) {
return 8;
}
}
else if (RsDiff >= 3) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (Rs < 1020) {
return 2;
}
else if (Rs >= 1020) {
return 4;
}
}
}
}
else if (Cs >= 1616) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
else if (TSC >= 6) {
if (diffTSC < 1) {
if (AFD < 10) {
if (Cs < 1319) {
return 1;
}
else if (Cs >= 1319) {
return 8;
}
}
else if (AFD >= 10) {
if (Rs < 1124) {
return 1;
}
else if (Rs >= 1124) {
return 4;
}
}
}
else if (diffTSC >= 1) {
if (Cs < 1588) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (Cs >= 1588) {
if (Rs < 1399) {
return 1;
}
else if (Rs >= 1399) {
return 4;
}
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 59) {
if (posBalance < 37) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (posBalance >= 37) {
if (Cs < 1361) {
return 8;
}
else if (Cs >= 1361) {
return 2;
}
}
}
else if (ssDCval >= 59) {
if (posBalance < 59) {
if (MVDiff < 16) {
if (refDCval < 77) {
if (Cs < 1319) {
if (Cs < 1318) {
return 1;
}
else if (Cs >= 1318) {
return 4;
}
}
else if (Cs >= 1319) {
return 1;
}
}
else if (refDCval >= 77) {
if (Rs < 1181) {
return 2;
}
else if (Rs >= 1181) {
return 4;
}
}
}
else if (MVDiff >= 16) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (posBalance >= 59) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 33) {
if (mcTcor < 97) {
if (TSC < 11) {
if (MVDiff < 8) {
if (ssDCval < 53) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (ssDCval >= 53) {
if (Rs < 975) {
if (Rs < 533) {
return 1;
}
else if (Rs >= 533) {
return 8;
}
}
else if (Rs >= 975) {
if (negBalance < 41) {
return 4;
}
else if (negBalance >= 41) {
if (mu_mv_mag_sq < 47) {
if (RsDiff < 6) {
if (SC < 2558) {
return 1;
}
else if (SC >= 2558) {
return 4;
}
}
else if (RsDiff >= 6) {
if (Rs < 1093) {
return 4;
}
else if (Rs >= 1093) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (SC < 2629) {
if (Cs < 1373) {
return 8;
}
else if (Cs >= 1373) {
return 4;
}
}
else if (SC >= 2629) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 8) {
if (TSC < 8) {
if (diffRsCsdiff < -21) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
return 8;
}
}
else if (diffRsCsdiff >= -21) {
if (refDCval < 75) {
if (ssDCval < 60) {
if (AFD < 11) {
if (Rs < 1074) {
return 4;
}
else if (Rs >= 1074) {
return 8;
}
}
else if (AFD >= 11) {
if (ssDCval < 58) {
return 2;
}
else if (ssDCval >= 58) {
return 4;
}
}
}
else if (ssDCval >= 60) {
if (mcTcor < 96) {
if (Rs < 629) {
return 2;
}
else if (Rs >= 629) {
return 8;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
}
else if (refDCval >= 75) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Rs < 985) {
return 4;
}
else if (Rs >= 985) {
return 8;
}
}
}
}
}
else if (TSC >= 8) {
if (diffMVdiffVal < 2) {
if (Rs < 1316) {
if (TSC < 10) {
if (Cs < 1385) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 2;
}
}
}
else if (Cs >= 1385) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (TSC >= 10) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
if (Rs < 1211) {
return 2;
}
else if (Rs >= 1211) {
return 8;
}
}
}
}
else if (Rs >= 1316) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
}
}
else if (TSC >= 11) {
if (CsDiff < 7) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (CsDiff >= 7) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
if (tcor < 61) {
return 4;
}
else if (tcor >= 61) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 18) {
if (diffAFD < 1) {
if (ssDCval < 62) {
if (Rs < 1057) {
return 1;
}
else if (Rs >= 1057) {
return 4;
}
}
else if (ssDCval >= 62) {
return 2;
}
}
else if (diffAFD >= 1) {
if (Rs < 821) {
return 1;
}
else if (Rs >= 821) {
return 2;
}
}
}
else if (MVDiff >= 18) {
if (Rs < 514) {
return 1;
}
else if (Rs >= 514) {
return 8;
}
}
}
}
}
else if (Rs >= 1454) {
if (RsDiff < 8) {
if (refDCval < 73) {
if (Cs < 2619) {
if (Rs < 3280) {
if (Cs < 1372) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (RsDiff < 5) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 1;
}
}
else if (RsDiff >= 5) {
if (RsCsDiff < 70) {
return 2;
}
else if (RsCsDiff >= 70) {
return 8;
}
}
}
}
else if (Cs >= 1372) {
if (ssDCval < 52) {
if (CsDiff < 7) {
if (SC < 3066) {
if (Rs < 1511) {
return 4;
}
else if (Rs >= 1511) {
if (Rs < 1571) {
return 8;
}
else if (Rs >= 1571) {
return 1;
}
}
}
else if (SC >= 3066) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (Rs < 1974) {
return 8;
}
else if (Rs >= 1974) {
return 4;
}
}
}
}
else if (CsDiff >= 7) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 1776) {
return 4;
}
else if (Rs >= 1776) {
if (Cs < 1782) {
return 8;
}
else if (Cs >= 1782) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 52) {
if (posBalance < 70) {
if (mcTcor < 92) {
if (refDCval < 57) {
if (Rs < 1713) {
return 4;
}
else if (Rs >= 1713) {
return 8;
}
}
else if (refDCval >= 57) {
return 8;
}
}
else if (mcTcor >= 92) {
if (MVDiff < 5) {
if (RsCsDiff < 11) {
return 1;
}
else if (RsCsDiff >= 11) {
return 8;
}
}
else if (MVDiff >= 5) {
if (refDCval < 65) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (refDCval >= 65) {
if (diffAFD < -1) {
if (Cs < 1461) {
if (Rs < 1660) {
return 4;
}
else if (Rs >= 1660) {
return 1;
}
}
else if (Cs >= 1461) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
else if (posBalance >= 70) {
return 4;
}
}
}
}
else if (Rs >= 3280) {
if (Rs < 3497) {
if (Rs < 3288) {
return 4;
}
else if (Rs >= 3288) {
return 8;
}
}
else if (Rs >= 3497) {
return 4;
}
}
}
else if (Cs >= 2619) {
return 8;
}
}
else if (refDCval >= 73) {
if (SC < 5086) {
if (mcTcor < 94) {
if (AFD < 16) {
if (Cs < 2401) {
if (diffAFD < -3) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
if (ssDCval < 80) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 2;
}
}
else if (ssDCval >= 80) {
if (Cs < 1454) {
return 2;
}
else if (Cs >= 1454) {
return 8;
}
}
}
}
else if (diffAFD >= -3) {
if (Cs < 1702) {
if (diffMVdiffVal < 12) {
if (diffRsCsdiff < -17) {
if (posBalance < 56) {
if (TSC < 8) {
if (diffRsCsdiff < -19) {
return 8;
}
else if (diffRsCsdiff >= -19) {
return 1;
}
}
else if (TSC >= 8) {
return 1;
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (diffRsCsdiff >= -17) {
return 8;
}
}
else if (diffMVdiffVal >= 12) {
return 2;
}
}
else if (Cs >= 1702) {
if (posBalance < 68) {
if (TSC < 9) {
if (MVDiff < 8) {
if (RsCsDiff < 73) {
return 8;
}
else if (RsCsDiff >= 73) {
if (Cs < 1875) {
return 4;
}
else if (Cs >= 1875) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 4;
}
}
}
}
else if (MVDiff >= 8) {
if (diffRsCsdiff < -10) {
return 4;
}
else if (diffRsCsdiff >= -10) {
return 8;
}
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (posBalance >= 68) {
return 4;
}
}
}
}
else if (Cs >= 2401) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 2178) {
return 1;
}
else if (Rs >= 2178) {
return 4;
}
}
}
}
else if (AFD >= 16) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 2016) {
return 8;
}
else if (Rs >= 2016) {
return 4;
}
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 2937) {
if (Cs < 1536) {
if (MVDiff < 9) {
if (ssDCval < 82) {
if (Cs < 1352) {
if (diffRsCsdiff < -7) {
return 1;
}
else if (diffRsCsdiff >= -7) {
if (diffRsCsdiff < 4) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
return 1;
}
}
else if (diffRsCsdiff >= 4) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
}
}
else if (Cs >= 1352) {
if (MVDiff < 3) {
if (gchDC < 1) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
if (Cs < 1369) {
return 8;
}
else if (Cs >= 1369) {
return 1;
}
}
}
else if (gchDC >= 1) {
if (Rs < 2535) {
return 1;
}
else if (Rs >= 2535) {
return 2;
}
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (ssDCval >= 82) {
if (mu_mv_mag_sq < 90) {
if (mu_mv_mag_sq < 52) {
if (Rs < 2519) {
return 8;
}
else if (Rs >= 2519) {
return 1;
}
}
else if (mu_mv_mag_sq >= 52) {
if (posBalance < 53) {
if (Rs < 2230) {
return 2;
}
else if (Rs >= 2230) {
return 4;
}
}
else if (posBalance >= 53) {
if (refDCval < 84) {
return 1;
}
else if (refDCval >= 84) {
if (Rs < 2169) {
return 2;
}
else if (Rs >= 2169) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (Cs >= 1536) {
if (AFD < 9) {
if (Cs < 1645) {
if (RsCsDiff < 51) {
return 8;
}
else if (RsCsDiff >= 51) {
if (Rs < 2011) {
return 2;
}
else if (Rs >= 2011) {
return 1;
}
}
}
else if (Cs >= 1645) {
if (Cs < 1741) {
if (Cs < 1740) {
return 8;
}
else if (Cs >= 1740) {
if (Rs < 2528) {
return 1;
}
else if (Rs >= 2528) {
return 8;
}
}
}
else if (Cs >= 1741) {
return 8;
}
}
}
else if (AFD >= 9) {
if (Cs < 2246) {
if (refDCval < 81) {
if (negBalance < 49) {
if (Cs < 1843) {
return 4;
}
else if (Cs >= 1843) {
if (RsDiff < 6) {
if (Rs < 1968) {
return 1;
}
else if (Rs >= 1968) {
return 4;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (Rs < 1675) {
return 2;
}
else if (Rs >= 1675) {
return 8;
}
}
}
else if (refDCval >= 81) {
if (posBalance < 59) {
if (Rs < 2831) {
if (MVDiff < 3) {
if (Rs < 2600) {
return 1;
}
else if (Rs >= 2600) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (Rs >= 2831) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 1;
}
}
}
else if (posBalance >= 59) {
if (ssDCval < 88) {
if (Cs < 1669) {
return 1;
}
else if (Cs >= 1669) {
return 4;
}
}
else if (ssDCval >= 88) {
if (Rs < 2191) {
return 2;
}
else if (Rs >= 2191) {
return 8;
}
}
}
}
}
else if (Cs >= 2246) {
if (Rs < 2336) {
return 1;
}
else if (Rs >= 2336) {
return 8;
}
}
}
}
}
else if (Rs >= 2937) {
if (Cs < 1406) {
if (RsCsDiff < 51) {
if (Rs < 3514) {
return 4;
}
else if (Rs >= 3514) {
return 1;
}
}
else if (RsCsDiff >= 51) {
return 2;
}
}
else if (Cs >= 1406) {
return 4;
}
}
}
}
else if (SC >= 5086) {
if (Rs < 2689) {
return 4;
}
else if (Rs >= 2689) {
if (Cs < 2046) {
if (posBalance < 45) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 4;
}
}
else if (posBalance >= 45) {
if (Rs < 3655) {
return 8;
}
else if (Rs >= 3655) {
if (Rs < 3732) {
return 2;
}
else if (Rs >= 3732) {
return 1;
}
}
}
}
else if (Cs >= 2046) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (negBalance < 50) {
if (mu_mv_mag_sq < 85) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (Rs < 3205) {
return 4;
}
else if (Rs >= 3205) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 4;
}
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (tcor < 92) {
if (TSC < 8) {
if (Cs < 1538) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
}
else if (Cs >= 1538) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
return 4;
}
}
}
else if (TSC >= 8) {
if (posBalance < 53) {
if (Cs < 1518) {
if (Rs < 2427) {
if (Rs < 2029) {
return 1;
}
else if (Rs >= 2029) {
return 2;
}
}
else if (Rs >= 2427) {
return 4;
}
}
else if (Cs >= 1518) {
return 8;
}
}
else if (posBalance >= 53) {
if (AFD < 19) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
if (TSC < 10) {
if (diffRsCsdiff < -9) {
if (Rs < 2675) {
return 4;
}
else if (Rs >= 2675) {
return 8;
}
}
else if (diffRsCsdiff >= -9) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (Cs < 1866) {
return 4;
}
else if (Cs >= 1866) {
return 8;
}
}
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (AFD >= 19) {
return 4;
}
}
}
}
else if (tcor >= 92) {
if (Cs < 1759) {
if (Rs < 2749) {
return 8;
}
else if (Rs >= 2749) {
return 2;
}
}
else if (Cs >= 1759) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 90) {
if (Rs < 4406) {
if (refDCval < 162) {
if (ssDCval < 109) {
if (mcTcor < 96) {
if (Rs < 1443) {
if (RsCsDiff < 69) {
if (mu_mv_mag_sq < 52) {
if (refDCval < 94) {
if (Cs < 1829) {
if (MVDiff < 11) {
if (Cs < 1762) {
return 4;
}
else if (Cs >= 1762) {
return 8;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (Cs >= 1829) {
return 4;
}
}
else if (refDCval >= 94) {
return 4;
}
}
else if (mu_mv_mag_sq >= 52) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Rs < 1325) {
return 4;
}
else if (Rs >= 1325) {
return 8;
}
}
}
}
else if (RsCsDiff >= 69) {
if (ssDCval < 100) {
if (Rs < 825) {
return 4;
}
else if (Rs >= 825) {
if (Cs < 1958) {
if (RsCsDiff < 94) {
return 8;
}
else if (RsCsDiff >= 94) {
if (Cs < 1519) {
return 2;
}
else if (Cs >= 1519) {
return 8;
}
}
}
else if (Cs >= 1958) {
return 2;
}
}
}
else if (ssDCval >= 100) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
}
}
else if (Rs >= 1443) {
if (Cs < 2136) {
if (Rs < 2225) {
if (TSC < 9) {
if (diffMVdiffVal < 3) {
if (mu_mv_mag_sq < 16) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
if (tcor < 86) {
return 2;
}
else if (tcor >= 86) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (ssDCval < 92) {
if (posBalance < 60) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
return 8;
}
}
else if (posBalance >= 60) {
return 2;
}
}
else if (ssDCval >= 92) {
return 8;
}
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1814) {
return 8;
}
else if (Rs >= 1814) {
if (Cs < 1809) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
return 2;
}
}
else if (Cs >= 1809) {
return 8;
}
}
}
}
else if (TSC >= 9) {
if (TSC < 13) {
if (refDCval < 97) {
return 8;
}
else if (refDCval >= 97) {
if (Rs < 1697) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (Rs >= 1697) {
if (ssDCval < 98) {
return 1;
}
else if (ssDCval >= 98) {
if (Cs < 1346) {
return 1;
}
else if (Cs >= 1346) {
return 8;
}
}
}
}
}
else if (TSC >= 13) {
if (MVDiff < 28) {
return 8;
}
else if (MVDiff >= 28) {
return 1;
}
}
}
}
else if (Rs >= 2225) {
if (Rs < 3909) {
if (refDCval < 103) {
if (negBalance < 34) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (negBalance >= 34) {
if (diffRsCsdiff < -53) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
else if (diffRsCsdiff >= -53) {
return 8;
}
}
}
else if (refDCval >= 103) {
if (Rs < 3100) {
if (MVDiff < 9) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
else if (MVDiff >= 9) {
if (mu_mv_mag_sq < 71) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (Cs < 1472) {
return 4;
}
else if (Cs >= 1472) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 71) {
return 8;
}
}
}
else if (Rs >= 3100) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
}
}
else if (Rs >= 3909) {
if (mu_mv_mag_sq < 85) {
if (Cs < 1466) {
if (Rs < 4222) {
return 4;
}
else if (Rs >= 4222) {
if (Rs < 4314) {
return 8;
}
else if (Rs >= 4314) {
return 8;
}
}
}
else if (Cs >= 1466) {
return 8;
}
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 4227) {
return 2;
}
else if (Rs >= 4227) {
return 1;
}
}
}
}
}
else if (Cs >= 2136) {
if (MVDiff < 6) {
if (Cs < 2428) {
if (diffMVdiffVal < 2) {
if (SC < 6436) {
if (negBalance < 56) {
if (TSC < 11) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 3179) {
return 8;
}
else if (Rs >= 3179) {
return 1;
}
}
}
else if (TSC >= 11) {
if (Cs < 2254) {
return 4;
}
else if (Cs >= 2254) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (SC >= 6436) {
return 1;
}
}
else if (diffMVdiffVal >= 2) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 8;
}
}
}
else if (Cs >= 2428) {
if (refDCval < 98) {
return 4;
}
else if (refDCval >= 98) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (Cs < 2616) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
else if (Cs >= 2616) {
return 8;
}
}
}
}
}
else if (MVDiff >= 6) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (SC < 4089) {
if (negBalance < 53) {
if (Rs < 1732) {
return 4;
}
else if (Rs >= 1732) {
return 1;
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (SC >= 4089) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (RsCsDiff < 55) {
if (MVDiff < 4) {
if (posBalance < 41) {
if (tcor < 97) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
if (Rs < 3084) {
return 1;
}
else if (Rs >= 3084) {
return 4;
}
}
}
else if (tcor >= 97) {
if (RsDiff < 5) {
if (posBalance < 10) {
return 1;
}
else if (posBalance >= 10) {
if (mu_mv_mag_sq < 106) {
if (posBalance < 37) {
return 4;
}
else if (posBalance >= 37) {
if (Cs < 1477) {
return 4;
}
else if (Cs >= 1477) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 106) {
return 8;
}
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (posBalance >= 41) {
if (SC < 2286) {
return 4;
}
else if (SC >= 2286) {
if (negBalance < 50) {
if (Cs < 2819) {
if (posBalance < 44) {
if (Cs < 1520) {
if (Rs < 3532) {
return 4;
}
else if (Rs >= 3532) {
return 2;
}
}
else if (Cs >= 1520) {
if (Rs < 2756) {
return 4;
}
else if (Rs >= 2756) {
return 8;
}
}
}
else if (posBalance >= 44) {
if (Rs < 1984) {
if (Rs < 1310) {
return 8;
}
else if (Rs >= 1310) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 2;
}
}
}
else if (Rs >= 1984) {
if (Rs < 3553) {
return 8;
}
else if (Rs >= 3553) {
if (Rs < 4113) {
return 2;
}
else if (Rs >= 4113) {
return 8;
}
}
}
}
}
else if (Cs >= 2819) {
if (Rs < 3825) {
return 4;
}
else if (Rs >= 3825) {
return 8;
}
}
}
else if (negBalance >= 50) {
return 4;
}
}
}
}
else if (MVDiff >= 4) {
if (CsDiff < 1) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 2298) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (Rs >= 2298) {
if (refDCval < 92) {
return 4;
}
else if (refDCval >= 92) {
if (Rs < 2319) {
return 4;
}
else if (Rs >= 2319) {
return 8;
}
}
}
}
}
else if (CsDiff >= 1) {
if (ssDCval < 107) {
if (tcor < 92) {
if (ssDCval < 103) {
if (Cs < 1392) {
return 4;
}
else if (Cs >= 1392) {
return 8;
}
}
else if (ssDCval >= 103) {
if (Rs < 3026) {
return 4;
}
else if (Rs >= 3026) {
return 8;
}
}
}
else if (tcor >= 92) {
if (Rs < 2127) {
if (SC < 3926) {
if (AFD < 11) {
if (Cs < 1375) {
if (Rs < 1410) {
return 4;
}
else if (Rs >= 1410) {
return 8;
}
}
else if (Cs >= 1375) {
return 8;
}
}
else if (AFD >= 11) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 2;
}
}
}
else if (SC >= 3926) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 1;
}
}
}
else if (Rs >= 2127) {
if (MVDiff < 5) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 2586) {
return 8;
}
else if (Rs >= 2586) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (Rs < 4146) {
if (tcor < 94) {
if (Cs < 1980) {
return 8;
}
else if (Cs >= 1980) {
return 1;
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (Rs >= 4146) {
if (Rs < 4188) {
return 4;
}
else if (Rs >= 4188) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 107) {
if (Cs < 2066) {
if (Cs < 1383) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 2716) {
return 8;
}
else if (Rs >= 2716) {
return 4;
}
}
}
else if (Cs >= 1383) {
return 4;
}
}
else if (Cs >= 2066) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 55) {
if (RsDiff < 5) {
if (Rs < 1303) {
if (Cs < 1768) {
return 8;
}
else if (Cs >= 1768) {
return 4;
}
}
else if (Rs >= 1303) {
if (Rs < 1943) {
return 2;
}
else if (Rs >= 1943) {
return 1;
}
}
}
else if (RsDiff >= 5) {
if (AFD < 13) {
if (Rs < 3466) {
if (SC < 3888) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Cs < 1373) {
return 8;
}
else if (Cs >= 1373) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
}
else if (SC >= 3888) {
if (SC < 4649) {
return 8;
}
else if (SC >= 4649) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (SC < 4704) {
return 4;
}
else if (SC >= 4704) {
return 8;
}
}
}
}
}
else if (Rs >= 3466) {
if (SC < 6112) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (Rs < 3962) {
return 8;
}
else if (Rs >= 3962) {
return 2;
}
}
}
}
else if (SC >= 6112) {
return 8;
}
}
}
else if (AFD >= 13) {
if (refDCval < 98) {
if (SC < 3603) {
if (Cs < 1626) {
return 1;
}
else if (Cs >= 1626) {
return 8;
}
}
else if (SC >= 3603) {
return 4;
}
}
else if (refDCval >= 98) {
if (Rs < 1390) {
return 4;
}
else if (Rs >= 1390) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 1529) {
if (mu_mv_mag_sq < 46) {
if (refDCval < 151) {
if (AFD < 19) {
if (SC < 2924) {
if (refDCval < 125) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (Rs < 1178) {
return 8;
}
else if (Rs >= 1178) {
return 2;
}
}
}
else if (refDCval >= 125) {
if (AFD < 15) {
if (mcTcor < 90) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (mcTcor >= 90) {
if (MVDiff < 5) {
if (Rs < 1318) {
return 8;
}
else if (Rs >= 1318) {
return 1;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (SC >= 2924) {
if (negBalance < 27) {
return 2;
}
else if (negBalance >= 27) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
if (TSC < 9) {
if (Cs < 1577) {
return 8;
}
else if (Cs >= 1577) {
return 4;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
else if (AFD >= 19) {
if (Rs < 899) {
return 2;
}
else if (Rs >= 899) {
if (Cs < 1442) {
return 1;
}
else if (Cs >= 1442) {
return 4;
}
}
}
}
else if (refDCval >= 151) {
if (mu_mv_mag_sq < 38) {
return 2;
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 1364) {
return 4;
}
else if (Rs >= 1364) {
if (Rs < 1509) {
return 8;
}
else if (Rs >= 1509) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (Rs < 693) {
return 1;
}
else if (Rs >= 693) {
if (Rs < 1284) {
if (TSC < 7) {
if (mcTcor < 94) {
if (RsCsDiff < 46) {
if (Cs < 1379) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 1379) {
if (Rs < 1191) {
return 4;
}
else if (Rs >= 1191) {
return 8;
}
}
}
else if (RsCsDiff >= 46) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 1030) {
return 8;
}
else if (Rs >= 1030) {
if (RsDiff < 3) {
if (MVDiff < 7) {
if (Rs < 1076) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
else if (Rs >= 1076) {
return 4;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (RsDiff >= 3) {
return 8;
}
}
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (Rs >= 1284) {
if (SC < 3154) {
return 8;
}
else if (SC >= 3154) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (Rs < 1488) {
return 4;
}
else if (Rs >= 1488) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 1529) {
if (ssDCval < 140) {
if (ssDCval < 126) {
if (mu_mv_mag_sq < 37) {
if (mcTcor < 98) {
if (negBalance < 45) {
if (diffRsCsdiff < -6) {
if (SC < 4174) {
return 4;
}
else if (SC >= 4174) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 8;
}
}
}
else if (negBalance >= 45) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (mcTcor < 96) {
if (MVDiff < 13) {
return 1;
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Rs < 1802) {
return 1;
}
else if (Rs >= 1802) {
if (Rs < 1870) {
return 4;
}
else if (Rs >= 1870) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (Cs < 2164) {
return 2;
}
else if (Cs >= 2164) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (TSC < 11) {
if (ssDCval < 121) {
if (Rs < 4225) {
if (refDCval < 109) {
if (Cs < 1591) {
if (Rs < 2650) {
return 8;
}
else if (Rs >= 2650) {
return 4;
}
}
else if (Cs >= 1591) {
return 8;
}
}
else if (refDCval >= 109) {
return 8;
}
}
else if (Rs >= 4225) {
if (Cs < 2480) {
return 1;
}
else if (Cs >= 2480) {
return 8;
}
}
}
else if (ssDCval >= 121) {
if (Cs < 1360) {
if (Rs < 2375) {
return 8;
}
else if (Rs >= 2375) {
return 2;
}
}
else if (Cs >= 1360) {
if (diffRsCsdiff < -41) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
return 1;
}
}
}
else if (diffRsCsdiff >= -41) {
if (RsDiff < 1) {
if (Rs < 2752) {
return 8;
}
else if (Rs >= 2752) {
return 2;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
}
}
}
else if (TSC >= 11) {
if (ssDCval < 112) {
if (Rs < 2837) {
return 1;
}
else if (Rs >= 2837) {
return 8;
}
}
else if (ssDCval >= 112) {
return 8;
}
}
}
}
else if (ssDCval >= 126) {
if (mcTcor < 92) {
if (AFD < 12) {
if (Rs < 3579) {
if (CsDiff < 5) {
if (TSC < 8) {
if (Rs < 2474) {
return 1;
}
else if (Rs >= 2474) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (CsDiff >= 5) {
if (CsDiff < 6) {
if (mcTcor < 89) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (mcTcor >= 89) {
return 8;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (Rs >= 3579) {
return 1;
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (mcTcor >= 92) {
if (ssDCval < 127) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 3562) {
return 1;
}
else if (Rs >= 3562) {
return 8;
}
}
}
else if (ssDCval >= 127) {
if (negBalance < 37) {
if (SC < 4354) {
if (ssDCval < 136) {
if (SC < 3791) {
if (Cs < 1330) {
return 1;
}
else if (Cs >= 1330) {
return 8;
}
}
else if (SC >= 3791) {
return 1;
}
}
else if (ssDCval >= 136) {
if (Rs < 2683) {
return 8;
}
else if (Rs >= 2683) {
return 4;
}
}
}
else if (SC >= 4354) {
if (Cs < 1348) {
return 2;
}
else if (Cs >= 1348) {
return 8;
}
}
}
else if (negBalance >= 37) {
return 8;
}
}
}
}
}
else if (ssDCval >= 140) {
if (posBalance < 63) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1764) {
return 2;
}
else if (Rs >= 1764) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (posBalance >= 63) {
if (Rs < 1925) {
if (Rs < 1788) {
return 2;
}
else if (Rs >= 1788) {
return 4;
}
}
else if (Rs >= 1925) {
if (Cs < 1321) {
return 2;
}
else if (Cs >= 1321) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 162) {
if (Rs < 2266) {
if (Rs < 1098) {
return 1;
}
else if (Rs >= 1098) {
return 2;
}
}
else if (Rs >= 2266) {
return 8;
}
}
}
else if (Rs >= 4406) {
if (Cs < 1525) {
return 2;
}
else if (Cs >= 1525) {
if (tcor < 94) {
if (mu_mv_mag_sq < 25) {
return 8;
}
else if (mu_mv_mag_sq >= 25) {
if (posBalance < 56) {
if (Cs < 1888) {
if (Rs < 4704) {
return 1;
}
else if (Rs >= 4704) {
return 4;
}
}
else if (Cs >= 1888) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 1;
}
}
}
else if (posBalance >= 56) {
return 1;
}
}
}
else if (tcor >= 94) {
if (Cs < 1854) {
if (tcor < 97) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
else if (tcor >= 97) {
if (refDCval < 92) {
if (Rs < 4757) {
return 2;
}
else if (Rs >= 4757) {
return 8;
}
}
else if (refDCval >= 92) {
return 2;
}
}
}
else if (Cs >= 1854) {
if (Rs < 6251) {
return 8;
}
else if (Rs >= 6251) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1926) {
if (Rs < 3338) {
if (tcor < 78) {
if (mcTcor < 87) {
if (Rs < 1620) {
if (ssDCval < 87) {
if (refDCval < 46) {
return 8;
}
else if (refDCval >= 46) {
if (ssDCval < 51) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
else if (ssDCval >= 51) {
if (mu_mv_mag_sq < 9) {
if (SC < 2973) {
if (diffMVdiffVal < 2) {
if (TSC < 12) {
if (Rs < 790) {
return 8;
}
else if (Rs >= 790) {
return 2;
}
}
else if (TSC >= 12) {
if (Rs < 1065) {
return 8;
}
else if (Rs >= 1065) {
return 4;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1292) {
return 4;
}
else if (Rs >= 1292) {
return 1;
}
}
}
else if (SC >= 2973) {
return 4;
}
}
else if (mu_mv_mag_sq >= 9) {
if (RsCsDiff < 173) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
if (Rs < 1448) {
return 4;
}
else if (Rs >= 1448) {
return 8;
}
}
}
else if (RsCsDiff >= 173) {
if (Cs < 1513) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 4;
}
}
else if (Cs >= 1513) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (diffMVdiffVal < 17) {
if (MVDiff < 36) {
if (Cs < 1564) {
return 8;
}
else if (Cs >= 1564) {
return 2;
}
}
else if (MVDiff >= 36) {
if (posBalance < 74) {
return 1;
}
else if (posBalance >= 74) {
if (Rs < 1508) {
return 4;
}
else if (Rs >= 1508) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 17) {
return 8;
}
}
}
else if (Rs >= 1620) {
if (AFD < 19) {
if (Cs < 1749) {
if (negBalance < 50) {
if (mu_mv_mag_sq < 29) {
if (mcTcor < 77) {
if (Cs < 1593) {
return 8;
}
else if (Cs >= 1593) {
return 4;
}
}
else if (mcTcor >= 77) {
if (ssDCval < 85) {
return 1;
}
else if (ssDCval >= 85) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (mu_mv_mag_sq < 61) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
return 8;
}
}
else if (mu_mv_mag_sq >= 61) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
if (Rs < 2236) {
return 4;
}
else if (Rs >= 2236) {
return 8;
}
}
}
}
}
else if (negBalance >= 50) {
if (MVDiff < 31) {
if (diffRsCsdiff < 71) {
if (MVDiff < 6) {
if (Rs < 2195) {
return 8;
}
else if (Rs >= 2195) {
return 4;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (diffRsCsdiff >= 71) {
return 4;
}
}
else if (MVDiff >= 31) {
if (Cs < 1676) {
if (MVDiff < 33) {
return 1;
}
else if (MVDiff >= 33) {
return 8;
}
}
else if (Cs >= 1676) {
return 1;
}
}
}
}
else if (Cs >= 1749) {
if (ssDCval < 57) {
if (Rs < 1996) {
return 4;
}
else if (Rs >= 1996) {
if (Cs < 1846) {
return 8;
}
else if (Cs >= 1846) {
return 4;
}
}
}
else if (ssDCval >= 57) {
if (TSC < 14) {
if (diffRsCsdiff < -146) {
return 4;
}
else if (diffRsCsdiff >= -146) {
return 8;
}
}
else if (TSC >= 14) {
if (Rs < 1900) {
return 4;
}
else if (Rs >= 1900) {
return 1;
}
}
}
}
}
else if (AFD >= 19) {
if (ssDCval < 121) {
if (refDCval < 104) {
if (diffMVdiffVal < -4) {
if (tcor < 58) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 1;
}
}
else if (tcor >= 58) {
if (diffRsCsdiff < 22) {
return 4;
}
else if (diffRsCsdiff >= 22) {
return 2;
}
}
}
else if (diffMVdiffVal >= -4) {
if (SC < 3770) {
if (diffMVdiffVal < 9) {
if (Cs < 1659) {
return 8;
}
else if (Cs >= 1659) {
return 4;
}
}
else if (diffMVdiffVal >= 9) {
if (Cs < 1401) {
return 8;
}
else if (Cs >= 1401) {
return 1;
}
}
}
else if (SC >= 3770) {
if (ssDCval < 99) {
if (TSC < 14) {
if (ssDCval < 88) {
if (RsCsDiff < 323) {
if (Rs < 2514) {
if (diffTSC < 1) {
if (Cs < 1860) {
return 1;
}
else if (Cs >= 1860) {
return 8;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (Rs >= 2514) {
return 4;
}
}
else if (RsCsDiff >= 323) {
return 2;
}
}
else if (ssDCval >= 88) {
return 8;
}
}
else if (TSC >= 14) {
if (Rs < 3204) {
return 8;
}
else if (Rs >= 3204) {
return 4;
}
}
}
else if (ssDCval >= 99) {
if (MVDiff < 107) {
if (Rs < 2595) {
return 4;
}
else if (Rs >= 2595) {
return 2;
}
}
else if (MVDiff >= 107) {
return 4;
}
}
}
}
}
else if (refDCval >= 104) {
if (Cs < 1571) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (Rs < 1970) {
return 2;
}
else if (Rs >= 1970) {
return 4;
}
}
}
}
else if (Cs >= 1571) {
if (mu_mv_mag_sq < 7) {
if (Rs < 2607) {
if (Rs < 2439) {
return 4;
}
else if (Rs >= 2439) {
return 1;
}
}
else if (Rs >= 2607) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
}
}
else if (ssDCval >= 121) {
if (refDCval < 141) {
if (posBalance < 56) {
if (Cs < 1454) {
return 4;
}
else if (Cs >= 1454) {
return 8;
}
}
else if (posBalance >= 56) {
if (posBalance < 121) {
return 1;
}
else if (posBalance >= 121) {
return 8;
}
}
}
else if (refDCval >= 141) {
if (TSC < 12) {
return 1;
}
else if (TSC >= 12) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 87) {
if (posBalance < 50) {
if (ssDCval < 48) {
if (Rs < 2221) {
return 2;
}
else if (Rs >= 2221) {
return 1;
}
}
else if (ssDCval >= 48) {
if (negBalance < 52) {
if (RsDiff < 9) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
else if (RsDiff >= 9) {
if (Rs < 3216) {
return 2;
}
else if (Rs >= 3216) {
return 4;
}
}
}
else if (negBalance >= 52) {
if (Cs < 1748) {
return 4;
}
else if (Cs >= 1748) {
return 8;
}
}
}
}
else if (posBalance >= 50) {
if (RsCsDiff < 290) {
if (AFD < 18) {
if (refDCval < 86) {
if (MVDiff < 11) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
return 4;
}
}
else if (MVDiff >= 11) {
if (posBalance < 58) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (posBalance >= 58) {
if (TSC < 11) {
if (Cs < 1330) {
return 1;
}
else if (Cs >= 1330) {
return 8;
}
}
else if (TSC >= 11) {
return 1;
}
}
}
}
else if (refDCval >= 86) {
if (ssDCval < 101) {
if (Cs < 1618) {
return 2;
}
else if (Cs >= 1618) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
return 8;
}
}
}
else if (ssDCval >= 101) {
if (Rs < 3138) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 4;
}
}
else if (Rs >= 3138) {
if (Rs < 3265) {
return 4;
}
else if (Rs >= 3265) {
return 1;
}
}
}
}
}
else if (AFD >= 18) {
if (SC < 3333) {
if (diffAFD < 3) {
if (negBalance < 52) {
if (Rs < 1739) {
if (CsDiff < 12) {
return 4;
}
else if (CsDiff >= 12) {
return 2;
}
}
else if (Rs >= 1739) {
return 8;
}
}
else if (negBalance >= 52) {
if (Rs < 1644) {
if (negBalance < 57) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
if (RsCsDiff < 133) {
if (Rs < 1401) {
return 4;
}
else if (Rs >= 1401) {
return 8;
}
}
else if (RsCsDiff >= 133) {
return 8;
}
}
}
else if (negBalance >= 57) {
if (SC < 3039) {
if (RsCsDiff < 181) {
if (ssDCval < 70) {
return 4;
}
else if (ssDCval >= 70) {
return 2;
}
}
else if (RsCsDiff >= 181) {
if (Cs < 1622) {
return 8;
}
else if (Cs >= 1622) {
return 4;
}
}
}
else if (SC >= 3039) {
if (Rs < 1584) {
return 8;
}
else if (Rs >= 1584) {
return 1;
}
}
}
}
else if (Rs >= 1644) {
if (gchDC < 2) {
if (Rs < 1738) {
return 4;
}
else if (Rs >= 1738) {
return 8;
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (diffAFD >= 3) {
if (tcor < 65) {
if (Rs < 696) {
return 2;
}
else if (Rs >= 696) {
if (refDCval < 64) {
return 4;
}
else if (refDCval >= 64) {
return 8;
}
}
}
else if (tcor >= 65) {
if (Cs < 1656) {
return 2;
}
else if (Cs >= 1656) {
return 1;
}
}
}
}
else if (SC >= 3333) {
if (ssDCval < 89) {
if (ssDCval < 77) {
if (mu_mv_mag_sq < 37) {
if (RsDiff < 12) {
if (diffMVdiffVal < 4) {
if (Rs < 2582) {
return 8;
}
else if (Rs >= 2582) {
return 2;
}
}
else if (diffMVdiffVal >= 4) {
if (Cs < 1797) {
return 4;
}
else if (Cs >= 1797) {
return 8;
}
}
}
else if (RsDiff >= 12) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 37) {
return 1;
}
}
else if (ssDCval >= 77) {
if (Cs < 1639) {
if (diffRsCsdiff < 42) {
if (mu_mv_mag_sq < 3) {
return 8;
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
else if (diffRsCsdiff >= 42) {
return 8;
}
}
else if (Cs >= 1639) {
if (diffAFD < -9) {
return 2;
}
else if (diffAFD >= -9) {
return 4;
}
}
}
}
else if (ssDCval >= 89) {
if (tcor < 60) {
if (mu_mv_mag_sq < 15) {
return 4;
}
else if (mu_mv_mag_sq >= 15) {
if (Rs < 1994) {
return 8;
}
else if (Rs >= 1994) {
return 2;
}
}
}
else if (tcor >= 60) {
if (ssDCval < 130) {
if (refDCval < 89) {
return 1;
}
else if (refDCval >= 89) {
if (Cs < 1355) {
return 1;
}
else if (Cs >= 1355) {
return 8;
}
}
}
else if (ssDCval >= 130) {
if (Cs < 1689) {
return 2;
}
else if (Cs >= 1689) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 290) {
if (tcor < 64) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
if (Cs < 1880) {
return 2;
}
else if (Cs >= 1880) {
return 8;
}
}
}
else if (tcor >= 64) {
if (mcTcor < 95) {
if (posBalance < 59) {
if (gchDC < 1) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
else if (gchDC >= 1) {
if (CsDiff < 14) {
return 8;
}
else if (CsDiff >= 14) {
return 4;
}
}
}
else if (posBalance >= 59) {
if (RsDiff < 15) {
if (Rs < 1996) {
return 2;
}
else if (Rs >= 1996) {
return 8;
}
}
else if (RsDiff >= 15) {
return 2;
}
}
}
else if (mcTcor >= 95) {
if (Rs < 2431) {
return 2;
}
else if (Rs >= 2431) {
if (Rs < 3170) {
return 1;
}
else if (Rs >= 3170) {
return 2;
}
}
}
}
}
}
}
}
else if (tcor >= 78) {
if (Rs < 2935) {
if (ssDCval < 83) {
if (posBalance < 50) {
if (negBalance < 53) {
if (AFD < 16) {
if (mu_mv_mag_sq < 82) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (SC < 3571) {
return 1;
}
else if (SC >= 3571) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 82) {
return 8;
}
}
else if (AFD >= 16) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
return 2;
}
}
}
else if (negBalance >= 53) {
if (mu_mv_mag_sq < 30) {
if (Rs < 1671) {
if (Rs < 985) {
return 2;
}
else if (Rs >= 985) {
return 1;
}
}
else if (Rs >= 1671) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (Rs < 1439) {
return 4;
}
else if (Rs >= 1439) {
return 8;
}
}
}
}
}
else if (posBalance >= 50) {
if (Cs < 1511) {
if (Rs < 1899) {
if (mu_mv_mag_sq < 12) {
if (Cs < 1332) {
if (Rs < 1183) {
if (Rs < 1035) {
return 8;
}
else if (Rs >= 1035) {
return 2;
}
}
else if (Rs >= 1183) {
return 1;
}
}
else if (Cs >= 1332) {
return 2;
}
}
else if (mu_mv_mag_sq >= 12) {
if (Cs < 1466) {
return 8;
}
else if (Cs >= 1466) {
return 1;
}
}
}
else if (Rs >= 1899) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Rs < 2592) {
return 4;
}
else if (Rs >= 2592) {
return 2;
}
}
}
}
else if (Cs >= 1511) {
if (ssDCval < 74) {
if (negBalance < 34) {
if (Rs < 1107) {
return 1;
}
else if (Rs >= 1107) {
if (Rs < 2159) {
return 4;
}
else if (Rs >= 2159) {
return 2;
}
}
}
else if (negBalance >= 34) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (diffMVdiffVal < 6) {
if (RsDiff < 13) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 2;
}
}
else if (RsDiff >= 13) {
return 4;
}
}
else if (diffMVdiffVal >= 6) {
if (Rs < 1755) {
return 1;
}
else if (Rs >= 1755) {
return 4;
}
}
}
}
}
else if (ssDCval >= 74) {
if (diffRsCsdiff < 21) {
if (TSC < 11) {
if (posBalance < 53) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
else if (posBalance >= 53) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
return 1;
}
}
}
else if (TSC >= 11) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
return 8;
}
}
}
else if (diffRsCsdiff >= 21) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 83) {
if (TSC < 10) {
if (AFD < 19) {
if (Rs < 1426) {
if (posBalance < 63) {
if (refDCval < 89) {
return 1;
}
else if (refDCval >= 89) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
}
else if (posBalance >= 63) {
return 4;
}
}
else if (Rs >= 1426) {
if (refDCval < 98) {
if (CsDiff < 8) {
if (TSC < 9) {
if (Rs < 2596) {
if (Cs < 1599) {
return 2;
}
else if (Cs >= 1599) {
return 8;
}
}
else if (Rs >= 2596) {
if (Rs < 2801) {
return 4;
}
else if (Rs >= 2801) {
return 2;
}
}
}
else if (TSC >= 9) {
if (Cs < 1853) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 1;
}
}
else if (Cs >= 1853) {
return 2;
}
}
}
else if (CsDiff >= 8) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (MVDiff < 11) {
if (Rs < 2176) {
return 2;
}
else if (Rs >= 2176) {
return 4;
}
}
else if (MVDiff >= 11) {
if (Rs < 2178) {
return 8;
}
else if (Rs >= 2178) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 98) {
if (ssDCval < 161) {
if (tcor < 81) {
if (Cs < 1736) {
return 8;
}
else if (Cs >= 1736) {
if (Rs < 2032) {
return 8;
}
else if (Rs >= 2032) {
return 2;
}
}
}
else if (tcor >= 81) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (RsDiff < 13) {
if (Cs < 1323) {
if (Rs < 1920) {
return 8;
}
else if (Rs >= 1920) {
return 4;
}
}
else if (Cs >= 1323) {
if (posBalance < 84) {
if (diffRsCsdiff < -56) {
if (Rs < 2000) {
return 4;
}
else if (Rs >= 2000) {
return 8;
}
}
else if (diffRsCsdiff >= -56) {
return 8;
}
}
else if (posBalance >= 84) {
return 4;
}
}
}
else if (RsDiff >= 13) {
return 4;
}
}
}
}
else if (ssDCval >= 161) {
return 2;
}
}
}
}
else if (AFD >= 19) {
if (refDCval < 96) {
if (Rs < 1499) {
return 8;
}
else if (Rs >= 1499) {
return 2;
}
}
else if (refDCval >= 96) {
if (refDCval < 108) {
return 4;
}
else if (refDCval >= 108) {
return 1;
}
}
}
}
else if (TSC >= 10) {
if (MVDiff < 24) {
if (RsCsDiff < 263) {
if (Rs < 1636) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
if (Rs < 1416) {
return 2;
}
else if (Rs >= 1416) {
return 8;
}
}
}
else if (Rs >= 1636) {
if (posBalance < 43) {
if (Rs < 2814) {
return 8;
}
else if (Rs >= 2814) {
return 2;
}
}
else if (posBalance >= 43) {
if (diffAFD < -4) {
if (Rs < 1967) {
return 1;
}
else if (Rs >= 1967) {
return 8;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
}
}
else if (RsCsDiff >= 263) {
if (Rs < 2222) {
return 8;
}
else if (Rs >= 2222) {
return 2;
}
}
}
else if (MVDiff >= 24) {
if (mu_mv_mag_sq < 18) {
if (TSC < 12) {
if (Rs < 1961) {
if (Cs < 1650) {
return 2;
}
else if (Cs >= 1650) {
return 1;
}
}
else if (Rs >= 1961) {
return 8;
}
}
else if (TSC >= 12) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
if (Cs < 1355) {
return 1;
}
else if (Cs >= 1355) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (Rs < 1473) {
if (Rs < 1236) {
return 4;
}
else if (Rs >= 1236) {
return 2;
}
}
else if (Rs >= 1473) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2935) {
if (refDCval < 97) {
if (negBalance < 47) {
if (MVDiff < 11) {
if (Cs < 1649) {
return 2;
}
else if (Cs >= 1649) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (Rs < 3004) {
return 1;
}
else if (Rs >= 3004) {
return 2;
}
}
}
}
else if (MVDiff >= 11) {
if (diffAFD < 1) {
if (Cs < 1413) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
else if (Cs >= 1413) {
return 2;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (AFD < 15) {
if (RsDiff < 13) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (Rs < 3274) {
return 2;
}
else if (Rs >= 3274) {
return 4;
}
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (AFD >= 15) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
return 8;
}
}
}
}
else if (refDCval >= 97) {
if (refDCval < 108) {
if (MVDiff < 15) {
if (TSC < 8) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (TSC >= 8) {
if (Rs < 3238) {
return 4;
}
else if (Rs >= 3238) {
return 8;
}
}
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (refDCval >= 108) {
if (refDCval < 158) {
if (Cs < 1335) {
if (Cs < 1313) {
return 8;
}
else if (Cs >= 1313) {
if (Rs < 3119) {
return 4;
}
else if (Rs >= 3119) {
return 1;
}
}
}
else if (Cs >= 1335) {
return 8;
}
}
else if (refDCval >= 158) {
if (Rs < 2990) {
return 8;
}
else if (Rs >= 2990) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 3338) {
if (refDCval < 100) {
if (mcTcor < 92) {
if (RsDiff < 13) {
if (negBalance < 42) {
if (MVDiff < 27) {
if (refDCval < 64) {
if (Cs < 1671) {
return 2;
}
else if (Cs >= 1671) {
return 8;
}
}
else if (refDCval >= 64) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (MVDiff >= 27) {
return 8;
}
}
else if (negBalance >= 42) {
return 8;
}
}
else if (RsDiff >= 13) {
if (AFD < 10) {
if (Rs < 4586) {
return 1;
}
else if (Rs >= 4586) {
return 4;
}
}
else if (AFD >= 10) {
if (Cs < 1658) {
if (MVDiff < 37) {
return 2;
}
else if (MVDiff >= 37) {
if (Cs < 1408) {
return 2;
}
else if (Cs >= 1408) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 4;
}
}
}
}
else if (Cs >= 1658) {
if (mu_mv_mag_sq < 10) {
return 2;
}
else if (mu_mv_mag_sq >= 10) {
if (gchDC < 2) {
if (MVDiff < 21) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 4;
}
}
else if (MVDiff >= 21) {
return 4;
}
}
else if (gchDC >= 2) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (posBalance < 47) {
if (negBalance < 36) {
if (gchDC < 1) {
if (Rs < 4950) {
return 2;
}
else if (Rs >= 4950) {
return 1;
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (negBalance >= 36) {
if (tcor < 89) {
if (Rs < 3649) {
return 4;
}
else if (Rs >= 3649) {
return 2;
}
}
else if (tcor >= 89) {
if (MVDiff < 2) {
if (Rs < 4378) {
return 2;
}
else if (Rs >= 4378) {
return 8;
}
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
else if (posBalance >= 47) {
if (diffMVdiffVal < 5) {
if (TSC < 6) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
return 1;
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (diffMVdiffVal >= 5) {
if (AFD < 20) {
if (diffRsCsdiff < 82) {
return 2;
}
else if (diffRsCsdiff >= 82) {
if (Rs < 4703) {
return 2;
}
else if (Rs >= 4703) {
return 4;
}
}
}
else if (AFD >= 20) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 100) {
if (RsDiff < 13) {
if (mu_mv_mag_sq < 84) {
if (tcor < 93) {
if (ssDCval < 137) {
if (tcor < 67) {
return 1;
}
else if (tcor >= 67) {
if (Cs < 1323) {
return 4;
}
else if (Cs >= 1323) {
if (Rs < 4412) {
if (diffMVdiffVal < -6) {
if (Cs < 1798) {
return 8;
}
else if (Cs >= 1798) {
return 2;
}
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
else if (Rs >= 4412) {
if (SC < 6233) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 8;
}
}
else if (SC >= 6233) {
if (SC < 6800) {
return 8;
}
else if (SC >= 6800) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 137) {
if (Rs < 4312) {
return 4;
}
else if (Rs >= 4312) {
return 8;
}
}
}
else if (tcor >= 93) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (Rs < 3666) {
return 8;
}
else if (Rs >= 3666) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (CsDiff < 6) {
if (Rs < 3707) {
return 4;
}
else if (Rs >= 3707) {
return 2;
}
}
else if (CsDiff >= 6) {
return 1;
}
}
}
else if (RsDiff >= 13) {
if (MVDiff < 17) {
if (tcor < 84) {
if (RsDiff < 16) {
if (Rs < 4263) {
return 8;
}
else if (Rs >= 4263) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (Cs < 1676) {
return 4;
}
else if (Cs >= 1676) {
return 8;
}
}
}
}
else if (RsDiff >= 16) {
if (Cs < 1784) {
if (negBalance < 55) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
if (Rs < 4427) {
return 8;
}
else if (Rs >= 4427) {
return 4;
}
}
}
else if (negBalance >= 55) {
return 2;
}
}
else if (Cs >= 1784) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 2;
}
}
}
}
else if (tcor >= 84) {
if (negBalance < 64) {
if (Rs < 4019) {
if (gchDC < 1) {
if (Rs < 3906) {
return 8;
}
else if (Rs >= 3906) {
return 4;
}
}
else if (gchDC >= 1) {
return 2;
}
}
else if (Rs >= 4019) {
return 2;
}
}
else if (negBalance >= 64) {
if (RsDiff < 17) {
return 1;
}
else if (RsDiff >= 17) {
return 2;
}
}
}
}
else if (MVDiff >= 17) {
if (TSC < 14) {
if (mu_mv_mag_sq < 18) {
if (posBalance < 50) {
if (ssDCval < 109) {
return 2;
}
else if (ssDCval >= 109) {
if (Rs < 6128) {
return 1;
}
else if (Rs >= 6128) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (Rs < 5116) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (diffMVdiffVal < -1) {
if (RsDiff < 21) {
return 8;
}
else if (RsDiff >= 21) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 5086) {
return 2;
}
else if (Rs >= 5086) {
return 1;
}
}
}
}
else if (Rs >= 5116) {
if (posBalance < 54) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (Cs < 1421) {
return 8;
}
else if (Cs >= 1421) {
return 2;
}
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (AFD < 22) {
if (Cs < 1656) {
if (Rs < 4281) {
return 1;
}
else if (Rs >= 4281) {
return 4;
}
}
else if (Cs >= 1656) {
return 8;
}
}
else if (AFD >= 22) {
if (AFD < 27) {
if (Rs < 5132) {
return 2;
}
else if (Rs >= 5132) {
return 1;
}
}
else if (AFD >= 27) {
if (negBalance < 57) {
if (Cs < 1454) {
if (Rs < 3928) {
return 8;
}
else if (Rs >= 3928) {
return 1;
}
}
else if (Cs >= 1454) {
if (RsDiff < 14) {
return 1;
}
else if (RsDiff >= 14) {
return 4;
}
}
}
else if (negBalance >= 57) {
if (Rs < 3951) {
return 8;
}
else if (Rs >= 3951) {
return 2;
}
}
}
}
}
}
else if (TSC >= 14) {
if (tcor < 73) {
if (CsDiff < 10) {
if (MVDiff < 61) {
if (negBalance < 55) {
if (Rs < 3738) {
return 2;
}
else if (Rs >= 3738) {
return 4;
}
}
else if (negBalance >= 55) {
return 2;
}
}
else if (MVDiff >= 61) {
return 4;
}
}
else if (CsDiff >= 10) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
if (MVDiff < 208) {
return 1;
}
else if (MVDiff >= 208) {
return 2;
}
}
}
}
else if (tcor >= 73) {
if (AFD < 26) {
if (TSC < 15) {
if (Rs < 5117) {
return 8;
}
else if (Rs >= 5117) {
return 2;
}
}
else if (TSC >= 15) {
return 1;
}
}
else if (AFD >= 26) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 1926) {
if (AFD < 12) {
if (ssDCval < 100) {
if (Rs < 4004) {
if (ssDCval < 74) {
if (mcTcor < 93) {
if (Cs < 2753) {
if (TSC < 9) {
if (MVDiff < 5) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 2256) {
return 4;
}
else if (Rs >= 2256) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (diffAFD < -2) {
if (Cs < 2214) {
return 1;
}
else if (Cs >= 2214) {
return 4;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
}
else if (TSC >= 9) {
if (Rs < 2478) {
return 4;
}
else if (Rs >= 2478) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 1;
}
}
}
}
else if (Cs >= 2753) {
if (Rs < 1468) {
return 8;
}
else if (Rs >= 1468) {
return 4;
}
}
}
else if (mcTcor >= 93) {
if (Cs < 2292) {
if (diffRsCsdiff < -69) {
return 8;
}
else if (diffRsCsdiff >= -69) {
return 4;
}
}
else if (Cs >= 2292) {
return 1;
}
}
}
else if (ssDCval >= 74) {
if (Cs < 2639) {
if (ssDCval < 89) {
if (mcTcor < 92) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
else if (mcTcor >= 92) {
if (TSC < 10) {
if (Rs < 2420) {
if (Rs < 2252) {
return 8;
}
else if (Rs >= 2252) {
return 1;
}
}
else if (Rs >= 2420) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (ssDCval >= 89) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
}
}
else if (Cs >= 2639) {
if (refDCval < 81) {
if (mu_mv_mag_sq < 78) {
return 4;
}
else if (mu_mv_mag_sq >= 78) {
return 8;
}
}
else if (refDCval >= 81) {
if (Rs < 3284) {
if (posBalance < 58) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Cs < 3332) {
return 4;
}
else if (Cs >= 3332) {
return 2;
}
}
}
else if (posBalance >= 58) {
return 4;
}
}
else if (Rs >= 3284) {
return 4;
}
}
}
}
}
else if (Rs >= 4004) {
if (tcor < 91) {
if (Rs < 4948) {
if (MVDiff < 7) {
if (mu_mv_mag_sq < 88) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
else if (mu_mv_mag_sq >= 88) {
if (ssDCval < 88) {
return 8;
}
else if (ssDCval >= 88) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 68) {
if (Cs < 3136) {
return 4;
}
else if (Cs >= 3136) {
return 8;
}
}
else if (ssDCval >= 68) {
return 8;
}
}
}
else if (Rs >= 4948) {
if (ssDCval < 74) {
if (SC < 8895) {
return 4;
}
else if (SC >= 8895) {
if (Rs < 7226) {
return 8;
}
else if (Rs >= 7226) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
}
else if (tcor >= 91) {
if (RsDiff < 9) {
if (posBalance < 44) {
if (Rs < 4799) {
return 8;
}
else if (Rs >= 4799) {
return 1;
}
}
else if (posBalance >= 44) {
if (Rs < 5235) {
return 4;
}
else if (Rs >= 5235) {
return 2;
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 72) {
if (Cs < 3422) {
return 4;
}
else if (Cs >= 3422) {
return 8;
}
}
else if (ssDCval >= 72) {
if (Rs < 4256) {
if (SC < 6339) {
return 2;
}
else if (SC >= 6339) {
if (Rs < 4190) {
return 4;
}
else if (Rs >= 4190) {
return 8;
}
}
}
else if (Rs >= 4256) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 100) {
if (negBalance < 46) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 93) {
if (Rs < 5239) {
if (ssDCval < 115) {
if (tcor < 82) {
return 8;
}
else if (tcor >= 82) {
return 2;
}
}
else if (ssDCval >= 115) {
if (mcTcor < 76) {
return 2;
}
else if (mcTcor >= 76) {
if (refDCval < 130) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
return 8;
}
}
else if (refDCval >= 130) {
if (diffMVdiffVal < -4) {
if (tcor < 82) {
if (RsCsDiff < 115) {
return 8;
}
else if (RsCsDiff >= 115) {
return 1;
}
}
else if (tcor >= 82) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
}
}
else if (Rs >= 5239) {
if (CsDiff < 9) {
return 2;
}
else if (CsDiff >= 9) {
if (ssDCval < 131) {
return 1;
}
else if (ssDCval >= 131) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (posBalance < 68) {
return 8;
}
else if (posBalance >= 68) {
return 2;
}
}
}
else if (negBalance >= 37) {
if (Rs < 6363) {
if (refDCval < 102) {
if (Rs < 4526) {
if (Cs < 2530) {
return 8;
}
else if (Cs >= 2530) {
if (Rs < 3339) {
return 4;
}
else if (Rs >= 3339) {
return 2;
}
}
}
else if (Rs >= 4526) {
return 1;
}
}
else if (refDCval >= 102) {
if (Rs < 5897) {
if (mu_mv_mag_sq < 100) {
if (diffRsCsdiff < -179) {
return 2;
}
else if (diffRsCsdiff >= -179) {
return 8;
}
}
else if (mu_mv_mag_sq >= 100) {
if (mcTcor < 93) {
if (Rs < 4385) {
return 8;
}
else if (Rs >= 4385) {
return 1;
}
}
else if (mcTcor >= 93) {
return 8;
}
}
}
else if (Rs >= 5897) {
if (Cs < 4773) {
return 4;
}
else if (Cs >= 4773) {
return 8;
}
}
}
}
else if (Rs >= 6363) {
if (Cs < 2943) {
return 2;
}
else if (Cs >= 2943) {
if (Rs < 8019) {
return 4;
}
else if (Rs >= 8019) {
return 1;
}
}
}
}
}
else if (negBalance >= 46) {
if (mcTcor < 93) {
if (mu_mv_mag_sq < 83) {
if (refDCval < 102) {
if (Rs < 4007) {
return 2;
}
else if (Rs >= 4007) {
if (Rs < 4495) {
return 8;
}
else if (Rs >= 4495) {
return 1;
}
}
}
else if (refDCval >= 102) {
return 8;
}
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 3740) {
if (Cs < 2062) {
return 1;
}
else if (Cs >= 2062) {
return 8;
}
}
else if (Rs >= 3740) {
if (Rs < 5654) {
if (mu_mv_mag_sq < 92) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
else if (Rs >= 5654) {
return 8;
}
}
}
}
else if (mcTcor >= 93) {
if (diffAFD < -2) {
if (posBalance < 53) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
else if (posBalance >= 53) {
if (Rs < 3179) {
return 1;
}
else if (Rs >= 3179) {
return 4;
}
}
}
else if (diffAFD >= -2) {
if (Rs < 7746) {
return 8;
}
else if (Rs >= 7746) {
return 1;
}
}
}
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 76) {
if (RsDiff < 17) {
if (refDCval < 101) {
if (tcor < 64) {
if (ssDCval < 73) {
if (refDCval < 63) {
if (MVDiff < 19) {
if (mu_mv_mag_sq < 4) {
return 4;
}
else if (mu_mv_mag_sq >= 4) {
if (Cs < 2223) {
return 4;
}
else if (Cs >= 2223) {
return 2;
}
}
}
else if (MVDiff >= 19) {
if (RsCsDiff < 294) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
else if (RsCsDiff >= 294) {
if (diffAFD < 4) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
return 4;
}
}
else if (diffAFD >= 4) {
return 2;
}
}
}
}
else if (refDCval >= 63) {
if (Rs < 1759) {
return 2;
}
else if (Rs >= 1759) {
if (mu_mv_mag_sq < 9) {
return 1;
}
else if (mu_mv_mag_sq >= 9) {
if (Cs < 2170) {
if (Rs < 2257) {
return 1;
}
else if (Rs >= 2257) {
return 2;
}
}
else if (Cs >= 2170) {
return 4;
}
}
}
}
}
else if (ssDCval >= 73) {
if (gchDC < 3) {
if (RsDiff < 11) {
if (ssDCval < 95) {
return 4;
}
else if (ssDCval >= 95) {
if (Cs < 2929) {
if (Rs < 1509) {
return 4;
}
else if (Rs >= 1509) {
return 8;
}
}
else if (Cs >= 2929) {
return 4;
}
}
}
else if (RsDiff >= 11) {
if (mcTcor < 74) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
else if (mcTcor >= 74) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
if (CsDiff < 12) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 1;
}
}
else if (CsDiff >= 12) {
if (diffRsCsdiff < -13) {
if (Rs < 2501) {
return 8;
}
else if (Rs >= 2501) {
if (Rs < 2677) {
return 1;
}
else if (Rs >= 2677) {
return 8;
}
}
}
else if (diffRsCsdiff >= -13) {
return 8;
}
}
}
}
}
}
else if (gchDC >= 3) {
if (SC < 4937) {
if (Cs < 2782) {
if (Rs < 1666) {
return 8;
}
else if (Rs >= 1666) {
if (Cs < 2005) {
return 4;
}
else if (Cs >= 2005) {
return 1;
}
}
}
else if (Cs >= 2782) {
if (CsDiff < 15) {
return 1;
}
else if (CsDiff >= 15) {
return 2;
}
}
}
else if (SC >= 4937) {
if (negBalance < 64) {
return 8;
}
else if (negBalance >= 64) {
return 2;
}
}
}
}
}
else if (tcor >= 64) {
if (negBalance < 47) {
if (posBalance < 49) {
if (ssDCval < 73) {
if (SC < 6253) {
if (CsDiff < 11) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
else if (CsDiff >= 11) {
if (Rs < 2229) {
return 8;
}
else if (Rs >= 2229) {
return 2;
}
}
}
else if (SC >= 6253) {
return 4;
}
}
else if (ssDCval >= 73) {
if (ssDCval < 88) {
if (Rs < 2481) {
return 8;
}
else if (Rs >= 2481) {
return 2;
}
}
else if (ssDCval >= 88) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
}
}
else if (posBalance >= 49) {
if (Rs < 3884) {
if (Cs < 2397) {
if (TSC < 8) {
if (RsCsDiff < 190) {
return 8;
}
else if (RsCsDiff >= 190) {
return 2;
}
}
else if (TSC >= 8) {
if (MVDiff < 30) {
if (negBalance < 44) {
if (Rs < 3688) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 8;
}
}
else if (Rs >= 3688) {
if (Rs < 3730) {
return 1;
}
else if (Rs >= 3730) {
if (Rs < 3778) {
return 2;
}
else if (Rs >= 3778) {
return 8;
}
}
}
}
else if (negBalance >= 44) {
if (posBalance < 59) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (Cs < 2208) {
return 1;
}
else if (Cs >= 2208) {
return 8;
}
}
}
else if (posBalance >= 59) {
if (Cs < 1954) {
return 1;
}
else if (Cs >= 1954) {
return 4;
}
}
}
}
else if (MVDiff >= 30) {
if (refDCval < 89) {
return 4;
}
else if (refDCval >= 89) {
return 8;
}
}
}
}
else if (Cs >= 2397) {
if (mu_mv_mag_sq < 51) {
if (ssDCval < 74) {
if (Cs < 4135) {
return 2;
}
else if (Cs >= 4135) {
return 1;
}
}
else if (ssDCval >= 74) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 51) {
if (RsCsDiff < 167) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (RsCsDiff >= 167) {
if (mcTcor < 94) {
if (tcor < 75) {
if (Rs < 3603) {
if (Rs < 2225) {
return 1;
}
else if (Rs >= 2225) {
return 8;
}
}
else if (Rs >= 3603) {
return 2;
}
}
else if (tcor >= 75) {
if (Cs < 2450) {
return 2;
}
else if (Cs >= 2450) {
if (CsDiff < 13) {
return 4;
}
else if (CsDiff >= 13) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
return 8;
}
}
}
}
}
else if (Rs >= 3884) {
if (MVDiff < 7) {
if (Rs < 5871) {
return 2;
}
else if (Rs >= 5871) {
return 8;
}
}
else if (MVDiff >= 7) {
if (posBalance < 53) {
if (refDCval < 94) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
else if (refDCval >= 94) {
return 1;
}
}
else if (posBalance >= 53) {
if (tcor < 88) {
if (MVDiff < 13) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (tcor >= 88) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 47) {
if (SC < 8280) {
if (SC < 5083) {
if (AFD < 17) {
if (Rs < 1575) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 4;
}
}
else if (Rs >= 1575) {
if (diffMVdiffVal < -8) {
if (CsDiff < 11) {
if (Rs < 2155) {
return 2;
}
else if (Rs >= 2155) {
return 8;
}
}
else if (CsDiff >= 11) {
return 1;
}
}
else if (diffMVdiffVal >= -8) {
if (tcor < 80) {
if (mcTcor < 72) {
return 2;
}
else if (mcTcor >= 72) {
if (MVDiff < 3) {
if (Cs < 2740) {
return 8;
}
else if (Cs >= 2740) {
return 1;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (tcor >= 80) {
if (negBalance < 53) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (Rs < 2507) {
return 4;
}
else if (Rs >= 2507) {
if (Rs < 2550) {
return 8;
}
else if (Rs >= 2550) {
return 2;
}
}
}
}
}
else if (negBalance >= 53) {
if (RsCsDiff < 138) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (RsCsDiff >= 138) {
if (RsDiff < 12) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
else if (RsDiff >= 12) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 17) {
if (ssDCval < 67) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (diffRsCsdiff < 15) {
if (diffMVdiffVal < 6) {
return 2;
}
else if (diffMVdiffVal >= 6) {
if (Rs < 1120) {
return 2;
}
else if (Rs >= 1120) {
return 1;
}
}
}
else if (diffRsCsdiff >= 15) {
if (SC < 4336) {
return 4;
}
else if (SC >= 4336) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (diffRsCsdiff < 62) {
return 2;
}
else if (diffRsCsdiff >= 62) {
if (Rs < 2566) {
return 4;
}
else if (Rs >= 2566) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (mcTcor < 94) {
if (mcTcor < 81) {
if (Cs < 2338) {
if (Rs < 1973) {
return 2;
}
else if (Rs >= 1973) {
return 8;
}
}
else if (Cs >= 2338) {
return 4;
}
}
else if (mcTcor >= 81) {
if (CsDiff < 10) {
if (Cs < 2509) {
if (MVDiff < 60) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
if (SC < 4290) {
return 8;
}
else if (SC >= 4290) {
return 4;
}
}
}
}
else if (MVDiff >= 60) {
return 4;
}
}
else if (Cs >= 2509) {
return 4;
}
}
else if (CsDiff >= 10) {
if (RsCsDiff < 201) {
if (posBalance < 61) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Cs < 2321) {
return 1;
}
else if (Cs >= 2321) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (diffRsCsdiff < 27) {
return 8;
}
else if (diffRsCsdiff >= 27) {
if (Rs < 1362) {
return 2;
}
else if (Rs >= 1362) {
return 1;
}
}
}
}
else if (RsCsDiff >= 201) {
if (gchDC < 2) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
if (posBalance < 66) {
if (mcTcor < 88) {
if (mu_mv_mag_sq < 7) {
if (Rs < 2385) {
return 8;
}
else if (Rs >= 2385) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
else if (mcTcor >= 88) {
return 8;
}
}
else if (posBalance >= 66) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (posBalance < 57) {
if (Rs < 2469) {
return 1;
}
else if (Rs >= 2469) {
return 4;
}
}
else if (posBalance >= 57) {
if (diffRsCsdiff < -4) {
if (MVDiff < 28) {
if (SC < 4843) {
return 2;
}
else if (SC >= 4843) {
return 1;
}
}
else if (MVDiff >= 28) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
if (SC < 4013) {
return 4;
}
else if (SC >= 4013) {
return 8;
}
}
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 1717) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 1;
}
}
else if (Rs >= 1717) {
if (Rs < 2652) {
if (diffRsCsdiff < 65) {
return 2;
}
else if (diffRsCsdiff >= 65) {
return 4;
}
}
else if (Rs >= 2652) {
return 8;
}
}
}
}
}
}
else if (SC >= 5083) {
if (ssDCval < 92) {
if (MVDiff < 5) {
if (ssDCval < 87) {
if (RsDiff < 11) {
if (Cs < 3357) {
if (Cs < 2961) {
if (ssDCval < 83) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 2264) {
return 4;
}
else if (Cs >= 2264) {
return 8;
}
}
}
else if (ssDCval >= 83) {
return 4;
}
}
else if (Cs >= 2961) {
if (Rs < 3278) {
if (mcTcor < 89) {
if (Rs < 2385) {
return 1;
}
else if (Rs >= 2385) {
return 4;
}
}
else if (mcTcor >= 89) {
return 1;
}
}
else if (Rs >= 3278) {
return 2;
}
}
}
else if (Cs >= 3357) {
if (Cs < 3899) {
return 2;
}
else if (Cs >= 3899) {
if (Rs < 3885) {
return 8;
}
else if (Rs >= 3885) {
return 2;
}
}
}
}
else if (RsDiff >= 11) {
if (Cs < 2429) {
return 4;
}
else if (Cs >= 2429) {
if (Rs < 2938) {
return 4;
}
else if (Rs >= 2938) {
return 8;
}
}
}
}
else if (ssDCval >= 87) {
if (negBalance < 54) {
if (Cs < 3811) {
return 4;
}
else if (Cs >= 3811) {
return 8;
}
}
else if (negBalance >= 54) {
if (RsDiff < 12) {
if (Rs < 2373) {
return 4;
}
else if (Rs >= 2373) {
return 8;
}
}
else if (RsDiff >= 12) {
if (Cs < 3787) {
return 4;
}
else if (Cs >= 3787) {
return 8;
}
}
}
}
}
else if (MVDiff >= 5) {
if (Cs < 2453) {
if (refDCval < 88) {
if (RsCsDiff < 157) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (RsCsDiff >= 157) {
if (mcTcor < 91) {
if (mcTcor < 88) {
if (RsCsDiff < 259) {
if (AFD < 17) {
if (Rs < 3008) {
return 8;
}
else if (Rs >= 3008) {
return 4;
}
}
else if (AFD >= 17) {
return 8;
}
}
else if (RsCsDiff >= 259) {
return 4;
}
}
else if (mcTcor >= 88) {
if (RsDiff < 13) {
if (RsCsDiff < 193) {
return 8;
}
else if (RsCsDiff >= 193) {
return 4;
}
}
else if (RsDiff >= 13) {
return 8;
}
}
}
else if (mcTcor >= 91) {
if (negBalance < 55) {
if (Cs < 2332) {
return 4;
}
else if (Cs >= 2332) {
return 2;
}
}
else if (negBalance >= 55) {
if (Rs < 4305) {
return 2;
}
else if (Rs >= 4305) {
return 8;
}
}
}
}
}
else if (refDCval >= 88) {
if (Rs < 2976) {
return 2;
}
else if (Rs >= 2976) {
if (tcor < 77) {
if (Rs < 3848) {
return 4;
}
else if (Rs >= 3848) {
return 8;
}
}
else if (tcor >= 77) {
return 8;
}
}
}
}
else if (Cs >= 2453) {
if (posBalance < 55) {
if (refDCval < 68) {
if (diffAFD < -6) {
if (Rs < 3484) {
return 2;
}
else if (Rs >= 3484) {
return 4;
}
}
else if (diffAFD >= -6) {
return 8;
}
}
else if (refDCval >= 68) {
if (diffRsCsdiff < -130) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
else if (diffRsCsdiff >= -130) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (mcTcor < 90) {
if (refDCval < 77) {
if (TSC < 12) {
if (AFD < 14) {
if (Cs < 2585) {
return 4;
}
else if (Cs >= 2585) {
return 8;
}
}
else if (AFD >= 14) {
return 4;
}
}
else if (TSC >= 12) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (Rs < 3000) {
return 1;
}
else if (Rs >= 3000) {
return 8;
}
}
}
}
else if (refDCval >= 77) {
if (Cs < 3234) {
if (RsCsDiff < 191) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
return 8;
}
}
else if (RsCsDiff >= 191) {
if (AFD < 16) {
if (Rs < 2792) {
return 8;
}
else if (Rs >= 2792) {
return 4;
}
}
else if (AFD >= 16) {
if (AFD < 22) {
if (Rs < 3308) {
return 8;
}
else if (Rs >= 3308) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (AFD >= 22) {
return 1;
}
}
}
}
else if (Cs >= 3234) {
if (tcor < 68) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Cs < 4279) {
return 4;
}
else if (Cs >= 4279) {
return 8;
}
}
}
else if (tcor >= 68) {
return 8;
}
}
}
}
else if (mcTcor >= 90) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 92) {
if (Cs < 2465) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
if (refDCval < 98) {
if (negBalance < 59) {
if (SC < 6084) {
if (mu_mv_mag_sq < 7) {
return 4;
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
else if (SC >= 6084) {
return 8;
}
}
else if (negBalance >= 59) {
return 2;
}
}
else if (refDCval >= 98) {
if (Cs < 2290) {
return 2;
}
else if (Cs >= 2290) {
return 4;
}
}
}
}
else if (Cs >= 2465) {
if (negBalance < 52) {
if (tcor < 83) {
if (diffRsCsdiff < 38) {
if (SC < 6952) {
return 4;
}
else if (SC >= 6952) {
return 8;
}
}
else if (diffRsCsdiff >= 38) {
if (Cs < 2547) {
return 8;
}
else if (Cs >= 2547) {
return 4;
}
}
}
else if (tcor >= 83) {
if (SC < 6230) {
return 2;
}
else if (SC >= 6230) {
return 4;
}
}
}
else if (negBalance >= 52) {
if (Rs < 4760) {
if (diffAFD < -5) {
if (Rs < 2463) {
return 4;
}
else if (Rs >= 2463) {
return 8;
}
}
else if (diffAFD >= -5) {
return 4;
}
}
else if (Rs >= 4760) {
return 8;
}
}
}
}
}
}
else if (SC >= 8280) {
if (diffRsCsdiff < -158) {
return 2;
}
else if (diffRsCsdiff >= -158) {
return 8;
}
}
}
}
}
else if (refDCval >= 101) {
if (AFD < 21) {
if (refDCval < 158) {
if (MVDiff < 9) {
if (Cs < 2801) {
if (mcTcor < 94) {
if (Rs < 3410) {
if (diffRsCsdiff < 136) {
if (tcor < 77) {
if (Rs < 3063) {
return 8;
}
else if (Rs >= 3063) {
return 4;
}
}
else if (tcor >= 77) {
if (AFD < 14) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (AFD >= 14) {
if (AFD < 18) {
if (tcor < 84) {
if (ssDCval < 129) {
return 4;
}
else if (ssDCval >= 129) {
if (Rs < 2814) {
return 4;
}
else if (Rs >= 2814) {
return 8;
}
}
}
else if (tcor >= 84) {
if (diffRsCsdiff < 8) {
if (Rs < 2695) {
if (Rs < 2214) {
return 1;
}
else if (Rs >= 2214) {
return 8;
}
}
else if (Rs >= 2695) {
return 4;
}
}
else if (diffRsCsdiff >= 8) {
return 8;
}
}
}
else if (AFD >= 18) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 136) {
return 1;
}
}
else if (Rs >= 3410) {
if (mu_mv_mag_sq < 74) {
return 8;
}
else if (mu_mv_mag_sq >= 74) {
if (Rs < 4154) {
return 8;
}
else if (Rs >= 4154) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 4804) {
if (negBalance < 48) {
if (Cs < 2254) {
if (Rs < 2044) {
return 1;
}
else if (Rs >= 2044) {
return 4;
}
}
else if (Cs >= 2254) {
return 2;
}
}
else if (negBalance >= 48) {
if (Rs < 2481) {
if (Rs < 2002) {
return 8;
}
else if (Rs >= 2002) {
return 2;
}
}
else if (Rs >= 2481) {
return 8;
}
}
}
else if (Rs >= 4804) {
if (Rs < 6826) {
return 4;
}
else if (Rs >= 6826) {
return 2;
}
}
}
}
else if (Cs >= 2801) {
if (tcor < 76) {
if (ssDCval < 107) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
else if (ssDCval >= 107) {
return 8;
}
}
else if (tcor >= 76) {
if (mcTcor < 89) {
if (SC < 8023) {
if (MVDiff < 5) {
if (Rs < 4257) {
return 8;
}
else if (Rs >= 4257) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (SC >= 8023) {
return 8;
}
}
else if (mcTcor >= 89) {
if (Cs < 2832) {
if (Cs < 2828) {
return 4;
}
else if (Cs >= 2828) {
return 8;
}
}
else if (Cs >= 2832) {
return 4;
}
}
}
}
}
else if (MVDiff >= 9) {
if (ssDCval < 121) {
if (SC < 3973) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
return 8;
}
}
else if (SC >= 3973) {
if (mcTcor < 87) {
if (Cs < 1958) {
return 4;
}
else if (Cs >= 1958) {
return 8;
}
}
else if (mcTcor >= 87) {
if (RsCsDiff < 197) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (RsCsDiff >= 197) {
if (negBalance < 46) {
if (mu_mv_mag_sq < 18) {
return 2;
}
else if (mu_mv_mag_sq >= 18) {
if (Rs < 5663) {
return 8;
}
else if (Rs >= 5663) {
return 4;
}
}
}
else if (negBalance >= 46) {
if (ssDCval < 102) {
return 2;
}
else if (ssDCval >= 102) {
if (negBalance < 57) {
if (Cs < 2023) {
return 2;
}
else if (Cs >= 2023) {
return 8;
}
}
else if (negBalance >= 57) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (RsCsDiff < 201) {
return 4;
}
else if (RsCsDiff >= 201) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 121) {
if (ssDCval < 144) {
if (posBalance < 60) {
if (mu_mv_mag_sq < 45) {
if (negBalance < 50) {
if (MVDiff < 13) {
if (CsDiff < 17) {
return 1;
}
else if (CsDiff >= 17) {
return 2;
}
}
else if (MVDiff >= 13) {
if (mu_mv_mag_sq < 22) {
if (diffMVdiffVal < 6) {
if (Rs < 2068) {
return 1;
}
else if (Rs >= 2068) {
if (Cs < 3640) {
return 4;
}
else if (Cs >= 3640) {
return 1;
}
}
}
else if (diffMVdiffVal >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 22) {
if (SC < 7304) {
if (diffAFD < -1) {
if (Rs < 2503) {
return 8;
}
else if (Rs >= 2503) {
return 2;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (SC >= 7304) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
if (Rs < 3930) {
return 8;
}
else if (Rs >= 3930) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 50) {
if (Cs < 1957) {
return 1;
}
else if (Cs >= 1957) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (Rs < 4163) {
if (posBalance < 55) {
if (Cs < 3835) {
return 4;
}
else if (Cs >= 3835) {
return 2;
}
}
else if (posBalance >= 55) {
if (Cs < 3694) {
if (Rs < 2473) {
return 4;
}
else if (Rs >= 2473) {
return 8;
}
}
else if (Cs >= 3694) {
return 1;
}
}
}
else if (Rs >= 4163) {
if (tcor < 89) {
return 1;
}
else if (tcor >= 89) {
return 2;
}
}
}
}
else if (posBalance >= 60) {
if (TSC < 15) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
if (mcTcor < 72) {
return 2;
}
else if (mcTcor >= 72) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (Rs < 2782) {
return 2;
}
else if (Rs >= 2782) {
return 8;
}
}
}
}
}
else if (TSC >= 15) {
if (Rs < 2521) {
return 8;
}
else if (Rs >= 2521) {
return 4;
}
}
}
}
else if (ssDCval >= 144) {
if (negBalance < 60) {
if (diffAFD < 6) {
return 8;
}
else if (diffAFD >= 6) {
if (Rs < 3122) {
return 8;
}
else if (Rs >= 3122) {
return 1;
}
}
}
else if (negBalance >= 60) {
return 1;
}
}
}
}
}
else if (refDCval >= 158) {
if (posBalance < 52) {
if (Cs < 7979) {
return 1;
}
else if (Cs >= 7979) {
return 8;
}
}
else if (posBalance >= 52) {
return 4;
}
}
}
else if (AFD >= 21) {
if (RsCsDiff < 189) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
if (Rs < 2718) {
return 2;
}
else if (Rs >= 2718) {
return 1;
}
}
}
else if (RsCsDiff >= 189) {
if (negBalance < 57) {
if (ssDCval < 119) {
if (MVDiff < 32) {
if (diffMVdiffVal < -6) {
if (Cs < 2666) {
return 1;
}
else if (Cs >= 2666) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 4462) {
if (Rs < 1967) {
return 4;
}
else if (Rs >= 1967) {
if (negBalance < 54) {
return 2;
}
else if (negBalance >= 54) {
if (Rs < 2880) {
return 2;
}
else if (Rs >= 2880) {
return 4;
}
}
}
}
else if (Rs >= 4462) {
return 1;
}
}
}
}
else if (MVDiff >= 32) {
return 8;
}
}
else if (ssDCval >= 119) {
if (posBalance < 59) {
if (Rs < 2927) {
return 8;
}
else if (Rs >= 2927) {
if (Cs < 3821) {
return 4;
}
else if (Cs >= 3821) {
return 2;
}
}
}
else if (posBalance >= 59) {
if (TSC < 16) {
return 8;
}
else if (TSC >= 16) {
if (Rs < 3352) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 2;
}
}
else if (Rs >= 3352) {
return 8;
}
}
}
}
}
else if (negBalance >= 57) {
if (AFD < 28) {
if (Rs < 2708) {
if (mu_mv_mag_sq < 8) {
if (Rs < 1973) {
if (Cs < 2292) {
return 2;
}
else if (Cs >= 2292) {
return 8;
}
}
else if (Rs >= 1973) {
return 2;
}
}
else if (mu_mv_mag_sq >= 8) {
if (MVDiff < 35) {
return 1;
}
else if (MVDiff >= 35) {
if (Cs < 3445) {
return 8;
}
else if (Cs >= 3445) {
return 2;
}
}
}
}
else if (Rs >= 2708) {
if (SC < 5616) {
if (diffRsCsdiff < -59) {
if (Rs < 3011) {
return 1;
}
else if (Rs >= 3011) {
if (Rs < 3205) {
return 4;
}
else if (Rs >= 3205) {
return 8;
}
}
}
else if (diffRsCsdiff >= -59) {
return 8;
}
}
else if (SC >= 5616) {
if (tcor < 77) {
if (Cs < 3307) {
if (Rs < 3219) {
if (Cs < 3080) {
if (Rs < 3204) {
return 1;
}
else if (Rs >= 3204) {
return 4;
}
}
else if (Cs >= 3080) {
return 8;
}
}
else if (Rs >= 3219) {
return 1;
}
}
else if (Cs >= 3307) {
if (Rs < 4428) {
return 4;
}
else if (Rs >= 4428) {
return 1;
}
}
}
else if (tcor >= 77) {
if (Rs < 3349) {
return 4;
}
else if (Rs >= 3349) {
return 8;
}
}
}
}
}
else if (AFD >= 28) {
if (refDCval < 106) {
if (Rs < 3527) {
return 2;
}
else if (Rs >= 3527) {
return 8;
}
}
else if (refDCval >= 106) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 17) {
if (TSC < 8) {
if (refDCval < 126) {
if (negBalance < 33) {
if (Cs < 2534) {
return 4;
}
else if (Cs >= 2534) {
return 8;
}
}
else if (negBalance >= 33) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 1;
}
}
}
else if (refDCval >= 126) {
if (Cs < 2549) {
return 2;
}
else if (Cs >= 2549) {
return 8;
}
}
}
else if (TSC >= 8) {
if (refDCval < 140) {
if (ssDCval < 111) {
if (Cs < 4698) {
if (TSC < 14) {
if (tcor < 79) {
if (posBalance < 64) {
if (diffAFD < 5) {
if (Cs < 3019) {
if (mcTcor < 83) {
if (Rs < 2881) {
return 4;
}
else if (Rs >= 2881) {
return 8;
}
}
else if (mcTcor >= 83) {
if (MVDiff < 15) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
if (RsCsDiff < 525) {
return 2;
}
else if (RsCsDiff >= 525) {
return 4;
}
}
}
else if (MVDiff >= 15) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 2133) {
if (Cs < 2015) {
if (Rs < 3534) {
return 2;
}
else if (Rs >= 3534) {
return 4;
}
}
else if (Cs >= 2015) {
return 8;
}
}
else if (Cs >= 2133) {
return 4;
}
}
}
}
}
else if (Cs >= 3019) {
if (RsDiff < 26) {
if (negBalance < 46) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
return 8;
}
}
else if (negBalance >= 46) {
return 4;
}
}
else if (RsDiff >= 26) {
return 8;
}
}
}
else if (diffAFD >= 5) {
if (diffMVdiffVal < 5) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
}
else if (posBalance >= 64) {
if (SC < 6134) {
return 8;
}
else if (SC >= 6134) {
return 1;
}
}
}
else if (tcor >= 79) {
if (ssDCval < 71) {
if (Cs < 3255) {
return 4;
}
else if (Cs >= 3255) {
return 2;
}
}
else if (ssDCval >= 71) {
if (CsDiff < 15) {
if (negBalance < 54) {
if (Rs < 3737) {
if (Rs < 3365) {
return 8;
}
else if (Rs >= 3365) {
return 4;
}
}
else if (Rs >= 3737) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
if (Rs < 6073) {
return 8;
}
else if (Rs >= 6073) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
if (ssDCval < 100) {
return 8;
}
else if (ssDCval >= 100) {
if (Rs < 4039) {
return 8;
}
else if (Rs >= 4039) {
return 4;
}
}
}
}
}
else if (CsDiff >= 15) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 2;
}
}
}
}
}
else if (TSC >= 14) {
if (diffAFD < 3) {
if (Rs < 4690) {
if (AFD < 26) {
if (Rs < 2839) {
return 4;
}
else if (Rs >= 2839) {
if (RsDiff < 22) {
return 8;
}
else if (RsDiff >= 22) {
return 2;
}
}
}
else if (AFD >= 26) {
if (posBalance < 59) {
if (MVDiff < 51) {
return 2;
}
else if (MVDiff >= 51) {
if (Cs < 2130) {
return 1;
}
else if (Cs >= 2130) {
return 8;
}
}
}
else if (posBalance >= 59) {
if (gchDC < 5) {
return 1;
}
else if (gchDC >= 5) {
if (Rs < 4119) {
return 8;
}
else if (Rs >= 4119) {
return 1;
}
}
}
}
}
else if (Rs >= 4690) {
if (negBalance < 46) {
if (MVDiff < 24) {
return 1;
}
else if (MVDiff >= 24) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 1;
}
}
}
else if (negBalance >= 46) {
if (MVDiff < 23) {
if (AFD < 23) {
if (negBalance < 57) {
if (Rs < 5757) {
return 8;
}
else if (Rs >= 5757) {
return 2;
}
}
else if (negBalance >= 57) {
return 2;
}
}
else if (AFD >= 23) {
return 4;
}
}
else if (MVDiff >= 23) {
if (diffMVdiffVal < 5) {
if (Cs < 3781) {
return 1;
}
else if (Cs >= 3781) {
return 2;
}
}
else if (diffMVdiffVal >= 5) {
return 2;
}
}
}
}
}
else if (diffAFD >= 3) {
if (mu_mv_mag_sq < 12) {
if (Rs < 4011) {
if (Rs < 3692) {
return 4;
}
else if (Rs >= 3692) {
return 8;
}
}
else if (Rs >= 4011) {
if (RsDiff < 29) {
return 2;
}
else if (RsDiff >= 29) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 12) {
if (gchDC < 7) {
return 8;
}
else if (gchDC >= 7) {
if (RsCsDiff < 517) {
return 8;
}
else if (RsCsDiff >= 517) {
if (AFD < 47) {
return 4;
}
else if (AFD >= 47) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 4698) {
if (CsDiff < 26) {
if (TSC < 21) {
if (Cs < 7580) {
return 4;
}
else if (Cs >= 7580) {
return 8;
}
}
else if (TSC >= 21) {
return 8;
}
}
else if (CsDiff >= 26) {
if (Rs < 3601) {
return 1;
}
else if (Rs >= 3601) {
return 8;
}
}
}
}
else if (ssDCval >= 111) {
if (TSC < 18) {
if (Cs < 3311) {
if (TSC < 16) {
if (tcor < 88) {
if (RsCsDiff < 749) {
if (MVDiff < 56) {
if (diffRsCsdiff < -33) {
if (RsCsDiff < 425) {
return 8;
}
else if (RsCsDiff >= 425) {
if (gchDC < 2) {
if (Cs < 1943) {
return 8;
}
else if (Cs >= 1943) {
return 2;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (diffRsCsdiff >= -33) {
return 8;
}
}
else if (MVDiff >= 56) {
if (Cs < 2515) {
return 8;
}
else if (Cs >= 2515) {
return 2;
}
}
}
else if (RsCsDiff >= 749) {
if (Rs < 5569) {
return 2;
}
else if (Rs >= 5569) {
return 4;
}
}
}
else if (tcor >= 88) {
return 2;
}
}
else if (TSC >= 16) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 8;
}
}
}
else if (Cs >= 3311) {
if (Cs < 3859) {
if (MVDiff < 28) {
return 4;
}
else if (MVDiff >= 28) {
return 2;
}
}
else if (Cs >= 3859) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
return 1;
}
}
}
}
else if (TSC >= 18) {
if (Rs < 4433) {
if (Rs < 3690) {
return 8;
}
else if (Rs >= 3690) {
return 4;
}
}
else if (Rs >= 4433) {
return 8;
}
}
}
}
else if (refDCval >= 140) {
if (refDCval < 166) {
if (gchDC < 3) {
if (ssDCval < 156) {
if (mcTcor < 88) {
if (Rs < 4548) {
return 8;
}
else if (Rs >= 4548) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
if (Rs < 7019) {
return 8;
}
else if (Rs >= 7019) {
return 4;
}
}
}
}
else if (mcTcor >= 88) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
if (negBalance < 53) {
if (Cs < 2031) {
return 4;
}
else if (Cs >= 2031) {
return 2;
}
}
else if (negBalance >= 53) {
if (Rs < 5661) {
return 8;
}
else if (Rs >= 5661) {
if (Rs < 8501) {
return 4;
}
else if (Rs >= 8501) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 156) {
if (Cs < 1993) {
return 2;
}
else if (Cs >= 1993) {
return 4;
}
}
}
else if (gchDC >= 3) {
if (diffAFD < 1) {
if (SC < 6458) {
return 4;
}
else if (SC >= 6458) {
if (diffRsCsdiff < 44) {
if (Rs < 4847) {
return 8;
}
else if (Rs >= 4847) {
return 2;
}
}
else if (diffRsCsdiff >= 44) {
return 2;
}
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
else if (refDCval >= 166) {
if (Cs < 3575) {
return 2;
}
else if (Cs >= 3575) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (SC < 7008) {
if (Rs < 3989) {
if (TSC < 14) {
if (mcTcor < 95) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (refDCval < 81) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (Cs < 3588) {
return 8;
}
else if (Cs >= 3588) {
return 4;
}
}
}
else if (refDCval >= 81) {
if (Cs < 2667) {
if (refDCval < 86) {
if (Rs < 3056) {
if (Cs < 2570) {
return 8;
}
else if (Cs >= 2570) {
return 4;
}
}
else if (Rs >= 3056) {
return 4;
}
}
else if (refDCval >= 86) {
if (diffTSC < -3) {
if (Rs < 3425) {
return 2;
}
else if (Rs >= 3425) {
return 8;
}
}
else if (diffTSC >= -3) {
return 8;
}
}
}
else if (Cs >= 2667) {
if (ssDCval < 108) {
return 4;
}
else if (ssDCval >= 108) {
if (Rs < 3459) {
return 8;
}
else if (Rs >= 3459) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 95) {
return 1;
}
}
else if (TSC >= 14) {
if (Rs < 3333) {
return 4;
}
else if (Rs >= 3333) {
if (Cs < 2745) {
return 1;
}
else if (Cs >= 2745) {
return 8;
}
}
}
}
else if (Rs >= 3989) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 1;
}
}
}
else if (SC >= 7008) {
if (negBalance < 52) {
if (Rs < 5441) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
if (ssDCval < 70) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
return 4;
}
}
else if (ssDCval >= 70) {
if (AFD < 16) {
if (ssDCval < 92) {
if (ssDCval < 91) {
if (Cs < 2843) {
return 4;
}
else if (Cs >= 2843) {
return 8;
}
}
else if (ssDCval >= 91) {
return 4;
}
}
else if (ssDCval >= 92) {
return 8;
}
}
else if (AFD >= 16) {
if (Rs < 4640) {
if (Cs < 3403) {
if (Rs < 4346) {
return 1;
}
else if (Rs >= 4346) {
return 2;
}
}
else if (Cs >= 3403) {
return 8;
}
}
else if (Rs >= 4640) {
return 4;
}
}
}
}
}
else if (Rs >= 5441) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (TSC < 13) {
if (Rs < 6218) {
return 4;
}
else if (Rs >= 6218) {
return 8;
}
}
else if (TSC >= 13) {
if (Rs < 6567) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
if (Cs < 2937) {
return 4;
}
else if (Cs >= 2937) {
return 2;
}
}
}
else if (Rs >= 6567) {
return 4;
}
}
}
}
}
else if (negBalance >= 52) {
if (posBalance < 57) {
if (Cs < 5775) {
if (diffMVdiffVal < -1) {
if (Rs < 4249) {
return 1;
}
else if (Rs >= 4249) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (Cs >= 5775) {
return 4;
}
}
else if (posBalance >= 57) {
if (ssDCval < 125) {
if (Rs < 3188) {
return 4;
}
else if (Rs >= 3188) {
return 8;
}
}
else if (ssDCval >= 125) {
return 4;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (Cs < 1644) {
if (Rs < 3094) {
if (Rs < 1470) {
if (ssDCval < 122) {
if (negBalance < 1) {
return 2;
}
else if (negBalance >= 1) {
if (refDCval < 64) {
if (AFD < 3) {
if (mu_mv_mag_sq < 128) {
if (Rs < 300) {
return 4;
}
else if (Rs >= 300) {
if (TSC < 2) {
if (mu_mv_mag_sq < 127) {
if (diffAFD < -2) {
if (Rs < 992) {
return 8;
}
else if (Rs >= 992) {
return 4;
}
}
else if (diffAFD >= -2) {
if (Cs < 1312) {
if (Rs < 1189) {
return 1;
}
else if (Rs >= 1189) {
return 4;
}
}
else if (Cs >= 1312) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (refDCval < 52) {
return 8;
}
else if (refDCval >= 52) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
if (Rs < 1036) {
return 8;
}
else if (Rs >= 1036) {
return 4;
}
}
}
}
}
else if (TSC >= 2) {
if (Rs < 802) {
return 1;
}
else if (Rs >= 802) {
if (Rs < 1111) {
return 8;
}
else if (Rs >= 1111) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1498) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (Rs < 1162) {
return 8;
}
else if (Rs >= 1162) {
return 4;
}
}
}
else if (Cs >= 1498) {
return 1;
}
}
}
else if (AFD >= 3) {
if (SC < 2584) {
return 8;
}
else if (SC >= 2584) {
if (tcor < 98) {
if (Rs < 1346) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 8;
}
}
else if (Rs >= 1346) {
return 8;
}
}
else if (tcor >= 98) {
return 4;
}
}
}
}
else if (refDCval >= 64) {
if (refDCval < 77) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (CsDiff < 3) {
if (Rs < 1280) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (Rs >= 1280) {
return 1;
}
}
else if (CsDiff >= 3) {
if (Rs < 1418) {
return 4;
}
else if (Rs >= 1418) {
return 8;
}
}
}
}
else if (refDCval >= 77) {
if (Rs < 979) {
if (MVDiff < 2) {
if (Rs < 905) {
return 1;
}
else if (Rs >= 905) {
return 4;
}
}
else if (MVDiff >= 2) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (Rs >= 979) {
if (ssDCval < 79) {
return 4;
}
else if (ssDCval >= 79) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 122) {
if (RsCsDiff < 2) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (Rs < 907) {
return 8;
}
else if (Rs >= 907) {
return 4;
}
}
}
else if (RsCsDiff >= 2) {
if (posBalance < 35) {
if (ssDCval < 131) {
return 8;
}
else if (ssDCval >= 131) {
return 1;
}
}
else if (posBalance >= 35) {
if (Rs < 1274) {
if (Rs < 989) {
if (diffMVdiffVal < -1) {
if (Rs < 933) {
return 8;
}
else if (Rs >= 933) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (Rs >= 989) {
if (refDCval < 131) {
if (Cs < 1462) {
if (Cs < 1367) {
return 4;
}
else if (Cs >= 1367) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (Cs >= 1462) {
return 4;
}
}
else if (refDCval >= 131) {
return 8;
}
}
}
else if (Rs >= 1274) {
return 8;
}
}
}
}
}
else if (Rs >= 1470) {
if (ssDCval < 86) {
if (mcTcor < 98) {
if (posBalance < 45) {
if (posBalance < 20) {
return 8;
}
else if (posBalance >= 20) {
if (Rs < 1703) {
return 8;
}
else if (Rs >= 1703) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (negBalance < 23) {
if (Rs < 2069) {
return 1;
}
else if (Rs >= 2069) {
return 4;
}
}
else if (negBalance >= 23) {
if (posBalance < 43) {
if (gchDC < 2) {
if (Rs < 2568) {
return 8;
}
else if (Rs >= 2568) {
return 4;
}
}
else if (gchDC >= 2) {
return 4;
}
}
else if (posBalance >= 43) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2791) {
if (negBalance < 20) {
if (Rs < 1681) {
return 8;
}
else if (Rs >= 1681) {
return 4;
}
}
else if (negBalance >= 20) {
return 8;
}
}
else if (Rs >= 2791) {
if (Rs < 3071) {
return 2;
}
else if (Rs >= 3071) {
return 8;
}
}
}
}
else if (mcTcor >= 98) {
if (posBalance < 39) {
if (SC < 4475) {
if (negBalance < 20) {
if (Cs < 1351) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (Rs < 2342) {
if (Cs < 1317) {
if (Rs < 1936) {
return 8;
}
else if (Rs >= 1936) {
return 2;
}
}
else if (Cs >= 1317) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 2302) {
if (Rs < 2287) {
return 8;
}
else if (Rs >= 2287) {
if (Rs < 2297) {
return 2;
}
else if (Rs >= 2297) {
if (Rs < 2301) {
return 8;
}
else if (Rs >= 2301) {
return 2;
}
}
}
}
else if (Rs >= 2302) {
return 8;
}
}
}
}
else if (Rs >= 2342) {
return 2;
}
}
}
else if (Cs >= 1351) {
if (mu_mv_mag_sq < 118) {
if (Rs < 1752) {
return 8;
}
else if (Rs >= 1752) {
return 1;
}
}
else if (mu_mv_mag_sq >= 118) {
if (diffRsCsdiff < 2) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 2032) {
return 1;
}
else if (Rs >= 2032) {
return 8;
}
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 1889) {
return 1;
}
else if (Rs >= 1889) {
return 8;
}
}
}
}
}
else if (negBalance >= 20) {
if (Rs < 1519) {
return 1;
}
else if (Rs >= 1519) {
if (Rs < 2278) {
return 4;
}
else if (Rs >= 2278) {
return 2;
}
}
}
}
else if (SC >= 4475) {
return 1;
}
}
else if (posBalance >= 39) {
if (refDCval < 79) {
if (Cs < 1339) {
if (Rs < 1486) {
return 4;
}
else if (Rs >= 1486) {
return 8;
}
}
else if (Cs >= 1339) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
return 8;
}
}
}
else if (refDCval >= 79) {
if (Rs < 2455) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 4;
}
}
else if (Rs >= 2455) {
return 4;
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (diffAFD < -20) {
if (Rs < 2540) {
return 4;
}
else if (Rs >= 2540) {
return 8;
}
}
else if (diffAFD >= -20) {
if (ssDCval < 117) {
if (MVDiff < 2) {
if (AFD < 11) {
if (diffAFD < -2) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 2;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (MVDiff >= 2) {
if (SC < 3528) {
if (Rs < 1914) {
return 8;
}
else if (Rs >= 1914) {
return 2;
}
}
else if (SC >= 3528) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (Rs < 2864) {
if (CsDiff < 2) {
if (Rs < 2829) {
return 8;
}
else if (Rs >= 2829) {
if (Rs < 2844) {
return 1;
}
else if (Rs >= 2844) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 119) {
return 8;
}
else if (mu_mv_mag_sq >= 119) {
if (ssDCval < 128) {
if (mu_mv_mag_sq < 125) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 125) {
return 1;
}
}
else if (ssDCval >= 128) {
return 8;
}
}
}
}
else if (Rs >= 2864) {
if (Rs < 3069) {
return 2;
}
else if (Rs >= 3069) {
return 8;
}
}
}
}
}
else if (ssDCval >= 140) {
if (diffAFD < -1) {
if (Rs < 2739) {
return 8;
}
else if (Rs >= 2739) {
return 2;
}
}
else if (diffAFD >= -1) {
if (Cs < 1568) {
return 2;
}
else if (Cs >= 1568) {
return 1;
}
}
}
}
}
}
else if (Rs >= 3094) {
if (refDCval < 78) {
if (mu_mv_mag_sq < 126) {
if (AFD < 4) {
if (mu_mv_mag_sq < 119) {
if (Rs < 4278) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 1;
}
}
else if (Rs >= 4278) {
return 8;
}
}
else if (mu_mv_mag_sq >= 119) {
if (Cs < 1548) {
if (negBalance < 13) {
if (Cs < 1381) {
return 4;
}
else if (Cs >= 1381) {
if (Rs < 4584) {
return 1;
}
else if (Rs >= 4584) {
return 8;
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 99) {
if (Rs < 4430) {
return 4;
}
else if (Rs >= 4430) {
return 8;
}
}
else if (mcTcor >= 99) {
if (posBalance < 28) {
return 4;
}
else if (posBalance >= 28) {
return 8;
}
}
}
}
else if (Cs >= 1548) {
return 4;
}
}
}
else if (AFD >= 4) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1577) {
if (Rs < 3910) {
if (diffRsCsdiff < -1) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (ssDCval < 58) {
if (Rs < 3570) {
return 8;
}
else if (Rs >= 3570) {
return 4;
}
}
else if (ssDCval >= 58) {
if (Rs < 3808) {
return 8;
}
else if (Rs >= 3808) {
return 4;
}
}
}
}
else if (Rs >= 3910) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (Cs >= 1577) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 4;
}
}
}
}
else if (refDCval >= 78) {
if (ssDCval < 113) {
if (RsDiff < 5) {
if (Rs < 4498) {
if (Cs < 1559) {
return 4;
}
else if (Cs >= 1559) {
if (ssDCval < 96) {
if (Rs < 3613) {
return 4;
}
else if (Rs >= 3613) {
return 8;
}
}
else if (ssDCval >= 96) {
if (Rs < 3268) {
return 8;
}
else if (Rs >= 3268) {
return 4;
}
}
}
}
else if (Rs >= 4498) {
if (Cs < 1571) {
return 2;
}
else if (Cs >= 1571) {
return 8;
}
}
}
else if (RsDiff >= 5) {
if (Rs < 3721) {
return 2;
}
else if (Rs >= 3721) {
if (Rs < 4086) {
if (ssDCval < 84) {
return 2;
}
else if (ssDCval >= 84) {
return 4;
}
}
else if (Rs >= 4086) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (mcTcor < 94) {
if (Rs < 3256) {
return 1;
}
else if (Rs >= 3256) {
return 4;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
}
}
}
else if (Cs >= 1644) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (refDCval < 65) {
if (Rs < 8850) {
return 8;
}
else if (Rs >= 8850) {
return 2;
}
}
else if (refDCval >= 65) {
if (diffTSC < -1) {
if (ssDCval < 72) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
return 4;
}
}
else if (ssDCval >= 72) {
if (Rs < 2268) {
return 4;
}
else if (Rs >= 2268) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (ssDCval < 87) {
if (Rs < 3473) {
return 8;
}
else if (Rs >= 3473) {
return 4;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 3319) {
if (negBalance < 21) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (SC < 5030) {
if (mu_mv_mag_sq < 123) {
if (Cs < 1774) {
if (Rs < 1029) {
return 8;
}
else if (Rs >= 1029) {
if (Rs < 1795) {
return 4;
}
else if (Rs >= 1795) {
return 8;
}
}
}
else if (Cs >= 1774) {
return 8;
}
}
else if (mu_mv_mag_sq >= 123) {
if (SC < 4789) {
if (diffAFD < -1) {
if (ssDCval < 66) {
if (Cs < 1952) {
return 8;
}
else if (Cs >= 1952) {
return 1;
}
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (diffAFD >= -1) {
if (diffRsCsdiff < -2) {
if (Rs < 2421) {
return 8;
}
else if (Rs >= 2421) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
if (Cs < 1727) {
if (Rs < 1425) {
return 1;
}
else if (Rs >= 1425) {
return 8;
}
}
else if (Cs >= 1727) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
return 1;
}
}
}
}
}
}
else if (SC >= 4789) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 4;
}
}
}
}
else if (SC >= 5030) {
if (negBalance < 19) {
return 4;
}
else if (negBalance >= 19) {
return 8;
}
}
}
}
}
else if (negBalance >= 21) {
if (Rs < 1661) {
if (ssDCval < 58) {
if (Rs < 1049) {
return 8;
}
else if (Rs >= 1049) {
return 4;
}
}
else if (ssDCval >= 58) {
if (mcTcor < 98) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
return 1;
}
}
else if (mcTcor >= 98) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (Rs >= 1661) {
if (refDCval < 66) {
if (posBalance < 41) {
if (ssDCval < 65) {
if (Rs < 2668) {
return 8;
}
else if (Rs >= 2668) {
return 4;
}
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (posBalance >= 41) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 2203) {
return 8;
}
else if (Rs >= 2203) {
return 1;
}
}
}
}
else if (refDCval >= 66) {
if (Rs < 3459) {
return 1;
}
else if (Rs >= 3459) {
return 8;
}
}
}
}
}
else if (Cs >= 3319) {
return 8;
}
}
else if (ssDCval >= 67) {
if (mu_mv_mag_sq < 121) {
if (negBalance < 13) {
if (ssDCval < 92) {
if (posBalance < 8) {
return 2;
}
else if (posBalance >= 8) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Rs < 2368) {
return 1;
}
else if (Rs >= 2368) {
if (Rs < 2475) {
return 8;
}
else if (Rs >= 2475) {
if (Rs < 3079) {
return 1;
}
else if (Rs >= 3079) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 92) {
if (TSCindex < 1) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
if (negBalance < 7) {
return 2;
}
else if (negBalance >= 7) {
if (Rs < 2320) {
return 2;
}
else if (Rs >= 2320) {
if (Rs < 2357) {
return 4;
}
else if (Rs >= 2357) {
return 2;
}
}
}
}
}
else if (TSCindex >= 1) {
if (Rs < 2487) {
return 4;
}
else if (Rs >= 2487) {
return 8;
}
}
}
}
else if (negBalance >= 13) {
if (posBalance < 18) {
if (Rs < 2302) {
return 2;
}
else if (Rs >= 2302) {
return 4;
}
}
else if (posBalance >= 18) {
if (diffTSC < -1) {
if (mu_mv_mag_sq < 120) {
return 8;
}
else if (mu_mv_mag_sq >= 120) {
return 2;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (SC < 7112) {
if (Rs < 1109) {
if (Rs < 623) {
return 8;
}
else if (Rs >= 623) {
return 1;
}
}
else if (Rs >= 1109) {
if (mu_mv_mag_sq < 125) {
if (diffTSC < -1) {
if (Rs < 3005) {
return 8;
}
else if (Rs >= 3005) {
if (Cs < 2326) {
return 2;
}
else if (Cs >= 2326) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (negBalance < 27) {
return 8;
}
else if (negBalance >= 27) {
if (Cs < 2156) {
return 2;
}
else if (Cs >= 2156) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (diffRsCsdiff < -210) {
return 4;
}
else if (diffRsCsdiff >= -210) {
if (SC < 4544) {
if (SC < 4439) {
return 8;
}
else if (SC >= 4439) {
if (AFD < 3) {
if (negBalance < 17) {
if (Rs < 2267) {
return 8;
}
else if (Rs >= 2267) {
return 4;
}
}
else if (negBalance >= 17) {
return 4;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (SC >= 4544) {
if (Cs < 3280) {
return 8;
}
else if (Cs >= 3280) {
if (Rs < 3284) {
return 8;
}
else if (Rs >= 3284) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 7112) {
if (SC < 7935) {
if (ssDCval < 85) {
if (Rs < 4975) {
return 8;
}
else if (Rs >= 4975) {
return 4;
}
}
else if (ssDCval >= 85) {
return 4;
}
}
else if (SC >= 7935) {
if (diffRsCsdiff < -6) {
if (ssDCval < 73) {
if (Rs < 5856) {
return 1;
}
else if (Rs >= 5856) {
return 4;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6650) {
if (SC < 8416) {
if (Rs < 3558) {
if (negBalance < 54) {
if (AFD < 10) {
if (negBalance < 36) {
if (Cs < 1889) {
if (Rs < 2566) {
if (mu_mv_mag_sq < 120) {
if (mcTcor < 99) {
if (gchDC < 4) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 2550) {
return 8;
}
else if (Rs >= 2550) {
return 4;
}
}
}
else if (gchDC >= 4) {
return 4;
}
}
else if (mcTcor >= 99) {
return 4;
}
}
else if (mu_mv_mag_sq >= 120) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 2471) {
return 8;
}
else if (Rs >= 2471) {
if (Cs < 1728) {
return 8;
}
else if (Cs >= 1728) {
return 1;
}
}
}
}
}
else if (Rs >= 2566) {
if (mcTcor < 97) {
if (Rs < 3078) {
return 8;
}
else if (Rs >= 3078) {
return 2;
}
}
else if (mcTcor >= 97) {
if (posBalance < 41) {
if (Cs < 1859) {
if (Rs < 2646) {
return 8;
}
else if (Rs >= 2646) {
if (Rs < 2980) {
if (Rs < 2950) {
if (Rs < 2790) {
return 4;
}
else if (Rs >= 2790) {
if (Rs < 2935) {
return 8;
}
else if (Rs >= 2935) {
return 4;
}
}
}
else if (Rs >= 2950) {
return 4;
}
}
else if (Rs >= 2980) {
if (Cs < 1767) {
return 4;
}
else if (Cs >= 1767) {
if (Rs < 3049) {
return 8;
}
else if (Rs >= 3049) {
if (Cs < 1820) {
if (Rs < 3201) {
return 8;
}
else if (Rs >= 3201) {
return 8;
}
}
else if (Cs >= 1820) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1859) {
if (Rs < 2814) {
return 8;
}
else if (Rs >= 2814) {
return 4;
}
}
}
else if (posBalance >= 41) {
if (Cs < 1823) {
return 4;
}
else if (Cs >= 1823) {
if (negBalance < 26) {
return 4;
}
else if (negBalance >= 26) {
if (Cs < 1864) {
return 8;
}
else if (Cs >= 1864) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
}
}
}
}
else if (Cs >= 1889) {
if (refDCval < 73) {
if (SC < 5416) {
if (Cs < 2116) {
if (Rs < 2764) {
return 8;
}
else if (Rs >= 2764) {
if (RsDiff < 5) {
if (negBalance < 24) {
if (Cs < 1995) {
if (Rs < 2976) {
return 8;
}
else if (Rs >= 2976) {
return 8;
}
}
else if (Cs >= 1995) {
return 4;
}
}
else if (negBalance >= 24) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
else if (RsDiff >= 5) {
return 8;
}
}
}
else if (Cs >= 2116) {
if (AFD < 4) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (SC >= 5416) {
if (posBalance < 44) {
if (mu_mv_mag_sq < 122) {
if (Rs < 3455) {
return 1;
}
else if (Rs >= 3455) {
return 8;
}
}
else if (mu_mv_mag_sq >= 122) {
return 8;
}
}
else if (posBalance >= 44) {
return 8;
}
}
}
else if (refDCval >= 73) {
if (Cs < 3060) {
if (tcor < 99) {
if (Cs < 1925) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
return 8;
}
}
}
else if (Cs >= 1925) {
if (Cs < 2653) {
return 8;
}
else if (Cs >= 2653) {
if (Rs < 2735) {
return 4;
}
else if (Rs >= 2735) {
return 8;
}
}
}
}
else if (tcor >= 99) {
if (Cs < 2260) {
if (Rs < 3041) {
if (ssDCval < 81) {
return 4;
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (Rs >= 3041) {
if (mu_mv_mag_sq < 126) {
return 2;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 3238) {
return 2;
}
else if (Rs >= 3238) {
return 8;
}
}
}
}
else if (Cs >= 2260) {
return 8;
}
}
}
else if (Cs >= 3060) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
return 4;
}
}
}
}
}
else if (negBalance >= 36) {
if (Cs < 4086) {
if (Rs < 1029) {
return 1;
}
else if (Rs >= 1029) {
if (SC < 4257) {
return 8;
}
else if (SC >= 4257) {
if (Rs < 2231) {
if (ssDCval < 71) {
if (Cs < 2368) {
return 8;
}
else if (Cs >= 2368) {
return 4;
}
}
else if (ssDCval >= 71) {
if (diffAFD < 2) {
if (Cs < 2724) {
return 8;
}
else if (Cs >= 2724) {
return 4;
}
}
else if (diffAFD >= 2) {
if (Rs < 2219) {
return 4;
}
else if (Rs >= 2219) {
return 8;
}
}
}
}
else if (Rs >= 2231) {
if (Cs < 1878) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Cs >= 1878) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Cs < 2545) {
if (MVDiff < 3) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
else if (MVDiff >= 3) {
if (Cs < 2032) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 8;
}
}
else if (Cs >= 2032) {
if (Rs < 3277) {
return 8;
}
else if (Rs >= 3277) {
if (Rs < 3289) {
return 4;
}
else if (Rs >= 3289) {
return 8;
}
}
}
}
}
else if (Cs >= 2545) {
if (refDCval < 98) {
if (Rs < 2527) {
return 4;
}
else if (Rs >= 2527) {
if (ssDCval < 93) {
if (SC < 6440) {
return 8;
}
else if (SC >= 6440) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
if (Rs < 3531) {
return 8;
}
else if (Rs >= 3531) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
}
else if (ssDCval >= 93) {
if (Cs < 3232) {
if (Rs < 3145) {
return 8;
}
else if (Rs >= 3145) {
return 4;
}
}
else if (Cs >= 3232) {
return 8;
}
}
}
}
else if (refDCval >= 98) {
return 4;
}
}
}
}
}
}
}
}
else if (Cs >= 4086) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 4;
}
}
}
}
else if (AFD >= 10) {
if (Cs < 2348) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
return 8;
}
}
else if (Cs >= 2348) {
if (RsDiff < 10) {
if (posBalance < 62) {
if (Rs < 3164) {
if (Rs < 2591) {
return 4;
}
else if (Rs >= 2591) {
return 8;
}
}
else if (Rs >= 3164) {
if (Cs < 3129) {
return 4;
}
else if (Cs >= 3129) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
else if (posBalance >= 62) {
return 4;
}
}
else if (RsDiff >= 10) {
return 8;
}
}
}
}
else if (negBalance >= 54) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Rs < 3035) {
return 4;
}
else if (Rs >= 3035) {
if (Rs < 3096) {
return 8;
}
else if (Rs >= 3096) {
return 4;
}
}
}
}
}
}
else if (Rs >= 3558) {
if (refDCval < 81) {
if (refDCval < 60) {
if (Rs < 3950) {
return 4;
}
else if (Rs >= 3950) {
if (Cs < 2767) {
if (Cs < 2141) {
if (diffTSC < 1) {
if (Cs < 2136) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 4389) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 4372) {
if (diffAFD < 1) {
if (ssDCval < 54) {
if (tcor < 94) {
if (Rs < 4366) {
return 8;
}
else if (Rs >= 4366) {
return 8;
}
}
else if (tcor >= 94) {
if (Cs < 2113) {
return 4;
}
else if (Cs >= 2113) {
return 8;
}
}
}
else if (ssDCval >= 54) {
if (Rs < 4204) {
return 4;
}
else if (Rs >= 4204) {
return 4;
}
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (Rs >= 4372) {
return 4;
}
}
}
else if (Rs >= 4389) {
if (Rs < 4417) {
return 8;
}
else if (Rs >= 4417) {
return 8;
}
}
}
}
else if (Cs >= 2136) {
return 4;
}
}
else if (diffTSC >= 1) {
if (Cs < 2111) {
return 4;
}
else if (Cs >= 2111) {
return 4;
}
}
}
else if (Cs >= 2141) {
if (Cs < 2725) {
if (Cs < 2149) {
return 8;
}
else if (Cs >= 2149) {
if (Cs < 2174) {
if (Rs < 4366) {
return 4;
}
else if (Rs >= 4366) {
return 4;
}
}
else if (Cs >= 2174) {
if (mcTcor < 97) {
if (Cs < 2613) {
if (Cs < 2552) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 4764) {
if (MVDiff < 3) {
if (negBalance < 18) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Rs < 4579) {
return 8;
}
else if (Rs >= 4579) {
return 8;
}
}
}
else if (negBalance >= 18) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Rs < 4644) {
return 8;
}
else if (Rs >= 4644) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (Rs >= 4764) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (Cs >= 2552) {
if (RsCsDiff < 37) {
return 4;
}
else if (RsCsDiff >= 37) {
return 4;
}
}
}
else if (Cs >= 2613) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (mu_mv_mag_sq < 115) {
if (Cs < 2367) {
return 4;
}
else if (Cs >= 2367) {
return 4;
}
}
else if (mu_mv_mag_sq >= 115) {
if (Cs < 2530) {
if (Rs < 4385) {
return 8;
}
else if (Rs >= 4385) {
if (Rs < 4682) {
return 8;
}
else if (Rs >= 4682) {
return 8;
}
}
}
else if (Cs >= 2530) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 2725) {
return 4;
}
}
}
else if (Cs >= 2767) {
return 8;
}
}
}
else if (refDCval >= 60) {
if (Cs < 1795) {
if (RsCsDiff < 41) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (RsCsDiff >= 41) {
return 2;
}
}
else if (Cs >= 1795) {
if (SC < 5666) {
return 2;
}
else if (SC >= 5666) {
return 8;
}
}
}
}
else if (refDCval >= 81) {
if (Cs < 2457) {
if (AFD < 8) {
if (negBalance < 28) {
if (Rs < 3860) {
if (Cs < 2217) {
return 2;
}
else if (Cs >= 2217) {
return 8;
}
}
else if (Rs >= 3860) {
if (SC < 6231) {
if (ssDCval < 95) {
return 2;
}
else if (ssDCval >= 95) {
return 4;
}
}
else if (SC >= 6231) {
if (Rs < 4967) {
if (Cs < 1985) {
if (Rs < 4480) {
if (Rs < 4446) {
return 4;
}
else if (Rs >= 4446) {
return 8;
}
}
else if (Rs >= 4480) {
return 2;
}
}
else if (Cs >= 1985) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (Rs >= 4967) {
return 4;
}
}
}
}
else if (negBalance >= 28) {
if (RsDiff < 7) {
if (AFD < 4) {
if (tcor < 98) {
if (Rs < 3768) {
return 8;
}
else if (Rs >= 3768) {
return 4;
}
}
else if (tcor >= 98) {
return 2;
}
}
else if (AFD >= 4) {
if (ssDCval < 91) {
if (Rs < 4634) {
return 2;
}
else if (Rs >= 4634) {
return 8;
}
}
else if (ssDCval >= 91) {
if (negBalance < 29) {
return 2;
}
else if (negBalance >= 29) {
return 8;
}
}
}
}
else if (RsDiff >= 7) {
return 2;
}
}
}
else if (AFD >= 8) {
return 1;
}
}
else if (Cs >= 2457) {
if (ssDCval < 98) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (RsCsDiff < 38) {
if (SC < 7101) {
if (diffAFD < -3) {
if (Rs < 3708) {
return 4;
}
else if (Rs >= 3708) {
return 8;
}
}
else if (diffAFD >= -3) {
if (Cs < 3285) {
return 8;
}
else if (Cs >= 3285) {
return 4;
}
}
}
else if (SC >= 7101) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 97) {
if (Rs < 4933) {
return 8;
}
else if (Rs >= 4933) {
return 4;
}
}
else if (ssDCval >= 97) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
return 4;
}
}
}
else if (RsCsDiff >= 38) {
if (Cs < 3718) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (tcor < 91) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (tcor >= 91) {
return 4;
}
}
}
else if (Cs >= 3718) {
return 8;
}
}
}
}
else if (ssDCval >= 98) {
if (mu_mv_mag_sq < 112) {
return 1;
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
}
}
}
}
else if (SC >= 8416) {
if (ssDCval < 79) {
if (Rs < 5968) {
if (negBalance < 29) {
if (Rs < 4653) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (Rs >= 4653) {
return 8;
}
}
else if (negBalance >= 29) {
return 8;
}
}
else if (Rs >= 5968) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 4;
}
}
}
}
else if (ssDCval >= 79) {
if (Cs < 3664) {
if (Rs < 5223) {
return 4;
}
else if (Rs >= 5223) {
return 8;
}
}
else if (Cs >= 3664) {
return 8;
}
}
}
}
else if (Rs >= 6650) {
if (ssDCval < 95) {
if (RsDiff < 43) {
return 4;
}
else if (RsDiff >= 43) {
return 8;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
else if (ssDCval >= 102) {
if (posBalance < 53) {
if (ssDCval < 200) {
if (ssDCval < 117) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 116) {
if (posBalance < 40) {
if (mu_mv_mag_sq < 115) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 115) {
if (MVDiff < 4) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (posBalance >= 40) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
else if (negBalance >= 37) {
if (RsCsDiff < 38) {
if (Rs < 4864) {
if (diffRsCsdiff < 16) {
if (Rs < 1394) {
return 2;
}
else if (Rs >= 1394) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
return 8;
}
}
}
else if (diffRsCsdiff >= 16) {
return 4;
}
}
else if (Rs >= 4864) {
if (Rs < 5376) {
return 2;
}
else if (Rs >= 5376) {
return 8;
}
}
}
else if (RsCsDiff >= 38) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (mu_mv_mag_sq < 125) {
if (posBalance < 32) {
if (ssDCval < 133) {
if (Rs < 2770) {
if (Rs < 2347) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Rs < 1925) {
return 2;
}
else if (Rs >= 1925) {
return 1;
}
}
}
else if (Rs >= 2347) {
if (posBalance < 16) {
if (Rs < 2638) {
return 4;
}
else if (Rs >= 2638) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (posBalance >= 16) {
return 4;
}
}
}
else if (Rs >= 2770) {
return 2;
}
}
else if (ssDCval >= 133) {
if (Cs < 1877) {
return 2;
}
else if (Cs >= 1877) {
return 8;
}
}
}
else if (posBalance >= 32) {
if (Rs < 6351) {
if (negBalance < 42) {
if (Cs < 1675) {
if (Rs < 3979) {
return 8;
}
else if (Rs >= 3979) {
return 2;
}
}
else if (Cs >= 1675) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 1;
}
}
}
else if (negBalance >= 42) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 4;
}
}
}
else if (Rs >= 6351) {
if (SC < 10080) {
return 4;
}
else if (SC >= 10080) {
if (Rs < 6970) {
return 8;
}
else if (Rs >= 6970) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 4112) {
if (Cs < 3891) {
if (negBalance < 24) {
if (diffMVdiffVal < -3) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (Rs < 2297) {
return 1;
}
else if (Rs >= 2297) {
return 8;
}
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (negBalance >= 24) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 8;
}
}
}
else if (Cs >= 3891) {
return 1;
}
}
else if (Rs >= 4112) {
return 8;
}
}
}
}
else if (ssDCval >= 200) {
if (Cs < 4080) {
return 8;
}
else if (Cs >= 4080) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (Rs < 1582) {
if (Cs < 1850) {
if (mu_mv_mag_sq < 128) {
if (CsDiff < 3) {
if (Rs < 1143) {
return 8;
}
else if (Rs >= 1143) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 723) {
return 1;
}
else if (Rs >= 723) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 2;
}
}
else if (Cs >= 1850) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
if (Rs < 794) {
return 8;
}
else if (Rs >= 794) {
return 4;
}
}
}
}
else if (Rs >= 1582) {
if (ssDCval < 137) {
if (refDCval < 105) {
if (Cs < 2816) {
if (CsDiff < 6) {
if (Rs < 4222) {
return 8;
}
else if (Rs >= 4222) {
return 2;
}
}
else if (CsDiff >= 6) {
if (Rs < 4961) {
return 1;
}
else if (Rs >= 4961) {
return 2;
}
}
}
else if (Cs >= 2816) {
return 8;
}
}
else if (refDCval >= 105) {
if (RsDiff < 2) {
if (Cs < 1953) {
return 8;
}
else if (Cs >= 1953) {
if (negBalance < 27) {
return 8;
}
else if (negBalance >= 27) {
return 2;
}
}
}
else if (RsDiff >= 2) {
if (posBalance < 63) {
if (gchDC < 1) {
if (diffAFD < -3) {
if (mu_mv_mag_sq < 119) {
if (Rs < 3528) {
return 1;
}
else if (Rs >= 3528) {
return 4;
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
else if (diffAFD >= -3) {
if (diffAFD < 3) {
if (negBalance < 36) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (negBalance >= 36) {
if (negBalance < 55) {
if (mu_mv_mag_sq < 110) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
if (Rs < 3362) {
return 1;
}
else if (Rs >= 3362) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 110) {
return 8;
}
}
else if (negBalance >= 55) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
}
else if (diffAFD >= 3) {
if (Rs < 3566) {
return 4;
}
else if (Rs >= 3566) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
if (diffRsCsdiff < -1) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Cs < 3465) {
if (AFD < 12) {
if (SC < 6517) {
return 8;
}
else if (SC >= 6517) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (AFD >= 12) {
return 2;
}
}
else if (Cs >= 3465) {
return 8;
}
}
}
else if (diffRsCsdiff >= -1) {
if (AFD < 29) {
if (diffTSC < -1) {
if (Cs < 1836) {
return 1;
}
else if (Cs >= 1836) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (AFD >= 29) {
return 1;
}
}
}
}
else if (posBalance >= 63) {
if (Cs < 1943) {
if (Rs < 2345) {
return 8;
}
else if (Rs >= 2345) {
return 1;
}
}
else if (Cs >= 1943) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
if (mcTcor < 96) {
if (Rs < 5949) {
return 1;
}
else if (Rs >= 5949) {
if (Rs < 8232) {
return 8;
}
else if (Rs >= 8232) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (tcor < 97) {
if (Rs < 5426) {
return 1;
}
else if (Rs >= 5426) {
return 8;
}
}
else if (tcor >= 97) {
if (diffAFD < -1) {
if (Rs < 5513) {
return 1;
}
else if (Rs >= 5513) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect11(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1313) {
if (Rs < 2377) {
if (Cs < 533) {
if (negBalance < 14) {
if (MVDiff < 7) {
if (SC < 793) {
if (ssDCval < 84) {
if (refDCval < 47) {
if (SC < 287) {
if (negBalance < 8) {
if (Rs < 3) {
if (ssDCval < 22) {
if (ssDCval < 9) {
return 4;
}
else if (ssDCval >= 9) {
return 8;
}
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (Rs >= 3) {
return 8;
}
}
else if (negBalance >= 8) {
if (ssDCval < 38) {
if (Rs < 97) {
if (ssDCval < 24) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
else if (ssDCval >= 24) {
if (mu_mv_mag_sq < 85) {
if (Rs < 60) {
return 1;
}
else if (Rs >= 60) {
return 8;
}
}
else if (mu_mv_mag_sq >= 85) {
if (posBalance < 40) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (posBalance >= 40) {
if (Rs < 29) {
return 4;
}
else if (Rs >= 29) {
if (Rs < 53) {
return 1;
}
else if (Rs >= 53) {
return 8;
}
}
}
}
}
}
else if (Rs >= 97) {
if (posBalance < 33) {
if (tcor < 91) {
return 4;
}
else if (tcor >= 91) {
return 8;
}
}
else if (posBalance >= 33) {
if (mu_mv_mag_sq < 97) {
return 8;
}
else if (mu_mv_mag_sq >= 97) {
return 1;
}
}
}
}
else if (ssDCval >= 38) {
if (Rs < 89) {
return 8;
}
else if (Rs >= 89) {
return 1;
}
}
}
}
else if (SC >= 287) {
if (Rs < 223) {
if (SC < 439) {
if (ssDCval < 35) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (Cs < 219) {
return 1;
}
else if (Cs >= 219) {
return 8;
}
}
}
else if (ssDCval >= 35) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 1;
}
}
}
else if (SC >= 439) {
if (mu_mv_mag_sq < 108) {
if (Rs < 169) {
return 1;
}
else if (Rs >= 169) {
return 4;
}
}
else if (mu_mv_mag_sq >= 108) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
return 1;
}
}
}
}
else if (Rs >= 223) {
if (tcor < 102) {
if (TSCindex < 1) {
if (Rs < 427) {
return 4;
}
else if (Rs >= 427) {
return 8;
}
}
else if (TSCindex >= 1) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 40) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (ssDCval >= 40) {
if (Cs < 224) {
if (Rs < 348) {
return 4;
}
else if (Rs >= 348) {
return 8;
}
}
else if (Cs >= 224) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
else if (tcor >= 102) {
return 8;
}
}
}
}
else if (refDCval >= 47) {
if (Cs < 7) {
return 2;
}
else if (Cs >= 7) {
if (ssDCval < 61) {
if (Cs < 276) {
if (ssDCval < 58) {
if (refDCval < 53) {
if (SC < 386) {
return 8;
}
else if (SC >= 386) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (refDCval >= 53) {
if (negBalance < 10) {
if (Cs < 123) {
return 8;
}
else if (Cs >= 123) {
if (Rs < 230) {
return 4;
}
else if (Rs >= 230) {
return 8;
}
}
}
else if (negBalance >= 10) {
return 4;
}
}
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (Cs >= 276) {
if (Rs < 252) {
return 4;
}
else if (Rs >= 252) {
if (SC < 692) {
return 1;
}
else if (SC >= 692) {
return 8;
}
}
}
}
else if (ssDCval >= 61) {
if (SC < 531) {
if (Rs < 78) {
if (Cs < 56) {
return 2;
}
else if (Cs >= 56) {
return 4;
}
}
else if (Rs >= 78) {
if (AFD < 2) {
if (ssDCval < 80) {
if (Cs < 284) {
return 2;
}
else if (Cs >= 284) {
return 4;
}
}
else if (ssDCval >= 80) {
return 8;
}
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (SC >= 531) {
if (Rs < 386) {
if (Cs < 288) {
return 4;
}
else if (Cs >= 288) {
return 8;
}
}
else if (Rs >= 386) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
if (TSC < 2) {
if (posBalance < 12) {
return 2;
}
else if (posBalance >= 12) {
return 4;
}
}
else if (TSC >= 2) {
if (Rs < 366) {
return 8;
}
else if (Rs >= 366) {
return 1;
}
}
}
}
}
}
else if (SC >= 793) {
if (mu_mv_mag_sq < 128) {
if (Cs < 268) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
if (diffMVdiffVal < 3) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 1060) {
return 8;
}
else if (Rs >= 1060) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
else if (Cs >= 268) {
if (Rs < 474) {
if (RsCsDiff < 1) {
if (Rs < 348) {
return 1;
}
else if (Rs >= 348) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
return 1;
}
}
}
else if (RsCsDiff >= 1) {
return 1;
}
}
else if (Rs >= 474) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Cs < 374) {
if (AFD < 4) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (tcor < 100) {
return 8;
}
else if (tcor >= 100) {
if (Rs < 844) {
return 8;
}
else if (Rs >= 844) {
return 1;
}
}
}
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 374) {
if (Cs < 485) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
if (Rs < 1308) {
if (Rs < 770) {
if (Rs < 689) {
return 8;
}
else if (Rs >= 689) {
return 1;
}
}
else if (Rs >= 770) {
return 8;
}
}
else if (Rs >= 1308) {
if (SC < 1889) {
return 1;
}
else if (SC >= 1889) {
return 8;
}
}
}
}
else if (Cs >= 485) {
if (Rs < 1011) {
return 2;
}
else if (Rs >= 1011) {
return 1;
}
}
}
}
else if (ssDCval >= 60) {
if (Cs < 457) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (Cs >= 457) {
if (Rs < 1471) {
return 2;
}
else if (Rs >= 1471) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (Rs < 547) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (posBalance < 13) {
return 8;
}
else if (posBalance >= 13) {
if (Rs < 533) {
return 1;
}
else if (Rs >= 533) {
if (Cs < 395) {
return 1;
}
else if (Cs >= 395) {
return 8;
}
}
}
}
}
else if (Rs >= 547) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (RsDiff < 1) {
if (Rs < 1300) {
return 4;
}
else if (Rs >= 1300) {
return 8;
}
}
else if (RsDiff >= 1) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (refDCval < 60) {
return 8;
}
else if (refDCval >= 60) {
if (ssDCval < 84) {
if (Cs < 349) {
if (Rs < 1101) {
return 1;
}
else if (Rs >= 1101) {
return 2;
}
}
else if (Cs >= 349) {
if (Rs < 1182) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 4;
}
}
else if (Rs >= 1182) {
return 2;
}
}
}
else if (ssDCval >= 84) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (SC < 1092) {
if (Rs < 575) {
return 8;
}
else if (Rs >= 575) {
return 4;
}
}
else if (SC >= 1092) {
return 8;
}
}
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (tcor < 94) {
if (negBalance < 5) {
if (AFD < 8) {
if (refDCval < 15) {
return 4;
}
else if (refDCval >= 15) {
if (Rs < 55) {
if (ssDCval < 52) {
if (MVDiff < 207) {
if (mu_mv_mag_sq < 82) {
return 8;
}
else if (mu_mv_mag_sq >= 82) {
if (TSC < 1) {
if (ssDCval < 21) {
return 8;
}
else if (ssDCval >= 21) {
return 4;
}
}
else if (TSC >= 1) {
return 8;
}
}
}
else if (MVDiff >= 207) {
if (negBalance < 1) {
return 8;
}
else if (negBalance >= 1) {
return 2;
}
}
}
else if (ssDCval >= 52) {
return 2;
}
}
else if (Rs >= 55) {
if (Cs < 63) {
if (tcor < 82) {
return 8;
}
else if (tcor >= 82) {
if (Rs < 1561) {
return 4;
}
else if (Rs >= 1561) {
return 8;
}
}
}
else if (Cs >= 63) {
if (TSC < 4) {
if (mu_mv_mag_sq < 114) {
if (MVDiff < 619) {
return 8;
}
else if (MVDiff >= 619) {
return 4;
}
}
else if (mu_mv_mag_sq >= 114) {
if (Rs < 199) {
return 8;
}
else if (Rs >= 199) {
return 4;
}
}
}
else if (TSC >= 4) {
return 4;
}
}
}
}
}
else if (AFD >= 8) {
if (mcTcor < 67) {
if (tcor < 4) {
if (Rs < 884) {
return 4;
}
else if (Rs >= 884) {
return 8;
}
}
else if (tcor >= 4) {
return 1;
}
}
else if (mcTcor >= 67) {
if (tcor < 71) {
if (MVDiff < 331) {
if (Cs < 262) {
return 2;
}
else if (Cs >= 262) {
return 4;
}
}
else if (MVDiff >= 331) {
return 8;
}
}
else if (tcor >= 71) {
return 8;
}
}
}
}
else if (negBalance >= 5) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (posBalance < 67) {
if (ssDCval < 49) {
if (Cs < 176) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 2;
}
}
else if (Cs >= 176) {
if (tcor < 91) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
if (mu_mv_mag_sq < 46) {
if (Rs < 283) {
return 4;
}
else if (Rs >= 283) {
return 1;
}
}
else if (mu_mv_mag_sq >= 46) {
return 1;
}
}
}
else if (tcor >= 91) {
if (Cs < 318) {
return 4;
}
else if (Cs >= 318) {
return 1;
}
}
}
}
else if (ssDCval >= 49) {
if (Rs < 417) {
return 2;
}
else if (Rs >= 417) {
return 8;
}
}
}
else if (posBalance >= 67) {
if (mu_mv_mag_sq < 24) {
if (refDCval < 58) {
if (Rs < 224) {
if (Rs < 42) {
if (Cs < 1) {
return 2;
}
else if (Cs >= 1) {
return 8;
}
}
else if (Rs >= 42) {
if (Cs < 162) {
return 1;
}
else if (Cs >= 162) {
return 8;
}
}
}
else if (Rs >= 224) {
if (diffRsCsdiff < 8) {
if (refDCval < 53) {
return 2;
}
else if (refDCval >= 53) {
return 8;
}
}
else if (diffRsCsdiff >= 8) {
if (Rs < 1339) {
return 1;
}
else if (Rs >= 1339) {
return 2;
}
}
}
}
else if (refDCval >= 58) {
return 1;
}
}
else if (mu_mv_mag_sq >= 24) {
if (posBalance < 111) {
if (MVDiff < 55) {
if (diffMVdiffVal < -5) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (Rs < 192) {
return 2;
}
else if (Rs >= 192) {
return 1;
}
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (MVDiff >= 55) {
if (gchDC < 11) {
return 2;
}
else if (gchDC >= 11) {
return 1;
}
}
}
else if (posBalance >= 111) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 1;
}
}
}
}
}
}
}
}
else if (tcor >= 94) {
if (negBalance < 6) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
return 8;
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (refDCval < 29) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (Rs < 91) {
return 8;
}
else if (Rs >= 91) {
return 4;
}
}
}
else if (refDCval >= 29) {
if (SC < 1309) {
if (Rs < 10) {
if (MVDiff < 445) {
return 2;
}
else if (MVDiff >= 445) {
return 1;
}
}
else if (Rs >= 10) {
if (diffRsCsdiff < -1) {
if (ssDCval < 159) {
if (SC < 651) {
if (Rs < 353) {
return 4;
}
else if (Rs >= 353) {
return 8;
}
}
else if (SC >= 651) {
if (ssDCval < 109) {
if (MVDiff < 292) {
return 4;
}
else if (MVDiff >= 292) {
return 1;
}
}
else if (ssDCval >= 109) {
if (MVDiff < 57) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 4;
}
}
else if (MVDiff >= 57) {
return 4;
}
}
}
}
else if (ssDCval >= 159) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
if (MVDiff < 37) {
if (Rs < 822) {
if (MVDiff < 8) {
if (negBalance < 4) {
return 8;
}
else if (negBalance >= 4) {
return 1;
}
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (Rs >= 822) {
return 4;
}
}
else if (MVDiff >= 37) {
if (Cs < 104) {
return 4;
}
else if (Cs >= 104) {
if (Cs < 148) {
if (Rs < 253) {
return 8;
}
else if (Rs >= 253) {
return 4;
}
}
else if (Cs >= 148) {
return 8;
}
}
}
}
}
}
else if (SC >= 1309) {
if (diffAFD < -8) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 2;
}
}
else if (diffAFD >= -8) {
if (Rs < 1381) {
return 1;
}
else if (Rs >= 1381) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 6) {
if (SC < 875) {
if (Cs < 118) {
if (ssDCval < 40) {
if (posBalance < 17) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
return 4;
}
}
else if (posBalance >= 17) {
if (Cs < 107) {
return 1;
}
else if (Cs >= 107) {
return 8;
}
}
}
else if (ssDCval >= 40) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (diffAFD < -2) {
if (Rs < 31) {
return 2;
}
else if (Rs >= 31) {
return 8;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
}
}
else if (Cs >= 118) {
if (mu_mv_mag_sq < 66) {
return 4;
}
else if (mu_mv_mag_sq >= 66) {
if (mu_mv_mag_sq < 117) {
if (diffMVdiffVal < 5) {
if (ssDCval < 133) {
if (ssDCval < 82) {
if (ssDCval < 39) {
if (Rs < 152) {
return 8;
}
else if (Rs >= 152) {
return 4;
}
}
else if (ssDCval >= 39) {
if (diffMVdiffVal < 4) {
return 1;
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
}
else if (ssDCval >= 82) {
return 4;
}
}
else if (ssDCval >= 133) {
if (ssDCval < 161) {
return 1;
}
else if (ssDCval >= 161) {
return 4;
}
}
}
else if (diffMVdiffVal >= 5) {
if (negBalance < 10) {
if (diffMVdiffVal < 6) {
if (Rs < 114) {
return 4;
}
else if (Rs >= 114) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
return 4;
}
}
else if (negBalance >= 10) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
}
}
else if (SC >= 875) {
if (Cs < 397) {
if (Cs < 48) {
return 8;
}
else if (Cs >= 48) {
if (ssDCval < 76) {
if (Rs < 1208) {
return 8;
}
else if (Rs >= 1208) {
return 1;
}
}
else if (ssDCval >= 76) {
return 1;
}
}
}
else if (Cs >= 397) {
if (SC < 924) {
return 4;
}
else if (SC >= 924) {
if (Rs < 1453) {
return 8;
}
else if (Rs >= 1453) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (mcTcor < 83) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (ssDCval < 23) {
if (tcor < 55) {
if (Rs < 1) {
return 8;
}
else if (Rs >= 1) {
return 1;
}
}
else if (tcor >= 55) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
return 4;
}
}
}
}
else if (ssDCval >= 23) {
if (Cs < 377) {
if (Rs < 506) {
if (negBalance < 44) {
if (SC < 302) {
if (TSCindex < 2) {
if (Rs < 26) {
return 1;
}
else if (Rs >= 26) {
return 2;
}
}
else if (TSCindex >= 2) {
if (Cs < 206) {
return 1;
}
else if (Cs >= 206) {
return 2;
}
}
}
else if (SC >= 302) {
if (posBalance < 46) {
if (Rs < 419) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 1;
}
}
else if (Rs >= 419) {
return 8;
}
}
else if (posBalance >= 46) {
if (Cs < 240) {
if (Cs < 164) {
if (Rs < 180) {
return 8;
}
else if (Rs >= 180) {
return 1;
}
}
else if (Cs >= 164) {
return 8;
}
}
else if (Cs >= 240) {
if (Rs < 339) {
return 2;
}
else if (Rs >= 339) {
return 1;
}
}
}
}
}
else if (negBalance >= 44) {
if (tcor < 67) {
if (MVDiff < 377) {
return 8;
}
else if (MVDiff >= 377) {
return 4;
}
}
else if (tcor >= 67) {
if (Rs < 37) {
return 4;
}
else if (Rs >= 37) {
if (Rs < 229) {
return 2;
}
else if (Rs >= 229) {
return 1;
}
}
}
}
}
else if (Rs >= 506) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
else if (Cs >= 377) {
if (TSC < 3) {
if (RsDiff < 4) {
if (Rs < 568) {
return 2;
}
else if (Rs >= 568) {
return 8;
}
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (TSC >= 3) {
if (mcTcor < 70) {
if (Rs < 436) {
return 2;
}
else if (Rs >= 436) {
return 8;
}
}
else if (mcTcor >= 70) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 83) {
if (posBalance < 27) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 116) {
if (negBalance < 24) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (negBalance >= 24) {
if (Rs < 82) {
if (Cs < 1) {
return 4;
}
else if (Cs >= 1) {
return 8;
}
}
else if (Rs >= 82) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
else if (MVDiff >= 5) {
if (SC < 23) {
if (refDCval < 24) {
return 8;
}
else if (refDCval >= 24) {
if (negBalance < 77) {
if (Rs < 1) {
return 4;
}
else if (Rs >= 1) {
if (Rs < 7) {
return 8;
}
else if (Rs >= 7) {
return 4;
}
}
}
else if (negBalance >= 77) {
if (Rs < 14) {
return 8;
}
else if (Rs >= 14) {
return 4;
}
}
}
}
else if (SC >= 23) {
if (ssDCval < 31) {
if (posBalance < 4) {
return 4;
}
else if (posBalance >= 4) {
if (ssDCval < 28) {
return 1;
}
else if (ssDCval >= 28) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
if (Rs < 22) {
return 8;
}
else if (Rs >= 22) {
return 1;
}
}
}
}
}
else if (ssDCval >= 31) {
if (mu_mv_mag_sq < 74) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 74) {
if (Rs < 382) {
return 1;
}
else if (Rs >= 382) {
if (Rs < 462) {
return 2;
}
else if (Rs >= 462) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 27) {
if (Rs < 69) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
if (Rs < 3) {
return 4;
}
else if (Rs >= 3) {
return 2;
}
}
}
else if (Rs >= 69) {
if (SC < 217) {
if (mu_mv_mag_sq < 37) {
if (Cs < 83) {
if (posBalance < 42) {
if (gchDC < 2) {
if (mcTcor < 88) {
return 1;
}
else if (mcTcor >= 88) {
return 4;
}
}
else if (gchDC >= 2) {
if (Cs < 48) {
return 1;
}
else if (Cs >= 48) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (MVDiff < 79) {
return 1;
}
else if (MVDiff >= 79) {
return 8;
}
}
}
else if (Cs >= 83) {
return 8;
}
}
else if (mu_mv_mag_sq >= 37) {
if (refDCval < 31) {
if (MVDiff < 2) {
if (Rs < 85) {
return 1;
}
else if (Rs >= 85) {
return 4;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (refDCval >= 31) {
if (Cs < 93) {
return 1;
}
else if (Cs >= 93) {
return 8;
}
}
}
}
else if (SC >= 217) {
if (posBalance < 39) {
if (refDCval < 29) {
if (Cs < 301) {
if (Cs < 32) {
return 4;
}
else if (Cs >= 32) {
return 1;
}
}
else if (Cs >= 301) {
if (mu_mv_mag_sq < 60) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 494) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (Rs >= 494) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 60) {
return 1;
}
}
}
else if (refDCval >= 29) {
if (mu_mv_mag_sq < 18) {
return 4;
}
else if (mu_mv_mag_sq >= 18) {
if (AFD < 4) {
if (mcTcor < 96) {
if (Cs < 237) {
if (refDCval < 33) {
if (refDCval < 30) {
if (Rs < 365) {
return 4;
}
else if (Rs >= 365) {
return 1;
}
}
else if (refDCval >= 30) {
if (Cs < 139) {
return 1;
}
else if (Cs >= 139) {
if (Rs < 318) {
return 8;
}
else if (Rs >= 318) {
if (Rs < 325) {
return 4;
}
else if (Rs >= 325) {
return 8;
}
}
}
}
}
else if (refDCval >= 33) {
if (Rs < 245) {
return 8;
}
else if (Rs >= 245) {
return 4;
}
}
}
else if (Cs >= 237) {
return 1;
}
}
else if (mcTcor >= 96) {
if (negBalance < 15) {
return 4;
}
else if (negBalance >= 15) {
if (Rs < 373) {
return 1;
}
else if (Rs >= 373) {
if (Rs < 750) {
return 4;
}
else if (Rs >= 750) {
return 1;
}
}
}
}
}
else if (AFD >= 4) {
if (posBalance < 30) {
if (Cs < 156) {
return 2;
}
else if (Cs >= 156) {
return 1;
}
}
else if (posBalance >= 30) {
if (negBalance < 33) {
if (Cs < 128) {
if (Cs < 103) {
return 1;
}
else if (Cs >= 103) {
if (Rs < 210) {
return 1;
}
else if (Rs >= 210) {
return 4;
}
}
}
else if (Cs >= 128) {
return 1;
}
}
else if (negBalance >= 33) {
if (Rs < 336) {
return 1;
}
else if (Rs >= 336) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 39) {
if (negBalance < 28) {
if (TSCindex < 3) {
if (diffMVdiffVal < 9) {
return 1;
}
else if (diffMVdiffVal >= 9) {
if (Rs < 481) {
return 8;
}
else if (Rs >= 481) {
return 4;
}
}
}
else if (TSCindex >= 3) {
if (MVDiff < 23) {
if (mcTcor < 90) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 77) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (mu_mv_mag_sq >= 77) {
return 8;
}
}
else if (MVDiff >= 6) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (mcTcor >= 90) {
if (ssDCval < 36) {
if (posBalance < 40) {
if (Rs < 229) {
return 1;
}
else if (Rs >= 229) {
return 8;
}
}
else if (posBalance >= 40) {
return 1;
}
}
else if (ssDCval >= 36) {
return 4;
}
}
}
else if (MVDiff >= 23) {
if (Cs < 310) {
return 8;
}
else if (Cs >= 310) {
if (Rs < 502) {
return 1;
}
else if (Rs >= 502) {
return 8;
}
}
}
}
}
else if (negBalance >= 28) {
if (SC < 369) {
if (ssDCval < 33) {
if (SC < 226) {
return 4;
}
else if (SC >= 226) {
return 1;
}
}
else if (ssDCval >= 33) {
if (Cs < 156) {
if (Rs < 200) {
return 8;
}
else if (Rs >= 200) {
return 1;
}
}
else if (Cs >= 156) {
if (Rs < 151) {
return 4;
}
else if (Rs >= 151) {
return 2;
}
}
}
}
else if (SC >= 369) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 307) {
if (ssDCval < 50) {
if (mcTcor < 93) {
if (posBalance < 47) {
if (negBalance < 29) {
if (mu_mv_mag_sq < 65) {
if (Rs < 294) {
return 8;
}
else if (Rs >= 294) {
return 1;
}
}
else if (mu_mv_mag_sq >= 65) {
if (Rs < 55) {
return 8;
}
else if (Rs >= 55) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
return 4;
}
}
}
}
else if (negBalance >= 29) {
if (MVDiff < 30) {
if (Cs < 59) {
return 2;
}
else if (Cs >= 59) {
return 8;
}
}
else if (MVDiff >= 30) {
if (negBalance < 123) {
if (RsCsDiff < 1) {
if (diffAFD < -1) {
if (mu_mv_mag_sq < 38) {
return 2;
}
else if (mu_mv_mag_sq >= 38) {
return 1;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (RsCsDiff >= 1) {
if (Cs < 22) {
return 1;
}
else if (Cs >= 22) {
if (diffAFD < 3) {
if (mcTcor < 84) {
if (ssDCval < 35) {
if (diffRsCsdiff < -39) {
return 2;
}
else if (diffRsCsdiff >= -39) {
return 1;
}
}
else if (ssDCval >= 35) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
return 1;
}
}
}
else if (mcTcor >= 84) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Rs < 106) {
return 1;
}
else if (Rs >= 106) {
return 2;
}
}
}
}
else if (diffAFD >= 3) {
if (Cs < 347) {
return 2;
}
else if (Cs >= 347) {
return 1;
}
}
}
}
}
else if (negBalance >= 123) {
return 8;
}
}
}
}
else if (posBalance >= 47) {
if (ssDCval < 41) {
if (MVDiff < 30) {
if (RsCsDiff < 1) {
if (posBalance < 67) {
if (ssDCval < 34) {
return 2;
}
else if (ssDCval >= 34) {
return 8;
}
}
else if (posBalance >= 67) {
return 4;
}
}
else if (RsCsDiff >= 1) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
if (mcTcor < 89) {
if (Rs < 252) {
if (Rs < 140) {
return 8;
}
else if (Rs >= 140) {
if (posBalance < 54) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
else if (posBalance >= 54) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (Rs >= 252) {
return 8;
}
}
else if (mcTcor >= 89) {
if (Rs < 181) {
if (ssDCval < 35) {
return 2;
}
else if (ssDCval >= 35) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (Rs >= 181) {
if (Cs < 240) {
return 1;
}
else if (Cs >= 240) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 30) {
if (mcTcor < 84) {
if (mu_mv_mag_sq < 14) {
if (posBalance < 57) {
if (Rs < 58) {
return 1;
}
else if (Rs >= 58) {
return 8;
}
}
else if (posBalance >= 57) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (refDCval < 41) {
return 8;
}
else if (refDCval >= 41) {
return 4;
}
}
}
else if (mcTcor >= 84) {
if (mu_mv_mag_sq < 36) {
if (MVDiff < 32) {
return 2;
}
else if (MVDiff >= 32) {
return 1;
}
}
else if (mu_mv_mag_sq >= 36) {
if (refDCval < 37) {
return 1;
}
else if (refDCval >= 37) {
return 8;
}
}
}
}
}
else if (ssDCval >= 41) {
if (posBalance < 64) {
if (ssDCval < 45) {
if (Cs < 106) {
if (SC < 111) {
if (Rs < 43) {
return 8;
}
else if (Rs >= 43) {
return 2;
}
}
else if (SC >= 111) {
if (diffRsCsdiff < -4) {
return 4;
}
else if (diffRsCsdiff >= -4) {
if (refDCval < 46) {
return 1;
}
else if (refDCval >= 46) {
if (Rs < 206) {
return 4;
}
else if (Rs >= 206) {
return 8;
}
}
}
}
}
else if (Cs >= 106) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
if (Rs < 201) {
return 1;
}
else if (Rs >= 201) {
return 2;
}
}
}
}
}
else if (ssDCval >= 45) {
if (refDCval < 48) {
if (diffTSC < -1) {
if (Rs < 100) {
return 4;
}
else if (Rs >= 100) {
return 1;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (refDCval >= 48) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
}
else if (posBalance >= 64) {
if (MVDiff < 59) {
if (Rs < 282) {
return 4;
}
else if (Rs >= 282) {
return 8;
}
}
else if (MVDiff >= 59) {
if (mu_mv_mag_sq < 103) {
if (Rs < 8) {
return 8;
}
else if (Rs >= 8) {
return 2;
}
}
else if (mu_mv_mag_sq >= 103) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (mu_mv_mag_sq < 92) {
if (posBalance < 21) {
if (mcTcor < 105) {
if (negBalance < 32) {
if (Rs < 49) {
return 2;
}
else if (Rs >= 49) {
return 1;
}
}
else if (negBalance >= 32) {
if (negBalance < 110) {
if (mcTcor < 102) {
return 8;
}
else if (mcTcor >= 102) {
if (ssDCval < 34) {
return 1;
}
else if (ssDCval >= 34) {
return 8;
}
}
}
else if (negBalance >= 110) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (mcTcor >= 105) {
if (Cs < 126) {
return 4;
}
else if (Cs >= 126) {
return 8;
}
}
}
else if (posBalance >= 21) {
if (mcTcor < 97) {
if (AFD < 4) {
if (Rs < 181) {
if (Rs < 17) {
return 8;
}
else if (Rs >= 17) {
return 1;
}
}
else if (Rs >= 181) {
if (diffMVdiffVal < -5) {
return 1;
}
else if (diffMVdiffVal >= -5) {
if (negBalance < 37) {
return 8;
}
else if (negBalance >= 37) {
return 2;
}
}
}
}
else if (AFD >= 4) {
if (posBalance < 68) {
if (Cs < 94) {
if (SC < 97) {
if (diffMVdiffVal < -6) {
if (Rs < 27) {
return 4;
}
else if (Rs >= 27) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
return 2;
}
}
else if (SC >= 97) {
if (posBalance < 45) {
if (gchDC < 1) {
if (Rs < 92) {
return 4;
}
else if (Rs >= 92) {
return 1;
}
}
else if (gchDC >= 1) {
if (Cs < 36) {
if (Rs < 156) {
return 2;
}
else if (Rs >= 156) {
return 1;
}
}
else if (Cs >= 36) {
return 1;
}
}
}
else if (posBalance >= 45) {
return 1;
}
}
}
else if (Cs >= 94) {
if (AFD < 10) {
if (ssDCval < 42) {
if (MVDiff < 9) {
if (tcor < 86) {
return 1;
}
else if (tcor >= 86) {
return 8;
}
}
else if (MVDiff >= 9) {
if (mu_mv_mag_sq < 36) {
if (Rs < 102) {
return 4;
}
else if (Rs >= 102) {
return 2;
}
}
else if (mu_mv_mag_sq >= 36) {
if (Cs < 260) {
if (Rs < 167) {
return 8;
}
else if (Rs >= 167) {
return 4;
}
}
else if (Cs >= 260) {
return 8;
}
}
}
}
else if (ssDCval >= 42) {
if (SC < 487) {
if (ssDCval < 47) {
if (Cs < 214) {
return 1;
}
else if (Cs >= 214) {
if (Rs < 176) {
return 1;
}
else if (Rs >= 176) {
return 2;
}
}
}
else if (ssDCval >= 47) {
if (Rs < 176) {
return 1;
}
else if (Rs >= 176) {
return 4;
}
}
}
else if (SC >= 487) {
if (ssDCval < 48) {
return 4;
}
else if (ssDCval >= 48) {
if (refDCval < 49) {
return 1;
}
else if (refDCval >= 49) {
if (Rs < 239) {
return 4;
}
else if (Rs >= 239) {
return 2;
}
}
}
}
}
}
else if (AFD >= 10) {
if (refDCval < 55) {
if (SC < 232) {
return 2;
}
else if (SC >= 232) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 4;
}
}
}
else if (refDCval >= 55) {
if (Rs < 191) {
return 4;
}
else if (Rs >= 191) {
return 8;
}
}
}
}
}
else if (posBalance >= 68) {
return 2;
}
}
}
else if (mcTcor >= 97) {
if (mcTcor < 99) {
if (MVDiff < 13) {
if (mu_mv_mag_sq < 85) {
return 1;
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 144) {
if (Rs < 85) {
return 8;
}
else if (Rs >= 85) {
return 1;
}
}
else if (Rs >= 144) {
return 4;
}
}
}
else if (MVDiff >= 13) {
if (Cs < 101) {
if (gchDC < 5) {
if (Cs < 6) {
return 2;
}
else if (Cs >= 6) {
return 1;
}
}
else if (gchDC >= 5) {
if (Rs < 123) {
return 2;
}
else if (Rs >= 123) {
return 8;
}
}
}
else if (Cs >= 101) {
if (AFD < 7) {
if (Rs < 45) {
return 1;
}
else if (Rs >= 45) {
if (AFD < 6) {
if (negBalance < 38) {
return 4;
}
else if (negBalance >= 38) {
return 2;
}
}
else if (AFD >= 6) {
return 2;
}
}
}
else if (AFD >= 7) {
if (Rs < 69) {
return 4;
}
else if (Rs >= 69) {
return 1;
}
}
}
}
}
else if (mcTcor >= 99) {
if (tcor < 80) {
if (Rs < 153) {
return 2;
}
else if (Rs >= 153) {
return 1;
}
}
else if (tcor >= 80) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (SC < 504) {
if (negBalance < 56) {
if (RsCsDiff < 2) {
if (Rs < 238) {
if (refDCval < 43) {
if (posBalance < 37) {
if (Cs < 227) {
if (Rs < 114) {
if (Cs < 172) {
return 4;
}
else if (Cs >= 172) {
return 8;
}
}
else if (Rs >= 114) {
if (posBalance < 24) {
return 4;
}
else if (posBalance >= 24) {
if (SC < 367) {
return 1;
}
else if (SC >= 367) {
return 4;
}
}
}
}
else if (Cs >= 227) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
return 1;
}
}
}
else if (posBalance >= 37) {
if (MVDiff < 8) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
else if (MVDiff >= 8) {
if (Cs < 78) {
return 4;
}
else if (Cs >= 78) {
return 1;
}
}
}
}
else if (refDCval >= 43) {
if (Cs < 172) {
return 8;
}
else if (Cs >= 172) {
return 1;
}
}
}
else if (Rs >= 238) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (RsCsDiff >= 2) {
if (negBalance < 19) {
return 4;
}
else if (negBalance >= 19) {
return 1;
}
}
}
else if (negBalance >= 56) {
return 4;
}
}
else if (SC >= 504) {
if (mu_mv_mag_sq < 93) {
return 4;
}
else if (mu_mv_mag_sq >= 93) {
return 1;
}
}
}
}
}
else if (ssDCval >= 50) {
if (refDCval < 70) {
if (TSCindex < 2) {
if (Rs < 113) {
if (posBalance < 42) {
if (refDCval < 56) {
if (posBalance < 34) {
if (Rs < 16) {
return 2;
}
else if (Rs >= 16) {
return 8;
}
}
else if (posBalance >= 34) {
return 4;
}
}
else if (refDCval >= 56) {
return 2;
}
}
else if (posBalance >= 42) {
if (Rs < 64) {
return 1;
}
else if (Rs >= 64) {
return 4;
}
}
}
else if (Rs >= 113) {
if (refDCval < 65) {
if (mu_mv_mag_sq < 104) {
return 4;
}
else if (mu_mv_mag_sq >= 104) {
if (mu_mv_mag_sq < 126) {
if (posBalance < 39) {
if (Rs < 145) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (Rs >= 145) {
return 1;
}
}
else if (posBalance >= 39) {
if (ssDCval < 59) {
if (Rs < 139) {
return 4;
}
else if (Rs >= 139) {
if (gchDC < 1) {
if (Cs < 192) {
if (Rs < 146) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Rs >= 146) {
return 1;
}
}
else if (Cs >= 192) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
return 4;
}
}
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (ssDCval >= 59) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 126) {
return 4;
}
}
}
else if (refDCval >= 65) {
if (mu_mv_mag_sq < 116) {
if (posBalance < 27) {
return 1;
}
else if (posBalance >= 27) {
return 2;
}
}
else if (mu_mv_mag_sq >= 116) {
if (Cs < 378) {
if (Cs < 204) {
return 1;
}
else if (Cs >= 204) {
return 4;
}
}
else if (Cs >= 378) {
return 1;
}
}
}
}
}
else if (TSCindex >= 2) {
if (RsCsDiff < 1) {
if (SC < 338) {
if (tcor < 100) {
if (mcTcor < 95) {
if (negBalance < 50) {
if (Cs < 87) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 8;
}
}
else if (Cs >= 87) {
return 8;
}
}
else if (negBalance >= 50) {
if (MVDiff < 52) {
if (Rs < 101) {
return 1;
}
else if (Rs >= 101) {
return 8;
}
}
else if (MVDiff >= 52) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (MVDiff < 4) {
if (refDCval < 63) {
if (Rs < 81) {
return 2;
}
else if (Rs >= 81) {
if (AFD < 2) {
if (Rs < 141) {
if (Rs < 121) {
return 1;
}
else if (Rs >= 121) {
return 4;
}
}
else if (Rs >= 141) {
return 1;
}
}
else if (AFD >= 2) {
if (mu_mv_mag_sq < 65) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 65) {
return 1;
}
}
}
}
else if (refDCval >= 63) {
if (Rs < 110) {
return 1;
}
else if (Rs >= 110) {
return 2;
}
}
}
else if (MVDiff >= 4) {
if (diffAFD < -2) {
if (Rs < 103) {
if (Rs < 91) {
return 2;
}
else if (Rs >= 91) {
return 1;
}
}
else if (Rs >= 103) {
return 8;
}
}
else if (diffAFD >= -2) {
if (Rs < 64) {
if (MVDiff < 60) {
return 1;
}
else if (MVDiff >= 60) {
return 4;
}
}
else if (Rs >= 64) {
if (MVDiff < 20) {
if (Cs < 220) {
if (tcor < 94) {
if (mu_mv_mag_sq < 44) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 44) {
return 4;
}
}
else if (tcor >= 94) {
if (Cs < 193) {
if (negBalance < 21) {
return 4;
}
else if (negBalance >= 21) {
if (posBalance < 38) {
if (Rs < 141) {
return 1;
}
else if (Rs >= 141) {
return 4;
}
}
else if (posBalance >= 38) {
return 1;
}
}
}
else if (Cs >= 193) {
return 4;
}
}
}
else if (Cs >= 220) {
return 4;
}
}
else if (MVDiff >= 20) {
return 4;
}
}
}
}
}
}
else if (tcor >= 100) {
if (Cs < 17) {
return 2;
}
else if (Cs >= 17) {
return 8;
}
}
}
else if (SC >= 338) {
if (negBalance < 28) {
if (ssDCval < 60) {
if (MVDiff < 1) {
if (Rs < 172) {
return 4;
}
else if (Rs >= 172) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (ssDCval < 59) {
if (Rs < 258) {
return 4;
}
else if (Rs >= 258) {
return 1;
}
}
else if (ssDCval >= 59) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 68) {
return 1;
}
else if (mu_mv_mag_sq >= 68) {
return 4;
}
}
}
else if (ssDCval >= 60) {
if (mu_mv_mag_sq < 100) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 2;
}
}
else if (mu_mv_mag_sq >= 100) {
if (Cs < 338) {
if (SC < 390) {
if (MVDiff < 6) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (Rs < 199) {
return 1;
}
else if (Rs >= 199) {
return 4;
}
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (SC >= 390) {
if (mu_mv_mag_sq < 127) {
return 4;
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 209) {
return 4;
}
else if (Rs >= 209) {
return 1;
}
}
}
}
else if (Cs >= 338) {
if (ssDCval < 69) {
if (Rs < 185) {
return 1;
}
else if (Rs >= 185) {
return 4;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
}
else if (negBalance >= 28) {
if (mu_mv_mag_sq < 87) {
if (negBalance < 39) {
if (ssDCval < 64) {
if (ssDCval < 57) {
if (Rs < 180) {
return 4;
}
else if (Rs >= 180) {
return 1;
}
}
else if (ssDCval >= 57) {
if (Cs < 215) {
if (Cs < 205) {
return 4;
}
else if (Cs >= 205) {
return 1;
}
}
else if (Cs >= 215) {
return 4;
}
}
}
else if (ssDCval >= 64) {
if (MVDiff < 2) {
if (Cs < 306) {
return 4;
}
else if (Cs >= 306) {
if (Rs < 170) {
return 1;
}
else if (Rs >= 170) {
return 4;
}
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (negBalance >= 39) {
if (Rs < 183) {
return 2;
}
else if (Rs >= 183) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (ssDCval < 67) {
if (posBalance < 48) {
if (mu_mv_mag_sq < 123) {
if (negBalance < 33) {
if (Cs < 435) {
if (Cs < 196) {
return 1;
}
else if (Cs >= 196) {
return 4;
}
}
else if (Cs >= 435) {
if (Rs < 186) {
return 4;
}
else if (Rs >= 186) {
return 1;
}
}
}
else if (negBalance >= 33) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (Cs < 503) {
return 1;
}
else if (Cs >= 503) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
return 1;
}
}
else if (posBalance >= 48) {
if (Cs < 491) {
if (Rs < 287) {
return 1;
}
else if (Rs >= 287) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 292) {
return 4;
}
else if (Rs >= 292) {
return 2;
}
}
}
}
else if (Cs >= 491) {
return 4;
}
}
}
else if (ssDCval >= 67) {
if (Cs < 218) {
return 2;
}
else if (Cs >= 218) {
if (Rs < 142) {
return 1;
}
else if (Rs >= 142) {
if (Rs < 222) {
return 4;
}
else if (Rs >= 222) {
if (Cs < 407) {
return 1;
}
else if (Cs >= 407) {
return 4;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (TSC < 4) {
if (mcTcor < 91) {
if (negBalance < 21) {
if (Rs < 92) {
return 8;
}
else if (Rs >= 92) {
return 1;
}
}
else if (negBalance >= 21) {
return 8;
}
}
else if (mcTcor >= 91) {
if (MVDiff < 3) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (posBalance < 54) {
if (posBalance < 45) {
if (Rs < 298) {
return 1;
}
else if (Rs >= 298) {
return 4;
}
}
else if (posBalance >= 45) {
if (tcor < 97) {
if (Rs < 225) {
if (negBalance < 33) {
if (posBalance < 50) {
if (Cs < 279) {
return 4;
}
else if (Cs >= 279) {
return 1;
}
}
else if (posBalance >= 50) {
return 4;
}
}
else if (negBalance >= 33) {
return 4;
}
}
else if (Rs >= 225) {
if (Cs < 343) {
return 1;
}
else if (Cs >= 343) {
return 4;
}
}
}
else if (tcor >= 97) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 99) {
if (Rs < 128) {
return 4;
}
else if (Rs >= 128) {
return 1;
}
}
else if (mu_mv_mag_sq >= 99) {
if (Rs < 240) {
return 4;
}
else if (Rs >= 240) {
return 1;
}
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 37) {
return 1;
}
else if (Rs >= 37) {
if (RsCsDiff < 5) {
if (Cs < 326) {
if (TSCindex < 3) {
if (Cs < 185) {
return 1;
}
else if (Cs >= 185) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (RsCsDiff < 2) {
return 4;
}
else if (RsCsDiff >= 2) {
return 1;
}
}
}
}
else if (TSCindex >= 3) {
if (SC < 521) {
if (mu_mv_mag_sq < 5) {
if (mcTcor < 94) {
if (Rs < 97) {
return 1;
}
else if (Rs >= 97) {
return 2;
}
}
else if (mcTcor >= 94) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (negBalance < 46) {
if (negBalance < 33) {
if (Rs < 165) {
return 4;
}
else if (Rs >= 165) {
return 1;
}
}
else if (negBalance >= 33) {
return 4;
}
}
else if (negBalance >= 46) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (MVDiff < 21) {
if (mu_mv_mag_sq < 82) {
if (Cs < 45) {
return 1;
}
else if (Cs >= 45) {
return 4;
}
}
else if (mu_mv_mag_sq >= 82) {
if (Rs < 212) {
if (Rs < 186) {
return 4;
}
else if (Rs >= 186) {
return 1;
}
}
else if (Rs >= 212) {
return 4;
}
}
}
else if (MVDiff >= 21) {
if (Cs < 274) {
if (diffMVdiffVal < -10) {
if (Rs < 110) {
return 1;
}
else if (Rs >= 110) {
return 8;
}
}
else if (diffMVdiffVal >= -10) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
else if (Cs >= 274) {
return 2;
}
}
}
}
else if (SC >= 521) {
if (ssDCval < 63) {
if (Rs < 278) {
return 8;
}
else if (Rs >= 278) {
return 2;
}
}
else if (ssDCval >= 63) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
return 4;
}
}
}
}
}
else if (Cs >= 326) {
if (negBalance < 41) {
if (MVDiff < 42) {
if (negBalance < 39) {
return 4;
}
else if (negBalance >= 39) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (MVDiff >= 42) {
if (Rs < 241) {
return 1;
}
else if (Rs >= 241) {
return 4;
}
}
}
else if (negBalance >= 41) {
if (posBalance < 52) {
if (refDCval < 68) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 4;
}
}
else if (refDCval >= 68) {
return 8;
}
}
else if (posBalance >= 52) {
return 8;
}
}
}
}
else if (RsCsDiff >= 5) {
if (AFD < 7) {
if (Rs < 250) {
return 4;
}
else if (Rs >= 250) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (CsDiff < 3) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
return 4;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
}
}
else if (AFD >= 7) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 1;
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (refDCval < 66) {
if (RsCsDiff < 19) {
if (CsDiff < 2) {
if (Cs < 94) {
if (gchDC < 6) {
return 1;
}
else if (gchDC >= 6) {
if (Cs < 68) {
if (Rs < 130) {
return 1;
}
else if (Rs >= 130) {
return 4;
}
}
else if (Cs >= 68) {
if (diffAFD < 9) {
return 8;
}
else if (diffAFD >= 9) {
return 1;
}
}
}
}
else if (Cs >= 94) {
if (mu_mv_mag_sq < 47) {
if (MVDiff < 22) {
if (Rs < 209) {
return 1;
}
else if (Rs >= 209) {
if (Cs < 298) {
if (SC < 482) {
return 1;
}
else if (SC >= 482) {
return 2;
}
}
else if (Cs >= 298) {
return 8;
}
}
}
else if (MVDiff >= 22) {
if (AFD < 12) {
if (RsDiff < 1) {
if (Rs < 150) {
return 1;
}
else if (Rs >= 150) {
return 8;
}
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (AFD >= 12) {
if (Cs < 205) {
if (refDCval < 64) {
if (negBalance < 33) {
if (Rs < 87) {
return 1;
}
else if (Rs >= 87) {
return 4;
}
}
else if (negBalance >= 33) {
return 1;
}
}
else if (refDCval >= 64) {
return 2;
}
}
else if (Cs >= 205) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (AFD < 12) {
if (Cs < 264) {
return 4;
}
else if (Cs >= 264) {
return 1;
}
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (CsDiff >= 2) {
if (mcTcor < 86) {
if (Cs < 423) {
if (refDCval < 48) {
return 2;
}
else if (refDCval >= 48) {
return 4;
}
}
else if (Cs >= 423) {
return 1;
}
}
else if (mcTcor >= 86) {
if (Rs < 269) {
return 1;
}
else if (Rs >= 269) {
if (Cs < 282) {
if (Cs < 218) {
return 4;
}
else if (Cs >= 218) {
return 2;
}
}
else if (Cs >= 282) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 19) {
if (SC < 320) {
if (Rs < 167) {
return 1;
}
else if (Rs >= 167) {
return 2;
}
}
else if (SC >= 320) {
return 8;
}
}
}
else if (refDCval >= 66) {
if (CsDiff < 1) {
if (Rs < 95) {
return 8;
}
else if (Rs >= 95) {
return 1;
}
}
else if (CsDiff >= 1) {
if (TSC < 5) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
if (Cs < 281) {
if (Cs < 119) {
return 4;
}
else if (Cs >= 119) {
return 8;
}
}
else if (Cs >= 281) {
if (Rs < 161) {
return 2;
}
else if (Rs >= 161) {
return 4;
}
}
}
}
else if (TSC >= 5) {
if (AFD < 20) {
if (diffMVdiffVal < -14) {
return 4;
}
else if (diffMVdiffVal >= -14) {
if (SC < 601) {
if (RsDiff < 3) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 4;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (SC >= 601) {
return 4;
}
}
}
else if (AFD >= 20) {
if (Cs < 95) {
return 4;
}
else if (Cs >= 95) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 109) {
if (TSCindex < 2) {
if (ssDCval < 81) {
if (mu_mv_mag_sq < 90) {
if (Rs < 129) {
if (Rs < 98) {
return 1;
}
else if (Rs >= 98) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Cs < 102) {
return 2;
}
else if (Cs >= 102) {
return 1;
}
}
}
}
else if (Rs >= 129) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 90) {
return 2;
}
}
else if (ssDCval >= 81) {
if (Cs < 159) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
if (Rs < 101) {
if (Rs < 5) {
return 1;
}
else if (Rs >= 5) {
return 4;
}
}
else if (Rs >= 101) {
return 1;
}
}
}
else if (Cs >= 159) {
if (MVDiff < 42) {
return 4;
}
else if (MVDiff >= 42) {
return 2;
}
}
}
}
else if (TSCindex >= 2) {
if (mcTcor < 97) {
if (Rs < 118) {
if (mu_mv_mag_sq < 72) {
if (SC < 39) {
if (tcor < 38) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (MVDiff < 593) {
return 2;
}
else if (MVDiff >= 593) {
return 1;
}
}
}
else if (tcor >= 38) {
if (diffRsCsdiff < -1) {
if (Rs < 3) {
return 2;
}
else if (Rs >= 3) {
if (Rs < 8) {
return 1;
}
else if (Rs >= 8) {
return 4;
}
}
}
else if (diffRsCsdiff >= -1) {
if (ssDCval < 160) {
return 1;
}
else if (ssDCval >= 160) {
if (MVDiff < 101) {
return 8;
}
else if (MVDiff >= 101) {
return 1;
}
}
}
}
}
else if (SC >= 39) {
if (ssDCval < 123) {
if (negBalance < 67) {
if (negBalance < 43) {
if (Rs < 7) {
return 2;
}
else if (Rs >= 7) {
if (gchDC < 6) {
return 1;
}
else if (gchDC >= 6) {
return 8;
}
}
}
else if (negBalance >= 43) {
if (Cs < 255) {
if (gchDC < 1) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
if (Rs < 76) {
if (Rs < 52) {
return 2;
}
else if (Rs >= 52) {
return 1;
}
}
else if (Rs >= 76) {
return 2;
}
}
}
else if (gchDC >= 1) {
if (mcTcor < 86) {
if (Rs < 54) {
return 4;
}
else if (Rs >= 54) {
if (Rs < 110) {
return 8;
}
else if (Rs >= 110) {
return 2;
}
}
}
else if (mcTcor >= 86) {
if (MVDiff < 33) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
if (Rs < 81) {
return 1;
}
else if (Rs >= 81) {
return 4;
}
}
}
else if (MVDiff >= 33) {
if (RsCsDiff < 5) {
if (SC < 170) {
if (mu_mv_mag_sq < 24) {
return 1;
}
else if (mu_mv_mag_sq >= 24) {
return 2;
}
}
else if (SC >= 170) {
return 2;
}
}
else if (RsCsDiff >= 5) {
return 1;
}
}
}
}
}
else if (Cs >= 255) {
if (Cs < 384) {
return 2;
}
else if (Cs >= 384) {
return 8;
}
}
}
}
else if (negBalance >= 67) {
if (mu_mv_mag_sq < 4) {
return 1;
}
else if (mu_mv_mag_sq >= 4) {
return 8;
}
}
}
else if (ssDCval >= 123) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (refDCval < 91) {
return 8;
}
else if (refDCval >= 91) {
return 4;
}
}
}
else if (Rs >= 118) {
if (RsCsDiff < 3) {
if (mcTcor < 90) {
if (TSC < 5) {
if (Cs < 361) {
if (Rs < 210) {
if (posBalance < 57) {
if (Rs < 130) {
return 4;
}
else if (Rs >= 130) {
if (diffMVdiffVal < 17) {
return 2;
}
else if (diffMVdiffVal >= 17) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (mcTcor < 79) {
return 2;
}
else if (mcTcor >= 79) {
if (diffMVdiffVal < -15) {
return 2;
}
else if (diffMVdiffVal >= -15) {
if (RsCsDiff < 1) {
if (Rs < 191) {
return 2;
}
else if (Rs >= 191) {
return 8;
}
}
else if (RsCsDiff >= 1) {
return 8;
}
}
}
}
}
else if (Rs >= 210) {
if (Cs < 328) {
return 2;
}
else if (Cs >= 328) {
if (Cs < 335) {
return 8;
}
else if (Cs >= 335) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 2;
}
}
}
}
}
else if (Cs >= 361) {
return 8;
}
}
else if (TSC >= 5) {
if (mcTcor < 80) {
return 4;
}
else if (mcTcor >= 80) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (ssDCval < 110) {
if (MVDiff < 5) {
if (diffMVdiffVal < -6) {
return 2;
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
else if (MVDiff >= 5) {
if (negBalance < 39) {
if (Cs < 323) {
return 1;
}
else if (Cs >= 323) {
return 2;
}
}
else if (negBalance >= 39) {
if (mu_mv_mag_sq < 6) {
if (tcor < 78) {
if (tcor < 57) {
return 8;
}
else if (tcor >= 57) {
return 2;
}
}
else if (tcor >= 78) {
if (Rs < 128) {
return 4;
}
else if (Rs >= 128) {
if (Rs < 241) {
return 1;
}
else if (Rs >= 241) {
if (Rs < 298) {
return 2;
}
else if (Rs >= 298) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 6) {
if (Rs < 183) {
if (mcTcor < 95) {
if (Rs < 175) {
if (MVDiff < 33) {
return 8;
}
else if (MVDiff >= 33) {
if (diffMVdiffVal < 19) {
return 2;
}
else if (diffMVdiffVal >= 19) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (Rs >= 175) {
return 1;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
else if (Rs >= 183) {
if (Cs < 56) {
return 1;
}
else if (Cs >= 56) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 110) {
if (posBalance < 58) {
if (tcor < 86) {
return 2;
}
else if (tcor >= 86) {
if (Cs < 440) {
return 8;
}
else if (Cs >= 440) {
return 2;
}
}
}
else if (posBalance >= 58) {
return 2;
}
}
}
}
else if (RsCsDiff >= 3) {
if (MVDiff < 57) {
if (AFD < 15) {
if (AFD < 10) {
if (Cs < 314) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
return 1;
}
}
else if (Cs >= 314) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 230) {
return 4;
}
else if (Rs >= 230) {
return 2;
}
}
}
}
else if (AFD >= 10) {
if (TSC < 7) {
if (negBalance < 58) {
if (Cs < 12) {
if (Rs < 149) {
return 2;
}
else if (Rs >= 149) {
return 8;
}
}
else if (Cs >= 12) {
return 8;
}
}
else if (negBalance >= 58) {
if (Rs < 274) {
return 1;
}
else if (Rs >= 274) {
return 8;
}
}
}
else if (TSC >= 7) {
if (Cs < 327) {
return 4;
}
else if (Cs >= 327) {
return 2;
}
}
}
}
else if (AFD >= 15) {
if (MVDiff < 43) {
if (Cs < 233) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (Cs >= 233) {
return 1;
}
}
else if (MVDiff >= 43) {
if (gchDC < 6) {
return 2;
}
else if (gchDC >= 6) {
return 8;
}
}
}
}
else if (MVDiff >= 57) {
if (SC < 265) {
if (AFD < 33) {
return 4;
}
else if (AFD >= 33) {
if (AFD < 50) {
return 8;
}
else if (AFD >= 50) {
return 1;
}
}
}
else if (SC >= 265) {
if (SC < 402) {
if (diffTSC < -1) {
if (diffTSC < -3) {
return 1;
}
else if (diffTSC >= -3) {
if (SC < 305) {
if (Rs < 262) {
return 1;
}
else if (Rs >= 262) {
return 4;
}
}
else if (SC >= 305) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (mcTcor < 83) {
return 2;
}
else if (mcTcor >= 83) {
if (Cs < 49) {
return 2;
}
else if (Cs >= 49) {
return 1;
}
}
}
}
else if (SC >= 402) {
if (diffRsCsdiff < 8) {
if (Rs < 303) {
if (Rs < 284) {
if (Rs < 255) {
if (ssDCval < 87) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 4;
}
}
else if (ssDCval >= 87) {
if (Cs < 299) {
return 1;
}
else if (Cs >= 299) {
return 2;
}
}
}
else if (Rs >= 255) {
if (MVDiff < 181) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
return 1;
}
}
else if (MVDiff >= 181) {
return 1;
}
}
}
else if (Rs >= 284) {
return 2;
}
}
else if (Rs >= 303) {
return 4;
}
}
else if (diffRsCsdiff >= 8) {
if (gchDC < 5) {
return 8;
}
else if (gchDC >= 5) {
if (Rs < 224) {
return 1;
}
else if (Rs >= 224) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 106) {
if (Cs < 246) {
if (Rs < 82) {
if (ssDCval < 94) {
if (TSC < 3) {
if (Rs < 74) {
return 1;
}
else if (Rs >= 74) {
if (ssDCval < 72) {
if (Rs < 77) {
return 4;
}
else if (Rs >= 77) {
return 2;
}
}
else if (ssDCval >= 72) {
return 1;
}
}
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 7) {
return 4;
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 94) {
return 8;
}
else if (Cs >= 94) {
return 1;
}
}
}
}
else if (ssDCval >= 94) {
if (MVDiff < 20) {
if (Cs < 80) {
if (Cs < 24) {
return 1;
}
else if (Cs >= 24) {
return 2;
}
}
else if (Cs >= 80) {
return 4;
}
}
else if (MVDiff >= 20) {
if (refDCval < 95) {
if (Rs < 32) {
return 2;
}
else if (Rs >= 32) {
return 4;
}
}
else if (refDCval >= 95) {
if (ssDCval < 104) {
return 1;
}
else if (ssDCval >= 104) {
return 2;
}
}
}
}
}
else if (Rs >= 82) {
if (TSC < 2) {
if (ssDCval < 98) {
if (Rs < 152) {
if (SC < 267) {
if (ssDCval < 75) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
return 2;
}
}
else if (ssDCval >= 75) {
if (Cs < 99) {
return 2;
}
else if (Cs >= 99) {
return 1;
}
}
}
else if (SC >= 267) {
if (Cs < 201) {
return 1;
}
else if (Cs >= 201) {
return 4;
}
}
}
else if (Rs >= 152) {
if (Cs < 129) {
return 8;
}
else if (Cs >= 129) {
if (Rs < 247) {
return 2;
}
else if (Rs >= 247) {
if (Rs < 255) {
return 4;
}
else if (Rs >= 255) {
return 2;
}
}
}
}
}
else if (ssDCval >= 98) {
return 1;
}
}
else if (TSC >= 2) {
if (CsDiff < 2) {
if (posBalance < 57) {
if (ssDCval < 91) {
if (Rs < 136) {
if (mu_mv_mag_sq < 69) {
if (MVDiff < 33) {
if (Rs < 90) {
if (Rs < 84) {
return 1;
}
else if (Rs >= 84) {
return 4;
}
}
else if (Rs >= 90) {
return 1;
}
}
else if (MVDiff >= 33) {
if (Rs < 113) {
return 2;
}
else if (Rs >= 113) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (Cs < 201) {
return 2;
}
else if (Cs >= 201) {
if (Rs < 130) {
return 1;
}
else if (Rs >= 130) {
return 4;
}
}
}
}
else if (Rs >= 136) {
if (Cs < 147) {
if (AFD < 5) {
if (RsDiff < 1) {
if (Rs < 143) {
return 1;
}
else if (Rs >= 143) {
return 8;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (AFD >= 5) {
if (SC < 264) {
if (Rs < 173) {
return 8;
}
else if (Rs >= 173) {
if (Rs < 230) {
return 1;
}
else if (Rs >= 230) {
return 2;
}
}
}
else if (SC >= 264) {
return 1;
}
}
}
else if (Cs >= 147) {
if (SC < 375) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (SC >= 375) {
if (Cs < 187) {
if (SC < 465) {
if (MVDiff < 26) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (MVDiff >= 26) {
return 8;
}
}
else if (SC >= 465) {
if (Rs < 297) {
return 8;
}
else if (Rs >= 297) {
return 4;
}
}
}
else if (Cs >= 187) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
if (Rs < 238) {
return 2;
}
else if (Rs >= 238) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (Rs < 186) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 2;
}
}
else if (Rs >= 186) {
if (ssDCval < 101) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
else if (ssDCval >= 101) {
return 1;
}
}
}
}
else if (posBalance >= 57) {
if (MVDiff < 14) {
if (ssDCval < 93) {
if (ssDCval < 82) {
return 2;
}
else if (ssDCval >= 82) {
return 1;
}
}
else if (ssDCval >= 93) {
return 2;
}
}
else if (MVDiff >= 14) {
if (Cs < 186) {
if (ssDCval < 103) {
if (mcTcor < 100) {
return 1;
}
else if (mcTcor >= 100) {
return 8;
}
}
else if (ssDCval >= 103) {
return 2;
}
}
else if (Cs >= 186) {
if (Cs < 214) {
return 2;
}
else if (Cs >= 214) {
return 4;
}
}
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 71) {
if (Cs < 224) {
if (MVDiff < 79) {
return 2;
}
else if (MVDiff >= 79) {
return 8;
}
}
else if (Cs >= 224) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 71) {
return 8;
}
}
}
}
}
else if (Cs >= 246) {
if (ssDCval < 76) {
if (Rs < 128) {
if (refDCval < 72) {
return 4;
}
else if (refDCval >= 72) {
if (Rs < 125) {
return 1;
}
else if (Rs >= 125) {
return 4;
}
}
}
else if (Rs >= 128) {
if (tcor < 97) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 86) {
return 1;
}
else if (mu_mv_mag_sq >= 86) {
return 4;
}
}
else if (MVDiff >= 5) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (tcor >= 97) {
return 1;
}
}
}
else if (ssDCval >= 76) {
if (SC < 487) {
if (gchDC < 2) {
if (Rs < 182) {
if (posBalance < 51) {
if (Rs < 93) {
return 1;
}
else if (Rs >= 93) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
else if (posBalance >= 51) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (Rs >= 182) {
return 2;
}
}
else if (gchDC >= 2) {
if (Rs < 85) {
return 2;
}
else if (Rs >= 85) {
if (SC < 389) {
return 8;
}
else if (SC >= 389) {
return 1;
}
}
}
}
else if (SC >= 487) {
if (TSC < 3) {
if (Cs < 366) {
if (AFD < 3) {
if (Rs < 242) {
return 4;
}
else if (Rs >= 242) {
return 2;
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 366) {
return 1;
}
}
else if (TSC >= 3) {
if (RsDiff < 2) {
if (posBalance < 55) {
if (diffRsCsdiff < 1) {
if (Rs < 198) {
return 4;
}
else if (Rs >= 198) {
return 1;
}
}
else if (diffRsCsdiff >= 1) {
return 2;
}
}
else if (posBalance >= 55) {
return 2;
}
}
else if (RsDiff >= 2) {
if (TSC < 6) {
if (Cs < 440) {
return 1;
}
else if (Cs >= 440) {
return 2;
}
}
else if (TSC >= 6) {
if (Rs < 256) {
return 8;
}
else if (Rs >= 256) {
if (Rs < 277) {
return 4;
}
else if (Rs >= 277) {
return 1;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 106) {
if (AFD < 3) {
if (negBalance < 24) {
if (ssDCval < 136) {
return 4;
}
else if (ssDCval >= 136) {
return 2;
}
}
else if (negBalance >= 24) {
if (Rs < 17) {
return 4;
}
else if (Rs >= 17) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 34) {
return 4;
}
else if (mu_mv_mag_sq >= 34) {
return 2;
}
}
}
}
}
else if (AFD >= 3) {
if (MVDiff < 75) {
if (posBalance < 15) {
return 8;
}
else if (posBalance >= 15) {
if (ssDCval < 113) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
if (tcor < 97) {
if (posBalance < 69) {
return 1;
}
else if (posBalance >= 69) {
return 2;
}
}
else if (tcor >= 97) {
return 2;
}
}
}
else if (ssDCval >= 113) {
if (Rs < 288) {
if (tcor < 86) {
if (Cs < 130) {
return 2;
}
else if (Cs >= 130) {
return 1;
}
}
else if (tcor >= 86) {
if (TSCindex < 3) {
if (SC < 283) {
return 1;
}
else if (SC >= 283) {
return 2;
}
}
else if (TSCindex >= 3) {
return 2;
}
}
}
else if (Rs >= 288) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
return 8;
}
}
}
}
}
else if (MVDiff >= 75) {
if (refDCval < 120) {
if (diffRsCsdiff < -1) {
return 1;
}
else if (diffRsCsdiff >= -1) {
if (MVDiff < 135) {
if (Rs < 33) {
return 2;
}
else if (Rs >= 33) {
return 8;
}
}
else if (MVDiff >= 135) {
return 4;
}
}
}
else if (refDCval >= 120) {
if (AFD < 6) {
if (Rs < 107) {
return 1;
}
else if (Rs >= 107) {
return 2;
}
}
else if (AFD >= 6) {
if (refDCval < 121) {
return 2;
}
else if (refDCval >= 121) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 131) {
if (refDCval < 84) {
if (Rs < 86) {
return 4;
}
else if (Rs >= 86) {
if (Cs < 123) {
return 8;
}
else if (Cs >= 123) {
if (Cs < 289) {
return 2;
}
else if (Cs >= 289) {
return 4;
}
}
}
}
else if (refDCval >= 84) {
if (refDCval < 101) {
return 1;
}
else if (refDCval >= 101) {
if (refDCval < 115) {
return 4;
}
else if (refDCval >= 115) {
if (Cs < 167) {
return 1;
}
else if (Cs >= 167) {
return 8;
}
}
}
}
}
else if (Rs >= 131) {
if (Cs < 214) {
if (ssDCval < 80) {
if (mcTcor < 99) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 115) {
return 2;
}
else if (mu_mv_mag_sq >= 115) {
if (refDCval < 74) {
return 2;
}
else if (refDCval >= 74) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
}
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (mcTcor >= 99) {
if (MVDiff < 5) {
if (Cs < 210) {
if (Cs < 122) {
if (Rs < 154) {
return 2;
}
else if (Rs >= 154) {
return 8;
}
}
else if (Cs >= 122) {
return 2;
}
}
else if (Cs >= 210) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
else if (MVDiff >= 5) {
return 1;
}
}
}
else if (ssDCval >= 80) {
if (refDCval < 87) {
if (Rs < 262) {
return 8;
}
else if (Rs >= 262) {
return 1;
}
}
else if (refDCval >= 87) {
if (SC < 287) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 2;
}
}
else if (SC >= 287) {
return 2;
}
}
}
}
else if (Cs >= 214) {
if (negBalance < 22) {
if (Cs < 266) {
if (Rs < 199) {
return 1;
}
else if (Rs >= 199) {
return 2;
}
}
else if (Cs >= 266) {
if (Rs < 174) {
return 4;
}
else if (Rs >= 174) {
return 1;
}
}
}
else if (negBalance >= 22) {
if (ssDCval < 78) {
if (negBalance < 35) {
if (negBalance < 25) {
if (Cs < 249) {
if (ssDCval < 74) {
if (Rs < 248) {
return 1;
}
else if (Rs >= 248) {
return 2;
}
}
else if (ssDCval >= 74) {
return 2;
}
}
else if (Cs >= 249) {
return 1;
}
}
else if (negBalance >= 25) {
return 1;
}
}
else if (negBalance >= 35) {
return 8;
}
}
else if (ssDCval >= 78) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 307) {
if (RsDiff < 4) {
if (ssDCval < 75) {
if (mcTcor < 89) {
if (tcor < 73) {
if (posBalance < 54) {
if (ssDCval < 40) {
if (Cs < 383) {
return 1;
}
else if (Cs >= 383) {
return 8;
}
}
else if (ssDCval >= 40) {
if (negBalance < 38) {
if (Cs < 458) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (Cs >= 458) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 2;
}
}
}
else if (negBalance >= 38) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (Rs < 716) {
if (Cs < 141) {
return 1;
}
else if (Cs >= 141) {
return 2;
}
}
else if (Rs >= 716) {
return 8;
}
}
}
}
}
else if (posBalance >= 54) {
if (MVDiff < 67) {
if (AFD < 9) {
if (negBalance < 44) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
return 2;
}
}
}
else if (negBalance >= 44) {
return 8;
}
}
else if (AFD >= 9) {
if (negBalance < 45) {
if (Cs < 501) {
if (ssDCval < 38) {
return 2;
}
else if (ssDCval >= 38) {
if (mcTcor < 67) {
return 4;
}
else if (mcTcor >= 67) {
return 8;
}
}
}
else if (Cs >= 501) {
return 1;
}
}
else if (negBalance >= 45) {
if (MVDiff < 52) {
if (mu_mv_mag_sq < 26) {
if (diffRsCsdiff < -13) {
return 2;
}
else if (diffRsCsdiff >= -13) {
if (ssDCval < 53) {
if (Rs < 618) {
if (diffMVdiffVal < -7) {
return 4;
}
else if (diffMVdiffVal >= -7) {
if (Rs < 357) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
return 4;
}
}
else if (Rs >= 357) {
return 8;
}
}
}
else if (Rs >= 618) {
return 4;
}
}
else if (ssDCval >= 53) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (Rs < 673) {
return 4;
}
else if (Rs >= 673) {
return 2;
}
}
}
else if (MVDiff >= 52) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 515) {
return 1;
}
else if (Rs >= 515) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 67) {
if (mu_mv_mag_sq < 40) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
if (Cs < 473) {
return 8;
}
else if (Cs >= 473) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 40) {
return 4;
}
}
}
}
else if (tcor >= 73) {
if (Cs < 156) {
if (RsDiff < 2) {
if (Rs < 450) {
return 4;
}
else if (Rs >= 450) {
return 8;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Cs >= 156) {
if (ssDCval < 41) {
if (negBalance < 27) {
if (Cs < 377) {
return 8;
}
else if (Cs >= 377) {
return 1;
}
}
else if (negBalance >= 27) {
return 8;
}
}
else if (ssDCval >= 41) {
if (mu_mv_mag_sq < 111) {
if (gchDC < 2) {
if (posBalance < 47) {
if (Rs < 478) {
return 8;
}
else if (Rs >= 478) {
return 2;
}
}
else if (posBalance >= 47) {
if (RsCsDiff < 7) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 4;
}
}
else if (RsCsDiff >= 7) {
if (negBalance < 51) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (Rs < 1008) {
if (diffRsCsdiff < 9) {
return 4;
}
else if (diffRsCsdiff >= 9) {
return 8;
}
}
else if (Rs >= 1008) {
return 8;
}
}
}
}
else if (negBalance >= 51) {
if (Rs < 682) {
return 4;
}
else if (Rs >= 682) {
return 1;
}
}
}
}
}
else if (gchDC >= 2) {
if (diffRsCsdiff < 4) {
if (Rs < 493) {
return 1;
}
else if (Rs >= 493) {
return 4;
}
}
else if (diffRsCsdiff >= 4) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 111) {
return 8;
}
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1295) {
if (tcor < 101) {
if (TSC < 1) {
if (mcTcor < 99) {
return 4;
}
else if (mcTcor >= 99) {
if (SC < 1079) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 4;
}
}
else if (SC >= 1079) {
if (Cs < 394) {
if (Rs < 789) {
return 8;
}
else if (Rs >= 789) {
return 1;
}
}
else if (Cs >= 394) {
return 8;
}
}
}
}
else if (TSC >= 1) {
if (MVDiff < 2) {
if (ssDCval < 69) {
if (mcTcor < 97) {
if (Rs < 473) {
if (tcor < 90) {
if (Cs < 357) {
return 4;
}
else if (Cs >= 357) {
return 1;
}
}
else if (tcor >= 90) {
if (mu_mv_mag_sq < 115) {
if (Cs < 513) {
if (Cs < 237) {
return 8;
}
else if (Cs >= 237) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (Cs >= 513) {
return 8;
}
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
}
else if (Rs >= 473) {
if (Cs < 528) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 1;
}
}
}
}
}
else if (Cs >= 528) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 468) {
if (ssDCval < 66) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 266) {
return 8;
}
else if (Cs >= 266) {
return 4;
}
}
}
else if (ssDCval >= 66) {
if (posBalance < 53) {
if (Rs < 398) {
return 4;
}
else if (Rs >= 398) {
return 1;
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
else if (Rs >= 468) {
if (refDCval < 56) {
if (ssDCval < 47) {
if (Rs < 531) {
return 8;
}
else if (Rs >= 531) {
return 4;
}
}
else if (ssDCval >= 47) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
return 4;
}
}
}
else if (refDCval >= 56) {
if (Rs < 906) {
if (Rs < 861) {
if (ssDCval < 66) {
if (Cs < 350) {
if (posBalance < 44) {
if (Rs < 591) {
return 8;
}
else if (Rs >= 591) {
return 1;
}
}
else if (posBalance >= 44) {
if (Rs < 728) {
return 1;
}
else if (Rs >= 728) {
return 2;
}
}
}
else if (Cs >= 350) {
if (SC < 1003) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
if (mu_mv_mag_sq < 92) {
if (Cs < 416) {
return 8;
}
else if (Cs >= 416) {
return 1;
}
}
else if (mu_mv_mag_sq >= 92) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 4;
}
}
}
}
else if (SC >= 1003) {
if (SC < 1112) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (Rs < 629) {
if (Rs < 564) {
return 8;
}
else if (Rs >= 564) {
return 1;
}
}
else if (Rs >= 629) {
return 8;
}
}
}
else if (SC >= 1112) {
if (SC < 1253) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
else if (SC >= 1253) {
return 8;
}
}
}
}
}
else if (ssDCval >= 66) {
if (Cs < 394) {
if (mu_mv_mag_sq < 111) {
if (Rs < 619) {
return 8;
}
else if (Rs >= 619) {
return 1;
}
}
else if (mu_mv_mag_sq >= 111) {
return 2;
}
}
else if (Cs >= 394) {
if (posBalance < 52) {
if (diffMVdiffVal < -2) {
return 2;
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
else if (posBalance >= 52) {
return 2;
}
}
}
}
else if (Rs >= 861) {
if (tcor < 98) {
if (Cs < 476) {
return 1;
}
else if (Cs >= 476) {
return 8;
}
}
else if (tcor >= 98) {
return 2;
}
}
}
else if (Rs >= 906) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 901) {
if (Rs < 597) {
if (refDCval < 74) {
if (SC < 953) {
if (negBalance < 26) {
return 2;
}
else if (negBalance >= 26) {
if (Rs < 414) {
return 2;
}
else if (Rs >= 414) {
return 1;
}
}
}
else if (SC >= 953) {
if (Cs < 387) {
return 2;
}
else if (Cs >= 387) {
return 1;
}
}
}
else if (refDCval >= 74) {
return 8;
}
}
else if (Rs >= 597) {
if (posBalance < 44) {
if (Rs < 603) {
if (Rs < 599) {
return 4;
}
else if (Rs >= 599) {
return 1;
}
}
else if (Rs >= 603) {
return 4;
}
}
else if (posBalance >= 44) {
if (mcTcor < 99) {
if (ssDCval < 72) {
if (Cs < 336) {
return 4;
}
else if (Cs >= 336) {
return 1;
}
}
else if (ssDCval >= 72) {
if (Rs < 686) {
return 4;
}
else if (Rs >= 686) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 4;
}
}
}
}
}
else if (Rs >= 901) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (SC < 1495) {
return 1;
}
else if (SC >= 1495) {
if (Rs < 1243) {
return 8;
}
else if (Rs >= 1243) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 138) {
if (mu_mv_mag_sq < 49) {
if (Cs < 95) {
return 1;
}
else if (Cs >= 95) {
if (ssDCval < 69) {
if (Rs < 887) {
if (Cs < 133) {
if (posBalance < 48) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (Rs < 622) {
return 8;
}
else if (Rs >= 622) {
return 1;
}
}
}
else if (posBalance >= 48) {
return 1;
}
}
else if (Cs >= 133) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (Cs < 136) {
return 4;
}
else if (Cs >= 136) {
return 1;
}
}
}
}
else if (Rs >= 887) {
if (mcTcor < 96) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
}
else if (ssDCval >= 69) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (Rs < 500) {
return 8;
}
else if (Rs >= 500) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (AFD < 3) {
if (Rs < 572) {
return 1;
}
else if (Rs >= 572) {
return 4;
}
}
else if (AFD >= 3) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Cs < 83) {
if (Rs < 405) {
return 1;
}
else if (Rs >= 405) {
return 8;
}
}
else if (Cs >= 83) {
if (Rs < 432) {
return 8;
}
else if (Rs >= 432) {
return 1;
}
}
}
}
}
}
else if (Cs >= 138) {
if (refDCval < 72) {
if (refDCval < 58) {
if (Cs < 428) {
if (posBalance < 34) {
if (mu_mv_mag_sq < 82) {
if (AFD < 4) {
if (Cs < 312) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (Cs >= 312) {
if (Rs < 444) {
return 1;
}
else if (Rs >= 444) {
return 8;
}
}
}
else if (AFD >= 4) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 82) {
return 1;
}
}
else if (posBalance >= 34) {
if (SC < 1056) {
if (refDCval < 45) {
if (RsCsDiff < 1) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (RsCsDiff >= 1) {
if (MVDiff < 28) {
if (gchDC < 1) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (diffTSC < 1) {
if (RsCsDiff < 2) {
if (Rs < 406) {
return 4;
}
else if (Rs >= 406) {
return 8;
}
}
else if (RsCsDiff >= 2) {
if (diffMVdiffVal < -6) {
return 4;
}
else if (diffMVdiffVal >= -6) {
return 1;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 387) {
return 1;
}
else if (Rs >= 387) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 61) {
if (AFD < 6) {
if (Cs < 336) {
if (Rs < 846) {
return 4;
}
else if (Rs >= 846) {
return 1;
}
}
else if (Cs >= 336) {
return 1;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 61) {
return 1;
}
}
}
else if (MVDiff >= 28) {
if (refDCval < 38) {
if (Rs < 654) {
return 1;
}
else if (Rs >= 654) {
return 4;
}
}
else if (refDCval >= 38) {
if (Rs < 377) {
return 2;
}
else if (Rs >= 377) {
return 8;
}
}
}
}
}
else if (refDCval >= 45) {
if (MVDiff < 4) {
if (mcTcor < 95) {
if (Rs < 498) {
return 8;
}
else if (Rs >= 498) {
if (Rs < 654) {
return 4;
}
else if (Rs >= 654) {
return 2;
}
}
}
else if (mcTcor >= 95) {
if (Cs < 394) {
return 1;
}
else if (Cs >= 394) {
if (Rs < 426) {
return 1;
}
else if (Rs >= 426) {
if (Rs < 523) {
return 8;
}
else if (Rs >= 523) {
return 4;
}
}
}
}
}
else if (MVDiff >= 4) {
if (SC < 910) {
if (tcor < 90) {
if (posBalance < 56) {
if (posBalance < 47) {
if (ssDCval < 51) {
if (Rs < 461) {
return 4;
}
else if (Rs >= 461) {
return 1;
}
}
else if (ssDCval >= 51) {
return 2;
}
}
else if (posBalance >= 47) {
return 4;
}
}
else if (posBalance >= 56) {
if (Cs < 333) {
if (mu_mv_mag_sq < 47) {
if (Rs < 483) {
return 4;
}
else if (Rs >= 483) {
if (Cs < 286) {
return 1;
}
else if (Cs >= 286) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (Cs >= 333) {
return 2;
}
}
}
else if (tcor >= 90) {
if (AFD < 3) {
if (diffTSC < -1) {
if (Rs < 470) {
return 8;
}
else if (Rs >= 470) {
return 1;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (AFD >= 3) {
if (diffAFD < -1) {
if (diffMVdiffVal < -21) {
return 4;
}
else if (diffMVdiffVal >= -21) {
if (Cs < 203) {
return 1;
}
else if (Cs >= 203) {
if (mu_mv_mag_sq < 58) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (Cs < 264) {
return 1;
}
else if (Cs >= 264) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 58) {
return 1;
}
}
}
}
else if (diffAFD >= -1) {
if (CsDiff < 1) {
if (Rs < 413) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 1;
}
}
else if (Rs >= 413) {
return 1;
}
}
else if (CsDiff >= 1) {
if (tcor < 94) {
if (Rs < 393) {
return 4;
}
else if (Rs >= 393) {
if (refDCval < 55) {
return 1;
}
else if (refDCval >= 55) {
if (Rs < 406) {
return 1;
}
else if (Rs >= 406) {
return 4;
}
}
}
}
else if (tcor >= 94) {
if (Rs < 574) {
return 4;
}
else if (Rs >= 574) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 910) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (MVDiff < 6) {
if (Rs < 648) {
return 4;
}
else if (Rs >= 648) {
return 1;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
}
}
}
}
else if (SC >= 1056) {
if (negBalance < 48) {
if (diffRsCsdiff < 7) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (Rs < 713) {
if (Rs < 701) {
return 1;
}
else if (Rs >= 701) {
return 2;
}
}
else if (Rs >= 713) {
return 1;
}
}
}
else if (diffRsCsdiff >= 7) {
return 8;
}
}
else if (negBalance >= 48) {
if (Rs < 815) {
return 8;
}
else if (Rs >= 815) {
return 1;
}
}
}
}
}
else if (Cs >= 428) {
if (negBalance < 37) {
if (Rs < 627) {
if (RsCsDiff < 2) {
if (Cs < 456) {
return 8;
}
else if (Cs >= 456) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (RsCsDiff >= 2) {
if (posBalance < 43) {
if (MVDiff < 9) {
if (Rs < 567) {
if (Rs < 345) {
return 2;
}
else if (Rs >= 345) {
return 8;
}
}
else if (Rs >= 567) {
return 4;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (posBalance >= 43) {
if (SC < 986) {
if (ssDCval < 58) {
if (Rs < 362) {
return 4;
}
else if (Rs >= 362) {
if (TSC < 3) {
if (Cs < 454) {
return 1;
}
else if (Cs >= 454) {
return 8;
}
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (ssDCval >= 58) {
return 4;
}
}
else if (SC >= 986) {
if (diffMVdiffVal < 3) {
if (SC < 1064) {
return 4;
}
else if (SC >= 1064) {
return 1;
}
}
else if (diffMVdiffVal >= 3) {
if (Cs < 442) {
return 4;
}
else if (Cs >= 442) {
return 1;
}
}
}
}
}
}
else if (Rs >= 627) {
if (TSC < 3) {
if (RsDiff < 2) {
if (posBalance < 34) {
if (Rs < 754) {
return 2;
}
else if (Rs >= 754) {
return 8;
}
}
else if (posBalance >= 34) {
return 1;
}
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (TSC >= 3) {
if (MVDiff < 15) {
if (Rs < 933) {
return 8;
}
else if (Rs >= 933) {
if (Rs < 1067) {
return 4;
}
else if (Rs >= 1067) {
return 2;
}
}
}
else if (MVDiff >= 15) {
if (SC < 1239) {
return 1;
}
else if (SC >= 1239) {
return 2;
}
}
}
}
}
else if (negBalance >= 37) {
if (mu_mv_mag_sq < 26) {
if (Cs < 496) {
if (MVDiff < 59) {
return 1;
}
else if (MVDiff >= 59) {
if (Rs < 586) {
return 2;
}
else if (Rs >= 586) {
return 8;
}
}
}
else if (Cs >= 496) {
if (Cs < 512) {
return 8;
}
else if (Cs >= 512) {
if (Rs < 542) {
return 2;
}
else if (Rs >= 542) {
if (Rs < 800) {
return 8;
}
else if (Rs >= 800) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (Cs < 529) {
if (refDCval < 49) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (Rs < 773) {
return 4;
}
else if (Rs >= 773) {
return 2;
}
}
}
else if (refDCval >= 49) {
return 8;
}
}
else if (Cs >= 529) {
return 4;
}
}
}
}
}
else if (refDCval >= 58) {
if (mu_mv_mag_sq < 25) {
if (mcTcor < 96) {
if (refDCval < 62) {
if (posBalance < 59) {
if (Cs < 202) {
return 1;
}
else if (Cs >= 202) {
return 2;
}
}
else if (posBalance >= 59) {
if (SC < 759) {
return 2;
}
else if (SC >= 759) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
}
else if (refDCval >= 62) {
if (Rs < 374) {
return 1;
}
else if (Rs >= 374) {
if (MVDiff < 27) {
return 4;
}
else if (MVDiff >= 27) {
if (diffAFD < 1) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (diffAFD >= 1) {
if (MVDiff < 59) {
return 4;
}
else if (MVDiff >= 59) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (TSC < 4) {
if (Rs < 852) {
if (posBalance < 50) {
if (negBalance < 44) {
if (MVDiff < 3) {
if (Rs < 490) {
return 1;
}
else if (Rs >= 490) {
return 4;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (negBalance >= 44) {
return 1;
}
}
else if (posBalance >= 50) {
if (ssDCval < 64) {
if (tcor < 94) {
return 2;
}
else if (tcor >= 94) {
return 4;
}
}
else if (ssDCval >= 64) {
return 4;
}
}
}
else if (Rs >= 852) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
else if (TSC >= 4) {
if (RsCsDiff < 4) {
if (Rs < 393) {
return 2;
}
else if (Rs >= 393) {
return 4;
}
}
else if (RsCsDiff >= 4) {
if (RsCsDiff < 7) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (RsCsDiff >= 7) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (tcor < 99) {
if (posBalance < 43) {
if (mcTcor < 98) {
if (gchDC < 1) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 843) {
return 1;
}
else if (Rs >= 843) {
return 4;
}
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (mcTcor >= 98) {
if (mcTcor < 100) {
if (Rs < 672) {
if (ssDCval < 62) {
if (Rs < 460) {
return 4;
}
else if (Rs >= 460) {
return 8;
}
}
else if (ssDCval >= 62) {
if (Cs < 354) {
return 8;
}
else if (Cs >= 354) {
return 1;
}
}
}
else if (Rs >= 672) {
return 1;
}
}
else if (mcTcor >= 100) {
if (Rs < 1047) {
return 1;
}
else if (Rs >= 1047) {
return 2;
}
}
}
}
else if (posBalance >= 43) {
if (SC < 1050) {
if (negBalance < 28) {
if (mcTcor < 99) {
if (Rs < 401) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
if (Rs < 335) {
return 4;
}
else if (Rs >= 335) {
return 8;
}
}
}
else if (Rs >= 401) {
return 1;
}
}
else if (mcTcor >= 99) {
if (Cs < 290) {
if (Rs < 585) {
return 4;
}
else if (Rs >= 585) {
return 1;
}
}
else if (Cs >= 290) {
if (Rs < 568) {
return 8;
}
else if (Rs >= 568) {
return 4;
}
}
}
}
else if (negBalance >= 28) {
if (MVDiff < 4) {
if (Cs < 451) {
if (posBalance < 50) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
if (Rs < 491) {
return 4;
}
else if (Rs >= 491) {
return 1;
}
}
}
else if (posBalance >= 50) {
return 1;
}
}
else if (Cs >= 451) {
return 8;
}
}
else if (MVDiff >= 4) {
if (MVDiff < 11) {
if (mcTcor < 97) {
if (TSC < 6) {
if (ssDCval < 70) {
if (MVDiff < 5) {
if (Rs < 406) {
return 2;
}
else if (Rs >= 406) {
return 4;
}
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (mcTcor >= 97) {
if (ssDCval < 62) {
if (posBalance < 47) {
if (Rs < 346) {
return 8;
}
else if (Rs >= 346) {
return 1;
}
}
else if (posBalance >= 47) {
if (mu_mv_mag_sq < 90) {
return 2;
}
else if (mu_mv_mag_sq >= 90) {
if (Cs < 389) {
return 1;
}
else if (Cs >= 389) {
return 2;
}
}
}
}
else if (ssDCval >= 62) {
if (TSC < 3) {
if (Rs < 436) {
return 4;
}
else if (Rs >= 436) {
return 1;
}
}
else if (TSC >= 3) {
if (SC < 824) {
return 1;
}
else if (SC >= 824) {
return 4;
}
}
}
}
}
else if (MVDiff >= 11) {
if (diffMVdiffVal < -20) {
return 2;
}
else if (diffMVdiffVal >= -20) {
if (negBalance < 45) {
if (SC < 643) {
if (Rs < 321) {
return 4;
}
else if (Rs >= 321) {
return 8;
}
}
else if (SC >= 643) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (Rs < 402) {
return 1;
}
else if (Rs >= 402) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 588) {
return 4;
}
else if (Rs >= 588) {
return 1;
}
}
}
}
}
}
}
}
else if (SC >= 1050) {
if (negBalance < 28) {
if (posBalance < 47) {
if (Rs < 787) {
return 8;
}
else if (Rs >= 787) {
return 4;
}
}
else if (posBalance >= 47) {
if (ssDCval < 67) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 2;
}
}
else if (ssDCval >= 67) {
return 2;
}
}
}
else if (negBalance >= 28) {
if (RsCsDiff < 7) {
if (Rs < 979) {
if (Cs < 423) {
return 8;
}
else if (Cs >= 423) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
else if (Rs >= 979) {
return 4;
}
}
else if (RsCsDiff >= 7) {
if (diffRsCsdiff < -1) {
if (SC < 1359) {
return 1;
}
else if (SC >= 1359) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 99) {
if (Rs < 420) {
if (Rs < 341) {
if (Rs < 322) {
return 4;
}
else if (Rs >= 322) {
return 2;
}
}
else if (Rs >= 341) {
return 4;
}
}
else if (Rs >= 420) {
if (mu_mv_mag_sq < 108) {
if (Cs < 371) {
return 8;
}
else if (Cs >= 371) {
if (Rs < 630) {
return 1;
}
else if (Rs >= 630) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 874) {
return 2;
}
else if (Rs >= 874) {
if (Cs < 381) {
return 1;
}
else if (Cs >= 381) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 72) {
if (tcor < 98) {
if (ssDCval < 73) {
if (mu_mv_mag_sq < 81) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 501) {
return 2;
}
else if (Rs >= 501) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 81) {
return 8;
}
}
else if (ssDCval >= 73) {
if (Rs < 495) {
if (Rs < 373) {
return 2;
}
else if (Rs >= 373) {
return 8;
}
}
else if (Rs >= 495) {
if (SC < 1247) {
return 4;
}
else if (SC >= 1247) {
if (Rs < 771) {
return 4;
}
else if (Rs >= 771) {
return 8;
}
}
}
}
}
else if (tcor >= 98) {
if (Rs < 875) {
return 8;
}
else if (Rs >= 875) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 101) {
if (RsDiff < 1) {
if (mcTcor < 103) {
if (Rs < 793) {
if (Rs < 404) {
return 4;
}
else if (Rs >= 404) {
return 8;
}
}
else if (Rs >= 793) {
return 4;
}
}
else if (mcTcor >= 103) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (RsDiff >= 1) {
if (MVDiff < 21) {
if (mu_mv_mag_sq < 72) {
if (Cs < 86) {
return 1;
}
else if (Cs >= 86) {
return 8;
}
}
else if (mu_mv_mag_sq >= 72) {
if (gchDC < 2) {
if (SC < 698) {
return 1;
}
else if (SC >= 698) {
return 2;
}
}
else if (gchDC >= 2) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
else if (MVDiff >= 21) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1295) {
if (SC < 1587) {
if (mu_mv_mag_sq < 59) {
return 1;
}
else if (mu_mv_mag_sq >= 59) {
return 8;
}
}
else if (SC >= 1587) {
if (mu_mv_mag_sq < 126) {
if (Rs < 1759) {
if (negBalance < 20) {
if (mu_mv_mag_sq < 90) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (negBalance < 18) {
return 2;
}
else if (negBalance >= 18) {
if (Rs < 1394) {
return 2;
}
else if (Rs >= 1394) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (ssDCval < 64) {
if (ssDCval < 58) {
if (Rs < 1503) {
return 1;
}
else if (Rs >= 1503) {
return 2;
}
}
else if (ssDCval >= 58) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (negBalance >= 20) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (RsDiff < 3) {
if (mu_mv_mag_sq < 124) {
if (TSC < 2) {
if (Rs < 1326) {
return 8;
}
else if (Rs >= 1326) {
return 1;
}
}
else if (TSC >= 2) {
if (diffAFD < -1) {
if (Rs < 1322) {
return 8;
}
else if (Rs >= 1322) {
return 1;
}
}
else if (diffAFD >= -1) {
if (mu_mv_mag_sq < 75) {
return 8;
}
else if (mu_mv_mag_sq >= 75) {
if (refDCval < 71) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
if (Cs < 466) {
return 8;
}
else if (Cs >= 466) {
if (Rs < 1401) {
return 8;
}
else if (Rs >= 1401) {
return 1;
}
}
}
}
else if (refDCval >= 71) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (posBalance < 26) {
return 2;
}
else if (posBalance >= 26) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 1460) {
return 2;
}
else if (Rs >= 1460) {
return 1;
}
}
}
}
}
else if (Rs >= 1759) {
if (Cs < 430) {
if (RsCsDiff < 10) {
return 2;
}
else if (RsCsDiff >= 10) {
if (ssDCval < 67) {
return 2;
}
else if (ssDCval >= 67) {
return 1;
}
}
}
else if (Cs >= 430) {
if (negBalance < 29) {
if (SC < 2821) {
return 4;
}
else if (SC >= 2821) {
return 2;
}
}
else if (negBalance >= 29) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (ssDCval < 62) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
return 8;
}
}
else if (ssDCval >= 62) {
if (Cs < 493) {
if (ssDCval < 65) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 1459) {
return 2;
}
else if (Rs >= 1459) {
return 1;
}
}
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (Cs >= 493) {
if (refDCval < 64) {
if (negBalance < 22) {
return 2;
}
else if (negBalance >= 22) {
return 8;
}
}
else if (refDCval >= 64) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (Cs < 173) {
if (ssDCval < 107) {
if (RsCsDiff < 2) {
if (SC < 1368) {
if (ssDCval < 101) {
if (RsDiff < 1) {
if (Cs < 168) {
return 8;
}
else if (Cs >= 168) {
if (Rs < 763) {
return 2;
}
else if (Rs >= 763) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (Cs < 78) {
return 1;
}
else if (Cs >= 78) {
return 8;
}
}
}
else if (ssDCval >= 101) {
if (Rs < 963) {
return 1;
}
else if (Rs >= 963) {
return 8;
}
}
}
else if (SC >= 1368) {
if (ssDCval < 95) {
return 1;
}
else if (ssDCval >= 95) {
return 8;
}
}
}
else if (RsCsDiff >= 2) {
if (tcor < 65) {
if (Rs < 326) {
if (Rs < 320) {
return 4;
}
else if (Rs >= 320) {
return 8;
}
}
else if (Rs >= 326) {
if (Rs < 358) {
return 2;
}
else if (Rs >= 358) {
return 8;
}
}
}
else if (tcor >= 65) {
if (ssDCval < 78) {
if (Cs < 146) {
return 1;
}
else if (Cs >= 146) {
return 4;
}
}
else if (ssDCval >= 78) {
if (CsDiff < 2) {
if (negBalance < 79) {
return 1;
}
else if (negBalance >= 79) {
if (Rs < 833) {
return 8;
}
else if (Rs >= 833) {
return 1;
}
}
}
else if (CsDiff >= 2) {
return 8;
}
}
}
}
}
else if (ssDCval >= 107) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (mcTcor < 97) {
if (AFD < 18) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (AFD >= 18) {
return 2;
}
}
else if (mcTcor >= 97) {
if (Rs < 1431) {
return 8;
}
else if (Rs >= 1431) {
return 1;
}
}
}
}
}
else if (Cs >= 173) {
if (SC < 853) {
if (mcTcor < 97) {
if (MVDiff < 20) {
if (Cs < 486) {
if (Rs < 563) {
if (diffAFD < -2) {
if (Cs < 235) {
return 8;
}
else if (Cs >= 235) {
return 1;
}
}
else if (diffAFD >= -2) {
if (diffAFD < 3) {
if (Cs < 401) {
if (ssDCval < 88) {
if (Cs < 351) {
if (SC < 689) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
return 2;
}
}
else if (SC >= 689) {
if (refDCval < 88) {
return 1;
}
else if (refDCval >= 88) {
return 2;
}
}
}
else if (Cs >= 351) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (ssDCval >= 88) {
if (negBalance < 64) {
if (Cs < 391) {
if (diffMVdiffVal < -9) {
return 2;
}
else if (diffMVdiffVal >= -9) {
return 8;
}
}
else if (Cs >= 391) {
return 2;
}
}
else if (negBalance >= 64) {
return 2;
}
}
}
else if (Cs >= 401) {
return 8;
}
}
else if (diffAFD >= 3) {
return 1;
}
}
}
else if (Rs >= 563) {
return 4;
}
}
else if (Cs >= 486) {
return 2;
}
}
else if (MVDiff >= 20) {
if (mu_mv_mag_sq < 27) {
if (diffMVdiffVal < 49) {
if (diffRsCsdiff < -2) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Cs < 377) {
if (Rs < 314) {
return 1;
}
else if (Rs >= 314) {
if (Rs < 510) {
return 2;
}
else if (Rs >= 510) {
return 1;
}
}
}
else if (Cs >= 377) {
if (Rs < 379) {
return 4;
}
else if (Rs >= 379) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= -2) {
if (Rs < 472) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
if (Rs < 340) {
return 1;
}
else if (Rs >= 340) {
return 2;
}
}
}
else if (Rs >= 472) {
return 1;
}
}
}
else if (diffMVdiffVal >= 49) {
if (Rs < 332) {
return 8;
}
else if (Rs >= 332) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (ssDCval < 119) {
if (tcor < 69) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
return 8;
}
}
else if (tcor >= 69) {
if (Cs < 377) {
return 8;
}
else if (Cs >= 377) {
if (Rs < 335) {
return 2;
}
else if (Rs >= 335) {
return 8;
}
}
}
}
else if (ssDCval >= 119) {
if (Rs < 333) {
return 8;
}
else if (Rs >= 333) {
return 2;
}
}
}
}
}
else if (mcTcor >= 97) {
if (tcor < 88) {
if (ssDCval < 131) {
return 4;
}
else if (ssDCval >= 131) {
if (Cs < 345) {
return 2;
}
else if (Cs >= 345) {
return 8;
}
}
}
else if (tcor >= 88) {
if (MVDiff < 28) {
if (ssDCval < 78) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
else if (ssDCval >= 78) {
if (ssDCval < 152) {
if (mu_mv_mag_sq < 43) {
if (ssDCval < 86) {
return 4;
}
else if (ssDCval >= 86) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (ssDCval < 122) {
if (ssDCval < 81) {
if (Rs < 405) {
return 2;
}
else if (Rs >= 405) {
return 1;
}
}
else if (ssDCval >= 81) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
if (Rs < 451) {
if (ssDCval < 91) {
return 1;
}
else if (ssDCval >= 91) {
if (posBalance < 34) {
if (Rs < 396) {
return 1;
}
else if (Rs >= 396) {
return 8;
}
}
else if (posBalance >= 34) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
}
}
else if (Rs >= 451) {
return 1;
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (ssDCval >= 152) {
if (mcTcor < 98) {
if (Rs < 372) {
return 2;
}
else if (Rs >= 372) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (mcTcor >= 98) {
return 4;
}
}
}
}
else if (MVDiff >= 28) {
if (mu_mv_mag_sq < 14) {
if (tcor < 95) {
if (MVDiff < 98) {
return 2;
}
else if (MVDiff >= 98) {
if (Rs < 345) {
return 2;
}
else if (Rs >= 345) {
return 4;
}
}
}
else if (tcor >= 95) {
return 4;
}
}
else if (mu_mv_mag_sq >= 14) {
return 8;
}
}
}
}
}
else if (SC >= 853) {
if (SC < 1886) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1385) {
if (tcor < 89) {
if (ssDCval < 89) {
if (refDCval < 81) {
if (MVDiff < 26) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (refDCval >= 81) {
if (Cs < 432) {
if (Rs < 525) {
return 8;
}
else if (Rs >= 525) {
return 1;
}
}
else if (Cs >= 432) {
if (tcor < 87) {
return 2;
}
else if (tcor >= 87) {
if (Rs < 649) {
return 1;
}
else if (Rs >= 649) {
return 2;
}
}
}
}
}
else if (ssDCval >= 89) {
if (Cs < 464) {
return 8;
}
else if (Cs >= 464) {
if (posBalance < 60) {
if (SC < 862) {
return 8;
}
else if (SC >= 862) {
if (Rs < 562) {
return 2;
}
else if (Rs >= 562) {
return 8;
}
}
}
else if (posBalance >= 60) {
if (ssDCval < 115) {
return 8;
}
else if (ssDCval >= 115) {
return 4;
}
}
}
}
}
else if (tcor >= 89) {
if (Cs < 385) {
if (ssDCval < 134) {
if (Cs < 283) {
if (Cs < 215) {
if (Cs < 203) {
return 8;
}
else if (Cs >= 203) {
return 1;
}
}
else if (Cs >= 215) {
return 4;
}
}
else if (Cs >= 283) {
if (ssDCval < 95) {
return 1;
}
else if (ssDCval >= 95) {
if (Rs < 830) {
return 8;
}
else if (Rs >= 830) {
return 1;
}
}
}
}
else if (ssDCval >= 134) {
return 4;
}
}
else if (Cs >= 385) {
if (Cs < 490) {
if (negBalance < 57) {
return 1;
}
else if (negBalance >= 57) {
if (Cs < 412) {
return 8;
}
else if (Cs >= 412) {
return 1;
}
}
}
else if (Cs >= 490) {
if (posBalance < 50) {
if (Rs < 409) {
return 8;
}
else if (Rs >= 409) {
return 2;
}
}
else if (posBalance >= 50) {
if (AFD < 10) {
if (mcTcor < 100) {
return 1;
}
else if (mcTcor >= 100) {
if (Rs < 485) {
return 1;
}
else if (Rs >= 485) {
return 2;
}
}
}
else if (AFD >= 10) {
if (Rs < 586) {
return 2;
}
else if (Rs >= 586) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1385) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (ssDCval < 115) {
if (Cs < 284) {
if (mu_mv_mag_sq < 120) {
if (mcTcor < 98) {
if (Rs < 776) {
return 4;
}
else if (Rs >= 776) {
return 1;
}
}
else if (mcTcor >= 98) {
if (Cs < 218) {
return 8;
}
else if (Cs >= 218) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (Cs < 239) {
return 1;
}
else if (Cs >= 239) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 983) {
return 1;
}
else if (Rs >= 983) {
if (Rs < 1001) {
return 8;
}
else if (Rs >= 1001) {
if (Rs < 1470) {
return 2;
}
else if (Rs >= 1470) {
return 8;
}
}
}
}
}
else if (Cs >= 284) {
if (refDCval < 96) {
if (tcor < 98) {
if (mcTcor < 95) {
if (AFD < 14) {
if (TSC < 4) {
if (Rs < 717) {
return 2;
}
else if (Rs >= 717) {
return 8;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (AFD >= 14) {
if (Rs < 425) {
return 1;
}
else if (Rs >= 425) {
return 2;
}
}
}
else if (mcTcor >= 95) {
if (Rs < 1113) {
if (mu_mv_mag_sq < 70) {
if (refDCval < 87) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (refDCval >= 87) {
if (Cs < 521) {
if (SC < 882) {
return 1;
}
else if (SC >= 882) {
return 8;
}
}
else if (Cs >= 521) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (negBalance < 41) {
if (AFD < 4) {
if (Rs < 990) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (negBalance < 28) {
return 1;
}
else if (negBalance >= 28) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 1;
}
}
}
}
}
else if (Rs >= 990) {
return 2;
}
}
else if (AFD >= 4) {
if (Rs < 777) {
return 1;
}
else if (Rs >= 777) {
return 8;
}
}
}
else if (negBalance >= 41) {
if (refDCval < 77) {
return 2;
}
else if (refDCval >= 77) {
return 8;
}
}
}
}
else if (Rs >= 1113) {
return 1;
}
}
}
else if (tcor >= 98) {
if (negBalance < 31) {
if (ssDCval < 94) {
if (Rs < 576) {
return 1;
}
else if (Rs >= 576) {
if (Cs < 350) {
return 8;
}
else if (Cs >= 350) {
if (Rs < 754) {
return 1;
}
else if (Rs >= 754) {
return 8;
}
}
}
}
else if (ssDCval >= 94) {
if (Rs < 863) {
return 8;
}
else if (Rs >= 863) {
return 1;
}
}
}
else if (negBalance >= 31) {
if (Rs < 540) {
if (Rs < 446) {
return 8;
}
else if (Rs >= 446) {
return 1;
}
}
else if (Rs >= 540) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Cs < 449) {
return 8;
}
else if (Cs >= 449) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 96) {
if (ssDCval < 98) {
if (MVDiff < 4) {
if (Cs < 473) {
if (Rs < 574) {
return 8;
}
else if (Rs >= 574) {
return 4;
}
}
else if (Cs >= 473) {
return 8;
}
}
else if (MVDiff >= 4) {
if (Rs < 510) {
return 2;
}
else if (Rs >= 510) {
return 8;
}
}
}
else if (ssDCval >= 98) {
if (negBalance < 54) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (negBalance >= 54) {
if (MVDiff < 8) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (MVDiff >= 8) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (Rs < 476) {
return 8;
}
else if (Rs >= 476) {
if (Rs < 565) {
return 2;
}
else if (Rs >= 565) {
return 1;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 115) {
if (refDCval < 118) {
if (Cs < 478) {
if (tcor < 95) {
if (Rs < 835) {
return 2;
}
else if (Rs >= 835) {
return 8;
}
}
else if (tcor >= 95) {
return 1;
}
}
else if (Cs >= 478) {
return 2;
}
}
else if (refDCval >= 118) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 50) {
if (MVDiff < 32) {
if (Rs < 429) {
if (Cs < 499) {
return 2;
}
else if (Cs >= 499) {
return 8;
}
}
else if (Rs >= 429) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (Rs < 1013) {
return 8;
}
else if (Rs >= 1013) {
return 1;
}
}
}
}
else if (MVDiff >= 32) {
if (Rs < 962) {
return 2;
}
else if (Rs >= 962) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (mu_mv_mag_sq < 79) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
return 4;
}
}
else if (mu_mv_mag_sq >= 79) {
if (Cs < 385) {
return 1;
}
else if (Cs >= 385) {
return 8;
}
}
}
}
else if (mcTcor >= 99) {
if (Rs < 383) {
if (Rs < 370) {
return 8;
}
else if (Rs >= 370) {
return 4;
}
}
else if (Rs >= 383) {
if (refDCval < 120) {
return 2;
}
else if (refDCval >= 120) {
if (gchDC < 6) {
return 1;
}
else if (gchDC >= 6) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 1886) {
if (Cs < 304) {
if (ssDCval < 104) {
return 1;
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (Cs >= 304) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 85) {
if (tcor < 97) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 1;
}
}
else if (tcor >= 97) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (refDCval < 80) {
return 1;
}
else if (refDCval >= 80) {
if (Rs < 1512) {
return 1;
}
else if (Rs >= 1512) {
if (negBalance < 26) {
if (Cs < 511) {
if (MVDiff < 5) {
if (Rs < 1561) {
return 1;
}
else if (Rs >= 1561) {
return 8;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (Cs >= 511) {
if (Rs < 2067) {
return 4;
}
else if (Rs >= 2067) {
return 8;
}
}
}
else if (negBalance >= 26) {
if (Rs < 1964) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (diffRsCsdiff < 4) {
return 4;
}
else if (diffRsCsdiff >= 4) {
if (Rs < 1694) {
return 4;
}
else if (Rs >= 1694) {
return 1;
}
}
}
}
else if (Rs >= 1964) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 55) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 2154) {
return 1;
}
else if (Rs >= 2154) {
return 2;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (mu_mv_mag_sq < 78) {
if (tcor < 74) {
if (negBalance < 46) {
if (SC < 2018) {
if (diffAFD < 1) {
if (posBalance < 60) {
if (ssDCval < 54) {
if (tcor < 68) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
if (Cs < 254) {
return 1;
}
else if (Cs >= 254) {
return 2;
}
}
}
else if (tcor >= 68) {
return 2;
}
}
else if (ssDCval >= 54) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 1;
}
}
}
else if (posBalance >= 60) {
if (ssDCval < 72) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (Rs < 1316) {
return 1;
}
else if (Rs >= 1316) {
return 2;
}
}
}
else if (ssDCval >= 72) {
return 2;
}
}
}
else if (diffAFD >= 1) {
if (Cs < 364) {
if (refDCval < 44) {
if (ssDCval < 42) {
return 1;
}
else if (ssDCval >= 42) {
if (MVDiff < 66) {
return 2;
}
else if (MVDiff >= 66) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
return 8;
}
}
}
}
else if (refDCval >= 44) {
return 1;
}
}
else if (Cs >= 364) {
if (posBalance < 75) {
if (MVDiff < 47) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (MVDiff >= 47) {
if (RsCsDiff < 51) {
return 8;
}
else if (RsCsDiff >= 51) {
if (SC < 1031) {
return 1;
}
else if (SC >= 1031) {
return 2;
}
}
}
}
else if (posBalance >= 75) {
if (MVDiff < 111) {
return 8;
}
else if (MVDiff >= 111) {
if (Rs < 375) {
return 4;
}
else if (Rs >= 375) {
return 2;
}
}
}
}
}
}
else if (SC >= 2018) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (tcor < 73) {
if (diffRsCsdiff < -24) {
if (AFD < 19) {
if (Rs < 1977) {
return 2;
}
else if (Rs >= 1977) {
return 1;
}
}
else if (AFD >= 19) {
return 4;
}
}
else if (diffRsCsdiff >= -24) {
return 1;
}
}
else if (tcor >= 73) {
return 2;
}
}
}
}
else if (negBalance >= 46) {
if (Cs < 353) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 7) {
if (MVDiff < 32) {
if (Cs < 300) {
return 2;
}
else if (Cs >= 300) {
return 8;
}
}
else if (MVDiff >= 32) {
if (Cs < 259) {
return 1;
}
else if (Cs >= 259) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (RsDiff < 20) {
if (diffMVdiffVal < 30) {
if (mu_mv_mag_sq < 31) {
if (MVDiff < 10) {
if (Rs < 1321) {
return 1;
}
else if (Rs >= 1321) {
return 2;
}
}
else if (MVDiff >= 10) {
if (Rs < 656) {
if (Rs < 603) {
return 1;
}
else if (Rs >= 603) {
return 2;
}
}
else if (Rs >= 656) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 31) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (mu_mv_mag_sq < 35) {
if (Rs < 1159) {
return 1;
}
else if (Rs >= 1159) {
return 2;
}
}
else if (mu_mv_mag_sq >= 35) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 30) {
if (negBalance < 62) {
if (Cs < 242) {
return 1;
}
else if (Cs >= 242) {
return 8;
}
}
else if (negBalance >= 62) {
return 1;
}
}
}
else if (RsDiff >= 20) {
if (Cs < 67) {
return 1;
}
else if (Cs >= 67) {
return 2;
}
}
}
}
else if (posBalance >= 54) {
if (refDCval < 131) {
if (RsCsDiff < 56) {
if (MVDiff < 32) {
if (diffAFD < -1) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (diffAFD >= -1) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (MVDiff >= 32) {
if (negBalance < 54) {
if (refDCval < 56) {
if (Cs < 302) {
return 4;
}
else if (Cs >= 302) {
return 8;
}
}
else if (refDCval >= 56) {
if (mu_mv_mag_sq < 3) {
return 8;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 516) {
if (Cs < 80) {
return 8;
}
else if (Cs >= 80) {
if (Rs < 498) {
return 2;
}
else if (Rs >= 498) {
return 8;
}
}
}
else if (Rs >= 516) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (Cs < 327) {
if (MVDiff < 69) {
if (mu_mv_mag_sq < 7) {
if (Rs < 651) {
return 4;
}
else if (Rs >= 651) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (MVDiff < 35) {
return 8;
}
else if (MVDiff >= 35) {
return 1;
}
}
}
else if (MVDiff >= 69) {
if (MVDiff < 136) {
return 2;
}
else if (MVDiff >= 136) {
return 1;
}
}
}
else if (Cs >= 327) {
return 2;
}
}
}
}
else if (RsCsDiff >= 56) {
if (ssDCval < 115) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (diffRsCsdiff < -34) {
if (MVDiff < 72) {
return 2;
}
else if (MVDiff >= 72) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 2;
}
}
}
else if (diffRsCsdiff >= -34) {
if (posBalance < 69) {
if (Rs < 373) {
return 4;
}
else if (Rs >= 373) {
if (Cs < 79) {
return 4;
}
else if (Cs >= 79) {
return 2;
}
}
}
else if (posBalance >= 69) {
if (Rs < 1468) {
return 1;
}
else if (Rs >= 1468) {
return 2;
}
}
}
}
}
else if (ssDCval >= 115) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
if (Cs < 308) {
return 1;
}
else if (Cs >= 308) {
return 2;
}
}
}
}
}
else if (refDCval >= 131) {
if (tcor < 68) {
if (ssDCval < 149) {
return 1;
}
else if (ssDCval >= 149) {
if (ssDCval < 171) {
if (Cs < 256) {
return 2;
}
else if (Cs >= 256) {
return 4;
}
}
else if (ssDCval >= 171) {
return 1;
}
}
}
else if (tcor >= 68) {
if (Rs < 1517) {
return 4;
}
else if (Rs >= 1517) {
return 8;
}
}
}
}
}
else if (Cs >= 353) {
if (AFD < 22) {
if (MVDiff < 44) {
if (tcor < 61) {
if (AFD < 18) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (TSC < 8) {
if (ssDCval < 119) {
if (mu_mv_mag_sq < 13) {
if (Rs < 854) {
return 2;
}
else if (Rs >= 854) {
return 8;
}
}
else if (mu_mv_mag_sq >= 13) {
return 4;
}
}
else if (ssDCval >= 119) {
if (Rs < 1038) {
if (Rs < 736) {
return 8;
}
else if (Rs >= 736) {
return 1;
}
}
else if (Rs >= 1038) {
return 2;
}
}
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (AFD >= 18) {
if (Rs < 541) {
return 4;
}
else if (Rs >= 541) {
return 8;
}
}
}
else if (tcor >= 61) {
if (diffRsCsdiff < 4) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
return 4;
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 1286) {
if (Rs < 628) {
if (Rs < 481) {
return 8;
}
else if (Rs >= 481) {
if (Rs < 524) {
return 2;
}
else if (Rs >= 524) {
return 1;
}
}
}
else if (Rs >= 628) {
return 8;
}
}
else if (Rs >= 1286) {
return 2;
}
}
}
}
else if (MVDiff >= 44) {
if (Cs < 403) {
if (gchDC < 2) {
if (MVDiff < 53) {
if (Rs < 1057) {
return 2;
}
else if (Rs >= 1057) {
return 1;
}
}
else if (MVDiff >= 53) {
return 4;
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (Cs >= 403) {
if (negBalance < 57) {
if (ssDCval < 124) {
if (mcTcor < 83) {
return 8;
}
else if (mcTcor >= 83) {
if (CsDiff < 4) {
if (Rs < 1415) {
if (Rs < 891) {
return 2;
}
else if (Rs >= 891) {
return 8;
}
}
else if (Rs >= 1415) {
return 2;
}
}
else if (CsDiff >= 4) {
if (Rs < 847) {
if (Rs < 771) {
if (Rs < 692) {
return 4;
}
else if (Rs >= 692) {
return 8;
}
}
else if (Rs >= 771) {
return 1;
}
}
else if (Rs >= 847) {
return 8;
}
}
}
}
else if (ssDCval >= 124) {
if (tcor < 40) {
return 1;
}
else if (tcor >= 40) {
if (Rs < 1050) {
return 8;
}
else if (Rs >= 1050) {
return 4;
}
}
}
}
else if (negBalance >= 57) {
if (Cs < 441) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 1;
}
}
else if (Cs >= 441) {
return 4;
}
}
}
}
}
else if (AFD >= 22) {
if (ssDCval < 116) {
if (mu_mv_mag_sq < 33) {
if (gchDC < 9) {
if (AFD < 24) {
if (MVDiff < 129) {
return 2;
}
else if (MVDiff >= 129) {
return 4;
}
}
else if (AFD >= 24) {
if (RsCsDiff < 273) {
if (CsDiff < 6) {
if (posBalance < 65) {
return 1;
}
else if (posBalance >= 65) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (CsDiff >= 6) {
if (Rs < 496) {
return 2;
}
else if (Rs >= 496) {
if (Rs < 667) {
return 4;
}
else if (Rs >= 667) {
return 1;
}
}
}
}
else if (RsCsDiff >= 273) {
return 2;
}
}
}
else if (gchDC >= 9) {
if (Cs < 476) {
if (Rs < 475) {
return 1;
}
else if (Rs >= 475) {
return 8;
}
}
else if (Cs >= 476) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (negBalance < 60) {
if (posBalance < 64) {
if (tcor < 66) {
return 4;
}
else if (tcor >= 66) {
return 1;
}
}
else if (posBalance >= 64) {
if (Cs < 512) {
return 8;
}
else if (Cs >= 512) {
return 4;
}
}
}
else if (negBalance >= 60) {
if (ssDCval < 97) {
if (refDCval < 69) {
return 4;
}
else if (refDCval >= 69) {
return 2;
}
}
else if (ssDCval >= 97) {
return 4;
}
}
}
}
else if (ssDCval >= 116) {
if (RsDiff < 11) {
if (Rs < 791) {
if (AFD < 31) {
return 2;
}
else if (AFD >= 31) {
return 1;
}
}
else if (Rs >= 791) {
return 8;
}
}
else if (RsDiff >= 11) {
if (Cs < 363) {
return 2;
}
else if (Cs >= 363) {
if (MVDiff < 412) {
return 1;
}
else if (MVDiff >= 412) {
return 2;
}
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (gchDC < 2) {
if (refDCval < 139) {
if (RsCsDiff < 77) {
if (Cs < 411) {
if (ssDCval < 86) {
if (posBalance < 40) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (ssDCval < 51) {
return 2;
}
else if (ssDCval >= 51) {
if (Rs < 1807) {
return 8;
}
else if (Rs >= 1807) {
return 4;
}
}
}
}
else if (posBalance >= 40) {
if (Cs < 206) {
return 1;
}
else if (Cs >= 206) {
if (SC < 1089) {
if (posBalance < 57) {
if (posBalance < 48) {
if (Cs < 265) {
return 1;
}
else if (Cs >= 265) {
return 8;
}
}
else if (posBalance >= 48) {
return 1;
}
}
else if (posBalance >= 57) {
if (Rs < 726) {
if (Rs < 586) {
return 4;
}
else if (Rs >= 586) {
return 1;
}
}
else if (Rs >= 726) {
return 4;
}
}
}
else if (SC >= 1089) {
if (TSC < 7) {
if (Rs < 2341) {
if (posBalance < 45) {
if (diffTSC < 1) {
if (Cs < 404) {
return 1;
}
else if (Cs >= 404) {
if (Rs < 1266) {
return 2;
}
else if (Rs >= 1266) {
return 1;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 1166) {
return 1;
}
else if (Rs >= 1166) {
return 4;
}
}
}
else if (posBalance >= 45) {
return 1;
}
}
else if (Rs >= 2341) {
return 2;
}
}
else if (TSC >= 7) {
if (Rs < 1463) {
return 1;
}
else if (Rs >= 1463) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (Cs < 243) {
if (tcor < 89) {
if (Rs < 1654) {
if (ssDCval < 97) {
if (ssDCval < 89) {
if (ssDCval < 88) {
return 2;
}
else if (ssDCval >= 88) {
return 1;
}
}
else if (ssDCval >= 89) {
return 2;
}
}
else if (ssDCval >= 97) {
return 1;
}
}
else if (Rs >= 1654) {
return 1;
}
}
else if (tcor >= 89) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
if (refDCval < 86) {
return 8;
}
else if (refDCval >= 86) {
return 1;
}
}
}
}
else if (Cs >= 243) {
if (RsDiff < 5) {
if (ssDCval < 99) {
if (Rs < 618) {
return 2;
}
else if (Rs >= 618) {
return 1;
}
}
else if (ssDCval >= 99) {
if (Rs < 617) {
return 4;
}
else if (Rs >= 617) {
return 8;
}
}
}
else if (RsDiff >= 5) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 1;
}
}
}
}
}
}
else if (Cs >= 411) {
if (ssDCval < 123) {
if (negBalance < 30) {
if (Rs < 953) {
if (Rs < 671) {
if (Rs < 610) {
return 8;
}
else if (Rs >= 610) {
return 1;
}
}
else if (Rs >= 671) {
return 2;
}
}
else if (Rs >= 953) {
if (diffRsCsdiff < -10) {
if (Rs < 1928) {
return 1;
}
else if (Rs >= 1928) {
return 4;
}
}
else if (diffRsCsdiff >= -10) {
return 4;
}
}
}
else if (negBalance >= 30) {
if (Rs < 817) {
if (AFD < 19) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (Cs < 456) {
if (Rs < 598) {
return 4;
}
else if (Rs >= 598) {
return 8;
}
}
else if (Cs >= 456) {
return 4;
}
}
}
else if (AFD >= 19) {
if (Rs < 520) {
return 2;
}
else if (Rs >= 520) {
return 1;
}
}
}
else if (Rs >= 817) {
if (ssDCval < 66) {
if (MVDiff < 11) {
if (Cs < 453) {
if (Rs < 1880) {
return 1;
}
else if (Rs >= 1880) {
return 4;
}
}
else if (Cs >= 453) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (Rs < 1653) {
return 8;
}
else if (Rs >= 1653) {
return 2;
}
}
}
}
else if (MVDiff >= 11) {
if (Rs < 1230) {
if (Cs < 482) {
if (Rs < 1126) {
if (Rs < 941) {
return 2;
}
else if (Rs >= 941) {
return 1;
}
}
else if (Rs >= 1126) {
return 2;
}
}
else if (Cs >= 482) {
if (Cs < 515) {
return 4;
}
else if (Cs >= 515) {
return 8;
}
}
}
else if (Rs >= 1230) {
return 1;
}
}
}
else if (ssDCval >= 66) {
if (mcTcor < 90) {
if (Rs < 1298) {
return 4;
}
else if (Rs >= 1298) {
return 8;
}
}
else if (mcTcor >= 90) {
if (diffMVdiffVal < 5) {
if (ssDCval < 96) {
if (MVDiff < 2) {
if (Rs < 1673) {
return 4;
}
else if (Rs >= 1673) {
return 1;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 96) {
if (ssDCval < 111) {
if (Cs < 503) {
return 2;
}
else if (Cs >= 503) {
return 8;
}
}
else if (ssDCval >= 111) {
return 1;
}
}
}
else if (diffMVdiffVal >= 5) {
if (Rs < 2252) {
if (mu_mv_mag_sq < 43) {
return 1;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 1424) {
return 8;
}
else if (Rs >= 1424) {
return 2;
}
}
}
else if (Rs >= 2252) {
if (Cs < 471) {
return 4;
}
else if (Cs >= 471) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 123) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (Cs < 475) {
return 1;
}
else if (Cs >= 475) {
if (Rs < 1169) {
return 8;
}
else if (Rs >= 1169) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (refDCval < 111) {
if (CsDiff < 3) {
if (ssDCval < 95) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (Cs < 374) {
if (diffMVdiffVal < -21) {
if (Rs < 1814) {
return 1;
}
else if (Rs >= 1814) {
return 2;
}
}
else if (diffMVdiffVal >= -21) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 1;
}
}
}
else if (Cs >= 374) {
if (Rs < 1612) {
return 1;
}
else if (Rs >= 1612) {
return 2;
}
}
}
}
}
else if (ssDCval >= 95) {
if (negBalance < 55) {
return 2;
}
else if (negBalance >= 55) {
if (diffMVdiffVal < 13) {
return 8;
}
else if (diffMVdiffVal >= 13) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
if (MVDiff < 39) {
if (diffMVdiffVal < -26) {
return 4;
}
else if (diffMVdiffVal >= -26) {
if (Rs < 1577) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 2;
}
}
else if (Rs >= 1577) {
if (SC < 2838) {
if (diffRsCsdiff < -70) {
if (tcor < 79) {
return 1;
}
else if (tcor >= 79) {
return 2;
}
}
else if (diffRsCsdiff >= -70) {
return 2;
}
}
else if (SC >= 2838) {
return 1;
}
}
}
}
else if (MVDiff >= 39) {
if (mu_mv_mag_sq < 24) {
if (Rs < 2205) {
if (Rs < 957) {
return 1;
}
else if (Rs >= 957) {
return 8;
}
}
else if (Rs >= 2205) {
return 1;
}
}
else if (mu_mv_mag_sq >= 24) {
if (Cs < 464) {
return 2;
}
else if (Cs >= 464) {
return 4;
}
}
}
}
}
else if (refDCval >= 111) {
if (mcTcor < 97) {
if (negBalance < 54) {
if (Cs < 371) {
return 1;
}
else if (Cs >= 371) {
return 2;
}
}
else if (negBalance >= 54) {
if (AFD < 19) {
if (Rs < 2150) {
return 8;
}
else if (Rs >= 2150) {
return 1;
}
}
else if (AFD >= 19) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (Rs < 2257) {
return 1;
}
else if (Rs >= 2257) {
if (Rs < 2309) {
return 8;
}
else if (Rs >= 2309) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 139) {
if (CsDiff < 2) {
if (MVDiff < 119) {
return 4;
}
else if (MVDiff >= 119) {
return 1;
}
}
else if (CsDiff >= 2) {
if (Rs < 1125) {
if (ssDCval < 147) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 1;
}
}
else if (ssDCval >= 147) {
if (refDCval < 179) {
if (Rs < 666) {
return 8;
}
else if (Rs >= 666) {
return 1;
}
}
else if (refDCval >= 179) {
return 2;
}
}
}
else if (Rs >= 1125) {
if (MVDiff < 98) {
return 1;
}
else if (MVDiff >= 98) {
if (Rs < 1482) {
return 2;
}
else if (Rs >= 1482) {
return 8;
}
}
}
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 111) {
if (RsDiff < 9) {
if (Cs < 299) {
if (MVDiff < 22) {
if (Rs < 1414) {
if (diffRsCsdiff < -10) {
if (Rs < 1051) {
return 2;
}
else if (Rs >= 1051) {
return 1;
}
}
else if (diffRsCsdiff >= -10) {
return 1;
}
}
else if (Rs >= 1414) {
if (Cs < 165) {
return 2;
}
else if (Cs >= 165) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
if (Cs < 170) {
return 1;
}
else if (Cs >= 170) {
return 2;
}
}
}
}
}
else if (MVDiff >= 22) {
if (posBalance < 97) {
if (Cs < 163) {
if (SC < 809) {
if (SC < 791) {
return 1;
}
else if (SC >= 791) {
return 8;
}
}
else if (SC >= 809) {
if (Rs < 1987) {
return 1;
}
else if (Rs >= 1987) {
if (Rs < 2129) {
return 2;
}
else if (Rs >= 2129) {
return 1;
}
}
}
}
else if (Cs >= 163) {
if (diffRsCsdiff < -27) {
if (tcor < 83) {
return 8;
}
else if (tcor >= 83) {
if (Rs < 599) {
return 1;
}
else if (Rs >= 599) {
return 2;
}
}
}
else if (diffRsCsdiff >= -27) {
if (refDCval < 100) {
if (Rs < 575) {
if (Rs < 461) {
return 1;
}
else if (Rs >= 461) {
if (Rs < 526) {
return 8;
}
else if (Rs >= 526) {
return 2;
}
}
}
else if (Rs >= 575) {
return 1;
}
}
else if (refDCval >= 100) {
if (Rs < 1253) {
return 2;
}
else if (Rs >= 1253) {
if (Rs < 1318) {
return 1;
}
else if (Rs >= 1318) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 97) {
return 2;
}
}
}
else if (Cs >= 299) {
if (Rs < 797) {
if (diffMVdiffVal < -10) {
if (Cs < 373) {
return 4;
}
else if (Cs >= 373) {
return 2;
}
}
else if (diffMVdiffVal >= -10) {
if (mu_mv_mag_sq < 10) {
if (ssDCval < 91) {
if (diffMVdiffVal < -7) {
return 8;
}
else if (diffMVdiffVal >= -7) {
return 1;
}
}
else if (ssDCval >= 91) {
if (Rs < 631) {
return 8;
}
else if (Rs >= 631) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (Cs < 374) {
if (diffRsCsdiff < 7) {
return 1;
}
else if (diffRsCsdiff >= 7) {
return 8;
}
}
else if (Cs >= 374) {
if (SC < 767) {
return 1;
}
else if (SC >= 767) {
return 8;
}
}
}
}
}
else if (Rs >= 797) {
if (mu_mv_mag_sq < 23) {
if (posBalance < 77) {
if (Cs < 531) {
if (AFD < 19) {
if (tcor < 79) {
if (RsCsDiff < 38) {
if (Rs < 1219) {
return 4;
}
else if (Rs >= 1219) {
return 1;
}
}
else if (RsCsDiff >= 38) {
return 1;
}
}
else if (tcor >= 79) {
if (negBalance < 57) {
return 1;
}
else if (negBalance >= 57) {
if (Cs < 498) {
return 2;
}
else if (Cs >= 498) {
return 1;
}
}
}
}
else if (AFD >= 19) {
return 2;
}
}
else if (Cs >= 531) {
return 2;
}
}
else if (posBalance >= 77) {
if (Rs < 1326) {
return 4;
}
else if (Rs >= 1326) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (refDCval < 53) {
return 2;
}
else if (refDCval >= 53) {
if (diffMVdiffVal < 3) {
if (diffAFD < 1) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (MVDiff < 57) {
if (ssDCval < 59) {
if (Rs < 1233) {
return 8;
}
else if (Rs >= 1233) {
return 2;
}
}
else if (ssDCval >= 59) {
return 8;
}
}
else if (MVDiff >= 57) {
return 1;
}
}
}
else if (diffAFD >= 1) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (diffMVdiffVal >= 3) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (MVDiff < 37) {
if (posBalance < 73) {
if (diffRsCsdiff < 28) {
if (diffMVdiffVal < -15) {
if (Cs < 404) {
return 1;
}
else if (Cs >= 404) {
return 2;
}
}
else if (diffMVdiffVal >= -15) {
return 2;
}
}
else if (diffRsCsdiff >= 28) {
if (TSC < 8) {
if (Rs < 1905) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
else if (Rs >= 1905) {
if (RsDiff < 12) {
return 1;
}
else if (RsDiff >= 12) {
return 2;
}
}
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (posBalance >= 73) {
if (Cs < 161) {
return 2;
}
else if (Cs >= 161) {
return 1;
}
}
}
else if (MVDiff >= 37) {
if (mu_mv_mag_sq < 6) {
if (ssDCval < 98) {
if (Cs < 276) {
if (Cs < 113) {
return 1;
}
else if (Cs >= 113) {
return 2;
}
}
else if (Cs >= 276) {
if (Rs < 2131) {
return 8;
}
else if (Rs >= 2131) {
return 2;
}
}
}
else if (ssDCval >= 98) {
return 1;
}
}
else if (mu_mv_mag_sq >= 6) {
if (Rs < 987) {
if (refDCval < 99) {
return 2;
}
else if (refDCval >= 99) {
return 1;
}
}
else if (Rs >= 987) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
if (posBalance < 5) {
return 2;
}
else if (posBalance >= 5) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 111) {
if (RsDiff < 6) {
if (Rs < 902) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (MVDiff < 111) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (MVDiff >= 111) {
if (Cs < 137) {
return 1;
}
else if (Cs >= 137) {
return 2;
}
}
}
}
else if (Rs >= 902) {
if (Cs < 467) {
return 1;
}
else if (Cs >= 467) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
}
}
else if (RsDiff >= 6) {
if (mcTcor < 96) {
if (MVDiff < 94) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 907) {
return 2;
}
else if (Rs >= 907) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
if (Rs < 2303) {
return 8;
}
else if (Rs >= 2303) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 94) {
if (Rs < 1598) {
if (Rs < 862) {
return 2;
}
else if (Rs >= 862) {
if (Cs < 443) {
return 1;
}
else if (Cs >= 443) {
return 2;
}
}
}
else if (Rs >= 1598) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (ssDCval < 180) {
return 1;
}
else if (ssDCval >= 180) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 1575) {
if (AFD < 18) {
if (ssDCval < 43) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
else if (ssDCval >= 43) {
if (tcor < 66) {
if (Rs < 809) {
return 4;
}
else if (Rs >= 809) {
if (Cs < 295) {
return 1;
}
else if (Cs >= 295) {
return 4;
}
}
}
else if (tcor >= 66) {
if (MVDiff < 4) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (MVDiff >= 4) {
if (posBalance < 44) {
if (Rs < 1264) {
return 8;
}
else if (Rs >= 1264) {
return 1;
}
}
else if (posBalance >= 44) {
if (Rs < 656) {
return 8;
}
else if (Rs >= 656) {
return 2;
}
}
}
}
}
}
else if (AFD >= 18) {
if (tcor < 49) {
if (CsDiff < 3) {
if (Cs < 73) {
return 8;
}
else if (Cs >= 73) {
return 2;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (tcor >= 49) {
if (diffMVdiffVal < 61) {
if (AFD < 21) {
if (MVDiff < 53) {
return 2;
}
else if (MVDiff >= 53) {
return 1;
}
}
else if (AFD >= 21) {
if (negBalance < 51) {
if (Rs < 837) {
return 1;
}
else if (Rs >= 837) {
return 2;
}
}
else if (negBalance >= 51) {
return 2;
}
}
}
else if (diffMVdiffVal >= 61) {
return 1;
}
}
}
}
else if (Rs >= 1575) {
if (AFD < 8) {
if (Cs < 500) {
return 8;
}
else if (Cs >= 500) {
return 1;
}
}
else if (AFD >= 8) {
if (diffAFD < 10) {
return 2;
}
else if (diffAFD >= 10) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Cs >= 533) {
if (negBalance < 13) {
if (Rs < 701) {
if (ssDCval < 56) {
if (refDCval < 32) {
if (Cs < 845) {
return 8;
}
else if (Cs >= 845) {
return 2;
}
}
else if (refDCval >= 32) {
if (mu_mv_mag_sq < 127) {
if (Rs < 245) {
return 2;
}
else if (Rs >= 245) {
return 1;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 351) {
return 1;
}
else if (Rs >= 351) {
return 8;
}
}
}
}
else if (ssDCval >= 56) {
if (Rs < 450) {
if (SC < 1434) {
if (ssDCval < 157) {
return 4;
}
else if (ssDCval >= 157) {
if (Rs < 429) {
return 2;
}
else if (Rs >= 429) {
return 1;
}
}
}
else if (SC >= 1434) {
return 8;
}
}
else if (Rs >= 450) {
if (SC < 1384) {
if (Rs < 495) {
if (TSC < 2) {
return 2;
}
else if (TSC >= 2) {
return 1;
}
}
else if (Rs >= 495) {
if (Cs < 779) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (Rs < 695) {
return 8;
}
else if (Rs >= 695) {
return 1;
}
}
}
else if (Cs >= 779) {
if (Rs < 591) {
return 8;
}
else if (Rs >= 591) {
return 2;
}
}
}
}
else if (SC >= 1384) {
if (Cs < 778) {
return 1;
}
else if (Cs >= 778) {
if (Cs < 801) {
return 2;
}
else if (Cs >= 801) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 701) {
if (diffAFD < -3) {
if (Rs < 1598) {
if (mcTcor < 99) {
if (diffAFD < -16) {
if (ssDCval < 139) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (ssDCval >= 139) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
return 8;
}
}
}
else if (diffAFD >= -16) {
if (diffRsCsdiff < 4) {
return 8;
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
}
else if (mcTcor >= 99) {
if (Cs < 754) {
if (Cs < 633) {
if (Rs < 1073) {
if (Cs < 586) {
return 2;
}
else if (Cs >= 586) {
return 8;
}
}
else if (Rs >= 1073) {
return 1;
}
}
else if (Cs >= 633) {
if (Rs < 951) {
if (Rs < 855) {
return 1;
}
else if (Rs >= 855) {
return 4;
}
}
else if (Rs >= 951) {
if (posBalance < 21) {
return 8;
}
else if (posBalance >= 21) {
return 4;
}
}
}
}
else if (Cs >= 754) {
if (Cs < 1010) {
if (diffRsCsdiff < -51) {
if (diffAFD < -21) {
return 4;
}
else if (diffAFD >= -21) {
return 8;
}
}
else if (diffRsCsdiff >= -51) {
if (diffMVdiffVal < -44) {
return 8;
}
else if (diffMVdiffVal >= -44) {
if (AFD < 1) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (AFD >= 1) {
return 1;
}
}
}
}
else if (Cs >= 1010) {
if (refDCval < 128) {
return 8;
}
else if (refDCval >= 128) {
if (ssDCval < 137) {
if (SC < 2677) {
return 1;
}
else if (SC >= 2677) {
return 4;
}
}
else if (ssDCval >= 137) {
if (ssDCval < 139) {
return 4;
}
else if (ssDCval >= 139) {
if (diffAFD < -20) {
return 4;
}
else if (diffAFD >= -20) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1598) {
if (Cs < 1266) {
return 8;
}
else if (Cs >= 1266) {
if (diffMVdiffVal < -19) {
return 4;
}
else if (diffMVdiffVal >= -19) {
if (Cs < 1290) {
return 8;
}
else if (Cs >= 1290) {
return 1;
}
}
}
}
}
else if (diffAFD >= -3) {
if (Rs < 1841) {
if (Rs < 965) {
if (posBalance < 12) {
if (ssDCval < 152) {
if (Cs < 1112) {
return 8;
}
else if (Cs >= 1112) {
return 2;
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (posBalance >= 12) {
if (Cs < 1175) {
if (refDCval < 93) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
if (SC < 1474) {
if (tcor < 99) {
if (Cs < 712) {
return 8;
}
else if (Cs >= 712) {
return 2;
}
}
else if (tcor >= 99) {
return 2;
}
}
else if (SC >= 1474) {
if (Rs < 731) {
return 1;
}
else if (Rs >= 731) {
return 8;
}
}
}
}
else if (refDCval >= 93) {
if (CsDiff < 1) {
if (mcTcor < 99) {
if (posBalance < 28) {
if (Rs < 725) {
return 8;
}
else if (Rs >= 725) {
return 1;
}
}
else if (posBalance >= 28) {
return 8;
}
}
else if (mcTcor >= 99) {
return 8;
}
}
else if (CsDiff >= 1) {
if (ssDCval < 135) {
return 8;
}
else if (ssDCval >= 135) {
return 1;
}
}
}
}
else if (Cs >= 1175) {
if (Rs < 808) {
return 2;
}
else if (Rs >= 808) {
return 8;
}
}
}
}
else if (Rs >= 965) {
if (MVDiff < 3) {
if (ssDCval < 51) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (ssDCval >= 51) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 33) {
if (negBalance < 2) {
return 1;
}
else if (negBalance >= 2) {
if (Rs < 1012) {
return 2;
}
else if (Rs >= 1012) {
if (posBalance < 29) {
if (tcor < 101) {
return 8;
}
else if (tcor >= 101) {
return 4;
}
}
else if (posBalance >= 29) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 33) {
if (ssDCval < 75) {
if (Cs < 1162) {
return 1;
}
else if (Cs >= 1162) {
return 8;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (diffMVdiffVal < -4) {
if (Rs < 1103) {
return 8;
}
else if (Rs >= 1103) {
if (Rs < 1126) {
return 4;
}
else if (Rs >= 1126) {
return 1;
}
}
}
else if (diffMVdiffVal >= -4) {
if (ssDCval < 62) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 1;
}
}
else if (ssDCval >= 62) {
return 8;
}
}
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
if (Rs < 1136) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 1012) {
return 8;
}
else if (Rs >= 1012) {
return 1;
}
}
}
}
else if (Rs >= 1136) {
if (ssDCval < 99) {
if (posBalance < 51) {
if (Cs < 1152) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
return 4;
}
}
else if (Cs >= 1152) {
return 8;
}
}
else if (posBalance >= 51) {
return 4;
}
}
else if (ssDCval >= 99) {
if (SC < 2220) {
return 1;
}
else if (SC >= 2220) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1841) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (refDCval < 79) {
if (Cs < 1002) {
if (Cs < 710) {
return 2;
}
else if (Cs >= 710) {
if (SC < 2951) {
return 1;
}
else if (SC >= 2951) {
if (SC < 2960) {
return 4;
}
else if (SC >= 2960) {
return 1;
}
}
}
}
else if (Cs >= 1002) {
if (Cs < 1143) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (Cs >= 1143) {
if (SC < 3567) {
if (Rs < 1945) {
return 4;
}
else if (Rs >= 1945) {
return 8;
}
}
else if (SC >= 3567) {
return 2;
}
}
}
}
else if (refDCval >= 79) {
if (Cs < 1300) {
return 8;
}
else if (Cs >= 1300) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 45) {
if (ssDCval < 47) {
if (mu_mv_mag_sq < 52) {
if (refDCval < 32) {
return 8;
}
else if (refDCval >= 32) {
if (negBalance < 32) {
if (mcTcor < 92) {
if (SC < 2157) {
if (ssDCval < 45) {
return 1;
}
else if (ssDCval >= 45) {
return 2;
}
}
else if (SC >= 2157) {
if (Rs < 1698) {
return 8;
}
else if (Rs >= 1698) {
return 4;
}
}
}
else if (mcTcor >= 92) {
if (Cs < 535) {
return 2;
}
else if (Cs >= 535) {
return 4;
}
}
}
else if (negBalance >= 32) {
if (Rs < 1482) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
else if (Rs >= 1482) {
if (Cs < 936) {
return 4;
}
else if (Cs >= 936) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (Rs < 1502) {
if (refDCval < 29) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
return 4;
}
}
else if (refDCval >= 29) {
if (SC < 1653) {
if (RsDiff < 3) {
if (Cs < 973) {
if (Cs < 708) {
if (Rs < 890) {
return 2;
}
else if (Rs >= 890) {
return 1;
}
}
else if (Cs >= 708) {
return 4;
}
}
else if (Cs >= 973) {
return 1;
}
}
else if (RsDiff >= 3) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
if (Cs < 640) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Cs >= 640) {
return 8;
}
}
}
}
else if (SC >= 1653) {
if (ssDCval < 46) {
if (diffRsCsdiff < 19) {
return 8;
}
else if (diffRsCsdiff >= 19) {
if (Rs < 1311) {
return 1;
}
else if (Rs >= 1311) {
return 8;
}
}
}
else if (ssDCval >= 46) {
if (Rs < 1375) {
return 2;
}
else if (Rs >= 1375) {
if (Rs < 1438) {
return 4;
}
else if (Rs >= 1438) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1502) {
if (MVDiff < 6) {
if (TSC < 3) {
if (Rs < 1817) {
return 4;
}
else if (Rs >= 1817) {
return 8;
}
}
else if (TSC >= 3) {
return 4;
}
}
else if (MVDiff >= 6) {
if (Cs < 827) {
return 4;
}
else if (Cs >= 827) {
return 1;
}
}
}
}
}
else if (ssDCval >= 47) {
if (mu_mv_mag_sq < 86) {
if (Cs < 890) {
if (negBalance < 62) {
if (Rs < 1544) {
if (ssDCval < 81) {
if (TSC < 5) {
if (RsCsDiff < 3) {
if (ssDCval < 63) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
return 8;
}
}
else if (ssDCval >= 63) {
return 4;
}
}
else if (RsCsDiff >= 3) {
if (Rs < 885) {
if (AFD < 9) {
if (posBalance < 33) {
return 1;
}
else if (posBalance >= 33) {
if (diffMVdiffVal < -1) {
if (RsCsDiff < 7) {
return 1;
}
else if (RsCsDiff >= 7) {
if (Rs < 603) {
return 1;
}
else if (Rs >= 603) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (Rs >= 885) {
if (negBalance < 32) {
if (Rs < 1332) {
return 1;
}
else if (Rs >= 1332) {
if (Rs < 1341) {
return 8;
}
else if (Rs >= 1341) {
return 1;
}
}
}
else if (negBalance >= 32) {
if (TSC < 4) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
else if (TSC >= 4) {
if (Cs < 592) {
return 2;
}
else if (Cs >= 592) {
if (Cs < 767) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (Cs >= 767) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 5) {
if (SC < 1772) {
if (Rs < 700) {
return 2;
}
else if (Rs >= 700) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (SC >= 1772) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
if (diffRsCsdiff < 12) {
return 2;
}
else if (diffRsCsdiff >= 12) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (ssDCval < 133) {
if (Rs < 1085) {
return 2;
}
else if (Rs >= 1085) {
return 1;
}
}
else if (ssDCval >= 133) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (Rs < 990) {
return 2;
}
else if (Rs >= 990) {
return 4;
}
}
}
}
}
else if (Rs >= 1544) {
if (negBalance < 29) {
if (Rs < 1758) {
return 2;
}
else if (Rs >= 1758) {
return 4;
}
}
else if (negBalance >= 29) {
if (tcor < 85) {
if (SC < 2753) {
if (RsDiff < 6) {
if (Cs < 722) {
return 4;
}
else if (Cs >= 722) {
return 2;
}
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (SC >= 2753) {
return 4;
}
}
else if (tcor >= 85) {
if (ssDCval < 69) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
return 2;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
}
else if (negBalance >= 62) {
if (ssDCval < 93) {
if (Rs < 359) {
return 1;
}
else if (Rs >= 359) {
return 8;
}
}
else if (ssDCval >= 93) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
else if (Cs >= 890) {
if (mu_mv_mag_sq < 66) {
if (Cs < 1150) {
if (gchDC < 3) {
if (mcTcor < 88) {
if (Rs < 995) {
if (Rs < 763) {
return 4;
}
else if (Rs >= 763) {
return 1;
}
}
else if (Rs >= 995) {
return 8;
}
}
else if (mcTcor >= 88) {
if (posBalance < 37) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (posBalance >= 37) {
if (tcor < 91) {
if (negBalance < 35) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 2;
}
}
else if (negBalance >= 35) {
if (refDCval < 56) {
return 4;
}
else if (refDCval >= 56) {
if (MVDiff < 24) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Rs < 879) {
return 8;
}
else if (Rs >= 879) {
return 4;
}
}
}
else if (MVDiff >= 24) {
return 2;
}
}
}
}
else if (tcor >= 91) {
if (diffMVdiffVal < -1) {
if (diffAFD < 1) {
if (MVDiff < 6) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (MVDiff >= 6) {
if (ssDCval < 136) {
return 8;
}
else if (ssDCval >= 136) {
return 2;
}
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
if (Cs < 977) {
if (Rs < 1050) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 2;
}
}
else if (Rs >= 1050) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
return 4;
}
}
}
else if (Cs >= 977) {
return 4;
}
}
}
}
}
}
else if (gchDC >= 3) {
if (ssDCval < 130) {
return 8;
}
else if (ssDCval >= 130) {
if (Rs < 1332) {
return 2;
}
else if (Rs >= 1332) {
return 8;
}
}
}
}
else if (Cs >= 1150) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (gchDC < 3) {
if (tcor < 92) {
return 2;
}
else if (tcor >= 92) {
return 8;
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (negBalance < 44) {
if (ssDCval < 139) {
if (SC < 2358) {
if (Rs < 1035) {
if (Cs < 934) {
if (Rs < 600) {
return 4;
}
else if (Rs >= 600) {
return 2;
}
}
else if (Cs >= 934) {
return 4;
}
}
else if (Rs >= 1035) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (SC >= 2358) {
if (CsDiff < 5) {
if (RsDiff < 3) {
if (Rs < 1729) {
return 4;
}
else if (Rs >= 1729) {
return 8;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (CsDiff >= 5) {
if (Rs < 2028) {
return 8;
}
else if (Rs >= 2028) {
return 4;
}
}
}
}
else if (ssDCval >= 139) {
return 8;
}
}
else if (negBalance >= 44) {
if (Cs < 913) {
return 1;
}
else if (Cs >= 913) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (ssDCval < 56) {
if (TSC < 3) {
if (Rs < 505) {
return 2;
}
else if (Rs >= 505) {
if (CsDiff < 2) {
if (ssDCval < 51) {
if (Rs < 1985) {
return 4;
}
else if (Rs >= 1985) {
return 8;
}
}
else if (ssDCval >= 51) {
return 8;
}
}
else if (CsDiff >= 2) {
return 4;
}
}
}
else if (TSC >= 3) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (negBalance < 35) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (RsDiff < 2) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (RsDiff >= 2) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (negBalance >= 35) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 114) {
if (ssDCval < 137) {
if (negBalance < 29) {
if (Rs < 866) {
return 4;
}
else if (Rs >= 866) {
if (ssDCval < 77) {
if (Rs < 1411) {
return 4;
}
else if (Rs >= 1411) {
if (SC < 2974) {
return 1;
}
else if (SC >= 2974) {
return 2;
}
}
}
else if (ssDCval >= 77) {
if (TSC < 4) {
if (Cs < 797) {
return 8;
}
else if (Cs >= 797) {
return 4;
}
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (negBalance >= 29) {
if (TSC < 3) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
if (Rs < 1350) {
return 2;
}
else if (Rs >= 1350) {
return 8;
}
}
}
else if (TSC >= 3) {
if (RsDiff < 6) {
if (gchDC < 2) {
if (Rs < 862) {
if (RsDiff < 3) {
if (Rs < 470) {
return 1;
}
else if (Rs >= 470) {
return 8;
}
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (Rs >= 862) {
return 8;
}
}
else if (gchDC >= 2) {
if (mu_mv_mag_sq < 104) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 2049) {
return 4;
}
else if (Rs >= 2049) {
return 4;
}
}
}
}
else if (RsDiff >= 6) {
if (Rs < 1968) {
return 8;
}
else if (Rs >= 1968) {
return 2;
}
}
}
}
}
else if (ssDCval >= 137) {
return 8;
}
}
else if (mu_mv_mag_sq >= 114) {
if (ssDCval < 133) {
return 8;
}
else if (ssDCval >= 133) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 45) {
if (mu_mv_mag_sq < 32) {
if (negBalance < 47) {
if (refDCval < 55) {
if (Rs < 881) {
if (ssDCval < 36) {
if (Cs < 919) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (Cs >= 919) {
return 1;
}
}
else if (ssDCval >= 36) {
if (refDCval < 43) {
if (RsCsDiff < 11) {
return 8;
}
else if (RsCsDiff >= 11) {
if (MVDiff < 12) {
if (Cs < 740) {
return 8;
}
else if (Cs >= 740) {
return 2;
}
}
else if (MVDiff >= 12) {
if (Cs < 1232) {
return 2;
}
else if (Cs >= 1232) {
return 1;
}
}
}
}
else if (refDCval >= 43) {
if (TSC < 5) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
if (Rs < 772) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
else if (Rs >= 772) {
return 4;
}
}
}
else if (TSC >= 5) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
if (Cs < 653) {
if (Rs < 738) {
return 1;
}
else if (Rs >= 738) {
return 4;
}
}
else if (Cs >= 653) {
return 8;
}
}
}
}
}
}
else if (Rs >= 881) {
if (Cs < 1146) {
if (Cs < 668) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (Rs < 903) {
return 8;
}
else if (Rs >= 903) {
return 2;
}
}
}
else if (Cs >= 668) {
return 8;
}
}
else if (Cs >= 1146) {
if (ssDCval < 48) {
return 8;
}
else if (ssDCval >= 48) {
if (Rs < 1287) {
return 4;
}
else if (Rs >= 1287) {
if (Rs < 1368) {
return 8;
}
else if (Rs >= 1368) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 55) {
if (mcTcor < 86) {
if (Cs < 925) {
if (MVDiff < 91) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (MVDiff >= 91) {
return 2;
}
}
else if (Cs >= 925) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
if (Cs < 1174) {
return 8;
}
else if (Cs >= 1174) {
return 1;
}
}
}
}
}
else if (mcTcor >= 86) {
if (RsDiff < 7) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 21) {
if (AFD < 8) {
if (ssDCval < 65) {
if (Rs < 1788) {
if (Cs < 962) {
return 4;
}
else if (Cs >= 962) {
if (MVDiff < 3) {
if (Rs < 1034) {
return 8;
}
else if (Rs >= 1034) {
return 1;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (Rs >= 1788) {
if (Rs < 2119) {
return 1;
}
else if (Rs >= 2119) {
return 2;
}
}
}
else if (ssDCval >= 65) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (SC < 1336) {
if (Cs < 606) {
return 4;
}
else if (Cs >= 606) {
if (Rs < 614) {
return 8;
}
else if (Rs >= 614) {
return 2;
}
}
}
else if (SC >= 1336) {
if (MVDiff < 24) {
if (Rs < 1951) {
return 1;
}
else if (Rs >= 1951) {
return 2;
}
}
else if (MVDiff >= 24) {
return 4;
}
}
}
}
}
else if (AFD >= 8) {
if (MVDiff < 4) {
if (Cs < 823) {
return 1;
}
else if (Cs >= 823) {
if (Cs < 1271) {
if (Rs < 2364) {
return 8;
}
else if (Rs >= 2364) {
return 2;
}
}
else if (Cs >= 1271) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (TSC < 6) {
if (ssDCval < 63) {
if (Rs < 484) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Rs >= 484) {
if (diffAFD < 1) {
if (MVDiff < 6) {
if (Rs < 1663) {
return 1;
}
else if (Rs >= 1663) {
return 2;
}
}
else if (MVDiff >= 6) {
if (Rs < 512) {
return 2;
}
else if (Rs >= 512) {
return 4;
}
}
}
else if (diffAFD >= 1) {
if (Rs < 707) {
return 2;
}
else if (Rs >= 707) {
return 8;
}
}
}
}
else if (ssDCval >= 63) {
if (Rs < 1271) {
if (mu_mv_mag_sq < 7) {
if (Rs < 892) {
return 2;
}
else if (Rs >= 892) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 7) {
return 1;
}
}
else if (Rs >= 1271) {
return 1;
}
}
}
else if (TSC >= 6) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (posBalance < 69) {
if (Cs < 1058) {
if (diffAFD < 2) {
if (tcor < 80) {
return 1;
}
else if (tcor >= 80) {
if (Rs < 2076) {
if (Rs < 745) {
return 2;
}
else if (Rs >= 745) {
if (diffRsCsdiff < -10) {
return 2;
}
else if (diffRsCsdiff >= -10) {
if (Cs < 1007) {
return 4;
}
else if (Cs >= 1007) {
return 2;
}
}
}
}
else if (Rs >= 2076) {
if (Rs < 2282) {
return 8;
}
else if (Rs >= 2282) {
return 1;
}
}
}
}
else if (diffAFD >= 2) {
if (Rs < 855) {
return 2;
}
else if (Rs >= 855) {
return 1;
}
}
}
else if (Cs >= 1058) {
if (SC < 2157) {
return 8;
}
else if (SC >= 2157) {
return 4;
}
}
}
else if (posBalance >= 69) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (Cs < 561) {
return 1;
}
else if (Cs >= 561) {
if (negBalance < 34) {
if (Rs < 1088) {
return 1;
}
else if (Rs >= 1088) {
return 8;
}
}
else if (negBalance >= 34) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 1;
}
}
}
}
}
else if (ssDCval >= 84) {
if (negBalance < 36) {
if (posBalance < 55) {
if (Cs < 947) {
return 2;
}
else if (Cs >= 947) {
if (Rs < 1363) {
return 1;
}
else if (Rs >= 1363) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (Rs < 762) {
return 8;
}
else if (Rs >= 762) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (Cs < 727) {
if (MVDiff < 5) {
if (Cs < 650) {
if (Rs < 627) {
return 1;
}
else if (Rs >= 627) {
return 2;
}
}
else if (Cs >= 650) {
if (Rs < 736) {
return 8;
}
else if (Rs >= 736) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (posBalance < 76) {
return 8;
}
else if (posBalance >= 76) {
return 2;
}
}
}
else if (Cs >= 727) {
if (tcor < 91) {
if (RsCsDiff < 19) {
if (ssDCval < 150) {
return 8;
}
else if (ssDCval >= 150) {
return 2;
}
}
else if (RsCsDiff >= 19) {
if (MVDiff < 33) {
if (Rs < 1422) {
return 2;
}
else if (Rs >= 1422) {
if (ssDCval < 101) {
return 2;
}
else if (ssDCval >= 101) {
return 8;
}
}
}
else if (MVDiff >= 33) {
return 4;
}
}
}
else if (tcor >= 91) {
if (Cs < 1082) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (Rs < 1188) {
return 4;
}
else if (Rs >= 1188) {
if (Rs < 1921) {
return 8;
}
else if (Rs >= 1921) {
return 4;
}
}
}
}
else if (Cs >= 1082) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (ssDCval < 73) {
if (Rs < 2062) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 2;
}
}
}
else if (Rs >= 2062) {
if (Cs < 1080) {
if (diffRsCsdiff < -318) {
return 4;
}
else if (diffRsCsdiff >= -318) {
return 2;
}
}
else if (Cs >= 1080) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 4;
}
}
}
}
else if (ssDCval >= 73) {
if (Cs < 753) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (Rs < 1979) {
return 1;
}
else if (Rs >= 1979) {
return 2;
}
}
}
else if (Cs >= 753) {
if (mu_mv_mag_sq < 3) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1947) {
return 2;
}
else if (Rs >= 1947) {
return 4;
}
}
}
}
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 908) {
if (CsDiff < 5) {
if (MVDiff < 5) {
if (ssDCval < 89) {
if (ssDCval < 59) {
if (refDCval < 52) {
if (mcTcor < 92) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
return 1;
}
}
else if (mcTcor >= 92) {
if (SC < 1405) {
return 4;
}
else if (SC >= 1405) {
return 8;
}
}
}
else if (refDCval >= 52) {
return 4;
}
}
else if (ssDCval >= 59) {
return 8;
}
}
else if (ssDCval >= 89) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (mcTcor < 95) {
if (AFD < 10) {
if (diffRsCsdiff < -15) {
if (Rs < 724) {
return 4;
}
else if (Rs >= 724) {
return 2;
}
}
else if (diffRsCsdiff >= -15) {
if (tcor < 72) {
if (mcTcor < 80) {
if (Cs < 766) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
if (Rs < 816) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (Rs >= 816) {
return 1;
}
}
}
else if (Cs >= 766) {
return 8;
}
}
else if (mcTcor >= 80) {
if (Rs < 710) {
return 4;
}
else if (Rs >= 710) {
return 1;
}
}
}
else if (tcor >= 72) {
if (MVDiff < 69) {
if (mcTcor < 80) {
if (Rs < 742) {
return 2;
}
else if (Rs >= 742) {
return 8;
}
}
else if (mcTcor >= 80) {
return 8;
}
}
else if (MVDiff >= 69) {
return 1;
}
}
}
}
else if (AFD >= 10) {
if (mu_mv_mag_sq < 2) {
if (posBalance < 57) {
if (Cs < 638) {
return 8;
}
else if (Cs >= 638) {
return 2;
}
}
else if (posBalance >= 57) {
if (mcTcor < 75) {
if (tcor < 54) {
return 4;
}
else if (tcor >= 54) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
}
else if (mcTcor >= 75) {
if (TSC < 9) {
if (Cs < 756) {
if (negBalance < 52) {
if (Rs < 467) {
return 2;
}
else if (Rs >= 467) {
return 4;
}
}
else if (negBalance >= 52) {
if (tcor < 61) {
return 4;
}
else if (tcor >= 61) {
return 1;
}
}
}
else if (Cs >= 756) {
return 4;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
if (RsCsDiff < 23) {
if (Rs < 472) {
return 1;
}
else if (Rs >= 472) {
return 8;
}
}
else if (RsCsDiff >= 23) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (Cs < 587) {
return 2;
}
else if (Cs >= 587) {
if (Cs < 695) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 1;
}
}
else if (Cs >= 695) {
if (ssDCval < 158) {
return 1;
}
else if (ssDCval >= 158) {
return 2;
}
}
}
}
}
}
else if (CsDiff >= 5) {
if (mcTcor < 84) {
if (Cs < 664) {
return 8;
}
else if (Cs >= 664) {
if (negBalance < 49) {
if (Cs < 981) {
return 2;
}
else if (Cs >= 981) {
return 4;
}
}
else if (negBalance >= 49) {
if (MVDiff < 13) {
if (Cs < 1029) {
return 8;
}
else if (Cs >= 1029) {
return 4;
}
}
else if (MVDiff >= 13) {
return 4;
}
}
}
}
else if (mcTcor >= 84) {
if (RsCsDiff < 51) {
if (SC < 1567) {
if (Rs < 648) {
return 4;
}
else if (Rs >= 648) {
return 2;
}
}
else if (SC >= 1567) {
if (Cs < 1265) {
return 1;
}
else if (Cs >= 1265) {
return 2;
}
}
}
else if (RsCsDiff >= 51) {
return 4;
}
}
}
}
else if (Rs >= 908) {
if (RsDiff < 7) {
if (tcor < 79) {
if (ssDCval < 125) {
if (MVDiff < 12) {
if (refDCval < 47) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
else if (refDCval >= 47) {
if (Cs < 1296) {
if (SC < 2678) {
return 8;
}
else if (SC >= 2678) {
if (SC < 2694) {
return 2;
}
else if (SC >= 2694) {
return 8;
}
}
}
else if (Cs >= 1296) {
return 4;
}
}
}
else if (MVDiff >= 12) {
if (Rs < 1142) {
if (Cs < 1185) {
if (tcor < 62) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (tcor >= 62) {
return 8;
}
}
else if (Cs >= 1185) {
return 2;
}
}
else if (Rs >= 1142) {
if (SC < 1985) {
return 2;
}
else if (SC >= 1985) {
if (posBalance < 58) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
if (Cs < 1184) {
return 2;
}
else if (Cs >= 1184) {
return 8;
}
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
}
else if (ssDCval >= 125) {
if (Cs < 899) {
return 1;
}
else if (Cs >= 899) {
return 8;
}
}
}
else if (tcor >= 79) {
if (refDCval < 62) {
if (Cs < 1044) {
if (Rs < 1514) {
return 8;
}
else if (Rs >= 1514) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 1571) {
return 4;
}
else if (Rs >= 1571) {
return 1;
}
}
}
}
else if (Cs >= 1044) {
if (Rs < 1362) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1041) {
return 4;
}
else if (Rs >= 1041) {
return 8;
}
}
}
else if (Rs >= 1362) {
return 8;
}
}
}
else if (refDCval >= 62) {
if (MVDiff < 15) {
if (ssDCval < 97) {
if (RsCsDiff < 19) {
return 8;
}
else if (RsCsDiff >= 19) {
if (TSC < 6) {
if (MVDiff < 7) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (TSC >= 6) {
if (Cs < 646) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 1078) {
return 1;
}
else if (Rs >= 1078) {
return 8;
}
}
}
else if (Cs >= 646) {
if (RsDiff < 6) {
if (negBalance < 50) {
if (Rs < 1006) {
return 4;
}
else if (Rs >= 1006) {
return 8;
}
}
else if (negBalance >= 50) {
if (AFD < 11) {
if (SC < 2198) {
return 8;
}
else if (SC >= 2198) {
return 1;
}
}
else if (AFD >= 11) {
if (posBalance < 55) {
if (gchDC < 2) {
if (Cs < 804) {
return 8;
}
else if (Cs >= 804) {
return 1;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
else if (RsDiff >= 6) {
if (tcor < 84) {
return 4;
}
else if (tcor >= 84) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 97) {
if (Rs < 938) {
return 4;
}
else if (Rs >= 938) {
if (TSC < 11) {
if (diffRsCsdiff < 5) {
return 8;
}
else if (diffRsCsdiff >= 5) {
if (AFD < 12) {
if (diffMVdiffVal < -6) {
return 2;
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
else if (AFD >= 12) {
return 2;
}
}
}
else if (TSC >= 11) {
return 1;
}
}
}
}
else if (MVDiff >= 15) {
if (Rs < 1097) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
else if (Rs >= 1097) {
if (mu_mv_mag_sq < 7) {
if (TSC < 8) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (MVDiff < 18) {
if (Cs < 869) {
return 8;
}
else if (Cs >= 869) {
if (Rs < 1402) {
return 2;
}
else if (Rs >= 1402) {
return 4;
}
}
}
else if (MVDiff >= 18) {
return 2;
}
}
}
else if (TSC >= 8) {
if (ssDCval < 106) {
return 8;
}
else if (ssDCval >= 106) {
if (Cs < 795) {
return 1;
}
else if (Cs >= 795) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (mcTcor < 95) {
if (AFD < 12) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
return 8;
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (mcTcor >= 95) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (AFD < 11) {
if (posBalance < 50) {
if (SC < 2341) {
if (SC < 2023) {
return 1;
}
else if (SC >= 2023) {
return 8;
}
}
else if (SC >= 2341) {
return 4;
}
}
else if (posBalance >= 50) {
if (Cs < 729) {
return 2;
}
else if (Cs >= 729) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 1;
}
}
}
}
else if (AFD >= 11) {
if (Rs < 2158) {
if (diffRsCsdiff < 16) {
if (tcor < 83) {
if (Cs < 1069) {
if (Rs < 1772) {
return 8;
}
else if (Rs >= 1772) {
return 4;
}
}
else if (Cs >= 1069) {
return 2;
}
}
else if (tcor >= 83) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
if (CsDiff < 6) {
if (mu_mv_mag_sq < 7) {
return 1;
}
else if (mu_mv_mag_sq >= 7) {
if (Rs < 1702) {
return 4;
}
else if (Rs >= 1702) {
if (Rs < 1820) {
return 2;
}
else if (Rs >= 1820) {
return 1;
}
}
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 16) {
if (tcor < 68) {
return 4;
}
else if (tcor >= 68) {
if (diffAFD < 1) {
if (Rs < 2002) {
return 2;
}
else if (Rs >= 2002) {
return 8;
}
}
else if (diffAFD >= 1) {
if (RsCsDiff < 113) {
return 8;
}
else if (RsCsDiff >= 113) {
return 4;
}
}
}
}
}
else if (Rs >= 2158) {
if (Cs < 876) {
return 2;
}
else if (Cs >= 876) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (Rs < 1596) {
if (TSC < 7) {
if (refDCval < 41) {
if (Rs < 813) {
if (Cs < 715) {
if (Rs < 404) {
return 1;
}
else if (Rs >= 404) {
return 8;
}
}
else if (Cs >= 715) {
return 1;
}
}
else if (Rs >= 813) {
if (refDCval < 39) {
return 8;
}
else if (refDCval >= 39) {
if (mu_mv_mag_sq < 59) {
if (Cs < 976) {
return 1;
}
else if (Cs >= 976) {
return 8;
}
}
else if (mu_mv_mag_sq >= 59) {
return 8;
}
}
}
}
else if (refDCval >= 41) {
if (ssDCval < 158) {
if (Rs < 456) {
if (ssDCval < 96) {
if (refDCval < 60) {
if (MVDiff < 14) {
if (Cs < 573) {
return 2;
}
else if (Cs >= 573) {
if (Cs < 1055) {
return 1;
}
else if (Cs >= 1055) {
if (Rs < 407) {
return 2;
}
else if (Rs >= 407) {
return 1;
}
}
}
}
else if (MVDiff >= 14) {
if (Rs < 401) {
return 1;
}
else if (Rs >= 401) {
if (Cs < 694) {
return 4;
}
else if (Cs >= 694) {
return 8;
}
}
}
}
else if (refDCval >= 60) {
if (negBalance < 41) {
if (SC < 1494) {
if (Rs < 306) {
return 2;
}
else if (Rs >= 306) {
if (Cs < 672) {
if (Cs < 665) {
return 8;
}
else if (Cs >= 665) {
return 4;
}
}
else if (Cs >= 672) {
return 8;
}
}
}
else if (SC >= 1494) {
return 1;
}
}
else if (negBalance >= 41) {
if (diffMVdiffVal < -3) {
if (Rs < 355) {
return 1;
}
else if (Rs >= 355) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (Cs < 637) {
if (Cs < 546) {
return 1;
}
else if (Cs >= 546) {
return 8;
}
}
else if (Cs >= 637) {
return 1;
}
}
}
}
}
else if (ssDCval >= 96) {
if (MVDiff < 5) {
if (Cs < 757) {
return 1;
}
else if (Cs >= 757) {
return 4;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (Rs >= 456) {
if (ssDCval < 77) {
if (TSCindex < 3) {
if (RsDiff < 2) {
if (Cs < 958) {
return 8;
}
else if (Cs >= 958) {
return 1;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (TSCindex >= 3) {
if (ssDCval < 64) {
if (Rs < 797) {
if (refDCval < 55) {
if (Rs < 521) {
if (Rs < 476) {
if (Rs < 469) {
return 2;
}
else if (Rs >= 469) {
return 4;
}
}
else if (Rs >= 476) {
return 1;
}
}
else if (Rs >= 521) {
if (Rs < 666) {
if (Rs < 661) {
return 4;
}
else if (Rs >= 661) {
return 1;
}
}
else if (Rs >= 666) {
if (Cs < 658) {
return 8;
}
else if (Cs >= 658) {
if (mu_mv_mag_sq < 98) {
return 4;
}
else if (mu_mv_mag_sq >= 98) {
if (Cs < 750) {
return 4;
}
else if (Cs >= 750) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 55) {
if (mu_mv_mag_sq < 83) {
if (negBalance < 41) {
if (CsDiff < 2) {
if (Rs < 784) {
return 8;
}
else if (Rs >= 784) {
return 4;
}
}
else if (CsDiff >= 2) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (negBalance >= 41) {
return 8;
}
}
else if (mu_mv_mag_sq >= 83) {
if (refDCval < 58) {
return 8;
}
else if (refDCval >= 58) {
if (mu_mv_mag_sq < 109) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (mu_mv_mag_sq >= 109) {
return 4;
}
}
}
}
}
else if (Rs >= 797) {
if (mu_mv_mag_sq < 115) {
if (refDCval < 60) {
if (Cs < 1043) {
if (RsDiff < 5) {
if (negBalance < 39) {
if (Rs < 1133) {
if (diffRsCsdiff < 13) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (diffRsCsdiff >= 13) {
return 1;
}
}
else if (Rs >= 1133) {
if (Cs < 736) {
if (AFD < 6) {
if (Rs < 1368) {
return 4;
}
else if (Rs >= 1368) {
return 4;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 736) {
if (Cs < 859) {
if (CsDiff < 2) {
if (Cs < 851) {
return 8;
}
else if (Cs >= 851) {
return 1;
}
}
else if (CsDiff >= 2) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 2;
}
}
}
else if (Cs >= 859) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (Cs < 958) {
if (Rs < 1220) {
return 8;
}
else if (Rs >= 1220) {
return 2;
}
}
else if (Cs >= 958) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 1005) {
if (negBalance < 47) {
if (Cs < 744) {
return 8;
}
else if (Cs >= 744) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (tcor < 88) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
else if (tcor >= 88) {
return 8;
}
}
}
}
else if (negBalance >= 47) {
if (RsDiff < 3) {
if (mcTcor < 90) {
return 1;
}
else if (mcTcor >= 90) {
return 8;
}
}
else if (RsDiff >= 3) {
return 8;
}
}
}
else if (Rs >= 1005) {
if (diffRsCsdiff < -11) {
if (Cs < 985) {
return 8;
}
else if (Cs >= 985) {
return 4;
}
}
else if (diffRsCsdiff >= -11) {
return 8;
}
}
}
}
else if (RsDiff >= 5) {
if (negBalance < 34) {
if (AFD < 10) {
if (Rs < 969) {
return 4;
}
else if (Rs >= 969) {
return 1;
}
}
else if (AFD >= 10) {
return 2;
}
}
else if (negBalance >= 34) {
if (tcor < 80) {
if (mu_mv_mag_sq < 94) {
return 8;
}
else if (mu_mv_mag_sq >= 94) {
if (Cs < 773) {
return 8;
}
else if (Cs >= 773) {
return 1;
}
}
}
else if (tcor >= 80) {
if (refDCval < 50) {
if (Rs < 963) {
return 8;
}
else if (Rs >= 963) {
return 2;
}
}
else if (refDCval >= 50) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1043) {
if (Rs < 1303) {
if (refDCval < 50) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (Rs < 848) {
return 2;
}
else if (Rs >= 848) {
return 8;
}
}
}
else if (refDCval >= 50) {
if (negBalance < 36) {
if (Cs < 1274) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
return 8;
}
}
else if (Cs >= 1274) {
return 4;
}
}
else if (negBalance >= 36) {
if (Rs < 951) {
return 8;
}
else if (Rs >= 951) {
if (Rs < 1205) {
return 4;
}
else if (Rs >= 1205) {
if (Cs < 1203) {
if (Rs < 1232) {
return 1;
}
else if (Rs >= 1232) {
return 8;
}
}
else if (Cs >= 1203) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1303) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 1;
}
}
}
}
else if (refDCval >= 60) {
if (posBalance < 48) {
if (negBalance < 33) {
return 2;
}
else if (negBalance >= 33) {
if (SC < 2348) {
return 4;
}
else if (SC >= 2348) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (Rs < 990) {
if (Rs < 921) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (Rs >= 921) {
if (mu_mv_mag_sq < 60) {
return 8;
}
else if (mu_mv_mag_sq >= 60) {
return 4;
}
}
}
else if (Rs >= 990) {
if (negBalance < 36) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (negBalance >= 36) {
if (negBalance < 44) {
if (Cs < 819) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (Cs >= 819) {
if (Rs < 1085) {
return 8;
}
else if (Rs >= 1085) {
if (AFD < 8) {
if (Rs < 1314) {
return 1;
}
else if (Rs >= 1314) {
if (Cs < 1222) {
return 8;
}
else if (Cs >= 1222) {
return 1;
}
}
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 1108) {
if (Cs < 911) {
return 8;
}
else if (Cs >= 911) {
if (Rs < 1061) {
return 1;
}
else if (Rs >= 1061) {
return 4;
}
}
}
else if (Rs >= 1108) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (tcor < 87) {
if (posBalance < 56) {
if (Rs < 1029) {
return 8;
}
else if (Rs >= 1029) {
return 1;
}
}
else if (posBalance >= 56) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
return 1;
}
}
}
else if (tcor >= 87) {
if (tcor < 97) {
if (tcor < 89) {
if (Cs < 857) {
if (Rs < 1178) {
return 8;
}
else if (Rs >= 1178) {
return 1;
}
}
else if (Cs >= 857) {
return 8;
}
}
else if (tcor >= 89) {
return 8;
}
}
else if (tcor >= 97) {
return 1;
}
}
}
}
}
else if (ssDCval >= 64) {
if (Cs < 1100) {
if (posBalance < 56) {
if (negBalance < 47) {
if (Cs < 656) {
if (diffRsCsdiff < -2) {
return 1;
}
else if (diffRsCsdiff >= -2) {
if (Rs < 636) {
if (Rs < 546) {
return 2;
}
else if (Rs >= 546) {
return 1;
}
}
else if (Rs >= 636) {
return 4;
}
}
}
else if (Cs >= 656) {
if (MVDiff < 6) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 93) {
if (ssDCval < 71) {
if (Rs < 1000) {
return 2;
}
else if (Rs >= 1000) {
return 1;
}
}
else if (ssDCval >= 71) {
if (negBalance < 38) {
if (Cs < 924) {
return 1;
}
else if (Cs >= 924) {
return 8;
}
}
else if (negBalance >= 38) {
if (Cs < 1038) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 8;
}
}
else if (Cs >= 1038) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
return 2;
}
}
}
}
else if (MVDiff >= 6) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
if (Rs < 1434) {
return 4;
}
else if (Rs >= 1434) {
return 8;
}
}
}
}
}
else if (negBalance >= 47) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
else if (posBalance >= 56) {
if (negBalance < 45) {
if (RsCsDiff < 16) {
if (AFD < 5) {
if (Rs < 747) {
if (Rs < 696) {
return 1;
}
else if (Rs >= 696) {
return 2;
}
}
else if (Rs >= 747) {
if (Rs < 897) {
return 4;
}
else if (Rs >= 897) {
return 8;
}
}
}
else if (AFD >= 5) {
if (posBalance < 70) {
return 8;
}
else if (posBalance >= 70) {
if (Rs < 1040) {
return 2;
}
else if (Rs >= 1040) {
return 8;
}
}
}
}
else if (RsCsDiff >= 16) {
if (Cs < 941) {
if (Rs < 963) {
return 8;
}
else if (Rs >= 963) {
if (Rs < 1535) {
return 1;
}
else if (Rs >= 1535) {
return 8;
}
}
}
else if (Cs >= 941) {
if (mu_mv_mag_sq < 47) {
return 8;
}
else if (mu_mv_mag_sq >= 47) {
return 4;
}
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (Cs >= 1100) {
if (Rs < 877) {
if (mu_mv_mag_sq < 81) {
if (Rs < 851) {
return 1;
}
else if (Rs >= 851) {
return 2;
}
}
else if (mu_mv_mag_sq >= 81) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (Rs >= 877) {
if (SC < 2531) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 1214) {
if (mu_mv_mag_sq < 62) {
if (Rs < 1092) {
return 8;
}
else if (Rs >= 1092) {
return 2;
}
}
else if (mu_mv_mag_sq >= 62) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
return 8;
}
}
}
else if (Rs >= 1214) {
if (Cs < 1173) {
if (Rs < 1344) {
if (diffRsCsdiff < 5) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (diffRsCsdiff >= 5) {
return 4;
}
}
else if (Rs >= 1344) {
return 4;
}
}
else if (Cs >= 1173) {
if (ssDCval < 67) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
return 4;
}
}
else if (ssDCval >= 67) {
return 1;
}
}
}
}
}
else if (SC >= 2531) {
if (mu_mv_mag_sq < 72) {
if (Cs < 1163) {
return 4;
}
else if (Cs >= 1163) {
return 1;
}
}
else if (mu_mv_mag_sq >= 72) {
if (diffTSC < 1) {
if (Rs < 1327) {
return 1;
}
else if (Rs >= 1327) {
return 8;
}
}
else if (diffTSC >= 1) {
if (Rs < 1341) {
return 1;
}
else if (Rs >= 1341) {
if (posBalance < 57) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 77) {
if (ssDCval < 130) {
if (Cs < 957) {
if (tcor < 79) {
if (Rs < 884) {
return 2;
}
else if (Rs >= 884) {
return 1;
}
}
else if (tcor >= 79) {
if (TSC < 5) {
if (refDCval < 98) {
if (ssDCval < 91) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
if (posBalance < 50) {
if (Rs < 669) {
return 2;
}
else if (Rs >= 669) {
return 1;
}
}
else if (posBalance >= 50) {
return 8;
}
}
}
else if (ssDCval >= 91) {
if (Rs < 705) {
return 2;
}
else if (Rs >= 705) {
return 1;
}
}
}
else if (refDCval >= 98) {
return 8;
}
}
else if (TSC >= 5) {
if (AFD < 10) {
if (RsCsDiff < 7) {
if (MVDiff < 3) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 1;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (RsCsDiff >= 7) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (diffRsCsdiff < -4) {
return 1;
}
else if (diffRsCsdiff >= -4) {
return 8;
}
}
}
}
else if (AFD >= 10) {
return 2;
}
}
}
}
else if (Cs >= 957) {
if (posBalance < 48) {
if (Rs < 990) {
return 4;
}
else if (Rs >= 990) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (posBalance >= 48) {
if (Rs < 940) {
if (refDCval < 91) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
else if (refDCval >= 91) {
if (Rs < 895) {
return 8;
}
else if (Rs >= 895) {
if (Rs < 898) {
return 4;
}
else if (Rs >= 898) {
return 8;
}
}
}
}
else if (Rs >= 940) {
if (RsDiff < 6) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 1275) {
return 8;
}
else if (Rs >= 1275) {
return 4;
}
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
else if (ssDCval >= 130) {
if (Rs < 800) {
return 8;
}
else if (Rs >= 800) {
if (AFD < 7) {
if (MVDiff < 24) {
if (AFD < 6) {
if (Rs < 962) {
return 8;
}
else if (Rs >= 962) {
if (posBalance < 65) {
if (Rs < 1486) {
return 4;
}
else if (Rs >= 1486) {
return 8;
}
}
else if (posBalance >= 65) {
return 1;
}
}
}
else if (AFD >= 6) {
if (diffMVdiffVal < -7) {
if (Cs < 859) {
return 8;
}
else if (Cs >= 859) {
return 4;
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
}
else if (MVDiff >= 24) {
return 4;
}
}
else if (AFD >= 7) {
if (SC < 1932) {
if (Rs < 981) {
return 1;
}
else if (Rs >= 981) {
return 4;
}
}
else if (SC >= 1932) {
if (tcor < 89) {
if (Cs < 996) {
if (Rs < 1197) {
return 1;
}
else if (Rs >= 1197) {
return 4;
}
}
else if (Cs >= 996) {
return 8;
}
}
else if (tcor >= 89) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 158) {
if (Rs < 913) {
return 4;
}
else if (Rs >= 913) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (Rs < 1192) {
if (ssDCval < 79) {
if (refDCval < 50) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (mcTcor < 75) {
return 1;
}
else if (mcTcor >= 75) {
if (Rs < 726) {
return 2;
}
else if (Rs >= 726) {
return 8;
}
}
}
}
else if (refDCval >= 50) {
if (mu_mv_mag_sq < 99) {
if (ssDCval < 67) {
if (TSC < 8) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (SC < 1886) {
if (Rs < 1029) {
if (ssDCval < 53) {
if (Rs < 894) {
return 4;
}
else if (Rs >= 894) {
return 8;
}
}
else if (ssDCval >= 53) {
return 4;
}
}
else if (Rs >= 1029) {
if (Rs < 1112) {
return 8;
}
else if (Rs >= 1112) {
return 4;
}
}
}
else if (SC >= 1886) {
if (Cs < 1153) {
return 8;
}
else if (Cs >= 1153) {
if (Rs < 1022) {
return 8;
}
else if (Rs >= 1022) {
return 4;
}
}
}
}
}
else if (TSC >= 8) {
if (RsCsDiff < 19) {
return 4;
}
else if (RsCsDiff >= 19) {
return 8;
}
}
}
else if (ssDCval >= 67) {
if (MVDiff < 7) {
if (refDCval < 73) {
return 8;
}
else if (refDCval >= 73) {
return 4;
}
}
else if (MVDiff >= 7) {
if (diffRsCsdiff < 3) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (diffRsCsdiff >= 3) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
return 8;
}
}
}
else if (ssDCval >= 79) {
if (diffAFD < -2) {
if (Rs < 858) {
return 8;
}
else if (Rs >= 858) {
return 1;
}
}
else if (diffAFD >= -2) {
if (ssDCval < 151) {
return 8;
}
else if (ssDCval >= 151) {
return 4;
}
}
}
}
else if (Rs >= 1192) {
if (Cs < 1124) {
if (tcor < 66) {
if (Cs < 598) {
return 4;
}
else if (Cs >= 598) {
return 8;
}
}
else if (tcor >= 66) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 40) {
return 2;
}
else if (mu_mv_mag_sq >= 40) {
if (MVDiff < 4) {
if (Rs < 1298) {
return 1;
}
else if (Rs >= 1298) {
return 8;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (Cs >= 1124) {
if (mcTcor < 96) {
if (negBalance < 45) {
if (mcTcor < 88) {
if (MVDiff < 6) {
if (Cs < 1252) {
return 8;
}
else if (Cs >= 1252) {
if (Rs < 1390) {
return 1;
}
else if (Rs >= 1390) {
return 4;
}
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (mcTcor >= 88) {
if (diffRsCsdiff < 21) {
return 8;
}
else if (diffRsCsdiff >= 21) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
return 2;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1420) {
return 4;
}
else if (Rs >= 1420) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1596) {
if (ssDCval < 71) {
if (ssDCval < 62) {
if (Cs < 891) {
if (refDCval < 55) {
if (MVDiff < 29) {
if (Rs < 1986) {
if (Rs < 1801) {
return 4;
}
else if (Rs >= 1801) {
return 4;
}
}
else if (Rs >= 1986) {
return 4;
}
}
else if (MVDiff >= 29) {
if (Rs < 2283) {
return 2;
}
else if (Rs >= 2283) {
return 4;
}
}
}
else if (refDCval >= 55) {
if (SC < 2698) {
if (posBalance < 55) {
if (SC < 2405) {
if (Rs < 1641) {
return 4;
}
else if (Rs >= 1641) {
return 2;
}
}
else if (SC >= 2405) {
if (Cs < 605) {
return 8;
}
else if (Cs >= 605) {
return 1;
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (SC >= 2698) {
return 2;
}
}
}
else if (Cs >= 891) {
if (diffAFD < -5) {
if (Rs < 1688) {
return 1;
}
else if (Rs >= 1688) {
return 2;
}
}
else if (diffAFD >= -5) {
if (mu_mv_mag_sq < 119) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (mu_mv_mag_sq >= 119) {
if (Cs < 1070) {
return 2;
}
else if (Cs >= 1070) {
if (Rs < 1947) {
return 8;
}
else if (Rs >= 1947) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 62) {
if (CsDiff < 5) {
if (MVDiff < 14) {
if (RsDiff < 5) {
if (TSC < 5) {
if (Cs < 1091) {
return 2;
}
else if (Cs >= 1091) {
return 1;
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 81) {
return 8;
}
else if (mu_mv_mag_sq >= 81) {
if (Cs < 895) {
return 8;
}
else if (Cs >= 895) {
if (Rs < 1923) {
return 2;
}
else if (Rs >= 1923) {
return 8;
}
}
}
}
}
else if (RsDiff >= 5) {
if (mcTcor < 91) {
if (Rs < 1682) {
return 8;
}
else if (Rs >= 1682) {
return 2;
}
}
else if (mcTcor >= 91) {
if (posBalance < 67) {
return 2;
}
else if (posBalance >= 67) {
if (Rs < 2278) {
return 8;
}
else if (Rs >= 2278) {
return 2;
}
}
}
}
}
else if (MVDiff >= 14) {
if (gchDC < 1) {
if (Rs < 1856) {
return 4;
}
else if (Rs >= 1856) {
return 8;
}
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 2170) {
return 8;
}
else if (Rs >= 2170) {
return 2;
}
}
}
}
else if (ssDCval >= 71) {
if (RsDiff < 8) {
if (refDCval < 79) {
if (negBalance < 47) {
if (mcTcor < 93) {
if (Rs < 2191) {
return 8;
}
else if (Rs >= 2191) {
return 4;
}
}
else if (mcTcor >= 93) {
if (MVDiff < 1) {
if (Rs < 1808) {
return 8;
}
else if (Rs >= 1808) {
return 1;
}
}
else if (MVDiff >= 1) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
else if (negBalance >= 47) {
if (diffMVdiffVal < 9) {
return 8;
}
else if (diffMVdiffVal >= 9) {
return 1;
}
}
}
else if (refDCval >= 79) {
if (MVDiff < 4) {
if (negBalance < 45) {
if (mu_mv_mag_sq < 103) {
if (ssDCval < 95) {
if (ssDCval < 86) {
if (AFD < 8) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (AFD >= 8) {
return 8;
}
}
else if (ssDCval >= 86) {
if (Cs < 1069) {
if (Rs < 2038) {
return 8;
}
else if (Rs >= 2038) {
return 1;
}
}
else if (Cs >= 1069) {
return 2;
}
}
}
else if (ssDCval >= 95) {
if (diffAFD < 2) {
if (Rs < 2372) {
return 8;
}
else if (Rs >= 2372) {
return 1;
}
}
else if (diffAFD >= 2) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (ssDCval < 129) {
return 8;
}
else if (ssDCval >= 129) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (posBalance < 56) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (Cs < 1288) {
return 8;
}
else if (Cs >= 1288) {
if (Rs < 1843) {
return 1;
}
else if (Rs >= 1843) {
return 8;
}
}
}
}
else if (posBalance >= 56) {
if (Cs < 1052) {
return 1;
}
else if (Cs >= 1052) {
if (AFD < 10) {
if (ssDCval < 85) {
if (Rs < 1695) {
return 8;
}
else if (Rs >= 1695) {
return 1;
}
}
else if (ssDCval >= 85) {
return 8;
}
}
else if (AFD >= 10) {
return 1;
}
}
}
}
}
else if (MVDiff >= 4) {
if (RsDiff < 6) {
if (Rs < 2270) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 100) {
return 8;
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 1739) {
return 2;
}
else if (Rs >= 1739) {
return 8;
}
}
}
}
else if (Rs >= 2270) {
if (tcor < 95) {
if (mu_mv_mag_sq < 70) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 8;
}
}
else if (mu_mv_mag_sq >= 70) {
return 8;
}
}
else if (tcor >= 95) {
return 1;
}
}
}
else if (RsDiff >= 6) {
if (Cs < 791) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 2;
}
}
else if (Cs >= 791) {
if (ssDCval < 101) {
if (TSC < 8) {
if (mcTcor < 95) {
return 2;
}
else if (mcTcor >= 95) {
if (Rs < 1857) {
return 4;
}
else if (Rs >= 1857) {
return 8;
}
}
}
else if (TSC >= 8) {
if (SC < 2900) {
if (Rs < 1662) {
return 8;
}
else if (Rs >= 1662) {
return 4;
}
}
else if (SC >= 2900) {
if (Rs < 2243) {
return 8;
}
else if (Rs >= 2243) {
return 2;
}
}
}
}
else if (ssDCval >= 101) {
if (MVDiff < 5) {
if (CsDiff < 4) {
if (Rs < 1944) {
return 2;
}
else if (Rs >= 1944) {
return 1;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (Cs < 804) {
return 1;
}
else if (Cs >= 804) {
if (RsDiff < 9) {
if (MVDiff < 10) {
if (Rs < 1935) {
return 4;
}
else if (Rs >= 1935) {
return 8;
}
}
else if (MVDiff >= 10) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 8;
}
}
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (refDCval < 114) {
if (negBalance < 46) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 33) {
if (refDCval < 61) {
return 1;
}
else if (refDCval >= 61) {
if (Rs < 2297) {
if (Rs < 1823) {
return 8;
}
else if (Rs >= 1823) {
return 2;
}
}
else if (Rs >= 2297) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
return 2;
}
}
}
else if (posBalance >= 49) {
if (SC < 1655) {
if (CsDiff < 6) {
if (MVDiff < 68) {
if (SC < 1527) {
if (diffMVdiffVal < -12) {
if (Cs < 651) {
return 2;
}
else if (Cs >= 651) {
return 8;
}
}
else if (diffMVdiffVal >= -12) {
if (diffMVdiffVal < 17) {
if (MVDiff < 30) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 8;
}
}
else if (MVDiff >= 30) {
return 1;
}
}
else if (diffMVdiffVal >= 17) {
return 8;
}
}
}
else if (SC >= 1527) {
if (Rs < 902) {
return 4;
}
else if (Rs >= 902) {
return 2;
}
}
}
else if (MVDiff >= 68) {
if (mcTcor < 85) {
return 4;
}
else if (mcTcor >= 85) {
if (Rs < 339) {
return 4;
}
else if (Rs >= 339) {
return 1;
}
}
}
}
else if (CsDiff >= 6) {
if (refDCval < 53) {
if (Cs < 827) {
return 2;
}
else if (Cs >= 827) {
return 8;
}
}
else if (refDCval >= 53) {
if (refDCval < 72) {
if (Cs < 561) {
return 8;
}
else if (Cs >= 561) {
if (AFD < 29) {
return 4;
}
else if (AFD >= 29) {
return 2;
}
}
}
else if (refDCval >= 72) {
return 2;
}
}
}
}
else if (SC >= 1655) {
if (Cs < 783) {
if (tcor < 76) {
if (Rs < 1061) {
return 1;
}
else if (Rs >= 1061) {
if (ssDCval < 73) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 4;
}
}
}
else if (ssDCval >= 73) {
return 8;
}
}
}
else if (tcor >= 76) {
if (AFD < 19) {
if (tcor < 87) {
if (Rs < 1239) {
return 4;
}
else if (Rs >= 1239) {
return 2;
}
}
else if (tcor >= 87) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
}
else if (AFD >= 19) {
return 1;
}
}
}
else if (Cs >= 783) {
if (posBalance < 65) {
if (ssDCval < 60) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (Rs < 1287) {
return 1;
}
else if (Rs >= 1287) {
if (Rs < 1445) {
return 8;
}
else if (Rs >= 1445) {
return 4;
}
}
}
}
else if (ssDCval >= 60) {
if (Cs < 1121) {
if (mu_mv_mag_sq < 5) {
if (AFD < 14) {
if (Cs < 1074) {
return 2;
}
else if (Cs >= 1074) {
if (Rs < 1178) {
return 4;
}
else if (Rs >= 1178) {
return 1;
}
}
}
else if (AFD >= 14) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
if (Rs < 899) {
return 4;
}
else if (Rs >= 899) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (mu_mv_mag_sq < 58) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (Cs < 806) {
return 1;
}
else if (Cs >= 806) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 1807) {
return 4;
}
else if (Rs >= 1807) {
return 8;
}
}
}
}
else if (Cs >= 1121) {
if (AFD < 17) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (AFD >= 17) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 4;
}
}
}
}
}
else if (posBalance >= 65) {
if (diffRsCsdiff < -15) {
if (posBalance < 75) {
if (diffRsCsdiff < -71) {
return 2;
}
else if (diffRsCsdiff >= -71) {
if (CsDiff < 6) {
if (Cs < 868) {
return 1;
}
else if (Cs >= 868) {
return 8;
}
}
else if (CsDiff >= 6) {
return 1;
}
}
}
else if (posBalance >= 75) {
if (tcor < 48) {
if (Rs < 1373) {
return 1;
}
else if (Rs >= 1373) {
return 4;
}
}
else if (tcor >= 48) {
return 2;
}
}
}
else if (diffRsCsdiff >= -15) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (tcor < 75) {
if (SC < 1893) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 2;
}
}
else if (SC >= 1893) {
if (RsDiff < 12) {
if (Cs < 885) {
return 2;
}
else if (Cs >= 885) {
if (tcor < 43) {
if (Rs < 1747) {
return 1;
}
else if (Rs >= 1747) {
return 8;
}
}
else if (tcor >= 43) {
return 8;
}
}
}
else if (RsDiff >= 12) {
if (Rs < 1610) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 4;
}
}
else if (Rs >= 1610) {
return 2;
}
}
}
}
else if (tcor >= 75) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (SC < 3261) {
return 2;
}
else if (SC >= 3261) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (negBalance >= 46) {
if (tcor < 77) {
if (ssDCval < 90) {
if (MVDiff < 16) {
if (Rs < 1137) {
if (RsCsDiff < 16) {
if (Rs < 327) {
return 2;
}
else if (Rs >= 327) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (posBalance < 61) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
if (SC < 2016) {
return 4;
}
else if (SC >= 2016) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (CsDiff < 9) {
if (diffAFD < 3) {
if (ssDCval < 77) {
if (Rs < 929) {
return 4;
}
else if (Rs >= 929) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (ssDCval >= 77) {
return 4;
}
}
else if (diffAFD >= 3) {
return 8;
}
}
else if (CsDiff >= 9) {
return 2;
}
}
}
}
else if (Rs >= 1137) {
if (mcTcor < 89) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Cs < 1284) {
return 8;
}
else if (Cs >= 1284) {
if (Rs < 1604) {
return 4;
}
else if (Rs >= 1604) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (posBalance < 58) {
if (Rs < 1718) {
return 8;
}
else if (Rs >= 1718) {
return 4;
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
else if (mcTcor >= 89) {
if (negBalance < 58) {
if (mu_mv_mag_sq < 3) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (Rs < 1757) {
return 8;
}
else if (Rs >= 1757) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1812) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
return 2;
}
}
else if (Rs >= 1812) {
if (Rs < 2111) {
return 4;
}
else if (Rs >= 2111) {
return 2;
}
}
}
}
else if (negBalance >= 58) {
if (Rs < 1462) {
return 2;
}
else if (Rs >= 1462) {
return 4;
}
}
}
}
}
else if (MVDiff >= 16) {
if (mcTcor < 88) {
if (tcor < 57) {
if (MVDiff < 52) {
if (ssDCval < 46) {
if (Cs < 927) {
return 8;
}
else if (Cs >= 927) {
return 2;
}
}
else if (ssDCval >= 46) {
if (Rs < 1019) {
if (mcTcor < 85) {
return 4;
}
else if (mcTcor >= 85) {
if (Rs < 616) {
return 4;
}
else if (Rs >= 616) {
return 8;
}
}
}
else if (Rs >= 1019) {
if (CsDiff < 6) {
if (Rs < 1679) {
return 8;
}
else if (Rs >= 1679) {
if (Rs < 2002) {
return 4;
}
else if (Rs >= 2002) {
return 1;
}
}
}
else if (CsDiff >= 6) {
if (diffRsCsdiff < -14) {
if (Cs < 1240) {
return 2;
}
else if (Cs >= 1240) {
return 8;
}
}
else if (diffRsCsdiff >= -14) {
if (Rs < 1818) {
return 4;
}
else if (Rs >= 1818) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 52) {
if (refDCval < 68) {
if (TSC < 11) {
if (SC < 2107) {
if (diffTSC < 1) {
if (diffRsCsdiff < -23) {
if (Cs < 762) {
return 1;
}
else if (Cs >= 762) {
return 4;
}
}
else if (diffRsCsdiff >= -23) {
return 8;
}
}
else if (diffTSC >= 1) {
if (RsCsDiff < 61) {
return 8;
}
else if (RsCsDiff >= 61) {
return 4;
}
}
}
else if (SC >= 2107) {
if (Rs < 1591) {
return 1;
}
else if (Rs >= 1591) {
return 2;
}
}
}
else if (TSC >= 11) {
if (TSC < 17) {
return 8;
}
else if (TSC >= 17) {
return 1;
}
}
}
else if (refDCval >= 68) {
if (Cs < 693) {
if (RsDiff < 10) {
if (Cs < 552) {
return 4;
}
else if (Cs >= 552) {
return 2;
}
}
else if (RsDiff >= 10) {
if (Rs < 1300) {
return 4;
}
else if (Rs >= 1300) {
return 1;
}
}
}
else if (Cs >= 693) {
if (Rs < 855) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (Rs >= 855) {
if (Rs < 1222) {
if (CsDiff < 9) {
return 4;
}
else if (CsDiff >= 9) {
return 1;
}
}
else if (Rs >= 1222) {
if (Rs < 1903) {
if (Cs < 1137) {
return 2;
}
else if (Cs >= 1137) {
return 8;
}
}
else if (Rs >= 1903) {
if (Rs < 2079) {
return 1;
}
else if (Rs >= 2079) {
return 8;
}
}
}
}
}
}
}
}
else if (tcor >= 57) {
if (AFD < 19) {
if (TSC < 10) {
if (Cs < 1000) {
if (mcTcor < 83) {
if (SC < 1126) {
return 1;
}
else if (SC >= 1126) {
return 8;
}
}
else if (mcTcor >= 83) {
if (SC < 2254) {
if (negBalance < 56) {
if (RsDiff < 5) {
if (diffMVdiffVal < -1) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (negBalance >= 56) {
if (tcor < 68) {
if (Rs < 615) {
if (Rs < 553) {
return 4;
}
else if (Rs >= 553) {
return 2;
}
}
else if (Rs >= 615) {
if (Rs < 953) {
return 8;
}
else if (Rs >= 953) {
return 4;
}
}
}
else if (tcor >= 68) {
return 2;
}
}
}
else if (SC >= 2254) {
return 8;
}
}
}
else if (Cs >= 1000) {
if (mu_mv_mag_sq < 16) {
if (SC < 2923) {
return 2;
}
else if (SC >= 2923) {
return 4;
}
}
else if (mu_mv_mag_sq >= 16) {
return 8;
}
}
}
else if (TSC >= 10) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (SC < 2452) {
if (SC < 2042) {
return 8;
}
else if (SC >= 2042) {
return 2;
}
}
else if (SC >= 2452) {
return 8;
}
}
}
}
else if (AFD >= 19) {
if (negBalance < 68) {
if (mcTcor < 87) {
if (ssDCval < 74) {
if (gchDC < 1) {
if (Cs < 970) {
return 4;
}
else if (Cs >= 970) {
return 1;
}
}
else if (gchDC >= 1) {
if (negBalance < 51) {
if (Rs < 1302) {
return 1;
}
else if (Rs >= 1302) {
return 4;
}
}
else if (negBalance >= 51) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (Rs < 1311) {
return 2;
}
else if (Rs >= 1311) {
if (SC < 2917) {
return 4;
}
else if (SC >= 2917) {
return 2;
}
}
}
}
else if (mcTcor >= 87) {
if (mu_mv_mag_sq < 17) {
if (Rs < 1462) {
return 4;
}
else if (Rs >= 1462) {
return 8;
}
}
else if (mu_mv_mag_sq >= 17) {
return 4;
}
}
}
else if (negBalance >= 68) {
return 8;
}
}
}
}
else if (mcTcor >= 88) {
if (AFD < 23) {
if (MVDiff < 31) {
if (posBalance < 62) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 3) {
if (mcTcor < 90) {
return 2;
}
else if (mcTcor >= 90) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (tcor < 73) {
if (mu_mv_mag_sq < 12) {
if (Rs < 2194) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 4;
}
}
}
else if (Rs >= 2194) {
return 4;
}
}
else if (mu_mv_mag_sq >= 12) {
if (Rs < 496) {
return 8;
}
else if (Rs >= 496) {
if (Rs < 580) {
return 2;
}
else if (Rs >= 580) {
return 1;
}
}
}
}
else if (tcor >= 73) {
if (RsDiff < 13) {
if (negBalance < 58) {
return 4;
}
else if (negBalance >= 58) {
return 2;
}
}
else if (RsDiff >= 13) {
return 2;
}
}
}
}
}
else if (posBalance >= 62) {
if (tcor < 74) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (mu_mv_mag_sq < 7) {
return 2;
}
else if (mu_mv_mag_sq >= 7) {
if (Rs < 894) {
return 2;
}
else if (Rs >= 894) {
return 8;
}
}
}
}
else if (tcor >= 74) {
if (Rs < 1323) {
return 1;
}
else if (Rs >= 1323) {
if (Rs < 1628) {
return 8;
}
else if (Rs >= 1628) {
return 4;
}
}
}
}
}
else if (MVDiff >= 31) {
if (RsDiff < 10) {
if (RsCsDiff < 73) {
if (mu_mv_mag_sq < 5) {
if (Rs < 1014) {
if (posBalance < 64) {
if (diffRsCsdiff < 5) {
if (MVDiff < 74) {
return 4;
}
else if (MVDiff >= 74) {
return 2;
}
}
else if (diffRsCsdiff >= 5) {
if (Rs < 735) {
return 8;
}
else if (Rs >= 735) {
return 2;
}
}
}
else if (posBalance >= 64) {
return 1;
}
}
else if (Rs >= 1014) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (Rs < 1217) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (Rs >= 1217) {
if (Rs < 1606) {
return 8;
}
else if (Rs >= 1606) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 73) {
if (diffRsCsdiff < -1) {
if (negBalance < 52) {
return 4;
}
else if (negBalance >= 52) {
if (diffAFD < 2) {
if (refDCval < 83) {
return 2;
}
else if (refDCval >= 83) {
if (Rs < 790) {
return 2;
}
else if (Rs >= 790) {
return 1;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 970) {
return 8;
}
else if (Rs >= 970) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= -1) {
if (Rs < 1001) {
if (RsCsDiff < 131) {
return 4;
}
else if (RsCsDiff >= 131) {
if (Rs < 907) {
return 8;
}
else if (Rs >= 907) {
return 1;
}
}
}
else if (Rs >= 1001) {
return 1;
}
}
}
}
else if (RsDiff >= 10) {
if (Cs < 1075) {
return 8;
}
else if (Cs >= 1075) {
return 4;
}
}
}
}
else if (AFD >= 23) {
if (Cs < 923) {
if (diffMVdiffVal < -21) {
if (negBalance < 55) {
if (MVDiff < 94) {
return 1;
}
else if (MVDiff >= 94) {
return 8;
}
}
else if (negBalance >= 55) {
if (Cs < 737) {
return 2;
}
else if (Cs >= 737) {
return 1;
}
}
}
else if (diffMVdiffVal >= -21) {
if (Cs < 800) {
if (RsCsDiff < 87) {
if (TSC < 8) {
if (Rs < 580) {
return 8;
}
else if (Rs >= 580) {
return 4;
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (RsCsDiff >= 87) {
if (diffRsCsdiff < -24) {
if (AFD < 28) {
return 4;
}
else if (AFD >= 28) {
return 8;
}
}
else if (diffRsCsdiff >= -24) {
if (mu_mv_mag_sq < 39) {
if (Rs < 620) {
return 4;
}
else if (Rs >= 620) {
return 2;
}
}
else if (mu_mv_mag_sq >= 39) {
if (Cs < 696) {
return 4;
}
else if (Cs >= 696) {
if (Rs < 846) {
return 8;
}
else if (Rs >= 846) {
return 2;
}
}
}
}
}
}
else if (Cs >= 800) {
if (MVDiff < 37) {
return 4;
}
else if (MVDiff >= 37) {
return 2;
}
}
}
}
else if (Cs >= 923) {
if (mu_mv_mag_sq < 2) {
if (Cs < 1013) {
return 4;
}
else if (Cs >= 1013) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (diffMVdiffVal < 13) {
if (diffRsCsdiff < -45) {
if (Rs < 1361) {
return 4;
}
else if (Rs >= 1361) {
return 1;
}
}
else if (diffRsCsdiff >= -45) {
return 4;
}
}
else if (diffMVdiffVal >= 13) {
if (ssDCval < 77) {
if (Rs < 1405) {
return 8;
}
else if (Rs >= 1405) {
return 2;
}
}
else if (ssDCval >= 77) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 90) {
if (MVDiff < 16) {
if (AFD < 15) {
if (posBalance < 61) {
if (AFD < 14) {
if (Rs < 1138) {
return 1;
}
else if (Rs >= 1138) {
return 2;
}
}
else if (AFD >= 14) {
if (Cs < 1272) {
return 4;
}
else if (Cs >= 1272) {
return 8;
}
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (AFD >= 15) {
if (Rs < 855) {
if (Cs < 785) {
if (Rs < 598) {
return 2;
}
else if (Rs >= 598) {
return 1;
}
}
else if (Cs >= 785) {
if (Cs < 850) {
if (Rs < 836) {
return 4;
}
else if (Rs >= 836) {
return 8;
}
}
else if (Cs >= 850) {
return 8;
}
}
}
else if (Rs >= 855) {
if (SC < 2947) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
if (tcor < 73) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (tcor >= 73) {
if (Rs < 901) {
return 2;
}
else if (Rs >= 901) {
return 8;
}
}
}
}
else if (SC >= 2947) {
if (TSC < 10) {
if (Rs < 2005) {
return 2;
}
else if (Rs >= 2005) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
else if (MVDiff >= 16) {
if (refDCval < 100) {
if (AFD < 26) {
if (mu_mv_mag_sq < 37) {
if (AFD < 20) {
if (RsCsDiff < 107) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 1173) {
return 4;
}
else if (Rs >= 1173) {
return 2;
}
}
}
else if (RsCsDiff >= 107) {
return 2;
}
}
else if (AFD >= 20) {
if (tcor < 76) {
if (RsDiff < 7) {
if (Rs < 543) {
return 4;
}
else if (Rs >= 543) {
return 2;
}
}
else if (RsDiff >= 7) {
if (MVDiff < 24) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (MVDiff >= 24) {
if (TSC < 11) {
if (refDCval < 91) {
return 1;
}
else if (refDCval >= 91) {
return 8;
}
}
else if (TSC >= 11) {
if (diffRsCsdiff < 16) {
return 1;
}
else if (diffRsCsdiff >= 16) {
if (Rs < 994) {
return 1;
}
else if (Rs >= 994) {
return 8;
}
}
}
}
}
}
else if (tcor >= 76) {
if (Cs < 864) {
return 1;
}
else if (Cs >= 864) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (RsCsDiff < 108) {
return 4;
}
else if (RsCsDiff >= 108) {
return 1;
}
}
}
else if (AFD >= 26) {
if (tcor < 61) {
if (mcTcor < 85) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
if (Rs < 2080) {
return 4;
}
else if (Rs >= 2080) {
return 2;
}
}
}
}
else if (mcTcor >= 85) {
if (Cs < 613) {
return 4;
}
else if (Cs >= 613) {
return 1;
}
}
}
else if (tcor >= 61) {
if (Rs < 967) {
if (posBalance < 68) {
return 2;
}
else if (posBalance >= 68) {
if (Rs < 638) {
return 1;
}
else if (Rs >= 638) {
return 8;
}
}
}
else if (Rs >= 967) {
if (AFD < 31) {
return 4;
}
else if (AFD >= 31) {
return 2;
}
}
}
}
}
else if (refDCval >= 100) {
if (negBalance < 61) {
if (mcTcor < 84) {
return 4;
}
else if (mcTcor >= 84) {
if (mcTcor < 92) {
if (Cs < 920) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
return 8;
}
}
else if (Cs >= 920) {
if (ssDCval < 110) {
return 2;
}
else if (ssDCval >= 110) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (MVDiff < 44) {
if (Cs < 550) {
return 8;
}
else if (Cs >= 550) {
return 4;
}
}
else if (MVDiff >= 44) {
if (Rs < 1980) {
return 2;
}
else if (Rs >= 1980) {
return 8;
}
}
}
}
}
else if (negBalance >= 61) {
if (mcTcor < 92) {
return 2;
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 45) {
if (Cs < 831) {
return 1;
}
else if (Cs >= 831) {
return 2;
}
}
else if (mu_mv_mag_sq >= 45) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 77) {
if (Cs < 1007) {
if (Rs < 1977) {
if (TSC < 10) {
if (mu_mv_mag_sq < 37) {
if (diffRsCsdiff < -43) {
if (RsDiff < 9) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (RsDiff >= 9) {
return 8;
}
}
else if (diffRsCsdiff >= -43) {
if (Rs < 593) {
if (negBalance < 61) {
return 2;
}
else if (negBalance >= 61) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (Rs < 262) {
return 4;
}
else if (Rs >= 262) {
if (Rs < 330) {
return 8;
}
else if (Rs >= 330) {
return 2;
}
}
}
}
}
else if (Rs >= 593) {
if (RsCsDiff < 41) {
if (Rs < 875) {
if (gchDC < 5) {
return 4;
}
else if (gchDC >= 5) {
return 8;
}
}
else if (Rs >= 875) {
return 8;
}
}
else if (RsCsDiff >= 41) {
if (mu_mv_mag_sq < 2) {
if (Cs < 872) {
if (ssDCval < 67) {
if (Rs < 956) {
return 4;
}
else if (Rs >= 956) {
return 2;
}
}
else if (ssDCval >= 67) {
if (RsCsDiff < 109) {
if (negBalance < 60) {
if (SC < 1666) {
return 2;
}
else if (SC >= 1666) {
if (diffRsCsdiff < -33) {
return 2;
}
else if (diffRsCsdiff >= -33) {
return 1;
}
}
}
else if (negBalance >= 60) {
if (Rs < 948) {
return 1;
}
else if (Rs >= 948) {
return 2;
}
}
}
else if (RsCsDiff >= 109) {
return 2;
}
}
}
else if (Cs >= 872) {
if (RsCsDiff < 57) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
if (Cs < 946) {
return 2;
}
else if (Cs >= 946) {
return 8;
}
}
}
else if (RsCsDiff >= 57) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsCsDiff < 83) {
if (AFD < 16) {
if (MVDiff < 16) {
if (CsDiff < 5) {
if (Cs < 684) {
return 8;
}
else if (Cs >= 684) {
return 2;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (MVDiff >= 16) {
if (RsDiff < 6) {
if (MVDiff < 27) {
return 8;
}
else if (MVDiff >= 27) {
if (Rs < 784) {
return 8;
}
else if (Rs >= 784) {
return 4;
}
}
}
else if (RsDiff >= 6) {
if (tcor < 88) {
if (mcTcor < 91) {
return 2;
}
else if (mcTcor >= 91) {
return 4;
}
}
else if (tcor >= 88) {
return 8;
}
}
}
}
else if (AFD >= 16) {
if (SC < 1759) {
if (ssDCval < 70) {
return 4;
}
else if (ssDCval >= 70) {
if (Cs < 767) {
return 1;
}
else if (Cs >= 767) {
return 4;
}
}
}
else if (SC >= 1759) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (SC < 2112) {
return 8;
}
else if (SC >= 2112) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 83) {
if (mcTcor < 96) {
if (TSC < 9) {
if (Cs < 830) {
if (Cs < 628) {
return 8;
}
else if (Cs >= 628) {
if (SC < 2570) {
return 2;
}
else if (SC >= 2570) {
return 1;
}
}
}
else if (Cs >= 830) {
if (Cs < 970) {
if (Rs < 1138) {
return 8;
}
else if (Rs >= 1138) {
return 4;
}
}
else if (Cs >= 970) {
return 1;
}
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Cs < 758) {
return 1;
}
else if (Cs >= 758) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (Cs < 913) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (Cs >= 913) {
return 4;
}
}
}
}
else if (TSC >= 10) {
if (gchDC < 3) {
if (ssDCval < 90) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (ssDCval >= 90) {
if (Rs < 1127) {
return 8;
}
else if (Rs >= 1127) {
return 1;
}
}
}
else if (gchDC >= 3) {
if (RsDiff < 11) {
if (diffAFD < -6) {
return 2;
}
else if (diffAFD >= -6) {
return 8;
}
}
else if (RsDiff >= 11) {
return 4;
}
}
}
}
else if (Rs >= 1977) {
if (refDCval < 86) {
if (negBalance < 47) {
if (Rs < 2311) {
return 2;
}
else if (Rs >= 2311) {
return 8;
}
}
else if (negBalance >= 47) {
return 2;
}
}
else if (refDCval >= 86) {
if (MVDiff < 27) {
if (Cs < 963) {
return 4;
}
else if (Cs >= 963) {
return 8;
}
}
else if (MVDiff >= 27) {
if (Rs < 2017) {
return 2;
}
else if (Rs >= 2017) {
return 8;
}
}
}
}
}
else if (Cs >= 1007) {
if (Rs < 1342) {
if (mu_mv_mag_sq < 7) {
if (refDCval < 70) {
if (diffMVdiffVal < 1) {
if (Rs < 850) {
return 8;
}
else if (Rs >= 850) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (refDCval >= 70) {
if (TSC < 12) {
if (RsDiff < 6) {
if (mcTcor < 92) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 4;
}
}
else if (mcTcor >= 92) {
if (negBalance < 52) {
if (SC < 2070) {
return 2;
}
else if (SC >= 2070) {
return 4;
}
}
else if (negBalance >= 52) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
}
}
}
else if (RsDiff >= 6) {
if (diffRsCsdiff < 7) {
if (Cs < 1057) {
if (Rs < 914) {
return 2;
}
else if (Rs >= 914) {
return 8;
}
}
else if (Cs >= 1057) {
return 2;
}
}
else if (diffRsCsdiff >= 7) {
if (SC < 2170) {
return 8;
}
else if (SC >= 2170) {
return 1;
}
}
}
}
else if (TSC >= 12) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (Rs < 859) {
if (Rs < 722) {
if (Rs < 463) {
return 4;
}
else if (Rs >= 463) {
return 8;
}
}
else if (Rs >= 722) {
if (Rs < 827) {
return 1;
}
else if (Rs >= 827) {
return 2;
}
}
}
else if (Rs >= 859) {
if (diffRsCsdiff < 18) {
return 8;
}
else if (diffRsCsdiff >= 18) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (Rs < 1050) {
return 1;
}
else if (Rs >= 1050) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1342) {
if (MVDiff < 7) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (MVDiff >= 7) {
if (AFD < 21) {
if (TSC < 9) {
if (SC < 3077) {
if (MVDiff < 14) {
if (diffMVdiffVal < 0) {
if (Cs < 1248) {
return 8;
}
else if (Cs >= 1248) {
return 4;
}
}
else if (diffMVdiffVal >= 0) {
if (Rs < 1508) {
return 8;
}
else if (Rs >= 1508) {
return 1;
}
}
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (SC >= 3077) {
if (Cs < 1215) {
if (Cs < 1057) {
return 1;
}
else if (Cs >= 1057) {
return 2;
}
}
else if (Cs >= 1215) {
if (Rs < 1918) {
return 8;
}
else if (Rs >= 1918) {
return 4;
}
}
}
}
else if (TSC >= 9) {
if (AFD < 17) {
if (mu_mv_mag_sq < 56) {
if (Cs < 1030) {
if (RsCsDiff < 112) {
if (diffRsCsdiff < -21) {
return 2;
}
else if (diffRsCsdiff >= -21) {
return 8;
}
}
else if (RsCsDiff >= 112) {
return 1;
}
}
else if (Cs >= 1030) {
return 8;
}
}
else if (mu_mv_mag_sq >= 56) {
return 1;
}
}
else if (AFD >= 17) {
if (diffRsCsdiff < 24) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (diffTSC < 2) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (Rs < 1931) {
return 2;
}
else if (Rs >= 1931) {
return 8;
}
}
}
else if (diffTSC >= 2) {
return 1;
}
}
}
else if (diffRsCsdiff >= 24) {
return 4;
}
}
}
}
else if (AFD >= 21) {
if (MVDiff < 44) {
return 1;
}
else if (MVDiff >= 44) {
if (Rs < 2109) {
return 8;
}
else if (Rs >= 2109) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (MVDiff < 24) {
if (Rs < 535) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
else if (Rs >= 535) {
if (tcor < 85) {
if (negBalance < 54) {
if (diffMVdiffVal < -20) {
return 1;
}
else if (diffMVdiffVal >= -20) {
if (RsDiff < 12) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 8;
}
}
else if (RsDiff >= 12) {
return 2;
}
}
}
else if (negBalance >= 54) {
if (MVDiff < 9) {
if (Rs < 1738) {
return 4;
}
else if (Rs >= 1738) {
if (Cs < 770) {
return 2;
}
else if (Cs >= 770) {
return 8;
}
}
}
else if (MVDiff >= 9) {
if (mcTcor < 86) {
return 2;
}
else if (mcTcor >= 86) {
if (posBalance < 54) {
if (Cs < 868) {
return 2;
}
else if (Cs >= 868) {
if (Rs < 1783) {
return 8;
}
else if (Rs >= 1783) {
return 2;
}
}
}
else if (posBalance >= 54) {
if (Rs < 611) {
return 2;
}
else if (Rs >= 611) {
return 8;
}
}
}
}
}
}
else if (tcor >= 85) {
if (MVDiff < 12) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 2;
}
}
}
}
else if (MVDiff >= 12) {
if (TSC < 8) {
if (Rs < 1629) {
return 2;
}
else if (Rs >= 1629) {
return 1;
}
}
else if (TSC >= 8) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (ssDCval < 115) {
return 8;
}
else if (ssDCval >= 115) {
if (mu_mv_mag_sq < 38) {
return 1;
}
else if (mu_mv_mag_sq >= 38) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 24) {
if (tcor < 76) {
if (tcor < 39) {
if (diffMVdiffVal < 29) {
return 1;
}
else if (diffMVdiffVal >= 29) {
if (Cs < 1071) {
if (TSC < 19) {
if (MVDiff < 382) {
return 2;
}
else if (MVDiff >= 382) {
return 1;
}
}
else if (TSC >= 19) {
return 4;
}
}
else if (Cs >= 1071) {
return 1;
}
}
}
else if (tcor >= 39) {
if (gchDC < 4) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (diffRsCsdiff < 103) {
if (Rs < 614) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
else if (Rs >= 614) {
if (SC < 3034) {
if (mcTcor < 94) {
if (diffAFD < -7) {
return 4;
}
else if (diffAFD >= -7) {
if (SC < 1989) {
if (SC < 1970) {
return 8;
}
else if (SC >= 1970) {
return 4;
}
}
else if (SC >= 1989) {
return 8;
}
}
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (SC >= 3034) {
return 4;
}
}
}
else if (diffRsCsdiff >= 103) {
if (Cs < 1203) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
return 1;
}
}
else if (Cs >= 1203) {
if (Rs < 1786) {
return 4;
}
else if (Rs >= 1786) {
return 8;
}
}
}
}
}
else if (gchDC >= 4) {
if (mcTcor < 80) {
if (diffMVdiffVal < 24) {
if (SC < 2523) {
return 1;
}
else if (SC >= 2523) {
return 8;
}
}
else if (diffMVdiffVal >= 24) {
return 4;
}
}
else if (mcTcor >= 80) {
if (TSC < 14) {
if (Rs < 1338) {
if (diffRsCsdiff < -34) {
return 1;
}
else if (diffRsCsdiff >= -34) {
return 2;
}
}
else if (Rs >= 1338) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
if (negBalance < 73) {
if (RsDiff < 15) {
return 2;
}
else if (RsDiff >= 15) {
return 1;
}
}
else if (negBalance >= 73) {
return 8;
}
}
}
}
else if (TSC >= 14) {
if (Rs < 1561) {
return 4;
}
else if (Rs >= 1561) {
return 1;
}
}
}
}
}
}
else if (tcor >= 76) {
if (Rs < 1709) {
if (tcor < 87) {
if (mcTcor < 95) {
if (negBalance < 48) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (negBalance >= 48) {
if (diffMVdiffVal < 12) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Cs < 589) {
return 8;
}
else if (Cs >= 589) {
return 1;
}
}
}
else if (diffMVdiffVal >= 12) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (RsDiff < 14) {
if (diffMVdiffVal < -11) {
return 8;
}
else if (diffMVdiffVal >= -11) {
return 2;
}
}
else if (RsDiff >= 14) {
return 1;
}
}
}
else if (tcor >= 87) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (Rs >= 1709) {
if (ssDCval < 151) {
if (tcor < 90) {
if (TSC < 8) {
if (Cs < 809) {
return 1;
}
else if (Cs >= 809) {
return 2;
}
}
else if (TSC >= 8) {
return 1;
}
}
else if (tcor >= 90) {
if (Rs < 1994) {
return 8;
}
else if (Rs >= 1994) {
return 1;
}
}
}
else if (ssDCval >= 151) {
if (Cs < 1140) {
return 8;
}
else if (Cs >= 1140) {
return 2;
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 88) {
if (SC < 2164) {
if (Rs < 959) {
if (refDCval < 64) {
if (MVDiff < 1) {
if (Rs < 543) {
if (ssDCval < 49) {
if (posBalance < 18) {
return 8;
}
else if (posBalance >= 18) {
if (Rs < 440) {
return 1;
}
else if (Rs >= 440) {
if (Rs < 497) {
return 4;
}
else if (Rs >= 497) {
return 1;
}
}
}
}
else if (ssDCval >= 49) {
if (negBalance < 21) {
return 2;
}
else if (negBalance >= 21) {
if (Rs < 419) {
if (Rs < 307) {
return 4;
}
else if (Rs >= 307) {
return 2;
}
}
else if (Rs >= 419) {
return 4;
}
}
}
}
else if (Rs >= 543) {
if (SC < 1671) {
if (tcor < 97) {
if (refDCval < 57) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
else if (refDCval >= 57) {
return 8;
}
}
else if (tcor >= 97) {
if (Rs < 921) {
if (Cs < 1015) {
return 8;
}
else if (Cs >= 1015) {
if (Rs < 586) {
return 4;
}
else if (Rs >= 586) {
return 1;
}
}
}
else if (Rs >= 921) {
if (ssDCval < 57) {
if (Rs < 924) {
return 4;
}
else if (Rs >= 924) {
return 1;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
}
else if (SC >= 1671) {
if (Rs < 648) {
if (Cs < 1076) {
return 1;
}
else if (Cs >= 1076) {
return 4;
}
}
else if (Rs >= 648) {
if (TSCindex < 3) {
if (posBalance < 32) {
if (Rs < 932) {
return 8;
}
else if (Rs >= 932) {
return 1;
}
}
else if (posBalance >= 32) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (Cs < 988) {
return 8;
}
else if (Cs >= 988) {
if (Rs < 812) {
return 1;
}
else if (Rs >= 812) {
if (Cs < 1056) {
return 8;
}
else if (Cs >= 1056) {
return 1;
}
}
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (refDCval < 37) {
if (Rs < 835) {
return 2;
}
else if (Rs >= 835) {
return 8;
}
}
else if (refDCval >= 37) {
if (Rs < 787) {
if (AFD < 8) {
if (negBalance < 28) {
if (mu_mv_mag_sq < 91) {
if (Rs < 721) {
if (MVDiff < 3) {
if (Rs < 624) {
return 8;
}
else if (Rs >= 624) {
return 4;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (Rs >= 721) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (ssDCval < 57) {
if (diffRsCsdiff < -1) {
if (Rs < 411) {
if (Rs < 279) {
return 2;
}
else if (Rs >= 279) {
return 1;
}
}
else if (Rs >= 411) {
return 4;
}
}
else if (diffRsCsdiff >= -1) {
if (Cs < 563) {
return 4;
}
else if (Cs >= 563) {
if (Cs < 794) {
if (Rs < 460) {
return 1;
}
else if (Rs >= 460) {
return 4;
}
}
else if (Cs >= 794) {
return 1;
}
}
}
}
else if (ssDCval >= 57) {
if (refDCval < 59) {
if (Rs < 564) {
return 2;
}
else if (Rs >= 564) {
return 8;
}
}
else if (refDCval >= 59) {
return 4;
}
}
}
}
else if (negBalance >= 28) {
if (refDCval < 50) {
if (Cs < 638) {
if (Cs < 552) {
return 4;
}
else if (Cs >= 552) {
return 1;
}
}
else if (Cs >= 638) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (refDCval >= 50) {
if (Rs < 705) {
if (MVDiff < 16) {
if (diffRsCsdiff < -6) {
if (Rs < 571) {
return 2;
}
else if (Rs >= 571) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
if (diffTSC < -1) {
if (Cs < 733) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (Cs >= 733) {
return 4;
}
}
else if (diffTSC >= -1) {
if (Rs < 238) {
return 4;
}
else if (Rs >= 238) {
if (mu_mv_mag_sq < 25) {
if (tcor < 95) {
if (Rs < 634) {
return 1;
}
else if (Rs >= 634) {
return 2;
}
}
else if (tcor >= 95) {
return 4;
}
}
else if (mu_mv_mag_sq >= 25) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 71) {
if (Rs < 695) {
if (Rs < 590) {
return 4;
}
else if (Rs >= 590) {
return 8;
}
}
else if (Rs >= 695) {
return 1;
}
}
else if (mu_mv_mag_sq >= 71) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 85) {
return 1;
}
else if (mu_mv_mag_sq >= 85) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
return 1;
}
}
}
}
}
}
}
}
else if (MVDiff >= 16) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (Rs >= 705) {
if (RsDiff < 3) {
if (Cs < 876) {
return 8;
}
else if (Cs >= 876) {
if (ssDCval < 58) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (ssDCval >= 58) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (Cs < 937) {
return 4;
}
else if (Cs >= 937) {
return 2;
}
}
}
}
}
}
else if (AFD >= 8) {
if (SC < 1409) {
if (AFD < 19) {
if (MVDiff < 8) {
if (Rs < 661) {
return 2;
}
else if (Rs >= 661) {
return 4;
}
}
else if (MVDiff >= 8) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
return 1;
}
}
}
else if (AFD >= 19) {
if (Cs < 743) {
return 2;
}
else if (Cs >= 743) {
return 4;
}
}
}
else if (SC >= 1409) {
if (Rs < 687) {
return 2;
}
else if (Rs >= 687) {
if (Cs < 844) {
return 4;
}
else if (Cs >= 844) {
return 8;
}
}
}
}
}
else if (Rs >= 787) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (Cs < 718) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
if (Rs < 894) {
return 1;
}
else if (Rs >= 894) {
return 8;
}
}
}
else if (Cs >= 718) {
if (Rs < 903) {
if (AFD < 9) {
if (Cs < 1037) {
if (SC < 1558) {
return 1;
}
else if (SC >= 1558) {
return 4;
}
}
else if (Cs >= 1037) {
return 1;
}
}
else if (AFD >= 9) {
return 8;
}
}
else if (Rs >= 903) {
if (Cs < 1014) {
return 1;
}
else if (Cs >= 1014) {
if (Rs < 925) {
return 8;
}
else if (Rs >= 925) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 64) {
if (RsCsDiff < 2) {
if (Cs < 695) {
if (Rs < 545) {
if (posBalance < 46) {
if (Rs < 312) {
return 1;
}
else if (Rs >= 312) {
return 2;
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (Rs >= 545) {
if (ssDCval < 76) {
if (Rs < 715) {
return 1;
}
else if (Rs >= 715) {
return 2;
}
}
else if (ssDCval >= 76) {
if (Rs < 770) {
return 2;
}
else if (Rs >= 770) {
return 8;
}
}
}
}
else if (Cs >= 695) {
if (ssDCval < 72) {
if (Rs < 732) {
if (Cs < 765) {
if (negBalance < 23) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (negBalance >= 23) {
return 8;
}
}
else if (Cs >= 765) {
return 8;
}
}
else if (Rs >= 732) {
if (Cs < 754) {
return 2;
}
else if (Cs >= 754) {
if (AFD < 3) {
if (SC < 1789) {
return 4;
}
else if (SC >= 1789) {
return 1;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (ssDCval >= 72) {
if (Rs < 655) {
if (SC < 1330) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
else if (SC >= 1330) {
if (Cs < 743) {
return 1;
}
else if (Cs >= 743) {
return 2;
}
}
}
else if (Rs >= 655) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (Rs < 721) {
return 2;
}
else if (Rs >= 721) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 2) {
if (ssDCval < 83) {
if (RsCsDiff < 12) {
if (refDCval < 75) {
if (Rs < 807) {
if (Rs < 459) {
if (Rs < 337) {
return 1;
}
else if (Rs >= 337) {
if (Rs < 444) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 1;
}
}
else if (Rs >= 444) {
return 2;
}
}
}
else if (Rs >= 459) {
if (ssDCval < 67) {
if (Rs < 718) {
if (Cs < 600) {
if (Cs < 576) {
return 4;
}
else if (Cs >= 576) {
return 1;
}
}
else if (Cs >= 600) {
if (MVDiff < 2) {
if (Cs < 768) {
return 1;
}
else if (Cs >= 768) {
return 8;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (Rs >= 718) {
return 4;
}
}
else if (ssDCval >= 67) {
if (Rs < 646) {
if (Cs < 620) {
if (refDCval < 68) {
return 1;
}
else if (refDCval >= 68) {
if (Rs < 563) {
return 8;
}
else if (Rs >= 563) {
return 4;
}
}
}
else if (Cs >= 620) {
return 8;
}
}
else if (Rs >= 646) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (TSC < 3) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
else if (TSC >= 3) {
if (negBalance < 36) {
if (Rs < 773) {
return 1;
}
else if (Rs >= 773) {
if (Cs < 1191) {
if (Rs < 794) {
return 2;
}
else if (Rs >= 794) {
return 4;
}
}
else if (Cs >= 1191) {
return 1;
}
}
}
else if (negBalance >= 36) {
if (Cs < 947) {
return 4;
}
else if (Cs >= 947) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 807) {
if (mcTcor < 98) {
if (MVDiff < 1) {
if (Cs < 1125) {
return 8;
}
else if (Cs >= 1125) {
return 4;
}
}
else if (MVDiff >= 1) {
if (Cs < 1095) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 1;
}
}
else if (Cs >= 1095) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 107) {
if (refDCval < 70) {
return 1;
}
else if (refDCval >= 70) {
return 2;
}
}
else if (mu_mv_mag_sq >= 107) {
if (Cs < 659) {
return 2;
}
else if (Cs >= 659) {
if (Rs < 825) {
return 1;
}
else if (Rs >= 825) {
if (AFD < 4) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 2;
}
}
else if (AFD >= 4) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (mu_mv_mag_sq < 88) {
if (Cs < 782) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
return 4;
}
}
else if (Cs >= 782) {
return 2;
}
}
else if (mu_mv_mag_sq >= 88) {
if (SC < 2007) {
return 8;
}
else if (SC >= 2007) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 12) {
if (Rs < 516) {
if (Cs < 678) {
return 8;
}
else if (Cs >= 678) {
if (mu_mv_mag_sq < 13) {
if (Rs < 485) {
return 2;
}
else if (Rs >= 485) {
return 1;
}
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 481) {
return 1;
}
else if (Rs >= 481) {
return 2;
}
}
}
}
else if (Rs >= 516) {
if (Cs < 1061) {
if (TSC < 5) {
if (posBalance < 57) {
if (MVDiff < 2) {
if (Rs < 811) {
return 1;
}
else if (Rs >= 811) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
return 4;
}
}
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (posBalance >= 57) {
return 4;
}
}
else if (TSC >= 5) {
if (Cs < 584) {
if (Rs < 717) {
return 1;
}
else if (Rs >= 717) {
return 4;
}
}
else if (Cs >= 584) {
if (refDCval < 77) {
if (negBalance < 41) {
return 8;
}
else if (negBalance >= 41) {
return 2;
}
}
else if (refDCval >= 77) {
return 8;
}
}
}
}
else if (Cs >= 1061) {
if (refDCval < 83) {
if (RsCsDiff < 16) {
if (Rs < 806) {
return 2;
}
else if (Rs >= 806) {
if (ssDCval < 78) {
return 4;
}
else if (ssDCval >= 78) {
return 2;
}
}
}
else if (RsCsDiff >= 16) {
return 2;
}
}
else if (refDCval >= 83) {
return 4;
}
}
}
}
}
else if (ssDCval >= 83) {
if (Cs < 893) {
if (TSC < 4) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 576) {
return 2;
}
else if (Rs >= 576) {
return 1;
}
}
}
else if (TSC >= 4) {
if (diffMVdiffVal < -1) {
if (Cs < 777) {
return 8;
}
else if (Cs >= 777) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
return 4;
}
}
}
}
else if (Cs >= 893) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 2;
}
}
}
}
}
}
else if (Rs >= 959) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 69) {
if (ssDCval < 51) {
if (Rs < 1183) {
if (Cs < 593) {
return 1;
}
else if (Cs >= 593) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (Rs >= 1183) {
if (Cs < 755) {
if (Cs < 629) {
if (Rs < 1357) {
return 4;
}
else if (Rs >= 1357) {
return 4;
}
}
else if (Cs >= 629) {
return 4;
}
}
else if (Cs >= 755) {
if (Rs < 1243) {
return 8;
}
else if (Rs >= 1243) {
return 1;
}
}
}
}
else if (ssDCval >= 51) {
if (posBalance < 55) {
if (SC < 1806) {
if (ssDCval < 65) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (ssDCval >= 65) {
if (tcor < 98) {
if (SC < 1727) {
if (Cs < 555) {
return 4;
}
else if (Cs >= 555) {
return 1;
}
}
else if (SC >= 1727) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (tcor >= 98) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
return 2;
}
}
}
}
else if (SC >= 1806) {
if (Cs < 1088) {
if (posBalance < 30) {
if (Cs < 719) {
return 1;
}
else if (Cs >= 719) {
return 8;
}
}
else if (posBalance >= 30) {
if (RsDiff < 3) {
if (MVDiff < 12) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (ssDCval < 61) {
if (Rs < 1231) {
if (refDCval < 51) {
return 8;
}
else if (refDCval >= 51) {
if (Cs < 763) {
if (Cs < 709) {
return 1;
}
else if (Cs >= 709) {
return 8;
}
}
else if (Cs >= 763) {
return 1;
}
}
}
else if (Rs >= 1231) {
if (Cs < 756) {
return 1;
}
else if (Cs >= 756) {
if (Cs < 861) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (Cs >= 861) {
return 1;
}
}
}
}
else if (ssDCval >= 61) {
if (diffTSC < -1) {
if (SC < 1855) {
if (Rs < 1127) {
return 4;
}
else if (Rs >= 1127) {
return 1;
}
}
else if (SC >= 1855) {
return 1;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
}
}
else if (MVDiff >= 12) {
if (Rs < 1248) {
return 8;
}
else if (Rs >= 1248) {
return 1;
}
}
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 62) {
if (Cs < 557) {
return 2;
}
else if (Cs >= 557) {
return 1;
}
}
else if (mu_mv_mag_sq >= 62) {
if (CsDiff < 3) {
if (Cs < 682) {
return 8;
}
else if (Cs >= 682) {
return 1;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
}
else if (Cs >= 1088) {
if (ssDCval < 65) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (Cs < 1177) {
return 4;
}
else if (Cs >= 1177) {
return 8;
}
}
}
else if (ssDCval >= 65) {
if (Rs < 974) {
return 2;
}
else if (Rs >= 974) {
return 8;
}
}
}
}
}
else if (posBalance >= 55) {
if (CsDiff < 3) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (Cs < 704) {
return 8;
}
else if (Cs >= 704) {
return 1;
}
}
}
else if (CsDiff >= 3) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 1125) {
if (refDCval < 75) {
if (mu_mv_mag_sq < 109) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (tcor < 93) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
return 2;
}
}
else if (tcor >= 93) {
if (TSC < 4) {
if (Rs < 993) {
return 8;
}
else if (Rs >= 993) {
return 1;
}
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (RsCsDiff < 7) {
return 2;
}
else if (RsCsDiff >= 7) {
if (Rs < 1078) {
return 1;
}
else if (Rs >= 1078) {
return 4;
}
}
}
}
else if (refDCval >= 75) {
if (mu_mv_mag_sq < 112) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 2;
}
}
else if (mu_mv_mag_sq >= 112) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
return 4;
}
}
}
}
}
else if (Rs >= 1125) {
if (refDCval < 72) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
return 4;
}
}
else if (refDCval >= 72) {
if (Cs < 809) {
if (ssDCval < 73) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (ssDCval >= 73) {
if (MVDiff < 40) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (MVDiff >= 40) {
return 4;
}
}
}
else if (Cs >= 809) {
if (Cs < 939) {
return 8;
}
else if (Cs >= 939) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (SC < 1588) {
return 2;
}
else if (SC >= 1588) {
if (Cs < 698) {
if (posBalance < 34) {
if (Rs < 1309) {
return 1;
}
else if (Rs >= 1309) {
return 8;
}
}
else if (posBalance >= 34) {
return 8;
}
}
else if (Cs >= 698) {
if (Cs < 736) {
if (Rs < 1167) {
return 4;
}
else if (Rs >= 1167) {
return 1;
}
}
else if (Cs >= 736) {
return 8;
}
}
}
}
}
}
else if (SC >= 2164) {
if (Rs < 1787) {
if (Cs < 1083) {
if (mu_mv_mag_sq < 122) {
if (SC < 2453) {
if (Cs < 940) {
if (mu_mv_mag_sq < 71) {
if (ssDCval < 70) {
if (ssDCval < 64) {
if (TSC < 5) {
if (SC < 2424) {
return 1;
}
else if (SC >= 2424) {
return 4;
}
}
else if (TSC >= 5) {
if (Rs < 1527) {
return 4;
}
else if (Rs >= 1527) {
return 2;
}
}
}
else if (ssDCval >= 64) {
return 2;
}
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (mu_mv_mag_sq >= 71) {
if (ssDCval < 71) {
if (ssDCval < 61) {
if (Cs < 840) {
if (mu_mv_mag_sq < 106) {
return 1;
}
else if (mu_mv_mag_sq >= 106) {
return 8;
}
}
else if (Cs >= 840) {
return 8;
}
}
else if (ssDCval >= 61) {
if (SC < 2389) {
return 1;
}
else if (SC >= 2389) {
if (Rs < 1653) {
return 8;
}
else if (Rs >= 1653) {
return 1;
}
}
}
}
else if (ssDCval >= 71) {
if (Rs < 1700) {
return 8;
}
else if (Rs >= 1700) {
if (Rs < 1776) {
return 1;
}
else if (Rs >= 1776) {
return 4;
}
}
}
}
}
else if (Cs >= 940) {
if (Rs < 1310) {
if (ssDCval < 56) {
if (Rs < 1205) {
return 1;
}
else if (Rs >= 1205) {
return 4;
}
}
else if (ssDCval >= 56) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 1178) {
return 8;
}
else if (Rs >= 1178) {
return 1;
}
}
}
}
else if (Rs >= 1310) {
if (Cs < 1021) {
return 1;
}
else if (Cs >= 1021) {
return 4;
}
}
}
}
else if (SC >= 2453) {
if (posBalance < 43) {
if (MVDiff < 11) {
if (ssDCval < 70) {
if (ssDCval < 61) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (posBalance >= 43) {
if (Cs < 789) {
return 1;
}
else if (Cs >= 789) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (Cs < 1000) {
return 8;
}
else if (Cs >= 1000) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (RsCsDiff < 1) {
if (ssDCval < 70) {
if (SC < 2237) {
return 8;
}
else if (SC >= 2237) {
return 1;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (RsCsDiff >= 1) {
if (ssDCval < 49) {
if (SC < 2793) {
if (Cs < 1016) {
if (Rs < 1547) {
return 8;
}
else if (Rs >= 1547) {
return 4;
}
}
else if (Cs >= 1016) {
return 4;
}
}
else if (SC >= 2793) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
if (Cs < 1061) {
return 4;
}
else if (Cs >= 1061) {
return 8;
}
}
}
}
else if (ssDCval >= 49) {
if (RsCsDiff < 7) {
if (Rs < 1335) {
if (CsDiff < 1) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
return 1;
}
}
else if (CsDiff >= 1) {
if (SC < 2181) {
if (Rs < 1281) {
return 1;
}
else if (Rs >= 1281) {
return 8;
}
}
else if (SC >= 2181) {
return 8;
}
}
}
else if (Rs >= 1335) {
if (Rs < 1747) {
return 8;
}
else if (Rs >= 1747) {
return 1;
}
}
}
else if (RsCsDiff >= 7) {
return 4;
}
}
}
}
}
else if (Cs >= 1083) {
if (AFD < 3) {
if (ssDCval < 55) {
if (SC < 2920) {
if (negBalance < 15) {
return 8;
}
else if (negBalance >= 15) {
if (negBalance < 27) {
if (diffRsCsdiff < -3) {
if (posBalance < 35) {
if (Cs < 1150) {
return 4;
}
else if (Cs >= 1150) {
return 4;
}
}
else if (posBalance >= 35) {
return 4;
}
}
else if (diffRsCsdiff >= -3) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (refDCval < 47) {
return 4;
}
else if (refDCval >= 47) {
if (Rs < 1389) {
return 4;
}
else if (Rs >= 1389) {
if (ssDCval < 49) {
if (posBalance < 41) {
if (Cs < 1180) {
if (Rs < 1761) {
if (Rs < 1744) {
if (Rs < 1706) {
if (SC < 2813) {
if (Cs < 1126) {
if (mu_mv_mag_sq < 128) {
if (Cs < 1094) {
return 4;
}
else if (Cs >= 1094) {
if (Cs < 1116) {
return 4;
}
else if (Cs >= 1116) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1678) {
return 4;
}
else if (Rs >= 1678) {
return 4;
}
}
}
else if (Cs >= 1126) {
return 4;
}
}
else if (SC >= 2813) {
if (posBalance < 37) {
if (AFD < 2) {
if (posBalance < 36) {
if (Rs < 1694) {
return 8;
}
else if (Rs >= 1694) {
return 4;
}
}
else if (posBalance >= 36) {
return 8;
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (posBalance >= 37) {
return 4;
}
}
}
else if (Rs >= 1706) {
if (Rs < 1731) {
return 4;
}
else if (Rs >= 1731) {
if (posBalance < 34) {
if (Rs < 1733) {
return 4;
}
else if (Rs >= 1733) {
return 4;
}
}
else if (posBalance >= 34) {
if (Rs < 1739) {
return 4;
}
else if (Rs >= 1739) {
return 4;
}
}
}
}
}
else if (Rs >= 1744) {
if (Cs < 1142) {
if (negBalance < 20) {
if (Cs < 1112) {
return 4;
}
else if (Cs >= 1112) {
return 4;
}
}
else if (negBalance >= 20) {
return 8;
}
}
else if (Cs >= 1142) {
return 8;
}
}
}
else if (Rs >= 1761) {
if (Rs < 1763) {
return 4;
}
else if (Rs >= 1763) {
if (negBalance < 21) {
if (Cs < 1137) {
return 4;
}
else if (Cs >= 1137) {
return 4;
}
}
else if (negBalance >= 21) {
if (Cs < 1095) {
return 8;
}
else if (Cs >= 1095) {
if (Cs < 1123) {
return 4;
}
else if (Cs >= 1123) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1180) {
return 4;
}
}
else if (posBalance >= 41) {
return 4;
}
}
else if (ssDCval >= 49) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (SC >= 2920) {
return 4;
}
}
else if (ssDCval >= 55) {
if (MVDiff < 8) {
if (ssDCval < 70) {
if (diffRsCsdiff < -11) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 4;
}
}
else if (diffRsCsdiff >= -11) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (Rs < 1233) {
return 4;
}
else if (Rs >= 1233) {
return 1;
}
}
}
}
else if (ssDCval >= 70) {
if (tcor < 99) {
if (Rs < 1375) {
if (Cs < 1139) {
return 8;
}
else if (Cs >= 1139) {
return 4;
}
}
else if (Rs >= 1375) {
return 8;
}
}
else if (tcor >= 99) {
if (Cs < 1106) {
return 8;
}
else if (Cs >= 1106) {
return 4;
}
}
}
}
else if (MVDiff >= 8) {
return 4;
}
}
}
else if (AFD >= 3) {
if (Rs < 1399) {
if (mu_mv_mag_sq < 114) {
if (RsDiff < 2) {
if (Rs < 1035) {
if (ssDCval < 75) {
if (Rs < 986) {
if (Rs < 961) {
return 2;
}
else if (Rs >= 961) {
return 4;
}
}
else if (Rs >= 986) {
return 8;
}
}
else if (ssDCval >= 75) {
return 4;
}
}
else if (Rs >= 1035) {
return 8;
}
}
else if (RsDiff >= 2) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (diffMVdiffVal < 1) {
if (ssDCval < 70) {
if (diffRsCsdiff < 3) {
return 4;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 1027) {
return 2;
}
else if (Rs >= 1027) {
if (Rs < 1128) {
return 4;
}
else if (Rs >= 1128) {
return 8;
}
}
}
}
else if (ssDCval >= 70) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 1270) {
return 4;
}
else if (Rs >= 1270) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 1) {
if (negBalance < 42) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (diffMVdiffVal < 7) {
return 2;
}
else if (diffMVdiffVal >= 7) {
return 4;
}
}
}
else if (negBalance >= 42) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (negBalance < 34) {
if (posBalance < 50) {
if (Rs < 1289) {
return 8;
}
else if (Rs >= 1289) {
return 4;
}
}
else if (posBalance >= 50) {
if (Cs < 1276) {
return 4;
}
else if (Cs >= 1276) {
if (Rs < 1153) {
return 8;
}
else if (Rs >= 1153) {
return 4;
}
}
}
}
else if (negBalance >= 34) {
if (RsCsDiff < 11) {
if (Rs < 1112) {
return 4;
}
else if (Rs >= 1112) {
if (Rs < 1323) {
return 1;
}
else if (Rs >= 1323) {
return 4;
}
}
}
else if (RsCsDiff >= 11) {
return 4;
}
}
}
}
else if (Rs >= 1399) {
if (MVDiff < 3) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
if (Rs < 1682) {
return 8;
}
else if (Rs >= 1682) {
return 2;
}
}
}
else if (MVDiff >= 3) {
if (Cs < 1231) {
if (diffAFD < -1) {
if (Rs < 1487) {
return 8;
}
else if (Rs >= 1487) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Cs >= 1231) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1787) {
if (TSC < 4) {
if (ssDCval < 82) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 97) {
return 4;
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 841) {
return 8;
}
else if (Cs >= 841) {
if (Cs < 1146) {
if (refDCval < 48) {
if (Rs < 1820) {
if (Rs < 1812) {
if (Rs < 1797) {
return 8;
}
else if (Rs >= 1797) {
if (posBalance < 36) {
return 4;
}
else if (posBalance >= 36) {
if (Rs < 1802) {
return 4;
}
else if (Rs >= 1802) {
return 8;
}
}
}
}
else if (Rs >= 1812) {
return 8;
}
}
else if (Rs >= 1820) {
if (Rs < 1825) {
return 4;
}
else if (Rs >= 1825) {
return 4;
}
}
}
else if (refDCval >= 48) {
return 4;
}
}
else if (Cs >= 1146) {
if (SC < 3323) {
return 4;
}
else if (SC >= 3323) {
if (Cs < 1241) {
return 4;
}
else if (Cs >= 1241) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 56) {
if (refDCval < 71) {
if (Cs < 1139) {
if (mcTcor < 99) {
if (Cs < 1002) {
return 1;
}
else if (Cs >= 1002) {
if (mu_mv_mag_sq < 94) {
return 4;
}
else if (mu_mv_mag_sq >= 94) {
if (MVDiff < 4) {
if (negBalance < 14) {
if (Rs < 1896) {
return 1;
}
else if (Rs >= 1896) {
return 4;
}
}
else if (negBalance >= 14) {
return 1;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
else if (mcTcor >= 99) {
if (Rs < 1901) {
if (Cs < 999) {
return 1;
}
else if (Cs >= 999) {
if (SC < 2947) {
if (Rs < 1877) {
if (Cs < 1046) {
return 4;
}
else if (Cs >= 1046) {
if (Cs < 1049) {
return 1;
}
else if (Cs >= 1049) {
return 4;
}
}
}
else if (Rs >= 1877) {
return 1;
}
}
else if (SC >= 2947) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
}
else if (Rs >= 1901) {
if (SC < 2873) {
if (Cs < 745) {
return 4;
}
else if (Cs >= 745) {
return 1;
}
}
else if (SC >= 2873) {
return 4;
}
}
}
}
else if (Cs >= 1139) {
if (Rs < 2050) {
return 2;
}
else if (Rs >= 2050) {
if (Rs < 2282) {
return 8;
}
else if (Rs >= 2282) {
return 2;
}
}
}
}
else if (refDCval >= 71) {
if (Cs < 980) {
return 1;
}
else if (Cs >= 980) {
if (AFD < 3) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
return 8;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
else if (ssDCval >= 82) {
if (MVDiff < 2) {
if (Cs < 939) {
if (Rs < 2195) {
return 4;
}
else if (Rs >= 2195) {
return 8;
}
}
else if (Cs >= 939) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Cs < 801) {
return 1;
}
else if (Cs >= 801) {
return 8;
}
}
}
}
else if (TSC >= 4) {
if (Cs < 950) {
if (posBalance < 53) {
if (negBalance < 47) {
if (ssDCval < 69) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (Rs < 2240) {
return 2;
}
else if (Rs >= 2240) {
return 4;
}
}
}
else if (ssDCval >= 69) {
if (diffRsCsdiff < -17) {
if (diffRsCsdiff < -24) {
return 1;
}
else if (diffRsCsdiff >= -24) {
if (Rs < 2193) {
return 2;
}
else if (Rs >= 2193) {
return 1;
}
}
}
else if (diffRsCsdiff >= -17) {
return 1;
}
}
}
else if (negBalance >= 47) {
if (RsDiff < 5) {
if (Cs < 617) {
return 4;
}
else if (Cs >= 617) {
if (Rs < 2023) {
return 8;
}
else if (Rs >= 2023) {
return 4;
}
}
}
else if (RsDiff >= 5) {
if (TSC < 6) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
else if (TSC >= 6) {
if (Cs < 606) {
return 8;
}
else if (Cs >= 606) {
return 2;
}
}
}
}
}
else if (posBalance >= 53) {
if (negBalance < 35) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (negBalance >= 35) {
return 2;
}
}
}
else if (Cs >= 950) {
if (Cs < 1061) {
if (mu_mv_mag_sq < 65) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 2226) {
return 4;
}
else if (Rs >= 2226) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 65) {
if (CsDiff < 3) {
if (SC < 2862) {
return 1;
}
else if (SC >= 2862) {
if (Rs < 2101) {
return 4;
}
else if (Rs >= 2101) {
return 4;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 2039) {
return 8;
}
else if (Rs >= 2039) {
return 1;
}
}
}
}
else if (Cs >= 1061) {
if (Rs < 1804) {
return 1;
}
else if (Rs >= 1804) {
if (Rs < 2174) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
else if (Rs >= 2174) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 88) {
if (mu_mv_mag_sq < 101) {
if (SC < 1302) {
if (negBalance < 70) {
if (Rs < 431) {
if (MVDiff < 6) {
if (Rs < 308) {
if (Rs < 254) {
return 2;
}
else if (Rs >= 254) {
if (Rs < 286) {
return 4;
}
else if (Rs >= 286) {
return 1;
}
}
}
else if (Rs >= 308) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
if (ssDCval < 118) {
if (Rs < 288) {
return 8;
}
else if (Rs >= 288) {
return 2;
}
}
else if (ssDCval >= 118) {
return 4;
}
}
}
}
else if (Rs >= 431) {
if (negBalance < 25) {
if (ssDCval < 139) {
if (mu_mv_mag_sq < 96) {
if (SC < 1004) {
return 2;
}
else if (SC >= 1004) {
return 1;
}
}
else if (mu_mv_mag_sq >= 96) {
if (Cs < 581) {
return 1;
}
else if (Cs >= 581) {
if (Rs < 515) {
return 4;
}
else if (Rs >= 515) {
return 2;
}
}
}
}
else if (ssDCval >= 139) {
if (Rs < 709) {
return 4;
}
else if (Rs >= 709) {
return 8;
}
}
}
else if (negBalance >= 25) {
if (TSC < 4) {
if (ssDCval < 121) {
if (posBalance < 43) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 2;
}
}
else if (posBalance >= 43) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
return 1;
}
}
}
else if (ssDCval >= 121) {
if (AFD < 7) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 80) {
if (SC < 1232) {
if (SC < 1092) {
if (ssDCval < 116) {
if (Rs < 448) {
return 1;
}
else if (Rs >= 448) {
return 8;
}
}
else if (ssDCval >= 116) {
if (Rs < 467) {
return 8;
}
else if (Rs >= 467) {
return 2;
}
}
}
else if (SC >= 1092) {
return 2;
}
}
else if (SC >= 1232) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (Rs < 603) {
return 8;
}
else if (Rs >= 603) {
if (Cs < 604) {
return 8;
}
else if (Cs >= 604) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 70) {
if (AFD < 22) {
if (Rs < 603) {
return 8;
}
else if (Rs >= 603) {
if (Rs < 609) {
return 1;
}
else if (Rs >= 609) {
return 8;
}
}
}
else if (AFD >= 22) {
return 2;
}
}
}
else if (SC >= 1302) {
if (RsCsDiff < 77) {
if (MVDiff < 5) {
if (SC < 3011) {
if (SC < 2739) {
if (Rs < 662) {
if (Rs < 481) {
if (Cs < 871) {
return 4;
}
else if (Cs >= 871) {
return 8;
}
}
else if (Rs >= 481) {
if (TSC < 4) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
return 1;
}
}
else if (TSC >= 4) {
if (ssDCval < 143) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
return 8;
}
}
else if (ssDCval >= 143) {
return 1;
}
}
}
}
else if (Rs >= 662) {
if (ssDCval < 97) {
if (Rs < 1325) {
if (Cs < 863) {
if (Rs < 677) {
return 1;
}
else if (Rs >= 677) {
return 8;
}
}
else if (Cs >= 863) {
if (Rs < 743) {
if (Rs < 723) {
return 4;
}
else if (Rs >= 723) {
return 1;
}
}
else if (Rs >= 743) {
return 4;
}
}
}
else if (Rs >= 1325) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 8;
}
}
}
}
else if (ssDCval >= 97) {
if (Cs < 601) {
if (Rs < 1988) {
return 1;
}
else if (Rs >= 1988) {
return 8;
}
}
else if (Cs >= 601) {
if (SC < 2529) {
if (Rs < 893) {
if (Rs < 678) {
if (Rs < 665) {
return 4;
}
else if (Rs >= 665) {
return 8;
}
}
else if (Rs >= 678) {
if (mcTcor < 98) {
if (Rs < 695) {
return 4;
}
else if (Rs >= 695) {
return 8;
}
}
else if (mcTcor >= 98) {
return 4;
}
}
}
else if (Rs >= 893) {
if (posBalance < 64) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
if (Rs < 1051) {
return 1;
}
else if (Rs >= 1051) {
return 8;
}
}
}
else if (posBalance >= 64) {
return 4;
}
}
}
else if (SC >= 2529) {
if (SC < 2716) {
return 4;
}
else if (SC >= 2716) {
if (SC < 2731) {
return 2;
}
else if (SC >= 2731) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 2739) {
if (ssDCval < 112) {
return 1;
}
else if (ssDCval >= 112) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 2;
}
}
}
}
else if (SC >= 3011) {
return 8;
}
}
else if (MVDiff >= 5) {
if (refDCval < 102) {
if (TSC < 5) {
if (SC < 2371) {
if (SC < 1447) {
return 2;
}
else if (SC >= 1447) {
if (posBalance < 51) {
if (Rs < 1698) {
return 1;
}
else if (Rs >= 1698) {
return 8;
}
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (SC >= 2371) {
return 1;
}
}
else if (TSC >= 5) {
if (tcor < 94) {
if (Cs < 1199) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (RsCsDiff < 63) {
if (Rs < 1626) {
return 2;
}
else if (Rs >= 1626) {
return 4;
}
}
else if (RsCsDiff >= 63) {
return 1;
}
}
}
else if (Cs >= 1199) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
else if (tcor >= 94) {
if (MVDiff < 23) {
if (ssDCval < 99) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 8;
}
}
else if (ssDCval >= 99) {
if (Rs < 1983) {
if (Cs < 1015) {
return 8;
}
else if (Cs >= 1015) {
return 2;
}
}
else if (Rs >= 1983) {
return 4;
}
}
}
else if (MVDiff >= 23) {
if (diffMVdiffVal < -5) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 1761) {
return 4;
}
else if (Rs >= 1761) {
return 1;
}
}
}
else if (diffMVdiffVal >= -5) {
return 1;
}
}
}
}
}
else if (refDCval >= 102) {
if (ssDCval < 158) {
if (posBalance < 67) {
if (mu_mv_mag_sq < 5) {
if (Cs < 965) {
if (negBalance < 50) {
if (ssDCval < 141) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 1515) {
return 4;
}
else if (Rs >= 1515) {
return 1;
}
}
}
else if (ssDCval >= 141) {
return 1;
}
}
else if (negBalance >= 50) {
if (tcor < 90) {
if (Rs < 1393) {
return 8;
}
else if (Rs >= 1393) {
return 2;
}
}
else if (tcor >= 90) {
if (Rs < 761) {
return 8;
}
else if (Rs >= 761) {
return 1;
}
}
}
}
else if (Cs >= 965) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
if (negBalance < 28) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (negBalance >= 28) {
if (TSC < 6) {
if (MVDiff < 8) {
if (SC < 1648) {
if (Rs < 662) {
if (Rs < 552) {
return 8;
}
else if (Rs >= 552) {
return 1;
}
}
else if (Rs >= 662) {
return 2;
}
}
else if (SC >= 1648) {
if (Rs < 612) {
if (Cs < 1199) {
return 8;
}
else if (Cs >= 1199) {
return 1;
}
}
else if (Rs >= 612) {
return 8;
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (TSC >= 6) {
if (ssDCval < 122) {
if (Rs < 2071) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
return 8;
}
}
else if (Rs >= 2071) {
if (Cs < 876) {
return 2;
}
else if (Cs >= 876) {
return 8;
}
}
}
else if (ssDCval >= 122) {
if (AFD < 16) {
if (Rs < 865) {
return 2;
}
else if (Rs >= 865) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 2;
}
}
}
else if (AFD >= 16) {
if (Rs < 1066) {
return 8;
}
else if (Rs >= 1066) {
return 2;
}
}
}
}
}
}
}
else if (posBalance >= 67) {
if (mu_mv_mag_sq < 15) {
return 2;
}
else if (mu_mv_mag_sq >= 15) {
if (AFD < 10) {
if (Rs < 1314) {
return 4;
}
else if (Rs >= 1314) {
return 2;
}
}
else if (AFD >= 10) {
return 8;
}
}
}
}
else if (ssDCval >= 158) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (Cs < 647) {
if (SC < 2298) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (SC >= 2298) {
if (Cs < 641) {
return 1;
}
else if (Cs >= 641) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (Cs >= 647) {
if (tcor < 98) {
if (Rs < 1941) {
if (ssDCval < 107) {
if (CsDiff < 8) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
if (Rs < 1739) {
return 1;
}
else if (Rs >= 1739) {
return 8;
}
}
}
else if (CsDiff >= 8) {
if (ssDCval < 95) {
if (Rs < 721) {
return 2;
}
else if (Rs >= 721) {
return 8;
}
}
else if (ssDCval >= 95) {
return 1;
}
}
}
else if (ssDCval >= 107) {
if (diffMVdiffVal < -4) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
if (mu_mv_mag_sq < 5) {
if (Rs < 1737) {
return 2;
}
else if (Rs >= 1737) {
if (Rs < 1767) {
return 1;
}
else if (Rs >= 1767) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (Rs < 1262) {
return 2;
}
else if (Rs >= 1262) {
if (Rs < 1649) {
return 8;
}
else if (Rs >= 1649) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 1941) {
if (Cs < 975) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
return 1;
}
}
else if (Cs >= 975) {
if (Rs < 2233) {
return 8;
}
else if (Rs >= 2233) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (refDCval < 97) {
return 4;
}
else if (refDCval >= 97) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (ssDCval < 133) {
if (Cs < 1227) {
if (Rs < 350) {
if (ssDCval < 122) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
else if (ssDCval >= 122) {
if (Rs < 324) {
return 2;
}
else if (Rs >= 324) {
return 8;
}
}
}
else if (Rs >= 350) {
if (Cs < 928) {
if (refDCval < 97) {
if (mu_mv_mag_sq < 127) {
if (Rs < 1006) {
if (mu_mv_mag_sq < 112) {
if (Rs < 554) {
return 8;
}
else if (Rs >= 554) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
else if (Rs >= 1006) {
if (ssDCval < 92) {
if (refDCval < 89) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (Rs < 1588) {
return 8;
}
else if (Rs >= 1588) {
return 4;
}
}
}
else if (refDCval >= 89) {
if (Cs < 665) {
return 1;
}
else if (Cs >= 665) {
return 8;
}
}
}
else if (ssDCval >= 92) {
if (negBalance < 36) {
return 4;
}
else if (negBalance >= 36) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 2177) {
if (posBalance < 25) {
return 4;
}
else if (posBalance >= 25) {
return 8;
}
}
else if (Rs >= 2177) {
return 4;
}
}
}
else if (refDCval >= 97) {
if (SC < 2417) {
if (mu_mv_mag_sq < 105) {
if (TSC < 4) {
if (posBalance < 48) {
if (Rs < 1187) {
return 4;
}
else if (Rs >= 1187) {
return 8;
}
}
else if (posBalance >= 48) {
if (Rs < 1004) {
return 1;
}
else if (Rs >= 1004) {
return 8;
}
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 105) {
if (Cs < 756) {
if (diffAFD < -2) {
if (RsCsDiff < 4) {
return 1;
}
else if (RsCsDiff >= 4) {
return 8;
}
}
else if (diffAFD >= -2) {
if (Rs < 523) {
if (ssDCval < 121) {
return 1;
}
else if (ssDCval >= 121) {
return 8;
}
}
else if (Rs >= 523) {
if (ssDCval < 97) {
return 1;
}
else if (ssDCval >= 97) {
return 8;
}
}
}
}
else if (Cs >= 756) {
if (Rs < 369) {
if (Rs < 364) {
return 8;
}
else if (Rs >= 364) {
return 4;
}
}
else if (Rs >= 369) {
return 8;
}
}
}
}
else if (SC >= 2417) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
return 1;
}
}
}
}
}
else if (Cs >= 928) {
if (ssDCval < 121) {
if (Cs < 1225) {
if (mu_mv_mag_sq < 109) {
if (SC < 1816) {
if (Rs < 679) {
if (Rs < 385) {
return 4;
}
else if (Rs >= 385) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
return 8;
}
}
}
else if (Rs >= 679) {
if (Rs < 750) {
return 4;
}
else if (Rs >= 750) {
return 1;
}
}
}
else if (SC >= 1816) {
return 8;
}
}
else if (mu_mv_mag_sq >= 109) {
if (diffRsCsdiff < -50) {
return 2;
}
else if (diffRsCsdiff >= -50) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Cs < 948) {
return 1;
}
else if (Cs >= 948) {
return 8;
}
}
}
}
}
else if (Cs >= 1225) {
return 2;
}
}
else if (ssDCval >= 121) {
if (Rs < 1491) {
if (Cs < 1153) {
return 4;
}
else if (Cs >= 1153) {
return 8;
}
}
else if (Rs >= 1491) {
return 2;
}
}
}
}
}
else if (Cs >= 1227) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1162) {
if (Rs < 977) {
return 8;
}
else if (Rs >= 977) {
return 2;
}
}
else if (Rs >= 1162) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2377) {
if (Rs < 3455) {
if (RsCsDiff < 53) {
if (Cs < 1075) {
if (MVDiff < 8) {
if (SC < 3383) {
if (posBalance < 51) {
if (diffRsCsdiff < -14) {
if (posBalance < 41) {
if (Cs < 494) {
return 4;
}
else if (Cs >= 494) {
return 2;
}
}
else if (posBalance >= 41) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 1;
}
}
}
else if (diffRsCsdiff >= -14) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (AFD < 6) {
if (CsDiff < 1) {
if (MVDiff < 6) {
if (Cs < 604) {
return 1;
}
else if (Cs >= 604) {
return 4;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (AFD >= 6) {
if (Rs < 2651) {
if (mu_mv_mag_sq < 57) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (mu_mv_mag_sq >= 57) {
return 8;
}
}
else if (Rs >= 2651) {
return 8;
}
}
}
}
}
else if (posBalance >= 51) {
if (ssDCval < 86) {
if (Rs < 2621) {
if (negBalance < 35) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
if (SC < 3171) {
return 2;
}
else if (SC >= 3171) {
return 1;
}
}
}
else if (negBalance >= 35) {
return 2;
}
}
else if (Rs >= 2621) {
if (Cs < 367) {
return 1;
}
else if (Cs >= 367) {
return 4;
}
}
}
else if (ssDCval >= 86) {
if (tcor < 96) {
if (Rs < 2637) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (Rs >= 2637) {
if (Cs < 484) {
return 4;
}
else if (Cs >= 484) {
if (Rs < 2702) {
return 4;
}
else if (Rs >= 2702) {
return 4;
}
}
}
}
else if (tcor >= 96) {
return 2;
}
}
}
}
else if (SC >= 3383) {
if (negBalance < 26) {
if (ssDCval < 120) {
if (ssDCval < 81) {
if (negBalance < 21) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
if (Cs < 910) {
return 4;
}
else if (Cs >= 910) {
return 4;
}
}
}
else if (negBalance >= 21) {
return 2;
}
}
else if (ssDCval >= 81) {
if (SC < 4017) {
return 8;
}
else if (SC >= 4017) {
if (mu_mv_mag_sq < 125) {
if (Rs < 3427) {
return 8;
}
else if (Rs >= 3427) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 3166) {
return 4;
}
else if (Rs >= 3166) {
return 2;
}
}
}
}
}
else if (ssDCval >= 120) {
return 2;
}
}
else if (negBalance >= 26) {
if (refDCval < 74) {
if (Rs < 3143) {
if (refDCval < 63) {
if (mu_mv_mag_sq < 103) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (mu_mv_mag_sq >= 103) {
if (Cs < 961) {
return 2;
}
else if (Cs >= 961) {
return 4;
}
}
}
else if (refDCval >= 63) {
if (mcTcor < 95) {
if (Rs < 2846) {
return 4;
}
else if (Rs >= 2846) {
return 2;
}
}
else if (mcTcor >= 95) {
if (posBalance < 44) {
if (diffRsCsdiff < 7) {
return 2;
}
else if (diffRsCsdiff >= 7) {
return 4;
}
}
else if (posBalance >= 44) {
return 2;
}
}
}
}
else if (Rs >= 3143) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
return 4;
}
}
}
else if (refDCval >= 74) {
if (Rs < 2801) {
if (diffRsCsdiff < -17) {
return 2;
}
else if (diffRsCsdiff >= -17) {
if (ssDCval < 88) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (SC < 3581) {
return 1;
}
else if (SC >= 3581) {
return 2;
}
}
}
else if (ssDCval >= 88) {
return 8;
}
}
}
else if (Rs >= 2801) {
if (tcor < 96) {
if (ssDCval < 109) {
if (gchDC < 1) {
if (Rs < 3352) {
return 1;
}
else if (Rs >= 3352) {
return 4;
}
}
else if (gchDC >= 1) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (Cs < 679) {
return 8;
}
else if (Cs >= 679) {
if (Cs < 923) {
return 4;
}
else if (Cs >= 923) {
if (Rs < 3218) {
return 8;
}
else if (Rs >= 3218) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 109) {
return 2;
}
}
else if (tcor >= 96) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 8) {
if (negBalance < 28) {
if (ssDCval < 90) {
if (SC < 3176) {
return 8;
}
else if (SC >= 3176) {
if (Rs < 2442) {
if (Rs < 2422) {
return 4;
}
else if (Rs >= 2422) {
return 8;
}
}
else if (Rs >= 2442) {
return 4;
}
}
}
else if (ssDCval >= 90) {
if (posBalance < 80) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Cs < 881) {
return 4;
}
else if (Cs >= 881) {
return 4;
}
}
}
else if (posBalance >= 80) {
return 2;
}
}
}
else if (negBalance >= 28) {
if (ssDCval < 76) {
if (Cs < 553) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
if (Rs < 2785) {
return 1;
}
else if (Rs >= 2785) {
return 4;
}
}
}
else if (Cs >= 553) {
if (RsDiff < 6) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 1;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (ssDCval >= 76) {
if (mu_mv_mag_sq < 66) {
if (Cs < 961) {
if (mu_mv_mag_sq < 19) {
if (mu_mv_mag_sq < 5) {
if (TSC < 8) {
if (SC < 2888) {
return 8;
}
else if (SC >= 2888) {
return 1;
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 19) {
if (SC < 3565) {
if (diffRsCsdiff < -1) {
return 1;
}
else if (diffRsCsdiff >= -1) {
if (Rs < 3107) {
if (RsCsDiff < 21) {
return 8;
}
else if (RsCsDiff >= 21) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (Cs < 436) {
return 8;
}
else if (Cs >= 436) {
return 1;
}
}
}
}
else if (Rs >= 3107) {
return 1;
}
}
}
else if (SC >= 3565) {
if (Cs < 739) {
return 8;
}
else if (Cs >= 739) {
return 1;
}
}
}
}
else if (Cs >= 961) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (diffMVdiffVal < 1) {
if (Cs < 795) {
return 1;
}
else if (Cs >= 795) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1075) {
if (SC < 4095) {
if (RsCsDiff < 12) {
if (ssDCval < 128) {
return 8;
}
else if (ssDCval >= 128) {
if (Rs < 2470) {
return 8;
}
else if (Rs >= 2470) {
return 1;
}
}
}
else if (RsCsDiff >= 12) {
if (mu_mv_mag_sq < 89) {
if (ssDCval < 85) {
if (Rs < 2558) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Rs >= 2558) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Cs < 1213) {
return 2;
}
else if (Cs >= 1213) {
return 8;
}
}
}
}
else if (ssDCval >= 85) {
if (Rs < 2382) {
return 4;
}
else if (Rs >= 2382) {
if (diffRsCsdiff < -34) {
if (Cs < 1112) {
return 2;
}
else if (Cs >= 1112) {
return 8;
}
}
else if (diffRsCsdiff >= -34) {
if (Cs < 1299) {
return 8;
}
else if (Cs >= 1299) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (ssDCval < 72) {
if (Cs < 1171) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 2;
}
}
else if (Cs >= 1171) {
if (Cs < 1279) {
if (AFD < 6) {
if (mu_mv_mag_sq < 106) {
return 4;
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 2648) {
if (Rs < 2514) {
return 4;
}
else if (Rs >= 2514) {
return 8;
}
}
else if (Rs >= 2648) {
return 8;
}
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 1279) {
return 2;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 2573) {
if (CsDiff < 3) {
if (Rs < 2541) {
return 8;
}
else if (Rs >= 2541) {
return 1;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (Rs >= 2573) {
if (negBalance < 35) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
else if (negBalance >= 35) {
return 8;
}
}
}
}
}
}
else if (SC >= 4095) {
if (MVDiff < 7) {
if (ssDCval < 108) {
if (RsCsDiff < 7) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
return 2;
}
}
else if (RsCsDiff >= 7) {
if (ssDCval < 93) {
if (posBalance < 49) {
if (Rs < 3358) {
if (diffAFD < -2) {
if (Rs < 2870) {
return 1;
}
else if (Rs >= 2870) {
return 2;
}
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Rs >= 3358) {
if (mu_mv_mag_sq < 110) {
return 1;
}
else if (mu_mv_mag_sq >= 110) {
return 2;
}
}
}
else if (posBalance >= 49) {
if (RsDiff < 5) {
if (Rs < 3073) {
return 4;
}
else if (Rs >= 3073) {
return 4;
}
}
else if (RsDiff >= 5) {
if (mu_mv_mag_sq < 65) {
if (Rs < 2978) {
return 1;
}
else if (Rs >= 2978) {
return 2;
}
}
else if (mu_mv_mag_sq >= 65) {
return 2;
}
}
}
}
else if (ssDCval >= 93) {
if (Rs < 3331) {
return 8;
}
else if (Rs >= 3331) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
}
else if (ssDCval >= 108) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
return 2;
}
}
}
else if (MVDiff >= 7) {
if (diffRsCsdiff < -14) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (tcor < 99) {
return 2;
}
else if (tcor >= 99) {
if (Rs < 3198) {
return 1;
}
else if (Rs >= 3198) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -14) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 53) {
if (ssDCval < 91) {
if (posBalance < 50) {
if (negBalance < 45) {
if (refDCval < 58) {
if (SC < 4432) {
if (AFD < 16) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (SC >= 4432) {
return 8;
}
}
else if (refDCval >= 58) {
if (tcor < 89) {
if (CsDiff < 5) {
if (diffRsCsdiff < 70) {
if (negBalance < 34) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (negBalance >= 34) {
return 4;
}
}
else if (diffRsCsdiff >= 70) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
if (diffAFD < 5) {
return 4;
}
else if (diffAFD >= 5) {
return 1;
}
}
}
}
else if (CsDiff >= 5) {
if (ssDCval < 74) {
if (Cs < 1043) {
if (mcTcor < 91) {
if (ssDCval < 70) {
if (Rs < 3070) {
return 2;
}
else if (Rs >= 3070) {
return 4;
}
}
else if (ssDCval >= 70) {
return 4;
}
}
else if (mcTcor >= 91) {
if (posBalance < 47) {
if (tcor < 80) {
return 2;
}
else if (tcor >= 80) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (posBalance >= 47) {
return 4;
}
}
}
else if (Cs >= 1043) {
if (SC < 3850) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
else if (SC >= 3850) {
if (ssDCval < 67) {
if (RsCsDiff < 112) {
return 4;
}
else if (RsCsDiff >= 112) {
return 2;
}
}
else if (ssDCval >= 67) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 2;
}
}
}
}
}
else if (ssDCval >= 74) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
if (Rs < 2436) {
return 8;
}
else if (Rs >= 2436) {
return 1;
}
}
}
}
}
else if (tcor >= 89) {
if (mcTcor < 95) {
if (ssDCval < 66) {
if (CsDiff < 5) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
else if (CsDiff >= 5) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (mcTcor >= 95) {
if (CsDiff < 5) {
if (posBalance < 45) {
if (Rs < 3057) {
if (Cs < 342) {
return 2;
}
else if (Cs >= 342) {
if (ssDCval < 65) {
if (Rs < 2765) {
return 4;
}
else if (Rs >= 2765) {
return 1;
}
}
else if (ssDCval >= 65) {
return 1;
}
}
}
else if (Rs >= 3057) {
return 2;
}
}
else if (posBalance >= 45) {
if (negBalance < 37) {
if (SC < 4126) {
return 4;
}
else if (SC >= 4126) {
return 1;
}
}
else if (negBalance >= 37) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (Rs < 2813) {
return 1;
}
else if (Rs >= 2813) {
return 2;
}
}
}
}
}
else if (CsDiff >= 5) {
if (mu_mv_mag_sq < 52) {
if (Cs < 1116) {
if (tcor < 90) {
return 2;
}
else if (tcor >= 90) {
if (ssDCval < 78) {
return 4;
}
else if (ssDCval >= 78) {
return 2;
}
}
}
else if (Cs >= 1116) {
if (Rs < 2887) {
return 1;
}
else if (Rs >= 2887) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 52) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 45) {
if (negBalance < 51) {
if (tcor < 81) {
if (mcTcor < 94) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (mcTcor >= 94) {
return 4;
}
}
else if (tcor >= 81) {
if (posBalance < 46) {
if (mu_mv_mag_sq < 57) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
if (Rs < 3051) {
return 2;
}
else if (Rs >= 3051) {
if (Rs < 3261) {
return 4;
}
else if (Rs >= 3261) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (Cs < 1021) {
return 4;
}
else if (Cs >= 1021) {
return 2;
}
}
}
else if (posBalance >= 46) {
return 2;
}
}
}
else if (negBalance >= 51) {
if (ssDCval < 58) {
if (Rs < 2481) {
return 8;
}
else if (Rs >= 2481) {
return 4;
}
}
else if (ssDCval >= 58) {
if (CsDiff < 7) {
if (AFD < 9) {
if (Cs < 875) {
return 1;
}
else if (Cs >= 875) {
return 2;
}
}
else if (AFD >= 9) {
return 2;
}
}
else if (CsDiff >= 7) {
return 8;
}
}
}
}
}
else if (posBalance >= 50) {
if (MVDiff < 35) {
if (ssDCval < 61) {
if (MVDiff < 23) {
if (Cs < 983) {
return 2;
}
else if (Cs >= 983) {
if (Rs < 3187) {
return 4;
}
else if (Rs >= 3187) {
if (Rs < 3227) {
return 2;
}
else if (Rs >= 3227) {
return 4;
}
}
}
}
else if (MVDiff >= 23) {
return 2;
}
}
else if (ssDCval >= 61) {
if (CsDiff < 3) {
if (refDCval < 70) {
if (diffMVdiffVal < -3) {
if (Rs < 2701) {
return 2;
}
else if (Rs >= 2701) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
if (Cs < 352) {
return 4;
}
else if (Cs >= 352) {
return 2;
}
}
}
else if (refDCval >= 70) {
if (TSC < 6) {
if (Rs < 2500) {
return 2;
}
else if (Rs >= 2500) {
return 1;
}
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (negBalance < 41) {
if (posBalance < 55) {
if (tcor < 90) {
if (mu_mv_mag_sq < 35) {
if (Cs < 1066) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
if (RsCsDiff < 193) {
return 2;
}
else if (RsCsDiff >= 193) {
return 1;
}
}
}
else if (Cs >= 1066) {
return 1;
}
}
else if (mu_mv_mag_sq >= 35) {
return 2;
}
}
else if (tcor >= 90) {
if (diffMVdiffVal < -3) {
if (Rs < 3174) {
return 1;
}
else if (Rs >= 3174) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
if (refDCval < 63) {
if (Rs < 2489) {
return 2;
}
else if (Rs >= 2489) {
return 4;
}
}
else if (refDCval >= 63) {
return 2;
}
}
}
}
else if (posBalance >= 55) {
if (tcor < 79) {
if (TSC < 9) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 4;
}
}
else if (TSC >= 9) {
return 2;
}
}
else if (tcor >= 79) {
if (mcTcor < 99) {
if (mcTcor < 87) {
return 1;
}
else if (mcTcor >= 87) {
return 2;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
}
else if (negBalance >= 41) {
if (mcTcor < 81) {
return 1;
}
else if (mcTcor >= 81) {
if (mcTcor < 93) {
if (RsCsDiff < 157) {
if (Cs < 1047) {
if (RsCsDiff < 142) {
return 2;
}
else if (RsCsDiff >= 142) {
if (Rs < 2463) {
return 1;
}
else if (Rs >= 2463) {
return 4;
}
}
}
else if (Cs >= 1047) {
return 4;
}
}
else if (RsCsDiff >= 157) {
if (refDCval < 91) {
if (mu_mv_mag_sq < 13) {
return 2;
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 2523) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 4;
}
}
else if (Rs >= 2523) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 8;
}
}
}
}
else if (refDCval >= 91) {
return 4;
}
}
}
else if (mcTcor >= 93) {
if (MVDiff < 32) {
if (RsDiff < 20) {
if (diffAFD < -8) {
if (Rs < 2769) {
return 4;
}
else if (Rs >= 2769) {
return 2;
}
}
else if (diffAFD >= -8) {
return 2;
}
}
else if (RsDiff >= 20) {
if (Cs < 834) {
return 2;
}
else if (Cs >= 834) {
return 4;
}
}
}
else if (MVDiff >= 32) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 35) {
if (Rs < 2578) {
if (Cs < 793) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
if (Cs < 204) {
return 1;
}
else if (Cs >= 204) {
return 2;
}
}
}
else if (Cs >= 793) {
if (CsDiff < 6) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 1;
}
}
else if (CsDiff >= 6) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 2;
}
}
}
}
else if (Rs >= 2578) {
if (CsDiff < 3) {
if (Cs < 148) {
return 2;
}
else if (Cs >= 148) {
if (Rs < 2645) {
return 1;
}
else if (Rs >= 2645) {
return 4;
}
}
}
else if (CsDiff >= 3) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
if (MVDiff < 63) {
if (RsDiff < 20) {
return 1;
}
else if (RsDiff >= 20) {
return 8;
}
}
else if (MVDiff >= 63) {
if (Cs < 503) {
return 2;
}
else if (Cs >= 503) {
if (Rs < 3208) {
return 4;
}
else if (Rs >= 3208) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (mcTcor < 69) {
if (MVDiff < 57) {
return 1;
}
else if (MVDiff >= 57) {
if (Cs < 883) {
if (Rs < 2641) {
return 2;
}
else if (Rs >= 2641) {
return 8;
}
}
else if (Cs >= 883) {
return 4;
}
}
}
else if (mcTcor >= 69) {
if (mcTcor < 95) {
if (RsDiff < 10) {
if (Cs < 791) {
if (negBalance < 54) {
if (diffRsCsdiff < -73) {
return 2;
}
else if (diffRsCsdiff >= -73) {
if (Rs < 3232) {
return 1;
}
else if (Rs >= 3232) {
return 8;
}
}
}
else if (negBalance >= 54) {
return 8;
}
}
else if (Cs >= 791) {
if (diffRsCsdiff < -52) {
if (Rs < 2648) {
return 8;
}
else if (Rs >= 2648) {
return 4;
}
}
else if (diffRsCsdiff >= -52) {
return 8;
}
}
}
else if (RsDiff >= 10) {
if (MVDiff < 29) {
if (ssDCval < 119) {
if (Cs < 838) {
if (negBalance < 42) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 2;
}
}
else if (negBalance >= 42) {
return 2;
}
}
else if (Cs >= 838) {
if (negBalance < 50) {
if (ssDCval < 106) {
return 2;
}
else if (ssDCval >= 106) {
if (Rs < 2810) {
return 1;
}
else if (Rs >= 2810) {
return 4;
}
}
}
else if (negBalance >= 50) {
if (RsCsDiff < 367) {
if (mu_mv_mag_sq < 21) {
if (AFD < 18) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 2;
}
}
}
else if (AFD >= 18) {
if (posBalance < 58) {
if (Rs < 2749) {
return 2;
}
else if (Rs >= 2749) {
return 8;
}
}
else if (posBalance >= 58) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
}
else if (RsCsDiff >= 367) {
if (AFD < 28) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (AFD >= 28) {
return 2;
}
}
}
}
}
else if (ssDCval >= 119) {
if (AFD < 17) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
if (Rs < 2633) {
return 1;
}
else if (Rs >= 2633) {
return 4;
}
}
}
else if (AFD >= 17) {
if (RsDiff < 20) {
if (Cs < 960) {
return 1;
}
else if (Cs >= 960) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 1;
}
}
}
else if (RsDiff >= 20) {
return 2;
}
}
}
}
else if (MVDiff >= 29) {
if (AFD < 27) {
if (mu_mv_mag_sq < 36) {
if (diffAFD < -7) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 1;
}
}
else if (diffAFD >= -7) {
if (RsDiff < 16) {
if (refDCval < 90) {
if (Rs < 2878) {
if (Rs < 2603) {
return 1;
}
else if (Rs >= 2603) {
return 8;
}
}
else if (Rs >= 2878) {
return 2;
}
}
else if (refDCval >= 90) {
if (refDCval < 136) {
if (SC < 2771) {
return 1;
}
else if (SC >= 2771) {
if (mcTcor < 73) {
if (Rs < 2587) {
return 4;
}
else if (Rs >= 2587) {
return 2;
}
}
else if (mcTcor >= 73) {
if (diffRsCsdiff < -61) {
if (mu_mv_mag_sq < 26) {
if (Rs < 2804) {
if (Rs < 2690) {
return 2;
}
else if (Rs >= 2690) {
return 1;
}
}
else if (Rs >= 2804) {
if (Rs < 3382) {
return 4;
}
else if (Rs >= 3382) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 26) {
return 8;
}
}
else if (diffRsCsdiff >= -61) {
if (TSC < 11) {
if (negBalance < 55) {
if (diffAFD < 2) {
if (MVDiff < 68) {
return 1;
}
else if (MVDiff >= 68) {
return 8;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
else if (negBalance >= 55) {
return 8;
}
}
else if (TSC >= 11) {
if (tcor < 58) {
return 2;
}
else if (tcor >= 58) {
if (Cs < 1279) {
return 8;
}
else if (Cs >= 1279) {
if (Rs < 2789) {
return 8;
}
else if (Rs >= 2789) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 136) {
if (MVDiff < 59) {
if (SC < 3721) {
return 1;
}
else if (SC >= 3721) {
return 4;
}
}
else if (MVDiff >= 59) {
return 8;
}
}
}
}
else if (RsDiff >= 16) {
if (refDCval < 112) {
if (negBalance < 61) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (SC < 3191) {
return 8;
}
else if (SC >= 3191) {
return 4;
}
}
}
else if (negBalance >= 61) {
return 2;
}
}
else if (refDCval >= 112) {
if (Cs < 503) {
return 2;
}
else if (Cs >= 503) {
if (RsCsDiff < 287) {
return 8;
}
else if (RsCsDiff >= 287) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 36) {
if (diffAFD < 2) {
if (Cs < 141) {
return 2;
}
else if (Cs >= 141) {
return 4;
}
}
else if (diffAFD >= 2) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 1;
}
}
}
}
else if (AFD >= 27) {
if (gchDC < 2) {
if (SC < 4417) {
if (RsCsDiff < 134) {
return 4;
}
else if (RsCsDiff >= 134) {
return 2;
}
}
else if (SC >= 4417) {
if (Cs < 1185) {
return 4;
}
else if (Cs >= 1185) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (SC < 4296) {
if (Rs < 3288) {
if (MVDiff < 61) {
return 1;
}
else if (MVDiff >= 61) {
if (SC < 3042) {
return 2;
}
else if (SC >= 3042) {
if (SC < 3519) {
return 1;
}
else if (SC >= 3519) {
if (gchDC < 7) {
return 1;
}
else if (gchDC >= 7) {
return 2;
}
}
}
}
}
else if (Rs >= 3288) {
if (Cs < 786) {
return 8;
}
else if (Cs >= 786) {
return 1;
}
}
}
else if (SC >= 4296) {
if (RsDiff < 20) {
return 4;
}
else if (RsDiff >= 20) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (MVDiff < 14) {
if (MVDiff < 7) {
if (mcTcor < 96) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 2725) {
return 1;
}
else if (Rs >= 2725) {
return 2;
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 110) {
if (Rs < 2805) {
return 8;
}
else if (Rs >= 2805) {
if (Cs < 991) {
return 2;
}
else if (Cs >= 991) {
if (Rs < 3222) {
return 2;
}
else if (Rs >= 3222) {
return 4;
}
}
}
}
else if (refDCval >= 110) {
return 2;
}
}
}
else if (MVDiff >= 7) {
if (RsCsDiff < 183) {
if (CsDiff < 4) {
if (Cs < 639) {
if (negBalance < 48) {
if (Rs < 3007) {
return 1;
}
else if (Rs >= 3007) {
if (Cs < 477) {
return 8;
}
else if (Cs >= 477) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (Cs < 562) {
return 2;
}
else if (Cs >= 562) {
return 1;
}
}
}
else if (Cs >= 639) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
if (Rs < 2896) {
if (Rs < 2555) {
return 1;
}
else if (Rs >= 2555) {
return 2;
}
}
else if (Rs >= 2896) {
return 8;
}
}
}
}
else if (CsDiff >= 4) {
if (posBalance < 59) {
if (Cs < 1060) {
if (CsDiff < 5) {
if (Rs < 2655) {
return 8;
}
else if (Rs >= 2655) {
return 2;
}
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (Cs >= 1060) {
return 8;
}
}
else if (posBalance >= 59) {
if (Rs < 2637) {
return 1;
}
else if (Rs >= 2637) {
if (Rs < 3307) {
return 4;
}
else if (Rs >= 3307) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 183) {
if (refDCval < 123) {
if (posBalance < 42) {
if (Rs < 3293) {
return 2;
}
else if (Rs >= 3293) {
return 8;
}
}
else if (posBalance >= 42) {
return 2;
}
}
else if (refDCval >= 123) {
if (RsDiff < 18) {
return 1;
}
else if (RsDiff >= 18) {
return 2;
}
}
}
}
}
else if (MVDiff >= 14) {
if (refDCval < 108) {
if (RsDiff < 11) {
if (ssDCval < 104) {
if (Rs < 2650) {
if (MVDiff < 29) {
return 1;
}
else if (MVDiff >= 29) {
return 8;
}
}
else if (Rs >= 2650) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 1;
}
}
}
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (RsDiff >= 11) {
if (gchDC < 7) {
if (RsCsDiff < 248) {
if (Cs < 517) {
if (Rs < 2646) {
return 2;
}
else if (Rs >= 2646) {
return 8;
}
}
else if (Cs >= 517) {
if (mu_mv_mag_sq < 38) {
return 2;
}
else if (mu_mv_mag_sq >= 38) {
if (Rs < 3117) {
return 8;
}
else if (Rs >= 3117) {
return 2;
}
}
}
}
else if (RsCsDiff >= 248) {
if (Rs < 2398) {
return 1;
}
else if (Rs >= 2398) {
return 2;
}
}
}
else if (gchDC >= 7) {
if (AFD < 30) {
return 1;
}
else if (AFD >= 30) {
return 8;
}
}
}
}
else if (refDCval >= 108) {
if (mu_mv_mag_sq < 35) {
if (RsCsDiff < 94) {
if (diffAFD < -3) {
if (tcor < 93) {
return 2;
}
else if (tcor >= 93) {
return 1;
}
}
else if (diffAFD >= -3) {
if (gchDC < 5) {
return 8;
}
else if (gchDC >= 5) {
if (Rs < 2826) {
return 1;
}
else if (Rs >= 2826) {
return 2;
}
}
}
}
else if (RsCsDiff >= 94) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (CsDiff < 3) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (SC < 3569) {
return 1;
}
else if (SC >= 3569) {
if (diffMVdiffVal < 9) {
if (MVDiff < 55) {
return 1;
}
else if (MVDiff >= 55) {
if (Rs < 2805) {
return 2;
}
else if (Rs >= 2805) {
return 8;
}
}
}
else if (diffMVdiffVal >= 9) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (Rs < 2993) {
return 1;
}
else if (Rs >= 2993) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (ssDCval < 120) {
if (MVDiff < 80) {
if (CsDiff < 3) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
if (Rs < 2998) {
return 2;
}
else if (Rs >= 2998) {
return 8;
}
}
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (MVDiff >= 80) {
return 1;
}
}
else if (ssDCval >= 120) {
if (negBalance < 39) {
if (refDCval < 124) {
if (SC < 3320) {
return 2;
}
else if (SC >= 3320) {
return 8;
}
}
else if (refDCval >= 124) {
return 2;
}
}
else if (negBalance >= 39) {
if (SC < 3329) {
if (refDCval < 127) {
return 2;
}
else if (refDCval >= 127) {
return 1;
}
}
else if (SC >= 3329) {
if (Rs < 2741) {
if (SC < 3811) {
return 2;
}
else if (SC >= 3811) {
return 8;
}
}
else if (Rs >= 2741) {
if (gchDC < 4) {
if (Cs < 645) {
return 2;
}
else if (Cs >= 645) {
return 1;
}
}
else if (gchDC >= 4) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 3455) {
if (refDCval < 92) {
if (posBalance < 48) {
if (Cs < 1146) {
if (TSC < 6) {
if (SC < 5558) {
if (negBalance < 46) {
if (Rs < 3482) {
return 2;
}
else if (Rs >= 3482) {
return 4;
}
}
else if (negBalance >= 46) {
return 2;
}
}
else if (SC >= 5558) {
if (mu_mv_mag_sq < 73) {
return 4;
}
else if (mu_mv_mag_sq >= 73) {
if (Rs < 5934) {
return 1;
}
else if (Rs >= 5934) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 80) {
if (ssDCval < 71) {
if (Rs < 5669) {
return 4;
}
else if (Rs >= 5669) {
return 8;
}
}
else if (ssDCval >= 71) {
if (MVDiff < 7) {
if (Cs < 1062) {
return 4;
}
else if (Cs >= 1062) {
return 1;
}
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (ssDCval >= 80) {
if (Cs < 1074) {
return 2;
}
else if (Cs >= 1074) {
if (Rs < 3899) {
return 4;
}
else if (Rs >= 3899) {
return 4;
}
}
}
}
}
else if (Cs >= 1146) {
if (RsCsDiff < 12) {
if (SC < 4980) {
return 8;
}
else if (SC >= 4980) {
if (SC < 5334) {
if (Cs < 1260) {
return 4;
}
else if (Cs >= 1260) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 3876) {
return 4;
}
else if (Rs >= 3876) {
return 8;
}
}
}
}
else if (SC >= 5334) {
if (Rs < 4576) {
return 8;
}
else if (Rs >= 4576) {
return 2;
}
}
}
}
else if (RsCsDiff >= 12) {
if (ssDCval < 76) {
if (negBalance < 29) {
if (RsCsDiff < 26) {
return 4;
}
else if (RsCsDiff >= 26) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (negBalance >= 29) {
if (Rs < 3821) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
return 1;
}
}
}
else if (Rs >= 3821) {
return 4;
}
}
}
else if (ssDCval >= 76) {
if (Rs < 3831) {
if (diffRsCsdiff < -3) {
return 2;
}
else if (diffRsCsdiff >= -3) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (Rs >= 3831) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
if (Rs < 4046) {
return 2;
}
else if (Rs >= 4046) {
if (Rs < 4287) {
return 4;
}
else if (Rs >= 4287) {
return 4;
}
}
}
}
}
}
}
}
else if (posBalance >= 48) {
if (CsDiff < 3) {
if (MVDiff < 65) {
if (Cs < 685) {
return 4;
}
else if (Cs >= 685) {
if (Rs < 4113) {
return 2;
}
else if (Rs >= 4113) {
return 4;
}
}
}
else if (MVDiff >= 65) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (negBalance < 32) {
if (refDCval < 90) {
if (Cs < 942) {
return 4;
}
else if (Cs >= 942) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (Rs < 5111) {
if (Rs < 3952) {
return 8;
}
else if (Rs >= 3952) {
return 4;
}
}
else if (Rs >= 5111) {
return 2;
}
}
}
}
else if (refDCval >= 90) {
if (Rs < 4021) {
return 2;
}
else if (Rs >= 4021) {
if (Rs < 4310) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 8;
}
}
else if (Rs >= 4310) {
return 4;
}
}
}
}
else if (negBalance >= 32) {
if (MVDiff < 22) {
if (tcor < 79) {
if (RsDiff < 16) {
if (SC < 4710) {
return 4;
}
else if (SC >= 4710) {
return 1;
}
}
else if (RsDiff >= 16) {
return 2;
}
}
else if (tcor >= 79) {
if (diffAFD < -4) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 2;
}
}
else if (diffAFD >= -4) {
return 2;
}
}
}
else if (MVDiff >= 22) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
if (TSC < 11) {
if (diffTSC < -1) {
if (Rs < 3734) {
return 2;
}
else if (Rs >= 3734) {
return 1;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 92) {
if (MVDiff < 7) {
if (refDCval < 108) {
if (Cs < 1016) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 4056) {
return 8;
}
else if (Rs >= 4056) {
return 2;
}
}
}
else if (Cs >= 1016) {
if (refDCval < 100) {
if (negBalance < 32) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
if (tcor < 96) {
if (Cs < 1099) {
if (Rs < 3821) {
return 4;
}
else if (Rs >= 3821) {
return 4;
}
}
else if (Cs >= 1099) {
return 8;
}
}
else if (tcor >= 96) {
return 4;
}
}
}
else if (negBalance >= 32) {
return 2;
}
}
else if (refDCval >= 100) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
if (Cs < 1183) {
if (Rs < 3531) {
return 4;
}
else if (Rs >= 3531) {
return 8;
}
}
else if (Cs >= 1183) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 108) {
if (ssDCval < 142) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (SC < 4336) {
if (TSC < 5) {
if (Rs < 3535) {
return 1;
}
else if (Rs >= 3535) {
return 8;
}
}
else if (TSC >= 5) {
return 2;
}
}
else if (SC >= 4336) {
if (diffMVdiffVal < -12) {
if (Rs < 5339) {
return 1;
}
else if (Rs >= 5339) {
return 2;
}
}
else if (diffMVdiffVal >= -12) {
if (diffMVdiffVal < -3) {
if (ssDCval < 134) {
return 2;
}
else if (ssDCval >= 134) {
if (RsCsDiff < 157) {
return 2;
}
else if (RsCsDiff >= 157) {
if (Rs < 5406) {
return 1;
}
else if (Rs >= 5406) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -3) {
if (Cs < 1283) {
return 2;
}
else if (Cs >= 1283) {
if (Rs < 3924) {
return 8;
}
else if (Rs >= 3924) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 142) {
if (Rs < 3981) {
if (Rs < 3669) {
return 2;
}
else if (Rs >= 3669) {
if (posBalance < 36) {
if (Cs < 1202) {
return 2;
}
else if (Cs >= 1202) {
return 1;
}
}
else if (posBalance >= 36) {
if (Cs < 1108) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (Cs >= 1108) {
return 1;
}
}
}
}
else if (Rs >= 3981) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 1;
}
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 135) {
if (Cs < 588) {
if (Rs < 3888) {
if (diffMVdiffVal < 20) {
if (diffMVdiffVal < -20) {
if (Rs < 3653) {
return 8;
}
else if (Rs >= 3653) {
return 2;
}
}
else if (diffMVdiffVal >= -20) {
if (Rs < 3853) {
return 2;
}
else if (Rs >= 3853) {
return 8;
}
}
}
else if (diffMVdiffVal >= 20) {
if (Rs < 3635) {
return 1;
}
else if (Rs >= 3635) {
return 4;
}
}
}
else if (Rs >= 3888) {
return 2;
}
}
else if (Cs >= 588) {
if (mcTcor < 67) {
if (AFD < 27) {
return 1;
}
else if (AFD >= 27) {
return 2;
}
}
else if (mcTcor >= 67) {
if (ssDCval < 110) {
if (CsDiff < 5) {
if (RsDiff < 16) {
if (mu_mv_mag_sq < 40) {
return 8;
}
else if (mu_mv_mag_sq >= 40) {
if (negBalance < 37) {
if (Cs < 1067) {
if (RsDiff < 7) {
if (AFD < 9) {
if (diffRsCsdiff < -4) {
return 8;
}
else if (diffRsCsdiff >= -4) {
return 4;
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (RsDiff >= 7) {
return 4;
}
}
else if (Cs >= 1067) {
if (Rs < 3689) {
return 8;
}
else if (Rs >= 3689) {
if (Rs < 3852) {
return 8;
}
else if (Rs >= 3852) {
return 4;
}
}
}
}
else if (negBalance >= 37) {
if (Cs < 1170) {
return 2;
}
else if (Cs >= 1170) {
return 4;
}
}
}
}
else if (RsDiff >= 16) {
if (MVDiff < 104) {
if (ssDCval < 109) {
if (AFD < 31) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
return 2;
}
}
else if (AFD >= 31) {
return 4;
}
}
else if (ssDCval >= 109) {
return 8;
}
}
else if (MVDiff >= 104) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (gchDC < 6) {
if (ssDCval < 105) {
if (refDCval < 101) {
if (MVDiff < 66) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
if (mcTcor < 92) {
if (Rs < 3790) {
return 8;
}
else if (Rs >= 3790) {
return 2;
}
}
else if (mcTcor >= 92) {
return 2;
}
}
}
else if (MVDiff >= 66) {
return 1;
}
}
else if (refDCval >= 101) {
if (negBalance < 57) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
if (Rs < 4044) {
return 2;
}
else if (Rs >= 4044) {
return 8;
}
}
}
else if (negBalance >= 57) {
return 2;
}
}
}
else if (ssDCval >= 105) {
if (Cs < 1213) {
return 2;
}
else if (Cs >= 1213) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
if (Cs < 1240) {
return 4;
}
else if (Cs >= 1240) {
return 2;
}
}
}
}
}
else if (gchDC >= 6) {
if (Rs < 4012) {
return 8;
}
else if (Rs >= 4012) {
return 1;
}
}
}
}
else if (ssDCval >= 110) {
if (MVDiff < 23) {
if (TSC < 6) {
if (RsDiff < 5) {
if (Rs < 3517) {
return 2;
}
else if (Rs >= 3517) {
return 1;
}
}
else if (RsDiff >= 5) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (Rs < 4297) {
return 8;
}
else if (Rs >= 4297) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (refDCval < 133) {
if (diffMVdiffVal < -18) {
if (RsCsDiff < 481) {
if (Rs < 5375) {
return 1;
}
else if (Rs >= 5375) {
return 2;
}
}
else if (RsCsDiff >= 481) {
return 2;
}
}
else if (diffMVdiffVal >= -18) {
if (Cs < 1237) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
if (Rs < 4665) {
return 1;
}
else if (Rs >= 4665) {
return 2;
}
}
}
else if (Cs >= 1237) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (AFD < 31) {
return 2;
}
else if (AFD >= 31) {
return 4;
}
}
}
}
}
else if (refDCval >= 133) {
if (AFD < 19) {
if (Rs < 3823) {
if (Cs < 1194) {
return 2;
}
else if (Cs >= 1194) {
return 8;
}
}
else if (Rs >= 3823) {
return 1;
}
}
else if (AFD >= 19) {
return 2;
}
}
}
}
else if (MVDiff >= 23) {
if (CsDiff < 7) {
if (mu_mv_mag_sq < 47) {
if (Rs < 4683) {
if (diffMVdiffVal < 6) {
if (MVDiff < 28) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 2;
}
}
else if (MVDiff >= 28) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
if (mcTcor < 96) {
if (RsDiff < 26) {
if (Rs < 4018) {
return 8;
}
else if (Rs >= 4018) {
if (Rs < 4387) {
return 1;
}
else if (Rs >= 4387) {
return 8;
}
}
}
else if (RsDiff >= 26) {
return 2;
}
}
else if (mcTcor >= 96) {
if (diffRsCsdiff < -11) {
if (Cs < 878) {
return 2;
}
else if (Cs >= 878) {
return 8;
}
}
else if (diffRsCsdiff >= -11) {
return 2;
}
}
}
}
else if (Rs >= 4683) {
if (Cs < 946) {
if (posBalance < 75) {
return 2;
}
else if (posBalance >= 75) {
if (Rs < 4841) {
return 2;
}
else if (Rs >= 4841) {
if (Rs < 5255) {
return 1;
}
else if (Rs >= 5255) {
return 2;
}
}
}
}
else if (Cs >= 946) {
if (diffMVdiffVal < 21) {
if (TSC < 12) {
if (diffMVdiffVal < -16) {
return 2;
}
else if (diffMVdiffVal >= -16) {
if (negBalance < 52) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
if (diffTSC < -3) {
return 8;
}
else if (diffTSC >= -3) {
return 2;
}
}
}
else if (negBalance >= 52) {
return 8;
}
}
}
else if (TSC >= 12) {
return 2;
}
}
else if (diffMVdiffVal >= 21) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 47) {
return 2;
}
}
else if (CsDiff >= 7) {
if (tcor < 69) {
return 2;
}
else if (tcor >= 69) {
if (negBalance < 54) {
if (mu_mv_mag_sq < 7) {
if (Rs < 4367) {
return 2;
}
else if (Rs >= 4367) {
if (MVDiff < 43) {
if (Rs < 6003) {
return 8;
}
else if (Rs >= 6003) {
return 2;
}
}
else if (MVDiff >= 43) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (RsCsDiff < 569) {
return 1;
}
else if (RsCsDiff >= 569) {
return 2;
}
}
}
else if (negBalance >= 54) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 135) {
if (mcTcor < 94) {
if (AFD < 19) {
if (posBalance < 54) {
if (Rs < 6290) {
return 1;
}
else if (Rs >= 6290) {
return 2;
}
}
else if (posBalance >= 54) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
return 2;
}
}
}
else if (AFD >= 19) {
if (posBalance < 47) {
if (Cs < 1115) {
return 2;
}
else if (Cs >= 1115) {
if (Rs < 5624) {
return 8;
}
else if (Rs >= 5624) {
return 2;
}
}
}
else if (posBalance >= 47) {
if (TSC < 17) {
if (ssDCval < 149) {
if (diffMVdiffVal < 9) {
if (MVDiff < 37) {
return 2;
}
else if (MVDiff >= 37) {
return 8;
}
}
else if (diffMVdiffVal >= 9) {
if (Rs < 6021) {
return 1;
}
else if (Rs >= 6021) {
return 2;
}
}
}
else if (ssDCval >= 149) {
return 1;
}
}
else if (TSC >= 17) {
return 4;
}
}
}
}
else if (mcTcor >= 94) {
if (mu_mv_mag_sq < 35) {
if (RsCsDiff < 698) {
if (AFD < 21) {
if (tcor < 94) {
if (RsDiff < 19) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (Rs < 4003) {
return 1;
}
else if (Rs >= 4003) {
return 2;
}
}
}
else if (RsDiff >= 19) {
if (Rs < 4990) {
if (Cs < 867) {
return 2;
}
else if (Cs >= 867) {
return 1;
}
}
else if (Rs >= 4990) {
return 2;
}
}
}
else if (tcor >= 94) {
if (Cs < 1010) {
return 2;
}
else if (Cs >= 1010) {
if (Rs < 4372) {
return 1;
}
else if (Rs >= 4372) {
return 2;
}
}
}
}
else if (AFD >= 21) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
if (RsCsDiff < 315) {
return 8;
}
else if (RsCsDiff >= 315) {
if (Rs < 5750) {
return 1;
}
else if (Rs >= 5750) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 698) {
return 2;
}
}
else if (mu_mv_mag_sq >= 35) {
if (TSC < 5) {
if (Rs < 5675) {
return 8;
}
else if (Rs >= 5675) {
return 2;
}
}
else if (TSC >= 5) {
if (Cs < 928) {
return 2;
}
else if (Cs >= 928) {
if (CsDiff < 4) {
if (diffTSC < -3) {
if (SC < 4835) {
return 2;
}
else if (SC >= 4835) {
return 1;
}
}
else if (diffTSC >= -3) {
return 2;
}
}
else if (CsDiff >= 4) {
if (Rs < 4873) {
return 1;
}
else if (Rs >= 4873) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1313) {
if (mu_mv_mag_sq < 107) {
if (RsCsDiff < 101) {
if (ssDCval < 109) {
if (negBalance < 39) {
if (Rs < 746) {
if (RsDiff < 1) {
if (Rs < 353) {
return 4;
}
else if (Rs >= 353) {
return 8;
}
}
else if (RsDiff >= 1) {
if (TSC < 5) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (TSC >= 5) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Rs < 698) {
return 2;
}
else if (Rs >= 698) {
return 1;
}
}
}
}
}
else if (Rs >= 746) {
if (SC < 10708) {
if (ssDCval < 87) {
if (posBalance < 46) {
if (MVDiff < 7) {
if (Rs < 2870) {
if (ssDCval < 77) {
if (mu_mv_mag_sq < 97) {
if (MVDiff < 4) {
if (Rs < 1877) {
if (Rs < 1149) {
return 4;
}
else if (Rs >= 1149) {
return 8;
}
}
else if (Rs >= 1877) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (MVDiff < 1) {
if (Rs < 2675) {
return 8;
}
else if (Rs >= 2675) {
return 1;
}
}
else if (MVDiff >= 1) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 2;
}
}
}
}
}
else if (MVDiff >= 4) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 97) {
return 8;
}
}
else if (ssDCval >= 77) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
return 8;
}
}
}
}
else if (Rs >= 2870) {
if (CsDiff < 6) {
if (SC < 5664) {
if (Cs < 1994) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (posBalance < 38) {
return 2;
}
else if (posBalance >= 38) {
if (SC < 4801) {
if (Rs < 3138) {
return 4;
}
else if (Rs >= 3138) {
return 2;
}
}
else if (SC >= 4801) {
return 4;
}
}
}
}
else if (Cs >= 1994) {
if (Rs < 2929) {
return 8;
}
else if (Rs >= 2929) {
return 1;
}
}
}
else if (SC >= 5664) {
if (SC < 7015) {
if (MVDiff < 3) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (MVDiff >= 3) {
if (SC < 6216) {
return 4;
}
else if (SC >= 6216) {
return 4;
}
}
}
else if (SC >= 7015) {
if (Cs < 3978) {
return 4;
}
else if (Cs >= 3978) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (Rs < 3459) {
if (Cs < 2962) {
return 1;
}
else if (Cs >= 2962) {
return 8;
}
}
else if (Rs >= 3459) {
if (RsCsDiff < 87) {
if (Cs < 1469) {
return 4;
}
else if (Cs >= 1469) {
return 8;
}
}
else if (RsCsDiff >= 87) {
if (Rs < 4187) {
return 4;
}
else if (Rs >= 4187) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 64) {
if (Rs < 3302) {
if (AFD < 7) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (Rs < 1318) {
return 1;
}
else if (Rs >= 1318) {
return 2;
}
}
}
else if (AFD >= 7) {
if (posBalance < 43) {
if (Rs < 1020) {
return 1;
}
else if (Rs >= 1020) {
return 2;
}
}
else if (posBalance >= 43) {
if (Rs < 1097) {
return 1;
}
else if (Rs >= 1097) {
return 4;
}
}
}
}
else if (Rs >= 3302) {
return 4;
}
}
else if (ssDCval >= 64) {
if (ssDCval < 73) {
return 4;
}
else if (ssDCval >= 73) {
if (SC < 3633) {
return 8;
}
else if (SC >= 3633) {
if (Cs < 1422) {
return 4;
}
else if (Cs >= 1422) {
if (Rs < 3255) {
return 1;
}
else if (Rs >= 3255) {
return 2;
}
}
}
}
}
}
}
else if (posBalance >= 46) {
if (Rs < 2937) {
if (mu_mv_mag_sq < 82) {
if (SC < 2639) {
if (tcor < 90) {
if (posBalance < 59) {
if (TSC < 4) {
if (Rs < 1083) {
return 1;
}
else if (Rs >= 1083) {
return 4;
}
}
else if (TSC >= 4) {
return 4;
}
}
else if (posBalance >= 59) {
if (Cs < 1325) {
return 2;
}
else if (Cs >= 1325) {
return 8;
}
}
}
else if (tcor >= 90) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 1571) {
return 2;
}
else if (Cs >= 1571) {
return 8;
}
}
}
}
else if (SC >= 2639) {
if (MVDiff < 7) {
if (tcor < 92) {
if (ssDCval < 56) {
if (Cs < 1475) {
return 8;
}
else if (Cs >= 1475) {
return 4;
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (tcor >= 92) {
if (SC < 3475) {
return 8;
}
else if (SC >= 3475) {
if (Cs < 1647) {
if (Rs < 2421) {
return 8;
}
else if (Rs >= 2421) {
return 2;
}
}
else if (Cs >= 1647) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (TSC < 6) {
if (negBalance < 36) {
if (gchDC < 2) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 1;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (negBalance >= 36) {
return 8;
}
}
else if (TSC >= 6) {
if (negBalance < 38) {
if (RsCsDiff < 43) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (RsCsDiff >= 43) {
if (mu_mv_mag_sq < 41) {
if (Cs < 2093) {
return 8;
}
else if (Cs >= 2093) {
return 2;
}
}
else if (mu_mv_mag_sq >= 41) {
if (MVDiff < 26) {
return 4;
}
else if (MVDiff >= 26) {
return 8;
}
}
}
}
else if (negBalance >= 38) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (SC < 4747) {
if (Cs < 1605) {
if (mcTcor < 96) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
else if (mcTcor >= 96) {
if (Rs < 2296) {
if (Rs < 1571) {
if (Rs < 1193) {
return 8;
}
else if (Rs >= 1193) {
return 4;
}
}
else if (Rs >= 1571) {
return 8;
}
}
else if (Rs >= 2296) {
return 4;
}
}
}
else if (Cs >= 1605) {
if (Rs < 2680) {
if (MVDiff < 6) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (MVDiff >= 6) {
if (tcor < 96) {
if (mu_mv_mag_sq < 100) {
return 8;
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 1068) {
if (Cs < 1693) {
return 4;
}
else if (Cs >= 1693) {
return 8;
}
}
else if (Rs >= 1068) {
return 4;
}
}
}
else if (tcor >= 96) {
if (refDCval < 62) {
return 8;
}
else if (refDCval >= 62) {
if (Rs < 1593) {
return 1;
}
else if (Rs >= 1593) {
return 8;
}
}
}
}
}
else if (Rs >= 2680) {
if (Cs < 1866) {
if (Cs < 1756) {
return 8;
}
else if (Cs >= 1756) {
return 4;
}
}
else if (Cs >= 1866) {
return 1;
}
}
}
}
else if (SC >= 4747) {
if (MVDiff < 2) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
}
else if (Rs >= 2937) {
if (Cs < 2011) {
if (RsCsDiff < 56) {
if (Cs < 1544) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
if (Rs < 3095) {
return 4;
}
else if (Rs >= 3095) {
return 1;
}
}
}
else if (Cs >= 1544) {
if (Rs < 7350) {
return 4;
}
else if (Rs >= 7350) {
return 2;
}
}
}
else if (RsCsDiff >= 56) {
if (refDCval < 84) {
if (tcor < 91) {
if (Rs < 4350) {
if (Cs < 1572) {
return 2;
}
else if (Cs >= 1572) {
return 8;
}
}
else if (Rs >= 4350) {
return 4;
}
}
else if (tcor >= 91) {
return 2;
}
}
else if (refDCval >= 84) {
if (Cs < 1449) {
return 2;
}
else if (Cs >= 1449) {
return 4;
}
}
}
}
else if (Cs >= 2011) {
if (posBalance < 48) {
if (Rs < 3389) {
return 1;
}
else if (Rs >= 3389) {
if (Rs < 4943) {
return 8;
}
else if (Rs >= 4943) {
return 2;
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 55) {
if (Rs < 3312) {
return 8;
}
else if (Rs >= 3312) {
if (Rs < 4131) {
return 4;
}
else if (Rs >= 4131) {
return 4;
}
}
}
else if (ssDCval >= 55) {
if (Rs < 5406) {
return 8;
}
else if (Rs >= 5406) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (MVDiff < 6) {
if (posBalance < 43) {
if (Rs < 2780) {
if (AFD < 5) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
if (MVDiff < 3) {
if (Rs < 2355) {
return 8;
}
else if (Rs >= 2355) {
if (Rs < 2387) {
return 4;
}
else if (Rs >= 2387) {
return 8;
}
}
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (Rs >= 2780) {
if (RsCsDiff < 46) {
if (ssDCval < 105) {
if (Cs < 1577) {
if (diffTSC < -1) {
if (Rs < 3961) {
return 2;
}
else if (Rs >= 3961) {
return 4;
}
}
else if (diffTSC >= -1) {
if (negBalance < 36) {
return 4;
}
else if (negBalance >= 36) {
if (Rs < 3983) {
return 2;
}
else if (Rs >= 3983) {
return 4;
}
}
}
}
else if (Cs >= 1577) {
if (ssDCval < 91) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
return 2;
}
}
else if (ssDCval >= 91) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
if (Cs < 1653) {
return 8;
}
else if (Cs >= 1653) {
return 4;
}
}
}
}
}
else if (ssDCval >= 105) {
return 8;
}
}
else if (RsCsDiff >= 46) {
if (mu_mv_mag_sq < 88) {
if (Rs < 4180) {
return 2;
}
else if (Rs >= 4180) {
return 4;
}
}
else if (mu_mv_mag_sq >= 88) {
return 8;
}
}
}
}
else if (posBalance >= 43) {
if (Rs < 3420) {
if (Cs < 2705) {
if (Cs < 1515) {
if (Rs < 1420) {
if (Rs < 927) {
return 4;
}
else if (Rs >= 927) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (Rs >= 1420) {
if (mu_mv_mag_sq < 78) {
return 8;
}
else if (mu_mv_mag_sq >= 78) {
if (gchDC < 2) {
if (Cs < 1424) {
return 2;
}
else if (Cs >= 1424) {
return 4;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
else if (Cs >= 1515) {
if (refDCval < 87) {
return 4;
}
else if (refDCval >= 87) {
return 8;
}
}
}
else if (Cs >= 2705) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (Rs >= 3420) {
if (RsCsDiff < 51) {
if (MVDiff < 2) {
if (Rs < 4135) {
return 2;
}
else if (Rs >= 4135) {
return 8;
}
}
else if (MVDiff >= 2) {
if (diffTSC < -1) {
if (Rs < 3677) {
return 4;
}
else if (Rs >= 3677) {
return 2;
}
}
else if (diffTSC >= -1) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (RsCsDiff >= 51) {
if (mcTcor < 95) {
if (Rs < 4281) {
return 1;
}
else if (Rs >= 4281) {
return 4;
}
}
else if (mcTcor >= 95) {
if (diffRsCsdiff < 13) {
return 2;
}
else if (diffRsCsdiff >= 13) {
if (Rs < 4137) {
return 2;
}
else if (Rs >= 4137) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (Rs < 2934) {
if (diffAFD < -3) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 2;
}
}
}
else if (diffAFD >= -3) {
if (refDCval < 103) {
if (posBalance < 38) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
else if (posBalance >= 38) {
if (Cs < 1518) {
if (mu_mv_mag_sq < 70) {
if (Rs < 1990) {
return 8;
}
else if (Rs >= 1990) {
if (Rs < 2468) {
return 2;
}
else if (Rs >= 2468) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (Cs < 1382) {
return 4;
}
else if (Cs >= 1382) {
if (Rs < 1819) {
return 2;
}
else if (Rs >= 1819) {
return 8;
}
}
}
}
else if (Cs >= 1518) {
return 8;
}
}
}
else if (refDCval >= 103) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
}
}
else if (Rs >= 2934) {
if (posBalance < 50) {
if (Rs < 4146) {
if (Cs < 1323) {
return 4;
}
else if (Cs >= 1323) {
return 8;
}
}
else if (Rs >= 4146) {
if (Cs < 1952) {
if (Cs < 1408) {
return 4;
}
else if (Cs >= 1408) {
return 4;
}
}
else if (Cs >= 1952) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (ssDCval < 95) {
if (AFD < 8) {
if (Cs < 1696) {
return 2;
}
else if (Cs >= 1696) {
return 8;
}
}
else if (AFD >= 8) {
return 1;
}
}
else if (ssDCval >= 95) {
if (Rs < 4174) {
return 8;
}
else if (Rs >= 4174) {
return 2;
}
}
}
}
}
}
}
else if (SC >= 10708) {
if (Rs < 5733) {
return 8;
}
else if (Rs >= 5733) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 1456) {
if (ssDCval < 53) {
if (TSC < 6) {
if (Rs < 1072) {
return 2;
}
else if (Rs >= 1072) {
return 8;
}
}
else if (TSC >= 6) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
else if (ssDCval >= 53) {
if (ssDCval < 77) {
if (mcTcor < 96) {
if (negBalance < 44) {
if (Rs < 1297) {
if (Cs < 1508) {
if (diffRsCsdiff < 26) {
return 4;
}
else if (diffRsCsdiff >= 26) {
return 8;
}
}
else if (Cs >= 1508) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 4;
}
}
}
else if (Rs >= 1297) {
if (tcor < 89) {
return 1;
}
else if (tcor >= 89) {
if (Rs < 1357) {
if (Rs < 1325) {
return 1;
}
else if (Rs >= 1325) {
return 4;
}
}
else if (Rs >= 1357) {
if (Rs < 1445) {
return 8;
}
else if (Rs >= 1445) {
return 4;
}
}
}
}
}
else if (negBalance >= 44) {
if (RsCsDiff < 77) {
if (posBalance < 53) {
if (Cs < 1422) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (Rs < 1104) {
return 2;
}
else if (Rs >= 1104) {
return 4;
}
}
}
else if (Cs >= 1422) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
if (Rs < 1173) {
return 4;
}
else if (Rs >= 1173) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
if (Rs < 739) {
if (Rs < 564) {
return 1;
}
else if (Rs >= 564) {
return 4;
}
}
else if (Rs >= 739) {
if (diffRsCsdiff < -23) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (Rs < 1178) {
return 4;
}
else if (Rs >= 1178) {
return 8;
}
}
}
else if (diffRsCsdiff >= -23) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
if (tcor < 88) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (diffMVdiffVal < -2) {
if (Rs < 863) {
return 8;
}
else if (Rs >= 863) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
else if (tcor >= 88) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (posBalance < 58) {
if (RsCsDiff < 98) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1444) {
return 4;
}
else if (Rs >= 1444) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
else if (RsCsDiff >= 98) {
if (Rs < 1275) {
return 2;
}
else if (Rs >= 1275) {
return 1;
}
}
}
else if (posBalance >= 58) {
if (Cs < 1392) {
return 2;
}
else if (Cs >= 1392) {
if (Cs < 1658) {
return 4;
}
else if (Cs >= 1658) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (TSC < 5) {
if (Rs < 1071) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 4;
}
}
else if (Rs >= 1071) {
if (posBalance < 50) {
return 4;
}
else if (posBalance >= 50) {
if (Rs < 1233) {
return 2;
}
else if (Rs >= 1233) {
return 1;
}
}
}
}
else if (TSC >= 5) {
if (CsDiff < 6) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (diffRsCsdiff < -10) {
if (MVDiff < 21) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
else if (MVDiff >= 21) {
return 8;
}
}
else if (diffRsCsdiff >= -10) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
return 1;
}
}
}
}
else if (CsDiff >= 6) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
if (Rs < 769) {
return 2;
}
else if (Rs >= 769) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 77) {
if (mu_mv_mag_sq < 52) {
if (TSC < 7) {
if (mu_mv_mag_sq < 3) {
if (Cs < 1493) {
if (diffRsCsdiff < -5) {
if (Rs < 715) {
return 8;
}
else if (Rs >= 715) {
return 2;
}
}
else if (diffRsCsdiff >= -5) {
if (Rs < 683) {
return 8;
}
else if (Rs >= 683) {
return 4;
}
}
}
else if (Cs >= 1493) {
if (Rs < 1347) {
return 8;
}
else if (Rs >= 1347) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (gchDC < 3) {
if (ssDCval < 81) {
return 2;
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (gchDC >= 3) {
return 2;
}
}
}
else if (TSC >= 7) {
if (negBalance < 52) {
if (Cs < 1531) {
if (AFD < 17) {
if (diffMVdiffVal < 3) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (Cs >= 1531) {
if (RsCsDiff < 95) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (diffMVdiffVal < 8) {
return 4;
}
else if (diffMVdiffVal >= 8) {
if (Rs < 1068) {
return 4;
}
else if (Rs >= 1068) {
return 8;
}
}
}
}
else if (RsCsDiff >= 95) {
if (Rs < 1230) {
return 2;
}
else if (Rs >= 1230) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (CsDiff < 6) {
if (RsDiff < 3) {
if (Rs < 776) {
return 8;
}
else if (Rs >= 776) {
return 1;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (CsDiff >= 6) {
if (diffAFD < 1) {
if (posBalance < 63) {
if (refDCval < 84) {
return 4;
}
else if (refDCval >= 84) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (RsCsDiff < 99) {
return 8;
}
else if (RsCsDiff >= 99) {
return 2;
}
}
}
}
else if (posBalance >= 63) {
return 4;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (Rs < 855) {
if (Cs < 1460) {
return 4;
}
else if (Cs >= 1460) {
return 1;
}
}
else if (Rs >= 855) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (diffRsCsdiff < -25) {
if (Rs < 1247) {
return 4;
}
else if (Rs >= 1247) {
return 8;
}
}
else if (diffRsCsdiff >= -25) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1456) {
if (refDCval < 72) {
if (RsDiff < 7) {
if (negBalance < 41) {
if (diffAFD < 1) {
if (mu_mv_mag_sq < 102) {
return 8;
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 2370) {
return 8;
}
else if (Rs >= 2370) {
return 4;
}
}
}
else if (diffAFD >= 1) {
if (ssDCval < 69) {
if (Cs < 1380) {
return 8;
}
else if (Cs >= 1380) {
return 4;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
}
else if (negBalance >= 41) {
if (Rs < 2132) {
if (ssDCval < 63) {
if (AFD < 13) {
if (Cs < 2727) {
if (Rs < 1495) {
if (Rs < 1476) {
return 8;
}
else if (Rs >= 1476) {
return 4;
}
}
else if (Rs >= 1495) {
return 8;
}
}
else if (Cs >= 2727) {
return 4;
}
}
else if (AFD >= 13) {
if (TSC < 9) {
if (Cs < 1668) {
return 4;
}
else if (Cs >= 1668) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (ssDCval >= 63) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
}
}
else if (Rs >= 2132) {
return 8;
}
}
}
else if (RsDiff >= 7) {
if (posBalance < 46) {
if (Rs < 4237) {
if (Rs < 2482) {
return 8;
}
else if (Rs >= 2482) {
return 2;
}
}
else if (Rs >= 4237) {
return 4;
}
}
else if (posBalance >= 46) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 89) {
if (Rs < 2179) {
if (mcTcor < 94) {
if (Cs < 1518) {
if (posBalance < 52) {
if (Rs < 1864) {
return 1;
}
else if (Rs >= 1864) {
return 8;
}
}
else if (posBalance >= 52) {
return 8;
}
}
else if (Cs >= 1518) {
if (Rs < 1594) {
return 4;
}
else if (Rs >= 1594) {
if (diffRsCsdiff < -14) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (Rs < 1808) {
return 4;
}
else if (Rs >= 1808) {
return 8;
}
}
}
else if (diffRsCsdiff >= -14) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 1651) {
return 8;
}
else if (Rs >= 1651) {
return 4;
}
}
}
else if (Rs >= 2179) {
return 8;
}
}
else if (mu_mv_mag_sq >= 89) {
if (ssDCval < 55) {
if (Cs < 1830) {
if (Rs < 1677) {
return 4;
}
else if (Rs >= 1677) {
return 8;
}
}
else if (Cs >= 1830) {
return 4;
}
}
else if (ssDCval >= 55) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 72) {
if (Cs < 2848) {
if (Rs < 4245) {
if (mcTcor < 94) {
if (AFD < 16) {
if (SC < 3801) {
if (diffRsCsdiff < 27) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (TSC < 9) {
if (Rs < 1932) {
if (negBalance < 58) {
if (diffAFD < -3) {
if (diffRsCsdiff < -14) {
return 8;
}
else if (diffRsCsdiff >= -14) {
return 4;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (negBalance >= 58) {
return 2;
}
}
else if (Rs >= 1932) {
if (mu_mv_mag_sq < 86) {
if (ssDCval < 82) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
return 1;
}
}
else if (ssDCval >= 82) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 86) {
return 8;
}
}
}
else if (TSC >= 9) {
if (ssDCval < 107) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (ssDCval >= 107) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 27) {
if (TSC < 9) {
if (diffTSC < 1) {
if (Rs < 1751) {
return 8;
}
else if (Rs >= 1751) {
return 4;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (TSC >= 9) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
return 8;
}
}
}
}
else if (SC >= 3801) {
if (mu_mv_mag_sq < 92) {
if (diffRsCsdiff < -44) {
if (diffRsCsdiff < -58) {
return 8;
}
else if (diffRsCsdiff >= -58) {
if (refDCval < 84) {
return 8;
}
else if (refDCval >= 84) {
return 4;
}
}
}
else if (diffRsCsdiff >= -44) {
if (RsCsDiff < 99) {
if (Rs < 1760) {
if (Cs < 2327) {
return 8;
}
else if (Cs >= 2327) {
if (Rs < 1657) {
return 2;
}
else if (Rs >= 1657) {
return 4;
}
}
}
else if (Rs >= 1760) {
if (negBalance < 56) {
if (Rs < 2071) {
if (Rs < 2063) {
return 8;
}
else if (Rs >= 2063) {
return 1;
}
}
else if (Rs >= 2071) {
return 8;
}
}
else if (negBalance >= 56) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 8;
}
}
}
}
else if (RsCsDiff >= 99) {
if (refDCval < 83) {
if (mu_mv_mag_sq < 71) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 2384) {
return 8;
}
else if (Rs >= 2384) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 71) {
return 4;
}
}
else if (refDCval >= 83) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (MVDiff < 9) {
if (Rs < 3323) {
if (ssDCval < 81) {
if (RsCsDiff < 62) {
return 8;
}
else if (RsCsDiff >= 62) {
if (Rs < 2255) {
return 8;
}
else if (Rs >= 2255) {
return 4;
}
}
}
else if (ssDCval >= 81) {
if (Rs < 1964) {
return 4;
}
else if (Rs >= 1964) {
return 8;
}
}
}
else if (Rs >= 3323) {
if (Rs < 3673) {
return 4;
}
else if (Rs >= 3673) {
return 1;
}
}
}
else if (MVDiff >= 9) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 8;
}
}
}
}
}
else if (AFD >= 16) {
if (Rs < 1666) {
if (mu_mv_mag_sq < 15) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 8;
}
}
else if (mu_mv_mag_sq >= 15) {
if (Rs < 1662) {
return 8;
}
else if (Rs >= 1662) {
return 4;
}
}
}
else if (Rs >= 1666) {
if (MVDiff < 23) {
if (negBalance < 47) {
return 4;
}
else if (negBalance >= 47) {
if (Cs < 1832) {
return 8;
}
else if (Cs >= 1832) {
if (SC < 4006) {
if (Rs < 1705) {
return 8;
}
else if (Rs >= 1705) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 4;
}
}
}
else if (SC >= 4006) {
return 8;
}
}
}
}
else if (MVDiff >= 23) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (diffMVdiffVal < 4) {
if (Cs < 1674) {
return 8;
}
else if (Cs >= 1674) {
return 1;
}
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (refDCval < 88) {
if (Cs < 1677) {
if (Rs < 2736) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 89) {
if (CsDiff < 4) {
if (mu_mv_mag_sq < 83) {
if (mu_mv_mag_sq < 64) {
return 1;
}
else if (mu_mv_mag_sq >= 64) {
if (Rs < 1995) {
return 1;
}
else if (Rs >= 1995) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 83) {
return 4;
}
}
else if (CsDiff >= 4) {
if (refDCval < 80) {
if (posBalance < 57) {
if (Rs < 2013) {
return 1;
}
else if (Rs >= 2013) {
return 8;
}
}
else if (posBalance >= 57) {
if (Rs < 2079) {
return 4;
}
else if (Rs >= 2079) {
return 8;
}
}
}
else if (refDCval >= 80) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (diffMVdiffVal < 2) {
if (Cs < 1607) {
return 1;
}
else if (Cs >= 1607) {
if (Cs < 1637) {
return 8;
}
else if (Cs >= 1637) {
return 1;
}
}
}
else if (diffMVdiffVal >= 2) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (RsDiff < 6) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (Rs < 2099) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Cs < 1353) {
return 1;
}
else if (Cs >= 1353) {
return 8;
}
}
}
else if (Rs >= 2099) {
return 1;
}
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (MVDiff >= 4) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (TSC < 7) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (Cs < 1390) {
return 8;
}
else if (Cs >= 1390) {
return 2;
}
}
}
else if (TSC >= 7) {
if (mu_mv_mag_sq < 73) {
if (Cs < 1341) {
if (Rs < 1999) {
return 1;
}
else if (Rs >= 1999) {
return 8;
}
}
else if (Cs >= 1341) {
if (tcor < 89) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 1822) {
return 8;
}
else if (Rs >= 1822) {
return 4;
}
}
}
else if (tcor >= 89) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 73) {
if (diffMVdiffVal < -5) {
if (Rs < 2188) {
return 8;
}
else if (Rs >= 2188) {
return 2;
}
}
else if (diffMVdiffVal >= -5) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2736) {
if (AFD < 10) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
if (Rs < 3197) {
return 4;
}
else if (Rs >= 3197) {
return 2;
}
}
}
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (Cs >= 1677) {
if (AFD < 9) {
if (CsDiff < 7) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
if (Cs < 2782) {
if (ssDCval < 75) {
if (Rs < 2083) {
return 4;
}
else if (Rs >= 2083) {
return 8;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
else if (Cs >= 2782) {
if (Rs < 2986) {
return 4;
}
else if (Rs >= 2986) {
return 8;
}
}
}
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (AFD >= 9) {
if (Cs < 2266) {
if (refDCval < 78) {
if (refDCval < 75) {
return 8;
}
else if (refDCval >= 75) {
return 4;
}
}
else if (refDCval >= 78) {
if (AFD < 15) {
if (posBalance < 45) {
if (Rs < 3501) {
return 1;
}
else if (Rs >= 3501) {
return 4;
}
}
else if (posBalance >= 45) {
if (Rs < 2085) {
if (diffAFD < -1) {
if (Rs < 2005) {
return 1;
}
else if (Rs >= 2005) {
return 4;
}
}
else if (diffAFD >= -1) {
if (Cs < 2145) {
return 8;
}
else if (Cs >= 2145) {
return 1;
}
}
}
else if (Rs >= 2085) {
return 8;
}
}
}
else if (AFD >= 15) {
return 4;
}
}
}
else if (Cs >= 2266) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Rs < 3328) {
return 1;
}
else if (Rs >= 3328) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 88) {
if (Rs < 2186) {
if (posBalance < 58) {
if (diffRsCsdiff < 31) {
if (refDCval < 95) {
if (Cs < 2322) {
return 8;
}
else if (Cs >= 2322) {
return 1;
}
}
else if (refDCval >= 95) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 91) {
if (ssDCval < 95) {
if (Cs < 1534) {
return 1;
}
else if (Cs >= 1534) {
return 8;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
else if (mu_mv_mag_sq >= 91) {
if (mu_mv_mag_sq < 100) {
return 2;
}
else if (mu_mv_mag_sq >= 100) {
return 1;
}
}
}
else if (posBalance >= 53) {
if (mcTcor < 97) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (mcTcor >= 97) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
}
else if (diffRsCsdiff >= 31) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (posBalance >= 58) {
if (mcTcor < 96) {
if (TSC < 7) {
if (Cs < 1601) {
return 2;
}
else if (Cs >= 1601) {
if (Rs < 1863) {
return 8;
}
else if (Rs >= 1863) {
return 2;
}
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (mcTcor >= 96) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
if (Rs < 1981) {
return 2;
}
else if (Rs >= 1981) {
if (Rs < 2042) {
return 8;
}
else if (Rs >= 2042) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 2186) {
if (posBalance < 48) {
if (tcor < 93) {
if (mu_mv_mag_sq < 50) {
if (Rs < 2469) {
return 8;
}
else if (Rs >= 2469) {
return 4;
}
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 2537) {
return 4;
}
else if (Rs >= 2537) {
return 8;
}
}
}
else if (tcor >= 93) {
if (AFD < 5) {
if (refDCval < 96) {
if (Rs < 2292) {
if (Rs < 2238) {
return 2;
}
else if (Rs >= 2238) {
return 1;
}
}
else if (Rs >= 2292) {
return 8;
}
}
else if (refDCval >= 96) {
return 4;
}
}
else if (AFD >= 5) {
if (Rs < 3601) {
return 8;
}
else if (Rs >= 3601) {
return 2;
}
}
}
}
else if (posBalance >= 48) {
if (negBalance < 53) {
if (refDCval < 107) {
if (diffMVdiffVal < -43) {
return 4;
}
else if (diffMVdiffVal >= -43) {
return 8;
}
}
else if (refDCval >= 107) {
if (mcTcor < 95) {
if (Cs < 1836) {
return 8;
}
else if (Cs >= 1836) {
return 4;
}
}
else if (mcTcor >= 95) {
return 4;
}
}
}
else if (negBalance >= 53) {
if (mu_mv_mag_sq < 67) {
if (Cs < 2393) {
return 8;
}
else if (Cs >= 2393) {
return 4;
}
}
else if (mu_mv_mag_sq >= 67) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 4245) {
if (mu_mv_mag_sq < 24) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
return 1;
}
}
else if (mu_mv_mag_sq >= 24) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
return 1;
}
}
}
}
else if (Cs >= 2848) {
if (Rs < 2833) {
return 4;
}
else if (Rs >= 2833) {
if (RsCsDiff < 67) {
if (tcor < 91) {
if (Rs < 4251) {
if (Rs < 3081) {
return 4;
}
else if (Rs >= 3081) {
return 1;
}
}
else if (Rs >= 4251) {
return 8;
}
}
else if (tcor >= 91) {
return 8;
}
}
else if (RsCsDiff >= 67) {
if (ssDCval < 84) {
if (negBalance < 50) {
if (Rs < 3921) {
return 1;
}
else if (Rs >= 3921) {
return 8;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (ssDCval >= 84) {
if (refDCval < 100) {
return 4;
}
else if (refDCval >= 100) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 4426) {
if (ssDCval < 163) {
if (Rs < 1488) {
if (mu_mv_mag_sq < 38) {
if (ssDCval < 140) {
if (diffRsCsdiff < 16) {
if (AFD < 19) {
if (negBalance < 41) {
if (Cs < 1568) {
return 8;
}
else if (Cs >= 1568) {
return 4;
}
}
else if (negBalance >= 41) {
return 8;
}
}
else if (AFD >= 19) {
if (Rs < 1012) {
if (Rs < 663) {
return 2;
}
else if (Rs >= 663) {
return 8;
}
}
else if (Rs >= 1012) {
return 4;
}
}
}
else if (diffRsCsdiff >= 16) {
if (AFD < 13) {
if (Cs < 1652) {
return 1;
}
else if (Cs >= 1652) {
return 8;
}
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (ssDCval >= 140) {
if (ssDCval < 147) {
return 1;
}
else if (ssDCval >= 147) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (SC < 2058) {
if (Rs < 693) {
return 1;
}
else if (Rs >= 693) {
return 8;
}
}
else if (SC >= 2058) {
if (Rs < 1262) {
if (MVDiff < 20) {
if (Rs < 1018) {
if (ssDCval < 115) {
if (Cs < 1324) {
return 8;
}
else if (Cs >= 1324) {
return 4;
}
}
else if (ssDCval >= 115) {
if (Cs < 2094) {
if (Rs < 989) {
return 8;
}
else if (Rs >= 989) {
if (Rs < 996) {
return 4;
}
else if (Rs >= 996) {
return 8;
}
}
}
else if (Cs >= 2094) {
return 1;
}
}
}
else if (Rs >= 1018) {
if (ssDCval < 127) {
if (negBalance < 35) {
if (Rs < 1180) {
return 4;
}
else if (Rs >= 1180) {
return 8;
}
}
else if (negBalance >= 35) {
return 8;
}
}
else if (ssDCval >= 127) {
if (negBalance < 38) {
if (mu_mv_mag_sq < 102) {
if (SC < 2687) {
return 8;
}
else if (SC >= 2687) {
return 4;
}
}
else if (mu_mv_mag_sq >= 102) {
return 8;
}
}
else if (negBalance >= 38) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
}
}
}
else if (MVDiff >= 20) {
if (posBalance < 121) {
return 8;
}
else if (posBalance >= 121) {
return 1;
}
}
}
else if (Rs >= 1262) {
if (SC < 3133) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (SC >= 3133) {
if (Rs < 1384) {
return 8;
}
else if (Rs >= 1384) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1488) {
if (refDCval < 140) {
if (mcTcor < 96) {
if (refDCval < 138) {
if (Rs < 2947) {
if (refDCval < 125) {
if (TSC < 9) {
if (ssDCval < 111) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (Rs < 1687) {
return 8;
}
else if (Rs >= 1687) {
return 4;
}
}
}
else if (ssDCval >= 111) {
if (diffRsCsdiff < 46) {
return 8;
}
else if (diffRsCsdiff >= 46) {
return 2;
}
}
}
else if (TSC >= 9) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (refDCval < 123) {
if (Rs < 1595) {
return 1;
}
else if (Rs >= 1595) {
return 8;
}
}
else if (refDCval >= 123) {
return 1;
}
}
}
}
}
else if (refDCval >= 125) {
if (mcTcor < 90) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (posBalance < 53) {
if (Rs < 2020) {
return 1;
}
else if (Rs >= 2020) {
return 8;
}
}
else if (posBalance >= 53) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (refDCval < 128) {
if (RsDiff < 7) {
if (Rs < 2848) {
return 8;
}
else if (Rs >= 2848) {
if (Rs < 2882) {
return 1;
}
else if (Rs >= 2882) {
return 8;
}
}
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (refDCval >= 128) {
if (AFD < 15) {
if (mcTcor < 91) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (mcTcor >= 91) {
return 8;
}
}
else if (AFD >= 15) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2947) {
if (diffAFD < -5) {
if (Cs < 2407) {
return 8;
}
else if (Cs >= 2407) {
return 1;
}
}
else if (diffAFD >= -5) {
return 8;
}
}
}
else if (refDCval >= 138) {
if (Rs < 3334) {
if (tcor < 82) {
return 1;
}
else if (tcor >= 82) {
return 8;
}
}
else if (Rs >= 3334) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (Cs < 1391) {
if (RsCsDiff < 53) {
if (negBalance < 35) {
if (MVDiff < 6) {
if (CsDiff < 3) {
if (Rs < 2594) {
return 8;
}
else if (Rs >= 2594) {
return 2;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (negBalance >= 35) {
if (Rs < 2765) {
return 8;
}
else if (Rs >= 2765) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 8;
}
}
}
}
else if (RsCsDiff >= 53) {
if (Rs < 2054) {
return 2;
}
else if (Rs >= 2054) {
if (Rs < 4060) {
return 8;
}
else if (Rs >= 4060) {
return 2;
}
}
}
}
else if (Cs >= 1391) {
if (mu_mv_mag_sq < 16) {
if (Rs < 2113) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (Rs >= 2113) {
return 8;
}
}
else if (mu_mv_mag_sq >= 16) {
if (posBalance < 35) {
if (Rs < 3903) {
if (tcor < 105) {
return 8;
}
else if (tcor >= 105) {
if (Rs < 2014) {
return 8;
}
else if (Rs >= 2014) {
return 2;
}
}
}
else if (Rs >= 3903) {
if (Rs < 4356) {
return 1;
}
else if (Rs >= 4356) {
return 2;
}
}
}
else if (posBalance >= 35) {
return 8;
}
}
}
}
}
else if (refDCval >= 140) {
if (negBalance < 59) {
if (SC < 5884) {
if (MVDiff < 22) {
if (negBalance < 32) {
if (AFD < 11) {
if (mu_mv_mag_sq < 29) {
return 2;
}
else if (mu_mv_mag_sq >= 29) {
return 8;
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (negBalance >= 32) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1791) {
return 2;
}
else if (Rs >= 1791) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (MVDiff >= 22) {
if (Rs < 2702) {
if (Rs < 1524) {
if (Rs < 1501) {
return 8;
}
else if (Rs >= 1501) {
return 2;
}
}
else if (Rs >= 1524) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
if (Rs < 1925) {
return 4;
}
else if (Rs >= 1925) {
return 8;
}
}
}
}
else if (Rs >= 2702) {
return 2;
}
}
}
else if (SC >= 5884) {
return 1;
}
}
else if (negBalance >= 59) {
if (Rs < 1522) {
return 1;
}
else if (Rs >= 1522) {
if (Rs < 1777) {
return 2;
}
else if (Rs >= 1777) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 1017) {
return 1;
}
else if (Rs >= 1017) {
if (Rs < 2311) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
else if (Rs >= 2311) {
return 8;
}
}
}
}
else if (Rs >= 4426) {
if (Cs < 1851) {
if (tcor < 95) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
else if (tcor >= 95) {
return 2;
}
}
else if (Cs >= 1851) {
if (Cs < 3458) {
if (Rs < 5697) {
if (Rs < 4771) {
if (Rs < 4455) {
return 1;
}
else if (Rs >= 4455) {
return 8;
}
}
else if (Rs >= 4771) {
return 1;
}
}
else if (Rs >= 5697) {
return 4;
}
}
else if (Cs >= 3458) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 101) {
if (CsDiff < 7) {
if (tcor < 89) {
if (RsCsDiff < 165) {
if (refDCval < 90) {
if (negBalance < 48) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (Cs < 1754) {
if (ssDCval < 85) {
if (diffAFD < 1) {
if (TSC < 8) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (TSC >= 8) {
return 2;
}
}
else if (diffAFD >= 1) {
if (Cs < 1463) {
return 4;
}
else if (Cs >= 1463) {
return 8;
}
}
}
else if (ssDCval >= 85) {
return 2;
}
}
else if (Cs >= 1754) {
if (Cs < 3434) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (Cs >= 3434) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (CsDiff < 6) {
if (Rs < 2279) {
return 2;
}
else if (Rs >= 2279) {
if (Rs < 2517) {
return 4;
}
else if (Rs >= 2517) {
return 8;
}
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
}
else if (refDCval >= 90) {
if (Cs < 1356) {
if (tcor < 79) {
if (Rs < 3020) {
return 2;
}
else if (Rs >= 3020) {
return 1;
}
}
else if (tcor >= 79) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (Cs >= 1356) {
if (MVDiff < 28) {
if (diffRsCsdiff < -47) {
if (diffRsCsdiff < -55) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (diffRsCsdiff >= -55) {
return 2;
}
}
else if (diffRsCsdiff >= -47) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
if (diffMVdiffVal < -91) {
return 4;
}
else if (diffMVdiffVal >= -91) {
if (ssDCval < 95) {
if (Cs < 1650) {
return 4;
}
else if (Cs >= 1650) {
return 8;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
else if (MVDiff >= 28) {
if (diffAFD < -6) {
return 1;
}
else if (diffAFD >= -6) {
if (AFD < 23) {
if (mu_mv_mag_sq < 55) {
return 8;
}
else if (mu_mv_mag_sq >= 55) {
return 4;
}
}
else if (AFD >= 23) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 165) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (tcor < 82) {
if (diffMVdiffVal < -4) {
if (mcTcor < 85) {
if (Cs < 1523) {
return 2;
}
else if (Cs >= 1523) {
if (RsDiff < 18) {
return 4;
}
else if (RsDiff >= 18) {
return 1;
}
}
}
else if (mcTcor >= 85) {
if (ssDCval < 92) {
if (Rs < 2941) {
return 8;
}
else if (Rs >= 2941) {
return 2;
}
}
else if (ssDCval >= 92) {
if (Cs < 1372) {
if (Rs < 4104) {
return 4;
}
else if (Rs >= 4104) {
if (Rs < 5122) {
return 1;
}
else if (Rs >= 5122) {
return 8;
}
}
}
else if (Cs >= 1372) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
return 1;
}
}
}
}
}
else if (diffMVdiffVal >= -4) {
if (SC < 4553) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 4;
}
}
}
else if (SC >= 4553) {
if (Rs < 4416) {
if (diffRsCsdiff < -64) {
if (Rs < 3907) {
return 2;
}
else if (Rs >= 3907) {
return 4;
}
}
else if (diffRsCsdiff >= -64) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (Rs >= 4416) {
if (gchDC < 2) {
if (tcor < 65) {
return 4;
}
else if (tcor >= 65) {
return 8;
}
}
else if (gchDC >= 2) {
if (Rs < 5454) {
return 4;
}
else if (Rs >= 5454) {
return 1;
}
}
}
}
}
}
else if (tcor >= 82) {
if (refDCval < 94) {
if (Cs < 2089) {
return 2;
}
else if (Cs >= 2089) {
return 8;
}
}
else if (refDCval >= 94) {
if (RsCsDiff < 227) {
if (negBalance < 55) {
if (MVDiff < 6) {
if (Rs < 4856) {
return 4;
}
else if (Rs >= 4856) {
return 2;
}
}
else if (MVDiff >= 6) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (Rs < 4053) {
return 4;
}
else if (Rs >= 4053) {
return 8;
}
}
}
}
else if (negBalance >= 55) {
return 4;
}
}
else if (RsCsDiff >= 227) {
if (MVDiff < 15) {
if (posBalance < 63) {
return 2;
}
else if (posBalance >= 63) {
return 4;
}
}
else if (MVDiff >= 15) {
if (refDCval < 114) {
if (Rs < 2879) {
return 8;
}
else if (Rs >= 2879) {
return 2;
}
}
else if (refDCval >= 114) {
return 8;
}
}
}
}
}
}
}
}
else if (tcor >= 89) {
if (Rs < 2854) {
if (Rs < 2278) {
if (Rs < 2068) {
return 8;
}
else if (Rs >= 2068) {
return 1;
}
}
else if (Rs >= 2278) {
return 8;
}
}
else if (Rs >= 2854) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Cs < 2176) {
if (ssDCval < 112) {
if (Rs < 3244) {
if (ssDCval < 95) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
if (Cs < 1739) {
return 2;
}
else if (Cs >= 1739) {
return 1;
}
}
}
else if (ssDCval >= 95) {
if (Rs < 3208) {
return 4;
}
else if (Rs >= 3208) {
return 8;
}
}
}
else if (Rs >= 3244) {
if (negBalance < 48) {
if (Cs < 1956) {
return 2;
}
else if (Cs >= 1956) {
if (Rs < 4388) {
if (Rs < 4055) {
return 2;
}
else if (Rs >= 4055) {
return 8;
}
}
else if (Rs >= 4388) {
return 2;
}
}
}
else if (negBalance >= 48) {
if (diffRsCsdiff < 66) {
if (ssDCval < 105) {
return 2;
}
else if (ssDCval >= 105) {
return 8;
}
}
else if (diffRsCsdiff >= 66) {
if (Cs < 1425) {
return 2;
}
else if (Cs >= 1425) {
if (SC < 5178) {
return 8;
}
else if (SC >= 5178) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (ssDCval < 120) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
else if (ssDCval >= 120) {
if (SC < 6158) {
return 1;
}
else if (SC >= 6158) {
if (Rs < 4692) {
return 8;
}
else if (Rs >= 4692) {
return 2;
}
}
}
}
}
else if (Cs >= 2176) {
if (negBalance < 38) {
if (Cs < 2181) {
return 4;
}
else if (Cs >= 2181) {
return 2;
}
}
else if (negBalance >= 38) {
if (ssDCval < 73) {
if (Cs < 3350) {
return 4;
}
else if (Cs >= 3350) {
return 8;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
}
}
}
}
}
else if (CsDiff >= 7) {
if (mcTcor < 90) {
if (mu_mv_mag_sq < 26) {
if (RsDiff < 16) {
if (Rs < 3800) {
if (tcor < 64) {
if (mu_mv_mag_sq < 5) {
if (AFD < 13) {
if (Rs < 1812) {
return 1;
}
else if (Rs >= 1812) {
return 8;
}
}
else if (AFD >= 13) {
if (refDCval < 87) {
if (refDCval < 77) {
if (MVDiff < 19) {
if (ssDCval < 63) {
if (Cs < 1599) {
if (Rs < 1497) {
return 4;
}
else if (Rs >= 1497) {
return 8;
}
}
else if (Cs >= 1599) {
return 4;
}
}
else if (ssDCval >= 63) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 2;
}
}
}
}
else if (MVDiff >= 19) {
if (Rs < 2109) {
if (Cs < 1729) {
if (refDCval < 70) {
if (diffRsCsdiff < 34) {
return 4;
}
else if (diffRsCsdiff >= 34) {
if (Cs < 1449) {
return 8;
}
else if (Cs >= 1449) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (Rs < 2025) {
return 1;
}
else if (Rs >= 2025) {
return 4;
}
}
}
}
}
else if (refDCval >= 70) {
if (MVDiff < 46) {
return 2;
}
else if (MVDiff >= 46) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 4;
}
}
}
}
else if (Cs >= 1729) {
if (TSC < 14) {
if (RsDiff < 11) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
return 2;
}
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (TSC >= 14) {
if (Rs < 1296) {
if (Rs < 1003) {
return 2;
}
else if (Rs >= 1003) {
return 1;
}
}
else if (Rs >= 1296) {
return 4;
}
}
}
}
else if (Rs >= 2109) {
if (Cs < 1731) {
if (Rs < 2749) {
return 8;
}
else if (Rs >= 2749) {
return 2;
}
}
else if (Cs >= 1731) {
if (MVDiff < 21) {
if (Cs < 3089) {
return 8;
}
else if (Cs >= 3089) {
return 4;
}
}
else if (MVDiff >= 21) {
return 4;
}
}
}
}
}
else if (refDCval >= 77) {
if (MVDiff < 30) {
return 4;
}
else if (MVDiff >= 30) {
if (SC < 3826) {
if (mcTcor < 76) {
if (Cs < 1818) {
return 1;
}
else if (Cs >= 1818) {
return 4;
}
}
else if (mcTcor >= 76) {
return 4;
}
}
else if (SC >= 3826) {
return 8;
}
}
}
}
else if (refDCval >= 87) {
if (negBalance < 53) {
if (ssDCval < 127) {
return 8;
}
else if (ssDCval >= 127) {
return 1;
}
}
else if (negBalance >= 53) {
if (TSC < 14) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (ssDCval < 115) {
if (Rs < 1512) {
return 8;
}
else if (Rs >= 1512) {
return 4;
}
}
else if (ssDCval >= 115) {
if (Rs < 2249) {
return 2;
}
else if (Rs >= 2249) {
return 1;
}
}
}
}
else if (TSC >= 14) {
if (tcor < 57) {
if (Cs < 3743) {
return 2;
}
else if (Cs >= 3743) {
return 8;
}
}
else if (tcor >= 57) {
if (ssDCval < 117) {
return 1;
}
else if (ssDCval >= 117) {
if (Rs < 2756) {
return 2;
}
else if (Rs >= 2756) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Cs < 1673) {
if (Rs < 2424) {
if (MVDiff < 47) {
if (negBalance < 59) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (RsDiff < 10) {
if (Rs < 1984) {
return 4;
}
else if (Rs >= 1984) {
return 8;
}
}
else if (RsDiff >= 10) {
return 8;
}
}
}
else if (negBalance >= 59) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 4;
}
}
}
}
else if (MVDiff >= 47) {
if (AFD < 35) {
return 8;
}
else if (AFD >= 35) {
return 1;
}
}
}
else if (Rs >= 2424) {
if (Cs < 1409) {
return 4;
}
else if (Cs >= 1409) {
return 1;
}
}
}
else if (Cs >= 1673) {
if (RsCsDiff < 339) {
if (CsDiff < 10) {
if (RsDiff < 12) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
else if (RsDiff >= 12) {
if (Rs < 2749) {
return 4;
}
else if (Rs >= 2749) {
return 8;
}
}
}
else if (CsDiff >= 10) {
if (MVDiff < 22) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
return 2;
}
}
else if (MVDiff >= 22) {
if (ssDCval < 77) {
if (CsDiff < 15) {
return 4;
}
else if (CsDiff >= 15) {
return 2;
}
}
else if (ssDCval >= 77) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 339) {
if (tcor < 60) {
if (Rs < 1341) {
return 2;
}
else if (Rs >= 1341) {
return 8;
}
}
else if (tcor >= 60) {
if (mu_mv_mag_sq < 20) {
return 2;
}
else if (mu_mv_mag_sq >= 20) {
if (Rs < 2403) {
return 1;
}
else if (Rs >= 2403) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 64) {
if (refDCval < 127) {
if (Cs < 1409) {
if (ssDCval < 118) {
if (mu_mv_mag_sq < 9) {
if (MVDiff < 17) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
else if (MVDiff >= 17) {
return 2;
}
}
else if (mu_mv_mag_sq >= 9) {
return 8;
}
}
else if (ssDCval >= 118) {
return 1;
}
}
else if (Cs >= 1409) {
if (posBalance < 64) {
if (RsCsDiff < 358) {
if (ssDCval < 107) {
if (mu_mv_mag_sq < 3) {
if (negBalance < 58) {
if (Cs < 2013) {
if (refDCval < 83) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
if (ssDCval < 50) {
if (Rs < 1863) {
return 8;
}
else if (Rs >= 1863) {
return 2;
}
}
else if (ssDCval >= 50) {
if (TSC < 14) {
if (RsDiff < 14) {
return 4;
}
else if (RsDiff >= 14) {
return 2;
}
}
else if (TSC >= 14) {
return 8;
}
}
}
}
else if (refDCval >= 83) {
if (RsDiff < 10) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
else if (RsDiff >= 10) {
if (AFD < 18) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
return 4;
}
}
else if (AFD >= 18) {
return 8;
}
}
}
}
else if (Cs >= 2013) {
if (Rs < 1574) {
return 4;
}
else if (Rs >= 1574) {
if (ssDCval < 80) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
if (RsCsDiff < 289) {
return 8;
}
else if (RsCsDiff >= 289) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (Cs < 3010) {
return 1;
}
else if (Cs >= 3010) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 80) {
if (SC < 5437) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (diffRsCsdiff < -2) {
if (Rs < 2653) {
return 4;
}
else if (Rs >= 2653) {
return 8;
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
else if (SC >= 5437) {
if (SC < 7181) {
if (mu_mv_mag_sq < 2) {
if (negBalance < 50) {
if (Cs < 3146) {
if (Rs < 3085) {
return 8;
}
else if (Rs >= 3085) {
return 2;
}
}
else if (Cs >= 3146) {
return 4;
}
}
else if (negBalance >= 50) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
else if (SC >= 7181) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 58) {
if (SC < 6382) {
if (MVDiff < 24) {
if (Cs < 1636) {
return 8;
}
else if (Cs >= 1636) {
if (RsCsDiff < 277) {
if (TSC < 10) {
if (Rs < 2224) {
return 4;
}
else if (Rs >= 2224) {
return 8;
}
}
else if (TSC >= 10) {
return 4;
}
}
else if (RsCsDiff >= 277) {
return 1;
}
}
}
else if (MVDiff >= 24) {
if (Rs < 1235) {
return 1;
}
else if (Rs >= 1235) {
return 2;
}
}
}
else if (SC >= 6382) {
if (Rs < 2438) {
return 2;
}
else if (Rs >= 2438) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 2508) {
if (mcTcor < 80) {
if (Cs < 1932) {
return 1;
}
else if (Cs >= 1932) {
if (Rs < 2052) {
return 1;
}
else if (Rs >= 2052) {
return 8;
}
}
}
else if (mcTcor >= 80) {
if (Rs < 2721) {
if (Rs < 1805) {
if (tcor < 72) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
return 8;
}
}
else if (tcor >= 72) {
if (Rs < 1408) {
return 2;
}
else if (Rs >= 1408) {
return 8;
}
}
}
else if (Rs >= 1805) {
if (MVDiff < 11) {
if (Cs < 2205) {
return 8;
}
else if (Cs >= 2205) {
return 1;
}
}
else if (MVDiff >= 11) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 8;
}
}
}
}
}
else if (Rs >= 2721) {
if (AFD < 25) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 4;
}
}
else if (AFD >= 25) {
return 8;
}
}
}
}
else if (Cs >= 2508) {
if (Rs < 2086) {
return 8;
}
else if (Rs >= 2086) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 107) {
if (negBalance < 59) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
if (Rs < 3718) {
return 8;
}
else if (Rs >= 3718) {
return 4;
}
}
}
else if (negBalance >= 59) {
if (Rs < 2451) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
return 1;
}
}
else if (Rs >= 2451) {
if (Rs < 2783) {
return 2;
}
else if (Rs >= 2783) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 358) {
if (MVDiff < 20) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (RsCsDiff < 380) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (RsCsDiff >= 380) {
return 4;
}
}
}
else if (MVDiff >= 20) {
if (MVDiff < 29) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 4;
}
}
else if (MVDiff >= 29) {
if (diffMVdiffVal < 8) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
return 2;
}
}
else if (diffMVdiffVal >= 8) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 64) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (CsDiff < 13) {
if (Rs < 2786) {
if (TSC < 9) {
if (Cs < 2306) {
return 2;
}
else if (Cs >= 2306) {
return 8;
}
}
else if (TSC >= 9) {
if (CsDiff < 12) {
if (Cs < 1977) {
if (Cs < 1812) {
if (Cs < 1603) {
return 8;
}
else if (Cs >= 1603) {
return 4;
}
}
else if (Cs >= 1812) {
return 8;
}
}
else if (Cs >= 1977) {
return 4;
}
}
else if (CsDiff >= 12) {
return 2;
}
}
}
else if (Rs >= 2786) {
if (negBalance < 53) {
return 2;
}
else if (negBalance >= 53) {
if (Rs < 3420) {
return 8;
}
else if (Rs >= 3420) {
return 4;
}
}
}
}
else if (CsDiff >= 13) {
return 8;
}
}
}
}
}
else if (refDCval >= 127) {
if (TSC < 13) {
if (CsDiff < 8) {
if (Rs < 2473) {
return 8;
}
else if (Rs >= 2473) {
return 4;
}
}
else if (CsDiff >= 8) {
if (mu_mv_mag_sq < 25) {
return 8;
}
else if (mu_mv_mag_sq >= 25) {
if (Rs < 2699) {
return 8;
}
else if (Rs >= 2699) {
return 4;
}
}
}
}
else if (TSC >= 13) {
if (MVDiff < 22) {
if (Cs < 2865) {
if (Rs < 2013) {
return 2;
}
else if (Rs >= 2013) {
return 8;
}
}
else if (Cs >= 2865) {
return 1;
}
}
else if (MVDiff >= 22) {
return 1;
}
}
}
}
}
else if (Rs >= 3800) {
if (negBalance < 55) {
if (diffAFD < -2) {
if (Cs < 2546) {
return 8;
}
else if (Cs >= 2546) {
return 4;
}
}
else if (diffAFD >= -2) {
if (ssDCval < 140) {
if (MVDiff < 45) {
return 8;
}
else if (MVDiff >= 45) {
return 1;
}
}
else if (ssDCval >= 140) {
return 4;
}
}
}
else if (negBalance >= 55) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
if (Cs < 2799) {
if (AFD < 20) {
if (Rs < 4000) {
return 8;
}
else if (Rs >= 4000) {
return 4;
}
}
else if (AFD >= 20) {
return 1;
}
}
else if (Cs >= 2799) {
if (Rs < 4801) {
return 4;
}
else if (Rs >= 4801) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 16) {
if (SC < 11119) {
if (AFD < 29) {
if (Cs < 1858) {
if (ssDCval < 101) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
return 2;
}
}
else if (ssDCval >= 101) {
if (Rs < 4009) {
return 8;
}
else if (Rs >= 4009) {
if (Cs < 1585) {
return 4;
}
else if (Cs >= 1585) {
return 2;
}
}
}
}
else if (Cs >= 1858) {
if (posBalance < 55) {
if (ssDCval < 72) {
if (tcor < 53) {
return 4;
}
else if (tcor >= 53) {
if (mcTcor < 89) {
return 2;
}
else if (mcTcor >= 89) {
return 4;
}
}
}
else if (ssDCval >= 72) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (Rs < 6653) {
if (mu_mv_mag_sq < 21) {
return 8;
}
else if (mu_mv_mag_sq >= 21) {
return 4;
}
}
else if (Rs >= 6653) {
if (Rs < 7549) {
return 1;
}
else if (Rs >= 7549) {
return 2;
}
}
}
}
}
else if (posBalance >= 55) {
if (TSC < 16) {
if (tcor < 51) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 8;
}
}
else if (tcor >= 51) {
if (diffRsCsdiff < 159) {
if (diffAFD < -2) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
return 4;
}
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < -94) {
if (Cs < 3156) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
return 2;
}
}
else if (Cs >= 3156) {
return 4;
}
}
else if (diffRsCsdiff >= -94) {
if (MVDiff < 9) {
if (AFD < 27) {
if (Rs < 4747) {
return 2;
}
else if (Rs >= 4747) {
return 8;
}
}
else if (AFD >= 27) {
return 4;
}
}
else if (MVDiff >= 9) {
if (CsDiff < 14) {
if (diffRsCsdiff < 122) {
return 4;
}
else if (diffRsCsdiff >= 122) {
if (Rs < 4976) {
return 8;
}
else if (Rs >= 4976) {
return 4;
}
}
}
else if (CsDiff >= 14) {
if (RsCsDiff < 615) {
if (diffRsCsdiff < -36) {
return 8;
}
else if (diffRsCsdiff >= -36) {
if (mcTcor < 79) {
return 8;
}
else if (mcTcor >= 79) {
return 4;
}
}
}
else if (RsCsDiff >= 615) {
return 4;
}
}
}
}
}
}
else if (diffRsCsdiff >= 159) {
if (tcor < 67) {
return 4;
}
else if (tcor >= 67) {
if (Rs < 4255) {
return 8;
}
else if (Rs >= 4255) {
return 2;
}
}
}
}
}
else if (TSC >= 16) {
if (MVDiff < 14) {
if (Rs < 5689) {
if (Rs < 4966) {
return 1;
}
else if (Rs >= 4966) {
return 4;
}
}
else if (Rs >= 5689) {
return 2;
}
}
else if (MVDiff >= 14) {
if (negBalance < 56) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
return 4;
}
}
else if (negBalance >= 56) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
if (RsCsDiff < 446) {
return 4;
}
else if (RsCsDiff >= 446) {
return 8;
}
}
}
}
}
}
}
}
else if (AFD >= 29) {
if (mu_mv_mag_sq < 20) {
if (CsDiff < 15) {
if (ssDCval < 83) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
return 1;
}
}
else if (ssDCval >= 83) {
if (tcor < 58) {
if (ssDCval < 122) {
return 2;
}
else if (ssDCval >= 122) {
if (Rs < 4694) {
return 1;
}
else if (Rs >= 4694) {
return 4;
}
}
}
else if (tcor >= 58) {
if (TSC < 17) {
if (Rs < 3984) {
return 8;
}
else if (Rs >= 3984) {
return 2;
}
}
else if (TSC >= 17) {
if (TSC < 20) {
return 1;
}
else if (TSC >= 20) {
return 8;
}
}
}
}
}
else if (CsDiff >= 15) {
if (RsDiff < 22) {
if (TSC < 19) {
return 4;
}
else if (TSC >= 19) {
return 8;
}
}
else if (RsDiff >= 22) {
if (negBalance < 59) {
if (AFD < 47) {
return 2;
}
else if (AFD >= 47) {
return 8;
}
}
else if (negBalance >= 59) {
if (RsDiff < 23) {
return 2;
}
else if (RsDiff >= 23) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 20) {
if (tcor < 58) {
if (RsCsDiff < 747) {
if (Cs < 1541) {
return 2;
}
else if (Cs >= 1541) {
return 4;
}
}
else if (RsCsDiff >= 747) {
return 1;
}
}
else if (tcor >= 58) {
if (RsDiff < 19) {
return 8;
}
else if (RsDiff >= 19) {
return 4;
}
}
}
}
}
else if (SC >= 11119) {
if (AFD < 41) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 4;
}
}
else if (AFD >= 41) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (tcor < 57) {
if (AFD < 31) {
if (diffAFD < 3) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
if (RsCsDiff < 748) {
if (ssDCval < 58) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 1;
}
}
else if (ssDCval >= 58) {
if (AFD < 29) {
if (AFD < 24) {
if (Cs < 2517) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
if (Rs < 1612) {
return 4;
}
else if (Rs >= 1612) {
return 8;
}
}
}
else if (Cs >= 2517) {
if (Cs < 3325) {
return 2;
}
else if (Cs >= 3325) {
return 8;
}
}
}
else if (AFD >= 24) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (AFD >= 29) {
if (RsCsDiff < 338) {
if (Rs < 2215) {
return 8;
}
else if (Rs >= 2215) {
return 1;
}
}
else if (RsCsDiff >= 338) {
return 4;
}
}
}
}
else if (RsCsDiff >= 748) {
return 4;
}
}
}
else if (diffAFD >= 3) {
if (RsDiff < 7) {
if (Rs < 822) {
return 2;
}
else if (Rs >= 822) {
return 1;
}
}
else if (RsDiff >= 7) {
if (diffAFD < 4) {
if (Rs < 2157) {
return 2;
}
else if (Rs >= 2157) {
return 1;
}
}
else if (diffAFD >= 4) {
if (Rs < 5390) {
return 8;
}
else if (Rs >= 5390) {
if (Rs < 7050) {
return 4;
}
else if (Rs >= 7050) {
return 8;
}
}
}
}
}
}
else if (AFD >= 31) {
if (CsDiff < 11) {
if (AFD < 32) {
return 4;
}
else if (AFD >= 32) {
return 2;
}
}
else if (CsDiff >= 11) {
if (posBalance < 51) {
if (Rs < 2507) {
return 2;
}
else if (Rs >= 2507) {
return 8;
}
}
else if (posBalance >= 51) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (AFD < 38) {
if (tcor < 38) {
return 4;
}
else if (tcor >= 38) {
if (Rs < 1097) {
return 4;
}
else if (Rs >= 1097) {
return 8;
}
}
}
else if (AFD >= 38) {
return 1;
}
}
}
}
}
}
else if (tcor >= 57) {
if (posBalance < 51) {
if (ssDCval < 84) {
if (SC < 5641) {
if (negBalance < 51) {
if (AFD < 14) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (Cs < 1972) {
if (Rs < 2065) {
if (AFD < 11) {
if (Cs < 1554) {
return 8;
}
else if (Cs >= 1554) {
return 4;
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (Rs >= 2065) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (Cs >= 1972) {
if (Rs < 1867) {
return 1;
}
else if (Rs >= 1867) {
return 4;
}
}
}
}
else if (AFD >= 14) {
if (Rs < 2319) {
return 1;
}
else if (Rs >= 2319) {
return 2;
}
}
}
else if (negBalance >= 51) {
if (Rs < 1717) {
if (SC < 2835) {
return 8;
}
else if (SC >= 2835) {
return 1;
}
}
else if (Rs >= 1717) {
if (diffAFD < -18) {
return 4;
}
else if (diffAFD >= -18) {
return 8;
}
}
}
}
else if (SC >= 5641) {
if (mu_mv_mag_sq < 56) {
if (RsCsDiff < 679) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
return 4;
}
}
else if (RsCsDiff >= 679) {
if (Cs < 3791) {
return 4;
}
else if (Cs >= 3791) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (Cs < 2369) {
if (Rs < 4008) {
return 8;
}
else if (Rs >= 4008) {
if (Rs < 4165) {
return 1;
}
else if (Rs >= 4165) {
return 4;
}
}
}
else if (Cs >= 2369) {
if (SC < 9858) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
if (TSC < 15) {
return 4;
}
else if (TSC >= 15) {
if (Rs < 5147) {
return 2;
}
else if (Rs >= 5147) {
return 4;
}
}
}
}
else if (SC >= 9858) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
if (Rs < 6088) {
return 4;
}
else if (Rs >= 6088) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (TSC < 14) {
if (ssDCval < 124) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (posBalance < 40) {
if (Rs < 6016) {
return 2;
}
else if (Rs >= 6016) {
return 8;
}
}
else if (posBalance >= 40) {
if (tcor < 77) {
return 2;
}
else if (tcor >= 77) {
return 8;
}
}
}
}
else if (ssDCval >= 124) {
if (ssDCval < 135) {
if (Rs < 2643) {
if (Rs < 2011) {
return 8;
}
else if (Rs >= 2011) {
return 1;
}
}
else if (Rs >= 2643) {
return 4;
}
}
else if (ssDCval >= 135) {
if (mu_mv_mag_sq < 37) {
if (RsCsDiff < 199) {
return 1;
}
else if (RsCsDiff >= 199) {
if (Rs < 6730) {
return 8;
}
else if (Rs >= 6730) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (negBalance < 50) {
if (diffMVdiffVal < -6) {
if (CsDiff < 10) {
return 1;
}
else if (CsDiff >= 10) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
else if (negBalance >= 50) {
if (Rs < 3552) {
return 8;
}
else if (Rs >= 3552) {
return 4;
}
}
}
}
}
}
else if (TSC >= 14) {
if (tcor < 78) {
if (Rs < 6178) {
if (Cs < 1784) {
return 2;
}
else if (Cs >= 1784) {
return 8;
}
}
else if (Rs >= 6178) {
return 1;
}
}
else if (tcor >= 78) {
if (Rs < 4624) {
return 8;
}
else if (Rs >= 4624) {
return 1;
}
}
}
}
}
else if (posBalance >= 51) {
if (Cs < 3529) {
if (mu_mv_mag_sq < 67) {
if (tcor < 72) {
if (MVDiff < 42) {
if (negBalance < 57) {
if (RsDiff < 20) {
if (MVDiff < 20) {
if (Rs < 2010) {
if (Cs < 1793) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (Cs >= 1793) {
return 2;
}
}
else if (Rs >= 2010) {
if (ssDCval < 137) {
if (Rs < 2871) {
if (MVDiff < 15) {
if (RsDiff < 11) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 4;
}
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (MVDiff >= 15) {
return 1;
}
}
else if (Rs >= 2871) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 4;
}
}
}
}
else if (ssDCval >= 137) {
if (Rs < 4562) {
if (Rs < 3153) {
return 8;
}
else if (Rs >= 3153) {
if (Rs < 4406) {
return 1;
}
else if (Rs >= 4406) {
return 8;
}
}
}
else if (Rs >= 4562) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
return 8;
}
}
}
}
}
else if (MVDiff >= 20) {
if (ssDCval < 76) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (tcor < 67) {
if (Cs < 2201) {
return 4;
}
else if (Cs >= 2201) {
return 2;
}
}
else if (tcor >= 67) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (MVDiff < 31) {
return 4;
}
else if (MVDiff >= 31) {
return 8;
}
}
}
}
}
else if (ssDCval >= 76) {
return 8;
}
}
}
else if (RsDiff >= 20) {
if (Rs < 5922) {
return 4;
}
else if (Rs >= 5922) {
if (Cs < 3306) {
return 8;
}
else if (Cs >= 3306) {
return 2;
}
}
}
}
else if (negBalance >= 57) {
if (Rs < 2509) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
if (RsCsDiff < 145) {
if (Rs < 1286) {
return 2;
}
else if (Rs >= 1286) {
return 1;
}
}
else if (RsCsDiff >= 145) {
if (ssDCval < 106) {
return 4;
}
else if (ssDCval >= 106) {
return 1;
}
}
}
}
else if (Rs >= 2509) {
if (SC < 6079) {
if (ssDCval < 134) {
return 8;
}
else if (ssDCval >= 134) {
return 2;
}
}
else if (SC >= 6079) {
return 4;
}
}
}
}
else if (MVDiff >= 42) {
if (Cs < 2197) {
if (tcor < 69) {
if (Cs < 1432) {
if (Rs < 1723) {
return 8;
}
else if (Rs >= 1723) {
return 4;
}
}
else if (Cs >= 1432) {
return 8;
}
}
else if (tcor >= 69) {
if (ssDCval < 120) {
if (Rs < 2330) {
return 1;
}
else if (Rs >= 2330) {
return 8;
}
}
else if (ssDCval >= 120) {
return 2;
}
}
}
else if (Cs >= 2197) {
if (ssDCval < 141) {
return 2;
}
else if (ssDCval >= 141) {
return 8;
}
}
}
}
else if (tcor >= 72) {
if (Cs < 3045) {
if (CsDiff < 11) {
if (MVDiff < 9) {
if (ssDCval < 108) {
if (ssDCval < 104) {
if (Cs < 2911) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (SC < 6520) {
return 8;
}
else if (SC >= 6520) {
return 4;
}
}
}
else if (Cs >= 2911) {
return 4;
}
}
else if (ssDCval >= 104) {
return 4;
}
}
else if (ssDCval >= 108) {
return 8;
}
}
else if (MVDiff >= 9) {
if (Rs < 3116) {
if (mcTcor < 89) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
if (Rs < 2379) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
else if (Rs >= 2379) {
return 4;
}
}
}
else if (mcTcor >= 89) {
if (diffTSC < -1) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
else if (diffTSC >= -1) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (Rs < 2685) {
return 2;
}
else if (Rs >= 2685) {
return 8;
}
}
}
}
}
else if (Rs >= 3116) {
if (TSC < 12) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
if (ssDCval < 78) {
if (Rs < 3562) {
return 4;
}
else if (Rs >= 3562) {
return 8;
}
}
else if (ssDCval >= 78) {
return 8;
}
}
}
else if (TSC >= 12) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (Rs < 3194) {
return 2;
}
else if (Rs >= 3194) {
if (TSC < 13) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 4;
}
}
else if (TSC >= 13) {
return 8;
}
}
}
}
}
}
}
else if (CsDiff >= 11) {
if (mu_mv_mag_sq < 45) {
if (RsDiff < 19) {
return 8;
}
else if (RsDiff >= 19) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (Cs < 2913) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
}
}
}
else if (Cs >= 2913) {
return 1;
}
}
}
}
else if (Cs >= 3045) {
if (ssDCval < 78) {
if (Rs < 4692) {
return 8;
}
else if (Rs >= 4692) {
return 2;
}
}
else if (ssDCval >= 78) {
if (Rs < 4529) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
return 8;
}
}
else if (Rs >= 4529) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 2652) {
if (MVDiff < 12) {
if (Cs < 2702) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 91) {
return 8;
}
else if (mu_mv_mag_sq >= 91) {
return 4;
}
}
else if (MVDiff >= 8) {
if (Rs < 1961) {
return 1;
}
else if (Rs >= 1961) {
return 4;
}
}
}
}
else if (Cs >= 2702) {
return 4;
}
}
else if (MVDiff >= 12) {
if (diffRsCsdiff < -35) {
return 4;
}
else if (diffRsCsdiff >= -35) {
if (ssDCval < 49) {
if (Rs < 1718) {
return 8;
}
else if (Rs >= 1718) {
return 4;
}
}
else if (ssDCval >= 49) {
return 8;
}
}
}
}
else if (Rs >= 2652) {
if (ssDCval < 94) {
if (mu_mv_mag_sq < 76) {
if (refDCval < 67) {
return 8;
}
else if (refDCval >= 67) {
if (SC < 6627) {
if (Rs < 2782) {
return 8;
}
else if (Rs >= 2782) {
if (MVDiff < 6) {
if (Rs < 2962) {
return 4;
}
else if (Rs >= 2962) {
return 8;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (SC >= 6627) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (diffAFD < -3) {
if (SC < 6484) {
return 4;
}
else if (SC >= 6484) {
return 1;
}
}
else if (diffAFD >= -3) {
if (refDCval < 52) {
return 4;
}
else if (refDCval >= 52) {
return 8;
}
}
}
}
else if (ssDCval >= 94) {
if (Cs < 2990) {
if (tcor < 87) {
if (CsDiff < 8) {
if (mu_mv_mag_sq < 68) {
if (Rs < 3391) {
return 4;
}
else if (Rs >= 3391) {
return 8;
}
}
else if (mu_mv_mag_sq >= 68) {
return 8;
}
}
else if (CsDiff >= 8) {
if (negBalance < 48) {
if (Rs < 3501) {
return 2;
}
else if (Rs >= 3501) {
return 8;
}
}
else if (negBalance >= 48) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
else if (tcor >= 87) {
return 1;
}
}
else if (Cs >= 2990) {
if (Rs < 4466) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
else if (Rs >= 4466) {
return 8;
}
}
}
}
}
}
else if (Cs >= 3529) {
if (negBalance < 52) {
if (SC < 10079) {
if (SC < 8476) {
if (tcor < 75) {
if (Cs < 3933) {
if (ssDCval < 134) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
return 1;
}
}
else if (ssDCval >= 134) {
return 8;
}
}
else if (Cs >= 3933) {
return 1;
}
}
else if (tcor >= 75) {
if (TSC < 13) {
if (SC < 7253) {
return 4;
}
else if (SC >= 7253) {
if (Rs < 4264) {
if (Cs < 3578) {
return 4;
}
else if (Cs >= 3578) {
return 8;
}
}
else if (Rs >= 4264) {
return 4;
}
}
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (SC >= 8476) {
return 8;
}
}
else if (SC >= 10079) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
if (SC < 10097) {
return 2;
}
else if (SC >= 10097) {
if (Rs < 4759) {
if (Rs < 4647) {
return 4;
}
else if (Rs >= 4647) {
return 8;
}
}
else if (Rs >= 4759) {
if (diffMVdiffVal < 7) {
if (Rs < 6132) {
if (Rs < 5517) {
return 4;
}
else if (Rs >= 5517) {
return 2;
}
}
else if (Rs >= 6132) {
return 4;
}
}
else if (diffMVdiffVal >= 7) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 52) {
if (Rs < 7515) {
if (posBalance < 55) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (posBalance >= 55) {
if (Rs < 3191) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
return 4;
}
}
else if (Rs >= 3191) {
if (diffMVdiffVal < -1) {
if (tcor < 67) {
if (TSC < 16) {
return 4;
}
else if (TSC >= 16) {
return 8;
}
}
else if (tcor >= 67) {
if (posBalance < 57) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (posBalance >= 57) {
if (Rs < 3479) {
if (Rs < 3454) {
return 8;
}
else if (Rs >= 3454) {
return 4;
}
}
else if (Rs >= 3479) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
}
else if (Rs >= 7515) {
if (Cs < 5802) {
return 8;
}
else if (Cs >= 5802) {
return 4;
}
}
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (RsDiff < 10) {
if (Rs < 3964) {
if (Rs < 1475) {
if (negBalance < 44) {
if (posBalance < 61) {
if (mu_mv_mag_sq < 32) {
if (Rs < 1305) {
if (diffRsCsdiff < 72) {
return 1;
}
else if (diffRsCsdiff >= 72) {
if (Rs < 882) {
return 1;
}
else if (Rs >= 882) {
return 2;
}
}
}
else if (Rs >= 1305) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Cs < 1732) {
return 2;
}
else if (Cs >= 1732) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (RsCsDiff < 110) {
return 2;
}
else if (RsCsDiff >= 110) {
return 8;
}
}
}
}
else if (posBalance >= 61) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
if (Rs < 1462) {
return 2;
}
else if (Rs >= 1462) {
return 4;
}
}
}
}
else if (negBalance >= 44) {
if (ssDCval < 59) {
if (Rs < 580) {
return 4;
}
else if (Rs >= 580) {
return 2;
}
}
else if (ssDCval >= 59) {
if (SC < 2756) {
if (refDCval < 80) {
if (negBalance < 51) {
if (AFD < 18) {
if (Rs < 1031) {
return 1;
}
else if (Rs >= 1031) {
return 8;
}
}
else if (AFD >= 18) {
return 2;
}
}
else if (negBalance >= 51) {
if (Rs < 600) {
return 4;
}
else if (Rs >= 600) {
return 8;
}
}
}
else if (refDCval >= 80) {
if (negBalance < 50) {
if (RsDiff < 9) {
if (Rs < 1040) {
return 4;
}
else if (Rs >= 1040) {
if (Rs < 1297) {
return 2;
}
else if (Rs >= 1297) {
return 8;
}
}
}
else if (RsDiff >= 9) {
return 1;
}
}
else if (negBalance >= 50) {
if (SC < 2262) {
return 1;
}
else if (SC >= 2262) {
return 2;
}
}
}
}
else if (SC >= 2756) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 10) {
if (Rs < 1365) {
return 4;
}
else if (Rs >= 1365) {
return 8;
}
}
else if (MVDiff >= 10) {
if (RsCsDiff < 201) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 1;
}
}
else if (RsCsDiff >= 201) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (MVDiff < 20) {
if (tcor < 81) {
if (Rs < 1008) {
return 2;
}
else if (Rs >= 1008) {
return 8;
}
}
else if (tcor >= 81) {
return 4;
}
}
else if (MVDiff >= 20) {
if (diffAFD < 3) {
if (RsCsDiff < 161) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 4;
}
}
else if (RsCsDiff >= 161) {
if (Rs < 1460) {
return 8;
}
else if (Rs >= 1460) {
return 4;
}
}
}
else if (diffAFD >= 3) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (refDCval < 69) {
return 4;
}
else if (refDCval >= 69) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1475) {
if (Cs < 2624) {
if (ssDCval < 85) {
if (posBalance < 46) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
else if (posBalance >= 46) {
if (Rs < 2752) {
if (Cs < 1803) {
if (ssDCval < 82) {
if (mu_mv_mag_sq < 12) {
if (ssDCval < 72) {
if (refDCval < 73) {
return 1;
}
else if (refDCval >= 73) {
return 4;
}
}
else if (ssDCval >= 72) {
if (mcTcor < 92) {
if (diffRsCsdiff < -17) {
if (Rs < 1912) {
return 4;
}
else if (Rs >= 1912) {
return 2;
}
}
else if (diffRsCsdiff >= -17) {
return 8;
}
}
else if (mcTcor >= 92) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
if (Rs < 1800) {
return 1;
}
else if (Rs >= 1800) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 12) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1545) {
return 8;
}
else if (Rs >= 1545) {
return 1;
}
}
}
}
}
else if (ssDCval >= 82) {
if (TSC < 10) {
if (CsDiff < 9) {
return 4;
}
else if (CsDiff >= 9) {
return 8;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (Cs >= 1803) {
if (ssDCval < 66) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (tcor < 81) {
return 1;
}
else if (tcor >= 81) {
return 4;
}
}
}
else if (ssDCval >= 66) {
if (Cs < 1833) {
if (Cs < 1828) {
return 8;
}
else if (Cs >= 1828) {
return 4;
}
}
else if (Cs >= 1833) {
return 8;
}
}
}
}
else if (Rs >= 2752) {
if (mu_mv_mag_sq < 82) {
if (negBalance < 40) {
if (Rs < 3674) {
return 2;
}
else if (Rs >= 3674) {
return 4;
}
}
else if (negBalance >= 40) {
return 4;
}
}
else if (mu_mv_mag_sq >= 82) {
if (RsCsDiff < 138) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (RsCsDiff >= 138) {
return 4;
}
}
}
}
}
else if (ssDCval >= 85) {
if (AFD < 28) {
if (Rs < 3686) {
if (AFD < 19) {
if (mcTcor < 95) {
if (tcor < 83) {
if (posBalance < 54) {
if (diffAFD < 1) {
if (Rs < 1746) {
return 8;
}
else if (Rs >= 1746) {
return 2;
}
}
else if (diffAFD >= 1) {
if (Cs < 1789) {
return 8;
}
else if (Cs >= 1789) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
if (SC < 3195) {
if (Rs < 1522) {
return 8;
}
else if (Rs >= 1522) {
return 4;
}
}
else if (SC >= 3195) {
if (diffMVdiffVal < 5) {
if (diffMVdiffVal < -4) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 1668) {
return 8;
}
else if (Rs >= 1668) {
return 4;
}
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (diffMVdiffVal >= 5) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 1670) {
return 8;
}
else if (Rs >= 1670) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 83) {
if (Cs < 2078) {
return 8;
}
else if (Cs >= 2078) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (mcTcor < 93) {
if (MVDiff < 3) {
if (Rs < 2943) {
return 8;
}
else if (Rs >= 2943) {
return 1;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (mcTcor >= 93) {
if (Rs < 2420) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (Rs >= 2420) {
if (ssDCval < 105) {
return 8;
}
else if (ssDCval >= 105) {
if (Rs < 3039) {
return 4;
}
else if (Rs >= 3039) {
if (Rs < 3358) {
return 1;
}
else if (Rs >= 3358) {
return 8;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (ssDCval < 97) {
if (mu_mv_mag_sq < 58) {
if (negBalance < 48) {
if (posBalance < 58) {
if (Cs < 1432) {
return 1;
}
else if (Cs >= 1432) {
return 8;
}
}
else if (posBalance >= 58) {
return 2;
}
}
else if (negBalance >= 48) {
if (mu_mv_mag_sq < 2) {
if (Rs < 2351) {
return 8;
}
else if (Rs >= 2351) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1674) {
return 8;
}
else if (Rs >= 1674) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 58) {
return 1;
}
}
else if (ssDCval >= 97) {
if (RsDiff < 8) {
if (mu_mv_mag_sq < 3) {
if (Cs < 2102) {
return 2;
}
else if (Cs >= 2102) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (RsDiff >= 8) {
if (tcor < 80) {
return 2;
}
else if (tcor >= 80) {
if (Rs < 1628) {
if (Rs < 1582) {
return 8;
}
else if (Rs >= 1582) {
return 4;
}
}
else if (Rs >= 1628) {
return 8;
}
}
}
}
}
}
else if (AFD >= 19) {
if (negBalance < 43) {
return 2;
}
else if (negBalance >= 43) {
if (posBalance < 60) {
if (mu_mv_mag_sq < 5) {
return 8;
}
else if (mu_mv_mag_sq >= 5) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (posBalance >= 60) {
if (ssDCval < 121) {
if (mcTcor < 94) {
if (refDCval < 90) {
return 1;
}
else if (refDCval >= 90) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (SC < 3681) {
return 1;
}
else if (SC >= 3681) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (ssDCval >= 121) {
return 2;
}
}
}
}
}
else if (Rs >= 3686) {
if (mu_mv_mag_sq < 88) {
if (ssDCval < 98) {
if (AFD < 10) {
if (Cs < 1970) {
return 2;
}
else if (Cs >= 1970) {
return 8;
}
}
else if (AFD >= 10) {
if (Cs < 2322) {
return 1;
}
else if (Cs >= 2322) {
return 4;
}
}
}
else if (ssDCval >= 98) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (Rs < 3856) {
return 1;
}
else if (Rs >= 3856) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 88) {
return 1;
}
}
}
else if (AFD >= 28) {
return 4;
}
}
}
else if (Cs >= 2624) {
if (tcor < 90) {
if (refDCval < 91) {
if (RsDiff < 8) {
if (Rs < 2586) {
if (Cs < 4001) {
return 4;
}
else if (Cs >= 4001) {
return 8;
}
}
else if (Rs >= 2586) {
if (mcTcor < 92) {
if (Cs < 3490) {
return 8;
}
else if (Cs >= 3490) {
return 4;
}
}
else if (mcTcor >= 92) {
return 1;
}
}
}
else if (RsDiff >= 8) {
if (ssDCval < 73) {
if (Rs < 2905) {
return 1;
}
else if (Rs >= 2905) {
return 4;
}
}
else if (ssDCval >= 73) {
if (negBalance < 52) {
if (MVDiff < 4) {
if (AFD < 16) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
else if (AFD >= 16) {
return 1;
}
}
else if (MVDiff >= 4) {
if (Rs < 3292) {
return 8;
}
else if (Rs >= 3292) {
return 4;
}
}
}
else if (negBalance >= 52) {
if (posBalance < 59) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
}
else if (refDCval >= 91) {
if (SC < 5703) {
if (diffTSC < -1) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
return 4;
}
}
else if (diffTSC >= -1) {
if (Rs < 2995) {
return 8;
}
else if (Rs >= 2995) {
return 1;
}
}
}
else if (SC >= 5703) {
if (Cs < 4211) {
return 4;
}
else if (Cs >= 4211) {
if (Rs < 2319) {
return 4;
}
else if (Rs >= 2319) {
return 8;
}
}
}
}
}
else if (tcor >= 90) {
if (ssDCval < 104) {
if (ssDCval < 87) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (ssDCval >= 87) {
if (SC < 6748) {
if (Rs < 2684) {
return 2;
}
else if (Rs >= 2684) {
return 4;
}
}
else if (SC >= 6748) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
}
}
else if (ssDCval >= 104) {
if (Rs < 3449) {
return 8;
}
else if (Rs >= 3449) {
return 4;
}
}
}
}
}
}
else if (Rs >= 3964) {
if (AFD < 12) {
if (TSC < 9) {
if (posBalance < 41) {
if (mu_mv_mag_sq < 65) {
return 1;
}
else if (mu_mv_mag_sq >= 65) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (posBalance >= 41) {
if (ssDCval < 97) {
if (negBalance < 44) {
if (Rs < 4411) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Rs >= 4411) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
else if (negBalance >= 44) {
if (SC < 7120) {
return 4;
}
else if (SC >= 7120) {
return 8;
}
}
}
else if (ssDCval >= 97) {
if (Cs < 1906) {
return 1;
}
else if (Cs >= 1906) {
if (Rs < 5860) {
if (Cs < 2152) {
if (Rs < 4711) {
return 1;
}
else if (Rs >= 4711) {
return 8;
}
}
else if (Cs >= 2152) {
return 8;
}
}
else if (Rs >= 5860) {
if (Rs < 7062) {
return 4;
}
else if (Rs >= 7062) {
return 8;
}
}
}
}
}
}
else if (TSC >= 9) {
if (Cs < 3210) {
if (mu_mv_mag_sq < 92) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (RsCsDiff < 121) {
if (mu_mv_mag_sq < 84) {
return 8;
}
else if (mu_mv_mag_sq >= 84) {
return 1;
}
}
else if (RsCsDiff >= 121) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
else if (Cs >= 3210) {
return 8;
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 61) {
return 8;
}
else if (mu_mv_mag_sq >= 61) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (Rs < 4213) {
return 2;
}
else if (Rs >= 4213) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 10) {
if (refDCval < 109) {
if (negBalance < 45) {
if (refDCval < 73) {
if (mu_mv_mag_sq < 56) {
if (refDCval < 46) {
return 1;
}
else if (refDCval >= 46) {
if (Cs < 2630) {
if (diffRsCsdiff < 50) {
if (Rs < 2402) {
return 8;
}
else if (Rs >= 2402) {
return 2;
}
}
else if (diffRsCsdiff >= 50) {
if (Rs < 2988) {
return 2;
}
else if (Rs >= 2988) {
if (Rs < 3908) {
return 1;
}
else if (Rs >= 3908) {
return 4;
}
}
}
}
else if (Cs >= 2630) {
if (Cs < 3160) {
return 4;
}
else if (Cs >= 3160) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (negBalance < 34) {
if (RsCsDiff < 226) {
if (Rs < 3547) {
return 8;
}
else if (Rs >= 3547) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 2;
}
}
}
else if (RsCsDiff >= 226) {
return 8;
}
}
else if (negBalance >= 34) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (RsDiff < 12) {
if (tcor < 89) {
if (Rs < 3502) {
return 1;
}
else if (Rs >= 3502) {
return 8;
}
}
else if (tcor >= 89) {
return 4;
}
}
else if (RsDiff >= 12) {
return 8;
}
}
}
}
}
else if (refDCval >= 73) {
if (AFD < 14) {
if (negBalance < 43) {
if (Cs < 4693) {
if (RsCsDiff < 183) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 2;
}
}
}
else if (RsCsDiff >= 183) {
if (CsDiff < 12) {
return 2;
}
else if (CsDiff >= 12) {
if (Rs < 3101) {
return 2;
}
else if (Rs >= 3101) {
return 4;
}
}
}
}
else if (Cs >= 4693) {
return 8;
}
}
else if (negBalance >= 43) {
if (Cs < 2279) {
if (Rs < 2910) {
return 8;
}
else if (Rs >= 2910) {
return 2;
}
}
else if (Cs >= 2279) {
if (mu_mv_mag_sq < 57) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
else if (mu_mv_mag_sq >= 57) {
return 4;
}
}
}
}
else if (AFD >= 14) {
if (mu_mv_mag_sq < 15) {
if (Cs < 1892) {
if (diffMVdiffVal < -7) {
if (Rs < 2158) {
return 8;
}
else if (Rs >= 2158) {
if (Rs < 3539) {
return 1;
}
else if (Rs >= 3539) {
return 2;
}
}
}
else if (diffMVdiffVal >= -7) {
return 2;
}
}
else if (Cs >= 1892) {
if (refDCval < 95) {
if (diffMVdiffVal < -6) {
return 2;
}
else if (diffMVdiffVal >= -6) {
if (Rs < 2146) {
return 4;
}
else if (Rs >= 2146) {
return 8;
}
}
}
else if (refDCval >= 95) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
if (Cs < 1957) {
return 1;
}
else if (Cs >= 1957) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (posBalance < 50) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (CsDiff < 15) {
if (RsCsDiff < 183) {
if (tcor < 80) {
return 2;
}
else if (tcor >= 80) {
return 8;
}
}
else if (RsCsDiff >= 183) {
return 2;
}
}
else if (CsDiff >= 15) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (Rs < 4449) {
if (refDCval < 96) {
if (diffRsCsdiff < 72) {
return 8;
}
else if (diffRsCsdiff >= 72) {
if (AFD < 17) {
if (Rs < 2440) {
return 8;
}
else if (Rs >= 2440) {
return 4;
}
}
else if (AFD >= 17) {
return 8;
}
}
}
else if (refDCval >= 96) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (Rs >= 4449) {
if (Cs < 3092) {
if (RsCsDiff < 445) {
return 2;
}
else if (RsCsDiff >= 445) {
return 8;
}
}
else if (Cs >= 3092) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (Cs < 2427) {
if (RsDiff < 15) {
if (mu_mv_mag_sq < 5) {
if (RsCsDiff < 213) {
if (tcor < 87) {
if (Rs < 3551) {
if (Rs < 2598) {
if (SC < 3662) {
if (ssDCval < 93) {
if (Cs < 1638) {
return 4;
}
else if (Cs >= 1638) {
return 8;
}
}
else if (ssDCval >= 93) {
return 2;
}
}
else if (SC >= 3662) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
else if (Rs >= 2598) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
if (SC < 4413) {
return 4;
}
else if (SC >= 4413) {
return 8;
}
}
}
}
}
else if (Rs >= 3551) {
return 8;
}
}
else if (tcor >= 87) {
if (ssDCval < 98) {
if (tcor < 88) {
if (Rs < 2526) {
return 8;
}
else if (Rs >= 2526) {
return 2;
}
}
else if (tcor >= 88) {
return 2;
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
else if (RsCsDiff >= 213) {
if (AFD < 17) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
return 8;
}
}
else if (AFD >= 17) {
if (posBalance < 58) {
if (MVDiff < 10) {
if (RsCsDiff < 305) {
if (diffRsCsdiff < -78) {
return 8;
}
else if (diffRsCsdiff >= -78) {
return 4;
}
}
else if (RsCsDiff >= 305) {
return 8;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (posBalance >= 58) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (mcTcor < 93) {
if (ssDCval < 81) {
if (RsDiff < 12) {
return 4;
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (ssDCval >= 81) {
if (Cs < 2311) {
return 4;
}
else if (Cs >= 2311) {
if (Rs < 2304) {
return 1;
}
else if (Rs >= 2304) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
if (Rs < 2685) {
if (CsDiff < 13) {
return 2;
}
else if (CsDiff >= 13) {
return 4;
}
}
else if (Rs >= 2685) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (AFD < 22) {
if (ssDCval < 102) {
if (mu_mv_mag_sq < 45) {
if (mu_mv_mag_sq < 9) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (mu_mv_mag_sq >= 9) {
if (diffAFD < -2) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
return 8;
}
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < -74) {
if (Cs < 2319) {
return 8;
}
else if (Cs >= 2319) {
return 2;
}
}
else if (diffRsCsdiff >= -74) {
if (diffRsCsdiff < 148) {
return 8;
}
else if (diffRsCsdiff >= 148) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (diffRsCsdiff < 86) {
if (ssDCval < 93) {
if (Cs < 1745) {
if (diffRsCsdiff < -16) {
if (Rs < 3437) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
return 8;
}
}
else if (Rs >= 3437) {
return 2;
}
}
else if (diffRsCsdiff >= -16) {
return 2;
}
}
else if (Cs >= 1745) {
return 8;
}
}
else if (ssDCval >= 93) {
if (diffMVdiffVal < 5) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 4;
}
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
}
else if (diffRsCsdiff >= 86) {
if (Rs < 2704) {
return 8;
}
else if (Rs >= 2704) {
return 4;
}
}
}
}
else if (ssDCval >= 102) {
if (Rs < 3604) {
if (Cs < 1465) {
return 1;
}
else if (Cs >= 1465) {
return 8;
}
}
else if (Rs >= 3604) {
if (Rs < 4822) {
return 1;
}
else if (Rs >= 4822) {
return 4;
}
}
}
}
else if (AFD >= 22) {
if (gchDC < 6) {
if (RsDiff < 11) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 4;
}
}
else if (RsDiff >= 11) {
if (negBalance < 56) {
if (Rs < 2719) {
return 8;
}
else if (Rs >= 2719) {
return 2;
}
}
else if (negBalance >= 56) {
if (AFD < 28) {
if (RsDiff < 14) {
return 4;
}
else if (RsDiff >= 14) {
return 8;
}
}
else if (AFD >= 28) {
if (Rs < 1835) {
return 1;
}
else if (Rs >= 1835) {
if (Rs < 2002) {
return 2;
}
else if (Rs >= 2002) {
return 8;
}
}
}
}
}
}
else if (gchDC >= 6) {
return 2;
}
}
}
}
else if (RsDiff >= 15) {
if (Rs < 3862) {
if (SC < 5276) {
if (Cs < 2081) {
if (refDCval < 65) {
return 4;
}
else if (refDCval >= 65) {
if (ssDCval < 102) {
if (TSC < 12) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
if (Rs < 3683) {
if (SC < 5122) {
return 8;
}
else if (SC >= 5122) {
return 4;
}
}
else if (Rs >= 3683) {
return 2;
}
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (ssDCval >= 102) {
return 2;
}
}
}
else if (Cs >= 2081) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 4;
}
}
}
else if (SC >= 5276) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (Rs >= 3862) {
if (TSC < 13) {
if (Cs < 2290) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
if (Rs < 4185) {
return 2;
}
else if (Rs >= 4185) {
return 4;
}
}
}
else if (Cs >= 2290) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (TSC >= 13) {
if (AFD < 18) {
if (Cs < 2092) {
return 2;
}
else if (Cs >= 2092) {
return 4;
}
}
else if (AFD >= 18) {
return 1;
}
}
}
}
}
else if (Cs >= 2427) {
if (refDCval < 80) {
if (ssDCval < 73) {
if (Rs < 2744) {
return 1;
}
else if (Rs >= 2744) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 2;
}
}
}
else if (ssDCval >= 73) {
if (Cs < 3228) {
return 8;
}
else if (Cs >= 3228) {
if (Rs < 2512) {
return 8;
}
else if (Rs >= 2512) {
return 4;
}
}
}
}
else if (refDCval >= 80) {
if (Rs < 4394) {
if (MVDiff < 16) {
if (Cs < 3310) {
if (CsDiff < 10) {
if (SC < 5888) {
return 8;
}
else if (SC >= 5888) {
if (Cs < 2553) {
return 2;
}
else if (Cs >= 2553) {
if (ssDCval < 86) {
if (Rs < 3812) {
return 2;
}
else if (Rs >= 3812) {
return 1;
}
}
else if (ssDCval >= 86) {
return 4;
}
}
}
}
else if (CsDiff >= 10) {
if (AFD < 28) {
if (SC < 6761) {
return 4;
}
else if (SC >= 6761) {
if (SC < 6774) {
return 8;
}
else if (SC >= 6774) {
return 4;
}
}
}
else if (AFD >= 28) {
return 1;
}
}
}
else if (Cs >= 3310) {
if (ssDCval < 87) {
return 2;
}
else if (ssDCval >= 87) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 4;
}
}
}
}
else if (MVDiff >= 16) {
if (CsDiff < 16) {
if (Rs < 4187) {
return 8;
}
else if (Rs >= 4187) {
return 2;
}
}
else if (CsDiff >= 16) {
if (ssDCval < 99) {
return 2;
}
else if (ssDCval >= 99) {
return 1;
}
}
}
}
else if (Rs >= 4394) {
if (SC < 9706) {
if (SC < 8280) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (CsDiff < 10) {
if (Rs < 4484) {
return 8;
}
else if (Rs >= 4484) {
return 4;
}
}
else if (CsDiff >= 10) {
return 8;
}
}
}
else if (SC >= 8280) {
return 8;
}
}
else if (SC >= 9706) {
if (Rs < 8132) {
if (mu_mv_mag_sq < 29) {
return 4;
}
else if (mu_mv_mag_sq >= 29) {
return 8;
}
}
else if (Rs >= 8132) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (RsDiff < 14) {
if (Rs < 5297) {
if (posBalance < 52) {
if (negBalance < 32) {
if (CsDiff < 14) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (CsDiff >= 14) {
if (Rs < 2575) {
return 4;
}
else if (Rs >= 2575) {
if (Rs < 3259) {
return 8;
}
else if (Rs >= 3259) {
return 1;
}
}
}
}
else if (negBalance >= 32) {
if (tcor < 88) {
if (Rs < 2697) {
return 8;
}
else if (Rs >= 2697) {
if (TSC < 10) {
if (Rs < 3476) {
return 2;
}
else if (Rs >= 3476) {
return 8;
}
}
else if (TSC >= 10) {
return 4;
}
}
}
else if (tcor >= 88) {
if (Rs < 2407) {
return 4;
}
else if (Rs >= 2407) {
if (Cs < 3738) {
return 8;
}
else if (Cs >= 3738) {
return 2;
}
}
}
}
}
else if (posBalance >= 52) {
if (AFD < 22) {
if (SC < 3694) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (MVDiff < 30) {
return 8;
}
else if (MVDiff >= 30) {
if (Rs < 1909) {
return 2;
}
else if (Rs >= 1909) {
return 1;
}
}
}
}
else if (SC >= 3694) {
if (negBalance < 33) {
if (Rs < 2786) {
return 4;
}
else if (Rs >= 2786) {
return 2;
}
}
else if (negBalance >= 33) {
if (ssDCval < 163) {
if (RsCsDiff < 189) {
if (mcTcor < 92) {
if (mu_mv_mag_sq < 86) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
else if (mu_mv_mag_sq >= 86) {
return 1;
}
}
else if (mcTcor >= 92) {
if (ssDCval < 110) {
return 1;
}
else if (ssDCval >= 110) {
return 8;
}
}
}
else if (RsCsDiff >= 189) {
if (negBalance < 60) {
if (posBalance < 53) {
if (Rs < 3612) {
return 4;
}
else if (Rs >= 3612) {
return 8;
}
}
else if (posBalance >= 53) {
if (diffRsCsdiff < -92) {
if (Cs < 2717) {
return 8;
}
else if (Cs >= 2717) {
return 4;
}
}
else if (diffRsCsdiff >= -92) {
return 8;
}
}
}
else if (negBalance >= 60) {
return 4;
}
}
}
else if (ssDCval >= 163) {
return 4;
}
}
}
}
else if (AFD >= 22) {
if (Rs < 2824) {
if (Rs < 1988) {
return 1;
}
else if (Rs >= 1988) {
if (negBalance < 60) {
return 8;
}
else if (negBalance >= 60) {
if (Rs < 2269) {
return 1;
}
else if (Rs >= 2269) {
return 8;
}
}
}
}
else if (Rs >= 2824) {
return 1;
}
}
}
}
else if (Rs >= 5297) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (Cs < 4504) {
if (diffMVdiffVal < -11) {
return 8;
}
else if (diffMVdiffVal >= -11) {
return 4;
}
}
else if (Cs >= 4504) {
return 8;
}
}
}
}
else if (RsDiff >= 14) {
if (SC < 11960) {
if (negBalance < 58) {
if (negBalance < 42) {
if (Cs < 3141) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
if (AFD < 23) {
if (Rs < 7130) {
return 8;
}
else if (Rs >= 7130) {
if (Cs < 1580) {
return 8;
}
else if (Cs >= 1580) {
return 2;
}
}
}
else if (AFD >= 23) {
if (Cs < 1796) {
return 1;
}
else if (Cs >= 1796) {
return 2;
}
}
}
}
else if (Cs >= 3141) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
}
else if (negBalance >= 42) {
if (ssDCval < 147) {
if (MVDiff < 11) {
if (Cs < 2025) {
if (CsDiff < 11) {
return 2;
}
else if (CsDiff >= 11) {
return 4;
}
}
else if (Cs >= 2025) {
if (Rs < 5449) {
if (Cs < 3206) {
if (Cs < 2912) {
return 2;
}
else if (Cs >= 2912) {
return 4;
}
}
else if (Cs >= 3206) {
return 1;
}
}
else if (Rs >= 5449) {
if (Cs < 4743) {
return 4;
}
else if (Cs >= 4743) {
return 8;
}
}
}
}
else if (MVDiff >= 11) {
if (Rs < 5610) {
if (diffAFD < -2) {
if (gchDC < 4) {
if (Rs < 3144) {
if (Rs < 2845) {
return 8;
}
else if (Rs >= 2845) {
return 1;
}
}
else if (Rs >= 3144) {
return 2;
}
}
else if (gchDC >= 4) {
return 1;
}
}
else if (diffAFD >= -2) {
if (SC < 7005) {
if (RsDiff < 19) {
if (tcor < 70) {
if (Rs < 3308) {
return 2;
}
else if (Rs >= 3308) {
return 4;
}
}
else if (tcor >= 70) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (RsDiff >= 19) {
if (RsCsDiff < 641) {
if (Rs < 4746) {
return 2;
}
else if (Rs >= 4746) {
if (Rs < 5052) {
return 4;
}
else if (Rs >= 5052) {
return 8;
}
}
}
else if (RsCsDiff >= 641) {
return 8;
}
}
}
else if (SC >= 7005) {
if (RsCsDiff < 451) {
return 1;
}
else if (RsCsDiff >= 451) {
return 2;
}
}
}
}
else if (Rs >= 5610) {
if (RsCsDiff < 609) {
if (SC < 7689) {
if (Rs < 5974) {
return 8;
}
else if (Rs >= 5974) {
return 2;
}
}
else if (SC >= 7689) {
return 8;
}
}
else if (RsCsDiff >= 609) {
if (diffMVdiffVal < -13) {
return 8;
}
else if (diffMVdiffVal >= -13) {
return 2;
}
}
}
}
}
else if (ssDCval >= 147) {
if (ssDCval < 163) {
if (SC < 5665) {
if (Rs < 3814) {
return 2;
}
else if (Rs >= 3814) {
return 1;
}
}
else if (SC >= 5665) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
if (RsCsDiff < 857) {
if (MVDiff < 153) {
return 4;
}
else if (MVDiff >= 153) {
return 8;
}
}
else if (RsCsDiff >= 857) {
if (Cs < 2300) {
return 4;
}
else if (Cs >= 2300) {
if (Rs < 5467) {
return 8;
}
else if (Rs >= 5467) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 163) {
if (AFD < 32) {
return 2;
}
else if (AFD >= 32) {
return 1;
}
}
}
}
}
else if (negBalance >= 58) {
if (AFD < 18) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
if (Rs < 3591) {
return 8;
}
else if (Rs >= 3591) {
return 1;
}
}
}
else if (AFD >= 18) {
if (RsCsDiff < 359) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
else if (RsCsDiff >= 359) {
if (Cs < 1476) {
return 1;
}
else if (Cs >= 1476) {
if (SC < 6403) {
if (Rs < 3742) {
if (Cs < 2296) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
else if (Cs >= 2296) {
if (AFD < 27) {
return 1;
}
else if (AFD >= 27) {
return 8;
}
}
}
else if (Rs >= 3742) {
return 1;
}
}
else if (SC >= 6403) {
if (Cs < 3686) {
return 8;
}
else if (Cs >= 3686) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 11960) {
if (ssDCval < 128) {
if (RsDiff < 44) {
return 1;
}
else if (RsDiff >= 44) {
return 8;
}
}
else if (ssDCval >= 128) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Cs < 1641) {
if (Rs < 3127) {
if (Rs < 1459) {
if (ssDCval < 122) {
if (refDCval < 63) {
if (negBalance < 26) {
if (mu_mv_mag_sq < 128) {
if (Cs < 1345) {
if (ssDCval < 55) {
if (Rs < 1391) {
if (Rs < 1139) {
return 8;
}
else if (Rs >= 1139) {
return 4;
}
}
else if (Rs >= 1391) {
return 8;
}
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (Cs >= 1345) {
if (MVDiff < 7) {
if (negBalance < 24) {
if (AFD < 3) {
if (Cs < 1353) {
if (Rs < 1307) {
return 1;
}
else if (Rs >= 1307) {
return 8;
}
}
else if (Cs >= 1353) {
return 1;
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (negBalance >= 24) {
if (Rs < 1207) {
return 1;
}
else if (Rs >= 1207) {
if (Cs < 1512) {
return 4;
}
else if (Cs >= 1512) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 57) {
if (Rs < 758) {
return 1;
}
else if (Rs >= 758) {
return 4;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1498) {
if (Cs < 1332) {
return 4;
}
else if (Cs >= 1332) {
return 8;
}
}
else if (Cs >= 1498) {
return 1;
}
}
}
else if (negBalance >= 26) {
if (Rs < 1129) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 664) {
return 1;
}
else if (Rs >= 664) {
if (Rs < 911) {
return 8;
}
else if (Rs >= 911) {
return 4;
}
}
}
}
else if (Rs >= 1129) {
if (mu_mv_mag_sq < 121) {
if (ssDCval < 58) {
if (SC < 2856) {
return 4;
}
else if (SC >= 2856) {
if (Rs < 1436) {
return 8;
}
else if (Rs >= 1436) {
return 4;
}
}
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (mu_mv_mag_sq >= 121) {
if (Cs < 1484) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (Cs >= 1484) {
return 4;
}
}
}
}
}
else if (refDCval >= 63) {
if (AFD < 1) {
if (Rs < 605) {
return 4;
}
else if (Rs >= 605) {
return 2;
}
}
else if (AFD >= 1) {
if (ssDCval < 77) {
if (negBalance < 31) {
if (Rs < 1414) {
return 8;
}
else if (Rs >= 1414) {
return 1;
}
}
else if (negBalance >= 31) {
if (MVDiff < 2) {
if (Cs < 1337) {
if (Rs < 1129) {
return 2;
}
else if (Rs >= 1129) {
return 1;
}
}
else if (Cs >= 1337) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (ssDCval >= 77) {
if (Rs < 1072) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Rs >= 1072) {
return 8;
}
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 2) {
if (posBalance < 28) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (posBalance >= 28) {
if (Rs < 1383) {
return 4;
}
else if (Rs >= 1383) {
return 8;
}
}
}
else if (AFD >= 2) {
if (tcor < 96) {
if (Rs < 945) {
if (Rs < 582) {
return 2;
}
else if (Rs >= 582) {
return 8;
}
}
else if (Rs >= 945) {
if (Rs < 1205) {
if (ssDCval < 129) {
return 4;
}
else if (ssDCval >= 129) {
if (Rs < 1080) {
return 8;
}
else if (Rs >= 1080) {
return 4;
}
}
}
else if (Rs >= 1205) {
if (Rs < 1296) {
return 8;
}
else if (Rs >= 1296) {
return 1;
}
}
}
}
else if (tcor >= 96) {
if (mu_mv_mag_sq < 121) {
return 8;
}
else if (mu_mv_mag_sq >= 121) {
if (Rs < 1024) {
return 8;
}
else if (Rs >= 1024) {
if (Rs < 1265) {
return 4;
}
else if (Rs >= 1265) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1459) {
if (ssDCval < 86) {
if (mcTcor < 96) {
if (ssDCval < 45) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 2022) {
return 8;
}
else if (Rs >= 2022) {
return 4;
}
}
}
else if (ssDCval >= 45) {
return 8;
}
}
else if (mcTcor >= 96) {
if (negBalance < 18) {
if (SC < 4469) {
if (Rs < 1981) {
if (Cs < 1427) {
return 8;
}
else if (Cs >= 1427) {
if (CsDiff < 2) {
if (ssDCval < 38) {
return 8;
}
else if (ssDCval >= 38) {
return 1;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (Rs >= 1981) {
if (Cs < 1350) {
if (Rs < 2339) {
if (Rs < 2287) {
return 8;
}
else if (Rs >= 2287) {
if (mu_mv_mag_sq < 127) {
if (Cs < 1329) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (posBalance < 31) {
if (Rs < 2293) {
return 2;
}
else if (Rs >= 2293) {
return 8;
}
}
else if (posBalance >= 31) {
return 2;
}
}
}
else if (Cs >= 1329) {
return 8;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
return 2;
}
}
}
}
else if (Rs >= 2339) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (Rs < 2366) {
return 2;
}
else if (Rs >= 2366) {
return 4;
}
}
}
}
else if (Cs >= 1350) {
return 8;
}
}
}
else if (SC >= 4469) {
if (Rs < 3088) {
return 1;
}
else if (Rs >= 3088) {
return 8;
}
}
}
else if (negBalance >= 18) {
if (negBalance < 32) {
if (Rs < 1545) {
if (ssDCval < 77) {
if (negBalance < 24) {
if (Rs < 1521) {
return 1;
}
else if (Rs >= 1521) {
return 8;
}
}
else if (negBalance >= 24) {
return 4;
}
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (Rs >= 1545) {
if (Rs < 1879) {
return 4;
}
else if (Rs >= 1879) {
if (Rs < 2426) {
if (RsDiff < 3) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
return 8;
}
}
else if (RsDiff >= 3) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
return 4;
}
}
}
else if (Rs >= 2426) {
if (Cs < 1396) {
return 2;
}
else if (Cs >= 1396) {
return 4;
}
}
}
}
}
else if (negBalance >= 32) {
if (posBalance < 48) {
if (RsDiff < 6) {
if (Rs < 1930) {
return 4;
}
else if (Rs >= 1930) {
if (diffMVdiffVal < 1) {
if (Rs < 2466) {
return 8;
}
else if (Rs >= 2466) {
if (Rs < 2673) {
return 4;
}
else if (Rs >= 2673) {
if (gchDC < 2) {
if (Rs < 2802) {
return 8;
}
else if (Rs >= 2802) {
return 8;
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (posBalance >= 48) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (MVDiff < 2) {
if (ssDCval < 117) {
if (posBalance < 60) {
if (diffAFD < -2) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 2;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (posBalance >= 60) {
if (SC < 3593) {
return 8;
}
else if (SC >= 3593) {
if (Rs < 2304) {
return 2;
}
else if (Rs >= 2304) {
return 1;
}
}
}
}
else if (ssDCval >= 117) {
if (Rs < 2864) {
if (ssDCval < 119) {
if (mu_mv_mag_sq < 120) {
return 8;
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 2032) {
return 1;
}
else if (Rs >= 2032) {
return 8;
}
}
}
else if (ssDCval >= 119) {
return 8;
}
}
else if (Rs >= 2864) {
return 2;
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 94) {
if (Cs < 1495) {
if (Rs < 1914) {
return 8;
}
else if (Rs >= 1914) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
}
else if (Cs >= 1495) {
return 8;
}
}
else if (ssDCval >= 94) {
if (mcTcor < 77) {
return 4;
}
else if (mcTcor >= 77) {
if (tcor < 101) {
return 8;
}
else if (tcor >= 101) {
if (Rs < 1855) {
return 8;
}
else if (Rs >= 1855) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Cs < 1568) {
return 2;
}
else if (Cs >= 1568) {
return 1;
}
}
}
}
}
}
else if (Rs >= 3127) {
if (refDCval < 78) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 63) {
if (negBalance < 21) {
if (Rs < 5513) {
if (diffAFD < -1) {
if (negBalance < 14) {
return 4;
}
else if (negBalance >= 14) {
return 8;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (Rs >= 5513) {
return 8;
}
}
else if (negBalance >= 21) {
return 8;
}
}
else if (ssDCval >= 63) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (ssDCval < 74) {
if (Rs < 3262) {
return 2;
}
else if (Rs >= 3262) {
return 4;
}
}
else if (ssDCval >= 74) {
if (RsDiff < 5) {
if (Rs < 4355) {
return 4;
}
else if (Rs >= 4355) {
return 8;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1570) {
if (SC < 4996) {
if (Cs < 1321) {
return 8;
}
else if (Cs >= 1321) {
return 4;
}
}
else if (SC >= 4996) {
if (RsCsDiff < 12) {
return 8;
}
else if (RsCsDiff >= 12) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 4;
}
}
}
}
else if (Cs >= 1570) {
if (Rs < 3265) {
return 8;
}
else if (Rs >= 3265) {
return 4;
}
}
}
}
else if (refDCval >= 78) {
if (ssDCval < 110) {
if (posBalance < 42) {
if (Rs < 4493) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
if (ssDCval < 96) {
if (Rs < 3899) {
if (Rs < 3602) {
return 2;
}
else if (Rs >= 3602) {
return 4;
}
}
else if (Rs >= 3899) {
return 8;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (Rs >= 4493) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (refDCval < 90) {
if (negBalance < 26) {
if (Rs < 3403) {
return 4;
}
else if (Rs >= 3403) {
return 2;
}
}
else if (negBalance >= 26) {
return 2;
}
}
else if (refDCval >= 90) {
if (Cs < 1534) {
if (Rs < 3263) {
return 2;
}
else if (Rs >= 3263) {
if (Rs < 4483) {
return 4;
}
else if (Rs >= 4483) {
return 2;
}
}
}
else if (Cs >= 1534) {
if (ssDCval < 101) {
return 8;
}
else if (ssDCval >= 101) {
if (Rs < 4319) {
return 4;
}
else if (Rs >= 4319) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 110) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (Rs < 3348) {
return 2;
}
else if (Rs >= 3348) {
if (Cs < 1549) {
return 4;
}
else if (Cs >= 1549) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1641) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (refDCval < 65) {
if (Rs < 8850) {
if (tcor < 92) {
if (Cs < 2088) {
return 8;
}
else if (Cs >= 2088) {
return 4;
}
}
else if (tcor >= 92) {
return 8;
}
}
else if (Rs >= 8850) {
return 2;
}
}
else if (refDCval >= 65) {
if (diffTSC < -1) {
if (diffRsCsdiff < -1) {
if (RsDiff < 1) {
if (ssDCval < 69) {
if (Rs < 3548) {
return 4;
}
else if (Rs >= 3548) {
return 8;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (RsDiff >= 1) {
return 4;
}
}
else if (diffRsCsdiff >= -1) {
return 4;
}
}
else if (diffTSC >= -1) {
if (ssDCval < 93) {
if (Cs < 2038) {
return 8;
}
else if (Cs >= 2038) {
return 4;
}
}
else if (ssDCval >= 93) {
return 8;
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (refDCval < 67) {
if (posBalance < 40) {
if (ssDCval < 41) {
return 8;
}
else if (ssDCval >= 41) {
if (SC < 6372) {
if (negBalance < 16) {
if (Rs < 1372) {
return 8;
}
else if (Rs >= 1372) {
if (Rs < 3257) {
return 1;
}
else if (Rs >= 3257) {
if (Rs < 3262) {
return 4;
}
else if (Rs >= 3262) {
return 1;
}
}
}
}
else if (negBalance >= 16) {
if (mu_mv_mag_sq < 121) {
if (Rs < 1002) {
return 8;
}
else if (Rs >= 1002) {
return 4;
}
}
else if (mu_mv_mag_sq >= 121) {
if (posBalance < 39) {
if (refDCval < 66) {
if (negBalance < 26) {
return 8;
}
else if (negBalance >= 26) {
return 1;
}
}
else if (refDCval >= 66) {
return 1;
}
}
else if (posBalance >= 39) {
if (Rs < 2637) {
return 1;
}
else if (Rs >= 2637) {
return 4;
}
}
}
}
}
else if (SC >= 6372) {
if (Rs < 3506) {
return 8;
}
else if (Rs >= 3506) {
return 4;
}
}
}
}
else if (posBalance >= 40) {
if (Rs < 1668) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (ssDCval < 59) {
if (Rs < 994) {
return 8;
}
else if (Rs >= 994) {
return 4;
}
}
else if (ssDCval >= 59) {
if (Cs < 1983) {
return 1;
}
else if (Cs >= 1983) {
return 8;
}
}
}
}
else if (Rs >= 1668) {
if (negBalance < 19) {
if (Rs < 2651) {
return 8;
}
else if (Rs >= 2651) {
if (Cs < 3441) {
if (Cs < 2085) {
return 8;
}
else if (Cs >= 2085) {
return 4;
}
}
else if (Cs >= 3441) {
return 8;
}
}
}
else if (negBalance >= 19) {
if (Cs < 1815) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
if (Rs < 2412) {
return 8;
}
else if (Rs >= 2412) {
return 1;
}
}
}
else if (Cs >= 1815) {
return 8;
}
}
}
}
}
else if (refDCval >= 67) {
if (mu_mv_mag_sq < 122) {
if (posBalance < 30) {
if (ssDCval < 92) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (MVDiff < 2) {
if (Rs < 2373) {
return 1;
}
else if (Rs >= 2373) {
if (Rs < 2476) {
return 8;
}
else if (Rs >= 2476) {
return 1;
}
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (ssDCval >= 92) {
if (diffMVdiffVal < -1) {
if (tcor < 99) {
if (Rs < 2324) {
return 1;
}
else if (Rs >= 2324) {
return 2;
}
}
else if (tcor >= 99) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2323) {
return 2;
}
else if (Rs >= 2323) {
if (Cs < 2117) {
return 4;
}
else if (Cs >= 2117) {
if (Cs < 2140) {
return 2;
}
else if (Cs >= 2140) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 30) {
if (Cs < 2232) {
if (Rs < 2107) {
return 8;
}
else if (Rs >= 2107) {
if (Cs < 2110) {
return 2;
}
else if (Cs >= 2110) {
if (Rs < 3935) {
return 8;
}
else if (Rs >= 3935) {
return 4;
}
}
}
}
else if (Cs >= 2232) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 1112) {
if (Rs < 623) {
return 8;
}
else if (Rs >= 623) {
return 1;
}
}
else if (Rs >= 1112) {
if (SC < 7112) {
if (MVDiff < 4) {
if (SC < 4488) {
if (Cs < 2096) {
if (SC < 4440) {
return 8;
}
else if (SC >= 4440) {
if (Rs < 2636) {
return 4;
}
else if (Rs >= 2636) {
return 8;
}
}
}
else if (Cs >= 2096) {
return 4;
}
}
else if (SC >= 4488) {
if (mu_mv_mag_sq < 124) {
if (Cs < 2137) {
if (Rs < 3229) {
return 8;
}
else if (Rs >= 3229) {
return 2;
}
}
else if (Cs >= 2137) {
return 8;
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
return 8;
}
}
}
else if (SC >= 7112) {
if (SC < 7964) {
if (Rs < 3296) {
return 8;
}
else if (Rs >= 3296) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 4;
}
}
}
else if (SC >= 7964) {
if (ssDCval < 72) {
if (Rs < 4925) {
return 1;
}
else if (Rs >= 4925) {
return 8;
}
}
else if (ssDCval >= 72) {
if (Rs < 6705) {
return 8;
}
else if (Rs >= 6705) {
return 4;
}
}
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6807) {
if (SC < 8083) {
if (Rs < 3674) {
if (AFD < 10) {
if (tcor < 99) {
if (SC < 2897) {
if (ssDCval < 72) {
if (Rs < 1137) {
if (SC < 2830) {
return 8;
}
else if (SC >= 2830) {
return 4;
}
}
else if (Rs >= 1137) {
return 4;
}
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (SC >= 2897) {
if (SC < 4097) {
if (negBalance < 63) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (negBalance >= 63) {
if (SC < 3788) {
return 8;
}
else if (SC >= 3788) {
return 4;
}
}
}
else if (SC >= 4097) {
if (Cs < 1881) {
if (mcTcor < 97) {
if (Rs < 3079) {
return 8;
}
else if (Rs >= 3079) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (Rs < 3352) {
return 2;
}
else if (Rs >= 3352) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 80) {
if (SC < 4805) {
if (mcTcor < 98) {
if (ssDCval < 65) {
if (Rs < 2636) {
return 4;
}
else if (Rs >= 2636) {
return 4;
}
}
else if (ssDCval >= 65) {
return 8;
}
}
else if (mcTcor >= 98) {
if (posBalance < 39) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
return 4;
}
}
else if (posBalance >= 39) {
return 8;
}
}
}
else if (SC >= 4805) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (Rs < 3110) {
if (diffRsCsdiff < -2) {
if (Rs < 3039) {
return 4;
}
else if (Rs >= 3039) {
return 4;
}
}
else if (diffRsCsdiff >= -2) {
return 4;
}
}
else if (Rs >= 3110) {
return 8;
}
}
}
}
else if (ssDCval >= 80) {
if (MVDiff < 1) {
if (RsDiff < 3) {
if (Rs < 2671) {
return 4;
}
else if (Rs >= 2671) {
return 8;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (MVDiff >= 1) {
if (Rs < 2943) {
return 8;
}
else if (Rs >= 2943) {
if (Cs < 1677) {
return 2;
}
else if (Cs >= 1677) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1881) {
if (Rs < 2145) {
if (SC < 4346) {
if (SC < 4100) {
return 4;
}
else if (SC >= 4100) {
return 8;
}
}
else if (SC >= 4346) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 4;
}
}
}
}
else if (Rs >= 2145) {
if (SC < 6962) {
if (mu_mv_mag_sq < 121) {
if (Cs < 2458) {
if (ssDCval < 70) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 4;
}
}
else if (ssDCval >= 70) {
if (ssDCval < 84) {
if (Cs < 1954) {
if (Cs < 1928) {
return 8;
}
else if (Cs >= 1928) {
if (Rs < 2350) {
return 8;
}
else if (Rs >= 2350) {
return 4;
}
}
}
else if (Cs >= 1954) {
if (posBalance < 65) {
if (Cs < 2377) {
return 8;
}
else if (Cs >= 2377) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (negBalance < 43) {
if (Rs < 3292) {
return 4;
}
else if (Rs >= 3292) {
return 8;
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
else if (posBalance >= 65) {
return 4;
}
}
}
else if (ssDCval >= 84) {
if (negBalance < 62) {
return 8;
}
else if (negBalance >= 62) {
return 4;
}
}
}
}
else if (Cs >= 2458) {
if (Rs < 2503) {
return 4;
}
else if (Rs >= 2503) {
if (ssDCval < 87) {
if (RsCsDiff < 92) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 4;
}
}
else if (RsCsDiff >= 92) {
if (Rs < 3081) {
return 8;
}
else if (Rs >= 3081) {
return 4;
}
}
}
else if (ssDCval >= 87) {
if (Cs < 2685) {
if (negBalance < 47) {
if (SC < 6255) {
return 8;
}
else if (SC >= 6255) {
return 4;
}
}
else if (negBalance >= 47) {
return 4;
}
}
else if (Cs >= 2685) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (Cs < 2014) {
if (ssDCval < 66) {
if (Rs < 2697) {
return 8;
}
else if (Rs >= 2697) {
if (Cs < 1913) {
if (Rs < 2949) {
return 8;
}
else if (Rs >= 2949) {
if (Rs < 2968) {
return 8;
}
else if (Rs >= 2968) {
return 4;
}
}
}
else if (Cs >= 1913) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
return 4;
}
}
}
}
else if (ssDCval >= 66) {
if (Rs < 3108) {
if (mu_mv_mag_sq < 122) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 122) {
return 8;
}
}
else if (Rs >= 3108) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
return 2;
}
}
}
}
else if (Cs >= 2014) {
if (diffAFD < -2) {
if (SC < 6288) {
if (AFD < 8) {
if (diffMVdiffVal < -5) {
if (Rs < 2592) {
return 4;
}
else if (Rs >= 2592) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (AFD >= 8) {
if (Rs < 2999) {
return 8;
}
else if (Rs >= 2999) {
return 4;
}
}
}
else if (SC >= 6288) {
return 4;
}
}
else if (diffAFD >= -2) {
if (SC < 6840) {
if (TSC < 9) {
if (Cs < 3554) {
if (refDCval < 67) {
if (posBalance < 44) {
if (Rs < 3000) {
return 8;
}
else if (Rs >= 3000) {
if (Rs < 3246) {
return 4;
}
else if (Rs >= 3246) {
if (Rs < 3390) {
return 1;
}
else if (Rs >= 3390) {
return 8;
}
}
}
}
else if (posBalance >= 44) {
return 8;
}
}
else if (refDCval >= 67) {
if (Rs < 2186) {
if (Rs < 2182) {
return 8;
}
else if (Rs >= 2182) {
return 4;
}
}
else if (Rs >= 2186) {
if (refDCval < 96) {
return 8;
}
else if (refDCval >= 96) {
if (Cs < 2655) {
return 8;
}
else if (Cs >= 2655) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Cs < 3128) {
return 4;
}
else if (Cs >= 3128) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 3554) {
if (Rs < 2877) {
return 4;
}
else if (Rs >= 2877) {
return 8;
}
}
}
else if (TSC >= 9) {
if (Cs < 3167) {
if (Rs < 3038) {
return 8;
}
else if (Rs >= 3038) {
return 4;
}
}
else if (Cs >= 3167) {
return 8;
}
}
}
else if (SC >= 6840) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (SC < 6845) {
return 4;
}
else if (SC >= 6845) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 6962) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
if (AFD < 6) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (AFD >= 6) {
if (Cs < 4245) {
return 8;
}
else if (Cs >= 4245) {
return 4;
}
}
}
}
}
}
}
}
}
else if (tcor >= 99) {
if (Cs < 2284) {
if (SC < 5137) {
if (Rs < 2544) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
else if (Rs >= 2544) {
if (Cs < 1902) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
return 4;
}
}
else if (Cs >= 1902) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 3020) {
return 4;
}
else if (Rs >= 3020) {
return 2;
}
}
}
}
}
else if (SC >= 5137) {
if (mu_mv_mag_sq < 126) {
return 2;
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 2188) {
return 2;
}
else if (Cs >= 2188) {
return 8;
}
}
}
}
else if (Cs >= 2284) {
if (diffAFD < -3) {
if (Cs < 2782) {
return 8;
}
else if (Cs >= 2782) {
return 4;
}
}
else if (diffAFD >= -3) {
if (SC < 7195) {
if (Rs < 2558) {
return 4;
}
else if (Rs >= 2558) {
return 8;
}
}
else if (SC >= 7195) {
return 4;
}
}
}
}
}
else if (AFD >= 10) {
if (Cs < 2301) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 1838) {
return 8;
}
else if (Rs >= 1838) {
return 4;
}
}
}
else if (Cs >= 2301) {
if (negBalance < 55) {
if (Cs < 3156) {
if (Cs < 2678) {
if (RsDiff < 8) {
if (diffAFD < 1) {
if (Rs < 3010) {
return 8;
}
else if (Rs >= 3010) {
return 4;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (Cs >= 2678) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
if (Rs < 3085) {
return 8;
}
else if (Rs >= 3085) {
return 4;
}
}
}
}
}
else if (Cs >= 3156) {
if (posBalance < 62) {
if (refDCval < 92) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
if (diffAFD < -1) {
if (Rs < 3541) {
return 4;
}
else if (Rs >= 3541) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (refDCval >= 92) {
return 8;
}
}
else if (posBalance >= 62) {
if (Rs < 1593) {
return 2;
}
else if (Rs >= 1593) {
return 4;
}
}
}
}
else if (negBalance >= 55) {
return 4;
}
}
}
}
else if (Rs >= 3674) {
if (refDCval < 85) {
if (Cs < 2739) {
if (SC < 5895) {
if (Rs < 4203) {
return 2;
}
else if (Rs >= 4203) {
return 4;
}
}
else if (SC >= 5895) {
if (RsCsDiff < 12) {
if (posBalance < 39) {
if (Cs < 2246) {
return 4;
}
else if (Cs >= 2246) {
if (Rs < 4891) {
return 8;
}
else if (Rs >= 4891) {
if (Rs < 5065) {
return 4;
}
else if (Rs >= 5065) {
return 8;
}
}
}
}
else if (posBalance >= 39) {
return 8;
}
}
else if (RsCsDiff >= 12) {
if (Rs < 5046) {
if (ssDCval < 73) {
if (Rs < 4026) {
return 4;
}
else if (Rs >= 4026) {
if (mu_mv_mag_sq < 123) {
if (diffRsCsdiff < 4) {
if (RsCsDiff < 46) {
if (diffAFD < 1) {
if (mcTcor < 96) {
if (Cs < 2608) {
return 8;
}
else if (Cs >= 2608) {
if (Rs < 4875) {
return 8;
}
else if (Rs >= 4875) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (mcTcor < 98) {
if (SC < 7396) {
if (Rs < 4850) {
if (diffMVdiffVal < 2) {
if (ssDCval < 54) {
if (MVDiff < 2) {
if (Cs < 2507) {
return 4;
}
else if (Cs >= 2507) {
if (Rs < 4711) {
return 4;
}
else if (Rs >= 4711) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (Rs < 4494) {
return 8;
}
else if (Rs >= 4494) {
return 4;
}
}
}
else if (ssDCval >= 54) {
if (posBalance < 27) {
return 8;
}
else if (posBalance >= 27) {
if (AFD < 4) {
if (posBalance < 31) {
return 4;
}
else if (posBalance >= 31) {
if (Cs < 2488) {
if (Rs < 4310) {
return 8;
}
else if (Rs >= 4310) {
return 8;
}
}
else if (Cs >= 2488) {
return 4;
}
}
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 2) {
return 4;
}
}
else if (Rs >= 4850) {
return 8;
}
}
else if (SC >= 7396) {
if (Rs < 4964) {
return 4;
}
else if (Rs >= 4964) {
return 4;
}
}
}
else if (mcTcor >= 98) {
return 8;
}
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (RsCsDiff >= 46) {
if (Rs < 4338) {
return 4;
}
else if (Rs >= 4338) {
if (Cs < 2252) {
if (Rs < 4484) {
return 8;
}
else if (Rs >= 4484) {
return 4;
}
}
else if (Cs >= 2252) {
if (gchDC < 1) {
if (Cs < 2431) {
return 4;
}
else if (Cs >= 2431) {
return 4;
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
}
else if (diffRsCsdiff >= 4) {
if (MVDiff < 4) {
if (diffMVdiffVal < 1) {
if (mu_mv_mag_sq < 119) {
if (RsDiff < 6) {
if (diffMVdiffVal < -1) {
if (Rs < 4463) {
return 8;
}
else if (Rs >= 4463) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (RsDiff >= 6) {
if (posBalance < 44) {
if (TSC < 4) {
if (RsDiff < 8) {
if (SC < 6742) {
return 4;
}
else if (SC >= 6742) {
if (Rs < 4701) {
return 4;
}
else if (Rs >= 4701) {
return 4;
}
}
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (TSC >= 4) {
return 4;
}
}
else if (posBalance >= 44) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 4249) {
if (Rs < 4165) {
return 4;
}
else if (Rs >= 4165) {
return 4;
}
}
else if (Rs >= 4249) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
if (tcor < 94) {
if (Rs < 4575) {
return 4;
}
else if (Rs >= 4575) {
return 4;
}
}
else if (tcor >= 94) {
return 4;
}
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (Rs >= 5046) {
if (mu_mv_mag_sq < 118) {
return 4;
}
else if (mu_mv_mag_sq >= 118) {
if (mcTcor < 98) {
if (Cs < 2625) {
return 4;
}
else if (Cs >= 2625) {
return 4;
}
}
else if (mcTcor >= 98) {
return 4;
}
}
}
}
}
}
else if (Cs >= 2739) {
return 8;
}
}
else if (refDCval >= 85) {
if (Cs < 2740) {
if (CsDiff < 7) {
if (negBalance < 28) {
if (Cs < 2229) {
if (ssDCval < 96) {
if (Rs < 5070) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 2;
}
}
else if (Rs >= 5070) {
return 4;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
else if (Cs >= 2229) {
return 8;
}
}
else if (negBalance >= 28) {
if (ssDCval < 91) {
if (Cs < 1731) {
return 8;
}
else if (Cs >= 1731) {
return 2;
}
}
else if (ssDCval >= 91) {
if (RsDiff < 7) {
if (ssDCval < 99) {
return 8;
}
else if (ssDCval >= 99) {
return 4;
}
}
else if (RsDiff >= 7) {
if (RsDiff < 8) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
}
else if (CsDiff >= 7) {
if (Rs < 3987) {
return 4;
}
else if (Rs >= 3987) {
return 1;
}
}
}
else if (Cs >= 2740) {
if (ssDCval < 95) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
else if (ssDCval >= 95) {
if (Rs < 3973) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (Rs >= 3973) {
if (Cs < 3525) {
return 4;
}
else if (Cs >= 3525) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 8083) {
if (posBalance < 48) {
if (mu_mv_mag_sq < 114) {
if (Rs < 5546) {
return 8;
}
else if (Rs >= 5546) {
return 2;
}
}
else if (mu_mv_mag_sq >= 114) {
if (Rs < 4471) {
if (Cs < 4210) {
return 8;
}
else if (Cs >= 4210) {
return 1;
}
}
else if (Rs >= 4471) {
if (Cs < 2358) {
return 4;
}
else if (Cs >= 2358) {
if (mu_mv_mag_sq < 119) {
if (Cs < 4376) {
return 8;
}
else if (Cs >= 4376) {
return 2;
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
}
}
}
else if (posBalance >= 48) {
if (Cs < 3666) {
if (ssDCval < 98) {
if (Rs < 4595) {
return 8;
}
else if (Rs >= 4595) {
return 4;
}
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (Cs >= 3666) {
if (Rs < 6621) {
if (Rs < 3758) {
if (Rs < 3318) {
return 8;
}
else if (Rs >= 3318) {
return 4;
}
}
else if (Rs >= 3758) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
return 8;
}
}
}
else if (Rs >= 6621) {
return 4;
}
}
}
}
}
else if (Rs >= 6807) {
if (Rs < 9093) {
return 4;
}
else if (Rs >= 9093) {
return 2;
}
}
}
}
}
else if (ssDCval >= 102) {
if (mcTcor < 96) {
if (ssDCval < 137) {
if (refDCval < 105) {
if (negBalance < 43) {
return 8;
}
else if (negBalance >= 43) {
if (SC < 7207) {
if (Rs < 3699) {
return 8;
}
else if (Rs >= 3699) {
return 1;
}
}
else if (SC >= 7207) {
if (Cs < 2995) {
return 2;
}
else if (Cs >= 2995) {
return 8;
}
}
}
}
else if (refDCval >= 105) {
if (posBalance < 63) {
if (TSC < 9) {
if (refDCval < 123) {
if (diffTSC < 4) {
return 8;
}
else if (diffTSC >= 4) {
if (Cs < 2051) {
return 4;
}
else if (Cs >= 2051) {
return 8;
}
}
}
else if (refDCval >= 123) {
if (diffMVdiffVal < -1) {
if (ssDCval < 124) {
if (Rs < 3498) {
return 1;
}
else if (Rs >= 3498) {
return 4;
}
}
else if (ssDCval >= 124) {
if (mu_mv_mag_sq < 117) {
if (Rs < 3715) {
return 1;
}
else if (Rs >= 3715) {
return 8;
}
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
}
else if (diffMVdiffVal >= -1) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 5345) {
return 4;
}
else if (Rs >= 5345) {
return 8;
}
}
}
}
}
else if (TSC >= 9) {
if (SC < 7077) {
if (Rs < 3153) {
if (AFD < 19) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (AFD >= 19) {
if (Rs < 925) {
return 2;
}
else if (Rs >= 925) {
if (Rs < 2055) {
return 4;
}
else if (Rs >= 2055) {
return 1;
}
}
}
}
else if (Rs >= 3153) {
if (negBalance < 54) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
if (Cs < 1997) {
return 2;
}
else if (Cs >= 1997) {
return 8;
}
}
}
else if (negBalance >= 54) {
return 2;
}
}
}
else if (SC >= 7077) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (Cs < 1945) {
return 1;
}
else if (Cs >= 1945) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
if (posBalance < 52) {
if (refDCval < 200) {
if (Rs < 1306) {
return 1;
}
else if (Rs >= 1306) {
return 8;
}
}
else if (refDCval >= 200) {
if (Rs < 2415) {
return 8;
}
else if (Rs >= 2415) {
return 4;
}
}
}
else if (posBalance >= 52) {
if (Cs < 4127) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 4127) {
if (Rs < 8232) {
return 8;
}
else if (Rs >= 8232) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1891) {
if (ssDCval < 115) {
if (Cs < 1948) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Cs < 1847) {
if (Rs < 1607) {
if (Rs < 1312) {
return 8;
}
else if (Rs >= 1312) {
return 2;
}
}
else if (Rs >= 1607) {
return 8;
}
}
else if (Cs >= 1847) {
return 2;
}
}
}
else if (Cs >= 1948) {
if (Cs < 2539) {
return 2;
}
else if (Cs >= 2539) {
return 8;
}
}
}
else if (ssDCval >= 115) {
if (Rs < 1257) {
if (posBalance < 54) {
if (Rs < 928) {
return 1;
}
else if (Rs >= 928) {
return 8;
}
}
else if (posBalance >= 54) {
if (Rs < 1143) {
return 8;
}
else if (Rs >= 1143) {
return 4;
}
}
}
else if (Rs >= 1257) {
return 8;
}
}
}
else if (Rs >= 1891) {
if (ssDCval < 117) {
if (diffTSC < -2) {
if (Cs < 1949) {
if (Rs < 3151) {
if (MVDiff < 44) {
return 8;
}
else if (MVDiff >= 44) {
return 4;
}
}
else if (Rs >= 3151) {
return 4;
}
}
else if (Cs >= 1949) {
if (mcTcor < 98) {
if (Cs < 3251) {
if (Cs < 2232) {
return 8;
}
else if (Cs >= 2232) {
return 2;
}
}
else if (Cs >= 3251) {
return 8;
}
}
else if (mcTcor >= 98) {
if (SC < 4202) {
return 2;
}
else if (SC >= 4202) {
return 8;
}
}
}
}
else if (diffTSC >= -2) {
if (refDCval < 105) {
if (negBalance < 39) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 8;
}
}
else if (negBalance >= 39) {
if (Rs < 4917) {
if (Cs < 2306) {
if (Rs < 3739) {
return 8;
}
else if (Rs >= 3739) {
return 1;
}
}
else if (Cs >= 2306) {
return 8;
}
}
else if (Rs >= 4917) {
if (Rs < 5688) {
return 2;
}
else if (Rs >= 5688) {
return 8;
}
}
}
}
else if (refDCval >= 105) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (gchDC < 3) {
if (refDCval < 117) {
return 8;
}
else if (refDCval >= 117) {
return 4;
}
}
else if (gchDC >= 3) {
if (Rs < 2491) {
return 1;
}
else if (Rs >= 2491) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 117) {
if (diffTSC < -6) {
if (Cs < 5888) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (mcTcor < 97) {
if (Rs < 3796) {
return 8;
}
else if (Rs >= 3796) {
return 1;
}
}
else if (mcTcor >= 97) {
return 8;
}
}
}
else if (Cs >= 5888) {
return 1;
}
}
else if (diffTSC >= -6) {
if (posBalance < 35) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 146) {
if (SC < 5751) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (posBalance < 11) {
if (Rs < 2635) {
return 4;
}
else if (Rs >= 2635) {
return 8;
}
}
else if (posBalance >= 11) {
if (Cs < 2938) {
return 4;
}
else if (Cs >= 2938) {
if (Rs < 2757) {
return 8;
}
else if (Rs >= 2757) {
return 4;
}
}
}
}
}
else if (SC >= 5751) {
return 2;
}
}
else if (ssDCval >= 146) {
return 8;
}
}
else if (mu_mv_mag_sq >= 125) {
if (ssDCval < 201) {
if (Rs < 2358) {
if (Cs < 3474) {
if (Cs < 1811) {
return 1;
}
else if (Cs >= 1811) {
return 8;
}
}
else if (Cs >= 3474) {
return 1;
}
}
else if (Rs >= 2358) {
if (negBalance < 16) {
if (Cs < 1650) {
return 2;
}
else if (Cs >= 1650) {
return 8;
}
}
else if (negBalance >= 16) {
if (Rs < 3765) {
return 8;
}
else if (Rs >= 3765) {
return 1;
}
}
}
}
else if (ssDCval >= 201) {
if (Rs < 2433) {
return 8;
}
else if (Rs >= 2433) {
return 4;
}
}
}
}
else if (posBalance >= 35) {
if (Rs < 6449) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 5103) {
if (Rs < 3280) {
return 8;
}
else if (Rs >= 3280) {
return 1;
}
}
else if (Rs >= 5103) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (diffTSC < 2) {
if (Cs < 1658) {
if (Rs < 3850) {
return 8;
}
else if (Rs >= 3850) {
return 2;
}
}
else if (Cs >= 1658) {
return 8;
}
}
else if (diffTSC >= 2) {
if (Rs < 2803) {
return 8;
}
else if (Rs >= 2803) {
return 2;
}
}
}
}
else if (Rs >= 6449) {
if (SC < 10686) {
if (Rs < 6922) {
return 4;
}
else if (Rs >= 6922) {
if (Rs < 7377) {
return 8;
}
else if (Rs >= 7377) {
return 2;
}
}
}
else if (SC >= 10686) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect12(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1313) {
if (Rs < 2525) {
if (Cs < 534) {
if (negBalance < 14) {
if (mu_mv_mag_sq < 107) {
if (posBalance < 52) {
if (MVDiff < 9) {
if (SC < 923) {
if (TSCindex < 1) {
if (ssDCval < 26) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
return 4;
}
}
else if (ssDCval >= 26) {
if (negBalance < 3) {
if (Rs < 48) {
return 2;
}
else if (Rs >= 48) {
return 1;
}
}
else if (negBalance >= 3) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 1;
}
}
}
}
else if (TSCindex >= 1) {
if (AFD < 1) {
if (ssDCval < 37) {
if (Rs < 66) {
if (Rs < 22) {
return 1;
}
else if (Rs >= 22) {
return 4;
}
}
else if (Rs >= 66) {
return 8;
}
}
else if (ssDCval >= 37) {
if (posBalance < 31) {
return 1;
}
else if (posBalance >= 31) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
return 2;
}
}
}
}
else if (AFD >= 1) {
if (ssDCval < 27) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
return 8;
}
}
else if (ssDCval >= 27) {
if (Cs < 331) {
if (Rs < 89) {
if (Cs < 46) {
if (Rs < 40) {
return 4;
}
else if (Rs >= 40) {
if (Rs < 56) {
return 1;
}
else if (Rs >= 56) {
return 8;
}
}
}
else if (Cs >= 46) {
return 4;
}
}
else if (Rs >= 89) {
if (Cs < 120) {
if (ssDCval < 31) {
return 4;
}
else if (ssDCval >= 31) {
return 8;
}
}
else if (Cs >= 120) {
if (TSC < 2) {
if (Rs < 125) {
return 1;
}
else if (Rs >= 125) {
if (posBalance < 22) {
return 1;
}
else if (posBalance >= 22) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
if (AFD < 2) {
if (Rs < 297) {
return 2;
}
else if (Rs >= 297) {
return 4;
}
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
else if (TSC >= 2) {
if (Rs < 300) {
if (Rs < 240) {
return 1;
}
else if (Rs >= 240) {
return 4;
}
}
else if (Rs >= 300) {
return 8;
}
}
}
}
}
else if (Cs >= 331) {
if (SC < 567) {
return 4;
}
else if (SC >= 567) {
if (ssDCval < 161) {
if (ssDCval < 36) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
return 4;
}
}
else if (ssDCval >= 36) {
return 1;
}
}
else if (ssDCval >= 161) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 923) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (tcor < 101) {
return 8;
}
else if (tcor >= 101) {
return 1;
}
}
}
}
else if (MVDiff >= 9) {
if (refDCval < 22) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (Rs < 226) {
if (Rs < 6) {
return 8;
}
else if (Rs >= 6) {
return 1;
}
}
else if (Rs >= 226) {
return 8;
}
}
}
else if (refDCval >= 22) {
if (ssDCval < 112) {
if (tcor < 94) {
if (MVDiff < 29) {
if (Rs < 485) {
return 1;
}
else if (Rs >= 485) {
if (Rs < 928) {
return 4;
}
else if (Rs >= 928) {
return 8;
}
}
}
else if (MVDiff >= 29) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
return 4;
}
}
}
else if (tcor >= 94) {
if (Cs < 81) {
if (Cs < 4) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (Cs >= 4) {
return 1;
}
}
else if (Cs >= 81) {
if (mu_mv_mag_sq < 89) {
return 4;
}
else if (mu_mv_mag_sq >= 89) {
if (diffMVdiffVal < 4) {
if (diffMVdiffVal < -2) {
if (SC < 329) {
if (SC < 304) {
return 4;
}
else if (SC >= 304) {
return 1;
}
}
else if (SC >= 329) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
else if (diffMVdiffVal >= 4) {
if (Cs < 204) {
return 4;
}
else if (Cs >= 204) {
if (Rs < 117) {
return 1;
}
else if (Rs >= 117) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (Rs < 526) {
if (Rs < 408) {
return 2;
}
else if (Rs >= 408) {
if (ssDCval < 161) {
return 1;
}
else if (ssDCval >= 161) {
return 4;
}
}
}
else if (Rs >= 526) {
if (Rs < 1207) {
return 8;
}
else if (Rs >= 1207) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 52) {
if (negBalance < 8) {
if (RsCsDiff < 9) {
if (Rs < 57) {
if (refDCval < 49) {
if (mu_mv_mag_sq < 40) {
if (mcTcor < 69) {
return 8;
}
else if (mcTcor >= 69) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (ssDCval < 17) {
return 4;
}
else if (ssDCval >= 17) {
if (mu_mv_mag_sq < 102) {
if (AFD < 9) {
if (diffMVdiffVal < 16) {
return 8;
}
else if (diffMVdiffVal >= 16) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
return 4;
}
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (mu_mv_mag_sq >= 102) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
}
}
}
else if (refDCval >= 49) {
if (MVDiff < 75) {
return 2;
}
else if (MVDiff >= 75) {
return 1;
}
}
}
else if (Rs >= 57) {
if (refDCval < 33) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (MVDiff < 865) {
return 4;
}
else if (MVDiff >= 865) {
return 1;
}
}
}
else if (refDCval >= 33) {
if (MVDiff < 10) {
if (negBalance < 1) {
if (mu_mv_mag_sq < 88) {
return 8;
}
else if (mu_mv_mag_sq >= 88) {
if (Rs < 582) {
return 2;
}
else if (Rs >= 582) {
return 4;
}
}
}
else if (negBalance >= 1) {
if (ssDCval < 114) {
return 8;
}
else if (ssDCval >= 114) {
return 1;
}
}
}
else if (MVDiff >= 10) {
if (AFD < 5) {
if (tcor < 89) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (tcor >= 89) {
if (Cs < 112) {
if (Rs < 62) {
return 8;
}
else if (Rs >= 62) {
return 4;
}
}
else if (Cs >= 112) {
if (posBalance < 83) {
if (Rs < 927) {
return 4;
}
else if (Rs >= 927) {
return 1;
}
}
else if (posBalance >= 83) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
return 8;
}
}
}
}
}
else if (AFD >= 5) {
if (mu_mv_mag_sq < 78) {
if (AFD < 45) {
if (Cs < 4) {
if (Rs < 1336) {
return 4;
}
else if (Rs >= 1336) {
return 8;
}
}
else if (Cs >= 4) {
return 8;
}
}
else if (AFD >= 45) {
return 1;
}
}
else if (mu_mv_mag_sq >= 78) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 9) {
if (mcTcor < 92) {
if (MVDiff < 42) {
return 4;
}
else if (MVDiff >= 42) {
return 2;
}
}
else if (mcTcor >= 92) {
if (Rs < 60) {
return 2;
}
else if (Rs >= 60) {
return 1;
}
}
}
}
else if (negBalance >= 8) {
if (MVDiff < 10) {
if (posBalance < 78) {
if (Cs < 211) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (Rs < 1104) {
return 8;
}
else if (Rs >= 1104) {
return 1;
}
}
}
else if (Cs >= 211) {
if (AFD < 2) {
if (Rs < 266) {
if (Rs < 93) {
return 4;
}
else if (Rs >= 93) {
return 1;
}
}
else if (Rs >= 266) {
return 8;
}
}
else if (AFD >= 2) {
if (Rs < 276) {
return 8;
}
else if (Rs >= 276) {
return 1;
}
}
}
}
else if (posBalance >= 78) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
return 2;
}
}
}
else if (MVDiff >= 10) {
if (tcor < 94) {
if (diffMVdiffVal < -19) {
if (Cs < 153) {
return 2;
}
else if (Cs >= 153) {
return 4;
}
}
else if (diffMVdiffVal >= -19) {
if (RsCsDiff < 5) {
if (Rs < 2) {
return 8;
}
else if (Rs >= 2) {
return 1;
}
}
else if (RsCsDiff >= 5) {
if (mcTcor < 96) {
if (TSC < 7) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (TSC >= 7) {
if (TSC < 9) {
if (diffAFD < 7) {
return 2;
}
else if (diffAFD >= 7) {
if (Cs < 103) {
return 2;
}
else if (Cs >= 103) {
return 1;
}
}
}
else if (TSC >= 9) {
if (Rs < 261) {
return 8;
}
else if (Rs >= 261) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
}
}
else if (tcor >= 94) {
if (Rs < 116) {
if (Rs < 45) {
return 1;
}
else if (Rs >= 45) {
return 4;
}
}
else if (Rs >= 116) {
if (mu_mv_mag_sq < 88) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
if (Rs < 463) {
return 1;
}
else if (Rs >= 463) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 88) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 386) {
if (ssDCval < 47) {
if (Cs < 105) {
if (ssDCval < 38) {
if (refDCval < 22) {
if (ssDCval < 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 10) {
if (MVDiff < 1076) {
return 8;
}
else if (MVDiff >= 1076) {
return 4;
}
}
}
else if (refDCval >= 22) {
if (Rs < 15) {
if (Cs < 11) {
return 4;
}
else if (Cs >= 11) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (Rs >= 15) {
if (negBalance < 10) {
if (Rs < 29) {
if (Rs < 21) {
return 8;
}
else if (Rs >= 21) {
return 4;
}
}
else if (Rs >= 29) {
return 8;
}
}
else if (negBalance >= 10) {
if (posBalance < 11) {
return 8;
}
else if (posBalance >= 11) {
if (Rs < 101) {
return 4;
}
else if (Rs >= 101) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 38) {
if (Cs < 96) {
return 8;
}
else if (Cs >= 96) {
return 1;
}
}
}
else if (Cs >= 105) {
if (ssDCval < 44) {
if (refDCval < 36) {
if (SC < 558) {
if (posBalance < 27) {
if (Cs < 108) {
return 4;
}
else if (Cs >= 108) {
return 8;
}
}
else if (posBalance >= 27) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (SC >= 558) {
if (ssDCval < 23) {
return 8;
}
else if (ssDCval >= 23) {
return 4;
}
}
}
else if (refDCval >= 36) {
if (negBalance < 3) {
return 8;
}
else if (negBalance >= 3) {
if (CsDiff < 1) {
if (SC < 440) {
if (Cs < 176) {
return 1;
}
else if (Cs >= 176) {
if (negBalance < 12) {
if (Rs < 206) {
return 4;
}
else if (Rs >= 206) {
return 8;
}
}
else if (negBalance >= 12) {
return 1;
}
}
}
else if (SC >= 440) {
if (Rs < 344) {
return 1;
}
else if (Rs >= 344) {
return 4;
}
}
}
else if (CsDiff >= 1) {
if (ssDCval < 40) {
return 4;
}
else if (ssDCval >= 40) {
return 1;
}
}
}
}
}
else if (ssDCval >= 44) {
return 4;
}
}
}
else if (ssDCval >= 47) {
if (ssDCval < 82) {
if (Rs < 23) {
return 2;
}
else if (Rs >= 23) {
if (ssDCval < 64) {
if (Cs < 269) {
if (ssDCval < 58) {
if (Rs < 124) {
return 8;
}
else if (Rs >= 124) {
if (negBalance < 11) {
if (Rs < 233) {
return 8;
}
else if (Rs >= 233) {
return 4;
}
}
else if (negBalance >= 11) {
return 4;
}
}
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (Cs >= 269) {
if (Rs < 251) {
return 4;
}
else if (Rs >= 251) {
if (Rs < 325) {
return 1;
}
else if (Rs >= 325) {
return 8;
}
}
}
}
else if (ssDCval >= 64) {
if (Rs < 80) {
return 4;
}
else if (Rs >= 80) {
if (Cs < 238) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
return 8;
}
}
else if (Cs >= 238) {
if (SC < 644) {
if (Cs < 245) {
return 1;
}
else if (Cs >= 245) {
return 4;
}
}
else if (SC >= 644) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 122) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 306) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (Rs >= 306) {
if (ssDCval < 130) {
return 2;
}
else if (ssDCval >= 130) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 386) {
if (diffTSC < -3) {
if (Rs < 933) {
if (posBalance < 1) {
return 4;
}
else if (posBalance >= 1) {
if (Cs < 335) {
return 8;
}
else if (Cs >= 335) {
if (diffAFD < -10) {
return 8;
}
else if (diffAFD >= -10) {
if (MVDiff < 86) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 4;
}
}
else if (MVDiff >= 86) {
return 4;
}
}
}
}
}
else if (Rs >= 933) {
if (diffAFD < -8) {
return 2;
}
else if (diffAFD >= -8) {
if (Rs < 1532) {
return 1;
}
else if (Rs >= 1532) {
return 2;
}
}
}
}
else if (diffTSC >= -3) {
if (mu_mv_mag_sq < 128) {
if (SC < 962) {
if (ssDCval < 33) {
if (Cs < 258) {
if (diffMVdiffVal < -164) {
return 8;
}
else if (diffMVdiffVal >= -164) {
return 4;
}
}
else if (Cs >= 258) {
return 8;
}
}
else if (ssDCval >= 33) {
if (CsDiff < 1) {
if (posBalance < 24) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
return 1;
}
}
else if (posBalance >= 24) {
if (Cs < 405) {
if (Cs < 283) {
return 8;
}
else if (Cs >= 283) {
if (SC < 865) {
if (SC < 692) {
return 1;
}
else if (SC >= 692) {
return 8;
}
}
else if (SC >= 865) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 1;
}
}
}
}
else if (Cs >= 405) {
return 8;
}
}
}
else if (CsDiff >= 1) {
if (Rs < 517) {
return 1;
}
else if (Rs >= 517) {
return 8;
}
}
}
}
else if (SC >= 962) {
if (ssDCval < 67) {
if (refDCval < 60) {
if (Cs < 480) {
if (SC < 1170) {
return 8;
}
else if (SC >= 1170) {
if (SC < 1282) {
if (Rs < 727) {
return 2;
}
else if (Rs >= 727) {
return 1;
}
}
else if (SC >= 1282) {
if (posBalance < 35) {
if (Rs < 1153) {
return 8;
}
else if (Rs >= 1153) {
return 1;
}
}
else if (posBalance >= 35) {
return 8;
}
}
}
}
else if (Cs >= 480) {
if (Rs < 1388) {
if (mu_mv_mag_sq < 114) {
if (Rs < 636) {
return 1;
}
else if (Rs >= 636) {
return 2;
}
}
else if (mu_mv_mag_sq >= 114) {
return 1;
}
}
else if (Rs >= 1388) {
return 8;
}
}
}
else if (refDCval >= 60) {
if (Cs < 446) {
return 1;
}
else if (Cs >= 446) {
return 2;
}
}
}
else if (ssDCval >= 67) {
if (SC < 1041) {
if (Rs < 559) {
if (SC < 995) {
return 1;
}
else if (SC >= 995) {
return 8;
}
}
else if (Rs >= 559) {
return 4;
}
}
else if (SC >= 1041) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 1113) {
if (SC < 1260) {
return 8;
}
else if (SC >= 1260) {
if (Rs < 1006) {
return 8;
}
else if (Rs >= 1006) {
return 2;
}
}
}
else if (Rs >= 1113) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 137) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Cs < 275) {
if (ssDCval < 87) {
if (Cs < 250) {
return 2;
}
else if (Cs >= 250) {
return 1;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
else if (Cs >= 275) {
if (negBalance < 12) {
if (ssDCval < 66) {
if (Rs < 1109) {
if (Cs < 353) {
return 1;
}
else if (Cs >= 353) {
return 8;
}
}
else if (Rs >= 1109) {
return 2;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (negBalance >= 12) {
if (ssDCval < 98) {
if (Rs < 580) {
return 8;
}
else if (Rs >= 580) {
if (Rs < 730) {
return 4;
}
else if (Rs >= 730) {
if (Rs < 872) {
return 1;
}
else if (Rs >= 872) {
return 8;
}
}
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (mcTcor < 82) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (Cs < 361) {
if (Rs < 506) {
if (negBalance < 41) {
if (CsDiff < 2) {
if (posBalance < 66) {
if (AFD < 2) {
if (Rs < 151) {
if (Rs < 30) {
return 4;
}
else if (Rs >= 30) {
return 2;
}
}
else if (Rs >= 151) {
return 1;
}
}
else if (AFD >= 2) {
if (Cs < 221) {
return 1;
}
else if (Cs >= 221) {
if (AFD < 6) {
if (Cs < 257) {
return 8;
}
else if (Cs >= 257) {
return 2;
}
}
else if (AFD >= 6) {
return 1;
}
}
}
}
else if (posBalance >= 66) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (tcor < 38) {
if (mcTcor < 38) {
return 8;
}
else if (mcTcor >= 38) {
return 1;
}
}
else if (tcor >= 38) {
if (Cs < 206) {
return 8;
}
else if (Cs >= 206) {
if (Rs < 413) {
return 2;
}
else if (Rs >= 413) {
return 8;
}
}
}
}
}
else if (negBalance >= 41) {
if (tcor < 71) {
if (diffTSC < 1) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
return 8;
}
}
else if (diffTSC >= 1) {
if (Rs < 91) {
return 4;
}
else if (Rs >= 91) {
if (Rs < 329) {
if (ssDCval < 27) {
return 1;
}
else if (ssDCval >= 27) {
return 8;
}
}
else if (Rs >= 329) {
return 1;
}
}
}
}
else if (tcor >= 71) {
if (Rs < 22) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
return 4;
}
}
else if (Rs >= 22) {
return 2;
}
}
}
}
else if (Rs >= 506) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
else if (Cs >= 361) {
if (RsDiff < 4) {
if (ssDCval < 24) {
return 4;
}
else if (ssDCval >= 24) {
if (negBalance < 26) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (negBalance >= 26) {
return 2;
}
}
}
else if (RsDiff >= 4) {
if (Rs < 1164) {
return 8;
}
else if (Rs >= 1164) {
return 1;
}
}
}
}
}
else if (mcTcor >= 82) {
if (posBalance < 27) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 110) {
if (negBalance < 22) {
if (ssDCval < 32) {
return 4;
}
else if (ssDCval >= 32) {
if (Rs < 100) {
return 8;
}
else if (Rs >= 100) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (Rs < 13) {
if (Cs < 1) {
return 4;
}
else if (Cs >= 1) {
return 8;
}
}
else if (Rs >= 13) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 110) {
return 8;
}
}
else if (MVDiff >= 5) {
if (posBalance < 18) {
if (Rs < 37) {
if (TSCindex < 1) {
return 4;
}
else if (TSCindex >= 1) {
if (mu_mv_mag_sq < 33) {
if (mcTcor < 93) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (mcTcor >= 93) {
return 8;
}
}
else if (mu_mv_mag_sq >= 33) {
if (Cs < 1) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
return 4;
}
}
else if (Cs >= 1) {
return 8;
}
}
}
}
else if (Rs >= 37) {
if (mcTcor < 98) {
if (MVDiff < 438) {
return 1;
}
else if (MVDiff >= 438) {
return 4;
}
}
else if (mcTcor >= 98) {
return 4;
}
}
}
else if (posBalance >= 18) {
if (ssDCval < 31) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
return 1;
}
}
}
else if (ssDCval >= 31) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (refDCval < 33) {
return 2;
}
else if (refDCval >= 33) {
if (Rs < 302) {
return 4;
}
else if (Rs >= 302) {
return 8;
}
}
}
}
}
}
}
}
else if (posBalance >= 27) {
if (posBalance < 73) {
if (Rs < 69) {
if (posBalance < 69) {
return 1;
}
else if (posBalance >= 69) {
return 4;
}
}
else if (Rs >= 69) {
if (posBalance < 42) {
if (ssDCval < 30) {
if (mu_mv_mag_sq < 62) {
if (Cs < 324) {
if (mu_mv_mag_sq < 41) {
if (Rs < 123) {
if (Rs < 96) {
return 8;
}
else if (Rs >= 96) {
return 4;
}
}
else if (Rs >= 123) {
if (mu_mv_mag_sq < 37) {
if (mcTcor < 87) {
if (Cs < 185) {
return 4;
}
else if (Cs >= 185) {
return 1;
}
}
else if (mcTcor >= 87) {
return 1;
}
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 206) {
return 1;
}
else if (Rs >= 206) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (Cs < 299) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
if (diffRsCsdiff < 1) {
if (diffMVdiffVal < -8) {
return 4;
}
else if (diffMVdiffVal >= -8) {
return 1;
}
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
}
else if (Cs >= 299) {
if (Cs < 306) {
return 4;
}
else if (Cs >= 306) {
return 1;
}
}
}
}
else if (Cs >= 324) {
return 4;
}
}
else if (mu_mv_mag_sq >= 62) {
return 1;
}
}
else if (ssDCval >= 30) {
if (Cs < 257) {
if (Rs < 307) {
if (mu_mv_mag_sq < 19) {
if (Rs < 109) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (Rs >= 109) {
return 4;
}
}
else if (mu_mv_mag_sq >= 19) {
if (mcTcor < 97) {
if (AFD < 3) {
if (negBalance < 19) {
return 8;
}
else if (negBalance >= 19) {
if (Rs < 206) {
return 1;
}
else if (Rs >= 206) {
return 8;
}
}
}
else if (AFD >= 3) {
if (mu_mv_mag_sq < 57) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (mu_mv_mag_sq < 46) {
if (negBalance < 31) {
if (Rs < 241) {
return 4;
}
else if (Rs >= 241) {
return 1;
}
}
else if (negBalance >= 31) {
return 1;
}
}
else if (mu_mv_mag_sq >= 46) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (Rs < 182) {
return 8;
}
else if (Rs >= 182) {
if (Cs < 190) {
return 4;
}
else if (Cs >= 190) {
return 1;
}
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 104) {
if (Cs < 93) {
return 1;
}
else if (Cs >= 93) {
return 8;
}
}
else if (Rs >= 104) {
return 1;
}
}
}
}
else if (Rs >= 307) {
if (mcTcor < 93) {
if (MVDiff < 6) {
if (Rs < 483) {
return 8;
}
else if (Rs >= 483) {
return 1;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (mcTcor >= 93) {
if (Cs < 143) {
return 2;
}
else if (Cs >= 143) {
return 4;
}
}
}
}
else if (Cs >= 257) {
return 1;
}
}
}
else if (posBalance >= 42) {
if (tcor < 64) {
if (Rs < 231) {
if (Cs < 182) {
return 8;
}
else if (Cs >= 182) {
return 4;
}
}
else if (Rs >= 231) {
if (MVDiff < 66) {
return 1;
}
else if (MVDiff >= 66) {
return 8;
}
}
}
else if (tcor >= 64) {
if (negBalance < 26) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (mcTcor < 89) {
if (ssDCval < 28) {
if (Rs < 177) {
return 8;
}
else if (Rs >= 177) {
return 1;
}
}
else if (ssDCval >= 28) {
return 8;
}
}
else if (mcTcor >= 89) {
if (MVDiff < 24) {
if (diffMVdiffVal < -12) {
return 4;
}
else if (diffMVdiffVal >= -12) {
return 1;
}
}
else if (MVDiff >= 24) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
}
else if (negBalance >= 26) {
if (Rs < 237) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (RsCsDiff < 2) {
return 2;
}
else if (RsCsDiff >= 2) {
if (Rs < 234) {
return 1;
}
else if (Rs >= 234) {
return 8;
}
}
}
}
else if (Rs >= 237) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 73) {
return 2;
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 312) {
if (ssDCval < 50) {
if (mcTcor < 92) {
if (posBalance < 47) {
if (tcor < 65) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (diffTSC < 2) {
if (Rs < 223) {
return 8;
}
else if (Rs >= 223) {
return 1;
}
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (tcor >= 65) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 35) {
if (negBalance < 94) {
if (SC < 439) {
return 2;
}
else if (SC >= 439) {
return 8;
}
}
else if (negBalance >= 94) {
if (Rs < 25) {
return 8;
}
else if (Rs >= 25) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (mu_mv_mag_sq < 109) {
if (ssDCval < 45) {
if (mu_mv_mag_sq < 66) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (Cs < 129) {
return 1;
}
else if (Cs >= 129) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (ssDCval >= 45) {
return 8;
}
}
else if (mu_mv_mag_sq >= 109) {
return 2;
}
}
}
}
}
else if (posBalance >= 47) {
if (negBalance < 31) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (mcTcor < 77) {
if (Rs < 251) {
return 8;
}
else if (Rs >= 251) {
return 2;
}
}
else if (mcTcor >= 77) {
return 1;
}
}
}
else if (negBalance >= 31) {
if (AFD < 8) {
if (refDCval < 39) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (TSC < 4) {
if (Cs < 161) {
return 8;
}
else if (Cs >= 161) {
return 2;
}
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (refDCval >= 39) {
if (Cs < 234) {
if (mu_mv_mag_sq < 40) {
if (Cs < 131) {
return 1;
}
else if (Cs >= 131) {
return 8;
}
}
else if (mu_mv_mag_sq >= 40) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 1;
}
}
}
}
else if (Cs >= 234) {
if (Rs < 295) {
return 2;
}
else if (Rs >= 295) {
return 4;
}
}
}
}
else if (AFD >= 8) {
if (TSC < 4) {
if (tcor < 49) {
if (ssDCval < 37) {
if (Rs < 24) {
return 4;
}
else if (Rs >= 24) {
return 1;
}
}
else if (ssDCval >= 37) {
if (Rs < 9) {
return 8;
}
else if (Rs >= 9) {
return 2;
}
}
}
else if (tcor >= 49) {
return 4;
}
}
else if (TSC >= 4) {
if (MVDiff < 42) {
if (ssDCval < 45) {
if (negBalance < 45) {
if (Rs < 173) {
return 8;
}
else if (Rs >= 173) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (Rs < 218) {
return 1;
}
else if (Rs >= 218) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (ssDCval >= 45) {
if (Cs < 178) {
return 2;
}
else if (Cs >= 178) {
return 4;
}
}
}
else if (MVDiff >= 42) {
if (mcTcor < 79) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 45) {
return 4;
}
else if (negBalance >= 45) {
if (mcTcor < 60) {
if (Rs < 128) {
return 8;
}
else if (Rs >= 128) {
return 1;
}
}
else if (mcTcor >= 60) {
if (tcor < 68) {
if (Rs < 301) {
if (SC < 168) {
if (Rs < 66) {
return 8;
}
else if (Rs >= 66) {
return 4;
}
}
else if (SC >= 168) {
return 8;
}
}
else if (Rs >= 301) {
return 4;
}
}
else if (tcor >= 68) {
return 4;
}
}
}
}
}
else if (mcTcor >= 79) {
if (mu_mv_mag_sq < 82) {
if (diffRsCsdiff < -8) {
if (Rs < 143) {
return 4;
}
else if (Rs >= 143) {
return 8;
}
}
else if (diffRsCsdiff >= -8) {
if (ssDCval < 41) {
if (Cs < 97) {
return 1;
}
else if (Cs >= 97) {
return 8;
}
}
else if (ssDCval >= 41) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (Rs < 36) {
return 2;
}
else if (Rs >= 36) {
return 4;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 50) {
if (Cs < 3) {
if (tcor < 79) {
if (Rs < 136) {
return 2;
}
else if (Rs >= 136) {
return 1;
}
}
else if (tcor >= 79) {
return 8;
}
}
else if (Cs >= 3) {
if (posBalance < 69) {
if (mcTcor < 98) {
if (posBalance < 43) {
if (AFD < 9) {
if (mu_mv_mag_sq < 6) {
if (Rs < 126) {
return 1;
}
else if (Rs >= 126) {
return 2;
}
}
else if (mu_mv_mag_sq >= 6) {
if (AFD < 5) {
if (Rs < 118) {
return 1;
}
else if (Rs >= 118) {
if (Rs < 198) {
return 8;
}
else if (Rs >= 198) {
return 4;
}
}
}
else if (AFD >= 5) {
if (TSC < 5) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 2;
}
}
else if (TSC >= 5) {
return 2;
}
}
}
}
else if (AFD >= 9) {
if (Rs < 250) {
return 1;
}
else if (Rs >= 250) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (Cs < 73) {
if (MVDiff < 50) {
return 1;
}
else if (MVDiff >= 50) {
if (MVDiff < 65) {
return 2;
}
else if (MVDiff >= 65) {
return 1;
}
}
}
else if (Cs >= 73) {
if (Cs < 337) {
if (Rs < 95) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (negBalance < 47) {
return 4;
}
else if (negBalance >= 47) {
return 1;
}
}
}
else if (Rs >= 95) {
if (AFD < 7) {
if (ssDCval < 43) {
if (Cs < 133) {
return 4;
}
else if (Cs >= 133) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (ssDCval >= 43) {
if (Cs < 163) {
return 1;
}
else if (Cs >= 163) {
if (Cs < 274) {
return 4;
}
else if (Cs >= 274) {
return 2;
}
}
}
}
else if (AFD >= 7) {
if (ssDCval < 44) {
if (SC < 340) {
if (Rs < 171) {
return 1;
}
else if (Rs >= 171) {
return 2;
}
}
else if (SC >= 340) {
return 1;
}
}
else if (ssDCval >= 44) {
if (refDCval < 44) {
return 1;
}
else if (refDCval >= 44) {
if (Rs < 185) {
return 2;
}
else if (Rs >= 185) {
return 4;
}
}
}
}
}
}
else if (Cs >= 337) {
if (Cs < 487) {
return 1;
}
else if (Cs >= 487) {
return 4;
}
}
}
}
}
else if (mcTcor >= 98) {
if (diffMVdiffVal < 18) {
if (negBalance < 69) {
if (Rs < 306) {
if (Cs < 7) {
return 2;
}
else if (Cs >= 7) {
if (AFD < 15) {
if (tcor < 90) {
if (SC < 222) {
if (Rs < 43) {
return 1;
}
else if (Rs >= 43) {
return 4;
}
}
else if (SC >= 222) {
return 1;
}
}
else if (tcor >= 90) {
return 1;
}
}
else if (AFD >= 15) {
return 4;
}
}
}
else if (Rs >= 306) {
return 4;
}
}
else if (negBalance >= 69) {
if (Rs < 142) {
return 4;
}
else if (Rs >= 142) {
return 8;
}
}
}
else if (diffMVdiffVal >= 18) {
if (Rs < 186) {
return 2;
}
else if (Rs >= 186) {
if (Rs < 255) {
return 8;
}
else if (Rs >= 255) {
return 4;
}
}
}
}
}
else if (posBalance >= 69) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (Cs < 297) {
if (mu_mv_mag_sq < 102) {
if (negBalance < 33) {
if (ssDCval < 39) {
if (posBalance < 31) {
if (Rs < 122) {
return 1;
}
else if (Rs >= 122) {
if (Cs < 277) {
return 4;
}
else if (Cs >= 277) {
return 8;
}
}
}
else if (posBalance >= 31) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 96) {
return 8;
}
else if (mu_mv_mag_sq >= 96) {
return 1;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
}
else if (ssDCval >= 39) {
if (ssDCval < 47) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 87) {
return 1;
}
else if (mu_mv_mag_sq >= 87) {
return 8;
}
}
}
else if (ssDCval >= 47) {
if (Rs < 172) {
if (Rs < 13) {
return 2;
}
else if (Rs >= 13) {
return 8;
}
}
else if (Rs >= 172) {
return 4;
}
}
}
}
else if (negBalance >= 33) {
if (AFD < 6) {
if (TSC < 2) {
if (ssDCval < 42) {
return 1;
}
else if (ssDCval >= 42) {
return 8;
}
}
else if (TSC >= 2) {
if (mcTcor < 105) {
if (ssDCval < 42) {
return 8;
}
else if (ssDCval >= 42) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (mcTcor >= 105) {
return 4;
}
}
}
else if (AFD >= 6) {
if (Rs < 132) {
return 4;
}
else if (Rs >= 132) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (negBalance < 55) {
if (posBalance < 37) {
if (tcor < 101) {
if (Cs < 196) {
if (Cs < 89) {
if (ssDCval < 42) {
return 4;
}
else if (ssDCval >= 42) {
return 8;
}
}
else if (Cs >= 89) {
if (posBalance < 24) {
return 4;
}
else if (posBalance >= 24) {
return 1;
}
}
}
else if (Cs >= 196) {
return 8;
}
}
else if (tcor >= 101) {
if (Rs < 92) {
return 8;
}
else if (Rs >= 92) {
if (Cs < 231) {
return 4;
}
else if (Cs >= 231) {
return 8;
}
}
}
}
else if (posBalance >= 37) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
return 1;
}
}
}
else if (negBalance >= 55) {
if (Rs < 8) {
return 2;
}
else if (Rs >= 8) {
return 4;
}
}
}
}
else if (Cs >= 297) {
if (mu_mv_mag_sq < 58) {
if (AFD < 6) {
if (Rs < 281) {
return 2;
}
else if (Rs >= 281) {
return 1;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 234) {
if (RsCsDiff < 1) {
if (Rs < 205) {
if (Cs < 300) {
return 8;
}
else if (Cs >= 300) {
return 1;
}
}
else if (Rs >= 205) {
return 8;
}
}
else if (RsCsDiff >= 1) {
return 1;
}
}
else if (Rs >= 234) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 50) {
if (refDCval < 70) {
if (TSCindex < 2) {
if (Rs < 110) {
if (posBalance < 42) {
if (ssDCval < 55) {
if (Rs < 29) {
if (Rs < 14) {
return 2;
}
else if (Rs >= 14) {
return 1;
}
}
else if (Rs >= 29) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
return 4;
}
}
}
else if (ssDCval >= 55) {
return 2;
}
}
else if (posBalance >= 42) {
if (MVDiff < 4) {
if (Rs < 90) {
return 2;
}
else if (Rs >= 90) {
return 4;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
}
else if (Rs >= 110) {
if (ssDCval < 66) {
if (posBalance < 28) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (posBalance >= 28) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
if (SC < 352) {
if (ssDCval < 58) {
if (MVDiff < 3) {
if (Rs < 157) {
if (Rs < 145) {
return 4;
}
else if (Rs >= 145) {
return 1;
}
}
else if (Rs >= 157) {
return 4;
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (ssDCval >= 58) {
if (Cs < 141) {
return 8;
}
else if (Cs >= 141) {
return 1;
}
}
}
else if (SC >= 352) {
if (Rs < 246) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 175) {
return 4;
}
else if (Rs >= 175) {
return 1;
}
}
}
else if (Rs >= 246) {
if (SC < 569) {
return 4;
}
else if (SC >= 569) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (Cs < 360) {
return 1;
}
else if (Cs >= 360) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 66) {
if (mu_mv_mag_sq < 119) {
return 2;
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 184) {
return 4;
}
else if (Rs >= 184) {
return 1;
}
}
}
}
}
else if (TSCindex >= 2) {
if (RsCsDiff < 1) {
if (SC < 347) {
if (tcor < 100) {
if (MVDiff < 3) {
if (ssDCval < 64) {
if (tcor < 95) {
if (Rs < 83) {
return 2;
}
else if (Rs >= 83) {
if (Cs < 117) {
return 1;
}
else if (Cs >= 117) {
return 8;
}
}
}
else if (tcor >= 95) {
if (negBalance < 22) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
return 1;
}
}
else if (negBalance >= 22) {
return 1;
}
}
}
else if (ssDCval >= 64) {
if (Cs < 138) {
return 1;
}
else if (Cs >= 138) {
if (posBalance < 27) {
return 1;
}
else if (posBalance >= 27) {
return 2;
}
}
}
}
else if (MVDiff >= 3) {
if (mcTcor < 97) {
if (MVDiff < 17) {
if (Rs < 127) {
if (Rs < 51) {
if (Rs < 34) {
return 8;
}
else if (Rs >= 34) {
return 1;
}
}
else if (Rs >= 51) {
if (Rs < 112) {
return 2;
}
else if (Rs >= 112) {
return 4;
}
}
}
else if (Rs >= 127) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (MVDiff >= 17) {
if (tcor < 92) {
if (posBalance < 56) {
if (MVDiff < 71) {
if (Rs < 65) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
if (MVDiff < 62) {
return 4;
}
else if (MVDiff >= 62) {
return 1;
}
}
}
else if (Rs >= 65) {
return 4;
}
}
else if (MVDiff >= 71) {
return 4;
}
}
else if (posBalance >= 56) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
}
else if (tcor >= 92) {
if (Rs < 119) {
return 8;
}
else if (Rs >= 119) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 194) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
if (MVDiff < 13) {
if (negBalance < 28) {
if (negBalance < 27) {
if (ssDCval < 57) {
if (Rs < 132) {
return 1;
}
else if (Rs >= 132) {
return 4;
}
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (negBalance >= 27) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 1;
}
}
}
else if (negBalance >= 28) {
if (tcor < 94) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Rs < 110) {
return 1;
}
else if (Rs >= 110) {
return 4;
}
}
}
else if (tcor >= 94) {
return 1;
}
}
}
else if (MVDiff >= 13) {
if (TSCindex < 3) {
if (Rs < 77) {
return 8;
}
else if (Rs >= 77) {
return 1;
}
}
else if (TSCindex >= 3) {
return 4;
}
}
}
}
else if (Cs >= 194) {
if (ssDCval < 62) {
if (SC < 307) {
return 2;
}
else if (SC >= 307) {
return 4;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
}
}
}
else if (tcor >= 100) {
if (Rs < 63) {
return 2;
}
else if (Rs >= 63) {
return 8;
}
}
}
else if (SC >= 347) {
if (negBalance < 31) {
if (refDCval < 60) {
if (MVDiff < 1) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 240) {
return 1;
}
else if (Rs >= 240) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 66) {
if (Rs < 156) {
if (Rs < 91) {
return 1;
}
else if (Rs >= 91) {
return 4;
}
}
else if (Rs >= 156) {
return 1;
}
}
else if (mu_mv_mag_sq >= 66) {
return 4;
}
}
}
else if (refDCval >= 60) {
if (negBalance < 25) {
if (Cs < 343) {
if (SC < 363) {
return 2;
}
else if (SC >= 363) {
if (MVDiff < 3) {
if (Rs < 147) {
if (mu_mv_mag_sq < 121) {
if (Cs < 256) {
return 4;
}
else if (Cs >= 256) {
return 1;
}
}
else if (mu_mv_mag_sq >= 121) {
return 4;
}
}
else if (Rs >= 147) {
return 4;
}
}
else if (MVDiff >= 3) {
if (Rs < 170) {
return 2;
}
else if (Rs >= 170) {
if (Rs < 200) {
return 1;
}
else if (Rs >= 200) {
return 4;
}
}
}
}
}
else if (Cs >= 343) {
if (ssDCval < 69) {
if (SC < 541) {
return 1;
}
else if (SC >= 541) {
return 4;
}
}
else if (ssDCval >= 69) {
if (Rs < 158) {
return 2;
}
else if (Rs >= 158) {
return 1;
}
}
}
}
else if (negBalance >= 25) {
if (mu_mv_mag_sq < 125) {
if (Rs < 278) {
if (posBalance < 44) {
if (Cs < 175) {
return 2;
}
else if (Cs >= 175) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
return 4;
}
}
}
else if (posBalance >= 44) {
if (ssDCval < 63) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 63) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 97) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (Cs < 278) {
return 4;
}
else if (Cs >= 278) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 380) {
if (mu_mv_mag_sq < 110) {
if (posBalance < 48) {
if (Rs < 132) {
return 1;
}
else if (Rs >= 132) {
if (diffMVdiffVal < -1) {
if (Rs < 168) {
return 1;
}
else if (Rs >= 168) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
else if (posBalance >= 48) {
if (tcor < 98) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 4;
}
}
else if (tcor >= 98) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 110) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
}
else if (Cs >= 380) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (tcor < 97) {
if (Cs < 224) {
return 1;
}
else if (Cs >= 224) {
return 4;
}
}
else if (tcor >= 97) {
return 1;
}
}
}
}
}
else if (Rs >= 278) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 148) {
return 4;
}
else if (Rs >= 148) {
if (Rs < 277) {
return 1;
}
else if (Rs >= 277) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 31) {
if (ssDCval < 63) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (Rs < 289) {
if (Cs < 215) {
if (tcor < 97) {
if (mu_mv_mag_sq < 12) {
return 1;
}
else if (mu_mv_mag_sq >= 12) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
return 4;
}
}
}
else if (tcor >= 97) {
return 1;
}
}
else if (Cs >= 215) {
return 4;
}
}
else if (Rs >= 289) {
if (Rs < 305) {
return 1;
}
else if (Rs >= 305) {
return 2;
}
}
}
}
else if (ssDCval >= 63) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (diffRsCsdiff < -1) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
if (Rs < 207) {
return 4;
}
else if (Rs >= 207) {
return 1;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (TSC < 4) {
if (mcTcor < 91) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (diffTSC < 1) {
if (Rs < 305) {
return 8;
}
else if (Rs >= 305) {
return 2;
}
}
else if (diffTSC >= 1) {
if (Rs < 166) {
return 4;
}
else if (Rs >= 166) {
return 8;
}
}
}
}
else if (mcTcor >= 91) {
if (Rs < 274) {
if (MVDiff < 3) {
if (SC < 655) {
if (Cs < 218) {
return 1;
}
else if (Cs >= 218) {
if (ssDCval < 69) {
if (refDCval < 52) {
return 1;
}
else if (refDCval >= 52) {
if (negBalance < 33) {
if (Rs < 167) {
return 1;
}
else if (Rs >= 167) {
return 4;
}
}
else if (negBalance >= 33) {
if (mu_mv_mag_sq < 32) {
return 4;
}
else if (mu_mv_mag_sq >= 32) {
if (Cs < 354) {
return 1;
}
else if (Cs >= 354) {
if (Rs < 217) {
return 4;
}
else if (Rs >= 217) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (posBalance < 49) {
if (Cs < 330) {
return 4;
}
else if (Cs >= 330) {
return 1;
}
}
else if (posBalance >= 49) {
if (Rs < 154) {
return 1;
}
else if (Rs >= 154) {
return 4;
}
}
}
}
}
else if (SC >= 655) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 37) {
if (TSCindex < 3) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (refDCval < 52) {
return 1;
}
else if (refDCval >= 52) {
if (MVDiff < 27) {
if (RsCsDiff < 7) {
if (CsDiff < 1) {
if (TSC < 3) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (TSC >= 3) {
return 4;
}
}
else if (CsDiff >= 1) {
if (refDCval < 69) {
if (AFD < 7) {
if (Cs < 283) {
if (Rs < 231) {
if (SC < 365) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (Rs < 23) {
return 1;
}
else if (Rs >= 23) {
return 4;
}
}
}
else if (SC >= 365) {
return 4;
}
}
else if (Rs >= 231) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
}
else if (Cs >= 283) {
return 4;
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (refDCval >= 69) {
if (Rs < 138) {
return 1;
}
else if (Rs >= 138) {
return 4;
}
}
}
}
else if (RsCsDiff >= 7) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 226) {
return 4;
}
else if (Rs >= 226) {
return 2;
}
}
}
}
else if (MVDiff >= 27) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (CsDiff < 4) {
if (SC < 171) {
if (Rs < 26) {
return 1;
}
else if (Rs >= 26) {
return 4;
}
}
else if (SC >= 171) {
return 1;
}
}
else if (CsDiff >= 4) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (Cs < 170) {
if (Cs < 133) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
return 4;
}
}
else if (Cs >= 133) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
if (Rs < 130) {
return 4;
}
else if (Rs >= 130) {
return 8;
}
}
}
}
else if (Cs >= 170) {
if (negBalance < 39) {
if (ssDCval < 69) {
if (posBalance < 27) {
return 1;
}
else if (posBalance >= 27) {
if (ssDCval < 54) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 2;
}
}
}
else if (ssDCval >= 54) {
return 4;
}
}
}
else if (ssDCval >= 69) {
if (Rs < 161) {
return 1;
}
else if (Rs >= 161) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 74) {
if (Rs < 192) {
return 2;
}
else if (Rs >= 192) {
return 4;
}
}
else if (mu_mv_mag_sq >= 74) {
return 1;
}
}
else if (mcTcor >= 98) {
if (posBalance < 50) {
return 4;
}
else if (posBalance >= 50) {
if (Cs < 503) {
return 8;
}
else if (Cs >= 503) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 274) {
if (gchDC < 1) {
if (tcor < 90) {
if (SC < 665) {
return 1;
}
else if (SC >= 665) {
return 2;
}
}
else if (tcor >= 90) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (AFD < 6) {
if (posBalance < 58) {
if (negBalance < 41) {
return 4;
}
else if (negBalance >= 41) {
if (Rs < 292) {
return 1;
}
else if (Rs >= 292) {
return 4;
}
}
}
else if (posBalance >= 58) {
return 1;
}
}
else if (AFD >= 6) {
return 1;
}
}
}
}
else if (gchDC >= 1) {
if (diffMVdiffVal < -6) {
if (Cs < 317) {
return 4;
}
else if (Cs >= 317) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
if (TSC < 3) {
if (negBalance < 35) {
if (Cs < 209) {
return 4;
}
else if (Cs >= 209) {
return 2;
}
}
else if (negBalance >= 35) {
return 4;
}
}
else if (TSC >= 3) {
if (diffAFD < -1) {
if (Rs < 292) {
return 2;
}
else if (Rs >= 292) {
return 1;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 26) {
if (tcor < 75) {
if (SC < 322) {
if (MVDiff < 106) {
return 2;
}
else if (MVDiff >= 106) {
return 1;
}
}
else if (SC >= 322) {
if (diffTSC < -1) {
if (Rs < 275) {
if (Cs < 395) {
return 8;
}
else if (Cs >= 395) {
return 2;
}
}
else if (Rs >= 275) {
return 1;
}
}
else if (diffTSC >= -1) {
if (refDCval < 63) {
if (Cs < 192) {
if (Cs < 103) {
return 1;
}
else if (Cs >= 103) {
return 4;
}
}
else if (Cs >= 192) {
if (mu_mv_mag_sq < 17) {
return 1;
}
else if (mu_mv_mag_sq >= 17) {
if (Rs < 160) {
return 1;
}
else if (Rs >= 160) {
if (Rs < 226) {
return 2;
}
else if (Rs >= 226) {
return 4;
}
}
}
}
}
else if (refDCval >= 63) {
if (Cs < 232) {
return 2;
}
else if (Cs >= 232) {
return 4;
}
}
}
}
}
else if (tcor >= 75) {
if (MVDiff < 60) {
if (AFD < 15) {
if (mcTcor < 96) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (AFD < 7) {
if (Cs < 332) {
return 1;
}
else if (Cs >= 332) {
return 2;
}
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (mcTcor >= 96) {
if (Cs < 332) {
return 1;
}
else if (Cs >= 332) {
if (Rs < 254) {
return 8;
}
else if (Rs >= 254) {
return 1;
}
}
}
}
else if (AFD >= 15) {
return 2;
}
}
else if (MVDiff >= 60) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Cs < 35) {
return 1;
}
else if (Cs >= 35) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (Rs < 141) {
return 1;
}
else if (Rs >= 141) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (AFD < 8) {
if (MVDiff < 6) {
if (Rs < 279) {
if (Cs < 394) {
return 4;
}
else if (Cs >= 394) {
return 1;
}
}
else if (Rs >= 279) {
return 8;
}
}
else if (MVDiff >= 6) {
if (MVDiff < 28) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
if (posBalance < 46) {
if (Rs < 223) {
return 2;
}
else if (Rs >= 223) {
return 1;
}
}
else if (posBalance >= 46) {
return 8;
}
}
}
else if (MVDiff >= 28) {
return 8;
}
}
}
else if (AFD >= 8) {
if (negBalance < 58) {
if (diffMVdiffVal < 32) {
if (SC < 201) {
if (Rs < 68) {
return 8;
}
else if (Rs >= 68) {
return 1;
}
}
else if (SC >= 201) {
if (Cs < 320) {
if (refDCval < 49) {
return 8;
}
else if (refDCval >= 49) {
return 4;
}
}
else if (Cs >= 320) {
if (Cs < 485) {
return 1;
}
else if (Cs >= 485) {
if (Rs < 218) {
return 4;
}
else if (Rs >= 218) {
return 8;
}
}
}
}
}
else if (diffMVdiffVal >= 32) {
if (MVDiff < 662) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (MVDiff >= 662) {
return 4;
}
}
}
else if (negBalance >= 58) {
if (tcor < 16) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (tcor >= 16) {
if (Rs < 238) {
if (Rs < 42) {
return 1;
}
else if (Rs >= 42) {
return 8;
}
}
else if (Rs >= 238) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 107) {
if (TSCindex < 2) {
if (ssDCval < 81) {
if (Cs < 263) {
if (SC < 257) {
if (Rs < 129) {
if (Rs < 76) {
return 1;
}
else if (Rs >= 76) {
return 2;
}
}
else if (Rs >= 129) {
return 1;
}
}
else if (SC >= 257) {
return 2;
}
}
else if (Cs >= 263) {
return 1;
}
}
else if (ssDCval >= 81) {
if (Cs < 159) {
if (Rs < 94) {
if (Rs < 5) {
return 1;
}
else if (Rs >= 5) {
return 4;
}
}
else if (Rs >= 94) {
if (Rs < 202) {
if (Rs < 168) {
if (Rs < 142) {
return 1;
}
else if (Rs >= 142) {
return 8;
}
}
else if (Rs >= 168) {
return 2;
}
}
else if (Rs >= 202) {
return 1;
}
}
}
else if (Cs >= 159) {
if (MVDiff < 41) {
return 4;
}
else if (MVDiff >= 41) {
return 2;
}
}
}
}
else if (TSCindex >= 2) {
if (mcTcor < 97) {
if (mcTcor < 72) {
if (diffAFD < 2) {
if (ssDCval < 77) {
if (Rs < 128) {
return 8;
}
else if (Rs >= 128) {
if (Rs < 218) {
return 1;
}
else if (Rs >= 218) {
return 4;
}
}
}
else if (ssDCval >= 77) {
if (ssDCval < 168) {
return 4;
}
else if (ssDCval >= 168) {
return 2;
}
}
}
else if (diffAFD >= 2) {
if (AFD < 36) {
if (mcTcor < 58) {
return 1;
}
else if (mcTcor >= 58) {
if (ssDCval < 140) {
if (Rs < 75) {
return 1;
}
else if (Rs >= 75) {
return 4;
}
}
else if (ssDCval >= 140) {
return 2;
}
}
}
else if (AFD >= 36) {
if (Rs < 269) {
return 8;
}
else if (Rs >= 269) {
return 2;
}
}
}
}
else if (mcTcor >= 72) {
if (MVDiff < 23) {
if (ssDCval < 86) {
if (MVDiff < 5) {
if (Rs < 107) {
return 4;
}
else if (Rs >= 107) {
return 8;
}
}
else if (MVDiff >= 5) {
if (Cs < 344) {
if (Rs < 130) {
if (ssDCval < 78) {
if (Rs < 57) {
return 1;
}
else if (Rs >= 57) {
return 4;
}
}
else if (ssDCval >= 78) {
return 1;
}
}
else if (Rs >= 130) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (Cs >= 344) {
if (Rs < 244) {
return 4;
}
else if (Rs >= 244) {
return 8;
}
}
}
}
else if (ssDCval >= 86) {
if (SC < 134) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
else if (SC >= 134) {
if (mcTcor < 94) {
if (negBalance < 56) {
if (Cs < 361) {
if (TSC < 5) {
if (tcor < 78) {
if (diffRsCsdiff < -1) {
if (Rs < 149) {
return 2;
}
else if (Rs >= 149) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
if (diffMVdiffVal < -12) {
return 2;
}
else if (diffMVdiffVal >= -12) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
if (Cs < 279) {
return 2;
}
else if (Cs >= 279) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 78) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
if (Rs < 191) {
return 2;
}
else if (Rs >= 191) {
return 8;
}
}
}
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (Cs >= 361) {
return 8;
}
}
else if (negBalance >= 56) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
return 1;
}
}
}
else if (mcTcor >= 94) {
if (negBalance < 36) {
if (Cs < 205) {
if (SC < 449) {
return 1;
}
else if (SC >= 449) {
return 4;
}
}
else if (Cs >= 205) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (Rs < 310) {
return 2;
}
else if (Rs >= 310) {
return 8;
}
}
}
}
else if (negBalance >= 36) {
if (AFD < 11) {
if (refDCval < 93) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
return 8;
}
}
}
else if (refDCval >= 93) {
return 8;
}
}
else if (AFD >= 11) {
if (posBalance < 72) {
return 1;
}
else if (posBalance >= 72) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 23) {
if (ssDCval < 134) {
if (AFD < 10) {
if (diffTSC < -1) {
if (negBalance < 40) {
if (mcTcor < 93) {
return 1;
}
else if (mcTcor >= 93) {
if (Rs < 121) {
return 2;
}
else if (Rs >= 121) {
return 8;
}
}
}
else if (negBalance >= 40) {
if (diffMVdiffVal < 13) {
return 8;
}
else if (diffMVdiffVal >= 13) {
if (Rs < 221) {
return 2;
}
else if (Rs >= 221) {
return 8;
}
}
}
}
else if (diffTSC >= -1) {
if (Cs < 155) {
if (mu_mv_mag_sq < 26) {
if (ssDCval < 117) {
if (tcor < 58) {
if (Rs < 96) {
return 1;
}
else if (Rs >= 96) {
return 2;
}
}
else if (tcor >= 58) {
return 1;
}
}
else if (ssDCval >= 117) {
if (Rs < 167) {
return 2;
}
else if (Rs >= 167) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (ssDCval < 91) {
if (refDCval < 78) {
return 1;
}
else if (refDCval >= 78) {
return 8;
}
}
else if (ssDCval >= 91) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
if (Rs < 103) {
return 2;
}
else if (Rs >= 103) {
return 1;
}
}
}
}
}
else if (Cs >= 155) {
if (ssDCval < 82) {
return 1;
}
else if (ssDCval >= 82) {
return 2;
}
}
}
}
else if (AFD >= 10) {
if (tcor < 90) {
if (MVDiff < 335) {
if (Cs < 25) {
if (RsDiff < 2) {
if (ssDCval < 104) {
return 2;
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (RsDiff >= 2) {
if (diffMVdiffVal < -4) {
if (Rs < 247) {
return 8;
}
else if (Rs >= 247) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (Cs >= 25) {
if (mu_mv_mag_sq < 17) {
if (MVDiff < 53) {
if (ssDCval < 116) {
if (AFD < 20) {
if (Rs < 280) {
return 1;
}
else if (Rs >= 280) {
return 2;
}
}
else if (AFD >= 20) {
return 8;
}
}
else if (ssDCval >= 116) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
}
else if (MVDiff >= 53) {
if (Cs < 218) {
if (RsCsDiff < 4) {
return 2;
}
else if (RsCsDiff >= 4) {
if (refDCval < 85) {
if (ssDCval < 80) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 1;
}
}
else if (ssDCval >= 80) {
return 2;
}
}
else if (refDCval >= 85) {
return 1;
}
}
}
else if (Cs >= 218) {
if (Cs < 254) {
if (Rs < 190) {
return 4;
}
else if (Rs >= 190) {
return 2;
}
}
else if (Cs >= 254) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (refDCval < 81) {
if (diffMVdiffVal < -2) {
if (Cs < 86) {
return 8;
}
else if (Cs >= 86) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
if (mu_mv_mag_sq < 25) {
return 2;
}
else if (mu_mv_mag_sq >= 25) {
if (Rs < 106) {
return 1;
}
else if (Rs >= 106) {
return 4;
}
}
}
}
else if (refDCval >= 81) {
if (RsDiff < 3) {
if (ssDCval < 117) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
if (TSC < 13) {
return 1;
}
else if (TSC >= 13) {
return 2;
}
}
}
else if (ssDCval >= 117) {
if (Rs < 240) {
return 8;
}
else if (Rs >= 240) {
return 2;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 229) {
return 1;
}
else if (Rs >= 229) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 335) {
if (Rs < 272) {
return 1;
}
else if (Rs >= 272) {
return 4;
}
}
}
else if (tcor >= 90) {
if (Cs < 175) {
return 8;
}
else if (Cs >= 175) {
if (Rs < 138) {
return 2;
}
else if (Rs >= 138) {
if (Rs < 204) {
return 8;
}
else if (Rs >= 204) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 134) {
if (ssDCval < 161) {
if (Rs < 1) {
return 1;
}
else if (Rs >= 1) {
return 2;
}
}
else if (ssDCval >= 161) {
if (MVDiff < 69) {
if (Rs < 60) {
return 8;
}
else if (Rs >= 60) {
return 2;
}
}
else if (MVDiff >= 69) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 106) {
if (Cs < 251) {
if (Rs < 111) {
if (Cs < 125) {
if (TSC < 3) {
if (mu_mv_mag_sq < 2) {
if (CsDiff < 1) {
if (Cs < 114) {
return 1;
}
else if (Cs >= 114) {
if (Rs < 56) {
return 4;
}
else if (Rs >= 56) {
return 1;
}
}
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 108) {
return 1;
}
else if (Rs >= 108) {
return 2;
}
}
}
else if (TSC >= 3) {
if (refDCval < 96) {
if (Rs < 59) {
if (Cs < 87) {
return 4;
}
else if (Cs >= 87) {
return 8;
}
}
else if (Rs >= 59) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 1;
}
}
}
else if (refDCval >= 96) {
return 2;
}
}
}
else if (Cs >= 125) {
if (ssDCval < 97) {
if (mu_mv_mag_sq < 7) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (Rs < 101) {
return 4;
}
else if (Rs >= 101) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (Rs < 80) {
if (Rs < 23) {
return 2;
}
else if (Rs >= 23) {
return 1;
}
}
else if (Rs >= 80) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 97) {
if (TSC < 2) {
if (Rs < 20) {
return 2;
}
else if (Rs >= 20) {
return 4;
}
}
else if (TSC >= 2) {
return 1;
}
}
}
}
else if (Rs >= 111) {
if (Rs < 248) {
if (mu_mv_mag_sq < 90) {
if (negBalance < 40) {
if (tcor < 98) {
if (diffAFD < 1) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
if (AFD < 11) {
if (tcor < 97) {
if (posBalance < 48) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (Rs < 238) {
return 8;
}
else if (Rs >= 238) {
return 1;
}
}
}
else if (posBalance >= 48) {
return 1;
}
}
else if (tcor >= 97) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (Cs < 95) {
return 8;
}
else if (Cs >= 95) {
return 1;
}
}
}
}
else if (AFD >= 11) {
if (Rs < 144) {
return 2;
}
else if (Rs >= 144) {
return 8;
}
}
}
}
else if (diffAFD >= 1) {
if (ssDCval < 82) {
if (Rs < 206) {
return 8;
}
else if (Rs >= 206) {
return 1;
}
}
else if (ssDCval >= 82) {
return 1;
}
}
}
else if (tcor >= 98) {
if (SC < 234) {
if (mu_mv_mag_sq < 85) {
return 2;
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 118) {
return 2;
}
else if (Rs >= 118) {
return 1;
}
}
}
else if (SC >= 234) {
if (Cs < 116) {
return 1;
}
else if (Cs >= 116) {
if (Cs < 166) {
if (Rs < 218) {
if (Cs < 126) {
if (Rs < 132) {
return 2;
}
else if (Rs >= 132) {
return 1;
}
}
else if (Cs >= 126) {
return 2;
}
}
else if (Rs >= 218) {
return 8;
}
}
else if (Cs >= 166) {
if (SC < 449) {
return 1;
}
else if (SC >= 449) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 40) {
if (ssDCval < 76) {
if (Cs < 228) {
return 1;
}
else if (Cs >= 228) {
return 4;
}
}
else if (ssDCval >= 76) {
if (TSC < 4) {
if (negBalance < 44) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 2;
}
}
else if (negBalance >= 44) {
return 2;
}
}
else if (TSC >= 4) {
if (refDCval < 93) {
return 1;
}
else if (refDCval >= 93) {
if (tcor < 82) {
return 1;
}
else if (tcor >= 82) {
if (ssDCval < 92) {
return 4;
}
else if (ssDCval >= 92) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (TSC < 2) {
if (Cs < 129) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Cs >= 129) {
return 2;
}
}
else if (TSC >= 2) {
if (tcor < 97) {
if (Rs < 120) {
return 1;
}
else if (Rs >= 120) {
if (mu_mv_mag_sq < 100) {
return 8;
}
else if (mu_mv_mag_sq >= 100) {
return 2;
}
}
}
else if (tcor >= 97) {
if (MVDiff < 3) {
if (diffMVdiffVal < -1) {
if (posBalance < 47) {
if (Rs < 193) {
return 2;
}
else if (Rs >= 193) {
return 8;
}
}
else if (posBalance >= 47) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
}
else if (Rs >= 248) {
if (Cs < 186) {
if (Cs < 163) {
if (Rs < 249) {
return 8;
}
else if (Rs >= 249) {
if (negBalance < 52) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (negBalance >= 52) {
if (mu_mv_mag_sq < 40) {
return 1;
}
else if (mu_mv_mag_sq >= 40) {
return 8;
}
}
}
}
else if (Cs >= 163) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (Cs >= 186) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (Cs < 187) {
return 2;
}
else if (Cs >= 187) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 251) {
if (SC < 491) {
if (ssDCval < 75) {
if (ssDCval < 72) {
if (Rs < 148) {
if (Cs < 275) {
if (Rs < 130) {
return 4;
}
else if (Rs >= 130) {
return 1;
}
}
else if (Cs >= 275) {
return 4;
}
}
else if (Rs >= 148) {
if (Cs < 304) {
return 1;
}
else if (Cs >= 304) {
return 4;
}
}
}
else if (ssDCval >= 72) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
else if (ssDCval >= 75) {
if (RsDiff < 1) {
if (refDCval < 82) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
return 4;
}
}
else if (refDCval >= 82) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 92) {
return 2;
}
else if (Rs >= 92) {
return 1;
}
}
}
}
else if (RsDiff >= 1) {
if (Rs < 170) {
if (Cs < 332) {
return 1;
}
else if (Cs >= 332) {
return 2;
}
}
else if (Rs >= 170) {
return 2;
}
}
}
}
else if (SC >= 491) {
if (ssDCval < 79) {
if (tcor < 97) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (tcor >= 97) {
return 1;
}
}
else if (ssDCval >= 79) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 67) {
return 2;
}
else if (mu_mv_mag_sq >= 67) {
if (Cs < 471) {
return 1;
}
else if (Cs >= 471) {
return 2;
}
}
}
else if (RsDiff >= 2) {
if (RsDiff < 4) {
if (Cs < 294) {
return 4;
}
else if (Cs >= 294) {
if (TSC < 6) {
if (negBalance < 62) {
return 1;
}
else if (negBalance >= 62) {
return 2;
}
}
else if (TSC >= 6) {
if (Rs < 254) {
return 8;
}
else if (Rs >= 254) {
return 2;
}
}
}
}
else if (RsDiff >= 4) {
if (Rs < 309) {
return 8;
}
else if (Rs >= 309) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 106) {
if (AFD < 3) {
if (mu_mv_mag_sq < 49) {
return 4;
}
else if (mu_mv_mag_sq >= 49) {
if (SC < 183) {
return 4;
}
else if (SC >= 183) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 2;
}
}
}
}
else if (AFD >= 3) {
if (posBalance < 10) {
return 8;
}
else if (posBalance >= 10) {
if (ssDCval < 119) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
if (tcor < 93) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
if (MVDiff < 134) {
return 1;
}
else if (MVDiff >= 134) {
if (Rs < 94) {
return 4;
}
else if (Rs >= 94) {
if (Rs < 181) {
return 2;
}
else if (Rs >= 181) {
return 1;
}
}
}
}
}
else if (tcor >= 93) {
return 2;
}
}
}
else if (ssDCval >= 119) {
if (refDCval < 163) {
if (posBalance < 57) {
if (Cs < 5) {
return 8;
}
else if (Cs >= 5) {
return 2;
}
}
else if (posBalance >= 57) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
if (AFD < 22) {
if (Rs < 300) {
return 2;
}
else if (Rs >= 300) {
return 8;
}
}
else if (AFD >= 22) {
return 8;
}
}
}
}
else if (refDCval >= 163) {
if (refDCval < 173) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (refDCval >= 173) {
if (Cs < 324) {
return 2;
}
else if (Cs >= 324) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 131) {
if (refDCval < 75) {
if (Rs < 112) {
return 4;
}
else if (Rs >= 112) {
if (Cs < 123) {
return 8;
}
else if (Cs >= 123) {
return 2;
}
}
}
else if (refDCval >= 75) {
if (posBalance < 64) {
if (SC < 452) {
return 1;
}
else if (SC >= 452) {
return 4;
}
}
else if (posBalance >= 64) {
if (Rs < 48) {
return 4;
}
else if (Rs >= 48) {
return 8;
}
}
}
}
else if (Rs >= 131) {
if (Cs < 262) {
if (refDCval < 84) {
if (negBalance < 24) {
if (tcor < 99) {
if (ssDCval < 79) {
if (Cs < 125) {
if (Rs < 139) {
return 2;
}
else if (Rs >= 139) {
return 8;
}
}
else if (Cs >= 125) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
}
}
else if (ssDCval >= 79) {
return 8;
}
}
else if (tcor >= 99) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
if (Rs < 190) {
return 2;
}
else if (Rs >= 190) {
if (Rs < 211) {
return 1;
}
else if (Rs >= 211) {
return 2;
}
}
}
}
}
else if (negBalance >= 24) {
if (Rs < 173) {
if (Rs < 133) {
return 8;
}
else if (Rs >= 133) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
return 1;
}
}
}
}
else if (Rs >= 173) {
if (ssDCval < 74) {
if (Cs < 163) {
if (Rs < 192) {
return 2;
}
else if (Rs >= 192) {
return 1;
}
}
else if (Cs >= 163) {
return 1;
}
}
else if (ssDCval >= 74) {
if (Cs < 169) {
return 8;
}
else if (Cs >= 169) {
return 2;
}
}
}
}
}
else if (refDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (SC < 231) {
return 1;
}
else if (SC >= 231) {
return 2;
}
}
}
}
else if (Cs >= 262) {
if (MVDiff < 4) {
if (Cs < 516) {
return 1;
}
else if (Cs >= 516) {
return 2;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 312) {
if (RsDiff < 4) {
if (refDCval < 75) {
if (mcTcor < 90) {
if (refDCval < 41) {
if (gchDC < 2) {
if (posBalance < 52) {
if (tcor < 82) {
if (Cs < 383) {
if (Rs < 732) {
return 1;
}
else if (Rs >= 732) {
return 2;
}
}
else if (Cs >= 383) {
return 8;
}
}
else if (tcor >= 82) {
return 8;
}
}
else if (posBalance >= 52) {
if (gchDC < 1) {
if (Cs < 133) {
return 1;
}
else if (Cs >= 133) {
return 8;
}
}
else if (gchDC >= 1) {
if (tcor < 70) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
if (Cs < 437) {
return 8;
}
else if (Cs >= 437) {
return 4;
}
}
}
else if (tcor >= 70) {
if (diffTSC < 1) {
if (Rs < 684) {
return 8;
}
else if (Rs >= 684) {
return 2;
}
}
else if (diffTSC >= 1) {
if (Rs < 463) {
return 8;
}
else if (Rs >= 463) {
return 1;
}
}
}
}
}
}
else if (gchDC >= 2) {
if (posBalance < 29) {
if (Rs < 387) {
return 8;
}
else if (Rs >= 387) {
return 4;
}
}
else if (posBalance >= 29) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
}
}
}
else if (refDCval >= 41) {
if (Cs < 133) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
return 2;
}
}
else if (Cs >= 133) {
if (mu_mv_mag_sq < 102) {
if (Cs < 201) {
if (Rs < 415) {
return 8;
}
else if (Rs >= 415) {
return 2;
}
}
else if (Cs >= 201) {
if (Rs < 683) {
if (posBalance < 56) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (MVDiff < 17) {
if (MVDiff < 9) {
if (Cs < 334) {
return 8;
}
else if (Cs >= 334) {
return 2;
}
}
else if (MVDiff >= 9) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (MVDiff >= 17) {
if (mcTcor < 77) {
if (Rs < 427) {
return 8;
}
else if (Rs >= 427) {
return 2;
}
}
else if (mcTcor >= 77) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (diffTSC < 1) {
if (Rs < 481) {
return 4;
}
else if (Rs >= 481) {
return 2;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
}
}
}
}
else if (posBalance >= 56) {
if (mcTcor < 75) {
if (tcor < 51) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
if (Rs < 509) {
return 8;
}
else if (Rs >= 509) {
return 1;
}
}
}
else if (tcor >= 51) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (mcTcor >= 75) {
if (diffRsCsdiff < 6) {
if (diffMVdiffVal < -2) {
if (RsDiff < 3) {
if (Rs < 368) {
if (Rs < 313) {
return 2;
}
else if (Rs >= 313) {
if (Rs < 339) {
return 4;
}
else if (Rs >= 339) {
return 1;
}
}
}
else if (Rs >= 368) {
return 4;
}
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (posBalance < 59) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
return 4;
}
}
else if (posBalance >= 59) {
return 4;
}
}
}
else if (diffRsCsdiff >= 6) {
if (Rs < 434) {
if (Rs < 341) {
return 4;
}
else if (Rs >= 341) {
return 2;
}
}
else if (Rs >= 434) {
if (Rs < 485) {
return 1;
}
else if (Rs >= 485) {
return 8;
}
}
}
}
}
}
else if (Rs >= 683) {
if (ssDCval < 56) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
return 1;
}
}
else if (ssDCval >= 56) {
if (Rs < 733) {
return 1;
}
else if (Rs >= 733) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 574) {
return 8;
}
else if (Rs >= 574) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (Rs < 1292) {
if (tcor < 100) {
if (TSC < 1) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
return 4;
}
}
else if (TSC >= 1) {
if (MVDiff < 2) {
if (refDCval < 69) {
if (negBalance < 40) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 113) {
if (tcor < 90) {
return 4;
}
else if (tcor >= 90) {
if (Cs < 431) {
if (Cs < 271) {
return 2;
}
else if (Cs >= 271) {
return 1;
}
}
else if (Cs >= 431) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (Cs < 440) {
if (Rs < 618) {
return 1;
}
else if (Rs >= 618) {
return 8;
}
}
else if (Cs >= 440) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 363) {
return 4;
}
else if (Rs >= 363) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 475) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (Rs < 378) {
if (Rs < 344) {
return 4;
}
else if (Rs >= 344) {
return 1;
}
}
else if (Rs >= 378) {
return 2;
}
}
}
else if (Rs >= 475) {
if (posBalance < 44) {
if (Rs < 645) {
if (mu_mv_mag_sq < 118) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (posBalance < 35) {
if (TSCindex < 2) {
if (Rs < 511) {
return 1;
}
else if (Rs >= 511) {
return 8;
}
}
else if (TSCindex >= 2) {
if (Rs < 605) {
return 1;
}
else if (Rs >= 605) {
return 8;
}
}
}
else if (posBalance >= 35) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 580) {
return 8;
}
else if (Rs >= 580) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (Cs < 388) {
return 2;
}
else if (Cs >= 388) {
if (posBalance < 38) {
if (Rs < 573) {
return 8;
}
else if (Rs >= 573) {
return 1;
}
}
else if (posBalance >= 38) {
return 1;
}
}
}
}
else if (Rs >= 645) {
if (Cs < 453) {
if (posBalance < 43) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 1;
}
}
else if (posBalance >= 43) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (Cs >= 453) {
if (diffRsCsdiff < 1) {
if (Rs < 721) {
return 1;
}
else if (Rs >= 721) {
return 8;
}
}
else if (diffRsCsdiff >= 1) {
return 1;
}
}
}
}
else if (posBalance >= 44) {
if (CsDiff < 1) {
if (Cs < 393) {
if (negBalance < 25) {
if (negBalance < 16) {
return 1;
}
else if (negBalance >= 16) {
if (Rs < 1068) {
return 2;
}
else if (Rs >= 1068) {
return 1;
}
}
}
else if (negBalance >= 25) {
return 1;
}
}
else if (Cs >= 393) {
return 1;
}
}
else if (CsDiff >= 1) {
if (Cs < 436) {
if (mu_mv_mag_sq < 102) {
return 1;
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 550) {
return 8;
}
else if (Rs >= 550) {
return 1;
}
}
}
else if (Cs >= 436) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (Rs < 491) {
return 2;
}
else if (Rs >= 491) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 40) {
if (Rs < 867) {
if (Rs < 515) {
if (Cs < 234) {
return 8;
}
else if (Cs >= 234) {
return 4;
}
}
else if (Rs >= 515) {
return 8;
}
}
else if (Rs >= 867) {
if (Rs < 1002) {
return 4;
}
else if (Rs >= 1002) {
return 1;
}
}
}
}
else if (refDCval >= 69) {
if (Rs < 901) {
if (Rs < 617) {
if (ssDCval < 74) {
if (SC < 959) {
if (tcor < 98) {
if (posBalance < 56) {
if (Rs < 384) {
return 4;
}
else if (Rs >= 384) {
return 1;
}
}
else if (posBalance >= 56) {
if (Rs < 434) {
return 2;
}
else if (Rs >= 434) {
return 4;
}
}
}
else if (tcor >= 98) {
return 2;
}
}
else if (SC >= 959) {
if (Cs < 394) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 2;
}
}
else if (Cs >= 394) {
if (RsDiff < 2) {
if (Rs < 611) {
return 1;
}
else if (Rs >= 611) {
return 2;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (Rs >= 617) {
if (Rs < 741) {
if (ssDCval < 71) {
if (Cs < 335) {
return 4;
}
else if (Cs >= 335) {
if (Cs < 418) {
return 2;
}
else if (Cs >= 418) {
return 1;
}
}
}
else if (ssDCval >= 71) {
return 4;
}
}
else if (Rs >= 741) {
if (Cs < 520) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 520) {
return 2;
}
}
}
}
else if (Rs >= 901) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (Rs < 1116) {
return 1;
}
else if (Rs >= 1116) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 139) {
if (RsDiff < 2) {
if (ssDCval < 55) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (posBalance < 53) {
if (Cs < 86) {
return 8;
}
else if (Cs >= 86) {
return 1;
}
}
else if (posBalance >= 53) {
return 1;
}
}
}
else if (ssDCval >= 55) {
if (diffMVdiffVal < -6) {
return 1;
}
else if (diffMVdiffVal >= -6) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
}
else if (RsDiff >= 2) {
if (SC < 974) {
if (MVDiff < 33) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (RsDiff < 3) {
if (gchDC < 3) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 1;
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (MVDiff >= 33) {
if (posBalance < 56) {
if (AFD < 10) {
if (negBalance < 39) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
else if (negBalance >= 39) {
return 1;
}
}
else if (AFD >= 10) {
if (Rs < 710) {
return 8;
}
else if (Rs >= 710) {
return 1;
}
}
}
else if (posBalance >= 56) {
return 1;
}
}
}
else if (SC >= 974) {
if (mcTcor < 96) {
if (Cs < 120) {
return 2;
}
else if (Cs >= 120) {
return 1;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
}
}
else if (Cs >= 139) {
if (ssDCval < 71) {
if (refDCval < 57) {
if (Cs < 399) {
if (mcTcor < 92) {
if (posBalance < 50) {
if (SC < 685) {
return 4;
}
else if (SC >= 685) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
if (Rs < 691) {
return 8;
}
else if (Rs >= 691) {
return 1;
}
}
}
}
else if (posBalance >= 50) {
if (ssDCval < 41) {
return 8;
}
else if (ssDCval >= 41) {
if (tcor < 80) {
if (Rs < 393) {
return 2;
}
else if (Rs >= 393) {
return 8;
}
}
else if (tcor >= 80) {
return 1;
}
}
}
}
else if (mcTcor >= 92) {
if (posBalance < 35) {
if (tcor < 96) {
if (Rs < 496) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
return 4;
}
}
else if (Rs >= 496) {
return 2;
}
}
else if (tcor >= 96) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
return 1;
}
}
}
}
else if (posBalance >= 35) {
if (ssDCval < 46) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 70) {
if (posBalance < 55) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (posBalance >= 55) {
return 1;
}
}
else if (mu_mv_mag_sq >= 70) {
if (ssDCval < 38) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (ssDCval >= 38) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (MVDiff < 37) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (Rs < 549) {
return 1;
}
else if (Rs >= 549) {
return 8;
}
}
}
else if (MVDiff >= 37) {
return 2;
}
}
}
else if (ssDCval >= 46) {
if (MVDiff < 4) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (MVDiff >= 4) {
if (Rs < 506) {
if (tcor < 90) {
if (Cs < 379) {
return 4;
}
else if (Cs >= 379) {
return 2;
}
}
else if (tcor >= 90) {
if (Cs < 163) {
if (Rs < 350) {
return 4;
}
else if (Rs >= 350) {
return 1;
}
}
else if (Cs >= 163) {
if (refDCval < 53) {
if (AFD < 4) {
if (Cs < 343) {
return 4;
}
else if (Cs >= 343) {
return 1;
}
}
else if (AFD >= 4) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (diffMVdiffVal < -1) {
if (Cs < 225) {
return 4;
}
else if (Cs >= 225) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
}
else if (refDCval >= 53) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
}
}
else if (Rs >= 506) {
if (negBalance < 16) {
return 4;
}
else if (negBalance >= 16) {
if (tcor < 79) {
return 2;
}
else if (tcor >= 79) {
if (mcTcor < 99) {
if (diffRsCsdiff < 2) {
if (diffMVdiffVal < 12) {
return 1;
}
else if (diffMVdiffVal >= 12) {
return 4;
}
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
else if (mcTcor >= 99) {
if (Rs < 683) {
return 1;
}
else if (Rs >= 683) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 399) {
if (Rs < 670) {
if (MVDiff < 17) {
if (RsCsDiff < 2) {
if (ssDCval < 52) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 596) {
return 8;
}
else if (Rs >= 596) {
return 1;
}
}
}
else if (ssDCval >= 52) {
return 8;
}
}
else if (RsCsDiff >= 2) {
if (ssDCval < 43) {
if (SC < 985) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (SC >= 985) {
return 8;
}
}
else if (ssDCval >= 43) {
if (refDCval < 53) {
if (AFD < 8) {
if (negBalance < 20) {
return 1;
}
else if (negBalance >= 20) {
if (diffMVdiffVal < 5) {
return 4;
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
}
else if (AFD >= 8) {
if (Rs < 487) {
return 8;
}
else if (Rs >= 487) {
return 1;
}
}
}
else if (refDCval >= 53) {
if (mu_mv_mag_sq < 2) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (MVDiff < 10) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
if (refDCval < 56) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 8;
}
}
else if (refDCval >= 56) {
return 1;
}
}
}
else if (MVDiff >= 10) {
if (Cs < 476) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
else if (Cs >= 476) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 17) {
if (negBalance < 47) {
if (SC < 1027) {
if (posBalance < 42) {
if (Rs < 463) {
return 4;
}
else if (Rs >= 463) {
return 1;
}
}
else if (posBalance >= 42) {
return 1;
}
}
else if (SC >= 1027) {
return 4;
}
}
else if (negBalance >= 47) {
if (tcor < 80) {
if (Rs < 515) {
return 2;
}
else if (Rs >= 515) {
return 1;
}
}
else if (tcor >= 80) {
return 8;
}
}
}
}
else if (Rs >= 670) {
if (AFD < 4) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (Cs < 528) {
return 1;
}
else if (Cs >= 528) {
return 4;
}
}
}
else if (AFD >= 4) {
if (posBalance < 45) {
if (mu_mv_mag_sq < 57) {
return 2;
}
else if (mu_mv_mag_sq >= 57) {
if (Cs < 449) {
return 1;
}
else if (Cs >= 449) {
return 4;
}
}
}
else if (posBalance >= 45) {
if (diffRsCsdiff < -8) {
if (Cs < 491) {
return 1;
}
else if (Cs >= 491) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (diffRsCsdiff >= -8) {
if (ssDCval < 39) {
if (Rs < 737) {
return 8;
}
else if (Rs >= 737) {
return 1;
}
}
else if (ssDCval >= 39) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 57) {
if (mu_mv_mag_sq < 78) {
if (Rs < 698) {
if (tcor < 94) {
if (ssDCval < 62) {
if (TSC < 4) {
if (SC < 968) {
if (Rs < 333) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
return 2;
}
}
else if (Rs >= 333) {
return 2;
}
}
else if (SC >= 968) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 524) {
return 2;
}
else if (Rs >= 524) {
if (Rs < 624) {
return 1;
}
else if (Rs >= 624) {
return 8;
}
}
}
}
}
else if (TSC >= 4) {
if (MVDiff < 48) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 1;
}
}
else if (MVDiff >= 48) {
if (mu_mv_mag_sq < 23) {
return 2;
}
else if (mu_mv_mag_sq >= 23) {
if (Rs < 401) {
return 1;
}
else if (Rs >= 401) {
return 4;
}
}
}
}
}
else if (ssDCval >= 62) {
if (mu_mv_mag_sq < 5) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
if (TSC < 6) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 327) {
return 4;
}
else if (Rs >= 327) {
if (Rs < 403) {
return 1;
}
else if (Rs >= 403) {
return 2;
}
}
}
}
else if (TSC >= 6) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (tcor < 84) {
if (MVDiff < 42) {
if (tcor < 72) {
return 2;
}
else if (tcor >= 72) {
return 4;
}
}
else if (MVDiff >= 42) {
return 1;
}
}
else if (tcor >= 84) {
if (negBalance < 52) {
if (SC < 960) {
if (SC < 747) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (SC >= 747) {
return 1;
}
}
else if (SC >= 960) {
return 8;
}
}
else if (negBalance >= 52) {
return 2;
}
}
}
}
}
}
else if (tcor >= 94) {
if (mu_mv_mag_sq < 71) {
if (negBalance < 32) {
if (SC < 911) {
return 4;
}
else if (SC >= 911) {
if (Rs < 543) {
return 8;
}
else if (Rs >= 543) {
return 4;
}
}
}
else if (negBalance >= 32) {
if (mu_mv_mag_sq < 62) {
if (SC < 806) {
if (SC < 779) {
return 4;
}
else if (SC >= 779) {
if (Rs < 494) {
return 1;
}
else if (Rs >= 494) {
return 4;
}
}
}
else if (SC >= 806) {
return 4;
}
}
else if (mu_mv_mag_sq >= 62) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
}
}
}
else if (Rs >= 698) {
if (mcTcor < 96) {
if (SC < 1077) {
return 2;
}
else if (SC >= 1077) {
if (gchDC < 1) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (gchDC >= 1) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (SC < 1530) {
if (Rs < 809) {
return 4;
}
else if (Rs >= 809) {
return 1;
}
}
else if (SC >= 1530) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (tcor < 97) {
if (tcor < 87) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 1;
}
}
else if (tcor >= 87) {
if (Cs < 508) {
return 1;
}
else if (Cs >= 508) {
return 8;
}
}
}
else if (tcor >= 97) {
if (negBalance < 22) {
if (Cs < 313) {
return 4;
}
else if (Cs >= 313) {
return 2;
}
}
else if (negBalance >= 22) {
if (mcTcor < 97) {
return 4;
}
else if (mcTcor >= 97) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (Cs < 279) {
if (refDCval < 63) {
if (Rs < 698) {
return 2;
}
else if (Rs >= 698) {
return 1;
}
}
else if (refDCval >= 63) {
if (posBalance < 37) {
if (Cs < 228) {
if (MVDiff < 34) {
return 4;
}
else if (MVDiff >= 34) {
return 8;
}
}
else if (Cs >= 228) {
return 8;
}
}
else if (posBalance >= 37) {
return 4;
}
}
}
else if (Cs >= 279) {
if (tcor < 91) {
if (SC < 1002) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 2;
}
}
else if (SC >= 1002) {
if (Rs < 640) {
return 8;
}
else if (Rs >= 640) {
return 4;
}
}
}
else if (tcor >= 91) {
if (tcor < 99) {
if (negBalance < 41) {
if (RsCsDiff < 1) {
if (ssDCval < 62) {
if (AFD < 3) {
if (Rs < 508) {
return 2;
}
else if (Rs >= 508) {
return 1;
}
}
else if (AFD >= 3) {
if (Cs < 398) {
return 1;
}
else if (Cs >= 398) {
return 2;
}
}
}
else if (ssDCval >= 62) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (RsCsDiff >= 1) {
if (diffRsCsdiff < -5) {
return 2;
}
else if (diffRsCsdiff >= -5) {
if (refDCval < 69) {
if (MVDiff < 7) {
if (Cs < 472) {
if (negBalance < 26) {
if (Rs < 606) {
return 8;
}
else if (Rs >= 606) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
else if (negBalance >= 26) {
if (Rs < 345) {
return 8;
}
else if (Rs >= 345) {
return 1;
}
}
}
else if (Cs >= 472) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
else if (MVDiff >= 7) {
if (mu_mv_mag_sq < 88) {
return 2;
}
else if (mu_mv_mag_sq >= 88) {
if (Rs < 1096) {
return 8;
}
else if (Rs >= 1096) {
return 1;
}
}
}
}
else if (refDCval >= 69) {
if (mu_mv_mag_sq < 98) {
if (Rs < 451) {
return 8;
}
else if (Rs >= 451) {
return 1;
}
}
else if (mu_mv_mag_sq >= 98) {
return 2;
}
}
}
}
}
else if (negBalance >= 41) {
return 2;
}
}
else if (tcor >= 99) {
if (mu_mv_mag_sq < 104) {
if (Rs < 413) {
return 4;
}
else if (Rs >= 413) {
return 8;
}
}
else if (mu_mv_mag_sq >= 104) {
if (posBalance < 40) {
if (Cs < 400) {
return 1;
}
else if (Cs >= 400) {
return 8;
}
}
else if (posBalance >= 40) {
if (Rs < 991) {
if (Cs < 518) {
return 2;
}
else if (Cs >= 518) {
return 8;
}
}
else if (Rs >= 991) {
return 1;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (mu_mv_mag_sq < 96) {
if (Rs < 531) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (Rs >= 531) {
if (Rs < 736) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Rs >= 736) {
if (MVDiff < 23) {
if (Cs < 471) {
return 1;
}
else if (Cs >= 471) {
return 8;
}
}
else if (MVDiff >= 23) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 96) {
if (posBalance < 49) {
if (Cs < 396) {
return 8;
}
else if (Cs >= 396) {
return 4;
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 100) {
if (Rs < 904) {
if (RsDiff < 1) {
if (SC < 853) {
if (refDCval < 39) {
return 1;
}
else if (refDCval >= 39) {
if (refDCval < 65) {
if (MVDiff < 9) {
if (Rs < 386) {
return 4;
}
else if (Rs >= 386) {
return 8;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (refDCval >= 65) {
if (Rs < 364) {
return 4;
}
else if (Rs >= 364) {
return 8;
}
}
}
}
else if (SC >= 853) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (negBalance < 15) {
return 1;
}
else if (negBalance >= 15) {
if (Cs < 277) {
if (Rs < 734) {
return 8;
}
else if (Rs >= 734) {
return 4;
}
}
else if (Cs >= 277) {
if (Rs < 863) {
return 4;
}
else if (Rs >= 863) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 1) {
if (negBalance < 44) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
if (Rs < 441) {
return 8;
}
else if (Rs >= 441) {
return 2;
}
}
}
else if (negBalance >= 44) {
if (Cs < 113) {
return 1;
}
else if (Cs >= 113) {
if (MVDiff < 26) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (diffMVdiffVal < -2) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
else if (MVDiff >= 26) {
return 4;
}
}
}
}
}
else if (Rs >= 904) {
if (ssDCval < 68) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 1129) {
return 2;
}
else if (Rs >= 1129) {
return 1;
}
}
}
else if (ssDCval >= 68) {
return 8;
}
}
}
}
else if (Rs >= 1292) {
if (SC < 1600) {
if (mu_mv_mag_sq < 58) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (mu_mv_mag_sq >= 58) {
if (Cs < 222) {
return 8;
}
else if (Cs >= 222) {
if (Rs < 1324) {
return 8;
}
else if (Rs >= 1324) {
return 1;
}
}
}
}
else if (SC >= 1600) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1760) {
if (refDCval < 64) {
if (ssDCval < 56) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (ssDCval >= 56) {
if (Cs < 424) {
if (posBalance < 69) {
if (Cs < 146) {
if (Rs < 1517) {
return 8;
}
else if (Rs >= 1517) {
return 1;
}
}
else if (Cs >= 146) {
return 1;
}
}
else if (posBalance >= 69) {
return 2;
}
}
else if (Cs >= 424) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (negBalance < 26) {
if (Rs < 1476) {
return 2;
}
else if (Rs >= 1476) {
return 1;
}
}
else if (negBalance >= 26) {
return 8;
}
}
}
}
}
else if (refDCval >= 64) {
if (mu_mv_mag_sq < 56) {
if (AFD < 5) {
if (Cs < 380) {
if (Cs < 275) {
return 1;
}
else if (Cs >= 275) {
return 4;
}
}
else if (Cs >= 380) {
return 2;
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 56) {
if (refDCval < 69) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
if (Rs < 1423) {
return 8;
}
else if (Rs >= 1423) {
return 1;
}
}
}
else if (refDCval >= 69) {
if (MVDiff < 4) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (MVDiff >= 4) {
if (diffMVdiffVal < -2) {
if (Rs < 1578) {
return 8;
}
else if (Rs >= 1578) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1760) {
if (Cs < 434) {
if (diffMVdiffVal < -4) {
if (Rs < 1792) {
return 2;
}
else if (Rs >= 1792) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 2;
}
}
else if (Cs >= 434) {
if (mu_mv_mag_sq < 62) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 1;
}
}
else if (mu_mv_mag_sq >= 62) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 62) {
if (Cs < 517) {
return 8;
}
else if (Cs >= 517) {
return 1;
}
}
else if (ssDCval >= 62) {
if (Rs < 1412) {
return 8;
}
else if (Rs >= 1412) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (Cs < 142) {
if (posBalance < 16) {
if (MVDiff < 164) {
return 8;
}
else if (MVDiff >= 164) {
return 1;
}
}
else if (posBalance >= 16) {
if (ssDCval < 106) {
if (Rs < 1126) {
if (refDCval < 96) {
if (TSC < 4) {
if (Cs < 9) {
return 1;
}
else if (Cs >= 9) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (Cs < 85) {
return 8;
}
else if (Cs >= 85) {
if (Rs < 325) {
return 8;
}
else if (Rs >= 325) {
return 2;
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 44) {
if (Cs < 91) {
return 1;
}
else if (Cs >= 91) {
if (Rs < 428) {
return 8;
}
else if (Rs >= 428) {
if (MVDiff < 37) {
return 1;
}
else if (MVDiff >= 37) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (refDCval < 81) {
return 1;
}
else if (refDCval >= 81) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
}
else if (refDCval >= 96) {
if (Cs < 19) {
return 4;
}
else if (Cs >= 19) {
return 1;
}
}
}
else if (Rs >= 1126) {
return 1;
}
}
else if (ssDCval >= 106) {
if (MVDiff < 20) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (MVDiff >= 20) {
if (SC < 525) {
if (Cs < 46) {
return 8;
}
else if (Cs >= 46) {
return 2;
}
}
else if (SC >= 525) {
return 1;
}
}
}
}
}
else if (Cs >= 142) {
if (SC < 1892) {
if (SC < 855) {
if (AFD < 2) {
if (ssDCval < 94) {
if (Cs < 273) {
if (Rs < 313) {
if (Cs < 200) {
return 1;
}
else if (Cs >= 200) {
return 2;
}
}
else if (Rs >= 313) {
return 2;
}
}
else if (Cs >= 273) {
return 8;
}
}
else if (ssDCval >= 94) {
if (Cs < 310) {
return 1;
}
else if (Cs >= 310) {
return 4;
}
}
}
else if (AFD >= 2) {
if (mcTcor < 96) {
if (MVDiff < 22) {
if (diffAFD < 1) {
if (diffAFD < -1) {
if (mu_mv_mag_sq < 3) {
if (Rs < 338) {
return 2;
}
else if (Rs >= 338) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (CsDiff < 1) {
if (Rs < 495) {
return 1;
}
else if (Rs >= 495) {
return 8;
}
}
else if (CsDiff >= 1) {
if (tcor < 86) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (tcor >= 86) {
return 8;
}
}
}
}
else if (diffAFD >= -1) {
if (Rs < 471) {
return 8;
}
else if (Rs >= 471) {
return 1;
}
}
}
else if (diffAFD >= 1) {
if (gchDC < 2) {
if (diffAFD < 3) {
if (Rs < 339) {
return 1;
}
else if (Rs >= 339) {
return 8;
}
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (MVDiff >= 22) {
if (Rs < 464) {
if (gchDC < 1) {
if (Rs < 322) {
return 4;
}
else if (Rs >= 322) {
return 8;
}
}
else if (gchDC >= 1) {
if (AFD < 9) {
if (tcor < 82) {
return 2;
}
else if (tcor >= 82) {
if (Rs < 331) {
return 8;
}
else if (Rs >= 331) {
if (Cs < 400) {
return 4;
}
else if (Cs >= 400) {
return 8;
}
}
}
}
else if (AFD >= 9) {
if (refDCval < 79) {
if (SC < 760) {
return 4;
}
else if (SC >= 760) {
return 2;
}
}
else if (refDCval >= 79) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
if (Cs < 170) {
return 8;
}
else if (Cs >= 170) {
return 2;
}
}
}
}
}
}
else if (Rs >= 464) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (ssDCval < 78) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (SC < 768) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
else if (SC >= 768) {
if (Rs < 617) {
return 4;
}
else if (Rs >= 617) {
return 1;
}
}
}
}
else if (ssDCval >= 78) {
if (MVDiff < 30) {
if (MVDiff < 9) {
if (Cs < 311) {
if (ssDCval < 98) {
if (mu_mv_mag_sq < 78) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
if (Rs < 393) {
return 1;
}
else if (Rs >= 393) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
return 1;
}
}
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (Cs >= 311) {
if (posBalance < 36) {
if (Rs < 414) {
return 1;
}
else if (Rs >= 414) {
return 4;
}
}
else if (posBalance >= 36) {
if (Cs < 407) {
if (TSC < 5) {
if (negBalance < 26) {
if (Rs < 446) {
return 4;
}
else if (Rs >= 446) {
return 1;
}
}
else if (negBalance >= 26) {
return 1;
}
}
else if (TSC >= 5) {
if (Rs < 467) {
return 4;
}
else if (Rs >= 467) {
return 8;
}
}
}
else if (Cs >= 407) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
}
else if (MVDiff >= 9) {
if (Rs < 391) {
if (ssDCval < 112) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (ssDCval >= 112) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
if (Rs < 367) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 2;
}
}
else if (Rs >= 367) {
return 1;
}
}
}
}
else if (Rs >= 391) {
if (posBalance < 42) {
if (Cs < 386) {
return 2;
}
else if (Cs >= 386) {
return 1;
}
}
else if (posBalance >= 42) {
if (AFD < 17) {
if (ssDCval < 80) {
if (Rs < 568) {
return 8;
}
else if (Rs >= 568) {
return 1;
}
}
else if (ssDCval >= 80) {
return 1;
}
}
else if (AFD >= 17) {
return 2;
}
}
}
}
}
else if (MVDiff >= 30) {
if (mu_mv_mag_sq < 14) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (posBalance < 72) {
return 8;
}
else if (posBalance >= 72) {
return 2;
}
}
}
}
}
}
}
else if (SC >= 855) {
if (refDCval < 116) {
if (ssDCval < 95) {
if (mu_mv_mag_sq < 40) {
if (tcor < 75) {
if (refDCval < 86) {
return 8;
}
else if (refDCval >= 86) {
if (Cs < 509) {
return 2;
}
else if (Cs >= 509) {
if (Rs < 516) {
return 8;
}
else if (Rs >= 516) {
return 2;
}
}
}
}
else if (tcor >= 75) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Cs < 440) {
if (Cs < 250) {
if (mu_mv_mag_sq < 15) {
if (Rs < 1059) {
return 4;
}
else if (Rs >= 1059) {
return 1;
}
}
else if (mu_mv_mag_sq >= 15) {
return 1;
}
}
else if (Cs >= 250) {
return 1;
}
}
else if (Cs >= 440) {
if (MVDiff < 16) {
if (tcor < 89) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (tcor >= 89) {
return 1;
}
}
else if (MVDiff >= 16) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (MVDiff < 42) {
if (Rs < 658) {
return 8;
}
else if (Rs >= 658) {
return 4;
}
}
else if (MVDiff >= 42) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (negBalance < 31) {
if (SC < 1239) {
if (ssDCval < 94) {
if (MVDiff < 3) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (posBalance < 23) {
return 8;
}
else if (posBalance >= 23) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (Cs < 267) {
return 4;
}
else if (Cs >= 267) {
return 8;
}
}
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (SC >= 1239) {
if (mu_mv_mag_sq < 117) {
if (Rs < 946) {
return 8;
}
else if (Rs >= 946) {
return 1;
}
}
else if (mu_mv_mag_sq >= 117) {
if (Rs < 1034) {
if (Cs < 268) {
return 2;
}
else if (Cs >= 268) {
return 1;
}
}
else if (Rs >= 1034) {
return 2;
}
}
}
}
else if (negBalance >= 31) {
if (refDCval < 80) {
if (RsCsDiff < 11) {
if (posBalance < 49) {
if (MVDiff < 4) {
if (CsDiff < 1) {
if (Rs < 974) {
return 8;
}
else if (Rs >= 974) {
return 1;
}
}
else if (CsDiff >= 1) {
if (Cs < 238) {
return 4;
}
else if (Cs >= 238) {
return 1;
}
}
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (posBalance >= 49) {
return 8;
}
}
else if (RsCsDiff >= 11) {
if (SC < 942) {
return 2;
}
else if (SC >= 942) {
if (Rs < 881) {
return 4;
}
else if (Rs >= 881) {
return 1;
}
}
}
}
else if (refDCval >= 80) {
if (negBalance < 40) {
if (MVDiff < 3) {
if (Rs < 748) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 1;
}
}
}
else if (Rs >= 748) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Rs < 1154) {
return 8;
}
else if (Rs >= 1154) {
return 1;
}
}
}
else if (negBalance >= 40) {
if (mu_mv_mag_sq < 62) {
if (negBalance < 48) {
if (Rs < 551) {
if (Rs < 514) {
return 8;
}
else if (Rs >= 514) {
return 1;
}
}
else if (Rs >= 551) {
return 2;
}
}
else if (negBalance >= 48) {
if (refDCval < 82) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (refDCval >= 82) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 62) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 95) {
if (ssDCval < 98) {
if (Cs < 452) {
if (posBalance < 51) {
if (Rs < 914) {
return 4;
}
else if (Rs >= 914) {
return 1;
}
}
else if (posBalance >= 51) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (Cs >= 452) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (Rs < 491) {
return 2;
}
else if (Rs >= 491) {
if (Rs < 577) {
return 1;
}
else if (Rs >= 577) {
return 8;
}
}
}
}
}
else if (ssDCval >= 98) {
if (Rs < 755) {
if (ssDCval < 102) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (ssDCval >= 102) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (refDCval < 115) {
if (gchDC < 3) {
if (ssDCval < 111) {
return 1;
}
else if (ssDCval >= 111) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 432) {
return 1;
}
else if (Rs >= 432) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
if (Rs < 555) {
return 2;
}
else if (Rs >= 555) {
return 8;
}
}
}
else if (refDCval >= 115) {
return 8;
}
}
}
}
}
else if (Rs >= 755) {
return 8;
}
}
}
}
else if (refDCval >= 116) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 80) {
if (Rs < 483) {
return 1;
}
else if (Rs >= 483) {
return 2;
}
}
else if (mu_mv_mag_sq >= 80) {
return 8;
}
}
else if (MVDiff >= 3) {
if (ssDCval < 119) {
if (Rs < 368) {
return 4;
}
else if (Rs >= 368) {
if (mu_mv_mag_sq < 35) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Rs < 456) {
return 2;
}
else if (Rs >= 456) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (Rs < 605) {
return 2;
}
else if (Rs >= 605) {
return 8;
}
}
}
}
else if (ssDCval >= 119) {
if (mcTcor < 99) {
if (MVDiff < 21) {
if (Cs < 382) {
return 4;
}
else if (Cs >= 382) {
if (posBalance < 55) {
if (diffAFD < 1) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Rs < 1111) {
return 8;
}
else if (Rs >= 1111) {
return 4;
}
}
}
else if (diffAFD >= 1) {
if (Cs < 474) {
if (ssDCval < 156) {
return 1;
}
else if (ssDCval >= 156) {
return 8;
}
}
else if (Cs >= 474) {
return 2;
}
}
}
else if (posBalance >= 55) {
if (AFD < 6) {
if (SC < 926) {
return 8;
}
else if (SC >= 926) {
return 4;
}
}
else if (AFD >= 6) {
if (Rs < 439) {
return 4;
}
else if (Rs >= 439) {
return 1;
}
}
}
}
}
else if (MVDiff >= 21) {
if (AFD < 8) {
if (Rs < 748) {
return 4;
}
else if (Rs >= 748) {
return 1;
}
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 405) {
return 4;
}
else if (Rs >= 405) {
if (gchDC < 5) {
return 1;
}
else if (gchDC >= 5) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 1892) {
if (posBalance < 54) {
if (Cs < 482) {
if (refDCval < 96) {
if (AFD < 3) {
if (Rs < 2111) {
return 1;
}
else if (Rs >= 2111) {
return 8;
}
}
else if (AFD >= 3) {
if (AFD < 5) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 69) {
return 8;
}
else if (mu_mv_mag_sq >= 69) {
if (MVDiff < 4) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (Rs < 1630) {
return 8;
}
else if (Rs >= 1630) {
return 1;
}
}
}
else if (MVDiff >= 4) {
return 1;
}
}
}
else if (ssDCval >= 84) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Cs < 313) {
return 1;
}
else if (Cs >= 313) {
return 8;
}
}
}
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (refDCval >= 96) {
if (ssDCval < 111) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (ssDCval >= 111) {
if (Rs < 1779) {
return 4;
}
else if (Rs >= 1779) {
return 1;
}
}
}
}
else if (Cs >= 482) {
if (Rs < 1940) {
if (MVDiff < 1) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (Rs >= 1940) {
if (AFD < 6) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (AFD >= 6) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
if (Cs < 344) {
if (ssDCval < 96) {
return 1;
}
else if (ssDCval >= 96) {
if (Rs < 1867) {
return 8;
}
else if (Rs >= 1867) {
return 4;
}
}
}
else if (Cs >= 344) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (Cs < 387) {
if (RsCsDiff < 67) {
if (negBalance < 47) {
if (mu_mv_mag_sq < 57) {
if (MVDiff < 13) {
if (Rs < 583) {
return 1;
}
else if (Rs >= 583) {
return 2;
}
}
else if (MVDiff >= 13) {
if (mcTcor < 72) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
if (Rs < 415) {
return 8;
}
else if (Rs >= 415) {
return 2;
}
}
}
else if (mcTcor >= 72) {
if (Rs < 520) {
if (tcor < 54) {
return 1;
}
else if (tcor >= 54) {
if (Cs < 217) {
return 2;
}
else if (Cs >= 217) {
return 4;
}
}
}
else if (Rs >= 520) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (Cs < 256) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 2;
}
}
else if (Cs >= 256) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (posBalance < 46) {
if (MVDiff < 33) {
if (Rs < 519) {
return 8;
}
else if (Rs >= 519) {
return 2;
}
}
else if (MVDiff >= 33) {
if (diffAFD < 12) {
return 1;
}
else if (diffAFD >= 12) {
return 8;
}
}
}
else if (posBalance >= 46) {
if (refDCval < 76) {
if (Rs < 724) {
if (diffRsCsdiff < -6) {
if (negBalance < 50) {
if (Rs < 458) {
return 8;
}
else if (Rs >= 458) {
return 4;
}
}
else if (negBalance >= 50) {
if (Cs < 360) {
return 8;
}
else if (Cs >= 360) {
if (Rs < 460) {
return 4;
}
else if (Rs >= 460) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= -6) {
if (Rs < 408) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
else if (Rs >= 408) {
if (posBalance < 54) {
if (Rs < 651) {
return 8;
}
else if (Rs >= 651) {
return 1;
}
}
else if (posBalance >= 54) {
if (diffRsCsdiff < 13) {
if (diffRsCsdiff < -3) {
return 1;
}
else if (diffRsCsdiff >= -3) {
return 4;
}
}
else if (diffRsCsdiff >= 13) {
return 1;
}
}
}
}
}
else if (Rs >= 724) {
if (MVDiff < 131) {
return 1;
}
else if (MVDiff >= 131) {
return 8;
}
}
}
else if (refDCval >= 76) {
if (tcor < 37) {
if (AFD < 17) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 4;
}
}
else if (AFD >= 17) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 1;
}
}
}
else if (tcor >= 37) {
if (MVDiff < 59) {
if (Rs < 1233) {
if (mu_mv_mag_sq < 6) {
return 8;
}
else if (mu_mv_mag_sq >= 6) {
if (AFD < 12) {
if (MVDiff < 36) {
if (diffMVdiffVal < -3) {
if (Rs < 492) {
return 8;
}
else if (Rs >= 492) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (MVDiff >= 36) {
return 2;
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 50) {
if (diffRsCsdiff < -3) {
if (Cs < 259) {
return 1;
}
else if (Cs >= 259) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 8;
}
}
}
else if (diffRsCsdiff >= -3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 50) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 2;
}
}
}
}
}
else if (Rs >= 1233) {
if (Rs < 1451) {
return 2;
}
else if (Rs >= 1451) {
return 8;
}
}
}
else if (MVDiff >= 59) {
if (Rs < 675) {
return 2;
}
else if (Rs >= 675) {
if (mu_mv_mag_sq < 31) {
if (Rs < 1156) {
return 1;
}
else if (Rs >= 1156) {
return 2;
}
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 67) {
if (MVDiff < 28) {
if (refDCval < 76) {
if (Rs < 1398) {
return 1;
}
else if (Rs >= 1398) {
if (TSC < 7) {
if (ssDCval < 69) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
else if (ssDCval >= 69) {
if (Rs < 2488) {
return 1;
}
else if (Rs >= 2488) {
return 2;
}
}
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (refDCval >= 76) {
if (ssDCval < 116) {
return 2;
}
else if (ssDCval >= 116) {
if (Cs < 265) {
if (Rs < 1516) {
return 1;
}
else if (Rs >= 1516) {
return 8;
}
}
else if (Cs >= 265) {
return 2;
}
}
}
}
else if (MVDiff >= 28) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 30) {
if (SC < 2446) {
if (RsDiff < 22) {
return 1;
}
else if (RsDiff >= 22) {
if (Rs < 1772) {
return 2;
}
else if (Rs >= 1772) {
return 1;
}
}
}
else if (SC >= 2446) {
if (AFD < 24) {
if (Rs < 2278) {
return 4;
}
else if (Rs >= 2278) {
return 2;
}
}
else if (AFD >= 24) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (diffMVdiffVal < -1) {
if (Rs < 1094) {
return 2;
}
else if (Rs >= 1094) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (diffTSC < 1) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (diffAFD < 4) {
return 2;
}
else if (diffAFD >= 4) {
if (Rs < 1348) {
return 2;
}
else if (Rs >= 1348) {
return 1;
}
}
}
}
else if (diffTSC >= 1) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
else if (posBalance >= 54) {
if (ssDCval < 112) {
if (Rs < 1278) {
if (ssDCval < 83) {
if (RsDiff < 11) {
if (Cs < 253) {
if (Rs < 939) {
return 8;
}
else if (Rs >= 939) {
return 1;
}
}
else if (Cs >= 253) {
if (MVDiff < 50) {
if (Rs < 1046) {
return 2;
}
else if (Rs >= 1046) {
return 1;
}
}
else if (MVDiff >= 50) {
return 1;
}
}
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (ssDCval >= 83) {
if (RsDiff < 13) {
return 2;
}
else if (RsDiff >= 13) {
if (Rs < 1203) {
return 4;
}
else if (Rs >= 1203) {
return 1;
}
}
}
}
else if (Rs >= 1278) {
if (Cs < 194) {
if (tcor < 6) {
return 4;
}
else if (tcor >= 6) {
return 2;
}
}
else if (Cs >= 194) {
if (diffRsCsdiff < 48) {
if (TSC < 11) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (TSC >= 11) {
return 2;
}
}
else if (diffRsCsdiff >= 48) {
if (AFD < 32) {
return 2;
}
else if (AFD >= 32) {
if (Rs < 2328) {
return 1;
}
else if (Rs >= 2328) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (diffAFD < 6) {
if (mcTcor < 89) {
if (Rs < 1933) {
return 2;
}
else if (Rs >= 1933) {
return 8;
}
}
else if (mcTcor >= 89) {
if (refDCval < 151) {
if (Cs < 77) {
return 2;
}
else if (Cs >= 77) {
return 1;
}
}
else if (refDCval >= 151) {
return 8;
}
}
}
else if (diffAFD >= 6) {
return 1;
}
}
}
}
}
}
else if (Cs >= 387) {
if (negBalance < 51) {
if (AFD < 14) {
if (tcor < 71) {
if (TSC < 7) {
if (posBalance < 59) {
if (negBalance < 46) {
if (Cs < 520) {
return 1;
}
else if (Cs >= 520) {
if (Rs < 779) {
return 1;
}
else if (Rs >= 779) {
return 2;
}
}
}
else if (negBalance >= 46) {
return 2;
}
}
else if (posBalance >= 59) {
if (tcor < 47) {
return 1;
}
else if (tcor >= 47) {
if (diffMVdiffVal < -2) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 528) {
return 8;
}
else if (Cs >= 528) {
return 1;
}
}
}
}
}
else if (TSC >= 7) {
if (Cs < 482) {
return 8;
}
else if (Cs >= 482) {
return 2;
}
}
}
else if (tcor >= 71) {
if (diffRsCsdiff < 12) {
return 8;
}
else if (diffRsCsdiff >= 12) {
return 1;
}
}
}
else if (AFD >= 14) {
if (posBalance < 67) {
if (RsDiff < 12) {
if (Cs < 476) {
if (posBalance < 61) {
if (ssDCval < 51) {
if (Rs < 787) {
return 1;
}
else if (Rs >= 787) {
return 2;
}
}
else if (ssDCval >= 51) {
return 1;
}
}
else if (posBalance >= 61) {
if (Rs < 1147) {
return 8;
}
else if (Rs >= 1147) {
return 2;
}
}
}
else if (Cs >= 476) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (posBalance >= 67) {
if (negBalance < 37) {
if (tcor < 28) {
return 1;
}
else if (tcor >= 28) {
return 2;
}
}
else if (negBalance >= 37) {
if (diffRsCsdiff < 4) {
if (SC < 1201) {
return 4;
}
else if (SC >= 1201) {
if (Rs < 996) {
return 8;
}
else if (Rs >= 996) {
return 1;
}
}
}
else if (diffRsCsdiff >= 4) {
if (RsDiff < 16) {
if (gchDC < 4) {
return 1;
}
else if (gchDC >= 4) {
return 8;
}
}
else if (RsDiff >= 16) {
if (Cs < 455) {
return 2;
}
else if (Cs >= 455) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 51) {
if (AFD < 21) {
if (ssDCval < 61) {
if (SC < 961) {
if (Rs < 470) {
return 4;
}
else if (Rs >= 470) {
return 2;
}
}
else if (SC >= 961) {
if (mu_mv_mag_sq < 18) {
return 8;
}
else if (mu_mv_mag_sq >= 18) {
if (negBalance < 60) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
else if (negBalance >= 60) {
return 2;
}
}
}
}
else if (ssDCval >= 61) {
if (RsCsDiff < 37) {
if (Rs < 666) {
if (TSC < 9) {
if (mu_mv_mag_sq < 31) {
return 1;
}
else if (mu_mv_mag_sq >= 31) {
return 8;
}
}
else if (TSC >= 9) {
return 4;
}
}
else if (Rs >= 666) {
return 8;
}
}
else if (RsCsDiff >= 37) {
if (CsDiff < 3) {
if (mcTcor < 79) {
return 4;
}
else if (mcTcor >= 79) {
if (Rs < 2199) {
return 2;
}
else if (Rs >= 2199) {
return 4;
}
}
}
else if (CsDiff >= 3) {
if (Rs < 1826) {
if (TSC < 7) {
if (Cs < 465) {
return 8;
}
else if (Cs >= 465) {
if (Rs < 1297) {
return 4;
}
else if (Rs >= 1297) {
return 8;
}
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (Rs >= 1826) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
}
}
else if (AFD >= 21) {
if (gchDC < 5) {
if (mu_mv_mag_sq < 28) {
if (Rs < 2105) {
if (AFD < 24) {
if (Rs < 1286) {
return 4;
}
else if (Rs >= 1286) {
return 8;
}
}
else if (AFD >= 24) {
if (tcor < 38) {
return 2;
}
else if (tcor >= 38) {
if (diffTSC < -3) {
if (Rs < 723) {
return 1;
}
else if (Rs >= 723) {
return 2;
}
}
else if (diffTSC >= -3) {
return 1;
}
}
}
}
else if (Rs >= 2105) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 28) {
if (RsDiff < 9) {
if (negBalance < 58) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 1;
}
}
else if (negBalance >= 58) {
if (Cs < 427) {
return 8;
}
else if (Cs >= 427) {
return 2;
}
}
}
else if (RsDiff >= 9) {
if (AFD < 34) {
if (Rs < 2124) {
if (mu_mv_mag_sq < 51) {
return 4;
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 1780) {
return 8;
}
else if (Rs >= 1780) {
return 4;
}
}
}
else if (Rs >= 2124) {
return 2;
}
}
else if (AFD >= 34) {
return 2;
}
}
}
}
else if (gchDC >= 5) {
if (negBalance < 64) {
if (TSC < 10) {
if (Rs < 1079) {
return 2;
}
else if (Rs >= 1079) {
return 1;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (negBalance >= 64) {
if (diffTSC < 4) {
if (Rs < 538) {
return 2;
}
else if (Rs >= 538) {
return 1;
}
}
else if (diffTSC >= 4) {
if (Rs < 766) {
return 4;
}
else if (Rs >= 766) {
return 8;
}
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 78) {
if (RsDiff < 9) {
if (ssDCval < 139) {
if (Cs < 269) {
if (refDCval < 79) {
if (MVDiff < 60) {
if (negBalance < 32) {
if (refDCval < 56) {
if (diffMVdiffVal < -9) {
return 1;
}
else if (diffMVdiffVal >= -9) {
return 2;
}
}
else if (refDCval >= 56) {
return 1;
}
}
else if (negBalance >= 32) {
if (SC < 2106) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
if (Cs < 220) {
return 1;
}
else if (Cs >= 220) {
if (Rs < 483) {
return 1;
}
else if (Rs >= 483) {
return 4;
}
}
}
}
else if (SC >= 2106) {
if (Rs < 1922) {
return 2;
}
else if (Rs >= 1922) {
return 1;
}
}
}
}
else if (MVDiff >= 60) {
if (Cs < 198) {
if (MVDiff < 229) {
if (diffAFD < -13) {
return 8;
}
else if (diffAFD >= -13) {
return 1;
}
}
else if (MVDiff >= 229) {
return 2;
}
}
else if (Cs >= 198) {
if (Rs < 888) {
return 8;
}
else if (Rs >= 888) {
return 1;
}
}
}
}
else if (refDCval >= 79) {
if (SC < 807) {
if (MVDiff < 199) {
if (MVDiff < 53) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
return 1;
}
}
else if (MVDiff >= 53) {
return 8;
}
}
else if (MVDiff >= 199) {
if (Rs < 441) {
return 1;
}
else if (Rs >= 441) {
return 2;
}
}
}
else if (SC >= 807) {
if (MVDiff < 18) {
if (diffRsCsdiff < -12) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (diffRsCsdiff >= -12) {
if (negBalance < 51) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (Rs < 1500) {
return 1;
}
else if (Rs >= 1500) {
return 2;
}
}
}
else if (negBalance >= 51) {
if (Rs < 1089) {
return 1;
}
else if (Rs >= 1089) {
return 2;
}
}
}
}
else if (MVDiff >= 18) {
if (posBalance < 101) {
if (mu_mv_mag_sq < 14) {
if (tcor < 96) {
if (negBalance < 56) {
if (tcor < 90) {
if (mu_mv_mag_sq < 2) {
if (Rs < 792) {
if (Rs < 776) {
return 1;
}
else if (Rs >= 776) {
return 4;
}
}
else if (Rs >= 792) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (negBalance < 42) {
if (Rs < 977) {
return 2;
}
else if (Rs >= 977) {
return 1;
}
}
else if (negBalance >= 42) {
return 2;
}
}
}
else if (tcor >= 90) {
return 1;
}
}
else if (negBalance >= 56) {
if (Rs < 1328) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
else if (Rs >= 1328) {
return 8;
}
}
}
else if (tcor >= 96) {
return 2;
}
}
else if (mu_mv_mag_sq >= 14) {
if (refDCval < 132) {
if (diffRsCsdiff < -91) {
if (Rs < 1883) {
return 1;
}
else if (Rs >= 1883) {
return 2;
}
}
else if (diffRsCsdiff >= -91) {
if (negBalance < 86) {
return 1;
}
else if (negBalance >= 86) {
if (Rs < 1175) {
return 1;
}
else if (Rs >= 1175) {
return 2;
}
}
}
}
else if (refDCval >= 132) {
return 2;
}
}
}
else if (posBalance >= 101) {
return 2;
}
}
}
}
}
else if (Cs >= 269) {
if (Cs < 454) {
if (tcor < 93) {
if (refDCval < 96) {
if (negBalance < 54) {
if (CsDiff < 3) {
if (Cs < 391) {
if (tcor < 76) {
if (Cs < 278) {
return 4;
}
else if (Cs >= 278) {
return 1;
}
}
else if (tcor >= 76) {
return 1;
}
}
else if (Cs >= 391) {
if (refDCval < 58) {
if (negBalance < 36) {
if (Cs < 445) {
return 4;
}
else if (Cs >= 445) {
return 1;
}
}
else if (negBalance >= 36) {
if (Rs < 941) {
return 2;
}
else if (Rs >= 941) {
return 1;
}
}
}
else if (refDCval >= 58) {
if (Cs < 393) {
return 2;
}
else if (Cs >= 393) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 40) {
if (TSC < 8) {
if (diffTSC < -1) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (RsDiff < 7) {
if (Cs < 437) {
if (Rs < 1781) {
return 2;
}
else if (Rs >= 1781) {
return 4;
}
}
else if (Cs >= 437) {
return 4;
}
}
else if (RsDiff >= 7) {
if (Rs < 1836) {
return 1;
}
else if (Rs >= 1836) {
return 2;
}
}
}
}
else if (diffTSC >= -1) {
if (refDCval < 65) {
return 8;
}
else if (refDCval >= 65) {
return 1;
}
}
}
else if (TSC >= 8) {
if (RsCsDiff < 60) {
return 4;
}
else if (RsCsDiff >= 60) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (diffMVdiffVal < 7) {
if (Cs < 432) {
if (Rs < 410) {
return 4;
}
else if (Rs >= 410) {
return 8;
}
}
else if (Cs >= 432) {
return 1;
}
}
else if (diffMVdiffVal >= 7) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (ssDCval < 67) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
return 1;
}
}
else if (ssDCval >= 67) {
if (tcor < 79) {
if (Rs < 848) {
return 8;
}
else if (Rs >= 848) {
return 4;
}
}
else if (tcor >= 79) {
return 1;
}
}
}
}
}
else if (refDCval >= 96) {
if (RsDiff < 6) {
if (ssDCval < 124) {
if (CsDiff < 2) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (ssDCval >= 124) {
return 1;
}
}
else if (RsDiff >= 6) {
if (ssDCval < 103) {
if (Rs < 1641) {
return 2;
}
else if (Rs >= 1641) {
return 8;
}
}
else if (ssDCval >= 103) {
if (Rs < 2053) {
if (mcTcor < 93) {
return 2;
}
else if (mcTcor >= 93) {
return 1;
}
}
else if (Rs >= 2053) {
if (AFD < 20) {
if (diffMVdiffVal < 49) {
if (diffMVdiffVal < -22) {
if (Rs < 2249) {
return 1;
}
else if (Rs >= 2249) {
return 8;
}
}
else if (diffMVdiffVal >= -22) {
return 1;
}
}
else if (diffMVdiffVal >= 49) {
return 8;
}
}
else if (AFD >= 20) {
return 8;
}
}
}
}
}
}
else if (tcor >= 93) {
if (AFD < 7) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (MVDiff < 4) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (MVDiff >= 4) {
if (Rs < 1838) {
if (MVDiff < 17) {
if (Cs < 410) {
return 4;
}
else if (Cs >= 410) {
return 2;
}
}
else if (MVDiff >= 17) {
return 1;
}
}
else if (Rs >= 1838) {
return 4;
}
}
}
}
else if (AFD >= 7) {
if (Rs < 620) {
if (Rs < 478) {
return 8;
}
else if (Rs >= 478) {
return 4;
}
}
else if (Rs >= 620) {
if (refDCval < 54) {
return 4;
}
else if (refDCval >= 54) {
if (mcTcor < 101) {
return 1;
}
else if (mcTcor >= 101) {
return 8;
}
}
}
}
}
}
else if (Cs >= 454) {
if (refDCval < 67) {
if (MVDiff < 11) {
if (posBalance < 45) {
if (Rs < 1583) {
return 8;
}
else if (Rs >= 1583) {
return 4;
}
}
else if (posBalance >= 45) {
if (negBalance < 47) {
if (negBalance < 28) {
if (Rs < 837) {
return 8;
}
else if (Rs >= 837) {
return 4;
}
}
else if (negBalance >= 28) {
return 2;
}
}
else if (negBalance >= 47) {
if (Rs < 1758) {
return 8;
}
else if (Rs >= 1758) {
return 2;
}
}
}
}
else if (MVDiff >= 11) {
if (MVDiff < 26) {
if (Rs < 1242) {
if (negBalance < 40) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (negBalance >= 40) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (refDCval < 64) {
return 4;
}
else if (refDCval >= 64) {
return 1;
}
}
}
}
else if (Rs >= 1242) {
return 1;
}
}
else if (MVDiff >= 26) {
if (MVDiff < 32) {
return 2;
}
else if (MVDiff >= 32) {
return 4;
}
}
}
}
else if (refDCval >= 67) {
if (ssDCval < 119) {
if (refDCval < 91) {
if (tcor < 85) {
if (negBalance < 60) {
if (mu_mv_mag_sq < 7) {
return 1;
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 17) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (posBalance < 70) {
return 1;
}
else if (posBalance >= 70) {
if (Rs < 1391) {
return 2;
}
else if (Rs >= 1391) {
return 4;
}
}
}
}
else if (AFD >= 17) {
if (SC < 1174) {
return 8;
}
else if (SC >= 1174) {
return 1;
}
}
}
}
else if (negBalance >= 60) {
return 2;
}
}
else if (tcor >= 85) {
if (MVDiff < 28) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (RsCsDiff < 71) {
return 1;
}
else if (RsCsDiff >= 71) {
if (Rs < 2239) {
return 1;
}
else if (Rs >= 2239) {
return 2;
}
}
}
}
else if (MVDiff >= 28) {
if (tcor < 90) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (tcor >= 90) {
return 2;
}
}
}
}
else if (refDCval >= 91) {
if (tcor < 87) {
if (RsDiff < 6) {
if (Cs < 521) {
return 8;
}
else if (Cs >= 521) {
return 4;
}
}
else if (RsDiff >= 6) {
if (Rs < 1133) {
return 1;
}
else if (Rs >= 1133) {
return 2;
}
}
}
else if (tcor >= 87) {
if (mu_mv_mag_sq < 37) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 37) {
return 8;
}
}
}
}
else if (ssDCval >= 119) {
if (Rs < 1706) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
if (refDCval < 133) {
return 2;
}
else if (refDCval >= 133) {
return 1;
}
}
}
}
else if (Rs >= 1706) {
if (Cs < 459) {
return 2;
}
else if (Cs >= 459) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 139) {
if (gchDC < 2) {
if (ssDCval < 145) {
if (mcTcor < 98) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 1;
}
}
else if (mcTcor >= 98) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (ssDCval >= 145) {
if (ssDCval < 161) {
return 1;
}
else if (ssDCval >= 161) {
if (diffAFD < 0) {
return 8;
}
else if (diffAFD >= 0) {
if (Rs < 1129) {
return 2;
}
else if (Rs >= 1129) {
return 1;
}
}
}
}
}
else if (gchDC >= 2) {
if (Rs < 848) {
if (Cs < 190) {
return 1;
}
else if (Cs >= 190) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
return 2;
}
}
}
else if (Rs >= 848) {
if (Cs < 125) {
return 2;
}
else if (Cs >= 125) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 106) {
if (Rs < 2060) {
if (negBalance < 48) {
if (Cs < 137) {
if (tcor < 77) {
return 1;
}
else if (tcor >= 77) {
if (Rs < 1361) {
return 1;
}
else if (Rs >= 1361) {
return 2;
}
}
}
else if (Cs >= 137) {
if (Cs < 478) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
return 2;
}
}
else if (Cs >= 478) {
if (ssDCval < 87) {
if (Cs < 532) {
return 2;
}
else if (Cs >= 532) {
return 1;
}
}
else if (ssDCval >= 87) {
return 1;
}
}
}
}
else if (negBalance >= 48) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (MVDiff < 45) {
if (Cs < 112) {
return 1;
}
else if (Cs >= 112) {
if (ssDCval < 90) {
if (Rs < 1222) {
return 1;
}
else if (Rs >= 1222) {
return 2;
}
}
else if (ssDCval >= 90) {
if (diffMVdiffVal < -13) {
return 1;
}
else if (diffMVdiffVal >= -13) {
if (refDCval < 104) {
return 2;
}
else if (refDCval >= 104) {
return 1;
}
}
}
}
}
else if (MVDiff >= 45) {
if (SC < 1576) {
return 2;
}
else if (SC >= 1576) {
if (Cs < 415) {
if (Rs < 2025) {
return 1;
}
else if (Rs >= 2025) {
return 8;
}
}
else if (Cs >= 415) {
if (Cs < 495) {
return 8;
}
else if (Cs >= 495) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 2060) {
if (refDCval < 81) {
if (MVDiff < 55) {
if (diffMVdiffVal < -27) {
return 4;
}
else if (diffMVdiffVal >= -27) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
if (Rs < 2349) {
return 2;
}
else if (Rs >= 2349) {
return 1;
}
}
}
}
else if (MVDiff >= 55) {
if (Cs < 182) {
return 1;
}
else if (Cs >= 182) {
return 4;
}
}
}
else if (refDCval >= 81) {
if (mcTcor < 94) {
if (Cs < 258) {
return 4;
}
else if (Cs >= 258) {
if (ssDCval < 86) {
return 1;
}
else if (ssDCval >= 86) {
if (Rs < 2405) {
return 8;
}
else if (Rs >= 2405) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (RsCsDiff < 155) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 1;
}
}
}
}
else if (RsCsDiff >= 155) {
if (diffMVdiffVal < -25) {
return 1;
}
else if (diffMVdiffVal >= -25) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 106) {
if (CsDiff < 2) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (Rs < 1039) {
return 2;
}
else if (Rs >= 1039) {
return 1;
}
}
}
}
else if (CsDiff >= 2) {
if (mcTcor < 96) {
if (refDCval < 120) {
if (SC < 2775) {
if (MVDiff < 96) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (MVDiff >= 96) {
return 8;
}
}
else if (SC >= 2775) {
return 8;
}
}
else if (refDCval >= 120) {
if (RsCsDiff < 88) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
return 1;
}
}
else if (RsCsDiff >= 88) {
if (AFD < 35) {
if (gchDC < 1) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 1;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (AFD >= 35) {
return 2;
}
}
}
}
else if (mcTcor >= 96) {
if (ssDCval < 106) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 1;
}
}
else if (ssDCval >= 106) {
if (mu_mv_mag_sq < 67) {
if (CsDiff < 3) {
if (refDCval < 109) {
return 2;
}
else if (refDCval >= 109) {
if (Cs < 423) {
return 1;
}
else if (Cs >= 423) {
if (MVDiff < 33) {
return 1;
}
else if (MVDiff >= 33) {
if (Rs < 1955) {
return 1;
}
else if (Rs >= 1955) {
return 2;
}
}
}
}
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 67) {
if (Cs < 358) {
return 1;
}
else if (Cs >= 358) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 1530) {
if (Cs < 379) {
if (refDCval < 141) {
return 1;
}
else if (refDCval >= 141) {
if (Cs < 148) {
return 2;
}
else if (Cs >= 148) {
if (Rs < 996) {
return 1;
}
else if (Rs >= 996) {
return 4;
}
}
}
}
else if (Cs >= 379) {
if (Cs < 502) {
return 8;
}
else if (Cs >= 502) {
if (Cs < 521) {
if (Rs < 1156) {
return 2;
}
else if (Rs >= 1156) {
return 1;
}
}
else if (Cs >= 521) {
return 4;
}
}
}
}
else if (Rs >= 1530) {
if (Rs < 2252) {
return 2;
}
else if (Rs >= 2252) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Cs >= 534) {
if (negBalance < 13) {
if (Rs < 699) {
if (ssDCval < 58) {
if (mcTcor < 97) {
if (tcor < 81) {
if (MVDiff < 73) {
return 1;
}
else if (MVDiff >= 73) {
return 2;
}
}
else if (tcor >= 81) {
return 8;
}
}
else if (mcTcor >= 97) {
if (Rs < 261) {
if (Cs < 1035) {
return 2;
}
else if (Cs >= 1035) {
return 1;
}
}
else if (Rs >= 261) {
if (ssDCval < 30) {
return 8;
}
else if (ssDCval >= 30) {
return 1;
}
}
}
}
else if (ssDCval >= 58) {
if (refDCval < 85) {
if (refDCval < 75) {
if (SC < 1168) {
return 4;
}
else if (SC >= 1168) {
if (Rs < 287) {
return 4;
}
else if (Rs >= 287) {
return 8;
}
}
}
else if (refDCval >= 75) {
if (Rs < 535) {
return 1;
}
else if (Rs >= 535) {
return 2;
}
}
}
else if (refDCval >= 85) {
if (TSC < 1) {
return 8;
}
else if (TSC >= 1) {
if (posBalance < 27) {
if (Rs < 646) {
return 8;
}
else if (Rs >= 646) {
return 1;
}
}
else if (posBalance >= 27) {
if (SC < 1070) {
if (Cs < 558) {
return 2;
}
else if (Cs >= 558) {
return 1;
}
}
else if (SC >= 1070) {
if (ssDCval < 91) {
return 1;
}
else if (ssDCval >= 91) {
if (diffMVdiffVal < -1) {
if (Rs < 563) {
return 1;
}
else if (Rs >= 563) {
if (Rs < 627) {
return 8;
}
else if (Rs >= 627) {
return 2;
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 699) {
if (diffAFD < -3) {
if (Rs < 1511) {
if (tcor < 99) {
if (diffAFD < -19) {
if (posBalance < 6) {
return 4;
}
else if (posBalance >= 6) {
if (Rs < 1210) {
return 8;
}
else if (Rs >= 1210) {
return 1;
}
}
}
else if (diffAFD >= -19) {
if (mcTcor < 99) {
if (mcTcor < 77) {
return 1;
}
else if (mcTcor >= 77) {
return 8;
}
}
else if (mcTcor >= 99) {
if (Rs < 1273) {
return 8;
}
else if (Rs >= 1273) {
return 1;
}
}
}
}
else if (tcor >= 99) {
if (diffAFD < -7) {
if (ssDCval < 133) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
if (Cs < 571) {
return 2;
}
else if (Cs >= 571) {
if (Rs < 1396) {
if (Rs < 843) {
return 8;
}
else if (Rs >= 843) {
if (ssDCval < 131) {
return 1;
}
else if (ssDCval >= 131) {
if (Cs < 1018) {
return 8;
}
else if (Cs >= 1018) {
return 1;
}
}
}
}
else if (Rs >= 1396) {
return 8;
}
}
}
}
else if (ssDCval >= 133) {
if (ssDCval < 141) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
if (Cs < 1196) {
return 4;
}
else if (Cs >= 1196) {
return 1;
}
}
}
else if (ssDCval >= 141) {
if (diffMVdiffVal < -2) {
if (Cs < 942) {
if (posBalance < 13) {
return 8;
}
else if (posBalance >= 13) {
return 4;
}
}
else if (Cs >= 942) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
}
else if (diffAFD >= -7) {
if (ssDCval < 69) {
if (Rs < 1280) {
return 8;
}
else if (Rs >= 1280) {
return 1;
}
}
else if (ssDCval >= 69) {
if (Cs < 1016) {
return 4;
}
else if (Cs >= 1016) {
if (Rs < 1123) {
return 8;
}
else if (Rs >= 1123) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1511) {
if (MVDiff < 16) {
if (Cs < 1290) {
return 8;
}
else if (Cs >= 1290) {
return 1;
}
}
else if (MVDiff >= 16) {
if (Cs < 1034) {
return 8;
}
else if (Cs >= 1034) {
if (Rs < 1559) {
return 8;
}
else if (Rs >= 1559) {
if (ssDCval < 144) {
if (Cs < 1310) {
return 4;
}
else if (Cs >= 1310) {
return 1;
}
}
else if (ssDCval >= 144) {
return 8;
}
}
}
}
}
}
else if (diffAFD >= -3) {
if (Rs < 1844) {
if (Rs < 965) {
if (posBalance < 13) {
if (ssDCval < 118) {
return 8;
}
else if (ssDCval >= 118) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (posBalance >= 13) {
if (MVDiff < 2) {
if (negBalance < 11) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (Rs < 713) {
return 1;
}
else if (Rs >= 713) {
return 8;
}
}
}
else if (negBalance >= 11) {
if (posBalance < 29) {
if (Rs < 766) {
if (Cs < 599) {
return 8;
}
else if (Cs >= 599) {
return 1;
}
}
else if (Rs >= 766) {
return 8;
}
}
else if (posBalance >= 29) {
if (ssDCval < 73) {
if (ssDCval < 62) {
if (Cs < 971) {
return 8;
}
else if (Cs >= 971) {
return 1;
}
}
else if (ssDCval >= 62) {
return 2;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (posBalance < 29) {
if (Rs < 845) {
if (ssDCval < 36) {
if (Rs < 735) {
return 1;
}
else if (Rs >= 735) {
return 8;
}
}
else if (ssDCval >= 36) {
return 2;
}
}
else if (Rs >= 845) {
return 8;
}
}
else if (posBalance >= 29) {
if (CsDiff < 3) {
if (Rs < 703) {
return 1;
}
else if (Rs >= 703) {
return 8;
}
}
else if (CsDiff >= 3) {
if (Cs < 1112) {
return 1;
}
else if (Cs >= 1112) {
return 8;
}
}
}
}
}
}
else if (Rs >= 965) {
if (MVDiff < 3) {
if (ssDCval < 53) {
if (Cs < 550) {
return 1;
}
else if (Cs >= 550) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
}
else if (ssDCval >= 53) {
if (SC < 2841) {
if (ssDCval < 64) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (Rs < 1268) {
return 1;
}
else if (Rs >= 1268) {
if (SC < 2385) {
return 8;
}
else if (SC >= 2385) {
return 1;
}
}
}
}
else if (ssDCval >= 64) {
if (negBalance < 1) {
if (Cs < 1137) {
if (diffMVdiffVal < -4) {
if (Rs < 1110) {
return 8;
}
else if (Rs >= 1110) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (Cs >= 1137) {
return 1;
}
}
else if (negBalance >= 1) {
if (tcor < 101) {
if (ssDCval < 152) {
if (Rs < 1686) {
if (refDCval < 66) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (refDCval >= 66) {
return 8;
}
}
else if (Rs >= 1686) {
if (Rs < 1690) {
return 4;
}
else if (Rs >= 1690) {
return 8;
}
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (tcor >= 101) {
if (Rs < 1316) {
return 4;
}
else if (Rs >= 1316) {
return 8;
}
}
}
}
}
else if (SC >= 2841) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
if (Rs < 1740) {
return 8;
}
else if (Rs >= 1740) {
return 4;
}
}
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (Rs < 1140) {
if (tcor < 76) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Rs < 1020) {
return 4;
}
else if (Rs >= 1020) {
return 1;
}
}
}
else if (tcor >= 76) {
return 8;
}
}
else if (Rs >= 1140) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (SC < 2202) {
return 4;
}
else if (SC >= 2202) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1844) {
if (Rs < 2038) {
if (negBalance < 5) {
return 8;
}
else if (negBalance >= 5) {
if (SC < 2885) {
if (mcTcor < 98) {
if (Rs < 1898) {
return 8;
}
else if (Rs >= 1898) {
return 2;
}
}
else if (mcTcor >= 98) {
if (Rs < 1848) {
return 4;
}
else if (Rs >= 1848) {
return 1;
}
}
}
else if (SC >= 2885) {
if (refDCval < 69) {
if (Cs < 940) {
return 1;
}
else if (Cs >= 940) {
return 4;
}
}
else if (refDCval >= 69) {
if (Cs < 1066) {
return 8;
}
else if (Cs >= 1066) {
return 1;
}
}
}
}
}
else if (Rs >= 2038) {
if (Cs < 1289) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (Rs < 2143) {
return 1;
}
else if (Rs >= 2143) {
return 8;
}
}
}
else if (Cs >= 1289) {
if (CsDiff < 1) {
if (Rs < 2202) {
return 8;
}
else if (Rs >= 2202) {
return 2;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (RsDiff < 7) {
if (AFD < 14) {
if (posBalance < 43) {
if (refDCval < 47) {
if (negBalance < 37) {
if (TSCindex < 3) {
if (ssDCval < 38) {
return 8;
}
else if (ssDCval >= 38) {
if (Cs < 1003) {
return 4;
}
else if (Cs >= 1003) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (diffMVdiffVal < 4) {
if (RsCsDiff < 49) {
if (Rs < 633) {
if (Cs < 708) {
return 4;
}
else if (Cs >= 708) {
return 1;
}
}
else if (Rs >= 633) {
if (negBalance < 32) {
if (AFD < 3) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (AFD >= 3) {
if (AFD < 10) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Cs < 1245) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (Rs < 777) {
return 4;
}
else if (Rs >= 777) {
return 1;
}
}
}
else if (Cs >= 1245) {
return 8;
}
}
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (negBalance >= 32) {
if (Rs < 1277) {
return 8;
}
else if (Rs >= 1277) {
if (Cs < 1192) {
return 4;
}
else if (Cs >= 1192) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 49) {
if (Rs < 1705) {
return 8;
}
else if (Rs >= 1705) {
return 4;
}
}
}
else if (diffMVdiffVal >= 4) {
if (TSC < 4) {
if (ssDCval < 28) {
return 4;
}
else if (ssDCval >= 28) {
return 1;
}
}
else if (TSC >= 4) {
if (Rs < 820) {
return 8;
}
else if (Rs >= 820) {
return 2;
}
}
}
}
}
else if (negBalance >= 37) {
if (diffRsCsdiff < -60) {
return 1;
}
else if (diffRsCsdiff >= -60) {
return 8;
}
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 72) {
if (negBalance < 63) {
if (tcor < 89) {
if (diffRsCsdiff < -23) {
if (Rs < 619) {
return 2;
}
else if (Rs >= 619) {
return 8;
}
}
else if (diffRsCsdiff >= -23) {
if (Cs < 717) {
if (CsDiff < 3) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 2;
}
}
else if (CsDiff >= 3) {
if (negBalance < 35) {
if (Rs < 1628) {
return 2;
}
else if (Rs >= 1628) {
return 1;
}
}
else if (negBalance >= 35) {
return 1;
}
}
}
else if (Cs >= 717) {
if (tcor < 75) {
if (Rs < 1239) {
return 1;
}
else if (Rs >= 1239) {
return 8;
}
}
else if (tcor >= 75) {
if (ssDCval < 82) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
if (Rs < 666) {
return 2;
}
else if (Rs >= 666) {
return 8;
}
}
}
else if (ssDCval >= 82) {
if (Rs < 1918) {
return 2;
}
else if (Rs >= 1918) {
return 4;
}
}
}
}
}
}
else if (tcor >= 89) {
if (Cs < 974) {
if (Rs < 739) {
if (negBalance < 36) {
return 4;
}
else if (negBalance >= 36) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (Rs >= 739) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Cs < 572) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (Cs >= 572) {
if (negBalance < 27) {
if (Rs < 1551) {
return 1;
}
else if (Rs >= 1551) {
return 2;
}
}
else if (negBalance >= 27) {
return 1;
}
}
}
}
}
else if (Cs >= 974) {
if (RsCsDiff < 17) {
if (Rs < 1179) {
return 2;
}
else if (Rs >= 1179) {
return 8;
}
}
else if (RsCsDiff >= 17) {
if (tcor < 92) {
if (Rs < 875) {
return 4;
}
else if (Rs >= 875) {
return 8;
}
}
else if (tcor >= 92) {
return 4;
}
}
}
}
}
else if (negBalance >= 63) {
if (gchDC < 2) {
if (Rs < 425) {
return 1;
}
else if (Rs >= 425) {
return 4;
}
}
else if (gchDC >= 2) {
if (ssDCval < 120) {
if (ssDCval < 46) {
if (Rs < 759) {
return 4;
}
else if (Rs >= 759) {
return 8;
}
}
else if (ssDCval >= 46) {
return 8;
}
}
else if (ssDCval >= 120) {
if (MVDiff < 40) {
return 2;
}
else if (MVDiff >= 40) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (refDCval < 56) {
if (TSC < 3) {
if (Rs < 613) {
if (Rs < 442) {
return 2;
}
else if (Rs >= 442) {
return 1;
}
}
else if (Rs >= 613) {
if (negBalance < 30) {
if (posBalance < 36) {
return 8;
}
else if (posBalance >= 36) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (negBalance >= 30) {
return 4;
}
}
}
else if (TSC >= 3) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (ssDCval < 49) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
else if (ssDCval >= 49) {
if (negBalance < 39) {
if (SC < 2844) {
return 1;
}
else if (SC >= 2844) {
return 8;
}
}
else if (negBalance >= 39) {
return 4;
}
}
}
}
}
else if (refDCval >= 56) {
if (mu_mv_mag_sq < 101) {
if (AFD < 5) {
if (ssDCval < 139) {
if (SC < 3076) {
if (mu_mv_mag_sq < 89) {
return 4;
}
else if (mu_mv_mag_sq >= 89) {
if (TSC < 4) {
if (Cs < 729) {
return 8;
}
else if (Cs >= 729) {
if (ssDCval < 57) {
if (Rs < 1147) {
return 4;
}
else if (Rs >= 1147) {
return 1;
}
}
else if (ssDCval >= 57) {
return 4;
}
}
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (SC >= 3076) {
if (SC < 3181) {
if (Rs < 2024) {
if (Rs < 1869) {
return 2;
}
else if (Rs >= 1869) {
return 1;
}
}
else if (Rs >= 2024) {
return 4;
}
}
else if (SC >= 3181) {
return 8;
}
}
}
else if (ssDCval >= 139) {
return 8;
}
}
else if (AFD >= 5) {
if (ssDCval < 97) {
if (negBalance < 34) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 2;
}
}
else if (negBalance >= 34) {
if (MVDiff < 6) {
if (Rs < 1874) {
if (Rs < 1214) {
return 2;
}
else if (Rs >= 1214) {
return 1;
}
}
else if (Rs >= 1874) {
if (Rs < 2355) {
return 4;
}
else if (Rs >= 2355) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 75) {
if (Rs < 982) {
return 8;
}
else if (Rs >= 982) {
if (Rs < 1910) {
return 4;
}
else if (Rs >= 1910) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 75) {
return 8;
}
}
}
}
else if (ssDCval >= 97) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (Rs < 1853) {
return 8;
}
else if (Rs >= 1853) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (RsCsDiff < 43) {
if (diffRsCsdiff < 6) {
if (ssDCval < 64) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (ssDCval >= 64) {
if (refDCval < 93) {
if (RsDiff < 3) {
if (Rs < 636) {
return 2;
}
else if (Rs >= 636) {
return 4;
}
}
else if (RsDiff >= 3) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
if (Rs < 1628) {
return 4;
}
else if (Rs >= 1628) {
if (Rs < 2008) {
return 8;
}
else if (Rs >= 2008) {
return 4;
}
}
}
}
}
else if (refDCval >= 93) {
return 8;
}
}
}
else if (diffRsCsdiff >= 6) {
if (mu_mv_mag_sq < 121) {
if (Rs < 1749) {
return 8;
}
else if (Rs >= 1749) {
return 1;
}
}
else if (mu_mv_mag_sq >= 121) {
if (Rs < 1574) {
return 1;
}
else if (Rs >= 1574) {
return 4;
}
}
}
}
else if (RsCsDiff >= 43) {
if (Rs < 1535) {
return 1;
}
else if (Rs >= 1535) {
return 2;
}
}
}
}
}
}
}
else if (posBalance >= 43) {
if (ssDCval < 156) {
if (Rs < 405) {
if (ssDCval < 96) {
if (ssDCval < 60) {
if (RsDiff < 3) {
if (gchDC < 3) {
if (posBalance < 48) {
if (Rs < 357) {
return 2;
}
else if (Rs >= 357) {
if (Cs < 860) {
return 4;
}
else if (Cs >= 860) {
return 1;
}
}
}
else if (posBalance >= 48) {
if (mcTcor < 96) {
if (Cs < 612) {
if (Rs < 344) {
return 1;
}
else if (Rs >= 344) {
return 2;
}
}
else if (Cs >= 612) {
return 1;
}
}
else if (mcTcor >= 96) {
if (Rs < 362) {
if (Rs < 303) {
return 1;
}
else if (Rs >= 303) {
return 2;
}
}
else if (Rs >= 362) {
return 1;
}
}
}
}
else if (gchDC >= 3) {
return 2;
}
}
else if (RsDiff >= 3) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 4;
}
}
}
}
else if (ssDCval >= 60) {
if (negBalance < 42) {
if (Cs < 672) {
if (Rs < 241) {
return 1;
}
else if (Rs >= 241) {
return 4;
}
}
else if (Cs >= 672) {
return 8;
}
}
else if (negBalance >= 42) {
if (TSC < 7) {
if (diffMVdiffVal < -12) {
return 8;
}
else if (diffMVdiffVal >= -12) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
else if (ssDCval >= 96) {
if (Cs < 719) {
if (Rs < 372) {
if (ssDCval < 133) {
return 2;
}
else if (ssDCval >= 133) {
return 8;
}
}
else if (Rs >= 372) {
return 1;
}
}
else if (Cs >= 719) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (Rs >= 405) {
if (mu_mv_mag_sq < 32) {
if (negBalance < 45) {
if (refDCval < 46) {
if (Rs < 846) {
if (negBalance < 37) {
if (Cs < 1027) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 503) {
return 8;
}
else if (Rs >= 503) {
if (Rs < 621) {
return 2;
}
else if (Rs >= 621) {
return 4;
}
}
}
}
else if (Cs >= 1027) {
if (Cs < 1111) {
return 2;
}
else if (Cs >= 1111) {
return 1;
}
}
}
else if (negBalance >= 37) {
if (gchDC < 1) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
else if (Rs >= 846) {
if (mu_mv_mag_sq < 5) {
if (Cs < 697) {
if (SC < 1643) {
return 8;
}
else if (SC >= 1643) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 2;
}
}
}
else if (Cs >= 697) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 1393) {
return 2;
}
else if (Rs >= 1393) {
return 8;
}
}
}
}
else if (refDCval >= 46) {
if (SC < 2540) {
if (Rs < 797) {
if (tcor < 92) {
if (SC < 1406) {
if (posBalance < 55) {
if (Rs < 767) {
return 2;
}
else if (Rs >= 767) {
if (Rs < 780) {
return 8;
}
else if (Rs >= 780) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (diffRsCsdiff < -16) {
if (Rs < 555) {
return 4;
}
else if (Rs >= 555) {
return 1;
}
}
else if (diffRsCsdiff >= -16) {
if (SC < 1251) {
return 8;
}
else if (SC >= 1251) {
if (Cs < 713) {
return 2;
}
else if (Cs >= 713) {
if (Rs < 582) {
return 8;
}
else if (Rs >= 582) {
return 4;
}
}
}
}
}
}
else if (SC >= 1406) {
if (diffRsCsdiff < 4) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (CsDiff < 7) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 559) {
return 2;
}
else if (Rs >= 559) {
if (Rs < 692) {
return 8;
}
else if (Rs >= 692) {
return 4;
}
}
}
}
else if (CsDiff >= 7) {
return 4;
}
}
}
else if (diffRsCsdiff >= 4) {
if (tcor < 87) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (tcor >= 87) {
if (Rs < 628) {
return 4;
}
else if (Rs >= 628) {
return 2;
}
}
}
}
}
else if (tcor >= 92) {
if (ssDCval < 73) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
if (Rs < 682) {
return 2;
}
else if (Rs >= 682) {
return 8;
}
}
}
else if (ssDCval >= 73) {
if (Rs < 579) {
return 8;
}
else if (Rs >= 579) {
return 1;
}
}
}
}
else if (Rs >= 797) {
if (AFD < 9) {
if (refDCval < 64) {
if (RsCsDiff < 16) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 4;
}
}
else if (RsCsDiff >= 16) {
if (tcor < 90) {
if (MVDiff < 7) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 4;
}
}
else if (MVDiff >= 7) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
else if (tcor >= 90) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1788) {
return 4;
}
else if (Rs >= 1788) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
return 1;
}
}
}
}
else if (refDCval >= 64) {
if (Rs < 1124) {
if (ssDCval < 125) {
if (Cs < 618) {
if (Rs < 934) {
if (Rs < 866) {
return 8;
}
else if (Rs >= 866) {
return 4;
}
}
else if (Rs >= 934) {
return 1;
}
}
else if (Cs >= 618) {
return 8;
}
}
else if (ssDCval >= 125) {
if (Rs < 973) {
return 1;
}
else if (Rs >= 973) {
return 4;
}
}
}
else if (Rs >= 1124) {
if (ssDCval < 94) {
return 1;
}
else if (ssDCval >= 94) {
if (negBalance < 39) {
if (diffAFD < 1) {
if (Rs < 1390) {
return 4;
}
else if (Rs >= 1390) {
return 8;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (negBalance >= 39) {
return 8;
}
}
}
}
}
else if (AFD >= 9) {
if (posBalance < 69) {
if (negBalance < 36) {
if (CsDiff < 7) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (CsDiff >= 7) {
return 8;
}
}
else if (negBalance >= 36) {
if (tcor < 80) {
if (Cs < 571) {
return 2;
}
else if (Cs >= 571) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (MVDiff < 38) {
if (refDCval < 51) {
return 8;
}
else if (refDCval >= 51) {
return 1;
}
}
else if (MVDiff >= 38) {
return 8;
}
}
}
}
else if (tcor >= 80) {
if (ssDCval < 63) {
if (RsCsDiff < 22) {
return 8;
}
else if (RsCsDiff >= 22) {
if (diffRsCsdiff < -37) {
if (Rs < 1124) {
return 2;
}
else if (Rs >= 1124) {
return 8;
}
}
else if (diffRsCsdiff >= -37) {
return 4;
}
}
}
else if (ssDCval >= 63) {
if (ssDCval < 108) {
if (CsDiff < 5) {
if (Cs < 647) {
if (tcor < 89) {
if (Rs < 852) {
return 2;
}
else if (Rs >= 852) {
return 1;
}
}
else if (tcor >= 89) {
if (Rs < 1420) {
return 4;
}
else if (Rs >= 1420) {
return 2;
}
}
}
else if (Cs >= 647) {
if (SC < 2469) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 1026) {
return 2;
}
else if (Rs >= 1026) {
return 4;
}
}
}
else if (SC >= 2469) {
if (Rs < 1815) {
return 1;
}
else if (Rs >= 1815) {
return 4;
}
}
}
}
else if (CsDiff >= 5) {
if (Rs < 1257) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
return 8;
}
}
else if (Rs >= 1257) {
return 1;
}
}
}
else if (ssDCval >= 108) {
if (Rs < 1057) {
return 2;
}
else if (Rs >= 1057) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 69) {
if (Cs < 788) {
return 2;
}
else if (Cs >= 788) {
return 8;
}
}
}
}
}
else if (SC >= 2540) {
if (Cs < 865) {
if (AFD < 10) {
if (Cs < 567) {
return 1;
}
else if (Cs >= 567) {
return 2;
}
}
else if (AFD >= 10) {
if (ssDCval < 73) {
return 4;
}
else if (ssDCval >= 73) {
return 1;
}
}
}
else if (Cs >= 865) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (MVDiff < 11) {
if (RsDiff < 6) {
if (diffRsCsdiff < -18) {
return 2;
}
else if (diffRsCsdiff >= -18) {
if (posBalance < 52) {
if (Rs < 2174) {
return 1;
}
else if (Rs >= 2174) {
return 4;
}
}
else if (posBalance >= 52) {
return 8;
}
}
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 959) {
if (CsDiff < 5) {
if (AFD < 11) {
if (MVDiff < 7) {
if (Cs < 654) {
if (AFD < 9) {
if (refDCval < 110) {
if (posBalance < 55) {
if (Rs < 566) {
return 8;
}
else if (Rs >= 566) {
if (Rs < 657) {
return 1;
}
else if (Rs >= 657) {
return 2;
}
}
}
else if (posBalance >= 55) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
}
else if (refDCval >= 110) {
if (Cs < 587) {
return 2;
}
else if (Cs >= 587) {
return 8;
}
}
}
else if (AFD >= 9) {
if (Rs < 729) {
if (Rs < 578) {
return 8;
}
else if (Rs >= 578) {
return 4;
}
}
else if (Rs >= 729) {
if (mcTcor < 86) {
return 1;
}
else if (mcTcor >= 86) {
return 8;
}
}
}
}
else if (Cs >= 654) {
if (tcor < 69) {
if (posBalance < 60) {
if (Cs < 719) {
return 1;
}
else if (Cs >= 719) {
return 8;
}
}
else if (posBalance >= 60) {
return 1;
}
}
else if (tcor >= 69) {
if (refDCval < 54) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
return 4;
}
}
else if (refDCval >= 54) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (Cs < 874) {
return 1;
}
else if (Cs >= 874) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (SC < 1477) {
if (diffMVdiffVal < 8) {
if (ssDCval < 49) {
if (mcTcor < 82) {
if (TSC < 6) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 683) {
return 2;
}
else if (Rs >= 683) {
return 1;
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (mcTcor >= 82) {
if (Cs < 571) {
if (Rs < 639) {
return 2;
}
else if (Rs >= 639) {
return 8;
}
}
else if (Cs >= 571) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 4;
}
}
}
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (diffMVdiffVal >= 8) {
if (Rs < 583) {
return 8;
}
else if (Rs >= 583) {
return 2;
}
}
}
else if (SC >= 1477) {
if (ssDCval < 111) {
if (Cs < 766) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (mu_mv_mag_sq < 2) {
if (tcor < 63) {
return 8;
}
else if (tcor >= 63) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
}
else if (Cs >= 766) {
if (RsCsDiff < 11) {
return 1;
}
else if (RsCsDiff >= 11) {
return 8;
}
}
}
else if (ssDCval >= 111) {
return 4;
}
}
}
}
else if (AFD >= 11) {
if (TSC < 6) {
if (Cs < 551) {
if (Rs < 700) {
return 1;
}
else if (Rs >= 700) {
return 4;
}
}
else if (Cs >= 551) {
if (tcor < 77) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 2;
}
}
else if (tcor >= 77) {
if (ssDCval < 66) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (ssDCval >= 66) {
if (negBalance < 61) {
if (Rs < 861) {
return 1;
}
else if (Rs >= 861) {
return 4;
}
}
else if (negBalance >= 61) {
return 2;
}
}
}
}
}
else if (TSC >= 6) {
if (Cs < 627) {
if (mcTcor < 77) {
if (Cs < 612) {
return 4;
}
else if (Cs >= 612) {
return 8;
}
}
else if (mcTcor >= 77) {
return 8;
}
}
else if (Cs >= 627) {
if (negBalance < 55) {
if (diffMVdiffVal < 0) {
if (Cs < 801) {
if (tcor < 72) {
return 4;
}
else if (tcor >= 72) {
if (SC < 1602) {
return 8;
}
else if (SC >= 1602) {
return 4;
}
}
}
else if (Cs >= 801) {
return 8;
}
}
else if (diffMVdiffVal >= 0) {
if (Cs < 701) {
return 1;
}
else if (Cs >= 701) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
return 4;
}
}
}
}
else if (negBalance >= 55) {
if (tcor < 80) {
if (Rs < 855) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 8;
}
}
else if (Rs >= 855) {
return 8;
}
}
else if (tcor >= 80) {
if (ssDCval < 103) {
if (Rs < 594) {
return 8;
}
else if (Rs >= 594) {
return 2;
}
}
else if (ssDCval >= 103) {
return 4;
}
}
}
}
}
}
}
else if (CsDiff >= 5) {
if (negBalance < 50) {
if (tcor < 73) {
if (MVDiff < 28) {
if (Rs < 527) {
return 4;
}
else if (Rs >= 527) {
return 2;
}
}
else if (MVDiff >= 28) {
if (Cs < 1022) {
return 4;
}
else if (Cs >= 1022) {
return 8;
}
}
}
else if (tcor >= 73) {
if (TSC < 7) {
if (Rs < 943) {
return 8;
}
else if (Rs >= 943) {
return 2;
}
}
else if (TSC >= 7) {
return 1;
}
}
}
else if (negBalance >= 50) {
if (mcTcor < 87) {
if (Cs < 827) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (Rs < 578) {
return 1;
}
else if (Rs >= 578) {
return 8;
}
}
}
else if (Cs >= 827) {
return 4;
}
}
else if (mcTcor >= 87) {
if (Cs < 1282) {
if (negBalance < 53) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (Rs < 810) {
return 1;
}
else if (Rs >= 810) {
return 8;
}
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (Cs >= 1282) {
return 2;
}
}
}
}
}
else if (Rs >= 959) {
if (MVDiff < 5) {
if (AFD < 13) {
if (diffMVdiffVal < -9) {
return 4;
}
else if (diffMVdiffVal >= -9) {
if (Rs < 1550) {
if (Rs < 988) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 979) {
return 1;
}
else if (Rs >= 979) {
return 4;
}
}
}
else if (Rs >= 988) {
return 8;
}
}
else if (Rs >= 1550) {
if (Cs < 748) {
if (Cs < 594) {
return 2;
}
else if (Cs >= 594) {
return 1;
}
}
else if (Cs >= 748) {
if (diffRsCsdiff < 28) {
if (TSC < 10) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1733) {
return 8;
}
else if (Rs >= 1733) {
return 1;
}
}
}
else if (TSC >= 10) {
return 1;
}
}
else if (diffRsCsdiff >= 28) {
if (Rs < 1924) {
return 2;
}
else if (Rs >= 1924) {
return 1;
}
}
}
}
}
}
else if (AFD >= 13) {
if (Rs < 1388) {
return 4;
}
else if (Rs >= 1388) {
if (negBalance < 52) {
if (Rs < 1929) {
return 4;
}
else if (Rs >= 1929) {
return 8;
}
}
else if (negBalance >= 52) {
return 8;
}
}
}
}
else if (MVDiff >= 5) {
if (tcor < 80) {
if (MVDiff < 12) {
if (negBalance < 46) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (negBalance >= 46) {
if (mcTcor < 93) {
if (AFD < 9) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (Rs < 1421) {
return 2;
}
else if (Rs >= 1421) {
return 8;
}
}
}
else if (AFD >= 9) {
return 8;
}
}
else if (mcTcor >= 93) {
if (Rs < 1488) {
return 2;
}
else if (Rs >= 1488) {
return 4;
}
}
}
}
else if (MVDiff >= 12) {
if (ssDCval < 120) {
if (CsDiff < 6) {
if (mcTcor < 91) {
if (Rs < 1347) {
if (RsCsDiff < 57) {
return 8;
}
else if (RsCsDiff >= 57) {
return 4;
}
}
else if (Rs >= 1347) {
if (TSC < 8) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
if (diffMVdiffVal < -12) {
return 8;
}
else if (diffMVdiffVal >= -12) {
if (diffMVdiffVal < 4) {
return 2;
}
else if (diffMVdiffVal >= 4) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
return 8;
}
}
}
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (mcTcor >= 91) {
if (Cs < 833) {
return 2;
}
else if (Cs >= 833) {
return 4;
}
}
}
else if (CsDiff >= 6) {
if (negBalance < 50) {
if (Rs < 1175) {
return 2;
}
else if (Rs >= 1175) {
return 1;
}
}
else if (negBalance >= 50) {
return 4;
}
}
}
else if (ssDCval >= 120) {
if (Rs < 1774) {
return 1;
}
else if (Rs >= 1774) {
return 8;
}
}
}
}
else if (tcor >= 80) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 54) {
if (ssDCval < 60) {
if (Cs < 576) {
return 1;
}
else if (Cs >= 576) {
if (diffAFD < -9) {
return 4;
}
else if (diffAFD >= -9) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 4;
}
}
}
}
else if (ssDCval >= 60) {
if (negBalance < 51) {
if (RsCsDiff < 29) {
return 1;
}
else if (RsCsDiff >= 29) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (negBalance >= 51) {
if (AFD < 9) {
if (Rs < 1644) {
return 4;
}
else if (Rs >= 1644) {
return 2;
}
}
else if (AFD >= 9) {
if (Rs < 1947) {
return 1;
}
else if (Rs >= 1947) {
return 4;
}
}
}
}
}
else if (posBalance >= 54) {
if (MVDiff < 27) {
if (tcor < 88) {
if (RsCsDiff < 51) {
if (Cs < 647) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (Cs >= 647) {
if (ssDCval < 111) {
if (negBalance < 48) {
if (Cs < 1121) {
if (Cs < 772) {
if (Rs < 1425) {
return 8;
}
else if (Rs >= 1425) {
return 4;
}
}
else if (Cs >= 772) {
return 8;
}
}
else if (Cs >= 1121) {
if (Rs < 1054) {
return 4;
}
else if (Rs >= 1054) {
return 2;
}
}
}
else if (negBalance >= 48) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (ssDCval >= 111) {
if (Rs < 1206) {
return 4;
}
else if (Rs >= 1206) {
return 8;
}
}
}
}
else if (RsCsDiff >= 51) {
if (SC < 2246) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 1;
}
}
else if (SC >= 2246) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (Rs < 1532) {
return 4;
}
else if (Rs >= 1532) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
}
else if (tcor >= 88) {
if (RsCsDiff < 23) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
if (Rs < 1579) {
return 8;
}
else if (Rs >= 1579) {
return 2;
}
}
}
else if (RsCsDiff >= 23) {
if (negBalance < 51) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (negBalance >= 51) {
if (SC < 2503) {
if (diffMVdiffVal < 5) {
if (RsCsDiff < 35) {
return 8;
}
else if (RsCsDiff >= 35) {
return 1;
}
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
else if (SC >= 2503) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1670) {
return 2;
}
else if (Rs >= 1670) {
if (Rs < 1949) {
return 1;
}
else if (Rs >= 1949) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 27) {
if (Cs < 628) {
return 1;
}
else if (Cs >= 628) {
if (Rs < 1670) {
return 4;
}
else if (Rs >= 1670) {
if (Rs < 2171) {
return 2;
}
else if (Rs >= 2171) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (refDCval < 138) {
if (diffRsCsdiff < -26) {
if (Rs < 1894) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Rs < 1357) {
return 8;
}
else if (Rs >= 1357) {
return 4;
}
}
}
else if (Rs >= 1894) {
return 2;
}
}
else if (diffRsCsdiff >= -26) {
if (diffRsCsdiff < 24) {
if (Rs < 1594) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
else if (Rs >= 1594) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 8;
}
}
}
else if (diffRsCsdiff >= 24) {
return 4;
}
}
}
else if (refDCval >= 138) {
if (Rs < 1034) {
return 1;
}
else if (Rs >= 1034) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (negBalance < 42) {
if (ssDCval < 78) {
if (ssDCval < 46) {
if (mu_mv_mag_sq < 117) {
if (Rs < 569) {
if (RsDiff < 3) {
if (Cs < 655) {
return 2;
}
else if (Cs >= 655) {
return 4;
}
}
else if (RsDiff >= 3) {
if (Cs < 539) {
return 8;
}
else if (Cs >= 539) {
return 1;
}
}
}
else if (Rs >= 569) {
if (posBalance < 49) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (ssDCval < 39) {
if (Rs < 827) {
return 1;
}
else if (Rs >= 827) {
return 8;
}
}
else if (ssDCval >= 39) {
return 1;
}
}
}
else if (posBalance >= 49) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (RsCsDiff < 22) {
return 8;
}
else if (RsCsDiff >= 22) {
if (Cs < 708) {
return 8;
}
else if (Cs >= 708) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (RsCsDiff < 16) {
if (Rs < 860) {
return 1;
}
else if (Rs >= 860) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 1490) {
return 1;
}
else if (Rs >= 1490) {
return 8;
}
}
}
}
}
else if (ssDCval >= 46) {
if (RsCsDiff < 19) {
if (Rs < 869) {
if (negBalance < 33) {
if (refDCval < 48) {
return 1;
}
else if (refDCval >= 48) {
if (mu_mv_mag_sq < 66) {
if (diffRsCsdiff < 7) {
if (Rs < 808) {
return 4;
}
else if (Rs >= 808) {
return 1;
}
}
else if (diffRsCsdiff >= 7) {
return 1;
}
}
else if (mu_mv_mag_sq >= 66) {
if (mu_mv_mag_sq < 99) {
if (diffRsCsdiff < 2) {
if (ssDCval < 65) {
if (Cs < 552) {
if (Rs < 720) {
return 1;
}
else if (Rs >= 720) {
return 4;
}
}
else if (Cs >= 552) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (SC < 1330) {
return 8;
}
else if (SC >= 1330) {
return 1;
}
}
}
}
else if (ssDCval >= 65) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (diffRsCsdiff >= 2) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (Rs < 558) {
return 2;
}
else if (Rs >= 558) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (posBalance < 52) {
if (diffTSC < 1) {
if (Cs < 578) {
if (Rs < 649) {
return 1;
}
else if (Rs >= 649) {
return 8;
}
}
else if (Cs >= 578) {
if (Rs < 561) {
return 8;
}
else if (Rs >= 561) {
return 4;
}
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (posBalance >= 52) {
return 4;
}
}
}
}
}
else if (negBalance >= 33) {
if (Rs < 798) {
if (RsDiff < 3) {
if (Cs < 723) {
if (posBalance < 45) {
if (Rs < 541) {
return 8;
}
else if (Rs >= 541) {
if (Rs < 609) {
return 2;
}
else if (Rs >= 609) {
return 1;
}
}
}
else if (posBalance >= 45) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
return 1;
}
}
}
else if (Cs >= 723) {
if (ssDCval < 68) {
if (Rs < 526) {
return 1;
}
else if (Rs >= 526) {
if (AFD < 7) {
if (Rs < 766) {
if (Rs < 726) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (Rs >= 726) {
return 4;
}
}
else if (Rs >= 766) {
return 8;
}
}
else if (AFD >= 7) {
if (Rs < 691) {
return 8;
}
else if (Rs >= 691) {
return 2;
}
}
}
}
else if (ssDCval >= 68) {
return 1;
}
}
}
else if (RsDiff >= 3) {
if (gchDC < 1) {
if (Rs < 684) {
return 4;
}
else if (Rs >= 684) {
if (Rs < 759) {
return 8;
}
else if (Rs >= 759) {
return 1;
}
}
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (Rs >= 798) {
if (ssDCval < 62) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
return 8;
}
}
else if (ssDCval >= 62) {
return 2;
}
}
}
}
else if (Rs >= 869) {
if (refDCval < 64) {
if (MVDiff < 3) {
if (AFD < 7) {
if (mcTcor < 96) {
if (Rs < 1567) {
if (diffRsCsdiff < -11) {
if (Rs < 1116) {
return 4;
}
else if (Rs >= 1116) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
return 8;
}
}
}
else if (diffRsCsdiff >= -11) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Cs < 579) {
if (Rs < 1013) {
return 8;
}
else if (Rs >= 1013) {
return 1;
}
}
else if (Cs >= 579) {
return 8;
}
}
}
}
else if (Rs >= 1567) {
if (SC < 2767) {
if (Cs < 1002) {
if (Rs < 1801) {
if (Rs < 1642) {
return 4;
}
else if (Rs >= 1642) {
return 4;
}
}
else if (Rs >= 1801) {
return 8;
}
}
else if (Cs >= 1002) {
return 8;
}
}
else if (SC >= 2767) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (SC < 2285) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (RsDiff < 3) {
if (Rs < 1340) {
return 8;
}
else if (Rs >= 1340) {
return 4;
}
}
else if (RsDiff >= 3) {
if (Rs < 1184) {
return 4;
}
else if (Rs >= 1184) {
return 1;
}
}
}
}
else if (SC >= 2285) {
return 8;
}
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (MVDiff >= 3) {
if (MVDiff < 11) {
if (Cs < 679) {
if (refDCval < 55) {
if (Rs < 891) {
return 8;
}
else if (Rs >= 891) {
return 1;
}
}
else if (refDCval >= 55) {
return 8;
}
}
else if (Cs >= 679) {
if (Rs < 1305) {
if (Rs < 1104) {
if (Cs < 842) {
if (negBalance < 29) {
return 1;
}
else if (negBalance >= 29) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Rs < 1008) {
return 4;
}
else if (Rs >= 1008) {
return 1;
}
}
}
}
else if (Cs >= 842) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
return 8;
}
}
}
else if (Rs >= 1104) {
if (mu_mv_mag_sq < 45) {
if (Rs < 1139) {
return 1;
}
else if (Rs >= 1139) {
return 8;
}
}
else if (mu_mv_mag_sq >= 45) {
if (diffMVdiffVal < -2) {
if (Rs < 1170) {
return 8;
}
else if (Rs >= 1170) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
}
}
else if (Rs >= 1305) {
if (AFD < 7) {
if (Cs < 832) {
return 2;
}
else if (Cs >= 832) {
return 8;
}
}
else if (AFD >= 7) {
return 4;
}
}
}
}
else if (MVDiff >= 11) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Cs < 785) {
if (Rs < 971) {
if (Rs < 910) {
return 8;
}
else if (Rs >= 910) {
return 4;
}
}
else if (Rs >= 971) {
if (Rs < 1096) {
return 1;
}
else if (Rs >= 1096) {
return 2;
}
}
}
else if (Cs >= 785) {
return 8;
}
}
}
}
}
else if (refDCval >= 64) {
if (Rs < 1157) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
if (Cs < 1141) {
return 8;
}
else if (Cs >= 1141) {
return 4;
}
}
}
}
else if (Rs >= 1157) {
if (diffAFD < -1) {
if (ssDCval < 67) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (ssDCval >= 67) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
return 1;
}
}
}
else if (diffAFD >= -1) {
if (negBalance < 40) {
if (RsDiff < 3) {
if (diffAFD < 1) {
if (Rs < 1262) {
if (Rs < 1213) {
return 8;
}
else if (Rs >= 1213) {
return 1;
}
}
else if (Rs >= 1262) {
return 8;
}
}
else if (diffAFD >= 1) {
if (Rs < 1248) {
return 8;
}
else if (Rs >= 1248) {
return 1;
}
}
}
else if (RsDiff >= 3) {
if (Cs < 1145) {
if (Cs < 1074) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (Cs >= 1074) {
return 4;
}
}
else if (Cs >= 1145) {
if (posBalance < 52) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 4;
}
}
else if (posBalance >= 52) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
else if (negBalance >= 40) {
if (Rs < 1345) {
return 1;
}
else if (Rs >= 1345) {
if (Rs < 1536) {
return 8;
}
else if (Rs >= 1536) {
if (Rs < 1707) {
return 2;
}
else if (Rs >= 1707) {
return 1;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 19) {
if (Cs < 955) {
if (SC < 2657) {
if (tcor < 96) {
if (diffMVdiffVal < 2) {
if (TSC < 5) {
if (refDCval < 48) {
return 1;
}
else if (refDCval >= 48) {
if (diffTSC < -1) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 4;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
}
else if (TSC >= 5) {
if (posBalance < 46) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
return 4;
}
}
else if (posBalance >= 46) {
if (tcor < 91) {
if (negBalance < 36) {
if (gchDC < 1) {
if (Rs < 1144) {
return 8;
}
else if (Rs >= 1144) {
return 2;
}
}
else if (gchDC >= 1) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (Cs < 639) {
return 1;
}
else if (Cs >= 639) {
return 8;
}
}
}
else if (tcor >= 91) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (posBalance < 53) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (tcor < 87) {
if (SC < 1440) {
return 1;
}
else if (SC >= 1440) {
return 8;
}
}
else if (tcor >= 87) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
return 8;
}
}
}
}
}
}
else if (tcor >= 96) {
return 8;
}
}
else if (SC >= 2657) {
if (refDCval < 69) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 2;
}
}
else if (refDCval >= 69) {
if (Cs < 848) {
return 1;
}
else if (Cs >= 848) {
return 4;
}
}
}
}
else if (Cs >= 955) {
if (Rs < 1333) {
if (Rs < 696) {
return 1;
}
else if (Rs >= 696) {
if (AFD < 10) {
if (RsDiff < 4) {
if (diffMVdiffVal < 1) {
if (Rs < 1204) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (Rs >= 1204) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Cs < 1194) {
return 2;
}
else if (Cs >= 1194) {
return 4;
}
}
}
}
else if (RsDiff >= 4) {
if (mu_mv_mag_sq < 121) {
if (SC < 2483) {
return 4;
}
else if (SC >= 2483) {
if (Cs < 1217) {
return 8;
}
else if (Cs >= 1217) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 121) {
return 8;
}
}
}
else if (AFD >= 10) {
if (Rs < 1144) {
return 8;
}
else if (Rs >= 1144) {
if (Rs < 1244) {
return 4;
}
else if (Rs >= 1244) {
return 2;
}
}
}
}
}
else if (Rs >= 1333) {
if (refDCval < 60) {
if (negBalance < 35) {
if (tcor < 90) {
if (RsCsDiff < 38) {
if (negBalance < 27) {
if (Rs < 1920) {
return 8;
}
else if (Rs >= 1920) {
return 4;
}
}
else if (negBalance >= 27) {
return 2;
}
}
else if (RsCsDiff >= 38) {
return 8;
}
}
else if (tcor >= 90) {
if (Cs < 1077) {
return 1;
}
else if (Cs >= 1077) {
return 8;
}
}
}
else if (negBalance >= 35) {
return 8;
}
}
else if (refDCval >= 60) {
if (refDCval < 68) {
if (SC < 3470) {
if (CsDiff < 5) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
return 4;
}
}
else if (CsDiff >= 5) {
if (Rs < 1694) {
return 8;
}
else if (Rs >= 1694) {
if (Rs < 1881) {
return 2;
}
else if (Rs >= 1881) {
return 4;
}
}
}
}
else if (SC >= 3470) {
return 2;
}
}
else if (refDCval >= 68) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 98) {
if (refDCval < 84) {
if (Rs < 1744) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
return 8;
}
}
else if (Rs >= 1744) {
if (mu_mv_mag_sq < 103) {
if (negBalance < 40) {
if (Rs < 2042) {
return 8;
}
else if (Rs >= 2042) {
return 1;
}
}
else if (negBalance >= 40) {
return 1;
}
}
else if (mu_mv_mag_sq >= 103) {
return 8;
}
}
}
else if (refDCval >= 84) {
if (TSC < 5) {
if (posBalance < 52) {
if (Rs < 451) {
return 8;
}
else if (Rs >= 451) {
if (Rs < 743) {
return 2;
}
else if (Rs >= 743) {
if (Cs < 844) {
return 1;
}
else if (Cs >= 844) {
return 2;
}
}
}
}
else if (posBalance >= 52) {
if (ssDCval < 91) {
if (negBalance < 30) {
if (Cs < 842) {
return 2;
}
else if (Cs >= 842) {
return 8;
}
}
else if (negBalance >= 30) {
return 8;
}
}
else if (ssDCval >= 91) {
if (Rs < 678) {
return 2;
}
else if (Rs >= 678) {
if (Rs < 1288) {
return 1;
}
else if (Rs >= 1288) {
return 8;
}
}
}
}
}
else if (TSC >= 5) {
if (RsCsDiff < 29) {
if (tcor < 79) {
if (Rs < 516) {
return 2;
}
else if (Rs >= 516) {
return 8;
}
}
else if (tcor >= 79) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (RsCsDiff < 11) {
return 4;
}
else if (RsCsDiff >= 11) {
return 8;
}
}
}
}
else if (RsCsDiff >= 29) {
if (Rs < 1936) {
if (diffMVdiffVal < -2) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 1223) {
return 1;
}
else if (Rs >= 1223) {
return 8;
}
}
}
else if (diffMVdiffVal >= -2) {
if (RsCsDiff < 57) {
return 8;
}
else if (RsCsDiff >= 57) {
return 2;
}
}
}
else if (Rs >= 1936) {
if (Rs < 2302) {
return 1;
}
else if (Rs >= 2302) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 98) {
if (SC < 2722) {
if (Rs < 724) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (Cs < 617) {
if (Cs < 615) {
return 8;
}
else if (Cs >= 615) {
return 2;
}
}
else if (Cs >= 617) {
return 8;
}
}
}
else if (Rs >= 724) {
if (ssDCval < 130) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (refDCval < 101) {
if (diffMVdiffVal < -2) {
if (SC < 2492) {
return 2;
}
else if (SC >= 2492) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (refDCval >= 101) {
if (SC < 1584) {
return 1;
}
else if (SC >= 1584) {
if (negBalance < 26) {
if (Cs < 1183) {
return 8;
}
else if (Cs >= 1183) {
return 4;
}
}
else if (negBalance >= 26) {
return 8;
}
}
}
}
}
else if (ssDCval >= 130) {
if (TSC < 6) {
if (tcor < 92) {
if (Rs < 1594) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Rs < 866) {
return 8;
}
else if (Rs >= 866) {
return 4;
}
}
}
else if (Rs >= 1594) {
if (Rs < 1621) {
return 8;
}
else if (Rs >= 1621) {
return 1;
}
}
}
else if (tcor >= 92) {
if (negBalance < 37) {
if (Rs < 1696) {
return 8;
}
else if (Rs >= 1696) {
return 2;
}
}
else if (negBalance >= 37) {
if (MVDiff < 20) {
if (Rs < 1420) {
return 1;
}
else if (Rs >= 1420) {
return 8;
}
}
else if (MVDiff >= 20) {
return 4;
}
}
}
}
else if (TSC >= 6) {
if (Rs < 1654) {
return 8;
}
else if (Rs >= 1654) {
return 2;
}
}
}
}
}
else if (SC >= 2722) {
if (SC < 3580) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (MVDiff < 4) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (SC >= 3580) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (Rs < 2451) {
return 1;
}
else if (Rs >= 2451) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 42) {
if (Rs < 1564) {
if (mu_mv_mag_sq < 51) {
if (ssDCval < 124) {
if (ssDCval < 81) {
if (tcor < 84) {
if (Rs < 900) {
if (Cs < 777) {
if (MVDiff < 31) {
return 8;
}
else if (MVDiff >= 31) {
return 4;
}
}
else if (Cs >= 777) {
if (SC < 1788) {
return 4;
}
else if (SC >= 1788) {
return 2;
}
}
}
else if (Rs >= 900) {
return 8;
}
}
else if (tcor >= 84) {
if (RsCsDiff < 9) {
return 8;
}
else if (RsCsDiff >= 9) {
if (AFD < 9) {
if (Cs < 834) {
if (Rs < 826) {
return 4;
}
else if (Rs >= 826) {
return 8;
}
}
else if (Cs >= 834) {
if (MVDiff < 4) {
if (Rs < 1352) {
return 8;
}
else if (Rs >= 1352) {
return 4;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (AFD >= 9) {
if (Cs < 676) {
return 4;
}
else if (Cs >= 676) {
if (Cs < 1203) {
if (refDCval < 73) {
return 8;
}
else if (refDCval >= 73) {
if (Rs < 1362) {
return 4;
}
else if (Rs >= 1362) {
return 8;
}
}
}
else if (Cs >= 1203) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (negBalance < 43) {
if (ssDCval < 103) {
return 2;
}
else if (ssDCval >= 103) {
return 8;
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
else if (ssDCval >= 124) {
if (TSC < 6) {
if (refDCval < 138) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (refDCval >= 138) {
return 4;
}
}
else if (TSC >= 6) {
if (Rs < 1244) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (Rs >= 1244) {
if (Cs < 632) {
return 4;
}
else if (Cs >= 632) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 467) {
if (MVDiff < 4) {
if (Rs < 407) {
return 4;
}
else if (Rs >= 407) {
if (Cs < 598) {
return 8;
}
else if (Cs >= 598) {
return 1;
}
}
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (Rs >= 467) {
if (diffTSC < 1) {
if (TSC < 6) {
if (mcTcor < 94) {
if (posBalance < 56) {
if (Cs < 1293) {
if (Cs < 571) {
if (Rs < 896) {
return 8;
}
else if (Rs >= 896) {
return 2;
}
}
else if (Cs >= 571) {
return 8;
}
}
else if (Cs >= 1293) {
if (Rs < 1190) {
return 8;
}
else if (Rs >= 1190) {
return 1;
}
}
}
else if (posBalance >= 56) {
if (ssDCval < 74) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (ssDCval < 48) {
if (Rs < 751) {
return 4;
}
else if (Rs >= 751) {
return 1;
}
}
else if (ssDCval >= 48) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
return 4;
}
}
}
}
else if (ssDCval >= 74) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (Cs < 1021) {
if (ssDCval < 62) {
if (diffMVdiffVal < -1) {
if (Cs < 772) {
return 4;
}
else if (Cs >= 772) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 718) {
return 4;
}
else if (Rs >= 718) {
return 8;
}
}
}
else if (ssDCval >= 62) {
if (Rs < 606) {
return 8;
}
else if (Rs >= 606) {
if (MVDiff < 2) {
if (posBalance < 51) {
if (Rs < 894) {
return 2;
}
else if (Rs >= 894) {
return 8;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Cs < 885) {
if (Cs < 780) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (Cs >= 780) {
return 1;
}
}
else if (Cs >= 885) {
return 8;
}
}
}
}
}
else if (Cs >= 1021) {
if (mu_mv_mag_sq < 65) {
return 4;
}
else if (mu_mv_mag_sq >= 65) {
if (ssDCval < 71) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 1;
}
}
else if (ssDCval >= 71) {
if (ssDCval < 95) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 4;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
}
else if (TSC >= 6) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (refDCval < 80) {
if (refDCval < 53) {
if (tcor < 75) {
if (mu_mv_mag_sq < 102) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 858) {
return 8;
}
else if (Rs >= 858) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 102) {
return 1;
}
}
else if (tcor >= 75) {
return 8;
}
}
else if (refDCval >= 53) {
if (Rs < 1253) {
if (diffAFD < 1) {
if (posBalance < 53) {
if (negBalance < 45) {
if (Rs < 1138) {
return 1;
}
else if (Rs >= 1138) {
return 2;
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (posBalance >= 53) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Cs < 872) {
if (Rs < 832) {
return 8;
}
else if (Rs >= 832) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (Cs >= 872) {
return 8;
}
}
}
}
else if (diffAFD >= 1) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
else if (Rs >= 1253) {
if (negBalance < 46) {
if (AFD < 9) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
else if (AFD >= 9) {
if (Rs < 1329) {
return 8;
}
else if (Rs >= 1329) {
return 4;
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
}
else if (refDCval >= 80) {
if (CsDiff < 7) {
if (SC < 2612) {
if (diffRsCsdiff < -34) {
return 1;
}
else if (diffRsCsdiff >= -34) {
return 8;
}
}
else if (SC >= 2612) {
if (Cs < 1129) {
return 1;
}
else if (Cs >= 1129) {
return 8;
}
}
}
else if (CsDiff >= 7) {
return 2;
}
}
}
}
}
else if (diffTSC >= 1) {
if (Rs < 1218) {
if (MVDiff < 9) {
if (mcTcor < 95) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 72) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 1130) {
return 2;
}
else if (Rs >= 1130) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (Rs < 613) {
return 8;
}
else if (Rs >= 613) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (posBalance < 45) {
if (Rs < 578) {
return 2;
}
else if (Rs >= 578) {
return 8;
}
}
else if (posBalance >= 45) {
return 8;
}
}
}
}
else if (mcTcor >= 95) {
if (tcor < 96) {
if (negBalance < 47) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 8;
}
}
else if (negBalance >= 47) {
return 1;
}
}
else if (tcor >= 96) {
if (RsCsDiff < 16) {
return 4;
}
else if (RsCsDiff >= 16) {
return 8;
}
}
}
}
else if (MVDiff >= 9) {
if (ssDCval < 79) {
if (mu_mv_mag_sq < 66) {
if (SC < 1506) {
return 4;
}
else if (SC >= 1506) {
return 1;
}
}
else if (mu_mv_mag_sq >= 66) {
return 4;
}
}
else if (ssDCval >= 79) {
if (Rs < 1157) {
return 8;
}
else if (Rs >= 1157) {
return 4;
}
}
}
}
else if (Rs >= 1218) {
if (Cs < 1223) {
if (Rs < 1226) {
return 1;
}
else if (Rs >= 1226) {
if (ssDCval < 123) {
if (SC < 2602) {
return 8;
}
else if (SC >= 2602) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (Rs < 1460) {
return 8;
}
else if (Rs >= 1460) {
if (Rs < 1490) {
return 1;
}
else if (Rs >= 1490) {
return 4;
}
}
}
}
}
else if (ssDCval >= 123) {
return 4;
}
}
}
else if (Cs >= 1223) {
if (Cs < 1250) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 4;
}
}
else if (Cs >= 1250) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 1564) {
if (refDCval < 72) {
if (refDCval < 62) {
if (MVDiff < 21) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (MVDiff >= 21) {
if (Rs < 1583) {
return 8;
}
else if (Rs >= 1583) {
return 4;
}
}
}
else if (refDCval >= 62) {
if (tcor < 89) {
if (Rs < 2017) {
return 8;
}
else if (Rs >= 2017) {
return 4;
}
}
else if (tcor >= 89) {
if (TSC < 8) {
if (Cs < 1037) {
if (Rs < 1679) {
if (Cs < 830) {
return 2;
}
else if (Cs >= 830) {
return 8;
}
}
else if (Rs >= 1679) {
return 2;
}
}
else if (Cs >= 1037) {
return 8;
}
}
else if (TSC >= 8) {
return 2;
}
}
}
}
else if (refDCval >= 72) {
if (tcor < 85) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (diffAFD < -1) {
if (Cs < 1176) {
return 8;
}
else if (Cs >= 1176) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (tcor >= 85) {
if (posBalance < 54) {
if (ssDCval < 100) {
if (TSC < 8) {
if (CsDiff < 3) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (CsDiff >= 3) {
if (ssDCval < 86) {
if (Rs < 1842) {
return 8;
}
else if (Rs >= 1842) {
return 1;
}
}
else if (ssDCval >= 86) {
return 2;
}
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (ssDCval >= 100) {
if (SC < 3778) {
return 8;
}
else if (SC >= 3778) {
return 1;
}
}
}
else if (posBalance >= 54) {
if (mu_mv_mag_sq < 63) {
if (SC < 2284) {
if (Rs < 1591) {
return 8;
}
else if (Rs >= 1591) {
return 1;
}
}
else if (SC >= 2284) {
return 8;
}
}
else if (mu_mv_mag_sq >= 63) {
if (RsCsDiff < 23) {
if (ssDCval < 111) {
return 8;
}
else if (ssDCval >= 111) {
return 1;
}
}
else if (RsCsDiff >= 23) {
if (tcor < 92) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
return 1;
}
}
else if (tcor >= 92) {
if (negBalance < 50) {
if (ssDCval < 82) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (ssDCval >= 82) {
if (posBalance < 60) {
if (MVDiff < 2) {
if (Rs < 1774) {
return 8;
}
else if (Rs >= 1774) {
return 1;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (posBalance >= 60) {
return 1;
}
}
}
else if (negBalance >= 50) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 156) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
if (posBalance < 61) {
if (Rs < 1841) {
return 2;
}
else if (Rs >= 1841) {
return 8;
}
}
else if (posBalance >= 61) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
}
else if (AFD >= 14) {
if (MVDiff < 8) {
if (TSC < 9) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (tcor < 80) {
if (diffRsCsdiff < 4) {
return 8;
}
else if (diffRsCsdiff >= 4) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (SC < 1637) {
if (Rs < 573) {
return 8;
}
else if (Rs >= 573) {
return 1;
}
}
else if (SC >= 1637) {
return 8;
}
}
}
}
else if (tcor >= 80) {
if (ssDCval < 78) {
if (Rs < 847) {
return 8;
}
else if (Rs >= 847) {
return 1;
}
}
else if (ssDCval >= 78) {
if (Rs < 1745) {
return 4;
}
else if (Rs >= 1745) {
return 8;
}
}
}
}
}
else if (TSC >= 9) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (diffRsCsdiff < -10) {
return 8;
}
else if (diffRsCsdiff >= -10) {
if (Rs < 1511) {
return 4;
}
else if (Rs >= 1511) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 8) {
if (Rs < 1415) {
if (refDCval < 119) {
if (mcTcor < 88) {
if (TSC < 11) {
if (refDCval < 98) {
if (Rs < 598) {
if (TSC < 9) {
if (RsCsDiff < 48) {
if (mu_mv_mag_sq < 7) {
return 4;
}
else if (mu_mv_mag_sq >= 7) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
if (Rs < 504) {
return 4;
}
else if (Rs >= 504) {
return 1;
}
}
}
}
else if (RsCsDiff >= 48) {
if (diffRsCsdiff < 53) {
return 4;
}
else if (diffRsCsdiff >= 53) {
if (Rs < 369) {
return 4;
}
else if (Rs >= 369) {
return 8;
}
}
}
}
else if (TSC >= 9) {
if (mu_mv_mag_sq < 8) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (mu_mv_mag_sq >= 8) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
}
else if (Rs >= 598) {
if (diffMVdiffVal < 17) {
if (negBalance < 44) {
if (Rs < 1067) {
return 1;
}
else if (Rs >= 1067) {
return 8;
}
}
else if (negBalance >= 44) {
if (posBalance < 59) {
if (diffAFD < 2) {
if (Rs < 1017) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 1;
}
}
else if (Rs >= 1017) {
return 2;
}
}
else if (diffAFD >= 2) {
if (Cs < 673) {
if (Rs < 839) {
return 8;
}
else if (Rs >= 839) {
return 2;
}
}
else if (Cs >= 673) {
return 1;
}
}
}
else if (posBalance >= 59) {
if (negBalance < 51) {
if (CsDiff < 5) {
if (Rs < 859) {
return 4;
}
else if (Rs >= 859) {
return 8;
}
}
else if (CsDiff >= 5) {
if (Rs < 1102) {
return 2;
}
else if (Rs >= 1102) {
return 4;
}
}
}
else if (negBalance >= 51) {
if (Cs < 652) {
return 8;
}
else if (Cs >= 652) {
if (tcor < 57) {
if (ssDCval < 55) {
if (MVDiff < 48) {
return 4;
}
else if (MVDiff >= 48) {
return 8;
}
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (tcor >= 57) {
return 4;
}
}
}
}
}
}
else if (diffMVdiffVal >= 17) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 4;
}
}
}
}
else if (refDCval >= 98) {
return 4;
}
}
else if (TSC >= 11) {
if (ssDCval < 67) {
if (RsCsDiff < 79) {
if (refDCval < 43) {
return 2;
}
else if (refDCval >= 43) {
return 8;
}
}
else if (RsCsDiff >= 79) {
if (diffRsCsdiff < -7) {
if (Rs < 1019) {
return 8;
}
else if (Rs >= 1019) {
return 1;
}
}
else if (diffRsCsdiff >= -7) {
return 2;
}
}
}
else if (ssDCval >= 67) {
if (RsCsDiff < 48) {
if (tcor < 74) {
if (gchDC < 3) {
if (Cs < 577) {
return 8;
}
else if (Cs >= 577) {
return 4;
}
}
else if (gchDC >= 3) {
if (MVDiff < 181) {
return 8;
}
else if (MVDiff >= 181) {
return 4;
}
}
}
else if (tcor >= 74) {
return 8;
}
}
else if (RsCsDiff >= 48) {
if (mcTcor < 84) {
if (RsCsDiff < 63) {
if (diffAFD < 11) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (diffAFD >= 11) {
return 2;
}
}
else if (RsCsDiff >= 63) {
if (mcTcor < 80) {
return 4;
}
else if (mcTcor >= 80) {
if (Rs < 704) {
return 1;
}
else if (Rs >= 704) {
return 8;
}
}
}
}
else if (mcTcor >= 84) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
if (Rs < 487) {
return 8;
}
else if (Rs >= 487) {
if (Cs < 1235) {
return 2;
}
else if (Cs >= 1235) {
return 4;
}
}
}
}
}
}
}
}
else if (mcTcor >= 88) {
if (Rs < 836) {
if (Cs < 942) {
if (Rs < 326) {
if (negBalance < 62) {
if (Cs < 553) {
return 4;
}
else if (Cs >= 553) {
return 2;
}
}
else if (negBalance >= 62) {
return 4;
}
}
else if (Rs >= 326) {
if (MVDiff < 57) {
if (Rs < 534) {
if (AFD < 22) {
if (refDCval < 100) {
if (RsDiff < 5) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
return 8;
}
}
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (refDCval >= 100) {
if (Cs < 670) {
return 4;
}
else if (Cs >= 670) {
return 2;
}
}
}
else if (AFD >= 22) {
if (SC < 1038) {
return 1;
}
else if (SC >= 1038) {
return 2;
}
}
}
else if (Rs >= 534) {
if (tcor < 69) {
if (diffMVdiffVal < -9) {
if (Rs < 715) {
return 4;
}
else if (Rs >= 715) {
return 1;
}
}
else if (diffMVdiffVal >= -9) {
if (mu_mv_mag_sq < 12) {
if (Rs < 741) {
return 8;
}
else if (Rs >= 741) {
return 1;
}
}
else if (mu_mv_mag_sq >= 12) {
return 1;
}
}
}
else if (tcor >= 69) {
if (posBalance < 59) {
if (mu_mv_mag_sq < 12) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
if (tcor < 80) {
return 1;
}
else if (tcor >= 80) {
if (Rs < 679) {
return 1;
}
else if (Rs >= 679) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 12) {
if (SC < 1518) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (SC >= 1518) {
return 1;
}
}
}
else if (posBalance >= 59) {
if (mcTcor < 93) {
if (Rs < 652) {
return 2;
}
else if (Rs >= 652) {
if (gchDC < 5) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
else if (gchDC >= 5) {
return 4;
}
}
}
else if (mcTcor >= 93) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 57) {
if (ssDCval < 89) {
if (gchDC < 5) {
if (diffRsCsdiff < -27) {
if (Cs < 787) {
return 8;
}
else if (Cs >= 787) {
return 2;
}
}
else if (diffRsCsdiff >= -27) {
return 1;
}
}
else if (gchDC >= 5) {
if (refDCval < 89) {
if (Rs < 630) {
return 2;
}
else if (Rs >= 630) {
return 8;
}
}
else if (refDCval >= 89) {
if (Rs < 459) {
return 1;
}
else if (Rs >= 459) {
return 8;
}
}
}
}
else if (ssDCval >= 89) {
if (diffAFD < -6) {
return 1;
}
else if (diffAFD >= -6) {
return 2;
}
}
}
}
}
else if (Cs >= 942) {
if (CsDiff < 6) {
if (Rs < 825) {
return 2;
}
else if (Rs >= 825) {
return 1;
}
}
else if (CsDiff >= 6) {
if (AFD < 18) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 15) {
if (AFD < 15) {
if (Rs < 706) {
return 8;
}
else if (Rs >= 706) {
return 1;
}
}
else if (AFD >= 15) {
return 4;
}
}
else if (mu_mv_mag_sq >= 15) {
return 1;
}
}
else if (gchDC >= 2) {
if (diffRsCsdiff < -5) {
if (Rs < 766) {
return 4;
}
else if (Rs >= 766) {
return 1;
}
}
else if (diffRsCsdiff >= -5) {
if (mu_mv_mag_sq < 28) {
if (SC < 1661) {
return 8;
}
else if (SC >= 1661) {
return 2;
}
}
else if (mu_mv_mag_sq >= 28) {
return 8;
}
}
}
}
else if (AFD >= 18) {
if (MVDiff < 60) {
if (mcTcor < 91) {
return 2;
}
else if (mcTcor >= 91) {
if (MVDiff < 50) {
return 4;
}
else if (MVDiff >= 50) {
if (Rs < 630) {
return 4;
}
else if (Rs >= 630) {
return 2;
}
}
}
}
else if (MVDiff >= 60) {
if (ssDCval < 97) {
return 4;
}
else if (ssDCval >= 97) {
return 1;
}
}
}
}
}
}
else if (Rs >= 836) {
if (CsDiff < 7) {
if (mu_mv_mag_sq < 7) {
if (diffAFD < -1) {
if (Cs < 832) {
if (refDCval < 84) {
return 4;
}
else if (refDCval >= 84) {
return 2;
}
}
else if (Cs >= 832) {
if (Rs < 1229) {
if (MVDiff < 15) {
return 2;
}
else if (MVDiff >= 15) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 2;
}
}
}
else if (Rs >= 1229) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 8;
}
}
}
}
else if (diffAFD >= -1) {
if (Cs < 739) {
return 4;
}
else if (Cs >= 739) {
if (SC < 2254) {
if (ssDCval < 101) {
if (Rs < 1037) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
return 4;
}
}
else if (Rs >= 1037) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
if (RsDiff < 5) {
if (Rs < 1078) {
return 8;
}
else if (Rs >= 1078) {
return 4;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
else if (ssDCval >= 101) {
if (SC < 2169) {
return 4;
}
else if (SC >= 2169) {
return 8;
}
}
}
else if (SC >= 2254) {
if (Rs < 1340) {
return 4;
}
else if (Rs >= 1340) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (posBalance < 48) {
if (AFD < 17) {
if (Rs < 1103) {
return 8;
}
else if (Rs >= 1103) {
return 1;
}
}
else if (AFD >= 17) {
return 2;
}
}
else if (posBalance >= 48) {
if (negBalance < 48) {
if (gchDC < 3) {
if (Rs < 1038) {
return 1;
}
else if (Rs >= 1038) {
return 4;
}
}
else if (gchDC >= 3) {
return 8;
}
}
else if (negBalance >= 48) {
if (AFD < 20) {
if (posBalance < 53) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 4;
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (AFD >= 20) {
return 1;
}
}
}
}
}
else if (CsDiff >= 7) {
if (gchDC < 5) {
if (refDCval < 68) {
return 8;
}
else if (refDCval >= 68) {
if (ssDCval < 96) {
return 2;
}
else if (ssDCval >= 96) {
return 8;
}
}
}
else if (gchDC >= 5) {
return 4;
}
}
}
}
}
else if (refDCval >= 119) {
if (TSC < 7) {
if (Rs < 441) {
return 2;
}
else if (Rs >= 441) {
return 8;
}
}
else if (TSC >= 7) {
if (posBalance < 70) {
if (Rs < 547) {
return 2;
}
else if (Rs >= 547) {
if (TSC < 9) {
if (Cs < 1017) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Rs < 770) {
return 8;
}
else if (Rs >= 770) {
return 2;
}
}
}
else if (Cs >= 1017) {
return 1;
}
}
else if (TSC >= 9) {
if (refDCval < 152) {
return 1;
}
else if (refDCval >= 152) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
}
}
}
else if (posBalance >= 70) {
if (Cs < 679) {
return 2;
}
else if (Cs >= 679) {
return 8;
}
}
}
}
}
else if (Rs >= 1415) {
if (negBalance < 46) {
if (SC < 2494) {
return 4;
}
else if (SC >= 2494) {
if (MVDiff < 67) {
return 1;
}
else if (MVDiff >= 67) {
return 8;
}
}
}
else if (negBalance >= 46) {
if (tcor < 69) {
if (ssDCval < 89) {
if (Rs < 1561) {
return 8;
}
else if (Rs >= 1561) {
return 1;
}
}
else if (ssDCval >= 89) {
if (Rs < 1961) {
return 4;
}
else if (Rs >= 1961) {
return 8;
}
}
}
else if (tcor >= 69) {
if (negBalance < 49) {
if (RsCsDiff < 48) {
return 8;
}
else if (RsCsDiff >= 48) {
return 4;
}
}
else if (negBalance >= 49) {
if (diffMVdiffVal < -14) {
return 2;
}
else if (diffMVdiffVal >= -14) {
return 8;
}
}
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (refDCval < 38) {
return 8;
}
else if (refDCval >= 38) {
if (Cs < 888) {
if (refDCval < 82) {
if (mcTcor < 92) {
if (mcTcor < 85) {
if (TSC < 9) {
if (Rs < 1000) {
if (MVDiff < 34) {
if (Rs < 905) {
return 8;
}
else if (Rs >= 905) {
return 1;
}
}
else if (MVDiff >= 34) {
return 4;
}
}
else if (Rs >= 1000) {
if (SC < 2593) {
return 8;
}
else if (SC >= 2593) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
return 8;
}
}
}
}
else if (TSC >= 9) {
if (Cs < 770) {
if (gchDC < 2) {
if (RsCsDiff < 89) {
return 8;
}
else if (RsCsDiff >= 89) {
return 1;
}
}
else if (gchDC >= 2) {
if (mcTcor < 75) {
return 8;
}
else if (mcTcor >= 75) {
if (Cs < 752) {
return 4;
}
else if (Cs >= 752) {
return 1;
}
}
}
}
else if (Cs >= 770) {
if (ssDCval < 72) {
if (Rs < 2232) {
return 4;
}
else if (Rs >= 2232) {
return 2;
}
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
else if (mcTcor >= 85) {
if (RsDiff < 11) {
if (tcor < 74) {
if (CsDiff < 5) {
if (Rs < 1116) {
if (Rs < 944) {
return 2;
}
else if (Rs >= 944) {
return 4;
}
}
else if (Rs >= 1116) {
if (Rs < 2258) {
return 8;
}
else if (Rs >= 2258) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 1038) {
if (CsDiff < 7) {
if (Rs < 811) {
return 4;
}
else if (Rs >= 811) {
if (Rs < 918) {
return 8;
}
else if (Rs >= 918) {
return 1;
}
}
}
else if (CsDiff >= 7) {
return 8;
}
}
else if (Rs >= 1038) {
if (RsCsDiff < 127) {
if (gchDC < 2) {
if (Cs < 813) {
return 2;
}
else if (Cs >= 813) {
return 4;
}
}
else if (gchDC >= 2) {
return 4;
}
}
else if (RsCsDiff >= 127) {
if (RsCsDiff < 143) {
return 2;
}
else if (RsCsDiff >= 143) {
return 4;
}
}
}
}
}
else if (tcor >= 74) {
if (ssDCval < 57) {
if (Rs < 1345) {
return 8;
}
else if (Rs >= 1345) {
if (Cs < 596) {
return 1;
}
else if (Cs >= 596) {
return 4;
}
}
}
else if (ssDCval >= 57) {
if (mcTcor < 90) {
if (MVDiff < 20) {
if (negBalance < 39) {
if (Rs < 1872) {
return 8;
}
else if (Rs >= 1872) {
return 4;
}
}
else if (negBalance >= 39) {
return 2;
}
}
else if (MVDiff >= 20) {
if (mcTcor < 89) {
if (TSC < 7) {
if (Rs < 1754) {
return 2;
}
else if (Rs >= 1754) {
return 1;
}
}
else if (TSC >= 7) {
if (Rs < 2207) {
return 4;
}
else if (Rs >= 2207) {
return 8;
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1249) {
return 1;
}
else if (Rs >= 1249) {
return 8;
}
}
}
}
else if (mcTcor >= 90) {
if (diffMVdiffVal < -11) {
if (Rs < 2222) {
return 4;
}
else if (Rs >= 2222) {
return 8;
}
}
else if (diffMVdiffVal >= -11) {
if (diffRsCsdiff < 35) {
return 2;
}
else if (diffRsCsdiff >= 35) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 11) {
if (SC < 3157) {
if (tcor < 50) {
return 4;
}
else if (tcor >= 50) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 8;
}
}
}
else if (SC >= 3157) {
return 1;
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 2104) {
if (TSC < 7) {
if (mu_mv_mag_sq < 34) {
if (negBalance < 58) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 5) {
if (Rs < 824) {
return 1;
}
else if (Rs >= 824) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
return 1;
}
}
}
else if (negBalance >= 58) {
return 8;
}
}
else if (mu_mv_mag_sq >= 34) {
if (Rs < 1176) {
if (Rs < 922) {
return 8;
}
else if (Rs >= 922) {
return 1;
}
}
else if (Rs >= 1176) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (posBalance < 36) {
return 1;
}
else if (posBalance >= 36) {
return 2;
}
}
}
}
}
else if (TSC >= 7) {
if (RsDiff < 8) {
if (AFD < 20) {
if (tcor < 82) {
if (CsDiff < 5) {
if (Rs < 1138) {
return 8;
}
else if (Rs >= 1138) {
return 4;
}
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (tcor >= 82) {
if (diffRsCsdiff < -29) {
return 2;
}
else if (diffRsCsdiff >= -29) {
if (Rs < 1774) {
return 8;
}
else if (Rs >= 1774) {
if (Rs < 2057) {
return 1;
}
else if (Rs >= 2057) {
return 8;
}
}
}
}
}
else if (AFD >= 20) {
return 2;
}
}
else if (RsDiff >= 8) {
if (Rs < 1353) {
if (mu_mv_mag_sq < 14) {
if (Cs < 731) {
return 2;
}
else if (Cs >= 731) {
return 8;
}
}
else if (mu_mv_mag_sq >= 14) {
return 4;
}
}
else if (Rs >= 1353) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 2;
}
}
}
}
}
else if (Rs >= 2104) {
if (TSC < 6) {
if (Cs < 676) {
return 2;
}
else if (Cs >= 676) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (TSC >= 6) {
if (ssDCval < 59) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
else if (ssDCval >= 59) {
return 2;
}
}
}
}
}
else if (refDCval >= 82) {
if (tcor < 80) {
if (SC < 3009) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 14) {
if (Cs < 691) {
if (tcor < 75) {
return 1;
}
else if (tcor >= 75) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (Rs < 1525) {
return 2;
}
else if (Rs >= 1525) {
return 1;
}
}
}
}
else if (Cs >= 691) {
if (Rs < 1953) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Rs < 1290) {
return 8;
}
else if (Rs >= 1290) {
return 2;
}
}
}
else if (Rs >= 1953) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (ssDCval < 136) {
if (mu_mv_mag_sq < 40) {
return 2;
}
else if (mu_mv_mag_sq >= 40) {
if (Rs < 1373) {
if (Rs < 1035) {
return 8;
}
else if (Rs >= 1035) {
return 4;
}
}
else if (Rs >= 1373) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
}
else if (ssDCval >= 136) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (posBalance < 66) {
if (MVDiff < 82) {
if (mcTcor < 92) {
if (MVDiff < 24) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 1;
}
}
else if (MVDiff >= 24) {
if (ssDCval < 99) {
if (ssDCval < 89) {
if (Rs < 1681) {
return 4;
}
else if (Rs >= 1681) {
return 8;
}
}
else if (ssDCval >= 89) {
if (Cs < 820) {
return 1;
}
else if (Cs >= 820) {
return 2;
}
}
}
else if (ssDCval >= 99) {
if (diffTSC < 2) {
if (Rs < 2232) {
if (MVDiff < 30) {
return 8;
}
else if (MVDiff >= 30) {
return 4;
}
}
else if (Rs >= 2232) {
return 8;
}
}
else if (diffTSC >= 2) {
if (SC < 1902) {
if (Rs < 1116) {
return 1;
}
else if (Rs >= 1116) {
return 4;
}
}
else if (SC >= 1902) {
return 8;
}
}
}
}
}
else if (mcTcor >= 92) {
if (mu_mv_mag_sq < 34) {
if (Rs < 1442) {
if (MVDiff < 34) {
return 8;
}
else if (MVDiff >= 34) {
if (Rs < 1395) {
return 1;
}
else if (Rs >= 1395) {
return 8;
}
}
}
else if (Rs >= 1442) {
return 1;
}
}
else if (mu_mv_mag_sq >= 34) {
if (RsDiff < 14) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
if (Rs < 823) {
return 1;
}
else if (Rs >= 823) {
return 4;
}
}
}
else if (RsDiff >= 14) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
if (AFD < 25) {
return 1;
}
else if (AFD >= 25) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 82) {
if (posBalance < 63) {
if (Cs < 662) {
return 8;
}
else if (Cs >= 662) {
if (Cs < 827) {
return 2;
}
else if (Cs >= 827) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (Rs < 1201) {
return 1;
}
else if (Rs >= 1201) {
return 4;
}
}
}
}
else if (posBalance >= 66) {
if (diffRsCsdiff < 80) {
if (posBalance < 68) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (Rs < 2121) {
return 2;
}
else if (Rs >= 2121) {
return 1;
}
}
}
else if (posBalance >= 68) {
if (ssDCval < 92) {
if (Cs < 661) {
return 2;
}
else if (Cs >= 661) {
return 1;
}
}
else if (ssDCval >= 92) {
if (RsCsDiff < 186) {
if (diffRsCsdiff < 14) {
if (RsCsDiff < 117) {
return 8;
}
else if (RsCsDiff >= 117) {
if (Cs < 685) {
return 8;
}
else if (Cs >= 685) {
return 2;
}
}
}
else if (diffRsCsdiff >= 14) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Rs < 882) {
return 8;
}
else if (Rs >= 882) {
return 1;
}
}
}
}
else if (RsCsDiff >= 186) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 80) {
return 2;
}
}
}
}
else if (SC >= 3009) {
if (mu_mv_mag_sq < 37) {
if (ssDCval < 124) {
if (gchDC < 2) {
if (TSC < 12) {
if (Cs < 852) {
return 8;
}
else if (Cs >= 852) {
return 2;
}
}
else if (TSC >= 12) {
return 1;
}
}
else if (gchDC >= 2) {
if (SC < 3164) {
return 2;
}
else if (SC >= 3164) {
if (Rs < 2441) {
return 1;
}
else if (Rs >= 2441) {
return 8;
}
}
}
}
else if (ssDCval >= 124) {
if (Rs < 2298) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 8;
}
}
else if (Rs >= 2298) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 4;
}
}
}
}
else if (tcor >= 80) {
if (ssDCval < 107) {
if (MVDiff < 44) {
if (RsCsDiff < 137) {
if (diffTSC < -2) {
if (Rs < 1381) {
return 8;
}
else if (Rs >= 1381) {
return 4;
}
}
else if (diffTSC >= -2) {
if (ssDCval < 90) {
if (mcTcor < 93) {
if (diffMVdiffVal < -2) {
if (Rs < 1709) {
return 8;
}
else if (Rs >= 1709) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
else if (ssDCval >= 90) {
if (Cs < 838) {
if (tcor < 89) {
if (mcTcor < 91) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
else if (mcTcor >= 91) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
if (Rs < 1619) {
return 2;
}
else if (Rs >= 1619) {
return 1;
}
}
}
}
else if (tcor >= 89) {
return 1;
}
}
else if (Cs >= 838) {
if (Rs < 1864) {
if (Rs < 1587) {
return 8;
}
else if (Rs >= 1587) {
return 1;
}
}
else if (Rs >= 1864) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 137) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
if (mu_mv_mag_sq < 26) {
return 2;
}
else if (mu_mv_mag_sq >= 26) {
return 1;
}
}
}
}
else if (MVDiff >= 44) {
if (diffMVdiffVal < 3) {
if (refDCval < 93) {
return 8;
}
else if (refDCval >= 93) {
if (Rs < 2019) {
return 2;
}
else if (Rs >= 2019) {
return 8;
}
}
}
else if (diffMVdiffVal >= 3) {
if (Cs < 631) {
return 1;
}
else if (Cs >= 631) {
if (Rs < 1923) {
return 8;
}
else if (Rs >= 1923) {
return 2;
}
}
}
}
}
else if (ssDCval >= 107) {
if (TSC < 7) {
if (Rs < 1640) {
return 8;
}
else if (Rs >= 1640) {
return 2;
}
}
else if (TSC >= 7) {
if (RsCsDiff < 134) {
if (ssDCval < 130) {
if (SC < 2849) {
return 1;
}
else if (SC >= 2849) {
if (refDCval < 114) {
return 8;
}
else if (refDCval >= 114) {
if (Rs < 2323) {
return 1;
}
else if (Rs >= 2323) {
return 2;
}
}
}
}
else if (ssDCval >= 130) {
if (Cs < 664) {
if (Rs < 1475) {
return 8;
}
else if (Rs >= 1475) {
return 1;
}
}
else if (Cs >= 664) {
if (Rs < 2194) {
return 8;
}
else if (Rs >= 2194) {
return 1;
}
}
}
}
else if (RsCsDiff >= 134) {
if (MVDiff < 21) {
if (Rs < 2048) {
return 2;
}
else if (Rs >= 2048) {
return 1;
}
}
else if (MVDiff >= 21) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 888) {
if (refDCval < 114) {
if (SC < 3033) {
if (AFD < 19) {
if (tcor < 54) {
if (negBalance < 48) {
if (diffAFD < 4) {
return 1;
}
else if (diffAFD >= 4) {
if (Rs < 1098) {
return 4;
}
else if (Rs >= 1098) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (RsCsDiff < 94) {
return 8;
}
else if (RsCsDiff >= 94) {
if (Rs < 693) {
return 8;
}
else if (Rs >= 693) {
return 4;
}
}
}
}
else if (tcor >= 54) {
if (Cs < 978) {
if (ssDCval < 98) {
if (tcor < 67) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
return 8;
}
}
else if (tcor >= 67) {
if (Rs < 1718) {
if (tcor < 73) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 1;
}
}
else if (tcor >= 73) {
if (AFD < 17) {
if (diffRsCsdiff < 55) {
return 8;
}
else if (diffRsCsdiff >= 55) {
return 2;
}
}
else if (AFD >= 17) {
return 4;
}
}
}
else if (Rs >= 1718) {
if (mu_mv_mag_sq < 32) {
return 4;
}
else if (mu_mv_mag_sq >= 32) {
if (Rs < 2070) {
return 8;
}
else if (Rs >= 2070) {
return 1;
}
}
}
}
}
else if (ssDCval >= 98) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (Cs >= 978) {
if (refDCval < 84) {
if (TSC < 10) {
if (refDCval < 49) {
if (Rs < 948) {
return 1;
}
else if (Rs >= 948) {
return 8;
}
}
else if (refDCval >= 49) {
if (SC < 2869) {
if (mu_mv_mag_sq < 46) {
if (posBalance < 59) {
if (Rs < 1370) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
return 4;
}
}
}
else if (Rs >= 1370) {
if (negBalance < 51) {
if (Cs < 1108) {
if (Rs < 1711) {
if (Rs < 1588) {
return 2;
}
else if (Rs >= 1588) {
return 8;
}
}
else if (Rs >= 1711) {
return 2;
}
}
else if (Cs >= 1108) {
return 1;
}
}
else if (negBalance >= 51) {
if (tcor < 65) {
return 2;
}
else if (tcor >= 65) {
if (MVDiff < 12) {
if (Rs < 1666) {
return 8;
}
else if (Rs >= 1666) {
return 1;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
}
}
}
else if (posBalance >= 59) {
if (mu_mv_mag_sq < 2) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
if (SC < 2443) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (SC >= 2443) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (mcTcor < 89) {
if (Rs < 1582) {
return 8;
}
else if (Rs >= 1582) {
return 2;
}
}
else if (mcTcor >= 89) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 8;
}
}
}
}
else if (SC >= 2869) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (SC < 2962) {
if (Rs < 1733) {
return 4;
}
else if (Rs >= 1733) {
return 2;
}
}
else if (SC >= 2962) {
return 1;
}
}
}
}
}
}
else if (TSC >= 10) {
if (Rs < 1279) {
return 2;
}
else if (Rs >= 1279) {
return 8;
}
}
}
else if (refDCval >= 84) {
if (tcor < 73) {
if (tcor < 66) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (tcor >= 66) {
return 4;
}
}
else if (tcor >= 73) {
if (TSC < 9) {
if (ssDCval < 104) {
if (MVDiff < 9) {
if (Rs < 1607) {
return 1;
}
else if (Rs >= 1607) {
if (Rs < 1695) {
return 8;
}
else if (Rs >= 1695) {
return 4;
}
}
}
else if (MVDiff >= 9) {
if (mcTcor < 95) {
if (diffAFD < 5) {
return 2;
}
else if (diffAFD >= 5) {
return 8;
}
}
else if (mcTcor >= 95) {
if (Rs < 1377) {
return 2;
}
else if (Rs >= 1377) {
return 8;
}
}
}
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (TSC >= 9) {
if (Cs < 1094) {
if (RsDiff < 8) {
if (Rs < 1454) {
return 8;
}
else if (Rs >= 1454) {
return 2;
}
}
else if (RsDiff >= 8) {
return 1;
}
}
else if (Cs >= 1094) {
if (Rs < 1149) {
return 1;
}
else if (Rs >= 1149) {
return 8;
}
}
}
}
}
}
}
}
else if (AFD >= 19) {
if (gchDC < 5) {
if (RsCsDiff < 129) {
if (Rs < 1158) {
if (posBalance < 61) {
if (MVDiff < 71) {
return 4;
}
else if (MVDiff >= 71) {
return 8;
}
}
else if (posBalance >= 61) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
return 4;
}
}
}
else if (Rs >= 1158) {
if (mcTcor < 85) {
if (Rs < 1685) {
if (Rs < 1230) {
return 2;
}
else if (Rs >= 1230) {
return 8;
}
}
else if (Rs >= 1685) {
return 4;
}
}
else if (mcTcor >= 85) {
if (diffRsCsdiff < -18) {
if (CsDiff < 7) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (Rs < 1677) {
return 8;
}
else if (Rs >= 1677) {
return 1;
}
}
}
else if (CsDiff >= 7) {
if (diffTSC < -1) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (diffRsCsdiff >= -18) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 129) {
if (AFD < 23) {
if (posBalance < 65) {
if (negBalance < 52) {
return 4;
}
else if (negBalance >= 52) {
if (AFD < 21) {
if (diffAFD < 2) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (AFD >= 21) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 1;
}
}
}
}
else if (posBalance >= 65) {
if (diffMVdiffVal < 7) {
return 8;
}
else if (diffMVdiffVal >= 7) {
return 2;
}
}
}
else if (AFD >= 23) {
if (mcTcor < 81) {
if (Rs < 1356) {
return 8;
}
else if (Rs >= 1356) {
return 1;
}
}
else if (mcTcor >= 81) {
if (SC < 2842) {
if (Cs < 947) {
if (diffRsCsdiff < 14) {
return 4;
}
else if (diffRsCsdiff >= 14) {
return 2;
}
}
else if (Cs >= 947) {
if (RsCsDiff < 217) {
if (tcor < 40) {
return 2;
}
else if (tcor >= 40) {
if (tcor < 79) {
return 4;
}
else if (tcor >= 79) {
return 1;
}
}
}
else if (RsCsDiff >= 217) {
return 8;
}
}
}
else if (SC >= 2842) {
if (RsCsDiff < 215) {
if (Cs < 1188) {
return 8;
}
else if (Cs >= 1188) {
return 2;
}
}
else if (RsCsDiff >= 215) {
if (RsCsDiff < 264) {
return 4;
}
else if (RsCsDiff >= 264) {
return 2;
}
}
}
}
}
}
}
else if (gchDC >= 5) {
if (Cs < 1182) {
if (mu_mv_mag_sq < 17) {
if (diffAFD < 10) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (tcor < 82) {
if (Rs < 889) {
return 8;
}
else if (Rs >= 889) {
return 2;
}
}
else if (tcor >= 82) {
return 8;
}
}
}
else if (diffAFD >= 10) {
if (Rs < 916) {
return 2;
}
else if (Rs >= 916) {
if (Rs < 1116) {
return 1;
}
else if (Rs >= 1116) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (CsDiff < 11) {
return 1;
}
else if (CsDiff >= 11) {
return 8;
}
}
}
else if (Cs >= 1182) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
if (RsCsDiff < 195) {
if (Rs < 1275) {
return 1;
}
else if (Rs >= 1275) {
return 8;
}
}
else if (RsCsDiff >= 195) {
if (Cs < 1213) {
return 2;
}
else if (Cs >= 1213) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 3033) {
if (ssDCval < 45) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 2;
}
}
else if (ssDCval >= 45) {
if (gchDC < 4) {
if (Rs < 2216) {
if (MVDiff < 21) {
if (tcor < 91) {
if (mcTcor < 87) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
}
else if (mcTcor >= 87) {
if (tcor < 80) {
if (posBalance < 60) {
if (tcor < 79) {
return 2;
}
else if (tcor >= 79) {
if (Rs < 2019) {
return 2;
}
else if (Rs >= 2019) {
return 4;
}
}
}
else if (posBalance >= 60) {
if (mu_mv_mag_sq < 3) {
if (Cs < 1249) {
return 8;
}
else if (Cs >= 1249) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (tcor >= 80) {
if (TSC < 9) {
if (ssDCval < 83) {
if (Rs < 2178) {
if (ssDCval < 55) {
if (Rs < 2139) {
return 4;
}
else if (Rs >= 2139) {
return 8;
}
}
else if (ssDCval >= 55) {
if (negBalance < 53) {
return 8;
}
else if (negBalance >= 53) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
else if (Rs >= 2178) {
return 2;
}
}
else if (ssDCval >= 83) {
if (Rs < 2122) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 8;
}
}
else if (Rs >= 2122) {
if (Rs < 2170) {
return 4;
}
else if (Rs >= 2170) {
if (Rs < 2198) {
return 2;
}
else if (Rs >= 2198) {
return 8;
}
}
}
}
}
else if (TSC >= 9) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 1;
}
}
}
}
}
else if (tcor >= 91) {
if (Cs < 1231) {
if (ssDCval < 83) {
if (Rs < 2142) {
return 8;
}
else if (Rs >= 2142) {
return 4;
}
}
else if (ssDCval >= 83) {
return 4;
}
}
else if (Cs >= 1231) {
if (Cs < 1301) {
return 8;
}
else if (Cs >= 1301) {
return 1;
}
}
}
}
else if (MVDiff >= 21) {
if (gchDC < 2) {
if (tcor < 63) {
if (RsDiff < 12) {
if (Rs < 1880) {
return 2;
}
else if (Rs >= 1880) {
return 1;
}
}
else if (RsDiff >= 12) {
return 4;
}
}
else if (tcor >= 63) {
if (Rs < 1944) {
if (Rs < 1845) {
return 2;
}
else if (Rs >= 1845) {
return 8;
}
}
else if (Rs >= 1944) {
if (Cs < 1044) {
if (Rs < 2117) {
return 2;
}
else if (Rs >= 2117) {
return 4;
}
}
else if (Cs >= 1044) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
if (RsCsDiff < 221) {
if (Rs < 2058) {
return 8;
}
else if (Rs >= 2058) {
if (diffAFD < -1) {
if (Rs < 2129) {
return 4;
}
else if (Rs >= 2129) {
return 8;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (RsCsDiff >= 221) {
if (RsCsDiff < 247) {
return 2;
}
else if (RsCsDiff >= 247) {
return 8;
}
}
}
}
}
else if (Rs >= 2216) {
if (MVDiff < 26) {
if (mcTcor < 93) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
if (posBalance < 41) {
if (Rs < 2406) {
return 2;
}
else if (Rs >= 2406) {
return 4;
}
}
else if (posBalance >= 41) {
if (RsDiff < 11) {
if (ssDCval < 107) {
return 4;
}
else if (ssDCval >= 107) {
return 8;
}
}
else if (RsDiff >= 11) {
if (ssDCval < 69) {
if (Cs < 1083) {
return 2;
}
else if (Cs >= 1083) {
return 8;
}
}
else if (ssDCval >= 69) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 104) {
if (Cs < 1064) {
if (posBalance < 44) {
if (Rs < 2462) {
return 1;
}
else if (Rs >= 2462) {
return 2;
}
}
else if (posBalance >= 44) {
return 2;
}
}
else if (Cs >= 1064) {
if (refDCval < 72) {
return 2;
}
else if (refDCval >= 72) {
if (Rs < 2414) {
return 4;
}
else if (Rs >= 2414) {
return 2;
}
}
}
}
else if (ssDCval >= 104) {
if (Rs < 2407) {
return 8;
}
else if (Rs >= 2407) {
return 1;
}
}
}
}
else if (MVDiff >= 26) {
if (ssDCval < 108) {
if (MVDiff < 30) {
if (Cs < 1164) {
if (Rs < 2343) {
return 8;
}
else if (Rs >= 2343) {
return 1;
}
}
else if (Cs >= 1164) {
return 4;
}
}
else if (MVDiff >= 30) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
if (AFD < 22) {
if (RsDiff < 8) {
if (Rs < 2449) {
if (Cs < 993) {
return 1;
}
else if (Cs >= 993) {
return 8;
}
}
else if (Rs >= 2449) {
return 2;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (AFD >= 22) {
if (MVDiff < 61) {
if (Rs < 2310) {
return 4;
}
else if (Rs >= 2310) {
if (Rs < 2397) {
return 1;
}
else if (Rs >= 2397) {
return 8;
}
}
}
else if (MVDiff >= 61) {
return 2;
}
}
}
}
}
else if (ssDCval >= 108) {
if (Cs < 924) {
return 8;
}
else if (Cs >= 924) {
return 4;
}
}
}
}
}
else if (gchDC >= 4) {
if (RsCsDiff < 141) {
if (Rs < 2158) {
if (Cs < 1066) {
return 8;
}
else if (Cs >= 1066) {
return 4;
}
}
else if (Rs >= 2158) {
return 8;
}
}
else if (RsCsDiff >= 141) {
if (Rs < 2471) {
if (SC < 3255) {
if (Rs < 1784) {
return 8;
}
else if (Rs >= 1784) {
return 2;
}
}
else if (SC >= 3255) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
return 8;
}
}
}
else if (Rs >= 2471) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 114) {
if (MVDiff < 24) {
if (mcTcor < 95) {
if (posBalance < 66) {
if (ssDCval < 120) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 1;
}
}
}
else if (ssDCval >= 120) {
if (negBalance < 35) {
if (Cs < 1032) {
return 8;
}
else if (Cs >= 1032) {
return 1;
}
}
else if (negBalance >= 35) {
return 8;
}
}
}
else if (posBalance >= 66) {
if (Rs < 1698) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
else if (Rs >= 1698) {
return 2;
}
}
}
else if (mcTcor >= 95) {
if (TSC < 9) {
if (RsDiff < 10) {
if (MVDiff < 5) {
if (Rs < 2260) {
return 2;
}
else if (Rs >= 2260) {
return 8;
}
}
else if (MVDiff >= 5) {
if (Rs < 1358) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 8;
}
}
else if (Rs >= 1358) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
return 8;
}
}
}
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (TSC >= 9) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
else if (MVDiff >= 24) {
if (refDCval < 137) {
if (diffRsCsdiff < -50) {
if (mcTcor < 90) {
if (mu_mv_mag_sq < 16) {
if (Rs < 1423) {
return 8;
}
else if (Rs >= 1423) {
return 4;
}
}
else if (mu_mv_mag_sq >= 16) {
return 1;
}
}
else if (mcTcor >= 90) {
if (diffRsCsdiff < -110) {
return 8;
}
else if (diffRsCsdiff >= -110) {
return 2;
}
}
}
else if (diffRsCsdiff >= -50) {
if (gchDC < 11) {
if (RsCsDiff < 99) {
if (SC < 2591) {
if (Rs < 1515) {
return 4;
}
else if (Rs >= 1515) {
return 1;
}
}
else if (SC >= 2591) {
return 8;
}
}
else if (RsCsDiff >= 99) {
if (negBalance < 67) {
if (Cs < 1306) {
return 1;
}
else if (Cs >= 1306) {
if (Rs < 1585) {
return 8;
}
else if (Rs >= 1585) {
return 1;
}
}
}
else if (negBalance >= 67) {
if (Rs < 1905) {
return 4;
}
else if (Rs >= 1905) {
return 8;
}
}
}
}
else if (gchDC >= 11) {
if (Cs < 1216) {
if (Rs < 1000) {
return 1;
}
else if (Rs >= 1000) {
if (Rs < 1703) {
return 2;
}
else if (Rs >= 1703) {
return 8;
}
}
}
else if (Cs >= 1216) {
return 4;
}
}
}
}
else if (refDCval >= 137) {
if (AFD < 29) {
if (negBalance < 27) {
return 2;
}
else if (negBalance >= 27) {
if (ssDCval < 155) {
if (Cs < 1076) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (Rs < 1180) {
return 8;
}
else if (Rs >= 1180) {
return 4;
}
}
}
else if (Cs >= 1076) {
return 8;
}
}
else if (ssDCval >= 155) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
}
}
else if (AFD >= 29) {
if (Rs < 1908) {
if (Rs < 1180) {
return 2;
}
else if (Rs >= 1180) {
return 4;
}
}
else if (Rs >= 1908) {
return 1;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (Rs < 1778) {
if (Cs < 1081) {
if (SC < 1910) {
if (mu_mv_mag_sq < 121) {
if (TSCindex < 2) {
if (Cs < 709) {
if (posBalance < 37) {
if (negBalance < 16) {
return 1;
}
else if (negBalance >= 16) {
if (Rs < 465) {
return 1;
}
else if (Rs >= 465) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (posBalance >= 37) {
return 8;
}
}
else if (Cs >= 709) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 1;
}
}
}
else if (TSCindex >= 2) {
if (Rs < 946) {
if (ssDCval < 85) {
if (Cs < 921) {
if (mu_mv_mag_sq < 83) {
if (negBalance < 35) {
if (Rs < 752) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (TSC < 5) {
if (mu_mv_mag_sq < 76) {
if (AFD < 7) {
if (tcor < 93) {
if (Rs < 485) {
return 1;
}
else if (Rs >= 485) {
if (Rs < 719) {
return 8;
}
else if (Rs >= 719) {
return 4;
}
}
}
else if (tcor >= 93) {
return 4;
}
}
else if (AFD >= 7) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (posBalance < 51) {
if (Rs < 327) {
return 1;
}
else if (Rs >= 327) {
return 4;
}
}
else if (posBalance >= 51) {
return 1;
}
}
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (Rs >= 752) {
if (Cs < 731) {
if (MVDiff < 56) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
else if (MVDiff >= 56) {
return 2;
}
}
else if (Cs >= 731) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 847) {
return 4;
}
else if (Rs >= 847) {
return 1;
}
}
}
}
}
else if (negBalance >= 35) {
if (MVDiff < 4) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
if (Cs < 585) {
return 4;
}
else if (Cs >= 585) {
if (diffRsCsdiff < 3) {
if (posBalance < 46) {
if (Rs < 757) {
return 4;
}
else if (Rs >= 757) {
return 8;
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (diffRsCsdiff >= 3) {
return 1;
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 590) {
if (TSC < 5) {
if (Cs < 553) {
if (Rs < 186) {
return 1;
}
else if (Rs >= 186) {
return 4;
}
}
else if (Cs >= 553) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 8;
}
}
}
else if (TSC >= 5) {
if (ssDCval < 70) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (SC < 1237) {
return 2;
}
else if (SC >= 1237) {
return 4;
}
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
else if (Rs >= 590) {
if (Cs < 673) {
if (MVDiff < 17) {
if (mu_mv_mag_sq < 20) {
if (Rs < 703) {
return 4;
}
else if (Rs >= 703) {
return 8;
}
}
else if (mu_mv_mag_sq >= 20) {
if (Rs < 871) {
return 8;
}
else if (Rs >= 871) {
return 1;
}
}
}
else if (MVDiff >= 17) {
return 1;
}
}
else if (Cs >= 673) {
if (ssDCval < 58) {
if (Rs < 850) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 2;
}
}
else if (Rs >= 850) {
return 1;
}
}
else if (ssDCval >= 58) {
if (SC < 1647) {
if (Rs < 682) {
if (Rs < 647) {
return 8;
}
else if (Rs >= 647) {
return 2;
}
}
else if (Rs >= 682) {
if (Cs < 848) {
return 4;
}
else if (Cs >= 848) {
return 2;
}
}
}
else if (SC >= 1647) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 735) {
if (ssDCval < 51) {
if (ssDCval < 43) {
if (Rs < 618) {
return 4;
}
else if (Rs >= 618) {
return 2;
}
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (ssDCval >= 51) {
if (ssDCval < 66) {
if (Rs < 531) {
if (Rs < 435) {
if (tcor < 99) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (tcor >= 99) {
return 4;
}
}
else if (Rs >= 435) {
if (Cs < 606) {
if (Cs < 556) {
return 4;
}
else if (Cs >= 556) {
return 2;
}
}
else if (Cs >= 606) {
return 4;
}
}
}
else if (Rs >= 531) {
if (Rs < 642) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (Rs >= 642) {
if (MVDiff < 7) {
if (posBalance < 53) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Cs < 743) {
return 4;
}
else if (Cs >= 743) {
if (Cs < 861) {
return 8;
}
else if (Cs >= 861) {
return 4;
}
}
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
if (Rs < 493) {
if (TSC < 3) {
if (Rs < 279) {
return 1;
}
else if (Rs >= 279) {
return 4;
}
}
else if (TSC >= 3) {
if (Rs < 431) {
return 8;
}
else if (Rs >= 431) {
return 2;
}
}
}
else if (Rs >= 493) {
if (CsDiff < 1) {
if (ssDCval < 76) {
return 8;
}
else if (ssDCval >= 76) {
return 1;
}
}
else if (CsDiff >= 1) {
if (Rs < 646) {
if (negBalance < 32) {
if (SC < 1222) {
return 1;
}
else if (SC >= 1222) {
if (Cs < 676) {
return 8;
}
else if (Cs >= 676) {
return 2;
}
}
}
else if (negBalance >= 32) {
return 8;
}
}
else if (Rs >= 646) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 735) {
if (Cs < 761) {
if (TSCindex < 3) {
if (SC < 1568) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
return 2;
}
}
else if (SC >= 1568) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
return 2;
}
}
}
else if (TSCindex >= 3) {
if (ssDCval < 65) {
if (negBalance < 40) {
if (mu_mv_mag_sq < 108) {
if (Cs < 724) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 724) {
return 4;
}
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 903) {
return 8;
}
else if (Rs >= 903) {
return 1;
}
}
}
else if (negBalance >= 40) {
return 4;
}
}
else if (ssDCval >= 65) {
if (MVDiff < 7) {
if (Rs < 807) {
if (posBalance < 49) {
if (Rs < 745) {
return 1;
}
else if (Rs >= 745) {
return 2;
}
}
else if (posBalance >= 49) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 4;
}
}
}
else if (Rs >= 807) {
return 2;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
else if (Cs >= 761) {
if (SC < 1789) {
if (ssDCval < 74) {
if (ssDCval < 59) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (ssDCval >= 74) {
if (Rs < 853) {
if (Cs < 817) {
return 1;
}
else if (Cs >= 817) {
return 2;
}
}
else if (Rs >= 853) {
return 8;
}
}
}
else if (SC >= 1789) {
if (posBalance < 39) {
if (Rs < 937) {
return 8;
}
else if (Rs >= 937) {
return 4;
}
}
else if (posBalance >= 39) {
return 1;
}
}
}
}
}
}
else if (Cs >= 921) {
if (Rs < 744) {
if (Cs < 1021) {
if (diffTSC < -1) {
if (Rs < 609) {
return 2;
}
else if (Rs >= 609) {
return 4;
}
}
else if (diffTSC >= -1) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
return 8;
}
}
}
else if (Cs >= 1021) {
if (mu_mv_mag_sq < 4) {
return 2;
}
else if (mu_mv_mag_sq >= 4) {
if (ssDCval < 53) {
if (Rs < 486) {
return 2;
}
else if (Rs >= 486) {
return 1;
}
}
else if (ssDCval >= 53) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
}
}
}
else if (Rs >= 744) {
if (mu_mv_mag_sq < 79) {
if (ssDCval < 68) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
else if (ssDCval >= 68) {
return 2;
}
}
else if (mu_mv_mag_sq >= 79) {
if (AFD < 4) {
if (ssDCval < 67) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
else if (ssDCval >= 67) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
else if (AFD >= 4) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 85) {
if (CsDiff < 2) {
if (Rs < 644) {
if (Cs < 745) {
if (mu_mv_mag_sq < 87) {
return 2;
}
else if (mu_mv_mag_sq >= 87) {
if (Rs < 580) {
return 2;
}
else if (Rs >= 580) {
return 1;
}
}
}
else if (Cs >= 745) {
return 2;
}
}
else if (Rs >= 644) {
if (Rs < 724) {
return 1;
}
else if (Rs >= 724) {
return 4;
}
}
}
else if (CsDiff >= 2) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
if (SC < 1327) {
return 1;
}
else if (SC >= 1327) {
return 4;
}
}
}
}
}
else if (Rs >= 946) {
if (refDCval < 76) {
if (negBalance < 40) {
if (Cs < 672) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (ssDCval < 72) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
if (negBalance < 33) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
return 2;
}
}
else if (negBalance >= 33) {
if (Rs < 1211) {
return 1;
}
else if (Rs >= 1211) {
return 2;
}
}
}
}
else if (ssDCval >= 72) {
return 1;
}
}
}
else if (Cs >= 672) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
if (TSC < 4) {
if (MVDiff < 10) {
if (SC < 1739) {
if (Rs < 1025) {
return 1;
}
else if (Rs >= 1025) {
return 2;
}
}
else if (SC >= 1739) {
return 1;
}
}
else if (MVDiff >= 10) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (TSC >= 4) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
return 8;
}
}
}
}
}
else if (negBalance >= 40) {
if (posBalance < 48) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (posBalance >= 48) {
if (MVDiff < 10) {
if (Rs < 952) {
return 1;
}
else if (Rs >= 952) {
return 8;
}
}
else if (MVDiff >= 10) {
if (Rs < 988) {
return 1;
}
else if (Rs >= 988) {
return 2;
}
}
}
}
}
else if (refDCval >= 76) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (ssDCval < 64) {
if (Rs < 532) {
if (Cs < 676) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (Cs >= 676) {
if (Cs < 894) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
return 4;
}
}
else if (Cs >= 894) {
if (Rs < 370) {
return 2;
}
else if (Rs >= 370) {
return 4;
}
}
}
}
else if (Rs >= 532) {
if (SC < 1664) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (SC >= 1664) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (Rs < 754) {
if (Rs < 666) {
return 4;
}
else if (Rs >= 666) {
return 1;
}
}
else if (Rs >= 754) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 50) {
if (Rs < 789) {
return 1;
}
else if (Rs >= 789) {
return 8;
}
}
else if (ssDCval >= 50) {
return 1;
}
}
else if (mu_mv_mag_sq >= 126) {
if (negBalance < 17) {
if (Rs < 1037) {
return 1;
}
else if (Rs >= 1037) {
return 8;
}
}
else if (negBalance >= 17) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (SC < 1806) {
if (AFD < 1) {
if (Rs < 924) {
if (Cs < 614) {
return 8;
}
else if (Cs >= 614) {
return 2;
}
}
else if (Rs >= 924) {
return 8;
}
}
else if (AFD >= 1) {
if (Rs < 802) {
if (ssDCval < 81) {
if (Rs < 715) {
if (SC < 1152) {
if (Rs < 224) {
return 1;
}
else if (Rs >= 224) {
return 4;
}
}
else if (SC >= 1152) {
return 8;
}
}
else if (Rs >= 715) {
if (Cs < 608) {
return 2;
}
else if (Cs >= 608) {
return 1;
}
}
}
else if (ssDCval >= 81) {
if (Rs < 712) {
return 2;
}
else if (Rs >= 712) {
return 4;
}
}
}
else if (Rs >= 802) {
if (tcor < 99) {
if (Rs < 810) {
return 1;
}
else if (Rs >= 810) {
return 4;
}
}
else if (tcor >= 99) {
if (Cs < 795) {
if (Rs < 1137) {
if (posBalance < 37) {
if (Rs < 1001) {
return 2;
}
else if (Rs >= 1001) {
return 1;
}
}
else if (posBalance >= 37) {
return 2;
}
}
else if (Rs >= 1137) {
if (Rs < 1147) {
return 8;
}
else if (Rs >= 1147) {
return 1;
}
}
}
else if (Cs >= 795) {
return 4;
}
}
}
}
}
else if (SC >= 1806) {
if (Cs < 724) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 8;
}
}
}
else if (Cs >= 724) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
}
}
else if (SC >= 1910) {
if (refDCval < 69) {
if (mu_mv_mag_sq < 128) {
if (SC < 2389) {
if (negBalance < 42) {
if (ssDCval < 52) {
if (posBalance < 40) {
if (Cs < 946) {
if (Rs < 1352) {
if (Cs < 844) {
return 8;
}
else if (Cs >= 844) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (Rs >= 1352) {
if (Rs < 1434) {
return 1;
}
else if (Rs >= 1434) {
return 4;
}
}
}
else if (Cs >= 946) {
return 1;
}
}
else if (posBalance >= 40) {
if (Rs < 879) {
return 1;
}
else if (Rs >= 879) {
if (Cs < 770) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (Rs < 1348) {
return 4;
}
else if (Rs >= 1348) {
return 4;
}
}
}
else if (Cs >= 770) {
return 8;
}
}
}
}
else if (ssDCval >= 52) {
if (Cs < 713) {
if (diffMVdiffVal < -3) {
if (Rs < 1570) {
return 1;
}
else if (Rs >= 1570) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (Cs >= 713) {
if (AFD < 3) {
if (posBalance < 32) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
}
else if (posBalance >= 32) {
if (Rs < 1343) {
return 1;
}
else if (Rs >= 1343) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 1;
}
}
}
}
else if (AFD >= 3) {
if (posBalance < 57) {
if (negBalance < 36) {
if (mu_mv_mag_sq < 88) {
if (mcTcor < 98) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
return 8;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (mu_mv_mag_sq >= 88) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (RsDiff < 2) {
if (tcor < 97) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (tcor >= 97) {
return 1;
}
}
else if (RsDiff >= 2) {
if (Rs < 1056) {
return 4;
}
else if (Rs >= 1056) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 36) {
if (tcor < 95) {
if (Rs < 1285) {
return 8;
}
else if (Rs >= 1285) {
return 4;
}
}
else if (tcor >= 95) {
return 1;
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
}
else if (negBalance >= 42) {
if (RsCsDiff < 45) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1179) {
return 4;
}
else if (Rs >= 1179) {
return 8;
}
}
}
else if (RsCsDiff >= 45) {
if (Rs < 1665) {
return 4;
}
else if (Rs >= 1665) {
return 2;
}
}
}
}
else if (SC >= 2389) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Cs < 800) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (Cs >= 800) {
if (ssDCval < 51) {
if (posBalance < 29) {
return 1;
}
else if (posBalance >= 29) {
if (mu_mv_mag_sq < 126) {
return 4;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 1673) {
return 8;
}
else if (Rs >= 1673) {
return 8;
}
}
}
}
else if (ssDCval >= 51) {
if (Rs < 1769) {
return 8;
}
else if (Rs >= 1769) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (tcor < 99) {
if (Rs < 1616) {
return 8;
}
else if (Rs >= 1616) {
if (Rs < 1734) {
return 8;
}
else if (Rs >= 1734) {
return 4;
}
}
}
else if (tcor >= 99) {
if (ssDCval < 58) {
if (Rs < 1738) {
return 8;
}
else if (Rs >= 1738) {
return 8;
}
}
else if (ssDCval >= 58) {
if (Cs < 922) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
return 1;
}
}
}
else if (Cs >= 922) {
if (SC < 2308) {
return 1;
}
else if (SC >= 2308) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 69) {
if (mu_mv_mag_sq < 66) {
if (Rs < 1253) {
if (Rs < 1113) {
return 2;
}
else if (Rs >= 1113) {
return 8;
}
}
else if (Rs >= 1253) {
if (TSC < 8) {
if (MVDiff < 21) {
return 1;
}
else if (MVDiff >= 21) {
if (Rs < 1407) {
return 4;
}
else if (Rs >= 1407) {
return 1;
}
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (Cs < 642) {
return 8;
}
else if (Cs >= 642) {
if (negBalance < 30) {
if (SC < 2166) {
if (Cs < 968) {
if (CsDiff < 2) {
if (ssDCval < 76) {
if (Rs < 1466) {
return 1;
}
else if (Rs >= 1466) {
return 8;
}
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (Cs >= 968) {
if (refDCval < 79) {
return 8;
}
else if (refDCval >= 79) {
return 4;
}
}
}
else if (SC >= 2166) {
if (posBalance < 41) {
if (Cs < 933) {
return 8;
}
else if (Cs >= 933) {
return 1;
}
}
else if (posBalance >= 41) {
return 8;
}
}
}
else if (negBalance >= 30) {
if (Cs < 716) {
if (Rs < 1612) {
return 4;
}
else if (Rs >= 1612) {
return 1;
}
}
else if (Cs >= 716) {
if (Rs < 1108) {
if (posBalance < 48) {
if (Rs < 1102) {
return 8;
}
else if (Rs >= 1102) {
return 4;
}
}
else if (posBalance >= 48) {
if (ssDCval < 79) {
if (Rs < 898) {
return 1;
}
else if (Rs >= 898) {
if (Rs < 1002) {
return 8;
}
else if (Rs >= 1002) {
return 4;
}
}
}
else if (ssDCval >= 79) {
return 4;
}
}
}
else if (Rs >= 1108) {
if (negBalance < 69) {
if (Rs < 1571) {
if (Cs < 1077) {
return 8;
}
else if (Cs >= 1077) {
return 4;
}
}
else if (Rs >= 1571) {
if (Cs < 1043) {
if (Rs < 1711) {
return 2;
}
else if (Rs >= 1711) {
return 4;
}
}
else if (Cs >= 1043) {
return 8;
}
}
}
else if (negBalance >= 69) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Cs >= 1081) {
if (RsDiff < 1) {
if (Rs < 601) {
return 8;
}
else if (Rs >= 601) {
if (mu_mv_mag_sq < 104) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (mu_mv_mag_sq >= 104) {
if (ssDCval < 55) {
if (Rs < 950) {
return 1;
}
else if (Rs >= 950) {
return 8;
}
}
else if (ssDCval >= 55) {
if (ssDCval < 80) {
if (Rs < 1436) {
return 1;
}
else if (Rs >= 1436) {
return 4;
}
}
else if (ssDCval >= 80) {
return 4;
}
}
}
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 62) {
if (negBalance < 46) {
if (ssDCval < 60) {
if (Rs < 1284) {
return 4;
}
else if (Rs >= 1284) {
return 2;
}
}
else if (ssDCval >= 60) {
if (Cs < 1291) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (Cs >= 1291) {
return 1;
}
}
}
else if (negBalance >= 46) {
if (CsDiff < 5) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 1;
}
}
else if (CsDiff >= 5) {
if (CsDiff < 9) {
if (Rs < 778) {
return 2;
}
else if (Rs >= 778) {
return 4;
}
}
else if (CsDiff >= 9) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 62) {
if (MVDiff < 10) {
if (mu_mv_mag_sq < 124) {
if (refDCval < 76) {
if (Rs < 1346) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 106) {
if (Rs < 905) {
if (ssDCval < 68) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (Rs >= 905) {
return 8;
}
}
else if (mu_mv_mag_sq >= 106) {
if (ssDCval < 56) {
if (mu_mv_mag_sq < 121) {
if (Rs < 934) {
return 1;
}
else if (Rs >= 934) {
return 8;
}
}
else if (mu_mv_mag_sq >= 121) {
if (ssDCval < 51) {
if (Rs < 959) {
return 1;
}
else if (Rs >= 959) {
return 8;
}
}
else if (ssDCval >= 51) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (ssDCval < 70) {
if (Rs < 988) {
if (Rs < 789) {
return 1;
}
else if (Rs >= 789) {
return 2;
}
}
else if (Rs >= 988) {
return 1;
}
}
else if (ssDCval >= 70) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
}
else if (RsDiff >= 2) {
if (diffRsCsdiff < 2) {
if (ssDCval < 71) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (Rs < 860) {
if (Cs < 1181) {
if (Rs < 799) {
return 4;
}
else if (Rs >= 799) {
return 2;
}
}
else if (Cs >= 1181) {
return 8;
}
}
else if (Rs >= 860) {
if (Rs < 1292) {
return 4;
}
else if (Rs >= 1292) {
if (Rs < 1309) {
return 8;
}
else if (Rs >= 1309) {
return 4;
}
}
}
}
}
else if (ssDCval >= 71) {
if (Rs < 993) {
return 2;
}
else if (Rs >= 993) {
return 8;
}
}
}
else if (diffRsCsdiff >= 2) {
if (mu_mv_mag_sq < 102) {
if (Cs < 1243) {
return 8;
}
else if (Cs >= 1243) {
return 4;
}
}
else if (mu_mv_mag_sq >= 102) {
if (diffAFD < 1) {
if (Rs < 989) {
if (Rs < 764) {
return 4;
}
else if (Rs >= 764) {
return 2;
}
}
else if (Rs >= 989) {
if (Cs < 1209) {
return 8;
}
else if (Cs >= 1209) {
return 1;
}
}
}
else if (diffAFD >= 1) {
if (Cs < 1269) {
return 1;
}
else if (Cs >= 1269) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1346) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 1759) {
if (Cs < 1120) {
return 8;
}
else if (Cs >= 1120) {
return 1;
}
}
else if (Rs >= 1759) {
return 2;
}
}
}
}
else if (refDCval >= 76) {
if (diffRsCsdiff < 4) {
if (posBalance < 45) {
if (Rs < 1289) {
return 8;
}
else if (Rs >= 1289) {
return 4;
}
}
else if (posBalance >= 45) {
return 4;
}
}
else if (diffRsCsdiff >= 4) {
if (negBalance < 37) {
return 8;
}
else if (negBalance >= 37) {
if (Rs < 1457) {
return 4;
}
else if (Rs >= 1457) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (ssDCval < 74) {
if (negBalance < 35) {
if (refDCval < 51) {
if (Rs < 1504) {
if (Rs < 904) {
return 1;
}
else if (Rs >= 904) {
return 8;
}
}
else if (Rs >= 1504) {
if (SC < 2952) {
if (posBalance < 42) {
if (refDCval < 47) {
return 4;
}
else if (refDCval >= 47) {
if (posBalance < 35) {
if (SC < 2770) {
return 4;
}
else if (SC >= 2770) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (diffRsCsdiff < 1) {
if (posBalance < 34) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (Rs < 1766) {
if (negBalance < 19) {
return 8;
}
else if (negBalance >= 19) {
if (posBalance < 33) {
if (Cs < 1155) {
if (SC < 2873) {
return 4;
}
else if (SC >= 2873) {
return 4;
}
}
else if (Cs >= 1155) {
return 8;
}
}
else if (posBalance >= 33) {
if (Rs < 1709) {
return 8;
}
else if (Rs >= 1709) {
return 8;
}
}
}
}
else if (Rs >= 1766) {
return 8;
}
}
}
else if (posBalance >= 34) {
if (Cs < 1151) {
if (Rs < 1701) {
return 4;
}
else if (Rs >= 1701) {
return 4;
}
}
else if (Cs >= 1151) {
return 4;
}
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 1664) {
return 4;
}
else if (Rs >= 1664) {
return 8;
}
}
}
}
}
else if (posBalance >= 35) {
if (posBalance < 36) {
return 4;
}
else if (posBalance >= 36) {
if (SC < 2841) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
if (Rs < 1596) {
return 4;
}
else if (Rs >= 1596) {
if (Rs < 1669) {
return 4;
}
else if (Rs >= 1669) {
if (Rs < 1691) {
return 4;
}
else if (Rs >= 1691) {
return 4;
}
}
}
}
}
else if (SC >= 2841) {
if (TSCindex < 3) {
if (Rs < 1702) {
return 4;
}
else if (Rs >= 1702) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
if (Cs < 1120) {
return 4;
}
else if (Cs >= 1120) {
if (SC < 2884) {
if (Rs < 1735) {
if (Rs < 1717) {
return 4;
}
else if (Rs >= 1717) {
return 8;
}
}
else if (Rs >= 1735) {
return 8;
}
}
else if (SC >= 2884) {
if (Cs < 1135) {
return 4;
}
else if (Cs >= 1135) {
if (Rs < 1766) {
if (Rs < 1752) {
return 4;
}
else if (Rs >= 1752) {
return 4;
}
}
else if (Rs >= 1766) {
return 4;
}
}
}
}
}
}
}
else if (TSCindex >= 3) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 42) {
return 8;
}
}
else if (SC >= 2952) {
return 4;
}
}
}
else if (refDCval >= 51) {
if (posBalance < 47) {
if (Cs < 1104) {
return 8;
}
else if (Cs >= 1104) {
return 4;
}
}
else if (posBalance >= 47) {
if (Cs < 1225) {
return 4;
}
else if (Cs >= 1225) {
if (Cs < 1226) {
return 1;
}
else if (Cs >= 1226) {
return 4;
}
}
}
}
}
else if (negBalance >= 35) {
if (Rs < 1518) {
return 1;
}
else if (Rs >= 1518) {
return 8;
}
}
}
else if (ssDCval >= 74) {
return 8;
}
}
}
else if (MVDiff >= 10) {
if (Rs < 1271) {
if (Rs < 814) {
return 1;
}
else if (Rs >= 814) {
return 4;
}
}
else if (Rs >= 1271) {
if (Cs < 1264) {
return 8;
}
else if (Cs >= 1264) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1778) {
if (TSC < 4) {
if (refDCval < 83) {
if (ssDCval < 56) {
if (Rs < 2438) {
if (mu_mv_mag_sq < 103) {
return 4;
}
else if (mu_mv_mag_sq >= 103) {
if (Cs < 935) {
return 8;
}
else if (Cs >= 935) {
if (SC < 2895) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Rs < 1780) {
return 4;
}
else if (Rs >= 1780) {
if (Rs < 1849) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (Rs < 1813) {
if (Rs < 1782) {
return 8;
}
else if (Rs >= 1782) {
if (AFD < 2) {
if (Rs < 1805) {
return 4;
}
else if (Rs >= 1805) {
return 8;
}
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (Rs >= 1813) {
return 8;
}
}
}
}
else if (Rs >= 1849) {
return 4;
}
}
}
}
else if (SC >= 2895) {
if (Rs < 2177) {
if (mu_mv_mag_sq < 128) {
return 4;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1799) {
return 4;
}
else if (Rs >= 1799) {
if (Rs < 1831) {
if (Cs < 1141) {
return 4;
}
else if (Cs >= 1141) {
if (Rs < 1811) {
return 4;
}
else if (Rs >= 1811) {
return 4;
}
}
}
else if (Rs >= 1831) {
return 4;
}
}
}
}
else if (Rs >= 2177) {
return 8;
}
}
}
}
}
else if (Rs >= 2438) {
if (Cs < 691) {
return 1;
}
else if (Cs >= 691) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (posBalance < 47) {
if (SC < 3028) {
if (ssDCval < 68) {
if (MVDiff < 3) {
if (Cs < 955) {
return 1;
}
else if (Cs >= 955) {
if (tcor < 99) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Cs < 1077) {
return 4;
}
else if (Cs >= 1077) {
return 1;
}
}
}
else if (tcor >= 99) {
if (Cs < 1055) {
if (refDCval < 57) {
if (Rs < 1942) {
return 1;
}
else if (Rs >= 1942) {
return 4;
}
}
else if (refDCval >= 57) {
return 4;
}
}
else if (Cs >= 1055) {
if (mu_mv_mag_sq < 125) {
if (Cs < 1106) {
return 4;
}
else if (Cs >= 1106) {
return 1;
}
}
else if (mu_mv_mag_sq >= 125) {
return 1;
}
}
}
}
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (SC >= 3028) {
if (Rs < 2018) {
if (posBalance < 40) {
if (Rs < 1887) {
return 2;
}
else if (Rs >= 1887) {
return 4;
}
}
else if (posBalance >= 40) {
if (Rs < 1933) {
return 1;
}
else if (Rs >= 1933) {
return 8;
}
}
}
else if (Rs >= 2018) {
if (Rs < 2287) {
if (Rs < 2096) {
if (Rs < 2024) {
return 8;
}
else if (Rs >= 2024) {
return 1;
}
}
else if (Rs >= 2096) {
return 8;
}
}
else if (Rs >= 2287) {
if (Cs < 819) {
return 4;
}
else if (Cs >= 819) {
return 2;
}
}
}
}
}
else if (posBalance >= 47) {
if (Cs < 871) {
if (Cs < 680) {
if (refDCval < 82) {
return 1;
}
else if (refDCval >= 82) {
return 8;
}
}
else if (Cs >= 680) {
return 4;
}
}
else if (Cs >= 871) {
return 8;
}
}
}
}
else if (refDCval >= 83) {
if (MVDiff < 2) {
if (Cs < 939) {
return 4;
}
else if (Cs >= 939) {
return 8;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (TSC >= 4) {
if (Cs < 1020) {
if (ssDCval < 71) {
if (Rs < 1982) {
if (Cs < 812) {
return 1;
}
else if (Cs >= 812) {
if (Rs < 1832) {
return 8;
}
else if (Rs >= 1832) {
return 4;
}
}
}
else if (Rs >= 1982) {
if (RsCsDiff < 141) {
if (refDCval < 59) {
return 1;
}
else if (refDCval >= 59) {
if (refDCval < 71) {
return 2;
}
else if (refDCval >= 71) {
if (Rs < 2222) {
return 1;
}
else if (Rs >= 2222) {
return 2;
}
}
}
}
else if (RsCsDiff >= 141) {
if (SC < 3085) {
if (Rs < 2187) {
return 1;
}
else if (Rs >= 2187) {
return 2;
}
}
else if (SC >= 3085) {
return 4;
}
}
}
}
else if (ssDCval >= 71) {
if (TSC < 5) {
if (mu_mv_mag_sq < 92) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 92) {
return 4;
}
}
else if (TSC >= 5) {
if (RsCsDiff < 37) {
if (AFD < 7) {
if (Rs < 2368) {
return 8;
}
else if (Rs >= 2368) {
return 2;
}
}
else if (AFD >= 7) {
return 1;
}
}
else if (RsCsDiff >= 37) {
if (tcor < 88) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 2;
}
}
else if (tcor >= 88) {
if (posBalance < 50) {
if (Cs < 709) {
return 2;
}
else if (Cs >= 709) {
if (mu_mv_mag_sq < 54) {
return 1;
}
else if (mu_mv_mag_sq >= 54) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1020) {
if (diffRsCsdiff < -4) {
if (negBalance < 34) {
if (diffTSC < -1) {
if (Rs < 1996) {
return 8;
}
else if (Rs >= 1996) {
return 1;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (negBalance >= 34) {
if (AFD < 8) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (diffRsCsdiff >= -4) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 100) {
if (SC < 1291) {
if (MVDiff < 4) {
if (Cs < 721) {
if (Rs < 523) {
if (SC < 1001) {
if (Rs < 341) {
return 2;
}
else if (Rs >= 341) {
return 8;
}
}
else if (SC >= 1001) {
if (AFD < 12) {
if (mcTcor < 98) {
if (Rs < 420) {
return 1;
}
else if (Rs >= 420) {
return 8;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (AFD >= 12) {
return 4;
}
}
}
else if (Rs >= 523) {
if (ssDCval < 114) {
return 1;
}
else if (ssDCval >= 114) {
if (ssDCval < 124) {
if (Cs < 547) {
return 8;
}
else if (Cs >= 547) {
return 2;
}
}
else if (ssDCval >= 124) {
if (Rs < 613) {
return 8;
}
else if (Rs >= 613) {
return 4;
}
}
}
}
}
else if (Cs >= 721) {
if (Rs < 420) {
return 4;
}
else if (Rs >= 420) {
return 1;
}
}
}
else if (MVDiff >= 4) {
if (negBalance < 72) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
if (TSC < 4) {
if (MVDiff < 6) {
if (Cs < 750) {
if (negBalance < 21) {
return 4;
}
else if (negBalance >= 21) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (Cs >= 750) {
return 4;
}
}
else if (MVDiff >= 6) {
if (Rs < 477) {
return 2;
}
else if (Rs >= 477) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (Rs < 492) {
return 1;
}
else if (Rs >= 492) {
return 2;
}
}
}
}
}
else if (TSC >= 4) {
if (diffRsCsdiff < -1) {
return 2;
}
else if (diffRsCsdiff >= -1) {
if (RsDiff < 2) {
if (Rs < 543) {
return 8;
}
else if (Rs >= 543) {
return 2;
}
}
else if (RsDiff >= 2) {
if (ssDCval < 115) {
if (Cs < 576) {
return 1;
}
else if (Cs >= 576) {
if (Cs < 719) {
return 8;
}
else if (Cs >= 719) {
return 2;
}
}
}
else if (ssDCval >= 115) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 72) {
if (Rs < 597) {
return 8;
}
else if (Rs >= 597) {
return 1;
}
}
}
}
else if (SC >= 1291) {
if (Cs < 694) {
if (Rs < 1818) {
if (Rs < 1043) {
if (AFD < 9) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (AFD >= 9) {
if (refDCval < 110) {
return 1;
}
else if (refDCval >= 110) {
if (negBalance < 33) {
if (Rs < 789) {
return 4;
}
else if (Rs >= 789) {
return 8;
}
}
else if (negBalance >= 33) {
if (ssDCval < 118) {
return 4;
}
else if (ssDCval >= 118) {
return 2;
}
}
}
}
}
else if (Rs >= 1043) {
if (TSC < 8) {
if (posBalance < 64) {
if (Cs < 562) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
return 1;
}
}
else if (Cs >= 562) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1425) {
if (MVDiff < 27) {
return 8;
}
else if (MVDiff >= 27) {
return 1;
}
}
else if (Rs >= 1425) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (posBalance >= 64) {
if (SC < 2238) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
else if (SC >= 2238) {
return 1;
}
}
}
else if (TSC >= 8) {
if (ssDCval < 113) {
if (Rs < 1341) {
return 4;
}
else if (Rs >= 1341) {
if (Rs < 1652) {
return 8;
}
else if (Rs >= 1652) {
return 1;
}
}
}
else if (ssDCval >= 113) {
return 1;
}
}
}
}
else if (Rs >= 1818) {
if (TSC < 3) {
if (Rs < 2016) {
return 4;
}
else if (Rs >= 2016) {
return 8;
}
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 32) {
if (RsDiff < 7) {
if (Rs < 1917) {
return 1;
}
else if (Rs >= 1917) {
if (Rs < 2101) {
return 8;
}
else if (Rs >= 2101) {
return 4;
}
}
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (mu_mv_mag_sq >= 32) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 1;
}
}
}
}
}
}
else if (Cs >= 694) {
if (refDCval < 102) {
if (SC < 3363) {
if (SC < 1540) {
if (SC < 1356) {
if (Rs < 496) {
return 8;
}
else if (Rs >= 496) {
return 2;
}
}
else if (SC >= 1356) {
if (ssDCval < 93) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (ssDCval >= 93) {
if (SC < 1402) {
return 1;
}
else if (SC >= 1402) {
return 2;
}
}
}
}
else if (SC >= 1540) {
if (posBalance < 58) {
if (TSC < 5) {
if (Rs < 1337) {
if (Rs < 1091) {
if (Cs < 1168) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (Cs >= 1168) {
return 4;
}
}
else if (Rs >= 1091) {
return 8;
}
}
else if (Rs >= 1337) {
return 1;
}
}
else if (TSC >= 5) {
if (MVDiff < 4) {
if (Rs < 1134) {
return 4;
}
else if (Rs >= 1134) {
return 1;
}
}
else if (MVDiff >= 4) {
if (Rs < 2202) {
if (tcor < 93) {
if (gchDC < 2) {
if (SC < 2316) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 4;
}
}
else if (SC >= 2316) {
if (SC < 3049) {
return 4;
}
else if (SC >= 3049) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (Rs < 1478) {
return 8;
}
else if (Rs >= 1478) {
return 2;
}
}
}
else if (tcor >= 93) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (diffRsCsdiff < -18) {
return 4;
}
else if (diffRsCsdiff >= -18) {
if (Rs < 854) {
return 4;
}
else if (Rs >= 854) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2202) {
return 4;
}
}
}
}
else if (posBalance >= 58) {
if (MVDiff < 9) {
if (Rs < 1001) {
return 4;
}
else if (Rs >= 1001) {
return 1;
}
}
else if (MVDiff >= 9) {
if (Cs < 975) {
if (Cs < 882) {
if (Cs < 795) {
return 2;
}
else if (Cs >= 795) {
return 8;
}
}
else if (Cs >= 882) {
if (Rs < 826) {
return 1;
}
else if (Rs >= 826) {
return 4;
}
}
}
else if (Cs >= 975) {
return 2;
}
}
}
}
}
else if (SC >= 3363) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 1;
}
}
}
else if (refDCval >= 102) {
if (Rs < 656) {
if (TSC < 4) {
if (posBalance < 41) {
if (Rs < 556) {
return 4;
}
else if (Rs >= 556) {
return 8;
}
}
else if (posBalance >= 41) {
return 1;
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 21) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (ssDCval < 152) {
return 8;
}
else if (ssDCval >= 152) {
return 4;
}
}
}
}
else if (Rs >= 656) {
if (CsDiff < 6) {
if (MVDiff < 5) {
if (SC < 2641) {
if (mu_mv_mag_sq < 82) {
if (SC < 2271) {
if (Rs < 677) {
return 8;
}
else if (Rs >= 677) {
return 4;
}
}
else if (SC >= 2271) {
if (SC < 2528) {
return 8;
}
else if (SC >= 2528) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 82) {
return 8;
}
}
else if (SC >= 2641) {
if (Rs < 2012) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
return 2;
}
}
else if (Rs >= 2012) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (refDCval < 156) {
if (SC < 1641) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
return 2;
}
}
else if (SC >= 1641) {
if (Cs < 866) {
if (diffRsCsdiff < -18) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
else if (diffRsCsdiff >= -18) {
if (negBalance < 36) {
if (Rs < 1671) {
return 4;
}
else if (Rs >= 1671) {
return 8;
}
}
else if (negBalance >= 36) {
if (negBalance < 51) {
if (SC < 1662) {
return 1;
}
else if (SC >= 1662) {
return 8;
}
}
else if (negBalance >= 51) {
if (mu_mv_mag_sq < 3) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (refDCval < 143) {
return 8;
}
else if (refDCval >= 143) {
if (Cs < 797) {
return 8;
}
else if (Cs >= 797) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 866) {
if (negBalance < 27) {
if (Rs < 1229) {
return 8;
}
else if (Rs >= 1229) {
return 1;
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
}
else if (refDCval >= 156) {
if (Cs < 1134) {
return 2;
}
else if (Cs >= 1134) {
return 4;
}
}
}
}
else if (CsDiff >= 6) {
if (tcor < 92) {
if (ssDCval < 142) {
if (Cs < 1214) {
return 2;
}
else if (Cs >= 1214) {
return 8;
}
}
else if (ssDCval >= 142) {
if (Rs < 2169) {
return 8;
}
else if (Rs >= 2169) {
return 2;
}
}
}
else if (tcor >= 92) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (ssDCval < 133) {
if (Cs < 1232) {
if (Cs < 915) {
if (mu_mv_mag_sq < 126) {
if (CsDiff < 1) {
if (Rs < 1581) {
if (SC < 1545) {
if (mu_mv_mag_sq < 113) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
else if (SC >= 1545) {
if (ssDCval < 106) {
return 4;
}
else if (ssDCval >= 106) {
if (Rs < 1018) {
return 4;
}
else if (Rs >= 1018) {
return 8;
}
}
}
}
else if (Rs >= 1581) {
if (Cs < 595) {
return 1;
}
else if (Cs >= 595) {
return 8;
}
}
}
else if (CsDiff >= 1) {
if (Rs < 331) {
if (Rs < 260) {
return 2;
}
else if (Rs >= 260) {
return 4;
}
}
else if (Rs >= 331) {
if (negBalance < 35) {
if (Cs < 643) {
if (mu_mv_mag_sq < 107) {
if (Rs < 438) {
return 4;
}
else if (Rs >= 438) {
return 1;
}
}
else if (mu_mv_mag_sq >= 107) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (mu_mv_mag_sq < 114) {
if (Rs < 433) {
return 8;
}
else if (Rs >= 433) {
return 1;
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
}
}
else if (Cs >= 643) {
if (SC < 2297) {
if (Rs < 394) {
return 4;
}
else if (Rs >= 394) {
if (MVDiff < 3) {
if (diffAFD < -1) {
if (diffRsCsdiff < -1) {
return 8;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 106) {
return 4;
}
else if (mu_mv_mag_sq >= 106) {
if (Cs < 769) {
return 4;
}
else if (Cs >= 769) {
return 8;
}
}
}
}
}
else if (SC >= 2297) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (Rs < 1582) {
return 4;
}
else if (Rs >= 1582) {
return 8;
}
}
}
}
}
else if (negBalance >= 35) {
if (ssDCval < 101) {
if (SC < 1272) {
return 1;
}
else if (SC >= 1272) {
if (mu_mv_mag_sq < 112) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
}
else if (ssDCval >= 101) {
if (Rs < 513) {
if (mu_mv_mag_sq < 104) {
if (AFD < 5) {
if (Rs < 410) {
return 4;
}
else if (Rs >= 410) {
return 1;
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 104) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
if (Cs < 859) {
return 8;
}
else if (Cs >= 859) {
return 4;
}
}
}
}
else if (Rs >= 513) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (posBalance < 44) {
if (posBalance < 25) {
return 4;
}
else if (posBalance >= 25) {
if (SC < 2895) {
return 8;
}
else if (SC >= 2895) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
return 8;
}
}
}
}
else if (posBalance >= 44) {
if (ssDCval < 123) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
if (Rs < 1605) {
if (Rs < 511) {
return 1;
}
else if (Rs >= 511) {
return 8;
}
}
else if (Rs >= 1605) {
return 1;
}
}
}
else if (ssDCval >= 123) {
if (Rs < 696) {
return 2;
}
else if (Rs >= 696) {
return 1;
}
}
}
}
}
else if (Cs >= 915) {
if (mu_mv_mag_sq < 107) {
if (Rs < 1045) {
if (Rs < 679) {
if (Rs < 378) {
return 4;
}
else if (Rs >= 378) {
if (tcor < 98) {
if (Rs < 463) {
return 8;
}
else if (Rs >= 463) {
return 1;
}
}
else if (tcor >= 98) {
return 8;
}
}
}
else if (Rs >= 679) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 8;
}
}
}
else if (Rs >= 1045) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (ssDCval < 122) {
if (Cs < 1222) {
if (diffRsCsdiff < -50) {
if (Rs < 1256) {
return 8;
}
else if (Rs >= 1256) {
return 2;
}
}
else if (diffRsCsdiff >= -50) {
return 8;
}
}
else if (Cs >= 1222) {
if (ssDCval < 95) {
return 2;
}
else if (ssDCval >= 95) {
return 8;
}
}
}
else if (ssDCval >= 122) {
if (Rs < 1489) {
return 8;
}
else if (Rs >= 1489) {
return 2;
}
}
}
}
}
else if (Cs >= 1232) {
if (RsDiff < 1) {
if (Rs < 817) {
return 1;
}
else if (Rs >= 817) {
return 4;
}
}
else if (RsDiff >= 1) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
if (Cs < 1233) {
return 1;
}
else if (Cs >= 1233) {
return 8;
}
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1170) {
return 2;
}
else if (Rs >= 1170) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2525) {
if (MVDiff < 14) {
if (ssDCval < 109) {
if (posBalance < 48) {
if (Rs < 3818) {
if (RsCsDiff < 12) {
if (Cs < 758) {
if (Cs < 666) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 8;
}
}
else if (Cs >= 666) {
return 4;
}
}
else if (Cs >= 758) {
if (Cs < 1227) {
if (negBalance < 23) {
if (RsDiff < 2) {
if (negBalance < 20) {
return 8;
}
else if (negBalance >= 20) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (Rs < 3023) {
return 8;
}
else if (Rs >= 3023) {
return 2;
}
}
}
}
else if (RsDiff >= 2) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 4;
}
}
}
else if (negBalance >= 23) {
if (mu_mv_mag_sq < 124) {
if (Cs < 1016) {
return 2;
}
else if (Cs >= 1016) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (Cs < 1113) {
return 2;
}
else if (Cs >= 1113) {
return 8;
}
}
}
}
else if (Cs >= 1227) {
if (mu_mv_mag_sq < 127) {
if (Rs < 3589) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (Rs >= 3589) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
}
}
else if (RsCsDiff >= 12) {
if (ssDCval < 96) {
if (negBalance < 46) {
if (mu_mv_mag_sq < 57) {
if (ssDCval < 75) {
if (Cs < 1008) {
if (Cs < 684) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
return 8;
}
}
else if (Cs >= 684) {
if (Rs < 2739) {
if (mu_mv_mag_sq < 19) {
return 1;
}
else if (mu_mv_mag_sq >= 19) {
if (Rs < 2577) {
return 2;
}
else if (Rs >= 2577) {
return 4;
}
}
}
else if (Rs >= 2739) {
if (MVDiff < 2) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
else if (MVDiff >= 2) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
if (refDCval < 75) {
return 4;
}
else if (refDCval >= 75) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1008) {
if (ssDCval < 72) {
if (CsDiff < 5) {
if (MVDiff < 5) {
if (Rs < 3569) {
return 2;
}
else if (Rs >= 3569) {
return 4;
}
}
else if (MVDiff >= 5) {
if (Rs < 2626) {
return 8;
}
else if (Rs >= 2626) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
return 4;
}
}
}
else if (ssDCval >= 72) {
if (negBalance < 41) {
if (Rs < 3345) {
return 2;
}
else if (Rs >= 3345) {
return 1;
}
}
else if (negBalance >= 41) {
return 4;
}
}
}
}
else if (ssDCval >= 75) {
if (MVDiff < 5) {
if (Rs < 3268) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Cs < 1215) {
return 4;
}
else if (Cs >= 1215) {
return 1;
}
}
}
else if (Rs >= 3268) {
return 4;
}
}
else if (MVDiff >= 5) {
if (negBalance < 40) {
if (diffRsCsdiff < 17) {
if (MVDiff < 9) {
if (Rs < 3314) {
return 8;
}
else if (Rs >= 3314) {
return 4;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (diffRsCsdiff >= 17) {
return 4;
}
}
else if (negBalance >= 40) {
if (Rs < 3068) {
return 8;
}
else if (Rs >= 3068) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (refDCval < 77) {
if (Rs < 3158) {
if (MVDiff < 4) {
if (Cs < 1256) {
if (Cs < 815) {
return 1;
}
else if (Cs >= 815) {
if (mcTcor < 95) {
if (Rs < 2735) {
return 4;
}
else if (Rs >= 2735) {
return 1;
}
}
else if (mcTcor >= 95) {
if (mu_mv_mag_sq < 122) {
if (Rs < 3006) {
return 2;
}
else if (Rs >= 3006) {
if (RsDiff < 6) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 4;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 122) {
return 4;
}
}
}
}
else if (Cs >= 1256) {
if (Rs < 2774) {
return 8;
}
else if (Rs >= 2774) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 69) {
if (posBalance < 30) {
return 1;
}
else if (posBalance >= 30) {
if (posBalance < 47) {
if (Rs < 3120) {
return 4;
}
else if (Rs >= 3120) {
return 8;
}
}
else if (posBalance >= 47) {
return 2;
}
}
}
else if (ssDCval >= 69) {
if (SC < 4278) {
return 2;
}
else if (SC >= 4278) {
return 4;
}
}
}
}
else if (Rs >= 3158) {
if (RsDiff < 6) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (RsDiff >= 6) {
if (Cs < 896) {
return 4;
}
else if (Cs >= 896) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (tcor < 94) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (tcor >= 94) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 77) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
}
}
else if (negBalance >= 46) {
if (posBalance < 34) {
if (mcTcor < 98) {
if (Rs < 3099) {
if (Cs < 902) {
return 8;
}
else if (Cs >= 902) {
if (Rs < 2785) {
return 8;
}
else if (Rs >= 2785) {
return 4;
}
}
}
else if (Rs >= 3099) {
return 8;
}
}
else if (mcTcor >= 98) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (posBalance >= 34) {
if (Cs < 304) {
return 8;
}
else if (Cs >= 304) {
return 2;
}
}
}
}
else if (ssDCval >= 96) {
if (negBalance < 31) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (mu_mv_mag_sq < 84) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (MVDiff < 9) {
if (Cs < 1016) {
return 4;
}
else if (Cs >= 1016) {
return 8;
}
}
else if (MVDiff >= 9) {
if (Rs < 3442) {
return 4;
}
else if (Rs >= 3442) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (Rs < 3630) {
return 4;
}
else if (Rs >= 3630) {
return 2;
}
}
}
}
else if (negBalance >= 31) {
if (negBalance < 52) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (negBalance >= 52) {
return 2;
}
}
}
}
}
else if (Rs >= 3818) {
if (Cs < 1265) {
if (refDCval < 91) {
if (posBalance < 46) {
if (ssDCval < 69) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (posBalance >= 46) {
if (Cs < 1136) {
return 4;
}
else if (Cs >= 1136) {
if (Rs < 4219) {
return 8;
}
else if (Rs >= 4219) {
if (Rs < 4286) {
return 4;
}
else if (Rs >= 4286) {
return 8;
}
}
}
}
}
else if (refDCval >= 91) {
if (negBalance < 42) {
if (tcor < 94) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (Rs < 3856) {
return 8;
}
else if (Rs >= 3856) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (Rs < 4139) {
if (posBalance < 41) {
if (Rs < 4021) {
return 4;
}
else if (Rs >= 4021) {
return 4;
}
}
else if (posBalance >= 41) {
if (Rs < 4071) {
return 4;
}
else if (Rs >= 4071) {
return 8;
}
}
}
else if (Rs >= 4139) {
return 4;
}
}
}
}
}
else if (tcor >= 94) {
if (Rs < 4247) {
return 4;
}
else if (Rs >= 4247) {
return 2;
}
}
}
else if (negBalance >= 42) {
return 2;
}
}
}
else if (Cs >= 1265) {
if (CsDiff < 2) {
if (Rs < 3900) {
return 4;
}
else if (Rs >= 3900) {
return 8;
}
}
else if (CsDiff >= 2) {
if (refDCval < 90) {
if (Cs < 1299) {
if (Cs < 1269) {
return 2;
}
else if (Cs >= 1269) {
return 4;
}
}
else if (Cs >= 1299) {
return 1;
}
}
else if (refDCval >= 90) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Cs < 646) {
if (SC < 3424) {
if (AFD < 10) {
if (Cs < 362) {
return 1;
}
else if (Cs >= 362) {
if (Cs < 598) {
return 4;
}
else if (Cs >= 598) {
return 1;
}
}
}
else if (AFD >= 10) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
}
else if (SC >= 3424) {
return 4;
}
}
else if (Cs >= 646) {
if (negBalance < 22) {
if (Cs < 1121) {
if (Rs < 2626) {
return 8;
}
else if (Rs >= 2626) {
if (Rs < 2934) {
if (posBalance < 74) {
if (Rs < 2750) {
return 4;
}
else if (Rs >= 2750) {
return 8;
}
}
else if (posBalance >= 74) {
return 4;
}
}
else if (Rs >= 2934) {
return 4;
}
}
}
else if (Cs >= 1121) {
return 8;
}
}
else if (negBalance >= 22) {
if (ssDCval < 63) {
if (Cs < 1028) {
return 2;
}
else if (Cs >= 1028) {
return 4;
}
}
else if (ssDCval >= 63) {
if (Rs < 3634) {
return 2;
}
else if (Rs >= 3634) {
if (Rs < 3777) {
return 4;
}
else if (Rs >= 3777) {
return 2;
}
}
}
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 65) {
if (SC < 3478) {
return 2;
}
else if (SC >= 3478) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
if (Cs < 994) {
return 2;
}
else if (Cs >= 994) {
if (diffRsCsdiff < 2) {
if (mu_mv_mag_sq < 100) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Rs < 2557) {
return 4;
}
else if (Rs >= 2557) {
if (Rs < 4236) {
if (RsCsDiff < 38) {
return 4;
}
else if (RsCsDiff >= 38) {
return 2;
}
}
else if (Rs >= 4236) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
return 4;
}
}
else if (diffRsCsdiff >= 2) {
if (mu_mv_mag_sq < 25) {
if (Rs < 3708) {
return 2;
}
else if (Rs >= 3708) {
return 4;
}
}
else if (mu_mv_mag_sq >= 25) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 65) {
if (negBalance < 38) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 77) {
if (tcor < 92) {
if (posBalance < 50) {
if (Rs < 3468) {
return 1;
}
else if (Rs >= 3468) {
return 4;
}
}
else if (posBalance >= 50) {
return 4;
}
}
else if (tcor >= 92) {
if (refDCval < 77) {
if (Rs < 3046) {
return 2;
}
else if (Rs >= 3046) {
return 4;
}
}
else if (refDCval >= 77) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 77) {
return 2;
}
}
else if (posBalance >= 53) {
if (AFD < 18) {
if (RsDiff < 5) {
if (Rs < 3949) {
return 4;
}
else if (Rs >= 3949) {
return 2;
}
}
else if (RsDiff >= 5) {
if (mcTcor < 99) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
if (Rs < 2893) {
return 1;
}
else if (Rs >= 2893) {
return 2;
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
else if (AFD >= 18) {
if (Rs < 3005) {
return 1;
}
else if (Rs >= 3005) {
return 4;
}
}
}
}
else if (negBalance >= 38) {
if (diffRsCsdiff < -191) {
if (Rs < 3988) {
return 4;
}
else if (Rs >= 3988) {
return 2;
}
}
else if (diffRsCsdiff >= -191) {
if (TSC < 5) {
if (Rs < 2603) {
return 4;
}
else if (Rs >= 2603) {
if (ssDCval < 84) {
return 2;
}
else if (ssDCval >= 84) {
return 1;
}
}
}
else if (TSC >= 5) {
if (tcor < 77) {
if (RsDiff < 14) {
if (Rs < 3663) {
return 4;
}
else if (Rs >= 3663) {
return 1;
}
}
else if (RsDiff >= 14) {
return 2;
}
}
else if (tcor >= 77) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsCsDiff < 102) {
if (posBalance < 63) {
if (Cs < 768) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (Cs >= 768) {
if (diffTSC < -3) {
if (Rs < 3521) {
return 4;
}
else if (Rs >= 3521) {
return 2;
}
}
else if (diffTSC >= -3) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (posBalance >= 63) {
if (tcor < 96) {
if (posBalance < 71) {
if (Rs < 3313) {
return 4;
}
else if (Rs >= 3313) {
return 2;
}
}
else if (posBalance >= 71) {
if (Cs < 530) {
if (Cs < 484) {
return 8;
}
else if (Cs >= 484) {
if (Rs < 2696) {
return 4;
}
else if (Rs >= 2696) {
return 8;
}
}
}
else if (Cs >= 530) {
return 8;
}
}
}
else if (tcor >= 96) {
return 2;
}
}
}
else if (RsCsDiff >= 102) {
if (negBalance < 52) {
if (RsCsDiff < 171) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 4;
}
}
else if (RsCsDiff >= 171) {
return 2;
}
}
else if (negBalance >= 52) {
if (Cs < 826) {
return 2;
}
else if (Cs >= 826) {
if (refDCval < 102) {
if (Rs < 3439) {
if (CsDiff < 6) {
if (mu_mv_mag_sq < 3) {
if (Rs < 3006) {
return 8;
}
else if (Rs >= 3006) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (Rs >= 3439) {
return 8;
}
}
else if (refDCval >= 102) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 3429) {
if (mcTcor < 97) {
if (RsDiff < 10) {
if (diffMVdiffVal < 3) {
if (Rs < 3148) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
else if (Rs >= 3148) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 2662) {
return 1;
}
else if (Rs >= 2662) {
if (Rs < 2861) {
return 8;
}
else if (Rs >= 2861) {
return 2;
}
}
}
}
else if (RsDiff >= 10) {
if (ssDCval < 117) {
return 2;
}
else if (ssDCval >= 117) {
if (Rs < 3259) {
if (diffMVdiffVal < -1) {
if (RsCsDiff < 113) {
return 2;
}
else if (RsCsDiff >= 113) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2955) {
return 8;
}
else if (Rs >= 2955) {
return 2;
}
}
}
else if (Rs >= 3259) {
if (Rs < 3415) {
return 4;
}
else if (Rs >= 3415) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 7) {
if (Cs < 536) {
if (RsDiff < 5) {
if (Rs < 2865) {
if (Cs < 478) {
return 1;
}
else if (Cs >= 478) {
return 2;
}
}
else if (Rs >= 2865) {
return 8;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (Cs >= 536) {
if (refDCval < 108) {
return 4;
}
else if (refDCval >= 108) {
if (Rs < 2584) {
if (Rs < 2546) {
return 2;
}
else if (Rs >= 2546) {
return 1;
}
}
else if (Rs >= 2584) {
return 2;
}
}
}
}
else if (MVDiff >= 7) {
if (TSC < 6) {
if (CsDiff < 3) {
if (posBalance < 50) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 3092) {
return 8;
}
else if (Rs >= 3092) {
return 2;
}
}
}
else if (posBalance >= 50) {
return 2;
}
}
else if (CsDiff >= 3) {
if (SC < 3527) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
else if (SC >= 3527) {
if (posBalance < 88) {
return 8;
}
else if (posBalance >= 88) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (refDCval < 135) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
if (Cs < 758) {
if (Rs < 2750) {
return 1;
}
else if (Rs >= 2750) {
return 2;
}
}
else if (Cs >= 758) {
return 4;
}
}
}
else if (refDCval >= 135) {
if (Cs < 963) {
return 1;
}
else if (Cs >= 963) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3429) {
if (MVDiff < 7) {
if (refDCval < 138) {
if (mu_mv_mag_sq < 24) {
if (Rs < 3875) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 2;
}
}
}
else if (Rs >= 3875) {
return 2;
}
}
else if (mu_mv_mag_sq >= 24) {
if (diffMVdiffVal < -12) {
if (Rs < 4354) {
return 1;
}
else if (Rs >= 4354) {
return 2;
}
}
else if (diffMVdiffVal >= -12) {
return 2;
}
}
}
else if (refDCval >= 138) {
if (Rs < 3866) {
if (SC < 4953) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 2;
}
}
}
else if (SC >= 4953) {
if (Cs < 1254) {
return 1;
}
else if (Cs >= 1254) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 1;
}
}
}
}
else if (Rs >= 3866) {
if (diffMVdiffVal < -5) {
if (Cs < 1204) {
return 2;
}
else if (Cs >= 1204) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
if (diffRsCsdiff < -24) {
if (Rs < 4010) {
return 1;
}
else if (Rs >= 4010) {
if (Rs < 6352) {
return 2;
}
else if (Rs >= 6352) {
return 1;
}
}
}
else if (diffRsCsdiff >= -24) {
return 2;
}
}
}
}
}
else if (MVDiff >= 7) {
if (negBalance < 54) {
if (TSC < 5) {
if (RsDiff < 5) {
if (Rs < 3513) {
return 2;
}
else if (Rs >= 3513) {
return 1;
}
}
else if (RsDiff >= 5) {
if (Rs < 3956) {
return 8;
}
else if (Rs >= 3956) {
return 2;
}
}
}
else if (TSC >= 5) {
if (ssDCval < 138) {
if (ssDCval < 127) {
return 2;
}
else if (ssDCval >= 127) {
if (AFD < 8) {
if (Rs < 3982) {
return 8;
}
else if (Rs >= 3982) {
return 2;
}
}
else if (AFD >= 8) {
if (Cs < 1002) {
return 2;
}
else if (Cs >= 1002) {
if (Cs < 1031) {
return 1;
}
else if (Cs >= 1031) {
if (tcor < 85) {
return 1;
}
else if (tcor >= 85) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 138) {
if (mu_mv_mag_sq < 54) {
if (tcor < 88) {
return 2;
}
else if (tcor >= 88) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (Rs < 3758) {
return 2;
}
else if (Rs >= 3758) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 54) {
if (posBalance < 57) {
if (negBalance < 55) {
if (Rs < 4907) {
return 1;
}
else if (Rs >= 4907) {
return 2;
}
}
else if (negBalance >= 55) {
return 2;
}
}
else if (posBalance >= 57) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
if (Cs < 1109) {
return 2;
}
else if (Cs >= 1109) {
if (Rs < 4624) {
return 4;
}
else if (Rs >= 4624) {
return 2;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 14) {
if (mcTcor < 68) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
if (Rs < 4400) {
return 8;
}
else if (Rs >= 4400) {
return 4;
}
}
}
else if (mcTcor >= 68) {
if (refDCval < 133) {
if (ssDCval < 80) {
if (TSC < 8) {
if (posBalance < 58) {
if (mu_mv_mag_sq < 87) {
if (Rs < 2784) {
if (AFD < 9) {
if (mu_mv_mag_sq < 48) {
if (Cs < 607) {
return 8;
}
else if (Cs >= 607) {
return 1;
}
}
else if (mu_mv_mag_sq >= 48) {
return 4;
}
}
else if (AFD >= 9) {
if (diffRsCsdiff < 11) {
if (negBalance < 45) {
if (Cs < 952) {
if (RsDiff < 12) {
return 4;
}
else if (RsDiff >= 12) {
if (Rs < 2674) {
return 2;
}
else if (Rs >= 2674) {
return 4;
}
}
}
else if (Cs >= 952) {
return 2;
}
}
else if (negBalance >= 45) {
return 2;
}
}
else if (diffRsCsdiff >= 11) {
if (Rs < 2717) {
return 4;
}
else if (Rs >= 2717) {
return 2;
}
}
}
}
else if (Rs >= 2784) {
if (Rs < 6448) {
if (refDCval < 80) {
if (RsCsDiff < 51) {
if (Rs < 3203) {
if (Rs < 2932) {
return 2;
}
else if (Rs >= 2932) {
return 8;
}
}
else if (Rs >= 3203) {
return 4;
}
}
else if (RsCsDiff >= 51) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (posBalance < 35) {
return 1;
}
else if (posBalance >= 35) {
if (Cs < 439) {
return 1;
}
else if (Cs >= 439) {
return 4;
}
}
}
}
}
else if (refDCval >= 80) {
if (Rs < 4083) {
return 2;
}
else if (Rs >= 4083) {
return 4;
}
}
}
else if (Rs >= 6448) {
if (Rs < 9150) {
return 8;
}
else if (Rs >= 9150) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
return 8;
}
}
else if (posBalance >= 58) {
if (AFD < 12) {
if (Rs < 2969) {
return 2;
}
else if (Rs >= 2969) {
return 4;
}
}
else if (AFD >= 12) {
return 2;
}
}
}
else if (TSC >= 8) {
if (SC < 4368) {
if (mcTcor < 82) {
if (RsDiff < 20) {
if (Rs < 2748) {
if (Rs < 2556) {
return 1;
}
else if (Rs >= 2556) {
if (Rs < 2661) {
return 4;
}
else if (Rs >= 2661) {
return 2;
}
}
}
else if (Rs >= 2748) {
return 1;
}
}
else if (RsDiff >= 20) {
if (Rs < 2675) {
return 2;
}
else if (Rs >= 2675) {
return 8;
}
}
}
else if (mcTcor >= 82) {
if (AFD < 13) {
if (posBalance < 54) {
if (Rs < 2866) {
return 1;
}
else if (Rs >= 2866) {
return 2;
}
}
else if (posBalance >= 54) {
if (negBalance < 38) {
return 2;
}
else if (negBalance >= 38) {
return 4;
}
}
}
else if (AFD >= 13) {
if (posBalance < 54) {
if (diffRsCsdiff < 35) {
if (posBalance < 53) {
if (Cs < 279) {
return 1;
}
else if (Cs >= 279) {
return 2;
}
}
else if (posBalance >= 53) {
if (Rs < 2747) {
return 2;
}
else if (Rs >= 2747) {
return 4;
}
}
}
else if (diffRsCsdiff >= 35) {
if (Rs < 2739) {
return 2;
}
else if (Rs >= 2739) {
if (negBalance < 51) {
if (diffAFD < 5) {
return 4;
}
else if (diffAFD >= 5) {
return 2;
}
}
else if (negBalance >= 51) {
return 2;
}
}
}
}
else if (posBalance >= 54) {
if (Cs < 215) {
return 1;
}
else if (Cs >= 215) {
return 2;
}
}
}
}
}
else if (SC >= 4368) {
if (mu_mv_mag_sq < 43) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (refDCval < 77) {
return 4;
}
else if (refDCval >= 77) {
if (Rs < 4398) {
return 2;
}
else if (Rs >= 4398) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (MVDiff < 17) {
if (Rs < 3769) {
return 1;
}
else if (Rs >= 3769) {
return 4;
}
}
else if (MVDiff >= 17) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 80) {
if (RsCsDiff < 210) {
if (Rs < 4309) {
if (refDCval < 97) {
if (mcTcor < 92) {
if (posBalance < 61) {
if (CsDiff < 7) {
if (mcTcor < 86) {
if (Rs < 2635) {
return 8;
}
else if (Rs >= 2635) {
return 2;
}
}
else if (mcTcor >= 86) {
return 4;
}
}
else if (CsDiff >= 7) {
return 1;
}
}
else if (posBalance >= 61) {
if (mcTcor < 89) {
if (Cs < 1081) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
if (Rs < 2879) {
return 4;
}
else if (Rs >= 2879) {
return 8;
}
}
}
else if (Cs >= 1081) {
if (Rs < 2783) {
return 8;
}
else if (Rs >= 2783) {
return 4;
}
}
}
else if (mcTcor >= 89) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
}
}
else if (mcTcor >= 92) {
if (AFD < 10) {
if (ssDCval < 85) {
if (Cs < 869) {
return 1;
}
else if (Cs >= 869) {
return 2;
}
}
else if (ssDCval >= 85) {
if (Rs < 2686) {
if (Rs < 2602) {
return 1;
}
else if (Rs >= 2602) {
return 2;
}
}
else if (Rs >= 2686) {
if (Cs < 846) {
return 8;
}
else if (Cs >= 846) {
return 2;
}
}
}
}
else if (AFD >= 10) {
if (MVDiff < 32) {
if (tcor < 80) {
if (SC < 3869) {
return 2;
}
else if (SC >= 3869) {
return 4;
}
}
else if (tcor >= 80) {
if (CsDiff < 7) {
if (RsCsDiff < 113) {
if (Rs < 2847) {
return 1;
}
else if (Rs >= 2847) {
return 4;
}
}
else if (RsCsDiff >= 113) {
return 2;
}
}
else if (CsDiff >= 7) {
return 1;
}
}
}
else if (MVDiff >= 32) {
if (diffRsCsdiff < -7) {
if (Rs < 3752) {
if (Rs < 2641) {
return 8;
}
else if (Rs >= 2641) {
return 2;
}
}
else if (Rs >= 3752) {
return 1;
}
}
else if (diffRsCsdiff >= -7) {
if (SC < 3243) {
return 1;
}
else if (SC >= 3243) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 97) {
if (diffRsCsdiff < -104) {
if (RsDiff < 9) {
if (SC < 4057) {
if (Cs < 367) {
return 8;
}
else if (Cs >= 367) {
return 1;
}
}
else if (SC >= 4057) {
return 8;
}
}
else if (RsDiff >= 9) {
if (diffAFD < -6) {
if (refDCval < 115) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
if (gchDC < 3) {
if (Rs < 2652) {
return 8;
}
else if (Rs >= 2652) {
return 2;
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
else if (refDCval >= 115) {
return 1;
}
}
else if (diffAFD >= -6) {
if (mcTcor < 99) {
return 2;
}
else if (mcTcor >= 99) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -104) {
if (SC < 3472) {
if (TSC < 10) {
if (ssDCval < 105) {
if (RsDiff < 11) {
if (Cs < 449) {
return 1;
}
else if (Cs >= 449) {
return 8;
}
}
else if (RsDiff >= 11) {
if (Cs < 284) {
return 4;
}
else if (Cs >= 284) {
return 2;
}
}
}
else if (ssDCval >= 105) {
if (Rs < 2862) {
if (tcor < 84) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
else if (tcor >= 84) {
return 1;
}
}
else if (Rs >= 2862) {
return 8;
}
}
}
else if (TSC >= 10) {
if (diffMVdiffVal < 19) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (diffMVdiffVal >= 19) {
if (Rs < 2603) {
return 2;
}
else if (Rs >= 2603) {
return 1;
}
}
}
}
else if (SC >= 3472) {
if (AFD < 23) {
if (gchDC < 2) {
if (mcTcor < 98) {
if (Cs < 480) {
return 4;
}
else if (Cs >= 480) {
if (MVDiff < 16) {
if (Cs < 945) {
return 2;
}
else if (Cs >= 945) {
return 8;
}
}
else if (MVDiff >= 16) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (SC < 4530) {
if (Cs < 366) {
return 4;
}
else if (Cs >= 366) {
if (MVDiff < 33) {
if (diffRsCsdiff < -3) {
if (SC < 3859) {
return 2;
}
else if (SC >= 3859) {
if (Cs < 523) {
return 2;
}
else if (Cs >= 523) {
return 8;
}
}
}
else if (diffRsCsdiff >= -3) {
if (diffAFD < 1) {
if (Cs < 771) {
if (Rs < 3305) {
return 1;
}
else if (Rs >= 3305) {
return 2;
}
}
else if (Cs >= 771) {
if (Rs < 3114) {
return 8;
}
else if (Rs >= 3114) {
return 4;
}
}
}
else if (diffAFD >= 1) {
if (Rs < 2541) {
return 1;
}
else if (Rs >= 2541) {
return 8;
}
}
}
}
else if (MVDiff >= 33) {
if (tcor < 95) {
if (diffRsCsdiff < -44) {
if (Rs < 3365) {
return 8;
}
else if (Rs >= 3365) {
return 4;
}
}
else if (diffRsCsdiff >= -44) {
return 8;
}
}
else if (tcor >= 95) {
if (Rs < 2789) {
return 8;
}
else if (Rs >= 2789) {
return 1;
}
}
}
}
}
else if (SC >= 4530) {
if (mu_mv_mag_sq < 42) {
if (diffMVdiffVal < 11) {
return 8;
}
else if (diffMVdiffVal >= 11) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 42) {
return 2;
}
}
}
}
else if (AFD >= 23) {
if (refDCval < 105) {
if (Cs < 1096) {
return 2;
}
else if (Cs >= 1096) {
return 4;
}
}
else if (refDCval >= 105) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (Rs < 3736) {
if (Rs < 2668) {
return 1;
}
else if (Rs >= 2668) {
return 8;
}
}
else if (Rs >= 3736) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 4309) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 210) {
if (mcTcor < 91) {
if (RsCsDiff < 382) {
if (SC < 3824) {
if (TSC < 10) {
if (Cs < 1071) {
return 2;
}
else if (Cs >= 1071) {
return 1;
}
}
else if (TSC >= 10) {
if (diffMVdiffVal < -2) {
if (posBalance < 69) {
if (Rs < 2772) {
return 2;
}
else if (Rs >= 2772) {
return 4;
}
}
else if (posBalance >= 69) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
if (ssDCval < 93) {
if (Rs < 2656) {
return 8;
}
else if (Rs >= 2656) {
return 2;
}
}
else if (ssDCval >= 93) {
if (Rs < 2945) {
return 4;
}
else if (Rs >= 2945) {
return 8;
}
}
}
}
}
else if (SC >= 3824) {
if (mcTcor < 80) {
return 4;
}
else if (mcTcor >= 80) {
if (refDCval < 109) {
if (CsDiff < 7) {
if (RsCsDiff < 256) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
else if (RsCsDiff >= 256) {
if (MVDiff < 20) {
if (Rs < 3235) {
return 4;
}
else if (Rs >= 3235) {
return 2;
}
}
else if (MVDiff >= 20) {
return 4;
}
}
}
else if (CsDiff >= 7) {
if (AFD < 28) {
return 8;
}
else if (AFD >= 28) {
return 2;
}
}
}
else if (refDCval >= 109) {
if (Cs < 494) {
return 4;
}
else if (Cs >= 494) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 382) {
if (Cs < 928) {
return 2;
}
else if (Cs >= 928) {
if (AFD < 26) {
if (tcor < 69) {
return 8;
}
else if (tcor >= 69) {
if (RsDiff < 19) {
return 1;
}
else if (RsDiff >= 19) {
return 2;
}
}
}
else if (AFD >= 26) {
if (MVDiff < 108) {
return 2;
}
else if (MVDiff >= 108) {
if (Rs < 3265) {
return 1;
}
else if (Rs >= 3265) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (ssDCval < 111) {
if (gchDC < 6) {
if (ssDCval < 82) {
if (Rs < 4454) {
if (Cs < 397) {
if (Rs < 3434) {
return 2;
}
else if (Rs >= 3434) {
return 4;
}
}
else if (Cs >= 397) {
return 2;
}
}
else if (Rs >= 4454) {
return 1;
}
}
else if (ssDCval >= 82) {
if (RsCsDiff < 248) {
if (refDCval < 105) {
if (tcor < 75) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
if (Rs < 2753) {
return 2;
}
else if (Rs >= 2753) {
return 1;
}
}
}
else if (tcor >= 75) {
if (Rs < 3468) {
return 2;
}
else if (Rs >= 3468) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 8;
}
}
}
}
else if (refDCval >= 105) {
return 8;
}
}
else if (RsCsDiff >= 248) {
if (MVDiff < 42) {
if (refDCval < 81) {
return 1;
}
else if (refDCval >= 81) {
return 2;
}
}
else if (MVDiff >= 42) {
if (Cs < 610) {
return 2;
}
else if (Cs >= 610) {
if (Cs < 996) {
if (MVDiff < 46) {
return 8;
}
else if (MVDiff >= 46) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (Cs >= 996) {
if (Rs < 4724) {
return 2;
}
else if (Rs >= 4724) {
return 1;
}
}
}
}
}
}
}
else if (gchDC >= 6) {
if (ssDCval < 103) {
if (mu_mv_mag_sq < 28) {
if (MVDiff < 29) {
return 2;
}
else if (MVDiff >= 29) {
return 1;
}
}
else if (mu_mv_mag_sq >= 28) {
if (Cs < 711) {
return 2;
}
else if (Cs >= 711) {
return 8;
}
}
}
else if (ssDCval >= 103) {
return 8;
}
}
}
else if (ssDCval >= 111) {
if (Rs < 3621) {
if (Cs < 319) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 1;
}
}
else if (Cs >= 319) {
if (posBalance < 67) {
if (refDCval < 116) {
if (posBalance < 60) {
if (diffTSC < -2) {
return 1;
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (posBalance >= 60) {
if (RsDiff < 20) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 8;
}
}
else if (RsDiff >= 20) {
if (Rs < 3465) {
return 1;
}
else if (Rs >= 3465) {
return 2;
}
}
}
}
else if (refDCval >= 116) {
if (Cs < 1061) {
if (MVDiff < 34) {
if (mu_mv_mag_sq < 9) {
if (diffMVdiffVal < -32) {
return 1;
}
else if (diffMVdiffVal >= -32) {
if (Rs < 2591) {
return 1;
}
else if (Rs >= 2591) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 9) {
return 1;
}
}
else if (MVDiff >= 34) {
return 1;
}
}
else if (Cs >= 1061) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
if (Rs < 2616) {
return 2;
}
else if (Rs >= 2616) {
return 1;
}
}
}
}
}
else if (posBalance >= 67) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
return 2;
}
}
}
}
else if (Rs >= 3621) {
if (Cs < 771) {
if (Rs < 4254) {
if (MVDiff < 35) {
return 2;
}
else if (MVDiff >= 35) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (Rs >= 4254) {
return 2;
}
}
else if (Cs >= 771) {
if (mu_mv_mag_sq < 18) {
if (MVDiff < 45) {
if (tcor < 85) {
if (refDCval < 118) {
return 2;
}
else if (refDCval >= 118) {
if (ssDCval < 126) {
if (ssDCval < 121) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
return 2;
}
}
else if (ssDCval >= 121) {
return 8;
}
}
else if (ssDCval >= 126) {
if (MVDiff < 31) {
if (Cs < 1046) {
return 2;
}
else if (Cs >= 1046) {
return 8;
}
}
else if (MVDiff >= 31) {
if (Rs < 3667) {
return 8;
}
else if (Rs >= 3667) {
return 2;
}
}
}
}
}
else if (tcor >= 85) {
if (diffAFD < -1) {
if (negBalance < 53) {
return 2;
}
else if (negBalance >= 53) {
return 8;
}
}
else if (diffAFD >= -1) {
if (Cs < 834) {
if (Rs < 4063) {
return 2;
}
else if (Rs >= 4063) {
return 8;
}
}
else if (Cs >= 834) {
return 1;
}
}
}
}
else if (MVDiff >= 45) {
if (diffAFD < 2) {
if (tcor < 69) {
return 8;
}
else if (tcor >= 69) {
if (posBalance < 61) {
if (MVDiff < 58) {
return 8;
}
else if (MVDiff >= 58) {
if (Rs < 4637) {
return 8;
}
else if (Rs >= 4637) {
return 2;
}
}
}
else if (posBalance >= 61) {
return 2;
}
}
}
else if (diffAFD >= 2) {
if (diffRsCsdiff < 228) {
return 8;
}
else if (diffRsCsdiff >= 228) {
if (Rs < 4843) {
return 8;
}
else if (Rs >= 4843) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (CsDiff < 5) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
else if (CsDiff >= 5) {
if (tcor < 79) {
return 2;
}
else if (tcor >= 79) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 133) {
if (mcTcor < 93) {
if (AFD < 19) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
if (posBalance < 55) {
if (Rs < 4850) {
return 1;
}
else if (Rs >= 4850) {
return 2;
}
}
else if (posBalance >= 55) {
if (diffMVdiffVal < -28) {
return 1;
}
else if (diffMVdiffVal >= -28) {
return 4;
}
}
}
}
else if (AFD >= 19) {
if (RsDiff < 26) {
if (mu_mv_mag_sq < 27) {
if (refDCval < 137) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 1;
}
}
else if (refDCval >= 137) {
if (posBalance < 64) {
if (RsDiff < 25) {
return 1;
}
else if (RsDiff >= 25) {
if (Rs < 4260) {
return 1;
}
else if (Rs >= 4260) {
return 2;
}
}
}
else if (posBalance >= 64) {
if (ssDCval < 162) {
return 8;
}
else if (ssDCval >= 162) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
return 1;
}
}
}
else if (RsDiff >= 26) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (CsDiff < 3) {
if (SC < 3272) {
return 1;
}
else if (SC >= 3272) {
if (gchDC < 1) {
if (Rs < 5002) {
return 4;
}
else if (Rs >= 5002) {
return 2;
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (SC < 3766) {
if (MVDiff < 87) {
return 1;
}
else if (MVDiff >= 87) {
if (Cs < 763) {
return 1;
}
else if (Cs >= 763) {
return 8;
}
}
}
else if (SC >= 3766) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (CsDiff < 4) {
if (negBalance < 46) {
if (ssDCval < 145) {
return 2;
}
else if (ssDCval >= 145) {
if (Cs < 607) {
return 1;
}
else if (Cs >= 607) {
return 2;
}
}
}
else if (negBalance >= 46) {
if (Cs < 911) {
if (negBalance < 65) {
return 2;
}
else if (negBalance >= 65) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (Cs >= 911) {
if (Cs < 1006) {
return 4;
}
else if (Cs >= 1006) {
return 1;
}
}
}
}
else if (CsDiff >= 4) {
if (SC < 7404) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (SC < 6155) {
if (mu_mv_mag_sq < 31) {
if (Rs < 4116) {
if (posBalance < 67) {
return 1;
}
else if (posBalance >= 67) {
if (Rs < 3782) {
return 1;
}
else if (Rs >= 3782) {
return 8;
}
}
}
else if (Rs >= 4116) {
if (Cs < 908) {
return 2;
}
else if (Cs >= 908) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 31) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
return 2;
}
}
}
else if (SC >= 6155) {
if (posBalance < 44) {
if (Rs < 5269) {
return 1;
}
else if (Rs >= 5269) {
return 2;
}
}
else if (posBalance >= 44) {
return 8;
}
}
}
}
else if (SC >= 7404) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1313) {
if (mu_mv_mag_sq < 111) {
if (RsDiff < 9) {
if (CsDiff < 8) {
if (negBalance < 32) {
if (ssDCval < 104) {
if (Rs < 822) {
if (Rs < 301) {
return 4;
}
else if (Rs >= 301) {
if (MVDiff < 84) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (MVDiff >= 84) {
return 8;
}
}
}
else if (Rs >= 822) {
if (SC < 10165) {
if (posBalance < 44) {
if (Cs < 2010) {
if (ssDCval < 99) {
if (ssDCval < 87) {
if (ssDCval < 64) {
if (ssDCval < 59) {
if (ssDCval < 55) {
if (Rs < 4011) {
return 8;
}
else if (Rs >= 4011) {
return 4;
}
}
else if (ssDCval >= 55) {
if (Rs < 1027) {
return 8;
}
else if (Rs >= 1027) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
}
}
else if (ssDCval >= 59) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (ssDCval >= 64) {
if (MVDiff < 9) {
if (Rs < 3031) {
if (Cs < 1882) {
return 8;
}
else if (Cs >= 1882) {
return 4;
}
}
else if (Rs >= 3031) {
if (Rs < 4298) {
return 4;
}
else if (Rs >= 4298) {
if (Cs < 1756) {
return 8;
}
else if (Cs >= 1756) {
return 4;
}
}
}
}
else if (MVDiff >= 9) {
return 4;
}
}
}
else if (ssDCval >= 87) {
if (Cs < 1722) {
if (Cs < 1426) {
return 2;
}
else if (Cs >= 1426) {
return 8;
}
}
else if (Cs >= 1722) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (ssDCval >= 99) {
return 4;
}
}
else if (Cs >= 2010) {
if (ssDCval < 82) {
if (ssDCval < 56) {
if (Rs < 4181) {
return 8;
}
else if (Rs >= 4181) {
if (Rs < 4768) {
if (ssDCval < 55) {
if (posBalance < 38) {
if (Rs < 4583) {
return 8;
}
else if (Rs >= 4583) {
return 4;
}
}
else if (posBalance >= 38) {
if (RsDiff < 7) {
if (Rs < 4583) {
return 4;
}
else if (Rs >= 4583) {
return 4;
}
}
else if (RsDiff >= 7) {
return 4;
}
}
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (Rs >= 4768) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (Rs < 1369) {
return 8;
}
else if (Rs >= 1369) {
return 4;
}
}
}
else if (ssDCval >= 82) {
if (Rs < 4848) {
if (Rs < 2739) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
if (Rs < 2402) {
if (Rs < 2338) {
if (Rs < 2327) {
if (Rs < 2320) {
return 4;
}
else if (Rs >= 2320) {
return 2;
}
}
else if (Rs >= 2327) {
return 1;
}
}
else if (Rs >= 2338) {
return 4;
}
}
else if (Rs >= 2402) {
if (Rs < 2419) {
return 8;
}
else if (Rs >= 2419) {
if (Rs < 2479) {
return 4;
}
else if (Rs >= 2479) {
return 1;
}
}
}
}
}
else if (Rs >= 2739) {
return 8;
}
}
else if (Rs >= 4848) {
return 4;
}
}
}
}
else if (posBalance >= 44) {
if (Rs < 2952) {
if (mu_mv_mag_sq < 87) {
if (SC < 2449) {
return 2;
}
else if (SC >= 2449) {
if (posBalance < 68) {
if (Cs < 1414) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
return 4;
}
}
else if (Cs >= 1414) {
if (Cs < 2008) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (MVDiff < 12) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (Rs < 2021) {
return 1;
}
else if (Rs >= 2021) {
return 8;
}
}
}
else if (MVDiff >= 12) {
if (Rs < 1462) {
return 4;
}
else if (Rs >= 1462) {
return 8;
}
}
}
}
else if (Cs >= 2008) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 4;
}
}
}
}
else if (posBalance >= 68) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (Rs < 2714) {
if (Cs < 1683) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Rs < 2261) {
if (negBalance < 23) {
if (RsDiff < 2) {
if (Cs < 1471) {
return 4;
}
else if (Cs >= 1471) {
return 8;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (negBalance >= 23) {
if (Rs < 831) {
return 4;
}
else if (Rs >= 831) {
return 8;
}
}
}
else if (Rs >= 2261) {
if (Cs < 1529) {
return 4;
}
else if (Cs >= 1529) {
return 8;
}
}
}
}
else if (Cs >= 1683) {
if (AFD < 10) {
if (Cs < 3908) {
if (posBalance < 50) {
if (Cs < 1739) {
return 4;
}
else if (Cs >= 1739) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 4;
}
}
}
else if (posBalance >= 50) {
if (SC < 4891) {
return 8;
}
else if (SC >= 4891) {
if (SC < 4929) {
return 4;
}
else if (SC >= 4929) {
return 8;
}
}
}
}
else if (Cs >= 3908) {
return 1;
}
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (Rs >= 2714) {
if (negBalance < 29) {
if (Rs < 2721) {
return 1;
}
else if (Rs >= 2721) {
return 8;
}
}
else if (negBalance >= 29) {
return 4;
}
}
}
}
else if (Rs >= 2952) {
if (Cs < 1818) {
if (diffRsCsdiff < -26) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
else if (diffRsCsdiff >= -26) {
if (posBalance < 48) {
if (MVDiff < 1) {
if (Rs < 4364) {
return 8;
}
else if (Rs >= 4364) {
return 2;
}
}
else if (MVDiff >= 1) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Rs < 4284) {
return 8;
}
else if (Rs >= 4284) {
return 4;
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 71) {
if (Rs < 3486) {
return 4;
}
else if (Rs >= 3486) {
return 4;
}
}
else if (ssDCval >= 71) {
if (MVDiff < 1) {
if (Rs < 3991) {
return 4;
}
else if (Rs >= 3991) {
return 2;
}
}
else if (MVDiff >= 1) {
if (posBalance < 53) {
if (diffRsCsdiff < 11) {
return 4;
}
else if (diffRsCsdiff >= 11) {
return 2;
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
}
}
}
else if (Cs >= 1818) {
if (RsDiff < 7) {
if (Rs < 4248) {
return 8;
}
else if (Rs >= 4248) {
if (Rs < 4379) {
return 8;
}
else if (Rs >= 4379) {
return 8;
}
}
}
else if (RsDiff >= 7) {
if (gchDC < 1) {
if (Rs < 3583) {
return 1;
}
else if (Rs >= 3583) {
return 2;
}
}
else if (gchDC >= 1) {
if (Cs < 2247) {
if (diffAFD < -1) {
if (Cs < 2057) {
return 2;
}
else if (Cs >= 2057) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Cs >= 2247) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 10165) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 2;
}
}
}
}
else if (ssDCval >= 104) {
if (refDCval < 125) {
if (Rs < 1272) {
if (posBalance < 57) {
if (ssDCval < 111) {
return 1;
}
else if (ssDCval >= 111) {
return 4;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (Rs >= 1272) {
if (Rs < 4183) {
if (ssDCval < 105) {
if (MVDiff < 3) {
if (Rs < 2922) {
return 8;
}
else if (Rs >= 2922) {
return 4;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (ssDCval >= 105) {
if (mu_mv_mag_sq < 7) {
return 2;
}
else if (mu_mv_mag_sq >= 7) {
if (posBalance < 10) {
if (Rs < 2167) {
return 2;
}
else if (Rs >= 2167) {
return 8;
}
}
else if (posBalance >= 10) {
return 8;
}
}
}
}
else if (Rs >= 4183) {
if (Cs < 2362) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
else if (Cs >= 2362) {
return 8;
}
}
}
}
else if (refDCval >= 125) {
if (mu_mv_mag_sq < 69) {
if (AFD < 10) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (AFD >= 10) {
if (Rs < 1875) {
return 1;
}
else if (Rs >= 1875) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
if (Rs < 1058) {
if (Cs < 1863) {
return 8;
}
else if (Cs >= 1863) {
return 1;
}
}
else if (Rs >= 1058) {
if (RsDiff < 8) {
if (ssDCval < 136) {
if (ssDCval < 135) {
if (tcor < 93) {
if (Rs < 2466) {
return 1;
}
else if (Rs >= 2466) {
return 8;
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (ssDCval >= 135) {
return 1;
}
}
else if (ssDCval >= 136) {
return 8;
}
}
else if (RsDiff >= 8) {
if (Rs < 3433) {
return 1;
}
else if (Rs >= 3433) {
return 2;
}
}
}
}
}
}
}
}
else if (negBalance >= 32) {
if (Rs < 1497) {
if (ssDCval < 161) {
if (refDCval < 84) {
if (TSC < 11) {
if (Cs < 1766) {
if (Rs < 570) {
return 1;
}
else if (Rs >= 570) {
if (MVDiff < 5) {
if (ssDCval < 80) {
if (TSC < 6) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Cs < 1544) {
if (AFD < 6) {
if (negBalance < 39) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 1351) {
return 4;
}
else if (Rs >= 1351) {
return 8;
}
}
}
else if (negBalance >= 39) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
return 4;
}
}
}
else if (AFD >= 6) {
if (diffRsCsdiff < -17) {
if (Rs < 1213) {
return 2;
}
else if (Rs >= 1213) {
return 4;
}
}
else if (diffRsCsdiff >= -17) {
return 4;
}
}
}
else if (Cs >= 1544) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (negBalance < 42) {
if (Rs < 1043) {
return 8;
}
else if (Rs >= 1043) {
return 4;
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
}
else if (ssDCval >= 80) {
return 1;
}
}
else if (MVDiff >= 5) {
if (SC < 2514) {
if (ssDCval < 52) {
if (Rs < 1070) {
return 4;
}
else if (Rs >= 1070) {
return 2;
}
}
else if (ssDCval >= 52) {
if (tcor < 70) {
return 8;
}
else if (tcor >= 70) {
if (Cs < 1528) {
if (ssDCval < 75) {
if (TSC < 4) {
if (Rs < 1062) {
return 1;
}
else if (Rs >= 1062) {
return 4;
}
}
else if (TSC >= 4) {
if (AFD < 14) {
if (MVDiff < 6) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
return 2;
}
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (AFD >= 14) {
if (Rs < 959) {
return 8;
}
else if (Rs >= 959) {
return 2;
}
}
}
}
else if (ssDCval >= 75) {
if (posBalance < 59) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
else if (posBalance >= 59) {
return 4;
}
}
}
else if (Cs >= 1528) {
if (tcor < 92) {
if (TSC < 7) {
if (Cs < 1738) {
return 8;
}
else if (Cs >= 1738) {
return 1;
}
}
else if (TSC >= 7) {
if (Rs < 918) {
return 4;
}
else if (Rs >= 918) {
return 8;
}
}
}
else if (tcor >= 92) {
return 2;
}
}
}
}
}
else if (SC >= 2514) {
if (negBalance < 46) {
if (mcTcor < 92) {
if (mcTcor < 75) {
return 8;
}
else if (mcTcor >= 75) {
if (Rs < 1264) {
if (Rs < 1032) {
return 8;
}
else if (Rs >= 1032) {
return 4;
}
}
else if (Rs >= 1264) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (CsDiff < 5) {
if (Rs < 1098) {
if (ssDCval < 60) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Rs < 1019) {
return 2;
}
else if (Rs >= 1019) {
return 8;
}
}
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (Rs >= 1098) {
return 4;
}
}
else if (CsDiff >= 5) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (ssDCval < 54) {
return 2;
}
else if (ssDCval >= 54) {
return 8;
}
}
}
}
}
else if (negBalance >= 46) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
if (negBalance < 59) {
if (MVDiff < 10) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (TSC < 8) {
if (mu_mv_mag_sq < 28) {
return 2;
}
else if (mu_mv_mag_sq >= 28) {
if (Rs < 1385) {
return 4;
}
else if (Rs >= 1385) {
return 8;
}
}
}
else if (TSC >= 8) {
if (diffRsCsdiff < 25) {
if (RsCsDiff < 63) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 2;
}
}
else if (RsCsDiff >= 63) {
return 4;
}
}
else if (diffRsCsdiff >= 25) {
return 2;
}
}
}
}
else if (negBalance >= 59) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 1766) {
if (mcTcor < 98) {
if (Rs < 1186) {
if (Cs < 2061) {
if (TSC < 5) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (Cs >= 2061) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (ssDCval < 70) {
return 2;
}
else if (ssDCval >= 70) {
return 8;
}
}
}
}
else if (Rs >= 1186) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
}
}
else if (TSC >= 11) {
if (CsDiff < 7) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
return 8;
}
}
else if (CsDiff >= 7) {
if (Cs < 1509) {
return 4;
}
else if (Cs >= 1509) {
if (Cs < 1727) {
return 8;
}
else if (Cs >= 1727) {
return 4;
}
}
}
}
}
else if (refDCval >= 84) {
if (mu_mv_mag_sq < 48) {
if (mcTcor < 90) {
if (refDCval < 123) {
if (diffRsCsdiff < 14) {
if (ssDCval < 111) {
if (RsCsDiff < 67) {
return 4;
}
else if (RsCsDiff >= 67) {
return 8;
}
}
else if (ssDCval >= 111) {
if (Rs < 1304) {
return 8;
}
else if (Rs >= 1304) {
return 4;
}
}
}
else if (diffRsCsdiff >= 14) {
return 8;
}
}
else if (refDCval >= 123) {
if (AFD < 16) {
if (Cs < 1585) {
return 1;
}
else if (Cs >= 1585) {
return 2;
}
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (mcTcor >= 90) {
if (TSC < 8) {
if (mu_mv_mag_sq < 3) {
if (tcor < 96) {
if (Rs < 1333) {
if (TSC < 7) {
if (ssDCval < 140) {
return 8;
}
else if (ssDCval >= 140) {
return 2;
}
}
else if (TSC >= 7) {
if (negBalance < 47) {
if (RsCsDiff < 46) {
return 8;
}
else if (RsCsDiff >= 46) {
return 4;
}
}
else if (negBalance >= 47) {
return 8;
}
}
}
else if (Rs >= 1333) {
if (Cs < 1666) {
return 2;
}
else if (Cs >= 1666) {
return 8;
}
}
}
else if (tcor >= 96) {
if (Rs < 651) {
return 8;
}
else if (Rs >= 651) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 3002) {
return 8;
}
else if (SC >= 3002) {
if (AFD < 14) {
if (Cs < 1595) {
return 1;
}
else if (Cs >= 1595) {
return 8;
}
}
else if (AFD >= 14) {
return 2;
}
}
}
}
else if (TSC >= 8) {
if (negBalance < 56) {
if (diffRsCsdiff < -25) {
return 1;
}
else if (diffRsCsdiff >= -25) {
if (AFD < 16) {
if (negBalance < 52) {
if (RsCsDiff < 46) {
if (Cs < 1362) {
return 4;
}
else if (Cs >= 1362) {
return 8;
}
}
else if (RsCsDiff >= 46) {
if (RsCsDiff < 79) {
return 4;
}
else if (RsCsDiff >= 79) {
return 8;
}
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (AFD >= 16) {
if (Cs < 1345) {
return 1;
}
else if (Cs >= 1345) {
return 4;
}
}
}
}
else if (negBalance >= 56) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
if (ssDCval < 139) {
return 8;
}
else if (ssDCval >= 139) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (Rs < 1274) {
if (negBalance < 46) {
if (Rs < 693) {
return 1;
}
else if (Rs >= 693) {
if (Rs < 1006) {
if (posBalance < 50) {
if (Cs < 1328) {
return 8;
}
else if (Cs >= 1328) {
return 4;
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 74) {
if (Rs < 802) {
return 4;
}
else if (Rs >= 802) {
return 8;
}
}
else if (mu_mv_mag_sq >= 74) {
if (negBalance < 35) {
if (ssDCval < 117) {
return 4;
}
else if (ssDCval >= 117) {
return 8;
}
}
else if (negBalance >= 35) {
return 8;
}
}
}
}
else if (Rs >= 1006) {
if (RsDiff < 3) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
if (Rs < 1020) {
return 8;
}
else if (Rs >= 1020) {
if (Rs < 1263) {
if (Rs < 1043) {
if (TSC < 6) {
if (Rs < 1029) {
return 4;
}
else if (Rs >= 1029) {
return 8;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (Rs >= 1043) {
return 4;
}
}
else if (Rs >= 1263) {
return 8;
}
}
}
}
else if (RsDiff >= 3) {
if (negBalance < 41) {
if (ssDCval < 112) {
return 4;
}
else if (ssDCval >= 112) {
if (SC < 2824) {
return 8;
}
else if (SC >= 2824) {
return 4;
}
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
else if (Rs >= 1274) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
}
else if (ssDCval >= 161) {
if (Rs < 1050) {
return 1;
}
else if (Rs >= 1050) {
return 2;
}
}
}
else if (Rs >= 1497) {
if (mcTcor < 96) {
if (CsDiff < 7) {
if (refDCval < 110) {
if (mcTcor < 93) {
if (Cs < 1537) {
if (ssDCval < 108) {
if (negBalance < 45) {
if (Rs < 2920) {
if (mu_mv_mag_sq < 70) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (AFD < 12) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
}
else if (Rs >= 2920) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
return 8;
}
}
}
else if (negBalance >= 45) {
if (Rs < 1603) {
if (mu_mv_mag_sq < 3) {
if (SC < 2954) {
if (tcor < 64) {
if (Rs < 1499) {
return 8;
}
else if (Rs >= 1499) {
return 4;
}
}
else if (tcor >= 64) {
return 8;
}
}
else if (SC >= 2954) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (Rs >= 1603) {
if (posBalance < 56) {
if (refDCval < 75) {
if (AFD < 13) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (AFD >= 13) {
return 1;
}
}
else if (refDCval >= 75) {
if (diffRsCsdiff < 27) {
if (Cs < 1339) {
return 2;
}
else if (Cs >= 1339) {
return 8;
}
}
else if (diffRsCsdiff >= 27) {
if (Rs < 1896) {
return 8;
}
else if (Rs >= 1896) {
return 2;
}
}
}
}
else if (posBalance >= 56) {
if (negBalance < 49) {
if (mu_mv_mag_sq < 23) {
if (Rs < 1678) {
return 4;
}
else if (Rs >= 1678) {
return 8;
}
}
else if (mu_mv_mag_sq >= 23) {
return 8;
}
}
else if (negBalance >= 49) {
return 8;
}
}
}
}
}
else if (ssDCval >= 108) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (Cs >= 1537) {
if (diffRsCsdiff < -24) {
if (mu_mv_mag_sq < 61) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
return 8;
}
}
else if (mu_mv_mag_sq >= 61) {
if (Rs < 2300) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 2074) {
if (Cs < 1771) {
return 8;
}
else if (Cs >= 1771) {
return 4;
}
}
else if (Rs >= 2074) {
return 4;
}
}
}
else if (Rs >= 2300) {
if (AFD < 14) {
if (Cs < 3021) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
if (Rs < 2730) {
return 4;
}
else if (Rs >= 2730) {
return 8;
}
}
}
else if (Cs >= 3021) {
return 4;
}
}
else if (AFD >= 14) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= -24) {
if (MVDiff < 43) {
if (posBalance < 69) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (refDCval < 90) {
if (posBalance < 51) {
if (mcTcor < 90) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (diffTSC < -1) {
if (Rs < 2097) {
return 8;
}
else if (Rs >= 2097) {
return 4;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (mcTcor >= 90) {
return 8;
}
}
else if (posBalance >= 51) {
if (refDCval < 89) {
if (diffRsCsdiff < -16) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 8;
}
}
else if (diffRsCsdiff >= -16) {
return 8;
}
}
else if (refDCval >= 89) {
return 4;
}
}
}
else if (refDCval >= 90) {
if (Rs < 2091) {
if (SC < 3632) {
return 8;
}
else if (SC >= 3632) {
if (SC < 4075) {
return 2;
}
else if (SC >= 4075) {
return 8;
}
}
}
else if (Rs >= 2091) {
return 8;
}
}
}
}
else if (posBalance >= 69) {
if (Rs < 1893) {
return 8;
}
else if (Rs >= 1893) {
return 4;
}
}
}
else if (MVDiff >= 43) {
if (Rs < 2011) {
return 1;
}
else if (Rs >= 2011) {
if (Rs < 4646) {
return 8;
}
else if (Rs >= 4646) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (RsCsDiff < 41) {
if (Cs < 1428) {
if (TSC < 6) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
return 2;
}
}
else if (TSC >= 6) {
if (tcor < 94) {
if (posBalance < 56) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (tcor >= 94) {
if (Rs < 1602) {
return 8;
}
else if (Rs >= 1602) {
return 1;
}
}
}
}
else if (Cs >= 1428) {
if (mu_mv_mag_sq < 85) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Rs < 3125) {
return 2;
}
else if (Rs >= 3125) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (tcor < 91) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (tcor >= 91) {
if (posBalance < 56) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
else if (posBalance >= 56) {
if (SC < 4660) {
return 8;
}
else if (SC >= 4660) {
if (SC < 5772) {
return 4;
}
else if (SC >= 5772) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 41) {
if (Cs < 1760) {
if (refDCval < 80) {
if (AFD < 10) {
if (Rs < 2605) {
if (Cs < 1635) {
if (MVDiff < 2) {
if (Rs < 1894) {
return 8;
}
else if (Rs >= 1894) {
if (Rs < 2058) {
return 1;
}
else if (Rs >= 2058) {
return 2;
}
}
}
else if (MVDiff >= 2) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 1754) {
return 4;
}
else if (Rs >= 1754) {
return 8;
}
}
}
}
else if (Cs >= 1635) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
if (Cs < 1722) {
return 8;
}
else if (Cs >= 1722) {
return 4;
}
}
}
}
else if (Rs >= 2605) {
if (Cs < 1702) {
return 2;
}
else if (Cs >= 1702) {
return 8;
}
}
}
else if (AFD >= 10) {
if (SC < 3758) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Cs < 1433) {
if (Rs < 1823) {
return 4;
}
else if (Rs >= 1823) {
return 8;
}
}
else if (Cs >= 1433) {
return 4;
}
}
}
}
else if (SC >= 3758) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 4;
}
}
}
}
else if (refDCval >= 80) {
if (TSC < 8) {
if (ssDCval < 99) {
if (RsDiff < 8) {
if (TSC < 7) {
if (posBalance < 54) {
if (Rs < 3048) {
if (MVDiff < 3) {
if (Rs < 2055) {
return 1;
}
else if (Rs >= 2055) {
return 4;
}
}
else if (MVDiff >= 3) {
if (refDCval < 89) {
return 8;
}
else if (refDCval >= 89) {
if (Rs < 2217) {
return 2;
}
else if (Rs >= 2217) {
return 4;
}
}
}
}
else if (Rs >= 3048) {
return 1;
}
}
else if (posBalance >= 54) {
if (refDCval < 84) {
if (Rs < 2266) {
return 1;
}
else if (Rs >= 2266) {
return 8;
}
}
else if (refDCval >= 84) {
return 2;
}
}
}
else if (TSC >= 7) {
if (diffRsCsdiff < -10) {
if (Rs < 1596) {
return 2;
}
else if (Rs >= 1596) {
if (diffMVdiffVal < -8) {
return 2;
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
}
else if (diffRsCsdiff >= -10) {
if (AFD < 10) {
if (ssDCval < 84) {
if (Cs < 1585) {
if (Cs < 1372) {
return 8;
}
else if (Cs >= 1372) {
return 1;
}
}
else if (Cs >= 1585) {
if (Rs < 2481) {
return 8;
}
else if (Rs >= 2481) {
return 1;
}
}
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (AFD >= 10) {
if (diffRsCsdiff < -6) {
return 1;
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
}
}
}
else if (RsDiff >= 8) {
if (Rs < 2907) {
return 4;
}
else if (Rs >= 2907) {
return 2;
}
}
}
else if (ssDCval >= 99) {
if (Rs < 3100) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (MVDiff < 24) {
if (Rs < 3058) {
return 8;
}
else if (Rs >= 3058) {
return 4;
}
}
else if (MVDiff >= 24) {
return 4;
}
}
}
else if (Rs >= 3100) {
return 8;
}
}
}
else if (TSC >= 8) {
if (SC < 3327) {
if (RsCsDiff < 63) {
if (ssDCval < 99) {
if (SC < 3313) {
return 8;
}
else if (SC >= 3313) {
if (Rs < 1894) {
return 1;
}
else if (Rs >= 1894) {
return 8;
}
}
}
else if (ssDCval >= 99) {
if (Rs < 1608) {
return 8;
}
else if (Rs >= 1608) {
return 2;
}
}
}
else if (RsCsDiff >= 63) {
if (Rs < 1819) {
if (Rs < 1728) {
return 4;
}
else if (Rs >= 1728) {
return 8;
}
}
else if (Rs >= 1819) {
return 1;
}
}
}
else if (SC >= 3327) {
if (Rs < 4488) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (SC < 3674) {
if (Cs < 1543) {
if (Rs < 2152) {
return 8;
}
else if (Rs >= 2152) {
return 1;
}
}
else if (Cs >= 1543) {
if (Rs < 2039) {
return 1;
}
else if (Rs >= 2039) {
return 2;
}
}
}
else if (SC >= 3674) {
return 8;
}
}
}
else if (Rs >= 4488) {
return 1;
}
}
}
}
}
else if (Cs >= 1760) {
if (Rs < 2112) {
if (SC < 4108) {
if (MVDiff < 19) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
return 8;
}
}
else if (MVDiff >= 19) {
return 1;
}
}
else if (SC >= 4108) {
return 1;
}
}
else if (Rs >= 2112) {
if (MVDiff < 2) {
if (SC < 4995) {
return 8;
}
else if (SC >= 4995) {
if (Rs < 2735) {
if (Cs < 2544) {
return 8;
}
else if (Cs >= 2544) {
return 4;
}
}
else if (Rs >= 2735) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
if (mu_mv_mag_sq < 107) {
if (ssDCval < 94) {
if (negBalance < 45) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (ssDCval >= 94) {
if (mu_mv_mag_sq < 105) {
return 4;
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 3372) {
return 8;
}
else if (Rs >= 3372) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 3275) {
return 8;
}
else if (Rs >= 3275) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (CsDiff < 4) {
if (Rs < 4242) {
return 8;
}
else if (Rs >= 4242) {
return 4;
}
}
else if (CsDiff >= 4) {
if (negBalance < 55) {
if (negBalance < 39) {
if (AFD < 9) {
if (Cs < 2815) {
return 8;
}
else if (Cs >= 2815) {
return 4;
}
}
else if (AFD >= 9) {
if (Rs < 2780) {
return 8;
}
else if (Rs >= 2780) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 1;
}
}
}
}
else if (negBalance >= 39) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (Cs < 3104) {
if (ssDCval < 70) {
if (Cs < 2009) {
return 8;
}
else if (Cs >= 2009) {
return 4;
}
}
else if (ssDCval >= 70) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (Rs < 2125) {
if (Rs < 2124) {
return 8;
}
else if (Rs >= 2124) {
return 4;
}
}
else if (Rs >= 2125) {
if (MVDiff < 17) {
if (Cs < 3029) {
return 8;
}
else if (Cs >= 3029) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 4;
}
}
}
else if (MVDiff >= 17) {
return 4;
}
}
}
}
}
else if (Cs >= 3104) {
return 8;
}
}
}
}
else if (negBalance >= 55) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 110) {
if (refDCval < 165) {
if (posBalance < 56) {
if (mu_mv_mag_sq < 64) {
if (ssDCval < 136) {
if (Rs < 2862) {
if (negBalance < 49) {
if (mcTcor < 90) {
return 1;
}
else if (mcTcor >= 90) {
if (posBalance < 52) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (posBalance >= 52) {
if (Rs < 1585) {
return 4;
}
else if (Rs >= 1585) {
return 8;
}
}
}
}
else if (negBalance >= 49) {
return 8;
}
}
else if (Rs >= 2862) {
return 8;
}
}
else if (ssDCval >= 136) {
if (SC < 5778) {
if (SC < 2893) {
return 1;
}
else if (SC >= 2893) {
if (posBalance < 39) {
return 2;
}
else if (posBalance >= 39) {
if (diffMVdiffVal < -15) {
return 2;
}
else if (diffMVdiffVal >= -15) {
if (Rs < 1604) {
if (Rs < 1601) {
return 8;
}
else if (Rs >= 1601) {
return 2;
}
}
else if (Rs >= 1604) {
return 8;
}
}
}
}
}
else if (SC >= 5778) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 64) {
return 8;
}
}
else if (posBalance >= 56) {
if (mu_mv_mag_sq < 87) {
if (posBalance < 66) {
if (Rs < 2049) {
if (ssDCval < 140) {
if (CsDiff < 5) {
if (ssDCval < 124) {
if (Cs < 1549) {
return 8;
}
else if (Cs >= 1549) {
return 4;
}
}
else if (ssDCval >= 124) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
}
else if (CsDiff >= 5) {
if (diffMVdiffVal < -6) {
if (Rs < 1808) {
if (Rs < 1584) {
return 4;
}
else if (Rs >= 1584) {
return 8;
}
}
else if (Rs >= 1808) {
return 1;
}
}
else if (diffMVdiffVal >= -6) {
if (ssDCval < 135) {
return 8;
}
else if (ssDCval >= 135) {
if (tcor < 83) {
return 1;
}
else if (tcor >= 83) {
return 8;
}
}
}
}
}
else if (ssDCval >= 140) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (Rs >= 2049) {
if (diffMVdiffVal < -31) {
if (SC < 4295) {
return 8;
}
else if (SC >= 4295) {
return 2;
}
}
else if (diffMVdiffVal >= -31) {
if (Cs < 2443) {
if (ssDCval < 110) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 1;
}
}
else if (ssDCval >= 110) {
return 8;
}
}
else if (Cs >= 2443) {
if (Rs < 3510) {
return 1;
}
else if (Rs >= 3510) {
return 8;
}
}
}
}
}
else if (posBalance >= 66) {
if (Cs < 2004) {
if (Rs < 1799) {
return 4;
}
else if (Rs >= 1799) {
return 8;
}
}
else if (Cs >= 2004) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (negBalance < 47) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 8;
}
}
else if (negBalance >= 47) {
if (Rs < 2396) {
return 8;
}
else if (Rs >= 2396) {
if (Cs < 2117) {
return 1;
}
else if (Cs >= 2117) {
if (ssDCval < 118) {
return 8;
}
else if (ssDCval >= 118) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 165) {
return 2;
}
}
}
else if (CsDiff >= 7) {
if (SC < 5259) {
if (Cs < 2297) {
if (tcor < 86) {
if (negBalance < 59) {
if (ssDCval < 53) {
if (Cs < 1730) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 1700) {
if (Cs < 1536) {
return 8;
}
else if (Cs >= 1536) {
return 4;
}
}
else if (Rs >= 1700) {
return 8;
}
}
}
else if (Cs >= 1730) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
return 4;
}
}
}
else if (ssDCval >= 53) {
if (mcTcor < 90) {
if (Rs < 3338) {
if (MVDiff < 72) {
if (SC < 4940) {
return 8;
}
else if (SC >= 4940) {
if (SC < 4956) {
return 4;
}
else if (SC >= 4956) {
return 8;
}
}
}
else if (MVDiff >= 72) {
if (Rs < 1671) {
return 8;
}
else if (Rs >= 1671) {
return 1;
}
}
}
else if (Rs >= 3338) {
return 2;
}
}
else if (mcTcor >= 90) {
if (ssDCval < 88) {
if (refDCval < 82) {
if (negBalance < 48) {
if (SC < 3574) {
if (Cs < 1583) {
return 4;
}
else if (Cs >= 1583) {
return 8;
}
}
else if (SC >= 3574) {
if (Rs < 2397) {
return 1;
}
else if (Rs >= 2397) {
return 4;
}
}
}
else if (negBalance >= 48) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (refDCval >= 82) {
if (Cs < 1533) {
return 8;
}
else if (Cs >= 1533) {
return 4;
}
}
}
else if (ssDCval >= 88) {
if (Rs < 1600) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (Rs >= 1600) {
if (TSC < 7) {
if (Rs < 2451) {
return 2;
}
else if (Rs >= 2451) {
return 8;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 59) {
if (ssDCval < 91) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
if (Rs < 2072) {
return 2;
}
else if (Rs >= 2072) {
return 4;
}
}
}
else if (ssDCval >= 91) {
if (Rs < 1742) {
return 2;
}
else if (Rs >= 1742) {
return 1;
}
}
}
}
else if (tcor >= 86) {
if (gchDC < 2) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
if (Rs < 2842) {
return 8;
}
else if (Rs >= 2842) {
if (Rs < 2856) {
return 4;
}
else if (Rs >= 2856) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (gchDC < 3) {
if (Rs < 2030) {
return 2;
}
else if (Rs >= 2030) {
return 8;
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
}
else if (Cs >= 2297) {
if (mcTcor < 94) {
if (posBalance < 61) {
if (Rs < 1821) {
return 4;
}
else if (Rs >= 1821) {
return 8;
}
}
else if (posBalance >= 61) {
if (Rs < 2358) {
return 4;
}
else if (Rs >= 2358) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 2305) {
return 1;
}
else if (Rs >= 2305) {
if (Rs < 2577) {
return 2;
}
else if (Rs >= 2577) {
return 8;
}
}
}
}
}
else if (SC >= 5259) {
if (mu_mv_mag_sq < 74) {
if (Cs < 2605) {
return 8;
}
else if (Cs >= 2605) {
if (Rs < 3400) {
if (SC < 6242) {
return 4;
}
else if (SC >= 6242) {
return 1;
}
}
else if (Rs >= 3400) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 2465) {
if (MVDiff < 7) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (Cs >= 2465) {
if (refDCval < 102) {
if (ssDCval < 81) {
if (negBalance < 46) {
if (Rs < 3940) {
if (Rs < 3156) {
return 8;
}
else if (Rs >= 3156) {
return 1;
}
}
else if (Rs >= 3940) {
return 4;
}
}
else if (negBalance >= 46) {
if (Rs < 2694) {
return 4;
}
else if (Rs >= 2694) {
return 8;
}
}
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (refDCval >= 102) {
if (Rs < 4285) {
return 8;
}
else if (Rs >= 4285) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (RsCsDiff < 33) {
if (SC < 4038) {
if (posBalance < 61) {
if (tcor < 95) {
if (refDCval < 80) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 8;
}
}
else if (refDCval >= 80) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 2;
}
}
}
else if (tcor >= 95) {
if (Cs < 1335) {
if (negBalance < 39) {
return 2;
}
else if (negBalance >= 39) {
return 8;
}
}
else if (Cs >= 1335) {
return 8;
}
}
}
else if (posBalance >= 61) {
if (Rs < 2104) {
return 2;
}
else if (Rs >= 2104) {
return 8;
}
}
}
else if (SC >= 4038) {
if (refDCval < 109) {
if (Cs < 3109) {
if (RsCsDiff < 16) {
if (posBalance < 14) {
if (Cs < 2065) {
if (ssDCval < 103) {
if (tcor < 98) {
if (Rs < 2166) {
return 1;
}
else if (Rs >= 2166) {
return 2;
}
}
else if (tcor >= 98) {
return 1;
}
}
else if (ssDCval >= 103) {
return 4;
}
}
else if (Cs >= 2065) {
return 8;
}
}
else if (posBalance >= 14) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 2901) {
return 1;
}
else if (Rs >= 2901) {
return 8;
}
}
}
}
else if (RsCsDiff >= 16) {
if (SC < 4959) {
if (ssDCval < 106) {
if (MVDiff < 5) {
if (Rs < 2643) {
if (SC < 4046) {
return 4;
}
else if (SC >= 4046) {
return 8;
}
}
else if (Rs >= 2643) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 4;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (ssDCval >= 106) {
return 4;
}
}
else if (SC >= 4959) {
if (SC < 6345) {
if (ssDCval < 102) {
if (gchDC < 2) {
if (SC < 5876) {
return 4;
}
else if (SC >= 5876) {
if (SC < 5928) {
return 8;
}
else if (SC >= 5928) {
return 4;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (ssDCval >= 102) {
return 8;
}
}
else if (SC >= 6345) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Cs < 2031) {
return 8;
}
else if (Cs >= 2031) {
return 4;
}
}
}
}
}
}
else if (Cs >= 3109) {
return 8;
}
}
else if (refDCval >= 109) {
if (CsDiff < 3) {
if (Cs < 1990) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (Cs < 1336) {
return 2;
}
else if (Cs >= 1336) {
return 8;
}
}
}
else if (Cs >= 1990) {
if (mu_mv_mag_sq < 109) {
if (SC < 4309) {
return 4;
}
else if (SC >= 4309) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
if (Cs < 2778) {
return 8;
}
else if (Cs >= 2778) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 2701) {
if (Rs < 2524) {
return 8;
}
else if (Rs >= 2524) {
return 4;
}
}
else if (Rs >= 2701) {
return 2;
}
}
}
}
else if (CsDiff >= 3) {
if (Rs < 4936) {
return 8;
}
else if (Rs >= 4936) {
if (Rs < 5224) {
return 2;
}
else if (Rs >= 5224) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 33) {
if (refDCval < 110) {
if (Cs < 2638) {
if (Rs < 2935) {
if (Rs < 1891) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
if (Cs < 1509) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Cs < 1365) {
return 8;
}
else if (Cs >= 1365) {
return 2;
}
}
}
else if (Cs >= 1509) {
if (ssDCval < 87) {
return 4;
}
else if (ssDCval >= 87) {
return 2;
}
}
}
}
else if (Rs >= 1891) {
if (posBalance < 42) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (posBalance >= 42) {
if (ssDCval < 103) {
if (diffRsCsdiff < -23) {
if (TSC < 7) {
if (AFD < 13) {
if (negBalance < 39) {
return 2;
}
else if (negBalance >= 39) {
if (Rs < 2072) {
return 2;
}
else if (Rs >= 2072) {
return 1;
}
}
}
else if (AFD >= 13) {
return 4;
}
}
else if (TSC >= 7) {
if (posBalance < 54) {
if (Rs < 2586) {
return 1;
}
else if (Rs >= 2586) {
return 8;
}
}
else if (posBalance >= 54) {
if (Rs < 2106) {
return 2;
}
else if (Rs >= 2106) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -23) {
if (refDCval < 88) {
if (ssDCval < 86) {
if (posBalance < 49) {
if (Rs < 2622) {
return 8;
}
else if (Rs >= 2622) {
return 1;
}
}
else if (posBalance >= 49) {
if (negBalance < 42) {
if (Cs < 1489) {
if (Cs < 1423) {
return 4;
}
else if (Cs >= 1423) {
return 2;
}
}
else if (Cs >= 1489) {
if (MVDiff < 3) {
if (TSC < 5) {
if (Rs < 2140) {
return 8;
}
else if (Rs >= 2140) {
return 4;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Cs < 1565) {
return 8;
}
else if (Cs >= 1565) {
return 4;
}
}
}
}
else if (negBalance >= 42) {
if (negBalance < 45) {
if (Cs < 1569) {
return 1;
}
else if (Cs >= 1569) {
if (Rs < 2385) {
return 8;
}
else if (Rs >= 2385) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (diffAFD < -2) {
if (Rs < 2164) {
return 1;
}
else if (Rs >= 2164) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (Rs < 2230) {
return 2;
}
else if (Rs >= 2230) {
return 1;
}
}
}
else if (refDCval >= 88) {
if (Rs < 1940) {
if (Cs < 1486) {
return 1;
}
else if (Cs >= 1486) {
if (Cs < 1685) {
return 8;
}
else if (Cs >= 1685) {
return 2;
}
}
}
else if (Rs >= 1940) {
if (diffRsCsdiff < 37) {
if (negBalance < 52) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (negBalance >= 52) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 2369) {
return 2;
}
else if (Rs >= 2369) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 37) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 103) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
return 8;
}
}
}
}
}
else if (Rs >= 2935) {
if (refDCval < 91) {
if (RsDiff < 8) {
if (CsDiff < 6) {
if (Rs < 3755) {
if (Cs < 1913) {
if (RsDiff < 6) {
if (mu_mv_mag_sq < 99) {
return 4;
}
else if (mu_mv_mag_sq >= 99) {
if (Cs < 1517) {
return 2;
}
else if (Cs >= 1517) {
if (Rs < 3117) {
return 8;
}
else if (Rs >= 3117) {
return 4;
}
}
}
}
else if (RsDiff >= 6) {
if (mu_mv_mag_sq < 83) {
if (Cs < 1455) {
return 2;
}
else if (Cs >= 1455) {
return 4;
}
}
else if (mu_mv_mag_sq >= 83) {
return 2;
}
}
}
else if (Cs >= 1913) {
if (Cs < 2139) {
if (Rs < 3086) {
return 1;
}
else if (Rs >= 3086) {
return 8;
}
}
else if (Cs >= 2139) {
return 4;
}
}
}
else if (Rs >= 3755) {
if (AFD < 7) {
if (Rs < 4398) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (Rs >= 4398) {
if (Cs < 1934) {
return 2;
}
else if (Cs >= 1934) {
return 8;
}
}
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (CsDiff >= 6) {
if (diffRsCsdiff < -21) {
if (Cs < 1801) {
return 8;
}
else if (Cs >= 1801) {
return 1;
}
}
else if (diffRsCsdiff >= -21) {
if (ssDCval < 80) {
if (Rs < 3751) {
return 2;
}
else if (Rs >= 3751) {
return 4;
}
}
else if (ssDCval >= 80) {
return 8;
}
}
}
}
else if (RsDiff >= 8) {
if (Cs < 1909) {
return 2;
}
else if (Cs >= 1909) {
if (Rs < 4247) {
return 1;
}
else if (Rs >= 4247) {
return 4;
}
}
}
}
else if (refDCval >= 91) {
if (tcor < 92) {
if (Rs < 3062) {
return 4;
}
else if (Rs >= 3062) {
if (ssDCval < 99) {
return 1;
}
else if (ssDCval >= 99) {
return 8;
}
}
}
else if (tcor >= 92) {
if (Rs < 4215) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 84) {
if (Rs < 3055) {
if (Rs < 2956) {
return 8;
}
else if (Rs >= 2956) {
return 4;
}
}
else if (Rs >= 3055) {
return 8;
}
}
else if (mu_mv_mag_sq >= 84) {
if (Cs < 1688) {
return 8;
}
else if (Cs >= 1688) {
return 2;
}
}
}
}
else if (Rs >= 4215) {
if (Cs < 2012) {
return 8;
}
else if (Cs >= 2012) {
return 2;
}
}
}
}
}
}
else if (Cs >= 2638) {
if (Rs < 3437) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
else if (Rs >= 3437) {
if (Rs < 3539) {
if (Rs < 3493) {
return 8;
}
else if (Rs >= 3493) {
return 4;
}
}
else if (Rs >= 3539) {
return 8;
}
}
}
}
else if (refDCval >= 110) {
if (mcTcor < 98) {
if (Rs < 4808) {
if (Rs < 1845) {
if (Rs < 1638) {
return 2;
}
else if (Rs >= 1638) {
if (Rs < 1767) {
return 8;
}
else if (Rs >= 1767) {
return 4;
}
}
}
else if (Rs >= 1845) {
if (RsDiff < 5) {
if (Cs < 1487) {
return 1;
}
else if (Cs >= 1487) {
return 8;
}
}
else if (RsDiff >= 5) {
return 8;
}
}
}
else if (Rs >= 4808) {
if (Rs < 6292) {
if (Rs < 5499) {
return 2;
}
else if (Rs >= 5499) {
return 8;
}
}
else if (Rs >= 6292) {
return 4;
}
}
}
else if (mcTcor >= 98) {
if (Cs < 2030) {
if (Rs < 3485) {
if (diffMVdiffVal < -1) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
}
else if (Rs >= 3485) {
return 2;
}
}
else if (Cs >= 2030) {
if (Rs < 5791) {
return 8;
}
else if (Rs >= 5791) {
return 4;
}
}
}
}
}
}
}
}
}
else if (CsDiff >= 8) {
if (Rs < 3802) {
if (refDCval < 93) {
if (Rs < 1531) {
if (TSC < 5) {
if (Cs < 2128) {
if (Rs < 465) {
return 4;
}
else if (Rs >= 465) {
return 1;
}
}
else if (Cs >= 2128) {
return 8;
}
}
else if (TSC >= 5) {
if (MVDiff < 57) {
if (tcor < 67) {
if (negBalance < 51) {
if (posBalance < 69) {
if (diffRsCsdiff < -72) {
if (Rs < 1027) {
return 4;
}
else if (Rs >= 1027) {
if (Rs < 1490) {
return 8;
}
else if (Rs >= 1490) {
return 1;
}
}
}
else if (diffRsCsdiff >= -72) {
if (MVDiff < 49) {
return 2;
}
else if (MVDiff >= 49) {
if (Rs < 780) {
return 4;
}
else if (Rs >= 780) {
return 2;
}
}
}
}
else if (posBalance >= 69) {
if (MVDiff < 46) {
return 4;
}
else if (MVDiff >= 46) {
if (Rs < 1105) {
return 8;
}
else if (Rs >= 1105) {
return 1;
}
}
}
}
else if (negBalance >= 51) {
if (Cs < 1408) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 4;
}
}
}
else if (Cs >= 1408) {
if (diffRsCsdiff < 55) {
if (Cs < 4111) {
if (SC < 2980) {
if (Cs < 1734) {
if (posBalance < 64) {
return 4;
}
else if (posBalance >= 64) {
return 2;
}
}
else if (Cs >= 1734) {
return 2;
}
}
else if (SC >= 2980) {
return 4;
}
}
else if (Cs >= 4111) {
return 8;
}
}
else if (diffRsCsdiff >= 55) {
if (Rs < 1384) {
if (Rs < 1347) {
return 4;
}
else if (Rs >= 1347) {
return 8;
}
}
else if (Rs >= 1384) {
return 2;
}
}
}
}
}
else if (tcor >= 67) {
if (Rs < 1371) {
if (refDCval < 89) {
if (AFD < 18) {
if (posBalance < 51) {
if (Rs < 1328) {
return 8;
}
else if (Rs >= 1328) {
return 4;
}
}
else if (posBalance >= 51) {
if (negBalance < 50) {
if (tcor < 84) {
if (Rs < 899) {
return 4;
}
else if (Rs >= 899) {
if (Cs < 2118) {
return 8;
}
else if (Cs >= 2118) {
return 4;
}
}
}
else if (tcor >= 84) {
if (diffRsCsdiff < 42) {
return 8;
}
else if (diffRsCsdiff >= 42) {
if (Rs < 701) {
return 1;
}
else if (Rs >= 701) {
return 2;
}
}
}
}
else if (negBalance >= 50) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 2;
}
}
}
}
else if (AFD >= 18) {
if (RsCsDiff < 110) {
if (Cs < 1555) {
return 8;
}
else if (Cs >= 1555) {
return 4;
}
}
else if (RsCsDiff >= 110) {
if (RsCsDiff < 133) {
return 2;
}
else if (RsCsDiff >= 133) {
if (Rs < 1098) {
return 8;
}
else if (Rs >= 1098) {
if (Rs < 1314) {
return 2;
}
else if (Rs >= 1314) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 89) {
return 4;
}
}
else if (Rs >= 1371) {
if (diffRsCsdiff < -9) {
if (CsDiff < 14) {
return 1;
}
else if (CsDiff >= 14) {
return 2;
}
}
else if (diffRsCsdiff >= -9) {
if (AFD < 16) {
if (Rs < 1386) {
return 1;
}
else if (Rs >= 1386) {
return 8;
}
}
else if (AFD >= 16) {
if (Cs < 1512) {
return 2;
}
else if (Cs >= 1512) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 57) {
if (mcTcor < 85) {
if (CsDiff < 13) {
return 8;
}
else if (CsDiff >= 13) {
if (Rs < 960) {
return 4;
}
else if (Rs >= 960) {
if (Rs < 1106) {
return 1;
}
else if (Rs >= 1106) {
return 8;
}
}
}
}
else if (mcTcor >= 85) {
if (mu_mv_mag_sq < 27) {
if (refDCval < 46) {
return 1;
}
else if (refDCval >= 46) {
return 2;
}
}
else if (mu_mv_mag_sq >= 27) {
if (Rs < 652) {
return 8;
}
else if (Rs >= 652) {
if (Rs < 867) {
return 4;
}
else if (Rs >= 867) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 1531) {
if (Cs < 2803) {
if (TSC < 6) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
if (Rs < 3359) {
return 1;
}
else if (Rs >= 3359) {
return 8;
}
}
}
else if (TSC >= 6) {
if (AFD < 17) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 32) {
if (Rs < 2427) {
if (Rs < 2268) {
return 1;
}
else if (Rs >= 2268) {
return 8;
}
}
else if (Rs >= 2427) {
return 2;
}
}
else if (mu_mv_mag_sq >= 32) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Rs < 3349) {
if (SC < 5608) {
return 8;
}
else if (SC >= 5608) {
if (Rs < 3204) {
return 4;
}
else if (Rs >= 3204) {
return 8;
}
}
}
else if (Rs >= 3349) {
return 2;
}
}
}
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 69) {
if (ssDCval < 52) {
if (Cs < 1707) {
if (Rs < 1584) {
return 4;
}
else if (Rs >= 1584) {
return 8;
}
}
else if (Cs >= 1707) {
if (Cs < 1951) {
return 4;
}
else if (Cs >= 1951) {
return 2;
}
}
}
else if (ssDCval >= 52) {
if (SC < 3886) {
if (SC < 3883) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
if (SC < 3622) {
if (Cs < 1482) {
return 8;
}
else if (Cs >= 1482) {
return 1;
}
}
else if (SC >= 3622) {
return 8;
}
}
}
else if (SC >= 3883) {
return 1;
}
}
else if (SC >= 3886) {
if (Rs < 1729) {
return 4;
}
else if (Rs >= 1729) {
if (MVDiff < 38) {
return 8;
}
else if (MVDiff >= 38) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (ssDCval < 85) {
if (Cs < 2304) {
if (diffRsCsdiff < 8) {
return 4;
}
else if (diffRsCsdiff >= 8) {
if (mu_mv_mag_sq < 82) {
return 8;
}
else if (mu_mv_mag_sq >= 82) {
if (Rs < 2555) {
return 4;
}
else if (Rs >= 2555) {
return 8;
}
}
}
}
else if (Cs >= 2304) {
if (Rs < 3163) {
if (diffRsCsdiff < -35) {
return 4;
}
else if (diffRsCsdiff >= -35) {
return 8;
}
}
else if (Rs >= 3163) {
return 4;
}
}
}
else if (ssDCval >= 85) {
return 8;
}
}
}
}
else if (AFD >= 17) {
if (tcor < 80) {
if (Rs < 2177) {
if (mu_mv_mag_sq < 29) {
if (tcor < 78) {
if (Rs < 1660) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (Rs < 1632) {
if (Cs < 1465) {
return 2;
}
else if (Cs >= 1465) {
return 4;
}
}
else if (Rs >= 1632) {
return 1;
}
}
}
else if (Rs >= 1660) {
if (Rs < 2126) {
return 4;
}
else if (Rs >= 2126) {
return 1;
}
}
}
else if (tcor >= 78) {
if (Rs < 1849) {
return 1;
}
else if (Rs >= 1849) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
if (Rs < 1752) {
return 8;
}
else if (Rs >= 1752) {
return 2;
}
}
}
}
else if (Rs >= 2177) {
if (negBalance < 53) {
if (ssDCval < 80) {
if (Rs < 2284) {
return 1;
}
else if (Rs >= 2284) {
return 2;
}
}
else if (ssDCval >= 80) {
return 4;
}
}
else if (negBalance >= 53) {
return 8;
}
}
}
else if (tcor >= 80) {
if (negBalance < 51) {
if (refDCval < 88) {
return 2;
}
else if (refDCval >= 88) {
return 1;
}
}
else if (negBalance >= 51) {
return 8;
}
}
}
}
}
else if (Cs >= 2803) {
if (negBalance < 48) {
if (Rs < 2849) {
return 4;
}
else if (Rs >= 2849) {
if (MVDiff < 6) {
if (Rs < 2961) {
if (Rs < 2942) {
return 8;
}
else if (Rs >= 2942) {
return 4;
}
}
else if (Rs >= 2961) {
if (Rs < 3599) {
return 1;
}
else if (Rs >= 3599) {
return 8;
}
}
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (negBalance >= 48) {
if (mcTcor < 93) {
if (MVDiff < 12) {
if (SC < 6716) {
if (refDCval < 89) {
if (TSC < 11) {
if (diffRsCsdiff < -25) {
if (Cs < 3041) {
return 1;
}
else if (Cs >= 3041) {
return 8;
}
}
else if (diffRsCsdiff >= -25) {
if (Rs < 2823) {
if (Rs < 2271) {
return 1;
}
else if (Rs >= 2271) {
return 4;
}
}
else if (Rs >= 2823) {
if (Rs < 3196) {
return 1;
}
else if (Rs >= 3196) {
return 8;
}
}
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (refDCval >= 89) {
return 4;
}
}
else if (SC >= 6716) {
return 8;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (mcTcor >= 93) {
if (SC < 6675) {
if (RsDiff < 7) {
if (Rs < 2800) {
return 4;
}
else if (Rs >= 2800) {
return 1;
}
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (SC >= 6675) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 93) {
if (AFD < 20) {
if (Rs < 2990) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 30) {
if (posBalance < 58) {
if (negBalance < 26) {
return 2;
}
else if (negBalance >= 26) {
if (MVDiff < 24) {
if (SC < 2852) {
return 2;
}
else if (SC >= 2852) {
return 1;
}
}
else if (MVDiff >= 24) {
if (Rs < 978) {
return 1;
}
else if (Rs >= 978) {
return 8;
}
}
}
}
else if (posBalance >= 58) {
if (SC < 3375) {
if (Rs < 1241) {
return 2;
}
else if (Rs >= 1241) {
return 4;
}
}
else if (SC >= 3375) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (ssDCval < 98) {
return 1;
}
else if (ssDCval >= 98) {
if (diffMVdiffVal < -2) {
if (tcor < 89) {
if (Rs < 1883) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (Rs >= 1883) {
if (Rs < 2521) {
return 1;
}
else if (Rs >= 2521) {
return 4;
}
}
}
else if (tcor >= 89) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
}
else if (negBalance >= 44) {
if (mcTcor < 79) {
if (TSC < 14) {
if (Rs < 2098) {
return 2;
}
else if (Rs >= 2098) {
return 8;
}
}
else if (TSC >= 14) {
return 1;
}
}
else if (mcTcor >= 79) {
if (Rs < 980) {
if (Rs < 929) {
return 2;
}
else if (Rs >= 929) {
return 1;
}
}
else if (Rs >= 980) {
if (TSC < 11) {
if (mcTcor < 95) {
if (Rs < 1004) {
return 2;
}
else if (Rs >= 1004) {
return 8;
}
}
else if (mcTcor >= 95) {
if (Rs < 2009) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (RsDiff < 7) {
return 2;
}
else if (RsDiff >= 7) {
if (Rs < 1912) {
return 8;
}
else if (Rs >= 1912) {
return 2;
}
}
}
}
else if (Rs >= 2009) {
return 8;
}
}
}
else if (TSC >= 11) {
if (MVDiff < 10) {
if (posBalance < 61) {
if (TSC < 13) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (TSC >= 13) {
return 4;
}
}
else if (posBalance >= 61) {
return 4;
}
}
else if (MVDiff >= 10) {
if (Cs < 5331) {
if (SC < 2705) {
return 2;
}
else if (SC >= 2705) {
return 8;
}
}
else if (Cs >= 5331) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 2990) {
if (ssDCval < 103) {
if (Rs < 3036) {
return 8;
}
else if (Rs >= 3036) {
return 4;
}
}
else if (ssDCval >= 103) {
if (Rs < 3116) {
if (ssDCval < 121) {
if (Rs < 3034) {
return 4;
}
else if (Rs >= 3034) {
return 8;
}
}
else if (ssDCval >= 121) {
return 1;
}
}
else if (Rs >= 3116) {
return 8;
}
}
}
}
else if (AFD >= 20) {
if (ssDCval < 117) {
if (Rs < 1654) {
if (TSC < 11) {
if (AFD < 25) {
if (Cs < 1898) {
return 1;
}
else if (Cs >= 1898) {
if (ssDCval < 100) {
return 8;
}
else if (ssDCval >= 100) {
return 2;
}
}
}
else if (AFD >= 25) {
if (ssDCval < 97) {
return 1;
}
else if (ssDCval >= 97) {
return 4;
}
}
}
else if (TSC >= 11) {
if (Rs < 949) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 2;
}
}
else if (Rs >= 949) {
if (mu_mv_mag_sq < 28) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 28) {
if (SC < 3583) {
return 4;
}
else if (SC >= 3583) {
return 8;
}
}
}
}
}
else if (Rs >= 1654) {
if (Cs < 2526) {
if (Rs < 1693) {
return 4;
}
else if (Rs >= 1693) {
return 8;
}
}
else if (Cs >= 2526) {
return 4;
}
}
}
else if (ssDCval >= 117) {
if (MVDiff < 18) {
if (Rs < 1916) {
return 2;
}
else if (Rs >= 1916) {
return 8;
}
}
else if (MVDiff >= 18) {
if (CsDiff < 10) {
if (posBalance < 61) {
return 1;
}
else if (posBalance >= 61) {
if (Rs < 1229) {
return 8;
}
else if (Rs >= 1229) {
return 4;
}
}
}
else if (CsDiff >= 10) {
if (CsDiff < 12) {
return 2;
}
else if (CsDiff >= 12) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 3802) {
if (AFD < 12) {
if (TSC < 9) {
if (ssDCval < 117) {
if (Rs < 4391) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (Rs >= 4391) {
if (Cs < 4019) {
if (diffMVdiffVal < -8) {
return 8;
}
else if (diffMVdiffVal >= -8) {
if (Cs < 1650) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
if (Rs < 4502) {
return 1;
}
else if (Rs >= 4502) {
return 8;
}
}
}
else if (Cs >= 1650) {
return 1;
}
}
}
else if (Cs >= 4019) {
if (Rs < 5971) {
return 8;
}
else if (Rs >= 5971) {
return 2;
}
}
}
}
else if (ssDCval >= 117) {
return 8;
}
}
else if (TSC >= 9) {
if (mu_mv_mag_sq < 95) {
if (diffAFD < -1) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 1;
}
}
else if (diffAFD >= -1) {
if (mu_mv_mag_sq < 79) {
if (Rs < 4552) {
return 1;
}
else if (Rs >= 4552) {
return 8;
}
}
else if (mu_mv_mag_sq >= 79) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 95) {
return 1;
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 83) {
return 8;
}
else if (mu_mv_mag_sq >= 83) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
if (Cs < 2096) {
if (tcor < 88) {
if (RsDiff < 15) {
if (ssDCval < 86) {
if (mcTcor < 89) {
if (MVDiff < 23) {
if (negBalance < 44) {
if (ssDCval < 42) {
return 8;
}
else if (ssDCval >= 42) {
if (Cs < 1849) {
if (MVDiff < 12) {
if (Rs < 2786) {
return 8;
}
else if (Rs >= 2786) {
return 1;
}
}
else if (MVDiff >= 12) {
if (Rs < 1577) {
return 4;
}
else if (Rs >= 1577) {
return 1;
}
}
}
else if (Cs >= 1849) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 3825) {
return 8;
}
else if (Rs >= 3825) {
return 1;
}
}
}
}
}
else if (negBalance >= 44) {
if (tcor < 69) {
if (negBalance < 57) {
if (diffMVdiffVal < 2) {
if (RsDiff < 12) {
if (posBalance < 63) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 4;
}
}
else if (posBalance >= 63) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 8;
}
}
}
else if (RsDiff >= 12) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
if (Cs < 1591) {
return 2;
}
else if (Cs >= 1591) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1954) {
return 4;
}
else if (Rs >= 1954) {
return 8;
}
}
}
else if (negBalance >= 57) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
return 4;
}
}
}
else if (tcor >= 69) {
if (refDCval < 67) {
return 8;
}
else if (refDCval >= 67) {
if (negBalance < 53) {
if (diffRsCsdiff < 21) {
if (Rs < 2656) {
if (refDCval < 74) {
return 1;
}
else if (refDCval >= 74) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 2;
}
}
}
else if (Rs >= 2656) {
return 8;
}
}
else if (diffRsCsdiff >= 21) {
if (Rs < 3046) {
return 4;
}
else if (Rs >= 3046) {
return 8;
}
}
}
else if (negBalance >= 53) {
if (Rs < 2480) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
else if (Rs >= 2480) {
if (Rs < 3834) {
return 4;
}
else if (Rs >= 3834) {
return 8;
}
}
}
}
}
}
}
else if (MVDiff >= 23) {
if (mcTcor < 71) {
if (gchDC < 11) {
if (diffTSC < 2) {
return 4;
}
else if (diffTSC >= 2) {
return 2;
}
}
else if (gchDC >= 11) {
if (Rs < 1563) {
return 4;
}
else if (Rs >= 1563) {
return 8;
}
}
}
else if (mcTcor >= 71) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 68) {
if (Rs < 2292) {
return 1;
}
else if (Rs >= 2292) {
return 8;
}
}
else if (ssDCval >= 68) {
if (RsCsDiff < 198) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
if (Rs < 1848) {
return 1;
}
else if (Rs >= 1848) {
return 8;
}
}
}
else if (RsCsDiff >= 198) {
if (diffRsCsdiff < 9) {
if (CsDiff < 9) {
return 2;
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (diffRsCsdiff >= 9) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (posBalance < 62) {
if (RsCsDiff < 193) {
if (negBalance < 45) {
if (Rs < 3020) {
return 1;
}
else if (Rs >= 3020) {
return 2;
}
}
else if (negBalance >= 45) {
if (AFD < 22) {
if (Cs < 1328) {
return 1;
}
else if (Cs >= 1328) {
return 8;
}
}
else if (AFD >= 22) {
return 1;
}
}
}
else if (RsCsDiff >= 193) {
if (mcTcor < 81) {
return 1;
}
else if (mcTcor >= 81) {
if (tcor < 60) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
if (Rs < 1372) {
return 4;
}
else if (Rs >= 1372) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
return 4;
}
}
}
}
else if (tcor >= 60) {
if (Cs < 1703) {
return 2;
}
else if (Cs >= 1703) {
if (MVDiff < 29) {
if (SC < 4762) {
return 8;
}
else if (SC >= 4762) {
return 4;
}
}
else if (MVDiff >= 29) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 62) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (RsCsDiff < 173) {
if (Rs < 2456) {
return 1;
}
else if (Rs >= 2456) {
return 8;
}
}
else if (RsCsDiff >= 173) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 89) {
if (negBalance < 48) {
if (Rs < 3220) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (Rs < 2693) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (RsCsDiff < 237) {
if (ssDCval < 85) {
if (mcTcor < 90) {
return 2;
}
else if (mcTcor >= 90) {
if (MVDiff < 43) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
return 4;
}
}
else if (MVDiff >= 43) {
return 2;
}
}
}
else if (ssDCval >= 85) {
return 2;
}
}
else if (RsCsDiff >= 237) {
if (RsCsDiff < 329) {
return 2;
}
else if (RsCsDiff >= 329) {
return 4;
}
}
}
}
else if (Rs >= 2693) {
if (MVDiff < 5) {
if (Cs < 1651) {
return 2;
}
else if (Cs >= 1651) {
return 1;
}
}
else if (MVDiff >= 5) {
if (RsCsDiff < 124) {
return 4;
}
else if (RsCsDiff >= 124) {
return 2;
}
}
}
}
}
else if (Rs >= 3220) {
if (posBalance < 47) {
if (diffTSC < -2) {
return 8;
}
else if (diffTSC >= -2) {
if (Cs < 1753) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
return 4;
}
}
else if (Cs >= 1753) {
return 4;
}
}
}
else if (posBalance >= 47) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
return 2;
}
}
}
}
else if (negBalance >= 48) {
if (mu_mv_mag_sq < 9) {
if (RsCsDiff < 183) {
if (RsCsDiff < 163) {
if (TSC < 11) {
if (Rs < 2724) {
if (refDCval < 74) {
if (Rs < 2071) {
return 2;
}
else if (Rs >= 2071) {
return 8;
}
}
else if (refDCval >= 74) {
if (negBalance < 60) {
if (Rs < 1484) {
return 8;
}
else if (Rs >= 1484) {
if (ssDCval < 85) {
return 4;
}
else if (ssDCval >= 85) {
return 8;
}
}
}
else if (negBalance >= 60) {
return 2;
}
}
}
else if (Rs >= 2724) {
return 8;
}
}
else if (TSC >= 11) {
if (Rs < 1489) {
return 2;
}
else if (Rs >= 1489) {
return 8;
}
}
}
else if (RsCsDiff >= 163) {
if (Cs < 1938) {
if (Rs < 1426) {
return 8;
}
else if (Rs >= 1426) {
return 4;
}
}
else if (Cs >= 1938) {
if (Rs < 2796) {
return 8;
}
else if (Rs >= 2796) {
return 4;
}
}
}
}
else if (RsCsDiff >= 183) {
if (Cs < 1564) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 1500) {
return 1;
}
else if (Rs >= 1500) {
return 8;
}
}
}
else if (Cs >= 1564) {
if (SC < 4753) {
if (TSC < 11) {
if (MVDiff < 44) {
if (Rs < 2350) {
return 8;
}
else if (Rs >= 2350) {
return 4;
}
}
else if (MVDiff >= 44) {
return 4;
}
}
else if (TSC >= 11) {
return 4;
}
}
else if (SC >= 4753) {
if (RsCsDiff < 222) {
return 2;
}
else if (RsCsDiff >= 222) {
if (Rs < 3690) {
return 4;
}
else if (Rs >= 3690) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (AFD < 20) {
if (diffAFD < -2) {
if (Rs < 3625) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (Rs < 2279) {
return 8;
}
else if (Rs >= 2279) {
return 4;
}
}
}
else if (Rs >= 3625) {
return 2;
}
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 52) {
return 8;
}
else if (mu_mv_mag_sq >= 52) {
if (mu_mv_mag_sq < 59) {
if (Rs < 2849) {
return 1;
}
else if (Rs >= 2849) {
return 4;
}
}
else if (mu_mv_mag_sq >= 59) {
return 8;
}
}
}
}
else if (AFD >= 20) {
if (mcTcor < 92) {
if (ssDCval < 83) {
if (Rs < 1333) {
return 1;
}
else if (Rs >= 1333) {
return 8;
}
}
else if (ssDCval >= 83) {
return 4;
}
}
else if (mcTcor >= 92) {
if (Cs < 2032) {
return 4;
}
else if (Cs >= 2032) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (negBalance < 54) {
if (refDCval < 96) {
if (TSC < 10) {
if (AFD < 19) {
if (mu_mv_mag_sq < 18) {
if (TSC < 9) {
if (CsDiff < 9) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (Rs < 2337) {
return 2;
}
else if (Rs >= 2337) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
return 4;
}
}
}
}
else if (CsDiff >= 9) {
if (Rs < 2361) {
return 8;
}
else if (Rs >= 2361) {
return 2;
}
}
}
else if (TSC >= 9) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (Rs < 2145) {
return 8;
}
else if (Rs >= 2145) {
return 4;
}
}
}
}
else if (AFD >= 19) {
return 2;
}
}
else if (TSC >= 10) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
if (AFD < 20) {
if (Cs < 1931) {
return 8;
}
else if (Cs >= 1931) {
if (Rs < 2598) {
return 8;
}
else if (Rs >= 2598) {
return 4;
}
}
}
else if (AFD >= 20) {
if (RsCsDiff < 271) {
if (Cs < 1348) {
return 4;
}
else if (Cs >= 1348) {
if (diffRsCsdiff < -74) {
if (Rs < 2193) {
return 4;
}
else if (Rs >= 2193) {
return 1;
}
}
else if (diffRsCsdiff >= -74) {
return 8;
}
}
}
else if (RsCsDiff >= 271) {
return 4;
}
}
}
}
}
else if (refDCval >= 96) {
if (refDCval < 161) {
if (tcor < 76) {
if (mcTcor < 88) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 139) {
if (TSC < 13) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (TSC >= 13) {
return 4;
}
}
else if (ssDCval >= 139) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffRsCsdiff < 89) {
if (Cs < 1364) {
return 4;
}
else if (Cs >= 1364) {
return 8;
}
}
else if (diffRsCsdiff >= 89) {
if (Rs < 2139) {
return 1;
}
else if (Rs >= 2139) {
return 8;
}
}
}
}
else if (mcTcor >= 88) {
if (Rs < 3044) {
if (posBalance < 57) {
if (mcTcor < 90) {
if (Cs < 1578) {
return 2;
}
else if (Cs >= 1578) {
return 8;
}
}
else if (mcTcor >= 90) {
return 2;
}
}
else if (posBalance >= 57) {
if (negBalance < 43) {
return 4;
}
else if (negBalance >= 43) {
return 8;
}
}
}
else if (Rs >= 3044) {
if (ssDCval < 135) {
if (MVDiff < 38) {
return 1;
}
else if (MVDiff >= 38) {
return 8;
}
}
else if (ssDCval >= 135) {
if (Cs < 1770) {
return 4;
}
else if (Cs >= 1770) {
return 8;
}
}
}
}
}
else if (tcor >= 76) {
if (diffAFD < -1) {
if (posBalance < 56) {
if (MVDiff < 14) {
if (Rs < 3424) {
return 2;
}
else if (Rs >= 3424) {
return 8;
}
}
else if (MVDiff >= 14) {
if (MVDiff < 82) {
return 8;
}
else if (MVDiff >= 82) {
return 4;
}
}
}
else if (posBalance >= 56) {
if (ssDCval < 109) {
if (TSC < 11) {
if (diffAFD < -3) {
if (Rs < 3248) {
return 4;
}
else if (Rs >= 3248) {
return 8;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (TSC >= 11) {
if (tcor < 84) {
return 4;
}
else if (tcor >= 84) {
if (Rs < 2967) {
return 1;
}
else if (Rs >= 2967) {
return 8;
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 1868) {
return 1;
}
else if (Rs >= 1868) {
if (Cs < 2052) {
return 8;
}
else if (Cs >= 2052) {
return 1;
}
}
}
}
}
else if (diffAFD >= -1) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (posBalance < 52) {
if (Cs < 1815) {
return 2;
}
else if (Cs >= 1815) {
return 4;
}
}
else if (posBalance >= 52) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
if (Rs < 2630) {
return 2;
}
else if (Rs >= 2630) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 161) {
if (Rs < 2929) {
return 2;
}
else if (Rs >= 2929) {
if (Cs < 1588) {
return 8;
}
else if (Cs >= 1588) {
return 4;
}
}
}
}
}
else if (negBalance >= 54) {
if (refDCval < 113) {
if (mcTcor < 94) {
if (mcTcor < 87) {
if (CsDiff < 8) {
if (ssDCval < 97) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Rs < 2531) {
return 1;
}
else if (Rs >= 2531) {
return 4;
}
}
}
else if (ssDCval >= 97) {
if (Cs < 1864) {
return 4;
}
else if (Cs >= 1864) {
return 8;
}
}
}
else if (CsDiff >= 8) {
if (tcor < 62) {
if (ssDCval < 100) {
if (tcor < 52) {
return 8;
}
else if (tcor >= 52) {
return 1;
}
}
else if (ssDCval >= 100) {
return 4;
}
}
else if (tcor >= 62) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
else if (mcTcor >= 87) {
if (MVDiff < 13) {
if (RsCsDiff < 220) {
if (TSC < 9) {
if (Rs < 3740) {
if (Rs < 2829) {
return 2;
}
else if (Rs >= 2829) {
return 8;
}
}
else if (Rs >= 3740) {
return 4;
}
}
else if (TSC >= 9) {
if (Cs < 2052) {
if (Rs < 1804) {
return 4;
}
else if (Rs >= 1804) {
return 8;
}
}
else if (Cs >= 2052) {
return 4;
}
}
}
else if (RsCsDiff >= 220) {
if (RsCsDiff < 244) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (RsCsDiff >= 244) {
return 4;
}
}
}
else if (MVDiff >= 13) {
if (diffMVdiffVal < -7) {
if (Rs < 2144) {
return 4;
}
else if (Rs >= 2144) {
return 1;
}
}
else if (diffMVdiffVal >= -7) {
if (TSC < 13) {
if (Rs < 4474) {
return 8;
}
else if (Rs >= 4474) {
return 1;
}
}
else if (TSC >= 13) {
if (Rs < 2186) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 1;
}
}
else if (Rs >= 2186) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (TSC < 11) {
if (diffMVdiffVal < 3) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 1;
}
}
else if (diffMVdiffVal >= 3) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
if (refDCval < 96) {
return 2;
}
else if (refDCval >= 96) {
return 4;
}
}
}
}
else if (TSC >= 11) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
return 2;
}
}
}
}
else if (refDCval >= 113) {
if (mu_mv_mag_sq < 18) {
if (ssDCval < 142) {
if (MVDiff < 20) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (Cs < 1527) {
return 1;
}
else if (Cs >= 1527) {
return 8;
}
}
}
}
else if (MVDiff >= 20) {
if (diffMVdiffVal < 14) {
if (diffTSC < -1) {
if (SC < 3277) {
return 4;
}
else if (SC >= 3277) {
if (mcTcor < 91) {
if (Rs < 1692) {
return 8;
}
else if (Rs >= 1692) {
return 1;
}
}
else if (mcTcor >= 91) {
return 8;
}
}
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (diffMVdiffVal >= 14) {
return 8;
}
}
}
else if (ssDCval >= 142) {
if (refDCval < 162) {
if (Rs < 1161) {
return 2;
}
else if (Rs >= 1161) {
return 8;
}
}
else if (refDCval >= 162) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (posBalance < 61) {
if (Rs < 3256) {
return 8;
}
else if (Rs >= 3256) {
if (Rs < 4365) {
return 4;
}
else if (Rs >= 4365) {
return 1;
}
}
}
else if (posBalance >= 61) {
if (ssDCval < 132) {
return 8;
}
else if (ssDCval >= 132) {
if (Cs < 1380) {
return 1;
}
else if (Cs >= 1380) {
return 2;
}
}
}
}
}
}
}
}
else if (RsDiff >= 15) {
if (MVDiff < 22) {
if (Cs < 1747) {
if (tcor < 83) {
if (RsDiff < 16) {
if (diffRsCsdiff < -74) {
return 2;
}
else if (diffRsCsdiff >= -74) {
if (MVDiff < 12) {
if (SC < 4635) {
return 4;
}
else if (SC >= 4635) {
if (Rs < 3150) {
return 2;
}
else if (Rs >= 3150) {
return 8;
}
}
}
else if (MVDiff >= 12) {
if (tcor < 66) {
return 8;
}
else if (tcor >= 66) {
if (Rs < 3294) {
return 1;
}
else if (Rs >= 3294) {
return 8;
}
}
}
}
}
else if (RsDiff >= 16) {
if (mcTcor < 96) {
if (SC < 6276) {
if (SC < 4634) {
if (TSC < 14) {
if (Rs < 2068) {
return 4;
}
else if (Rs >= 2068) {
if (tcor < 77) {
return 2;
}
else if (tcor >= 77) {
return 4;
}
}
}
else if (TSC >= 14) {
return 8;
}
}
else if (SC >= 4634) {
return 2;
}
}
else if (SC >= 6276) {
if (tcor < 81) {
if (diffMVdiffVal < 1) {
if (RsDiff < 19) {
return 1;
}
else if (RsDiff >= 19) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
if (Rs < 8284) {
return 4;
}
else if (Rs >= 8284) {
return 2;
}
}
}
else if (tcor >= 81) {
if (Cs < 1580) {
return 8;
}
else if (Cs >= 1580) {
return 2;
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 4241) {
return 4;
}
else if (Rs >= 4241) {
if (Rs < 4787) {
return 1;
}
else if (Rs >= 4787) {
return 8;
}
}
}
}
}
else if (tcor >= 83) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
if (RsDiff < 17) {
return 4;
}
else if (RsDiff >= 17) {
return 2;
}
}
}
}
}
else if (Cs >= 1747) {
if (mu_mv_mag_sq < 5) {
if (AFD < 23) {
if (ssDCval < 142) {
if (diffMVdiffVal < -6) {
if (Cs < 2020) {
return 2;
}
else if (Cs >= 2020) {
return 4;
}
}
else if (diffMVdiffVal >= -6) {
return 2;
}
}
else if (ssDCval >= 142) {
return 4;
}
}
else if (AFD >= 23) {
if (Rs < 3127) {
return 4;
}
else if (Rs >= 3127) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (CsDiff < 9) {
if (diffRsCsdiff < 22) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (diffRsCsdiff >= 22) {
if (ssDCval < 103) {
return 4;
}
else if (ssDCval >= 103) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 4;
}
}
}
}
else if (CsDiff >= 9) {
if (RsCsDiff < 438) {
if (negBalance < 42) {
if (Rs < 3281) {
return 2;
}
else if (Rs >= 3281) {
return 1;
}
}
else if (negBalance >= 42) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
return 8;
}
}
}
else if (RsCsDiff >= 438) {
if (AFD < 25) {
if (Rs < 6452) {
return 2;
}
else if (Rs >= 6452) {
return 8;
}
}
else if (AFD >= 25) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Cs < 2049) {
return 4;
}
else if (Cs >= 2049) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 22) {
if (AFD < 30) {
if (mu_mv_mag_sq < 7) {
if (refDCval < 107) {
if (mcTcor < 80) {
if (Rs < 2666) {
return 8;
}
else if (Rs >= 2666) {
return 4;
}
}
else if (mcTcor >= 80) {
if (MVDiff < 52) {
if (Cs < 1354) {
return 8;
}
else if (Cs >= 1354) {
return 2;
}
}
else if (MVDiff >= 52) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 2;
}
}
}
}
else if (refDCval >= 107) {
if (diffRsCsdiff < -134) {
if (SC < 7677) {
if (negBalance < 56) {
if (Rs < 5664) {
return 2;
}
else if (Rs >= 5664) {
return 4;
}
}
else if (negBalance >= 56) {
if (Rs < 2431) {
return 2;
}
else if (Rs >= 2431) {
return 1;
}
}
}
else if (SC >= 7677) {
return 8;
}
}
else if (diffRsCsdiff >= -134) {
if (diffAFD < -8) {
return 2;
}
else if (diffAFD >= -8) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (gchDC < 1) {
if (posBalance < 58) {
if (Cs < 1421) {
return 8;
}
else if (Cs >= 1421) {
return 2;
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (gchDC >= 1) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (SC < 5188) {
if (TSC < 11) {
if (diffRsCsdiff < 240) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 2;
}
}
else if (diffRsCsdiff >= 240) {
return 1;
}
}
else if (TSC >= 11) {
if (mcTcor < 86) {
if (Cs < 1478) {
if (Cs < 1325) {
return 8;
}
else if (Cs >= 1325) {
return 2;
}
}
else if (Cs >= 1478) {
return 8;
}
}
else if (mcTcor >= 86) {
if (diffRsCsdiff < -6) {
if (AFD < 26) {
return 4;
}
else if (AFD >= 26) {
if (Rs < 3260) {
return 2;
}
else if (Rs >= 3260) {
return 1;
}
}
}
else if (diffRsCsdiff >= -6) {
if (Rs < 2404) {
return 4;
}
else if (Rs >= 2404) {
return 8;
}
}
}
}
}
else if (SC >= 5188) {
if (gchDC < 2) {
if (ssDCval < 103) {
return 4;
}
else if (ssDCval >= 103) {
if (RsDiff < 22) {
return 8;
}
else if (RsDiff >= 22) {
if (ssDCval < 137) {
if (Rs < 5909) {
return 2;
}
else if (Rs >= 5909) {
return 4;
}
}
else if (ssDCval >= 137) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 109) {
if (diffAFD < 3) {
if (MVDiff < 26) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 4;
}
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (diffAFD >= 3) {
if (Rs < 7243) {
return 8;
}
else if (Rs >= 7243) {
return 2;
}
}
}
else if (ssDCval >= 109) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
if (Rs < 4338) {
return 4;
}
else if (Rs >= 4338) {
return 2;
}
}
}
}
}
}
}
}
else if (AFD >= 30) {
if (CsDiff < 9) {
if (Rs < 3505) {
return 2;
}
else if (Rs >= 3505) {
if (negBalance < 62) {
if (Rs < 5678) {
return 4;
}
else if (Rs >= 5678) {
return 8;
}
}
else if (negBalance >= 62) {
return 2;
}
}
}
else if (CsDiff >= 9) {
if (tcor < 54) {
if (mcTcor < 75) {
if (MVDiff < 104) {
return 8;
}
else if (MVDiff >= 104) {
return 1;
}
}
else if (mcTcor >= 75) {
if (AFD < 50) {
return 2;
}
else if (AFD >= 50) {
return 1;
}
}
}
else if (tcor >= 54) {
if (mu_mv_mag_sq < 17) {
return 1;
}
else if (mu_mv_mag_sq >= 17) {
if (diffRsCsdiff < 178) {
if (AFD < 36) {
return 4;
}
else if (AFD >= 36) {
if (Cs < 1720) {
return 2;
}
else if (Cs >= 1720) {
return 8;
}
}
}
else if (diffRsCsdiff >= 178) {
if (TSC < 19) {
return 1;
}
else if (TSC >= 19) {
return 8;
}
}
}
}
}
}
}
}
}
else if (tcor >= 88) {
if (Rs < 3245) {
if (CsDiff < 8) {
if (ssDCval < 96) {
if (Cs < 1528) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (Cs >= 1528) {
if (mcTcor < 95) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (RsCsDiff < 124) {
return 4;
}
else if (RsCsDiff >= 124) {
return 2;
}
}
}
}
else if (ssDCval >= 96) {
if (MVDiff < 10) {
if (mu_mv_mag_sq < 64) {
if (diffTSC < 1) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (diffTSC >= 1) {
if (Rs < 2638) {
return 1;
}
else if (Rs >= 2638) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 64) {
return 4;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
else if (CsDiff >= 8) {
if (diffRsCsdiff < -62) {
if (mu_mv_mag_sq < 27) {
return 2;
}
else if (mu_mv_mag_sq >= 27) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (diffRsCsdiff >= -62) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 2177) {
return 2;
}
else if (Rs >= 2177) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3245) {
if (ssDCval < 112) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (mcTcor < 93) {
if (negBalance < 48) {
if (RsCsDiff < 152) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (Rs < 4341) {
return 2;
}
else if (Rs >= 4341) {
return 1;
}
}
}
else if (RsCsDiff >= 152) {
if (TSC < 8) {
if (Rs < 4515) {
return 4;
}
else if (Rs >= 4515) {
return 2;
}
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (negBalance >= 48) {
if (Rs < 3632) {
return 8;
}
else if (Rs >= 3632) {
if (Cs < 1812) {
return 1;
}
else if (Cs >= 1812) {
if (Rs < 4041) {
return 8;
}
else if (Rs >= 4041) {
return 2;
}
}
}
}
}
else if (mcTcor >= 93) {
if (Cs < 1870) {
if (refDCval < 73) {
if (tcor < 92) {
if (Rs < 4303) {
return 2;
}
else if (Rs >= 4303) {
return 4;
}
}
else if (tcor >= 92) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
return 8;
}
}
}
else if (refDCval >= 73) {
if (ssDCval < 102) {
return 2;
}
else if (ssDCval >= 102) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
if (Rs < 4386) {
return 2;
}
else if (Rs >= 4386) {
return 1;
}
}
}
}
}
else if (Cs >= 1870) {
if (RsDiff < 10) {
if (tcor < 93) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (Cs < 2030) {
return 8;
}
else if (Cs >= 2030) {
return 1;
}
}
}
else if (tcor >= 93) {
if (diffRsCsdiff < 43) {
return 2;
}
else if (diffRsCsdiff >= 43) {
if (Rs < 3910) {
return 4;
}
else if (Rs >= 3910) {
return 8;
}
}
}
}
else if (RsDiff >= 10) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (Cs < 2063) {
if (Rs < 5300) {
return 2;
}
else if (Rs >= 5300) {
return 4;
}
}
else if (Cs >= 2063) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (ssDCval < 135) {
if (posBalance < 41) {
if (Rs < 4071) {
return 8;
}
else if (Rs >= 4071) {
return 2;
}
}
else if (posBalance >= 41) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 5072) {
return 1;
}
else if (Rs >= 5072) {
return 2;
}
}
}
}
else if (ssDCval >= 135) {
if (SC < 6739) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (SC >= 6739) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 2096) {
if (mu_mv_mag_sq < 26) {
if (ssDCval < 152) {
if (Rs < 5958) {
if (AFD < 21) {
if (ssDCval < 71) {
if (negBalance < 57) {
if (CsDiff < 11) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 1;
}
}
else if (CsDiff >= 11) {
if (RsCsDiff < 267) {
if (ssDCval < 58) {
if (Cs < 2460) {
return 2;
}
else if (Cs >= 2460) {
return 4;
}
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (RsCsDiff >= 267) {
return 4;
}
}
}
else if (negBalance >= 57) {
if (Cs < 2116) {
return 4;
}
else if (Cs >= 2116) {
return 2;
}
}
}
else if (ssDCval >= 71) {
if (negBalance < 57) {
if (MVDiff < 5) {
if (tcor < 81) {
if (CsDiff < 12) {
if (TSC < 11) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 8;
}
}
else if (TSC >= 11) {
return 4;
}
}
else if (CsDiff >= 12) {
return 8;
}
}
else if (tcor >= 81) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
if (ssDCval < 86) {
if (Rs < 3169) {
return 1;
}
else if (Rs >= 3169) {
if (SC < 5980) {
return 4;
}
else if (SC >= 5980) {
if (Rs < 4476) {
return 2;
}
else if (Rs >= 4476) {
return 8;
}
}
}
}
else if (ssDCval >= 86) {
if (mcTcor < 90) {
if (Rs < 3902) {
if (Rs < 2672) {
return 8;
}
else if (Rs >= 2672) {
return 4;
}
}
else if (Rs >= 3902) {
return 8;
}
}
else if (mcTcor >= 90) {
if (RsCsDiff < 201) {
if (Rs < 3688) {
if (Cs < 2361) {
return 8;
}
else if (Cs >= 2361) {
return 4;
}
}
else if (Rs >= 3688) {
return 8;
}
}
else if (RsCsDiff >= 201) {
if (Cs < 3625) {
if (negBalance < 50) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 2;
}
}
else if (negBalance >= 50) {
return 4;
}
}
else if (Cs >= 3625) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 5) {
if (RsDiff < 17) {
if (Rs < 3963) {
if (MVDiff < 10) {
if (tcor < 81) {
if (posBalance < 63) {
if (RsDiff < 14) {
if (ssDCval < 76) {
if (Rs < 3353) {
return 8;
}
else if (Rs >= 3353) {
return 1;
}
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (RsDiff >= 14) {
return 4;
}
}
else if (posBalance >= 63) {
if (ssDCval < 92) {
return 4;
}
else if (ssDCval >= 92) {
return 8;
}
}
}
else if (tcor >= 81) {
if (ssDCval < 101) {
if (Cs < 2146) {
return 2;
}
else if (Cs >= 2146) {
if (diffAFD < 4) {
return 4;
}
else if (diffAFD >= 4) {
return 8;
}
}
}
else if (ssDCval >= 101) {
if (Rs < 3401) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
else if (Rs >= 3401) {
return 8;
}
}
}
}
else if (MVDiff >= 10) {
if (negBalance < 31) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 2;
}
}
else if (negBalance >= 31) {
if (CsDiff < 16) {
if (TSC < 15) {
if (Rs < 3694) {
if (AFD < 15) {
if (ssDCval < 91) {
if (Rs < 2742) {
return 8;
}
else if (Rs >= 2742) {
return 2;
}
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (AFD >= 15) {
if (Rs < 1503) {
return 1;
}
else if (Rs >= 1503) {
if (ssDCval < 81) {
if (Rs < 3394) {
if (diffAFD < -2) {
if (Rs < 2541) {
return 8;
}
else if (Rs >= 2541) {
if (Rs < 2707) {
return 2;
}
else if (Rs >= 2707) {
return 1;
}
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (Rs >= 3394) {
return 1;
}
}
else if (ssDCval >= 81) {
if (ssDCval < 97) {
if (Rs < 2190) {
return 4;
}
else if (Rs >= 2190) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (refDCval < 80) {
return 4;
}
else if (refDCval >= 80) {
return 8;
}
}
}
}
else if (ssDCval >= 97) {
return 8;
}
}
}
}
}
else if (Rs >= 3694) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (CsDiff < 14) {
return 4;
}
else if (CsDiff >= 14) {
return 8;
}
}
}
}
else if (TSC >= 15) {
return 4;
}
}
else if (CsDiff >= 16) {
if (Rs < 3099) {
return 1;
}
else if (Rs >= 3099) {
return 2;
}
}
}
}
}
else if (Rs >= 3963) {
if (negBalance < 44) {
if (ssDCval < 99) {
if (ssDCval < 93) {
if (Cs < 3014) {
return 2;
}
else if (Cs >= 3014) {
return 8;
}
}
else if (ssDCval >= 93) {
return 1;
}
}
else if (ssDCval >= 99) {
return 8;
}
}
else if (negBalance >= 44) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
if (AFD < 16) {
if (Cs < 2230) {
return 4;
}
else if (Cs >= 2230) {
return 8;
}
}
else if (AFD >= 16) {
return 4;
}
}
}
}
}
else if (RsDiff >= 17) {
if (CsDiff < 13) {
if (tcor < 78) {
if (ssDCval < 81) {
return 2;
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (tcor >= 78) {
if (Rs < 3525) {
return 8;
}
else if (Rs >= 3525) {
return 2;
}
}
}
else if (CsDiff >= 13) {
if (Rs < 2663) {
return 2;
}
else if (Rs >= 2663) {
return 8;
}
}
}
}
}
else if (negBalance >= 57) {
if (mcTcor < 91) {
if (Cs < 2932) {
if (posBalance < 61) {
if (diffRsCsdiff < 1) {
if (diffAFD < 1) {
if (Cs < 2894) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
if (Rs < 3495) {
return 8;
}
else if (Rs >= 3495) {
return 4;
}
}
}
else if (Cs >= 2894) {
return 4;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (diffRsCsdiff >= 1) {
if (Cs < 2469) {
return 1;
}
else if (Cs >= 2469) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
else if (posBalance >= 61) {
if (mcTcor < 81) {
return 4;
}
else if (mcTcor >= 81) {
if (Rs < 2240) {
return 4;
}
else if (Rs >= 2240) {
return 8;
}
}
}
}
else if (Cs >= 2932) {
return 4;
}
}
else if (mcTcor >= 91) {
if (posBalance < 53) {
if (Rs < 3747) {
return 4;
}
else if (Rs >= 3747) {
return 8;
}
}
else if (posBalance >= 53) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 2915) {
return 2;
}
else if (Rs >= 2915) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 21) {
if (RsCsDiff < 393) {
if (diffAFD < 2) {
if (refDCval < 99) {
if (SC < 6040) {
if (mu_mv_mag_sq < 5) {
if (Rs < 2678) {
if (mcTcor < 81) {
return 4;
}
else if (mcTcor >= 81) {
if (Rs < 2388) {
if (CsDiff < 11) {
return 1;
}
else if (CsDiff >= 11) {
if (Cs < 2528) {
return 4;
}
else if (Cs >= 2528) {
return 1;
}
}
}
else if (Rs >= 2388) {
return 1;
}
}
}
else if (Rs >= 2678) {
if (Rs < 3030) {
if (Rs < 2861) {
return 8;
}
else if (Rs >= 2861) {
return 2;
}
}
else if (Rs >= 3030) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (mcTcor < 88) {
if (Rs < 2031) {
return 2;
}
else if (Rs >= 2031) {
return 1;
}
}
else if (mcTcor >= 88) {
if (Rs < 1870) {
return 4;
}
else if (Rs >= 1870) {
return 8;
}
}
}
}
else if (SC >= 6040) {
if (Rs < 4513) {
if (Cs < 3393) {
if (Rs < 3195) {
return 8;
}
else if (Rs >= 3195) {
return 4;
}
}
else if (Cs >= 3393) {
return 8;
}
}
else if (Rs >= 4513) {
return 8;
}
}
}
else if (refDCval >= 99) {
if (Rs < 3126) {
if (Rs < 2216) {
if (TSC < 15) {
return 1;
}
else if (TSC >= 15) {
if (Cs < 2533) {
return 2;
}
else if (Cs >= 2533) {
return 8;
}
}
}
else if (Rs >= 2216) {
if (Rs < 2740) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
else if (Rs >= 2740) {
if (Cs < 2719) {
return 1;
}
else if (Cs >= 2719) {
if (Rs < 2897) {
return 2;
}
else if (Rs >= 2897) {
return 4;
}
}
}
}
}
else if (Rs >= 3126) {
if (posBalance < 62) {
if (tcor < 66) {
if (Rs < 4496) {
return 1;
}
else if (Rs >= 4496) {
return 8;
}
}
else if (tcor >= 66) {
return 8;
}
}
else if (posBalance >= 62) {
return 1;
}
}
}
}
else if (diffAFD >= 2) {
if (Rs < 1478) {
return 2;
}
else if (Rs >= 1478) {
if (diffRsCsdiff < -32) {
if (Rs < 2201) {
return 4;
}
else if (Rs >= 2201) {
if (Rs < 2333) {
return 8;
}
else if (Rs >= 2333) {
return 2;
}
}
}
else if (diffRsCsdiff >= -32) {
if (Rs < 5037) {
return 8;
}
else if (Rs >= 5037) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 393) {
if (negBalance < 59) {
if (posBalance < 62) {
if (ssDCval < 114) {
if (RsCsDiff < 644) {
if (posBalance < 56) {
if (Cs < 2314) {
return 4;
}
else if (Cs >= 2314) {
return 2;
}
}
else if (posBalance >= 56) {
if (Cs < 5224) {
if (tcor < 67) {
return 8;
}
else if (tcor >= 67) {
if (ssDCval < 98) {
if (CsDiff < 14) {
return 2;
}
else if (CsDiff >= 14) {
return 8;
}
}
else if (ssDCval >= 98) {
return 2;
}
}
}
else if (Cs >= 5224) {
return 2;
}
}
}
else if (RsCsDiff >= 644) {
if (Cs < 3753) {
if (Rs < 3503) {
return 4;
}
else if (Rs >= 3503) {
return 2;
}
}
else if (Cs >= 3753) {
if (Cs < 4506) {
return 8;
}
else if (Cs >= 4506) {
return 4;
}
}
}
}
else if (ssDCval >= 114) {
if (Cs < 3079) {
if (AFD < 28) {
if (Rs < 5272) {
return 8;
}
else if (Rs >= 5272) {
return 1;
}
}
else if (AFD >= 28) {
return 2;
}
}
else if (Cs >= 3079) {
if (Rs < 4506) {
return 4;
}
else if (Rs >= 4506) {
if (Cs < 3445) {
return 8;
}
else if (Cs >= 3445) {
return 4;
}
}
}
}
}
else if (posBalance >= 62) {
if (TSC < 16) {
if (CsDiff < 14) {
if (refDCval < 113) {
if (ssDCval < 108) {
return 4;
}
else if (ssDCval >= 108) {
return 2;
}
}
else if (refDCval >= 113) {
if (Cs < 2462) {
return 1;
}
else if (Cs >= 2462) {
return 8;
}
}
}
else if (CsDiff >= 14) {
if (Rs < 3927) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
return 8;
}
}
}
else if (Rs >= 3927) {
return 1;
}
}
}
else if (TSC >= 16) {
if (posBalance < 65) {
if (CsDiff < 16) {
if (Rs < 3944) {
return 4;
}
else if (Rs >= 3944) {
return 8;
}
}
else if (CsDiff >= 16) {
return 4;
}
}
else if (posBalance >= 65) {
if (posBalance < 88) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 1;
}
}
else if (posBalance >= 88) {
if (Cs < 2769) {
return 4;
}
else if (Cs >= 2769) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 59) {
if (posBalance < 54) {
if (Rs < 2597) {
return 2;
}
else if (Rs >= 2597) {
return 8;
}
}
else if (posBalance >= 54) {
if (RsDiff < 26) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
if (RsCsDiff < 423) {
return 1;
}
else if (RsCsDiff >= 423) {
return 8;
}
}
}
else if (RsDiff >= 26) {
return 1;
}
}
}
}
}
}
else if (Rs >= 5958) {
if (negBalance < 51) {
if (MVDiff < 10) {
if (negBalance < 41) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (Rs < 6891) {
return 8;
}
else if (Rs >= 6891) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (tcor < 70) {
return 8;
}
else if (tcor >= 70) {
if (Cs < 2323) {
return 2;
}
else if (Cs >= 2323) {
return 4;
}
}
}
}
else if (MVDiff >= 10) {
if (Rs < 6548) {
return 2;
}
else if (Rs >= 6548) {
return 1;
}
}
}
else if (negBalance >= 51) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (AFD < 40) {
if (RsCsDiff < 442) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 4;
}
}
else if (RsCsDiff >= 442) {
return 4;
}
}
else if (AFD >= 40) {
return 8;
}
}
}
}
}
else if (ssDCval >= 152) {
if (ssDCval < 168) {
if (posBalance < 55) {
if (Rs < 6482) {
return 8;
}
else if (Rs >= 6482) {
return 4;
}
}
else if (posBalance >= 55) {
if (Rs < 3335) {
return 8;
}
else if (Rs >= 3335) {
return 4;
}
}
}
else if (ssDCval >= 168) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (negBalance < 42) {
if (ssDCval < 107) {
if (tcor < 89) {
if (refDCval < 77) {
if (mu_mv_mag_sq < 54) {
if (diffTSC < 2) {
if (diffRsCsdiff < 68) {
if (ssDCval < 64) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
else if (ssDCval >= 64) {
return 2;
}
}
else if (diffRsCsdiff >= 68) {
return 4;
}
}
else if (diffTSC >= 2) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (diffRsCsdiff < 55) {
if (diffRsCsdiff < -167) {
if (Rs < 3714) {
return 1;
}
else if (Rs >= 3714) {
return 8;
}
}
else if (diffRsCsdiff >= -167) {
if (mu_mv_mag_sq < 93) {
if (posBalance < 53) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (Rs < 4097) {
return 1;
}
else if (Rs >= 4097) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
if (diffRsCsdiff < -19) {
if (Rs < 2580) {
return 1;
}
else if (Rs >= 2580) {
return 4;
}
}
else if (diffRsCsdiff >= -19) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 93) {
return 1;
}
}
}
else if (diffRsCsdiff >= 55) {
if (SC < 8773) {
if (SC < 6364) {
if (Cs < 2173) {
return 4;
}
else if (Cs >= 2173) {
return 8;
}
}
else if (SC >= 6364) {
return 4;
}
}
else if (SC >= 8773) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
else if (refDCval >= 77) {
if (RsDiff < 13) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Cs < 2293) {
return 2;
}
else if (Cs >= 2293) {
return 8;
}
}
}
else if (RsDiff >= 13) {
if (posBalance < 49) {
if (gchDC < 3) {
if (Cs < 2404) {
return 8;
}
else if (Cs >= 2404) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 1;
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
else if (posBalance >= 49) {
if (tcor < 74) {
return 8;
}
else if (tcor >= 74) {
if (Cs < 3004) {
if (SC < 5630) {
return 4;
}
else if (SC >= 5630) {
return 8;
}
}
else if (Cs >= 3004) {
if (Rs < 6651) {
return 4;
}
else if (Rs >= 6651) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 89) {
if (ssDCval < 72) {
if (RsCsDiff < 281) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Rs < 4122) {
return 1;
}
else if (Rs >= 4122) {
return 8;
}
}
}
else if (RsCsDiff >= 281) {
return 8;
}
}
else if (ssDCval >= 72) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
if (Rs < 4414) {
if (Rs < 4164) {
if (Cs < 3360) {
return 4;
}
else if (Cs >= 3360) {
return 8;
}
}
else if (Rs >= 4164) {
if (Rs < 4209) {
return 2;
}
else if (Rs >= 4209) {
return 8;
}
}
}
else if (Rs >= 4414) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 107) {
if (TSC < 8) {
if (ssDCval < 136) {
if (SC < 8237) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
if (CsDiff < 15) {
return 2;
}
else if (CsDiff >= 15) {
return 4;
}
}
}
else if (SC >= 8237) {
if (Rs < 5767) {
return 8;
}
else if (Rs >= 5767) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 8;
}
}
}
}
else if (ssDCval >= 136) {
if (gchDC < 7) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (gchDC >= 7) {
if (Cs < 3244) {
return 1;
}
else if (Cs >= 3244) {
return 8;
}
}
}
}
else if (TSC >= 8) {
if (mu_mv_mag_sq < 37) {
if (mcTcor < 86) {
if (MVDiff < 74) {
if (ssDCval < 142) {
return 8;
}
else if (ssDCval >= 142) {
return 1;
}
}
else if (MVDiff >= 74) {
return 2;
}
}
else if (mcTcor >= 86) {
if (CsDiff < 16) {
return 2;
}
else if (CsDiff >= 16) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (SC < 8062) {
if (ssDCval < 133) {
return 8;
}
else if (ssDCval >= 133) {
if (MVDiff < 57) {
return 4;
}
else if (MVDiff >= 57) {
return 8;
}
}
}
else if (SC >= 8062) {
if (ssDCval < 120) {
if (Rs < 5414) {
return 8;
}
else if (Rs >= 5414) {
return 4;
}
}
else if (ssDCval >= 120) {
return 2;
}
}
}
}
}
}
}
}
}
else if (negBalance >= 42) {
if (Cs < 3619) {
if (refDCval < 137) {
if (mu_mv_mag_sq < 57) {
if (TSC < 11) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (mcTcor < 94) {
if (diffRsCsdiff < 183) {
if (diffTSC < -1) {
if (ssDCval < 132) {
if (MVDiff < 39) {
if (MVDiff < 6) {
if (Rs < 2721) {
return 8;
}
else if (Rs >= 2721) {
return 4;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (MVDiff >= 39) {
return 4;
}
}
else if (ssDCval >= 132) {
return 4;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (diffRsCsdiff >= 183) {
return 4;
}
}
else if (mcTcor >= 94) {
if (TSC < 10) {
if (Rs < 5796) {
return 8;
}
else if (Rs >= 5796) {
if (Rs < 7119) {
return 4;
}
else if (Rs >= 7119) {
return 8;
}
}
}
else if (TSC >= 10) {
if (Rs < 4021) {
return 2;
}
else if (Rs >= 4021) {
return 4;
}
}
}
}
}
else if (TSC >= 11) {
if (MVDiff < 11) {
if (ssDCval < 99) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
if (SC < 4948) {
return 2;
}
else if (SC >= 4948) {
if (diffMVdiffVal < 4) {
if (diffMVdiffVal < -3) {
if (Rs < 5106) {
return 4;
}
else if (Rs >= 5106) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (diffMVdiffVal >= 4) {
if (Rs < 2926) {
return 4;
}
else if (Rs >= 2926) {
return 1;
}
}
}
}
}
else if (ssDCval >= 99) {
if (Cs < 2541) {
if (mu_mv_mag_sq < 54) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
return 8;
}
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 3153) {
return 4;
}
else if (Rs >= 3153) {
return 2;
}
}
}
else if (Cs >= 2541) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (diffRsCsdiff < -38) {
if (Rs < 4003) {
return 4;
}
else if (Rs >= 4003) {
return 8;
}
}
else if (diffRsCsdiff >= -38) {
return 4;
}
}
}
}
}
else if (MVDiff >= 11) {
if (CsDiff < 12) {
if (Rs < 5506) {
if (posBalance < 68) {
if (negBalance < 58) {
if (mu_mv_mag_sq < 47) {
if (MVDiff < 12) {
if (Cs < 2239) {
return 2;
}
else if (Cs >= 2239) {
return 8;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 47) {
if (RsCsDiff < 205) {
return 8;
}
else if (RsCsDiff >= 205) {
if (diffRsCsdiff < 39) {
if (Rs < 4516) {
return 8;
}
else if (Rs >= 4516) {
if (Rs < 4730) {
return 4;
}
else if (Rs >= 4730) {
return 8;
}
}
}
else if (diffRsCsdiff >= 39) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
return 4;
}
}
}
}
}
else if (negBalance >= 58) {
if (MVDiff < 27) {
if (Cs < 2421) {
return 2;
}
else if (Cs >= 2421) {
return 4;
}
}
else if (MVDiff >= 27) {
return 8;
}
}
}
else if (posBalance >= 68) {
if (Rs < 3260) {
return 2;
}
else if (Rs >= 3260) {
return 8;
}
}
}
else if (Rs >= 5506) {
if (Rs < 6321) {
return 1;
}
else if (Rs >= 6321) {
return 8;
}
}
}
else if (CsDiff >= 12) {
if (refDCval < 68) {
if (CsDiff < 14) {
if (MVDiff < 71) {
return 2;
}
else if (MVDiff >= 71) {
return 4;
}
}
else if (CsDiff >= 14) {
return 4;
}
}
else if (refDCval >= 68) {
if (posBalance < 61) {
if (RsCsDiff < 315) {
if (tcor < 72) {
return 1;
}
else if (tcor >= 72) {
return 8;
}
}
else if (RsCsDiff >= 315) {
if (SC < 7380) {
if (tcor < 69) {
if (AFD < 31) {
return 8;
}
else if (AFD >= 31) {
if (Cs < 2577) {
return 1;
}
else if (Cs >= 2577) {
return 8;
}
}
}
else if (tcor >= 69) {
if (negBalance < 54) {
if (RsCsDiff < 441) {
return 4;
}
else if (RsCsDiff >= 441) {
return 2;
}
}
else if (negBalance >= 54) {
if (SC < 7260) {
if (tcor < 70) {
return 4;
}
else if (tcor >= 70) {
if (AFD < 40) {
return 8;
}
else if (AFD >= 40) {
return 4;
}
}
}
else if (SC >= 7260) {
return 4;
}
}
}
}
else if (SC >= 7380) {
if (mcTcor < 82) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
if (Rs < 6026) {
return 8;
}
else if (Rs >= 6026) {
return 2;
}
}
}
else if (mcTcor >= 82) {
if (posBalance < 46) {
if (Rs < 5310) {
return 4;
}
else if (Rs >= 5310) {
return 1;
}
}
else if (posBalance >= 46) {
if (Cs < 3472) {
return 2;
}
else if (Cs >= 3472) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 61) {
if (AFD < 31) {
if (diffAFD < 1) {
if (posBalance < 67) {
return 4;
}
else if (posBalance >= 67) {
return 8;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (AFD >= 31) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (ssDCval < 101) {
if (refDCval < 92) {
if (posBalance < 57) {
if (AFD < 13) {
if (ssDCval < 75) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 4;
}
}
else if (ssDCval >= 75) {
return 4;
}
}
else if (AFD >= 13) {
if (MVDiff < 8) {
if (diffRsCsdiff < 48) {
if (Rs < 4349) {
return 1;
}
else if (Rs >= 4349) {
return 8;
}
}
else if (diffRsCsdiff >= 48) {
if (Rs < 5108) {
return 8;
}
else if (Rs >= 5108) {
return 2;
}
}
}
else if (MVDiff >= 8) {
if (mcTcor < 90) {
if (Rs < 3666) {
if (Rs < 2826) {
return 2;
}
else if (Rs >= 2826) {
return 8;
}
}
else if (Rs >= 3666) {
if (Cs < 2150) {
return 8;
}
else if (Cs >= 2150) {
if (TSC < 15) {
return 4;
}
else if (TSC >= 15) {
return 2;
}
}
}
}
else if (mcTcor >= 90) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 8;
}
}
}
}
}
else if (posBalance >= 57) {
if (SC < 6517) {
if (mcTcor < 76) {
return 4;
}
else if (mcTcor >= 76) {
if (mcTcor < 90) {
if (MVDiff < 9) {
if (Rs < 2507) {
return 4;
}
else if (Rs >= 2507) {
return 8;
}
}
else if (MVDiff >= 9) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
if (tcor < 86) {
return 4;
}
else if (tcor >= 86) {
return 8;
}
}
}
}
else if (mcTcor >= 90) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 8;
}
}
}
}
else if (SC >= 6517) {
if (Rs < 4296) {
if (negBalance < 48) {
if (Rs < 3597) {
return 8;
}
else if (Rs >= 3597) {
return 1;
}
}
else if (negBalance >= 48) {
return 8;
}
}
else if (Rs >= 4296) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
if (SC < 7553) {
return 1;
}
else if (SC >= 7553) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 92) {
if (Cs < 2570) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (TSC < 13) {
if (Rs < 3283) {
return 1;
}
else if (Rs >= 3283) {
return 4;
}
}
else if (TSC >= 13) {
if (Rs < 3601) {
return 2;
}
else if (Rs >= 3601) {
return 8;
}
}
}
}
}
else if (Cs >= 2570) {
if (MVDiff < 23) {
if (refDCval < 101) {
return 4;
}
else if (refDCval >= 101) {
return 2;
}
}
else if (MVDiff >= 23) {
return 8;
}
}
}
}
else if (ssDCval >= 101) {
if (ssDCval < 105) {
if (posBalance < 57) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (negBalance < 53) {
return 8;
}
else if (negBalance >= 53) {
return 2;
}
}
}
else if (posBalance >= 57) {
if (Cs < 2598) {
return 1;
}
else if (Cs >= 2598) {
return 2;
}
}
}
else if (ssDCval >= 105) {
if (Rs < 5413) {
if (mcTcor < 90) {
if (AFD < 18) {
if (ssDCval < 119) {
if (Rs < 3614) {
if (Cs < 3031) {
return 4;
}
else if (Cs >= 3031) {
return 8;
}
}
else if (Rs >= 3614) {
return 8;
}
}
else if (ssDCval >= 119) {
if (Rs < 3714) {
return 8;
}
else if (Rs >= 3714) {
return 4;
}
}
}
else if (AFD >= 18) {
return 1;
}
}
else if (mcTcor >= 90) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (Rs < 3083) {
return 4;
}
else if (Rs >= 3083) {
return 8;
}
}
}
}
else if (Rs >= 5413) {
if (Rs < 6968) {
return 4;
}
else if (Rs >= 6968) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 137) {
if (Rs < 3641) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Rs < 2534) {
return 4;
}
else if (Rs >= 2534) {
if (Rs < 2664) {
return 2;
}
else if (Rs >= 2664) {
return 8;
}
}
}
}
else if (Rs >= 3641) {
if (RsCsDiff < 389) {
if (mcTcor < 92) {
if (Rs < 5841) {
return 1;
}
else if (Rs >= 5841) {
return 8;
}
}
else if (mcTcor >= 92) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
return 8;
}
}
}
else if (RsCsDiff >= 389) {
if (Rs < 4624) {
return 2;
}
else if (Rs >= 4624) {
return 4;
}
}
}
}
}
else if (Cs >= 3619) {
if (posBalance < 50) {
if (MVDiff < 18) {
if (TSC < 15) {
if (Rs < 5467) {
return 1;
}
else if (Rs >= 5467) {
return 4;
}
}
else if (TSC >= 15) {
return 1;
}
}
else if (MVDiff >= 18) {
if (AFD < 21) {
if (ssDCval < 86) {
if (Rs < 4722) {
return 8;
}
else if (Rs >= 4722) {
return 2;
}
}
else if (ssDCval >= 86) {
if (RsDiff < 23) {
return 1;
}
else if (RsDiff >= 23) {
return 2;
}
}
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (posBalance >= 50) {
if (SC < 13972) {
if (negBalance < 61) {
if (negBalance < 53) {
if (Rs < 5798) {
if (RsCsDiff < 601) {
if (tcor < 68) {
if (ssDCval < 105) {
return 8;
}
else if (ssDCval >= 105) {
if (gchDC < 1) {
if (Rs < 4183) {
return 8;
}
else if (Rs >= 4183) {
return 1;
}
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (tcor >= 68) {
if (diffTSC < -2) {
if (Cs < 3787) {
return 4;
}
else if (Cs >= 3787) {
return 8;
}
}
else if (diffTSC >= -2) {
if (SC < 7871) {
if (MVDiff < 22) {
if (RsDiff < 10) {
if (Rs < 3838) {
return 8;
}
else if (Rs >= 3838) {
return 2;
}
}
else if (RsDiff >= 10) {
return 8;
}
}
else if (MVDiff >= 22) {
return 1;
}
}
else if (SC >= 7871) {
if (Cs < 3718) {
if (ssDCval < 95) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
else if (Cs >= 3718) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 601) {
if (Rs < 5332) {
return 4;
}
else if (Rs >= 5332) {
return 8;
}
}
}
else if (Rs >= 5798) {
if (TSC < 13) {
if (SC < 10432) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (SC >= 10432) {
return 8;
}
}
else if (TSC >= 13) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
if (Rs < 6435) {
return 2;
}
else if (Rs >= 6435) {
return 4;
}
}
}
}
}
else if (negBalance >= 53) {
if (Rs < 3176) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
else if (Rs >= 3176) {
if (posBalance < 57) {
if (diffAFD < 3) {
if (diffRsCsdiff < -76) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Cs < 3706) {
return 8;
}
else if (Cs >= 3706) {
return 4;
}
}
}
else if (diffRsCsdiff >= -76) {
return 8;
}
}
else if (diffAFD >= 3) {
if (Rs < 4193) {
return 1;
}
else if (Rs >= 4193) {
return 4;
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
}
else if (negBalance >= 61) {
if (Rs < 6061) {
if (Rs < 5212) {
return 2;
}
else if (Rs >= 5212) {
return 8;
}
}
else if (Rs >= 6061) {
return 4;
}
}
}
else if (SC >= 13972) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1652) {
if (Rs < 3081) {
if (Rs < 1342) {
if (Rs < 1107) {
if (AFD < 1) {
if (Rs < 599) {
return 4;
}
else if (Rs >= 599) {
return 2;
}
}
else if (AFD >= 1) {
if (refDCval < 83) {
if (ssDCval < 77) {
if (ssDCval < 63) {
if (SC < 2385) {
if (mu_mv_mag_sq < 128) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (SC >= 2385) {
if (mu_mv_mag_sq < 117) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (Rs < 1029) {
return 8;
}
else if (Rs >= 1029) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
}
else if (ssDCval >= 63) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
else if (ssDCval >= 77) {
return 1;
}
}
else if (refDCval >= 83) {
if (ssDCval < 129) {
if (SC < 2254) {
if (Rs < 485) {
return 4;
}
else if (Rs >= 485) {
return 8;
}
}
else if (SC >= 2254) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (mu_mv_mag_sq < 122) {
return 4;
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 986) {
return 8;
}
else if (Rs >= 986) {
return 4;
}
}
}
}
}
else if (ssDCval >= 129) {
return 8;
}
}
}
}
else if (Rs >= 1107) {
if (mu_mv_mag_sq < 122) {
if (Cs < 1472) {
if (TSC < 2) {
if (posBalance < 9) {
return 8;
}
else if (posBalance >= 9) {
if (Cs < 1444) {
return 1;
}
else if (Cs >= 1444) {
return 4;
}
}
}
else if (TSC >= 2) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Rs < 1226) {
return 4;
}
else if (Rs >= 1226) {
if (Rs < 1326) {
return 8;
}
else if (Rs >= 1326) {
return 1;
}
}
}
}
}
else if (Cs >= 1472) {
if (SC < 2900) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
if (SC < 2857) {
return 4;
}
else if (SC >= 2857) {
if (Rs < 1265) {
return 8;
}
else if (Rs >= 1265) {
return 4;
}
}
}
}
else if (SC >= 2900) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (AFD < 1) {
if (Rs < 1136) {
return 1;
}
else if (Rs >= 1136) {
return 8;
}
}
else if (AFD >= 1) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (diffAFD < -1) {
if (Cs < 1328) {
return 8;
}
else if (Cs >= 1328) {
return 4;
}
}
else if (diffAFD >= -1) {
if (Rs < 1315) {
return 8;
}
else if (Rs >= 1315) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1342) {
if (refDCval < 61) {
if (TSCindex < 3) {
if (ssDCval < 52) {
if (negBalance < 18) {
if (mu_mv_mag_sq < 117) {
return 1;
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
else if (negBalance >= 18) {
return 4;
}
}
else if (ssDCval >= 52) {
if (Cs < 1464) {
if (Rs < 1464) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 4;
}
}
else if (Rs >= 1464) {
return 8;
}
}
else if (Cs >= 1464) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (mcTcor < 98) {
if (Rs < 2226) {
return 8;
}
else if (Rs >= 2226) {
if (SC < 3864) {
return 4;
}
else if (SC >= 3864) {
if (Cs < 1466) {
return 4;
}
else if (Cs >= 1466) {
return 8;
}
}
}
}
else if (mcTcor >= 98) {
return 4;
}
}
}
else if (refDCval >= 61) {
if (ssDCval < 86) {
if (mcTcor < 97) {
if (Rs < 3009) {
if (negBalance < 62) {
return 8;
}
else if (negBalance >= 62) {
return 4;
}
}
else if (Rs >= 3009) {
return 2;
}
}
else if (mcTcor >= 97) {
if (negBalance < 18) {
if (SC < 2845) {
return 1;
}
else if (SC >= 2845) {
if (Cs < 1350) {
if (Cs < 1317) {
if (Rs < 2467) {
return 2;
}
else if (Rs >= 2467) {
return 8;
}
}
else if (Cs >= 1317) {
if (Rs < 2339) {
if (Cs < 1337) {
if (Rs < 2287) {
return 8;
}
else if (Rs >= 2287) {
if (SC < 3619) {
return 2;
}
else if (SC >= 3619) {
if (Cs < 1335) {
if (Rs < 2297) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
return 2;
}
}
else if (Rs >= 2297) {
return 8;
}
}
else if (Cs >= 1335) {
if (Rs < 2306) {
return 2;
}
else if (Rs >= 2306) {
return 8;
}
}
}
}
}
else if (Cs >= 1337) {
return 8;
}
}
else if (Rs >= 2339) {
if (posBalance < 35) {
return 2;
}
else if (posBalance >= 35) {
return 8;
}
}
}
}
else if (Cs >= 1350) {
return 8;
}
}
}
else if (negBalance >= 18) {
if (ssDCval < 75) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (RsDiff < 4) {
if (ssDCval < 70) {
if (Cs < 1598) {
if (Rs < 2050) {
return 4;
}
else if (Rs >= 2050) {
if (Rs < 2802) {
return 4;
}
else if (Rs >= 2802) {
return 4;
}
}
}
else if (Cs >= 1598) {
return 8;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (RsDiff >= 4) {
return 2;
}
}
}
else if (ssDCval >= 75) {
if (Rs < 2455) {
if (RsDiff < 4) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 2242) {
return 8;
}
else if (Rs >= 2242) {
return 4;
}
}
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (Rs >= 2455) {
return 4;
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (CsDiff < 2) {
if (Rs < 3053) {
if (diffAFD < -19) {
if (Cs < 1457) {
return 1;
}
else if (Cs >= 1457) {
return 8;
}
}
else if (diffAFD >= -19) {
if (Rs < 1362) {
if (Rs < 1358) {
return 8;
}
else if (Rs >= 1358) {
return 4;
}
}
else if (Rs >= 1362) {
return 8;
}
}
}
else if (Rs >= 3053) {
if (Rs < 3055) {
return 2;
}
else if (Rs >= 3055) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (ssDCval < 117) {
if (AFD < 11) {
if (MVDiff < 2) {
if (diffAFD < -2) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 2;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (MVDiff >= 2) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Rs < 1817) {
return 8;
}
else if (Rs >= 1817) {
return 2;
}
}
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (ssDCval >= 117) {
if (mu_mv_mag_sq < 120) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (mu_mv_mag_sq >= 120) {
if (SC < 3495) {
if (Cs < 1355) {
return 8;
}
else if (Cs >= 1355) {
return 1;
}
}
else if (SC >= 3495) {
if (ssDCval < 123) {
if (Rs < 2368) {
return 1;
}
else if (Rs >= 2368) {
return 8;
}
}
else if (ssDCval >= 123) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (MVDiff < 3) {
if (Rs < 2008) {
return 2;
}
else if (Rs >= 2008) {
return 1;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3081) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 126) {
if (Rs < 3616) {
if (Cs < 1562) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
else if (Cs >= 1562) {
if (Rs < 3180) {
return 8;
}
else if (Rs >= 3180) {
return 4;
}
}
}
else if (Rs >= 3616) {
if (tcor < 99) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
if (Rs < 4315) {
return 2;
}
else if (Rs >= 4315) {
return 8;
}
}
}
else if (tcor >= 99) {
if (posBalance < 33) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 33) {
if (negBalance < 17) {
if (Rs < 5497) {
if (Rs < 3638) {
return 8;
}
else if (Rs >= 3638) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 4085) {
return 4;
}
else if (Rs >= 4085) {
return 8;
}
}
}
}
else if (Rs >= 5497) {
return 8;
}
}
else if (negBalance >= 17) {
if (AFD < 3) {
if (mu_mv_mag_sq < 115) {
return 1;
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
else if (AFD >= 3) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (RsCsDiff < 10) {
if (Cs < 1577) {
if (Rs < 3679) {
if (diffRsCsdiff < -1) {
return 8;
}
else if (diffRsCsdiff >= -1) {
return 4;
}
}
else if (Rs >= 3679) {
return 8;
}
}
else if (Cs >= 1577) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 4;
}
}
}
else if (RsCsDiff >= 10) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 113) {
if (posBalance < 42) {
if (negBalance < 28) {
if (Rs < 3224) {
return 8;
}
else if (Rs >= 3224) {
if (Rs < 4493) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (Rs >= 4493) {
return 2;
}
}
}
else if (negBalance >= 28) {
if (Cs < 1505) {
return 4;
}
else if (Cs >= 1505) {
if (Rs < 4738) {
return 8;
}
else if (Rs >= 4738) {
return 2;
}
}
}
}
else if (posBalance >= 42) {
if (refDCval < 91) {
if (Rs < 3127) {
if (Rs < 3103) {
return 2;
}
else if (Rs >= 3103) {
return 4;
}
}
else if (Rs >= 3127) {
return 2;
}
}
else if (refDCval >= 91) {
if (Cs < 1532) {
if (Rs < 3270) {
return 2;
}
else if (Rs >= 3270) {
if (Rs < 4455) {
return 4;
}
else if (Rs >= 4455) {
return 2;
}
}
}
else if (Cs >= 1532) {
if (Rs < 3745) {
return 2;
}
else if (Rs >= 3745) {
return 8;
}
}
}
}
}
else if (ssDCval >= 113) {
if (MVDiff < 12) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (MVDiff >= 12) {
if (Rs < 3585) {
if (Rs < 3152) {
return 8;
}
else if (Rs >= 3152) {
return 2;
}
}
else if (Rs >= 3585) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1652) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (Rs < 3796) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 4;
}
}
}
else if (Rs >= 3796) {
if (refDCval < 65) {
if (Rs < 8850) {
return 8;
}
else if (Rs >= 8850) {
return 2;
}
}
else if (refDCval >= 65) {
if (Rs < 4218) {
if (ssDCval < 90) {
if (MVDiff < 39) {
return 4;
}
else if (MVDiff >= 39) {
return 8;
}
}
else if (ssDCval >= 90) {
return 8;
}
}
else if (Rs >= 4218) {
return 8;
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (SC < 4364) {
if (Cs < 1790) {
if (MVDiff < 2) {
if (refDCval < 63) {
if (Rs < 1670) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 1248) {
return 1;
}
else if (Rs >= 1248) {
return 4;
}
}
}
else if (Rs >= 1670) {
return 8;
}
}
else if (refDCval >= 63) {
if (Cs < 1737) {
return 8;
}
else if (Cs >= 1737) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 1;
}
}
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
return 4;
}
}
}
else if (Cs >= 1790) {
if (posBalance < 37) {
if (Cs < 1812) {
return 8;
}
else if (Cs >= 1812) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 1826) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 8;
}
}
else if (Rs >= 1826) {
return 1;
}
}
}
}
else if (posBalance >= 37) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (AFD < 3) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
else if (SC >= 4364) {
if (negBalance < 19) {
if (Cs < 3335) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (ssDCval < 60) {
if (Cs < 1667) {
if (Rs < 3139) {
return 1;
}
else if (Rs >= 3139) {
return 4;
}
}
else if (Cs >= 1667) {
return 1;
}
}
else if (ssDCval >= 60) {
if (Cs < 2085) {
return 8;
}
else if (Cs >= 2085) {
return 4;
}
}
}
}
else if (Cs >= 3335) {
return 8;
}
}
else if (negBalance >= 19) {
if (SC < 4779) {
if (Rs < 2939) {
return 4;
}
else if (Rs >= 2939) {
return 8;
}
}
else if (SC >= 4779) {
return 8;
}
}
}
}
else if (ssDCval >= 67) {
if (mu_mv_mag_sq < 123) {
if (TSCindex < 1) {
if (ssDCval < 92) {
if (Rs < 2373) {
return 1;
}
else if (Rs >= 2373) {
if (Cs < 2143) {
return 8;
}
else if (Cs >= 2143) {
if (Rs < 2487) {
return 2;
}
else if (Rs >= 2487) {
return 1;
}
}
}
}
else if (ssDCval >= 92) {
if (negBalance < 7) {
return 2;
}
else if (negBalance >= 7) {
if (Rs < 2318) {
return 2;
}
else if (Rs >= 2318) {
if (Rs < 2353) {
return 4;
}
else if (Rs >= 2353) {
return 2;
}
}
}
}
}
else if (TSCindex >= 1) {
if (Cs < 2289) {
if (Rs < 3067) {
if (Cs < 2249) {
return 8;
}
else if (Cs >= 2249) {
return 1;
}
}
else if (Rs >= 3067) {
if (Rs < 4318) {
return 2;
}
else if (Rs >= 4318) {
return 4;
}
}
}
else if (Cs >= 2289) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1109) {
if (Rs < 615) {
return 8;
}
else if (Rs >= 615) {
return 1;
}
}
else if (Rs >= 1109) {
if (SC < 6962) {
if (SC < 4488) {
if (SC < 4407) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 4;
}
}
else if (SC >= 4407) {
if (posBalance < 22) {
return 8;
}
else if (posBalance >= 22) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (SC >= 4488) {
if (diffRsCsdiff < -210) {
if (Rs < 3481) {
return 8;
}
else if (Rs >= 3481) {
return 4;
}
}
else if (diffRsCsdiff >= -210) {
return 8;
}
}
}
else if (SC >= 6962) {
if (SC < 7762) {
if (SC < 7182) {
if (SC < 6995) {
return 4;
}
else if (SC >= 6995) {
return 8;
}
}
else if (SC >= 7182) {
return 4;
}
}
else if (SC >= 7762) {
if (Rs < 6705) {
if (ssDCval < 72) {
if (Rs < 4925) {
return 1;
}
else if (Rs >= 4925) {
return 8;
}
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (Rs >= 6705) {
return 4;
}
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6642) {
if (SC < 8342) {
if (ssDCval < 82) {
if (negBalance < 34) {
if (mu_mv_mag_sq < 120) {
if (tcor < 99) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (Cs < 1827) {
if (RsDiff < 6) {
if (Rs < 2257) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (Rs >= 2257) {
return 4;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (Cs >= 1827) {
if (Rs < 1980) {
return 8;
}
else if (Rs >= 1980) {
if (posBalance < 54) {
if (ssDCval < 71) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Rs < 4003) {
return 4;
}
else if (Rs >= 4003) {
if (SC < 7396) {
if (diffMVdiffVal < -1) {
if (posBalance < 27) {
return 4;
}
else if (posBalance >= 27) {
if (RsDiff < 6) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
if (Cs < 2153) {
return 8;
}
else if (Cs >= 2153) {
return 4;
}
}
}
else if (RsDiff >= 6) {
if (posBalance < 29) {
return 4;
}
else if (posBalance >= 29) {
if (Rs < 4384) {
return 4;
}
else if (Rs >= 4384) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (tcor < 95) {
if (mu_mv_mag_sq < 113) {
return 8;
}
else if (mu_mv_mag_sq >= 113) {
if (posBalance < 26) {
return 4;
}
else if (posBalance >= 26) {
if (Cs < 2068) {
if (Cs < 1971) {
return 8;
}
else if (Cs >= 1971) {
return 8;
}
}
else if (Cs >= 2068) {
if (Cs < 2141) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
return 4;
}
}
else if (Cs >= 2141) {
if (posBalance < 31) {
if (diffRsCsdiff < -5) {
if (Rs < 4775) {
return 8;
}
else if (Rs >= 4775) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
return 8;
}
}
else if (posBalance >= 31) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 95) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
else if (SC >= 7396) {
if (posBalance < 35) {
if (RsCsDiff < 22) {
return 4;
}
else if (RsCsDiff >= 22) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 4988) {
return 4;
}
else if (Rs >= 4988) {
return 8;
}
}
}
}
else if (posBalance >= 35) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
if (Cs < 2680) {
return 4;
}
else if (Cs >= 2680) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (Rs < 4587) {
return 8;
}
else if (Rs >= 4587) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
return 4;
}
}
}
}
}
}
else if (tcor >= 99) {
return 8;
}
}
else if (mu_mv_mag_sq >= 120) {
if (Cs < 2462) {
if (posBalance < 21) {
return 8;
}
else if (posBalance >= 21) {
if (SC < 4258) {
if (Rs < 2463) {
return 8;
}
else if (Rs >= 2463) {
if (Rs < 2481) {
return 1;
}
else if (Rs >= 2481) {
return 8;
}
}
}
else if (SC >= 4258) {
if (SC < 5109) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (RsCsDiff < 16) {
if (Rs < 2580) {
if (Rs < 2238) {
return 4;
}
else if (Rs >= 2238) {
return 8;
}
}
else if (Rs >= 2580) {
if (Rs < 2760) {
return 4;
}
else if (Rs >= 2760) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (ssDCval < 65) {
if (gchDC < 1) {
if (Rs < 2959) {
if (Rs < 2935) {
return 4;
}
else if (Rs >= 2935) {
return 4;
}
}
else if (Rs >= 2959) {
if (SC < 4814) {
if (SC < 4781) {
return 8;
}
else if (SC >= 4781) {
return 8;
}
}
else if (SC >= 4814) {
if (Rs < 3055) {
return 4;
}
else if (Rs >= 3055) {
if (Rs < 3142) {
return 4;
}
else if (Rs >= 3142) {
return 4;
}
}
}
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (ssDCval >= 65) {
if (Cs < 1881) {
if (Rs < 2956) {
if (Rs < 2881) {
return 8;
}
else if (Rs >= 2881) {
return 4;
}
}
else if (Rs >= 2956) {
return 4;
}
}
else if (Cs >= 1881) {
if (CsDiff < 3) {
if (Cs < 1890) {
return 8;
}
else if (Cs >= 1890) {
return 8;
}
}
else if (CsDiff >= 3) {
if (Rs < 2967) {
return 8;
}
else if (Rs >= 2967) {
return 4;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 16) {
if (Cs < 1739) {
return 8;
}
else if (Cs >= 1739) {
return 4;
}
}
}
}
else if (SC >= 5109) {
if (ssDCval < 60) {
if (Rs < 4436) {
if (Rs < 4065) {
return 4;
}
else if (Rs >= 4065) {
if (CsDiff < 4) {
if (Rs < 4184) {
return 4;
}
else if (Rs >= 4184) {
if (mu_mv_mag_sq < 122) {
return 8;
}
else if (mu_mv_mag_sq >= 122) {
return 4;
}
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (Rs >= 4436) {
if (Rs < 4456) {
return 4;
}
else if (Rs >= 4456) {
return 4;
}
}
}
else if (ssDCval >= 60) {
if (Rs < 5043) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (Rs >= 5043) {
return 4;
}
}
}
}
}
}
else if (Cs >= 2462) {
if (negBalance < 20) {
if (mcTcor < 98) {
if (Rs < 5144) {
return 4;
}
else if (Rs >= 5144) {
return 4;
}
}
else if (mcTcor >= 98) {
if (SC < 6021) {
return 1;
}
else if (SC >= 6021) {
return 8;
}
}
}
else if (negBalance >= 20) {
return 8;
}
}
}
}
else if (negBalance >= 34) {
if (MVDiff < 4) {
if (Cs < 3523) {
if (SC < 5497) {
if (Cs < 2825) {
if (SC < 4346) {
if (Rs < 1029) {
return 1;
}
else if (Rs >= 1029) {
return 8;
}
}
else if (SC >= 4346) {
if (posBalance < 53) {
if (ssDCval < 71) {
if (CsDiff < 3) {
if (Rs < 2931) {
return 8;
}
else if (Rs >= 2931) {
if (Rs < 3085) {
return 4;
}
else if (Rs >= 3085) {
return 8;
}
}
}
else if (CsDiff >= 3) {
return 4;
}
}
else if (ssDCval >= 71) {
if (Cs < 1828) {
return 4;
}
else if (Cs >= 1828) {
return 8;
}
}
}
else if (posBalance >= 53) {
if (gchDC < 2) {
if (Rs < 2141) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Rs >= 2141) {
return 8;
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (Cs >= 2825) {
return 4;
}
}
else if (SC >= 5497) {
if (Cs < 1976) {
return 2;
}
else if (Cs >= 1976) {
return 8;
}
}
}
else if (Cs >= 3523) {
if (Rs < 3395) {
return 4;
}
else if (Rs >= 3395) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (SC < 5094) {
if (Cs < 1972) {
return 8;
}
else if (Cs >= 1972) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (SC >= 5094) {
if (Cs < 3091) {
return 8;
}
else if (Cs >= 3091) {
return 4;
}
}
}
}
}
else if (ssDCval >= 82) {
if (Cs < 2615) {
if (Rs < 3756) {
if (tcor < 98) {
if (negBalance < 48) {
if (diffRsCsdiff < -38) {
if (mu_mv_mag_sq < 115) {
if (Rs < 2130) {
return 2;
}
else if (Rs >= 2130) {
if (Rs < 2952) {
return 4;
}
else if (Rs >= 2952) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
else if (diffRsCsdiff >= -38) {
return 8;
}
}
else if (negBalance >= 48) {
if (mu_mv_mag_sq < 118) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 118) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
else if (tcor >= 98) {
if (Cs < 1800) {
if (ssDCval < 86) {
if (SC < 4192) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (SC >= 4192) {
return 4;
}
}
else if (ssDCval >= 86) {
if (Rs < 3150) {
return 8;
}
else if (Rs >= 3150) {
return 2;
}
}
}
else if (Cs >= 1800) {
if (posBalance < 43) {
if (refDCval < 90) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (Rs < 2308) {
return 8;
}
else if (Rs >= 2308) {
return 2;
}
}
}
else if (refDCval >= 90) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (posBalance >= 43) {
if (diffAFD < 2) {
if (Cs < 1887) {
if (Rs < 2834) {
return 8;
}
else if (Rs >= 2834) {
return 4;
}
}
else if (Cs >= 1887) {
return 8;
}
}
else if (diffAFD >= 2) {
if (Rs < 2238) {
return 4;
}
else if (Rs >= 2238) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3756) {
if (RsCsDiff < 59) {
if (negBalance < 29) {
if (Cs < 2316) {
if (ssDCval < 99) {
if (Rs < 4889) {
if (tcor < 97) {
if (Rs < 4346) {
return 4;
}
else if (Rs >= 4346) {
return 8;
}
}
else if (tcor >= 97) {
if (Rs < 4667) {
return 2;
}
else if (Rs >= 4667) {
return 8;
}
}
}
else if (Rs >= 4889) {
return 4;
}
}
else if (ssDCval >= 99) {
return 4;
}
}
else if (Cs >= 2316) {
if (Rs < 4704) {
return 8;
}
else if (Rs >= 4704) {
return 4;
}
}
}
else if (negBalance >= 29) {
if (ssDCval < 91) {
if (Rs < 4651) {
if (Rs < 4132) {
return 8;
}
else if (Rs >= 4132) {
return 2;
}
}
else if (Rs >= 4651) {
return 8;
}
}
else if (ssDCval >= 91) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (RsCsDiff >= 59) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
else if (Cs >= 2615) {
if (Cs < 3637) {
if (RsCsDiff < 69) {
if (SC < 7102) {
if (Rs < 2813) {
return 4;
}
else if (Rs >= 2813) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (refDCval < 97) {
if (SC < 6382) {
return 8;
}
else if (SC >= 6382) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (ssDCval < 95) {
if (Cs < 3544) {
return 4;
}
else if (Cs >= 3544) {
return 8;
}
}
else if (ssDCval >= 95) {
if (Rs < 3709) {
return 4;
}
else if (Rs >= 3709) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 97) {
return 8;
}
}
}
}
else if (SC >= 7102) {
if (mu_mv_mag_sq < 128) {
if (Rs < 3962) {
if (Rs < 3600) {
return 4;
}
else if (Rs >= 3600) {
return 8;
}
}
else if (Rs >= 3962) {
return 4;
}
}
else if (mu_mv_mag_sq >= 128) {
return 4;
}
}
}
else if (RsCsDiff >= 69) {
if (Cs < 3348) {
if (Rs < 3276) {
if (Rs < 2679) {
return 4;
}
else if (Rs >= 2679) {
if (refDCval < 97) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (refDCval >= 97) {
if (SC < 6203) {
return 4;
}
else if (SC >= 6203) {
if (Cs < 3123) {
return 4;
}
else if (Cs >= 3123) {
return 8;
}
}
}
}
}
else if (Rs >= 3276) {
return 4;
}
}
else if (Cs >= 3348) {
if (negBalance < 45) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 2528) {
return 2;
}
else if (Rs >= 2528) {
return 8;
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
}
else if (Cs >= 3637) {
if (Cs < 4086) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (SC < 6697) {
return 4;
}
else if (SC >= 6697) {
return 8;
}
}
}
else if (Cs >= 4086) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 4;
}
}
}
}
}
}
else if (SC >= 8342) {
if (MVDiff < 3) {
if (Cs < 3666) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
if (Rs < 5509) {
return 4;
}
else if (Rs >= 5509) {
return 8;
}
}
}
else if (Cs >= 3666) {
if (negBalance < 29) {
if (Rs < 4533) {
if (Rs < 4382) {
return 8;
}
else if (Rs >= 4382) {
return 1;
}
}
else if (Rs >= 4533) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Cs < 4239) {
return 8;
}
else if (Cs >= 4239) {
return 2;
}
}
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 5858) {
return 8;
}
else if (Rs >= 5858) {
if (ssDCval < 82) {
return 2;
}
else if (ssDCval >= 82) {
return 8;
}
}
}
}
}
else if (Rs >= 6642) {
if (ssDCval < 95) {
if (Rs < 9152) {
return 4;
}
else if (Rs >= 9152) {
return 2;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
else if (ssDCval >= 102) {
if (posBalance < 53) {
if (refDCval < 117) {
if (Cs < 2313) {
if (Rs < 3929) {
if (ssDCval < 108) {
if (diffTSC < -10) {
return 4;
}
else if (diffTSC >= -10) {
if (posBalance < 1) {
if (Rs < 2568) {
return 2;
}
else if (Rs >= 2568) {
return 8;
}
}
else if (posBalance >= 1) {
if (Rs < 3286) {
return 8;
}
else if (Rs >= 3286) {
if (Rs < 3449) {
return 4;
}
else if (Rs >= 3449) {
return 8;
}
}
}
}
}
else if (ssDCval >= 108) {
return 8;
}
}
else if (Rs >= 3929) {
if (Rs < 4501) {
return 4;
}
else if (Rs >= 4501) {
return 2;
}
}
}
else if (Cs >= 2313) {
return 8;
}
}
else if (refDCval >= 117) {
if (Cs < 2841) {
if (Rs < 4885) {
if (diffTSC < 3) {
if (tcor < 100) {
if (posBalance < 18) {
if (posBalance < 17) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Cs < 1818) {
return 1;
}
else if (Cs >= 1818) {
return 8;
}
}
}
else if (posBalance >= 17) {
if (Rs < 2564) {
return 8;
}
else if (Rs >= 2564) {
return 1;
}
}
}
else if (posBalance >= 18) {
return 8;
}
}
else if (tcor >= 100) {
if (negBalance < 24) {
if (ssDCval < 127) {
if (ssDCval < 125) {
return 8;
}
else if (ssDCval >= 125) {
if (Rs < 2432) {
return 1;
}
else if (Rs >= 2432) {
return 8;
}
}
}
else if (ssDCval >= 127) {
return 8;
}
}
else if (negBalance >= 24) {
return 1;
}
}
}
else if (diffTSC >= 3) {
if (Rs < 2038) {
return 8;
}
else if (Rs >= 2038) {
return 1;
}
}
}
else if (Rs >= 4885) {
if (mu_mv_mag_sq < 125) {
return 2;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 4893) {
return 2;
}
else if (Rs >= 4893) {
return 8;
}
}
}
}
else if (Cs >= 2841) {
if (ssDCval < 200) {
if (Rs < 2824) {
if (Cs < 3938) {
if (Cs < 2936) {
if (mu_mv_mag_sq < 121) {
if (RsDiff < 2) {
if (posBalance < 15) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 2635) {
return 4;
}
else if (Rs >= 2635) {
return 8;
}
}
}
else if (posBalance >= 15) {
return 4;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 121) {
return 8;
}
}
else if (Cs >= 2936) {
if (ssDCval < 118) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 8;
}
}
else if (ssDCval >= 118) {
if (Rs < 2816) {
return 8;
}
else if (Rs >= 2816) {
return 2;
}
}
}
}
else if (Cs >= 3938) {
return 1;
}
}
else if (Rs >= 2824) {
if (diffTSC < -7) {
if (MVDiff < 12) {
return 1;
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (diffTSC >= -7) {
if (negBalance < 41) {
if (TSC < 4) {
if (diffMVdiffVal < -8) {
if (Rs < 4087) {
if (Rs < 3769) {
return 8;
}
else if (Rs >= 3769) {
return 1;
}
}
else if (Rs >= 4087) {
return 8;
}
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
else if (TSC >= 4) {
if (Rs < 6518) {
return 8;
}
else if (Rs >= 6518) {
if (Cs < 4491) {
return 4;
}
else if (Cs >= 4491) {
return 8;
}
}
}
}
else if (negBalance >= 41) {
return 4;
}
}
}
}
else if (ssDCval >= 200) {
return 4;
}
}
}
}
else if (posBalance >= 53) {
if (ssDCval < 137) {
if (Rs < 1649) {
if (Cs < 1823) {
if (negBalance < 34) {
if (Rs < 1170) {
return 8;
}
else if (Rs >= 1170) {
return 4;
}
}
else if (negBalance >= 34) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
return 2;
}
}
}
else if (Cs >= 1823) {
if (mcTcor < 92) {
if (Rs < 818) {
return 8;
}
else if (Rs >= 818) {
return 4;
}
}
else if (mcTcor >= 92) {
if (Cs < 1870) {
if (Rs < 1318) {
return 2;
}
else if (Rs >= 1318) {
return 8;
}
}
else if (Cs >= 1870) {
return 2;
}
}
}
}
else if (Rs >= 1649) {
if (refDCval < 105) {
if (Cs < 2962) {
if (Rs < 4905) {
if (Rs < 3796) {
return 8;
}
else if (Rs >= 3796) {
return 1;
}
}
else if (Rs >= 4905) {
return 2;
}
}
else if (Cs >= 2962) {
return 8;
}
}
else if (refDCval >= 105) {
if (RsDiff < 2) {
if (Rs < 1916) {
if (Cs < 1851) {
return 8;
}
else if (Cs >= 1851) {
return 2;
}
}
else if (Rs >= 1916) {
return 8;
}
}
else if (RsDiff >= 2) {
if (gchDC < 1) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
if (posBalance < 63) {
if (mu_mv_mag_sq < 114) {
if (diffAFD < 2) {
if (negBalance < 58) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (negBalance >= 58) {
if (Rs < 5133) {
return 2;
}
else if (Rs >= 5133) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 4352) {
return 8;
}
else if (Rs >= 4352) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Rs < 4155) {
if (Rs < 2472) {
return 8;
}
else if (Rs >= 2472) {
return 2;
}
}
else if (Rs >= 4155) {
return 8;
}
}
}
}
else if (posBalance >= 63) {
if (Cs < 1824) {
if (Rs < 2366) {
return 8;
}
else if (Rs >= 2366) {
return 1;
}
}
else if (Cs >= 1824) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
if (SC < 7484) {
if (negBalance < 53) {
if (diffAFD < -2) {
if (RsCsDiff < 56) {
return 8;
}
else if (RsCsDiff >= 56) {
if (Rs < 3809) {
return 1;
}
else if (Rs >= 3809) {
return 2;
}
}
}
else if (diffAFD >= -2) {
if (ssDCval < 105) {
return 2;
}
else if (ssDCval >= 105) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (Rs < 3034) {
return 1;
}
else if (Rs >= 3034) {
return 8;
}
}
}
}
}
else if (negBalance >= 53) {
if (Cs < 2657) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
return 1;
}
}
else if (Cs >= 2657) {
return 2;
}
}
}
else if (SC >= 7484) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 137) {
if (tcor < 95) {
if (Rs < 2032) {
return 8;
}
else if (Rs >= 2032) {
return 1;
}
}
else if (tcor >= 95) {
if (mcTcor < 97) {
if (Cs < 3347) {
return 1;
}
else if (Cs >= 3347) {
return 8;
}
}
else if (mcTcor >= 97) {
return 8;
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect13(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1332) {
if (Rs < 2356) {
if (Cs < 535) {
if (negBalance < 14) {
if (mu_mv_mag_sq < 107) {
if (posBalance < 52) {
if (MVDiff < 8) {
if (SC < 926) {
if (TSCindex < 1) {
if (ssDCval < 26) {
if (ssDCval < 21) {
return 8;
}
else if (ssDCval >= 21) {
return 4;
}
}
else if (ssDCval >= 26) {
if (Cs < 196) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
return 8;
}
}
else if (Cs >= 196) {
return 1;
}
}
}
else if (TSCindex >= 1) {
if (AFD < 1) {
if (ssDCval < 76) {
if (Rs < 84) {
if (Rs < 51) {
return 4;
}
else if (Rs >= 51) {
return 1;
}
}
else if (Rs >= 84) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
return 2;
}
}
}
}
else if (ssDCval >= 76) {
if (posBalance < 29) {
return 1;
}
else if (posBalance >= 29) {
return 4;
}
}
}
else if (AFD >= 1) {
if (ssDCval < 29) {
if (ssDCval < 25) {
return 8;
}
else if (ssDCval >= 25) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
return 1;
}
}
}
else if (ssDCval >= 29) {
if (Cs < 116) {
if (ssDCval < 31) {
return 4;
}
else if (ssDCval >= 31) {
return 8;
}
}
else if (Cs >= 116) {
if (TSCindex < 2) {
if (SC < 785) {
if (diffMVdiffVal < 3) {
if (tcor < 100) {
if (Rs < 125) {
if (Rs < 115) {
return 4;
}
else if (Rs >= 115) {
return 1;
}
}
else if (Rs >= 125) {
return 4;
}
}
else if (tcor >= 100) {
return 1;
}
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
else if (SC >= 785) {
if (Rs < 576) {
return 1;
}
else if (Rs >= 576) {
return 4;
}
}
}
else if (TSCindex >= 2) {
if (ssDCval < 40) {
if (Cs < 394) {
return 4;
}
else if (Cs >= 394) {
return 1;
}
}
else if (ssDCval >= 40) {
if (tcor < 95) {
return 1;
}
else if (tcor >= 95) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
}
}
}
}
}
}
else if (SC >= 926) {
if (AFD < 3) {
if (Rs < 547) {
if (Cs < 392) {
return 1;
}
else if (Cs >= 392) {
return 8;
}
}
else if (Rs >= 547) {
return 8;
}
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (MVDiff >= 8) {
if (tcor < 94) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (refDCval < 31) {
if (Cs < 179) {
if (Rs < 8) {
return 4;
}
else if (Rs >= 8) {
return 1;
}
}
else if (Cs >= 179) {
if (tcor < 91) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 249) {
if (Rs < 200) {
return 1;
}
else if (Rs >= 200) {
return 4;
}
}
else if (Rs >= 249) {
return 8;
}
}
}
else if (tcor >= 91) {
return 4;
}
}
}
else if (refDCval >= 31) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
return 2;
}
}
}
}
else if (tcor >= 94) {
if (ssDCval < 112) {
if (SC < 149) {
if (posBalance < 11) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (posBalance >= 11) {
if (Rs < 7) {
return 2;
}
else if (Rs >= 7) {
return 1;
}
}
}
else if (SC >= 149) {
if (mu_mv_mag_sq < 89) {
return 4;
}
else if (mu_mv_mag_sq >= 89) {
if (ssDCval < 66) {
if (Rs < 151) {
return 1;
}
else if (Rs >= 151) {
return 4;
}
}
else if (ssDCval >= 66) {
return 4;
}
}
}
}
else if (ssDCval >= 112) {
if (Cs < 401) {
if (ssDCval < 161) {
return 1;
}
else if (ssDCval >= 161) {
if (Rs < 197) {
return 2;
}
else if (Rs >= 197) {
return 4;
}
}
}
else if (Cs >= 401) {
return 8;
}
}
}
}
}
else if (posBalance >= 52) {
if (negBalance < 8) {
if (TSC < 6) {
if (posBalance < 94) {
if (ssDCval < 77) {
if (MVDiff < 105) {
if (SC < 121) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 17) {
return 2;
}
else if (mu_mv_mag_sq >= 17) {
if (tcor < 83) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
if (Cs < 1) {
return 4;
}
else if (Cs >= 1) {
return 8;
}
}
}
else if (tcor >= 83) {
if (mu_mv_mag_sq < 42) {
return 2;
}
else if (mu_mv_mag_sq >= 42) {
return 8;
}
}
}
}
else if (ssDCval >= 50) {
return 2;
}
}
else if (SC >= 121) {
return 8;
}
}
else if (MVDiff >= 105) {
if (Rs < 138) {
return 2;
}
else if (Rs >= 138) {
if (Rs < 501) {
return 1;
}
else if (Rs >= 501) {
return 4;
}
}
}
}
else if (ssDCval >= 77) {
if (Rs < 180) {
return 4;
}
else if (Rs >= 180) {
if (Rs < 1081) {
return 2;
}
else if (Rs >= 1081) {
if (Rs < 1971) {
return 1;
}
else if (Rs >= 1971) {
return 8;
}
}
}
}
}
else if (posBalance >= 94) {
if (Cs < 92) {
if (tcor < 90) {
if (TSC < 3) {
if (MVDiff < 47) {
if (posBalance < 95) {
return 4;
}
else if (posBalance >= 95) {
return 2;
}
}
else if (MVDiff >= 47) {
if (Rs < 200) {
return 4;
}
else if (Rs >= 200) {
return 1;
}
}
}
else if (TSC >= 3) {
if (gchDC < 6) {
return 8;
}
else if (gchDC >= 6) {
if (Rs < 3) {
return 1;
}
else if (Rs >= 3) {
return 2;
}
}
}
}
else if (tcor >= 90) {
if (AFD < 6) {
if (tcor < 100) {
return 4;
}
else if (tcor >= 100) {
return 1;
}
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (Cs >= 92) {
if (ssDCval < 48) {
if (Rs < 286) {
return 1;
}
else if (Rs >= 286) {
return 4;
}
}
else if (ssDCval >= 48) {
if (Rs < 702) {
if (tcor < 99) {
if (Cs < 101) {
if (Rs < 103) {
return 8;
}
else if (Rs >= 103) {
return 4;
}
}
else if (Cs >= 101) {
return 8;
}
}
else if (tcor >= 99) {
return 1;
}
}
else if (Rs >= 702) {
return 4;
}
}
}
}
}
else if (TSC >= 6) {
if (Cs < 120) {
if (RsDiff < 9) {
if (mu_mv_mag_sq < 4) {
if (Rs < 328) {
return 8;
}
else if (Rs >= 328) {
return 2;
}
}
else if (mu_mv_mag_sq >= 4) {
return 1;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (Cs >= 120) {
if (negBalance < 5) {
if (gchDC < 10) {
if (Cs < 239) {
return 4;
}
else if (Cs >= 239) {
return 1;
}
}
else if (gchDC >= 10) {
return 8;
}
}
else if (negBalance >= 5) {
return 2;
}
}
}
}
else if (negBalance >= 8) {
if (MVDiff < 7) {
if (Cs < 214) {
return 8;
}
else if (Cs >= 214) {
if (Rs < 2029) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (Rs >= 2029) {
return 2;
}
}
}
else if (MVDiff >= 7) {
if (tcor < 92) {
if (posBalance < 68) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (Rs < 440) {
return 2;
}
else if (Rs >= 440) {
return 8;
}
}
}
else if (posBalance >= 68) {
if (gchDC < 12) {
if (diffRsCsdiff < 6) {
if (MVDiff < 27) {
if (MVDiff < 17) {
if (Rs < 1305) {
return 8;
}
else if (Rs >= 1305) {
return 2;
}
}
else if (MVDiff >= 17) {
if (Rs < 102) {
return 2;
}
else if (Rs >= 102) {
return 1;
}
}
}
else if (MVDiff >= 27) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Cs < 184) {
return 4;
}
else if (Cs >= 184) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 6) {
if (Cs < 375) {
if (MVDiff < 163) {
return 2;
}
else if (MVDiff >= 163) {
return 1;
}
}
else if (Cs >= 375) {
if (Rs < 1163) {
return 1;
}
else if (Rs >= 1163) {
return 8;
}
}
}
}
else if (gchDC >= 12) {
if (Rs < 78) {
return 8;
}
else if (Rs >= 78) {
if (ssDCval < 111) {
return 1;
}
else if (ssDCval >= 111) {
if (MVDiff < 361) {
return 2;
}
else if (MVDiff >= 361) {
return 1;
}
}
}
}
}
}
else if (tcor >= 92) {
if (Rs < 977) {
if (SC < 724) {
if (Cs < 48) {
return 1;
}
else if (Cs >= 48) {
if (Rs < 172) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
return 4;
}
}
else if (Rs >= 172) {
return 1;
}
}
}
else if (SC >= 724) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 8;
}
}
}
else if (Rs >= 977) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (SC < 782) {
if (refDCval < 47) {
if (Cs < 105) {
if (ssDCval < 38) {
if (refDCval < 22) {
if (ssDCval < 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 10) {
if (MVDiff < 1076) {
return 8;
}
else if (MVDiff >= 1076) {
return 4;
}
}
}
else if (refDCval >= 22) {
if (Rs < 56) {
if (SC < 16) {
return 4;
}
else if (SC >= 16) {
if (mu_mv_mag_sq < 119) {
return 4;
}
else if (mu_mv_mag_sq >= 119) {
if (ssDCval < 36) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (ssDCval >= 36) {
return 4;
}
}
}
}
else if (Rs >= 56) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
}
}
else if (ssDCval >= 38) {
if (Cs < 84) {
return 8;
}
else if (Cs >= 84) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
return 1;
}
}
}
}
else if (Cs >= 105) {
if (Rs < 249) {
if (ssDCval < 36) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (posBalance < 35) {
if (Rs < 93) {
return 4;
}
else if (Rs >= 93) {
return 8;
}
}
else if (posBalance >= 35) {
return 1;
}
}
}
else if (ssDCval >= 36) {
if (MVDiff < 2) {
if (SC < 422) {
if (Cs < 180) {
return 1;
}
else if (Cs >= 180) {
if (negBalance < 12) {
return 4;
}
else if (negBalance >= 12) {
return 1;
}
}
}
else if (SC >= 422) {
if (tcor < 87) {
return 4;
}
else if (tcor >= 87) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (Cs < 221) {
return 8;
}
else if (Cs >= 221) {
if (Rs < 180) {
return 1;
}
else if (Rs >= 180) {
return 4;
}
}
}
}
}
else if (Rs >= 249) {
if (Rs < 419) {
if (refDCval < 39) {
if (SC < 529) {
return 8;
}
else if (SC >= 529) {
if (mu_mv_mag_sq < 128) {
return 4;
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 33) {
if (ssDCval < 23) {
return 8;
}
else if (ssDCval >= 23) {
if (Cs < 243) {
return 8;
}
else if (Cs >= 243) {
return 4;
}
}
}
else if (ssDCval >= 33) {
return 8;
}
}
}
}
else if (refDCval >= 39) {
if (Cs < 187) {
return 4;
}
else if (Cs >= 187) {
if (ssDCval < 41) {
return 4;
}
else if (ssDCval >= 41) {
return 1;
}
}
}
}
else if (Rs >= 419) {
return 8;
}
}
}
}
else if (refDCval >= 47) {
if (ssDCval < 84) {
if (Rs < 23) {
return 2;
}
else if (Rs >= 23) {
if (ssDCval < 66) {
if (Cs < 283) {
if (MVDiff < 1) {
if (negBalance < 4) {
if (Rs < 178) {
return 8;
}
else if (Rs >= 178) {
return 4;
}
}
else if (negBalance >= 4) {
return 8;
}
}
else if (MVDiff >= 1) {
if (SC < 248) {
return 8;
}
else if (SC >= 248) {
if (negBalance < 7) {
if (ssDCval < 50) {
if (Rs < 239) {
return 8;
}
else if (Rs >= 239) {
return 4;
}
}
else if (ssDCval >= 50) {
return 8;
}
}
else if (negBalance >= 7) {
if (Cs < 182) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 8;
}
}
else if (Cs >= 182) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 4;
}
}
}
}
}
}
else if (Cs >= 283) {
if (Rs < 252) {
return 4;
}
else if (Rs >= 252) {
if (Rs < 404) {
return 1;
}
else if (Rs >= 404) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
if (Cs < 238) {
if (Rs < 79) {
return 4;
}
else if (Rs >= 79) {
return 2;
}
}
else if (Cs >= 238) {
if (Cs < 267) {
if (Rs < 320) {
return 1;
}
else if (Rs >= 320) {
return 4;
}
}
else if (Cs >= 267) {
return 8;
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (ssDCval < 94) {
if (Rs < 128) {
return 4;
}
else if (Rs >= 128) {
return 2;
}
}
else if (ssDCval >= 94) {
if (negBalance < 7) {
if (ssDCval < 130) {
if (Rs < 341) {
return 8;
}
else if (Rs >= 341) {
if (Rs < 403) {
return 2;
}
else if (Rs >= 403) {
return 4;
}
}
}
else if (ssDCval >= 130) {
return 4;
}
}
else if (negBalance >= 7) {
return 4;
}
}
}
}
}
}
else if (SC >= 782) {
if (diffTSC < -3) {
if (SC < 980) {
if (diffMVdiffVal < -12) {
return 4;
}
else if (diffMVdiffVal >= -12) {
if (Cs < 359) {
return 8;
}
else if (Cs >= 359) {
return 2;
}
}
}
else if (SC >= 980) {
if (diffAFD < -8) {
if (ssDCval < 147) {
return 2;
}
else if (ssDCval >= 147) {
return 8;
}
}
else if (diffAFD >= -8) {
if (Rs < 1532) {
return 1;
}
else if (Rs >= 1532) {
return 2;
}
}
}
}
else if (diffTSC >= -3) {
if (mu_mv_mag_sq < 128) {
if (SC < 956) {
if (Rs < 388) {
return 1;
}
else if (Rs >= 388) {
if (ssDCval < 36) {
if (Cs < 258) {
return 4;
}
else if (Cs >= 258) {
return 8;
}
}
else if (ssDCval >= 36) {
if (ssDCval < 122) {
if (mu_mv_mag_sq < 122) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (ssDCval < 63) {
if (Cs < 454) {
return 1;
}
else if (Cs >= 454) {
return 8;
}
}
else if (ssDCval >= 63) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 122) {
return 8;
}
}
else if (ssDCval >= 122) {
if (Rs < 483) {
return 1;
}
else if (Rs >= 483) {
return 8;
}
}
}
}
}
else if (SC >= 956) {
if (ssDCval < 67) {
if (refDCval < 60) {
if (Rs < 1314) {
if (CsDiff < 1) {
if (mcTcor < 100) {
if (mu_mv_mag_sq < 116) {
if (refDCval < 54) {
return 8;
}
else if (refDCval >= 54) {
return 1;
}
}
else if (mu_mv_mag_sq >= 116) {
return 8;
}
}
else if (mcTcor >= 100) {
if (Rs < 829) {
return 8;
}
else if (Rs >= 829) {
return 1;
}
}
}
else if (CsDiff >= 1) {
if (Rs < 654) {
return 1;
}
else if (Rs >= 654) {
return 2;
}
}
}
else if (Rs >= 1314) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (refDCval >= 60) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
if (Rs < 1471) {
return 2;
}
else if (Rs >= 1471) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (Rs < 489) {
return 1;
}
else if (Rs >= 489) {
if (SC < 1049) {
if (SC < 1014) {
return 8;
}
else if (SC >= 1014) {
return 4;
}
}
else if (SC >= 1049) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 1113) {
if (SC < 1260) {
return 8;
}
else if (SC >= 1260) {
if (Rs < 1006) {
return 8;
}
else if (Rs >= 1006) {
return 2;
}
}
}
else if (Rs >= 1113) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (Cs < 317) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (ssDCval < 87) {
if (Cs < 243) {
return 2;
}
else if (Cs >= 243) {
return 1;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
}
else if (Cs >= 317) {
if (SC < 1882) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 590) {
return 8;
}
else if (Rs >= 590) {
return 4;
}
}
}
}
else if (SC >= 1882) {
if (Rs < 1452) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (Rs >= 1452) {
return 8;
}
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (mcTcor < 84) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (tcor < 60) {
if (diffTSC < 1) {
if (Cs < 317) {
if (negBalance < 45) {
if (posBalance < 69) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 69) {
return 8;
}
}
else if (negBalance >= 45) {
if (Rs < 842) {
return 8;
}
else if (Rs >= 842) {
return 1;
}
}
}
else if (Cs >= 317) {
if (RsCsDiff < 16) {
return 2;
}
else if (RsCsDiff >= 16) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (negBalance < 49) {
if (Cs < 389) {
if (diffMVdiffVal < -6) {
if (diffMVdiffVal < -15) {
return 1;
}
else if (diffMVdiffVal >= -15) {
if (MVDiff < 123) {
return 8;
}
else if (MVDiff >= 123) {
return 1;
}
}
}
else if (diffMVdiffVal >= -6) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (Rs < 324) {
return 8;
}
else if (Rs >= 324) {
return 1;
}
}
}
}
else if (Cs >= 389) {
return 2;
}
}
else if (negBalance >= 49) {
if (Rs < 50) {
return 4;
}
else if (Rs >= 50) {
return 8;
}
}
}
}
else if (tcor >= 60) {
if (refDCval < 23) {
if (Cs < 236) {
if (Rs < 298) {
return 8;
}
else if (Rs >= 298) {
return 1;
}
}
else if (Cs >= 236) {
return 4;
}
}
else if (refDCval >= 23) {
if (mcTcor < 72) {
if (ssDCval < 28) {
if (ssDCval < 21) {
return 1;
}
else if (ssDCval >= 21) {
return 4;
}
}
else if (ssDCval >= 28) {
return 2;
}
}
else if (mcTcor >= 72) {
if (ssDCval < 33) {
if (AFD < 5) {
if (mu_mv_mag_sq < 43) {
if (Cs < 256) {
if (Rs < 553) {
return 1;
}
else if (Rs >= 553) {
return 4;
}
}
else if (Cs >= 256) {
if (Rs < 441) {
return 2;
}
else if (Rs >= 441) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 536) {
return 8;
}
else if (Rs >= 536) {
return 1;
}
}
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (ssDCval >= 33) {
if (Rs < 357) {
return 8;
}
else if (Rs >= 357) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 84) {
if (posBalance < 27) {
if (MVDiff < 5) {
if (tcor < 102) {
if (negBalance < 22) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
return 8;
}
}
else if (negBalance >= 22) {
return 1;
}
}
else if (tcor >= 102) {
if (Rs < 209) {
return 8;
}
else if (Rs >= 209) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (SC < 23) {
if (mu_mv_mag_sq < 96) {
if (diffAFD < 1) {
if (TSCindex < 1) {
return 4;
}
else if (TSCindex >= 1) {
if (MVDiff < 28) {
if (MVDiff < 18) {
if (posBalance < 4) {
return 8;
}
else if (posBalance >= 4) {
return 4;
}
}
else if (MVDiff >= 18) {
return 4;
}
}
else if (MVDiff >= 28) {
return 8;
}
}
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 96) {
return 4;
}
}
else if (SC >= 23) {
if (posBalance < 4) {
return 4;
}
else if (posBalance >= 4) {
if (ssDCval < 31) {
if (mu_mv_mag_sq < 80) {
if (negBalance < 61) {
return 1;
}
else if (negBalance >= 61) {
if (Rs < 93) {
return 8;
}
else if (Rs >= 93) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (Rs < 205) {
if (posBalance < 19) {
return 4;
}
else if (posBalance >= 19) {
return 1;
}
}
else if (Rs >= 205) {
return 4;
}
}
}
else if (ssDCval >= 31) {
if (Rs < 382) {
if (mcTcor < 98) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (Rs >= 382) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 27) {
if (ssDCval < 29) {
if (Cs < 359) {
if (MVDiff < 33) {
if (mu_mv_mag_sq < 41) {
if (posBalance < 71) {
if (posBalance < 41) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (mcTcor < 87) {
if (ssDCval < 24) {
return 1;
}
else if (ssDCval >= 24) {
return 4;
}
}
else if (mcTcor >= 87) {
return 1;
}
}
}
else if (posBalance >= 41) {
return 1;
}
}
else if (posBalance >= 71) {
return 2;
}
}
else if (mu_mv_mag_sq >= 41) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (mu_mv_mag_sq < 89) {
return 1;
}
else if (mu_mv_mag_sq >= 89) {
return 8;
}
}
}
}
else if (MVDiff >= 33) {
if (Cs < 69) {
return 1;
}
else if (Cs >= 69) {
if (MVDiff < 43) {
if (Rs < 269) {
return 4;
}
else if (Rs >= 269) {
return 1;
}
}
else if (MVDiff >= 43) {
if (MVDiff < 421) {
if (posBalance < 33) {
return 1;
}
else if (posBalance >= 33) {
return 8;
}
}
else if (MVDiff >= 421) {
return 2;
}
}
}
}
}
else if (Cs >= 359) {
if (Rs < 633) {
return 4;
}
else if (Rs >= 633) {
if (Cs < 403) {
return 1;
}
else if (Cs >= 403) {
return 8;
}
}
}
}
else if (ssDCval >= 29) {
if (posBalance < 42) {
if (Cs < 254) {
if (AFD < 3) {
if (mcTcor < 97) {
if (ssDCval < 31) {
if (MVDiff < 3) {
if (Rs < 175) {
return 4;
}
else if (Rs >= 175) {
return 8;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (ssDCval >= 31) {
if (Rs < 82) {
return 1;
}
else if (Rs >= 82) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 32) {
return 1;
}
else if (ssDCval >= 32) {
if (Rs < 112) {
if (Rs < 86) {
return 1;
}
else if (Rs >= 86) {
return 8;
}
}
else if (Rs >= 112) {
if (Rs < 438) {
return 1;
}
else if (Rs >= 438) {
return 4;
}
}
}
}
}
else if (AFD >= 3) {
if (mu_mv_mag_sq < 36) {
if (AFD < 7) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
if (tcor < 71) {
return 4;
}
else if (tcor >= 71) {
return 1;
}
}
}
else if (AFD >= 7) {
return 1;
}
}
else if (mu_mv_mag_sq >= 36) {
if (Rs < 187) {
if (Cs < 121) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
if (Rs < 119) {
return 1;
}
else if (Rs >= 119) {
return 4;
}
}
}
else if (Cs >= 121) {
if (Rs < 180) {
return 8;
}
else if (Rs >= 180) {
return 1;
}
}
}
else if (Rs >= 187) {
if (TSC < 3) {
if (MVDiff < 12) {
if (TSC < 2) {
return 2;
}
else if (TSC >= 2) {
if (mcTcor < 92) {
if (Rs < 285) {
return 4;
}
else if (Rs >= 285) {
return 8;
}
}
else if (mcTcor >= 92) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (MVDiff >= 12) {
if (diffMVdiffVal < 9) {
return 1;
}
else if (diffMVdiffVal >= 9) {
return 4;
}
}
}
else if (TSC >= 3) {
if (posBalance < 34) {
if (Rs < 342) {
return 4;
}
else if (Rs >= 342) {
return 1;
}
}
else if (posBalance >= 34) {
return 1;
}
}
}
}
}
}
else if (Cs >= 254) {
if (negBalance < 15) {
if (Rs < 268) {
return 1;
}
else if (Rs >= 268) {
return 4;
}
}
else if (negBalance >= 15) {
return 1;
}
}
}
else if (posBalance >= 42) {
if (negBalance < 17) {
if (mcTcor < 88) {
if (Rs < 109) {
return 4;
}
else if (Rs >= 109) {
return 8;
}
}
else if (mcTcor >= 88) {
if (Cs < 3) {
return 2;
}
else if (Cs >= 3) {
return 1;
}
}
}
else if (negBalance >= 17) {
if (tcor < 62) {
if (mu_mv_mag_sq < 10) {
if (Cs < 231) {
return 8;
}
else if (Cs >= 231) {
return 1;
}
}
else if (mu_mv_mag_sq >= 10) {
if (Cs < 174) {
return 1;
}
else if (Cs >= 174) {
return 4;
}
}
}
else if (tcor >= 62) {
if (ssDCval < 33) {
return 1;
}
else if (ssDCval >= 33) {
if (SC < 454) {
if (Cs < 123) {
if (Rs < 28) {
return 2;
}
else if (Rs >= 28) {
return 1;
}
}
else if (Cs >= 123) {
if (Rs < 192) {
return 2;
}
else if (Rs >= 192) {
if (Rs < 213) {
return 4;
}
else if (Rs >= 213) {
return 8;
}
}
}
}
else if (SC >= 454) {
if (Cs < 512) {
return 1;
}
else if (Cs >= 512) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 243) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 46) {
if (negBalance < 36) {
if (tcor < 92) {
if (CsDiff < 1) {
if (negBalance < 33) {
if (Rs < 39) {
return 1;
}
else if (Rs >= 39) {
if (Rs < 182) {
if (diffMVdiffVal < -16) {
return 4;
}
else if (diffMVdiffVal >= -16) {
return 2;
}
}
else if (Rs >= 182) {
return 4;
}
}
}
else if (negBalance >= 33) {
if (posBalance < 54) {
if (Rs < 96) {
return 4;
}
else if (Rs >= 96) {
return 2;
}
}
else if (posBalance >= 54) {
return 1;
}
}
}
else if (CsDiff >= 1) {
if (negBalance < 32) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 144) {
return 1;
}
else if (Rs >= 144) {
return 2;
}
}
}
else if (negBalance >= 32) {
if (Cs < 232) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 1;
}
}
else if (Cs >= 232) {
return 1;
}
}
}
}
else if (tcor >= 92) {
return 1;
}
}
else if (negBalance >= 36) {
if (mcTcor < 93) {
if (tcor < 65) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (Cs < 65) {
if (mu_mv_mag_sq < 9) {
return 4;
}
else if (mu_mv_mag_sq >= 9) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 4;
}
}
}
else if (Cs >= 65) {
if (mu_mv_mag_sq < 14) {
if (posBalance < 57) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (posBalance >= 57) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 14) {
return 8;
}
}
}
}
else if (tcor >= 65) {
if (SC < 257) {
if (MVDiff < 109) {
if (Cs < 25) {
if (Rs < 194) {
return 2;
}
else if (Rs >= 194) {
return 1;
}
}
else if (Cs >= 25) {
if (mcTcor < 84) {
if (Rs < 151) {
return 8;
}
else if (Rs >= 151) {
return 1;
}
}
else if (mcTcor >= 84) {
if (Rs < 22) {
return 4;
}
else if (Rs >= 22) {
return 1;
}
}
}
}
else if (MVDiff >= 109) {
if (posBalance < 16) {
return 8;
}
else if (posBalance >= 16) {
return 2;
}
}
}
else if (SC >= 257) {
if (ssDCval < 38) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (ssDCval < 37) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
if (Cs < 206) {
if (Rs < 131) {
return 8;
}
else if (Rs >= 131) {
return 2;
}
}
else if (Cs >= 206) {
return 8;
}
}
}
else if (ssDCval >= 37) {
return 8;
}
}
}
else if (ssDCval >= 38) {
if (mcTcor < 83) {
if (Cs < 363) {
return 8;
}
else if (Cs >= 363) {
return 1;
}
}
else if (mcTcor >= 83) {
if (mu_mv_mag_sq < 5) {
return 2;
}
else if (mu_mv_mag_sq >= 5) {
if (Cs < 290) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 1;
}
}
else if (Cs >= 290) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 93) {
if (negBalance < 77) {
if (Rs < 16) {
if (Rs < 7) {
return 8;
}
else if (Rs >= 7) {
return 4;
}
}
else if (Rs >= 16) {
if (diffRsCsdiff < -19) {
if (Rs < 142) {
return 8;
}
else if (Rs >= 142) {
return 4;
}
}
else if (diffRsCsdiff >= -19) {
if (RsCsDiff < 2) {
if (ssDCval < 39) {
if (ssDCval < 38) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 4;
}
}
}
else if (ssDCval >= 38) {
return 4;
}
}
else if (ssDCval >= 39) {
if (Cs < 138) {
return 1;
}
else if (Cs >= 138) {
if (Rs < 84) {
return 4;
}
else if (Rs >= 84) {
return 1;
}
}
}
}
else if (RsCsDiff >= 2) {
if (mu_mv_mag_sq < 6) {
if (ssDCval < 44) {
if (Rs < 84) {
return 1;
}
else if (Rs >= 84) {
return 2;
}
}
else if (ssDCval >= 44) {
return 1;
}
}
else if (mu_mv_mag_sq >= 6) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (Cs < 270) {
if (Rs < 188) {
return 2;
}
else if (Rs >= 188) {
return 4;
}
}
else if (Cs >= 270) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 77) {
if (posBalance < 2) {
if (Rs < 2) {
return 8;
}
else if (Rs >= 2) {
return 4;
}
}
else if (posBalance >= 2) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (negBalance < 27) {
if (mu_mv_mag_sq < 104) {
if (ssDCval < 39) {
if (Rs < 119) {
return 1;
}
else if (Rs >= 119) {
if (posBalance < 44) {
if (mu_mv_mag_sq < 89) {
return 8;
}
else if (mu_mv_mag_sq >= 89) {
if (Rs < 213) {
return 4;
}
else if (Rs >= 213) {
return 8;
}
}
}
else if (posBalance >= 44) {
if (Rs < 163) {
return 8;
}
else if (Rs >= 163) {
if (mu_mv_mag_sq < 52) {
return 8;
}
else if (mu_mv_mag_sq >= 52) {
return 1;
}
}
}
}
}
else if (ssDCval >= 39) {
if (Cs < 69) {
if (refDCval < 42) {
return 1;
}
else if (refDCval >= 42) {
return 8;
}
}
else if (Cs >= 69) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Cs < 89) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 8;
}
}
else if (Cs >= 89) {
if (negBalance < 22) {
if (tcor < 94) {
if (Rs < 99) {
return 4;
}
else if (Rs >= 99) {
return 1;
}
}
else if (tcor >= 94) {
if (ssDCval < 35) {
if (Rs < 121) {
return 4;
}
else if (Rs >= 121) {
return 1;
}
}
else if (ssDCval >= 35) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (ssDCval < 43) {
return 1;
}
else if (ssDCval >= 43) {
if (Rs < 206) {
return 8;
}
else if (Rs >= 206) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 22) {
return 8;
}
}
}
}
else if (negBalance >= 27) {
if (Cs < 93) {
if (TSCindex < 3) {
if (ssDCval < 41) {
return 1;
}
else if (ssDCval >= 41) {
return 8;
}
}
else if (TSCindex >= 3) {
if (Rs < 163) {
if (mu_mv_mag_sq < 81) {
if (diffMVdiffVal < 22) {
if (refDCval < 36) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
else if (refDCval >= 36) {
if (Rs < 53) {
if (gchDC < 9) {
return 8;
}
else if (gchDC >= 9) {
return 4;
}
}
else if (Rs >= 53) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 22) {
if (Rs < 10) {
return 8;
}
else if (Rs >= 10) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (diffMVdiffVal < 4) {
if (ssDCval < 48) {
return 4;
}
else if (ssDCval >= 48) {
return 2;
}
}
else if (diffMVdiffVal >= 4) {
if (posBalance < 66) {
if (posBalance < 55) {
if (refDCval < 41) {
if (Rs < 101) {
return 4;
}
else if (Rs >= 101) {
return 2;
}
}
else if (refDCval >= 41) {
if (Rs < 16) {
return 2;
}
else if (Rs >= 16) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (posBalance >= 66) {
return 4;
}
}
}
}
else if (Rs >= 163) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (Cs >= 93) {
if (refDCval < 40) {
if (ssDCval < 34) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
return 8;
}
}
else if (ssDCval >= 34) {
if (posBalance < 68) {
if (negBalance < 29) {
if (tcor < 83) {
return 1;
}
else if (tcor >= 83) {
return 8;
}
}
else if (negBalance >= 29) {
return 8;
}
}
else if (posBalance >= 68) {
return 4;
}
}
}
else if (refDCval >= 40) {
if (Cs < 181) {
if (posBalance < 57) {
if (diffRsCsdiff < -2) {
if (Rs < 153) {
return 4;
}
else if (Rs >= 153) {
return 8;
}
}
else if (diffRsCsdiff >= -2) {
if (MVDiff < 1) {
if (Rs < 175) {
return 4;
}
else if (Rs >= 175) {
return 8;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (Rs < 89) {
return 1;
}
else if (Rs >= 89) {
if (Rs < 161) {
return 2;
}
else if (Rs >= 161) {
return 4;
}
}
}
}
else if (Cs >= 181) {
if (TSC < 4) {
if (Cs < 211) {
return 4;
}
else if (Cs >= 211) {
return 1;
}
}
else if (TSC >= 4) {
if (Cs < 309) {
return 4;
}
else if (Cs >= 309) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 50) {
if (TSCindex < 2) {
if (ssDCval < 65) {
if (posBalance < 30) {
if (Rs < 23) {
return 2;
}
else if (Rs >= 23) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
return 1;
}
}
}
else if (posBalance >= 30) {
if (Rs < 66) {
return 1;
}
else if (Rs >= 66) {
if (diffMVdiffVal < -22) {
if (Rs < 127) {
return 4;
}
else if (Rs >= 127) {
return 8;
}
}
else if (diffMVdiffVal >= -22) {
if (Rs < 139) {
return 4;
}
else if (Rs >= 139) {
if (SC < 352) {
if (MVDiff < 3) {
if (Rs < 157) {
return 1;
}
else if (Rs >= 157) {
return 4;
}
}
else if (MVDiff >= 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 1;
}
}
}
else if (SC >= 352) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 65) {
if (ssDCval < 93) {
if (Cs < 259) {
if (SC < 254) {
if (SC < 201) {
if (Rs < 88) {
return 4;
}
else if (Rs >= 88) {
return 1;
}
}
else if (SC >= 201) {
if (Rs < 122) {
if (Rs < 96) {
return 1;
}
else if (Rs >= 96) {
return 2;
}
}
else if (Rs >= 122) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
return 1;
}
}
}
}
else if (SC >= 254) {
if (posBalance < 25) {
return 1;
}
else if (posBalance >= 25) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
if (Cs < 153) {
if (Rs < 196) {
return 2;
}
else if (Rs >= 196) {
return 8;
}
}
else if (Cs >= 153) {
return 2;
}
}
}
}
}
else if (Cs >= 259) {
if (Cs < 378) {
if (negBalance < 19) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
return 4;
}
}
}
else if (negBalance >= 19) {
if (Rs < 156) {
return 4;
}
else if (Rs >= 156) {
if (Cs < 306) {
return 1;
}
else if (Cs >= 306) {
if (Rs < 158) {
return 1;
}
else if (Rs >= 158) {
return 4;
}
}
}
}
}
else if (Cs >= 378) {
return 1;
}
}
}
else if (ssDCval >= 93) {
if (Cs < 139) {
if (mu_mv_mag_sq < 46) {
if (Rs < 88) {
return 4;
}
else if (Rs >= 88) {
return 1;
}
}
else if (mu_mv_mag_sq >= 46) {
if (diffMVdiffVal < -24) {
return 8;
}
else if (diffMVdiffVal >= -24) {
return 1;
}
}
}
else if (Cs >= 139) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
}
}
}
else if (TSCindex >= 2) {
if (refDCval < 71) {
if (Cs < 192) {
if (RsCsDiff < 1) {
if (MVDiff < 1) {
if (SC < 245) {
if (ssDCval < 58) {
if (Cs < 104) {
return 2;
}
else if (Cs >= 104) {
return 8;
}
}
else if (ssDCval >= 58) {
if (Cs < 119) {
return 1;
}
else if (Cs >= 119) {
return 8;
}
}
}
else if (SC >= 245) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 2;
}
}
}
else if (MVDiff >= 1) {
if (tcor < 73) {
if (negBalance < 38) {
if (Rs < 67) {
return 8;
}
else if (Rs >= 67) {
return 2;
}
}
else if (negBalance >= 38) {
return 4;
}
}
else if (tcor >= 73) {
if (tcor < 100) {
if (mcTcor < 95) {
if (diffTSC < 1) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Rs < 58) {
return 1;
}
else if (Rs >= 58) {
return 2;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 70) {
return 2;
}
else if (Rs >= 70) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (tcor < 94) {
if (TSCindex < 3) {
if (Rs < 187) {
return 1;
}
else if (Rs >= 187) {
return 4;
}
}
else if (TSCindex >= 3) {
if (Cs < 66) {
return 1;
}
else if (Cs >= 66) {
if (ssDCval < 59) {
if (mu_mv_mag_sq < 35) {
return 1;
}
else if (mu_mv_mag_sq >= 35) {
return 4;
}
}
else if (ssDCval >= 59) {
if (TSC < 3) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (Rs < 170) {
return 1;
}
else if (Rs >= 170) {
return 4;
}
}
}
else if (TSC >= 3) {
if (Cs < 73) {
return 4;
}
else if (Cs >= 73) {
return 1;
}
}
}
}
}
}
else if (tcor >= 94) {
if (ssDCval < 67) {
if (mu_mv_mag_sq < 84) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (mu_mv_mag_sq < 74) {
return 1;
}
else if (mu_mv_mag_sq >= 74) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (posBalance < 44) {
if (Cs < 127) {
if (Rs < 90) {
return 8;
}
else if (Rs >= 90) {
return 1;
}
}
else if (Cs >= 127) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
}
else if (posBalance >= 44) {
if (negBalance < 19) {
if (Rs < 135) {
return 8;
}
else if (Rs >= 135) {
return 4;
}
}
else if (negBalance >= 19) {
if (Rs < 208) {
if (ssDCval < 57) {
if (mcTcor < 99) {
if (Rs < 148) {
return 4;
}
else if (Rs >= 148) {
return 1;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (Rs >= 208) {
if (Cs < 187) {
return 4;
}
else if (Cs >= 187) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
return 2;
}
}
}
}
}
}
else if (tcor >= 100) {
if (Rs < 45) {
if (Rs < 3) {
return 2;
}
else if (Rs >= 3) {
return 1;
}
}
else if (Rs >= 45) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (TSC < 5) {
if (posBalance < 48) {
if (mu_mv_mag_sq < 89) {
if (Cs < 89) {
return 2;
}
else if (Cs >= 89) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (negBalance < 19) {
return 1;
}
else if (negBalance >= 19) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
return 4;
}
}
}
}
else if (posBalance >= 48) {
if (SC < 178) {
if (negBalance < 52) {
if (diffMVdiffVal < 14) {
return 1;
}
else if (diffMVdiffVal >= 14) {
return 4;
}
}
else if (negBalance >= 52) {
if (Rs < 62) {
return 8;
}
else if (Rs >= 62) {
return 4;
}
}
}
else if (SC >= 178) {
if (MVDiff < 7) {
if (mcTcor < 99) {
if (Rs < 124) {
return 1;
}
else if (Rs >= 124) {
return 4;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (MVDiff >= 7) {
if (posBalance < 68) {
if (negBalance < 30) {
if (Rs < 115) {
return 1;
}
else if (Rs >= 115) {
if (SC < 322) {
return 8;
}
else if (SC >= 322) {
return 4;
}
}
}
else if (negBalance >= 30) {
if (mcTcor < 91) {
if (Rs < 131) {
return 4;
}
else if (Rs >= 131) {
if (Rs < 160) {
return 8;
}
else if (Rs >= 160) {
return 2;
}
}
}
else if (mcTcor >= 91) {
if (mu_mv_mag_sq < 93) {
return 4;
}
else if (mu_mv_mag_sq >= 93) {
return 1;
}
}
}
}
else if (posBalance >= 68) {
if (Cs < 165) {
return 4;
}
else if (Cs >= 165) {
return 1;
}
}
}
}
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 21) {
if (AFD < 20) {
if (diffAFD < 4) {
if (diffMVdiffVal < 27) {
if (Cs < 44) {
return 1;
}
else if (Cs >= 44) {
return 2;
}
}
else if (diffMVdiffVal >= 27) {
if (Rs < 170) {
return 1;
}
else if (Rs >= 170) {
return 2;
}
}
}
else if (diffAFD >= 4) {
return 1;
}
}
else if (AFD >= 20) {
if (diffAFD < 0) {
return 1;
}
else if (diffAFD >= 0) {
if (Rs < 182) {
return 4;
}
else if (Rs >= 182) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (TSC < 8) {
if (SC < 180) {
if (Cs < 68) {
return 1;
}
else if (Cs >= 68) {
return 8;
}
}
else if (SC >= 180) {
return 4;
}
}
else if (TSC >= 8) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
}
else if (Cs >= 192) {
if (TSC < 4) {
if (Rs < 170) {
if (Rs < 64) {
return 1;
}
else if (Rs >= 64) {
if (negBalance < 29) {
if (negBalance < 17) {
if (Rs < 144) {
return 2;
}
else if (Rs >= 144) {
return 1;
}
}
else if (negBalance >= 17) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (mu_mv_mag_sq < 96) {
if (Cs < 340) {
if (Rs < 102) {
return 4;
}
else if (Rs >= 102) {
return 1;
}
}
else if (Cs >= 340) {
return 2;
}
}
else if (mu_mv_mag_sq >= 96) {
if (Cs < 228) {
return 2;
}
else if (Cs >= 228) {
return 4;
}
}
}
}
}
else if (negBalance >= 29) {
if (mu_mv_mag_sq < 67) {
if (negBalance < 50) {
if (MVDiff < 2) {
if (Rs < 167) {
return 1;
}
else if (Rs >= 167) {
return 4;
}
}
else if (MVDiff >= 2) {
if (diffMVdiffVal < 5) {
return 4;
}
else if (diffMVdiffVal >= 5) {
if (ssDCval < 56) {
return 2;
}
else if (ssDCval >= 56) {
return 4;
}
}
}
}
else if (negBalance >= 50) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Rs < 101) {
return 2;
}
else if (Rs >= 101) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 133) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (Rs >= 133) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (ssDCval < 66) {
if (Cs < 262) {
if (Rs < 163) {
return 1;
}
else if (Rs >= 163) {
if (Cs < 250) {
return 4;
}
else if (Cs >= 250) {
return 1;
}
}
}
else if (Cs >= 262) {
return 4;
}
}
else if (ssDCval >= 66) {
if (Cs < 421) {
if (mcTcor < 98) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (Rs < 158) {
return 4;
}
else if (Rs >= 158) {
return 1;
}
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (Cs >= 421) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 170) {
if (Cs < 308) {
if (tcor < 98) {
if (mu_mv_mag_sq < 110) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
if (mcTcor < 97) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (mcTcor >= 97) {
if (Cs < 214) {
if (Cs < 204) {
return 4;
}
else if (Cs >= 204) {
return 1;
}
}
else if (Cs >= 214) {
if (mu_mv_mag_sq < 104) {
if (Rs < 236) {
if (SC < 400) {
if (SC < 396) {
return 4;
}
else if (SC >= 396) {
return 1;
}
}
else if (SC >= 400) {
return 4;
}
}
else if (Rs >= 236) {
return 2;
}
}
else if (mu_mv_mag_sq >= 104) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 110) {
return 1;
}
}
else if (tcor >= 98) {
if (negBalance < 26) {
if (mu_mv_mag_sq < 122) {
return 4;
}
else if (mu_mv_mag_sq >= 122) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
return 1;
}
}
}
else if (negBalance >= 26) {
if (ssDCval < 62) {
if (mu_mv_mag_sq < 108) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 108) {
return 1;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
}
}
else if (Cs >= 308) {
if (ssDCval < 59) {
if (Cs < 391) {
if (mu_mv_mag_sq < 69) {
if (SC < 625) {
return 4;
}
else if (SC >= 625) {
return 8;
}
}
else if (mu_mv_mag_sq >= 69) {
if (Cs < 386) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (Cs >= 386) {
return 4;
}
}
}
else if (Cs >= 391) {
if (Rs < 184) {
return 4;
}
else if (Rs >= 184) {
return 1;
}
}
}
else if (ssDCval >= 59) {
if (ssDCval < 69) {
if (Rs < 177) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (Cs < 340) {
return 4;
}
else if (Cs >= 340) {
if (mu_mv_mag_sq < 119) {
if (Rs < 174) {
return 4;
}
else if (Rs >= 174) {
return 1;
}
}
else if (mu_mv_mag_sq >= 119) {
return 1;
}
}
}
}
else if (Rs >= 177) {
if (MVDiff < 32) {
if (Cs < 319) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
return 1;
}
}
else if (Cs >= 319) {
return 4;
}
}
else if (MVDiff >= 32) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
return 4;
}
}
}
}
else if (ssDCval >= 69) {
if (Cs < 376) {
if (Rs < 173) {
return 1;
}
else if (Rs >= 173) {
if (Rs < 202) {
return 4;
}
else if (Rs >= 202) {
return 1;
}
}
}
else if (Cs >= 376) {
return 1;
}
}
}
}
}
}
else if (TSC >= 4) {
if (mcTcor < 82) {
if (TSC < 11) {
if (Rs < 228) {
return 4;
}
else if (Rs >= 228) {
return 2;
}
}
else if (TSC >= 11) {
if (Rs < 106) {
return 2;
}
else if (Rs >= 106) {
return 8;
}
}
}
else if (mcTcor >= 82) {
if (posBalance < 54) {
if (diffAFD < 1) {
if (mcTcor < 95) {
if (Rs < 193) {
return 1;
}
else if (Rs >= 193) {
return 8;
}
}
else if (mcTcor >= 95) {
if (MVDiff < 5) {
if (Rs < 228) {
return 4;
}
else if (Rs >= 228) {
return 1;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (negBalance < 64) {
return 2;
}
else if (negBalance >= 64) {
if (Rs < 162) {
return 1;
}
else if (Rs >= 162) {
return 8;
}
}
}
}
else if (posBalance >= 54) {
if (diffTSC < -1) {
if (MVDiff < 27) {
return 1;
}
else if (MVDiff >= 27) {
if (Cs < 459) {
return 2;
}
else if (Cs >= 459) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (refDCval < 64) {
if (Cs < 441) {
return 1;
}
else if (Cs >= 441) {
if (Rs < 213) {
if (Rs < 189) {
return 4;
}
else if (Rs >= 189) {
return 8;
}
}
else if (Rs >= 213) {
return 1;
}
}
}
else if (refDCval >= 64) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (Rs < 98) {
return 1;
}
else if (Rs >= 98) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 71) {
if (posBalance < 46) {
if (tcor < 101) {
if (MVDiff < 5) {
if (Cs < 124) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 87) {
return 4;
}
else if (Rs >= 87) {
return 1;
}
}
}
else if (Cs >= 124) {
if (Cs < 259) {
if (negBalance < 29) {
if (Cs < 214) {
if (Rs < 237) {
return 2;
}
else if (Rs >= 237) {
return 8;
}
}
else if (Cs >= 214) {
if (Cs < 219) {
return 1;
}
else if (Cs >= 219) {
return 2;
}
}
}
else if (negBalance >= 29) {
if (diffAFD < 1) {
if (Cs < 215) {
return 1;
}
else if (Cs >= 215) {
if (Rs < 209) {
return 1;
}
else if (Rs >= 209) {
return 2;
}
}
}
else if (diffAFD >= 1) {
return 2;
}
}
}
else if (Cs >= 259) {
if (Rs < 155) {
return 4;
}
else if (Rs >= 155) {
return 1;
}
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 106) {
if (Cs < 102) {
if (diffMVdiffVal < -14) {
if (Rs < 21) {
return 1;
}
else if (Rs >= 21) {
if (Rs < 59) {
return 4;
}
else if (Rs >= 59) {
return 8;
}
}
}
else if (diffMVdiffVal >= -14) {
return 1;
}
}
else if (Cs >= 102) {
if (SC < 264) {
if (MVDiff < 15) {
if (Rs < 65) {
return 4;
}
else if (Rs >= 65) {
return 2;
}
}
else if (MVDiff >= 15) {
if (Rs < 26) {
return 2;
}
else if (Rs >= 26) {
if (diffMVdiffVal < 12) {
return 1;
}
else if (diffMVdiffVal >= 12) {
if (Rs < 61) {
return 1;
}
else if (Rs >= 61) {
return 2;
}
}
}
}
}
else if (SC >= 264) {
if (refDCval < 80) {
if (diffTSC < -4) {
return 8;
}
else if (diffTSC >= -4) {
return 1;
}
}
else if (refDCval >= 80) {
if (Cs < 151) {
return 1;
}
else if (Cs >= 151) {
if (TSCindex < 3) {
if (Cs < 271) {
return 2;
}
else if (Cs >= 271) {
return 4;
}
}
else if (TSCindex >= 3) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 106) {
if (Rs < 1) {
return 1;
}
else if (Rs >= 1) {
if (MVDiff < 236) {
return 2;
}
else if (MVDiff >= 236) {
if (Rs < 69) {
return 2;
}
else if (Rs >= 69) {
return 8;
}
}
}
}
}
}
else if (tcor >= 101) {
if (MVDiff < 245) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Rs < 181) {
return 8;
}
else if (Rs >= 181) {
return 1;
}
}
}
else if (MVDiff >= 245) {
if (Rs < 23) {
return 1;
}
else if (Rs >= 23) {
return 4;
}
}
}
}
else if (posBalance >= 46) {
if (tcor < 88) {
if (mcTcor < 80) {
if (AFD < 13) {
if (Rs < 6) {
return 1;
}
else if (Rs >= 6) {
return 4;
}
}
else if (AFD >= 13) {
if (tcor < 28) {
if (MVDiff < 452) {
if (Rs < 156) {
return 8;
}
else if (Rs >= 156) {
return 1;
}
}
else if (MVDiff >= 452) {
return 4;
}
}
else if (tcor >= 28) {
if (diffAFD < 5) {
return 1;
}
else if (diffAFD >= 5) {
if (Rs < 112) {
return 2;
}
else if (Rs >= 112) {
return 4;
}
}
}
}
}
else if (mcTcor >= 80) {
if (negBalance < 40) {
if (diffMVdiffVal < 14) {
if (MVDiff < 4) {
if (Rs < 185) {
return 8;
}
else if (Rs >= 185) {
return 4;
}
}
else if (MVDiff >= 4) {
if (AFD < 24) {
if (Rs < 32) {
if (Cs < 75) {
if (mu_mv_mag_sq < 28) {
return 1;
}
else if (mu_mv_mag_sq >= 28) {
return 8;
}
}
else if (Cs >= 75) {
return 2;
}
}
else if (Rs >= 32) {
return 1;
}
}
else if (AFD >= 24) {
return 8;
}
}
}
else if (diffMVdiffVal >= 14) {
if (Cs < 242) {
return 2;
}
else if (Cs >= 242) {
return 1;
}
}
}
else if (negBalance >= 40) {
if (TSC < 4) {
if (ssDCval < 75) {
if (MVDiff < 13) {
if (Rs < 177) {
return 4;
}
else if (Rs >= 177) {
return 8;
}
}
else if (MVDiff >= 13) {
return 1;
}
}
else if (ssDCval >= 75) {
if (ssDCval < 105) {
if (SC < 462) {
if (mcTcor < 84) {
if (Rs < 128) {
if (Rs < 110) {
return 8;
}
else if (Rs >= 110) {
return 2;
}
}
else if (Rs >= 128) {
return 8;
}
}
else if (mcTcor >= 84) {
if (diffRsCsdiff < -1) {
if (MVDiff < 45) {
return 8;
}
else if (MVDiff >= 45) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
if (MVDiff < 3) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (MVDiff >= 3) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (mcTcor < 95) {
if (mcTcor < 91) {
if (SC < 125) {
return 8;
}
else if (SC >= 125) {
if (negBalance < 55) {
if (Cs < 21) {
return 2;
}
else if (Cs >= 21) {
if (Cs < 27) {
return 8;
}
else if (Cs >= 27) {
return 2;
}
}
}
else if (negBalance >= 55) {
if (Rs < 134) {
return 8;
}
else if (Rs >= 134) {
if (Rs < 141) {
return 2;
}
else if (Rs >= 141) {
return 8;
}
}
}
}
}
else if (mcTcor >= 91) {
if (diffMVdiffVal < 1) {
if (Rs < 119) {
return 8;
}
else if (Rs >= 119) {
if (Rs < 143) {
return 2;
}
else if (Rs >= 143) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
else if (mcTcor >= 95) {
return 2;
}
}
}
}
}
}
else if (SC >= 462) {
if (Rs < 216) {
return 8;
}
else if (Rs >= 216) {
if (Rs < 229) {
return 2;
}
else if (Rs >= 229) {
return 8;
}
}
}
}
else if (ssDCval >= 105) {
if (Cs < 183) {
return 2;
}
else if (Cs >= 183) {
return 1;
}
}
}
}
else if (TSC >= 4) {
if (SC < 522) {
if (MVDiff < 43) {
if (mu_mv_mag_sq < 72) {
if (Rs < 228) {
if (diffMVdiffVal < 7) {
return 1;
}
else if (diffMVdiffVal >= 7) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (Rs < 154) {
return 1;
}
else if (Rs >= 154) {
return 8;
}
}
}
}
else if (Rs >= 228) {
if (Rs < 236) {
return 2;
}
else if (Rs >= 236) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 72) {
return 8;
}
}
else if (MVDiff >= 43) {
if (ssDCval < 81) {
if (posBalance < 65) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
else if (posBalance >= 65) {
return 4;
}
}
else if (ssDCval >= 81) {
if (diffAFD < -2) {
if (MVDiff < 120) {
return 1;
}
else if (MVDiff >= 120) {
if (Rs < 188) {
return 2;
}
else if (Rs >= 188) {
return 1;
}
}
}
else if (diffAFD >= -2) {
if (Rs < 79) {
if (Cs < 52) {
return 2;
}
else if (Cs >= 52) {
return 1;
}
}
else if (Rs >= 79) {
if (tcor < 65) {
if (MVDiff < 89) {
return 8;
}
else if (MVDiff >= 89) {
if (Rs < 138) {
return 1;
}
else if (Rs >= 138) {
return 8;
}
}
}
else if (tcor >= 65) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
if (MVDiff < 60) {
return 2;
}
else if (MVDiff >= 60) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
}
}
}
else if (SC >= 522) {
if (negBalance < 50) {
if (SC < 573) {
if (negBalance < 46) {
if (Rs < 144) {
return 2;
}
else if (Rs >= 144) {
return 8;
}
}
else if (negBalance >= 46) {
return 2;
}
}
else if (SC >= 573) {
return 8;
}
}
else if (negBalance >= 50) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 88) {
if (Cs < 233) {
if (Rs < 86) {
if (ssDCval < 114) {
if (SC < 123) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Cs < 57) {
return 2;
}
else if (Cs >= 57) {
return 1;
}
}
}
else if (SC >= 123) {
if (Rs < 56) {
if (Rs < 20) {
return 2;
}
else if (Rs >= 20) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
if (MVDiff < 179) {
return 1;
}
else if (MVDiff >= 179) {
if (Rs < 40) {
return 1;
}
else if (Rs >= 40) {
return 4;
}
}
}
}
}
else if (Rs >= 56) {
if (AFD < 2) {
if (Rs < 78) {
return 4;
}
else if (Rs >= 78) {
return 2;
}
}
else if (AFD >= 2) {
if (diffRsCsdiff < 1) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (ssDCval < 94) {
return 1;
}
else if (ssDCval >= 94) {
if (Rs < 71) {
return 4;
}
else if (Rs >= 71) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 1) {
return 4;
}
}
}
}
}
else if (ssDCval >= 114) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 1) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Rs >= 1) {
return 2;
}
}
}
}
else if (Rs >= 86) {
if (Cs < 65) {
if (ssDCval < 97) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
return 1;
}
}
else if (ssDCval >= 97) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
if (Rs < 96) {
return 8;
}
else if (Rs >= 96) {
return 2;
}
}
}
}
}
else if (Cs >= 65) {
if (refDCval < 104) {
if (TSCindex < 3) {
if (ssDCval < 86) {
if (negBalance < 22) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 2;
}
}
else if (negBalance >= 22) {
if (SC < 403) {
if (ssDCval < 75) {
if (refDCval < 73) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (Rs < 98) {
return 2;
}
else if (Rs >= 98) {
return 1;
}
}
}
}
else if (refDCval >= 73) {
return 2;
}
}
else if (ssDCval >= 75) {
if (Rs < 160) {
if (Cs < 97) {
return 2;
}
else if (Cs >= 97) {
return 1;
}
}
else if (Rs >= 160) {
return 2;
}
}
}
else if (SC >= 403) {
if (SC < 455) {
return 1;
}
else if (SC >= 455) {
return 2;
}
}
}
}
else if (ssDCval >= 86) {
if (tcor < 99) {
if (SC < 360) {
return 1;
}
else if (SC >= 360) {
return 4;
}
}
else if (tcor >= 99) {
if (Cs < 182) {
if (Rs < 226) {
return 2;
}
else if (Rs >= 226) {
return 1;
}
}
else if (Cs >= 182) {
return 4;
}
}
}
}
else if (TSCindex >= 3) {
if (negBalance < 43) {
if (MVDiff < 2) {
if (Cs < 160) {
return 8;
}
else if (Cs >= 160) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (AFD < 6) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
if (Rs < 228) {
return 2;
}
else if (Rs >= 228) {
return 1;
}
}
}
}
}
else if (AFD >= 6) {
if (refDCval < 76) {
return 8;
}
else if (refDCval >= 76) {
if (Cs < 230) {
return 2;
}
else if (Cs >= 230) {
return 8;
}
}
}
}
}
else if (negBalance >= 43) {
if (mu_mv_mag_sq < 23) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
if (gchDC < 2) {
if (refDCval < 78) {
if (Rs < 164) {
return 4;
}
else if (Rs >= 164) {
return 2;
}
}
else if (refDCval >= 78) {
if (tcor < 97) {
return 2;
}
else if (tcor >= 97) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 23) {
return 8;
}
}
}
}
else if (refDCval >= 104) {
if (ssDCval < 118) {
return 2;
}
else if (ssDCval >= 118) {
if (Rs < 181) {
return 4;
}
else if (Rs >= 181) {
return 2;
}
}
}
}
}
}
else if (Cs >= 233) {
if (ssDCval < 76) {
if (refDCval < 72) {
if (Cs < 302) {
if (Rs < 110) {
return 4;
}
else if (Rs >= 110) {
return 1;
}
}
else if (Cs >= 302) {
return 4;
}
}
else if (refDCval >= 72) {
if (negBalance < 41) {
if (mcTcor < 98) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (mcTcor >= 98) {
if (ssDCval < 72) {
if (Rs < 132) {
return 4;
}
else if (Rs >= 132) {
return 1;
}
}
else if (ssDCval >= 72) {
return 1;
}
}
}
else if (negBalance >= 41) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 1;
}
}
}
}
else if (ssDCval >= 76) {
if (TSC < 3) {
if (ssDCval < 135) {
if (RsDiff < 1) {
if (AFD < 2) {
if (Cs < 265) {
return 2;
}
else if (Cs >= 265) {
return 4;
}
}
else if (AFD >= 2) {
if (ssDCval < 89) {
if (Cs < 435) {
return 4;
}
else if (Cs >= 435) {
return 1;
}
}
else if (ssDCval >= 89) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 135) {
return 2;
}
}
else if (TSC >= 3) {
if (negBalance < 52) {
if (SC < 430) {
if (ssDCval < 79) {
return 4;
}
else if (ssDCval >= 79) {
return 2;
}
}
else if (SC >= 430) {
return 2;
}
}
else if (negBalance >= 52) {
if (Rs < 186) {
if (Rs < 125) {
return 1;
}
else if (Rs >= 125) {
return 2;
}
}
else if (Rs >= 186) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 243) {
if (RsDiff < 4) {
if (refDCval < 74) {
if (mcTcor < 89) {
if (ssDCval < 41) {
if (tcor < 60) {
if (posBalance < 53) {
if (refDCval < 46) {
return 1;
}
else if (refDCval >= 46) {
if (Rs < 404) {
return 8;
}
else if (Rs >= 404) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (Rs < 286) {
return 4;
}
else if (Rs >= 286) {
if (Cs < 278) {
return 8;
}
else if (Cs >= 278) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 8;
}
}
}
}
}
else if (tcor >= 60) {
if (gchDC < 3) {
if (Cs < 152) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 732) {
return 1;
}
else if (Rs >= 732) {
return 2;
}
}
}
else if (Cs >= 152) {
if (negBalance < 28) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
return 1;
}
}
else if (negBalance >= 28) {
if (diffRsCsdiff < -6) {
if (ssDCval < 38) {
if (Rs < 641) {
return 2;
}
else if (Rs >= 641) {
return 1;
}
}
else if (ssDCval >= 38) {
if (refDCval < 41) {
return 8;
}
else if (refDCval >= 41) {
return 4;
}
}
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
}
}
else if (gchDC >= 3) {
if (Cs < 163) {
return 4;
}
else if (Cs >= 163) {
return 1;
}
}
}
}
else if (ssDCval >= 41) {
if (mu_mv_mag_sq < 101) {
if (refDCval < 51) {
if (MVDiff < 56) {
if (TSC < 6) {
if (AFD < 9) {
if (MVDiff < 19) {
if (mu_mv_mag_sq < 75) {
if (Cs < 523) {
return 8;
}
else if (Cs >= 523) {
return 1;
}
}
else if (mu_mv_mag_sq >= 75) {
return 1;
}
}
else if (MVDiff >= 19) {
if (Rs < 378) {
return 2;
}
else if (Rs >= 378) {
return 1;
}
}
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 45) {
if (AFD < 11) {
if (Rs < 424) {
return 4;
}
else if (Rs >= 424) {
return 2;
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (mu_mv_mag_sq >= 45) {
return 4;
}
}
}
else if (TSC >= 6) {
if (SC < 1194) {
if (SC < 560) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (SC >= 560) {
return 8;
}
}
else if (SC >= 1194) {
return 1;
}
}
}
else if (MVDiff >= 56) {
if (tcor < 49) {
return 1;
}
else if (tcor >= 49) {
if (Rs < 659) {
return 2;
}
else if (Rs >= 659) {
return 1;
}
}
}
}
else if (refDCval >= 51) {
if (ssDCval < 56) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
if (negBalance < 49) {
if (refDCval < 54) {
if (Rs < 570) {
if (SC < 455) {
return 4;
}
else if (SC >= 455) {
return 2;
}
}
else if (Rs >= 570) {
return 8;
}
}
else if (refDCval >= 54) {
return 1;
}
}
else if (negBalance >= 49) {
if (diffMVdiffVal < 0) {
return 8;
}
else if (diffMVdiffVal >= 0) {
if (Cs < 163) {
return 8;
}
else if (Cs >= 163) {
return 4;
}
}
}
}
}
else if (ssDCval >= 56) {
if (mcTcor < 81) {
if (refDCval < 66) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (Cs < 237) {
return 2;
}
else if (Cs >= 237) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 610) {
return 4;
}
else if (Rs >= 610) {
return 8;
}
}
}
}
}
else if (refDCval >= 66) {
if (ssDCval < 68) {
if (Rs < 260) {
return 1;
}
else if (Rs >= 260) {
return 4;
}
}
else if (ssDCval >= 68) {
return 2;
}
}
}
else if (mcTcor >= 81) {
if (posBalance < 41) {
if (Cs < 346) {
return 8;
}
else if (Cs >= 346) {
if (Rs < 410) {
return 1;
}
else if (Rs >= 410) {
return 2;
}
}
}
else if (posBalance >= 41) {
if (diffTSC < 1) {
if (negBalance < 52) {
if (mu_mv_mag_sq < 38) {
if (tcor < 52) {
return 4;
}
else if (tcor >= 52) {
return 1;
}
}
else if (mu_mv_mag_sq >= 38) {
if (AFD < 8) {
if (Rs < 743) {
return 2;
}
else if (Rs >= 743) {
return 4;
}
}
else if (AFD >= 8) {
if (Rs < 487) {
return 1;
}
else if (Rs >= 487) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (Rs < 364) {
return 2;
}
else if (Rs >= 364) {
return 4;
}
}
}
else if (diffTSC >= 1) {
if (Cs < 445) {
return 2;
}
else if (Cs >= 445) {
if (Rs < 529) {
return 8;
}
else if (Rs >= 529) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (AFD < 8) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
return 8;
}
}
else if (AFD >= 8) {
return 4;
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1273) {
if (posBalance < 26) {
if (diffRsCsdiff < -1) {
if (posBalance < 16) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (posBalance >= 16) {
if (Rs < 1121) {
return 2;
}
else if (Rs >= 1121) {
return 1;
}
}
}
else if (diffRsCsdiff >= -1) {
if (Rs < 793) {
if (ssDCval < 42) {
if (Rs < 327) {
return 1;
}
else if (Rs >= 327) {
return 4;
}
}
else if (ssDCval >= 42) {
if (TSC < 4) {
if (SC < 1142) {
if (Rs < 293) {
if (Rs < 273) {
return 8;
}
else if (Rs >= 273) {
return 4;
}
}
else if (Rs >= 293) {
return 8;
}
}
else if (SC >= 1142) {
return 1;
}
}
else if (TSC >= 4) {
if (Cs < 287) {
if (Rs < 539) {
return 8;
}
else if (Rs >= 539) {
return 1;
}
}
else if (Cs >= 287) {
return 4;
}
}
}
}
else if (Rs >= 793) {
if (Cs < 163) {
return 4;
}
else if (Cs >= 163) {
if (Cs < 394) {
return 1;
}
else if (Cs >= 394) {
return 8;
}
}
}
}
}
else if (posBalance >= 26) {
if (TSC < 1) {
if (mu_mv_mag_sq < 105) {
return 4;
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 357) {
return 2;
}
else if (Rs >= 357) {
if (Rs < 1042) {
return 8;
}
else if (Rs >= 1042) {
return 1;
}
}
}
}
else if (TSC >= 1) {
if (Rs < 401) {
if (refDCval < 71) {
if (MVDiff < 4) {
if (Cs < 261) {
if (tcor < 99) {
if (ssDCval < 38) {
return 8;
}
else if (ssDCval >= 38) {
if (RsDiff < 2) {
if (SC < 492) {
return 1;
}
else if (SC >= 492) {
if (Rs < 329) {
if (Rs < 283) {
if (Cs < 245) {
return 4;
}
else if (Cs >= 245) {
return 1;
}
}
else if (Rs >= 283) {
return 4;
}
}
else if (Rs >= 329) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (Rs < 298) {
return 8;
}
else if (Rs >= 298) {
return 2;
}
}
}
}
else if (tcor >= 99) {
if (Cs < 191) {
return 8;
}
else if (Cs >= 191) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 2;
}
}
}
}
else if (Cs >= 261) {
if (ssDCval < 68) {
if (SC < 730) {
if (gchDC < 2) {
if (TSCindex < 2) {
if (posBalance < 41) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (posBalance >= 41) {
return 4;
}
}
else if (TSCindex >= 2) {
if (diffMVdiffVal < -2) {
if (SC < 637) {
return 1;
}
else if (SC >= 637) {
if (ssDCval < 58) {
if (Rs < 261) {
return 4;
}
else if (Rs >= 261) {
return 1;
}
}
else if (ssDCval >= 58) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= -2) {
if (mcTcor < 96) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
return 4;
}
}
else if (mcTcor >= 96) {
if (Rs < 247) {
if (negBalance < 26) {
if (mu_mv_mag_sq < 121) {
return 4;
}
else if (mu_mv_mag_sq >= 121) {
return 1;
}
}
else if (negBalance >= 26) {
return 1;
}
}
else if (Rs >= 247) {
return 1;
}
}
}
}
}
else if (gchDC >= 2) {
if (Cs < 325) {
return 4;
}
else if (Cs >= 325) {
return 2;
}
}
}
else if (SC >= 730) {
if (tcor < 97) {
if (SC < 801) {
if (mu_mv_mag_sq < 117) {
return 1;
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
else if (SC >= 801) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (Cs < 430) {
return 8;
}
else if (Cs >= 430) {
return 1;
}
}
}
}
else if (tcor >= 97) {
return 1;
}
}
}
else if (ssDCval >= 68) {
if (Rs < 345) {
if (Rs < 292) {
return 1;
}
else if (Rs >= 292) {
return 4;
}
}
else if (Rs >= 345) {
return 2;
}
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 44) {
if (negBalance < 33) {
if (posBalance < 41) {
if (RsCsDiff < 2) {
if (Rs < 303) {
return 2;
}
else if (Rs >= 303) {
return 1;
}
}
else if (RsCsDiff >= 2) {
if (Cs < 301) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (Cs >= 301) {
return 1;
}
}
}
else if (posBalance >= 41) {
return 1;
}
}
else if (negBalance >= 33) {
if (RsCsDiff < 1) {
return 8;
}
else if (RsCsDiff >= 1) {
if (posBalance < 52) {
if (Cs < 318) {
if (refDCval < 40) {
return 1;
}
else if (refDCval >= 40) {
if (tcor < 86) {
return 1;
}
else if (tcor >= 86) {
if (Cs < 160) {
return 1;
}
else if (Cs >= 160) {
return 2;
}
}
}
}
else if (Cs >= 318) {
if (Cs < 414) {
return 2;
}
else if (Cs >= 414) {
return 8;
}
}
}
else if (posBalance >= 52) {
if (TSC < 5) {
if (Rs < 385) {
return 1;
}
else if (Rs >= 385) {
if (Rs < 393) {
return 4;
}
else if (Rs >= 393) {
return 1;
}
}
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (ssDCval >= 44) {
if (Cs < 114) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
if (Rs < 268) {
if (Rs < 258) {
return 1;
}
else if (Rs >= 258) {
return 4;
}
}
else if (Rs >= 268) {
return 1;
}
}
}
else if (Cs >= 114) {
if (MVDiff < 30) {
if (Cs < 245) {
if (AFD < 15) {
if (posBalance < 45) {
if (AFD < 3) {
if (Cs < 160) {
return 2;
}
else if (Cs >= 160) {
return 4;
}
}
else if (AFD >= 3) {
if (Rs < 285) {
return 4;
}
else if (Rs >= 285) {
return 1;
}
}
}
else if (posBalance >= 45) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
if (tcor < 97) {
if (ssDCval < 48) {
if (Cs < 216) {
if (refDCval < 46) {
return 1;
}
else if (refDCval >= 46) {
if (Rs < 264) {
return 1;
}
else if (Rs >= 264) {
return 4;
}
}
}
else if (Cs >= 216) {
return 4;
}
}
else if (ssDCval >= 48) {
if (posBalance < 64) {
return 4;
}
else if (posBalance >= 64) {
return 2;
}
}
}
else if (tcor >= 97) {
if (negBalance < 32) {
if (Rs < 267) {
return 1;
}
else if (Rs >= 267) {
return 4;
}
}
else if (negBalance >= 32) {
return 1;
}
}
}
}
}
else if (AFD >= 15) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 8;
}
}
}
else if (Cs >= 245) {
if (SC < 560) {
if (AFD < 5) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
return 1;
}
}
else if (AFD >= 5) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
if (Rs < 285) {
return 4;
}
else if (Rs >= 285) {
return 1;
}
}
}
}
else if (SC >= 560) {
if (MVDiff < 17) {
if (refDCval < 66) {
if (negBalance < 22) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 4;
}
}
else if (negBalance >= 22) {
if (ssDCval < 57) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
if (tcor < 97) {
if (posBalance < 67) {
if (posBalance < 41) {
if (Rs < 360) {
return 4;
}
else if (Rs >= 360) {
return 1;
}
}
else if (posBalance >= 41) {
return 4;
}
}
else if (posBalance >= 67) {
return 2;
}
}
else if (tcor >= 97) {
return 2;
}
}
}
else if (ssDCval >= 57) {
if (ssDCval < 62) {
if (Rs < 290) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (Rs >= 290) {
if (AFD < 6) {
if (SC < 701) {
return 2;
}
else if (SC >= 701) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 80) {
return 4;
}
else if (mu_mv_mag_sq >= 80) {
if (Rs < 346) {
return 8;
}
else if (Rs >= 346) {
return 1;
}
}
}
}
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (ssDCval >= 62) {
if (posBalance < 47) {
if (Rs < 366) {
return 4;
}
else if (Rs >= 366) {
return 8;
}
}
else if (posBalance >= 47) {
if (gchDC < 1) {
if (Rs < 351) {
return 1;
}
else if (Rs >= 351) {
if (Rs < 375) {
return 2;
}
else if (Rs >= 375) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (Cs < 398) {
return 2;
}
else if (Cs >= 398) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 66) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (Cs < 311) {
return 1;
}
else if (Cs >= 311) {
return 4;
}
}
}
}
else if (MVDiff >= 17) {
if (ssDCval < 49) {
if (Rs < 298) {
return 1;
}
else if (Rs >= 298) {
return 4;
}
}
else if (ssDCval >= 49) {
if (mu_mv_mag_sq < 2) {
if (AFD < 6) {
if (Rs < 267) {
return 4;
}
else if (Rs >= 267) {
return 2;
}
}
else if (AFD >= 6) {
if (Cs < 410) {
return 4;
}
else if (Cs >= 410) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 304) {
if (Rs < 317) {
return 8;
}
else if (Rs >= 317) {
if (Rs < 336) {
return 1;
}
else if (Rs >= 336) {
return 2;
}
}
}
else if (Cs >= 304) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 30) {
if (CsDiff < 2) {
if (MVDiff < 53) {
if (Cs < 271) {
if (SC < 529) {
return 1;
}
else if (SC >= 529) {
return 4;
}
}
else if (Cs >= 271) {
if (Rs < 333) {
return 2;
}
else if (Rs >= 333) {
return 1;
}
}
}
else if (MVDiff >= 53) {
return 2;
}
}
else if (CsDiff >= 2) {
if (gchDC < 2) {
if (diffAFD < -1) {
if (Cs < 283) {
return 4;
}
else if (Cs >= 283) {
return 1;
}
}
else if (diffAFD >= -1) {
if (Rs < 343) {
return 2;
}
else if (Rs >= 343) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (CsDiff < 6) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (RsCsDiff < 7) {
return 2;
}
else if (RsCsDiff >= 7) {
return 1;
}
}
}
else if (CsDiff >= 6) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 71) {
if (MVDiff < 1) {
if (Cs < 473) {
return 2;
}
else if (Cs >= 473) {
return 1;
}
}
else if (MVDiff >= 1) {
if (Rs < 359) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (Rs >= 359) {
return 8;
}
}
}
}
else if (Rs >= 401) {
if (ssDCval < 71) {
if (CsDiff < 1) {
if (ssDCval < 68) {
if (MVDiff < 10) {
if (negBalance < 24) {
if (Rs < 831) {
if (ssDCval < 53) {
if (tcor < 100) {
if (Cs < 317) {
if (AFD < 3) {
if (Rs < 501) {
return 4;
}
else if (Rs >= 501) {
return 8;
}
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 317) {
if (Rs < 425) {
return 8;
}
else if (Rs >= 425) {
return 1;
}
}
}
else if (tcor >= 100) {
return 8;
}
}
else if (ssDCval >= 53) {
if (Cs < 199) {
if (Rs < 753) {
return 2;
}
else if (Rs >= 753) {
return 4;
}
}
else if (Cs >= 199) {
if (ssDCval < 63) {
if (Cs < 389) {
if (mu_mv_mag_sq < 91) {
if (Rs < 703) {
return 8;
}
else if (Rs >= 703) {
return 1;
}
}
else if (mu_mv_mag_sq >= 91) {
return 1;
}
}
else if (Cs >= 389) {
if (mu_mv_mag_sq < 106) {
if (Cs < 452) {
return 8;
}
else if (Cs >= 452) {
if (Rs < 438) {
return 8;
}
else if (Rs >= 438) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Cs < 431) {
return 1;
}
else if (Cs >= 431) {
if (Rs < 689) {
return 8;
}
else if (Rs >= 689) {
return 1;
}
}
}
}
}
else if (ssDCval >= 63) {
if (Rs < 476) {
if (Rs < 441) {
return 8;
}
else if (Rs >= 441) {
return 1;
}
}
else if (Rs >= 476) {
return 8;
}
}
}
}
}
else if (Rs >= 831) {
if (posBalance < 44) {
if (Cs < 522) {
return 1;
}
else if (Cs >= 522) {
return 8;
}
}
else if (posBalance >= 44) {
if (Cs < 282) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (Cs >= 282) {
return 2;
}
}
}
}
else if (negBalance >= 24) {
if (mcTcor < 98) {
if (MVDiff < 7) {
if (Cs < 269) {
if (mu_mv_mag_sq < 64) {
if (Rs < 411) {
return 2;
}
else if (Rs >= 411) {
return 1;
}
}
else if (mu_mv_mag_sq >= 64) {
if (negBalance < 28) {
return 1;
}
else if (negBalance >= 28) {
return 2;
}
}
}
else if (Cs >= 269) {
if (Rs < 640) {
return 1;
}
else if (Rs >= 640) {
if (Cs < 346) {
return 1;
}
else if (Cs >= 346) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
if (RsDiff < 2) {
if (Rs < 491) {
return 1;
}
else if (Rs >= 491) {
return 8;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (tcor < 101) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (AFD < 3) {
if (Rs < 468) {
if (Rs < 458) {
return 2;
}
else if (Rs >= 458) {
return 8;
}
}
else if (Rs >= 468) {
if (Cs < 237) {
return 1;
}
else if (Cs >= 237) {
return 4;
}
}
}
else if (AFD >= 3) {
if (Rs < 402) {
return 8;
}
else if (Rs >= 402) {
return 1;
}
}
}
}
else if (tcor >= 101) {
return 8;
}
}
}
}
else if (MVDiff >= 10) {
if (AFD < 4) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Cs < 368) {
return 1;
}
else if (Cs >= 368) {
return 4;
}
}
}
}
else if (AFD >= 4) {
if (mu_mv_mag_sq < 52) {
if (Cs < 178) {
return 1;
}
else if (Cs >= 178) {
if (Cs < 217) {
return 4;
}
else if (Cs >= 217) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (negBalance < 30) {
if (SC < 1039) {
return 1;
}
else if (SC >= 1039) {
return 4;
}
}
else if (negBalance >= 30) {
if (Cs < 85) {
return 8;
}
else if (Cs >= 85) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 68) {
if (tcor < 99) {
if (ssDCval < 69) {
if (Rs < 481) {
if (Rs < 443) {
return 2;
}
else if (Rs >= 443) {
return 8;
}
}
else if (Rs >= 481) {
return 1;
}
}
else if (ssDCval >= 69) {
if (negBalance < 27) {
if (Rs < 689) {
if (Rs < 678) {
return 1;
}
else if (Rs >= 678) {
return 2;
}
}
else if (Rs >= 689) {
if (Rs < 844) {
return 4;
}
else if (Rs >= 844) {
return 8;
}
}
}
else if (negBalance >= 27) {
if (SC < 651) {
return 4;
}
else if (SC >= 651) {
return 1;
}
}
}
}
else if (tcor >= 99) {
if (posBalance < 42) {
if (Cs < 420) {
if (mu_mv_mag_sq < 120) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 587) {
return 2;
}
else if (Rs >= 587) {
if (Rs < 774) {
return 1;
}
else if (Rs >= 774) {
return 8;
}
}
}
}
else if (Cs >= 420) {
if (Rs < 813) {
return 1;
}
else if (Rs >= 813) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (MVDiff < 4) {
if (negBalance < 18) {
return 1;
}
else if (negBalance >= 18) {
return 2;
}
}
else if (MVDiff >= 4) {
if (Rs < 638) {
return 8;
}
else if (Rs >= 638) {
return 4;
}
}
}
}
}
}
else if (CsDiff >= 1) {
if (Cs < 480) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 2) {
if (TSC < 3) {
if (Cs < 355) {
return 1;
}
else if (Cs >= 355) {
return 8;
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (MVDiff >= 2) {
if (RsCsDiff < 7) {
if (TSC < 3) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (diffRsCsdiff < 2) {
if (negBalance < 30) {
return 2;
}
else if (negBalance >= 30) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
}
else if (diffRsCsdiff >= 2) {
return 4;
}
}
}
else if (TSC >= 3) {
if (refDCval < 55) {
if (gchDC < 1) {
if (Cs < 278) {
return 8;
}
else if (Cs >= 278) {
return 1;
}
}
else if (gchDC >= 1) {
if (Cs < 331) {
if (Rs < 518) {
return 1;
}
else if (Rs >= 518) {
return 8;
}
}
else if (Cs >= 331) {
return 4;
}
}
}
else if (refDCval >= 55) {
if (posBalance < 66) {
if (Rs < 714) {
return 4;
}
else if (Rs >= 714) {
if (Rs < 747) {
return 2;
}
else if (Rs >= 747) {
return 4;
}
}
}
else if (posBalance >= 66) {
return 2;
}
}
}
}
else if (RsCsDiff >= 7) {
if (refDCval < 53) {
if (gchDC < 1) {
if (posBalance < 56) {
if (Cs < 460) {
return 1;
}
else if (Cs >= 460) {
return 8;
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (gchDC >= 1) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (refDCval >= 53) {
if (Cs < 298) {
if (Rs < 881) {
return 1;
}
else if (Rs >= 881) {
if (Rs < 1027) {
return 2;
}
else if (Rs >= 1027) {
return 1;
}
}
}
else if (Cs >= 298) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
if (Rs < 494) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
else if (Rs >= 494) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
if (Rs < 712) {
return 4;
}
else if (Rs >= 712) {
return 1;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 99) {
if (mcTcor < 97) {
if (refDCval < 58) {
if (mu_mv_mag_sq < 110) {
if (posBalance < 34) {
if (AFD < 4) {
if (Rs < 801) {
return 4;
}
else if (Rs >= 801) {
return 1;
}
}
else if (AFD >= 4) {
return 2;
}
}
else if (posBalance >= 34) {
if (negBalance < 35) {
if (refDCval < 47) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Rs < 846) {
if (negBalance < 27) {
return 1;
}
else if (negBalance >= 27) {
return 8;
}
}
else if (Rs >= 846) {
return 2;
}
}
}
else if (refDCval >= 47) {
if (refDCval < 56) {
if (Rs < 550) {
return 4;
}
else if (Rs >= 550) {
if (Cs < 448) {
if (diffTSC < 1) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
if (Rs < 606) {
return 1;
}
else if (Rs >= 606) {
return 2;
}
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (Cs >= 448) {
return 4;
}
}
}
else if (refDCval >= 56) {
if (tcor < 81) {
return 4;
}
else if (tcor >= 81) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
return 2;
}
}
}
}
}
else if (negBalance >= 35) {
if (RsCsDiff < 7) {
if (TSC < 4) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
if (Rs < 578) {
return 4;
}
else if (Rs >= 578) {
return 1;
}
}
}
}
else if (TSC >= 4) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 4;
}
}
}
else if (RsCsDiff >= 7) {
if (tcor < 70) {
if (Rs < 645) {
return 2;
}
else if (Rs >= 645) {
return 8;
}
}
else if (tcor >= 70) {
if (Cs < 380) {
if (mcTcor < 92) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
return 1;
}
}
else if (mcTcor >= 92) {
return 1;
}
}
else if (Cs >= 380) {
if (mu_mv_mag_sq < 10) {
return 1;
}
else if (mu_mv_mag_sq >= 10) {
if (Rs < 432) {
if (Rs < 413) {
return 4;
}
else if (Rs >= 413) {
return 1;
}
}
else if (Rs >= 432) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 459) {
return 4;
}
else if (Rs >= 459) {
if (Rs < 648) {
return 8;
}
else if (Rs >= 648) {
return 2;
}
}
}
}
else if (refDCval >= 58) {
if (mu_mv_mag_sq < 67) {
if (tcor < 93) {
if (diffRsCsdiff < -10) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 1;
}
}
else if (diffRsCsdiff >= -10) {
if (mcTcor < 96) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (Cs < 401) {
if (negBalance < 39) {
if (Rs < 463) {
return 2;
}
else if (Rs >= 463) {
return 1;
}
}
else if (negBalance >= 39) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 1;
}
}
}
else if (Cs >= 401) {
if (Rs < 567) {
return 2;
}
else if (Rs >= 567) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
}
else if (tcor >= 93) {
return 1;
}
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 544) {
if (AFD < 9) {
if (Rs < 433) {
return 1;
}
else if (Rs >= 433) {
return 2;
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (Rs >= 544) {
if (Rs < 628) {
return 8;
}
else if (Rs >= 628) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 40) {
if (Rs < 730) {
if (SC < 876) {
if (posBalance < 31) {
if (Rs < 495) {
return 1;
}
else if (Rs >= 495) {
return 2;
}
}
else if (posBalance >= 31) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
return 4;
}
}
}
else if (SC >= 876) {
if (diffRsCsdiff < 1) {
if (ssDCval < 53) {
if (Rs < 639) {
return 4;
}
else if (Rs >= 639) {
if (Rs < 714) {
return 2;
}
else if (Rs >= 714) {
return 1;
}
}
}
else if (ssDCval >= 53) {
if (Rs < 583) {
if (SC < 897) {
return 1;
}
else if (SC >= 897) {
if (Rs < 554) {
return 8;
}
else if (Rs >= 554) {
return 1;
}
}
}
else if (Rs >= 583) {
if (diffMVdiffVal < 1) {
if (Cs < 465) {
return 4;
}
else if (Cs >= 465) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 1) {
if (negBalance < 34) {
if (Rs < 692) {
return 1;
}
else if (Rs >= 692) {
return 4;
}
}
else if (negBalance >= 34) {
if (Rs < 474) {
return 2;
}
else if (Rs >= 474) {
return 1;
}
}
}
}
}
else if (Rs >= 730) {
return 1;
}
}
else if (negBalance >= 40) {
if (refDCval < 58) {
if (mu_mv_mag_sq < 28) {
if (Rs < 430) {
return 8;
}
else if (Rs >= 430) {
return 1;
}
}
else if (mu_mv_mag_sq >= 28) {
return 8;
}
}
else if (refDCval >= 58) {
if (diffAFD < -1) {
if (Rs < 741) {
return 8;
}
else if (Rs >= 741) {
return 1;
}
}
else if (diffAFD >= -1) {
if (diffMVdiffVal < 3) {
if (ssDCval < 70) {
return 4;
}
else if (ssDCval >= 70) {
return 1;
}
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (RsDiff < 2) {
if (Rs < 567) {
if (Rs < 515) {
if (ssDCval < 61) {
if (Cs < 400) {
return 1;
}
else if (Cs >= 400) {
return 8;
}
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (Rs >= 515) {
if (posBalance < 35) {
return 4;
}
else if (posBalance >= 35) {
if (Cs < 379) {
return 4;
}
else if (Cs >= 379) {
return 8;
}
}
}
}
else if (Rs >= 567) {
if (mu_mv_mag_sq < 89) {
return 1;
}
else if (mu_mv_mag_sq >= 89) {
if (mu_mv_mag_sq < 97) {
return 8;
}
else if (mu_mv_mag_sq >= 97) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
}
else if (RsDiff >= 2) {
if (Cs < 334) {
if (Rs < 523) {
return 2;
}
else if (Rs >= 523) {
return 1;
}
}
else if (Cs >= 334) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
return 4;
}
}
}
}
}
}
else if (Cs >= 480) {
if (posBalance < 39) {
if (Rs < 1013) {
if (mu_mv_mag_sq < 70) {
if (Rs < 690) {
return 1;
}
else if (Rs >= 690) {
return 2;
}
}
else if (mu_mv_mag_sq >= 70) {
if (Cs < 504) {
return 2;
}
else if (Cs >= 504) {
return 4;
}
}
}
else if (Rs >= 1013) {
return 1;
}
}
else if (posBalance >= 39) {
if (ssDCval < 65) {
if (SC < 995) {
if (ssDCval < 58) {
if (Cs < 510) {
return 1;
}
else if (Cs >= 510) {
if (Rs < 460) {
return 8;
}
else if (Rs >= 460) {
return 1;
}
}
}
else if (ssDCval >= 58) {
if (AFD < 6) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (SC >= 995) {
if (diffRsCsdiff < -6) {
if (AFD < 8) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (Cs < 533) {
return 2;
}
else if (Cs >= 533) {
return 4;
}
}
}
else if (AFD >= 8) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
else if (diffRsCsdiff >= -6) {
if (negBalance < 29) {
if (ssDCval < 57) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 1;
}
}
else if (ssDCval >= 57) {
if (Cs < 527) {
return 8;
}
else if (Cs >= 527) {
return 4;
}
}
}
else if (negBalance >= 29) {
if (mu_mv_mag_sq < 68) {
if (tcor < 97) {
if (ssDCval < 57) {
if (MVDiff < 16) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Rs < 896) {
return 4;
}
else if (Rs >= 896) {
return 8;
}
}
}
else if (MVDiff >= 16) {
if (AFD < 11) {
if (Cs < 517) {
return 8;
}
else if (Cs >= 517) {
return 1;
}
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (ssDCval >= 57) {
if (CsDiff < 2) {
if (Rs < 641) {
return 4;
}
else if (Rs >= 641) {
return 8;
}
}
else if (CsDiff >= 2) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 4;
}
}
}
}
else if (tcor >= 97) {
return 4;
}
}
else if (mu_mv_mag_sq >= 68) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 65) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (mu_mv_mag_sq < 119) {
if (tcor < 97) {
if (Cs < 432) {
if (SC < 804) {
if (mu_mv_mag_sq < 15) {
return 4;
}
else if (mu_mv_mag_sq >= 15) {
if (Cs < 174) {
return 1;
}
else if (Cs >= 174) {
if (SC < 789) {
return 8;
}
else if (SC >= 789) {
return 1;
}
}
}
}
else if (SC >= 804) {
return 1;
}
}
else if (Cs >= 432) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
if (diffAFD < -2) {
if (Rs < 694) {
return 1;
}
else if (Rs >= 694) {
return 4;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
}
else if (tcor >= 97) {
if (mu_mv_mag_sq < 32) {
if (Rs < 690) {
return 1;
}
else if (Rs >= 690) {
return 4;
}
}
else if (mu_mv_mag_sq >= 32) {
if (posBalance < 29) {
return 4;
}
else if (posBalance >= 29) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (ssDCval < 73) {
if (Rs < 598) {
return 1;
}
else if (Rs >= 598) {
return 4;
}
}
else if (ssDCval >= 73) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 1273) {
if (SC < 1625) {
if (mu_mv_mag_sq < 58) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (mu_mv_mag_sq >= 58) {
if (Cs < 225) {
return 8;
}
else if (Cs >= 225) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
}
}
else if (SC >= 1625) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1703) {
if (mu_mv_mag_sq < 70) {
if (negBalance < 24) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (Cs < 184) {
return 2;
}
else if (Cs >= 184) {
return 1;
}
}
}
}
else if (negBalance >= 24) {
if (TSC < 5) {
if (posBalance < 31) {
if (Rs < 1460) {
return 2;
}
else if (Rs >= 1460) {
return 1;
}
}
else if (posBalance >= 31) {
return 1;
}
}
else if (TSC >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (TSCindex < 2) {
if (ssDCval < 64) {
if (ssDCval < 60) {
if (SC < 1818) {
if (Rs < 1329) {
return 1;
}
else if (Rs >= 1329) {
return 8;
}
}
else if (SC >= 1818) {
return 1;
}
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (TSCindex >= 2) {
if (posBalance < 68) {
if (mu_mv_mag_sq < 118) {
if (ssDCval < 70) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
if (Cs < 426) {
return 1;
}
else if (Cs >= 426) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (negBalance < 18) {
if (Rs < 1392) {
return 1;
}
else if (Rs >= 1392) {
return 2;
}
}
else if (negBalance >= 18) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
if (Rs < 1348) {
return 8;
}
else if (Rs >= 1348) {
return 1;
}
}
}
}
}
else if (posBalance >= 68) {
return 2;
}
}
}
}
else if (Rs >= 1703) {
if (Cs < 421) {
if (RsCsDiff < 10) {
return 2;
}
else if (RsCsDiff >= 10) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (Cs >= 421) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 62) {
if (Cs < 517) {
return 8;
}
else if (Cs >= 517) {
return 1;
}
}
else if (ssDCval >= 62) {
if (Rs < 1407) {
return 8;
}
else if (Rs >= 1407) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 74) {
if (Cs < 140) {
if (refDCval < 106) {
if (posBalance < 10) {
return 8;
}
else if (posBalance >= 10) {
if (Rs < 1145) {
if (ssDCval < 98) {
if (mu_mv_mag_sq < 37) {
if (TSC < 4) {
if (Cs < 9) {
return 1;
}
else if (Cs >= 9) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (Rs < 356) {
if (Rs < 297) {
return 1;
}
else if (Rs >= 297) {
return 8;
}
}
else if (Rs >= 356) {
if (Rs < 419) {
return 2;
}
else if (Rs >= 419) {
return 4;
}
}
}
}
}
else if (TSC >= 4) {
if (RsDiff < 2) {
if (Cs < 88) {
return 1;
}
else if (Cs >= 88) {
return 8;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (diffRsCsdiff < -2) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 2;
}
}
else if (diffRsCsdiff >= -2) {
if (Cs < 68) {
if (TSC < 4) {
if (negBalance < 53) {
return 8;
}
else if (negBalance >= 53) {
if (MVDiff < 29) {
return 8;
}
else if (MVDiff >= 29) {
return 2;
}
}
}
else if (TSC >= 4) {
return 1;
}
}
else if (Cs >= 68) {
return 8;
}
}
}
}
else if (ssDCval >= 98) {
return 1;
}
}
else if (Rs >= 1145) {
return 1;
}
}
}
else if (refDCval >= 106) {
if (MVDiff < 20) {
if (ssDCval < 130) {
return 8;
}
else if (ssDCval >= 130) {
if (Rs < 808) {
return 2;
}
else if (Rs >= 808) {
return 4;
}
}
}
else if (MVDiff >= 20) {
if (diffMVdiffVal < 42) {
if (mu_mv_mag_sq < 9) {
if (Rs < 472) {
return 8;
}
else if (Rs >= 472) {
return 1;
}
}
else if (mu_mv_mag_sq >= 9) {
return 1;
}
}
else if (diffMVdiffVal >= 42) {
if (Cs < 32) {
return 8;
}
else if (Cs >= 32) {
return 2;
}
}
}
}
}
else if (Cs >= 140) {
if (SC < 1889) {
if (SC < 884) {
if (mcTcor < 97) {
if (MVDiff < 22) {
if (TSC < 6) {
if (AFD < 9) {
if (Cs < 186) {
if (MVDiff < 12) {
if (Rs < 268) {
if (Cs < 177) {
return 2;
}
else if (Cs >= 177) {
return 8;
}
}
else if (Rs >= 268) {
return 4;
}
}
else if (MVDiff >= 12) {
return 1;
}
}
else if (Cs >= 186) {
if (diffAFD < -2) {
if (Cs < 474) {
return 1;
}
else if (Cs >= 474) {
return 8;
}
}
else if (diffAFD >= -2) {
if (Rs < 516) {
if (MVDiff < 5) {
if (AFD < 7) {
if (Cs < 453) {
return 8;
}
else if (Cs >= 453) {
if (Rs < 330) {
return 2;
}
else if (Rs >= 330) {
return 8;
}
}
}
else if (AFD >= 7) {
return 1;
}
}
else if (MVDiff >= 5) {
if (ssDCval < 133) {
if (ssDCval < 95) {
if (Cs < 368) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (Cs < 325) {
return 2;
}
else if (Cs >= 325) {
if (Cs < 342) {
return 8;
}
else if (Cs >= 342) {
return 2;
}
}
}
}
else if (Cs >= 368) {
return 8;
}
}
else if (ssDCval >= 95) {
return 2;
}
}
else if (ssDCval >= 133) {
if (Cs < 254) {
return 2;
}
else if (Cs >= 254) {
return 8;
}
}
}
}
else if (Rs >= 516) {
if (Rs < 562) {
return 1;
}
else if (Rs >= 562) {
return 4;
}
}
}
}
}
else if (AFD >= 9) {
if (posBalance < 52) {
if (Rs < 294) {
return 1;
}
else if (Rs >= 294) {
return 2;
}
}
else if (posBalance >= 52) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 8;
}
}
}
}
}
else if (TSC >= 6) {
if (Rs < 481) {
if (ssDCval < 129) {
return 8;
}
else if (ssDCval >= 129) {
return 2;
}
}
else if (Rs >= 481) {
if (Rs < 511) {
return 4;
}
else if (Rs >= 511) {
return 2;
}
}
}
}
else if (MVDiff >= 22) {
if (Rs < 474) {
if (tcor < 83) {
if (AFD < 16) {
if (tcor < 79) {
if (MVDiff < 30) {
if (Cs < 430) {
return 4;
}
else if (Cs >= 430) {
return 8;
}
}
else if (MVDiff >= 30) {
if (Cs < 224) {
return 8;
}
else if (Cs >= 224) {
if (ssDCval < 83) {
if (Rs < 327) {
return 4;
}
else if (Rs >= 327) {
return 8;
}
}
else if (ssDCval >= 83) {
return 2;
}
}
}
}
else if (tcor >= 79) {
if (Rs < 392) {
return 1;
}
else if (Rs >= 392) {
return 8;
}
}
}
else if (AFD >= 16) {
if (Rs < 316) {
if (MVDiff < 174) {
if (diffMVdiffVal < 10) {
if (Cs < 446) {
return 2;
}
else if (Cs >= 446) {
return 1;
}
}
else if (diffMVdiffVal >= 10) {
if (mu_mv_mag_sq < 28) {
return 1;
}
else if (mu_mv_mag_sq >= 28) {
return 8;
}
}
}
else if (MVDiff >= 174) {
if (Cs < 260) {
return 1;
}
else if (Cs >= 260) {
return 4;
}
}
}
else if (Rs >= 316) {
if (SC < 869) {
if (Cs < 515) {
return 2;
}
else if (Cs >= 515) {
return 8;
}
}
else if (SC >= 869) {
return 1;
}
}
}
}
else if (tcor >= 83) {
if (posBalance < 36) {
if (Rs < 431) {
return 8;
}
else if (Rs >= 431) {
return 4;
}
}
else if (posBalance >= 36) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (diffRsCsdiff < -6) {
if (Rs < 393) {
return 8;
}
else if (Rs >= 393) {
return 2;
}
}
else if (diffRsCsdiff >= -6) {
return 2;
}
}
}
}
}
else if (Rs >= 474) {
if (TSC < 5) {
if (Rs < 504) {
return 2;
}
else if (Rs >= 504) {
return 8;
}
}
else if (TSC >= 5) {
if (Cs < 231) {
return 1;
}
else if (Cs >= 231) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 32) {
if (ssDCval < 150) {
if (Cs < 448) {
return 8;
}
else if (Cs >= 448) {
if (Rs < 376) {
return 4;
}
else if (Rs >= 376) {
return 1;
}
}
}
else if (ssDCval >= 150) {
return 4;
}
}
else if (posBalance >= 32) {
if (mu_mv_mag_sq < 109) {
if (refDCval < 117) {
if (ssDCval < 84) {
if (refDCval < 75) {
if (MVDiff < 30) {
return 1;
}
else if (MVDiff >= 30) {
return 8;
}
}
else if (refDCval >= 75) {
if (Rs < 344) {
if (Cs < 189) {
if (Rs < 257) {
return 1;
}
else if (Rs >= 257) {
return 8;
}
}
else if (Cs >= 189) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 295) {
return 1;
}
else if (Rs >= 295) {
return 2;
}
}
}
}
else if (Rs >= 344) {
if (Cs < 234) {
if (Rs < 464) {
return 8;
}
else if (Rs >= 464) {
if (mu_mv_mag_sq < 34) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (mu_mv_mag_sq >= 34) {
if (MVDiff < 29) {
return 2;
}
else if (MVDiff >= 29) {
return 8;
}
}
}
}
else if (Cs >= 234) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (mu_mv_mag_sq < 65) {
if (Rs < 493) {
return 4;
}
else if (Rs >= 493) {
return 1;
}
}
else if (mu_mv_mag_sq >= 65) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (gchDC < 2) {
if (MVDiff < 6) {
if (negBalance < 34) {
if (Cs < 343) {
if (diffMVdiffVal < 4) {
return 4;
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
else if (Cs >= 343) {
if (Rs < 258) {
return 4;
}
else if (Rs >= 258) {
return 1;
}
}
}
else if (negBalance >= 34) {
return 1;
}
}
else if (MVDiff >= 6) {
if (TSC < 7) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (Rs < 445) {
if (ssDCval < 92) {
if (posBalance < 45) {
if (Rs < 396) {
return 8;
}
else if (Rs >= 396) {
return 2;
}
}
else if (posBalance >= 45) {
return 1;
}
}
else if (ssDCval >= 92) {
if (Rs < 434) {
if (ssDCval < 103) {
return 2;
}
else if (ssDCval >= 103) {
if (mu_mv_mag_sq < 11) {
return 4;
}
else if (mu_mv_mag_sq >= 11) {
return 2;
}
}
}
else if (Rs >= 434) {
return 1;
}
}
}
else if (Rs >= 445) {
return 8;
}
}
}
}
else if (refDCval >= 117) {
if (mcTcor < 99) {
if (tcor < 97) {
if (MVDiff < 13) {
if (ssDCval < 134) {
if (Cs < 315) {
return 8;
}
else if (Cs >= 315) {
return 2;
}
}
else if (ssDCval >= 134) {
if (Cs < 391) {
return 1;
}
else if (Cs >= 391) {
return 8;
}
}
}
else if (MVDiff >= 13) {
if (MVDiff < 155) {
return 2;
}
else if (MVDiff >= 155) {
return 4;
}
}
}
else if (tcor >= 97) {
if (negBalance < 48) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
return 2;
}
}
else if (negBalance >= 48) {
return 2;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 370) {
return 2;
}
else if (Rs >= 370) {
if (TSC < 4) {
if (Rs < 485) {
return 1;
}
else if (Rs >= 485) {
return 4;
}
}
else if (TSC >= 4) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (Cs < 237) {
if (ssDCval < 94) {
if (SC < 512) {
if (Rs < 271) {
return 2;
}
else if (Rs >= 271) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
return 2;
}
}
}
else if (SC >= 512) {
return 2;
}
}
else if (ssDCval >= 94) {
return 1;
}
}
else if (Cs >= 237) {
if (Rs < 436) {
if (Rs < 292) {
return 1;
}
else if (Rs >= 292) {
return 8;
}
}
else if (Rs >= 436) {
return 1;
}
}
}
}
}
}
else if (SC >= 884) {
if (CsDiff < 2) {
if (refDCval < 94) {
if (ssDCval < 81) {
if (negBalance < 32) {
if (SC < 1234) {
if (refDCval < 77) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (Cs < 312) {
return 4;
}
else if (Cs >= 312) {
return 8;
}
}
}
else if (refDCval >= 77) {
if (Rs < 730) {
if (Cs < 332) {
return 8;
}
else if (Cs >= 332) {
return 1;
}
}
else if (Rs >= 730) {
return 8;
}
}
}
else if (SC >= 1234) {
if (Cs < 340) {
return 1;
}
else if (Cs >= 340) {
return 8;
}
}
}
else if (negBalance >= 32) {
if (tcor < 97) {
if (posBalance < 47) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (posBalance >= 47) {
if (diffAFD < -1) {
if (Rs < 860) {
return 4;
}
else if (Rs >= 860) {
return 1;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (tcor >= 97) {
if (mu_mv_mag_sq < 30) {
if (Rs < 737) {
return 1;
}
else if (Rs >= 737) {
return 4;
}
}
else if (mu_mv_mag_sq >= 30) {
if (Rs < 1591) {
if (MVDiff < 1) {
if (Rs < 646) {
return 8;
}
else if (Rs >= 646) {
return 1;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (Rs >= 1591) {
return 1;
}
}
}
}
}
else if (ssDCval >= 81) {
if (SC < 1234) {
if (gchDC < 2) {
if (tcor < 88) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
if (Rs < 698) {
return 2;
}
else if (Rs >= 698) {
return 4;
}
}
}
else if (tcor >= 88) {
if (posBalance < 23) {
return 8;
}
else if (posBalance >= 23) {
if (MVDiff < 4) {
if (SC < 1219) {
return 1;
}
else if (SC >= 1219) {
if (Rs < 878) {
return 8;
}
else if (Rs >= 878) {
return 1;
}
}
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 53) {
return 1;
}
else if (mu_mv_mag_sq >= 53) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
return 1;
}
}
}
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (SC >= 1234) {
if (AFD < 2) {
if (Rs < 1044) {
if (negBalance < 18) {
return 2;
}
else if (negBalance >= 18) {
return 1;
}
}
else if (Rs >= 1044) {
return 2;
}
}
else if (AFD >= 2) {
if (Cs < 424) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Rs < 1605) {
return 8;
}
else if (Rs >= 1605) {
return 1;
}
}
}
else if (Cs >= 424) {
return 8;
}
}
}
}
}
else if (refDCval >= 94) {
if (ssDCval < 115) {
if (ssDCval < 98) {
if (Rs < 583) {
return 8;
}
else if (Rs >= 583) {
if (MVDiff < 4) {
if (SC < 1108) {
return 4;
}
else if (SC >= 1108) {
return 8;
}
}
else if (MVDiff >= 4) {
if (Rs < 851) {
if (Cs < 296) {
return 1;
}
else if (Cs >= 296) {
return 8;
}
}
else if (Rs >= 851) {
return 1;
}
}
}
}
else if (ssDCval >= 98) {
if (diffRsCsdiff < -6) {
if (Rs < 869) {
return 1;
}
else if (Rs >= 869) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
if (Rs < 482) {
return 1;
}
else if (Rs >= 482) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 818) {
return 1;
}
else if (Rs >= 818) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 115) {
if (refDCval < 118) {
if (Rs < 628) {
return 2;
}
else if (Rs >= 628) {
return 8;
}
}
else if (refDCval >= 118) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (AFD < 3) {
if (Rs < 465) {
return 4;
}
else if (Rs >= 465) {
return 1;
}
}
else if (AFD >= 3) {
if (mu_mv_mag_sq < 32) {
if (Cs < 432) {
if (Rs < 1045) {
return 8;
}
else if (Rs >= 1045) {
return 4;
}
}
else if (Cs >= 432) {
return 1;
}
}
else if (mu_mv_mag_sq >= 32) {
if (ssDCval < 158) {
if (diffMVdiffVal < -8) {
return 4;
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
else if (ssDCval >= 158) {
return 4;
}
}
}
}
}
}
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 63) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 7) {
if (tcor < 92) {
if (posBalance < 59) {
if (diffAFD < -2) {
if (Rs < 683) {
if (Rs < 484) {
return 2;
}
else if (Rs >= 484) {
return 8;
}
}
else if (Rs >= 683) {
return 4;
}
}
else if (diffAFD >= -2) {
if (AFD < 11) {
if (ssDCval < 85) {
if (SC < 1055) {
return 1;
}
else if (SC >= 1055) {
return 2;
}
}
else if (ssDCval >= 85) {
return 2;
}
}
else if (AFD >= 11) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 993) {
return 2;
}
else if (Rs >= 993) {
return 1;
}
}
}
}
}
else if (posBalance >= 59) {
if (ssDCval < 91) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
return 2;
}
}
else if (ssDCval >= 91) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
}
}
}
else if (tcor >= 92) {
if (MVDiff < 18) {
if (negBalance < 40) {
if (Rs < 752) {
return 1;
}
else if (Rs >= 752) {
return 4;
}
}
else if (negBalance >= 40) {
return 8;
}
}
else if (MVDiff >= 18) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (TSC < 6) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 419) {
return 8;
}
else if (Rs >= 419) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
return 1;
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 81) {
if (Rs < 459) {
return 2;
}
else if (Rs >= 459) {
return 1;
}
}
else if (ssDCval >= 81) {
if (diffRsCsdiff < -6) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
}
}
}
else if (mcTcor >= 98) {
if (Cs < 498) {
if (diffMVdiffVal < 5) {
if (Rs < 1459) {
return 1;
}
else if (Rs >= 1459) {
return 8;
}
}
else if (diffMVdiffVal >= 5) {
if (Rs < 487) {
return 2;
}
else if (Rs >= 487) {
if (Rs < 650) {
return 1;
}
else if (Rs >= 650) {
return 4;
}
}
}
}
else if (Cs >= 498) {
if (diffMVdiffVal < -2) {
return 2;
}
else if (diffMVdiffVal >= -2) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (Rs < 1361) {
return 8;
}
else if (Rs >= 1361) {
return 1;
}
}
}
}
}
else if (SC >= 1889) {
if (SC < 2542) {
if (refDCval < 96) {
if (Cs < 486) {
if (Cs < 319) {
return 1;
}
else if (Cs >= 319) {
if (mu_mv_mag_sq < 84) {
if (AFD < 5) {
if (Rs < 2071) {
return 8;
}
else if (Rs >= 2071) {
return 1;
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 84) {
if (Rs < 2114) {
return 1;
}
else if (Rs >= 2114) {
return 2;
}
}
}
}
else if (Cs >= 486) {
if (mu_mv_mag_sq < 96) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 1668) {
return 1;
}
else if (Rs >= 1668) {
return 4;
}
}
}
}
else if (refDCval >= 96) {
if (diffRsCsdiff < -6) {
if (Rs < 1764) {
return 2;
}
else if (Rs >= 1764) {
return 4;
}
}
else if (diffRsCsdiff >= -6) {
if (ssDCval < 111) {
if (Cs < 235) {
return 1;
}
else if (Cs >= 235) {
return 8;
}
}
else if (ssDCval >= 111) {
if (Rs < 1779) {
return 4;
}
else if (Rs >= 1779) {
return 1;
}
}
}
}
}
else if (SC >= 2542) {
if (posBalance < 54) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (posBalance >= 54) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (Rs < 1419) {
if (Cs < 403) {
if (tcor < 70) {
if (posBalance < 55) {
if (Cs < 256) {
if (negBalance < 58) {
if (diffRsCsdiff < -32) {
if (Rs < 719) {
return 2;
}
else if (Rs >= 719) {
return 1;
}
}
else if (diffRsCsdiff >= -32) {
return 1;
}
}
else if (negBalance >= 58) {
if (TSC < 6) {
if (Rs < 697) {
return 8;
}
else if (Rs >= 697) {
return 2;
}
}
else if (TSC >= 6) {
if (tcor < 65) {
if (Cs < 197) {
return 1;
}
else if (Cs >= 197) {
if (Rs < 1062) {
return 2;
}
else if (Rs >= 1062) {
return 1;
}
}
}
else if (tcor >= 65) {
if (MVDiff < 141) {
return 2;
}
else if (MVDiff >= 141) {
return 1;
}
}
}
}
}
else if (Cs >= 256) {
if (Rs < 890) {
if (tcor < 38) {
if (Rs < 358) {
return 4;
}
else if (Rs >= 358) {
return 1;
}
}
else if (tcor >= 38) {
if (TSC < 5) {
if (Rs < 399) {
return 8;
}
else if (Rs >= 399) {
if (Rs < 483) {
return 1;
}
else if (Rs >= 483) {
return 2;
}
}
}
else if (TSC >= 5) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 1;
}
}
}
}
else if (Rs >= 890) {
if (AFD < 27) {
return 1;
}
else if (AFD >= 27) {
if (Rs < 1219) {
return 8;
}
else if (Rs >= 1219) {
if (Rs < 1299) {
return 1;
}
else if (Rs >= 1299) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 55) {
if (mcTcor < 77) {
if (diffAFD < 2) {
if (RsCsDiff < 19) {
if (Rs < 377) {
return 1;
}
else if (Rs >= 377) {
return 4;
}
}
else if (RsCsDiff >= 19) {
if (tcor < 43) {
if (Cs < 376) {
return 4;
}
else if (Cs >= 376) {
return 8;
}
}
else if (tcor >= 43) {
if (Rs < 751) {
if (Rs < 554) {
return 4;
}
else if (Rs >= 554) {
return 1;
}
}
else if (Rs >= 751) {
return 2;
}
}
}
}
else if (diffAFD >= 2) {
if (posBalance < 65) {
if (negBalance < 55) {
return 1;
}
else if (negBalance >= 55) {
return 2;
}
}
else if (posBalance >= 65) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
if (Rs < 662) {
return 2;
}
else if (Rs >= 662) {
return 1;
}
}
}
}
}
else if (mcTcor >= 77) {
if (RsCsDiff < 37) {
if (gchDC < 3) {
if (posBalance < 56) {
if (Cs < 353) {
return 2;
}
else if (Cs >= 353) {
return 4;
}
}
else if (posBalance >= 56) {
if (refDCval < 57) {
if (diffMVdiffVal < 37) {
if (AFD < 15) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
return 4;
}
}
else if (AFD >= 15) {
if (SC < 799) {
return 4;
}
else if (SC >= 799) {
return 1;
}
}
}
else if (diffMVdiffVal >= 37) {
return 1;
}
}
else if (refDCval >= 57) {
if (MVDiff < 66) {
if (ssDCval < 102) {
if (TSC < 5) {
if (diffMVdiffVal < -5) {
return 8;
}
else if (diffMVdiffVal >= -5) {
if (diffMVdiffVal < 7) {
return 2;
}
else if (diffMVdiffVal >= 7) {
return 8;
}
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (ssDCval >= 102) {
if (Rs < 696) {
return 1;
}
else if (Rs >= 696) {
return 8;
}
}
}
else if (MVDiff >= 66) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
if (diffMVdiffVal < -93) {
return 2;
}
else if (diffMVdiffVal >= -93) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
else if (RsCsDiff >= 37) {
if (refDCval < 83) {
if (CsDiff < 4) {
if (diffAFD < -2) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Rs < 635) {
return 8;
}
else if (Rs >= 635) {
return 1;
}
}
}
else if (CsDiff >= 4) {
if (diffAFD < 1) {
if (posBalance < 64) {
return 4;
}
else if (posBalance >= 64) {
if (Rs < 382) {
return 2;
}
else if (Rs >= 382) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (diffRsCsdiff < -2) {
if (Rs < 470) {
return 4;
}
else if (Rs >= 470) {
return 8;
}
}
else if (diffRsCsdiff >= -2) {
if (Rs < 362) {
return 8;
}
else if (Rs >= 362) {
return 1;
}
}
}
}
}
else if (refDCval >= 83) {
if (posBalance < 59) {
if (ssDCval < 90) {
return 2;
}
else if (ssDCval >= 90) {
if (SC < 1173) {
if (Rs < 580) {
return 1;
}
else if (Rs >= 580) {
return 4;
}
}
else if (SC >= 1173) {
return 1;
}
}
}
else if (posBalance >= 59) {
if (ssDCval < 131) {
if (mcTcor < 90) {
return 2;
}
else if (mcTcor >= 90) {
if (mu_mv_mag_sq < 45) {
if (MVDiff < 75) {
return 1;
}
else if (MVDiff >= 75) {
if (AFD < 37) {
return 2;
}
else if (AFD >= 37) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 45) {
return 2;
}
}
}
else if (ssDCval >= 131) {
if (Cs < 172) {
return 1;
}
else if (Cs >= 172) {
if (diffAFD < 5) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (diffAFD >= 5) {
return 1;
}
}
}
}
}
}
}
}
}
else if (tcor >= 70) {
if (refDCval < 139) {
if (Rs < 577) {
if (Cs < 133) {
if (TSC < 13) {
if (refDCval < 81) {
return 1;
}
else if (refDCval >= 81) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 1;
}
}
}
else if (TSC >= 13) {
return 2;
}
}
else if (Cs >= 133) {
if (diffAFD < 1) {
if (diffMVdiffVal < 12) {
if (negBalance < 47) {
if (posBalance < 38) {
return 2;
}
else if (posBalance >= 38) {
if (Rs < 508) {
return 8;
}
else if (Rs >= 508) {
return 1;
}
}
}
else if (negBalance >= 47) {
if (Rs < 537) {
return 8;
}
else if (Rs >= 537) {
if (Rs < 556) {
return 4;
}
else if (Rs >= 556) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 12) {
if (Rs < 314) {
return 4;
}
else if (Rs >= 314) {
return 2;
}
}
}
else if (diffAFD >= 1) {
if (diffTSC < 1) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 1;
}
}
}
else if (diffTSC >= 1) {
if (Cs < 375) {
return 8;
}
else if (Cs >= 375) {
return 4;
}
}
}
}
}
else if (Rs >= 577) {
if (Cs < 273) {
if (AFD < 11) {
if (MVDiff < 12) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Cs < 226) {
return 2;
}
else if (Cs >= 226) {
return 1;
}
}
}
else if (MVDiff >= 12) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (Cs < 158) {
return 1;
}
else if (Cs >= 158) {
return 8;
}
}
}
}
else if (AFD >= 11) {
if (diffRsCsdiff < -18) {
if (refDCval < 104) {
if (MVDiff < 177) {
if (mu_mv_mag_sq < 30) {
if (Cs < 175) {
if (gchDC < 2) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 1;
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (Cs >= 175) {
if (AFD < 16) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 30) {
return 2;
}
}
else if (MVDiff >= 177) {
return 8;
}
}
else if (refDCval >= 104) {
return 1;
}
}
else if (diffRsCsdiff >= -18) {
if (ssDCval < 85) {
if (MVDiff < 225) {
if (RsDiff < 12) {
return 1;
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (MVDiff >= 225) {
return 2;
}
}
else if (ssDCval >= 85) {
if (refDCval < 110) {
if (MVDiff < 28) {
if (Rs < 898) {
return 1;
}
else if (Rs >= 898) {
if (Cs < 244) {
return 2;
}
else if (Cs >= 244) {
return 1;
}
}
}
else if (MVDiff >= 28) {
if (SC < 837) {
if (Rs < 624) {
return 1;
}
else if (Rs >= 624) {
return 8;
}
}
else if (SC >= 837) {
if (Cs < 209) {
return 1;
}
else if (Cs >= 209) {
return 2;
}
}
}
}
else if (refDCval >= 110) {
return 1;
}
}
}
}
}
else if (Cs >= 273) {
if (Rs < 983) {
if (diffMVdiffVal < 13) {
if (negBalance < 52) {
if (MVDiff < 37) {
return 1;
}
else if (MVDiff >= 37) {
if (Rs < 916) {
return 1;
}
else if (Rs >= 916) {
return 4;
}
}
}
else if (negBalance >= 52) {
if (posBalance < 47) {
if (Rs < 660) {
return 4;
}
else if (Rs >= 660) {
return 8;
}
}
else if (posBalance >= 47) {
if (SC < 968) {
return 8;
}
else if (SC >= 968) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 13) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
if (mu_mv_mag_sq < 7) {
return 1;
}
else if (mu_mv_mag_sq >= 7) {
return 2;
}
}
}
}
else if (Rs >= 983) {
if (MVDiff < 11) {
if (posBalance < 45) {
if (Cs < 336) {
return 4;
}
else if (Cs >= 336) {
if (Rs < 1313) {
return 8;
}
else if (Rs >= 1313) {
return 1;
}
}
}
else if (posBalance >= 45) {
if (tcor < 80) {
if (Cs < 320) {
return 1;
}
else if (Cs >= 320) {
return 8;
}
}
else if (tcor >= 80) {
return 1;
}
}
}
else if (MVDiff >= 11) {
return 1;
}
}
}
}
}
else if (refDCval >= 139) {
if (gchDC < 2) {
if (refDCval < 145) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (Cs < 342) {
return 1;
}
else if (Cs >= 342) {
return 2;
}
}
}
else if (refDCval >= 145) {
if (ssDCval < 174) {
return 1;
}
else if (ssDCval >= 174) {
if (Rs < 742) {
return 8;
}
else if (Rs >= 742) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
if (mu_mv_mag_sq < 60) {
if (negBalance < 73) {
return 1;
}
else if (negBalance >= 73) {
if (mu_mv_mag_sq < 18) {
return 1;
}
else if (mu_mv_mag_sq >= 18) {
if (Rs < 1091) {
return 2;
}
else if (Rs >= 1091) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 60) {
if (MVDiff < 146) {
return 2;
}
else if (MVDiff >= 146) {
return 1;
}
}
}
}
}
}
else if (Cs >= 403) {
if (mcTcor < 91) {
if (Rs < 939) {
if (negBalance < 46) {
if (AFD < 8) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (AFD >= 8) {
if (posBalance < 70) {
if (diffRsCsdiff < -12) {
if (Rs < 543) {
return 1;
}
else if (Rs >= 543) {
return 8;
}
}
else if (diffRsCsdiff >= -12) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
return 8;
}
}
}
else if (posBalance >= 70) {
if (Cs < 434) {
return 8;
}
else if (Cs >= 434) {
if (Rs < 633) {
return 2;
}
else if (Rs >= 633) {
return 4;
}
}
}
}
}
else if (negBalance >= 46) {
if (AFD < 25) {
if (diffMVdiffVal < 5) {
if (mcTcor < 89) {
if (mu_mv_mag_sq < 91) {
if (Cs < 490) {
return 8;
}
else if (Cs >= 490) {
if (Rs < 685) {
if (Rs < 602) {
if (Rs < 596) {
return 2;
}
else if (Rs >= 596) {
return 8;
}
}
else if (Rs >= 602) {
return 4;
}
}
else if (Rs >= 685) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 91) {
return 2;
}
}
else if (mcTcor >= 89) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 5) {
if (CsDiff < 4) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (CsDiff >= 4) {
if (mcTcor < 85) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (mcTcor >= 85) {
if (Rs < 687) {
return 4;
}
else if (Rs >= 687) {
if (Rs < 811) {
return 8;
}
else if (Rs >= 811) {
return 2;
}
}
}
}
}
}
else if (AFD >= 25) {
if (MVDiff < 181) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
else if (MVDiff >= 181) {
if (Rs < 397) {
return 2;
}
else if (Rs >= 397) {
return 8;
}
}
}
}
}
else if (Rs >= 939) {
if (RsCsDiff < 49) {
if (posBalance < 55) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (posBalance < 50) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 2;
}
}
else if (posBalance >= 50) {
return 2;
}
}
}
else if (posBalance >= 55) {
if (diffRsCsdiff < 10) {
if (diffMVdiffVal < 8) {
if (TSC < 5) {
if (Cs < 497) {
return 1;
}
else if (Cs >= 497) {
return 8;
}
}
else if (TSC >= 5) {
if (diffMVdiffVal < -8) {
if (SC < 1544) {
return 2;
}
else if (SC >= 1544) {
return 8;
}
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
}
else if (diffMVdiffVal >= 8) {
return 4;
}
}
else if (diffRsCsdiff >= 10) {
if (Cs < 470) {
return 4;
}
else if (Cs >= 470) {
return 1;
}
}
}
}
else if (RsCsDiff >= 49) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (AFD < 23) {
if (ssDCval < 120) {
return 4;
}
else if (ssDCval >= 120) {
if (Rs < 1342) {
return 2;
}
else if (Rs >= 1342) {
return 4;
}
}
}
else if (AFD >= 23) {
if (Rs < 1393) {
if (gchDC < 8) {
return 2;
}
else if (gchDC >= 8) {
return 8;
}
}
else if (Rs >= 1393) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (MVDiff < 51) {
if (posBalance < 47) {
if (mu_mv_mag_sq < 76) {
if (tcor < 90) {
if (SC < 1393) {
return 2;
}
else if (SC >= 1393) {
if (mu_mv_mag_sq < 36) {
return 1;
}
else if (mu_mv_mag_sq >= 36) {
if (AFD < 12) {
if (Rs < 949) {
return 2;
}
else if (Rs >= 949) {
return 1;
}
}
else if (AFD >= 12) {
return 2;
}
}
}
}
else if (tcor >= 90) {
if (diffRsCsdiff < 9) {
if (MVDiff < 10) {
if (Rs < 1362) {
return 8;
}
else if (Rs >= 1362) {
return 1;
}
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (diffRsCsdiff >= 9) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (ssDCval < 61) {
if (Rs < 936) {
return 2;
}
else if (Rs >= 936) {
return 4;
}
}
else if (ssDCval >= 61) {
return 8;
}
}
}
else if (posBalance >= 47) {
if (tcor < 90) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 3) {
if (RsCsDiff < 62) {
return 1;
}
else if (RsCsDiff >= 62) {
if (Rs < 905) {
return 1;
}
else if (Rs >= 905) {
if (Rs < 1014) {
return 2;
}
else if (Rs >= 1014) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (RsCsDiff < 31) {
if (gchDC < 3) {
if (Rs < 586) {
return 4;
}
else if (Rs >= 586) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (gchDC >= 3) {
return 1;
}
}
else if (RsCsDiff >= 31) {
if (Cs < 508) {
if (negBalance < 57) {
if (MVDiff < 18) {
if (Rs < 887) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
return 4;
}
}
else if (Rs >= 887) {
return 1;
}
}
else if (MVDiff >= 18) {
if (diffMVdiffVal < 4) {
if (tcor < 74) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (tcor >= 74) {
return 1;
}
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
}
else if (negBalance >= 57) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 4;
}
}
}
else if (Cs >= 508) {
if (tcor < 80) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
else if (tcor >= 80) {
if (Cs < 529) {
return 8;
}
else if (Cs >= 529) {
return 4;
}
}
}
}
}
}
else if (diffAFD >= 2) {
if (Rs < 766) {
if (MVDiff < 25) {
if (Rs < 483) {
return 1;
}
else if (Rs >= 483) {
return 4;
}
}
else if (MVDiff >= 25) {
if (Rs < 647) {
return 1;
}
else if (Rs >= 647) {
return 2;
}
}
}
else if (Rs >= 766) {
return 1;
}
}
}
else if (tcor >= 90) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
return 4;
}
}
}
}
else if (MVDiff >= 51) {
if (negBalance < 62) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (RsDiff < 9) {
if (diffRsCsdiff < -33) {
return 8;
}
else if (diffRsCsdiff >= -33) {
if (diffTSC < 1) {
if (Rs < 512) {
if (Rs < 357) {
return 8;
}
else if (Rs >= 357) {
return 2;
}
}
else if (Rs >= 512) {
if (tcor < 66) {
return 1;
}
else if (tcor >= 66) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 1;
}
}
}
}
else if (RsDiff >= 9) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (diffAFD < 6) {
return 2;
}
else if (diffAFD >= 6) {
return 1;
}
}
}
}
}
else if (negBalance >= 62) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
if (Rs < 819) {
if (SC < 1194) {
return 8;
}
else if (SC >= 1194) {
return 2;
}
}
else if (Rs >= 819) {
if (Cs < 438) {
return 1;
}
else if (Cs >= 438) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 1419) {
if (ssDCval < 113) {
if (RsCsDiff < 147) {
if (mu_mv_mag_sq < 87) {
if (posBalance < 54) {
if (ssDCval < 69) {
if (mcTcor < 97) {
if (tcor < 88) {
if (Rs < 1642) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
return 2;
}
}
else if (Rs >= 1642) {
return 2;
}
}
else if (tcor >= 88) {
if (SC < 2602) {
if (MVDiff < 7) {
if (Rs < 2102) {
return 2;
}
else if (Rs >= 2102) {
return 1;
}
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (SC >= 2602) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (diffRsCsdiff < 2) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
return 1;
}
}
else if (diffRsCsdiff >= 2) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
}
}
else if (ssDCval >= 69) {
if (RsCsDiff < 82) {
if (ssDCval < 98) {
if (Rs < 1440) {
if (Rs < 1423) {
return 1;
}
else if (Rs >= 1423) {
if (Rs < 1429) {
return 4;
}
else if (Rs >= 1429) {
return 2;
}
}
}
else if (Rs >= 1440) {
if (Cs < 207) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (Cs >= 207) {
return 1;
}
}
}
else if (ssDCval >= 98) {
if (TSC < 4) {
if (Cs < 200) {
return 2;
}
else if (Cs >= 200) {
return 8;
}
}
else if (TSC >= 4) {
return 1;
}
}
}
else if (RsCsDiff >= 82) {
if (Cs < 435) {
if (TSC < 7) {
if (ssDCval < 81) {
if (RsCsDiff < 84) {
return 2;
}
else if (RsCsDiff >= 84) {
return 1;
}
}
else if (ssDCval >= 81) {
if (Cs < 416) {
return 2;
}
else if (Cs >= 416) {
return 1;
}
}
}
else if (TSC >= 7) {
if (mu_mv_mag_sq < 24) {
return 1;
}
else if (mu_mv_mag_sq >= 24) {
return 2;
}
}
}
else if (Cs >= 435) {
if (posBalance < 46) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (Rs < 1971) {
return 1;
}
else if (Rs >= 1971) {
return 4;
}
}
}
else if (posBalance >= 46) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 54) {
if (MVDiff < 10) {
if (posBalance < 56) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
if (Rs < 2260) {
return 2;
}
else if (Rs >= 2260) {
return 1;
}
}
}
else if (posBalance >= 56) {
return 2;
}
}
else if (MVDiff >= 10) {
if (mu_mv_mag_sq < 36) {
if (Cs < 413) {
if (refDCval < 94) {
if (RsDiff < 12) {
if (SC < 1954) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 1;
}
}
}
else if (SC >= 1954) {
if (MVDiff < 38) {
if (negBalance < 36) {
if (Cs < 187) {
return 2;
}
else if (Cs >= 187) {
return 1;
}
}
else if (negBalance >= 36) {
return 2;
}
}
else if (MVDiff >= 38) {
return 1;
}
}
}
else if (RsDiff >= 12) {
if (Rs < 1972) {
return 2;
}
else if (Rs >= 1972) {
return 4;
}
}
}
else if (refDCval >= 94) {
if (Rs < 1516) {
if (diffMVdiffVal < 15) {
return 2;
}
else if (diffMVdiffVal >= 15) {
return 1;
}
}
else if (Rs >= 1516) {
if (Cs < 200) {
if (MVDiff < 56) {
return 2;
}
else if (MVDiff >= 56) {
return 1;
}
}
else if (Cs >= 200) {
if (mcTcor < 96) {
if (diffAFD < -1) {
if (MVDiff < 62) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
if (Rs < 1552) {
return 1;
}
else if (Rs >= 1552) {
return 8;
}
}
}
else if (MVDiff >= 62) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (mcTcor >= 96) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
}
}
else if (Cs >= 413) {
if (diffMVdiffVal < 6) {
if (mcTcor < 94) {
return 2;
}
else if (mcTcor >= 94) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (diffMVdiffVal >= 6) {
if (diffRsCsdiff < 46) {
if (Cs < 458) {
return 1;
}
else if (Cs >= 458) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (diffRsCsdiff >= 46) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 36) {
if (Cs < 269) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
else if (Cs >= 269) {
if (diffRsCsdiff < 10) {
if (posBalance < 62) {
return 4;
}
else if (posBalance >= 62) {
if (Rs < 1941) {
return 8;
}
else if (Rs >= 1941) {
return 1;
}
}
}
else if (diffRsCsdiff >= 10) {
if (Rs < 1698) {
if (Rs < 1576) {
return 4;
}
else if (Rs >= 1576) {
return 2;
}
}
else if (Rs >= 1698) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
return 2;
}
}
else if (RsCsDiff >= 147) {
if (MVDiff < 35) {
if (ssDCval < 86) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (tcor < 55) {
if (SC < 2198) {
return 4;
}
else if (SC >= 2198) {
return 2;
}
}
else if (tcor >= 55) {
if (diffMVdiffVal < -10) {
if (Cs < 228) {
return 2;
}
else if (Cs >= 228) {
return 1;
}
}
else if (diffMVdiffVal >= -10) {
return 2;
}
}
}
}
else if (ssDCval >= 86) {
if (tcor < 70) {
return 2;
}
else if (tcor >= 70) {
if (MVDiff < 27) {
if (Rs < 1769) {
return 2;
}
else if (Rs >= 1769) {
if (Rs < 2128) {
return 1;
}
else if (Rs >= 2128) {
if (diffMVdiffVal < -2) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
}
}
else if (MVDiff >= 27) {
return 2;
}
}
}
}
else if (MVDiff >= 35) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 28) {
if (TSC < 15) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (TSC >= 15) {
if (Cs < 238) {
return 2;
}
else if (Cs >= 238) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 28) {
if (RsCsDiff < 182) {
if (Cs < 373) {
return 1;
}
else if (Cs >= 373) {
return 2;
}
}
else if (RsCsDiff >= 182) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (Cs < 365) {
if (RsDiff < 14) {
if (refDCval < 100) {
if (Rs < 1654) {
return 2;
}
else if (Rs >= 1654) {
if (Rs < 2080) {
return 4;
}
else if (Rs >= 2080) {
return 8;
}
}
}
else if (refDCval >= 100) {
return 1;
}
}
else if (RsDiff >= 14) {
if (MVDiff < 48) {
if (Cs < 79) {
return 2;
}
else if (Cs >= 79) {
return 1;
}
}
else if (MVDiff >= 48) {
return 2;
}
}
}
else if (Cs >= 365) {
if (diffAFD < 2) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 8;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (mcTcor < 93) {
if (ssDCval < 126) {
if (posBalance < 74) {
return 8;
}
else if (posBalance >= 74) {
if (Rs < 1737) {
return 2;
}
else if (Rs >= 1737) {
return 1;
}
}
}
else if (ssDCval >= 126) {
if (RsDiff < 11) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (Rs < 1534) {
return 2;
}
else if (Rs >= 1534) {
return 8;
}
}
}
else if (RsDiff >= 11) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
if (AFD < 12) {
if (Rs < 1726) {
if (Cs < 461) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (Cs >= 461) {
return 2;
}
}
else if (Rs >= 1726) {
if (negBalance < 26) {
if (Rs < 2291) {
return 2;
}
else if (Rs >= 2291) {
return 8;
}
}
else if (negBalance >= 26) {
if (mu_mv_mag_sq < 66) {
return 1;
}
else if (mu_mv_mag_sq >= 66) {
if (Rs < 1811) {
return 2;
}
else if (Rs >= 1811) {
return 1;
}
}
}
}
}
else if (AFD >= 12) {
if (RsCsDiff < 345) {
if (MVDiff < 6) {
if (Rs < 2012) {
return 1;
}
else if (Rs >= 2012) {
return 2;
}
}
else if (MVDiff >= 6) {
if (Rs < 2268) {
if (Cs < 408) {
return 1;
}
else if (Cs >= 408) {
if (mu_mv_mag_sq < 10) {
if (ssDCval < 159) {
return 1;
}
else if (ssDCval >= 159) {
return 8;
}
}
else if (mu_mv_mag_sq >= 10) {
if (ssDCval < 134) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (negBalance < 45) {
if (Rs < 1806) {
return 1;
}
else if (Rs >= 1806) {
return 2;
}
}
else if (negBalance >= 45) {
return 1;
}
}
}
else if (ssDCval >= 134) {
return 1;
}
}
}
}
else if (Rs >= 2268) {
if (Rs < 2272) {
return 8;
}
else if (Rs >= 2272) {
if (CsDiff < 2) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 8;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 345) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 535) {
if (negBalance < 13) {
if (Rs < 699) {
if (ssDCval < 56) {
if (ssDCval < 30) {
return 8;
}
else if (ssDCval >= 30) {
if (mu_mv_mag_sq < 127) {
if (SC < 1635) {
return 1;
}
else if (SC >= 1635) {
return 2;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 351) {
if (Rs < 258) {
return 2;
}
else if (Rs >= 258) {
return 1;
}
}
else if (Rs >= 351) {
return 8;
}
}
}
}
else if (ssDCval >= 56) {
if (Rs < 450) {
if (SC < 1435) {
if (Cs < 568) {
return 2;
}
else if (Cs >= 568) {
if (posBalance < 120) {
return 4;
}
else if (posBalance >= 120) {
return 8;
}
}
}
else if (SC >= 1435) {
return 8;
}
}
else if (Rs >= 450) {
if (negBalance < 10) {
if (Rs < 508) {
if (diffRsCsdiff < -9) {
if (Rs < 486) {
return 8;
}
else if (Rs >= 486) {
return 4;
}
}
else if (diffRsCsdiff >= -9) {
return 1;
}
}
else if (Rs >= 508) {
return 8;
}
}
else if (negBalance >= 10) {
if (Rs < 666) {
if (diffMVdiffVal < -1) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 626) {
if (SC < 1073) {
return 2;
}
else if (SC >= 1073) {
return 8;
}
}
else if (Rs >= 626) {
if (Rs < 647) {
return 2;
}
else if (Rs >= 647) {
if (Rs < 658) {
return 1;
}
else if (Rs >= 658) {
return 8;
}
}
}
}
}
else if (Rs >= 666) {
if (posBalance < 31) {
return 1;
}
else if (posBalance >= 31) {
if (Rs < 678) {
return 1;
}
else if (Rs >= 678) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 699) {
if (diffAFD < -3) {
if (SC < 2306) {
if (mu_mv_mag_sq < 128) {
if (SC < 1956) {
return 8;
}
else if (SC >= 1956) {
if (diffRsCsdiff < -11) {
return 1;
}
else if (diffRsCsdiff >= -11) {
if (Rs < 1002) {
return 8;
}
else if (Rs >= 1002) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 640) {
if (SC < 1480) {
return 2;
}
else if (SC >= 1480) {
if (Rs < 1088) {
return 8;
}
else if (Rs >= 1088) {
if (Rs < 1481) {
return 1;
}
else if (Rs >= 1481) {
return 8;
}
}
}
}
else if (Cs >= 640) {
if (diffRsCsdiff < -68) {
if (diffTSC < -8) {
if (SC < 1919) {
return 8;
}
else if (SC >= 1919) {
return 4;
}
}
else if (diffTSC >= -8) {
if (Rs < 950) {
return 1;
}
else if (Rs >= 950) {
return 8;
}
}
}
else if (diffRsCsdiff >= -68) {
if (Cs < 1065) {
if (MVDiff < 5) {
if (Rs < 1064) {
return 1;
}
else if (Rs >= 1064) {
return 8;
}
}
else if (MVDiff >= 5) {
if (diffMVdiffVal < -44) {
if (Rs < 874) {
return 4;
}
else if (Rs >= 874) {
return 8;
}
}
else if (diffMVdiffVal >= -44) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
return 4;
}
}
}
}
else if (Cs >= 1065) {
if (Rs < 1074) {
return 1;
}
else if (Rs >= 1074) {
return 8;
}
}
}
}
}
}
else if (SC >= 2306) {
if (MVDiff < 16) {
if (diffRsCsdiff < -169) {
return 1;
}
else if (diffRsCsdiff >= -169) {
if (Rs < 1151) {
return 1;
}
else if (Rs >= 1151) {
return 8;
}
}
}
else if (MVDiff >= 16) {
if (refDCval < 121) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Rs < 1222) {
return 8;
}
else if (Rs >= 1222) {
return 1;
}
}
}
else if (refDCval >= 121) {
if (MVDiff < 80) {
if (MVDiff < 23) {
if (diffAFD < -8) {
return 8;
}
else if (diffAFD >= -8) {
return 4;
}
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (MVDiff >= 80) {
return 1;
}
}
}
}
}
else if (diffAFD >= -3) {
if (Rs < 1841) {
if (ssDCval < 119) {
if (MVDiff < 3) {
if (negBalance < 11) {
if (diffRsCsdiff < -1) {
if (Cs < 714) {
if (Rs < 1004) {
if (Cs < 698) {
return 8;
}
else if (Cs >= 698) {
return 2;
}
}
else if (Rs >= 1004) {
if (Rs < 1732) {
return 4;
}
else if (Rs >= 1732) {
return 1;
}
}
}
else if (Cs >= 714) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
if (SC < 2887) {
if (diffMVdiffVal < -1) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (SC >= 2887) {
if (Rs < 1753) {
return 8;
}
else if (Rs >= 1753) {
return 1;
}
}
}
}
else if (negBalance >= 11) {
if (Cs < 692) {
if (Rs < 1148) {
return 8;
}
else if (Rs >= 1148) {
if (Rs < 1519) {
return 1;
}
else if (Rs >= 1519) {
return 8;
}
}
}
else if (Cs >= 692) {
if (ssDCval < 57) {
if (posBalance < 45) {
if (Rs < 748) {
return 2;
}
else if (Rs >= 748) {
return 8;
}
}
else if (posBalance >= 45) {
if (Rs < 1133) {
return 1;
}
else if (Rs >= 1133) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (Cs < 991) {
if (Rs < 742) {
return 1;
}
else if (Rs >= 742) {
if (AFD < 1) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (Cs >= 991) {
return 1;
}
}
}
}
}
else if (MVDiff >= 3) {
if (refDCval < 33) {
return 8;
}
else if (refDCval >= 33) {
if (ssDCval < 39) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
else if (ssDCval >= 39) {
if (SC < 1768) {
if (Rs < 724) {
return 1;
}
else if (Rs >= 724) {
return 8;
}
}
else if (SC >= 1768) {
if (ssDCval < 100) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (gchDC < 1) {
if (Rs < 1114) {
if (Rs < 1059) {
return 1;
}
else if (Rs >= 1059) {
return 8;
}
}
else if (Rs >= 1114) {
return 1;
}
}
else if (gchDC >= 1) {
if (Cs < 877) {
if (Cs < 735) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (Cs >= 735) {
return 4;
}
}
else if (Cs >= 877) {
if (Rs < 1551) {
return 1;
}
else if (Rs >= 1551) {
return 8;
}
}
}
}
}
else if (ssDCval >= 100) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 119) {
if (Cs < 1183) {
if (posBalance < 10) {
if (ssDCval < 147) {
if (ssDCval < 130) {
return 1;
}
else if (ssDCval >= 130) {
if (diffMVdiffVal < -11) {
return 4;
}
else if (diffMVdiffVal >= -11) {
return 8;
}
}
}
else if (ssDCval >= 147) {
return 2;
}
}
else if (posBalance >= 10) {
if (SC < 1395) {
return 1;
}
else if (SC >= 1395) {
if (RsCsDiff < 5) {
if (diffMVdiffVal < -18) {
if (Rs < 1102) {
return 8;
}
else if (Rs >= 1102) {
return 4;
}
}
else if (diffMVdiffVal >= -18) {
return 8;
}
}
else if (RsCsDiff >= 5) {
if (mu_mv_mag_sq < 120) {
if (ssDCval < 150) {
return 8;
}
else if (ssDCval >= 150) {
return 2;
}
}
else if (mu_mv_mag_sq >= 120) {
return 1;
}
}
}
}
}
else if (Cs >= 1183) {
if (Rs < 872) {
return 2;
}
else if (Rs >= 872) {
if (posBalance < 9) {
return 1;
}
else if (posBalance >= 9) {
if (AFD < 1) {
if (Rs < 1112) {
return 8;
}
else if (Rs >= 1112) {
return 1;
}
}
else if (AFD >= 1) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1841) {
if (Rs < 2021) {
if (negBalance < 4) {
return 8;
}
else if (negBalance >= 4) {
if (Cs < 1000) {
if (tcor < 96) {
if (Rs < 1964) {
return 2;
}
else if (Rs >= 1964) {
return 8;
}
}
else if (tcor >= 96) {
if (Rs < 2009) {
return 1;
}
else if (Rs >= 2009) {
return 4;
}
}
}
else if (Cs >= 1000) {
if (ssDCval < 69) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 4;
}
}
else if (ssDCval >= 69) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
return 8;
}
}
}
}
}
else if (Rs >= 2021) {
if (posBalance < 29) {
if (mu_mv_mag_sq < 97) {
return 4;
}
else if (mu_mv_mag_sq >= 97) {
return 8;
}
}
else if (posBalance >= 29) {
if (ssDCval < 78) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (Cs < 1303) {
return 2;
}
else if (Cs >= 1303) {
if (SC < 3612) {
return 8;
}
else if (SC >= 3612) {
if (Cs < 1323) {
return 2;
}
else if (Cs >= 1323) {
if (Rs < 2290) {
return 2;
}
else if (Rs >= 2290) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 78) {
if (Rs < 2043) {
return 8;
}
else if (Rs >= 2043) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 14) {
if (posBalance < 42) {
if (refDCval < 47) {
if (MVDiff < 11) {
if (TSCindex < 3) {
if (Rs < 575) {
return 4;
}
else if (Rs >= 575) {
return 8;
}
}
else if (TSCindex >= 3) {
if (AFD < 9) {
if (Rs < 1073) {
if (ssDCval < 29) {
return 4;
}
else if (ssDCval >= 29) {
if (SC < 1538) {
if (ssDCval < 37) {
return 1;
}
else if (ssDCval >= 37) {
if (Rs < 582) {
return 8;
}
else if (Rs >= 582) {
return 4;
}
}
}
else if (SC >= 1538) {
return 8;
}
}
}
else if (Rs >= 1073) {
if (RsDiff < 6) {
if (RsDiff < 3) {
if (Rs < 1296) {
return 8;
}
else if (Rs >= 1296) {
return 4;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (RsDiff >= 6) {
if (Rs < 1727) {
return 8;
}
else if (Rs >= 1727) {
return 4;
}
}
}
}
else if (AFD >= 9) {
if (Rs < 1510) {
return 2;
}
else if (Rs >= 1510) {
return 1;
}
}
}
}
else if (MVDiff >= 11) {
if (Cs < 1001) {
if (ssDCval < 44) {
if (refDCval < 23) {
return 4;
}
else if (refDCval >= 23) {
if (diffRsCsdiff < -53) {
return 1;
}
else if (diffRsCsdiff >= -53) {
return 8;
}
}
}
else if (ssDCval >= 44) {
if (Rs < 1399) {
return 2;
}
else if (Rs >= 1399) {
return 4;
}
}
}
else if (Cs >= 1001) {
if (refDCval < 46) {
if (MVDiff < 30) {
return 1;
}
else if (MVDiff >= 30) {
return 8;
}
}
else if (refDCval >= 46) {
return 8;
}
}
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 70) {
if (refDCval < 88) {
if (Cs < 843) {
if (refDCval < 58) {
if (mu_mv_mag_sq < 49) {
if (refDCval < 54) {
if (Cs < 729) {
return 4;
}
else if (Cs >= 729) {
return 8;
}
}
else if (refDCval >= 54) {
return 2;
}
}
else if (mu_mv_mag_sq >= 49) {
if (Rs < 762) {
return 4;
}
else if (Rs >= 762) {
return 1;
}
}
}
else if (refDCval >= 58) {
if (Rs < 1491) {
if (TSC < 4) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (TSC >= 4) {
if (RsCsDiff < 31) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (RsCsDiff >= 31) {
if (ssDCval < 60) {
if (Rs < 1397) {
return 1;
}
else if (Rs >= 1397) {
return 4;
}
}
else if (ssDCval >= 60) {
return 8;
}
}
}
}
else if (Rs >= 1491) {
if (ssDCval < 69) {
if (SC < 2439) {
return 2;
}
else if (SC >= 2439) {
return 1;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
else if (Cs >= 843) {
if (negBalance < 56) {
if (AFD < 7) {
if (TSC < 4) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (TSC >= 4) {
if (Rs < 1108) {
return 4;
}
else if (Rs >= 1108) {
if (Cs < 979) {
return 1;
}
else if (Cs >= 979) {
return 4;
}
}
}
}
else if (AFD >= 7) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (tcor < 91) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
if (Rs < 2170) {
if (Rs < 1800) {
return 8;
}
else if (Rs >= 1800) {
return 4;
}
}
else if (Rs >= 2170) {
return 2;
}
}
}
else if (tcor >= 91) {
if (CsDiff < 5) {
if (Cs < 923) {
return 1;
}
else if (Cs >= 923) {
return 2;
}
}
else if (CsDiff >= 5) {
return 4;
}
}
}
}
}
else if (negBalance >= 56) {
return 8;
}
}
}
else if (refDCval >= 88) {
if (ssDCval < 153) {
if (Rs < 1790) {
if (Cs < 556) {
return 2;
}
else if (Cs >= 556) {
if (posBalance < 40) {
if (mu_mv_mag_sq < 3) {
if (refDCval < 141) {
if (Rs < 1523) {
return 8;
}
else if (Rs >= 1523) {
return 1;
}
}
else if (refDCval >= 141) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
return 8;
}
}
}
else if (posBalance >= 40) {
return 2;
}
}
}
else if (Rs >= 1790) {
if (AFD < 10) {
if (Rs < 2193) {
return 1;
}
else if (Rs >= 2193) {
return 8;
}
}
else if (AFD >= 10) {
return 2;
}
}
}
else if (ssDCval >= 153) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (refDCval < 58) {
if (SC < 2377) {
if (TSC < 3) {
if (negBalance < 25) {
if (Rs < 698) {
return 2;
}
else if (Rs >= 698) {
if (Rs < 1071) {
return 1;
}
else if (Rs >= 1071) {
return 8;
}
}
}
else if (negBalance >= 25) {
return 4;
}
}
else if (TSC >= 3) {
if (diffAFD < -2) {
if (Rs < 869) {
return 8;
}
else if (Rs >= 869) {
return 4;
}
}
else if (diffAFD >= -2) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
return 4;
}
}
}
}
else if (SC >= 2377) {
if (ssDCval < 49) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
return 4;
}
}
}
else if (ssDCval >= 49) {
if (CsDiff < 3) {
if (Rs < 2195) {
return 1;
}
else if (Rs >= 2195) {
return 8;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
else if (refDCval >= 58) {
if (mu_mv_mag_sq < 89) {
if (posBalance < 34) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
if (MVDiff < 4) {
if (Rs < 2275) {
return 8;
}
else if (Rs >= 2275) {
return 8;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (posBalance >= 34) {
if (negBalance < 29) {
if (Cs < 1259) {
return 4;
}
else if (Cs >= 1259) {
return 2;
}
}
else if (negBalance >= 29) {
if (MVDiff < 17) {
if (negBalance < 31) {
return 2;
}
else if (negBalance >= 31) {
if (Rs < 690) {
if (Cs < 613) {
return 8;
}
else if (Cs >= 613) {
return 2;
}
}
else if (Rs >= 690) {
if (diffRsCsdiff < 13) {
return 8;
}
else if (diffRsCsdiff >= 13) {
if (Rs < 1974) {
return 1;
}
else if (Rs >= 1974) {
return 8;
}
}
}
}
}
else if (MVDiff >= 17) {
if (diffAFD < -1) {
if (Rs < 1723) {
return 8;
}
else if (Rs >= 1723) {
return 2;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (Rs < 582) {
if (Cs < 841) {
if (Rs < 505) {
return 4;
}
else if (Rs >= 505) {
if (Rs < 579) {
return 8;
}
else if (Rs >= 579) {
return 2;
}
}
}
else if (Cs >= 841) {
return 4;
}
}
else if (Rs >= 582) {
if (negBalance < 50) {
if (RsCsDiff < 23) {
if (MVDiff < 26) {
if (diffRsCsdiff < 6) {
if (Rs < 1010) {
if (negBalance < 27) {
return 8;
}
else if (negBalance >= 27) {
return 4;
}
}
else if (Rs >= 1010) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 8;
}
}
}
else if (diffRsCsdiff >= 6) {
if (Cs < 755) {
return 8;
}
else if (Cs >= 755) {
return 1;
}
}
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (RsCsDiff >= 23) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (MVDiff < 5) {
if (tcor < 92) {
return 2;
}
else if (tcor >= 92) {
return 8;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (negBalance >= 50) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
if (Rs < 1872) {
return 4;
}
else if (Rs >= 1872) {
if (Rs < 2262) {
return 4;
}
else if (Rs >= 2262) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (posBalance >= 42) {
if (ssDCval < 156) {
if (RsDiff < 8) {
if (Rs < 903) {
if (refDCval < 75) {
if (ssDCval < 41) {
if (Cs < 831) {
if (refDCval < 39) {
if (Rs < 316) {
return 1;
}
else if (Rs >= 316) {
if (posBalance < 56) {
if (posBalance < 45) {
if (RsCsDiff < 30) {
return 1;
}
else if (RsCsDiff >= 30) {
return 8;
}
}
else if (posBalance >= 45) {
if (Rs < 882) {
return 8;
}
else if (Rs >= 882) {
if (Rs < 884) {
return 2;
}
else if (Rs >= 884) {
return 8;
}
}
}
}
else if (posBalance >= 56) {
if (MVDiff < 14) {
if (Rs < 597) {
return 2;
}
else if (Rs >= 597) {
return 8;
}
}
else if (MVDiff >= 14) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
}
}
}
else if (refDCval >= 39) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (Cs >= 831) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (negBalance < 39) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
else if (negBalance >= 39) {
return 2;
}
}
}
}
else if (ssDCval >= 41) {
if (Rs < 511) {
if (Cs < 784) {
if (SC < 952) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 1;
}
}
else if (SC >= 952) {
if (MVDiff < 5) {
if (AFD < 3) {
if (Cs < 625) {
return 1;
}
else if (Cs >= 625) {
return 4;
}
}
else if (AFD >= 3) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (posBalance < 61) {
if (negBalance < 34) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
if (Cs < 700) {
return 2;
}
else if (Cs >= 700) {
return 4;
}
}
}
else if (negBalance >= 34) {
if (ssDCval < 54) {
if (negBalance < 57) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (negBalance >= 57) {
return 4;
}
}
else if (ssDCval >= 54) {
if (tcor < 86) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (tcor >= 86) {
if (Cs < 624) {
return 8;
}
else if (Cs >= 624) {
return 2;
}
}
}
}
}
else if (posBalance >= 61) {
if (ssDCval < 46) {
return 4;
}
else if (ssDCval >= 46) {
if (SC < 1028) {
if (Rs < 470) {
return 2;
}
else if (Rs >= 470) {
return 4;
}
}
else if (SC >= 1028) {
return 8;
}
}
}
}
}
}
else if (Cs >= 784) {
if (ssDCval < 54) {
if (TSC < 4) {
if (ssDCval < 49) {
if (ssDCval < 46) {
return 4;
}
else if (ssDCval >= 46) {
return 1;
}
}
else if (ssDCval >= 49) {
if (Rs < 327) {
return 1;
}
else if (Rs >= 327) {
return 2;
}
}
}
else if (TSC >= 4) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (mcTcor < 87) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (mcTcor >= 87) {
return 2;
}
}
}
}
else if (ssDCval >= 54) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
if (Rs < 443) {
return 4;
}
else if (Rs >= 443) {
return 2;
}
}
}
}
}
}
else if (Rs >= 511) {
if (CsDiff < 6) {
if (negBalance < 31) {
if (refDCval < 58) {
if (diffTSC < 1) {
if (negBalance < 20) {
if (Rs < 571) {
return 4;
}
else if (Rs >= 571) {
return 1;
}
}
else if (negBalance >= 20) {
if (MVDiff < 2) {
if (Rs < 667) {
return 4;
}
else if (Rs >= 667) {
return 8;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 706) {
if (refDCval < 51) {
return 1;
}
else if (refDCval >= 51) {
return 4;
}
}
else if (Rs >= 706) {
return 8;
}
}
}
else if (refDCval >= 58) {
if (RsDiff < 2) {
if (Cs < 690) {
if (Rs < 748) {
return 2;
}
else if (Rs >= 748) {
return 8;
}
}
else if (Cs >= 690) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 616) {
return 4;
}
else if (Rs >= 616) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 75) {
if (Rs < 591) {
return 8;
}
else if (Rs >= 591) {
if (Cs < 733) {
return 1;
}
else if (Cs >= 733) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Cs < 851) {
return 2;
}
else if (Cs >= 851) {
if (Rs < 794) {
return 4;
}
else if (Rs >= 794) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 31) {
if (AFD < 11) {
if (mcTcor < 76) {
if (Rs < 819) {
return 8;
}
else if (Rs >= 819) {
if (Cs < 806) {
if (refDCval < 46) {
return 1;
}
else if (refDCval >= 46) {
return 8;
}
}
else if (Cs >= 806) {
return 8;
}
}
}
else if (mcTcor >= 76) {
if (tcor < 71) {
if (RsCsDiff < 29) {
if (ssDCval < 50) {
if (MVDiff < 19) {
return 1;
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (ssDCval >= 50) {
if (Rs < 596) {
return 1;
}
else if (Rs >= 596) {
return 4;
}
}
}
else if (RsCsDiff >= 29) {
if (Rs < 750) {
if (Cs < 655) {
return 1;
}
else if (Cs >= 655) {
return 2;
}
}
else if (Rs >= 750) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
if (Rs < 876) {
return 4;
}
else if (Rs >= 876) {
return 1;
}
}
}
}
}
else if (tcor >= 71) {
if (refDCval < 68) {
if (MVDiff < 6) {
if (Cs < 632) {
if (mu_mv_mag_sq < 24) {
if (Rs < 774) {
return 1;
}
else if (Rs >= 774) {
return 8;
}
}
else if (mu_mv_mag_sq >= 24) {
if (refDCval < 62) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (Rs < 574) {
return 2;
}
else if (Rs >= 574) {
return 8;
}
}
}
else if (refDCval >= 62) {
return 4;
}
}
}
else if (Cs >= 632) {
if (Cs < 800) {
if (Rs < 775) {
if (ssDCval < 62) {
if (mcTcor < 79) {
return 8;
}
else if (mcTcor >= 79) {
return 4;
}
}
else if (ssDCval >= 62) {
if (Cs < 712) {
return 2;
}
else if (Cs >= 712) {
return 8;
}
}
}
else if (Rs >= 775) {
if (refDCval < 55) {
if (posBalance < 58) {
if (tcor < 94) {
if (Cs < 775) {
return 8;
}
else if (Cs >= 775) {
return 4;
}
}
else if (tcor >= 94) {
return 4;
}
}
else if (posBalance >= 58) {
if (Rs < 799) {
return 1;
}
else if (Rs >= 799) {
return 4;
}
}
}
else if (refDCval >= 55) {
return 8;
}
}
}
else if (Cs >= 800) {
if (negBalance < 43) {
if (Rs < 606) {
if (Cs < 1016) {
return 8;
}
else if (Cs >= 1016) {
return 1;
}
}
else if (Rs >= 606) {
if (SC < 1982) {
if (Cs < 895) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (Rs < 793) {
return 8;
}
else if (Rs >= 793) {
return 4;
}
}
}
else if (Cs >= 895) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (Rs < 869) {
return 2;
}
else if (Rs >= 869) {
return 8;
}
}
}
}
else if (SC >= 1982) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
if (RsDiff < 2) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
return 8;
}
}
else if (RsDiff >= 2) {
if (tcor < 86) {
if (mcTcor < 84) {
if (negBalance < 44) {
if (Rs < 672) {
return 8;
}
else if (Rs >= 672) {
if (Rs < 698) {
return 2;
}
else if (Rs >= 698) {
return 1;
}
}
}
else if (negBalance >= 44) {
if (mcTcor < 77) {
if (Rs < 846) {
return 8;
}
else if (Rs >= 846) {
return 1;
}
}
else if (mcTcor >= 77) {
return 8;
}
}
}
else if (mcTcor >= 84) {
if (negBalance < 48) {
if (mu_mv_mag_sq < 56) {
if (MVDiff < 23) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 4;
}
}
else if (MVDiff >= 23) {
return 2;
}
}
else if (mu_mv_mag_sq >= 56) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
return 1;
}
}
}
}
else if (negBalance >= 48) {
if (SC < 1480) {
return 4;
}
else if (SC >= 1480) {
return 1;
}
}
}
}
else if (tcor >= 86) {
if (mu_mv_mag_sq < 41) {
if (gchDC < 1) {
if (Cs < 980) {
return 8;
}
else if (Cs >= 980) {
return 4;
}
}
else if (gchDC >= 1) {
if (TSC < 5) {
if (Rs < 783) {
return 8;
}
else if (Rs >= 783) {
return 4;
}
}
else if (TSC >= 5) {
if (refDCval < 54) {
if (SC < 1649) {
return 2;
}
else if (SC >= 1649) {
return 8;
}
}
else if (refDCval >= 54) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (mu_mv_mag_sq < 56) {
if (Rs < 756) {
if (Rs < 691) {
return 1;
}
else if (Rs >= 691) {
return 2;
}
}
else if (Rs >= 756) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 68) {
if (negBalance < 38) {
if (Rs < 777) {
if (Cs < 567) {
return 4;
}
else if (Cs >= 567) {
return 1;
}
}
else if (Rs >= 777) {
if (Cs < 787) {
return 4;
}
else if (Cs >= 787) {
return 2;
}
}
}
else if (negBalance >= 38) {
if (diffRsCsdiff < 9) {
if (ssDCval < 69) {
if (Rs < 842) {
return 1;
}
else if (Rs >= 842) {
return 4;
}
}
else if (ssDCval >= 69) {
if (posBalance < 54) {
if (Rs < 831) {
return 4;
}
else if (Rs >= 831) {
return 1;
}
}
else if (posBalance >= 54) {
if (mu_mv_mag_sq < 79) {
if (mu_mv_mag_sq < 72) {
return 8;
}
else if (mu_mv_mag_sq >= 72) {
return 1;
}
}
else if (mu_mv_mag_sq >= 79) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 9) {
return 2;
}
}
}
}
}
}
else if (AFD >= 11) {
if (refDCval < 58) {
if (negBalance < 44) {
if (Cs < 729) {
return 1;
}
else if (Cs >= 729) {
if (SC < 1599) {
if (Rs < 737) {
return 2;
}
else if (Rs >= 737) {
return 8;
}
}
else if (SC >= 1599) {
return 4;
}
}
}
else if (negBalance >= 44) {
if (mcTcor < 87) {
if (TSC < 9) {
if (diffRsCsdiff < -10) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
return 4;
}
}
else if (diffRsCsdiff >= -10) {
return 4;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (mcTcor >= 87) {
if (RsCsDiff < 19) {
return 4;
}
else if (RsCsDiff >= 19) {
return 8;
}
}
}
}
else if (refDCval >= 58) {
if (refDCval < 69) {
if (TSC < 6) {
if (Rs < 800) {
return 8;
}
else if (Rs >= 800) {
return 2;
}
}
else if (TSC >= 6) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (refDCval >= 69) {
if (MVDiff < 31) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (Rs < 659) {
return 8;
}
else if (Rs >= 659) {
return 2;
}
}
}
else if (MVDiff >= 31) {
return 4;
}
}
}
}
}
}
else if (CsDiff >= 6) {
if (negBalance < 50) {
if (RsCsDiff < 63) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (RsCsDiff >= 63) {
if (MVDiff < 26) {
if (Cs < 1305) {
return 2;
}
else if (Cs >= 1305) {
return 4;
}
}
else if (MVDiff >= 26) {
return 4;
}
}
}
else if (negBalance >= 50) {
if (Cs < 829) {
return 8;
}
else if (Cs >= 829) {
if (mcTcor < 66) {
return 2;
}
else if (mcTcor >= 66) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (Rs < 400) {
if (MVDiff < 4) {
if (Cs < 758) {
if (Cs < 745) {
if (Cs < 725) {
return 1;
}
else if (Cs >= 725) {
if (Rs < 373) {
return 1;
}
else if (Rs >= 373) {
return 4;
}
}
}
else if (Cs >= 745) {
return 8;
}
}
else if (Cs >= 758) {
return 4;
}
}
else if (MVDiff >= 4) {
if (AFD < 9) {
if (Rs < 334) {
return 2;
}
else if (Rs >= 334) {
return 8;
}
}
else if (AFD >= 9) {
if (ssDCval < 109) {
return 1;
}
else if (ssDCval >= 109) {
if (Rs < 282) {
return 8;
}
else if (Rs >= 282) {
return 2;
}
}
}
}
}
else if (Rs >= 400) {
if (MVDiff < 5) {
if (Rs < 457) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
if (SC < 1240) {
if (Rs < 407) {
if (Rs < 405) {
return 1;
}
else if (Rs >= 405) {
return 4;
}
}
else if (Rs >= 407) {
return 1;
}
}
else if (SC >= 1240) {
if (Rs < 430) {
return 4;
}
else if (Rs >= 430) {
return 8;
}
}
}
}
else if (Rs >= 457) {
if (Cs < 971) {
if (mcTcor < 95) {
if (refDCval < 90) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (negBalance < 27) {
return 2;
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (refDCval >= 90) {
if (mu_mv_mag_sq < 97) {
if (Rs < 610) {
if (Cs < 593) {
return 1;
}
else if (Cs >= 593) {
return 8;
}
}
else if (Rs >= 610) {
if (Rs < 752) {
return 2;
}
else if (Rs >= 752) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 97) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (negBalance < 43) {
if (mu_mv_mag_sq < 82) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
else if (mu_mv_mag_sq >= 82) {
if (Rs < 674) {
if (Rs < 461) {
return 2;
}
else if (Rs >= 461) {
if (Rs < 588) {
return 8;
}
else if (Rs >= 588) {
if (Rs < 601) {
if (Cs < 833) {
return 2;
}
else if (Cs >= 833) {
return 8;
}
}
else if (Rs >= 601) {
return 8;
}
}
}
}
else if (Rs >= 674) {
if (ssDCval < 88) {
return 2;
}
else if (ssDCval >= 88) {
if (Rs < 826) {
return 1;
}
else if (Rs >= 826) {
return 8;
}
}
}
}
}
else if (negBalance >= 43) {
if (ssDCval < 105) {
if (TSC < 6) {
if (posBalance < 53) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (TSC >= 6) {
return 1;
}
}
else if (ssDCval >= 105) {
if (Cs < 546) {
return 2;
}
else if (Cs >= 546) {
return 1;
}
}
}
}
}
else if (Cs >= 971) {
if (ssDCval < 93) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (ssDCval >= 93) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
return 8;
}
}
}
}
}
else if (MVDiff >= 5) {
if (AFD < 10) {
if (Cs < 1120) {
if (posBalance < 52) {
if (tcor < 91) {
if (negBalance < 52) {
if (Rs < 628) {
return 2;
}
else if (Rs >= 628) {
if (Rs < 702) {
return 8;
}
else if (Rs >= 702) {
return 1;
}
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (tcor >= 91) {
return 8;
}
}
else if (posBalance >= 52) {
if (mcTcor < 80) {
if (Rs < 764) {
return 2;
}
else if (Rs >= 764) {
return 8;
}
}
else if (mcTcor >= 80) {
if (negBalance < 32) {
if (posBalance < 60) {
if (Cs < 740) {
return 2;
}
else if (Cs >= 740) {
return 8;
}
}
else if (posBalance >= 60) {
return 8;
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
}
else if (Cs >= 1120) {
if (posBalance < 51) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (AFD >= 10) {
if (ssDCval < 126) {
if (TSC < 6) {
if (refDCval < 86) {
if (negBalance < 50) {
return 4;
}
else if (negBalance >= 50) {
if (ssDCval < 77) {
return 1;
}
else if (ssDCval >= 77) {
return 8;
}
}
}
else if (refDCval >= 86) {
if (Rs < 794) {
return 2;
}
else if (Rs >= 794) {
return 1;
}
}
}
else if (TSC >= 6) {
if (Cs < 745) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 2;
}
}
}
else if (Cs >= 745) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
if (diffRsCsdiff < -20) {
if (Rs < 677) {
return 4;
}
else if (Rs >= 677) {
return 2;
}
}
else if (diffRsCsdiff >= -20) {
return 8;
}
}
}
}
}
else if (ssDCval >= 126) {
if (Cs < 873) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
else if (Cs >= 873) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 903) {
if (mcTcor < 87) {
if (Rs < 1290) {
if (ssDCval < 45) {
if (diffRsCsdiff < 30) {
if (AFD < 13) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 119) {
return 8;
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 1157) {
return 1;
}
else if (Rs >= 1157) {
return 8;
}
}
}
}
else if (AFD >= 13) {
return 2;
}
}
else if (diffRsCsdiff >= 30) {
if (Rs < 1034) {
return 1;
}
else if (Rs >= 1034) {
if (Rs < 1192) {
return 8;
}
else if (Rs >= 1192) {
return 2;
}
}
}
}
else if (ssDCval >= 45) {
if (SC < 1894) {
if (mu_mv_mag_sq < 72) {
if (diffRsCsdiff < -8) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (Cs < 706) {
if (Rs < 1111) {
return 8;
}
else if (Rs >= 1111) {
return 2;
}
}
else if (Cs >= 706) {
return 1;
}
}
}
else if (diffRsCsdiff >= -8) {
if (Rs < 1239) {
return 8;
}
else if (Rs >= 1239) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (MVDiff < 6) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
else if (MVDiff >= 6) {
if (Rs < 1235) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 4;
}
}
else if (Rs >= 1235) {
return 8;
}
}
}
}
else if (SC >= 1894) {
if (tcor < 74) {
if (ssDCval < 59) {
if (AFD < 13) {
if (mu_mv_mag_sq < 77) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (Rs < 1068) {
return 4;
}
else if (Rs >= 1068) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 1224) {
return 4;
}
else if (Rs >= 1224) {
return 1;
}
}
}
else if (AFD >= 13) {
if (Rs < 1277) {
if (SC < 2451) {
return 4;
}
else if (SC >= 2451) {
return 2;
}
}
else if (Rs >= 1277) {
return 8;
}
}
}
else if (ssDCval >= 59) {
if (MVDiff < 8) {
if (Rs < 1252) {
return 8;
}
else if (Rs >= 1252) {
return 4;
}
}
else if (MVDiff >= 8) {
if (Rs < 1267) {
if (Rs < 1108) {
return 4;
}
else if (Rs >= 1108) {
return 1;
}
}
else if (Rs >= 1267) {
return 8;
}
}
}
}
else if (tcor >= 74) {
if (negBalance < 45) {
if (MVDiff < 9) {
if (Cs < 936) {
if (SC < 1938) {
return 8;
}
else if (SC >= 1938) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (Cs >= 936) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 40) {
return 1;
}
else if (mu_mv_mag_sq >= 40) {
return 8;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
}
else if (MVDiff >= 9) {
if (Cs < 1016) {
return 8;
}
else if (Cs >= 1016) {
return 4;
}
}
}
else if (negBalance >= 45) {
if (tcor < 75) {
if (Rs < 1140) {
return 1;
}
else if (Rs >= 1140) {
return 8;
}
}
else if (tcor >= 75) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1290) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 59) {
if (CsDiff < 4) {
if (SC < 2422) {
return 2;
}
else if (SC >= 2422) {
if (refDCval < 98) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
return 2;
}
}
else if (refDCval >= 98) {
return 2;
}
}
}
else if (CsDiff >= 4) {
if (negBalance < 47) {
if (Rs < 1329) {
return 8;
}
else if (Rs >= 1329) {
return 1;
}
}
else if (negBalance >= 47) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
}
}
else if (posBalance >= 59) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 45) {
if (TSC < 7) {
if (TSC < 5) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
return 4;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (TSC >= 7) {
if (MVDiff < 6) {
if (Rs < 1387) {
return 4;
}
else if (Rs >= 1387) {
return 8;
}
}
else if (MVDiff >= 6) {
if (MVDiff < 17) {
if (Cs < 1305) {
return 1;
}
else if (Cs >= 1305) {
return 8;
}
}
else if (MVDiff >= 17) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (Cs < 544) {
return 4;
}
else if (Cs >= 544) {
if (SC < 2896) {
return 8;
}
else if (SC >= 2896) {
if (SC < 2905) {
return 4;
}
else if (SC >= 2905) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 87) {
if (RsCsDiff < 19) {
if (ssDCval < 125) {
if (RsCsDiff < 7) {
if (negBalance < 28) {
if (Rs < 1235) {
if (refDCval < 57) {
if (diffMVdiffVal < -1) {
if (Rs < 1145) {
return 8;
}
else if (Rs >= 1145) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (refDCval >= 57) {
if (SC < 1871) {
if (MVDiff < 17) {
return 8;
}
else if (MVDiff >= 17) {
if (Rs < 970) {
return 8;
}
else if (Rs >= 970) {
return 4;
}
}
}
else if (SC >= 1871) {
if (Rs < 1132) {
return 8;
}
else if (Rs >= 1132) {
return 1;
}
}
}
}
else if (Rs >= 1235) {
return 8;
}
}
else if (negBalance >= 28) {
if (mcTcor < 95) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
if (Rs < 1124) {
return 4;
}
else if (Rs >= 1124) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (Cs < 549) {
return 1;
}
else if (Cs >= 549) {
if (MVDiff < 32) {
if (tcor < 97) {
if (diffRsCsdiff < -14) {
return 1;
}
else if (diffRsCsdiff >= -14) {
if (CsDiff < 2) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
if (Cs < 680) {
return 1;
}
else if (Cs >= 680) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (AFD < 4) {
if (Cs < 1198) {
return 1;
}
else if (Cs >= 1198) {
return 8;
}
}
else if (AFD >= 4) {
return 8;
}
}
}
}
else if (tcor >= 97) {
return 1;
}
}
else if (MVDiff >= 32) {
if (Rs < 1483) {
return 8;
}
else if (Rs >= 1483) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 7) {
if (Rs < 970) {
if (SC < 2094) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (RsCsDiff < 16) {
if (mu_mv_mag_sq < 79) {
if (MVDiff < 10) {
if (MVDiff < 8) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (MVDiff >= 10) {
if (Cs < 547) {
return 4;
}
else if (Cs >= 547) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (Rs < 915) {
return 8;
}
else if (Rs >= 915) {
return 4;
}
}
}
else if (RsCsDiff >= 16) {
if (Rs < 911) {
return 8;
}
else if (Rs >= 911) {
return 4;
}
}
}
}
}
else if (SC >= 2094) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (Rs >= 970) {
if (posBalance < 56) {
if (ssDCval < 75) {
if (refDCval < 63) {
if (MVDiff < 12) {
if (Rs < 1307) {
if (ssDCval < 59) {
if (Cs < 931) {
if (negBalance < 39) {
if (Rs < 1164) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 63) {
return 1;
}
else if (mu_mv_mag_sq >= 63) {
return 8;
}
}
else if (posBalance >= 49) {
if (SC < 1997) {
return 8;
}
else if (SC >= 1997) {
if (Rs < 1125) {
return 1;
}
else if (Rs >= 1125) {
return 8;
}
}
}
}
else if (Rs >= 1164) {
return 1;
}
}
else if (negBalance >= 39) {
return 8;
}
}
else if (Cs >= 931) {
if (Rs < 1078) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (Rs >= 1078) {
return 8;
}
}
}
else if (ssDCval >= 59) {
if (Cs < 1220) {
if (posBalance < 52) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (posBalance >= 52) {
if (Rs < 1111) {
return 4;
}
else if (Rs >= 1111) {
return 1;
}
}
}
else if (Cs >= 1220) {
return 8;
}
}
}
else if (Rs >= 1307) {
if (Cs < 928) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Rs < 1354) {
return 8;
}
else if (Rs >= 1354) {
if (Rs < 1520) {
return 4;
}
else if (Rs >= 1520) {
if (Rs < 1801) {
if (Rs < 1642) {
return 4;
}
else if (Rs >= 1642) {
return 4;
}
}
else if (Rs >= 1801) {
return 8;
}
}
}
}
}
else if (Cs >= 928) {
if (negBalance < 23) {
if (Rs < 1848) {
return 1;
}
else if (Rs >= 1848) {
return 4;
}
}
else if (negBalance >= 23) {
return 8;
}
}
}
}
else if (MVDiff >= 12) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Cs < 913) {
return 2;
}
else if (Cs >= 913) {
return 8;
}
}
}
}
else if (refDCval >= 63) {
if (Rs < 1350) {
if (refDCval < 67) {
if (diffRsCsdiff < 6) {
if (mu_mv_mag_sq < 114) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (diffAFD < 1) {
if (SC < 2206) {
if (SC < 1778) {
if (Rs < 1174) {
return 8;
}
else if (Rs >= 1174) {
return 4;
}
}
else if (SC >= 1778) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
return 4;
}
}
}
else if (SC >= 2206) {
if (Rs < 1339) {
return 8;
}
else if (Rs >= 1339) {
return 1;
}
}
}
else if (diffAFD >= 1) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
if (Rs < 1254) {
return 8;
}
else if (Rs >= 1254) {
return 1;
}
}
}
else if (refDCval >= 67) {
if (mu_mv_mag_sq < 116) {
if (Cs < 942) {
if (MVDiff < 20) {
return 1;
}
else if (MVDiff >= 20) {
return 4;
}
}
else if (Cs >= 942) {
if (Cs < 1167) {
return 8;
}
else if (Cs >= 1167) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 116) {
if (Cs < 1175) {
return 4;
}
else if (Cs >= 1175) {
return 1;
}
}
}
}
else if (Rs >= 1350) {
if (RsCsDiff < 16) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
if (Rs < 1452) {
return 1;
}
else if (Rs >= 1452) {
if (Rs < 1710) {
return 8;
}
else if (Rs >= 1710) {
return 1;
}
}
}
}
else if (RsCsDiff >= 16) {
if (Rs < 1533) {
if (Cs < 1216) {
return 4;
}
else if (Cs >= 1216) {
return 8;
}
}
else if (Rs >= 1533) {
if (Rs < 1692) {
if (Cs < 1022) {
return 2;
}
else if (Cs >= 1022) {
return 8;
}
}
else if (Rs >= 1692) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 75) {
if (mu_mv_mag_sq < 64) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (Rs < 1797) {
return 8;
}
else if (Rs >= 1797) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
if (Rs < 1850) {
return 2;
}
else if (Rs >= 1850) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 56) {
if (negBalance < 43) {
if (RsCsDiff < 18) {
if (Rs < 1069) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (Rs >= 1069) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
return 8;
}
}
}
else if (RsCsDiff >= 18) {
if (ssDCval < 63) {
if (diffRsCsdiff < -19) {
return 2;
}
else if (diffRsCsdiff >= -19) {
if (diffRsCsdiff < 8) {
if (Cs < 792) {
if (Rs < 1174) {
return 8;
}
else if (Rs >= 1174) {
return 2;
}
}
else if (Cs >= 792) {
return 8;
}
}
else if (diffRsCsdiff >= 8) {
if (Rs < 1302) {
return 8;
}
else if (Rs >= 1302) {
return 2;
}
}
}
}
else if (ssDCval >= 63) {
if (ssDCval < 78) {
if (negBalance < 36) {
if (Rs < 1690) {
return 8;
}
else if (Rs >= 1690) {
return 1;
}
}
else if (negBalance >= 36) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
if (Rs < 1878) {
return 4;
}
else if (Rs >= 1878) {
return 8;
}
}
}
}
else if (ssDCval >= 78) {
return 8;
}
}
}
}
else if (negBalance >= 43) {
if (Rs < 977) {
if (Rs < 973) {
return 1;
}
else if (Rs >= 973) {
if (Cs < 775) {
return 8;
}
else if (Cs >= 775) {
return 4;
}
}
}
else if (Rs >= 977) {
if (TSC < 6) {
if (diffTSC < 1) {
if (MVDiff < 37) {
if (SC < 2615) {
return 8;
}
else if (SC >= 2615) {
if (Rs < 1460) {
return 8;
}
else if (Rs >= 1460) {
return 4;
}
}
}
else if (MVDiff >= 37) {
if (Cs < 725) {
return 4;
}
else if (Cs >= 725) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (Rs < 1215) {
return 8;
}
else if (Rs >= 1215) {
if (Rs < 1311) {
return 1;
}
else if (Rs >= 1311) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (refDCval < 116) {
return 8;
}
else if (refDCval >= 116) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 125) {
if (MVDiff < 20) {
if (ssDCval < 132) {
if (RsCsDiff < 16) {
return 4;
}
else if (RsCsDiff >= 16) {
if (Cs < 891) {
return 8;
}
else if (Cs >= 891) {
return 1;
}
}
}
else if (ssDCval >= 132) {
if (posBalance < 67) {
if (Rs < 911) {
return 1;
}
else if (Rs >= 911) {
return 8;
}
}
else if (posBalance >= 67) {
return 1;
}
}
}
else if (MVDiff >= 20) {
if (Cs < 634) {
return 8;
}
else if (Cs >= 634) {
return 4;
}
}
}
}
else if (RsCsDiff >= 19) {
if (ssDCval < 73) {
if (ssDCval < 63) {
if (negBalance < 41) {
if (Cs < 894) {
if (AFD < 9) {
if (RsCsDiff < 31) {
if (refDCval < 45) {
if (posBalance < 49) {
if (Rs < 1306) {
return 8;
}
else if (Rs >= 1306) {
return 4;
}
}
else if (posBalance >= 49) {
return 2;
}
}
else if (refDCval >= 45) {
if (SC < 1982) {
if (TSC < 5) {
if (Cs < 665) {
return 8;
}
else if (Cs >= 665) {
return 1;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (SC >= 1982) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 4;
}
}
}
}
else if (RsCsDiff >= 31) {
if (RsDiff < 6) {
if (Rs < 1015) {
return 2;
}
else if (Rs >= 1015) {
return 4;
}
}
else if (RsDiff >= 6) {
if (Rs < 1743) {
return 1;
}
else if (Rs >= 1743) {
return 2;
}
}
}
}
else if (AFD >= 9) {
if (Rs < 1964) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (Rs < 1612) {
return 8;
}
else if (Rs >= 1612) {
return 2;
}
}
}
else if (Rs >= 1964) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 4;
}
}
}
}
else if (Cs >= 894) {
if (refDCval < 58) {
if (negBalance < 30) {
if (mu_mv_mag_sq < 101) {
if (Cs < 1270) {
if (diffAFD < -6) {
return 1;
}
else if (diffAFD >= -6) {
return 4;
}
}
else if (Cs >= 1270) {
return 2;
}
}
else if (mu_mv_mag_sq >= 101) {
return 8;
}
}
else if (negBalance >= 30) {
if (Rs < 1152) {
if (refDCval < 47) {
return 8;
}
else if (refDCval >= 47) {
if (AFD < 10) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (Rs >= 1152) {
if (Rs < 1987) {
if (MVDiff < 9) {
if (diffMVdiffVal < -5) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (MVDiff >= 9) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
return 8;
}
}
}
else if (Rs >= 1987) {
if (Rs < 2180) {
return 4;
}
else if (Rs >= 2180) {
return 8;
}
}
}
}
}
else if (refDCval >= 58) {
if (mu_mv_mag_sq < 2) {
if (Cs < 1273) {
return 8;
}
else if (Cs >= 1273) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
}
}
else if (negBalance >= 41) {
if (SC < 2562) {
if (MVDiff < 15) {
if (Cs < 978) {
if (MVDiff < 8) {
if (ssDCval < 47) {
if (SC < 2159) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
return 1;
}
}
else if (SC >= 2159) {
return 2;
}
}
else if (ssDCval >= 47) {
return 8;
}
}
else if (MVDiff >= 8) {
if (AFD < 10) {
if (ssDCval < 59) {
if (Cs < 918) {
return 8;
}
else if (Cs >= 918) {
return 4;
}
}
else if (ssDCval >= 59) {
return 4;
}
}
else if (AFD >= 10) {
if (Rs < 1564) {
return 8;
}
else if (Rs >= 1564) {
return 2;
}
}
}
}
else if (Cs >= 978) {
if (Rs < 1178) {
if (ssDCval < 49) {
if (Cs < 1162) {
return 8;
}
else if (Cs >= 1162) {
if (Rs < 1131) {
return 4;
}
else if (Rs >= 1131) {
return 8;
}
}
}
else if (ssDCval >= 49) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
else if (Rs >= 1178) {
if (AFD < 11) {
if (Rs < 1275) {
if (Cs < 1108) {
return 8;
}
else if (Cs >= 1108) {
if (Rs < 1188) {
return 8;
}
else if (Rs >= 1188) {
return 4;
}
}
}
else if (Rs >= 1275) {
return 8;
}
}
else if (AFD >= 11) {
return 4;
}
}
}
}
else if (MVDiff >= 15) {
if (diffRsCsdiff < 12) {
if (tcor < 85) {
if (Rs < 1275) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
return 2;
}
}
else if (Rs >= 1275) {
return 8;
}
}
else if (tcor >= 85) {
if (Rs < 1723) {
return 4;
}
else if (Rs >= 1723) {
return 2;
}
}
}
else if (diffRsCsdiff >= 12) {
if (Rs < 1034) {
return 2;
}
else if (Rs >= 1034) {
return 8;
}
}
}
}
else if (SC >= 2562) {
if (Cs < 1315) {
if (refDCval < 62) {
if (tcor < 85) {
if (Cs < 1266) {
if (refDCval < 59) {
return 8;
}
else if (refDCval >= 59) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
else if (Cs >= 1266) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 4;
}
}
}
else if (tcor >= 85) {
return 8;
}
}
else if (refDCval >= 62) {
if (Cs < 1044) {
return 2;
}
else if (Cs >= 1044) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
}
else if (Cs >= 1315) {
return 1;
}
}
}
}
else if (ssDCval >= 63) {
if (Rs < 1602) {
if (AFD < 9) {
if (MVDiff < 4) {
if (diffRsCsdiff < 15) {
if (diffMVdiffVal < -1) {
if (ssDCval < 68) {
if (Rs < 1563) {
return 8;
}
else if (Rs >= 1563) {
return 2;
}
}
else if (ssDCval >= 68) {
if (tcor < 92) {
if (Rs < 1164) {
return 8;
}
else if (Rs >= 1164) {
return 1;
}
}
else if (tcor >= 92) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (TSC < 6) {
if (negBalance < 41) {
if (ssDCval < 69) {
if (Rs < 1078) {
return 8;
}
else if (Rs >= 1078) {
if (Rs < 1155) {
return 4;
}
else if (Rs >= 1155) {
return 1;
}
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (negBalance >= 41) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (diffRsCsdiff >= 15) {
if (negBalance < 44) {
return 4;
}
else if (negBalance >= 44) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (mcTcor < 96) {
if (Rs < 1001) {
return 1;
}
else if (Rs >= 1001) {
if (mu_mv_mag_sq < 14) {
if (Cs < 809) {
return 4;
}
else if (Cs >= 809) {
return 1;
}
}
else if (mu_mv_mag_sq >= 14) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 1316) {
return 8;
}
else if (Rs >= 1316) {
return 1;
}
}
}
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 3) {
if (TSC < 8) {
if (negBalance < 46) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
else if (negBalance >= 46) {
if (Rs < 1063) {
if (Rs < 952) {
return 2;
}
else if (Rs >= 952) {
return 1;
}
}
else if (Rs >= 1063) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
else if (TSC >= 8) {
if (Rs < 1134) {
if (Rs < 1107) {
return 4;
}
else if (Rs >= 1107) {
return 2;
}
}
else if (Rs >= 1134) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 49) {
if (refDCval < 65) {
return 8;
}
else if (refDCval >= 65) {
if (Cs < 787) {
return 1;
}
else if (Cs >= 787) {
if (Cs < 1151) {
if (Rs < 914) {
return 1;
}
else if (Rs >= 914) {
return 8;
}
}
else if (Cs >= 1151) {
if (mu_mv_mag_sq < 43) {
return 1;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 952) {
return 8;
}
else if (Rs >= 952) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 49) {
return 8;
}
}
}
}
else if (Rs >= 1602) {
if (Cs < 1143) {
if (posBalance < 53) {
if (negBalance < 37) {
if (tcor < 94) {
if (tcor < 83) {
return 2;
}
else if (tcor >= 83) {
if (Cs < 605) {
return 2;
}
else if (Cs >= 605) {
return 1;
}
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (negBalance >= 37) {
if (mcTcor < 95) {
if (MVDiff < 24) {
return 4;
}
else if (MVDiff >= 24) {
if (Rs < 1703) {
return 2;
}
else if (Rs >= 1703) {
return 8;
}
}
}
else if (mcTcor >= 95) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (RsCsDiff < 60) {
if (mcTcor < 91) {
if (Cs < 748) {
return 2;
}
else if (Cs >= 748) {
return 8;
}
}
else if (mcTcor >= 91) {
if (RsCsDiff < 27) {
if (Rs < 1729) {
return 8;
}
else if (Rs >= 1729) {
return 2;
}
}
else if (RsCsDiff >= 27) {
return 2;
}
}
}
else if (RsCsDiff >= 60) {
if (Rs < 1764) {
return 8;
}
else if (Rs >= 1764) {
return 1;
}
}
}
}
else if (Cs >= 1143) {
if (mu_mv_mag_sq < 80) {
if (mu_mv_mag_sq < 48) {
if (Rs < 1760) {
return 8;
}
else if (Rs >= 1760) {
return 1;
}
}
else if (mu_mv_mag_sq >= 48) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 73) {
if (Cs < 783) {
if (RsDiff < 5) {
if (negBalance < 40) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 1309) {
return 4;
}
else if (Rs >= 1309) {
return 1;
}
}
}
else if (negBalance >= 40) {
if (negBalance < 57) {
if (Rs < 1868) {
if (diffAFD < 2) {
if (diffRsCsdiff < -12) {
if (Rs < 1089) {
return 4;
}
else if (Rs >= 1089) {
return 8;
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (Rs >= 1868) {
return 2;
}
}
else if (negBalance >= 57) {
if (Rs < 1092) {
return 2;
}
else if (Rs >= 1092) {
return 1;
}
}
}
}
else if (RsDiff >= 5) {
if (ssDCval < 116) {
if (tcor < 85) {
if (negBalance < 50) {
if (Rs < 1392) {
return 2;
}
else if (Rs >= 1392) {
return 1;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (tcor >= 85) {
if (Rs < 1035) {
if (Rs < 958) {
return 4;
}
else if (Rs >= 958) {
return 8;
}
}
else if (Rs >= 1035) {
if (refDCval < 91) {
if (Rs < 1359) {
if (Rs < 1148) {
return 1;
}
else if (Rs >= 1148) {
if (Rs < 1209) {
return 4;
}
else if (Rs >= 1209) {
return 8;
}
}
}
else if (Rs >= 1359) {
return 1;
}
}
else if (refDCval >= 91) {
if (Rs < 1784) {
if (RsDiff < 6) {
if (Cs < 766) {
return 8;
}
else if (Cs >= 766) {
return 1;
}
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (Rs >= 1784) {
if (Cs < 771) {
return 2;
}
else if (Cs >= 771) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 116) {
if (MVDiff < 32) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
else if (MVDiff >= 32) {
return 4;
}
}
}
}
else if (Cs >= 783) {
if (MVDiff < 2) {
if (tcor < 92) {
if (negBalance < 41) {
if (RsCsDiff < 38) {
return 8;
}
else if (RsCsDiff >= 38) {
return 2;
}
}
else if (negBalance >= 41) {
return 8;
}
}
else if (tcor >= 92) {
if (negBalance < 41) {
if (Cs < 1155) {
return 8;
}
else if (Cs >= 1155) {
if (ssDCval < 84) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 1;
}
}
else if (ssDCval >= 84) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (Rs < 1742) {
if (SC < 2617) {
if (Rs < 1282) {
return 4;
}
else if (Rs >= 1282) {
return 1;
}
}
else if (SC >= 2617) {
return 8;
}
}
else if (Rs >= 1742) {
if (diffRsCsdiff < -18) {
if (Cs < 1103) {
return 1;
}
else if (Cs >= 1103) {
return 8;
}
}
else if (diffRsCsdiff >= -18) {
return 1;
}
}
}
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 33) {
if (negBalance < 35) {
if (Rs < 1216) {
return 1;
}
else if (Rs >= 1216) {
return 2;
}
}
else if (negBalance >= 35) {
if (MVDiff < 27) {
if (posBalance < 53) {
if (Cs < 1123) {
if (MVDiff < 8) {
if (Cs < 1080) {
return 1;
}
else if (Cs >= 1080) {
return 8;
}
}
else if (MVDiff >= 8) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
if (Cs < 1058) {
return 8;
}
else if (Cs >= 1058) {
return 4;
}
}
}
}
else if (Cs >= 1123) {
if (Rs < 1072) {
return 1;
}
else if (Rs >= 1072) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
if (Rs < 1352) {
if (negBalance < 43) {
if (CsDiff < 5) {
if (Rs < 1269) {
return 4;
}
else if (Rs >= 1269) {
return 1;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (negBalance >= 43) {
if (ssDCval < 128) {
if (MVDiff < 12) {
if (diffRsCsdiff < 15) {
if (posBalance < 67) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
if (posBalance < 55) {
if (Rs < 1195) {
return 4;
}
else if (Rs >= 1195) {
return 8;
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (posBalance >= 67) {
if (Rs < 1157) {
return 2;
}
else if (Rs >= 1157) {
return 8;
}
}
}
else if (diffRsCsdiff >= 15) {
return 1;
}
}
else if (MVDiff >= 12) {
if (Rs < 1020) {
return 4;
}
else if (Rs >= 1020) {
if (gchDC < 2) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
return 4;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
else if (ssDCval >= 128) {
return 1;
}
}
}
else if (Rs >= 1352) {
if (negBalance < 43) {
if (MVDiff < 6) {
if (Rs < 1612) {
return 8;
}
else if (Rs >= 1612) {
return 2;
}
}
else if (MVDiff >= 6) {
if (ssDCval < 101) {
if (Rs < 1802) {
return 8;
}
else if (Rs >= 1802) {
if (Rs < 2058) {
return 2;
}
else if (Rs >= 2058) {
return 4;
}
}
}
else if (ssDCval >= 101) {
return 8;
}
}
}
else if (negBalance >= 43) {
if (mcTcor < 89) {
if (Rs < 1646) {
return 8;
}
else if (Rs >= 1646) {
return 4;
}
}
else if (mcTcor >= 89) {
if (SC < 3577) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (Cs < 1246) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
return 4;
}
}
else if (Cs >= 1246) {
return 2;
}
}
}
else if (SC >= 3577) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 27) {
if (RsCsDiff < 43) {
return 4;
}
else if (RsCsDiff >= 43) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Rs < 1318) {
return 4;
}
else if (Rs >= 1318) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (refDCval < 100) {
if (refDCval < 85) {
if (mcTcor < 94) {
if (MVDiff < 29) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
return 8;
}
}
else if (MVDiff >= 29) {
return 1;
}
}
else if (mcTcor >= 94) {
if (Rs < 1647) {
if (RsDiff < 7) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
if (Rs < 1235) {
return 4;
}
else if (Rs >= 1235) {
return 8;
}
}
}
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (Rs >= 1647) {
if (Rs < 2123) {
if (Cs < 968) {
return 1;
}
else if (Cs >= 968) {
if (Cs < 1317) {
return 8;
}
else if (Cs >= 1317) {
return 1;
}
}
}
else if (Rs >= 2123) {
if (Cs < 1037) {
return 2;
}
else if (Cs >= 1037) {
return 1;
}
}
}
}
}
else if (refDCval >= 85) {
if (diffAFD < -2) {
if (posBalance < 59) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 2214) {
return 2;
}
else if (Rs >= 2214) {
return 8;
}
}
}
else if (posBalance >= 59) {
return 4;
}
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < 8) {
if (AFD < 8) {
if (Rs < 1791) {
return 2;
}
else if (Rs >= 1791) {
if (Rs < 2043) {
return 8;
}
else if (Rs >= 2043) {
return 1;
}
}
}
else if (AFD >= 8) {
if (negBalance < 41) {
if (mu_mv_mag_sq < 58) {
if (Rs < 1647) {
return 8;
}
else if (Rs >= 1647) {
return 4;
}
}
else if (mu_mv_mag_sq >= 58) {
return 8;
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
else if (diffRsCsdiff >= 8) {
if (tcor < 93) {
if (RsCsDiff < 51) {
return 8;
}
else if (RsCsDiff >= 51) {
if (TSC < 8) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (tcor >= 93) {
if (AFD < 8) {
if (Rs < 1524) {
return 8;
}
else if (Rs >= 1524) {
return 2;
}
}
else if (AFD >= 8) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 100) {
if (negBalance < 49) {
if (MVDiff < 25) {
if (diffRsCsdiff < 26) {
return 8;
}
else if (diffRsCsdiff >= 26) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 2;
}
}
}
else if (MVDiff >= 25) {
if (RsCsDiff < 37) {
return 4;
}
else if (RsCsDiff >= 37) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (MVDiff < 13) {
return 1;
}
else if (MVDiff >= 13) {
if (Rs < 1119) {
return 1;
}
else if (Rs >= 1119) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (ssDCval < 47) {
if (Rs < 774) {
return 1;
}
else if (Rs >= 774) {
return 8;
}
}
else if (ssDCval >= 47) {
if (mcTcor < 92) {
if (SC < 2292) {
if (negBalance < 50) {
if (Rs < 1011) {
return 1;
}
else if (Rs >= 1011) {
return 8;
}
}
else if (negBalance >= 50) {
if (CsDiff < 5) {
if (Cs < 637) {
if (Rs < 1543) {
return 2;
}
else if (Rs >= 1543) {
return 4;
}
}
else if (Cs >= 637) {
return 8;
}
}
else if (CsDiff >= 5) {
return 1;
}
}
}
else if (SC >= 2292) {
if (TSC < 10) {
if (MVDiff < 14) {
if (diffAFD < 2) {
if (ssDCval < 57) {
if (RsCsDiff < 77) {
return 8;
}
else if (RsCsDiff >= 77) {
return 4;
}
}
else if (ssDCval >= 57) {
if (mu_mv_mag_sq < 67) {
if (Cs < 639) {
return 2;
}
else if (Cs >= 639) {
return 8;
}
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 2073) {
return 8;
}
else if (Rs >= 2073) {
return 4;
}
}
}
}
else if (diffAFD >= 2) {
if (diffRsCsdiff < 22) {
if (Cs < 1053) {
return 1;
}
else if (Cs >= 1053) {
return 2;
}
}
else if (diffRsCsdiff >= 22) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 2;
}
}
}
}
else if (MVDiff >= 14) {
if (tcor < 73) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
if (Rs < 2306) {
return 4;
}
else if (Rs >= 2306) {
return 1;
}
}
}
else if (tcor >= 73) {
if (diffMVdiffVal < -10) {
return 8;
}
else if (diffMVdiffVal >= -10) {
if (AFD < 12) {
if (Rs < 1904) {
return 2;
}
else if (Rs >= 1904) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
if (Rs < 2072) {
return 8;
}
else if (Rs >= 2072) {
return 1;
}
}
}
}
else if (AFD >= 12) {
if (Cs < 877) {
if (MVDiff < 29) {
if (Rs < 2152) {
return 4;
}
else if (Rs >= 2152) {
return 2;
}
}
else if (MVDiff >= 29) {
return 2;
}
}
else if (Cs >= 877) {
if (Rs < 2355) {
return 2;
}
else if (Rs >= 2355) {
return 1;
}
}
}
}
}
}
}
else if (TSC >= 10) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (refDCval < 72) {
if (Rs < 1537) {
return 1;
}
else if (Rs >= 1537) {
return 2;
}
}
else if (refDCval >= 72) {
if (Cs < 852) {
if (posBalance < 57) {
if (Rs < 1730) {
return 8;
}
else if (Rs >= 1730) {
if (mcTcor < 94) {
if (Rs < 1951) {
return 2;
}
else if (Rs >= 1951) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 1852) {
if (Rs < 1817) {
return 1;
}
else if (Rs >= 1817) {
return 2;
}
}
else if (Rs >= 1852) {
return 1;
}
}
}
}
else if (posBalance >= 57) {
if (diffAFD < -4) {
if (Rs < 2265) {
return 4;
}
else if (Rs >= 2265) {
return 1;
}
}
else if (diffAFD >= -4) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
else if (Cs >= 852) {
if (RsDiff < 11) {
if (negBalance < 33) {
if (Rs < 1698) {
return 2;
}
else if (Rs >= 1698) {
return 4;
}
}
else if (negBalance >= 33) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 16) {
if (Rs < 1639) {
return 8;
}
else if (Rs >= 1639) {
return 2;
}
}
else if (MVDiff >= 16) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 3425) {
return 8;
}
else if (SC >= 3425) {
return 2;
}
}
}
}
else if (RsDiff >= 11) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 156) {
if (AFD < 10) {
if (Rs < 790) {
return 4;
}
else if (Rs >= 790) {
return 2;
}
}
else if (AFD >= 10) {
if (MVDiff < 24) {
if (Cs < 1203) {
return 8;
}
else if (Cs >= 1203) {
return 4;
}
}
else if (MVDiff >= 24) {
return 2;
}
}
}
}
}
else if (AFD >= 14) {
if (ssDCval < 114) {
if (MVDiff < 9) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 81) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Rs < 1872) {
if (negBalance < 51) {
return 1;
}
else if (negBalance >= 51) {
return 8;
}
}
else if (Rs >= 1872) {
if (Rs < 2008) {
return 1;
}
else if (Rs >= 2008) {
return 4;
}
}
}
}
else if (ssDCval >= 81) {
if (tcor < 84) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
if (AFD < 15) {
if (Rs < 1325) {
return 8;
}
else if (Rs >= 1325) {
return 4;
}
}
else if (AFD >= 15) {
if (ssDCval < 84) {
if (Rs < 1636) {
return 8;
}
else if (Rs >= 1636) {
return 4;
}
}
else if (ssDCval >= 84) {
return 4;
}
}
}
}
else if (tcor >= 84) {
if (negBalance < 50) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
else if (negBalance >= 50) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 10) {
if (ssDCval < 102) {
if (mcTcor < 93) {
if (ssDCval < 54) {
if (Rs < 1128) {
return 4;
}
else if (Rs >= 1128) {
return 8;
}
}
else if (ssDCval >= 54) {
if (Rs < 1768) {
return 8;
}
else if (Rs >= 1768) {
return 1;
}
}
}
else if (mcTcor >= 93) {
if (refDCval < 71) {
if (Rs < 503) {
return 2;
}
else if (Rs >= 503) {
return 1;
}
}
else if (refDCval >= 71) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (diffRsCsdiff < 4) {
return 4;
}
else if (diffRsCsdiff >= 4) {
if (Rs < 1627) {
return 8;
}
else if (Rs >= 1627) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (Cs < 738) {
return 1;
}
else if (Cs >= 738) {
return 4;
}
}
}
else if (RsDiff >= 10) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
if (Rs < 1572) {
return 1;
}
else if (Rs >= 1572) {
return 2;
}
}
}
}
}
else if (MVDiff >= 9) {
if (mcTcor < 85) {
if (refDCval < 100) {
if (Rs < 1032) {
if (TSC < 10) {
if (ssDCval < 46) {
if (diffAFD < 2) {
if (Rs < 535) {
return 2;
}
else if (Rs >= 535) {
return 8;
}
}
else if (diffAFD >= 2) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
return 4;
}
}
}
else if (ssDCval >= 46) {
if (AFD < 24) {
if (AFD < 18) {
if (diffMVdiffVal < 6) {
if (diffMVdiffVal < -47) {
if (Rs < 527) {
return 4;
}
else if (Rs >= 527) {
return 1;
}
}
else if (diffMVdiffVal >= -47) {
return 4;
}
}
else if (diffMVdiffVal >= 6) {
if (mcTcor < 78) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 8;
}
}
else if (mcTcor >= 78) {
if (Rs < 751) {
return 8;
}
else if (Rs >= 751) {
return 1;
}
}
}
}
else if (AFD >= 18) {
if (tcor < 38) {
if (RsCsDiff < 54) {
return 8;
}
else if (RsCsDiff >= 54) {
return 4;
}
}
else if (tcor >= 38) {
if (RsCsDiff < 82) {
if (Rs < 581) {
return 4;
}
else if (Rs >= 581) {
return 2;
}
}
else if (RsCsDiff >= 82) {
return 8;
}
}
}
}
else if (AFD >= 24) {
if (Rs < 780) {
if (Rs < 553) {
return 2;
}
else if (Rs >= 553) {
return 8;
}
}
else if (Rs >= 780) {
return 1;
}
}
}
}
else if (TSC >= 10) {
if (diffTSC < -6) {
return 1;
}
else if (diffTSC >= -6) {
if (negBalance < 60) {
if (mu_mv_mag_sq < 3) {
return 8;
}
else if (mu_mv_mag_sq >= 3) {
if (mu_mv_mag_sq < 11) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
return 4;
}
}
else if (mu_mv_mag_sq >= 11) {
if (negBalance < 46) {
if (Cs < 1103) {
if (Cs < 971) {
if (CsDiff < 7) {
return 1;
}
else if (CsDiff >= 7) {
return 2;
}
}
else if (Cs >= 971) {
return 2;
}
}
else if (Cs >= 1103) {
return 8;
}
}
else if (negBalance >= 46) {
if (Cs < 545) {
return 4;
}
else if (Cs >= 545) {
return 8;
}
}
}
}
}
else if (negBalance >= 60) {
if (gchDC < 2) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (gchDC >= 2) {
if (diffMVdiffVal < -14) {
return 4;
}
else if (diffMVdiffVal >= -14) {
if (gchDC < 22) {
return 8;
}
else if (gchDC >= 22) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 1032) {
if (RsCsDiff < 83) {
if (Cs < 695) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 1;
}
}
else if (Cs >= 695) {
if (AFD < 17) {
if (ssDCval < 48) {
if (Rs < 1085) {
return 8;
}
else if (Rs >= 1085) {
return 4;
}
}
else if (ssDCval >= 48) {
return 8;
}
}
else if (AFD >= 17) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (MVDiff < 25) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (MVDiff >= 25) {
if (SC < 2215) {
return 2;
}
else if (SC >= 2215) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 83) {
if (mu_mv_mag_sq < 43) {
if (SC < 2872) {
if (diffMVdiffVal < 6) {
if (diffMVdiffVal < 3) {
if (negBalance < 56) {
if (diffMVdiffVal < -7) {
if (Rs < 1326) {
return 8;
}
else if (Rs >= 1326) {
return 2;
}
}
else if (diffMVdiffVal >= -7) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (MVDiff < 13) {
if (Rs < 1823) {
return 8;
}
else if (Rs >= 1823) {
return 4;
}
}
else if (MVDiff >= 13) {
return 4;
}
}
}
}
else if (negBalance >= 56) {
if (SC < 2490) {
if (SC < 2425) {
return 2;
}
else if (SC >= 2425) {
return 1;
}
}
else if (SC >= 2490) {
if (mu_mv_mag_sq < 25) {
return 8;
}
else if (mu_mv_mag_sq >= 25) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= 3) {
if (SC < 2836) {
return 2;
}
else if (SC >= 2836) {
return 4;
}
}
}
else if (diffMVdiffVal >= 6) {
if (AFD < 24) {
if (posBalance < 105) {
if (RsDiff < 8) {
if (Rs < 1248) {
return 4;
}
else if (Rs >= 1248) {
return 8;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (posBalance >= 105) {
return 4;
}
}
else if (AFD >= 24) {
if (Rs < 1158) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
else if (Rs >= 1158) {
return 1;
}
}
}
}
else if (SC >= 2872) {
if (RsCsDiff < 187) {
return 8;
}
else if (RsCsDiff >= 187) {
if (refDCval < 85) {
if (Cs < 832) {
if (Rs < 2251) {
return 1;
}
else if (Rs >= 2251) {
return 8;
}
}
else if (Cs >= 832) {
if (Rs < 1882) {
return 8;
}
else if (Rs >= 1882) {
return 2;
}
}
}
else if (refDCval >= 85) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (diffAFD < 4) {
return 1;
}
else if (diffAFD >= 4) {
if (Rs < 1745) {
return 8;
}
else if (Rs >= 1745) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 100) {
if (posBalance < 58) {
if (AFD < 38) {
if (ssDCval < 73) {
if (Rs < 987) {
return 4;
}
else if (Rs >= 987) {
return 8;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (AFD >= 38) {
if (Rs < 1000) {
return 1;
}
else if (Rs >= 1000) {
return 2;
}
}
}
else if (posBalance >= 58) {
if (tcor < 72) {
return 4;
}
else if (tcor >= 72) {
if (Rs < 1674) {
return 4;
}
else if (Rs >= 1674) {
return 2;
}
}
}
}
}
else if (mcTcor >= 85) {
if (tcor < 79) {
if (posBalance < 48) {
if (RsDiff < 7) {
if (gchDC < 8) {
if (negBalance < 63) {
if (MVDiff < 40) {
return 2;
}
else if (MVDiff >= 40) {
return 1;
}
}
else if (negBalance >= 63) {
if (SC < 1042) {
return 1;
}
else if (SC >= 1042) {
return 4;
}
}
}
else if (gchDC >= 8) {
if (gchDC < 17) {
return 8;
}
else if (gchDC >= 17) {
if (Rs < 437) {
return 2;
}
else if (Rs >= 437) {
return 1;
}
}
}
}
else if (RsDiff >= 7) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (refDCval < 87) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
return 2;
}
}
else if (refDCval >= 87) {
if (diffTSC < 2) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 1;
}
}
else if (diffTSC >= 2) {
return 2;
}
}
}
}
}
else if (posBalance >= 48) {
if (AFD < 23) {
if (Rs < 1330) {
if (MVDiff < 16) {
if (TSC < 9) {
if (Rs < 449) {
if (Rs < 216) {
return 4;
}
else if (Rs >= 216) {
return 2;
}
}
else if (Rs >= 449) {
if (ssDCval < 87) {
if (TSC < 8) {
if (gchDC < 2) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (gchDC >= 2) {
if (SC < 1680) {
return 8;
}
else if (SC >= 1680) {
return 4;
}
}
}
else if (TSC >= 8) {
if (Cs < 1309) {
return 4;
}
else if (Cs >= 1309) {
return 2;
}
}
}
else if (ssDCval >= 87) {
if (Cs < 677) {
return 1;
}
else if (Cs >= 677) {
return 8;
}
}
}
}
else if (TSC >= 9) {
if (AFD < 16) {
if (Rs < 1020) {
return 8;
}
else if (Rs >= 1020) {
if (Cs < 1224) {
return 4;
}
else if (Cs >= 1224) {
return 2;
}
}
}
else if (AFD >= 16) {
return 4;
}
}
}
else if (MVDiff >= 16) {
if (Cs < 1054) {
if (mu_mv_mag_sq < 10) {
if (MVDiff < 46) {
if (diffAFD < -4) {
if (Rs < 847) {
return 1;
}
else if (Rs >= 847) {
return 4;
}
}
else if (diffAFD >= -4) {
if (diffRsCsdiff < 19) {
if (RsCsDiff < 37) {
if (MVDiff < 34) {
return 8;
}
else if (MVDiff >= 34) {
if (Rs < 843) {
return 4;
}
else if (Rs >= 843) {
return 8;
}
}
}
else if (RsCsDiff >= 37) {
if (Rs < 1262) {
if (posBalance < 65) {
return 4;
}
else if (posBalance >= 65) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
return 8;
}
}
}
else if (Rs >= 1262) {
if (Rs < 1302) {
return 8;
}
else if (Rs >= 1302) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 19) {
if (diffRsCsdiff < 26) {
if (Rs < 768) {
return 8;
}
else if (Rs >= 768) {
return 1;
}
}
else if (diffRsCsdiff >= 26) {
if (Cs < 889) {
return 8;
}
else if (Cs >= 889) {
return 4;
}
}
}
}
}
else if (MVDiff >= 46) {
if (diffMVdiffVal < 17) {
if (diffAFD < 1) {
if (gchDC < 6) {
return 2;
}
else if (gchDC >= 6) {
return 1;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (diffMVdiffVal >= 17) {
if (diffRsCsdiff < 9) {
if (Cs < 920) {
return 2;
}
else if (Cs >= 920) {
return 4;
}
}
else if (diffRsCsdiff >= 9) {
if (Rs < 424) {
return 2;
}
else if (Rs >= 424) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (ssDCval < 66) {
if (CsDiff < 6) {
if (negBalance < 51) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 8;
}
}
else if (negBalance >= 51) {
if (Rs < 1270) {
return 8;
}
else if (Rs >= 1270) {
return 1;
}
}
}
else if (CsDiff >= 6) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
if (negBalance < 44) {
if (Rs < 468) {
return 4;
}
else if (Rs >= 468) {
return 1;
}
}
else if (negBalance >= 44) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
if (refDCval < 95) {
if (Cs < 566) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
return 4;
}
}
else if (Cs >= 566) {
if (AFD < 18) {
if (Rs < 988) {
return 2;
}
else if (Rs >= 988) {
return 1;
}
}
else if (AFD >= 18) {
if (RsCsDiff < 147) {
return 1;
}
else if (RsCsDiff >= 147) {
if (Rs < 771) {
return 4;
}
else if (Rs >= 771) {
return 8;
}
}
}
}
}
else if (refDCval >= 95) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
}
else if (Cs >= 1054) {
if (mcTcor < 87) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
return 1;
}
}
else if (mcTcor >= 87) {
if (diffRsCsdiff < -10) {
if (mu_mv_mag_sq < 2) {
return 2;
}
else if (mu_mv_mag_sq >= 2) {
return 1;
}
}
else if (diffRsCsdiff >= -10) {
if (Cs < 1256) {
if (RsDiff < 9) {
if (posBalance < 60) {
if (Rs < 898) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (Rs >= 898) {
return 2;
}
}
else if (posBalance >= 60) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (ssDCval < 67) {
if (Cs < 1201) {
return 8;
}
else if (Cs >= 1201) {
return 2;
}
}
else if (ssDCval >= 67) {
return 2;
}
}
}
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (Cs >= 1256) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1330) {
if (ssDCval < 83) {
if (RsCsDiff < 99) {
if (MVDiff < 37) {
if (Cs < 610) {
return 4;
}
else if (Cs >= 610) {
return 8;
}
}
else if (MVDiff >= 37) {
if (diffAFD < -1) {
if (Rs < 1486) {
return 4;
}
else if (Rs >= 1486) {
return 1;
}
}
else if (diffAFD >= -1) {
if (Rs < 1488) {
return 1;
}
else if (Rs >= 1488) {
return 8;
}
}
}
}
else if (RsCsDiff >= 99) {
if (negBalance < 45) {
if (diffRsCsdiff < 80) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (diffRsCsdiff >= 80) {
return 4;
}
}
else if (negBalance >= 45) {
if (mu_mv_mag_sq < 9) {
if (negBalance < 58) {
if (Cs < 929) {
if (Cs < 710) {
return 8;
}
else if (Cs >= 710) {
return 2;
}
}
else if (Cs >= 929) {
if (diffRsCsdiff < -44) {
return 2;
}
else if (diffRsCsdiff >= -44) {
if (Cs < 1258) {
if (gchDC < 2) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
if (Rs < 2072) {
return 8;
}
else if (Rs >= 2072) {
return 4;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
else if (Cs >= 1258) {
if (Rs < 1452) {
return 1;
}
else if (Rs >= 1452) {
return 8;
}
}
}
}
}
else if (negBalance >= 58) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (MVDiff < 28) {
if (Cs < 854) {
if (Rs < 1447) {
return 8;
}
else if (Rs >= 1447) {
return 2;
}
}
else if (Cs >= 854) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 1;
}
}
}
else if (MVDiff >= 28) {
if (Rs < 1450) {
return 4;
}
else if (Rs >= 1450) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 83) {
if (refDCval < 94) {
if (RsDiff < 10) {
if (mcTcor < 90) {
if (Rs < 1584) {
return 4;
}
else if (Rs >= 1584) {
return 8;
}
}
else if (mcTcor >= 90) {
if (tcor < 75) {
return 1;
}
else if (tcor >= 75) {
if (Rs < 1541) {
return 1;
}
else if (Rs >= 1541) {
return 8;
}
}
}
}
else if (RsDiff >= 10) {
if (AFD < 22) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 1839) {
return 8;
}
else if (Rs >= 1839) {
return 2;
}
}
}
else if (AFD >= 22) {
return 1;
}
}
}
else if (refDCval >= 94) {
if (MVDiff < 31) {
if (posBalance < 60) {
return 2;
}
else if (posBalance >= 60) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (MVDiff >= 31) {
if (RsCsDiff < 134) {
if (ssDCval < 97) {
if (Cs < 974) {
return 8;
}
else if (Cs >= 974) {
return 4;
}
}
else if (ssDCval >= 97) {
return 2;
}
}
else if (RsCsDiff >= 134) {
if (AFD < 21) {
if (Rs < 1614) {
return 8;
}
else if (Rs >= 1614) {
return 1;
}
}
else if (AFD >= 21) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 23) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (mcTcor < 92) {
if (Cs < 944) {
if (Rs < 1400) {
return 2;
}
else if (Rs >= 1400) {
return 1;
}
}
else if (Cs >= 944) {
return 2;
}
}
else if (mcTcor >= 92) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 1168) {
if (Rs < 682) {
if (Rs < 402) {
return 1;
}
else if (Rs >= 402) {
return 2;
}
}
else if (Rs >= 682) {
if (Rs < 1729) {
if (RsCsDiff < 133) {
if (tcor < 61) {
if (TSC < 9) {
if (Rs < 987) {
return 1;
}
else if (Rs >= 987) {
return 4;
}
}
else if (TSC >= 9) {
if (diffTSC < 2) {
if (mcTcor < 86) {
return 2;
}
else if (mcTcor >= 86) {
return 1;
}
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (tcor >= 61) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (Rs < 1145) {
return 8;
}
else if (Rs >= 1145) {
if (Cs < 1031) {
return 1;
}
else if (Cs >= 1031) {
if (Rs < 1339) {
return 8;
}
else if (Rs >= 1339) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 133) {
if (AFD < 27) {
if (AFD < 25) {
return 4;
}
else if (AFD >= 25) {
if (diffRsCsdiff < -8) {
return 4;
}
else if (diffRsCsdiff >= -8) {
if (diffRsCsdiff < 12) {
return 8;
}
else if (diffRsCsdiff >= 12) {
return 4;
}
}
}
}
else if (AFD >= 27) {
if (refDCval < 91) {
if (TSC < 12) {
if (tcor < 54) {
if (Rs < 1333) {
return 4;
}
else if (Rs >= 1333) {
return 2;
}
}
else if (tcor >= 54) {
return 2;
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (refDCval >= 91) {
return 4;
}
}
}
}
else if (Rs >= 1729) {
if (refDCval < 84) {
if (Cs < 1025) {
return 8;
}
else if (Cs >= 1025) {
return 4;
}
}
else if (refDCval >= 84) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
return 2;
}
}
}
}
}
else if (Cs >= 1168) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
if (MVDiff < 96) {
return 4;
}
else if (MVDiff >= 96) {
return 1;
}
}
}
}
}
else if (diffAFD >= 2) {
if (RsCsDiff < 153) {
if (SC < 2160) {
if (AFD < 25) {
if (diffRsCsdiff < -7) {
return 8;
}
else if (diffRsCsdiff >= -7) {
if (CsDiff < 8) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 2;
}
}
else if (CsDiff >= 8) {
return 8;
}
}
}
else if (AFD >= 25) {
if (diffAFD < 6) {
return 2;
}
else if (diffAFD >= 6) {
if (Cs < 889) {
return 2;
}
else if (Cs >= 889) {
return 1;
}
}
}
}
else if (SC >= 2160) {
if (TSC < 10) {
if (Rs < 1392) {
return 8;
}
else if (Rs >= 1392) {
return 4;
}
}
else if (TSC >= 10) {
if (Cs < 994) {
if (SC < 2280) {
return 2;
}
else if (SC >= 2280) {
return 8;
}
}
else if (Cs >= 994) {
return 1;
}
}
}
}
else if (RsCsDiff >= 153) {
if (RsCsDiff < 187) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 8;
}
}
else if (RsCsDiff >= 187) {
if (mu_mv_mag_sq < 15) {
return 2;
}
else if (mu_mv_mag_sq >= 15) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
return 4;
}
}
}
}
}
}
}
}
else if (tcor >= 79) {
if (Cs < 1041) {
if (negBalance < 64) {
if (ssDCval < 97) {
if (mu_mv_mag_sq < 3) {
if (diffAFD < -2) {
if (Rs < 1446) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 1;
}
}
else if (Rs >= 1446) {
if (MVDiff < 59) {
if (ssDCval < 90) {
return 2;
}
else if (ssDCval >= 90) {
return 1;
}
}
else if (MVDiff >= 59) {
return 8;
}
}
}
else if (diffAFD >= -2) {
if (negBalance < 51) {
if (ssDCval < 60) {
if (Rs < 896) {
return 2;
}
else if (Rs >= 896) {
return 4;
}
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (negBalance >= 51) {
if (Rs < 1720) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
if (TSC < 10) {
if (Cs < 649) {
return 4;
}
else if (Cs >= 649) {
return 2;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (Rs >= 1720) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1585) {
if (ssDCval < 85) {
if (diffRsCsdiff < -19) {
if (diffMVdiffVal < -8) {
if (Rs < 677) {
return 8;
}
else if (Rs >= 677) {
return 4;
}
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
else if (diffRsCsdiff >= -19) {
if (mu_mv_mag_sq < 7) {
if (AFD < 16) {
if (Rs < 1446) {
if (CsDiff < 5) {
if (SC < 1092) {
return 8;
}
else if (SC >= 1092) {
return 4;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (Rs >= 1446) {
return 4;
}
}
else if (AFD >= 16) {
return 4;
}
}
else if (mu_mv_mag_sq >= 7) {
if (mcTcor < 96) {
return 4;
}
else if (mcTcor >= 96) {
if (Rs < 210) {
return 4;
}
else if (Rs >= 210) {
return 2;
}
}
}
}
}
else if (ssDCval >= 85) {
if (mu_mv_mag_sq < 9) {
if (Rs < 1474) {
return 2;
}
else if (Rs >= 1474) {
return 4;
}
}
else if (mu_mv_mag_sq >= 9) {
return 8;
}
}
}
else if (Rs >= 1585) {
if (Cs < 814) {
if (RsDiff < 9) {
if (Cs < 762) {
return 1;
}
else if (Cs >= 762) {
return 2;
}
}
else if (RsDiff >= 9) {
if (RsDiff < 14) {
return 2;
}
else if (RsDiff >= 14) {
return 4;
}
}
}
else if (Cs >= 814) {
if (mu_mv_mag_sq < 26) {
return 4;
}
else if (mu_mv_mag_sq >= 26) {
if (Cs < 1031) {
return 8;
}
else if (Cs >= 1031) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 97) {
if (mcTcor < 92) {
if (Rs < 1973) {
return 1;
}
else if (Rs >= 1973) {
return 2;
}
}
else if (mcTcor >= 92) {
if (mcTcor < 95) {
return 2;
}
else if (mcTcor >= 95) {
if (Cs < 703) {
return 1;
}
else if (Cs >= 703) {
if (Cs < 810) {
return 2;
}
else if (Cs >= 810) {
if (Rs < 2027) {
return 1;
}
else if (Rs >= 2027) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 64) {
if (mcTcor < 94) {
if (Cs < 565) {
return 1;
}
else if (Cs >= 565) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 1447) {
if (ssDCval < 85) {
return 1;
}
else if (ssDCval >= 85) {
return 2;
}
}
else if (Rs >= 1447) {
return 8;
}
}
}
}
else if (Cs >= 1041) {
if (Rs < 1334) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1141) {
if (negBalance < 46) {
return 1;
}
else if (negBalance >= 46) {
return 2;
}
}
else if (Rs >= 1141) {
if (tcor < 81) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
if (Rs < 1262) {
return 4;
}
else if (Rs >= 1262) {
return 8;
}
}
}
else if (tcor >= 81) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsCsDiff < 80) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (Rs < 1077) {
return 4;
}
else if (Rs >= 1077) {
return 8;
}
}
}
}
else if (RsCsDiff >= 80) {
if (gchDC < 4) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 1059) {
return 4;
}
else if (Rs >= 1059) {
return 1;
}
}
}
else if (gchDC >= 4) {
return 2;
}
}
}
}
else if (Rs >= 1334) {
if (mu_mv_mag_sq < 52) {
if (TSC < 9) {
if (Rs < 1812) {
if (MVDiff < 37) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 8;
}
}
else if (MVDiff >= 37) {
return 2;
}
}
else if (Rs >= 1812) {
if (Cs < 1108) {
return 4;
}
else if (Cs >= 1108) {
return 2;
}
}
}
else if (TSC >= 9) {
if (AFD < 19) {
if (diffMVdiffVal < -11) {
return 4;
}
else if (diffMVdiffVal >= -11) {
if (Cs < 1051) {
if (Rs < 2100) {
return 1;
}
else if (Rs >= 2100) {
return 8;
}
}
else if (Cs >= 1051) {
return 8;
}
}
}
else if (AFD >= 19) {
if (MVDiff < 26) {
return 2;
}
else if (MVDiff >= 26) {
if (Rs < 2317) {
return 8;
}
else if (Rs >= 2317) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (Rs < 1865) {
return 1;
}
else if (Rs >= 1865) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 114) {
if (MVDiff < 22) {
if (Rs < 573) {
return 2;
}
else if (Rs >= 573) {
if (tcor < 84) {
if (TSC < 7) {
if (Rs < 646) {
return 1;
}
else if (Rs >= 646) {
return 8;
}
}
else if (TSC >= 7) {
if (diffRsCsdiff < 55) {
if (refDCval < 116) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 2;
}
}
else if (refDCval >= 116) {
if (mcTcor < 86) {
if (Rs < 1619) {
return 2;
}
else if (Rs >= 1619) {
return 8;
}
}
else if (mcTcor >= 86) {
if (posBalance < 52) {
if (Cs < 1249) {
return 2;
}
else if (Cs >= 1249) {
return 8;
}
}
else if (posBalance >= 52) {
if (MVDiff < 20) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (MVDiff >= 20) {
if (Rs < 1445) {
return 4;
}
else if (Rs >= 1445) {
return 8;
}
}
}
}
}
}
else if (diffRsCsdiff >= 55) {
return 2;
}
}
}
else if (tcor >= 84) {
if (TSC < 8) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
}
else if (TSC >= 8) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (diffMVdiffVal < 5) {
return 1;
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 22) {
if (tcor < 77) {
if (mu_mv_mag_sq < 200) {
if (AFD < 24) {
if (SC < 1603) {
if (CsDiff < 3) {
if (Rs < 589) {
return 8;
}
else if (Rs >= 589) {
return 1;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (SC >= 1603) {
if (negBalance < 44) {
if (tcor < 61) {
if (Cs < 954) {
return 8;
}
else if (Cs >= 954) {
return 1;
}
}
else if (tcor >= 61) {
if (Rs < 1287) {
return 4;
}
else if (Rs >= 1287) {
return 2;
}
}
}
else if (negBalance >= 44) {
if (AFD < 17) {
if (Cs < 774) {
return 8;
}
else if (Cs >= 774) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
return 4;
}
}
}
else if (AFD >= 17) {
if (diffAFD < -7) {
return 4;
}
else if (diffAFD >= -7) {
if (diffAFD < 11) {
if (gchDC < 7) {
return 8;
}
else if (gchDC >= 7) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (diffAFD >= 11) {
return 4;
}
}
}
}
}
}
else if (AFD >= 24) {
if (AFD < 35) {
if (CsDiff < 7) {
if (posBalance < 69) {
if (SC < 1268) {
if (AFD < 28) {
return 2;
}
else if (AFD >= 28) {
return 1;
}
}
else if (SC >= 1268) {
return 1;
}
}
else if (posBalance >= 69) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
return 1;
}
}
}
else if (CsDiff >= 7) {
if (RsCsDiff < 214) {
if (Rs < 759) {
return 2;
}
else if (Rs >= 759) {
if (mcTcor < 87) {
return 4;
}
else if (mcTcor >= 87) {
return 1;
}
}
}
else if (RsCsDiff >= 214) {
if (AFD < 32) {
return 8;
}
else if (AFD >= 32) {
return 4;
}
}
}
}
else if (AFD >= 35) {
if (refDCval < 148) {
if (Rs < 2200) {
return 2;
}
else if (Rs >= 2200) {
return 8;
}
}
else if (refDCval >= 148) {
if (Cs < 938) {
return 1;
}
else if (Cs >= 938) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 200) {
return 1;
}
}
else if (tcor >= 77) {
if (Rs < 1557) {
if (MVDiff < 33) {
if (MVDiff < 28) {
if (MVDiff < 25) {
return 4;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (MVDiff >= 28) {
return 1;
}
}
else if (MVDiff >= 33) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (RsCsDiff < 111) {
return 8;
}
else if (RsCsDiff >= 111) {
if (Rs < 1286) {
return 2;
}
else if (Rs >= 1286) {
return 8;
}
}
}
}
}
else if (Rs >= 1557) {
if (ssDCval < 151) {
if (diffTSC < 2) {
if (TSC < 8) {
if (Cs < 760) {
return 1;
}
else if (Cs >= 760) {
return 2;
}
}
else if (TSC >= 8) {
return 1;
}
}
else if (diffTSC >= 2) {
if (diffRsCsdiff < 18) {
return 1;
}
else if (diffRsCsdiff >= 18) {
if (Rs < 1830) {
return 8;
}
else if (Rs >= 1830) {
return 2;
}
}
}
}
else if (ssDCval >= 151) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (SC < 2158) {
if (Rs < 1002) {
if (refDCval < 64) {
if (MVDiff < 1) {
if (Rs < 543) {
if (Rs < 395) {
if (Cs < 910) {
if (ssDCval < 42) {
return 8;
}
else if (ssDCval >= 42) {
return 1;
}
}
else if (Cs >= 910) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 4;
}
}
}
else if (Rs >= 395) {
if (SC < 1150) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 4;
}
}
}
else if (SC >= 1150) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 4;
}
}
}
}
else if (Rs >= 543) {
if (SC < 1656) {
if (ssDCval < 57) {
if (mu_mv_mag_sq < 118) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (SC < 1534) {
if (SC < 1441) {
return 8;
}
else if (SC >= 1441) {
return 4;
}
}
else if (SC >= 1534) {
if (Rs < 893) {
return 8;
}
else if (Rs >= 893) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 994) {
return 8;
}
else if (Rs >= 994) {
return 4;
}
}
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (SC >= 1656) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (Rs < 648) {
if (Cs < 1067) {
return 1;
}
else if (Cs >= 1067) {
return 4;
}
}
else if (Rs >= 648) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (Cs < 817) {
return 1;
}
else if (Cs >= 817) {
if (Cs < 1092) {
if (diffRsCsdiff < 1) {
return 8;
}
else if (diffRsCsdiff >= 1) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 981) {
return 8;
}
else if (Rs >= 981) {
return 1;
}
}
}
}
else if (Cs >= 1092) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (TSCindex < 2) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
if (Cs < 584) {
return 8;
}
else if (Cs >= 584) {
return 1;
}
}
}
else if (TSCindex >= 2) {
if (Rs < 891) {
if (refDCval < 42) {
if (negBalance < 17) {
return 2;
}
else if (negBalance >= 17) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Rs < 576) {
return 2;
}
else if (Rs >= 576) {
return 1;
}
}
}
}
else if (refDCval >= 42) {
if (AFD < 9) {
if (AFD < 4) {
if (Cs < 607) {
if (diffRsCsdiff < -5) {
return 2;
}
else if (diffRsCsdiff >= -5) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
return 4;
}
}
}
}
else if (Cs >= 607) {
if (mu_mv_mag_sq < 98) {
if (diffRsCsdiff < -8) {
if (Rs < 783) {
return 8;
}
else if (Rs >= 783) {
return 4;
}
}
else if (diffRsCsdiff >= -8) {
if (tcor < 99) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
return 1;
}
}
else if (tcor >= 99) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (RsDiff < 1) {
if (Rs < 302) {
return 2;
}
else if (Rs >= 302) {
if (posBalance < 48) {
if (Cs < 621) {
if (Rs < 594) {
return 4;
}
else if (Rs >= 594) {
return 8;
}
}
else if (Cs >= 621) {
return 4;
}
}
else if (posBalance >= 48) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 1;
}
}
}
}
else if (RsDiff >= 1) {
if (Cs < 1004) {
if (posBalance < 31) {
return 4;
}
else if (posBalance >= 31) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (Cs >= 1004) {
if (Cs < 1132) {
return 2;
}
else if (Cs >= 1132) {
return 1;
}
}
}
}
}
}
else if (AFD >= 4) {
if (ssDCval < 62) {
if (negBalance < 27) {
if (tcor < 93) {
if (Rs < 678) {
return 4;
}
else if (Rs >= 678) {
return 1;
}
}
else if (tcor >= 93) {
return 4;
}
}
else if (negBalance >= 27) {
if (Rs < 824) {
if (diffTSC < -1) {
if (ssDCval < 57) {
if (mu_mv_mag_sq < 21) {
if (Rs < 610) {
return 1;
}
else if (Rs >= 610) {
return 8;
}
}
else if (mu_mv_mag_sq >= 21) {
if (MVDiff < 27) {
return 4;
}
else if (MVDiff >= 27) {
return 8;
}
}
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (diffTSC >= -1) {
if (Rs < 704) {
if (mu_mv_mag_sq < 19) {
if (tcor < 94) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (tcor >= 94) {
return 4;
}
}
else if (mu_mv_mag_sq >= 19) {
if (tcor < 94) {
if (Cs < 774) {
return 8;
}
else if (Cs >= 774) {
return 1;
}
}
else if (tcor >= 94) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
if (Rs < 494) {
return 4;
}
else if (Rs >= 494) {
return 1;
}
}
}
}
}
else if (Rs >= 704) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (AFD < 7) {
if (Rs < 717) {
return 4;
}
else if (Rs >= 717) {
return 8;
}
}
else if (AFD >= 7) {
return 4;
}
}
}
}
}
}
else if (Rs >= 824) {
if (Rs < 861) {
return 1;
}
else if (Rs >= 861) {
return 8;
}
}
}
}
else if (ssDCval >= 62) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 765) {
return 2;
}
else if (Rs >= 765) {
if (Rs < 866) {
return 1;
}
else if (Rs >= 866) {
return 4;
}
}
}
}
}
}
else if (AFD >= 9) {
if (SC < 1582) {
if (diffTSC < -2) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
else if (diffTSC >= -2) {
if (SC < 1429) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (SC >= 1429) {
if (Cs < 690) {
return 4;
}
else if (Cs >= 690) {
if (Rs < 713) {
return 2;
}
else if (Rs >= 713) {
return 1;
}
}
}
}
}
else if (SC >= 1582) {
if (posBalance < 53) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
else if (posBalance >= 53) {
return 8;
}
}
}
}
}
else if (Rs >= 891) {
if (ssDCval < 54) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 942) {
return 4;
}
else if (Rs >= 942) {
return 1;
}
}
}
else if (ssDCval >= 54) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
if (AFD < 6) {
if (posBalance < 34) {
return 4;
}
else if (posBalance >= 34) {
if (diffAFD < -1) {
if (Rs < 908) {
return 4;
}
else if (Rs >= 908) {
return 1;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (AFD >= 6) {
if (negBalance < 35) {
if (Rs < 957) {
return 4;
}
else if (Rs >= 957) {
return 1;
}
}
else if (negBalance >= 35) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 64) {
if (Cs < 685) {
if (negBalance < 28) {
if (mu_mv_mag_sq < 89) {
if (Cs < 615) {
return 1;
}
else if (Cs >= 615) {
return 4;
}
}
else if (mu_mv_mag_sq >= 89) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
return 2;
}
}
}
}
}
else if (negBalance >= 28) {
if (ssDCval < 74) {
if (negBalance < 39) {
if (Cs < 597) {
if (AFD < 4) {
if (Rs < 346) {
return 4;
}
else if (Rs >= 346) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
}
else if (AFD >= 4) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
return 1;
}
}
}
else if (Cs >= 597) {
if (Rs < 632) {
if (Rs < 512) {
if (Rs < 442) {
return 4;
}
else if (Rs >= 442) {
return 1;
}
}
else if (Rs >= 512) {
return 8;
}
}
else if (Rs >= 632) {
return 1;
}
}
}
else if (negBalance >= 39) {
if (tcor < 95) {
if (Cs < 644) {
return 4;
}
else if (Cs >= 644) {
return 2;
}
}
else if (tcor >= 95) {
if (AFD < 7) {
if (negBalance < 61) {
if (Cs < 579) {
return 4;
}
else if (Cs >= 579) {
return 8;
}
}
else if (negBalance >= 61) {
return 4;
}
}
else if (AFD >= 7) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
else if (ssDCval >= 74) {
if (ssDCval < 85) {
if (Rs < 607) {
if (Rs < 238) {
return 1;
}
else if (Rs >= 238) {
return 8;
}
}
else if (Rs >= 607) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (ssDCval >= 85) {
return 2;
}
}
}
}
else if (Cs >= 685) {
if (ssDCval < 81) {
if (AFD < 6) {
if (Rs < 690) {
if (SC < 1236) {
if (Rs < 340) {
return 8;
}
else if (Rs >= 340) {
return 4;
}
}
else if (SC >= 1236) {
if (RsDiff < 2) {
if (Cs < 698) {
if (Rs < 640) {
return 8;
}
else if (Rs >= 640) {
return 1;
}
}
else if (Cs >= 698) {
return 8;
}
}
else if (RsDiff >= 2) {
if (Rs < 615) {
return 2;
}
else if (Rs >= 615) {
return 1;
}
}
}
}
else if (Rs >= 690) {
if (ssDCval < 75) {
if (ssDCval < 71) {
if (mu_mv_mag_sq < 106) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (mcTcor < 98) {
if (diffMVdiffVal < 1) {
if (Rs < 900) {
if (Cs < 1132) {
return 1;
}
else if (Cs >= 1132) {
return 8;
}
}
else if (Rs >= 900) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
if (Rs < 917) {
return 4;
}
else if (Rs >= 917) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Cs < 717) {
if (Rs < 859) {
return 1;
}
else if (Rs >= 859) {
return 2;
}
}
else if (Cs >= 717) {
if (Rs < 730) {
if (Rs < 700) {
return 4;
}
else if (Rs >= 700) {
return 8;
}
}
else if (Rs >= 730) {
if (SC < 1913) {
return 4;
}
else if (SC >= 1913) {
if (Cs < 1012) {
return 1;
}
else if (Cs >= 1012) {
if (Rs < 886) {
return 2;
}
else if (Rs >= 886) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (Rs < 808) {
return 1;
}
else if (Rs >= 808) {
if (Rs < 836) {
return 4;
}
else if (Rs >= 836) {
return 1;
}
}
}
}
else if (ssDCval >= 75) {
if (Rs < 832) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
else if (Rs >= 832) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
if (SC < 1927) {
return 8;
}
else if (SC >= 1927) {
if (Rs < 945) {
return 8;
}
else if (Rs >= 945) {
return 4;
}
}
}
}
}
}
}
else if (AFD >= 6) {
if (Rs < 506) {
if (AFD < 13) {
if (refDCval < 65) {
return 2;
}
else if (refDCval >= 65) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (AFD >= 13) {
return 2;
}
}
else if (Rs >= 506) {
if (negBalance < 47) {
if (RsCsDiff < 23) {
if (MVDiff < 2) {
if (Rs < 862) {
return 1;
}
else if (Rs >= 862) {
return 4;
}
}
else if (MVDiff >= 2) {
if (posBalance < 54) {
if (Rs < 693) {
return 8;
}
else if (Rs >= 693) {
return 2;
}
}
else if (posBalance >= 54) {
if (RsCsDiff < 16) {
if (Cs < 750) {
return 4;
}
else if (Cs >= 750) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (RsCsDiff >= 16) {
return 4;
}
}
}
}
else if (RsCsDiff >= 23) {
if (diffRsCsdiff < -15) {
if (Rs < 833) {
return 2;
}
else if (Rs >= 833) {
return 8;
}
}
else if (diffRsCsdiff >= -15) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (Rs < 594) {
return 2;
}
else if (Rs >= 594) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (SC < 1483) {
if (Rs < 656) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
if (TSC < 4) {
if (Cs < 734) {
if (posBalance < 44) {
if (Rs < 569) {
return 2;
}
else if (Rs >= 569) {
return 1;
}
}
else if (posBalance >= 44) {
return 2;
}
}
else if (Cs >= 734) {
return 2;
}
}
else if (TSC >= 4) {
if (Rs < 475) {
return 8;
}
else if (Rs >= 475) {
return 4;
}
}
}
}
else if (Rs >= 656) {
return 1;
}
}
else if (SC >= 1483) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 1002) {
if (mu_mv_mag_sq < 123) {
if (ssDCval < 51) {
if (mu_mv_mag_sq < 79) {
if (Rs < 1370) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Rs >= 1370) {
return 4;
}
}
else if (mu_mv_mag_sq >= 79) {
if (Rs < 1155) {
return 8;
}
else if (Rs >= 1155) {
if (Rs < 1259) {
return 4;
}
else if (Rs >= 1259) {
if (Rs < 1338) {
return 4;
}
else if (Rs >= 1338) {
return 4;
}
}
}
}
}
else if (ssDCval >= 51) {
if (RsCsDiff < 12) {
if (SC < 1798) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (mu_mv_mag_sq < 108) {
if (Rs < 1028) {
return 8;
}
else if (Rs >= 1028) {
return 1;
}
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 1125) {
return 2;
}
else if (Rs >= 1125) {
if (Cs < 661) {
return 1;
}
else if (Cs >= 661) {
return 2;
}
}
}
}
}
else if (SC >= 1798) {
if (ssDCval < 70) {
if (MVDiff < 15) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
if (negBalance < 42) {
if (MVDiff < 8) {
if (diffRsCsdiff < -5) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Cs < 935) {
return 8;
}
else if (Cs >= 935) {
return 1;
}
}
}
else if (diffRsCsdiff >= -5) {
if (ssDCval < 55) {
if (refDCval < 54) {
return 1;
}
else if (refDCval >= 54) {
if (Rs < 1150) {
return 1;
}
else if (Rs >= 1150) {
return 8;
}
}
}
else if (ssDCval >= 55) {
return 1;
}
}
}
else if (MVDiff >= 8) {
if (posBalance < 44) {
if (diffRsCsdiff < 5) {
return 1;
}
else if (diffRsCsdiff >= 5) {
return 8;
}
}
else if (posBalance >= 44) {
return 8;
}
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
else if (MVDiff >= 15) {
if (Rs < 1073) {
return 4;
}
else if (Rs >= 1073) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (Rs < 1110) {
if (mu_mv_mag_sq < 118) {
return 8;
}
else if (mu_mv_mag_sq >= 118) {
return 4;
}
}
else if (Rs >= 1110) {
if (ssDCval < 73) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (ssDCval >= 73) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 12) {
if (mu_mv_mag_sq < 58) {
if (posBalance < 52) {
if (negBalance < 64) {
if (Cs < 574) {
if (Rs < 1394) {
return 2;
}
else if (Rs >= 1394) {
return 1;
}
}
else if (Cs >= 574) {
return 1;
}
}
else if (negBalance >= 64) {
return 8;
}
}
else if (posBalance >= 52) {
if (SC < 1913) {
if (mu_mv_mag_sq < 49) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (mu_mv_mag_sq >= 49) {
if (Rs < 1091) {
return 8;
}
else if (Rs >= 1091) {
return 4;
}
}
}
else if (SC >= 1913) {
if (MVDiff < 11) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (MVDiff >= 11) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (ssDCval < 66) {
if (Rs < 1234) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 1062) {
return 4;
}
else if (Rs >= 1062) {
if (Rs < 1117) {
return 2;
}
else if (Rs >= 1117) {
return 8;
}
}
}
}
else if (Rs >= 1234) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (ssDCval >= 66) {
if (negBalance < 69) {
return 8;
}
else if (negBalance >= 69) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1133) {
if (posBalance < 41) {
if (diffRsCsdiff < 1) {
if (Cs < 582) {
return 2;
}
else if (Cs >= 582) {
return 8;
}
}
else if (diffRsCsdiff >= 1) {
return 2;
}
}
else if (posBalance >= 41) {
if (ssDCval < 60) {
if (Rs < 1052) {
return 8;
}
else if (Rs >= 1052) {
return 1;
}
}
else if (ssDCval >= 60) {
if (SC < 1817) {
return 2;
}
else if (SC >= 1817) {
return 4;
}
}
}
}
else if (Rs >= 1133) {
if (Rs < 1383) {
if (Cs < 756) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
return 1;
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 707) {
return 8;
}
else if (Cs >= 707) {
return 4;
}
}
}
}
else if (Cs >= 756) {
if (Cs < 885) {
if (mu_mv_mag_sq < 124) {
return 1;
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (Cs >= 885) {
if (Rs < 1215) {
return 8;
}
else if (Rs >= 1215) {
return 1;
}
}
}
}
else if (Rs >= 1383) {
if (Cs < 670) {
return 8;
}
else if (Cs >= 670) {
return 1;
}
}
}
}
}
}
else if (SC >= 2158) {
if (Rs < 1761) {
if (Cs < 1058) {
if (mu_mv_mag_sq < 128) {
if (refDCval < 69) {
if (ssDCval < 66) {
if (posBalance < 42) {
if (mu_mv_mag_sq < 89) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (CsDiff < 1) {
if (MVDiff < 1) {
if (SC < 2230) {
return 8;
}
else if (SC >= 2230) {
return 1;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (CsDiff >= 1) {
if (Cs < 744) {
return 1;
}
else if (Cs >= 744) {
if (Cs < 950) {
return 8;
}
else if (Cs >= 950) {
if (Rs < 1541) {
return 1;
}
else if (Rs >= 1541) {
if (Cs < 1018) {
return 8;
}
else if (Cs >= 1018) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 42) {
if (SC < 2303) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (Cs < 929) {
if (mu_mv_mag_sq < 81) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 81) {
if (Rs < 1422) {
return 8;
}
else if (Rs >= 1422) {
return 4;
}
}
}
else if (Cs >= 929) {
return 1;
}
}
}
else if (SC >= 2303) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 1221) {
return 8;
}
else if (Rs >= 1221) {
if (Rs < 1483) {
return 1;
}
else if (Rs >= 1483) {
return 2;
}
}
}
}
}
else if (refDCval >= 69) {
if (mu_mv_mag_sq < 64) {
if (Rs < 1305) {
return 8;
}
else if (Rs >= 1305) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (negBalance < 21) {
if (Rs < 1455) {
return 8;
}
else if (Rs >= 1455) {
return 1;
}
}
else if (negBalance >= 21) {
if (posBalance < 55) {
if (diffRsCsdiff < -10) {
if (Rs < 1430) {
return 8;
}
else if (Rs >= 1430) {
return 1;
}
}
else if (diffRsCsdiff >= -10) {
if (diffMVdiffVal < -5) {
if (Rs < 1477) {
return 8;
}
else if (Rs >= 1477) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (Rs < 1489) {
if (Rs < 1342) {
return 8;
}
else if (Rs >= 1342) {
return 1;
}
}
else if (Rs >= 1489) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1318) {
if (Cs < 933) {
return 8;
}
else if (Cs >= 933) {
return 1;
}
}
else if (Rs >= 1318) {
return 8;
}
}
}
else if (Cs >= 1058) {
if (RsDiff < 1) {
if (MVDiff < 12) {
if (ssDCval < 55) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (Rs < 1183) {
return 1;
}
else if (Rs >= 1183) {
return 4;
}
}
}
else if (ssDCval >= 55) {
if (ssDCval < 70) {
if (Rs < 1469) {
return 1;
}
else if (Rs >= 1469) {
return 8;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
else if (MVDiff >= 12) {
return 4;
}
}
else if (RsDiff >= 1) {
if (AFD < 3) {
if (ssDCval < 53) {
if (negBalance < 26) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (SC < 2928) {
if (mcTcor < 98) {
return 4;
}
else if (mcTcor >= 98) {
if (ssDCval < 49) {
if (Cs < 1179) {
if (Rs < 1641) {
return 4;
}
else if (Rs >= 1641) {
if (negBalance < 19) {
if (posBalance < 36) {
if (Rs < 1694) {
if (Cs < 1114) {
return 8;
}
else if (Cs >= 1114) {
if (Rs < 1683) {
return 4;
}
else if (Rs >= 1683) {
return 8;
}
}
}
else if (Rs >= 1694) {
return 8;
}
}
else if (posBalance >= 36) {
if (Rs < 1696) {
return 4;
}
else if (Rs >= 1696) {
if (TSCindex < 3) {
if (Rs < 1758) {
if (SC < 2884) {
if (Rs < 1733) {
return 4;
}
else if (Rs >= 1733) {
if (Rs < 1745) {
return 4;
}
else if (Rs >= 1745) {
return 4;
}
}
}
else if (SC >= 2884) {
return 4;
}
}
else if (Rs >= 1758) {
return 8;
}
}
else if (TSCindex >= 3) {
return 8;
}
}
}
}
else if (negBalance >= 19) {
if (Cs < 1138) {
if (Rs < 1689) {
if (mu_mv_mag_sq < 126) {
return 4;
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1107) {
return 8;
}
else if (Cs >= 1107) {
if (Rs < 1660) {
return 8;
}
else if (Rs >= 1660) {
return 8;
}
}
}
}
else if (Rs >= 1689) {
if (Rs < 1724) {
return 4;
}
else if (Rs >= 1724) {
if (negBalance < 22) {
if (Rs < 1734) {
if (Rs < 1730) {
return 4;
}
else if (Rs >= 1730) {
return 4;
}
}
else if (Rs >= 1734) {
if (Rs < 1745) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
return 4;
}
}
else if (Rs >= 1745) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (Cs < 1065) {
return 4;
}
else if (Cs >= 1065) {
return 4;
}
}
}
}
}
else if (Cs >= 1138) {
if (posBalance < 32) {
return 4;
}
else if (posBalance >= 32) {
return 4;
}
}
}
}
}
else if (Cs >= 1179) {
return 4;
}
}
else if (ssDCval >= 49) {
if (Rs < 1354) {
if (Rs < 1145) {
return 8;
}
else if (Rs >= 1145) {
return 4;
}
}
else if (Rs >= 1354) {
return 8;
}
}
}
}
else if (SC >= 2928) {
return 4;
}
}
}
else if (negBalance >= 26) {
return 8;
}
}
else if (ssDCval >= 53) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
if (Rs < 1241) {
return 4;
}
else if (Rs >= 1241) {
if (mu_mv_mag_sq < 127) {
if (Cs < 1298) {
if (diffRsCsdiff < 5) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
if (Rs < 1547) {
return 8;
}
else if (Rs >= 1547) {
return 1;
}
}
}
else if (diffRsCsdiff >= 5) {
return 4;
}
}
else if (Cs >= 1298) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 1481) {
if (Cs < 1097) {
return 8;
}
else if (Cs >= 1097) {
return 4;
}
}
else if (Rs >= 1481) {
return 8;
}
}
}
}
}
}
else if (AFD >= 3) {
if (SC < 2704) {
if (RsCsDiff < 12) {
if (mu_mv_mag_sq < 113) {
if (Cs < 1184) {
if (SC < 2681) {
if (Cs < 1065) {
return 1;
}
else if (Cs >= 1065) {
return 8;
}
}
else if (SC >= 2681) {
return 1;
}
}
else if (Cs >= 1184) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
if (Rs < 1191) {
if (ssDCval < 74) {
if (Cs < 1194) {
return 4;
}
else if (Cs >= 1194) {
return 8;
}
}
else if (ssDCval >= 74) {
return 4;
}
}
else if (Rs >= 1191) {
if (Rs < 1364) {
return 4;
}
else if (Rs >= 1364) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (negBalance < 34) {
if (posBalance < 53) {
if (Rs < 1260) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
if (negBalance < 28) {
return 8;
}
else if (negBalance >= 28) {
return 4;
}
}
}
else if (Rs >= 1260) {
if (Cs < 1118) {
if (Rs < 1271) {
return 4;
}
else if (Rs >= 1271) {
return 8;
}
}
else if (Cs >= 1118) {
return 4;
}
}
}
else if (posBalance >= 53) {
if (Rs < 993) {
return 8;
}
else if (Rs >= 993) {
return 4;
}
}
}
else if (negBalance >= 34) {
if (ssDCval < 70) {
if (Rs < 1302) {
return 1;
}
else if (Rs >= 1302) {
if (Rs < 1323) {
return 8;
}
else if (Rs >= 1323) {
return 4;
}
}
}
else if (ssDCval >= 70) {
return 4;
}
}
}
}
else if (RsCsDiff >= 12) {
if (negBalance < 42) {
if (mu_mv_mag_sq < 124) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 105) {
return 8;
}
else if (mu_mv_mag_sq >= 105) {
if (MVDiff < 1) {
if (Rs < 965) {
return 2;
}
else if (Rs >= 965) {
return 8;
}
}
else if (MVDiff >= 1) {
if (Cs < 1253) {
return 4;
}
else if (Cs >= 1253) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 4;
}
}
else if (negBalance >= 42) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
else if (SC >= 2704) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
if (mu_mv_mag_sq < 86) {
if (Rs < 1652) {
return 8;
}
else if (Rs >= 1652) {
return 2;
}
}
else if (mu_mv_mag_sq >= 86) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1761) {
if (TSC < 4) {
if (refDCval < 83) {
if (ssDCval < 56) {
if (SC < 2989) {
if (mu_mv_mag_sq < 109) {
return 4;
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 1985) {
if (mu_mv_mag_sq < 128) {
if (diffAFD < 1) {
if (negBalance < 19) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
return 4;
}
}
else if (negBalance >= 19) {
return 4;
}
}
else if (diffAFD >= 1) {
if (Rs < 1775) {
return 8;
}
else if (Rs >= 1775) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (posBalance < 33) {
if (negBalance < 22) {
return 4;
}
else if (negBalance >= 22) {
if (Rs < 1791) {
if (Rs < 1775) {
return 4;
}
else if (Rs >= 1775) {
return 8;
}
}
else if (Rs >= 1791) {
return 4;
}
}
}
else if (posBalance >= 33) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
if (SC < 2848) {
return 4;
}
else if (SC >= 2848) {
if (Rs < 1817) {
if (Rs < 1778) {
if (Cs < 1157) {
if (Cs < 1137) {
if (Rs < 1770) {
return 8;
}
else if (Rs >= 1770) {
return 8;
}
}
else if (Cs >= 1137) {
return 8;
}
}
else if (Cs >= 1157) {
return 4;
}
}
else if (Rs >= 1778) {
if (Rs < 1781) {
return 4;
}
else if (Rs >= 1781) {
if (Rs < 1791) {
return 4;
}
else if (Rs >= 1791) {
if (Rs < 1807) {
if (Rs < 1802) {
return 4;
}
else if (Rs >= 1802) {
return 8;
}
}
else if (Rs >= 1807) {
return 4;
}
}
}
}
}
else if (Rs >= 1817) {
if (Cs < 1111) {
if (Rs < 1826) {
return 8;
}
else if (Rs >= 1826) {
return 4;
}
}
else if (Cs >= 1111) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1985) {
return 8;
}
}
}
else if (SC >= 2989) {
if (Cs < 838) {
return 8;
}
else if (Cs >= 838) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (Cs < 1136) {
if (TSC < 3) {
if (ssDCval < 79) {
if (ssDCval < 72) {
if (posBalance < 36) {
if (Rs < 1929) {
if (Cs < 993) {
return 1;
}
else if (Cs >= 993) {
if (SC < 2940) {
return 4;
}
else if (SC >= 2940) {
if (refDCval < 62) {
return 1;
}
else if (refDCval >= 62) {
return 4;
}
}
}
}
else if (Rs >= 1929) {
return 4;
}
}
else if (posBalance >= 36) {
if (mu_mv_mag_sq < 103) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 1;
}
}
else if (mu_mv_mag_sq >= 103) {
if (negBalance < 15) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (negBalance >= 15) {
return 1;
}
}
}
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (ssDCval >= 79) {
if (Cs < 871) {
return 4;
}
else if (Cs >= 871) {
return 8;
}
}
}
else if (TSC >= 3) {
if (Cs < 886) {
if (ssDCval < 82) {
return 1;
}
else if (ssDCval >= 82) {
return 8;
}
}
else if (Cs >= 886) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
}
else if (Cs >= 1136) {
if (ssDCval < 67) {
if (Cs < 1318) {
return 2;
}
else if (Cs >= 1318) {
if (Rs < 2318) {
return 8;
}
else if (Rs >= 2318) {
return 2;
}
}
}
else if (ssDCval >= 67) {
return 8;
}
}
}
}
else if (refDCval >= 83) {
if (Rs < 2195) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
else if (Rs >= 2195) {
return 8;
}
}
}
else if (TSC >= 4) {
if (Cs < 967) {
if (ssDCval < 71) {
if (refDCval < 61) {
return 1;
}
else if (refDCval >= 61) {
if (Cs < 853) {
return 2;
}
else if (Cs >= 853) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
return 4;
}
}
}
}
else if (ssDCval >= 71) {
if (mu_mv_mag_sq < 93) {
if (RsCsDiff < 56) {
if (Cs < 569) {
return 4;
}
else if (Cs >= 569) {
return 1;
}
}
else if (RsCsDiff >= 56) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (RsDiff < 4) {
if (Rs < 2056) {
return 4;
}
else if (Rs >= 2056) {
return 4;
}
}
else if (RsDiff >= 4) {
return 2;
}
}
}
}
else if (Cs >= 967) {
if (mu_mv_mag_sq < 21) {
if (SC < 3162) {
return 8;
}
else if (SC >= 3162) {
return 1;
}
}
else if (mu_mv_mag_sq >= 21) {
if (RsDiff < 7) {
if (Rs < 2152) {
if (Rs < 1801) {
if (Rs < 1780) {
return 8;
}
else if (Rs >= 1780) {
return 1;
}
}
else if (Rs >= 1801) {
return 8;
}
}
else if (Rs >= 2152) {
if (Rs < 2269) {
if (Rs < 2184) {
return 1;
}
else if (Rs >= 2184) {
return 4;
}
}
else if (Rs >= 2269) {
return 8;
}
}
}
else if (RsDiff >= 7) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 99) {
if (SC < 1310) {
if (ssDCval < 116) {
if (TSC < 5) {
if (Rs < 400) {
if (MVDiff < 7) {
if (ssDCval < 102) {
if (Rs < 384) {
return 1;
}
else if (Rs >= 384) {
return 8;
}
}
else if (ssDCval >= 102) {
return 4;
}
}
else if (MVDiff >= 7) {
return 2;
}
}
else if (Rs >= 400) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (posBalance < 31) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (posBalance >= 31) {
if (Cs < 563) {
return 1;
}
else if (Cs >= 563) {
return 2;
}
}
}
}
}
else if (TSC >= 5) {
if (Rs < 642) {
if (negBalance < 37) {
if (Rs < 549) {
return 2;
}
else if (Rs >= 549) {
return 8;
}
}
else if (negBalance >= 37) {
return 8;
}
}
else if (Rs >= 642) {
if (Rs < 667) {
return 2;
}
else if (Rs >= 667) {
return 4;
}
}
}
}
else if (ssDCval >= 116) {
if (mcTcor < 98) {
if (CsDiff < 3) {
if (ssDCval < 119) {
if (Rs < 491) {
return 4;
}
else if (Rs >= 491) {
return 2;
}
}
else if (ssDCval >= 119) {
return 8;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (mcTcor >= 98) {
if (ssDCval < 150) {
if (posBalance < 34) {
if (Rs < 540) {
return 1;
}
else if (Rs >= 540) {
return 8;
}
}
else if (posBalance >= 34) {
if (Cs < 586) {
if (negBalance < 25) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (negBalance >= 25) {
if (Cs < 585) {
return 2;
}
else if (Cs >= 585) {
return 4;
}
}
}
else if (Cs >= 586) {
return 2;
}
}
}
else if (ssDCval >= 150) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 506) {
return 1;
}
else if (Rs >= 506) {
return 2;
}
}
}
}
}
}
else if (SC >= 1310) {
if (RsCsDiff < 77) {
if (MVDiff < 5) {
if (SC < 3126) {
if (SC < 2661) {
if (Rs < 662) {
if (ssDCval < 95) {
if (Rs < 515) {
return 8;
}
else if (Rs >= 515) {
return 2;
}
}
else if (ssDCval >= 95) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 630) {
return 8;
}
else if (Rs >= 630) {
return 1;
}
}
}
}
else if (Rs >= 662) {
if (Cs < 833) {
if (Cs < 592) {
if (Rs < 2007) {
return 1;
}
else if (Rs >= 2007) {
return 8;
}
}
else if (Cs >= 592) {
if (Rs < 708) {
if (Rs < 680) {
return 1;
}
else if (Rs >= 680) {
return 4;
}
}
else if (Rs >= 708) {
if (refDCval < 90) {
if (Rs < 1917) {
return 1;
}
else if (Rs >= 1917) {
return 4;
}
}
else if (refDCval >= 90) {
if (gchDC < 2) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (gchDC >= 2) {
if (Rs < 877) {
return 4;
}
else if (Rs >= 877) {
return 1;
}
}
}
}
}
}
else if (Cs >= 833) {
if (SC < 1846) {
return 4;
}
else if (SC >= 1846) {
if (negBalance < 47) {
if (mu_mv_mag_sq < 82) {
if (diffAFD < 2) {
if (ssDCval < 141) {
return 4;
}
else if (ssDCval >= 141) {
return 8;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 82) {
if (ssDCval < 93) {
if (Rs < 1078) {
return 4;
}
else if (Rs >= 1078) {
return 8;
}
}
else if (ssDCval >= 93) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (Rs < 1310) {
return 8;
}
else if (Rs >= 1310) {
return 1;
}
}
}
}
}
}
else if (SC >= 2661) {
if (ssDCval < 120) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 8;
}
}
else if (ssDCval >= 120) {
if (Rs < 1949) {
return 2;
}
else if (Rs >= 1949) {
return 4;
}
}
}
}
else if (SC >= 3126) {
return 8;
}
}
else if (MVDiff >= 5) {
if (refDCval < 104) {
if (TSC < 5) {
if (Rs < 1436) {
if (mu_mv_mag_sq < 88) {
if (Cs < 721) {
return 1;
}
else if (Cs >= 721) {
return 2;
}
}
else if (mu_mv_mag_sq >= 88) {
return 8;
}
}
else if (Rs >= 1436) {
return 1;
}
}
else if (TSC >= 5) {
if (RsCsDiff < 27) {
if (tcor < 93) {
if (Cs < 1055) {
return 8;
}
else if (Cs >= 1055) {
return 4;
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (RsCsDiff >= 27) {
if (TSC < 7) {
if (negBalance < 52) {
if (MVDiff < 15) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Cs < 1240) {
return 4;
}
else if (Cs >= 1240) {
return 2;
}
}
}
else if (MVDiff >= 15) {
if (Rs < 1555) {
return 2;
}
else if (Rs >= 1555) {
return 8;
}
}
}
else if (negBalance >= 52) {
if (diffMVdiffVal < -3) {
if (Rs < 1979) {
return 2;
}
else if (Rs >= 1979) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
}
else if (TSC >= 7) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
if (AFD < 12) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
return 2;
}
}
else if (AFD >= 12) {
if (Cs < 1272) {
return 1;
}
else if (Cs >= 1272) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 104) {
if (ssDCval < 159) {
if (mu_mv_mag_sq < 5) {
if (negBalance < 51) {
if (Rs < 868) {
if (ssDCval < 142) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
else if (ssDCval >= 142) {
return 1;
}
}
else if (Rs >= 868) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
if (Rs < 1801) {
return 1;
}
else if (Rs >= 1801) {
return 2;
}
}
}
}
else if (negBalance >= 51) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
if (negBalance < 67) {
return 1;
}
else if (negBalance >= 67) {
if (Rs < 924) {
return 8;
}
else if (Rs >= 924) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Cs < 572) {
if (Rs < 1875) {
return 2;
}
else if (Rs >= 1875) {
return 1;
}
}
else if (Cs >= 572) {
if (SC < 1562) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (AFD < 7) {
if (Rs < 662) {
return 8;
}
else if (Rs >= 662) {
return 1;
}
}
else if (AFD >= 7) {
if (Rs < 871) {
return 2;
}
else if (Rs >= 871) {
return 8;
}
}
}
}
else if (SC >= 1562) {
if (Rs < 2191) {
if (negBalance < 36) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 879) {
if (Rs < 646) {
return 1;
}
else if (Rs >= 646) {
return 8;
}
}
else if (Rs >= 879) {
return 4;
}
}
}
else if (negBalance >= 36) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
if (diffMVdiffVal < -1) {
if (mu_mv_mag_sq < 11) {
return 8;
}
else if (mu_mv_mag_sq >= 11) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
}
}
else if (Rs >= 2191) {
if (Cs < 966) {
return 2;
}
else if (Cs >= 966) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 159) {
if (refDCval < 169) {
return 4;
}
else if (refDCval >= 169) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (Cs < 792) {
if (SC < 2007) {
if (ssDCval < 138) {
if (AFD < 25) {
return 1;
}
else if (AFD >= 25) {
if (Rs < 1113) {
return 2;
}
else if (Rs >= 1113) {
return 4;
}
}
}
else if (ssDCval >= 138) {
return 2;
}
}
else if (SC >= 2007) {
if (TSC < 6) {
if (Rs < 1956) {
return 8;
}
else if (Rs >= 1956) {
return 1;
}
}
else if (TSC >= 6) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 1;
}
}
}
}
}
else if (Cs >= 792) {
if (mu_mv_mag_sq < 17) {
if (Rs < 2027) {
if (Rs < 1197) {
if (Cs < 855) {
if (Rs < 1072) {
return 2;
}
else if (Rs >= 1072) {
return 8;
}
}
else if (Cs >= 855) {
if (MVDiff < 70) {
return 1;
}
else if (MVDiff >= 70) {
return 2;
}
}
}
else if (Rs >= 1197) {
if (AFD < 14) {
if (Rs < 1842) {
return 8;
}
else if (Rs >= 1842) {
return 2;
}
}
else if (AFD >= 14) {
return 2;
}
}
}
else if (Rs >= 2027) {
if (SC < 3023) {
return 1;
}
else if (SC >= 3023) {
if (Cs < 1029) {
return 4;
}
else if (Cs >= 1029) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (diffMVdiffVal < -1) {
if (Rs < 1893) {
if (Cs < 812) {
return 4;
}
else if (Cs >= 812) {
return 2;
}
}
else if (Rs >= 1893) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 99) {
if (ssDCval < 133) {
if (Rs < 630) {
if (Cs < 1233) {
if (mu_mv_mag_sq < 108) {
if (Rs < 391) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
else if (Rs >= 391) {
if (ssDCval < 117) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
if (TSC < 4) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (TSC >= 4) {
if (Rs < 423) {
return 1;
}
else if (Rs >= 423) {
return 8;
}
}
}
}
else if (ssDCval >= 117) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 350) {
if (ssDCval < 123) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
return 8;
}
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 350) {
if (AFD < 4) {
if (Cs < 1203) {
if (diffAFD < -1) {
if (Cs < 765) {
if (Cs < 728) {
return 8;
}
else if (Cs >= 728) {
return 1;
}
}
else if (Cs >= 765) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Cs >= 1203) {
if (Rs < 600) {
return 1;
}
else if (Rs >= 600) {
return 8;
}
}
}
else if (AFD >= 4) {
if (ssDCval < 102) {
if (Cs < 801) {
return 1;
}
else if (Cs >= 801) {
return 8;
}
}
else if (ssDCval >= 102) {
return 8;
}
}
}
}
}
else if (Cs >= 1233) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (Rs >= 630) {
if (refDCval < 97) {
if (SC < 2689) {
if (TSC < 3) {
if (mu_mv_mag_sq < 126) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
return 8;
}
}
else if (mu_mv_mag_sq >= 126) {
if (posBalance < 26) {
return 4;
}
else if (posBalance >= 26) {
return 8;
}
}
}
else if (TSC >= 3) {
if (Rs < 790) {
if (Cs < 939) {
return 1;
}
else if (Cs >= 939) {
return 4;
}
}
else if (Rs >= 790) {
if (Rs < 1501) {
if (Rs < 850) {
if (Rs < 809) {
return 8;
}
else if (Rs >= 809) {
return 1;
}
}
else if (Rs >= 850) {
return 8;
}
}
else if (Rs >= 1501) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
}
else if (SC >= 2689) {
if (Rs < 2143) {
if (diffMVdiffVal < -2) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (Rs >= 2143) {
if (ssDCval < 89) {
if (CsDiff < 1) {
if (Cs < 764) {
return 8;
}
else if (Cs >= 764) {
return 4;
}
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 89) {
return 8;
}
}
}
}
else if (refDCval >= 97) {
if (ssDCval < 121) {
if (refDCval < 115) {
if (refDCval < 113) {
if (SC < 1365) {
if (Cs < 616) {
return 8;
}
else if (Cs >= 616) {
return 1;
}
}
else if (SC >= 1365) {
if (diffRsCsdiff < 11) {
return 8;
}
else if (diffRsCsdiff >= 11) {
return 1;
}
}
}
else if (refDCval >= 113) {
if (SC < 2837) {
if (Rs < 1598) {
return 8;
}
else if (Rs >= 1598) {
return 1;
}
}
else if (SC >= 2837) {
return 8;
}
}
}
else if (refDCval >= 115) {
return 8;
}
}
else if (ssDCval >= 121) {
if (Cs < 774) {
if (Rs < 979) {
return 8;
}
else if (Rs >= 979) {
return 1;
}
}
else if (Cs >= 774) {
if (AFD < 2) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (AFD >= 2) {
if (Rs < 1534) {
return 8;
}
else if (Rs >= 1534) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 144) {
return 1;
}
else if (ssDCval >= 144) {
if (Rs < 995) {
return 8;
}
else if (Rs >= 995) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2356) {
if (MVDiff < 16) {
if (ssDCval < 109) {
if (posBalance < 48) {
if (RsDiff < 7) {
if (Cs < 1027) {
if (Rs < 3738) {
if (Cs < 669) {
if (Rs < 2505) {
if (mu_mv_mag_sq < 2) {
if (SC < 2927) {
return 4;
}
else if (SC >= 2927) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
return 8;
}
}
}
else if (Rs >= 2505) {
if (Cs < 550) {
if (Rs < 3159) {
return 8;
}
else if (Rs >= 3159) {
return 2;
}
}
else if (Cs >= 550) {
if (Rs < 3054) {
if (diffRsCsdiff < 4) {
return 1;
}
else if (diffRsCsdiff >= 4) {
return 8;
}
}
else if (Rs >= 3054) {
return 2;
}
}
}
}
else if (Cs >= 669) {
if (negBalance < 28) {
if (ssDCval < 98) {
if (SC < 3214) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (SC >= 3214) {
if (Cs < 807) {
return 4;
}
else if (Cs >= 807) {
if (tcor < 98) {
if (Rs < 3445) {
return 4;
}
else if (Rs >= 3445) {
return 4;
}
}
else if (tcor >= 98) {
if (Rs < 2418) {
return 2;
}
else if (Rs >= 2418) {
if (Cs < 935) {
return 8;
}
else if (Cs >= 935) {
if (Cs < 1001) {
return 4;
}
else if (Cs >= 1001) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 98) {
if (negBalance < 26) {
return 8;
}
else if (negBalance >= 26) {
if (Rs < 3439) {
if (Rs < 3392) {
return 8;
}
else if (Rs >= 3392) {
return 8;
}
}
else if (Rs >= 3439) {
return 4;
}
}
}
}
else if (negBalance >= 28) {
if (tcor < 96) {
if (ssDCval < 75) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
if (Rs < 2383) {
return 2;
}
else if (Rs >= 2383) {
return 4;
}
}
}
else if (ssDCval >= 75) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (Rs < 2456) {
return 4;
}
else if (Rs >= 2456) {
if (Rs < 2636) {
return 8;
}
else if (Rs >= 2636) {
if (ssDCval < 98) {
if (Rs < 2899) {
return 8;
}
else if (Rs >= 2899) {
return 4;
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 96) {
if (Rs < 2475) {
if (Rs < 2420) {
return 4;
}
else if (Rs >= 2420) {
return 1;
}
}
else if (Rs >= 2475) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 4;
}
}
}
}
}
}
else if (Rs >= 3738) {
if (ssDCval < 92) {
return 4;
}
else if (ssDCval >= 92) {
return 2;
}
}
}
else if (Cs >= 1027) {
if (RsCsDiff < 7) {
if (SC < 4956) {
if (negBalance < 23) {
if (RsCsDiff < 5) {
return 8;
}
else if (RsCsDiff >= 5) {
if (Cs < 1212) {
if (Rs < 3113) {
return 8;
}
else if (Rs >= 3113) {
return 4;
}
}
else if (Cs >= 1212) {
return 8;
}
}
}
else if (negBalance >= 23) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Rs < 2988) {
return 8;
}
else if (Rs >= 2988) {
return 2;
}
}
}
}
else if (SC >= 4956) {
if (SC < 5305) {
if (Cs < 1259) {
return 4;
}
else if (Cs >= 1259) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (Rs < 4017) {
return 8;
}
else if (Rs >= 4017) {
return 4;
}
}
}
}
else if (SC >= 5305) {
if (Cs < 1293) {
if (SC < 5438) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
return 8;
}
}
else if (SC >= 5438) {
if (Rs < 4490) {
return 8;
}
else if (Rs >= 4490) {
return 2;
}
}
}
else if (Cs >= 1293) {
return 8;
}
}
}
}
else if (RsCsDiff >= 7) {
if (refDCval < 103) {
if (TSC < 7) {
if (posBalance < 35) {
if (MVDiff < 2) {
if (Cs < 1211) {
return 8;
}
else if (Cs >= 1211) {
if (Rs < 4016) {
return 1;
}
else if (Rs >= 4016) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 95) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
else if (posBalance >= 35) {
if (Rs < 3017) {
if (ssDCval < 77) {
if (ssDCval < 48) {
return 4;
}
else if (ssDCval >= 48) {
if (negBalance < 39) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (Rs < 2771) {
return 2;
}
else if (Rs >= 2771) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (Rs < 2686) {
return 8;
}
else if (Rs >= 2686) {
return 4;
}
}
}
}
else if (ssDCval >= 77) {
if (Rs < 2950) {
return 1;
}
else if (Rs >= 2950) {
return 8;
}
}
}
else if (Rs >= 3017) {
if (diffAFD < -2) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 4;
}
}
else if (diffAFD >= -2) {
if (RsCsDiff < 27) {
if (MVDiff < 11) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 4;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (RsCsDiff >= 27) {
if (refDCval < 79) {
if (negBalance < 28) {
if (Rs < 3229) {
return 4;
}
else if (Rs >= 3229) {
return 1;
}
}
else if (negBalance >= 28) {
if (ssDCval < 65) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 2;
}
}
else if (ssDCval >= 65) {
return 4;
}
}
}
else if (refDCval >= 79) {
if (mu_mv_mag_sq < 71) {
if (AFD < 9) {
if (Rs < 3720) {
return 8;
}
else if (Rs >= 3720) {
if (Rs < 3855) {
return 8;
}
else if (Rs >= 3855) {
return 4;
}
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (mu_mv_mag_sq >= 71) {
return 2;
}
}
}
}
}
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (refDCval >= 103) {
if (Rs < 3761) {
return 8;
}
else if (Rs >= 3761) {
if (Rs < 4192) {
return 4;
}
else if (Rs >= 4192) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (negBalance < 46) {
if (ssDCval < 88) {
if (refDCval < 65) {
if (MVDiff < 6) {
if (mcTcor < 93) {
if (Cs < 1241) {
return 4;
}
else if (Cs >= 1241) {
return 2;
}
}
else if (mcTcor >= 93) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (MVDiff >= 6) {
if (Cs < 1026) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
if (Rs < 3269) {
return 2;
}
else if (Rs >= 3269) {
return 4;
}
}
}
else if (Cs >= 1026) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
if (Rs < 3504) {
return 2;
}
else if (Rs >= 3504) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 65) {
if (diffTSC < -1) {
if (TSC < 5) {
if (Rs < 3712) {
if (Rs < 2938) {
return 1;
}
else if (Rs >= 2938) {
if (Rs < 3367) {
return 8;
}
else if (Rs >= 3367) {
return 2;
}
}
}
else if (Rs >= 3712) {
return 4;
}
}
else if (TSC >= 5) {
if (negBalance < 40) {
if (diffRsCsdiff < -7) {
if (TSC < 7) {
if (negBalance < 38) {
return 2;
}
else if (negBalance >= 38) {
if (Rs < 3136) {
return 2;
}
else if (Rs >= 3136) {
return 4;
}
}
}
else if (TSC >= 7) {
if (Rs < 3526) {
return 4;
}
else if (Rs >= 3526) {
return 1;
}
}
}
else if (diffRsCsdiff >= -7) {
return 1;
}
}
else if (negBalance >= 40) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
else if (diffTSC >= -1) {
if (Cs < 1064) {
if (Rs < 3077) {
if (TSC < 6) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (Rs < 2693) {
return 1;
}
else if (Rs >= 2693) {
return 2;
}
}
}
else if (TSC >= 6) {
if (CsDiff < 4) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 2;
}
}
else if (CsDiff >= 4) {
if (Rs < 2509) {
return 2;
}
else if (Rs >= 2509) {
return 4;
}
}
}
}
else if (Rs >= 3077) {
if (MVDiff < 9) {
if (Cs < 751) {
return 1;
}
else if (Cs >= 751) {
return 4;
}
}
else if (MVDiff >= 9) {
if (Rs < 3336) {
return 4;
}
else if (Rs >= 3336) {
return 1;
}
}
}
}
else if (Cs >= 1064) {
if (Rs < 3444) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
if (ssDCval < 68) {
if (AFD < 11) {
if (Rs < 3101) {
return 4;
}
else if (Rs >= 3101) {
return 2;
}
}
else if (AFD >= 11) {
return 2;
}
}
else if (ssDCval >= 68) {
if (Rs < 2636) {
return 2;
}
else if (Rs >= 2636) {
return 4;
}
}
}
}
else if (Rs >= 3444) {
if (ssDCval < 78) {
if (diffAFD < 1) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (Rs < 4355) {
return 4;
}
else if (Rs >= 4355) {
return 1;
}
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (ssDCval >= 78) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (negBalance < 43) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Cs < 1066) {
return 4;
}
else if (Cs >= 1066) {
if (diffMVdiffVal < -1) {
if (negBalance < 31) {
if (Rs < 4160) {
if (Rs < 4032) {
return 8;
}
else if (Rs >= 4032) {
if (Rs < 4085) {
return 8;
}
else if (Rs >= 4085) {
return 8;
}
}
}
else if (Rs >= 4160) {
if (Rs < 4215) {
return 4;
}
else if (Rs >= 4215) {
return 4;
}
}
}
else if (negBalance >= 31) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (refDCval < 91) {
return 4;
}
else if (refDCval >= 91) {
return 4;
}
}
}
}
}
else if (negBalance >= 43) {
return 2;
}
}
}
else if (negBalance >= 46) {
if (negBalance < 51) {
if (SC < 3027) {
return 1;
}
else if (SC >= 3027) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (RsDiff < 11) {
if (posBalance < 43) {
if (MVDiff < 8) {
if (Rs < 3020) {
return 2;
}
else if (Rs >= 3020) {
return 4;
}
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (posBalance >= 43) {
return 2;
}
}
else if (RsDiff >= 11) {
if (SC < 3955) {
return 8;
}
else if (SC >= 3955) {
return 4;
}
}
}
}
}
else if (negBalance >= 51) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Rs < 3093) {
return 1;
}
else if (Rs >= 3093) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Rs < 3904) {
if (Cs < 215) {
if (Rs < 2630) {
return 2;
}
else if (Rs >= 2630) {
return 1;
}
}
else if (Cs >= 215) {
if (negBalance < 18) {
if (Cs < 630) {
return 1;
}
else if (Cs >= 630) {
if (diffTSC < 1) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (diffTSC >= 1) {
return 4;
}
}
}
else if (negBalance >= 18) {
if (mcTcor < 99) {
if (ssDCval < 69) {
if (Cs < 562) {
if (Rs < 2467) {
return 2;
}
else if (Rs >= 2467) {
return 4;
}
}
else if (Cs >= 562) {
if (Rs < 2760) {
return 2;
}
else if (Rs >= 2760) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Cs < 761) {
return 4;
}
else if (Cs >= 761) {
return 2;
}
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 81) {
if (posBalance < 53) {
if (Rs < 2844) {
return 1;
}
else if (Rs >= 2844) {
return 8;
}
}
else if (posBalance >= 53) {
if (SC < 2754) {
return 1;
}
else if (SC >= 2754) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 81) {
return 2;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 2510) {
return 2;
}
else if (Rs >= 2510) {
if (Rs < 3049) {
return 8;
}
else if (Rs >= 3049) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3904) {
return 4;
}
}
else if (CsDiff >= 3) {
if (refDCval < 65) {
if (SC < 3473) {
return 2;
}
else if (SC >= 3473) {
if (tcor < 87) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (tcor >= 87) {
if (mu_mv_mag_sq < 84) {
if (RsCsDiff < 102) {
if (Rs < 2490) {
return 8;
}
else if (Rs >= 2490) {
return 2;
}
}
else if (RsCsDiff >= 102) {
if (refDCval < 63) {
if (Rs < 6961) {
return 4;
}
else if (Rs >= 6961) {
return 2;
}
}
else if (refDCval >= 63) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (Cs < 993) {
return 2;
}
else if (Cs >= 993) {
if (Rs < 3375) {
return 4;
}
else if (Rs >= 3375) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 65) {
if (TSC < 5) {
if (Cs < 910) {
if (negBalance < 40) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 4;
}
}
else if (negBalance >= 40) {
if (Cs < 494) {
return 2;
}
else if (Cs >= 494) {
return 1;
}
}
}
else if (Cs >= 910) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (posBalance < 53) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (posBalance >= 53) {
return 2;
}
}
}
}
else if (TSC >= 5) {
if (tcor < 80) {
if (RsCsDiff < 246) {
if (SC < 3127) {
return 2;
}
else if (SC >= 3127) {
if (negBalance < 51) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (RsCsDiff < 179) {
return 1;
}
else if (RsCsDiff >= 179) {
if (Rs < 3408) {
return 2;
}
else if (Rs >= 3408) {
return 1;
}
}
}
}
else if (negBalance >= 51) {
return 4;
}
}
}
else if (RsCsDiff >= 246) {
if (negBalance < 44) {
if (Rs < 3036) {
return 2;
}
else if (Rs >= 3036) {
return 4;
}
}
else if (negBalance >= 44) {
return 2;
}
}
}
else if (tcor >= 80) {
if (Rs < 2469) {
if (ssDCval < 77) {
if (SC < 3536) {
return 2;
}
else if (SC >= 3536) {
return 4;
}
}
else if (ssDCval >= 77) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (Rs < 2450) {
if (diffRsCsdiff < 13) {
return 2;
}
else if (diffRsCsdiff >= 13) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (Rs >= 2450) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 4;
}
}
}
}
}
else if (Rs >= 2469) {
if (mcTcor < 89) {
if (Rs < 3477) {
return 1;
}
else if (Rs >= 3477) {
return 2;
}
}
else if (mcTcor >= 89) {
if (Rs < 4116) {
if (negBalance < 41) {
if (posBalance < 52) {
if (mu_mv_mag_sq < 46) {
if (Rs < 2803) {
return 1;
}
else if (Rs >= 2803) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 46) {
return 2;
}
}
else if (posBalance >= 52) {
if (AFD < 17) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (AFD >= 17) {
if (Rs < 3197) {
return 4;
}
else if (Rs >= 3197) {
return 2;
}
}
}
}
else if (negBalance >= 41) {
return 2;
}
}
else if (Rs >= 4116) {
if (TSC < 7) {
if (Rs < 4231) {
if (Rs < 4168) {
return 8;
}
else if (Rs >= 4168) {
return 2;
}
}
else if (Rs >= 4231) {
return 4;
}
}
else if (TSC >= 7) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsDiff < 11) {
if (mcTcor < 98) {
if (diffRsCsdiff < -50) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 4;
}
}
else if (diffRsCsdiff >= -50) {
if (refDCval < 94) {
if (Rs < 3006) {
if (posBalance < 52) {
if (Rs < 2460) {
return 4;
}
else if (Rs >= 2460) {
return 1;
}
}
else if (posBalance >= 52) {
if (SC < 3585) {
if (refDCval < 92) {
return 8;
}
else if (refDCval >= 92) {
if (Rs < 2570) {
return 4;
}
else if (Rs >= 2570) {
return 4;
}
}
}
else if (SC >= 3585) {
return 8;
}
}
}
else if (Rs >= 3006) {
if (Cs < 1169) {
return 2;
}
else if (Cs >= 1169) {
return 4;
}
}
}
else if (refDCval >= 94) {
if (Cs < 502) {
return 2;
}
else if (Cs >= 502) {
if (diffMVdiffVal < 8) {
if (RsCsDiff < 121) {
return 8;
}
else if (RsCsDiff >= 121) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (diffMVdiffVal >= 8) {
if (Rs < 4031) {
return 4;
}
else if (Rs >= 4031) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (Cs < 1015) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 2;
}
}
else if (Cs >= 1015) {
if (Rs < 3141) {
return 8;
}
else if (Rs >= 3141) {
return 2;
}
}
}
}
else if (RsDiff >= 11) {
if (negBalance < 49) {
if (RsCsDiff < 142) {
if (Rs < 3367) {
return 2;
}
else if (Rs >= 3367) {
return 8;
}
}
else if (RsCsDiff >= 142) {
return 2;
}
}
else if (negBalance >= 49) {
if (Cs < 837) {
return 2;
}
else if (Cs >= 837) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (CsDiff < 6) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
return 1;
}
}
else if (CsDiff >= 6) {
if (diffMVdiffVal < -3) {
if (Rs < 3235) {
return 2;
}
else if (Rs >= 3235) {
if (Rs < 3580) {
return 8;
}
else if (Rs >= 3580) {
return 2;
}
}
}
else if (diffMVdiffVal >= -3) {
if (diffAFD < 1) {
if (ssDCval < 103) {
if (Rs < 3089) {
return 4;
}
else if (Rs >= 3089) {
return 8;
}
}
else if (ssDCval >= 103) {
if (Cs < 1104) {
return 2;
}
else if (Cs >= 1104) {
return 4;
}
}
}
else if (diffAFD >= 1) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 109) {
if (Rs < 3356) {
if (mcTcor < 97) {
if (AFD < 14) {
if (mcTcor < 95) {
if (Rs < 3172) {
if (posBalance < 41) {
if (Rs < 2423) {
return 4;
}
else if (Rs >= 2423) {
return 8;
}
}
else if (posBalance >= 41) {
return 8;
}
}
else if (Rs >= 3172) {
if (Rs < 3244) {
return 2;
}
else if (Rs >= 3244) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (diffRsCsdiff < 17) {
if (diffMVdiffVal < -3) {
if (mu_mv_mag_sq < 76) {
return 8;
}
else if (mu_mv_mag_sq >= 76) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
if (diffAFD < -2) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
if (Rs < 2948) {
return 1;
}
else if (Rs >= 2948) {
return 8;
}
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (diffRsCsdiff >= 17) {
if (SC < 3762) {
return 1;
}
else if (SC >= 3762) {
if (Rs < 2814) {
return 8;
}
else if (Rs >= 2814) {
return 2;
}
}
}
}
}
else if (AFD >= 14) {
if (Cs < 644) {
return 2;
}
else if (Cs >= 644) {
if (diffMVdiffVal < 1) {
if (diffMVdiffVal < -4) {
if (refDCval < 111) {
return 8;
}
else if (refDCval >= 111) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
if (Cs < 731) {
return 1;
}
else if (Cs >= 731) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
if (mu_mv_mag_sq < 10) {
return 1;
}
else if (mu_mv_mag_sq >= 10) {
return 2;
}
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 7) {
if (TSCindex < 3) {
if (SC < 3273) {
return 1;
}
else if (SC >= 3273) {
if (SC < 4163) {
return 8;
}
else if (SC >= 4163) {
return 2;
}
}
}
else if (TSCindex >= 3) {
if (Rs < 2446) {
if (Cs < 424) {
return 2;
}
else if (Cs >= 424) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
if (Rs < 2370) {
return 8;
}
else if (Rs >= 2370) {
return 1;
}
}
}
}
else if (Rs >= 2446) {
if (ssDCval < 115) {
if (Cs < 630) {
return 8;
}
else if (Cs >= 630) {
if (refDCval < 108) {
return 4;
}
else if (refDCval >= 108) {
return 2;
}
}
}
else if (ssDCval >= 115) {
return 2;
}
}
}
}
else if (MVDiff >= 7) {
if (RsDiff < 9) {
if (CsDiff < 3) {
if (diffRsCsdiff < -33) {
return 1;
}
else if (diffRsCsdiff >= -33) {
if (mu_mv_mag_sq < 47) {
if (Rs < 2649) {
return 1;
}
else if (Rs >= 2649) {
return 8;
}
}
else if (mu_mv_mag_sq >= 47) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (ssDCval < 134) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 3030) {
return 8;
}
else if (Rs >= 3030) {
return 2;
}
}
}
else if (ssDCval >= 134) {
if (Rs < 2377) {
return 8;
}
else if (Rs >= 2377) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
if (SC < 3312) {
if (ssDCval < 111) {
return 2;
}
else if (ssDCval >= 111) {
return 1;
}
}
else if (SC >= 3312) {
if (Cs < 814) {
if (ssDCval < 130) {
return 2;
}
else if (ssDCval >= 130) {
if (Rs < 2919) {
return 2;
}
else if (Rs >= 2919) {
return 1;
}
}
}
else if (Cs >= 814) {
if (refDCval < 115) {
return 4;
}
else if (refDCval >= 115) {
if (mu_mv_mag_sq < 52) {
if (Cs < 1069) {
return 1;
}
else if (Cs >= 1069) {
return 8;
}
}
else if (mu_mv_mag_sq >= 52) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 3356) {
if (MVDiff < 7) {
if (refDCval < 136) {
if (mu_mv_mag_sq < 31) {
if (Rs < 3898) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 2;
}
}
}
else if (Rs >= 3898) {
return 2;
}
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
else if (refDCval >= 136) {
if (CsDiff < 4) {
if (Rs < 3822) {
if (SC < 4865) {
if (diffTSC < -2) {
return 1;
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (SC >= 4865) {
if (Cs < 1237) {
if (ssDCval < 141) {
return 2;
}
else if (ssDCval >= 141) {
if (mu_mv_mag_sq < 92) {
return 2;
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
}
else if (Cs >= 1237) {
return 2;
}
}
}
else if (Rs >= 3822) {
if (TSC < 10) {
if (diffTSC < -2) {
if (Rs < 4012) {
return 1;
}
else if (Rs >= 4012) {
return 2;
}
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (TSC >= 10) {
return 1;
}
}
}
else if (CsDiff >= 4) {
if (Rs < 4418) {
if (Cs < 1001) {
return 2;
}
else if (Cs >= 1001) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (SC < 5539) {
return 1;
}
else if (SC >= 5539) {
if (Rs < 4361) {
return 2;
}
else if (Rs >= 4361) {
return 1;
}
}
}
}
}
else if (Rs >= 4418) {
return 2;
}
}
}
}
else if (MVDiff >= 7) {
if (mcTcor < 94) {
if (ssDCval < 135) {
if (refDCval < 133) {
return 2;
}
else if (refDCval >= 133) {
if (Rs < 3753) {
return 8;
}
else if (Rs >= 3753) {
return 1;
}
}
}
else if (ssDCval >= 135) {
if (RsDiff < 18) {
return 4;
}
else if (RsDiff >= 18) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (TSC < 6) {
if (posBalance < 42) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (Rs < 3894) {
return 8;
}
else if (Rs >= 3894) {
if (SC < 4681) {
return 8;
}
else if (SC >= 4681) {
return 2;
}
}
}
}
else if (posBalance >= 42) {
if (Cs < 1130) {
return 2;
}
else if (Cs >= 1130) {
if (Cs < 1242) {
if (Rs < 3553) {
return 2;
}
else if (Rs >= 3553) {
return 1;
}
}
else if (Cs >= 1242) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 130) {
if (RsDiff < 9) {
if (Cs < 1226) {
return 2;
}
else if (Cs >= 1226) {
return 8;
}
}
else if (RsDiff >= 9) {
if (AFD < 33) {
return 2;
}
else if (AFD >= 33) {
if (Rs < 4193) {
return 4;
}
else if (Rs >= 4193) {
return 2;
}
}
}
}
else if (ssDCval >= 130) {
if (Cs < 1004) {
if (Rs < 5821) {
if (mcTcor < 95) {
if (Rs < 3864) {
return 1;
}
else if (Rs >= 3864) {
return 4;
}
}
else if (mcTcor >= 95) {
if (Rs < 3414) {
return 1;
}
else if (Rs >= 3414) {
return 2;
}
}
}
else if (Rs >= 5821) {
if (RsDiff < 19) {
return 1;
}
else if (RsDiff >= 19) {
return 2;
}
}
}
else if (Cs >= 1004) {
if (Rs < 5093) {
if (mu_mv_mag_sq < 15) {
return 1;
}
else if (mu_mv_mag_sq >= 15) {
if (AFD < 13) {
if (Rs < 4834) {
return 1;
}
else if (Rs >= 4834) {
return 2;
}
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (Rs >= 5093) {
return 2;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 16) {
if (mcTcor < 69) {
if (ssDCval < 101) {
if (Cs < 453) {
return 8;
}
else if (Cs >= 453) {
return 1;
}
}
else if (ssDCval >= 101) {
if (AFD < 28) {
return 4;
}
else if (AFD >= 28) {
if (Cs < 523) {
return 8;
}
else if (Cs >= 523) {
return 2;
}
}
}
}
else if (mcTcor >= 69) {
if (RsCsDiff < 210) {
if (SC < 3414) {
if (refDCval < 91) {
if (AFD < 15) {
if (ssDCval < 74) {
if (RsDiff < 7) {
if (SC < 3205) {
return 8;
}
else if (SC >= 3205) {
return 4;
}
}
else if (RsDiff >= 7) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (Cs < 372) {
return 4;
}
else if (Cs >= 372) {
return 2;
}
}
}
}
else if (ssDCval >= 74) {
if (diffRsCsdiff < -52) {
if (Rs < 2568) {
if (Rs < 2496) {
return 4;
}
else if (Rs >= 2496) {
return 8;
}
}
else if (Rs >= 2568) {
return 2;
}
}
else if (diffRsCsdiff >= -52) {
if (Rs < 2637) {
if (SC < 2883) {
if (Rs < 2406) {
return 1;
}
else if (Rs >= 2406) {
return 2;
}
}
else if (SC >= 2883) {
return 1;
}
}
else if (Rs >= 2637) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (Cs < 240) {
return 8;
}
else if (Cs >= 240) {
return 4;
}
}
}
}
}
}
else if (AFD >= 15) {
if (mcTcor < 89) {
if (SC < 2979) {
return 2;
}
else if (SC >= 2979) {
return 1;
}
}
else if (mcTcor >= 89) {
if (TSC < 12) {
if (Rs < 2373) {
return 1;
}
else if (Rs >= 2373) {
return 2;
}
}
else if (TSC >= 12) {
return 4;
}
}
}
}
else if (refDCval >= 91) {
if (refDCval < 134) {
if (mcTcor < 95) {
if (Cs < 320) {
return 1;
}
else if (Cs >= 320) {
if (MVDiff < 29) {
if (Rs < 2737) {
return 2;
}
else if (Rs >= 2737) {
return 1;
}
}
else if (MVDiff >= 29) {
if (diffMVdiffVal < -45) {
return 2;
}
else if (diffMVdiffVal >= -45) {
if (SC < 2771) {
return 1;
}
else if (SC >= 2771) {
if (Rs < 2366) {
return 2;
}
else if (Rs >= 2366) {
if (AFD < 27) {
if (gchDC < 1) {
if (Cs < 720) {
if (MVDiff < 38) {
return 1;
}
else if (MVDiff >= 38) {
return 8;
}
}
else if (Cs >= 720) {
return 1;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (AFD >= 27) {
return 1;
}
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (Cs < 442) {
if (posBalance < 59) {
if (diffRsCsdiff < -9) {
if (tcor < 90) {
if (Cs < 424) {
return 8;
}
else if (Cs >= 424) {
if (MVDiff < 34) {
return 8;
}
else if (MVDiff >= 34) {
return 1;
}
}
}
else if (tcor >= 90) {
if (Rs < 2907) {
return 1;
}
else if (Rs >= 2907) {
return 4;
}
}
}
else if (diffRsCsdiff >= -9) {
if (gchDC < 6) {
if (Cs < 407) {
return 8;
}
else if (Cs >= 407) {
return 1;
}
}
else if (gchDC >= 6) {
if (Rs < 2532) {
return 1;
}
else if (Rs >= 2532) {
return 2;
}
}
}
}
else if (posBalance >= 59) {
if (Rs < 2641) {
return 1;
}
else if (Rs >= 2641) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
if (Rs < 2764) {
return 1;
}
else if (Rs >= 2764) {
return 8;
}
}
}
}
}
else if (Cs >= 442) {
if (tcor < 79) {
return 2;
}
else if (tcor >= 79) {
if (CsDiff < 7) {
if (mu_mv_mag_sq < 54) {
if (gchDC < 2) {
if (MVDiff < 39) {
return 1;
}
else if (MVDiff >= 39) {
if (Rs < 2567) {
return 1;
}
else if (Rs >= 2567) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 54) {
return 2;
}
}
else if (CsDiff >= 7) {
return 2;
}
}
}
}
}
else if (refDCval >= 134) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
return 1;
}
}
}
}
else if (SC >= 3414) {
if (refDCval < 82) {
if (Cs < 743) {
if (Rs < 3010) {
if (Rs < 2885) {
return 4;
}
else if (Rs >= 2885) {
return 8;
}
}
else if (Rs >= 3010) {
if (Cs < 654) {
return 4;
}
else if (Cs >= 654) {
if (Rs < 3084) {
return 2;
}
else if (Rs >= 3084) {
return 4;
}
}
}
}
else if (Cs >= 743) {
if (Rs < 2616) {
if (TSC < 8) {
if (MVDiff < 27) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
return 4;
}
}
else if (MVDiff >= 27) {
return 8;
}
}
else if (TSC >= 8) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
if (SC < 3463) {
return 8;
}
else if (SC >= 3463) {
return 1;
}
}
}
}
else if (Rs >= 2616) {
if (posBalance < 47) {
if (CsDiff < 4) {
if (Rs < 3065) {
return 1;
}
else if (Rs >= 3065) {
return 8;
}
}
else if (CsDiff >= 4) {
if (diffRsCsdiff < 16) {
if (diffRsCsdiff < -86) {
if (Cs < 1281) {
return 4;
}
else if (Cs >= 1281) {
return 1;
}
}
else if (diffRsCsdiff >= -86) {
if (posBalance < 29) {
return 1;
}
else if (posBalance >= 29) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
return 4;
}
}
}
}
}
else if (diffRsCsdiff >= 16) {
return 4;
}
}
}
else if (posBalance >= 47) {
if (CsDiff < 8) {
if (tcor < 68) {
if (Rs < 2811) {
return 2;
}
else if (Rs >= 2811) {
return 4;
}
}
else if (tcor >= 68) {
if (Cs < 1191) {
if (MVDiff < 18) {
if (Cs < 903) {
if (Rs < 2794) {
return 2;
}
else if (Rs >= 2794) {
return 4;
}
}
else if (Cs >= 903) {
return 2;
}
}
else if (MVDiff >= 18) {
return 2;
}
}
else if (Cs >= 1191) {
if (Cs < 1220) {
return 4;
}
else if (Cs >= 1220) {
return 2;
}
}
}
}
else if (CsDiff >= 8) {
return 1;
}
}
}
}
}
else if (refDCval >= 82) {
if (mcTcor < 94) {
if (refDCval < 134) {
if (Cs < 692) {
if (Rs < 3992) {
if (TSC < 10) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (Rs >= 3992) {
return 2;
}
}
else if (Cs >= 692) {
if (AFD < 26) {
if (MVDiff < 29) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
if (tcor < 79) {
if (MVDiff < 18) {
if (Rs < 2602) {
return 8;
}
else if (Rs >= 2602) {
return 4;
}
}
else if (MVDiff >= 18) {
return 1;
}
}
else if (tcor >= 79) {
if (diffAFD < 6) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
return 2;
}
}
else if (diffAFD >= 6) {
return 8;
}
}
}
}
else if (MVDiff >= 29) {
if (SC < 3759) {
if (TSC < 13) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (TSC >= 13) {
return 2;
}
}
else if (SC >= 3759) {
if (mu_mv_mag_sq < 3) {
return 8;
}
else if (mu_mv_mag_sq >= 3) {
if (diffTSC < 1) {
if (mu_mv_mag_sq < 30) {
return 1;
}
else if (mu_mv_mag_sq >= 30) {
if (Cs < 1179) {
return 4;
}
else if (Cs >= 1179) {
return 8;
}
}
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
}
else if (AFD >= 26) {
if (mcTcor < 76) {
return 2;
}
else if (mcTcor >= 76) {
if (Cs < 1127) {
return 1;
}
else if (Cs >= 1127) {
return 4;
}
}
}
}
}
else if (refDCval >= 134) {
if (ssDCval < 148) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 8;
}
}
}
else if (ssDCval >= 148) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (RsDiff < 7) {
if (Rs < 3775) {
if (Rs < 2373) {
return 4;
}
else if (Rs >= 2373) {
if (diffMVdiffVal < -4) {
if (Cs < 985) {
if (Rs < 3334) {
return 1;
}
else if (Rs >= 3334) {
return 8;
}
}
else if (Cs >= 985) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (Rs >= 3775) {
return 2;
}
}
else if (RsDiff >= 7) {
if (negBalance < 49) {
if (SC < 3995) {
if (Cs < 556) {
if (Rs < 3298) {
return 8;
}
else if (Rs >= 3298) {
if (Rs < 3419) {
return 2;
}
else if (Rs >= 3419) {
return 8;
}
}
}
else if (Cs >= 556) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 48) {
if (diffMVdiffVal < 25) {
return 1;
}
else if (diffMVdiffVal >= 25) {
return 8;
}
}
else if (mu_mv_mag_sq >= 48) {
if (Rs < 2730) {
return 8;
}
else if (Rs >= 2730) {
return 2;
}
}
}
else if (diffAFD >= 2) {
return 2;
}
}
}
else if (SC >= 3995) {
if (ssDCval < 132) {
if (posBalance < 79) {
if (Cs < 1149) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
}
else if (Cs >= 1149) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (Rs < 4523) {
return 2;
}
else if (Rs >= 4523) {
return 8;
}
}
}
}
else if (posBalance >= 79) {
if (RsDiff < 13) {
return 2;
}
else if (RsDiff >= 13) {
return 1;
}
}
}
else if (ssDCval >= 132) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
if (diffMVdiffVal < 11) {
if (refDCval < 147) {
return 1;
}
else if (refDCval >= 147) {
return 2;
}
}
else if (diffMVdiffVal >= 11) {
if (Rs < 5695) {
return 8;
}
else if (Rs >= 5695) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 49) {
if (ssDCval < 95) {
if (MVDiff < 40) {
return 2;
}
else if (MVDiff >= 40) {
if (Rs < 3663) {
return 1;
}
else if (Rs >= 3663) {
return 8;
}
}
}
else if (ssDCval >= 95) {
if (refDCval < 123) {
if (diffRsCsdiff < -171) {
return 2;
}
else if (diffRsCsdiff >= -171) {
if (Cs < 471) {
if (Rs < 4868) {
return 4;
}
else if (Rs >= 4868) {
return 2;
}
}
else if (Cs >= 471) {
if (negBalance < 71) {
if (AFD < 19) {
if (refDCval < 97) {
return 4;
}
else if (refDCval >= 97) {
return 8;
}
}
else if (AFD >= 19) {
if (MVDiff < 40) {
return 1;
}
else if (MVDiff >= 40) {
return 8;
}
}
}
else if (negBalance >= 71) {
return 2;
}
}
}
}
else if (refDCval >= 123) {
if (Rs < 3045) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (Rs < 2784) {
return 8;
}
else if (Rs >= 2784) {
return 1;
}
}
}
else if (Rs >= 3045) {
if (RsCsDiff < 88) {
return 8;
}
else if (RsCsDiff >= 88) {
if (SC < 5090) {
if (MVDiff < 44) {
return 2;
}
else if (MVDiff >= 44) {
return 1;
}
}
else if (SC >= 5090) {
if (Rs < 4117) {
return 8;
}
else if (Rs >= 4117) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 210) {
if (ssDCval < 135) {
if (refDCval < 112) {
if (CsDiff < 7) {
if (refDCval < 74) {
if (Rs < 3409) {
if (AFD < 22) {
if (diffRsCsdiff < 124) {
if (posBalance < 52) {
if (negBalance < 45) {
return 4;
}
else if (negBalance >= 45) {
if (Rs < 2418) {
return 1;
}
else if (Rs >= 2418) {
return 2;
}
}
}
else if (posBalance >= 52) {
return 2;
}
}
else if (diffRsCsdiff >= 124) {
if (Rs < 2547) {
return 1;
}
else if (Rs >= 2547) {
return 4;
}
}
}
else if (AFD >= 22) {
if (RsDiff < 25) {
if (ssDCval < 71) {
if (AFD < 27) {
return 1;
}
else if (AFD >= 27) {
return 4;
}
}
else if (ssDCval >= 71) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 1;
}
}
}
else if (RsDiff >= 25) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
}
}
else if (Rs >= 3409) {
if (Rs < 5813) {
return 4;
}
else if (Rs >= 5813) {
return 8;
}
}
}
else if (refDCval >= 74) {
if (mcTcor < 91) {
if (Cs < 972) {
if (tcor < 77) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (Cs < 714) {
return 2;
}
else if (Cs >= 714) {
if (RsDiff < 19) {
if (Rs < 2745) {
return 4;
}
else if (Rs >= 2745) {
return 1;
}
}
else if (RsDiff >= 19) {
return 2;
}
}
}
}
else if (tcor >= 77) {
if (negBalance < 41) {
return 2;
}
else if (negBalance >= 41) {
if (refDCval < 110) {
if (Cs < 945) {
if (diffAFD < 4) {
return 4;
}
else if (diffAFD >= 4) {
if (Rs < 2960) {
return 4;
}
else if (Rs >= 2960) {
return 8;
}
}
}
else if (Cs >= 945) {
return 2;
}
}
else if (refDCval >= 110) {
return 8;
}
}
}
}
else if (Cs >= 972) {
if (diffAFD < -2) {
if (mu_mv_mag_sq < 16) {
return 2;
}
else if (mu_mv_mag_sq >= 16) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
if (Rs < 3204) {
return 2;
}
else if (Rs >= 3204) {
return 4;
}
}
}
}
else if (diffAFD >= -2) {
if (ssDCval < 100) {
return 4;
}
else if (ssDCval >= 100) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (Rs < 3913) {
return 4;
}
else if (Rs >= 3913) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (refDCval < 78) {
if (Rs < 3698) {
if (MVDiff < 63) {
return 2;
}
else if (MVDiff >= 63) {
return 4;
}
}
else if (Rs >= 3698) {
return 4;
}
}
else if (refDCval >= 78) {
if (Cs < 697) {
if (RsCsDiff < 286) {
if (AFD < 22) {
if (ssDCval < 112) {
return 2;
}
else if (ssDCval >= 112) {
return 4;
}
}
else if (AFD >= 22) {
if (Cs < 339) {
return 1;
}
else if (Cs >= 339) {
if (Rs < 3195) {
return 2;
}
else if (Rs >= 3195) {
return 4;
}
}
}
}
else if (RsCsDiff >= 286) {
return 2;
}
}
else if (Cs >= 697) {
if (MVDiff < 68) {
if (diffRsCsdiff < 56) {
if (Cs < 1271) {
return 2;
}
else if (Cs >= 1271) {
return 8;
}
}
else if (diffRsCsdiff >= 56) {
if (diffRsCsdiff < 59) {
return 8;
}
else if (diffRsCsdiff >= 59) {
if (Rs < 4872) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
else if (Rs >= 4872) {
return 8;
}
}
}
}
else if (MVDiff >= 68) {
if (Cs < 854) {
return 1;
}
else if (Cs >= 854) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
}
}
}
else if (CsDiff >= 7) {
if (tcor < 73) {
if (AFD < 26) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
if (mcTcor < 83) {
if (Rs < 2759) {
return 4;
}
else if (Rs >= 2759) {
return 8;
}
}
else if (mcTcor >= 83) {
return 8;
}
}
}
else if (AFD >= 26) {
if (Cs < 760) {
return 8;
}
else if (Cs >= 760) {
if (Cs < 932) {
if (AFD < 28) {
return 2;
}
else if (AFD >= 28) {
return 1;
}
}
else if (Cs >= 932) {
return 2;
}
}
}
}
else if (tcor >= 73) {
if (TSC < 12) {
if (negBalance < 39) {
if (posBalance < 55) {
if (Rs < 3008) {
return 1;
}
else if (Rs >= 3008) {
return 4;
}
}
else if (posBalance >= 55) {
if (Cs < 847) {
return 2;
}
else if (Cs >= 847) {
return 8;
}
}
}
else if (negBalance >= 39) {
if (ssDCval < 110) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
return 2;
}
}
else if (ssDCval >= 110) {
if (Cs < 596) {
return 2;
}
else if (Cs >= 596) {
return 1;
}
}
}
}
else if (TSC >= 12) {
if (Cs < 1277) {
return 1;
}
else if (Cs >= 1277) {
return 8;
}
}
}
}
}
else if (refDCval >= 112) {
if (RsCsDiff < 431) {
if (Cs < 668) {
if (TSC < 11) {
if (tcor < 75) {
return 2;
}
else if (tcor >= 75) {
if (Rs < 3294) {
if (mcTcor < 91) {
return 2;
}
else if (mcTcor >= 91) {
if (refDCval < 114) {
if (Rs < 3149) {
return 1;
}
else if (Rs >= 3149) {
return 8;
}
}
else if (refDCval >= 114) {
return 1;
}
}
}
else if (Rs >= 3294) {
if (posBalance < 63) {
return 2;
}
else if (posBalance >= 63) {
if (Rs < 4326) {
return 1;
}
else if (Rs >= 4326) {
return 2;
}
}
}
}
}
else if (TSC >= 11) {
if (Rs < 4709) {
return 1;
}
else if (Rs >= 4709) {
return 2;
}
}
}
else if (Cs >= 668) {
if (mu_mv_mag_sq < 40) {
if (tcor < 77) {
if (CsDiff < 8) {
if (gchDC < 11) {
return 8;
}
else if (gchDC >= 11) {
return 2;
}
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (tcor >= 77) {
if (AFD < 15) {
if (RsCsDiff < 269) {
if (Rs < 4165) {
return 2;
}
else if (Rs >= 4165) {
return 8;
}
}
else if (RsCsDiff >= 269) {
return 2;
}
}
else if (AFD >= 15) {
if (refDCval < 115) {
return 1;
}
else if (refDCval >= 115) {
if (ssDCval < 121) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (RsDiff < 18) {
return 1;
}
else if (RsDiff >= 18) {
if (Rs < 3963) {
return 8;
}
else if (Rs >= 3963) {
return 2;
}
}
}
}
else if (ssDCval >= 121) {
if (diffRsCsdiff < -42) {
if (RsDiff < 15) {
return 1;
}
else if (RsDiff >= 15) {
if (Rs < 4095) {
return 8;
}
else if (Rs >= 4095) {
if (Rs < 5681) {
return 2;
}
else if (Rs >= 5681) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -42) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
if (Rs < 3221) {
if (Rs < 2455) {
return 4;
}
else if (Rs >= 2455) {
return 1;
}
}
else if (Rs >= 3221) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 431) {
if (Cs < 832) {
if (mcTcor < 98) {
if (diffAFD < 12) {
return 2;
}
else if (diffAFD >= 12) {
if (Rs < 4335) {
return 8;
}
else if (Rs >= 4335) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 1;
}
}
}
else if (Cs >= 832) {
if (CsDiff < 7) {
if (posBalance < 84) {
if (refDCval < 118) {
return 2;
}
else if (refDCval >= 118) {
if (TSC < 13) {
if (TSC < 10) {
if (Rs < 4854) {
if (diffMVdiffVal < -18) {
if (Rs < 3981) {
return 8;
}
else if (Rs >= 3981) {
return 2;
}
}
else if (diffMVdiffVal >= -18) {
if (refDCval < 127) {
return 8;
}
else if (refDCval >= 127) {
if (Cs < 878) {
return 2;
}
else if (Cs >= 878) {
if (Rs < 4507) {
return 8;
}
else if (Rs >= 4507) {
return 2;
}
}
}
}
}
else if (Rs >= 4854) {
if (Cs < 989) {
return 8;
}
else if (Cs >= 989) {
if (diffTSC < -3) {
return 8;
}
else if (diffTSC >= -3) {
return 2;
}
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (TSC >= 13) {
return 2;
}
}
}
else if (posBalance >= 84) {
return 1;
}
}
else if (CsDiff >= 7) {
if (AFD < 36) {
if (mcTcor < 95) {
if (Rs < 6437) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (Cs < 1044) {
return 1;
}
else if (Cs >= 1044) {
return 2;
}
}
}
else if (Rs >= 6437) {
return 8;
}
}
else if (mcTcor >= 95) {
if (Rs < 4549) {
return 1;
}
else if (Rs >= 4549) {
if (Rs < 5532) {
return 8;
}
else if (Rs >= 5532) {
return 2;
}
}
}
}
else if (AFD >= 36) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 135) {
if (AFD < 19) {
if (mcTcor < 92) {
if (Rs < 5349) {
return 4;
}
else if (Rs >= 5349) {
return 2;
}
}
else if (mcTcor >= 92) {
if (Rs < 6459) {
if (CsDiff < 4) {
if (Rs < 3061) {
return 1;
}
else if (Rs >= 3061) {
if (Cs < 833) {
return 2;
}
else if (Cs >= 833) {
return 4;
}
}
}
else if (CsDiff >= 4) {
if (diffTSC < 3) {
return 1;
}
else if (diffTSC >= 3) {
return 2;
}
}
}
else if (Rs >= 6459) {
return 2;
}
}
}
else if (AFD >= 19) {
if (Rs < 4961) {
if (AFD < 32) {
if (ssDCval < 140) {
if (MVDiff < 42) {
if (diffAFD < 8) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 2;
}
}
else if (diffAFD >= 8) {
return 1;
}
}
else if (MVDiff >= 42) {
if (Cs < 986) {
return 1;
}
else if (Cs >= 986) {
return 8;
}
}
}
else if (ssDCval >= 140) {
if (negBalance < 75) {
if (tcor < 71) {
if (gchDC < 6) {
if (RsDiff < 16) {
return 8;
}
else if (RsDiff >= 16) {
return 1;
}
}
else if (gchDC >= 6) {
return 2;
}
}
else if (tcor >= 71) {
return 1;
}
}
else if (negBalance >= 75) {
return 2;
}
}
}
else if (AFD >= 32) {
if (diffRsCsdiff < -68) {
return 1;
}
else if (diffRsCsdiff >= -68) {
return 2;
}
}
}
else if (Rs >= 4961) {
if (Cs < 1011) {
return 2;
}
else if (Cs >= 1011) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
if (Rs < 5573) {
return 8;
}
else if (Rs >= 5573) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1332) {
if (mu_mv_mag_sq < 109) {
if (RsDiff < 9) {
if (CsDiff < 8) {
if (ssDCval < 89) {
if (negBalance < 38) {
if (Rs < 3478) {
if (MVDiff < 7) {
if (Rs < 2693) {
if (Rs < 741) {
if (Rs < 312) {
return 4;
}
else if (Rs >= 312) {
return 1;
}
}
else if (Rs >= 741) {
if (posBalance < 43) {
if (ssDCval < 76) {
if (mu_mv_mag_sq < 97) {
if (Rs < 1859) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (Rs >= 1859) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (SC < 2612) {
return 4;
}
else if (SC >= 2612) {
return 8;
}
}
}
else if (ssDCval >= 76) {
if (Rs < 1884) {
return 8;
}
else if (Rs >= 1884) {
return 1;
}
}
}
else if (posBalance >= 43) {
if (mu_mv_mag_sq < 76) {
if (SC < 2699) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 4;
}
}
else if (SC >= 2699) {
if (diffRsCsdiff < 3) {
if (Cs < 1357) {
return 2;
}
else if (Cs >= 1357) {
if (Rs < 1494) {
return 1;
}
else if (Rs >= 1494) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 3) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (Rs < 2378) {
if (SC < 4617) {
if (mu_mv_mag_sq < 88) {
if (Rs < 1020) {
return 1;
}
else if (Rs >= 1020) {
if (Cs < 2074) {
return 8;
}
else if (Cs >= 2074) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 88) {
if (diffRsCsdiff < 12) {
return 8;
}
else if (diffRsCsdiff >= 12) {
if (SC < 2832) {
return 4;
}
else if (SC >= 2832) {
return 8;
}
}
}
}
else if (SC >= 4617) {
if (Rs < 1768) {
return 1;
}
else if (Rs >= 1768) {
return 4;
}
}
}
else if (Rs >= 2378) {
if (ssDCval < 83) {
return 8;
}
else if (ssDCval >= 83) {
if (Rs < 2449) {
return 4;
}
else if (Rs >= 2449) {
if (Cs < 2762) {
return 8;
}
else if (Cs >= 2762) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 2693) {
if (Cs < 1858) {
if (RsCsDiff < 38) {
if (Rs < 3404) {
return 4;
}
else if (Rs >= 3404) {
return 2;
}
}
else if (RsCsDiff >= 38) {
if (Cs < 1543) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (Cs < 1443) {
return 2;
}
else if (Cs >= 1443) {
return 4;
}
}
}
else if (Cs >= 1543) {
if (ssDCval < 78) {
if (Rs < 3109) {
return 1;
}
else if (Rs >= 3109) {
return 2;
}
}
else if (ssDCval >= 78) {
return 4;
}
}
}
}
else if (Cs >= 1858) {
if (RsCsDiff < 57) {
if (tcor < 96) {
if (Cs < 3129) {
return 8;
}
else if (Cs >= 3129) {
return 4;
}
}
else if (tcor >= 96) {
if (Rs < 2791) {
return 1;
}
else if (Rs >= 2791) {
return 4;
}
}
}
else if (RsCsDiff >= 57) {
if (posBalance < 52) {
if (Cs < 2906) {
return 1;
}
else if (Cs >= 2906) {
return 8;
}
}
else if (posBalance >= 52) {
if (Rs < 3046) {
return 1;
}
else if (Rs >= 3046) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (mu_mv_mag_sq < 75) {
if (Rs < 1334) {
if (TSC < 5) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (TSC >= 5) {
if (Rs < 1172) {
if (Cs < 2326) {
return 8;
}
else if (Cs >= 2326) {
return 2;
}
}
else if (Rs >= 1172) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
return 1;
}
}
}
}
else if (Rs >= 1334) {
if (Rs < 1546) {
if (ssDCval < 58) {
if (Cs < 1365) {
return 2;
}
else if (Cs >= 1365) {
return 1;
}
}
else if (ssDCval >= 58) {
return 4;
}
}
else if (Rs >= 1546) {
if (Cs < 1807) {
if (posBalance < 41) {
return 2;
}
else if (posBalance >= 41) {
if (Cs < 1670) {
if (Rs < 2805) {
return 8;
}
else if (Rs >= 2805) {
return 4;
}
}
else if (Cs >= 1670) {
return 1;
}
}
}
else if (Cs >= 1807) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (refDCval < 63) {
if (Cs < 1468) {
if (mu_mv_mag_sq < 95) {
if (Rs < 2561) {
return 2;
}
else if (Rs >= 2561) {
return 4;
}
}
else if (mu_mv_mag_sq >= 95) {
if (Cs < 1408) {
return 8;
}
else if (Cs >= 1408) {
return 1;
}
}
}
else if (Cs >= 1468) {
if (mcTcor < 94) {
if (Rs < 1033) {
return 8;
}
else if (Rs >= 1033) {
if (tcor < 87) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 1944) {
return 4;
}
else if (Rs >= 1944) {
return 8;
}
}
}
else if (tcor >= 87) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (ssDCval < 60) {
if (diffAFD < 2) {
if (Rs < 675) {
return 1;
}
else if (Rs >= 675) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
return 4;
}
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (ssDCval >= 60) {
return 1;
}
}
}
}
else if (refDCval >= 63) {
if (ssDCval < 69) {
if (posBalance < 47) {
if (MVDiff < 8) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 1079) {
return 4;
}
else if (Rs >= 1079) {
return 2;
}
}
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (posBalance >= 47) {
if (SC < 3432) {
return 1;
}
else if (SC >= 3432) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
return 4;
}
}
}
}
else if (ssDCval >= 69) {
if (tcor < 91) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
else if (tcor >= 91) {
if (RsDiff < 6) {
if (Rs < 2375) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 82) {
return 1;
}
else if (mu_mv_mag_sq >= 82) {
return 8;
}
}
}
else if (Rs >= 2375) {
return 8;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 3478) {
if (posBalance < 42) {
if (CsDiff < 5) {
if (Cs < 2413) {
if (SC < 7108) {
if (Rs < 4572) {
if (Cs < 1949) {
return 4;
}
else if (Cs >= 1949) {
if (Rs < 4018) {
return 8;
}
else if (Rs >= 4018) {
return 4;
}
}
}
else if (Rs >= 4572) {
if (RsDiff < 5) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
else if (RsDiff >= 5) {
return 4;
}
}
}
else if (SC >= 7108) {
return 4;
}
}
else if (Cs >= 2413) {
if (Rs < 5459) {
return 8;
}
else if (Rs >= 5459) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 4197) {
if (Rs < 3523) {
return 8;
}
else if (Rs >= 3523) {
return 4;
}
}
else if (Rs >= 4197) {
if (diffTSC < -1) {
if (Rs < 4766) {
return 4;
}
else if (Rs >= 4766) {
return 1;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
else if (posBalance >= 42) {
if (Rs < 6022) {
if (posBalance < 46) {
if (diffRsCsdiff < -31) {
return 2;
}
else if (diffRsCsdiff >= -31) {
if (ssDCval < 87) {
if (SC < 5821) {
return 4;
}
else if (SC >= 5821) {
if (Rs < 4633) {
return 8;
}
else if (Rs >= 4633) {
if (Rs < 4675) {
return 4;
}
else if (Rs >= 4675) {
if (gchDC < 1) {
if (Rs < 4714) {
return 8;
}
else if (Rs >= 4714) {
return 4;
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
}
else if (ssDCval >= 87) {
return 2;
}
}
}
else if (posBalance >= 46) {
if (ssDCval < 71) {
if (Cs < 2935) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (Cs >= 2935) {
return 8;
}
}
else if (ssDCval >= 71) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
else if (Rs >= 6022) {
return 2;
}
}
}
}
else if (negBalance >= 38) {
if (Rs < 1462) {
if (ssDCval < 57) {
if (Cs < 1575) {
if (Rs < 1095) {
if (Rs < 851) {
return 4;
}
else if (Rs >= 851) {
if (gchDC < 1) {
if (negBalance < 46) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (negBalance >= 46) {
return 4;
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
else if (Rs >= 1095) {
if (posBalance < 57) {
if (Cs < 1443) {
return 8;
}
else if (Cs >= 1443) {
return 4;
}
}
else if (posBalance >= 57) {
if (Rs < 1284) {
if (Rs < 1210) {
return 4;
}
else if (Rs >= 1210) {
return 2;
}
}
else if (Rs >= 1284) {
return 4;
}
}
}
}
else if (Cs >= 1575) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
}
else if (ssDCval >= 57) {
if (AFD < 12) {
if (MVDiff < 5) {
if (Cs < 1582) {
if (ssDCval < 67) {
if (TSC < 7) {
if (Cs < 1444) {
return 4;
}
else if (Cs >= 1444) {
return 8;
}
}
else if (TSC >= 7) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 1340) {
return 1;
}
else if (Rs >= 1340) {
return 8;
}
}
}
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (Cs >= 1582) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (MVDiff >= 5) {
if (Rs < 1049) {
if (TSC < 5) {
if (mu_mv_mag_sq < 74) {
return 1;
}
else if (mu_mv_mag_sq >= 74) {
return 8;
}
}
else if (TSC >= 5) {
if (Cs < 1382) {
return 1;
}
else if (Cs >= 1382) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (RsCsDiff < 60) {
if (Rs < 718) {
return 2;
}
else if (Rs >= 718) {
return 8;
}
}
else if (RsCsDiff >= 60) {
return 2;
}
}
}
}
}
else if (Rs >= 1049) {
if (negBalance < 46) {
if (Rs < 1236) {
if (Cs < 1501) {
return 4;
}
else if (Cs >= 1501) {
if (Rs < 1104) {
return 1;
}
else if (Rs >= 1104) {
return 8;
}
}
}
else if (Rs >= 1236) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 1;
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
}
else if (AFD >= 12) {
if (ssDCval < 71) {
if (AFD < 18) {
if (negBalance < 44) {
return 4;
}
else if (negBalance >= 44) {
if (Rs < 910) {
return 2;
}
else if (Rs >= 910) {
return 8;
}
}
}
else if (AFD >= 18) {
if (Rs < 1172) {
if (Rs < 950) {
return 8;
}
else if (Rs >= 950) {
return 2;
}
}
else if (Rs >= 1172) {
return 4;
}
}
}
else if (ssDCval >= 71) {
if (SC < 2396) {
if (ssDCval < 76) {
if (Rs < 652) {
return 1;
}
else if (Rs >= 652) {
return 2;
}
}
else if (ssDCval >= 76) {
if (refDCval < 83) {
if (Rs < 738) {
return 8;
}
else if (Rs >= 738) {
return 4;
}
}
else if (refDCval >= 83) {
return 8;
}
}
}
else if (SC >= 2396) {
if (RsDiff < 8) {
if (mcTcor < 89) {
if (Rs < 1444) {
if (Rs < 1211) {
return 2;
}
else if (Rs >= 1211) {
return 8;
}
}
else if (Rs >= 1444) {
return 4;
}
}
else if (mcTcor >= 89) {
if (Rs < 981) {
if (SC < 2437) {
return 4;
}
else if (SC >= 2437) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (Rs >= 981) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
return 4;
}
}
}
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1462) {
if (Cs < 1537) {
if (Rs < 2788) {
if (mcTcor < 94) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (SC < 3561) {
if (refDCval < 88) {
if (diffRsCsdiff < -16) {
if (diffRsCsdiff < -19) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
return 8;
}
}
else if (diffRsCsdiff >= -19) {
if (ssDCval < 67) {
if (Rs < 1910) {
return 4;
}
else if (Rs >= 1910) {
return 8;
}
}
else if (ssDCval >= 67) {
return 1;
}
}
}
else if (diffRsCsdiff >= -16) {
if (posBalance < 73) {
if (tcor < 51) {
return 4;
}
else if (tcor >= 51) {
if (negBalance < 62) {
return 8;
}
else if (negBalance >= 62) {
if (Cs < 1481) {
return 8;
}
else if (Cs >= 1481) {
return 2;
}
}
}
}
else if (posBalance >= 73) {
return 4;
}
}
}
else if (refDCval >= 88) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
}
else if (SC >= 3561) {
if (posBalance < 42) {
return 2;
}
else if (posBalance >= 42) {
if (Rs < 2143) {
if (Cs < 1493) {
return 2;
}
else if (Cs >= 1493) {
return 8;
}
}
else if (Rs >= 2143) {
if (mu_mv_mag_sq < 2) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (tcor < 87) {
if (ssDCval < 80) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (ssDCval >= 80) {
if (MVDiff < 17) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (tcor >= 87) {
if (MVDiff < 7) {
if (ssDCval < 83) {
if (diffRsCsdiff < -15) {
return 1;
}
else if (diffRsCsdiff >= -15) {
if (ssDCval < 80) {
if (diffRsCsdiff < 5) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
else if (diffRsCsdiff >= 5) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (ssDCval >= 80) {
if (SC < 3250) {
return 1;
}
else if (SC >= 3250) {
if (RsDiff < 6) {
if (Rs < 1959) {
return 8;
}
else if (Rs >= 1959) {
if (Rs < 2108) {
return 1;
}
else if (Rs >= 2108) {
return 8;
}
}
}
else if (RsDiff >= 6) {
return 8;
}
}
}
}
}
else if (ssDCval >= 83) {
if (AFD < 8) {
if (Cs < 1430) {
return 4;
}
else if (Cs >= 1430) {
if (negBalance < 43) {
return 2;
}
else if (negBalance >= 43) {
if (Rs < 2097) {
return 1;
}
else if (Rs >= 2097) {
return 8;
}
}
}
}
else if (AFD >= 8) {
if (TSC < 8) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (Rs < 2179) {
return 2;
}
else if (Rs >= 2179) {
return 1;
}
}
}
else if (TSC >= 8) {
if (SC < 3558) {
return 1;
}
else if (SC >= 3558) {
return 8;
}
}
}
}
}
else if (MVDiff >= 7) {
if (negBalance < 46) {
if (Cs < 1385) {
return 1;
}
else if (Cs >= 1385) {
return 8;
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
}
}
else if (Rs >= 2788) {
if (tcor < 91) {
if (ssDCval < 80) {
return 4;
}
else if (ssDCval >= 80) {
if (Rs < 3285) {
return 8;
}
else if (Rs >= 3285) {
return 1;
}
}
}
else if (tcor >= 91) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
if (Rs < 3352) {
return 4;
}
else if (Rs >= 3352) {
return 2;
}
}
}
}
}
else if (Cs >= 1537) {
if (Cs < 3429) {
if (mcTcor < 93) {
if (Cs < 2980) {
if (posBalance < 70) {
if (ssDCval < 53) {
if (RsCsDiff < 80) {
return 8;
}
else if (RsCsDiff >= 80) {
if (tcor < 55) {
return 1;
}
else if (tcor >= 55) {
if (negBalance < 53) {
if (Cs < 1805) {
if (Rs < 1684) {
if (Rs < 1509) {
return 8;
}
else if (Rs >= 1509) {
return 4;
}
}
else if (Rs >= 1684) {
return 8;
}
}
else if (Cs >= 1805) {
if (Rs < 2164) {
return 4;
}
else if (Rs >= 2164) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
else if (negBalance >= 53) {
return 4;
}
}
}
}
else if (ssDCval >= 53) {
if (RsCsDiff < 87) {
if (Rs < 4231) {
if (diffMVdiffVal < -5) {
if (mu_mv_mag_sq < 90) {
return 8;
}
else if (mu_mv_mag_sq >= 90) {
if (Rs < 2558) {
if (Rs < 2151) {
return 8;
}
else if (Rs >= 2151) {
return 4;
}
}
else if (Rs >= 2558) {
return 8;
}
}
}
else if (diffMVdiffVal >= -5) {
if (gchDC < 6) {
if (Rs < 1911) {
if (Rs < 1907) {
if (Cs < 1548) {
if (Cs < 1543) {
return 8;
}
else if (Cs >= 1543) {
return 4;
}
}
else if (Cs >= 1548) {
return 8;
}
}
else if (Rs >= 1907) {
return 4;
}
}
else if (Rs >= 1911) {
return 8;
}
}
else if (gchDC >= 6) {
if (Rs < 1879) {
return 8;
}
else if (Rs >= 1879) {
return 4;
}
}
}
}
else if (Rs >= 4231) {
return 4;
}
}
else if (RsCsDiff >= 87) {
if (gchDC < 2) {
if (diffAFD < 2) {
if (Rs < 1594) {
if (Rs < 1551) {
return 8;
}
else if (Rs >= 1551) {
return 4;
}
}
else if (Rs >= 1594) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 74) {
if (AFD < 17) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (Rs < 2299) {
return 8;
}
else if (Rs >= 2299) {
return 4;
}
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (mu_mv_mag_sq >= 74) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 2015) {
return 1;
}
else if (Rs >= 2015) {
if (mu_mv_mag_sq < 68) {
return 8;
}
else if (mu_mv_mag_sq >= 68) {
if (Rs < 2331) {
return 8;
}
else if (Rs >= 2331) {
return 4;
}
}
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 77) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Rs < 1859) {
return 1;
}
else if (Rs >= 1859) {
return 2;
}
}
}
else if (ssDCval >= 77) {
return 4;
}
}
}
}
}
else if (posBalance >= 70) {
return 4;
}
}
else if (Cs >= 2980) {
if (Rs < 3067) {
return 4;
}
else if (Rs >= 3067) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (RsCsDiff < 37) {
if (RsDiff < 4) {
if (Rs < 1514) {
if (Rs < 1476) {
return 8;
}
else if (Rs >= 1476) {
return 1;
}
}
else if (Rs >= 1514) {
return 8;
}
}
else if (RsDiff >= 4) {
if (AFD < 10) {
if (CsDiff < 4) {
if (Rs < 4284) {
return 8;
}
else if (Rs >= 4284) {
return 4;
}
}
else if (CsDiff >= 4) {
if (Cs < 2007) {
return 8;
}
else if (Cs >= 2007) {
if (negBalance < 43) {
if (mu_mv_mag_sq < 100) {
return 4;
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 3754) {
return 8;
}
else if (Rs >= 3754) {
return 4;
}
}
}
else if (negBalance >= 43) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 8;
}
}
}
}
}
else if (AFD >= 10) {
if (Rs < 2991) {
return 4;
}
else if (Rs >= 2991) {
return 8;
}
}
}
}
else if (RsCsDiff >= 37) {
if (SC < 4934) {
if (Cs < 2269) {
if (tcor < 87) {
if (mu_mv_mag_sq < 57) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 4;
}
}
else if (mu_mv_mag_sq >= 57) {
return 8;
}
}
else if (tcor >= 87) {
if (refDCval < 80) {
if (SC < 3873) {
if (posBalance < 58) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Rs < 1688) {
return 8;
}
else if (Rs >= 1688) {
return 4;
}
}
}
else if (posBalance >= 58) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (SC >= 3873) {
if (mu_mv_mag_sq < 94) {
return 8;
}
else if (mu_mv_mag_sq >= 94) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
return 8;
}
}
}
}
else if (refDCval >= 80) {
if (Rs < 2945) {
if (posBalance < 44) {
if (Rs < 2302) {
if (Rs < 1682) {
return 8;
}
else if (Rs >= 1682) {
return 4;
}
}
else if (Rs >= 2302) {
return 1;
}
}
else if (posBalance >= 44) {
if (TSC < 8) {
if (SC < 4314) {
if (Rs < 2445) {
if (Cs < 2138) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
return 1;
}
}
else if (Cs >= 2138) {
return 1;
}
}
else if (Rs >= 2445) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (SC >= 4314) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (Rs >= 2945) {
if (Rs < 3142) {
return 1;
}
else if (Rs >= 3142) {
return 4;
}
}
}
}
}
else if (Cs >= 2269) {
if (Rs < 2193) {
if (Rs < 1737) {
return 8;
}
else if (Rs >= 1737) {
return 1;
}
}
else if (Rs >= 2193) {
return 8;
}
}
}
else if (SC >= 4934) {
if (Cs < 1956) {
if (RsCsDiff < 77) {
if (refDCval < 88) {
return 4;
}
else if (refDCval >= 88) {
if (Cs < 1756) {
return 1;
}
else if (Cs >= 1756) {
if (Rs < 4173) {
return 2;
}
else if (Rs >= 4173) {
return 4;
}
}
}
}
else if (RsCsDiff >= 77) {
return 2;
}
}
else if (Cs >= 1956) {
if (Rs < 2564) {
return 4;
}
else if (Rs >= 2564) {
if (ssDCval < 84) {
if (negBalance < 48) {
if (diffRsCsdiff < -14) {
if (Cs < 2916) {
return 4;
}
else if (Cs >= 2916) {
return 1;
}
}
else if (diffRsCsdiff >= -14) {
if (SC < 5505) {
return 8;
}
else if (SC >= 5505) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
if (mu_mv_mag_sq < 83) {
if (MVDiff < 12) {
if (SC < 5736) {
return 2;
}
else if (SC >= 5736) {
return 4;
}
}
else if (MVDiff >= 12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 83) {
return 8;
}
}
}
else if (ssDCval >= 84) {
if (Cs < 3008) {
return 8;
}
else if (Cs >= 3008) {
return 4;
}
}
}
}
}
}
}
}
else if (Cs >= 3429) {
return 8;
}
}
}
}
}
else if (ssDCval >= 89) {
if (ssDCval < 164) {
if (Rs < 4453) {
if (refDCval < 110) {
if (mcTcor < 96) {
if (Cs < 2658) {
if (Rs < 1422) {
if (RsCsDiff < 41) {
return 8;
}
else if (RsCsDiff >= 41) {
if (RsCsDiff < 63) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
if (SC < 2758) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
else if (SC >= 2758) {
return 8;
}
}
}
else if (RsCsDiff >= 63) {
if (Cs < 2176) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 1172) {
return 8;
}
else if (Rs >= 1172) {
return 4;
}
}
}
else if (Cs >= 2176) {
return 2;
}
}
}
}
else if (Rs >= 1422) {
if (SC < 4056) {
if (TSC < 13) {
if (TSC < 9) {
if (Cs < 2265) {
if (RsCsDiff < 94) {
if (mcTcor < 90) {
if (Rs < 1443) {
if (Rs < 1434) {
return 8;
}
else if (Rs >= 1434) {
return 4;
}
}
else if (Rs >= 1443) {
return 8;
}
}
else if (mcTcor >= 90) {
if (refDCval < 92) {
if (AFD < 8) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 2;
}
}
else if (AFD >= 8) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 2;
}
}
}
else if (refDCval >= 92) {
if (Cs < 1475) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (diffRsCsdiff < 7) {
return 8;
}
else if (diffRsCsdiff >= 7) {
if (Rs < 1529) {
return 8;
}
else if (Rs >= 1529) {
return 2;
}
}
}
}
}
else if (Cs >= 1475) {
return 8;
}
}
}
}
else if (RsCsDiff >= 94) {
if (Cs < 1607) {
return 2;
}
else if (Cs >= 1607) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (Cs >= 2265) {
if (Rs < 1666) {
return 2;
}
else if (Rs >= 1666) {
return 1;
}
}
}
else if (TSC >= 9) {
if (Rs < 1617) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
}
else if (Rs >= 1617) {
if (posBalance < 49) {
if (Rs < 1941) {
return 8;
}
else if (Rs >= 1941) {
return 1;
}
}
else if (posBalance >= 49) {
if (CsDiff < 5) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 1;
}
}
else if (CsDiff >= 5) {
if (Cs < 1344) {
if (Rs < 1784) {
return 8;
}
else if (Rs >= 1784) {
return 1;
}
}
else if (Cs >= 1344) {
return 8;
}
}
}
}
}
}
else if (TSC >= 13) {
if (Rs < 1723) {
return 8;
}
else if (Rs >= 1723) {
return 1;
}
}
}
else if (SC >= 4056) {
if (mcTcor < 95) {
if (Rs < 1980) {
if (Rs < 1864) {
return 4;
}
else if (Rs >= 1864) {
return 1;
}
}
else if (Rs >= 1980) {
if (RsDiff < 8) {
if (Rs < 4329) {
if (posBalance < 33) {
if (Rs < 2717) {
if (Rs < 2320) {
return 1;
}
else if (Rs >= 2320) {
return 4;
}
}
else if (Rs >= 2717) {
return 8;
}
}
else if (posBalance >= 33) {
if (refDCval < 107) {
return 8;
}
else if (refDCval >= 107) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
if (mu_mv_mag_sq < 77) {
if (Rs < 2719) {
return 8;
}
else if (Rs >= 2719) {
return 4;
}
}
else if (mu_mv_mag_sq >= 77) {
return 8;
}
}
}
}
}
else if (Rs >= 4329) {
if (Rs < 4382) {
return 1;
}
else if (Rs >= 4382) {
return 4;
}
}
}
else if (RsDiff >= 8) {
if (diffRsCsdiff < -53) {
return 4;
}
else if (diffRsCsdiff >= -53) {
if (diffTSC < 2) {
if (Cs < 2311) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Rs < 3346) {
return 8;
}
else if (Rs >= 3346) {
return 1;
}
}
}
else if (Cs >= 2311) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
if (ssDCval < 100) {
return 8;
}
else if (ssDCval >= 100) {
return 4;
}
}
}
}
else if (diffTSC >= 2) {
if (Rs < 3323) {
return 8;
}
else if (Rs >= 3323) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (Rs < 3306) {
if (SC < 4304) {
return 8;
}
else if (SC >= 4304) {
if (diffMVdiffVal < 4) {
if (negBalance < 54) {
if (posBalance < 47) {
if (ssDCval < 100) {
if (Rs < 2421) {
return 4;
}
else if (Rs >= 2421) {
return 8;
}
}
else if (ssDCval >= 100) {
return 4;
}
}
else if (posBalance >= 47) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (negBalance >= 54) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 4;
}
}
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
}
else if (Rs >= 3306) {
if (MVDiff < 5) {
if (AFD < 9) {
if (Rs < 4141) {
return 4;
}
else if (Rs >= 4141) {
return 1;
}
}
else if (AFD >= 9) {
return 2;
}
}
else if (MVDiff >= 5) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 2658) {
if (ssDCval < 101) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (RsCsDiff < 55) {
if (MVDiff < 4) {
if (posBalance < 42) {
if (Cs < 1513) {
if (RsCsDiff < 37) {
return 4;
}
else if (RsCsDiff >= 37) {
if (Rs < 3020) {
return 1;
}
else if (Rs >= 3020) {
return 2;
}
}
}
else if (Cs >= 1513) {
if (mcTcor < 97) {
return 4;
}
else if (mcTcor >= 97) {
if (posBalance < 17) {
if (diffMVdiffVal < -4) {
return 1;
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
else if (posBalance >= 17) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
if (MVDiff < 1) {
if (negBalance < 31) {
if (mu_mv_mag_sq < 107) {
if (Rs < 2275) {
return 8;
}
else if (Rs >= 2275) {
return 4;
}
}
else if (mu_mv_mag_sq >= 107) {
return 8;
}
}
else if (negBalance >= 31) {
return 8;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 42) {
if (AFD < 5) {
if (Rs < 1333) {
return 4;
}
else if (Rs >= 1333) {
return 8;
}
}
else if (AFD >= 5) {
if (refDCval < 97) {
if (negBalance < 51) {
if (negBalance < 39) {
if (Cs < 1630) {
if (MVDiff < 2) {
if (Rs < 4143) {
return 2;
}
else if (Rs >= 4143) {
return 8;
}
}
else if (MVDiff >= 2) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (Cs >= 1630) {
if (Cs < 2657) {
return 8;
}
else if (Cs >= 2657) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (diffMVdiffVal < -11) {
return 2;
}
else if (diffMVdiffVal >= -11) {
return 8;
}
}
}
else if (negBalance >= 51) {
return 4;
}
}
else if (refDCval >= 97) {
if (TSC < 6) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (Rs < 2716) {
if (Rs < 982) {
return 4;
}
else if (Rs >= 982) {
return 8;
}
}
else if (Rs >= 2716) {
return 4;
}
}
}
else if (TSC >= 6) {
if (refDCval < 103) {
if (Rs < 1985) {
return 2;
}
else if (Rs >= 1985) {
if (Rs < 2560) {
return 8;
}
else if (Rs >= 2560) {
return 4;
}
}
}
else if (refDCval >= 103) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (posBalance < 35) {
if (negBalance < 40) {
if (negBalance < 28) {
if (Rs < 2280) {
return 8;
}
else if (Rs >= 2280) {
return 4;
}
}
else if (negBalance >= 28) {
if (tcor < 101) {
return 1;
}
else if (tcor >= 101) {
return 4;
}
}
}
else if (negBalance >= 40) {
if (SC < 4240) {
if (Rs < 2116) {
if (Rs < 2009) {
return 8;
}
else if (Rs >= 2009) {
return 1;
}
}
else if (Rs >= 2116) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 1;
}
}
}
else if (SC >= 4240) {
return 8;
}
}
}
else if (posBalance >= 35) {
if (refDCval < 107) {
if (diffTSC < -1) {
if (Rs < 2077) {
if (Rs < 1645) {
if (Cs < 1377) {
return 4;
}
else if (Cs >= 1377) {
return 8;
}
}
else if (Rs >= 1645) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 2;
}
}
}
else if (Rs >= 2077) {
return 8;
}
}
else if (diffTSC >= -1) {
if (refDCval < 90) {
if (Cs < 1579) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
else if (Cs >= 1579) {
return 8;
}
}
else if (refDCval >= 90) {
if (tcor < 91) {
if (Rs < 2201) {
return 8;
}
else if (Rs >= 2201) {
return 4;
}
}
else if (tcor >= 91) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (Rs < 4146) {
return 8;
}
else if (Rs >= 4146) {
if (Rs < 4282) {
return 4;
}
else if (Rs >= 4282) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 107) {
if (posBalance < 54) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (diffRsCsdiff < 8) {
return 4;
}
else if (diffRsCsdiff >= 8) {
return 8;
}
}
}
else if (posBalance >= 54) {
if (Rs < 2326) {
return 2;
}
else if (Rs >= 2326) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 55) {
if (Rs < 3407) {
if (Rs < 2210) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 1852) {
if (Rs < 1916) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (Cs < 1495) {
if (Rs < 1694) {
return 8;
}
else if (Rs >= 1694) {
return 2;
}
}
else if (Cs >= 1495) {
return 2;
}
}
}
else if (Rs >= 1916) {
if (Cs < 1493) {
return 1;
}
else if (Cs >= 1493) {
return 8;
}
}
}
else if (Cs >= 1852) {
if (Rs < 1176) {
return 4;
}
else if (Rs >= 1176) {
return 2;
}
}
}
}
else if (Rs >= 2210) {
if (tcor < 92) {
if (negBalance < 38) {
if (Rs < 2461) {
return 2;
}
else if (Rs >= 2461) {
return 8;
}
}
else if (negBalance >= 38) {
return 4;
}
}
else if (tcor >= 92) {
if (Cs < 2636) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Cs < 1352) {
return 2;
}
else if (Cs >= 1352) {
if (SC < 4649) {
return 8;
}
else if (SC >= 4649) {
if (ssDCval < 97) {
return 1;
}
else if (ssDCval >= 97) {
return 8;
}
}
}
}
}
else if (Cs >= 2636) {
if (Cs < 3515) {
return 4;
}
else if (Cs >= 3515) {
return 8;
}
}
}
}
}
else if (Rs >= 3407) {
if (Cs < 1922) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 8;
}
}
}
else if (Cs >= 1922) {
return 8;
}
}
}
}
}
else if (refDCval >= 110) {
if (Rs < 1554) {
if (mu_mv_mag_sq < 46) {
if (ssDCval < 141) {
if (AFD < 17) {
if (Rs < 1538) {
if (CsDiff < 5) {
if (refDCval < 122) {
return 8;
}
else if (refDCval >= 122) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (CsDiff >= 5) {
if (RsCsDiff < 92) {
if (diffMVdiffVal < -7) {
if (Rs < 1398) {
return 8;
}
else if (Rs >= 1398) {
return 4;
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
else if (RsCsDiff >= 92) {
return 4;
}
}
}
else if (Rs >= 1538) {
if (Cs < 1580) {
if (Rs < 1546) {
return 2;
}
else if (Rs >= 1546) {
return 8;
}
}
else if (Cs >= 1580) {
return 4;
}
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (ssDCval >= 141) {
if (mu_mv_mag_sq < 38) {
if (Cs < 1456) {
return 1;
}
else if (Cs >= 1456) {
if (Rs < 1094) {
return 1;
}
else if (Rs >= 1094) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (TSC < 8) {
if (Rs < 1154) {
return 1;
}
else if (Rs >= 1154) {
return 8;
}
}
else if (TSC >= 8) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (Rs < 1262) {
if (Cs < 2031) {
if (SC < 2699) {
if (ssDCval < 116) {
if (Rs < 1042) {
if (Rs < 750) {
return 1;
}
else if (Rs >= 750) {
return 4;
}
}
else if (Rs >= 1042) {
return 8;
}
}
else if (ssDCval >= 116) {
if (Rs < 1016) {
return 8;
}
else if (Rs >= 1016) {
if (mu_mv_mag_sq < 96) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 1;
}
}
else if (mu_mv_mag_sq >= 96) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 1028) {
return 4;
}
else if (Rs >= 1028) {
return 8;
}
}
}
}
}
}
else if (SC >= 2699) {
if (Cs < 1774) {
if (RsDiff < 4) {
if (TSC < 4) {
if (Cs < 1612) {
return 4;
}
else if (Cs >= 1612) {
return 8;
}
}
else if (TSC >= 4) {
if (ssDCval < 134) {
return 4;
}
else if (ssDCval >= 134) {
return 8;
}
}
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (Cs >= 1774) {
return 8;
}
}
}
else if (Cs >= 2031) {
return 1;
}
}
else if (Rs >= 1262) {
if (AFD < 13) {
if (refDCval < 157) {
if (AFD < 3) {
if (Rs < 1302) {
return 4;
}
else if (Rs >= 1302) {
return 8;
}
}
else if (AFD >= 3) {
return 8;
}
}
else if (refDCval >= 157) {
if (Cs < 1804) {
return 2;
}
else if (Cs >= 1804) {
return 8;
}
}
}
else if (AFD >= 13) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (Rs < 1501) {
return 4;
}
else if (Rs >= 1501) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1554) {
if (mcTcor < 92) {
if (Rs < 3639) {
if (mu_mv_mag_sq < 16) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (Rs < 1936) {
if (ssDCval < 149) {
if (Cs < 1506) {
return 4;
}
else if (Cs >= 1506) {
return 1;
}
}
else if (ssDCval >= 149) {
return 2;
}
}
else if (Rs >= 1936) {
if (SC < 3796) {
return 1;
}
else if (SC >= 3796) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (CsDiff < 5) {
if (ssDCval < 139) {
if (refDCval < 124) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 1;
}
}
else if (refDCval >= 124) {
if (Rs < 2461) {
return 1;
}
else if (Rs >= 2461) {
return 8;
}
}
}
else if (ssDCval >= 139) {
return 8;
}
}
else if (CsDiff >= 5) {
if (ssDCval < 163) {
if (diffRsCsdiff < 73) {
return 8;
}
else if (diffRsCsdiff >= 73) {
return 1;
}
}
else if (ssDCval >= 163) {
if (Rs < 2333) {
return 2;
}
else if (Rs >= 2333) {
return 8;
}
}
}
}
}
else if (Rs >= 3639) {
if (ssDCval < 137) {
return 8;
}
else if (ssDCval >= 137) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (TSC < 8) {
if (posBalance < 38) {
if (tcor < 89) {
if (Rs < 3661) {
if (Rs < 2520) {
return 8;
}
else if (Rs >= 2520) {
return 4;
}
}
else if (Rs >= 3661) {
return 1;
}
}
else if (tcor >= 89) {
if (mu_mv_mag_sq < 108) {
if (mu_mv_mag_sq < 94) {
return 8;
}
else if (mu_mv_mag_sq >= 94) {
if (gchDC < 5) {
if (Rs < 2767) {
return 8;
}
else if (Rs >= 2767) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 2;
}
}
}
else if (gchDC >= 5) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (Cs < 2181) {
return 8;
}
else if (Cs >= 2181) {
return 4;
}
}
}
}
else if (posBalance >= 38) {
if (mu_mv_mag_sq < 7) {
if (Rs < 1900) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
if (Rs < 1727) {
return 2;
}
else if (Rs >= 1727) {
return 8;
}
}
}
}
else if (Rs >= 1900) {
if (SC < 4978) {
return 8;
}
else if (SC >= 4978) {
if (SC < 5076) {
return 2;
}
else if (SC >= 5076) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 1415) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
if (MVDiff < 6) {
if (Rs < 2710) {
return 1;
}
else if (Rs >= 2710) {
return 2;
}
}
else if (MVDiff >= 6) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
}
}
else if (Cs >= 1415) {
if (diffTSC < -2) {
if (MVDiff < 5) {
if (Rs < 3090) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (Rs >= 3090) {
return 1;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (diffTSC >= -2) {
return 8;
}
}
}
}
}
else if (TSC >= 8) {
if (posBalance < 65) {
if (mu_mv_mag_sq < 106) {
if (refDCval < 128) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
if (mu_mv_mag_sq < 71) {
if (SC < 4067) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
else if (SC >= 4067) {
return 1;
}
}
else if (mu_mv_mag_sq >= 71) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (Rs < 2341) {
return 2;
}
else if (Rs >= 2341) {
return 1;
}
}
}
}
}
else if (refDCval >= 128) {
return 8;
}
}
else if (mu_mv_mag_sq >= 106) {
if (posBalance < 56) {
if (Rs < 3480) {
return 8;
}
else if (Rs >= 3480) {
return 4;
}
}
else if (posBalance >= 56) {
return 1;
}
}
}
else if (posBalance >= 65) {
if (SC < 5266) {
if (Rs < 2556) {
return 4;
}
else if (Rs >= 2556) {
return 8;
}
}
else if (SC >= 5266) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 4453) {
if (Cs < 1590) {
if (tcor < 94) {
if (Rs < 4826) {
return 1;
}
else if (Rs >= 4826) {
return 8;
}
}
else if (tcor >= 94) {
return 2;
}
}
else if (Cs >= 1590) {
if (TSC < 8) {
if (Cs < 1892) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (MVDiff < 8) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (MVDiff >= 8) {
return 1;
}
}
}
else if (Cs >= 1892) {
if (Rs < 6274) {
return 8;
}
else if (Rs >= 6274) {
if (Rs < 8183) {
return 4;
}
else if (Rs >= 8183) {
return 8;
}
}
}
}
else if (TSC >= 8) {
if (AFD < 11) {
if (Rs < 4999) {
return 1;
}
else if (Rs >= 4999) {
return 8;
}
}
else if (AFD >= 11) {
return 8;
}
}
}
}
}
else if (ssDCval >= 164) {
if (Rs < 2283) {
if (Rs < 1009) {
return 1;
}
else if (Rs >= 1009) {
return 2;
}
}
else if (Rs >= 2283) {
return 8;
}
}
}
}
else if (CsDiff >= 8) {
if (Rs < 3833) {
if (Rs < 1525) {
if (negBalance < 44) {
if (MVDiff < 18) {
if (posBalance < 62) {
if (mcTcor < 97) {
if (Cs < 3382) {
if (ssDCval < 108) {
if (refDCval < 64) {
if (Rs < 1122) {
if (Rs < 900) {
return 8;
}
else if (Rs >= 900) {
return 1;
}
}
else if (Rs >= 1122) {
return 4;
}
}
else if (refDCval >= 64) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (ssDCval >= 108) {
return 1;
}
}
else if (Cs >= 3382) {
return 1;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (posBalance >= 62) {
return 4;
}
}
else if (MVDiff >= 18) {
if (RsCsDiff < 251) {
if (ssDCval < 60) {
if (RsDiff < 7) {
return 2;
}
else if (RsDiff >= 7) {
if (Rs < 1288) {
return 4;
}
else if (Rs >= 1288) {
return 1;
}
}
}
else if (ssDCval >= 60) {
if (Rs < 1417) {
if (ssDCval < 160) {
if (mu_mv_mag_sq < 11) {
if (CsDiff < 10) {
if (Rs < 1280) {
if (MVDiff < 19) {
return 4;
}
else if (MVDiff >= 19) {
return 8;
}
}
else if (Rs >= 1280) {
return 4;
}
}
else if (CsDiff >= 10) {
if (Rs < 691) {
return 8;
}
else if (Rs >= 691) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (Cs < 3977) {
if (Rs < 761) {
if (Rs < 584) {
return 8;
}
else if (Rs >= 584) {
return 4;
}
}
else if (Rs >= 761) {
return 8;
}
}
else if (Cs >= 3977) {
return 1;
}
}
}
else if (ssDCval >= 160) {
return 1;
}
}
else if (Rs >= 1417) {
if (Rs < 1438) {
return 2;
}
else if (Rs >= 1438) {
if (Rs < 1458) {
return 8;
}
else if (Rs >= 1458) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 251) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 2;
}
}
}
}
else if (negBalance >= 44) {
if (ssDCval < 93) {
if (posBalance < 60) {
if (mcTcor < 93) {
if (mu_mv_mag_sq < 28) {
if (negBalance < 64) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (ssDCval < 71) {
if (SC < 2649) {
return 4;
}
else if (SC >= 2649) {
return 2;
}
}
else if (ssDCval >= 71) {
if (Cs < 1701) {
return 1;
}
else if (Cs >= 1701) {
return 4;
}
}
}
}
else if (negBalance >= 64) {
return 2;
}
}
else if (mu_mv_mag_sq >= 28) {
if (CsDiff < 17) {
if (SC < 2744) {
if (Cs < 1617) {
return 8;
}
else if (Cs >= 1617) {
return 4;
}
}
else if (SC >= 2744) {
return 4;
}
}
else if (CsDiff >= 17) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (RsCsDiff < 93) {
if (Cs < 1799) {
return 8;
}
else if (Cs >= 1799) {
return 4;
}
}
else if (RsCsDiff >= 93) {
if (mu_mv_mag_sq < 5) {
if (Rs < 853) {
if (Cs < 1731) {
return 8;
}
else if (Cs >= 1731) {
return 1;
}
}
else if (Rs >= 853) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
if (Rs < 1189) {
return 2;
}
else if (Rs >= 1189) {
if (Rs < 1377) {
return 1;
}
else if (Rs >= 1377) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 60) {
if (mu_mv_mag_sq < 9) {
if (ssDCval < 92) {
if (Cs < 1335) {
return 2;
}
else if (Cs >= 1335) {
if (MVDiff < 29) {
return 4;
}
else if (MVDiff >= 29) {
if (TSC < 12) {
if (Rs < 670) {
return 4;
}
else if (Rs >= 670) {
return 2;
}
}
else if (TSC >= 12) {
return 4;
}
}
}
}
else if (ssDCval >= 92) {
return 8;
}
}
else if (mu_mv_mag_sq >= 9) {
if (mu_mv_mag_sq < 33) {
if (AFD < 18) {
if (Rs < 1353) {
return 4;
}
else if (Rs >= 1353) {
return 1;
}
}
else if (AFD >= 18) {
if (CsDiff < 10) {
if (Rs < 1345) {
return 4;
}
else if (Rs >= 1345) {
return 1;
}
}
else if (CsDiff >= 10) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 33) {
return 4;
}
}
}
}
else if (ssDCval >= 93) {
if (tcor < 66) {
if (Rs < 1298) {
if (CsDiff < 18) {
return 2;
}
else if (CsDiff >= 18) {
return 4;
}
}
else if (Rs >= 1298) {
if (AFD < 28) {
if (Cs < 2214) {
return 8;
}
else if (Cs >= 2214) {
return 1;
}
}
else if (AFD >= 28) {
return 4;
}
}
}
else if (tcor >= 66) {
if (Rs < 721) {
if (Cs < 1575) {
return 4;
}
else if (Cs >= 1575) {
return 2;
}
}
else if (Rs >= 721) {
if (RsDiff < 8) {
if (diffMVdiffVal < 8) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
return 4;
}
}
else if (diffMVdiffVal >= 8) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Cs < 1535) {
return 1;
}
else if (Cs >= 1535) {
return 4;
}
}
}
}
else if (RsDiff >= 8) {
if (Rs < 1441) {
if (Rs < 1318) {
return 2;
}
else if (Rs >= 1318) {
return 4;
}
}
else if (Rs >= 1441) {
if (Rs < 1478) {
return 1;
}
else if (Rs >= 1478) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1525) {
if (Cs < 2824) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (AFD < 17) {
if (Rs < 2973) {
if (ssDCval < 52) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (tcor < 56) {
if (Rs < 1595) {
return 4;
}
else if (Rs >= 1595) {
if (Rs < 1658) {
return 1;
}
else if (Rs >= 1658) {
return 2;
}
}
}
else if (tcor >= 56) {
if (Cs < 1844) {
if (Rs < 1921) {
return 4;
}
else if (Rs >= 1921) {
return 8;
}
}
else if (Cs >= 1844) {
return 4;
}
}
}
}
else if (ssDCval >= 52) {
if (mu_mv_mag_sq < 5) {
if (negBalance < 40) {
if (Rs < 1774) {
return 8;
}
else if (Rs >= 1774) {
return 1;
}
}
else if (negBalance >= 40) {
if (MVDiff < 4) {
if (Rs < 1984) {
return 8;
}
else if (Rs >= 1984) {
if (Rs < 2399) {
return 1;
}
else if (Rs >= 2399) {
return 2;
}
}
}
else if (MVDiff >= 4) {
if (Rs < 1906) {
if (AFD < 16) {
if (ssDCval < 80) {
if (Rs < 1669) {
return 4;
}
else if (Rs >= 1669) {
return 1;
}
}
else if (ssDCval >= 80) {
return 8;
}
}
else if (AFD >= 16) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
else if (Rs >= 1906) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (Rs < 2081) {
if (Cs < 2162) {
return 2;
}
else if (Cs >= 2162) {
return 8;
}
}
else if (Rs >= 2081) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (posBalance < 48) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
if (Cs < 2202) {
return 8;
}
else if (Cs >= 2202) {
return 1;
}
}
}
else if (posBalance >= 48) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (tcor < 89) {
if (mu_mv_mag_sq < 49) {
return 8;
}
else if (mu_mv_mag_sq >= 49) {
if (mcTcor < 93) {
if (Cs < 2717) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (tcor < 83) {
if (mu_mv_mag_sq < 50) {
return 1;
}
else if (mu_mv_mag_sq >= 50) {
return 8;
}
}
else if (tcor >= 83) {
if (Rs < 2608) {
if (Rs < 1914) {
return 4;
}
else if (Rs >= 1914) {
return 8;
}
}
else if (Rs >= 2608) {
return 4;
}
}
}
}
else if (Cs >= 2717) {
if (Rs < 2253) {
return 4;
}
else if (Rs >= 2253) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (Rs < 2381) {
return 1;
}
else if (Rs >= 2381) {
return 4;
}
}
}
}
}
else if (tcor >= 89) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2973) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (MVDiff < 8) {
if (posBalance < 51) {
if (Cs < 1801) {
return 2;
}
else if (Cs >= 1801) {
return 1;
}
}
else if (posBalance >= 51) {
if (ssDCval < 115) {
return 4;
}
else if (ssDCval >= 115) {
if (Rs < 3038) {
return 8;
}
else if (Rs >= 3038) {
return 1;
}
}
}
}
else if (MVDiff >= 8) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (Rs < 3375) {
return 4;
}
else if (Rs >= 3375) {
return 8;
}
}
}
}
}
}
else if (AFD >= 17) {
if (refDCval < 94) {
if (tcor < 78) {
if (diffRsCsdiff < 46) {
if (ssDCval < 64) {
if (Rs < 1967) {
return 2;
}
else if (Rs >= 1967) {
return 1;
}
}
else if (ssDCval >= 64) {
if (diffMVdiffVal < 10) {
if (diffMVdiffVal < -3) {
if (TSC < 10) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
return 2;
}
}
else if (TSC >= 10) {
if (mu_mv_mag_sq < 21) {
return 4;
}
else if (mu_mv_mag_sq >= 21) {
return 1;
}
}
}
else if (diffMVdiffVal >= -3) {
if (AFD < 22) {
if (diffRsCsdiff < -37) {
if (Cs < 2474) {
return 8;
}
else if (Cs >= 2474) {
return 4;
}
}
else if (diffRsCsdiff >= -37) {
if (Rs < 2193) {
return 4;
}
else if (Rs >= 2193) {
if (Rs < 2224) {
return 8;
}
else if (Rs >= 2224) {
return 4;
}
}
}
}
else if (AFD >= 22) {
return 8;
}
}
}
else if (diffMVdiffVal >= 10) {
return 1;
}
}
}
else if (diffRsCsdiff >= 46) {
return 8;
}
}
else if (tcor >= 78) {
if (negBalance < 51) {
if (ssDCval < 88) {
if (Rs < 1837) {
return 1;
}
else if (Rs >= 1837) {
return 2;
}
}
else if (ssDCval >= 88) {
return 1;
}
}
else if (negBalance >= 51) {
return 8;
}
}
}
else if (refDCval >= 94) {
if (negBalance < 55) {
if (mu_mv_mag_sq < 7) {
if (SC < 3653) {
return 4;
}
else if (SC >= 3653) {
if (refDCval < 108) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (Rs < 2796) {
return 8;
}
else if (Rs >= 2796) {
return 4;
}
}
}
else if (refDCval >= 108) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
else if (negBalance >= 55) {
if (ssDCval < 120) {
if (MVDiff < 48) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (refDCval < 95) {
if (Rs < 1715) {
return 1;
}
else if (Rs >= 1715) {
return 8;
}
}
else if (refDCval >= 95) {
return 8;
}
}
}
else if (MVDiff >= 48) {
return 1;
}
}
else if (ssDCval >= 120) {
if (TSC < 13) {
if (Cs < 2112) {
return 1;
}
else if (Cs >= 2112) {
return 8;
}
}
else if (TSC >= 13) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 2824) {
if (ssDCval < 107) {
if (mcTcor < 94) {
if (negBalance < 46) {
if (Rs < 3458) {
if (Cs < 2908) {
if (Cs < 2880) {
return 4;
}
else if (Cs >= 2880) {
return 8;
}
}
else if (Cs >= 2908) {
return 4;
}
}
else if (Rs >= 3458) {
if (Rs < 3599) {
return 1;
}
else if (Rs >= 3599) {
return 8;
}
}
}
else if (negBalance >= 46) {
if (ssDCval < 81) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
if (Rs < 2480) {
if (Rs < 2127) {
return 8;
}
else if (Rs >= 2127) {
return 1;
}
}
else if (Rs >= 2480) {
return 8;
}
}
}
else if (ssDCval >= 81) {
if (Cs < 3627) {
if (tcor < 78) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (tcor >= 78) {
return 4;
}
}
else if (Cs >= 3627) {
if (refDCval < 95) {
return 8;
}
else if (refDCval >= 95) {
if (Rs < 2618) {
return 4;
}
else if (Rs >= 2618) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (RsCsDiff < 128) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (Rs < 3004) {
return 4;
}
else if (Rs >= 3004) {
return 8;
}
}
}
}
else if (RsCsDiff >= 128) {
if (Rs < 2817) {
return 8;
}
else if (Rs >= 2817) {
return 1;
}
}
}
}
else if (ssDCval >= 107) {
if (mcTcor < 94) {
if (diffTSC < -1) {
if (Rs < 2306) {
return 2;
}
else if (Rs >= 2306) {
return 1;
}
}
else if (diffTSC >= -1) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (mcTcor >= 94) {
return 8;
}
}
}
}
}
else if (Rs >= 3833) {
if (AFD < 12) {
if (Cs < 3834) {
if (ssDCval < 99) {
if (Rs < 4548) {
if (Rs < 4217) {
return 4;
}
else if (Rs >= 4217) {
return 8;
}
}
else if (Rs >= 4548) {
return 1;
}
}
else if (ssDCval >= 99) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 47) {
if (Rs < 4896) {
return 8;
}
else if (Rs >= 4896) {
return 1;
}
}
else if (mu_mv_mag_sq >= 47) {
if (SC < 6725) {
if (Cs < 2508) {
return 1;
}
else if (Cs >= 2508) {
return 8;
}
}
else if (SC >= 6725) {
if (CsDiff < 10) {
if (Rs < 4877) {
if (mu_mv_mag_sq < 97) {
return 8;
}
else if (mu_mv_mag_sq >= 97) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (Rs >= 4877) {
return 1;
}
}
else if (CsDiff >= 10) {
return 1;
}
}
}
}
}
}
else if (Cs >= 3834) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 83) {
return 8;
}
else if (mu_mv_mag_sq >= 83) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
if (Cs < 1926) {
if (Rs < 3341) {
if (ssDCval < 111) {
if (negBalance < 48) {
if (mcTcor < 86) {
if (posBalance < 54) {
if (Cs < 1687) {
if (ssDCval < 42) {
if (Rs < 2216) {
return 8;
}
else if (Rs >= 2216) {
return 1;
}
}
else if (ssDCval >= 42) {
return 1;
}
}
else if (Cs >= 1687) {
if (Rs < 2072) {
return 4;
}
else if (Rs >= 2072) {
if (MVDiff < 25) {
return 8;
}
else if (MVDiff >= 25) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
if (Rs < 1979) {
if (AFD < 21) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
return 8;
}
}
else if (AFD >= 21) {
if (Rs < 1457) {
return 4;
}
else if (Rs >= 1457) {
return 2;
}
}
}
else if (Rs >= 1979) {
return 8;
}
}
}
else if (mcTcor >= 86) {
if (Rs < 2448) {
if (mu_mv_mag_sq < 14) {
if (tcor < 85) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (mcTcor < 87) {
return 1;
}
else if (mcTcor >= 87) {
return 2;
}
}
}
else if (tcor >= 85) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Cs < 1417) {
return 8;
}
else if (Cs >= 1417) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (AFD < 20) {
if (mu_mv_mag_sq < 56) {
return 8;
}
else if (mu_mv_mag_sq >= 56) {
if (AFD < 12) {
if (Cs < 1556) {
return 8;
}
else if (Cs >= 1556) {
return 4;
}
}
else if (AFD >= 12) {
return 1;
}
}
}
else if (AFD >= 20) {
if (RsDiff < 12) {
return 2;
}
else if (RsDiff >= 12) {
if (Rs < 1825) {
return 4;
}
else if (Rs >= 1825) {
return 1;
}
}
}
}
}
else if (Rs >= 2448) {
if (ssDCval < 101) {
if (posBalance < 54) {
if (posBalance < 38) {
if (Cs < 1653) {
return 8;
}
else if (Cs >= 1653) {
return 1;
}
}
else if (posBalance >= 38) {
if (Cs < 1377) {
if (Rs < 3172) {
return 2;
}
else if (Rs >= 3172) {
return 4;
}
}
else if (Cs >= 1377) {
if (Cs < 1907) {
return 2;
}
else if (Cs >= 1907) {
if (Rs < 3057) {
return 1;
}
else if (Rs >= 3057) {
return 2;
}
}
}
}
}
else if (posBalance >= 54) {
if (Rs < 2985) {
if (Cs < 1609) {
if (negBalance < 42) {
if (Rs < 2958) {
return 2;
}
else if (Rs >= 2958) {
return 8;
}
}
else if (negBalance >= 42) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 4;
}
}
}
else if (Cs >= 1609) {
if (tcor < 78) {
if (Rs < 2786) {
return 2;
}
else if (Rs >= 2786) {
return 8;
}
}
else if (tcor >= 78) {
return 8;
}
}
}
else if (Rs >= 2985) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
return 2;
}
}
}
}
else if (ssDCval >= 101) {
if (diffAFD < -6) {
return 1;
}
else if (diffAFD >= -6) {
return 8;
}
}
}
}
}
else if (negBalance >= 48) {
if (AFD < 20) {
if (mu_mv_mag_sq < 7) {
if (SC < 4496) {
if (ssDCval < 84) {
if (Rs < 2829) {
if (CsDiff < 7) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 4;
}
}
else if (CsDiff >= 7) {
if (negBalance < 62) {
if (Rs < 1492) {
if (Cs < 1443) {
return 4;
}
else if (Cs >= 1443) {
return 2;
}
}
else if (Rs >= 1492) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
if (mcTcor < 91) {
if (MVDiff < 22) {
if (RsCsDiff < 147) {
return 8;
}
else if (RsCsDiff >= 147) {
if (Cs < 1588) {
return 8;
}
else if (Cs >= 1588) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
return 8;
}
}
}
}
else if (MVDiff >= 22) {
return 4;
}
}
else if (mcTcor >= 91) {
return 4;
}
}
}
}
else if (negBalance >= 62) {
return 8;
}
}
}
else if (Rs >= 2829) {
return 2;
}
}
else if (ssDCval >= 84) {
if (negBalance < 50) {
if (Rs < 2574) {
if (Cs < 1778) {
return 2;
}
else if (Cs >= 1778) {
return 8;
}
}
else if (Rs >= 2574) {
return 4;
}
}
else if (negBalance >= 50) {
if (mcTcor < 97) {
if (gchDC < 5) {
if (Rs < 3064) {
if (AFD < 14) {
if (Rs < 2325) {
return 4;
}
else if (Rs >= 2325) {
return 8;
}
}
else if (AFD >= 14) {
return 8;
}
}
else if (Rs >= 3064) {
return 4;
}
}
else if (gchDC >= 5) {
if (Rs < 1515) {
return 2;
}
else if (Rs >= 1515) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (Rs < 1838) {
return 1;
}
else if (Rs >= 1838) {
return 2;
}
}
}
}
}
else if (SC >= 4496) {
if (refDCval < 74) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 2;
}
}
else if (refDCval >= 74) {
if (TSC < 9) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (Rs < 3200) {
return 4;
}
else if (Rs >= 3200) {
return 8;
}
}
}
else if (TSC >= 9) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (negBalance < 55) {
return 4;
}
else if (negBalance >= 55) {
if (Rs < 3022) {
if (Rs < 2743) {
return 4;
}
else if (Rs >= 2743) {
return 8;
}
}
else if (Rs >= 3022) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (TSC < 9) {
if (ssDCval < 75) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 1;
}
}
else if (ssDCval >= 75) {
if (tcor < 90) {
if (tcor < 78) {
if (Rs < 1811) {
return 8;
}
else if (Rs >= 1811) {
return 2;
}
}
else if (tcor >= 78) {
return 8;
}
}
else if (tcor >= 90) {
if (Rs < 2567) {
return 4;
}
else if (Rs >= 2567) {
return 2;
}
}
}
}
else if (TSC >= 9) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 55) {
if (tcor < 60) {
return 4;
}
else if (tcor >= 60) {
if (MVDiff < 19) {
if (posBalance < 51) {
if (Rs < 2469) {
return 8;
}
else if (Rs >= 2469) {
return 2;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (MVDiff >= 19) {
if (CsDiff < 8) {
if (Rs < 2459) {
return 8;
}
else if (Rs >= 2459) {
return 4;
}
}
else if (CsDiff >= 8) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (diffRsCsdiff < -36) {
if (Rs < 2500) {
return 1;
}
else if (Rs >= 2500) {
return 4;
}
}
else if (diffRsCsdiff >= -36) {
if (mu_mv_mag_sq < 74) {
if (Rs < 2096) {
return 1;
}
else if (Rs >= 2096) {
return 8;
}
}
else if (mu_mv_mag_sq >= 74) {
if (Rs < 2823) {
return 4;
}
else if (Rs >= 2823) {
return 8;
}
}
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
else if (AFD >= 20) {
if (RsDiff < 17) {
if (mcTcor < 82) {
if (refDCval < 88) {
if (mu_mv_mag_sq < 19) {
if (Rs < 2451) {
if (Rs < 2164) {
return 1;
}
else if (Rs >= 2164) {
if (Rs < 2331) {
return 8;
}
else if (Rs >= 2331) {
return 1;
}
}
}
else if (Rs >= 2451) {
return 4;
}
}
else if (mu_mv_mag_sq >= 19) {
return 8;
}
}
else if (refDCval >= 88) {
if (Cs < 1514) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 4;
}
}
else if (Cs >= 1514) {
return 8;
}
}
}
else if (mcTcor >= 82) {
if (mcTcor < 94) {
if (ssDCval < 89) {
if (ssDCval < 76) {
if (mu_mv_mag_sq < 7) {
if (diffAFD < 3) {
if (mcTcor < 91) {
if (SC < 2997) {
return 4;
}
else if (SC >= 2997) {
return 2;
}
}
else if (mcTcor >= 91) {
return 4;
}
}
else if (diffAFD >= 3) {
if (Rs < 1800) {
return 1;
}
else if (Rs >= 1800) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (tcor < 63) {
if (mu_mv_mag_sq < 11) {
if (Rs < 1821) {
return 8;
}
else if (Rs >= 1821) {
return 1;
}
}
else if (mu_mv_mag_sq >= 11) {
return 4;
}
}
else if (tcor >= 63) {
if (diffRsCsdiff < 54) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
if (Rs < 2789) {
return 2;
}
else if (Rs >= 2789) {
return 8;
}
}
}
else if (diffRsCsdiff >= 54) {
return 4;
}
}
}
}
else if (ssDCval >= 76) {
if (mu_mv_mag_sq < 17) {
if (TSC < 13) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
if (RsCsDiff < 201) {
if (posBalance < 57) {
if (Rs < 1882) {
return 8;
}
else if (Rs >= 1882) {
return 2;
}
}
else if (posBalance >= 57) {
return 4;
}
}
else if (RsCsDiff >= 201) {
if (TSC < 12) {
if (Rs < 2915) {
return 8;
}
else if (Rs >= 2915) {
return 2;
}
}
else if (TSC >= 12) {
return 4;
}
}
}
}
else if (TSC >= 13) {
if (Rs < 2449) {
if (Rs < 1908) {
if (Cs < 1578) {
if (Rs < 1796) {
return 1;
}
else if (Rs >= 1796) {
return 8;
}
}
else if (Cs >= 1578) {
return 4;
}
}
else if (Rs >= 1908) {
return 8;
}
}
else if (Rs >= 2449) {
if (Cs < 1617) {
return 2;
}
else if (Cs >= 1617) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (Cs < 1838) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (Cs >= 1838) {
return 8;
}
}
}
}
else if (ssDCval >= 89) {
if (ssDCval < 98) {
if (ssDCval < 92) {
if (RsDiff < 10) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
else if (RsDiff >= 10) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 1;
}
}
}
else if (ssDCval >= 92) {
return 8;
}
}
else if (ssDCval >= 98) {
if (posBalance < 63) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (Cs < 1594) {
if (Rs < 2050) {
return 8;
}
else if (Rs >= 2050) {
return 4;
}
}
else if (Cs >= 1594) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 4;
}
}
}
}
}
else if (mcTcor >= 94) {
if (gchDC < 4) {
if (AFD < 21) {
if (Rs < 2285) {
return 1;
}
else if (Rs >= 2285) {
return 4;
}
}
else if (AFD >= 21) {
if (Rs < 1112) {
if (Rs < 1037) {
return 4;
}
else if (Rs >= 1037) {
return 1;
}
}
else if (Rs >= 1112) {
if (Cs < 1870) {
return 2;
}
else if (Cs >= 1870) {
return 8;
}
}
}
}
else if (gchDC >= 4) {
return 4;
}
}
}
}
else if (RsDiff >= 17) {
if (diffMVdiffVal < -24) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
return 1;
}
}
else if (diffMVdiffVal >= -24) {
if (CsDiff < 11) {
if (MVDiff < 63) {
return 2;
}
else if (MVDiff >= 63) {
return 8;
}
}
else if (CsDiff >= 11) {
if (posBalance < 60) {
if (Cs < 1418) {
return 1;
}
else if (Cs >= 1418) {
return 8;
}
}
else if (posBalance >= 60) {
if (Rs < 3163) {
return 2;
}
else if (Rs >= 3163) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 111) {
if (AFD < 17) {
if (Rs < 3044) {
if (negBalance < 40) {
if (RsDiff < 10) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (RsDiff >= 10) {
if (Rs < 2417) {
return 2;
}
else if (Rs >= 2417) {
return 4;
}
}
}
else if (negBalance >= 40) {
return 8;
}
}
else if (Rs >= 3044) {
if (tcor < 77) {
if (Rs < 3276) {
return 4;
}
else if (Rs >= 3276) {
return 1;
}
}
else if (tcor >= 77) {
return 8;
}
}
}
else if (AFD >= 17) {
if (mu_mv_mag_sq < 18) {
if (AFD < 24) {
if (mcTcor < 89) {
if (Rs < 2066) {
if (negBalance < 65) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (negBalance >= 65) {
return 4;
}
}
else if (Rs >= 2066) {
if (ssDCval < 116) {
return 8;
}
else if (ssDCval >= 116) {
return 1;
}
}
}
else if (mcTcor >= 89) {
if (diffRsCsdiff < -114) {
return 1;
}
else if (diffRsCsdiff >= -114) {
if (TSC < 10) {
if (AFD < 19) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
return 8;
}
}
else if (AFD >= 19) {
if (Rs < 1631) {
if (Rs < 1269) {
return 8;
}
else if (Rs >= 1269) {
return 4;
}
}
else if (Rs >= 1631) {
return 1;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
else if (AFD >= 24) {
if (diffRsCsdiff < 202) {
if (mcTcor < 87) {
if (Cs < 1525) {
return 4;
}
else if (Cs >= 1525) {
return 8;
}
}
else if (mcTcor >= 87) {
if (RsDiff < 17) {
if (Rs < 1249) {
return 2;
}
else if (Rs >= 1249) {
return 8;
}
}
else if (RsDiff >= 17) {
return 2;
}
}
}
else if (diffRsCsdiff >= 202) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (refDCval < 153) {
if (RsDiff < 22) {
if (refDCval < 133) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (Rs < 1675) {
if (Rs < 1605) {
return 8;
}
else if (Rs >= 1605) {
return 4;
}
}
else if (Rs >= 1675) {
return 8;
}
}
}
else if (refDCval >= 133) {
if (negBalance < 55) {
if (MVDiff < 30) {
if (diffMVdiffVal < -4) {
return 2;
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (MVDiff >= 30) {
return 8;
}
}
else if (negBalance >= 55) {
return 2;
}
}
}
else if (RsDiff >= 22) {
if (Rs < 3150) {
return 1;
}
else if (Rs >= 3150) {
return 2;
}
}
}
else if (refDCval >= 153) {
if (SC < 4732) {
return 2;
}
else if (SC >= 4732) {
return 4;
}
}
}
}
}
}
else if (Rs >= 3341) {
if (tcor < 86) {
if (RsDiff < 13) {
if (negBalance < 47) {
if (refDCval < 87) {
if (posBalance < 47) {
if (mcTcor < 91) {
if (Rs < 3785) {
return 2;
}
else if (Rs >= 3785) {
return 4;
}
}
else if (mcTcor >= 91) {
return 4;
}
}
else if (posBalance >= 47) {
return 2;
}
}
else if (refDCval >= 87) {
if (ssDCval < 137) {
return 8;
}
else if (ssDCval >= 137) {
return 4;
}
}
}
else if (negBalance >= 47) {
if (diffRsCsdiff < -158) {
return 1;
}
else if (diffRsCsdiff >= -158) {
if (mu_mv_mag_sq < 78) {
if (diffMVdiffVal < -13) {
if (Rs < 3543) {
return 8;
}
else if (Rs >= 3543) {
return 1;
}
}
else if (diffMVdiffVal >= -13) {
return 8;
}
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 3609) {
return 8;
}
else if (Rs >= 3609) {
return 4;
}
}
}
}
}
else if (RsDiff >= 13) {
if (mu_mv_mag_sq < 10) {
if (ssDCval < 108) {
if (RsDiff < 16) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
if (AFD < 18) {
if (Cs < 1784) {
return 8;
}
else if (Cs >= 1784) {
return 2;
}
}
else if (AFD >= 18) {
return 4;
}
}
}
else if (RsDiff >= 16) {
if (mcTcor < 95) {
return 2;
}
else if (mcTcor >= 95) {
if (Rs < 3449) {
return 8;
}
else if (Rs >= 3449) {
return 2;
}
}
}
}
else if (ssDCval >= 108) {
if (TSC < 14) {
if (RsDiff < 26) {
if (MVDiff < 27) {
if (AFD < 21) {
if (Rs < 4461) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
else if (Rs >= 4461) {
if (Rs < 7303) {
if (Cs < 1421) {
return 8;
}
else if (Cs >= 1421) {
return 2;
}
}
else if (Rs >= 7303) {
return 8;
}
}
}
else if (AFD >= 21) {
if (Rs < 5460) {
return 1;
}
else if (Rs >= 5460) {
return 8;
}
}
}
else if (MVDiff >= 27) {
if (Rs < 5106) {
if (RsDiff < 21) {
return 8;
}
else if (RsDiff >= 21) {
return 2;
}
}
else if (Rs >= 5106) {
return 8;
}
}
}
else if (RsDiff >= 26) {
return 2;
}
}
else if (TSC >= 14) {
if (tcor < 55) {
return 4;
}
else if (tcor >= 55) {
if (Cs < 1549) {
return 1;
}
else if (Cs >= 1549) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (refDCval < 140) {
if (RsDiff < 17) {
if (MVDiff < 10) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
if (Rs < 3674) {
return 2;
}
else if (Rs >= 3674) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (RsDiff < 15) {
if (diffAFD < -1) {
if (Rs < 3877) {
return 2;
}
else if (Rs >= 3877) {
return 8;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (RsDiff >= 15) {
if (Cs < 1650) {
if (Rs < 3620) {
return 1;
}
else if (Rs >= 3620) {
return 2;
}
}
else if (Cs >= 1650) {
return 8;
}
}
}
}
else if (RsDiff >= 17) {
if (Cs < 1749) {
if (MVDiff < 46) {
if (ssDCval < 142) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
if (ssDCval < 106) {
if (refDCval < 100) {
return 2;
}
else if (refDCval >= 100) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
return 2;
}
}
}
else if (ssDCval >= 106) {
return 2;
}
}
}
else if (ssDCval >= 142) {
return 1;
}
}
else if (MVDiff >= 46) {
if (CsDiff < 7) {
if (Rs < 3531) {
return 2;
}
else if (Rs >= 3531) {
return 4;
}
}
else if (CsDiff >= 7) {
if (RsDiff < 29) {
return 2;
}
else if (RsDiff >= 29) {
return 1;
}
}
}
}
else if (Cs >= 1749) {
if (negBalance < 58) {
if (Rs < 3472) {
return 8;
}
else if (Rs >= 3472) {
if (AFD < 37) {
return 4;
}
else if (AFD >= 37) {
return 2;
}
}
}
else if (negBalance >= 58) {
if (Cs < 1890) {
return 2;
}
else if (Cs >= 1890) {
if (Rs < 3399) {
return 8;
}
else if (Rs >= 3399) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 140) {
if (gchDC < 3) {
if (tcor < 74) {
return 4;
}
else if (tcor >= 74) {
if (Cs < 1801) {
return 2;
}
else if (Cs >= 1801) {
if (Cs < 1914) {
return 4;
}
else if (Cs >= 1914) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
if (Rs < 4938) {
return 1;
}
else if (Rs >= 4938) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 86) {
if (CsDiff < 3) {
if (Rs < 4842) {
return 2;
}
else if (Rs >= 4842) {
return 4;
}
}
else if (CsDiff >= 3) {
if (refDCval < 101) {
if (posBalance < 44) {
if (negBalance < 40) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
if (diffRsCsdiff < -46) {
if (Cs < 1598) {
return 4;
}
else if (Cs >= 1598) {
return 2;
}
}
else if (diffRsCsdiff >= -46) {
return 2;
}
}
}
}
else if (negBalance >= 40) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 1;
}
}
}
else if (posBalance >= 44) {
if (mcTcor < 92) {
if (ssDCval < 96) {
if (CsDiff < 9) {
return 2;
}
else if (CsDiff >= 9) {
return 8;
}
}
else if (ssDCval >= 96) {
if (Rs < 4215) {
return 8;
}
else if (Rs >= 4215) {
return 2;
}
}
}
else if (mcTcor >= 92) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
return 8;
}
}
}
}
else if (refDCval >= 101) {
if (RsDiff < 11) {
if (posBalance < 37) {
return 2;
}
else if (posBalance >= 37) {
if (Rs < 4993) {
return 8;
}
else if (Rs >= 4993) {
return 1;
}
}
}
else if (RsDiff >= 11) {
if (MVDiff < 24) {
if (posBalance < 59) {
if (negBalance < 64) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
if (Rs < 4136) {
return 2;
}
else if (Rs >= 4136) {
return 8;
}
}
}
else if (negBalance >= 64) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
else if (posBalance >= 59) {
if (MVDiff < 10) {
if (Rs < 4278) {
return 4;
}
else if (Rs >= 4278) {
return 1;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
else if (MVDiff >= 24) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
if (Rs < 3947) {
return 8;
}
else if (Rs >= 3947) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Cs >= 1926) {
if (TSC < 8) {
if (SC < 11189) {
if (ssDCval < 72) {
if (CsDiff < 8) {
if (MVDiff < 5) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 3993) {
return 1;
}
else if (Rs >= 3993) {
return 4;
}
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (CsDiff >= 8) {
if (Rs < 3891) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (diffRsCsdiff < -121) {
return 2;
}
else if (diffRsCsdiff >= -121) {
return 1;
}
}
}
else if (Rs >= 3891) {
return 8;
}
}
}
else if (ssDCval >= 72) {
if (ssDCval < 112) {
if (negBalance < 32) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
else if (negBalance >= 32) {
if (tcor < 91) {
if (Rs < 6110) {
if (RsCsDiff < 298) {
if (mu_mv_mag_sq < 2) {
if (Rs < 3312) {
return 4;
}
else if (Rs >= 3312) {
if (Rs < 3654) {
return 2;
}
else if (Rs >= 3654) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
else if (RsCsDiff >= 298) {
return 2;
}
}
else if (Rs >= 6110) {
return 2;
}
}
else if (tcor >= 91) {
if (Cs < 3137) {
return 2;
}
else if (Cs >= 3137) {
if (Cs < 3317) {
return 4;
}
else if (Cs >= 3317) {
if (Rs < 5811) {
return 8;
}
else if (Rs >= 5811) {
if (Rs < 6534) {
return 1;
}
else if (Rs >= 6534) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (Rs < 6059) {
if (refDCval < 132) {
if (mcTcor < 94) {
if (negBalance < 24) {
if (Cs < 2487) {
return 2;
}
else if (Cs >= 2487) {
return 4;
}
}
else if (negBalance >= 24) {
if (Rs < 2025) {
return 4;
}
else if (Rs >= 2025) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 5424) {
return 8;
}
else if (Rs >= 5424) {
if (Rs < 5931) {
return 4;
}
else if (Rs >= 5931) {
return 8;
}
}
}
}
else if (refDCval >= 132) {
if (diffTSC < -4) {
return 1;
}
else if (diffTSC >= -4) {
if (tcor < 70) {
return 2;
}
else if (tcor >= 70) {
if (gchDC < 7) {
return 8;
}
else if (gchDC >= 7) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 8;
}
}
}
}
}
}
else if (Rs >= 6059) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
if (Rs < 7148) {
return 8;
}
else if (Rs >= 7148) {
return 2;
}
}
}
}
}
}
else if (SC >= 11189) {
if (ssDCval < 129) {
if (Rs < 7845) {
return 8;
}
else if (Rs >= 7845) {
if (RsDiff < 41) {
return 1;
}
else if (RsDiff >= 41) {
return 8;
}
}
}
else if (ssDCval >= 129) {
return 8;
}
}
}
else if (TSC >= 8) {
if (refDCval < 156) {
if (mu_mv_mag_sq < 84) {
if (RsDiff < 16) {
if (AFD < 21) {
if (refDCval < 101) {
if (mcTcor < 75) {
if (Rs < 2091) {
if (mu_mv_mag_sq < 33) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
if (Rs < 2080) {
return 4;
}
else if (Rs >= 2080) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (tcor < 66) {
return 4;
}
else if (tcor >= 66) {
return 2;
}
}
}
else if (Rs >= 2091) {
if (AFD < 17) {
if (mu_mv_mag_sq < 23) {
return 8;
}
else if (mu_mv_mag_sq >= 23) {
if (diffMVdiffVal < 6) {
return 4;
}
else if (diffMVdiffVal >= 6) {
if (Rs < 2828) {
return 1;
}
else if (Rs >= 2828) {
return 8;
}
}
}
}
else if (AFD >= 17) {
if (Rs < 3633) {
if (Rs < 2849) {
return 4;
}
else if (Rs >= 2849) {
if (Cs < 3123) {
return 8;
}
else if (Cs >= 3123) {
return 4;
}
}
}
else if (Rs >= 3633) {
return 2;
}
}
}
}
else if (mcTcor >= 75) {
if (negBalance < 46) {
if (ssDCval < 73) {
if (mu_mv_mag_sq < 55) {
if (AFD < 17) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 8;
}
}
else if (AFD >= 17) {
return 2;
}
}
else if (mu_mv_mag_sq >= 55) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (Rs < 3740) {
return 8;
}
else if (Rs >= 3740) {
return 4;
}
}
}
}
else if (ssDCval >= 73) {
if (posBalance < 57) {
if (TSC < 10) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (MVDiff < 6) {
if (CsDiff < 10) {
return 2;
}
else if (CsDiff >= 10) {
if (Rs < 3087) {
return 8;
}
else if (Rs >= 3087) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 77) {
if (Rs < 4719) {
return 4;
}
else if (Rs >= 4719) {
return 8;
}
}
else if (ssDCval >= 77) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (posBalance < 53) {
if (MVDiff < 19) {
if (Cs < 2011) {
return 2;
}
else if (Cs >= 2011) {
return 1;
}
}
else if (MVDiff >= 19) {
return 2;
}
}
else if (posBalance >= 53) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (ssDCval < 89) {
if (Rs < 4109) {
if (diffRsCsdiff < -86) {
return 8;
}
else if (diffRsCsdiff >= -86) {
if (gchDC < 5) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (gchDC >= 5) {
if (Cs < 2322) {
return 8;
}
else if (Cs >= 2322) {
return 4;
}
}
}
}
else if (Rs >= 4109) {
return 2;
}
}
else if (ssDCval >= 89) {
if (MVDiff < 12) {
if (Cs < 2638) {
return 2;
}
else if (Cs >= 2638) {
return 4;
}
}
else if (MVDiff >= 12) {
if (ssDCval < 104) {
if (Rs < 4730) {
if (Rs < 3264) {
if (Rs < 2845) {
return 8;
}
else if (Rs >= 2845) {
return 2;
}
}
else if (Rs >= 3264) {
return 8;
}
}
else if (Rs >= 4730) {
return 2;
}
}
else if (ssDCval >= 104) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (ssDCval < 81) {
if (tcor < 64) {
if (AFD < 20) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
else if (AFD >= 20) {
return 1;
}
}
else if (tcor >= 64) {
if (Cs < 3139) {
if (Rs < 2846) {
if (AFD < 18) {
if (gchDC < 3) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (gchDC >= 3) {
return 1;
}
}
else if (AFD >= 18) {
if (ssDCval < 73) {
if (MVDiff < 22) {
return 1;
}
else if (MVDiff >= 22) {
if (Rs < 2633) {
return 8;
}
else if (Rs >= 2633) {
return 2;
}
}
}
else if (ssDCval >= 73) {
if (posBalance < 64) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 1;
}
}
else if (posBalance >= 64) {
if (Rs < 2682) {
return 2;
}
else if (Rs >= 2682) {
return 1;
}
}
}
}
}
else if (Rs >= 2846) {
if (SC < 6154) {
if (SC < 5342) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
return 8;
}
}
else if (SC >= 5342) {
if (TSC < 10) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
return 8;
}
}
else if (TSC >= 10) {
if (Cs < 2470) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (Cs < 2074) {
if (Rs < 3745) {
return 8;
}
else if (Rs >= 3745) {
return 2;
}
}
else if (Cs >= 2074) {
if (Rs < 3705) {
return 4;
}
else if (Rs >= 3705) {
return 8;
}
}
}
}
else if (Cs >= 2470) {
return 8;
}
}
}
}
else if (SC >= 6154) {
if (posBalance < 57) {
if (ssDCval < 78) {
if (TSC < 10) {
if (Cs < 2612) {
return 4;
}
else if (Cs >= 2612) {
return 8;
}
}
else if (TSC >= 10) {
if (Rs < 3743) {
return 4;
}
else if (Rs >= 3743) {
return 8;
}
}
}
else if (ssDCval >= 78) {
return 1;
}
}
else if (posBalance >= 57) {
if (tcor < 80) {
if (mu_mv_mag_sq < 13) {
return 1;
}
else if (mu_mv_mag_sq >= 13) {
return 8;
}
}
else if (tcor >= 80) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (Rs < 3178) {
return 8;
}
else if (Rs >= 3178) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Rs < 3388) {
return 2;
}
else if (Rs >= 3388) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Cs >= 3139) {
if (TSC < 11) {
if (Rs < 4109) {
return 4;
}
else if (Rs >= 4109) {
return 8;
}
}
else if (TSC >= 11) {
return 8;
}
}
}
}
else if (ssDCval >= 81) {
if (SC < 8191) {
if (Cs < 2530) {
if (RsDiff < 11) {
if (mcTcor < 94) {
if (AFD < 15) {
if (ssDCval < 89) {
if (diffMVdiffVal < 2) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
else if (diffMVdiffVal >= 2) {
if (Cs < 2226) {
return 4;
}
else if (Cs >= 2226) {
return 2;
}
}
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (AFD >= 15) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (mu_mv_mag_sq < 18) {
if (Rs < 2288) {
return 8;
}
else if (Rs >= 2288) {
if (Rs < 3114) {
return 2;
}
else if (Rs >= 3114) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 18) {
return 8;
}
}
}
else if (RsDiff >= 11) {
if (mu_mv_mag_sq < 2) {
if (diffAFD < 3) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 2;
}
}
}
else if (diffAFD >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 2030) {
if (Rs < 2942) {
if (SC < 4333) {
return 4;
}
else if (SC >= 4333) {
return 8;
}
}
else if (Rs >= 2942) {
return 4;
}
}
else if (Cs >= 2030) {
if (Cs < 2364) {
if (mcTcor < 87) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 2;
}
}
else if (mcTcor >= 87) {
return 8;
}
}
else if (Cs >= 2364) {
if (Rs < 3394) {
return 2;
}
else if (Rs >= 3394) {
if (Rs < 4013) {
return 8;
}
else if (Rs >= 4013) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 2530) {
if (Cs < 3347) {
if (RsCsDiff < 255) {
if (SC < 6917) {
if (posBalance < 60) {
if (negBalance < 58) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (Rs < 2732) {
return 4;
}
else if (Rs >= 2732) {
return 8;
}
}
}
else if (negBalance >= 58) {
return 8;
}
}
else if (posBalance >= 60) {
if (mcTcor < 83) {
return 4;
}
else if (mcTcor >= 83) {
if (negBalance < 55) {
if (Rs < 2782) {
return 8;
}
else if (Rs >= 2782) {
if (MVDiff < 22) {
if (Rs < 3421) {
if (Rs < 3101) {
return 4;
}
else if (Rs >= 3101) {
return 2;
}
}
else if (Rs >= 3421) {
return 4;
}
}
else if (MVDiff >= 22) {
return 8;
}
}
}
else if (negBalance >= 55) {
return 8;
}
}
}
}
else if (SC >= 6917) {
if (diffRsCsdiff < -20) {
return 4;
}
else if (diffRsCsdiff >= -20) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
return 4;
}
}
}
}
else if (RsCsDiff >= 255) {
if (Rs < 2707) {
return 8;
}
else if (Rs >= 2707) {
return 4;
}
}
}
else if (Cs >= 3347) {
if (mcTcor < 90) {
if (refDCval < 95) {
if (Cs < 3723) {
return 4;
}
else if (Cs >= 3723) {
return 8;
}
}
else if (refDCval >= 95) {
return 4;
}
}
else if (mcTcor >= 90) {
if (Rs < 3248) {
return 8;
}
else if (Rs >= 3248) {
return 2;
}
}
}
}
}
else if (SC >= 8191) {
return 8;
}
}
}
}
}
else if (refDCval >= 101) {
if (mcTcor < 85) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
if (Cs < 2904) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 1;
}
}
else if (Cs >= 2904) {
if (negBalance < 50) {
if (negBalance < 45) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
if (Cs < 3967) {
return 8;
}
else if (Cs >= 3967) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (diffRsCsdiff < -39) {
if (Rs < 4298) {
return 8;
}
else if (Rs >= 4298) {
return 4;
}
}
else if (diffRsCsdiff >= -39) {
return 1;
}
}
}
else if (negBalance >= 50) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
}
}
}
else if (mcTcor >= 85) {
if (Cs < 2618) {
if (mcTcor < 94) {
if (SC < 6062) {
if (Rs < 2766) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (posBalance < 53) {
if (ssDCval < 129) {
if (Cs < 2081) {
return 1;
}
else if (Cs >= 2081) {
return 8;
}
}
else if (ssDCval >= 129) {
return 4;
}
}
else if (posBalance >= 53) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 1;
}
}
}
}
else if (Rs >= 2766) {
if (diffAFD < 10) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
if (TSC < 12) {
if (Rs < 3261) {
if (ssDCval < 131) {
if (Rs < 2983) {
if (MVDiff < 7) {
if (Cs < 2039) {
return 8;
}
else if (Cs >= 2039) {
return 4;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (Rs >= 2983) {
if (Cs < 1991) {
return 8;
}
else if (Cs >= 1991) {
return 4;
}
}
}
else if (ssDCval >= 131) {
return 8;
}
}
else if (Rs >= 3261) {
if (Rs < 4048) {
return 8;
}
else if (Rs >= 4048) {
return 4;
}
}
}
else if (TSC >= 12) {
if (Cs < 2474) {
return 8;
}
else if (Cs >= 2474) {
if (Rs < 2956) {
return 8;
}
else if (Rs >= 2956) {
return 2;
}
}
}
}
}
else if (diffAFD >= 10) {
if (Rs < 3245) {
return 1;
}
else if (Rs >= 3245) {
return 8;
}
}
}
}
else if (SC >= 6062) {
if (mu_mv_mag_sq < 74) {
return 8;
}
else if (mu_mv_mag_sq >= 74) {
if (Rs < 4774) {
return 1;
}
else if (Rs >= 4774) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (MVDiff < 8) {
if (RsCsDiff < 189) {
if (Rs < 2427) {
return 2;
}
else if (Rs >= 2427) {
return 8;
}
}
else if (RsCsDiff >= 189) {
if (Rs < 2899) {
if (Rs < 2732) {
return 2;
}
else if (Rs >= 2732) {
return 8;
}
}
else if (Rs >= 2899) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 4;
}
}
}
}
else if (MVDiff >= 8) {
if (tcor < 87) {
if (RsCsDiff < 236) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
else if (RsCsDiff >= 236) {
return 1;
}
}
else if (tcor >= 87) {
return 8;
}
}
}
}
else if (Cs >= 2618) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
if (negBalance < 56) {
if (MVDiff < 9) {
if (mcTcor < 95) {
if (AFD < 14) {
if (negBalance < 50) {
if (diffAFD < -2) {
if (Rs < 5166) {
return 4;
}
else if (Rs >= 5166) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (negBalance >= 50) {
return 4;
}
}
else if (AFD >= 14) {
if (ssDCval < 117) {
if (refDCval < 105) {
return 8;
}
else if (refDCval >= 105) {
if (ssDCval < 112) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (Rs < 2712) {
return 8;
}
else if (Rs >= 2712) {
return 4;
}
}
}
else if (ssDCval >= 112) {
if (Rs < 5975) {
return 8;
}
else if (Rs >= 5975) {
return 4;
}
}
}
}
else if (ssDCval >= 117) {
return 4;
}
}
}
else if (mcTcor >= 95) {
if (Rs < 4690) {
return 2;
}
else if (Rs >= 4690) {
return 4;
}
}
}
else if (MVDiff >= 9) {
if (CsDiff < 14) {
if (MVDiff < 11) {
if (SC < 5705) {
return 2;
}
else if (SC >= 5705) {
if (diffMVdiffVal < 4) {
if (Rs < 2858) {
return 4;
}
else if (Rs >= 2858) {
return 8;
}
}
else if (diffMVdiffVal >= 4) {
if (Cs < 2853) {
return 8;
}
else if (Cs >= 2853) {
return 4;
}
}
}
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (CsDiff >= 14) {
if (ssDCval < 123) {
return 2;
}
else if (ssDCval >= 123) {
if (Rs < 5170) {
return 4;
}
else if (Rs >= 5170) {
return 8;
}
}
}
}
}
else if (negBalance >= 56) {
return 4;
}
}
}
}
}
}
else if (AFD >= 21) {
if (refDCval < 100) {
if (ssDCval < 91) {
if (Cs < 3031) {
if (Rs < 2075) {
if (Rs < 1609) {
if (negBalance < 58) {
return 2;
}
else if (negBalance >= 58) {
if (Rs < 1296) {
if (Rs < 1257) {
return 1;
}
else if (Rs >= 1257) {
return 4;
}
}
else if (Rs >= 1296) {
return 8;
}
}
}
else if (Rs >= 1609) {
if (CsDiff < 15) {
if (CsDiff < 10) {
if (Rs < 1877) {
return 4;
}
else if (Rs >= 1877) {
if (Rs < 2039) {
return 8;
}
else if (Rs >= 2039) {
return 1;
}
}
}
else if (CsDiff >= 10) {
return 4;
}
}
else if (CsDiff >= 15) {
return 1;
}
}
}
else if (Rs >= 2075) {
if (RsDiff < 14) {
if (diffAFD < 2) {
if (mcTcor < 90) {
if (negBalance < 53) {
if (Cs < 2477) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 4;
}
}
else if (Cs >= 2477) {
return 1;
}
}
else if (negBalance >= 53) {
return 1;
}
}
else if (mcTcor >= 90) {
if (ssDCval < 70) {
if (Cs < 2418) {
return 4;
}
else if (Cs >= 2418) {
return 1;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
else if (diffAFD >= 2) {
return 8;
}
}
else if (RsDiff >= 14) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
if (Rs < 3345) {
return 8;
}
else if (Rs >= 3345) {
if (refDCval < 79) {
return 8;
}
else if (refDCval >= 79) {
return 4;
}
}
}
}
}
}
else if (Cs >= 3031) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (ssDCval >= 91) {
if (mu_mv_mag_sq < 3) {
if (Rs < 4706) {
if (Cs < 1938) {
return 8;
}
else if (Cs >= 1938) {
return 4;
}
}
else if (Rs >= 4706) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 56) {
if (Rs < 2162) {
if (Rs < 1776) {
return 8;
}
else if (Rs >= 1776) {
return 4;
}
}
else if (Rs >= 2162) {
return 8;
}
}
else if (negBalance >= 56) {
if (Rs < 2124) {
return 8;
}
else if (Rs >= 2124) {
if (tcor < 72) {
return 4;
}
else if (tcor >= 72) {
if (Rs < 2867) {
return 1;
}
else if (Rs >= 2867) {
if (Rs < 4380) {
return 2;
}
else if (Rs >= 4380) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 100) {
if (Rs < 3087) {
if (diffRsCsdiff < 60) {
if (SC < 5140) {
if (MVDiff < 23) {
if (Rs < 2934) {
if (mu_mv_mag_sq < 8) {
if (Rs < 2786) {
if (Rs < 1632) {
return 8;
}
else if (Rs >= 1632) {
return 2;
}
}
else if (Rs >= 2786) {
return 1;
}
}
else if (mu_mv_mag_sq >= 8) {
if (negBalance < 61) {
if (Rs < 2760) {
return 8;
}
else if (Rs >= 2760) {
return 2;
}
}
else if (negBalance >= 61) {
return 1;
}
}
}
else if (Rs >= 2934) {
return 8;
}
}
else if (MVDiff >= 23) {
if (Rs < 2842) {
if (RsCsDiff < 235) {
return 1;
}
else if (RsCsDiff >= 235) {
return 8;
}
}
else if (Rs >= 2842) {
return 2;
}
}
}
else if (SC >= 5140) {
if (Rs < 2212) {
return 4;
}
else if (Rs >= 2212) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 60) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (Rs >= 3087) {
if (CsDiff < 17) {
if (negBalance < 55) {
if (MVDiff < 9) {
if (Cs < 2367) {
return 8;
}
else if (Cs >= 2367) {
return 4;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (negBalance >= 55) {
if (posBalance < 58) {
if (CsDiff < 13) {
return 8;
}
else if (CsDiff >= 13) {
return 4;
}
}
else if (posBalance >= 58) {
if (Rs < 3231) {
if (Cs < 3059) {
return 4;
}
else if (Cs >= 3059) {
return 1;
}
}
else if (Rs >= 3231) {
if (ssDCval < 137) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (ssDCval >= 137) {
return 8;
}
}
}
}
}
else if (CsDiff >= 17) {
return 2;
}
}
}
}
}
else if (RsDiff >= 16) {
if (SC < 10979) {
if (mcTcor < 91) {
if (TSC < 15) {
if (ssDCval < 138) {
if (ssDCval < 66) {
if (AFD < 21) {
if (ssDCval < 49) {
return 2;
}
else if (ssDCval >= 49) {
if (mu_mv_mag_sq < 53) {
return 4;
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 4124) {
return 8;
}
else if (Rs >= 4124) {
return 4;
}
}
}
}
else if (AFD >= 21) {
if (diffAFD < 3) {
if (diffRsCsdiff < -71) {
return 2;
}
else if (diffRsCsdiff >= -71) {
return 4;
}
}
else if (diffAFD >= 3) {
return 2;
}
}
}
else if (ssDCval >= 66) {
if (mcTcor < 83) {
if (ssDCval < 83) {
if (Rs < 3675) {
return 8;
}
else if (Rs >= 3675) {
if (Rs < 4322) {
return 1;
}
else if (Rs >= 4322) {
return 4;
}
}
}
else if (ssDCval >= 83) {
if (Rs < 2586) {
return 2;
}
else if (Rs >= 2586) {
return 8;
}
}
}
else if (mcTcor >= 83) {
if (refDCval < 74) {
if (Cs < 3892) {
if (posBalance < 49) {
if (Cs < 2241) {
return 2;
}
else if (Cs >= 2241) {
return 4;
}
}
else if (posBalance >= 49) {
if (Rs < 4543) {
if (tcor < 72) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
return 4;
}
}
else if (tcor >= 72) {
if (Cs < 3444) {
return 8;
}
else if (Cs >= 3444) {
return 2;
}
}
}
else if (Rs >= 4543) {
if (Rs < 5518) {
return 2;
}
else if (Rs >= 5518) {
return 8;
}
}
}
}
else if (Cs >= 3892) {
return 8;
}
}
else if (refDCval >= 74) {
if (posBalance < 65) {
if (tcor < 79) {
if (diffRsCsdiff < 120) {
if (TSC < 14) {
if (diffAFD < -2) {
if (negBalance < 58) {
return 2;
}
else if (negBalance >= 58) {
if (Rs < 5080) {
return 8;
}
else if (Rs >= 5080) {
return 4;
}
}
}
else if (diffAFD >= -2) {
if (ssDCval < 136) {
if (Rs < 3135) {
if (Rs < 3049) {
return 4;
}
else if (Rs >= 3049) {
return 2;
}
}
else if (Rs >= 3135) {
return 4;
}
}
else if (ssDCval >= 136) {
return 8;
}
}
}
else if (TSC >= 14) {
if (CsDiff < 17) {
return 8;
}
else if (CsDiff >= 17) {
return 4;
}
}
}
else if (diffRsCsdiff >= 120) {
if (negBalance < 54) {
if (Cs < 3180) {
return 8;
}
else if (Cs >= 3180) {
if (Rs < 5561) {
return 4;
}
else if (Rs >= 5561) {
if (Rs < 6096) {
return 8;
}
else if (Rs >= 6096) {
return 4;
}
}
}
}
else if (negBalance >= 54) {
if (Rs < 3352) {
return 1;
}
else if (Rs >= 3352) {
return 2;
}
}
}
}
else if (tcor >= 79) {
if (Rs < 4989) {
if (mu_mv_mag_sq < 43) {
return 8;
}
else if (mu_mv_mag_sq >= 43) {
if (Cs < 2212) {
return 8;
}
else if (Cs >= 2212) {
return 4;
}
}
}
else if (Rs >= 4989) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 1;
}
}
}
}
else if (posBalance >= 65) {
if (Rs < 3329) {
return 2;
}
else if (Rs >= 3329) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 138) {
return 4;
}
}
else if (TSC >= 15) {
if (RsCsDiff < 941) {
if (Rs < 5804) {
if (AFD < 31) {
if (posBalance < 52) {
if (tcor < 73) {
return 2;
}
else if (tcor >= 73) {
if (Rs < 5134) {
if (Cs < 3303) {
return 8;
}
else if (Cs >= 3303) {
return 2;
}
}
else if (Rs >= 5134) {
if (Rs < 5632) {
return 1;
}
else if (Rs >= 5632) {
return 4;
}
}
}
}
else if (posBalance >= 52) {
if (Cs < 2989) {
if (ssDCval < 80) {
if (Rs < 3927) {
return 1;
}
else if (Rs >= 3927) {
return 4;
}
}
else if (ssDCval >= 80) {
if (MVDiff < 27) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
return 1;
}
}
else if (MVDiff >= 27) {
return 8;
}
}
}
else if (Cs >= 2989) {
if (diffAFD < 4) {
if (tcor < 64) {
if (CsDiff < 17) {
if (Rs < 5634) {
return 2;
}
else if (Rs >= 5634) {
return 8;
}
}
else if (CsDiff >= 17) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 4;
}
}
}
else if (tcor >= 64) {
if (negBalance < 51) {
if (Rs < 5155) {
return 8;
}
else if (Rs >= 5155) {
return 2;
}
}
else if (negBalance >= 51) {
if (gchDC < 5) {
return 8;
}
else if (gchDC >= 5) {
return 2;
}
}
}
}
else if (diffAFD >= 4) {
if (Rs < 5303) {
return 4;
}
else if (Rs >= 5303) {
return 2;
}
}
}
}
}
else if (AFD >= 31) {
if (posBalance < 59) {
if (Rs < 3640) {
return 4;
}
else if (Rs >= 3640) {
return 2;
}
}
else if (posBalance >= 59) {
if (ssDCval < 109) {
return 1;
}
else if (ssDCval >= 109) {
return 8;
}
}
}
}
else if (Rs >= 5804) {
if (mcTcor < 74) {
return 2;
}
else if (mcTcor >= 74) {
if (posBalance < 56) {
if (diffMVdiffVal < -2) {
if (Rs < 6131) {
return 2;
}
else if (Rs >= 6131) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
else if (posBalance >= 56) {
if (Cs < 3873) {
return 4;
}
else if (Cs >= 3873) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 941) {
if (tcor < 22) {
return 1;
}
else if (tcor >= 22) {
if (RsDiff < 31) {
if (diffAFD < 8) {
if (AFD < 26) {
if (Rs < 5529) {
return 4;
}
else if (Rs >= 5529) {
return 8;
}
}
else if (AFD >= 26) {
return 8;
}
}
else if (diffAFD >= 8) {
if (Rs < 3928) {
return 8;
}
else if (Rs >= 3928) {
return 4;
}
}
}
else if (RsDiff >= 31) {
if (AFD < 48) {
return 4;
}
else if (AFD >= 48) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (negBalance < 52) {
if (gchDC < 3) {
if (refDCval < 77) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
if (Rs < 4797) {
return 2;
}
else if (Rs >= 4797) {
return 4;
}
}
}
else if (refDCval >= 77) {
if (Cs < 1987) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
return 8;
}
}
else if (Cs >= 1987) {
if (mu_mv_mag_sq < 62) {
if (diffRsCsdiff < -404) {
return 8;
}
else if (diffRsCsdiff >= -404) {
if (TSC < 9) {
if (Rs < 5318) {
return 8;
}
else if (Rs >= 5318) {
return 2;
}
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 62) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 4;
}
}
}
}
}
else if (gchDC >= 3) {
if (CsDiff < 15) {
return 8;
}
else if (CsDiff >= 15) {
if (RsDiff < 17) {
if (Rs < 3529) {
return 2;
}
else if (Rs >= 3529) {
return 8;
}
}
else if (RsDiff >= 17) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (RsCsDiff < 859) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
if (mu_mv_mag_sq < 57) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (Cs < 2033) {
return 1;
}
else if (Cs >= 2033) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (Rs < 4611) {
return 4;
}
else if (Rs >= 4611) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 859) {
if (SC < 9510) {
return 2;
}
else if (SC >= 9510) {
if (Rs < 5227) {
return 1;
}
else if (Rs >= 5227) {
return 4;
}
}
}
}
}
}
else if (SC >= 10979) {
if (mu_mv_mag_sq < 37) {
if (AFD < 41) {
if (RsCsDiff < 621) {
return 8;
}
else if (RsCsDiff >= 621) {
if (Rs < 9676) {
return 4;
}
else if (Rs >= 9676) {
return 1;
}
}
}
else if (AFD >= 41) {
return 8;
}
}
else if (mu_mv_mag_sq >= 37) {
if (AFD < 22) {
if (TSC < 12) {
if (Rs < 6104) {
return 8;
}
else if (Rs >= 6104) {
return 1;
}
}
else if (TSC >= 12) {
return 4;
}
}
else if (AFD >= 22) {
if (Cs < 4520) {
if (Rs < 7327) {
return 2;
}
else if (Rs >= 7327) {
return 4;
}
}
else if (Cs >= 4520) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (Cs < 3192) {
if (ssDCval < 137) {
if (Cs < 2707) {
if (Rs < 3596) {
if (SC < 5965) {
if (negBalance < 54) {
if (Rs < 3311) {
return 8;
}
else if (Rs >= 3311) {
if (Rs < 3329) {
if (Rs < 3316) {
return 1;
}
else if (Rs >= 3316) {
return 4;
}
}
else if (Rs >= 3329) {
return 8;
}
}
}
else if (negBalance >= 54) {
if (Rs < 2654) {
return 4;
}
else if (Rs >= 2654) {
return 8;
}
}
}
else if (SC >= 5965) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (Rs >= 3596) {
if (TSC < 10) {
if (Cs < 2083) {
if (ssDCval < 79) {
if (Rs < 3935) {
return 4;
}
else if (Rs >= 3935) {
return 1;
}
}
else if (ssDCval >= 79) {
return 2;
}
}
else if (Cs >= 2083) {
if (Rs < 3694) {
return 2;
}
else if (Rs >= 3694) {
if (Rs < 5016) {
return 8;
}
else if (Rs >= 5016) {
return 2;
}
}
}
}
else if (TSC >= 10) {
if (Rs < 3788) {
if (Rs < 3648) {
return 1;
}
else if (Rs >= 3648) {
return 8;
}
}
else if (Rs >= 3788) {
if (Rs < 5593) {
return 1;
}
else if (Rs >= 5593) {
return 4;
}
}
}
}
}
else if (Cs >= 2707) {
if (ssDCval < 107) {
if (refDCval < 80) {
if (negBalance < 44) {
return 4;
}
else if (negBalance >= 44) {
return 8;
}
}
else if (refDCval >= 80) {
if (Rs < 4005) {
if (ssDCval < 81) {
return 1;
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (Rs >= 4005) {
if (Rs < 4276) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
else if (Rs >= 4276) {
return 1;
}
}
}
}
else if (ssDCval >= 107) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Rs < 3101) {
return 8;
}
else if (Rs >= 3101) {
return 4;
}
}
}
}
}
else if (ssDCval >= 137) {
if (Rs < 6023) {
return 1;
}
else if (Rs >= 6023) {
return 2;
}
}
}
else if (Cs >= 3192) {
if (MVDiff < 15) {
if (Rs < 6033) {
if (RsCsDiff < 201) {
if (Rs < 4948) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (AFD < 16) {
if (MVDiff < 3) {
if (ssDCval < 90) {
if (Rs < 3232) {
return 4;
}
else if (Rs >= 3232) {
return 8;
}
}
else if (ssDCval >= 90) {
return 4;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (AFD >= 16) {
return 4;
}
}
}
else if (Rs >= 4948) {
if (Cs < 3820) {
return 1;
}
else if (Cs >= 3820) {
return 8;
}
}
}
else if (RsCsDiff >= 201) {
if (Rs < 3187) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (Rs >= 3187) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
if (diffAFD < -7) {
return 4;
}
else if (diffAFD >= -7) {
if (Cs < 3335) {
if (Cs < 3332) {
return 8;
}
else if (Cs >= 3332) {
return 4;
}
}
else if (Cs >= 3335) {
return 8;
}
}
}
}
}
}
else if (Rs >= 6033) {
if (AFD < 21) {
if (TSC < 13) {
if (Cs < 4255) {
return 4;
}
else if (Cs >= 4255) {
return 8;
}
}
else if (TSC >= 13) {
if (Rs < 6231) {
return 2;
}
else if (Rs >= 6231) {
return 4;
}
}
}
else if (AFD >= 21) {
return 8;
}
}
}
else if (MVDiff >= 15) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
if (AFD < 21) {
if (Rs < 4677) {
return 1;
}
else if (Rs >= 4677) {
return 2;
}
}
else if (AFD >= 21) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 156) {
if (refDCval < 166) {
if (negBalance < 43) {
if (Cs < 2470) {
if (Rs < 4479) {
return 4;
}
else if (Rs >= 4479) {
return 2;
}
}
else if (Cs >= 2470) {
return 8;
}
}
else if (negBalance >= 43) {
return 4;
}
}
else if (refDCval >= 166) {
if (Cs < 4423) {
if (Rs < 5439) {
return 4;
}
else if (Rs >= 5439) {
return 2;
}
}
else if (Cs >= 4423) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (Cs < 1641) {
if (Rs < 3098) {
if (Rs < 1345) {
if (ssDCval < 122) {
if (negBalance < 2) {
return 2;
}
else if (negBalance >= 2) {
if (refDCval < 63) {
if (AFD < 3) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (SC < 2401) {
return 1;
}
else if (SC >= 2401) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
if (SC < 2601) {
return 8;
}
else if (SC >= 2601) {
return 4;
}
}
}
}
}
else if (AFD >= 3) {
if (MVDiff < 6) {
if (tcor < 99) {
if (Cs < 1340) {
return 4;
}
else if (Cs >= 1340) {
return 8;
}
}
else if (tcor >= 99) {
return 4;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (refDCval >= 63) {
if (refDCval < 76) {
if (Cs < 1431) {
if (mu_mv_mag_sq < 116) {
if (Rs < 321) {
return 4;
}
else if (Rs >= 321) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
if (Rs < 1030) {
return 2;
}
else if (Rs >= 1030) {
return 4;
}
}
}
else if (Cs >= 1431) {
return 8;
}
}
else if (refDCval >= 76) {
if (Rs < 1074) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (Rs >= 1074) {
return 8;
}
}
}
}
}
else if (ssDCval >= 122) {
if (refDCval < 127) {
if (AFD < 2) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
else if (AFD >= 2) {
if (Cs < 1462) {
if (mu_mv_mag_sq < 124) {
return 8;
}
else if (mu_mv_mag_sq >= 124) {
return 4;
}
}
else if (Cs >= 1462) {
return 4;
}
}
}
else if (refDCval >= 127) {
if (refDCval < 133) {
if (Rs < 1002) {
return 8;
}
else if (Rs >= 1002) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
else if (refDCval >= 133) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
else if (Rs >= 1345) {
if (ssDCval < 86) {
if (mcTcor < 98) {
if (TSC < 4) {
if (refDCval < 51) {
if (RsDiff < 3) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (RsDiff >= 3) {
if (Rs < 2224) {
return 8;
}
else if (Rs >= 2224) {
return 4;
}
}
}
else if (refDCval >= 51) {
if (posBalance < 54) {
if (Cs < 1633) {
return 8;
}
else if (Cs >= 1633) {
return 4;
}
}
else if (posBalance >= 54) {
if (ssDCval < 75) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
}
}
else if (TSC >= 4) {
if (negBalance < 62) {
return 8;
}
else if (negBalance >= 62) {
return 4;
}
}
}
else if (mcTcor >= 98) {
if (posBalance < 39) {
if (Cs < 1458) {
if (negBalance < 29) {
if (SC < 2778) {
if (Rs < 1411) {
return 8;
}
else if (Rs >= 1411) {
return 1;
}
}
else if (SC >= 2778) {
if (Rs < 2342) {
if (Cs < 1343) {
if (Cs < 1342) {
return 8;
}
else if (Cs >= 1342) {
if (Rs < 1958) {
return 8;
}
else if (Rs >= 1958) {
return 2;
}
}
}
else if (Cs >= 1343) {
return 8;
}
}
else if (Rs >= 2342) {
if (Rs < 2371) {
return 2;
}
else if (Rs >= 2371) {
return 8;
}
}
}
}
else if (negBalance >= 29) {
if (Rs < 2163) {
return 4;
}
else if (Rs >= 2163) {
return 2;
}
}
}
else if (Cs >= 1458) {
if (ssDCval < 61) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
return 1;
}
}
else if (ssDCval >= 61) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Rs < 2276) {
return 4;
}
else if (Rs >= 2276) {
return 4;
}
}
}
}
}
else if (posBalance >= 39) {
if (refDCval < 79) {
if (ssDCval < 66) {
if (Rs < 1974) {
if (Cs < 1357) {
return 8;
}
else if (Cs >= 1357) {
return 4;
}
}
else if (Rs >= 1974) {
return 8;
}
}
else if (ssDCval >= 66) {
return 4;
}
}
else if (refDCval >= 79) {
if (Rs < 2483) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (Rs >= 2483) {
return 4;
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 139) {
if (mcTcor < 78) {
return 4;
}
else if (mcTcor >= 78) {
if (ssDCval < 117) {
if (RsCsDiff < 16) {
return 8;
}
else if (RsCsDiff >= 16) {
if (diffAFD < -2) {
if (Rs < 1915) {
return 8;
}
else if (Rs >= 1915) {
return 2;
}
}
else if (diffAFD >= -2) {
if (diffMVdiffVal < -5) {
return 2;
}
else if (diffMVdiffVal >= -5) {
if (TSC < 11) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
else if (ssDCval >= 117) {
if (ssDCval < 119) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (ssDCval >= 119) {
if (Rs < 2819) {
if (diffMVdiffVal < -149) {
return 1;
}
else if (diffMVdiffVal >= -149) {
if (Rs < 1362) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (Rs >= 1362) {
return 8;
}
}
}
else if (Rs >= 2819) {
if (Rs < 2924) {
return 1;
}
else if (Rs >= 2924) {
if (Rs < 3070) {
return 2;
}
else if (Rs >= 3070) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 139) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
if (diffRsCsdiff < -9) {
if (Rs < 2776) {
return 8;
}
else if (Rs >= 2776) {
return 2;
}
}
else if (diffRsCsdiff >= -9) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3098) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 126) {
if (Rs < 4249) {
if (AFD < 3) {
if (ssDCval < 63) {
if (Rs < 3683) {
return 4;
}
else if (Rs >= 3683) {
return 8;
}
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (AFD >= 3) {
if (RsDiff < 6) {
if (Rs < 3300) {
return 2;
}
else if (Rs >= 3300) {
return 4;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (Rs >= 4249) {
if (Cs < 1580) {
return 8;
}
else if (Cs >= 1580) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1570) {
if (RsCsDiff < 12) {
if (Rs < 3659) {
if (Rs < 3270) {
return 8;
}
else if (Rs >= 3270) {
if (Rs < 3543) {
return 1;
}
else if (Rs >= 3543) {
return 4;
}
}
}
else if (Rs >= 3659) {
return 8;
}
}
else if (RsCsDiff >= 12) {
return 4;
}
}
else if (Cs >= 1570) {
if (Rs < 3265) {
return 8;
}
else if (Rs >= 3265) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 113) {
if (ssDCval < 88) {
if (mu_mv_mag_sq < 126) {
return 2;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 3377) {
return 4;
}
else if (Rs >= 3377) {
return 2;
}
}
}
else if (ssDCval >= 88) {
if (SC < 5682) {
if (Rs < 3279) {
if (Rs < 3154) {
return 8;
}
else if (Rs >= 3154) {
return 2;
}
}
else if (Rs >= 3279) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 3499) {
return 2;
}
else if (Rs >= 3499) {
return 4;
}
}
}
}
else if (SC >= 5682) {
if (ssDCval < 96) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 2;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
}
else if (ssDCval >= 113) {
if (MVDiff < 8) {
if (Rs < 3127) {
return 1;
}
else if (Rs >= 3127) {
return 2;
}
}
else if (MVDiff >= 8) {
if (Cs < 1488) {
return 4;
}
else if (Cs >= 1488) {
if (Rs < 3325) {
return 2;
}
else if (Rs >= 3325) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1641) {
if (ssDCval < 102) {
if (posBalance < 3) {
if (refDCval < 65) {
if (Rs < 8850) {
return 8;
}
else if (Rs >= 8850) {
return 2;
}
}
else if (refDCval >= 65) {
if (diffTSC < -1) {
if (ssDCval < 72) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
return 4;
}
}
else if (ssDCval >= 72) {
if (Rs < 2268) {
return 4;
}
else if (Rs >= 2268) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (ssDCval < 87) {
if (Rs < 3298) {
return 8;
}
else if (Rs >= 3298) {
return 4;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
}
}
else if (posBalance >= 3) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 3332) {
if (posBalance < 36) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (Rs < 3549) {
if (Rs < 1819) {
if (Cs < 1697) {
if (Rs < 863) {
return 1;
}
else if (Rs >= 863) {
return 4;
}
}
else if (Cs >= 1697) {
return 8;
}
}
else if (Rs >= 1819) {
if (ssDCval < 62) {
if (Rs < 3233) {
return 1;
}
else if (Rs >= 3233) {
if (Cs < 2221) {
return 4;
}
else if (Cs >= 2221) {
return 1;
}
}
}
else if (ssDCval >= 62) {
if (Rs < 2254) {
return 8;
}
else if (Rs >= 2254) {
return 1;
}
}
}
}
else if (Rs >= 3549) {
return 4;
}
}
}
else if (posBalance >= 36) {
if (Rs < 1659) {
if (ssDCval < 57) {
if (Rs < 1002) {
if (Rs < 791) {
return 1;
}
else if (Rs >= 791) {
return 8;
}
}
else if (Rs >= 1002) {
return 4;
}
}
else if (ssDCval >= 57) {
if (tcor < 98) {
if (SC < 2874) {
return 8;
}
else if (SC >= 2874) {
if (SC < 2946) {
if (Rs < 1003) {
return 8;
}
else if (Rs >= 1003) {
return 1;
}
}
else if (SC >= 2946) {
return 8;
}
}
}
else if (tcor >= 98) {
if (Cs < 1701) {
return 8;
}
else if (Cs >= 1701) {
return 1;
}
}
}
}
else if (Rs >= 1659) {
if (SC < 4374) {
if (SC < 4093) {
return 8;
}
else if (SC >= 4093) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
else if (SC >= 4374) {
if (negBalance < 19) {
if (Cs < 2083) {
return 8;
}
else if (Cs >= 2083) {
return 4;
}
}
else if (negBalance >= 19) {
if (SC < 4908) {
if (diffRsCsdiff < -5) {
if (Rs < 2924) {
return 4;
}
else if (Rs >= 2924) {
return 4;
}
}
else if (diffRsCsdiff >= -5) {
return 8;
}
}
else if (SC >= 4908) {
return 8;
}
}
}
}
}
}
else if (Cs >= 3332) {
return 8;
}
}
else if (ssDCval >= 67) {
if (mu_mv_mag_sq < 123) {
if (posBalance < 36) {
if (Cs < 2318) {
if (ssDCval < 92) {
if (Rs < 3080) {
if (negBalance < 7) {
return 1;
}
else if (negBalance >= 7) {
if (Cs < 2143) {
if (posBalance < 11) {
return 1;
}
else if (posBalance >= 11) {
return 8;
}
}
else if (Cs >= 2143) {
if (Rs < 2468) {
return 2;
}
else if (Rs >= 2468) {
return 1;
}
}
}
}
else if (Rs >= 3080) {
if (Rs < 4312) {
return 2;
}
else if (Rs >= 4312) {
return 4;
}
}
}
else if (ssDCval >= 92) {
if (negBalance < 7) {
if (mu_mv_mag_sq < 119) {
return 2;
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 2479) {
return 1;
}
else if (Rs >= 2479) {
return 8;
}
}
}
else if (negBalance >= 7) {
if (Rs < 2320) {
return 2;
}
else if (Rs >= 2320) {
if (Rs < 2381) {
return 4;
}
else if (Rs >= 2381) {
return 2;
}
}
}
}
}
else if (Cs >= 2318) {
return 8;
}
}
else if (posBalance >= 36) {
if (TSCindex < 1) {
return 2;
}
else if (TSCindex >= 1) {
if (posBalance < 39) {
if (Cs < 2305) {
if (Rs < 4354) {
return 2;
}
else if (Rs >= 4354) {
return 4;
}
}
else if (Cs >= 2305) {
return 8;
}
}
else if (posBalance >= 39) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (SC < 6962) {
if (Rs < 1114) {
if (Rs < 623) {
return 8;
}
else if (Rs >= 623) {
return 1;
}
}
else if (Rs >= 1114) {
if (SC < 4488) {
if (SC < 4439) {
if (posBalance < 59) {
if (Cs < 1644) {
if (Rs < 2186) {
return 4;
}
else if (Rs >= 2186) {
return 8;
}
}
else if (Cs >= 1644) {
if (Rs < 2637) {
return 8;
}
else if (Rs >= 2637) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
}
else if (posBalance >= 59) {
return 4;
}
}
else if (SC >= 4439) {
if (AFD < 3) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (SC >= 4488) {
if (diffRsCsdiff < -210) {
if (Rs < 3479) {
return 8;
}
else if (Rs >= 3479) {
return 4;
}
}
else if (diffRsCsdiff >= -210) {
if (mu_mv_mag_sq < 124) {
if (diffAFD < -1) {
if (Cs < 2232) {
return 2;
}
else if (Cs >= 2232) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
}
}
}
else if (SC >= 6962) {
if (SC < 7964) {
if (SC < 7205) {
if (SC < 6999) {
return 4;
}
else if (SC >= 6999) {
return 8;
}
}
else if (SC >= 7205) {
return 4;
}
}
else if (SC >= 7964) {
if (diffRsCsdiff < -14) {
if (Rs < 4396) {
return 1;
}
else if (Rs >= 4396) {
if (Rs < 6632) {
return 8;
}
else if (Rs >= 6632) {
return 4;
}
}
}
else if (diffRsCsdiff >= -14) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 6650) {
if (SC < 8342) {
if (Rs < 3697) {
if (AFD < 10) {
if (tcor < 99) {
if (posBalance < 60) {
if (mcTcor < 96) {
if (SC < 2882) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (SC >= 2882) {
if (Cs < 2403) {
if (diffRsCsdiff < 18) {
if (diffTSC < -3) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
return 4;
}
}
else if (diffTSC >= -3) {
return 8;
}
}
else if (diffRsCsdiff >= 18) {
if (mu_mv_mag_sq < 114) {
if (RsCsDiff < 67) {
if (Rs < 3231) {
return 4;
}
else if (Rs >= 3231) {
return 4;
}
}
else if (RsCsDiff >= 67) {
if (Rs < 3256) {
return 8;
}
else if (Rs >= 3256) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (diffRsCsdiff < 19) {
if (Rs < 2855) {
return 8;
}
else if (Rs >= 2855) {
return 4;
}
}
else if (diffRsCsdiff >= 19) {
return 8;
}
}
}
}
else if (Cs >= 2403) {
if (refDCval < 91) {
if (Rs < 2179) {
if (Cs < 2418) {
return 8;
}
else if (Cs >= 2418) {
return 4;
}
}
else if (Rs >= 2179) {
if (mu_mv_mag_sq < 111) {
if (Cs < 2563) {
return 8;
}
else if (Cs >= 2563) {
if (Rs < 2918) {
return 4;
}
else if (Rs >= 2918) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 3565) {
return 8;
}
else if (Cs >= 3565) {
if (Cs < 3653) {
return 4;
}
else if (Cs >= 3653) {
return 8;
}
}
}
}
}
else if (refDCval >= 91) {
if (negBalance < 49) {
if (Cs < 2511) {
return 8;
}
else if (Cs >= 2511) {
if (Cs < 3278) {
if (Rs < 3075) {
return 8;
}
else if (Rs >= 3075) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (Cs >= 3278) {
if (Rs < 2597) {
return 4;
}
else if (Rs >= 2597) {
return 8;
}
}
}
}
else if (negBalance >= 49) {
if (Cs < 3166) {
return 4;
}
else if (Cs >= 3166) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (SC < 4221) {
if (Rs < 975) {
return 1;
}
else if (Rs >= 975) {
if (refDCval < 88) {
return 8;
}
else if (refDCval >= 88) {
if (MVDiff < 4) {
if (ssDCval < 87) {
return 4;
}
else if (ssDCval >= 87) {
return 8;
}
}
else if (MVDiff >= 4) {
return 2;
}
}
}
}
else if (SC >= 4221) {
if (Cs < 1935) {
if (Cs < 1706) {
if (Cs < 1663) {
return 2;
}
else if (Cs >= 1663) {
return 4;
}
}
else if (Cs >= 1706) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
if (MVDiff < 1) {
if (CsDiff < 3) {
if (Rs < 2565) {
return 8;
}
else if (Rs >= 2565) {
if (Cs < 1900) {
if (ssDCval < 65) {
if (Cs < 1781) {
if (Cs < 1744) {
return 4;
}
else if (Cs >= 1744) {
return 8;
}
}
else if (Cs >= 1781) {
if (Rs < 2774) {
return 8;
}
else if (Rs >= 2774) {
if (negBalance < 23) {
if (Rs < 3057) {
return 4;
}
else if (Rs >= 3057) {
return 8;
}
}
else if (negBalance >= 23) {
if (Rs < 2969) {
return 4;
}
else if (Rs >= 2969) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (Rs < 3003) {
return 4;
}
else if (Rs >= 3003) {
if (posBalance < 39) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 4;
}
}
else if (posBalance >= 39) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 65) {
if (Cs < 1881) {
return 4;
}
else if (Cs >= 1881) {
if (Rs < 2961) {
return 4;
}
else if (Rs >= 2961) {
return 4;
}
}
}
}
else if (Cs >= 1900) {
if (Rs < 2687) {
return 4;
}
else if (Rs >= 2687) {
if (Rs < 2976) {
if (Cs < 1913) {
return 8;
}
else if (Cs >= 1913) {
return 8;
}
}
else if (Rs >= 2976) {
return 4;
}
}
}
}
}
else if (CsDiff >= 3) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
return 4;
}
}
}
else if (MVDiff >= 1) {
if (Rs < 2504) {
return 4;
}
else if (Rs >= 2504) {
return 8;
}
}
}
}
}
else if (Cs >= 1935) {
if (Rs < 2186) {
if (ssDCval < 81) {
return 4;
}
else if (ssDCval >= 81) {
if (Rs < 1979) {
return 4;
}
else if (Rs >= 1979) {
return 8;
}
}
}
else if (Rs >= 2186) {
if (SC < 6937) {
if (mu_mv_mag_sq < 121) {
if (diffRsCsdiff < -20) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
if (Cs < 2363) {
return 8;
}
else if (Cs >= 2363) {
if (Cs < 3216) {
return 4;
}
else if (Cs >= 3216) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -20) {
if (mu_mv_mag_sq < 113) {
if (Cs < 2837) {
return 8;
}
else if (Cs >= 2837) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (diffRsCsdiff < 20) {
if (tcor < 95) {
if (SC < 5471) {
return 4;
}
else if (SC >= 5471) {
return 8;
}
}
else if (tcor >= 95) {
if (ssDCval < 67) {
if (Rs < 3033) {
return 4;
}
else if (Rs >= 3033) {
return 8;
}
}
else if (ssDCval >= 67) {
if (negBalance < 60) {
return 8;
}
else if (negBalance >= 60) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 20) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (Cs < 2096) {
if (SC < 5073) {
if (Cs < 1945) {
if (Cs < 1942) {
return 8;
}
else if (Cs >= 1942) {
return 4;
}
}
else if (Cs >= 1945) {
return 8;
}
}
else if (SC >= 5073) {
if (Cs < 2050) {
if (Cs < 2017) {
return 2;
}
else if (Cs >= 2017) {
return 8;
}
}
else if (Cs >= 2050) {
return 4;
}
}
}
else if (Cs >= 2096) {
if (negBalance < 47) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
else if (negBalance >= 47) {
if (Rs < 3330) {
return 8;
}
else if (Rs >= 3330) {
return 4;
}
}
}
}
}
else if (SC >= 6937) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
}
}
}
}
else if (posBalance >= 60) {
if (Rs < 3426) {
if (diffTSC < 2) {
if (diffAFD < -2) {
if (Rs < 3154) {
return 4;
}
else if (Rs >= 3154) {
return 8;
}
}
else if (diffAFD >= -2) {
if (Cs < 2911) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (diffMVdiffVal < -1) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (Cs >= 2911) {
if (Cs < 3022) {
return 4;
}
else if (Cs >= 3022) {
if (SC < 6775) {
return 8;
}
else if (SC >= 6775) {
return 4;
}
}
}
}
}
else if (diffTSC >= 2) {
if (Rs < 2664) {
return 4;
}
else if (Rs >= 2664) {
if (Rs < 3324) {
return 8;
}
else if (Rs >= 3324) {
return 4;
}
}
}
}
else if (Rs >= 3426) {
if (gchDC < 1) {
if (Cs < 3258) {
return 4;
}
else if (Cs >= 3258) {
if (Rs < 3481) {
return 4;
}
else if (Rs >= 3481) {
return 8;
}
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
else if (tcor >= 99) {
if (Cs < 2284) {
if (SC < 5125) {
if (Rs < 2477) {
if (AFD < 8) {
if (Rs < 2471) {
return 8;
}
else if (Rs >= 2471) {
if (Cs < 1728) {
return 8;
}
else if (Cs >= 1728) {
return 1;
}
}
}
else if (AFD >= 8) {
return 4;
}
}
else if (Rs >= 2477) {
if (posBalance < 39) {
if (Rs < 2931) {
return 8;
}
else if (Rs >= 2931) {
return 8;
}
}
else if (posBalance >= 39) {
if (Cs < 1839) {
return 4;
}
else if (Cs >= 1839) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 2595) {
return 8;
}
else if (Rs >= 2595) {
return 4;
}
}
}
}
}
}
else if (SC >= 5125) {
if (mu_mv_mag_sq < 127) {
return 2;
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
}
else if (Cs >= 2284) {
if (diffRsCsdiff < -23) {
if (Cs < 2739) {
return 8;
}
else if (Cs >= 2739) {
return 4;
}
}
else if (diffRsCsdiff >= -23) {
if (posBalance < 37) {
if (Cs < 2558) {
return 8;
}
else if (Cs >= 2558) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
else if (posBalance >= 37) {
return 8;
}
}
}
}
}
else if (AFD >= 10) {
if (Cs < 2278) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
return 8;
}
}
else if (Cs >= 2278) {
if (negBalance < 55) {
if (Rs < 1377) {
return 2;
}
else if (Rs >= 1377) {
if (Cs < 3165) {
if (Cs < 2678) {
if (mcTcor < 94) {
if (diffTSC < 1) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (mcTcor >= 94) {
return 8;
}
}
else if (Cs >= 2678) {
return 4;
}
}
else if (Cs >= 3165) {
if (posBalance < 62) {
if (diffTSC < -1) {
if (SC < 6941) {
return 4;
}
else if (SC >= 6941) {
return 8;
}
}
else if (diffTSC >= -1) {
if (posBalance < 57) {
if (Rs < 3234) {
return 8;
}
else if (Rs >= 3234) {
return 4;
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (posBalance >= 62) {
return 4;
}
}
}
}
else if (negBalance >= 55) {
return 4;
}
}
}
}
else if (Rs >= 3697) {
if (refDCval < 81) {
if (refDCval < 60) {
if (RsDiff < 4) {
if (Rs < 4271) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Cs < 2106) {
if (Rs < 4222) {
return 4;
}
else if (Rs >= 4222) {
return 4;
}
}
else if (Cs >= 2106) {
return 4;
}
}
}
else if (Rs >= 4271) {
if (Cs < 2095) {
return 4;
}
else if (Cs >= 2095) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (tcor < 96) {
if (TSC < 3) {
if (Cs < 2589) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (MVDiff < 2) {
if (Rs < 4341) {
return 8;
}
else if (Rs >= 4341) {
return 4;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (Cs >= 2589) {
return 8;
}
}
else if (TSC >= 3) {
return 4;
}
}
else if (tcor >= 96) {
if (Cs < 2755) {
return 4;
}
else if (Cs >= 2755) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (Rs < 4358) {
if (SC < 6438) {
if (Cs < 2031) {
if (RsCsDiff < 38) {
return 8;
}
else if (RsCsDiff >= 38) {
return 4;
}
}
else if (Cs >= 2031) {
if (Rs < 4252) {
return 8;
}
else if (Rs >= 4252) {
if (Rs < 4327) {
return 8;
}
else if (Rs >= 4327) {
return 8;
}
}
}
}
else if (SC >= 6438) {
return 8;
}
}
else if (Rs >= 4358) {
if (tcor < 93) {
if (RsDiff < 6) {
if (Cs < 2608) {
return 4;
}
else if (Cs >= 2608) {
return 4;
}
}
else if (RsDiff >= 6) {
if (mu_mv_mag_sq < 115) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Cs < 2172) {
return 8;
}
else if (Cs >= 2172) {
if (Rs < 4848) {
return 4;
}
else if (Rs >= 4848) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
}
else if (tcor >= 93) {
if (Rs < 5099) {
if (diffRsCsdiff < 15) {
if (posBalance < 25) {
return 8;
}
else if (posBalance >= 25) {
if (diffAFD < -1) {
if (RsDiff < 6) {
if (Rs < 4390) {
return 8;
}
else if (Rs >= 4390) {
if (Rs < 4645) {
return 8;
}
else if (Rs >= 4645) {
return 4;
}
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (diffAFD >= -1) {
if (Rs < 4384) {
return 4;
}
else if (Rs >= 4384) {
if (mu_mv_mag_sq < 111) {
return 8;
}
else if (mu_mv_mag_sq >= 111) {
if (mu_mv_mag_sq < 117) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (posBalance < 27) {
return 8;
}
else if (posBalance >= 27) {
if (mu_mv_mag_sq < 114) {
return 4;
}
else if (mu_mv_mag_sq >= 114) {
if (Rs < 4547) {
return 4;
}
else if (Rs >= 4547) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (Rs < 4763) {
return 4;
}
else if (Rs >= 4763) {
return 4;
}
}
}
}
}
}
}
}
else if (diffRsCsdiff >= 15) {
return 8;
}
}
else if (Rs >= 5099) {
return 4;
}
}
}
}
}
else if (refDCval >= 60) {
if (SC < 5930) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 2;
}
}
else if (SC >= 5930) {
if (Rs < 5916) {
if (SC < 6043) {
if (Rs < 3993) {
return 8;
}
else if (Rs >= 3993) {
return 2;
}
}
else if (SC >= 6043) {
return 8;
}
}
else if (Rs >= 5916) {
return 4;
}
}
}
}
else if (refDCval >= 81) {
if (mcTcor < 93) {
if (Cs < 2439) {
return 1;
}
else if (Cs >= 2439) {
if (ssDCval < 101) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Rs < 3798) {
return 8;
}
else if (Rs >= 3798) {
return 4;
}
}
}
else if (ssDCval >= 101) {
if (SC < 6780) {
return 8;
}
else if (SC >= 6780) {
return 1;
}
}
}
}
else if (mcTcor >= 93) {
if (Cs < 2143) {
if (ssDCval < 98) {
if (MVDiff < 1) {
if (AFD < 4) {
if (Rs < 4717) {
return 2;
}
else if (Rs >= 4717) {
if (Rs < 5377) {
return 8;
}
else if (Rs >= 5377) {
return 4;
}
}
}
else if (AFD >= 4) {
if (ssDCval < 91) {
if (Rs < 4648) {
return 2;
}
else if (Rs >= 4648) {
return 8;
}
}
else if (ssDCval >= 91) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
else if (MVDiff >= 1) {
if (posBalance < 42) {
if (Rs < 4481) {
return 8;
}
else if (Rs >= 4481) {
return 2;
}
}
else if (posBalance >= 42) {
return 2;
}
}
}
else if (ssDCval >= 98) {
if (RsDiff < 6) {
if (Cs < 1926) {
return 4;
}
else if (Cs >= 1926) {
return 2;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (Cs >= 2143) {
if (ssDCval < 98) {
if (RsCsDiff < 46) {
if (SC < 7192) {
if (diffRsCsdiff < -29) {
return 4;
}
else if (diffRsCsdiff >= -29) {
if (Rs < 3716) {
return 4;
}
else if (Rs >= 3716) {
return 8;
}
}
}
else if (SC >= 7192) {
if (negBalance < 39) {
if (Cs < 3611) {
if (SC < 7207) {
if (Rs < 4500) {
return 8;
}
else if (Rs >= 4500) {
return 4;
}
}
else if (SC >= 7207) {
return 4;
}
}
else if (Cs >= 3611) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
return 4;
}
}
}
else if (negBalance >= 39) {
return 8;
}
}
}
else if (RsCsDiff >= 46) {
if (Cs < 2571) {
return 2;
}
else if (Cs >= 2571) {
return 4;
}
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
}
}
else if (SC >= 8342) {
if (MVDiff < 2) {
if (TSCindex < 3) {
if (Rs < 4529) {
return 1;
}
else if (Rs >= 4529) {
return 8;
}
}
else if (TSCindex >= 3) {
if (Cs < 3666) {
if (Rs < 5401) {
if (Rs < 4811) {
return 8;
}
else if (Rs >= 4811) {
return 4;
}
}
else if (Rs >= 5401) {
return 8;
}
}
else if (Cs >= 3666) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (Rs < 5860) {
return 8;
}
else if (Rs >= 5860) {
if (ssDCval < 82) {
if (Cs < 3885) {
return 8;
}
else if (Cs >= 3885) {
return 2;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
}
}
}
else if (Rs >= 6650) {
if (ssDCval < 95) {
if (ssDCval < 68) {
if (Rs < 8744) {
return 8;
}
else if (Rs >= 8744) {
return 2;
}
}
else if (ssDCval >= 68) {
return 4;
}
}
else if (ssDCval >= 95) {
return 8;
}
}
}
}
}
else if (ssDCval >= 102) {
if (ssDCval < 129) {
if (Rs < 1889) {
if (refDCval < 118) {
if (Cs < 2015) {
if (Rs < 1581) {
if (posBalance < 54) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 8;
}
}
else if (posBalance >= 54) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (Rs >= 1581) {
return 8;
}
}
else if (Cs >= 2015) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 4;
}
}
}
else if (refDCval >= 118) {
if (SC < 2886) {
return 4;
}
else if (SC >= 2886) {
return 8;
}
}
}
else if (Rs >= 1889) {
if (refDCval < 105) {
if (posBalance < 56) {
if (negBalance < 38) {
if (Cs < 1857) {
if (Rs < 3251) {
return 8;
}
else if (Rs >= 3251) {
return 4;
}
}
else if (Cs >= 1857) {
return 8;
}
}
else if (negBalance >= 38) {
if (CsDiff < 4) {
if (Rs < 4089) {
return 8;
}
else if (Rs >= 4089) {
return 2;
}
}
else if (CsDiff >= 4) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (Rs < 3655) {
return 8;
}
else if (Rs >= 3655) {
return 2;
}
}
}
}
}
else if (posBalance >= 56) {
if (Cs < 2984) {
if (Rs < 4902) {
if (Rs < 3849) {
return 8;
}
else if (Rs >= 3849) {
return 1;
}
}
else if (Rs >= 4902) {
return 2;
}
}
else if (Cs >= 2984) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (diffAFD < -2) {
if (diffRsCsdiff < -978) {
return 1;
}
else if (diffRsCsdiff >= -978) {
if (ssDCval < 123) {
if (posBalance < 63) {
if (negBalance < 16) {
if (diffMVdiffVal < -104) {
return 4;
}
else if (diffMVdiffVal >= -104) {
return 8;
}
}
else if (negBalance >= 16) {
if (RsCsDiff < 2) {
if (Rs < 2361) {
return 2;
}
else if (Rs >= 2361) {
return 4;
}
}
else if (RsCsDiff >= 2) {
if (posBalance < 56) {
if (Cs < 1913) {
if (Rs < 2843) {
return 8;
}
else if (Rs >= 2843) {
return 4;
}
}
else if (Cs >= 1913) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
return 4;
}
}
}
else if (posBalance >= 56) {
if (gchDC < 1) {
if (ssDCval < 115) {
return 8;
}
else if (ssDCval >= 115) {
if (Rs < 2899) {
return 8;
}
else if (Rs >= 2899) {
return 2;
}
}
}
else if (gchDC >= 1) {
if (Rs < 5824) {
return 2;
}
else if (Rs >= 5824) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 63) {
if (Rs < 5171) {
return 1;
}
else if (Rs >= 5171) {
return 8;
}
}
}
else if (ssDCval >= 123) {
if (MVDiff < 4) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (MVDiff >= 4) {
if (Cs < 2696) {
return 8;
}
else if (Cs >= 2696) {
if (Rs < 2656) {
return 4;
}
else if (Rs >= 2656) {
return 8;
}
}
}
}
}
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 125) {
if (CsDiff < 1) {
if (SC < 5745) {
if (Cs < 2777) {
return 8;
}
else if (Cs >= 2777) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 2647) {
return 4;
}
else if (Rs >= 2647) {
return 8;
}
}
}
}
else if (SC >= 5745) {
if (Rs < 5216) {
return 2;
}
else if (Rs >= 5216) {
return 8;
}
}
}
else if (CsDiff >= 1) {
if (TSC < 10) {
if (diffAFD < 2) {
if (Rs < 3444) {
if (refDCval < 123) {
return 8;
}
else if (refDCval >= 123) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (Rs >= 3444) {
if (Rs < 3446) {
return 4;
}
else if (Rs >= 3446) {
if (Rs < 6432) {
return 8;
}
else if (Rs >= 6432) {
if (Cs < 3936) {
return 4;
}
else if (Cs >= 3936) {
return 8;
}
}
}
}
}
else if (diffAFD >= 2) {
if (AFD < 5) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
if (Rs < 3134) {
return 1;
}
else if (Rs >= 3134) {
return 8;
}
}
}
else if (AFD >= 5) {
if (Rs < 3165) {
return 8;
}
else if (Rs >= 3165) {
if (Rs < 3538) {
return 4;
}
else if (Rs >= 3538) {
if (Rs < 6457) {
return 8;
}
else if (Rs >= 6457) {
return 4;
}
}
}
}
}
}
else if (TSC >= 10) {
if (SC < 7020) {
if (Rs < 2403) {
return 8;
}
else if (Rs >= 2403) {
if (negBalance < 51) {
if (Rs < 2928) {
return 1;
}
else if (Rs >= 2928) {
return 8;
}
}
else if (negBalance >= 51) {
return 2;
}
}
}
else if (SC >= 7020) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (negBalance < 56) {
if (diffAFD < 3) {
if (posBalance < 1) {
if (Rs < 2482) {
return 1;
}
else if (Rs >= 2482) {
return 8;
}
}
else if (posBalance >= 1) {
return 8;
}
}
else if (diffAFD >= 3) {
if (ssDCval < 107) {
return 2;
}
else if (ssDCval >= 107) {
return 8;
}
}
}
else if (negBalance >= 56) {
if (Rs < 4305) {
if (Rs < 3247) {
return 1;
}
else if (Rs >= 3247) {
return 2;
}
}
else if (Rs >= 4305) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 129) {
if (mcTcor < 96) {
if (negBalance < 38) {
if (ssDCval < 200) {
if (ssDCval < 130) {
return 1;
}
else if (ssDCval >= 130) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 1;
}
}
}
else if (ssDCval >= 200) {
if (Rs < 2424) {
if (Rs < 1185) {
return 1;
}
else if (Rs >= 1185) {
return 8;
}
}
else if (Rs >= 2424) {
return 4;
}
}
}
else if (negBalance >= 38) {
if (ssDCval < 137) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
else if (ssDCval >= 137) {
if (Rs < 2759) {
return 8;
}
else if (Rs >= 2759) {
if (Rs < 5949) {
return 1;
}
else if (Rs >= 5949) {
return 8;
}
}
}
}
}
else if (mcTcor >= 96) {
if (SC < 7857) {
if (Cs < 3891) {
if (Rs < 4487) {
if (SC < 3058) {
if (posBalance < 55) {
if (Rs < 888) {
return 1;
}
else if (Rs >= 888) {
return 8;
}
}
else if (posBalance >= 55) {
if (ssDCval < 132) {
return 4;
}
else if (ssDCval >= 132) {
return 8;
}
}
}
else if (SC >= 3058) {
if (mu_mv_mag_sq < 111) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 111) {
if (diffAFD < -20) {
if (Rs < 2420) {
return 8;
}
else if (Rs >= 2420) {
return 1;
}
}
else if (diffAFD >= -20) {
return 8;
}
}
}
}
else if (Rs >= 4487) {
if (Rs < 4973) {
return 1;
}
else if (Rs >= 4973) {
return 2;
}
}
}
else if (Cs >= 3891) {
if (Cs < 4280) {
return 1;
}
else if (Cs >= 4280) {
return 4;
}
}
}
else if (SC >= 7857) {
if (Cs < 2853) {
return 2;
}
else if (Cs >= 2853) {
return 8;
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect14(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1336) {
if (Rs < 2376) {
if (Cs < 535) {
if (negBalance < 14) {
if (MVDiff < 7) {
if (SC < 792) {
if (ssDCval < 48) {
if (SC < 385) {
if (negBalance < 8) {
if (Rs < 2) {
if (ssDCval < 22) {
if (ssDCval < 10) {
return 4;
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (ssDCval >= 22) {
if (ssDCval < 38) {
return 4;
}
else if (ssDCval >= 38) {
return 2;
}
}
}
else if (Rs >= 2) {
if (Cs < 185) {
if (SC < 377) {
return 8;
}
else if (SC >= 377) {
return 4;
}
}
else if (Cs >= 185) {
if (Cs < 195) {
return 4;
}
else if (Cs >= 195) {
if (Rs < 114) {
return 1;
}
else if (Rs >= 114) {
return 8;
}
}
}
}
}
else if (negBalance >= 8) {
if (SC < 204) {
if (ssDCval < 31) {
if (ssDCval < 24) {
if (Rs < 61) {
return 8;
}
else if (Rs >= 61) {
return 1;
}
}
else if (ssDCval >= 24) {
if (mu_mv_mag_sq < 77) {
if (Rs < 67) {
return 1;
}
else if (Rs >= 67) {
return 8;
}
}
else if (mu_mv_mag_sq >= 77) {
return 4;
}
}
}
else if (ssDCval >= 31) {
if (ssDCval < 39) {
if (Rs < 74) {
return 4;
}
else if (Rs >= 74) {
return 8;
}
}
else if (ssDCval >= 39) {
if (diffMVdiffVal < -4) {
if (Rs < 54) {
return 8;
}
else if (Rs >= 54) {
return 1;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
}
else if (SC >= 204) {
if (ssDCval < 34) {
if (posBalance < 29) {
return 8;
}
else if (posBalance >= 29) {
if (Cs < 109) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (Cs >= 109) {
return 1;
}
}
}
else if (ssDCval >= 34) {
if (refDCval < 37) {
return 4;
}
else if (refDCval >= 37) {
if (Rs < 81) {
return 8;
}
else if (Rs >= 81) {
if (Rs < 150) {
return 1;
}
else if (Rs >= 150) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 385) {
if (Cs < 155) {
if (ssDCval < 19) {
if (Cs < 138) {
return 8;
}
else if (Cs >= 138) {
return 4;
}
}
else if (ssDCval >= 19) {
return 4;
}
}
else if (Cs >= 155) {
if (ssDCval < 36) {
if (ssDCval < 29) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
else if (ssDCval >= 29) {
if (Cs < 213) {
return 8;
}
else if (Cs >= 213) {
if (mu_mv_mag_sq < 128) {
return 4;
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 258) {
return 8;
}
else if (Cs >= 258) {
return 4;
}
}
}
}
}
else if (ssDCval >= 36) {
if (tcor < 96) {
if (ssDCval < 40) {
if (Rs < 160) {
if (Cs < 316) {
return 4;
}
else if (Cs >= 316) {
return 1;
}
}
else if (Rs >= 160) {
if (posBalance < 50) {
return 4;
}
else if (posBalance >= 50) {
return 8;
}
}
}
else if (ssDCval >= 40) {
return 1;
}
}
else if (tcor >= 96) {
if (Rs < 363) {
return 1;
}
else if (Rs >= 363) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 48) {
if (ssDCval < 84) {
if (Cs < 7) {
return 2;
}
else if (Cs >= 7) {
if (ssDCval < 61) {
if (Cs < 283) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (mcTcor < 99) {
if (tcor < 83) {
return 1;
}
else if (tcor >= 83) {
return 8;
}
}
else if (mcTcor >= 99) {
if (Cs < 122) {
return 8;
}
else if (Cs >= 122) {
if (Cs < 186) {
return 4;
}
else if (Cs >= 186) {
if (Rs < 141) {
return 8;
}
else if (Rs >= 141) {
return 4;
}
}
}
}
}
}
else if (Cs >= 283) {
if (Rs < 252) {
return 4;
}
else if (Rs >= 252) {
if (Rs < 404) {
return 1;
}
else if (Rs >= 404) {
return 8;
}
}
}
}
else if (ssDCval >= 61) {
if (Rs < 335) {
if (Rs < 79) {
if (Cs < 57) {
return 2;
}
else if (Cs >= 57) {
return 4;
}
}
else if (Rs >= 79) {
if (Cs < 284) {
if (ssDCval < 80) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
else if (ssDCval >= 80) {
return 8;
}
}
else if (Cs >= 284) {
return 4;
}
}
}
else if (Rs >= 335) {
if (Rs < 386) {
if (Cs < 271) {
return 4;
}
else if (Cs >= 271) {
return 8;
}
}
else if (Rs >= 386) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
if (Rs < 271) {
return 4;
}
else if (Rs >= 271) {
if (ssDCval < 161) {
if (Rs < 396) {
if (Rs < 334) {
return 8;
}
else if (Rs >= 334) {
return 2;
}
}
else if (Rs >= 396) {
return 1;
}
}
else if (ssDCval >= 161) {
return 4;
}
}
}
}
}
}
}
else if (SC >= 792) {
if (mu_mv_mag_sq < 128) {
if (Cs < 268) {
if (ssDCval < 39) {
if (Rs < 562) {
return 8;
}
else if (Rs >= 562) {
return 4;
}
}
else if (ssDCval >= 39) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (mu_mv_mag_sq < 124) {
if (Rs < 1664) {
return 8;
}
else if (Rs >= 1664) {
return 1;
}
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 1020) {
return 8;
}
else if (Rs >= 1020) {
return 2;
}
}
}
}
}
else if (Cs >= 268) {
if (negBalance < 6) {
if (Rs < 410) {
return 1;
}
else if (Rs >= 410) {
if (RsDiff < 2) {
if (mcTcor < 100) {
return 8;
}
else if (mcTcor >= 100) {
if (Rs < 662) {
return 8;
}
else if (Rs >= 662) {
return 1;
}
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (negBalance >= 6) {
if (ssDCval < 67) {
if (ssDCval < 61) {
if (Cs < 480) {
if (CsDiff < 1) {
if (ssDCval < 60) {
if (mu_mv_mag_sq < 125) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (posBalance < 47) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 1;
}
}
else if (posBalance >= 47) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1375) {
return 8;
}
else if (Rs >= 1375) {
if (Rs < 1464) {
return 1;
}
else if (Rs >= 1464) {
return 8;
}
}
}
}
else if (ssDCval >= 60) {
if (Rs < 607) {
return 8;
}
else if (Rs >= 607) {
return 1;
}
}
}
else if (CsDiff >= 1) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (Cs >= 480) {
if (diffMVdiffVal < -1) {
if (Rs < 583) {
return 8;
}
else if (Rs >= 583) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (tcor < 98) {
if (Rs < 676) {
return 1;
}
else if (Rs >= 676) {
return 8;
}
}
else if (tcor >= 98) {
return 1;
}
}
}
}
else if (ssDCval >= 61) {
if (Cs < 444) {
if (Rs < 529) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (Rs >= 529) {
return 1;
}
}
else if (Cs >= 444) {
if (Rs < 1492) {
return 2;
}
else if (Rs >= 1492) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (ssDCval < 114) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (Rs < 1219) {
return 4;
}
else if (Rs >= 1219) {
if (Rs < 1977) {
return 1;
}
else if (Rs >= 1977) {
return 8;
}
}
}
}
else if (ssDCval >= 114) {
if (Rs < 534) {
return 1;
}
else if (Rs >= 534) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
if (Rs < 1335) {
return 1;
}
else if (Rs >= 1335) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (refDCval < 60) {
return 8;
}
else if (refDCval >= 60) {
if (ssDCval < 84) {
if (Cs < 356) {
if (Rs < 1055) {
return 1;
}
else if (Rs >= 1055) {
return 2;
}
}
else if (Cs >= 356) {
if (Rs < 1203) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
else if (Rs >= 1203) {
return 2;
}
}
}
else if (ssDCval >= 84) {
return 8;
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (tcor < 93) {
if (negBalance < 4) {
if (MVDiff < 648) {
if (posBalance < 94) {
if (posBalance < 71) {
if (ssDCval < 20) {
return 8;
}
else if (ssDCval >= 20) {
if (tcor < 76) {
return 4;
}
else if (tcor >= 76) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
return 4;
}
}
}
}
else if (posBalance >= 71) {
if (mu_mv_mag_sq < 17) {
if (Rs < 60) {
return 2;
}
else if (Rs >= 60) {
return 8;
}
}
else if (mu_mv_mag_sq >= 17) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (Rs < 46) {
return 2;
}
else if (Rs >= 46) {
return 8;
}
}
}
}
}
else if (posBalance >= 94) {
if (posBalance < 95) {
return 4;
}
else if (posBalance >= 95) {
if (Rs < 41) {
if (diffMVdiffVal < -38) {
if (Rs < 24) {
return 8;
}
else if (Rs >= 24) {
return 1;
}
}
else if (diffMVdiffVal >= -38) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (Rs >= 41) {
if (refDCval < 34) {
return 4;
}
else if (refDCval >= 34) {
if (negBalance < 3) {
if (Rs < 914) {
return 8;
}
else if (Rs >= 914) {
return 4;
}
}
else if (negBalance >= 3) {
if (Rs < 282) {
return 4;
}
else if (Rs >= 282) {
if (Rs < 521) {
return 1;
}
else if (Rs >= 521) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 648) {
if (AFD < 7) {
if (ssDCval < 11) {
return 1;
}
else if (ssDCval >= 11) {
return 4;
}
}
else if (AFD >= 7) {
if (Rs < 1068) {
return 1;
}
else if (Rs >= 1068) {
return 8;
}
}
}
}
else if (negBalance >= 4) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (mu_mv_mag_sq < 42) {
if (posBalance < 78) {
if (refDCval < 21) {
return 1;
}
else if (refDCval >= 21) {
if (Rs < 583) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Cs < 119) {
return 1;
}
else if (Cs >= 119) {
return 2;
}
}
}
else if (Rs >= 583) {
return 8;
}
}
}
else if (posBalance >= 78) {
if (ssDCval < 87) {
if (refDCval < 36) {
if (Rs < 403) {
if (Rs < 125) {
if (Rs < 29) {
return 8;
}
else if (Rs >= 29) {
return 2;
}
}
else if (Rs >= 125) {
return 1;
}
}
else if (Rs >= 403) {
if (Rs < 635) {
return 4;
}
else if (Rs >= 635) {
return 2;
}
}
}
else if (refDCval >= 36) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Rs < 664) {
if (Rs < 298) {
return 8;
}
else if (Rs >= 298) {
return 2;
}
}
else if (Rs >= 664) {
return 1;
}
}
}
}
else if (ssDCval >= 87) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (refDCval < 50) {
if (posBalance < 61) {
if (MVDiff < 187) {
if (mu_mv_mag_sq < 45) {
if (Rs < 132) {
return 1;
}
else if (Rs >= 132) {
return 4;
}
}
else if (mu_mv_mag_sq >= 45) {
if (posBalance < 26) {
if (Rs < 103) {
return 1;
}
else if (Rs >= 103) {
if (Rs < 358) {
return 8;
}
else if (Rs >= 358) {
return 1;
}
}
}
else if (posBalance >= 26) {
return 1;
}
}
}
else if (MVDiff >= 187) {
return 4;
}
}
else if (posBalance >= 61) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
else if (refDCval >= 50) {
if (mu_mv_mag_sq < 60) {
if (Rs < 948) {
return 2;
}
else if (Rs >= 948) {
return 1;
}
}
else if (mu_mv_mag_sq >= 60) {
if (Rs < 1995) {
return 8;
}
else if (Rs >= 1995) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 93) {
if (mu_mv_mag_sq < 108) {
if (SC < 930) {
if (SC < 171) {
if (posBalance < 4) {
if (ssDCval < 22) {
return 8;
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (posBalance >= 4) {
if (negBalance < 5) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (MVDiff < 445) {
if (negBalance < 2) {
return 2;
}
else if (negBalance >= 2) {
return 4;
}
}
else if (MVDiff >= 445) {
return 1;
}
}
}
else if (negBalance >= 5) {
if (MVDiff < 36) {
if (Cs < 100) {
return 1;
}
else if (Cs >= 100) {
return 8;
}
}
else if (MVDiff >= 36) {
if (Rs < 40) {
return 1;
}
else if (Rs >= 40) {
return 4;
}
}
}
}
}
else if (SC >= 171) {
if (posBalance < 57) {
if (mu_mv_mag_sq < 85) {
return 4;
}
else if (mu_mv_mag_sq >= 85) {
if (posBalance < 37) {
if (ssDCval < 38) {
if (Rs < 138) {
return 8;
}
else if (Rs >= 138) {
return 4;
}
}
else if (ssDCval >= 38) {
if (ssDCval < 76) {
if (Cs < 253) {
if (Rs < 154) {
return 1;
}
else if (Rs >= 154) {
return 8;
}
}
else if (Cs >= 253) {
return 2;
}
}
else if (ssDCval >= 76) {
if (Rs < 473) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 4;
}
}
else if (Rs >= 473) {
return 1;
}
}
}
}
else if (posBalance >= 37) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 4;
}
}
}
}
else if (posBalance >= 57) {
if (SC < 327) {
if (AFD < 5) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (SC >= 327) {
if (Rs < 209) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
if (Cs < 190) {
return 8;
}
else if (Cs >= 190) {
return 1;
}
}
}
else if (Rs >= 209) {
if (SC < 908) {
return 8;
}
else if (SC >= 908) {
return 4;
}
}
}
}
}
}
else if (SC >= 930) {
if (Rs < 1031) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
else if (Rs >= 1031) {
if (Cs < 47) {
return 8;
}
else if (Cs >= 47) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (ssDCval < 59) {
if (diffMVdiffVal < -11) {
if (diffMVdiffVal < -25) {
return 8;
}
else if (diffMVdiffVal >= -25) {
return 4;
}
}
else if (diffMVdiffVal >= -11) {
if (MVDiff < 82) {
return 8;
}
else if (MVDiff >= 82) {
if (MVDiff < 825) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (MVDiff >= 825) {
return 4;
}
}
}
}
else if (ssDCval >= 59) {
if (Rs < 819) {
if (ssDCval < 94) {
if (Rs < 57) {
return 2;
}
else if (Rs >= 57) {
if (Rs < 576) {
return 1;
}
else if (Rs >= 576) {
return 8;
}
}
}
else if (ssDCval >= 94) {
if (diffRsCsdiff < -1) {
if (ssDCval < 159) {
if (MVDiff < 60) {
if (diffAFD < -4) {
if (ssDCval < 130) {
return 2;
}
else if (ssDCval >= 130) {
return 4;
}
}
else if (diffAFD >= -4) {
return 4;
}
}
else if (MVDiff >= 60) {
return 4;
}
}
else if (ssDCval >= 159) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
if (Rs < 157) {
return 4;
}
else if (Rs >= 157) {
return 8;
}
}
}
}
else if (Rs >= 819) {
if (diffAFD < -8) {
return 2;
}
else if (diffAFD >= -8) {
if (Rs < 1612) {
return 1;
}
else if (Rs >= 1612) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (posBalance < 27) {
if (posBalance < 8) {
if (posBalance < 2) {
if (mu_mv_mag_sq < 28) {
if (TSCindex < 2) {
if (refDCval < 24) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 8;
}
}
else if (refDCval >= 24) {
return 4;
}
}
else if (TSCindex >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 28) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (posBalance >= 2) {
return 8;
}
}
else if (posBalance >= 8) {
if (negBalance < 18) {
if (ssDCval < 27) {
if (Rs < 114) {
return 1;
}
else if (Rs >= 114) {
return 8;
}
}
else if (ssDCval >= 27) {
if (SC < 191) {
return 4;
}
else if (SC >= 191) {
if (Cs < 109) {
return 8;
}
else if (Cs >= 109) {
if (ssDCval < 30) {
return 8;
}
else if (ssDCval >= 30) {
return 4;
}
}
}
}
}
else if (negBalance >= 18) {
if (negBalance < 58) {
if (ssDCval < 31) {
if (mu_mv_mag_sq < 95) {
if (negBalance < 19) {
if (Rs < 89) {
return 4;
}
else if (Rs >= 89) {
return 1;
}
}
else if (negBalance >= 19) {
if (Rs < 1) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Rs >= 1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (MVDiff < 15) {
if (Rs < 13) {
return 4;
}
else if (Rs >= 13) {
return 1;
}
}
else if (MVDiff >= 15) {
return 4;
}
}
}
else if (ssDCval >= 31) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (Rs < 382) {
return 1;
}
else if (Rs >= 382) {
return 2;
}
}
}
}
else if (negBalance >= 58) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 27) {
if (mcTcor < 88) {
if (ssDCval < 23) {
if (Cs < 254) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
return 1;
}
}
else if (Cs >= 254) {
return 4;
}
}
else if (ssDCval >= 23) {
if (tcor < 58) {
if (diffTSC < 1) {
if (Cs < 317) {
if (negBalance < 41) {
if (diffMVdiffVal < -16) {
if (posBalance < 51) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (diffMVdiffVal >= -16) {
return 1;
}
}
else if (negBalance >= 41) {
if (Rs < 870) {
return 8;
}
else if (Rs >= 870) {
return 1;
}
}
}
else if (Cs >= 317) {
if (Rs < 548) {
return 2;
}
else if (Rs >= 548) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (Cs < 389) {
if (diffMVdiffVal < -8) {
if (Rs < 265) {
if (Cs < 204) {
return 8;
}
else if (Cs >= 204) {
return 4;
}
}
else if (Rs >= 265) {
return 1;
}
}
else if (diffMVdiffVal >= -8) {
if (negBalance < 46) {
return 1;
}
else if (negBalance >= 46) {
if (Rs < 328) {
return 8;
}
else if (Rs >= 328) {
return 1;
}
}
}
}
else if (Cs >= 389) {
if (Rs < 565) {
return 2;
}
else if (Rs >= 565) {
return 8;
}
}
}
}
else if (tcor >= 58) {
if (Rs < 330) {
if (ssDCval < 28) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (diffMVdiffVal < 2) {
if (Rs < 200) {
return 8;
}
else if (Rs >= 200) {
return 4;
}
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
}
else if (ssDCval >= 28) {
if (tcor < 83) {
if (posBalance < 53) {
if (AFD < 7) {
if (posBalance < 36) {
if (Rs < 137) {
return 2;
}
else if (Rs >= 137) {
return 8;
}
}
else if (posBalance >= 36) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 1;
}
}
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (posBalance >= 53) {
if (Cs < 184) {
if (Rs < 43) {
if (Rs < 8) {
return 4;
}
else if (Rs >= 8) {
return 1;
}
}
else if (Rs >= 43) {
return 8;
}
}
else if (Cs >= 184) {
if (Rs < 304) {
return 2;
}
else if (Rs >= 304) {
return 8;
}
}
}
}
else if (tcor >= 83) {
if (Cs < 276) {
return 8;
}
else if (Cs >= 276) {
return 1;
}
}
}
}
else if (Rs >= 330) {
if (TSC < 3) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (Rs < 666) {
return 2;
}
else if (Rs >= 666) {
return 1;
}
}
}
}
}
else if (TSC >= 3) {
return 1;
}
}
}
}
}
else if (mcTcor >= 88) {
if (Rs < 84) {
if (posBalance < 68) {
return 1;
}
else if (posBalance >= 68) {
if (Rs < 48) {
return 2;
}
else if (Rs >= 48) {
return 1;
}
}
}
else if (Rs >= 84) {
if (posBalance < 47) {
if (refDCval < 30) {
if (MVDiff < 23) {
if (Cs < 46) {
if (AFD < 4) {
if (Cs < 24) {
return 4;
}
else if (Cs >= 24) {
return 1;
}
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 46) {
if (diffMVdiffVal < -10) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 4;
}
}
else if (diffMVdiffVal >= -10) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 4;
}
}
}
}
else if (MVDiff >= 23) {
if (Cs < 238) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (Cs < 60) {
return 4;
}
else if (Cs >= 60) {
return 8;
}
}
}
}
else if (Cs >= 238) {
if (negBalance < 17) {
return 4;
}
else if (negBalance >= 17) {
if (Rs < 434) {
return 1;
}
else if (Rs >= 434) {
return 8;
}
}
}
}
}
else if (refDCval >= 30) {
if (Rs < 110) {
if (SC < 187) {
if (AFD < 2) {
if (Rs < 86) {
return 8;
}
else if (Rs >= 86) {
return 4;
}
}
else if (AFD >= 2) {
return 1;
}
}
else if (SC >= 187) {
return 8;
}
}
else if (Rs >= 110) {
if (Cs < 253) {
if (Cs < 149) {
if (mcTcor < 96) {
if (negBalance < 22) {
if (Rs < 155) {
return 8;
}
else if (Rs >= 155) {
return 4;
}
}
else if (negBalance >= 22) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (mcTcor < 93) {
if (Cs < 142) {
return 1;
}
else if (Cs >= 142) {
return 8;
}
}
else if (mcTcor >= 93) {
if (Rs < 277) {
return 4;
}
else if (Rs >= 277) {
return 1;
}
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 259) {
return 1;
}
else if (Rs >= 259) {
if (Rs < 297) {
return 2;
}
else if (Rs >= 297) {
return 1;
}
}
}
}
else if (Cs >= 149) {
if (AFD < 4) {
if (SC < 517) {
if (Cs < 181) {
return 8;
}
else if (Cs >= 181) {
if (MVDiff < 9) {
if (Rs < 288) {
return 4;
}
else if (Rs >= 288) {
return 8;
}
}
else if (MVDiff >= 9) {
return 1;
}
}
}
else if (SC >= 517) {
if (mcTcor < 92) {
if (Rs < 497) {
return 8;
}
else if (Rs >= 497) {
return 1;
}
}
else if (mcTcor >= 92) {
return 4;
}
}
}
else if (AFD >= 4) {
if (posBalance < 37) {
if (ssDCval < 31) {
if (Rs < 330) {
return 1;
}
else if (Rs >= 330) {
return 4;
}
}
else if (ssDCval >= 31) {
return 4;
}
}
else if (posBalance >= 37) {
if (ssDCval < 35) {
return 1;
}
else if (ssDCval >= 35) {
return 4;
}
}
}
}
}
else if (Cs >= 253) {
if (MVDiff < 22) {
return 1;
}
else if (MVDiff >= 22) {
if (Rs < 652) {
return 4;
}
else if (Rs >= 652) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 47) {
if (SC < 344) {
if (tcor < 74) {
if (Cs < 109) {
return 1;
}
else if (Cs >= 109) {
return 8;
}
}
else if (tcor >= 74) {
return 1;
}
}
else if (SC >= 344) {
if (MVDiff < 45) {
return 1;
}
else if (MVDiff >= 45) {
if (Rs < 198) {
return 2;
}
else if (Rs >= 198) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 282) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 50) {
if (mcTcor < 90) {
if (MVDiff < 42) {
if (tcor < 67) {
if (gchDC < 1) {
if (Cs < 242) {
return 4;
}
else if (Cs >= 242) {
return 2;
}
}
else if (gchDC >= 1) {
if (RsDiff < 2) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Cs < 191) {
return 1;
}
else if (Cs >= 191) {
if (Cs < 282) {
return 4;
}
else if (Cs >= 282) {
return 1;
}
}
}
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (tcor >= 67) {
if (posBalance < 54) {
if (TSC < 4) {
if (Rs < 142) {
return 8;
}
else if (Rs >= 142) {
return 2;
}
}
else if (TSC >= 4) {
return 8;
}
}
else if (posBalance >= 54) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (Cs < 272) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 2;
}
}
else if (Cs >= 272) {
return 2;
}
}
}
}
}
else if (MVDiff >= 42) {
if (tcor < 68) {
if (mu_mv_mag_sq < 22) {
if (posBalance < 70) {
if (AFD < 12) {
if (refDCval < 35) {
return 8;
}
else if (refDCval >= 35) {
if (MVDiff < 45) {
return 1;
}
else if (MVDiff >= 45) {
return 4;
}
}
}
else if (AFD >= 12) {
if (Cs < 126) {
if (tcor < 62) {
if (AFD < 35) {
return 1;
}
else if (AFD >= 35) {
return 8;
}
}
else if (tcor >= 62) {
return 4;
}
}
else if (Cs >= 126) {
return 8;
}
}
}
else if (posBalance >= 70) {
return 2;
}
}
else if (mu_mv_mag_sq >= 22) {
if (mu_mv_mag_sq < 36) {
return 1;
}
else if (mu_mv_mag_sq >= 36) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
return 8;
}
}
}
}
else if (tcor >= 68) {
if (diffMVdiffVal < 6) {
if (MVDiff < 88) {
if (Cs < 30) {
return 2;
}
else if (Cs >= 30) {
return 1;
}
}
else if (MVDiff >= 88) {
return 2;
}
}
else if (diffMVdiffVal >= 6) {
if (MVDiff < 77) {
if (Rs < 196) {
return 2;
}
else if (Rs >= 196) {
return 8;
}
}
else if (MVDiff >= 77) {
if (posBalance < 3) {
return 8;
}
else if (posBalance >= 3) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (posBalance < 8) {
if (refDCval < 36) {
if (Cs < 2) {
return 8;
}
else if (Cs >= 2) {
return 4;
}
}
else if (refDCval >= 36) {
return 8;
}
}
else if (posBalance >= 8) {
if (tcor < 91) {
if (posBalance < 67) {
if (Cs < 1) {
if (Rs < 29) {
return 8;
}
else if (Rs >= 29) {
return 2;
}
}
else if (Cs >= 1) {
if (mcTcor < 93) {
if (refDCval < 42) {
if (TSC < 3) {
if (Rs < 51) {
return 4;
}
else if (Rs >= 51) {
return 1;
}
}
else if (TSC >= 3) {
if (RsCsDiff < 4) {
if (Rs < 137) {
return 8;
}
else if (Rs >= 137) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 2;
}
}
}
else if (RsCsDiff >= 4) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
if (Cs < 239) {
return 1;
}
else if (Cs >= 239) {
return 4;
}
}
}
}
}
else if (refDCval >= 42) {
if (Rs < 278) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (Rs < 180) {
return 2;
}
else if (Rs >= 180) {
return 1;
}
}
}
else if (Rs >= 278) {
return 4;
}
}
}
else if (mcTcor >= 93) {
if (Cs < 54) {
if (Rs < 128) {
if (ssDCval < 38) {
return 1;
}
else if (ssDCval >= 38) {
return 2;
}
}
else if (Rs >= 128) {
return 1;
}
}
else if (Cs >= 54) {
if (SC < 279) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
if (ssDCval < 47) {
if (Rs < 46) {
if (Rs < 18) {
return 4;
}
else if (Rs >= 18) {
return 1;
}
}
else if (Rs >= 46) {
return 4;
}
}
else if (ssDCval >= 47) {
if (Rs < 85) {
return 1;
}
else if (Rs >= 85) {
return 2;
}
}
}
}
else if (SC >= 279) {
if (mu_mv_mag_sq < 4) {
if (RsCsDiff < 3) {
if (ssDCval < 44) {
return 2;
}
else if (ssDCval >= 44) {
if (Rs < 215) {
return 1;
}
else if (Rs >= 215) {
return 4;
}
}
}
else if (RsCsDiff >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 4) {
if (Cs < 154) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
return 1;
}
}
else if (Cs >= 154) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (Cs < 439) {
return 1;
}
else if (Cs >= 439) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (posBalance >= 67) {
return 2;
}
}
else if (tcor >= 91) {
if (MVDiff < 105) {
if (SC < 707) {
if (ssDCval < 41) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
if (Rs < 141) {
if (Rs < 34) {
return 8;
}
else if (Rs >= 34) {
return 1;
}
}
else if (Rs >= 141) {
return 2;
}
}
}
else if (ssDCval >= 41) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Rs < 172) {
return 1;
}
else if (Rs >= 172) {
return 4;
}
}
}
}
else if (SC >= 707) {
return 8;
}
}
else if (MVDiff >= 105) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (Cs < 178) {
if (gchDC < 1) {
if (posBalance < 40) {
if (ssDCval < 43) {
if (ssDCval < 38) {
if (Rs < 124) {
return 4;
}
else if (Rs >= 124) {
if (Rs < 142) {
if (Rs < 128) {
return 2;
}
else if (Rs >= 128) {
return 8;
}
}
else if (Rs >= 142) {
return 1;
}
}
}
else if (ssDCval >= 38) {
return 1;
}
}
else if (ssDCval >= 43) {
return 8;
}
}
else if (posBalance >= 40) {
if (Rs < 246) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 114) {
return 1;
}
else if (Rs >= 114) {
return 8;
}
}
}
}
else if (Rs >= 246) {
if (SC < 426) {
if (Rs < 264) {
return 4;
}
else if (Rs >= 264) {
return 1;
}
}
else if (SC >= 426) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
if (posBalance < 23) {
if (mcTcor < 105) {
if (refDCval < 36) {
if (Rs < 108) {
return 1;
}
else if (Rs >= 108) {
return 4;
}
}
else if (refDCval >= 36) {
if (Rs < 8) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
else if (Rs >= 8) {
return 8;
}
}
}
else if (mcTcor >= 105) {
return 4;
}
}
else if (posBalance >= 23) {
if (mu_mv_mag_sq < 90) {
if (Cs < 106) {
if (posBalance < 50) {
if (SC < 97) {
return 2;
}
else if (SC >= 97) {
if (Rs < 247) {
return 1;
}
else if (Rs >= 247) {
return 2;
}
}
}
else if (posBalance >= 50) {
if (Rs < 140) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (Rs < 77) {
return 2;
}
else if (Rs >= 77) {
return 4;
}
}
}
else if (Rs >= 140) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (Cs >= 106) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
if (gchDC < 5) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 4;
}
}
else if (gchDC >= 5) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (ssDCval < 44) {
if (MVDiff < 70) {
if (Rs < 93) {
return 4;
}
else if (Rs >= 93) {
if (diffMVdiffVal < 2) {
if (ssDCval < 38) {
return 1;
}
else if (ssDCval >= 38) {
return 4;
}
}
else if (diffMVdiffVal >= 2) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (MVDiff >= 70) {
return 8;
}
}
else if (ssDCval >= 44) {
if (negBalance < 43) {
if (Cs < 35) {
return 4;
}
else if (Cs >= 35) {
return 8;
}
}
else if (negBalance >= 43) {
if (Rs < 26) {
return 2;
}
else if (Rs >= 26) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 178) {
if (ssDCval < 40) {
if (Cs < 311) {
if (negBalance < 26) {
if (mcTcor < 94) {
if (MVDiff < 5) {
if (Rs < 208) {
return 1;
}
else if (Rs >= 208) {
return 4;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (mcTcor >= 94) {
if (posBalance < 27) {
if (Rs < 131) {
return 1;
}
else if (Rs >= 131) {
return 4;
}
}
else if (posBalance >= 27) {
if (Rs < 238) {
if (Cs < 294) {
if (negBalance < 17) {
if (Cs < 233) {
return 4;
}
else if (Cs >= 233) {
return 8;
}
}
else if (negBalance >= 17) {
return 8;
}
}
else if (Cs >= 294) {
if (Rs < 214) {
return 1;
}
else if (Rs >= 214) {
return 8;
}
}
}
else if (Rs >= 238) {
if (Cs < 269) {
return 4;
}
else if (Cs >= 269) {
return 1;
}
}
}
}
}
else if (negBalance >= 26) {
if (Rs < 270) {
return 8;
}
else if (Rs >= 270) {
return 1;
}
}
}
else if (Cs >= 311) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
return 1;
}
}
}
else if (ssDCval >= 40) {
if (diffRsCsdiff < -1) {
if (MVDiff < 13) {
if (Cs < 427) {
return 1;
}
else if (Cs >= 427) {
return 2;
}
}
else if (MVDiff >= 13) {
if (Cs < 189) {
return 8;
}
else if (Cs >= 189) {
return 4;
}
}
}
else if (diffRsCsdiff >= -1) {
if (mu_mv_mag_sq < 58) {
if (Rs < 156) {
return 1;
}
else if (Rs >= 156) {
return 4;
}
}
else if (mu_mv_mag_sq >= 58) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 50) {
if (ssDCval < 70) {
if (posBalance < 33) {
if (tcor < 98) {
if (refDCval < 60) {
if (Cs < 235) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
return 1;
}
}
else if (Cs >= 235) {
if (Cs < 367) {
return 1;
}
else if (Cs >= 367) {
return 4;
}
}
}
else if (refDCval >= 60) {
if (MVDiff < 86) {
if (SC < 388) {
return 1;
}
else if (SC >= 388) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (MVDiff >= 86) {
if (Rs < 97) {
return 4;
}
else if (Rs >= 97) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (SC < 79) {
return 2;
}
else if (SC >= 79) {
if (SC < 430) {
if (ssDCval < 61) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (Rs < 130) {
return 4;
}
else if (Rs >= 130) {
return 8;
}
}
}
else if (ssDCval >= 61) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (SC >= 430) {
if (negBalance < 88) {
if (Cs < 489) {
return 4;
}
else if (Cs >= 489) {
return 1;
}
}
else if (negBalance >= 88) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 33) {
if (TSCindex < 2) {
if (ssDCval < 65) {
if (MVDiff < 7) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (Cs < 189) {
if (MVDiff < 3) {
if (Rs < 146) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Rs >= 146) {
return 1;
}
}
else if (MVDiff >= 3) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 1;
}
}
}
else if (Cs >= 189) {
if (Rs < 245) {
return 4;
}
else if (Rs >= 245) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
if (Cs < 316) {
return 4;
}
else if (Cs >= 316) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 89) {
return 1;
}
else if (Rs >= 89) {
return 8;
}
}
}
else if (ssDCval >= 65) {
if (Cs < 240) {
return 2;
}
else if (Cs >= 240) {
if (Cs < 378) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (Cs >= 378) {
return 1;
}
}
}
}
else if (TSCindex >= 2) {
if (TSC < 4) {
if (RsCsDiff < 1) {
if (Cs < 192) {
if (MVDiff < 2) {
if (refDCval < 64) {
if (mcTcor < 96) {
if (Rs < 75) {
return 2;
}
else if (Rs >= 75) {
if (Rs < 107) {
return 1;
}
else if (Rs >= 107) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Rs < 132) {
return 2;
}
else if (Rs >= 132) {
return 1;
}
}
}
}
else if (refDCval >= 64) {
if (Rs < 90) {
return 1;
}
else if (Rs >= 90) {
return 2;
}
}
}
else if (MVDiff >= 2) {
if (mcTcor < 95) {
if (AFD < 5) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
return 2;
}
}
}
else if (AFD >= 5) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (mcTcor >= 95) {
if (negBalance < 19) {
if (MVDiff < 25) {
return 2;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (negBalance >= 19) {
if (Rs < 216) {
if (diffMVdiffVal < -7) {
if (Rs < 97) {
return 1;
}
else if (Rs >= 97) {
return 4;
}
}
else if (diffMVdiffVal >= -7) {
if (MVDiff < 48) {
if (tcor < 99) {
if (Cs < 90) {
return 1;
}
else if (Cs >= 90) {
if (MVDiff < 16) {
if (mu_mv_mag_sq < 50) {
return 1;
}
else if (mu_mv_mag_sq >= 50) {
if (tcor < 94) {
return 4;
}
else if (tcor >= 94) {
if (negBalance < 24) {
if (mu_mv_mag_sq < 104) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
return 1;
}
}
else if (mu_mv_mag_sq >= 104) {
return 4;
}
}
else if (negBalance >= 24) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 172) {
return 4;
}
else if (Rs >= 172) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 16) {
if (Cs < 152) {
return 4;
}
else if (Cs >= 152) {
return 1;
}
}
}
}
else if (tcor >= 99) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (MVDiff >= 48) {
return 4;
}
}
}
else if (Rs >= 216) {
if (Cs < 185) {
return 4;
}
else if (Cs >= 185) {
return 1;
}
}
}
}
}
}
else if (Cs >= 192) {
if (negBalance < 30) {
if (ssDCval < 60) {
if (MVDiff < 1) {
if (mcTcor < 98) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 4;
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 22) {
return 4;
}
else if (negBalance >= 22) {
return 1;
}
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 4;
}
}
}
else if (ssDCval >= 60) {
if (Rs < 158) {
if (Rs < 126) {
if (Rs < 103) {
if (Rs < 88) {
return 4;
}
else if (Rs >= 88) {
return 1;
}
}
else if (Rs >= 103) {
return 2;
}
}
else if (Rs >= 126) {
if (Cs < 313) {
if (mu_mv_mag_sq < 98) {
if (Cs < 255) {
return 4;
}
else if (Cs >= 255) {
return 1;
}
}
else if (mu_mv_mag_sq >= 98) {
return 4;
}
}
else if (Cs >= 313) {
if (Rs < 146) {
return 2;
}
else if (Rs >= 146) {
if (Rs < 155) {
return 1;
}
else if (Rs >= 155) {
return 4;
}
}
}
}
}
else if (Rs >= 158) {
if (ssDCval < 69) {
if (negBalance < 26) {
if (Rs < 177) {
if (TSC < 2) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
return 4;
}
}
else if (TSC >= 2) {
return 1;
}
}
else if (Rs >= 177) {
return 4;
}
}
else if (negBalance >= 26) {
if (mu_mv_mag_sq < 125) {
if (SC < 527) {
if (Cs < 335) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (Rs < 234) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (ssDCval < 65) {
if (mu_mv_mag_sq < 118) {
return 4;
}
else if (mu_mv_mag_sq >= 118) {
return 1;
}
}
else if (ssDCval >= 65) {
return 1;
}
}
}
else if (Rs >= 234) {
return 1;
}
}
}
else if (Cs >= 335) {
return 1;
}
}
else if (SC >= 527) {
return 4;
}
}
else if (mu_mv_mag_sq >= 125) {
return 1;
}
}
}
else if (ssDCval >= 69) {
if (Cs < 345) {
return 4;
}
else if (Cs >= 345) {
return 1;
}
}
}
}
}
else if (negBalance >= 30) {
if (tcor < 97) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 87) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
if (Cs < 235) {
return 1;
}
else if (Cs >= 235) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (posBalance < 49) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (posBalance >= 49) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
else if (negBalance >= 37) {
if (Cs < 250) {
return 2;
}
else if (Cs >= 250) {
return 1;
}
}
}
else if (tcor >= 97) {
if (posBalance < 47) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
if (posBalance < 45) {
if (Rs < 236) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
if (Cs < 245) {
return 1;
}
else if (Cs >= 245) {
return 4;
}
}
}
else if (Rs >= 236) {
return 1;
}
}
else if (posBalance >= 45) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (ssDCval < 63) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (Cs < 220) {
return 1;
}
else if (Cs >= 220) {
return 4;
}
}
}
else if (ssDCval >= 63) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (MVDiff < 2) {
if (Cs < 404) {
if (mu_mv_mag_sq < 116) {
if (Rs < 202) {
if (Cs < 376) {
if (Rs < 175) {
return 1;
}
else if (Rs >= 175) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (Cs >= 376) {
return 1;
}
}
else if (Rs >= 202) {
return 4;
}
}
else if (mu_mv_mag_sq >= 116) {
return 1;
}
}
else if (Cs >= 404) {
return 1;
}
}
else if (MVDiff >= 2) {
if (Rs < 37) {
return 1;
}
else if (Rs >= 37) {
if (RsDiff < 1) {
if (negBalance < 51) {
if (MVDiff < 45) {
if (AFD < 5) {
if (refDCval < 64) {
if (ssDCval < 53) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 4;
}
}
}
else if (ssDCval >= 53) {
if (Rs < 227) {
return 4;
}
else if (Rs >= 227) {
if (Cs < 306) {
return 1;
}
else if (Cs >= 306) {
return 4;
}
}
}
}
else if (refDCval >= 64) {
if (Cs < 265) {
return 1;
}
else if (Cs >= 265) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 4;
}
}
}
}
else if (AFD >= 5) {
if (refDCval < 56) {
if (Rs < 96) {
return 1;
}
else if (Rs >= 96) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 2;
}
}
}
else if (refDCval >= 56) {
return 4;
}
}
}
else if (MVDiff >= 45) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (negBalance >= 51) {
if (Cs < 198) {
return 4;
}
else if (Cs >= 198) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (RsDiff >= 1) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (mu_mv_mag_sq < 18) {
if (mu_mv_mag_sq < 3) {
if (negBalance < 40) {
if (negBalance < 37) {
if (Rs < 273) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (Cs < 257) {
return 1;
}
else if (Cs >= 257) {
return 4;
}
}
}
else if (Rs >= 273) {
return 2;
}
}
else if (negBalance >= 37) {
return 2;
}
}
else if (negBalance >= 40) {
if (Rs < 266) {
return 4;
}
else if (Rs >= 266) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 610) {
if (mcTcor < 99) {
if (MVDiff < 7) {
if (Cs < 219) {
return 1;
}
else if (Cs >= 219) {
return 4;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (mcTcor >= 99) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (SC >= 610) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 18) {
if (posBalance < 37) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
return 8;
}
}
}
else if (posBalance >= 37) {
if (ssDCval < 64) {
if (mu_mv_mag_sq < 51) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (mu_mv_mag_sq >= 51) {
return 4;
}
}
else if (ssDCval >= 64) {
if (gchDC < 2) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (Rs < 234) {
return 8;
}
else if (Rs >= 234) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 43) {
if (tcor < 75) {
if (MVDiff < 59) {
if (AFD < 15) {
if (AFD < 10) {
if (Rs < 222) {
return 1;
}
else if (Rs >= 222) {
return 4;
}
}
else if (AFD >= 10) {
return 4;
}
}
else if (AFD >= 15) {
if (RsCsDiff < 7) {
return 2;
}
else if (RsCsDiff >= 7) {
if (Rs < 246) {
return 1;
}
else if (Rs >= 246) {
return 8;
}
}
}
}
else if (MVDiff >= 59) {
if (diffMVdiffVal < 19) {
if (Rs < 263) {
if (AFD < 23) {
if (ssDCval < 65) {
if (gchDC < 5) {
return 1;
}
else if (gchDC >= 5) {
if (Rs < 88) {
return 1;
}
else if (Rs >= 88) {
return 2;
}
}
}
else if (ssDCval >= 65) {
return 2;
}
}
else if (AFD >= 23) {
if (Cs < 200) {
return 1;
}
else if (Cs >= 200) {
return 2;
}
}
}
else if (Rs >= 263) {
return 4;
}
}
else if (diffMVdiffVal >= 19) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (AFD < 28) {
return 8;
}
else if (AFD >= 28) {
return 1;
}
}
}
}
}
else if (tcor >= 75) {
if (MVDiff < 60) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 24) {
if (Cs < 494) {
return 1;
}
else if (Cs >= 494) {
return 2;
}
}
else if (mu_mv_mag_sq >= 24) {
if (AFD < 10) {
if (Rs < 129) {
return 8;
}
else if (Rs >= 129) {
return 2;
}
}
else if (AFD >= 10) {
if (Rs < 248) {
return 1;
}
else if (Rs >= 248) {
return 4;
}
}
}
}
else if (mcTcor >= 95) {
if (AFD < 10) {
if (Cs < 313) {
return 2;
}
else if (Cs >= 313) {
return 8;
}
}
else if (AFD >= 10) {
if (AFD < 13) {
if (Rs < 266) {
return 1;
}
else if (Rs >= 266) {
return 8;
}
}
else if (AFD >= 13) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (MVDiff >= 60) {
if (diffAFD < -4) {
if (Rs < 156) {
return 1;
}
else if (Rs >= 156) {
return 8;
}
}
else if (diffAFD >= -4) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (diffRsCsdiff < 1) {
if (diffMVdiffVal < -2) {
if (Rs < 239) {
if (refDCval < 68) {
return 8;
}
else if (refDCval >= 68) {
return 4;
}
}
else if (Rs >= 239) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
if (posBalance < 67) {
if (gchDC < 6) {
if (tcor < 97) {
if (SC < 158) {
return 1;
}
else if (SC >= 158) {
if (Rs < 243) {
return 4;
}
else if (Rs >= 243) {
return 1;
}
}
}
else if (tcor >= 97) {
if (Rs < 123) {
return 8;
}
else if (Rs >= 123) {
if (Rs < 221) {
return 2;
}
else if (Rs >= 221) {
return 1;
}
}
}
}
else if (gchDC >= 6) {
if (Rs < 109) {
if (Rs < 49) {
return 8;
}
else if (Rs >= 49) {
return 4;
}
}
else if (Rs >= 109) {
return 8;
}
}
}
else if (posBalance >= 67) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 1) {
if (diffMVdiffVal < -8) {
if (Rs < 97) {
return 1;
}
else if (Rs >= 97) {
return 4;
}
}
else if (diffMVdiffVal >= -8) {
if (mu_mv_mag_sq < 69) {
return 1;
}
else if (mu_mv_mag_sq >= 69) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 70) {
if (TSC < 2) {
if (refDCval < 81) {
if (MVDiff < 6) {
if (Rs < 118) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
return 4;
}
}
else if (Rs >= 118) {
if (Cs < 253) {
if (Cs < 121) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Cs < 111) {
return 1;
}
else if (Cs >= 111) {
return 2;
}
}
}
else if (Cs >= 121) {
if (negBalance < 22) {
if (RsDiff < 1) {
if (mcTcor < 98) {
if (Rs < 188) {
return 8;
}
else if (Rs >= 188) {
return 2;
}
}
else if (mcTcor >= 98) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
if (Cs < 188) {
return 8;
}
else if (Cs >= 188) {
return 2;
}
}
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (negBalance >= 22) {
if (SC < 407) {
if (Cs < 130) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
return 8;
}
}
else if (Cs >= 130) {
return 2;
}
}
else if (SC >= 407) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
return 2;
}
}
}
}
}
else if (Cs >= 253) {
if (Rs < 184) {
return 4;
}
else if (Rs >= 184) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 74) {
return 1;
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 252) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Rs < 160) {
if (SC < 223) {
return 2;
}
else if (SC >= 223) {
return 1;
}
}
else if (Rs >= 160) {
return 2;
}
}
}
else if (Cs >= 252) {
return 1;
}
}
}
}
else if (refDCval >= 81) {
if (Cs < 142) {
if (refDCval < 110) {
if (Cs < 117) {
if (ssDCval < 106) {
if (refDCval < 82) {
if (Rs < 109) {
return 1;
}
else if (Rs >= 109) {
return 8;
}
}
else if (refDCval >= 82) {
return 1;
}
}
else if (ssDCval >= 106) {
if (Rs < 47) {
return 1;
}
else if (Rs >= 47) {
return 2;
}
}
}
else if (Cs >= 117) {
if (Rs < 89) {
if (Rs < 21) {
return 2;
}
else if (Rs >= 21) {
return 4;
}
}
else if (Rs >= 89) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 1;
}
}
}
}
else if (refDCval >= 110) {
if (Rs < 120) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
else if (Rs >= 120) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
}
}
else if (Cs >= 142) {
if (ssDCval < 136) {
if (Cs < 185) {
if (ssDCval < 96) {
if (diffTSC < -1) {
if (Rs < 192) {
return 1;
}
else if (Rs >= 192) {
return 4;
}
}
else if (diffTSC >= -1) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
if (Rs < 54) {
return 2;
}
else if (Rs >= 54) {
return 1;
}
}
}
}
else if (ssDCval >= 96) {
if (Rs < 195) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Rs >= 195) {
return 1;
}
}
}
else if (Cs >= 185) {
if (negBalance < 29) {
if (diffTSC < -1) {
if (Cs < 208) {
return 1;
}
else if (Cs >= 208) {
return 4;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (negBalance >= 29) {
if (tcor < 91) {
if (Rs < 119) {
return 2;
}
else if (Rs >= 119) {
return 8;
}
}
else if (tcor >= 91) {
if (Cs < 215) {
return 1;
}
else if (Cs >= 215) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 136) {
return 2;
}
}
}
}
else if (TSC >= 2) {
if (ssDCval < 86) {
if (Cs < 234) {
if (MVDiff < 4) {
if (Cs < 163) {
if (Cs < 124) {
return 8;
}
else if (Cs >= 124) {
if (posBalance < 49) {
if (mcTcor < 98) {
if (Cs < 142) {
return 8;
}
else if (Cs >= 142) {
return 2;
}
}
else if (mcTcor >= 98) {
return 2;
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
else if (Cs >= 163) {
if (ssDCval < 76) {
if (SC < 377) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 1;
}
}
else if (SC >= 377) {
return 1;
}
}
else if (ssDCval >= 76) {
if (Cs < 189) {
if (Rs < 220) {
return 2;
}
else if (Rs >= 220) {
return 1;
}
}
else if (Cs >= 189) {
return 2;
}
}
}
}
else if (MVDiff >= 4) {
if (MVDiff < 49) {
if (Cs < 141) {
if (mu_mv_mag_sq < 41) {
if (Cs < 96) {
if (tcor < 102) {
return 1;
}
else if (tcor >= 102) {
if (Rs < 223) {
return 8;
}
else if (Rs >= 223) {
return 1;
}
}
}
else if (Cs >= 96) {
if (gchDC < 3) {
if (refDCval < 73) {
return 4;
}
else if (refDCval >= 73) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Rs < 71) {
return 1;
}
else if (Rs >= 71) {
return 4;
}
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
if (AFD < 6) {
if (TSC < 3) {
if (Cs < 82) {
return 8;
}
else if (Cs >= 82) {
if (Rs < 121) {
return 2;
}
else if (Rs >= 121) {
return 1;
}
}
}
else if (TSC >= 3) {
return 1;
}
}
else if (AFD >= 6) {
return 8;
}
}
}
}
else if (Cs >= 141) {
if (Rs < 129) {
if (MVDiff < 12) {
if (SC < 290) {
return 2;
}
else if (SC >= 290) {
if (Cs < 188) {
return 1;
}
else if (Cs >= 188) {
return 4;
}
}
}
else if (MVDiff >= 12) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (Rs >= 129) {
if (SC < 462) {
if (mu_mv_mag_sq < 116) {
return 1;
}
else if (mu_mv_mag_sq >= 116) {
return 2;
}
}
else if (SC >= 462) {
return 2;
}
}
}
}
else if (MVDiff >= 49) {
if (MVDiff < 219) {
if (diffMVdiffVal < 42) {
if (mu_mv_mag_sq < 40) {
if (MVDiff < 96) {
return 2;
}
else if (MVDiff >= 96) {
if (diffRsCsdiff < -1) {
return 2;
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 42) {
if (Cs < 170) {
return 8;
}
else if (Cs >= 170) {
return 4;
}
}
}
else if (MVDiff >= 219) {
if (mcTcor < 69) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
return 4;
}
}
else if (mcTcor >= 69) {
return 1;
}
}
}
}
}
else if (Cs >= 234) {
if (SC < 486) {
if (refDCval < 75) {
if (Rs < 136) {
if (Cs < 288) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (Rs < 79) {
return 2;
}
else if (Rs >= 79) {
return 4;
}
}
}
else if (Cs >= 288) {
if (Rs < 82) {
return 2;
}
else if (Rs >= 82) {
return 4;
}
}
}
else if (Rs >= 136) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
return 4;
}
}
}
else if (refDCval >= 75) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (tcor < 93) {
if (MVDiff < 8) {
if (Rs < 158) {
return 1;
}
else if (Rs >= 158) {
return 2;
}
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (tcor >= 93) {
return 2;
}
}
}
}
}
else if (SC >= 486) {
if (negBalance < 42) {
if (mcTcor < 97) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 80) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 181) {
return 1;
}
else if (Rs >= 181) {
return 4;
}
}
}
else if (ssDCval >= 80) {
return 4;
}
}
}
else if (negBalance >= 42) {
if (RsCsDiff < 4) {
if (Rs < 207) {
return 4;
}
else if (Rs >= 207) {
if (Cs < 297) {
return 2;
}
else if (Cs >= 297) {
return 8;
}
}
}
else if (RsCsDiff >= 4) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (mcTcor < 71) {
if (Rs < 230) {
return 8;
}
else if (Rs >= 230) {
return 4;
}
}
else if (mcTcor >= 71) {
if (diffTSC < 2) {
return 2;
}
else if (diffTSC >= 2) {
if (SC < 626) {
return 1;
}
else if (SC >= 626) {
return 2;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (mcTcor < 72) {
if (gchDC < 7) {
return 4;
}
else if (gchDC >= 7) {
if (ssDCval < 179) {
if (ssDCval < 94) {
return 4;
}
else if (ssDCval >= 94) {
return 1;
}
}
else if (ssDCval >= 179) {
return 2;
}
}
}
else if (mcTcor >= 72) {
if (Rs < 142) {
if (ssDCval < 100) {
if (AFD < 8) {
if (posBalance < 60) {
if (tcor < 71) {
return 8;
}
else if (tcor >= 71) {
return 1;
}
}
else if (posBalance >= 60) {
if (Cs < 99) {
if (SC < 123) {
return 1;
}
else if (SC >= 123) {
return 2;
}
}
else if (Cs >= 99) {
return 1;
}
}
}
else if (AFD >= 8) {
if (tcor < 90) {
if (gchDC < 1) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
return 8;
}
}
else if (gchDC >= 1) {
if (Cs < 319) {
if (TSC < 4) {
if (Rs < 135) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
return 2;
}
}
else if (Rs >= 135) {
return 1;
}
}
else if (TSC >= 4) {
if (refDCval < 78) {
return 8;
}
else if (refDCval >= 78) {
return 1;
}
}
}
else if (Cs >= 319) {
return 2;
}
}
}
else if (tcor >= 90) {
if (Cs < 138) {
return 8;
}
else if (Cs >= 138) {
return 2;
}
}
}
}
else if (ssDCval >= 100) {
if (SC < 5) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
if (MVDiff < 164) {
return 8;
}
else if (MVDiff >= 164) {
return 1;
}
}
}
else if (SC >= 5) {
if (AFD < 12) {
if (mu_mv_mag_sq < 68) {
if (ssDCval < 108) {
if (Rs < 84) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Rs < 23) {
return 1;
}
else if (Rs >= 23) {
return 2;
}
}
}
else if (Rs >= 84) {
return 2;
}
}
else if (ssDCval >= 108) {
if (ssDCval < 159) {
return 2;
}
else if (ssDCval >= 159) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 68) {
return 1;
}
}
else if (AFD >= 12) {
if (ssDCval < 117) {
if (Cs < 350) {
if (gchDC < 6) {
return 1;
}
else if (gchDC >= 6) {
if (refDCval < 103) {
return 8;
}
else if (refDCval >= 103) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (Cs >= 350) {
return 2;
}
}
else if (ssDCval >= 117) {
if (refDCval < 138) {
if (MVDiff < 136) {
return 8;
}
else if (MVDiff >= 136) {
if (Cs < 14) {
return 8;
}
else if (Cs >= 14) {
return 2;
}
}
}
else if (refDCval >= 138) {
if (Rs < 131) {
return 2;
}
else if (Rs >= 131) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 142) {
if (RsDiff < 2) {
if (negBalance < 31) {
if (posBalance < 50) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
if (Rs < 259) {
return 1;
}
else if (Rs >= 259) {
return 8;
}
}
}
else if (posBalance >= 50) {
if (refDCval < 102) {
if (Cs < 8) {
return 2;
}
else if (Cs >= 8) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 8;
}
}
}
else if (refDCval >= 102) {
if (Cs < 90) {
return 1;
}
else if (Cs >= 90) {
return 2;
}
}
}
}
else if (negBalance >= 31) {
if (Cs < 72) {
if (mu_mv_mag_sq < 34) {
if (SC < 218) {
if (Rs < 173) {
return 2;
}
else if (Rs >= 173) {
return 1;
}
}
else if (SC >= 218) {
return 1;
}
}
else if (mu_mv_mag_sq >= 34) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
return 2;
}
}
}
else if (Cs >= 72) {
if (negBalance < 66) {
if (mcTcor < 97) {
if (TSC < 5) {
if (diffMVdiffVal < -5) {
if (mcTcor < 94) {
if (diffRsCsdiff < -3) {
return 8;
}
else if (diffRsCsdiff >= -3) {
return 2;
}
}
else if (mcTcor >= 94) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
if (RsCsDiff < 1) {
if (Rs < 229) {
if (tcor < 66) {
return 8;
}
else if (tcor >= 66) {
return 2;
}
}
else if (Rs >= 229) {
return 8;
}
}
else if (RsCsDiff >= 1) {
if (Rs < 265) {
if (Rs < 239) {
if (tcor < 82) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
if (AFD < 8) {
if (MVDiff < 22) {
if (Cs < 353) {
return 8;
}
else if (Cs >= 353) {
return 2;
}
}
else if (MVDiff >= 22) {
return 2;
}
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (tcor >= 82) {
if (Cs < 353) {
return 2;
}
else if (Cs >= 353) {
return 8;
}
}
}
else if (Rs >= 239) {
return 8;
}
}
else if (Rs >= 265) {
if (Rs < 275) {
return 2;
}
else if (Rs >= 275) {
return 8;
}
}
}
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 5) {
if (Cs < 338) {
return 2;
}
else if (Cs >= 338) {
return 1;
}
}
else if (mu_mv_mag_sq >= 5) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (refDCval < 91) {
return 1;
}
else if (refDCval >= 91) {
if (posBalance < 65) {
return 2;
}
else if (posBalance >= 65) {
if (Rs < 175) {
if (Rs < 152) {
return 2;
}
else if (Rs >= 152) {
return 1;
}
}
else if (Rs >= 175) {
return 8;
}
}
}
}
}
else if (negBalance >= 66) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (negBalance < 53) {
if (diffAFD < 3) {
if (tcor < 75) {
if (Rs < 273) {
return 8;
}
else if (Rs >= 273) {
return 2;
}
}
else if (tcor >= 75) {
if (SC < 562) {
return 1;
}
else if (SC >= 562) {
if (Rs < 256) {
return 8;
}
else if (Rs >= 256) {
return 2;
}
}
}
}
else if (diffAFD >= 3) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
if (Rs < 245) {
return 2;
}
else if (Rs >= 245) {
return 1;
}
}
}
}
else if (negBalance >= 53) {
if (Rs < 242) {
if (diffMVdiffVal < 1) {
if (Cs < 68) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (Cs >= 68) {
if (posBalance < 54) {
return 2;
}
else if (posBalance >= 54) {
if (Cs < 242) {
return 8;
}
else if (Cs >= 242) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 232) {
if (mu_mv_mag_sq < 10) {
return 4;
}
else if (mu_mv_mag_sq >= 10) {
if (Rs < 167) {
return 4;
}
else if (Rs >= 167) {
return 2;
}
}
}
else if (Cs >= 232) {
return 8;
}
}
}
else if (Rs >= 242) {
if (MVDiff < 25) {
if (Rs < 251) {
return 4;
}
else if (Rs >= 251) {
return 2;
}
}
else if (MVDiff >= 25) {
if (Cs < 489) {
return 1;
}
else if (Cs >= 489) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 282) {
if (RsDiff < 4) {
if (mcTcor < 90) {
if (Cs < 145) {
if (RsDiff < 2) {
if (Rs < 554) {
return 8;
}
else if (Rs >= 554) {
return 1;
}
}
else if (RsDiff >= 2) {
if (TSC < 4) {
if (Rs < 327) {
return 8;
}
else if (Rs >= 327) {
if (MVDiff < 35) {
return 2;
}
else if (MVDiff >= 35) {
if (Rs < 434) {
return 8;
}
else if (Rs >= 434) {
return 1;
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 67) {
if (Cs < 135) {
if (diffAFD < 4) {
if (tcor < 49) {
if (Rs < 337) {
return 8;
}
else if (Rs >= 337) {
return 1;
}
}
else if (tcor >= 49) {
return 1;
}
}
else if (diffAFD >= 4) {
if (Rs < 405) {
return 2;
}
else if (Rs >= 405) {
return 8;
}
}
}
else if (Cs >= 135) {
if (Rs < 381) {
return 8;
}
else if (Rs >= 381) {
if (Rs < 662) {
return 1;
}
else if (Rs >= 662) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 315) {
return 2;
}
else if (Rs >= 315) {
return 4;
}
}
}
}
}
else if (Cs >= 145) {
if (tcor < 66) {
if (refDCval < 81) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
if (negBalance < 47) {
if (tcor < 58) {
if (diffRsCsdiff < 2) {
if (tcor < 51) {
return 4;
}
else if (tcor >= 51) {
if (Rs < 387) {
return 8;
}
else if (Rs >= 387) {
return 1;
}
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 372) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 1;
}
}
else if (Rs >= 372) {
return 8;
}
}
}
else if (tcor >= 58) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (refDCval < 54) {
if (TSC < 5) {
if (Cs < 201) {
return 8;
}
else if (Cs >= 201) {
return 4;
}
}
else if (TSC >= 5) {
if (negBalance < 52) {
if (tcor < 58) {
if (Cs < 274) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (Cs >= 274) {
return 4;
}
}
else if (tcor >= 58) {
return 8;
}
}
else if (negBalance >= 52) {
if (diffRsCsdiff < -4) {
if (Rs < 394) {
return 8;
}
else if (Rs >= 394) {
return 2;
}
}
else if (diffRsCsdiff >= -4) {
return 8;
}
}
}
}
else if (refDCval >= 54) {
if (MVDiff < 64) {
return 4;
}
else if (MVDiff >= 64) {
if (Cs < 411) {
if (Cs < 162) {
return 8;
}
else if (Cs >= 162) {
return 2;
}
}
else if (Cs >= 411) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 81) {
if (Rs < 323) {
if (Rs < 315) {
return 1;
}
else if (Rs >= 315) {
if (Rs < 318) {
return 2;
}
else if (Rs >= 318) {
return 4;
}
}
}
else if (Rs >= 323) {
return 2;
}
}
}
else if (tcor >= 66) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (ssDCval < 45) {
if (gchDC < 1) {
if (posBalance < 51) {
if (ssDCval < 36) {
return 1;
}
else if (ssDCval >= 36) {
if (negBalance < 23) {
return 1;
}
else if (negBalance >= 23) {
return 8;
}
}
}
else if (posBalance >= 51) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (gchDC >= 1) {
if (MVDiff < 81) {
if (negBalance < 44) {
if (posBalance < 62) {
if (tcor < 83) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (MVDiff < 17) {
if (Rs < 415) {
return 1;
}
else if (Rs >= 415) {
return 8;
}
}
else if (MVDiff >= 17) {
return 2;
}
}
}
else if (tcor >= 83) {
return 8;
}
}
else if (posBalance >= 62) {
return 8;
}
}
else if (negBalance >= 44) {
if (Rs < 616) {
return 8;
}
else if (Rs >= 616) {
if (Rs < 627) {
return 4;
}
else if (Rs >= 627) {
return 8;
}
}
}
}
else if (MVDiff >= 81) {
if (Cs < 243) {
return 4;
}
else if (Cs >= 243) {
return 1;
}
}
}
}
else if (ssDCval >= 45) {
if (refDCval < 89) {
if (mu_mv_mag_sq < 92) {
if (TSC < 4) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
return 1;
}
}
else if (TSC >= 4) {
if (refDCval < 78) {
if (refDCval < 58) {
if (Cs < 356) {
if (negBalance < 44) {
if (Cs < 327) {
return 2;
}
else if (Cs >= 327) {
return 8;
}
}
else if (negBalance >= 44) {
if (Cs < 213) {
return 4;
}
else if (Cs >= 213) {
return 8;
}
}
}
else if (Cs >= 356) {
if (posBalance < 62) {
if (ssDCval < 49) {
if (Rs < 848) {
return 1;
}
else if (Rs >= 848) {
return 8;
}
}
else if (ssDCval >= 49) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 4;
}
}
}
}
else if (posBalance >= 62) {
return 8;
}
}
}
else if (refDCval >= 58) {
if (refDCval < 62) {
return 2;
}
else if (refDCval >= 62) {
if (SC < 674) {
if (Rs < 317) {
return 2;
}
else if (Rs >= 317) {
return 1;
}
}
else if (SC >= 674) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
return 4;
}
}
}
}
}
else if (refDCval >= 78) {
if (Cs < 322) {
if (Rs < 388) {
return 2;
}
else if (Rs >= 388) {
return 1;
}
}
else if (Cs >= 322) {
if (ssDCval < 87) {
if (AFD < 11) {
if (mu_mv_mag_sq < 69) {
return 8;
}
else if (mu_mv_mag_sq >= 69) {
if (Rs < 329) {
return 2;
}
else if (Rs >= 329) {
return 8;
}
}
}
else if (AFD >= 11) {
return 2;
}
}
else if (ssDCval >= 87) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (refDCval < 52) {
return 4;
}
else if (refDCval >= 52) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
else if (refDCval >= 89) {
if (diffMVdiffVal < -18) {
return 2;
}
else if (diffMVdiffVal >= -18) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (ssDCval < 73) {
if (Rs < 1292) {
if (posBalance < 26) {
if (diffAFD < -3) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (diffAFD >= -3) {
if (tcor < 101) {
if (posBalance < 24) {
if (Cs < 354) {
if (SC < 978) {
if (posBalance < 19) {
return 8;
}
else if (posBalance >= 19) {
return 4;
}
}
else if (SC >= 978) {
return 1;
}
}
else if (Cs >= 354) {
return 8;
}
}
else if (posBalance >= 24) {
if (tcor < 98) {
if (Rs < 432) {
return 1;
}
else if (Rs >= 432) {
return 8;
}
}
else if (tcor >= 98) {
return 1;
}
}
}
else if (tcor >= 101) {
if (SC < 447) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Rs < 350) {
return 8;
}
else if (Rs >= 350) {
return 1;
}
}
}
else if (SC >= 447) {
if (negBalance < 81) {
return 8;
}
else if (negBalance >= 81) {
if (mu_mv_mag_sq < 67) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
return 4;
}
}
else if (mu_mv_mag_sq >= 67) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 26) {
if (refDCval < 69) {
if (MVDiff < 2) {
if (tcor < 95) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
if (TSC < 3) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (Rs < 683) {
return 8;
}
else if (Rs >= 683) {
if (SC < 958) {
return 2;
}
else if (SC >= 958) {
return 1;
}
}
}
}
else if (TSC >= 3) {
if (Rs < 461) {
if (Rs < 376) {
if (Rs < 325) {
return 4;
}
else if (Rs >= 325) {
return 8;
}
}
else if (Rs >= 376) {
if (Rs < 415) {
return 1;
}
else if (Rs >= 415) {
return 2;
}
}
}
else if (Rs >= 461) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (Rs < 976) {
return 8;
}
else if (Rs >= 976) {
return 4;
}
}
}
}
}
}
else if (tcor >= 95) {
if (negBalance < 40) {
if (posBalance < 44) {
if (Rs < 895) {
if (tcor < 100) {
if (Cs < 407) {
if (tcor < 99) {
if (posBalance < 35) {
if (Rs < 682) {
if (Rs < 554) {
return 1;
}
else if (Rs >= 554) {
return 8;
}
}
else if (Rs >= 682) {
return 1;
}
}
else if (posBalance >= 35) {
return 1;
}
}
else if (tcor >= 99) {
if (ssDCval < 65) {
if (ssDCval < 61) {
if (Rs < 469) {
return 1;
}
else if (Rs >= 469) {
if (Rs < 738) {
return 8;
}
else if (Rs >= 738) {
return 2;
}
}
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (ssDCval >= 65) {
if (Rs < 562) {
return 8;
}
else if (Rs >= 562) {
return 2;
}
}
}
}
else if (Cs >= 407) {
if (ssDCval < 64) {
if (ssDCval < 53) {
return 1;
}
else if (ssDCval >= 53) {
if (Rs < 508) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (Rs >= 508) {
if (refDCval < 57) {
if (Rs < 679) {
return 8;
}
else if (Rs >= 679) {
return 1;
}
}
else if (refDCval >= 57) {
return 8;
}
}
}
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (tcor >= 100) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
if (Rs < 553) {
return 4;
}
else if (Rs >= 553) {
return 1;
}
}
}
}
else if (Rs >= 895) {
if (SC < 1694) {
return 1;
}
else if (SC >= 1694) {
if (Rs < 1253) {
return 1;
}
else if (Rs >= 1253) {
return 8;
}
}
}
}
else if (posBalance >= 44) {
if (Rs < 577) {
if (posBalance < 54) {
if (Cs < 274) {
if (ssDCval < 55) {
if (Rs < 402) {
return 1;
}
else if (Rs >= 402) {
return 8;
}
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (Cs >= 274) {
if (diffAFD < 2) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 115) {
return 1;
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
else if (mcTcor >= 97) {
if (SC < 1029) {
if (refDCval < 68) {
return 1;
}
else if (refDCval >= 68) {
if (Rs < 500) {
return 1;
}
else if (Rs >= 500) {
return 8;
}
}
}
else if (SC >= 1029) {
if (Rs < 563) {
return 8;
}
else if (Rs >= 563) {
return 1;
}
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (Rs >= 577) {
if (tcor < 99) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (AFD < 2) {
if (Rs < 736) {
return 1;
}
else if (Rs >= 736) {
return 2;
}
}
else if (AFD >= 2) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 613) {
return 2;
}
else if (Rs >= 613) {
return 8;
}
}
}
}
}
else if (tcor >= 99) {
if (Rs < 1079) {
return 2;
}
else if (Rs >= 1079) {
return 1;
}
}
}
}
}
else if (negBalance >= 40) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 127) {
if (mu_mv_mag_sq < 49) {
if (tcor < 81) {
if (Rs < 429) {
return 1;
}
else if (Rs >= 429) {
if (RsDiff < 3) {
if (Rs < 943) {
return 8;
}
else if (Rs >= 943) {
return 2;
}
}
else if (RsDiff >= 3) {
if (Cs < 121) {
return 1;
}
else if (Cs >= 121) {
return 8;
}
}
}
}
else if (tcor >= 81) {
if (tcor < 85) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
return 1;
}
}
else if (tcor >= 85) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (TSC < 3) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 4;
}
}
else if (TSC >= 3) {
if (Cs < 86) {
return 8;
}
else if (Cs >= 86) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (Rs < 591) {
return 8;
}
else if (Rs >= 591) {
return 2;
}
}
}
}
}
}
else if (Cs >= 127) {
if (ssDCval < 45) {
if (posBalance < 32) {
if (mu_mv_mag_sq < 88) {
if (ssDCval < 34) {
return 4;
}
else if (ssDCval >= 34) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 88) {
return 1;
}
}
else if (posBalance >= 32) {
if (Cs < 447) {
if (posBalance < 50) {
if (mu_mv_mag_sq < 70) {
if (TSC < 2) {
if (Cs < 333) {
return 4;
}
else if (Cs >= 333) {
return 2;
}
}
else if (TSC >= 2) {
if (MVDiff < 35) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 13) {
return 4;
}
else if (mu_mv_mag_sq >= 13) {
if (posBalance < 35) {
return 4;
}
else if (posBalance >= 35) {
return 1;
}
}
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (MVDiff >= 35) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 70) {
return 1;
}
}
else if (posBalance >= 50) {
if (MVDiff < 8) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (negBalance < 43) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
if (Cs < 283) {
return 1;
}
else if (Cs >= 283) {
return 4;
}
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (SC < 925) {
return 1;
}
else if (SC >= 925) {
if (Rs < 584) {
return 8;
}
else if (Rs >= 584) {
return 1;
}
}
}
}
}
else if (Cs >= 447) {
if (Rs < 912) {
if (tcor < 78) {
return 1;
}
else if (tcor >= 78) {
if (SC < 1066) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Cs < 456) {
return 8;
}
else if (Cs >= 456) {
return 1;
}
}
}
else if (SC >= 1066) {
return 8;
}
}
}
else if (Rs >= 912) {
return 2;
}
}
}
}
else if (ssDCval >= 45) {
if (SC < 709) {
if (Cs < 265) {
if (negBalance < 30) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 66) {
return 4;
}
else if (mu_mv_mag_sq >= 66) {
if (TSC < 3) {
if (ssDCval < 49) {
if (Cs < 239) {
return 1;
}
else if (Cs >= 239) {
return 4;
}
}
else if (ssDCval >= 49) {
return 4;
}
}
else if (TSC >= 3) {
return 1;
}
}
}
}
else if (negBalance >= 30) {
if (tcor < 96) {
if (posBalance < 62) {
if (mu_mv_mag_sq < 51) {
if (tcor < 85) {
if (Rs < 294) {
return 2;
}
else if (Rs >= 294) {
return 4;
}
}
else if (tcor >= 85) {
if (SC < 569) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (Rs < 290) {
return 4;
}
else if (Rs >= 290) {
return 1;
}
}
}
else if (SC >= 569) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 51) {
if (Cs < 180) {
return 2;
}
else if (Cs >= 180) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
return 4;
}
}
}
}
}
else if (posBalance >= 62) {
if (Cs < 213) {
if (Cs < 138) {
return 8;
}
else if (Cs >= 138) {
return 1;
}
}
else if (Cs >= 213) {
return 2;
}
}
}
else if (tcor >= 96) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
if (MVDiff < 31) {
return 1;
}
else if (MVDiff >= 31) {
return 4;
}
}
}
}
}
else if (Cs >= 265) {
if (ssDCval < 65) {
if (ssDCval < 58) {
if (mu_mv_mag_sq < 52) {
if (negBalance < 42) {
if (Rs < 292) {
return 2;
}
else if (Rs >= 292) {
return 4;
}
}
else if (negBalance >= 42) {
if (tcor < 83) {
if (Rs < 303) {
return 8;
}
else if (Rs >= 303) {
return 1;
}
}
else if (tcor >= 83) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 1;
}
}
}
else if (ssDCval >= 58) {
if (SC < 584) {
if (Rs < 286) {
return 1;
}
else if (Rs >= 286) {
return 8;
}
}
else if (SC >= 584) {
if (mu_mv_mag_sq < 109) {
if (SC < 700) {
return 2;
}
else if (SC >= 700) {
if (Rs < 325) {
return 2;
}
else if (Rs >= 325) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 2;
}
}
}
}
}
else if (ssDCval >= 65) {
if (Cs < 311) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 4;
}
}
else if (Cs >= 311) {
if (Rs < 322) {
return 4;
}
else if (Rs >= 322) {
return 8;
}
}
}
}
}
else if (SC >= 709) {
if (mu_mv_mag_sq < 25) {
if (tcor < 93) {
if (ssDCval < 64) {
if (MVDiff < 55) {
if (refDCval < 57) {
if (RsCsDiff < 12) {
if (Rs < 508) {
if (RsDiff < 2) {
if (Cs < 367) {
return 1;
}
else if (Cs >= 367) {
return 4;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (Rs >= 508) {
return 1;
}
}
else if (RsCsDiff >= 12) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
if (Rs < 586) {
return 2;
}
else if (Rs >= 586) {
return 8;
}
}
}
}
}
else if (refDCval >= 57) {
if (AFD < 10) {
if (Rs < 797) {
if (Rs < 327) {
return 1;
}
else if (Rs >= 327) {
if (TSC < 5) {
if (Cs < 501) {
if (mcTcor < 98) {
return 2;
}
else if (mcTcor >= 98) {
return 4;
}
}
else if (Cs >= 501) {
return 4;
}
}
else if (TSC >= 5) {
if (Rs < 648) {
return 1;
}
else if (Rs >= 648) {
return 4;
}
}
}
}
else if (Rs >= 797) {
return 1;
}
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (MVDiff >= 55) {
return 2;
}
}
else if (ssDCval >= 64) {
if (negBalance < 52) {
if (negBalance < 47) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (tcor < 90) {
if (Cs < 168) {
return 1;
}
else if (Cs >= 168) {
return 4;
}
}
else if (tcor >= 90) {
return 1;
}
}
}
}
else if (negBalance >= 47) {
return 4;
}
}
else if (negBalance >= 52) {
return 2;
}
}
}
else if (tcor >= 93) {
if (TSC < 4) {
if (negBalance < 30) {
if (SC < 1120) {
if (Rs < 915) {
return 4;
}
else if (Rs >= 915) {
return 1;
}
}
else if (SC >= 1120) {
if (Rs < 907) {
return 2;
}
else if (Rs >= 907) {
return 4;
}
}
}
else if (negBalance >= 30) {
if (RsCsDiff < 3) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (negBalance < 36) {
return 4;
}
else if (negBalance >= 36) {
if (Cs < 202) {
return 4;
}
else if (Cs >= 202) {
return 1;
}
}
}
}
else if (RsCsDiff >= 3) {
if (AFD < 9) {
if (Rs < 934) {
return 4;
}
else if (Rs >= 934) {
return 1;
}
}
else if (AFD >= 9) {
return 1;
}
}
}
}
else if (TSC >= 4) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (tcor < 91) {
if (negBalance < 44) {
if (MVDiff < 10) {
if (Rs < 465) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
else if (Rs >= 465) {
if (Cs < 239) {
if (mu_mv_mag_sq < 74) {
return 1;
}
else if (mu_mv_mag_sq >= 74) {
return 2;
}
}
else if (Cs >= 239) {
if (diffAFD < 1) {
if (negBalance < 23) {
return 1;
}
else if (negBalance >= 23) {
return 4;
}
}
else if (diffAFD >= 1) {
if (Cs < 405) {
if (Rs < 709) {
return 4;
}
else if (Rs >= 709) {
return 1;
}
}
else if (Cs >= 405) {
if (Rs < 921) {
return 8;
}
else if (Rs >= 921) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 10) {
if (refDCval < 55) {
if (Rs < 553) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (Rs >= 553) {
if (Cs < 452) {
return 1;
}
else if (Cs >= 452) {
if (Rs < 651) {
return 4;
}
else if (Rs >= 651) {
if (Rs < 853) {
return 1;
}
else if (Rs >= 853) {
return 2;
}
}
}
}
}
else if (refDCval >= 55) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
}
}
}
else if (negBalance >= 44) {
if (SC < 972) {
if (diffMVdiffVal < 17) {
if (Rs < 547) {
return 4;
}
else if (Rs >= 547) {
if (Rs < 574) {
return 8;
}
else if (Rs >= 574) {
return 1;
}
}
}
else if (diffMVdiffVal >= 17) {
return 1;
}
}
else if (SC >= 972) {
if (Cs < 265) {
return 2;
}
else if (Cs >= 265) {
return 8;
}
}
}
}
else if (tcor >= 91) {
if (mu_mv_mag_sq < 66) {
if (Rs < 674) {
if (Rs < 400) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
if (Cs < 410) {
return 4;
}
else if (Cs >= 410) {
return 1;
}
}
}
}
else if (Rs >= 400) {
if (Cs < 415) {
if (Cs < 336) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 336) {
if (Rs < 427) {
return 1;
}
else if (Rs >= 427) {
if (diffRsCsdiff < 1) {
return 4;
}
else if (diffRsCsdiff >= 1) {
return 1;
}
}
}
}
else if (Cs >= 415) {
if (diffAFD < -1) {
if (Rs < 524) {
if (SC < 883) {
return 2;
}
else if (SC >= 883) {
return 1;
}
}
else if (Rs >= 524) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (Rs >= 674) {
if (posBalance < 37) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
return 1;
}
}
}
else if (posBalance >= 37) {
if (negBalance < 45) {
if (mu_mv_mag_sq < 65) {
return 1;
}
else if (mu_mv_mag_sq >= 65) {
if (Rs < 1014) {
return 1;
}
else if (Rs >= 1014) {
return 8;
}
}
}
else if (negBalance >= 45) {
if (Rs < 1001) {
return 4;
}
else if (Rs >= 1001) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (TSC < 4) {
if (mcTcor < 99) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 82) {
if (Rs < 660) {
return 2;
}
else if (Rs >= 660) {
return 4;
}
}
else if (mu_mv_mag_sq >= 82) {
if (mu_mv_mag_sq < 105) {
if (negBalance < 29) {
if (Rs < 440) {
return 8;
}
else if (Rs >= 440) {
return 1;
}
}
else if (negBalance >= 29) {
if (Rs < 445) {
if (Rs < 403) {
return 2;
}
else if (Rs >= 403) {
return 1;
}
}
else if (Rs >= 445) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Cs < 506) {
return 8;
}
else if (Cs >= 506) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
if (SC < 1043) {
if (mu_mv_mag_sq < 89) {
if (tcor < 94) {
if (Rs < 442) {
return 8;
}
else if (Rs >= 442) {
return 1;
}
}
else if (tcor >= 94) {
if (negBalance < 39) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (SC < 792) {
if (Rs < 516) {
return 4;
}
else if (Rs >= 516) {
return 1;
}
}
else if (SC >= 792) {
return 1;
}
}
}
else if (negBalance >= 39) {
if (Rs < 426) {
return 1;
}
else if (Rs >= 426) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (posBalance < 39) {
if (Cs < 473) {
return 8;
}
else if (Cs >= 473) {
return 1;
}
}
else if (posBalance >= 39) {
if (Cs < 519) {
return 1;
}
else if (Cs >= 519) {
return 2;
}
}
}
}
else if (SC >= 1043) {
return 1;
}
}
}
else if (mcTcor >= 99) {
if (SC < 985) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (CsDiff < 1) {
if (ssDCval < 53) {
if (Rs < 649) {
return 1;
}
else if (Rs >= 649) {
return 8;
}
}
else if (ssDCval >= 53) {
if (mu_mv_mag_sq < 105) {
return 8;
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 621) {
if (Cs < 346) {
return 2;
}
else if (Cs >= 346) {
return 1;
}
}
else if (Rs >= 621) {
return 8;
}
}
}
}
else if (CsDiff >= 1) {
if (Rs < 379) {
return 4;
}
else if (Rs >= 379) {
if (Rs < 503) {
return 8;
}
else if (Rs >= 503) {
if (Cs < 409) {
return 4;
}
else if (Cs >= 409) {
return 8;
}
}
}
}
}
}
else if (SC >= 985) {
if (ssDCval < 68) {
if (mu_mv_mag_sq < 119) {
return 1;
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 586) {
return 1;
}
else if (Rs >= 586) {
return 8;
}
}
}
else if (ssDCval >= 68) {
return 2;
}
}
}
}
else if (TSC >= 4) {
if (RsDiff < 3) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 695) {
return 4;
}
else if (Rs >= 695) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 829) {
return 2;
}
else if (Rs >= 829) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 69) {
if (mu_mv_mag_sq < 91) {
if (Rs < 523) {
if (MVDiff < 5) {
if (Rs < 386) {
return 4;
}
else if (Rs >= 386) {
return 1;
}
}
else if (MVDiff >= 5) {
if (mu_mv_mag_sq < 30) {
if (posBalance < 50) {
if (Rs < 494) {
return 4;
}
else if (Rs >= 494) {
return 2;
}
}
else if (posBalance >= 50) {
return 1;
}
}
else if (mu_mv_mag_sq >= 30) {
if (diffTSC < -1) {
if (Rs < 365) {
return 8;
}
else if (Rs >= 365) {
return 1;
}
}
else if (diffTSC >= -1) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
if (mcTcor < 95) {
if (Rs < 435) {
return 8;
}
else if (Rs >= 435) {
return 2;
}
}
else if (mcTcor >= 95) {
return 8;
}
}
}
}
}
}
else if (Rs >= 523) {
if (Rs < 745) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Rs < 547) {
if (Cs < 263) {
return 4;
}
else if (Cs >= 263) {
return 8;
}
}
else if (Rs >= 547) {
if (Cs < 177) {
return 1;
}
else if (Cs >= 177) {
if (Rs < 578) {
return 1;
}
else if (Rs >= 578) {
return 4;
}
}
}
}
}
else if (Rs >= 745) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (mu_mv_mag_sq < 116) {
if (MVDiff < 3) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (posBalance < 47) {
if (posBalance < 43) {
if (Rs < 652) {
return 2;
}
else if (Rs >= 652) {
return 4;
}
}
else if (posBalance >= 43) {
return 1;
}
}
else if (posBalance >= 47) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 70) {
if (Rs < 880) {
return 2;
}
else if (Rs >= 880) {
return 8;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 116) {
if (Rs < 597) {
if (Cs < 384) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (Cs >= 384) {
if (Rs < 435) {
if (Rs < 321) {
return 1;
}
else if (Rs >= 321) {
return 2;
}
}
else if (Rs >= 435) {
return 1;
}
}
}
else if (Rs >= 597) {
if (negBalance < 20) {
if (ssDCval < 69) {
if (Rs < 769) {
return 1;
}
else if (Rs >= 769) {
return 8;
}
}
else if (ssDCval >= 69) {
if (ssDCval < 70) {
if (Rs < 645) {
return 1;
}
else if (Rs >= 645) {
return 4;
}
}
else if (ssDCval >= 70) {
return 4;
}
}
}
else if (negBalance >= 20) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
if (Cs < 313) {
return 4;
}
else if (Cs >= 313) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1292) {
if (SC < 1587) {
if (mu_mv_mag_sq < 58) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
else if (mu_mv_mag_sq >= 58) {
return 8;
}
}
else if (SC >= 1587) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1718) {
if (negBalance < 24) {
if (refDCval < 64) {
if (Rs < 1367) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
if (Rs < 1300) {
return 1;
}
else if (Rs >= 1300) {
return 8;
}
}
}
else if (Rs >= 1367) {
if (Rs < 1423) {
if (Cs < 508) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (Cs >= 508) {
return 1;
}
}
else if (Rs >= 1423) {
if (Rs < 1446) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
return 2;
}
}
else if (Rs >= 1446) {
if (posBalance < 70) {
if (mu_mv_mag_sq < 126) {
return 1;
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
else if (posBalance >= 70) {
return 2;
}
}
}
}
}
else if (refDCval >= 64) {
if (mu_mv_mag_sq < 91) {
if (Rs < 1401) {
if (Rs < 1319) {
return 8;
}
else if (Rs >= 1319) {
return 2;
}
}
else if (Rs >= 1401) {
return 4;
}
}
else if (mu_mv_mag_sq >= 91) {
if (ssDCval < 64) {
if (Rs < 1462) {
return 2;
}
else if (Rs >= 1462) {
return 1;
}
}
else if (ssDCval >= 64) {
return 1;
}
}
}
}
else if (negBalance >= 24) {
if (tcor < 98) {
if (TSC < 5) {
if (diffRsCsdiff < -29) {
return 2;
}
else if (diffRsCsdiff >= -29) {
return 1;
}
}
else if (TSC >= 5) {
return 4;
}
}
else if (tcor >= 98) {
if (MVDiff < 6) {
if (posBalance < 27) {
return 1;
}
else if (posBalance >= 27) {
return 8;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
}
}
else if (Rs >= 1718) {
if (Cs < 434) {
if (RsCsDiff < 10) {
return 2;
}
else if (RsCsDiff >= 10) {
if (Cs < 261) {
return 1;
}
else if (Cs >= 261) {
return 2;
}
}
}
else if (Cs >= 434) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (diffTSC < -1) {
if (Cs < 521) {
if (Rs < 1754) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
else if (Rs >= 1754) {
return 2;
}
}
else if (Cs >= 521) {
return 1;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
}
else if (ssDCval >= 73) {
if (Cs < 285) {
if (refDCval < 107) {
if (mu_mv_mag_sq < 115) {
if (SC < 1259) {
if (ssDCval < 102) {
if (RsCsDiff < 2) {
if (AFD < 4) {
if (ssDCval < 82) {
if (ssDCval < 80) {
if (refDCval < 73) {
return 1;
}
else if (refDCval >= 73) {
return 8;
}
}
else if (ssDCval >= 80) {
return 1;
}
}
else if (ssDCval >= 82) {
if (mu_mv_mag_sq < 82) {
return 8;
}
else if (mu_mv_mag_sq >= 82) {
return 2;
}
}
}
else if (AFD >= 4) {
if (Cs < 86) {
return 1;
}
else if (Cs >= 86) {
if (TSC < 3) {
if (Cs < 158) {
return 8;
}
else if (Cs >= 158) {
return 1;
}
}
else if (TSC >= 3) {
return 8;
}
}
}
}
else if (RsCsDiff >= 2) {
if (RsDiff < 2) {
if (MVDiff < 6) {
if (diffRsCsdiff < 1) {
if (refDCval < 92) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
return 8;
}
}
else if (refDCval >= 92) {
return 8;
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 426) {
return 1;
}
else if (Rs >= 426) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (Rs < 289) {
return 1;
}
else if (Rs >= 289) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (AFD < 10) {
if (TSC < 5) {
if (Cs < 142) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
return 1;
}
}
else if (Cs >= 142) {
if (Rs < 669) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 464) {
return 1;
}
else if (Rs >= 464) {
return 4;
}
}
}
else if (Rs >= 669) {
if (Cs < 245) {
return 4;
}
else if (Cs >= 245) {
return 8;
}
}
}
}
else if (TSC >= 5) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (AFD >= 10) {
if (Rs < 469) {
if (mu_mv_mag_sq < 40) {
if (ssDCval < 88) {
if (MVDiff < 54) {
return 1;
}
else if (MVDiff >= 54) {
return 8;
}
}
else if (ssDCval >= 88) {
if (MVDiff < 20) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (MVDiff >= 20) {
if (Rs < 325) {
return 8;
}
else if (Rs >= 325) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 40) {
if (RsCsDiff < 7) {
return 4;
}
else if (RsCsDiff >= 7) {
return 2;
}
}
}
else if (Rs >= 469) {
if (mu_mv_mag_sq < 3) {
if (Rs < 566) {
return 1;
}
else if (Rs >= 566) {
if (Rs < 709) {
return 8;
}
else if (Rs >= 709) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 102) {
return 1;
}
}
else if (SC >= 1259) {
if (refDCval < 102) {
if (posBalance < 19) {
if (Rs < 1202) {
return 8;
}
else if (Rs >= 1202) {
return 1;
}
}
else if (posBalance >= 19) {
return 1;
}
}
else if (refDCval >= 102) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (ssDCval < 79) {
if (Cs < 204) {
return 8;
}
else if (Cs >= 204) {
return 1;
}
}
else if (ssDCval >= 79) {
if (Rs < 310) {
if (Cs < 200) {
return 1;
}
else if (Cs >= 200) {
return 2;
}
}
else if (Rs >= 310) {
if (posBalance < 46) {
if (posBalance < 33) {
return 1;
}
else if (posBalance >= 33) {
if (ssDCval < 94) {
if (Cs < 251) {
if (Rs < 316) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (Rs >= 316) {
return 2;
}
}
else if (Cs >= 251) {
if (Rs < 950) {
return 1;
}
else if (Rs >= 950) {
return 2;
}
}
}
else if (ssDCval >= 94) {
return 1;
}
}
}
else if (posBalance >= 46) {
return 1;
}
}
}
}
}
else if (refDCval >= 107) {
if (AFD < 6) {
if (MVDiff < 31) {
if (SC < 429) {
return 2;
}
else if (SC >= 429) {
return 8;
}
}
else if (MVDiff >= 31) {
if (Rs < 551) {
return 2;
}
else if (Rs >= 551) {
return 1;
}
}
}
else if (AFD >= 6) {
if (ssDCval < 117) {
if (RsCsDiff < 12) {
if (SC < 428) {
return 8;
}
else if (SC >= 428) {
if (mu_mv_mag_sq < 37) {
return 1;
}
else if (mu_mv_mag_sq >= 37) {
return 8;
}
}
}
else if (RsCsDiff >= 12) {
return 4;
}
}
else if (ssDCval >= 117) {
if (negBalance < 55) {
if (Rs < 333) {
return 2;
}
else if (Rs >= 333) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 2;
}
}
}
else if (negBalance >= 55) {
if (Rs < 352) {
return 8;
}
else if (Rs >= 352) {
return 1;
}
}
}
}
}
}
else if (Cs >= 285) {
if (refDCval < 108) {
if (Rs < 958) {
if (mu_mv_mag_sq < 2) {
if (TSC < 6) {
if (tcor < 86) {
if (Rs < 602) {
return 1;
}
else if (Rs >= 602) {
return 2;
}
}
else if (tcor >= 86) {
return 1;
}
}
else if (TSC >= 6) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Rs < 513) {
return 8;
}
else if (Rs >= 513) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 619) {
if (ssDCval < 94) {
if (refDCval < 92) {
if (Rs < 447) {
if (tcor < 88) {
if (TSC < 5) {
if (mcTcor < 96) {
if (Rs < 342) {
return 2;
}
else if (Rs >= 342) {
return 1;
}
}
else if (mcTcor >= 96) {
return 4;
}
}
else if (TSC >= 5) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 327) {
return 4;
}
else if (Rs >= 327) {
return 1;
}
}
}
}
else if (tcor >= 88) {
if (SC < 664) {
return 2;
}
else if (SC >= 664) {
if (Cs < 521) {
if (Cs < 357) {
if (Rs < 410) {
return 8;
}
else if (Rs >= 410) {
return 1;
}
}
else if (Cs >= 357) {
return 8;
}
}
else if (Cs >= 521) {
return 1;
}
}
}
}
else if (Rs >= 447) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 42) {
if (SC < 1006) {
return 1;
}
else if (SC >= 1006) {
if (Rs < 576) {
return 8;
}
else if (Rs >= 576) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 42) {
if (SC < 999) {
if (posBalance < 53) {
if (MVDiff < 5) {
if (Rs < 563) {
return 2;
}
else if (Rs >= 563) {
return 1;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (SC >= 999) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 92) {
if (mcTcor < 95) {
if (Rs < 460) {
return 8;
}
else if (Rs >= 460) {
return 2;
}
}
else if (mcTcor >= 95) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (ssDCval >= 94) {
if (Cs < 443) {
if (RsCsDiff < 4) {
if (Rs < 547) {
return 8;
}
else if (Rs >= 547) {
return 4;
}
}
else if (RsCsDiff >= 4) {
if (mcTcor < 97) {
if (MVDiff < 35) {
return 8;
}
else if (MVDiff >= 35) {
return 2;
}
}
else if (mcTcor >= 97) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (Rs < 413) {
return 2;
}
else if (Rs >= 413) {
return 1;
}
}
}
}
}
else if (Cs >= 443) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 439) {
return 1;
}
else if (Rs >= 439) {
return 2;
}
}
}
}
}
else if (Rs >= 619) {
if (MVDiff < 1) {
if (Rs < 695) {
if (ssDCval < 90) {
if (ssDCval < 78) {
if (Rs < 629) {
return 8;
}
else if (Rs >= 629) {
return 4;
}
}
else if (ssDCval >= 78) {
if (Cs < 318) {
return 8;
}
else if (Cs >= 318) {
return 1;
}
}
}
else if (ssDCval >= 90) {
if (Cs < 498) {
return 4;
}
else if (Cs >= 498) {
return 8;
}
}
}
else if (Rs >= 695) {
if (negBalance < 16) {
if (Rs < 754) {
return 1;
}
else if (Rs >= 754) {
return 8;
}
}
else if (negBalance >= 16) {
return 8;
}
}
}
else if (MVDiff >= 1) {
if (RsDiff < 2) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (mu_mv_mag_sq < 101) {
return 8;
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 798) {
return 1;
}
else if (Rs >= 798) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (mcTcor < 95) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
else if (mcTcor >= 95) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
if (AFD < 8) {
if (Rs < 657) {
return 2;
}
else if (Rs >= 657) {
if (Rs < 808) {
return 4;
}
else if (Rs >= 808) {
return 8;
}
}
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 958) {
if (mu_mv_mag_sq < 91) {
if (tcor < 98) {
if (posBalance < 55) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (refDCval < 75) {
return 8;
}
else if (refDCval >= 75) {
return 1;
}
}
}
}
else if (posBalance >= 55) {
if (Rs < 1436) {
if (Rs < 1034) {
return 1;
}
else if (Rs >= 1034) {
return 8;
}
}
else if (Rs >= 1436) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (mu_mv_mag_sq < 80) {
if (Cs < 300) {
if (Rs < 1786) {
return 8;
}
else if (Rs >= 1786) {
return 1;
}
}
else if (Cs >= 300) {
return 8;
}
}
else if (mu_mv_mag_sq >= 80) {
if (Cs < 468) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (Cs >= 468) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (Cs < 497) {
if (Rs < 2043) {
if (ssDCval < 92) {
if (RsDiff < 3) {
if (Rs < 1070) {
return 2;
}
else if (Rs >= 1070) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
return 2;
}
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (ssDCval >= 92) {
return 8;
}
}
else if (Rs >= 2043) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
return 8;
}
}
}
else if (Cs >= 497) {
if (Rs < 2069) {
return 4;
}
else if (Rs >= 2069) {
return 8;
}
}
}
}
}
else if (refDCval >= 108) {
if (refDCval < 130) {
if (mu_mv_mag_sq < 104) {
if (negBalance < 40) {
if (posBalance < 54) {
if (Cs < 505) {
if (negBalance < 38) {
if (SC < 2184) {
return 1;
}
else if (SC >= 2184) {
return 2;
}
}
else if (negBalance >= 38) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
if (Rs < 1280) {
return 1;
}
else if (Rs >= 1280) {
return 8;
}
}
}
}
else if (Cs >= 505) {
if (posBalance < 50) {
if (Rs < 596) {
if (Rs < 484) {
if (Rs < 409) {
return 4;
}
else if (Rs >= 409) {
return 1;
}
}
else if (Rs >= 484) {
return 2;
}
}
else if (Rs >= 596) {
return 1;
}
}
else if (posBalance >= 50) {
return 2;
}
}
}
else if (posBalance >= 54) {
if (MVDiff < 58) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (mcTcor < 96) {
return 1;
}
else if (mcTcor >= 96) {
return 2;
}
}
}
else if (MVDiff >= 58) {
if (Cs < 335) {
return 2;
}
else if (Cs >= 335) {
return 4;
}
}
}
}
else if (negBalance >= 40) {
if (posBalance < 60) {
if (CsDiff < 2) {
if (Rs < 548) {
if (refDCval < 124) {
if (ssDCval < 114) {
if (SC < 968) {
return 1;
}
else if (SC >= 968) {
return 8;
}
}
else if (ssDCval >= 114) {
return 2;
}
}
else if (refDCval >= 124) {
if (Cs < 408) {
if (MVDiff < 39) {
return 8;
}
else if (MVDiff >= 39) {
return 2;
}
}
else if (Cs >= 408) {
return 4;
}
}
}
else if (Rs >= 548) {
if (ssDCval < 125) {
return 8;
}
else if (ssDCval >= 125) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (MVDiff < 34) {
if (TSC < 6) {
if (RsCsDiff < 11) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (RsCsDiff >= 11) {
if (diffMVdiffVal < -6) {
if (Rs < 731) {
return 8;
}
else if (Rs >= 731) {
return 2;
}
}
else if (diffMVdiffVal >= -6) {
if (Rs < 1654) {
return 1;
}
else if (Rs >= 1654) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (Rs < 439) {
return 2;
}
else if (Rs >= 439) {
return 8;
}
}
}
else if (MVDiff >= 34) {
return 2;
}
}
}
else if (posBalance >= 60) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (Rs < 984) {
return 4;
}
else if (Rs >= 984) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
if (diffRsCsdiff < 3) {
return 8;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 1155) {
return 8;
}
else if (Rs >= 1155) {
return 1;
}
}
}
}
}
else if (refDCval >= 130) {
if (AFD < 7) {
if (ssDCval < 159) {
if (Cs < 386) {
if (RsCsDiff < 4) {
if (TSC < 4) {
if (MVDiff < 2) {
if (Rs < 395) {
return 2;
}
else if (Rs >= 395) {
return 8;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (TSC >= 4) {
return 4;
}
}
else if (RsCsDiff >= 4) {
if (ssDCval < 150) {
return 4;
}
else if (ssDCval >= 150) {
return 8;
}
}
}
else if (Cs >= 386) {
if (mcTcor < 99) {
if (refDCval < 133) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Rs < 634) {
return 2;
}
else if (Rs >= 634) {
return 1;
}
}
}
else if (refDCval >= 133) {
return 8;
}
}
else if (mcTcor >= 99) {
if (ssDCval < 130) {
return 8;
}
else if (ssDCval >= 130) {
return 1;
}
}
}
}
else if (ssDCval >= 159) {
return 4;
}
}
else if (AFD >= 7) {
if (Rs < 633) {
if (TSC < 5) {
if (ssDCval < 181) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (diffRsCsdiff < -16) {
return 8;
}
else if (diffRsCsdiff >= -16) {
return 1;
}
}
}
else if (ssDCval >= 181) {
return 2;
}
}
else if (TSC >= 5) {
if (Cs < 406) {
return 1;
}
else if (Cs >= 406) {
return 2;
}
}
}
else if (Rs >= 633) {
if (Cs < 403) {
return 4;
}
else if (Cs >= 403) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (Cs < 393) {
if (posBalance < 53) {
if (MVDiff < 35) {
if (Rs < 1285) {
if (CsDiff < 1) {
if (Rs < 583) {
return 1;
}
else if (Rs >= 583) {
return 2;
}
}
else if (CsDiff >= 1) {
if (negBalance < 57) {
if (Rs < 597) {
if (Rs < 532) {
return 1;
}
else if (Rs >= 532) {
return 8;
}
}
else if (Rs >= 597) {
return 1;
}
}
else if (negBalance >= 57) {
return 2;
}
}
}
else if (Rs >= 1285) {
if (RsCsDiff < 71) {
if (Cs < 197) {
return 2;
}
else if (Cs >= 197) {
return 1;
}
}
else if (RsCsDiff >= 71) {
if (diffMVdiffVal < -11) {
if (Rs < 2128) {
return 2;
}
else if (Rs >= 2128) {
return 1;
}
}
else if (diffMVdiffVal >= -11) {
return 2;
}
}
}
}
else if (MVDiff >= 35) {
if (mu_mv_mag_sq < 31) {
if (gchDC < 5) {
return 1;
}
else if (gchDC >= 5) {
if (MVDiff < 160) {
if (mu_mv_mag_sq < 10) {
return 8;
}
else if (mu_mv_mag_sq >= 10) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (Rs < 480) {
return 1;
}
else if (Rs >= 480) {
return 8;
}
}
}
}
else if (MVDiff >= 160) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 31) {
if (mu_mv_mag_sq < 65) {
if (Cs < 370) {
if (diffTSC < -1) {
if (diffRsCsdiff < -36) {
return 2;
}
else if (diffRsCsdiff >= -36) {
if (AFD < 35) {
if (diffMVdiffVal < -12) {
return 2;
}
else if (diffMVdiffVal >= -12) {
return 1;
}
}
else if (AFD >= 35) {
return 2;
}
}
}
else if (diffTSC >= -1) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (Cs >= 370) {
return 2;
}
}
else if (mu_mv_mag_sq >= 65) {
if (AFD < 25) {
if (mu_mv_mag_sq < 119) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (mu_mv_mag_sq >= 119) {
return 1;
}
}
else if (AFD >= 25) {
return 2;
}
}
}
}
}
else if (posBalance >= 53) {
if (RsCsDiff < 39) {
if (negBalance < 46) {
if (SC < 445) {
return 8;
}
else if (SC >= 445) {
if (Cs < 351) {
if (diffAFD < 1) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 1;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (Cs >= 351) {
return 2;
}
}
}
else if (negBalance >= 46) {
if (ssDCval < 75) {
if (diffMVdiffVal < -1) {
if (Rs < 403) {
return 1;
}
else if (Rs >= 403) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (refDCval < 60) {
if (Cs < 250) {
return 4;
}
else if (Cs >= 250) {
if (Cs < 354) {
return 8;
}
else if (Cs >= 354) {
return 1;
}
}
}
else if (refDCval >= 60) {
if (Rs < 563) {
return 4;
}
else if (Rs >= 563) {
return 2;
}
}
}
}
else if (ssDCval >= 75) {
if (tcor < 35) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
else if (tcor >= 35) {
if (MVDiff < 24) {
if (tcor < 66) {
if (Rs < 600) {
return 1;
}
else if (Rs >= 600) {
return 8;
}
}
else if (tcor >= 66) {
return 1;
}
}
else if (MVDiff >= 24) {
if (negBalance < 54) {
if (negBalance < 50) {
if (Cs < 237) {
if (Rs < 605) {
return 8;
}
else if (Rs >= 605) {
return 2;
}
}
else if (Cs >= 237) {
return 8;
}
}
else if (negBalance >= 50) {
if (Rs < 479) {
return 8;
}
else if (Rs >= 479) {
if (Cs < 78) {
return 4;
}
else if (Cs >= 78) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (mu_mv_mag_sq < 15) {
if (Cs < 92) {
return 8;
}
else if (Cs >= 92) {
return 1;
}
}
else if (mu_mv_mag_sq >= 15) {
if (AFD < 36) {
return 2;
}
else if (AFD >= 36) {
return 1;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 39) {
if (MVDiff < 20) {
if (ssDCval < 112) {
return 2;
}
else if (ssDCval >= 112) {
if (Rs < 1646) {
return 1;
}
else if (Rs >= 1646) {
return 8;
}
}
}
else if (MVDiff >= 20) {
if (tcor < 62) {
if (diffAFD < 7) {
if (negBalance < 60) {
if (Cs < 320) {
if (mcTcor < 84) {
if (posBalance < 65) {
if (negBalance < 53) {
if (Rs < 1800) {
return 1;
}
else if (Rs >= 1800) {
return 4;
}
}
else if (negBalance >= 53) {
return 4;
}
}
else if (posBalance >= 65) {
if (MVDiff < 131) {
return 1;
}
else if (MVDiff >= 131) {
if (Cs < 259) {
return 2;
}
else if (Cs >= 259) {
return 1;
}
}
}
}
else if (mcTcor >= 84) {
if (ssDCval < 84) {
if (tcor < 53) {
if (SC < 2157) {
return 1;
}
else if (SC >= 2157) {
return 4;
}
}
else if (tcor >= 53) {
if (Rs < 2201) {
return 2;
}
else if (Rs >= 2201) {
return 4;
}
}
}
else if (ssDCval >= 84) {
if (diffRsCsdiff < -101) {
return 1;
}
else if (diffRsCsdiff >= -101) {
if (ssDCval < 131) {
return 2;
}
else if (ssDCval >= 131) {
if (Rs < 1390) {
return 1;
}
else if (Rs >= 1390) {
return 2;
}
}
}
}
}
}
else if (Cs >= 320) {
if (refDCval < 57) {
if (posBalance < 64) {
if (Cs < 351) {
return 1;
}
else if (Cs >= 351) {
return 2;
}
}
else if (posBalance >= 64) {
if (Rs < 383) {
return 1;
}
else if (Rs >= 383) {
return 8;
}
}
}
else if (refDCval >= 57) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
if (Rs < 2013) {
return 4;
}
else if (Rs >= 2013) {
return 8;
}
}
}
}
}
else if (negBalance >= 60) {
if (diffTSC < -1) {
if (gchDC < 2) {
if (Rs < 1113) {
return 4;
}
else if (Rs >= 1113) {
return 2;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
}
else if (diffAFD >= 7) {
if (diffTSC < 7) {
if (diffAFD < 10) {
if (RsDiff < 12) {
return 1;
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (diffAFD >= 10) {
return 1;
}
}
else if (diffTSC >= 7) {
if (Rs < 846) {
return 1;
}
else if (Rs >= 846) {
return 8;
}
}
}
}
else if (tcor >= 62) {
if (mu_mv_mag_sq < 61) {
if (AFD < 19) {
if (ssDCval < 116) {
return 1;
}
else if (ssDCval >= 116) {
return 4;
}
}
else if (AFD >= 19) {
if (AFD < 24) {
if (diffMVdiffVal < 16) {
if (mcTcor < 85) {
return 1;
}
else if (mcTcor >= 85) {
if (mu_mv_mag_sq < 37) {
return 2;
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 1547) {
return 1;
}
else if (Rs >= 1547) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 16) {
if (Rs < 1577) {
return 1;
}
else if (Rs >= 1577) {
return 4;
}
}
}
else if (AFD >= 24) {
if (Rs < 345) {
return 4;
}
else if (Rs >= 345) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
if (diffMVdiffVal < -56) {
return 2;
}
else if (diffMVdiffVal >= -56) {
if (refDCval < 149) {
return 1;
}
else if (refDCval >= 149) {
if (Rs < 1780) {
return 1;
}
else if (Rs >= 1780) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 393) {
if (MVDiff < 20) {
if (posBalance < 55) {
if (negBalance < 52) {
if (diffRsCsdiff < 11) {
if (TSC < 6) {
if (Rs < 746) {
return 2;
}
else if (Rs >= 746) {
return 1;
}
}
else if (TSC >= 6) {
return 2;
}
}
else if (diffRsCsdiff >= 11) {
return 1;
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (posBalance >= 55) {
if (mu_mv_mag_sq < 21) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (posBalance < 62) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
else if (posBalance >= 62) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (Cs < 470) {
if (Rs < 1433) {
return 8;
}
else if (Rs >= 1433) {
return 2;
}
}
else if (Cs >= 470) {
return 4;
}
}
}
}
else if (MVDiff >= 20) {
if (refDCval < 140) {
if (posBalance < 52) {
if (ssDCval < 59) {
if (Rs < 946) {
if (Cs < 434) {
return 2;
}
else if (Cs >= 434) {
if (Rs < 621) {
return 4;
}
else if (Rs >= 621) {
return 8;
}
}
}
else if (Rs >= 946) {
return 2;
}
}
else if (ssDCval >= 59) {
if (tcor < 66) {
if (SC < 1027) {
if (Rs < 498) {
return 4;
}
else if (Rs >= 498) {
return 2;
}
}
else if (SC >= 1027) {
if (AFD < 30) {
return 1;
}
else if (AFD >= 30) {
return 4;
}
}
}
else if (tcor >= 66) {
if (Rs < 933) {
return 4;
}
else if (Rs >= 933) {
return 2;
}
}
}
}
else if (posBalance >= 52) {
if (RsCsDiff < 277) {
if (mcTcor < 87) {
if (tcor < 40) {
if (AFD < 17) {
if (ssDCval < 40) {
return 8;
}
else if (ssDCval >= 40) {
if (mu_mv_mag_sq < 75) {
return 1;
}
else if (mu_mv_mag_sq >= 75) {
return 4;
}
}
}
else if (AFD >= 17) {
if (AFD < 25) {
if (mu_mv_mag_sq < 37) {
if (Cs < 485) {
return 8;
}
else if (Cs >= 485) {
return 4;
}
}
else if (mu_mv_mag_sq >= 37) {
return 8;
}
}
else if (AFD >= 25) {
if (Cs < 488) {
if (refDCval < 70) {
return 2;
}
else if (refDCval >= 70) {
return 4;
}
}
else if (Cs >= 488) {
return 1;
}
}
}
}
else if (tcor >= 40) {
if (Rs < 1869) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (tcor < 44) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (tcor >= 44) {
if (Rs < 1448) {
if (ssDCval < 48) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
else if (ssDCval >= 48) {
if (negBalance < 57) {
return 2;
}
else if (negBalance >= 57) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
if (Cs < 420) {
return 4;
}
else if (Cs >= 420) {
return 2;
}
}
}
}
}
else if (Rs >= 1448) {
return 4;
}
}
}
}
else if (Rs >= 1869) {
return 8;
}
}
}
else if (mcTcor >= 87) {
if (gchDC < 9) {
if (MVDiff < 48) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
if (Rs < 1637) {
if (RsCsDiff < 60) {
if (mcTcor < 91) {
if (posBalance < 58) {
return 1;
}
else if (posBalance >= 58) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
}
else if (mcTcor >= 91) {
return 8;
}
}
else if (RsCsDiff >= 60) {
if (MVDiff < 31) {
if (AFD < 23) {
if (Rs < 519) {
return 4;
}
else if (Rs >= 519) {
return 8;
}
}
else if (AFD >= 23) {
return 4;
}
}
else if (MVDiff >= 31) {
return 8;
}
}
}
else if (Rs >= 1637) {
return 1;
}
}
}
else if (MVDiff >= 48) {
if (posBalance < 65) {
if (gchDC < 2) {
if (diffAFD < -2) {
if (Rs < 812) {
return 1;
}
else if (Rs >= 812) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Cs < 404) {
return 2;
}
else if (Cs >= 404) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (tcor < 64) {
if (Cs < 462) {
return 1;
}
else if (Cs >= 462) {
return 2;
}
}
else if (tcor >= 64) {
if (TSC < 9) {
if (mu_mv_mag_sq < 14) {
return 4;
}
else if (mu_mv_mag_sq >= 14) {
return 1;
}
}
else if (TSC >= 9) {
return 4;
}
}
}
}
else if (posBalance >= 65) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (Rs < 472) {
return 2;
}
else if (Rs >= 472) {
return 4;
}
}
}
}
}
else if (gchDC >= 9) {
return 8;
}
}
}
else if (RsCsDiff >= 277) {
return 2;
}
}
}
else if (refDCval >= 140) {
if (MVDiff < 273) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 1;
}
}
else if (MVDiff >= 273) {
return 8;
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 81) {
if (RsDiff < 9) {
if (Cs < 293) {
if (refDCval < 139) {
if (refDCval < 79) {
if (negBalance < 31) {
if (ssDCval < 58) {
return 2;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (negBalance >= 31) {
if (Cs < 199) {
return 1;
}
else if (Cs >= 199) {
if (Rs < 795) {
if (diffAFD < -1) {
if (negBalance < 51) {
if (Cs < 263) {
return 2;
}
else if (Cs >= 263) {
return 1;
}
}
else if (negBalance >= 51) {
return 8;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (Rs >= 795) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
if (Rs < 1520) {
return 4;
}
else if (Rs >= 1520) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 79) {
if (MVDiff < 22) {
if (RsDiff < 5) {
if (Cs < 277) {
return 1;
}
else if (Cs >= 277) {
return 4;
}
}
else if (RsDiff >= 5) {
if (ssDCval < 117) {
if (diffAFD < 1) {
if (AFD < 15) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 2;
}
}
else if (AFD >= 15) {
return 1;
}
}
else if (diffAFD >= 1) {
if (ssDCval < 97) {
if (diffMVdiffVal < -4) {
return 2;
}
else if (diffMVdiffVal >= -4) {
if (diffAFD < 4) {
return 1;
}
else if (diffAFD >= 4) {
if (Cs < 87) {
return 1;
}
else if (Cs >= 87) {
return 2;
}
}
}
}
else if (ssDCval >= 97) {
return 2;
}
}
}
else if (ssDCval >= 117) {
return 1;
}
}
}
else if (MVDiff >= 22) {
if (Rs < 786) {
if (Rs < 776) {
if (mcTcor < 95) {
if (MVDiff < 119) {
return 1;
}
else if (MVDiff >= 119) {
return 2;
}
}
else if (mcTcor >= 95) {
if (Cs < 214) {
if (RsDiff < 5) {
if (MVDiff < 29) {
return 1;
}
else if (MVDiff >= 29) {
return 8;
}
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (Cs >= 214) {
return 1;
}
}
}
else if (Rs >= 776) {
return 4;
}
}
else if (Rs >= 786) {
if (mu_mv_mag_sq < 3) {
if (SC < 2268) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 2) {
if (Cs < 101) {
return 2;
}
else if (Cs >= 101) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1438) {
return 2;
}
else if (Rs >= 1438) {
return 1;
}
}
}
}
else if (SC >= 2268) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 1114) {
if (Cs < 158) {
if (posBalance < 102) {
return 1;
}
else if (posBalance >= 102) {
return 2;
}
}
else if (Cs >= 158) {
return 2;
}
}
else if (SC >= 1114) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 139) {
if (gchDC < 2) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 4;
}
}
else if (gchDC >= 2) {
if (negBalance < 81) {
return 1;
}
else if (negBalance >= 81) {
if (Rs < 684) {
return 1;
}
else if (Rs >= 684) {
return 2;
}
}
}
}
}
else if (Cs >= 293) {
if (Rs < 783) {
if (ssDCval < 80) {
if (AFD < 10) {
if (tcor < 91) {
if (negBalance < 31) {
if (Rs < 648) {
return 8;
}
else if (Rs >= 648) {
return 2;
}
}
else if (negBalance >= 31) {
return 8;
}
}
else if (tcor >= 91) {
return 1;
}
}
else if (AFD >= 10) {
if (SC < 1176) {
if (MVDiff < 19) {
if (MVDiff < 10) {
if (Rs < 650) {
return 8;
}
else if (Rs >= 650) {
return 1;
}
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (MVDiff >= 19) {
if (negBalance < 55) {
if (tcor < 79) {
if (Rs < 398) {
if (Rs < 364) {
return 4;
}
else if (Rs >= 364) {
return 2;
}
}
else if (Rs >= 398) {
return 1;
}
}
else if (tcor >= 79) {
return 4;
}
}
else if (negBalance >= 55) {
if (negBalance < 58) {
return 8;
}
else if (negBalance >= 58) {
return 1;
}
}
}
}
else if (SC >= 1176) {
if (Rs < 761) {
return 4;
}
else if (Rs >= 761) {
return 8;
}
}
}
}
else if (ssDCval >= 80) {
if (MVDiff < 21) {
if (mcTcor < 97) {
if (Rs < 677) {
return 2;
}
else if (Rs >= 677) {
return 1;
}
}
else if (mcTcor >= 97) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
else if (MVDiff >= 21) {
if (mu_mv_mag_sq < 52) {
if (diffAFD < -4) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
else if (diffAFD >= -4) {
if (mu_mv_mag_sq < 9) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 8;
}
}
else if (mu_mv_mag_sq >= 9) {
if (AFD < 23) {
if (Cs < 365) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
return 8;
}
}
else if (Cs >= 365) {
return 8;
}
}
else if (AFD >= 23) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
return 1;
}
}
}
}
else if (Rs >= 783) {
if (AFD < 10) {
if (ssDCval < 97) {
if (refDCval < 72) {
if (diffAFD < -1) {
if (MVDiff < 9) {
if (posBalance < 46) {
if (Rs < 1846) {
return 8;
}
else if (Rs >= 1846) {
return 1;
}
}
else if (posBalance >= 46) {
return 2;
}
}
else if (MVDiff >= 9) {
if (negBalance < 45) {
if (Cs < 526) {
return 1;
}
else if (Cs >= 526) {
if (Rs < 2048) {
return 4;
}
else if (Rs >= 2048) {
return 2;
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (posBalance < 46) {
if (mu_mv_mag_sq < 58) {
if (diffRsCsdiff < 6) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (diffRsCsdiff >= 6) {
if (diffRsCsdiff < 24) {
return 4;
}
else if (diffRsCsdiff >= 24) {
if (Rs < 1616) {
return 2;
}
else if (Rs >= 1616) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (Cs < 512) {
return 1;
}
else if (Cs >= 512) {
return 2;
}
}
}
else if (posBalance >= 46) {
if (ssDCval < 56) {
if (Rs < 1009) {
return 2;
}
else if (Rs >= 1009) {
return 1;
}
}
else if (ssDCval >= 56) {
if (mcTcor < 93) {
if (Rs < 933) {
return 2;
}
else if (Rs >= 933) {
return 4;
}
}
else if (mcTcor >= 93) {
if (RsCsDiff < 23) {
return 1;
}
else if (RsCsDiff >= 23) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 72) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
if (Cs < 433) {
return 1;
}
else if (Cs >= 433) {
return 2;
}
}
}
}
else if (ssDCval >= 97) {
if (Cs < 445) {
if (negBalance < 37) {
if (Rs < 1751) {
return 1;
}
else if (Rs >= 1751) {
return 8;
}
}
else if (negBalance >= 37) {
if (tcor < 85) {
return 8;
}
else if (tcor >= 85) {
if (MVDiff < 34) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (MVDiff >= 34) {
return 1;
}
}
}
}
else if (Cs >= 445) {
if (ssDCval < 135) {
if (ssDCval < 121) {
if (Rs < 1818) {
return 8;
}
else if (Rs >= 1818) {
return 1;
}
}
else if (ssDCval >= 121) {
return 2;
}
}
else if (ssDCval >= 135) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
else if (AFD >= 10) {
if (MVDiff < 88) {
if (ssDCval < 66) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 7) {
if (Cs < 398) {
if (Rs < 1629) {
return 1;
}
else if (Rs >= 1629) {
return 2;
}
}
else if (Cs >= 398) {
return 4;
}
}
else if (mu_mv_mag_sq >= 7) {
return 1;
}
}
else if (CsDiff >= 3) {
if (tcor < 89) {
if (gchDC < 3) {
if (Cs < 507) {
if (TSC < 5) {
if (Rs < 1801) {
return 2;
}
else if (Rs >= 1801) {
return 4;
}
}
else if (TSC >= 5) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
if (Rs < 1587) {
return 1;
}
else if (Rs >= 1587) {
return 2;
}
}
}
}
else if (Cs >= 507) {
return 4;
}
}
else if (gchDC >= 3) {
if (Cs < 511) {
return 1;
}
else if (Cs >= 511) {
return 2;
}
}
}
else if (tcor >= 89) {
if (Rs < 1389) {
return 8;
}
else if (Rs >= 1389) {
return 1;
}
}
}
}
else if (ssDCval >= 66) {
if (mcTcor < 91) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 4;
}
}
}
else if (mcTcor >= 91) {
if (Rs < 2319) {
if (gchDC < 3) {
if (ssDCval < 84) {
if (diffRsCsdiff < 42) {
return 1;
}
else if (diffRsCsdiff >= 42) {
return 2;
}
}
else if (ssDCval >= 84) {
if (mcTcor < 96) {
if (RsCsDiff < 43) {
if (diffRsCsdiff < 16) {
return 1;
}
else if (diffRsCsdiff >= 16) {
return 4;
}
}
else if (RsCsDiff >= 43) {
if (Rs < 1765) {
return 2;
}
else if (Rs >= 1765) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (MVDiff < 45) {
if (Rs < 2308) {
if (posBalance < 74) {
if (Rs < 972) {
return 2;
}
else if (Rs >= 972) {
return 1;
}
}
else if (posBalance >= 74) {
return 2;
}
}
else if (Rs >= 2308) {
return 2;
}
}
else if (MVDiff >= 45) {
return 2;
}
}
}
}
}
else if (gchDC >= 3) {
if (refDCval < 67) {
if (Rs < 1286) {
return 4;
}
else if (Rs >= 1286) {
return 1;
}
}
else if (refDCval >= 67) {
return 1;
}
}
}
else if (Rs >= 2319) {
return 8;
}
}
}
}
else if (MVDiff >= 88) {
if (tcor < 91) {
if (Rs < 1048) {
return 1;
}
else if (Rs >= 1048) {
return 8;
}
}
else if (tcor >= 91) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 113) {
if (Rs < 1930) {
if (negBalance < 29) {
return 2;
}
else if (negBalance >= 29) {
if (AFD < 12) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
return 2;
}
}
else if (AFD >= 12) {
if (MVDiff < 37) {
if (diffMVdiffVal < -10) {
if (ssDCval < 109) {
return 1;
}
else if (ssDCval >= 109) {
return 2;
}
}
else if (diffMVdiffVal >= -10) {
if (CsDiff < 3) {
if (diffAFD < 8) {
return 2;
}
else if (diffAFD >= 8) {
return 1;
}
}
else if (CsDiff >= 3) {
if (diffMVdiffVal < -2) {
if (Rs < 1917) {
return 2;
}
else if (Rs >= 1917) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
if (gchDC < 5) {
if (negBalance < 64) {
if (posBalance < 75) {
return 1;
}
else if (posBalance >= 75) {
return 2;
}
}
else if (negBalance >= 64) {
return 2;
}
}
else if (gchDC >= 5) {
return 2;
}
}
}
}
}
else if (MVDiff >= 37) {
if (gchDC < 30) {
if (refDCval < 96) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
if (Cs < 208) {
if (diffMVdiffVal < -8) {
return 8;
}
else if (diffMVdiffVal >= -8) {
return 1;
}
}
else if (Cs >= 208) {
if (Rs < 1700) {
return 2;
}
else if (Rs >= 1700) {
return 8;
}
}
}
}
else if (refDCval >= 96) {
return 1;
}
}
else if (gchDC >= 30) {
return 2;
}
}
}
}
}
else if (Rs >= 1930) {
if (refDCval < 82) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 2;
}
}
else if (refDCval >= 82) {
if (Cs < 381) {
if (mcTcor < 90) {
if (Rs < 2303) {
return 4;
}
else if (Rs >= 2303) {
return 8;
}
}
else if (mcTcor >= 90) {
if (mu_mv_mag_sq < 23) {
if (posBalance < 58) {
if (Rs < 2191) {
if (diffMVdiffVal < -46) {
return 1;
}
else if (diffMVdiffVal >= -46) {
return 2;
}
}
else if (Rs >= 2191) {
if (RsDiff < 12) {
return 1;
}
else if (RsDiff >= 12) {
return 2;
}
}
}
else if (posBalance >= 58) {
return 1;
}
}
else if (mu_mv_mag_sq >= 23) {
return 2;
}
}
}
else if (Cs >= 381) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 38) {
return 1;
}
else if (mu_mv_mag_sq >= 38) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 113) {
if (TSC < 6) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
return 1;
}
}
else if (TSC >= 6) {
if (mcTcor < 96) {
if (CsDiff < 3) {
if (MVDiff < 97) {
if (Rs < 1316) {
return 4;
}
else if (Rs >= 1316) {
if (CsDiff < 2) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 8;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
}
else if (MVDiff >= 97) {
if (Rs < 1658) {
return 1;
}
else if (Rs >= 1658) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (RsCsDiff < 94) {
if (Cs < 466) {
return 1;
}
else if (Cs >= 466) {
return 8;
}
}
else if (RsCsDiff >= 94) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (AFD < 33) {
if (refDCval < 117) {
if (mu_mv_mag_sq < 31) {
return 1;
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
else if (refDCval >= 117) {
return 1;
}
}
else if (AFD >= 33) {
if (Cs < 436) {
return 1;
}
else if (Cs >= 436) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 53) {
if (posBalance < 65) {
if (ssDCval < 59) {
if (Rs < 1112) {
return 4;
}
else if (Rs >= 1112) {
return 1;
}
}
else if (ssDCval >= 59) {
if (Cs < 519) {
return 8;
}
else if (Cs >= 519) {
return 1;
}
}
}
else if (posBalance >= 65) {
if (MVDiff < 119) {
if (Rs < 729) {
return 1;
}
else if (Rs >= 729) {
return 2;
}
}
else if (MVDiff >= 119) {
return 1;
}
}
}
else if (negBalance >= 53) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 535) {
if (AFD < 13) {
if (mcTcor < 96) {
if (posBalance < 43) {
if (refDCval < 47) {
if (TSCindex < 3) {
if (Rs < 644) {
if (Rs < 632) {
return 8;
}
else if (Rs >= 632) {
return 1;
}
}
else if (Rs >= 644) {
return 8;
}
}
else if (TSCindex >= 3) {
if (MVDiff < 7) {
if (negBalance < 31) {
if (negBalance < 15) {
return 1;
}
else if (negBalance >= 15) {
if (AFD < 4) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (negBalance >= 31) {
if (Rs < 1482) {
if (Cs < 672) {
if (Rs < 1448) {
return 4;
}
else if (Rs >= 1448) {
return 4;
}
}
else if (Cs >= 672) {
return 8;
}
}
else if (Rs >= 1482) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 28) {
if (Rs < 364) {
return 8;
}
else if (Rs >= 364) {
return 4;
}
}
else if (ssDCval >= 28) {
if (mu_mv_mag_sq < 73) {
if (Rs < 1664) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
if (Rs < 729) {
if (Rs < 541) {
return 1;
}
else if (Rs >= 541) {
return 4;
}
}
else if (Rs >= 729) {
if (Rs < 883) {
return 8;
}
else if (Rs >= 883) {
return 2;
}
}
}
}
else if (Rs >= 1664) {
if (Cs < 827) {
return 4;
}
else if (Cs >= 827) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 73) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 68) {
if (Cs < 734) {
if (refDCval < 94) {
if (Cs < 593) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 27) {
return 1;
}
else if (mu_mv_mag_sq >= 27) {
return 8;
}
}
}
else if (Cs >= 593) {
if (diffMVdiffVal < -7) {
if (Rs < 855) {
return 1;
}
else if (Rs >= 855) {
return 2;
}
}
else if (diffMVdiffVal >= -7) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
else if (refDCval >= 94) {
if (diffMVdiffVal < 3) {
if (Cs < 710) {
return 2;
}
else if (Cs >= 710) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 955) {
return 8;
}
else if (Rs >= 955) {
return 2;
}
}
}
}
else if (Cs >= 734) {
if (Cs < 1155) {
if (gchDC < 3) {
if (Rs < 1905) {
if (mcTcor < 93) {
if (ssDCval < 152) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
return 2;
}
}
else if (ssDCval >= 152) {
return 2;
}
}
else if (mcTcor >= 93) {
if (diffTSC < -1) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
else if (diffTSC >= -1) {
if (Cs < 947) {
if (MVDiff < 25) {
if (diffRsCsdiff < 13) {
return 2;
}
else if (diffRsCsdiff >= 13) {
return 1;
}
}
else if (MVDiff >= 25) {
return 4;
}
}
else if (Cs >= 947) {
if (gchDC < 1) {
if (Rs < 1259) {
return 1;
}
else if (Rs >= 1259) {
return 8;
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
}
else if (Rs >= 1905) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
else if (Cs >= 1155) {
if (Cs < 1279) {
return 2;
}
else if (Cs >= 1279) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (mu_mv_mag_sq < 101) {
if (ssDCval < 73) {
if (TSC < 3) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
return 4;
}
}
else if (TSC >= 3) {
if (refDCval < 60) {
if (Cs < 909) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
if (Rs < 1046) {
return 8;
}
else if (Rs >= 1046) {
return 1;
}
}
}
else if (Cs >= 909) {
if (Rs < 2074) {
return 8;
}
else if (Rs >= 2074) {
return 4;
}
}
}
else if (refDCval >= 60) {
if (mu_mv_mag_sq < 79) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (mu_mv_mag_sq >= 79) {
if (TSC < 7) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
else if (ssDCval >= 73) {
if (Cs < 805) {
if (Rs < 1869) {
return 8;
}
else if (Rs >= 1869) {
return 8;
}
}
else if (Cs >= 805) {
if (negBalance < 30) {
if (refDCval < 139) {
return 4;
}
else if (refDCval >= 139) {
return 8;
}
}
else if (negBalance >= 30) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (RsDiff < 4) {
if (negBalance < 36) {
if (MVDiff < 27) {
if (Rs < 1136) {
if (ssDCval < 153) {
if (refDCval < 56) {
if (Rs < 901) {
return 8;
}
else if (Rs >= 901) {
return 1;
}
}
else if (refDCval >= 56) {
if (posBalance < 33) {
if (ssDCval < 138) {
return 1;
}
else if (ssDCval >= 138) {
return 8;
}
}
else if (posBalance >= 33) {
return 8;
}
}
}
else if (ssDCval >= 153) {
return 2;
}
}
else if (Rs >= 1136) {
if (refDCval < 48) {
return 4;
}
else if (refDCval >= 48) {
if (diffMVdiffVal < -19) {
return 4;
}
else if (diffMVdiffVal >= -19) {
return 8;
}
}
}
}
else if (MVDiff >= 27) {
if (diffRsCsdiff < -182) {
return 8;
}
else if (diffRsCsdiff >= -182) {
return 4;
}
}
}
else if (negBalance >= 36) {
if (Rs < 1128) {
return 8;
}
else if (Rs >= 1128) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
}
else if (RsDiff >= 4) {
if (Cs < 1037) {
if (SC < 2907) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (MVDiff < 5) {
if (Cs < 724) {
return 8;
}
else if (Cs >= 724) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (MVDiff >= 5) {
return 1;
}
}
}
else if (SC >= 2907) {
return 2;
}
}
else if (Cs >= 1037) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 43) {
if (mu_mv_mag_sq < 32) {
if (Cs < 708) {
if (Rs < 1341) {
if (Rs < 404) {
if (diffRsCsdiff < -4) {
if (Rs < 278) {
return 8;
}
else if (Rs >= 278) {
return 2;
}
}
else if (diffRsCsdiff >= -4) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (Rs >= 404) {
if (ssDCval < 94) {
if (refDCval < 89) {
if (Rs < 508) {
if (diffAFD < -1) {
if (negBalance < 55) {
return 2;
}
else if (negBalance >= 55) {
return 4;
}
}
else if (diffAFD >= -1) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
else if (Rs >= 508) {
if (ssDCval < 41) {
if (negBalance < 37) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (negBalance >= 37) {
if (mcTcor < 87) {
if (Rs < 595) {
return 2;
}
else if (Rs >= 595) {
return 8;
}
}
else if (mcTcor >= 87) {
if (tcor < 87) {
if (MVDiff < 33) {
return 2;
}
else if (MVDiff >= 33) {
return 8;
}
}
else if (tcor >= 87) {
return 8;
}
}
}
}
else if (ssDCval >= 41) {
if (refDCval < 72) {
if (TSC < 6) {
if (MVDiff < 26) {
if (ssDCval < 45) {
if (SC < 1436) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (SC >= 1436) {
return 8;
}
}
else if (ssDCval >= 45) {
if (ssDCval < 51) {
if (Rs < 950) {
return 1;
}
else if (Rs >= 950) {
return 8;
}
}
else if (ssDCval >= 51) {
if (MVDiff < 7) {
if (Rs < 1028) {
return 4;
}
else if (Rs >= 1028) {
if (Rs < 1089) {
return 1;
}
else if (Rs >= 1089) {
return 8;
}
}
}
else if (MVDiff >= 7) {
if (ssDCval < 64) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (ssDCval >= 64) {
return 1;
}
}
}
}
}
else if (MVDiff >= 26) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 62) {
if (Rs < 856) {
if (AFD < 12) {
if (SC < 1484) {
if (TSC < 7) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
if (Rs < 591) {
return 8;
}
else if (Rs >= 591) {
return 1;
}
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (SC >= 1484) {
return 1;
}
}
else if (AFD >= 12) {
if (MVDiff < 26) {
if (Cs < 583) {
return 8;
}
else if (Cs >= 583) {
return 4;
}
}
else if (MVDiff >= 26) {
if (Rs < 552) {
return 8;
}
else if (Rs >= 552) {
return 1;
}
}
}
}
else if (Rs >= 856) {
return 8;
}
}
else if (ssDCval >= 62) {
if (Rs < 916) {
return 2;
}
else if (Rs >= 916) {
return 4;
}
}
}
}
else if (refDCval >= 72) {
if (RsCsDiff < 38) {
if (negBalance < 43) {
if (Cs < 560) {
return 8;
}
else if (Cs >= 560) {
return 2;
}
}
else if (negBalance >= 43) {
if (Cs < 575) {
if (Rs < 1197) {
if (Cs < 570) {
return 8;
}
else if (Cs >= 570) {
return 2;
}
}
else if (Rs >= 1197) {
return 2;
}
}
else if (Cs >= 575) {
return 8;
}
}
}
else if (RsCsDiff >= 38) {
if (Rs < 1072) {
return 1;
}
else if (Rs >= 1072) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 89) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (Rs < 814) {
return 8;
}
else if (Rs >= 814) {
return 4;
}
}
}
}
else if (ssDCval >= 94) {
if (MVDiff < 6) {
if (mcTcor < 93) {
if (Rs < 618) {
return 1;
}
else if (Rs >= 618) {
return 8;
}
}
else if (mcTcor >= 93) {
return 1;
}
}
else if (MVDiff >= 6) {
if (diffRsCsdiff < -7) {
if (CsDiff < 4) {
if (negBalance < 50) {
if (Rs < 1169) {
if (Rs < 649) {
return 8;
}
else if (Rs >= 649) {
return 2;
}
}
else if (Rs >= 1169) {
return 8;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (CsDiff >= 4) {
if (Rs < 965) {
return 2;
}
else if (Rs >= 965) {
return 1;
}
}
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
}
}
}
else if (Rs >= 1341) {
if (mcTcor < 90) {
if (ssDCval < 124) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Cs < 585) {
return 2;
}
else if (Cs >= 585) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (ssDCval >= 124) {
return 4;
}
}
else if (mcTcor >= 90) {
if (Rs < 2129) {
if (posBalance < 56) {
if (Cs < 685) {
if (diffRsCsdiff < -43) {
return 8;
}
else if (diffRsCsdiff >= -43) {
if (negBalance < 58) {
return 1;
}
else if (negBalance >= 58) {
if (Rs < 1670) {
return 1;
}
else if (Rs >= 1670) {
return 2;
}
}
}
}
else if (Cs >= 685) {
return 2;
}
}
else if (posBalance >= 56) {
if (ssDCval < 107) {
if (SC < 2234) {
if (Rs < 1607) {
if (Rs < 1526) {
return 4;
}
else if (Rs >= 1526) {
return 2;
}
}
else if (Rs >= 1607) {
return 1;
}
}
else if (SC >= 2234) {
return 2;
}
}
else if (ssDCval >= 107) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
else if (Rs >= 2129) {
if (Rs < 2365) {
return 2;
}
else if (Rs >= 2365) {
return 1;
}
}
}
}
}
else if (Cs >= 708) {
if (negBalance < 35) {
if (tcor < 85) {
if (Cs < 969) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 515) {
return 8;
}
else if (Rs >= 515) {
return 2;
}
}
}
else if (Cs >= 969) {
if (Rs < 1210) {
if (gchDC < 3) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
if (Rs < 962) {
return 1;
}
else if (Rs >= 962) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
if (Rs < 741) {
return 2;
}
else if (Rs >= 741) {
return 8;
}
}
}
else if (Rs >= 1210) {
if (ssDCval < 61) {
if (Rs < 1562) {
return 1;
}
else if (Rs >= 1562) {
if (Rs < 1973) {
return 8;
}
else if (Rs >= 1973) {
return 4;
}
}
}
else if (ssDCval >= 61) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
}
else if (tcor >= 85) {
if (posBalance < 58) {
if (mcTcor < 95) {
if (tcor < 88) {
if (Rs < 569) {
return 2;
}
else if (Rs >= 569) {
return 8;
}
}
else if (tcor >= 88) {
if (Rs < 985) {
return 4;
}
else if (Rs >= 985) {
return 2;
}
}
}
else if (mcTcor >= 95) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (posBalance >= 58) {
if (MVDiff < 5) {
if (Rs < 1287) {
return 1;
}
else if (Rs >= 1287) {
return 2;
}
}
else if (MVDiff >= 5) {
if (ssDCval < 139) {
return 8;
}
else if (ssDCval >= 139) {
return 2;
}
}
}
}
}
else if (negBalance >= 35) {
if (CsDiff < 5) {
if (refDCval < 52) {
if (refDCval < 42) {
if (Rs < 844) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (Cs < 721) {
return 8;
}
else if (Cs >= 721) {
return 2;
}
}
}
else if (Rs >= 844) {
return 8;
}
}
else if (refDCval >= 42) {
if (negBalance < 52) {
if (diffRsCsdiff < -13) {
if (Rs < 1197) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
else if (Rs >= 1197) {
return 8;
}
}
else if (diffRsCsdiff >= -13) {
if (Cs < 1134) {
return 8;
}
else if (Cs >= 1134) {
if (Rs < 1202) {
return 4;
}
else if (Rs >= 1202) {
return 8;
}
}
}
}
else if (negBalance >= 52) {
if (mcTcor < 78) {
if (refDCval < 46) {
if (Cs < 769) {
if (Rs < 810) {
return 8;
}
else if (Rs >= 810) {
return 1;
}
}
else if (Cs >= 769) {
return 8;
}
}
else if (refDCval >= 46) {
return 8;
}
}
else if (mcTcor >= 78) {
if (refDCval < 48) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (refDCval >= 48) {
return 8;
}
}
}
}
}
else if (refDCval >= 52) {
if (MVDiff < 6) {
if (tcor < 91) {
if (refDCval < 73) {
if (RsCsDiff < 38) {
if (posBalance < 46) {
if (Rs < 914) {
return 8;
}
else if (Rs >= 914) {
return 4;
}
}
else if (posBalance >= 46) {
if (tcor < 74) {
return 4;
}
else if (tcor >= 74) {
if (Cs < 1083) {
return 8;
}
else if (Cs >= 1083) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 38) {
if (ssDCval < 62) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (ssDCval >= 62) {
if (Rs < 1972) {
return 1;
}
else if (Rs >= 1972) {
return 2;
}
}
}
}
else if (refDCval >= 73) {
if (negBalance < 42) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
if (Rs < 2012) {
return 2;
}
else if (Rs >= 2012) {
return 1;
}
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
else if (tcor >= 91) {
if (posBalance < 56) {
if (posBalance < 47) {
if (Rs < 2338) {
return 2;
}
else if (Rs >= 2338) {
return 8;
}
}
else if (posBalance >= 47) {
return 8;
}
}
else if (posBalance >= 56) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (Rs < 2053) {
return 8;
}
else if (Rs >= 2053) {
return 1;
}
}
}
}
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 2) {
if (tcor < 81) {
if (tcor < 68) {
if (Rs < 1024) {
return 4;
}
else if (Rs >= 1024) {
return 1;
}
}
else if (tcor >= 68) {
if (ssDCval < 95) {
if (Cs < 1182) {
if (refDCval < 91) {
if (Rs < 791) {
if (Rs < 776) {
if (tcor < 73) {
return 4;
}
else if (tcor >= 73) {
return 8;
}
}
else if (Rs >= 776) {
return 2;
}
}
else if (Rs >= 791) {
return 8;
}
}
else if (refDCval >= 91) {
return 4;
}
}
else if (Cs >= 1182) {
if (Rs < 1507) {
return 1;
}
else if (Rs >= 1507) {
return 4;
}
}
}
else if (ssDCval >= 95) {
if (ssDCval < 99) {
if (mcTcor < 85) {
if (AFD < 10) {
if (Rs < 1766) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
return 8;
}
}
else if (Rs >= 1766) {
return 8;
}
}
else if (AFD >= 10) {
return 8;
}
}
else if (mcTcor >= 85) {
return 2;
}
}
else if (ssDCval >= 99) {
return 8;
}
}
}
}
else if (tcor >= 81) {
if (posBalance < 55) {
if (gchDC < 1) {
if (RsDiff < 4) {
if (Rs < 1004) {
return 2;
}
else if (Rs >= 1004) {
return 8;
}
}
else if (RsDiff >= 4) {
if (Cs < 760) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
else if (Cs >= 760) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (Cs < 1163) {
if (AFD < 11) {
if (Rs < 962) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (Rs >= 962) {
if (ssDCval < 146) {
if (MVDiff < 8) {
if (Cs < 972) {
return 2;
}
else if (Cs >= 972) {
return 4;
}
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (ssDCval >= 146) {
return 2;
}
}
}
else if (AFD >= 11) {
if (tcor < 88) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
return 4;
}
}
else if (tcor >= 88) {
if (Rs < 495) {
return 1;
}
else if (Rs >= 495) {
return 2;
}
}
}
}
else if (Cs >= 1163) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (Rs < 1194) {
if (posBalance < 59) {
if (Cs < 906) {
return 8;
}
else if (Cs >= 906) {
return 4;
}
}
else if (posBalance >= 59) {
if (diffAFD < -1) {
if (negBalance < 44) {
return 4;
}
else if (negBalance >= 44) {
if (Cs < 964) {
return 2;
}
else if (Cs >= 964) {
return 8;
}
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (Rs >= 1194) {
if (RsDiff < 7) {
if (AFD < 9) {
if (Cs < 832) {
if (ssDCval < 80) {
return 4;
}
else if (ssDCval >= 80) {
if (Rs < 1687) {
return 8;
}
else if (Rs >= 1687) {
return 2;
}
}
}
else if (Cs >= 832) {
return 8;
}
}
else if (AFD >= 9) {
return 8;
}
}
else if (RsDiff >= 7) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (ssDCval < 136) {
if (Rs < 699) {
if (Rs < 696) {
return 1;
}
else if (Rs >= 696) {
return 4;
}
}
else if (Rs >= 699) {
if (RsCsDiff < 23) {
if (diffAFD < 6) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 1164) {
return 8;
}
else if (Rs >= 1164) {
if (Rs < 1378) {
return 2;
}
else if (Rs >= 1378) {
return 8;
}
}
}
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (RsCsDiff >= 23) {
if (SC < 2535) {
if (tcor < 68) {
if (Rs < 985) {
return 4;
}
else if (Rs >= 985) {
return 1;
}
}
else if (tcor >= 68) {
if (posBalance < 60) {
if (negBalance < 54) {
if (SC < 2154) {
return 8;
}
else if (SC >= 2154) {
if (ssDCval < 94) {
return 4;
}
else if (ssDCval >= 94) {
return 8;
}
}
}
else if (negBalance >= 54) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
else if (SC >= 2535) {
if (posBalance < 55) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (Rs < 2118) {
return 1;
}
else if (Rs >= 2118) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
if (Rs < 1793) {
return 8;
}
else if (Rs >= 1793) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 136) {
if (Rs < 1949) {
return 4;
}
else if (Rs >= 1949) {
return 8;
}
}
}
}
}
}
else if (CsDiff >= 5) {
if (Rs < 1043) {
if (negBalance < 49) {
if (Cs < 955) {
if (ssDCval < 57) {
if (diffRsCsdiff < 12) {
return 2;
}
else if (diffRsCsdiff >= 12) {
if (Cs < 848) {
return 8;
}
else if (Cs >= 848) {
return 2;
}
}
}
else if (ssDCval >= 57) {
return 8;
}
}
else if (Cs >= 955) {
if (ssDCval < 68) {
if (ssDCval < 47) {
if (Cs < 1304) {
return 2;
}
else if (Cs >= 1304) {
return 4;
}
}
else if (ssDCval >= 47) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
if (TSC < 7) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (TSC >= 7) {
return 1;
}
}
}
}
else if (ssDCval >= 68) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
}
}
else if (negBalance >= 49) {
if (Rs < 415) {
return 1;
}
else if (Rs >= 415) {
if (Rs < 1008) {
if (negBalance < 53) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (negBalance >= 53) {
if (Rs < 888) {
if (mcTcor < 72) {
if (Cs < 898) {
return 8;
}
else if (Cs >= 898) {
return 4;
}
}
else if (mcTcor >= 72) {
return 4;
}
}
else if (Rs >= 888) {
return 8;
}
}
}
else if (Rs >= 1008) {
return 1;
}
}
}
}
else if (Rs >= 1043) {
if (SC < 2271) {
if (TSC < 8) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
return 2;
}
}
else if (TSC >= 8) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
return 1;
}
}
}
else if (SC >= 2271) {
if (ssDCval < 92) {
if (tcor < 79) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (tcor >= 79) {
if (refDCval < 59) {
return 4;
}
else if (refDCval >= 59) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
if (TSC < 9) {
if (gchDC < 2) {
if (Cs < 1114) {
return 4;
}
else if (Cs >= 1114) {
return 2;
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
else if (ssDCval >= 92) {
if (RsDiff < 8) {
if (ssDCval < 93) {
if (Rs < 1804) {
return 2;
}
else if (Rs >= 1804) {
return 8;
}
}
else if (ssDCval >= 93) {
if (MVDiff < 26) {
return 8;
}
else if (MVDiff >= 26) {
return 2;
}
}
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (ssDCval < 47) {
if (mu_mv_mag_sq < 114) {
if (Rs < 569) {
if (MVDiff < 35) {
return 1;
}
else if (MVDiff >= 35) {
return 2;
}
}
else if (Rs >= 569) {
if (ssDCval < 39) {
if (MVDiff < 5) {
if (Rs < 867) {
return 1;
}
else if (Rs >= 867) {
return 8;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (ssDCval >= 39) {
if (TSC < 4) {
if (Cs < 638) {
if (Rs < 1032) {
return 1;
}
else if (Rs >= 1032) {
return 2;
}
}
else if (Cs >= 638) {
return 8;
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 101) {
if (negBalance < 35) {
if (Rs < 1272) {
if (MVDiff < 24) {
return 8;
}
else if (MVDiff >= 24) {
return 2;
}
}
else if (Rs >= 1272) {
if (Rs < 1717) {
return 2;
}
else if (Rs >= 1717) {
return 4;
}
}
}
else if (negBalance >= 35) {
if (TSC < 10) {
if (Rs < 690) {
if (Rs < 678) {
return 8;
}
else if (Rs >= 678) {
return 4;
}
}
else if (Rs >= 690) {
if (TSC < 7) {
if (RsDiff < 5) {
if (diffRsCsdiff < -17) {
return 2;
}
else if (diffRsCsdiff >= -17) {
return 8;
}
}
else if (RsDiff >= 5) {
if (negBalance < 44) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
else if (negBalance >= 44) {
if (Rs < 1248) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (Rs >= 1248) {
return 8;
}
}
}
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (TSC >= 10) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (Cs < 821) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
return 1;
}
}
else if (Cs >= 821) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (negBalance < 41) {
if (Rs < 879) {
if (Rs < 437) {
if (Rs < 419) {
return 8;
}
else if (Rs >= 419) {
return 4;
}
}
else if (Rs >= 437) {
return 1;
}
}
else if (Rs >= 879) {
if (tcor < 87) {
if (Rs < 1032) {
return 8;
}
else if (Rs >= 1032) {
return 1;
}
}
else if (tcor >= 87) {
return 8;
}
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
else if (ssDCval >= 47) {
if (Rs < 1302) {
if (Rs < 381) {
if (ssDCval < 96) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
if (negBalance < 42) {
if (Cs < 692) {
return 4;
}
else if (Cs >= 692) {
return 8;
}
}
else if (negBalance >= 42) {
if (Cs < 613) {
return 8;
}
else if (Cs >= 613) {
return 1;
}
}
}
}
else if (ssDCval >= 96) {
if (Rs < 332) {
return 2;
}
else if (Rs >= 332) {
return 4;
}
}
}
else if (Rs >= 381) {
if (RsDiff < 3) {
if (ssDCval < 55) {
if (negBalance < 31) {
if (posBalance < 47) {
return 1;
}
else if (posBalance >= 47) {
if (RsCsDiff < 4) {
return 1;
}
else if (RsCsDiff >= 4) {
return 4;
}
}
}
else if (negBalance >= 31) {
if (Rs < 859) {
if (ssDCval < 51) {
if (Rs < 839) {
return 8;
}
else if (Rs >= 839) {
return 1;
}
}
else if (ssDCval >= 51) {
if (Cs < 827) {
return 4;
}
else if (Cs >= 827) {
if (Rs < 434) {
return 1;
}
else if (Rs >= 434) {
return 2;
}
}
}
}
else if (Rs >= 859) {
if (tcor < 90) {
if (ssDCval < 50) {
if (Rs < 911) {
return 8;
}
else if (Rs >= 911) {
return 1;
}
}
else if (ssDCval >= 50) {
return 8;
}
}
else if (tcor >= 90) {
if (Cs < 1328) {
return 8;
}
else if (Cs >= 1328) {
return 4;
}
}
}
}
}
else if (ssDCval >= 55) {
if (Cs < 1309) {
if (mcTcor < 94) {
if (refDCval < 131) {
if (diffTSC < 1) {
if (Cs < 1048) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 643) {
if (TSC < 6) {
if (Cs < 649) {
return 8;
}
else if (Cs >= 649) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (Cs < 704) {
return 8;
}
else if (Cs >= 704) {
return 2;
}
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (Rs >= 643) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 835) {
return 8;
}
else if (Rs >= 835) {
return 1;
}
}
}
}
}
else if (Cs >= 1048) {
if (refDCval < 61) {
if (Cs < 1158) {
if (Rs < 545) {
return 1;
}
else if (Rs >= 545) {
return 8;
}
}
else if (Cs >= 1158) {
return 1;
}
}
else if (refDCval >= 61) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (posBalance < 68) {
if (posBalance < 56) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
if (Cs < 567) {
if (Rs < 973) {
return 4;
}
else if (Rs >= 973) {
return 8;
}
}
else if (Cs >= 567) {
return 8;
}
}
}
else if (posBalance >= 56) {
if (SC < 1314) {
if (Cs < 542) {
return 1;
}
else if (Cs >= 542) {
return 8;
}
}
else if (SC >= 1314) {
if (mu_mv_mag_sq < 91) {
return 1;
}
else if (mu_mv_mag_sq >= 91) {
return 8;
}
}
}
}
else if (posBalance >= 68) {
return 8;
}
}
}
else if (refDCval >= 131) {
if (Rs < 981) {
return 8;
}
else if (Rs >= 981) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (Rs < 998) {
return 8;
}
else if (Rs >= 998) {
return 2;
}
}
}
}
}
else if (mcTcor >= 94) {
if (MVDiff < 5) {
if (Rs < 528) {
if (ssDCval < 96) {
if (AFD < 6) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 8;
}
}
else if (AFD >= 6) {
if (Cs < 596) {
return 8;
}
else if (Cs >= 596) {
return 1;
}
}
}
else if (ssDCval >= 96) {
if (Cs < 770) {
return 1;
}
else if (Cs >= 770) {
return 4;
}
}
}
else if (Rs >= 528) {
if (mu_mv_mag_sq < 111) {
if (MVDiff < 4) {
if (mcTcor < 95) {
if (Cs < 664) {
return 8;
}
else if (Cs >= 664) {
if (TSC < 5) {
if (mu_mv_mag_sq < 104) {
return 2;
}
else if (mu_mv_mag_sq >= 104) {
return 8;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (Cs < 1220) {
if (negBalance < 30) {
if (Rs < 836) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 2;
}
}
else if (Rs >= 836) {
return 8;
}
}
else if (negBalance >= 30) {
if (Cs < 867) {
if (Cs < 634) {
if (Rs < 1014) {
return 8;
}
else if (Rs >= 1014) {
return 1;
}
}
else if (Cs >= 634) {
return 1;
}
}
else if (Cs >= 867) {
if (tcor < 96) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 661) {
return 8;
}
else if (Rs >= 661) {
return 1;
}
}
}
else if (tcor >= 96) {
return 4;
}
}
}
}
else if (Cs >= 1220) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (posBalance < 55) {
if (Cs < 817) {
return 8;
}
else if (Cs >= 817) {
return 1;
}
}
else if (posBalance >= 55) {
if (Cs < 1221) {
return 8;
}
else if (Cs >= 1221) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (SC < 1513) {
return 4;
}
else if (SC >= 1513) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 825) {
return 8;
}
else if (Rs >= 825) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 5) {
if (MVDiff < 21) {
if (SC < 2066) {
if (Cs < 1187) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
if (AFD < 4) {
if (Rs < 758) {
return 2;
}
else if (Rs >= 758) {
return 8;
}
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (Cs >= 1187) {
return 4;
}
}
else if (SC >= 2066) {
if (Cs < 1065) {
if (Rs < 1157) {
return 4;
}
else if (Rs >= 1157) {
return 8;
}
}
else if (Cs >= 1065) {
return 1;
}
}
}
else if (MVDiff >= 21) {
if (ssDCval < 135) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (ssDCval >= 135) {
return 4;
}
}
}
}
}
else if (Cs >= 1309) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
return 8;
}
}
}
}
else if (RsDiff >= 3) {
if (negBalance < 43) {
if (mu_mv_mag_sq < 61) {
if (posBalance < 58) {
if (ssDCval < 79) {
if (MVDiff < 9) {
if (RsCsDiff < 22) {
return 1;
}
else if (RsCsDiff >= 22) {
if (Rs < 1190) {
return 4;
}
else if (Rs >= 1190) {
return 1;
}
}
}
else if (MVDiff >= 9) {
if (Cs < 706) {
if (TSC < 5) {
if (Rs < 821) {
return 4;
}
else if (Rs >= 821) {
return 2;
}
}
else if (TSC >= 5) {
return 1;
}
}
else if (Cs >= 706) {
if (SC < 2167) {
if (diffMVdiffVal < 5) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (diffMVdiffVal >= 5) {
if (Rs < 1189) {
return 8;
}
else if (Rs >= 1189) {
return 2;
}
}
}
else if (SC >= 2167) {
return 8;
}
}
}
}
else if (ssDCval >= 79) {
if (RsDiff < 7) {
if (mu_mv_mag_sq < 57) {
return 2;
}
else if (mu_mv_mag_sq >= 57) {
return 8;
}
}
else if (RsDiff >= 7) {
return 1;
}
}
}
else if (posBalance >= 58) {
if (ssDCval < 129) {
if (Cs < 1104) {
if (posBalance < 59) {
if (SC < 1529) {
return 4;
}
else if (SC >= 1529) {
if (Rs < 1173) {
return 1;
}
else if (Rs >= 1173) {
return 8;
}
}
}
else if (posBalance >= 59) {
if (diffMVdiffVal < -3) {
if (Rs < 1060) {
return 1;
}
else if (Rs >= 1060) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
}
else if (Cs >= 1104) {
if (AFD < 11) {
if (Rs < 1189) {
return 4;
}
else if (Rs >= 1189) {
return 1;
}
}
else if (AFD >= 11) {
if (Cs < 1178) {
return 8;
}
else if (Cs >= 1178) {
return 2;
}
}
}
}
else if (ssDCval >= 129) {
if (Rs < 1214) {
return 4;
}
else if (Rs >= 1214) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (TSC < 4) {
if (diffMVdiffVal < 3) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
if (Rs < 924) {
return 1;
}
else if (Rs >= 924) {
return 8;
}
}
}
else if (TSC >= 4) {
if (Cs < 1116) {
if (Cs < 696) {
if (posBalance < 55) {
if (ssDCval < 83) {
if (negBalance < 26) {
return 2;
}
else if (negBalance >= 26) {
return 1;
}
}
else if (ssDCval >= 83) {
if (Rs < 1143) {
return 8;
}
else if (Rs >= 1143) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (Rs < 1207) {
return 4;
}
else if (Rs >= 1207) {
return 8;
}
}
}
else if (Cs >= 696) {
if (posBalance < 46) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
else if (posBalance >= 46) {
if (tcor < 83) {
if (SC < 1958) {
return 4;
}
else if (SC >= 1958) {
return 8;
}
}
else if (tcor >= 83) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (diffAFD < -1) {
if (mu_mv_mag_sq < 70) {
if (SC < 2109) {
return 1;
}
else if (SC >= 2109) {
return 2;
}
}
else if (mu_mv_mag_sq >= 70) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1116) {
if (ssDCval < 80) {
if (CsDiff < 4) {
if (gchDC < 1) {
if (posBalance < 53) {
if (Rs < 1258) {
return 8;
}
else if (Rs >= 1258) {
return 4;
}
}
else if (posBalance >= 53) {
if (Rs < 1062) {
return 4;
}
else if (Rs >= 1062) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
return 4;
}
}
}
}
else if (gchDC >= 1) {
return 4;
}
}
else if (CsDiff >= 4) {
if (ssDCval < 61) {
if (Rs < 789) {
return 8;
}
else if (Rs >= 789) {
if (Cs < 1203) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
return 8;
}
}
else if (Cs >= 1203) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
else if (ssDCval >= 61) {
return 4;
}
}
}
else if (ssDCval >= 80) {
if (SC < 2059) {
return 4;
}
else if (SC >= 2059) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 43) {
if (MVDiff < 3) {
if (SC < 2520) {
if (SC < 1521) {
return 4;
}
else if (SC >= 1521) {
return 8;
}
}
else if (SC >= 2520) {
if (Rs < 1249) {
return 8;
}
else if (Rs >= 1249) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (ssDCval < 76) {
if (refDCval < 53) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
if (Rs < 932) {
return 4;
}
else if (Rs >= 932) {
return 8;
}
}
}
else if (refDCval >= 53) {
if (Rs < 1164) {
if (refDCval < 62) {
if (Rs < 927) {
return 4;
}
else if (Rs >= 927) {
if (Cs < 929) {
if (mu_mv_mag_sq < 86) {
return 8;
}
else if (mu_mv_mag_sq >= 86) {
return 4;
}
}
else if (Cs >= 929) {
return 4;
}
}
}
else if (refDCval >= 62) {
if (MVDiff < 9) {
if (Rs < 941) {
return 8;
}
else if (Rs >= 941) {
if (Rs < 1075) {
return 1;
}
else if (Rs >= 1075) {
return 8;
}
}
}
else if (MVDiff >= 9) {
if (posBalance < 57) {
if (Rs < 1007) {
return 4;
}
else if (Rs >= 1007) {
return 1;
}
}
else if (posBalance >= 57) {
return 4;
}
}
}
}
else if (Rs >= 1164) {
if (tcor < 88) {
if (AFD < 10) {
if (Cs < 1108) {
return 8;
}
else if (Cs >= 1108) {
if (Rs < 1248) {
return 4;
}
else if (Rs >= 1248) {
return 1;
}
}
}
else if (AFD >= 10) {
return 8;
}
}
else if (tcor >= 88) {
return 4;
}
}
}
}
else if (ssDCval >= 76) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (MVDiff < 25) {
if (tcor < 75) {
if (Rs < 1220) {
return 1;
}
else if (Rs >= 1220) {
return 4;
}
}
else if (tcor >= 75) {
return 8;
}
}
else if (MVDiff >= 25) {
if (Cs < 730) {
return 1;
}
else if (Cs >= 730) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1302) {
if (RsCsDiff < 33) {
if (Cs < 1259) {
if (TSC < 7) {
if (mu_mv_mag_sq < 71) {
if (ssDCval < 68) {
if (negBalance < 30) {
return 1;
}
else if (negBalance >= 30) {
if (posBalance < 45) {
if (Rs < 1458) {
return 8;
}
else if (Rs >= 1458) {
return 4;
}
}
else if (posBalance >= 45) {
if (Rs < 1309) {
return 2;
}
else if (Rs >= 1309) {
return 8;
}
}
}
}
else if (ssDCval >= 68) {
if (refDCval < 73) {
if (Cs < 922) {
if (Rs < 1458) {
if (Rs < 1372) {
return 4;
}
else if (Rs >= 1372) {
return 1;
}
}
else if (Rs >= 1458) {
if (Cs < 674) {
return 4;
}
else if (Cs >= 674) {
return 8;
}
}
}
else if (Cs >= 922) {
if (Rs < 1553) {
return 4;
}
else if (Rs >= 1553) {
return 1;
}
}
}
else if (refDCval >= 73) {
if (ssDCval < 134) {
if (Cs < 679) {
if (Cs < 575) {
return 8;
}
else if (Cs >= 575) {
return 1;
}
}
else if (Cs >= 679) {
if (Cs < 1191) {
return 8;
}
else if (Cs >= 1191) {
return 4;
}
}
}
else if (ssDCval >= 134) {
if (MVDiff < 22) {
if (Cs < 1084) {
return 8;
}
else if (Cs >= 1084) {
return 2;
}
}
else if (MVDiff >= 22) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (SC < 2083) {
if (ssDCval < 119) {
return 8;
}
else if (ssDCval >= 119) {
return 4;
}
}
else if (SC >= 2083) {
if (diffRsCsdiff < 16) {
if (Rs < 1595) {
if (ssDCval < 48) {
if (Rs < 1494) {
return 2;
}
else if (Rs >= 1494) {
return 4;
}
}
else if (ssDCval >= 48) {
if (diffAFD < 2) {
if (MVDiff < 17) {
if (Cs < 1252) {
return 8;
}
else if (Cs >= 1252) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (MVDiff >= 17) {
return 4;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
}
else if (Rs >= 1595) {
if (mu_mv_mag_sq < 127) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (mcTcor < 95) {
if (Cs < 1251) {
if (negBalance < 25) {
if (Rs < 2226) {
return 4;
}
else if (Rs >= 2226) {
return 8;
}
}
else if (negBalance >= 25) {
return 8;
}
}
else if (Cs >= 1251) {
return 1;
}
}
else if (mcTcor >= 95) {
if (Cs < 1209) {
if (diffAFD < 2) {
if (Rs < 1632) {
return 1;
}
else if (Rs >= 1632) {
if (Cs < 1163) {
return 8;
}
else if (Cs >= 1163) {
if (Rs < 1756) {
return 8;
}
else if (Rs >= 1756) {
return 1;
}
}
}
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (Cs >= 1209) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 2022) {
return 1;
}
else if (Rs >= 2022) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 16) {
return 4;
}
}
}
}
else if (TSC >= 7) {
if (negBalance < 43) {
if (mcTcor < 87) {
if (Rs < 1429) {
return 1;
}
else if (Rs >= 1429) {
return 4;
}
}
else if (mcTcor >= 87) {
return 8;
}
}
else if (negBalance >= 43) {
if (MVDiff < 31) {
if (posBalance < 68) {
return 8;
}
else if (posBalance >= 68) {
return 4;
}
}
else if (MVDiff >= 31) {
return 4;
}
}
}
}
else if (Cs >= 1259) {
if (RsCsDiff < 22) {
if (diffRsCsdiff < -12) {
if (Rs < 1834) {
return 1;
}
else if (Rs >= 1834) {
return 8;
}
}
else if (diffRsCsdiff >= -12) {
return 8;
}
}
else if (RsCsDiff >= 22) {
if (Cs < 1267) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
return 2;
}
}
else if (Cs >= 1267) {
if (tcor < 86) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 1;
}
}
else if (tcor >= 86) {
if (Rs < 1736) {
if (diffTSC < -1) {
if (Rs < 1428) {
return 1;
}
else if (Rs >= 1428) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (Rs >= 1736) {
if (TSC < 6) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1807) {
return 8;
}
else if (Rs >= 1807) {
if (Rs < 2115) {
return 2;
}
else if (Rs >= 2115) {
return 1;
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 2124) {
return 1;
}
else if (Rs >= 2124) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 33) {
if (Cs < 876) {
if (ssDCval < 68) {
if (SC < 2588) {
if (negBalance < 38) {
if (tcor < 79) {
return 8;
}
else if (tcor >= 79) {
if (SC < 2044) {
return 1;
}
else if (SC >= 2044) {
return 2;
}
}
}
else if (negBalance >= 38) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Rs < 1548) {
if (Rs < 1421) {
return 8;
}
else if (Rs >= 1421) {
return 2;
}
}
else if (Rs >= 1548) {
return 4;
}
}
}
}
else if (SC >= 2588) {
if (mcTcor < 93) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (MVDiff < 25) {
return 4;
}
else if (MVDiff >= 25) {
if (refDCval < 54) {
return 4;
}
else if (refDCval >= 54) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
return 2;
}
}
}
else if (ssDCval >= 68) {
if (Rs < 1567) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
if (Rs < 1418) {
return 2;
}
else if (Rs >= 1418) {
return 4;
}
}
}
else if (Rs >= 1567) {
if (MVDiff < 7) {
if (diffMVdiffVal < -2) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
else if (MVDiff >= 7) {
if (Cs < 727) {
return 1;
}
else if (Cs >= 727) {
if (Rs < 1591) {
return 4;
}
else if (Rs >= 1591) {
return 8;
}
}
}
}
}
}
else if (Cs >= 876) {
if (posBalance < 45) {
if (tcor < 86) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (Rs < 2228) {
return 2;
}
else if (Rs >= 2228) {
return 8;
}
}
}
else if (tcor >= 86) {
if (Rs < 1823) {
return 8;
}
else if (Rs >= 1823) {
return 4;
}
}
}
else if (posBalance >= 45) {
if (mu_mv_mag_sq < 83) {
if (refDCval < 162) {
if (refDCval < 108) {
if (ssDCval < 88) {
if (posBalance < 47) {
if (Cs < 1069) {
if (Rs < 1850) {
return 8;
}
else if (Rs >= 1850) {
return 4;
}
}
else if (Cs >= 1069) {
return 1;
}
}
else if (posBalance >= 47) {
if (mu_mv_mag_sq < 47) {
if (negBalance < 47) {
if (mu_mv_mag_sq < 44) {
if (Rs < 1704) {
return 1;
}
else if (Rs >= 1704) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (Rs < 1325) {
return 4;
}
else if (Rs >= 1325) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 47) {
if (mu_mv_mag_sq < 74) {
return 8;
}
else if (mu_mv_mag_sq >= 74) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 78) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (Rs < 1831) {
return 2;
}
else if (Rs >= 1831) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (Rs < 1596) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (Rs >= 1596) {
if (ssDCval < 99) {
if (AFD < 11) {
if (tcor < 92) {
return 2;
}
else if (tcor >= 92) {
return 4;
}
}
else if (AFD >= 11) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (ssDCval >= 99) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
}
else if (refDCval >= 108) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 1565) {
return 1;
}
else if (Rs >= 1565) {
return 8;
}
}
}
}
else if (refDCval >= 162) {
if (Rs < 1319) {
return 8;
}
else if (Rs >= 1319) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (tcor < 72) {
return 4;
}
else if (tcor >= 72) {
if (refDCval < 71) {
if (Cs < 1197) {
if (refDCval < 62) {
if (RsCsDiff < 49) {
return 8;
}
else if (RsCsDiff >= 49) {
return 2;
}
}
else if (refDCval >= 62) {
return 2;
}
}
else if (Cs >= 1197) {
if (Rs < 1754) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (Rs >= 1754) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
else if (refDCval >= 71) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (posBalance < 57) {
if (TSC < 7) {
if (Rs < 1903) {
return 2;
}
else if (Rs >= 1903) {
return 8;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (posBalance >= 57) {
if (Rs < 1634) {
if (Rs < 1491) {
return 8;
}
else if (Rs >= 1491) {
return 2;
}
}
else if (Rs >= 1634) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (negBalance < 13) {
if (Rs < 702) {
if (ssDCval < 57) {
if (tcor < 99) {
if (negBalance < 6) {
if (Rs < 309) {
return 4;
}
else if (Rs >= 309) {
return 8;
}
}
else if (negBalance >= 6) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
return 2;
}
}
}
else if (tcor >= 99) {
if (ssDCval < 30) {
return 8;
}
else if (ssDCval >= 30) {
return 1;
}
}
}
else if (ssDCval >= 57) {
if (RsCsDiff < 1) {
if (Cs < 685) {
if (Rs < 522) {
if (Rs < 487) {
return 8;
}
else if (Rs >= 487) {
return 1;
}
}
else if (Rs >= 522) {
return 8;
}
}
else if (Cs >= 685) {
if (Rs < 663) {
if (SC < 1173) {
return 4;
}
else if (SC >= 1173) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
return 2;
}
}
}
else if (Rs >= 663) {
return 1;
}
}
}
else if (RsCsDiff >= 1) {
if (AFD < 5) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 1;
}
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (Rs >= 702) {
if (diffAFD < -3) {
if (Rs < 1599) {
if (diffRsCsdiff < -19) {
if (ssDCval < 133) {
if (diffRsCsdiff < -38) {
if (diffAFD < -14) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (diffTSC < -9) {
if (Rs < 1308) {
return 4;
}
else if (Rs >= 1308) {
return 1;
}
}
else if (diffTSC >= -9) {
return 1;
}
}
}
else if (diffAFD >= -14) {
return 8;
}
}
else if (diffRsCsdiff >= -38) {
if (Cs < 572) {
return 2;
}
else if (Cs >= 572) {
return 1;
}
}
}
else if (ssDCval >= 133) {
if (ssDCval < 141) {
if (MVDiff < 16) {
if (Rs < 1126) {
return 8;
}
else if (Rs >= 1126) {
return 1;
}
}
else if (MVDiff >= 16) {
if (diffMVdiffVal < 33) {
return 4;
}
else if (diffMVdiffVal >= 33) {
return 1;
}
}
}
else if (ssDCval >= 141) {
if (diffAFD < -21) {
if (Rs < 1173) {
return 8;
}
else if (Rs >= 1173) {
return 4;
}
}
else if (diffAFD >= -21) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -19) {
if (TSCindex < 1) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (Rs < 859) {
if (Rs < 805) {
return 8;
}
else if (Rs >= 805) {
return 2;
}
}
else if (Rs >= 859) {
if (diffAFD < -10) {
return 8;
}
else if (diffAFD >= -10) {
return 4;
}
}
}
}
else if (TSCindex >= 1) {
return 8;
}
}
}
else if (Rs >= 1599) {
if (diffMVdiffVal < -10) {
if (diffAFD < -15) {
return 4;
}
else if (diffAFD >= -15) {
if (Rs < 1850) {
return 8;
}
else if (Rs >= 1850) {
return 1;
}
}
}
else if (diffMVdiffVal >= -10) {
return 8;
}
}
}
else if (diffAFD >= -3) {
if (SC < 2893) {
if (refDCval < 153) {
if (Rs < 783) {
if (MVDiff < 1) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (Rs < 763) {
if (ssDCval < 106) {
return 8;
}
else if (ssDCval >= 106) {
return 1;
}
}
else if (Rs >= 763) {
return 8;
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 121) {
if (SC < 1447) {
if (Cs < 679) {
if (SC < 1364) {
return 2;
}
else if (SC >= 1364) {
return 8;
}
}
else if (Cs >= 679) {
return 1;
}
}
else if (SC >= 1447) {
return 2;
}
}
else if (ssDCval >= 121) {
return 8;
}
}
}
else if (Rs >= 783) {
if (Rs < 1847) {
if (Rs < 1058) {
if (Rs < 874) {
if (Rs < 870) {
if (RsDiff < 2) {
if (Cs < 1148) {
return 8;
}
else if (Cs >= 1148) {
if (Rs < 864) {
return 1;
}
else if (Rs >= 864) {
return 8;
}
}
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Rs >= 870) {
return 2;
}
}
else if (Rs >= 874) {
return 8;
}
}
else if (Rs >= 1058) {
if (negBalance < 10) {
if (Rs < 1134) {
if (Rs < 1130) {
if (Cs < 627) {
return 4;
}
else if (Cs >= 627) {
if (diffMVdiffVal < -11) {
return 4;
}
else if (diffMVdiffVal >= -11) {
return 8;
}
}
}
else if (Rs >= 1130) {
return 1;
}
}
else if (Rs >= 1134) {
if (diffRsCsdiff < 2) {
if (Cs < 682) {
if (Rs < 1480) {
return 8;
}
else if (Rs >= 1480) {
return 4;
}
}
else if (Cs >= 682) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
else if (negBalance >= 10) {
if (ssDCval < 64) {
if (ssDCval < 57) {
if (TSCindex < 2) {
if (Cs < 577) {
return 1;
}
else if (Cs >= 577) {
return 8;
}
}
else if (TSCindex >= 2) {
if (Cs < 1155) {
return 1;
}
else if (Cs >= 1155) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (Rs < 1597) {
return 8;
}
else if (Rs >= 1597) {
return 4;
}
}
}
}
else if (ssDCval >= 64) {
if (Rs < 1792) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (Rs < 1432) {
return 4;
}
else if (Rs >= 1432) {
return 8;
}
}
}
else if (Rs >= 1792) {
return 1;
}
}
}
}
}
else if (Rs >= 1847) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (ssDCval < 87) {
if (Rs < 2020) {
return 1;
}
else if (Rs >= 2020) {
return 2;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
}
}
}
else if (refDCval >= 153) {
if (Rs < 836) {
return 8;
}
else if (Rs >= 836) {
return 2;
}
}
}
else if (SC >= 2893) {
if (Rs < 2055) {
if (ssDCval < 68) {
if (Rs < 1839) {
return 8;
}
else if (Rs >= 1839) {
if (Cs < 938) {
return 1;
}
else if (Cs >= 938) {
return 4;
}
}
}
else if (ssDCval >= 68) {
if (ssDCval < 83) {
return 1;
}
else if (ssDCval >= 83) {
return 8;
}
}
}
else if (Rs >= 2055) {
if (SC < 3573) {
return 8;
}
else if (SC >= 3573) {
if (SC < 3631) {
if (Cs < 1317) {
return 2;
}
else if (Cs >= 1317) {
if (SC < 3612) {
return 8;
}
else if (SC >= 3612) {
if (Rs < 2290) {
return 2;
}
else if (Rs >= 2290) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
if (Rs < 2302) {
return 2;
}
else if (Rs >= 2302) {
return 8;
}
}
}
}
}
}
else if (SC >= 3631) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (ssDCval < 88) {
if (SC < 2166) {
if (Rs < 962) {
if (refDCval < 62) {
if (TSCindex < 2) {
if (Rs < 613) {
if (Rs < 563) {
if (Cs < 800) {
return 1;
}
else if (Cs >= 800) {
if (Rs < 369) {
return 2;
}
else if (Rs >= 369) {
return 4;
}
}
}
else if (Rs >= 563) {
return 8;
}
}
else if (Rs >= 613) {
if (ssDCval < 48) {
return 8;
}
else if (ssDCval >= 48) {
if (Cs < 643) {
return 8;
}
else if (Cs >= 643) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 1;
}
}
}
}
}
else if (TSCindex >= 2) {
if (Rs < 528) {
if (Cs < 906) {
if (refDCval < 46) {
if (Cs < 671) {
return 1;
}
else if (Cs >= 671) {
return 8;
}
}
else if (refDCval >= 46) {
if (AFD < 8) {
if (ssDCval < 56) {
if (AFD < 5) {
if (SC < 1028) {
return 1;
}
else if (SC >= 1028) {
if (TSC < 3) {
if (Rs < 497) {
return 4;
}
else if (Rs >= 497) {
return 8;
}
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (AFD >= 5) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 1;
}
}
}
else if (ssDCval >= 56) {
if (tcor < 96) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (tcor >= 96) {
if (Rs < 518) {
return 4;
}
else if (Rs >= 518) {
if (Cs < 701) {
return 2;
}
else if (Cs >= 701) {
return 4;
}
}
}
}
}
else if (AFD >= 8) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
return 1;
}
}
}
}
else if (Cs >= 906) {
if (Rs < 434) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
return 1;
}
}
else if (Rs >= 434) {
if (mu_mv_mag_sq < 119) {
return 1;
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 504) {
return 4;
}
else if (Rs >= 504) {
if (Rs < 513) {
return 1;
}
else if (Rs >= 513) {
return 4;
}
}
}
}
}
}
else if (Rs >= 528) {
if (MVDiff < 1) {
if (SC < 1467) {
if (Rs < 562) {
if (Cs < 668) {
return 8;
}
else if (Cs >= 668) {
return 4;
}
}
else if (Rs >= 562) {
return 8;
}
}
else if (SC >= 1467) {
if (Rs < 681) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (ssDCval < 55) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (ssDCval >= 55) {
if (Cs < 973) {
return 8;
}
else if (Cs >= 973) {
return 1;
}
}
}
}
else if (Rs >= 681) {
if (TSCindex < 3) {
if (ssDCval < 49) {
if (Cs < 670) {
return 1;
}
else if (Cs >= 670) {
return 8;
}
}
else if (ssDCval >= 49) {
if (negBalance < 26) {
if (SC < 1671) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (SC >= 1671) {
return 1;
}
}
else if (negBalance >= 26) {
if (Rs < 792) {
return 1;
}
else if (Rs >= 792) {
return 8;
}
}
}
}
else if (TSCindex >= 3) {
if (Cs < 975) {
if (SC < 1477) {
return 4;
}
else if (SC >= 1477) {
return 8;
}
}
else if (Cs >= 975) {
if (Rs < 790) {
return 1;
}
else if (Rs >= 790) {
if (Cs < 1077) {
return 8;
}
else if (Cs >= 1077) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (Rs < 793) {
if (negBalance < 17) {
return 2;
}
else if (negBalance >= 17) {
if (negBalance < 35) {
if (refDCval < 51) {
if (mu_mv_mag_sq < 82) {
if (negBalance < 25) {
if (Rs < 628) {
return 8;
}
else if (Rs >= 628) {
return 1;
}
}
else if (negBalance >= 25) {
return 4;
}
}
else if (mu_mv_mag_sq >= 82) {
return 1;
}
}
else if (refDCval >= 51) {
if (TSC < 4) {
if (Rs < 686) {
if (MVDiff < 2) {
if (Rs < 634) {
return 8;
}
else if (Rs >= 634) {
if (Rs < 657) {
return 1;
}
else if (Rs >= 657) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Rs < 605) {
return 4;
}
else if (Rs >= 605) {
return 8;
}
}
}
}
else if (Rs >= 686) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
if (posBalance < 48) {
if (Rs < 708) {
return 1;
}
else if (Rs >= 708) {
if (Rs < 770) {
return 4;
}
else if (Rs >= 770) {
return 1;
}
}
}
else if (posBalance >= 48) {
return 8;
}
}
}
}
else if (TSC >= 4) {
if (SC < 1333) {
if (Rs < 623) {
return 1;
}
else if (Rs >= 623) {
return 4;
}
}
else if (SC >= 1333) {
if (Cs < 1160) {
return 4;
}
else if (Cs >= 1160) {
return 1;
}
}
}
}
}
else if (negBalance >= 35) {
if (Cs < 675) {
if (AFD < 8) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 4;
}
}
else if (AFD >= 8) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
return 8;
}
}
}
else if (Cs >= 675) {
if (MVDiff < 7) {
if (ssDCval < 55) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
else if (ssDCval >= 55) {
if (Cs < 782) {
return 4;
}
else if (Cs >= 782) {
return 8;
}
}
}
else if (MVDiff >= 7) {
if (Cs < 800) {
return 4;
}
else if (Cs >= 800) {
return 2;
}
}
}
}
}
}
else if (Rs >= 793) {
if (negBalance < 38) {
if (ssDCval < 52) {
if (mu_mv_mag_sq < 103) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (Rs < 813) {
return 2;
}
else if (Rs >= 813) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 103) {
return 8;
}
}
else if (ssDCval >= 52) {
if (RsCsDiff < 7) {
return 1;
}
else if (RsCsDiff >= 7) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
}
else if (negBalance >= 38) {
if (tcor < 96) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
else if (tcor >= 96) {
if (Rs < 899) {
return 4;
}
else if (Rs >= 899) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 62) {
if (Rs < 357) {
if (mu_mv_mag_sq < 5) {
if (Cs < 625) {
if (Rs < 293) {
return 1;
}
else if (Rs >= 293) {
return 4;
}
}
else if (Cs >= 625) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
if (Cs < 539) {
return 4;
}
else if (Cs >= 539) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
}
}
else if (Rs >= 357) {
if (posBalance < 50) {
if (Cs < 697) {
if (mcTcor < 99) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (SC < 1385) {
if (Cs < 615) {
if (AFD < 9) {
if (Rs < 792) {
return 4;
}
else if (Rs >= 792) {
return 2;
}
}
else if (AFD >= 9) {
return 8;
}
}
else if (Cs >= 615) {
if (Rs < 608) {
return 8;
}
else if (Rs >= 608) {
return 2;
}
}
}
else if (SC >= 1385) {
if (TSC < 4) {
if (negBalance < 23) {
return 2;
}
else if (negBalance >= 23) {
return 1;
}
}
else if (TSC >= 4) {
if (SC < 1410) {
if (Rs < 788) {
return 4;
}
else if (Rs >= 788) {
return 1;
}
}
else if (SC >= 1410) {
if (Rs < 812) {
return 2;
}
else if (Rs >= 812) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (posBalance < 35) {
if (tcor < 99) {
if (Rs < 770) {
return 2;
}
else if (Rs >= 770) {
return 1;
}
}
else if (tcor >= 99) {
if (SC < 1504) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
else if (SC >= 1504) {
return 8;
}
}
}
else if (posBalance >= 35) {
if (Rs < 548) {
if (Rs < 487) {
if (Rs < 421) {
return 4;
}
else if (Rs >= 421) {
return 2;
}
}
else if (Rs >= 487) {
return 8;
}
}
else if (Rs >= 548) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
return 8;
}
}
}
}
}
else if (Cs >= 697) {
if (MVDiff < 1) {
if (ssDCval < 81) {
if (ssDCval < 69) {
if (Rs < 718) {
if (Rs < 517) {
return 4;
}
else if (Rs >= 517) {
return 8;
}
}
else if (Rs >= 718) {
if (Rs < 869) {
if (Cs < 912) {
return 4;
}
else if (Cs >= 912) {
if (Rs < 803) {
return 2;
}
else if (Rs >= 803) {
return 8;
}
}
}
else if (Rs >= 869) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
if (Rs < 916) {
return 8;
}
else if (Rs >= 916) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 560) {
return 8;
}
else if (Rs >= 560) {
if (Rs < 853) {
return 1;
}
else if (Rs >= 853) {
return 8;
}
}
}
}
else if (ssDCval >= 81) {
if (Rs < 662) {
return 2;
}
else if (Rs >= 662) {
if (Rs < 700) {
return 1;
}
else if (Rs >= 700) {
return 4;
}
}
}
}
else if (MVDiff >= 1) {
if (TSCindex < 2) {
return 1;
}
else if (TSCindex >= 2) {
if (negBalance < 30) {
if (ssDCval < 72) {
if (Rs < 913) {
return 4;
}
else if (Rs >= 913) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 2;
}
}
}
else if (ssDCval >= 72) {
if (tcor < 98) {
return 2;
}
else if (tcor >= 98) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (Rs < 590) {
return 1;
}
else if (Rs >= 590) {
return 2;
}
}
}
}
}
else if (negBalance >= 30) {
if (negBalance < 57) {
if (Rs < 871) {
if (RsCsDiff < 9) {
if (mu_mv_mag_sq < 80) {
if (mu_mv_mag_sq < 53) {
if (Rs < 716) {
return 2;
}
else if (Rs >= 716) {
return 1;
}
}
else if (mu_mv_mag_sq >= 53) {
return 1;
}
}
else if (mu_mv_mag_sq >= 80) {
if (ssDCval < 67) {
return 8;
}
else if (ssDCval >= 67) {
if (posBalance < 46) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 2;
}
}
else if (posBalance >= 46) {
if (Rs < 739) {
return 8;
}
else if (Rs >= 739) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 9) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
if (Rs < 763) {
return 1;
}
else if (Rs >= 763) {
return 2;
}
}
}
}
else if (Rs >= 871) {
if (Cs < 986) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
if (Rs < 943) {
return 8;
}
else if (Rs >= 943) {
if (Rs < 946) {
return 4;
}
else if (Rs >= 946) {
return 1;
}
}
}
}
else if (Cs >= 986) {
if (Cs < 1095) {
return 4;
}
else if (Cs >= 1095) {
if (Rs < 873) {
return 4;
}
else if (Rs >= 873) {
return 2;
}
}
}
}
}
else if (negBalance >= 57) {
if (Rs < 588) {
return 2;
}
else if (Rs >= 588) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 50) {
if (MVDiff < 2) {
if (ssDCval < 82) {
if (Rs < 884) {
if (Rs < 639) {
if (Rs < 491) {
if (Rs < 426) {
return 4;
}
else if (Rs >= 426) {
return 2;
}
}
else if (Rs >= 491) {
return 8;
}
}
else if (Rs >= 639) {
if (RsDiff < 1) {
if (Rs < 789) {
return 8;
}
else if (Rs >= 789) {
return 4;
}
}
else if (RsDiff >= 1) {
if (negBalance < 35) {
if (mcTcor < 98) {
if (Rs < 763) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (ssDCval < 75) {
return 1;
}
else if (ssDCval >= 75) {
return 4;
}
}
}
else if (Rs >= 763) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 4;
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 27) {
return 2;
}
else if (negBalance >= 27) {
return 1;
}
}
}
else if (negBalance >= 35) {
if (negBalance < 41) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
if (Cs < 1132) {
return 1;
}
else if (Cs >= 1132) {
return 8;
}
}
}
else if (negBalance >= 41) {
if (Cs < 836) {
return 2;
}
else if (Cs >= 836) {
return 8;
}
}
}
}
}
}
else if (Rs >= 884) {
if (SC < 2006) {
if (Cs < 666) {
return 2;
}
else if (Cs >= 666) {
return 8;
}
}
else if (SC >= 2006) {
if (AFD < 7) {
if (Rs < 957) {
return 4;
}
else if (Rs >= 957) {
return 2;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
}
else if (ssDCval >= 82) {
if (negBalance < 28) {
return 2;
}
else if (negBalance >= 28) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (Cs < 740) {
if (mcTcor < 98) {
if (mu_mv_mag_sq < 44) {
if (negBalance < 42) {
if (negBalance < 39) {
if (Cs < 686) {
return 4;
}
else if (Cs >= 686) {
return 8;
}
}
else if (negBalance >= 39) {
return 2;
}
}
else if (negBalance >= 42) {
if (diffTSC < -1) {
if (Rs < 676) {
return 1;
}
else if (Rs >= 676) {
return 4;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
if (Rs < 657) {
return 1;
}
else if (Rs >= 657) {
return 2;
}
}
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 25) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 487) {
return 8;
}
else if (Rs >= 487) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (AFD < 4) {
if (Rs < 910) {
return 2;
}
else if (Rs >= 910) {
return 8;
}
}
else if (AFD >= 4) {
if (Rs < 538) {
return 1;
}
else if (Rs >= 538) {
return 8;
}
}
}
}
}
}
else if (Cs >= 740) {
if (mu_mv_mag_sq < 11) {
if (RsCsDiff < 23) {
if (MVDiff < 8) {
if (Cs < 751) {
return 4;
}
else if (Cs >= 751) {
if (ssDCval < 87) {
return 2;
}
else if (ssDCval >= 87) {
return 4;
}
}
}
else if (MVDiff >= 8) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (Rs < 835) {
return 2;
}
else if (Rs >= 835) {
return 4;
}
}
}
}
else if (RsCsDiff >= 23) {
return 2;
}
}
else if (mu_mv_mag_sq >= 11) {
if (negBalance < 31) {
if (ssDCval < 72) {
if (negBalance < 30) {
if (SC < 1714) {
if (Rs < 738) {
return 8;
}
else if (Rs >= 738) {
return 4;
}
}
else if (SC >= 1714) {
return 1;
}
}
else if (negBalance >= 30) {
if (Rs < 759) {
return 8;
}
else if (Rs >= 759) {
return 2;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 527) {
if (Rs < 462) {
return 8;
}
else if (Rs >= 462) {
return 1;
}
}
else if (Rs >= 527) {
if (Cs < 745) {
return 1;
}
else if (Cs >= 745) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 2;
}
}
}
}
}
else if (negBalance >= 31) {
if (mu_mv_mag_sq < 77) {
if (tcor < 94) {
if (Rs < 914) {
return 1;
}
else if (Rs >= 914) {
return 8;
}
}
else if (tcor >= 94) {
if (Rs < 822) {
return 2;
}
else if (Rs >= 822) {
if (Rs < 848) {
return 1;
}
else if (Rs >= 848) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 798) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 1;
}
}
else if (Rs >= 798) {
if (Rs < 809) {
if (Rs < 805) {
return 1;
}
else if (Rs >= 805) {
return 2;
}
}
else if (Rs >= 809) {
if (Rs < 857) {
return 8;
}
else if (Rs >= 857) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 962) {
if (mu_mv_mag_sq < 125) {
if (refDCval < 78) {
if (posBalance < 38) {
if (ssDCval < 51) {
if (Rs < 1199) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (Rs >= 1199) {
return 1;
}
}
else if (ssDCval >= 51) {
if (Cs < 682) {
if (SC < 1577) {
return 4;
}
else if (SC >= 1577) {
return 1;
}
}
else if (Cs >= 682) {
if (RsDiff < 3) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (negBalance < 14) {
return 8;
}
else if (negBalance >= 14) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
return 1;
}
}
}
}
else if (RsDiff >= 3) {
if (SC < 1921) {
return 1;
}
else if (SC >= 1921) {
return 8;
}
}
}
}
}
else if (posBalance >= 38) {
if (posBalance < 55) {
if (ssDCval < 51) {
if (tcor < 94) {
if (Rs < 1319) {
return 1;
}
else if (Rs >= 1319) {
return 4;
}
}
else if (tcor >= 94) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 1033) {
return 4;
}
else if (Rs >= 1033) {
if (Rs < 1180) {
return 8;
}
else if (Rs >= 1180) {
if (Rs < 1357) {
return 8;
}
else if (Rs >= 1357) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 51) {
if (RsCsDiff < 23) {
if (SC < 1998) {
if (tcor < 99) {
if (MVDiff < 11) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (tcor < 98) {
if (negBalance < 40) {
if (diffRsCsdiff < -9) {
if (SC < 1951) {
return 1;
}
else if (SC >= 1951) {
return 8;
}
}
else if (diffRsCsdiff >= -9) {
return 1;
}
}
else if (negBalance >= 40) {
if (Rs < 1014) {
return 1;
}
else if (Rs >= 1014) {
if (diffRsCsdiff < 2) {
if (Rs < 1220) {
return 1;
}
else if (Rs >= 1220) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
}
}
else if (tcor >= 98) {
if (negBalance < 28) {
return 2;
}
else if (negBalance >= 28) {
if (Rs < 1163) {
return 1;
}
else if (Rs >= 1163) {
return 8;
}
}
}
}
}
else if (MVDiff >= 11) {
if (TSC < 4) {
if (Rs < 1261) {
return 8;
}
else if (Rs >= 1261) {
return 2;
}
}
else if (TSC >= 4) {
if (Cs < 708) {
if (diffRsCsdiff < -10) {
return 2;
}
else if (diffRsCsdiff >= -10) {
return 1;
}
}
else if (Cs >= 708) {
return 2;
}
}
}
}
else if (tcor >= 99) {
if (SC < 1802) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
return 2;
}
}
else if (SC >= 1802) {
if (Rs < 1360) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
else if (Rs >= 1360) {
return 8;
}
}
}
}
else if (SC >= 1998) {
if (refDCval < 69) {
if (Cs < 1071) {
if (mcTcor < 97) {
if (posBalance < 48) {
if (Rs < 1302) {
return 8;
}
else if (Rs >= 1302) {
return 4;
}
}
else if (posBalance >= 48) {
return 1;
}
}
else if (mcTcor >= 97) {
if (Rs < 1263) {
return 1;
}
else if (Rs >= 1263) {
if (refDCval < 61) {
if (CsDiff < 2) {
if (Rs < 1363) {
return 8;
}
else if (Rs >= 1363) {
return 1;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (refDCval >= 61) {
return 1;
}
}
}
}
else if (Cs >= 1071) {
if (negBalance < 28) {
if (Rs < 1015) {
return 1;
}
else if (Rs >= 1015) {
return 4;
}
}
else if (negBalance >= 28) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (refDCval >= 69) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 1148) {
return 8;
}
else if (Rs >= 1148) {
if (CsDiff < 1) {
if (Rs < 1476) {
return 1;
}
else if (Rs >= 1476) {
return 8;
}
}
else if (CsDiff >= 1) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 23) {
if (tcor < 93) {
if (MVDiff < 15) {
if (MVDiff < 5) {
if (Rs < 1036) {
return 4;
}
else if (Rs >= 1036) {
return 8;
}
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (MVDiff >= 15) {
if (Rs < 1072) {
return 4;
}
else if (Rs >= 1072) {
return 2;
}
}
}
else if (tcor >= 93) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
if (MVDiff < 3) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
else if (MVDiff >= 3) {
if (negBalance < 43) {
if (posBalance < 58) {
if (SC < 1877) {
return 8;
}
else if (SC >= 1877) {
if (Cs < 1011) {
return 1;
}
else if (Cs >= 1011) {
return 8;
}
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (negBalance >= 43) {
if (posBalance < 60) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
return 8;
}
}
else if (posBalance >= 60) {
if (Rs < 1042) {
return 1;
}
else if (Rs >= 1042) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 78) {
if (mu_mv_mag_sq < 40) {
if (Cs < 747) {
return 1;
}
else if (Cs >= 747) {
return 2;
}
}
else if (mu_mv_mag_sq >= 40) {
if (mu_mv_mag_sq < 112) {
if (Rs < 1390) {
return 8;
}
else if (Rs >= 1390) {
return 4;
}
}
else if (mu_mv_mag_sq >= 112) {
if (negBalance < 31) {
if (Cs < 1036) {
if (mu_mv_mag_sq < 123) {
return 8;
}
else if (mu_mv_mag_sq >= 123) {
return 4;
}
}
else if (Cs >= 1036) {
return 4;
}
}
else if (negBalance >= 31) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1138) {
if (posBalance < 40) {
if (Cs < 581) {
return 2;
}
else if (Cs >= 581) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 1060) {
return 8;
}
else if (Rs >= 1060) {
return 1;
}
}
}
}
else if (posBalance >= 40) {
if (SC < 1819) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
else if (SC >= 1819) {
if (Rs < 1025) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
return 4;
}
}
else if (Rs >= 1025) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
return 4;
}
}
}
}
}
else if (Rs >= 1138) {
if (SC < 1924) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
}
else if (SC >= 1924) {
if (ssDCval < 74) {
if (negBalance < 18) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
return 1;
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (negBalance >= 18) {
return 8;
}
}
else if (ssDCval >= 74) {
if (Rs < 1344) {
return 8;
}
else if (Rs >= 1344) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 2166) {
if (TSC < 3) {
if (ssDCval < 81) {
if (ssDCval < 52) {
if (Rs < 1629) {
if (MVDiff < 9) {
if (negBalance < 20) {
return 8;
}
else if (negBalance >= 20) {
if (negBalance < 25) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
return 4;
}
}
else if (negBalance >= 25) {
if (Cs < 1286) {
return 8;
}
else if (Cs >= 1286) {
return 4;
}
}
}
}
else if (MVDiff >= 9) {
if (Rs < 1209) {
return 4;
}
else if (Rs >= 1209) {
return 1;
}
}
}
else if (Rs >= 1629) {
if (SC < 2981) {
if (ssDCval < 47) {
if (posBalance < 30) {
if (Rs < 1741) {
return 4;
}
else if (Rs >= 1741) {
return 8;
}
}
else if (posBalance >= 30) {
return 4;
}
}
else if (ssDCval >= 47) {
if (Rs < 2198) {
if (RsCsDiff < 4) {
if (Cs < 1143) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1749) {
if (Rs < 1742) {
if (ssDCval < 48) {
if (Cs < 1036) {
return 8;
}
else if (Cs >= 1036) {
if (Cs < 1126) {
if (mcTcor < 99) {
return 8;
}
else if (mcTcor >= 99) {
if (Rs < 1734) {
return 8;
}
else if (Rs >= 1734) {
return 8;
}
}
}
else if (Cs >= 1126) {
return 4;
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
else if (Rs >= 1742) {
return 4;
}
}
else if (Rs >= 1749) {
if (SC < 2904) {
return 8;
}
else if (SC >= 2904) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (SC < 2780) {
return 4;
}
else if (SC >= 2780) {
if (posBalance < 41) {
if (posBalance < 35) {
if (SC < 2932) {
if (SC < 2860) {
if (Cs < 1104) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (Rs < 1756) {
return 4;
}
else if (Rs >= 1756) {
if (posBalance < 34) {
return 4;
}
else if (posBalance >= 34) {
return 4;
}
}
}
}
else if (Cs >= 1104) {
return 4;
}
}
else if (SC >= 2860) {
if (Rs < 1781) {
return 8;
}
else if (Rs >= 1781) {
if (Rs < 1813) {
return 4;
}
else if (Rs >= 1813) {
return 4;
}
}
}
}
else if (SC >= 2932) {
return 4;
}
}
else if (posBalance >= 35) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (diffAFD < 1) {
if (Rs < 1763) {
if (Rs < 1735) {
return 4;
}
else if (Rs >= 1735) {
return 4;
}
}
else if (Rs >= 1763) {
if (Cs < 1070) {
return 4;
}
else if (Cs >= 1070) {
if (Cs < 1111) {
return 8;
}
else if (Cs >= 1111) {
if (Rs < 1801) {
return 8;
}
else if (Rs >= 1801) {
return 8;
}
}
}
}
}
else if (diffAFD >= 1) {
if (Cs < 1079) {
return 8;
}
else if (Cs >= 1079) {
return 8;
}
}
}
}
}
else if (posBalance >= 41) {
return 4;
}
}
}
}
else if (Cs >= 1143) {
if (Rs < 1674) {
return 8;
}
else if (Rs >= 1674) {
if (posBalance < 36) {
if (TSCindex < 3) {
if (Rs < 1795) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
if (Rs < 1698) {
return 4;
}
else if (Rs >= 1698) {
return 4;
}
}
}
else if (Rs >= 1795) {
return 8;
}
}
else if (TSCindex >= 3) {
return 4;
}
}
else if (posBalance >= 36) {
if (Rs < 1768) {
return 4;
}
else if (Rs >= 1768) {
if (Rs < 1795) {
return 4;
}
else if (Rs >= 1795) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 4) {
if (ssDCval < 51) {
if (Cs < 1151) {
if (negBalance < 22) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (SC < 2823) {
if (SC < 2749) {
return 4;
}
else if (SC >= 2749) {
return 4;
}
}
else if (SC >= 2823) {
if (Cs < 1142) {
if (Rs < 1929) {
return 8;
}
else if (Rs >= 1929) {
return 8;
}
}
else if (Cs >= 1142) {
if (Rs < 1735) {
return 4;
}
else if (Rs >= 1735) {
return 4;
}
}
}
}
}
else if (negBalance >= 22) {
if (Rs < 1638) {
return 4;
}
else if (Rs >= 1638) {
return 4;
}
}
}
else if (Cs >= 1151) {
return 4;
}
}
else if (ssDCval >= 51) {
return 8;
}
}
}
else if (Rs >= 2198) {
return 8;
}
}
}
else if (SC >= 2981) {
if (SC < 3393) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (SC >= 3393) {
return 8;
}
}
}
}
else if (ssDCval >= 52) {
if (MVDiff < 9) {
if (SC < 2493) {
if (refDCval < 69) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 127) {
if (AFD < 3) {
if (Cs < 1314) {
return 1;
}
else if (Cs >= 1314) {
return 8;
}
}
else if (AFD >= 3) {
if (Cs < 759) {
return 1;
}
else if (Cs >= 759) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
else if (mcTcor >= 99) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (mu_mv_mag_sq < 128) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 128) {
if (posBalance < 34) {
if (Rs < 1262) {
return 1;
}
else if (Rs >= 1262) {
return 8;
}
}
else if (posBalance >= 34) {
if (Rs < 1170) {
return 4;
}
else if (Rs >= 1170) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 69) {
if (Cs < 1103) {
if (posBalance < 40) {
if (Rs < 1388) {
return 8;
}
else if (Rs >= 1388) {
return 1;
}
}
else if (posBalance >= 40) {
return 8;
}
}
else if (Cs >= 1103) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
return 4;
}
}
}
}
else if (SC >= 2493) {
if (posBalance < 42) {
if (Rs < 2252) {
if (Rs < 1800) {
if (tcor < 97) {
if (Rs < 1738) {
return 8;
}
else if (Rs >= 1738) {
return 2;
}
}
else if (tcor >= 97) {
if (mu_mv_mag_sq < 111) {
if (Rs < 1671) {
return 1;
}
else if (Rs >= 1671) {
return 8;
}
}
else if (mu_mv_mag_sq >= 111) {
return 1;
}
}
}
else if (Rs >= 1800) {
if (ssDCval < 72) {
if (mu_mv_mag_sq < 99) {
return 4;
}
else if (mu_mv_mag_sq >= 99) {
if (Cs < 931) {
return 1;
}
else if (Cs >= 931) {
if (tcor < 99) {
if (gchDC < 1) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
else if (gchDC >= 1) {
if (Cs < 1079) {
return 4;
}
else if (Cs >= 1079) {
return 1;
}
}
}
else if (tcor >= 99) {
if (Rs < 1901) {
if (RsCsDiff < 1) {
return 1;
}
else if (RsCsDiff >= 1) {
if (Cs < 1070) {
if (Rs < 1882) {
return 4;
}
else if (Rs >= 1882) {
return 1;
}
}
else if (Cs >= 1070) {
return 1;
}
}
}
else if (Rs >= 1901) {
return 4;
}
}
}
}
}
else if (ssDCval >= 72) {
return 1;
}
}
}
else if (Rs >= 2252) {
if (Cs < 1329) {
if (AFD < 1) {
if (Rs < 2298) {
return 2;
}
else if (Rs >= 2298) {
if (Rs < 2310) {
return 8;
}
else if (Rs >= 2310) {
return 2;
}
}
}
else if (AFD >= 1) {
return 8;
}
}
else if (Cs >= 1329) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (Rs < 1481) {
if (mu_mv_mag_sq < 127) {
if (Rs < 1391) {
if (Rs < 1250) {
return 4;
}
else if (Rs >= 1250) {
return 1;
}
}
else if (Rs >= 1391) {
return 8;
}
}
else if (mu_mv_mag_sq >= 127) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
else if (Rs >= 1481) {
if (mu_mv_mag_sq < 104) {
if (Cs < 1076) {
return 1;
}
else if (Cs >= 1076) {
return 4;
}
}
else if (mu_mv_mag_sq >= 104) {
return 8;
}
}
}
}
}
else if (MVDiff >= 9) {
if (SC < 2924) {
if (Cs < 562) {
return 1;
}
else if (Cs >= 562) {
return 4;
}
}
else if (SC >= 2924) {
if (Rs < 2095) {
return 1;
}
else if (Rs >= 2095) {
return 8;
}
}
}
}
}
else if (ssDCval >= 81) {
if (SC < 2982) {
if (Rs < 2195) {
if (posBalance < 57) {
if (diffRsCsdiff < -5) {
return 8;
}
else if (diffRsCsdiff >= -5) {
return 4;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (Rs >= 2195) {
return 8;
}
}
else if (SC >= 2982) {
return 8;
}
}
}
else if (TSC >= 3) {
if (Cs < 870) {
if (RsDiff < 5) {
if (ssDCval < 85) {
if (TSC < 5) {
if (Cs < 796) {
if (ssDCval < 53) {
if (Rs < 2045) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Rs >= 2045) {
return 4;
}
}
else if (ssDCval >= 53) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
if (ssDCval < 69) {
if (ssDCval < 67) {
if (Rs < 2123) {
return 1;
}
else if (Rs >= 2123) {
return 2;
}
}
else if (ssDCval >= 67) {
return 2;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
else if (Cs >= 796) {
if (Rs < 1963) {
if (mu_mv_mag_sq < 105) {
if (diffRsCsdiff < -4) {
if (tcor < 96) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (tcor >= 96) {
return 1;
}
}
else if (diffRsCsdiff >= -4) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (Rs < 1706) {
return 8;
}
else if (Rs >= 1706) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (tcor < 100) {
return 8;
}
else if (tcor >= 100) {
return 4;
}
}
}
else if (Rs >= 1963) {
if (Rs < 2120) {
return 1;
}
else if (Rs >= 2120) {
if (Rs < 2196) {
return 2;
}
else if (Rs >= 2196) {
return 1;
}
}
}
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 54) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (mu_mv_mag_sq >= 54) {
if (posBalance < 48) {
if (TSC < 6) {
if (Rs < 2254) {
if (Rs < 1816) {
return 4;
}
else if (Rs >= 1816) {
return 4;
}
}
else if (Rs >= 2254) {
return 8;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (posBalance >= 48) {
if (Rs < 1852) {
return 8;
}
else if (Rs >= 1852) {
return 1;
}
}
}
}
}
else if (ssDCval >= 85) {
if (SC < 2552) {
return 4;
}
else if (SC >= 2552) {
return 1;
}
}
}
else if (RsDiff >= 5) {
if (refDCval < 71) {
if (Rs < 1805) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Cs < 667) {
return 2;
}
else if (Cs >= 667) {
return 1;
}
}
}
else if (Rs >= 1805) {
if (diffMVdiffVal < 7) {
if (Cs < 835) {
return 2;
}
else if (Cs >= 835) {
if (Rs < 2329) {
if (Rs < 2266) {
return 4;
}
else if (Rs >= 2266) {
return 1;
}
}
else if (Rs >= 2329) {
return 2;
}
}
}
else if (diffMVdiffVal >= 7) {
return 4;
}
}
}
else if (refDCval >= 71) {
if (RsCsDiff < 56) {
if (diffRsCsdiff < -42) {
return 2;
}
else if (diffRsCsdiff >= -42) {
if (mu_mv_mag_sq < 59) {
return 1;
}
else if (mu_mv_mag_sq >= 59) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
}
}
else if (RsCsDiff >= 56) {
if (mu_mv_mag_sq < 12) {
if (diffAFD < 1) {
if (Rs < 2166) {
return 8;
}
else if (Rs >= 2166) {
return 2;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 12) {
return 2;
}
}
}
}
}
else if (Cs >= 870) {
if (posBalance < 45) {
if (mu_mv_mag_sq < 78) {
if (AFD < 9) {
if (Cs < 945) {
return 1;
}
else if (Cs >= 945) {
if (mu_mv_mag_sq < 54) {
if (Rs < 2229) {
return 4;
}
else if (Rs >= 2229) {
return 1;
}
}
else if (mu_mv_mag_sq >= 54) {
if (refDCval < 51) {
return 1;
}
else if (refDCval >= 51) {
if (Rs < 1900) {
return 4;
}
else if (Rs >= 1900) {
return 1;
}
}
}
}
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 2) {
if (Rs < 2090) {
return 1;
}
else if (Rs >= 2090) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (ssDCval < 58) {
if (Rs < 2104) {
if (mcTcor < 97) {
return 4;
}
else if (mcTcor >= 97) {
if (Cs < 1087) {
if (Rs < 1671) {
if (Rs < 1547) {
return 4;
}
else if (Rs >= 1547) {
return 8;
}
}
else if (Rs >= 1671) {
return 4;
}
}
else if (Cs >= 1087) {
return 8;
}
}
}
else if (Rs >= 2104) {
return 4;
}
}
else if (ssDCval >= 58) {
if (Cs < 1217) {
if (diffAFD < 2) {
if (posBalance < 43) {
if (diffRsCsdiff < 5) {
return 8;
}
else if (diffRsCsdiff >= 5) {
if (Cs < 1082) {
return 8;
}
else if (Cs >= 1082) {
return 1;
}
}
}
else if (posBalance >= 43) {
if (RsDiff < 3) {
if (Cs < 994) {
return 1;
}
else if (Cs >= 994) {
return 8;
}
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (Cs >= 1217) {
return 4;
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 1427) {
if (Rs < 1155) {
if (RsCsDiff < 19) {
if (Rs < 992) {
if (posBalance < 53) {
if (Rs < 913) {
return 1;
}
else if (Rs >= 913) {
return 2;
}
}
else if (posBalance >= 53) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
return 8;
}
}
}
else if (Rs >= 992) {
if (RsCsDiff < 4) {
if (Cs < 1117) {
return 1;
}
else if (Cs >= 1117) {
return 8;
}
}
else if (RsCsDiff >= 4) {
if (Cs < 1238) {
if (Rs < 1030) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (Rs < 1018) {
return 4;
}
else if (Rs >= 1018) {
return 2;
}
}
}
else if (Rs >= 1030) {
if (mu_mv_mag_sq < 111) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Cs < 1084) {
return 8;
}
else if (Cs >= 1084) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 111) {
return 4;
}
}
}
else if (Cs >= 1238) {
return 8;
}
}
}
}
else if (RsCsDiff >= 19) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (MVDiff < 2) {
if (Rs < 957) {
return 4;
}
else if (Rs >= 957) {
return 8;
}
}
else if (MVDiff >= 2) {
if (ssDCval < 81) {
return 2;
}
else if (ssDCval >= 81) {
return 4;
}
}
}
}
}
else if (Rs >= 1155) {
if (AFD < 4) {
if (mu_mv_mag_sq < 125) {
if (negBalance < 34) {
if (SC < 2512) {
return 8;
}
else if (SC >= 2512) {
return 4;
}
}
else if (negBalance >= 34) {
if (ssDCval < 70) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (ssDCval >= 70) {
if (Rs < 1330) {
return 4;
}
else if (Rs >= 1330) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (ssDCval < 74) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (ssDCval >= 74) {
return 8;
}
}
}
else if (AFD >= 4) {
if (diffRsCsdiff < -11) {
if (SC < 2236) {
return 8;
}
else if (SC >= 2236) {
return 4;
}
}
else if (diffRsCsdiff >= -11) {
if (Cs < 1223) {
if (mu_mv_mag_sq < 123) {
if (RsCsDiff < 46) {
if (diffAFD < 1) {
if (diffTSC < -1) {
if (Rs < 1347) {
return 8;
}
else if (Rs >= 1347) {
if (Cs < 892) {
return 8;
}
else if (Cs >= 892) {
return 4;
}
}
}
else if (diffTSC >= -1) {
if (Rs < 1173) {
return 4;
}
else if (Rs >= 1173) {
if (mu_mv_mag_sq < 85) {
if (gchDC < 1) {
if (SC < 2283) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 8;
}
}
else if (SC >= 2283) {
return 1;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 85) {
return 8;
}
}
}
}
else if (diffAFD >= 1) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
return 8;
}
}
}
else if (RsCsDiff >= 46) {
return 4;
}
}
else if (mu_mv_mag_sq >= 123) {
return 4;
}
}
else if (Cs >= 1223) {
if (AFD < 6) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
else if (AFD >= 6) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1427) {
if (RsDiff < 4) {
if (posBalance < 56) {
if (SC < 3048) {
return 8;
}
else if (SC >= 3048) {
if (Rs < 2011) {
return 1;
}
else if (Rs >= 2011) {
return 8;
}
}
}
else if (posBalance >= 56) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (posBalance < 58) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 8;
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
else if (RsDiff >= 4) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
if (tcor < 92) {
if (Cs < 1304) {
return 4;
}
else if (Cs >= 1304) {
return 2;
}
}
else if (tcor >= 92) {
if (TSC < 5) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
if (Rs < 2012) {
return 8;
}
else if (Rs >= 2012) {
return 4;
}
}
}
else if (TSC >= 5) {
if (AFD < 10) {
if (diffMVdiffVal < 2) {
if (posBalance < 54) {
if (Cs < 881) {
return 2;
}
else if (Cs >= 881) {
return 1;
}
}
else if (posBalance >= 54) {
if (Cs < 1155) {
return 8;
}
else if (Cs >= 1155) {
if (Rs < 2003) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
return 1;
}
}
else if (Rs >= 2003) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 2) {
return 8;
}
}
else if (AFD >= 10) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (SC < 1310) {
if (mu_mv_mag_sq < 106) {
if (MVDiff < 5) {
if (Cs < 766) {
if (RsDiff < 1) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 2;
}
}
else if (RsDiff >= 1) {
if (ssDCval < 123) {
if (Rs < 520) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (Rs < 446) {
if (AFD < 10) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
return 4;
}
}
else if (AFD >= 10) {
return 4;
}
}
else if (Rs >= 446) {
return 8;
}
}
}
else if (Rs >= 520) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
if (Cs < 622) {
return 1;
}
else if (Cs >= 622) {
return 2;
}
}
}
}
}
else if (ssDCval >= 123) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Rs < 579) {
if (Rs < 358) {
return 2;
}
else if (Rs >= 358) {
return 1;
}
}
else if (Rs >= 579) {
return 4;
}
}
}
}
}
else if (Cs >= 766) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
if (SC < 1286) {
return 4;
}
else if (SC >= 1286) {
return 1;
}
}
}
}
else if (MVDiff >= 5) {
if (negBalance < 38) {
if (Rs < 598) {
if (mcTcor < 98) {
if (MVDiff < 11) {
if (Cs < 709) {
return 2;
}
else if (Cs >= 709) {
return 4;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (mcTcor >= 98) {
if (posBalance < 35) {
return 4;
}
else if (posBalance >= 35) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
return 2;
}
}
}
}
else if (Rs >= 598) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Rs < 673) {
return 4;
}
else if (Rs >= 673) {
return 2;
}
}
}
}
else if (negBalance >= 38) {
if (mcTcor < 97) {
if (Rs < 403) {
return 2;
}
else if (Rs >= 403) {
return 8;
}
}
else if (mcTcor >= 97) {
if (negBalance < 60) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (Rs < 512) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (Rs >= 512) {
return 2;
}
}
}
else if (negBalance >= 60) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 331) {
if (ssDCval < 122) {
return 4;
}
else if (ssDCval >= 122) {
return 2;
}
}
else if (Rs >= 331) {
if (posBalance < 53) {
if (Cs < 895) {
return 8;
}
else if (Cs >= 895) {
if (Rs < 373) {
return 4;
}
else if (Rs >= 373) {
return 8;
}
}
}
else if (posBalance >= 53) {
if (negBalance < 40) {
if (SC < 1172) {
if (ssDCval < 124) {
return 1;
}
else if (ssDCval >= 124) {
return 8;
}
}
else if (SC >= 1172) {
if (mu_mv_mag_sq < 108) {
return 2;
}
else if (mu_mv_mag_sq >= 108) {
return 8;
}
}
}
else if (negBalance >= 40) {
return 1;
}
}
}
}
}
else if (SC >= 1310) {
if (mu_mv_mag_sq < 128) {
if (mu_mv_mag_sq < 21) {
if (ssDCval < 124) {
if (TSC < 7) {
if (ssDCval < 102) {
if (Rs < 2062) {
if (Rs < 1358) {
if (CsDiff < 3) {
if (Rs < 619) {
return 2;
}
else if (Rs >= 619) {
return 1;
}
}
else if (CsDiff >= 3) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (Rs < 1056) {
return 4;
}
else if (Rs >= 1056) {
return 2;
}
}
}
}
else if (Rs >= 1358) {
if (Cs < 1119) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
if (Rs < 1994) {
return 2;
}
else if (Rs >= 1994) {
return 1;
}
}
}
else if (Cs >= 1119) {
if (Rs < 1682) {
return 8;
}
else if (Rs >= 1682) {
return 1;
}
}
}
}
else if (Rs >= 2062) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (ssDCval >= 102) {
if (Rs < 1235) {
if (negBalance < 52) {
if (RsCsDiff < 38) {
return 4;
}
else if (RsCsDiff >= 38) {
return 1;
}
}
else if (negBalance >= 52) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
return 1;
}
}
}
else if (Rs >= 1235) {
if (posBalance < 65) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (posBalance >= 65) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (Rs < 2292) {
if (MVDiff < 30) {
if (negBalance < 53) {
return 8;
}
else if (negBalance >= 53) {
if (refDCval < 96) {
if (Rs < 1150) {
return 8;
}
else if (Rs >= 1150) {
return 1;
}
}
else if (refDCval >= 96) {
return 8;
}
}
}
else if (MVDiff >= 30) {
return 1;
}
}
else if (Rs >= 2292) {
return 4;
}
}
}
else if (ssDCval >= 124) {
if (Cs < 829) {
if (mu_mv_mag_sq < 2) {
if (posBalance < 68) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (posBalance >= 68) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
return 2;
}
}
else if (Cs >= 829) {
if (Rs < 1959) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (AFD < 7) {
if (Rs < 1779) {
return 8;
}
else if (Rs >= 1779) {
return 2;
}
}
else if (AFD >= 7) {
if (Cs < 1333) {
return 2;
}
else if (Cs >= 1333) {
return 8;
}
}
}
}
else if (Rs >= 1959) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (refDCval < 114) {
if (refDCval < 100) {
if (SC < 1592) {
if (ssDCval < 89) {
if (negBalance < 29) {
return 2;
}
else if (negBalance >= 29) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
return 8;
}
}
}
else if (ssDCval >= 89) {
if (posBalance < 38) {
if (Rs < 719) {
return 1;
}
else if (Rs >= 719) {
return 8;
}
}
else if (posBalance >= 38) {
if (mu_mv_mag_sq < 75) {
if (Rs < 685) {
return 2;
}
else if (Rs >= 685) {
return 8;
}
}
else if (mu_mv_mag_sq >= 75) {
return 8;
}
}
}
}
else if (SC >= 1592) {
if (RsDiff < 3) {
if (SC < 2401) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 919) {
if (Cs < 961) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
else if (Cs >= 961) {
if (Rs < 649) {
return 8;
}
else if (Rs >= 649) {
return 4;
}
}
}
else if (Rs >= 919) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (SC >= 2401) {
if (ssDCval < 89) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (ssDCval >= 89) {
if (MVDiff < 3) {
if (SC < 2449) {
if (Rs < 1414) {
return 8;
}
else if (Rs >= 1414) {
return 4;
}
}
else if (SC >= 2449) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Rs < 1871) {
if (Rs < 1543) {
return 4;
}
else if (Rs >= 1543) {
return 8;
}
}
else if (Rs >= 1871) {
return 1;
}
}
}
}
}
else if (RsDiff >= 3) {
if (TSC < 5) {
if (MVDiff < 2) {
if (diffMVdiffVal < -1) {
if (Cs < 982) {
return 8;
}
else if (Cs >= 982) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 103) {
return 1;
}
else if (mu_mv_mag_sq >= 103) {
if (Rs < 1793) {
return 2;
}
else if (Rs >= 1793) {
return 4;
}
}
}
}
else if (TSC >= 5) {
if (negBalance < 31) {
if (Rs < 1911) {
return 4;
}
else if (Rs >= 1911) {
return 1;
}
}
else if (negBalance >= 31) {
if (MVDiff < 4) {
if (mu_mv_mag_sq < 91) {
if (Rs < 2248) {
if (Rs < 1184) {
return 8;
}
else if (Rs >= 1184) {
return 1;
}
}
else if (Rs >= 2248) {
return 8;
}
}
else if (mu_mv_mag_sq >= 91) {
return 8;
}
}
else if (MVDiff >= 4) {
if (negBalance < 52) {
if (tcor < 89) {
if (Rs < 1502) {
return 4;
}
else if (Rs >= 1502) {
return 8;
}
}
else if (tcor >= 89) {
return 8;
}
}
else if (negBalance >= 52) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 100) {
if (RsDiff < 1) {
if (Rs < 845) {
if (ssDCval < 112) {
return 1;
}
else if (ssDCval >= 112) {
return 8;
}
}
else if (Rs >= 845) {
return 8;
}
}
else if (RsDiff >= 1) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 102) {
if (mcTcor < 98) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (mcTcor >= 98) {
if (Cs < 786) {
return 8;
}
else if (Cs >= 786) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (diffMVdiffVal < -1) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
if (Rs < 534) {
return 8;
}
else if (Rs >= 534) {
return 1;
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (negBalance >= 37) {
if (ssDCval < 113) {
if (Rs < 2288) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (TSC < 4) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (Rs < 1369) {
return 1;
}
else if (Rs >= 1369) {
return 8;
}
}
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (Rs >= 2288) {
if (Cs < 889) {
return 2;
}
else if (Cs >= 889) {
return 8;
}
}
}
else if (ssDCval >= 113) {
if (Rs < 1639) {
return 8;
}
else if (Rs >= 1639) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 114) {
if (Cs < 609) {
if (Rs < 1759) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Rs < 1182) {
return 1;
}
else if (Rs >= 1182) {
return 2;
}
}
}
else if (Rs >= 1759) {
return 1;
}
}
else if (Cs >= 609) {
if (TSC < 6) {
if (negBalance < 28) {
if (mu_mv_mag_sq < 97) {
if (TSC < 4) {
if (mu_mv_mag_sq < 70) {
return 2;
}
else if (mu_mv_mag_sq >= 70) {
return 4;
}
}
else if (TSC >= 4) {
if (TSC < 5) {
if (Rs < 2014) {
return 2;
}
else if (Rs >= 2014) {
return 8;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (Cs < 1329) {
if (Rs < 551) {
return 4;
}
else if (Rs >= 551) {
if (diffAFD < -2) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 2;
}
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < 8) {
return 8;
}
else if (diffRsCsdiff >= 8) {
return 2;
}
}
}
}
else if (Cs >= 1329) {
if (Rs < 1218) {
return 4;
}
else if (Rs >= 1218) {
return 8;
}
}
}
}
else if (negBalance >= 28) {
if (RsCsDiff < 1) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
return 8;
}
}
else if (RsCsDiff >= 1) {
if (ssDCval < 138) {
if (Rs < 2346) {
if (RsCsDiff < 12) {
if (Cs < 875) {
if (Rs < 1684) {
if (Cs < 646) {
if (Rs < 1225) {
return 1;
}
else if (Rs >= 1225) {
return 8;
}
}
else if (Cs >= 646) {
return 8;
}
}
else if (Rs >= 1684) {
if (Cs < 857) {
return 8;
}
else if (Cs >= 857) {
return 1;
}
}
}
else if (Cs >= 875) {
return 8;
}
}
else if (RsCsDiff >= 12) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 71) {
return 2;
}
else if (mu_mv_mag_sq >= 71) {
if (Rs < 1794) {
if (Rs < 1569) {
return 8;
}
else if (Rs >= 1569) {
if (Cs < 974) {
return 1;
}
else if (Cs >= 974) {
return 2;
}
}
}
else if (Rs >= 1794) {
return 8;
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
else if (Rs >= 2346) {
if (Cs < 1282) {
return 1;
}
else if (Cs >= 1282) {
return 8;
}
}
}
else if (ssDCval >= 138) {
if (MVDiff < 23) {
if (ssDCval < 158) {
return 8;
}
else if (ssDCval >= 158) {
if (Rs < 606) {
return 4;
}
else if (Rs >= 606) {
return 2;
}
}
}
else if (MVDiff >= 23) {
return 4;
}
}
}
}
}
else if (TSC >= 6) {
if (ssDCval < 122) {
if (Cs < 749) {
return 2;
}
else if (Cs >= 749) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (ssDCval >= 122) {
if (mu_mv_mag_sq < 49) {
if (ssDCval < 157) {
return 1;
}
else if (ssDCval >= 157) {
return 4;
}
}
else if (mu_mv_mag_sq >= 49) {
if (diffRsCsdiff < -9) {
if (Rs < 1186) {
return 1;
}
else if (Rs >= 1186) {
return 8;
}
}
else if (diffRsCsdiff >= -9) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 123) {
if (Rs < 2155) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Cs < 897) {
return 1;
}
else if (Cs >= 897) {
return 8;
}
}
}
else if (Rs >= 2155) {
return 4;
}
}
else if (ssDCval >= 123) {
if (Rs < 1026) {
return 8;
}
else if (Rs >= 1026) {
return 1;
}
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (refDCval < 114) {
if (mcTcor < 88) {
if (refDCval < 95) {
if (Rs < 1265) {
if (MVDiff < 16) {
if (refDCval < 88) {
if (ssDCval < 57) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 8;
}
}
else if (ssDCval >= 57) {
if (Rs < 884) {
return 4;
}
else if (Rs >= 884) {
return 8;
}
}
}
else if (refDCval >= 88) {
if (Cs < 820) {
return 8;
}
else if (Cs >= 820) {
return 4;
}
}
}
else if (MVDiff >= 16) {
if (TSC < 7) {
if (Rs < 770) {
return 4;
}
else if (Rs >= 770) {
if (Rs < 1055) {
return 8;
}
else if (Rs >= 1055) {
return 1;
}
}
}
else if (TSC >= 7) {
if (mcTcor < 77) {
if (mu_mv_mag_sq < 2) {
if (CsDiff < 6) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (CsDiff >= 6) {
if (posBalance < 67) {
if (Cs < 755) {
return 1;
}
else if (Cs >= 755) {
return 4;
}
}
else if (posBalance >= 67) {
if (AFD < 36) {
return 8;
}
else if (AFD >= 36) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (refDCval < 44) {
if (CsDiff < 6) {
if (negBalance < 46) {
return 1;
}
else if (negBalance >= 46) {
return 8;
}
}
else if (CsDiff >= 6) {
if (Rs < 1024) {
return 2;
}
else if (Rs >= 1024) {
return 8;
}
}
}
else if (refDCval >= 44) {
if (refDCval < 79) {
if (MVDiff < 65) {
if (AFD < 24) {
if (RsDiff < 8) {
if (ssDCval < 46) {
return 2;
}
else if (ssDCval >= 46) {
return 4;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (AFD >= 24) {
return 2;
}
}
else if (MVDiff >= 65) {
if (tcor < 28) {
if (Cs < 1010) {
return 4;
}
else if (Cs >= 1010) {
return 8;
}
}
else if (tcor >= 28) {
if (diffAFD < -1) {
if (Rs < 681) {
return 2;
}
else if (Rs >= 681) {
return 4;
}
}
else if (diffAFD >= -1) {
if (tcor < 61) {
return 8;
}
else if (tcor >= 61) {
if (Rs < 713) {
return 4;
}
else if (Rs >= 713) {
if (Rs < 1017) {
return 1;
}
else if (Rs >= 1017) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 79) {
if (mcTcor < 57) {
if (Rs < 745) {
return 4;
}
else if (Rs >= 745) {
return 1;
}
}
else if (mcTcor >= 57) {
if (Rs < 826) {
if (Rs < 596) {
return 8;
}
else if (Rs >= 596) {
return 1;
}
}
else if (Rs >= 826) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 77) {
if (negBalance < 56) {
if (ssDCval < 59) {
if (posBalance < 70) {
if (refDCval < 50) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
return 2;
}
}
else if (refDCval >= 50) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (Rs < 625) {
return 4;
}
else if (Rs >= 625) {
return 8;
}
}
}
}
else if (posBalance >= 70) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (ssDCval >= 59) {
if (diffMVdiffVal < 14) {
if (diffRsCsdiff < -12) {
if (Rs < 531) {
return 8;
}
else if (Rs >= 531) {
return 4;
}
}
else if (diffRsCsdiff >= -12) {
if (RsDiff < 8) {
if (diffRsCsdiff < 8) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
return 4;
}
}
else if (diffRsCsdiff >= 8) {
return 1;
}
}
else if (RsDiff >= 8) {
if (ssDCval < 68) {
return 4;
}
else if (ssDCval >= 68) {
if (Rs < 1067) {
return 2;
}
else if (Rs >= 1067) {
return 1;
}
}
}
}
}
else if (diffMVdiffVal >= 14) {
if (posBalance < 66) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 2;
}
}
else if (posBalance >= 66) {
if (posBalance < 78) {
if (Rs < 1137) {
return 8;
}
else if (Rs >= 1137) {
return 2;
}
}
else if (posBalance >= 78) {
if (Rs < 448) {
return 4;
}
else if (Rs >= 448) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 56) {
if (diffTSC < 1) {
if (SC < 2209) {
if (tcor < 72) {
if (refDCval < 78) {
if (refDCval < 66) {
if (refDCval < 54) {
if (Cs < 704) {
return 8;
}
else if (Cs >= 704) {
return 4;
}
}
else if (refDCval >= 54) {
if (diffRsCsdiff < -1) {
return 8;
}
else if (diffRsCsdiff >= -1) {
if (CsDiff < 9) {
return 2;
}
else if (CsDiff >= 9) {
return 8;
}
}
}
}
else if (refDCval >= 66) {
return 4;
}
}
else if (refDCval >= 78) {
if (MVDiff < 55) {
return 8;
}
else if (MVDiff >= 55) {
if (Rs < 396) {
return 8;
}
else if (Rs >= 396) {
return 2;
}
}
}
}
else if (tcor >= 72) {
if (Rs < 482) {
return 2;
}
else if (Rs >= 482) {
return 8;
}
}
}
else if (SC >= 2209) {
if (Rs < 1046) {
return 4;
}
else if (Rs >= 1046) {
if (Rs < 1193) {
return 2;
}
else if (Rs >= 1193) {
return 1;
}
}
}
}
else if (diffTSC >= 1) {
if (tcor < 43) {
return 4;
}
else if (tcor >= 43) {
if (negBalance < 62) {
return 2;
}
else if (negBalance >= 62) {
if (Rs < 1032) {
if (Rs < 802) {
return 4;
}
else if (Rs >= 802) {
return 2;
}
}
else if (Rs >= 1032) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1265) {
if (RsCsDiff < 127) {
if (ssDCval < 65) {
if (MVDiff < 42) {
if (Cs < 913) {
if (refDCval < 56) {
if (Rs < 1336) {
return 8;
}
else if (Rs >= 1336) {
return 4;
}
}
else if (refDCval >= 56) {
return 2;
}
}
else if (Cs >= 913) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (posBalance < 60) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
return 8;
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
}
else if (MVDiff >= 42) {
if (tcor < 57) {
if (SC < 2661) {
return 1;
}
else if (SC >= 2661) {
return 8;
}
}
else if (tcor >= 57) {
return 8;
}
}
}
else if (ssDCval >= 65) {
if (ssDCval < 86) {
if (AFD < 23) {
if (negBalance < 39) {
if (tcor < 65) {
return 8;
}
else if (tcor >= 65) {
if (Rs < 2099) {
return 1;
}
else if (Rs >= 2099) {
return 4;
}
}
}
else if (negBalance >= 39) {
return 8;
}
}
else if (AFD >= 23) {
if (Rs < 1495) {
return 2;
}
else if (Rs >= 1495) {
return 1;
}
}
}
else if (ssDCval >= 86) {
if (CsDiff < 5) {
if (AFD < 16) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 1;
}
}
else if (AFD >= 16) {
return 4;
}
}
else if (CsDiff >= 5) {
if (Cs < 851) {
return 4;
}
else if (Cs >= 851) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 127) {
if (Rs < 2229) {
if (posBalance < 50) {
if (refDCval < 82) {
if (tcor < 77) {
return 2;
}
else if (tcor >= 77) {
return 8;
}
}
else if (refDCval >= 82) {
return 8;
}
}
else if (posBalance >= 50) {
if (AFD < 19) {
if (RsDiff < 12) {
if (Rs < 1353) {
return 4;
}
else if (Rs >= 1353) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (AFD >= 19) {
if (AFD < 26) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
if (SC < 2743) {
if (diffTSC < -2) {
return 8;
}
else if (diffTSC >= -2) {
return 4;
}
}
else if (SC >= 2743) {
if (Rs < 2097) {
if (negBalance < 58) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 4;
}
}
else if (negBalance >= 58) {
return 4;
}
}
else if (Rs >= 2097) {
return 4;
}
}
}
}
else if (AFD >= 26) {
if (refDCval < 70) {
if (RsCsDiff < 239) {
return 4;
}
else if (RsCsDiff >= 239) {
if (Rs < 1633) {
return 8;
}
else if (Rs >= 1633) {
return 1;
}
}
}
else if (refDCval >= 70) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2229) {
if (negBalance < 53) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 8;
}
}
else if (negBalance >= 53) {
if (Cs < 1244) {
return 1;
}
else if (Cs >= 1244) {
if (Rs < 2330) {
return 2;
}
else if (Rs >= 2330) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 95) {
if (tcor < 74) {
if (gchDC < 2) {
if (AFD < 24) {
if (AFD < 14) {
if (Rs < 2111) {
return 4;
}
else if (Rs >= 2111) {
return 8;
}
}
else if (AFD >= 14) {
return 4;
}
}
else if (AFD >= 24) {
if (tcor < 48) {
return 4;
}
else if (tcor >= 48) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
if (Rs < 1116) {
return 1;
}
else if (Rs >= 1116) {
return 8;
}
}
}
}
}
else if (gchDC >= 2) {
if (gchDC < 9) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
if (Rs < 675) {
return 2;
}
else if (Rs >= 675) {
return 1;
}
}
}
}
else if (gchDC >= 9) {
if (Cs < 667) {
return 4;
}
else if (Cs >= 667) {
return 8;
}
}
}
}
else if (tcor >= 74) {
if (RsDiff < 7) {
if (ssDCval < 106) {
if (Rs < 701) {
return 8;
}
else if (Rs >= 701) {
return 4;
}
}
else if (ssDCval >= 106) {
return 8;
}
}
else if (RsDiff >= 7) {
if (Rs < 2141) {
return 1;
}
else if (Rs >= 2141) {
return 2;
}
}
}
}
}
else if (mcTcor >= 88) {
if (AFD < 23) {
if (Cs < 876) {
if (Rs < 1531) {
if (Cs < 819) {
if (ssDCval < 92) {
if (AFD < 18) {
if (TSC < 5) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (Rs < 260) {
if (Rs < 169) {
return 1;
}
else if (Rs >= 169) {
return 2;
}
}
else if (Rs >= 260) {
if (Rs < 376) {
return 8;
}
else if (Rs >= 376) {
return 4;
}
}
}
}
else if (TSC >= 5) {
if (MVDiff < 16) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
else if (MVDiff >= 16) {
if (diffRsCsdiff < -46) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 2;
}
}
else if (diffRsCsdiff >= -46) {
if (posBalance < 57) {
if (MVDiff < 30) {
if (Rs < 832) {
if (Rs < 569) {
return 2;
}
else if (Rs >= 569) {
return 8;
}
}
else if (Rs >= 832) {
if (Rs < 927) {
return 1;
}
else if (Rs >= 927) {
return 2;
}
}
}
else if (MVDiff >= 30) {
if (Cs < 662) {
if (ssDCval < 74) {
if (diffAFD < -2) {
if (Rs < 431) {
return 8;
}
else if (Rs >= 431) {
return 4;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (ssDCval >= 74) {
if (Cs < 575) {
return 4;
}
else if (Cs >= 575) {
return 8;
}
}
}
else if (Cs >= 662) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
else if (posBalance >= 57) {
if (Cs < 771) {
if (diffMVdiffVal < 6) {
if (ssDCval < 61) {
if (RsCsDiff < 47) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
else if (RsCsDiff >= 47) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 1;
}
}
}
else if (ssDCval >= 61) {
return 4;
}
}
else if (diffMVdiffVal >= 6) {
if (Rs < 325) {
return 4;
}
else if (Rs >= 325) {
if (Rs < 997) {
return 8;
}
else if (Rs >= 997) {
if (Rs < 1205) {
return 1;
}
else if (Rs >= 1205) {
return 8;
}
}
}
}
}
else if (Cs >= 771) {
return 8;
}
}
}
}
}
}
else if (AFD >= 18) {
if (diffRsCsdiff < -13) {
if (AFD < 21) {
if (SC < 1083) {
return 1;
}
else if (SC >= 1083) {
if (posBalance < 55) {
if (Rs < 1311) {
return 2;
}
else if (Rs >= 1311) {
return 8;
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
else if (AFD >= 21) {
if (Rs < 393) {
return 4;
}
else if (Rs >= 393) {
if (MVDiff < 68) {
return 1;
}
else if (MVDiff >= 68) {
if (Rs < 617) {
return 1;
}
else if (Rs >= 617) {
return 2;
}
}
}
}
}
else if (diffRsCsdiff >= -13) {
if (diffRsCsdiff < 19) {
if (negBalance < 60) {
if (RsCsDiff < 60) {
return 4;
}
else if (RsCsDiff >= 60) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
if (posBalance < 59) {
if (Rs < 777) {
return 4;
}
else if (Rs >= 777) {
if (Rs < 1348) {
return 8;
}
else if (Rs >= 1348) {
return 2;
}
}
}
else if (posBalance >= 59) {
return 1;
}
}
}
}
else if (negBalance >= 60) {
if (Rs < 542) {
return 8;
}
else if (Rs >= 542) {
return 1;
}
}
}
else if (diffRsCsdiff >= 19) {
if (Rs < 508) {
return 2;
}
else if (Rs >= 508) {
if (Rs < 892) {
if (Rs < 658) {
return 1;
}
else if (Rs >= 658) {
return 8;
}
}
else if (Rs >= 892) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 92) {
if (tcor < 81) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (refDCval < 105) {
if (mu_mv_mag_sq < 41) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
return 8;
}
}
else if (mu_mv_mag_sq >= 41) {
return 2;
}
}
else if (refDCval >= 105) {
if (diffMVdiffVal < 2) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 8;
}
}
else if (diffMVdiffVal >= 2) {
if (negBalance < 60) {
return 8;
}
else if (negBalance >= 60) {
if (Rs < 1087) {
return 1;
}
else if (Rs >= 1087) {
return 2;
}
}
}
}
}
}
else if (tcor >= 81) {
if (MVDiff < 20) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (Rs < 955) {
return 8;
}
else if (Rs >= 955) {
return 2;
}
}
}
else if (MVDiff >= 20) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Cs < 551) {
return 1;
}
else if (Cs >= 551) {
if (Rs < 1428) {
return 2;
}
else if (Rs >= 1428) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 819) {
if (mu_mv_mag_sq < 3) {
if (gchDC < 2) {
if (Cs < 835) {
if (Rs < 788) {
return 8;
}
else if (Rs >= 788) {
return 1;
}
}
else if (Cs >= 835) {
return 4;
}
}
else if (gchDC >= 2) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 83) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (posBalance < 62) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
return 8;
}
}
else if (posBalance >= 62) {
return 4;
}
}
}
else if (ssDCval >= 83) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (MVDiff < 7) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 4;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1531) {
if (refDCval < 79) {
if (tcor < 53) {
return 4;
}
else if (tcor >= 53) {
if (MVDiff < 44) {
if (tcor < 60) {
if (Rs < 2087) {
return 8;
}
else if (Rs >= 2087) {
return 2;
}
}
else if (tcor >= 60) {
if (Cs < 863) {
return 2;
}
else if (Cs >= 863) {
if (Cs < 867) {
return 4;
}
else if (Cs >= 867) {
return 2;
}
}
}
}
else if (MVDiff >= 44) {
if (Rs < 1626) {
return 4;
}
else if (Rs >= 1626) {
return 8;
}
}
}
}
else if (refDCval >= 79) {
if (negBalance < 59) {
if (negBalance < 49) {
if (CsDiff < 4) {
if (Cs < 728) {
return 1;
}
else if (Cs >= 728) {
return 2;
}
}
else if (CsDiff >= 4) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
return 2;
}
}
}
else if (negBalance >= 49) {
if (Rs < 2189) {
if (Cs < 660) {
if (tcor < 87) {
if (ssDCval < 87) {
return 2;
}
else if (ssDCval >= 87) {
return 4;
}
}
else if (tcor >= 87) {
if (Cs < 631) {
return 1;
}
else if (Cs >= 631) {
return 2;
}
}
}
else if (Cs >= 660) {
if (RsCsDiff < 77) {
if (RsCsDiff < 69) {
if (Rs < 1694) {
return 1;
}
else if (Rs >= 1694) {
return 8;
}
}
else if (RsCsDiff >= 69) {
return 4;
}
}
else if (RsCsDiff >= 77) {
return 1;
}
}
}
else if (Rs >= 2189) {
if (Cs < 786) {
return 4;
}
else if (Cs >= 786) {
return 8;
}
}
}
}
else if (negBalance >= 59) {
if (MVDiff < 35) {
return 2;
}
else if (MVDiff >= 35) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 8;
}
}
}
}
}
}
else if (Cs >= 876) {
if (MVDiff < 7) {
if (mu_mv_mag_sq < 3) {
if (tcor < 85) {
if (diffMVdiffVal < -2) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
if (Rs < 1034) {
return 8;
}
else if (Rs >= 1034) {
return 4;
}
}
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
else if (tcor >= 85) {
if (RsCsDiff < 73) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (Cs < 1225) {
return 8;
}
else if (Cs >= 1225) {
return 1;
}
}
}
else if (RsCsDiff >= 73) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (tcor < 75) {
if (diffRsCsdiff < -13) {
if (Rs < 1383) {
return 8;
}
else if (Rs >= 1383) {
return 4;
}
}
else if (diffRsCsdiff >= -13) {
if (Cs < 885) {
return 4;
}
else if (Cs >= 885) {
return 8;
}
}
}
else if (tcor >= 75) {
if (Cs < 926) {
return 1;
}
else if (Cs >= 926) {
return 4;
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 1338) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 88) {
if (refDCval < 68) {
if (Rs < 847) {
if (MVDiff < 40) {
return 1;
}
else if (MVDiff >= 40) {
return 2;
}
}
else if (Rs >= 847) {
if (SC < 2174) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 4;
}
}
else if (SC >= 2174) {
if (diffRsCsdiff < -7) {
if (Rs < 1122) {
return 2;
}
else if (Rs >= 1122) {
return 8;
}
}
else if (diffRsCsdiff >= -7) {
return 4;
}
}
}
}
else if (refDCval >= 68) {
if (tcor < 85) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
if (Rs < 930) {
return 2;
}
else if (Rs >= 930) {
return 8;
}
}
}
}
else if (tcor >= 85) {
if (refDCval < 75) {
if (Rs < 947) {
return 8;
}
else if (Rs >= 947) {
return 1;
}
}
else if (refDCval >= 75) {
if (Cs < 1305) {
return 4;
}
else if (Cs >= 1305) {
return 2;
}
}
}
}
}
else if (ssDCval >= 88) {
if (RsCsDiff < 46) {
if (AFD < 15) {
if (Rs < 1082) {
if (Rs < 824) {
return 1;
}
else if (Rs >= 824) {
return 8;
}
}
else if (Rs >= 1082) {
return 4;
}
}
else if (AFD >= 15) {
return 4;
}
}
else if (RsCsDiff >= 46) {
if (negBalance < 51) {
if (Rs < 1148) {
return 1;
}
else if (Rs >= 1148) {
return 2;
}
}
else if (negBalance >= 51) {
if (Rs < 970) {
return 1;
}
else if (Rs >= 970) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsCsDiff < 67) {
if (Rs < 425) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
if (SC < 1178) {
return 1;
}
else if (SC >= 1178) {
return 2;
}
}
}
else if (Rs >= 425) {
if (TSC < 8) {
if (Cs < 1249) {
if (tcor < 86) {
if (mu_mv_mag_sq < 66) {
return 8;
}
else if (mu_mv_mag_sq >= 66) {
return 4;
}
}
else if (tcor >= 86) {
if (diffMVdiffVal < -1) {
if (refDCval < 69) {
return 8;
}
else if (refDCval >= 69) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (Cs >= 1249) {
if (diffMVdiffVal < 3) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (Rs < 1089) {
return 8;
}
else if (Rs >= 1089) {
return 4;
}
}
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
}
else if (TSC >= 8) {
if (MVDiff < 31) {
if (ssDCval < 73) {
if (negBalance < 49) {
if (Rs < 883) {
return 4;
}
else if (Rs >= 883) {
return 1;
}
}
else if (negBalance >= 49) {
return 8;
}
}
else if (ssDCval >= 73) {
if (Rs < 655) {
return 2;
}
else if (Rs >= 655) {
return 4;
}
}
}
else if (MVDiff >= 31) {
if (Rs < 1276) {
return 8;
}
else if (Rs >= 1276) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 67) {
if (refDCval < 91) {
if (tcor < 77) {
if (TSC < 9) {
if (Rs < 456) {
return 4;
}
else if (Rs >= 456) {
if (diffMVdiffVal < 3) {
if (Cs < 1228) {
return 8;
}
else if (Cs >= 1228) {
return 1;
}
}
else if (diffMVdiffVal >= 3) {
if (Cs < 1117) {
if (Rs < 831) {
return 8;
}
else if (Rs >= 831) {
return 4;
}
}
else if (Cs >= 1117) {
return 2;
}
}
}
}
else if (TSC >= 9) {
if (Rs < 903) {
return 2;
}
else if (Rs >= 903) {
return 4;
}
}
}
else if (tcor >= 77) {
if (Rs < 1254) {
if (mu_mv_mag_sq < 5) {
if (Rs < 936) {
if (Rs < 858) {
return 1;
}
else if (Rs >= 858) {
return 8;
}
}
else if (Rs >= 936) {
if (MVDiff < 24) {
return 2;
}
else if (MVDiff >= 24) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (MVDiff < 26) {
if (diffTSC < 1) {
if (SC < 1901) {
return 2;
}
else if (SC >= 1901) {
return 1;
}
}
else if (diffTSC >= 1) {
if (Rs < 836) {
return 4;
}
else if (Rs >= 836) {
return 2;
}
}
}
else if (MVDiff >= 26) {
return 4;
}
}
}
else if (Rs >= 1254) {
if (Cs < 1017) {
return 4;
}
else if (Cs >= 1017) {
return 8;
}
}
}
}
else if (refDCval >= 91) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 1;
}
}
}
}
}
}
else if (Rs >= 1338) {
if (TSC < 9) {
if (RsCsDiff < 87) {
if (mu_mv_mag_sq < 3) {
if (ssDCval < 79) {
if (Rs < 1691) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
return 4;
}
}
else if (Rs >= 1691) {
return 1;
}
}
else if (ssDCval >= 79) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (mu_mv_mag_sq < 62) {
if (tcor < 70) {
if (Rs < 1566) {
return 1;
}
else if (Rs >= 1566) {
if (Rs < 1603) {
return 8;
}
else if (Rs >= 1603) {
return 2;
}
}
}
else if (tcor >= 70) {
if (diffAFD < 6) {
if (Cs < 918) {
if (Cs < 893) {
return 8;
}
else if (Cs >= 893) {
return 4;
}
}
else if (Cs >= 918) {
return 8;
}
}
else if (diffAFD >= 6) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 62) {
if (Rs < 2007) {
return 1;
}
else if (Rs >= 2007) {
return 4;
}
}
}
}
else if (RsCsDiff >= 87) {
if (mu_mv_mag_sq < 5) {
if (ssDCval < 88) {
if (SC < 2773) {
return 8;
}
else if (SC >= 2773) {
if (CsDiff < 6) {
if (Cs < 908) {
return 4;
}
else if (Cs >= 908) {
return 2;
}
}
else if (CsDiff >= 6) {
return 4;
}
}
}
else if (ssDCval >= 88) {
if (ssDCval < 102) {
return 2;
}
else if (ssDCval >= 102) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (ssDCval < 65) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (posBalance < 57) {
if (SC < 3418) {
return 2;
}
else if (SC >= 3418) {
return 8;
}
}
else if (posBalance >= 57) {
if (Rs < 1416) {
return 8;
}
else if (Rs >= 1416) {
return 1;
}
}
}
}
else if (ssDCval >= 65) {
if (diffMVdiffVal < -5) {
return 8;
}
else if (diffMVdiffVal >= -5) {
if (Rs < 1622) {
if (CsDiff < 7) {
if (Cs < 1135) {
return 2;
}
else if (Cs >= 1135) {
return 1;
}
}
else if (CsDiff >= 7) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (Rs >= 1622) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
}
}
}
}
}
}
else if (TSC >= 9) {
if (tcor < 76) {
if (RsDiff < 9) {
if (negBalance < 54) {
if (Rs < 1375) {
return 4;
}
else if (Rs >= 1375) {
return 8;
}
}
else if (negBalance >= 54) {
if (Rs < 1868) {
return 1;
}
else if (Rs >= 1868) {
return 2;
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 76) {
if (MVDiff < 28) {
if (Cs < 1010) {
return 4;
}
else if (Cs >= 1010) {
return 2;
}
}
else if (MVDiff >= 28) {
return 4;
}
}
else if (ssDCval >= 76) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (MVDiff < 24) {
if (Cs < 1228) {
return 4;
}
else if (Cs >= 1228) {
return 2;
}
}
else if (MVDiff >= 24) {
return 8;
}
}
}
}
}
else if (tcor >= 76) {
if (Cs < 985) {
if (ssDCval < 99) {
if (refDCval < 91) {
return 4;
}
else if (refDCval >= 91) {
return 8;
}
}
else if (ssDCval >= 99) {
return 1;
}
}
else if (Cs >= 985) {
if (Rs < 2151) {
if (MVDiff < 24) {
if (posBalance < 52) {
if (Rs < 1944) {
return 1;
}
else if (Rs >= 1944) {
return 8;
}
}
else if (posBalance >= 52) {
if (diffMVdiffVal < -7) {
if (Cs < 1143) {
return 2;
}
else if (Cs >= 1143) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
}
else if (MVDiff >= 24) {
if (Rs < 1435) {
return 4;
}
else if (Rs >= 1435) {
if (AFD < 14) {
if (Rs < 1917) {
return 2;
}
else if (Rs >= 1917) {
return 1;
}
}
else if (AFD >= 14) {
if (diffRsCsdiff < -44) {
return 2;
}
else if (diffRsCsdiff >= -44) {
return 8;
}
}
}
}
}
else if (Rs >= 2151) {
if (tcor < 83) {
if (diffAFD < 2) {
if (Rs < 2264) {
return 1;
}
else if (Rs >= 2264) {
return 2;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (tcor >= 83) {
return 8;
}
}
}
}
}
}
}
}
}
else if (AFD >= 23) {
if (MVDiff < 34) {
if (Rs < 735) {
if (Cs < 718) {
return 8;
}
else if (Cs >= 718) {
return 1;
}
}
else if (Rs >= 735) {
if (Rs < 1726) {
if (diffRsCsdiff < 26) {
if (RsDiff < 8) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
if (diffRsCsdiff < -6) {
if (Rs < 880) {
return 4;
}
else if (Rs >= 880) {
return 8;
}
}
else if (diffRsCsdiff >= -6) {
return 2;
}
}
}
else if (RsDiff >= 8) {
if (Rs < 985) {
return 1;
}
else if (Rs >= 985) {
if (diffMVdiffVal < 11) {
return 4;
}
else if (diffMVdiffVal >= 11) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 26) {
if (refDCval < 91) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
return 2;
}
}
else if (refDCval >= 91) {
if (Rs < 1198) {
return 1;
}
else if (Rs >= 1198) {
return 4;
}
}
}
}
else if (Rs >= 1726) {
if (ssDCval < 85) {
if (refDCval < 74) {
return 2;
}
else if (refDCval >= 74) {
return 4;
}
}
else if (ssDCval >= 85) {
if (ssDCval < 101) {
return 1;
}
else if (ssDCval >= 101) {
if (ssDCval < 109) {
return 2;
}
else if (ssDCval >= 109) {
if (Rs < 2095) {
return 1;
}
else if (Rs >= 2095) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 34) {
if (refDCval < 89) {
if (Cs < 923) {
if (RsCsDiff < 51) {
if (AFD < 27) {
return 1;
}
else if (AFD >= 27) {
return 4;
}
}
else if (RsCsDiff >= 51) {
if (diffRsCsdiff < 47) {
if (MVDiff < 199) {
if (SC < 1499) {
if (posBalance < 66) {
if (Rs < 881) {
return 4;
}
else if (Rs >= 881) {
return 1;
}
}
else if (posBalance >= 66) {
return 2;
}
}
else if (SC >= 1499) {
if (refDCval < 64) {
return 8;
}
else if (refDCval >= 64) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
return 8;
}
}
}
}
else if (MVDiff >= 199) {
return 8;
}
}
else if (diffRsCsdiff >= 47) {
if (Cs < 820) {
return 8;
}
else if (Cs >= 820) {
return 4;
}
}
}
}
else if (Cs >= 923) {
if (diffAFD < 4) {
if (mu_mv_mag_sq < 2) {
if (Cs < 1011) {
return 4;
}
else if (Cs >= 1011) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (tcor < 55) {
if (Rs < 1098) {
return 1;
}
else if (Rs >= 1098) {
if (Rs < 1402) {
return 2;
}
else if (Rs >= 1402) {
return 4;
}
}
}
else if (tcor >= 55) {
return 4;
}
}
}
else if (diffAFD >= 4) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 858) {
return 4;
}
else if (Rs >= 858) {
return 2;
}
}
}
}
}
else if (refDCval >= 89) {
if (SC < 2189) {
if (RsDiff < 9) {
if (diffAFD < 3) {
if (CsDiff < 6) {
if (AFD < 40) {
return 1;
}
else if (AFD >= 40) {
return 8;
}
}
else if (CsDiff >= 6) {
if (Rs < 1135) {
return 2;
}
else if (Rs >= 1135) {
return 1;
}
}
}
else if (diffAFD >= 3) {
if (Rs < 376) {
return 8;
}
else if (Rs >= 376) {
return 1;
}
}
}
else if (RsDiff >= 9) {
if (Rs < 1474) {
if (AFD < 31) {
return 4;
}
else if (AFD >= 31) {
return 2;
}
}
else if (Rs >= 1474) {
return 1;
}
}
}
else if (SC >= 2189) {
if (Cs < 1092) {
if (gchDC < 3) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (gchDC >= 3) {
if (Cs < 593) {
return 1;
}
else if (Cs >= 593) {
return 8;
}
}
}
else if (Cs >= 1092) {
if (Rs < 1680) {
return 4;
}
else if (Rs >= 1680) {
if (Cs < 1128) {
return 8;
}
else if (Cs >= 1128) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (RsDiff < 9) {
if (SC < 1398) {
if (tcor < 103) {
if (TSC < 6) {
if (Rs < 469) {
return 4;
}
else if (Rs >= 469) {
if (Rs < 544) {
return 1;
}
else if (Rs >= 544) {
return 8;
}
}
}
else if (TSC >= 6) {
if (MVDiff < 126) {
if (refDCval < 119) {
return 4;
}
else if (refDCval >= 119) {
return 2;
}
}
else if (MVDiff >= 126) {
if (Cs < 593) {
return 1;
}
else if (Cs >= 593) {
return 4;
}
}
}
}
else if (tcor >= 103) {
return 8;
}
}
else if (SC >= 1398) {
if (AFD < 23) {
if (refDCval < 152) {
if (posBalance < 68) {
if (diffAFD < -1) {
if (posBalance < 63) {
if (Rs < 913) {
if (Cs < 765) {
return 2;
}
else if (Cs >= 765) {
return 1;
}
}
else if (Rs >= 913) {
if (TSC < 8) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (gchDC < 5) {
return 8;
}
else if (gchDC >= 5) {
return 2;
}
}
}
else if (TSC >= 8) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (diffMVdiffVal < 1) {
if (Cs < 716) {
if (Rs < 1719) {
return 8;
}
else if (Rs >= 1719) {
return 1;
}
}
else if (Cs >= 716) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 1042) {
return 1;
}
else if (Cs >= 1042) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 63) {
if (tcor < 84) {
return 4;
}
else if (tcor >= 84) {
if (SC < 2432) {
if (Rs < 1150) {
return 4;
}
else if (Rs >= 1150) {
return 8;
}
}
else if (SC >= 2432) {
return 1;
}
}
}
}
else if (diffAFD >= -1) {
if (posBalance < 65) {
if (tcor < 82) {
if (tcor < 43) {
if (Rs < 1124) {
return 1;
}
else if (Rs >= 1124) {
return 4;
}
}
else if (tcor >= 43) {
if (CsDiff < 2) {
if (Rs < 1053) {
return 8;
}
else if (Rs >= 1053) {
return 4;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
}
else if (tcor >= 82) {
if (ssDCval < 139) {
if (Rs < 1681) {
if (ssDCval < 133) {
return 8;
}
else if (ssDCval >= 133) {
return 2;
}
}
else if (Rs >= 1681) {
if (MVDiff < 16) {
if (Rs < 2196) {
return 8;
}
else if (Rs >= 2196) {
return 1;
}
}
else if (MVDiff >= 16) {
return 1;
}
}
}
else if (ssDCval >= 139) {
if (Rs < 1472) {
return 1;
}
else if (Rs >= 1472) {
return 8;
}
}
}
}
else if (posBalance >= 65) {
if (Cs < 1026) {
return 8;
}
else if (Cs >= 1026) {
return 4;
}
}
}
}
else if (posBalance >= 68) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (diffTSC < 2) {
return 2;
}
else if (diffTSC >= 2) {
return 8;
}
}
}
}
else if (refDCval >= 152) {
if (Rs < 1102) {
if (Cs < 807) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 8;
}
}
else if (Cs >= 807) {
return 2;
}
}
else if (Rs >= 1102) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
else if (AFD >= 23) {
if (diffMVdiffVal < -41) {
return 1;
}
else if (diffMVdiffVal >= -41) {
if (diffTSC < -5) {
return 1;
}
else if (diffTSC >= -5) {
if (gchDC < 14) {
return 2;
}
else if (gchDC >= 14) {
if (MVDiff < 211) {
return 8;
}
else if (MVDiff >= 211) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (tcor < 69) {
if (tcor < 36) {
if (AFD < 35) {
return 1;
}
else if (AFD >= 35) {
return 2;
}
}
else if (tcor >= 36) {
if (diffAFD < 11) {
if (mu_mv_mag_sq < 16) {
if (Rs < 1352) {
return 8;
}
else if (Rs >= 1352) {
if (diffRsCsdiff < 85) {
if (Rs < 2269) {
return 4;
}
else if (Rs >= 2269) {
if (Rs < 2297) {
return 8;
}
else if (Rs >= 2297) {
if (Rs < 2300) {
return 4;
}
else if (Rs >= 2300) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 85) {
if (Rs < 1650) {
return 2;
}
else if (Rs >= 1650) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (negBalance < 52) {
if (ssDCval < 138) {
return 1;
}
else if (ssDCval >= 138) {
if (Rs < 2004) {
return 8;
}
else if (Rs >= 2004) {
return 1;
}
}
}
else if (negBalance >= 52) {
if (Rs < 1674) {
if (Rs < 1412) {
if (Rs < 1327) {
return 8;
}
else if (Rs >= 1327) {
return 2;
}
}
else if (Rs >= 1412) {
return 4;
}
}
else if (Rs >= 1674) {
return 8;
}
}
}
}
else if (diffAFD >= 11) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
}
}
else if (tcor >= 69) {
if (Rs < 1713) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 25) {
if (Rs < 1657) {
return 8;
}
else if (Rs >= 1657) {
return 2;
}
}
else if (MVDiff >= 25) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 22) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
if (Rs < 1494) {
return 2;
}
else if (Rs >= 1494) {
return 8;
}
}
}
else if (AFD >= 22) {
if (gchDC < 6) {
if (Rs < 1270) {
return 2;
}
else if (Rs >= 1270) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (gchDC >= 6) {
if (SC < 2363) {
return 8;
}
else if (SC >= 2363) {
return 2;
}
}
}
}
}
else if (Rs >= 1713) {
if (CsDiff < 6) {
if (Cs < 844) {
if (RsCsDiff < 102) {
if (Cs < 653) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (Cs >= 653) {
return 8;
}
}
else if (RsCsDiff >= 102) {
return 1;
}
}
else if (Cs >= 844) {
if (Rs < 2179) {
return 8;
}
else if (Rs >= 2179) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (TSC < 11) {
if (Rs < 2343) {
if (ssDCval < 129) {
if (diffAFD < -1) {
if (diffRsCsdiff < -42) {
if (Rs < 1980) {
return 8;
}
else if (Rs >= 1980) {
return 1;
}
}
else if (diffRsCsdiff >= -42) {
return 2;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (ssDCval >= 129) {
if (Cs < 1239) {
return 2;
}
else if (Cs >= 1239) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 2;
}
}
}
}
else if (Rs >= 2343) {
return 1;
}
}
else if (TSC >= 11) {
if (gchDC < 2) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
return 8;
}
}
else if (gchDC >= 2) {
if (negBalance < 58) {
if (diffRsCsdiff < -85) {
if (Rs < 2001) {
return 2;
}
else if (Rs >= 2001) {
return 8;
}
}
else if (diffRsCsdiff >= -85) {
return 1;
}
}
else if (negBalance >= 58) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2376) {
if (MVDiff < 15) {
if (refDCval < 109) {
if (posBalance < 48) {
if (negBalance < 22) {
if (Cs < 1053) {
if (SC < 3162) {
return 1;
}
else if (SC >= 3162) {
if (mu_mv_mag_sq < 110) {
if (Rs < 2391) {
return 8;
}
else if (Rs >= 2391) {
return 4;
}
}
else if (mu_mv_mag_sq >= 110) {
if (posBalance < 34) {
if (Cs < 730) {
return 4;
}
else if (Cs >= 730) {
return 8;
}
}
else if (posBalance >= 34) {
if (Rs < 3417) {
return 1;
}
else if (Rs >= 3417) {
return 4;
}
}
}
}
}
else if (Cs >= 1053) {
if (SC < 4860) {
if (RsCsDiff < 5) {
if (negBalance < 21) {
return 8;
}
else if (negBalance >= 21) {
if (Cs < 1112) {
return 2;
}
else if (Cs >= 1112) {
return 8;
}
}
}
else if (RsCsDiff >= 5) {
if (RsDiff < 3) {
if (posBalance < 5) {
return 4;
}
else if (posBalance >= 5) {
return 8;
}
}
else if (RsDiff >= 3) {
return 2;
}
}
}
else if (SC >= 4860) {
if (SC < 5300) {
if (posBalance < 39) {
if (mu_mv_mag_sq < 126) {
return 4;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 4007) {
if (Rs < 3973) {
if (SC < 4966) {
return 8;
}
else if (SC >= 4966) {
return 4;
}
}
else if (Rs >= 3973) {
return 8;
}
}
else if (Rs >= 4007) {
return 4;
}
}
}
else if (posBalance >= 39) {
if (Rs < 3969) {
return 8;
}
else if (Rs >= 3969) {
return 1;
}
}
}
else if (SC >= 5300) {
return 8;
}
}
}
}
else if (negBalance >= 22) {
if (Rs < 3806) {
if (negBalance < 51) {
if (ssDCval < 103) {
if (tcor < 99) {
if (SC < 3404) {
if (Rs < 2484) {
if (ssDCval < 73) {
if (posBalance < 44) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (posBalance >= 44) {
return 2;
}
}
else if (ssDCval >= 73) {
if (diffRsCsdiff < -55) {
return 2;
}
else if (diffRsCsdiff >= -55) {
return 1;
}
}
}
else if (Rs >= 2484) {
if (Cs < 553) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 1;
}
}
else if (Cs >= 553) {
if (gchDC < 1) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
return 1;
}
}
else if (gchDC >= 1) {
if (Rs < 2500) {
return 8;
}
else if (Rs >= 2500) {
return 4;
}
}
}
}
}
else if (SC >= 3404) {
if (refDCval < 96) {
if (refDCval < 79) {
if (refDCval < 64) {
if (RsDiff < 6) {
if (Cs < 1120) {
if (Cs < 979) {
return 2;
}
else if (Cs >= 979) {
if (MVDiff < 1) {
return 2;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (Cs >= 1120) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (diffRsCsdiff < -15) {
return 4;
}
else if (diffRsCsdiff >= -15) {
return 8;
}
}
}
}
else if (RsDiff >= 6) {
if (MVDiff < 5) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 2;
}
}
else if (MVDiff >= 5) {
if (CsDiff < 6) {
if (diffAFD < -1) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 4;
}
}
}
else if (diffAFD >= -1) {
if (AFD < 17) {
if (Cs < 1289) {
if (Rs < 2590) {
if (Rs < 2523) {
return 4;
}
else if (Rs >= 2523) {
return 8;
}
}
else if (Rs >= 2590) {
return 4;
}
}
else if (Cs >= 1289) {
return 8;
}
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 2;
}
}
}
}
}
else if (refDCval >= 64) {
if (mu_mv_mag_sq < 105) {
if (posBalance < 40) {
if (tcor < 83) {
return 2;
}
else if (tcor >= 83) {
if (Cs < 915) {
if (Cs < 707) {
return 8;
}
else if (Cs >= 707) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 1;
}
}
}
else if (Cs >= 915) {
if (SC < 4451) {
return 1;
}
else if (SC >= 4451) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Cs < 1208) {
return 4;
}
else if (Cs >= 1208) {
return 2;
}
}
}
}
}
}
else if (posBalance >= 40) {
if (RsCsDiff < 37) {
return 4;
}
else if (RsCsDiff >= 37) {
if (diffTSC < -1) {
if (Cs < 925) {
return 4;
}
else if (Cs >= 925) {
if (diffRsCsdiff < -111) {
if (Rs < 3499) {
return 4;
}
else if (Rs >= 3499) {
return 1;
}
}
else if (diffRsCsdiff >= -111) {
if (ssDCval < 65) {
if (Rs < 3141) {
return 4;
}
else if (Rs >= 3141) {
return 2;
}
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
else if (diffTSC >= -1) {
if (ssDCval < 69) {
if (refDCval < 65) {
return 4;
}
else if (refDCval >= 65) {
if (Cs < 1049) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
return 4;
}
}
else if (Cs >= 1049) {
if (diffTSC < 1) {
if (CsDiff < 6) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
return 2;
}
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 78) {
if (SC < 4156) {
if (CsDiff < 5) {
if (negBalance < 44) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (negBalance >= 44) {
return 2;
}
}
else if (CsDiff >= 5) {
if (Rs < 2624) {
return 2;
}
else if (Rs >= 2624) {
return 4;
}
}
}
else if (SC >= 4156) {
if (RsDiff < 12) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
if (Rs < 3212) {
return 2;
}
else if (Rs >= 3212) {
return 4;
}
}
}
else if (RsDiff >= 12) {
if (RsCsDiff < 233) {
if (Rs < 3418) {
return 2;
}
else if (Rs >= 3418) {
return 1;
}
}
else if (RsCsDiff >= 233) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 3328) {
return 2;
}
else if (Rs >= 3328) {
return 4;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 3648) {
if (Rs < 3181) {
return 2;
}
else if (Rs >= 3181) {
if (Rs < 3394) {
return 4;
}
else if (Rs >= 3394) {
return 2;
}
}
}
else if (Rs >= 3648) {
return 1;
}
}
}
}
else if (refDCval >= 79) {
if (mu_mv_mag_sq < 58) {
if (ssDCval < 83) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
if (Rs < 3036) {
return 2;
}
else if (Rs >= 3036) {
return 4;
}
}
}
else if (ssDCval >= 83) {
if (Rs < 3519) {
return 2;
}
else if (Rs >= 3519) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (negBalance < 47) {
return 2;
}
else if (negBalance >= 47) {
if (Cs < 1069) {
if (Rs < 2987) {
return 4;
}
else if (Rs >= 2987) {
return 4;
}
}
else if (Cs >= 1069) {
return 2;
}
}
}
}
}
else if (refDCval >= 96) {
if (mu_mv_mag_sq < 91) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (negBalance < 26) {
if (Rs < 3439) {
return 8;
}
else if (Rs >= 3439) {
return 8;
}
}
else if (negBalance >= 26) {
if (Rs < 3358) {
return 4;
}
else if (Rs >= 3358) {
if (Cs < 976) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (Cs >= 976) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
return 4;
}
}
}
}
else if (tcor >= 99) {
if (ssDCval < 92) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (ssDCval >= 92) {
return 2;
}
}
}
else if (ssDCval >= 103) {
if (MVDiff < 3) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 2;
}
}
else if (MVDiff >= 3) {
if (posBalance < 35) {
if (Rs < 3047) {
return 8;
}
else if (Rs >= 3047) {
return 1;
}
}
else if (posBalance >= 35) {
return 8;
}
}
}
}
else if (negBalance >= 51) {
if (mu_mv_mag_sq < 117) {
if (RsDiff < 5) {
if (Rs < 2955) {
return 1;
}
else if (Rs >= 2955) {
return 2;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (mu_mv_mag_sq >= 117) {
if (Rs < 2568) {
return 4;
}
else if (Rs >= 2568) {
return 8;
}
}
}
}
else if (Rs >= 3806) {
if (negBalance < 46) {
if (Cs < 1085) {
if (ssDCval < 99) {
return 4;
}
else if (ssDCval >= 99) {
return 2;
}
}
else if (Cs >= 1085) {
if (ssDCval < 68) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (MVDiff < 6) {
if (Rs < 3948) {
return 2;
}
else if (Rs >= 3948) {
return 4;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (ssDCval >= 68) {
if (mu_mv_mag_sq < 125) {
if (RsCsDiff < 48) {
return 4;
}
else if (RsCsDiff >= 48) {
if (ssDCval < 93) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (negBalance < 33) {
if (Rs < 3928) {
return 4;
}
else if (Rs >= 3928) {
if (posBalance < 45) {
if (Cs < 1231) {
return 8;
}
else if (Cs >= 1231) {
return 4;
}
}
else if (posBalance >= 45) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
else if (negBalance >= 33) {
if (Cs < 1301) {
return 4;
}
else if (Cs >= 1301) {
return 4;
}
}
}
}
else if (ssDCval >= 93) {
if (Rs < 3842) {
return 8;
}
else if (Rs >= 3842) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1278) {
if (Rs < 4347) {
return 4;
}
else if (Rs >= 4347) {
return 2;
}
}
else if (Cs >= 1278) {
return 8;
}
}
}
}
}
else if (negBalance >= 46) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
return 2;
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Rs < 3904) {
if (CsDiff < 1) {
if (Cs < 867) {
return 1;
}
else if (Cs >= 867) {
return 8;
}
}
else if (CsDiff >= 1) {
if (ssDCval < 74) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Cs < 558) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
else if (Cs >= 558) {
if (Cs < 1033) {
if (Rs < 2975) {
return 2;
}
else if (Rs >= 2975) {
return 4;
}
}
else if (Cs >= 1033) {
if (Cs < 1074) {
return 4;
}
else if (Cs >= 1074) {
return 2;
}
}
}
}
}
else if (ssDCval >= 74) {
if (posBalance < 72) {
if (mu_mv_mag_sq < 29) {
if (Rs < 2749) {
if (Cs < 662) {
return 1;
}
else if (Cs >= 662) {
return 2;
}
}
else if (Rs >= 2749) {
return 8;
}
}
else if (mu_mv_mag_sq >= 29) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (Rs < 2655) {
return 1;
}
else if (Rs >= 2655) {
return 2;
}
}
}
}
else if (posBalance >= 72) {
if (diffTSC < 1) {
if (Cs < 620) {
return 4;
}
else if (Cs >= 620) {
return 4;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
}
}
}
else if (Rs >= 3904) {
return 4;
}
}
else if (CsDiff >= 3) {
if (refDCval < 65) {
if (SC < 3643) {
if (RsDiff < 13) {
if (RsDiff < 5) {
if (Rs < 2432) {
return 8;
}
else if (Rs >= 2432) {
return 4;
}
}
else if (RsDiff >= 5) {
if (SC < 3575) {
return 2;
}
else if (SC >= 3575) {
if (refDCval < 63) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 2;
}
}
else if (refDCval >= 63) {
return 2;
}
}
}
}
else if (RsDiff >= 13) {
if (Cs < 715) {
return 2;
}
else if (Cs >= 715) {
return 4;
}
}
}
else if (SC >= 3643) {
if (refDCval < 58) {
return 4;
}
else if (refDCval >= 58) {
if (tcor < 84) {
if (diffTSC < 2) {
return 4;
}
else if (diffTSC >= 2) {
return 2;
}
}
else if (tcor >= 84) {
if (mu_mv_mag_sq < 84) {
if (RsDiff < 12) {
if (Rs < 2593) {
return 4;
}
else if (Rs >= 2593) {
return 2;
}
}
else if (RsDiff >= 12) {
return 4;
}
}
else if (mu_mv_mag_sq >= 84) {
if (Cs < 996) {
return 2;
}
else if (Cs >= 996) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 65) {
if (mcTcor < 90) {
if (RsCsDiff < 238) {
if (AFD < 16) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 2;
}
}
else if (AFD >= 16) {
return 4;
}
}
else if (RsCsDiff >= 238) {
return 2;
}
}
else if (mcTcor >= 90) {
if (TSC < 6) {
if (posBalance < 53) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (negBalance < 42) {
if (tcor < 92) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (Cs < 1048) {
return 4;
}
else if (Cs >= 1048) {
if (Rs < 3207) {
return 2;
}
else if (Rs >= 3207) {
return 4;
}
}
}
}
else if (tcor >= 92) {
if (negBalance < 39) {
if (Rs < 3246) {
return 1;
}
else if (Rs >= 3246) {
return 4;
}
}
else if (negBalance >= 39) {
if (Cs < 618) {
return 1;
}
else if (Cs >= 618) {
return 2;
}
}
}
}
else if (negBalance >= 42) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (diffRsCsdiff < -176) {
return 4;
}
else if (diffRsCsdiff >= -176) {
if (gchDC < 3) {
if (mcTcor < 93) {
return 1;
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (Rs < 4116) {
if (ssDCval < 84) {
if (diffRsCsdiff < -359) {
return 4;
}
else if (diffRsCsdiff >= -359) {
if (mcTcor < 98) {
if (posBalance < 51) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
if (Rs < 2434) {
if (Rs < 2418) {
return 2;
}
else if (Rs >= 2418) {
return 4;
}
}
else if (Rs >= 2434) {
return 2;
}
}
}
else if (posBalance >= 51) {
return 2;
}
}
else if (mcTcor >= 98) {
if (Rs < 3383) {
return 4;
}
else if (Rs >= 3383) {
return 2;
}
}
}
}
else if (ssDCval >= 84) {
if (RsDiff < 8) {
if (Rs < 2879) {
if (Cs < 1261) {
return 1;
}
else if (Cs >= 1261) {
return 8;
}
}
else if (Rs >= 2879) {
return 2;
}
}
else if (RsDiff >= 8) {
if (mcTcor < 99) {
if (tcor < 68) {
if (Rs < 3480) {
return 2;
}
else if (Rs >= 3480) {
return 1;
}
}
else if (tcor >= 68) {
return 2;
}
}
else if (mcTcor >= 99) {
if (Rs < 2870) {
return 2;
}
else if (Rs >= 2870) {
return 1;
}
}
}
}
}
else if (Rs >= 4116) {
if (ssDCval < 89) {
return 2;
}
else if (ssDCval >= 89) {
if (Rs < 4305) {
return 4;
}
else if (Rs >= 4305) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsDiff < 11) {
if (Cs < 1018) {
if (tcor < 96) {
if (TSC < 9) {
if (RsDiff < 7) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
if (Rs < 2696) {
return 8;
}
else if (Rs >= 2696) {
return 4;
}
}
}
}
else if (RsDiff >= 7) {
if (gchDC < 4) {
if (Cs < 947) {
return 2;
}
else if (Cs >= 947) {
return 4;
}
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (tcor >= 96) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
}
else if (Cs >= 1018) {
if (Rs < 3281) {
if (diffAFD < -3) {
if (Rs < 2738) {
return 8;
}
else if (Rs >= 2738) {
return 4;
}
}
else if (diffAFD >= -3) {
if (refDCval < 108) {
if (Rs < 2406) {
if (Cs < 1210) {
return 4;
}
else if (Cs >= 1210) {
return 8;
}
}
else if (Rs >= 2406) {
return 8;
}
}
else if (refDCval >= 108) {
return 2;
}
}
}
else if (Rs >= 3281) {
if (ssDCval < 96) {
if (Rs < 4235) {
return 2;
}
else if (Rs >= 4235) {
if (Rs < 4481) {
if (Rs < 4377) {
return 4;
}
else if (Rs >= 4377) {
return 8;
}
}
else if (Rs >= 4481) {
return 2;
}
}
}
else if (ssDCval >= 96) {
if (MVDiff < 12) {
if (Rs < 3316) {
return 2;
}
else if (Rs >= 3316) {
return 8;
}
}
else if (MVDiff >= 12) {
return 2;
}
}
}
}
}
else if (RsDiff >= 11) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (Cs < 837) {
return 2;
}
else if (Cs >= 837) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (refDCval < 102) {
if (SC < 4069) {
return 4;
}
else if (SC >= 4069) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
if (Rs < 3108) {
return 2;
}
else if (Rs >= 3108) {
return 8;
}
}
}
}
else if (refDCval >= 102) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 2;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 3387) {
if (mcTcor < 97) {
if (RsDiff < 9) {
if (mcTcor < 95) {
if (posBalance < 40) {
if (Rs < 2394) {
return 4;
}
else if (Rs >= 2394) {
return 8;
}
}
else if (posBalance >= 40) {
if (Rs < 3277) {
return 8;
}
else if (Rs >= 3277) {
return 2;
}
}
}
else if (mcTcor >= 95) {
if (TSC < 5) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 2;
}
}
else if (TSC >= 5) {
if (gchDC < 2) {
if (diffAFD < 2) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
if (diffMVdiffVal < 3) {
if (Cs < 1296) {
return 8;
}
else if (Cs >= 1296) {
if (Rs < 2475) {
return 8;
}
else if (Rs >= 2475) {
return 1;
}
}
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (gchDC >= 2) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (Rs < 2907) {
return 1;
}
else if (Rs >= 2907) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (Cs < 644) {
if (negBalance < 45) {
if (Cs < 595) {
return 1;
}
else if (Cs >= 595) {
return 2;
}
}
else if (negBalance >= 45) {
return 2;
}
}
else if (Cs >= 644) {
if (mcTcor < 94) {
if (mu_mv_mag_sq < 13) {
if (TSC < 9) {
if (Rs < 2769) {
return 8;
}
else if (Rs >= 2769) {
return 1;
}
}
else if (TSC >= 9) {
return 4;
}
}
else if (mu_mv_mag_sq >= 13) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 3192) {
return 1;
}
else if (Rs >= 3192) {
return 2;
}
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 5) {
if (TSCindex < 3) {
if (Rs < 2692) {
return 1;
}
else if (Rs >= 2692) {
if (SC < 4167) {
return 8;
}
else if (SC >= 4167) {
return 2;
}
}
}
else if (TSCindex >= 3) {
if (Rs < 2587) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 2;
}
}
}
else if (Rs >= 2587) {
if (Cs < 536) {
if (ssDCval < 119) {
return 8;
}
else if (ssDCval >= 119) {
return 2;
}
}
else if (Cs >= 536) {
return 2;
}
}
}
}
else if (MVDiff >= 5) {
if (TSC < 6) {
if (mu_mv_mag_sq < 87) {
if (Cs < 982) {
if (SC < 3471) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
}
else if (SC >= 3471) {
if (ssDCval < 118) {
return 8;
}
else if (ssDCval >= 118) {
return 2;
}
}
}
else if (Cs >= 982) {
if (posBalance < 86) {
return 8;
}
else if (posBalance >= 86) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (Rs < 2855) {
if (Rs < 2748) {
return 1;
}
else if (Rs >= 2748) {
return 8;
}
}
else if (Rs >= 2855) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (SC < 3320) {
if (AFD < 23) {
return 1;
}
else if (AFD >= 23) {
return 2;
}
}
else if (SC >= 3320) {
if (refDCval < 116) {
if (Cs < 799) {
return 2;
}
else if (Cs >= 799) {
if (Cs < 1062) {
return 4;
}
else if (Cs >= 1062) {
if (Rs < 3201) {
return 8;
}
else if (Rs >= 3201) {
return 2;
}
}
}
}
else if (refDCval >= 116) {
if (diffRsCsdiff < -50) {
return 1;
}
else if (diffRsCsdiff >= -50) {
if (diffMVdiffVal < -16) {
return 1;
}
else if (diffMVdiffVal >= -16) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 3387) {
if (MVDiff < 8) {
if (refDCval < 138) {
if (mu_mv_mag_sq < 48) {
if (Rs < 3878) {
if (RsDiff < 9) {
if (posBalance < 58) {
if (Rs < 3530) {
if (Rs < 3485) {
return 2;
}
else if (Rs >= 3485) {
return 1;
}
}
else if (Rs >= 3530) {
return 8;
}
}
else if (posBalance >= 58) {
return 2;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (Rs >= 3878) {
return 2;
}
}
else if (mu_mv_mag_sq >= 48) {
return 2;
}
}
else if (refDCval >= 138) {
if (Rs < 3889) {
if (SC < 4885) {
if (diffRsCsdiff < -64) {
return 1;
}
else if (diffRsCsdiff >= -64) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (SC >= 4885) {
if (Cs < 1249) {
if (posBalance < 48) {
if (Cs < 1193) {
return 2;
}
else if (Cs >= 1193) {
if (Rs < 3677) {
return 2;
}
else if (Rs >= 3677) {
return 1;
}
}
}
else if (posBalance >= 48) {
return 1;
}
}
else if (Cs >= 1249) {
return 2;
}
}
}
else if (Rs >= 3889) {
if (mcTcor < 97) {
if (diffMVdiffVal < 1) {
if (RsDiff < 16) {
return 1;
}
else if (RsDiff >= 16) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
else if (mcTcor >= 97) {
if (diffMVdiffVal < -8) {
if (Rs < 4165) {
return 2;
}
else if (Rs >= 4165) {
return 1;
}
}
else if (diffMVdiffVal >= -8) {
return 2;
}
}
}
}
}
else if (MVDiff >= 8) {
if (negBalance < 54) {
if (Rs < 3950) {
if (CsDiff < 4) {
if (posBalance < 42) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
else if (posBalance >= 42) {
if (Cs < 1271) {
return 2;
}
else if (Cs >= 1271) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (negBalance < 46) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
return 8;
}
}
else if (negBalance >= 46) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
}
else if (Rs >= 3950) {
if (ssDCval < 136) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
if (diffMVdiffVal < -17) {
if (SC < 5472) {
return 1;
}
else if (SC >= 5472) {
return 2;
}
}
else if (diffMVdiffVal >= -17) {
return 2;
}
}
}
else if (ssDCval >= 136) {
if (mcTcor < 97) {
if (RsDiff < 19) {
if (ssDCval < 149) {
return 1;
}
else if (ssDCval >= 149) {
return 2;
}
}
else if (RsDiff >= 19) {
return 2;
}
}
else if (mcTcor >= 97) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (posBalance < 57) {
if (Cs < 1301) {
if (negBalance < 55) {
if (Rs < 4309) {
return 8;
}
else if (Rs >= 4309) {
return 1;
}
}
else if (negBalance >= 55) {
return 2;
}
}
else if (Cs >= 1301) {
if (Cs < 1319) {
return 1;
}
else if (Cs >= 1319) {
return 2;
}
}
}
else if (posBalance >= 57) {
if (RsDiff < 19) {
return 4;
}
else if (RsDiff >= 19) {
if (Cs < 1212) {
return 2;
}
else if (Cs >= 1212) {
return 4;
}
}
}
}
}
}
}
}
else if (MVDiff >= 15) {
if (mcTcor < 69) {
if (MVDiff < 51) {
return 1;
}
else if (MVDiff >= 51) {
if (Cs < 1038) {
if (Rs < 2453) {
return 2;
}
else if (Rs >= 2453) {
return 8;
}
}
else if (Cs >= 1038) {
if (Cs < 1128) {
return 1;
}
else if (Cs >= 1128) {
return 4;
}
}
}
}
else if (mcTcor >= 69) {
if (RsCsDiff < 243) {
if (ssDCval < 88) {
if (Rs < 3652) {
if (AFD < 11) {
if (ssDCval < 72) {
if (posBalance < 33) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (Rs < 3055) {
return 4;
}
else if (Rs >= 3055) {
return 2;
}
}
}
else if (posBalance >= 33) {
if (TSC < 6) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Cs < 979) {
return 4;
}
else if (Cs >= 979) {
return 8;
}
}
}
else if (TSC >= 6) {
if (TSC < 8) {
if (Rs < 2560) {
return 2;
}
else if (Rs >= 2560) {
return 4;
}
}
else if (TSC >= 8) {
return 2;
}
}
}
}
else if (ssDCval >= 72) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
if (diffRsCsdiff < -51) {
if (Rs < 2581) {
return 8;
}
else if (Rs >= 2581) {
return 2;
}
}
else if (diffRsCsdiff >= -51) {
if (MVDiff < 69) {
return 1;
}
else if (MVDiff >= 69) {
return 8;
}
}
}
}
}
else if (AFD >= 11) {
if (TSC < 7) {
if (diffRsCsdiff < -56) {
return 4;
}
else if (diffRsCsdiff >= -56) {
if (Rs < 2417) {
if (SC < 2576) {
return 2;
}
else if (SC >= 2576) {
return 1;
}
}
else if (Rs >= 2417) {
if (MVDiff < 24) {
if (refDCval < 66) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 4;
}
}
else if (refDCval >= 66) {
if (Rs < 2513) {
return 2;
}
else if (Rs >= 2513) {
return 4;
}
}
}
else if (MVDiff >= 24) {
return 4;
}
}
}
}
else if (TSC >= 7) {
if (MVDiff < 43) {
if (mu_mv_mag_sq < 34) {
if (tcor < 74) {
if (Rs < 2442) {
return 8;
}
else if (Rs >= 2442) {
return 1;
}
}
else if (tcor >= 74) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
if (mu_mv_mag_sq < 17) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (mu_mv_mag_sq >= 17) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (tcor < 83) {
if (tcor < 73) {
if (Rs < 3021) {
return 2;
}
else if (Rs >= 3021) {
return 4;
}
}
else if (tcor >= 73) {
if (ssDCval < 68) {
if (RsCsDiff < 133) {
if (Cs < 1102) {
if (RsDiff < 7) {
return 2;
}
else if (RsDiff >= 7) {
return 4;
}
}
else if (Cs >= 1102) {
return 2;
}
}
else if (RsCsDiff >= 133) {
return 2;
}
}
else if (ssDCval >= 68) {
return 4;
}
}
}
else if (tcor >= 83) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 2;
}
}
}
}
else if (MVDiff >= 43) {
if (Cs < 402) {
return 4;
}
else if (Cs >= 402) {
if (Cs < 603) {
return 1;
}
else if (Cs >= 603) {
if (diffAFD < 0) {
return 8;
}
else if (diffAFD >= 0) {
if (Rs < 2466) {
return 2;
}
else if (Rs >= 2466) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 3652) {
if (Cs < 1041) {
return 4;
}
else if (Cs >= 1041) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
return 4;
}
}
}
}
else if (ssDCval >= 88) {
if (mcTcor < 94) {
if (refDCval < 133) {
if (MVDiff < 32) {
if (RsDiff < 10) {
if (diffAFD < 2) {
if (Rs < 3223) {
return 8;
}
else if (Rs >= 3223) {
return 2;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (RsDiff >= 10) {
if (AFD < 23) {
if (tcor < 58) {
return 1;
}
else if (tcor >= 58) {
if (Cs < 1186) {
if (SC < 3710) {
if (mu_mv_mag_sq < 3) {
if (diffRsCsdiff < -5) {
return 2;
}
else if (diffRsCsdiff >= -5) {
if (Rs < 2615) {
return 8;
}
else if (Rs >= 2615) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
if (tcor < 84) {
return 2;
}
else if (tcor >= 84) {
if (Rs < 2719) {
return 4;
}
else if (Rs >= 2719) {
return 8;
}
}
}
}
}
else if (SC >= 3710) {
return 2;
}
}
else if (Cs >= 1186) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 4;
}
}
}
}
else if (AFD >= 23) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 1;
}
}
}
}
else if (MVDiff >= 32) {
if (AFD < 24) {
if (SC < 3167) {
if (negBalance < 51) {
return 1;
}
else if (negBalance >= 51) {
if (MVDiff < 64) {
if (posBalance < 60) {
return 1;
}
else if (posBalance >= 60) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
return 1;
}
}
}
else if (MVDiff >= 64) {
if (Cs < 298) {
return 1;
}
else if (Cs >= 298) {
return 8;
}
}
}
}
else if (SC >= 3167) {
if (diffRsCsdiff < -30) {
if (RsDiff < 13) {
if (Cs < 376) {
return 2;
}
else if (Cs >= 376) {
return 8;
}
}
else if (RsDiff >= 13) {
if (AFD < 18) {
if (Cs < 778) {
return 4;
}
else if (Cs >= 778) {
return 2;
}
}
else if (AFD >= 18) {
return 8;
}
}
}
else if (diffRsCsdiff >= -30) {
if (diffAFD < -4) {
if (Cs < 847) {
return 1;
}
else if (Cs >= 847) {
return 8;
}
}
else if (diffAFD >= -4) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (posBalance < 70) {
return 8;
}
else if (posBalance >= 70) {
if (Cs < 533) {
return 1;
}
else if (Cs >= 533) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 24) {
if (Rs < 2787) {
if (refDCval < 96) {
return 1;
}
else if (refDCval >= 96) {
if (ssDCval < 112) {
return 2;
}
else if (ssDCval >= 112) {
return 8;
}
}
}
else if (Rs >= 2787) {
if (MVDiff < 85) {
if (RsDiff < 12) {
return 1;
}
else if (RsDiff >= 12) {
return 8;
}
}
else if (MVDiff >= 85) {
return 4;
}
}
}
}
}
else if (refDCval >= 133) {
if (ssDCval < 148) {
if (gchDC < 3) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
if (Cs < 1063) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 1;
}
}
else if (Cs >= 1063) {
return 1;
}
}
}
else if (gchDC >= 3) {
if (Rs < 3347) {
if (Rs < 2707) {
return 1;
}
else if (Rs >= 2707) {
return 8;
}
}
else if (Rs >= 3347) {
return 2;
}
}
}
else if (ssDCval >= 148) {
if (RsCsDiff < 165) {
return 8;
}
else if (RsCsDiff >= 165) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (SC < 3433) {
if (CsDiff < 3) {
if (Rs < 2698) {
if (diffAFD < -3) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (diffAFD >= -3) {
if (diffTSC < -1) {
if (MVDiff < 30) {
return 1;
}
else if (MVDiff >= 30) {
return 8;
}
}
else if (diffTSC >= -1) {
if (Cs < 13) {
return 8;
}
else if (Cs >= 13) {
if (Rs < 2633) {
return 1;
}
else if (Rs >= 2633) {
return 8;
}
}
}
}
}
else if (Rs >= 2698) {
if (posBalance < 65) {
if (Rs < 2964) {
if (Cs < 197) {
if (Rs < 2739) {
return 8;
}
else if (Rs >= 2739) {
return 1;
}
}
else if (Cs >= 197) {
return 8;
}
}
else if (Rs >= 2964) {
return 4;
}
}
else if (posBalance >= 65) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 113) {
if (Rs < 2524) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (Rs >= 2524) {
if (diffMVdiffVal < -13) {
return 2;
}
else if (diffMVdiffVal >= -13) {
return 1;
}
}
}
else if (refDCval >= 113) {
if (mu_mv_mag_sq < 32) {
return 1;
}
else if (mu_mv_mag_sq >= 32) {
if (Cs < 531) {
return 2;
}
else if (Cs >= 531) {
return 1;
}
}
}
}
}
else if (SC >= 3433) {
if (RsCsDiff < 71) {
if (Rs < 3735) {
if (SC < 3591) {
if (Rs < 2565) {
return 8;
}
else if (Rs >= 2565) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (SC >= 3591) {
if (refDCval < 104) {
if (MVDiff < 30) {
if (Rs < 3060) {
return 8;
}
else if (Rs >= 3060) {
if (Rs < 3514) {
return 4;
}
else if (Rs >= 3514) {
return 8;
}
}
}
else if (MVDiff >= 30) {
return 1;
}
}
else if (refDCval >= 104) {
if (diffRsCsdiff < -17) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
return 2;
}
}
}
else if (diffRsCsdiff >= -17) {
return 8;
}
}
}
}
else if (Rs >= 3735) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
else if (RsCsDiff >= 71) {
if (refDCval < 123) {
if (TSC < 8) {
if (MVDiff < 33) {
if (mu_mv_mag_sq < 14) {
if (ssDCval < 99) {
if (SC < 4241) {
return 2;
}
else if (SC >= 4241) {
return 8;
}
}
else if (ssDCval >= 99) {
return 2;
}
}
else if (mu_mv_mag_sq >= 14) {
if (diffAFD < -5) {
if (Rs < 3069) {
return 8;
}
else if (Rs >= 3069) {
return 2;
}
}
else if (diffAFD >= -5) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
return 2;
}
}
}
}
else if (MVDiff >= 33) {
return 2;
}
}
else if (TSC >= 8) {
if (MVDiff < 24) {
if (TSC < 9) {
if (ssDCval < 98) {
if (Rs < 2964) {
return 4;
}
else if (Rs >= 2964) {
return 2;
}
}
else if (ssDCval >= 98) {
if (Cs < 672) {
return 2;
}
else if (Cs >= 672) {
return 8;
}
}
}
else if (TSC >= 9) {
if (SC < 3814) {
if (ssDCval < 108) {
return 2;
}
else if (ssDCval >= 108) {
return 1;
}
}
else if (SC >= 3814) {
return 2;
}
}
}
else if (MVDiff >= 24) {
if (diffAFD < -5) {
if (Cs < 876) {
return 1;
}
else if (Cs >= 876) {
if (Rs < 2835) {
return 8;
}
else if (Rs >= 2835) {
return 2;
}
}
}
else if (diffAFD >= -5) {
if (negBalance < 48) {
if (SC < 4054) {
if (SC < 3788) {
return 1;
}
else if (SC >= 3788) {
return 8;
}
}
else if (SC >= 4054) {
return 2;
}
}
else if (negBalance >= 48) {
if (Cs < 516) {
if (Rs < 3226) {
return 2;
}
else if (Rs >= 3226) {
return 4;
}
}
else if (Cs >= 516) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 123) {
if (CsDiff < 4) {
if (diffAFD < -7) {
if (Cs < 473) {
return 8;
}
else if (Cs >= 473) {
return 1;
}
}
else if (diffAFD >= -7) {
if (SC < 5142) {
if (diffMVdiffVal < -7) {
if (negBalance < 61) {
if (Rs < 2947) {
return 1;
}
else if (Rs >= 2947) {
return 2;
}
}
else if (negBalance >= 61) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
return 2;
}
}
else if (SC >= 5142) {
if (Rs < 4354) {
return 1;
}
else if (Rs >= 4354) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 1;
}
}
}
}
}
else if (CsDiff >= 4) {
if (tcor < 91) {
if (Rs < 2519) {
return 8;
}
else if (Rs >= 2519) {
if (posBalance < 67) {
return 1;
}
else if (posBalance >= 67) {
if (Rs < 3579) {
return 1;
}
else if (Rs >= 3579) {
return 8;
}
}
}
}
else if (tcor >= 91) {
if (Cs < 1200) {
if (Rs < 4280) {
return 1;
}
else if (Rs >= 4280) {
if (diffAFD < -8) {
return 1;
}
else if (diffAFD >= -8) {
return 2;
}
}
}
else if (Cs >= 1200) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 243) {
if (refDCval < 111) {
if (refDCval < 74) {
if (Rs < 3312) {
if (diffRsCsdiff < 172) {
if (AFD < 22) {
if (Rs < 3208) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 2;
}
}
}
else if (Rs >= 3208) {
if (Rs < 3256) {
return 4;
}
else if (Rs >= 3256) {
return 1;
}
}
}
else if (AFD >= 22) {
if (posBalance < 54) {
if (Rs < 2834) {
return 1;
}
else if (Rs >= 2834) {
return 4;
}
}
else if (posBalance >= 54) {
if (Rs < 2742) {
return 2;
}
else if (Rs >= 2742) {
if (Cs < 1115) {
return 1;
}
else if (Cs >= 1115) {
return 2;
}
}
}
}
}
else if (diffRsCsdiff >= 172) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
if (Rs < 2685) {
return 1;
}
else if (Rs >= 2685) {
return 2;
}
}
}
}
else if (Rs >= 3312) {
if (Rs < 5979) {
if (Cs < 362) {
if (Rs < 3491) {
return 4;
}
else if (Rs >= 3491) {
return 1;
}
}
else if (Cs >= 362) {
return 4;
}
}
else if (Rs >= 5979) {
return 8;
}
}
}
else if (refDCval >= 74) {
if (mcTcor < 91) {
if (RsCsDiff < 372) {
if (TSC < 10) {
if (CsDiff < 8) {
if (Rs < 4390) {
return 2;
}
else if (Rs >= 4390) {
return 4;
}
}
else if (CsDiff >= 8) {
if (Rs < 2549) {
return 8;
}
else if (Rs >= 2549) {
return 1;
}
}
}
else if (TSC >= 10) {
if (refDCval < 81) {
if (MVDiff < 32) {
return 2;
}
else if (MVDiff >= 32) {
if (Rs < 3405) {
return 1;
}
else if (Rs >= 3405) {
return 4;
}
}
}
else if (refDCval >= 81) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (SC < 4896) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
if (Rs < 2842) {
return 4;
}
else if (Rs >= 2842) {
return 8;
}
}
}
else if (SC >= 4896) {
if (AFD < 31) {
return 8;
}
else if (AFD >= 31) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 372) {
if (AFD < 23) {
if (Cs < 923) {
if (Cs < 621) {
return 2;
}
else if (Cs >= 621) {
if (Rs < 3766) {
return 2;
}
else if (Rs >= 3766) {
return 4;
}
}
}
else if (Cs >= 923) {
return 8;
}
}
else if (AFD >= 23) {
if (MVDiff < 85) {
return 2;
}
else if (MVDiff >= 85) {
if (Cs < 1118) {
return 2;
}
else if (Cs >= 1118) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (gchDC < 6) {
if (MVDiff < 23) {
if (AFD < 28) {
return 2;
}
else if (AFD >= 28) {
if (Rs < 3103) {
return 1;
}
else if (Rs >= 3103) {
return 2;
}
}
}
else if (MVDiff >= 23) {
if (Cs < 730) {
if (ssDCval < 82) {
if (AFD < 20) {
if (Rs < 3492) {
return 2;
}
else if (Rs >= 3492) {
return 4;
}
}
else if (AFD >= 20) {
return 2;
}
}
else if (ssDCval >= 82) {
if (SC < 2612) {
return 1;
}
else if (SC >= 2612) {
return 2;
}
}
}
else if (Cs >= 730) {
if (Rs < 4481) {
if (CsDiff < 4) {
if (Rs < 3717) {
return 2;
}
else if (Rs >= 3717) {
return 8;
}
}
else if (CsDiff >= 4) {
if (Rs < 3121) {
if (TSC < 10) {
if (Rs < 3092) {
return 2;
}
else if (Rs >= 3092) {
return 8;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (Rs >= 3121) {
return 2;
}
}
}
else if (Rs >= 4481) {
if (Cs < 1088) {
return 8;
}
else if (Cs >= 1088) {
return 1;
}
}
}
}
}
else if (gchDC >= 6) {
if (TSC < 11) {
if (AFD < 30) {
if (RsCsDiff < 271) {
return 1;
}
else if (RsCsDiff >= 271) {
return 2;
}
}
else if (AFD >= 30) {
return 8;
}
}
else if (TSC >= 11) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (refDCval < 89) {
return 8;
}
else if (refDCval >= 89) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 111) {
if (RsCsDiff < 425) {
if (mcTcor < 92) {
if (AFD < 19) {
if (posBalance < 59) {
if (Cs < 610) {
return 2;
}
else if (Cs >= 610) {
return 8;
}
}
else if (posBalance >= 59) {
return 4;
}
}
else if (AFD >= 19) {
if (CsDiff < 6) {
if (SC < 3886) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (Rs < 3133) {
return 1;
}
else if (Rs >= 3133) {
return 2;
}
}
}
else if (SC >= 3886) {
if (Rs < 3519) {
return 8;
}
else if (Rs >= 3519) {
if (Rs < 3599) {
return 2;
}
else if (Rs >= 3599) {
return 4;
}
}
}
}
else if (CsDiff >= 6) {
if (ssDCval < 114) {
if (Rs < 3010) {
return 4;
}
else if (Rs >= 3010) {
return 8;
}
}
else if (ssDCval >= 114) {
if (diffTSC < 0) {
if (MVDiff < 76) {
return 8;
}
else if (MVDiff >= 76) {
return 1;
}
}
else if (diffTSC >= 0) {
return 1;
}
}
}
}
}
else if (mcTcor >= 92) {
if (AFD < 16) {
if (Rs < 4209) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 4;
}
}
else if (Rs >= 4209) {
if (RsDiff < 18) {
if (diffMVdiffVal < 1) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 5379) {
return 2;
}
else if (Rs >= 5379) {
return 8;
}
}
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
else if (RsDiff >= 18) {
return 2;
}
}
}
else if (AFD >= 16) {
if (Rs < 3025) {
if (tcor < 76) {
if (diffRsCsdiff < 110) {
if (negBalance < 50) {
if (Cs < 1162) {
return 1;
}
else if (Cs >= 1162) {
return 2;
}
}
else if (negBalance >= 50) {
if (Rs < 2453) {
return 1;
}
else if (Rs >= 2453) {
return 2;
}
}
}
else if (diffRsCsdiff >= 110) {
return 1;
}
}
else if (tcor >= 76) {
if (CsDiff < 4) {
if (Cs < 425) {
return 1;
}
else if (Cs >= 425) {
return 2;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
}
else if (Rs >= 3025) {
if (tcor < 88) {
if (Cs < 1049) {
if (MVDiff < 29) {
if (Rs < 3367) {
return 1;
}
else if (Rs >= 3367) {
return 2;
}
}
else if (MVDiff >= 29) {
if (negBalance < 58) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (AFD < 27) {
return 1;
}
else if (AFD >= 27) {
return 8;
}
}
}
else if (negBalance >= 58) {
if (gchDC < 3) {
if (Rs < 3358) {
return 1;
}
else if (Rs >= 3358) {
return 2;
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
else if (Cs >= 1049) {
if (TSC < 12) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (TSC >= 12) {
return 1;
}
}
}
else if (tcor >= 88) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 425) {
if (Cs < 853) {
if (Rs < 4335) {
if (SC < 4726) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (negBalance < 49) {
if (MVDiff < 59) {
return 2;
}
else if (MVDiff >= 59) {
return 1;
}
}
else if (negBalance >= 49) {
if (gchDC < 5) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (CsDiff < 4) {
if (Cs < 341) {
return 1;
}
else if (Cs >= 341) {
return 2;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
}
else if (gchDC >= 5) {
return 1;
}
}
}
}
else if (SC >= 4726) {
if (AFD < 29) {
if (Rs < 4066) {
return 1;
}
else if (Rs >= 4066) {
return 2;
}
}
else if (AFD >= 29) {
return 8;
}
}
}
else if (Rs >= 4335) {
return 2;
}
}
else if (Cs >= 853) {
if (TSC < 10) {
if (posBalance < 66) {
if (Rs < 4507) {
if (RsCsDiff < 560) {
return 8;
}
else if (RsCsDiff >= 560) {
return 2;
}
}
else if (Rs >= 4507) {
if (diffMVdiffVal < -1) {
if (SC < 5984) {
return 2;
}
else if (SC >= 5984) {
if (Cs < 1154) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (Cs >= 1154) {
return 2;
}
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 5366) {
return 8;
}
else if (Rs >= 5366) {
return 2;
}
}
}
}
}
else if (posBalance >= 66) {
if (Rs < 4288) {
return 2;
}
else if (Rs >= 4288) {
return 1;
}
}
}
else if (TSC >= 10) {
if (gchDC < 2) {
if (ssDCval < 122) {
return 2;
}
else if (ssDCval >= 122) {
if (diffAFD < -2) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 1;
}
}
else if (diffAFD >= -2) {
if (ssDCval < 134) {
return 8;
}
else if (ssDCval >= 134) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
}
else if (gchDC >= 2) {
if (Rs < 5348) {
if (diffAFD < -2) {
if (Rs < 3363) {
return 1;
}
else if (Rs >= 3363) {
return 2;
}
}
else if (diffAFD >= -2) {
if (tcor < 67) {
if (diffTSC < 1) {
if (SC < 5437) {
return 2;
}
else if (SC >= 5437) {
return 1;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (tcor >= 67) {
if (ssDCval < 122) {
if (mcTcor < 95) {
if (Rs < 5119) {
return 2;
}
else if (Rs >= 5119) {
return 1;
}
}
else if (mcTcor >= 95) {
return 1;
}
}
else if (ssDCval >= 122) {
return 1;
}
}
}
}
else if (Rs >= 5348) {
if (RsDiff < 26) {
return 8;
}
else if (RsDiff >= 26) {
if (Cs < 1234) {
return 2;
}
else if (Cs >= 1234) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1336) {
if (mu_mv_mag_sq < 112) {
if (RsCsDiff < 101) {
if (refDCval < 105) {
if (posBalance < 45) {
if (ssDCval < 87) {
if (SC < 10484) {
if (Rs < 857) {
if (SC < 1842) {
if (Rs < 319) {
return 4;
}
else if (Rs >= 319) {
return 8;
}
}
else if (SC >= 1842) {
return 1;
}
}
else if (Rs >= 857) {
if (ssDCval < 64) {
if (Rs < 3226) {
if (mu_mv_mag_sq < 97) {
if (posBalance < 33) {
return 8;
}
else if (posBalance >= 33) {
if (RsDiff < 4) {
if (tcor < 96) {
if (Rs < 1136) {
return 1;
}
else if (Rs >= 1136) {
return 2;
}
}
else if (tcor >= 96) {
return 8;
}
}
else if (RsDiff >= 4) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (TSC < 6) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
else if (TSC >= 6) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (SC < 2958) {
if (Rs < 981) {
return 8;
}
else if (Rs >= 981) {
if (Cs < 1523) {
return 1;
}
else if (Cs >= 1523) {
return 4;
}
}
}
else if (SC >= 2958) {
if (Cs < 1517) {
if (refDCval < 54) {
if (Rs < 2335) {
return 1;
}
else if (Rs >= 2335) {
return 4;
}
}
else if (refDCval >= 54) {
return 8;
}
}
else if (Cs >= 1517) {
return 8;
}
}
}
}
else if (Rs >= 3226) {
if (mu_mv_mag_sq < 103) {
return 4;
}
else if (mu_mv_mag_sq >= 103) {
if (negBalance < 22) {
if (diffRsCsdiff < 19) {
if (Rs < 4024) {
return 4;
}
else if (Rs >= 4024) {
if (RsDiff < 5) {
if (Rs < 5496) {
return 8;
}
else if (Rs >= 5496) {
return 8;
}
}
else if (RsDiff >= 5) {
return 4;
}
}
}
else if (diffRsCsdiff >= 19) {
return 4;
}
}
else if (negBalance >= 22) {
return 4;
}
}
}
}
else if (ssDCval >= 64) {
if (refDCval < 68) {
if (mu_mv_mag_sq < 17) {
return 2;
}
else if (mu_mv_mag_sq >= 17) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 1167) {
return 1;
}
else if (Rs >= 1167) {
return 4;
}
}
}
}
else if (refDCval >= 68) {
if (Rs < 3201) {
if (Rs < 1908) {
if (SC < 2575) {
return 2;
}
else if (SC >= 2575) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 1729) {
return 4;
}
else if (Rs >= 1729) {
return 8;
}
}
}
}
else if (Rs >= 1908) {
if (negBalance < 41) {
if (MVDiff < 10) {
if (Cs < 2089) {
if (mu_mv_mag_sq < 87) {
if (Cs < 2014) {
return 1;
}
else if (Cs >= 2014) {
return 8;
}
}
else if (mu_mv_mag_sq >= 87) {
return 8;
}
}
else if (Cs >= 2089) {
return 1;
}
}
else if (MVDiff >= 10) {
if (Rs < 3115) {
return 4;
}
else if (Rs >= 3115) {
return 8;
}
}
}
else if (negBalance >= 41) {
if (posBalance < 37) {
if (Rs < 2539) {
return 4;
}
else if (Rs >= 2539) {
return 2;
}
}
else if (posBalance >= 37) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
}
}
}
else if (Rs >= 3201) {
if (SC < 5722) {
if (RsDiff < 7) {
if (negBalance < 45) {
if (Rs < 3545) {
if (tcor < 97) {
if (Cs < 2004) {
return 4;
}
else if (Cs >= 2004) {
return 1;
}
}
else if (tcor >= 97) {
if (Rs < 3398) {
return 4;
}
else if (Rs >= 3398) {
return 2;
}
}
}
else if (Rs >= 3545) {
return 4;
}
}
else if (negBalance >= 45) {
return 1;
}
}
else if (RsDiff >= 7) {
if (AFD < 9) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (SC >= 5722) {
if (Cs < 1786) {
if (AFD < 9) {
if (Rs < 4274) {
return 4;
}
else if (Rs >= 4274) {
return 8;
}
}
else if (AFD >= 9) {
if (Rs < 4841) {
return 4;
}
else if (Rs >= 4841) {
return 1;
}
}
}
else if (Cs >= 1786) {
if (CsDiff < 6) {
if (mu_mv_mag_sq < 106) {
if (Cs < 4163) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
else if (Cs >= 4163) {
return 8;
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 5174) {
return 8;
}
else if (Rs >= 5174) {
return 4;
}
}
}
else if (CsDiff >= 6) {
if (tcor < 94) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (tcor >= 94) {
return 8;
}
}
}
}
}
}
}
}
}
else if (SC >= 10484) {
if (Rs < 5715) {
return 8;
}
else if (Rs >= 5715) {
return 2;
}
}
}
else if (ssDCval >= 87) {
if (MVDiff < 7) {
if (SC < 5833) {
if (SC < 4440) {
if (Rs < 2243) {
if (Rs < 1245) {
return 4;
}
else if (Rs >= 1245) {
return 8;
}
}
else if (Rs >= 2243) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
if (Rs < 2300) {
return 1;
}
else if (Rs >= 2300) {
if (Rs < 2329) {
return 2;
}
else if (Rs >= 2329) {
return 8;
}
}
}
}
}
else if (SC >= 4440) {
if (RsCsDiff < 37) {
if (tcor < 98) {
if (Cs < 1578) {
return 4;
}
else if (Cs >= 1578) {
if (posBalance < 42) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 108) {
return 4;
}
else if (mu_mv_mag_sq >= 108) {
if (SC < 5517) {
return 4;
}
else if (SC >= 5517) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (posBalance >= 42) {
return 8;
}
}
}
else if (tcor >= 98) {
if (Rs < 2392) {
if (posBalance < 14) {
return 1;
}
else if (posBalance >= 14) {
return 4;
}
}
else if (Rs >= 2392) {
if (Rs < 3531) {
return 8;
}
else if (Rs >= 3531) {
if (Rs < 3890) {
return 2;
}
else if (Rs >= 3890) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 37) {
if (ssDCval < 98) {
if (ssDCval < 93) {
return 8;
}
else if (ssDCval >= 93) {
return 2;
}
}
else if (ssDCval >= 98) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
}
else if (SC >= 5833) {
if (refDCval < 89) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (refDCval >= 89) {
if (Cs < 1754) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (Cs >= 1754) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Rs < 3977) {
return 8;
}
else if (Rs >= 3977) {
if (Rs < 4743) {
return 2;
}
else if (Rs >= 4743) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (CsDiff < 1) {
if (Cs < 2121) {
if (Rs < 2306) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 1;
}
}
else if (Rs >= 2306) {
return 8;
}
}
else if (Cs >= 2121) {
return 4;
}
}
else if (CsDiff >= 1) {
if (Rs < 4146) {
if (posBalance < 13) {
if (Rs < 2673) {
return 2;
}
else if (Rs >= 2673) {
return 8;
}
}
else if (posBalance >= 13) {
return 8;
}
}
else if (Rs >= 4146) {
if (Cs < 1887) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 1;
}
}
else if (Cs >= 1887) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 45) {
if (mcTcor < 96) {
if (CsDiff < 7) {
if (Rs < 1479) {
if (AFD < 10) {
if (refDCval < 68) {
if (AFD < 7) {
if (Cs < 1732) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (Rs < 1021) {
return 8;
}
else if (Rs >= 1021) {
if (Rs < 1403) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (Rs >= 1403) {
return 8;
}
}
}
}
else if (Cs >= 1732) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (AFD >= 7) {
if (MVDiff < 6) {
if (diffRsCsdiff < 2) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 1332) {
return 4;
}
else if (Rs >= 1332) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
return 8;
}
}
}
}
else if (MVDiff >= 6) {
if (CsDiff < 5) {
if (Rs < 1449) {
return 1;
}
else if (Rs >= 1449) {
return 8;
}
}
else if (CsDiff >= 5) {
if (TSC < 6) {
if (Cs < 2009) {
return 1;
}
else if (Cs >= 2009) {
if (Rs < 1163) {
return 8;
}
else if (Rs >= 1163) {
return 4;
}
}
}
else if (TSC >= 6) {
if (Cs < 1364) {
return 4;
}
else if (Cs >= 1364) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 68) {
if (Cs < 1583) {
return 8;
}
else if (Cs >= 1583) {
if (diffTSC < 1) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (diffTSC >= 1) {
if (Rs < 1117) {
return 1;
}
else if (Rs >= 1117) {
return 8;
}
}
}
}
}
else if (AFD >= 10) {
if (TSC < 11) {
if (MVDiff < 19) {
if (tcor < 87) {
if (ssDCval < 75) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
if (Rs < 1328) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (Rs < 1036) {
return 2;
}
else if (Rs >= 1036) {
return 8;
}
}
}
else if (Rs >= 1328) {
return 4;
}
}
}
else if (ssDCval >= 75) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 1622) {
return 8;
}
else if (Cs >= 1622) {
return 4;
}
}
}
}
else if (tcor >= 87) {
if (diffMVdiffVal < -3) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
}
else if (MVDiff >= 19) {
if (refDCval < 60) {
return 2;
}
else if (refDCval >= 60) {
if (Cs < 1596) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (Cs >= 1596) {
return 8;
}
}
}
}
else if (TSC >= 11) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Rs < 1415) {
return 8;
}
else if (Rs >= 1415) {
return 4;
}
}
}
}
}
else if (Rs >= 1479) {
if (mcTcor < 93) {
if (Rs < 2517) {
if (Cs < 2490) {
if (gchDC < 1) {
if (mcTcor < 91) {
if (AFD < 16) {
if (posBalance < 56) {
if (Cs < 1751) {
return 8;
}
else if (Cs >= 1751) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (AFD >= 16) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Rs < 1609) {
return 4;
}
else if (Rs >= 1609) {
return 8;
}
}
}
}
else if (mcTcor >= 91) {
if (diffRsCsdiff < 18) {
if (Cs < 1484) {
if (Cs < 1481) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 2009) {
return 2;
}
else if (Rs >= 2009) {
return 8;
}
}
}
else if (Cs >= 1481) {
return 2;
}
}
else if (Cs >= 1484) {
return 8;
}
}
else if (diffRsCsdiff >= 18) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
if (MVDiff < 17) {
if (ssDCval < 94) {
if (posBalance < 69) {
if (Cs < 1784) {
if (AFD < 16) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
return 4;
}
}
else if (AFD >= 16) {
if (Rs < 1601) {
if (Rs < 1531) {
return 8;
}
else if (Rs >= 1531) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 8;
}
}
}
else if (Rs >= 1601) {
return 8;
}
}
}
else if (Cs >= 1784) {
if (Cs < 1859) {
if (diffRsCsdiff < -3) {
return 4;
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
else if (Cs >= 1859) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (posBalance >= 69) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Rs < 1663) {
return 8;
}
else if (Rs >= 1663) {
return 4;
}
}
}
}
else if (ssDCval >= 94) {
if (mu_mv_mag_sq < 22) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
if (Rs < 1711) {
return 4;
}
else if (Rs >= 1711) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 22) {
return 8;
}
}
}
else if (MVDiff >= 17) {
if (mu_mv_mag_sq < 3) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (MVDiff < 20) {
if (Rs < 1870) {
return 4;
}
else if (Rs >= 1870) {
return 8;
}
}
else if (MVDiff >= 20) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mu_mv_mag_sq < 59) {
return 8;
}
else if (mu_mv_mag_sq >= 59) {
if (SC < 3441) {
return 1;
}
else if (SC >= 3441) {
return 8;
}
}
}
}
}
}
else if (Cs >= 2490) {
return 4;
}
}
else if (Rs >= 2517) {
if (posBalance < 46) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 98) {
return 8;
}
else if (mu_mv_mag_sq >= 98) {
if (diffRsCsdiff < 45) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
return 4;
}
}
}
else if (diffRsCsdiff >= 45) {
return 4;
}
}
}
}
}
else if (mcTcor >= 93) {
if (Cs < 2529) {
if (RsDiff < 8) {
if (Cs < 1479) {
if (mcTcor < 95) {
if (Cs < 1429) {
if (diffTSC < 1) {
if (diffAFD < 2) {
if (RsCsDiff < 33) {
if (mu_mv_mag_sq < 81) {
if (Rs < 2408) {
return 1;
}
else if (Rs >= 2408) {
return 8;
}
}
else if (mu_mv_mag_sq >= 81) {
return 8;
}
}
else if (RsCsDiff >= 33) {
return 8;
}
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (diffTSC >= 1) {
if (Rs < 1953) {
return 8;
}
else if (Rs >= 1953) {
if (TSC < 8) {
if (Rs < 3736) {
return 8;
}
else if (Rs >= 3736) {
return 4;
}
}
else if (TSC >= 8) {
return 1;
}
}
}
}
else if (Cs >= 1429) {
if (SC < 3574) {
if (ssDCval < 82) {
if (Rs < 1757) {
return 4;
}
else if (Rs >= 1757) {
return 8;
}
}
else if (ssDCval >= 82) {
if (TSC < 9) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (SC >= 3574) {
if (Cs < 1477) {
if (diffAFD < -1) {
if (Rs < 2731) {
return 1;
}
else if (Rs >= 2731) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Cs >= 1477) {
return 1;
}
}
}
}
else if (mcTcor >= 95) {
if (negBalance < 43) {
if (Rs < 2253) {
if (refDCval < 90) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 1;
}
}
else if (refDCval >= 90) {
if (Rs < 1980) {
return 2;
}
else if (Rs >= 1980) {
return 8;
}
}
}
else if (Rs >= 2253) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (Rs < 2344) {
return 2;
}
else if (Rs >= 2344) {
return 4;
}
}
}
}
else if (negBalance >= 43) {
if (ssDCval < 89) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
if (MVDiff < 7) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
else if (MVDiff >= 7) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
}
else if (ssDCval >= 89) {
if (Cs < 1342) {
return 2;
}
else if (Cs >= 1342) {
return 8;
}
}
}
}
}
else if (Cs >= 1479) {
if (Cs < 2274) {
if (Rs < 4159) {
if (negBalance < 56) {
if (refDCval < 80) {
if (AFD < 10) {
if (MVDiff < 4) {
if (diffMVdiffVal < -3) {
if (Cs < 1893) {
return 8;
}
else if (Cs >= 1893) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (MVDiff >= 4) {
if (Cs < 1690) {
if (Rs < 3217) {
return 8;
}
else if (Rs >= 3217) {
return 2;
}
}
else if (Cs >= 1690) {
if (tcor < 94) {
if (Rs < 3387) {
return 4;
}
else if (Rs >= 3387) {
return 8;
}
}
else if (tcor >= 94) {
return 8;
}
}
}
}
else if (AFD >= 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (AFD < 13) {
if (diffRsCsdiff < 14) {
return 8;
}
else if (diffRsCsdiff >= 14) {
return 4;
}
}
else if (AFD >= 13) {
if (Rs < 1831) {
if (Rs < 1735) {
return 8;
}
else if (Rs >= 1735) {
return 1;
}
}
else if (Rs >= 1831) {
return 4;
}
}
}
}
}
else if (refDCval >= 80) {
if (refDCval < 84) {
if (negBalance < 36) {
if (Rs < 2690) {
return 8;
}
else if (Rs >= 2690) {
return 1;
}
}
else if (negBalance >= 36) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
else if (refDCval >= 84) {
if (MVDiff < 28) {
if (diffRsCsdiff < -20) {
if (TSC < 7) {
if (Cs < 1683) {
if (Cs < 1615) {
if (Cs < 1551) {
return 2;
}
else if (Cs >= 1551) {
return 8;
}
}
else if (Cs >= 1615) {
return 2;
}
}
else if (Cs >= 1683) {
return 8;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (diffRsCsdiff >= -20) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Rs < 2989) {
return 8;
}
else if (Rs >= 2989) {
return 4;
}
}
}
}
else if (MVDiff >= 28) {
if (Rs < 2406) {
return 8;
}
else if (Rs >= 2406) {
return 4;
}
}
}
}
}
else if (negBalance >= 56) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
if (Rs < 1756) {
return 2;
}
else if (Rs >= 1756) {
return 8;
}
}
}
}
else if (Rs >= 4159) {
if (ssDCval < 84) {
if (Rs < 4562) {
return 2;
}
else if (Rs >= 4562) {
return 4;
}
}
else if (ssDCval >= 84) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
}
}
else if (Cs >= 2274) {
if (Rs < 2204) {
if (Rs < 1814) {
return 8;
}
else if (Rs >= 1814) {
return 1;
}
}
else if (Rs >= 2204) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
}
else if (RsDiff >= 8) {
if (AFD < 9) {
if (Cs < 2105) {
return 2;
}
else if (Cs >= 2105) {
if (Rs < 3825) {
return 1;
}
else if (Rs >= 3825) {
return 8;
}
}
}
else if (AFD >= 9) {
if (ssDCval < 88) {
if (tcor < 87) {
if (ssDCval < 85) {
return 4;
}
else if (ssDCval >= 85) {
return 2;
}
}
else if (tcor >= 87) {
if (Rs < 3024) {
return 8;
}
else if (Rs >= 3024) {
return 2;
}
}
}
else if (ssDCval >= 88) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Rs < 3313) {
return 8;
}
else if (Rs >= 3313) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 2529) {
if (refDCval < 81) {
if (mu_mv_mag_sq < 100) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
return 8;
}
}
else if (mu_mv_mag_sq >= 100) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 4502) {
return 4;
}
else if (Rs >= 4502) {
return 2;
}
}
}
}
else if (refDCval >= 81) {
if (Rs < 3843) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (Rs >= 3843) {
return 8;
}
}
}
}
}
}
else if (CsDiff >= 7) {
if (ssDCval < 101) {
if (Rs < 1436) {
if (TSC < 7) {
if (mu_mv_mag_sq < 2) {
if (Rs < 868) {
return 8;
}
else if (Rs >= 868) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (TSC < 6) {
if (Cs < 1953) {
return 1;
}
else if (Cs >= 1953) {
return 8;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (TSC >= 7) {
if (Cs < 1991) {
if (MVDiff < 13) {
if (tcor < 77) {
return 4;
}
else if (tcor >= 77) {
if (RsCsDiff < 77) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
return 8;
}
}
else if (RsCsDiff >= 77) {
if (Rs < 940) {
return 4;
}
else if (Rs >= 940) {
return 8;
}
}
}
}
else if (MVDiff >= 13) {
if (diffAFD < -2) {
if (tcor < 74) {
return 2;
}
else if (tcor >= 74) {
if (Rs < 961) {
return 8;
}
else if (Rs >= 961) {
return 4;
}
}
}
else if (diffAFD >= -2) {
if (diffTSC < -1) {
if (Rs < 1150) {
if (Rs < 954) {
return 8;
}
else if (Rs >= 954) {
return 2;
}
}
else if (Rs >= 1150) {
return 4;
}
}
else if (diffTSC >= -1) {
if (posBalance < 78) {
if (diffRsCsdiff < -38) {
return 8;
}
else if (diffRsCsdiff >= -38) {
return 4;
}
}
else if (posBalance >= 78) {
return 8;
}
}
}
}
}
else if (Cs >= 1991) {
return 2;
}
}
}
else if (Rs >= 1436) {
if (Cs < 2767) {
if (MVDiff < 55) {
if (Cs < 2327) {
if (ssDCval < 55) {
if (TSC < 9) {
if (TSC < 6) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (TSC >= 6) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
else if (TSC >= 9) {
if (Cs < 1492) {
return 8;
}
else if (Cs >= 1492) {
return 4;
}
}
}
else if (ssDCval >= 55) {
if (AFD < 19) {
if (refDCval < 98) {
if (gchDC < 2) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 39) {
if (negBalance < 48) {
if (Rs < 2063) {
return 8;
}
else if (Rs >= 2063) {
return 4;
}
}
else if (negBalance >= 48) {
if (posBalance < 64) {
if (AFD < 18) {
if (diffRsCsdiff < -22) {
if (Rs < 1666) {
return 8;
}
else if (Rs >= 1666) {
return 4;
}
}
else if (diffRsCsdiff >= -22) {
return 8;
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (posBalance >= 64) {
return 4;
}
}
}
else if (MVDiff >= 39) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
}
}
}
else if (gchDC >= 2) {
if (Cs < 1347) {
return 1;
}
else if (Cs >= 1347) {
if (diffAFD < -4) {
if (Rs < 2439) {
return 2;
}
else if (Rs >= 2439) {
return 8;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
}
}
else if (refDCval >= 98) {
if (Rs < 2133) {
if (Rs < 1925) {
return 4;
}
else if (Rs >= 1925) {
return 2;
}
}
else if (Rs >= 2133) {
return 8;
}
}
}
else if (AFD >= 19) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
return 4;
}
}
}
}
else if (Cs >= 2327) {
if (mcTcor < 91) {
if (Rs < 1728) {
return 4;
}
else if (Rs >= 1728) {
return 8;
}
}
else if (mcTcor >= 91) {
if (tcor < 91) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (tcor >= 91) {
if (posBalance < 59) {
if (Rs < 2286) {
return 1;
}
else if (Rs >= 2286) {
return 8;
}
}
else if (posBalance >= 59) {
return 4;
}
}
}
}
}
else if (MVDiff >= 55) {
return 1;
}
}
else if (Cs >= 2767) {
if (SC < 7261) {
if (ssDCval < 81) {
if (Cs < 2973) {
return 8;
}
else if (Cs >= 2973) {
if (Rs < 2863) {
return 4;
}
else if (Rs >= 2863) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
return 8;
}
}
}
}
else if (ssDCval >= 81) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (Rs < 2925) {
return 4;
}
else if (Rs >= 2925) {
return 1;
}
}
}
}
else if (SC >= 7261) {
if (Rs < 6017) {
return 8;
}
else if (Rs >= 6017) {
return 2;
}
}
}
}
}
else if (ssDCval >= 101) {
if (mu_mv_mag_sq < 88) {
return 8;
}
else if (mu_mv_mag_sq >= 88) {
if (Cs < 2901) {
return 1;
}
else if (Cs >= 2901) {
return 8;
}
}
}
}
}
else if (mcTcor >= 96) {
if (RsDiff < 6) {
if (Rs < 760) {
if (ssDCval < 73) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 612) {
if (Rs < 526) {
return 1;
}
else if (Rs >= 526) {
return 8;
}
}
else if (Rs >= 612) {
if (Cs < 1923) {
return 2;
}
else if (Cs >= 1923) {
return 1;
}
}
}
}
else if (ssDCval >= 73) {
if (Rs < 386) {
return 1;
}
else if (Rs >= 386) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (Rs < 700) {
return 4;
}
else if (Rs >= 700) {
return 8;
}
}
}
}
}
else if (Rs >= 760) {
if (Rs < 6046) {
if (Rs < 1503) {
if (Cs < 1609) {
if (MVDiff < 5) {
if (refDCval < 102) {
if (Rs < 1224) {
if (diffRsCsdiff < -14) {
return 4;
}
else if (diffRsCsdiff >= -14) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
else if (Rs >= 1224) {
return 4;
}
}
else if (refDCval >= 102) {
return 4;
}
}
else if (MVDiff >= 5) {
if (Rs < 1122) {
if (tcor < 89) {
if (AFD < 11) {
if (Cs < 1363) {
return 2;
}
else if (Cs >= 1363) {
return 1;
}
}
else if (AFD >= 11) {
return 8;
}
}
else if (tcor >= 89) {
if (TSC < 4) {
if (Rs < 924) {
return 8;
}
else if (Rs >= 924) {
return 1;
}
}
else if (TSC >= 4) {
return 2;
}
}
}
else if (Rs >= 1122) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
}
else if (Cs >= 1609) {
if (TSC < 6) {
if (mu_mv_mag_sq < 91) {
if (ssDCval < 69) {
if (Cs < 1982) {
if (Rs < 1004) {
return 1;
}
else if (Rs >= 1004) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (Cs >= 1982) {
return 1;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (mu_mv_mag_sq >= 91) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (TSC >= 6) {
if (tcor < 89) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 2;
}
}
else if (tcor >= 89) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (Cs < 1800) {
return 1;
}
else if (Cs >= 1800) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1503) {
if (Cs < 3299) {
if (Rs < 2676) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2152) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 2;
}
}
else if (Rs >= 2152) {
if (Cs < 1495) {
return 1;
}
else if (Cs >= 1495) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 53) {
if (ssDCval < 88) {
if (mu_mv_mag_sq < 88) {
if (TSC < 6) {
if (Cs < 2524) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
}
}
else if (Cs >= 2524) {
return 4;
}
}
else if (TSC >= 6) {
if (SC < 3460) {
return 8;
}
else if (SC >= 3460) {
if (Rs < 2411) {
if (negBalance < 42) {
if (Rs < 1990) {
return 1;
}
else if (Rs >= 1990) {
return 4;
}
}
else if (negBalance >= 42) {
return 1;
}
}
else if (Rs >= 2411) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 88) {
if (Cs < 2587) {
if (mu_mv_mag_sq < 111) {
if (Cs < 1580) {
if (Rs < 2147) {
if (Cs < 1554) {
return 8;
}
else if (Cs >= 1554) {
return 4;
}
}
else if (Rs >= 2147) {
if (Rs < 2318) {
if (Rs < 2188) {
return 1;
}
else if (Rs >= 2188) {
return 2;
}
}
else if (Rs >= 2318) {
return 4;
}
}
}
else if (Cs >= 1580) {
if (diffMVdiffVal < -2) {
if (AFD < 8) {
if (mu_mv_mag_sq < 107) {
return 8;
}
else if (mu_mv_mag_sq >= 107) {
return 4;
}
}
else if (AFD >= 8) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 111) {
return 4;
}
}
else if (Cs >= 2587) {
return 4;
}
}
}
else if (ssDCval >= 88) {
if (Rs < 2069) {
if (diffAFD < -2) {
if (MVDiff < 9) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
return 2;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (diffAFD >= -2) {
if (CsDiff < 6) {
if (Cs < 1424) {
if (Rs < 1737) {
return 8;
}
else if (Rs >= 1737) {
return 2;
}
}
else if (Cs >= 1424) {
return 8;
}
}
else if (CsDiff >= 6) {
if (Rs < 1733) {
return 2;
}
else if (Rs >= 1733) {
return 8;
}
}
}
}
else if (Rs >= 2069) {
return 8;
}
}
}
else if (negBalance >= 53) {
return 4;
}
}
}
else if (Rs >= 2676) {
if (Cs < 2030) {
if (Cs < 1482) {
if (CsDiff < 4) {
if (Cs < 1460) {
return 8;
}
else if (Cs >= 1460) {
if (Rs < 3036) {
return 8;
}
else if (Rs >= 3036) {
return 2;
}
}
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (Cs >= 1482) {
if (negBalance < 37) {
if (ssDCval < 87) {
if (RsCsDiff < 46) {
if (SC < 4510) {
if (Rs < 2906) {
return 4;
}
else if (Rs >= 2906) {
return 2;
}
}
else if (SC >= 4510) {
return 4;
}
}
else if (RsCsDiff >= 46) {
return 1;
}
}
else if (ssDCval >= 87) {
return 2;
}
}
else if (negBalance >= 37) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
if (Rs < 3562) {
return 4;
}
else if (Rs >= 3562) {
return 8;
}
}
}
}
}
}
else if (Cs >= 2030) {
if (negBalance < 41) {
if (AFD < 7) {
if (RsCsDiff < 46) {
if (ssDCval < 54) {
if (Rs < 3401) {
return 8;
}
else if (Rs >= 3401) {
return 4;
}
}
else if (ssDCval >= 54) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 3635) {
return 4;
}
else if (Rs >= 3635) {
return 8;
}
}
}
}
else if (RsCsDiff >= 46) {
return 4;
}
}
else if (AFD >= 7) {
return 4;
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
}
else if (Cs >= 3299) {
return 8;
}
}
}
else if (Rs >= 6046) {
return 2;
}
}
}
else if (RsDiff >= 6) {
if (negBalance < 41) {
if (Rs < 2791) {
if (posBalance < 58) {
if (Rs < 2683) {
if (diffRsCsdiff < 26) {
return 8;
}
else if (diffRsCsdiff >= 26) {
return 1;
}
}
else if (Rs >= 2683) {
return 1;
}
}
else if (posBalance >= 58) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
if (refDCval < 93) {
return 4;
}
else if (refDCval >= 93) {
return 8;
}
}
}
}
else if (Rs >= 2791) {
if (refDCval < 94) {
if (AFD < 10) {
if (diffRsCsdiff < -9) {
return 2;
}
else if (diffRsCsdiff >= -9) {
if (RsCsDiff < 74) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
if (Cs < 1496) {
return 2;
}
else if (Cs >= 1496) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
return 2;
}
}
}
}
else if (RsCsDiff >= 74) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (AFD >= 10) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
else if (refDCval >= 94) {
if (mu_mv_mag_sq < 84) {
return 8;
}
else if (mu_mv_mag_sq >= 84) {
if (Rs < 3492) {
return 4;
}
else if (Rs >= 3492) {
return 2;
}
}
}
}
}
else if (negBalance >= 41) {
if (Rs < 2010) {
if (mu_mv_mag_sq < 7) {
return 2;
}
else if (mu_mv_mag_sq >= 7) {
if (AFD < 13) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (Rs < 1717) {
return 8;
}
else if (Rs >= 1717) {
return 2;
}
}
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (Rs >= 2010) {
if (Cs < 1656) {
if (mu_mv_mag_sq < 75) {
if (mu_mv_mag_sq < 62) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 62) {
if (ssDCval < 90) {
if (Rs < 2942) {
return 1;
}
else if (Rs >= 2942) {
return 2;
}
}
else if (ssDCval >= 90) {
if (Rs < 2267) {
return 1;
}
else if (Rs >= 2267) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (Rs < 2816) {
return 1;
}
else if (Rs >= 2816) {
return 2;
}
}
}
else if (Cs >= 1656) {
if (ssDCval < 90) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
return 8;
}
}
else if (ssDCval >= 90) {
if (Cs < 1977) {
return 8;
}
else if (Cs >= 1977) {
if (Rs < 2460) {
return 2;
}
else if (Rs >= 2460) {
return 4;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 105) {
if (ssDCval < 163) {
if (Rs < 4405) {
if (mu_mv_mag_sq < 46) {
if (Rs < 1844) {
if (ssDCval < 146) {
if (refDCval < 128) {
if (AFD < 13) {
if (negBalance < 46) {
if (SC < 3372) {
if (diffRsCsdiff < -5) {
return 1;
}
else if (diffRsCsdiff >= -5) {
return 8;
}
}
else if (SC >= 3372) {
if (Rs < 1526) {
return 4;
}
else if (Rs >= 1526) {
return 2;
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
else if (AFD >= 13) {
if (tcor < 90) {
if (CsDiff < 6) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 8;
}
}
else if (CsDiff >= 6) {
if (refDCval < 122) {
return 8;
}
else if (refDCval >= 122) {
if (Rs < 1388) {
if (Rs < 1041) {
return 2;
}
else if (Rs >= 1041) {
return 8;
}
}
else if (Rs >= 1388) {
return 4;
}
}
}
}
else if (tcor >= 90) {
if (RsCsDiff < 67) {
return 8;
}
else if (RsCsDiff >= 67) {
return 2;
}
}
}
}
else if (refDCval >= 128) {
if (mcTcor < 90) {
if (TSC < 8) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (mcTcor >= 90) {
if (CsDiff < 7) {
if (refDCval < 129) {
return 1;
}
else if (refDCval >= 129) {
return 8;
}
}
else if (CsDiff >= 7) {
if (Rs < 1324) {
return 1;
}
else if (Rs >= 1324) {
if (Rs < 1447) {
return 4;
}
else if (Rs >= 1447) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 146) {
if (Rs < 1185) {
return 4;
}
else if (Rs >= 1185) {
if (Rs < 1692) {
if (mu_mv_mag_sq < 38) {
return 2;
}
else if (mu_mv_mag_sq >= 38) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
if (Rs < 1336) {
return 4;
}
else if (Rs >= 1336) {
return 8;
}
}
}
}
else if (Rs >= 1692) {
return 8;
}
}
}
}
else if (Rs >= 1844) {
if (refDCval < 110) {
if (mcTcor < 95) {
if (Cs < 1517) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
if (Rs < 2486) {
return 8;
}
else if (Rs >= 2486) {
return 4;
}
}
}
else if (Cs >= 1517) {
return 8;
}
}
else if (mcTcor >= 95) {
if (MVDiff < 10) {
if (SC < 4180) {
if (Cs < 1702) {
if (Rs < 2332) {
return 8;
}
else if (Rs >= 2332) {
return 4;
}
}
else if (Cs >= 1702) {
return 2;
}
}
else if (SC >= 4180) {
return 8;
}
}
else if (MVDiff >= 10) {
if (Cs < 1396) {
if (Rs < 2730) {
return 4;
}
else if (Rs >= 2730) {
return 8;
}
}
else if (Cs >= 1396) {
return 4;
}
}
}
}
else if (refDCval >= 110) {
if (Cs < 2820) {
if (mcTcor < 98) {
if (Rs < 2017) {
if (negBalance < 49) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (negBalance >= 49) {
return 8;
}
}
else if (Rs >= 2017) {
if (Rs < 4179) {
return 8;
}
else if (Rs >= 4179) {
if (Rs < 4230) {
return 1;
}
else if (Rs >= 4230) {
return 8;
}
}
}
}
else if (mcTcor >= 98) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Cs < 2215) {
return 2;
}
else if (Cs >= 2215) {
return 8;
}
}
}
}
else if (Cs >= 2820) {
if (Rs < 3187) {
return 2;
}
else if (Rs >= 3187) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (Rs < 1285) {
if (Cs < 1498) {
if (AFD < 2) {
if (Rs < 884) {
return 1;
}
else if (Rs >= 884) {
return 4;
}
}
else if (AFD >= 2) {
if (tcor < 76) {
return 1;
}
else if (tcor >= 76) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
}
else if (Cs >= 1498) {
if (negBalance < 45) {
if (Rs < 989) {
if (ssDCval < 114) {
return 4;
}
else if (ssDCval >= 114) {
return 8;
}
}
else if (Rs >= 989) {
if (RsDiff < 3) {
if (ssDCval < 118) {
return 8;
}
else if (ssDCval >= 118) {
if (Rs < 1043) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 4;
}
}
else if (Rs >= 1043) {
if (ssDCval < 135) {
return 4;
}
else if (ssDCval >= 135) {
return 8;
}
}
}
}
else if (RsDiff >= 3) {
if (tcor < 85) {
if (Rs < 1059) {
return 8;
}
else if (Rs >= 1059) {
return 4;
}
}
else if (tcor >= 85) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
else if (Rs >= 1285) {
if (RsDiff < 7) {
if (posBalance < 35) {
if (negBalance < 46) {
if (mcTcor < 97) {
if (negBalance < 41) {
return 8;
}
else if (negBalance >= 41) {
return 4;
}
}
else if (mcTcor >= 97) {
if (mu_mv_mag_sq < 98) {
return 1;
}
else if (mu_mv_mag_sq >= 98) {
if (posBalance < 9) {
if (diffAFD < 1) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
else if (diffAFD >= 1) {
if (Rs < 2166) {
return 2;
}
else if (Rs >= 2166) {
return 4;
}
}
}
else if (posBalance >= 9) {
if (posBalance < 11) {
return 8;
}
else if (posBalance >= 11) {
if (negBalance < 38) {
if (Rs < 2804) {
return 8;
}
else if (Rs >= 2804) {
return 2;
}
}
else if (negBalance >= 38) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
else if (posBalance >= 35) {
if (posBalance < 57) {
if (diffTSC < -4) {
if (MVDiff < 5) {
if (Rs < 3180) {
return 2;
}
else if (Rs >= 3180) {
return 4;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (diffTSC >= -4) {
if (refDCval < 109) {
if (diffRsCsdiff < 22) {
if (ssDCval < 105) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (ssDCval >= 105) {
return 8;
}
}
else if (diffRsCsdiff >= 22) {
if (Cs < 2546) {
return 4;
}
else if (Cs >= 2546) {
return 8;
}
}
}
else if (refDCval >= 109) {
if (Cs < 1371) {
if (SC < 3821) {
return 8;
}
else if (SC >= 3821) {
if (mu_mv_mag_sq < 79) {
return 8;
}
else if (mu_mv_mag_sq >= 79) {
if (Rs < 2687) {
return 1;
}
else if (Rs >= 2687) {
return 2;
}
}
}
}
else if (Cs >= 1371) {
if (diffRsCsdiff < -11) {
if (refDCval < 130) {
if (refDCval < 126) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 2044) {
return 2;
}
else if (Rs >= 2044) {
return 8;
}
}
}
else if (refDCval >= 126) {
if (SC < 3737) {
return 1;
}
else if (SC >= 3737) {
if (ssDCval < 126) {
return 1;
}
else if (ssDCval >= 126) {
return 8;
}
}
}
}
else if (refDCval >= 130) {
return 8;
}
}
else if (diffRsCsdiff >= -11) {
if (Cs < 2999) {
return 8;
}
else if (Cs >= 2999) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 57) {
if (refDCval < 123) {
return 8;
}
else if (refDCval >= 123) {
if (mu_mv_mag_sq < 102) {
if (mcTcor < 89) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 8;
}
}
else if (mcTcor >= 89) {
if (Rs < 1448) {
if (Rs < 1371) {
return 8;
}
else if (Rs >= 1371) {
return 1;
}
}
else if (Rs >= 1448) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (ssDCval < 144) {
return 1;
}
else if (ssDCval >= 144) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 7) {
if (negBalance < 49) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (ssDCval < 128) {
if (Rs < 3099) {
return 1;
}
else if (Rs >= 3099) {
return 8;
}
}
else if (ssDCval >= 128) {
if (Rs < 3974) {
if (posBalance < 46) {
if (mu_mv_mag_sq < 61) {
if (Rs < 1823) {
return 8;
}
else if (Rs >= 1823) {
return 1;
}
}
else if (mu_mv_mag_sq >= 61) {
if (Rs < 2596) {
return 8;
}
else if (Rs >= 2596) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (Rs >= 3974) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
}
}
}
else if (negBalance >= 49) {
if (Rs < 3641) {
if (ssDCval < 107) {
if (SC < 5011) {
return 8;
}
else if (SC >= 5011) {
if (Cs < 2578) {
return 4;
}
else if (Cs >= 2578) {
return 8;
}
}
}
else if (ssDCval >= 107) {
if (MVDiff < 1) {
if (Rs < 2095) {
return 8;
}
else if (Rs >= 2095) {
return 1;
}
}
else if (MVDiff >= 1) {
if (CsDiff < 5) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 1;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
else if (Rs >= 3641) {
if (Cs < 2286) {
return 1;
}
else if (Cs >= 2286) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 4405) {
if (Cs < 1774) {
if (tcor < 95) {
if (Cs < 1654) {
return 1;
}
else if (Cs >= 1654) {
return 8;
}
}
else if (tcor >= 95) {
return 2;
}
}
else if (Cs >= 1774) {
if (Cs < 3559) {
if (tcor < 96) {
if (SC < 6859) {
return 8;
}
else if (SC >= 6859) {
if (Cs < 3079) {
return 1;
}
else if (Cs >= 3079) {
return 8;
}
}
}
else if (tcor >= 96) {
if (Rs < 5408) {
return 2;
}
else if (Rs >= 5408) {
return 4;
}
}
}
else if (Cs >= 3559) {
return 8;
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2266) {
if (Rs < 1101) {
return 1;
}
else if (Rs >= 1101) {
return 2;
}
}
else if (Rs >= 2266) {
return 8;
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 2005) {
if (Rs < 3333) {
if (ssDCval < 87) {
if (negBalance < 47) {
if (mcTcor < 86) {
if (posBalance < 56) {
if (AFD < 9) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 1;
}
}
}
else if (AFD >= 9) {
if (MVDiff < 22) {
if (Cs < 1698) {
if (Rs < 1694) {
return 4;
}
else if (Rs >= 1694) {
return 1;
}
}
else if (Cs >= 1698) {
return 4;
}
}
else if (MVDiff >= 22) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
else if (posBalance >= 56) {
if (diffTSC < -2) {
if (Rs < 1482) {
return 4;
}
else if (Rs >= 1482) {
return 1;
}
}
else if (diffTSC >= -2) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
if (Rs < 527) {
return 8;
}
else if (Rs >= 527) {
if (Rs < 1341) {
return 4;
}
else if (Rs >= 1341) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 86) {
if (TSC < 6) {
if (diffMVdiffVal < -9) {
if (Rs < 637) {
return 2;
}
else if (Rs >= 637) {
return 8;
}
}
else if (diffMVdiffVal >= -9) {
return 1;
}
}
else if (TSC >= 6) {
if (RsDiff < 10) {
if (tcor < 78) {
if (mcTcor < 87) {
return 1;
}
else if (mcTcor >= 87) {
if (MVDiff < 19) {
if (Rs < 3101) {
return 4;
}
else if (Rs >= 3101) {
return 2;
}
}
else if (MVDiff >= 19) {
if (diffTSC < 2) {
return 2;
}
else if (diffTSC >= 2) {
return 8;
}
}
}
}
else if (tcor >= 78) {
if (CsDiff < 8) {
if (posBalance < 59) {
if (Cs < 1675) {
if (mu_mv_mag_sq < 49) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 1;
}
}
else if (mu_mv_mag_sq >= 49) {
if (Cs < 1552) {
return 2;
}
else if (Cs >= 1552) {
return 4;
}
}
}
else if (Cs >= 1675) {
return 1;
}
}
else if (posBalance >= 59) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
if (Cs < 1754) {
return 2;
}
else if (Cs >= 1754) {
return 4;
}
}
}
}
else if (CsDiff >= 8) {
if (Cs < 1523) {
if (Rs < 1886) {
return 1;
}
else if (Rs >= 1886) {
return 2;
}
}
else if (Cs >= 1523) {
if (Rs < 1495) {
return 4;
}
else if (Rs >= 1495) {
if (Rs < 2962) {
return 8;
}
else if (Rs >= 2962) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 10) {
if (posBalance < 63) {
if (diffMVdiffVal < -6) {
if (tcor < 82) {
if (Rs < 2124) {
return 1;
}
else if (Rs >= 2124) {
return 2;
}
}
else if (tcor >= 82) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
if (ssDCval < 48) {
return 1;
}
else if (ssDCval >= 48) {
if (Rs < 3202) {
return 2;
}
else if (Rs >= 3202) {
if (SC < 4607) {
return 4;
}
else if (SC >= 4607) {
return 2;
}
}
}
}
}
else if (posBalance >= 63) {
if (diffMVdiffVal < 8) {
if (Rs < 1964) {
return 1;
}
else if (Rs >= 1964) {
return 8;
}
}
else if (diffMVdiffVal >= 8) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 47) {
if (mu_mv_mag_sq < 23) {
if (tcor < 60) {
if (Rs < 1873) {
if (ssDCval < 69) {
if (MVDiff < 57) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (MVDiff >= 57) {
return 8;
}
}
else if (ssDCval >= 69) {
if (RsCsDiff < 163) {
if (Cs < 1544) {
return 2;
}
else if (Cs >= 1544) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 1;
}
}
}
else if (RsCsDiff >= 163) {
if (negBalance < 59) {
if (AFD < 30) {
if (Cs < 1718) {
return 8;
}
else if (Cs >= 1718) {
return 4;
}
}
else if (AFD >= 30) {
return 2;
}
}
else if (negBalance >= 59) {
return 4;
}
}
}
}
else if (Rs >= 1873) {
if (MVDiff < 17) {
if (Rs < 2175) {
return 8;
}
else if (Rs >= 2175) {
return 2;
}
}
else if (MVDiff >= 17) {
if (RsCsDiff < 363) {
if (SC < 3656) {
if (Rs < 2105) {
return 4;
}
else if (Rs >= 2105) {
return 8;
}
}
else if (SC >= 3656) {
return 1;
}
}
else if (RsCsDiff >= 363) {
if (TSC < 14) {
if (Rs < 2476) {
return 4;
}
else if (Rs >= 2476) {
return 2;
}
}
else if (TSC >= 14) {
return 8;
}
}
}
}
}
else if (tcor >= 60) {
if (Rs < 2582) {
if (SC < 3494) {
if (tcor < 79) {
if (negBalance < 53) {
if (RsCsDiff < 167) {
if (Cs < 1382) {
return 8;
}
else if (Cs >= 1382) {
return 4;
}
}
else if (RsCsDiff >= 167) {
if (Cs < 1381) {
return 4;
}
else if (Cs >= 1381) {
return 8;
}
}
}
else if (negBalance >= 53) {
if (Cs < 1684) {
if (diffMVdiffVal < 7) {
if (AFD < 22) {
if (mcTcor < 94) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (RsDiff < 9) {
if (Rs < 1240) {
return 8;
}
else if (Rs >= 1240) {
return 1;
}
}
else if (RsDiff >= 9) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (Cs < 1398) {
if (Rs < 1467) {
return 8;
}
else if (Rs >= 1467) {
return 1;
}
}
else if (Cs >= 1398) {
return 2;
}
}
}
else if (AFD >= 22) {
if (Rs < 1582) {
if (Rs < 1329) {
return 8;
}
else if (Rs >= 1329) {
return 1;
}
}
else if (Rs >= 1582) {
return 4;
}
}
}
else if (diffMVdiffVal >= 7) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
if (Cs < 1455) {
return 8;
}
else if (Cs >= 1455) {
return 4;
}
}
}
}
else if (Cs >= 1684) {
if (ssDCval < 77) {
if (Rs < 1479) {
if (Rs < 972) {
return 4;
}
else if (Rs >= 972) {
return 2;
}
}
else if (Rs >= 1479) {
return 4;
}
}
else if (ssDCval >= 77) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 4;
}
}
}
}
}
else if (tcor >= 79) {
if (mcTcor < 92) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
if (Cs < 1637) {
return 8;
}
else if (Cs >= 1637) {
return 4;
}
}
}
else if (mcTcor >= 92) {
if (RsCsDiff < 123) {
return 2;
}
else if (RsCsDiff >= 123) {
if (Cs < 1441) {
return 2;
}
else if (Cs >= 1441) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 2;
}
}
}
}
}
}
else if (SC >= 3494) {
if (mcTcor < 96) {
if (RsCsDiff < 129) {
if (diffAFD < -1) {
if (Rs < 1808) {
return 1;
}
else if (Rs >= 1808) {
if (Rs < 2001) {
return 8;
}
else if (Rs >= 2001) {
return 2;
}
}
}
else if (diffAFD >= -1) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 4;
}
}
}
else if (RsCsDiff >= 129) {
if (Cs < 1855) {
if (mcTcor < 92) {
if (Cs < 1711) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (negBalance < 58) {
return 4;
}
else if (negBalance >= 58) {
if (Rs < 1880) {
return 4;
}
else if (Rs >= 1880) {
return 8;
}
}
}
}
else if (Cs >= 1711) {
if (diffAFD < 2) {
if (RsDiff < 10) {
return 8;
}
else if (RsDiff >= 10) {
if (diffRsCsdiff < -85) {
if (Rs < 2168) {
return 2;
}
else if (Rs >= 2168) {
return 8;
}
}
else if (diffRsCsdiff >= -85) {
return 4;
}
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 85) {
return 4;
}
else if (ssDCval >= 85) {
return 8;
}
}
}
else if (Cs >= 1855) {
if (gchDC < 2) {
if (AFD < 14) {
if (Rs < 2358) {
return 8;
}
else if (Rs >= 2358) {
return 4;
}
}
else if (AFD >= 14) {
return 8;
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
return 2;
}
}
}
else if (Rs >= 2582) {
if (ssDCval < 80) {
if (RsCsDiff < 198) {
if (Cs < 1764) {
if (Rs < 2897) {
return 8;
}
else if (Rs >= 2897) {
return 4;
}
}
else if (Cs >= 1764) {
return 4;
}
}
else if (RsCsDiff >= 198) {
if (AFD < 26) {
if (Cs < 1987) {
if (refDCval < 63) {
return 8;
}
else if (refDCval >= 63) {
return 2;
}
}
else if (Cs >= 1987) {
return 4;
}
}
else if (AFD >= 26) {
return 8;
}
}
}
else if (ssDCval >= 80) {
if (CsDiff < 9) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (CsDiff >= 9) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2788) {
return 4;
}
else if (Rs >= 2788) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 23) {
if (Rs < 2009) {
if (tcor < 69) {
if (diffMVdiffVal < -15) {
if (Rs < 1660) {
return 1;
}
else if (Rs >= 1660) {
return 2;
}
}
else if (diffMVdiffVal >= -15) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
if (Rs < 1190) {
return 8;
}
else if (Rs >= 1190) {
return 4;
}
}
}
}
}
}
else if (tcor >= 69) {
if (MVDiff < 17) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 1991) {
return 4;
}
else if (Rs >= 1991) {
return 8;
}
}
}
else if (MVDiff >= 17) {
if (ssDCval < 76) {
if (mu_mv_mag_sq < 50) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
if (AFD < 25) {
return 8;
}
else if (AFD >= 25) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 1114) {
return 8;
}
else if (Rs >= 1114) {
return 1;
}
}
}
else if (ssDCval >= 76) {
if (Rs < 1129) {
return 4;
}
else if (Rs >= 1129) {
return 1;
}
}
}
}
}
else if (Rs >= 2009) {
if (mcTcor < 94) {
if (TSC < 12) {
if (diffAFD < 2) {
if (Cs < 1907) {
return 8;
}
else if (Cs >= 1907) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 3244) {
return 4;
}
else if (Rs >= 3244) {
return 8;
}
}
}
}
else if (diffAFD >= 2) {
if (Cs < 1598) {
return 8;
}
else if (Cs >= 1598) {
if (tcor < 84) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (tcor >= 84) {
return 8;
}
}
}
}
else if (TSC >= 12) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 57) {
return 8;
}
else if (mu_mv_mag_sq >= 57) {
if (Rs < 2236) {
return 8;
}
else if (Rs >= 2236) {
return 1;
}
}
}
else if (ssDCval >= 84) {
if (Rs < 2403) {
return 8;
}
else if (Rs >= 2403) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 3088) {
if (Cs < 1688) {
return 4;
}
else if (Cs >= 1688) {
return 8;
}
}
else if (Rs >= 3088) {
if (Rs < 3326) {
return 2;
}
else if (Rs >= 3326) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 87) {
if (AFD < 19) {
if (TSC < 7) {
if (ssDCval < 103) {
if (mcTcor < 96) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
return 2;
}
}
else if (mcTcor >= 96) {
if (Rs < 2853) {
return 1;
}
else if (Rs >= 2853) {
return 2;
}
}
}
else if (ssDCval >= 103) {
if (SC < 3396) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (SC < 3228) {
return 2;
}
else if (SC >= 3228) {
if (Rs < 1449) {
return 1;
}
else if (Rs >= 1449) {
return 8;
}
}
}
}
}
else if (SC >= 3396) {
if (posBalance < 44) {
if (Cs < 1584) {
return 4;
}
else if (Cs >= 1584) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
return 8;
}
}
}
else if (posBalance >= 44) {
return 8;
}
}
}
}
else if (TSC >= 7) {
if (mu_mv_mag_sq < 7) {
if (refDCval < 100) {
if (MVDiff < 19) {
if (negBalance < 50) {
if (RsCsDiff < 115) {
if (Rs < 2738) {
return 8;
}
else if (Rs >= 2738) {
return 4;
}
}
else if (RsCsDiff >= 115) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (tcor < 71) {
return 4;
}
else if (tcor >= 71) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
if (Cs < 1442) {
return 4;
}
else if (Cs >= 1442) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 50) {
if (RsDiff < 12) {
if (tcor < 89) {
if (posBalance < 55) {
if (Rs < 1826) {
return 8;
}
else if (Rs >= 1826) {
return 2;
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (tcor >= 89) {
if (Rs < 2160) {
return 2;
}
else if (Rs >= 2160) {
return 8;
}
}
}
else if (RsDiff >= 12) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
if (Rs < 3109) {
return 2;
}
else if (Rs >= 3109) {
return 8;
}
}
}
}
}
else if (MVDiff >= 19) {
if (Rs < 1816) {
return 1;
}
else if (Rs >= 1816) {
return 8;
}
}
}
else if (refDCval >= 100) {
if (Rs < 3088) {
if (negBalance < 43) {
if (Rs < 1637) {
return 4;
}
else if (Rs >= 1637) {
return 8;
}
}
else if (negBalance >= 43) {
if (Cs < 1502) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
if (Rs < 2672) {
return 1;
}
else if (Rs >= 2672) {
return 8;
}
}
}
else if (Cs >= 1502) {
if (CsDiff < 7) {
if (Rs < 2876) {
return 8;
}
else if (Rs >= 2876) {
return 4;
}
}
else if (CsDiff >= 7) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Rs < 1938) {
return 4;
}
else if (Rs >= 1938) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3088) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
if (diffAFD < -3) {
if (Cs < 1483) {
if (Rs < 3221) {
return 4;
}
else if (Rs >= 3221) {
return 1;
}
}
else if (Cs >= 1483) {
return 2;
}
}
else if (diffAFD >= -3) {
if (Rs < 3310) {
return 4;
}
else if (Rs >= 3310) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 1811) {
if (SC < 4242) {
if (TSC < 13) {
if (Cs < 1570) {
if (mu_mv_mag_sq < 59) {
if (Rs < 1401) {
if (Rs < 1323) {
return 2;
}
else if (Rs >= 1323) {
return 4;
}
}
else if (Rs >= 1401) {
if (RsCsDiff < 201) {
return 8;
}
else if (RsCsDiff >= 201) {
if (mu_mv_mag_sq < 13) {
if (Rs < 2426) {
return 2;
}
else if (Rs >= 2426) {
return 1;
}
}
else if (mu_mv_mag_sq >= 13) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Rs < 2094) {
return 1;
}
else if (Rs >= 2094) {
if (Rs < 2263) {
return 4;
}
else if (Rs >= 2263) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (mcTcor < 95) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 4;
}
}
else if (mcTcor >= 95) {
if (Rs < 2224) {
return 1;
}
else if (Rs >= 2224) {
return 4;
}
}
}
}
else if (Cs >= 1570) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
return 2;
}
}
}
else if (TSC >= 13) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
return 1;
}
}
}
else if (SC >= 4242) {
if (refDCval < 97) {
if (RsCsDiff < 132) {
if (SC < 4350) {
return 8;
}
else if (SC >= 4350) {
return 4;
}
}
else if (RsCsDiff >= 132) {
if (refDCval < 87) {
return 8;
}
else if (refDCval >= 87) {
return 2;
}
}
}
else if (refDCval >= 97) {
if (Rs < 2613) {
if (Rs < 2559) {
return 2;
}
else if (Rs >= 2559) {
return 4;
}
}
else if (Rs >= 2613) {
if (mu_mv_mag_sq < 66) {
if (diffMVdiffVal < 24) {
return 8;
}
else if (diffMVdiffVal >= 24) {
if (Rs < 2836) {
return 2;
}
else if (Rs >= 2836) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (Cs < 1599) {
return 8;
}
else if (Cs >= 1599) {
if (Rs < 3173) {
return 4;
}
else if (Rs >= 3173) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 1811) {
return 8;
}
}
}
}
else if (AFD >= 19) {
if (TSC < 10) {
if (Rs < 1886) {
if (AFD < 25) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1730) {
return 4;
}
else if (Rs >= 1730) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1753) {
if (Cs < 1412) {
return 8;
}
else if (Cs >= 1412) {
return 2;
}
}
else if (Rs >= 1753) {
if (Rs < 1843) {
return 8;
}
else if (Rs >= 1843) {
return 4;
}
}
}
}
else if (AFD >= 25) {
return 4;
}
}
else if (Rs >= 1886) {
if (mu_mv_mag_sq < 8) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (mu_mv_mag_sq >= 8) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 2;
}
}
}
}
else if (TSC >= 10) {
if (refDCval < 121) {
if (mcTcor < 94) {
if (refDCval < 98) {
if (Cs < 1780) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
if (CsDiff < 6) {
if (Rs < 3076) {
return 2;
}
else if (Rs >= 3076) {
return 4;
}
}
else if (CsDiff >= 6) {
if (TSC < 13) {
if (Rs < 1368) {
return 2;
}
else if (Rs >= 1368) {
return 8;
}
}
else if (TSC >= 13) {
if (diffMVdiffVal < 7) {
if (CsDiff < 8) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
return 8;
}
}
else if (CsDiff >= 8) {
if (Rs < 1756) {
return 4;
}
else if (Rs >= 1756) {
return 1;
}
}
}
else if (diffMVdiffVal >= 7) {
if (MVDiff < 45) {
if (Rs < 2287) {
return 2;
}
else if (Rs >= 2287) {
return 8;
}
}
else if (MVDiff >= 45) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1780) {
if (RsCsDiff < 217) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
else if (RsCsDiff >= 217) {
if (RsCsDiff < 343) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 4;
}
}
else if (RsCsDiff >= 343) {
return 8;
}
}
}
}
else if (refDCval >= 98) {
if (tcor < 55) {
if (CsDiff < 12) {
if (AFD < 22) {
return 4;
}
else if (AFD >= 22) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 2;
}
}
}
else if (CsDiff >= 12) {
return 8;
}
}
else if (tcor >= 55) {
if (RsDiff < 16) {
if (refDCval < 111) {
if (Cs < 1714) {
if (CsDiff < 9) {
if (Rs < 3108) {
return 8;
}
else if (Rs >= 3108) {
return 4;
}
}
else if (CsDiff >= 9) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
if (Rs < 2268) {
return 8;
}
else if (Rs >= 2268) {
return 1;
}
}
}
}
else if (Cs >= 1714) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (refDCval >= 111) {
if (Rs < 1676) {
return 4;
}
else if (Rs >= 1676) {
return 8;
}
}
}
else if (RsDiff >= 16) {
if (refDCval < 106) {
if (AFD < 28) {
return 2;
}
else if (AFD >= 28) {
return 8;
}
}
else if (refDCval >= 106) {
return 8;
}
}
}
}
}
else if (mcTcor >= 94) {
if (tcor < 86) {
if (Rs < 2543) {
if (negBalance < 50) {
if (Rs < 1796) {
return 1;
}
else if (Rs >= 1796) {
if (Rs < 2456) {
return 8;
}
else if (Rs >= 2456) {
return 2;
}
}
}
else if (negBalance >= 50) {
return 2;
}
}
else if (Rs >= 2543) {
if (Cs < 1465) {
return 1;
}
else if (Cs >= 1465) {
return 8;
}
}
}
else if (tcor >= 86) {
return 8;
}
}
}
else if (refDCval >= 121) {
if (MVDiff < 40) {
if (Rs < 1492) {
if (Cs < 1925) {
return 2;
}
else if (Cs >= 1925) {
return 1;
}
}
else if (Rs >= 1492) {
if (mcTcor < 85) {
return 1;
}
else if (mcTcor >= 85) {
if (mu_mv_mag_sq < 14) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (ssDCval < 153) {
return 1;
}
else if (ssDCval >= 153) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (tcor < 68) {
if (Rs < 2249) {
return 8;
}
else if (Rs >= 2249) {
return 2;
}
}
else if (tcor >= 68) {
return 8;
}
}
}
}
}
else if (MVDiff >= 40) {
if (ssDCval < 155) {
if (refDCval < 126) {
if (MVDiff < 58) {
return 8;
}
else if (MVDiff >= 58) {
return 1;
}
}
else if (refDCval >= 126) {
if (CsDiff < 6) {
if (Rs < 3037) {
return 4;
}
else if (Rs >= 3037) {
return 8;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (ssDCval >= 155) {
if (Rs < 2981) {
return 2;
}
else if (Rs >= 2981) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 3333) {
if (tcor < 88) {
if (RsDiff < 12) {
if (posBalance < 49) {
if (refDCval < 87) {
if (ssDCval < 66) {
if (Cs < 1804) {
return 2;
}
else if (Cs >= 1804) {
if (Rs < 4621) {
return 8;
}
else if (Rs >= 4621) {
return 4;
}
}
}
else if (ssDCval >= 66) {
return 4;
}
}
else if (refDCval >= 87) {
if (Cs < 1530) {
return 4;
}
else if (Cs >= 1530) {
if (Rs < 4325) {
return 8;
}
else if (Rs >= 4325) {
return 1;
}
}
}
}
else if (posBalance >= 49) {
if (ssDCval < 90) {
if (diffRsCsdiff < 4) {
if (Cs < 1850) {
if (Rs < 4290) {
return 2;
}
else if (Rs >= 4290) {
return 8;
}
}
else if (Cs >= 1850) {
return 4;
}
}
else if (diffRsCsdiff >= 4) {
if (Cs < 1546) {
return 1;
}
else if (Cs >= 1546) {
return 8;
}
}
}
else if (ssDCval >= 90) {
return 8;
}
}
}
else if (RsDiff >= 12) {
if (MVDiff < 12) {
if (mu_mv_mag_sq < 70) {
if (RsCsDiff < 231) {
if (ssDCval < 92) {
if (Rs < 4946) {
return 2;
}
else if (Rs >= 4946) {
return 4;
}
}
else if (ssDCval >= 92) {
if (Cs < 1343) {
return 4;
}
else if (Cs >= 1343) {
return 8;
}
}
}
else if (RsCsDiff >= 231) {
if (tcor < 83) {
if (ssDCval < 91) {
if (ssDCval < 66) {
return 8;
}
else if (ssDCval >= 66) {
return 2;
}
}
else if (ssDCval >= 91) {
if (RsDiff < 17) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
else if (RsDiff >= 17) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 2;
}
}
}
}
else if (tcor >= 83) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
if (Rs < 3716) {
if (Rs < 3582) {
return 8;
}
else if (Rs >= 3582) {
return 2;
}
}
else if (Rs >= 3716) {
return 1;
}
}
}
}
else if (MVDiff >= 12) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (refDCval < 133) {
if (TSC < 13) {
if (mu_mv_mag_sq < 69) {
if (refDCval < 95) {
if (Cs < 1722) {
if (Rs < 3453) {
return 1;
}
else if (Rs >= 3453) {
return 2;
}
}
else if (Cs >= 1722) {
if (diffMVdiffVal < 1) {
if (CsDiff < 9) {
return 4;
}
else if (CsDiff >= 9) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
if (Rs < 3367) {
return 2;
}
else if (Rs >= 3367) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 4;
}
}
}
}
}
else if (refDCval >= 95) {
if (RsDiff < 17) {
if (AFD < 21) {
if (diffAFD < -6) {
return 2;
}
else if (diffAFD >= -6) {
return 8;
}
}
else if (AFD >= 21) {
if (Rs < 4633) {
return 8;
}
else if (Rs >= 4633) {
return 1;
}
}
}
else if (RsDiff >= 17) {
if (SC < 6750) {
if (AFD < 29) {
if (MVDiff < 53) {
if (SC < 5236) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 8;
}
}
else if (SC >= 5236) {
return 2;
}
}
else if (MVDiff >= 53) {
return 8;
}
}
else if (AFD >= 29) {
return 4;
}
}
else if (SC >= 6750) {
if (Rs < 6608) {
return 8;
}
else if (Rs >= 6608) {
if (Rs < 6782) {
return 2;
}
else if (Rs >= 6782) {
if (Cs < 1580) {
return 8;
}
else if (Cs >= 1580) {
if (Cs < 1751) {
return 2;
}
else if (Cs >= 1751) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (Cs < 1846) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 1;
}
}
else if (Cs >= 1846) {
return 8;
}
}
}
else if (TSC >= 13) {
if (tcor < 52) {
if (Rs < 3507) {
if (Rs < 3453) {
return 8;
}
else if (Rs >= 3453) {
return 1;
}
}
else if (Rs >= 3507) {
return 2;
}
}
else if (tcor >= 52) {
if (mu_mv_mag_sq < 35) {
if (mcTcor < 93) {
if (diffMVdiffVal < -3) {
if (negBalance < 52) {
return 1;
}
else if (negBalance >= 52) {
if (Rs < 4237) {
return 4;
}
else if (Rs >= 4237) {
return 2;
}
}
}
else if (diffMVdiffVal >= -3) {
if (ssDCval < 118) {
if (Cs < 1963) {
return 2;
}
else if (Cs >= 1963) {
return 8;
}
}
else if (ssDCval >= 118) {
if (TSC < 26) {
if (TSC < 15) {
return 8;
}
else if (TSC >= 15) {
return 1;
}
}
else if (TSC >= 26) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
if (MVDiff < 183) {
return 1;
}
else if (MVDiff >= 183) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (Rs < 3820) {
return 1;
}
else if (Rs >= 3820) {
return 4;
}
}
}
}
}
else if (refDCval >= 133) {
if (CsDiff < 9) {
if (RsCsDiff < 619) {
if (Rs < 4397) {
if (Rs < 3557) {
if (Rs < 3407) {
return 4;
}
else if (Rs >= 3407) {
return 2;
}
}
else if (Rs >= 3557) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 8;
}
}
}
else if (Rs >= 4397) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (ssDCval < 137) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
if (Rs < 6489) {
return 8;
}
else if (Rs >= 6489) {
if (Rs < 6496) {
return 2;
}
else if (Rs >= 6496) {
return 8;
}
}
}
}
else if (ssDCval >= 137) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
if (Rs < 4681) {
return 4;
}
else if (Rs >= 4681) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 619) {
if (ssDCval < 140) {
return 2;
}
else if (ssDCval >= 140) {
return 4;
}
}
}
else if (CsDiff >= 9) {
if (gchDC < 2) {
if (negBalance < 54) {
return 4;
}
else if (negBalance >= 54) {
if (Rs < 3763) {
return 1;
}
else if (Rs >= 3763) {
if (Rs < 5313) {
return 2;
}
else if (Rs >= 5313) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (TSC < 12) {
if (RsCsDiff < 667) {
return 8;
}
else if (RsCsDiff >= 667) {
if (Rs < 5586) {
return 2;
}
else if (Rs >= 5586) {
return 4;
}
}
}
else if (TSC >= 12) {
return 1;
}
}
}
}
}
}
}
}
else if (tcor >= 88) {
if (RsDiff < 9) {
if (diffTSC < 2) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (Rs < 3623) {
return 2;
}
else if (Rs >= 3623) {
if (tcor < 89) {
if (Rs < 4337) {
return 1;
}
else if (Rs >= 4337) {
return 8;
}
}
else if (tcor >= 89) {
return 1;
}
}
}
}
else if (diffTSC >= 2) {
return 8;
}
}
else if (RsDiff >= 9) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (ssDCval < 101) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (posBalance < 40) {
if (negBalance < 39) {
return 4;
}
else if (negBalance >= 39) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
if (Cs < 1548) {
return 1;
}
else if (Cs >= 1548) {
return 8;
}
}
}
}
else if (posBalance >= 40) {
if (Cs < 1801) {
if (diffRsCsdiff < 83) {
return 2;
}
else if (diffRsCsdiff >= 83) {
if (diffRsCsdiff < 87) {
if (Rs < 5344) {
return 2;
}
else if (Rs >= 5344) {
return 4;
}
}
else if (diffRsCsdiff >= 87) {
return 2;
}
}
}
else if (Cs >= 1801) {
if (Rs < 4135) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (Rs < 3527) {
return 8;
}
else if (Rs >= 3527) {
return 2;
}
}
}
else if (Rs >= 4135) {
return 2;
}
}
}
}
}
else if (ssDCval >= 101) {
if (RsDiff < 12) {
if (MVDiff < 6) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (Rs < 3455) {
return 4;
}
else if (Rs >= 3455) {
return 1;
}
}
}
else if (MVDiff >= 6) {
if (diffAFD < -6) {
return 2;
}
else if (diffAFD >= -6) {
return 8;
}
}
}
else if (RsDiff >= 12) {
if (MVDiff < 15) {
if (Rs < 3849) {
if (Rs < 3825) {
return 2;
}
else if (Rs >= 3825) {
return 8;
}
}
else if (Rs >= 3849) {
return 2;
}
}
else if (MVDiff >= 15) {
if (ssDCval < 110) {
return 2;
}
else if (ssDCval >= 110) {
if (mu_mv_mag_sq < 33) {
if (Rs < 3703) {
return 8;
}
else if (Rs >= 3703) {
return 1;
}
}
else if (mu_mv_mag_sq >= 33) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 2005) {
if (AFD < 12) {
if (ssDCval < 100) {
if (Rs < 4189) {
if (Cs < 2636) {
if (ssDCval < 84) {
if (negBalance < 39) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (negBalance >= 39) {
if (mcTcor < 92) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (Cs < 2067) {
return 4;
}
else if (Cs >= 2067) {
if (Rs < 3827) {
return 8;
}
else if (Rs >= 3827) {
return 4;
}
}
}
}
else if (mcTcor >= 92) {
if (Rs < 2726) {
return 8;
}
else if (Rs >= 2726) {
if (Rs < 3623) {
return 4;
}
else if (Rs >= 3623) {
if (Rs < 3834) {
return 8;
}
else if (Rs >= 3834) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (RsDiff < 11) {
if (posBalance < 60) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (Rs < 2486) {
return 8;
}
else if (Rs >= 2486) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
else if (posBalance >= 60) {
if (Rs < 3512) {
return 4;
}
else if (Rs >= 3512) {
return 8;
}
}
}
else if (RsDiff >= 11) {
return 2;
}
}
}
else if (Cs >= 2636) {
if (ssDCval < 55) {
if (Rs < 2295) {
return 1;
}
else if (Rs >= 2295) {
return 8;
}
}
else if (ssDCval >= 55) {
if (negBalance < 53) {
if (SC < 5773) {
if (TSC < 10) {
if (Rs < 1468) {
return 8;
}
else if (Rs >= 1468) {
if (mu_mv_mag_sq < 100) {
return 4;
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 2270) {
return 4;
}
else if (Rs >= 2270) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (Rs < 2423) {
return 4;
}
else if (Rs >= 2423) {
return 8;
}
}
}
else if (SC >= 5773) {
if (Cs < 2679) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
else if (Cs >= 2679) {
if (negBalance < 50) {
if (Cs < 4339) {
if (MVDiff < 30) {
if (gchDC < 2) {
if (RsDiff < 7) {
if (Rs < 3458) {
return 4;
}
else if (Rs >= 3458) {
return 1;
}
}
else if (RsDiff >= 7) {
return 4;
}
}
else if (gchDC >= 2) {
if (Rs < 3679) {
return 8;
}
else if (Rs >= 3679) {
return 4;
}
}
}
else if (MVDiff >= 30) {
return 2;
}
}
else if (Cs >= 4339) {
return 8;
}
}
else if (negBalance >= 50) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
return 4;
}
}
}
}
}
else if (negBalance >= 53) {
if (Rs < 3253) {
return 4;
}
else if (Rs >= 3253) {
return 8;
}
}
}
}
}
else if (Rs >= 4189) {
if (tcor < 90) {
if (AFD < 8) {
if (Rs < 5671) {
return 8;
}
else if (Rs >= 5671) {
return 2;
}
}
else if (AFD >= 8) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (mu_mv_mag_sq < 82) {
return 8;
}
else if (mu_mv_mag_sq >= 82) {
if (RsDiff < 14) {
if (Rs < 4486) {
return 1;
}
else if (Rs >= 4486) {
return 4;
}
}
else if (RsDiff >= 14) {
return 8;
}
}
}
}
}
}
else if (tcor >= 90) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 4409) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Cs < 2708) {
return 2;
}
else if (Cs >= 2708) {
if (Rs < 4212) {
return 8;
}
else if (Rs >= 4212) {
return 4;
}
}
}
}
else if (Rs >= 4409) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
if (Rs < 4584) {
return 2;
}
else if (Rs >= 4584) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 100) {
if (negBalance < 46) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 93) {
if (ssDCval < 109) {
return 2;
}
else if (ssDCval >= 109) {
if (negBalance < 27) {
if (ssDCval < 138) {
if (Cs < 2952) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (Cs >= 2952) {
if (Rs < 3093) {
return 4;
}
else if (Rs >= 3093) {
return 2;
}
}
}
else if (ssDCval >= 138) {
if (Rs < 1907) {
return 1;
}
else if (Rs >= 1907) {
return 8;
}
}
}
else if (negBalance >= 27) {
if (Cs < 2704) {
if (RsDiff < 10) {
if (tcor < 86) {
if (Rs < 3379) {
return 8;
}
else if (Rs >= 3379) {
return 1;
}
}
else if (tcor >= 86) {
return 8;
}
}
else if (RsDiff >= 10) {
if (Rs < 2206) {
return 8;
}
else if (Rs >= 2206) {
return 2;
}
}
}
else if (Cs >= 2704) {
if (MVDiff < 30) {
return 1;
}
else if (MVDiff >= 30) {
if (Rs < 2576) {
return 8;
}
else if (Rs >= 2576) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (tcor < 94) {
if (Rs < 1841) {
return 2;
}
else if (Rs >= 1841) {
return 8;
}
}
else if (tcor >= 94) {
return 2;
}
}
}
else if (negBalance >= 37) {
if (Rs < 4486) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (MVDiff < 5) {
if (diffAFD < 2) {
if (ssDCval < 104) {
if (Rs < 3443) {
if (Cs < 2633) {
return 8;
}
else if (Cs >= 2633) {
return 4;
}
}
else if (Rs >= 3443) {
if (SC < 6633) {
return 2;
}
else if (SC >= 6633) {
return 1;
}
}
}
else if (ssDCval >= 104) {
return 8;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (Rs >= 4486) {
if (Cs < 3688) {
if (Rs < 6129) {
if (mu_mv_mag_sq < 90) {
return 8;
}
else if (mu_mv_mag_sq >= 90) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
return 1;
}
}
}
else if (Rs >= 6129) {
if (Rs < 6970) {
return 4;
}
else if (Rs >= 6970) {
if (Rs < 8305) {
return 2;
}
else if (Rs >= 8305) {
return 1;
}
}
}
}
else if (Cs >= 3688) {
if (mu_mv_mag_sq < 66) {
if (Rs < 5692) {
return 8;
}
else if (Rs >= 5692) {
if (Rs < 8171) {
return 4;
}
else if (Rs >= 8171) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 66) {
return 8;
}
}
}
}
}
else if (negBalance >= 46) {
if (tcor < 93) {
if (mu_mv_mag_sq < 83) {
if (tcor < 88) {
return 8;
}
else if (tcor >= 88) {
if (diffTSC < 2) {
if (SC < 6406) {
return 1;
}
else if (SC >= 6406) {
if (Rs < 4686) {
return 8;
}
else if (Rs >= 4686) {
return 1;
}
}
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 83) {
if (Cs < 3533) {
if (Rs < 3666) {
if (Cs < 2324) {
return 8;
}
else if (Cs >= 2324) {
return 1;
}
}
else if (Rs >= 3666) {
return 1;
}
}
else if (Cs >= 3533) {
return 8;
}
}
}
else if (tcor >= 93) {
return 8;
}
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 75) {
if (RsDiff < 16) {
if (refDCval < 101) {
if (tcor < 64) {
if (refDCval < 63) {
if (RsCsDiff < 294) {
if (MVDiff < 19) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
if (diffMVdiffVal < -2) {
return 2;
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
}
else if (MVDiff >= 19) {
if (diffRsCsdiff < -36) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
return 4;
}
}
else if (diffRsCsdiff >= -36) {
return 2;
}
}
}
else if (RsCsDiff >= 294) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
return 4;
}
}
}
else if (refDCval >= 63) {
if (Rs < 950) {
if (Cs < 4124) {
return 2;
}
else if (Cs >= 4124) {
return 8;
}
}
else if (Rs >= 950) {
if (RsDiff < 9) {
if (negBalance < 54) {
if (Rs < 1517) {
if (Rs < 1253) {
return 2;
}
else if (Rs >= 1253) {
return 4;
}
}
else if (Rs >= 1517) {
return 8;
}
}
else if (negBalance >= 54) {
if (Cs < 4282) {
return 4;
}
else if (Cs >= 4282) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (mcTcor < 74) {
if (negBalance < 58) {
if (Cs < 3108) {
if (Rs < 2320) {
return 4;
}
else if (Rs >= 2320) {
return 8;
}
}
else if (Cs >= 3108) {
return 4;
}
}
else if (negBalance >= 58) {
return 4;
}
}
else if (mcTcor >= 74) {
if (ssDCval < 85) {
if (RsDiff < 11) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (diffAFD < 1) {
if (Rs < 2276) {
return 4;
}
else if (Rs >= 2276) {
return 8;
}
}
else if (diffAFD >= 1) {
if (Rs < 1625) {
return 2;
}
else if (Rs >= 1625) {
return 8;
}
}
}
}
else if (RsDiff >= 11) {
if (CsDiff < 12) {
return 1;
}
else if (CsDiff >= 12) {
if (gchDC < 3) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (Cs < 2493) {
if (Rs < 2827) {
return 1;
}
else if (Rs >= 2827) {
return 8;
}
}
else if (Cs >= 2493) {
return 8;
}
}
}
else if (gchDC >= 3) {
return 1;
}
}
}
}
else if (ssDCval >= 85) {
if (negBalance < 67) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
return 8;
}
}
else if (negBalance >= 67) {
return 2;
}
}
}
}
}
}
}
else if (tcor >= 64) {
if (MVDiff < 6) {
if (ssDCval < 86) {
if (mu_mv_mag_sq < 3) {
if (Cs < 2521) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
else if (Cs >= 2521) {
if (mcTcor < 94) {
if (Rs < 2464) {
return 4;
}
else if (Rs >= 2464) {
if (Cs < 4033) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 2;
}
}
else if (Cs >= 4033) {
return 8;
}
}
}
else if (mcTcor >= 94) {
if (Rs < 3461) {
return 1;
}
else if (Rs >= 3461) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 12) {
if (mu_mv_mag_sq < 19) {
return 1;
}
else if (mu_mv_mag_sq >= 19) {
if (Cs < 3049) {
if (Cs < 2794) {
if (Rs < 2970) {
return 8;
}
else if (Rs >= 2970) {
if (Rs < 4038) {
return 4;
}
else if (Rs >= 4038) {
return 8;
}
}
}
else if (Cs >= 2794) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (Rs < 3872) {
return 1;
}
else if (Rs >= 3872) {
return 8;
}
}
}
}
else if (Cs >= 3049) {
return 1;
}
}
}
else if (TSC >= 12) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 8;
}
}
}
}
else if (ssDCval >= 86) {
if (SC < 5593) {
if (diffMVdiffVal < 2) {
if (Cs < 2864) {
if (RsCsDiff < 255) {
return 8;
}
else if (RsCsDiff >= 255) {
if (Rs < 2358) {
return 8;
}
else if (Rs >= 2358) {
return 4;
}
}
}
else if (Cs >= 2864) {
if (RsCsDiff < 140) {
return 4;
}
else if (RsCsDiff >= 140) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 3257) {
if (Rs < 2668) {
return 1;
}
else if (Rs >= 2668) {
return 8;
}
}
else if (Rs >= 3257) {
return 2;
}
}
}
else if (SC >= 5593) {
if (AFD < 16) {
if (Cs < 2678) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 4;
}
}
else if (Cs >= 2678) {
return 4;
}
}
else if (AFD >= 16) {
if (refDCval < 91) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (refDCval >= 91) {
if (Rs < 4564) {
if (TSC < 14) {
return 4;
}
else if (TSC >= 14) {
if (refDCval < 92) {
return 8;
}
else if (refDCval >= 92) {
return 4;
}
}
}
else if (Rs >= 4564) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (posBalance < 49) {
if (Rs < 4248) {
if (Rs < 2707) {
if (ssDCval < 58) {
if (Rs < 1146) {
return 2;
}
else if (Rs >= 1146) {
return 1;
}
}
else if (ssDCval >= 58) {
if (Cs < 2024) {
return 1;
}
else if (Cs >= 2024) {
return 8;
}
}
}
else if (Rs >= 2707) {
if (negBalance < 56) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (mu_mv_mag_sq < 72) {
return 2;
}
else if (mu_mv_mag_sq >= 72) {
return 4;
}
}
}
else if (negBalance >= 56) {
return 8;
}
}
}
else if (Rs >= 4248) {
if (ssDCval < 74) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (ssDCval >= 74) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (Rs < 4936) {
return 2;
}
else if (Rs >= 4936) {
return 1;
}
}
}
}
}
else if (posBalance >= 49) {
if (AFD < 18) {
if (Rs < 1737) {
if (negBalance < 44) {
if (ssDCval < 64) {
if (Rs < 1362) {
return 1;
}
else if (Rs >= 1362) {
return 2;
}
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (negBalance >= 44) {
if (SC < 3955) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 2;
}
}
else if (SC >= 3955) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 4;
}
}
}
}
}
else if (Rs >= 1737) {
if (RsCsDiff < 205) {
if (mu_mv_mag_sq < 62) {
if (refDCval < 84) {
if (Rs < 3013) {
return 8;
}
else if (Rs >= 3013) {
if (SC < 5342) {
return 4;
}
else if (SC >= 5342) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
return 4;
}
}
}
}
else if (refDCval >= 84) {
if (negBalance < 50) {
if (negBalance < 38) {
if (Rs < 3394) {
return 8;
}
else if (Rs >= 3394) {
return 1;
}
}
else if (negBalance >= 38) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (Cs < 2466) {
return 2;
}
else if (Cs >= 2466) {
return 8;
}
}
}
}
else if (negBalance >= 50) {
if (tcor < 87) {
if (posBalance < 63) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
if (diffAFD < 3) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
if (Cs < 2540) {
return 8;
}
else if (Cs >= 2540) {
return 4;
}
}
}
else if (diffAFD >= 3) {
return 8;
}
}
}
else if (posBalance >= 63) {
return 4;
}
}
else if (tcor >= 87) {
if (diffRsCsdiff < 11) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (Rs < 2538) {
return 2;
}
else if (Rs >= 2538) {
return 4;
}
}
}
else if (diffRsCsdiff >= 11) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 62) {
if (Rs < 3363) {
if (negBalance < 51) {
if (Cs < 2508) {
if (Rs < 2731) {
return 8;
}
else if (Rs >= 2731) {
return 4;
}
}
else if (Cs >= 2508) {
return 8;
}
}
else if (negBalance >= 51) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
else if (Rs >= 3363) {
if (Rs < 4588) {
return 4;
}
else if (Rs >= 4588) {
return 2;
}
}
}
}
else if (RsCsDiff >= 205) {
if (refDCval < 54) {
return 4;
}
else if (refDCval >= 54) {
if (posBalance < 64) {
if (SC < 5001) {
if (Rs < 2235) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
else if (Rs >= 2235) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (SC >= 5001) {
if (MVDiff < 11) {
if (refDCval < 78) {
if (RsCsDiff < 341) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (Rs < 3868) {
if (Rs < 3307) {
return 8;
}
else if (Rs >= 3307) {
return 1;
}
}
else if (Rs >= 3868) {
return 2;
}
}
}
else if (RsCsDiff >= 341) {
if (Cs < 3599) {
return 8;
}
else if (Cs >= 3599) {
return 4;
}
}
}
else if (refDCval >= 78) {
if (Rs < 4618) {
return 4;
}
else if (Rs >= 4618) {
return 1;
}
}
}
else if (MVDiff >= 11) {
if (mu_mv_mag_sq < 57) {
if (refDCval < 79) {
return 8;
}
else if (refDCval >= 79) {
if (negBalance < 48) {
if (Rs < 2395) {
return 4;
}
else if (Rs >= 2395) {
return 8;
}
}
else if (negBalance >= 48) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 64) {
if (refDCval < 90) {
if (TSC < 12) {
if (diffMVdiffVal < -2) {
if (MVDiff < 13) {
if (Rs < 3967) {
return 8;
}
else if (Rs >= 3967) {
return 2;
}
}
else if (MVDiff >= 13) {
return 4;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (refDCval >= 90) {
return 2;
}
}
}
}
}
}
else if (AFD >= 18) {
if (SC < 6439) {
if (refDCval < 93) {
if (ssDCval < 68) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
if (negBalance < 55) {
if (SC < 4574) {
if (Rs < 999) {
return 2;
}
else if (Rs >= 999) {
return 1;
}
}
else if (SC >= 4574) {
if (Cs < 2687) {
return 2;
}
else if (Cs >= 2687) {
return 4;
}
}
}
else if (negBalance >= 55) {
if (Rs < 1329) {
return 8;
}
else if (Rs >= 1329) {
return 4;
}
}
}
}
else if (ssDCval >= 68) {
if (refDCval < 78) {
if (Rs < 3382) {
if (diffAFD < -2) {
if (Rs < 2313) {
return 8;
}
else if (Rs >= 2313) {
return 1;
}
}
else if (diffAFD >= -2) {
if (posBalance < 58) {
if (tcor < 76) {
if (MVDiff < 28) {
if (Rs < 3010) {
return 1;
}
else if (Rs >= 3010) {
return 2;
}
}
else if (MVDiff >= 28) {
return 8;
}
}
else if (tcor >= 76) {
return 8;
}
}
else if (posBalance >= 58) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
if (Cs < 2251) {
return 2;
}
else if (Cs >= 2251) {
if (Rs < 920) {
return 2;
}
else if (Rs >= 920) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3382) {
if (Cs < 2405) {
return 8;
}
else if (Cs >= 2405) {
return 1;
}
}
}
else if (refDCval >= 78) {
if (MVDiff < 17) {
if (TSC < 14) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (mu_mv_mag_sq < 3) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (TSC >= 14) {
if (refDCval < 87) {
if (Rs < 2666) {
return 4;
}
else if (Rs >= 2666) {
return 8;
}
}
else if (refDCval >= 87) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
return 1;
}
}
}
}
else if (MVDiff >= 17) {
if (diffRsCsdiff < 14) {
if (CsDiff < 12) {
return 4;
}
else if (CsDiff >= 12) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
if (Cs < 2070) {
return 8;
}
else if (Cs >= 2070) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 14) {
if (Cs < 2624) {
if (Rs < 2076) {
return 4;
}
else if (Rs >= 2076) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
return 8;
}
}
}
else if (Cs >= 2624) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 93) {
if (SC < 5386) {
if (mcTcor < 94) {
if (ssDCval < 96) {
if (Cs < 2689) {
if (diffRsCsdiff < 1) {
if (tcor < 75) {
if (Rs < 2479) {
return 4;
}
else if (Rs >= 2479) {
return 8;
}
}
else if (tcor >= 75) {
return 8;
}
}
else if (diffRsCsdiff >= 1) {
if (Cs < 2043) {
return 1;
}
else if (Cs >= 2043) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 1;
}
}
}
}
else if (Cs >= 2689) {
return 4;
}
}
else if (ssDCval >= 96) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (Rs < 951) {
return 2;
}
else if (Rs >= 951) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 1851) {
return 4;
}
else if (Rs >= 1851) {
return 2;
}
}
}
else if (SC >= 5386) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
return 4;
}
}
}
}
else if (SC >= 6439) {
if (CsDiff < 17) {
if (negBalance < 44) {
if (Rs < 4692) {
return 2;
}
else if (Rs >= 4692) {
return 1;
}
}
else if (negBalance >= 44) {
if (diffRsCsdiff < -145) {
return 2;
}
else if (diffRsCsdiff >= -145) {
if (AFD < 19) {
if (Rs < 3356) {
return 4;
}
else if (Rs >= 3356) {
return 8;
}
}
else if (AFD >= 19) {
return 8;
}
}
}
}
else if (CsDiff >= 17) {
if (Rs < 1852) {
return 8;
}
else if (Rs >= 1852) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 101) {
if (AFD < 21) {
if (ssDCval < 105) {
if (Rs < 3713) {
if (Cs < 2637) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 2;
}
}
}
}
else if (Cs >= 2637) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
return 4;
}
}
}
else if (Rs >= 3713) {
if (mcTcor < 94) {
if (mu_mv_mag_sq < 71) {
return 8;
}
else if (mu_mv_mag_sq >= 71) {
return 1;
}
}
else if (mcTcor >= 94) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (Rs < 4764) {
return 2;
}
else if (Rs >= 4764) {
return 4;
}
}
}
}
}
else if (ssDCval >= 105) {
if (tcor < 70) {
if (refDCval < 121) {
return 8;
}
else if (refDCval >= 121) {
if (MVDiff < 60) {
if (AFD < 19) {
if (posBalance < 59) {
if (negBalance < 45) {
if (Rs < 2753) {
return 1;
}
else if (Rs >= 2753) {
if (MVDiff < 13) {
return 1;
}
else if (MVDiff >= 13) {
return 8;
}
}
}
else if (negBalance >= 45) {
if (mu_mv_mag_sq < 15) {
return 8;
}
else if (mu_mv_mag_sq >= 15) {
return 1;
}
}
}
else if (posBalance >= 59) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
}
else if (AFD >= 19) {
if (MVDiff < 12) {
if (Rs < 2583) {
return 2;
}
else if (Rs >= 2583) {
if (Rs < 5335) {
return 1;
}
else if (Rs >= 5335) {
return 8;
}
}
}
else if (MVDiff >= 12) {
return 8;
}
}
}
else if (MVDiff >= 60) {
return 2;
}
}
}
else if (tcor >= 70) {
if (SC < 4901) {
if (negBalance < 43) {
if (mcTcor < 92) {
if (CsDiff < 11) {
if (Rs < 2682) {
return 8;
}
else if (Rs >= 2682) {
return 2;
}
}
else if (CsDiff >= 11) {
return 4;
}
}
else if (mcTcor >= 92) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
if (Rs < 1271) {
return 2;
}
else if (Rs >= 1271) {
if (Rs < 1789) {
return 1;
}
else if (Rs >= 1789) {
return 8;
}
}
}
}
}
else if (negBalance >= 43) {
if (SC < 3689) {
return 1;
}
else if (SC >= 3689) {
if (SC < 3989) {
if (Rs < 1897) {
if (Rs < 1580) {
return 4;
}
else if (Rs >= 1580) {
return 8;
}
}
else if (Rs >= 1897) {
if (Rs < 1908) {
return 1;
}
else if (Rs >= 1908) {
return 2;
}
}
}
else if (SC >= 3989) {
return 8;
}
}
}
}
else if (SC >= 4901) {
if (mcTcor < 85) {
if (AFD < 20) {
if (diffAFD < -3) {
if (Rs < 3517) {
return 8;
}
else if (Rs >= 3517) {
return 1;
}
}
else if (diffAFD >= -3) {
if (diffMVdiffVal < -7) {
return 1;
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
}
else if (AFD >= 20) {
return 4;
}
}
else if (mcTcor >= 85) {
if (MVDiff < 9) {
if (Cs < 2573) {
if (ssDCval < 108) {
return 4;
}
else if (ssDCval >= 108) {
if (diffAFD < 9) {
if (diffRsCsdiff < -31) {
if (Cs < 2288) {
return 4;
}
else if (Cs >= 2288) {
return 8;
}
}
else if (diffRsCsdiff >= -31) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
if (Rs < 5482) {
return 8;
}
else if (Rs >= 5482) {
return 4;
}
}
}
}
}
else if (diffAFD >= 9) {
return 1;
}
}
}
else if (Cs >= 2573) {
if (refDCval < 115) {
if (tcor < 87) {
if (Rs < 3889) {
if (Rs < 2871) {
return 8;
}
else if (Rs >= 2871) {
return 4;
}
}
else if (Rs >= 3889) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 4;
}
}
}
else if (tcor >= 87) {
return 4;
}
}
else if (refDCval >= 115) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
else if (MVDiff >= 9) {
if (refDCval < 158) {
if (CsDiff < 13) {
if (negBalance < 29) {
if (Rs < 3316) {
return 2;
}
else if (Rs >= 3316) {
return 8;
}
}
else if (negBalance >= 29) {
if (MVDiff < 15) {
if (CsDiff < 10) {
if (RsDiff < 14) {
if (Rs < 5381) {
if (Cs < 2025) {
return 4;
}
else if (Cs >= 2025) {
return 8;
}
}
else if (Rs >= 5381) {
return 4;
}
}
else if (RsDiff >= 14) {
return 2;
}
}
else if (CsDiff >= 10) {
if (tcor < 89) {
if (RsCsDiff < 280) {
return 4;
}
else if (RsCsDiff >= 280) {
return 8;
}
}
else if (tcor >= 89) {
return 8;
}
}
}
else if (MVDiff >= 15) {
return 8;
}
}
}
else if (CsDiff >= 13) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
if (CsDiff < 17) {
if (Rs < 4422) {
return 8;
}
else if (Rs >= 4422) {
return 1;
}
}
else if (CsDiff >= 17) {
return 2;
}
}
}
}
else if (refDCval >= 158) {
if (posBalance < 31) {
if (AFD < 16) {
return 1;
}
else if (AFD >= 16) {
return 8;
}
}
else if (posBalance >= 31) {
return 4;
}
}
}
}
}
}
}
}
else if (AFD >= 21) {
if (MVDiff < 29) {
if (RsCsDiff < 189) {
return 2;
}
else if (RsCsDiff >= 189) {
if (negBalance < 55) {
if (refDCval < 122) {
if (Cs < 3417) {
if (Rs < 2966) {
if (Rs < 2016) {
return 1;
}
else if (Rs >= 2016) {
return 2;
}
}
else if (Rs >= 2966) {
return 4;
}
}
else if (Cs >= 3417) {
return 2;
}
}
else if (refDCval >= 122) {
if (Cs < 2096) {
return 2;
}
else if (Cs >= 2096) {
if (diffRsCsdiff < -220) {
return 4;
}
else if (diffRsCsdiff >= -220) {
return 8;
}
}
}
}
else if (negBalance >= 55) {
if (RsCsDiff < 243) {
if (Rs < 2760) {
return 2;
}
else if (Rs >= 2760) {
return 1;
}
}
else if (RsCsDiff >= 243) {
if (AFD < 23) {
if (Rs < 4275) {
if (diffTSC < 1) {
if (CsDiff < 14) {
return 4;
}
else if (CsDiff >= 14) {
return 8;
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (Rs >= 4275) {
return 1;
}
}
else if (AFD >= 23) {
if (tcor < 78) {
if (diffRsCsdiff < -65) {
if (Rs < 3883) {
return 8;
}
else if (Rs >= 3883) {
return 1;
}
}
else if (diffRsCsdiff >= -65) {
if (Rs < 1807) {
return 2;
}
else if (Rs >= 1807) {
return 1;
}
}
}
else if (tcor >= 78) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 29) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
if (tcor < 60) {
return 8;
}
else if (tcor >= 60) {
if (Cs < 4972) {
return 2;
}
else if (Cs >= 4972) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 16) {
if (TSC < 8) {
if (refDCval < 126) {
if (Rs < 8215) {
if (Cs < 3036) {
return 4;
}
else if (Cs >= 3036) {
return 8;
}
}
else if (Rs >= 8215) {
return 1;
}
}
else if (refDCval >= 126) {
if (RsCsDiff < 429) {
return 2;
}
else if (RsCsDiff >= 429) {
return 8;
}
}
}
else if (TSC >= 8) {
if (refDCval < 140) {
if (Cs < 4613) {
if (TSC < 14) {
if (tcor < 80) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 9) {
if (AFD < 20) {
if (RsCsDiff < 473) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
if (Rs < 3132) {
return 2;
}
else if (Rs >= 3132) {
return 4;
}
}
}
else if (RsCsDiff >= 473) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
return 8;
}
}
}
else if (AFD >= 20) {
if (SC < 8023) {
if (negBalance < 57) {
return 2;
}
else if (negBalance >= 57) {
if (Rs < 4179) {
return 8;
}
else if (Rs >= 4179) {
return 2;
}
}
}
else if (SC >= 8023) {
if (Rs < 4508) {
return 4;
}
else if (Rs >= 4508) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (mcTcor < 83) {
if (Rs < 2629) {
if (Rs < 2328) {
return 4;
}
else if (Rs >= 2328) {
return 2;
}
}
else if (Rs >= 2629) {
return 8;
}
}
else if (mcTcor >= 83) {
if (SC < 6904) {
if (negBalance < 46) {
if (Rs < 4595) {
return 2;
}
else if (Rs >= 4595) {
return 1;
}
}
else if (negBalance >= 46) {
if (ssDCval < 98) {
if (diffMVdiffVal < 4) {
return 4;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 3216) {
if (Rs < 2875) {
return 2;
}
else if (Rs >= 2875) {
return 4;
}
}
else if (Rs >= 3216) {
return 8;
}
}
}
else if (ssDCval >= 98) {
if (Cs < 2266) {
return 2;
}
else if (Cs >= 2266) {
return 1;
}
}
}
}
else if (SC >= 6904) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (tcor < 78) {
return 4;
}
else if (tcor >= 78) {
return 2;
}
}
}
}
}
}
else if (diffAFD >= 2) {
if (AFD < 24) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (Rs < 4715) {
if (diffTSC < 4) {
return 4;
}
else if (diffTSC >= 4) {
return 8;
}
}
else if (Rs >= 4715) {
return 8;
}
}
}
else if (AFD >= 24) {
if (mu_mv_mag_sq < 22) {
if (SC < 5153) {
return 1;
}
else if (SC >= 5153) {
return 2;
}
}
else if (mu_mv_mag_sq >= 22) {
if (Rs < 3291) {
return 4;
}
else if (Rs >= 3291) {
return 8;
}
}
}
}
}
else if (tcor >= 80) {
if (negBalance < 52) {
if (gchDC < 3) {
if (RsCsDiff < 345) {
if (tcor < 87) {
if (Rs < 5334) {
return 8;
}
else if (Rs >= 5334) {
return 4;
}
}
else if (tcor >= 87) {
return 2;
}
}
else if (RsCsDiff >= 345) {
if (Cs < 4064) {
return 2;
}
else if (Cs >= 4064) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (Rs < 6446) {
return 8;
}
else if (Rs >= 6446) {
return 2;
}
}
}
else if (negBalance >= 52) {
if (RsDiff < 22) {
if (AFD < 21) {
if (posBalance < 61) {
if (ssDCval < 72) {
if (Rs < 4414) {
return 4;
}
else if (Rs >= 4414) {
if (Rs < 4763) {
return 8;
}
else if (Rs >= 4763) {
return 2;
}
}
}
else if (ssDCval >= 72) {
return 8;
}
}
else if (posBalance >= 61) {
return 4;
}
}
else if (AFD >= 21) {
if (Rs < 3593) {
return 8;
}
else if (Rs >= 3593) {
return 4;
}
}
}
else if (RsDiff >= 22) {
return 2;
}
}
}
}
else if (TSC >= 14) {
if (CsDiff < 17) {
if (negBalance < 51) {
if (diffAFD < 3) {
if (ssDCval < 83) {
if (Rs < 7262) {
return 2;
}
else if (Rs >= 7262) {
return 1;
}
}
else if (ssDCval >= 83) {
if (refDCval < 97) {
return 1;
}
else if (refDCval >= 97) {
if (tcor < 72) {
if (TSC < 19) {
return 8;
}
else if (TSC >= 19) {
return 1;
}
}
else if (tcor >= 72) {
if (Cs < 3933) {
return 1;
}
else if (Cs >= 3933) {
return 2;
}
}
}
}
}
else if (diffAFD >= 3) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
if (RsDiff < 19) {
if (Rs < 3931) {
return 8;
}
else if (Rs >= 3931) {
return 4;
}
}
else if (RsDiff >= 19) {
return 4;
}
}
}
}
else if (negBalance >= 51) {
if (AFD < 29) {
if (MVDiff < 13) {
if (mu_mv_mag_sq < 45) {
if (negBalance < 57) {
return 4;
}
else if (negBalance >= 57) {
if (Rs < 4335) {
return 8;
}
else if (Rs >= 4335) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 45) {
return 8;
}
}
else if (MVDiff >= 13) {
if (diffMVdiffVal < 5) {
return 8;
}
else if (diffMVdiffVal >= 5) {
if (TSC < 15) {
if (Cs < 2359) {
return 8;
}
else if (Cs >= 2359) {
return 2;
}
}
else if (TSC >= 15) {
if (refDCval < 129) {
return 8;
}
else if (refDCval >= 129) {
return 2;
}
}
}
}
}
else if (AFD >= 29) {
if (refDCval < 106) {
if (Rs < 4132) {
return 1;
}
else if (Rs >= 4132) {
if (AFD < 38) {
return 2;
}
else if (AFD >= 38) {
return 1;
}
}
}
else if (refDCval >= 106) {
return 8;
}
}
}
}
else if (CsDiff >= 17) {
if (TSC < 18) {
if (posBalance < 62) {
if (tcor < 57) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
return 4;
}
}
else if (tcor >= 57) {
if (AFD < 23) {
if (Rs < 4954) {
return 8;
}
else if (Rs >= 4954) {
return 4;
}
}
else if (AFD >= 23) {
if (Rs < 3329) {
return 4;
}
else if (Rs >= 3329) {
return 2;
}
}
}
}
else if (posBalance >= 62) {
if (negBalance < 54) {
return 4;
}
else if (negBalance >= 54) {
if (Rs < 5638) {
return 1;
}
else if (Rs >= 5638) {
return 4;
}
}
}
}
else if (TSC >= 18) {
if (RsDiff < 32) {
if (diffTSC < 3) {
if (Cs < 4570) {
return 8;
}
else if (Cs >= 4570) {
return 2;
}
}
else if (diffTSC >= 3) {
if (Rs < 4254) {
if (Cs < 3089) {
return 8;
}
else if (Cs >= 3089) {
return 4;
}
}
else if (Rs >= 4254) {
return 4;
}
}
}
else if (RsDiff >= 32) {
if (Rs < 4875) {
return 1;
}
else if (Rs >= 4875) {
return 8;
}
}
}
}
}
}
else if (Cs >= 4613) {
if (ssDCval < 111) {
if (posBalance < 53) {
if (Cs < 4627) {
return 4;
}
else if (Cs >= 4627) {
return 8;
}
}
else if (posBalance >= 53) {
if (mu_mv_mag_sq < 64) {
return 4;
}
else if (mu_mv_mag_sq >= 64) {
if (Rs < 6151) {
return 8;
}
else if (Rs >= 6151) {
return 4;
}
}
}
}
else if (ssDCval >= 111) {
if (tcor < 83) {
return 8;
}
else if (tcor >= 83) {
return 1;
}
}
}
}
else if (refDCval >= 140) {
if (refDCval < 166) {
if (gchDC < 2) {
if (ssDCval < 156) {
if (mcTcor < 89) {
if (RsDiff < 18) {
if (Cs < 2362) {
return 4;
}
else if (Cs >= 2362) {
return 8;
}
}
else if (RsDiff >= 18) {
if (posBalance < 54) {
if (Cs < 2758) {
return 4;
}
else if (Cs >= 2758) {
return 8;
}
}
else if (posBalance >= 54) {
return 4;
}
}
}
else if (mcTcor >= 89) {
if (MVDiff < 18) {
if (Cs < 4067) {
return 2;
}
else if (Cs >= 4067) {
return 4;
}
}
else if (MVDiff >= 18) {
if (Rs < 3824) {
return 2;
}
else if (Rs >= 3824) {
return 8;
}
}
}
}
else if (ssDCval >= 156) {
return 4;
}
}
else if (gchDC >= 2) {
if (ssDCval < 157) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
if (Rs < 5825) {
return 2;
}
else if (Rs >= 5825) {
return 8;
}
}
}
else if (ssDCval >= 157) {
if (Rs < 4507) {
return 8;
}
else if (Rs >= 4507) {
return 4;
}
}
}
}
else if (refDCval >= 166) {
if (Rs < 8679) {
return 2;
}
else if (Rs >= 8679) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (SC < 7053) {
if (negBalance < 55) {
if (Rs < 4138) {
if (ssDCval < 89) {
if (ssDCval < 67) {
if (diffRsCsdiff < -9) {
return 8;
}
else if (diffRsCsdiff >= -9) {
if (RsDiff < 12) {
if (negBalance < 54) {
return 4;
}
else if (negBalance >= 54) {
return 8;
}
}
else if (RsDiff >= 12) {
return 8;
}
}
}
else if (ssDCval >= 67) {
if (Cs < 3583) {
if (diffMVdiffVal < -5) {
if (TSC < 12) {
if (Rs < 3749) {
return 1;
}
else if (Rs >= 3749) {
return 4;
}
}
else if (TSC >= 12) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
else if (Cs >= 3583) {
if (Rs < 2837) {
return 4;
}
else if (Rs >= 2837) {
return 1;
}
}
}
}
else if (ssDCval >= 89) {
if (Cs < 2667) {
if (SC < 6091) {
return 8;
}
else if (SC >= 6091) {
return 1;
}
}
else if (Cs >= 2667) {
if (ssDCval < 107) {
return 4;
}
else if (ssDCval >= 107) {
if (SC < 6898) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (SC >= 6898) {
return 4;
}
}
}
}
}
else if (Rs >= 4138) {
return 1;
}
}
else if (negBalance >= 55) {
if (Cs < 2675) {
if (AFD < 34) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
if (RsCsDiff < 302) {
return 4;
}
else if (RsCsDiff >= 302) {
return 8;
}
}
}
else if (AFD >= 34) {
return 1;
}
}
else if (Cs >= 2675) {
if (Rs < 3140) {
return 4;
}
else if (Rs >= 3140) {
return 8;
}
}
}
}
else if (SC >= 7053) {
if (negBalance < 52) {
if (tcor < 85) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
if (Rs < 4824) {
if (MVDiff < 10) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
return 4;
}
}
else if (MVDiff >= 10) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (Cs < 3267) {
return 8;
}
else if (Cs >= 3267) {
return 1;
}
}
}
}
else if (Rs >= 4824) {
if (TSC < 13) {
if (MVDiff < 12) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 8;
}
}
else if (MVDiff >= 12) {
return 4;
}
}
else if (TSC >= 13) {
if (gchDC < 3) {
if (Rs < 6351) {
if (Rs < 5960) {
return 4;
}
else if (Rs >= 5960) {
return 2;
}
}
else if (Rs >= 6351) {
return 4;
}
}
else if (gchDC >= 3) {
return 2;
}
}
}
}
}
else if (tcor >= 85) {
if (SC < 7649) {
if (mcTcor < 92) {
if (negBalance < 37) {
return 8;
}
else if (negBalance >= 37) {
return 4;
}
}
else if (mcTcor >= 92) {
return 8;
}
}
else if (SC >= 7649) {
if (diffRsCsdiff < -140) {
if (Cs < 3583) {
return 8;
}
else if (Cs >= 3583) {
if (Cs < 3712) {
return 2;
}
else if (Cs >= 3712) {
return 4;
}
}
}
else if (diffRsCsdiff >= -140) {
if (TSC < 11) {
if (MVDiff < 4) {
if (Cs < 2735) {
return 8;
}
else if (Cs >= 2735) {
return 4;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (TSC >= 11) {
return 8;
}
}
}
}
}
else if (negBalance >= 52) {
if (posBalance < 57) {
if (AFD < 17) {
if (tcor < 82) {
return 1;
}
else if (tcor >= 82) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 8;
}
}
}
else if (AFD >= 17) {
return 4;
}
}
else if (posBalance >= 57) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
if (Rs < 3188) {
return 4;
}
else if (Rs >= 3188) {
if (Rs < 8611) {
if (Rs < 3465) {
if (Cs < 4401) {
return 8;
}
else if (Cs >= 4401) {
return 4;
}
}
else if (Rs >= 3465) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (posBalance < 59) {
if (Rs < 5893) {
return 8;
}
else if (Rs >= 5893) {
return 4;
}
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (Rs >= 8611) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (Cs < 1643) {
if (Rs < 3145) {
if (ssDCval < 122) {
if (Rs < 1506) {
if (negBalance < 1) {
return 2;
}
else if (negBalance >= 1) {
if (refDCval < 63) {
if (negBalance < 25) {
if (refDCval < 53) {
if (Cs < 1433) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (Rs < 1087) {
if (Rs < 683) {
return 1;
}
else if (Rs >= 683) {
return 8;
}
}
else if (Rs >= 1087) {
return 4;
}
}
}
else if (Cs >= 1433) {
if (ssDCval < 33) {
return 8;
}
else if (ssDCval >= 33) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
if (Rs < 969) {
return 1;
}
else if (Rs >= 969) {
return 4;
}
}
}
}
}
else if (refDCval >= 53) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (Rs < 942) {
return 8;
}
else if (Rs >= 942) {
return 4;
}
}
}
}
else if (negBalance >= 25) {
if (Rs < 786) {
return 1;
}
else if (Rs >= 786) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Rs < 1225) {
return 8;
}
else if (Rs >= 1225) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 63) {
if (ssDCval < 77) {
if (Cs < 1433) {
if (Cs < 1424) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (MVDiff < 2) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (Cs >= 1424) {
return 2;
}
}
else if (Cs >= 1433) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (ssDCval >= 77) {
if (Rs < 1074) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (Rs >= 1074) {
return 8;
}
}
}
}
}
else if (Rs >= 1506) {
if (negBalance < 18) {
if (Rs < 2988) {
if (MVDiff < 105) {
if (negBalance < 15) {
if (Cs < 1338) {
if (Rs < 2339) {
return 8;
}
else if (Rs >= 2339) {
return 2;
}
}
else if (Cs >= 1338) {
return 8;
}
}
else if (negBalance >= 15) {
if (Rs < 2925) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (Rs >= 2925) {
return 1;
}
}
}
else if (MVDiff >= 105) {
return 4;
}
}
else if (Rs >= 2988) {
if (ssDCval < 82) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
return 1;
}
}
else if (ssDCval >= 82) {
return 8;
}
}
}
else if (negBalance >= 18) {
if (mcTcor < 98) {
if (posBalance < 45) {
if (Rs < 1874) {
return 8;
}
else if (Rs >= 1874) {
if (Cs < 1537) {
return 4;
}
else if (Cs >= 1537) {
if (Rs < 2738) {
return 8;
}
else if (Rs >= 2738) {
return 8;
}
}
}
}
else if (posBalance >= 45) {
if (SC < 4159) {
if (negBalance < 21) {
if (Cs < 1588) {
return 4;
}
else if (Cs >= 1588) {
return 8;
}
}
else if (negBalance >= 21) {
if (diffAFD < -2) {
if (diffRsCsdiff < -10) {
return 8;
}
else if (diffRsCsdiff >= -10) {
return 2;
}
}
else if (diffAFD >= -2) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 2;
}
}
}
}
}
else if (SC >= 4159) {
if (Cs < 1547) {
if (Rs < 2917) {
return 4;
}
else if (Rs >= 2917) {
return 2;
}
}
else if (Cs >= 1547) {
if (Rs < 2599) {
return 1;
}
else if (Rs >= 2599) {
return 8;
}
}
}
}
}
else if (mcTcor >= 98) {
if (ssDCval < 85) {
if (RsDiff < 6) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (posBalance < 57) {
if (Rs < 1512) {
return 8;
}
else if (Rs >= 1512) {
if (posBalance < 29) {
return 8;
}
else if (posBalance >= 29) {
if (Cs < 1617) {
if (Cs < 1354) {
return 8;
}
else if (Cs >= 1354) {
return 4;
}
}
else if (Cs >= 1617) {
if (Cs < 1627) {
if (Rs < 2613) {
return 8;
}
else if (Rs >= 2613) {
return 4;
}
}
else if (Cs >= 1627) {
return 4;
}
}
}
}
}
else if (posBalance >= 57) {
return 8;
}
}
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (ssDCval >= 85) {
if (ssDCval < 117) {
if (Rs < 2483) {
return 8;
}
else if (Rs >= 2483) {
if (Rs < 2776) {
return 4;
}
else if (Rs >= 2776) {
return 8;
}
}
}
else if (ssDCval >= 117) {
if (mu_mv_mag_sq < 124) {
return 8;
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 2004) {
return 1;
}
else if (Rs >= 2004) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 122) {
if (Rs < 1339) {
if (refDCval < 127) {
if (mu_mv_mag_sq < 121) {
if (Cs < 1477) {
return 8;
}
else if (Cs >= 1477) {
return 4;
}
}
else if (mu_mv_mag_sq >= 121) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (Rs < 933) {
return 8;
}
else if (Rs >= 933) {
return 4;
}
}
}
}
else if (refDCval >= 127) {
if (tcor < 95) {
if (ssDCval < 132) {
if (Cs < 1358) {
return 4;
}
else if (Cs >= 1358) {
return 8;
}
}
else if (ssDCval >= 132) {
if (Rs < 867) {
return 2;
}
else if (Rs >= 867) {
return 1;
}
}
}
else if (tcor >= 95) {
if (refDCval < 128) {
if (Rs < 972) {
return 8;
}
else if (Rs >= 972) {
return 4;
}
}
else if (refDCval >= 128) {
return 8;
}
}
}
}
else if (Rs >= 1339) {
if (refDCval < 133) {
if (Rs < 2717) {
if (RsDiff < 1) {
if (Rs < 1489) {
return 4;
}
else if (Rs >= 1489) {
if (Cs < 1390) {
return 4;
}
else if (Cs >= 1390) {
return 8;
}
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (Rs >= 2717) {
if (Rs < 2915) {
return 1;
}
else if (Rs >= 2915) {
return 2;
}
}
}
else if (refDCval >= 133) {
if (MVDiff < 1) {
if (Rs < 2363) {
return 2;
}
else if (Rs >= 2363) {
return 1;
}
}
else if (MVDiff >= 1) {
if (diffRsCsdiff < -19) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 2;
}
}
else if (diffRsCsdiff >= -19) {
if (Cs < 1439) {
return 8;
}
else if (Cs >= 1439) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 3145) {
if (ssDCval < 77) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 63) {
if (Cs < 1558) {
if (posBalance < 27) {
return 4;
}
else if (posBalance >= 27) {
return 8;
}
}
else if (Cs >= 1558) {
return 4;
}
}
else if (ssDCval >= 63) {
if (negBalance < 23) {
if (Rs < 4281) {
return 1;
}
else if (Rs >= 4281) {
return 8;
}
}
else if (negBalance >= 23) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1570) {
if (AFD < 3) {
if (Rs < 3556) {
return 1;
}
else if (Rs >= 3556) {
return 8;
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (Cs >= 1570) {
if (Rs < 3265) {
return 8;
}
else if (Rs >= 3265) {
return 4;
}
}
}
}
else if (ssDCval >= 77) {
if (ssDCval < 111) {
if (posBalance < 42) {
if (SC < 5682) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 4;
}
}
else if (SC >= 5682) {
if (ssDCval < 96) {
if (negBalance < 21) {
return 2;
}
else if (negBalance >= 21) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (posBalance >= 42) {
if (refDCval < 91) {
return 2;
}
else if (refDCval >= 91) {
if (Cs < 1532) {
if (Rs < 4455) {
if (Rs < 3270) {
return 2;
}
else if (Rs >= 3270) {
return 4;
}
}
else if (Rs >= 4455) {
return 2;
}
}
else if (Cs >= 1532) {
if (Rs < 3745) {
return 2;
}
else if (Rs >= 3745) {
return 8;
}
}
}
}
}
else if (ssDCval >= 111) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (Rs < 3657) {
if (Rs < 3363) {
return 2;
}
else if (Rs >= 3363) {
return 1;
}
}
else if (Rs >= 3657) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1643) {
if (ssDCval < 102) {
if (SC < 4332) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 1789) {
if (MVDiff < 3) {
if (ssDCval < 63) {
if (Rs < 831) {
return 1;
}
else if (Rs >= 831) {
if (Rs < 1670) {
if (Rs < 1377) {
return 8;
}
else if (Rs >= 1377) {
return 4;
}
}
else if (Rs >= 1670) {
return 8;
}
}
}
else if (ssDCval >= 63) {
if (Rs < 2383) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 985) {
return 8;
}
else if (Rs >= 985) {
return 1;
}
}
}
else if (Rs >= 2383) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 991) {
return 8;
}
else if (Rs >= 991) {
return 4;
}
}
}
else if (Cs >= 1789) {
if (diffRsCsdiff < -2) {
if (ssDCval < 62) {
if (RsDiff < 1) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 1;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (ssDCval >= 62) {
if (Rs < 1022) {
return 8;
}
else if (Rs >= 1022) {
return 1;
}
}
}
else if (diffRsCsdiff >= -2) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (posBalance < 36) {
if (Cs < 1816) {
return 8;
}
else if (Cs >= 1816) {
return 1;
}
}
else if (posBalance >= 36) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 34) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (Rs < 2280) {
return 2;
}
else if (Rs >= 2280) {
return 8;
}
}
}
else if (negBalance >= 34) {
if (Rs < 1546) {
return 1;
}
else if (Rs >= 1546) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 2580) {
if (Rs < 1019) {
return 1;
}
else if (Rs >= 1019) {
if (gchDC < 2) {
if (Rs < 2404) {
if (Cs < 2390) {
if (SC < 2860) {
if (Rs < 1132) {
return 8;
}
else if (Rs >= 1132) {
return 4;
}
}
else if (SC >= 2860) {
return 8;
}
}
else if (Cs >= 2390) {
return 4;
}
}
else if (Rs >= 2404) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
if (ssDCval < 80) {
if (SC < 4234) {
return 1;
}
else if (SC >= 4234) {
return 8;
}
}
else if (ssDCval >= 80) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
if (diffRsCsdiff < -26) {
return 8;
}
else if (diffRsCsdiff >= -26) {
return 4;
}
}
}
}
}
else if (Rs >= 2580) {
return 4;
}
}
}
else if (SC >= 4332) {
if (Cs < 2197) {
if (Rs < 3835) {
if (posBalance < 23) {
if (Rs < 2389) {
if (ssDCval < 92) {
if (ssDCval < 91) {
if (Rs < 2349) {
return 8;
}
else if (Rs >= 2349) {
return 4;
}
}
else if (ssDCval >= 91) {
return 1;
}
}
else if (ssDCval >= 92) {
if (mu_mv_mag_sq < 120) {
if (negBalance < 8) {
return 2;
}
else if (negBalance >= 8) {
if (Rs < 2360) {
return 4;
}
else if (Rs >= 2360) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 120) {
return 4;
}
}
}
else if (Rs >= 2389) {
if (posBalance < 1) {
if (SC < 5183) {
return 8;
}
else if (SC >= 5183) {
if (Rs < 3362) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (Rs >= 3362) {
return 8;
}
}
}
else if (posBalance >= 1) {
return 8;
}
}
}
else if (posBalance >= 23) {
if (Rs < 3085) {
if (Cs < 1935) {
if (RsDiff < 2) {
if (diffRsCsdiff < -2) {
if (Rs < 2758) {
return 4;
}
else if (Rs >= 2758) {
if (Cs < 1750) {
return 8;
}
else if (Cs >= 1750) {
return 8;
}
}
}
else if (diffRsCsdiff >= -2) {
if (Cs < 1818) {
return 1;
}
else if (Cs >= 1818) {
if (Rs < 2518) {
return 1;
}
else if (Rs >= 2518) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (diffMVdiffVal < -1) {
if (Rs < 2946) {
return 8;
}
else if (Rs >= 2946) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2573) {
if (Rs < 2458) {
return 4;
}
else if (Rs >= 2458) {
return 8;
}
}
else if (Rs >= 2573) {
if (MVDiff < 1) {
if (ssDCval < 73) {
if (SC < 4880) {
if (negBalance < 24) {
if (posBalance < 40) {
if (SC < 4871) {
if (Rs < 2999) {
return 4;
}
else if (Rs >= 2999) {
return 4;
}
}
else if (SC >= 4871) {
return 4;
}
}
else if (posBalance >= 40) {
if (Rs < 2983) {
if (Rs < 2950) {
return 8;
}
else if (Rs >= 2950) {
return 4;
}
}
else if (Rs >= 2983) {
return 8;
}
}
}
else if (negBalance >= 24) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (posBalance < 37) {
if (Rs < 2932) {
return 4;
}
else if (Rs >= 2932) {
return 8;
}
}
else if (posBalance >= 37) {
return 4;
}
}
}
}
else if (SC >= 4880) {
return 4;
}
}
else if (ssDCval >= 73) {
if (Cs < 1871) {
return 4;
}
else if (Cs >= 1871) {
if (ssDCval < 84) {
return 4;
}
else if (ssDCval >= 84) {
return 8;
}
}
}
}
else if (MVDiff >= 1) {
if (Rs < 3021) {
return 8;
}
else if (Rs >= 3021) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 1935) {
if (negBalance < 19) {
if (Cs < 2095) {
if (Cs < 2068) {
return 8;
}
else if (Cs >= 2068) {
if (Rs < 2325) {
return 2;
}
else if (Rs >= 2325) {
return 1;
}
}
}
else if (Cs >= 2095) {
if (Rs < 2974) {
return 4;
}
else if (Rs >= 2974) {
return 8;
}
}
}
else if (negBalance >= 19) {
if (MVDiff < 5) {
if (ssDCval < 65) {
if (Rs < 2859) {
return 8;
}
else if (Rs >= 2859) {
return 4;
}
}
else if (ssDCval >= 65) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 2949) {
return 8;
}
else if (Rs >= 2949) {
return 4;
}
}
}
}
else if (MVDiff >= 5) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 8;
}
}
}
}
}
else if (Rs >= 3085) {
if (ssDCval < 86) {
if (RsDiff < 6) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
if (Rs < 3137) {
return 8;
}
else if (Rs >= 3137) {
if (Rs < 3232) {
return 4;
}
else if (Rs >= 3232) {
return 4;
}
}
}
}
else if (RsDiff >= 6) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 3236) {
return 4;
}
else if (Rs >= 3236) {
return 2;
}
}
}
}
else if (ssDCval >= 86) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
return 2;
}
}
}
}
}
else if (Rs >= 3835) {
if (posBalance < 41) {
if (Rs < 8367) {
if (CsDiff < 3) {
if (ssDCval < 66) {
if (Rs < 4591) {
return 8;
}
else if (Rs >= 4591) {
return 4;
}
}
else if (ssDCval >= 66) {
return 4;
}
}
else if (CsDiff >= 3) {
if (ssDCval < 95) {
if (ssDCval < 78) {
if (diffRsCsdiff < -8) {
if (SC < 6285) {
return 4;
}
else if (SC >= 6285) {
if (SC < 6472) {
return 4;
}
else if (SC >= 6472) {
if (Rs < 4432) {
return 4;
}
else if (Rs >= 4432) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= -8) {
if (Cs < 1814) {
return 4;
}
else if (Cs >= 1814) {
if (diffAFD < 1) {
if (TSC < 3) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (refDCval < 54) {
if (Rs < 4234) {
return 4;
}
else if (Rs >= 4234) {
if (Cs < 2097) {
return 8;
}
else if (Cs >= 2097) {
if (Cs < 2115) {
return 4;
}
else if (Cs >= 2115) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 4;
}
}
}
}
}
else if (refDCval >= 54) {
if (Rs < 4329) {
return 8;
}
else if (Rs >= 4329) {
return 8;
}
}
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (diffAFD >= 1) {
if (tcor < 95) {
if (Cs < 2183) {
if (Rs < 4202) {
return 4;
}
else if (Rs >= 4202) {
return 4;
}
}
else if (Cs >= 2183) {
return 4;
}
}
else if (tcor >= 95) {
return 4;
}
}
}
}
}
else if (ssDCval >= 78) {
if (Cs < 1867) {
return 8;
}
else if (Cs >= 1867) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
else if (ssDCval >= 95) {
if (Cs < 1920) {
return 4;
}
else if (Cs >= 1920) {
return 2;
}
}
}
}
else if (Rs >= 8367) {
return 2;
}
}
else if (posBalance >= 41) {
if (RsCsDiff < 59) {
if (mcTcor < 99) {
if (negBalance < 27) {
if (Cs < 1937) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 1937) {
if (AFD < 4) {
if (Rs < 4297) {
return 4;
}
else if (Rs >= 4297) {
return 4;
}
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (negBalance >= 27) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
return 8;
}
}
}
else if (mcTcor >= 99) {
return 8;
}
}
else if (RsCsDiff >= 59) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
else if (Cs >= 2197) {
if (SC < 8396) {
if (SC < 6958) {
if (AFD < 7) {
if (Rs < 2406) {
if (negBalance < 31) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
else if (negBalance >= 31) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
if (Cs < 2451) {
return 8;
}
else if (Cs >= 2451) {
return 4;
}
}
}
}
else if (Rs >= 2406) {
if (ssDCval < 47) {
if (ssDCval < 43) {
return 8;
}
else if (ssDCval >= 43) {
return 1;
}
}
else if (ssDCval >= 47) {
if (SC < 6393) {
if (Cs < 2283) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (mu_mv_mag_sq < 126) {
if (Rs < 3007) {
return 8;
}
else if (Rs >= 3007) {
return 2;
}
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
}
else if (Cs >= 2283) {
if (negBalance < 65) {
return 8;
}
else if (negBalance >= 65) {
return 4;
}
}
}
else if (SC >= 6393) {
if (TSC < 5) {
if (ssDCval < 58) {
if (Cs < 2318) {
if (Rs < 4504) {
if (Rs < 4427) {
return 4;
}
else if (Rs >= 4427) {
return 4;
}
}
else if (Rs >= 4504) {
return 4;
}
}
else if (Cs >= 2318) {
if (Rs < 3942) {
return 4;
}
else if (Rs >= 3942) {
return 8;
}
}
}
else if (ssDCval >= 58) {
if (diffAFD < -2) {
if (Rs < 3778) {
return 4;
}
else if (Rs >= 3778) {
return 8;
}
}
else if (diffAFD >= -2) {
if (SC < 6846) {
return 8;
}
else if (SC >= 6846) {
if (negBalance < 26) {
if (AFD < 3) {
if (SC < 6848) {
return 4;
}
else if (SC >= 6848) {
return 8;
}
}
else if (AFD >= 3) {
return 4;
}
}
else if (negBalance >= 26) {
return 8;
}
}
}
}
}
else if (TSC >= 5) {
if (ssDCval < 85) {
if (Rs < 2977) {
return 4;
}
else if (Rs >= 2977) {
return 8;
}
}
else if (ssDCval >= 85) {
if (Rs < 3205) {
return 8;
}
else if (Rs >= 3205) {
return 4;
}
}
}
}
}
}
}
else if (AFD >= 7) {
if (negBalance < 54) {
if (Rs < 3393) {
if (posBalance < 63) {
if (Rs < 3094) {
if (Rs < 1961) {
return 4;
}
else if (Rs >= 1961) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 2564) {
return 4;
}
else if (Rs >= 2564) {
return 8;
}
}
}
}
else if (Rs >= 3094) {
if (posBalance < 61) {
if (Cs < 3251) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 4;
}
}
else if (Cs >= 3251) {
return 8;
}
}
else if (posBalance >= 61) {
return 4;
}
}
}
else if (posBalance >= 63) {
if (diffTSC < -2) {
return 8;
}
else if (diffTSC >= -2) {
return 4;
}
}
}
else if (Rs >= 3393) {
if (Cs < 2614) {
if (Rs < 4086) {
return 8;
}
else if (Rs >= 4086) {
return 1;
}
}
else if (Cs >= 2614) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
}
else if (negBalance >= 54) {
return 4;
}
}
}
else if (SC >= 6958) {
if (refDCval < 85) {
if (negBalance < 26) {
if (negBalance < 4) {
return 4;
}
else if (negBalance >= 4) {
if (AFD < 2) {
if (TSCindex < 1) {
if (Rs < 3594) {
return 8;
}
else if (Rs >= 3594) {
return 4;
}
}
else if (TSCindex >= 1) {
if (Cs < 2386) {
return 4;
}
else if (Cs >= 2386) {
return 8;
}
}
}
else if (AFD >= 2) {
if (posBalance < 40) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
if (diffTSC < 1) {
if (mu_mv_mag_sq < 120) {
if (RsDiff < 4) {
if (Rs < 5069) {
return 8;
}
else if (Rs >= 5069) {
return 8;
}
}
else if (RsDiff >= 4) {
if (CsDiff < 4) {
if (Rs < 5096) {
return 4;
}
else if (Rs >= 5096) {
return 4;
}
}
else if (CsDiff >= 4) {
if (posBalance < 38) {
if (Rs < 4966) {
if (Cs < 2568) {
return 4;
}
else if (Cs >= 2568) {
return 8;
}
}
else if (Rs >= 4966) {
return 4;
}
}
else if (posBalance >= 38) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 120) {
return 4;
}
}
else if (diffTSC >= 1) {
if (Cs < 2399) {
return 4;
}
else if (Cs >= 2399) {
return 4;
}
}
}
}
else if (posBalance >= 40) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Rs < 4883) {
if (diffRsCsdiff < -4) {
return 4;
}
else if (diffRsCsdiff >= -4) {
if (Rs < 4788) {
if (Rs < 4659) {
return 4;
}
else if (Rs >= 4659) {
return 4;
}
}
else if (Rs >= 4788) {
return 4;
}
}
}
else if (Rs >= 4883) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 26) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
if (Rs < 3611) {
return 4;
}
else if (Rs >= 3611) {
return 8;
}
}
}
}
else if (refDCval >= 85) {
if (ssDCval < 98) {
if (mu_mv_mag_sq < 128) {
if (Cs < 3599) {
if (posBalance < 50) {
if (ssDCval < 94) {
return 4;
}
else if (ssDCval >= 94) {
if (Rs < 4289) {
return 8;
}
else if (Rs >= 4289) {
return 4;
}
}
}
else if (posBalance >= 50) {
return 4;
}
}
else if (Cs >= 3599) {
if (ssDCval < 96) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 4;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 3183) {
if (Rs < 4097) {
return 8;
}
else if (Rs >= 4097) {
return 4;
}
}
else if (Cs >= 3183) {
return 4;
}
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
else if (SC >= 8396) {
if (Rs < 6807) {
if (ssDCval < 78) {
if (Rs < 5995) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 27) {
if (Rs < 4780) {
return 1;
}
else if (Rs >= 4780) {
return 8;
}
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (Rs >= 5995) {
if (Cs < 4028) {
return 8;
}
else if (Cs >= 4028) {
return 2;
}
}
}
else if (ssDCval >= 78) {
if (Cs < 3666) {
if (Rs < 5223) {
return 4;
}
else if (Rs >= 5223) {
return 8;
}
}
else if (Cs >= 3666) {
return 8;
}
}
}
else if (Rs >= 6807) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (negBalance < 37) {
if (ssDCval < 117) {
if (Rs < 1628) {
if (Cs < 2089) {
return 8;
}
else if (Cs >= 2089) {
return 2;
}
}
else if (Rs >= 1628) {
if (diffMVdiffVal < 2) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (Cs < 1885) {
if (Rs < 3142) {
return 8;
}
else if (Rs >= 3142) {
return 4;
}
}
else if (Cs >= 1885) {
return 8;
}
}
}
else if (diffMVdiffVal >= 2) {
if (refDCval < 104) {
if (Rs < 2476) {
return 2;
}
else if (Rs >= 2476) {
return 8;
}
}
else if (refDCval >= 104) {
if (ssDCval < 116) {
return 8;
}
else if (ssDCval >= 116) {
return 4;
}
}
}
}
}
else if (ssDCval >= 117) {
if (ssDCval < 200) {
if (refDCval < 119) {
if (Cs < 4847) {
if (Rs < 2807) {
if (Cs < 2968) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (RsDiff < 2) {
if (Cs < 2939) {
return 4;
}
else if (Cs >= 2939) {
return 8;
}
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (Cs >= 2968) {
if (MVDiff < 1) {
if (Rs < 2767) {
return 8;
}
else if (Rs >= 2767) {
return 2;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
}
else if (Rs >= 2807) {
return 8;
}
}
else if (Cs >= 4847) {
return 1;
}
}
else if (refDCval >= 119) {
if (posBalance < 55) {
if (Cs < 1680) {
if (Rs < 4372) {
return 8;
}
else if (Rs >= 4372) {
return 2;
}
}
else if (Cs >= 1680) {
if (ssDCval < 136) {
if (refDCval < 123) {
return 8;
}
else if (refDCval >= 123) {
if (Rs < 3965) {
if (Cs < 3862) {
if (tcor < 100) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 1;
}
}
else if (tcor >= 100) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (Cs >= 3862) {
return 1;
}
}
else if (Rs >= 3965) {
return 8;
}
}
}
else if (ssDCval >= 136) {
if (TSC < 6) {
if (Cs < 1872) {
if (Rs < 4490) {
return 8;
}
else if (Rs >= 4490) {
return 2;
}
}
else if (Cs >= 1872) {
return 8;
}
}
else if (TSC >= 6) {
if (Rs < 4890) {
return 8;
}
else if (Rs >= 4890) {
return 2;
}
}
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 132) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
if (Rs < 5735) {
return 8;
}
else if (Rs >= 5735) {
return 4;
}
}
}
else if (ssDCval >= 132) {
return 8;
}
}
}
}
else if (ssDCval >= 200) {
if (Cs < 4080) {
if (Rs < 1150) {
return 1;
}
else if (Rs >= 1150) {
return 8;
}
}
else if (Cs >= 4080) {
return 4;
}
}
}
}
else if (negBalance >= 37) {
if (ssDCval < 137) {
if (RsDiff < 2) {
if (ssDCval < 117) {
if (Rs < 2017) {
if (ssDCval < 106) {
return 8;
}
else if (ssDCval >= 106) {
return 2;
}
}
else if (Rs >= 2017) {
return 8;
}
}
else if (ssDCval >= 117) {
if (tcor < 97) {
return 4;
}
else if (tcor >= 97) {
return 8;
}
}
}
else if (RsDiff >= 2) {
if (refDCval < 105) {
if (Cs < 2908) {
if (ssDCval < 104) {
if (Rs < 3642) {
if (Rs < 1762) {
return 2;
}
else if (Rs >= 1762) {
return 8;
}
}
else if (Rs >= 3642) {
return 1;
}
}
else if (ssDCval >= 104) {
if (Cs < 2461) {
return 2;
}
else if (Cs >= 2461) {
return 1;
}
}
}
else if (Cs >= 2908) {
return 8;
}
}
else if (refDCval >= 105) {
if (posBalance < 63) {
if (SC < 9937) {
if (AFD < 13) {
if (diffRsCsdiff < -44) {
if (RsDiff < 8) {
if (mu_mv_mag_sq < 114) {
return 1;
}
else if (mu_mv_mag_sq >= 114) {
if (ssDCval < 128) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
if (Cs < 2002) {
return 4;
}
else if (Cs >= 2002) {
return 8;
}
}
}
else if (ssDCval >= 128) {
return 8;
}
}
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (diffRsCsdiff >= -44) {
if (diffAFD < 2) {
if (mcTcor < 99) {
if (negBalance < 56) {
if (AFD < 11) {
if (diffAFD < -3) {
if (ssDCval < 119) {
return 8;
}
else if (ssDCval >= 119) {
return 1;
}
}
else if (diffAFD >= -3) {
if (Cs < 1650) {
if (Rs < 1911) {
return 1;
}
else if (Rs >= 1911) {
return 8;
}
}
else if (Cs >= 1650) {
return 8;
}
}
}
else if (AFD >= 11) {
if (mu_mv_mag_sq < 120) {
return 8;
}
else if (mu_mv_mag_sq >= 120) {
return 2;
}
}
}
else if (negBalance >= 56) {
if (Rs < 4305) {
return 2;
}
else if (Rs >= 4305) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 1759) {
return 2;
}
else if (Rs >= 1759) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 4870) {
if (Cs < 1926) {
if (Rs < 2801) {
return 8;
}
else if (Rs >= 2801) {
return 4;
}
}
else if (Cs >= 1926) {
return 8;
}
}
else if (Rs >= 4870) {
if (Rs < 5812) {
return 2;
}
else if (Rs >= 5812) {
return 4;
}
}
}
}
}
else if (AFD >= 13) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
if (Rs < 2055) {
return 4;
}
else if (Rs >= 2055) {
return 1;
}
}
}
}
else if (SC >= 9937) {
return 8;
}
}
else if (posBalance >= 63) {
if (CsDiff < 5) {
if (Rs < 2366) {
return 8;
}
else if (Rs >= 2366) {
return 1;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
}
else if (ssDCval >= 137) {
if (mcTcor < 96) {
if (ssDCval < 146) {
return 1;
}
else if (ssDCval >= 146) {
return 8;
}
}
else if (mcTcor >= 96) {
if (refDCval < 144) {
if (Rs < 4567) {
return 8;
}
else if (Rs >= 4567) {
if (mu_mv_mag_sq < 122) {
if (Rs < 5663) {
return 8;
}
else if (Rs >= 5663) {
return 2;
}
}
else if (mu_mv_mag_sq >= 122) {
return 1;
}
}
}
else if (refDCval >= 144) {
return 8;
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect15(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1308) {
if (Rs < 2587) {
if (Cs < 535) {
if (negBalance < 14) {
if (SC < 796) {
if (Cs < 201) {
if (refDCval < 48) {
if (AFD < 1) {
if (refDCval < 22) {
if (ssDCval < 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 10) {
if (MVDiff < 1076) {
if (Rs < 392) {
if (negBalance < 7) {
return 8;
}
else if (negBalance >= 7) {
if (posBalance < 11) {
return 8;
}
else if (posBalance >= 11) {
if (ssDCval < 18) {
return 8;
}
else if (ssDCval >= 18) {
return 1;
}
}
}
}
else if (Rs >= 392) {
if (Rs < 432) {
return 4;
}
else if (Rs >= 432) {
return 8;
}
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
}
else if (refDCval >= 22) {
if (ssDCval < 27) {
if (posBalance < 59) {
if (MVDiff < 3) {
if (Rs < 38) {
return 4;
}
else if (Rs >= 38) {
return 8;
}
}
else if (MVDiff >= 3) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (Rs < 73) {
return 4;
}
else if (Rs >= 73) {
if (Rs < 115) {
return 1;
}
else if (Rs >= 115) {
return 4;
}
}
}
}
}
else if (posBalance >= 59) {
return 8;
}
}
else if (ssDCval >= 27) {
if (Rs < 233) {
if (negBalance < 7) {
if (Rs < 8) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (Rs >= 8) {
if (Cs < 186) {
return 8;
}
else if (Cs >= 186) {
return 4;
}
}
}
else if (negBalance >= 7) {
if (ssDCval < 38) {
if (SC < 307) {
if (Rs < 60) {
return 4;
}
else if (Rs >= 60) {
return 8;
}
}
else if (SC >= 307) {
return 4;
}
}
else if (ssDCval >= 38) {
if (tcor < 98) {
if (Rs < 95) {
return 8;
}
else if (Rs >= 95) {
return 1;
}
}
else if (tcor >= 98) {
return 8;
}
}
}
}
else if (Rs >= 233) {
if (Rs < 297) {
return 4;
}
else if (Rs >= 297) {
return 8;
}
}
}
}
}
else if (AFD >= 1) {
if (posBalance < 61) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
if (refDCval < 25) {
if (MVDiff < 4) {
if (Rs < 74) {
return 8;
}
else if (Rs >= 74) {
return 1;
}
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (refDCval >= 25) {
if (refDCval < 31) {
if (posBalance < 43) {
if (MVDiff < 19) {
if (posBalance < 27) {
if (Rs < 188) {
return 4;
}
else if (Rs >= 188) {
return 8;
}
}
else if (posBalance >= 27) {
return 1;
}
}
else if (MVDiff >= 19) {
return 4;
}
}
else if (posBalance >= 43) {
if (MVDiff < 7) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (refDCval >= 31) {
if (SC < 218) {
if (posBalance < 23) {
return 1;
}
else if (posBalance >= 23) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Cs < 58) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
return 8;
}
}
else if (Cs >= 58) {
return 8;
}
}
}
}
else if (SC >= 218) {
if (mu_mv_mag_sq < 121) {
if (refDCval < 40) {
if (posBalance < 38) {
if (Cs < 129) {
return 8;
}
else if (Cs >= 129) {
if (TSC < 1) {
return 8;
}
else if (TSC >= 1) {
return 4;
}
}
}
else if (posBalance >= 38) {
return 1;
}
}
else if (refDCval >= 40) {
return 4;
}
}
else if (mu_mv_mag_sq >= 121) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 61) {
if (RsCsDiff < 1) {
if (ssDCval < 34) {
if (mu_mv_mag_sq < 29) {
if (Rs < 142) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
if (Rs < 57) {
return 8;
}
else if (Rs >= 57) {
return 2;
}
}
}
else if (Rs >= 142) {
if (Rs < 282) {
return 1;
}
else if (Rs >= 282) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (ssDCval < 20) {
if (Rs < 172) {
return 4;
}
else if (Rs >= 172) {
return 1;
}
}
else if (ssDCval >= 20) {
if (mu_mv_mag_sq < 102) {
if (negBalance < 1) {
if (Rs < 57) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (Rs >= 57) {
return 4;
}
}
else if (negBalance >= 1) {
if (diffMVdiffVal < -5) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (MVDiff < 1) {
if (Rs < 3) {
return 4;
}
else if (Rs >= 3) {
return 8;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
}
}
}
else if (ssDCval >= 34) {
if (Rs < 216) {
if (Rs < 9) {
return 1;
}
else if (Rs >= 9) {
return 8;
}
}
else if (Rs >= 216) {
if (negBalance < 3) {
return 4;
}
else if (negBalance >= 3) {
return 8;
}
}
}
}
else if (RsCsDiff >= 1) {
if (refDCval < 37) {
if (posBalance < 64) {
if (Rs < 104) {
return 4;
}
else if (Rs >= 104) {
return 2;
}
}
else if (posBalance >= 64) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
if (RsDiff < 4) {
if (Rs < 245) {
return 8;
}
else if (Rs >= 245) {
return 2;
}
}
else if (RsDiff >= 4) {
return 1;
}
}
}
}
else if (refDCval >= 37) {
return 8;
}
}
}
}
}
else if (refDCval >= 48) {
if (ssDCval < 82) {
if (SC < 61) {
return 2;
}
else if (SC >= 61) {
if (ssDCval < 61) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (mcTcor < 99) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
if (Rs < 90) {
return 4;
}
else if (Rs >= 90) {
return 1;
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 54) {
if (Rs < 428) {
return 8;
}
else if (Rs >= 428) {
return 4;
}
}
else if (ssDCval >= 54) {
if (Rs < 92) {
return 8;
}
else if (Rs >= 92) {
return 4;
}
}
}
}
}
else if (ssDCval >= 61) {
if (Cs < 119) {
if (Rs < 154) {
if (Cs < 24) {
return 2;
}
else if (Cs >= 24) {
return 4;
}
}
else if (Rs >= 154) {
return 8;
}
}
else if (Cs >= 119) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 87) {
if (Cs < 58) {
return 4;
}
else if (Cs >= 58) {
return 1;
}
}
else if (ssDCval >= 87) {
if (SC < 311) {
if (Cs < 40) {
if (negBalance < 2) {
if (MVDiff < 293) {
return 2;
}
else if (MVDiff >= 293) {
if (Rs < 115) {
return 1;
}
else if (Rs >= 115) {
return 4;
}
}
}
else if (negBalance >= 2) {
return 1;
}
}
else if (Cs >= 40) {
if (AFD < 6) {
if (posBalance < 32) {
return 8;
}
else if (posBalance >= 32) {
return 4;
}
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (SC >= 311) {
if (refDCval < 102) {
return 2;
}
else if (refDCval >= 102) {
if (Cs < 170) {
return 8;
}
else if (Cs >= 170) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 201) {
if (mu_mv_mag_sq < 111) {
if (AFD < 2) {
if (tcor < 92) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (tcor >= 92) {
if (mu_mv_mag_sq < 72) {
return 4;
}
else if (mu_mv_mag_sq >= 72) {
if (RsDiff < 1) {
if (ssDCval < 82) {
if (Rs < 158) {
if (Rs < 132) {
return 1;
}
else if (Rs >= 132) {
return 2;
}
}
else if (Rs >= 158) {
if (ssDCval < 67) {
return 4;
}
else if (ssDCval >= 67) {
if (Rs < 292) {
return 1;
}
else if (Rs >= 292) {
return 8;
}
}
}
}
else if (ssDCval >= 82) {
return 4;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
else if (AFD >= 2) {
if (RsDiff < 1) {
if (gchDC < 2) {
if (negBalance < 13) {
return 4;
}
else if (negBalance >= 13) {
if (Cs < 294) {
return 1;
}
else if (Cs >= 294) {
return 4;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (RsDiff >= 1) {
if (ssDCval < 21) {
return 8;
}
else if (ssDCval >= 21) {
if (mu_mv_mag_sq < 41) {
if (refDCval < 46) {
if (SC < 592) {
return 1;
}
else if (SC >= 592) {
return 4;
}
}
else if (refDCval >= 46) {
if (Rs < 218) {
return 1;
}
else if (Rs >= 218) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (Cs < 250) {
return 1;
}
else if (Cs >= 250) {
if (Rs < 264) {
if (Rs < 224) {
return 2;
}
else if (Rs >= 224) {
return 1;
}
}
else if (Rs >= 264) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 239) {
if (ssDCval < 64) {
if (SC < 376) {
return 4;
}
else if (SC >= 376) {
return 8;
}
}
else if (ssDCval >= 64) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (Cs >= 239) {
if (Rs < 353) {
if (SC < 491) {
if (posBalance < 32) {
if (SC < 422) {
return 4;
}
else if (SC >= 422) {
return 1;
}
}
else if (posBalance >= 32) {
if (Rs < 162) {
return 4;
}
else if (Rs >= 162) {
return 8;
}
}
}
else if (SC >= 491) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 88) {
if (Rs < 308) {
if (Cs < 326) {
if (SC < 562) {
if (Rs < 241) {
return 4;
}
else if (Rs >= 241) {
return 1;
}
}
else if (SC >= 562) {
return 4;
}
}
else if (Cs >= 326) {
return 1;
}
}
else if (Rs >= 308) {
if (Cs < 268) {
return 1;
}
else if (Cs >= 268) {
return 4;
}
}
}
else if (ssDCval >= 88) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 321) {
if (negBalance < 3) {
if (Cs < 287) {
return 4;
}
else if (Cs >= 287) {
if (Cs < 311) {
return 8;
}
else if (Cs >= 311) {
return 4;
}
}
}
else if (negBalance >= 3) {
return 4;
}
}
else if (Cs >= 321) {
if (Rs < 171) {
return 1;
}
else if (Rs >= 171) {
return 8;
}
}
}
}
}
else if (Rs >= 353) {
if (ssDCval < 84) {
if (Rs < 378) {
return 4;
}
else if (Rs >= 378) {
if (SC < 634) {
if (Rs < 384) {
return 8;
}
else if (Rs >= 384) {
return 4;
}
}
else if (SC >= 634) {
return 8;
}
}
}
else if (ssDCval >= 84) {
if (MVDiff < 57) {
if (Cs < 293) {
return 4;
}
else if (Cs >= 293) {
return 2;
}
}
else if (MVDiff >= 57) {
if (Rs < 446) {
return 4;
}
else if (Rs >= 446) {
return 1;
}
}
}
}
}
}
}
}
else if (SC >= 796) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 128) {
if (Cs < 267) {
if (ssDCval < 39) {
return 4;
}
else if (ssDCval >= 39) {
if (MVDiff < 7) {
if (negBalance < 12) {
if (posBalance < 120) {
return 8;
}
else if (posBalance >= 120) {
return 4;
}
}
else if (negBalance >= 12) {
if (Cs < 203) {
return 8;
}
else if (Cs >= 203) {
if (Rs < 907) {
return 8;
}
else if (Rs >= 907) {
if (Rs < 1511) {
return 2;
}
else if (Rs >= 1511) {
return 1;
}
}
}
}
}
else if (MVDiff >= 7) {
if (Rs < 1741) {
return 1;
}
else if (Rs >= 1741) {
return 2;
}
}
}
}
else if (Cs >= 267) {
if (Rs < 698) {
if (RsCsDiff < 1) {
if (Rs < 490) {
if (Cs < 485) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
return 1;
}
}
else if (Cs >= 485) {
if (Rs < 482) {
return 1;
}
else if (Rs >= 482) {
return 8;
}
}
}
else if (Rs >= 490) {
if (Rs < 605) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (Cs < 397) {
if (ssDCval < 152) {
return 8;
}
else if (ssDCval >= 152) {
return 1;
}
}
else if (Cs >= 397) {
if (Rs < 491) {
return 1;
}
else if (Rs >= 491) {
return 8;
}
}
}
}
else if (Rs >= 605) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (Cs < 420) {
return 4;
}
else if (Cs >= 420) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (Rs < 513) {
return 1;
}
else if (Rs >= 513) {
if (ssDCval < 62) {
if (Rs < 621) {
return 1;
}
else if (Rs >= 621) {
return 8;
}
}
else if (ssDCval >= 62) {
return 8;
}
}
}
}
else if (Rs >= 698) {
if (ssDCval < 67) {
if (refDCval < 60) {
if (Cs < 485) {
if (SC < 1180) {
if (Cs < 430) {
return 8;
}
else if (Cs >= 430) {
return 2;
}
}
else if (SC >= 1180) {
if (mu_mv_mag_sq < 127) {
if (ssDCval < 57) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
return 1;
}
}
else if (ssDCval >= 57) {
if (Rs < 1183) {
return 4;
}
else if (Rs >= 1183) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 127) {
return 8;
}
}
}
else if (Cs >= 485) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (Rs < 1046) {
return 2;
}
else if (Rs >= 1046) {
return 8;
}
}
}
}
else if (refDCval >= 60) {
if (Cs < 448) {
return 1;
}
else if (Cs >= 448) {
if (Rs < 1457) {
return 2;
}
else if (Rs >= 1457) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (mu_mv_mag_sq < 94) {
if (Rs < 2043) {
return 1;
}
else if (Rs >= 2043) {
return 2;
}
}
else if (mu_mv_mag_sq >= 94) {
if (negBalance < 13) {
return 8;
}
else if (negBalance >= 13) {
if (Rs < 1061) {
return 8;
}
else if (Rs >= 1061) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 136) {
if (refDCval < 60) {
return 8;
}
else if (refDCval >= 60) {
if (ssDCval < 88) {
if (Cs < 355) {
if (Cs < 250) {
return 2;
}
else if (Cs >= 250) {
return 1;
}
}
else if (Cs >= 355) {
if (Rs < 1186) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 521) {
return 8;
}
else if (Rs >= 521) {
return 4;
}
}
}
else if (Rs >= 1186) {
return 2;
}
}
}
else if (ssDCval >= 88) {
return 8;
}
}
}
else if (ssDCval >= 136) {
return 1;
}
}
}
else if (MVDiff >= 8) {
if (SC < 937) {
if (SC < 861) {
if (posBalance < 19) {
return 2;
}
else if (posBalance >= 19) {
if (CsDiff < 2) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
if (tcor < 90) {
if (Rs < 553) {
return 1;
}
else if (Rs >= 553) {
return 4;
}
}
else if (tcor >= 90) {
return 4;
}
}
}
else if (CsDiff >= 2) {
if (Rs < 371) {
return 8;
}
else if (Rs >= 371) {
return 2;
}
}
}
}
else if (SC >= 861) {
if (SC < 897) {
if (negBalance < 8) {
return 8;
}
else if (negBalance >= 8) {
if (Rs < 445) {
return 4;
}
else if (Rs >= 445) {
return 1;
}
}
}
else if (SC >= 897) {
if (MVDiff < 46) {
if (ssDCval < 141) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
return 4;
}
}
else if (ssDCval >= 141) {
return 1;
}
}
else if (MVDiff >= 46) {
return 8;
}
}
}
}
else if (SC >= 937) {
if (diffAFD < -8) {
if (ssDCval < 131) {
if (ssDCval < 125) {
if (MVDiff < 103) {
return 8;
}
else if (MVDiff >= 103) {
if (Rs < 867) {
return 4;
}
else if (Rs >= 867) {
return 2;
}
}
}
else if (ssDCval >= 125) {
return 2;
}
}
else if (ssDCval >= 131) {
if (Rs < 1349) {
return 8;
}
else if (Rs >= 1349) {
return 1;
}
}
}
else if (diffAFD >= -8) {
if (posBalance < 94) {
if (refDCval < 156) {
if (mu_mv_mag_sq < 25) {
return 1;
}
else if (mu_mv_mag_sq >= 25) {
if (diffTSC < -3) {
return 1;
}
else if (diffTSC >= -3) {
if (negBalance < 11) {
if (AFD < 9) {
if (Cs < 227) {
return 1;
}
else if (Cs >= 227) {
if (tcor < 101) {
return 8;
}
else if (tcor >= 101) {
return 1;
}
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (negBalance >= 11) {
if (AFD < 7) {
if (posBalance < 24) {
return 8;
}
else if (posBalance >= 24) {
return 1;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
}
}
else if (refDCval >= 156) {
return 8;
}
}
else if (posBalance >= 94) {
if (RsCsDiff < 6) {
return 8;
}
else if (RsCsDiff >= 6) {
if (Cs < 222) {
return 2;
}
else if (Cs >= 222) {
if (mu_mv_mag_sq < 47) {
if (Rs < 809) {
return 8;
}
else if (Rs >= 809) {
return 1;
}
}
else if (mu_mv_mag_sq >= 47) {
return 4;
}
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 34) {
if (mcTcor < 83) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (ssDCval < 23) {
if (tcor < 55) {
if (Rs < 1) {
return 8;
}
else if (Rs >= 1) {
return 1;
}
}
else if (tcor >= 55) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
if (ssDCval < 20) {
return 1;
}
else if (ssDCval >= 20) {
return 4;
}
}
}
}
else if (ssDCval >= 23) {
if (Cs < 385) {
if (tcor < 47) {
if (negBalance < 41) {
if (diffMVdiffVal < -5) {
if (ssDCval < 36) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
else if (ssDCval >= 36) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
return 1;
}
}
else if (negBalance >= 41) {
if (tcor < 18) {
if (Rs < 192) {
return 4;
}
else if (Rs >= 192) {
return 8;
}
}
else if (tcor >= 18) {
if (Rs < 963) {
return 8;
}
else if (Rs >= 963) {
return 1;
}
}
}
}
else if (tcor >= 47) {
if (AFD < 6) {
if (Cs < 208) {
if (MVDiff < 35) {
return 1;
}
else if (MVDiff >= 35) {
if (Rs < 37) {
return 4;
}
else if (Rs >= 37) {
return 2;
}
}
}
else if (Cs >= 208) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
return 2;
}
}
}
else if (AFD >= 6) {
if (Rs < 172) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 8;
}
}
else if (Rs >= 172) {
return 1;
}
}
}
}
else if (Cs >= 385) {
if (RsDiff < 4) {
if (tcor < 59) {
return 2;
}
else if (tcor >= 59) {
if (Rs < 705) {
return 8;
}
else if (Rs >= 705) {
return 1;
}
}
}
else if (RsDiff >= 4) {
if (Rs < 1164) {
return 8;
}
else if (Rs >= 1164) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 83) {
if (posBalance < 27) {
if (posBalance < 13) {
if (Rs < 155) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (mu_mv_mag_sq < 38) {
if (diffTSC < -1) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
else if (diffTSC >= -1) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 38) {
return 4;
}
}
}
else if (Rs >= 155) {
if (refDCval < 30) {
return 4;
}
else if (refDCval >= 30) {
return 1;
}
}
}
else if (posBalance >= 13) {
if (MVDiff < 5) {
if (negBalance < 22) {
if (tcor < 102) {
return 4;
}
else if (tcor >= 102) {
return 8;
}
}
else if (negBalance >= 22) {
return 1;
}
}
else if (MVDiff >= 5) {
if (ssDCval < 31) {
if (mu_mv_mag_sq < 78) {
return 1;
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 213) {
if (Rs < 55) {
return 4;
}
else if (Rs >= 55) {
return 1;
}
}
else if (Rs >= 213) {
return 4;
}
}
}
else if (ssDCval >= 31) {
if (mu_mv_mag_sq < 90) {
if (Rs < 267) {
return 8;
}
else if (Rs >= 267) {
return 2;
}
}
else if (mu_mv_mag_sq >= 90) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 27) {
if (Rs < 72) {
if (posBalance < 69) {
return 1;
}
else if (posBalance >= 69) {
if (TSC < 2) {
return 2;
}
else if (TSC >= 2) {
return 4;
}
}
}
else if (Rs >= 72) {
if (posBalance < 42) {
if (refDCval < 30) {
if (MVDiff < 22) {
if (Cs < 23) {
if (Rs < 93) {
return 1;
}
else if (Rs >= 93) {
return 4;
}
}
else if (Cs >= 23) {
if (Cs < 376) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
if (diffRsCsdiff < 1) {
if (diffMVdiffVal < -8) {
return 4;
}
else if (diffMVdiffVal >= -8) {
return 1;
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 203) {
return 4;
}
else if (Rs >= 203) {
return 1;
}
}
}
}
else if (Cs >= 376) {
if (Rs < 411) {
return 1;
}
else if (Rs >= 411) {
return 4;
}
}
}
}
else if (MVDiff >= 22) {
if (negBalance < 15) {
if (Rs < 196) {
return 1;
}
else if (Rs >= 196) {
return 4;
}
}
else if (negBalance >= 15) {
if (mu_mv_mag_sq < 40) {
if (Rs < 354) {
if (Rs < 98) {
return 8;
}
else if (Rs >= 98) {
return 4;
}
}
else if (Rs >= 354) {
return 1;
}
}
else if (mu_mv_mag_sq >= 40) {
return 1;
}
}
}
}
else if (refDCval >= 30) {
if (Cs < 253) {
if (SC < 525) {
if (mcTcor < 96) {
if (AFD < 4) {
if (tcor < 87) {
return 1;
}
else if (tcor >= 87) {
if (posBalance < 41) {
if (diffMVdiffVal < -14) {
return 1;
}
else if (diffMVdiffVal >= -14) {
return 8;
}
}
else if (posBalance >= 41) {
return 1;
}
}
}
else if (AFD >= 4) {
if (MVDiff < 16) {
if (TSC < 2) {
if (Rs < 243) {
return 8;
}
else if (Rs >= 243) {
return 2;
}
}
else if (TSC >= 2) {
return 4;
}
}
else if (MVDiff >= 16) {
if (Cs < 65) {
if (Rs < 133) {
return 1;
}
else if (Rs >= 133) {
return 4;
}
}
else if (Cs >= 65) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 23) {
if (Rs < 104) {
return 8;
}
else if (Rs >= 104) {
return 4;
}
}
else if (mu_mv_mag_sq >= 23) {
if (Rs < 108) {
if (Cs < 93) {
return 1;
}
else if (Cs >= 93) {
if (SC < 191) {
return 4;
}
else if (SC >= 191) {
return 8;
}
}
}
else if (Rs >= 108) {
if (posBalance < 30) {
if (Rs < 265) {
return 1;
}
else if (Rs >= 265) {
return 2;
}
}
else if (posBalance >= 30) {
return 1;
}
}
}
}
}
else if (SC >= 525) {
if (mcTcor < 92) {
if (Rs < 360) {
return 8;
}
else if (Rs >= 360) {
if (Rs < 486) {
return 4;
}
else if (Rs >= 486) {
return 1;
}
}
}
else if (mcTcor >= 92) {
return 4;
}
}
}
else if (Cs >= 253) {
if (TSC < 2) {
if (posBalance < 38) {
if (Rs < 251) {
return 1;
}
else if (Rs >= 251) {
return 4;
}
}
else if (posBalance >= 38) {
return 1;
}
}
else if (TSC >= 2) {
return 1;
}
}
}
}
else if (posBalance >= 42) {
if (MVDiff < 58) {
if (SC < 369) {
if (tcor < 67) {
if (negBalance < 40) {
if (ssDCval < 29) {
return 8;
}
else if (ssDCval >= 29) {
if (Cs < 174) {
return 1;
}
else if (Cs >= 174) {
return 4;
}
}
}
else if (negBalance >= 40) {
return 8;
}
}
else if (tcor >= 67) {
if (mcTcor < 88) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Cs < 186) {
return 1;
}
else if (Cs >= 186) {
return 2;
}
}
}
else if (mcTcor >= 88) {
if (tcor < 72) {
if (Cs < 64) {
return 1;
}
else if (Cs >= 64) {
if (Rs < 172) {
return 8;
}
else if (Rs >= 172) {
return 2;
}
}
}
else if (tcor >= 72) {
return 1;
}
}
}
}
else if (SC >= 369) {
if (negBalance < 26) {
if (mcTcor < 88) {
if (Cs < 239) {
return 8;
}
else if (Cs >= 239) {
return 1;
}
}
else if (mcTcor >= 88) {
return 1;
}
}
else if (negBalance >= 26) {
return 1;
}
}
}
else if (MVDiff >= 58) {
if (negBalance < 36) {
if (MVDiff < 423) {
return 8;
}
else if (MVDiff >= 423) {
return 2;
}
}
else if (negBalance >= 36) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 34) {
if (Rs < 300) {
if (ssDCval < 51) {
if (mu_mv_mag_sq < 52) {
if (mcTcor < 88) {
if (MVDiff < 68) {
if (refDCval < 42) {
if (Cs < 72) {
if (Cs < 25) {
return 2;
}
else if (Cs >= 25) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 4;
}
}
}
else if (Cs >= 72) {
if (CsDiff < 2) {
if (posBalance < 54) {
if (Cs < 191) {
return 8;
}
else if (Cs >= 191) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
if (Rs < 234) {
return 2;
}
else if (Rs >= 234) {
return 8;
}
}
}
}
else if (posBalance >= 54) {
return 8;
}
}
else if (CsDiff >= 2) {
if (TSC < 5) {
if (Cs < 221) {
if (Rs < 179) {
return 4;
}
else if (Rs >= 179) {
return 1;
}
}
else if (Cs >= 221) {
return 2;
}
}
else if (TSC >= 5) {
if (ssDCval < 36) {
if (Rs < 160) {
return 8;
}
else if (Rs >= 160) {
return 4;
}
}
else if (ssDCval >= 36) {
return 8;
}
}
}
}
}
else if (refDCval >= 42) {
if (MVDiff < 33) {
if (Cs < 286) {
return 4;
}
else if (Cs >= 286) {
return 8;
}
}
else if (MVDiff >= 33) {
if (Rs < 138) {
return 8;
}
else if (Rs >= 138) {
return 1;
}
}
}
}
else if (MVDiff >= 68) {
if (posBalance < 47) {
if (tcor < 103) {
if (diffRsCsdiff < -2) {
if (ssDCval < 38) {
return 1;
}
else if (ssDCval >= 38) {
if (Rs < 134) {
return 8;
}
else if (Rs >= 134) {
return 1;
}
}
}
else if (diffRsCsdiff >= -2) {
if (ssDCval < 41) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 1;
}
}
else if (ssDCval >= 41) {
return 1;
}
}
}
else if (tcor >= 103) {
if (Rs < 80) {
return 8;
}
else if (Rs >= 80) {
return 2;
}
}
}
else if (posBalance >= 47) {
if (negBalance < 39) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
else if (negBalance >= 39) {
if (RsDiff < 3) {
if (CsDiff < 2) {
if (diffRsCsdiff < 2) {
if (Cs < 37) {
if (MVDiff < 348) {
return 8;
}
else if (MVDiff >= 348) {
return 4;
}
}
else if (Cs >= 37) {
return 4;
}
}
else if (diffRsCsdiff >= 2) {
return 1;
}
}
else if (CsDiff >= 2) {
if (Rs < 223) {
if (MVDiff < 120) {
return 8;
}
else if (MVDiff >= 120) {
return 1;
}
}
else if (Rs >= 223) {
return 4;
}
}
}
else if (RsDiff >= 3) {
return 1;
}
}
}
}
}
else if (mcTcor >= 88) {
if (negBalance < 75) {
if (mcTcor < 97) {
if (TSC < 4) {
if (AFD < 9) {
if (posBalance < 49) {
if (Rs < 130) {
if (SC < 130) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
if (Rs < 46) {
return 8;
}
else if (Rs >= 46) {
return 2;
}
}
}
else if (SC >= 130) {
if (MVDiff < 25) {
if (Rs < 69) {
return 1;
}
else if (Rs >= 69) {
return 4;
}
}
else if (MVDiff >= 25) {
return 1;
}
}
}
else if (Rs >= 130) {
if (negBalance < 44) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
else if (negBalance >= 44) {
if (Rs < 252) {
return 2;
}
else if (Rs >= 252) {
return 4;
}
}
}
}
else if (posBalance >= 49) {
if (AFD < 4) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
return 8;
}
}
else if (AFD >= 4) {
if (diffTSC < -1) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (negBalance < 36) {
if (Rs < 256) {
return 2;
}
else if (Rs >= 256) {
return 1;
}
}
else if (negBalance >= 36) {
return 1;
}
}
}
else if (diffTSC >= -1) {
if (MVDiff < 5) {
if (ssDCval < 45) {
return 1;
}
else if (ssDCval >= 45) {
return 4;
}
}
else if (MVDiff >= 5) {
if (diffMVdiffVal < 9) {
if (negBalance < 49) {
if (Rs < 48) {
if (Cs < 53) {
return 2;
}
else if (Cs >= 53) {
return 1;
}
}
else if (Rs >= 48) {
if (diffAFD < 2) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (negBalance >= 49) {
if (Rs < 142) {
return 8;
}
else if (Rs >= 142) {
return 1;
}
}
}
else if (diffMVdiffVal >= 9) {
if (RsDiff < 2) {
if (Rs < 81) {
return 2;
}
else if (Rs >= 81) {
return 4;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
}
}
}
else if (AFD >= 9) {
if (posBalance < 64) {
if (mu_mv_mag_sq < 23) {
if (diffAFD < 0) {
return 1;
}
else if (diffAFD >= 0) {
return 4;
}
}
else if (mu_mv_mag_sq >= 23) {
return 1;
}
}
else if (posBalance >= 64) {
if (Rs < 73) {
return 2;
}
else if (Rs >= 73) {
return 4;
}
}
}
}
else if (TSC >= 4) {
if (ssDCval < 43) {
if (diffMVdiffVal < 9) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (tcor < 51) {
return 1;
}
else if (tcor >= 51) {
if (SC < 342) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 8;
}
}
else if (SC >= 342) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 9) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (Rs < 207) {
return 1;
}
else if (Rs >= 207) {
return 4;
}
}
}
}
else if (ssDCval >= 43) {
if (negBalance < 41) {
return 2;
}
else if (negBalance >= 41) {
if (Rs < 285) {
if (Rs < 88) {
if (Rs < 67) {
return 2;
}
else if (Rs >= 67) {
return 4;
}
}
else if (Rs >= 88) {
if (diffAFD < 3) {
if (mcTcor < 90) {
if (Rs < 154) {
return 1;
}
else if (Rs >= 154) {
return 2;
}
}
else if (mcTcor >= 90) {
return 1;
}
}
else if (diffAFD >= 3) {
return 2;
}
}
}
else if (Rs >= 285) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (AFD < 5) {
if (Cs < 5) {
return 2;
}
else if (Cs >= 5) {
if (diffAFD < -2) {
if (Rs < 136) {
return 4;
}
else if (Rs >= 136) {
return 1;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
}
else if (AFD >= 5) {
if (negBalance < 38) {
if (posBalance < 52) {
if (Rs < 43) {
return 1;
}
else if (Rs >= 43) {
if (mu_mv_mag_sq < 35) {
return 4;
}
else if (mu_mv_mag_sq >= 35) {
if (Rs < 225) {
return 1;
}
else if (Rs >= 225) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
else if (posBalance >= 52) {
if (Rs < 165) {
if (Rs < 45) {
return 1;
}
else if (Rs >= 45) {
return 2;
}
}
else if (Rs >= 165) {
return 1;
}
}
}
else if (negBalance >= 38) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
if (Rs < 122) {
return 2;
}
else if (Rs >= 122) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 75) {
if (mcTcor < 95) {
if (tcor < 130) {
return 2;
}
else if (tcor >= 130) {
if (Rs < 6) {
return 4;
}
else if (Rs >= 6) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (posBalance < 1) {
return 4;
}
else if (posBalance >= 1) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (SC < 525) {
if (negBalance < 31) {
if (mu_mv_mag_sq < 102) {
if (refDCval < 47) {
if (posBalance < 29) {
if (negBalance < 18) {
if (posBalance < 21) {
return 1;
}
else if (posBalance >= 21) {
return 4;
}
}
else if (negBalance >= 18) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (Cs < 70) {
return 2;
}
else if (Cs >= 70) {
return 1;
}
}
}
}
else if (posBalance >= 29) {
if (mcTcor < 97) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (tcor < 87) {
if (mcTcor < 96) {
if (ssDCval < 35) {
if (Rs < 153) {
return 8;
}
else if (Rs >= 153) {
return 1;
}
}
else if (ssDCval >= 35) {
return 1;
}
}
else if (mcTcor >= 96) {
return 8;
}
}
else if (tcor >= 87) {
if (refDCval < 43) {
if (Rs < 57) {
return 1;
}
else if (Rs >= 57) {
return 8;
}
}
else if (refDCval >= 43) {
return 1;
}
}
}
}
else if (mcTcor >= 97) {
return 1;
}
}
}
else if (refDCval >= 47) {
if (Rs < 167) {
if (Rs < 14) {
return 2;
}
else if (Rs >= 14) {
return 8;
}
}
else if (Rs >= 167) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 214) {
if (refDCval < 40) {
if (Cs < 188) {
if (SC < 291) {
return 4;
}
else if (SC >= 291) {
return 1;
}
}
else if (Cs >= 188) {
if (Cs < 303) {
return 8;
}
else if (Cs >= 303) {
return 1;
}
}
}
else if (refDCval >= 40) {
if (SC < 190) {
return 8;
}
else if (SC >= 190) {
if (ssDCval < 43) {
return 1;
}
else if (ssDCval >= 43) {
return 8;
}
}
}
}
else if (Rs >= 214) {
return 4;
}
}
}
else if (negBalance >= 31) {
if (Cs < 93) {
if (SC < 149) {
if (Cs < 34) {
if (mu_mv_mag_sq < 77) {
if (Cs < 16) {
return 8;
}
else if (Cs >= 16) {
return 2;
}
}
else if (mu_mv_mag_sq >= 77) {
if (Rs < 4) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
else if (Rs >= 4) {
if (diffAFD < 2) {
if (Rs < 116) {
return 4;
}
else if (Rs >= 116) {
return 2;
}
}
else if (diffAFD >= 2) {
if (Rs < 16) {
return 2;
}
else if (Rs >= 16) {
return 1;
}
}
}
}
}
else if (Cs >= 34) {
if (Rs < 37) {
if (diffMVdiffVal < 21) {
return 8;
}
else if (diffMVdiffVal >= 21) {
if (Rs < 27) {
return 2;
}
else if (Rs >= 27) {
return 1;
}
}
}
else if (Rs >= 37) {
if (mu_mv_mag_sq < 106) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (Cs < 84) {
return 8;
}
else if (Cs >= 84) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 106) {
return 4;
}
}
}
}
else if (SC >= 149) {
if (mu_mv_mag_sq < 71) {
if (Cs < 22) {
return 1;
}
else if (Cs >= 22) {
return 8;
}
}
else if (mu_mv_mag_sq >= 71) {
if (Rs < 171) {
if (Cs < 85) {
return 4;
}
else if (Cs >= 85) {
return 1;
}
}
else if (Rs >= 171) {
return 1;
}
}
}
}
else if (Cs >= 93) {
if (refDCval < 41) {
if (negBalance < 44) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 155) {
return 8;
}
else if (Rs >= 155) {
return 1;
}
}
}
else if (negBalance >= 44) {
if (Rs < 99) {
return 8;
}
else if (Rs >= 99) {
return 2;
}
}
}
else if (refDCval >= 41) {
if (TSC < 1) {
return 4;
}
else if (TSC >= 1) {
if (refDCval < 44) {
if (Rs < 222) {
return 1;
}
else if (Rs >= 222) {
if (Rs < 270) {
return 4;
}
else if (Rs >= 270) {
return 8;
}
}
}
else if (refDCval >= 44) {
if (Cs < 178) {
if (mcTcor < 106) {
return 8;
}
else if (mcTcor >= 106) {
return 4;
}
}
else if (Cs >= 178) {
if (Cs < 186) {
return 1;
}
else if (Cs >= 186) {
return 4;
}
}
}
}
}
}
}
}
else if (SC >= 525) {
if (posBalance < 40) {
if (Rs < 232) {
if (Cs < 420) {
if (Rs < 191) {
return 4;
}
else if (Rs >= 191) {
return 8;
}
}
else if (Cs >= 420) {
return 1;
}
}
else if (Rs >= 232) {
return 1;
}
}
else if (posBalance >= 40) {
return 1;
}
}
}
}
else if (ssDCval >= 51) {
if (refDCval < 70) {
if (TSCindex < 2) {
if (ssDCval < 66) {
if (posBalance < 31) {
if (ssDCval < 63) {
if (Cs < 40) {
return 2;
}
else if (Cs >= 40) {
return 8;
}
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (posBalance >= 31) {
if (Rs < 65) {
return 1;
}
else if (Rs >= 65) {
if (Rs < 254) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
if (SC < 342) {
if (MVDiff < 3) {
if (posBalance < 43) {
if (SC < 314) {
return 1;
}
else if (SC >= 314) {
return 4;
}
}
else if (posBalance >= 43) {
if (Rs < 146) {
if (Rs < 144) {
return 1;
}
else if (Rs >= 144) {
return 4;
}
}
else if (Rs >= 146) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 151) {
return 4;
}
else if (Rs >= 151) {
return 1;
}
}
}
else if (SC >= 342) {
if (MVDiff < 6) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (diffAFD < -1) {
if (refDCval < 58) {
return 1;
}
else if (refDCval >= 58) {
return 4;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (MVDiff >= 6) {
return 1;
}
}
}
}
else if (Rs >= 254) {
if (ssDCval < 56) {
if (Rs < 295) {
return 4;
}
else if (Rs >= 295) {
return 2;
}
}
else if (ssDCval >= 56) {
if (Rs < 288) {
return 1;
}
else if (Rs >= 288) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 66) {
if (Cs < 241) {
return 2;
}
else if (Cs >= 241) {
if (MVDiff < 2) {
if (Cs < 379) {
return 4;
}
else if (Cs >= 379) {
return 1;
}
}
else if (MVDiff >= 2) {
return 2;
}
}
}
}
else if (TSCindex >= 2) {
if (RsCsDiff < 1) {
if (SC < 370) {
if (tcor < 100) {
if (MVDiff < 1) {
if (Cs < 129) {
if (mcTcor < 97) {
if (Rs < 94) {
return 2;
}
else if (Rs >= 94) {
return 8;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
else if (Cs >= 129) {
if (Cs < 221) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (Cs >= 221) {
return 4;
}
}
}
else if (MVDiff >= 1) {
if (Cs < 200) {
if (mcTcor < 96) {
if (negBalance < 45) {
if (Cs < 60) {
return 2;
}
else if (Cs >= 60) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (MVDiff < 55) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
else if (MVDiff >= 55) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 68) {
return 1;
}
else if (Rs >= 68) {
if (tcor < 93) {
if (AFD < 5) {
if (tcor < 91) {
return 1;
}
else if (tcor >= 91) {
if (TSC < 2) {
return 1;
}
else if (TSC >= 2) {
return 4;
}
}
}
else if (AFD >= 5) {
return 4;
}
}
else if (tcor >= 93) {
if (ssDCval < 66) {
if (negBalance < 24) {
if (ssDCval < 57) {
if (mu_mv_mag_sq < 93) {
return 1;
}
else if (mu_mv_mag_sq >= 93) {
if (Rs < 160) {
return 4;
}
else if (Rs >= 160) {
return 1;
}
}
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (negBalance >= 24) {
if (posBalance < 39) {
if (Rs < 116) {
return 8;
}
else if (Rs >= 116) {
return 4;
}
}
else if (posBalance >= 39) {
if (diffAFD < 1) {
if (ssDCval < 65) {
if (posBalance < 43) {
if (Cs < 177) {
return 1;
}
else if (Cs >= 177) {
return 4;
}
}
else if (posBalance >= 43) {
return 1;
}
}
else if (ssDCval >= 65) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (diffAFD >= 1) {
if (SC < 320) {
return 1;
}
else if (SC >= 320) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 66) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
}
}
else if (Cs >= 200) {
if (MVDiff < 16) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (MVDiff >= 16) {
if (Rs < 108) {
if (Rs < 52) {
return 1;
}
else if (Rs >= 52) {
return 2;
}
}
else if (Rs >= 108) {
return 8;
}
}
}
}
}
else if (tcor >= 100) {
if (Rs < 63) {
return 2;
}
else if (Rs >= 63) {
return 8;
}
}
}
else if (SC >= 370) {
if (negBalance < 31) {
if (ssDCval < 60) {
if (MVDiff < 1) {
if (Rs < 171) {
return 4;
}
else if (Rs >= 171) {
if (ssDCval < 53) {
if (Rs < 243) {
return 4;
}
else if (Rs >= 243) {
return 1;
}
}
else if (ssDCval >= 53) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
}
else if (MVDiff >= 1) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 1;
}
}
}
else if (ssDCval >= 60) {
if (MVDiff < 4) {
if (negBalance < 23) {
if (Cs < 377) {
return 4;
}
else if (Cs >= 377) {
if (Cs < 408) {
return 1;
}
else if (Cs >= 408) {
return 4;
}
}
}
else if (negBalance >= 23) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 68) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 123) {
return 4;
}
else if (mu_mv_mag_sq >= 123) {
return 1;
}
}
else if (posBalance >= 49) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
return 1;
}
}
}
}
else if (ssDCval >= 68) {
if (Rs < 185) {
if (Cs < 353) {
if (ssDCval < 69) {
if (Rs < 148) {
return 1;
}
else if (Rs >= 148) {
if (Cs < 330) {
if (mu_mv_mag_sq < 105) {
return 1;
}
else if (mu_mv_mag_sq >= 105) {
return 4;
}
}
else if (Cs >= 330) {
if (Rs < 168) {
return 1;
}
else if (Rs >= 168) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (Cs >= 353) {
return 1;
}
}
else if (Rs >= 185) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
if (Rs < 156) {
return 4;
}
else if (Rs >= 156) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (ssDCval < 63) {
if (Rs < 263) {
return 4;
}
else if (Rs >= 263) {
return 2;
}
}
else if (ssDCval >= 63) {
if (Cs < 288) {
if (Cs < 175) {
return 4;
}
else if (Cs >= 175) {
if (Rs < 253) {
return 1;
}
else if (Rs >= 253) {
return 4;
}
}
}
else if (Cs >= 288) {
if (Rs < 181) {
return 2;
}
else if (Rs >= 181) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 31) {
if (refDCval < 63) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (SC < 403) {
if (Rs < 193) {
return 1;
}
else if (Rs >= 193) {
return 8;
}
}
else if (SC >= 403) {
if (posBalance < 52) {
return 4;
}
else if (posBalance >= 52) {
if (mu_mv_mag_sq < 68) {
if (Cs < 173) {
return 4;
}
else if (Cs >= 173) {
return 1;
}
}
else if (mu_mv_mag_sq >= 68) {
return 4;
}
}
}
}
}
else if (refDCval >= 63) {
if (refDCval < 67) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 1;
}
}
}
else if (refDCval >= 67) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (refDCval < 69) {
return 1;
}
else if (refDCval >= 69) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (TSC < 4) {
if (mcTcor < 91) {
if (negBalance < 19) {
return 1;
}
else if (negBalance >= 19) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
if (Cs < 65) {
return 2;
}
else if (Cs >= 65) {
if (MVDiff < 6) {
if (Rs < 204) {
return 1;
}
else if (Rs >= 204) {
return 8;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
}
}
}
else if (mcTcor >= 91) {
if (Rs < 280) {
if (refDCval < 52) {
if (Rs < 234) {
return 1;
}
else if (Rs >= 234) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (Cs < 297) {
return 1;
}
else if (Cs >= 297) {
return 4;
}
}
}
}
else if (refDCval >= 52) {
if (Rs < 37) {
return 1;
}
else if (Rs >= 37) {
if (MVDiff < 9) {
if (SC < 358) {
if (diffRsCsdiff < -4) {
return 8;
}
else if (diffRsCsdiff >= -4) {
if (diffAFD < 1) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (diffAFD >= 1) {
if (Cs < 155) {
return 4;
}
else if (Cs >= 155) {
return 1;
}
}
}
}
else if (SC >= 358) {
if (MVDiff < 3) {
if (mu_mv_mag_sq < 116) {
if (ssDCval < 69) {
if (SC < 678) {
if (SC < 463) {
if (SC < 412) {
return 4;
}
else if (SC >= 412) {
return 1;
}
}
else if (SC >= 463) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
return 4;
}
}
}
else if (SC >= 678) {
return 1;
}
}
else if (ssDCval >= 69) {
if (diffMVdiffVal < -1) {
if (Rs < 228) {
if (Cs < 362) {
return 4;
}
else if (Cs >= 362) {
return 1;
}
}
else if (Rs >= 228) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 116) {
return 1;
}
}
else if (MVDiff >= 3) {
if (Cs < 296) {
if (Rs < 258) {
if (Cs < 285) {
if (negBalance < 35) {
if (Cs < 238) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
if (SC < 426) {
return 4;
}
else if (SC >= 426) {
return 1;
}
}
}
else if (Cs >= 238) {
return 4;
}
}
else if (negBalance >= 35) {
return 4;
}
}
else if (Cs >= 285) {
return 1;
}
}
else if (Rs >= 258) {
return 1;
}
}
else if (Cs >= 296) {
if (tcor < 105) {
if (Rs < 235) {
return 4;
}
else if (Rs >= 235) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (Rs < 241) {
return 8;
}
else if (Rs >= 241) {
return 4;
}
}
}
}
else if (tcor >= 105) {
return 8;
}
}
}
}
}
else if (MVDiff >= 9) {
if (RsCsDiff < 5) {
if (TSC < 2) {
if (mu_mv_mag_sq < 33) {
return 1;
}
else if (mu_mv_mag_sq >= 33) {
return 4;
}
}
else if (TSC >= 2) {
if (Cs < 46) {
return 2;
}
else if (Cs >= 46) {
if (mcTcor < 100) {
if (RsDiff < 1) {
if (MVDiff < 78) {
if (ssDCval < 54) {
if (Rs < 127) {
return 2;
}
else if (Rs >= 127) {
return 4;
}
}
else if (ssDCval >= 54) {
return 4;
}
}
else if (MVDiff >= 78) {
return 1;
}
}
else if (RsDiff >= 1) {
if (AFD < 5) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (mcTcor >= 100) {
if (Rs < 179) {
return 8;
}
else if (Rs >= 179) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 5) {
if (AFD < 9) {
if (Rs < 226) {
if (MVDiff < 18) {
return 4;
}
else if (MVDiff >= 18) {
return 2;
}
}
else if (Rs >= 226) {
return 2;
}
}
else if (AFD >= 9) {
if (Cs < 286) {
if (Rs < 184) {
return 2;
}
else if (Rs >= 184) {
return 4;
}
}
else if (Cs >= 286) {
return 1;
}
}
}
}
}
}
}
else if (Rs >= 280) {
if (AFD < 5) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
if (Rs < 288) {
return 4;
}
else if (Rs >= 288) {
return 1;
}
}
}
else if (AFD >= 5) {
if (gchDC < 1) {
if (Rs < 285) {
return 1;
}
else if (Rs >= 285) {
return 4;
}
}
else if (gchDC >= 1) {
if (Rs < 293) {
if (Cs < 167) {
return 4;
}
else if (Cs >= 167) {
return 2;
}
}
else if (Rs >= 293) {
return 4;
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 25) {
if (mcTcor < 76) {
if (Cs < 217) {
return 1;
}
else if (Cs >= 217) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (mcTcor >= 76) {
if (MVDiff < 56) {
if (tcor < 90) {
if (AFD < 15) {
if (negBalance < 33) {
return 4;
}
else if (negBalance >= 33) {
return 1;
}
}
else if (AFD >= 15) {
if (diffRsCsdiff < -6) {
if (Cs < 257) {
return 8;
}
else if (Cs >= 257) {
return 1;
}
}
else if (diffRsCsdiff >= -6) {
return 2;
}
}
}
else if (tcor >= 90) {
if (ssDCval < 58) {
if (Rs < 235) {
return 8;
}
else if (Rs >= 235) {
return 1;
}
}
else if (ssDCval >= 58) {
return 2;
}
}
}
else if (MVDiff >= 56) {
if (posBalance < 62) {
if (diffRsCsdiff < -13) {
return 1;
}
else if (diffRsCsdiff >= -13) {
return 2;
}
}
else if (posBalance >= 62) {
if (MVDiff < 114) {
if (AFD < 15) {
if (Cs < 29) {
return 1;
}
else if (Cs >= 29) {
if (Rs < 279) {
if (refDCval < 61) {
return 2;
}
else if (refDCval >= 61) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (Rs >= 279) {
return 1;
}
}
}
else if (AFD >= 15) {
return 1;
}
}
else if (MVDiff >= 114) {
if (Rs < 216) {
return 1;
}
else if (Rs >= 216) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (posBalance < 63) {
if (Rs < 261) {
if (MVDiff < 28) {
if (refDCval < 65) {
if (diffMVdiffVal < -7) {
if (Rs < 192) {
return 4;
}
else if (Rs >= 192) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
return 2;
}
}
else if (refDCval >= 65) {
if (Rs < 154) {
return 2;
}
else if (Rs >= 154) {
return 4;
}
}
}
else if (MVDiff >= 28) {
if (gchDC < 3) {
if (Rs < 155) {
return 8;
}
else if (Rs >= 155) {
return 4;
}
}
else if (gchDC >= 3) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
return 8;
}
}
}
}
}
else if (Rs >= 261) {
if (refDCval < 62) {
return 2;
}
else if (refDCval >= 62) {
if (Cs < 207) {
return 8;
}
else if (Cs >= 207) {
return 1;
}
}
}
}
else if (posBalance >= 63) {
if (ssDCval < 70) {
if (ssDCval < 63) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (ssDCval >= 63) {
if (negBalance < 49) {
if (AFD < 35) {
return 1;
}
else if (AFD >= 35) {
return 8;
}
}
else if (negBalance >= 49) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (MVDiff < 523) {
if (mu_mv_mag_sq < 58) {
return 2;
}
else if (mu_mv_mag_sq >= 58) {
return 1;
}
}
else if (MVDiff >= 523) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 70) {
if (TSC < 2) {
if (refDCval < 81) {
if (mu_mv_mag_sq < 94) {
if (mu_mv_mag_sq < 72) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (mu_mv_mag_sq < 62) {
return 1;
}
else if (mu_mv_mag_sq >= 62) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (Cs < 251) {
if (ssDCval < 75) {
if (Rs < 71) {
return 1;
}
else if (Rs >= 71) {
return 2;
}
}
else if (ssDCval >= 75) {
if (Rs < 165) {
if (Cs < 98) {
return 2;
}
else if (Cs >= 98) {
if (SC < 222) {
return 2;
}
else if (SC >= 222) {
return 1;
}
}
}
else if (Rs >= 165) {
return 2;
}
}
}
else if (Cs >= 251) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (Rs < 121) {
return 4;
}
else if (Rs >= 121) {
if (Cs < 210) {
if (Cs < 125) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 130) {
return 1;
}
else if (Rs >= 130) {
return 2;
}
}
}
else if (Cs >= 125) {
if (ssDCval < 80) {
if (diffTSC < -1) {
if (refDCval < 74) {
return 2;
}
else if (refDCval >= 74) {
if (Rs < 160) {
return 2;
}
else if (Rs >= 160) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (Cs < 127) {
if (Rs < 147) {
return 2;
}
else if (Rs >= 147) {
return 8;
}
}
else if (Cs >= 127) {
return 2;
}
}
}
else if (ssDCval >= 80) {
if (Rs < 209) {
return 2;
}
else if (Rs >= 209) {
return 8;
}
}
}
}
else if (Cs >= 210) {
if (negBalance < 23) {
if (Cs < 268) {
if (SC < 413) {
if (Rs < 193) {
return 2;
}
else if (Rs >= 193) {
return 1;
}
}
else if (SC >= 413) {
return 2;
}
}
else if (Cs >= 268) {
return 1;
}
}
else if (negBalance >= 23) {
if (mu_mv_mag_sq < 121) {
if (ssDCval < 74) {
if (Rs < 170) {
return 2;
}
else if (Rs >= 170) {
return 1;
}
}
else if (ssDCval >= 74) {
return 2;
}
}
else if (mu_mv_mag_sq >= 121) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 81) {
if (Cs < 142) {
if (ssDCval < 110) {
if (Cs < 117) {
if (refDCval < 106) {
if (ssDCval < 82) {
if (Rs < 109) {
return 1;
}
else if (Rs >= 109) {
return 8;
}
}
else if (ssDCval >= 82) {
return 1;
}
}
else if (refDCval >= 106) {
if (Rs < 47) {
return 1;
}
else if (Rs >= 47) {
return 2;
}
}
}
else if (Cs >= 117) {
if (Rs < 89) {
if (Rs < 27) {
return 2;
}
else if (Rs >= 27) {
return 4;
}
}
else if (Rs >= 89) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
}
else if (ssDCval >= 110) {
if (Rs < 120) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
else if (Rs >= 120) {
if (ssDCval < 138) {
return 8;
}
else if (ssDCval >= 138) {
return 2;
}
}
}
}
else if (Cs >= 142) {
if (ssDCval < 132) {
if (Cs < 185) {
if (ssDCval < 96) {
if (Rs < 153) {
if (Rs < 93) {
if (Rs < 49) {
return 2;
}
else if (Rs >= 49) {
return 4;
}
}
else if (Rs >= 93) {
return 1;
}
}
else if (Rs >= 153) {
if (Rs < 233) {
return 2;
}
else if (Rs >= 233) {
return 4;
}
}
}
else if (ssDCval >= 96) {
if (diffTSC < -1) {
if (Rs < 90) {
return 1;
}
else if (Rs >= 90) {
return 2;
}
}
else if (diffTSC >= -1) {
if (Rs < 15) {
return 2;
}
else if (Rs >= 15) {
if (Rs < 208) {
return 4;
}
else if (Rs >= 208) {
return 1;
}
}
}
}
}
else if (Cs >= 185) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (tcor < 95) {
if (Rs < 31) {
return 2;
}
else if (Rs >= 31) {
return 1;
}
}
else if (tcor >= 95) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
if (Rs < 145) {
return 1;
}
else if (Rs >= 145) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 132) {
if (Cs < 198) {
return 8;
}
else if (Cs >= 198) {
return 2;
}
}
}
}
}
else if (TSC >= 2) {
if (mcTcor < 97) {
if (MVDiff < 21) {
if (mcTcor < 87) {
if (Cs < 361) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (Rs < 210) {
return 8;
}
else if (Rs >= 210) {
if (TSC < 5) {
if (Cs < 346) {
return 2;
}
else if (Cs >= 346) {
if (Rs < 256) {
return 8;
}
else if (Rs >= 256) {
return 2;
}
}
}
else if (TSC >= 5) {
return 8;
}
}
}
}
else if (Cs >= 361) {
return 8;
}
}
else if (mcTcor >= 87) {
if (Rs < 109) {
if (Cs < 197) {
if (Rs < 9) {
return 4;
}
else if (Rs >= 9) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
return 2;
}
}
}
else if (Cs >= 197) {
if (Rs < 64) {
return 2;
}
else if (Rs >= 64) {
return 4;
}
}
}
else if (Rs >= 109) {
if (tcor < 75) {
if (Rs < 290) {
if (diffMVdiffVal < -8) {
return 2;
}
else if (diffMVdiffVal >= -8) {
if (tcor < 57) {
return 8;
}
else if (tcor >= 57) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
if (diffMVdiffVal < 0) {
return 8;
}
else if (diffMVdiffVal >= 0) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
return 8;
}
}
}
}
}
}
else if (Rs >= 290) {
return 1;
}
}
else if (tcor >= 75) {
if (tcor < 84) {
if (posBalance < 57) {
if (ssDCval < 116) {
return 1;
}
else if (ssDCval >= 116) {
if (Rs < 191) {
return 1;
}
else if (Rs >= 191) {
return 2;
}
}
}
else if (posBalance >= 57) {
if (Rs < 136) {
return 1;
}
else if (Rs >= 136) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
return 1;
}
}
}
}
else if (tcor >= 84) {
if (Cs < 64) {
if (Cs < 5) {
return 8;
}
else if (Cs >= 5) {
return 1;
}
}
else if (Cs >= 64) {
if (diffTSC < 1) {
if (Cs < 423) {
if (mu_mv_mag_sq < 3) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Rs < 212) {
return 4;
}
else if (Rs >= 212) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 6) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (ssDCval < 123) {
if (diffMVdiffVal < 1) {
return 1;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 240) {
return 8;
}
else if (Rs >= 240) {
return 4;
}
}
}
else if (ssDCval >= 123) {
return 8;
}
}
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (Cs >= 423) {
return 2;
}
}
else if (diffTSC >= 1) {
if (Rs < 268) {
if (Cs < 81) {
return 1;
}
else if (Cs >= 81) {
return 8;
}
}
else if (Rs >= 268) {
return 1;
}
}
}
}
}
}
}
}
else if (MVDiff >= 21) {
if (mcTcor < 71) {
if (gchDC < 6) {
if (AFD < 29) {
return 4;
}
else if (AFD >= 29) {
if (Rs < 81) {
return 4;
}
else if (Rs >= 81) {
return 8;
}
}
}
else if (gchDC >= 6) {
if (ssDCval < 186) {
if (refDCval < 81) {
if (Rs < 251) {
return 8;
}
else if (Rs >= 251) {
if (Rs < 270) {
return 4;
}
else if (Rs >= 270) {
return 2;
}
}
}
else if (refDCval >= 81) {
return 1;
}
}
else if (ssDCval >= 186) {
return 2;
}
}
}
else if (mcTcor >= 71) {
if (ssDCval < 134) {
if (posBalance < 42) {
if (tcor < 99) {
if (SC < 552) {
if (tcor < 90) {
return 1;
}
else if (tcor >= 90) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
return 1;
}
}
}
}
else if (SC >= 552) {
return 8;
}
}
else if (tcor >= 99) {
if (Cs < 300) {
return 8;
}
else if (Cs >= 300) {
return 4;
}
}
}
else if (posBalance >= 42) {
if (negBalance < 44) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (Cs < 323) {
if (diffTSC < 4) {
if (Rs < 282) {
return 1;
}
else if (Rs >= 282) {
if (Rs < 285) {
return 8;
}
else if (Rs >= 285) {
return 1;
}
}
}
else if (diffTSC >= 4) {
if (MVDiff < 31) {
return 8;
}
else if (MVDiff >= 31) {
return 1;
}
}
}
else if (Cs >= 323) {
if (Rs < 254) {
return 8;
}
else if (Rs >= 254) {
return 2;
}
}
}
}
else if (negBalance >= 44) {
if (Cs < 45) {
if (tcor < 55) {
if (MVDiff < 488) {
if (Cs < 18) {
if (Rs < 122) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (Rs >= 122) {
if (MVDiff < 47) {
return 2;
}
else if (MVDiff >= 47) {
return 8;
}
}
}
else if (Cs >= 18) {
if (mu_mv_mag_sq < 43) {
if (SC < 125) {
return 8;
}
else if (SC >= 125) {
return 2;
}
}
else if (mu_mv_mag_sq >= 43) {
return 8;
}
}
}
else if (MVDiff >= 488) {
return 1;
}
}
else if (tcor >= 55) {
if (MVDiff < 105) {
return 1;
}
else if (MVDiff >= 105) {
if (Cs < 28) {
return 8;
}
else if (Cs >= 28) {
return 2;
}
}
}
}
else if (Cs >= 45) {
if (AFD < 17) {
if (RsCsDiff < 5) {
if (tcor < 72) {
if (TSC < 7) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
return 2;
}
}
else if (TSC >= 7) {
if (Rs < 188) {
return 8;
}
else if (Rs >= 188) {
return 4;
}
}
}
else if (tcor >= 72) {
if (CsDiff < 1) {
if (ssDCval < 107) {
return 1;
}
else if (ssDCval >= 107) {
return 2;
}
}
else if (CsDiff >= 1) {
return 2;
}
}
}
else if (RsCsDiff >= 5) {
if (negBalance < 52) {
if (SC < 626) {
return 8;
}
else if (SC >= 626) {
return 2;
}
}
else if (negBalance >= 52) {
if (tcor < 69) {
if (Rs < 288) {
return 8;
}
else if (Rs >= 288) {
return 2;
}
}
else if (tcor >= 69) {
if (refDCval < 73) {
return 2;
}
else if (refDCval >= 73) {
return 4;
}
}
}
}
}
else if (AFD >= 17) {
if (refDCval < 76) {
if (CsDiff < 3) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (CsDiff >= 3) {
if (Rs < 224) {
return 8;
}
else if (Rs >= 224) {
return 2;
}
}
}
else if (refDCval >= 76) {
if (mu_mv_mag_sq < 17) {
if (SC < 446) {
if (RsCsDiff < 12) {
if (mu_mv_mag_sq < 9) {
if (Rs < 148) {
return 1;
}
else if (Rs >= 148) {
if (Cs < 181) {
return 2;
}
else if (Cs >= 181) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 9) {
return 2;
}
}
else if (RsCsDiff >= 12) {
return 8;
}
}
else if (SC >= 446) {
return 2;
}
}
else if (mu_mv_mag_sq >= 17) {
if (Cs < 132) {
return 2;
}
else if (Cs >= 132) {
if (negBalance < 65) {
if (gchDC < 9) {
return 1;
}
else if (gchDC >= 9) {
return 8;
}
}
else if (negBalance >= 65) {
if (diffMVdiffVal < -2) {
return 2;
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 134) {
if (Rs < 6) {
if (mu_mv_mag_sq < 16) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 1) {
return 1;
}
else if (Rs >= 1) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 16) {
return 1;
}
}
else if (Rs >= 6) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 100) {
if (Cs < 234) {
if (Rs < 111) {
if (AFD < 8) {
if (TSC < 3) {
if (refDCval < 75) {
if (mu_mv_mag_sq < 8) {
return 4;
}
else if (mu_mv_mag_sq >= 8) {
return 1;
}
}
else if (refDCval >= 75) {
return 1;
}
}
else if (TSC >= 3) {
if (MVDiff < 18) {
if (RsCsDiff < 1) {
return 2;
}
else if (RsCsDiff >= 1) {
if (Rs < 90) {
return 4;
}
else if (Rs >= 90) {
return 1;
}
}
}
else if (MVDiff >= 18) {
return 1;
}
}
}
else if (AFD >= 8) {
if (Cs < 98) {
return 4;
}
else if (Cs >= 98) {
if (SC < 268) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
else if (SC >= 268) {
return 2;
}
}
}
}
else if (Rs >= 111) {
if (refDCval < 91) {
if (MVDiff < 4) {
if (SC < 377) {
if (Cs < 123) {
return 8;
}
else if (Cs >= 123) {
if (tcor < 96) {
if (Rs < 183) {
return 8;
}
else if (Rs >= 183) {
return 1;
}
}
else if (tcor >= 96) {
if (Cs < 137) {
if (Rs < 155) {
return 2;
}
else if (Rs >= 155) {
return 8;
}
}
else if (Cs >= 137) {
return 2;
}
}
}
}
else if (SC >= 377) {
if (ssDCval < 76) {
if (Cs < 188) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (Cs >= 188) {
return 1;
}
}
else if (ssDCval >= 76) {
if (Cs < 197) {
return 1;
}
else if (Cs >= 197) {
return 2;
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 130) {
if (mu_mv_mag_sq < 67) {
if (SC < 302) {
return 1;
}
else if (SC >= 302) {
return 2;
}
}
else if (mu_mv_mag_sq >= 67) {
return 2;
}
}
else if (Rs >= 130) {
if (MVDiff < 49) {
if (ssDCval < 70) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 218) {
return 8;
}
else if (Rs >= 218) {
return 4;
}
}
}
else if (ssDCval >= 70) {
if (Cs < 220) {
if (negBalance < 24) {
if (Rs < 196) {
return 8;
}
else if (Rs >= 196) {
return 1;
}
}
else if (negBalance >= 24) {
return 1;
}
}
else if (Cs >= 220) {
return 8;
}
}
}
else if (MVDiff >= 49) {
if (Rs < 259) {
return 2;
}
else if (Rs >= 259) {
return 8;
}
}
}
}
}
else if (refDCval >= 91) {
if (SC < 332) {
if (posBalance < 61) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
return 1;
}
}
else if (posBalance >= 61) {
if (mu_mv_mag_sq < 7) {
return 2;
}
else if (mu_mv_mag_sq >= 7) {
if (Rs < 118) {
return 2;
}
else if (Rs >= 118) {
return 1;
}
}
}
}
else if (SC >= 332) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
if (tcor < 91) {
if (Rs < 178) {
return 8;
}
else if (Rs >= 178) {
return 4;
}
}
else if (tcor >= 91) {
if (Rs < 283) {
return 2;
}
else if (Rs >= 283) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 234) {
if (refDCval < 76) {
if (Rs < 121) {
if (refDCval < 72) {
return 4;
}
else if (refDCval >= 72) {
return 1;
}
}
else if (Rs >= 121) {
if (tcor < 97) {
if (MVDiff < 4) {
if (Rs < 152) {
return 4;
}
else if (Rs >= 152) {
return 1;
}
}
else if (MVDiff >= 4) {
if (RsDiff < 1) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (tcor >= 97) {
if (Rs < 126) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 4;
}
}
else if (Rs >= 126) {
return 1;
}
}
}
}
else if (refDCval >= 76) {
if (Cs < 361) {
if (ssDCval < 84) {
if (Rs < 181) {
return 4;
}
else if (Rs >= 181) {
if (Cs < 289) {
return 2;
}
else if (Cs >= 289) {
if (Rs < 219) {
return 4;
}
else if (Rs >= 219) {
return 1;
}
}
}
}
else if (ssDCval >= 84) {
if (TSC < 6) {
if (mu_mv_mag_sq < 21) {
return 2;
}
else if (mu_mv_mag_sq >= 21) {
if (SC < 384) {
return 2;
}
else if (SC >= 384) {
return 1;
}
}
}
else if (TSC >= 6) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Rs < 134) {
return 2;
}
else if (Rs >= 134) {
return 1;
}
}
}
}
}
else if (Cs >= 361) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 8) {
return 2;
}
else if (mu_mv_mag_sq >= 8) {
if (tcor < 90) {
return 1;
}
else if (tcor >= 90) {
if (Rs < 174) {
return 2;
}
else if (Rs >= 174) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 100) {
if (tcor < 80) {
if (gchDC < 6) {
return 1;
}
else if (gchDC >= 6) {
return 8;
}
}
else if (tcor >= 80) {
if (posBalance < 12) {
return 8;
}
else if (posBalance >= 12) {
if (Cs < 394) {
if (MVDiff < 8) {
if (mcTcor < 98) {
if (Rs < 291) {
return 8;
}
else if (Rs >= 291) {
return 2;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (MVDiff >= 8) {
if (MVDiff < 56) {
if (ssDCval < 113) {
if (SC < 368) {
if (Rs < 43) {
if (Cs < 34) {
return 2;
}
else if (Cs >= 34) {
return 1;
}
}
else if (Rs >= 43) {
if (Cs < 216) {
return 2;
}
else if (Cs >= 216) {
return 1;
}
}
}
else if (SC >= 368) {
if (posBalance < 63) {
return 1;
}
else if (posBalance >= 63) {
if (Rs < 233) {
return 8;
}
else if (Rs >= 233) {
return 2;
}
}
}
}
else if (ssDCval >= 113) {
return 2;
}
}
else if (MVDiff >= 56) {
if (tcor < 92) {
if (diffRsCsdiff < -1) {
if (Rs < 230) {
return 1;
}
else if (Rs >= 230) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
if (refDCval < 98) {
return 8;
}
else if (refDCval >= 98) {
return 2;
}
}
}
else if (tcor >= 92) {
if (Cs < 6) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 6) {
return 1;
}
}
}
}
}
else if (Cs >= 394) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 300) {
if (RsDiff < 4) {
if (refDCval < 73) {
if (mcTcor < 89) {
if (refDCval < 41) {
if (tcor < 70) {
if (negBalance < 47) {
if (mcTcor < 79) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (Cs < 263) {
return 8;
}
else if (Cs >= 263) {
return 2;
}
}
}
else if (mcTcor >= 79) {
return 1;
}
}
else if (negBalance >= 47) {
if (mu_mv_mag_sq < 31) {
return 8;
}
else if (mu_mv_mag_sq >= 31) {
if (Rs < 347) {
return 1;
}
else if (Rs >= 347) {
if (Rs < 497) {
return 4;
}
else if (Rs >= 497) {
return 8;
}
}
}
}
}
else if (tcor >= 70) {
if (Cs < 156) {
if (Cs < 61) {
return 8;
}
else if (Cs >= 61) {
if (Cs < 141) {
return 1;
}
else if (Cs >= 141) {
return 4;
}
}
}
else if (Cs >= 156) {
if (negBalance < 27) {
if (posBalance < 61) {
return 1;
}
else if (posBalance >= 61) {
return 8;
}
}
else if (negBalance >= 27) {
if (Rs < 845) {
return 8;
}
else if (Rs >= 845) {
return 1;
}
}
}
}
}
else if (refDCval >= 41) {
if (mu_mv_mag_sq < 93) {
if (posBalance < 56) {
if (Rs < 423) {
if (Cs < 359) {
if (AFD < 15) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (AFD >= 15) {
return 1;
}
}
else if (Cs >= 359) {
if (diffRsCsdiff < 1) {
if (Rs < 348) {
return 8;
}
else if (Rs >= 348) {
return 2;
}
}
else if (diffRsCsdiff >= 1) {
return 1;
}
}
}
else if (Rs >= 423) {
if (Rs < 630) {
if (mu_mv_mag_sq < 38) {
if (RsCsDiff < 12) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (RsCsDiff >= 12) {
return 2;
}
}
else if (mu_mv_mag_sq >= 38) {
if (posBalance < 52) {
return 2;
}
else if (posBalance >= 52) {
return 1;
}
}
}
else if (Rs >= 630) {
if (mu_mv_mag_sq < 41) {
return 1;
}
else if (mu_mv_mag_sq >= 41) {
if (AFD < 7) {
if (Rs < 747) {
return 1;
}
else if (Rs >= 747) {
return 4;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
}
}
else if (posBalance >= 56) {
if (MVDiff < 36) {
if (MVDiff < 26) {
if (Rs < 675) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
}
else if (Rs >= 675) {
if (Rs < 758) {
return 1;
}
else if (Rs >= 758) {
return 8;
}
}
}
else if (MVDiff >= 26) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
if (Rs < 438) {
return 1;
}
else if (Rs >= 438) {
return 4;
}
}
}
}
else if (MVDiff >= 36) {
if (posBalance < 73) {
if (Rs < 319) {
if (Rs < 306) {
return 2;
}
else if (Rs >= 306) {
return 8;
}
}
else if (Rs >= 319) {
if (gchDC < 5) {
if (MVDiff < 63) {
return 4;
}
else if (MVDiff >= 63) {
if (Rs < 361) {
return 4;
}
else if (Rs >= 361) {
if (Cs < 485) {
return 2;
}
else if (Cs >= 485) {
return 4;
}
}
}
}
else if (gchDC >= 5) {
return 2;
}
}
}
else if (posBalance >= 73) {
if (Cs < 465) {
return 1;
}
else if (Cs >= 465) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (negBalance < 39) {
return 4;
}
else if (negBalance >= 39) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 4;
}
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1292) {
if (TSC < 1) {
if (mcTcor < 99) {
return 4;
}
else if (mcTcor >= 99) {
if (Rs < 1007) {
if (Cs < 15) {
return 4;
}
else if (Cs >= 15) {
if (ssDCval < 71) {
if (Rs < 789) {
return 8;
}
else if (Rs >= 789) {
if (Rs < 860) {
return 1;
}
else if (Rs >= 860) {
return 8;
}
}
}
else if (ssDCval >= 71) {
return 4;
}
}
}
else if (Rs >= 1007) {
return 1;
}
}
}
else if (TSC >= 1) {
if (MVDiff < 2) {
if (ssDCval < 69) {
if (mcTcor < 97) {
if (posBalance < 52) {
if (mu_mv_mag_sq < 112) {
if (mu_mv_mag_sq < 61) {
return 8;
}
else if (mu_mv_mag_sq >= 61) {
if (Rs < 812) {
return 1;
}
else if (Rs >= 812) {
if (Rs < 1047) {
return 4;
}
else if (Rs >= 1047) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (ssDCval < 63) {
if (Cs < 203) {
return 2;
}
else if (Cs >= 203) {
return 8;
}
}
else if (ssDCval >= 63) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (SC < 804) {
return 4;
}
else if (SC >= 804) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
if (Cs < 512) {
if (SC < 1114) {
return 2;
}
else if (SC >= 1114) {
return 8;
}
}
else if (Cs >= 512) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
if (tcor < 101) {
if (mu_mv_mag_sq < 108) {
if (Rs < 733) {
if (negBalance < 25) {
if (diffMVdiffVal < -3) {
if (posBalance < 45) {
return 8;
}
else if (posBalance >= 45) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (tcor < 99) {
if (Rs < 667) {
return 8;
}
else if (Rs >= 667) {
return 1;
}
}
else if (tcor >= 99) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (SC < 884) {
return 8;
}
else if (SC >= 884) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 25) {
if (negBalance < 41) {
if (posBalance < 43) {
if (tcor < 99) {
if (Rs < 678) {
return 8;
}
else if (Rs >= 678) {
return 1;
}
}
else if (tcor >= 99) {
return 1;
}
}
else if (posBalance >= 43) {
return 1;
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
else if (Rs >= 733) {
return 1;
}
}
else if (mu_mv_mag_sq >= 108) {
if (posBalance < 43) {
if (TSCindex < 3) {
if (Cs < 256) {
return 4;
}
else if (Cs >= 256) {
if (refDCval < 64) {
if (Cs < 407) {
if (tcor < 98) {
if (Rs < 860) {
return 8;
}
else if (Rs >= 860) {
return 1;
}
}
else if (tcor >= 98) {
return 1;
}
}
else if (Cs >= 407) {
if (posBalance < 42) {
if (ssDCval < 56) {
if (Rs < 1271) {
return 1;
}
else if (Rs >= 1271) {
return 8;
}
}
else if (ssDCval >= 56) {
return 8;
}
}
else if (posBalance >= 42) {
return 1;
}
}
}
else if (refDCval >= 64) {
return 1;
}
}
}
else if (TSCindex >= 3) {
if (RsDiff < 1) {
if (Cs < 409) {
return 1;
}
else if (Cs >= 409) {
return 8;
}
}
else if (RsDiff >= 1) {
if (Cs < 529) {
return 4;
}
else if (Cs >= 529) {
return 2;
}
}
}
}
else if (posBalance >= 43) {
if (Rs < 599) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
if (AFD < 3) {
if (Cs < 367) {
if (Rs < 454) {
return 4;
}
else if (Rs >= 454) {
return 2;
}
}
else if (Cs >= 367) {
return 1;
}
}
else if (AFD >= 3) {
return 2;
}
}
}
else if (Rs >= 599) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
}
else if (tcor >= 101) {
if (Cs < 456) {
return 8;
}
else if (Cs >= 456) {
return 4;
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 597) {
if (SC < 952) {
if (Rs < 414) {
return 2;
}
else if (Rs >= 414) {
if (Rs < 524) {
return 1;
}
else if (Rs >= 524) {
return 2;
}
}
}
else if (SC >= 952) {
return 1;
}
}
else if (Rs >= 597) {
if (Rs < 792) {
if (mu_mv_mag_sq < 122) {
if (AFD < 3) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 4;
}
}
else if (AFD >= 3) {
if (Rs < 618) {
return 1;
}
else if (Rs >= 618) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (mu_mv_mag_sq < 125) {
if (Rs < 628) {
return 1;
}
else if (Rs >= 628) {
return 4;
}
}
else if (mu_mv_mag_sq >= 125) {
return 4;
}
}
}
else if (Rs >= 792) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
if (Cs < 507) {
return 1;
}
else if (Cs >= 507) {
if (Rs < 836) {
return 8;
}
else if (Rs >= 836) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 138) {
if (RsDiff < 2) {
if (ssDCval < 64) {
if (TSC < 3) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (Rs < 752) {
if (Rs < 335) {
return 2;
}
else if (Rs >= 335) {
if (Rs < 377) {
return 8;
}
else if (Rs >= 377) {
return 4;
}
}
}
else if (Rs >= 752) {
return 1;
}
}
}
else if (TSC >= 3) {
if (posBalance < 51) {
if (ssDCval < 57) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 8;
}
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (posBalance >= 51) {
return 1;
}
}
}
else if (ssDCval >= 64) {
if (Rs < 512) {
return 8;
}
else if (Rs >= 512) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 618) {
return 4;
}
else if (Rs >= 618) {
return 1;
}
}
}
}
}
else if (RsDiff >= 2) {
if (MVDiff < 7) {
if (ssDCval < 68) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (ssDCval >= 68) {
return 4;
}
}
else if (MVDiff >= 7) {
if (tcor < 83) {
if (RsDiff < 3) {
if (ssDCval < 52) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 8;
}
}
else if (ssDCval >= 52) {
return 8;
}
}
else if (RsDiff >= 3) {
if (Cs < 126) {
return 1;
}
else if (Cs >= 126) {
return 8;
}
}
}
else if (tcor >= 83) {
if (Cs < 133) {
return 1;
}
else if (Cs >= 133) {
if (Rs < 446) {
return 4;
}
else if (Rs >= 446) {
return 1;
}
}
}
}
}
}
else if (Cs >= 138) {
if (tcor < 99) {
if (mu_mv_mag_sq < 90) {
if (ssDCval < 57) {
if (negBalance < 41) {
if (posBalance < 34) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (mcTcor < 95) {
if (Rs < 491) {
return 4;
}
else if (Rs >= 491) {
return 2;
}
}
else if (mcTcor >= 95) {
if (mu_mv_mag_sq < 71) {
return 2;
}
else if (mu_mv_mag_sq >= 71) {
if (Cs < 415) {
return 1;
}
else if (Cs >= 415) {
return 2;
}
}
}
}
}
else if (posBalance >= 34) {
if (Rs < 657) {
if (refDCval < 42) {
if (mu_mv_mag_sq < 50) {
if (Rs < 619) {
if (Cs < 317) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (Cs < 290) {
return 8;
}
else if (Cs >= 290) {
return 1;
}
}
}
else if (Cs >= 317) {
if (posBalance < 38) {
return 8;
}
else if (posBalance >= 38) {
if (Cs < 336) {
if (Rs < 348) {
return 1;
}
else if (Rs >= 348) {
return 4;
}
}
else if (Cs >= 336) {
return 1;
}
}
}
}
else if (Rs >= 619) {
return 8;
}
}
else if (mu_mv_mag_sq >= 50) {
return 1;
}
}
else if (refDCval >= 42) {
if (AFD < 3) {
if (Cs < 347) {
return 4;
}
else if (Cs >= 347) {
return 8;
}
}
else if (AFD >= 3) {
if (mu_mv_mag_sq < 80) {
if (TSC < 4) {
if (CsDiff < 1) {
if (Rs < 425) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 1;
}
}
else if (Rs >= 425) {
return 1;
}
}
else if (CsDiff >= 1) {
if (AFD < 4) {
if (Cs < 356) {
return 4;
}
else if (Cs >= 356) {
return 8;
}
}
else if (AFD >= 4) {
if (posBalance < 46) {
if (tcor < 91) {
return 1;
}
else if (tcor >= 91) {
if (diffAFD < 1) {
if (tcor < 94) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 4;
}
}
else if (tcor >= 94) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
return 1;
}
}
}
else if (diffAFD >= 1) {
return 4;
}
}
}
else if (posBalance >= 46) {
if (Rs < 605) {
if (Rs < 513) {
if (SC < 880) {
if (MVDiff < 50) {
if (MVDiff < 6) {
if (Rs < 397) {
return 4;
}
else if (Rs >= 397) {
return 1;
}
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (MVDiff >= 50) {
return 1;
}
}
else if (SC >= 880) {
return 1;
}
}
else if (Rs >= 513) {
return 4;
}
}
else if (Rs >= 605) {
if (Cs < 395) {
return 1;
}
else if (Cs >= 395) {
return 4;
}
}
}
}
}
}
else if (TSC >= 4) {
if (mcTcor < 96) {
if (mu_mv_mag_sq < 56) {
if (diffRsCsdiff < 2) {
if (Cs < 275) {
return 8;
}
else if (Cs >= 275) {
return 4;
}
}
else if (diffRsCsdiff >= 2) {
if (Rs < 379) {
return 1;
}
else if (Rs >= 379) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (Cs < 510) {
return 1;
}
else if (Cs >= 510) {
return 4;
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (Rs < 338) {
return 2;
}
else if (Rs >= 338) {
return 1;
}
}
}
}
}
else if (Rs >= 657) {
if (mcTcor < 99) {
if (Cs < 432) {
if (Rs < 697) {
if (Rs < 665) {
return 2;
}
else if (Rs >= 665) {
if (Rs < 682) {
return 1;
}
else if (Rs >= 682) {
return 8;
}
}
}
else if (Rs >= 697) {
return 1;
}
}
else if (Cs >= 432) {
if (mcTcor < 96) {
if (ssDCval < 54) {
if (mu_mv_mag_sq < 19) {
return 1;
}
else if (mu_mv_mag_sq >= 19) {
return 2;
}
}
else if (ssDCval >= 54) {
return 4;
}
}
else if (mcTcor >= 96) {
if (posBalance < 52) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 904) {
return 2;
}
else if (Rs >= 904) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (Rs < 771) {
return 1;
}
else if (Rs >= 771) {
return 8;
}
}
}
}
}
else if (mcTcor >= 99) {
if (Rs < 807) {
return 2;
}
else if (Rs >= 807) {
return 4;
}
}
}
}
}
else if (negBalance >= 41) {
if (Rs < 381) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Cs < 477) {
if (tcor < 93) {
if (ssDCval < 42) {
return 1;
}
else if (ssDCval >= 42) {
return 4;
}
}
else if (tcor >= 93) {
if (Cs < 346) {
return 1;
}
else if (Cs >= 346) {
return 4;
}
}
}
else if (Cs >= 477) {
return 2;
}
}
}
else if (Rs >= 381) {
if (TSC < 4) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (MVDiff < 28) {
if (Rs < 890) {
return 1;
}
else if (Rs >= 890) {
return 2;
}
}
else if (MVDiff >= 28) {
return 8;
}
}
}
else if (TSC >= 4) {
if (SC < 924) {
if (gchDC < 1) {
if (Rs < 480) {
return 8;
}
else if (Rs >= 480) {
return 1;
}
}
else if (gchDC >= 1) {
if (MVDiff < 30) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (MVDiff >= 30) {
if (Rs < 515) {
return 2;
}
else if (Rs >= 515) {
return 4;
}
}
}
}
else if (SC >= 924) {
if (AFD < 12) {
if (ssDCval < 56) {
if (negBalance < 42) {
if (Rs < 740) {
return 4;
}
else if (Rs >= 740) {
return 8;
}
}
else if (negBalance >= 42) {
return 8;
}
}
else if (ssDCval >= 56) {
return 1;
}
}
else if (AFD >= 12) {
if (Rs < 612) {
return 1;
}
else if (Rs >= 612) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 57) {
if (Rs < 674) {
if (ssDCval < 62) {
if (TSC < 3) {
if (Cs < 413) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
return 2;
}
}
else if (Cs >= 413) {
return 8;
}
}
else if (TSC >= 3) {
if (SC < 715) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (mu_mv_mag_sq < 63) {
return 2;
}
else if (mu_mv_mag_sq >= 63) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 4;
}
}
}
}
else if (SC >= 715) {
if (Cs < 389) {
if (mu_mv_mag_sq < 61) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 496) {
return 1;
}
else if (Rs >= 496) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 61) {
return 1;
}
}
else if (Cs >= 389) {
if (TSC < 4) {
if (SC < 1013) {
if (Cs < 391) {
return 1;
}
else if (Cs >= 391) {
if (RsDiff < 3) {
if (refDCval < 57) {
return 4;
}
else if (refDCval >= 57) {
return 2;
}
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (SC >= 1013) {
return 4;
}
}
else if (TSC >= 4) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 389) {
return 1;
}
else if (Rs >= 389) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 62) {
if (Rs < 334) {
if (TSC < 3) {
if (MVDiff < 8) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (TSC >= 3) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 320) {
return 1;
}
else if (Rs >= 320) {
if (Rs < 325) {
return 4;
}
else if (Rs >= 325) {
return 1;
}
}
}
}
}
else if (Rs >= 334) {
if (mcTcor < 96) {
if (negBalance < 41) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
if (Rs < 497) {
return 8;
}
else if (Rs >= 497) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (mu_mv_mag_sq < 8) {
if (refDCval < 72) {
return 4;
}
else if (refDCval >= 72) {
return 8;
}
}
else if (mu_mv_mag_sq >= 8) {
if (SC < 787) {
return 4;
}
else if (SC >= 787) {
if (diffAFD < 3) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
if (Cs < 452) {
return 1;
}
else if (Cs >= 452) {
return 2;
}
}
}
else if (diffAFD >= 3) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (MVDiff < 12) {
if (mu_mv_mag_sq < 77) {
if (refDCval < 69) {
if (posBalance < 67) {
if (mcTcor < 97) {
if (Rs < 514) {
return 4;
}
else if (Rs >= 514) {
return 1;
}
}
else if (mcTcor >= 97) {
return 4;
}
}
else if (posBalance >= 67) {
return 2;
}
}
else if (refDCval >= 69) {
if (Rs < 519) {
if (Cs < 432) {
return 1;
}
else if (Cs >= 432) {
return 4;
}
}
else if (Rs >= 519) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 77) {
return 1;
}
}
else if (MVDiff >= 12) {
if (AFD < 8) {
if (Rs < 474) {
return 2;
}
else if (Rs >= 474) {
return 8;
}
}
else if (AFD >= 8) {
return 1;
}
}
}
}
}
}
else if (Rs >= 674) {
if (mcTcor < 96) {
if (refDCval < 64) {
if (Cs < 380) {
if (diffRsCsdiff < -6) {
return 1;
}
else if (diffRsCsdiff >= -6) {
return 2;
}
}
else if (Cs >= 380) {
if (negBalance < 28) {
if (Rs < 966) {
return 2;
}
else if (Rs >= 966) {
return 1;
}
}
else if (negBalance >= 28) {
if (refDCval < 58) {
if (Rs < 963) {
return 1;
}
else if (Rs >= 963) {
return 8;
}
}
else if (refDCval >= 58) {
return 4;
}
}
}
}
else if (refDCval >= 64) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (Cs < 440) {
return 1;
}
else if (Cs >= 440) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 2) {
if (negBalance < 36) {
if (Cs < 164) {
return 1;
}
else if (Cs >= 164) {
return 4;
}
}
else if (negBalance >= 36) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (SC < 1475) {
if (TSC < 4) {
if (RsCsDiff < 1) {
if (Cs < 268) {
return 4;
}
else if (Cs >= 268) {
return 1;
}
}
else if (RsCsDiff >= 1) {
return 1;
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 40) {
return 1;
}
else if (mu_mv_mag_sq >= 40) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
else if (SC >= 1475) {
if (Cs < 386) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 386) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (negBalance < 22) {
if (ssDCval < 56) {
if (TSCindex < 2) {
if (Rs < 458) {
return 8;
}
else if (Rs >= 458) {
return 4;
}
}
else if (TSCindex >= 2) {
if (diffMVdiffVal < -6) {
return 4;
}
else if (diffMVdiffVal >= -6) {
return 1;
}
}
}
else if (ssDCval >= 56) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
return 8;
}
}
}
else if (negBalance >= 22) {
if (ssDCval < 71) {
if (mcTcor < 97) {
if (negBalance < 26) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
if (mu_mv_mag_sq < 93) {
return 1;
}
else if (mu_mv_mag_sq >= 93) {
if (Rs < 394) {
return 8;
}
else if (Rs >= 394) {
return 4;
}
}
}
}
else if (negBalance >= 26) {
if (ssDCval < 63) {
if (MVDiff < 4) {
if (ssDCval < 57) {
if (Cs < 339) {
return 8;
}
else if (Cs >= 339) {
return 4;
}
}
else if (ssDCval >= 57) {
if (Rs < 427) {
return 1;
}
else if (Rs >= 427) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (RsDiff < 3) {
if (Rs < 1010) {
return 8;
}
else if (Rs >= 1010) {
return 4;
}
}
else if (RsDiff >= 3) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 605) {
return 1;
}
else if (Rs >= 605) {
return 2;
}
}
}
}
}
else if (ssDCval >= 63) {
return 4;
}
}
}
else if (mcTcor >= 97) {
if (Cs < 275) {
if (Rs < 730) {
return 4;
}
else if (Rs >= 730) {
return 1;
}
}
else if (Cs >= 275) {
if (negBalance < 28) {
if (mu_mv_mag_sq < 95) {
if (mcTcor < 99) {
return 2;
}
else if (mcTcor >= 99) {
return 8;
}
}
else if (mu_mv_mag_sq >= 95) {
if (ssDCval < 67) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (negBalance < 24) {
if (Rs < 624) {
return 4;
}
else if (Rs >= 624) {
return 8;
}
}
else if (negBalance >= 24) {
if (Rs < 379) {
return 2;
}
else if (Rs >= 379) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
return 2;
}
}
}
else if (negBalance >= 28) {
if (AFD < 3) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Cs < 435) {
return 2;
}
else if (Cs >= 435) {
return 8;
}
}
}
else if (AFD >= 3) {
return 1;
}
}
}
}
}
else if (ssDCval >= 71) {
if (Rs < 484) {
return 1;
}
else if (Rs >= 484) {
return 8;
}
}
}
}
}
else if (tcor >= 99) {
if (Rs < 400) {
if (Rs < 355) {
if (posBalance < 44) {
if (Cs < 254) {
return 1;
}
else if (Cs >= 254) {
return 4;
}
}
else if (posBalance >= 44) {
if (Cs < 329) {
return 2;
}
else if (Cs >= 329) {
return 4;
}
}
}
else if (Rs >= 355) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (Rs < 383) {
return 4;
}
else if (Rs >= 383) {
return 8;
}
}
}
}
else if (Rs >= 400) {
if (ssDCval < 68) {
if (negBalance < 31) {
if (SC < 817) {
return 1;
}
else if (SC >= 817) {
if (SC < 1170) {
return 8;
}
else if (SC >= 1170) {
return 1;
}
}
}
else if (negBalance >= 31) {
if (CsDiff < 2) {
if (RsDiff < 1) {
if (posBalance < 1) {
return 4;
}
else if (posBalance >= 1) {
return 8;
}
}
else if (RsDiff >= 1) {
if (refDCval < 65) {
if (Rs < 482) {
return 1;
}
else if (Rs >= 482) {
return 4;
}
}
else if (refDCval >= 65) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (Rs < 561) {
return 1;
}
else if (Rs >= 561) {
return 8;
}
}
}
}
}
else if (CsDiff >= 2) {
if (negBalance < 45) {
if (Cs < 455) {
return 2;
}
else if (Cs >= 455) {
return 1;
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
}
else if (ssDCval >= 68) {
if (mu_mv_mag_sq < 100) {
if (mcTcor < 99) {
return 4;
}
else if (mcTcor >= 99) {
return 8;
}
}
else if (mu_mv_mag_sq >= 100) {
if (negBalance < 20) {
if (Rs < 792) {
return 4;
}
else if (Rs >= 792) {
return 8;
}
}
else if (negBalance >= 20) {
return 2;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1292) {
if (SC < 1634) {
if (mu_mv_mag_sq < 61) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
else if (mu_mv_mag_sq >= 61) {
if (Cs < 223) {
return 8;
}
else if (Cs >= 223) {
if (Rs < 1329) {
return 8;
}
else if (Rs >= 1329) {
return 1;
}
}
}
}
else if (SC >= 1634) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1716) {
if (mu_mv_mag_sq < 126) {
if (mu_mv_mag_sq < 56) {
if (negBalance < 28) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
else if (negBalance >= 28) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (TSCindex < 2) {
if (ssDCval < 64) {
if (ssDCval < 59) {
return 1;
}
else if (ssDCval >= 59) {
return 2;
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (TSCindex >= 2) {
if (mu_mv_mag_sq < 113) {
if (ssDCval < 68) {
if (posBalance < 70) {
return 1;
}
else if (posBalance >= 70) {
return 2;
}
}
else if (ssDCval >= 68) {
if (Cs < 413) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (Cs >= 413) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (AFD < 2) {
if (mu_mv_mag_sq < 123) {
return 1;
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
else if (AFD >= 2) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (ssDCval < 62) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (ssDCval >= 62) {
if (Rs < 1446) {
if (negBalance < 25) {
return 2;
}
else if (negBalance >= 25) {
return 8;
}
}
else if (Rs >= 1446) {
return 1;
}
}
}
}
else if (Rs >= 1716) {
if (Cs < 434) {
if (Rs < 2386) {
if (RsCsDiff < 10) {
return 2;
}
else if (RsCsDiff >= 10) {
if (Cs < 261) {
return 1;
}
else if (Cs >= 261) {
return 2;
}
}
}
else if (Rs >= 2386) {
return 8;
}
}
else if (Cs >= 434) {
if (mu_mv_mag_sq < 33) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
return 1;
}
}
else if (mu_mv_mag_sq >= 33) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (Rs < 1407) {
return 8;
}
else if (Rs >= 1407) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 73) {
if (Cs < 277) {
if (ssDCval < 106) {
if (mu_mv_mag_sq < 117) {
if (SC < 1290) {
if (RsCsDiff < 2) {
if (ssDCval < 102) {
if (mu_mv_mag_sq < 26) {
if (MVDiff < 4) {
if (Cs < 209) {
return 8;
}
else if (Cs >= 209) {
return 4;
}
}
else if (MVDiff >= 4) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (Rs < 321) {
return 1;
}
else if (Rs >= 321) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (mu_mv_mag_sq < 91) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Cs < 76) {
return 1;
}
else if (Cs >= 76) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
if (Rs < 757) {
return 2;
}
else if (Rs >= 757) {
return 1;
}
}
}
}
}
else if (ssDCval >= 102) {
return 1;
}
}
else if (RsCsDiff >= 2) {
if (tcor < 63) {
if (Rs < 451) {
if (Rs < 326) {
return 8;
}
else if (Rs >= 326) {
if (Cs < 96) {
return 2;
}
else if (Cs >= 96) {
if (Cs < 103) {
return 8;
}
else if (Cs >= 103) {
return 2;
}
}
}
}
else if (Rs >= 451) {
return 8;
}
}
else if (tcor >= 63) {
if (CsDiff < 1) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
if (Rs < 372) {
return 8;
}
else if (Rs >= 372) {
if (Cs < 128) {
return 1;
}
else if (Cs >= 128) {
if (Rs < 500) {
return 1;
}
else if (Rs >= 500) {
return 4;
}
}
}
}
}
else if (CsDiff >= 1) {
if (posBalance < 60) {
if (gchDC < 3) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (Cs < 221) {
if (tcor < 90) {
if (Cs < 105) {
return 1;
}
else if (Cs >= 105) {
return 4;
}
}
else if (tcor >= 90) {
if (MVDiff < 26) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 1;
}
}
else if (MVDiff >= 26) {
return 4;
}
}
}
else if (Cs >= 221) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
else if (gchDC >= 3) {
if (gchDC < 4) {
if (MVDiff < 19) {
return 2;
}
else if (MVDiff >= 19) {
if (Rs < 487) {
return 4;
}
else if (Rs >= 487) {
return 1;
}
}
}
else if (gchDC >= 4) {
if (Cs < 272) {
return 8;
}
else if (Cs >= 272) {
return 2;
}
}
}
}
else if (posBalance >= 60) {
if (Rs < 432) {
if (diffTSC < -1) {
if (Rs < 333) {
return 1;
}
else if (Rs >= 333) {
return 2;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (Rs >= 432) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
return 2;
}
}
}
}
}
}
}
else if (SC >= 1290) {
if (ssDCval < 102) {
if (Rs < 2465) {
if (diffMVdiffVal < -163) {
return 8;
}
else if (diffMVdiffVal >= -163) {
return 1;
}
}
else if (Rs >= 2465) {
return 8;
}
}
else if (ssDCval >= 102) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (ssDCval < 79) {
if (Cs < 210) {
return 8;
}
else if (Cs >= 210) {
return 1;
}
}
else if (ssDCval >= 79) {
if (posBalance < 46) {
if (negBalance < 23) {
if (SC < 507) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (SC >= 507) {
if (Cs < 251) {
if (Rs < 313) {
if (Rs < 311) {
return 2;
}
else if (Rs >= 311) {
return 1;
}
}
else if (Rs >= 313) {
return 2;
}
}
else if (Cs >= 251) {
if (Rs < 960) {
return 1;
}
else if (Rs >= 960) {
return 2;
}
}
}
}
else if (negBalance >= 23) {
if (Rs < 981) {
if (Cs < 196) {
return 2;
}
else if (Cs >= 196) {
return 1;
}
}
else if (Rs >= 981) {
return 2;
}
}
}
else if (posBalance >= 46) {
if (Rs < 397) {
return 4;
}
else if (Rs >= 397) {
return 1;
}
}
}
}
}
else if (ssDCval >= 106) {
if (MVDiff < 33) {
if (RsCsDiff < 7) {
if (SC < 620) {
if (ssDCval < 110) {
return 1;
}
else if (ssDCval >= 110) {
return 8;
}
}
else if (SC >= 620) {
return 8;
}
}
else if (RsCsDiff >= 7) {
if (Cs < 193) {
if (Cs < 110) {
return 1;
}
else if (Cs >= 110) {
return 4;
}
}
else if (Cs >= 193) {
if (Rs < 384) {
return 2;
}
else if (Rs >= 384) {
return 8;
}
}
}
}
else if (MVDiff >= 33) {
if (Rs < 447) {
if (Cs < 60) {
return 1;
}
else if (Cs >= 60) {
return 2;
}
}
else if (Rs >= 447) {
if (posBalance < 9) {
return 8;
}
else if (posBalance >= 9) {
if (Rs < 474) {
if (Rs < 463) {
return 1;
}
else if (Rs >= 463) {
return 8;
}
}
else if (Rs >= 474) {
return 1;
}
}
}
}
}
}
else if (Cs >= 277) {
if (refDCval < 115) {
if (SC < 1924) {
if (mu_mv_mag_sq < 75) {
if (mcTcor < 94) {
if (mcTcor < 85) {
if (ssDCval < 107) {
if (gchDC < 1) {
if (Cs < 512) {
if (tcor < 74) {
if (Rs < 480) {
if (Rs < 410) {
return 8;
}
else if (Rs >= 410) {
return 2;
}
}
else if (Rs >= 480) {
return 8;
}
}
else if (tcor >= 74) {
return 2;
}
}
else if (Cs >= 512) {
return 2;
}
}
else if (gchDC >= 1) {
if (refDCval < 80) {
if (Cs < 406) {
return 2;
}
else if (Cs >= 406) {
return 4;
}
}
else if (refDCval >= 80) {
return 2;
}
}
}
else if (ssDCval >= 107) {
return 8;
}
}
else if (mcTcor >= 85) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 2) {
if (tcor < 87) {
if (Rs < 404) {
return 1;
}
else if (Rs >= 404) {
return 2;
}
}
else if (tcor >= 87) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 409) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
else if (Cs >= 409) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
}
else if (posBalance >= 54) {
if (negBalance < 34) {
if (AFD < 8) {
if (Rs < 428) {
return 8;
}
else if (Rs >= 428) {
return 1;
}
}
else if (AFD >= 8) {
return 2;
}
}
else if (negBalance >= 34) {
if (AFD < 10) {
if (diffRsCsdiff < 6) {
if (Rs < 306) {
if (Rs < 305) {
return 8;
}
else if (Rs >= 305) {
return 2;
}
}
else if (Rs >= 306) {
return 8;
}
}
else if (diffRsCsdiff >= 6) {
return 1;
}
}
else if (AFD >= 10) {
if (ssDCval < 104) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (negBalance < 49) {
if (posBalance < 69) {
return 1;
}
else if (posBalance >= 69) {
return 8;
}
}
else if (negBalance >= 49) {
return 8;
}
}
}
else if (ssDCval >= 104) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (tcor < 98) {
if (TSC < 7) {
if (mcTcor < 97) {
if (MVDiff < 4) {
if (SC < 1096) {
if (posBalance < 51) {
if (Rs < 535) {
return 1;
}
else if (Rs >= 535) {
return 2;
}
}
else if (posBalance >= 51) {
return 1;
}
}
else if (SC >= 1096) {
if (tcor < 90) {
return 2;
}
else if (tcor >= 90) {
if (Cs < 515) {
return 1;
}
else if (Cs >= 515) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 705) {
if (gchDC < 2) {
if (posBalance < 70) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (gchDC < 1) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (gchDC >= 1) {
if (SC < 968) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 1;
}
}
else if (SC >= 968) {
return 8;
}
}
}
}
else if (posBalance >= 70) {
return 2;
}
}
else if (gchDC >= 2) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
return 2;
}
}
}
else if (Rs >= 705) {
if (Cs < 458) {
return 1;
}
else if (Cs >= 458) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 310) {
if (mcTcor < 98) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (mcTcor >= 98) {
if (Rs < 639) {
return 2;
}
else if (Rs >= 639) {
return 1;
}
}
}
else if (Cs >= 310) {
if (RsCsDiff < 7) {
if (negBalance < 50) {
if (Rs < 1387) {
return 1;
}
else if (Rs >= 1387) {
return 8;
}
}
else if (negBalance >= 50) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 393) {
return 8;
}
else if (Rs >= 393) {
return 1;
}
}
}
}
else if (RsCsDiff >= 7) {
if (diffMVdiffVal < -9) {
if (Rs < 581) {
return 1;
}
else if (Rs >= 581) {
return 4;
}
}
else if (diffMVdiffVal >= -9) {
return 1;
}
}
}
}
}
else if (TSC >= 7) {
if (AFD < 17) {
if (refDCval < 83) {
return 1;
}
else if (refDCval >= 83) {
return 8;
}
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (tcor >= 98) {
if (ssDCval < 94) {
if (Rs < 695) {
if (MVDiff < 10) {
if (Rs < 466) {
return 4;
}
else if (Rs >= 466) {
return 1;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (Rs >= 695) {
return 8;
}
}
else if (ssDCval >= 94) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (Cs < 495) {
return 1;
}
else if (Cs >= 495) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (ssDCval < 96) {
if (negBalance < 33) {
if (ssDCval < 94) {
if (refDCval < 81) {
if (posBalance < 47) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
if (Rs < 440) {
return 8;
}
else if (Rs >= 440) {
return 1;
}
}
}
else if (posBalance >= 47) {
if (Rs < 957) {
if (RsDiff < 2) {
if (posBalance < 48) {
if (SC < 818) {
return 1;
}
else if (SC >= 818) {
return 8;
}
}
else if (posBalance >= 48) {
return 8;
}
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (Rs >= 957) {
return 1;
}
}
}
else if (refDCval >= 81) {
if (TSC < 2) {
if (Rs < 1034) {
if (negBalance < 23) {
return 1;
}
else if (negBalance >= 23) {
return 8;
}
}
else if (Rs >= 1034) {
return 2;
}
}
else if (TSC >= 2) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
return 8;
}
}
}
}
else if (ssDCval >= 94) {
if (Rs < 869) {
return 8;
}
else if (Rs >= 869) {
return 1;
}
}
}
else if (negBalance >= 33) {
if (tcor < 97) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (MVDiff < 6) {
if (Rs < 413) {
return 8;
}
else if (Rs >= 413) {
if (ssDCval < 83) {
if (Cs < 331) {
return 1;
}
else if (Cs >= 331) {
return 4;
}
}
else if (ssDCval >= 83) {
return 1;
}
}
}
else if (MVDiff >= 6) {
return 8;
}
}
}
else if (tcor >= 97) {
if (Rs < 524) {
if (Rs < 471) {
return 8;
}
else if (Rs >= 471) {
return 1;
}
}
else if (Rs >= 524) {
if (diffRsCsdiff < -3) {
if (Cs < 500) {
return 1;
}
else if (Cs >= 500) {
return 8;
}
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
}
}
}
else if (ssDCval >= 96) {
if (refDCval < 98) {
if (Cs < 466) {
return 4;
}
else if (Cs >= 466) {
return 8;
}
}
else if (refDCval >= 98) {
if (Rs < 412) {
return 1;
}
else if (Rs >= 412) {
return 8;
}
}
}
}
}
else if (SC >= 1924) {
if (posBalance < 54) {
if (Cs < 504) {
if (mu_mv_mag_sq < 72) {
if (tcor < 98) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (tcor >= 98) {
if (Rs < 2092) {
return 8;
}
else if (Rs >= 2092) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 72) {
if (SC < 2486) {
if (ssDCval < 100) {
if (mu_mv_mag_sq < 84) {
if (negBalance < 30) {
if (Rs < 1870) {
return 1;
}
else if (Rs >= 1870) {
return 8;
}
}
else if (negBalance >= 30) {
return 1;
}
}
else if (mu_mv_mag_sq >= 84) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
else if (ssDCval >= 100) {
if (Rs < 1772) {
return 8;
}
else if (Rs >= 1772) {
return 1;
}
}
}
else if (SC >= 2486) {
if (negBalance < 31) {
if (Cs < 452) {
return 1;
}
else if (Cs >= 452) {
return 8;
}
}
else if (negBalance >= 31) {
return 1;
}
}
}
}
else if (Cs >= 504) {
if (mu_mv_mag_sq < 99) {
if (Rs < 2087) {
if (Cs < 507) {
return 8;
}
else if (Cs >= 507) {
return 1;
}
}
else if (Rs >= 2087) {
return 8;
}
}
else if (mu_mv_mag_sq >= 99) {
return 4;
}
}
}
else if (posBalance >= 54) {
if (SC < 2288) {
if (Rs < 1731) {
if (Rs < 1584) {
return 1;
}
else if (Rs >= 1584) {
return 8;
}
}
else if (Rs >= 1731) {
return 1;
}
}
else if (SC >= 2288) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 115) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 92) {
if (ssDCval < 151) {
if (Rs < 864) {
if (mcTcor < 98) {
if (MVDiff < 22) {
if (tcor < 83) {
if (Rs < 392) {
return 2;
}
else if (Rs >= 392) {
return 4;
}
}
else if (tcor >= 83) {
if (Cs < 328) {
return 2;
}
else if (Cs >= 328) {
if (Rs < 797) {
return 8;
}
else if (Rs >= 797) {
return 2;
}
}
}
}
else if (MVDiff >= 22) {
if (posBalance < 57) {
if (posBalance < 27) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Cs < 436) {
return 4;
}
else if (Cs >= 436) {
return 2;
}
}
}
else if (posBalance >= 27) {
return 2;
}
}
else if (posBalance >= 57) {
if (refDCval < 124) {
return 2;
}
else if (refDCval >= 124) {
return 4;
}
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 62) {
if (ssDCval < 132) {
return 2;
}
else if (ssDCval >= 132) {
if (Cs < 450) {
return 2;
}
else if (Cs >= 450) {
return 8;
}
}
}
else if (negBalance >= 62) {
return 8;
}
}
}
else if (Rs >= 864) {
if (ssDCval < 136) {
if (posBalance < 50) {
if (Rs < 1022) {
return 1;
}
else if (Rs >= 1022) {
return 2;
}
}
else if (posBalance >= 50) {
if (Cs < 347) {
return 8;
}
else if (Cs >= 347) {
return 1;
}
}
}
else if (ssDCval >= 136) {
return 4;
}
}
}
else if (ssDCval >= 151) {
if (Cs < 345) {
if (Rs < 393) {
return 2;
}
else if (Rs >= 393) {
return 8;
}
}
else if (Cs >= 345) {
if (Cs < 478) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (Rs < 476) {
return 1;
}
else if (Rs >= 476) {
return 4;
}
}
}
else if (Cs >= 478) {
if (Rs < 501) {
return 2;
}
else if (Rs >= 501) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (Rs < 343) {
if (Rs < 322) {
return 2;
}
else if (Rs >= 322) {
return 4;
}
}
else if (Rs >= 343) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (ssDCval < 127) {
if (Rs < 1167) {
return 4;
}
else if (Rs >= 1167) {
return 1;
}
}
else if (ssDCval >= 127) {
return 8;
}
}
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 119) {
if (Rs < 509) {
if (ssDCval < 117) {
return 1;
}
else if (ssDCval >= 117) {
return 2;
}
}
else if (Rs >= 509) {
return 2;
}
}
else if (ssDCval >= 119) {
if (TSC < 4) {
if (mu_mv_mag_sq < 109) {
if (ssDCval < 155) {
if (Rs < 375) {
if (Cs < 398) {
return 4;
}
else if (Cs >= 398) {
return 2;
}
}
else if (Rs >= 375) {
return 1;
}
}
else if (ssDCval >= 155) {
return 4;
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 707) {
return 8;
}
else if (Rs >= 707) {
return 1;
}
}
}
else if (TSC >= 4) {
if (Rs < 422) {
return 4;
}
else if (Rs >= 422) {
return 1;
}
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 75) {
if (RsCsDiff < 209) {
if (Cs < 321) {
if (mu_mv_mag_sq < 50) {
if (MVDiff < 17) {
if (Rs < 1176) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
if (Rs < 618) {
return 2;
}
else if (Rs >= 618) {
return 8;
}
}
}
else if (Rs >= 1176) {
if (AFD < 21) {
if (diffMVdiffVal < -10) {
return 1;
}
else if (diffMVdiffVal >= -10) {
return 2;
}
}
else if (AFD >= 21) {
return 8;
}
}
}
else if (MVDiff >= 17) {
if (Cs < 78) {
if (ssDCval < 97) {
return 1;
}
else if (ssDCval >= 97) {
if (AFD < 23) {
return 4;
}
else if (AFD >= 23) {
return 1;
}
}
}
else if (Cs >= 78) {
if (refDCval < 41) {
if (Cs < 91) {
return 8;
}
else if (Cs >= 91) {
return 1;
}
}
else if (refDCval >= 41) {
if (ssDCval < 125) {
if (posBalance < 51) {
if (Cs < 218) {
return 1;
}
else if (Cs >= 218) {
if (Rs < 1834) {
if (Rs < 968) {
if (Rs < 487) {
return 1;
}
else if (Rs >= 487) {
return 2;
}
}
else if (Rs >= 968) {
if (ssDCval < 46) {
return 2;
}
else if (ssDCval >= 46) {
return 1;
}
}
}
else if (Rs >= 1834) {
return 2;
}
}
}
else if (posBalance >= 51) {
if (Cs < 302) {
if (gchDC < 3) {
if (AFD < 20) {
if (refDCval < 86) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (MVDiff < 34) {
if (Rs < 500) {
return 8;
}
else if (Rs >= 500) {
return 1;
}
}
else if (MVDiff >= 34) {
return 4;
}
}
}
else if (refDCval >= 86) {
if (RsCsDiff < 28) {
if (Rs < 490) {
return 8;
}
else if (Rs >= 490) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 8;
}
}
}
else if (RsCsDiff >= 28) {
if (AFD < 18) {
if (Rs < 1716) {
return 1;
}
else if (Rs >= 1716) {
return 8;
}
}
else if (AFD >= 18) {
if (Rs < 1775) {
if (Cs < 202) {
return 8;
}
else if (Cs >= 202) {
if (Rs < 1618) {
return 1;
}
else if (Rs >= 1618) {
return 8;
}
}
}
else if (Rs >= 1775) {
return 2;
}
}
}
}
}
else if (AFD >= 20) {
if (Cs < 258) {
if (negBalance < 56) {
if (Rs < 1127) {
return 1;
}
else if (Rs >= 1127) {
return 2;
}
}
else if (negBalance >= 56) {
return 2;
}
}
else if (Cs >= 258) {
if (Rs < 1338) {
return 1;
}
else if (Rs >= 1338) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
if (diffAFD < -2) {
if (posBalance < 64) {
if (Rs < 1381) {
return 2;
}
else if (Rs >= 1381) {
return 8;
}
}
else if (posBalance >= 64) {
if (AFD < 29) {
return 1;
}
else if (AFD >= 29) {
return 2;
}
}
}
else if (diffAFD >= -2) {
if (gchDC < 13) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 1;
}
}
else if (gchDC >= 13) {
return 8;
}
}
}
}
else if (Cs >= 302) {
if (RsCsDiff < 69) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
return 4;
}
}
else if (RsCsDiff >= 69) {
if (Rs < 578) {
return 4;
}
else if (Rs >= 578) {
return 2;
}
}
}
}
}
else if (ssDCval >= 125) {
if (diffMVdiffVal < -4) {
if (Rs < 535) {
return 2;
}
else if (Rs >= 535) {
if (ssDCval < 131) {
return 2;
}
else if (ssDCval >= 131) {
return 1;
}
}
}
else if (diffMVdiffVal >= -4) {
if (tcor < 68) {
return 1;
}
else if (tcor >= 68) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 1;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (tcor < 37) {
if (AFD < 23) {
return 4;
}
else if (AFD >= 23) {
if (Rs < 469) {
return 8;
}
else if (Rs >= 469) {
return 1;
}
}
}
else if (tcor >= 37) {
if (posBalance < 63) {
if (AFD < 25) {
if (diffRsCsdiff < 8) {
if (mu_mv_mag_sq < 96) {
if (SC < 733) {
if (Rs < 327) {
return 8;
}
else if (Rs >= 327) {
return 4;
}
}
else if (SC >= 733) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 713) {
return 1;
}
else if (Rs >= 713) {
return 8;
}
}
}
else if (diffRsCsdiff >= 8) {
if (Cs < 259) {
return 1;
}
else if (Cs >= 259) {
return 2;
}
}
}
else if (AFD >= 25) {
if (SC < 1205) {
if (Rs < 419) {
return 1;
}
else if (Rs >= 419) {
return 2;
}
}
else if (SC >= 1205) {
return 1;
}
}
}
else if (posBalance >= 63) {
if (tcor < 73) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 1;
}
}
else if (tcor >= 73) {
return 1;
}
}
}
}
}
else if (Cs >= 321) {
if (mu_mv_mag_sq < 6) {
if (posBalance < 58) {
if (negBalance < 45) {
if (Rs < 891) {
return 2;
}
else if (Rs >= 891) {
return 1;
}
}
else if (negBalance >= 45) {
if (Cs < 496) {
if (Rs < 2178) {
if (AFD < 18) {
if (MVDiff < 56) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 1;
}
}
else if (MVDiff >= 56) {
if (Rs < 662) {
return 4;
}
else if (Rs >= 662) {
return 1;
}
}
}
else if (AFD >= 18) {
if (Rs < 624) {
return 2;
}
else if (Rs >= 624) {
return 4;
}
}
}
else if (Rs >= 2178) {
return 8;
}
}
else if (Cs >= 496) {
if (tcor < 65) {
if (Cs < 500) {
return 8;
}
else if (Cs >= 500) {
return 2;
}
}
else if (tcor >= 65) {
return 8;
}
}
}
}
else if (posBalance >= 58) {
if (negBalance < 37) {
if (Rs < 570) {
return 4;
}
else if (Rs >= 570) {
return 1;
}
}
else if (negBalance >= 37) {
if (MVDiff < 23) {
if (Cs < 523) {
return 8;
}
else if (Cs >= 523) {
return 4;
}
}
else if (MVDiff >= 23) {
if (RsCsDiff < 29) {
if (Rs < 581) {
return 2;
}
else if (Rs >= 581) {
return 4;
}
}
else if (RsCsDiff >= 29) {
if (posBalance < 62) {
if (negBalance < 57) {
if (Cs < 518) {
if (SC < 1109) {
return 4;
}
else if (SC >= 1109) {
return 8;
}
}
else if (Cs >= 518) {
return 4;
}
}
else if (negBalance >= 57) {
return 4;
}
}
else if (posBalance >= 62) {
if (ssDCval < 124) {
if (Rs < 816) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 1;
}
}
else if (Rs >= 816) {
return 8;
}
}
else if (ssDCval >= 124) {
if (Rs < 2258) {
return 1;
}
else if (Rs >= 2258) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 6) {
if (diffAFD < 4) {
if (posBalance < 51) {
if (TSC < 5) {
if (mcTcor < 90) {
if (gchDC < 3) {
return 8;
}
else if (gchDC >= 3) {
if (Rs < 572) {
return 1;
}
else if (Rs >= 572) {
return 2;
}
}
}
else if (mcTcor >= 90) {
return 1;
}
}
else if (TSC >= 5) {
if (mcTcor < 82) {
return 1;
}
else if (mcTcor >= 82) {
if (TSC < 6) {
if (ssDCval < 55) {
return 2;
}
else if (ssDCval >= 55) {
if (Rs < 646) {
return 4;
}
else if (Rs >= 646) {
return 1;
}
}
}
else if (TSC >= 6) {
if (MVDiff < 46) {
return 2;
}
else if (MVDiff >= 46) {
if (mcTcor < 90) {
if (Rs < 1862) {
return 2;
}
else if (Rs >= 1862) {
return 1;
}
}
else if (mcTcor >= 90) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 51) {
if (mcTcor < 89) {
if (tcor < 42) {
if (AFD < 19) {
if (MVDiff < 43) {
if (AFD < 14) {
if (Rs < 672) {
return 1;
}
else if (Rs >= 672) {
return 8;
}
}
else if (AFD >= 14) {
return 4;
}
}
else if (MVDiff >= 43) {
return 1;
}
}
else if (AFD >= 19) {
if (diffRsCsdiff < 25) {
if (diffRsCsdiff < -43) {
return 4;
}
else if (diffRsCsdiff >= -43) {
return 8;
}
}
else if (diffRsCsdiff >= 25) {
return 4;
}
}
}
else if (tcor >= 42) {
if (Rs < 1601) {
if (posBalance < 60) {
if (diffRsCsdiff < 2) {
if (Cs < 355) {
return 4;
}
else if (Cs >= 355) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
if (diffMVdiffVal < 4) {
return 2;
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
}
else if (posBalance >= 60) {
if (TSC < 9) {
if (CsDiff < 3) {
if (Cs < 522) {
return 2;
}
else if (Cs >= 522) {
return 8;
}
}
else if (CsDiff >= 3) {
if (AFD < 16) {
if (ssDCval < 44) {
if (Rs < 619) {
return 8;
}
else if (Rs >= 619) {
return 2;
}
}
else if (ssDCval >= 44) {
return 4;
}
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (TSC >= 9) {
if (ssDCval < 88) {
if (Rs < 679) {
return 4;
}
else if (Rs >= 679) {
return 8;
}
}
else if (ssDCval >= 88) {
return 2;
}
}
}
}
else if (Rs >= 1601) {
if (RsCsDiff < 176) {
return 4;
}
else if (RsCsDiff >= 176) {
return 8;
}
}
}
}
else if (mcTcor >= 89) {
if (Cs < 491) {
if (posBalance < 60) {
if (ssDCval < 106) {
if (RsCsDiff < 46) {
if (diffMVdiffVal < 17) {
return 1;
}
else if (diffMVdiffVal >= 17) {
return 8;
}
}
else if (RsCsDiff >= 46) {
if (Cs < 362) {
return 1;
}
else if (Cs >= 362) {
if (Rs < 1821) {
return 4;
}
else if (Rs >= 1821) {
return 1;
}
}
}
}
else if (ssDCval >= 106) {
if (Rs < 755) {
return 2;
}
else if (Rs >= 755) {
return 8;
}
}
}
else if (posBalance >= 60) {
if (gchDC < 4) {
if (diffAFD < 1) {
if (ssDCval < 92) {
if (Cs < 481) {
return 2;
}
else if (Cs >= 481) {
return 1;
}
}
else if (ssDCval >= 92) {
return 1;
}
}
else if (diffAFD >= 1) {
if (Rs < 541) {
return 4;
}
else if (Rs >= 541) {
return 1;
}
}
}
else if (gchDC >= 4) {
if (CsDiff < 4) {
if (Rs < 847) {
return 8;
}
else if (Rs >= 847) {
return 1;
}
}
else if (CsDiff >= 4) {
if (Rs < 397) {
return 8;
}
else if (Rs >= 397) {
return 2;
}
}
}
}
}
else if (Cs >= 491) {
if (MVDiff < 36) {
if (diffRsCsdiff < 12) {
return 4;
}
else if (diffRsCsdiff >= 12) {
if (Rs < 504) {
return 8;
}
else if (Rs >= 504) {
return 1;
}
}
}
else if (MVDiff >= 36) {
if (Cs < 519) {
return 8;
}
else if (Cs >= 519) {
return 2;
}
}
}
}
}
}
else if (diffAFD >= 4) {
if (MVDiff < 147) {
if (Rs < 2005) {
if (Cs < 469) {
if (AFD < 24) {
if (diffMVdiffVal < 30) {
if (Cs < 453) {
return 1;
}
else if (Cs >= 453) {
if (Rs < 1108) {
return 8;
}
else if (Rs >= 1108) {
return 1;
}
}
}
else if (diffMVdiffVal >= 30) {
return 8;
}
}
else if (AFD >= 24) {
if (diffRsCsdiff < -16) {
if (Cs < 412) {
return 1;
}
else if (Cs >= 412) {
return 4;
}
}
else if (diffRsCsdiff >= -16) {
return 1;
}
}
}
else if (Cs >= 469) {
if (Rs < 735) {
return 1;
}
else if (Rs >= 735) {
if (Rs < 1103) {
return 8;
}
else if (Rs >= 1103) {
return 2;
}
}
}
}
else if (Rs >= 2005) {
return 2;
}
}
else if (MVDiff >= 147) {
if (Rs < 1006) {
if (mcTcor < 88) {
if (mu_mv_mag_sq < 46) {
return 4;
}
else if (mu_mv_mag_sq >= 46) {
if (Rs < 570) {
return 2;
}
else if (Rs >= 570) {
return 1;
}
}
}
else if (mcTcor >= 88) {
return 8;
}
}
else if (Rs >= 1006) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 209) {
if (MVDiff < 29) {
return 2;
}
else if (MVDiff >= 29) {
if (mu_mv_mag_sq < 61) {
if (tcor < 43) {
if (Rs < 2053) {
if (Rs < 1645) {
return 2;
}
else if (Rs >= 1645) {
if (SC < 2246) {
return 4;
}
else if (SC >= 2246) {
return 2;
}
}
}
else if (Rs >= 2053) {
return 8;
}
}
else if (tcor >= 43) {
if (AFD < 23) {
if (Cs < 133) {
return 1;
}
else if (Cs >= 133) {
if (Cs < 406) {
return 2;
}
else if (Cs >= 406) {
return 8;
}
}
}
else if (AFD >= 23) {
if (posBalance < 54) {
if (AFD < 46) {
return 1;
}
else if (AFD >= 46) {
return 8;
}
}
else if (posBalance >= 54) {
if (diffAFD < 5) {
if (diffAFD < 2) {
if (diffMVdiffVal < -6) {
if (MVDiff < 137) {
return 8;
}
else if (MVDiff >= 137) {
return 2;
}
}
else if (diffMVdiffVal >= -6) {
if (negBalance < 60) {
return 1;
}
else if (negBalance >= 60) {
return 2;
}
}
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (diffAFD >= 5) {
if (AFD < 28) {
return 2;
}
else if (AFD >= 28) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (AFD < 38) {
return 2;
}
else if (AFD >= 38) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
return 1;
}
}
}
}
}
}
else if (tcor >= 75) {
if (MVDiff < 9) {
if (posBalance < 52) {
if (negBalance < 31) {
if (ssDCval < 58) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
else if (ssDCval >= 58) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
return 8;
}
}
}
else if (negBalance >= 31) {
if (negBalance < 51) {
if (ssDCval < 67) {
if (diffRsCsdiff < 4) {
if (Rs < 1434) {
return 8;
}
else if (Rs >= 1434) {
if (Rs < 2104) {
return 2;
}
else if (Rs >= 2104) {
return 4;
}
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 2020) {
return 1;
}
else if (Rs >= 2020) {
return 2;
}
}
}
else if (ssDCval >= 67) {
if (ssDCval < 96) {
if (Rs < 954) {
return 4;
}
else if (Rs >= 954) {
return 1;
}
}
else if (ssDCval >= 96) {
if (TSC < 4) {
if (SC < 2211) {
return 2;
}
else if (SC >= 2211) {
return 8;
}
}
else if (TSC >= 4) {
return 1;
}
}
}
}
else if (negBalance >= 51) {
if (ssDCval < 106) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 2;
}
}
else if (ssDCval >= 106) {
if (Rs < 1768) {
return 1;
}
else if (Rs >= 1768) {
return 8;
}
}
}
}
}
else if (posBalance >= 52) {
if (Rs < 1537) {
if (ssDCval < 134) {
if (Cs < 157) {
return 2;
}
else if (Cs >= 157) {
if (mcTcor < 96) {
if (negBalance < 48) {
if (mcTcor < 90) {
if (Cs < 483) {
return 4;
}
else if (Cs >= 483) {
if (Rs < 1071) {
return 8;
}
else if (Rs >= 1071) {
return 2;
}
}
}
else if (mcTcor >= 90) {
if (Rs < 670) {
return 2;
}
else if (Rs >= 670) {
return 1;
}
}
}
else if (negBalance >= 48) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (Rs < 662) {
if (Cs < 334) {
return 1;
}
else if (Cs >= 334) {
return 4;
}
}
else if (Rs >= 662) {
return 1;
}
}
}
}
else if (ssDCval >= 134) {
return 4;
}
}
else if (Rs >= 1537) {
if (negBalance < 34) {
if (TSC < 5) {
if (Rs < 2399) {
return 2;
}
else if (Rs >= 2399) {
return 8;
}
}
else if (TSC >= 5) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (negBalance >= 34) {
if (tcor < 91) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
return 1;
}
}
else if (tcor >= 91) {
if (ssDCval < 83) {
return 2;
}
else if (ssDCval >= 83) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 9) {
if (RsDiff < 9) {
if (Rs < 820) {
if (CsDiff < 2) {
if (refDCval < 79) {
if (diffAFD < -5) {
if (Cs < 160) {
if (Cs < 52) {
return 2;
}
else if (Cs >= 52) {
return 1;
}
}
else if (Cs >= 160) {
return 8;
}
}
else if (diffAFD >= -5) {
return 1;
}
}
else if (refDCval >= 79) {
if (negBalance < 55) {
if (Rs < 781) {
return 1;
}
else if (Rs >= 781) {
if (Cs < 129) {
return 1;
}
else if (Cs >= 129) {
return 4;
}
}
}
else if (negBalance >= 55) {
if (MVDiff < 41) {
return 1;
}
else if (MVDiff >= 41) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
}
}
}
else if (CsDiff >= 2) {
if (Cs < 331) {
if (refDCval < 70) {
if (Rs < 646) {
if (posBalance < 55) {
if (Cs < 256) {
return 2;
}
else if (Cs >= 256) {
return 1;
}
}
else if (posBalance >= 55) {
return 4;
}
}
else if (Rs >= 646) {
return 1;
}
}
else if (refDCval >= 70) {
if (mu_mv_mag_sq < 12) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (mu_mv_mag_sq >= 12) {
if (mu_mv_mag_sq < 24) {
return 8;
}
else if (mu_mv_mag_sq >= 24) {
if (Rs < 472) {
if (SC < 559) {
if (Rs < 343) {
return 8;
}
else if (Rs >= 343) {
return 2;
}
}
else if (SC >= 559) {
return 1;
}
}
else if (Rs >= 472) {
return 2;
}
}
}
}
}
else if (Cs >= 331) {
if (mu_mv_mag_sq < 14) {
if (diffAFD < -1) {
if (negBalance < 48) {
if (Cs < 400) {
return 1;
}
else if (Cs >= 400) {
return 2;
}
}
else if (negBalance >= 48) {
if (negBalance < 54) {
if (Rs < 642) {
return 8;
}
else if (Rs >= 642) {
return 4;
}
}
else if (negBalance >= 54) {
if (diffMVdiffVal < -11) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
return 8;
}
}
else if (diffMVdiffVal >= -11) {
return 8;
}
}
}
}
else if (diffAFD >= -1) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (Cs < 477) {
if (Cs < 376) {
return 8;
}
else if (Cs >= 376) {
return 1;
}
}
else if (Cs >= 477) {
if (Rs < 630) {
return 4;
}
else if (Rs >= 630) {
if (Rs < 724) {
return 2;
}
else if (Rs >= 724) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (diffMVdiffVal < -38) {
if (Rs < 584) {
return 4;
}
else if (Rs >= 584) {
return 2;
}
}
else if (diffMVdiffVal >= -38) {
if (SC < 1145) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (AFD < 11) {
if (SC < 1085) {
return 8;
}
else if (SC >= 1085) {
return 1;
}
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (SC >= 1145) {
if (negBalance < 30) {
if (Rs < 795) {
return 2;
}
else if (Rs >= 795) {
return 8;
}
}
else if (negBalance >= 30) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 820) {
if (ssDCval < 139) {
if (Cs < 254) {
if (refDCval < 79) {
if (Rs < 1412) {
return 1;
}
else if (Rs >= 1412) {
if (Rs < 1584) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 2;
}
}
else if (Rs >= 1584) {
return 1;
}
}
}
else if (refDCval >= 79) {
if (MVDiff < 50) {
if (Cs < 164) {
if (diffRsCsdiff < -6) {
if (ssDCval < 86) {
if (Rs < 1640) {
return 1;
}
else if (Rs >= 1640) {
return 2;
}
}
else if (ssDCval >= 86) {
if (Rs < 2248) {
return 2;
}
else if (Rs >= 2248) {
return 1;
}
}
}
else if (diffRsCsdiff >= -6) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
if (ssDCval < 103) {
return 1;
}
else if (ssDCval >= 103) {
return 2;
}
}
}
}
else if (Cs >= 164) {
if (Rs < 1126) {
if (Rs < 1102) {
if (Cs < 190) {
return 2;
}
else if (Cs >= 190) {
return 1;
}
}
else if (Rs >= 1102) {
return 2;
}
}
else if (Rs >= 1126) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
return 1;
}
}
}
}
else if (MVDiff >= 50) {
if (ssDCval < 83) {
if (diffRsCsdiff < -30) {
return 2;
}
else if (diffRsCsdiff >= -30) {
return 1;
}
}
else if (ssDCval >= 83) {
return 1;
}
}
}
}
else if (Cs >= 254) {
if (SC < 2322) {
if (Cs < 462) {
if (negBalance < 24) {
return 8;
}
else if (negBalance >= 24) {
if (Rs < 1925) {
if (tcor < 84) {
if (MVDiff < 23) {
if (posBalance < 39) {
return 2;
}
else if (posBalance >= 39) {
if (SC < 1572) {
if (RsCsDiff < 28) {
return 4;
}
else if (RsCsDiff >= 28) {
if (Cs < 403) {
return 1;
}
else if (Cs >= 403) {
if (Cs < 444) {
return 4;
}
else if (Cs >= 444) {
return 1;
}
}
}
}
else if (SC >= 1572) {
return 1;
}
}
}
else if (MVDiff >= 23) {
return 1;
}
}
else if (tcor >= 84) {
if (negBalance < 56) {
if (diffTSC < -4) {
return 8;
}
else if (diffTSC >= -4) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (ssDCval < 67) {
if (Rs < 1125) {
return 1;
}
else if (Rs >= 1125) {
return 2;
}
}
else if (ssDCval >= 67) {
return 1;
}
}
}
}
else if (negBalance >= 56) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Cs < 430) {
return 2;
}
else if (Cs >= 430) {
return 1;
}
}
}
}
}
else if (Rs >= 1925) {
if (Rs < 1979) {
return 2;
}
else if (Rs >= 1979) {
if (Rs < 2019) {
return 8;
}
else if (Rs >= 2019) {
return 4;
}
}
}
}
}
else if (Cs >= 462) {
if (tcor < 90) {
if (posBalance < 47) {
if (diffTSC < 1) {
if (diffMVdiffVal < -10) {
return 1;
}
else if (diffMVdiffVal >= -10) {
return 2;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (posBalance >= 47) {
if (Cs < 496) {
if (Cs < 472) {
if (Rs < 1524) {
return 4;
}
else if (Rs >= 1524) {
if (Rs < 1681) {
return 2;
}
else if (Rs >= 1681) {
return 1;
}
}
}
else if (Cs >= 472) {
if (mu_mv_mag_sq < 36) {
return 1;
}
else if (mu_mv_mag_sq >= 36) {
return 8;
}
}
}
else if (Cs >= 496) {
if (tcor < 87) {
if (Rs < 1127) {
if (refDCval < 62) {
if (Rs < 900) {
return 2;
}
else if (Rs >= 900) {
return 8;
}
}
else if (refDCval >= 62) {
if (Cs < 507) {
return 8;
}
else if (Cs >= 507) {
return 4;
}
}
}
else if (Rs >= 1127) {
return 2;
}
}
else if (tcor >= 87) {
if (Cs < 531) {
return 8;
}
else if (Cs >= 531) {
return 1;
}
}
}
}
}
else if (tcor >= 90) {
if (ssDCval < 120) {
if (posBalance < 62) {
if (posBalance < 14) {
return 2;
}
else if (posBalance >= 14) {
return 1;
}
}
else if (posBalance >= 62) {
if (Cs < 518) {
return 1;
}
else if (Cs >= 518) {
return 8;
}
}
}
else if (ssDCval >= 120) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 1;
}
}
}
}
}
else if (SC >= 2322) {
if (gchDC < 2) {
if (negBalance < 45) {
if (negBalance < 31) {
return 4;
}
else if (negBalance >= 31) {
if (mcTcor < 97) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 1;
}
}
else if (mcTcor >= 97) {
if (Rs < 2464) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < 18) {
return 4;
}
else if (diffRsCsdiff >= 18) {
if (Rs < 2311) {
return 1;
}
else if (Rs >= 2311) {
return 4;
}
}
}
}
else if (Rs >= 2464) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (Cs < 355) {
return 2;
}
else if (Cs >= 355) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
if (TSC < 4) {
if (Rs < 2183) {
return 1;
}
else if (Rs >= 2183) {
return 8;
}
}
else if (TSC >= 4) {
if (AFD < 23) {
if (RsCsDiff < 28) {
if (Rs < 2502) {
return 1;
}
else if (Rs >= 2502) {
return 8;
}
}
else if (RsCsDiff >= 28) {
return 1;
}
}
else if (AFD >= 23) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 139) {
if (gchDC < 2) {
if (SC < 1845) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
if (Cs < 177) {
return 2;
}
else if (Cs >= 177) {
return 1;
}
}
}
else if (SC >= 1845) {
if (MVDiff < 68) {
return 1;
}
else if (MVDiff >= 68) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (tcor < 83) {
if (diffMVdiffVal < 30) {
if (Rs < 1235) {
return 2;
}
else if (Rs >= 1235) {
return 8;
}
}
else if (diffMVdiffVal >= 30) {
return 1;
}
}
else if (tcor >= 83) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 111) {
if (MVDiff < 34) {
if (negBalance < 37) {
if (Cs < 351) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (Cs >= 351) {
return 1;
}
}
else if (negBalance >= 37) {
if (ssDCval < 61) {
if (MVDiff < 32) {
return 1;
}
else if (MVDiff >= 32) {
return 2;
}
}
else if (ssDCval >= 61) {
if (ssDCval < 77) {
if (Cs < 91) {
return 1;
}
else if (Cs >= 91) {
return 2;
}
}
else if (ssDCval >= 77) {
if (negBalance < 60) {
if (MVDiff < 27) {
if (mu_mv_mag_sq < 20) {
if (Rs < 2403) {
if (Rs < 1640) {
return 2;
}
else if (Rs >= 1640) {
if (RsCsDiff < 113) {
return 1;
}
else if (RsCsDiff >= 113) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (diffAFD < 4) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 1;
}
}
else if (diffAFD >= 4) {
return 2;
}
}
}
}
}
else if (Rs >= 2403) {
return 2;
}
}
else if (mu_mv_mag_sq >= 20) {
return 1;
}
}
else if (MVDiff >= 27) {
return 2;
}
}
else if (negBalance >= 60) {
return 2;
}
}
}
}
}
else if (MVDiff >= 34) {
if (RsCsDiff < 223) {
if (TSC < 6) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 2;
}
}
else if (TSC >= 6) {
if (Cs < 211) {
return 1;
}
else if (Cs >= 211) {
if (posBalance < 69) {
if (RsDiff < 10) {
if (Rs < 947) {
return 8;
}
else if (Rs >= 947) {
return 1;
}
}
else if (RsDiff >= 10) {
if (MVDiff < 53) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 1972) {
return 8;
}
else if (Rs >= 1972) {
return 2;
}
}
}
else if (MVDiff >= 53) {
if (Cs < 487) {
if (Cs < 231) {
return 2;
}
else if (Cs >= 231) {
if (diffMVdiffVal < 17) {
if (Rs < 2035) {
if (Rs < 1662) {
return 1;
}
else if (Rs >= 1662) {
return 8;
}
}
else if (Rs >= 2035) {
return 1;
}
}
else if (diffMVdiffVal >= 17) {
return 8;
}
}
}
else if (Cs >= 487) {
return 2;
}
}
}
}
else if (posBalance >= 69) {
if (Rs < 2388) {
return 1;
}
else if (Rs >= 2388) {
if (Rs < 2475) {
return 4;
}
else if (Rs >= 2475) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 223) {
if (diffMVdiffVal < -24) {
if (Rs < 2524) {
return 4;
}
else if (Rs >= 2524) {
return 1;
}
}
else if (diffMVdiffVal >= -24) {
if (AFD < 31) {
if (refDCval < 76) {
return 1;
}
else if (refDCval >= 76) {
return 2;
}
}
else if (AFD >= 31) {
return 1;
}
}
}
}
}
else if (refDCval >= 111) {
if (CsDiff < 3) {
if (mu_mv_mag_sq < 67) {
if (MVDiff < 61) {
if (Cs < 162) {
if (Rs < 1513) {
return 4;
}
else if (Rs >= 1513) {
return 2;
}
}
else if (Cs >= 162) {
return 1;
}
}
else if (MVDiff >= 61) {
if (refDCval < 121) {
if (negBalance < 47) {
return 1;
}
else if (negBalance >= 47) {
if (AFD < 19) {
if (Cs < 352) {
return 1;
}
else if (Cs >= 352) {
if (Cs < 357) {
return 8;
}
else if (Cs >= 357) {
return 1;
}
}
}
else if (AFD >= 19) {
return 8;
}
}
}
else if (refDCval >= 121) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (SC < 2656) {
return 1;
}
else if (SC >= 2656) {
if (Rs < 2280) {
return 1;
}
else if (Rs >= 2280) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 67) {
return 2;
}
}
else if (CsDiff >= 3) {
if (diffMVdiffVal < 58) {
if (mcTcor < 96) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Cs < 498) {
return 1;
}
else if (Cs >= 498) {
if (Cs < 500) {
return 8;
}
else if (Cs >= 500) {
return 1;
}
}
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (diffMVdiffVal >= 58) {
if (RsDiff < 10) {
return 2;
}
else if (RsDiff >= 10) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 535) {
if (RsCsDiff < 43) {
if (Rs < 959) {
if (ssDCval < 78) {
if (AFD < 1) {
if (Rs < 710) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 361) {
return 1;
}
else if (Rs >= 361) {
return 8;
}
}
}
else if (Rs >= 710) {
if (ssDCval < 61) {
if (Rs < 760) {
if (Rs < 733) {
return 8;
}
else if (Rs >= 733) {
return 2;
}
}
else if (Rs >= 760) {
return 8;
}
}
else if (ssDCval >= 61) {
return 2;
}
}
}
else if (AFD >= 1) {
if (ssDCval < 64) {
if (Rs < 534) {
if (Cs < 894) {
if (SC < 1009) {
if (MVDiff < 5) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
return 1;
}
}
else if (MVDiff >= 5) {
if (tcor < 94) {
if (AFD < 14) {
if (Rs < 352) {
return 1;
}
else if (Rs >= 352) {
if (SC < 976) {
if (Cs < 542) {
return 4;
}
else if (Cs >= 542) {
return 2;
}
}
else if (SC >= 976) {
if (Rs < 400) {
return 1;
}
else if (Rs >= 400) {
return 8;
}
}
}
}
else if (AFD >= 14) {
if (Rs < 261) {
if (Rs < 192) {
return 4;
}
else if (Rs >= 192) {
return 2;
}
}
else if (Rs >= 261) {
if (Cs < 560) {
return 1;
}
else if (Cs >= 560) {
return 8;
}
}
}
}
else if (tcor >= 94) {
if (Cs < 636) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
return 1;
}
}
else if (Cs >= 636) {
return 1;
}
}
}
}
else if (SC >= 1009) {
if (RsDiff < 1) {
if (SC < 1117) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Rs < 432) {
return 1;
}
else if (Rs >= 432) {
return 4;
}
}
}
else if (SC >= 1117) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (TSC < 3) {
if (Rs < 520) {
return 4;
}
else if (Rs >= 520) {
if (Rs < 526) {
return 2;
}
else if (Rs >= 526) {
return 4;
}
}
}
else if (TSC >= 3) {
if (Rs < 360) {
return 1;
}
else if (Rs >= 360) {
return 8;
}
}
}
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 30) {
if (tcor < 82) {
if (refDCval < 40) {
if (Rs < 398) {
return 8;
}
else if (Rs >= 398) {
return 2;
}
}
else if (refDCval >= 40) {
if (tcor < 47) {
return 4;
}
else if (tcor >= 47) {
if (ssDCval < 45) {
return 4;
}
else if (ssDCval >= 45) {
if (diffRsCsdiff < -1) {
if (Rs < 472) {
return 4;
}
else if (Rs >= 472) {
return 8;
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
}
}
}
else if (tcor >= 82) {
if (negBalance < 37) {
if (Rs < 381) {
return 2;
}
else if (Rs >= 381) {
return 4;
}
}
else if (negBalance >= 37) {
if (mcTcor < 93) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
else if (mcTcor >= 93) {
if (Cs < 749) {
return 2;
}
else if (Cs >= 749) {
if (Cs < 837) {
return 1;
}
else if (Cs >= 837) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (negBalance < 31) {
if (Cs < 611) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
return 2;
}
}
else if (Cs >= 611) {
if (negBalance < 25) {
if (tcor < 82) {
if (Rs < 519) {
return 1;
}
else if (Rs >= 519) {
return 2;
}
}
else if (tcor >= 82) {
return 4;
}
}
else if (negBalance >= 25) {
if (Rs < 462) {
return 1;
}
else if (Rs >= 462) {
return 4;
}
}
}
}
else if (negBalance >= 31) {
if (diffRsCsdiff < -15) {
return 8;
}
else if (diffRsCsdiff >= -15) {
if (ssDCval < 61) {
if (diffRsCsdiff < 1) {
return 1;
}
else if (diffRsCsdiff >= 1) {
if (Cs < 606) {
return 1;
}
else if (Cs >= 606) {
if (Rs < 489) {
return 2;
}
else if (Rs >= 489) {
return 4;
}
}
}
}
else if (ssDCval >= 61) {
if (SC < 1110) {
if (SC < 1077) {
return 4;
}
else if (SC >= 1077) {
return 1;
}
}
else if (SC >= 1110) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 894) {
if (tcor < 95) {
if (Cs < 996) {
if (Rs < 447) {
if (Rs < 350) {
return 4;
}
else if (Rs >= 350) {
if (Rs < 421) {
return 1;
}
else if (Rs >= 421) {
return 4;
}
}
}
else if (Rs >= 447) {
return 2;
}
}
else if (Cs >= 996) {
return 1;
}
}
else if (tcor >= 95) {
if (Rs < 361) {
return 2;
}
else if (Rs >= 361) {
if (tcor < 98) {
if (ssDCval < 45) {
return 4;
}
else if (ssDCval >= 45) {
return 1;
}
}
else if (tcor >= 98) {
return 4;
}
}
}
}
}
else if (Rs >= 534) {
if (ssDCval < 45) {
if (mu_mv_mag_sq < 92) {
if (ssDCval < 25) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (ssDCval >= 25) {
if (ssDCval < 41) {
if (ssDCval < 36) {
if (mu_mv_mag_sq < 67) {
return 8;
}
else if (mu_mv_mag_sq >= 67) {
if (Rs < 871) {
return 1;
}
else if (Rs >= 871) {
return 8;
}
}
}
else if (ssDCval >= 36) {
if (mcTcor < 87) {
if (ssDCval < 40) {
if (Rs < 569) {
return 2;
}
else if (Rs >= 569) {
return 8;
}
}
else if (ssDCval >= 40) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (mcTcor >= 87) {
if (CsDiff < 3) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 8;
}
}
else if (CsDiff >= 3) {
if (Rs < 884) {
if (mu_mv_mag_sq < 41) {
return 2;
}
else if (mu_mv_mag_sq >= 41) {
return 1;
}
}
else if (Rs >= 884) {
return 8;
}
}
}
}
}
else if (ssDCval >= 41) {
if (Rs < 822) {
if (mcTcor < 91) {
if (negBalance < 47) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Rs < 768) {
return 2;
}
else if (Rs >= 768) {
return 8;
}
}
}
else if (negBalance >= 47) {
if (AFD < 10) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
else if (AFD >= 10) {
if (Cs < 593) {
return 8;
}
else if (Cs >= 593) {
return 4;
}
}
}
}
else if (mcTcor >= 91) {
if (Cs < 821) {
return 4;
}
else if (Cs >= 821) {
return 8;
}
}
}
else if (Rs >= 822) {
if (diffRsCsdiff < -27) {
if (Cs < 700) {
return 4;
}
else if (Cs >= 700) {
return 8;
}
}
else if (diffRsCsdiff >= -27) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (AFD < 3) {
if (ssDCval < 36) {
if (ssDCval < 29) {
return 8;
}
else if (ssDCval >= 29) {
if (refDCval < 35) {
return 1;
}
else if (refDCval >= 35) {
return 8;
}
}
}
else if (ssDCval >= 36) {
if (ssDCval < 38) {
return 2;
}
else if (ssDCval >= 38) {
return 8;
}
}
}
else if (AFD >= 3) {
if (ssDCval < 37) {
return 1;
}
else if (ssDCval >= 37) {
return 8;
}
}
}
}
else if (ssDCval >= 45) {
if (RsDiff < 2) {
if (SC < 1491) {
if (refDCval < 53) {
if (SC < 1230) {
return 4;
}
else if (SC >= 1230) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (MVDiff < 22) {
if (mu_mv_mag_sq < 117) {
return 1;
}
else if (mu_mv_mag_sq >= 117) {
return 8;
}
}
else if (MVDiff >= 22) {
return 8;
}
}
}
}
else if (refDCval >= 53) {
if (MVDiff < 2) {
if (ssDCval < 55) {
if (Rs < 657) {
return 8;
}
else if (Rs >= 657) {
return 4;
}
}
else if (ssDCval >= 55) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (mcTcor < 96) {
if (Cs < 660) {
return 8;
}
else if (Cs >= 660) {
return 4;
}
}
else if (mcTcor >= 96) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (negBalance < 30) {
if (Rs < 751) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
return 8;
}
}
else if (Rs >= 751) {
if (Rs < 891) {
return 1;
}
else if (Rs >= 891) {
return 8;
}
}
}
else if (negBalance >= 30) {
return 8;
}
}
}
}
else if (SC >= 1491) {
if (Rs < 723) {
if (negBalance < 26) {
if (mu_mv_mag_sq < 123) {
if (Cs < 1099) {
return 1;
}
else if (Cs >= 1099) {
return 2;
}
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
else if (negBalance >= 26) {
if (ssDCval < 59) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
return 4;
}
}
else if (ssDCval >= 59) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (Rs < 708) {
return 8;
}
else if (Rs >= 708) {
return 4;
}
}
}
}
}
else if (Rs >= 723) {
if (negBalance < 41) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (mcTcor < 96) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Rs < 843) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (MVDiff < 4) {
if (Cs < 898) {
return 8;
}
else if (Cs >= 898) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
if (Rs < 749) {
return 1;
}
else if (Rs >= 749) {
return 8;
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 773) {
return 4;
}
else if (Rs >= 773) {
return 1;
}
}
}
}
else if (Rs >= 843) {
if (negBalance < 12) {
return 8;
}
else if (negBalance >= 12) {
if (Rs < 956) {
if (Cs < 587) {
if (Rs < 948) {
return 8;
}
else if (Rs >= 948) {
return 1;
}
}
else if (Cs >= 587) {
if (ssDCval < 52) {
if (Cs < 1061) {
return 8;
}
else if (Cs >= 1061) {
return 1;
}
}
else if (ssDCval >= 52) {
return 1;
}
}
}
else if (Rs >= 956) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 41) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
return 4;
}
}
}
}
}
else if (RsDiff >= 2) {
if (refDCval < 52) {
if (mcTcor < 93) {
if (MVDiff < 19) {
if (Cs < 823) {
if (mu_mv_mag_sq < 2) {
if (negBalance < 54) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
if (Cs < 766) {
if (Rs < 701) {
return 8;
}
else if (Rs >= 701) {
if (Rs < 926) {
return 1;
}
else if (Rs >= 926) {
return 8;
}
}
}
else if (Cs >= 766) {
return 8;
}
}
}
else if (negBalance >= 54) {
if (RsCsDiff < 22) {
if (Rs < 808) {
return 8;
}
else if (Rs >= 808) {
return 1;
}
}
else if (RsCsDiff >= 22) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 808) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 4;
}
}
else if (Rs >= 808) {
if (tcor < 77) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (tcor >= 77) {
if (negBalance < 47) {
return 8;
}
else if (negBalance >= 47) {
if (Cs < 626) {
return 8;
}
else if (Cs >= 626) {
return 1;
}
}
}
}
}
}
else if (Cs >= 823) {
if (Rs < 574) {
return 4;
}
else if (Rs >= 574) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 909) {
return 1;
}
else if (Rs >= 909) {
return 8;
}
}
}
}
}
else if (MVDiff >= 19) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < 11) {
if (diffRsCsdiff < -17) {
if (Rs < 687) {
return 4;
}
else if (Rs >= 687) {
return 2;
}
}
else if (diffRsCsdiff >= -17) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Rs < 762) {
return 4;
}
else if (Rs >= 762) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 11) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
else if (mcTcor >= 93) {
if (MVDiff < 4) {
if (Cs < 1042) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
return 8;
}
}
else if (Cs >= 1042) {
return 1;
}
}
else if (MVDiff >= 4) {
if (negBalance < 41) {
if (AFD < 12) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
return 4;
}
}
else if (AFD >= 12) {
return 1;
}
}
else if (negBalance >= 41) {
if (Cs < 753) {
return 1;
}
else if (Cs >= 753) {
return 8;
}
}
}
}
}
else if (refDCval >= 52) {
if (negBalance < 27) {
if (Rs < 779) {
return 4;
}
else if (Rs >= 779) {
if (tcor < 84) {
return 4;
}
else if (tcor >= 84) {
if (Cs < 751) {
return 1;
}
else if (Cs >= 751) {
return 4;
}
}
}
}
else if (negBalance >= 27) {
if (mu_mv_mag_sq < 88) {
if (negBalance < 40) {
if (MVDiff < 15) {
if (Rs < 842) {
if (TSC < 5) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
if (Cs < 629) {
if (Rs < 618) {
return 2;
}
else if (Rs >= 618) {
return 8;
}
}
else if (Cs >= 629) {
if (MVDiff < 3) {
if (Rs < 706) {
return 2;
}
else if (Rs >= 706) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Cs < 1149) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (Cs < 639) {
return 4;
}
else if (Cs >= 639) {
return 1;
}
}
}
else if (Cs >= 1149) {
if (Rs < 673) {
return 1;
}
else if (Rs >= 673) {
return 2;
}
}
}
}
}
}
else if (TSC >= 5) {
if (Rs < 729) {
if (diffRsCsdiff < 9) {
return 2;
}
else if (diffRsCsdiff >= 9) {
return 1;
}
}
else if (Rs >= 729) {
if (posBalance < 43) {
if (Rs < 776) {
return 4;
}
else if (Rs >= 776) {
return 2;
}
}
else if (posBalance >= 43) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
}
else if (Rs >= 842) {
if (Cs < 901) {
if (negBalance < 38) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (Cs >= 901) {
if (Rs < 907) {
return 8;
}
else if (Rs >= 907) {
return 4;
}
}
}
}
else if (MVDiff >= 15) {
if (SC < 1542) {
return 1;
}
else if (SC >= 1542) {
if (Rs < 800) {
return 8;
}
else if (Rs >= 800) {
if (Rs < 918) {
return 1;
}
else if (Rs >= 918) {
return 2;
}
}
}
}
}
else if (negBalance >= 40) {
if (mcTcor < 85) {
if (tcor < 81) {
return 4;
}
else if (tcor >= 81) {
return 8;
}
}
else if (mcTcor >= 85) {
if (posBalance < 42) {
if (Cs < 679) {
return 1;
}
else if (Cs >= 679) {
return 2;
}
}
else if (posBalance >= 42) {
if (mu_mv_mag_sq < 71) {
if (AFD < 11) {
if (Rs < 638) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 605) {
return 8;
}
else if (Rs >= 605) {
return 1;
}
}
}
else if (Rs >= 638) {
if (Cs < 975) {
if (SC < 1499) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (AFD < 10) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
return 8;
}
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (SC >= 1499) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 791) {
return 2;
}
else if (Rs >= 791) {
return 8;
}
}
}
}
else if (Cs >= 975) {
if (Rs < 889) {
return 1;
}
else if (Rs >= 889) {
return 4;
}
}
}
}
else if (AFD >= 11) {
if (negBalance < 52) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (ssDCval < 55) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 4;
}
}
else if (ssDCval >= 55) {
if (Rs < 828) {
return 1;
}
else if (Rs >= 828) {
return 2;
}
}
}
}
else if (negBalance >= 52) {
if (diffMVdiffVal < 8) {
if (refDCval < 54) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
else if (refDCval >= 54) {
return 8;
}
}
else if (diffMVdiffVal >= 8) {
if (Rs < 616) {
return 2;
}
else if (Rs >= 616) {
if (Rs < 680) {
return 8;
}
else if (Rs >= 680) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (Rs < 660) {
return 8;
}
else if (Rs >= 660) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 88) {
if (tcor < 85) {
if (mu_mv_mag_sq < 107) {
if (Rs < 599) {
return 8;
}
else if (Rs >= 599) {
return 4;
}
}
else if (mu_mv_mag_sq >= 107) {
return 8;
}
}
else if (tcor >= 85) {
if (ssDCval < 56) {
if (Rs < 847) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
return 4;
}
}
else if (Rs >= 847) {
return 8;
}
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 91) {
return 4;
}
else if (mu_mv_mag_sq >= 91) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (tcor < 99) {
if (Rs < 423) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
if (tcor < 81) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
return 2;
}
}
else if (tcor >= 81) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 1;
}
}
}
}
else if (Rs >= 423) {
if (CsDiff < 4) {
if (SC < 2103) {
if (refDCval < 75) {
if (RsCsDiff < 1) {
if (mu_mv_mag_sq < 119) {
return 8;
}
else if (mu_mv_mag_sq >= 119) {
return 4;
}
}
else if (RsCsDiff >= 1) {
if (negBalance < 42) {
if (Rs < 641) {
if (Rs < 497) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Rs < 475) {
if (Rs < 453) {
return 8;
}
else if (Rs >= 453) {
return 1;
}
}
else if (Rs >= 475) {
return 2;
}
}
}
else if (Rs >= 497) {
if (mu_mv_mag_sq < 80) {
if (Cs < 834) {
if (Rs < 558) {
return 8;
}
else if (Rs >= 558) {
return 4;
}
}
else if (Cs >= 834) {
return 1;
}
}
else if (mu_mv_mag_sq >= 80) {
if (Cs < 594) {
return 1;
}
else if (Cs >= 594) {
return 8;
}
}
}
}
else if (Rs >= 641) {
if (posBalance < 58) {
if (Rs < 793) {
if (ssDCval < 66) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (Rs < 660) {
return 2;
}
else if (Rs >= 660) {
return 8;
}
}
}
else if (ssDCval >= 66) {
if (MVDiff < 2) {
if (Cs < 586) {
if (Rs < 783) {
return 4;
}
else if (Rs >= 783) {
return 2;
}
}
else if (Cs >= 586) {
if (RsDiff < 1) {
if (Rs < 679) {
return 1;
}
else if (Rs >= 679) {
return 8;
}
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
if (Rs < 690) {
return 2;
}
else if (Rs >= 690) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 793) {
if (mu_mv_mag_sq < 106) {
if (mcTcor < 98) {
if (CsDiff < 3) {
if (TSC < 5) {
if (Rs < 938) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 4;
}
}
else if (Rs >= 938) {
return 1;
}
}
else if (TSC >= 5) {
return 1;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (mcTcor >= 98) {
if (Cs < 574) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 574) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (Rs < 807) {
if (Rs < 803) {
return 1;
}
else if (Rs >= 803) {
return 4;
}
}
else if (Rs >= 807) {
return 2;
}
}
}
}
}
else if (posBalance >= 58) {
if (RsDiff < 2) {
if (negBalance < 29) {
if (Rs < 741) {
return 2;
}
else if (Rs >= 741) {
return 4;
}
}
else if (negBalance >= 29) {
return 8;
}
}
else if (RsDiff >= 2) {
if (mcTcor < 95) {
if (Cs < 631) {
return 8;
}
else if (Cs >= 631) {
return 1;
}
}
else if (mcTcor >= 95) {
if (negBalance < 36) {
if (AFD < 6) {
if (Rs < 798) {
if (Rs < 773) {
return 1;
}
else if (Rs >= 773) {
return 4;
}
}
else if (Rs >= 798) {
return 2;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (negBalance >= 36) {
if (Cs < 1125) {
return 2;
}
else if (Cs >= 1125) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 42) {
if (MVDiff < 5) {
if (SC < 1464) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (SC >= 1464) {
return 8;
}
}
else if (MVDiff >= 5) {
if (TSC < 7) {
if (refDCval < 72) {
if (mu_mv_mag_sq < 78) {
return 1;
}
else if (mu_mv_mag_sq >= 78) {
return 8;
}
}
else if (refDCval >= 72) {
if (Rs < 854) {
return 8;
}
else if (Rs >= 854) {
return 4;
}
}
}
else if (TSC >= 7) {
return 4;
}
}
}
}
}
else if (refDCval >= 75) {
if (Cs < 727) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
if (SC < 1579) {
return 8;
}
else if (SC >= 1579) {
return 4;
}
}
}
else if (Cs >= 727) {
if (Rs < 837) {
return 2;
}
else if (Rs >= 837) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 2;
}
}
}
}
}
else if (SC >= 2103) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
if (Rs < 874) {
return 8;
}
else if (Rs >= 874) {
return 4;
}
}
}
}
else if (CsDiff >= 4) {
if (Cs < 944) {
if (mu_mv_mag_sq < 14) {
if (Rs < 704) {
if (TSC < 9) {
if (mu_mv_mag_sq < 3) {
if (Cs < 658) {
return 4;
}
else if (Cs >= 658) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (SC < 1229) {
return 8;
}
else if (SC >= 1229) {
if (Rs < 599) {
return 2;
}
else if (Rs >= 599) {
return 1;
}
}
}
}
else if (TSC >= 9) {
return 4;
}
}
else if (Rs >= 704) {
if (refDCval < 58) {
return 8;
}
else if (refDCval >= 58) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Cs < 923) {
return 4;
}
else if (Cs >= 923) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (MVDiff < 32) {
if (Cs < 667) {
return 8;
}
else if (Cs >= 667) {
if (AFD < 11) {
if (Cs < 685) {
return 2;
}
else if (Cs >= 685) {
if (Cs < 845) {
return 8;
}
else if (Cs >= 845) {
return 2;
}
}
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (MVDiff >= 32) {
if (Rs < 829) {
return 1;
}
else if (Rs >= 829) {
return 4;
}
}
}
}
else if (Cs >= 944) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (Rs < 901) {
if (mu_mv_mag_sq < 44) {
if (Cs < 998) {
if (Cs < 979) {
return 2;
}
else if (Cs >= 979) {
return 8;
}
}
else if (Cs >= 998) {
if (refDCval < 75) {
if (diffAFD < -1) {
if (Rs < 811) {
return 2;
}
else if (Rs >= 811) {
return 1;
}
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (refDCval >= 75) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (Cs < 1154) {
return 2;
}
else if (Cs >= 1154) {
return 1;
}
}
}
else if (Rs >= 901) {
if (Cs < 1201) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (Cs >= 1201) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 99) {
if (Rs < 731) {
if (SC < 1176) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 124) {
return 4;
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (MVDiff >= 5) {
if (Rs < 198) {
return 1;
}
else if (Rs >= 198) {
return 8;
}
}
}
else if (SC >= 1176) {
if (Rs < 718) {
return 8;
}
else if (Rs >= 718) {
if (Rs < 721) {
if (Cs < 694) {
return 1;
}
else if (Cs >= 694) {
return 4;
}
}
else if (Rs >= 721) {
return 8;
}
}
}
}
else if (Rs >= 731) {
if (Cs < 766) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Cs < 645) {
return 8;
}
else if (Cs >= 645) {
return 4;
}
}
}
else if (Cs >= 766) {
if (Cs < 897) {
if (Rs < 785) {
return 1;
}
else if (Rs >= 785) {
return 4;
}
}
else if (Cs >= 897) {
if (Rs < 881) {
return 1;
}
else if (Rs >= 881) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 78) {
if (mu_mv_mag_sq < 115) {
if (mcTcor < 98) {
if (Cs < 726) {
if (refDCval < 92) {
if (mu_mv_mag_sq < 49) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (MVDiff < 24) {
if (Rs < 846) {
return 8;
}
else if (Rs >= 846) {
if (Cs < 583) {
return 8;
}
else if (Cs >= 583) {
return 4;
}
}
}
else if (MVDiff >= 24) {
if (Rs < 550) {
if (Rs < 410) {
return 4;
}
else if (Rs >= 410) {
return 2;
}
}
else if (Rs >= 550) {
if (SC < 1158) {
return 2;
}
else if (SC >= 1158) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (posBalance < 41) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
else if (posBalance >= 41) {
if (tcor < 79) {
return 2;
}
else if (tcor >= 79) {
return 8;
}
}
}
}
else if (refDCval >= 92) {
if (refDCval < 111) {
if (MVDiff < 5) {
if (posBalance < 52) {
if (Rs < 624) {
return 8;
}
else if (Rs >= 624) {
return 2;
}
}
else if (posBalance >= 52) {
if (Cs < 543) {
return 2;
}
else if (Cs >= 543) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (mu_mv_mag_sq < 22) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (negBalance < 63) {
if (AFD < 12) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (AFD >= 12) {
return 1;
}
}
else if (negBalance >= 63) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 22) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (diffAFD < 3) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
return 8;
}
}
else if (diffAFD >= 3) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 111) {
if (Rs < 372) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
else if (Rs >= 372) {
if (AFD < 9) {
if (TSC < 2) {
return 2;
}
else if (TSC >= 2) {
if (ssDCval < 119) {
if (MVDiff < 6) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (Rs < 616) {
return 8;
}
else if (Rs >= 616) {
return 2;
}
}
}
else if (MVDiff >= 6) {
if (negBalance < 37) {
return 2;
}
else if (negBalance >= 37) {
return 8;
}
}
}
else if (ssDCval >= 119) {
if (Rs < 926) {
return 8;
}
else if (Rs >= 926) {
if (Rs < 941) {
return 1;
}
else if (Rs >= 941) {
return 8;
}
}
}
}
}
else if (AFD >= 9) {
if (diffRsCsdiff < -8) {
if (Rs < 529) {
return 4;
}
else if (Rs >= 529) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 1;
}
}
}
else if (diffRsCsdiff >= -8) {
if (RsCsDiff < 11) {
return 8;
}
else if (RsCsDiff >= 11) {
if (posBalance < 61) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 8;
}
}
else if (posBalance >= 61) {
if (mu_mv_mag_sq < 13) {
if (Cs < 661) {
return 2;
}
else if (Cs >= 661) {
return 4;
}
}
else if (mu_mv_mag_sq >= 13) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Cs >= 726) {
if (MVDiff < 2) {
if (Rs < 415) {
return 4;
}
else if (Rs >= 415) {
if (Cs < 872) {
if (posBalance < 50) {
if (ssDCval < 115) {
return 2;
}
else if (ssDCval >= 115) {
return 8;
}
}
else if (posBalance >= 50) {
if (ssDCval < 89) {
return 8;
}
else if (ssDCval >= 89) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (Cs >= 872) {
if (ssDCval < 94) {
if (negBalance < 37) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (negBalance >= 37) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (Rs < 884) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Rs >= 884) {
return 8;
}
}
}
}
else if (ssDCval >= 94) {
return 8;
}
}
}
}
else if (MVDiff >= 2) {
if (refDCval < 143) {
if (posBalance < 46) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
if (negBalance < 52) {
if (Cs < 960) {
return 2;
}
else if (Cs >= 960) {
return 8;
}
}
else if (negBalance >= 52) {
if (mu_mv_mag_sq < 13) {
if (Rs < 846) {
return 4;
}
else if (Rs >= 846) {
return 8;
}
}
else if (mu_mv_mag_sq >= 13) {
return 8;
}
}
}
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 38) {
if (TSC < 7) {
if (posBalance < 63) {
if (Cs < 1059) {
if (mcTcor < 95) {
if (ssDCval < 117) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 860) {
return 8;
}
else if (Rs >= 860) {
return 2;
}
}
}
else if (ssDCval >= 117) {
return 1;
}
}
else if (mcTcor >= 95) {
if (Rs < 912) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
if (Rs < 707) {
return 1;
}
else if (Rs >= 707) {
return 4;
}
}
}
else if (Rs >= 912) {
return 8;
}
}
}
else if (Cs >= 1059) {
if (Rs < 931) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
else if (Rs >= 931) {
return 2;
}
}
}
else if (posBalance >= 63) {
if (Rs < 633) {
if (posBalance < 68) {
return 8;
}
else if (posBalance >= 68) {
if (AFD < 14) {
if (Rs < 543) {
return 4;
}
else if (Rs >= 543) {
return 8;
}
}
else if (AFD >= 14) {
return 2;
}
}
}
else if (Rs >= 633) {
return 8;
}
}
}
else if (TSC >= 7) {
if (ssDCval < 126) {
if (gchDC < 2) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (gchDC >= 2) {
if (Rs < 711) {
if (Cs < 746) {
return 8;
}
else if (Cs >= 746) {
return 4;
}
}
else if (Rs >= 711) {
return 8;
}
}
}
else if (ssDCval >= 126) {
if (ssDCval < 144) {
if (diffRsCsdiff < -30) {
return 8;
}
else if (diffRsCsdiff >= -30) {
return 1;
}
}
else if (ssDCval >= 144) {
if (Cs < 818) {
return 2;
}
else if (Cs >= 818) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (MVDiff < 5) {
if (Cs < 898) {
if (TSC < 6) {
if (negBalance < 28) {
return 2;
}
else if (negBalance >= 28) {
if (refDCval < 89) {
return 8;
}
else if (refDCval >= 89) {
if (ssDCval < 102) {
return 1;
}
else if (ssDCval >= 102) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
return 8;
}
}
}
}
}
else if (TSC >= 6) {
if (Rs < 435) {
return 4;
}
else if (Rs >= 435) {
return 1;
}
}
}
else if (Cs >= 898) {
if (diffAFD < 2) {
if (posBalance < 55) {
if (negBalance < 39) {
return 4;
}
else if (negBalance >= 39) {
if (Cs < 939) {
return 1;
}
else if (Cs >= 939) {
return 8;
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (diffAFD >= 2) {
return 1;
}
}
}
else if (MVDiff >= 5) {
if (Rs < 831) {
if (AFD < 10) {
if (posBalance < 49) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (posBalance >= 49) {
return 8;
}
}
else if (AFD >= 10) {
if (refDCval < 116) {
return 8;
}
else if (refDCval >= 116) {
return 2;
}
}
}
else if (Rs >= 831) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 940) {
return 1;
}
else if (Rs >= 940) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 143) {
if (refDCval < 155) {
if (RsCsDiff < 16) {
if (tcor < 31) {
return 1;
}
else if (tcor >= 31) {
return 8;
}
}
else if (RsCsDiff >= 16) {
if (Cs < 771) {
return 2;
}
else if (Cs >= 771) {
return 1;
}
}
}
else if (refDCval >= 155) {
if (tcor < 93) {
return 2;
}
else if (tcor >= 93) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (MVDiff < 4) {
if (Rs < 415) {
if (Rs < 349) {
if (Rs < 281) {
return 2;
}
else if (Rs >= 281) {
return 4;
}
}
else if (Rs >= 349) {
if (mu_mv_mag_sq < 108) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 4;
}
}
else if (mu_mv_mag_sq >= 108) {
return 8;
}
}
}
else if (Rs >= 415) {
if (Rs < 685) {
if (Rs < 470) {
if (refDCval < 99) {
return 8;
}
else if (refDCval >= 99) {
if (ssDCval < 125) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
return 1;
}
}
else if (ssDCval >= 125) {
return 8;
}
}
}
else if (Rs >= 470) {
if (TSC < 4) {
if (ssDCval < 116) {
if (ssDCval < 92) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (ssDCval < 84) {
if (Rs < 565) {
return 1;
}
else if (Rs >= 565) {
return 8;
}
}
else if (ssDCval >= 84) {
if (diffAFD < 1) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
if (Rs < 596) {
return 2;
}
else if (Rs >= 596) {
return 1;
}
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
else if (ssDCval >= 92) {
if (mu_mv_mag_sq < 113) {
return 1;
}
else if (mu_mv_mag_sq >= 113) {
if (Cs < 1154) {
return 8;
}
else if (Cs >= 1154) {
return 1;
}
}
}
}
else if (ssDCval >= 116) {
if (Rs < 518) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
else if (Rs >= 518) {
if (ssDCval < 120) {
return 2;
}
else if (ssDCval >= 120) {
if (Cs < 874) {
return 4;
}
else if (Cs >= 874) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
if (ssDCval < 122) {
if (ssDCval < 94) {
return 4;
}
else if (ssDCval >= 94) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
else if (ssDCval >= 122) {
if (Rs < 668) {
return 1;
}
else if (Rs >= 668) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 685) {
if (Rs < 837) {
if (Cs < 770) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
else if (Cs >= 770) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 8;
}
}
}
else if (Rs >= 837) {
if (Cs < 1149) {
return 8;
}
else if (Cs >= 1149) {
return 4;
}
}
}
}
}
else if (MVDiff >= 4) {
if (negBalance < 69) {
if (SC < 1483) {
if (negBalance < 20) {
if (ssDCval < 112) {
return 1;
}
else if (ssDCval >= 112) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
return 8;
}
}
}
}
else if (negBalance >= 20) {
if (SC < 1290) {
if (RsDiff < 1) {
if (ssDCval < 125) {
return 2;
}
else if (ssDCval >= 125) {
if (Rs < 388) {
return 2;
}
else if (Rs >= 388) {
return 1;
}
}
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 71) {
if (SC < 1239) {
if (ssDCval < 151) {
if (diffTSC < -1) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (ssDCval >= 151) {
if (Rs < 475) {
return 4;
}
else if (Rs >= 475) {
return 1;
}
}
}
else if (SC >= 1239) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (Cs < 715) {
return 4;
}
else if (Cs >= 715) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 71) {
if (Rs < 494) {
if (SC < 1027) {
if (SC < 860) {
return 4;
}
else if (SC >= 860) {
return 2;
}
}
else if (SC >= 1027) {
return 8;
}
}
else if (Rs >= 494) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 8;
}
}
}
}
}
else if (SC >= 1290) {
if (refDCval < 86) {
if (diffRsCsdiff < -1) {
if (Rs < 585) {
return 1;
}
else if (Rs >= 585) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
return 1;
}
}
else if (refDCval >= 86) {
if (ssDCval < 95) {
if (Rs < 660) {
return 2;
}
else if (Rs >= 660) {
if (Rs < 709) {
return 8;
}
else if (Rs >= 709) {
return 4;
}
}
}
else if (ssDCval >= 95) {
if (mu_mv_mag_sq < 99) {
if (TSC < 4) {
if (Rs < 465) {
return 2;
}
else if (Rs >= 465) {
return 1;
}
}
else if (TSC >= 4) {
if (Rs < 637) {
return 4;
}
else if (Rs >= 637) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 99) {
return 8;
}
}
}
}
}
}
else if (SC >= 1483) {
if (RsDiff < 2) {
if (Rs < 649) {
if (mu_mv_mag_sq < 94) {
if (Rs < 606) {
return 8;
}
else if (Rs >= 606) {
return 2;
}
}
else if (mu_mv_mag_sq >= 94) {
return 1;
}
}
else if (Rs >= 649) {
if (mu_mv_mag_sq < 100) {
return 8;
}
else if (mu_mv_mag_sq >= 100) {
if (Rs < 807) {
return 8;
}
else if (Rs >= 807) {
return 4;
}
}
}
}
else if (RsDiff >= 2) {
if (mu_mv_mag_sq < 3) {
if (negBalance < 42) {
return 2;
}
else if (negBalance >= 42) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 144) {
return 8;
}
else if (ssDCval >= 144) {
return 2;
}
}
}
}
}
else if (negBalance >= 69) {
if (Cs < 666) {
if (Rs < 599) {
return 8;
}
else if (Rs >= 599) {
return 1;
}
}
else if (Cs >= 666) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 115) {
if (posBalance < 10) {
if (diffTSC < -1) {
if (Cs < 639) {
return 8;
}
else if (Cs >= 639) {
return 4;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (posBalance >= 10) {
if (ssDCval < 88) {
if (posBalance < 48) {
if (Rs < 768) {
return 2;
}
else if (Rs >= 768) {
return 8;
}
}
else if (posBalance >= 48) {
if (Cs < 807) {
return 1;
}
else if (Cs >= 807) {
return 4;
}
}
}
else if (ssDCval >= 88) {
if (negBalance < 2) {
if (Cs < 1099) {
if (diffTSC < -3) {
if (Rs < 859) {
return 2;
}
else if (Rs >= 859) {
return 4;
}
}
else if (diffTSC >= -3) {
return 8;
}
}
else if (Cs >= 1099) {
return 2;
}
}
else if (negBalance >= 2) {
if (Rs < 745) {
if (posBalance < 27) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (Rs < 646) {
return 8;
}
else if (Rs >= 646) {
if (Cs < 688) {
if (Cs < 650) {
return 1;
}
else if (Cs >= 650) {
return 8;
}
}
else if (Cs >= 688) {
return 1;
}
}
}
}
else if (posBalance >= 27) {
if (Rs < 369) {
if (Rs < 322) {
return 2;
}
else if (Rs >= 322) {
if (ssDCval < 109) {
return 4;
}
else if (ssDCval >= 109) {
return 8;
}
}
}
else if (Rs >= 369) {
if (Cs < 1240) {
if (Cs < 767) {
if (Cs < 709) {
if (ssDCval < 135) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 1;
}
}
else if (ssDCval >= 135) {
return 1;
}
}
else if (Cs >= 709) {
if (ssDCval < 117) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 8;
}
}
else if (ssDCval >= 117) {
return 8;
}
}
}
else if (Cs >= 767) {
return 8;
}
}
else if (Cs >= 1240) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
else if (Rs >= 745) {
if (diffRsCsdiff < -13) {
return 2;
}
else if (diffRsCsdiff >= -13) {
if (SC < 1373) {
if (Cs < 614) {
return 8;
}
else if (Cs >= 614) {
return 1;
}
}
else if (SC >= 1373) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 959) {
if (posBalance < 55) {
if (negBalance < 14) {
if (ssDCval < 55) {
if (TSCindex < 2) {
if (Cs < 557) {
if (Rs < 1183) {
return 8;
}
else if (Rs >= 1183) {
return 1;
}
}
else if (Cs >= 557) {
if (Cs < 1059) {
return 8;
}
else if (Cs >= 1059) {
if (mu_mv_mag_sq < 119) {
if (Cs < 1164) {
return 1;
}
else if (Cs >= 1164) {
return 8;
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
}
}
else if (TSCindex >= 2) {
if (diffMVdiffVal < -3) {
if (Rs < 1033) {
return 8;
}
else if (Rs >= 1033) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 1501) {
return 1;
}
else if (Rs >= 1501) {
return 8;
}
}
}
}
}
else if (ssDCval >= 55) {
if (MVDiff < 6) {
if (ssDCval < 65) {
if (SC < 2865) {
if (SC < 2416) {
if (Rs < 1265) {
if (ssDCval < 59) {
if (Cs < 983) {
return 8;
}
else if (Cs >= 983) {
return 1;
}
}
else if (ssDCval >= 59) {
return 1;
}
}
else if (Rs >= 1265) {
return 8;
}
}
else if (SC >= 2416) {
if (diffRsCsdiff < -2) {
if (Rs < 1847) {
return 4;
}
else if (Rs >= 1847) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
else if (SC >= 2865) {
if (Rs < 2146) {
if (Cs < 984) {
return 1;
}
else if (Cs >= 984) {
if (Rs < 1891) {
if (Rs < 1879) {
if (Rs < 1846) {
if (Rs < 1831) {
return 4;
}
else if (Rs >= 1831) {
return 1;
}
}
else if (Rs >= 1846) {
return 4;
}
}
else if (Rs >= 1879) {
return 1;
}
}
else if (Rs >= 1891) {
return 4;
}
}
}
else if (Rs >= 2146) {
if (SC < 3570) {
return 8;
}
else if (SC >= 3570) {
return 2;
}
}
}
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 22) {
if (negBalance < 2) {
if (Rs < 1110) {
if (Rs < 1025) {
return 2;
}
else if (Rs >= 1025) {
return 8;
}
}
else if (Rs >= 1110) {
return 1;
}
}
else if (negBalance >= 2) {
if (Rs < 2566) {
if (ssDCval < 151) {
return 8;
}
else if (ssDCval >= 151) {
return 2;
}
}
else if (Rs >= 2566) {
return 4;
}
}
}
else if (posBalance >= 22) {
if (ssDCval < 83) {
return 1;
}
else if (ssDCval >= 83) {
if (Rs < 1166) {
return 1;
}
else if (Rs >= 1166) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (diffMVdiffVal < -4) {
if (SC < 2031) {
if (Cs < 809) {
return 4;
}
else if (Cs >= 809) {
return 1;
}
}
else if (SC >= 2031) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (diffMVdiffVal >= -4) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (posBalance < 37) {
if (posBalance < 32) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (posBalance >= 32) {
if (Rs < 1113) {
return 8;
}
else if (Rs >= 1113) {
return 1;
}
}
}
else if (posBalance >= 37) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (diffRsCsdiff < -57) {
if (diffMVdiffVal < -27) {
return 4;
}
else if (diffMVdiffVal >= -27) {
if (diffMVdiffVal < 12) {
if (AFD < 2) {
if (Cs < 1290) {
if (MVDiff < 31) {
return 8;
}
else if (MVDiff >= 31) {
if (Cs < 952) {
return 8;
}
else if (Cs >= 952) {
return 4;
}
}
}
else if (Cs >= 1290) {
return 1;
}
}
else if (AFD >= 2) {
return 1;
}
}
else if (diffMVdiffVal >= 12) {
if (Rs < 1228) {
return 4;
}
else if (Rs >= 1228) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= -57) {
if (Cs < 976) {
if (Cs < 670) {
if (Rs < 1500) {
if (MVDiff < 30) {
return 1;
}
else if (MVDiff >= 30) {
return 4;
}
}
else if (Rs >= 1500) {
return 8;
}
}
else if (Cs >= 670) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
if (diffTSC < -6) {
return 1;
}
else if (diffTSC >= -6) {
return 4;
}
}
}
}
else if (Cs >= 976) {
if (ssDCval < 137) {
if (Rs < 2161) {
return 8;
}
else if (Rs >= 2161) {
return 4;
}
}
else if (ssDCval >= 137) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (Cs < 964) {
if (Rs < 1793) {
if (mcTcor < 96) {
if (refDCval < 29) {
return 4;
}
else if (refDCval >= 29) {
if (CsDiff < 2) {
if (ssDCval < 64) {
if (ssDCval < 50) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 1209) {
if (Rs < 1000) {
return 8;
}
else if (Rs >= 1000) {
return 1;
}
}
else if (Rs >= 1209) {
return 4;
}
}
}
else if (ssDCval >= 50) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 992) {
return 8;
}
else if (Rs >= 992) {
return 1;
}
}
}
}
else if (ssDCval >= 64) {
if (MVDiff < 15) {
if (mu_mv_mag_sq < 115) {
if (ssDCval < 79) {
return 1;
}
else if (ssDCval >= 79) {
return 8;
}
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 1123) {
return 8;
}
else if (Rs >= 1123) {
return 4;
}
}
}
else if (MVDiff >= 15) {
return 4;
}
}
}
else if (CsDiff >= 2) {
if (refDCval < 138) {
if (Rs < 1317) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (MVDiff < 10) {
if (mu_mv_mag_sq < 124) {
if (negBalance < 44) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (refDCval < 50) {
if (Cs < 907) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
if (Rs < 1109) {
return 8;
}
else if (Rs >= 1109) {
return 1;
}
}
}
else if (Cs >= 907) {
if (Rs < 1198) {
return 8;
}
else if (Rs >= 1198) {
return 1;
}
}
}
else if (refDCval >= 50) {
if (mu_mv_mag_sq < 89) {
if (AFD < 8) {
if (posBalance < 47) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
return 4;
}
}
}
else if (posBalance >= 47) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (AFD >= 8) {
if (Cs < 875) {
return 1;
}
else if (Cs >= 875) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (Cs < 647) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (Cs < 619) {
return 8;
}
else if (Cs >= 619) {
if (Rs < 1167) {
return 1;
}
else if (Rs >= 1167) {
return 4;
}
}
}
}
else if (Cs >= 647) {
return 8;
}
}
}
}
}
else if (negBalance >= 44) {
if (diffMVdiffVal < 3) {
if (Cs < 572) {
return 2;
}
else if (Cs >= 572) {
if (Rs < 1284) {
return 8;
}
else if (Rs >= 1284) {
if (Rs < 1300) {
return 1;
}
else if (Rs >= 1300) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (MVDiff >= 10) {
if (refDCval < 58) {
if (mcTcor < 89) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (mcTcor >= 89) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
if (negBalance < 44) {
if (Cs < 870) {
return 2;
}
else if (Cs >= 870) {
return 8;
}
}
else if (negBalance >= 44) {
return 8;
}
}
}
}
else if (refDCval >= 58) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (TSC < 7) {
if (negBalance < 46) {
if (RsCsDiff < 23) {
return 8;
}
else if (RsCsDiff >= 23) {
return 1;
}
}
else if (negBalance >= 46) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (Rs < 1032) {
return 8;
}
else if (Rs >= 1032) {
if (Rs < 1141) {
return 4;
}
else if (Rs >= 1141) {
return 1;
}
}
}
}
}
else if (TSC >= 7) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1317) {
if (ssDCval < 44) {
return 4;
}
else if (ssDCval >= 44) {
if (Cs < 718) {
if (refDCval < 51) {
if (Rs < 1401) {
return 8;
}
else if (Rs >= 1401) {
return 8;
}
}
else if (refDCval >= 51) {
if (tcor < 86) {
if (Rs < 1568) {
return 8;
}
else if (Rs >= 1568) {
return 4;
}
}
else if (tcor >= 86) {
return 1;
}
}
}
else if (Cs >= 718) {
if (negBalance < 38) {
if (Rs < 1587) {
if (Cs < 858) {
if (Rs < 1494) {
return 2;
}
else if (Rs >= 1494) {
return 4;
}
}
else if (Cs >= 858) {
return 8;
}
}
else if (Rs >= 1587) {
if (Rs < 1715) {
return 1;
}
else if (Rs >= 1715) {
return 8;
}
}
}
else if (negBalance >= 38) {
if (MVDiff < 7) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (Rs < 1657) {
return 1;
}
else if (Rs >= 1657) {
return 2;
}
}
}
else if (MVDiff >= 7) {
if (mu_mv_mag_sq < 41) {
if (negBalance < 43) {
return 4;
}
else if (negBalance >= 43) {
if (tcor < 90) {
if (Rs < 1508) {
return 4;
}
else if (Rs >= 1508) {
return 8;
}
}
else if (tcor >= 90) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 138) {
if (ssDCval < 141) {
if (Cs < 703) {
return 8;
}
else if (Cs >= 703) {
return 4;
}
}
else if (ssDCval >= 141) {
if (ssDCval < 150) {
return 8;
}
else if (ssDCval >= 150) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 69) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 48) {
if (Rs < 1271) {
if (Cs < 647) {
return 1;
}
else if (Cs >= 647) {
return 8;
}
}
else if (Rs >= 1271) {
if (Cs < 912) {
if (posBalance < 43) {
return 4;
}
else if (posBalance >= 43) {
if (Rs < 1385) {
return 4;
}
else if (Rs >= 1385) {
if (Rs < 1465) {
return 4;
}
else if (Rs >= 1465) {
return 4;
}
}
}
}
else if (Cs >= 912) {
return 1;
}
}
}
else if (ssDCval >= 48) {
if (mcTcor < 98) {
if (ssDCval < 66) {
if (negBalance < 45) {
if (Cs < 656) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 2;
}
}
else if (Cs >= 656) {
if (CsDiff < 2) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
if (negBalance < 25) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (negBalance >= 25) {
return 8;
}
}
}
else if (CsDiff >= 2) {
if (Rs < 1333) {
if (mu_mv_mag_sq < 106) {
return 1;
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 1158) {
return 1;
}
else if (Rs >= 1158) {
return 8;
}
}
}
else if (Rs >= 1333) {
if (refDCval < 54) {
return 1;
}
else if (refDCval >= 54) {
if (negBalance < 36) {
return 8;
}
else if (negBalance >= 36) {
if (Cs < 706) {
return 8;
}
else if (Cs >= 706) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (negBalance < 48) {
if (Cs < 759) {
return 4;
}
else if (Cs >= 759) {
return 8;
}
}
else if (negBalance >= 48) {
if (Rs < 1081) {
return 1;
}
else if (Rs >= 1081) {
return 8;
}
}
}
}
else if (ssDCval >= 66) {
if (Rs < 1186) {
if (Cs < 662) {
return 1;
}
else if (Cs >= 662) {
return 8;
}
}
else if (Rs >= 1186) {
if (Rs < 1462) {
return 1;
}
else if (Rs >= 1462) {
return 2;
}
}
}
}
else if (mcTcor >= 98) {
if (SC < 2427) {
if (SC < 1800) {
if (refDCval < 65) {
if (ssDCval < 51) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (ssDCval >= 51) {
return 1;
}
}
else if (refDCval >= 65) {
if (AFD < 3) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
return 2;
}
}
else if (AFD >= 3) {
return 1;
}
}
}
else if (SC >= 1800) {
if (RsCsDiff < 7) {
if (SC < 2354) {
if (posBalance < 31) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (posBalance >= 31) {
return 1;
}
}
else if (SC >= 2354) {
if (Rs < 1551) {
return 8;
}
else if (Rs >= 1551) {
return 1;
}
}
}
else if (RsCsDiff >= 7) {
if (mu_mv_mag_sq < 85) {
if (negBalance < 38) {
if (diffRsCsdiff < -6) {
if (RsCsDiff < 16) {
return 2;
}
else if (RsCsDiff >= 16) {
return 1;
}
}
else if (diffRsCsdiff >= -6) {
return 1;
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (mu_mv_mag_sq >= 85) {
if (Cs < 879) {
if (Rs < 1474) {
return 8;
}
else if (Rs >= 1474) {
return 1;
}
}
else if (Cs >= 879) {
return 1;
}
}
}
}
}
else if (SC >= 2427) {
if (mu_mv_mag_sq < 91) {
if (Rs < 1694) {
return 4;
}
else if (Rs >= 1694) {
return 1;
}
}
else if (mu_mv_mag_sq >= 91) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (ssDCval < 64) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (ssDCval < 58) {
if (diffRsCsdiff < -3) {
if (Rs < 1411) {
return 4;
}
else if (Rs >= 1411) {
return 1;
}
}
else if (diffRsCsdiff >= -3) {
if (Rs < 963) {
return 1;
}
else if (Rs >= 963) {
return 8;
}
}
}
else if (ssDCval >= 58) {
if (posBalance < 36) {
if (negBalance < 18) {
return 8;
}
else if (negBalance >= 18) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
return 8;
}
}
}
else if (posBalance >= 36) {
return 1;
}
}
}
}
else if (ssDCval >= 64) {
if (posBalance < 39) {
if (Rs < 1102) {
if (Rs < 1025) {
return 2;
}
else if (Rs >= 1025) {
return 8;
}
}
else if (Rs >= 1102) {
return 1;
}
}
else if (posBalance >= 39) {
if (Rs < 1076) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
return 4;
}
}
else if (Rs >= 1076) {
if (Cs < 676) {
return 2;
}
else if (Cs >= 676) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 69) {
if (ssDCval < 82) {
if (mu_mv_mag_sq < 59) {
if (MVDiff < 42) {
return 1;
}
else if (MVDiff >= 42) {
if (Rs < 1104) {
return 4;
}
else if (Rs >= 1104) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (SC < 1716) {
if (ssDCval < 75) {
if (Cs < 580) {
return 1;
}
else if (Cs >= 580) {
return 2;
}
}
else if (ssDCval >= 75) {
return 8;
}
}
else if (SC >= 1716) {
if (refDCval < 73) {
if (mu_mv_mag_sq < 109) {
if (RsDiff < 4) {
if (SC < 2521) {
return 1;
}
else if (SC >= 2521) {
return 8;
}
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 109) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
return 8;
}
}
}
else if (refDCval >= 73) {
if (Cs < 768) {
return 1;
}
else if (Cs >= 768) {
if (negBalance < 61) {
if (mu_mv_mag_sq < 68) {
if (Rs < 1394) {
return 8;
}
else if (Rs >= 1394) {
return 1;
}
}
else if (mu_mv_mag_sq >= 68) {
if (Rs < 991) {
if (Cs < 884) {
return 8;
}
else if (Cs >= 884) {
return 4;
}
}
else if (Rs >= 991) {
return 8;
}
}
}
else if (negBalance >= 61) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 82) {
if (tcor < 99) {
if (SC < 2552) {
if (mu_mv_mag_sq < 66) {
if (negBalance < 57) {
if (MVDiff < 31) {
if (mu_mv_mag_sq < 58) {
if (SC < 1908) {
if (SC < 1890) {
if (Cs < 759) {
return 8;
}
else if (Cs >= 759) {
return 1;
}
}
else if (SC >= 1890) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
}
else if (SC >= 1908) {
return 8;
}
}
else if (mu_mv_mag_sq >= 58) {
return 1;
}
}
else if (MVDiff >= 31) {
if (Rs < 1216) {
return 4;
}
else if (Rs >= 1216) {
return 1;
}
}
}
else if (negBalance >= 57) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (Rs < 1012) {
return 8;
}
else if (Rs >= 1012) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (Rs < 1333) {
return 8;
}
else if (Rs >= 1333) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (AFD < 5) {
if (Cs < 850) {
if (SC < 2451) {
return 4;
}
else if (SC >= 2451) {
return 8;
}
}
else if (Cs >= 850) {
return 8;
}
}
else if (AFD >= 5) {
return 8;
}
}
}
}
}
else if (SC >= 2552) {
if (mcTcor < 98) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 8;
}
}
else if (mcTcor >= 98) {
return 1;
}
}
}
else if (tcor >= 99) {
if (ssDCval < 99) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
else if (ssDCval >= 99) {
if (mu_mv_mag_sq < 122) {
return 8;
}
else if (mu_mv_mag_sq >= 122) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
return 1;
}
}
}
}
}
}
}
}
else if (Rs >= 1793) {
if (ssDCval < 88) {
if (refDCval < 82) {
if (ssDCval < 58) {
if (mu_mv_mag_sq < 96) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
if (mu_mv_mag_sq < 115) {
if (Cs < 812) {
return 1;
}
else if (Cs >= 812) {
return 4;
}
}
else if (mu_mv_mag_sq >= 115) {
if (mcTcor < 99) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
return 1;
}
}
else if (mcTcor >= 99) {
if (tcor < 100) {
if (Cs < 953) {
return 1;
}
else if (Cs >= 953) {
return 4;
}
}
else if (tcor >= 100) {
if (Rs < 2078) {
return 4;
}
else if (Rs >= 2078) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 58) {
if (ssDCval < 71) {
if (MVDiff < 1) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (MVDiff >= 1) {
if (Cs < 686) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
}
else if (Cs >= 686) {
if (posBalance < 43) {
if (mu_mv_mag_sq < 78) {
if (diffMVdiffVal < 1) {
if (Rs < 2072) {
return 8;
}
else if (Rs >= 2072) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (mu_mv_mag_sq >= 78) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
return 8;
}
}
}
else if (posBalance >= 43) {
return 2;
}
}
}
}
else if (ssDCval >= 71) {
if (negBalance < 42) {
if (diffRsCsdiff < -45) {
return 2;
}
else if (diffRsCsdiff >= -45) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
if (Cs < 678) {
return 1;
}
else if (Cs >= 678) {
return 2;
}
}
}
}
else if (negBalance >= 42) {
if (SC < 2820) {
return 1;
}
else if (SC >= 2820) {
if (diffAFD < 1) {
if (diffAFD < -1) {
if (Rs < 2254) {
return 8;
}
else if (Rs >= 2254) {
return 8;
}
}
else if (diffAFD >= -1) {
if (Cs < 954) {
return 4;
}
else if (Cs >= 954) {
return 4;
}
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 82) {
if (mu_mv_mag_sq < 90) {
if (SC < 3102) {
return 1;
}
else if (SC >= 3102) {
return 4;
}
}
else if (mu_mv_mag_sq >= 90) {
if (Cs < 817) {
return 4;
}
else if (Cs >= 817) {
return 1;
}
}
}
}
else if (ssDCval >= 88) {
if (ssDCval < 122) {
if (mu_mv_mag_sq < 96) {
if (refDCval < 102) {
if (MVDiff < 21) {
if (RsCsDiff < 12) {
if (Rs < 2051) {
return 8;
}
else if (Rs >= 2051) {
return 1;
}
}
else if (RsCsDiff >= 12) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (posBalance < 52) {
if (Rs < 2006) {
return 8;
}
else if (Rs >= 2006) {
return 4;
}
}
else if (posBalance >= 52) {
return 1;
}
}
}
}
else if (MVDiff >= 21) {
if (Rs < 1846) {
return 1;
}
else if (Rs >= 1846) {
return 2;
}
}
}
else if (refDCval >= 102) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
if (Rs < 2143) {
if (SC < 2648) {
if (Rs < 1810) {
return 8;
}
else if (Rs >= 1810) {
return 4;
}
}
else if (SC >= 2648) {
return 8;
}
}
else if (Rs >= 2143) {
if (refDCval < 89) {
if (posBalance < 36) {
if (Rs < 2164) {
return 4;
}
else if (Rs >= 2164) {
return 8;
}
}
else if (posBalance >= 36) {
return 4;
}
}
else if (refDCval >= 89) {
return 8;
}
}
}
}
else if (ssDCval >= 122) {
if (tcor < 98) {
if (posBalance < 41) {
return 4;
}
else if (posBalance >= 41) {
if (RsCsDiff < 32) {
return 8;
}
else if (RsCsDiff >= 32) {
if (Rs < 2251) {
return 2;
}
else if (Rs >= 2251) {
return 1;
}
}
}
}
else if (tcor >= 98) {
return 1;
}
}
}
}
}
else if (Cs >= 964) {
if (posBalance < 40) {
if (refDCval < 83) {
if (ssDCval < 49) {
if (SC < 2960) {
if (posBalance < 28) {
if (ssDCval < 42) {
return 8;
}
else if (ssDCval >= 42) {
if (Rs < 1234) {
return 8;
}
else if (Rs >= 1234) {
return 1;
}
}
}
else if (posBalance >= 28) {
if (Rs < 1185) {
return 8;
}
else if (Rs >= 1185) {
if (diffAFD < 1) {
if (Cs < 1232) {
if (refDCval < 47) {
if (diffMVdiffVal < -1) {
if (Rs < 1825) {
return 8;
}
else if (Rs >= 1825) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (refDCval >= 47) {
if (negBalance < 19) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
if (Cs < 1031) {
return 8;
}
else if (Cs >= 1031) {
if (Rs < 1678) {
return 8;
}
else if (Rs >= 1678) {
if (posBalance < 39) {
if (diffAFD < -1) {
if (Rs < 1769) {
return 8;
}
else if (Rs >= 1769) {
return 4;
}
}
else if (diffAFD >= -1) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1737) {
return 4;
}
else if (Rs >= 1737) {
return 4;
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1037) {
return 4;
}
else if (Cs >= 1037) {
if (Cs < 1083) {
return 8;
}
else if (Cs >= 1083) {
return 4;
}
}
}
}
}
else if (posBalance >= 39) {
return 8;
}
}
}
}
}
else if (negBalance >= 19) {
if (Rs < 1410) {
return 8;
}
else if (Rs >= 1410) {
if (ssDCval < 48) {
if (Cs < 1155) {
if (mu_mv_mag_sq < 126) {
if (Rs < 1705) {
return 8;
}
else if (Rs >= 1705) {
return 4;
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1100) {
if (negBalance < 22) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
if (AFD < 2) {
if (Rs < 1743) {
if (Rs < 1730) {
if (Rs < 1715) {
return 4;
}
else if (Rs >= 1715) {
return 8;
}
}
else if (Rs >= 1730) {
return 8;
}
}
else if (Rs >= 1743) {
return 4;
}
}
else if (AFD >= 2) {
if (negBalance < 20) {
if (Rs < 1685) {
return 8;
}
else if (Rs >= 1685) {
if (Rs < 1735) {
return 4;
}
else if (Rs >= 1735) {
return 4;
}
}
}
else if (negBalance >= 20) {
return 4;
}
}
}
}
else if (negBalance >= 22) {
if (Rs < 1758) {
return 4;
}
else if (Rs >= 1758) {
return 4;
}
}
}
else if (Cs >= 1100) {
if (Cs < 1107) {
return 4;
}
else if (Cs >= 1107) {
if (Rs < 1814) {
if (Cs < 1148) {
if (RsDiff < 2) {
if (Rs < 1688) {
return 4;
}
else if (Rs >= 1688) {
if (posBalance < 34) {
if (Rs < 1742) {
return 4;
}
else if (Rs >= 1742) {
return 4;
}
}
else if (posBalance >= 34) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 1795) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 4;
}
}
else if (Rs >= 1795) {
return 4;
}
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 1722) {
if (posBalance < 35) {
return 4;
}
else if (posBalance >= 35) {
return 8;
}
}
else if (Rs >= 1722) {
return 8;
}
}
}
else if (Cs >= 1148) {
return 4;
}
}
else if (Rs >= 1814) {
return 8;
}
}
}
}
}
else if (Cs >= 1155) {
if (posBalance < 31) {
return 4;
}
else if (posBalance >= 31) {
return 4;
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
}
}
}
else if (Cs >= 1232) {
return 8;
}
}
else if (diffAFD >= 1) {
if (Cs < 1079) {
return 8;
}
else if (Cs >= 1079) {
if (Rs < 1764) {
return 4;
}
else if (Rs >= 1764) {
if (Rs < 1775) {
return 4;
}
else if (Rs >= 1775) {
if (Rs < 1783) {
return 8;
}
else if (Rs >= 1783) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 2960) {
if (posBalance < 31) {
if (Cs < 1095) {
return 4;
}
else if (Cs >= 1095) {
return 8;
}
}
else if (posBalance >= 31) {
return 4;
}
}
}
else if (ssDCval >= 49) {
if (mu_mv_mag_sq < 104) {
if (tcor < 98) {
if (refDCval < 66) {
if (diffRsCsdiff < 3) {
if (SC < 3062) {
if (Cs < 1157) {
return 4;
}
else if (Cs >= 1157) {
return 2;
}
}
else if (SC >= 3062) {
return 2;
}
}
else if (diffRsCsdiff >= 3) {
return 1;
}
}
else if (refDCval >= 66) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (Rs < 1432) {
return 8;
}
else if (Rs >= 1432) {
return 1;
}
}
}
}
else if (tcor >= 98) {
if (Rs < 2052) {
if (ssDCval < 58) {
if (Rs < 1879) {
return 4;
}
else if (Rs >= 1879) {
return 1;
}
}
else if (ssDCval >= 58) {
return 4;
}
}
else if (Rs >= 2052) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Rs < 1760) {
if (Rs < 1373) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
return 8;
}
}
else if (Rs >= 1373) {
if (ssDCval < 68) {
if (Rs < 1434) {
if (Cs < 1124) {
return 8;
}
else if (Cs >= 1124) {
return 1;
}
}
else if (Rs >= 1434) {
return 8;
}
}
else if (ssDCval >= 68) {
return 1;
}
}
}
else if (Rs >= 1760) {
if (AFD < 4) {
if (ssDCval < 60) {
if (mu_mv_mag_sq < 114) {
if (Rs < 1874) {
return 4;
}
else if (Rs >= 1874) {
return 1;
}
}
else if (mu_mv_mag_sq >= 114) {
return 4;
}
}
else if (ssDCval >= 60) {
if (Cs < 1054) {
if (mcTcor < 99) {
return 1;
}
else if (mcTcor >= 99) {
return 4;
}
}
else if (Cs >= 1054) {
if (mu_mv_mag_sq < 114) {
if (Rs < 1913) {
return 1;
}
else if (Rs >= 1913) {
return 4;
}
}
else if (mu_mv_mag_sq >= 114) {
return 1;
}
}
}
}
else if (AFD >= 4) {
if (negBalance < 54) {
if (Rs < 2280) {
return 8;
}
else if (Rs >= 2280) {
return 4;
}
}
else if (negBalance >= 54) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 83) {
if (ssDCval < 128) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
if (Cs < 972) {
if (Rs < 1487) {
return 8;
}
else if (Rs >= 1487) {
return 1;
}
}
else if (Cs >= 972) {
return 8;
}
}
}
else if (ssDCval >= 128) {
if (negBalance < 30) {
if (Rs < 1748) {
if (Rs < 1559) {
return 8;
}
else if (Rs >= 1559) {
return 2;
}
}
else if (Rs >= 1748) {
return 4;
}
}
else if (negBalance >= 30) {
if (Rs < 1914) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
return 2;
}
}
else if (Rs >= 1914) {
return 8;
}
}
}
}
}
else if (posBalance >= 40) {
if (Rs < 1454) {
if (ssDCval < 157) {
if (refDCval < 50) {
if (posBalance < 48) {
if (SC < 2383) {
if (Rs < 1040) {
return 4;
}
else if (Rs >= 1040) {
return 8;
}
}
else if (SC >= 2383) {
return 2;
}
}
else if (posBalance >= 48) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
return 8;
}
}
}
else if (refDCval >= 50) {
if (MVDiff < 5) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 86) {
if (CsDiff < 1) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (CsDiff >= 1) {
if (ssDCval < 81) {
if (mcTcor < 94) {
if (diffAFD < 2) {
if (Rs < 1253) {
if (Cs < 1106) {
return 8;
}
else if (Cs >= 1106) {
return 1;
}
}
else if (Rs >= 1253) {
return 8;
}
}
else if (diffAFD >= 2) {
return 4;
}
}
else if (mcTcor >= 94) {
if (RsCsDiff < 7) {
if (ssDCval < 79) {
if (negBalance < 31) {
if (Cs < 1091) {
if (SC < 1972) {
return 1;
}
else if (SC >= 1972) {
return 8;
}
}
else if (Cs >= 1091) {
if (mu_mv_mag_sq < 126) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
if (Cs < 1286) {
return 8;
}
else if (Cs >= 1286) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 126) {
return 4;
}
}
}
else if (negBalance >= 31) {
if (mu_mv_mag_sq < 98) {
return 8;
}
else if (mu_mv_mag_sq >= 98) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (Rs < 1365) {
if (ssDCval < 70) {
if (diffMVdiffVal < -2) {
if (Rs < 1161) {
return 8;
}
else if (Rs >= 1161) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 1;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (Rs >= 1365) {
return 8;
}
}
}
}
}
else if (ssDCval >= 79) {
return 4;
}
}
else if (RsCsDiff >= 7) {
if (SC < 2560) {
if (mu_mv_mag_sq < 123) {
if (RsCsDiff < 16) {
if (Rs < 1158) {
if (Rs < 1093) {
if (diffRsCsdiff < -4) {
if (ssDCval < 69) {
if (Cs < 1261) {
return 4;
}
else if (Cs >= 1261) {
return 8;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
return 8;
}
}
}
else if (Rs >= 1093) {
if (Cs < 1041) {
return 1;
}
else if (Cs >= 1041) {
return 4;
}
}
}
else if (Rs >= 1158) {
if (Cs < 1214) {
if (mu_mv_mag_sq < 118) {
if (SC < 2390) {
if (SC < 2325) {
return 8;
}
else if (SC >= 2325) {
if (negBalance < 42) {
if (diffRsCsdiff < -3) {
return 8;
}
else if (diffRsCsdiff >= -3) {
return 1;
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
else if (SC >= 2390) {
return 8;
}
}
else if (mu_mv_mag_sq >= 118) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
if (Rs < 1325) {
return 4;
}
else if (Rs >= 1325) {
return 8;
}
}
}
}
else if (Cs >= 1214) {
if (Cs < 1256) {
return 1;
}
else if (Cs >= 1256) {
return 4;
}
}
}
}
else if (RsCsDiff >= 16) {
if (mu_mv_mag_sq < 90) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (posBalance < 50) {
if (Rs < 1055) {
return 2;
}
else if (Rs >= 1055) {
return 8;
}
}
else if (posBalance >= 50) {
if (RsCsDiff < 38) {
return 4;
}
else if (RsCsDiff >= 38) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (SC < 2440) {
return 4;
}
else if (SC >= 2440) {
if (SC < 2446) {
return 8;
}
else if (SC >= 2446) {
if (SC < 2493) {
return 4;
}
else if (SC >= 2493) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
return 1;
}
}
}
else if (SC >= 2560) {
if (Rs < 1335) {
return 1;
}
else if (Rs >= 1335) {
return 8;
}
}
}
}
}
else if (ssDCval >= 81) {
return 4;
}
}
}
else if (ssDCval >= 86) {
if (diffAFD < -2) {
if (Rs < 1279) {
return 4;
}
else if (Rs >= 1279) {
return 1;
}
}
else if (diffAFD >= -2) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1037) {
return 8;
}
else if (Cs >= 1037) {
if (ssDCval < 75) {
return 4;
}
else if (ssDCval >= 75) {
return 8;
}
}
}
}
else if (MVDiff >= 5) {
if (Cs < 1058) {
if (mu_mv_mag_sq < 13) {
if (Rs < 1340) {
return 1;
}
else if (Rs >= 1340) {
return 8;
}
}
else if (mu_mv_mag_sq >= 13) {
if (negBalance < 39) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
if (MVDiff < 27) {
return 8;
}
else if (MVDiff >= 27) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (tcor < 89) {
return 8;
}
else if (tcor >= 89) {
if (Cs < 966) {
return 8;
}
else if (Cs >= 966) {
if (Cs < 1049) {
if (mcTcor < 93) {
if (Rs < 1345) {
return 8;
}
else if (Rs >= 1345) {
return 4;
}
}
else if (mcTcor >= 93) {
return 4;
}
}
else if (Cs >= 1049) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1058) {
if (posBalance < 44) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
else if (posBalance >= 44) {
if (SC < 2258) {
if (Rs < 1026) {
if (mcTcor < 96) {
if (Cs < 1242) {
return 1;
}
else if (Cs >= 1242) {
return 4;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (Rs >= 1026) {
if (Rs < 1139) {
return 4;
}
else if (Rs >= 1139) {
return 1;
}
}
}
else if (SC >= 2258) {
if (negBalance < 44) {
if (mcTcor < 91) {
return 1;
}
else if (mcTcor >= 91) {
if (mu_mv_mag_sq < 101) {
if (AFD < 8) {
if (tcor < 95) {
if (Rs < 1410) {
return 8;
}
else if (Rs >= 1410) {
if (Rs < 1425) {
return 1;
}
else if (Rs >= 1425) {
return 8;
}
}
}
else if (tcor >= 95) {
return 4;
}
}
else if (AFD >= 8) {
if (Rs < 1305) {
return 4;
}
else if (Rs >= 1305) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 101) {
return 1;
}
}
}
else if (negBalance >= 44) {
if (TSC < 7) {
if (mu_mv_mag_sq < 81) {
return 4;
}
else if (mu_mv_mag_sq >= 81) {
return 8;
}
}
else if (TSC >= 7) {
if (RsCsDiff < 22) {
return 4;
}
else if (RsCsDiff >= 22) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 157) {
return 2;
}
}
else if (Rs >= 1454) {
if (ssDCval < 48) {
if (RsDiff < 5) {
if (posBalance < 44) {
if (Rs < 1749) {
if (negBalance < 18) {
return 8;
}
else if (negBalance >= 18) {
return 4;
}
}
else if (Rs >= 1749) {
if (negBalance < 18) {
if (Rs < 2203) {
if (Rs < 1770) {
if (Cs < 1153) {
return 4;
}
else if (Cs >= 1153) {
return 4;
}
}
else if (Rs >= 1770) {
if (Rs < 1822) {
return 4;
}
else if (Rs >= 1822) {
return 4;
}
}
}
else if (Rs >= 2203) {
return 8;
}
}
else if (negBalance >= 18) {
return 4;
}
}
}
else if (posBalance >= 44) {
return 8;
}
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (ssDCval >= 48) {
if (mu_mv_mag_sq < 104) {
if (negBalance < 25) {
if (ssDCval < 82) {
if (MVDiff < 2) {
if (Rs < 1974) {
return 1;
}
else if (Rs >= 1974) {
return 4;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (ssDCval >= 82) {
return 1;
}
}
else if (negBalance >= 25) {
if (Cs < 1071) {
if (posBalance < 46) {
if (TSC < 5) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (Cs < 1016) {
return 1;
}
else if (Cs >= 1016) {
return 2;
}
}
}
else if (TSC >= 5) {
return 4;
}
}
else if (posBalance >= 46) {
if (Rs < 2253) {
if (ssDCval < 124) {
if (Rs < 1602) {
return 8;
}
else if (Rs >= 1602) {
if (Rs < 1608) {
return 2;
}
else if (Rs >= 1608) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 2;
}
}
}
}
}
else if (ssDCval >= 124) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (Rs < 1646) {
return 4;
}
else if (Rs >= 1646) {
return 8;
}
}
}
}
else if (Rs >= 2253) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 4;
}
}
}
}
}
else if (Cs >= 1071) {
if (TSC < 6) {
if (Rs < 2365) {
if (Rs < 2045) {
if (ssDCval < 67) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (ssDCval >= 67) {
if (MVDiff < 4) {
if (RsCsDiff < 16) {
if (SC < 3003) {
return 2;
}
else if (SC >= 3003) {
return 8;
}
}
else if (RsCsDiff >= 16) {
return 2;
}
}
else if (MVDiff >= 4) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (TSC < 5) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (Rs >= 2045) {
return 8;
}
}
else if (Rs >= 2365) {
if (diffAFD < 1) {
if (diffAFD < -2) {
if (Rs < 2472) {
return 8;
}
else if (Rs >= 2472) {
return 1;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (diffAFD >= 1) {
if (ssDCval < 139) {
if (tcor < 94) {
if (Cs < 1179) {
return 8;
}
else if (Cs >= 1179) {
return 1;
}
}
else if (tcor >= 94) {
return 1;
}
}
else if (ssDCval >= 139) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (refDCval < 92) {
if (posBalance < 45) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
else if (posBalance >= 45) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (TSC < 8) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
if (Rs < 1822) {
if (Rs < 1512) {
return 1;
}
else if (Rs >= 1512) {
return 8;
}
}
else if (Rs >= 1822) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (MVDiff < 4) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
return 8;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
}
}
else if (TSC >= 8) {
if (Cs < 1106) {
return 2;
}
else if (Cs >= 1106) {
return 8;
}
}
}
}
}
else if (refDCval >= 92) {
if (Cs < 1156) {
if (Cs < 1152) {
return 8;
}
else if (Cs >= 1152) {
return 4;
}
}
else if (Cs >= 1156) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (ssDCval < 120) {
if (RsDiff < 4) {
if (Rs < 1466) {
if (Cs < 1267) {
return 8;
}
else if (Cs >= 1267) {
return 4;
}
}
else if (Rs >= 1466) {
if (diffMVdiffVal < -3) {
if (Cs < 1104) {
return 1;
}
else if (Cs >= 1104) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (diffTSC < -1) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (diffMVdiffVal < -1) {
if (Rs < 1896) {
return 2;
}
else if (Rs >= 1896) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (diffTSC >= -1) {
return 8;
}
}
}
}
else if (RsDiff >= 4) {
if (diffRsCsdiff < 19) {
if (Rs < 2502) {
if (Rs < 1803) {
return 8;
}
else if (Rs >= 1803) {
if (SC < 3174) {
return 2;
}
else if (SC >= 3174) {
if (Cs < 1301) {
if (mu_mv_mag_sq < 118) {
return 8;
}
else if (mu_mv_mag_sq >= 118) {
return 4;
}
}
else if (Cs >= 1301) {
return 2;
}
}
}
}
else if (Rs >= 2502) {
if (Cs < 1065) {
return 4;
}
else if (Cs >= 1065) {
return 2;
}
}
}
else if (diffRsCsdiff >= 19) {
return 4;
}
}
}
else if (ssDCval >= 120) {
if (Rs < 1918) {
return 2;
}
else if (Rs >= 1918) {
if (Rs < 2367) {
return 8;
}
else if (Rs >= 2367) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (posBalance >= 55) {
if (AFD < 15) {
if (negBalance < 46) {
if (RsDiff < 4) {
if (Rs < 1183) {
if (refDCval < 130) {
if (ssDCval < 51) {
if (diffRsCsdiff < -15) {
return 2;
}
else if (diffRsCsdiff >= -15) {
return 8;
}
}
else if (ssDCval >= 51) {
if (Cs < 1036) {
if (posBalance < 57) {
if (Cs < 596) {
return 4;
}
else if (Cs >= 596) {
if (tcor < 99) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
if (mu_mv_mag_sq < 91) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (Rs < 1143) {
return 1;
}
else if (Rs >= 1143) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 91) {
return 8;
}
}
}
else if (tcor >= 99) {
return 4;
}
}
}
else if (posBalance >= 57) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (SC < 1804) {
return 4;
}
else if (SC >= 1804) {
if (TSC < 8) {
if (diffRsCsdiff < -6) {
if (Rs < 1117) {
return 8;
}
else if (Rs >= 1117) {
return 1;
}
}
else if (diffRsCsdiff >= -6) {
return 8;
}
}
else if (TSC >= 8) {
if (Rs < 1120) {
return 8;
}
else if (Rs >= 1120) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1036) {
if (negBalance < 35) {
if (mu_mv_mag_sq < 111) {
if (RsCsDiff < 23) {
return 8;
}
else if (RsCsDiff >= 23) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 111) {
return 4;
}
}
else if (negBalance >= 35) {
if (CsDiff < 5) {
if (RsDiff < 3) {
if (Cs < 1206) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (Cs >= 1206) {
return 4;
}
}
else if (RsDiff >= 3) {
if (diffMVdiffVal < -4) {
return 8;
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 1092) {
return 2;
}
else if (Rs >= 1092) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 130) {
if (mu_mv_mag_sq < 16) {
if (Rs < 1019) {
return 8;
}
else if (Rs >= 1019) {
return 2;
}
}
else if (mu_mv_mag_sq >= 16) {
return 4;
}
}
}
else if (Rs >= 1183) {
if (refDCval < 78) {
if (refDCval < 64) {
if (SC < 2181) {
if (negBalance < 38) {
if (posBalance < 63) {
if (Rs < 1246) {
if (Cs < 819) {
return 2;
}
else if (Cs >= 819) {
return 1;
}
}
else if (Rs >= 1246) {
return 2;
}
}
else if (posBalance >= 63) {
return 8;
}
}
else if (negBalance >= 38) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
return 8;
}
}
}
else if (SC >= 2181) {
return 8;
}
}
else if (refDCval >= 64) {
if (SC < 2566) {
if (RsCsDiff < 16) {
if (Cs < 1222) {
return 8;
}
else if (Cs >= 1222) {
return 1;
}
}
else if (RsCsDiff >= 16) {
if (negBalance < 44) {
if (negBalance < 35) {
if (Rs < 1584) {
return 8;
}
else if (Rs >= 1584) {
return 1;
}
}
else if (negBalance >= 35) {
return 1;
}
}
else if (negBalance >= 44) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 4;
}
}
}
}
else if (SC >= 2566) {
if (mu_mv_mag_sq < 103) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
if (diffTSC < -1) {
if (Rs < 2026) {
return 2;
}
else if (Rs >= 2026) {
return 8;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (diffTSC < -1) {
if (Rs < 1546) {
return 4;
}
else if (Rs >= 1546) {
return 2;
}
}
else if (diffTSC >= -1) {
if (Rs < 1349) {
return 1;
}
else if (Rs >= 1349) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 78) {
if (ssDCval < 122) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (MVDiff < 12) {
if (Cs < 876) {
if (SC < 2546) {
return 8;
}
else if (SC >= 2546) {
if (Rs < 1856) {
return 1;
}
else if (Rs >= 1856) {
return 4;
}
}
}
else if (Cs >= 876) {
return 8;
}
}
else if (MVDiff >= 12) {
return 1;
}
}
}
else if (ssDCval >= 122) {
if (mu_mv_mag_sq < 94) {
if (MVDiff < 24) {
if (posBalance < 67) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
return 4;
}
}
else if (posBalance >= 67) {
if (Rs < 1455) {
return 8;
}
else if (Rs >= 1455) {
return 2;
}
}
}
else if (MVDiff >= 24) {
if (Rs < 1941) {
return 4;
}
else if (Rs >= 1941) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 94) {
if (ssDCval < 133) {
if (Rs < 1659) {
return 4;
}
else if (Rs >= 1659) {
return 1;
}
}
else if (ssDCval >= 133) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
return 1;
}
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (negBalance < 41) {
if (mcTcor < 95) {
if (ssDCval < 156) {
if (ssDCval < 54) {
if (Cs < 754) {
if (Rs < 1231) {
return 8;
}
else if (Rs >= 1231) {
return 2;
}
}
else if (Cs >= 754) {
if (mcTcor < 94) {
if (MVDiff < 1) {
if (Rs < 1301) {
return 1;
}
else if (Rs >= 1301) {
return 8;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (mcTcor >= 94) {
return 4;
}
}
}
else if (ssDCval >= 54) {
if (ssDCval < 76) {
if (Cs < 951) {
if (tcor < 89) {
if (tcor < 83) {
if (Rs < 1162) {
return 4;
}
else if (Rs >= 1162) {
return 8;
}
}
else if (tcor >= 83) {
return 2;
}
}
else if (tcor >= 89) {
return 1;
}
}
else if (Cs >= 951) {
if (Rs < 1518) {
return 4;
}
else if (Rs >= 1518) {
if (Rs < 1633) {
return 8;
}
else if (Rs >= 1633) {
return 2;
}
}
}
}
else if (ssDCval >= 76) {
if (ssDCval < 126) {
if (diffRsCsdiff < -16) {
if (Rs < 1568) {
return 8;
}
else if (Rs >= 1568) {
return 2;
}
}
else if (diffRsCsdiff >= -16) {
return 8;
}
}
else if (ssDCval >= 126) {
if (Rs < 1947) {
return 4;
}
else if (Rs >= 1947) {
return 8;
}
}
}
}
}
else if (ssDCval >= 156) {
return 2;
}
}
else if (mcTcor >= 95) {
if (ssDCval < 70) {
if (Cs < 949) {
return 2;
}
else if (Cs >= 949) {
if (ssDCval < 64) {
if (Rs < 1712) {
return 8;
}
else if (Rs >= 1712) {
return 4;
}
}
else if (ssDCval >= 64) {
if (Rs < 1908) {
return 1;
}
else if (Rs >= 1908) {
return 2;
}
}
}
}
else if (ssDCval >= 70) {
if (Cs < 825) {
if (posBalance < 73) {
if (MVDiff < 3) {
if (Rs < 1899) {
return 8;
}
else if (Rs >= 1899) {
return 2;
}
}
else if (MVDiff >= 3) {
if (ssDCval < 137) {
return 1;
}
else if (ssDCval >= 137) {
return 8;
}
}
}
else if (posBalance >= 73) {
if (Cs < 686) {
return 2;
}
else if (Cs >= 686) {
return 8;
}
}
}
else if (Cs >= 825) {
if (MVDiff < 12) {
if (SC < 2997) {
if (Rs < 1645) {
if (Rs < 1044) {
return 2;
}
else if (Rs >= 1044) {
return 8;
}
}
else if (Rs >= 1645) {
return 2;
}
}
else if (SC >= 2997) {
if (negBalance < 33) {
if (Cs < 1274) {
return 8;
}
else if (Cs >= 1274) {
return 4;
}
}
else if (negBalance >= 33) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
}
}
else if (MVDiff >= 12) {
return 8;
}
}
}
}
}
else if (negBalance >= 41) {
if (ssDCval < 93) {
if (refDCval < 52) {
if (Cs < 665) {
if (Rs < 976) {
return 8;
}
else if (Rs >= 976) {
return 2;
}
}
else if (Cs >= 665) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 86) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
return 4;
}
}
else if (mu_mv_mag_sq >= 86) {
if (Rs < 1317) {
return 1;
}
else if (Rs >= 1317) {
return 2;
}
}
}
}
}
else if (refDCval >= 52) {
if (Rs < 1200) {
if (AFD < 10) {
if (Cs < 895) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (Cs >= 895) {
return 4;
}
}
else if (AFD >= 10) {
if (diffTSC < -1) {
if (Cs < 979) {
return 2;
}
else if (Cs >= 979) {
return 1;
}
}
else if (diffTSC >= -1) {
if (Cs < 1219) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 1;
}
}
else if (Cs >= 1219) {
return 4;
}
}
}
}
else if (Rs >= 1200) {
if (Cs < 1168) {
if (mu_mv_mag_sq < 25) {
if (MVDiff < 7) {
if (Rs < 1964) {
return 1;
}
else if (Rs >= 1964) {
return 2;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (mu_mv_mag_sq >= 25) {
if (diffRsCsdiff < 8) {
if (Rs < 1478) {
if (Rs < 1311) {
return 8;
}
else if (Rs >= 1311) {
if (RsCsDiff < 37) {
if (Cs < 1120) {
return 4;
}
else if (Cs >= 1120) {
return 8;
}
}
else if (RsCsDiff >= 37) {
return 8;
}
}
}
else if (Rs >= 1478) {
if (Rs < 2224) {
return 8;
}
else if (Rs >= 2224) {
return 4;
}
}
}
else if (diffRsCsdiff >= 8) {
if (SC < 2614) {
if (Cs < 773) {
return 8;
}
else if (Cs >= 773) {
return 1;
}
}
else if (SC >= 2614) {
return 4;
}
}
}
}
else if (Cs >= 1168) {
if (tcor < 91) {
if (ssDCval < 69) {
if (SC < 2770) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (SC >= 2770) {
if (Rs < 1589) {
return 4;
}
else if (Rs >= 1589) {
return 8;
}
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (tcor >= 91) {
return 1;
}
}
}
}
}
else if (ssDCval >= 93) {
if (posBalance < 63) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
if (ssDCval < 136) {
return 8;
}
else if (ssDCval >= 136) {
return 4;
}
}
}
else if (posBalance >= 63) {
if (Cs < 1164) {
if (Rs < 1439) {
if (Cs < 822) {
return 8;
}
else if (Cs >= 822) {
return 4;
}
}
else if (Rs >= 1439) {
return 8;
}
}
else if (Cs >= 1164) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (ssDCval < 112) {
if (mcTcor < 92) {
if (TSC < 7) {
if (ssDCval < 68) {
if (Rs < 1131) {
if (Rs < 1128) {
if (negBalance < 52) {
if (RsDiff < 3) {
if (Rs < 1025) {
return 1;
}
else if (Rs >= 1025) {
return 8;
}
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (negBalance >= 52) {
return 1;
}
}
else if (Rs >= 1128) {
return 1;
}
}
else if (Rs >= 1131) {
return 8;
}
}
else if (ssDCval >= 68) {
if (refDCval < 73) {
if (Cs < 766) {
return 2;
}
else if (Cs >= 766) {
return 8;
}
}
else if (refDCval >= 73) {
if (AFD < 10) {
if (refDCval < 98) {
if (Cs < 948) {
if (MVDiff < 24) {
return 2;
}
else if (MVDiff >= 24) {
return 8;
}
}
else if (Cs >= 948) {
return 8;
}
}
else if (refDCval >= 98) {
if (Cs < 1074) {
return 8;
}
else if (Cs >= 1074) {
return 2;
}
}
}
else if (AFD >= 10) {
return 8;
}
}
}
}
else if (TSC >= 7) {
if (diffRsCsdiff < -8) {
if (SC < 1699) {
if (SC < 1620) {
return 8;
}
else if (SC >= 1620) {
return 4;
}
}
else if (SC >= 1699) {
if (diffTSC < 2) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 4;
}
}
else if (diffTSC >= 2) {
return 4;
}
}
}
else if (diffRsCsdiff >= -8) {
if (Rs < 1617) {
if (tcor < 67) {
if (SC < 1761) {
return 4;
}
else if (SC >= 1761) {
return 8;
}
}
else if (tcor >= 67) {
return 8;
}
}
else if (Rs >= 1617) {
if (AFD < 10) {
if (mcTcor < 87) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
return 2;
}
}
else if (mcTcor >= 87) {
return 2;
}
}
else if (AFD >= 10) {
return 8;
}
}
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 83) {
if (RsCsDiff < 19) {
return 8;
}
else if (RsCsDiff >= 19) {
if (refDCval < 62) {
if (Cs < 1038) {
return 8;
}
else if (Cs >= 1038) {
if (Rs < 1564) {
return 4;
}
else if (Rs >= 1564) {
return 8;
}
}
}
else if (refDCval >= 62) {
if (ssDCval < 75) {
if (MVDiff < 8) {
if (AFD < 8) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (Rs < 1641) {
return 8;
}
else if (Rs >= 1641) {
return 2;
}
}
}
else if (AFD >= 8) {
if (RsDiff < 4) {
if (Rs < 1082) {
return 1;
}
else if (Rs >= 1082) {
return 8;
}
}
else if (RsDiff >= 4) {
return 2;
}
}
}
else if (MVDiff >= 8) {
if (negBalance < 51) {
if (mu_mv_mag_sq < 56) {
if (Cs < 566) {
return 2;
}
else if (Cs >= 566) {
return 4;
}
}
else if (mu_mv_mag_sq >= 56) {
if (Rs < 2002) {
return 8;
}
else if (Rs >= 2002) {
return 1;
}
}
}
else if (negBalance >= 51) {
return 8;
}
}
}
else if (ssDCval >= 75) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (Rs < 1243) {
return 2;
}
else if (Rs >= 1243) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 83) {
if (posBalance < 58) {
if (MVDiff < 8) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
else if (MVDiff >= 8) {
if (Cs < 844) {
if (Rs < 1551) {
return 8;
}
else if (Rs >= 1551) {
return 2;
}
}
else if (Cs >= 844) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (SC < 2471) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (Rs < 1377) {
return 8;
}
else if (Rs >= 1377) {
return 1;
}
}
}
else if (SC >= 2471) {
if (Rs < 1580) {
return 2;
}
else if (Rs >= 1580) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 112) {
if (MVDiff < 17) {
if (mu_mv_mag_sq < 106) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Cs < 877) {
return 8;
}
else if (Cs >= 877) {
if (Rs < 1370) {
return 4;
}
else if (Rs >= 1370) {
if (Rs < 1389) {
return 1;
}
else if (Rs >= 1389) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (Rs < 1493) {
if (Rs < 1330) {
return 8;
}
else if (Rs >= 1330) {
return 4;
}
}
else if (Rs >= 1493) {
return 1;
}
}
}
else if (MVDiff >= 17) {
if (TSC < 10) {
if (negBalance < 54) {
if (posBalance < 58) {
if (Cs < 808) {
return 8;
}
else if (Cs >= 808) {
if (Rs < 1419) {
return 4;
}
else if (Rs >= 1419) {
return 8;
}
}
}
else if (posBalance >= 58) {
return 1;
}
}
else if (negBalance >= 54) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
else if (AFD >= 15) {
if (refDCval < 88) {
if (posBalance < 57) {
if (Rs < 1220) {
return 1;
}
else if (Rs >= 1220) {
return 4;
}
}
else if (posBalance >= 57) {
if (ssDCval < 90) {
if (diffRsCsdiff < -27) {
return 4;
}
else if (diffRsCsdiff >= -27) {
return 8;
}
}
else if (ssDCval >= 90) {
return 2;
}
}
}
else if (refDCval >= 88) {
if (mu_mv_mag_sq < 2) {
if (Cs < 820) {
return 8;
}
else if (Cs >= 820) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 1178) {
return 8;
}
else if (Cs >= 1178) {
return 4;
}
}
}
}
}
}
}
else if (RsCsDiff >= 43) {
if (ssDCval < 41) {
if (AFD < 8) {
if (diffAFD < -2) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (AFD >= 8) {
if (MVDiff < 33) {
if (negBalance < 40) {
return 1;
}
else if (negBalance >= 40) {
return 8;
}
}
else if (MVDiff >= 33) {
if (Rs < 597) {
return 2;
}
else if (Rs >= 597) {
return 1;
}
}
}
}
else if (ssDCval >= 41) {
if (Cs < 872) {
if (Rs < 1762) {
if (refDCval < 123) {
if (tcor < 81) {
if (CsDiff < 5) {
if (AFD < 14) {
if (AFD < 13) {
if (ssDCval < 45) {
return 1;
}
else if (ssDCval >= 45) {
if (negBalance < 58) {
if (mcTcor < 91) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
return 8;
}
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (negBalance >= 58) {
if (Rs < 1648) {
return 2;
}
else if (Rs >= 1648) {
return 8;
}
}
}
}
else if (AFD >= 13) {
if (Rs < 1057) {
if (Rs < 893) {
return 1;
}
else if (Rs >= 893) {
return 4;
}
}
else if (Rs >= 1057) {
return 2;
}
}
}
else if (AFD >= 14) {
if (mu_mv_mag_sq < 19) {
if (diffMVdiffVal < -1) {
if (negBalance < 49) {
if (diffMVdiffVal < -12) {
if (Rs < 1195) {
return 8;
}
else if (Rs >= 1195) {
return 2;
}
}
else if (diffMVdiffVal >= -12) {
return 1;
}
}
else if (negBalance >= 49) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
if (AFD < 24) {
if (SC < 1925) {
return 4;
}
else if (SC >= 1925) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (ssDCval < 114) {
return 8;
}
else if (ssDCval >= 114) {
return 4;
}
}
}
}
else if (AFD >= 24) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 19) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
if (mu_mv_mag_sq < 39) {
return 4;
}
else if (mu_mv_mag_sq >= 39) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
}
}
}
}
else if (CsDiff >= 5) {
if (gchDC < 3) {
if (ssDCval < 50) {
if (refDCval < 44) {
if (Cs < 725) {
return 8;
}
else if (Cs >= 725) {
return 2;
}
}
else if (refDCval >= 44) {
if (diffRsCsdiff < -10) {
return 4;
}
else if (diffRsCsdiff >= -10) {
if (refDCval < 51) {
if (Rs < 538) {
return 4;
}
else if (Rs >= 538) {
return 8;
}
}
else if (refDCval >= 51) {
return 4;
}
}
}
}
else if (ssDCval >= 50) {
if (tcor < 55) {
if (MVDiff < 63) {
if (TSC < 9) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
return 4;
}
}
else if (TSC >= 9) {
if (mcTcor < 88) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (diffRsCsdiff < 33) {
return 1;
}
else if (diffRsCsdiff >= 33) {
return 4;
}
}
}
else if (mcTcor >= 88) {
if (Rs < 1491) {
return 2;
}
else if (Rs >= 1491) {
return 4;
}
}
}
}
else if (MVDiff >= 63) {
if (refDCval < 61) {
return 8;
}
else if (refDCval >= 61) {
if (diffMVdiffVal < 18) {
if (Rs < 965) {
return 1;
}
else if (Rs >= 965) {
return 2;
}
}
else if (diffMVdiffVal >= 18) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 2;
}
}
}
}
}
else if (tcor >= 55) {
if (RsCsDiff < 60) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
if (Rs < 936) {
return 4;
}
else if (Rs >= 936) {
return 1;
}
}
}
}
else if (RsCsDiff >= 60) {
if (diffRsCsdiff < -17) {
if (negBalance < 48) {
if (Rs < 849) {
return 4;
}
else if (Rs >= 849) {
return 2;
}
}
else if (negBalance >= 48) {
if (SC < 1489) {
if (Rs < 530) {
return 4;
}
else if (Rs >= 530) {
return 1;
}
}
else if (SC >= 1489) {
if (tcor < 74) {
if (Rs < 979) {
return 8;
}
else if (Rs >= 979) {
if (refDCval < 57) {
return 8;
}
else if (refDCval >= 57) {
if (MVDiff < 121) {
return 4;
}
else if (MVDiff >= 121) {
return 8;
}
}
}
}
else if (tcor >= 74) {
if (Rs < 1496) {
return 8;
}
else if (Rs >= 1496) {
return 1;
}
}
}
}
}
else if (diffRsCsdiff >= -17) {
if (mu_mv_mag_sq < 47) {
if (MVDiff < 79) {
if (mcTcor < 95) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (AFD < 20) {
if (diffRsCsdiff < 30) {
if (mcTcor < 92) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
if (gchDC < 1) {
if (Rs < 980) {
return 1;
}
else if (Rs >= 980) {
return 8;
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
else if (mcTcor >= 92) {
return 4;
}
}
else if (diffRsCsdiff >= 30) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 4;
}
}
}
else if (AFD >= 20) {
if (gchDC < 2) {
if (RsCsDiff < 118) {
return 1;
}
else if (RsCsDiff >= 118) {
if (Rs < 1450) {
return 2;
}
else if (Rs >= 1450) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (Cs < 638) {
return 4;
}
else if (Cs >= 638) {
return 2;
}
}
}
}
}
else if (mcTcor >= 95) {
if (negBalance < 53) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 1;
}
}
else if (negBalance >= 53) {
return 8;
}
}
}
else if (MVDiff >= 79) {
return 2;
}
}
else if (mu_mv_mag_sq >= 47) {
if (Cs < 773) {
return 4;
}
else if (Cs >= 773) {
if (Rs < 1088) {
return 8;
}
else if (Rs >= 1088) {
if (Rs < 1544) {
return 4;
}
else if (Rs >= 1544) {
return 2;
}
}
}
}
}
}
}
}
}
else if (gchDC >= 3) {
if (Rs < 1115) {
if (Rs < 1017) {
if (posBalance < 80) {
if (ssDCval < 63) {
if (RsCsDiff < 60) {
if (Cs < 644) {
return 4;
}
else if (Cs >= 644) {
return 8;
}
}
else if (RsCsDiff >= 60) {
if (Cs < 544) {
return 2;
}
else if (Cs >= 544) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 8;
}
}
}
}
else if (ssDCval >= 63) {
if (ssDCval < 116) {
if (mu_mv_mag_sq < 78) {
if (mu_mv_mag_sq < 19) {
if (mcTcor < 89) {
if (TSC < 19) {
return 1;
}
else if (TSC >= 19) {
return 4;
}
}
else if (mcTcor >= 89) {
if (AFD < 21) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (AFD >= 21) {
if (MVDiff < 68) {
if (Rs < 674) {
return 1;
}
else if (Rs >= 674) {
return 8;
}
}
else if (MVDiff >= 68) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 19) {
if (CsDiff < 7) {
if (Cs < 774) {
return 8;
}
else if (Cs >= 774) {
return 2;
}
}
else if (CsDiff >= 7) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (diffAFD < 7) {
if (Rs < 691) {
return 1;
}
else if (Rs >= 691) {
return 4;
}
}
else if (diffAFD >= 7) {
return 2;
}
}
}
else if (ssDCval >= 116) {
return 2;
}
}
}
else if (posBalance >= 80) {
if (negBalance < 30) {
if (Rs < 378) {
return 2;
}
else if (Rs >= 378) {
return 1;
}
}
else if (negBalance >= 30) {
return 4;
}
}
}
else if (Rs >= 1017) {
if (Rs < 1043) {
return 2;
}
else if (Rs >= 1043) {
if (SC < 1820) {
return 1;
}
else if (SC >= 1820) {
return 2;
}
}
}
}
else if (Rs >= 1115) {
if (Cs < 705) {
if (diffTSC < 2) {
return 4;
}
else if (diffTSC >= 2) {
if (Rs < 1481) {
return 2;
}
else if (Rs >= 1481) {
return 1;
}
}
}
else if (Cs >= 705) {
if (posBalance < 65) {
if (ssDCval < 86) {
if (mcTcor < 89) {
if (Cs < 854) {
return 8;
}
else if (Cs >= 854) {
return 2;
}
}
else if (mcTcor >= 89) {
return 2;
}
}
else if (ssDCval >= 86) {
return 8;
}
}
else if (posBalance >= 65) {
if (SC < 1955) {
return 4;
}
else if (SC >= 1955) {
if (MVDiff < 100) {
return 2;
}
else if (MVDiff >= 100) {
if (Rs < 1678) {
return 8;
}
else if (Rs >= 1678) {
return 4;
}
}
}
}
}
}
}
}
}
else if (tcor >= 81) {
if (Rs < 1357) {
if (mu_mv_mag_sq < 7) {
if (tcor < 86) {
if (mcTcor < 92) {
if (Cs < 757) {
return 1;
}
else if (Cs >= 757) {
return 2;
}
}
else if (mcTcor >= 92) {
if (MVDiff < 88) {
return 2;
}
else if (MVDiff >= 88) {
return 1;
}
}
}
else if (tcor >= 86) {
if (Cs < 727) {
if (negBalance < 44) {
return 2;
}
else if (negBalance >= 44) {
return 1;
}
}
else if (Cs >= 727) {
if (negBalance < 54) {
if (Cs < 802) {
return 1;
}
else if (Cs >= 802) {
return 4;
}
}
else if (negBalance >= 54) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (negBalance < 48) {
if (Cs < 707) {
if (mu_mv_mag_sq < 43) {
return 4;
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 511) {
return 2;
}
else if (Rs >= 511) {
if (Cs < 557) {
if (Rs < 1054) {
return 8;
}
else if (Rs >= 1054) {
return 2;
}
}
else if (Cs >= 557) {
return 8;
}
}
}
}
else if (Cs >= 707) {
if (RsCsDiff < 51) {
if (Rs < 1175) {
return 1;
}
else if (Rs >= 1175) {
return 2;
}
}
else if (RsCsDiff >= 51) {
if (diffAFD < 1) {
if (Cs < 729) {
if (Rs < 928) {
return 8;
}
else if (Rs >= 928) {
return 1;
}
}
else if (Cs >= 729) {
return 2;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
if (AFD < 22) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
if (SC < 1421) {
return 8;
}
else if (SC >= 1421) {
return 4;
}
}
}
else if (AFD >= 22) {
return 2;
}
}
}
}
else if (Rs >= 1357) {
if (Cs < 648) {
if (RsDiff < 10) {
if (diffAFD < -2) {
if (Rs < 1722) {
return 2;
}
else if (Rs >= 1722) {
return 1;
}
}
else if (diffAFD >= -2) {
if (gchDC < 1) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
return 1;
}
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (RsDiff >= 10) {
if (Cs < 593) {
return 2;
}
else if (Cs >= 593) {
return 8;
}
}
}
else if (Cs >= 648) {
if (ssDCval < 90) {
if (mcTcor < 95) {
if (RsCsDiff < 60) {
if (negBalance < 49) {
if (Rs < 1564) {
return 2;
}
else if (Rs >= 1564) {
return 4;
}
}
else if (negBalance >= 49) {
if (SC < 2217) {
if (Rs < 1445) {
return 2;
}
else if (Rs >= 1445) {
return 1;
}
}
else if (SC >= 2217) {
return 8;
}
}
}
else if (RsCsDiff >= 60) {
if (posBalance < 63) {
return 4;
}
else if (posBalance >= 63) {
return 2;
}
}
}
else if (mcTcor >= 95) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (ssDCval >= 90) {
if (posBalance < 54) {
if (AFD < 15) {
if (Rs < 1479) {
return 1;
}
else if (Rs >= 1479) {
return 4;
}
}
else if (AFD >= 15) {
return 2;
}
}
else if (posBalance >= 54) {
if (diffMVdiffVal < -5) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 123) {
if (Cs < 575) {
if (mu_mv_mag_sq < 18) {
if (Rs < 955) {
return 8;
}
else if (Rs >= 955) {
return 1;
}
}
else if (mu_mv_mag_sq >= 18) {
if (Rs < 1493) {
if (Rs < 1058) {
return 8;
}
else if (Rs >= 1058) {
return 2;
}
}
else if (Rs >= 1493) {
return 4;
}
}
}
else if (Cs >= 575) {
if (SC < 1837) {
if (gchDC < 2) {
if (Rs < 1019) {
if (ssDCval < 134) {
return 2;
}
else if (ssDCval >= 134) {
return 8;
}
}
else if (Rs >= 1019) {
return 2;
}
}
else if (gchDC >= 2) {
if (CsDiff < 11) {
if (posBalance < 74) {
return 2;
}
else if (posBalance >= 74) {
return 8;
}
}
else if (CsDiff >= 11) {
return 8;
}
}
}
else if (SC >= 1837) {
if (RsCsDiff < 183) {
if (diffTSC < 2) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
}
else if (diffTSC >= 2) {
if (CsDiff < 5) {
if (diffMVdiffVal < 7) {
return 1;
}
else if (diffMVdiffVal >= 7) {
return 8;
}
}
else if (CsDiff >= 5) {
if (Rs < 1491) {
return 8;
}
else if (Rs >= 1491) {
return 2;
}
}
}
}
else if (RsCsDiff >= 183) {
if (Rs < 1730) {
return 2;
}
else if (Rs >= 1730) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1762) {
if (refDCval < 77) {
if (ssDCval < 57) {
if (MVDiff < 6) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
else if (MVDiff >= 6) {
if (RsCsDiff < 142) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (mcTcor < 92) {
if (Rs < 1870) {
return 2;
}
else if (Rs >= 1870) {
return 4;
}
}
else if (mcTcor >= 92) {
return 2;
}
}
}
else if (RsCsDiff >= 142) {
return 2;
}
}
}
else if (ssDCval >= 57) {
if (MVDiff < 35) {
if (posBalance < 47) {
if (RsDiff < 9) {
if (ssDCval < 71) {
if (MVDiff < 16) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (Rs < 2205) {
return 8;
}
else if (Rs >= 2205) {
return 4;
}
}
}
else if (MVDiff >= 16) {
return 1;
}
}
else if (ssDCval >= 71) {
return 1;
}
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (posBalance >= 47) {
if (mcTcor < 90) {
if (posBalance < 61) {
if (Rs < 1814) {
return 8;
}
else if (Rs >= 1814) {
return 4;
}
}
else if (posBalance >= 61) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (mcTcor >= 90) {
if (diffRsCsdiff < -335) {
return 4;
}
else if (diffRsCsdiff >= -335) {
if (Cs < 847) {
return 2;
}
else if (Cs >= 847) {
if (Cs < 850) {
return 8;
}
else if (Cs >= 850) {
if (MVDiff < 21) {
return 2;
}
else if (MVDiff >= 21) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 35) {
if (AFD < 25) {
if (diffRsCsdiff < 25) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (diffRsCsdiff >= 25) {
if (tcor < 76) {
return 8;
}
else if (tcor >= 76) {
return 2;
}
}
}
else if (AFD >= 25) {
if (Rs < 2522) {
return 1;
}
else if (Rs >= 2522) {
return 4;
}
}
}
}
}
else if (refDCval >= 77) {
if (refDCval < 113) {
if (MVDiff < 36) {
if (TSC < 13) {
if (TSC < 6) {
if (ssDCval < 90) {
if (Cs < 811) {
if (AFD < 12) {
if (mu_mv_mag_sq < 58) {
return 1;
}
else if (mu_mv_mag_sq >= 58) {
if (Rs < 2368) {
return 1;
}
else if (Rs >= 2368) {
return 2;
}
}
}
else if (AFD >= 12) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 4;
}
}
}
else if (Cs >= 811) {
return 2;
}
}
else if (ssDCval >= 90) {
if (negBalance < 50) {
return 4;
}
else if (negBalance >= 50) {
if (Rs < 1891) {
return 1;
}
else if (Rs >= 1891) {
return 2;
}
}
}
}
else if (TSC >= 6) {
if (Cs < 781) {
if (RsCsDiff < 94) {
if (mu_mv_mag_sq < 12) {
if (gchDC < 2) {
if (ssDCval < 101) {
if (AFD < 11) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 2;
}
}
else if (AFD >= 11) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 1;
}
}
}
else if (ssDCval >= 101) {
return 2;
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 12) {
if (diffRsCsdiff < 31) {
return 2;
}
else if (diffRsCsdiff >= 31) {
return 1;
}
}
}
else if (RsCsDiff >= 94) {
if (AFD < 23) {
if (mcTcor < 90) {
if (Rs < 2390) {
return 4;
}
else if (Rs >= 2390) {
return 2;
}
}
else if (mcTcor >= 90) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
if (diffAFD < 5) {
if (SC < 2373) {
return 1;
}
else if (SC >= 2373) {
return 2;
}
}
else if (diffAFD >= 5) {
if (Rs < 2323) {
return 4;
}
else if (Rs >= 2323) {
return 2;
}
}
}
}
}
else if (AFD >= 23) {
if (refDCval < 98) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 1;
}
}
else if (refDCval >= 98) {
return 1;
}
}
}
}
else if (Cs >= 781) {
if (ssDCval < 102) {
if (diffRsCsdiff < 19) {
if (negBalance < 52) {
if (tcor < 82) {
if (Rs < 1836) {
return 2;
}
else if (Rs >= 1836) {
return 8;
}
}
else if (tcor >= 82) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 2;
}
}
}
}
else if (negBalance >= 52) {
return 1;
}
}
else if (diffRsCsdiff >= 19) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 4;
}
}
}
else if (ssDCval >= 102) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
}
}
else if (TSC >= 13) {
return 1;
}
}
else if (MVDiff >= 36) {
if (mu_mv_mag_sq < 10) {
if (mcTcor < 94) {
if (posBalance < 88) {
return 8;
}
else if (posBalance >= 88) {
return 2;
}
}
else if (mcTcor >= 94) {
if (negBalance < 57) {
if (MVDiff < 41) {
return 2;
}
else if (MVDiff >= 41) {
return 1;
}
}
else if (negBalance >= 57) {
if (diffAFD < -6) {
if (Rs < 2284) {
return 2;
}
else if (Rs >= 2284) {
return 1;
}
}
else if (diffAFD >= -6) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (MVDiff < 50) {
if (gchDC < 4) {
return 1;
}
else if (gchDC >= 4) {
return 8;
}
}
else if (MVDiff >= 50) {
if (Rs < 2296) {
return 4;
}
else if (Rs >= 2296) {
return 2;
}
}
}
}
}
else if (refDCval >= 113) {
if (mcTcor < 91) {
if (AFD < 27) {
if (Rs < 2143) {
if (Cs < 552) {
return 8;
}
else if (Cs >= 552) {
return 4;
}
}
else if (Rs >= 2143) {
if (AFD < 19) {
if (SC < 3091) {
if (Rs < 2203) {
return 8;
}
else if (Rs >= 2203) {
return 4;
}
}
else if (SC >= 3091) {
return 1;
}
}
else if (AFD >= 19) {
return 8;
}
}
}
else if (AFD >= 27) {
if (Rs < 1896) {
return 1;
}
else if (Rs >= 1896) {
return 2;
}
}
}
else if (mcTcor >= 91) {
if (TSC < 6) {
if (Rs < 2459) {
return 8;
}
else if (Rs >= 2459) {
return 1;
}
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 35) {
if (Rs < 1769) {
return 2;
}
else if (Rs >= 1769) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 35) {
if (tcor < 90) {
if (SC < 2560) {
return 1;
}
else if (SC >= 2560) {
if (Cs < 715) {
return 2;
}
else if (Cs >= 715) {
if (Rs < 1949) {
return 2;
}
else if (Rs >= 1949) {
return 8;
}
}
}
}
else if (tcor >= 90) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 872) {
if (RsCsDiff < 87) {
if (Rs < 1298) {
if (ssDCval < 63) {
if (Rs < 1003) {
if (MVDiff < 50) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
if (RsDiff < 6) {
if (tcor < 63) {
if (refDCval < 43) {
return 2;
}
else if (refDCval >= 43) {
return 4;
}
}
else if (tcor >= 63) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (RsDiff < 5) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (MVDiff < 14) {
return 4;
}
else if (MVDiff >= 14) {
if (Cs < 1018) {
return 8;
}
else if (Cs >= 1018) {
return 2;
}
}
}
}
else if (RsDiff >= 5) {
if (Cs < 1200) {
return 8;
}
else if (Cs >= 1200) {
return 2;
}
}
}
}
}
else if (RsDiff >= 6) {
if (Cs < 1034) {
return 4;
}
else if (Cs >= 1034) {
return 2;
}
}
}
}
}
else if (MVDiff >= 50) {
if (Cs < 1006) {
return 8;
}
else if (Cs >= 1006) {
return 2;
}
}
}
else if (Rs >= 1003) {
if (ssDCval < 46) {
return 8;
}
else if (ssDCval >= 46) {
if (refDCval < 57) {
if (RsDiff < 6) {
if (diffMVdiffVal < 3) {
if (Cs < 1041) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
else if (Cs >= 1041) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
return 8;
}
}
else if (RsDiff >= 6) {
return 4;
}
}
else if (refDCval >= 57) {
if (Rs < 1234) {
if (Cs < 1275) {
return 8;
}
else if (Cs >= 1275) {
return 4;
}
}
else if (Rs >= 1234) {
if (Rs < 1246) {
return 1;
}
else if (Rs >= 1246) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 63) {
if (Rs < 821) {
if (MVDiff < 23) {
if (mcTcor < 96) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (mcTcor >= 96) {
if (diffRsCsdiff < -18) {
return 1;
}
else if (diffRsCsdiff >= -18) {
return 2;
}
}
}
else if (MVDiff >= 23) {
if (TSC < 13) {
if (TSC < 9) {
if (AFD < 26) {
if (Cs < 882) {
return 2;
}
else if (Cs >= 882) {
return 4;
}
}
else if (AFD >= 26) {
return 1;
}
}
else if (TSC >= 9) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
if (ssDCval < 66) {
return 2;
}
else if (ssDCval >= 66) {
return 4;
}
}
}
}
}
else if (TSC >= 13) {
return 8;
}
}
}
else if (Rs >= 821) {
if (diffRsCsdiff < 19) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 7) {
if (Cs < 1036) {
if (Rs < 1038) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
return 2;
}
}
else if (Rs >= 1038) {
if (RsCsDiff < 51) {
if (Rs < 1160) {
return 4;
}
else if (Rs >= 1160) {
return 8;
}
}
else if (RsCsDiff >= 51) {
return 8;
}
}
}
else if (Cs >= 1036) {
if (mcTcor < 79) {
return 8;
}
else if (mcTcor >= 79) {
if (ssDCval < 74) {
if (Rs < 875) {
if (Cs < 1246) {
return 8;
}
else if (Cs >= 1246) {
return 2;
}
}
else if (Rs >= 875) {
if (RsCsDiff < 57) {
return 2;
}
else if (RsCsDiff >= 57) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 2;
}
}
}
}
}
else if (ssDCval >= 74) {
if (diffMVdiffVal < -3) {
if (AFD < 14) {
if (Cs < 1164) {
return 8;
}
else if (Cs >= 1164) {
return 1;
}
}
else if (AFD >= 14) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
if (MVDiff < 18) {
if (ssDCval < 99) {
if (RsCsDiff < 57) {
return 4;
}
else if (RsCsDiff >= 57) {
if (Cs < 1159) {
return 4;
}
else if (Cs >= 1159) {
return 8;
}
}
}
else if (ssDCval >= 99) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
else if (MVDiff >= 18) {
if (mu_mv_mag_sq < 4) {
return 2;
}
else if (mu_mv_mag_sq >= 4) {
if (Rs < 1089) {
return 8;
}
else if (Rs >= 1089) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (ssDCval < 134) {
if (Cs < 937) {
if (Rs < 1149) {
return 1;
}
else if (Rs >= 1149) {
return 4;
}
}
else if (Cs >= 937) {
if (Rs < 973) {
if (Rs < 968) {
return 8;
}
else if (Rs >= 968) {
return 1;
}
}
else if (Rs >= 973) {
return 8;
}
}
}
else if (ssDCval >= 134) {
if (ssDCval < 152) {
return 1;
}
else if (ssDCval >= 152) {
if (Rs < 1075) {
return 4;
}
else if (Rs >= 1075) {
return 2;
}
}
}
}
}
else if (gchDC >= 2) {
if (SC < 2189) {
if (gchDC < 3) {
if (Cs < 1212) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (diffMVdiffVal < -9) {
return 8;
}
else if (diffMVdiffVal >= -9) {
return 2;
}
}
}
else if (Cs >= 1212) {
return 1;
}
}
else if (gchDC >= 3) {
if (tcor < 78) {
if (AFD < 33) {
return 4;
}
else if (AFD >= 33) {
return 1;
}
}
else if (tcor >= 78) {
if (diffAFD < -3) {
if (Rs < 877) {
return 8;
}
else if (Rs >= 877) {
if (Rs < 999) {
return 4;
}
else if (Rs >= 999) {
return 1;
}
}
}
else if (diffAFD >= -3) {
return 8;
}
}
}
}
else if (SC >= 2189) {
if (gchDC < 6) {
return 2;
}
else if (gchDC >= 6) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 19) {
if (ssDCval < 126) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (Cs < 1096) {
if (diffTSC < 1) {
if (Rs < 1173) {
if (Rs < 947) {
return 2;
}
else if (Rs >= 947) {
if (Rs < 1122) {
return 4;
}
else if (Rs >= 1122) {
return 2;
}
}
}
else if (Rs >= 1173) {
return 1;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (Cs >= 1096) {
if (mu_mv_mag_sq < 14) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (mu_mv_mag_sq >= 14) {
if (Cs < 1232) {
return 1;
}
else if (Cs >= 1232) {
return 8;
}
}
}
}
}
else if (ssDCval >= 126) {
return 1;
}
}
}
}
}
else if (Rs >= 1298) {
if (posBalance < 53) {
if (refDCval < 108) {
if (CsDiff < 5) {
if (negBalance < 41) {
if (ssDCval < 65) {
if (TSC < 5) {
if (Cs < 1054) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 2;
}
}
else if (Cs >= 1054) {
if (Cs < 1150) {
return 1;
}
else if (Cs >= 1150) {
return 8;
}
}
}
else if (TSC >= 5) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (Rs < 1901) {
return 8;
}
else if (Rs >= 1901) {
return 2;
}
}
}
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 60) {
return 1;
}
else if (mu_mv_mag_sq >= 60) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (mu_mv_mag_sq < 2) {
if (Cs < 936) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
else if (Cs >= 936) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
if (Cs < 1130) {
if (Rs < 1609) {
return 1;
}
else if (Rs >= 1609) {
return 4;
}
}
else if (Cs >= 1130) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsDiff < 7) {
if (ssDCval < 96) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 8;
}
}
}
else if (ssDCval >= 96) {
if (Rs < 1959) {
return 2;
}
else if (Rs >= 1959) {
if (Rs < 2324) {
return 4;
}
else if (Rs >= 2324) {
return 8;
}
}
}
}
else if (RsDiff >= 7) {
if (tcor < 88) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
return 2;
}
}
else if (tcor >= 88) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 4;
}
}
}
}
}
}
else if (CsDiff >= 5) {
if (tcor < 63) {
if (Rs < 1474) {
return 1;
}
else if (Rs >= 1474) {
return 4;
}
}
else if (tcor >= 63) {
if (negBalance < 35) {
if (SC < 2838) {
if (SC < 2477) {
return 1;
}
else if (SC >= 2477) {
return 2;
}
}
else if (SC >= 2838) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Cs < 1280) {
return 1;
}
else if (Cs >= 1280) {
return 8;
}
}
}
}
else if (negBalance >= 35) {
if (Cs < 1104) {
if (tcor < 90) {
if (ssDCval < 82) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 82) {
if (Rs < 1723) {
return 1;
}
else if (Rs >= 1723) {
return 4;
}
}
}
else if (tcor >= 90) {
if (diffAFD < -1) {
if (Rs < 2443) {
return 1;
}
else if (Rs >= 2443) {
return 2;
}
}
else if (diffAFD >= -1) {
if (Rs < 1427) {
return 8;
}
else if (Rs >= 1427) {
return 4;
}
}
}
}
else if (Cs >= 1104) {
if (posBalance < 52) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (posBalance >= 52) {
if (Rs < 1646) {
return 2;
}
else if (Rs >= 1646) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 108) {
if (mu_mv_mag_sq < 7) {
if (SC < 2574) {
return 8;
}
else if (SC >= 2574) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
if (ssDCval < 160) {
if (MVDiff < 4) {
if (Rs < 2110) {
return 2;
}
else if (Rs >= 2110) {
return 8;
}
}
else if (MVDiff >= 4) {
if (mu_mv_mag_sq < 70) {
return 8;
}
else if (mu_mv_mag_sq >= 70) {
if (Rs < 2087) {
return 8;
}
else if (Rs >= 2087) {
return 4;
}
}
}
}
else if (ssDCval >= 160) {
return 2;
}
}
}
}
else if (posBalance >= 53) {
if (negBalance < 44) {
if (ssDCval < 102) {
if (Rs < 2266) {
if (ssDCval < 91) {
if (diffRsCsdiff < 1) {
if (mcTcor < 93) {
if (Rs < 1409) {
return 4;
}
else if (Rs >= 1409) {
return 8;
}
}
else if (mcTcor >= 93) {
if (RsCsDiff < 57) {
return 8;
}
else if (RsCsDiff >= 57) {
return 4;
}
}
}
else if (diffRsCsdiff >= 1) {
if (mu_mv_mag_sq < 41) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (mu_mv_mag_sq >= 41) {
if (SC < 2552) {
if (Rs < 1418) {
return 8;
}
else if (Rs >= 1418) {
return 2;
}
}
else if (SC >= 2552) {
return 8;
}
}
}
}
else if (ssDCval >= 91) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
if (Cs < 1033) {
return 4;
}
else if (Cs >= 1033) {
return 2;
}
}
}
}
}
else if (Rs >= 2266) {
if (Cs < 1222) {
return 2;
}
else if (Cs >= 1222) {
return 1;
}
}
}
else if (ssDCval >= 102) {
if (MVDiff < 6) {
if (Rs < 1722) {
return 8;
}
else if (Rs >= 1722) {
return 2;
}
}
else if (MVDiff >= 6) {
if (ssDCval < 159) {
return 8;
}
else if (ssDCval >= 159) {
return 2;
}
}
}
}
else if (negBalance >= 44) {
if (AFD < 20) {
if (tcor < 85) {
if (mcTcor < 90) {
if (mcTcor < 86) {
if (Rs < 1374) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (TSC < 11) {
if (Cs < 975) {
return 8;
}
else if (Cs >= 975) {
return 4;
}
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (Rs >= 1374) {
return 8;
}
}
else if (mcTcor >= 86) {
if (refDCval < 92) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
if (SC < 2429) {
return 2;
}
else if (SC >= 2429) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 60) {
if (diffAFD < 1) {
if (diffMVdiffVal < -6) {
return 2;
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
else if (diffAFD >= 1) {
return 2;
}
}
else if (posBalance >= 60) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
else if (refDCval >= 92) {
if (ssDCval < 96) {
return 4;
}
else if (ssDCval >= 96) {
return 8;
}
}
}
}
else if (mcTcor >= 90) {
if (MVDiff < 22) {
if (tcor < 84) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Cs < 924) {
return 4;
}
else if (Cs >= 924) {
return 8;
}
}
}
else if (tcor >= 84) {
if (SC < 2927) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 4;
}
}
else if (SC >= 2927) {
return 8;
}
}
}
else if (MVDiff >= 22) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
else if (tcor >= 85) {
if (TSC < 11) {
if (mu_mv_mag_sq < 93) {
if (mcTcor < 98) {
if (Rs < 1656) {
if (Cs < 1255) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 4;
}
}
else if (Cs >= 1255) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
if (Cs < 1300) {
return 8;
}
else if (Cs >= 1300) {
return 2;
}
}
}
}
else if (Rs >= 1656) {
if (diffAFD < 2) {
if (AFD < 11) {
if (Rs < 1668) {
return 2;
}
else if (Rs >= 1668) {
return 8;
}
}
else if (AFD >= 11) {
if (mu_mv_mag_sq < 56) {
if (Cs < 1139) {
if (TSC < 9) {
if (Cs < 1120) {
return 8;
}
else if (Cs >= 1120) {
return 1;
}
}
else if (TSC >= 9) {
if (Rs < 1906) {
return 1;
}
else if (Rs >= 1906) {
return 2;
}
}
}
else if (Cs >= 1139) {
return 8;
}
}
else if (mu_mv_mag_sq >= 56) {
return 1;
}
}
}
else if (diffAFD >= 2) {
if (ssDCval < 80) {
if (Cs < 1130) {
return 2;
}
else if (Cs >= 1130) {
return 1;
}
}
else if (ssDCval >= 80) {
return 8;
}
}
}
}
else if (mcTcor >= 98) {
if (Rs < 1508) {
return 8;
}
else if (Rs >= 1508) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (Cs < 1211) {
if (Rs < 1850) {
return 2;
}
else if (Rs >= 1850) {
return 1;
}
}
else if (Cs >= 1211) {
return 8;
}
}
}
else if (TSC >= 11) {
return 1;
}
}
}
else if (AFD >= 20) {
if (Rs < 1800) {
if (Rs < 1464) {
return 4;
}
else if (Rs >= 1464) {
if (Rs < 1585) {
return 8;
}
else if (Rs >= 1585) {
return 1;
}
}
}
else if (Rs >= 1800) {
return 2;
}
}
}
}
}
}
else if (RsCsDiff >= 87) {
if (refDCval < 114) {
if (Rs < 2106) {
if (AFD < 23) {
if (negBalance < 44) {
if (TSC < 7) {
if (ssDCval < 67) {
if (diffRsCsdiff < 27) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
if (ssDCval < 66) {
return 1;
}
else if (ssDCval >= 66) {
return 8;
}
}
}
else if (diffRsCsdiff >= 27) {
if (posBalance < 67) {
return 8;
}
else if (posBalance >= 67) {
if (Rs < 932) {
return 4;
}
else if (Rs >= 932) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
if (Cs < 1178) {
return 1;
}
else if (Cs >= 1178) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (ssDCval < 102) {
if (diffTSC < 2) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
if (Rs < 2041) {
if (tcor < 57) {
return 8;
}
else if (tcor >= 57) {
if (Rs < 1516) {
if (Rs < 1377) {
return 2;
}
else if (Rs >= 1377) {
return 8;
}
}
else if (Rs >= 1516) {
return 2;
}
}
}
else if (Rs >= 2041) {
return 8;
}
}
}
}
else if (diffTSC >= 2) {
if (Rs < 1606) {
if (Rs < 1515) {
return 1;
}
else if (Rs >= 1515) {
return 2;
}
}
else if (Rs >= 1606) {
return 8;
}
}
}
else if (ssDCval >= 102) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 8;
}
}
}
}
else if (negBalance >= 44) {
if (Cs < 1013) {
if (ssDCval < 98) {
if (MVDiff < 44) {
if (Rs < 1687) {
if (refDCval < 84) {
if (Cs < 937) {
if (MVDiff < 27) {
if (Cs < 884) {
return 1;
}
else if (Cs >= 884) {
return 2;
}
}
else if (MVDiff >= 27) {
return 4;
}
}
else if (Cs >= 937) {
if (mcTcor < 93) {
if (RsDiff < 8) {
if (Rs < 967) {
return 4;
}
else if (Rs >= 967) {
return 8;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (mcTcor >= 93) {
return 4;
}
}
}
else if (refDCval >= 84) {
return 8;
}
}
else if (Rs >= 1687) {
if (tcor < 69) {
return 8;
}
else if (tcor >= 69) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
else if (MVDiff >= 44) {
if (mu_mv_mag_sq < 5) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Rs < 817) {
return 2;
}
else if (Rs >= 817) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (CsDiff < 10) {
if (Rs < 1607) {
return 1;
}
else if (Rs >= 1607) {
return 2;
}
}
else if (CsDiff >= 10) {
if (Rs < 813) {
return 4;
}
else if (Rs >= 813) {
return 8;
}
}
}
}
}
else if (ssDCval >= 98) {
if (mcTcor < 83) {
return 4;
}
else if (mcTcor >= 83) {
if (mcTcor < 97) {
return 1;
}
else if (mcTcor >= 97) {
return 2;
}
}
}
}
else if (Cs >= 1013) {
if (tcor < 62) {
if (AFD < 19) {
if (MVDiff < 16) {
if (TSC < 9) {
if (Cs < 1205) {
return 8;
}
else if (Cs >= 1205) {
return 1;
}
}
else if (TSC >= 9) {
return 4;
}
}
else if (MVDiff >= 16) {
if (RsDiff < 13) {
return 4;
}
else if (RsDiff >= 13) {
return 2;
}
}
}
else if (AFD >= 19) {
if (SC < 3154) {
if (Rs < 1337) {
if (Rs < 1040) {
return 8;
}
else if (Rs >= 1040) {
return 2;
}
}
else if (Rs >= 1337) {
return 8;
}
}
else if (SC >= 3154) {
return 4;
}
}
}
else if (tcor >= 62) {
if (ssDCval < 84) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1524) {
if (posBalance < 59) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 4;
}
}
else if (posBalance >= 59) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 1;
}
}
}
else if (Rs >= 1524) {
if (ssDCval < 57) {
if (Rs < 1626) {
return 8;
}
else if (Rs >= 1626) {
return 4;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (RsCsDiff < 115) {
if (MVDiff < 70) {
return 8;
}
else if (MVDiff >= 70) {
return 4;
}
}
else if (RsCsDiff >= 115) {
if (MVDiff < 15) {
return 8;
}
else if (MVDiff >= 15) {
if (TSC < 8) {
if (diffMVdiffVal < 10) {
if (Rs < 1677) {
return 1;
}
else if (Rs >= 1677) {
return 2;
}
}
else if (diffMVdiffVal >= 10) {
return 4;
}
}
else if (TSC >= 8) {
if (negBalance < 56) {
if (SC < 3078) {
if (Cs < 1101) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 2;
}
}
else if (Cs >= 1101) {
if (Rs < 1180) {
if (Rs < 944) {
return 2;
}
else if (Rs >= 944) {
return 1;
}
}
else if (Rs >= 1180) {
return 4;
}
}
}
else if (SC >= 3078) {
if (RsDiff < 9) {
return 2;
}
else if (RsDiff >= 9) {
if (tcor < 64) {
return 1;
}
else if (tcor >= 64) {
return 8;
}
}
}
}
else if (negBalance >= 56) {
if (Rs < 1257) {
return 1;
}
else if (Rs >= 1257) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (TSC < 9) {
if (MVDiff < 8) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 1;
}
}
else if (MVDiff >= 8) {
if (posBalance < 62) {
if (refDCval < 105) {
if (CsDiff < 9) {
return 2;
}
else if (CsDiff >= 9) {
if (Rs < 1183) {
return 2;
}
else if (Rs >= 1183) {
return 8;
}
}
}
else if (refDCval >= 105) {
return 8;
}
}
else if (posBalance >= 62) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (Cs < 1086) {
return 4;
}
else if (Cs >= 1086) {
return 8;
}
}
}
}
}
else if (TSC >= 9) {
if (tcor < 76) {
if (Cs < 1270) {
if (CsDiff < 7) {
if (ssDCval < 95) {
if (Rs < 1688) {
return 1;
}
else if (Rs >= 1688) {
return 8;
}
}
else if (ssDCval >= 95) {
return 4;
}
}
else if (CsDiff >= 7) {
return 1;
}
}
else if (Cs >= 1270) {
if (Cs < 1277) {
return 2;
}
else if (Cs >= 1277) {
return 8;
}
}
}
else if (tcor >= 76) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 23) {
if (mcTcor < 87) {
if (MVDiff < 28) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 2;
}
}
else if (MVDiff >= 28) {
if (CsDiff < 8) {
if (SC < 2386) {
if (Rs < 1101) {
return 4;
}
else if (Rs >= 1101) {
return 2;
}
}
else if (SC >= 2386) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 8;
}
}
}
else if (CsDiff >= 8) {
if (Rs < 1320) {
if (diffTSC < 2) {
if (negBalance < 71) {
if (Cs < 1010) {
return 1;
}
else if (Cs >= 1010) {
return 8;
}
}
else if (negBalance >= 71) {
return 1;
}
}
else if (diffTSC >= 2) {
if (diffRsCsdiff < 115) {
if (MVDiff < 188) {
return 4;
}
else if (MVDiff >= 188) {
if (Rs < 718) {
return 2;
}
else if (Rs >= 718) {
return 1;
}
}
}
else if (diffRsCsdiff >= 115) {
if (Cs < 974) {
return 4;
}
else if (Cs >= 974) {
return 8;
}
}
}
}
else if (Rs >= 1320) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
if (AFD < 35) {
return 4;
}
else if (AFD >= 35) {
if (Rs < 1636) {
return 2;
}
else if (Rs >= 1636) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 87) {
if (RsCsDiff < 273) {
if (diffAFD < -3) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 2;
}
}
else if (diffAFD >= -3) {
if (diffAFD < 6) {
if (refDCval < 92) {
if (diffTSC < -2) {
if (RsCsDiff < 115) {
return 1;
}
else if (RsCsDiff >= 115) {
if (RsDiff < 10) {
return 2;
}
else if (RsDiff >= 10) {
return 4;
}
}
}
else if (diffTSC >= -2) {
if (mu_mv_mag_sq < 2) {
if (Rs < 1454) {
return 8;
}
else if (Rs >= 1454) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (SC < 2456) {
if (ssDCval < 88) {
if (diffMVdiffVal < 12) {
return 4;
}
else if (diffMVdiffVal >= 12) {
if (tcor < 64) {
if (Cs < 1080) {
return 2;
}
else if (Cs >= 1080) {
return 8;
}
}
else if (tcor >= 64) {
return 4;
}
}
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (SC >= 2456) {
return 4;
}
}
}
}
else if (refDCval >= 92) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (Cs < 1092) {
if (Rs < 1506) {
return 8;
}
else if (Rs >= 1506) {
return 1;
}
}
else if (Cs >= 1092) {
if (Rs < 1093) {
return 1;
}
else if (Rs >= 1093) {
return 4;
}
}
}
}
}
else if (diffAFD >= 6) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
if (TSC < 13) {
return 1;
}
else if (TSC >= 13) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 273) {
return 2;
}
}
}
}
else if (Rs >= 2106) {
if (MVDiff < 25) {
if (tcor < 86) {
if (mu_mv_mag_sq < 5) {
if (mcTcor < 90) {
if (TSC < 12) {
if (diffRsCsdiff < 4) {
if (Rs < 2241) {
return 8;
}
else if (Rs >= 2241) {
return 4;
}
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
else if (TSC >= 12) {
return 2;
}
}
else if (mcTcor >= 90) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (Cs < 1144) {
if (Rs < 2147) {
return 4;
}
else if (Rs >= 2147) {
return 2;
}
}
else if (Cs >= 1144) {
if (Cs < 1252) {
return 4;
}
else if (Cs >= 1252) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Rs < 2196) {
if (Cs < 1183) {
return 2;
}
else if (Cs >= 1183) {
return 4;
}
}
else if (Rs >= 2196) {
if (diffRsCsdiff < -2) {
if (Cs < 1267) {
return 2;
}
else if (Cs >= 1267) {
return 8;
}
}
else if (diffRsCsdiff >= -2) {
if (diffRsCsdiff < 93) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
if (gchDC < 4) {
return 4;
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (diffRsCsdiff >= 93) {
return 8;
}
}
}
}
}
}
else if (tcor >= 86) {
if (refDCval < 89) {
if (RsDiff < 9) {
if (AFD < 10) {
if (Cs < 1013) {
return 1;
}
else if (Cs >= 1013) {
return 2;
}
}
else if (AFD >= 10) {
return 1;
}
}
else if (RsDiff >= 9) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (refDCval >= 89) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 1;
}
}
}
}
else if (MVDiff >= 25) {
if (tcor < 53) {
if (ssDCval < 102) {
if (gchDC < 8) {
return 1;
}
else if (gchDC >= 8) {
return 8;
}
}
else if (ssDCval >= 102) {
return 4;
}
}
else if (tcor >= 53) {
if (ssDCval < 76) {
if (Cs < 1060) {
return 2;
}
else if (Cs >= 1060) {
if (negBalance < 39) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 2278) {
return 4;
}
else if (Rs >= 2278) {
return 8;
}
}
}
else if (negBalance >= 39) {
if (mcTcor < 88) {
return 1;
}
else if (mcTcor >= 88) {
if (Rs < 2530) {
return 4;
}
else if (Rs >= 2530) {
return 1;
}
}
}
}
}
else if (ssDCval >= 76) {
if (RsCsDiff < 131) {
if (Rs < 2115) {
return 4;
}
else if (Rs >= 2115) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
if (Rs < 2347) {
return 8;
}
else if (Rs >= 2347) {
return 2;
}
}
}
}
else if (RsCsDiff >= 131) {
if (TSC < 11) {
if (MVDiff < 41) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (MVDiff >= 41) {
return 8;
}
}
else if (TSC >= 11) {
if (mu_mv_mag_sq < 13) {
if (Rs < 2564) {
return 2;
}
else if (Rs >= 2564) {
return 1;
}
}
else if (mu_mv_mag_sq >= 13) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 114) {
if (mcTcor < 94) {
if (MVDiff < 22) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (posBalance < 66) {
if (negBalance < 38) {
if (Cs < 1032) {
return 8;
}
else if (Cs >= 1032) {
return 1;
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (posBalance >= 66) {
if (Rs < 2040) {
return 4;
}
else if (Rs >= 2040) {
return 1;
}
}
}
}
else if (MVDiff >= 22) {
if (ssDCval < 134) {
if (gchDC < 5) {
if (Rs < 2559) {
return 1;
}
else if (Rs >= 2559) {
return 8;
}
}
else if (gchDC >= 5) {
if (negBalance < 72) {
if (mcTcor < 80) {
return 1;
}
else if (mcTcor >= 80) {
return 2;
}
}
else if (negBalance >= 72) {
return 8;
}
}
}
else if (ssDCval >= 134) {
if (mcTcor < 87) {
if (mu_mv_mag_sq < 17) {
if (Rs < 1266) {
if (Rs < 1063) {
return 1;
}
else if (Rs >= 1063) {
return 8;
}
}
else if (Rs >= 1266) {
return 4;
}
}
else if (mu_mv_mag_sq >= 17) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
if (Rs < 1514) {
return 4;
}
else if (Rs >= 1514) {
return 8;
}
}
}
}
else if (mcTcor >= 87) {
if (AFD < 23) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 8;
}
}
else if (AFD >= 23) {
return 1;
}
}
}
}
}
else if (mcTcor >= 94) {
if (ssDCval < 129) {
if (mcTcor < 96) {
if (AFD < 18) {
if (Rs < 1721) {
return 8;
}
else if (Rs >= 1721) {
if (Rs < 2279) {
return 2;
}
else if (Rs >= 2279) {
return 1;
}
}
}
else if (AFD >= 18) {
if (TSC < 12) {
return 1;
}
else if (TSC >= 12) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (gchDC < 2) {
if (Cs < 1000) {
return 8;
}
else if (Cs >= 1000) {
if (Rs < 1483) {
return 2;
}
else if (Rs >= 1483) {
return 1;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
else if (ssDCval >= 129) {
if (posBalance < 49) {
if (Rs < 1005) {
return 8;
}
else if (Rs >= 1005) {
return 2;
}
}
else if (posBalance >= 49) {
if (mcTcor < 96) {
if (Rs < 2001) {
return 8;
}
else if (Rs >= 2001) {
if (Rs < 2250) {
return 2;
}
else if (Rs >= 2250) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (CsDiff < 7) {
if (Rs < 1292) {
return 2;
}
else if (Rs >= 1292) {
return 8;
}
}
else if (CsDiff >= 7) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 2587) {
if (MVDiff < 17) {
if (refDCval < 108) {
if (posBalance < 48) {
if (negBalance < 22) {
if (SC < 4838) {
if (RsCsDiff < 5) {
if (Cs < 758) {
if (Rs < 2626) {
return 4;
}
else if (Rs >= 2626) {
return 1;
}
}
else if (Cs >= 758) {
if (negBalance < 20) {
return 8;
}
else if (negBalance >= 20) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
return 2;
}
}
}
}
else if (RsCsDiff >= 5) {
if (SC < 4401) {
if (Cs < 1009) {
if (Rs < 2801) {
return 8;
}
else if (Rs >= 2801) {
return 4;
}
}
else if (Cs >= 1009) {
return 8;
}
}
else if (SC >= 4401) {
if (Rs < 3457) {
if (Rs < 3347) {
return 2;
}
else if (Rs >= 3347) {
return 1;
}
}
else if (Rs >= 3457) {
return 4;
}
}
}
}
else if (SC >= 4838) {
if (SC < 5393) {
if (Cs < 1269) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
else if (Cs >= 1269) {
if (Rs < 3900) {
return 4;
}
else if (Rs >= 3900) {
return 8;
}
}
}
else if (SC >= 5393) {
if (mu_mv_mag_sq < 119) {
if (Rs < 5085) {
return 1;
}
else if (Rs >= 5085) {
return 2;
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
}
}
else if (negBalance >= 22) {
if (Rs < 3876) {
if (negBalance < 51) {
if (RsCsDiff < 28) {
if (Cs < 1125) {
if (SC < 3882) {
if (ssDCval < 74) {
if (Cs < 840) {
if (Cs < 602) {
return 8;
}
else if (Cs >= 602) {
return 1;
}
}
else if (Cs >= 840) {
return 2;
}
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (SC >= 3882) {
if (tcor < 97) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (Rs < 3476) {
return 1;
}
else if (Rs >= 3476) {
return 2;
}
}
}
else if (tcor >= 97) {
if (diffRsCsdiff < -38) {
if (Cs < 1059) {
if (Rs < 3342) {
return 1;
}
else if (Rs >= 3342) {
return 4;
}
}
else if (Cs >= 1059) {
return 2;
}
}
else if (diffRsCsdiff >= -38) {
return 2;
}
}
}
}
else if (Cs >= 1125) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
if (negBalance < 28) {
if (Rs < 3659) {
return 4;
}
else if (Rs >= 3659) {
return 1;
}
}
else if (negBalance >= 28) {
if (Rs < 3752) {
return 8;
}
else if (Rs >= 3752) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 28) {
if (refDCval < 89) {
if (ssDCval < 77) {
if (posBalance < 39) {
if (ssDCval < 59) {
if (Cs < 1276) {
return 4;
}
else if (Cs >= 1276) {
return 8;
}
}
else if (ssDCval >= 59) {
if (Rs < 3717) {
if (Cs < 1245) {
return 1;
}
else if (Cs >= 1245) {
return 4;
}
}
else if (Rs >= 3717) {
if (Cs < 1066) {
return 4;
}
else if (Cs >= 1066) {
return 1;
}
}
}
}
else if (posBalance >= 39) {
if (ssDCval < 68) {
if (CsDiff < 6) {
if (MVDiff < 4) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
return 2;
}
}
}
else if (MVDiff >= 4) {
if (Cs < 845) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
else if (Cs >= 845) {
if (negBalance < 42) {
if (SC < 4946) {
if (Cs < 851) {
return 2;
}
else if (Cs >= 851) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
if (Cs < 1110) {
return 4;
}
else if (Cs >= 1110) {
return 2;
}
}
}
}
else if (SC >= 4946) {
if (SC < 5008) {
return 2;
}
else if (SC >= 5008) {
return 4;
}
}
}
else if (negBalance >= 42) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 2;
}
}
}
}
}
else if (CsDiff >= 6) {
if (Cs < 893) {
return 4;
}
else if (Cs >= 893) {
return 2;
}
}
}
else if (ssDCval >= 68) {
if (Rs < 3126) {
if (Cs < 1063) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (CsDiff < 5) {
if (SC < 4073) {
return 1;
}
else if (SC >= 4073) {
return 4;
}
}
else if (CsDiff >= 5) {
return 4;
}
}
}
else if (Cs >= 1063) {
if (MVDiff < 8) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 4;
}
}
else if (MVDiff >= 8) {
return 4;
}
}
}
else if (Rs >= 3126) {
if (MVDiff < 6) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (Rs < 3532) {
if (Rs < 3190) {
return 1;
}
else if (Rs >= 3190) {
return 2;
}
}
else if (Rs >= 3532) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (Rs < 3474) {
return 4;
}
else if (Rs >= 3474) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
if (SC < 4529) {
return 1;
}
else if (SC >= 4529) {
if (Rs < 3590) {
return 4;
}
else if (Rs >= 3590) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 77) {
if (Rs < 2967) {
if (RsDiff < 10) {
if (refDCval < 86) {
return 1;
}
else if (refDCval >= 86) {
return 4;
}
}
else if (RsDiff >= 10) {
if (Rs < 2767) {
return 4;
}
else if (Rs >= 2767) {
return 2;
}
}
}
else if (Rs >= 2967) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
}
}
}
else if (refDCval >= 89) {
if (Rs < 3847) {
if (mcTcor < 97) {
if (ssDCval < 103) {
if (diffTSC < 1) {
if (Cs < 835) {
return 8;
}
else if (Cs >= 835) {
if (diffRsCsdiff < 4) {
if (diffMVdiffVal < 1) {
if (Cs < 881) {
return 4;
}
else if (Cs >= 881) {
if (tcor < 94) {
if (Rs < 3462) {
return 4;
}
else if (Rs >= 3462) {
return 4;
}
}
else if (tcor >= 94) {
if (Rs < 3404) {
return 8;
}
else if (Rs >= 3404) {
if (diffMVdiffVal < -1) {
if (Rs < 3569) {
return 8;
}
else if (Rs >= 3569) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 3472) {
return 8;
}
else if (Rs >= 3472) {
return 8;
}
}
}
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (ssDCval >= 103) {
return 8;
}
}
else if (mcTcor >= 97) {
if (Cs < 948) {
if (Cs < 781) {
if (Rs < 2886) {
return 8;
}
else if (Rs >= 2886) {
return 2;
}
}
else if (Cs >= 781) {
return 1;
}
}
else if (Cs >= 948) {
return 8;
}
}
}
else if (Rs >= 3847) {
return 2;
}
}
}
}
else if (negBalance >= 51) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Rs < 2820) {
return 1;
}
else if (Rs >= 2820) {
return 2;
}
}
}
}
else if (Rs >= 3876) {
if (Cs < 1036) {
if (Cs < 406) {
return 2;
}
else if (Cs >= 406) {
return 4;
}
}
else if (Cs >= 1036) {
if (refDCval < 76) {
if (diffRsCsdiff < 21) {
return 4;
}
else if (diffRsCsdiff >= 21) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 4;
}
}
}
else if (refDCval >= 76) {
if (CsDiff < 4) {
if (Cs < 1263) {
return 4;
}
else if (Cs >= 1263) {
if (Cs < 1282) {
if (Rs < 4412) {
return 4;
}
else if (Rs >= 4412) {
return 2;
}
}
else if (Cs >= 1282) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (AFD < 12) {
if (mu_mv_mag_sq < 59) {
if (Cs < 1182) {
if (Rs < 4119) {
if (Rs < 4062) {
if (Rs < 4032) {
return 4;
}
else if (Rs >= 4032) {
return 8;
}
}
else if (Rs >= 4062) {
return 4;
}
}
else if (Rs >= 4119) {
return 8;
}
}
else if (Cs >= 1182) {
if (Rs < 4287) {
return 4;
}
else if (Rs >= 4287) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (Cs < 1222) {
return 8;
}
else if (Cs >= 1222) {
return 8;
}
}
}
else if (AFD >= 12) {
return 2;
}
}
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Rs < 3904) {
if (Cs < 215) {
return 1;
}
else if (Cs >= 215) {
if (negBalance < 22) {
if (AFD < 3) {
if (Rs < 2596) {
return 1;
}
else if (Rs >= 2596) {
return 8;
}
}
else if (AFD >= 3) {
if (negBalance < 15) {
return 4;
}
else if (negBalance >= 15) {
if (Rs < 2617) {
return 8;
}
else if (Rs >= 2617) {
if (Rs < 2741) {
return 4;
}
else if (Rs >= 2741) {
return 4;
}
}
}
}
}
else if (negBalance >= 22) {
if (gchDC < 2) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (Cs < 651) {
if (tcor < 84) {
return 2;
}
else if (tcor >= 84) {
if (RsDiff < 12) {
return 4;
}
else if (RsDiff >= 12) {
return 2;
}
}
}
else if (Cs >= 651) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (Rs < 2780) {
return 1;
}
else if (Rs >= 2780) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3904) {
return 4;
}
}
else if (CsDiff >= 3) {
if (refDCval < 64) {
if (Cs < 993) {
if (tcor < 84) {
if (Rs < 2666) {
return 2;
}
else if (Rs >= 2666) {
return 4;
}
}
else if (tcor >= 84) {
return 2;
}
}
else if (Cs >= 993) {
if (tcor < 94) {
if (SC < 3671) {
return 2;
}
else if (SC >= 3671) {
return 4;
}
}
else if (tcor >= 94) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
else if (refDCval >= 64) {
if (negBalance < 38) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 79) {
if (Rs < 3007) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (Cs < 1084) {
return 4;
}
else if (Cs >= 1084) {
return 2;
}
}
}
else if (Rs >= 3007) {
if (Rs < 3955) {
return 4;
}
else if (Rs >= 3955) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 79) {
return 2;
}
}
else if (posBalance >= 53) {
if (AFD < 18) {
if (RsDiff < 5) {
if (Rs < 3949) {
return 4;
}
else if (Rs >= 3949) {
return 2;
}
}
else if (RsDiff >= 5) {
if (gchDC < 4) {
if (mcTcor < 99) {
return 2;
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (gchDC >= 4) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 2;
}
}
}
}
else if (AFD >= 18) {
return 4;
}
}
}
else if (negBalance >= 38) {
if (tcor < 79) {
if (posBalance < 55) {
if (refDCval < 74) {
return 4;
}
else if (refDCval >= 74) {
if (Cs < 1153) {
return 2;
}
else if (Cs >= 1153) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (SC < 4255) {
return 2;
}
else if (SC >= 4255) {
if (Cs < 1147) {
return 1;
}
else if (Cs >= 1147) {
return 2;
}
}
}
}
else if (tcor >= 79) {
if (Rs < 4726) {
if (RsDiff < 6) {
if (Rs < 2711) {
return 4;
}
else if (Rs >= 2711) {
return 2;
}
}
else if (RsDiff >= 6) {
if (mcTcor < 99) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (diffRsCsdiff < 202) {
return 2;
}
else if (diffRsCsdiff >= 202) {
if (Rs < 2962) {
return 2;
}
else if (Rs >= 2962) {
return 4;
}
}
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
else if (Rs >= 4726) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsCsDiff < 121) {
if (Cs < 972) {
if (tcor < 95) {
if (RsCsDiff < 63) {
if (mu_mv_mag_sq < 49) {
return 8;
}
else if (mu_mv_mag_sq >= 49) {
if (Cs < 484) {
return 8;
}
else if (Cs >= 484) {
if (Cs < 496) {
return 4;
}
else if (Cs >= 496) {
if (Rs < 3187) {
return 4;
}
else if (Rs >= 3187) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 63) {
if (AFD < 11) {
if (refDCval < 96) {
return 8;
}
else if (refDCval >= 96) {
return 2;
}
}
else if (AFD >= 11) {
return 8;
}
}
}
else if (tcor >= 95) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 2;
}
}
}
else if (Cs >= 972) {
if (diffRsCsdiff < -45) {
if (Rs < 3198) {
return 4;
}
else if (Rs >= 3198) {
return 2;
}
}
else if (diffRsCsdiff >= -45) {
if (Rs < 3998) {
return 8;
}
else if (Rs >= 3998) {
if (Rs < 4362) {
return 4;
}
else if (Rs >= 4362) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 121) {
if (negBalance < 49) {
if (ssDCval < 107) {
return 2;
}
else if (ssDCval >= 107) {
if (RsDiff < 13) {
if (Cs < 1146) {
return 4;
}
else if (Cs >= 1146) {
return 8;
}
}
else if (RsDiff >= 13) {
return 2;
}
}
}
else if (negBalance >= 49) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (Cs < 806) {
return 2;
}
else if (Cs >= 806) {
if (CsDiff < 6) {
if (mcTcor < 86) {
if (Rs < 2866) {
return 8;
}
else if (Rs >= 2866) {
return 1;
}
}
else if (mcTcor >= 86) {
if (MVDiff < 15) {
return 4;
}
else if (MVDiff >= 15) {
if (Rs < 3365) {
return 2;
}
else if (Rs >= 3365) {
return 4;
}
}
}
}
else if (CsDiff >= 6) {
if (Rs < 3263) {
return 4;
}
else if (Rs >= 3263) {
if (ssDCval < 100) {
return 2;
}
else if (ssDCval >= 100) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 108) {
if (MVDiff < 7) {
if (mcTcor < 92) {
if (Rs < 3074) {
return 8;
}
else if (Rs >= 3074) {
if (Rs < 3575) {
if (Rs < 3219) {
return 1;
}
else if (Rs >= 3219) {
return 4;
}
}
else if (Rs >= 3575) {
return 2;
}
}
}
else if (mcTcor >= 92) {
if (Cs < 619) {
if (RsDiff < 7) {
if (mu_mv_mag_sq < 52) {
return 8;
}
else if (mu_mv_mag_sq >= 52) {
if (Cs < 468) {
return 1;
}
else if (Cs >= 468) {
return 2;
}
}
}
else if (RsDiff >= 7) {
if (diffMVdiffVal < -59) {
return 8;
}
else if (diffMVdiffVal >= -59) {
return 2;
}
}
}
else if (Cs >= 619) {
if (refDCval < 136) {
if (MVDiff < 6) {
if (mcTcor < 96) {
if (Rs < 4008) {
return 8;
}
else if (Rs >= 4008) {
return 2;
}
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (MVDiff >= 6) {
if (AFD < 9) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (AFD >= 9) {
return 2;
}
}
}
else if (refDCval >= 136) {
if (mcTcor < 98) {
if (Cs < 978) {
if (Rs < 6340) {
return 2;
}
else if (Rs >= 6340) {
if (Rs < 6984) {
return 1;
}
else if (Rs >= 6984) {
return 2;
}
}
}
else if (Cs >= 978) {
if (Rs < 4420) {
if (RsCsDiff < 71) {
if (Rs < 3210) {
return 8;
}
else if (Rs >= 3210) {
return 2;
}
}
else if (RsCsDiff >= 71) {
return 1;
}
}
else if (Rs >= 4420) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 3963) {
if (Cs < 1139) {
return 2;
}
else if (Cs >= 1139) {
if (Rs < 3668) {
return 2;
}
else if (Rs >= 3668) {
if (ssDCval < 140) {
return 2;
}
else if (ssDCval >= 140) {
if (negBalance < 47) {
if (Cs < 1251) {
return 1;
}
else if (Cs >= 1251) {
if (Rs < 3725) {
return 2;
}
else if (Rs >= 3725) {
return 1;
}
}
}
else if (negBalance >= 47) {
return 2;
}
}
}
}
}
else if (Rs >= 3963) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (RsDiff < 9) {
if (mu_mv_mag_sq < 84) {
if (TSC < 5) {
if (posBalance < 36) {
if (ssDCval < 117) {
return 1;
}
else if (ssDCval >= 117) {
return 8;
}
}
else if (posBalance >= 36) {
if (Rs < 4212) {
return 8;
}
else if (Rs >= 4212) {
return 2;
}
}
}
else if (TSC >= 5) {
if (Cs < 1113) {
if (AFD < 8) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (Cs < 1050) {
return 2;
}
else if (Cs >= 1050) {
return 8;
}
}
}
else if (AFD >= 8) {
return 2;
}
}
else if (Cs >= 1113) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (Cs < 1161) {
return 8;
}
else if (Cs >= 1161) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (diffAFD < -6) {
if (Cs < 1282) {
return 1;
}
else if (Cs >= 1282) {
return 8;
}
}
else if (diffAFD >= -6) {
return 2;
}
}
}
else if (RsDiff >= 9) {
if (ssDCval < 134) {
if (Rs < 3953) {
if (Cs < 997) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 113) {
return 2;
}
else if (ssDCval >= 113) {
if (Rs < 3526) {
return 1;
}
else if (Rs >= 3526) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (ssDCval < 131) {
if (Cs < 876) {
return 2;
}
else if (Cs >= 876) {
if (Rs < 3159) {
return 4;
}
else if (Rs >= 3159) {
return 2;
}
}
}
else if (ssDCval >= 131) {
if (Cs < 557) {
return 1;
}
else if (Cs >= 557) {
return 2;
}
}
}
}
}
else if (Cs >= 997) {
if (mu_mv_mag_sq < 27) {
if (ssDCval < 121) {
return 4;
}
else if (ssDCval >= 121) {
return 1;
}
}
else if (mu_mv_mag_sq >= 27) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 2;
}
}
}
}
else if (Rs >= 3953) {
if (diffMVdiffVal < -18) {
if (posBalance < 65) {
return 2;
}
else if (posBalance >= 65) {
return 1;
}
}
else if (diffMVdiffVal >= -18) {
return 2;
}
}
}
else if (ssDCval >= 134) {
if (mcTcor < 95) {
if (negBalance < 53) {
if (RsDiff < 21) {
return 1;
}
else if (RsDiff >= 21) {
return 2;
}
}
else if (negBalance >= 53) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
return 2;
}
}
}
else if (mcTcor >= 95) {
if (mcTcor < 98) {
if (RsDiff < 17) {
if (mu_mv_mag_sq < 38) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
return 2;
}
}
else if (mu_mv_mag_sq >= 38) {
if (Cs < 1036) {
if (Rs < 2900) {
return 1;
}
else if (Rs >= 2900) {
return 2;
}
}
else if (Cs >= 1036) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 2;
}
}
}
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (mcTcor >= 98) {
if (Cs < 1101) {
return 2;
}
else if (Cs >= 1101) {
if (Rs < 4220) {
return 1;
}
else if (Rs >= 4220) {
return 2;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 17) {
if (mcTcor < 68) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
if (AFD < 33) {
return 4;
}
else if (AFD >= 33) {
return 8;
}
}
}
else if (mcTcor >= 68) {
if (RsCsDiff < 345) {
if (refDCval < 133) {
if (ssDCval < 92) {
if (CsDiff < 3) {
if (refDCval < 86) {
if (AFD < 6) {
if (Cs < 181) {
return 1;
}
else if (Cs >= 181) {
return 8;
}
}
else if (AFD >= 6) {
if (RsDiff < 18) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
if (Rs < 4021) {
return 8;
}
else if (Rs >= 4021) {
return 4;
}
}
}
else if (RsDiff >= 18) {
if (Rs < 3675) {
return 2;
}
else if (Rs >= 3675) {
return 4;
}
}
}
}
else if (refDCval >= 86) {
if (mu_mv_mag_sq < 5) {
return 2;
}
else if (mu_mv_mag_sq >= 5) {
if (tcor < 90) {
if (RsDiff < 14) {
return 4;
}
else if (RsDiff >= 14) {
return 2;
}
}
else if (tcor >= 90) {
return 8;
}
}
}
}
else if (CsDiff >= 3) {
if (posBalance < 48) {
if (negBalance < 55) {
if (SC < 3822) {
if (diffAFD < 4) {
if (RsCsDiff < 138) {
if (SC < 3522) {
return 4;
}
else if (SC >= 3522) {
if (Rs < 2892) {
return 2;
}
else if (Rs >= 2892) {
return 8;
}
}
}
else if (RsCsDiff >= 138) {
return 2;
}
}
else if (diffAFD >= 4) {
return 1;
}
}
else if (SC >= 3822) {
if (diffMVdiffVal < -4) {
if (Rs < 3083) {
return 2;
}
else if (Rs >= 3083) {
if (Rs < 3212) {
return 4;
}
else if (Rs >= 3212) {
return 8;
}
}
}
else if (diffMVdiffVal >= -4) {
if (ssDCval < 86) {
return 4;
}
else if (ssDCval >= 86) {
return 2;
}
}
}
}
else if (negBalance >= 55) {
if (Rs < 2623) {
return 8;
}
else if (Rs >= 2623) {
return 2;
}
}
}
else if (posBalance >= 48) {
if (Rs < 3606) {
if (CsDiff < 8) {
if (AFD < 23) {
if (MVDiff < 36) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 2;
}
}
else if (MVDiff >= 36) {
if (Rs < 2600) {
return 1;
}
else if (Rs >= 2600) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 1;
}
}
}
}
else if (AFD >= 23) {
return 1;
}
}
else if (CsDiff >= 8) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
return 2;
}
}
}
else if (Rs >= 3606) {
if (Cs < 823) {
return 4;
}
else if (Cs >= 823) {
if (Rs < 4210) {
return 2;
}
else if (Rs >= 4210) {
if (Cs < 1155) {
return 1;
}
else if (Cs >= 1155) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 92) {
if (Rs < 5022) {
if (SC < 3146) {
if (RsDiff < 16) {
if (gchDC < 1) {
if (Rs < 2723) {
return 2;
}
else if (Rs >= 2723) {
return 1;
}
}
else if (gchDC >= 1) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (diffRsCsdiff < 42) {
if (MVDiff < 33) {
if (Rs < 2648) {
return 8;
}
else if (Rs >= 2648) {
return 1;
}
}
else if (MVDiff >= 33) {
return 8;
}
}
else if (diffRsCsdiff >= 42) {
return 1;
}
}
}
}
else if (RsDiff >= 16) {
if (Cs < 293) {
return 2;
}
else if (Cs >= 293) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 1;
}
}
}
}
else if (SC >= 3146) {
if (mcTcor < 92) {
if (mcTcor < 82) {
if (Cs < 1025) {
if (refDCval < 102) {
return 1;
}
else if (refDCval >= 102) {
if (SC < 3358) {
return 8;
}
else if (SC >= 3358) {
return 2;
}
}
}
else if (Cs >= 1025) {
if (tcor < 56) {
return 4;
}
else if (tcor >= 56) {
if (Rs < 2808) {
return 2;
}
else if (Rs >= 2808) {
if (Rs < 3032) {
return 4;
}
else if (Rs >= 3032) {
return 1;
}
}
}
}
}
else if (mcTcor >= 82) {
if (MVDiff < 31) {
if (Cs < 727) {
if (AFD < 14) {
if (Rs < 2910) {
return 1;
}
else if (Rs >= 2910) {
return 4;
}
}
else if (AFD >= 14) {
return 2;
}
}
else if (Cs >= 727) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
if (Rs < 2796) {
return 2;
}
else if (Rs >= 2796) {
return 8;
}
}
}
}
}
else if (MVDiff >= 31) {
if (AFD < 26) {
if (CsDiff < 3) {
if (diffRsCsdiff < 83) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (diffMVdiffVal < -20) {
return 4;
}
else if (diffMVdiffVal >= -20) {
return 8;
}
}
}
else if (diffRsCsdiff >= 83) {
return 4;
}
}
else if (CsDiff >= 3) {
if (diffAFD < -4) {
if (Cs < 847) {
return 1;
}
else if (Cs >= 847) {
return 8;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
}
else if (AFD >= 26) {
if (Cs < 1134) {
return 1;
}
else if (Cs >= 1134) {
return 4;
}
}
}
}
}
else if (mcTcor >= 92) {
if (CsDiff < 3) {
if (negBalance < 49) {
if (TSC < 4) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (TSC >= 4) {
if (refDCval < 112) {
if (diffTSC < -2) {
if (ssDCval < 106) {
return 2;
}
else if (ssDCval >= 106) {
return 8;
}
}
else if (diffTSC >= -2) {
return 4;
}
}
else if (refDCval >= 112) {
return 2;
}
}
}
else if (negBalance >= 49) {
if (MVDiff < 20) {
return 2;
}
else if (MVDiff >= 20) {
if (Cs < 377) {
return 4;
}
else if (Cs >= 377) {
if (diffRsCsdiff < -39) {
if (RsCsDiff < 84) {
return 8;
}
else if (RsCsDiff >= 84) {
if (Rs < 3222) {
return 1;
}
else if (Rs >= 3222) {
return 4;
}
}
}
else if (diffRsCsdiff >= -39) {
return 8;
}
}
}
}
}
else if (CsDiff >= 3) {
if (Cs < 1231) {
if (refDCval < 99) {
if (MVDiff < 36) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
return 2;
}
}
else if (MVDiff >= 36) {
if (Rs < 3219) {
return 2;
}
else if (Rs >= 3219) {
if (Cs < 1177) {
return 8;
}
else if (Cs >= 1177) {
return 2;
}
}
}
}
else if (refDCval >= 99) {
if (posBalance < 62) {
if (CsDiff < 7) {
if (diffRsCsdiff < -127) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
return 1;
}
}
}
else if (diffRsCsdiff >= -127) {
if (Rs < 3263) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (mcTcor < 95) {
if (mu_mv_mag_sq < 2) {
return 8;
}
else if (mu_mv_mag_sq >= 2) {
if (MVDiff < 29) {
if (Rs < 2861) {
return 2;
}
else if (Rs >= 2861) {
return 8;
}
}
else if (MVDiff >= 29) {
if (Cs < 524) {
return 8;
}
else if (Cs >= 524) {
return 1;
}
}
}
}
else if (mcTcor >= 95) {
if (Rs < 2737) {
return 8;
}
else if (Rs >= 2737) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
}
}
}
else if (Rs >= 3263) {
if (Rs < 3335) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
return 8;
}
}
else if (Rs >= 3335) {
if (MVDiff < 69) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
if (gchDC < 5) {
return 8;
}
else if (gchDC >= 5) {
return 2;
}
}
}
else if (MVDiff >= 69) {
return 2;
}
}
}
}
}
else if (CsDiff >= 7) {
if (refDCval < 105) {
return 2;
}
else if (refDCval >= 105) {
if (Cs < 752) {
return 2;
}
else if (Cs >= 752) {
return 1;
}
}
}
}
else if (posBalance >= 62) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (RsCsDiff < 332) {
if (MVDiff < 24) {
if (mu_mv_mag_sq < 10) {
if (ssDCval < 114) {
return 2;
}
else if (ssDCval >= 114) {
return 1;
}
}
else if (mu_mv_mag_sq >= 10) {
if (diffRsCsdiff < -1) {
if (Rs < 2868) {
return 8;
}
else if (Rs >= 2868) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
return 2;
}
}
}
else if (MVDiff >= 24) {
if (tcor < 86) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (Rs < 2780) {
return 1;
}
else if (Rs >= 2780) {
return 8;
}
}
}
}
else if (tcor >= 86) {
return 1;
}
}
}
else if (RsCsDiff >= 332) {
return 2;
}
}
}
}
}
else if (Cs >= 1231) {
if (Rs < 4634) {
return 8;
}
else if (Rs >= 4634) {
if (Rs < 4732) {
return 1;
}
else if (Rs >= 4732) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 5022) {
return 2;
}
}
}
else if (refDCval >= 133) {
if (mcTcor < 93) {
if (AFD < 19) {
if (ssDCval < 150) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 4;
}
}
else if (ssDCval >= 150) {
if (Cs < 998) {
return 1;
}
else if (Cs >= 998) {
return 8;
}
}
}
else if (AFD >= 19) {
if (MVDiff < 64) {
if (diffMVdiffVal < 10) {
return 1;
}
else if (diffMVdiffVal >= 10) {
if (Rs < 3236) {
return 8;
}
else if (Rs >= 3236) {
return 1;
}
}
}
else if (MVDiff >= 64) {
if (AFD < 31) {
return 8;
}
else if (AFD >= 31) {
return 2;
}
}
}
}
else if (mcTcor >= 93) {
if (Rs < 2978) {
if (Cs < 798) {
return 1;
}
else if (Cs >= 798) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (Rs >= 2978) {
if (diffAFD < 2) {
if (negBalance < 35) {
return 2;
}
else if (negBalance >= 35) {
if (gchDC < 4) {
if (diffRsCsdiff < -252) {
if (diffTSC < -1) {
if (Cs < 620) {
return 8;
}
else if (Cs >= 620) {
return 2;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (diffRsCsdiff >= -252) {
if (mcTcor < 94) {
if (Rs < 4230) {
return 4;
}
else if (Rs >= 4230) {
return 1;
}
}
else if (mcTcor >= 94) {
if (diffRsCsdiff < 65) {
return 1;
}
else if (diffRsCsdiff >= 65) {
if (Rs < 3334) {
return 1;
}
else if (Rs >= 3334) {
return 2;
}
}
}
}
}
else if (gchDC >= 4) {
if (ssDCval < 144) {
if (RsDiff < 16) {
return 2;
}
else if (RsDiff >= 16) {
return 1;
}
}
else if (ssDCval >= 144) {
if (Rs < 4271) {
return 1;
}
else if (Rs >= 4271) {
return 8;
}
}
}
}
}
else if (diffAFD >= 2) {
if (RsCsDiff < 265) {
return 2;
}
else if (RsCsDiff >= 265) {
if (Rs < 4310) {
return 2;
}
else if (Rs >= 4310) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 345) {
if (CsDiff < 6) {
if (Rs < 5154) {
if (refDCval < 120) {
if (refDCval < 74) {
if (diffRsCsdiff < 182) {
if (AFD < 21) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 2;
}
}
else if (AFD >= 21) {
return 1;
}
}
else if (diffRsCsdiff >= 182) {
return 4;
}
}
else if (refDCval >= 74) {
if (Cs < 1102) {
if (RsCsDiff < 382) {
if (gchDC < 4) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (diffRsCsdiff < -44) {
return 2;
}
else if (diffRsCsdiff >= -44) {
return 8;
}
}
}
}
else if (gchDC >= 4) {
if (Rs < 2836) {
return 2;
}
else if (Rs >= 2836) {
return 1;
}
}
}
else if (RsCsDiff >= 382) {
if (ssDCval < 121) {
if (diffAFD < -9) {
return 1;
}
else if (diffAFD >= -9) {
return 2;
}
}
else if (ssDCval >= 121) {
return 1;
}
}
}
else if (Cs >= 1102) {
if (MVDiff < 62) {
if (Rs < 4373) {
return 2;
}
else if (Rs >= 4373) {
return 8;
}
}
else if (MVDiff >= 62) {
if (Cs < 1258) {
return 4;
}
else if (Cs >= 1258) {
return 2;
}
}
}
}
}
else if (refDCval >= 120) {
if (mcTcor < 90) {
if (MVDiff < 43) {
return 4;
}
else if (MVDiff >= 43) {
return 2;
}
}
else if (mcTcor >= 90) {
if (SC < 4314) {
if (Cs < 230) {
return 2;
}
else if (Cs >= 230) {
return 1;
}
}
else if (SC >= 4314) {
if (tcor < 76) {
if (Cs < 493) {
return 2;
}
else if (Cs >= 493) {
return 8;
}
}
else if (tcor >= 76) {
if (Rs < 4510) {
if (CsDiff < 4) {
if (Rs < 4353) {
return 2;
}
else if (Rs >= 4353) {
return 4;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (Rs >= 4510) {
return 2;
}
}
}
}
}
}
else if (Rs >= 5154) {
if (ssDCval < 55) {
return 8;
}
else if (ssDCval >= 55) {
if (Cs < 958) {
return 2;
}
else if (Cs >= 958) {
if (Cs < 1085) {
return 8;
}
else if (Cs >= 1085) {
return 2;
}
}
}
}
}
else if (CsDiff >= 6) {
if (gchDC < 5) {
if (ssDCval < 121) {
if (tcor < 63) {
if (AFD < 26) {
if (MVDiff < 21) {
return 2;
}
else if (MVDiff >= 21) {
return 8;
}
}
else if (AFD >= 26) {
return 2;
}
}
else if (tcor >= 63) {
if (diffRsCsdiff < 44) {
if (MVDiff < 35) {
if (SC < 4453) {
if (Rs < 3190) {
return 2;
}
else if (Rs >= 3190) {
return 1;
}
}
else if (SC >= 4453) {
if (diffRsCsdiff < -7) {
if (MVDiff < 31) {
return 2;
}
else if (MVDiff >= 31) {
if (Rs < 3755) {
return 2;
}
else if (Rs >= 3755) {
return 8;
}
}
}
else if (diffRsCsdiff >= -7) {
if (Rs < 4535) {
return 8;
}
else if (Rs >= 4535) {
return 2;
}
}
}
}
else if (MVDiff >= 35) {
if (diffAFD < -7) {
return 2;
}
else if (diffAFD >= -7) {
if (Rs < 5729) {
return 1;
}
else if (Rs >= 5729) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= 44) {
if (diffMVdiffVal < -31) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 1;
}
}
else if (diffMVdiffVal >= -31) {
if (RsDiff < 18) {
return 8;
}
else if (RsDiff >= 18) {
if (ssDCval < 120) {
return 2;
}
else if (ssDCval >= 120) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 121) {
if (Rs < 5668) {
if (diffAFD < 6) {
if (mu_mv_mag_sq < 2) {
if (RsDiff < 25) {
return 1;
}
else if (RsDiff >= 25) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
if (gchDC < 2) {
if (diffAFD < -1) {
if (Rs < 4507) {
return 8;
}
else if (Rs >= 4507) {
return 2;
}
}
else if (diffAFD >= -1) {
if (MVDiff < 66) {
return 8;
}
else if (MVDiff >= 66) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (AFD < 21) {
return 1;
}
else if (AFD >= 21) {
if (Rs < 3506) {
if (Cs < 825) {
return 2;
}
else if (Cs >= 825) {
return 1;
}
}
else if (Rs >= 3506) {
if (AFD < 34) {
return 2;
}
else if (AFD >= 34) {
if (Rs < 4080) {
return 8;
}
else if (Rs >= 4080) {
return 2;
}
}
}
}
}
}
}
else if (diffAFD >= 6) {
if (Rs < 5088) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
return 1;
}
}
else if (Rs >= 5088) {
return 2;
}
}
}
else if (Rs >= 5668) {
if (negBalance < 50) {
return 2;
}
else if (negBalance >= 50) {
if (diffMVdiffVal < 11) {
if (RsDiff < 27) {
return 8;
}
else if (RsDiff >= 27) {
return 2;
}
}
else if (diffMVdiffVal >= 11) {
return 2;
}
}
}
}
}
else if (gchDC >= 5) {
if (tcor < 75) {
if (MVDiff < 56) {
if (Rs < 4727) {
if (posBalance < 63) {
if (Cs < 1148) {
return 1;
}
else if (Cs >= 1148) {
return 2;
}
}
else if (posBalance >= 63) {
return 2;
}
}
else if (Rs >= 4727) {
if (AFD < 26) {
return 2;
}
else if (AFD >= 26) {
return 8;
}
}
}
else if (MVDiff >= 56) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
return 1;
}
}
}
else if (tcor >= 75) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Cs >= 1308) {
if (mu_mv_mag_sq < 111) {
if (RsDiff < 9) {
if (CsDiff < 7) {
if (ssDCval < 90) {
if (negBalance < 39) {
if (Rs < 2831) {
if (MVDiff < 5) {
if (posBalance < 48) {
if (refDCval < 79) {
if (Rs < 1168) {
if (MVDiff < 4) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (MVDiff >= 4) {
if (TSCindex < 2) {
if (Cs < 1776) {
return 4;
}
else if (Cs >= 1776) {
return 8;
}
}
else if (TSCindex >= 2) {
return 1;
}
}
}
else if (Rs >= 1168) {
if (mu_mv_mag_sq < 96) {
if (RsDiff < 4) {
if (tcor < 96) {
if (Rs < 1852) {
return 8;
}
else if (Rs >= 1852) {
return 2;
}
}
else if (tcor >= 96) {
return 8;
}
}
else if (RsDiff >= 4) {
if (Rs < 2609) {
return 8;
}
else if (Rs >= 2609) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 96) {
return 8;
}
}
}
else if (refDCval >= 79) {
if (Rs < 2105) {
return 8;
}
else if (Rs >= 2105) {
if (negBalance < 36) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 1;
}
}
else if (negBalance >= 36) {
if (Rs < 2785) {
return 8;
}
else if (Rs >= 2785) {
return 1;
}
}
}
}
}
else if (posBalance >= 48) {
if (SC < 4989) {
if (diffAFD < -2) {
if (Cs < 1493) {
return 2;
}
else if (Cs >= 1493) {
if (Rs < 1703) {
return 1;
}
else if (Rs >= 1703) {
return 8;
}
}
}
else if (diffAFD >= -2) {
if (ssDCval < 57) {
if (ssDCval < 55) {
if (Rs < 1124) {
if (diffRsCsdiff < -2) {
if (Cs < 1607) {
return 4;
}
else if (Cs >= 1607) {
return 8;
}
}
else if (diffRsCsdiff >= -2) {
if (Rs < 756) {
return 1;
}
else if (Rs >= 756) {
return 2;
}
}
}
else if (Rs >= 1124) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 4;
}
}
}
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (ssDCval >= 57) {
if (Rs < 2359) {
if (TSC < 8) {
if (diffRsCsdiff < 12) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
if (Rs < 1988) {
return 8;
}
else if (Rs >= 1988) {
return 2;
}
}
}
else if (diffRsCsdiff >= 12) {
if (Rs < 1519) {
return 1;
}
else if (Rs >= 1519) {
return 8;
}
}
}
else if (TSC >= 8) {
return 4;
}
}
else if (Rs >= 2359) {
if (tcor < 97) {
if (MVDiff < 3) {
if (Rs < 2382) {
return 4;
}
else if (Rs >= 2382) {
return 8;
}
}
else if (MVDiff >= 3) {
if (Rs < 2527) {
return 4;
}
else if (Rs >= 2527) {
return 2;
}
}
}
else if (tcor >= 97) {
if (Rs < 2623) {
if (Cs < 1594) {
return 4;
}
else if (Cs >= 1594) {
return 8;
}
}
else if (Rs >= 2623) {
return 4;
}
}
}
}
}
}
else if (SC >= 4989) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (Rs < 829) {
if (Rs < 300) {
return 4;
}
else if (Rs >= 300) {
if (MVDiff < 39) {
return 1;
}
else if (MVDiff >= 39) {
if (Rs < 532) {
return 8;
}
else if (Rs >= 532) {
return 2;
}
}
}
}
else if (Rs >= 829) {
if (ssDCval < 64) {
if (negBalance < 21) {
if (refDCval < 55) {
if (Cs < 1514) {
if (Rs < 1693) {
return 8;
}
else if (Rs >= 1693) {
return 1;
}
}
else if (Cs >= 1514) {
return 8;
}
}
else if (refDCval >= 55) {
if (Rs < 1045) {
return 8;
}
else if (Rs >= 1045) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 8;
}
}
}
}
else if (negBalance >= 21) {
if (posBalance < 43) {
if (mu_mv_mag_sq < 102) {
if (Rs < 2277) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
return 8;
}
}
else if (Rs >= 2277) {
if (Rs < 2641) {
return 8;
}
else if (Rs >= 2641) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 102) {
if (Rs < 1046) {
return 8;
}
else if (Rs >= 1046) {
if (Rs < 1695) {
return 4;
}
else if (Rs >= 1695) {
return 8;
}
}
}
}
else if (posBalance >= 43) {
if (SC < 2903) {
if (SC < 2604) {
if (mu_mv_mag_sq < 65) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 8;
}
}
else if (mu_mv_mag_sq >= 65) {
if (ssDCval < 56) {
return 2;
}
else if (ssDCval >= 56) {
if (Rs < 928) {
return 8;
}
else if (Rs >= 928) {
return 1;
}
}
}
}
else if (SC >= 2604) {
if (gchDC < 1) {
if (SC < 2764) {
if (Rs < 1275) {
return 4;
}
else if (Rs >= 1275) {
return 1;
}
}
else if (SC >= 2764) {
return 8;
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 80) {
return 1;
}
else if (mu_mv_mag_sq >= 80) {
if (Rs < 1005) {
return 1;
}
else if (Rs >= 1005) {
return 8;
}
}
}
}
}
else if (SC >= 2903) {
if (gchDC < 1) {
if (Rs < 2083) {
if (SC < 3237) {
return 1;
}
else if (SC >= 3237) {
return 8;
}
}
else if (Rs >= 2083) {
if (Rs < 2184) {
return 4;
}
else if (Rs >= 2184) {
return 8;
}
}
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 22) {
return 1;
}
else if (mu_mv_mag_sq >= 22) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (refDCval < 69) {
if (posBalance < 45) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
else if (posBalance >= 45) {
if (mcTcor < 96) {
if (Rs < 1184) {
return 8;
}
else if (Rs >= 1184) {
return 4;
}
}
else if (mcTcor >= 96) {
if (Cs < 1418) {
return 4;
}
else if (Cs >= 1418) {
return 1;
}
}
}
}
else if (refDCval >= 69) {
if (refDCval < 76) {
if (mu_mv_mag_sq < 88) {
return 8;
}
else if (mu_mv_mag_sq >= 88) {
if (MVDiff < 87) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (MVDiff >= 87) {
return 8;
}
}
}
else if (refDCval >= 76) {
if (Cs < 1421) {
if (Rs < 2100) {
return 1;
}
else if (Rs >= 2100) {
return 4;
}
}
else if (Cs >= 1421) {
if (RsCsDiff < 27) {
if (ssDCval < 81) {
return 1;
}
else if (ssDCval >= 81) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (RsCsDiff >= 27) {
if (RsCsDiff < 79) {
return 8;
}
else if (RsCsDiff >= 79) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2831) {
if (posBalance < 42) {
if (mu_mv_mag_sq < 84) {
if (ssDCval < 77) {
return 4;
}
else if (ssDCval >= 77) {
if (TSC < 6) {
if (Rs < 4885) {
if (Cs < 1451) {
return 4;
}
else if (Cs >= 1451) {
return 8;
}
}
else if (Rs >= 4885) {
return 1;
}
}
else if (TSC >= 6) {
if (Rs < 3730) {
return 8;
}
else if (Rs >= 3730) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (posBalance < 31) {
return 8;
}
else if (posBalance >= 31) {
if (SC < 4889) {
if (Rs < 3113) {
return 4;
}
else if (Rs >= 3113) {
return 2;
}
}
else if (SC >= 4889) {
if (Rs < 4401) {
if (negBalance < 23) {
return 8;
}
else if (negBalance >= 23) {
if (RsDiff < 7) {
return 4;
}
else if (RsDiff >= 7) {
if (Rs < 3185) {
return 1;
}
else if (Rs >= 3185) {
return 4;
}
}
}
}
else if (Rs >= 4401) {
if (SC < 7250) {
if (refDCval < 84) {
if (Cs < 2053) {
return 8;
}
else if (Cs >= 2053) {
if (Cs < 2253) {
return 4;
}
else if (Cs >= 2253) {
if (Rs < 4555) {
return 8;
}
else if (Rs >= 4555) {
if (Rs < 4703) {
return 4;
}
else if (Rs >= 4703) {
return 4;
}
}
}
}
}
else if (refDCval >= 84) {
return 4;
}
}
else if (SC >= 7250) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 42) {
if (Cs < 1826) {
if (Rs < 6818) {
if (posBalance < 46) {
if (Rs < 3607) {
if (RsDiff < 6) {
if (Rs < 3087) {
return 8;
}
else if (Rs >= 3087) {
return 4;
}
}
else if (RsDiff >= 6) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (Rs >= 3607) {
if (diffRsCsdiff < -39) {
return 2;
}
else if (diffRsCsdiff >= -39) {
return 4;
}
}
}
else if (posBalance >= 46) {
if (RsCsDiff < 37) {
if (Cs < 1472) {
if (Rs < 3103) {
return 4;
}
else if (Rs >= 3103) {
return 2;
}
}
else if (Cs >= 1472) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 4;
}
}
}
else if (RsCsDiff >= 37) {
if (ssDCval < 84) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (RsCsDiff < 62) {
if (mu_mv_mag_sq < 76) {
return 4;
}
else if (mu_mv_mag_sq >= 76) {
return 2;
}
}
else if (RsCsDiff >= 62) {
return 2;
}
}
}
else if (ssDCval >= 84) {
if (Rs < 3314) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 3049) {
return 2;
}
else if (Rs >= 3049) {
return 1;
}
}
}
else if (Rs >= 3314) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
return 2;
}
}
}
}
}
}
else if (Rs >= 6818) {
return 2;
}
}
else if (Cs >= 1826) {
if (RsCsDiff < 38) {
if (Cs < 2309) {
if (negBalance < 36) {
return 4;
}
else if (negBalance >= 36) {
return 8;
}
}
else if (Cs >= 2309) {
if (ssDCval < 56) {
if (Rs < 3428) {
return 8;
}
else if (Rs >= 3428) {
if (Rs < 4214) {
return 4;
}
else if (Rs >= 4214) {
return 4;
}
}
}
else if (ssDCval >= 56) {
return 8;
}
}
}
else if (RsCsDiff >= 38) {
if (Rs < 3949) {
if (TSC < 5) {
if (Rs < 3377) {
return 4;
}
else if (Rs >= 3377) {
if (Cs < 2379) {
return 8;
}
else if (Cs >= 2379) {
return 4;
}
}
}
else if (TSC >= 5) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Rs < 3046) {
return 1;
}
else if (Rs >= 3046) {
return 8;
}
}
}
}
else if (Rs >= 3949) {
if (ssDCval < 71) {
if (Rs < 5077) {
return 4;
}
else if (Rs >= 5077) {
return 4;
}
}
else if (ssDCval >= 71) {
if (negBalance < 35) {
return 2;
}
else if (negBalance >= 35) {
if (Rs < 4401) {
return 2;
}
else if (Rs >= 4401) {
if (Rs < 4808) {
return 4;
}
else if (Rs >= 4808) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (negBalance >= 39) {
if (SC < 6954) {
if (mcTcor < 94) {
if (Rs < 1454) {
if (TSC < 11) {
if (AFD < 10) {
if (MVDiff < 5) {
if (negBalance < 42) {
return 4;
}
else if (negBalance >= 42) {
if (ssDCval < 48) {
return 4;
}
else if (ssDCval >= 48) {
return 8;
}
}
}
else if (MVDiff >= 5) {
if (tcor < 81) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
return 8;
}
}
else if (tcor >= 81) {
if (Rs < 1250) {
return 4;
}
else if (Rs >= 1250) {
if (Cs < 1323) {
return 8;
}
else if (Cs >= 1323) {
return 1;
}
}
}
}
}
else if (AFD >= 10) {
if (refDCval < 55) {
if (Cs < 1572) {
return 4;
}
else if (Cs >= 1572) {
return 2;
}
}
else if (refDCval >= 55) {
if (refDCval < 72) {
if (posBalance < 59) {
if (Rs < 967) {
return 2;
}
else if (Rs >= 967) {
if (Rs < 1111) {
return 1;
}
else if (Rs >= 1111) {
return 8;
}
}
}
else if (posBalance >= 59) {
return 8;
}
}
else if (refDCval >= 72) {
if (RsDiff < 3) {
if (Rs < 776) {
return 8;
}
else if (Rs >= 776) {
return 1;
}
}
else if (RsDiff >= 3) {
if (CsDiff < 6) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
return 4;
}
}
else if (CsDiff >= 6) {
if (MVDiff < 17) {
if (Cs < 1503) {
if (Cs < 1314) {
return 2;
}
else if (Cs >= 1314) {
return 8;
}
}
else if (Cs >= 1503) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
}
else if (MVDiff >= 17) {
return 4;
}
}
}
}
}
}
}
else if (TSC >= 11) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (Rs < 1292) {
return 2;
}
else if (Rs >= 1292) {
if (Rs < 1405) {
return 8;
}
else if (Rs >= 1405) {
return 4;
}
}
}
}
}
else if (Rs >= 1454) {
if (Cs < 2918) {
if (MVDiff < 17) {
if (RsDiff < 6) {
if (posBalance < 68) {
if (mu_mv_mag_sq < 3) {
if (diffMVdiffVal < -7) {
return 4;
}
else if (diffMVdiffVal >= -7) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (Rs < 1612) {
return 4;
}
else if (Rs >= 1612) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffRsCsdiff < -17) {
if (diffRsCsdiff < -19) {
return 8;
}
else if (diffRsCsdiff >= -19) {
if (Rs < 2429) {
return 1;
}
else if (Rs >= 2429) {
return 8;
}
}
}
else if (diffRsCsdiff >= -17) {
return 8;
}
}
}
else if (posBalance >= 68) {
return 4;
}
}
else if (RsDiff >= 6) {
if (diffRsCsdiff < -25) {
if (mu_mv_mag_sq < 90) {
if (ssDCval < 60) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
else if (ssDCval >= 60) {
if (Rs < 1551) {
return 4;
}
else if (Rs >= 1551) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (mu_mv_mag_sq < 102) {
if (Rs < 1719) {
return 8;
}
else if (Rs >= 1719) {
return 4;
}
}
else if (mu_mv_mag_sq >= 102) {
return 8;
}
}
}
else if (diffRsCsdiff >= -25) {
if (refDCval < 75) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Cs < 1392) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (Cs >= 1392) {
return 8;
}
}
}
else if (refDCval >= 75) {
if (Rs < 1688) {
if (MVDiff < 10) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (Rs >= 1688) {
if (RsCsDiff < 87) {
if (mu_mv_mag_sq < 93) {
if (diffAFD < -3) {
if (Rs < 2085) {
return 8;
}
else if (Rs >= 2085) {
return 2;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 93) {
if (diffMVdiffVal < -4) {
return 4;
}
else if (diffMVdiffVal >= -4) {
if (refDCval < 76) {
return 4;
}
else if (refDCval >= 76) {
return 8;
}
}
}
}
else if (RsCsDiff >= 87) {
if (refDCval < 83) {
if (MVDiff < 10) {
if (diffMVdiffVal < -1) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (MVDiff >= 10) {
if (Cs < 1441) {
return 8;
}
else if (Cs >= 1441) {
return 4;
}
}
}
else if (refDCval >= 83) {
if (TSC < 9) {
if (Rs < 2614) {
return 2;
}
else if (Rs >= 2614) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
}
}
}
}
}
else if (MVDiff >= 17) {
if (mcTcor < 89) {
if (ssDCval < 89) {
if (TSC < 6) {
if (Cs < 1444) {
return 1;
}
else if (Cs >= 1444) {
return 8;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (ssDCval >= 89) {
if (Cs < 1482) {
return 8;
}
else if (Cs >= 1482) {
return 4;
}
}
}
else if (mcTcor >= 89) {
if (Rs < 1988) {
if (tcor < 83) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 8;
}
}
else if (tcor >= 83) {
if (Cs < 1383) {
return 8;
}
else if (Cs >= 1383) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 8;
}
}
}
}
else if (Rs >= 1988) {
if (Cs < 2105) {
return 8;
}
else if (Cs >= 2105) {
return 4;
}
}
}
}
}
else if (Cs >= 2918) {
if (Rs < 3141) {
return 4;
}
else if (Rs >= 3141) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (Cs < 1606) {
if (Rs < 2604) {
if (tcor < 89) {
if (Rs < 1023) {
if (ssDCval < 77) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
if (Rs < 874) {
return 8;
}
else if (Rs >= 874) {
return 2;
}
}
}
else if (ssDCval >= 77) {
if (diffRsCsdiff < 19) {
return 8;
}
else if (diffRsCsdiff >= 19) {
return 4;
}
}
}
else if (Rs >= 1023) {
if (CsDiff < 6) {
if (Rs < 2445) {
if (negBalance < 52) {
return 4;
}
else if (negBalance >= 52) {
return 8;
}
}
else if (Rs >= 2445) {
return 2;
}
}
else if (CsDiff >= 6) {
if (diffMVdiffVal < -1) {
if (Cs < 1550) {
return 8;
}
else if (Cs >= 1550) {
return 4;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
}
}
else if (tcor >= 89) {
if (ssDCval < 59) {
if (SC < 2859) {
if (RsCsDiff < 22) {
if (Rs < 1075) {
return 2;
}
else if (Rs >= 1075) {
return 8;
}
}
else if (RsCsDiff >= 22) {
return 4;
}
}
else if (SC >= 2859) {
return 8;
}
}
else if (ssDCval >= 59) {
if (Cs < 1461) {
if (posBalance < 52) {
if (mu_mv_mag_sq < 62) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
if (Rs < 1405) {
return 1;
}
else if (Rs >= 1405) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 62) {
return 4;
}
}
else if (posBalance >= 52) {
if (refDCval < 74) {
if (TSC < 6) {
if (Rs < 1365) {
return 4;
}
else if (Rs >= 1365) {
return 8;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (refDCval >= 74) {
if (ssDCval < 85) {
if (MVDiff < 5) {
if (refDCval < 82) {
if (Cs < 1350) {
return 1;
}
else if (Cs >= 1350) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (tcor < 93) {
if (Rs < 1050) {
return 1;
}
else if (Rs >= 1050) {
return 8;
}
}
else if (tcor >= 93) {
if (ssDCval < 79) {
if (Rs < 2116) {
return 8;
}
else if (Rs >= 2116) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
}
else if (ssDCval >= 79) {
return 1;
}
}
}
}
}
else if (refDCval >= 82) {
return 1;
}
}
else if (MVDiff >= 5) {
if (TSC < 7) {
if (Rs < 1219) {
return 1;
}
else if (Rs >= 1219) {
return 2;
}
}
else if (TSC >= 7) {
if (mu_mv_mag_sq < 51) {
return 8;
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 8;
}
}
}
}
}
else if (ssDCval >= 85) {
if (CsDiff < 4) {
if (Rs < 1002) {
return 8;
}
else if (Rs >= 1002) {
return 4;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
}
else if (Cs >= 1461) {
if (mu_mv_mag_sq < 69) {
if (mcTcor < 97) {
if (Cs < 1584) {
if (Rs < 847) {
if (Rs < 776) {
return 8;
}
else if (Rs >= 776) {
return 2;
}
}
else if (Rs >= 847) {
return 8;
}
}
else if (Cs >= 1584) {
if (Rs < 2128) {
return 4;
}
else if (Rs >= 2128) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (Rs < 750) {
return 8;
}
else if (Rs >= 750) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (RsDiff < 6) {
if (posBalance < 59) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (negBalance < 46) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
return 8;
}
}
else if (negBalance >= 46) {
return 1;
}
}
}
}
else if (posBalance >= 59) {
if (Cs < 1512) {
return 8;
}
else if (Cs >= 1512) {
return 1;
}
}
}
else if (RsDiff >= 6) {
if (ssDCval < 81) {
if (Rs < 1872) {
return 4;
}
else if (Rs >= 1872) {
return 8;
}
}
else if (ssDCval >= 81) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2430) {
return 2;
}
else if (Rs >= 2430) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2604) {
if (RsDiff < 7) {
if (MVDiff < 4) {
if (tcor < 96) {
if (Rs < 2892) {
return 2;
}
else if (Rs >= 2892) {
return 4;
}
}
else if (tcor >= 96) {
return 2;
}
}
else if (MVDiff >= 4) {
if (Rs < 2690) {
return 8;
}
else if (Rs >= 2690) {
return 1;
}
}
}
else if (RsDiff >= 7) {
if (AFD < 13) {
if (RsCsDiff < 77) {
if (Cs < 1406) {
return 2;
}
else if (Cs >= 1406) {
return 4;
}
}
else if (RsCsDiff >= 77) {
return 2;
}
}
else if (AFD >= 13) {
return 4;
}
}
}
}
else if (Cs >= 1606) {
if (SC < 5099) {
if (Rs < 1465) {
if (refDCval < 58) {
return 8;
}
else if (refDCval >= 58) {
if (TSC < 7) {
if (SC < 2421) {
return 8;
}
else if (SC >= 2421) {
if (diffRsCsdiff < -13) {
if (SC < 2515) {
return 1;
}
else if (SC >= 2515) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
else if (diffRsCsdiff >= -13) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
else if (TSC >= 7) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
return 4;
}
}
}
}
else if (Rs >= 1465) {
if (AFD < 10) {
if (Rs < 3129) {
if (refDCval < 79) {
if (tcor < 94) {
if (MVDiff < 5) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (tcor >= 94) {
return 8;
}
}
else if (refDCval >= 79) {
if (mu_mv_mag_sq < 86) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Rs < 1973) {
return 1;
}
else if (Rs >= 1973) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
if (negBalance < 43) {
if (Cs < 1632) {
return 8;
}
else if (Cs >= 1632) {
return 1;
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
return 8;
}
}
}
else if (Rs >= 3129) {
if (Rs < 3240) {
return 4;
}
else if (Rs >= 3240) {
return 2;
}
}
}
else if (AFD >= 10) {
if (ssDCval < 79) {
if (Rs < 1688) {
return 8;
}
else if (Rs >= 1688) {
return 4;
}
}
else if (ssDCval >= 79) {
if (Cs < 2268) {
if (tcor < 84) {
return 4;
}
else if (tcor >= 84) {
if (MVDiff < 5) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
return 8;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (Cs >= 2268) {
return 1;
}
}
}
}
}
else if (SC >= 5099) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 2782) {
return 4;
}
else if (Rs >= 2782) {
if (posBalance < 54) {
if (Rs < 3287) {
return 8;
}
else if (Rs >= 3287) {
if (negBalance < 52) {
if (diffRsCsdiff < 23) {
return 4;
}
else if (diffRsCsdiff >= 23) {
return 8;
}
}
else if (negBalance >= 52) {
if (Rs < 3789) {
return 2;
}
else if (Rs >= 3789) {
return 8;
}
}
}
}
else if (posBalance >= 54) {
if (posBalance < 56) {
if (Rs < 3388) {
return 1;
}
else if (Rs >= 3388) {
if (Cs < 2542) {
if (Rs < 3796) {
return 4;
}
else if (Rs >= 3796) {
return 2;
}
}
else if (Cs >= 2542) {
return 8;
}
}
}
else if (posBalance >= 56) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 6954) {
if (Cs < 2649) {
return 4;
}
else if (Cs >= 2649) {
return 8;
}
}
}
}
else if (ssDCval >= 90) {
if (Rs < 4425) {
if (ssDCval < 165) {
if (refDCval < 109) {
if (mcTcor < 96) {
if (Rs < 2083) {
if (Rs < 1572) {
if (ssDCval < 100) {
if (Cs < 1691) {
if (refDCval < 96) {
return 8;
}
else if (refDCval >= 96) {
if (Rs < 1383) {
return 4;
}
else if (Rs >= 1383) {
return 8;
}
}
}
else if (Cs >= 1691) {
if (MVDiff < 7) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 4;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (ssDCval >= 100) {
if (refDCval < 107) {
return 4;
}
else if (refDCval >= 107) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
else if (Rs >= 1572) {
if (mcTcor < 89) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
else if (mcTcor >= 89) {
if (AFD < 12) {
if (RsCsDiff < 38) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
return 2;
}
}
else if (RsCsDiff >= 38) {
if (AFD < 8) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (AFD >= 8) {
if (RsDiff < 7) {
if (Cs < 1443) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Rs < 1618) {
return 8;
}
else if (Rs >= 1618) {
return 2;
}
}
}
else if (Cs >= 1443) {
return 8;
}
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (Rs >= 2083) {
if (Cs < 2698) {
if (mcTcor < 94) {
if (TSC < 11) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (mcTcor < 93) {
if (Rs < 2328) {
if (Rs < 2189) {
return 8;
}
else if (Rs >= 2189) {
return 2;
}
}
else if (Rs >= 2328) {
return 8;
}
}
else if (mcTcor >= 93) {
if (MVDiff < 11) {
if (Rs < 2615) {
return 8;
}
else if (Rs >= 2615) {
return 4;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
}
}
}
else if (TSC >= 11) {
if (ssDCval < 101) {
return 8;
}
else if (ssDCval >= 101) {
if (Rs < 2748) {
return 1;
}
else if (Rs >= 2748) {
return 4;
}
}
}
}
else if (mcTcor >= 94) {
if (negBalance < 55) {
if (negBalance < 34) {
if (AFD < 10) {
if (Rs < 4088) {
if (posBalance < 35) {
if (SC < 5562) {
return 4;
}
else if (SC >= 5562) {
return 8;
}
}
else if (posBalance >= 35) {
return 8;
}
}
else if (Rs >= 4088) {
if (diffRsCsdiff < 4) {
if (Cs < 1857) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
if (Rs < 4284) {
return 4;
}
else if (Rs >= 4284) {
return 4;
}
}
}
else if (Cs >= 1857) {
return 2;
}
}
else if (diffRsCsdiff >= 4) {
return 8;
}
}
}
else if (AFD >= 10) {
return 1;
}
}
else if (negBalance >= 34) {
if (refDCval < 103) {
if (AFD < 5) {
if (Rs < 2285) {
return 1;
}
else if (Rs >= 2285) {
return 4;
}
}
else if (AFD >= 5) {
if (Cs < 1330) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 8;
}
}
else if (Cs >= 1330) {
return 8;
}
}
}
else if (refDCval >= 103) {
if (posBalance < 47) {
if (Rs < 3219) {
return 4;
}
else if (Rs >= 3219) {
return 8;
}
}
else if (posBalance >= 47) {
if (diffMVdiffVal < -1) {
if (negBalance < 45) {
if (posBalance < 49) {
if (Rs < 2948) {
return 4;
}
else if (Rs >= 2948) {
return 8;
}
}
else if (posBalance >= 49) {
return 8;
}
}
else if (negBalance >= 45) {
if (Cs < 1815) {
return 4;
}
else if (Cs >= 1815) {
return 8;
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
}
}
else if (negBalance >= 55) {
if (Rs < 2694) {
return 8;
}
else if (Rs >= 2694) {
return 4;
}
}
}
}
else if (Cs >= 2698) {
if (ssDCval < 103) {
if (Rs < 4083) {
return 4;
}
else if (Rs >= 4083) {
return 8;
}
}
else if (ssDCval >= 103) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (RsCsDiff < 41) {
if (MVDiff < 4) {
if (Cs < 1685) {
if (ssDCval < 98) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (TSC < 5) {
if (Cs < 1519) {
if (Rs < 1635) {
return 8;
}
else if (Rs >= 1635) {
return 2;
}
}
else if (Cs >= 1519) {
return 8;
}
}
else if (TSC >= 5) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
}
}
else if (ssDCval >= 98) {
if (TSC < 6) {
if (ssDCval < 105) {
return 4;
}
else if (ssDCval >= 105) {
if (Cs < 1508) {
return 4;
}
else if (Cs >= 1508) {
return 8;
}
}
}
else if (TSC >= 6) {
return 8;
}
}
}
else if (Cs >= 1685) {
if (tcor < 97) {
if (Rs < 4063) {
if (Rs < 2533) {
if (Cs < 2153) {
return 8;
}
else if (Cs >= 2153) {
return 4;
}
}
else if (Rs >= 2533) {
return 4;
}
}
else if (Rs >= 4063) {
return 8;
}
}
else if (tcor >= 97) {
if (CsDiff < 1) {
if (Rs < 2383) {
if (mu_mv_mag_sq < 110) {
return 4;
}
else if (mu_mv_mag_sq >= 110) {
if (Rs < 2338) {
return 2;
}
else if (Rs >= 2338) {
return 1;
}
}
}
else if (Rs >= 2383) {
return 8;
}
}
else if (CsDiff >= 1) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
return 2;
}
}
}
}
}
else if (MVDiff >= 4) {
if (Rs < 2860) {
if (Rs < 2672) {
if (posBalance < 33) {
if (Rs < 2183) {
return 2;
}
else if (Rs >= 2183) {
if (refDCval < 92) {
return 4;
}
else if (refDCval >= 92) {
if (negBalance < 38) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 1;
}
}
else if (negBalance >= 38) {
if (Cs < 2043) {
return 4;
}
else if (Cs >= 2043) {
return 8;
}
}
}
}
}
else if (posBalance >= 33) {
if (Rs < 1899) {
if (Rs < 1643) {
if (Cs < 1361) {
return 4;
}
else if (Cs >= 1361) {
return 8;
}
}
else if (Rs >= 1643) {
return 2;
}
}
else if (Rs >= 1899) {
return 8;
}
}
}
else if (Rs >= 2672) {
if (diffRsCsdiff < 8) {
return 4;
}
else if (diffRsCsdiff >= 8) {
return 8;
}
}
}
else if (Rs >= 2860) {
if (refDCval < 90) {
return 4;
}
else if (refDCval >= 90) {
return 8;
}
}
}
}
else if (RsCsDiff >= 41) {
if (Rs < 3563) {
if (Rs < 2631) {
if (Rs < 1907) {
if (SC < 3093) {
if (MVDiff < 16) {
return 8;
}
else if (MVDiff >= 16) {
return 2;
}
}
else if (SC >= 3093) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (Rs >= 1907) {
if (mu_mv_mag_sq < 52) {
if (MVDiff < 6) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 1982) {
return 1;
}
else if (Rs >= 1982) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (Rs < 1920) {
return 8;
}
else if (Rs >= 1920) {
if (Rs < 2498) {
return 2;
}
else if (Rs >= 2498) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (diffRsCsdiff < -35) {
return 1;
}
else if (diffRsCsdiff >= -35) {
return 8;
}
}
}
}
else if (Rs >= 2631) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
if (MVDiff < 3) {
if (SC < 4527) {
return 8;
}
else if (SC >= 4527) {
return 4;
}
}
else if (MVDiff >= 3) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
if (AFD < 10) {
if (Cs < 1391) {
return 8;
}
else if (Cs >= 1391) {
return 4;
}
}
else if (AFD >= 10) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3563) {
if (Rs < 4153) {
if (MVDiff < 4) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
return 4;
}
}
else if (MVDiff >= 4) {
if (diffRsCsdiff < -33) {
return 2;
}
else if (diffRsCsdiff >= -33) {
return 8;
}
}
}
else if (Rs >= 4153) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 1265) {
if (ssDCval < 157) {
if (refDCval < 111) {
if (negBalance < 33) {
if (Cs < 1414) {
return 1;
}
else if (Cs >= 1414) {
return 8;
}
}
else if (negBalance >= 33) {
if (Cs < 1840) {
return 8;
}
else if (Cs >= 1840) {
return 4;
}
}
}
else if (refDCval >= 111) {
if (SC < 2658) {
if (mcTcor < 96) {
if (AFD < 7) {
if (ssDCval < 129) {
if (posBalance < 55) {
if (Rs < 1107) {
return 4;
}
else if (Rs >= 1107) {
return 8;
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (ssDCval >= 129) {
return 8;
}
}
else if (AFD >= 7) {
if (ssDCval < 150) {
return 8;
}
else if (ssDCval >= 150) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (SC >= 2658) {
if (Cs < 1742) {
if (ssDCval < 130) {
if (negBalance < 37) {
if (SC < 2751) {
return 4;
}
else if (SC >= 2751) {
return 8;
}
}
else if (negBalance >= 37) {
return 8;
}
}
else if (ssDCval >= 130) {
if (ssDCval < 134) {
return 4;
}
else if (ssDCval >= 134) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 4;
}
}
}
}
else if (Cs >= 1742) {
return 8;
}
}
}
}
else if (ssDCval >= 157) {
if (Cs < 1772) {
if (Rs < 1182) {
return 4;
}
else if (Rs >= 1182) {
return 2;
}
}
else if (Cs >= 1772) {
return 1;
}
}
}
else if (Rs >= 1265) {
if (mcTcor < 92) {
if (refDCval < 140) {
if (refDCval < 125) {
if (TSC < 11) {
if (ssDCval < 131) {
if (diffMVdiffVal < -7) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (diffMVdiffVal >= -7) {
return 8;
}
}
else if (ssDCval >= 131) {
return 1;
}
}
else if (TSC >= 11) {
if (ssDCval < 111) {
return 1;
}
else if (ssDCval >= 111) {
return 8;
}
}
}
else if (refDCval >= 125) {
if (mcTcor < 89) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (AFD < 11) {
if (Cs < 1759) {
return 8;
}
else if (Cs >= 1759) {
return 1;
}
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (mcTcor >= 89) {
if (Rs < 3584) {
if (SC < 3694) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (SC >= 3694) {
return 8;
}
}
else if (Rs >= 3584) {
return 1;
}
}
}
}
else if (refDCval >= 140) {
if (Rs < 1483) {
if (refDCval < 151) {
return 8;
}
else if (refDCval >= 151) {
return 2;
}
}
else if (Rs >= 1483) {
if (diffMVdiffVal < -12) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (Rs < 1636) {
return 1;
}
else if (Rs >= 1636) {
return 2;
}
}
}
else if (diffMVdiffVal >= -12) {
return 8;
}
}
}
}
else if (mcTcor >= 92) {
if (AFD < 12) {
if (posBalance < 36) {
if (negBalance < 49) {
if (negBalance < 40) {
if (SC < 5158) {
if (ssDCval < 107) {
return 1;
}
else if (ssDCval >= 107) {
return 8;
}
}
else if (SC >= 5158) {
if (Rs < 3424) {
return 4;
}
else if (Rs >= 3424) {
return 1;
}
}
}
else if (negBalance >= 40) {
if (Rs < 2138) {
return 2;
}
else if (Rs >= 2138) {
if (MVDiff < 3) {
if (SC < 4880) {
return 8;
}
else if (SC >= 4880) {
return 2;
}
}
else if (MVDiff >= 3) {
if (MVDiff < 6) {
if (negBalance < 43) {
return 4;
}
else if (negBalance >= 43) {
return 8;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
}
}
}
else if (negBalance >= 49) {
if (tcor < 94) {
if (Rs < 1774) {
return 2;
}
else if (Rs >= 1774) {
return 8;
}
}
else if (tcor >= 94) {
return 8;
}
}
}
else if (posBalance >= 36) {
if (ssDCval < 163) {
if (Cs < 1375) {
if (mu_mv_mag_sq < 88) {
if (Rs < 3626) {
return 8;
}
else if (Rs >= 3626) {
if (Rs < 3846) {
return 2;
}
else if (Rs >= 3846) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 88) {
if (Rs < 2298) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
else if (Rs >= 2298) {
if (Rs < 2678) {
if (Rs < 2425) {
return 1;
}
else if (Rs >= 2425) {
return 8;
}
}
else if (Rs >= 2678) {
if (Rs < 3696) {
return 2;
}
else if (Rs >= 3696) {
return 1;
}
}
}
}
}
else if (Cs >= 1375) {
if (mu_mv_mag_sq < 3) {
if (Rs < 2067) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
}
else if (Rs >= 2067) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 9) {
if (MVDiff < 6) {
if (ssDCval < 123) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (ssDCval >= 123) {
if (posBalance < 56) {
if (diffTSC < -1) {
if (negBalance < 32) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
return 1;
}
}
else if (negBalance >= 32) {
if (ssDCval < 127) {
if (Rs < 3545) {
return 1;
}
else if (Rs >= 3545) {
return 8;
}
}
else if (ssDCval >= 127) {
return 8;
}
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (posBalance >= 56) {
if (mcTcor < 94) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (ssDCval < 141) {
return 8;
}
else if (ssDCval >= 141) {
return 1;
}
}
}
else if (mcTcor >= 94) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 8;
}
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 113) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 2;
}
}
else if (ssDCval >= 113) {
return 8;
}
}
}
else if (TSC >= 9) {
if (Cs < 1809) {
if (MVDiff < 10) {
if (Rs < 2972) {
return 1;
}
else if (Rs >= 2972) {
return 8;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (Cs >= 1809) {
if (diffAFD < 7) {
return 8;
}
else if (diffAFD >= 7) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2262) {
return 2;
}
else if (Rs >= 2262) {
return 8;
}
}
}
}
else if (AFD >= 12) {
if (SC < 3267) {
if (negBalance < 56) {
if (Cs < 1655) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 8;
}
}
else if (Cs >= 1655) {
return 4;
}
}
else if (negBalance >= 56) {
if (Rs < 1716) {
return 2;
}
else if (Rs >= 1716) {
return 1;
}
}
}
else if (SC >= 3267) {
if (diffRsCsdiff < -20) {
if (diffMVdiffVal < -26) {
return 2;
}
else if (diffMVdiffVal >= -26) {
if (posBalance < 63) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 8;
}
}
else if (posBalance >= 63) {
if (Rs < 3409) {
return 4;
}
else if (Rs >= 3409) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -20) {
if (negBalance < 47) {
if (mu_mv_mag_sq < 53) {
if (tcor < 89) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 1;
}
}
else if (tcor >= 89) {
return 1;
}
}
else if (mu_mv_mag_sq >= 53) {
return 8;
}
}
else if (negBalance >= 47) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 165) {
if (Cs < 1914) {
return 2;
}
else if (Cs >= 1914) {
if (Rs < 2321) {
return 1;
}
else if (Rs >= 2321) {
return 8;
}
}
}
}
else if (Rs >= 4425) {
if (Cs < 1590) {
return 2;
}
else if (Cs >= 1590) {
if (AFD < 6) {
if (Cs < 2023) {
if (tcor < 97) {
if (Rs < 4559) {
return 2;
}
else if (Rs >= 4559) {
return 8;
}
}
else if (tcor >= 97) {
if (Rs < 4769) {
return 2;
}
else if (Rs >= 4769) {
if (Rs < 5059) {
return 8;
}
else if (Rs >= 5059) {
return 2;
}
}
}
}
else if (Cs >= 2023) {
return 8;
}
}
else if (AFD >= 6) {
if (Rs < 6251) {
if (diffAFD < -3) {
if (Rs < 4598) {
return 1;
}
else if (Rs >= 4598) {
return 2;
}
}
else if (diffAFD >= -3) {
if (mcTcor < 94) {
if (Rs < 4649) {
return 8;
}
else if (Rs >= 4649) {
return 1;
}
}
else if (mcTcor >= 94) {
return 8;
}
}
}
else if (Rs >= 6251) {
return 4;
}
}
}
}
}
}
else if (CsDiff >= 7) {
if (Rs < 3696) {
if (Rs < 1511) {
if (negBalance < 41) {
if (mcTcor < 96) {
if (diffRsCsdiff < -32) {
if (mcTcor < 76) {
return 4;
}
else if (mcTcor >= 76) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Cs < 2451) {
if (Rs < 1170) {
return 1;
}
else if (Rs >= 1170) {
return 4;
}
}
else if (Cs >= 2451) {
if (Rs < 1428) {
return 2;
}
else if (Rs >= 1428) {
return 8;
}
}
}
}
}
else if (diffRsCsdiff >= -32) {
if (mu_mv_mag_sq < 38) {
if (mcTcor < 79) {
return 8;
}
else if (mcTcor >= 79) {
if (RsDiff < 5) {
if (SC < 2024) {
return 4;
}
else if (SC >= 2024) {
if (diffAFD < 3) {
if (Rs < 1054) {
return 8;
}
else if (Rs >= 1054) {
return 2;
}
}
else if (diffAFD >= 3) {
return 2;
}
}
}
else if (RsDiff >= 5) {
if (mcTcor < 92) {
if (CsDiff < 16) {
if (TSC < 16) {
return 1;
}
else if (TSC >= 16) {
return 4;
}
}
else if (CsDiff >= 16) {
return 2;
}
}
else if (mcTcor >= 92) {
if (Cs < 1373) {
return 8;
}
else if (Cs >= 1373) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (MVDiff < 21) {
if (negBalance < 31) {
if (Cs < 2855) {
return 4;
}
else if (Cs >= 2855) {
return 1;
}
}
else if (negBalance >= 31) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (MVDiff >= 21) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (Rs < 891) {
return 1;
}
else if (Rs >= 891) {
if (Rs < 1204) {
return 2;
}
else if (Rs >= 1204) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 41) {
if (tcor < 60) {
if (mcTcor < 74) {
if (gchDC < 3) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (refDCval < 70) {
return 4;
}
else if (refDCval >= 70) {
if (Rs < 1249) {
return 2;
}
else if (Rs >= 1249) {
return 4;
}
}
}
}
else if (gchDC >= 3) {
if (Rs < 1081) {
return 1;
}
else if (Rs >= 1081) {
return 8;
}
}
}
else if (mcTcor >= 74) {
if (Rs < 950) {
if (Rs < 550) {
return 4;
}
else if (Rs >= 550) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (Rs >= 950) {
if (mcTcor < 85) {
if (tcor < 54) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 1;
}
}
else if (tcor >= 54) {
if (posBalance < 63) {
if (Rs < 1377) {
return 4;
}
else if (Rs >= 1377) {
return 2;
}
}
else if (posBalance >= 63) {
return 2;
}
}
}
else if (mcTcor >= 85) {
if (Rs < 1492) {
return 4;
}
else if (Rs >= 1492) {
return 2;
}
}
}
}
}
else if (tcor >= 60) {
if (refDCval < 93) {
if (TSC < 7) {
if (ssDCval < 63) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (ssDCval >= 63) {
if (tcor < 89) {
if (Cs < 1334) {
return 4;
}
else if (Cs >= 1334) {
return 8;
}
}
else if (tcor >= 89) {
if (Cs < 1528) {
return 2;
}
else if (Cs >= 1528) {
return 8;
}
}
}
}
else if (TSC >= 7) {
if (diffMVdiffVal < -8) {
if (AFD < 16) {
if (Rs < 1268) {
return 4;
}
else if (Rs >= 1268) {
if (Rs < 1441) {
return 1;
}
else if (Rs >= 1441) {
return 4;
}
}
}
else if (AFD >= 16) {
if (Cs < 1334) {
return 2;
}
else if (Cs >= 1334) {
if (ssDCval < 75) {
if (Rs < 1394) {
return 2;
}
else if (Rs >= 1394) {
return 4;
}
}
else if (ssDCval >= 75) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= -8) {
if (negBalance < 63) {
if (mcTcor < 90) {
if (Rs < 1451) {
if (diffTSC < 2) {
if (MVDiff < 86) {
if (diffAFD < -1) {
if (gchDC < 3) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 4;
}
}
else if (gchDC >= 3) {
return 8;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (MVDiff >= 86) {
return 8;
}
}
else if (diffTSC >= 2) {
if (Rs < 1301) {
return 2;
}
else if (Rs >= 1301) {
return 4;
}
}
}
else if (Rs >= 1451) {
if (Cs < 1729) {
if (Rs < 1499) {
return 8;
}
else if (Rs >= 1499) {
return 4;
}
}
else if (Cs >= 1729) {
return 4;
}
}
}
else if (mcTcor >= 90) {
if (refDCval < 75) {
if (ssDCval < 57) {
return 2;
}
else if (ssDCval >= 57) {
if (MVDiff < 18) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
else if (MVDiff >= 18) {
if (Cs < 3298) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (Rs < 727) {
return 4;
}
else if (Rs >= 727) {
return 8;
}
}
}
else if (Cs >= 3298) {
return 2;
}
}
}
}
else if (refDCval >= 75) {
if (AFD < 22) {
if (TSC < 10) {
if (Cs < 1369) {
if (Rs < 1336) {
return 2;
}
else if (Rs >= 1336) {
return 8;
}
}
else if (Cs >= 1369) {
if (Rs < 1438) {
return 4;
}
else if (Rs >= 1438) {
if (Rs < 1473) {
return 2;
}
else if (Rs >= 1473) {
return 4;
}
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (AFD >= 22) {
return 8;
}
}
}
}
else if (negBalance >= 63) {
if (ssDCval < 62) {
if (Rs < 1203) {
return 8;
}
else if (Rs >= 1203) {
return 1;
}
}
else if (ssDCval >= 62) {
return 2;
}
}
}
}
}
else if (refDCval >= 93) {
if (MVDiff < 13) {
if (AFD < 16) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
return 8;
}
}
else if (AFD >= 16) {
if (mu_mv_mag_sq < 5) {
if (Cs < 1979) {
return 8;
}
else if (Cs >= 1979) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
if (RsCsDiff < 86) {
return 8;
}
else if (RsCsDiff >= 86) {
return 2;
}
}
}
}
else if (MVDiff >= 13) {
if (mcTcor < 91) {
if (posBalance < 65) {
if (diffTSC < -2) {
return 1;
}
else if (diffTSC >= -2) {
return 8;
}
}
else if (posBalance >= 65) {
if (SC < 2794) {
return 2;
}
else if (SC >= 2794) {
return 4;
}
}
}
else if (mcTcor >= 91) {
if (AFD < 24) {
if (Cs < 1811) {
if (posBalance < 60) {
if (RsDiff < 8) {
if (ssDCval < 88) {
return 4;
}
else if (ssDCval >= 88) {
return 8;
}
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (posBalance >= 60) {
if (ssDCval < 102) {
return 2;
}
else if (ssDCval >= 102) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
}
}
else if (Cs >= 1811) {
if (CsDiff < 12) {
return 1;
}
else if (CsDiff >= 12) {
return 4;
}
}
}
else if (AFD >= 24) {
if (MVDiff < 24) {
return 1;
}
else if (MVDiff >= 24) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 1511) {
if (Cs < 2719) {
if (RsDiff < 5) {
if (Rs < 1950) {
if (Cs < 2286) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (Cs >= 2286) {
return 2;
}
}
else if (Rs >= 1950) {
return 1;
}
}
else if (RsDiff >= 5) {
if (AFD < 17) {
if (ssDCval < 52) {
if (TSC < 6) {
if (Rs < 1778) {
return 1;
}
else if (Rs >= 1778) {
return 8;
}
}
else if (TSC >= 6) {
if (Cs < 1707) {
if (Rs < 1684) {
if (Cs < 1548) {
return 8;
}
else if (Cs >= 1548) {
return 4;
}
}
else if (Rs >= 1684) {
return 8;
}
}
else if (Cs >= 1707) {
if (MVDiff < 19) {
if (Rs < 2797) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
else if (Rs >= 2797) {
return 8;
}
}
else if (MVDiff >= 19) {
if (Rs < 1855) {
return 2;
}
else if (Rs >= 1855) {
return 8;
}
}
}
}
}
else if (ssDCval >= 52) {
if (SC < 5436) {
if (mcTcor < 94) {
if (refDCval < 79) {
if (posBalance < 48) {
if (Cs < 2445) {
if (SC < 3265) {
return 8;
}
else if (SC >= 3265) {
if (Rs < 3217) {
return 1;
}
else if (Rs >= 3217) {
return 4;
}
}
}
else if (Cs >= 2445) {
return 4;
}
}
else if (posBalance >= 48) {
if (diffMVdiffVal < 6) {
if (Rs < 2827) {
if (negBalance < 35) {
if (Rs < 1919) {
return 8;
}
else if (Rs >= 1919) {
if (Rs < 2096) {
return 4;
}
else if (Rs >= 2096) {
return 1;
}
}
}
else if (negBalance >= 35) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
if (Cs < 2717) {
if (ssDCval < 79) {
if (diffAFD < -1) {
if (Cs < 2418) {
return 8;
}
else if (Cs >= 2418) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (ssDCval >= 79) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (Cs >= 2717) {
return 1;
}
}
}
}
else if (Rs >= 2827) {
if (Rs < 3273) {
return 4;
}
else if (Rs >= 3273) {
return 2;
}
}
}
else if (diffMVdiffVal >= 6) {
if (AFD < 14) {
if (Rs < 1931) {
return 8;
}
else if (Rs >= 1931) {
return 4;
}
}
else if (AFD >= 14) {
return 8;
}
}
}
}
else if (refDCval >= 79) {
if (posBalance < 62) {
if (diffRsCsdiff < -113) {
if (Rs < 2450) {
return 8;
}
else if (Rs >= 2450) {
return 4;
}
}
else if (diffRsCsdiff >= -113) {
if (Rs < 2743) {
if (diffTSC < 3) {
return 8;
}
else if (diffTSC >= 3) {
if (SC < 4119) {
return 4;
}
else if (SC >= 4119) {
return 8;
}
}
}
else if (Rs >= 2743) {
if (Rs < 2747) {
return 4;
}
else if (Rs >= 2747) {
if (mcTcor < 90) {
if (mcTcor < 89) {
return 8;
}
else if (mcTcor >= 89) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Rs < 2805) {
return 8;
}
else if (Rs >= 2805) {
return 4;
}
}
}
}
else if (mcTcor >= 90) {
return 8;
}
}
}
}
}
else if (posBalance >= 62) {
if (negBalance < 55) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Cs < 2018) {
return 8;
}
else if (Cs >= 2018) {
if (Rs < 2032) {
return 4;
}
else if (Rs >= 2032) {
return 8;
}
}
}
}
else if (negBalance >= 55) {
if (Rs < 1873) {
return 4;
}
else if (Rs >= 1873) {
return 8;
}
}
}
}
}
else if (mcTcor >= 94) {
if (ssDCval < 86) {
if (MVDiff < 4) {
if (Rs < 1868) {
return 8;
}
else if (Rs >= 1868) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (AFD < 15) {
if (Rs < 2790) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
else if (Rs >= 2790) {
return 4;
}
}
else if (AFD >= 15) {
if (Rs < 1610) {
return 1;
}
else if (Rs >= 1610) {
return 4;
}
}
}
}
else if (ssDCval >= 86) {
if (Rs < 2141) {
if (TSC < 8) {
if (negBalance < 55) {
if (Rs < 2037) {
if (Rs < 1529) {
return 2;
}
else if (Rs >= 1529) {
return 8;
}
}
else if (Rs >= 2037) {
if (Cs < 2120) {
return 2;
}
else if (Cs >= 2120) {
return 8;
}
}
}
else if (negBalance >= 55) {
return 2;
}
}
else if (TSC >= 8) {
if (SC < 3980) {
if (Cs < 1363) {
return 1;
}
else if (Cs >= 1363) {
if (AFD < 16) {
if (diffRsCsdiff < 45) {
return 8;
}
else if (diffRsCsdiff >= 45) {
return 1;
}
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (SC >= 3980) {
if (Rs < 1682) {
return 8;
}
else if (Rs >= 1682) {
return 1;
}
}
}
}
else if (Rs >= 2141) {
if (diffRsCsdiff < 37) {
return 8;
}
else if (diffRsCsdiff >= 37) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
}
}
else if (SC >= 5436) {
if (refDCval < 71) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 8;
}
}
else if (refDCval >= 71) {
if (posBalance < 59) {
if (ssDCval < 87) {
if (Cs < 2596) {
return 4;
}
else if (Cs >= 2596) {
return 8;
}
}
else if (ssDCval >= 87) {
return 8;
}
}
else if (posBalance >= 59) {
if (ssDCval < 116) {
if (mu_mv_mag_sq < 50) {
return 8;
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 3019) {
return 8;
}
else if (Rs >= 3019) {
return 4;
}
}
}
else if (ssDCval >= 116) {
return 1;
}
}
}
}
}
}
else if (AFD >= 17) {
if (negBalance < 59) {
if (ssDCval < 89) {
if (negBalance < 52) {
if (AFD < 18) {
if (Rs < 1716) {
return 8;
}
else if (Rs >= 1716) {
return 1;
}
}
else if (AFD >= 18) {
if (Cs < 1432) {
if (Rs < 1868) {
return 2;
}
else if (Rs >= 1868) {
return 8;
}
}
else if (Cs >= 1432) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (CsDiff < 9) {
return 4;
}
else if (CsDiff >= 9) {
return 2;
}
}
}
}
}
else if (negBalance >= 52) {
if (tcor < 78) {
if (Rs < 2221) {
if (tcor < 68) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
return 4;
}
}
else if (tcor >= 68) {
if (ssDCval < 79) {
if (posBalance < 56) {
if (Cs < 1766) {
return 4;
}
else if (Cs >= 1766) {
return 1;
}
}
else if (posBalance >= 56) {
return 1;
}
}
else if (ssDCval >= 79) {
return 4;
}
}
}
else if (Rs >= 2221) {
if (CsDiff < 10) {
return 8;
}
else if (CsDiff >= 10) {
return 1;
}
}
}
else if (tcor >= 78) {
return 8;
}
}
}
else if (ssDCval >= 89) {
if (ssDCval < 97) {
if (MVDiff < 42) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
if (Rs < 1759) {
return 1;
}
else if (Rs >= 1759) {
if (Rs < 2020) {
return 8;
}
else if (Rs >= 2020) {
return 4;
}
}
}
}
}
else if (MVDiff >= 42) {
return 1;
}
}
else if (ssDCval >= 97) {
if (MVDiff < 9) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
return 4;
}
}
else if (MVDiff >= 9) {
if (AFD < 32) {
return 8;
}
else if (AFD >= 32) {
return 4;
}
}
}
}
}
else if (negBalance >= 59) {
if (MVDiff < 15) {
if (mu_mv_mag_sq < 10) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
return 2;
}
}
else if (mu_mv_mag_sq >= 10) {
return 8;
}
}
else if (MVDiff >= 15) {
if (ssDCval < 94) {
if (Rs < 1657) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 4;
}
}
else if (Rs >= 1657) {
return 4;
}
}
else if (ssDCval >= 94) {
if (Cs < 2116) {
if (Rs < 2216) {
return 1;
}
else if (Rs >= 2216) {
return 8;
}
}
else if (Cs >= 2116) {
if (Cs < 2248) {
return 2;
}
else if (Cs >= 2248) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 2719) {
if (ssDCval < 102) {
if (negBalance < 49) {
if (TSC < 11) {
if (Rs < 3036) {
if (ssDCval < 94) {
if (Cs < 2908) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (Cs >= 2908) {
return 4;
}
}
else if (ssDCval >= 94) {
if (Rs < 2192) {
return 8;
}
else if (Rs >= 2192) {
return 4;
}
}
}
else if (Rs >= 3036) {
if (mu_mv_mag_sq < 85) {
return 1;
}
else if (mu_mv_mag_sq >= 85) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (SC < 6204) {
return 1;
}
else if (SC >= 6204) {
return 4;
}
}
}
}
}
else if (TSC >= 11) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
return 4;
}
}
}
else if (negBalance >= 49) {
if (ssDCval < 90) {
if (TSC < 10) {
if (Rs < 2652) {
if (AFD < 16) {
if (Rs < 2640) {
if (Cs < 2739) {
return 2;
}
else if (Cs >= 2739) {
return 4;
}
}
else if (Rs >= 2640) {
return 2;
}
}
else if (AFD >= 16) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 4;
}
}
}
else if (Rs >= 2652) {
if (mcTcor < 94) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Rs < 3450) {
return 1;
}
else if (Rs >= 3450) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (AFD < 14) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (AFD >= 14) {
if (diffRsCsdiff < -50) {
if (Rs < 2165) {
return 8;
}
else if (Rs >= 2165) {
return 1;
}
}
else if (diffRsCsdiff >= -50) {
return 8;
}
}
}
}
else if (ssDCval >= 90) {
if (negBalance < 51) {
if (Rs < 3289) {
if (refDCval < 90) {
return 4;
}
else if (refDCval >= 90) {
return 8;
}
}
else if (Rs >= 3289) {
return 4;
}
}
else if (negBalance >= 51) {
if (Cs < 3920) {
return 4;
}
else if (Cs >= 3920) {
if (CsDiff < 15) {
return 8;
}
else if (CsDiff >= 15) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (tcor < 88) {
if (MVDiff < 13) {
if (SC < 6471) {
if (tcor < 76) {
return 8;
}
else if (tcor >= 76) {
return 1;
}
}
else if (SC >= 6471) {
return 4;
}
}
else if (MVDiff >= 13) {
if (CsDiff < 9) {
return 2;
}
else if (CsDiff >= 9) {
return 8;
}
}
}
else if (tcor >= 88) {
return 8;
}
}
}
}
}
else if (Rs >= 3696) {
if (mu_mv_mag_sq < 89) {
if (TSC < 6) {
if (CsDiff < 8) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
return 2;
}
}
else if (CsDiff >= 8) {
return 1;
}
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 78) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (mu_mv_mag_sq < 76) {
return 8;
}
else if (mu_mv_mag_sq >= 76) {
if (Cs < 1916) {
return 1;
}
else if (Cs >= 1916) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (ssDCval < 107) {
if (Cs < 2503) {
return 1;
}
else if (Cs >= 2503) {
if (Cs < 2844) {
return 4;
}
else if (Cs >= 2844) {
return 8;
}
}
}
else if (ssDCval >= 107) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (Cs < 2824) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
if (posBalance < 61) {
if (mu_mv_mag_sq < 92) {
if (diffRsCsdiff < 30) {
return 1;
}
else if (diffRsCsdiff >= 30) {
return 8;
}
}
else if (mu_mv_mag_sq >= 92) {
return 1;
}
}
else if (posBalance >= 61) {
return 8;
}
}
}
else if (Cs >= 2824) {
if (Cs < 3701) {
if (ssDCval < 130) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
return 4;
}
}
else if (ssDCval >= 130) {
if (Rs < 4225) {
return 8;
}
else if (Rs >= 4225) {
return 1;
}
}
}
else if (Cs >= 3701) {
if (ssDCval < 79) {
if (Rs < 5477) {
return 8;
}
else if (Rs >= 5477) {
return 2;
}
}
else if (ssDCval >= 79) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (mcTcor < 92) {
if (RsDiff < 16) {
if (Cs < 3594) {
if (ssDCval < 101) {
if (negBalance < 48) {
if (Cs < 1825) {
if (ssDCval < 59) {
if (ssDCval < 42) {
if (MVDiff < 40) {
return 8;
}
else if (MVDiff >= 40) {
return 1;
}
}
else if (ssDCval >= 42) {
if (posBalance < 59) {
if (gchDC < 3) {
if (tcor < 52) {
if (Cs < 1562) {
if (Rs < 1657) {
return 2;
}
else if (Rs >= 1657) {
return 1;
}
}
else if (Cs >= 1562) {
return 4;
}
}
else if (tcor >= 52) {
if (MVDiff < 11) {
if (Rs < 3112) {
if (Rs < 2203) {
return 4;
}
else if (Rs >= 2203) {
return 8;
}
}
else if (Rs >= 3112) {
return 1;
}
}
else if (MVDiff >= 11) {
return 1;
}
}
}
else if (gchDC >= 3) {
if (Rs < 1912) {
return 1;
}
else if (Rs >= 1912) {
return 2;
}
}
}
else if (posBalance >= 59) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 1;
}
}
}
}
else if (ssDCval >= 59) {
if (MVDiff < 30) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (RsCsDiff < 160) {
if (Rs < 3073) {
if (mu_mv_mag_sq < 60) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 2210) {
return 8;
}
else if (Rs >= 2210) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 60) {
return 8;
}
}
else if (Rs >= 3073) {
if (posBalance < 64) {
if (Cs < 1755) {
return 4;
}
else if (Cs >= 1755) {
return 8;
}
}
else if (posBalance >= 64) {
return 2;
}
}
}
else if (RsCsDiff >= 160) {
if (posBalance < 40) {
if (Rs < 3732) {
return 2;
}
else if (Rs >= 3732) {
return 4;
}
}
else if (posBalance >= 40) {
if (ssDCval < 97) {
if (mu_mv_mag_sq < 71) {
return 2;
}
else if (mu_mv_mag_sq >= 71) {
if (Rs < 3816) {
return 4;
}
else if (Rs >= 3816) {
return 2;
}
}
}
else if (ssDCval >= 97) {
return 4;
}
}
}
}
}
else if (MVDiff >= 30) {
if (Rs < 1769) {
if (Cs < 1604) {
return 4;
}
else if (Cs >= 1604) {
return 2;
}
}
else if (Rs >= 1769) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
if (Rs < 2722) {
return 8;
}
else if (Rs >= 2722) {
return 1;
}
}
}
}
}
}
else if (Cs >= 1825) {
if (ssDCval < 78) {
if (posBalance < 56) {
if (AFD < 16) {
if (posBalance < 52) {
if (mu_mv_mag_sq < 93) {
if (SC < 5300) {
if (SC < 4806) {
return 4;
}
else if (SC >= 4806) {
return 1;
}
}
else if (SC >= 5300) {
return 4;
}
}
else if (mu_mv_mag_sq >= 93) {
if (Cs < 2230) {
return 8;
}
else if (Cs >= 2230) {
if (Rs < 3616) {
return 1;
}
else if (Rs >= 3616) {
return 4;
}
}
}
}
else if (posBalance >= 52) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
if (Cs < 2766) {
return 8;
}
else if (Cs >= 2766) {
if (Rs < 3838) {
return 1;
}
else if (Rs >= 3838) {
return 2;
}
}
}
}
}
else if (AFD >= 16) {
if (CsDiff < 15) {
if (Rs < 4297) {
if (MVDiff < 36) {
return 2;
}
else if (MVDiff >= 36) {
return 8;
}
}
else if (Rs >= 4297) {
return 4;
}
}
else if (CsDiff >= 15) {
return 4;
}
}
}
else if (posBalance >= 56) {
if (diffMVdiffVal < -4) {
if (RsCsDiff < 201) {
return 8;
}
else if (RsCsDiff >= 201) {
return 4;
}
}
else if (diffMVdiffVal >= -4) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Rs < 2891) {
return 1;
}
else if (Rs >= 2891) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 78) {
if (Cs < 3013) {
if (Rs < 3699) {
if (Rs < 2887) {
return 8;
}
else if (Rs >= 2887) {
if (mcTcor < 90) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (diffMVdiffVal < -6) {
return 4;
}
else if (diffMVdiffVal >= -6) {
if (mu_mv_mag_sq < 50) {
return 2;
}
else if (mu_mv_mag_sq >= 50) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
return 2;
}
}
}
}
}
else if (mcTcor >= 90) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
if (Rs < 3288) {
return 2;
}
else if (Rs >= 3288) {
return 8;
}
}
}
}
}
else if (Rs >= 3699) {
if (Rs < 5496) {
return 8;
}
else if (Rs >= 5496) {
return 4;
}
}
}
else if (Cs >= 3013) {
if (MVDiff < 19) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 4375) {
return 1;
}
else if (Rs >= 4375) {
return 4;
}
}
}
}
else if (MVDiff >= 19) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Rs < 4975) {
return 8;
}
else if (Rs >= 4975) {
return 1;
}
}
}
}
}
}
}
else if (negBalance >= 48) {
if (mcTcor < 73) {
if (refDCval < 61) {
if (refDCval < 50) {
return 4;
}
else if (refDCval >= 50) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
if (Rs < 2160) {
return 1;
}
else if (Rs >= 2160) {
return 4;
}
}
}
}
else if (refDCval >= 61) {
if (posBalance < 61) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (posBalance >= 61) {
if (Cs < 1982) {
return 1;
}
else if (Cs >= 1982) {
return 4;
}
}
}
}
else if (mcTcor >= 73) {
if (AFD < 20) {
if (TSC < 16) {
if (Cs < 1632) {
if (negBalance < 50) {
if (ssDCval < 76) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 1;
}
}
else if (ssDCval >= 76) {
return 2;
}
}
else if (negBalance >= 50) {
if (diffAFD < -3) {
if (refDCval < 74) {
if (mcTcor < 90) {
return 1;
}
else if (mcTcor >= 90) {
return 8;
}
}
else if (refDCval >= 74) {
return 8;
}
}
else if (diffAFD >= -3) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (Rs < 2623) {
return 8;
}
else if (Rs >= 2623) {
return 4;
}
}
}
}
}
else if (Cs >= 1632) {
if (ssDCval < 82) {
if (tcor < 63) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 1797) {
return 2;
}
else if (Rs >= 1797) {
return 8;
}
}
}
}
else if (tcor >= 63) {
if (Cs < 2168) {
if (mu_mv_mag_sq < 3) {
if (RsDiff < 10) {
if (Rs < 1597) {
return 2;
}
else if (Rs >= 1597) {
if (Rs < 1934) {
if (Rs < 1796) {
return 8;
}
else if (Rs >= 1796) {
return 1;
}
}
else if (Rs >= 1934) {
return 8;
}
}
}
else if (RsDiff >= 10) {
if (ssDCval < 73) {
if (MVDiff < 23) {
return 8;
}
else if (MVDiff >= 23) {
return 4;
}
}
else if (ssDCval >= 73) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 53) {
if (Cs < 2074) {
if (diffMVdiffVal < -2) {
if (mu_mv_mag_sq < 58) {
if (Cs < 1989) {
return 1;
}
else if (Cs >= 1989) {
return 2;
}
}
else if (mu_mv_mag_sq >= 58) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (mu_mv_mag_sq < 74) {
return 8;
}
else if (mu_mv_mag_sq >= 74) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 8;
}
}
}
}
else if (Cs >= 2074) {
if (TSC < 12) {
return 4;
}
else if (TSC >= 12) {
if (Rs < 2359) {
return 1;
}
else if (Rs >= 2359) {
return 8;
}
}
}
}
else if (negBalance >= 53) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (SC < 4289) {
return 4;
}
else if (SC >= 4289) {
return 8;
}
}
}
}
}
else if (Cs >= 2168) {
if (Rs < 3992) {
if (posBalance < 58) {
if (Cs < 2937) {
if (tcor < 79) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (diffAFD < -2) {
if (Rs < 3016) {
return 1;
}
else if (Rs >= 3016) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (tcor >= 79) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
if (gchDC < 4) {
if (Cs < 2196) {
if (Rs < 2803) {
return 4;
}
else if (Rs >= 2803) {
return 8;
}
}
else if (Cs >= 2196) {
return 8;
}
}
else if (gchDC >= 4) {
return 4;
}
}
}
}
else if (Cs >= 2937) {
if (Cs < 3112) {
return 1;
}
else if (Cs >= 3112) {
return 4;
}
}
}
else if (posBalance >= 58) {
if (negBalance < 51) {
if (MVDiff < 19) {
if (Rs < 2407) {
return 1;
}
else if (Rs >= 2407) {
if (negBalance < 49) {
if (Rs < 3494) {
return 8;
}
else if (Rs >= 3494) {
return 1;
}
}
else if (negBalance >= 49) {
if (diffMVdiffVal < -5) {
if (Rs < 3309) {
return 4;
}
else if (Rs >= 3309) {
return 8;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
}
}
else if (MVDiff >= 19) {
if (Rs < 3242) {
if (Cs < 2195) {
return 2;
}
else if (Cs >= 2195) {
return 8;
}
}
else if (Rs >= 3242) {
return 4;
}
}
}
else if (negBalance >= 51) {
if (Rs < 3516) {
return 8;
}
else if (Rs >= 3516) {
if (Cs < 2785) {
return 1;
}
else if (Cs >= 2785) {
return 8;
}
}
}
}
}
else if (Rs >= 3992) {
if (Cs < 3164) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 2;
}
}
else if (Cs >= 3164) {
return 8;
}
}
}
}
}
else if (ssDCval >= 82) {
if (SC < 7004) {
if (Cs < 2702) {
if (RsDiff < 11) {
if (TSC < 14) {
if (ssDCval < 89) {
if (TSC < 11) {
if (Cs < 2099) {
return 4;
}
else if (Cs >= 2099) {
return 8;
}
}
else if (TSC >= 11) {
if (Rs < 2976) {
return 8;
}
else if (Rs >= 2976) {
return 2;
}
}
}
else if (ssDCval >= 89) {
if (diffMVdiffVal < -9) {
return 2;
}
else if (diffMVdiffVal >= -9) {
if (diffRsCsdiff < -48) {
if (Rs < 3018) {
return 2;
}
else if (Rs >= 3018) {
return 8;
}
}
else if (diffRsCsdiff >= -48) {
return 8;
}
}
}
}
else if (TSC >= 14) {
return 1;
}
}
else if (RsDiff >= 11) {
if (refDCval < 88) {
if (mu_mv_mag_sq < 25) {
return 4;
}
else if (mu_mv_mag_sq >= 25) {
if (Rs < 3315) {
return 8;
}
else if (Rs >= 3315) {
return 4;
}
}
}
else if (refDCval >= 88) {
if (SC < 6540) {
if (RsCsDiff < 277) {
if (MVDiff < 17) {
if (tcor < 85) {
if (diffAFD < 7) {
return 2;
}
else if (diffAFD >= 7) {
return 8;
}
}
else if (tcor >= 85) {
return 8;
}
}
else if (MVDiff >= 17) {
return 8;
}
}
else if (RsCsDiff >= 277) {
if (tcor < 78) {
return 8;
}
else if (tcor >= 78) {
return 4;
}
}
}
else if (SC >= 6540) {
return 4;
}
}
}
}
else if (Cs >= 2702) {
if (refDCval < 91) {
if (refDCval < 88) {
if (Cs < 3298) {
return 4;
}
else if (Cs >= 3298) {
return 2;
}
}
else if (refDCval >= 88) {
return 8;
}
}
else if (refDCval >= 91) {
if (Rs < 2237) {
return 8;
}
else if (Rs >= 2237) {
return 4;
}
}
}
}
else if (SC >= 7004) {
if (negBalance < 50) {
if (Cs < 3061) {
return 8;
}
else if (Cs >= 3061) {
return 4;
}
}
else if (negBalance >= 50) {
if (mu_mv_mag_sq < 3) {
if (Rs < 5006) {
return 4;
}
else if (Rs >= 5006) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
}
}
}
else if (TSC >= 16) {
if (Rs < 3299) {
return 4;
}
else if (Rs >= 3299) {
return 8;
}
}
}
else if (AFD >= 20) {
if (refDCval < 76) {
if (Rs < 2730) {
if (SC < 3996) {
if (mcTcor < 81) {
if (Cs < 1543) {
return 8;
}
else if (Cs >= 1543) {
return 1;
}
}
else if (mcTcor >= 81) {
if (RsCsDiff < 154) {
if (Rs < 1549) {
if (Rs < 1460) {
return 1;
}
else if (Rs >= 1460) {
return 8;
}
}
else if (Rs >= 1549) {
return 2;
}
}
else if (RsCsDiff >= 154) {
if (AFD < 28) {
if (diffMVdiffVal < -6) {
if (Rs < 1439) {
return 2;
}
else if (Rs >= 1439) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
if (TSC < 11) {
if (Cs < 1424) {
return 4;
}
else if (Cs >= 1424) {
return 8;
}
}
else if (TSC >= 11) {
if (Cs < 1359) {
return 8;
}
else if (Cs >= 1359) {
return 4;
}
}
}
}
else if (AFD >= 28) {
return 2;
}
}
}
}
else if (SC >= 3996) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
if (RsCsDiff < 393) {
if (MVDiff < 30) {
return 2;
}
else if (MVDiff >= 30) {
return 4;
}
}
else if (RsCsDiff >= 393) {
return 8;
}
}
}
}
else if (Rs >= 2730) {
if (negBalance < 57) {
if (MVDiff < 15) {
if (Rs < 2940) {
return 1;
}
else if (Rs >= 2940) {
return 2;
}
}
else if (MVDiff >= 15) {
return 2;
}
}
else if (negBalance >= 57) {
if (Rs < 2834) {
return 4;
}
else if (Rs >= 2834) {
return 8;
}
}
}
}
else if (refDCval >= 76) {
if (Rs < 1626) {
if (CsDiff < 14) {
if (Cs < 1335) {
return 8;
}
else if (Cs >= 1335) {
return 1;
}
}
else if (CsDiff >= 14) {
if (CsDiff < 19) {
return 2;
}
else if (CsDiff >= 19) {
return 8;
}
}
}
else if (Rs >= 1626) {
if (mu_mv_mag_sq < 2) {
if (tcor < 77) {
if (ssDCval < 79) {
if (Cs < 1640) {
return 2;
}
else if (Cs >= 1640) {
return 8;
}
}
else if (ssDCval >= 79) {
if (SC < 3711) {
return 8;
}
else if (SC >= 3711) {
if (negBalance < 52) {
if (Rs < 1789) {
return 4;
}
else if (Rs >= 1789) {
return 8;
}
}
else if (negBalance >= 52) {
if (Rs < 4753) {
return 4;
}
else if (Rs >= 4753) {
return 8;
}
}
}
}
}
else if (tcor >= 77) {
if (TSC < 15) {
return 8;
}
else if (TSC >= 15) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (TSC < 13) {
if (diffRsCsdiff < 43) {
if (posBalance < 57) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
return 8;
}
}
else if (posBalance >= 57) {
if (diffRsCsdiff < 27) {
if (diffMVdiffVal < -3) {
if (Rs < 1672) {
return 4;
}
else if (Rs >= 1672) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (tcor < 65) {
if (Rs < 1768) {
return 4;
}
else if (Rs >= 1768) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
else if (tcor >= 65) {
if (Cs < 2773) {
if (CsDiff < 12) {
return 4;
}
else if (CsDiff >= 12) {
return 8;
}
}
else if (Cs >= 2773) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 27) {
return 4;
}
}
}
else if (diffRsCsdiff >= 43) {
if (diffAFD < -2) {
if (Rs < 1987) {
return 1;
}
else if (Rs >= 1987) {
return 8;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (TSC >= 13) {
if (refDCval < 92) {
if (Cs < 2513) {
if (tcor < 58) {
return 1;
}
else if (tcor >= 58) {
if (diffMVdiffVal < -3) {
if (Rs < 2647) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 1;
}
}
else if (Rs >= 2647) {
return 4;
}
}
else if (diffMVdiffVal >= -3) {
if (gchDC < 1) {
if (MVDiff < 22) {
return 8;
}
else if (MVDiff >= 22) {
return 4;
}
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
else if (Cs >= 2513) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 1;
}
}
}
else if (refDCval >= 92) {
if (tcor < 78) {
if (posBalance < 58) {
if (Rs < 2376) {
return 4;
}
else if (Rs >= 2376) {
return 8;
}
}
else if (posBalance >= 58) {
if (mu_mv_mag_sq < 13) {
return 1;
}
else if (mu_mv_mag_sq >= 13) {
return 4;
}
}
}
else if (tcor >= 78) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 101) {
if (mu_mv_mag_sq < 85) {
if (negBalance < 31) {
if (TSC < 10) {
if (MVDiff < 22) {
if (mu_mv_mag_sq < 78) {
return 4;
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 3447) {
return 2;
}
else if (Rs >= 3447) {
return 8;
}
}
}
else if (MVDiff >= 22) {
if (Rs < 1920) {
return 4;
}
else if (Rs >= 1920) {
return 2;
}
}
}
else if (TSC >= 10) {
if (mu_mv_mag_sq < 58) {
return 8;
}
else if (mu_mv_mag_sq >= 58) {
return 4;
}
}
}
else if (negBalance >= 31) {
if (negBalance < 54) {
if (mcTcor < 87) {
if (refDCval < 125) {
if (Cs < 1404) {
if (Rs < 2417) {
return 4;
}
else if (Rs >= 2417) {
return 8;
}
}
else if (Cs >= 1404) {
if (negBalance < 53) {
return 8;
}
else if (negBalance >= 53) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 4;
}
}
}
}
else if (refDCval >= 125) {
if (SC < 7183) {
if (mcTcor < 70) {
if (Rs < 2279) {
if (Rs < 1721) {
return 8;
}
else if (Rs >= 1721) {
return 1;
}
}
else if (Rs >= 2279) {
return 2;
}
}
else if (mcTcor >= 70) {
if (Cs < 1780) {
if (Rs < 3287) {
if (Cs < 1770) {
return 8;
}
else if (Cs >= 1770) {
return 1;
}
}
else if (Rs >= 3287) {
return 4;
}
}
else if (Cs >= 1780) {
return 8;
}
}
}
else if (SC >= 7183) {
if (refDCval < 136) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 4;
}
}
else if (refDCval >= 136) {
if (negBalance < 43) {
return 8;
}
else if (negBalance >= 43) {
return 1;
}
}
}
}
}
else if (mcTcor >= 87) {
if (RsCsDiff < 220) {
if (Cs < 2589) {
if (Cs < 1415) {
if (mu_mv_mag_sq < 16) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 16) {
if (Cs < 1398) {
return 8;
}
else if (Cs >= 1398) {
return 2;
}
}
}
else if (Cs >= 1415) {
if (SC < 5599) {
if (ssDCval < 109) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (SC < 5109) {
if (ssDCval < 104) {
return 8;
}
else if (ssDCval >= 104) {
return 4;
}
}
else if (SC >= 5109) {
return 8;
}
}
}
else if (ssDCval >= 109) {
if (posBalance < 50) {
if (Rs < 2812) {
return 8;
}
else if (Rs >= 2812) {
return 4;
}
}
else if (posBalance >= 50) {
if (SC < 3129) {
return 4;
}
else if (SC >= 3129) {
return 8;
}
}
}
}
else if (SC >= 5599) {
if (mu_mv_mag_sq < 75) {
if (Rs < 3263) {
return 1;
}
else if (Rs >= 3263) {
return 8;
}
}
else if (mu_mv_mag_sq >= 75) {
if (Rs < 4090) {
return 8;
}
else if (Rs >= 4090) {
return 1;
}
}
}
}
}
else if (Cs >= 2589) {
if (MVDiff < 10) {
if (ssDCval < 105) {
if (Rs < 3822) {
return 2;
}
else if (Rs >= 3822) {
return 8;
}
}
else if (ssDCval >= 105) {
if (ssDCval < 134) {
if (Rs < 4349) {
return 4;
}
else if (Rs >= 4349) {
return 8;
}
}
else if (ssDCval >= 134) {
return 8;
}
}
}
else if (MVDiff >= 10) {
if (diffAFD < -10) {
return 4;
}
else if (diffAFD >= -10) {
return 8;
}
}
}
}
else if (RsCsDiff >= 220) {
if (refDCval < 129) {
if (posBalance < 67) {
if (mcTcor < 91) {
if (MVDiff < 7) {
if (Rs < 3054) {
return 8;
}
else if (Rs >= 3054) {
return 4;
}
}
else if (MVDiff >= 7) {
if (posBalance < 63) {
return 8;
}
else if (posBalance >= 63) {
return 2;
}
}
}
else if (mcTcor >= 91) {
if (Rs < 3084) {
return 8;
}
else if (Rs >= 3084) {
return 4;
}
}
}
else if (posBalance >= 67) {
if (Cs < 2099) {
return 8;
}
else if (Cs >= 2099) {
return 2;
}
}
}
else if (refDCval >= 129) {
if (negBalance < 39) {
if (Rs < 2458) {
return 8;
}
else if (Rs >= 2458) {
return 2;
}
}
else if (negBalance >= 39) {
if (diffAFD < 2) {
if (Rs < 2328) {
return 8;
}
else if (Rs >= 2328) {
return 4;
}
}
else if (diffAFD >= 2) {
if (Cs < 2613) {
return 8;
}
else if (Cs >= 2613) {
return 4;
}
}
}
}
}
}
}
else if (negBalance >= 54) {
if (refDCval < 108) {
if (Rs < 2487) {
if (tcor < 63) {
if (AFD < 23) {
return 4;
}
else if (AFD >= 23) {
return 2;
}
}
else if (tcor >= 63) {
if (Cs < 1502) {
return 4;
}
else if (Cs >= 1502) {
return 8;
}
}
}
else if (Rs >= 2487) {
if (TSC < 14) {
if (Cs < 1456) {
return 8;
}
else if (Cs >= 1456) {
if (gchDC < 3) {
if (Rs < 3250) {
if (mcTcor < 84) {
if (Rs < 2878) {
return 4;
}
else if (Rs >= 2878) {
return 8;
}
}
else if (mcTcor >= 84) {
return 4;
}
}
else if (Rs >= 3250) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 8;
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (TSC >= 14) {
if (Rs < 3577) {
return 4;
}
else if (Rs >= 3577) {
return 1;
}
}
}
}
else if (refDCval >= 108) {
if (mu_mv_mag_sq < 13) {
if (SC < 6283) {
if (tcor < 60) {
if (SC < 4237) {
if (CsDiff < 11) {
if (Rs < 1551) {
return 8;
}
else if (Rs >= 1551) {
return 4;
}
}
else if (CsDiff >= 11) {
if (Rs < 978) {
return 2;
}
else if (Rs >= 978) {
return 8;
}
}
}
else if (SC >= 4237) {
if (Rs < 2731) {
return 2;
}
else if (Rs >= 2731) {
return 1;
}
}
}
else if (tcor >= 60) {
if (MVDiff < 27) {
if (diffMVdiffVal < -4) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
if (Rs < 2468) {
if (Cs < 2217) {
if (Rs < 2268) {
return 1;
}
else if (Rs >= 2268) {
return 8;
}
}
else if (Cs >= 2217) {
return 2;
}
}
else if (Rs >= 2468) {
return 1;
}
}
}
else if (diffMVdiffVal >= -4) {
if (negBalance < 59) {
if (diffAFD < 12) {
return 8;
}
else if (diffAFD >= 12) {
return 1;
}
}
else if (negBalance >= 59) {
if (diffMVdiffVal < 6) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
return 8;
}
}
else if (diffMVdiffVal >= 6) {
if (Cs < 2898) {
return 8;
}
else if (Cs >= 2898) {
return 2;
}
}
}
}
}
else if (MVDiff >= 27) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
return 1;
}
}
}
}
else if (SC >= 6283) {
if (Rs < 5305) {
return 4;
}
else if (Rs >= 5305) {
if (Rs < 5686) {
return 1;
}
else if (Rs >= 5686) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (Cs < 2415) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (Rs < 2233) {
return 8;
}
else if (Rs >= 2233) {
if (Cs < 1613) {
if (Rs < 2320) {
return 2;
}
else if (Rs >= 2320) {
return 4;
}
}
else if (Cs >= 1613) {
if (Rs < 3311) {
if (Cs < 2171) {
if (refDCval < 124) {
return 8;
}
else if (refDCval >= 124) {
if (Rs < 2586) {
return 1;
}
else if (Rs >= 2586) {
return 2;
}
}
}
else if (Cs >= 2171) {
return 8;
}
}
else if (Rs >= 3311) {
return 1;
}
}
}
}
}
else if (Cs >= 2415) {
if (diffAFD < 1) {
if (AFD < 21) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 8;
}
}
else if (AFD >= 21) {
if (Rs < 2597) {
return 1;
}
else if (Rs >= 2597) {
return 4;
}
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (negBalance < 44) {
if (MVDiff < 21) {
return 8;
}
else if (MVDiff >= 21) {
if (Rs < 2099) {
return 1;
}
else if (Rs >= 2099) {
return 2;
}
}
}
else if (negBalance >= 44) {
if (Rs < 3713) {
if (Cs < 2616) {
return 8;
}
else if (Cs >= 2616) {
if (Rs < 3648) {
return 4;
}
else if (Rs >= 3648) {
return 2;
}
}
}
else if (Rs >= 3713) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
}
}
}
}
else if (Cs >= 3594) {
if (AFD < 10) {
if (mu_mv_mag_sq < 101) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (mu_mv_mag_sq >= 101) {
return 8;
}
}
else if (AFD >= 10) {
if (refDCval < 99) {
if (Rs < 6056) {
if (Rs < 3465) {
if (mcTcor < 82) {
if (AFD < 20) {
return 4;
}
else if (AFD >= 20) {
return 8;
}
}
else if (mcTcor >= 82) {
if (Cs < 5125) {
return 8;
}
else if (Cs >= 5125) {
return 4;
}
}
}
else if (Rs >= 3465) {
if (negBalance < 48) {
if (Rs < 4392) {
if (Cs < 3926) {
return 4;
}
else if (Cs >= 3926) {
return 2;
}
}
else if (Rs >= 4392) {
return 8;
}
}
else if (negBalance >= 48) {
if (TSC < 12) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (Rs < 4086) {
return 2;
}
else if (Rs >= 4086) {
return 8;
}
}
}
else if (TSC >= 12) {
return 8;
}
}
}
}
else if (Rs >= 6056) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 4;
}
}
}
else if (refDCval >= 99) {
if (Rs < 4547) {
if (negBalance < 48) {
if (tcor < 71) {
if (ssDCval < 134) {
if (Cs < 4108) {
return 1;
}
else if (Cs >= 4108) {
return 8;
}
}
else if (ssDCval >= 134) {
if (mcTcor < 76) {
if (MVDiff < 17) {
return 1;
}
else if (MVDiff >= 17) {
return 8;
}
}
else if (mcTcor >= 76) {
return 8;
}
}
}
else if (tcor >= 71) {
if (Rs < 2772) {
return 4;
}
else if (Rs >= 2772) {
return 8;
}
}
}
else if (negBalance >= 48) {
if (mcTcor < 89) {
if (Cs < 4171) {
return 4;
}
else if (Cs >= 4171) {
if (Rs < 2183) {
return 4;
}
else if (Rs >= 2183) {
return 2;
}
}
}
else if (mcTcor >= 89) {
return 1;
}
}
}
else if (Rs >= 4547) {
if (refDCval < 100) {
if (Rs < 5386) {
return 8;
}
else if (Rs >= 5386) {
return 1;
}
}
else if (refDCval >= 100) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 16) {
if (Cs < 4680) {
if (refDCval < 134) {
if (Cs < 1862) {
if (diffAFD < 13) {
if (MVDiff < 33) {
if (ssDCval < 86) {
if (ssDCval < 61) {
if (TSC < 10) {
if (TSC < 9) {
return 4;
}
else if (TSC >= 9) {
return 1;
}
}
else if (TSC >= 10) {
if (Cs < 1744) {
return 2;
}
else if (Cs >= 1744) {
return 4;
}
}
}
else if (ssDCval >= 61) {
return 2;
}
}
else if (ssDCval >= 86) {
if (AFD < 23) {
if (tcor < 78) {
if (Cs < 1317) {
return 1;
}
else if (Cs >= 1317) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (tcor >= 78) {
return 2;
}
}
else if (AFD >= 23) {
if (refDCval < 117) {
if (negBalance < 56) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
return 4;
}
}
else if (negBalance >= 56) {
return 2;
}
}
else if (refDCval >= 117) {
if (Cs < 1710) {
return 2;
}
else if (Cs >= 1710) {
return 8;
}
}
}
}
}
else if (MVDiff >= 33) {
if (tcor < 69) {
if (AFD < 28) {
if (mu_mv_mag_sq < 21) {
if (tcor < 57) {
return 8;
}
else if (tcor >= 57) {
return 2;
}
}
else if (mu_mv_mag_sq >= 21) {
return 4;
}
}
else if (AFD >= 28) {
if (diffMVdiffVal < -6) {
if (Rs < 3337) {
if (Rs < 2719) {
return 2;
}
else if (Rs >= 2719) {
return 1;
}
}
else if (Rs >= 3337) {
return 4;
}
}
else if (diffMVdiffVal >= -6) {
return 2;
}
}
}
else if (tcor >= 69) {
if (Cs < 1436) {
return 1;
}
else if (Cs >= 1436) {
if (RsCsDiff < 349) {
if (Rs < 3955) {
return 8;
}
else if (Rs >= 3955) {
return 1;
}
}
else if (RsCsDiff >= 349) {
return 8;
}
}
}
}
}
else if (diffAFD >= 13) {
if (diffMVdiffVal < 33) {
if (Rs < 7043) {
return 1;
}
else if (Rs >= 7043) {
return 2;
}
}
else if (diffMVdiffVal >= 33) {
return 8;
}
}
}
else if (Cs >= 1862) {
if (TSC < 15) {
if (SC < 8446) {
if (MVDiff < 13) {
if (negBalance < 46) {
return 2;
}
else if (negBalance >= 46) {
if (AFD < 20) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (Rs < 4808) {
if (RsCsDiff < 401) {
return 1;
}
else if (RsCsDiff >= 401) {
return 8;
}
}
else if (Rs >= 4808) {
return 2;
}
}
}
else if (AFD >= 20) {
if (Cs < 2310) {
if (ssDCval < 108) {
if (Rs < 3924) {
return 4;
}
else if (Rs >= 3924) {
return 2;
}
}
else if (ssDCval >= 108) {
if (Rs < 4479) {
return 2;
}
else if (Rs >= 4479) {
return 4;
}
}
}
else if (Cs >= 2310) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 2;
}
}
}
}
}
else if (MVDiff >= 13) {
if (ssDCval < 98) {
if (diffAFD < 3) {
if (RsCsDiff < 880) {
if (MVDiff < 25) {
if (ssDCval < 89) {
if (mu_mv_mag_sq < 7) {
if (mcTcor < 81) {
return 4;
}
else if (mcTcor >= 81) {
if (ssDCval < 78) {
return 2;
}
else if (ssDCval >= 78) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (tcor < 79) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (tcor >= 79) {
if (Rs < 5217) {
if (Rs < 4086) {
return 8;
}
else if (Rs >= 4086) {
return 2;
}
}
else if (Rs >= 5217) {
return 4;
}
}
}
}
else if (ssDCval >= 89) {
return 4;
}
}
else if (MVDiff >= 25) {
if (mcTcor < 90) {
if (AFD < 19) {
if (Rs < 3814) {
return 8;
}
else if (Rs >= 3814) {
if (Rs < 4584) {
return 4;
}
else if (Rs >= 4584) {
return 1;
}
}
}
else if (AFD >= 19) {
if (RsDiff < 21) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
if (Rs < 3706) {
return 8;
}
else if (Rs >= 3706) {
return 4;
}
}
}
else if (RsDiff >= 21) {
return 8;
}
}
}
else if (mcTcor >= 90) {
return 8;
}
}
}
else if (RsCsDiff >= 880) {
if (negBalance < 56) {
return 1;
}
else if (negBalance >= 56) {
return 2;
}
}
}
else if (diffAFD >= 3) {
if (mu_mv_mag_sq < 26) {
return 2;
}
else if (mu_mv_mag_sq >= 26) {
if (Rs < 3456) {
if (Rs < 3389) {
return 8;
}
else if (Rs >= 3389) {
if (Cs < 2115) {
return 8;
}
else if (Cs >= 2115) {
return 4;
}
}
}
else if (Rs >= 3456) {
return 8;
}
}
}
}
else if (ssDCval >= 98) {
if (ssDCval < 122) {
if (Cs < 3267) {
return 8;
}
else if (Cs >= 3267) {
return 1;
}
}
else if (ssDCval >= 122) {
if (Rs < 4283) {
return 2;
}
else if (Rs >= 4283) {
return 4;
}
}
}
}
}
else if (SC >= 8446) {
if (AFD < 21) {
if (MVDiff < 13) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 2;
}
}
else if (MVDiff >= 13) {
if (refDCval < 73) {
if (diffRsCsdiff < 102) {
return 4;
}
else if (diffRsCsdiff >= 102) {
return 8;
}
}
else if (refDCval >= 73) {
if (Rs < 4900) {
return 4;
}
else if (Rs >= 4900) {
return 8;
}
}
}
}
else if (AFD >= 21) {
if (mu_mv_mag_sq < 7) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (refDCval < 75) {
return 8;
}
else if (refDCval >= 75) {
if (MVDiff < 61) {
return 4;
}
else if (MVDiff >= 61) {
return 2;
}
}
}
}
}
}
else if (TSC >= 15) {
if (diffTSC < 2) {
if (negBalance < 49) {
if (refDCval < 82) {
if (Rs < 7262) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
return 4;
}
}
else if (Rs >= 7262) {
return 1;
}
}
else if (refDCval >= 82) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (RsCsDiff < 740) {
return 1;
}
else if (RsCsDiff >= 740) {
if (Cs < 3429) {
return 8;
}
else if (Cs >= 3429) {
return 2;
}
}
}
}
}
else if (negBalance >= 49) {
if (tcor < 65) {
if (Cs < 3552) {
if (Rs < 4136) {
if (mcTcor < 77) {
return 8;
}
else if (mcTcor >= 77) {
if (Cs < 2082) {
return 2;
}
else if (Cs >= 2082) {
return 1;
}
}
}
else if (Rs >= 4136) {
if (AFD < 36) {
return 2;
}
else if (AFD >= 36) {
return 1;
}
}
}
else if (Cs >= 3552) {
if (Rs < 6344) {
return 8;
}
else if (Rs >= 6344) {
return 4;
}
}
}
else if (tcor >= 65) {
if (RsCsDiff < 693) {
if (Rs < 5919) {
if (refDCval < 131) {
return 8;
}
else if (refDCval >= 131) {
return 2;
}
}
else if (Rs >= 5919) {
return 4;
}
}
else if (RsCsDiff >= 693) {
if (Cs < 3494) {
return 4;
}
else if (Cs >= 3494) {
return 2;
}
}
}
}
}
else if (diffTSC >= 2) {
if (RsCsDiff < 521) {
if (Rs < 2247) {
return 2;
}
else if (Rs >= 2247) {
return 8;
}
}
else if (RsCsDiff >= 521) {
if (mcTcor < 74) {
if (diffMVdiffVal < 47) {
return 8;
}
else if (diffMVdiffVal >= 47) {
return 4;
}
}
else if (mcTcor >= 74) {
if (diffMVdiffVal < 11) {
if (posBalance < 58) {
if (Rs < 6215) {
return 2;
}
else if (Rs >= 6215) {
return 1;
}
}
else if (posBalance >= 58) {
if (Cs < 2702) {
return 1;
}
else if (Cs >= 2702) {
return 4;
}
}
}
else if (diffMVdiffVal >= 11) {
if (AFD < 30) {
return 8;
}
else if (AFD >= 30) {
if (diffAFD < 17) {
return 4;
}
else if (diffAFD >= 17) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 134) {
if (posBalance < 57) {
if (Rs < 5683) {
if (negBalance < 64) {
if (MVDiff < 17) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (diffMVdiffVal < -2) {
if (Cs < 1972) {
return 4;
}
else if (Cs >= 1972) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
}
else if (MVDiff >= 17) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
if (Cs < 1911) {
if (Rs < 3739) {
return 2;
}
else if (Rs >= 3739) {
return 4;
}
}
else if (Cs >= 1911) {
return 8;
}
}
}
}
else if (negBalance >= 64) {
return 2;
}
}
else if (Rs >= 5683) {
if (ssDCval < 165) {
if (tcor < 79) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
else if (tcor >= 79) {
if (Rs < 6214) {
return 1;
}
else if (Rs >= 6214) {
return 2;
}
}
}
else if (ssDCval >= 165) {
if (Rs < 6316) {
return 8;
}
else if (Rs >= 6316) {
return 2;
}
}
}
}
else if (posBalance >= 57) {
if (Rs < 3795) {
if (Cs < 1695) {
return 2;
}
else if (Cs >= 1695) {
return 8;
}
}
else if (Rs >= 3795) {
if (diffAFD < 3) {
if (diffRsCsdiff < 507) {
return 4;
}
else if (diffRsCsdiff >= 507) {
return 2;
}
}
else if (diffAFD >= 3) {
if (Rs < 5737) {
return 8;
}
else if (Rs >= 5737) {
return 2;
}
}
}
}
}
}
else if (Cs >= 4680) {
if (mu_mv_mag_sq < 15) {
if (negBalance < 60) {
if (TSC < 12) {
if (AFD < 28) {
return 8;
}
else if (AFD >= 28) {
return 4;
}
}
else if (TSC >= 12) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
return 4;
}
}
}
else if (negBalance >= 60) {
if (AFD < 26) {
return 4;
}
else if (AFD >= 26) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (posBalance < 60) {
if (posBalance < 41) {
if (RsDiff < 17) {
return 1;
}
else if (RsDiff >= 17) {
return 8;
}
}
else if (posBalance >= 41) {
if (negBalance < 34) {
return 8;
}
else if (negBalance >= 34) {
if (RsCsDiff < 615) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (RsCsDiff >= 615) {
return 4;
}
}
}
}
else if (posBalance >= 60) {
if (Rs < 8730) {
return 8;
}
else if (Rs >= 8730) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 101) {
if (CsDiff < 8) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Cs < 2160) {
if (Rs < 3231) {
if (TSC < 9) {
if (SC < 3653) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (AFD < 19) {
return 2;
}
else if (AFD >= 19) {
return 8;
}
}
}
else if (SC >= 3653) {
if (SC < 4989) {
if (mcTcor < 94) {
if (RsDiff < 10) {
if (SC < 4474) {
return 4;
}
else if (SC >= 4474) {
return 2;
}
}
else if (RsDiff >= 10) {
if (Rs < 2656) {
return 4;
}
else if (Rs >= 2656) {
return 2;
}
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (SC >= 4989) {
return 1;
}
}
}
else if (TSC >= 9) {
if (Rs < 2779) {
if (TSC < 12) {
if (ssDCval < 90) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (RsCsDiff < 143) {
return 8;
}
else if (RsCsDiff >= 143) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
if (Rs < 2150) {
return 2;
}
else if (Rs >= 2150) {
return 8;
}
}
}
}
}
else if (ssDCval >= 90) {
return 8;
}
}
else if (TSC >= 12) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
return 8;
}
}
}
else if (Rs >= 2779) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
}
}
else if (Rs >= 3231) {
if (posBalance < 46) {
if (tcor < 89) {
if (TSC < 8) {
if (RsDiff < 10) {
if (Rs < 3529) {
return 4;
}
else if (Rs >= 3529) {
return 2;
}
}
else if (RsDiff >= 10) {
return 4;
}
}
else if (TSC >= 8) {
if (RsCsDiff < 164) {
if (ssDCval < 66) {
return 2;
}
else if (ssDCval >= 66) {
return 4;
}
}
else if (RsCsDiff >= 164) {
return 2;
}
}
}
else if (tcor >= 89) {
if (ssDCval < 61) {
return 4;
}
else if (ssDCval >= 61) {
if (diffMVdiffVal < -11) {
return 8;
}
else if (diffMVdiffVal >= -11) {
return 2;
}
}
}
}
else if (posBalance >= 46) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
if (MVDiff < 8) {
if (ssDCval < 70) {
return 4;
}
else if (ssDCval >= 70) {
if (TSC < 6) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
return 2;
}
}
else if (TSC >= 6) {
return 2;
}
}
}
else if (MVDiff >= 8) {
if (Cs < 1936) {
if (diffMVdiffVal < -14) {
return 8;
}
else if (diffMVdiffVal >= -14) {
if (CsDiff < 4) {
if (Rs < 7893) {
return 4;
}
else if (Rs >= 7893) {
return 2;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (Cs >= 1936) {
if (tcor < 91) {
if (Rs < 3915) {
return 4;
}
else if (Rs >= 3915) {
return 8;
}
}
else if (tcor >= 91) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 2160) {
if (mu_mv_mag_sq < 55) {
if (posBalance < 46) {
if (Rs < 4662) {
return 2;
}
else if (Rs >= 4662) {
return 4;
}
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 6) {
if (Rs < 4253) {
return 8;
}
else if (Rs >= 4253) {
return 2;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 55) {
if (Rs < 4331) {
if (TSC < 11) {
if (posBalance < 46) {
return 1;
}
else if (posBalance >= 46) {
return 4;
}
}
else if (TSC >= 11) {
return 8;
}
}
else if (Rs >= 4331) {
if (Rs < 7090) {
return 2;
}
else if (Rs >= 7090) {
return 4;
}
}
}
}
}
}
else if (CsDiff >= 8) {
if (negBalance < 46) {
if (ssDCval < 72) {
if (Rs < 3620) {
if (negBalance < 37) {
return 1;
}
else if (negBalance >= 37) {
if (Cs < 1686) {
return 8;
}
else if (Cs >= 1686) {
return 2;
}
}
}
else if (Rs >= 3620) {
if (SC < 8273) {
if (Rs < 4202) {
return 8;
}
else if (Rs >= 4202) {
return 4;
}
}
else if (SC >= 8273) {
return 8;
}
}
}
else if (ssDCval >= 72) {
if (Rs < 2629) {
if (AFD < 18) {
if (posBalance < 71) {
return 8;
}
else if (posBalance >= 71) {
return 2;
}
}
else if (AFD >= 18) {
if (Cs < 1977) {
return 2;
}
else if (Cs >= 1977) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Rs < 2210) {
return 4;
}
else if (Rs >= 2210) {
return 1;
}
}
}
}
}
else if (Rs >= 2629) {
if (gchDC < 3) {
if (Cs < 2191) {
if (negBalance < 35) {
if (Rs < 4646) {
return 1;
}
else if (Rs >= 4646) {
return 2;
}
}
else if (negBalance >= 35) {
return 2;
}
}
else if (Cs >= 2191) {
if (negBalance < 34) {
return 2;
}
else if (negBalance >= 34) {
if (mu_mv_mag_sq < 64) {
if (RsCsDiff < 194) {
if (Rs < 3547) {
return 4;
}
else if (Rs >= 3547) {
return 8;
}
}
else if (RsCsDiff >= 194) {
if (Rs < 2917) {
return 8;
}
else if (Rs >= 2917) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 64) {
if (Cs < 3358) {
if (diffTSC < 2) {
return 4;
}
else if (diffTSC >= 2) {
return 2;
}
}
else if (Cs >= 3358) {
return 8;
}
}
}
}
}
else if (gchDC >= 3) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 4519) {
return 4;
}
else if (Rs >= 4519) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (mu_mv_mag_sq < 9) {
if (mcTcor < 94) {
if (Rs < 3716) {
if (MVDiff < 10) {
if (Rs < 2502) {
if (Cs < 1736) {
if (SC < 3466) {
if (Rs < 1854) {
return 2;
}
else if (Rs >= 1854) {
return 8;
}
}
else if (SC >= 3466) {
return 4;
}
}
else if (Cs >= 1736) {
if (Cs < 2435) {
return 8;
}
else if (Cs >= 2435) {
return 1;
}
}
}
else if (Rs >= 2502) {
if (TSC < 12) {
if (refDCval < 81) {
if (Rs < 3215) {
return 4;
}
else if (Rs >= 3215) {
return 2;
}
}
else if (refDCval >= 81) {
return 4;
}
}
else if (TSC >= 12) {
if (Rs < 3357) {
return 8;
}
else if (Rs >= 3357) {
return 4;
}
}
}
}
else if (MVDiff >= 10) {
if (Cs < 1871) {
if (ssDCval < 91) {
if (RsDiff < 12) {
return 4;
}
else if (RsDiff >= 12) {
if (Cs < 1825) {
return 2;
}
else if (Cs >= 1825) {
return 4;
}
}
}
else if (ssDCval >= 91) {
return 8;
}
}
else if (Cs >= 1871) {
if (gchDC < 2) {
if (tcor < 87) {
return 8;
}
else if (tcor >= 87) {
return 2;
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
}
else if (Rs >= 3716) {
if (ssDCval < 93) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
if (Rs < 4836) {
return 4;
}
else if (Rs >= 4836) {
return 2;
}
}
}
else if (ssDCval >= 93) {
if (Cs < 1918) {
return 2;
}
else if (Cs >= 1918) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (AFD < 17) {
if (SC < 4512) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
else if (SC >= 4512) {
if (Cs < 2296) {
if (Cs < 1444) {
return 8;
}
else if (Cs >= 1444) {
return 2;
}
}
else if (Cs >= 2296) {
if (RsDiff < 10) {
if (Rs < 3206) {
return 8;
}
else if (Rs >= 3206) {
return 2;
}
}
else if (RsDiff >= 10) {
return 4;
}
}
}
}
else if (AFD >= 17) {
if (tcor < 81) {
if (Cs < 1744) {
return 2;
}
else if (Cs >= 1744) {
if (RsDiff < 11) {
if (Rs < 1481) {
return 1;
}
else if (Rs >= 1481) {
return 8;
}
}
else if (RsDiff >= 11) {
if (Rs < 2948) {
return 2;
}
else if (Rs >= 2948) {
if (Cs < 3525) {
return 8;
}
else if (Cs >= 3525) {
return 2;
}
}
}
}
}
else if (tcor >= 81) {
if (tcor < 85) {
if (refDCval < 85) {
return 1;
}
else if (refDCval >= 85) {
if (Rs < 3878) {
if (AFD < 22) {
if (Cs < 2251) {
return 4;
}
else if (Cs >= 2251) {
if (Rs < 3134) {
return 2;
}
else if (Rs >= 3134) {
return 4;
}
}
}
else if (AFD >= 22) {
return 2;
}
}
else if (Rs >= 3878) {
return 2;
}
}
}
else if (tcor >= 85) {
if (Rs < 2433) {
return 2;
}
else if (Rs >= 2433) {
if (Rs < 2999) {
return 8;
}
else if (Rs >= 2999) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (mu_mv_mag_sq < 61) {
if (AFD < 24) {
if (Cs < 1490) {
if (tcor < 81) {
return 2;
}
else if (tcor >= 81) {
if (mcTcor < 93) {
if (Rs < 1129) {
return 2;
}
else if (Rs >= 1129) {
return 1;
}
}
else if (mcTcor >= 93) {
return 8;
}
}
}
else if (Cs >= 1490) {
if (SC < 4702) {
if (AFD < 22) {
if (diffAFD < -9) {
return 2;
}
else if (diffAFD >= -9) {
return 8;
}
}
else if (AFD >= 22) {
return 1;
}
}
else if (SC >= 4702) {
if (SC < 4714) {
return 2;
}
else if (SC >= 4714) {
if (Cs < 2080) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
else if (Cs >= 2080) {
return 8;
}
}
}
}
}
else if (AFD >= 24) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
if (gchDC < 6) {
if (diffRsCsdiff < 36) {
return 8;
}
else if (diffRsCsdiff >= 36) {
return 1;
}
}
else if (gchDC >= 6) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 61) {
if (Cs < 2537) {
if (RsDiff < 16) {
if (mu_mv_mag_sq < 70) {
if (Cs < 2363) {
return 4;
}
else if (Cs >= 2363) {
if (Rs < 3018) {
return 2;
}
else if (Rs >= 3018) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 70) {
return 8;
}
}
else if (RsDiff >= 16) {
return 2;
}
}
else if (Cs >= 2537) {
if (Cs < 3700) {
if (MVDiff < 25) {
return 4;
}
else if (MVDiff >= 25) {
return 8;
}
}
else if (Cs >= 3700) {
if (Rs < 5348) {
return 8;
}
else if (Rs >= 5348) {
return 4;
}
}
}
}
}
}
}
}
else if (ssDCval >= 101) {
if (RsDiff < 14) {
if (posBalance < 44) {
if (SC < 8524) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (negBalance < 32) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (negBalance >= 32) {
if (ssDCval < 106) {
if (Rs < 2746) {
return 8;
}
else if (Rs >= 2746) {
return 2;
}
}
else if (ssDCval >= 106) {
if (posBalance < 36) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (posBalance >= 36) {
return 8;
}
}
}
}
}
else if (SC >= 8524) {
if (ssDCval < 118) {
return 1;
}
else if (ssDCval >= 118) {
return 8;
}
}
}
else if (posBalance >= 44) {
if (Rs < 5509) {
if (AFD < 18) {
if (diffAFD < -3) {
if (MVDiff < 14) {
if (Rs < 3392) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
else if (Rs >= 3392) {
if (RsCsDiff < 182) {
return 8;
}
else if (RsCsDiff >= 182) {
return 2;
}
}
}
else if (MVDiff >= 14) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (diffAFD >= -3) {
if (posBalance < 51) {
if (tcor < 81) {
return 2;
}
else if (tcor >= 81) {
if (CsDiff < 9) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 8;
}
}
else if (CsDiff >= 9) {
if (Rs < 2904) {
return 4;
}
else if (Rs >= 2904) {
if (Rs < 4087) {
return 8;
}
else if (Rs >= 4087) {
return 2;
}
}
}
}
}
else if (posBalance >= 51) {
if (ssDCval < 112) {
if (negBalance < 33) {
return 4;
}
else if (negBalance >= 33) {
if (MVDiff < 6) {
if (RsCsDiff < 172) {
if (AFD < 11) {
if (Cs < 1870) {
return 1;
}
else if (Cs >= 1870) {
return 8;
}
}
else if (AFD >= 11) {
if (Rs < 2193) {
return 4;
}
else if (Rs >= 2193) {
return 8;
}
}
}
else if (RsCsDiff >= 172) {
if (refDCval < 104) {
return 4;
}
else if (refDCval >= 104) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (Rs < 3540) {
return 8;
}
else if (Rs >= 3540) {
return 4;
}
}
}
}
}
else if (MVDiff >= 6) {
return 8;
}
}
}
else if (ssDCval >= 112) {
if (MVDiff < 24) {
if (mcTcor < 93) {
if (MVDiff < 3) {
if (ssDCval < 119) {
return 8;
}
else if (ssDCval >= 119) {
return 1;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (mcTcor >= 93) {
if (diffMVdiffVal < -12) {
if (Rs < 3613) {
return 8;
}
else if (Rs >= 3613) {
return 1;
}
}
else if (diffMVdiffVal >= -12) {
return 8;
}
}
}
else if (MVDiff >= 24) {
if (mu_mv_mag_sq < 57) {
return 8;
}
else if (mu_mv_mag_sq >= 57) {
return 4;
}
}
}
}
}
}
else if (AFD >= 18) {
if (Rs < 2017) {
if (mcTcor < 95) {
if (Rs < 1669) {
if (Cs < 2194) {
return 4;
}
else if (Cs >= 2194) {
return 1;
}
}
else if (Rs >= 1669) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
return 1;
}
}
}
else if (mcTcor >= 95) {
if (RsDiff < 10) {
return 1;
}
else if (RsDiff >= 10) {
if (Cs < 1478) {
if (Rs < 1752) {
if (Rs < 1449) {
return 8;
}
else if (Rs >= 1449) {
return 2;
}
}
else if (Rs >= 1752) {
return 1;
}
}
else if (Cs >= 1478) {
return 2;
}
}
}
}
else if (Rs >= 2017) {
if (ssDCval < 110) {
if (mcTcor < 95) {
if (Rs < 2570) {
if (Cs < 2573) {
return 8;
}
else if (Cs >= 2573) {
return 2;
}
}
else if (Rs >= 2570) {
return 1;
}
}
else if (mcTcor >= 95) {
if (TSC < 10) {
if (Rs < 3578) {
return 1;
}
else if (Rs >= 3578) {
return 4;
}
}
else if (TSC >= 10) {
return 2;
}
}
}
else if (ssDCval >= 110) {
if (RsDiff < 13) {
if (diffRsCsdiff < -61) {
if (RsCsDiff < 138) {
return 1;
}
else if (RsCsDiff >= 138) {
if (SC < 5436) {
return 8;
}
else if (SC >= 5436) {
if (Cs < 2692) {
return 4;
}
else if (Cs >= 2692) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -61) {
return 8;
}
}
else if (RsDiff >= 13) {
if (SC < 4611) {
return 1;
}
else if (SC >= 4611) {
if (AFD < 22) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
if (Rs < 4045) {
return 8;
}
else if (Rs >= 4045) {
return 1;
}
}
}
else if (AFD >= 22) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 5509) {
if (AFD < 10) {
if (Rs < 7018) {
return 8;
}
else if (Rs >= 7018) {
if (Rs < 8617) {
return 2;
}
else if (Rs >= 8617) {
return 1;
}
}
}
else if (AFD >= 10) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
if (Rs < 5815) {
return 1;
}
else if (Rs >= 5815) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 14) {
if (Cs < 4357) {
if (MVDiff < 18) {
if (Cs < 2412) {
if (AFD < 23) {
if (negBalance < 53) {
if (Rs < 2928) {
return 4;
}
else if (Rs >= 2928) {
return 2;
}
}
else if (negBalance >= 53) {
if (gchDC < 1) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 2;
}
}
else if (gchDC >= 1) {
if (mcTcor < 95) {
return 2;
}
else if (mcTcor >= 95) {
if (Cs < 1458) {
return 1;
}
else if (Cs >= 1458) {
if (Rs < 5886) {
return 2;
}
else if (Rs >= 5886) {
return 8;
}
}
}
}
}
}
else if (AFD >= 23) {
if (mu_mv_mag_sq < 13) {
if (Cs < 1399) {
return 2;
}
else if (Cs >= 1399) {
return 8;
}
}
else if (mu_mv_mag_sq >= 13) {
if (Rs < 3990) {
if (Rs < 3031) {
if (Rs < 2917) {
return 8;
}
else if (Rs >= 2917) {
return 1;
}
}
else if (Rs >= 3031) {
return 2;
}
}
else if (Rs >= 3990) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
return 4;
}
}
}
}
}
else if (Cs >= 2412) {
if (RsDiff < 22) {
if (Rs < 5027) {
if (ssDCval < 126) {
if (RsCsDiff < 503) {
return 8;
}
else if (RsCsDiff >= 503) {
if (RsCsDiff < 626) {
return 2;
}
else if (RsCsDiff >= 626) {
if (Rs < 4297) {
return 1;
}
else if (Rs >= 4297) {
return 8;
}
}
}
}
else if (ssDCval >= 126) {
return 1;
}
}
else if (Rs >= 5027) {
if (Rs < 7233) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (Rs >= 7233) {
if (Rs < 8518) {
return 2;
}
else if (Rs >= 8518) {
return 1;
}
}
}
}
else if (RsDiff >= 22) {
if (Cs < 3420) {
return 2;
}
else if (Cs >= 3420) {
return 1;
}
}
}
}
else if (MVDiff >= 18) {
if (negBalance < 58) {
if (refDCval < 144) {
if (Rs < 5619) {
if (gchDC < 11) {
if (refDCval < 136) {
if (AFD < 20) {
if (RsCsDiff < 353) {
if (Cs < 1387) {
return 2;
}
else if (Cs >= 1387) {
return 8;
}
}
else if (RsCsDiff >= 353) {
return 2;
}
}
else if (AFD >= 20) {
if (negBalance < 55) {
if (diffMVdiffVal < 13) {
if (CsDiff < 14) {
if (Rs < 3980) {
if (MVDiff < 21) {
if (Rs < 2897) {
return 1;
}
else if (Rs >= 2897) {
return 8;
}
}
else if (MVDiff >= 21) {
return 2;
}
}
else if (Rs >= 3980) {
return 1;
}
}
else if (CsDiff >= 14) {
return 2;
}
}
else if (diffMVdiffVal >= 13) {
return 8;
}
}
else if (negBalance >= 55) {
return 8;
}
}
}
else if (refDCval >= 136) {
if (Cs < 1342) {
return 1;
}
else if (Cs >= 1342) {
return 2;
}
}
}
else if (gchDC >= 11) {
if (Cs < 1540) {
return 1;
}
else if (Cs >= 1540) {
return 4;
}
}
}
else if (Rs >= 5619) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (posBalance < 58) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
else if (refDCval >= 144) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (Cs < 1485) {
if (Rs < 3397) {
return 2;
}
else if (Rs >= 3397) {
return 1;
}
}
else if (Cs >= 1485) {
if (tcor < 83) {
if (Rs < 6425) {
return 4;
}
else if (Rs >= 6425) {
return 2;
}
}
else if (tcor >= 83) {
return 8;
}
}
}
}
}
else if (negBalance >= 58) {
if (Rs < 2252) {
return 2;
}
else if (Rs >= 2252) {
if (RsCsDiff < 556) {
if (Cs < 1695) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (RsDiff < 22) {
return 1;
}
else if (RsDiff >= 22) {
return 8;
}
}
}
else if (Cs >= 1695) {
if (ssDCval < 113) {
return 2;
}
else if (ssDCval >= 113) {
return 8;
}
}
}
else if (RsCsDiff >= 556) {
return 1;
}
}
}
}
}
else if (Cs >= 4357) {
if (refDCval < 126) {
if (Rs < 6866) {
if (Rs < 4443) {
return 1;
}
else if (Rs >= 4443) {
return 8;
}
}
else if (Rs >= 6866) {
return 1;
}
}
else if (refDCval >= 126) {
if (gchDC < 12) {
return 8;
}
else if (gchDC >= 12) {
return 1;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1633) {
if (Rs < 3124) {
if (Rs < 1455) {
if (ssDCval < 122) {
if (negBalance < 1) {
return 2;
}
else if (negBalance >= 1) {
if (ssDCval < 64) {
if (AFD < 3) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 49) {
if (posBalance < 17) {
if (Rs < 803) {
return 4;
}
else if (Rs >= 803) {
return 8;
}
}
else if (posBalance >= 17) {
if (posBalance < 44) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (posBalance >= 44) {
if (Rs < 791) {
return 1;
}
else if (Rs >= 791) {
if (Rs < 1053) {
return 8;
}
else if (Rs >= 1053) {
return 4;
}
}
}
}
}
else if (posBalance >= 49) {
if (Rs < 743) {
return 1;
}
else if (Rs >= 743) {
if (Rs < 1243) {
return 8;
}
else if (Rs >= 1243) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1508) {
return 8;
}
else if (Cs >= 1508) {
return 1;
}
}
}
else if (AFD >= 3) {
if (Rs < 1270) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (Rs >= 1270) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
return 4;
}
}
}
}
}
else if (ssDCval >= 64) {
if (ssDCval < 77) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (Rs < 1418) {
if (CsDiff < 3) {
if (Rs < 1155) {
return 2;
}
else if (Rs >= 1155) {
return 1;
}
}
else if (CsDiff >= 3) {
return 4;
}
}
else if (Rs >= 1418) {
return 8;
}
}
}
else if (ssDCval >= 77) {
if (Rs < 1154) {
if (Rs < 903) {
if (Rs < 346) {
return 4;
}
else if (Rs >= 346) {
return 1;
}
}
else if (Rs >= 903) {
if (Rs < 989) {
return 4;
}
else if (Rs >= 989) {
return 2;
}
}
}
else if (Rs >= 1154) {
return 8;
}
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 2) {
if (diffTSC < -1) {
if (negBalance < 22) {
return 8;
}
else if (negBalance >= 22) {
return 4;
}
}
else if (diffTSC >= -1) {
if (Rs < 1384) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Rs >= 1384) {
return 8;
}
}
}
else if (AFD >= 2) {
if (mcTcor < 96) {
if (ssDCval < 129) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
return 4;
}
}
else if (ssDCval >= 129) {
if (Rs < 1234) {
return 8;
}
else if (Rs >= 1234) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 120) {
return 8;
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 1009) {
return 8;
}
else if (Rs >= 1009) {
if (Rs < 1270) {
return 4;
}
else if (Rs >= 1270) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1455) {
if (ssDCval < 86) {
if (tcor < 97) {
if (posBalance < 46) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (posBalance < 38) {
if (Rs < 2208) {
if (Cs < 1436) {
return 8;
}
else if (Cs >= 1436) {
return 1;
}
}
else if (Rs >= 2208) {
return 4;
}
}
else if (posBalance >= 38) {
if (mu_mv_mag_sq < 113) {
if (Rs < 2002) {
return 4;
}
else if (Rs >= 2002) {
if (Rs < 2456) {
return 8;
}
else if (Rs >= 2456) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 113) {
if (Rs < 3064) {
return 8;
}
else if (Rs >= 3064) {
return 2;
}
}
}
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (tcor >= 97) {
if (negBalance < 19) {
if (SC < 4462) {
if (Cs < 1344) {
if (Rs < 2273) {
return 8;
}
else if (Rs >= 2273) {
if (mu_mv_mag_sq < 127) {
if (Rs < 2338) {
if (Cs < 1323) {
if (Rs < 2290) {
if (Rs < 2279) {
return 2;
}
else if (Rs >= 2279) {
return 8;
}
}
else if (Rs >= 2290) {
return 2;
}
}
else if (Cs >= 1323) {
if (Cs < 1329) {
if (Cs < 1327) {
return 8;
}
else if (Cs >= 1327) {
return 2;
}
}
else if (Cs >= 1329) {
return 8;
}
}
}
else if (Rs >= 2338) {
if (Rs < 2484) {
return 2;
}
else if (Rs >= 2484) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 11) {
return 8;
}
else if (negBalance >= 11) {
if (posBalance < 29) {
return 8;
}
else if (posBalance >= 29) {
return 2;
}
}
}
}
}
else if (Cs >= 1344) {
if (SC < 3319) {
if (Cs < 1441) {
return 8;
}
else if (Cs >= 1441) {
if (Cs < 1594) {
return 1;
}
else if (Cs >= 1594) {
return 8;
}
}
}
else if (SC >= 3319) {
return 8;
}
}
}
else if (SC >= 4462) {
if (Rs < 3084) {
return 1;
}
else if (Rs >= 3084) {
return 8;
}
}
}
else if (negBalance >= 19) {
if (posBalance < 39) {
if (Rs < 1519) {
return 1;
}
else if (Rs >= 1519) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
else if (posBalance >= 39) {
if (negBalance < 35) {
if (refDCval < 79) {
if (mcTcor < 98) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 2364) {
return 4;
}
else if (Rs >= 2364) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (posBalance < 42) {
if (Cs < 1595) {
return 4;
}
else if (Cs >= 1595) {
return 8;
}
}
else if (posBalance >= 42) {
return 4;
}
}
}
else if (refDCval >= 79) {
if (Rs < 2484) {
return 8;
}
else if (Rs >= 2484) {
return 4;
}
}
}
else if (negBalance >= 35) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (diffAFD < -20) {
if (Rs < 1750) {
return 1;
}
else if (Rs >= 1750) {
if (Rs < 2540) {
return 4;
}
else if (Rs >= 2540) {
return 8;
}
}
}
else if (diffAFD >= -20) {
if (posBalance < 47) {
if (Rs < 3048) {
return 8;
}
else if (Rs >= 3048) {
if (Rs < 3065) {
return 2;
}
else if (Rs >= 3065) {
return 8;
}
}
}
else if (posBalance >= 47) {
if (ssDCval < 117) {
if (diffAFD < -1) {
if (Cs < 1404) {
if (ssDCval < 98) {
return 2;
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (Cs >= 1404) {
if (AFD < 11) {
if (refDCval < 89) {
if (Rs < 2188) {
return 8;
}
else if (Rs >= 2188) {
return 2;
}
}
else if (refDCval >= 89) {
return 8;
}
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (ssDCval >= 117) {
if (CsDiff < 2) {
if (Rs < 2819) {
return 8;
}
else if (Rs >= 2819) {
return 1;
}
}
else if (CsDiff >= 2) {
if (mu_mv_mag_sq < 125) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 2476) {
return 1;
}
else if (Rs >= 2476) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 140) {
if (diffAFD < -2) {
return 8;
}
else if (diffAFD >= -2) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3124) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 125) {
if (SC < 4913) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
else if (SC >= 4913) {
if (posBalance < 45) {
if (ssDCval < 63) {
if (negBalance < 20) {
if (Rs < 5513) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (posBalance < 30) {
return 4;
}
else if (posBalance >= 30) {
return 8;
}
}
}
else if (Rs >= 5513) {
return 8;
}
}
else if (negBalance >= 20) {
return 8;
}
}
else if (ssDCval >= 63) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Rs < 4453) {
return 4;
}
else if (Rs >= 4453) {
return 8;
}
}
}
}
else if (posBalance >= 45) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1570) {
if (negBalance < 25) {
if (SC < 5043) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (SC >= 5043) {
return 8;
}
}
else if (negBalance >= 25) {
if (Rs < 4013) {
return 8;
}
else if (Rs >= 4013) {
return 4;
}
}
}
else if (Cs >= 1570) {
if (Rs < 3242) {
return 8;
}
else if (Rs >= 3242) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 111) {
if (posBalance < 44) {
if (Rs < 4079) {
if (ssDCval < 86) {
return 2;
}
else if (ssDCval >= 86) {
return 4;
}
}
else if (Rs >= 4079) {
if (ssDCval < 96) {
if (negBalance < 25) {
return 2;
}
else if (negBalance >= 25) {
return 8;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
else if (posBalance >= 44) {
if (MVDiff < 56) {
if (ssDCval < 92) {
return 2;
}
else if (ssDCval >= 92) {
if (Cs < 1444) {
return 2;
}
else if (Cs >= 1444) {
return 8;
}
}
}
else if (MVDiff >= 56) {
return 4;
}
}
}
else if (ssDCval >= 111) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (Rs < 3310) {
return 2;
}
else if (Rs >= 3310) {
if (Cs < 1549) {
return 4;
}
else if (Cs >= 1549) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 1633) {
if (ssDCval < 102) {
if (Rs < 3496) {
if (negBalance < 54) {
if (ssDCval < 67) {
if (posBalance < 35) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (ssDCval < 61) {
if (negBalance < 16) {
if (posBalance < 6) {
return 4;
}
else if (posBalance >= 6) {
return 1;
}
}
else if (negBalance >= 16) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
if (Rs < 2655) {
return 1;
}
else if (Rs >= 2655) {
return 4;
}
}
}
}
else if (ssDCval >= 61) {
if (Rs < 2711) {
if (Cs < 1707) {
return 8;
}
else if (Cs >= 1707) {
return 1;
}
}
else if (Rs >= 2711) {
return 8;
}
}
}
}
else if (posBalance >= 35) {
if (Rs < 1648) {
if (ssDCval < 60) {
if (Rs < 998) {
if (Rs < 791) {
return 1;
}
else if (Rs >= 791) {
return 8;
}
}
else if (Rs >= 998) {
return 4;
}
}
else if (ssDCval >= 60) {
if (AFD < 3) {
if (Rs < 1131) {
return 1;
}
else if (Rs >= 1131) {
return 8;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (Rs >= 1648) {
if (SC < 4724) {
if (SC < 4011) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (SC >= 4011) {
if (RsDiff < 2) {
if (ssDCval < 66) {
if (diffRsCsdiff < -2) {
if (Cs < 1851) {
return 8;
}
else if (Cs >= 1851) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
}
else if (ssDCval >= 66) {
return 1;
}
}
else if (RsDiff >= 2) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (refDCval < 66) {
return 8;
}
else if (refDCval >= 66) {
if (Cs < 1773) {
return 1;
}
else if (Cs >= 1773) {
return 8;
}
}
}
}
}
}
else if (SC >= 4724) {
if (Cs < 2037) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 119) {
return 8;
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 2951) {
return 4;
}
else if (Rs >= 2951) {
if (negBalance < 25) {
if (SC < 4873) {
if (tcor < 98) {
if (Rs < 2970) {
return 8;
}
else if (Rs >= 2970) {
return 4;
}
}
else if (tcor >= 98) {
if (Rs < 3054) {
return 8;
}
else if (Rs >= 3054) {
return 8;
}
}
}
else if (SC >= 4873) {
if (Cs < 1916) {
if (posBalance < 39) {
if (Rs < 3026) {
return 8;
}
else if (Rs >= 3026) {
if (Rs < 3059) {
return 4;
}
else if (Rs >= 3059) {
if (Rs < 3142) {
if (Rs < 3110) {
return 4;
}
else if (Rs >= 3110) {
return 4;
}
}
else if (Rs >= 3142) {
return 4;
}
}
}
}
else if (posBalance >= 39) {
if (Rs < 3023) {
return 4;
}
else if (Rs >= 3023) {
return 4;
}
}
}
else if (Cs >= 1916) {
return 8;
}
}
}
else if (negBalance >= 25) {
if (Cs < 1831) {
if (Rs < 3010) {
if (Rs < 2964) {
return 4;
}
else if (Rs >= 2964) {
return 4;
}
}
else if (Rs >= 3010) {
return 8;
}
}
else if (Cs >= 1831) {
return 4;
}
}
}
}
}
else if (posBalance >= 49) {
return 4;
}
}
else if (Cs >= 2037) {
if (Rs < 2022) {
return 4;
}
else if (Rs >= 2022) {
if (Cs < 2156) {
if (mu_mv_mag_sq < 127) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
if (SC < 4973) {
return 8;
}
else if (SC >= 4973) {
return 4;
}
}
}
else if (Cs >= 2156) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (posBalance < 43) {
if (mu_mv_mag_sq < 121) {
if (TSC < 1) {
if (ssDCval < 92) {
if (posBalance < 9) {
return 2;
}
else if (posBalance >= 9) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
if (negBalance < 9) {
return 1;
}
else if (negBalance >= 9) {
return 8;
}
}
}
}
else if (ssDCval >= 92) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
if (Rs < 2326) {
return 2;
}
else if (Rs >= 2326) {
if (Rs < 2343) {
return 4;
}
else if (Rs >= 2343) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 4;
}
}
}
}
}
}
else if (TSC >= 1) {
if (negBalance < 23) {
if (tcor < 100) {
return 8;
}
else if (tcor >= 100) {
if (Rs < 2637) {
return 1;
}
else if (Rs >= 2637) {
return 8;
}
}
}
else if (negBalance >= 23) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
if (CsDiff < 2) {
if (ssDCval < 88) {
return 2;
}
else if (ssDCval >= 88) {
if (Rs < 3387) {
return 8;
}
else if (Rs >= 3387) {
return 2;
}
}
}
else if (CsDiff >= 2) {
if (Cs < 1779) {
return 4;
}
else if (Cs >= 1779) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 121) {
if (posBalance < 32) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (posBalance >= 32) {
if (Cs < 2260) {
if (Rs < 3047) {
if (Rs < 2568) {
return 8;
}
else if (Rs >= 2568) {
if (Cs < 1881) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (Cs >= 1881) {
return 8;
}
}
}
else if (Rs >= 3047) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
return 2;
}
}
}
else if (Cs >= 2260) {
if (Cs < 3504) {
return 8;
}
else if (Cs >= 3504) {
return 4;
}
}
}
}
}
else if (posBalance >= 43) {
if (SC < 4142) {
if (SC < 2734) {
if (Rs < 637) {
return 8;
}
else if (Rs >= 637) {
return 1;
}
}
else if (SC >= 2734) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
return 8;
}
}
}
else if (SC >= 4142) {
if (Cs < 1819) {
if (TSC < 5) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
return 4;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (Cs >= 1819) {
if (posBalance < 63) {
if (diffAFD < -2) {
if (ssDCval < 95) {
if (ssDCval < 83) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
return 4;
}
}
else if (ssDCval >= 83) {
return 8;
}
}
else if (ssDCval >= 95) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 8;
}
}
}
else if (diffAFD >= -2) {
if (Cs < 4092) {
if (Rs < 2145) {
if (ssDCval < 71) {
if (Rs < 2008) {
return 8;
}
else if (Rs >= 2008) {
return 4;
}
}
else if (ssDCval >= 71) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
return 4;
}
}
}
else if (Rs >= 2145) {
if (Cs < 1925) {
if (negBalance < 33) {
if (refDCval < 84) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (refDCval >= 84) {
return 8;
}
}
else if (negBalance >= 33) {
return 8;
}
}
else if (Cs >= 1925) {
if (Cs < 2518) {
if (ssDCval < 71) {
if (diffRsCsdiff < 16) {
if (Rs < 2171) {
return 4;
}
else if (Rs >= 2171) {
return 8;
}
}
else if (diffRsCsdiff >= 16) {
if (Cs < 1988) {
return 8;
}
else if (Cs >= 1988) {
return 4;
}
}
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (Cs >= 2518) {
if (Rs < 2533) {
return 4;
}
else if (Rs >= 2533) {
if (ssDCval < 89) {
if (diffRsCsdiff < 44) {
return 8;
}
else if (diffRsCsdiff >= 44) {
if (Rs < 3039) {
return 8;
}
else if (Rs >= 3039) {
return 4;
}
}
}
else if (ssDCval >= 89) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (mu_mv_mag_sq < 122) {
if (Rs < 3082) {
return 8;
}
else if (Rs >= 3082) {
if (Cs < 3273) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 3273) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 3393) {
return 8;
}
else if (Rs >= 3393) {
if (Cs < 3576) {
if (ssDCval < 98) {
return 4;
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (Cs >= 3576) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 4092) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 4;
}
}
}
}
else if (posBalance >= 63) {
if (mcTcor < 92) {
if (Rs < 1377) {
return 2;
}
else if (Rs >= 1377) {
if (ssDCval < 71) {
return 8;
}
else if (ssDCval >= 71) {
return 4;
}
}
}
else if (mcTcor >= 92) {
if (Rs < 2199) {
return 4;
}
else if (Rs >= 2199) {
if (Rs < 3414) {
return 8;
}
else if (Rs >= 3414) {
return 4;
}
}
}
}
}
}
}
}
}
else if (negBalance >= 54) {
if (ssDCval < 69) {
return 8;
}
else if (ssDCval >= 69) {
return 4;
}
}
}
else if (Rs >= 3496) {
if (Cs < 2222) {
if (ssDCval < 69) {
if (Rs < 7392) {
if (ssDCval < 62) {
if (Rs < 3914) {
return 8;
}
else if (Rs >= 3914) {
if (Rs < 4662) {
if (SC < 6538) {
if (diffRsCsdiff < -14) {
if (Rs < 4223) {
return 4;
}
else if (Rs >= 4223) {
return 8;
}
}
else if (diffRsCsdiff >= -14) {
if (RsCsDiff < 38) {
if (diffTSC < 1) {
if (diffAFD < 1) {
if (diffRsCsdiff < 1) {
if (Rs < 4264) {
return 4;
}
else if (Rs >= 4264) {
if (Rs < 4417) {
if (Cs < 2118) {
return 4;
}
else if (Cs >= 2118) {
return 4;
}
}
else if (Rs >= 4417) {
return 4;
}
}
}
else if (diffRsCsdiff >= 1) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
if (Cs < 2083) {
return 4;
}
else if (Cs >= 2083) {
return 4;
}
}
}
}
else if (diffAFD >= 1) {
if (Rs < 4239) {
return 4;
}
else if (Rs >= 4239) {
return 4;
}
}
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (RsCsDiff >= 38) {
return 4;
}
}
}
else if (SC >= 6538) {
if (posBalance < 31) {
return 4;
}
else if (posBalance >= 31) {
if (Rs < 4384) {
return 8;
}
else if (Rs >= 4384) {
return 8;
}
}
}
}
else if (Rs >= 4662) {
return 4;
}
}
}
else if (ssDCval >= 62) {
return 8;
}
}
else if (Rs >= 7392) {
return 2;
}
}
else if (ssDCval >= 69) {
if (refDCval < 75) {
if (Rs < 3711) {
return 8;
}
else if (Rs >= 3711) {
return 4;
}
}
else if (refDCval >= 75) {
if (ssDCval < 99) {
if (Rs < 4927) {
if (RsDiff < 7) {
if (ssDCval < 91) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (Rs < 4651) {
return 2;
}
else if (Rs >= 4651) {
return 8;
}
}
}
else if (ssDCval >= 91) {
if (negBalance < 29) {
if (Cs < 1738) {
if (Rs < 4190) {
return 8;
}
else if (Rs >= 4190) {
return 4;
}
}
else if (Cs >= 1738) {
if (Cs < 2067) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 2;
}
}
else if (Cs >= 2067) {
return 8;
}
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
else if (RsDiff >= 7) {
if (Cs < 2142) {
return 2;
}
else if (Cs >= 2142) {
return 1;
}
}
}
else if (Rs >= 4927) {
return 4;
}
}
else if (ssDCval >= 99) {
return 4;
}
}
}
}
else if (Cs >= 2222) {
if (SC < 8344) {
if (Cs < 2720) {
if (mu_mv_mag_sq < 120) {
if (ssDCval < 101) {
if (CsDiff < 2) {
if (Cs < 2307) {
return 4;
}
else if (Cs >= 2307) {
return 8;
}
}
else if (CsDiff >= 2) {
if (SC < 6284) {
return 8;
}
else if (SC >= 6284) {
if (posBalance < 46) {
if (SC < 7723) {
if (mu_mv_mag_sq < 113) {
return 8;
}
else if (mu_mv_mag_sq >= 113) {
if (diffRsCsdiff < 18) {
if (Rs < 4868) {
if (Cs < 2425) {
if (Cs < 2373) {
if (Cs < 2310) {
return 4;
}
else if (Cs >= 2310) {
if (Rs < 4540) {
return 4;
}
else if (Rs >= 4540) {
return 4;
}
}
}
else if (Cs >= 2373) {
return 8;
}
}
else if (Cs >= 2425) {
return 4;
}
}
else if (Rs >= 4868) {
if (Rs < 4885) {
if (Rs < 4877) {
return 8;
}
else if (Rs >= 4877) {
return 8;
}
}
else if (Rs >= 4885) {
return 4;
}
}
}
else if (diffRsCsdiff >= 18) {
if (TSC < 4) {
if (Rs < 4752) {
return 8;
}
else if (Rs >= 4752) {
return 4;
}
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (SC >= 7723) {
return 4;
}
}
else if (posBalance >= 46) {
if (Rs < 4791) {
return 4;
}
else if (Rs >= 4791) {
return 2;
}
}
}
}
}
else if (ssDCval >= 101) {
return 1;
}
}
else if (mu_mv_mag_sq >= 120) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 60) {
if (SC < 6736) {
return 4;
}
else if (SC >= 6736) {
if (Rs < 4769) {
return 8;
}
else if (Rs >= 4769) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
if (Rs < 4299) {
return 8;
}
else if (Rs >= 4299) {
if (Rs < 4698) {
return 1;
}
else if (Rs >= 4698) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
return 8;
}
}
}
else if (Cs >= 2720) {
if (refDCval < 85) {
if (negBalance < 20) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (negBalance >= 20) {
return 8;
}
}
else if (refDCval >= 85) {
if (refDCval < 98) {
if (AFD < 12) {
if (refDCval < 95) {
if (Rs < 4403) {
if (diffRsCsdiff < -64) {
return 8;
}
else if (diffRsCsdiff >= -64) {
if (Cs < 2743) {
return 8;
}
else if (Cs >= 2743) {
return 4;
}
}
}
else if (Rs >= 4403) {
return 8;
}
}
else if (refDCval >= 95) {
if (Rs < 3994) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 8;
}
}
}
else if (Rs >= 3994) {
return 4;
}
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (refDCval >= 98) {
return 8;
}
}
}
}
else if (SC >= 8344) {
if (Rs < 6660) {
if (MVDiff < 3) {
if (negBalance < 29) {
if (Rs < 4482) {
if (Rs < 4373) {
return 8;
}
else if (Rs >= 4373) {
return 1;
}
}
else if (Rs >= 4482) {
if (mu_mv_mag_sq < 113) {
if (Rs < 5705) {
return 8;
}
else if (Rs >= 5705) {
return 2;
}
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
}
else if (negBalance >= 29) {
if (Cs < 3639) {
if (Rs < 5314) {
if (Rs < 4778) {
return 8;
}
else if (Rs >= 4778) {
return 4;
}
}
else if (Rs >= 5314) {
return 8;
}
}
else if (Cs >= 3639) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 5873) {
return 8;
}
else if (Rs >= 5873) {
if (Cs < 3980) {
return 8;
}
else if (Cs >= 3980) {
if (ssDCval < 82) {
return 2;
}
else if (ssDCval >= 82) {
return 8;
}
}
}
}
}
else if (Rs >= 6660) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 102) {
if (ssDCval < 129) {
if (Rs < 1520) {
if (refDCval < 116) {
if (SC < 3323) {
if (posBalance < 55) {
if (Rs < 1193) {
return 2;
}
else if (Rs >= 1193) {
return 8;
}
}
else if (posBalance >= 55) {
if (Rs < 1146) {
return 4;
}
else if (Rs >= 1146) {
return 2;
}
}
}
else if (SC >= 3323) {
return 2;
}
}
else if (refDCval >= 116) {
return 8;
}
}
else if (Rs >= 1520) {
if (refDCval < 105) {
if (negBalance < 43) {
if (posBalance < 54) {
if (Rs < 2116) {
return 2;
}
else if (Rs >= 2116) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (CsDiff < 3) {
if (Rs < 4077) {
return 8;
}
else if (Rs >= 4077) {
return 2;
}
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
else if (posBalance >= 54) {
if (Rs < 4875) {
return 8;
}
else if (Rs >= 4875) {
if (Rs < 5474) {
return 2;
}
else if (Rs >= 5474) {
return 8;
}
}
}
}
else if (negBalance >= 43) {
if (Cs < 3040) {
if (Rs < 4920) {
if (Rs < 3188) {
return 8;
}
else if (Rs >= 3188) {
return 1;
}
}
else if (Rs >= 4920) {
return 2;
}
}
else if (Cs >= 3040) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (diffAFD < -2) {
if (diffRsCsdiff < -999) {
return 1;
}
else if (diffRsCsdiff >= -999) {
if (negBalance < 24) {
if (diffRsCsdiff < -4) {
if (diffMVdiffVal < -101) {
return 4;
}
else if (diffMVdiffVal >= -101) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
if (Cs < 2678) {
return 8;
}
else if (Cs >= 2678) {
return 2;
}
}
}
else if (negBalance >= 24) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (Cs < 3496) {
if (posBalance < 55) {
if (Cs < 1983) {
return 4;
}
else if (Cs >= 1983) {
if (ssDCval < 120) {
return 8;
}
else if (ssDCval >= 120) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
if (Cs < 2338) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (Cs >= 2338) {
if (Rs < 3352) {
return 1;
}
else if (Rs >= 3352) {
return 2;
}
}
}
}
else if (Cs >= 3496) {
return 8;
}
}
}
}
}
else if (diffAFD >= -2) {
if (mu_mv_mag_sq < 122) {
if (CsDiff < 1) {
if (Cs < 2965) {
if (ssDCval < 118) {
if (Cs < 2938) {
if (Rs < 2679) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
return 4;
}
}
else if (Rs >= 2679) {
if (negBalance < 8) {
return 8;
}
else if (negBalance >= 8) {
return 4;
}
}
}
else if (Cs >= 2938) {
return 8;
}
}
else if (ssDCval >= 118) {
return 8;
}
}
else if (Cs >= 2965) {
if (Rs < 4034) {
return 2;
}
else if (Rs >= 4034) {
return 8;
}
}
}
else if (CsDiff >= 1) {
if (TSC < 10) {
if (Rs < 6639) {
if (diffTSC < 3) {
return 8;
}
else if (diffTSC >= 3) {
if (Cs < 2050) {
if (Rs < 2784) {
return 1;
}
else if (Rs >= 2784) {
return 4;
}
}
else if (Cs >= 2050) {
return 8;
}
}
}
else if (Rs >= 6639) {
if (Cs < 4021) {
return 4;
}
else if (Cs >= 4021) {
return 8;
}
}
}
else if (TSC >= 10) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (SC < 6907) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
else if (SC >= 6907) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Rs < 1908) {
if (refDCval < 113) {
if (Cs < 2011) {
if (SC < 3854) {
return 8;
}
else if (SC >= 3854) {
return 2;
}
}
else if (Cs >= 2011) {
if (Rs < 1896) {
return 2;
}
else if (Rs >= 1896) {
return 8;
}
}
}
else if (refDCval >= 113) {
if (Cs < 1652) {
if (Rs < 1859) {
return 1;
}
else if (Rs >= 1859) {
return 8;
}
}
else if (Cs >= 1652) {
return 8;
}
}
}
else if (Rs >= 1908) {
if (diffMVdiffVal < -1) {
if (posBalance < 1) {
return 1;
}
else if (posBalance >= 1) {
if (ssDCval < 115) {
return 8;
}
else if (ssDCval >= 115) {
if (ssDCval < 118) {
if (Cs < 2687) {
return 8;
}
else if (Cs >= 2687) {
if (Rs < 2760) {
return 8;
}
else if (Rs >= 2760) {
return 2;
}
}
}
else if (ssDCval >= 118) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (diffAFD < 3) {
if (AFD < 28) {
if (Cs < 1679) {
if (Rs < 4147) {
return 8;
}
else if (Rs >= 4147) {
return 2;
}
}
else if (Cs >= 1679) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
return 2;
}
}
}
}
else if (AFD >= 28) {
return 1;
}
}
else if (diffAFD >= 3) {
if (Cs < 2233) {
return 2;
}
else if (Cs >= 2233) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 129) {
if (mcTcor < 96) {
if (negBalance < 39) {
if (ssDCval < 200) {
if (ssDCval < 130) {
if (Rs < 2478) {
return 1;
}
else if (Rs >= 2478) {
return 8;
}
}
else if (ssDCval >= 130) {
return 8;
}
}
else if (ssDCval >= 200) {
if (Rs < 1688) {
return 1;
}
else if (Rs >= 1688) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (ssDCval < 137) {
return 8;
}
else if (ssDCval >= 137) {
if (Rs < 6023) {
return 1;
}
else if (Rs >= 6023) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 144) {
if (SC < 6378) {
if (SC < 3056) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Rs < 1143) {
return 8;
}
else if (Rs >= 1143) {
if (Rs < 1259) {
return 4;
}
else if (Rs >= 1259) {
return 8;
}
}
}
}
else if (SC >= 3056) {
if (diffAFD < -20) {
return 1;
}
else if (diffAFD >= -20) {
if (SC < 4969) {
return 8;
}
else if (SC >= 4969) {
if (SC < 4983) {
return 1;
}
else if (SC >= 4983) {
return 8;
}
}
}
}
}
else if (SC >= 6378) {
if (Cs < 2750) {
if (Rs < 4634) {
return 8;
}
else if (Rs >= 4634) {
return 2;
}
}
else if (Cs >= 2750) {
if (SC < 7857) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (posBalance < 15) {
if (Rs < 3018) {
return 1;
}
else if (Rs >= 3018) {
return 8;
}
}
else if (posBalance >= 15) {
return 1;
}
}
}
else if (SC >= 7857) {
return 8;
}
}
}
}
else if (refDCval >= 144) {
if (ssDCval < 201) {
return 8;
}
else if (ssDCval >= 201) {
if (Cs < 4080) {
return 8;
}
else if (Cs >= 4080) {
return 4;
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect16(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1313) {
if (Rs < 2340) {
if (Cs < 506) {
if (negBalance < 14) {
if (MVDiff < 7) {
if (SC < 798) {
if (ssDCval < 48) {
if (SC < 375) {
if (TSC < 1) {
if (SC < 59) {
if (ssDCval < 22) {
if (ssDCval < 9) {
return 4;
}
else if (ssDCval >= 9) {
return 8;
}
}
else if (ssDCval >= 22) {
if (posBalance < 16) {
if (negBalance < 13) {
return 4;
}
else if (negBalance >= 13) {
if (Rs < 24) {
return 8;
}
else if (Rs >= 24) {
return 4;
}
}
}
else if (posBalance >= 16) {
if (Rs < 5) {
return 2;
}
else if (Rs >= 5) {
return 8;
}
}
}
}
else if (SC >= 59) {
if (Cs < 183) {
if (negBalance < 12) {
if (Cs < 179) {
if (negBalance < 8) {
return 8;
}
else if (negBalance >= 8) {
if (tcor < 97) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
if (Rs < 93) {
return 8;
}
else if (Rs >= 93) {
return 1;
}
}
}
else if (tcor >= 97) {
return 8;
}
}
}
else if (Cs >= 179) {
if (Rs < 162) {
return 4;
}
else if (Rs >= 162) {
return 8;
}
}
}
else if (negBalance >= 12) {
if (ssDCval < 35) {
if (Rs < 70) {
return 4;
}
else if (Rs >= 70) {
return 8;
}
}
else if (ssDCval >= 35) {
return 8;
}
}
}
else if (Cs >= 183) {
if (Rs < 130) {
if (Cs < 206) {
return 8;
}
else if (Cs >= 206) {
return 1;
}
}
else if (Rs >= 130) {
return 4;
}
}
}
}
else if (TSC >= 1) {
if (SC < 239) {
if (posBalance < 37) {
if (ssDCval < 38) {
if (Rs < 90) {
if (Cs < 38) {
return 1;
}
else if (Cs >= 38) {
return 4;
}
}
else if (Rs >= 90) {
if (ssDCval < 31) {
return 4;
}
else if (ssDCval >= 31) {
return 8;
}
}
}
else if (ssDCval >= 38) {
if (Rs < 55) {
return 8;
}
else if (Rs >= 55) {
return 1;
}
}
}
else if (posBalance >= 37) {
if (Rs < 3) {
return 4;
}
else if (Rs >= 3) {
if (ssDCval < 29) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 8;
}
}
else if (ssDCval >= 29) {
return 8;
}
}
}
}
else if (SC >= 239) {
return 1;
}
}
}
else if (SC >= 375) {
if (Rs < 202) {
if (Cs < 271) {
if (mu_mv_mag_sq < 115) {
return 4;
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 165) {
return 8;
}
else if (Rs >= 165) {
if (Rs < 183) {
return 4;
}
else if (Rs >= 183) {
return 1;
}
}
}
}
else if (Cs >= 271) {
if (tcor < 90) {
return 4;
}
else if (tcor >= 90) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
return 1;
}
}
}
}
else if (Rs >= 202) {
if (mu_mv_mag_sq < 125) {
if (ssDCval < 29) {
if (Rs < 211) {
return 2;
}
else if (Rs >= 211) {
return 1;
}
}
else if (ssDCval >= 29) {
if (ssDCval < 40) {
if (Cs < 212) {
return 8;
}
else if (Cs >= 212) {
return 4;
}
}
else if (ssDCval >= 40) {
if (Cs < 221) {
if (Rs < 348) {
return 4;
}
else if (Rs >= 348) {
return 8;
}
}
else if (Cs >= 221) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (tcor < 101) {
if (Rs < 430) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (Rs >= 430) {
return 8;
}
}
else if (tcor >= 101) {
return 8;
}
}
}
}
}
else if (ssDCval >= 48) {
if (ssDCval < 84) {
if (Rs < 23) {
return 2;
}
else if (Rs >= 23) {
if (ssDCval < 67) {
if (Cs < 279) {
if (negBalance < 13) {
if (ssDCval < 61) {
if (ssDCval < 55) {
if (refDCval < 54) {
if (SC < 407) {
return 8;
}
else if (SC >= 407) {
if (Cs < 100) {
return 8;
}
else if (Cs >= 100) {
return 4;
}
}
}
else if (refDCval >= 54) {
if (Rs < 91) {
return 8;
}
else if (Rs >= 91) {
return 4;
}
}
}
else if (ssDCval >= 55) {
return 8;
}
}
else if (ssDCval >= 61) {
if (Rs < 307) {
return 2;
}
else if (Rs >= 307) {
return 8;
}
}
}
else if (negBalance >= 13) {
if (ssDCval < 53) {
return 8;
}
else if (ssDCval >= 53) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
}
else if (Cs >= 279) {
if (Rs < 251) {
return 4;
}
else if (Rs >= 251) {
if (Rs < 404) {
return 1;
}
else if (Rs >= 404) {
return 8;
}
}
}
}
else if (ssDCval >= 67) {
if (Rs < 78) {
return 4;
}
else if (Rs >= 78) {
if (Rs < 336) {
if (ssDCval < 80) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
return 1;
}
}
else if (ssDCval >= 80) {
return 8;
}
}
else if (Rs >= 336) {
if (SC < 646) {
return 4;
}
else if (SC >= 646) {
if (SC < 731) {
return 8;
}
else if (SC >= 731) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
if (ssDCval < 94) {
return 2;
}
else if (ssDCval >= 94) {
if (Rs < 241) {
return 4;
}
else if (Rs >= 241) {
if (Rs < 379) {
if (Rs < 357) {
return 8;
}
else if (Rs >= 357) {
return 2;
}
}
else if (Rs >= 379) {
if (ssDCval < 161) {
return 1;
}
else if (ssDCval >= 161) {
return 4;
}
}
}
}
}
}
}
}
else if (SC >= 798) {
if (ssDCval < 60) {
if (mu_mv_mag_sq < 125) {
if (Cs < 255) {
if (MVDiff < 1) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 8;
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (Cs >= 255) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (ssDCval < 57) {
if (SC < 1127) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (SC >= 1127) {
return 1;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 499) {
if (Rs < 1441) {
return 8;
}
else if (Rs >= 1441) {
if (Rs < 1449) {
return 1;
}
else if (Rs >= 1449) {
return 8;
}
}
}
else if (Cs >= 499) {
if (Rs < 1364) {
return 1;
}
else if (Rs >= 1364) {
return 8;
}
}
}
}
else if (ssDCval >= 60) {
if (ssDCval < 67) {
if (SC < 953) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (mu_mv_mag_sq < 116) {
return 8;
}
else if (mu_mv_mag_sq >= 116) {
return 1;
}
}
}
}
else if (SC >= 953) {
if (Cs < 445) {
return 1;
}
else if (Cs >= 445) {
if (Rs < 1466) {
if (Rs < 1092) {
return 8;
}
else if (Rs >= 1092) {
return 2;
}
}
else if (Rs >= 1466) {
return 1;
}
}
}
}
else if (ssDCval >= 67) {
if (ssDCval < 141) {
if (ssDCval < 84) {
if (Rs < 1113) {
if (Rs < 641) {
if (Cs < 387) {
if (Rs < 590) {
return 8;
}
else if (Rs >= 590) {
return 2;
}
}
else if (Cs >= 387) {
return 4;
}
}
else if (Rs >= 641) {
return 8;
}
}
else if (Rs >= 1113) {
if (Rs < 1432) {
return 2;
}
else if (Rs >= 1432) {
return 1;
}
}
}
else if (ssDCval >= 84) {
if (Cs < 457) {
return 8;
}
else if (Cs >= 457) {
if (Rs < 624) {
return 8;
}
else if (Rs >= 624) {
return 4;
}
}
}
}
else if (ssDCval >= 141) {
if (Rs < 547) {
if (posBalance < 13) {
return 8;
}
else if (posBalance >= 13) {
return 1;
}
}
else if (Rs >= 547) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (tcor < 94) {
if (negBalance < 5) {
if (TSC < 7) {
if (posBalance < 93) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
if (refDCval < 50) {
if (mu_mv_mag_sq < 39) {
if (diffMVdiffVal < 22) {
if (MVDiff < 33) {
return 2;
}
else if (MVDiff >= 33) {
return 8;
}
}
else if (diffMVdiffVal >= 22) {
if (Rs < 50) {
return 2;
}
else if (Rs >= 50) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (MVDiff < 8) {
if (Rs < 355) {
return 4;
}
else if (Rs >= 355) {
return 8;
}
}
else if (MVDiff >= 8) {
if (diffMVdiffVal < -24) {
if (ssDCval < 21) {
return 8;
}
else if (ssDCval >= 21) {
return 4;
}
}
else if (diffMVdiffVal >= -24) {
return 8;
}
}
}
}
else if (refDCval >= 50) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (posBalance >= 93) {
if (TSCindex < 3) {
if (MVDiff < 34) {
return 2;
}
else if (MVDiff >= 34) {
return 4;
}
}
else if (TSCindex >= 3) {
if (Rs < 385) {
if (TSC < 3) {
if (Cs < 65) {
if (Rs < 33) {
return 2;
}
else if (Rs >= 33) {
return 4;
}
}
else if (Cs >= 65) {
return 8;
}
}
else if (TSC >= 3) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (Rs >= 385) {
if (ssDCval < 91) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
return 4;
}
}
else if (ssDCval >= 91) {
return 8;
}
}
}
}
}
else if (TSC >= 7) {
if (posBalance < 117) {
if (ssDCval < 35) {
if (Rs < 411) {
return 1;
}
else if (Rs >= 411) {
return 2;
}
}
else if (ssDCval >= 35) {
return 8;
}
}
else if (posBalance >= 117) {
if (CsDiff < 3) {
if (Cs < 424) {
return 1;
}
else if (Cs >= 424) {
if (Rs < 668) {
return 4;
}
else if (Rs >= 668) {
return 8;
}
}
}
else if (CsDiff >= 3) {
return 4;
}
}
}
}
else if (negBalance >= 5) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (ssDCval < 111) {
if (posBalance < 67) {
if (Cs < 108) {
return 1;
}
else if (Cs >= 108) {
if (MVDiff < 28) {
if (MVDiff < 22) {
if (tcor < 93) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
return 8;
}
}
else if (tcor >= 93) {
if (Rs < 235) {
return 8;
}
else if (Rs >= 235) {
return 4;
}
}
}
else if (MVDiff >= 22) {
return 8;
}
}
else if (MVDiff >= 28) {
if (tcor < 78) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
return 1;
}
}
else if (tcor >= 78) {
return 4;
}
}
}
}
else if (posBalance >= 67) {
if (MVDiff < 16) {
if (posBalance < 88) {
return 8;
}
else if (posBalance >= 88) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 2;
}
}
}
else if (MVDiff >= 16) {
if (CsDiff < 1) {
if (negBalance < 6) {
if (diffRsCsdiff < -1) {
if (Rs < 106) {
return 4;
}
else if (Rs >= 106) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (Cs < 149) {
return 8;
}
else if (Cs >= 149) {
return 1;
}
}
}
else if (negBalance >= 6) {
if (diffMVdiffVal < 27) {
if (posBalance < 113) {
return 2;
}
else if (posBalance >= 113) {
return 4;
}
}
else if (diffMVdiffVal >= 27) {
if (Cs < 47) {
if (Rs < 1104) {
return 8;
}
else if (Rs >= 1104) {
return 1;
}
}
else if (Cs >= 47) {
return 4;
}
}
}
}
else if (CsDiff >= 1) {
if (posBalance < 109) {
if (mcTcor < 54) {
if (MVDiff < 136) {
return 2;
}
else if (MVDiff >= 136) {
return 1;
}
}
else if (mcTcor >= 54) {
return 1;
}
}
else if (posBalance >= 109) {
if (Cs < 307) {
return 8;
}
else if (Cs >= 307) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 111) {
if (MVDiff < 274) {
return 2;
}
else if (MVDiff >= 274) {
if (Rs < 1167) {
return 8;
}
else if (Rs >= 1167) {
return 1;
}
}
}
}
}
}
else if (tcor >= 94) {
if (negBalance < 6) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
if (posBalance < 14) {
return 8;
}
else if (posBalance >= 14) {
if (Rs < 19) {
return 8;
}
else if (Rs >= 19) {
return 1;
}
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (refDCval < 30) {
return 4;
}
else if (refDCval >= 30) {
if (SC < 943) {
if (ssDCval < 44) {
if (Rs < 500) {
return 8;
}
else if (Rs >= 500) {
return 4;
}
}
else if (ssDCval >= 44) {
if (Rs < 656) {
if (refDCval < 130) {
if (diffRsCsdiff < -1) {
if (Cs < 313) {
return 1;
}
else if (Cs >= 313) {
return 2;
}
}
else if (diffRsCsdiff >= -1) {
if (SC < 155) {
if (negBalance < 2) {
return 2;
}
else if (negBalance >= 2) {
return 4;
}
}
else if (SC >= 155) {
if (Cs < 133) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (Cs < 40) {
return 8;
}
else if (Cs >= 40) {
return 4;
}
}
}
else if (Cs >= 133) {
return 8;
}
}
}
}
else if (refDCval >= 130) {
if (Rs < 576) {
if (Rs < 155) {
return 1;
}
else if (Rs >= 155) {
return 4;
}
}
else if (Rs >= 576) {
return 8;
}
}
}
else if (Rs >= 656) {
return 4;
}
}
}
else if (SC >= 943) {
if (diffAFD < -7) {
if (ssDCval < 149) {
return 2;
}
else if (ssDCval >= 149) {
return 8;
}
}
else if (diffAFD >= -7) {
if (ssDCval < 119) {
if (Rs < 1806) {
return 8;
}
else if (Rs >= 1806) {
return 1;
}
}
else if (ssDCval >= 119) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 6) {
if (mu_mv_mag_sq < 70) {
if (Cs < 116) {
if (MVDiff < 17) {
if (mu_mv_mag_sq < 58) {
if (Rs < 304) {
return 1;
}
else if (Rs >= 304) {
return 8;
}
}
else if (mu_mv_mag_sq >= 58) {
return 1;
}
}
else if (MVDiff >= 17) {
if (ssDCval < 21) {
if (Rs < 17) {
return 8;
}
else if (Rs >= 17) {
return 1;
}
}
else if (ssDCval >= 21) {
if (AFD < 7) {
if (Cs < 93) {
return 4;
}
else if (Cs >= 93) {
return 8;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
}
else if (Cs >= 116) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (SC < 923) {
if (SC < 175) {
if (Rs < 7) {
if (ssDCval < 34) {
return 8;
}
else if (ssDCval >= 34) {
return 2;
}
}
else if (Rs >= 7) {
if (mu_mv_mag_sq < 103) {
return 1;
}
else if (mu_mv_mag_sq >= 103) {
return 8;
}
}
}
else if (SC >= 175) {
if (Rs < 88) {
return 4;
}
else if (Rs >= 88) {
if (mu_mv_mag_sq < 112) {
if (ssDCval < 133) {
if (SC < 334) {
if (negBalance < 9) {
if (Rs < 142) {
return 8;
}
else if (Rs >= 142) {
return 4;
}
}
else if (negBalance >= 9) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
return 4;
}
}
}
else if (SC >= 334) {
if (posBalance < 56) {
if (diffAFD < -4) {
if (Rs < 361) {
return 4;
}
else if (Rs >= 361) {
return 8;
}
}
else if (diffAFD >= -4) {
return 4;
}
}
else if (posBalance >= 56) {
if (Rs < 346) {
return 1;
}
else if (Rs >= 346) {
return 8;
}
}
}
}
else if (ssDCval >= 133) {
if (ssDCval < 161) {
return 1;
}
else if (ssDCval >= 161) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (MVDiff < 10) {
if (Cs < 403) {
return 1;
}
else if (Cs >= 403) {
return 8;
}
}
else if (MVDiff >= 10) {
return 8;
}
}
}
}
}
else if (SC >= 923) {
if (Cs < 397) {
return 1;
}
else if (Cs >= 397) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 35) {
if (posBalance < 27) {
if (MVDiff < 5) {
if (tcor < 102) {
if (Rs < 27) {
return 1;
}
else if (Rs >= 27) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
return 8;
}
}
}
else if (tcor >= 102) {
return 8;
}
}
else if (MVDiff >= 5) {
if (posBalance < 6) {
if (Cs < 2) {
if (negBalance < 80) {
if (ssDCval < 22) {
if (MVDiff < 247) {
return 8;
}
else if (MVDiff >= 247) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (ssDCval >= 22) {
return 4;
}
}
else if (negBalance >= 80) {
if (gchDC < 7) {
return 8;
}
else if (gchDC >= 7) {
return 4;
}
}
}
else if (Cs >= 2) {
return 4;
}
}
else if (posBalance >= 6) {
if (mu_mv_mag_sq < 91) {
if (ssDCval < 32) {
if (Rs < 19) {
if (negBalance < 58) {
if (refDCval < 23) {
return 8;
}
else if (refDCval >= 23) {
return 1;
}
}
else if (negBalance >= 58) {
return 8;
}
}
else if (Rs >= 19) {
if (CsDiff < 3) {
if (ssDCval < 31) {
if (mu_mv_mag_sq < 78) {
return 1;
}
else if (mu_mv_mag_sq >= 78) {
if (Cs < 55) {
return 1;
}
else if (Cs >= 55) {
return 4;
}
}
}
else if (ssDCval >= 31) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (CsDiff >= 3) {
return 2;
}
}
}
else if (ssDCval >= 32) {
if (Rs < 211) {
return 8;
}
else if (Rs >= 211) {
if (Rs < 345) {
return 1;
}
else if (Rs >= 345) {
if (Rs < 408) {
return 2;
}
else if (Rs >= 408) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (posBalance < 24) {
return 4;
}
else if (posBalance >= 24) {
if (Cs < 180) {
return 4;
}
else if (Cs >= 180) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 27) {
if (mcTcor < 81) {
if (ssDCval < 22) {
if (MVDiff < 40) {
return 4;
}
else if (MVDiff >= 40) {
return 8;
}
}
else if (ssDCval >= 22) {
if (CsDiff < 2) {
if (negBalance < 45) {
if (posBalance < 87) {
if (posBalance < 36) {
if (SC < 461) {
return 1;
}
else if (SC >= 461) {
if (Rs < 501) {
return 8;
}
else if (Rs >= 501) {
return 4;
}
}
}
else if (posBalance >= 36) {
if (diffMVdiffVal < -27) {
if (Rs < 89) {
return 1;
}
else if (Rs >= 89) {
return 8;
}
}
else if (diffMVdiffVal >= -27) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < -2) {
if (Rs < 226) {
return 2;
}
else if (Rs >= 226) {
return 1;
}
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
}
}
else if (posBalance >= 87) {
return 8;
}
}
else if (negBalance >= 45) {
if (tcor < 55) {
if (MVDiff < 309) {
return 8;
}
else if (MVDiff >= 309) {
return 4;
}
}
else if (tcor >= 55) {
if (Rs < 379) {
if (Rs < 249) {
return 1;
}
else if (Rs >= 249) {
return 8;
}
}
else if (Rs >= 379) {
return 1;
}
}
}
}
else if (CsDiff >= 2) {
if (TSC < 4) {
if (Rs < 455) {
if (MVDiff < 62) {
return 2;
}
else if (MVDiff >= 62) {
return 8;
}
}
else if (Rs >= 455) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (TSC >= 4) {
if (mcTcor < 54) {
return 8;
}
else if (mcTcor >= 54) {
if (Rs < 207) {
if (Rs < 121) {
return 1;
}
else if (Rs >= 121) {
return 8;
}
}
else if (Rs >= 207) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 81) {
if (ssDCval < 30) {
if (MVDiff < 60) {
if (Cs < 374) {
if (mu_mv_mag_sq < 43) {
if (posBalance < 42) {
if (ssDCval < 26) {
return 1;
}
else if (ssDCval >= 26) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
if (mu_mv_mag_sq < 36) {
if (Rs < 64) {
return 1;
}
else if (Rs >= 64) {
return 4;
}
}
else if (mu_mv_mag_sq >= 36) {
if (Rs < 230) {
return 1;
}
else if (Rs >= 230) {
return 4;
}
}
}
}
}
else if (posBalance >= 42) {
if (Cs < 323) {
if (posBalance < 71) {
if (tcor < 54) {
if (Rs < 73) {
return 1;
}
else if (Rs >= 73) {
return 8;
}
}
else if (tcor >= 54) {
return 1;
}
}
else if (posBalance >= 71) {
return 2;
}
}
else if (Cs >= 323) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (diffMVdiffVal < -5) {
if (Cs < 173) {
return 1;
}
else if (Cs >= 173) {
if (ssDCval < 28) {
return 1;
}
else if (ssDCval >= 28) {
if (Rs < 225) {
return 1;
}
else if (Rs >= 225) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= -5) {
if (refDCval < 30) {
return 1;
}
else if (refDCval >= 30) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
return 1;
}
}
}
}
}
else if (Cs >= 374) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (MVDiff >= 60) {
if (MVDiff < 101) {
if (Rs < 45) {
return 1;
}
else if (Rs >= 45) {
return 8;
}
}
else if (MVDiff >= 101) {
return 1;
}
}
}
else if (ssDCval >= 30) {
if (RsCsDiff < 3) {
if (Cs < 260) {
if (Cs < 84) {
if (Cs < 11) {
if (ssDCval < 32) {
if (posBalance < 66) {
return 1;
}
else if (posBalance >= 66) {
if (Rs < 9) {
return 4;
}
else if (Rs >= 9) {
return 2;
}
}
}
else if (ssDCval >= 32) {
return 2;
}
}
else if (Cs >= 11) {
if (MVDiff < 3) {
if (Cs < 35) {
return 1;
}
else if (Cs >= 35) {
return 4;
}
}
else if (MVDiff >= 3) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (mcTcor < 98) {
if (refDCval < 31) {
if (MVDiff < 25) {
return 1;
}
else if (MVDiff >= 25) {
return 4;
}
}
else if (refDCval >= 31) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 1;
}
}
}
else if (mcTcor >= 98) {
return 4;
}
}
}
}
}
else if (Cs >= 84) {
if (mcTcor < 95) {
if (posBalance < 56) {
if (AFD < 3) {
if (MVDiff < 25) {
return 8;
}
else if (MVDiff >= 25) {
return 1;
}
}
else if (AFD >= 3) {
if (RsCsDiff < 2) {
if (posBalance < 38) {
if (Rs < 184) {
return 8;
}
else if (Rs >= 184) {
return 4;
}
}
else if (posBalance >= 38) {
return 1;
}
}
else if (RsCsDiff >= 2) {
if (gchDC < 2) {
if (SC < 455) {
if (mu_mv_mag_sq < 62) {
if (Rs < 134) {
return 8;
}
else if (Rs >= 134) {
return 2;
}
}
else if (mu_mv_mag_sq >= 62) {
return 8;
}
}
else if (SC >= 455) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (Rs < 311) {
return 1;
}
else if (Rs >= 311) {
return 8;
}
}
}
}
}
else if (gchDC >= 2) {
if (SC < 341) {
return 4;
}
else if (SC >= 341) {
return 1;
}
}
}
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (mcTcor >= 95) {
if (SC < 598) {
if (Rs < 108) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Rs >= 108) {
if (ssDCval < 34) {
if (mcTcor < 96) {
if (posBalance < 37) {
if (negBalance < 22) {
return 8;
}
else if (negBalance >= 22) {
if (Rs < 164) {
return 1;
}
else if (Rs >= 164) {
return 4;
}
}
}
else if (posBalance >= 37) {
return 1;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (ssDCval >= 34) {
if (Cs < 159) {
return 1;
}
else if (Cs >= 159) {
return 8;
}
}
}
}
else if (SC >= 598) {
if (Cs < 149) {
return 1;
}
else if (Cs >= 149) {
return 4;
}
}
}
}
}
else if (Cs >= 260) {
if (negBalance < 16) {
if (Rs < 287) {
return 2;
}
else if (Rs >= 287) {
return 4;
}
}
else if (negBalance >= 16) {
if (mcTcor < 89) {
if (Rs < 252) {
return 1;
}
else if (Rs >= 252) {
return 8;
}
}
else if (mcTcor >= 89) {
return 1;
}
}
}
}
else if (RsCsDiff >= 3) {
if (TSCindex < 3) {
if (Rs < 427) {
if (Rs < 258) {
return 1;
}
else if (Rs >= 258) {
return 2;
}
}
else if (Rs >= 427) {
return 4;
}
}
else if (TSCindex >= 3) {
if (CsDiff < 2) {
if (posBalance < 32) {
return 4;
}
else if (posBalance >= 32) {
return 1;
}
}
else if (CsDiff >= 2) {
if (negBalance < 33) {
if (negBalance < 32) {
if (Rs < 821) {
return 1;
}
else if (Rs >= 821) {
return 8;
}
}
else if (negBalance >= 32) {
return 8;
}
}
else if (negBalance >= 33) {
if (CsDiff < 3) {
return 1;
}
else if (CsDiff >= 3) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 35) {
if (Rs < 304) {
if (ssDCval < 50) {
if (mu_mv_mag_sq < 52) {
if (mcTcor < 89) {
if (Cs < 88) {
if (SC < 14) {
return 8;
}
else if (SC >= 14) {
if (AFD < 8) {
if (Rs < 194) {
if (posBalance < 51) {
return 2;
}
else if (posBalance >= 51) {
if (Rs < 34) {
return 8;
}
else if (Rs >= 34) {
return 1;
}
}
}
else if (Rs >= 194) {
return 1;
}
}
else if (AFD >= 8) {
if (refDCval < 45) {
return 1;
}
else if (refDCval >= 45) {
if (posBalance < 51) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Rs < 22) {
return 1;
}
else if (Rs >= 22) {
return 8;
}
}
}
else if (posBalance >= 51) {
if (Rs < 64) {
return 2;
}
else if (Rs >= 64) {
return 4;
}
}
}
}
}
}
else if (Cs >= 88) {
if (posBalance < 48) {
if (CsDiff < 4) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
return 8;
}
}
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (posBalance >= 48) {
if (AFD < 9) {
if (Cs < 276) {
if (Cs < 191) {
return 8;
}
else if (Cs >= 191) {
if (negBalance < 36) {
return 2;
}
else if (negBalance >= 36) {
if (posBalance < 52) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
else if (posBalance >= 52) {
return 8;
}
}
}
}
else if (Cs >= 276) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (AFD >= 9) {
if (negBalance < 44) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (mu_mv_mag_sq < 28) {
return 4;
}
else if (mu_mv_mag_sq >= 28) {
if (Cs < 224) {
return 1;
}
else if (Cs >= 224) {
return 8;
}
}
}
}
else if (negBalance >= 44) {
if (RsCsDiff < 3) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (RsCsDiff >= 3) {
if (mu_mv_mag_sq < 2) {
if (Rs < 279) {
return 4;
}
else if (Rs >= 279) {
return 8;
}
}
else if (mu_mv_mag_sq >= 2) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 89) {
if (posBalance < 8) {
if (posBalance < 1) {
if (AFD < 8) {
if (Rs < 2) {
return 2;
}
else if (Rs >= 2) {
return 4;
}
}
else if (AFD >= 8) {
return 8;
}
}
else if (posBalance >= 1) {
return 8;
}
}
else if (posBalance >= 8) {
if (tcor < 93) {
if (posBalance < 68) {
if (TSC < 4) {
if (SC < 31) {
return 8;
}
else if (SC >= 31) {
if (ssDCval < 37) {
if (mu_mv_mag_sq < 37) {
if (Cs < 241) {
if (refDCval < 38) {
return 1;
}
else if (refDCval >= 38) {
return 4;
}
}
else if (Cs >= 241) {
if (Rs < 214) {
return 8;
}
else if (Rs >= 214) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 161) {
return 2;
}
else if (Rs >= 161) {
return 4;
}
}
}
else if (ssDCval >= 37) {
if (Cs < 342) {
if (SC < 466) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
if (Cs < 205) {
if (Cs < 161) {
if (Cs < 39) {
if (Rs < 161) {
return 2;
}
else if (Rs >= 161) {
return 1;
}
}
else if (Cs >= 39) {
if (MVDiff < 44) {
if (diffAFD < 2) {
if (SC < 317) {
if (refDCval < 37) {
return 4;
}
else if (refDCval >= 37) {
return 1;
}
}
else if (SC >= 317) {
if (Rs < 176) {
return 2;
}
else if (Rs >= 176) {
return 1;
}
}
}
else if (diffAFD >= 2) {
if (MVDiff < 29) {
return 2;
}
else if (MVDiff >= 29) {
if (Cs < 153) {
return 4;
}
else if (Cs >= 153) {
return 2;
}
}
}
}
else if (MVDiff >= 44) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 4;
}
}
}
}
else if (Cs >= 161) {
if (AFD < 5) {
return 2;
}
else if (AFD >= 5) {
if (Rs < 41) {
return 1;
}
else if (Rs >= 41) {
return 4;
}
}
}
}
else if (Cs >= 205) {
if (Rs < 181) {
return 1;
}
else if (Rs >= 181) {
return 2;
}
}
}
}
else if (SC >= 466) {
if (ssDCval < 44) {
if (AFD < 8) {
return 2;
}
else if (AFD >= 8) {
return 1;
}
}
else if (ssDCval >= 44) {
if (Rs < 299) {
return 4;
}
else if (Rs >= 299) {
return 2;
}
}
}
}
else if (Cs >= 342) {
return 1;
}
}
}
}
else if (TSC >= 4) {
if (SC < 342) {
if (ssDCval < 47) {
if (CsDiff < 3) {
if (mcTcor < 90) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
else if (mcTcor >= 90) {
return 1;
}
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (ssDCval >= 47) {
return 2;
}
}
else if (SC >= 342) {
if (refDCval < 44) {
if (mu_mv_mag_sq < 15) {
return 8;
}
else if (mu_mv_mag_sq >= 15) {
if (TSC < 5) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 4;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (refDCval >= 44) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (Cs < 186) {
return 1;
}
else if (Cs >= 186) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 68) {
return 2;
}
}
else if (tcor >= 93) {
if (SC < 709) {
if (diffRsCsdiff < -2) {
if (Rs < 131) {
return 1;
}
else if (Rs >= 131) {
return 2;
}
}
else if (diffRsCsdiff >= -2) {
if (MVDiff < 37) {
return 1;
}
else if (MVDiff >= 37) {
if (mu_mv_mag_sq < 12) {
return 1;
}
else if (mu_mv_mag_sq >= 12) {
return 4;
}
}
}
}
else if (SC >= 709) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (negBalance < 31) {
if (mu_mv_mag_sq < 111) {
if (Cs < 237) {
if (ssDCval < 45) {
if (ssDCval < 40) {
if (posBalance < 30) {
return 4;
}
else if (posBalance >= 30) {
if (MVDiff < 7) {
if (Cs < 225) {
return 4;
}
else if (Cs >= 225) {
return 8;
}
}
else if (MVDiff >= 7) {
if (Cs < 184) {
return 1;
}
else if (Cs >= 184) {
return 8;
}
}
}
}
else if (ssDCval >= 40) {
if (negBalance < 27) {
return 1;
}
else if (negBalance >= 27) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
}
}
else if (ssDCval >= 45) {
if (mu_mv_mag_sq < 92) {
if (Rs < 129) {
return 2;
}
else if (Rs >= 129) {
return 4;
}
}
else if (mu_mv_mag_sq >= 92) {
return 8;
}
}
}
else if (Cs >= 237) {
if (SC < 532) {
if (ssDCval < 39) {
if (tcor < 84) {
return 1;
}
else if (tcor >= 84) {
if (Rs < 123) {
return 1;
}
else if (Rs >= 123) {
return 8;
}
}
}
else if (ssDCval >= 39) {
return 1;
}
}
else if (SC >= 532) {
if (refDCval < 50) {
return 1;
}
else if (refDCval >= 50) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (negBalance < 15) {
if (Cs < 134) {
if (Rs < 38) {
return 8;
}
else if (Rs >= 38) {
return 4;
}
}
else if (Cs >= 134) {
if (ssDCval < 37) {
return 8;
}
else if (ssDCval >= 37) {
return 1;
}
}
}
else if (negBalance >= 15) {
if (Rs < 186) {
if (posBalance < 61) {
if (ssDCval < 36) {
if (Cs < 197) {
return 1;
}
else if (Cs >= 197) {
return 8;
}
}
else if (ssDCval >= 36) {
return 8;
}
}
else if (posBalance >= 61) {
return 4;
}
}
else if (Rs >= 186) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 1;
}
}
}
}
}
else if (negBalance >= 31) {
if (gchDC < 1) {
if (SC < 202) {
if (Rs < 47) {
return 8;
}
else if (Rs >= 47) {
return 1;
}
}
else if (SC >= 202) {
return 8;
}
}
else if (gchDC >= 1) {
if (SC < 167) {
if (diffAFD < 1) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 78) {
if (Cs < 74) {
return 8;
}
else if (Cs >= 74) {
return 4;
}
}
else if (mu_mv_mag_sq >= 78) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 2;
}
}
}
}
else if (diffAFD >= 1) {
if (ssDCval < 49) {
if (Rs < 135) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 2;
}
}
else if (Rs >= 135) {
return 2;
}
}
else if (ssDCval >= 49) {
return 2;
}
}
}
else if (SC >= 167) {
if (AFD < 6) {
if (refDCval < 41) {
if (Cs < 95) {
return 1;
}
else if (Cs >= 95) {
return 8;
}
}
else if (refDCval >= 41) {
if (posBalance < 26) {
if (Rs < 185) {
return 8;
}
else if (Rs >= 185) {
return 4;
}
}
else if (posBalance >= 26) {
if (mu_mv_mag_sq < 81) {
return 1;
}
else if (mu_mv_mag_sq >= 81) {
if (ssDCval < 45) {
if (Rs < 158) {
return 4;
}
else if (Rs >= 158) {
return 1;
}
}
else if (ssDCval >= 45) {
return 8;
}
}
}
}
}
else if (AFD >= 6) {
if (negBalance < 51) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
return 4;
}
}
else if (negBalance >= 51) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 50) {
if (TSCindex < 2) {
if (Cs < 250) {
if (mu_mv_mag_sq < 114) {
if (TSCindex < 1) {
return 4;
}
else if (TSCindex >= 1) {
if (Cs < 115) {
if (posBalance < 38) {
if (Cs < 40) {
if (MVDiff < 395) {
return 2;
}
else if (MVDiff >= 395) {
return 1;
}
}
else if (Cs >= 40) {
if (Cs < 110) {
return 8;
}
else if (Cs >= 110) {
return 2;
}
}
}
else if (posBalance >= 38) {
if (ssDCval < 118) {
if (SC < 223) {
if (Rs < 81) {
return 1;
}
else if (Rs >= 81) {
return 2;
}
}
else if (SC >= 223) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
return 1;
}
}
}
else if (ssDCval >= 118) {
return 4;
}
}
}
else if (Cs >= 115) {
if (Rs < 133) {
if (ssDCval < 84) {
if (ssDCval < 65) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
return 4;
}
}
else if (ssDCval >= 65) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
return 2;
}
}
}
else if (ssDCval >= 84) {
return 4;
}
}
else if (Rs >= 133) {
if (ssDCval < 63) {
if (MVDiff < 2) {
if (Cs < 125) {
return 8;
}
else if (Cs >= 125) {
return 1;
}
}
else if (MVDiff >= 2) {
if (ssDCval < 54) {
if (Rs < 223) {
return 1;
}
else if (Rs >= 223) {
return 2;
}
}
else if (ssDCval >= 54) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
if (MVDiff < 4) {
if (Rs < 165) {
return 1;
}
else if (Rs >= 165) {
return 4;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
}
else if (ssDCval >= 63) {
if (Rs < 268) {
if (SC < 300) {
if (AFD < 2) {
if (posBalance < 45) {
if (Rs < 147) {
return 1;
}
else if (Rs >= 147) {
return 2;
}
}
else if (posBalance >= 45) {
return 2;
}
}
else if (AFD >= 2) {
return 2;
}
}
else if (SC >= 300) {
if (Cs < 147) {
if (Rs < 203) {
return 2;
}
else if (Rs >= 203) {
return 8;
}
}
else if (Cs >= 147) {
return 2;
}
}
}
else if (Rs >= 268) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
if (SC < 251) {
if (Cs < 99) {
if (Rs < 77) {
return 8;
}
else if (Rs >= 77) {
return 1;
}
}
else if (Cs >= 99) {
return 4;
}
}
else if (SC >= 251) {
if (ssDCval < 66) {
if (mu_mv_mag_sq < 115) {
return 1;
}
else if (mu_mv_mag_sq >= 115) {
return 4;
}
}
else if (ssDCval >= 66) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
if (Rs < 261) {
if (Rs < 206) {
return 2;
}
else if (Rs >= 206) {
return 8;
}
}
else if (Rs >= 261) {
return 1;
}
}
}
}
}
}
else if (Cs >= 250) {
if (Rs < 159) {
if (Rs < 132) {
return 4;
}
else if (Rs >= 132) {
if (AFD < 2) {
if (mu_mv_mag_sq < 103) {
if (ssDCval < 84) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 1;
}
}
else if (ssDCval >= 84) {
return 4;
}
}
else if (mu_mv_mag_sq >= 103) {
return 4;
}
}
else if (AFD >= 2) {
return 2;
}
}
}
else if (Rs >= 159) {
if (Rs < 249) {
if (ssDCval < 68) {
if (refDCval < 58) {
return 1;
}
else if (refDCval >= 58) {
return 4;
}
}
else if (ssDCval >= 68) {
if (ssDCval < 78) {
if (Rs < 183) {
if (Cs < 295) {
return 1;
}
else if (Cs >= 295) {
return 4;
}
}
else if (Rs >= 183) {
return 1;
}
}
else if (ssDCval >= 78) {
return 4;
}
}
}
else if (Rs >= 249) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
return 4;
}
}
}
}
}
else if (TSCindex >= 2) {
if (refDCval < 72) {
if (Cs < 223) {
if (tcor < 101) {
if (RsCsDiff < 1) {
if (MVDiff < 2) {
if (refDCval < 64) {
if (Rs < 81) {
return 2;
}
else if (Rs >= 81) {
if (TSCindex < 3) {
if (Cs < 202) {
return 1;
}
else if (Cs >= 202) {
return 4;
}
}
else if (TSCindex >= 3) {
if (Rs < 139) {
if (Cs < 132) {
return 8;
}
else if (Cs >= 132) {
return 2;
}
}
else if (Rs >= 139) {
return 1;
}
}
}
}
else if (refDCval >= 64) {
if (SC < 279) {
if (ssDCval < 65) {
return 2;
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (SC >= 279) {
if (ssDCval < 65) {
if (Rs < 175) {
return 2;
}
else if (Rs >= 175) {
return 4;
}
}
else if (ssDCval >= 65) {
return 2;
}
}
}
}
else if (MVDiff >= 2) {
if (mcTcor < 97) {
if (Cs < 50) {
if (Rs < 38) {
if (MVDiff < 45) {
if (Rs < 9) {
return 1;
}
else if (Rs >= 9) {
return 8;
}
}
else if (MVDiff >= 45) {
return 4;
}
}
else if (Rs >= 38) {
return 2;
}
}
else if (Cs >= 50) {
if (diffRsCsdiff < -1) {
if (Rs < 161) {
if (Cs < 94) {
return 2;
}
else if (Cs >= 94) {
return 4;
}
}
else if (Rs >= 161) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (mu_mv_mag_sq < 41) {
if (MVDiff < 67) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 1;
}
}
}
else if (MVDiff >= 67) {
if (Rs < 36) {
return 4;
}
else if (Rs >= 36) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 41) {
if (MVDiff < 8) {
if (Cs < 168) {
return 1;
}
else if (Cs >= 168) {
return 4;
}
}
else if (MVDiff >= 8) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 68) {
if (negBalance < 44) {
return 1;
}
else if (negBalance >= 44) {
if (Cs < 59) {
return 1;
}
else if (Cs >= 59) {
return 4;
}
}
}
else if (Rs >= 68) {
if (MVDiff < 15) {
if (ssDCval < 66) {
if (Rs < 209) {
if (negBalance < 29) {
if (Cs < 189) {
if (negBalance < 23) {
if (Cs < 153) {
return 4;
}
else if (Cs >= 153) {
if (Rs < 148) {
return 4;
}
else if (Rs >= 148) {
return 1;
}
}
}
else if (negBalance >= 23) {
if (Cs < 180) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
else if (Cs >= 180) {
return 4;
}
}
}
else if (Cs >= 189) {
return 4;
}
}
else if (negBalance >= 29) {
if (tcor < 95) {
if (mu_mv_mag_sq < 45) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
if (Cs < 166) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
}
else if (Cs >= 166) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (tcor >= 95) {
return 1;
}
}
}
else if (Rs >= 209) {
if (Cs < 204) {
return 4;
}
else if (Cs >= 204) {
if (Cs < 215) {
return 1;
}
else if (Cs >= 215) {
return 4;
}
}
}
}
else if (ssDCval >= 66) {
if (AFD < 2) {
return 2;
}
else if (AFD >= 2) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
else if (MVDiff >= 15) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 1) {
if (TSC < 5) {
if (mcTcor < 88) {
if (TSC < 4) {
if (refDCval < 60) {
if (Cs < 65) {
return 2;
}
else if (Cs >= 65) {
return 8;
}
}
else if (refDCval >= 60) {
return 1;
}
}
else if (TSC >= 4) {
return 2;
}
}
else if (mcTcor >= 88) {
if (negBalance < 29) {
if (negBalance < 21) {
return 1;
}
else if (negBalance >= 21) {
if (RsCsDiff < 2) {
if (Rs < 132) {
return 1;
}
else if (Rs >= 132) {
return 4;
}
}
else if (RsCsDiff >= 2) {
if (refDCval < 70) {
return 8;
}
else if (refDCval >= 70) {
return 1;
}
}
}
}
else if (negBalance >= 29) {
if (mu_mv_mag_sq < 14) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (MVDiff < 54) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (mcTcor < 94) {
if (Cs < 160) {
if (Rs < 174) {
return 2;
}
else if (Rs >= 174) {
return 4;
}
}
else if (Cs >= 160) {
return 1;
}
}
else if (mcTcor >= 94) {
if (Rs < 130) {
return 4;
}
else if (Rs >= 130) {
if (Rs < 208) {
if (ssDCval < 64) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
return 4;
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (Rs >= 208) {
return 4;
}
}
}
}
}
else if (MVDiff >= 54) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (refDCval < 67) {
if (SC < 122) {
return 1;
}
else if (SC >= 122) {
if (posBalance < 38) {
if (Rs < 199) {
return 8;
}
else if (Rs >= 199) {
if (Rs < 273) {
return 1;
}
else if (Rs >= 273) {
return 4;
}
}
}
else if (posBalance >= 38) {
if (SC < 504) {
if (mu_mv_mag_sq < 79) {
return 4;
}
else if (mu_mv_mag_sq >= 79) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 1;
}
}
}
else if (SC >= 504) {
return 2;
}
}
}
}
else if (refDCval >= 67) {
if (RsDiff < 1) {
if (Rs < 92) {
return 4;
}
else if (Rs >= 92) {
return 1;
}
}
else if (RsDiff >= 1) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 4;
}
}
}
}
}
}
}
else if (TSC >= 5) {
if (AFD < 14) {
if (MVDiff < 42) {
return 1;
}
else if (MVDiff >= 42) {
if (Cs < 92) {
if (Rs < 123) {
return 2;
}
else if (Rs >= 123) {
return 1;
}
}
else if (Cs >= 92) {
if (Rs < 108) {
if (Cs < 144) {
return 2;
}
else if (Cs >= 144) {
return 1;
}
}
else if (Rs >= 108) {
return 2;
}
}
}
}
else if (AFD >= 14) {
if (MVDiff < 73) {
if (Cs < 179) {
return 4;
}
else if (Cs >= 179) {
if (MVDiff < 57) {
if (Rs < 215) {
return 1;
}
else if (Rs >= 215) {
return 8;
}
}
else if (MVDiff >= 57) {
return 2;
}
}
}
else if (MVDiff >= 73) {
if (diffMVdiffVal < -62) {
if (Cs < 86) {
return 8;
}
else if (Cs >= 86) {
if (Rs < 108) {
return 2;
}
else if (Rs >= 108) {
return 1;
}
}
}
else if (diffMVdiffVal >= -62) {
if (ssDCval < 55) {
if (SC < 215) {
if (Rs < 136) {
return 4;
}
else if (Rs >= 136) {
return 1;
}
}
else if (SC >= 215) {
return 1;
}
}
else if (ssDCval >= 55) {
if (ssDCval < 71) {
if (SC < 328) {
return 1;
}
else if (SC >= 328) {
if (Rs < 203) {
return 1;
}
else if (Rs >= 203) {
return 8;
}
}
}
else if (ssDCval >= 71) {
if (Rs < 97) {
if (Rs < 84) {
return 2;
}
else if (Rs >= 84) {
return 4;
}
}
else if (Rs >= 97) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (tcor >= 101) {
if (SC < 120) {
return 2;
}
else if (SC >= 120) {
return 8;
}
}
}
else if (Cs >= 223) {
if (MVDiff < 4) {
if (Rs < 184) {
if (TSCindex < 3) {
if (Cs < 339) {
if (Rs < 180) {
if (SC < 378) {
if (Rs < 127) {
return 1;
}
else if (Rs >= 127) {
return 4;
}
}
else if (SC >= 378) {
return 4;
}
}
else if (Rs >= 180) {
if (Rs < 181) {
return 1;
}
else if (Rs >= 181) {
return 4;
}
}
}
else if (Cs >= 339) {
if (Rs < 163) {
return 4;
}
else if (Rs >= 163) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Rs < 165) {
return 1;
}
else if (Rs >= 165) {
return 4;
}
}
}
}
}
else if (TSCindex >= 3) {
if (negBalance < 44) {
if (Rs < 136) {
return 4;
}
else if (Rs >= 136) {
if (negBalance < 34) {
if (Cs < 338) {
if (mu_mv_mag_sq < 105) {
if (Rs < 143) {
return 1;
}
else if (Rs >= 143) {
if (posBalance < 52) {
if (Cs < 268) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 268) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 1;
}
}
}
else if (posBalance >= 52) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
return 4;
}
}
else if (Cs >= 338) {
if (posBalance < 52) {
if (Rs < 182) {
if (tcor < 99) {
return 1;
}
else if (tcor >= 99) {
return 4;
}
}
else if (Rs >= 182) {
return 4;
}
}
else if (posBalance >= 52) {
return 4;
}
}
}
else if (negBalance >= 34) {
if (Cs < 386) {
return 1;
}
else if (Cs >= 386) {
return 4;
}
}
}
}
else if (negBalance >= 44) {
if (Cs < 466) {
return 4;
}
else if (Cs >= 466) {
return 8;
}
}
}
}
else if (Rs >= 184) {
if (negBalance < 28) {
if (ssDCval < 60) {
if (MVDiff < 1) {
if (diffTSC < 1) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 237) {
return 1;
}
else if (Rs >= 237) {
return 4;
}
}
}
else if (diffTSC >= 1) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 60) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 122) {
return 4;
}
else if (mu_mv_mag_sq >= 122) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
return 4;
}
}
}
else if (ssDCval >= 69) {
return 1;
}
}
}
else if (negBalance >= 28) {
if (tcor < 98) {
if (Cs < 425) {
if (Cs < 315) {
if (SC < 461) {
return 4;
}
else if (SC >= 461) {
return 1;
}
}
else if (Cs >= 315) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (negBalance < 31) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
if (Rs < 188) {
return 1;
}
else if (Rs >= 188) {
return 4;
}
}
}
else if (negBalance >= 31) {
return 4;
}
}
}
}
else if (Cs >= 425) {
return 1;
}
}
else if (tcor >= 98) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
if (diffTSC < -1) {
if (Rs < 235) {
if (negBalance < 32) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (negBalance >= 32) {
return 1;
}
}
else if (Rs >= 235) {
return 1;
}
}
else if (diffTSC >= -1) {
if (Cs < 497) {
if (negBalance < 30) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 1;
}
}
}
else if (negBalance >= 30) {
return 1;
}
}
else if (Cs >= 497) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 4) {
if (TSC < 4) {
if (Cs < 306) {
if (refDCval < 62) {
if (Rs < 234) {
if (ssDCval < 55) {
if (tcor < 83) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
return 2;
}
}
else if (tcor >= 83) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 2;
}
}
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (Rs >= 234) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (SC < 504) {
return 1;
}
else if (SC >= 504) {
return 2;
}
}
}
}
else if (refDCval >= 62) {
if (tcor < 97) {
if (Rs < 81) {
return 1;
}
else if (Rs >= 81) {
if (Cs < 256) {
return 4;
}
else if (Cs >= 256) {
if (ssDCval < 66) {
if (diffMVdiffVal < 4) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
return 4;
}
}
else if (diffMVdiffVal >= 4) {
return 4;
}
}
else if (ssDCval >= 66) {
return 4;
}
}
}
}
else if (tcor >= 97) {
if (posBalance < 43) {
if (Rs < 200) {
if (Rs < 148) {
return 2;
}
else if (Rs >= 148) {
return 1;
}
}
else if (Rs >= 200) {
return 4;
}
}
else if (posBalance >= 43) {
return 1;
}
}
}
}
else if (Cs >= 306) {
if (tcor < 89) {
if (RsCsDiff < 7) {
if (Rs < 249) {
return 4;
}
else if (Rs >= 249) {
return 2;
}
}
else if (RsCsDiff >= 7) {
return 1;
}
}
else if (tcor >= 89) {
if (negBalance < 21) {
if (TSC < 2) {
return 2;
}
else if (TSC >= 2) {
return 4;
}
}
else if (negBalance >= 21) {
if (ssDCval < 51) {
if (Rs < 187) {
return 1;
}
else if (Rs >= 187) {
if (Rs < 239) {
return 4;
}
else if (Rs >= 239) {
return 2;
}
}
}
else if (ssDCval >= 51) {
if (AFD < 6) {
if (RsDiff < 2) {
if (TSC < 2) {
if (mu_mv_mag_sq < 80) {
return 2;
}
else if (mu_mv_mag_sq >= 80) {
return 4;
}
}
else if (TSC >= 2) {
return 4;
}
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (AFD >= 6) {
if (refDCval < 58) {
if (Rs < 252) {
return 8;
}
else if (Rs >= 252) {
return 2;
}
}
else if (refDCval >= 58) {
if (MVDiff < 21) {
if (Cs < 482) {
return 4;
}
else if (Cs >= 482) {
return 1;
}
}
else if (MVDiff >= 21) {
return 1;
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (mcTcor < 75) {
if (diffAFD < 5) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (diffAFD >= 5) {
if (Rs < 198) {
return 1;
}
else if (Rs >= 198) {
return 2;
}
}
}
else if (mcTcor >= 75) {
if (mu_mv_mag_sq < 38) {
if (diffAFD < 4) {
if (diffAFD < -2) {
if (Cs < 312) {
if (diffRsCsdiff < -33) {
return 8;
}
else if (diffRsCsdiff >= -33) {
return 2;
}
}
else if (Cs >= 312) {
if (Cs < 372) {
return 1;
}
else if (Cs >= 372) {
return 2;
}
}
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < 5) {
if (refDCval < 58) {
return 1;
}
else if (refDCval >= 58) {
if (mcTcor < 96) {
if (CsDiff < 2) {
if (SC < 482) {
return 1;
}
else if (SC >= 482) {
return 2;
}
}
else if (CsDiff >= 2) {
if (mcTcor < 81) {
return 2;
}
else if (mcTcor >= 81) {
return 1;
}
}
}
else if (mcTcor >= 96) {
return 2;
}
}
}
else if (diffRsCsdiff >= 5) {
return 2;
}
}
}
else if (diffAFD >= 4) {
if (Cs < 461) {
if (Rs < 234) {
return 8;
}
else if (Rs >= 234) {
return 4;
}
}
else if (Cs >= 461) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 38) {
if (diffRsCsdiff < -1) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
if (Rs < 231) {
return 4;
}
else if (Rs >= 231) {
return 8;
}
}
}
else if (diffRsCsdiff >= -1) {
if (mu_mv_mag_sq < 76) {
if (SC < 520) {
return 4;
}
else if (SC >= 520) {
if (posBalance < 12) {
return 4;
}
else if (posBalance >= 12) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (Cs < 266) {
return 8;
}
else if (Cs >= 266) {
return 1;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 72) {
if (mcTcor < 97) {
if (MVDiff < 30) {
if (Rs < 89) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (SC < 240) {
return 1;
}
else if (SC >= 240) {
return 4;
}
}
}
}
else if (Rs >= 89) {
if (refDCval < 85) {
if (MVDiff < 6) {
return 8;
}
else if (MVDiff >= 6) {
if (Cs < 339) {
if (gchDC < 4) {
if (ssDCval < 85) {
if (Rs < 130) {
if (Rs < 121) {
return 1;
}
else if (Rs >= 121) {
return 4;
}
}
else if (Rs >= 130) {
return 1;
}
}
else if (ssDCval >= 85) {
if (Rs < 277) {
return 8;
}
else if (Rs >= 277) {
return 1;
}
}
}
else if (gchDC >= 4) {
return 8;
}
}
else if (Cs >= 339) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (Rs < 241) {
return 4;
}
else if (Rs >= 241) {
return 2;
}
}
}
}
}
else if (refDCval >= 85) {
if (AFD < 4) {
if (Cs < 164) {
return 1;
}
else if (Cs >= 164) {
return 4;
}
}
else if (AFD >= 4) {
if (AFD < 9) {
if (Rs < 239) {
if (Cs < 80) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 161) {
return 4;
}
else if (Rs >= 161) {
return 1;
}
}
}
else if (Cs >= 80) {
if (gchDC < 1) {
if (SC < 334) {
return 8;
}
else if (SC >= 334) {
if (Cs < 350) {
if (Rs < 213) {
if (Cs < 291) {
if (MVDiff < 4) {
if (tcor < 66) {
return 8;
}
else if (tcor >= 66) {
if (Rs < 184) {
return 2;
}
else if (Rs >= 184) {
return 8;
}
}
}
else if (MVDiff >= 4) {
return 2;
}
}
else if (Cs >= 291) {
return 8;
}
}
else if (Rs >= 213) {
return 2;
}
}
else if (Cs >= 350) {
return 8;
}
}
}
else if (gchDC >= 1) {
return 2;
}
}
}
else if (Rs >= 239) {
if (Cs < 295) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
return 1;
}
}
else if (Cs >= 295) {
if (Rs < 262) {
return 8;
}
else if (Rs >= 262) {
if (TSC < 5) {
if (diffMVdiffVal < 2) {
if (Rs < 279) {
if (Cs < 323) {
return 8;
}
else if (Cs >= 323) {
return 2;
}
}
else if (Rs >= 279) {
return 8;
}
}
else if (diffMVdiffVal >= 2) {
return 2;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (AFD >= 9) {
if (AFD < 15) {
if (TSC < 5) {
if (tcor < 77) {
if (diffRsCsdiff < 2) {
return 8;
}
else if (diffRsCsdiff >= 2) {
if (Rs < 161) {
return 8;
}
else if (Rs >= 161) {
return 2;
}
}
}
else if (tcor >= 77) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (TSC >= 5) {
if (tcor < 84) {
if (posBalance < 70) {
return 1;
}
else if (posBalance >= 70) {
return 8;
}
}
else if (tcor >= 84) {
if (Cs < 434) {
return 8;
}
else if (Cs >= 434) {
return 1;
}
}
}
}
else if (AFD >= 15) {
if (posBalance < 55) {
if (Rs < 179) {
return 1;
}
else if (Rs >= 179) {
return 2;
}
}
else if (posBalance >= 55) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 30) {
if (mcTcor < 72) {
if (posBalance < 37) {
if (ssDCval < 186) {
return 1;
}
else if (ssDCval >= 186) {
if (Rs < 2) {
return 1;
}
else if (Rs >= 2) {
return 2;
}
}
}
else if (posBalance >= 37) {
if (Cs < 118) {
return 4;
}
else if (Cs >= 118) {
if (tcor < 3) {
return 8;
}
else if (tcor >= 3) {
if (MVDiff < 172) {
return 1;
}
else if (MVDiff >= 172) {
return 4;
}
}
}
}
}
else if (mcTcor >= 72) {
if (posBalance < 3) {
return 8;
}
else if (posBalance >= 3) {
if (Cs < 11) {
if (diffRsCsdiff < -4) {
if (Cs < 4) {
return 8;
}
else if (Cs >= 4) {
return 2;
}
}
else if (diffRsCsdiff >= -4) {
if (diffMVdiffVal < -12) {
if (mcTcor < 93) {
return 1;
}
else if (mcTcor >= 93) {
if (posBalance < 36) {
return 1;
}
else if (posBalance >= 36) {
return 8;
}
}
}
else if (diffMVdiffVal >= -12) {
return 1;
}
}
}
else if (Cs >= 11) {
if (AFD < 24) {
if (refDCval < 118) {
if (tcor < 70) {
if (negBalance < 56) {
if (Rs < 85) {
return 1;
}
else if (Rs >= 85) {
if (ssDCval < 88) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (Rs < 182) {
return 1;
}
else if (Rs >= 182) {
return 2;
}
}
}
else if (ssDCval >= 88) {
if (MVDiff < 48) {
return 2;
}
else if (MVDiff >= 48) {
if (Rs < 252) {
return 8;
}
else if (Rs >= 252) {
return 2;
}
}
}
}
}
else if (negBalance >= 56) {
return 2;
}
}
else if (tcor >= 70) {
if (tcor < 96) {
if (CsDiff < 1) {
if (Cs < 144) {
if (ssDCval < 109) {
return 1;
}
else if (ssDCval >= 109) {
return 2;
}
}
else if (Cs >= 144) {
return 2;
}
}
else if (CsDiff >= 1) {
if (mu_mv_mag_sq < 6) {
if (diffMVdiffVal < -4) {
return 2;
}
else if (diffMVdiffVal >= -4) {
return 4;
}
}
else if (mu_mv_mag_sq >= 6) {
if (posBalance < 55) {
if (Cs < 131) {
return 4;
}
else if (Cs >= 131) {
return 2;
}
}
else if (posBalance >= 55) {
if (mu_mv_mag_sq < 95) {
if (diffTSC < -2) {
if (Rs < 187) {
return 8;
}
else if (Rs >= 187) {
return 1;
}
}
else if (diffTSC >= -2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 95) {
return 2;
}
}
}
}
}
else if (tcor >= 96) {
if (Cs < 230) {
return 8;
}
else if (Cs >= 230) {
return 4;
}
}
}
}
else if (refDCval >= 118) {
if (mcTcor < 89) {
if (Cs < 51) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (Rs < 159) {
return 2;
}
else if (Rs >= 159) {
return 1;
}
}
}
else if (Cs >= 51) {
if (MVDiff < 36) {
return 4;
}
else if (MVDiff >= 36) {
return 2;
}
}
}
else if (mcTcor >= 89) {
return 2;
}
}
}
else if (AFD >= 24) {
if (diffRsCsdiff < -27) {
if (Rs < 149) {
return 4;
}
else if (Rs >= 149) {
return 8;
}
}
else if (diffRsCsdiff >= -27) {
if (diffRsCsdiff < -1) {
if (tcor < 28) {
if (SC < 183) {
return 1;
}
else if (SC >= 183) {
return 8;
}
}
else if (tcor >= 28) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (tcor < 63) {
if (TSC < 13) {
if (Rs < 292) {
if (Cs < 53) {
if (Rs < 155) {
return 4;
}
else if (Rs >= 155) {
return 1;
}
}
else if (Cs >= 53) {
return 1;
}
}
else if (Rs >= 292) {
return 4;
}
}
else if (TSC >= 13) {
if (Cs < 234) {
return 1;
}
else if (Cs >= 234) {
return 2;
}
}
}
else if (tcor >= 63) {
return 2;
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (posBalance < 48) {
if (tcor < 101) {
if (mu_mv_mag_sq < 98) {
if (refDCval < 106) {
if (MVDiff < 13) {
if (Cs < 167) {
if (Cs < 62) {
return 8;
}
else if (Cs >= 62) {
if (ssDCval < 94) {
if (AFD < 3) {
if (ssDCval < 75) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
return 2;
}
}
else if (ssDCval >= 75) {
if (Rs < 159) {
if (Cs < 103) {
return 2;
}
else if (Cs >= 103) {
return 1;
}
}
else if (Rs >= 159) {
return 2;
}
}
}
else if (AFD >= 3) {
if (Rs < 230) {
return 1;
}
else if (Rs >= 230) {
return 8;
}
}
}
else if (ssDCval >= 94) {
if (negBalance < 25) {
if (Rs < 238) {
return 2;
}
else if (Rs >= 238) {
return 1;
}
}
else if (negBalance >= 25) {
return 4;
}
}
}
}
else if (Cs >= 167) {
if (tcor < 96) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
if (Rs < 170) {
return 1;
}
else if (Rs >= 170) {
return 2;
}
}
}
else if (tcor >= 96) {
if (refDCval < 77) {
return 1;
}
else if (refDCval >= 77) {
if (Rs < 149) {
return 1;
}
else if (Rs >= 149) {
return 4;
}
}
}
}
}
else if (MVDiff >= 13) {
if (MVDiff < 35) {
if (ssDCval < 72) {
return 2;
}
else if (ssDCval >= 72) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
if (Rs < 266) {
return 1;
}
else if (Rs >= 266) {
return 8;
}
}
}
}
else if (MVDiff >= 35) {
if (Rs < 146) {
if (Cs < 95) {
return 1;
}
else if (Cs >= 95) {
return 2;
}
}
else if (Rs >= 146) {
return 1;
}
}
}
}
else if (refDCval >= 106) {
if (mu_mv_mag_sq < 92) {
if (diffAFD < -1) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
if (Rs < 145) {
return 1;
}
else if (Rs >= 145) {
return 8;
}
}
}
else if (diffAFD >= -1) {
return 2;
}
}
else if (mu_mv_mag_sq >= 92) {
if (Rs < 110) {
return 1;
}
else if (Rs >= 110) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (Cs < 214) {
if (Cs < 125) {
if (SC < 227) {
if (Rs < 90) {
return 4;
}
else if (Rs >= 90) {
return 1;
}
}
else if (SC >= 227) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
if (Rs < 139) {
if (Cs < 111) {
return 8;
}
else if (Cs >= 111) {
return 2;
}
}
else if (Rs >= 139) {
return 8;
}
}
}
}
else if (Cs >= 125) {
if (ssDCval < 73) {
if (MVDiff < 3) {
if (Rs < 196) {
return 2;
}
else if (Rs >= 196) {
return 1;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (ssDCval >= 73) {
if (Rs < 135) {
if (Cs < 151) {
return 2;
}
else if (Cs >= 151) {
return 1;
}
}
else if (Rs >= 135) {
if (SC < 328) {
if (Cs < 156) {
if (negBalance < 23) {
return 2;
}
else if (negBalance >= 23) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
}
else if (Cs >= 156) {
return 8;
}
}
else if (SC >= 328) {
return 2;
}
}
}
}
}
else if (Cs >= 214) {
if (ssDCval < 78) {
return 1;
}
else if (ssDCval >= 78) {
if (Cs < 264) {
return 2;
}
else if (Cs >= 264) {
if (Rs < 191) {
return 1;
}
else if (Rs >= 191) {
return 4;
}
}
}
}
}
}
else if (tcor >= 101) {
if (Cs < 56) {
return 1;
}
else if (Cs >= 56) {
return 8;
}
}
}
else if (posBalance >= 48) {
if (refDCval < 76) {
if (Rs < 167) {
if (negBalance < 30) {
if (mu_mv_mag_sq < 95) {
if (Rs < 126) {
if (Rs < 57) {
if (Rs < 40) {
return 1;
}
else if (Rs >= 40) {
return 8;
}
}
else if (Rs >= 57) {
if (Rs < 102) {
return 2;
}
else if (Rs >= 102) {
if (Cs < 104) {
return 2;
}
else if (Cs >= 104) {
if (posBalance < 51) {
if (Cs < 212) {
return 2;
}
else if (Cs >= 212) {
return 1;
}
}
else if (posBalance >= 51) {
return 1;
}
}
}
}
}
else if (Rs >= 126) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (Rs < 144) {
if (Rs < 124) {
if (Rs < 95) {
return 4;
}
else if (Rs >= 95) {
return 2;
}
}
else if (Rs >= 124) {
return 8;
}
}
else if (Rs >= 144) {
return 2;
}
}
}
else if (negBalance >= 30) {
if (mcTcor < 99) {
if (negBalance < 36) {
return 1;
}
else if (negBalance >= 36) {
if (Cs < 74) {
return 1;
}
else if (Cs >= 74) {
return 4;
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
else if (Rs >= 167) {
if (Cs < 187) {
if (MVDiff < 3) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 1;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (Cs >= 187) {
if (ssDCval < 72) {
if (Rs < 213) {
return 4;
}
else if (Rs >= 213) {
if (Rs < 263) {
return 8;
}
else if (Rs >= 263) {
return 2;
}
}
}
else if (ssDCval >= 72) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
}
}
else if (refDCval >= 76) {
if (Cs < 201) {
if (negBalance < 29) {
if (Rs < 68) {
if (mu_mv_mag_sq < 18) {
if (Rs < 38) {
if (Cs < 5) {
return 8;
}
else if (Cs >= 5) {
return 1;
}
}
else if (Rs >= 38) {
return 4;
}
}
else if (mu_mv_mag_sq >= 18) {
if (MVDiff < 58) {
return 4;
}
else if (MVDiff >= 58) {
return 2;
}
}
}
else if (Rs >= 68) {
if (tcor < 84) {
return 4;
}
else if (tcor >= 84) {
if (gchDC < 8) {
if (ssDCval < 106) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 81) {
return 4;
}
else if (mu_mv_mag_sq >= 81) {
return 1;
}
}
else if (posBalance >= 53) {
return 1;
}
}
else if (ssDCval >= 106) {
if (Cs < 66) {
return 1;
}
else if (Cs >= 66) {
return 2;
}
}
}
else if (gchDC >= 8) {
if (Rs < 173) {
return 4;
}
else if (Rs >= 173) {
if (Rs < 225) {
return 2;
}
else if (Rs >= 225) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 29) {
if (refDCval < 98) {
if (TSC < 3) {
if (MVDiff < 35) {
return 1;
}
else if (MVDiff >= 35) {
return 2;
}
}
else if (TSC >= 3) {
if (refDCval < 93) {
if (ssDCval < 81) {
if (ssDCval < 77) {
if (Rs < 113) {
return 4;
}
else if (Rs >= 113) {
return 1;
}
}
else if (ssDCval >= 77) {
if (Rs < 72) {
return 4;
}
else if (Rs >= 72) {
if (Rs < 265) {
return 2;
}
else if (Rs >= 265) {
return 1;
}
}
}
}
else if (ssDCval >= 81) {
if (SC < 381) {
if (Cs < 11) {
return 2;
}
else if (Cs >= 11) {
return 1;
}
}
else if (SC >= 381) {
return 8;
}
}
}
else if (refDCval >= 93) {
if (Rs < 126) {
if (Rs < 94) {
return 8;
}
else if (Rs >= 94) {
return 1;
}
}
else if (Rs >= 126) {
return 2;
}
}
}
}
else if (refDCval >= 98) {
if (AFD < 12) {
if (posBalance < 51) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 1;
}
}
else if (posBalance >= 51) {
if (Rs < 85) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
else if (Rs >= 85) {
return 2;
}
}
}
else if (AFD >= 12) {
if (ssDCval < 102) {
if (Rs < 230) {
return 4;
}
else if (Rs >= 230) {
return 2;
}
}
else if (ssDCval >= 102) {
if (gchDC < 11) {
return 1;
}
else if (gchDC >= 11) {
return 8;
}
}
}
}
}
}
else if (Cs >= 201) {
if (RsDiff < 1) {
if (Rs < 186) {
if (MVDiff < 14) {
if (mu_mv_mag_sq < 78) {
return 4;
}
else if (mu_mv_mag_sq >= 78) {
if (Rs < 136) {
return 1;
}
else if (Rs >= 136) {
return 4;
}
}
}
else if (MVDiff >= 14) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (mu_mv_mag_sq < 62) {
return 2;
}
else if (mu_mv_mag_sq >= 62) {
return 1;
}
}
}
}
}
else if (Rs >= 186) {
if (MVDiff < 7) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
return 2;
}
}
else if (MVDiff >= 7) {
if (Cs < 371) {
return 4;
}
else if (Cs >= 371) {
return 2;
}
}
}
}
else if (RsDiff >= 1) {
if (RsDiff < 2) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (tcor < 98) {
if (Rs < 78) {
return 1;
}
else if (Rs >= 78) {
if (Rs < 288) {
return 2;
}
else if (Rs >= 288) {
if (Rs < 291) {
return 8;
}
else if (Rs >= 291) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (Rs < 190) {
return 1;
}
else if (Rs >= 190) {
if (Rs < 241) {
return 2;
}
else if (Rs >= 241) {
return 4;
}
}
}
}
}
else if (RsDiff >= 2) {
if (negBalance < 53) {
if (posBalance < 66) {
return 1;
}
else if (posBalance >= 66) {
if (Rs < 197) {
return 1;
}
else if (Rs >= 197) {
return 2;
}
}
}
else if (negBalance >= 53) {
if (Rs < 230) {
return 8;
}
else if (Rs >= 230) {
if (Cs < 410) {
if (Rs < 280) {
return 4;
}
else if (Rs >= 280) {
return 1;
}
}
else if (Cs >= 410) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 304) {
if (RsDiff < 4) {
if (refDCval < 73) {
if (mcTcor < 90) {
if (AFD < 9) {
if (negBalance < 33) {
if (RsDiff < 3) {
if (tcor < 72) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
else if (tcor >= 72) {
if (ssDCval < 57) {
if (Rs < 435) {
if (Cs < 404) {
return 4;
}
else if (Cs >= 404) {
return 1;
}
}
else if (Rs >= 435) {
if (SC < 781) {
return 1;
}
else if (SC >= 781) {
return 8;
}
}
}
else if (ssDCval >= 57) {
if (Cs < 166) {
return 1;
}
else if (Cs >= 166) {
return 8;
}
}
}
}
else if (RsDiff >= 3) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
if (Cs < 306) {
return 1;
}
else if (Cs >= 306) {
return 8;
}
}
}
}
}
else if (negBalance >= 33) {
if (Cs < 156) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (Cs >= 156) {
if (refDCval < 59) {
if (posBalance < 67) {
if (AFD < 7) {
if (Rs < 918) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (Rs >= 918) {
return 1;
}
}
else if (AFD >= 7) {
if (MVDiff < 19) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (MVDiff >= 19) {
if (MVDiff < 21) {
if (Rs < 444) {
return 2;
}
else if (Rs >= 444) {
return 1;
}
}
else if (MVDiff >= 21) {
if (ssDCval < 40) {
return 8;
}
else if (ssDCval >= 40) {
return 2;
}
}
}
}
}
else if (posBalance >= 67) {
return 4;
}
}
else if (refDCval >= 59) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (Rs < 556) {
return 2;
}
else if (Rs >= 556) {
return 4;
}
}
}
}
}
}
else if (AFD >= 9) {
if (negBalance < 48) {
if (diffMVdiffVal < -30) {
if (Cs < 410) {
return 2;
}
else if (Cs >= 410) {
return 8;
}
}
else if (diffMVdiffVal >= -30) {
if (mu_mv_mag_sq < 50) {
if (mcTcor < 87) {
if (MVDiff < 60) {
if (Cs < 138) {
if (Rs < 328) {
return 1;
}
else if (Rs >= 328) {
return 2;
}
}
else if (Cs >= 138) {
if (diffRsCsdiff < -16) {
return 2;
}
else if (diffRsCsdiff >= -16) {
if (diffMVdiffVal < 20) {
return 8;
}
else if (diffMVdiffVal >= 20) {
return 4;
}
}
}
}
else if (MVDiff >= 60) {
return 1;
}
}
else if (mcTcor >= 87) {
if (SC < 994) {
return 1;
}
else if (SC >= 994) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (Rs < 383) {
return 1;
}
else if (Rs >= 383) {
if (Cs < 488) {
return 4;
}
else if (Cs >= 488) {
return 1;
}
}
}
}
}
else if (negBalance >= 48) {
if (refDCval < 45) {
if (tcor < 34) {
return 4;
}
else if (tcor >= 34) {
if (TSC < 6) {
if (diffRsCsdiff < 4) {
return 4;
}
else if (diffRsCsdiff >= 4) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
else if (TSC >= 6) {
if (MVDiff < 46) {
return 8;
}
else if (MVDiff >= 46) {
return 4;
}
}
}
}
else if (refDCval >= 45) {
if (MVDiff < 24) {
return 4;
}
else if (MVDiff >= 24) {
if (Cs < 288) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
if (MVDiff < 89) {
return 8;
}
else if (MVDiff >= 89) {
if (Rs < 342) {
return 2;
}
else if (Rs >= 342) {
return 1;
}
}
}
}
else if (Cs >= 288) {
if (Rs < 360) {
if (MVDiff < 105) {
return 8;
}
else if (MVDiff >= 105) {
return 4;
}
}
else if (Rs >= 360) {
if (MVDiff < 73) {
if (Cs < 334) {
return 8;
}
else if (Cs >= 334) {
return 4;
}
}
else if (MVDiff >= 73) {
if (Rs < 365) {
return 4;
}
else if (Rs >= 365) {
if (Rs < 404) {
return 1;
}
else if (Rs >= 404) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 90) {
if (Rs < 1280) {
if (tcor < 101) {
if (TSC < 1) {
if (posBalance < 30) {
if (Rs < 1063) {
if (ssDCval < 72) {
if (Cs < 131) {
return 4;
}
else if (Cs >= 131) {
if (Cs < 283) {
if (Rs < 766) {
return 8;
}
else if (Rs >= 766) {
return 1;
}
}
else if (Cs >= 283) {
return 8;
}
}
}
else if (ssDCval >= 72) {
return 4;
}
}
else if (Rs >= 1063) {
return 1;
}
}
else if (posBalance >= 30) {
return 4;
}
}
else if (TSC >= 1) {
if (MVDiff < 2) {
if (ssDCval < 69) {
if (mcTcor < 96) {
if (posBalance < 50) {
if (mu_mv_mag_sq < 113) {
return 1;
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
else if (posBalance >= 50) {
if (Rs < 363) {
return 4;
}
else if (Rs >= 363) {
if (Cs < 236) {
return 2;
}
else if (Cs >= 236) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 106) {
if (Rs < 782) {
if (TSC < 2) {
if (Rs < 401) {
return 4;
}
else if (Rs >= 401) {
if (posBalance < 31) {
if (Cs < 355) {
return 8;
}
else if (Cs >= 355) {
return 1;
}
}
else if (posBalance >= 31) {
if (Rs < 482) {
if (Rs < 455) {
return 8;
}
else if (Rs >= 455) {
return 1;
}
}
else if (Rs >= 482) {
return 8;
}
}
}
}
else if (TSC >= 2) {
if (negBalance < 40) {
if (Cs < 407) {
if (SC < 702) {
if (Rs < 408) {
if (Rs < 350) {
return 1;
}
else if (Rs >= 350) {
return 4;
}
}
else if (Rs >= 408) {
return 8;
}
}
else if (SC >= 702) {
if (diffRsCsdiff < 2) {
return 1;
}
else if (diffRsCsdiff >= 2) {
if (ssDCval < 56) {
return 1;
}
else if (ssDCval >= 56) {
return 8;
}
}
}
}
else if (Cs >= 407) {
if (refDCval < 65) {
if (RsCsDiff < 1) {
return 1;
}
else if (RsCsDiff >= 1) {
if (ssDCval < 53) {
return 4;
}
else if (ssDCval >= 53) {
return 8;
}
}
}
else if (refDCval >= 65) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
return 2;
}
}
}
}
else if (negBalance >= 40) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
else if (Rs >= 782) {
if (RsDiff < 2) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (Rs < 888) {
return 2;
}
else if (Rs >= 888) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 106) {
if (refDCval < 59) {
if (Cs < 297) {
if (Cs < 259) {
return 2;
}
else if (Cs >= 259) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
return 1;
}
}
}
else if (Cs >= 297) {
if (ssDCval < 47) {
if (Rs < 560) {
return 8;
}
else if (Rs >= 560) {
return 4;
}
}
else if (ssDCval >= 47) {
if (negBalance < 19) {
if (mu_mv_mag_sq < 119) {
return 1;
}
else if (mu_mv_mag_sq >= 119) {
if (Cs < 452) {
return 8;
}
else if (Cs >= 452) {
return 1;
}
}
}
else if (negBalance >= 19) {
return 1;
}
}
}
}
else if (refDCval >= 59) {
if (mu_mv_mag_sq < 125) {
if (SC < 658) {
return 4;
}
else if (SC >= 658) {
if (SC < 1117) {
if (ssDCval < 66) {
if (SC < 1028) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 4;
}
}
else if (SC >= 1028) {
return 8;
}
}
else if (ssDCval >= 66) {
if (posBalance < 45) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (Cs < 431) {
return 2;
}
else if (Cs >= 431) {
return 1;
}
}
}
else if (posBalance >= 45) {
return 2;
}
}
}
else if (SC >= 1117) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
if (SC < 1251) {
return 2;
}
else if (SC >= 1251) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (MVDiff < 1) {
if (tcor < 99) {
if (Rs < 755) {
return 1;
}
else if (Rs >= 755) {
return 8;
}
}
else if (tcor >= 99) {
return 1;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 601) {
if (SC < 953) {
if (mu_mv_mag_sq < 78) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
else if (mu_mv_mag_sq >= 78) {
if (mu_mv_mag_sq < 125) {
return 2;
}
else if (mu_mv_mag_sq >= 125) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 1;
}
}
}
}
else if (SC >= 953) {
if (Rs < 561) {
return 4;
}
else if (Rs >= 561) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
return 4;
}
}
}
}
else if (Rs >= 601) {
if (Rs < 836) {
if (mu_mv_mag_sq < 123) {
if (CsDiff < 1) {
if (Cs < 411) {
return 2;
}
else if (Cs >= 411) {
return 4;
}
}
else if (CsDiff >= 1) {
if (Rs < 620) {
return 1;
}
else if (Rs >= 620) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 123) {
return 4;
}
}
else if (Rs >= 836) {
if (Cs < 273) {
return 1;
}
else if (Cs >= 273) {
return 8;
}
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 139) {
if (AFD < 5) {
if (TSC < 3) {
if (mcTcor < 97) {
if (gchDC < 1) {
if (Rs < 691) {
return 1;
}
else if (Rs >= 691) {
return 4;
}
}
else if (gchDC >= 1) {
return 2;
}
}
else if (mcTcor >= 97) {
if (ssDCval < 63) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
return 1;
}
}
}
else if (ssDCval >= 63) {
return 4;
}
}
}
else if (TSC >= 3) {
if (Cs < 83) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (Cs >= 83) {
return 1;
}
}
}
else if (AFD >= 5) {
if (tcor < 81) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Cs < 121) {
if (MVDiff < 43) {
return 1;
}
else if (MVDiff >= 43) {
return 8;
}
}
else if (Cs >= 121) {
return 8;
}
}
}
else if (tcor >= 81) {
if (ssDCval < 69) {
if (Cs < 114) {
return 1;
}
else if (Cs >= 114) {
if (mu_mv_mag_sq < 46) {
if (diffMVdiffVal < 5) {
return 1;
}
else if (diffMVdiffVal >= 5) {
if (Rs < 644) {
return 8;
}
else if (Rs >= 644) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
if (Rs < 602) {
return 4;
}
else if (Rs >= 602) {
return 2;
}
}
}
}
}
else if (ssDCval >= 69) {
if (Rs < 588) {
return 4;
}
else if (Rs >= 588) {
return 1;
}
}
}
}
}
else if (Cs >= 139) {
if (mu_mv_mag_sq < 84) {
if (Rs < 689) {
if (ssDCval < 57) {
if (Cs < 399) {
if (MVDiff < 37) {
if (posBalance < 31) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
else if (posBalance >= 31) {
if (mcTcor < 92) {
if (posBalance < 49) {
if (Rs < 424) {
return 4;
}
else if (Rs >= 424) {
return 1;
}
}
else if (posBalance >= 49) {
if (diffAFD < 1) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (mcTcor >= 92) {
if (ssDCval < 43) {
if (AFD < 6) {
if (negBalance < 37) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (mu_mv_mag_sq < 74) {
if (Rs < 335) {
return 1;
}
else if (Rs >= 335) {
return 4;
}
}
else if (mu_mv_mag_sq >= 74) {
return 1;
}
}
}
else if (negBalance >= 37) {
if (Rs < 412) {
return 1;
}
else if (Rs >= 412) {
return 8;
}
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (ssDCval >= 43) {
if (Rs < 470) {
if (MVDiff < 8) {
if (Rs < 367) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
if (Rs < 352) {
return 1;
}
else if (Rs >= 352) {
return 4;
}
}
}
else if (Rs >= 367) {
if (ssDCval < 55) {
return 1;
}
else if (ssDCval >= 55) {
return 4;
}
}
}
else if (MVDiff >= 8) {
if (negBalance < 50) {
if (tcor < 98) {
if (Cs < 387) {
return 4;
}
else if (Cs >= 387) {
return 1;
}
}
else if (tcor >= 98) {
if (MVDiff < 32) {
return 1;
}
else if (MVDiff >= 32) {
return 4;
}
}
}
else if (negBalance >= 50) {
if (Rs < 340) {
return 2;
}
else if (Rs >= 340) {
return 1;
}
}
}
}
else if (Rs >= 470) {
if (negBalance < 34) {
if (Rs < 657) {
if (Cs < 353) {
if (AFD < 3) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
return 1;
}
}
else if (AFD >= 3) {
if (diffMVdiffVal < -7) {
if (Rs < 513) {
return 4;
}
else if (Rs >= 513) {
return 1;
}
}
else if (diffMVdiffVal >= -7) {
return 1;
}
}
}
else if (Cs >= 353) {
if (Rs < 560) {
return 4;
}
else if (Rs >= 560) {
return 1;
}
}
}
else if (Rs >= 657) {
return 2;
}
}
else if (negBalance >= 34) {
if (mcTcor < 94) {
if (Rs < 536) {
return 8;
}
else if (Rs >= 536) {
return 1;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 37) {
if (diffMVdiffVal < 6) {
if (diffRsCsdiff < 1) {
return 2;
}
else if (diffRsCsdiff >= 1) {
if (Rs < 420) {
return 4;
}
else if (Rs >= 420) {
return 2;
}
}
}
else if (diffMVdiffVal >= 6) {
if (Cs < 183) {
return 4;
}
else if (Cs >= 183) {
return 1;
}
}
}
}
else if (Cs >= 399) {
if (MVDiff < 8) {
if (SC < 960) {
return 8;
}
else if (SC >= 960) {
if (Rs < 614) {
return 4;
}
else if (Rs >= 614) {
if (Rs < 653) {
return 8;
}
else if (Rs >= 653) {
return 1;
}
}
}
}
else if (MVDiff >= 8) {
if (negBalance < 43) {
if (SC < 878) {
if (diffRsCsdiff < -2) {
return 1;
}
else if (diffRsCsdiff >= -2) {
if (RsCsDiff < 4) {
return 1;
}
else if (RsCsDiff >= 4) {
return 4;
}
}
}
else if (SC >= 878) {
if (Rs < 548) {
return 1;
}
else if (Rs >= 548) {
if (Rs < 625) {
if (SC < 1040) {
if (Rs < 608) {
return 4;
}
else if (Rs >= 608) {
return 1;
}
}
else if (SC >= 1040) {
return 1;
}
}
else if (Rs >= 625) {
return 4;
}
}
}
}
else if (negBalance >= 43) {
if (AFD < 12) {
if (Rs < 401) {
return 4;
}
else if (Rs >= 401) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 1;
}
}
}
else if (AFD >= 12) {
return 1;
}
}
}
}
}
else if (ssDCval >= 57) {
if (refDCval < 65) {
if (negBalance < 24) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 4;
}
}
else if (negBalance >= 24) {
if (Cs < 353) {
if (SC < 563) {
if (Cs < 224) {
return 4;
}
else if (Cs >= 224) {
return 1;
}
}
else if (SC >= 563) {
if (tcor < 94) {
if (mu_mv_mag_sq < 67) {
return 2;
}
else if (mu_mv_mag_sq >= 67) {
if (Cs < 231) {
return 2;
}
else if (Cs >= 231) {
return 4;
}
}
}
else if (tcor >= 94) {
if (Rs < 385) {
if (Rs < 311) {
return 2;
}
else if (Rs >= 311) {
return 1;
}
}
else if (Rs >= 385) {
return 4;
}
}
}
}
else if (Cs >= 353) {
if (TSC < 4) {
if (negBalance < 38) {
if (SC < 881) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (SC < 873) {
if (Cs < 430) {
return 1;
}
else if (Cs >= 430) {
if (Rs < 327) {
return 1;
}
else if (Rs >= 327) {
return 2;
}
}
}
else if (SC >= 873) {
return 2;
}
}
}
else if (SC >= 881) {
if (mu_mv_mag_sq < 53) {
return 4;
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 540) {
return 2;
}
else if (Rs >= 540) {
return 1;
}
}
}
}
else if (negBalance >= 38) {
if (tcor < 93) {
return 2;
}
else if (tcor >= 93) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
}
else if (TSC >= 4) {
if (tcor < 85) {
if (negBalance < 48) {
if (negBalance < 43) {
return 1;
}
else if (negBalance >= 43) {
if (Rs < 382) {
return 1;
}
else if (Rs >= 382) {
return 4;
}
}
}
else if (negBalance >= 48) {
return 2;
}
}
else if (tcor >= 85) {
if (mu_mv_mag_sq < 65) {
return 1;
}
else if (mu_mv_mag_sq >= 65) {
if (Rs < 522) {
return 2;
}
else if (Rs >= 522) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 65) {
if (MVDiff < 6) {
if (mu_mv_mag_sq < 27) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 1;
}
}
else if (mu_mv_mag_sq >= 27) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (Rs < 494) {
return 4;
}
else if (Rs >= 494) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (Rs < 477) {
if (MVDiff < 19) {
if (MVDiff < 7) {
if (Rs < 387) {
return 4;
}
else if (Rs >= 387) {
return 2;
}
}
else if (MVDiff >= 7) {
if (SC < 881) {
if (negBalance < 51) {
return 1;
}
else if (negBalance >= 51) {
return 2;
}
}
else if (SC >= 881) {
if (Rs < 446) {
return 8;
}
else if (Rs >= 446) {
return 2;
}
}
}
}
else if (MVDiff >= 19) {
if (ssDCval < 68) {
if (Cs < 402) {
return 4;
}
else if (Cs >= 402) {
if (Rs < 418) {
return 1;
}
else if (Rs >= 418) {
return 2;
}
}
}
else if (ssDCval >= 68) {
if (Cs < 445) {
return 8;
}
else if (Cs >= 445) {
return 1;
}
}
}
}
else if (Rs >= 477) {
if (AFD < 8) {
if (Rs < 515) {
return 4;
}
else if (Rs >= 515) {
return 1;
}
}
else if (AFD >= 8) {
if (Cs < 460) {
return 4;
}
else if (Cs >= 460) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
}
}
}
}
else if (Rs >= 689) {
if (mcTcor < 96) {
if (ssDCval < 56) {
if (negBalance < 43) {
if (posBalance < 37) {
if (Rs < 1004) {
return 2;
}
else if (Rs >= 1004) {
return 1;
}
}
else if (posBalance >= 37) {
if (Cs < 469) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (Cs >= 469) {
return 2;
}
}
}
else if (negBalance >= 43) {
if (ssDCval < 39) {
return 1;
}
else if (ssDCval >= 39) {
return 8;
}
}
}
else if (ssDCval >= 56) {
if (SC < 1071) {
if (SC < 971) {
return 1;
}
else if (SC >= 971) {
return 2;
}
}
else if (SC >= 1071) {
if (diffRsCsdiff < -6) {
if (Rs < 841) {
return 4;
}
else if (Rs >= 841) {
if (negBalance < 42) {
return 1;
}
else if (negBalance >= 42) {
return 8;
}
}
}
else if (diffRsCsdiff >= -6) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
if (mcTcor < 99) {
if (ssDCval < 44) {
if (TSC < 2) {
if (Rs < 813) {
return 8;
}
else if (Rs >= 813) {
return 4;
}
}
else if (TSC >= 2) {
if (Rs < 818) {
return 2;
}
else if (Rs >= 818) {
return 1;
}
}
}
else if (ssDCval >= 44) {
if (SC < 1281) {
return 1;
}
else if (SC >= 1281) {
if (negBalance < 29) {
if (Rs < 939) {
if (Rs < 856) {
return 1;
}
else if (Rs >= 856) {
return 2;
}
}
else if (Rs >= 939) {
return 4;
}
}
else if (negBalance >= 29) {
if (diffRsCsdiff < 4) {
return 1;
}
else if (diffRsCsdiff >= 4) {
if (Rs < 1093) {
return 4;
}
else if (Rs >= 1093) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 56) {
return 2;
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1110) {
return 1;
}
else if (Rs >= 1110) {
if (Cs < 334) {
return 1;
}
else if (Cs >= 334) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (ssDCval < 58) {
if (negBalance < 38) {
if (diffRsCsdiff < -5) {
if (posBalance < 51) {
if (Rs < 705) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (Rs >= 705) {
return 4;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
if (TSCindex < 2) {
if (Rs < 458) {
return 8;
}
else if (Rs >= 458) {
return 4;
}
}
else if (TSCindex >= 2) {
if (negBalance < 28) {
if (refDCval < 44) {
return 1;
}
else if (refDCval >= 44) {
if (SC < 861) {
if (ssDCval < 49) {
if (mu_mv_mag_sq < 88) {
return 4;
}
else if (mu_mv_mag_sq >= 88) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
else if (ssDCval >= 49) {
if (Cs < 378) {
return 4;
}
else if (Cs >= 378) {
return 8;
}
}
}
else if (SC >= 861) {
if (ssDCval < 46) {
return 4;
}
else if (ssDCval >= 46) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 4;
}
}
}
}
}
else if (negBalance >= 28) {
if (mcTcor < 97) {
if (tcor < 88) {
return 1;
}
else if (tcor >= 88) {
if (posBalance < 42) {
return 1;
}
else if (posBalance >= 42) {
return 8;
}
}
}
else if (mcTcor >= 97) {
return 1;
}
}
}
}
}
else if (negBalance >= 38) {
return 8;
}
}
else if (ssDCval >= 58) {
if (SC < 706) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
if (AFD < 3) {
if (Rs < 423) {
return 4;
}
else if (Rs >= 423) {
return 8;
}
}
else if (AFD >= 3) {
if (Rs < 394) {
return 1;
}
else if (Rs >= 394) {
return 2;
}
}
}
}
else if (SC >= 706) {
if (refDCval < 71) {
if (negBalance < 31) {
if (mu_mv_mag_sq < 118) {
if (TSCindex < 3) {
if (ssDCval < 64) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (SC < 982) {
if (RsCsDiff < 1) {
return 1;
}
else if (RsCsDiff >= 1) {
return 8;
}
}
else if (SC >= 982) {
return 1;
}
}
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (TSCindex >= 3) {
if (Rs < 894) {
if (Rs < 528) {
if (tcor < 85) {
return 2;
}
else if (tcor >= 85) {
if (RsCsDiff < 2) {
return 8;
}
else if (RsCsDiff >= 2) {
if (Rs < 501) {
return 1;
}
else if (Rs >= 501) {
return 8;
}
}
}
}
else if (Rs >= 528) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (Rs < 580) {
return 1;
}
else if (Rs >= 580) {
return 2;
}
}
}
}
else if (Rs >= 894) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 556) {
return 1;
}
else if (Rs >= 556) {
return 2;
}
}
}
else if (negBalance >= 31) {
if (Rs < 656) {
if (mu_mv_mag_sq < 98) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (Cs < 376) {
if (Rs < 507) {
return 2;
}
else if (Rs >= 507) {
return 1;
}
}
else if (Cs >= 376) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 98) {
return 1;
}
}
else if (Rs >= 656) {
if (Rs < 798) {
return 4;
}
else if (Rs >= 798) {
if (Cs < 470) {
return 1;
}
else if (Cs >= 470) {
return 8;
}
}
}
}
}
else if (refDCval >= 71) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
return 8;
}
}
}
}
}
}
}
}
}
else if (tcor >= 101) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
if (TSC < 3) {
if (Rs < 407) {
if (Cs < 60) {
return 8;
}
else if (Cs >= 60) {
return 1;
}
}
else if (Rs >= 407) {
return 8;
}
}
else if (TSC >= 3) {
if (tcor < 105) {
if (SC < 778) {
if (Cs < 253) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
else if (Cs >= 253) {
return 4;
}
}
else if (SC >= 778) {
if (CsDiff < 2) {
return 4;
}
else if (CsDiff >= 2) {
if (Rs < 511) {
return 2;
}
else if (Rs >= 511) {
return 1;
}
}
}
}
else if (tcor >= 105) {
if (Rs < 391) {
if (Rs < 359) {
return 4;
}
else if (Rs >= 359) {
return 1;
}
}
else if (Rs >= 391) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1280) {
if (refDCval < 64) {
if (RsCsDiff < 2) {
if (SC < 1654) {
return 8;
}
else if (SC >= 1654) {
if (ssDCval < 62) {
if (ssDCval < 56) {
if (Rs < 1495) {
return 1;
}
else if (Rs >= 1495) {
return 2;
}
}
else if (ssDCval >= 56) {
if (Cs < 404) {
if (Rs < 1552) {
if (Rs < 1440) {
return 8;
}
else if (Rs >= 1440) {
return 1;
}
}
else if (Rs >= 1552) {
return 2;
}
}
else if (Cs >= 404) {
if (mu_mv_mag_sq < 120) {
if (Rs < 1492) {
return 1;
}
else if (Rs >= 1492) {
return 8;
}
}
else if (mu_mv_mag_sq >= 120) {
return 8;
}
}
}
}
else if (ssDCval >= 62) {
if (SC < 1879) {
if (Cs < 489) {
if (Rs < 1462) {
return 2;
}
else if (Rs >= 1462) {
return 1;
}
}
else if (Cs >= 489) {
return 8;
}
}
else if (SC >= 1879) {
if (diffMVdiffVal < -1) {
if (Rs < 1731) {
return 1;
}
else if (Rs >= 1731) {
return 2;
}
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 2) {
if (gchDC < 2) {
if (negBalance < 28) {
if (SC < 1910) {
return 1;
}
else if (SC >= 1910) {
if (Cs < 367) {
return 2;
}
else if (Cs >= 367) {
return 4;
}
}
}
else if (negBalance >= 28) {
return 1;
}
}
else if (gchDC >= 2) {
if (ssDCval < 63) {
return 2;
}
else if (ssDCval >= 63) {
return 1;
}
}
}
}
else if (refDCval >= 64) {
if (SC < 2072) {
if (mu_mv_mag_sq < 56) {
if (negBalance < 25) {
if (MVDiff < 9) {
if (Cs < 304) {
return 1;
}
else if (Cs >= 304) {
return 2;
}
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (negBalance >= 25) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (ssDCval < 70) {
if (ssDCval < 64) {
if (Rs < 1475) {
return 2;
}
else if (Rs >= 1475) {
return 1;
}
}
else if (ssDCval >= 64) {
if (mu_mv_mag_sq < 128) {
if (Rs < 1318) {
if (Rs < 1295) {
return 1;
}
else if (Rs >= 1295) {
return 8;
}
}
else if (Rs >= 1318) {
return 1;
}
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
}
else if (ssDCval >= 70) {
if (AFD < 4) {
if (Cs < 453) {
return 8;
}
else if (Cs >= 453) {
return 1;
}
}
else if (AFD >= 4) {
return 1;
}
}
}
}
else if (SC >= 2072) {
if (RsDiff < 3) {
return 2;
}
else if (RsDiff >= 3) {
if (Cs < 276) {
return 1;
}
else if (Cs >= 276) {
if (Rs < 1822) {
return 4;
}
else if (Rs >= 1822) {
return 2;
}
}
}
}
}
}
}
}
else if (refDCval >= 73) {
if (Rs < 1118) {
if (refDCval < 80) {
if (AFD < 5) {
if (MVDiff < 1) {
if (ssDCval < 74) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (ssDCval >= 74) {
if (ssDCval < 78) {
if (SC < 1272) {
return 8;
}
else if (SC >= 1272) {
return 1;
}
}
else if (ssDCval >= 78) {
if (negBalance < 33) {
return 1;
}
else if (negBalance >= 33) {
return 8;
}
}
}
}
else if (MVDiff >= 1) {
if (Rs < 843) {
if (posBalance < 49) {
if (CsDiff < 1) {
if (Cs < 317) {
return 8;
}
else if (Cs >= 317) {
if (Cs < 325) {
return 1;
}
else if (Cs >= 325) {
return 8;
}
}
}
else if (CsDiff >= 1) {
if (TSC < 3) {
if (Rs < 681) {
return 1;
}
else if (Rs >= 681) {
return 8;
}
}
else if (TSC >= 3) {
if (Rs < 533) {
return 8;
}
else if (Rs >= 533) {
if (Rs < 722) {
return 4;
}
else if (Rs >= 722) {
return 8;
}
}
}
}
}
else if (posBalance >= 49) {
return 8;
}
}
else if (Rs >= 843) {
return 1;
}
}
}
else if (AFD >= 5) {
if (tcor < 77) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (SC < 730) {
if (SC < 673) {
return 8;
}
else if (SC >= 673) {
return 4;
}
}
else if (SC >= 730) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
}
}
else if (tcor >= 77) {
if (mu_mv_mag_sq < 60) {
if (MVDiff < 9) {
if (Cs < 243) {
if (Rs < 393) {
return 1;
}
else if (Rs >= 393) {
if (Rs < 722) {
return 4;
}
else if (Rs >= 722) {
return 8;
}
}
}
else if (Cs >= 243) {
return 1;
}
}
else if (MVDiff >= 9) {
if (Rs < 489) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 1;
}
}
else if (Rs >= 489) {
if (Cs < 221) {
return 1;
}
else if (Cs >= 221) {
if (MVDiff < 24) {
return 1;
}
else if (MVDiff >= 24) {
if (Rs < 623) {
return 1;
}
else if (Rs >= 623) {
if (refDCval < 78) {
return 4;
}
else if (refDCval >= 78) {
return 8;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 60) {
return 2;
}
}
}
}
else if (refDCval >= 80) {
if (Cs < 168) {
if (Rs < 687) {
if (tcor < 58) {
if (mcTcor < 86) {
return 2;
}
else if (mcTcor >= 86) {
if (diffMVdiffVal < 8) {
if (diffMVdiffVal < -2) {
if (Rs < 369) {
return 8;
}
else if (Rs >= 369) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
return 2;
}
}
else if (diffMVdiffVal >= 8) {
return 8;
}
}
}
else if (tcor >= 58) {
if (ssDCval < 84) {
if (Cs < 78) {
if (MVDiff < 26) {
if (Rs < 497) {
return 2;
}
else if (Rs >= 497) {
return 8;
}
}
else if (MVDiff >= 26) {
return 1;
}
}
else if (Cs >= 78) {
return 8;
}
}
else if (ssDCval >= 84) {
if (Rs < 339) {
if (posBalance < 58) {
if (Cs < 103) {
if (Cs < 92) {
return 1;
}
else if (Cs >= 92) {
return 8;
}
}
else if (Cs >= 103) {
return 1;
}
}
else if (posBalance >= 58) {
if (Cs < 65) {
return 1;
}
else if (Cs >= 65) {
return 8;
}
}
}
else if (Rs >= 339) {
if (AFD < 17) {
if (CsDiff < 2) {
if (posBalance < 80) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
if (diffMVdiffVal < 4) {
if (posBalance < 57) {
return 1;
}
else if (posBalance >= 57) {
if (Rs < 410) {
return 8;
}
else if (Rs >= 410) {
return 1;
}
}
}
else if (diffMVdiffVal >= 4) {
return 1;
}
}
}
else if (posBalance >= 80) {
return 8;
}
}
else if (CsDiff >= 2) {
return 8;
}
}
else if (AFD >= 17) {
return 8;
}
}
}
}
}
else if (Rs >= 687) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
if (Rs < 739) {
return 4;
}
else if (Rs >= 739) {
if (Rs < 1101) {
return 1;
}
else if (Rs >= 1101) {
return 8;
}
}
}
}
}
else if (Cs >= 168) {
if (mcTcor < 97) {
if (MVDiff < 22) {
if (mu_mv_mag_sq < 2) {
if (TSC < 6) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
if (Cs < 484) {
if (mcTcor < 94) {
if (Rs < 499) {
return 2;
}
else if (Rs >= 499) {
return 1;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (Cs >= 484) {
if (Cs < 494) {
return 2;
}
else if (Cs >= 494) {
return 1;
}
}
}
}
else if (TSC >= 6) {
if (AFD < 13) {
if (Rs < 642) {
return 8;
}
else if (Rs >= 642) {
return 1;
}
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (ssDCval < 83) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (ssDCval >= 83) {
if (AFD < 11) {
if (diffRsCsdiff < -18) {
return 4;
}
else if (diffRsCsdiff >= -18) {
if (refDCval < 88) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 515) {
return 8;
}
else if (Rs >= 515) {
return 1;
}
}
}
else if (refDCval >= 88) {
if (RsCsDiff < 1) {
return 1;
}
else if (RsCsDiff >= 1) {
if (Rs < 306) {
if (Rs < 305) {
return 8;
}
else if (Rs >= 305) {
return 2;
}
}
else if (Rs >= 306) {
if (Cs < 184) {
return 1;
}
else if (Cs >= 184) {
return 8;
}
}
}
}
}
}
else if (AFD >= 11) {
if (TSC < 6) {
if (Cs < 347) {
if (Rs < 378) {
return 8;
}
else if (Rs >= 378) {
return 2;
}
}
else if (Cs >= 347) {
return 1;
}
}
else if (TSC >= 6) {
if (ssDCval < 85) {
return 2;
}
else if (ssDCval >= 85) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 22) {
if (Rs < 559) {
if (diffMVdiffVal < 48) {
if (TSC < 9) {
if (tcor < 89) {
if (RsCsDiff < 3) {
return 8;
}
else if (RsCsDiff >= 3) {
if (Rs < 418) {
if (SC < 700) {
if (Cs < 315) {
return 1;
}
else if (Cs >= 315) {
return 8;
}
}
else if (SC >= 700) {
if (diffMVdiffVal < 36) {
return 2;
}
else if (diffMVdiffVal >= 36) {
return 1;
}
}
}
else if (Rs >= 418) {
if (Rs < 518) {
return 8;
}
else if (Rs >= 518) {
return 1;
}
}
}
}
else if (tcor >= 89) {
return 2;
}
}
else if (TSC >= 9) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
return 1;
}
}
}
else if (diffMVdiffVal >= 48) {
if (Rs < 411) {
return 4;
}
else if (Rs >= 411) {
if (Rs < 459) {
return 2;
}
else if (Rs >= 459) {
if (Rs < 506) {
return 1;
}
else if (Rs >= 506) {
return 8;
}
}
}
}
}
else if (Rs >= 559) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
if (Rs < 590) {
return 1;
}
else if (Rs >= 590) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 232) {
if (ssDCval < 93) {
if (mu_mv_mag_sq < 76) {
if (negBalance < 35) {
return 4;
}
else if (negBalance >= 35) {
if (Rs < 422) {
return 1;
}
else if (Rs >= 422) {
if (Rs < 622) {
return 8;
}
else if (Rs >= 622) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 76) {
if (Rs < 313) {
if (Cs < 200) {
return 1;
}
else if (Cs >= 200) {
return 2;
}
}
else if (Rs >= 313) {
return 2;
}
}
}
else if (ssDCval >= 93) {
if (refDCval < 106) {
return 1;
}
else if (refDCval >= 106) {
if (Cs < 205) {
return 4;
}
else if (Cs >= 205) {
if (Rs < 383) {
return 2;
}
else if (Rs >= 383) {
return 8;
}
}
}
}
}
else if (Cs >= 232) {
if (mu_mv_mag_sq < 68) {
if (MVDiff < 6) {
if (Cs < 309) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
if (Rs < 466) {
return 1;
}
else if (Rs >= 466) {
return 8;
}
}
}
else if (Cs >= 309) {
if (AFD < 13) {
if (Rs < 1033) {
if (Cs < 332) {
if (Cs < 324) {
return 1;
}
else if (Cs >= 324) {
return 4;
}
}
else if (Cs >= 332) {
return 1;
}
}
else if (Rs >= 1033) {
return 4;
}
}
else if (AFD >= 13) {
return 8;
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 116) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 53) {
if (Cs < 240) {
return 4;
}
else if (Cs >= 240) {
return 1;
}
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 588) {
return 4;
}
else if (Rs >= 588) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (Rs < 555) {
if (mu_mv_mag_sq < 8) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 4;
}
}
else if (mu_mv_mag_sq >= 8) {
if (AFD < 15) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 8;
}
}
else if (AFD >= 15) {
return 2;
}
}
}
else if (Rs >= 555) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
}
else if (ssDCval >= 116) {
if (mcTcor < 99) {
if (Rs < 484) {
if (tcor < 87) {
if (Rs < 330) {
return 8;
}
else if (Rs >= 330) {
if (Rs < 348) {
return 1;
}
else if (Rs >= 348) {
return 4;
}
}
}
else if (tcor >= 87) {
if (refDCval < 129) {
return 2;
}
else if (refDCval >= 129) {
if (mu_mv_mag_sq < 3) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 377) {
return 4;
}
else if (Rs >= 377) {
return 8;
}
}
}
}
}
else if (Rs >= 484) {
if (Cs < 421) {
if (Rs < 1031) {
return 4;
}
else if (Rs >= 1031) {
return 1;
}
}
else if (Cs >= 421) {
if (MVDiff < 41) {
if (Cs < 482) {
return 1;
}
else if (Cs >= 482) {
return 2;
}
}
else if (MVDiff >= 41) {
return 2;
}
}
}
}
else if (mcTcor >= 99) {
if (MVDiff < 44) {
if (ssDCval < 155) {
if (refDCval < 123) {
if (Rs < 416) {
return 2;
}
else if (Rs >= 416) {
return 1;
}
}
else if (refDCval >= 123) {
return 1;
}
}
else if (ssDCval >= 155) {
return 4;
}
}
else if (MVDiff >= 44) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 68) {
if (ssDCval < 96) {
if (Cs < 278) {
if (Rs < 988) {
return 1;
}
else if (Rs >= 988) {
if (refDCval < 81) {
return 1;
}
else if (refDCval >= 81) {
return 2;
}
}
}
else if (Cs >= 278) {
if (posBalance < 43) {
if (Rs < 1040) {
return 8;
}
else if (Rs >= 1040) {
return 2;
}
}
else if (posBalance >= 43) {
if (MVDiff < 8) {
if (negBalance < 37) {
if (Cs < 493) {
return 1;
}
else if (Cs >= 493) {
return 8;
}
}
else if (negBalance >= 37) {
if (Rs < 539) {
return 4;
}
else if (Rs >= 539) {
return 8;
}
}
}
else if (MVDiff >= 8) {
return 8;
}
}
}
}
else if (ssDCval >= 96) {
if (ssDCval < 98) {
if (Cs < 468) {
if (Rs < 522) {
return 1;
}
else if (Rs >= 522) {
return 4;
}
}
else if (Cs >= 468) {
return 8;
}
}
else if (ssDCval >= 98) {
if (mu_mv_mag_sq < 103) {
if (negBalance < 31) {
if (ssDCval < 159) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 2;
}
}
else if (ssDCval >= 159) {
return 4;
}
}
else if (negBalance >= 31) {
if (posBalance < 54) {
if (Rs < 407) {
return 4;
}
else if (Rs >= 407) {
return 2;
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (diffAFD < -9) {
return 1;
}
else if (diffAFD >= -9) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 1118) {
if (ssDCval < 101) {
if (CsDiff < 1) {
if (TSCindex < 2) {
if (Cs < 149) {
if (Rs < 1252) {
return 8;
}
else if (Rs >= 1252) {
return 1;
}
}
else if (Cs >= 149) {
if (Rs < 1441) {
return 2;
}
else if (Rs >= 1441) {
return 1;
}
}
}
else if (TSCindex >= 2) {
if (negBalance < 81) {
if (Rs < 2118) {
return 1;
}
else if (Rs >= 2118) {
return 2;
}
}
else if (negBalance >= 81) {
if (Rs < 1380) {
return 8;
}
else if (Rs >= 1380) {
return 4;
}
}
}
}
else if (CsDiff >= 1) {
if (tcor < 98) {
if (SC < 2615) {
if (SC < 2442) {
if (ssDCval < 98) {
if (MVDiff < 2) {
if (ssDCval < 81) {
if (Rs < 1663) {
return 8;
}
else if (Rs >= 1663) {
return 1;
}
}
else if (ssDCval >= 81) {
return 1;
}
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 98) {
return 8;
}
}
else if (SC >= 2442) {
if (Rs < 2107) {
return 8;
}
else if (Rs >= 2107) {
return 1;
}
}
}
else if (SC >= 2615) {
return 2;
}
}
else if (tcor >= 98) {
if (mu_mv_mag_sq < 67) {
if (Cs < 229) {
return 1;
}
else if (Cs >= 229) {
return 8;
}
}
else if (mu_mv_mag_sq >= 67) {
if (SC < 2466) {
if (mu_mv_mag_sq < 74) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
return 1;
}
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 483) {
return 1;
}
else if (Cs >= 483) {
return 4;
}
}
}
else if (SC >= 2466) {
if (negBalance < 31) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
return 8;
}
}
else if (negBalance >= 31) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 101) {
if (MVDiff < 17) {
if (ssDCval < 134) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (ssDCval >= 134) {
return 4;
}
}
else if (MVDiff >= 17) {
if (RsCsDiff < 10) {
if (mu_mv_mag_sq < 14) {
return 8;
}
else if (mu_mv_mag_sq >= 14) {
return 1;
}
}
else if (RsCsDiff >= 10) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 74) {
if (RsDiff < 8) {
if (negBalance < 46) {
if (AFD < 8) {
if (Rs < 559) {
return 8;
}
else if (Rs >= 559) {
return 2;
}
}
else if (AFD >= 8) {
if (posBalance < 75) {
if (RsDiff < 5) {
if (AFD < 11) {
if (gchDC < 4) {
return 8;
}
else if (gchDC >= 4) {
if (Rs < 700) {
return 2;
}
else if (Rs >= 700) {
return 1;
}
}
}
else if (AFD >= 11) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
if (Cs < 229) {
return 1;
}
else if (Cs >= 229) {
return 8;
}
}
}
}
else if (RsDiff >= 5) {
if (negBalance < 45) {
if (tcor < 73) {
return 1;
}
else if (tcor >= 73) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (negBalance >= 45) {
if (Rs < 1011) {
return 2;
}
else if (Rs >= 1011) {
return 1;
}
}
}
}
else if (posBalance >= 75) {
if (AFD < 26) {
if (posBalance < 77) {
if (Rs < 376) {
return 1;
}
else if (Rs >= 376) {
return 8;
}
}
else if (posBalance >= 77) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (AFD >= 26) {
if (Rs < 378) {
return 8;
}
else if (Rs >= 378) {
if (Rs < 633) {
return 2;
}
else if (Rs >= 633) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 46) {
if (AFD < 22) {
if (Cs < 387) {
if (tcor < 39) {
if (TSC < 8) {
if (tcor < 23) {
if (Rs < 562) {
return 4;
}
else if (Rs >= 562) {
return 1;
}
}
else if (tcor >= 23) {
if (Rs < 395) {
return 1;
}
else if (Rs >= 395) {
return 4;
}
}
}
else if (TSC >= 8) {
return 8;
}
}
else if (tcor >= 39) {
if (AFD < 14) {
if (ssDCval < 48) {
if (Rs < 602) {
return 8;
}
else if (Rs >= 602) {
return 1;
}
}
else if (ssDCval >= 48) {
if (ssDCval < 153) {
if (MVDiff < 22) {
if (Rs < 731) {
return 2;
}
else if (Rs >= 731) {
return 8;
}
}
else if (MVDiff >= 22) {
if (diffMVdiffVal < 5) {
if (mu_mv_mag_sq < 31) {
return 2;
}
else if (mu_mv_mag_sq >= 31) {
if (Rs < 930) {
return 8;
}
else if (Rs >= 930) {
return 2;
}
}
}
else if (diffMVdiffVal >= 5) {
if (MVDiff < 36) {
return 8;
}
else if (MVDiff >= 36) {
return 2;
}
}
}
}
else if (ssDCval >= 153) {
return 1;
}
}
}
else if (AFD >= 14) {
if (mu_mv_mag_sq < 5) {
if (diffMVdiffVal < 5) {
if (AFD < 21) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 1;
}
}
else if (AFD >= 21) {
return 8;
}
}
else if (diffMVdiffVal >= 5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 5) {
if (SC < 778) {
if (SC < 705) {
return 1;
}
else if (SC >= 705) {
return 8;
}
}
else if (SC >= 778) {
if (posBalance < 67) {
if (SC < 928) {
if (diffRsCsdiff < 7) {
return 4;
}
else if (diffRsCsdiff >= 7) {
if (SC < 831) {
return 1;
}
else if (SC >= 831) {
return 2;
}
}
}
else if (SC >= 928) {
if (RsCsDiff < 19) {
return 2;
}
else if (RsCsDiff >= 19) {
return 1;
}
}
}
else if (posBalance >= 67) {
return 4;
}
}
}
}
}
}
else if (Cs >= 387) {
if (tcor < 60) {
if (AFD < 18) {
if (negBalance < 54) {
if (Rs < 677) {
return 8;
}
else if (Rs >= 677) {
if (Rs < 892) {
return 2;
}
else if (Rs >= 892) {
return 1;
}
}
}
else if (negBalance >= 54) {
if (ssDCval < 116) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (Rs < 944) {
return 2;
}
else if (Rs >= 944) {
return 4;
}
}
}
else if (ssDCval >= 116) {
if (Rs < 834) {
return 8;
}
else if (Rs >= 834) {
return 2;
}
}
}
}
else if (AFD >= 18) {
return 8;
}
}
else if (tcor >= 60) {
if (diffTSC < 1) {
if (posBalance < 55) {
if (mu_mv_mag_sq < 3) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 409) {
return 1;
}
else if (Cs >= 409) {
return 8;
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (diffTSC >= 1) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
if (Rs < 820) {
return 4;
}
else if (Rs >= 820) {
return 8;
}
}
}
}
}
}
else if (AFD >= 22) {
if (mu_mv_mag_sq < 50) {
if (AFD < 23) {
if (ssDCval < 149) {
return 2;
}
else if (ssDCval >= 149) {
return 1;
}
}
else if (AFD >= 23) {
if (tcor < 68) {
if (refDCval < 78) {
if (SC < 1016) {
return 1;
}
else if (SC >= 1016) {
if (SC < 1066) {
return 4;
}
else if (SC >= 1066) {
return 1;
}
}
}
else if (refDCval >= 78) {
if (Rs < 485) {
if (Rs < 378) {
return 1;
}
else if (Rs >= 378) {
return 2;
}
}
else if (Rs >= 485) {
if (negBalance < 61) {
return 1;
}
else if (negBalance >= 61) {
if (Rs < 635) {
return 1;
}
else if (Rs >= 635) {
return 2;
}
}
}
}
}
else if (tcor >= 68) {
if (diffMVdiffVal < -6) {
return 1;
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 50) {
if (ssDCval < 94) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (MVDiff < 41) {
return 2;
}
else if (MVDiff >= 41) {
return 8;
}
}
}
else if (ssDCval >= 94) {
if (Rs < 988) {
return 2;
}
else if (Rs >= 988) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (MVDiff < 30) {
if (Rs < 1535) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
if (SC < 1729) {
if (SC < 1288) {
return 1;
}
else if (SC >= 1288) {
return 2;
}
}
else if (SC >= 1729) {
if (Rs < 1467) {
return 4;
}
else if (Rs >= 1467) {
return 1;
}
}
}
}
else if (Rs >= 1535) {
if (ssDCval < 115) {
if (diffMVdiffVal < -8) {
if (Rs < 1747) {
return 1;
}
else if (Rs >= 1747) {
return 4;
}
}
else if (diffMVdiffVal >= -8) {
if (AFD < 25) {
return 2;
}
else if (AFD >= 25) {
if (Cs < 335) {
return 2;
}
else if (Cs >= 335) {
if (Rs < 1985) {
return 4;
}
else if (Rs >= 1985) {
if (Rs < 2226) {
return 2;
}
else if (Rs >= 2226) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 115) {
if (AFD < 23) {
return 8;
}
else if (AFD >= 23) {
if (Rs < 1796) {
return 2;
}
else if (Rs >= 1796) {
return 1;
}
}
}
}
}
else if (MVDiff >= 30) {
if (diffAFD < 6) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 30) {
if (gchDC < 8) {
return 1;
}
else if (gchDC >= 8) {
if (Rs < 1423) {
return 1;
}
else if (Rs >= 1423) {
if (Rs < 1962) {
return 2;
}
else if (Rs >= 1962) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 30) {
if (ssDCval < 49) {
return 2;
}
else if (ssDCval >= 49) {
if (Rs < 993) {
if (Cs < 340) {
return 2;
}
else if (Cs >= 340) {
return 8;
}
}
else if (Rs >= 993) {
if (posBalance < 43) {
if (Rs < 1827) {
return 2;
}
else if (Rs >= 1827) {
return 1;
}
}
else if (posBalance >= 43) {
return 1;
}
}
}
}
}
else if (posBalance >= 53) {
if (Cs < 326) {
if (diffRsCsdiff < -130) {
if (Cs < 157) {
return 2;
}
else if (Cs >= 157) {
return 4;
}
}
else if (diffRsCsdiff >= -130) {
if (tcor < 19) {
if (Rs < 1413) {
return 4;
}
else if (Rs >= 1413) {
return 1;
}
}
else if (tcor >= 19) {
if (refDCval < 71) {
if (MVDiff < 40) {
return 2;
}
else if (MVDiff >= 40) {
return 1;
}
}
else if (refDCval >= 71) {
if (ssDCval < 123) {
if (RsCsDiff < 67) {
return 8;
}
else if (RsCsDiff >= 67) {
if (MVDiff < 189) {
if (Rs < 768) {
return 1;
}
else if (Rs >= 768) {
return 2;
}
}
else if (MVDiff >= 189) {
if (Rs < 922) {
return 2;
}
else if (Rs >= 922) {
return 1;
}
}
}
}
else if (ssDCval >= 123) {
if (tcor < 60) {
return 2;
}
else if (tcor >= 60) {
if (Rs < 932) {
return 4;
}
else if (Rs >= 932) {
return 1;
}
}
}
}
}
}
}
else if (Cs >= 326) {
if (posBalance < 63) {
if (gchDC < 2) {
if (Rs < 2068) {
if (RsCsDiff < 94) {
if (Rs < 1450) {
return 8;
}
else if (Rs >= 1450) {
return 4;
}
}
else if (RsCsDiff >= 94) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (ssDCval < 115) {
return 4;
}
else if (ssDCval >= 115) {
return 8;
}
}
}
}
else if (Rs >= 2068) {
return 8;
}
}
else if (gchDC >= 2) {
if (ssDCval < 91) {
if (AFD < 18) {
if (Rs < 1016) {
return 1;
}
else if (Rs >= 1016) {
return 8;
}
}
else if (AFD >= 18) {
return 4;
}
}
else if (ssDCval >= 91) {
if (MVDiff < 52) {
return 1;
}
else if (MVDiff >= 52) {
return 2;
}
}
}
}
else if (posBalance >= 63) {
if (diffAFD < 2) {
if (posBalance < 68) {
return 2;
}
else if (posBalance >= 68) {
return 1;
}
}
else if (diffAFD >= 2) {
if (mcTcor < 89) {
if (posBalance < 65) {
if (Rs < 1147) {
return 1;
}
else if (Rs >= 1147) {
return 2;
}
}
else if (posBalance >= 65) {
return 8;
}
}
else if (mcTcor >= 89) {
return 1;
}
}
}
}
}
}
else if (diffAFD >= 6) {
if (Rs < 884) {
if (SC < 1216) {
if (AFD < 41) {
return 1;
}
else if (AFD >= 41) {
return 2;
}
}
else if (SC >= 1216) {
return 8;
}
}
else if (Rs >= 884) {
if (RsDiff < 19) {
if (AFD < 50) {
return 1;
}
else if (AFD >= 50) {
return 8;
}
}
else if (RsDiff >= 19) {
if (Rs < 2201) {
return 2;
}
else if (Rs >= 2201) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 74) {
if (mu_mv_mag_sq < 81) {
if (RsCsDiff < 77) {
if (Cs < 139) {
if (refDCval < 83) {
if (negBalance < 29) {
if (TSC < 5) {
if (Cs < 132) {
return 2;
}
else if (Cs >= 132) {
return 1;
}
}
else if (TSC >= 5) {
return 1;
}
}
else if (negBalance >= 29) {
return 1;
}
}
else if (refDCval >= 83) {
if (RsCsDiff < 30) {
if (negBalance < 67) {
return 1;
}
else if (negBalance >= 67) {
if (Rs < 776) {
return 8;
}
else if (Rs >= 776) {
return 1;
}
}
}
else if (RsCsDiff >= 30) {
if (AFD < 13) {
if (ssDCval < 126) {
return 2;
}
else if (ssDCval >= 126) {
return 4;
}
}
else if (AFD >= 13) {
if (refDCval < 124) {
if (posBalance < 102) {
return 1;
}
else if (posBalance >= 102) {
return 2;
}
}
else if (refDCval >= 124) {
return 2;
}
}
}
}
}
else if (Cs >= 139) {
if (Rs < 596) {
if (posBalance < 72) {
if (mcTcor < 97) {
if (mu_mv_mag_sq < 17) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (Cs < 479) {
return 1;
}
else if (Cs >= 479) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (negBalance < 32) {
if (SC < 835) {
return 2;
}
else if (SC >= 835) {
return 8;
}
}
else if (negBalance >= 32) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (diffMVdiffVal < 6) {
return 4;
}
else if (diffMVdiffVal >= 6) {
if (Rs < 484) {
return 1;
}
else if (Rs >= 484) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 524) {
if (mu_mv_mag_sq < 63) {
return 8;
}
else if (mu_mv_mag_sq >= 63) {
return 4;
}
}
else if (Rs >= 524) {
return 4;
}
}
}
else if (posBalance >= 72) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
return 1;
}
}
}
else if (Rs >= 596) {
if (ssDCval < 139) {
if (Cs < 254) {
if (MVDiff < 14) {
if (AFD < 9) {
if (diffAFD < -2) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (AFD >= 9) {
if (Rs < 1262) {
return 1;
}
else if (Rs >= 1262) {
if (mu_mv_mag_sq < 13) {
return 2;
}
else if (mu_mv_mag_sq >= 13) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
}
}
else if (MVDiff >= 14) {
if (refDCval < 79) {
return 1;
}
else if (refDCval >= 79) {
if (refDCval < 144) {
if (ssDCval < 104) {
if (SC < 1138) {
if (negBalance < 51) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 1;
}
}
else if (negBalance >= 51) {
if (Rs < 822) {
return 8;
}
else if (Rs >= 822) {
return 2;
}
}
}
else if (SC >= 1138) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
if (ssDCval < 101) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
return 1;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
}
}
else if (ssDCval >= 104) {
return 1;
}
}
else if (refDCval >= 144) {
return 2;
}
}
}
}
else if (Cs >= 254) {
if (ssDCval < 104) {
if (ssDCval < 102) {
if (negBalance < 29) {
if (TSC < 4) {
if (AFD < 9) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
if (Rs < 1616) {
return 2;
}
else if (Rs >= 1616) {
return 4;
}
}
}
else if (AFD >= 9) {
if (Rs < 1223) {
return 2;
}
else if (Rs >= 1223) {
return 8;
}
}
}
else if (TSC >= 4) {
if (diffMVdiffVal < 1) {
if (posBalance < 43) {
return 1;
}
else if (posBalance >= 43) {
if (diffRsCsdiff < 1) {
if (Rs < 1760) {
return 1;
}
else if (Rs >= 1760) {
return 2;
}
}
else if (diffRsCsdiff >= 1) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
if (RsCsDiff < 19) {
return 2;
}
else if (RsCsDiff >= 19) {
return 1;
}
}
}
}
else if (negBalance >= 29) {
if (Rs < 1023) {
if (posBalance < 55) {
if (diffRsCsdiff < 11) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (diffRsCsdiff < 4) {
if (AFD < 14) {
if (gchDC < 2) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Cs < 453) {
return 1;
}
else if (Cs >= 453) {
return 8;
}
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (AFD >= 14) {
return 1;
}
}
else if (diffRsCsdiff >= 4) {
if (Rs < 798) {
return 1;
}
else if (Rs >= 798) {
if (Rs < 961) {
return 2;
}
else if (Rs >= 961) {
return 1;
}
}
}
}
}
else if (diffRsCsdiff >= 11) {
if (Cs < 378) {
if (Rs < 843) {
return 8;
}
else if (Rs >= 843) {
return 1;
}
}
else if (Cs >= 378) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (tcor < 81) {
if (diffRsCsdiff < -1) {
if (negBalance < 46) {
if (Rs < 863) {
return 2;
}
else if (Rs >= 863) {
return 4;
}
}
else if (negBalance >= 46) {
return 1;
}
}
else if (diffRsCsdiff >= -1) {
if (MVDiff < 15) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 1;
}
}
else if (MVDiff >= 15) {
if (AFD < 24) {
if (negBalance < 45) {
return 1;
}
else if (negBalance >= 45) {
return 4;
}
}
else if (AFD >= 24) {
return 8;
}
}
}
}
else if (tcor >= 81) {
if (MVDiff < 30) {
return 1;
}
else if (MVDiff >= 30) {
return 4;
}
}
}
}
else if (Rs >= 1023) {
if (refDCval < 97) {
if (refDCval < 65) {
if (CsDiff < 3) {
if (MVDiff < 7) {
if (Rs < 2150) {
return 2;
}
else if (Rs >= 2150) {
return 4;
}
}
else if (MVDiff >= 7) {
if (Rs < 1885) {
if (Cs < 476) {
return 1;
}
else if (Cs >= 476) {
return 8;
}
}
else if (Rs >= 1885) {
if (Rs < 2031) {
return 2;
}
else if (Rs >= 2031) {
return 1;
}
}
}
}
else if (CsDiff >= 3) {
if (mcTcor < 97) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
if (MVDiff < 60) {
return 2;
}
else if (MVDiff >= 60) {
return 1;
}
}
}
else if (mcTcor >= 97) {
if (Rs < 1806) {
return 8;
}
else if (Rs >= 1806) {
return 4;
}
}
}
}
else if (refDCval >= 65) {
if (mcTcor < 94) {
if (mu_mv_mag_sq < 37) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
if (Rs < 1232) {
return 2;
}
else if (Rs >= 1232) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 1211) {
return 8;
}
else if (Rs >= 1211) {
return 4;
}
}
}
else if (mcTcor >= 94) {
if (diffRsCsdiff < -40) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Rs < 2200) {
return 2;
}
else if (Rs >= 2200) {
return 4;
}
}
}
else if (diffRsCsdiff >= -40) {
if (MVDiff < 34) {
if (SC < 2736) {
return 1;
}
else if (SC >= 2736) {
if (Cs < 458) {
return 2;
}
else if (Cs >= 458) {
return 1;
}
}
}
else if (MVDiff >= 34) {
if (MVDiff < 43) {
return 2;
}
else if (MVDiff >= 43) {
if (Rs < 1222) {
return 2;
}
else if (Rs >= 1222) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 97) {
if (Cs < 400) {
return 8;
}
else if (Cs >= 400) {
return 1;
}
}
}
}
}
else if (ssDCval >= 102) {
if (Rs < 1910) {
return 8;
}
else if (Rs >= 1910) {
return 4;
}
}
}
else if (ssDCval >= 104) {
if (Cs < 453) {
if (refDCval < 121) {
if (tcor < 91) {
if (mu_mv_mag_sq < 6) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (Cs < 302) {
return 1;
}
else if (Cs >= 302) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 6) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
else if (tcor >= 91) {
return 1;
}
}
else if (refDCval >= 121) {
return 1;
}
}
else if (Cs >= 453) {
if (Rs < 1826) {
if (AFD < 12) {
if (ssDCval < 129) {
return 2;
}
else if (ssDCval >= 129) {
if (Rs < 1402) {
return 1;
}
else if (Rs >= 1402) {
return 2;
}
}
}
else if (AFD >= 12) {
if (Rs < 1294) {
return 2;
}
else if (Rs >= 1294) {
return 1;
}
}
}
else if (Rs >= 1826) {
if (Cs < 486) {
return 1;
}
else if (Cs >= 486) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 139) {
if (gchDC < 2) {
if (SC < 1675) {
if (AFD < 13) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (AFD >= 13) {
if (Cs < 172) {
return 4;
}
else if (Cs >= 172) {
return 1;
}
}
}
else if (SC >= 1675) {
if (mcTcor < 96) {
if (Rs < 1670) {
return 4;
}
else if (Rs >= 1670) {
return 8;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
}
else if (gchDC >= 2) {
if (Rs < 848) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (Cs < 302) {
return 1;
}
else if (Cs >= 302) {
return 8;
}
}
}
else if (Rs >= 848) {
if (tcor < 78) {
if (Rs < 1490) {
return 1;
}
else if (Rs >= 1490) {
return 8;
}
}
else if (tcor >= 78) {
return 1;
}
}
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (ssDCval < 113) {
if (MVDiff < 37) {
if (Rs < 2058) {
if (diffRsCsdiff < 37) {
if (posBalance < 68) {
if (mu_mv_mag_sq < 17) {
if (posBalance < 56) {
if (Rs < 1540) {
return 1;
}
else if (Rs >= 1540) {
if (Cs < 352) {
return 2;
}
else if (Cs >= 352) {
if (SC < 2354) {
return 1;
}
else if (SC >= 2354) {
return 2;
}
}
}
}
else if (posBalance >= 56) {
return 2;
}
}
else if (mu_mv_mag_sq >= 17) {
if (mcTcor < 98) {
if (Cs < 365) {
if (SC < 1719) {
if (MVDiff < 32) {
return 1;
}
else if (MVDiff >= 32) {
return 2;
}
}
else if (SC >= 1719) {
return 2;
}
}
else if (Cs >= 365) {
return 1;
}
}
else if (mcTcor >= 98) {
return 2;
}
}
}
else if (posBalance >= 68) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 1;
}
}
}
else if (diffRsCsdiff >= 37) {
if (refDCval < 107) {
if (AFD < 25) {
return 1;
}
else if (AFD >= 25) {
return 2;
}
}
else if (refDCval >= 107) {
return 2;
}
}
}
else if (Rs >= 2058) {
if (negBalance < 42) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (negBalance >= 42) {
if (refDCval < 106) {
if (diffRsCsdiff < -97) {
if (Cs < 323) {
return 1;
}
else if (Cs >= 323) {
return 2;
}
}
else if (diffRsCsdiff >= -97) {
return 2;
}
}
else if (refDCval >= 106) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
}
else if (MVDiff >= 37) {
if (negBalance < 58) {
if (mu_mv_mag_sq < 7) {
if (mcTcor < 93) {
if (Cs < 250) {
if (Rs < 2153) {
return 1;
}
else if (Rs >= 2153) {
return 4;
}
}
else if (Cs >= 250) {
return 8;
}
}
else if (mcTcor >= 93) {
if (posBalance < 51) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
else if (posBalance >= 51) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (ssDCval < 82) {
if (Rs < 738) {
return 2;
}
else if (Rs >= 738) {
return 1;
}
}
else if (ssDCval >= 82) {
if (ssDCval < 107) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
}
else if (ssDCval >= 107) {
if (Rs < 2143) {
return 1;
}
else if (Rs >= 2143) {
return 2;
}
}
}
}
}
else if (negBalance >= 58) {
if (Cs < 432) {
if (mu_mv_mag_sq < 7) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (RsCsDiff < 86) {
if (Rs < 1552) {
return 1;
}
else if (Rs >= 1552) {
return 8;
}
}
else if (RsCsDiff >= 86) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (SC < 651) {
return 2;
}
else if (SC >= 651) {
return 1;
}
}
}
else if (Cs >= 432) {
if (Cs < 495) {
return 8;
}
else if (Cs >= 495) {
return 2;
}
}
}
}
}
else if (ssDCval >= 113) {
if (AFD < 13) {
if (tcor < 88) {
if (Rs < 1417) {
return 4;
}
else if (Rs >= 1417) {
return 2;
}
}
else if (tcor >= 88) {
return 1;
}
}
else if (AFD >= 13) {
if (mcTcor < 96) {
if (CsDiff < 3) {
if (MVDiff < 66) {
return 1;
}
else if (MVDiff >= 66) {
if (ssDCval < 132) {
if (Cs < 385) {
return 8;
}
else if (Cs >= 385) {
if (Rs < 2188) {
return 1;
}
else if (Rs >= 2188) {
return 8;
}
}
}
else if (ssDCval >= 132) {
if (Rs < 1604) {
return 1;
}
else if (Rs >= 1604) {
return 8;
}
}
}
}
else if (CsDiff >= 3) {
if (Cs < 499) {
if (mcTcor < 92) {
if (Rs < 1801) {
return 2;
}
else if (Rs >= 1801) {
return 1;
}
}
else if (mcTcor >= 92) {
return 1;
}
}
else if (Cs >= 499) {
if (Rs < 1736) {
return 1;
}
else if (Rs >= 1736) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (CsDiff < 5) {
if (SC < 2643) {
return 1;
}
else if (SC >= 2643) {
if (Cs < 398) {
return 8;
}
else if (Cs >= 398) {
return 1;
}
}
}
else if (CsDiff >= 5) {
if (Cs < 452) {
return 1;
}
else if (Cs >= 452) {
if (Rs < 1896) {
return 2;
}
else if (Rs >= 1896) {
return 1;
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 52) {
if (Cs < 408) {
if (Rs < 1485) {
if (ssDCval < 168) {
return 1;
}
else if (ssDCval >= 168) {
return 2;
}
}
else if (Rs >= 1485) {
if (Cs < 195) {
return 2;
}
else if (Cs >= 195) {
return 4;
}
}
}
else if (Cs >= 408) {
if (Cs < 502) {
return 8;
}
else if (Cs >= 502) {
return 1;
}
}
}
else if (negBalance >= 52) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 506) {
if (negBalance < 13) {
if (Rs < 709) {
if (ssDCval < 57) {
if (mu_mv_mag_sq < 127) {
if (refDCval < 32) {
if (Rs < 566) {
if (Rs < 463) {
return 4;
}
else if (Rs >= 463) {
return 2;
}
}
else if (Rs >= 566) {
return 8;
}
}
else if (refDCval >= 32) {
if (diffTSC < 4) {
if (Rs < 245) {
if (Rs < 226) {
return 1;
}
else if (Rs >= 226) {
return 2;
}
}
else if (Rs >= 245) {
return 1;
}
}
else if (diffTSC >= 4) {
if (Rs < 428) {
return 4;
}
else if (Rs >= 428) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 9) {
return 8;
}
else if (negBalance >= 9) {
if (Rs < 258) {
return 2;
}
else if (Rs >= 258) {
return 1;
}
}
}
}
else if (ssDCval >= 57) {
if (CsDiff < 1) {
if (Cs < 685) {
if (Rs < 507) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
if (Rs < 353) {
return 4;
}
else if (Rs >= 353) {
return 1;
}
}
}
else if (Rs >= 507) {
return 8;
}
}
else if (Cs >= 685) {
if (Rs < 656) {
if (Cs < 793) {
if (Rs < 512) {
return 4;
}
else if (Rs >= 512) {
return 2;
}
}
else if (Cs >= 793) {
return 8;
}
}
else if (Rs >= 656) {
return 1;
}
}
}
else if (CsDiff >= 1) {
if (refDCval < 143) {
if (tcor < 93) {
if (Rs < 565) {
if (SC < 1040) {
return 1;
}
else if (SC >= 1040) {
return 4;
}
}
else if (Rs >= 565) {
return 8;
}
}
else if (tcor >= 93) {
if (ssDCval < 101) {
return 8;
}
else if (ssDCval >= 101) {
return 1;
}
}
}
else if (refDCval >= 143) {
return 2;
}
}
}
}
else if (Rs >= 709) {
if (diffAFD < -3) {
if (Rs < 1601) {
if (ssDCval < 144) {
if (Cs < 642) {
if (Rs < 1184) {
return 2;
}
else if (Rs >= 1184) {
return 1;
}
}
else if (Cs >= 642) {
if (mcTcor < 99) {
if (diffAFD < -17) {
if (CsDiff < 1) {
return 4;
}
else if (CsDiff >= 1) {
if (Rs < 1210) {
return 8;
}
else if (Rs >= 1210) {
return 1;
}
}
}
else if (diffAFD >= -17) {
if (mcTcor < 77) {
if (Cs < 1076) {
return 1;
}
else if (Cs >= 1076) {
return 8;
}
}
else if (mcTcor >= 77) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
if (diffAFD < -9) {
if (tcor < 100) {
if (Rs < 996) {
return 8;
}
else if (Rs >= 996) {
if (Cs < 739) {
return 1;
}
else if (Cs >= 739) {
return 4;
}
}
}
else if (tcor >= 100) {
if (ssDCval < 138) {
if (posBalance < 5) {
if (Cs < 1258) {
if (Cs < 977) {
if (Rs < 980) {
return 1;
}
else if (Rs >= 980) {
return 8;
}
}
else if (Cs >= 977) {
return 1;
}
}
else if (Cs >= 1258) {
return 8;
}
}
else if (posBalance >= 5) {
return 4;
}
}
else if (ssDCval >= 138) {
return 8;
}
}
}
else if (diffAFD >= -9) {
if (diffRsCsdiff < -38) {
if (posBalance < 5) {
return 8;
}
else if (posBalance >= 5) {
return 4;
}
}
else if (diffRsCsdiff >= -38) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 144) {
if (diffAFD < -21) {
return 4;
}
else if (diffAFD >= -21) {
return 8;
}
}
}
else if (Rs >= 1601) {
if (Cs < 1266) {
return 8;
}
else if (Cs >= 1266) {
if (Rs < 1935) {
return 4;
}
else if (Rs >= 1935) {
return 1;
}
}
}
}
else if (diffAFD >= -3) {
if (Rs < 1841) {
if (ssDCval < 152) {
if (Rs < 874) {
if (Cs < 1167) {
if (refDCval < 93) {
if (ssDCval < 30) {
return 8;
}
else if (ssDCval >= 30) {
if (AFD < 2) {
if (Cs < 947) {
return 2;
}
else if (Cs >= 947) {
return 1;
}
}
else if (AFD >= 2) {
if (Rs < 833) {
return 1;
}
else if (Rs >= 833) {
return 8;
}
}
}
}
else if (refDCval >= 93) {
if (ssDCval < 135) {
if (mcTcor < 99) {
if (posBalance < 28) {
if (Rs < 719) {
return 8;
}
else if (Rs >= 719) {
return 1;
}
}
else if (posBalance >= 28) {
return 8;
}
}
else if (mcTcor >= 99) {
return 8;
}
}
else if (ssDCval >= 135) {
return 1;
}
}
}
else if (Cs >= 1167) {
return 2;
}
}
else if (Rs >= 874) {
if (SC < 1790) {
return 8;
}
else if (SC >= 1790) {
if (TSCindex < 1) {
if (diffMVdiffVal < -14) {
if (Rs < 1101) {
return 8;
}
else if (Rs >= 1101) {
return 4;
}
}
else if (diffMVdiffVal >= -14) {
if (Rs < 1196) {
if (Cs < 1238) {
if (diffMVdiffVal < -2) {
if (Rs < 1120) {
return 8;
}
else if (Rs >= 1120) {
return 1;
}
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (Cs >= 1238) {
return 1;
}
}
else if (Rs >= 1196) {
return 8;
}
}
}
else if (TSCindex >= 1) {
if (Cs < 657) {
if (Rs < 1515) {
return 1;
}
else if (Rs >= 1515) {
return 8;
}
}
else if (Cs >= 657) {
if (Cs < 848) {
if (TSC < 7) {
if (Cs < 845) {
if (Cs < 717) {
if (ssDCval < 66) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
return 8;
}
}
else if (ssDCval >= 66) {
return 8;
}
}
else if (Cs >= 717) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 1248) {
return 4;
}
else if (Rs >= 1248) {
return 8;
}
}
}
}
else if (Cs >= 845) {
return 4;
}
}
else if (TSC >= 7) {
return 4;
}
}
else if (Cs >= 848) {
if (mu_mv_mag_sq < 127) {
if (posBalance < 67) {
if (Rs < 1551) {
if (ssDCval < 31) {
return 8;
}
else if (ssDCval >= 31) {
if (ssDCval < 49) {
return 1;
}
else if (ssDCval >= 49) {
if (Cs < 1044) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Cs >= 1044) {
if (Rs < 1426) {
return 1;
}
else if (Rs >= 1426) {
return 8;
}
}
}
}
}
else if (Rs >= 1551) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 1827) {
return 1;
}
else if (Rs >= 1827) {
return 8;
}
}
}
}
else if (posBalance >= 67) {
return 8;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 1678) {
if (posBalance < 36) {
if (tcor < 100) {
return 8;
}
else if (tcor >= 100) {
if (ssDCval < 57) {
return 8;
}
else if (ssDCval >= 57) {
return 1;
}
}
}
else if (posBalance >= 36) {
if (Rs < 1077) {
return 8;
}
else if (Rs >= 1077) {
return 1;
}
}
}
else if (Rs >= 1678) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 152) {
if (Rs < 837) {
return 8;
}
else if (Rs >= 837) {
return 2;
}
}
}
else if (Rs >= 1841) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (refDCval < 79) {
if (Cs < 980) {
if (Cs < 638) {
return 2;
}
else if (Cs >= 638) {
if (SC < 2952) {
return 1;
}
else if (SC >= 2952) {
if (SC < 2962) {
return 4;
}
else if (SC >= 2962) {
return 1;
}
}
}
}
else if (Cs >= 980) {
if (refDCval < 61) {
return 4;
}
else if (refDCval >= 61) {
if (Rs < 1933) {
return 1;
}
else if (Rs >= 1933) {
if (Cs < 1233) {
if (Rs < 2140) {
return 4;
}
else if (Rs >= 2140) {
return 8;
}
}
else if (Cs >= 1233) {
return 2;
}
}
}
}
}
else if (refDCval >= 79) {
if (Cs < 627) {
return 2;
}
else if (Cs >= 627) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 13) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 45) {
if (refDCval < 47) {
if (Rs < 1575) {
if (refDCval < 29) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
return 8;
}
}
}
else if (refDCval >= 29) {
if (mcTcor < 89) {
if (MVDiff < 9) {
if (Rs < 958) {
return 1;
}
else if (Rs >= 958) {
return 8;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (mcTcor >= 89) {
if (mu_mv_mag_sq < 27) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 2;
}
}
else if (mu_mv_mag_sq >= 27) {
if (Cs < 615) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (Cs >= 615) {
if (Rs < 696) {
if (AFD < 5) {
if (Cs < 1003) {
return 4;
}
else if (Cs >= 1003) {
return 1;
}
}
else if (AFD >= 5) {
return 1;
}
}
else if (Rs >= 696) {
if (Rs < 1339) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
return 8;
}
}
else if (Rs >= 1339) {
if (Rs < 1390) {
return 2;
}
else if (Rs >= 1390) {
if (mu_mv_mag_sq < 118) {
if (Cs < 867) {
return 4;
}
else if (Cs >= 867) {
return 8;
}
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1575) {
if (TSC < 3) {
if (Cs < 969) {
return 4;
}
else if (Cs >= 969) {
return 8;
}
}
else if (TSC >= 3) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
if (Cs < 827) {
return 4;
}
else if (Cs >= 827) {
return 1;
}
}
}
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 101) {
if (refDCval < 88) {
if (mcTcor < 94) {
if (Rs < 1980) {
if (tcor < 73) {
if (Cs < 606) {
return 1;
}
else if (Cs >= 606) {
return 4;
}
}
else if (tcor >= 73) {
if (Rs < 1084) {
if (TSCindex < 2) {
return 4;
}
else if (TSCindex >= 2) {
if (negBalance < 32) {
if (Rs < 761) {
return 8;
}
else if (Rs >= 761) {
return 2;
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
else if (Rs >= 1084) {
if (Cs < 797) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (mcTcor < 91) {
if (Rs < 1362) {
return 8;
}
else if (Rs >= 1362) {
return 1;
}
}
else if (mcTcor >= 91) {
if (refDCval < 55) {
return 1;
}
else if (refDCval >= 55) {
if (negBalance < 39) {
return 2;
}
else if (negBalance >= 39) {
return 1;
}
}
}
}
}
else if (Cs >= 797) {
if (ssDCval < 59) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
if (Rs < 1366) {
return 1;
}
else if (Rs >= 1366) {
return 2;
}
}
}
}
else if (ssDCval >= 59) {
if (CsDiff < 5) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
return 4;
}
}
else if (CsDiff >= 5) {
if (Cs < 1061) {
return 2;
}
else if (Cs >= 1061) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1980) {
if (diffTSC < -1) {
if (Cs < 864) {
return 4;
}
else if (Cs >= 864) {
return 2;
}
}
else if (diffTSC >= -1) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
if (Cs < 902) {
return 4;
}
else if (Cs >= 902) {
if (Rs < 2249) {
return 2;
}
else if (Rs >= 2249) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 94) {
if (Cs < 990) {
if (Rs < 806) {
if (MVDiff < 25) {
if (CsDiff < 3) {
if (negBalance < 36) {
if (diffAFD < -2) {
if (Rs < 711) {
return 1;
}
else if (Rs >= 711) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Cs < 537) {
if (Rs < 600) {
return 2;
}
else if (Rs >= 600) {
return 4;
}
}
else if (Cs >= 537) {
return 4;
}
}
}
else if (negBalance >= 36) {
if (Cs < 519) {
return 4;
}
else if (Cs >= 519) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (SC < 1289) {
return 8;
}
else if (SC >= 1289) {
if (ssDCval < 53) {
if (Rs < 683) {
return 4;
}
else if (Rs >= 683) {
return 1;
}
}
else if (ssDCval >= 53) {
return 2;
}
}
}
}
else if (MVDiff >= 25) {
return 1;
}
}
else if (Rs >= 806) {
if (negBalance < 56) {
if (ssDCval < 69) {
if (refDCval < 55) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (diffRsCsdiff < -16) {
if (Rs < 1163) {
return 4;
}
else if (Rs >= 1163) {
return 2;
}
}
else if (diffRsCsdiff >= -16) {
if (Rs < 2024) {
return 1;
}
else if (Rs >= 2024) {
return 4;
}
}
}
}
else if (refDCval >= 55) {
if (posBalance < 36) {
return 1;
}
else if (posBalance >= 36) {
if (ssDCval < 61) {
if (MVDiff < 6) {
if (RsDiff < 3) {
if (Rs < 949) {
if (Rs < 851) {
return 8;
}
else if (Rs >= 851) {
return 4;
}
}
else if (Rs >= 949) {
if (Rs < 1228) {
return 1;
}
else if (Rs >= 1228) {
return 8;
}
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (MVDiff >= 6) {
if (diffRsCsdiff < 7) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (diffRsCsdiff >= 7) {
return 1;
}
}
}
else if (ssDCval >= 61) {
if (CsDiff < 4) {
if (Rs < 1481) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 2;
}
}
else if (Rs >= 1481) {
if (Rs < 1671) {
return 1;
}
else if (Rs >= 1671) {
return 2;
}
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
}
else if (ssDCval >= 69) {
if (mu_mv_mag_sq < 74) {
return 1;
}
else if (mu_mv_mag_sq >= 74) {
if (mu_mv_mag_sq < 82) {
if (Rs < 1296) {
return 8;
}
else if (Rs >= 1296) {
return 1;
}
}
else if (mu_mv_mag_sq >= 82) {
if (Rs < 1712) {
return 8;
}
else if (Rs >= 1712) {
return 4;
}
}
}
}
}
else if (negBalance >= 56) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
}
else if (Cs >= 990) {
if (tcor < 90) {
if (posBalance < 37) {
if (Rs < 1693) {
return 8;
}
else if (Rs >= 1693) {
return 1;
}
}
else if (posBalance >= 37) {
return 2;
}
}
else if (tcor >= 90) {
if (mu_mv_mag_sq < 85) {
if (Rs < 727) {
return 2;
}
else if (Rs >= 727) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (Rs < 2005) {
return 8;
}
else if (Rs >= 2005) {
if (Rs < 2312) {
return 2;
}
else if (Rs >= 2312) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
return 1;
}
}
}
}
}
else if (refDCval >= 88) {
if (negBalance < 30) {
if (Cs < 797) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Cs < 665) {
return 8;
}
else if (Cs >= 665) {
return 4;
}
}
}
else if (Cs >= 797) {
if (refDCval < 139) {
if (RsDiff < 6) {
if (mu_mv_mag_sq < 97) {
return 4;
}
else if (mu_mv_mag_sq >= 97) {
if (Rs < 864) {
return 4;
}
else if (Rs >= 864) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (Rs < 1378) {
return 1;
}
else if (Rs >= 1378) {
return 8;
}
}
}
else if (refDCval >= 139) {
if (Cs < 1246) {
return 8;
}
else if (Cs >= 1246) {
return 2;
}
}
}
}
else if (negBalance >= 30) {
if (tcor < 89) {
if (ssDCval < 154) {
if (negBalance < 45) {
return 8;
}
else if (negBalance >= 45) {
if (ssDCval < 93) {
if (Rs < 1878) {
return 8;
}
else if (Rs >= 1878) {
return 2;
}
}
else if (ssDCval >= 93) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
return 8;
}
}
}
}
else if (ssDCval >= 154) {
return 2;
}
}
else if (tcor >= 89) {
if (Cs < 1013) {
if (Rs < 977) {
if (ssDCval < 126) {
if (negBalance < 53) {
if (Cs < 594) {
return 8;
}
else if (Cs >= 594) {
return 2;
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (ssDCval >= 126) {
return 2;
}
}
else if (Rs >= 977) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (Rs < 1442) {
return 4;
}
else if (Rs >= 1442) {
return 8;
}
}
}
}
else if (Cs >= 1013) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (ssDCval < 56) {
if (ssDCval < 52) {
if (Rs < 529) {
return 2;
}
else if (Rs >= 529) {
if (ssDCval < 49) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
else if (ssDCval >= 49) {
return 8;
}
}
}
else if (ssDCval >= 52) {
if (mcTcor < 94) {
if (Rs < 1903) {
return 8;
}
else if (Rs >= 1903) {
return 1;
}
}
else if (mcTcor >= 94) {
return 1;
}
}
}
else if (ssDCval >= 56) {
if (posBalance < 33) {
if (CsDiff < 4) {
if (negBalance < 16) {
if (Rs < 972) {
return 1;
}
else if (Rs >= 972) {
return 8;
}
}
else if (negBalance >= 16) {
if (negBalance < 33) {
return 4;
}
else if (negBalance >= 33) {
if (Rs < 2262) {
return 8;
}
else if (Rs >= 2262) {
return 4;
}
}
}
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (posBalance >= 33) {
if (mu_mv_mag_sq < 114) {
if (SC < 2718) {
if (TSC < 3) {
if (Cs < 680) {
return 1;
}
else if (Cs >= 680) {
return 4;
}
}
else if (TSC >= 3) {
if (Rs < 583) {
if (SC < 1391) {
if (Rs < 494) {
return 1;
}
else if (Rs >= 494) {
return 2;
}
}
else if (SC >= 1391) {
return 8;
}
}
else if (Rs >= 583) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (negBalance < 28) {
if (Rs < 1066) {
return 1;
}
else if (Rs >= 1066) {
if (Rs < 1166) {
return 4;
}
else if (Rs >= 1166) {
return 8;
}
}
}
else if (negBalance >= 28) {
return 8;
}
}
}
}
}
else if (SC >= 2718) {
if (Rs < 1989) {
if (Rs < 1615) {
return 4;
}
else if (Rs >= 1615) {
return 1;
}
}
else if (Rs >= 1989) {
if (Rs < 2279) {
return 2;
}
else if (Rs >= 2279) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 114) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 45) {
if (mu_mv_mag_sq < 33) {
if (negBalance < 47) {
if (refDCval < 47) {
if (Rs < 825) {
if (posBalance < 56) {
if (Cs < 976) {
if (diffMVdiffVal < -3) {
if (ssDCval < 36) {
return 8;
}
else if (ssDCval >= 36) {
if (Cs < 687) {
return 2;
}
else if (Cs >= 687) {
return 4;
}
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (Cs >= 976) {
return 1;
}
}
else if (posBalance >= 56) {
if (negBalance < 22) {
if (Rs < 427) {
return 2;
}
else if (Rs >= 427) {
return 1;
}
}
else if (negBalance >= 22) {
if (SC < 1244) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (SC >= 1244) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
else if (Rs >= 825) {
if (mu_mv_mag_sq < 27) {
if (posBalance < 50) {
if (MVDiff < 17) {
if (Rs < 1866) {
return 8;
}
else if (Rs >= 1866) {
return 4;
}
}
else if (MVDiff >= 17) {
if (Cs < 699) {
return 2;
}
else if (Cs >= 699) {
return 8;
}
}
}
else if (posBalance >= 50) {
return 8;
}
}
else if (mu_mv_mag_sq >= 27) {
if (MVDiff < 24) {
return 2;
}
else if (MVDiff >= 24) {
if (Rs < 981) {
return 1;
}
else if (Rs >= 981) {
return 8;
}
}
}
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 3) {
if (mcTcor < 91) {
if (tcor < 78) {
if (Cs < 1033) {
if (mcTcor < 88) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
else if (mcTcor >= 88) {
if (ssDCval < 67) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
else if (ssDCval >= 67) {
return 8;
}
}
}
else if (Cs >= 1033) {
if (SC < 2552) {
if (diffMVdiffVal < 4) {
return 1;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 936) {
return 4;
}
else if (Rs >= 936) {
if (Rs < 1135) {
return 1;
}
else if (Rs >= 1135) {
return 8;
}
}
}
}
else if (SC >= 2552) {
return 8;
}
}
}
else if (tcor >= 78) {
if (RsDiff < 4) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 2;
}
}
else if (RsDiff >= 4) {
if (diffRsCsdiff < -27) {
return 8;
}
else if (diffRsCsdiff >= -27) {
if (ssDCval < 68) {
if (MVDiff < 7) {
if (Cs < 1215) {
return 8;
}
else if (Cs >= 1215) {
return 4;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (ssDCval >= 68) {
return 2;
}
}
}
}
}
else if (mcTcor >= 91) {
if (RsDiff < 7) {
if (RsDiff < 2) {
if (negBalance < 37) {
if (Rs < 645) {
return 4;
}
else if (Rs >= 645) {
if (Rs < 840) {
return 1;
}
else if (Rs >= 840) {
return 8;
}
}
}
else if (negBalance >= 37) {
return 8;
}
}
else if (RsDiff >= 2) {
if (TSC < 6) {
if (refDCval < 65) {
if (negBalance < 40) {
if (tcor < 88) {
if (TSC < 4) {
if (Rs < 621) {
return 1;
}
else if (Rs >= 621) {
return 4;
}
}
else if (TSC >= 4) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (tcor >= 88) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (Rs < 786) {
return 2;
}
else if (Rs >= 786) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
}
else if (negBalance >= 40) {
if (CsDiff < 4) {
if (diffMVdiffVal < -1) {
if (ssDCval < 54) {
return 8;
}
else if (ssDCval >= 54) {
if (Rs < 847) {
return 8;
}
else if (Rs >= 847) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (refDCval < 52) {
if (Cs < 684) {
return 1;
}
else if (Cs >= 684) {
return 8;
}
}
else if (refDCval >= 52) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 4;
}
}
}
}
else if (CsDiff >= 4) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 336) {
return 1;
}
else if (Rs >= 336) {
return 2;
}
}
}
}
}
else if (refDCval >= 65) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (AFD < 9) {
if (tcor < 90) {
if (negBalance < 34) {
if (negBalance < 31) {
return 2;
}
else if (negBalance >= 31) {
return 1;
}
}
else if (negBalance >= 34) {
if (posBalance < 55) {
if (Rs < 1754) {
return 2;
}
else if (Rs >= 1754) {
return 8;
}
}
else if (posBalance >= 55) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
return 8;
}
}
}
}
else if (tcor >= 90) {
if (MVDiff < 2) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (Rs < 622) {
return 4;
}
else if (Rs >= 622) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (diffTSC < 1) {
if (AFD < 7) {
if (ssDCval < 74) {
if (Rs < 1372) {
return 4;
}
else if (Rs >= 1372) {
return 1;
}
}
else if (ssDCval >= 74) {
return 1;
}
}
else if (AFD >= 7) {
return 1;
}
}
else if (diffTSC >= 1) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
return 2;
}
}
}
}
}
else if (AFD >= 9) {
if (tcor < 90) {
if (negBalance < 41) {
if (ssDCval < 83) {
return 1;
}
else if (ssDCval >= 83) {
if (negBalance < 35) {
return 1;
}
else if (negBalance >= 35) {
return 2;
}
}
}
else if (negBalance >= 41) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (tcor >= 90) {
return 2;
}
}
}
}
}
else if (TSC >= 6) {
if (diffAFD < 2) {
if (Cs < 606) {
if (Rs < 1553) {
if (AFD < 11) {
if (Rs < 1301) {
return 4;
}
else if (Rs >= 1301) {
return 1;
}
}
else if (AFD >= 11) {
return 1;
}
}
else if (Rs >= 1553) {
return 8;
}
}
else if (Cs >= 606) {
if (MVDiff < 7) {
if (negBalance < 42) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (Rs < 1527) {
return 1;
}
else if (Rs >= 1527) {
return 2;
}
}
}
else if (negBalance >= 42) {
if (Cs < 1117) {
return 8;
}
else if (Cs >= 1117) {
if (Rs < 1509) {
return 4;
}
else if (Rs >= 1509) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
if (Cs < 1157) {
if (RsDiff < 4) {
if (tcor < 90) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
return 8;
}
}
else if (tcor >= 90) {
return 4;
}
}
else if (RsDiff >= 4) {
if (gchDC < 2) {
if (CsDiff < 6) {
if (RsCsDiff < 29) {
return 2;
}
else if (RsCsDiff >= 29) {
if (ssDCval < 110) {
return 4;
}
else if (ssDCval >= 110) {
return 8;
}
}
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (Cs >= 1157) {
if (TSC < 7) {
if (Rs < 1073) {
return 2;
}
else if (Rs >= 1073) {
return 4;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
else if (diffAFD >= 2) {
if (Rs < 1375) {
return 8;
}
else if (Rs >= 1375) {
return 1;
}
}
}
}
}
else if (RsDiff >= 7) {
if (ssDCval < 73) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
return 2;
}
}
}
else if (ssDCval >= 73) {
if (Cs < 738) {
if (Rs < 1838) {
return 2;
}
else if (Rs >= 1838) {
return 1;
}
}
else if (Cs >= 738) {
if (TSC < 7) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
else if (TSC >= 7) {
if (RsCsDiff < 85) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (RsCsDiff >= 85) {
return 2;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (AFD < 11) {
if (Cs < 551) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
if (Rs < 1221) {
return 1;
}
else if (Rs >= 1221) {
return 4;
}
}
}
else if (Cs >= 551) {
if (Rs < 800) {
if (TSC < 5) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
if (ssDCval < 131) {
return 2;
}
else if (ssDCval >= 131) {
return 8;
}
}
}
else if (TSC >= 5) {
if (diffMVdiffVal < 3) {
return 2;
}
else if (diffMVdiffVal >= 3) {
if (Rs < 527) {
return 1;
}
else if (Rs >= 527) {
return 4;
}
}
}
}
else if (Rs >= 800) {
if (refDCval < 135) {
if (Cs < 1211) {
if (posBalance < 46) {
if (Cs < 800) {
return 4;
}
else if (Cs >= 800) {
return 8;
}
}
else if (posBalance >= 46) {
return 8;
}
}
else if (Cs >= 1211) {
if (Rs < 1358) {
return 1;
}
else if (Rs >= 1358) {
if (Rs < 1672) {
return 8;
}
else if (Rs >= 1672) {
return 2;
}
}
}
}
else if (refDCval >= 135) {
if (Cs < 1072) {
return 4;
}
else if (Cs >= 1072) {
if (Rs < 1431) {
return 1;
}
else if (Rs >= 1431) {
return 8;
}
}
}
}
}
}
else if (AFD >= 11) {
if (Cs < 1099) {
if (MVDiff < 16) {
if (RsDiff < 9) {
if (Rs < 546) {
return 2;
}
else if (Rs >= 546) {
if (Cs < 722) {
if (Rs < 1023) {
return 8;
}
else if (Rs >= 1023) {
return 1;
}
}
else if (Cs >= 722) {
return 8;
}
}
}
else if (RsDiff >= 9) {
return 2;
}
}
else if (MVDiff >= 16) {
if (RsCsDiff < 38) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (RsCsDiff >= 38) {
if (ssDCval < 71) {
if (Rs < 810) {
return 4;
}
else if (Rs >= 810) {
return 2;
}
}
else if (ssDCval >= 71) {
if (Rs < 793) {
return 2;
}
else if (Rs >= 793) {
return 1;
}
}
}
}
}
else if (Cs >= 1099) {
if (ssDCval < 106) {
return 1;
}
else if (ssDCval >= 106) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 908) {
if (Rs < 415) {
if (ssDCval < 110) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 110) {
return 8;
}
}
else if (Rs >= 415) {
if (AFD < 11) {
if (MVDiff < 6) {
if (ssDCval < 61) {
if (ssDCval < 52) {
if (ssDCval < 44) {
if (negBalance < 52) {
return 8;
}
else if (negBalance >= 52) {
return 4;
}
}
else if (ssDCval >= 44) {
if (Cs < 757) {
return 1;
}
else if (Cs >= 757) {
if (Rs < 849) {
return 8;
}
else if (Rs >= 849) {
return 1;
}
}
}
}
else if (ssDCval >= 52) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 4;
}
}
}
else if (ssDCval >= 61) {
if (ssDCval < 100) {
if (Cs < 526) {
return 1;
}
else if (Cs >= 526) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 1;
}
}
}
else if (ssDCval >= 100) {
if (posBalance < 55) {
return 2;
}
else if (posBalance >= 55) {
return 1;
}
}
}
}
else if (MVDiff >= 6) {
if (negBalance < 52) {
if (MVDiff < 16) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
if (ssDCval < 71) {
if (Rs < 708) {
return 4;
}
else if (Rs >= 708) {
if (Rs < 868) {
return 8;
}
else if (Rs >= 868) {
return 4;
}
}
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (MVDiff >= 16) {
if (Cs < 550) {
if (Rs < 516) {
return 8;
}
else if (Rs >= 516) {
return 2;
}
}
else if (Cs >= 550) {
if (Cs < 635) {
return 8;
}
else if (Cs >= 635) {
if (Rs < 666) {
return 2;
}
else if (Rs >= 666) {
return 4;
}
}
}
}
}
else if (negBalance >= 52) {
if (mu_mv_mag_sq < 2) {
if (Cs < 756) {
if (Rs < 729) {
if (SC < 1197) {
if (Cs < 542) {
if (Rs < 501) {
return 8;
}
else if (Rs >= 501) {
return 1;
}
}
else if (Cs >= 542) {
return 4;
}
}
else if (SC >= 1197) {
return 8;
}
}
else if (Rs >= 729) {
if (tcor < 77) {
if (MVDiff < 10) {
if (Rs < 757) {
return 1;
}
else if (Rs >= 757) {
return 8;
}
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (tcor >= 77) {
return 8;
}
}
}
else if (Cs >= 756) {
if (ssDCval < 95) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (Rs < 760) {
return 8;
}
else if (Rs >= 760) {
return 4;
}
}
}
else if (ssDCval >= 95) {
if (Rs < 893) {
return 1;
}
else if (Rs >= 893) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (ssDCval < 135) {
if (Cs < 1034) {
return 8;
}
else if (Cs >= 1034) {
return 4;
}
}
else if (ssDCval >= 135) {
return 2;
}
}
}
}
}
else if (AFD >= 11) {
if (mcTcor < 93) {
if (Cs < 819) {
if (ssDCval < 59) {
if (MVDiff < 11) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 4;
}
}
else if (MVDiff >= 11) {
if (mu_mv_mag_sq < 3) {
if (negBalance < 54) {
if (Cs < 531) {
return 8;
}
else if (Cs >= 531) {
if (SC < 1467) {
return 4;
}
else if (SC >= 1467) {
return 1;
}
}
}
else if (negBalance >= 54) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
}
else if (ssDCval >= 59) {
if (ssDCval < 69) {
return 1;
}
else if (ssDCval >= 69) {
return 8;
}
}
}
else if (Cs >= 819) {
if (Rs < 898) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (Cs < 837) {
return 2;
}
else if (Cs >= 837) {
return 4;
}
}
}
else if (Rs >= 898) {
if (Rs < 903) {
return 8;
}
else if (Rs >= 903) {
return 1;
}
}
}
}
else if (mcTcor >= 93) {
if (gchDC < 2) {
if (ssDCval < 93) {
if (negBalance < 54) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (TSC < 6) {
if (Cs < 558) {
return 4;
}
else if (Cs >= 558) {
return 2;
}
}
else if (TSC >= 6) {
return 4;
}
}
}
else if (negBalance >= 54) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 2;
}
}
}
else if (ssDCval >= 93) {
return 1;
}
}
else if (gchDC >= 2) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
}
}
}
}
else if (Rs >= 908) {
if (MVDiff < 12) {
if (tcor < 80) {
if (MVDiff < 3) {
if (ssDCval < 47) {
return 1;
}
else if (ssDCval >= 47) {
if (Cs < 1298) {
return 8;
}
else if (Cs >= 1298) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (mcTcor < 92) {
if (TSC < 9) {
if (diffMVdiffVal < -6) {
if (Cs < 905) {
return 8;
}
else if (Cs >= 905) {
return 2;
}
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
else if (TSC >= 9) {
if (Rs < 1779) {
return 2;
}
else if (Rs >= 1779) {
return 8;
}
}
}
else if (mcTcor >= 92) {
if (SC < 1673) {
return 8;
}
else if (SC >= 1673) {
return 2;
}
}
}
}
else if (tcor >= 80) {
if (SC < 2052) {
if (RsDiff < 5) {
if (diffRsCsdiff < 11) {
return 8;
}
else if (diffRsCsdiff >= 11) {
return 1;
}
}
else if (RsDiff >= 5) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (Rs < 934) {
return 4;
}
else if (Rs >= 934) {
return 8;
}
}
}
}
else if (SC >= 2052) {
if (Rs < 1138) {
if (Rs < 969) {
return 1;
}
else if (Rs >= 969) {
return 4;
}
}
else if (Rs >= 1138) {
if (Cs < 705) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
else if (Cs >= 705) {
if (posBalance < 49) {
if (Rs < 1883) {
if (Rs < 1469) {
return 8;
}
else if (Rs >= 1469) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 4;
}
}
}
else if (Rs >= 1883) {
return 4;
}
}
else if (posBalance >= 49) {
if (ssDCval < 98) {
if (refDCval < 64) {
if (AFD < 12) {
if (Rs < 1247) {
if (Cs < 1134) {
return 8;
}
else if (Cs >= 1134) {
return 4;
}
}
else if (Rs >= 1247) {
return 8;
}
}
else if (AFD >= 12) {
return 4;
}
}
else if (refDCval >= 64) {
if (TSC < 6) {
if (ssDCval < 72) {
if (Rs < 1722) {
return 4;
}
else if (Rs >= 1722) {
return 2;
}
}
else if (ssDCval >= 72) {
return 1;
}
}
else if (TSC >= 6) {
if (tcor < 84) {
if (Rs < 1542) {
if (Rs < 1321) {
return 2;
}
else if (Rs >= 1321) {
return 8;
}
}
else if (Rs >= 1542) {
return 4;
}
}
else if (tcor >= 84) {
if (diffRsCsdiff < 12) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
if (Rs < 1281) {
return 8;
}
else if (Rs >= 1281) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 12) {
if (diffRsCsdiff < 19) {
return 1;
}
else if (diffRsCsdiff >= 19) {
if (Rs < 1876) {
return 8;
}
else if (Rs >= 1876) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 98) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
}
}
}
else if (MVDiff >= 12) {
if (ssDCval < 63) {
if (Cs < 1285) {
if (posBalance < 51) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (posBalance >= 51) {
if (posBalance < 63) {
if (Rs < 1616) {
return 8;
}
else if (Rs >= 1616) {
if (Rs < 1645) {
return 1;
}
else if (Rs >= 1645) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (Rs < 1028) {
return 4;
}
else if (Rs >= 1028) {
return 8;
}
}
}
}
else if (Cs >= 1285) {
if (Rs < 977) {
return 4;
}
else if (Rs >= 977) {
return 2;
}
}
}
else if (ssDCval >= 63) {
if (mcTcor < 79) {
if (Rs < 1390) {
return 1;
}
else if (Rs >= 1390) {
return 4;
}
}
else if (mcTcor >= 79) {
if (tcor < 80) {
if (ssDCval < 98) {
if (tcor < 75) {
if (diffRsCsdiff < -22) {
return 2;
}
else if (diffRsCsdiff >= -22) {
if (AFD < 10) {
if (Rs < 1166) {
return 2;
}
else if (Rs >= 1166) {
return 8;
}
}
else if (AFD >= 10) {
return 8;
}
}
}
else if (tcor >= 75) {
if (Rs < 1159) {
return 8;
}
else if (Rs >= 1159) {
if (Rs < 1950) {
if (diffMVdiffVal < -12) {
return 8;
}
else if (diffMVdiffVal >= -12) {
return 2;
}
}
else if (Rs >= 1950) {
return 8;
}
}
}
}
else if (ssDCval >= 98) {
if (tcor < 71) {
if (Rs < 1480) {
if (Rs < 1325) {
return 1;
}
else if (Rs >= 1325) {
return 2;
}
}
else if (Rs >= 1480) {
if (Rs < 1683) {
return 8;
}
else if (Rs >= 1683) {
return 4;
}
}
}
else if (tcor >= 71) {
return 8;
}
}
}
else if (tcor >= 80) {
if (Rs < 1088) {
if (CsDiff < 5) {
if (RsCsDiff < 38) {
return 4;
}
else if (RsCsDiff >= 38) {
if (Cs < 1092) {
if (Rs < 971) {
return 8;
}
else if (Rs >= 971) {
return 1;
}
}
else if (Cs >= 1092) {
return 4;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 990) {
return 2;
}
else if (Rs >= 990) {
return 1;
}
}
}
else if (Rs >= 1088) {
if (Cs < 834) {
if (ssDCval < 112) {
if (negBalance < 55) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
if (Rs < 1882) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
if (Cs < 608) {
if (Rs < 1705) {
return 1;
}
else if (Rs >= 1705) {
return 8;
}
}
else if (Cs >= 608) {
if (negBalance < 54) {
return 4;
}
else if (negBalance >= 54) {
return 1;
}
}
}
}
else if (Rs >= 1882) {
if (Cs < 817) {
return 2;
}
else if (Cs >= 817) {
return 1;
}
}
}
}
else if (negBalance >= 55) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
return 2;
}
}
}
else if (ssDCval >= 112) {
if (RsCsDiff < 23) {
if (Rs < 1347) {
return 1;
}
else if (Rs >= 1347) {
return 4;
}
}
else if (RsCsDiff >= 23) {
return 8;
}
}
}
else if (Cs >= 834) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 3) {
if (tcor < 92) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (Cs < 1105) {
return 4;
}
else if (Cs >= 1105) {
if (Rs < 1620) {
return 8;
}
else if (Rs >= 1620) {
return 4;
}
}
}
}
else if (tcor >= 92) {
if (Cs < 1098) {
return 8;
}
else if (Cs >= 1098) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 132) {
if (Rs < 2095) {
return 8;
}
else if (Rs >= 2095) {
return 2;
}
}
else if (ssDCval >= 132) {
return 4;
}
}
}
else if (gchDC >= 2) {
if (Rs < 1132) {
return 8;
}
else if (Rs >= 1132) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 33) {
if (Rs < 1596) {
if (TSC < 7) {
if (refDCval < 47) {
if (mu_mv_mag_sq < 108) {
if (ssDCval < 39) {
if (Rs < 537) {
if (Rs < 256) {
return 1;
}
else if (Rs >= 256) {
return 2;
}
}
else if (Rs >= 537) {
if (Rs < 712) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (refDCval < 37) {
return 1;
}
else if (refDCval >= 37) {
return 8;
}
}
}
else if (Rs >= 712) {
return 8;
}
}
}
else if (ssDCval >= 39) {
if (SC < 1735) {
if (RsCsDiff < 19) {
if (mcTcor < 91) {
if (Rs < 497) {
return 1;
}
else if (Rs >= 497) {
return 8;
}
}
else if (mcTcor >= 91) {
if (Cs < 679) {
if (Rs < 1084) {
return 8;
}
else if (Rs >= 1084) {
return 2;
}
}
else if (Cs >= 679) {
return 4;
}
}
}
else if (RsCsDiff >= 19) {
if (negBalance < 34) {
if (Cs < 793) {
return 1;
}
else if (Cs >= 793) {
return 2;
}
}
else if (negBalance >= 34) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
if (Cs < 703) {
return 2;
}
else if (Cs >= 703) {
return 1;
}
}
}
}
}
else if (SC >= 1735) {
if (Rs < 1304) {
if (diffRsCsdiff < -15) {
if (diffMVdiffVal < -3) {
return 2;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (diffRsCsdiff >= -15) {
if (Rs < 1275) {
return 8;
}
else if (Rs >= 1275) {
if (Rs < 1291) {
return 1;
}
else if (Rs >= 1291) {
return 8;
}
}
}
}
else if (Rs >= 1304) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
if (Rs < 1377) {
return 2;
}
else if (Rs >= 1377) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (SC < 2210) {
if (Rs < 867) {
return 1;
}
else if (Rs >= 867) {
if (Rs < 1089) {
return 8;
}
else if (Rs >= 1089) {
if (ssDCval < 45) {
return 8;
}
else if (ssDCval >= 45) {
return 1;
}
}
}
}
else if (SC >= 2210) {
if (posBalance < 54) {
if (SC < 2249) {
return 1;
}
else if (SC >= 2249) {
return 8;
}
}
else if (posBalance >= 54) {
return 8;
}
}
}
}
else if (refDCval >= 47) {
if (ssDCval < 160) {
if (negBalance < 29) {
if (posBalance < 82) {
if (mu_mv_mag_sq < 95) {
if (CsDiff < 2) {
if (ssDCval < 57) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (ssDCval >= 57) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (Cs < 815) {
return 1;
}
else if (Cs >= 815) {
return 8;
}
}
}
}
else if (CsDiff >= 2) {
if (ssDCval < 57) {
if (mcTcor < 90) {
if (Rs < 1292) {
return 1;
}
else if (Rs >= 1292) {
return 2;
}
}
else if (mcTcor >= 90) {
if (Rs < 426) {
return 2;
}
else if (Rs >= 426) {
return 4;
}
}
}
else if (ssDCval >= 57) {
if (Cs < 896) {
if (negBalance < 25) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
else if (negBalance >= 25) {
if (diffAFD < 1) {
return 2;
}
else if (diffAFD >= 1) {
if (Cs < 764) {
return 8;
}
else if (Cs >= 764) {
return 2;
}
}
}
}
else if (Cs >= 896) {
if (Rs < 764) {
return 1;
}
else if (Rs >= 764) {
if (Rs < 940) {
return 4;
}
else if (Rs >= 940) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (tcor < 83) {
return 4;
}
else if (tcor >= 83) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 123) {
if (diffAFD < 1) {
if (SC < 1773) {
if (posBalance < 54) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (posBalance >= 54) {
if (ssDCval < 52) {
return 1;
}
else if (ssDCval >= 52) {
if (Rs < 485) {
return 1;
}
else if (Rs >= 485) {
return 4;
}
}
}
}
else if (SC >= 1773) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (Rs < 543) {
return 1;
}
else if (Rs >= 543) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
return 8;
}
}
else if (RsDiff >= 2) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (Cs < 1038) {
return 8;
}
else if (Cs >= 1038) {
if (Cs < 1249) {
return 4;
}
else if (Cs >= 1249) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 82) {
return 8;
}
}
else if (negBalance >= 29) {
if (negBalance < 51) {
if (refDCval < 75) {
if (Rs < 526) {
if (Cs < 804) {
if (tcor < 88) {
if (Cs < 729) {
return 1;
}
else if (Cs >= 729) {
return 8;
}
}
else if (tcor >= 88) {
if (SC < 1007) {
return 2;
}
else if (SC >= 1007) {
return 8;
}
}
}
else if (Cs >= 804) {
if (ssDCval < 53) {
return 2;
}
else if (ssDCval >= 53) {
return 1;
}
}
}
else if (Rs >= 526) {
if (ssDCval < 59) {
if (Rs < 1145) {
if (RsDiff < 3) {
if (negBalance < 46) {
if (Rs < 841) {
if (ssDCval < 55) {
if (mu_mv_mag_sq < 96) {
if (negBalance < 33) {
if (Rs < 649) {
return 1;
}
else if (Rs >= 649) {
return 4;
}
}
else if (negBalance >= 33) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
return 4;
}
}
else if (ssDCval >= 55) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 8;
}
}
}
else if (Rs >= 841) {
return 8;
}
}
else if (negBalance >= 46) {
if (mu_mv_mag_sq < 66) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
else if (mu_mv_mag_sq >= 66) {
return 4;
}
}
}
else if (RsDiff >= 3) {
if (Cs < 842) {
if (Rs < 757) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
return 1;
}
}
else if (Rs >= 757) {
if (negBalance < 37) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (negBalance >= 37) {
return 8;
}
}
}
else if (Cs >= 842) {
if (ssDCval < 50) {
if (Rs < 953) {
return 4;
}
else if (Rs >= 953) {
return 8;
}
}
else if (ssDCval >= 50) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
return 8;
}
}
}
}
}
else if (Rs >= 1145) {
if (MVDiff < 10) {
if (AFD < 8) {
if (refDCval < 48) {
if (Rs < 1361) {
return 8;
}
else if (Rs >= 1361) {
return 2;
}
}
else if (refDCval >= 48) {
return 8;
}
}
else if (AFD >= 8) {
if (Cs < 1205) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (Cs >= 1205) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (ssDCval < 55) {
if (Rs < 1486) {
if (Rs < 1168) {
return 4;
}
else if (Rs >= 1168) {
return 8;
}
}
else if (Rs >= 1486) {
return 1;
}
}
else if (ssDCval >= 55) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 59) {
if (Rs < 1306) {
if (diffRsCsdiff < 15) {
if (posBalance < 51) {
if (refDCval < 66) {
if (refDCval < 59) {
if (Rs < 1006) {
return 1;
}
else if (Rs >= 1006) {
return 2;
}
}
else if (refDCval >= 59) {
if (Rs < 1039) {
if (MVDiff < 5) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
if (Rs < 776) {
return 8;
}
else if (Rs >= 776) {
return 4;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (Rs >= 1039) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 8;
}
}
}
}
else if (refDCval >= 66) {
if (Cs < 844) {
if (Cs < 591) {
return 4;
}
else if (Cs >= 591) {
return 1;
}
}
else if (Cs >= 844) {
if (Rs < 982) {
return 2;
}
else if (Rs >= 982) {
if (Rs < 1243) {
return 8;
}
else if (Rs >= 1243) {
return 1;
}
}
}
}
}
else if (posBalance >= 51) {
if (posBalance < 58) {
if (mu_mv_mag_sq < 100) {
if (RsCsDiff < 7) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (RsCsDiff >= 7) {
if (mu_mv_mag_sq < 45) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (Cs < 630) {
if (negBalance < 38) {
if (mu_mv_mag_sq < 72) {
return 1;
}
else if (mu_mv_mag_sq >= 72) {
return 8;
}
}
else if (negBalance >= 38) {
if (Cs < 606) {
return 4;
}
else if (Cs >= 606) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
else if (Cs >= 630) {
if (Cs < 1102) {
if (TSC < 6) {
if (Rs < 597) {
return 8;
}
else if (Rs >= 597) {
if (MVDiff < 13) {
return 1;
}
else if (MVDiff >= 13) {
if (Rs < 1003) {
return 1;
}
else if (Rs >= 1003) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (Cs < 779) {
return 1;
}
else if (Cs >= 779) {
return 8;
}
}
}
else if (Cs >= 1102) {
if (tcor < 93) {
if (Cs < 1175) {
if (Rs < 1196) {
return 4;
}
else if (Rs >= 1196) {
return 8;
}
}
else if (Cs >= 1175) {
if (ssDCval < 70) {
if (diffTSC < -1) {
if (Rs < 973) {
return 4;
}
else if (Rs >= 973) {
return 1;
}
}
else if (diffTSC >= -1) {
return 1;
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
else if (tcor >= 93) {
if (Rs < 998) {
if (Rs < 845) {
return 8;
}
else if (Rs >= 845) {
return 4;
}
}
else if (Rs >= 998) {
if (Rs < 1153) {
return 8;
}
else if (Rs >= 1153) {
return 4;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 100) {
if (negBalance < 41) {
if (Rs < 1154) {
if (Cs < 1174) {
return 4;
}
else if (Cs >= 1174) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (Rs >= 1154) {
if (RsDiff < 3) {
if (diffRsCsdiff < -7) {
return 4;
}
else if (diffRsCsdiff >= -7) {
if (Rs < 1262) {
if (Cs < 1075) {
return 8;
}
else if (Cs >= 1075) {
return 1;
}
}
else if (Rs >= 1262) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
}
}
else if (negBalance >= 41) {
if (TSC < 6) {
if (Rs < 998) {
return 8;
}
else if (Rs >= 998) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (posBalance >= 58) {
if (Cs < 553) {
if (Rs < 1001) {
return 4;
}
else if (Rs >= 1001) {
return 1;
}
}
else if (Cs >= 553) {
if (negBalance < 45) {
if (RsCsDiff < 12) {
if (AFD < 5) {
if (Rs < 705) {
return 8;
}
else if (Rs >= 705) {
if (Rs < 817) {
return 4;
}
else if (Rs >= 817) {
return 2;
}
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (RsCsDiff >= 12) {
if (diffRsCsdiff < 3) {
if (AFD < 10) {
if (Rs < 863) {
return 8;
}
else if (Rs >= 863) {
return 4;
}
}
else if (AFD >= 10) {
if (Rs < 989) {
return 1;
}
else if (Rs >= 989) {
return 8;
}
}
}
else if (diffRsCsdiff >= 3) {
if (ssDCval < 62) {
if (Rs < 1167) {
return 4;
}
else if (Rs >= 1167) {
return 1;
}
}
else if (ssDCval >= 62) {
if (Rs < 1282) {
return 8;
}
else if (Rs >= 1282) {
return 1;
}
}
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
}
}
}
else if (diffRsCsdiff >= 15) {
if (Rs < 923) {
return 2;
}
else if (Rs >= 923) {
return 4;
}
}
}
else if (Rs >= 1306) {
if (RsCsDiff < 16) {
if (mu_mv_mag_sq < 123) {
if (ssDCval < 68) {
return 8;
}
else if (ssDCval >= 68) {
if (TSC < 5) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (Rs < 1487) {
return 8;
}
else if (Rs >= 1487) {
return 1;
}
}
}
else if (TSC >= 5) {
if (SC < 2478) {
return 8;
}
else if (SC >= 2478) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 1381) {
return 4;
}
else if (Rs >= 1381) {
return 8;
}
}
}
else if (RsCsDiff >= 16) {
if (refDCval < 65) {
if (diffTSC < -1) {
if (Rs < 1402) {
if (Cs < 1099) {
return 4;
}
else if (Cs >= 1099) {
return 1;
}
}
else if (Rs >= 1402) {
return 8;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (refDCval >= 65) {
if (mu_mv_mag_sq < 81) {
if (Cs < 919) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (Cs >= 919) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 43) {
if (diffTSC < 1) {
if (diffAFD < -1) {
if (Rs < 1345) {
return 1;
}
else if (Rs >= 1345) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (diffTSC >= 1) {
return 4;
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 75) {
if (refDCval < 130) {
if (RsCsDiff < 7) {
if (TSC < 6) {
if (ssDCval < 111) {
if (refDCval < 90) {
if (ssDCval < 77) {
return 2;
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (refDCval >= 90) {
if (diffMVdiffVal < -2) {
if (tcor < 93) {
if (Rs < 550) {
return 8;
}
else if (Rs >= 550) {
return 2;
}
}
else if (tcor >= 93) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 630) {
return 8;
}
else if (Cs >= 630) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (mu_mv_mag_sq < 112) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (mu_mv_mag_sq >= 112) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 111) {
return 8;
}
}
else if (TSC >= 6) {
if (Rs < 429) {
return 4;
}
else if (Rs >= 429) {
if (Rs < 522) {
return 1;
}
else if (Rs >= 522) {
return 8;
}
}
}
}
else if (RsCsDiff >= 7) {
if (Cs < 1013) {
if (posBalance < 50) {
if (ssDCval < 95) {
if (Cs < 859) {
return 2;
}
else if (Cs >= 859) {
if (Rs < 1166) {
return 1;
}
else if (Rs >= 1166) {
return 8;
}
}
}
else if (ssDCval >= 95) {
return 8;
}
}
else if (posBalance >= 50) {
if (tcor < 74) {
if (Rs < 830) {
return 2;
}
else if (Rs >= 830) {
return 1;
}
}
else if (tcor >= 74) {
if (TSC < 6) {
if (AFD < 7) {
if (MVDiff < 3) {
if (posBalance < 56) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (TSC < 5) {
if (Rs < 708) {
return 2;
}
else if (Rs >= 708) {
return 1;
}
}
else if (TSC >= 5) {
if (ssDCval < 80) {
if (Rs < 692) {
return 8;
}
else if (Rs >= 692) {
return 2;
}
}
else if (ssDCval >= 80) {
return 8;
}
}
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (AFD >= 7) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (RsCsDiff < 21) {
return 8;
}
else if (RsCsDiff >= 21) {
return 2;
}
}
}
}
else if (TSC >= 6) {
return 8;
}
}
}
}
else if (Cs >= 1013) {
if (mcTcor < 96) {
if (AFD < 10) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (AFD >= 10) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 1084) {
return 2;
}
else if (Rs >= 1084) {
return 4;
}
}
}
}
else if (mcTcor >= 96) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
if (MVDiff < 2) {
if (Rs < 940) {
return 4;
}
else if (Rs >= 940) {
return 8;
}
}
else if (MVDiff >= 2) {
if (Cs < 1293) {
return 8;
}
else if (Cs >= 1293) {
return 4;
}
}
}
}
}
}
}
else if (refDCval >= 130) {
if (Rs < 932) {
if (diffMVdiffVal < -8) {
return 4;
}
else if (diffMVdiffVal >= -8) {
return 8;
}
}
else if (Rs >= 932) {
if (TSC < 5) {
if (tcor < 95) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 4;
}
}
else if (tcor >= 95) {
if (Rs < 1420) {
return 1;
}
else if (Rs >= 1420) {
return 8;
}
}
}
else if (TSC >= 5) {
if (AFD < 7) {
if (MVDiff < 18) {
if (Cs < 703) {
return 4;
}
else if (Cs >= 703) {
return 8;
}
}
else if (MVDiff >= 18) {
if (Rs < 1405) {
return 4;
}
else if (Rs >= 1405) {
return 8;
}
}
}
else if (AFD >= 7) {
if (tcor < 88) {
if (Rs < 1041) {
return 8;
}
else if (Rs >= 1041) {
if (Rs < 1225) {
return 1;
}
else if (Rs >= 1225) {
return 4;
}
}
}
else if (tcor >= 88) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 51) {
if (Cs < 652) {
return 8;
}
else if (Cs >= 652) {
if (Rs < 445) {
if (ssDCval < 96) {
return 1;
}
else if (ssDCval >= 96) {
if (Rs < 424) {
return 4;
}
else if (Rs >= 424) {
return 1;
}
}
}
else if (Rs >= 445) {
if (SC < 2016) {
if (Cs < 657) {
return 4;
}
else if (Cs >= 657) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
if (Rs < 515) {
return 1;
}
else if (Rs >= 515) {
if (ssDCval < 117) {
return 8;
}
else if (ssDCval >= 117) {
return 1;
}
}
}
}
}
else if (SC >= 2016) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
if (ssDCval < 134) {
return 8;
}
else if (ssDCval >= 134) {
return 4;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 160) {
if (Rs < 842) {
return 4;
}
else if (Rs >= 842) {
return 2;
}
}
}
}
else if (TSC >= 7) {
if (negBalance < 45) {
if (refDCval < 77) {
if (refDCval < 53) {
if (tcor < 60) {
return 2;
}
else if (tcor >= 60) {
return 8;
}
}
else if (refDCval >= 53) {
if (posBalance < 60) {
if (posBalance < 57) {
if (tcor < 90) {
return 1;
}
else if (tcor >= 90) {
return 4;
}
}
else if (posBalance >= 57) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
if (Rs < 1228) {
return 8;
}
else if (Rs >= 1228) {
return 1;
}
}
}
}
else if (posBalance >= 60) {
if (gchDC < 3) {
if (Cs < 1242) {
return 4;
}
else if (Cs >= 1242) {
if (Rs < 1390) {
return 1;
}
else if (Rs >= 1390) {
return 4;
}
}
}
else if (gchDC >= 3) {
if (Rs < 875) {
return 1;
}
else if (Rs >= 875) {
return 8;
}
}
}
}
}
else if (refDCval >= 77) {
if (AFD < 11) {
if (MVDiff < 3) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
else if (AFD >= 11) {
if (TSC < 8) {
if (Rs < 1060) {
return 1;
}
else if (Rs >= 1060) {
return 2;
}
}
else if (TSC >= 8) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 2;
}
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 1015) {
if (Rs < 851) {
if (AFD < 10) {
if (MVDiff < 4) {
if (Cs < 753) {
return 4;
}
else if (Cs >= 753) {
return 8;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (AFD >= 10) {
if (diffRsCsdiff < -2) {
if (Rs < 746) {
return 1;
}
else if (Rs >= 746) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
else if (Rs >= 851) {
if (Cs < 945) {
if (ssDCval < 53) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 1;
}
}
else if (ssDCval >= 53) {
if (Rs < 1007) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (Rs >= 1007) {
return 1;
}
}
}
else if (Cs >= 945) {
if (RsCsDiff < 46) {
return 8;
}
else if (RsCsDiff >= 46) {
return 4;
}
}
}
}
else if (Rs >= 1015) {
if (MVDiff < 30) {
if (Cs < 1124) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (Rs < 1511) {
return 2;
}
else if (Rs >= 1511) {
return 8;
}
}
}
else if (Cs >= 1124) {
if (tcor < 93) {
if (Rs < 1312) {
if (diffRsCsdiff < -10) {
return 4;
}
else if (diffRsCsdiff >= -10) {
if (Rs < 1310) {
if (SC < 2528) {
return 8;
}
else if (SC >= 2528) {
return 1;
}
}
else if (Rs >= 1310) {
return 4;
}
}
}
else if (Rs >= 1312) {
return 8;
}
}
else if (tcor >= 93) {
return 1;
}
}
}
else if (MVDiff >= 30) {
if (Cs < 598) {
return 4;
}
else if (Cs >= 598) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1596) {
if (ssDCval < 72) {
if (refDCval < 62) {
if (Cs < 874) {
if (mu_mv_mag_sq < 55) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (mu_mv_mag_sq >= 55) {
if (diffTSC < -1) {
if (Cs < 830) {
return 2;
}
else if (Cs >= 830) {
return 1;
}
}
else if (diffTSC >= -1) {
if (Rs < 2281) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (Rs >= 2281) {
return 2;
}
}
}
}
else if (Cs >= 874) {
if (posBalance < 51) {
if (TSC < 6) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
return 8;
}
}
else if (TSC >= 6) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 4;
}
}
}
else if (posBalance >= 51) {
if (mu_mv_mag_sq < 89) {
return 8;
}
else if (mu_mv_mag_sq >= 89) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (Rs < 1797) {
return 2;
}
else if (Rs >= 1797) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 62) {
if (Cs < 1025) {
if (SC < 2517) {
if (SC < 2398) {
return 2;
}
else if (SC >= 2398) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 8;
}
}
}
else if (SC >= 2517) {
if (negBalance < 32) {
if (Rs < 1953) {
return 8;
}
else if (Rs >= 1953) {
return 1;
}
}
else if (negBalance >= 32) {
if (mcTcor < 91) {
if (Rs < 1750) {
return 2;
}
else if (Rs >= 1750) {
if (Rs < 2049) {
return 8;
}
else if (Rs >= 2049) {
return 4;
}
}
}
else if (mcTcor >= 91) {
return 2;
}
}
}
}
else if (Cs >= 1025) {
if (mu_mv_mag_sq < 77) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
else if (mu_mv_mag_sq >= 77) {
if (refDCval < 68) {
if (refDCval < 65) {
if (Cs < 1077) {
return 2;
}
else if (Cs >= 1077) {
return 8;
}
}
else if (refDCval >= 65) {
return 2;
}
}
else if (refDCval >= 68) {
return 8;
}
}
}
}
}
else if (ssDCval >= 72) {
if (refDCval < 79) {
if (negBalance < 47) {
if (mu_mv_mag_sq < 109) {
if (mcTcor < 94) {
if (Cs < 913) {
return 1;
}
else if (Cs >= 913) {
if (Rs < 2101) {
return 8;
}
else if (Rs >= 2101) {
return 4;
}
}
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (mu_mv_mag_sq >= 109) {
return 8;
}
}
else if (negBalance >= 47) {
if (TSC < 7) {
if (MVDiff < 8) {
return 1;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (refDCval >= 79) {
if (RsDiff < 9) {
if (MVDiff < 2) {
if (negBalance < 45) {
if (ssDCval < 126) {
if (diffAFD < 2) {
if (Cs < 1268) {
return 8;
}
else if (Cs >= 1268) {
if (Rs < 1765) {
return 1;
}
else if (Rs >= 1765) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (Rs < 1752) {
return 8;
}
else if (Rs >= 1752) {
return 2;
}
}
}
else if (ssDCval >= 126) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (negBalance >= 45) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (SC < 3068) {
return 1;
}
else if (SC >= 3068) {
if (mu_mv_mag_sq < 87) {
return 1;
}
else if (mu_mv_mag_sq >= 87) {
return 8;
}
}
}
}
}
else if (MVDiff >= 2) {
if (refDCval < 99) {
if (Cs < 851) {
if (Rs < 2003) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
else if (Rs >= 2003) {
return 1;
}
}
else if (Cs >= 851) {
if (refDCval < 90) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (ssDCval < 85) {
if (Rs < 2225) {
return 8;
}
else if (Rs >= 2225) {
return 2;
}
}
else if (ssDCval >= 85) {
if (Rs < 1968) {
return 4;
}
else if (Rs >= 1968) {
return 2;
}
}
}
}
else if (refDCval >= 90) {
if (TSC < 8) {
if (RsDiff < 5) {
if (Cs < 1187) {
return 8;
}
else if (Cs >= 1187) {
return 2;
}
}
else if (RsDiff >= 5) {
if (Rs < 2179) {
return 2;
}
else if (Rs >= 2179) {
return 4;
}
}
}
else if (TSC >= 8) {
if (RsCsDiff < 46) {
return 4;
}
else if (RsCsDiff >= 46) {
return 8;
}
}
}
}
}
else if (refDCval >= 99) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (negBalance < 49) {
if (MVDiff < 5) {
if (ssDCval < 123) {
return 8;
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (negBalance >= 49) {
if (TSC < 8) {
if (Rs < 2192) {
return 8;
}
else if (Rs >= 2192) {
return 2;
}
}
else if (TSC >= 8) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (Cs < 789) {
return 1;
}
else if (Cs >= 789) {
return 4;
}
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (refDCval < 111) {
if (MVDiff < 10) {
if (ssDCval < 90) {
if (Rs < 1781) {
if (refDCval < 58) {
if (negBalance < 50) {
if (Rs < 1283) {
return 1;
}
else if (Rs >= 1283) {
return 8;
}
}
else if (negBalance >= 50) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 1519) {
return 8;
}
else if (Rs >= 1519) {
return 4;
}
}
}
}
else if (refDCval >= 58) {
if (TSC < 9) {
if (posBalance < 58) {
if (negBalance < 57) {
if (MVDiff < 7) {
if (Rs < 1505) {
return 1;
}
else if (Rs >= 1505) {
return 4;
}
}
else if (MVDiff >= 7) {
if (Rs < 949) {
return 1;
}
else if (Rs >= 949) {
return 8;
}
}
}
else if (negBalance >= 57) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (posBalance >= 58) {
if (Rs < 1687) {
return 8;
}
else if (Rs >= 1687) {
if (Rs < 1696) {
return 2;
}
else if (Rs >= 1696) {
return 8;
}
}
}
}
else if (TSC >= 9) {
if (Rs < 1491) {
if (ssDCval < 87) {
if (tcor < 80) {
return 8;
}
else if (tcor >= 80) {
return 4;
}
}
else if (ssDCval >= 87) {
if (Rs < 1417) {
return 4;
}
else if (Rs >= 1417) {
if (Rs < 1476) {
return 8;
}
else if (Rs >= 1476) {
return 4;
}
}
}
}
else if (Rs >= 1491) {
return 8;
}
}
}
}
else if (Rs >= 1781) {
if (Cs < 724) {
return 2;
}
else if (Cs >= 724) {
if (MVDiff < 5) {
return 1;
}
else if (MVDiff >= 5) {
if (mu_mv_mag_sq < 25) {
return 4;
}
else if (mu_mv_mag_sq >= 25) {
return 8;
}
}
}
}
}
else if (ssDCval >= 90) {
if (AFD < 15) {
if (Cs < 735) {
return 1;
}
else if (Cs >= 735) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 1916) {
return 1;
}
else if (Rs >= 1916) {
return 8;
}
}
}
}
}
}
else if (AFD >= 15) {
if (negBalance < 55) {
if (SC < 2061) {
if (ssDCval < 103) {
return 8;
}
else if (ssDCval >= 103) {
if (Rs < 719) {
return 1;
}
else if (Rs >= 719) {
return 4;
}
}
}
else if (SC >= 2061) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 8;
}
}
}
else if (negBalance >= 55) {
if (SC < 2700) {
return 4;
}
else if (SC >= 2700) {
return 8;
}
}
}
}
}
else if (MVDiff >= 10) {
if (tcor < 52) {
if (TSC < 10) {
if (ssDCval < 43) {
if (Rs < 467) {
return 2;
}
else if (Rs >= 467) {
return 8;
}
}
else if (ssDCval >= 43) {
if (AFD < 24) {
if (Cs < 1063) {
if (posBalance < 58) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
if (RsDiff < 9) {
return 4;
}
else if (RsDiff >= 9) {
return 1;
}
}
}
else if (posBalance >= 58) {
if (negBalance < 60) {
return 4;
}
else if (negBalance >= 60) {
return 8;
}
}
}
else if (Cs >= 1063) {
if (posBalance < 65) {
if (Cs < 1085) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 1;
}
}
else if (Cs >= 1085) {
if (CsDiff < 6) {
if (Rs < 1874) {
return 1;
}
else if (Rs >= 1874) {
return 8;
}
}
else if (CsDiff >= 6) {
return 4;
}
}
}
else if (posBalance >= 65) {
return 1;
}
}
}
else if (AFD >= 24) {
if (ssDCval < 79) {
if (ssDCval < 59) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 4;
}
}
else if (ssDCval >= 59) {
return 2;
}
}
else if (ssDCval >= 79) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 1;
}
}
}
}
}
else if (TSC >= 10) {
if (mcTcor < 82) {
if (ssDCval < 101) {
if (mcTcor < 73) {
if (RsDiff < 10) {
if (mu_mv_mag_sq < 17) {
if (mu_mv_mag_sq < 2) {
if (Cs < 1236) {
return 8;
}
else if (Cs >= 1236) {
return 4;
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 1105) {
return 4;
}
else if (Rs >= 1105) {
if (Rs < 1185) {
return 8;
}
else if (Rs >= 1185) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (negBalance < 49) {
if (TSC < 12) {
return 1;
}
else if (TSC >= 12) {
return 8;
}
}
else if (negBalance >= 49) {
return 8;
}
}
}
else if (RsDiff >= 10) {
if (refDCval < 84) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 1;
}
}
else if (refDCval >= 84) {
if (RsCsDiff < 227) {
return 1;
}
else if (RsCsDiff >= 227) {
return 4;
}
}
}
}
else if (mcTcor >= 73) {
if (Cs < 804) {
if (tcor < 48) {
if (mcTcor < 78) {
return 8;
}
else if (mcTcor >= 78) {
return 2;
}
}
else if (tcor >= 48) {
return 4;
}
}
else if (Cs >= 804) {
return 8;
}
}
}
else if (ssDCval >= 101) {
if (TSC < 23) {
return 4;
}
else if (TSC >= 23) {
if (Rs < 1118) {
return 2;
}
else if (Rs >= 1118) {
return 8;
}
}
}
}
else if (mcTcor >= 82) {
if (diffTSC < 1) {
if (TSC < 11) {
if (Cs < 736) {
return 1;
}
else if (Cs >= 736) {
return 2;
}
}
else if (TSC >= 11) {
if (AFD < 34) {
return 4;
}
else if (AFD >= 34) {
return 8;
}
}
}
else if (diffTSC >= 1) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (diffMVdiffVal < 19) {
return 2;
}
else if (diffMVdiffVal >= 19) {
if (Rs < 378) {
return 8;
}
else if (Rs >= 378) {
if (Rs < 523) {
return 2;
}
else if (Rs >= 523) {
if (Rs < 593) {
return 4;
}
else if (Rs >= 593) {
return 1;
}
}
}
}
}
}
}
}
}
else if (tcor >= 52) {
if (negBalance < 45) {
if (posBalance < 53) {
if (diffAFD < 6) {
if (refDCval < 81) {
if (mu_mv_mag_sq < 32) {
if (MVDiff < 17) {
return 2;
}
else if (MVDiff >= 17) {
if (Rs < 1938) {
return 1;
}
else if (Rs >= 1938) {
if (Rs < 2276) {
return 8;
}
else if (Rs >= 2276) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (Rs < 863) {
return 1;
}
else if (Rs >= 863) {
return 2;
}
}
}
else if (refDCval >= 81) {
if (Cs < 892) {
if (Rs < 1445) {
return 2;
}
else if (Rs >= 1445) {
return 1;
}
}
else if (Cs >= 892) {
return 8;
}
}
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (posBalance >= 53) {
if (Cs < 817) {
if (RsCsDiff < 94) {
if (diffMVdiffVal < 7) {
if (refDCval < 71) {
if (diffAFD < 2) {
if (diffMVdiffVal < 0) {
if (Rs < 534) {
return 2;
}
else if (Rs >= 534) {
return 1;
}
}
else if (diffMVdiffVal >= 0) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (diffAFD >= 2) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (Rs < 419) {
return 4;
}
else if (Rs >= 419) {
return 2;
}
}
}
}
else if (refDCval >= 71) {
if (Rs < 1821) {
if (posBalance < 59) {
return 1;
}
else if (posBalance >= 59) {
return 2;
}
}
else if (Rs >= 1821) {
return 1;
}
}
}
else if (diffMVdiffVal >= 7) {
if (diffRsCsdiff < -40) {
return 2;
}
else if (diffRsCsdiff >= -40) {
if (SC < 1332) {
if (diffTSC < 1) {
if (Rs < 282) {
return 4;
}
else if (Rs >= 282) {
return 8;
}
}
else if (diffTSC >= 1) {
return 1;
}
}
else if (SC >= 1332) {
return 8;
}
}
}
}
else if (RsCsDiff >= 94) {
if (TSC < 11) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
return 1;
}
}
else if (TSC >= 11) {
if (Rs < 1485) {
return 4;
}
else if (Rs >= 1485) {
return 1;
}
}
}
}
else if (Cs >= 817) {
if (gchDC < 2) {
if (AFD < 16) {
if (diffAFD < -1) {
if (Rs < 1272) {
return 4;
}
else if (Rs >= 1272) {
return 2;
}
}
else if (diffAFD >= -1) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (AFD >= 16) {
if (Cs < 1272) {
if (Cs < 872) {
return 2;
}
else if (Cs >= 872) {
return 4;
}
}
else if (Cs >= 1272) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (TSC < 8) {
if (diffMVdiffVal < 13) {
if (diffAFD < 6) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (diffMVdiffVal >= 13) {
if (Rs < 699) {
return 4;
}
else if (Rs >= 699) {
return 8;
}
}
}
else if (TSC >= 8) {
if (SC < 2158) {
if (MVDiff < 32) {
if (Rs < 837) {
return 8;
}
else if (Rs >= 837) {
return 1;
}
}
else if (MVDiff >= 32) {
return 8;
}
}
else if (SC >= 2158) {
if (ssDCval < 108) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
if (Rs < 1678) {
return 8;
}
else if (Rs >= 1678) {
if (Cs < 1290) {
return 2;
}
else if (Cs >= 1290) {
return 8;
}
}
}
}
else if (ssDCval >= 108) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 45) {
if (AFD < 23) {
if (SC < 2442) {
if (mu_mv_mag_sq < 10) {
if (refDCval < 59) {
if (negBalance < 57) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
if (mcTcor < 89) {
if (Cs < 555) {
return 2;
}
else if (Cs >= 555) {
if (MVDiff < 47) {
return 4;
}
else if (MVDiff >= 47) {
if (MVDiff < 52) {
return 8;
}
else if (MVDiff >= 52) {
return 4;
}
}
}
}
else if (mcTcor >= 89) {
if (gchDC < 2) {
if (Cs < 835) {
return 8;
}
else if (Cs >= 835) {
return 4;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
else if (negBalance >= 57) {
if (diffRsCsdiff < -16) {
if (Rs < 1295) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
return 2;
}
}
else if (Rs >= 1295) {
return 1;
}
}
else if (diffRsCsdiff >= -16) {
if (diffAFD < 2) {
if (Cs < 948) {
return 8;
}
else if (Cs >= 948) {
if (Rs < 595) {
return 2;
}
else if (Rs >= 595) {
if (Rs < 1060) {
return 4;
}
else if (Rs >= 1060) {
return 8;
}
}
}
}
else if (diffAFD >= 2) {
return 2;
}
}
}
}
else if (refDCval >= 59) {
if (Cs < 822) {
if (tcor < 86) {
if (RsCsDiff < 77) {
if (mcTcor < 89) {
if (AFD < 15) {
if (CsDiff < 4) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
return 4;
}
}
else if (CsDiff >= 4) {
if (MVDiff < 25) {
if (Rs < 1127) {
return 8;
}
else if (Rs >= 1127) {
return 4;
}
}
else if (MVDiff >= 25) {
return 1;
}
}
}
else if (AFD >= 15) {
if (TSC < 9) {
if (Rs < 1254) {
return 4;
}
else if (Rs >= 1254) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (mcTcor >= 89) {
if (diffMVdiffVal < 3) {
if (MVDiff < 47) {
if (MVDiff < 12) {
if (Rs < 802) {
return 8;
}
else if (Rs >= 802) {
return 1;
}
}
else if (MVDiff >= 12) {
if (ssDCval < 91) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (ssDCval < 62) {
return 8;
}
else if (ssDCval >= 62) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (Rs < 1004) {
return 4;
}
else if (Rs >= 1004) {
return 8;
}
}
}
}
}
else if (ssDCval >= 91) {
if (Rs < 469) {
return 8;
}
else if (Rs >= 469) {
return 2;
}
}
}
}
else if (MVDiff >= 47) {
if (tcor < 74) {
if (Rs < 886) {
return 1;
}
else if (Rs >= 886) {
if (Rs < 1078) {
return 4;
}
else if (Rs >= 1078) {
return 2;
}
}
}
else if (tcor >= 74) {
return 8;
}
}
}
else if (diffMVdiffVal >= 3) {
if (ssDCval < 72) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 4;
}
}
else if (ssDCval >= 72) {
if (Rs < 1293) {
return 1;
}
else if (Rs >= 1293) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (mu_mv_mag_sq < 3) {
if (RsCsDiff < 107) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
if (Rs < 992) {
return 2;
}
else if (Rs >= 992) {
return 8;
}
}
}
else if (RsCsDiff >= 107) {
return 2;
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
if (AFD < 15) {
if (Cs < 718) {
return 4;
}
else if (Cs >= 718) {
return 1;
}
}
else if (AFD >= 15) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Rs < 889) {
return 1;
}
else if (Rs >= 889) {
return 4;
}
}
}
}
}
}
}
else if (tcor >= 86) {
if (CsDiff < 4) {
if (Cs < 643) {
return 1;
}
else if (Cs >= 643) {
return 4;
}
}
else if (CsDiff >= 4) {
return 1;
}
}
}
else if (Cs >= 822) {
if (refDCval < 98) {
if (mu_mv_mag_sq < 3) {
if (mcTcor < 85) {
if (Rs < 1068) {
if (Cs < 1115) {
if (Cs < 1000) {
return 8;
}
else if (Cs >= 1000) {
return 1;
}
}
else if (Cs >= 1115) {
return 2;
}
}
else if (Rs >= 1068) {
return 8;
}
}
else if (mcTcor >= 85) {
if (ssDCval < 68) {
if (gchDC < 2) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
return 1;
}
}
else if (gchDC >= 2) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 4;
}
}
}
else if (ssDCval >= 68) {
if (MVDiff < 16) {
if (Rs < 1011) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 8;
}
}
else if (Rs >= 1011) {
if (Rs < 1139) {
return 2;
}
else if (Rs >= 1139) {
if (Cs < 1121) {
return 2;
}
else if (Cs >= 1121) {
return 4;
}
}
}
}
else if (MVDiff >= 16) {
if (SC < 2268) {
if (diffAFD < -5) {
return 1;
}
else if (diffAFD >= -5) {
if (diffAFD < 3) {
return 2;
}
else if (diffAFD >= 3) {
if (Rs < 849) {
return 4;
}
else if (Rs >= 849) {
return 2;
}
}
}
}
else if (SC >= 2268) {
if (Rs < 1440) {
return 1;
}
else if (Rs >= 1440) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 10) {
if (refDCval < 72) {
if (Rs < 1056) {
if (Rs < 806) {
if (Cs < 999) {
return 4;
}
else if (Cs >= 999) {
return 8;
}
}
else if (Rs >= 806) {
if (tcor < 67) {
if (Rs < 958) {
return 4;
}
else if (Rs >= 958) {
return 8;
}
}
else if (tcor >= 67) {
return 1;
}
}
}
else if (Rs >= 1056) {
if (RsCsDiff < 73) {
return 8;
}
else if (RsCsDiff >= 73) {
return 1;
}
}
}
else if (refDCval >= 72) {
if (ssDCval < 93) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
if (ssDCval < 77) {
if (Rs < 886) {
return 2;
}
else if (Rs >= 886) {
return 4;
}
}
else if (ssDCval >= 77) {
if (mcTcor < 94) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
if (Rs < 1226) {
return 8;
}
else if (Rs >= 1226) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (tcor < 77) {
return 8;
}
else if (tcor >= 77) {
if (ssDCval < 82) {
return 8;
}
else if (ssDCval >= 82) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 93) {
return 8;
}
}
}
else if (TSC >= 10) {
if (AFD < 20) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 4;
}
}
else if (AFD >= 20) {
if (Rs < 854) {
return 2;
}
else if (Rs >= 854) {
return 1;
}
}
}
}
}
else if (refDCval >= 98) {
if (diffMVdiffVal < 3) {
return 4;
}
else if (diffMVdiffVal >= 3) {
if (AFD < 15) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
return 2;
}
}
else if (AFD >= 15) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
if (Rs < 883) {
return 4;
}
else if (Rs >= 883) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 10) {
if (ssDCval < 64) {
if (mu_mv_mag_sq < 44) {
if (tcor < 56) {
if (Cs < 824) {
if (Rs < 888) {
return 8;
}
else if (Rs >= 888) {
return 4;
}
}
else if (Cs >= 824) {
return 1;
}
}
else if (tcor >= 56) {
if (ssDCval < 47) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 8;
}
}
else if (ssDCval >= 47) {
if (Cs < 1241) {
if (diffAFD < -1) {
if (RsCsDiff < 49) {
return 8;
}
else if (RsCsDiff >= 49) {
if (SC < 1438) {
if (Rs < 293) {
return 4;
}
else if (Rs >= 293) {
return 8;
}
}
else if (SC >= 1438) {
return 4;
}
}
}
else if (diffAFD >= -1) {
if (SC < 1278) {
if (Rs < 462) {
return 8;
}
else if (Rs >= 462) {
return 4;
}
}
else if (SC >= 1278) {
return 8;
}
}
}
else if (Cs >= 1241) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (MVDiff < 32) {
return 8;
}
else if (MVDiff >= 32) {
return 4;
}
}
}
else if (ssDCval >= 64) {
if (TSC < 10) {
if (refDCval < 87) {
if (CsDiff < 4) {
if (tcor < 85) {
if (ssDCval < 67) {
return 2;
}
else if (ssDCval >= 67) {
return 4;
}
}
else if (tcor >= 85) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (TSC < 8) {
if (gchDC < 3) {
if (Cs < 902) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (CsDiff < 8) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
if (Rs < 1059) {
return 1;
}
else if (Rs >= 1059) {
return 2;
}
}
}
else if (CsDiff >= 8) {
return 4;
}
}
}
else if (Cs >= 902) {
if (Rs < 731) {
return 4;
}
else if (Rs >= 731) {
return 8;
}
}
}
else if (gchDC >= 3) {
if (Rs < 474) {
if (Rs < 457) {
return 4;
}
else if (Rs >= 457) {
return 8;
}
}
else if (Rs >= 474) {
return 1;
}
}
}
else if (TSC >= 8) {
if (Rs < 1170) {
if (Rs < 763) {
if (Cs < 864) {
return 4;
}
else if (Cs >= 864) {
return 2;
}
}
else if (Rs >= 763) {
if (Cs < 1286) {
return 1;
}
else if (Cs >= 1286) {
return 2;
}
}
}
else if (Rs >= 1170) {
return 4;
}
}
}
}
else if (refDCval >= 87) {
if (diffAFD < 1) {
if (ssDCval < 96) {
if (mu_mv_mag_sq < 30) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
if (Rs < 1318) {
return 8;
}
else if (Rs >= 1318) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 30) {
return 8;
}
}
else if (ssDCval >= 96) {
if (posBalance < 59) {
return 2;
}
else if (posBalance >= 59) {
if (Rs < 1379) {
return 8;
}
else if (Rs >= 1379) {
return 4;
}
}
}
}
else if (diffAFD >= 1) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (refDCval < 89) {
return 8;
}
else if (refDCval >= 89) {
return 2;
}
}
}
}
}
else if (TSC >= 10) {
if (RsDiff < 9) {
if (diffRsCsdiff < -17) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 8;
}
}
else if (diffRsCsdiff >= -17) {
if (Rs < 1468) {
return 8;
}
else if (Rs >= 1468) {
return 1;
}
}
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
}
else if (SC >= 2442) {
if (RsDiff < 9) {
if (AFD < 19) {
if (Rs < 1387) {
if (ssDCval < 74) {
if (MVDiff < 25) {
return 8;
}
else if (MVDiff >= 25) {
if (Cs < 1261) {
if (Rs < 1284) {
return 2;
}
else if (Rs >= 1284) {
return 4;
}
}
else if (Cs >= 1261) {
return 8;
}
}
}
else if (ssDCval >= 74) {
if (CsDiff < 7) {
if (Cs < 1133) {
return 2;
}
else if (Cs >= 1133) {
return 4;
}
}
else if (CsDiff >= 7) {
return 8;
}
}
}
else if (Rs >= 1387) {
if (posBalance < 58) {
if (AFD < 14) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
if (SC < 2523) {
return 2;
}
else if (SC >= 2523) {
if (Cs < 786) {
return 1;
}
else if (Cs >= 786) {
return 8;
}
}
}
}
else if (AFD >= 14) {
if (SC < 3332) {
if (tcor < 81) {
if (diffRsCsdiff < 5) {
if (diffMVdiffVal < -22) {
return 4;
}
else if (diffMVdiffVal >= -22) {
return 8;
}
}
else if (diffRsCsdiff >= 5) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 4;
}
}
}
else if (tcor >= 81) {
return 8;
}
}
else if (SC >= 3332) {
if (Rs < 2141) {
return 2;
}
else if (Rs >= 2141) {
return 4;
}
}
}
}
else if (posBalance >= 58) {
if (tcor < 85) {
if (diffAFD < -3) {
if (Rs < 1712) {
return 1;
}
else if (Rs >= 1712) {
if (Rs < 1979) {
return 8;
}
else if (Rs >= 1979) {
return 4;
}
}
}
else if (diffAFD >= -3) {
if (Cs < 1276) {
if (tcor < 83) {
return 8;
}
else if (tcor >= 83) {
if (Cs < 1043) {
return 4;
}
else if (Cs >= 1043) {
return 8;
}
}
}
else if (Cs >= 1276) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
}
}
}
else if (tcor >= 85) {
if (mcTcor < 93) {
if (Rs < 1718) {
return 8;
}
else if (Rs >= 1718) {
return 1;
}
}
else if (mcTcor >= 93) {
return 8;
}
}
}
}
}
else if (AFD >= 19) {
if (negBalance < 56) {
if (posBalance < 64) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (Rs < 1257) {
return 2;
}
else if (Rs >= 1257) {
return 4;
}
}
}
else if (posBalance >= 64) {
return 8;
}
}
else if (negBalance >= 56) {
if (Rs < 1748) {
return 1;
}
else if (Rs >= 1748) {
if (Rs < 1967) {
return 8;
}
else if (Rs >= 1967) {
return 2;
}
}
}
}
}
else if (RsDiff >= 9) {
if (mcTcor < 90) {
if (refDCval < 85) {
if (refDCval < 66) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
if (RsDiff < 12) {
if (Cs < 1240) {
if (SC < 2580) {
return 2;
}
else if (SC >= 2580) {
return 4;
}
}
else if (Cs >= 1240) {
return 8;
}
}
else if (RsDiff >= 12) {
return 2;
}
}
}
else if (refDCval >= 66) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
if (diffAFD < 5) {
if (Cs < 1139) {
return 4;
}
else if (Cs >= 1139) {
return 2;
}
}
else if (diffAFD >= 5) {
return 8;
}
}
}
}
else if (refDCval >= 85) {
if (gchDC < 3) {
if (ssDCval < 91) {
if (Rs < 1674) {
return 8;
}
else if (Rs >= 1674) {
return 4;
}
}
else if (ssDCval >= 91) {
if (Cs < 1031) {
return 1;
}
else if (Cs >= 1031) {
return 4;
}
}
}
else if (gchDC >= 3) {
if (AFD < 20) {
return 2;
}
else if (AFD >= 20) {
return 8;
}
}
}
}
else if (mcTcor >= 90) {
if (refDCval < 73) {
if (Rs < 1809) {
return 4;
}
else if (Rs >= 1809) {
return 2;
}
}
else if (refDCval >= 73) {
if (MVDiff < 46) {
if (CsDiff < 8) {
if (Rs < 2152) {
if (SC < 3024) {
if (Cs < 888) {
if (mcTcor < 96) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
return 8;
}
}
else if (mcTcor >= 96) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 2;
}
}
}
else if (Cs >= 888) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (RsCsDiff < 121) {
if (Rs < 1884) {
return 2;
}
else if (Rs >= 1884) {
return 1;
}
}
else if (RsCsDiff >= 121) {
return 8;
}
}
}
}
else if (SC >= 3024) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
return 4;
}
}
}
else if (Rs >= 2152) {
if (mcTcor < 93) {
return 4;
}
else if (mcTcor >= 93) {
if (Rs < 2213) {
return 4;
}
else if (Rs >= 2213) {
if (Rs < 2286) {
return 2;
}
else if (Rs >= 2286) {
return 8;
}
}
}
}
}
else if (CsDiff >= 8) {
if (posBalance < 66) {
if (Cs < 958) {
return 2;
}
else if (Cs >= 958) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (diffRsCsdiff < -37) {
return 2;
}
else if (diffRsCsdiff >= -37) {
return 8;
}
}
}
}
else if (posBalance >= 66) {
return 4;
}
}
}
else if (MVDiff >= 46) {
return 8;
}
}
}
}
}
}
else if (AFD >= 23) {
if (MVDiff < 47) {
if (Rs < 1733) {
if (RsCsDiff < 148) {
if (MVDiff < 36) {
if (RsDiff < 9) {
if (Cs < 776) {
if (Rs < 1004) {
return 8;
}
else if (Rs >= 1004) {
return 4;
}
}
else if (Cs >= 776) {
if (mu_mv_mag_sq < 5) {
if (RsCsDiff < 121) {
return 4;
}
else if (RsCsDiff >= 121) {
return 2;
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 1015) {
if (Cs < 800) {
return 2;
}
else if (Cs >= 800) {
return 1;
}
}
else if (Rs >= 1015) {
if (Rs < 1449) {
return 8;
}
else if (Rs >= 1449) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
if (Rs < 1613) {
return 4;
}
else if (Rs >= 1613) {
return 2;
}
}
}
else if (MVDiff >= 36) {
if (ssDCval < 84) {
if (Rs < 1230) {
return 4;
}
else if (Rs >= 1230) {
if (Rs < 1334) {
return 2;
}
else if (Rs >= 1334) {
return 1;
}
}
}
else if (ssDCval >= 84) {
return 1;
}
}
}
else if (RsCsDiff >= 148) {
if (AFD < 28) {
if (TSC < 9) {
if (Cs < 877) {
return 4;
}
else if (Cs >= 877) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (TSC >= 9) {
if (Rs < 1648) {
return 4;
}
else if (Rs >= 1648) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 8;
}
}
}
}
}
else if (AFD >= 28) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (Cs < 1028) {
return 2;
}
else if (Cs >= 1028) {
return 8;
}
}
}
}
}
else if (Rs >= 1733) {
if (negBalance < 52) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 2;
}
}
else if (negBalance >= 52) {
if (SC < 3304) {
if (AFD < 32) {
return 1;
}
else if (AFD >= 32) {
return 2;
}
}
else if (SC >= 3304) {
if (Rs < 2243) {
return 2;
}
else if (Rs >= 2243) {
return 4;
}
}
}
}
}
else if (MVDiff >= 47) {
if (Rs < 1523) {
if (refDCval < 74) {
if (tcor < 57) {
if (MVDiff < 62) {
return 4;
}
else if (MVDiff >= 62) {
return 1;
}
}
else if (tcor >= 57) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 8;
}
}
}
else if (refDCval >= 74) {
if (refDCval < 89) {
if (Rs < 1222) {
if (CsDiff < 7) {
if (mu_mv_mag_sq < 11) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 4;
}
}
else if (mu_mv_mag_sq >= 11) {
return 8;
}
}
else if (CsDiff >= 7) {
if (refDCval < 82) {
if (refDCval < 76) {
return 2;
}
else if (refDCval >= 76) {
return 4;
}
}
else if (refDCval >= 82) {
if (Rs < 635) {
return 4;
}
else if (Rs >= 635) {
return 2;
}
}
}
}
else if (Rs >= 1222) {
return 2;
}
}
else if (refDCval >= 89) {
if (Rs < 885) {
if (RsCsDiff < 87) {
if (RsDiff < 5) {
if (Rs < 430) {
return 1;
}
else if (Rs >= 430) {
return 4;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (RsCsDiff >= 87) {
if (Rs < 787) {
return 1;
}
else if (Rs >= 787) {
if (Cs < 693) {
return 1;
}
else if (Cs >= 693) {
return 2;
}
}
}
}
else if (Rs >= 885) {
if (RsCsDiff < 153) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 1;
}
}
else if (RsCsDiff >= 153) {
if (AFD < 31) {
return 4;
}
else if (AFD >= 31) {
return 2;
}
}
}
}
}
}
else if (Rs >= 1523) {
if (refDCval < 81) {
if (Rs < 1923) {
return 2;
}
else if (Rs >= 1923) {
return 1;
}
}
else if (refDCval >= 81) {
if (SC < 2838) {
return 8;
}
else if (SC >= 2838) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
if (ssDCval < 107) {
return 2;
}
else if (ssDCval >= 107) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 111) {
if (RsDiff < 9) {
if (Rs < 646) {
if (MVDiff < 126) {
if (diffRsCsdiff < 19) {
if (TSC < 8) {
if (Rs < 455) {
return 2;
}
else if (Rs >= 455) {
if (Rs < 546) {
return 4;
}
else if (Rs >= 546) {
if (Rs < 621) {
return 8;
}
else if (Rs >= 621) {
return 1;
}
}
}
}
else if (TSC >= 8) {
return 2;
}
}
else if (diffRsCsdiff >= 19) {
if (Rs < 268) {
return 1;
}
else if (Rs >= 268) {
return 8;
}
}
}
else if (MVDiff >= 126) {
if (Rs < 450) {
return 8;
}
else if (Rs >= 450) {
return 1;
}
}
}
else if (Rs >= 646) {
if (MVDiff < 12) {
if (tcor < 68) {
return 4;
}
else if (tcor >= 68) {
if (posBalance < 49) {
return 2;
}
else if (posBalance >= 49) {
return 8;
}
}
}
else if (MVDiff >= 12) {
if (refDCval < 152) {
if (gchDC < 6) {
if (ssDCval < 130) {
if (Rs < 1632) {
if (TSC < 8) {
if (posBalance < 59) {
if (SC < 1697) {
return 2;
}
else if (SC >= 1697) {
return 8;
}
}
else if (posBalance >= 59) {
return 2;
}
}
else if (TSC >= 8) {
if (ssDCval < 112) {
return 4;
}
else if (ssDCval >= 112) {
return 8;
}
}
}
else if (Rs >= 1632) {
if (negBalance < 58) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
else if (negBalance >= 58) {
return 8;
}
}
}
else if (ssDCval >= 130) {
if (RsCsDiff < 67) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (diffMVdiffVal < 3) {
return 1;
}
else if (diffMVdiffVal >= 3) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
return 1;
}
}
}
}
else if (RsCsDiff >= 67) {
if (AFD < 19) {
if (mcTcor < 80) {
return 1;
}
else if (mcTcor >= 80) {
return 8;
}
}
else if (AFD >= 19) {
if (Rs < 1203) {
return 1;
}
else if (Rs >= 1203) {
return 4;
}
}
}
}
}
else if (gchDC >= 6) {
if (mu_mv_mag_sq < 7) {
if (ssDCval < 147) {
return 2;
}
else if (ssDCval >= 147) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (refDCval < 123) {
if (Rs < 904) {
if (Rs < 744) {
return 8;
}
else if (Rs >= 744) {
return 4;
}
}
else if (Rs >= 904) {
return 2;
}
}
else if (refDCval >= 123) {
return 8;
}
}
}
}
else if (refDCval >= 152) {
if (mcTcor < 86) {
return 2;
}
else if (mcTcor >= 86) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 2;
}
}
}
}
}
}
else if (RsDiff >= 9) {
if (AFD < 32) {
if (tcor < 69) {
if (mu_mv_mag_sq < 179) {
if (diffRsCsdiff < 104) {
if (refDCval < 139) {
if (ssDCval < 131) {
if (ssDCval < 116) {
return 4;
}
else if (ssDCval >= 116) {
if (RsCsDiff < 142) {
return 4;
}
else if (RsCsDiff >= 142) {
return 8;
}
}
}
else if (ssDCval >= 131) {
if (Cs < 876) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 1;
}
}
else if (Cs >= 876) {
if (Cs < 908) {
return 8;
}
else if (Cs >= 908) {
return 4;
}
}
}
}
else if (refDCval >= 139) {
return 8;
}
}
else if (diffRsCsdiff >= 104) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 179) {
return 1;
}
}
else if (tcor >= 69) {
if (TSC < 9) {
if (CsDiff < 5) {
if (RsCsDiff < 94) {
return 8;
}
else if (RsCsDiff >= 94) {
return 1;
}
}
else if (CsDiff >= 5) {
if (RsDiff < 10) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
return 8;
}
}
else if (RsDiff >= 10) {
if (diffRsCsdiff < 63) {
return 2;
}
else if (diffRsCsdiff >= 63) {
if (Rs < 1862) {
return 2;
}
else if (Rs >= 1862) {
return 1;
}
}
}
}
}
else if (TSC >= 9) {
if (posBalance < 42) {
if (Rs < 2290) {
return 8;
}
else if (Rs >= 2290) {
return 2;
}
}
else if (posBalance >= 42) {
if (MVDiff < 22) {
if (gchDC < 3) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (gchDC >= 3) {
return 2;
}
}
else if (MVDiff >= 22) {
if (ssDCval < 132) {
if (SC < 2469) {
if (posBalance < 66) {
return 1;
}
else if (posBalance >= 66) {
if (Rs < 1594) {
return 2;
}
else if (Rs >= 1594) {
return 8;
}
}
}
else if (SC >= 2469) {
if (ssDCval < 111) {
if (Rs < 2048) {
return 8;
}
else if (Rs >= 2048) {
return 1;
}
}
else if (ssDCval >= 111) {
return 1;
}
}
}
else if (ssDCval >= 132) {
if (Rs < 2114) {
if (refDCval < 135) {
return 8;
}
else if (refDCval >= 135) {
if (ssDCval < 151) {
if (AFD < 25) {
return 1;
}
else if (AFD >= 25) {
if (Cs < 577) {
return 1;
}
else if (Cs >= 577) {
return 2;
}
}
}
else if (ssDCval >= 151) {
return 8;
}
}
}
else if (Rs >= 2114) {
return 1;
}
}
}
}
}
}
}
else if (AFD >= 32) {
if (diffAFD < 7) {
if (Cs < 542) {
return 1;
}
else if (Cs >= 542) {
return 2;
}
}
else if (diffAFD >= 7) {
if (Cs < 794) {
return 1;
}
else if (Cs >= 794) {
return 4;
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (Rs < 1787) {
if (Rs < 1098) {
if (refDCval < 64) {
if (Rs < 539) {
if (SC < 1090) {
if (MVDiff < 3) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
if (ssDCval < 62) {
if (Cs < 508) {
return 1;
}
else if (Cs >= 508) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 2;
}
}
}
else if (ssDCval >= 62) {
if (Rs < 311) {
return 4;
}
else if (Rs >= 311) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (mcTcor < 99) {
if (AFD < 7) {
if (SC < 884) {
return 4;
}
else if (SC >= 884) {
if (mu_mv_mag_sq < 57) {
if (Rs < 301) {
return 1;
}
else if (Rs >= 301) {
return 4;
}
}
else if (mu_mv_mag_sq >= 57) {
if (RsDiff < 2) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
return 1;
}
}
else if (RsDiff >= 2) {
return 2;
}
}
}
}
else if (AFD >= 7) {
return 1;
}
}
else if (mcTcor >= 99) {
if (Cs < 684) {
return 4;
}
else if (Cs >= 684) {
return 2;
}
}
}
}
else if (SC >= 1090) {
if (Rs < 388) {
if (ssDCval < 49) {
return 4;
}
else if (ssDCval >= 49) {
if (posBalance < 49) {
if (Rs < 363) {
return 2;
}
else if (Rs >= 363) {
if (Rs < 368) {
return 1;
}
else if (Rs >= 368) {
return 2;
}
}
}
else if (posBalance >= 49) {
return 1;
}
}
}
else if (Rs >= 388) {
if (diffRsCsdiff < -3) {
if (Cs < 691) {
return 2;
}
else if (Cs >= 691) {
return 4;
}
}
else if (diffRsCsdiff >= -3) {
if (negBalance < 36) {
if (negBalance < 20) {
if (Cs < 952) {
if (Cs < 651) {
return 2;
}
else if (Cs >= 651) {
return 1;
}
}
else if (Cs >= 952) {
return 4;
}
}
else if (negBalance >= 20) {
if (diffAFD < -1) {
if (Cs < 1011) {
return 1;
}
else if (Cs >= 1011) {
return 4;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (negBalance >= 36) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
}
else if (Rs >= 539) {
if (MVDiff < 1) {
if (SC < 1581) {
if (ssDCval < 54) {
if (mu_mv_mag_sq < 120) {
if (negBalance < 28) {
return 1;
}
else if (negBalance >= 28) {
return 4;
}
}
else if (mu_mv_mag_sq >= 120) {
return 8;
}
}
else if (ssDCval >= 54) {
return 8;
}
}
else if (SC >= 1581) {
if (ssDCval < 53) {
if (Rs < 905) {
if (Cs < 988) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (Rs < 734) {
if (Cs < 914) {
return 8;
}
else if (Cs >= 914) {
return 1;
}
}
else if (Rs >= 734) {
return 8;
}
}
}
else if (Cs >= 988) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (Rs >= 905) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
}
else if (ssDCval >= 53) {
if (Rs < 839) {
if (Rs < 738) {
if (ssDCval < 57) {
return 4;
}
else if (ssDCval >= 57) {
return 1;
}
}
else if (Rs >= 738) {
return 8;
}
}
else if (Rs >= 839) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
return 4;
}
}
}
}
}
else if (MVDiff >= 1) {
if (ssDCval < 50) {
if (ssDCval < 37) {
if (negBalance < 19) {
if (Rs < 825) {
return 2;
}
else if (Rs >= 825) {
return 8;
}
}
else if (negBalance >= 19) {
return 1;
}
}
else if (ssDCval >= 37) {
if (posBalance < 53) {
if (negBalance < 32) {
if (MVDiff < 7) {
if (Rs < 916) {
return 1;
}
else if (Rs >= 916) {
if (Cs < 583) {
return 1;
}
else if (Cs >= 583) {
return 8;
}
}
}
else if (MVDiff >= 7) {
if (Rs < 1029) {
return 4;
}
else if (Rs >= 1029) {
if (Rs < 1065) {
return 1;
}
else if (Rs >= 1065) {
return 8;
}
}
}
}
else if (negBalance >= 32) {
return 8;
}
}
else if (posBalance >= 53) {
if (Rs < 791) {
if (Cs < 803) {
return 4;
}
else if (Cs >= 803) {
return 2;
}
}
else if (Rs >= 791) {
return 8;
}
}
}
}
else if (ssDCval >= 50) {
if (negBalance < 22) {
if (posBalance < 36) {
return 1;
}
else if (posBalance >= 36) {
if (MVDiff < 7) {
if (ssDCval < 51) {
return 8;
}
else if (ssDCval >= 51) {
if (Rs < 615) {
return 4;
}
else if (Rs >= 615) {
if (diffRsCsdiff < -5) {
return 4;
}
else if (diffRsCsdiff >= -5) {
return 1;
}
}
}
}
else if (MVDiff >= 7) {
if (Cs < 1043) {
return 8;
}
else if (Cs >= 1043) {
return 4;
}
}
}
}
else if (negBalance >= 22) {
if (Rs < 779) {
if (negBalance < 30) {
if (ssDCval < 51) {
return 1;
}
else if (ssDCval >= 51) {
if (Cs < 512) {
return 1;
}
else if (Cs >= 512) {
if (RsDiff < 2) {
if (MVDiff < 2) {
if (Rs < 666) {
return 8;
}
else if (Rs >= 666) {
return 4;
}
}
else if (MVDiff >= 2) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
}
else if (negBalance >= 30) {
if (MVDiff < 11) {
if (mu_mv_mag_sq < 54) {
if (negBalance < 43) {
if (ssDCval < 62) {
return 4;
}
else if (ssDCval >= 62) {
return 2;
}
}
else if (negBalance >= 43) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (ssDCval < 56) {
if (Cs < 843) {
if (Cs < 567) {
return 4;
}
else if (Cs >= 567) {
return 8;
}
}
else if (Cs >= 843) {
if (Rs < 688) {
return 2;
}
else if (Rs >= 688) {
return 4;
}
}
}
else if (ssDCval >= 56) {
if (negBalance < 33) {
if (gchDC < 1) {
if (Rs < 658) {
return 1;
}
else if (Rs >= 658) {
return 2;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (negBalance >= 33) {
return 8;
}
}
}
}
else if (MVDiff >= 11) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (ssDCval < 59) {
if (SC < 1356) {
return 1;
}
else if (SC >= 1356) {
if (Rs < 641) {
return 2;
}
else if (Rs >= 641) {
return 8;
}
}
}
else if (ssDCval >= 59) {
if (Cs < 862) {
return 4;
}
else if (Cs >= 862) {
return 2;
}
}
}
}
}
}
else if (Rs >= 779) {
if (tcor < 99) {
if (SC < 2129) {
if (posBalance < 57) {
if (MVDiff < 9) {
if (TSC < 6) {
if (Rs < 928) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
if (tcor < 93) {
return 1;
}
else if (tcor >= 93) {
if (Rs < 921) {
return 4;
}
else if (Rs >= 921) {
return 2;
}
}
}
}
else if (Rs >= 928) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (MVDiff >= 9) {
if (diffMVdiffVal < 5) {
if (posBalance < 51) {
if (Rs < 1081) {
return 1;
}
else if (Rs >= 1081) {
return 8;
}
}
else if (posBalance >= 51) {
return 8;
}
}
else if (diffMVdiffVal >= 5) {
return 1;
}
}
}
else if (posBalance >= 57) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (SC >= 2129) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 4;
}
}
}
else if (tcor >= 99) {
if (MVDiff < 5) {
if (Rs < 883) {
return 4;
}
else if (Rs >= 883) {
return 1;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 64) {
if (SC < 2156) {
if (negBalance < 15) {
if (Rs < 916) {
if (Rs < 497) {
if (Rs < 241) {
return 1;
}
else if (Rs >= 241) {
return 8;
}
}
else if (Rs >= 497) {
return 2;
}
}
else if (Rs >= 916) {
return 8;
}
}
else if (negBalance >= 15) {
if (AFD < 3) {
if (Cs < 693) {
if (SC < 1213) {
if (Cs < 575) {
return 1;
}
else if (Cs >= 575) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 486) {
return 4;
}
else if (Rs >= 486) {
return 2;
}
}
}
}
else if (SC >= 1213) {
if (MVDiff < 7) {
if (ssDCval < 75) {
if (Rs < 1093) {
if (negBalance < 29) {
return 2;
}
else if (negBalance >= 29) {
return 1;
}
}
else if (Rs >= 1093) {
return 4;
}
}
else if (ssDCval >= 75) {
if (Rs < 770) {
return 2;
}
else if (Rs >= 770) {
return 8;
}
}
}
else if (MVDiff >= 7) {
if (SC < 1471) {
return 1;
}
else if (SC >= 1471) {
return 8;
}
}
}
}
else if (Cs >= 693) {
if (ssDCval < 84) {
if (Rs < 718) {
if (SC < 1177) {
return 4;
}
else if (SC >= 1177) {
if (mcTcor < 98) {
if (Rs < 712) {
return 4;
}
else if (Rs >= 712) {
return 8;
}
}
else if (mcTcor >= 98) {
if (diffRsCsdiff < -3) {
return 1;
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
}
}
else if (Rs >= 718) {
if (Cs < 752) {
if (Rs < 850) {
if (Cs < 730) {
if (Rs < 759) {
return 8;
}
else if (Rs >= 759) {
return 1;
}
}
else if (Cs >= 730) {
return 4;
}
}
else if (Rs >= 850) {
return 2;
}
}
else if (Cs >= 752) {
if (SC < 1870) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (Rs < 730) {
if (Rs < 724) {
return 1;
}
else if (Rs >= 724) {
return 8;
}
}
else if (Rs >= 730) {
if (diffMVdiffVal < -1) {
if (Rs < 737) {
return 4;
}
else if (Rs >= 737) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (negBalance < 20) {
return 1;
}
else if (negBalance >= 20) {
return 4;
}
}
}
}
}
else if (SC >= 1870) {
if (refDCval < 78) {
if (diffRsCsdiff < -2) {
return 8;
}
else if (diffRsCsdiff >= -2) {
if (negBalance < 21) {
if (Rs < 1001) {
return 8;
}
else if (Rs >= 1001) {
return 1;
}
}
else if (negBalance >= 21) {
return 1;
}
}
}
else if (refDCval >= 78) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (CsDiff < 1) {
if (Rs < 633) {
return 2;
}
else if (Rs >= 633) {
return 1;
}
}
else if (CsDiff >= 1) {
if (Rs < 696) {
return 2;
}
else if (Rs >= 696) {
return 4;
}
}
}
}
}
else if (AFD >= 3) {
if (MVDiff < 2) {
if (ssDCval < 81) {
if (refDCval < 75) {
if (negBalance < 40) {
if (Rs < 580) {
if (SC < 1163) {
if (Rs < 419) {
return 4;
}
else if (Rs >= 419) {
return 2;
}
}
else if (SC >= 1163) {
return 8;
}
}
else if (Rs >= 580) {
if (Rs < 732) {
if (Cs < 782) {
return 1;
}
else if (Cs >= 782) {
if (CsDiff < 2) {
return 8;
}
else if (CsDiff >= 2) {
return 1;
}
}
}
else if (Rs >= 732) {
if (tcor < 98) {
if (mu_mv_mag_sq < 91) {
if (Cs < 889) {
if (Rs < 837) {
return 4;
}
else if (Rs >= 837) {
return 1;
}
}
else if (Cs >= 889) {
if (diffRsCsdiff < -2) {
return 2;
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (mu_mv_mag_sq < 108) {
if (mu_mv_mag_sq < 95) {
return 4;
}
else if (mu_mv_mag_sq >= 95) {
if (Cs < 614) {
return 4;
}
else if (Cs >= 614) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (Cs < 808) {
if (Rs < 815) {
return 1;
}
else if (Rs >= 815) {
return 2;
}
}
else if (Cs >= 808) {
return 4;
}
}
}
}
else if (tcor >= 98) {
if (Cs < 675) {
if (Cs < 526) {
return 8;
}
else if (Cs >= 526) {
return 2;
}
}
else if (Cs >= 675) {
if (Cs < 1076) {
return 1;
}
else if (Cs >= 1076) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 40) {
if (gchDC < 2) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 564) {
return 8;
}
else if (Rs >= 564) {
return 4;
}
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (refDCval >= 75) {
if (mu_mv_mag_sq < 116) {
if (mu_mv_mag_sq < 75) {
if (Rs < 701) {
return 1;
}
else if (Rs >= 701) {
return 4;
}
}
else if (mu_mv_mag_sq >= 75) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
return 8;
}
}
}
}
}
else if (ssDCval >= 81) {
if (Rs < 617) {
return 2;
}
else if (Rs >= 617) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (mu_mv_mag_sq < 75) {
if (Cs < 803) {
if (ssDCval < 86) {
if (posBalance < 41) {
if (Cs < 661) {
return 1;
}
else if (Cs >= 661) {
return 2;
}
}
else if (posBalance >= 41) {
if (gchDC < 2) {
if (Cs < 775) {
if (posBalance < 55) {
if (mcTcor < 99) {
return 4;
}
else if (mcTcor >= 99) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 65) {
if (Rs < 387) {
return 2;
}
else if (Rs >= 387) {
return 1;
}
}
else if (ssDCval >= 65) {
if (mu_mv_mag_sq < 49) {
if (diffMVdiffVal < -6) {
if (Rs < 726) {
return 8;
}
else if (Rs >= 726) {
return 4;
}
}
else if (diffMVdiffVal >= -6) {
return 4;
}
}
else if (mu_mv_mag_sq >= 49) {
return 8;
}
}
}
}
else if (Cs >= 775) {
return 1;
}
}
else if (gchDC >= 2) {
if (diffMVdiffVal < 4) {
if (negBalance < 59) {
return 1;
}
else if (negBalance >= 59) {
return 8;
}
}
else if (diffMVdiffVal >= 4) {
return 8;
}
}
}
}
else if (ssDCval >= 86) {
if (mu_mv_mag_sq < 10) {
if (Cs < 585) {
return 1;
}
else if (Cs >= 585) {
return 4;
}
}
else if (mu_mv_mag_sq >= 10) {
return 2;
}
}
}
else if (Cs >= 803) {
if (RsDiff < 2) {
if (Cs < 835) {
return 2;
}
else if (Cs >= 835) {
if (posBalance < 30) {
return 2;
}
else if (posBalance >= 30) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (Rs < 750) {
if (Rs < 506) {
if (ssDCval < 66) {
return 2;
}
else if (ssDCval >= 66) {
if (AFD < 13) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
return 2;
}
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (Rs >= 506) {
return 2;
}
}
else if (Rs >= 750) {
if (SC < 1922) {
if (Rs < 884) {
if (Rs < 809) {
if (AFD < 9) {
if (Rs < 770) {
return 4;
}
else if (Rs >= 770) {
return 2;
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (Rs >= 809) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
if (Rs < 827) {
return 4;
}
else if (Rs >= 827) {
return 8;
}
}
}
}
else if (Rs >= 884) {
return 2;
}
}
else if (SC >= 1922) {
if (MVDiff < 3) {
if (Cs < 1133) {
return 8;
}
else if (Cs >= 1133) {
return 2;
}
}
else if (MVDiff >= 3) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (Cs < 921) {
if (MVDiff < 5) {
if (tcor < 98) {
if (Cs < 638) {
if (Rs < 597) {
return 1;
}
else if (Rs >= 597) {
return 2;
}
}
else if (Cs >= 638) {
if (AFD < 7) {
if (mu_mv_mag_sq < 108) {
return 8;
}
else if (mu_mv_mag_sq >= 108) {
if (Cs < 786) {
return 8;
}
else if (Cs >= 786) {
return 4;
}
}
}
else if (AFD >= 7) {
return 2;
}
}
}
else if (tcor >= 98) {
if (AFD < 5) {
if (Rs < 523) {
return 1;
}
else if (Rs >= 523) {
return 2;
}
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (mu_mv_mag_sq < 87) {
if (Cs < 684) {
if (Rs < 435) {
return 4;
}
else if (Rs >= 435) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
}
else if (Cs >= 684) {
if (Rs < 591) {
return 1;
}
else if (Rs >= 591) {
if (TSC < 6) {
if (refDCval < 66) {
return 1;
}
else if (refDCval >= 66) {
return 2;
}
}
else if (TSC >= 6) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 87) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
if (Rs < 762) {
return 1;
}
else if (Rs >= 762) {
return 8;
}
}
}
}
}
else if (Cs >= 921) {
if (TSC < 4) {
if (SC < 2025) {
return 1;
}
else if (SC >= 2025) {
return 2;
}
}
else if (TSC >= 4) {
if (Rs < 809) {
return 8;
}
else if (Rs >= 809) {
return 4;
}
}
}
}
}
}
}
}
else if (SC >= 2156) {
if (tcor < 98) {
if (RsCsDiff < 4) {
return 8;
}
else if (RsCsDiff >= 4) {
if (posBalance < 49) {
if (negBalance < 36) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 8;
}
}
else if (negBalance >= 36) {
return 2;
}
}
else if (posBalance >= 49) {
if (diffRsCsdiff < -7) {
return 2;
}
else if (diffRsCsdiff >= -7) {
return 4;
}
}
}
}
else if (tcor >= 98) {
return 4;
}
}
}
}
else if (Rs >= 1098) {
if (Cs < 1058) {
if (refDCval < 69) {
if (mu_mv_mag_sq < 128) {
if (SC < 2357) {
if (ssDCval < 61) {
if (TSC < 5) {
if (ssDCval < 47) {
if (Rs < 1326) {
return 8;
}
else if (Rs >= 1326) {
if (Rs < 1397) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (Rs < 1348) {
return 4;
}
else if (Rs >= 1348) {
return 4;
}
}
}
else if (Rs >= 1397) {
return 8;
}
}
}
else if (ssDCval >= 47) {
if (mu_mv_mag_sq < 125) {
if (Cs < 676) {
return 1;
}
else if (Cs >= 676) {
if (negBalance < 15) {
return 8;
}
else if (negBalance >= 15) {
if (AFD < 3) {
if (Cs < 767) {
return 4;
}
else if (Cs >= 767) {
return 1;
}
}
else if (AFD >= 3) {
if (mu_mv_mag_sq < 79) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 8;
}
}
else if (mu_mv_mag_sq >= 79) {
if (Rs < 1230) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
if (Rs < 1142) {
return 1;
}
else if (Rs >= 1142) {
return 8;
}
}
}
else if (Rs >= 1230) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 1;
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (SC < 1917) {
return 1;
}
else if (SC >= 1917) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
return 1;
}
}
}
}
}
else if (TSC >= 5) {
if (MVDiff < 8) {
if (Cs < 866) {
return 8;
}
else if (Cs >= 866) {
return 4;
}
}
else if (MVDiff >= 8) {
return 2;
}
}
}
else if (ssDCval >= 61) {
if (TSC < 3) {
if (Rs < 1137) {
if (posBalance < 47) {
if (Rs < 1132) {
return 1;
}
else if (Rs >= 1132) {
return 2;
}
}
else if (posBalance >= 47) {
if (Cs < 736) {
return 2;
}
else if (Cs >= 736) {
if (Rs < 1121) {
return 1;
}
else if (Rs >= 1121) {
return 8;
}
}
}
}
else if (Rs >= 1137) {
if (Cs < 507) {
return 2;
}
else if (Cs >= 507) {
return 1;
}
}
}
else if (TSC >= 3) {
if (Cs < 621) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (Cs >= 621) {
if (mu_mv_mag_sq < 56) {
if (Cs < 813) {
return 8;
}
else if (Cs >= 813) {
if (Rs < 1328) {
return 8;
}
else if (Rs >= 1328) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (mu_mv_mag_sq < 92) {
return 1;
}
else if (mu_mv_mag_sq >= 92) {
if (SC < 2198) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (SC >= 2198) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 2357) {
if (negBalance < 22) {
if (mcTcor < 99) {
if (Cs < 1007) {
if (SC < 2396) {
return 8;
}
else if (SC >= 2396) {
return 1;
}
}
else if (Cs >= 1007) {
return 8;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (negBalance >= 22) {
if (Cs < 829) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 4;
}
}
else if (Cs >= 829) {
if (MVDiff < 7) {
if (tcor < 101) {
return 8;
}
else if (tcor >= 101) {
return 8;
}
}
else if (MVDiff >= 7) {
if (Rs < 1612) {
if (Cs < 942) {
return 1;
}
else if (Cs >= 942) {
return 8;
}
}
else if (Rs >= 1612) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 67) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
if (Rs < 1143) {
return 8;
}
else if (Rs >= 1143) {
if (posBalance < 39) {
if (RsCsDiff < 2) {
if (ssDCval < 62) {
if (Rs < 1336) {
if (Cs < 920) {
return 8;
}
else if (Cs >= 920) {
return 1;
}
}
else if (Rs >= 1336) {
return 1;
}
}
else if (ssDCval >= 62) {
return 1;
}
}
else if (RsCsDiff >= 2) {
if (Rs < 1780) {
return 8;
}
else if (Rs >= 1780) {
return 1;
}
}
}
else if (posBalance >= 39) {
return 8;
}
}
}
}
else if (ssDCval >= 67) {
if (Cs < 692) {
return 8;
}
else if (Cs >= 692) {
return 4;
}
}
}
}
else if (refDCval >= 69) {
if (mu_mv_mag_sq < 56) {
if (MVDiff < 14) {
return 1;
}
else if (MVDiff >= 14) {
if (Cs < 789) {
if (Cs < 653) {
return 1;
}
else if (Cs >= 653) {
return 4;
}
}
else if (Cs >= 789) {
if (Rs < 1159) {
return 2;
}
else if (Rs >= 1159) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 56) {
if (refDCval < 73) {
if (posBalance < 56) {
if (posBalance < 41) {
if (Rs < 1554) {
return 8;
}
else if (Rs >= 1554) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
else if (posBalance >= 41) {
if (SC < 1890) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 1238) {
return 2;
}
else if (Rs >= 1238) {
return 1;
}
}
}
else if (SC >= 1890) {
if (Rs < 1768) {
return 8;
}
else if (Rs >= 1768) {
return 1;
}
}
}
}
else if (posBalance >= 56) {
return 2;
}
}
else if (refDCval >= 73) {
if (TSC < 3) {
if (ssDCval < 77) {
return 1;
}
else if (ssDCval >= 77) {
return 8;
}
}
else if (TSC >= 3) {
if (Cs < 827) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (Rs < 1473) {
return 8;
}
else if (Rs >= 1473) {
return 4;
}
}
}
else if (Cs >= 827) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 1058) {
if (RsCsDiff < 2) {
if (Rs < 1436) {
if (Rs < 1126) {
return 4;
}
else if (Rs >= 1126) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 8;
}
}
}
else if (Rs >= 1436) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
return 4;
}
}
}
}
else if (RsCsDiff >= 2) {
if (AFD < 4) {
if (Rs < 1401) {
if (mu_mv_mag_sq < 127) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 110) {
return 8;
}
else if (mu_mv_mag_sq >= 110) {
if (ssDCval < 70) {
if (ssDCval < 55) {
if (Cs < 1264) {
return 8;
}
else if (Cs >= 1264) {
return 4;
}
}
else if (ssDCval >= 55) {
if (mu_mv_mag_sq < 126) {
if (Rs < 1116) {
return 4;
}
else if (Rs >= 1116) {
return 1;
}
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 1301) {
return 4;
}
else if (Rs >= 1301) {
return 1;
}
}
}
}
else if (ssDCval >= 70) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
return 4;
}
}
}
}
else if (MVDiff >= 8) {
return 4;
}
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 1256) {
return 4;
}
else if (Cs >= 1256) {
return 1;
}
}
}
else if (Rs >= 1401) {
if (ssDCval < 49) {
if (Cs < 1154) {
if (Rs < 1646) {
return 4;
}
else if (Rs >= 1646) {
if (ssDCval < 48) {
if (SC < 2783) {
return 8;
}
else if (SC >= 2783) {
if (Rs < 1761) {
if (Cs < 1112) {
if (Rs < 1735) {
if (Rs < 1728) {
return 4;
}
else if (Rs >= 1728) {
return 8;
}
}
else if (Rs >= 1735) {
if (Rs < 1748) {
if (Rs < 1745) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (Rs >= 1745) {
return 4;
}
}
else if (Rs >= 1748) {
return 8;
}
}
}
else if (Cs >= 1112) {
if (negBalance < 21) {
if (Rs < 1684) {
if (Cs < 1114) {
return 4;
}
else if (Cs >= 1114) {
return 4;
}
}
else if (Rs >= 1684) {
if (Rs < 1720) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
if (Rs < 1713) {
if (Rs < 1695) {
if (Rs < 1689) {
return 8;
}
else if (Rs >= 1689) {
return 8;
}
}
else if (Rs >= 1695) {
return 4;
}
}
else if (Rs >= 1713) {
return 8;
}
}
}
else if (Rs >= 1720) {
if (Rs < 1733) {
return 4;
}
else if (Rs >= 1733) {
if (Cs < 1148) {
if (Rs < 1741) {
return 4;
}
else if (Rs >= 1741) {
if (Cs < 1134) {
return 4;
}
else if (Cs >= 1134) {
return 4;
}
}
}
else if (Cs >= 1148) {
return 4;
}
}
}
}
}
else if (negBalance >= 21) {
return 4;
}
}
}
else if (Rs >= 1761) {
if (SC < 2915) {
if (negBalance < 22) {
if (tcor < 99) {
if (Rs < 1780) {
if (SC < 2912) {
if (Rs < 1775) {
return 8;
}
else if (Rs >= 1775) {
return 4;
}
}
else if (SC >= 2912) {
return 8;
}
}
else if (Rs >= 1780) {
return 8;
}
}
else if (tcor >= 99) {
return 8;
}
}
else if (negBalance >= 22) {
return 4;
}
}
else if (SC >= 2915) {
if (Cs < 1140) {
return 4;
}
else if (Cs >= 1140) {
if (Rs < 1775) {
return 8;
}
else if (Rs >= 1775) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 48) {
return 4;
}
}
}
else if (Cs >= 1154) {
if (Rs < 1549) {
return 8;
}
else if (Rs >= 1549) {
if (Cs < 1176) {
if (Rs < 1719) {
return 4;
}
else if (Rs >= 1719) {
if (Cs < 1158) {
return 8;
}
else if (Cs >= 1158) {
if (Rs < 1746) {
return 4;
}
else if (Rs >= 1746) {
return 4;
}
}
}
}
else if (Cs >= 1176) {
return 4;
}
}
}
}
else if (ssDCval >= 49) {
if (Rs < 1481) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (Rs >= 1481) {
if (Cs < 1062) {
if (Rs < 1721) {
return 1;
}
else if (Rs >= 1721) {
return 8;
}
}
else if (Cs >= 1062) {
return 8;
}
}
}
}
}
else if (AFD >= 4) {
if (Rs < 1353) {
if (SC < 2471) {
if (ssDCval < 67) {
if (AFD < 6) {
if (mu_mv_mag_sq < 109) {
return 8;
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 1282) {
return 4;
}
else if (Rs >= 1282) {
return 8;
}
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (ssDCval >= 67) {
return 8;
}
}
else if (SC >= 2471) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1324) {
return 2;
}
else if (Rs >= 1324) {
return 8;
}
}
}
}
else if (Rs >= 1353) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (tcor < 96) {
return 2;
}
else if (tcor >= 96) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 1787) {
if (TSC < 4) {
if (refDCval < 82) {
if (ssDCval < 56) {
if (posBalance < 45) {
if (mu_mv_mag_sq < 109) {
return 4;
}
else if (mu_mv_mag_sq >= 109) {
if (Cs < 935) {
return 8;
}
else if (Cs >= 935) {
if (posBalance < 40) {
if (SC < 2924) {
if (Cs < 1060) {
return 4;
}
else if (Cs >= 1060) {
if (Cs < 1097) {
if (Rs < 1810) {
return 4;
}
else if (Rs >= 1810) {
return 4;
}
}
else if (Cs >= 1097) {
return 4;
}
}
}
else if (SC >= 2924) {
return 4;
}
}
else if (posBalance >= 40) {
if (Rs < 1822) {
return 4;
}
else if (Rs >= 1822) {
return 8;
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2242) {
return 1;
}
else if (Rs >= 2242) {
return 8;
}
}
}
else if (ssDCval >= 56) {
if (ssDCval < 71) {
if (Cs < 993) {
if (mu_mv_mag_sq < 31) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 31) {
if (SC < 2943) {
return 1;
}
else if (SC >= 2943) {
if (Rs < 2084) {
return 4;
}
else if (Rs >= 2084) {
return 1;
}
}
}
}
else if (Cs >= 993) {
if (Cs < 1090) {
if (posBalance < 36) {
if (Rs < 1870) {
if (Rs < 1865) {
return 4;
}
else if (Rs >= 1865) {
return 1;
}
}
else if (Rs >= 1870) {
return 4;
}
}
else if (posBalance >= 36) {
if (Rs < 1893) {
return 4;
}
else if (Rs >= 1893) {
if (negBalance < 18) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
else if (negBalance >= 18) {
return 1;
}
}
}
}
else if (Cs >= 1090) {
if (SC < 3058) {
if (refDCval < 61) {
return 1;
}
else if (refDCval >= 61) {
return 4;
}
}
else if (SC >= 3058) {
if (Cs < 1137) {
return 4;
}
else if (Cs >= 1137) {
if (Rs < 2287) {
if (Rs < 2050) {
return 2;
}
else if (Rs >= 2050) {
return 8;
}
}
else if (Rs >= 2287) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 71) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
if (SC < 2664) {
return 1;
}
else if (SC >= 2664) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 82) {
if (MVDiff < 1) {
if (Cs < 902) {
return 4;
}
else if (Cs >= 902) {
return 8;
}
}
else if (MVDiff >= 1) {
if (Rs < 1974) {
if (RsDiff < 3) {
return 4;
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (Rs >= 1974) {
if (Rs < 2291) {
return 8;
}
else if (Rs >= 2291) {
return 1;
}
}
}
}
}
else if (TSC >= 4) {
if (Cs < 1020) {
if (posBalance < 51) {
if (posBalance < 43) {
if (diffTSC < -1) {
if (ssDCval < 77) {
if (Cs < 585) {
return 1;
}
else if (Cs >= 585) {
return 2;
}
}
else if (ssDCval >= 77) {
if (Rs < 2163) {
return 8;
}
else if (Rs >= 2163) {
return 4;
}
}
}
else if (diffTSC >= -1) {
if (AFD < 7) {
if (RsDiff < 5) {
if (SC < 2758) {
if (Cs < 601) {
return 4;
}
else if (Cs >= 601) {
return 1;
}
}
else if (SC >= 2758) {
return 4;
}
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (AFD >= 7) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 2;
}
}
}
}
else if (posBalance >= 43) {
if (RsCsDiff < 79) {
if (mu_mv_mag_sq < 87) {
return 1;
}
else if (mu_mv_mag_sq >= 87) {
if (Rs < 2100) {
return 1;
}
else if (Rs >= 2100) {
return 2;
}
}
}
else if (RsCsDiff >= 79) {
if (SC < 2797) {
return 1;
}
else if (SC >= 2797) {
return 2;
}
}
}
}
else if (posBalance >= 51) {
if (ssDCval < 76) {
if (Rs < 1896) {
return 8;
}
else if (Rs >= 1896) {
return 2;
}
}
else if (ssDCval >= 76) {
if (Rs < 2252) {
if (Cs < 908) {
return 1;
}
else if (Cs >= 908) {
return 4;
}
}
else if (Rs >= 2252) {
return 2;
}
}
}
}
else if (Cs >= 1020) {
if (negBalance < 44) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
else if (negBalance >= 44) {
return 8;
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 101) {
if (SC < 1356) {
if (posBalance < 32) {
if (MVDiff < 9) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Rs < 653) {
return 8;
}
else if (Rs >= 653) {
return 1;
}
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (posBalance >= 32) {
if (RsCsDiff < 2) {
if (ssDCval < 123) {
if (negBalance < 15) {
return 1;
}
else if (negBalance >= 15) {
return 2;
}
}
else if (ssDCval >= 123) {
if (Rs < 326) {
return 2;
}
else if (Rs >= 326) {
if (Rs < 548) {
return 4;
}
else if (Rs >= 548) {
return 1;
}
}
}
}
else if (RsCsDiff >= 2) {
if (negBalance < 26) {
if (refDCval < 117) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 575) {
return 2;
}
else if (Rs >= 575) {
if (Rs < 628) {
return 1;
}
else if (Rs >= 628) {
return 8;
}
}
}
}
else if (refDCval >= 117) {
if (diffRsCsdiff < 3) {
return 4;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 593) {
return 1;
}
else if (Rs >= 593) {
return 8;
}
}
}
}
else if (negBalance >= 26) {
if (Cs < 748) {
if (posBalance < 57) {
if (mcTcor < 98) {
if (RsCsDiff < 7) {
if (ssDCval < 106) {
return 1;
}
else if (ssDCval >= 106) {
if (Cs < 560) {
return 8;
}
else if (Cs >= 560) {
if (Rs < 540) {
return 8;
}
else if (Rs >= 540) {
if (Cs < 670) {
return 2;
}
else if (Cs >= 670) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 7) {
if (ssDCval < 110) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
else if (ssDCval >= 110) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 69) {
if (Rs < 723) {
if (diffAFD < 4) {
if (ssDCval < 133) {
if (negBalance < 28) {
return 1;
}
else if (negBalance >= 28) {
if (diffAFD < -6) {
return 1;
}
else if (diffAFD >= -6) {
return 2;
}
}
}
else if (ssDCval >= 133) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 2;
}
}
}
else if (diffAFD >= 4) {
return 1;
}
}
else if (Rs >= 723) {
return 1;
}
}
else if (negBalance >= 69) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 2;
}
}
}
}
else if (posBalance >= 57) {
if (TSC < 7) {
if (mu_mv_mag_sq < 53) {
if (negBalance < 51) {
if (TSC < 4) {
if (Rs < 474) {
return 2;
}
else if (Rs >= 474) {
return 1;
}
}
else if (TSC >= 4) {
return 4;
}
}
else if (negBalance >= 51) {
if (Cs < 586) {
return 1;
}
else if (Cs >= 586) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 489) {
return 8;
}
else if (Rs >= 489) {
if (Cs < 736) {
return 2;
}
else if (Cs >= 736) {
return 8;
}
}
}
}
else if (TSC >= 7) {
return 8;
}
}
}
else if (Cs >= 748) {
if (Rs < 415) {
return 4;
}
else if (Rs >= 415) {
if (Rs < 493) {
if (ssDCval < 105) {
return 8;
}
else if (ssDCval >= 105) {
return 1;
}
}
else if (Rs >= 493) {
if (Rs < 582) {
return 2;
}
else if (Rs >= 582) {
return 4;
}
}
}
}
}
}
}
}
else if (SC >= 1356) {
if (RsCsDiff < 81) {
if (refDCval < 112) {
if (TSC < 5) {
if (SC < 2679) {
if (MVDiff < 1) {
if (Cs < 1110) {
return 4;
}
else if (Cs >= 1110) {
return 8;
}
}
else if (MVDiff >= 1) {
if (Cs < 1097) {
if (mu_mv_mag_sq < 97) {
if (Rs < 2002) {
if (SC < 1448) {
if (Rs < 761) {
if (Rs < 676) {
return 2;
}
else if (Rs >= 676) {
return 4;
}
}
else if (Rs >= 761) {
return 1;
}
}
else if (SC >= 1448) {
if (SC < 2371) {
if (diffMVdiffVal < 2) {
if (negBalance < 31) {
if (Rs < 1583) {
return 8;
}
else if (Rs >= 1583) {
return 1;
}
}
else if (negBalance >= 31) {
return 1;
}
}
else if (diffMVdiffVal >= 2) {
return 8;
}
}
else if (SC >= 2371) {
return 1;
}
}
}
else if (Rs >= 2002) {
return 8;
}
}
else if (mu_mv_mag_sq >= 97) {
return 8;
}
}
else if (Cs >= 1097) {
if (ssDCval < 106) {
if (Rs < 930) {
return 4;
}
else if (Rs >= 930) {
return 1;
}
}
else if (ssDCval >= 106) {
return 1;
}
}
}
}
else if (SC >= 2679) {
return 1;
}
}
else if (TSC >= 5) {
if (mu_mv_mag_sq < 7) {
if (MVDiff < 14) {
if (CsDiff < 4) {
if (TSC < 6) {
if (Rs < 811) {
return 4;
}
else if (Rs >= 811) {
return 1;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (CsDiff >= 4) {
if (Cs < 1095) {
if (diffRsCsdiff < -21) {
return 1;
}
else if (diffRsCsdiff >= -21) {
if (tcor < 97) {
return 4;
}
else if (tcor >= 97) {
return 8;
}
}
}
else if (Cs >= 1095) {
if (posBalance < 53) {
if (Cs < 1227) {
return 1;
}
else if (Cs >= 1227) {
return 8;
}
}
else if (posBalance >= 53) {
if (Rs < 1102) {
return 4;
}
else if (Rs >= 1102) {
if (Rs < 1641) {
return 2;
}
else if (Rs >= 1641) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 14) {
if (RsDiff < 7) {
if (Cs < 664) {
return 8;
}
else if (Cs >= 664) {
return 2;
}
}
else if (RsDiff >= 7) {
if (Cs < 954) {
return 1;
}
else if (Cs >= 954) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (MVDiff < 4) {
if (Rs < 935) {
if (Cs < 814) {
return 8;
}
else if (Cs >= 814) {
return 4;
}
}
else if (Rs >= 935) {
if (mu_mv_mag_sq < 88) {
return 1;
}
else if (mu_mv_mag_sq >= 88) {
return 8;
}
}
}
else if (MVDiff >= 4) {
if (tcor < 90) {
if (Cs < 1220) {
return 4;
}
else if (Cs >= 1220) {
return 1;
}
}
else if (tcor >= 90) {
if (mu_mv_mag_sq < 60) {
if (MVDiff < 7) {
if (Rs < 1780) {
return 1;
}
else if (Rs >= 1780) {
return 4;
}
}
else if (MVDiff >= 7) {
if (negBalance < 33) {
if (Rs < 1710) {
return 4;
}
else if (Rs >= 1710) {
return 1;
}
}
else if (negBalance >= 33) {
if (Rs < 2132) {
return 8;
}
else if (Rs >= 2132) {
if (Rs < 2200) {
return 4;
}
else if (Rs >= 2200) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 60) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (refDCval < 97) {
return 8;
}
else if (refDCval >= 97) {
return 2;
}
}
}
}
}
}
}
}
else if (refDCval >= 112) {
if (RsCsDiff < 15) {
if (negBalance < 26) {
return 4;
}
else if (negBalance >= 26) {
if (diffRsCsdiff < 10) {
if (Cs < 757) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (Cs < 754) {
return 8;
}
else if (Cs >= 754) {
return 1;
}
}
}
else if (Cs >= 757) {
if (AFD < 2) {
if (Cs < 990) {
return 8;
}
else if (Cs >= 990) {
return 4;
}
}
else if (AFD >= 2) {
if (ssDCval < 157) {
if (posBalance < 68) {
return 8;
}
else if (posBalance >= 68) {
return 4;
}
}
else if (ssDCval >= 157) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 10) {
return 2;
}
}
}
else if (RsCsDiff >= 15) {
if (refDCval < 163) {
if (MVDiff < 5) {
if (SC < 2544) {
if (mu_mv_mag_sq < 3) {
if (tcor < 92) {
return 4;
}
else if (tcor >= 92) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 751) {
return 1;
}
else if (Rs >= 751) {
return 8;
}
}
}
else if (SC >= 2544) {
if (Rs < 1997) {
return 2;
}
else if (Rs >= 1997) {
if (Rs < 2281) {
return 8;
}
else if (Rs >= 2281) {
return 1;
}
}
}
}
else if (MVDiff >= 5) {
if (diffAFD < -2) {
if (Cs < 778) {
if (Rs < 995) {
return 2;
}
else if (Rs >= 995) {
if (TSC < 6) {
if (ssDCval < 137) {
return 2;
}
else if (ssDCval >= 137) {
return 1;
}
}
else if (TSC >= 6) {
return 1;
}
}
}
else if (Cs >= 778) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 1;
}
}
}
else if (diffAFD >= -2) {
if (Rs < 2156) {
if (mu_mv_mag_sq < 3) {
if (MVDiff < 21) {
if (negBalance < 51) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 1;
}
}
else if (negBalance >= 51) {
if (tcor < 90) {
return 8;
}
else if (tcor >= 90) {
return 1;
}
}
}
else if (MVDiff >= 21) {
if (Rs < 1269) {
return 2;
}
else if (Rs >= 1269) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffAFD < -1) {
if (ssDCval < 137) {
if (Rs < 814) {
return 2;
}
else if (Rs >= 814) {
return 8;
}
}
else if (ssDCval >= 137) {
if (Cs < 1076) {
if (diffRsCsdiff < 5) {
return 1;
}
else if (diffRsCsdiff >= 5) {
return 8;
}
}
else if (Cs >= 1076) {
return 4;
}
}
}
else if (diffAFD >= -1) {
if (posBalance < 74) {
if (Cs < 592) {
if (Rs < 1081) {
return 8;
}
else if (Rs >= 1081) {
return 2;
}
}
else if (Cs >= 592) {
return 8;
}
}
else if (posBalance >= 74) {
if (SC < 1530) {
return 8;
}
else if (SC >= 1530) {
return 2;
}
}
}
}
}
else if (Rs >= 2156) {
if (Rs < 2228) {
return 2;
}
else if (Rs >= 2228) {
return 1;
}
}
}
}
}
else if (refDCval >= 163) {
return 2;
}
}
}
}
else if (RsCsDiff >= 81) {
if (Cs < 673) {
if (Cs < 578) {
return 1;
}
else if (Cs >= 578) {
if (tcor < 85) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (Cs < 623) {
return 2;
}
else if (Cs >= 623) {
return 4;
}
}
}
else if (tcor >= 85) {
if (Rs < 1936) {
if (Cs < 647) {
return 8;
}
else if (Cs >= 647) {
return 1;
}
}
else if (Rs >= 1936) {
return 1;
}
}
}
}
else if (Cs >= 673) {
if (Rs < 1947) {
if (ssDCval < 101) {
if (Rs < 1286) {
if (Cs < 887) {
if (Rs < 1080) {
if (Rs < 845) {
return 2;
}
else if (Rs >= 845) {
return 8;
}
}
else if (Rs >= 1080) {
return 4;
}
}
else if (Cs >= 887) {
return 1;
}
}
else if (Rs >= 1286) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (ssDCval >= 101) {
if (diffAFD < -1) {
if (Rs < 598) {
return 8;
}
else if (Rs >= 598) {
if (CsDiff < 6) {
return 1;
}
else if (CsDiff >= 6) {
return 2;
}
}
}
else if (diffAFD >= -1) {
if (Rs < 1559) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 764) {
return 8;
}
else if (Rs >= 764) {
return 2;
}
}
}
else if (Rs >= 1559) {
return 8;
}
}
}
}
else if (Rs >= 1947) {
if (mu_mv_mag_sq < 6) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 6) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 101) {
if (ssDCval < 133) {
if (Cs < 1221) {
if (Rs < 337) {
if (ssDCval < 123) {
return 4;
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 337) {
if (refDCval < 97) {
if (mu_mv_mag_sq < 126) {
if (Cs < 872) {
if (Rs < 1069) {
if (TSCindex < 3) {
return 8;
}
else if (TSCindex >= 3) {
if (Rs < 869) {
if (Rs < 479) {
return 8;
}
else if (Rs >= 479) {
return 1;
}
}
else if (Rs >= 869) {
return 8;
}
}
}
else if (Rs >= 1069) {
if (Rs < 1968) {
return 4;
}
else if (Rs >= 1968) {
if (Cs < 665) {
return 1;
}
else if (Cs >= 665) {
if (SC < 2959) {
if (Rs < 2163) {
return 8;
}
else if (Rs >= 2163) {
if (Rs < 2169) {
return 4;
}
else if (Rs >= 2169) {
return 8;
}
}
}
else if (SC >= 2959) {
return 4;
}
}
}
}
}
else if (Cs >= 872) {
if (posBalance < 55) {
if (negBalance < 36) {
if (mu_mv_mag_sq < 124) {
return 8;
}
else if (mu_mv_mag_sq >= 124) {
return 2;
}
}
else if (negBalance >= 36) {
if (Rs < 1037) {
return 1;
}
else if (Rs >= 1037) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (Rs < 1023) {
if (Rs < 600) {
return 8;
}
else if (Rs >= 600) {
return 4;
}
}
else if (Rs >= 1023) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 2161) {
if (posBalance < 23) {
return 4;
}
else if (posBalance >= 23) {
return 8;
}
}
else if (Rs >= 2161) {
if (Rs < 2253) {
return 4;
}
else if (Rs >= 2253) {
return 8;
}
}
}
}
else if (refDCval >= 97) {
if (mu_mv_mag_sq < 108) {
if (Rs < 1027) {
if (TSC < 4) {
if (ssDCval < 117) {
if (Rs < 400) {
return 4;
}
else if (Rs >= 400) {
if (Rs < 527) {
return 1;
}
else if (Rs >= 527) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (Cs < 1008) {
return 8;
}
else if (Cs >= 1008) {
return 1;
}
}
}
}
}
else if (ssDCval >= 117) {
if (Cs < 629) {
if (Rs < 475) {
return 8;
}
else if (Rs >= 475) {
return 1;
}
}
else if (Cs >= 629) {
return 4;
}
}
}
else if (TSC >= 4) {
if (SC < 1103) {
return 1;
}
else if (SC >= 1103) {
return 8;
}
}
}
else if (Rs >= 1027) {
if (ssDCval < 124) {
return 8;
}
else if (ssDCval >= 124) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 1676) {
if (Rs < 505) {
if (ssDCval < 102) {
if (Cs < 793) {
return 1;
}
else if (Cs >= 793) {
return 8;
}
}
else if (ssDCval >= 102) {
if (negBalance < 29) {
if (Rs < 427) {
return 4;
}
else if (Rs >= 427) {
return 1;
}
}
else if (negBalance >= 29) {
return 8;
}
}
}
else if (Rs >= 505) {
if (diffRsCsdiff < -19) {
if (Cs < 781) {
return 1;
}
else if (Cs >= 781) {
return 8;
}
}
else if (diffRsCsdiff >= -19) {
if (ssDCval < 97) {
if (Rs < 986) {
return 1;
}
else if (Rs >= 986) {
return 8;
}
}
else if (ssDCval >= 97) {
return 8;
}
}
}
}
else if (Rs >= 1676) {
if (Cs < 915) {
if (mu_mv_mag_sq < 113) {
return 8;
}
else if (mu_mv_mag_sq >= 113) {
return 1;
}
}
else if (Cs >= 915) {
if (ssDCval < 122) {
if (negBalance < 17) {
return 2;
}
else if (negBalance >= 17) {
return 8;
}
}
else if (ssDCval >= 122) {
return 2;
}
}
}
}
}
}
}
else if (Cs >= 1221) {
if (Rs < 634) {
return 1;
}
else if (Rs >= 634) {
if (ssDCval < 95) {
return 2;
}
else if (ssDCval >= 95) {
if (Cs < 1226) {
return 2;
}
else if (Cs >= 1226) {
return 8;
}
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 127) {
return 8;
}
else if (mu_mv_mag_sq >= 127) {
if (Rs < 1129) {
if (Rs < 972) {
return 8;
}
else if (Rs >= 972) {
return 2;
}
}
else if (Rs >= 1129) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2340) {
if (MVDiff < 17) {
if (refDCval < 109) {
if (posBalance < 48) {
if (RsCsDiff < 18) {
if (SC < 4980) {
if (SC < 3426) {
if (posBalance < 11) {
return 4;
}
else if (posBalance >= 11) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
if (diffRsCsdiff < -14) {
return 4;
}
else if (diffRsCsdiff >= -14) {
return 8;
}
}
}
}
}
else if (SC >= 3426) {
if (RsCsDiff < 5) {
if (negBalance < 22) {
if (SC < 3545) {
return 4;
}
else if (SC >= 3545) {
if (negBalance < 20) {
return 8;
}
else if (negBalance >= 20) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 2;
}
}
}
}
else if (negBalance >= 22) {
if (Cs < 1101) {
if (Rs < 2930) {
return 8;
}
else if (Rs >= 2930) {
return 2;
}
}
else if (Cs >= 1101) {
return 8;
}
}
}
else if (RsCsDiff >= 5) {
if (Cs < 1227) {
if (TSC < 5) {
if (diffRsCsdiff < -8) {
if (Rs < 2887) {
if (Rs < 2668) {
return 8;
}
else if (Rs >= 2668) {
return 2;
}
}
else if (Rs >= 2887) {
return 1;
}
}
else if (diffRsCsdiff >= -8) {
if (negBalance < 24) {
if (RsDiff < 3) {
if (diffRsCsdiff < 2) {
return 4;
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (negBalance >= 24) {
if (mu_mv_mag_sq < 124) {
if (RsCsDiff < 12) {
return 8;
}
else if (RsCsDiff >= 12) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
if (Rs < 2662) {
if (Rs < 2511) {
return 8;
}
else if (Rs >= 2511) {
return 4;
}
}
else if (Rs >= 2662) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 2;
}
}
}
}
else if (TSC >= 5) {
return 2;
}
}
else if (Cs >= 1227) {
if (ssDCval < 101) {
if (negBalance < 33) {
return 8;
}
else if (negBalance >= 33) {
if (Cs < 1253) {
if (Rs < 2877) {
return 4;
}
else if (Rs >= 2877) {
return 4;
}
}
else if (Cs >= 1253) {
return 1;
}
}
}
else if (ssDCval >= 101) {
if (Rs < 3007) {
return 8;
}
else if (Rs >= 3007) {
return 4;
}
}
}
}
}
}
else if (SC >= 4980) {
if (Cs < 1274) {
return 4;
}
else if (Cs >= 1274) {
if (Rs < 3900) {
return 4;
}
else if (Rs >= 3900) {
if (Rs < 4576) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (Rs >= 4576) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 18) {
if (ssDCval < 98) {
if (negBalance < 51) {
if (Rs < 3876) {
if (refDCval < 66) {
if (Cs < 1140) {
if (MVDiff < 3) {
if (ssDCval < 65) {
if (Rs < 2458) {
if (Rs < 2413) {
return 4;
}
else if (Rs >= 2413) {
return 8;
}
}
else if (Rs >= 2458) {
return 2;
}
}
else if (ssDCval >= 65) {
if (Rs < 2931) {
return 2;
}
else if (Rs >= 2931) {
return 4;
}
}
}
else if (MVDiff >= 3) {
if (AFD < 17) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (SC < 3535) {
if (MVDiff < 10) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (SC >= 3535) {
return 4;
}
}
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (Cs >= 1140) {
if (tcor < 89) {
if (mu_mv_mag_sq < 49) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
return 2;
}
}
else if (mu_mv_mag_sq >= 49) {
return 4;
}
}
else if (tcor >= 89) {
if (posBalance < 43) {
if (mu_mv_mag_sq < 115) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 115) {
return 8;
}
}
else if (posBalance >= 43) {
if (Rs < 2600) {
return 8;
}
else if (Rs >= 2600) {
if (RsDiff < 6) {
return 4;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 66) {
if (Rs < 3007) {
if (Cs < 508) {
if (Rs < 2576) {
if (Rs < 2355) {
return 2;
}
else if (Rs >= 2355) {
return 1;
}
}
else if (Rs >= 2576) {
return 8;
}
}
else if (Cs >= 508) {
if (mu_mv_mag_sq < 80) {
if (mcTcor < 97) {
if (RsCsDiff < 63) {
if (Rs < 2448) {
return 4;
}
else if (Rs >= 2448) {
if (Rs < 2601) {
return 1;
}
else if (Rs >= 2601) {
if (Cs < 1215) {
return 4;
}
else if (Cs >= 1215) {
return 1;
}
}
}
}
else if (RsCsDiff >= 63) {
if (Cs < 981) {
if (TSC < 7) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
return 1;
}
}
else if (TSC >= 7) {
if (ssDCval < 69) {
return 2;
}
else if (ssDCval >= 69) {
if (Rs < 2834) {
if (Rs < 2728) {
return 4;
}
else if (Rs >= 2728) {
return 2;
}
}
else if (Rs >= 2834) {
return 1;
}
}
}
}
else if (Cs >= 981) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
if (Rs < 2831) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
if (refDCval < 68) {
return 2;
}
else if (refDCval >= 68) {
return 4;
}
}
}
else if (Rs >= 2831) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (TSC < 6) {
if (ssDCval < 69) {
if (Rs < 2771) {
return 1;
}
else if (Rs >= 2771) {
return 4;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (TSC >= 6) {
if (Rs < 2740) {
return 4;
}
else if (Rs >= 2740) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (posBalance < 40) {
if (Rs < 2541) {
return 2;
}
else if (Rs >= 2541) {
return 1;
}
}
else if (posBalance >= 40) {
return 2;
}
}
}
}
else if (Rs >= 3007) {
if (refDCval < 77) {
if (negBalance < 45) {
if (diffAFD < -2) {
if (ssDCval < 74) {
if (Rs < 3082) {
return 2;
}
else if (Rs >= 3082) {
if (ssDCval < 68) {
return 2;
}
else if (ssDCval >= 68) {
return 1;
}
}
}
else if (ssDCval >= 74) {
if (Rs < 3571) {
return 4;
}
else if (Rs >= 3571) {
return 2;
}
}
}
else if (diffAFD >= -2) {
if (RsDiff < 12) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
if (diffMVdiffVal < -1) {
if (Rs < 3062) {
return 1;
}
else if (Rs >= 3062) {
if (SC < 4627) {
return 4;
}
else if (SC >= 4627) {
if (Rs < 3704) {
return 1;
}
else if (Rs >= 3704) {
return 4;
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (mu_mv_mag_sq < 105) {
return 4;
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 3188) {
return 2;
}
else if (Rs >= 3188) {
return 4;
}
}
}
}
}
else if (RsDiff >= 12) {
if (refDCval < 67) {
return 2;
}
else if (refDCval >= 67) {
if (Rs < 3433) {
return 4;
}
else if (Rs >= 3433) {
if (Cs < 1176) {
return 1;
}
else if (Cs >= 1176) {
return 4;
}
}
}
}
}
}
else if (negBalance >= 45) {
if (tcor < 89) {
if (Rs < 3307) {
return 4;
}
else if (Rs >= 3307) {
return 1;
}
}
else if (tcor >= 89) {
return 2;
}
}
}
else if (refDCval >= 77) {
if (tcor < 95) {
if (ssDCval < 86) {
if (TSC < 7) {
if (Cs < 1216) {
return 1;
}
else if (Cs >= 1216) {
return 4;
}
}
else if (TSC >= 7) {
return 2;
}
}
else if (ssDCval >= 86) {
if (Rs < 3847) {
if (Cs < 1077) {
if (AFD < 9) {
if (mu_mv_mag_sq < 52) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (Rs < 3480) {
return 4;
}
else if (Rs >= 3480) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 52) {
if (Rs < 3790) {
if (Rs < 3256) {
return 4;
}
else if (Rs >= 3256) {
if (Rs < 3630) {
return 4;
}
else if (Rs >= 3630) {
return 4;
}
}
}
else if (Rs >= 3790) {
return 4;
}
}
}
else if (AFD >= 9) {
return 4;
}
}
else if (Cs >= 1077) {
return 8;
}
}
else if (Rs >= 3847) {
return 2;
}
}
}
else if (tcor >= 95) {
if (negBalance < 31) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (negBalance >= 31) {
if (CsDiff < 5) {
return 2;
}
else if (CsDiff >= 5) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 3876) {
if (ssDCval < 90) {
if (mu_mv_mag_sq < 108) {
if (Cs < 1290) {
return 4;
}
else if (Cs >= 1290) {
if (Cs < 1296) {
return 1;
}
else if (Cs >= 1296) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 108) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 1;
}
}
}
else if (ssDCval >= 90) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 63) {
if (diffRsCsdiff < 12) {
if (negBalance < 29) {
return 4;
}
else if (negBalance >= 29) {
if (MVDiff < 5) {
if (Rs < 4175) {
return 4;
}
else if (Rs >= 4175) {
return 4;
}
}
else if (MVDiff >= 5) {
if (ssDCval < 92) {
if (Rs < 4223) {
return 4;
}
else if (Rs >= 4223) {
return 8;
}
}
else if (ssDCval >= 92) {
if (Rs < 4085) {
return 8;
}
else if (Rs >= 4085) {
return 8;
}
}
}
}
}
else if (diffRsCsdiff >= 12) {
return 8;
}
}
else if (mu_mv_mag_sq >= 63) {
return 4;
}
}
}
}
}
else if (negBalance >= 51) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
if (Rs < 3105) {
return 1;
}
else if (Rs >= 3105) {
return 2;
}
}
}
}
else if (ssDCval >= 98) {
if (negBalance < 47) {
if (Cs < 1019) {
if (RsDiff < 5) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (RsDiff >= 5) {
if (CsDiff < 2) {
return 2;
}
else if (CsDiff >= 2) {
if (negBalance < 42) {
if (RsDiff < 6) {
if (MVDiff < 7) {
return 8;
}
else if (MVDiff >= 7) {
if (Cs < 864) {
return 4;
}
else if (Cs >= 864) {
if (Rs < 3435) {
return 8;
}
else if (Rs >= 3435) {
return 4;
}
}
}
}
else if (RsDiff >= 6) {
return 8;
}
}
else if (negBalance >= 42) {
if (Rs < 2660) {
return 4;
}
else if (Rs >= 2660) {
return 2;
}
}
}
}
}
else if (Cs >= 1019) {
if (mu_mv_mag_sq < 100) {
if (Rs < 3775) {
return 8;
}
else if (Rs >= 3775) {
return 4;
}
}
else if (mu_mv_mag_sq >= 100) {
return 4;
}
}
}
else if (negBalance >= 47) {
return 2;
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Rs < 3809) {
if (negBalance < 21) {
if (SC < 3216) {
return 1;
}
else if (SC >= 3216) {
if (negBalance < 16) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (Rs < 2902) {
return 4;
}
else if (Rs >= 2902) {
return 4;
}
}
}
else if (negBalance >= 16) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
}
else if (negBalance >= 21) {
if (ssDCval < 74) {
if (Cs < 562) {
if (RsDiff < 12) {
return 4;
}
else if (RsDiff >= 12) {
if (Rs < 3015) {
return 2;
}
else if (Rs >= 3015) {
return 4;
}
}
}
else if (Cs >= 562) {
if (Cs < 1095) {
if (diffRsCsdiff < 33) {
if (MVDiff < 1) {
if (Cs < 1014) {
return 2;
}
else if (Cs >= 1014) {
return 4;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (diffRsCsdiff >= 33) {
return 4;
}
}
else if (Cs >= 1095) {
return 4;
}
}
}
else if (ssDCval >= 74) {
if (negBalance < 34) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (Rs < 3345) {
return 8;
}
else if (Rs >= 3345) {
return 4;
}
}
}
else if (negBalance >= 34) {
if (AFD < 19) {
if (Rs < 2377) {
return 2;
}
else if (Rs >= 2377) {
if (ssDCval < 85) {
return 1;
}
else if (ssDCval >= 85) {
return 2;
}
}
}
else if (AFD >= 19) {
return 2;
}
}
}
}
}
else if (Rs >= 3809) {
return 4;
}
}
else if (CsDiff >= 3) {
if (refDCval < 60) {
if (SC < 3481) {
return 2;
}
else if (SC >= 3481) {
if (Rs < 7231) {
return 4;
}
else if (Rs >= 7231) {
return 2;
}
}
}
else if (refDCval >= 60) {
if (negBalance < 39) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 77) {
if (refDCval < 62) {
return 2;
}
else if (refDCval >= 62) {
if (Cs < 1096) {
if (SC < 3254) {
return 1;
}
else if (SC >= 3254) {
if (tcor < 92) {
if (mcTcor < 89) {
return 1;
}
else if (mcTcor >= 89) {
return 4;
}
}
else if (tcor >= 92) {
if (negBalance < 38) {
if (refDCval < 73) {
return 4;
}
else if (refDCval >= 73) {
return 1;
}
}
else if (negBalance >= 38) {
return 2;
}
}
}
}
else if (Cs >= 1096) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (MVDiff < 7) {
if (Rs < 3936) {
return 4;
}
else if (Rs >= 3936) {
return 4;
}
}
else if (MVDiff >= 7) {
if (Rs < 2833) {
return 2;
}
else if (Rs >= 2833) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
return 2;
}
}
else if (posBalance >= 54) {
if (RsDiff < 5) {
if (ssDCval < 75) {
if (Rs < 2781) {
return 4;
}
else if (Rs >= 2781) {
return 4;
}
}
else if (ssDCval >= 75) {
if (Rs < 3624) {
return 1;
}
else if (Rs >= 3624) {
return 2;
}
}
}
else if (RsDiff >= 5) {
if (posBalance < 58) {
if (gchDC < 2) {
if (MVDiff < 1) {
if (Rs < 2764) {
return 2;
}
else if (Rs >= 2764) {
return 4;
}
}
else if (MVDiff >= 1) {
return 2;
}
}
else if (gchDC >= 2) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
return 4;
}
}
}
else if (posBalance >= 58) {
if (Rs < 4604) {
return 2;
}
else if (Rs >= 4604) {
return 4;
}
}
}
}
}
else if (negBalance >= 39) {
if (tcor < 77) {
if (Cs < 798) {
if (Rs < 3465) {
return 2;
}
else if (Rs >= 3465) {
return 4;
}
}
else if (Cs >= 798) {
if (SC < 4081) {
return 4;
}
else if (SC >= 4081) {
if (Rs < 3397) {
return 8;
}
else if (Rs >= 3397) {
return 1;
}
}
}
}
else if (tcor >= 77) {
if (mu_mv_mag_sq < 79) {
if (diffRsCsdiff < -196) {
if (RsDiff < 14) {
return 4;
}
else if (RsDiff >= 14) {
return 2;
}
}
else if (diffRsCsdiff >= -196) {
if (Rs < 2462) {
if (ssDCval < 82) {
return 2;
}
else if (ssDCval >= 82) {
if (diffRsCsdiff < 17) {
return 4;
}
else if (diffRsCsdiff >= 17) {
if (Rs < 2405) {
return 2;
}
else if (Rs >= 2405) {
return 1;
}
}
}
}
else if (Rs >= 2462) {
if (mcTcor < 99) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
if (Cs < 1184) {
return 2;
}
else if (Cs >= 1184) {
return 1;
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 79) {
if (refDCval < 65) {
if (Cs < 978) {
return 2;
}
else if (Cs >= 978) {
if (ssDCval < 65) {
if (TSC < 6) {
if (Cs < 1059) {
return 4;
}
else if (Cs >= 1059) {
return 2;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (ssDCval >= 65) {
return 2;
}
}
}
else if (refDCval >= 65) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsDiff < 11) {
if (AFD < 12) {
if (Cs < 998) {
if (negBalance < 20) {
if (Cs < 928) {
if (negBalance < 15) {
return 8;
}
else if (negBalance >= 15) {
if (diffRsCsdiff < 8) {
if (diffRsCsdiff < 3) {
return 4;
}
else if (diffRsCsdiff >= 3) {
if (Rs < 3109) {
return 4;
}
else if (Rs >= 3109) {
return 4;
}
}
}
else if (diffRsCsdiff >= 8) {
return 4;
}
}
}
else if (Cs >= 928) {
return 2;
}
}
else if (negBalance >= 20) {
if (posBalance < 53) {
return 8;
}
else if (posBalance >= 53) {
if (tcor < 92) {
if (Cs < 464) {
return 2;
}
else if (Cs >= 464) {
return 8;
}
}
else if (tcor >= 92) {
return 2;
}
}
}
}
else if (Cs >= 998) {
if (RsCsDiff < 94) {
if (mcTcor < 99) {
if (CsDiff < 5) {
if (refDCval < 108) {
return 8;
}
else if (refDCval >= 108) {
return 2;
}
}
else if (CsDiff >= 5) {
if (Rs < 3400) {
return 4;
}
else if (Rs >= 3400) {
return 8;
}
}
}
else if (mcTcor >= 99) {
if (Rs < 2945) {
return 8;
}
else if (Rs >= 2945) {
return 2;
}
}
}
else if (RsCsDiff >= 94) {
return 2;
}
}
}
else if (AFD >= 12) {
if (diffAFD < -3) {
if (Rs < 2620) {
return 8;
}
else if (Rs >= 2620) {
return 4;
}
}
else if (diffAFD >= -3) {
if (tcor < 71) {
return 4;
}
else if (tcor >= 71) {
return 8;
}
}
}
}
else if (RsDiff >= 11) {
if (negBalance < 49) {
if (ssDCval < 107) {
return 2;
}
else if (ssDCval >= 107) {
if (RsDiff < 14) {
if (Rs < 3628) {
return 4;
}
else if (Rs >= 3628) {
return 8;
}
}
else if (RsDiff >= 14) {
return 2;
}
}
}
else if (negBalance >= 49) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (Cs < 823) {
if (Rs < 2463) {
if (Rs < 2412) {
return 4;
}
else if (Rs >= 2412) {
return 1;
}
}
else if (Rs >= 2463) {
return 2;
}
}
else if (Cs >= 823) {
if (RsCsDiff < 215) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (negBalance < 57) {
return 4;
}
else if (negBalance >= 57) {
return 8;
}
}
}
else if (RsCsDiff >= 215) {
if (Rs < 3432) {
if (diffRsCsdiff < 5) {
if (mcTcor < 91) {
if (Rs < 3244) {
return 4;
}
else if (Rs >= 3244) {
return 1;
}
}
else if (mcTcor >= 91) {
return 2;
}
}
else if (diffRsCsdiff >= 5) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 4;
}
}
}
else if (Rs >= 3432) {
if (Rs < 3893) {
return 8;
}
else if (Rs >= 3893) {
if (Rs < 4156) {
return 4;
}
else if (Rs >= 4156) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (mcTcor < 98) {
if (RsDiff < 8) {
if (Rs < 3432) {
if (Rs < 2681) {
if (tcor < 92) {
return 8;
}
else if (tcor >= 92) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
if (diffMVdiffVal < 1) {
if (Cs < 1296) {
return 8;
}
else if (Cs >= 1296) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
if (mu_mv_mag_sq < 61) {
if (Rs < 2381) {
return 2;
}
else if (Rs >= 2381) {
return 8;
}
}
else if (mu_mv_mag_sq >= 61) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2681) {
if (CsDiff < 3) {
if (Cs < 1260) {
return 2;
}
else if (Cs >= 1260) {
return 8;
}
}
else if (CsDiff >= 3) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
}
else if (Rs >= 3432) {
if (tcor < 90) {
return 1;
}
else if (tcor >= 90) {
if (CsDiff < 4) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (diffMVdiffVal < 2) {
if (Rs < 4164) {
return 8;
}
else if (Rs >= 4164) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
return 2;
}
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (RsDiff >= 8) {
if (refDCval < 134) {
if (Rs < 3430) {
if (diffMVdiffVal < -3) {
if (refDCval < 113) {
return 2;
}
else if (refDCval >= 113) {
if (mu_mv_mag_sq < 34) {
if (Rs < 3176) {
return 1;
}
else if (Rs >= 3176) {
if (Rs < 3322) {
return 2;
}
else if (Rs >= 3322) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 34) {
return 2;
}
}
}
else if (diffMVdiffVal >= -3) {
if (Cs < 980) {
if (ssDCval < 130) {
if (Rs < 2436) {
return 1;
}
else if (Rs >= 2436) {
return 2;
}
}
else if (ssDCval >= 130) {
if (Cs < 834) {
return 1;
}
else if (Cs >= 834) {
return 2;
}
}
}
else if (Cs >= 980) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
if (ssDCval < 113) {
return 1;
}
else if (ssDCval >= 113) {
return 8;
}
}
}
}
}
else if (Rs >= 3430) {
if (TSC < 6) {
if (MVDiff < 9) {
if (Cs < 1221) {
return 2;
}
else if (Cs >= 1221) {
return 8;
}
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (TSC >= 6) {
if (Cs < 1227) {
return 2;
}
else if (Cs >= 1227) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 134) {
if (mcTcor < 94) {
if (RsDiff < 21) {
if (Rs < 3042) {
if (Rs < 2768) {
return 8;
}
else if (Rs >= 2768) {
return 1;
}
}
else if (Rs >= 3042) {
return 4;
}
}
else if (RsDiff >= 21) {
return 2;
}
}
else if (mcTcor >= 94) {
if (RsDiff < 16) {
if (CsDiff < 4) {
if (Rs < 5821) {
if (Rs < 2881) {
return 1;
}
else if (Rs >= 2881) {
if (Cs < 1004) {
return 2;
}
else if (Cs >= 1004) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 3864) {
return 1;
}
else if (Rs >= 3864) {
return 2;
}
}
}
}
}
else if (Rs >= 5821) {
return 1;
}
}
else if (CsDiff >= 4) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (mu_mv_mag_sq < 63) {
if (Cs < 1174) {
return 2;
}
else if (Cs >= 1174) {
return 1;
}
}
else if (mu_mv_mag_sq >= 63) {
return 1;
}
}
}
}
else if (RsDiff >= 16) {
if (ssDCval < 137) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
return 2;
}
}
}
else if (ssDCval >= 137) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (MVDiff < 7) {
if (Cs < 558) {
if (posBalance < 49) {
if (mu_mv_mag_sq < 37) {
return 8;
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 3189) {
return 1;
}
else if (Rs >= 3189) {
return 2;
}
}
}
else if (posBalance >= 49) {
return 2;
}
}
else if (Cs >= 558) {
if (refDCval < 142) {
return 2;
}
else if (refDCval >= 142) {
if (Rs < 3963) {
if (mu_mv_mag_sq < 97) {
if (diffMVdiffVal < -5) {
if (Rs < 3302) {
return 2;
}
else if (Rs >= 3302) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
return 2;
}
}
else if (mu_mv_mag_sq >= 97) {
if (Rs < 3661) {
return 2;
}
else if (Rs >= 3661) {
if (Cs < 1128) {
return 2;
}
else if (Cs >= 1128) {
if (posBalance < 36) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
else if (posBalance >= 36) {
return 1;
}
}
}
}
}
else if (Rs >= 3963) {
return 2;
}
}
}
}
else if (MVDiff >= 7) {
if (TSC < 5) {
if (mu_mv_mag_sq < 84) {
if (Cs < 353) {
return 1;
}
else if (Cs >= 353) {
if (Rs < 3981) {
if (refDCval < 110) {
return 1;
}
else if (refDCval >= 110) {
return 8;
}
}
else if (Rs >= 3981) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
}
else if (TSC >= 5) {
if (SC < 3312) {
return 1;
}
else if (SC >= 3312) {
if (diffTSC < -4) {
if (Rs < 5093) {
return 1;
}
else if (Rs >= 5093) {
return 2;
}
}
else if (diffTSC >= -4) {
if (refDCval < 113) {
if (Rs < 3151) {
return 4;
}
else if (Rs >= 3151) {
return 2;
}
}
else if (refDCval >= 113) {
return 2;
}
}
}
}
}
}
}
}
else if (MVDiff >= 17) {
if (mcTcor < 68) {
if (AFD < 21) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 1;
}
}
else if (AFD >= 21) {
if (gchDC < 7) {
if (diffMVdiffVal < -10) {
return 2;
}
else if (diffMVdiffVal >= -10) {
return 4;
}
}
else if (gchDC >= 7) {
return 8;
}
}
}
else if (mcTcor >= 68) {
if (RsCsDiff < 357) {
if (mcTcor < 92) {
if (refDCval < 133) {
if (MVDiff < 32) {
if (TSC < 12) {
if (AFD < 13) {
if (TSC < 8) {
if (posBalance < 43) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
return 8;
}
}
else if (posBalance >= 43) {
if (mu_mv_mag_sq < 25) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 2;
}
}
else if (mu_mv_mag_sq >= 25) {
return 4;
}
}
}
else if (TSC >= 8) {
if (MVDiff < 23) {
if (Rs < 2649) {
return 1;
}
else if (Rs >= 2649) {
return 2;
}
}
else if (MVDiff >= 23) {
return 1;
}
}
}
else if (AFD >= 13) {
if (ssDCval < 114) {
if (SC < 3819) {
if (diffTSC < 1) {
if (Rs < 2693) {
if (mcTcor < 89) {
if (tcor < 50) {
return 4;
}
else if (tcor >= 50) {
return 2;
}
}
else if (mcTcor >= 89) {
if (diffMVdiffVal < -1) {
if (Rs < 2405) {
return 8;
}
else if (Rs >= 2405) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
}
else if (Rs >= 2693) {
return 2;
}
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (SC >= 3819) {
if (ssDCval < 96) {
if (mu_mv_mag_sq < 42) {
if (gchDC < 2) {
if (AFD < 19) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (mcTcor < 91) {
if (Cs < 1246) {
return 2;
}
else if (Cs >= 1246) {
return 4;
}
}
else if (mcTcor >= 91) {
return 4;
}
}
}
else if (AFD >= 19) {
return 2;
}
}
else if (gchDC >= 2) {
return 1;
}
}
else if (mu_mv_mag_sq >= 42) {
if (Rs < 2710) {
return 8;
}
else if (Rs >= 2710) {
return 2;
}
}
}
else if (ssDCval >= 96) {
if (Rs < 3433) {
return 8;
}
else if (Rs >= 3433) {
return 2;
}
}
}
}
else if (ssDCval >= 114) {
if (AFD < 24) {
if (Cs < 632) {
return 2;
}
else if (Cs >= 632) {
return 1;
}
}
else if (AFD >= 24) {
return 8;
}
}
}
}
else if (TSC >= 12) {
if (mcTcor < 84) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (diffRsCsdiff < -17) {
if (AFD < 24) {
return 4;
}
else if (AFD >= 24) {
return 1;
}
}
else if (diffRsCsdiff >= -17) {
return 1;
}
}
}
else if (mcTcor >= 84) {
if (Rs < 4525) {
return 4;
}
else if (Rs >= 4525) {
return 2;
}
}
}
}
else if (MVDiff >= 32) {
if (tcor < 70) {
if (posBalance < 70) {
if (gchDC < 3) {
if (Cs < 1111) {
if (Cs < 705) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (Cs >= 705) {
if (mcTcor < 79) {
if (RsDiff < 15) {
return 2;
}
else if (RsDiff >= 15) {
return 4;
}
}
else if (mcTcor >= 79) {
if (Cs < 869) {
return 1;
}
else if (Cs >= 869) {
if (mu_mv_mag_sq < 4) {
return 8;
}
else if (mu_mv_mag_sq >= 4) {
if (Rs < 2951) {
if (Cs < 1094) {
return 2;
}
else if (Cs >= 1094) {
return 1;
}
}
else if (Rs >= 2951) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1111) {
if (diffRsCsdiff < -75) {
return 2;
}
else if (diffRsCsdiff >= -75) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 4;
}
}
}
}
else if (gchDC >= 3) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
if (Rs < 2424) {
if (Cs < 1146) {
return 2;
}
else if (Cs >= 1146) {
return 1;
}
}
else if (Rs >= 2424) {
return 1;
}
}
}
}
else if (posBalance >= 70) {
if (Cs < 335) {
return 2;
}
else if (Cs >= 335) {
if (TSC < 17) {
return 8;
}
else if (TSC >= 17) {
return 2;
}
}
}
}
else if (tcor >= 70) {
if (Cs < 452) {
if (AFD < 15) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 4;
}
}
else if (AFD >= 15) {
if (negBalance < 50) {
return 1;
}
else if (negBalance >= 50) {
if (RsDiff < 9) {
if (Rs < 2791) {
return 1;
}
else if (Rs >= 2791) {
return 4;
}
}
else if (RsDiff >= 9) {
return 8;
}
}
}
}
else if (Cs >= 452) {
if (diffMVdiffVal < 38) {
if (CsDiff < 5) {
if (MVDiff < 84) {
if (AFD < 13) {
if (Cs < 900) {
return 1;
}
else if (Cs >= 900) {
return 8;
}
}
else if (AFD >= 13) {
if (diffAFD < -9) {
return 1;
}
else if (diffAFD >= -9) {
if (diffMVdiffVal < -20) {
if (Rs < 2920) {
return 8;
}
else if (Rs >= 2920) {
return 4;
}
}
else if (diffMVdiffVal >= -20) {
return 8;
}
}
}
}
else if (MVDiff >= 84) {
if (Rs < 2494) {
return 2;
}
else if (Rs >= 2494) {
if (Rs < 3017) {
return 1;
}
else if (Rs >= 3017) {
return 8;
}
}
}
}
else if (CsDiff >= 5) {
if (mcTcor < 82) {
return 1;
}
else if (mcTcor >= 82) {
if (Rs < 2420) {
return 2;
}
else if (Rs >= 2420) {
if (refDCval < 71) {
return 2;
}
else if (refDCval >= 71) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (refDCval < 104) {
return 8;
}
else if (refDCval >= 104) {
return 2;
}
}
}
}
}
}
}
else if (diffMVdiffVal >= 38) {
if (Rs < 2672) {
return 1;
}
else if (Rs >= 2672) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 133) {
if (ssDCval < 147) {
if (tcor < 62) {
if (mu_mv_mag_sq < 17) {
return 1;
}
else if (mu_mv_mag_sq >= 17) {
return 8;
}
}
else if (tcor >= 62) {
if (AFD < 21) {
return 4;
}
else if (AFD >= 21) {
if (Rs < 3491) {
return 8;
}
else if (Rs >= 3491) {
return 1;
}
}
}
}
else if (ssDCval >= 147) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
if (AFD < 28) {
return 1;
}
else if (AFD >= 28) {
if (Cs < 789) {
return 8;
}
else if (Cs >= 789) {
return 2;
}
}
}
}
}
}
else if (mcTcor >= 92) {
if (refDCval < 95) {
if (TSC < 8) {
if (ssDCval < 80) {
if (Rs < 2905) {
if (posBalance < 54) {
if (mu_mv_mag_sq < 21) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (ssDCval < 68) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 4;
}
}
else if (ssDCval >= 68) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 21) {
if (ssDCval < 66) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
return 2;
}
}
else if (ssDCval >= 66) {
if (Rs < 2370) {
return 4;
}
else if (Rs >= 2370) {
return 1;
}
}
}
}
else if (posBalance >= 54) {
return 2;
}
}
else if (Rs >= 2905) {
if (tcor < 97) {
if (Cs < 849) {
return 4;
}
else if (Cs >= 849) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
return 4;
}
}
}
}
else if (tcor >= 97) {
if (Rs < 4200) {
return 8;
}
else if (Rs >= 4200) {
return 4;
}
}
}
}
else if (ssDCval >= 80) {
if (RsDiff < 12) {
if (Rs < 2740) {
return 1;
}
else if (Rs >= 2740) {
if (Rs < 3239) {
return 2;
}
else if (Rs >= 3239) {
return 8;
}
}
}
else if (RsDiff >= 12) {
if (diffRsCsdiff < 113) {
return 2;
}
else if (diffRsCsdiff >= 113) {
if (Rs < 2418) {
return 1;
}
else if (Rs >= 2418) {
return 2;
}
}
}
}
}
else if (TSC >= 8) {
if (Rs < 4183) {
if (MVDiff < 28) {
if (Cs < 1221) {
if (CsDiff < 4) {
if (Rs < 3582) {
if (ssDCval < 90) {
return 2;
}
else if (ssDCval >= 90) {
return 4;
}
}
else if (Rs >= 3582) {
return 4;
}
}
else if (CsDiff >= 4) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (Cs >= 1221) {
if (Rs < 3343) {
return 4;
}
else if (Rs >= 3343) {
return 2;
}
}
}
else if (MVDiff >= 28) {
if (refDCval < 91) {
if (ssDCval < 95) {
if (RsDiff < 13) {
if (TSC < 10) {
if (Rs < 2816) {
return 1;
}
else if (Rs >= 2816) {
return 2;
}
}
else if (TSC >= 10) {
return 2;
}
}
else if (RsDiff >= 13) {
if (Rs < 3606) {
return 2;
}
else if (Rs >= 3606) {
if (Rs < 3850) {
return 4;
}
else if (Rs >= 3850) {
return 2;
}
}
}
}
else if (ssDCval >= 95) {
if (Rs < 3636) {
return 8;
}
else if (Rs >= 3636) {
return 1;
}
}
}
else if (refDCval >= 91) {
if (Rs < 3376) {
return 8;
}
else if (Rs >= 3376) {
return 2;
}
}
}
}
else if (Rs >= 4183) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (refDCval >= 95) {
if (SC < 3437) {
if (CsDiff < 3) {
if (tcor < 83) {
if (MVDiff < 43) {
return 2;
}
else if (MVDiff >= 43) {
return 8;
}
}
else if (tcor >= 83) {
if (diffAFD < -3) {
if (diffAFD < -7) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (diffAFD >= -7) {
if (posBalance < 75) {
return 2;
}
else if (posBalance >= 75) {
return 1;
}
}
}
else if (diffAFD >= -3) {
if (Rs < 2632) {
if (gchDC < 3) {
if (ssDCval < 108) {
if (Rs < 2502) {
return 1;
}
else if (Rs >= 2502) {
return 8;
}
}
else if (ssDCval >= 108) {
if (Rs < 2622) {
return 8;
}
else if (Rs >= 2622) {
return 1;
}
}
}
else if (gchDC >= 3) {
return 1;
}
}
else if (Rs >= 2632) {
if (Rs < 2898) {
return 8;
}
else if (Rs >= 2898) {
return 4;
}
}
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 112) {
if (MVDiff < 26) {
return 2;
}
else if (MVDiff >= 26) {
if (CsDiff < 6) {
if (diffMVdiffVal < 32) {
return 1;
}
else if (diffMVdiffVal >= 32) {
return 2;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
else if (refDCval >= 112) {
if (ssDCval < 169) {
if (RsDiff < 12) {
if (Rs < 2616) {
if (mu_mv_mag_sq < 37) {
return 1;
}
else if (mu_mv_mag_sq >= 37) {
if (Rs < 2487) {
return 1;
}
else if (Rs >= 2487) {
return 2;
}
}
}
else if (Rs >= 2616) {
if (Rs < 2643) {
return 8;
}
else if (Rs >= 2643) {
return 1;
}
}
}
else if (RsDiff >= 12) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 1;
}
}
}
else if (ssDCval >= 169) {
if (Rs < 2422) {
return 1;
}
else if (Rs >= 2422) {
return 2;
}
}
}
}
}
else if (SC >= 3437) {
if (negBalance < 38) {
if (Rs < 3358) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (Cs < 1039) {
if (MVDiff < 41) {
return 2;
}
else if (MVDiff >= 41) {
return 1;
}
}
else if (Cs >= 1039) {
if (Cs < 1248) {
return 8;
}
else if (Cs >= 1248) {
return 1;
}
}
}
}
else if (Rs >= 3358) {
if (CsDiff < 5) {
if (Cs < 1057) {
return 2;
}
else if (Cs >= 1057) {
if (Cs < 1101) {
return 1;
}
else if (Cs >= 1101) {
return 2;
}
}
}
else if (CsDiff >= 5) {
if (Rs < 5387) {
return 1;
}
else if (Rs >= 5387) {
return 2;
}
}
}
}
else if (negBalance >= 38) {
if (ssDCval < 125) {
if (diffRsCsdiff < -215) {
if (ssDCval < 120) {
return 2;
}
else if (ssDCval >= 120) {
return 1;
}
}
else if (diffRsCsdiff >= -215) {
if (RsCsDiff < 233) {
if (diffAFD < -5) {
if (ssDCval < 116) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (ssDCval >= 116) {
if (Rs < 3145) {
return 1;
}
else if (Rs >= 3145) {
return 2;
}
}
}
else if (diffAFD >= -5) {
if (Cs < 349) {
return 4;
}
else if (Cs >= 349) {
if (negBalance < 39) {
return 1;
}
else if (negBalance >= 39) {
if (tcor < 73) {
return 1;
}
else if (tcor >= 73) {
if (refDCval < 104) {
if (mu_mv_mag_sq < 34) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Rs < 3310) {
return 1;
}
else if (Rs >= 3310) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 34) {
return 1;
}
}
else if (refDCval >= 104) {
if (MVDiff < 71) {
if (tcor < 76) {
if (Rs < 3144) {
return 2;
}
else if (Rs >= 3144) {
return 8;
}
}
else if (tcor >= 76) {
return 8;
}
}
else if (MVDiff >= 71) {
if (negBalance < 58) {
if (Cs < 855) {
return 4;
}
else if (Cs >= 855) {
return 2;
}
}
else if (negBalance >= 58) {
return 8;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 233) {
if (Rs < 3648) {
if (ssDCval < 104) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
return 2;
}
}
else if (ssDCval >= 104) {
if (posBalance < 51) {
if (Cs < 1111) {
return 2;
}
else if (Cs >= 1111) {
return 8;
}
}
else if (posBalance >= 51) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
if (RsDiff < 18) {
return 1;
}
else if (RsDiff >= 18) {
return 8;
}
}
}
}
}
else if (Rs >= 3648) {
if (CsDiff < 6) {
if (Rs < 5811) {
return 8;
}
else if (Rs >= 5811) {
return 2;
}
}
else if (CsDiff >= 6) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 125) {
if (RsCsDiff < 59) {
return 8;
}
else if (RsCsDiff >= 59) {
if (CsDiff < 4) {
if (mcTcor < 94) {
if (MVDiff < 24) {
return 4;
}
else if (MVDiff >= 24) {
if (Rs < 3158) {
return 8;
}
else if (Rs >= 3158) {
return 2;
}
}
}
else if (mcTcor >= 94) {
if (MVDiff < 39) {
if (Cs < 986) {
if (Rs < 2868) {
return 1;
}
else if (Rs >= 2868) {
if (MVDiff < 31) {
return 2;
}
else if (MVDiff >= 31) {
if (Rs < 3713) {
return 2;
}
else if (Rs >= 3713) {
return 1;
}
}
}
}
else if (Cs >= 986) {
return 1;
}
}
else if (MVDiff >= 39) {
if (Rs < 3455) {
return 1;
}
else if (Rs >= 3455) {
return 8;
}
}
}
}
else if (CsDiff >= 4) {
if (mu_mv_mag_sq < 26) {
if (tcor < 80) {
if (MVDiff < 53) {
return 1;
}
else if (MVDiff >= 53) {
return 8;
}
}
else if (tcor >= 80) {
if (Cs < 1230) {
if (RsCsDiff < 351) {
return 1;
}
else if (RsCsDiff >= 351) {
if (Rs < 5218) {
return 1;
}
else if (Rs >= 5218) {
return 8;
}
}
}
else if (Cs >= 1230) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (posBalance < 44) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 1;
}
}
else if (posBalance >= 44) {
if (Rs < 2671) {
return 2;
}
else if (Rs >= 2671) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 357) {
if (Cs < 797) {
if (ssDCval < 76) {
if (diffRsCsdiff < 170) {
if (negBalance < 36) {
if (Rs < 4589) {
return 4;
}
else if (Rs >= 4589) {
return 8;
}
}
else if (negBalance >= 36) {
if (Cs < 173) {
if (Rs < 3967) {
return 1;
}
else if (Rs >= 3967) {
return 2;
}
}
else if (Cs >= 173) {
return 2;
}
}
}
else if (diffRsCsdiff >= 170) {
if (AFD < 38) {
return 4;
}
else if (AFD >= 38) {
return 2;
}
}
}
else if (ssDCval >= 76) {
if (RsCsDiff < 462) {
if (refDCval < 125) {
if (ssDCval < 104) {
if (diffRsCsdiff < 254) {
return 2;
}
else if (diffRsCsdiff >= 254) {
return 4;
}
}
else if (ssDCval >= 104) {
if (gchDC < 6) {
if (ssDCval < 114) {
return 2;
}
else if (ssDCval >= 114) {
if (posBalance < 60) {
return 2;
}
else if (posBalance >= 60) {
if (posBalance < 65) {
return 8;
}
else if (posBalance >= 65) {
if (mcTcor < 91) {
return 2;
}
else if (mcTcor >= 91) {
return 1;
}
}
}
}
}
else if (gchDC >= 6) {
if (ssDCval < 114) {
return 1;
}
else if (ssDCval >= 114) {
return 8;
}
}
}
}
else if (refDCval >= 125) {
if (AFD < 19) {
if (Rs < 5416) {
return 4;
}
else if (Rs >= 5416) {
return 2;
}
}
else if (AFD >= 19) {
if (MVDiff < 43) {
if (Cs < 676) {
if (Cs < 309) {
return 2;
}
else if (Cs >= 309) {
return 1;
}
}
else if (Cs >= 676) {
return 2;
}
}
else if (MVDiff >= 43) {
return 1;
}
}
}
}
else if (RsCsDiff >= 462) {
if (negBalance < 66) {
return 2;
}
else if (negBalance >= 66) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
if (RsDiff < 25) {
return 1;
}
else if (RsDiff >= 25) {
return 2;
}
}
}
}
}
}
else if (Cs >= 797) {
if (refDCval < 119) {
if (Rs < 3471) {
if (CsDiff < 6) {
if (Rs < 2645) {
return 8;
}
else if (Rs >= 2645) {
if (Cs < 1235) {
return 4;
}
else if (Cs >= 1235) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (ssDCval < 110) {
if (AFD < 26) {
if (tcor < 70) {
if (Rs < 2620) {
return 2;
}
else if (Rs >= 2620) {
return 8;
}
}
else if (tcor >= 70) {
return 2;
}
}
else if (AFD >= 26) {
if (Cs < 1004) {
return 1;
}
else if (Cs >= 1004) {
return 2;
}
}
}
else if (ssDCval >= 110) {
if (mu_mv_mag_sq < 23) {
return 1;
}
else if (mu_mv_mag_sq >= 23) {
if (Rs < 3172) {
return 1;
}
else if (Rs >= 3172) {
return 2;
}
}
}
}
}
else if (Rs >= 3471) {
if (mcTcor < 94) {
if (CsDiff < 5) {
if (negBalance < 57) {
if (Rs < 4128) {
return 4;
}
else if (Rs >= 4128) {
return 8;
}
}
else if (negBalance >= 57) {
return 2;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
else if (mcTcor >= 94) {
if (gchDC < 3) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (CsDiff < 6) {
return 2;
}
else if (CsDiff >= 6) {
return 1;
}
}
}
else if (gchDC >= 3) {
if (Rs < 5976) {
return 1;
}
else if (Rs >= 5976) {
return 2;
}
}
}
}
}
else if (refDCval >= 119) {
if (mu_mv_mag_sq < 9) {
if (tcor < 82) {
if (CsDiff < 7) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (ssDCval < 136) {
if (diffMVdiffVal < -10) {
if (SC < 5990) {
return 2;
}
else if (SC >= 5990) {
return 8;
}
}
else if (diffMVdiffVal >= -10) {
if (negBalance < 58) {
if (RsDiff < 23) {
if (Rs < 5983) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 2;
}
}
else if (Rs >= 5983) {
return 2;
}
}
else if (RsDiff >= 23) {
return 2;
}
}
else if (negBalance >= 58) {
return 8;
}
}
}
else if (ssDCval >= 136) {
if (Rs < 3906) {
return 1;
}
else if (Rs >= 3906) {
return 8;
}
}
}
}
else if (CsDiff >= 7) {
if (Rs < 4801) {
return 1;
}
else if (Rs >= 4801) {
if (Cs < 1272) {
return 8;
}
else if (Cs >= 1272) {
return 1;
}
}
}
}
else if (tcor >= 82) {
if (Rs < 5894) {
if (diffMVdiffVal < 20) {
if (Cs < 1264) {
return 1;
}
else if (Cs >= 1264) {
return 2;
}
}
else if (diffMVdiffVal >= 20) {
if (Rs < 5655) {
return 2;
}
else if (Rs >= 5655) {
return 1;
}
}
}
else if (Rs >= 5894) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (tcor < 69) {
if (Rs < 3081) {
return 1;
}
else if (Rs >= 3081) {
if (diffAFD < 6) {
if (mcTcor < 96) {
return 2;
}
else if (mcTcor >= 96) {
if (Cs < 878) {
return 2;
}
else if (Cs >= 878) {
return 8;
}
}
}
else if (diffAFD >= 6) {
if (Cs < 1100) {
return 1;
}
else if (Cs >= 1100) {
return 2;
}
}
}
}
else if (tcor >= 69) {
if (CsDiff < 5) {
if (Rs < 3798) {
return 2;
}
else if (Rs >= 3798) {
return 4;
}
}
else if (CsDiff >= 5) {
if (diffMVdiffVal < -36) {
if (AFD < 30) {
return 2;
}
else if (AFD >= 30) {
return 1;
}
}
else if (diffMVdiffVal >= -36) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1313) {
if (mu_mv_mag_sq < 111) {
if (RsCsDiff < 101) {
if (refDCval < 90) {
if (negBalance < 39) {
if (Rs < 816) {
if (Rs < 309) {
return 4;
}
else if (Rs >= 309) {
if (ssDCval < 69) {
if (Rs < 741) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
else if (Rs >= 741) {
if (SC < 2311) {
return 8;
}
else if (SC >= 2311) {
return 1;
}
}
}
else if (ssDCval >= 69) {
if (Cs < 1698) {
return 8;
}
else if (Cs >= 1698) {
if (Rs < 636) {
return 1;
}
else if (Rs >= 636) {
return 2;
}
}
}
}
}
else if (Rs >= 816) {
if (SC < 10037) {
if (posBalance < 45) {
if (ssDCval < 64) {
if (Rs < 2997) {
if (mu_mv_mag_sq < 97) {
if (ssDCval < 50) {
if (AFD < 9) {
if (diffMVdiffVal < -3) {
if (Rs < 2184) {
return 2;
}
else if (Rs >= 2184) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (AFD >= 9) {
return 2;
}
}
else if (ssDCval >= 50) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
if (Rs < 2185) {
return 4;
}
else if (Rs >= 2185) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (SC < 2962) {
if (Rs < 952) {
return 8;
}
else if (Rs >= 952) {
return 4;
}
}
else if (SC >= 2962) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
}
else if (Rs >= 2997) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (Rs < 4181) {
return 8;
}
else if (Rs >= 4181) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
if (Rs < 4848) {
if (mu_mv_mag_sq < 108) {
return 4;
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 4764) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 4;
}
}
else if (Rs >= 4764) {
return 4;
}
}
}
else if (Rs >= 4848) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 64) {
if (refDCval < 68) {
if (MVDiff < 7) {
if (Cs < 1649) {
return 2;
}
else if (Cs >= 1649) {
if (Rs < 2688) {
return 8;
}
else if (Rs >= 2688) {
return 4;
}
}
}
else if (MVDiff >= 7) {
return 4;
}
}
else if (refDCval >= 68) {
if (Rs < 3519) {
if (MVDiff < 8) {
if (Rs < 1953) {
return 8;
}
else if (Rs >= 1953) {
if (Rs < 3192) {
if (negBalance < 38) {
return 1;
}
else if (negBalance >= 38) {
if (Rs < 2510) {
return 8;
}
else if (Rs >= 2510) {
return 4;
}
}
}
else if (Rs >= 3192) {
if (Cs < 1574) {
if (Cs < 1346) {
return 4;
}
else if (Cs >= 1346) {
return 2;
}
}
else if (Cs >= 1574) {
if (AFD < 7) {
if (ssDCval < 86) {
return 4;
}
else if (ssDCval >= 86) {
return 8;
}
}
else if (AFD >= 7) {
return 1;
}
}
}
}
}
else if (MVDiff >= 8) {
if (refDCval < 79) {
if (Rs < 2254) {
if (Rs < 1905) {
return 1;
}
else if (Rs >= 1905) {
return 8;
}
}
else if (Rs >= 2254) {
return 4;
}
}
else if (refDCval >= 79) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (Rs >= 3519) {
if (CsDiff < 5) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
if (RsCsDiff < 62) {
if (negBalance < 29) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
if (Rs < 4272) {
return 4;
}
else if (Rs >= 4272) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
}
}
else if (negBalance >= 29) {
if (SC < 5043) {
if (Rs < 3690) {
return 4;
}
else if (Rs >= 3690) {
return 2;
}
}
else if (SC >= 5043) {
return 4;
}
}
}
else if (RsCsDiff >= 62) {
return 2;
}
}
}
else if (CsDiff >= 5) {
if (ssDCval < 80) {
if (Rs < 4282) {
return 4;
}
else if (Rs >= 4282) {
return 8;
}
}
else if (ssDCval >= 80) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 4805) {
return 4;
}
else if (Rs >= 4805) {
return 1;
}
}
}
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2863) {
if (SC < 2583) {
if (tcor < 89) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (posBalance < 67) {
if (Rs < 919) {
return 2;
}
else if (Rs >= 919) {
return 1;
}
}
else if (posBalance >= 67) {
return 8;
}
}
}
else if (tcor >= 89) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Rs < 1155) {
return 2;
}
else if (Rs >= 1155) {
return 4;
}
}
}
}
else if (SC >= 2583) {
if (diffRsCsdiff < -8) {
if (SC < 2937) {
if (mu_mv_mag_sq < 81) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (Cs < 1461) {
return 4;
}
else if (Cs >= 1461) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (Rs < 1063) {
return 8;
}
else if (Rs >= 1063) {
return 4;
}
}
}
else if (SC >= 2937) {
if (Cs < 1755) {
if (mu_mv_mag_sq < 104) {
if (Cs < 1345) {
if (Rs < 2252) {
return 8;
}
else if (Rs >= 2252) {
return 2;
}
}
else if (Cs >= 1345) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
if (Rs < 2120) {
return 8;
}
else if (Rs >= 2120) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 104) {
if (Cs < 1633) {
return 2;
}
else if (Cs >= 1633) {
return 8;
}
}
}
else if (Cs >= 1755) {
if (posBalance < 47) {
if (Rs < 2681) {
return 8;
}
else if (Rs >= 2681) {
return 1;
}
}
else if (posBalance >= 47) {
if (SC < 5122) {
if (SC < 3728) {
if (AFD < 8) {
if (negBalance < 32) {
return 8;
}
else if (negBalance >= 32) {
return 1;
}
}
else if (AFD >= 8) {
return 4;
}
}
else if (SC >= 3728) {
return 8;
}
}
else if (SC >= 5122) {
if (Rs < 1800) {
return 1;
}
else if (Rs >= 1800) {
return 4;
}
}
}
}
}
}
else if (diffRsCsdiff >= -8) {
if (mu_mv_mag_sq < 91) {
if (Rs < 1489) {
if (posBalance < 66) {
if (TSC < 6) {
if (SC < 2863) {
if (MVDiff < 15) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
return 4;
}
}
else if (MVDiff >= 15) {
return 1;
}
}
else if (SC >= 2863) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (Rs < 923) {
return 8;
}
else if (Rs >= 923) {
return 1;
}
}
}
}
else if (TSC >= 6) {
if (SC < 2811) {
return 8;
}
else if (SC >= 2811) {
return 4;
}
}
}
else if (posBalance >= 66) {
return 8;
}
}
else if (Rs >= 1489) {
if (Cs < 1476) {
if (Rs < 2335) {
return 8;
}
else if (Rs >= 2335) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
else if (Cs >= 1476) {
if (Cs < 2547) {
if (posBalance < 48) {
if (Rs < 2512) {
return 8;
}
else if (Rs >= 2512) {
return 1;
}
}
else if (posBalance >= 48) {
if (Rs < 1987) {
if (SC < 3589) {
return 8;
}
else if (SC >= 3589) {
return 1;
}
}
else if (Rs >= 1987) {
return 8;
}
}
}
else if (Cs >= 2547) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 91) {
if (TSC < 8) {
if (SC < 4386) {
if (refDCval < 57) {
if (mu_mv_mag_sq < 101) {
return 8;
}
else if (mu_mv_mag_sq >= 101) {
if (Rs < 1031) {
return 8;
}
else if (Rs >= 1031) {
if (refDCval < 55) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (refDCval >= 55) {
return 4;
}
}
}
}
else if (refDCval >= 57) {
if (Cs < 1427) {
if (Cs < 1370) {
return 8;
}
else if (Cs >= 1370) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 4;
}
}
}
else if (Cs >= 1427) {
return 8;
}
}
}
else if (SC >= 4386) {
if (RsCsDiff < 22) {
return 8;
}
else if (RsCsDiff >= 22) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Rs < 2449) {
return 4;
}
else if (Rs >= 2449) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
return 4;
}
}
}
}
}
}
else if (TSC >= 8) {
return 4;
}
}
}
}
}
else if (Rs >= 2863) {
if (Cs < 2168) {
if (tcor < 91) {
if (posBalance < 54) {
if (Rs < 3171) {
return 1;
}
else if (Rs >= 3171) {
return 4;
}
}
else if (posBalance >= 54) {
if (Rs < 3872) {
return 8;
}
else if (Rs >= 3872) {
return 1;
}
}
}
else if (tcor >= 91) {
if (RsCsDiff < 55) {
if (Cs < 1594) {
if (RsCsDiff < 37) {
if (Rs < 3186) {
return 4;
}
else if (Rs >= 3186) {
if (Rs < 3404) {
return 4;
}
else if (Rs >= 3404) {
return 8;
}
}
}
else if (RsCsDiff >= 37) {
if (AFD < 9) {
return 2;
}
else if (AFD >= 9) {
return 1;
}
}
}
else if (Cs >= 1594) {
if (diffAFD < -2) {
if (Rs < 3625) {
return 8;
}
else if (Rs >= 3625) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Cs < 1999) {
return 4;
}
else if (Cs >= 1999) {
if (Rs < 3639) {
return 1;
}
else if (Rs >= 3639) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 55) {
if (mu_mv_mag_sq < 81) {
if (ssDCval < 84) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 2;
}
}
else if (ssDCval >= 84) {
if (Rs < 3295) {
return 4;
}
else if (Rs >= 3295) {
if (Rs < 3780) {
return 2;
}
else if (Rs >= 3780) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 81) {
if (Rs < 2936) {
return 1;
}
else if (Rs >= 2936) {
return 2;
}
}
}
}
}
else if (Cs >= 2168) {
if (CsDiff < 6) {
if (ssDCval < 55) {
if (Rs < 3334) {
return 8;
}
else if (Rs >= 3334) {
return 4;
}
}
else if (ssDCval >= 55) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (Cs < 2267) {
return 8;
}
else if (Cs >= 2267) {
return 1;
}
}
}
}
}
}
else if (SC >= 10037) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (SC < 7201) {
if (mcTcor < 94) {
if (Rs < 1438) {
if (RsCsDiff < 33) {
if (negBalance < 44) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
if (Rs < 1268) {
return 4;
}
else if (Rs >= 1268) {
return 8;
}
}
}
else if (negBalance >= 44) {
if (gchDC < 1) {
if (posBalance < 53) {
return 1;
}
else if (posBalance >= 53) {
return 8;
}
}
else if (gchDC >= 1) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (Rs < 714) {
return 2;
}
else if (Rs >= 714) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 33) {
if (refDCval < 54) {
if (MVDiff < 20) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (MVDiff >= 20) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 4;
}
}
}
else if (refDCval >= 54) {
if (RsCsDiff < 73) {
if (mcTcor < 85) {
if (Rs < 1430) {
return 8;
}
else if (Rs >= 1430) {
if (Cs < 1401) {
return 4;
}
else if (Cs >= 1401) {
return 8;
}
}
}
else if (mcTcor >= 85) {
if (ssDCval < 76) {
if (SC < 2640) {
if (SC < 2553) {
if (AFD < 15) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 2;
}
}
else if (AFD >= 15) {
if (Rs < 1121) {
return 2;
}
else if (Rs >= 1121) {
return 8;
}
}
}
else if (SC >= 2553) {
return 4;
}
}
else if (SC >= 2640) {
if (mcTcor < 87) {
if (Rs < 1195) {
return 2;
}
else if (Rs >= 1195) {
return 8;
}
}
else if (mcTcor >= 87) {
return 8;
}
}
}
else if (ssDCval >= 76) {
if (mu_mv_mag_sq < 3) {
return 4;
}
else if (mu_mv_mag_sq >= 3) {
if (Cs < 1671) {
return 8;
}
else if (Cs >= 1671) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 73) {
if (mu_mv_mag_sq < 5) {
if (MVDiff < 36) {
return 4;
}
else if (MVDiff >= 36) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (tcor < 77) {
if (mu_mv_mag_sq < 41) {
return 4;
}
else if (mu_mv_mag_sq >= 41) {
return 8;
}
}
else if (tcor >= 77) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1438) {
if (Cs < 2969) {
if (Rs < 3646) {
if (RsCsDiff < 87) {
if (AFD < 16) {
if (MVDiff < 17) {
if (mu_mv_mag_sq < 93) {
if (negBalance < 47) {
if (posBalance < 57) {
if (SC < 3413) {
if (SC < 3399) {
return 8;
}
else if (SC >= 3399) {
return 1;
}
}
else if (SC >= 3413) {
return 8;
}
}
else if (posBalance >= 57) {
if (RsDiff < 5) {
if (Rs < 1903) {
return 1;
}
else if (Rs >= 1903) {
return 4;
}
}
else if (RsDiff >= 5) {
if (ssDCval < 83) {
if (Rs < 1932) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
return 4;
}
}
else if (Rs >= 1932) {
return 8;
}
}
else if (ssDCval >= 83) {
if (Rs < 2271) {
return 2;
}
else if (Rs >= 2271) {
return 8;
}
}
}
}
}
else if (negBalance >= 47) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (tcor < 80) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (diffRsCsdiff < -36) {
return 4;
}
else if (diffRsCsdiff >= -36) {
return 8;
}
}
}
else if (tcor >= 80) {
if (posBalance < 66) {
if (diffMVdiffVal < -5) {
if (Rs < 1964) {
return 8;
}
else if (Rs >= 1964) {
return 4;
}
}
else if (diffMVdiffVal >= -5) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 2869) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (Rs < 2110) {
return 8;
}
else if (Rs >= 2110) {
return 2;
}
}
}
else if (Rs >= 2869) {
if (Rs < 2902) {
return 4;
}
else if (Rs >= 2902) {
return 8;
}
}
}
}
}
else if (posBalance >= 66) {
return 4;
}
}
}
}
else if (MVDiff >= 17) {
if (refDCval < 58) {
if (TSC < 8) {
if (Rs < 1693) {
return 1;
}
else if (Rs >= 1693) {
return 8;
}
}
else if (TSC >= 8) {
return 4;
}
}
else if (refDCval >= 58) {
if (Cs < 2105) {
if (posBalance < 56) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
if (Rs < 2030) {
return 1;
}
else if (Rs >= 2030) {
return 8;
}
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (Cs >= 2105) {
return 4;
}
}
}
}
else if (AFD >= 16) {
if (ssDCval < 78) {
return 8;
}
else if (ssDCval >= 78) {
if (negBalance < 57) {
if (ssDCval < 81) {
return 4;
}
else if (ssDCval >= 81) {
return 8;
}
}
else if (negBalance >= 57) {
return 4;
}
}
}
}
else if (RsCsDiff >= 87) {
if (TSC < 8) {
if (mcTcor < 91) {
if (tcor < 61) {
if (Rs < 1627) {
return 8;
}
else if (Rs >= 1627) {
return 1;
}
}
else if (tcor >= 61) {
return 8;
}
}
else if (mcTcor >= 91) {
if (MVDiff < 10) {
if (diffMVdiffVal < 1) {
if (Cs < 1437) {
return 4;
}
else if (Cs >= 1437) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 2;
}
}
else if (MVDiff >= 10) {
if (posBalance < 52) {
return 8;
}
else if (posBalance >= 52) {
return 4;
}
}
}
}
else if (TSC >= 8) {
if (mu_mv_mag_sq < 92) {
if (Rs < 1651) {
if (Cs < 1486) {
return 8;
}
else if (Cs >= 1486) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (MVDiff < 25) {
return 8;
}
else if (MVDiff >= 25) {
return 4;
}
}
}
}
else if (Rs >= 1651) {
if (TSC < 11) {
if (RsDiff < 7) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 1;
}
}
else if (RsDiff >= 7) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
if (tcor < 85) {
if (RsCsDiff < 99) {
if (Cs < 1533) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 8;
}
}
else if (Cs >= 1533) {
if (Rs < 1970) {
return 2;
}
else if (Rs >= 1970) {
return 4;
}
}
}
else if (RsCsDiff >= 99) {
if (Rs < 2910) {
return 8;
}
else if (Rs >= 2910) {
return 4;
}
}
}
else if (tcor >= 85) {
if (mcTcor < 90) {
return 4;
}
else if (mcTcor >= 90) {
if (Cs < 2911) {
return 8;
}
else if (Cs >= 2911) {
return 4;
}
}
}
}
}
}
else if (TSC >= 11) {
if (diffMVdiffVal < -6) {
if (Rs < 2063) {
return 1;
}
else if (Rs >= 2063) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 92) {
if (MVDiff < 5) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
else if (MVDiff >= 5) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
return 4;
}
}
}
}
}
}
else if (Rs >= 3646) {
return 4;
}
}
else if (Cs >= 2969) {
if (Rs < 3069) {
return 4;
}
else if (Rs >= 3069) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (RsDiff < 8) {
if (mu_mv_mag_sq < 85) {
if (MVDiff < 6) {
if (Cs < 2322) {
if (AFD < 10) {
if (Cs < 1658) {
if (RsDiff < 5) {
if (ssDCval < 59) {
return 4;
}
else if (ssDCval >= 59) {
if (diffRsCsdiff < -8) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 4;
}
}
else if (diffRsCsdiff >= -8) {
if (Cs < 1494) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
if (Rs < 787) {
return 1;
}
else if (Rs >= 787) {
if (posBalance < 55) {
if (Rs < 1350) {
return 8;
}
else if (Rs >= 1350) {
return 1;
}
}
else if (posBalance >= 55) {
return 8;
}
}
}
}
else if (Cs >= 1494) {
return 8;
}
}
}
}
else if (RsDiff >= 5) {
if (mcTcor < 97) {
if (ssDCval < 82) {
if (Cs < 1438) {
if (diffRsCsdiff < 5) {
return 8;
}
else if (diffRsCsdiff >= 5) {
return 1;
}
}
else if (Cs >= 1438) {
if (Rs < 1640) {
return 4;
}
else if (Rs >= 1640) {
return 8;
}
}
}
else if (ssDCval >= 82) {
if (ssDCval < 84) {
return 1;
}
else if (ssDCval >= 84) {
if (Rs < 2291) {
return 2;
}
else if (Rs >= 2291) {
if (RsCsDiff < 70) {
return 1;
}
else if (RsCsDiff >= 70) {
return 4;
}
}
}
}
}
else if (mcTcor >= 97) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 2;
}
}
}
}
else if (Cs >= 1658) {
if (MVDiff < 4) {
if (Rs < 3677) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (posBalance < 54) {
return 1;
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (Rs >= 3677) {
return 4;
}
}
else if (MVDiff >= 4) {
if (Cs < 1897) {
if (Rs < 737) {
return 2;
}
else if (Rs >= 737) {
return 8;
}
}
else if (Cs >= 1897) {
return 1;
}
}
}
}
else if (AFD >= 10) {
if (refDCval < 81) {
if (SC < 2545) {
if (Rs < 987) {
return 8;
}
else if (Rs >= 987) {
if (Rs < 1044) {
return 1;
}
else if (Rs >= 1044) {
return 2;
}
}
}
else if (SC >= 2545) {
if (diffRsCsdiff < -12) {
if (AFD < 11) {
if (Rs < 2820) {
return 8;
}
else if (Rs >= 2820) {
return 4;
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (diffRsCsdiff >= -12) {
return 4;
}
}
}
else if (refDCval >= 81) {
if (AFD < 11) {
if (diffMVdiffVal < -4) {
if (Rs < 3376) {
return 8;
}
else if (Rs >= 3376) {
return 4;
}
}
else if (diffMVdiffVal >= -4) {
if (Rs < 1660) {
return 8;
}
else if (Rs >= 1660) {
return 1;
}
}
}
else if (AFD >= 11) {
if (TSC < 8) {
if (Rs < 3465) {
return 1;
}
else if (Rs >= 3465) {
return 4;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
else if (Cs >= 2322) {
if (mu_mv_mag_sq < 82) {
return 1;
}
else if (mu_mv_mag_sq >= 82) {
if (Cs < 2806) {
if (Rs < 2467) {
return 1;
}
else if (Rs >= 2467) {
return 8;
}
}
else if (Cs >= 2806) {
return 4;
}
}
}
}
else if (MVDiff >= 6) {
if (Rs < 1533) {
if (posBalance < 60) {
if (TSC < 6) {
if (ssDCval < 57) {
if (Rs < 1075) {
return 2;
}
else if (Rs >= 1075) {
return 8;
}
}
else if (ssDCval >= 57) {
if (MVDiff < 22) {
if (posBalance < 37) {
return 2;
}
else if (posBalance >= 37) {
if (Cs < 2710) {
if (RsDiff < 4) {
return 1;
}
else if (RsDiff >= 4) {
if (Rs < 760) {
return 1;
}
else if (Rs >= 760) {
return 2;
}
}
}
else if (Cs >= 2710) {
return 8;
}
}
}
else if (MVDiff >= 22) {
return 8;
}
}
}
else if (TSC >= 6) {
if (mu_mv_mag_sq < 2) {
if (RsDiff < 4) {
if (Cs < 1693) {
return 2;
}
else if (Cs >= 1693) {
return 8;
}
}
else if (RsDiff >= 4) {
return 2;
}
}
else if (mu_mv_mag_sq >= 2) {
if (mu_mv_mag_sq < 30) {
return 8;
}
else if (mu_mv_mag_sq >= 30) {
if (Rs < 1069) {
return 2;
}
else if (Rs >= 1069) {
if (Rs < 1406) {
return 1;
}
else if (Rs >= 1406) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 60) {
if (tcor < 89) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
else if (tcor >= 89) {
return 8;
}
}
}
else if (Rs >= 1533) {
if (tcor < 90) {
if (refDCval < 80) {
if (ssDCval < 80) {
if (diffAFD < -2) {
if (Rs < 1978) {
if (Rs < 1767) {
return 8;
}
else if (Rs >= 1767) {
return 2;
}
}
else if (Rs >= 1978) {
return 4;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (ssDCval >= 80) {
return 4;
}
}
else if (refDCval >= 80) {
if (diffMVdiffVal < -4) {
if (Rs < 2534) {
return 8;
}
else if (Rs >= 2534) {
return 4;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (tcor >= 90) {
if (TSC < 8) {
if (negBalance < 48) {
if (diffRsCsdiff < -4) {
if (SC < 3594) {
return 8;
}
else if (SC >= 3594) {
if (Rs < 3341) {
return 1;
}
else if (Rs >= 3341) {
return 8;
}
}
}
else if (diffRsCsdiff >= -4) {
if (CsDiff < 5) {
if (Cs < 1354) {
return 1;
}
else if (Cs >= 1354) {
return 8;
}
}
else if (CsDiff >= 5) {
return 2;
}
}
}
else if (negBalance >= 48) {
return 8;
}
}
else if (TSC >= 8) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (Cs < 1661) {
if (Rs < 2758) {
if (posBalance < 55) {
if (negBalance < 44) {
if (Rs < 1501) {
if (Rs < 1010) {
return 8;
}
else if (Rs >= 1010) {
return 4;
}
}
else if (Rs >= 1501) {
if (AFD < 7) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
else if (AFD >= 7) {
if (SC < 3876) {
return 4;
}
else if (SC >= 3876) {
if (Rs < 2493) {
return 8;
}
else if (Rs >= 2493) {
return 2;
}
}
}
}
}
else if (negBalance >= 44) {
if (negBalance < 55) {
if (ssDCval < 74) {
return 8;
}
else if (ssDCval >= 74) {
if (ssDCval < 81) {
return 1;
}
else if (ssDCval >= 81) {
if (Rs < 1984) {
return 8;
}
else if (Rs >= 1984) {
return 1;
}
}
}
}
else if (negBalance >= 55) {
return 4;
}
}
}
else if (posBalance >= 55) {
if (ssDCval < 80) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
return 1;
}
}
}
else if (ssDCval >= 80) {
if (ssDCval < 83) {
return 1;
}
else if (ssDCval >= 83) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 2255) {
return 2;
}
else if (Rs >= 2255) {
return 1;
}
}
}
}
}
}
else if (Rs >= 2758) {
return 2;
}
}
else if (Cs >= 1661) {
if (SC < 5094) {
if (Rs < 1465) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (Rs >= 1465) {
if (posBalance < 56) {
if (tcor < 93) {
if (mu_mv_mag_sq < 94) {
return 8;
}
else if (mu_mv_mag_sq >= 94) {
return 4;
}
}
else if (tcor >= 93) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (ssDCval < 66) {
return 4;
}
else if (ssDCval >= 66) {
return 8;
}
}
}
}
}
else if (posBalance >= 56) {
if (diffAFD < -1) {
if (Cs < 1758) {
return 1;
}
else if (Cs >= 1758) {
if (mu_mv_mag_sq < 103) {
return 8;
}
else if (mu_mv_mag_sq >= 103) {
return 1;
}
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
else if (SC >= 5094) {
if (Rs < 2770) {
return 4;
}
else if (Rs >= 2770) {
if (ssDCval < 58) {
return 4;
}
else if (ssDCval >= 58) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (AFD < 10) {
if (Cs < 2382) {
return 2;
}
else if (Cs >= 2382) {
return 4;
}
}
else if (AFD >= 10) {
if (tcor < 87) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
else if (tcor >= 87) {
return 8;
}
}
}
}
}
else if (SC >= 7201) {
return 8;
}
}
}
else if (refDCval >= 90) {
if (Rs < 4357) {
if (ssDCval < 163) {
if (refDCval < 110) {
if (mcTcor < 96) {
if (SC < 3450) {
if (Rs < 1514) {
if (mu_mv_mag_sq < 27) {
if (Cs < 1991) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
if (mu_mv_mag_sq < 5) {
if (Cs < 1400) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
else if (Cs >= 1400) {
if (RsCsDiff < 63) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 4;
}
}
else if (RsCsDiff >= 63) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (Rs < 620) {
return 8;
}
else if (Rs >= 620) {
return 4;
}
}
}
}
else if (Cs >= 1991) {
return 2;
}
}
else if (mu_mv_mag_sq >= 27) {
if (Cs < 1758) {
if (Cs < 1337) {
if (mcTcor < 94) {
return 8;
}
else if (mcTcor >= 94) {
if (Rs < 1152) {
return 8;
}
else if (Rs >= 1152) {
return 4;
}
}
}
else if (Cs >= 1337) {
return 8;
}
}
else if (Cs >= 1758) {
if (mu_mv_mag_sq < 52) {
return 4;
}
else if (mu_mv_mag_sq >= 52) {
if (diffAFD < -1) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 8;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
}
}
else if (Rs >= 1514) {
if (Rs < 1858) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Cs < 1346) {
return 8;
}
else if (Cs >= 1346) {
return 2;
}
}
}
}
}
else if (Rs >= 1858) {
if (TSC < 8) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (Rs < 1884) {
return 2;
}
else if (Rs >= 1884) {
return 8;
}
}
}
else if (TSC >= 8) {
if (diffAFD < 2) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
else if (diffAFD >= 2) {
if (Cs < 1427) {
return 2;
}
else if (Cs >= 1427) {
return 8;
}
}
}
}
}
}
else if (SC >= 3450) {
if (TSC < 11) {
if (Cs < 2427) {
if (tcor < 89) {
if (Rs < 2227) {
if (negBalance < 48) {
if (mu_mv_mag_sq < 19) {
return 2;
}
else if (mu_mv_mag_sq >= 19) {
if (ssDCval < 93) {
return 2;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
else if (negBalance >= 48) {
return 8;
}
}
else if (Rs >= 2227) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (diffMVdiffVal < -14) {
if (Rs < 2624) {
return 8;
}
else if (Rs >= 2624) {
return 4;
}
}
else if (diffMVdiffVal >= -14) {
return 8;
}
}
}
}
else if (tcor >= 89) {
if (RsCsDiff < 99) {
if (diffRsCsdiff < -146) {
if (Rs < 2407) {
return 1;
}
else if (Rs >= 2407) {
return 4;
}
}
else if (diffRsCsdiff >= -146) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
if (ssDCval < 97) {
if (RsDiff < 8) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (Rs < 2179) {
return 2;
}
else if (Rs >= 2179) {
return 8;
}
}
}
else if (RsDiff >= 8) {
if (Rs < 4006) {
return 2;
}
else if (Rs >= 4006) {
return 8;
}
}
}
else if (ssDCval >= 97) {
if (mcTcor < 95) {
if (diffRsCsdiff < 32) {
if (negBalance < 55) {
if (Cs < 1443) {
if (MVDiff < 16) {
if (Rs < 2272) {
return 4;
}
else if (Rs >= 2272) {
return 8;
}
}
else if (MVDiff >= 16) {
return 4;
}
}
else if (Cs >= 1443) {
return 8;
}
}
else if (negBalance >= 55) {
return 4;
}
}
else if (diffRsCsdiff >= 32) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (posBalance < 47) {
if (Rs < 3163) {
return 4;
}
else if (Rs >= 3163) {
if (Rs < 3786) {
return 8;
}
else if (Rs >= 3786) {
return 4;
}
}
}
else if (posBalance >= 47) {
if (negBalance < 54) {
if (diffAFD < -6) {
return 4;
}
else if (diffAFD >= -6) {
return 8;
}
}
else if (negBalance >= 54) {
if (Rs < 2685) {
return 8;
}
else if (Rs >= 2685) {
return 4;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 99) {
if (Rs < 3254) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
return 4;
}
}
else if (Rs >= 3254) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (ssDCval < 105) {
return 1;
}
else if (ssDCval >= 105) {
return 8;
}
}
}
}
}
}
else if (Cs >= 2427) {
if (ssDCval < 105) {
if (MVDiff < 4) {
if (Cs < 2511) {
if (Rs < 3354) {
return 4;
}
else if (Rs >= 3354) {
return 8;
}
}
else if (Cs >= 2511) {
return 4;
}
}
else if (MVDiff >= 4) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 4;
}
}
}
else if (ssDCval >= 105) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
return 4;
}
}
}
}
else if (TSC >= 11) {
if (Cs < 1897) {
if (ssDCval < 106) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
else if (ssDCval >= 106) {
return 1;
}
}
else if (Cs >= 1897) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (diffAFD < -4) {
return 8;
}
else if (diffAFD >= -4) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 96) {
if (RsCsDiff < 56) {
if (posBalance < 40) {
if (MVDiff < 7) {
if (tcor < 98) {
if (negBalance < 31) {
if (Cs < 2147) {
return 4;
}
else if (Cs >= 2147) {
if (Rs < 2404) {
return 4;
}
else if (Rs >= 2404) {
if (Rs < 2530) {
return 8;
}
else if (Rs >= 2530) {
return 1;
}
}
}
}
else if (negBalance >= 31) {
if (Rs < 3171) {
return 1;
}
else if (Rs >= 3171) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
if (Rs < 4009) {
return 2;
}
else if (Rs >= 4009) {
return 4;
}
}
}
}
}
else if (tcor >= 98) {
if (CsDiff < 1) {
if (Rs < 2393) {
if (diffMVdiffVal < -5) {
return 1;
}
else if (diffMVdiffVal >= -5) {
return 4;
}
}
else if (Rs >= 2393) {
return 8;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
}
else if (MVDiff >= 7) {
if (SC < 4254) {
if (diffRsCsdiff < -2) {
if (Rs < 2026) {
return 8;
}
else if (Rs >= 2026) {
return 2;
}
}
else if (diffRsCsdiff >= -2) {
if (Rs < 2211) {
return 1;
}
else if (Rs >= 2211) {
return 2;
}
}
}
else if (SC >= 4254) {
if (ssDCval < 90) {
return 1;
}
else if (ssDCval >= 90) {
if (negBalance < 26) {
if (Rs < 2280) {
return 8;
}
else if (Rs >= 2280) {
return 4;
}
}
else if (negBalance >= 26) {
if (Cs < 1359) {
if (Rs < 3591) {
return 8;
}
else if (Rs >= 3591) {
return 4;
}
}
else if (Cs >= 1359) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 40) {
if (tcor < 93) {
if (Rs < 2604) {
return 8;
}
else if (Rs >= 2604) {
if (Rs < 3088) {
return 4;
}
else if (Rs >= 3088) {
if (Cs < 1637) {
return 4;
}
else if (Cs >= 1637) {
return 8;
}
}
}
}
else if (tcor >= 93) {
if (SC < 2324) {
if (TSC < 6) {
if (Rs < 681) {
return 1;
}
else if (Rs >= 681) {
return 4;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (SC >= 2324) {
if (Rs < 1923) {
if (Rs < 1408) {
return 8;
}
else if (Rs >= 1408) {
if (SC < 3435) {
if (negBalance < 39) {
return 2;
}
else if (negBalance >= 39) {
return 8;
}
}
else if (SC >= 3435) {
return 2;
}
}
}
else if (Rs >= 1923) {
if (Rs < 3797) {
if (negBalance < 52) {
if (Rs < 3307) {
if (tcor < 95) {
if (Rs < 2693) {
return 8;
}
else if (Rs >= 2693) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
return 4;
}
}
}
else if (tcor >= 95) {
if (diffAFD < -2) {
if (Rs < 2141) {
return 2;
}
else if (Rs >= 2141) {
return 8;
}
}
else if (diffAFD >= -2) {
if (TSC < 1) {
return 2;
}
else if (TSC >= 1) {
return 8;
}
}
}
}
else if (Rs >= 3307) {
if (MVDiff < 4) {
if (mcTcor < 98) {
if (Cs < 1390) {
return 2;
}
else if (Cs >= 1390) {
return 4;
}
}
else if (mcTcor >= 98) {
if (Cs < 1514) {
return 4;
}
else if (Cs >= 1514) {
return 8;
}
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (negBalance >= 52) {
return 4;
}
}
else if (Rs >= 3797) {
if (SC < 5739) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
else if (SC >= 5739) {
return 8;
}
}
}
}
}
}
}
else if (RsCsDiff >= 56) {
if (Rs < 1916) {
if (tcor < 89) {
if (diffRsCsdiff < 6) {
if (negBalance < 53) {
return 4;
}
else if (negBalance >= 53) {
return 2;
}
}
else if (diffRsCsdiff >= 6) {
if (Rs < 740) {
return 4;
}
else if (Rs >= 740) {
return 8;
}
}
}
else if (tcor >= 89) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 2;
}
}
}
else if (Rs >= 1916) {
if (Rs < 3466) {
if (tcor < 92) {
if (SC < 3961) {
return 2;
}
else if (SC >= 3961) {
if (Rs < 3074) {
return 4;
}
else if (Rs >= 3074) {
return 8;
}
}
}
else if (tcor >= 92) {
if (Cs < 2742) {
if (CsDiff < 7) {
if (Rs < 2220) {
if (ssDCval < 101) {
if (Cs < 1755) {
return 1;
}
else if (Cs >= 1755) {
return 2;
}
}
else if (ssDCval >= 101) {
return 8;
}
}
else if (Rs >= 2220) {
if (ssDCval < 94) {
if (Rs < 2737) {
return 8;
}
else if (Rs >= 2737) {
return 1;
}
}
else if (ssDCval >= 94) {
return 8;
}
}
}
else if (CsDiff >= 7) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (Cs >= 2742) {
if (Rs < 2941) {
return 8;
}
else if (Rs >= 2941) {
return 4;
}
}
}
}
else if (Rs >= 3466) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 110) {
if (Rs < 1262) {
if (mu_mv_mag_sq < 39) {
if (refDCval < 123) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
return 8;
}
}
else if (refDCval >= 123) {
if (RsDiff < 5) {
if (MVDiff < 36) {
return 2;
}
else if (MVDiff >= 36) {
return 8;
}
}
else if (RsDiff >= 5) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 39) {
if (Rs < 989) {
if (ssDCval < 119) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
if (refDCval < 111) {
if (Rs < 815) {
return 1;
}
else if (Rs >= 815) {
return 4;
}
}
else if (refDCval >= 111) {
return 8;
}
}
}
else if (ssDCval >= 119) {
if (Cs < 2094) {
return 8;
}
else if (Cs >= 2094) {
return 1;
}
}
}
else if (Rs >= 989) {
if (CsDiff < 6) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
if (posBalance < 62) {
if (diffRsCsdiff < 5) {
if (Cs < 1849) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (MVDiff < 5) {
if (Cs < 1651) {
if (Cs < 1534) {
return 4;
}
else if (Cs >= 1534) {
return 8;
}
}
else if (Cs >= 1651) {
return 4;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
}
else if (Cs >= 1849) {
return 8;
}
}
else if (diffRsCsdiff >= 5) {
if (diffRsCsdiff < 10) {
if (RsDiff < 3) {
if (negBalance < 38) {
return 8;
}
else if (negBalance >= 38) {
return 4;
}
}
else if (RsDiff >= 3) {
return 8;
}
}
else if (diffRsCsdiff >= 10) {
return 4;
}
}
}
else if (posBalance >= 62) {
return 8;
}
}
}
else if (CsDiff >= 6) {
return 8;
}
}
}
}
else if (Rs >= 1262) {
if (mu_mv_mag_sq < 32) {
if (Rs < 2026) {
if (ssDCval < 142) {
if (refDCval < 124) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (negBalance < 47) {
if (Cs < 1774) {
if (posBalance < 67) {
return 4;
}
else if (posBalance >= 67) {
return 2;
}
}
else if (Cs >= 1774) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (Rs < 1593) {
return 1;
}
else if (Rs >= 1593) {
return 2;
}
}
}
}
else if (negBalance >= 47) {
if (SC < 2900) {
return 2;
}
else if (SC >= 2900) {
return 8;
}
}
}
}
else if (refDCval >= 124) {
if (negBalance < 49) {
if (mcTcor < 91) {
return 1;
}
else if (mcTcor >= 91) {
if (mu_mv_mag_sq < 3) {
return 8;
}
else if (mu_mv_mag_sq >= 3) {
return 1;
}
}
}
else if (negBalance >= 49) {
return 8;
}
}
}
else if (ssDCval >= 142) {
if (Cs < 1871) {
if (Cs < 1420) {
if (Rs < 1441) {
return 2;
}
else if (Rs >= 1441) {
return 1;
}
}
else if (Cs >= 1420) {
return 2;
}
}
else if (Cs >= 1871) {
return 8;
}
}
}
else if (Rs >= 2026) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (tcor < 97) {
if (posBalance < 57) {
if (ssDCval < 142) {
if (Cs < 1872) {
if (refDCval < 125) {
if (mu_mv_mag_sq < 45) {
if (Cs < 1379) {
return 1;
}
else if (Cs >= 1379) {
return 8;
}
}
else if (mu_mv_mag_sq >= 45) {
return 8;
}
}
else if (refDCval >= 125) {
if (mcTcor < 90) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
if (Cs < 1759) {
return 8;
}
else if (Cs >= 1759) {
return 1;
}
}
}
else if (mcTcor >= 90) {
if (MVDiff < 8) {
if (ssDCval < 128) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
else if (ssDCval >= 128) {
if (mu_mv_mag_sq < 92) {
return 8;
}
else if (mu_mv_mag_sq >= 92) {
if (Cs < 1349) {
return 1;
}
else if (Cs >= 1349) {
if (diffRsCsdiff < -18) {
if (Cs < 1592) {
return 8;
}
else if (Cs >= 1592) {
return 1;
}
}
else if (diffRsCsdiff >= -18) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Rs < 2513) {
return 8;
}
else if (Rs >= 2513) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 8) {
if (posBalance < 38) {
if (Rs < 3545) {
return 4;
}
else if (Rs >= 3545) {
return 8;
}
}
else if (posBalance >= 38) {
if (diffAFD < -3) {
if (diffMVdiffVal < -1) {
if (Rs < 3720) {
if (Cs < 1359) {
return 8;
}
else if (Cs >= 1359) {
return 2;
}
}
else if (Rs >= 3720) {
return 8;
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1872) {
return 8;
}
}
else if (ssDCval >= 142) {
if (SC < 5778) {
if (posBalance < 30) {
return 2;
}
else if (posBalance >= 30) {
return 8;
}
}
else if (SC >= 5778) {
return 1;
}
}
}
else if (posBalance >= 57) {
if (negBalance < 48) {
if (AFD < 13) {
if (Rs < 1879) {
if (MVDiff < 13) {
if (diffAFD < 1) {
if (Cs < 1479) {
return 8;
}
else if (Cs >= 1479) {
return 1;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (Rs >= 1879) {
if (Rs < 3881) {
return 8;
}
else if (Rs >= 3881) {
if (Rs < 3902) {
return 1;
}
else if (Rs >= 3902) {
return 8;
}
}
}
}
else if (AFD >= 13) {
if (posBalance < 66) {
if (RsCsDiff < 63) {
return 1;
}
else if (RsCsDiff >= 63) {
return 8;
}
}
else if (posBalance >= 66) {
if (Rs < 2538) {
return 4;
}
else if (Rs >= 2538) {
return 8;
}
}
}
}
else if (negBalance >= 48) {
if (Rs < 2396) {
return 8;
}
else if (Rs >= 2396) {
if (mu_mv_mag_sq < 75) {
return 8;
}
else if (mu_mv_mag_sq >= 75) {
if (tcor < 94) {
if (Cs < 2117) {
return 1;
}
else if (Cs >= 2117) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (tcor >= 94) {
return 8;
}
}
}
}
}
}
else if (tcor >= 97) {
if (Cs < 1360) {
if (Rs < 2521) {
return 8;
}
else if (Rs >= 2521) {
if (CsDiff < 4) {
return 2;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (Cs >= 1360) {
if (CsDiff < 2) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (RsCsDiff < 2) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
if (Rs < 2600) {
return 8;
}
else if (Rs >= 2600) {
return 4;
}
}
}
else if (RsCsDiff >= 2) {
if (Rs < 2785) {
return 8;
}
else if (Rs >= 2785) {
if (Rs < 2996) {
return 2;
}
else if (Rs >= 2996) {
return 8;
}
}
}
}
}
else if (CsDiff >= 2) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (Rs < 2079) {
return 2;
}
else if (Rs >= 2079) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 163) {
if (Rs < 2289) {
if (Rs < 1064) {
return 1;
}
else if (Rs >= 1064) {
return 2;
}
}
else if (Rs >= 2289) {
return 8;
}
}
}
else if (Rs >= 4357) {
if (tcor < 94) {
if (mu_mv_mag_sq < 23) {
return 8;
}
else if (mu_mv_mag_sq >= 23) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 53) {
return 1;
}
else if (mu_mv_mag_sq >= 53) {
if (refDCval < 91) {
if (Rs < 4589) {
return 4;
}
else if (Rs >= 4589) {
return 4;
}
}
else if (refDCval >= 91) {
return 8;
}
}
}
else if (posBalance >= 53) {
return 1;
}
}
}
else if (tcor >= 94) {
if (Cs < 1570) {
return 2;
}
else if (Cs >= 1570) {
if (Cs < 2116) {
if (tcor < 97) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (Rs < 4577) {
return 8;
}
else if (Rs >= 4577) {
if (Rs < 4696) {
return 2;
}
else if (Rs >= 4696) {
return 8;
}
}
}
}
else if (tcor >= 97) {
if (Rs < 4645) {
return 2;
}
else if (Rs >= 4645) {
if (Rs < 5059) {
return 8;
}
else if (Rs >= 5059) {
return 2;
}
}
}
}
else if (Cs >= 2116) {
if (Rs < 6296) {
return 8;
}
else if (Rs >= 6296) {
if (Rs < 6602) {
return 4;
}
else if (Rs >= 6602) {
return 8;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1937) {
if (Rs < 3354) {
if (refDCval < 86) {
if (negBalance < 47) {
if (mcTcor < 87) {
if (CsDiff < 10) {
if (Cs < 1784) {
if (mu_mv_mag_sq < 41) {
if (TSC < 13) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 1;
}
}
else if (TSC >= 13) {
return 8;
}
}
else if (mu_mv_mag_sq >= 41) {
if (Cs < 1757) {
if (RsCsDiff < 198) {
return 8;
}
else if (RsCsDiff >= 198) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 1;
}
}
}
else if (Cs >= 1757) {
return 1;
}
}
}
else if (Cs >= 1784) {
if (Rs < 1996) {
return 4;
}
else if (Rs >= 1996) {
return 8;
}
}
}
else if (CsDiff >= 10) {
if (Rs < 1951) {
if (mcTcor < 81) {
if (refDCval < 46) {
return 8;
}
else if (refDCval >= 46) {
return 4;
}
}
else if (mcTcor >= 81) {
if (Cs < 1556) {
return 2;
}
else if (Cs >= 1556) {
return 1;
}
}
}
else if (Rs >= 1951) {
if (mu_mv_mag_sq < 40) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 8;
}
}
else if (mu_mv_mag_sq >= 40) {
if (Cs < 1427) {
return 8;
}
else if (Cs >= 1427) {
return 1;
}
}
}
}
}
else if (mcTcor >= 87) {
if (TSC < 7) {
if (posBalance < 62) {
return 1;
}
else if (posBalance >= 62) {
if (Rs < 1115) {
return 4;
}
else if (Rs >= 1115) {
return 8;
}
}
}
else if (TSC >= 7) {
if (posBalance < 48) {
if (SC < 4688) {
if (ssDCval < 82) {
if (Cs < 1325) {
if (Rs < 2796) {
return 2;
}
else if (Rs >= 2796) {
return 4;
}
}
else if (Cs >= 1325) {
return 2;
}
}
else if (ssDCval >= 82) {
if (Rs < 2498) {
return 8;
}
else if (Rs >= 2498) {
if (Rs < 3006) {
return 2;
}
else if (Rs >= 3006) {
return 4;
}
}
}
}
else if (SC >= 4688) {
if (Rs < 3092) {
return 1;
}
else if (Rs >= 3092) {
return 2;
}
}
}
else if (posBalance >= 48) {
if (Rs < 2711) {
if (tcor < 75) {
if (gchDC < 1) {
if (Rs < 2268) {
return 1;
}
else if (Rs >= 2268) {
return 8;
}
}
else if (gchDC >= 1) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 2;
}
}
}
else if (tcor >= 75) {
if (CsDiff < 8) {
if (diffMVdiffVal < -5) {
return 8;
}
else if (diffMVdiffVal >= -5) {
if (Cs < 1776) {
return 4;
}
else if (Cs >= 1776) {
return 1;
}
}
}
else if (CsDiff >= 8) {
return 8;
}
}
}
else if (Rs >= 2711) {
if (posBalance < 61) {
if (negBalance < 43) {
return 2;
}
else if (negBalance >= 43) {
if (Rs < 3205) {
if (Rs < 3113) {
return 4;
}
else if (Rs >= 3113) {
return 8;
}
}
else if (Rs >= 3205) {
return 2;
}
}
}
else if (posBalance >= 61) {
if (Rs < 3288) {
return 8;
}
else if (Rs >= 3288) {
return 2;
}
}
}
}
}
}
}
else if (negBalance >= 47) {
if (mu_mv_mag_sq < 9) {
if (RsDiff < 13) {
if (tcor < 57) {
if (Rs < 1966) {
if (diffMVdiffVal < -7) {
if (RsCsDiff < 154) {
return 4;
}
else if (RsCsDiff >= 154) {
if (Rs < 1667) {
return 8;
}
else if (Rs >= 1667) {
return 1;
}
}
}
else if (diffMVdiffVal >= -7) {
if (diffAFD < -1) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
}
else if (Rs >= 1966) {
return 1;
}
}
else if (tcor >= 57) {
if (Rs < 1637) {
if (refDCval < 71) {
if (Cs < 1429) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
if (Rs < 1388) {
return 4;
}
else if (Rs >= 1388) {
return 1;
}
}
}
else if (Cs >= 1429) {
if (Rs < 1630) {
return 4;
}
else if (Rs >= 1630) {
return 1;
}
}
}
else if (refDCval >= 71) {
if (negBalance < 58) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (MVDiff < 22) {
if (diffRsCsdiff < 24) {
return 1;
}
else if (diffRsCsdiff >= 24) {
if (Cs < 1529) {
return 2;
}
else if (Cs >= 1529) {
if (Rs < 1290) {
return 4;
}
else if (Rs >= 1290) {
return 1;
}
}
}
}
else if (MVDiff >= 22) {
if (CsDiff < 11) {
return 4;
}
else if (CsDiff >= 11) {
if (Cs < 1529) {
return 2;
}
else if (Cs >= 1529) {
return 8;
}
}
}
}
}
else if (negBalance >= 58) {
if (Rs < 1476) {
return 2;
}
else if (Rs >= 1476) {
return 8;
}
}
}
}
else if (Rs >= 1637) {
if (Rs < 2370) {
if (RsCsDiff < 141) {
if (Cs < 1490) {
if (tcor < 79) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (Cs < 1408) {
return 8;
}
else if (Cs >= 1408) {
return 4;
}
}
}
else if (tcor >= 79) {
if (Rs < 2117) {
return 2;
}
else if (Rs >= 2117) {
return 8;
}
}
}
else if (Cs >= 1490) {
if (TSC < 11) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (mcTcor < 91) {
if (Rs < 1808) {
return 4;
}
else if (Rs >= 1808) {
return 8;
}
}
else if (mcTcor >= 91) {
return 4;
}
}
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (RsCsDiff >= 141) {
if (Cs < 1862) {
if (negBalance < 53) {
if (Rs < 2034) {
if (diffMVdiffVal < 4) {
return 8;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 1813) {
return 4;
}
else if (Rs >= 1813) {
return 8;
}
}
}
else if (Rs >= 2034) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
if (tcor < 82) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
else if (tcor >= 82) {
return 8;
}
}
}
}
else if (negBalance >= 53) {
if (Rs < 1678) {
return 2;
}
else if (Rs >= 1678) {
if (Cs < 1613) {
if (mu_mv_mag_sq < 3) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffRsCsdiff < 28) {
return 4;
}
else if (diffRsCsdiff >= 28) {
if (Cs < 1481) {
return 4;
}
else if (Cs >= 1481) {
return 8;
}
}
}
}
else if (Cs >= 1613) {
if (CsDiff < 13) {
return 4;
}
else if (CsDiff >= 13) {
return 2;
}
}
}
}
}
else if (Cs >= 1862) {
if (ssDCval < 73) {
if (Rs < 2060) {
return 1;
}
else if (Rs >= 2060) {
return 4;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
}
}
else if (Rs >= 2370) {
if (refDCval < 74) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (Cs < 1703) {
return 2;
}
else if (Cs >= 1703) {
return 4;
}
}
}
else if (refDCval >= 74) {
if (Cs < 1432) {
return 2;
}
else if (Cs >= 1432) {
if (MVDiff < 24) {
return 4;
}
else if (MVDiff >= 24) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 13) {
if (TSC < 14) {
if (gchDC < 4) {
if (negBalance < 51) {
if (Rs < 2952) {
return 4;
}
else if (Rs >= 2952) {
return 2;
}
}
else if (negBalance >= 51) {
if (posBalance < 56) {
if (Rs < 2020) {
return 1;
}
else if (Rs >= 2020) {
return 4;
}
}
else if (posBalance >= 56) {
return 2;
}
}
}
else if (gchDC >= 4) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 2;
}
}
}
else if (TSC >= 14) {
if (Cs < 1536) {
return 1;
}
else if (Cs >= 1536) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 9) {
if (Rs < 1877) {
if (SC < 3383) {
if (MVDiff < 19) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Rs < 1345) {
return 8;
}
else if (Rs >= 1345) {
if (AFD < 23) {
return 2;
}
else if (AFD >= 23) {
return 4;
}
}
}
}
else if (MVDiff >= 19) {
if (MVDiff < 44) {
if (AFD < 26) {
if (Cs < 1339) {
if (Rs < 1611) {
return 8;
}
else if (Rs >= 1611) {
return 1;
}
}
else if (Cs >= 1339) {
if (TSC < 9) {
if (Cs < 1423) {
return 4;
}
else if (Cs >= 1423) {
if (Rs < 1461) {
return 1;
}
else if (Rs >= 1461) {
return 8;
}
}
}
else if (TSC >= 9) {
return 4;
}
}
}
else if (AFD >= 26) {
return 2;
}
}
else if (MVDiff >= 44) {
if (diffMVdiffVal < -34) {
return 1;
}
else if (diffMVdiffVal >= -34) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
if (gchDC < 2) {
if (Rs < 1224) {
return 4;
}
else if (Rs >= 1224) {
return 8;
}
}
else if (gchDC >= 2) {
return 8;
}
}
}
}
}
}
else if (SC >= 3383) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 1;
}
}
}
else if (Rs >= 1877) {
if (refDCval < 71) {
if (RsCsDiff < 347) {
if (SC < 3515) {
if (Rs < 1974) {
return 8;
}
else if (Rs >= 1974) {
return 1;
}
}
else if (SC >= 3515) {
if (Rs < 3325) {
return 8;
}
else if (Rs >= 3325) {
return 4;
}
}
}
else if (RsCsDiff >= 347) {
return 4;
}
}
else if (refDCval >= 71) {
if (Cs < 1750) {
if (negBalance < 54) {
if (diffMVdiffVal < -4) {
if (Rs < 2660) {
return 1;
}
else if (Rs >= 2660) {
return 4;
}
}
else if (diffMVdiffVal >= -4) {
if (tcor < 80) {
if (Rs < 1914) {
return 1;
}
else if (Rs >= 1914) {
return 2;
}
}
else if (tcor >= 80) {
if (Rs < 2757) {
return 8;
}
else if (Rs >= 2757) {
return 2;
}
}
}
}
else if (negBalance >= 54) {
if (MVDiff < 32) {
if (Rs < 3028) {
return 8;
}
else if (Rs >= 3028) {
return 2;
}
}
else if (MVDiff >= 32) {
if (Rs < 2359) {
return 1;
}
else if (Rs >= 2359) {
return 2;
}
}
}
}
else if (Cs >= 1750) {
if (diffMVdiffVal < -9) {
return 1;
}
else if (diffMVdiffVal >= -9) {
if (mcTcor < 89) {
if (CsDiff < 9) {
if (mcTcor < 82) {
return 8;
}
else if (mcTcor >= 82) {
return 4;
}
}
else if (CsDiff >= 9) {
return 8;
}
}
else if (mcTcor >= 89) {
if (ssDCval < 76) {
return 1;
}
else if (ssDCval >= 76) {
return 8;
}
}
}
}
}
}
}
}
}
else if (refDCval >= 86) {
if (AFD < 19) {
if (refDCval < 99) {
if (TSC < 9) {
if (Cs < 1675) {
if (posBalance < 53) {
if (AFD < 11) {
if (Rs < 2881) {
return 1;
}
else if (Rs >= 2881) {
return 2;
}
}
else if (AFD >= 11) {
return 8;
}
}
else if (posBalance >= 53) {
if (RsDiff < 9) {
if (diffRsCsdiff < 23) {
return 2;
}
else if (diffRsCsdiff >= 23) {
if (Rs < 1595) {
return 1;
}
else if (Rs >= 1595) {
return 8;
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 94) {
return 2;
}
else if (refDCval >= 94) {
if (Rs < 2556) {
return 2;
}
else if (Rs >= 2556) {
return 4;
}
}
}
}
}
else if (Cs >= 1675) {
if (TSC < 7) {
if (Rs < 2365) {
return 1;
}
else if (Rs >= 2365) {
return 2;
}
}
else if (TSC >= 7) {
if (posBalance < 65) {
if (AFD < 16) {
if (Rs < 2757) {
return 8;
}
else if (Rs >= 2757) {
if (Rs < 2809) {
return 2;
}
else if (Rs >= 2809) {
return 8;
}
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (posBalance >= 65) {
return 2;
}
}
}
}
else if (TSC >= 9) {
if (Rs < 2625) {
if (Rs < 1846) {
if (refDCval < 94) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
if (Rs < 1507) {
return 8;
}
else if (Rs >= 1507) {
return 4;
}
}
}
else if (refDCval >= 94) {
if (Rs < 1522) {
return 2;
}
else if (Rs >= 1522) {
return 1;
}
}
}
else if (Rs >= 1846) {
if (mu_mv_mag_sq < 7) {
if (mcTcor < 97) {
if (negBalance < 50) {
if (RsDiff < 10) {
if (Rs < 1979) {
return 1;
}
else if (Rs >= 1979) {
return 8;
}
}
else if (RsDiff >= 10) {
return 2;
}
}
else if (negBalance >= 50) {
return 8;
}
}
else if (mcTcor >= 97) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
}
else if (Rs >= 2625) {
if (negBalance < 49) {
if (gchDC < 2) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (gchDC >= 2) {
return 8;
}
}
else if (negBalance >= 49) {
if (SC < 4822) {
return 8;
}
else if (SC >= 4822) {
return 4;
}
}
}
}
}
else if (refDCval >= 99) {
if (mu_mv_mag_sq < 2) {
if (Rs < 2969) {
if (Cs < 1539) {
if (TSC < 12) {
if (posBalance < 63) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (ssDCval < 104) {
if (Rs < 2636) {
return 4;
}
else if (Rs >= 2636) {
return 8;
}
}
else if (ssDCval >= 104) {
return 8;
}
}
}
else if (posBalance >= 63) {
return 4;
}
}
else if (TSC >= 12) {
return 1;
}
}
else if (Cs >= 1539) {
if (SC < 3463) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
return 4;
}
}
else if (SC >= 3463) {
if (refDCval < 104) {
if (negBalance < 51) {
return 2;
}
else if (negBalance >= 51) {
return 8;
}
}
else if (refDCval >= 104) {
return 8;
}
}
}
}
else if (Rs >= 2969) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
if (diffMVdiffVal < -3) {
if (MVDiff < 16) {
if (Rs < 3234) {
return 4;
}
else if (Rs >= 3234) {
return 1;
}
}
else if (MVDiff >= 16) {
return 8;
}
}
else if (diffMVdiffVal >= -3) {
if (ssDCval < 156) {
return 4;
}
else if (ssDCval >= 156) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (SC < 3267) {
if (negBalance < 30) {
return 2;
}
else if (negBalance >= 30) {
if (mcTcor < 96) {
if (Cs < 1354) {
if (Rs < 1609) {
return 1;
}
else if (Rs >= 1609) {
return 2;
}
}
else if (Cs >= 1354) {
return 8;
}
}
else if (mcTcor >= 96) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 1506) {
return 4;
}
else if (Rs >= 1506) {
return 8;
}
}
}
}
}
else if (SC >= 3267) {
if (ssDCval < 111) {
if (mu_mv_mag_sq < 77) {
if (tcor < 59) {
if (Rs < 2425) {
return 4;
}
else if (Rs >= 2425) {
return 8;
}
}
else if (tcor >= 59) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
if (mu_mv_mag_sq < 5) {
if (Rs < 2777) {
return 8;
}
else if (Rs >= 2777) {
if (Cs < 1517) {
return 4;
}
else if (Cs >= 1517) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 5) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 77) {
if (Cs < 1722) {
return 4;
}
else if (Cs >= 1722) {
return 8;
}
}
}
else if (ssDCval >= 111) {
if (TSC < 13) {
if (posBalance < 17) {
return 1;
}
else if (posBalance >= 17) {
if (diffMVdiffVal < 25) {
if (posBalance < 47) {
if (mu_mv_mag_sq < 33) {
if (Rs < 2510) {
return 2;
}
else if (Rs >= 2510) {
return 8;
}
}
else if (mu_mv_mag_sq >= 33) {
return 8;
}
}
else if (posBalance >= 47) {
return 8;
}
}
else if (diffMVdiffVal >= 25) {
if (Rs < 2294) {
return 8;
}
else if (Rs >= 2294) {
return 2;
}
}
}
}
else if (TSC >= 13) {
if (Cs < 1502) {
return 1;
}
else if (Cs >= 1502) {
return 8;
}
}
}
}
}
}
}
else if (AFD >= 19) {
if (TSC < 10) {
if (mu_mv_mag_sq < 17) {
if (Rs < 1765) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
}
else if (Rs >= 1765) {
if (negBalance < 56) {
if (Rs < 2431) {
return 2;
}
else if (Rs >= 2431) {
return 1;
}
}
else if (negBalance >= 56) {
if (CsDiff < 10) {
return 1;
}
else if (CsDiff >= 10) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (mcTcor < 91) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 2;
}
}
else if (mcTcor >= 91) {
return 2;
}
}
}
else if (TSC >= 10) {
if (Rs < 1359) {
if (gchDC < 3) {
if (ssDCval < 87) {
return 4;
}
else if (ssDCval >= 87) {
if (posBalance < 67) {
return 2;
}
else if (posBalance >= 67) {
return 1;
}
}
}
else if (gchDC >= 3) {
if (diffAFD < 1) {
if (SC < 2717) {
if (Rs < 1187) {
return 1;
}
else if (Rs >= 1187) {
return 2;
}
}
else if (SC >= 2717) {
return 4;
}
}
else if (diffAFD >= 1) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
if (Rs < 1072) {
return 2;
}
else if (Rs >= 1072) {
return 8;
}
}
}
}
}
else if (Rs >= 1359) {
if (AFD < 33) {
if (refDCval < 127) {
if (tcor < 78) {
if (Cs < 1335) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 2;
}
}
else if (Cs >= 1335) {
if (posBalance < 47) {
if (Rs < 2518) {
return 8;
}
else if (Rs >= 2518) {
return 2;
}
}
else if (posBalance >= 47) {
if (refDCval < 90) {
if (Cs < 1747) {
if (diffAFD < 2) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
return 8;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
else if (Cs >= 1747) {
if (negBalance < 56) {
return 4;
}
else if (negBalance >= 56) {
return 1;
}
}
}
else if (refDCval >= 90) {
if (Cs < 1582) {
if (refDCval < 95) {
return 8;
}
else if (refDCval >= 95) {
if (mcTcor < 93) {
if (SC < 2902) {
return 8;
}
else if (SC >= 2902) {
if (ssDCval < 111) {
if (mu_mv_mag_sq < 11) {
if (ssDCval < 107) {
return 8;
}
else if (ssDCval >= 107) {
return 4;
}
}
else if (mu_mv_mag_sq >= 11) {
return 4;
}
}
else if (ssDCval >= 111) {
if (Rs < 1710) {
return 4;
}
else if (Rs >= 1710) {
return 8;
}
}
}
}
else if (mcTcor >= 93) {
if (Cs < 1405) {
return 1;
}
else if (Cs >= 1405) {
return 8;
}
}
}
}
else if (Cs >= 1582) {
if (posBalance < 63) {
if (RsCsDiff < 401) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (RsCsDiff >= 401) {
if (mu_mv_mag_sq < 4) {
return 2;
}
else if (mu_mv_mag_sq >= 4) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (RsCsDiff < 198) {
if (Cs < 1690) {
return 8;
}
else if (Cs >= 1690) {
if (Rs < 1592) {
return 8;
}
else if (Rs >= 1592) {
return 1;
}
}
}
else if (RsCsDiff >= 198) {
if (mu_mv_mag_sq < 11) {
if (tcor < 61) {
return 4;
}
else if (tcor >= 61) {
if (Rs < 3035) {
return 8;
}
else if (Rs >= 3035) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (Rs < 2024) {
return 4;
}
else if (Rs >= 2024) {
return 8;
}
}
}
}
}
}
}
}
}
else if (tcor >= 78) {
if (negBalance < 45) {
if (Rs < 2843) {
return 1;
}
else if (Rs >= 2843) {
return 8;
}
}
else if (negBalance >= 45) {
if (RsCsDiff < 400) {
return 8;
}
else if (RsCsDiff >= 400) {
return 2;
}
}
}
}
else if (refDCval >= 127) {
if (mu_mv_mag_sq < 14) {
if (RsDiff < 13) {
if (RsCsDiff < 148) {
if (Rs < 1619) {
return 2;
}
else if (Rs >= 1619) {
if (Rs < 2432) {
return 1;
}
else if (Rs >= 2432) {
return 8;
}
}
}
else if (RsCsDiff >= 148) {
return 1;
}
}
else if (RsDiff >= 13) {
if (ssDCval < 135) {
return 1;
}
else if (ssDCval >= 135) {
if (Cs < 1341) {
return 1;
}
else if (Cs >= 1341) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 14) {
if (Rs < 2630) {
return 8;
}
else if (Rs >= 2630) {
if (posBalance < 49) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 4;
}
}
else if (posBalance >= 49) {
return 2;
}
}
}
}
}
else if (AFD >= 33) {
if (gchDC < 5) {
if (Cs < 1774) {
return 2;
}
else if (Cs >= 1774) {
return 1;
}
}
else if (gchDC >= 5) {
if (diffMVdiffVal < -3) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
if (TSC < 16) {
return 4;
}
else if (TSC >= 16) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 3354) {
if (tcor < 85) {
if (RsDiff < 13) {
if (posBalance < 50) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (Rs < 3533) {
if (Rs < 3406) {
return 8;
}
else if (Rs >= 3406) {
return 2;
}
}
else if (Rs >= 3533) {
return 4;
}
}
}
}
else if (posBalance >= 50) {
if (diffMVdiffVal < -11) {
return 1;
}
else if (diffMVdiffVal >= -11) {
if (SC < 4947) {
if (Cs < 1426) {
return 8;
}
else if (Cs >= 1426) {
return 2;
}
}
else if (SC >= 4947) {
return 8;
}
}
}
}
else if (RsDiff >= 13) {
if (mu_mv_mag_sq < 13) {
if (refDCval < 100) {
return 2;
}
else if (refDCval >= 100) {
if (TSC < 14) {
if (RsCsDiff < 404) {
if (Cs < 1494) {
return 8;
}
else if (Cs >= 1494) {
if (CsDiff < 7) {
return 8;
}
else if (CsDiff >= 7) {
if (Cs < 1733) {
if (Rs < 3571) {
return 4;
}
else if (Rs >= 3571) {
return 1;
}
}
else if (Cs >= 1733) {
if (TSC < 10) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 404) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
if (Rs < 5106) {
if (ssDCval < 143) {
if (SC < 4902) {
return 8;
}
else if (SC >= 4902) {
return 2;
}
}
else if (ssDCval >= 143) {
return 1;
}
}
else if (Rs >= 5106) {
if (Rs < 6216) {
return 8;
}
else if (Rs >= 6216) {
if (posBalance < 59) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (MVDiff < 28) {
return 2;
}
else if (MVDiff >= 28) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
return 2;
}
}
}
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
}
}
else if (TSC >= 14) {
if (mcTcor < 85) {
return 4;
}
else if (mcTcor >= 85) {
if (negBalance < 54) {
if (diffTSC < 3) {
return 1;
}
else if (diffTSC >= 3) {
return 2;
}
}
else if (negBalance >= 54) {
if (Cs < 1320) {
return 2;
}
else if (Cs >= 1320) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 13) {
if (SC < 7290) {
if (RsCsDiff < 627) {
if (tcor < 70) {
if (AFD < 24) {
if (refDCval < 138) {
if (Cs < 1880) {
return 8;
}
else if (Cs >= 1880) {
return 4;
}
}
else if (refDCval >= 138) {
return 4;
}
}
else if (AFD >= 24) {
if (diffAFD < -3) {
if (Cs < 1408) {
return 2;
}
else if (Cs >= 1408) {
return 8;
}
}
else if (diffAFD >= -3) {
if (Rs < 3463) {
return 2;
}
else if (Rs >= 3463) {
if (diffTSC < 3) {
return 4;
}
else if (diffTSC >= 3) {
return 2;
}
}
}
}
}
else if (tcor >= 70) {
if (mu_mv_mag_sq < 24) {
if (AFD < 26) {
return 1;
}
else if (AFD >= 26) {
return 4;
}
}
else if (mu_mv_mag_sq >= 24) {
if (mcTcor < 85) {
if (Rs < 3935) {
return 1;
}
else if (Rs >= 3935) {
return 8;
}
}
else if (mcTcor >= 85) {
if (Rs < 4866) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
if (Cs < 1715) {
return 2;
}
else if (Cs >= 1715) {
return 8;
}
}
}
}
else if (Rs >= 4866) {
if (Rs < 5176) {
return 4;
}
else if (Rs >= 5176) {
return 1;
}
}
}
}
}
}
else if (RsCsDiff >= 627) {
if (gchDC < 3) {
if (Rs < 3435) {
return 8;
}
else if (Rs >= 3435) {
if (ssDCval < 146) {
if (Cs < 1809) {
return 2;
}
else if (Cs >= 1809) {
if (Rs < 4337) {
return 4;
}
else if (Rs >= 4337) {
return 2;
}
}
}
else if (ssDCval >= 146) {
if (Rs < 5129) {
return 4;
}
else if (Rs >= 5129) {
return 2;
}
}
}
}
else if (gchDC >= 3) {
if (RsDiff < 27) {
if (refDCval < 103) {
return 8;
}
else if (refDCval >= 103) {
return 1;
}
}
else if (RsDiff >= 27) {
if (CsDiff < 15) {
return 2;
}
else if (CsDiff >= 15) {
return 1;
}
}
}
}
}
else if (SC >= 7290) {
if (Cs < 1574) {
return 1;
}
else if (Cs >= 1574) {
return 4;
}
}
}
}
}
else if (tcor >= 85) {
if (RsDiff < 9) {
if (AFD < 12) {
if (tcor < 94) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
if (Rs < 4970) {
return 8;
}
else if (Rs >= 4970) {
return 1;
}
}
}
else if (tcor >= 94) {
if (Rs < 4751) {
return 2;
}
else if (Rs >= 4751) {
return 1;
}
}
}
else if (AFD >= 12) {
return 8;
}
}
else if (RsDiff >= 9) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (refDCval < 101) {
if (mcTcor < 92) {
if (negBalance < 50) {
if (RsDiff < 12) {
if (Rs < 3594) {
return 2;
}
else if (Rs >= 3594) {
if (Rs < 3829) {
return 4;
}
else if (Rs >= 3829) {
return 8;
}
}
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (negBalance >= 50) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (Rs < 3527) {
return 8;
}
else if (Rs >= 3527) {
if (Rs < 3792) {
return 1;
}
else if (Rs >= 3792) {
return 4;
}
}
}
}
}
else if (mcTcor >= 92) {
if (posBalance < 42) {
if (negBalance < 43) {
if (Cs < 1598) {
return 4;
}
else if (Cs >= 1598) {
if (Rs < 4161) {
return 2;
}
else if (Rs >= 4161) {
return 8;
}
}
}
else if (negBalance >= 43) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (ssDCval < 62) {
if (CsDiff < 7) {
return 2;
}
else if (CsDiff >= 7) {
if (Rs < 4028) {
return 8;
}
else if (Rs >= 4028) {
return 4;
}
}
}
else if (ssDCval >= 62) {
if (Cs < 1931) {
return 2;
}
else if (Cs >= 1931) {
if (Rs < 3882) {
return 8;
}
else if (Rs >= 3882) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 101) {
if (RsDiff < 11) {
if (mu_mv_mag_sq < 84) {
if (Rs < 4997) {
return 8;
}
else if (Rs >= 4997) {
return 2;
}
}
else if (mu_mv_mag_sq >= 84) {
return 1;
}
}
else if (RsDiff >= 11) {
if (MVDiff < 23) {
if (posBalance < 58) {
if (negBalance < 64) {
return 2;
}
else if (negBalance >= 64) {
if (Cs < 1402) {
return 1;
}
else if (Cs >= 1402) {
return 2;
}
}
}
else if (posBalance >= 58) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
if (Cs < 1378) {
return 4;
}
else if (Cs >= 1378) {
if (Rs < 4312) {
return 8;
}
else if (Rs >= 4312) {
return 1;
}
}
}
}
}
else if (MVDiff >= 23) {
if (RsCsDiff < 397) {
if (tcor < 86) {
if (Rs < 5249) {
return 1;
}
else if (Rs >= 5249) {
return 2;
}
}
else if (tcor >= 86) {
if (MVDiff < 86) {
return 8;
}
else if (MVDiff >= 86) {
return 1;
}
}
}
else if (RsCsDiff >= 397) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Cs >= 1937) {
if (AFD < 12) {
if (ssDCval < 100) {
if (Rs < 4079) {
if (Cs < 2664) {
if (ssDCval < 74) {
if (negBalance < 38) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (Cs < 1971) {
return 1;
}
else if (Cs >= 1971) {
return 4;
}
}
}
else if (negBalance >= 38) {
if (ssDCval < 58) {
if (Cs < 2297) {
return 4;
}
else if (Cs >= 2297) {
if (Rs < 3350) {
return 8;
}
else if (Rs >= 3350) {
return 4;
}
}
}
else if (ssDCval >= 58) {
if (refDCval < 72) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
else if (refDCval >= 72) {
if (Rs < 2174) {
return 1;
}
else if (Rs >= 2174) {
return 4;
}
}
}
}
}
else if (ssDCval >= 74) {
if (refDCval < 89) {
if (mcTcor < 94) {
if (TSC < 10) {
if (RsCsDiff < 147) {
if (diffAFD < -1) {
if (Rs < 2545) {
return 8;
}
else if (Rs >= 2545) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (RsCsDiff >= 147) {
if (Rs < 3605) {
return 4;
}
else if (Rs >= 3605) {
return 8;
}
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (mcTcor >= 94) {
if (Cs < 2304) {
if (Rs < 3139) {
return 1;
}
else if (Rs >= 3139) {
return 2;
}
}
else if (Cs >= 2304) {
return 4;
}
}
}
else if (refDCval >= 89) {
if (Rs < 3644) {
return 8;
}
else if (Rs >= 3644) {
if (tcor < 91) {
return 8;
}
else if (tcor >= 91) {
return 2;
}
}
}
}
}
else if (Cs >= 2664) {
if (negBalance < 52) {
if (TSC < 8) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
if (SC < 6402) {
if (posBalance < 55) {
if (Rs < 1159) {
return 8;
}
else if (Rs >= 1159) {
return 1;
}
}
else if (posBalance >= 55) {
return 4;
}
}
else if (SC >= 6402) {
if (Cs < 3925) {
if (CsDiff < 8) {
return 4;
}
else if (CsDiff >= 8) {
return 8;
}
}
else if (Cs >= 3925) {
return 4;
}
}
}
}
else if (TSC >= 8) {
if (TSC < 10) {
if (diffMVdiffVal < 1) {
return 4;
}
else if (diffMVdiffVal >= 1) {
if (negBalance < 44) {
if (Rs < 3328) {
if (Cs < 2900) {
return 8;
}
else if (Cs >= 2900) {
return 4;
}
}
else if (Rs >= 3328) {
return 1;
}
}
else if (negBalance >= 44) {
return 4;
}
}
}
else if (TSC >= 10) {
if (Rs < 3284) {
if (Rs < 2624) {
return 4;
}
else if (Rs >= 2624) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 4;
}
}
}
else if (Rs >= 3284) {
return 4;
}
}
}
}
else if (negBalance >= 52) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
else if (Rs >= 4079) {
if (tcor < 91) {
if (Rs < 5048) {
if (MVDiff < 7) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
if (mu_mv_mag_sq < 99) {
return 4;
}
else if (mu_mv_mag_sq >= 99) {
if (Cs < 3832) {
return 1;
}
else if (Cs >= 3832) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (Rs >= 5048) {
if (ssDCval < 74) {
if (SC < 9079) {
return 4;
}
else if (SC >= 9079) {
if (Rs < 7226) {
return 8;
}
else if (Rs >= 7226) {
return 4;
}
}
}
else if (ssDCval >= 74) {
return 2;
}
}
}
else if (tcor >= 91) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
if (Rs < 4226) {
if (Cs < 2187) {
return 2;
}
else if (Cs >= 2187) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Rs < 4122) {
return 4;
}
else if (Rs >= 4122) {
return 8;
}
}
}
}
else if (Rs >= 4226) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 100) {
if (negBalance < 46) {
if (posBalance < 39) {
if (ssDCval < 118) {
if (SC < 9032) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (diffRsCsdiff < -287) {
return 8;
}
else if (diffRsCsdiff >= -287) {
return 2;
}
}
}
else if (SC >= 9032) {
if (Rs < 5440) {
return 8;
}
else if (Rs >= 5440) {
return 1;
}
}
}
else if (ssDCval >= 118) {
if (ssDCval < 133) {
if (Cs < 3859) {
return 4;
}
else if (Cs >= 3859) {
return 1;
}
}
else if (ssDCval >= 133) {
return 8;
}
}
}
else if (posBalance >= 39) {
if (Rs < 4361) {
if (MVDiff < 30) {
if (ssDCval < 106) {
if (mu_mv_mag_sq < 73) {
return 8;
}
else if (mu_mv_mag_sq >= 73) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (diffRsCsdiff < -10) {
return 2;
}
else if (diffRsCsdiff >= -10) {
if (Cs < 2873) {
return 8;
}
else if (Cs >= 2873) {
return 2;
}
}
}
}
}
else if (ssDCval >= 106) {
if (Cs < 3139) {
return 8;
}
else if (Cs >= 3139) {
if (posBalance < 51) {
if (Rs < 3488) {
return 4;
}
else if (Rs >= 3488) {
return 2;
}
}
else if (posBalance >= 51) {
if (CsDiff < 8) {
if (Rs < 3173) {
return 8;
}
else if (Rs >= 3173) {
return 4;
}
}
else if (CsDiff >= 8) {
return 8;
}
}
}
}
}
else if (MVDiff >= 30) {
if (RsDiff < 10) {
if (Rs < 3204) {
return 8;
}
else if (Rs >= 3204) {
return 1;
}
}
else if (RsDiff >= 10) {
return 2;
}
}
}
else if (Rs >= 4361) {
if (Cs < 3703) {
if (AFD < 11) {
if (posBalance < 55) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Cs < 2971) {
return 2;
}
else if (Cs >= 2971) {
if (Rs < 8038) {
return 4;
}
else if (Rs >= 8038) {
return 1;
}
}
}
}
else if (posBalance >= 55) {
if (RsDiff < 11) {
return 1;
}
else if (RsDiff >= 11) {
if (Cs < 2661) {
return 2;
}
else if (Cs >= 2661) {
if (Rs < 7000) {
return 8;
}
else if (Rs >= 7000) {
return 1;
}
}
}
}
}
else if (AFD >= 11) {
if (Rs < 5775) {
return 8;
}
else if (Rs >= 5775) {
if (Cs < 2703) {
return 2;
}
else if (Cs >= 2703) {
if (Rs < 7917) {
return 4;
}
else if (Rs >= 7917) {
return 1;
}
}
}
}
}
else if (Cs >= 3703) {
if (Rs < 8517) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (Rs >= 8517) {
return 1;
}
}
}
}
}
else if (negBalance >= 46) {
if (tcor < 93) {
if (mu_mv_mag_sq < 83) {
if (MVDiff < 4) {
if (Cs < 2759) {
return 1;
}
else if (Cs >= 2759) {
return 8;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (mu_mv_mag_sq >= 83) {
if (Rs < 3740) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
return 8;
}
}
else if (Rs >= 3740) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
return 1;
}
}
}
}
else if (tcor >= 93) {
if (ssDCval < 103) {
if (Rs < 4034) {
return 4;
}
else if (Rs >= 4034) {
return 2;
}
}
else if (ssDCval >= 103) {
if (mcTcor < 94) {
if (Rs < 4036) {
return 8;
}
else if (Rs >= 4036) {
return 1;
}
}
else if (mcTcor >= 94) {
return 8;
}
}
}
}
}
}
else if (AFD >= 12) {
if (mu_mv_mag_sq < 78) {
if (RsDiff < 17) {
if (refDCval < 101) {
if (tcor < 64) {
if (ssDCval < 63) {
if (RsCsDiff < 267) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
if (TSC < 14) {
return 2;
}
else if (TSC >= 14) {
return 4;
}
}
}
else if (RsCsDiff >= 267) {
if (tcor < 55) {
return 4;
}
else if (tcor >= 55) {
if (posBalance < 57) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 4;
}
}
else if (posBalance >= 57) {
if (TSC < 13) {
return 4;
}
else if (TSC >= 13) {
return 2;
}
}
}
}
}
else if (ssDCval >= 63) {
if (AFD < 31) {
if (negBalance < 57) {
if (posBalance < 64) {
if (ssDCval < 74) {
if (Cs < 2538) {
return 8;
}
else if (Cs >= 2538) {
return 1;
}
}
else if (ssDCval >= 74) {
if (Rs < 1533) {
if (Rs < 1132) {
return 8;
}
else if (Rs >= 1132) {
if (Rs < 1461) {
return 4;
}
else if (Rs >= 1461) {
return 2;
}
}
}
else if (Rs >= 1533) {
return 8;
}
}
}
else if (posBalance >= 64) {
if (mu_mv_mag_sq < 15) {
if (Rs < 1264) {
return 1;
}
else if (Rs >= 1264) {
if (ssDCval < 92) {
return 4;
}
else if (ssDCval >= 92) {
if (Rs < 2028) {
return 8;
}
else if (Rs >= 2028) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (Rs < 1487) {
return 1;
}
else if (Rs >= 1487) {
return 8;
}
}
}
}
else if (negBalance >= 57) {
if (tcor < 46) {
return 1;
}
else if (tcor >= 46) {
if (posBalance < 56) {
if (Rs < 3394) {
return 1;
}
else if (Rs >= 3394) {
return 8;
}
}
else if (posBalance >= 56) {
if (Rs < 3133) {
if (Cs < 1971) {
return 1;
}
else if (Cs >= 1971) {
return 4;
}
}
else if (Rs >= 3133) {
return 1;
}
}
}
}
}
else if (AFD >= 31) {
if (CsDiff < 19) {
if (Cs < 2712) {
if (Rs < 1598) {
return 4;
}
else if (Rs >= 1598) {
return 1;
}
}
else if (Cs >= 2712) {
if (Rs < 3082) {
return 2;
}
else if (Rs >= 3082) {
return 8;
}
}
}
else if (CsDiff >= 19) {
return 8;
}
}
}
}
else if (tcor >= 64) {
if (negBalance < 47) {
if (posBalance < 45) {
if (ssDCval < 73) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 2;
}
}
else if (ssDCval >= 73) {
if (diffAFD < 1) {
if (Rs < 5559) {
return 2;
}
else if (Rs >= 5559) {
return 1;
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
else if (posBalance >= 45) {
if (RsCsDiff < 167) {
if (Rs < 1827) {
if (TSC < 8) {
if (Rs < 1087) {
return 2;
}
else if (Rs >= 1087) {
return 8;
}
}
else if (TSC >= 8) {
return 4;
}
}
else if (Rs >= 1827) {
if (AFD < 18) {
if (Rs < 2736) {
return 8;
}
else if (Rs >= 2736) {
if (negBalance < 41) {
if (AFD < 15) {
return 1;
}
else if (AFD >= 15) {
if (Rs < 3220) {
return 2;
}
else if (Rs >= 3220) {
return 8;
}
}
}
else if (negBalance >= 41) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
if (SC < 5317) {
return 4;
}
else if (SC >= 5317) {
return 8;
}
}
}
}
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (RsCsDiff >= 167) {
if (mu_mv_mag_sq < 57) {
if (Rs < 2995) {
if (posBalance < 54) {
if (Rs < 2834) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 4;
}
}
else if (Rs >= 2834) {
return 2;
}
}
else if (posBalance >= 54) {
if (refDCval < 75) {
if (diffMVdiffVal < -12) {
if (Rs < 1825) {
return 1;
}
else if (Rs >= 1825) {
return 4;
}
}
else if (diffMVdiffVal >= -12) {
if (SC < 3247) {
return 8;
}
else if (SC >= 3247) {
return 2;
}
}
}
else if (refDCval >= 75) {
if (negBalance < 41) {
if (diffTSC < 0) {
return 2;
}
else if (diffTSC >= 0) {
return 8;
}
}
else if (negBalance >= 41) {
if (TSC < 12) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
return 4;
}
}
else if (TSC >= 12) {
return 8;
}
}
}
}
}
else if (Rs >= 2995) {
if (posBalance < 53) {
if (ssDCval < 90) {
if (MVDiff < 8) {
if (Rs < 4211) {
return 1;
}
else if (Rs >= 4211) {
return 4;
}
}
else if (MVDiff >= 8) {
return 2;
}
}
else if (ssDCval >= 90) {
if (Rs < 4535) {
return 8;
}
else if (Rs >= 4535) {
return 1;
}
}
}
else if (posBalance >= 53) {
if (MVDiff < 17) {
if (posBalance < 63) {
if (MVDiff < 8) {
if (Cs < 2589) {
return 2;
}
else if (Cs >= 2589) {
return 4;
}
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (posBalance >= 63) {
return 2;
}
}
else if (MVDiff >= 17) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (Cs < 2286) {
if (Rs < 4109) {
return 8;
}
else if (Rs >= 4109) {
if (Rs < 4428) {
return 4;
}
else if (Rs >= 4428) {
return 2;
}
}
}
else if (Cs >= 2286) {
if (SC < 7140) {
if (MVDiff < 34) {
return 4;
}
else if (MVDiff >= 34) {
return 8;
}
}
else if (SC >= 7140) {
if (MVDiff < 16) {
return 2;
}
else if (MVDiff >= 16) {
if (Cs < 2989) {
return 4;
}
else if (Cs >= 2989) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 47) {
if (SC < 8280) {
if (refDCval < 91) {
if (TSC < 11) {
if (Cs < 2708) {
if (SC < 3884) {
if (AFD < 17) {
if (Rs < 1503) {
return 4;
}
else if (Rs >= 1503) {
return 1;
}
}
else if (AFD >= 17) {
if (MVDiff < 27) {
return 8;
}
else if (MVDiff >= 27) {
return 2;
}
}
}
else if (SC >= 3884) {
if (RsDiff < 11) {
if (MVDiff < 6) {
if (CsDiff < 9) {
if (Cs < 2686) {
return 8;
}
else if (Cs >= 2686) {
return 2;
}
}
else if (CsDiff >= 9) {
if (ssDCval < 71) {
if (Rs < 2375) {
return 1;
}
else if (Rs >= 2375) {
return 8;
}
}
else if (ssDCval >= 71) {
return 2;
}
}
}
else if (MVDiff >= 6) {
if (tcor < 88) {
if (mu_mv_mag_sq < 56) {
if (mu_mv_mag_sq < 3) {
if (diffAFD < 1) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (Rs < 3755) {
return 4;
}
else if (Rs >= 3755) {
return 8;
}
}
}
else if (diffAFD >= 1) {
if (Rs < 2674) {
return 4;
}
else if (Rs >= 2674) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 56) {
if (mcTcor < 88) {
return 8;
}
else if (mcTcor >= 88) {
return 4;
}
}
}
else if (tcor >= 88) {
if (Rs < 2694) {
return 8;
}
else if (Rs >= 2694) {
return 2;
}
}
}
}
else if (RsDiff >= 11) {
if (mu_mv_mag_sq < 10) {
if (Rs < 3733) {
if (MVDiff < 13) {
if (SC < 4787) {
if (Cs < 1961) {
return 4;
}
else if (Cs >= 1961) {
return 2;
}
}
else if (SC >= 4787) {
return 4;
}
}
else if (MVDiff >= 13) {
return 2;
}
}
else if (Rs >= 3733) {
return 2;
}
}
else if (mu_mv_mag_sq >= 10) {
if (negBalance < 54) {
if (RsCsDiff < 260) {
return 4;
}
else if (RsCsDiff >= 260) {
if (RsCsDiff < 316) {
if (Rs < 3650) {
return 1;
}
else if (Rs >= 3650) {
return 2;
}
}
else if (RsCsDiff >= 316) {
return 8;
}
}
}
else if (negBalance >= 54) {
return 8;
}
}
}
}
}
else if (Cs >= 2708) {
if (RsDiff < 11) {
if (AFD < 16) {
if (Rs < 2959) {
if (ssDCval < 80) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (CsDiff < 12) {
return 8;
}
else if (CsDiff >= 12) {
return 4;
}
}
}
else if (ssDCval >= 80) {
return 4;
}
}
else if (Rs >= 2959) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
if (AFD < 14) {
return 4;
}
else if (AFD >= 14) {
return 2;
}
}
}
}
else if (AFD >= 16) {
if (ssDCval < 86) {
if (MVDiff < 4) {
return 1;
}
else if (MVDiff >= 4) {
if (diffAFD < 2) {
if (MVDiff < 56) {
if (Cs < 3111) {
return 1;
}
else if (Cs >= 3111) {
return 2;
}
}
else if (MVDiff >= 56) {
return 1;
}
}
else if (diffAFD >= 2) {
return 8;
}
}
}
else if (ssDCval >= 86) {
return 8;
}
}
}
else if (RsDiff >= 11) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
if (Rs < 4349) {
return 4;
}
else if (Rs >= 4349) {
return 8;
}
}
}
}
}
else if (TSC >= 11) {
if (CsDiff < 10) {
if (RsDiff < 12) {
if (SC < 4142) {
if (Rs < 1924) {
return 4;
}
else if (Rs >= 1924) {
return 8;
}
}
else if (SC >= 4142) {
if (refDCval < 79) {
return 8;
}
else if (refDCval >= 79) {
if (Rs < 1993) {
return 4;
}
else if (Rs >= 1993) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
if (Rs < 2927) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 4;
}
}
else if (Rs >= 2927) {
if (Cs < 2266) {
return 4;
}
else if (Cs >= 2266) {
if (Rs < 4070) {
return 2;
}
else if (Rs >= 4070) {
return 8;
}
}
}
}
}
}
}
}
}
else if (RsDiff >= 12) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
return 4;
}
}
}
else if (CsDiff >= 10) {
if (Cs < 3172) {
if (Rs < 2226) {
if (mcTcor < 78) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
return 4;
}
}
else if (mcTcor >= 78) {
if (negBalance < 53) {
if (Rs < 1566) {
return 2;
}
else if (Rs >= 1566) {
return 1;
}
}
else if (negBalance >= 53) {
if (RsDiff < 9) {
return 8;
}
else if (RsDiff >= 9) {
if (SC < 3930) {
return 8;
}
else if (SC >= 3930) {
return 4;
}
}
}
}
}
else if (Rs >= 2226) {
if (mu_mv_mag_sq < 44) {
if (TSC < 15) {
if (diffAFD < -1) {
if (posBalance < 62) {
if (posBalance < 58) {
if (Rs < 2730) {
return 1;
}
else if (Rs >= 2730) {
return 4;
}
}
else if (posBalance >= 58) {
if (Rs < 3608) {
if (Rs < 2344) {
return 4;
}
else if (Rs >= 2344) {
return 8;
}
}
else if (Rs >= 3608) {
return 1;
}
}
}
else if (posBalance >= 62) {
if (CsDiff < 12) {
if (Rs < 2898) {
return 1;
}
else if (Rs >= 2898) {
return 8;
}
}
else if (CsDiff >= 12) {
return 2;
}
}
}
else if (diffAFD >= -1) {
if (ssDCval < 82) {
if (Rs < 3386) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
if (Cs < 2396) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (Cs >= 2396) {
if (Rs < 2679) {
return 1;
}
else if (Rs >= 2679) {
return 8;
}
}
}
}
else if (Rs >= 3386) {
if (Cs < 2869) {
return 1;
}
else if (Cs >= 2869) {
return 8;
}
}
}
else if (ssDCval >= 82) {
if (negBalance < 52) {
return 4;
}
else if (negBalance >= 52) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
if (Cs < 2340) {
return 8;
}
else if (Cs >= 2340) {
return 4;
}
}
}
}
}
}
else if (TSC >= 15) {
if (ssDCval < 86) {
if (Rs < 3443) {
return 8;
}
else if (Rs >= 3443) {
return 4;
}
}
else if (ssDCval >= 86) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (tcor < 83) {
if (refDCval < 73) {
if (Cs < 2578) {
if (Cs < 2414) {
return 8;
}
else if (Cs >= 2414) {
return 4;
}
}
else if (Cs >= 2578) {
return 8;
}
}
else if (refDCval >= 73) {
if (Cs < 2870) {
if (Rs < 2269) {
return 1;
}
else if (Rs >= 2269) {
return 4;
}
}
else if (Cs >= 2870) {
return 1;
}
}
}
else if (tcor >= 83) {
return 8;
}
}
}
}
else if (Cs >= 3172) {
if (tcor < 83) {
if (refDCval < 90) {
if (CsDiff < 11) {
if (Rs < 2890) {
return 4;
}
else if (Rs >= 2890) {
return 8;
}
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (refDCval >= 90) {
if (Rs < 3935) {
return 8;
}
else if (Rs >= 3935) {
return 4;
}
}
}
else if (tcor >= 83) {
if (Rs < 3766) {
return 8;
}
else if (Rs >= 3766) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 91) {
if (Cs < 2746) {
if (tcor < 78) {
if (AFD < 19) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 1;
}
}
else if (AFD >= 19) {
if (tcor < 75) {
if (mu_mv_mag_sq < 2) {
return 4;
}
else if (mu_mv_mag_sq >= 2) {
if (Cs < 2361) {
return 4;
}
else if (Cs >= 2361) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 4;
}
}
}
}
else if (tcor >= 75) {
if (diffRsCsdiff < -19) {
return 8;
}
else if (diffRsCsdiff >= -19) {
if (RsCsDiff < 140) {
return 8;
}
else if (RsCsDiff >= 140) {
if (Rs < 3134) {
return 1;
}
else if (Rs >= 3134) {
return 4;
}
}
}
}
}
}
else if (tcor >= 78) {
if (mu_mv_mag_sq < 61) {
if (TSC < 13) {
if (AFD < 15) {
if (RsCsDiff < 196) {
return 8;
}
else if (RsCsDiff >= 196) {
if (Rs < 2929) {
return 8;
}
else if (Rs >= 2929) {
return 2;
}
}
}
else if (AFD >= 15) {
if (Cs < 2154) {
if (AFD < 20) {
if (negBalance < 52) {
if (Rs < 2550) {
return 8;
}
else if (Rs >= 2550) {
return 2;
}
}
else if (negBalance >= 52) {
return 8;
}
}
else if (AFD >= 20) {
if (Rs < 2823) {
return 2;
}
else if (Rs >= 2823) {
if (Rs < 3388) {
return 4;
}
else if (Rs >= 3388) {
return 8;
}
}
}
}
else if (Cs >= 2154) {
if (AFD < 17) {
if (MVDiff < 13) {
return 4;
}
else if (MVDiff >= 13) {
return 8;
}
}
else if (AFD >= 17) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (posBalance < 58) {
return 2;
}
else if (posBalance >= 58) {
if (Rs < 2401) {
return 2;
}
else if (Rs >= 2401) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 13) {
return 8;
}
}
else if (mu_mv_mag_sq >= 61) {
if (SC < 6203) {
if (Rs < 3479) {
return 2;
}
else if (Rs >= 3479) {
return 8;
}
}
else if (SC >= 6203) {
return 4;
}
}
}
}
else if (Cs >= 2746) {
if (negBalance < 51) {
if (tcor < 84) {
if (RsDiff < 15) {
return 8;
}
else if (RsDiff >= 15) {
return 4;
}
}
else if (tcor >= 84) {
if (CsDiff < 9) {
return 8;
}
else if (CsDiff >= 9) {
return 4;
}
}
}
else if (negBalance >= 51) {
if (refDCval < 92) {
if (tcor < 78) {
if (Rs < 3345) {
if (Rs < 2701) {
return 4;
}
else if (Rs >= 2701) {
return 1;
}
}
else if (Rs >= 3345) {
return 4;
}
}
else if (tcor >= 78) {
if (Rs < 3073) {
return 4;
}
else if (Rs >= 3073) {
return 8;
}
}
}
else if (refDCval >= 92) {
if (TSC < 10) {
if (CsDiff < 11) {
return 4;
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (TSC >= 10) {
if (tcor < 66) {
return 8;
}
else if (tcor >= 66) {
if (diffAFD < -5) {
if (Rs < 2425) {
return 4;
}
else if (Rs >= 2425) {
return 8;
}
}
else if (diffAFD >= -5) {
return 4;
}
}
}
}
}
}
}
}
else if (SC >= 8280) {
if (diffRsCsdiff < -164) {
return 2;
}
else if (diffRsCsdiff >= -164) {
return 8;
}
}
}
}
}
else if (refDCval >= 101) {
if (AFD < 21) {
if (ssDCval < 119) {
if (mcTcor < 94) {
if (Rs < 3464) {
if (refDCval < 108) {
if (Rs < 2986) {
if (TSC < 10) {
return 8;
}
else if (TSC >= 10) {
if (MVDiff < 8) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
if (Rs < 2366) {
return 4;
}
else if (Rs >= 2366) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (mcTcor < 89) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 4;
}
}
else if (mcTcor >= 89) {
return 8;
}
}
}
}
else if (Rs >= 2986) {
if (posBalance < 54) {
if (Rs < 3383) {
return 8;
}
else if (Rs >= 3383) {
return 2;
}
}
else if (posBalance >= 54) {
return 4;
}
}
}
else if (refDCval >= 108) {
if (negBalance < 30) {
return 2;
}
else if (negBalance >= 30) {
return 8;
}
}
}
else if (Rs >= 3464) {
if (tcor < 86) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Cs < 2425) {
return 8;
}
else if (Cs >= 2425) {
return 2;
}
}
}
else if (tcor >= 86) {
if (RsCsDiff < 256) {
if (mcTcor < 90) {
if (Rs < 4653) {
return 1;
}
else if (Rs >= 4653) {
return 8;
}
}
else if (mcTcor >= 90) {
return 8;
}
}
else if (RsCsDiff >= 256) {
if (Rs < 5208) {
return 2;
}
else if (Rs >= 5208) {
return 4;
}
}
}
}
}
else if (mcTcor >= 94) {
if (SC < 7150) {
if (Rs < 1613) {
if (ssDCval < 108) {
if (Rs < 1501) {
return 2;
}
else if (Rs >= 1501) {
return 4;
}
}
else if (ssDCval >= 108) {
return 1;
}
}
else if (Rs >= 1613) {
if (RsDiff < 13) {
if (RsCsDiff < 232) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
return 2;
}
}
else if (RsCsDiff >= 232) {
if (refDCval < 110) {
return 2;
}
else if (refDCval >= 110) {
return 8;
}
}
}
else if (RsDiff >= 13) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
if (Rs < 4314) {
return 8;
}
else if (Rs >= 4314) {
return 1;
}
}
}
}
}
else if (SC >= 7150) {
return 4;
}
}
}
else if (ssDCval >= 119) {
if (mcTcor < 84) {
if (posBalance < 54) {
if (Rs < 4022) {
if (TSC < 10) {
return 1;
}
else if (TSC >= 10) {
if (Rs < 1457) {
return 1;
}
else if (Rs >= 1457) {
return 8;
}
}
}
else if (Rs >= 4022) {
return 1;
}
}
else if (posBalance >= 54) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (Cs < 3660) {
if (Rs < 4443) {
return 8;
}
else if (Rs >= 4443) {
return 4;
}
}
else if (Cs >= 3660) {
if (mu_mv_mag_sq < 19) {
return 8;
}
else if (mu_mv_mag_sq >= 19) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 84) {
if (RsCsDiff < 167) {
if (mu_mv_mag_sq < 30) {
if (refDCval < 136) {
if (MVDiff < 17) {
if (Rs < 3357) {
return 1;
}
else if (Rs >= 3357) {
return 8;
}
}
else if (MVDiff >= 17) {
return 8;
}
}
else if (refDCval >= 136) {
return 1;
}
}
else if (mu_mv_mag_sq >= 30) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
if (mcTcor < 92) {
return 8;
}
else if (mcTcor >= 92) {
if (ssDCval < 129) {
if (tcor < 85) {
if (Rs < 3775) {
return 1;
}
else if (Rs >= 3775) {
return 8;
}
}
else if (tcor >= 85) {
return 8;
}
}
else if (ssDCval >= 129) {
if (Rs < 3281) {
return 4;
}
else if (Rs >= 3281) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 167) {
if (CsDiff < 16) {
if (SC < 5947) {
if (posBalance < 53) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (negBalance < 34) {
if (Cs < 2180) {
return 2;
}
else if (Cs >= 2180) {
return 4;
}
}
else if (negBalance >= 34) {
if (Cs < 2062) {
if (Cs < 2022) {
return 8;
}
else if (Cs >= 2022) {
if (Rs < 2654) {
return 4;
}
else if (Rs >= 2654) {
return 2;
}
}
}
else if (Cs >= 2062) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
if (negBalance < 59) {
return 8;
}
else if (negBalance >= 59) {
return 1;
}
}
}
else if (SC >= 5947) {
if (MVDiff < 42) {
if (Rs < 6731) {
if (diffRsCsdiff < -4) {
if (Cs < 4378) {
return 4;
}
else if (Cs >= 4378) {
return 8;
}
}
else if (diffRsCsdiff >= -4) {
if (Rs < 4100) {
if (CsDiff < 14) {
if (Cs < 2222) {
return 4;
}
else if (Cs >= 2222) {
if (ssDCval < 131) {
return 4;
}
else if (ssDCval >= 131) {
return 8;
}
}
}
else if (CsDiff >= 14) {
return 4;
}
}
else if (Rs >= 4100) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
if (negBalance < 58) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
return 4;
}
}
else if (negBalance >= 58) {
return 8;
}
}
}
}
}
else if (Rs >= 6731) {
return 8;
}
}
else if (MVDiff >= 42) {
if (Rs < 3545) {
return 4;
}
else if (Rs >= 3545) {
return 8;
}
}
}
}
else if (CsDiff >= 16) {
if (refDCval < 131) {
if (Rs < 3033) {
return 4;
}
else if (Rs >= 3033) {
return 2;
}
}
else if (refDCval >= 131) {
if (gchDC < 8) {
return 8;
}
else if (gchDC >= 8) {
return 1;
}
}
}
}
}
}
}
else if (AFD >= 21) {
if (negBalance < 56) {
if (ssDCval < 109) {
if (Cs < 3178) {
if (diffRsCsdiff < -31) {
if (Rs < 3659) {
return 2;
}
else if (Rs >= 3659) {
return 1;
}
}
else if (diffRsCsdiff >= -31) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (Cs >= 3178) {
if (Rs < 3539) {
return 8;
}
else if (Rs >= 3539) {
return 2;
}
}
}
else if (ssDCval >= 109) {
if (Rs < 3448) {
if (RsCsDiff < 455) {
if (Rs < 2604) {
return 8;
}
else if (Rs >= 2604) {
if (SC < 5024) {
return 2;
}
else if (SC >= 5024) {
if (Cs < 3965) {
if (diffMVdiffVal < 30) {
return 8;
}
else if (diffMVdiffVal >= 30) {
return 2;
}
}
else if (Cs >= 3965) {
return 2;
}
}
}
}
else if (RsCsDiff >= 455) {
if (Cs < 11519) {
if (MVDiff < 16) {
return 4;
}
else if (MVDiff >= 16) {
return 2;
}
}
else if (Cs >= 11519) {
return 8;
}
}
}
else if (Rs >= 3448) {
if (Cs < 2188) {
return 4;
}
else if (Cs >= 2188) {
return 8;
}
}
}
}
else if (negBalance >= 56) {
if (RsCsDiff < 201) {
if (MVDiff < 27) {
return 2;
}
else if (MVDiff >= 27) {
return 8;
}
}
else if (RsCsDiff >= 201) {
if (Rs < 2820) {
if (mu_mv_mag_sq < 8) {
if (AFD < 25) {
if (RsCsDiff < 246) {
if (Rs < 1690) {
return 1;
}
else if (Rs >= 1690) {
return 2;
}
}
else if (RsCsDiff >= 246) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
else if (AFD >= 25) {
return 2;
}
}
else if (mu_mv_mag_sq >= 8) {
if (negBalance < 61) {
if (Rs < 2784) {
if (TSC < 11) {
return 1;
}
else if (TSC >= 11) {
return 8;
}
}
else if (Rs >= 2784) {
return 2;
}
}
else if (negBalance >= 61) {
return 1;
}
}
}
else if (Rs >= 2820) {
if (tcor < 77) {
if (AFD < 23) {
if (TSC < 12) {
if (Rs < 2992) {
return 1;
}
else if (Rs >= 2992) {
return 8;
}
}
else if (TSC >= 12) {
if (Cs < 2557) {
return 8;
}
else if (Cs >= 2557) {
return 4;
}
}
}
else if (AFD >= 23) {
if (posBalance < 58) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
if (CsDiff < 12) {
return 8;
}
else if (CsDiff >= 12) {
return 4;
}
}
}
else if (posBalance >= 58) {
return 1;
}
}
}
else if (tcor >= 77) {
return 8;
}
}
}
}
}
}
}
else if (RsDiff >= 17) {
if (TSC < 8) {
if (ssDCval < 127) {
if (Rs < 8171) {
if (Cs < 2632) {
return 4;
}
else if (Cs >= 2632) {
return 8;
}
}
else if (Rs >= 8171) {
return 1;
}
}
else if (ssDCval >= 127) {
if (Rs < 3384) {
return 2;
}
else if (Rs >= 3384) {
return 8;
}
}
}
else if (TSC >= 8) {
if (ssDCval < 139) {
if (SC < 10482) {
if (TSC < 15) {
if (mcTcor < 91) {
if (negBalance < 54) {
if (ssDCval < 71) {
if (mcTcor < 86) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
return 8;
}
}
else if (mcTcor >= 86) {
if (gchDC < 2) {
if (Rs < 4961) {
return 2;
}
else if (Rs >= 4961) {
return 4;
}
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (ssDCval >= 71) {
if (SC < 6044) {
if (Rs < 2791) {
return 2;
}
else if (Rs >= 2791) {
return 8;
}
}
else if (SC >= 6044) {
if (posBalance < 46) {
if (Rs < 4541) {
return 2;
}
else if (Rs >= 4541) {
return 4;
}
}
else if (posBalance >= 46) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (tcor < 78) {
if (diffAFD < 2) {
if (diffAFD < -3) {
return 8;
}
else if (diffAFD >= -3) {
if (diffRsCsdiff < -149) {
if (Rs < 5386) {
return 8;
}
else if (Rs >= 5386) {
return 4;
}
}
else if (diffRsCsdiff >= -149) {
return 4;
}
}
}
else if (diffAFD >= 2) {
if (MVDiff < 17) {
return 4;
}
else if (MVDiff >= 17) {
return 8;
}
}
}
else if (tcor >= 78) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 54) {
if (refDCval < 75) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
if (MVDiff < 8) {
return 8;
}
else if (MVDiff >= 8) {
if (Cs < 4083) {
return 2;
}
else if (Cs >= 4083) {
return 4;
}
}
}
}
else if (refDCval >= 75) {
if (SC < 7358) {
if (RsCsDiff < 446) {
return 4;
}
else if (RsCsDiff >= 446) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
if (negBalance < 61) {
return 1;
}
else if (negBalance >= 61) {
if (Rs < 3073) {
return 2;
}
else if (Rs >= 3073) {
return 8;
}
}
}
}
}
else if (SC >= 7358) {
if (diffAFD < 3) {
return 4;
}
else if (diffAFD >= 3) {
return 2;
}
}
}
}
}
else if (mcTcor >= 91) {
if (MVDiff < 23) {
if (tcor < 60) {
return 4;
}
else if (tcor >= 60) {
if (SC < 10170) {
if (negBalance < 39) {
if (RsDiff < 20) {
return 8;
}
else if (RsDiff >= 20) {
return 2;
}
}
else if (negBalance >= 39) {
if (negBalance < 50) {
if (refDCval < 75) {
return 4;
}
else if (refDCval >= 75) {
return 2;
}
}
else if (negBalance >= 50) {
if (tcor < 79) {
if (Rs < 5273) {
return 2;
}
else if (Rs >= 5273) {
return 4;
}
}
else if (tcor >= 79) {
if (MVDiff < 15) {
if (CsDiff < 11) {
if (Rs < 3758) {
return 8;
}
else if (Rs >= 3758) {
return 4;
}
}
else if (CsDiff >= 11) {
return 8;
}
}
else if (MVDiff >= 15) {
return 2;
}
}
}
}
}
else if (SC >= 10170) {
return 4;
}
}
}
else if (MVDiff >= 23) {
if (AFD < 22) {
return 8;
}
else if (AFD >= 22) {
if (SC < 6498) {
return 1;
}
else if (SC >= 6498) {
return 2;
}
}
}
}
}
else if (TSC >= 15) {
if (posBalance < 59) {
if (negBalance < 46) {
if (ssDCval < 83) {
if (Rs < 6964) {
return 2;
}
else if (Rs >= 6964) {
return 1;
}
}
else if (ssDCval >= 83) {
if (AFD < 40) {
return 1;
}
else if (AFD >= 40) {
return 8;
}
}
}
else if (negBalance >= 46) {
if (negBalance < 56) {
if (TSC < 17) {
if (Cs < 3649) {
if (AFD < 27) {
return 4;
}
else if (AFD >= 27) {
return 2;
}
}
else if (Cs >= 3649) {
if (Rs < 5715) {
return 2;
}
else if (Rs >= 5715) {
return 1;
}
}
}
else if (TSC >= 17) {
return 2;
}
}
else if (negBalance >= 56) {
if (RsCsDiff < 1091) {
if (refDCval < 131) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
if (Rs < 4180) {
return 8;
}
else if (Rs >= 4180) {
if (Rs < 4879) {
return 2;
}
else if (Rs >= 4879) {
return 8;
}
}
}
}
else if (refDCval >= 131) {
if (Rs < 4821) {
return 2;
}
else if (Rs >= 4821) {
return 8;
}
}
}
else if (RsCsDiff >= 1091) {
if (Rs < 4293) {
return 1;
}
else if (Rs >= 4293) {
return 4;
}
}
}
}
}
else if (posBalance >= 59) {
if (SC < 7705) {
if (AFD < 31) {
if (SC < 6636) {
if (Cs < 2930) {
return 8;
}
else if (Cs >= 2930) {
return 1;
}
}
else if (SC >= 6636) {
if (RsDiff < 18) {
if (Rs < 4025) {
return 8;
}
else if (Rs >= 4025) {
return 1;
}
}
else if (RsDiff >= 18) {
return 4;
}
}
}
else if (AFD >= 31) {
if (Rs < 3773) {
return 4;
}
else if (Rs >= 3773) {
return 1;
}
}
}
else if (SC >= 7705) {
if (negBalance < 54) {
if (Rs < 5230) {
if (ssDCval < 108) {
return 8;
}
else if (ssDCval >= 108) {
return 4;
}
}
else if (Rs >= 5230) {
if (CsDiff < 17) {
return 2;
}
else if (CsDiff >= 17) {
return 4;
}
}
}
else if (negBalance >= 54) {
return 8;
}
}
}
}
}
else if (SC >= 10482) {
if (ssDCval < 111) {
if (ssDCval < 86) {
return 8;
}
else if (ssDCval >= 86) {
if (RsDiff < 19) {
if (TSC < 15) {
if (refDCval < 92) {
return 4;
}
else if (refDCval >= 92) {
return 8;
}
}
else if (TSC >= 15) {
return 4;
}
}
else if (RsDiff >= 19) {
return 4;
}
}
}
else if (ssDCval >= 111) {
if (Rs < 9699) {
return 8;
}
else if (Rs >= 9699) {
return 1;
}
}
}
}
else if (ssDCval >= 139) {
if (refDCval < 166) {
if (mcTcor < 91) {
if (gchDC < 3) {
if (TSC < 10) {
if (Rs < 7292) {
return 4;
}
else if (Rs >= 7292) {
return 8;
}
}
else if (TSC >= 10) {
if (diffRsCsdiff < 491) {
if (CsDiff < 41) {
if (Rs < 5019) {
if (Rs < 4951) {
return 4;
}
else if (Rs >= 4951) {
return 8;
}
}
else if (Rs >= 5019) {
return 4;
}
}
else if (CsDiff >= 41) {
return 8;
}
}
else if (diffRsCsdiff >= 491) {
if (Rs < 8714) {
return 2;
}
else if (Rs >= 8714) {
return 4;
}
}
}
}
else if (gchDC >= 3) {
if (MVDiff < 21) {
return 4;
}
else if (MVDiff >= 21) {
return 8;
}
}
}
else if (mcTcor >= 91) {
if (ssDCval < 153) {
if (Rs < 6452) {
return 2;
}
else if (Rs >= 6452) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
return 8;
}
}
}
else if (ssDCval >= 153) {
if (refDCval < 163) {
return 4;
}
else if (refDCval >= 163) {
if (Rs < 4420) {
return 8;
}
else if (Rs >= 4420) {
return 2;
}
}
}
}
}
else if (refDCval >= 166) {
if (Cs < 3720) {
return 2;
}
else if (Cs >= 3720) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 78) {
if (SC < 7050) {
if (refDCval < 90) {
if (Rs < 2687) {
if (mu_mv_mag_sq < 84) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 8;
}
}
else if (mu_mv_mag_sq >= 84) {
if (Cs < 2000) {
if (Rs < 2102) {
return 4;
}
else if (Rs >= 2102) {
return 8;
}
}
else if (Cs >= 2000) {
return 4;
}
}
}
else if (Rs >= 2687) {
if (diffMVdiffVal < -5) {
if (Rs < 3695) {
return 4;
}
else if (Rs >= 3695) {
return 1;
}
}
else if (diffMVdiffVal >= -5) {
if (TSC < 10) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
return 4;
}
}
else if (TSC >= 10) {
if (Cs < 3625) {
if (mu_mv_mag_sq < 80) {
if (Cs < 3001) {
return 8;
}
else if (Cs >= 3001) {
return 1;
}
}
else if (mu_mv_mag_sq >= 80) {
return 8;
}
}
else if (Cs >= 3625) {
if (Rs < 2969) {
return 4;
}
else if (Rs >= 2969) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 90) {
if (Cs < 2676) {
if (posBalance < 57) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (Cs < 2619) {
return 4;
}
else if (Cs >= 2619) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (TSC < 11) {
return 8;
}
else if (TSC >= 11) {
if (Rs < 2326) {
return 4;
}
else if (Rs >= 2326) {
return 1;
}
}
}
}
else if (Cs >= 2676) {
if (negBalance < 46) {
if (Rs < 3623) {
return 8;
}
else if (Rs >= 3623) {
return 4;
}
}
else if (negBalance >= 46) {
return 4;
}
}
}
}
else if (SC >= 7050) {
if (MVDiff < 15) {
if (negBalance < 63) {
if (negBalance < 51) {
if (Rs < 5464) {
if (diffRsCsdiff < -79) {
if (refDCval < 123) {
if (Rs < 5352) {
return 8;
}
else if (Rs >= 5352) {
return 4;
}
}
else if (refDCval >= 123) {
return 4;
}
}
else if (diffRsCsdiff >= -79) {
if (RsCsDiff < 201) {
if (Cs < 3539) {
return 8;
}
else if (Cs >= 3539) {
if (Rs < 3512) {
return 8;
}
else if (Rs >= 3512) {
return 4;
}
}
}
else if (RsCsDiff >= 201) {
return 8;
}
}
}
else if (Rs >= 5464) {
if (ssDCval < 97) {
if (AFD < 18) {
return 4;
}
else if (AFD >= 18) {
return 2;
}
}
else if (ssDCval >= 97) {
if (Cs < 7237) {
return 8;
}
else if (Cs >= 7237) {
return 4;
}
}
}
}
else if (negBalance >= 51) {
if (Rs < 3188) {
if (Rs < 3090) {
return 8;
}
else if (Rs >= 3090) {
return 4;
}
}
else if (Rs >= 3188) {
if (ssDCval < 122) {
if (Rs < 7651) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (Rs < 3479) {
if (Rs < 3450) {
return 8;
}
else if (Rs >= 3450) {
return 4;
}
}
else if (Rs >= 3479) {
if (negBalance < 52) {
if (Rs < 5558) {
return 8;
}
else if (Rs >= 5558) {
return 4;
}
}
else if (negBalance >= 52) {
return 8;
}
}
}
}
else if (Rs >= 7651) {
return 4;
}
}
else if (ssDCval >= 122) {
return 4;
}
}
}
}
else if (negBalance >= 63) {
return 4;
}
}
else if (MVDiff >= 15) {
if (negBalance < 31) {
return 8;
}
else if (negBalance >= 31) {
if (diffAFD < 2) {
if (ssDCval < 118) {
if (ssDCval < 99) {
if (TSC < 15) {
return 4;
}
else if (TSC >= 15) {
return 2;
}
}
else if (ssDCval >= 99) {
return 8;
}
}
else if (ssDCval >= 118) {
return 4;
}
}
else if (diffAFD >= 2) {
if (Rs < 4575) {
return 1;
}
else if (Rs >= 4575) {
return 8;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1641) {
if (Rs < 3139) {
if (Rs < 1470) {
if (ssDCval < 122) {
if (negBalance < 2) {
if (Rs < 1194) {
return 2;
}
else if (Rs >= 1194) {
return 8;
}
}
else if (negBalance >= 2) {
if (refDCval < 63) {
if (AFD < 3) {
if (refDCval < 53) {
if (Cs < 1445) {
if (posBalance < 37) {
return 8;
}
else if (posBalance >= 37) {
if (Rs < 1168) {
if (Rs < 691) {
return 1;
}
else if (Rs >= 691) {
return 8;
}
}
else if (Rs >= 1168) {
return 4;
}
}
}
else if (Cs >= 1445) {
if (Rs < 1434) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 8;
}
}
else if (Rs >= 1434) {
if (Rs < 1444) {
return 4;
}
else if (Rs >= 1444) {
return 1;
}
}
}
}
else if (refDCval >= 53) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
if (Rs < 753) {
return 1;
}
else if (Rs >= 753) {
return 8;
}
}
}
}
else if (AFD >= 3) {
if (SC < 2640) {
return 8;
}
else if (SC >= 2640) {
if (mcTcor < 98) {
if (diffAFD < -1) {
if (Cs < 1525) {
return 8;
}
else if (Cs >= 1525) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (mcTcor >= 98) {
return 4;
}
}
}
}
else if (refDCval >= 63) {
if (ssDCval < 77) {
if (RsDiff < 3) {
if (Cs < 1431) {
if (AFD < 2) {
if (Cs < 1339) {
return 1;
}
else if (Cs >= 1339) {
return 4;
}
}
else if (AFD >= 2) {
if (Cs < 1426) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (Cs >= 1426) {
return 2;
}
}
}
else if (Cs >= 1431) {
return 8;
}
}
else if (RsDiff >= 3) {
if (Cs < 1330) {
return 1;
}
else if (Cs >= 1330) {
return 4;
}
}
}
else if (ssDCval >= 77) {
if (Rs < 1057) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Rs >= 1057) {
return 8;
}
}
}
}
}
else if (ssDCval >= 122) {
if (AFD < 2) {
if (posBalance < 28) {
if (ssDCval < 132) {
return 8;
}
else if (ssDCval >= 132) {
if (Rs < 1327) {
return 1;
}
else if (Rs >= 1327) {
return 8;
}
}
}
else if (posBalance >= 28) {
if (Rs < 1383) {
return 4;
}
else if (Rs >= 1383) {
return 8;
}
}
}
else if (AFD >= 2) {
if (negBalance < 4) {
return 1;
}
else if (negBalance >= 4) {
if (ssDCval < 129) {
if (Rs < 1239) {
if (Rs < 933) {
return 8;
}
else if (Rs >= 933) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
if (mu_mv_mag_sq < 114) {
if (Rs < 1088) {
return 4;
}
else if (Rs >= 1088) {
return 8;
}
}
else if (mu_mv_mag_sq >= 114) {
return 4;
}
}
}
}
else if (Rs >= 1239) {
return 8;
}
}
else if (ssDCval >= 129) {
return 8;
}
}
}
}
}
else if (Rs >= 1470) {
if (ssDCval < 86) {
if (mcTcor < 98) {
if (posBalance < 46) {
if (negBalance < 15) {
return 8;
}
else if (negBalance >= 15) {
if (Rs < 2226) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 4;
}
}
else if (Rs >= 2226) {
if (Cs < 1402) {
return 4;
}
else if (Cs >= 1402) {
if (Rs < 3064) {
if (Rs < 2497) {
return 8;
}
else if (Rs >= 2497) {
if (Rs < 2677) {
return 4;
}
else if (Rs >= 2677) {
if (Rs < 2918) {
return 8;
}
else if (Rs >= 2918) {
return 4;
}
}
}
}
else if (Rs >= 3064) {
return 2;
}
}
}
}
}
else if (posBalance >= 46) {
if (negBalance < 22) {
if (Cs < 1475) {
return 4;
}
else if (Cs >= 1475) {
return 8;
}
}
else if (negBalance >= 22) {
return 8;
}
}
}
else if (mcTcor >= 98) {
if (negBalance < 19) {
if (SC < 4464) {
if (MVDiff < 2) {
if (Cs < 1344) {
if (Rs < 2287) {
return 8;
}
else if (Rs >= 2287) {
if (Cs < 1342) {
if (mu_mv_mag_sq < 123) {
return 2;
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 2289) {
return 2;
}
else if (Rs >= 2289) {
if (Cs < 1317) {
return 2;
}
else if (Cs >= 1317) {
if (Rs < 2302) {
if (Rs < 2301) {
if (Rs < 2296) {
return 8;
}
else if (Rs >= 2296) {
if (Rs < 2297) {
return 2;
}
else if (Rs >= 2297) {
return 8;
}
}
}
else if (Rs >= 2301) {
return 2;
}
}
else if (Rs >= 2302) {
return 8;
}
}
}
}
}
else if (Cs >= 1342) {
if (Rs < 2305) {
return 2;
}
else if (Rs >= 2305) {
return 8;
}
}
}
}
else if (Cs >= 1344) {
if (Rs < 1506) {
if (Cs < 1551) {
return 8;
}
else if (Cs >= 1551) {
return 1;
}
}
else if (Rs >= 1506) {
return 8;
}
}
}
else if (MVDiff >= 2) {
if (SC < 3319) {
return 1;
}
else if (SC >= 3319) {
return 8;
}
}
}
else if (SC >= 4464) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
return 1;
}
}
}
else if (negBalance >= 19) {
if (posBalance < 38) {
if (Rs < 1519) {
return 1;
}
else if (Rs >= 1519) {
if (RsDiff < 5) {
if (Rs < 2335) {
return 4;
}
else if (Rs >= 2335) {
return 4;
}
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (posBalance >= 38) {
if (Cs < 1613) {
if (negBalance < 22) {
if (Cs < 1565) {
return 4;
}
else if (Cs >= 1565) {
if (Cs < 1593) {
return 8;
}
else if (Cs >= 1593) {
return 4;
}
}
}
else if (negBalance >= 22) {
if (posBalance < 60) {
if (Rs < 1491) {
if (Rs < 1481) {
return 4;
}
else if (Rs >= 1481) {
return 8;
}
}
else if (Rs >= 1491) {
return 4;
}
}
else if (posBalance >= 60) {
return 8;
}
}
}
else if (Cs >= 1613) {
if (Rs < 2471) {
if (ssDCval < 79) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
return 4;
}
}
else if (ssDCval >= 79) {
return 8;
}
}
else if (Rs >= 2471) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (Rs < 1945) {
if (ssDCval < 117) {
if (diffAFD < -3) {
if (Rs < 1710) {
return 8;
}
else if (Rs >= 1710) {
return 2;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (ssDCval >= 117) {
if (ssDCval < 119) {
if (mu_mv_mag_sq < 118) {
return 8;
}
else if (mu_mv_mag_sq >= 118) {
if (Rs < 1648) {
return 8;
}
else if (Rs >= 1648) {
return 1;
}
}
}
else if (ssDCval >= 119) {
if (diffMVdiffVal < 3) {
return 8;
}
else if (diffMVdiffVal >= 3) {
return 4;
}
}
}
}
else if (Rs >= 1945) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
if (Rs < 2825) {
if (diffAFD < -2) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 2;
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (Rs >= 2825) {
return 2;
}
}
}
}
else if (ssDCval >= 140) {
if (Rs < 2136) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (Rs >= 2136) {
return 1;
}
}
}
}
}
else if (Rs >= 3139) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 126) {
if (ssDCval < 63) {
if (negBalance < 21) {
if (Rs < 5497) {
if (ssDCval < 61) {
if (negBalance < 19) {
return 4;
}
else if (negBalance >= 19) {
if (Rs < 4773) {
return 8;
}
else if (Rs >= 4773) {
return 4;
}
}
}
else if (ssDCval >= 61) {
return 8;
}
}
else if (Rs >= 5497) {
return 8;
}
}
else if (negBalance >= 21) {
return 8;
}
}
else if (ssDCval >= 63) {
if (AFD < 3) {
if (ssDCval < 70) {
return 1;
}
else if (ssDCval >= 70) {
return 8;
}
}
else if (AFD >= 3) {
if (ssDCval < 74) {
if (Rs < 3491) {
return 2;
}
else if (Rs >= 3491) {
return 4;
}
}
else if (ssDCval >= 74) {
if (Rs < 4265) {
return 2;
}
else if (Rs >= 4265) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 1570) {
if (RsCsDiff < 12) {
if (Rs < 3657) {
if (Rs < 3619) {
if (Rs < 3520) {
return 1;
}
else if (Rs >= 3520) {
return 8;
}
}
else if (Rs >= 3619) {
return 4;
}
}
else if (Rs >= 3657) {
return 8;
}
}
else if (RsCsDiff >= 12) {
if (Rs < 3681) {
return 8;
}
else if (Rs >= 3681) {
return 4;
}
}
}
else if (Cs >= 1570) {
if (Rs < 3265) {
return 8;
}
else if (Rs >= 3265) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 110) {
if (ssDCval < 88) {
return 2;
}
else if (ssDCval >= 88) {
if (Rs < 4121) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
if (Rs < 3509) {
return 2;
}
else if (Rs >= 3509) {
return 4;
}
}
}
else if (Rs >= 4121) {
if (ssDCval < 96) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 2;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
}
else if (ssDCval >= 110) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
if (Rs < 3657) {
if (Rs < 3325) {
return 2;
}
else if (Rs >= 3325) {
return 1;
}
}
else if (Rs >= 3657) {
return 4;
}
}
}
}
}
}
else if (Cs >= 1641) {
if (ssDCval < 98) {
if (SC < 4361) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (Cs < 1789) {
if (MVDiff < 2) {
if (ssDCval < 63) {
if (Rs < 941) {
return 1;
}
else if (Rs >= 941) {
if (Rs < 1616) {
if (Rs < 1379) {
return 8;
}
else if (Rs >= 1379) {
return 4;
}
}
else if (Rs >= 1616) {
return 8;
}
}
}
else if (ssDCval >= 63) {
if (Rs < 2383) {
if (Cs < 1730) {
return 8;
}
else if (Cs >= 1730) {
if (Cs < 1757) {
return 1;
}
else if (Cs >= 1757) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
return 1;
}
}
}
}
else if (Rs >= 2383) {
return 1;
}
}
}
else if (MVDiff >= 2) {
if (Rs < 1022) {
return 8;
}
else if (Rs >= 1022) {
return 4;
}
}
}
else if (Cs >= 1789) {
if (diffRsCsdiff < -2) {
if (ssDCval < 64) {
if (diffMVdiffVal < 1) {
if (CsDiff < 1) {
if (Rs < 1879) {
return 1;
}
else if (Rs >= 1879) {
return 8;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
else if (ssDCval >= 64) {
if (Rs < 1022) {
return 8;
}
else if (Rs >= 1022) {
return 1;
}
}
}
else if (diffRsCsdiff >= -2) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
if (negBalance < 16) {
if (Rs < 1367) {
return 8;
}
else if (Rs >= 1367) {
return 1;
}
}
else if (negBalance >= 16) {
if (ssDCval < 59) {
return 8;
}
else if (ssDCval >= 59) {
if (AFD < 3) {
if (posBalance < 42) {
if (SC < 3966) {
return 8;
}
else if (SC >= 3966) {
return 1;
}
}
else if (posBalance >= 42) {
return 1;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
}
}
}
}
else if (ssDCval >= 67) {
if (TSC < 4) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
if (Cs < 1859) {
return 8;
}
else if (Cs >= 1859) {
if (Rs < 2297) {
return 2;
}
else if (Rs >= 2297) {
return 1;
}
}
}
}
else if (TSC >= 4) {
if (Rs < 1619) {
return 1;
}
else if (Rs >= 1619) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 2502) {
if (MVDiff < 3) {
if (negBalance < 61) {
if (Cs < 2305) {
if (Rs < 959) {
return 1;
}
else if (Rs >= 959) {
return 8;
}
}
else if (Cs >= 2305) {
if (Rs < 1834) {
return 8;
}
else if (Rs >= 1834) {
return 4;
}
}
}
else if (negBalance >= 61) {
if (Cs < 1865) {
return 4;
}
else if (Cs >= 1865) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (negBalance < 32) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 4;
}
}
else if (negBalance >= 32) {
if (Cs < 2207) {
return 8;
}
else if (Cs >= 2207) {
return 2;
}
}
}
}
else if (Rs >= 2502) {
if (ssDCval < 82) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 4;
}
}
else if (ssDCval >= 82) {
return 4;
}
}
}
}
else if (SC >= 4361) {
if (Cs < 2197) {
if (Rs < 3085) {
if (Cs < 1777) {
if (negBalance < 17) {
if (Rs < 2987) {
return 8;
}
else if (Rs >= 2987) {
return 1;
}
}
else if (negBalance >= 17) {
if (Rs < 2935) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
else if (Rs >= 2935) {
if (AFD < 3) {
if (RsDiff < 2) {
return 4;
}
else if (RsDiff >= 2) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 4;
}
}
}
else if (AFD >= 3) {
if (Rs < 3027) {
return 4;
}
else if (Rs >= 3027) {
return 4;
}
}
}
}
}
else if (Cs >= 1777) {
if (Rs < 2390) {
if (posBalance < 27) {
if (refDCval < 92) {
if (Rs < 2363) {
return 1;
}
else if (Rs >= 2363) {
if (Rs < 2384) {
return 4;
}
else if (Rs >= 2384) {
return 2;
}
}
}
else if (refDCval >= 92) {
if (Cs < 2117) {
if (Rs < 2331) {
if (Cs < 2068) {
return 4;
}
else if (Cs >= 2068) {
if (Rs < 2327) {
return 2;
}
else if (Rs >= 2327) {
return 1;
}
}
}
else if (Rs >= 2331) {
return 4;
}
}
else if (Cs >= 2117) {
return 2;
}
}
}
else if (posBalance >= 27) {
if (negBalance < 6) {
return 1;
}
else if (negBalance >= 6) {
return 8;
}
}
}
else if (Rs >= 2390) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
if (Cs < 2037) {
if (negBalance < 33) {
if (Rs < 2607) {
return 8;
}
else if (Rs >= 2607) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (mu_mv_mag_sq < 121) {
return 4;
}
else if (mu_mv_mag_sq >= 121) {
if (MVDiff < 1) {
if (ssDCval < 84) {
if (ssDCval < 74) {
if (posBalance < 38) {
return 4;
}
else if (posBalance >= 38) {
if (Cs < 1890) {
if (Cs < 1872) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (negBalance < 22) {
return 8;
}
else if (negBalance >= 22) {
if (Rs < 2987) {
return 4;
}
else if (Rs >= 2987) {
return 8;
}
}
}
}
else if (Cs >= 1872) {
return 4;
}
}
else if (Cs >= 1890) {
if (ssDCval < 64) {
return 4;
}
else if (ssDCval >= 64) {
return 8;
}
}
}
}
else if (ssDCval >= 74) {
return 4;
}
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (MVDiff >= 1) {
if (Rs < 3028) {
return 8;
}
else if (Rs >= 3028) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 33) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
if (negBalance < 59) {
return 8;
}
else if (negBalance >= 59) {
return 4;
}
}
}
}
else if (Cs >= 2037) {
if (negBalance < 19) {
if (Rs < 2941) {
return 4;
}
else if (Rs >= 2941) {
return 8;
}
}
else if (negBalance >= 19) {
if (Rs < 2971) {
if (ssDCval < 64) {
if (Rs < 2879) {
return 8;
}
else if (Rs >= 2879) {
return 4;
}
}
else if (ssDCval >= 64) {
return 8;
}
}
else if (Rs >= 2971) {
if (Cs < 2101) {
return 4;
}
else if (Cs >= 2101) {
return 8;
}
}
}
}
}
}
}
}
else if (Rs >= 3085) {
if (refDCval < 87) {
if (ssDCval < 66) {
if (Rs < 7250) {
if (CsDiff < 2) {
if (Cs < 1668) {
return 4;
}
else if (Cs >= 1668) {
if (Cs < 2185) {
if (Rs < 3099) {
return 4;
}
else if (Rs >= 3099) {
return 8;
}
}
else if (Cs >= 2185) {
return 4;
}
}
}
else if (CsDiff >= 2) {
if (Cs < 1849) {
if (Cs < 1764) {
return 4;
}
else if (Cs >= 1764) {
return 8;
}
}
else if (Cs >= 1849) {
if (ssDCval < 62) {
if (SC < 6297) {
if (SC < 6235) {
if (Rs < 3787) {
return 8;
}
else if (Rs >= 3787) {
if (Rs < 4187) {
if (Rs < 4130) {
return 4;
}
else if (Rs >= 4130) {
if (Rs < 4156) {
return 4;
}
else if (Rs >= 4156) {
return 4;
}
}
}
else if (Rs >= 4187) {
return 8;
}
}
}
else if (SC >= 6235) {
return 8;
}
}
else if (SC >= 6297) {
if (SC < 6376) {
return 4;
}
else if (SC >= 6376) {
if (diffTSC < 1) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Cs < 2114) {
if (Rs < 4417) {
return 4;
}
else if (Rs >= 4417) {
return 4;
}
}
else if (Cs >= 2114) {
if (MVDiff < 4) {
if (Rs < 4541) {
if (diffMVdiffVal < 1) {
if (Cs < 2136) {
return 4;
}
else if (Cs >= 2136) {
return 4;
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 2155) {
return 4;
}
else if (Cs >= 2155) {
return 8;
}
}
}
else if (Rs >= 4541) {
return 8;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
else if (diffTSC >= 1) {
if (Cs < 2169) {
return 8;
}
else if (Cs >= 2169) {
return 4;
}
}
}
}
}
else if (ssDCval >= 62) {
return 4;
}
}
}
}
else if (Rs >= 7250) {
return 2;
}
}
else if (ssDCval >= 66) {
if (SC < 5901) {
if (Cs < 1975) {
if (Rs < 3508) {
if (Rs < 3260) {
return 2;
}
else if (Rs >= 3260) {
return 4;
}
}
else if (Rs >= 3508) {
return 2;
}
}
else if (Cs >= 1975) {
if (refDCval < 68) {
return 4;
}
else if (refDCval >= 68) {
return 8;
}
}
}
else if (SC >= 5901) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Rs < 4486) {
return 2;
}
else if (Rs >= 4486) {
return 8;
}
}
}
}
}
else if (refDCval >= 87) {
if (posBalance < 27) {
return 8;
}
else if (posBalance >= 27) {
if (mu_mv_mag_sq < 117) {
if (negBalance < 27) {
return 2;
}
else if (negBalance >= 27) {
if (ssDCval < 91) {
if (Rs < 4648) {
return 2;
}
else if (Rs >= 4648) {
return 8;
}
}
else if (ssDCval >= 91) {
if (Rs < 4152) {
return 2;
}
else if (Rs >= 4152) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 117) {
return 2;
}
}
}
}
}
else if (Cs >= 2197) {
if (refDCval < 85) {
if (Rs < 6089) {
if (negBalance < 24) {
if (ssDCval < 53) {
if (refDCval < 43) {
return 8;
}
else if (refDCval >= 43) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (ssDCval >= 53) {
if (ssDCval < 66) {
if (Cs < 3457) {
if (Cs < 2927) {
if (SC < 6703) {
return 8;
}
else if (SC >= 6703) {
if (Rs < 4620) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (Rs < 4576) {
return 4;
}
else if (Rs >= 4576) {
return 4;
}
}
}
else if (Rs >= 4620) {
if (negBalance < 9) {
if (Rs < 4824) {
return 4;
}
else if (Rs >= 4824) {
return 4;
}
}
else if (negBalance >= 9) {
if (Cs < 2781) {
if (diffMVdiffVal < -1) {
if (refDCval < 55) {
return 8;
}
else if (refDCval >= 55) {
if (RsCsDiff < 37) {
if (Rs < 5102) {
return 8;
}
else if (Rs >= 5102) {
return 8;
}
}
else if (RsCsDiff >= 37) {
return 4;
}
}
}
else if (diffMVdiffVal >= -1) {
if (mcTcor < 98) {
if (Cs < 2729) {
if (Rs < 5027) {
if (CsDiff < 5) {
if (posBalance < 29) {
if (Rs < 4797) {
return 8;
}
else if (Rs >= 4797) {
return 8;
}
}
else if (posBalance >= 29) {
return 4;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (Rs >= 5027) {
if (Cs < 2602) {
return 4;
}
else if (Cs >= 2602) {
if (Rs < 5058) {
return 4;
}
else if (Rs >= 5058) {
if (Rs < 5105) {
return 4;
}
else if (Rs >= 5105) {
return 4;
}
}
}
}
}
else if (Cs >= 2729) {
if (Rs < 5117) {
return 8;
}
else if (Rs >= 5117) {
return 8;
}
}
}
else if (mcTcor >= 98) {
return 8;
}
}
}
else if (Cs >= 2781) {
return 4;
}
}
}
}
}
else if (Cs >= 2927) {
return 4;
}
}
else if (Cs >= 3457) {
return 8;
}
}
else if (ssDCval >= 66) {
if (Rs < 2909) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 4;
}
}
else if (Rs >= 2909) {
if (Cs < 4031) {
if (refDCval < 82) {
return 8;
}
else if (refDCval >= 82) {
if (Rs < 4391) {
return 8;
}
else if (Rs >= 4391) {
return 4;
}
}
}
else if (Cs >= 4031) {
return 1;
}
}
}
}
}
else if (negBalance >= 24) {
if (Rs < 2238) {
if (ssDCval < 72) {
if (SC < 5540) {
return 4;
}
else if (SC >= 5540) {
return 8;
}
}
else if (ssDCval >= 72) {
if (CsDiff < 5) {
if (Cs < 2368) {
return 8;
}
else if (Cs >= 2368) {
return 4;
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (Rs >= 2238) {
if (mu_mv_mag_sq < 116) {
if (Cs < 3063) {
if (posBalance < 62) {
if (mcTcor < 96) {
if (Rs < 2417) {
return 4;
}
else if (Rs >= 2417) {
return 8;
}
}
else if (mcTcor >= 96) {
if (mu_mv_mag_sq < 113) {
if (Cs < 2564) {
if (Rs < 3755) {
return 8;
}
else if (Rs >= 3755) {
return 4;
}
}
else if (Cs >= 2564) {
return 4;
}
}
else if (mu_mv_mag_sq >= 113) {
if (Cs < 2570) {
return 8;
}
else if (Cs >= 2570) {
return 8;
}
}
}
}
else if (posBalance >= 62) {
return 4;
}
}
else if (Cs >= 3063) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
if (diffRsCsdiff < 20) {
if (negBalance < 29) {
if (Cs < 4275) {
if (MVDiff < 6) {
if (refDCval < 81) {
return 8;
}
else if (refDCval >= 81) {
if (Rs < 4404) {
return 8;
}
else if (Rs >= 4404) {
return 4;
}
}
}
else if (MVDiff >= 6) {
if (Cs < 2514) {
return 4;
}
else if (Cs >= 2514) {
return 8;
}
}
}
else if (Cs >= 4275) {
if (Rs < 4376) {
return 8;
}
else if (Rs >= 4376) {
return 1;
}
}
}
else if (negBalance >= 29) {
if (Rs < 2867) {
if (Rs < 2865) {
return 8;
}
else if (Rs >= 2865) {
return 4;
}
}
else if (Rs >= 2867) {
return 8;
}
}
}
else if (diffRsCsdiff >= 20) {
if (diffTSC < 4) {
if (SC < 6990) {
if (SC < 6518) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (SC >= 6518) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 4;
}
}
}
else if (SC >= 6990) {
return 8;
}
}
else if (diffTSC >= 4) {
return 1;
}
}
}
}
}
}
else if (Rs >= 6089) {
if (negBalance < 26) {
return 8;
}
else if (negBalance >= 26) {
if (Rs < 6485) {
return 2;
}
else if (Rs >= 6485) {
return 4;
}
}
}
}
else if (refDCval >= 85) {
if (SC < 8349) {
if (Cs < 2724) {
if (negBalance < 52) {
if (Rs < 4523) {
if (Cs < 2622) {
if (Cs < 2283) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 2;
}
}
else if (Cs >= 2283) {
return 8;
}
}
else if (Cs >= 2622) {
if (mu_mv_mag_sq < 120) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 8;
}
}
else if (mu_mv_mag_sq >= 120) {
return 8;
}
}
}
else if (Rs >= 4523) {
if (Rs < 5093) {
return 2;
}
else if (Rs >= 5093) {
return 4;
}
}
}
else if (negBalance >= 52) {
return 4;
}
}
else if (Cs >= 2724) {
if (posBalance < 45) {
return 4;
}
else if (posBalance >= 45) {
if (Cs < 3095) {
if (AFD < 6) {
if (Rs < 4124) {
return 8;
}
else if (Rs >= 4124) {
return 4;
}
}
else if (AFD >= 6) {
return 4;
}
}
else if (Cs >= 3095) {
if (Rs < 3369) {
if (AFD < 11) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
else if (AFD >= 11) {
return 4;
}
}
else if (Rs >= 3369) {
if (RsDiff < 10) {
if (Rs < 3961) {
if (mcTcor < 99) {
if (negBalance < 46) {
if (tcor < 98) {
if (negBalance < 45) {
return 4;
}
else if (negBalance >= 45) {
if (diffRsCsdiff < 11) {
return 4;
}
else if (diffRsCsdiff >= 11) {
if (Rs < 3502) {
return 8;
}
else if (Rs >= 3502) {
return 4;
}
}
}
}
else if (tcor >= 98) {
if (Rs < 3674) {
return 4;
}
else if (Rs >= 3674) {
return 8;
}
}
}
else if (negBalance >= 46) {
if (negBalance < 49) {
if (Cs < 3372) {
return 4;
}
else if (Cs >= 3372) {
return 8;
}
}
else if (negBalance >= 49) {
return 4;
}
}
}
else if (mcTcor >= 99) {
return 8;
}
}
else if (Rs >= 3961) {
if (Cs < 3642) {
return 4;
}
else if (Cs >= 3642) {
return 8;
}
}
}
else if (RsDiff >= 10) {
return 8;
}
}
}
}
}
}
else if (SC >= 8349) {
if (Rs < 6938) {
if (Cs < 3722) {
return 4;
}
else if (Cs >= 3722) {
return 8;
}
}
else if (Rs >= 6938) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 98) {
if (ssDCval < 129) {
if (Rs < 1474) {
if (posBalance < 54) {
if (Cs < 2179) {
if (Rs < 1101) {
return 2;
}
else if (Rs >= 1101) {
return 8;
}
}
else if (Cs >= 2179) {
return 2;
}
}
else if (posBalance >= 54) {
if (SC < 3120) {
if (Rs < 1233) {
if (Rs < 671) {
return 8;
}
else if (Rs >= 671) {
return 4;
}
}
else if (Rs >= 1233) {
if (Rs < 1299) {
return 2;
}
else if (Rs >= 1299) {
return 8;
}
}
}
else if (SC >= 3120) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 4;
}
}
}
}
else if (Rs >= 1474) {
if (refDCval < 105) {
if (Cs < 1802) {
if (Rs < 3867) {
if (Rs < 3411) {
return 8;
}
else if (Rs >= 3411) {
return 2;
}
}
else if (Rs >= 3867) {
if (Rs < 4484) {
return 4;
}
else if (Rs >= 4484) {
return 1;
}
}
}
else if (Cs >= 1802) {
if (Rs < 3792) {
if (Rs < 2054) {
return 4;
}
else if (Rs >= 2054) {
if (negBalance < 40) {
return 8;
}
else if (negBalance >= 40) {
if (Cs < 2563) {
return 8;
}
else if (Cs >= 2563) {
if (Cs < 3181) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 3181) {
if (Rs < 2963) {
return 4;
}
else if (Rs >= 2963) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3792) {
if (Cs < 2587) {
if (CsDiff < 6) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (Cs < 1913) {
if (Rs < 4097) {
return 2;
}
else if (Rs >= 4097) {
return 4;
}
}
else if (Cs >= 1913) {
return 2;
}
}
}
else if (CsDiff >= 6) {
if (Rs < 4884) {
return 1;
}
else if (Rs >= 4884) {
return 2;
}
}
}
else if (Cs >= 2587) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
if (Rs < 4343) {
return 1;
}
else if (Rs >= 4343) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 105) {
if (diffRsCsdiff < -999) {
return 1;
}
else if (diffRsCsdiff >= -999) {
if (MVDiff < 1) {
if (diffTSC < -3) {
if (tcor < 100) {
if (mu_mv_mag_sq < 122) {
return 4;
}
else if (mu_mv_mag_sq >= 122) {
if (ssDCval < 108) {
return 4;
}
else if (ssDCval >= 108) {
return 8;
}
}
}
else if (tcor >= 100) {
return 1;
}
}
else if (diffTSC >= -3) {
if (posBalance < 63) {
if (Rs < 1893) {
if (tcor < 99) {
if (Cs < 1652) {
if (Rs < 1859) {
return 1;
}
else if (Rs >= 1859) {
return 8;
}
}
else if (Cs >= 1652) {
return 8;
}
}
else if (tcor >= 99) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
}
else if (Rs >= 1893) {
if (diffMVdiffVal < -2) {
if (ssDCval < 116) {
return 8;
}
else if (ssDCval >= 116) {
if (ssDCval < 119) {
if (Rs < 2768) {
return 4;
}
else if (Rs >= 2768) {
return 2;
}
}
else if (ssDCval >= 119) {
if (posBalance < 6) {
return 1;
}
else if (posBalance >= 6) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 3417) {
if (posBalance < 12) {
if (ssDCval < 114) {
return 8;
}
else if (ssDCval >= 114) {
return 2;
}
}
else if (posBalance >= 12) {
return 8;
}
}
else if (Rs >= 3417) {
if (Cs < 2053) {
if (Rs < 3600) {
return 4;
}
else if (Rs >= 3600) {
if (Rs < 4447) {
return 8;
}
else if (Rs >= 4447) {
return 2;
}
}
}
else if (Cs >= 2053) {
if (SC < 7279) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
if (RsDiff < 7) {
return 2;
}
else if (RsDiff >= 7) {
return 8;
}
}
}
else if (SC >= 7279) {
if (mu_mv_mag_sq < 121) {
if (Rs < 6395) {
return 8;
}
else if (Rs >= 6395) {
return 4;
}
}
else if (mu_mv_mag_sq >= 121) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 63) {
if (Cs < 1824) {
if (Rs < 2345) {
return 8;
}
else if (Rs >= 2345) {
return 1;
}
}
else if (Cs >= 1824) {
return 8;
}
}
}
}
else if (MVDiff >= 1) {
if (Cs < 2852) {
if (MVDiff < 58) {
if (tcor < 100) {
if (refDCval < 124) {
return 8;
}
else if (refDCval >= 124) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
}
else if (tcor >= 100) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
if (Rs < 2470) {
return 8;
}
else if (Rs >= 2470) {
if (Rs < 2702) {
return 4;
}
else if (Rs >= 2702) {
return 1;
}
}
}
}
}
else if (MVDiff >= 58) {
if (Rs < 2494) {
return 8;
}
else if (Rs >= 2494) {
if (Rs < 2790) {
return 4;
}
else if (Rs >= 2790) {
return 1;
}
}
}
}
else if (Cs >= 2852) {
if (SC < 5815) {
if (SC < 5755) {
if (posBalance < 39) {
if (posBalance < 15) {
if (mu_mv_mag_sq < 118) {
return 8;
}
else if (mu_mv_mag_sq >= 118) {
if (MVDiff < 7) {
return 4;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
else if (posBalance >= 15) {
if (Rs < 2580) {
return 8;
}
else if (Rs >= 2580) {
return 4;
}
}
}
else if (posBalance >= 39) {
return 8;
}
}
else if (SC >= 5755) {
if (Rs < 2841) {
return 2;
}
else if (Rs >= 2841) {
return 8;
}
}
}
else if (SC >= 5815) {
if (gchDC < 1) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (MVDiff < 4) {
if (Rs < 6363) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (posBalance < 48) {
return 4;
}
else if (posBalance >= 48) {
return 8;
}
}
}
else if (Rs >= 6363) {
return 4;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
else if (gchDC >= 1) {
if (Cs < 3322) {
if (AFD < 11) {
if (Cs < 2864) {
return 2;
}
else if (Cs >= 2864) {
return 8;
}
}
else if (AFD >= 11) {
return 2;
}
}
else if (Cs >= 3322) {
return 8;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 129) {
if (mcTcor < 96) {
if (negBalance < 39) {
if (ssDCval < 199) {
if (diffAFD < 5) {
if (ssDCval < 130) {
if (Rs < 2478) {
return 1;
}
else if (Rs >= 2478) {
return 8;
}
}
else if (ssDCval >= 130) {
return 8;
}
}
else if (diffAFD >= 5) {
return 1;
}
}
else if (ssDCval >= 199) {
if (Rs < 1688) {
return 1;
}
else if (Rs >= 1688) {
return 4;
}
}
}
else if (negBalance >= 39) {
if (ssDCval < 137) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 2;
}
}
else if (ssDCval >= 137) {
if (Rs < 5949) {
return 1;
}
else if (Rs >= 5949) {
if (Rs < 8232) {
return 8;
}
else if (Rs >= 8232) {
return 4;
}
}
}
}
}
else if (mcTcor >= 96) {
if (refDCval < 130) {
if (negBalance < 33) {
if (Rs < 1787) {
return 4;
}
else if (Rs >= 1787) {
return 1;
}
}
else if (negBalance >= 33) {
return 8;
}
}
else if (refDCval >= 130) {
if (ssDCval < 200) {
if (ssDCval < 144) {
if (Rs < 4571) {
if (diffAFD < -20) {
if (Rs < 2415) {
return 8;
}
else if (Rs >= 2415) {
return 1;
}
}
else if (diffAFD >= -20) {
if (Cs < 3890) {
if (ssDCval < 131) {
if (posBalance < 60) {
if (diffMVdiffVal < -2) {
return 4;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (posBalance >= 60) {
return 4;
}
}
else if (ssDCval >= 131) {
return 8;
}
}
else if (Cs >= 3890) {
if (Rs < 3973) {
return 1;
}
else if (Rs >= 3973) {
return 8;
}
}
}
}
else if (Rs >= 4571) {
if (Cs < 3560) {
if (Rs < 4874) {
return 1;
}
else if (Rs >= 4874) {
return 2;
}
}
else if (Cs >= 3560) {
return 8;
}
}
}
else if (ssDCval >= 144) {
return 8;
}
}
else if (ssDCval >= 200) {
if (Rs < 2684) {
return 8;
}
else if (Rs >= 2684) {
return 4;
}
}
}
}
}
}
}
}
}
return 8;
}
mfxU16 AGOPSelect17(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 SC, mfxU32 ssDCval) {
if (Cs < 1307) {
if (Rs < 2354) {
if (Cs < 491) {
if (negBalance < 14) {
if (SC < 810) {
if (Cs < 223) {
if (refDCval < 48) {
if (AFD < 1) {
if (mu_mv_mag_sq < 60) {
if (tcor < 97) {
if (Rs < 28) {
return 8;
}
else if (Rs >= 28) {
return 1;
}
}
else if (tcor >= 97) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 60) {
if (Rs < 254) {
if (SC < 59) {
if (ssDCval < 22) {
if (MVDiff < 1076) {
if (ssDCval < 10) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
return 8;
}
}
else if (ssDCval >= 10) {
return 8;
}
}
else if (MVDiff >= 1076) {
return 4;
}
}
else if (ssDCval >= 22) {
if (posBalance < 16) {
if (negBalance < 13) {
return 4;
}
else if (negBalance >= 13) {
if (Cs < 25) {
return 8;
}
else if (Cs >= 25) {
return 4;
}
}
}
else if (posBalance >= 16) {
if (posBalance < 58) {
if (Rs < 7) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
return 4;
}
}
else if (Rs >= 7) {
return 1;
}
}
else if (posBalance >= 58) {
return 8;
}
}
}
}
else if (SC >= 59) {
if (MVDiff < 7) {
if (negBalance < 8) {
if (Cs < 186) {
return 8;
}
else if (Cs >= 186) {
if (Rs < 198) {
return 4;
}
else if (Rs >= 198) {
return 8;
}
}
}
else if (negBalance >= 8) {
if (SC < 313) {
if (Cs < 101) {
if (mu_mv_mag_sq < 113) {
if (Rs < 24) {
return 8;
}
else if (Rs >= 24) {
return 4;
}
}
else if (mu_mv_mag_sq >= 113) {
return 8;
}
}
else if (Cs >= 101) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
if (ssDCval < 27) {
return 1;
}
else if (ssDCval >= 27) {
return 8;
}
}
}
}
else if (SC >= 313) {
return 4;
}
}
}
else if (MVDiff >= 7) {
if (Rs < 150) {
if (negBalance < 10) {
if (Rs < 94) {
if (ssDCval < 18) {
return 8;
}
else if (ssDCval >= 18) {
if (negBalance < 4) {
return 8;
}
else if (negBalance >= 4) {
return 1;
}
}
}
else if (Rs >= 94) {
return 8;
}
}
else if (negBalance >= 10) {
return 4;
}
}
else if (Rs >= 150) {
if (ssDCval < 32) {
return 4;
}
else if (ssDCval >= 32) {
return 8;
}
}
}
}
}
else if (Rs >= 254) {
if (posBalance < 16) {
if (tcor < 63) {
return 4;
}
else if (tcor >= 63) {
return 8;
}
}
else if (posBalance >= 16) {
return 4;
}
}
}
}
else if (AFD >= 1) {
if (tcor < 82) {
if (gchDC < 12) {
if (posBalance < 75) {
if (refDCval < 19) {
return 8;
}
else if (refDCval >= 19) {
if (mu_mv_mag_sq < 95) {
if (posBalance < 64) {
if (diffMVdiffVal < -12) {
if (Rs < 283) {
return 8;
}
else if (Rs >= 283) {
return 1;
}
}
else if (diffMVdiffVal >= -12) {
return 1;
}
}
else if (posBalance >= 64) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 95) {
if (ssDCval < 24) {
return 8;
}
else if (ssDCval >= 24) {
return 4;
}
}
}
}
else if (posBalance >= 75) {
if (AFD < 3) {
if (mu_mv_mag_sq < 48) {
if (mu_mv_mag_sq < 41) {
return 2;
}
else if (mu_mv_mag_sq >= 41) {
if (Rs < 57) {
return 8;
}
else if (Rs >= 57) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (MVDiff < 76) {
return 4;
}
else if (MVDiff >= 76) {
return 8;
}
}
}
else if (AFD >= 3) {
if (SC < 224) {
return 8;
}
else if (SC >= 224) {
if (mu_mv_mag_sq < 29) {
if (ssDCval < 42) {
return 2;
}
else if (ssDCval >= 42) {
return 8;
}
}
else if (mu_mv_mag_sq >= 29) {
if (Rs < 401) {
return 1;
}
else if (Rs >= 401) {
return 4;
}
}
}
}
}
}
else if (gchDC >= 12) {
return 1;
}
}
else if (tcor >= 82) {
if (refDCval < 31) {
if (mu_mv_mag_sq < 58) {
if (tcor < 96) {
if (MVDiff < 35) {
if (Cs < 7) {
if (Rs < 92) {
return 2;
}
else if (Rs >= 92) {
return 4;
}
}
else if (Cs >= 7) {
if (Cs < 188) {
return 1;
}
else if (Cs >= 188) {
return 4;
}
}
}
else if (MVDiff >= 35) {
if (TSC < 1) {
return 1;
}
else if (TSC >= 1) {
return 4;
}
}
}
else if (tcor >= 96) {
return 4;
}
}
else if (mu_mv_mag_sq >= 58) {
if (refDCval < 29) {
if (mcTcor < 89) {
if (posBalance < 68) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (MVDiff < 23) {
return 1;
}
else if (MVDiff >= 23) {
return 8;
}
}
}
else if (posBalance >= 68) {
if (TSCindex < 2) {
return 8;
}
else if (TSCindex >= 2) {
return 4;
}
}
}
else if (mcTcor >= 89) {
if (ssDCval < 19) {
if (ssDCval < 12) {
return 4;
}
else if (ssDCval >= 12) {
return 8;
}
}
else if (ssDCval >= 19) {
if (tcor < 96) {
return 1;
}
else if (tcor >= 96) {
if (Rs < 185) {
if (posBalance < 19) {
return 4;
}
else if (posBalance >= 19) {
return 1;
}
}
else if (Rs >= 185) {
return 4;
}
}
}
}
}
else if (refDCval >= 29) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 110) {
if (mu_mv_mag_sq < 74) {
return 1;
}
else if (mu_mv_mag_sq >= 74) {
return 4;
}
}
else if (Rs >= 110) {
return 1;
}
}
}
}
}
else if (refDCval >= 31) {
if (Rs < 259) {
if (ssDCval < 38) {
if (ssDCval < 33) {
return 8;
}
else if (ssDCval >= 33) {
if (negBalance < 10) {
if (ssDCval < 36) {
return 4;
}
else if (ssDCval >= 36) {
return 8;
}
}
else if (negBalance >= 10) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (ssDCval >= 38) {
if (posBalance < 29) {
return 1;
}
else if (posBalance >= 29) {
if (Cs < 169) {
return 8;
}
else if (Cs >= 169) {
return 1;
}
}
}
}
else if (Rs >= 259) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Rs < 405) {
return 1;
}
else if (Rs >= 405) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 48) {
if (ssDCval < 84) {
if (SC < 26) {
return 2;
}
else if (SC >= 26) {
if (refDCval < 67) {
if (MVDiff < 1) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (Rs < 98) {
return 2;
}
else if (Rs >= 98) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (mcTcor < 99) {
if (ssDCval < 58) {
return 8;
}
else if (ssDCval >= 58) {
if (Cs < 18) {
if (Rs < 448) {
return 2;
}
else if (Rs >= 448) {
return 4;
}
}
else if (Cs >= 18) {
if (tcor < 90) {
if (ssDCval < 60) {
return 8;
}
else if (ssDCval >= 60) {
return 1;
}
}
else if (tcor >= 90) {
if (SC < 168) {
return 4;
}
else if (SC >= 168) {
return 8;
}
}
}
}
}
else if (mcTcor >= 99) {
if (Cs < 132) {
if (Cs < 5) {
return 4;
}
else if (Cs >= 5) {
return 8;
}
}
else if (Cs >= 132) {
if (mu_mv_mag_sq < 119) {
return 4;
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 205) {
return 8;
}
else if (Rs >= 205) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 67) {
if (Cs < 120) {
if (Rs < 373) {
return 4;
}
else if (Rs >= 373) {
return 8;
}
}
else if (Cs >= 120) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
if (Rs < 385) {
return 8;
}
else if (Rs >= 385) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (ssDCval < 87) {
if (Cs < 58) {
return 4;
}
else if (Cs >= 58) {
return 1;
}
}
else if (ssDCval >= 87) {
if (SC < 310) {
if (AFD < 5) {
if (posBalance < 32) {
if (Rs < 80) {
return 2;
}
else if (Rs >= 80) {
return 8;
}
}
else if (posBalance >= 32) {
if (Rs < 22) {
return 1;
}
else if (Rs >= 22) {
return 4;
}
}
}
else if (AFD >= 5) {
if (Cs < 27) {
return 1;
}
else if (Cs >= 27) {
if (AFD < 32) {
return 8;
}
else if (AFD >= 32) {
return 2;
}
}
}
}
else if (SC >= 310) {
if (negBalance < 9) {
if (refDCval < 129) {
return 8;
}
else if (refDCval >= 129) {
return 4;
}
}
else if (negBalance >= 9) {
if (Rs < 146) {
return 4;
}
else if (Rs >= 146) {
if (Rs < 641) {
return 2;
}
else if (Rs >= 641) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 223) {
if (mu_mv_mag_sq < 109) {
if (AFD < 2) {
if (mu_mv_mag_sq < 71) {
return 4;
}
else if (mu_mv_mag_sq >= 71) {
if (mu_mv_mag_sq < 82) {
if (Rs < 135) {
return 4;
}
else if (Rs >= 135) {
if (Rs < 459) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
if (Rs < 247) {
return 8;
}
else if (Rs >= 247) {
return 4;
}
}
}
else if (Rs >= 459) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 82) {
if (Cs < 230) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
return 1;
}
}
else if (Cs >= 230) {
if (MVDiff < 6) {
if (Rs < 431) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (Rs < 137) {
return 4;
}
else if (Rs >= 137) {
return 2;
}
}
}
else if (Rs >= 431) {
if (Cs < 320) {
return 1;
}
else if (Cs >= 320) {
return 4;
}
}
}
else if (MVDiff >= 6) {
return 4;
}
}
}
}
}
else if (AFD >= 2) {
if (refDCval < 143) {
if (refDCval < 48) {
if (Rs < 300) {
if (ssDCval < 26) {
if (Rs < 192) {
return 8;
}
else if (Rs >= 192) {
return 1;
}
}
else if (ssDCval >= 26) {
if (SC < 496) {
if (Rs < 145) {
return 1;
}
else if (Rs >= 145) {
if (Rs < 198) {
return 4;
}
else if (Rs >= 198) {
return 2;
}
}
}
else if (SC >= 496) {
return 4;
}
}
}
else if (Rs >= 300) {
if (Rs < 305) {
return 2;
}
else if (Rs >= 305) {
if (Rs < 463) {
return 1;
}
else if (Rs >= 463) {
return 8;
}
}
}
}
else if (refDCval >= 48) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
}
else if (refDCval >= 143) {
if (MVDiff < 22) {
if (Rs < 389) {
return 8;
}
else if (Rs >= 389) {
return 1;
}
}
else if (MVDiff >= 22) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 378) {
if (refDCval < 58) {
if (Rs < 310) {
if (refDCval < 36) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (refDCval >= 36) {
if (posBalance < 40) {
if (tcor < 88) {
return 4;
}
else if (tcor >= 88) {
if (SC < 579) {
return 1;
}
else if (SC >= 579) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
}
}
else if (posBalance >= 40) {
return 8;
}
}
}
else if (Rs >= 310) {
if (Cs < 266) {
if (Rs < 334) {
return 8;
}
else if (Rs >= 334) {
return 1;
}
}
else if (Cs >= 266) {
return 4;
}
}
}
else if (refDCval >= 58) {
if (Rs < 252) {
if (Cs < 255) {
return 2;
}
else if (Cs >= 255) {
return 4;
}
}
else if (Rs >= 252) {
if (ssDCval < 130) {
if (Rs < 275) {
return 1;
}
else if (Rs >= 275) {
if (SC < 558) {
return 2;
}
else if (SC >= 558) {
if (SC < 662) {
return 8;
}
else if (SC >= 662) {
return 2;
}
}
}
}
else if (ssDCval >= 130) {
return 4;
}
}
}
}
else if (Rs >= 378) {
if (ssDCval < 85) {
if (negBalance < 13) {
if (tcor < 97) {
if (Rs < 413) {
return 4;
}
else if (Rs >= 413) {
return 8;
}
}
else if (tcor >= 97) {
return 8;
}
}
else if (negBalance >= 13) {
if (Rs < 388) {
return 4;
}
else if (Rs >= 388) {
return 1;
}
}
}
else if (ssDCval >= 85) {
if (ssDCval < 129) {
if (Rs < 450) {
return 2;
}
else if (Rs >= 450) {
return 1;
}
}
else if (ssDCval >= 129) {
if (Rs < 437) {
return 4;
}
else if (Rs >= 437) {
return 1;
}
}
}
}
}
}
}
else if (SC >= 810) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 128) {
if (ssDCval < 33) {
if (ssDCval < 29) {
if (Rs < 616) {
return 1;
}
else if (Rs >= 616) {
return 8;
}
}
else if (ssDCval >= 29) {
return 4;
}
}
else if (ssDCval >= 33) {
if (SC < 962) {
if (ssDCval < 122) {
if (RsDiff < 1) {
if (Rs < 515) {
return 8;
}
else if (Rs >= 515) {
if (Rs < 570) {
if (Cs < 390) {
return 1;
}
else if (Cs >= 390) {
return 8;
}
}
else if (Rs >= 570) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (Rs < 597) {
return 1;
}
else if (Rs >= 597) {
return 8;
}
}
}
else if (ssDCval >= 122) {
if (Rs < 536) {
if (posBalance < 10) {
return 8;
}
else if (posBalance >= 10) {
if (SC < 927) {
return 1;
}
else if (SC >= 927) {
return 8;
}
}
}
else if (Rs >= 536) {
return 8;
}
}
}
else if (SC >= 962) {
if (ssDCval < 67) {
if (ssDCval < 60) {
if (Rs < 1314) {
if (RsCsDiff < 2) {
if (posBalance < 39) {
return 8;
}
else if (posBalance >= 39) {
if (mcTcor < 100) {
if (Rs < 603) {
return 1;
}
else if (Rs >= 603) {
if (Rs < 873) {
return 8;
}
else if (Rs >= 873) {
if (Rs < 1117) {
return 1;
}
else if (Rs >= 1117) {
return 8;
}
}
}
}
else if (mcTcor >= 100) {
return 1;
}
}
}
else if (RsCsDiff >= 2) {
if (Rs < 891) {
return 2;
}
else if (Rs >= 891) {
return 4;
}
}
}
else if (Rs >= 1314) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (ssDCval >= 60) {
if (Cs < 199) {
return 8;
}
else if (Cs >= 199) {
return 1;
}
}
}
else if (ssDCval >= 67) {
if (ssDCval < 84) {
if (Rs < 1122) {
if (Rs < 646) {
return 4;
}
else if (Rs >= 646) {
if (ssDCval < 79) {
return 8;
}
else if (ssDCval >= 79) {
if (Rs < 1018) {
return 8;
}
else if (Rs >= 1018) {
return 2;
}
}
}
}
else if (Rs >= 1122) {
if (Rs < 1294) {
return 2;
}
else if (Rs >= 1294) {
return 1;
}
}
}
else if (ssDCval >= 84) {
if (Rs < 498) {
return 1;
}
else if (Rs >= 498) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (ssDCval < 135) {
if (refDCval < 60) {
return 8;
}
else if (refDCval >= 60) {
if (Cs < 253) {
if (Cs < 246) {
return 2;
}
else if (Cs >= 246) {
return 1;
}
}
else if (Cs >= 253) {
if (negBalance < 12) {
if (Cs < 317) {
if (Rs < 1026) {
return 1;
}
else if (Rs >= 1026) {
return 8;
}
}
else if (Cs >= 317) {
return 8;
}
}
else if (negBalance >= 12) {
if (Cs < 476) {
if (Cs < 391) {
return 8;
}
else if (Cs >= 391) {
if (Rs < 560) {
return 8;
}
else if (Rs >= 560) {
return 4;
}
}
}
else if (Cs >= 476) {
return 2;
}
}
}
}
}
else if (ssDCval >= 135) {
return 1;
}
}
}
else if (MVDiff >= 8) {
if (ssDCval < 66) {
if (tcor < 95) {
if (mu_mv_mag_sq < 19) {
if (diffAFD < 6) {
return 2;
}
else if (diffAFD >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 19) {
if (negBalance < 11) {
if (posBalance < 120) {
if (SC < 868) {
return 4;
}
else if (SC >= 868) {
return 8;
}
}
else if (posBalance >= 120) {
return 4;
}
}
else if (negBalance >= 11) {
if (Rs < 672) {
return 1;
}
else if (Rs >= 672) {
return 8;
}
}
}
}
else if (tcor >= 95) {
if (MVDiff < 51) {
if (Cs < 487) {
return 4;
}
else if (Cs >= 487) {
return 8;
}
}
else if (MVDiff >= 51) {
if (Rs < 738) {
return 8;
}
else if (Rs >= 738) {
return 4;
}
}
}
}
else if (ssDCval >= 66) {
if (diffTSC < -3) {
if (ssDCval < 130) {
if (diffAFD < -4) {
if (MVDiff < 15) {
if (Rs < 981) {
return 1;
}
else if (Rs >= 981) {
return 4;
}
}
else if (MVDiff >= 15) {
if (Rs < 999) {
return 2;
}
else if (Rs >= 999) {
if (Rs < 1428) {
return 1;
}
else if (Rs >= 1428) {
return 2;
}
}
}
}
else if (diffAFD >= -4) {
if (Rs < 1352) {
return 4;
}
else if (Rs >= 1352) {
return 1;
}
}
}
else if (ssDCval >= 130) {
if (Rs < 854) {
if (Rs < 712) {
return 8;
}
else if (Rs >= 712) {
return 4;
}
}
else if (Rs >= 854) {
return 1;
}
}
}
else if (diffTSC >= -3) {
if (Cs < 398) {
if (posBalance < 82) {
return 1;
}
else if (posBalance >= 82) {
if (AFD < 9) {
if (Cs < 358) {
return 8;
}
else if (Cs >= 358) {
return 1;
}
}
else if (AFD >= 9) {
if (MVDiff < 88) {
return 2;
}
else if (MVDiff >= 88) {
return 1;
}
}
}
}
else if (Cs >= 398) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 1;
}
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (refDCval < 35) {
if (mcTcor < 83) {
if (refDCval < 20) {
return 8;
}
else if (refDCval >= 20) {
if (Cs < 317) {
if (negBalance < 45) {
if (Rs < 1) {
return 4;
}
else if (Rs >= 1) {
if (diffMVdiffVal < 2) {
if (posBalance < 56) {
if (mu_mv_mag_sq < 9) {
if (Rs < 503) {
if (Rs < 223) {
return 8;
}
else if (Rs >= 223) {
return 1;
}
}
else if (Rs >= 503) {
return 4;
}
}
else if (mu_mv_mag_sq >= 9) {
if (AFD < 13) {
if (AFD < 2) {
if (Rs < 151) {
return 2;
}
else if (Rs >= 151) {
return 1;
}
}
else if (AFD >= 2) {
if (tcor < 78) {
return 1;
}
else if (tcor >= 78) {
if (Rs < 138) {
return 1;
}
else if (Rs >= 138) {
return 8;
}
}
}
}
else if (AFD >= 13) {
return 4;
}
}
}
else if (posBalance >= 56) {
if (Cs < 101) {
if (Rs < 41) {
return 8;
}
else if (Rs >= 41) {
return 1;
}
}
else if (Cs >= 101) {
if (Cs < 207) {
return 8;
}
else if (Cs >= 207) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 2) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 1;
}
}
}
}
else if (negBalance >= 45) {
if (Rs < 516) {
if (tcor < 64) {
if (MVDiff < 377) {
return 8;
}
else if (MVDiff >= 377) {
if (Rs < 187) {
return 4;
}
else if (Rs >= 187) {
return 1;
}
}
}
else if (tcor >= 64) {
if (tcor < 74) {
if (Rs < 133) {
return 2;
}
else if (Rs >= 133) {
return 1;
}
}
else if (tcor >= 74) {
if (MVDiff < 689) {
return 8;
}
else if (MVDiff >= 689) {
return 4;
}
}
}
}
else if (Rs >= 516) {
return 1;
}
}
}
else if (Cs >= 317) {
if (ssDCval < 23) {
return 4;
}
else if (ssDCval >= 23) {
if (RsDiff < 4) {
if (diffRsCsdiff < -12) {
return 8;
}
else if (diffRsCsdiff >= -12) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
return 2;
}
}
}
else if (RsDiff >= 4) {
if (Rs < 1149) {
return 8;
}
else if (Rs >= 1149) {
return 1;
}
}
}
}
}
}
else if (mcTcor >= 83) {
if (posBalance < 27) {
if (posBalance < 7) {
if (mu_mv_mag_sq < 103) {
if (Cs < 2) {
if (mu_mv_mag_sq < 31) {
if (AFD < 2) {
if (MVDiff < 229) {
return 4;
}
else if (MVDiff >= 229) {
return 8;
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (mu_mv_mag_sq >= 31) {
if (ssDCval < 19) {
return 8;
}
else if (ssDCval >= 19) {
return 4;
}
}
}
else if (Cs >= 2) {
return 4;
}
}
else if (mu_mv_mag_sq >= 103) {
return 8;
}
}
else if (posBalance >= 7) {
if (ssDCval < 29) {
if (negBalance < 63) {
if (negBalance < 17) {
if (MVDiff < 11) {
if (Rs < 113) {
return 4;
}
else if (Rs >= 113) {
return 8;
}
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (negBalance >= 17) {
if (Rs < 4) {
return 4;
}
else if (Rs >= 4) {
return 1;
}
}
}
else if (negBalance >= 63) {
return 8;
}
}
else if (ssDCval >= 29) {
if (TSCindex < 2) {
if (ssDCval < 31) {
return 4;
}
else if (ssDCval >= 31) {
if (mu_mv_mag_sq < 86) {
return 8;
}
else if (mu_mv_mag_sq >= 86) {
return 4;
}
}
}
else if (TSCindex >= 2) {
if (TSC < 2) {
if (negBalance < 25) {
if (Rs < 329) {
return 8;
}
else if (Rs >= 329) {
return 1;
}
}
else if (negBalance >= 25) {
return 1;
}
}
else if (TSC >= 2) {
if (MVDiff < 23) {
return 4;
}
else if (MVDiff >= 23) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 27) {
if (refDCval < 30) {
if (Cs < 375) {
if (MVDiff < 34) {
if (tcor < 77) {
if (posBalance < 39) {
if (Rs < 270) {
if (ssDCval < 29) {
return 1;
}
else if (ssDCval >= 29) {
return 4;
}
}
else if (Rs >= 270) {
if (negBalance < 24) {
return 4;
}
else if (negBalance >= 24) {
return 1;
}
}
}
else if (posBalance >= 39) {
if (mu_mv_mag_sq < 55) {
if (mcTcor < 84) {
if (Rs < 316) {
return 8;
}
else if (Rs >= 316) {
return 1;
}
}
else if (mcTcor >= 84) {
return 1;
}
}
else if (mu_mv_mag_sq >= 55) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
return 8;
}
}
}
}
else if (tcor >= 77) {
if (mu_mv_mag_sq < 36) {
if (mu_mv_mag_sq < 7) {
if (Rs < 6) {
return 2;
}
else if (Rs >= 6) {
return 1;
}
}
else if (mu_mv_mag_sq >= 7) {
if (MVDiff < 12) {
return 4;
}
else if (MVDiff >= 12) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 36) {
if (diffMVdiffVal < -11) {
if (Cs < 142) {
return 1;
}
else if (Cs >= 142) {
return 4;
}
}
else if (diffMVdiffVal >= -11) {
return 1;
}
}
}
}
else if (MVDiff >= 34) {
if (refDCval < 24) {
if (Rs < 51) {
return 1;
}
else if (Rs >= 51) {
return 8;
}
}
else if (refDCval >= 24) {
if (Cs < 232) {
if (ssDCval < 30) {
return 1;
}
else if (ssDCval >= 30) {
if (Rs < 256) {
return 4;
}
else if (Rs >= 256) {
return 1;
}
}
}
else if (Cs >= 232) {
if (Rs < 200) {
return 2;
}
else if (Rs >= 200) {
return 4;
}
}
}
}
}
else if (Cs >= 375) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 778) {
return 1;
}
else if (Rs >= 778) {
return 8;
}
}
}
}
else if (refDCval >= 30) {
if (RsCsDiff < 3) {
if (Cs < 260) {
if (Cs < 83) {
if (Cs < 10) {
if (Rs < 9) {
return 4;
}
else if (Rs >= 9) {
return 2;
}
}
else if (Cs >= 10) {
if (mu_mv_mag_sq < 99) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
if (gchDC < 2) {
if (posBalance < 38) {
if (Rs < 231) {
return 4;
}
else if (Rs >= 231) {
return 2;
}
}
else if (posBalance >= 38) {
if (Rs < 69) {
return 2;
}
else if (Rs >= 69) {
return 1;
}
}
}
else if (gchDC >= 2) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 99) {
return 4;
}
}
}
else if (Cs >= 83) {
if (SC < 576) {
if (mcTcor < 97) {
if (AFD < 4) {
if (posBalance < 36) {
if (ssDCval < 33) {
if (refDCval < 31) {
if (Cs < 124) {
return 4;
}
else if (Cs >= 124) {
return 8;
}
}
else if (refDCval >= 31) {
if (Rs < 318) {
return 8;
}
else if (Rs >= 318) {
if (Rs < 329) {
return 4;
}
else if (Rs >= 329) {
return 8;
}
}
}
}
else if (ssDCval >= 33) {
return 4;
}
}
else if (posBalance >= 36) {
if (RsCsDiff < 2) {
if (ssDCval < 32) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 32) {
return 8;
}
}
else if (RsCsDiff >= 2) {
return 8;
}
}
}
else if (AFD >= 4) {
if (negBalance < 42) {
if (ssDCval < 34) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (MVDiff < 10) {
if (Rs < 330) {
return 4;
}
else if (Rs >= 330) {
return 2;
}
}
else if (MVDiff >= 10) {
return 1;
}
}
}
else if (ssDCval >= 34) {
if (Cs < 150) {
if (Rs < 127) {
return 1;
}
else if (Rs >= 127) {
return 4;
}
}
else if (Cs >= 150) {
return 8;
}
}
}
else if (negBalance >= 42) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 109) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (Rs >= 109) {
if (mu_mv_mag_sq < 2) {
if (Rs < 326) {
return 4;
}
else if (Rs >= 326) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
return 1;
}
}
}
}
else if (SC >= 576) {
if (posBalance < 44) {
return 4;
}
else if (posBalance >= 44) {
return 1;
}
}
}
}
else if (Cs >= 260) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (RsCsDiff >= 3) {
if (TSCindex < 3) {
if (Cs < 305) {
return 4;
}
else if (Cs >= 305) {
return 1;
}
}
else if (TSCindex >= 3) {
if (AFD < 20) {
if (negBalance < 18) {
return 8;
}
else if (negBalance >= 18) {
return 1;
}
}
else if (AFD >= 20) {
return 8;
}
}
}
}
}
}
}
else if (refDCval >= 35) {
if (Rs < 300) {
if (ssDCval < 47) {
if (mu_mv_mag_sq < 49) {
if (mcTcor < 89) {
if (Cs < 64) {
if (SC < 14) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
else if (SC >= 14) {
if (diffMVdiffVal < 51) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
return 4;
}
}
else if (diffMVdiffVal >= 51) {
if (Rs < 147) {
return 4;
}
else if (Rs >= 147) {
return 8;
}
}
}
}
else if (Cs >= 64) {
if (gchDC < 4) {
if (tcor < 67) {
if (mcTcor < 86) {
if (mcTcor < 66) {
return 4;
}
else if (mcTcor >= 66) {
if (Rs < 205) {
if (Rs < 194) {
return 8;
}
else if (Rs >= 194) {
return 4;
}
}
else if (Rs >= 205) {
return 8;
}
}
}
else if (mcTcor >= 86) {
if (MVDiff < 27) {
return 4;
}
else if (MVDiff >= 27) {
return 1;
}
}
}
else if (tcor >= 67) {
if (Cs < 352) {
if (mcTcor < 86) {
return 8;
}
else if (mcTcor >= 86) {
if (Rs < 169) {
return 8;
}
else if (Rs >= 169) {
if (MVDiff < 10) {
return 8;
}
else if (MVDiff >= 10) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (Cs >= 352) {
return 1;
}
}
}
else if (gchDC >= 4) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
if (Rs < 100) {
if (Rs < 59) {
return 8;
}
else if (Rs >= 59) {
return 4;
}
}
else if (Rs >= 100) {
return 1;
}
}
}
}
}
else if (mcTcor >= 89) {
if (posBalance < 8) {
if (Rs < 230) {
if (mcTcor < 95) {
return 4;
}
else if (mcTcor >= 95) {
return 8;
}
}
else if (Rs >= 230) {
return 4;
}
}
else if (posBalance >= 8) {
if (mcTcor < 97) {
if (AFD < 9) {
if (Cs < 218) {
if (posBalance < 52) {
if (tcor < 90) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
if (negBalance < 44) {
if (SC < 244) {
if (Rs < 119) {
return 1;
}
else if (Rs >= 119) {
return 2;
}
}
else if (SC >= 244) {
return 4;
}
}
else if (negBalance >= 44) {
return 1;
}
}
}
else if (tcor >= 90) {
if (Rs < 143) {
return 1;
}
else if (Rs >= 143) {
return 2;
}
}
}
else if (posBalance >= 52) {
if (negBalance < 35) {
if (posBalance < 53) {
if (Cs < 42) {
return 2;
}
else if (Cs >= 42) {
return 1;
}
}
else if (posBalance >= 53) {
return 2;
}
}
else if (negBalance >= 35) {
return 1;
}
}
}
else if (Cs >= 218) {
if (Rs < 189) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
return 8;
}
}
else if (Rs >= 189) {
if (refDCval < 43) {
return 2;
}
else if (refDCval >= 43) {
if (Rs < 274) {
if (Rs < 255) {
return 2;
}
else if (Rs >= 255) {
return 8;
}
}
else if (Rs >= 274) {
return 4;
}
}
}
}
}
else if (AFD >= 9) {
if (diffRsCsdiff < -13) {
if (Cs < 150) {
return 1;
}
else if (Cs >= 150) {
return 8;
}
}
else if (diffRsCsdiff >= -13) {
if (Rs < 278) {
if (posBalance < 64) {
if (RsCsDiff < 1) {
if (Rs < 47) {
return 1;
}
else if (Rs >= 47) {
return 4;
}
}
else if (RsCsDiff >= 1) {
return 1;
}
}
else if (posBalance >= 64) {
return 4;
}
}
else if (Rs >= 278) {
return 4;
}
}
}
}
else if (mcTcor >= 97) {
if (MVDiff < 38) {
if (tcor < 91) {
if (Rs < 160) {
if (Rs < 54) {
return 1;
}
else if (Rs >= 54) {
if (Cs < 196) {
return 4;
}
else if (Cs >= 196) {
return 1;
}
}
}
else if (Rs >= 160) {
return 1;
}
}
else if (tcor >= 91) {
return 1;
}
}
else if (MVDiff >= 38) {
if (Rs < 180) {
return 2;
}
else if (Rs >= 180) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (negBalance < 30) {
if (tcor < 91) {
if (mu_mv_mag_sq < 58) {
if (posBalance < 38) {
if (Rs < 253) {
return 8;
}
else if (Rs >= 253) {
return 4;
}
}
else if (posBalance >= 38) {
if (tcor < 84) {
if (Rs < 180) {
return 1;
}
else if (Rs >= 180) {
return 8;
}
}
else if (tcor >= 84) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 58) {
if (Cs < 176) {
if (mu_mv_mag_sq < 94) {
return 1;
}
else if (mu_mv_mag_sq >= 94) {
if (ssDCval < 40) {
return 4;
}
else if (ssDCval >= 40) {
if (Rs < 180) {
return 1;
}
else if (Rs >= 180) {
return 4;
}
}
}
}
else if (Cs >= 176) {
return 1;
}
}
}
else if (tcor >= 91) {
if (MVDiff < 6) {
if (posBalance < 37) {
if (tcor < 101) {
if (Cs < 256) {
if (ssDCval < 43) {
if (ssDCval < 40) {
if (negBalance < 15) {
return 4;
}
else if (negBalance >= 15) {
if (tcor < 98) {
if (Rs < 69) {
return 4;
}
else if (Rs >= 69) {
return 1;
}
}
else if (tcor >= 98) {
if (Cs < 229) {
return 4;
}
else if (Cs >= 229) {
return 1;
}
}
}
}
else if (ssDCval >= 40) {
return 1;
}
}
else if (ssDCval >= 43) {
if (Rs < 222) {
return 8;
}
else if (Rs >= 222) {
return 4;
}
}
}
else if (Cs >= 256) {
if (ssDCval < 37) {
return 8;
}
else if (ssDCval >= 37) {
if (mcTcor < 97) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
return 1;
}
}
else if (mcTcor >= 97) {
return 1;
}
}
}
}
else if (tcor >= 101) {
return 8;
}
}
else if (posBalance >= 37) {
if (Rs < 229) {
return 8;
}
else if (Rs >= 229) {
return 1;
}
}
}
else if (MVDiff >= 6) {
if (refDCval < 37) {
if (Cs < 174) {
return 1;
}
else if (Cs >= 174) {
return 8;
}
}
else if (refDCval >= 37) {
return 1;
}
}
}
}
else if (negBalance >= 30) {
if (negBalance < 44) {
if (gchDC < 2) {
if (ssDCval < 41) {
if (RsDiff < 2) {
if (posBalance < 64) {
if (Cs < 306) {
return 8;
}
else if (Cs >= 306) {
if (Rs < 193) {
return 1;
}
else if (Rs >= 193) {
return 8;
}
}
}
else if (posBalance >= 64) {
return 4;
}
}
else if (RsDiff >= 2) {
return 1;
}
}
else if (ssDCval >= 41) {
if (mu_mv_mag_sq < 100) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 8;
}
}
else if (mu_mv_mag_sq >= 100) {
return 8;
}
}
}
else if (gchDC >= 2) {
if (Rs < 53) {
return 8;
}
else if (Rs >= 53) {
if (negBalance < 42) {
if (Rs < 265) {
return 1;
}
else if (Rs >= 265) {
return 4;
}
}
else if (negBalance >= 42) {
return 8;
}
}
}
}
else if (negBalance >= 44) {
if (mu_mv_mag_sq < 90) {
if (negBalance < 51) {
if (AFD < 7) {
if (Cs < 31) {
return 1;
}
else if (Cs >= 31) {
if (Rs < 144) {
if (Rs < 84) {
return 8;
}
else if (Rs >= 84) {
return 2;
}
}
else if (Rs >= 144) {
return 8;
}
}
}
else if (AFD >= 7) {
if (MVDiff < 55) {
return 4;
}
else if (MVDiff >= 55) {
return 1;
}
}
}
else if (negBalance >= 51) {
if (AFD < 14) {
if (MVDiff < 6) {
return 4;
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (AFD >= 14) {
if (AFD < 17) {
return 2;
}
else if (AFD >= 17) {
if (Rs < 25) {
return 1;
}
else if (Rs >= 25) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 90) {
if (MVDiff < 108) {
if (Cs < 196) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (TSC < 2) {
return 8;
}
else if (TSC >= 2) {
return 4;
}
}
}
else if (Cs >= 196) {
return 8;
}
}
else if (MVDiff >= 108) {
if (Rs < 16) {
return 2;
}
else if (Rs >= 16) {
return 1;
}
}
}
}
}
}
}
else if (ssDCval >= 47) {
if (TSCindex < 2) {
if (refDCval < 63) {
if (Rs < 80) {
if (posBalance < 37) {
if (Rs < 22) {
return 2;
}
else if (Rs >= 22) {
return 8;
}
}
else if (posBalance >= 37) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
return 1;
}
}
}
else if (Rs >= 80) {
if (diffMVdiffVal < -4) {
if (Rs < 230) {
return 8;
}
else if (Rs >= 230) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
if (Rs < 140) {
if (posBalance < 28) {
return 8;
}
else if (posBalance >= 28) {
return 4;
}
}
else if (Rs >= 140) {
if (SC < 342) {
if (Rs < 146) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Rs >= 146) {
if (mu_mv_mag_sq < 102) {
return 4;
}
else if (mu_mv_mag_sq >= 102) {
if (posBalance < 39) {
if (Rs < 147) {
return 4;
}
else if (Rs >= 147) {
return 1;
}
}
else if (posBalance >= 39) {
return 1;
}
}
}
}
else if (SC >= 342) {
if (SC < 563) {
if (diffAFD < -1) {
if (Rs < 210) {
return 1;
}
else if (Rs >= 210) {
return 4;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (SC >= 563) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (Cs < 360) {
return 1;
}
else if (Cs >= 360) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 63) {
if (ssDCval < 93) {
if (Cs < 259) {
if (SC < 256) {
if (Rs < 87) {
if (Rs < 61) {
if (Rs < 36) {
return 2;
}
else if (Rs >= 36) {
return 1;
}
}
else if (Rs >= 61) {
return 4;
}
}
else if (Rs >= 87) {
if (ssDCval < 75) {
if (Rs < 96) {
return 1;
}
else if (Rs >= 96) {
return 2;
}
}
else if (ssDCval >= 75) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
if (Cs < 106) {
return 2;
}
else if (Cs >= 106) {
return 1;
}
}
}
}
}
else if (SC >= 256) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
if (Cs < 130) {
if (Rs < 178) {
return 2;
}
else if (Rs >= 178) {
return 8;
}
}
else if (Cs >= 130) {
if (ssDCval < 79) {
return 2;
}
else if (ssDCval >= 79) {
if (tcor < 99) {
return 8;
}
else if (tcor >= 99) {
return 2;
}
}
}
}
}
}
else if (Cs >= 259) {
if (Rs < 159) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
if (Rs < 124) {
return 4;
}
else if (Rs >= 124) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 1;
}
}
}
}
else if (Rs >= 159) {
if (ssDCval < 78) {
if (Rs < 184) {
if (Cs < 286) {
return 1;
}
else if (Cs >= 286) {
return 4;
}
}
else if (Rs >= 184) {
return 1;
}
}
else if (ssDCval >= 78) {
return 4;
}
}
}
}
else if (ssDCval >= 93) {
if (Cs < 139) {
if (Rs < 78) {
if (Rs < 5) {
return 1;
}
else if (Rs >= 5) {
return 4;
}
}
else if (Rs >= 78) {
if (MVDiff < 33) {
return 1;
}
else if (MVDiff >= 33) {
return 8;
}
}
}
else if (Cs >= 139) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
return 2;
}
}
}
}
}
else if (TSCindex >= 2) {
if (refDCval < 72) {
if (posBalance < 32) {
if (negBalance < 19) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (Rs < 132) {
return 8;
}
else if (Rs >= 132) {
return 4;
}
}
}
else if (negBalance >= 19) {
if (mu_mv_mag_sq < 26) {
if (gchDC < 5) {
if (Rs < 21) {
return 4;
}
else if (Rs >= 21) {
return 1;
}
}
else if (gchDC >= 5) {
return 2;
}
}
else if (mu_mv_mag_sq >= 26) {
if (MVDiff < 10) {
if (ssDCval < 56) {
if (SC < 236) {
return 8;
}
else if (SC >= 236) {
return 4;
}
}
else if (ssDCval >= 56) {
if (ssDCval < 65) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (Rs < 224) {
return 1;
}
else if (Rs >= 224) {
return 8;
}
}
}
else if (ssDCval >= 65) {
return 1;
}
}
}
else if (MVDiff >= 10) {
if (Cs < 17) {
if (Rs < 101) {
return 2;
}
else if (Rs >= 101) {
return 8;
}
}
else if (Cs >= 17) {
return 8;
}
}
}
}
}
else if (posBalance >= 32) {
if (CsDiff < 1) {
if (SC < 367) {
if (MVDiff < 3) {
if (refDCval < 68) {
if (refDCval < 63) {
if (mcTcor < 96) {
if (Rs < 107) {
if (Rs < 79) {
return 2;
}
else if (Rs >= 79) {
return 1;
}
}
else if (Rs >= 107) {
return 8;
}
}
else if (mcTcor >= 96) {
if (ssDCval < 51) {
if (Rs < 140) {
return 8;
}
else if (Rs >= 140) {
return 4;
}
}
else if (ssDCval >= 51) {
if (negBalance < 22) {
if (Rs < 140) {
return 4;
}
else if (Rs >= 140) {
return 1;
}
}
else if (negBalance >= 22) {
return 1;
}
}
}
}
else if (refDCval >= 63) {
if (Cs < 123) {
return 1;
}
else if (Cs >= 123) {
return 2;
}
}
}
else if (refDCval >= 68) {
return 2;
}
}
else if (MVDiff >= 3) {
if (mcTcor < 95) {
if (gchDC < 2) {
if (mu_mv_mag_sq < 24) {
if (Cs < 121) {
return 2;
}
else if (Cs >= 121) {
return 1;
}
}
else if (mu_mv_mag_sq >= 24) {
if (diffRsCsdiff < 1) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 8;
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 119) {
return 2;
}
else if (Rs >= 119) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (Rs < 28) {
return 4;
}
else if (Rs >= 28) {
if (mu_mv_mag_sq < 54) {
if (mcTcor < 68) {
if (Rs < 79) {
return 8;
}
else if (Rs >= 79) {
return 2;
}
}
else if (mcTcor >= 68) {
if (SC < 108) {
if (Rs < 34) {
return 1;
}
else if (Rs >= 34) {
return 2;
}
}
else if (SC >= 108) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (diffTSC < 2) {
if (refDCval < 59) {
return 4;
}
else if (refDCval >= 59) {
return 8;
}
}
else if (diffTSC >= 2) {
if (Cs < 45) {
return 1;
}
else if (Cs >= 45) {
if (Rs < 132) {
return 8;
}
else if (Rs >= 132) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 95) {
if (Rs < 65) {
return 1;
}
else if (Rs >= 65) {
if (MVDiff < 13) {
if (Cs < 195) {
if (negBalance < 22) {
return 4;
}
else if (negBalance >= 22) {
if (diffRsCsdiff < -1) {
if (Rs < 185) {
if (Rs < 126) {
return 8;
}
else if (Rs >= 126) {
return 1;
}
}
else if (Rs >= 185) {
return 4;
}
}
else if (diffRsCsdiff >= -1) {
if (Rs < 144) {
if (posBalance < 50) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
if (ssDCval < 60) {
return 1;
}
else if (ssDCval >= 60) {
return 4;
}
}
}
}
else if (posBalance >= 50) {
if (ssDCval < 51) {
return 2;
}
else if (ssDCval >= 51) {
return 1;
}
}
}
else if (Rs >= 144) {
if (MVDiff < 10) {
return 1;
}
else if (MVDiff >= 10) {
if (posBalance < 50) {
return 1;
}
else if (posBalance >= 50) {
return 4;
}
}
}
}
}
}
else if (Cs >= 195) {
if (posBalance < 53) {
if (ssDCval < 65) {
return 4;
}
else if (ssDCval >= 65) {
return 1;
}
}
else if (posBalance >= 53) {
return 1;
}
}
}
else if (MVDiff >= 13) {
if (Cs < 127) {
if (AFD < 9) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 4;
}
}
else if (AFD >= 9) {
return 1;
}
}
else if (Cs >= 127) {
if (TSC < 3) {
if (diffMVdiffVal < 4) {
if (diffMVdiffVal < -3) {
if (Rs < 85) {
return 4;
}
else if (Rs >= 85) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
return 1;
}
}
else if (diffMVdiffVal >= 4) {
return 2;
}
}
else if (TSC >= 3) {
return 8;
}
}
}
}
}
}
}
else if (SC >= 367) {
if (negBalance < 31) {
if (refDCval < 60) {
if (mu_mv_mag_sq < 117) {
if (ssDCval < 57) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
if (mu_mv_mag_sq < 115) {
return 1;
}
else if (mu_mv_mag_sq >= 115) {
return 4;
}
}
}
else if (ssDCval >= 57) {
if (Cs < 212) {
return 1;
}
else if (Cs >= 212) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 117) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
if (negBalance < 24) {
if (Rs < 258) {
return 4;
}
else if (Rs >= 258) {
return 1;
}
}
else if (negBalance >= 24) {
return 1;
}
}
}
}
else if (refDCval >= 60) {
if (TSC < 2) {
if (mu_mv_mag_sq < 82) {
if (Rs < 166) {
return 2;
}
else if (Rs >= 166) {
return 4;
}
}
else if (mu_mv_mag_sq >= 82) {
if (Rs < 241) {
if (mu_mv_mag_sq < 108) {
return 4;
}
else if (mu_mv_mag_sq >= 108) {
if (Cs < 374) {
if (posBalance < 47) {
if (diffMVdiffVal < -1) {
if (Cs < 345) {
return 4;
}
else if (Cs >= 345) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
return 4;
}
}
else if (posBalance >= 47) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
if (Rs < 113) {
return 4;
}
else if (Rs >= 113) {
return 1;
}
}
}
}
else if (Cs >= 374) {
return 1;
}
}
}
else if (Rs >= 241) {
if (Cs < 218) {
return 2;
}
else if (Cs >= 218) {
if (Rs < 245) {
return 1;
}
else if (Rs >= 245) {
return 4;
}
}
}
}
}
else if (TSC >= 2) {
if (MVDiff < 6) {
if (SC < 377) {
if (Rs < 197) {
return 1;
}
else if (Rs >= 197) {
return 2;
}
}
else if (SC >= 377) {
if (mu_mv_mag_sq < 126) {
if (Rs < 242) {
if (posBalance < 51) {
if (mcTcor < 97) {
return 1;
}
else if (mcTcor >= 97) {
if (Cs < 219) {
if (Cs < 176) {
return 4;
}
else if (Cs >= 176) {
return 1;
}
}
else if (Cs >= 219) {
if (ssDCval < 68) {
if (mu_mv_mag_sq < 123) {
if (tcor < 99) {
return 4;
}
else if (tcor >= 99) {
if (Rs < 183) {
return 4;
}
else if (Rs >= 183) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 180) {
return 4;
}
else if (Rs >= 180) {
return 1;
}
}
}
else if (ssDCval >= 68) {
if (mcTcor < 99) {
if (negBalance < 26) {
if (Cs < 348) {
if (Rs < 163) {
return 1;
}
else if (Rs >= 163) {
return 4;
}
}
else if (Cs >= 348) {
return 1;
}
}
else if (negBalance >= 26) {
if (mu_mv_mag_sq < 102) {
if (Rs < 122) {
return 4;
}
else if (Rs >= 122) {
return 1;
}
}
else if (mu_mv_mag_sq >= 102) {
if (negBalance < 28) {
if (posBalance < 44) {
return 1;
}
else if (posBalance >= 44) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (Rs < 197) {
return 1;
}
else if (Rs >= 197) {
return 4;
}
}
}
}
else if (negBalance >= 28) {
return 4;
}
}
}
}
else if (mcTcor >= 99) {
if (Cs < 355) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (Cs >= 355) {
if (TSCindex < 3) {
return 1;
}
else if (TSCindex >= 3) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 51) {
if (posBalance < 57) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
if (Rs < 228) {
return 1;
}
else if (Rs >= 228) {
return 4;
}
}
}
else if (posBalance >= 57) {
return 1;
}
}
}
else if (Rs >= 242) {
return 1;
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 280) {
return 1;
}
else if (Cs >= 280) {
return 4;
}
}
}
}
else if (MVDiff >= 6) {
if (Rs < 250) {
return 1;
}
else if (Rs >= 250) {
if (Rs < 278) {
return 4;
}
else if (Rs >= 278) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 31) {
if (mu_mv_mag_sq < 85) {
if (ssDCval < 63) {
if (mu_mv_mag_sq < 14) {
if (Cs < 223) {
return 1;
}
else if (Cs >= 223) {
return 4;
}
}
else if (mu_mv_mag_sq >= 14) {
if (Cs < 367) {
return 4;
}
else if (Cs >= 367) {
return 1;
}
}
}
else if (ssDCval >= 63) {
if (Rs < 125) {
return 4;
}
else if (Rs >= 125) {
if (posBalance < 46) {
if (Cs < 287) {
return 1;
}
else if (Cs >= 287) {
return 4;
}
}
else if (posBalance >= 46) {
if (tcor < 95) {
return 4;
}
else if (tcor >= 95) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (posBalance < 44) {
if (Cs < 427) {
return 4;
}
else if (Cs >= 427) {
return 1;
}
}
else if (posBalance >= 44) {
if (tcor < 98) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (Rs < 275) {
return 1;
}
else if (Rs >= 275) {
return 4;
}
}
}
else if (tcor >= 98) {
return 1;
}
}
}
}
}
}
else if (CsDiff >= 1) {
if (TSC < 4) {
if (Rs < 50) {
if (Cs < 88) {
return 2;
}
else if (Cs >= 88) {
return 1;
}
}
else if (Rs >= 50) {
if (MVDiff < 2) {
if (SC < 611) {
if (Rs < 183) {
if (Cs < 398) {
if (Cs < 338) {
if (Rs < 164) {
return 1;
}
else if (Rs >= 164) {
return 4;
}
}
else if (Cs >= 338) {
return 1;
}
}
else if (Cs >= 398) {
return 4;
}
}
else if (Rs >= 183) {
if (Rs < 241) {
if (Cs < 388) {
return 4;
}
else if (Cs >= 388) {
if (Rs < 208) {
return 1;
}
else if (Rs >= 208) {
return 4;
}
}
}
else if (Rs >= 241) {
return 1;
}
}
}
else if (SC >= 611) {
return 1;
}
}
else if (MVDiff >= 2) {
if (mcTcor < 92) {
if (refDCval < 60) {
if (Rs < 131) {
return 4;
}
else if (Rs >= 131) {
return 8;
}
}
else if (refDCval >= 60) {
return 1;
}
}
else if (mcTcor >= 92) {
if (Rs < 263) {
if (Cs < 326) {
if (refDCval < 53) {
if (tcor < 81) {
if (Rs < 178) {
return 2;
}
else if (Rs >= 178) {
return 4;
}
}
else if (tcor >= 81) {
return 1;
}
}
else if (refDCval >= 53) {
if (posBalance < 38) {
if (Cs < 145) {
return 8;
}
else if (Cs >= 145) {
return 1;
}
}
else if (posBalance >= 38) {
if (refDCval < 62) {
if (ssDCval < 64) {
if (mcTcor < 99) {
return 4;
}
else if (mcTcor >= 99) {
if (Rs < 129) {
return 4;
}
else if (Rs >= 129) {
return 1;
}
}
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (refDCval >= 62) {
if (negBalance < 40) {
if (Cs < 210) {
if (posBalance < 57) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 1;
}
}
else if (posBalance >= 57) {
if (Rs < 146) {
return 4;
}
else if (Rs >= 146) {
return 8;
}
}
}
else if (Cs >= 210) {
if (Rs < 233) {
if (Cs < 317) {
if (SC < 362) {
if (Rs < 133) {
return 4;
}
else if (Rs >= 133) {
return 1;
}
}
else if (SC >= 362) {
return 4;
}
}
else if (Cs >= 317) {
return 1;
}
}
else if (Rs >= 233) {
return 1;
}
}
}
else if (negBalance >= 40) {
if (ssDCval < 62) {
return 1;
}
else if (ssDCval >= 62) {
if (Cs < 109) {
return 1;
}
else if (Cs >= 109) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
return 1;
}
}
}
}
}
}
}
}
else if (Cs >= 326) {
if (negBalance < 51) {
if (MVDiff < 3) {
if (Cs < 453) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (Cs >= 453) {
return 1;
}
}
else if (MVDiff >= 3) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
if (posBalance < 53) {
return 4;
}
else if (posBalance >= 53) {
if (Rs < 227) {
return 4;
}
else if (Rs >= 227) {
return 8;
}
}
}
}
}
else if (negBalance >= 51) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
return 4;
}
}
}
}
}
else if (Rs >= 263) {
if (MVDiff < 5) {
if (Cs < 337) {
return 1;
}
else if (Cs >= 337) {
if (Rs < 274) {
return 8;
}
else if (Rs >= 274) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (posBalance < 60) {
if (diffRsCsdiff < -2) {
if (diffMVdiffVal < 1) {
return 2;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 281) {
return 2;
}
else if (Rs >= 281) {
return 4;
}
}
}
else if (diffRsCsdiff >= -2) {
if (diffAFD < -1) {
if (Rs < 281) {
if (Cs < 451) {
return 1;
}
else if (Cs >= 451) {
return 2;
}
}
else if (Rs >= 281) {
return 4;
}
}
else if (diffAFD >= -1) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 2;
}
}
}
}
else if (posBalance >= 60) {
if (Cs < 366) {
return 2;
}
else if (Cs >= 366) {
return 1;
}
}
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 26) {
if (tcor < 77) {
if (refDCval < 66) {
if (ssDCval < 58) {
if (AFD < 14) {
if (mcTcor < 93) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 4;
}
}
else if (mcTcor >= 93) {
return 4;
}
}
else if (AFD >= 14) {
if (posBalance < 57) {
if (Cs < 198) {
return 1;
}
else if (Cs >= 198) {
if (Rs < 193) {
return 2;
}
else if (Rs >= 193) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
return 4;
}
}
}
}
else if (ssDCval >= 58) {
if (diffTSC < -1) {
if (MVDiff < 143) {
if (Cs < 374) {
return 1;
}
else if (Cs >= 374) {
return 8;
}
}
else if (MVDiff >= 143) {
return 8;
}
}
else if (diffTSC >= -1) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
return 1;
}
}
}
}
else if (refDCval >= 66) {
if (ssDCval < 67) {
return 1;
}
else if (ssDCval >= 67) {
if (Cs < 408) {
if (diffMVdiffVal < -43) {
return 4;
}
else if (diffMVdiffVal >= -43) {
return 2;
}
}
else if (Cs >= 408) {
return 4;
}
}
}
}
else if (tcor >= 77) {
if (MVDiff < 47) {
if (MVDiff < 13) {
if (AFD < 13) {
if (Rs < 209) {
return 1;
}
else if (Rs >= 209) {
return 8;
}
}
else if (AFD >= 13) {
return 2;
}
}
else if (MVDiff >= 13) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 1;
}
}
}
else if (MVDiff >= 47) {
if (diffTSC < 1) {
if (CsDiff < 3) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 2;
}
}
else if (CsDiff >= 3) {
if (Rs < 204) {
return 8;
}
else if (Rs >= 204) {
return 1;
}
}
}
else if (diffTSC >= 1) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (diffAFD < 6) {
if (Rs < 199) {
if (posBalance < 50) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
return 4;
}
}
else if (posBalance >= 50) {
if (Rs < 56) {
if (Rs < 39) {
return 1;
}
else if (Rs >= 39) {
return 8;
}
}
else if (Rs >= 56) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
if (negBalance < 50) {
return 4;
}
else if (negBalance >= 50) {
return 1;
}
}
}
}
}
else if (Rs >= 199) {
if (mcTcor < 90) {
if (negBalance < 51) {
if (Rs < 275) {
if (Cs < 144) {
return 4;
}
else if (Cs >= 144) {
return 2;
}
}
else if (Rs >= 275) {
return 8;
}
}
else if (negBalance >= 51) {
if (Cs < 262) {
return 1;
}
else if (Cs >= 262) {
return 4;
}
}
}
else if (mcTcor >= 90) {
if (ssDCval < 67) {
if (diffMVdiffVal < -17) {
return 8;
}
else if (diffMVdiffVal >= -17) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
return 4;
}
}
}
else if (ssDCval >= 67) {
return 4;
}
}
}
}
else if (diffAFD >= 6) {
if (TSC < 8) {
if (Rs < 250) {
return 8;
}
else if (Rs >= 250) {
return 1;
}
}
else if (TSC >= 8) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 72) {
if (Rs < 176) {
if (Cs < 215) {
if (Rs < 128) {
if (ssDCval < 108) {
if (Cs < 88) {
if (TSC < 3) {
if (SC < 2) {
return 4;
}
else if (SC >= 2) {
return 1;
}
}
else if (TSC >= 3) {
if (MVDiff < 143) {
if (mu_mv_mag_sq < 36) {
if (gchDC < 1) {
if (Cs < 58) {
return 2;
}
else if (Cs >= 58) {
if (Rs < 101) {
return 1;
}
else if (Rs >= 101) {
return 2;
}
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (mu_mv_mag_sq >= 36) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
if (Rs < 113) {
return 2;
}
else if (Rs >= 113) {
return 1;
}
}
}
}
else if (MVDiff >= 143) {
if (diffAFD < 2) {
if (mu_mv_mag_sq < 5) {
if (Rs < 36) {
return 1;
}
else if (Rs >= 36) {
return 4;
}
}
else if (mu_mv_mag_sq >= 5) {
if (mcTcor < 71) {
return 4;
}
else if (mcTcor >= 71) {
if (Cs < 72) {
return 2;
}
else if (Cs >= 72) {
return 4;
}
}
}
}
else if (diffAFD >= 2) {
return 1;
}
}
}
}
else if (Cs >= 88) {
if (Rs < 56) {
if (MVDiff < 20) {
return 4;
}
else if (MVDiff >= 20) {
if (Rs < 20) {
return 2;
}
else if (Rs >= 20) {
if (Cs < 95) {
return 8;
}
else if (Cs >= 95) {
return 1;
}
}
}
}
else if (Rs >= 56) {
if (MVDiff < 3) {
if (Rs < 97) {
return 4;
}
else if (Rs >= 97) {
return 8;
}
}
else if (MVDiff >= 3) {
if (refDCval < 82) {
if (negBalance < 44) {
if (mu_mv_mag_sq < 72) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
return 2;
}
}
else if (mu_mv_mag_sq >= 72) {
if (ssDCval < 75) {
return 2;
}
else if (ssDCval >= 75) {
if (Cs < 97) {
return 2;
}
else if (Cs >= 97) {
return 1;
}
}
}
}
else if (negBalance >= 44) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
if (AFD < 33) {
if (ssDCval < 75) {
if (refDCval < 77) {
return 1;
}
else if (refDCval >= 77) {
return 2;
}
}
else if (ssDCval >= 75) {
return 2;
}
}
else if (AFD >= 33) {
if (Rs < 81) {
return 4;
}
else if (Rs >= 81) {
return 8;
}
}
}
}
}
else if (refDCval >= 82) {
if (refDCval < 103) {
if (SC < 286) {
return 1;
}
else if (SC >= 286) {
if (mu_mv_mag_sq < 65) {
if (Cs < 208) {
return 2;
}
else if (Cs >= 208) {
return 4;
}
}
else if (mu_mv_mag_sq >= 65) {
return 1;
}
}
}
else if (refDCval >= 103) {
if (Rs < 91) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (Rs >= 91) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 108) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Cs < 2) {
if (posBalance < 41) {
return 1;
}
else if (posBalance >= 41) {
if (refDCval < 166) {
return 8;
}
else if (refDCval >= 166) {
return 1;
}
}
}
else if (Cs >= 2) {
if (posBalance < 72) {
if (Cs < 169) {
if (diffMVdiffVal < -43) {
if (mu_mv_mag_sq < 21) {
return 2;
}
else if (mu_mv_mag_sq >= 21) {
return 1;
}
}
else if (diffMVdiffVal >= -43) {
if (CsDiff < 2) {
if (Rs < 1) {
return 1;
}
else if (Rs >= 1) {
if (SC < 255) {
return 2;
}
else if (SC >= 255) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (Rs < 82) {
return 1;
}
else if (Rs >= 82) {
return 8;
}
}
}
}
else if (Cs >= 169) {
if (Cs < 211) {
return 1;
}
else if (Cs >= 211) {
return 8;
}
}
}
else if (posBalance >= 72) {
if (Rs < 73) {
if (Rs < 19) {
return 2;
}
else if (Rs >= 19) {
return 8;
}
}
else if (Rs >= 73) {
return 4;
}
}
}
}
}
}
else if (Rs >= 128) {
if (refDCval < 75) {
if (mcTcor < 98) {
if (AFD < 5) {
if (ssDCval < 74) {
if (Rs < 151) {
return 8;
}
else if (Rs >= 151) {
return 2;
}
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (AFD >= 5) {
if (Rs < 140) {
return 1;
}
else if (Rs >= 140) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (Cs < 127) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
return 8;
}
}
else if (Cs >= 127) {
if (MVDiff < 5) {
return 2;
}
else if (MVDiff >= 5) {
return 1;
}
}
}
}
else if (refDCval >= 75) {
if (refDCval < 94) {
if (tcor < 76) {
if (TSC < 10) {
if (mcTcor < 84) {
return 8;
}
else if (mcTcor >= 84) {
if (refDCval < 89) {
if (Rs < 172) {
return 2;
}
else if (Rs >= 172) {
if (Rs < 175) {
return 8;
}
else if (Rs >= 175) {
return 2;
}
}
}
else if (refDCval >= 89) {
if (diffMVdiffVal < 2) {
if (MVDiff < 19) {
return 8;
}
else if (MVDiff >= 19) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
}
else if (diffMVdiffVal >= 2) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (AFD < 33) {
return 4;
}
else if (AFD >= 33) {
if (Rs < 166) {
return 1;
}
else if (Rs >= 166) {
return 8;
}
}
}
}
else if (tcor >= 76) {
if (MVDiff < 41) {
if (SC < 355) {
if (tcor < 100) {
if (MVDiff < 9) {
if (Rs < 145) {
return 1;
}
else if (Rs >= 145) {
if (Cs < 61) {
if (Rs < 163) {
return 8;
}
else if (Rs >= 163) {
return 1;
}
}
else if (Cs >= 61) {
return 2;
}
}
}
else if (MVDiff >= 9) {
return 1;
}
}
else if (tcor >= 100) {
return 2;
}
}
else if (SC >= 355) {
return 2;
}
}
else if (MVDiff >= 41) {
return 8;
}
}
}
else if (refDCval >= 94) {
if (Cs < 41) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 1;
}
}
else if (Cs >= 41) {
if (AFD < 2) {
return 4;
}
else if (AFD >= 2) {
if (diffMVdiffVal < 12) {
if (diffMVdiffVal < -21) {
if (Rs < 148) {
return 8;
}
else if (Rs >= 148) {
return 4;
}
}
else if (diffMVdiffVal >= -21) {
if (mu_mv_mag_sq < 57) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
if (Rs < 158) {
return 1;
}
else if (Rs >= 158) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 57) {
if (Rs < 146) {
return 1;
}
else if (Rs >= 146) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= 12) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 215) {
if (tcor < 91) {
if (refDCval < 94) {
if (posBalance < 51) {
if (Rs < 101) {
return 4;
}
else if (Rs >= 101) {
return 1;
}
}
else if (posBalance >= 51) {
if (posBalance < 66) {
if (TSC < 3) {
if (Cs < 219) {
return 8;
}
else if (Cs >= 219) {
if (negBalance < 49) {
return 2;
}
else if (negBalance >= 49) {
if (Rs < 160) {
return 8;
}
else if (Rs >= 160) {
return 2;
}
}
}
}
else if (TSC >= 3) {
return 8;
}
}
else if (posBalance >= 66) {
if (Cs < 260) {
return 4;
}
else if (Cs >= 260) {
if (RsDiff < 2) {
return 2;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
}
}
else if (refDCval >= 94) {
if (gchDC < 4) {
if (tcor < 88) {
return 2;
}
else if (tcor >= 88) {
if (Cs < 337) {
return 2;
}
else if (Cs >= 337) {
return 1;
}
}
}
else if (gchDC >= 4) {
if (ssDCval < 96) {
if (Rs < 113) {
return 2;
}
else if (Rs >= 113) {
return 8;
}
}
else if (ssDCval >= 96) {
return 1;
}
}
}
}
else if (tcor >= 91) {
if (refDCval < 115) {
if (refDCval < 75) {
if (AFD < 5) {
if (Cs < 230) {
return 2;
}
else if (Cs >= 230) {
if (tcor < 98) {
if (Rs < 132) {
return 4;
}
else if (Rs >= 132) {
return 1;
}
}
else if (tcor >= 98) {
return 1;
}
}
}
else if (AFD >= 5) {
return 4;
}
}
else if (refDCval >= 75) {
if (gchDC < 2) {
if (Cs < 363) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
}
else if (Cs >= 363) {
return 1;
}
}
else if (gchDC >= 2) {
if (Rs < 87) {
return 1;
}
else if (Rs >= 87) {
return 2;
}
}
}
}
else if (refDCval >= 115) {
return 2;
}
}
}
}
else if (Rs >= 176) {
if (refDCval < 76) {
if (Cs < 210) {
if (MVDiff < 3) {
if (TSCindex < 3) {
return 2;
}
else if (TSCindex >= 3) {
if (negBalance < 29) {
return 8;
}
else if (negBalance >= 29) {
if (CsDiff < 1) {
return 2;
}
else if (CsDiff >= 1) {
return 1;
}
}
}
}
else if (MVDiff >= 3) {
if (negBalance < 24) {
return 2;
}
else if (negBalance >= 24) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
if (Rs < 229) {
return 4;
}
else if (Rs >= 229) {
return 1;
}
}
}
}
}
else if (Cs >= 210) {
if (tcor < 96) {
if (mu_mv_mag_sq < 72) {
if (AFD < 19) {
if (Cs < 353) {
return 1;
}
else if (Cs >= 353) {
if (Rs < 214) {
return 4;
}
else if (Rs >= 214) {
return 1;
}
}
}
else if (AFD >= 19) {
if (Rs < 224) {
return 8;
}
else if (Rs >= 224) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 72) {
return 8;
}
}
else if (tcor >= 96) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 4;
}
}
}
}
else if (refDCval >= 76) {
if (Cs < 71) {
if (MVDiff < 43) {
if (TSC < 4) {
if (Cs < 45) {
return 2;
}
else if (Cs >= 45) {
return 8;
}
}
else if (TSC >= 4) {
if (gchDC < 5) {
return 1;
}
else if (gchDC >= 5) {
if (Rs < 213) {
return 1;
}
else if (Rs >= 213) {
if (Rs < 221) {
return 8;
}
else if (Rs >= 221) {
return 2;
}
}
}
}
}
else if (MVDiff >= 43) {
if (Cs < 28) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (TSC < 5) {
if (diffMVdiffVal < -1) {
return 8;
}
else if (diffMVdiffVal >= -1) {
return 2;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (Cs >= 28) {
if (mu_mv_mag_sq < 31) {
return 1;
}
else if (mu_mv_mag_sq >= 31) {
return 2;
}
}
}
}
else if (Cs >= 71) {
if (posBalance < 27) {
return 8;
}
else if (posBalance >= 27) {
if (mcTcor < 97) {
if (MVDiff < 19) {
if (AFD < 11) {
if (TSC < 5) {
if (refDCval < 130) {
if (diffAFD < 1) {
if (negBalance < 50) {
if (TSCindex < 3) {
return 2;
}
else if (TSCindex >= 3) {
if (posBalance < 57) {
if (Cs < 283) {
return 8;
}
else if (Cs >= 283) {
if (Rs < 277) {
return 2;
}
else if (Rs >= 277) {
return 8;
}
}
}
else if (posBalance >= 57) {
if (Cs < 362) {
if (mcTcor < 85) {
return 2;
}
else if (mcTcor >= 85) {
if (Rs < 190) {
return 8;
}
else if (Rs >= 190) {
if (Rs < 191) {
return 2;
}
else if (Rs >= 191) {
if (RsCsDiff < 2) {
if (posBalance < 58) {
if (Rs < 212) {
return 8;
}
else if (Rs >= 212) {
return 2;
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (RsCsDiff >= 2) {
if (Rs < 213) {
return 8;
}
else if (Rs >= 213) {
return 2;
}
}
}
}
}
}
else if (Cs >= 362) {
return 8;
}
}
}
}
else if (negBalance >= 50) {
return 2;
}
}
else if (diffAFD >= 1) {
if (Cs < 317) {
return 1;
}
else if (Cs >= 317) {
return 8;
}
}
}
else if (refDCval >= 130) {
return 8;
}
}
else if (TSC >= 5) {
return 8;
}
}
else if (AFD >= 11) {
if (Cs < 330) {
return 1;
}
else if (Cs >= 330) {
if (Rs < 233) {
return 8;
}
else if (Rs >= 233) {
return 2;
}
}
}
}
else if (MVDiff >= 19) {
if (Rs < 286) {
if (diffTSC < 2) {
if (MVDiff < 24) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
return 1;
}
}
}
else if (MVDiff >= 24) {
if (diffMVdiffVal < 16) {
if (mu_mv_mag_sq < 64) {
if (posBalance < 67) {
if (ssDCval < 132) {
return 2;
}
else if (ssDCval >= 132) {
if (Rs < 213) {
return 4;
}
else if (Rs >= 213) {
return 2;
}
}
}
else if (posBalance >= 67) {
if (Rs < 232) {
return 8;
}
else if (Rs >= 232) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 64) {
return 8;
}
}
else if (diffMVdiffVal >= 16) {
if (negBalance < 49) {
if (Cs < 131) {
return 2;
}
else if (Cs >= 131) {
return 8;
}
}
else if (negBalance >= 49) {
if (Cs < 159) {
return 2;
}
else if (Cs >= 159) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 1;
}
}
}
}
}
}
else if (diffTSC >= 2) {
if (MVDiff < 155) {
return 1;
}
else if (MVDiff >= 155) {
return 4;
}
}
}
else if (Rs >= 286) {
if (posBalance < 62) {
return 2;
}
else if (posBalance >= 62) {
if (Rs < 291) {
return 4;
}
else if (Rs >= 291) {
return 2;
}
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 134) {
if (ssDCval < 100) {
if (mu_mv_mag_sq < 26) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 4;
}
}
else if (mu_mv_mag_sq >= 26) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
return 8;
}
}
}
else if (ssDCval >= 100) {
if (SC < 337) {
return 2;
}
else if (SC >= 337) {
return 1;
}
}
}
else if (Cs >= 134) {
if (SC < 421) {
if (Rs < 232) {
if (Rs < 230) {
return 2;
}
else if (Rs >= 230) {
if (Rs < 231) {
return 1;
}
else if (Rs >= 231) {
return 2;
}
}
}
else if (Rs >= 232) {
return 8;
}
}
else if (SC >= 421) {
if (AFD < 4) {
if (ssDCval < 81) {
if (Rs < 194) {
return 1;
}
else if (Rs >= 194) {
return 2;
}
}
else if (ssDCval >= 81) {
if (tcor < 99) {
if (posBalance < 57) {
return 4;
}
else if (posBalance >= 57) {
return 2;
}
}
else if (tcor >= 99) {
if (Rs < 230) {
return 4;
}
else if (Rs >= 230) {
return 1;
}
}
}
}
else if (AFD >= 4) {
if (RsDiff < 2) {
if (refDCval < 100) {
if (negBalance < 45) {
if (posBalance < 56) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (posBalance >= 56) {
if (Rs < 204) {
return 8;
}
else if (Rs >= 204) {
return 2;
}
}
}
else if (negBalance >= 45) {
return 2;
}
}
else if (refDCval >= 100) {
if (CsDiff < 1) {
if (Cs < 247) {
return 2;
}
else if (Cs >= 247) {
if (Rs < 274) {
return 4;
}
else if (Rs >= 274) {
return 8;
}
}
}
else if (CsDiff >= 1) {
return 2;
}
}
}
else if (RsDiff >= 2) {
if (ssDCval < 96) {
if (mu_mv_mag_sq < 8) {
return 1;
}
else if (mu_mv_mag_sq >= 8) {
if (Rs < 253) {
return 8;
}
else if (Rs >= 253) {
return 1;
}
}
}
else if (ssDCval >= 96) {
if (Rs < 277) {
if (Cs < 317) {
return 4;
}
else if (Cs >= 317) {
return 2;
}
}
else if (Rs >= 277) {
if (Cs < 305) {
return 2;
}
else if (Cs >= 305) {
return 1;
}
}
}
}
}
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 300) {
if (RsDiff < 4) {
if (refDCval < 74) {
if (Rs < 1295) {
if (mcTcor < 89) {
if (Cs < 155) {
if (MVDiff < 18) {
if (Cs < 45) {
return 8;
}
else if (Cs >= 45) {
if (Rs < 411) {
return 1;
}
else if (Rs >= 411) {
return 2;
}
}
}
else if (MVDiff >= 18) {
if (AFD < 12) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
return 8;
}
}
else if (AFD >= 12) {
if (Cs < 33) {
if (Rs < 414) {
return 2;
}
else if (Rs >= 414) {
return 1;
}
}
else if (Cs >= 33) {
return 8;
}
}
}
}
else if (Cs >= 155) {
if (refDCval < 43) {
if (tcor < 70) {
if (negBalance < 47) {
if (negBalance < 41) {
if (negBalance < 36) {
if (posBalance < 73) {
return 1;
}
else if (posBalance >= 73) {
if (Rs < 320) {
return 2;
}
else if (Rs >= 320) {
return 8;
}
}
}
else if (negBalance >= 36) {
if (Rs < 637) {
return 8;
}
else if (Rs >= 637) {
return 1;
}
}
}
else if (negBalance >= 41) {
return 2;
}
}
else if (negBalance >= 47) {
if (Cs < 306) {
return 8;
}
else if (Cs >= 306) {
if (tcor < 66) {
return 4;
}
else if (tcor >= 66) {
if (Rs < 468) {
return 1;
}
else if (Rs >= 468) {
return 8;
}
}
}
}
}
else if (tcor >= 70) {
if (negBalance < 29) {
if (posBalance < 61) {
if (Cs < 309) {
return 8;
}
else if (Cs >= 309) {
return 1;
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (negBalance >= 29) {
if (diffAFD < -3) {
return 2;
}
else if (diffAFD >= -3) {
return 8;
}
}
}
}
else if (refDCval >= 43) {
if (posBalance < 52) {
if (mcTcor < 84) {
if (SC < 730) {
return 8;
}
else if (SC >= 730) {
if (Rs < 348) {
return 8;
}
else if (Rs >= 348) {
return 2;
}
}
}
else if (mcTcor >= 84) {
if (posBalance < 46) {
if (Rs < 503) {
return 8;
}
else if (Rs >= 503) {
if (Rs < 637) {
return 4;
}
else if (Rs >= 637) {
if (Rs < 733) {
return 1;
}
else if (Rs >= 733) {
return 2;
}
}
}
}
else if (posBalance >= 46) {
if (Cs < 457) {
return 1;
}
else if (Cs >= 457) {
return 4;
}
}
}
}
else if (posBalance >= 52) {
if (mcTcor < 83) {
if (tcor < 59) {
if (AFD < 15) {
if (Cs < 186) {
return 8;
}
else if (Cs >= 186) {
return 4;
}
}
else if (AFD >= 15) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 4;
}
}
}
else if (tcor >= 59) {
if (diffMVdiffVal < 2) {
return 8;
}
else if (diffMVdiffVal >= 2) {
if (mcTcor < 78) {
return 8;
}
else if (mcTcor >= 78) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (Rs < 432) {
return 2;
}
else if (Rs >= 432) {
return 8;
}
}
}
}
}
}
else if (mcTcor >= 83) {
if (MVDiff < 36) {
if (diffMVdiffVal < -1) {
if (Cs < 467) {
return 4;
}
else if (Cs >= 467) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (posBalance < 60) {
return 8;
}
else if (posBalance >= 60) {
if (Rs < 495) {
return 4;
}
else if (Rs >= 495) {
if (Rs < 523) {
return 8;
}
else if (Rs >= 523) {
if (Rs < 592) {
return 4;
}
else if (Rs >= 592) {
return 8;
}
}
}
}
}
}
else if (MVDiff >= 36) {
if (MVDiff < 64) {
if (Rs < 481) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
return 4;
}
}
else if (Rs >= 481) {
return 2;
}
}
else if (MVDiff >= 64) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 89) {
if (posBalance < 27) {
if (tcor < 101) {
if (Cs < 404) {
if (Cs < 55) {
return 4;
}
else if (Cs >= 55) {
if (posBalance < 22) {
return 8;
}
else if (posBalance >= 22) {
if (ssDCval < 67) {
if (diffAFD < -2) {
if (Rs < 785) {
return 8;
}
else if (Rs >= 785) {
return 1;
}
}
else if (diffAFD >= -2) {
return 1;
}
}
else if (ssDCval >= 67) {
return 8;
}
}
}
}
else if (Cs >= 404) {
if (diffMVdiffVal < -1) {
if (Rs < 536) {
return 1;
}
else if (Rs >= 536) {
if (Rs < 822) {
return 8;
}
else if (Rs >= 822) {
return 2;
}
}
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
}
else if (tcor >= 101) {
if (posBalance < 2) {
return 4;
}
else if (posBalance >= 2) {
if (diffAFD < -3) {
if (Rs < 854) {
return 2;
}
else if (Rs >= 854) {
return 1;
}
}
else if (diffAFD >= -3) {
if (Cs < 13) {
return 4;
}
else if (Cs >= 13) {
if (mcTcor < 104) {
return 8;
}
else if (mcTcor >= 104) {
if (Cs < 296) {
return 8;
}
else if (Cs >= 296) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 27) {
if (TSC < 1) {
if (tcor < 100) {
if (Rs < 872) {
return 4;
}
else if (Rs >= 872) {
return 8;
}
}
else if (tcor >= 100) {
if (Rs < 846) {
return 8;
}
else if (Rs >= 846) {
return 1;
}
}
}
else if (TSC >= 1) {
if (ssDCval < 68) {
if (MVDiff < 2) {
if (mcTcor < 97) {
if (Rs < 418) {
if (mu_mv_mag_sq < 115) {
if (Cs < 447) {
return 1;
}
else if (Cs >= 447) {
return 8;
}
}
else if (mu_mv_mag_sq >= 115) {
if (Rs < 363) {
return 4;
}
else if (Rs >= 363) {
return 8;
}
}
}
else if (Rs >= 418) {
if (negBalance < 32) {
if (tcor < 89) {
return 4;
}
else if (tcor >= 89) {
if (Cs < 240) {
return 2;
}
else if (Cs >= 240) {
if (posBalance < 49) {
if (Cs < 343) {
return 8;
}
else if (Cs >= 343) {
return 1;
}
}
else if (posBalance >= 49) {
return 8;
}
}
}
}
else if (negBalance >= 32) {
return 8;
}
}
}
else if (mcTcor >= 97) {
if (negBalance < 40) {
if (Rs < 424) {
if (Cs < 278) {
if (Cs < 191) {
return 1;
}
else if (Cs >= 191) {
return 4;
}
}
else if (Cs >= 278) {
return 1;
}
}
else if (Rs >= 424) {
if (posBalance < 44) {
if (Cs < 407) {
if (Rs < 801) {
if (ssDCval < 52) {
return 8;
}
else if (ssDCval >= 52) {
if (Rs < 651) {
if (mu_mv_mag_sq < 104) {
return 8;
}
else if (mu_mv_mag_sq >= 104) {
if (Cs < 366) {
return 8;
}
else if (Cs >= 366) {
return 1;
}
}
}
else if (Rs >= 651) {
return 1;
}
}
}
else if (Rs >= 801) {
return 1;
}
}
else if (Cs >= 407) {
if (negBalance < 21) {
if (posBalance < 42) {
return 8;
}
else if (posBalance >= 42) {
if (ssDCval < 57) {
return 1;
}
else if (ssDCval >= 57) {
return 8;
}
}
}
else if (negBalance >= 21) {
if (Rs < 531) {
return 1;
}
else if (Rs >= 531) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 119) {
return 1;
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
else if (mcTcor >= 99) {
if (ssDCval < 54) {
return 1;
}
else if (ssDCval >= 54) {
return 8;
}
}
}
}
}
}
}
else if (posBalance >= 44) {
if (Cs < 323) {
if (ssDCval < 63) {
if (Rs < 612) {
return 8;
}
else if (Rs >= 612) {
return 2;
}
}
else if (ssDCval >= 63) {
return 1;
}
}
else if (Cs >= 323) {
if (Cs < 468) {
if (ssDCval < 67) {
if (ssDCval < 47) {
return 4;
}
else if (ssDCval >= 47) {
return 1;
}
}
else if (ssDCval >= 67) {
if (Rs < 476) {
return 2;
}
else if (Rs >= 476) {
return 1;
}
}
}
else if (Cs >= 468) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 40) {
if (posBalance < 40) {
return 4;
}
else if (posBalance >= 40) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
}
}
}
else if (MVDiff >= 2) {
if (Cs < 117) {
if (mu_mv_mag_sq < 49) {
if (tcor < 81) {
if (negBalance < 42) {
if (Rs < 943) {
return 8;
}
else if (Rs >= 943) {
return 2;
}
}
else if (negBalance >= 42) {
return 1;
}
}
else if (tcor >= 81) {
if (Rs < 981) {
if (tcor < 85) {
if (RsDiff < 3) {
if (posBalance < 53) {
if (Rs < 492) {
return 1;
}
else if (Rs >= 492) {
return 8;
}
}
else if (posBalance >= 53) {
return 1;
}
}
else if (RsDiff >= 3) {
return 1;
}
}
else if (tcor >= 85) {
return 1;
}
}
else if (Rs >= 981) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 49) {
if (mcTcor < 99) {
if (Rs < 574) {
if (diffMVdiffVal < 1) {
if (Cs < 93) {
return 1;
}
else if (Cs >= 93) {
return 2;
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 15) {
return 1;
}
else if (Cs >= 15) {
return 8;
}
}
}
else if (Rs >= 574) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
if (Rs < 753) {
return 1;
}
else if (Rs >= 753) {
return 4;
}
}
}
}
else if (mcTcor >= 99) {
return 1;
}
}
}
else if (Cs >= 117) {
if (ssDCval < 45) {
if (posBalance < 32) {
if (tcor < 97) {
return 2;
}
else if (tcor >= 97) {
return 1;
}
}
else if (posBalance >= 32) {
if (RsCsDiff < 1) {
if (posBalance < 49) {
if (Rs < 418) {
return 1;
}
else if (Rs >= 418) {
return 8;
}
}
else if (posBalance >= 49) {
return 8;
}
}
else if (RsCsDiff >= 1) {
if (negBalance < 35) {
if (mu_mv_mag_sq < 70) {
if (tcor < 90) {
if (Cs < 183) {
return 4;
}
else if (Cs >= 183) {
if (negBalance < 33) {
return 1;
}
else if (negBalance >= 33) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
return 1;
}
}
}
}
else if (tcor >= 90) {
if (negBalance < 25) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 1;
}
}
else if (negBalance >= 25) {
if (diffRsCsdiff < -2) {
return 1;
}
else if (diffRsCsdiff >= -2) {
if (gchDC < 1) {
if (Rs < 418) {
return 4;
}
else if (Rs >= 418) {
return 1;
}
}
else if (gchDC >= 1) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 70) {
if (mu_mv_mag_sq < 109) {
return 1;
}
else if (mu_mv_mag_sq >= 109) {
if (Rs < 404) {
return 1;
}
else if (Rs >= 404) {
if (Rs < 554) {
return 8;
}
else if (Rs >= 554) {
return 4;
}
}
}
}
}
else if (negBalance >= 35) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (mcTcor < 91) {
if (Cs < 332) {
return 1;
}
else if (Cs >= 332) {
return 8;
}
}
else if (mcTcor >= 91) {
if (tcor < 65) {
if (Rs < 658) {
return 1;
}
else if (Rs >= 658) {
return 8;
}
}
else if (tcor >= 65) {
if (mcTcor < 99) {
if (MVDiff < 3) {
if (Cs < 340) {
return 8;
}
else if (Cs >= 340) {
return 1;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (mcTcor >= 99) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 45) {
if (SC < 694) {
if (Cs < 268) {
if (tcor < 65) {
if (Rs < 412) {
return 8;
}
else if (Rs >= 412) {
return 2;
}
}
else if (tcor >= 65) {
if (posBalance < 67) {
if (Rs < 418) {
if (refDCval < 49) {
if (Cs < 227) {
if (tcor < 89) {
if (Rs < 397) {
return 4;
}
else if (Rs >= 397) {
return 1;
}
}
else if (tcor >= 89) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
return 1;
}
}
}
else if (Cs >= 227) {
if (mu_mv_mag_sq < 84) {
return 4;
}
else if (mu_mv_mag_sq >= 84) {
return 1;
}
}
}
else if (refDCval >= 49) {
if (negBalance < 39) {
if (Rs < 314) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
return 2;
}
}
else if (Rs >= 314) {
return 4;
}
}
else if (negBalance >= 39) {
if (mu_mv_mag_sq < 45) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
return 1;
}
}
else if (mu_mv_mag_sq >= 45) {
return 1;
}
}
}
}
else if (Rs >= 418) {
if (tcor < 92) {
if (Cs < 175) {
if (Rs < 492) {
return 2;
}
else if (Rs >= 492) {
return 1;
}
}
else if (Cs >= 175) {
return 4;
}
}
else if (tcor >= 92) {
if (RsDiff < 1) {
return 4;
}
else if (RsDiff >= 1) {
return 1;
}
}
}
}
else if (posBalance >= 67) {
return 2;
}
}
}
else if (Cs >= 268) {
if (ssDCval < 58) {
if (negBalance < 39) {
if (TSC < 2) {
return 4;
}
else if (TSC >= 2) {
return 1;
}
}
else if (negBalance >= 39) {
if (Rs < 318) {
return 2;
}
else if (Rs >= 318) {
return 4;
}
}
}
else if (ssDCval >= 58) {
if (ssDCval < 62) {
return 2;
}
else if (ssDCval >= 62) {
if (SC < 629) {
return 1;
}
else if (SC >= 629) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 2;
}
}
}
}
}
}
else if (SC >= 694) {
if (mu_mv_mag_sq < 3) {
if (tcor < 97) {
if (tcor < 87) {
if (ssDCval < 62) {
if (refDCval < 53) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (refDCval >= 53) {
if (Cs < 481) {
if (refDCval < 62) {
return 2;
}
else if (refDCval >= 62) {
return 1;
}
}
else if (Cs >= 481) {
return 8;
}
}
}
else if (ssDCval >= 62) {
if (Cs < 174) {
return 2;
}
else if (Cs >= 174) {
return 4;
}
}
}
else if (tcor >= 87) {
if (posBalance < 46) {
if (AFD < 6) {
if (Cs < 319) {
return 1;
}
else if (Cs >= 319) {
return 2;
}
}
else if (AFD >= 6) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
return 1;
}
}
}
else if (posBalance >= 46) {
if (Rs < 576) {
if (negBalance < 37) {
if (Rs < 427) {
return 2;
}
else if (Rs >= 427) {
return 4;
}
}
else if (negBalance >= 37) {
if (mcTcor < 96) {
if (RsCsDiff < 4) {
return 4;
}
else if (RsCsDiff >= 4) {
if (Rs < 451) {
return 2;
}
else if (Rs >= 451) {
return 8;
}
}
}
else if (mcTcor >= 96) {
if (tcor < 92) {
return 1;
}
else if (tcor >= 92) {
if (Rs < 363) {
return 1;
}
else if (Rs >= 363) {
if (Rs < 549) {
return 4;
}
else if (Rs >= 549) {
return 1;
}
}
}
}
}
}
else if (Rs >= 576) {
if (mcTcor < 94) {
if (Rs < 816) {
return 4;
}
else if (Rs >= 816) {
return 2;
}
}
else if (mcTcor >= 94) {
if (Cs < 402) {
return 1;
}
else if (Cs >= 402) {
if (Rs < 682) {
return 8;
}
else if (Rs >= 682) {
return 1;
}
}
}
}
}
}
}
else if (tcor >= 97) {
if (Cs < 155) {
return 1;
}
else if (Cs >= 155) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 99) {
if (mcTcor < 97) {
if (ssDCval < 54) {
if (SC < 908) {
if (refDCval < 48) {
if (SC < 770) {
return 8;
}
else if (SC >= 770) {
if (posBalance < 42) {
return 4;
}
else if (posBalance >= 42) {
return 1;
}
}
}
else if (refDCval >= 48) {
if (tcor < 88) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 2;
}
}
else if (tcor >= 88) {
if (Rs < 382) {
return 4;
}
else if (Rs >= 382) {
return 1;
}
}
}
}
else if (SC >= 908) {
if (posBalance < 43) {
if (Cs < 456) {
return 1;
}
else if (Cs >= 456) {
return 2;
}
}
else if (posBalance >= 43) {
if (refDCval < 48) {
if (negBalance < 31) {
return 1;
}
else if (negBalance >= 31) {
if (Cs < 413) {
return 8;
}
else if (Cs >= 413) {
return 4;
}
}
}
else if (refDCval >= 48) {
if (Rs < 506) {
return 2;
}
else if (Rs >= 506) {
return 8;
}
}
}
}
}
else if (ssDCval >= 54) {
if (SC < 1040) {
if (refDCval < 62) {
if (RsCsDiff < 12) {
if (refDCval < 60) {
if (tcor < 79) {
if (Cs < 133) {
return 8;
}
else if (Cs >= 133) {
return 2;
}
}
else if (tcor >= 79) {
if (TSC < 4) {
if (SC < 819) {
if (SC < 741) {
return 2;
}
else if (SC >= 741) {
return 1;
}
}
else if (SC >= 819) {
if (ssDCval < 57) {
if (Rs < 540) {
return 1;
}
else if (Rs >= 540) {
return 4;
}
}
else if (ssDCval >= 57) {
if (SC < 1025) {
if (RsDiff < 1) {
return 8;
}
else if (RsDiff >= 1) {
return 2;
}
}
else if (SC >= 1025) {
return 1;
}
}
}
}
else if (TSC >= 4) {
if (posBalance < 52) {
if (Rs < 627) {
return 4;
}
else if (Rs >= 627) {
return 8;
}
}
else if (posBalance >= 52) {
if (Rs < 367) {
return 4;
}
else if (Rs >= 367) {
return 1;
}
}
}
}
}
else if (refDCval >= 60) {
return 2;
}
}
else if (RsCsDiff >= 12) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (refDCval >= 62) {
if (ssDCval < 66) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
if (Cs < 302) {
return 4;
}
else if (Cs >= 302) {
return 2;
}
}
}
else if (ssDCval >= 66) {
if (posBalance < 56) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
if (Rs < 583) {
return 4;
}
else if (Rs >= 583) {
return 1;
}
}
}
else if (posBalance >= 56) {
return 2;
}
}
}
}
else if (SC >= 1040) {
if (Cs < 427) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (negBalance < 37) {
return 1;
}
else if (negBalance >= 37) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 1;
}
}
}
}
else if (Cs >= 427) {
if (Cs < 489) {
return 4;
}
else if (Cs >= 489) {
return 8;
}
}
}
}
}
else if (mcTcor >= 97) {
if (Rs < 634) {
if (mu_mv_mag_sq < 93) {
if (Rs < 460) {
if (refDCval < 61) {
if (ssDCval < 48) {
return 4;
}
else if (ssDCval >= 48) {
if (negBalance < 20) {
if (Rs < 370) {
return 4;
}
else if (Rs >= 370) {
return 8;
}
}
else if (negBalance >= 20) {
return 1;
}
}
}
else if (refDCval >= 61) {
return 4;
}
}
else if (Rs >= 460) {
if (ssDCval < 56) {
if (AFD < 6) {
if (Cs < 240) {
return 1;
}
else if (Cs >= 240) {
if (MVDiff < 13) {
if (negBalance < 34) {
return 4;
}
else if (negBalance >= 34) {
if (Rs < 560) {
return 1;
}
else if (Rs >= 560) {
return 4;
}
}
}
else if (MVDiff >= 13) {
return 1;
}
}
}
else if (AFD >= 6) {
if (Rs < 596) {
return 8;
}
else if (Rs >= 596) {
return 1;
}
}
}
else if (ssDCval >= 56) {
if (MVDiff < 7) {
if (SC < 934) {
return 2;
}
else if (SC >= 934) {
if (Cs < 440) {
return 8;
}
else if (Cs >= 440) {
return 1;
}
}
}
else if (MVDiff >= 7) {
if (tcor < 95) {
return 1;
}
else if (tcor >= 95) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 93) {
if (MVDiff < 90) {
return 1;
}
else if (MVDiff >= 90) {
return 8;
}
}
}
else if (Rs >= 634) {
if (diffMVdiffVal < 3) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
if (Cs < 188) {
return 1;
}
else if (Cs >= 188) {
if (Cs < 342) {
return 4;
}
else if (Cs >= 342) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 3) {
if (Cs < 413) {
if (Rs < 667) {
return 2;
}
else if (Rs >= 667) {
return 1;
}
}
else if (Cs >= 413) {
return 2;
}
}
}
}
}
else if (mcTcor >= 99) {
if (RsCsDiff < 5) {
if (SC < 972) {
if (negBalance < 32) {
if (refDCval < 54) {
if (Rs < 461) {
return 1;
}
else if (Rs >= 461) {
return 4;
}
}
else if (refDCval >= 54) {
if (posBalance < 58) {
if (mu_mv_mag_sq < 107) {
if (Rs < 686) {
if (Rs < 362) {
return 4;
}
else if (Rs >= 362) {
return 8;
}
}
else if (Rs >= 686) {
return 1;
}
}
else if (mu_mv_mag_sq >= 107) {
if (Rs < 564) {
return 1;
}
else if (Rs >= 564) {
return 8;
}
}
}
else if (posBalance >= 58) {
return 1;
}
}
}
else if (negBalance >= 32) {
return 1;
}
}
else if (SC >= 972) {
if (mu_mv_mag_sq < 118) {
return 1;
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
}
else if (RsCsDiff >= 5) {
if (AFD < 6) {
if (Rs < 1232) {
return 2;
}
else if (Rs >= 1232) {
return 1;
}
}
else if (AFD >= 6) {
if (Rs < 603) {
return 8;
}
else if (Rs >= 603) {
return 1;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 68) {
if (mu_mv_mag_sq < 112) {
if (mu_mv_mag_sq < 85) {
if (Rs < 522) {
if (MVDiff < 19) {
if (Cs < 313) {
if (gchDC < 2) {
if (tcor < 98) {
return 1;
}
else if (tcor >= 98) {
return 4;
}
}
else if (gchDC >= 2) {
if (Rs < 465) {
return 2;
}
else if (Rs >= 465) {
return 4;
}
}
}
else if (Cs >= 313) {
if (mu_mv_mag_sq < 68) {
if (AFD < 17) {
if (diffTSC < 1) {
return 4;
}
else if (diffTSC >= 1) {
return 2;
}
}
else if (AFD >= 17) {
return 8;
}
}
else if (mu_mv_mag_sq >= 68) {
if (negBalance < 32) {
return 1;
}
else if (negBalance >= 32) {
return 2;
}
}
}
}
else if (MVDiff >= 19) {
if (Cs < 350) {
if (Rs < 313) {
return 1;
}
else if (Rs >= 313) {
return 8;
}
}
else if (Cs >= 350) {
if (Rs < 431) {
return 1;
}
else if (Rs >= 431) {
return 2;
}
}
}
}
else if (Rs >= 522) {
if (TSC < 3) {
if (Rs < 1292) {
return 1;
}
else if (Rs >= 1292) {
return 8;
}
}
else if (TSC >= 3) {
if (MVDiff < 8) {
if (SC < 971) {
return 4;
}
else if (SC >= 971) {
if (tcor < 87) {
if (Rs < 930) {
return 1;
}
else if (Rs >= 930) {
return 4;
}
}
else if (tcor >= 87) {
if (TSC < 5) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 4;
}
}
else if (TSC >= 5) {
return 8;
}
}
}
}
else if (MVDiff >= 8) {
if (SC < 953) {
return 1;
}
else if (SC >= 953) {
if (MVDiff < 29) {
if (RsCsDiff < 7) {
return 4;
}
else if (RsCsDiff >= 7) {
return 1;
}
}
else if (MVDiff >= 29) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (tcor < 97) {
if (Rs < 608) {
if (Rs < 494) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 2;
}
}
else if (Rs >= 494) {
if (Cs < 304) {
return 8;
}
else if (Cs >= 304) {
return 1;
}
}
}
else if (Rs >= 608) {
return 4;
}
}
else if (tcor >= 97) {
if (Rs < 376) {
if (Cs < 326) {
return 4;
}
else if (Cs >= 326) {
return 2;
}
}
else if (Rs >= 376) {
if (ssDCval < 70) {
if (Cs < 355) {
if (Cs < 196) {
return 1;
}
else if (Cs >= 196) {
return 8;
}
}
else if (Cs >= 355) {
if (posBalance < 45) {
if (Rs < 430) {
return 8;
}
else if (Rs >= 430) {
return 1;
}
}
else if (posBalance >= 45) {
return 2;
}
}
}
else if (ssDCval >= 70) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (MVDiff < 1) {
if (SC < 947) {
if (posBalance < 46) {
if (Cs < 363) {
return 2;
}
else if (Cs >= 363) {
return 1;
}
}
else if (posBalance >= 46) {
if (Rs < 626) {
return 2;
}
else if (Rs >= 626) {
return 4;
}
}
}
else if (SC >= 947) {
if (Cs < 436) {
if (SC < 1070) {
if (ssDCval < 70) {
if (Rs < 629) {
return 1;
}
else if (Rs >= 629) {
if (Rs < 670) {
return 2;
}
else if (Rs >= 670) {
return 4;
}
}
}
else if (ssDCval >= 70) {
if (Rs < 591) {
return 1;
}
else if (Rs >= 591) {
return 4;
}
}
}
else if (SC >= 1070) {
if (Rs < 860) {
return 2;
}
else if (Rs >= 860) {
return 1;
}
}
}
else if (Cs >= 436) {
return 1;
}
}
}
else if (MVDiff >= 1) {
if (tcor < 98) {
if (Cs < 335) {
if (Rs < 519) {
return 8;
}
else if (Rs >= 519) {
return 4;
}
}
else if (Cs >= 335) {
return 1;
}
}
else if (tcor >= 98) {
if (ssDCval < 71) {
return 2;
}
else if (ssDCval >= 71) {
return 4;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1295) {
if (ssDCval < 64) {
if (RsCsDiff < 2) {
if (SC < 1642) {
return 8;
}
else if (SC >= 1642) {
if (ssDCval < 62) {
if (refDCval < 54) {
if (Rs < 1393) {
return 1;
}
else if (Rs >= 1393) {
return 2;
}
}
else if (refDCval >= 54) {
if (Cs < 415) {
if (Rs < 1440) {
return 8;
}
else if (Rs >= 1440) {
return 1;
}
}
else if (Cs >= 415) {
return 8;
}
}
}
else if (ssDCval >= 62) {
if (mu_mv_mag_sq < 121) {
if (Rs < 1660) {
return 1;
}
else if (Rs >= 1660) {
return 2;
}
}
else if (mu_mv_mag_sq >= 121) {
return 2;
}
}
}
}
else if (RsCsDiff >= 2) {
if (negBalance < 29) {
if (Rs < 1415) {
return 1;
}
else if (Rs >= 1415) {
if (SC < 2446) {
if (diffMVdiffVal < -3) {
if (MVDiff < 6) {
return 2;
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 2;
}
}
else if (SC >= 2446) {
return 4;
}
}
}
else if (negBalance >= 29) {
if (Cs < 135) {
return 2;
}
else if (Cs >= 135) {
return 1;
}
}
}
}
else if (ssDCval >= 64) {
if (Rs < 1799) {
if (ssDCval < 69) {
if (mu_mv_mag_sq < 59) {
if (TSCindex < 3) {
return 4;
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 23) {
return 1;
}
else if (mu_mv_mag_sq >= 23) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 59) {
if (mu_mv_mag_sq < 127) {
return 1;
}
else if (mu_mv_mag_sq >= 127) {
return 2;
}
}
}
else if (ssDCval >= 69) {
if (tcor < 98) {
if (tcor < 84) {
return 4;
}
else if (tcor >= 84) {
if (Rs < 1645) {
return 1;
}
else if (Rs >= 1645) {
if (Rs < 1672) {
return 8;
}
else if (Rs >= 1672) {
return 1;
}
}
}
}
else if (tcor >= 98) {
if (Cs < 186) {
return 1;
}
else if (Cs >= 186) {
return 8;
}
}
}
}
else if (Rs >= 1799) {
if (Cs < 434) {
return 2;
}
else if (Cs >= 434) {
return 4;
}
}
}
}
}
else if (refDCval >= 74) {
if (Rs < 1380) {
if (SC < 899) {
if (Cs < 171) {
if (refDCval < 84) {
if (AFD < 5) {
if (Cs < 168) {
return 8;
}
else if (Cs >= 168) {
return 2;
}
}
else if (AFD >= 5) {
if (Cs < 106) {
if (refDCval < 81) {
return 1;
}
else if (refDCval >= 81) {
return 2;
}
}
else if (Cs >= 106) {
if (Rs < 489) {
return 8;
}
else if (Rs >= 489) {
if (tcor < 96) {
return 4;
}
else if (tcor >= 96) {
return 1;
}
}
}
}
}
else if (refDCval >= 84) {
if (tcor < 58) {
if (mcTcor < 93) {
if (MVDiff < 20) {
return 8;
}
else if (MVDiff >= 20) {
return 2;
}
}
else if (mcTcor >= 93) {
return 8;
}
}
else if (tcor >= 58) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
if (RsDiff < 2) {
if (SC < 392) {
return 1;
}
else if (SC >= 392) {
return 8;
}
}
else if (RsDiff >= 2) {
if (AFD < 17) {
if (CsDiff < 2) {
return 1;
}
else if (CsDiff >= 2) {
if (Rs < 412) {
return 2;
}
else if (Rs >= 412) {
return 8;
}
}
}
else if (AFD >= 17) {
return 8;
}
}
}
}
}
}
else if (Cs >= 171) {
if (mcTcor < 97) {
if (MVDiff < 38) {
if (Rs < 516) {
if (diffAFD < -2) {
if (MVDiff < 11) {
return 1;
}
else if (MVDiff >= 11) {
if (negBalance < 49) {
return 8;
}
else if (negBalance >= 49) {
if (Cs < 352) {
return 1;
}
else if (Cs >= 352) {
return 8;
}
}
}
}
else if (diffAFD >= -2) {
if (diffAFD < 3) {
if (mu_mv_mag_sq < 16) {
if (mcTcor < 95) {
if (TSC < 6) {
if (Cs < 323) {
return 2;
}
else if (Cs >= 323) {
if (Cs < 451) {
if (negBalance < 48) {
return 2;
}
else if (negBalance >= 48) {
return 8;
}
}
else if (Cs >= 451) {
return 2;
}
}
}
else if (TSC >= 6) {
if (diffMVdiffVal < -19) {
return 1;
}
else if (diffMVdiffVal >= -19) {
return 8;
}
}
}
else if (mcTcor >= 95) {
if (MVDiff < 14) {
if (Cs < 211) {
return 8;
}
else if (Cs >= 211) {
return 1;
}
}
else if (MVDiff >= 14) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (diffRsCsdiff < 1) {
if (Rs < 306) {
if (Rs < 305) {
return 8;
}
else if (Rs >= 305) {
return 2;
}
}
else if (Rs >= 306) {
return 8;
}
}
else if (diffRsCsdiff >= 1) {
if (SC < 701) {
if (tcor < 87) {
if (Rs < 365) {
return 8;
}
else if (Rs >= 365) {
return 4;
}
}
else if (tcor >= 87) {
return 2;
}
}
else if (SC >= 701) {
if (mcTcor < 90) {
if (Rs < 367) {
return 2;
}
else if (Rs >= 367) {
if (Rs < 425) {
return 1;
}
else if (Rs >= 425) {
return 8;
}
}
}
else if (mcTcor >= 90) {
return 8;
}
}
}
}
}
else if (diffAFD >= 3) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 2;
}
}
}
}
else if (Rs >= 516) {
if (mu_mv_mag_sq < 69) {
if (RsCsDiff < 3) {
return 4;
}
else if (RsCsDiff >= 3) {
return 1;
}
}
else if (mu_mv_mag_sq >= 69) {
return 8;
}
}
}
else if (MVDiff >= 38) {
if (diffMVdiffVal < 49) {
if (ssDCval < 80) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 2;
}
}
}
else if (ssDCval >= 80) {
if (mu_mv_mag_sq < 42) {
if (Rs < 521) {
return 2;
}
else if (Rs >= 521) {
return 1;
}
}
else if (mu_mv_mag_sq >= 42) {
return 8;
}
}
}
else if (diffMVdiffVal >= 49) {
if (Rs < 483) {
if (ssDCval < 139) {
if (Rs < 304) {
return 2;
}
else if (Rs >= 304) {
return 4;
}
}
else if (ssDCval >= 139) {
return 8;
}
}
else if (Rs >= 483) {
return 8;
}
}
}
}
else if (mcTcor >= 97) {
if (Cs < 226) {
if (ssDCval < 84) {
if (ssDCval < 80) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
else if (ssDCval >= 80) {
return 2;
}
}
else if (ssDCval >= 84) {
if (posBalance < 43) {
if (ssDCval < 88) {
return 1;
}
else if (ssDCval >= 88) {
return 2;
}
}
else if (posBalance >= 43) {
if (Rs < 388) {
if (Cs < 197) {
if (ssDCval < 89) {
return 1;
}
else if (ssDCval >= 89) {
return 4;
}
}
else if (Cs >= 197) {
return 2;
}
}
else if (Rs >= 388) {
if (ssDCval < 88) {
return 4;
}
else if (ssDCval >= 88) {
return 1;
}
}
}
}
}
else if (Cs >= 226) {
if (ssDCval < 77) {
if (mu_mv_mag_sq < 51) {
if (Rs < 557) {
return 1;
}
else if (Rs >= 557) {
return 4;
}
}
else if (mu_mv_mag_sq >= 51) {
if (Rs < 324) {
return 2;
}
else if (Rs >= 324) {
return 8;
}
}
}
else if (ssDCval >= 77) {
if (posBalance < 32) {
if (Rs < 447) {
if (Cs < 448) {
return 8;
}
else if (Cs >= 448) {
return 1;
}
}
else if (Rs >= 447) {
return 4;
}
}
else if (posBalance >= 32) {
if (Rs < 388) {
if (TSC < 3) {
if (mcTcor < 99) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 8;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (TSC >= 3) {
if (ssDCval < 94) {
if (SC < 786) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 1;
}
}
else if (SC >= 786) {
return 8;
}
}
else if (ssDCval >= 94) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
if (tcor < 98) {
if (diffRsCsdiff < 7) {
return 2;
}
else if (diffRsCsdiff >= 7) {
return 1;
}
}
else if (tcor >= 98) {
if (ssDCval < 148) {
return 2;
}
else if (ssDCval >= 148) {
return 4;
}
}
}
}
}
}
else if (Rs >= 388) {
if (Cs < 293) {
if (ssDCval < 94) {
if (mu_mv_mag_sq < 63) {
return 4;
}
else if (mu_mv_mag_sq >= 63) {
if (posBalance < 50) {
return 2;
}
else if (posBalance >= 50) {
return 8;
}
}
}
else if (ssDCval >= 94) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
return 8;
}
}
}
else if (Cs >= 293) {
if (AFD < 5) {
if (TSC < 3) {
return 1;
}
else if (TSC >= 3) {
if (Rs < 453) {
return 2;
}
else if (Rs >= 453) {
return 1;
}
}
}
else if (AFD >= 5) {
if (mu_mv_mag_sq < 60) {
if (posBalance < 60) {
if (Cs < 337) {
if (Cs < 298) {
return 1;
}
else if (Cs >= 298) {
if (Rs < 496) {
return 4;
}
else if (Rs >= 496) {
return 8;
}
}
}
else if (Cs >= 337) {
return 1;
}
}
else if (posBalance >= 60) {
if (Rs < 393) {
return 2;
}
else if (Rs >= 393) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 60) {
if (Rs < 421) {
return 8;
}
else if (Rs >= 421) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (SC >= 899) {
if (mu_mv_mag_sq < 119) {
if (tcor < 98) {
if (mcTcor < 95) {
if (ssDCval < 93) {
if (RsCsDiff < 12) {
if (SC < 1045) {
if (RsCsDiff < 7) {
return 8;
}
else if (RsCsDiff >= 7) {
return 1;
}
}
else if (SC >= 1045) {
return 1;
}
}
else if (RsCsDiff >= 12) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 2;
}
}
}
else if (ssDCval >= 93) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 1;
}
}
}
else if (mcTcor >= 95) {
if (mu_mv_mag_sq < 69) {
if (refDCval < 110) {
if (ssDCval < 76) {
if (SC < 1001) {
if (Rs < 617) {
return 1;
}
else if (Rs >= 617) {
return 4;
}
}
else if (SC >= 1001) {
return 1;
}
}
else if (ssDCval >= 76) {
if (TSC < 7) {
if (AFD < 19) {
return 1;
}
else if (AFD >= 19) {
return 8;
}
}
else if (TSC >= 7) {
if (Cs < 380) {
return 1;
}
else if (Cs >= 380) {
return 8;
}
}
}
}
else if (refDCval >= 110) {
if (Cs < 421) {
if (ssDCval < 122) {
return 8;
}
else if (ssDCval >= 122) {
if (MVDiff < 8) {
if (Rs < 880) {
return 8;
}
else if (Rs >= 880) {
return 1;
}
}
else if (MVDiff >= 8) {
return 4;
}
}
}
else if (Cs >= 421) {
if (negBalance < 53) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
if (Rs < 496) {
return 2;
}
else if (Rs >= 496) {
return 1;
}
}
}
else if (negBalance >= 53) {
if (TSC < 5) {
return 2;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 69) {
if (negBalance < 41) {
if (refDCval < 94) {
if (mu_mv_mag_sq < 110) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (mu_mv_mag_sq >= 110) {
return 8;
}
}
else if (refDCval >= 94) {
if (Cs < 387) {
if (TSC < 2) {
if (Rs < 811) {
return 1;
}
else if (Rs >= 811) {
return 8;
}
}
else if (TSC >= 2) {
return 4;
}
}
else if (Cs >= 387) {
if (ssDCval < 159) {
return 8;
}
else if (ssDCval >= 159) {
return 4;
}
}
}
}
else if (negBalance >= 41) {
if (refDCval < 77) {
return 2;
}
else if (refDCval >= 77) {
return 8;
}
}
}
}
}
else if (tcor >= 98) {
if (Rs < 585) {
if (negBalance < 37) {
if (Rs < 578) {
return 1;
}
else if (Rs >= 578) {
return 4;
}
}
else if (negBalance >= 37) {
if (AFD < 5) {
if (Cs < 473) {
return 8;
}
else if (Cs >= 473) {
return 2;
}
}
else if (AFD >= 5) {
return 1;
}
}
}
else if (Rs >= 585) {
if (SC < 1059) {
if (diffRsCsdiff < -3) {
return 1;
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
else if (SC >= 1059) {
if (Rs < 675) {
return 1;
}
else if (Rs >= 675) {
if (negBalance < 30) {
if (Cs < 52) {
return 8;
}
else if (Cs >= 52) {
if (mu_mv_mag_sq < 109) {
if (Rs < 927) {
return 8;
}
else if (Rs >= 927) {
return 1;
}
}
else if (mu_mv_mag_sq >= 109) {
if (ssDCval < 81) {
return 8;
}
else if (ssDCval >= 81) {
if (Cs < 483) {
return 1;
}
else if (Cs >= 483) {
return 8;
}
}
}
}
}
else if (negBalance >= 30) {
if (MVDiff < 56) {
if (Rs < 1356) {
return 8;
}
else if (Rs >= 1356) {
if (Rs < 1371) {
return 1;
}
else if (Rs >= 1371) {
return 8;
}
}
}
else if (MVDiff >= 56) {
if (diffAFD < 2) {
return 1;
}
else if (diffAFD >= 2) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 119) {
if (Rs < 999) {
if (Cs < 388) {
if (Cs < 217) {
return 8;
}
else if (Cs >= 217) {
if (Rs < 598) {
return 4;
}
else if (Rs >= 598) {
if (posBalance < 24) {
return 8;
}
else if (posBalance >= 24) {
return 1;
}
}
}
}
else if (Cs >= 388) {
if (Rs < 539) {
return 8;
}
else if (Rs >= 539) {
if (AFD < 3) {
if (posBalance < 45) {
if (Rs < 567) {
return 8;
}
else if (Rs >= 567) {
return 4;
}
}
else if (posBalance >= 45) {
return 1;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
}
else if (Rs >= 999) {
if (ssDCval < 78) {
if (Cs < 188) {
return 8;
}
else if (Cs >= 188) {
return 1;
}
}
else if (ssDCval >= 78) {
return 2;
}
}
}
}
}
else if (Rs >= 1380) {
if (ssDCval < 102) {
if (Cs < 324) {
if (ssDCval < 74) {
if (Rs < 1561) {
return 8;
}
else if (Rs >= 1561) {
return 1;
}
}
else if (ssDCval >= 74) {
return 1;
}
}
else if (Cs >= 324) {
if (Rs < 2042) {
if (mu_mv_mag_sq < 84) {
if (tcor < 97) {
return 1;
}
else if (tcor >= 97) {
if (mu_mv_mag_sq < 80) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
return 1;
}
}
else if (mu_mv_mag_sq >= 80) {
if (Cs < 438) {
return 1;
}
else if (Cs >= 438) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 84) {
if (RsDiff < 3) {
return 1;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
else if (Rs >= 2042) {
if (Cs < 453) {
return 2;
}
else if (Cs >= 453) {
return 8;
}
}
}
}
else if (ssDCval >= 102) {
if (MVDiff < 112) {
if (ssDCval < 127) {
if (refDCval < 113) {
return 8;
}
else if (refDCval >= 113) {
if (posBalance < 20) {
return 8;
}
else if (posBalance >= 20) {
if (Cs < 23) {
return 8;
}
else if (Cs >= 23) {
return 1;
}
}
}
}
else if (ssDCval >= 127) {
if (Rs < 1536) {
return 2;
}
else if (Rs >= 1536) {
if (Rs < 1643) {
return 1;
}
else if (Rs >= 1643) {
return 4;
}
}
}
}
else if (MVDiff >= 112) {
return 1;
}
}
}
}
}
else if (RsDiff >= 4) {
if (tcor < 71) {
if (posBalance < 55) {
if (SC < 2166) {
if (mu_mv_mag_sq < 32) {
if (Rs < 777) {
if (mcTcor < 82) {
if (AFD < 13) {
return 8;
}
else if (AFD >= 13) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 1;
}
}
}
else if (mcTcor >= 82) {
if (AFD < 23) {
if (diffMVdiffVal < 28) {
if (Rs < 632) {
return 8;
}
else if (Rs >= 632) {
if (Cs < 455) {
return 2;
}
else if (Cs >= 455) {
return 8;
}
}
}
else if (diffMVdiffVal >= 28) {
if (Cs < 207) {
return 1;
}
else if (Cs >= 207) {
return 2;
}
}
}
else if (AFD >= 23) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
if (Rs < 506) {
return 8;
}
else if (Rs >= 506) {
return 4;
}
}
}
}
}
else if (Rs >= 777) {
if (mu_mv_mag_sq < 9) {
if (Rs < 1249) {
if (Rs < 890) {
return 8;
}
else if (Rs >= 890) {
return 1;
}
}
else if (Rs >= 1249) {
return 2;
}
}
else if (mu_mv_mag_sq >= 9) {
if (diffAFD < 12) {
return 1;
}
else if (diffAFD >= 12) {
if (Rs < 1342) {
return 1;
}
else if (Rs >= 1342) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 32) {
if (posBalance < 35) {
return 8;
}
else if (posBalance >= 35) {
if (diffAFD < -4) {
return 2;
}
else if (diffAFD >= -4) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (ssDCval < 39) {
return 1;
}
else if (ssDCval >= 39) {
if (diffAFD < 4) {
if (diffRsCsdiff < -3) {
return 2;
}
else if (diffRsCsdiff >= -3) {
if (MVDiff < 24) {
return 2;
}
else if (MVDiff >= 24) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (AFD < 20) {
if (ssDCval < 54) {
if (Rs < 946) {
return 8;
}
else if (Rs >= 946) {
return 2;
}
}
else if (ssDCval >= 54) {
return 1;
}
}
else if (AFD >= 20) {
return 2;
}
}
}
}
}
else if (diffAFD >= 4) {
if (MVDiff < 125) {
return 1;
}
else if (MVDiff >= 125) {
if (Rs < 1194) {
return 2;
}
else if (Rs >= 1194) {
return 1;
}
}
}
}
}
}
}
}
}
else if (SC >= 2166) {
if (diffMVdiffVal < 14) {
if (ssDCval < 111) {
if (AFD < 32) {
return 2;
}
else if (AFD >= 32) {
return 1;
}
}
else if (ssDCval >= 111) {
return 1;
}
}
else if (diffMVdiffVal >= 14) {
if (ssDCval < 91) {
if (posBalance < 52) {
return 1;
}
else if (posBalance >= 52) {
return 4;
}
}
else if (ssDCval >= 91) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
if (diffAFD < 4) {
if (Cs < 332) {
if (RsCsDiff < 147) {
if (tcor < 39) {
if (AFD < 18) {
if (tcor < 23) {
if (ssDCval < 42) {
return 4;
}
else if (ssDCval >= 42) {
return 1;
}
}
else if (tcor >= 23) {
return 4;
}
}
else if (AFD >= 18) {
if (AFD < 22) {
if (AFD < 20) {
return 8;
}
else if (AFD >= 20) {
return 4;
}
}
else if (AFD >= 22) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
else if (tcor >= 39) {
if (mu_mv_mag_sq < 45) {
if (TSC < 5) {
if (diffMVdiffVal < -11) {
return 8;
}
else if (diffMVdiffVal >= -11) {
if (diffMVdiffVal < 5) {
return 2;
}
else if (diffMVdiffVal >= 5) {
if (MVDiff < 36) {
return 8;
}
else if (MVDiff >= 36) {
return 2;
}
}
}
}
else if (TSC >= 5) {
if (ssDCval < 74) {
if (Rs < 710) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
else if (Rs >= 710) {
return 1;
}
}
else if (ssDCval >= 74) {
if (CsDiff < 2) {
if (MVDiff < 62) {
if (Rs < 741) {
return 1;
}
else if (Rs >= 741) {
return 8;
}
}
else if (MVDiff >= 62) {
if (MVDiff < 78) {
return 4;
}
else if (MVDiff >= 78) {
return 1;
}
}
}
else if (CsDiff >= 2) {
if (gchDC < 6) {
if (RsCsDiff < 62) {
if (negBalance < 57) {
return 1;
}
else if (negBalance >= 57) {
return 2;
}
}
else if (RsCsDiff >= 62) {
if (ssDCval < 132) {
return 2;
}
else if (ssDCval >= 132) {
return 1;
}
}
}
else if (gchDC >= 6) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (refDCval < 82) {
if (SC < 768) {
return 2;
}
else if (SC >= 768) {
return 1;
}
}
else if (refDCval >= 82) {
if (CsDiff < 1) {
if (Rs < 1221) {
return 8;
}
else if (Rs >= 1221) {
return 2;
}
}
else if (CsDiff >= 1) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 147) {
if (diffMVdiffVal < -31) {
if (Rs < 1790) {
return 1;
}
else if (Rs >= 1790) {
return 4;
}
}
else if (diffMVdiffVal >= -31) {
if (Rs < 1193) {
return 4;
}
else if (Rs >= 1193) {
if (mu_mv_mag_sq < 15) {
if (posBalance < 58) {
return 4;
}
else if (posBalance >= 58) {
return 2;
}
}
else if (mu_mv_mag_sq >= 15) {
return 2;
}
}
}
}
}
else if (Cs >= 332) {
if (mu_mv_mag_sq < 5) {
if (RsCsDiff < 41) {
return 8;
}
else if (RsCsDiff >= 41) {
if (mcTcor < 87) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (mcTcor >= 87) {
if (diffRsCsdiff < 21) {
return 1;
}
else if (diffRsCsdiff >= 21) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (CsDiff < 3) {
if (MVDiff < 22) {
if (diffMVdiffVal < -10) {
if (Rs < 1171) {
return 4;
}
else if (Rs >= 1171) {
return 1;
}
}
else if (diffMVdiffVal >= -10) {
return 8;
}
}
else if (MVDiff >= 22) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (mcTcor < 90) {
if (refDCval < 44) {
if (TSC < 6) {
if (SC < 1181) {
return 1;
}
else if (SC >= 1181) {
return 2;
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (refDCval >= 44) {
if (Rs < 1854) {
if (posBalance < 66) {
if (RsCsDiff < 38) {
if (Rs < 616) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 2;
}
}
else if (Rs >= 616) {
return 8;
}
}
else if (RsCsDiff >= 38) {
if (ssDCval < 54) {
return 2;
}
else if (ssDCval >= 54) {
if (MVDiff < 56) {
return 4;
}
else if (MVDiff >= 56) {
if (tcor < 48) {
return 4;
}
else if (tcor >= 48) {
return 2;
}
}
}
}
}
else if (posBalance >= 66) {
if (TSC < 10) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (TSC >= 10) {
return 8;
}
}
}
else if (Rs >= 1854) {
if (AFD < 17) {
return 1;
}
else if (AFD >= 17) {
return 8;
}
}
}
}
else if (mcTcor >= 90) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
if (AFD < 25) {
if (ssDCval < 78) {
if (TSC < 6) {
if (Rs < 821) {
return 1;
}
else if (Rs >= 821) {
return 8;
}
}
else if (TSC >= 6) {
return 4;
}
}
else if (ssDCval >= 78) {
return 1;
}
}
else if (AFD >= 25) {
if (gchDC < 8) {
if (diffAFD < -1) {
return 1;
}
else if (diffAFD >= -1) {
if (diffTSC < -3) {
return 1;
}
else if (diffTSC >= -3) {
return 2;
}
}
}
else if (gchDC >= 8) {
if (Cs < 466) {
return 8;
}
else if (Cs >= 466) {
return 2;
}
}
}
}
}
}
}
}
}
else if (diffAFD >= 4) {
if (RsDiff < 19) {
if (mu_mv_mag_sq < 45) {
if (negBalance < 57) {
if (CsDiff < 6) {
if (AFD < 46) {
if (posBalance < 57) {
if (Rs < 1375) {
return 1;
}
else if (Rs >= 1375) {
return 4;
}
}
else if (posBalance >= 57) {
return 1;
}
}
else if (AFD >= 46) {
return 4;
}
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (negBalance >= 57) {
if (diffRsCsdiff < 24) {
return 1;
}
else if (diffRsCsdiff >= 24) {
if (Rs < 1315) {
return 2;
}
else if (Rs >= 1315) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 45) {
if (RsDiff < 6) {
if (ssDCval < 85) {
return 8;
}
else if (ssDCval >= 85) {
return 2;
}
}
else if (RsDiff >= 6) {
if (MVDiff < 38) {
return 2;
}
else if (MVDiff >= 38) {
if (Cs < 416) {
return 1;
}
else if (Cs >= 416) {
if (Rs < 1282) {
return 8;
}
else if (Rs >= 1282) {
return 1;
}
}
}
}
}
}
else if (RsDiff >= 19) {
return 2;
}
}
}
}
else if (tcor >= 71) {
if (mu_mv_mag_sq < 86) {
if (RsCsDiff < 77) {
if (ssDCval < 139) {
if (Cs < 269) {
if (refDCval < 80) {
if (mcTcor < 101) {
if (diffAFD < -1) {
if (Cs < 217) {
if (negBalance < 29) {
if (Rs < 1045) {
return 1;
}
else if (Rs >= 1045) {
return 2;
}
}
else if (negBalance >= 29) {
if (diffAFD < -14) {
return 8;
}
else if (diffAFD >= -14) {
return 1;
}
}
}
else if (Cs >= 217) {
if (Rs < 833) {
return 2;
}
else if (Rs >= 833) {
if (Rs < 1623) {
return 1;
}
else if (Rs >= 1623) {
return 2;
}
}
}
}
else if (diffAFD >= -1) {
if (tcor < 72) {
return 4;
}
else if (tcor >= 72) {
return 1;
}
}
}
else if (mcTcor >= 101) {
return 2;
}
}
else if (refDCval >= 80) {
if (MVDiff < 16) {
if (tcor < 81) {
return 2;
}
else if (tcor >= 81) {
if (Cs < 176) {
if (Rs < 1111) {
return 1;
}
else if (Rs >= 1111) {
if (TSC < 4) {
if (Rs < 1818) {
return 1;
}
else if (Rs >= 1818) {
return 2;
}
}
else if (TSC >= 4) {
return 2;
}
}
}
else if (Cs >= 176) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
return 2;
}
}
}
}
else if (MVDiff >= 16) {
if (SC < 1364) {
if (tcor < 89) {
if (diffRsCsdiff < -27) {
if (ssDCval < 128) {
return 2;
}
else if (ssDCval >= 128) {
return 1;
}
}
else if (diffRsCsdiff >= -27) {
if (RsCsDiff < 26) {
if (ssDCval < 105) {
if (SC < 1313) {
return 1;
}
else if (SC >= 1313) {
return 2;
}
}
else if (ssDCval >= 105) {
return 8;
}
}
else if (RsCsDiff >= 26) {
if (CsDiff < 2) {
if (TSC < 7) {
if (Cs < 97) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
return 1;
}
}
else if (Cs >= 97) {
return 2;
}
}
else if (TSC >= 7) {
return 1;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
}
}
else if (tcor >= 89) {
if (refDCval < 86) {
if (diffMVdiffVal < -3) {
return 1;
}
else if (diffMVdiffVal >= -3) {
return 8;
}
}
else if (refDCval >= 86) {
return 1;
}
}
}
else if (SC >= 1364) {
if (negBalance < 57) {
if (posBalance < 32) {
return 2;
}
else if (posBalance >= 32) {
if (MVDiff < 18) {
if (Rs < 1597) {
return 1;
}
else if (Rs >= 1597) {
return 2;
}
}
else if (MVDiff >= 18) {
return 1;
}
}
}
else if (negBalance >= 57) {
if (Rs < 1716) {
if (AFD < 12) {
return 2;
}
else if (AFD >= 12) {
return 1;
}
}
else if (Rs >= 1716) {
return 8;
}
}
}
}
}
}
else if (Cs >= 269) {
if (SC < 1204) {
if (diffMVdiffVal < -10) {
if (Rs < 477) {
return 2;
}
else if (Rs >= 477) {
return 4;
}
}
else if (diffMVdiffVal >= -10) {
if (diffMVdiffVal < 1) {
if (mu_mv_mag_sq < 74) {
if (posBalance < 43) {
if (ssDCval < 51) {
if (Rs < 587) {
return 8;
}
else if (Rs >= 587) {
return 2;
}
}
else if (ssDCval >= 51) {
if (Rs < 721) {
return 1;
}
else if (Rs >= 721) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
if (MVDiff < 24) {
return 1;
}
else if (MVDiff >= 24) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 74) {
if (Rs < 630) {
return 4;
}
else if (Rs >= 630) {
return 8;
}
}
}
else if (diffMVdiffVal >= 1) {
if (gchDC < 2) {
if (TSC < 5) {
if (Rs < 499) {
return 4;
}
else if (Rs >= 499) {
return 8;
}
}
else if (TSC >= 5) {
if (diffAFD < 1) {
return 4;
}
else if (diffAFD >= 1) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (tcor < 80) {
if (Rs < 494) {
if (Rs < 407) {
return 8;
}
else if (Rs >= 407) {
return 4;
}
}
else if (Rs >= 494) {
return 8;
}
}
else if (tcor >= 80) {
if (diffAFD < -5) {
return 8;
}
else if (diffAFD >= -5) {
return 1;
}
}
}
}
}
}
else if (SC >= 1204) {
if (AFD < 7) {
if (ssDCval < 97) {
if (diffAFD < 1) {
if (diffMVdiffVal < -1) {
if (mcTcor < 90) {
if (Rs < 947) {
return 1;
}
else if (Rs >= 947) {
return 4;
}
}
else if (mcTcor >= 90) {
return 1;
}
}
else if (diffMVdiffVal >= -1) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (SC < 1749) {
return 1;
}
else if (SC >= 1749) {
if (tcor < 95) {
if (RsDiff < 5) {
return 2;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (tcor >= 95) {
return 2;
}
}
}
}
}
else if (diffAFD >= 1) {
if (Cs < 428) {
if (Rs < 2204) {
return 1;
}
else if (Rs >= 2204) {
return 2;
}
}
else if (Cs >= 428) {
return 4;
}
}
}
else if (ssDCval >= 97) {
if (Rs < 1857) {
return 2;
}
else if (Rs >= 1857) {
return 4;
}
}
}
else if (AFD >= 7) {
if (Cs < 391) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
if (negBalance < 48) {
return 1;
}
else if (negBalance >= 48) {
if (RsCsDiff < 19) {
return 8;
}
else if (RsCsDiff >= 19) {
if (tcor < 99) {
if (diffTSC < -3) {
return 8;
}
else if (diffTSC >= -3) {
if (tcor < 89) {
if (diffRsCsdiff < -12) {
if (AFD < 16) {
return 2;
}
else if (AFD >= 16) {
return 1;
}
}
else if (diffRsCsdiff >= -12) {
return 1;
}
}
else if (tcor >= 89) {
return 1;
}
}
}
else if (tcor >= 99) {
return 8;
}
}
}
}
}
else if (Cs >= 391) {
if (negBalance < 24) {
if (Rs < 1231) {
return 8;
}
else if (Rs >= 1231) {
return 4;
}
}
else if (negBalance >= 24) {
if (mcTcor < 89) {
if (RsDiff < 7) {
if (Rs < 1242) {
return 2;
}
else if (Rs >= 1242) {
return 8;
}
}
else if (RsDiff >= 7) {
if (Rs < 1266) {
return 4;
}
else if (Rs >= 1266) {
return 1;
}
}
}
else if (mcTcor >= 89) {
if (refDCval < 73) {
if (MVDiff < 7) {
if (Rs < 1376) {
return 8;
}
else if (Rs >= 1376) {
if (Cs < 477) {
return 4;
}
else if (Cs >= 477) {
return 1;
}
}
}
else if (MVDiff >= 7) {
if (diffMVdiffVal < -26) {
if (Rs < 1004) {
return 4;
}
else if (Rs >= 1004) {
return 2;
}
}
else if (diffMVdiffVal >= -26) {
if (ssDCval < 53) {
if (Rs < 1069) {
if (Rs < 798) {
return 8;
}
else if (Rs >= 798) {
return 2;
}
}
else if (Rs >= 1069) {
if (AFD < 11) {
return 1;
}
else if (AFD >= 11) {
return 2;
}
}
}
else if (ssDCval >= 53) {
if (Rs < 1894) {
if (mu_mv_mag_sq < 3) {
if (Rs < 947) {
return 1;
}
else if (Rs >= 947) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
return 1;
}
}
else if (Rs >= 1894) {
if (AFD < 13) {
if (Rs < 2131) {
return 2;
}
else if (Rs >= 2131) {
return 4;
}
}
else if (AFD >= 13) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 73) {
if (mcTcor < 96) {
if (refDCval < 85) {
return 1;
}
else if (refDCval >= 85) {
if (Rs < 1706) {
if (RsDiff < 8) {
if (mcTcor < 93) {
if (Rs < 961) {
return 2;
}
else if (Rs >= 961) {
return 8;
}
}
else if (mcTcor >= 93) {
return 2;
}
}
else if (RsDiff >= 8) {
return 1;
}
}
else if (Rs >= 1706) {
if (MVDiff < 87) {
return 1;
}
else if (MVDiff >= 87) {
return 8;
}
}
}
}
else if (mcTcor >= 96) {
if (Rs < 2299) {
if (TSC < 4) {
if (Rs < 1993) {
return 1;
}
else if (Rs >= 1993) {
return 8;
}
}
else if (TSC >= 4) {
if (refDCval < 124) {
return 1;
}
else if (refDCval >= 124) {
if (Rs < 1012) {
return 2;
}
else if (Rs >= 1012) {
if (AFD < 10) {
if (Rs < 1525) {
return 1;
}
else if (Rs >= 1525) {
return 2;
}
}
else if (AFD >= 10) {
return 1;
}
}
}
}
}
else if (Rs >= 2299) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 139) {
if (gchDC < 2) {
if (ssDCval < 146) {
if (mcTcor < 98) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 1;
}
}
else if (mcTcor >= 98) {
if (mu_mv_mag_sq < 35) {
return 1;
}
else if (mu_mv_mag_sq >= 35) {
if (Rs < 1243) {
return 2;
}
else if (Rs >= 1243) {
return 4;
}
}
}
}
else if (ssDCval >= 146) {
if (ssDCval < 173) {
return 1;
}
else if (ssDCval >= 173) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
if (Rs < 689) {
if (SC < 1041) {
if (diffTSC < -3) {
return 2;
}
else if (diffTSC >= -3) {
return 1;
}
}
else if (SC >= 1041) {
return 8;
}
}
else if (Rs >= 689) {
if (tcor < 78) {
if (Rs < 1342) {
return 1;
}
else if (Rs >= 1342) {
return 8;
}
}
else if (tcor >= 78) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 77) {
if (refDCval < 109) {
if (MVDiff < 38) {
if (AFD < 12) {
if (ssDCval < 99) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (ssDCval >= 99) {
return 2;
}
}
else if (AFD >= 12) {
if (ssDCval < 70) {
if (SC < 1971) {
if (SC < 1744) {
return 2;
}
else if (SC >= 1744) {
return 1;
}
}
else if (SC >= 1971) {
return 2;
}
}
else if (ssDCval >= 70) {
if (negBalance < 46) {
if (RsCsDiff < 213) {
if (AFD < 16) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 2;
}
}
else if (AFD >= 16) {
if (mcTcor < 91) {
if (Rs < 2065) {
return 1;
}
else if (Rs >= 2065) {
return 2;
}
}
else if (mcTcor >= 91) {
return 1;
}
}
}
else if (RsCsDiff >= 213) {
return 2;
}
}
else if (negBalance >= 46) {
if (AFD < 21) {
return 2;
}
else if (AFD >= 21) {
if (refDCval < 94) {
return 2;
}
else if (refDCval >= 94) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 38) {
if (mu_mv_mag_sq < 24) {
if (diffRsCsdiff < -108) {
if (Rs < 2193) {
return 2;
}
else if (Rs >= 2193) {
return 1;
}
}
else if (diffRsCsdiff >= -108) {
if (RsCsDiff < 113) {
if (tcor < 79) {
if (mu_mv_mag_sq < 3) {
return 8;
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 95) {
return 1;
}
else if (mcTcor >= 95) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 1;
}
}
}
}
else if (tcor >= 79) {
return 1;
}
}
else if (RsCsDiff >= 113) {
if (tcor < 75) {
return 2;
}
else if (tcor >= 75) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
if (posBalance < 55) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 1;
}
}
else if (posBalance >= 55) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (SC < 2627) {
if (Rs < 1265) {
if (Rs < 993) {
return 2;
}
else if (Rs >= 993) {
if (Cs < 418) {
return 1;
}
else if (Cs >= 418) {
return 4;
}
}
}
else if (Rs >= 1265) {
return 2;
}
}
else if (SC >= 2627) {
if (Rs < 2233) {
return 4;
}
else if (Rs >= 2233) {
return 8;
}
}
}
}
}
else if (refDCval >= 109) {
if (AFD < 13) {
if (Rs < 1340) {
return 4;
}
else if (Rs >= 1340) {
if (tcor < 95) {
if (Rs < 2150) {
return 2;
}
else if (Rs >= 2150) {
return 1;
}
}
else if (tcor >= 95) {
return 1;
}
}
}
else if (AFD >= 13) {
if (RsCsDiff < 128) {
if (TSC < 10) {
if (mcTcor < 96) {
if (mu_mv_mag_sq < 8) {
if (tcor < 76) {
return 8;
}
else if (tcor >= 76) {
return 1;
}
}
else if (mu_mv_mag_sq >= 8) {
if (SC < 1960) {
if (MVDiff < 158) {
return 2;
}
else if (MVDiff >= 158) {
return 1;
}
}
else if (SC >= 1960) {
return 1;
}
}
}
else if (mcTcor >= 96) {
if (CsDiff < 5) {
return 1;
}
else if (CsDiff >= 5) {
return 2;
}
}
}
else if (TSC >= 10) {
if (RsCsDiff < 88) {
return 8;
}
else if (RsCsDiff >= 88) {
if (Cs < 325) {
return 1;
}
else if (Cs >= 325) {
if (MVDiff < 95) {
if (Rs < 2298) {
return 1;
}
else if (Rs >= 2298) {
return 8;
}
}
else if (MVDiff >= 95) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 128) {
if (TSC < 4) {
return 2;
}
else if (TSC >= 4) {
if (MVDiff < 379) {
if (RsDiff < 17) {
return 1;
}
else if (RsDiff >= 17) {
if (MVDiff < 36) {
if (Rs < 1880) {
return 1;
}
else if (Rs >= 1880) {
return 2;
}
}
else if (MVDiff >= 36) {
return 1;
}
}
}
else if (MVDiff >= 379) {
if (Rs < 1536) {
return 1;
}
else if (Rs >= 1536) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (negBalance < 51) {
if (Cs < 379) {
if (RsDiff < 12) {
if (Rs < 1485) {
return 1;
}
else if (Rs >= 1485) {
return 4;
}
}
else if (RsDiff >= 12) {
return 2;
}
}
else if (Cs >= 379) {
return 8;
}
}
else if (negBalance >= 51) {
return 2;
}
}
}
}
}
}
}
}
else if (Cs >= 491) {
if (negBalance < 14) {
if (Rs < 701) {
if (ssDCval < 59) {
if (ssDCval < 33) {
if (ssDCval < 26) {
return 8;
}
else if (ssDCval >= 26) {
return 4;
}
}
else if (ssDCval >= 33) {
if (TSC < 1) {
return 1;
}
else if (TSC >= 1) {
if (ssDCval < 55) {
if (mu_mv_mag_sq < 125) {
if (Rs < 694) {
return 1;
}
else if (Rs >= 694) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (Rs < 327) {
return 2;
}
else if (Rs >= 327) {
return 8;
}
}
}
else if (ssDCval >= 55) {
return 4;
}
}
}
}
else if (ssDCval >= 59) {
if (ssDCval < 108) {
if (Cs < 701) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (Cs < 525) {
return 8;
}
else if (Cs >= 525) {
return 1;
}
}
}
else if (Cs >= 701) {
if (Cs < 802) {
if (Cs < 749) {
if (Rs < 547) {
return 4;
}
else if (Rs >= 547) {
return 1;
}
}
else if (Cs >= 749) {
return 2;
}
}
else if (Cs >= 802) {
return 8;
}
}
}
else if (ssDCval >= 108) {
if (refDCval < 143) {
if (negBalance < 10) {
if (Rs < 493) {
if (Cs < 537) {
return 8;
}
else if (Cs >= 537) {
return 1;
}
}
else if (Rs >= 493) {
return 8;
}
}
else if (negBalance >= 10) {
if (posBalance < 31) {
if (AFD < 4) {
if (tcor < 100) {
return 1;
}
else if (tcor >= 100) {
return 8;
}
}
else if (AFD >= 4) {
return 8;
}
}
else if (posBalance >= 31) {
if (SC < 1174) {
if (Rs < 335) {
return 4;
}
else if (Rs >= 335) {
return 1;
}
}
else if (SC >= 1174) {
if (ssDCval < 112) {
return 1;
}
else if (ssDCval >= 112) {
return 8;
}
}
}
}
}
else if (refDCval >= 143) {
return 2;
}
}
}
}
else if (Rs >= 701) {
if (MVDiff < 6) {
if (SC < 2835) {
if (ssDCval < 153) {
if (negBalance < 12) {
if (Cs < 1200) {
if (Cs < 512) {
if (Rs < 1499) {
return 2;
}
else if (Rs >= 1499) {
return 8;
}
}
else if (Cs >= 512) {
if (diffMVdiffVal < -2) {
if (RsDiff < 2) {
if (SC < 1984) {
if (SC < 1954) {
if (Rs < 1143) {
return 8;
}
else if (Rs >= 1143) {
if (Cs < 685) {
return 1;
}
else if (Cs >= 685) {
return 8;
}
}
}
else if (SC >= 1954) {
return 1;
}
}
else if (SC >= 1984) {
return 8;
}
}
else if (RsDiff >= 2) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 1;
}
}
}
else if (diffMVdiffVal >= -2) {
if (Rs < 1846) {
if (Rs < 766) {
if (MVDiff < 1) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (Rs < 718) {
return 8;
}
else if (Rs >= 718) {
if (Rs < 752) {
return 1;
}
else if (Rs >= 752) {
return 8;
}
}
}
}
else if (MVDiff >= 1) {
if (Cs < 655) {
return 8;
}
else if (Cs >= 655) {
if (Rs < 744) {
return 1;
}
else if (Rs >= 744) {
return 2;
}
}
}
}
else if (Rs >= 766) {
if (Cs < 1187) {
if (Cs < 679) {
if (Rs < 1132) {
return 8;
}
else if (Rs >= 1132) {
if (ssDCval < 74) {
return 1;
}
else if (ssDCval >= 74) {
if (Rs < 1480) {
return 8;
}
else if (Rs >= 1480) {
return 4;
}
}
}
}
else if (Cs >= 679) {
return 8;
}
}
else if (Cs >= 1187) {
if (Cs < 1189) {
return 1;
}
else if (Cs >= 1189) {
return 8;
}
}
}
}
else if (Rs >= 1846) {
if (negBalance < 7) {
return 8;
}
else if (negBalance >= 7) {
return 1;
}
}
}
}
}
else if (Cs >= 1200) {
if (Rs < 807) {
return 2;
}
else if (Rs >= 807) {
if (negBalance < 1) {
if (Rs < 1401) {
return 1;
}
else if (Rs >= 1401) {
return 8;
}
}
else if (negBalance >= 1) {
return 8;
}
}
}
}
else if (negBalance >= 12) {
if (SC < 1816) {
if (mu_mv_mag_sq < 110) {
if (posBalance < 33) {
if (Rs < 907) {
return 2;
}
else if (Rs >= 907) {
return 8;
}
}
else if (posBalance >= 33) {
return 1;
}
}
else if (mu_mv_mag_sq >= 110) {
if (SC < 1470) {
if (SC < 1448) {
return 8;
}
else if (SC >= 1448) {
if (RsDiff < 1) {
if (Rs < 908) {
return 2;
}
else if (Rs >= 908) {
return 8;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
}
else if (SC >= 1470) {
if (diffRsCsdiff < 6) {
return 8;
}
else if (diffRsCsdiff >= 6) {
return 1;
}
}
}
}
else if (SC >= 1816) {
if (posBalance < 37) {
if (Cs < 729) {
if (posBalance < 28) {
if (Rs < 1377) {
return 8;
}
else if (Rs >= 1377) {
return 2;
}
}
else if (posBalance >= 28) {
if (Cs < 492) {
return 2;
}
else if (Cs >= 492) {
return 1;
}
}
}
else if (Cs >= 729) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
if (mu_mv_mag_sq < 128) {
if (posBalance < 21) {
return 8;
}
else if (posBalance >= 21) {
if (Rs < 981) {
return 8;
}
else if (Rs >= 981) {
if (Cs < 915) {
return 8;
}
else if (Cs >= 915) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1068) {
return 8;
}
else if (Cs >= 1068) {
return 1;
}
}
}
}
}
else if (posBalance >= 37) {
if (Rs < 1345) {
if (SC < 1845) {
return 8;
}
else if (SC >= 1845) {
if (AFD < 3) {
if (SC < 1991) {
return 4;
}
else if (SC >= 1991) {
return 1;
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (Rs >= 1345) {
if (Rs < 1919) {
return 8;
}
else if (Rs >= 1919) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 153) {
if (Rs < 837) {
return 8;
}
else if (Rs >= 837) {
return 2;
}
}
}
else if (SC >= 2835) {
if (Rs < 2041) {
if (SC < 2884) {
if (negBalance < 8) {
if (Cs < 955) {
return 1;
}
else if (Cs >= 955) {
return 8;
}
}
else if (negBalance >= 8) {
if (Rs < 1689) {
return 8;
}
else if (Rs >= 1689) {
return 1;
}
}
}
else if (SC >= 2884) {
if (ssDCval < 65) {
if (Rs < 1859) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
return 1;
}
}
else if (Rs >= 1859) {
if (Cs < 975) {
if (Rs < 2008) {
return 1;
}
else if (Rs >= 2008) {
return 4;
}
}
else if (Cs >= 975) {
return 4;
}
}
}
else if (ssDCval >= 65) {
if (ssDCval < 80) {
if (Rs < 1941) {
return 1;
}
else if (Rs >= 1941) {
return 8;
}
}
else if (ssDCval >= 80) {
return 8;
}
}
}
}
else if (Rs >= 2041) {
if (SC < 3578) {
return 8;
}
else if (SC >= 3578) {
if (Cs < 1289) {
return 8;
}
else if (Cs >= 1289) {
return 2;
}
}
}
}
}
else if (MVDiff >= 6) {
if (Cs < 660) {
if (SC < 1714) {
if (AFD < 2) {
if (diffAFD < -11) {
return 8;
}
else if (diffAFD >= -11) {
return 2;
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (SC >= 1714) {
if (tcor < 99) {
if (Rs < 1881) {
return 8;
}
else if (Rs >= 1881) {
return 2;
}
}
else if (tcor >= 99) {
return 1;
}
}
}
else if (Cs >= 660) {
if (mu_mv_mag_sq < 128) {
if (negBalance < 6) {
if (CsDiff < 3) {
if (Cs < 1189) {
return 8;
}
else if (Cs >= 1189) {
if (Rs < 997) {
return 2;
}
else if (Rs >= 997) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (Rs < 1119) {
return 1;
}
else if (Rs >= 1119) {
return 8;
}
}
}
}
else if (negBalance >= 6) {
if (Rs < 2071) {
if (gchDC < 2) {
if (negBalance < 10) {
if (posBalance < 19) {
if (Rs < 1280) {
return 8;
}
else if (Rs >= 1280) {
return 1;
}
}
else if (posBalance >= 19) {
return 8;
}
}
else if (negBalance >= 10) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Rs < 1054) {
return 2;
}
else if (Rs >= 1054) {
return 1;
}
}
}
}
else if (gchDC >= 2) {
if (diffMVdiffVal < 2) {
if (AFD < 6) {
if (Rs < 1129) {
return 8;
}
else if (Rs >= 1129) {
return 4;
}
}
else if (AFD >= 6) {
if (Rs < 796) {
return 2;
}
else if (Rs >= 796) {
return 1;
}
}
}
else if (diffMVdiffVal >= 2) {
if (SC < 3054) {
return 8;
}
else if (SC >= 3054) {
return 4;
}
}
}
}
else if (Rs >= 2071) {
if (mcTcor < 91) {
return 4;
}
else if (mcTcor >= 91) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (diffRsCsdiff < -57) {
if (MVDiff < 23) {
if (Cs < 1290) {
return 8;
}
else if (Cs >= 1290) {
return 1;
}
}
else if (MVDiff >= 23) {
if (posBalance < 1) {
if (diffTSC < -9) {
if (ssDCval < 151) {
return 4;
}
else if (ssDCval >= 151) {
return 8;
}
}
else if (diffTSC >= -9) {
return 1;
}
}
else if (posBalance >= 1) {
if (ssDCval < 93) {
return 4;
}
else if (ssDCval >= 93) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -57) {
if (SC < 2740) {
if (ssDCval < 73) {
return 8;
}
else if (ssDCval >= 73) {
if (Rs < 983) {
if (Cs < 836) {
return 4;
}
else if (Cs >= 836) {
if (Cs < 1035) {
return 8;
}
else if (Cs >= 1035) {
return 1;
}
}
}
else if (Rs >= 983) {
if (diffTSC < -7) {
return 8;
}
else if (diffTSC >= -7) {
return 4;
}
}
}
}
else if (SC >= 2740) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 14) {
if (mcTcor < 97) {
if (AFD < 13) {
if (posBalance < 42) {
if (refDCval < 47) {
if (TSCindex < 3) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Cs < 599) {
if (Rs < 863) {
return 2;
}
else if (Rs >= 863) {
return 1;
}
}
else if (Cs >= 599) {
if (Cs < 807) {
return 8;
}
else if (Cs >= 807) {
return 4;
}
}
}
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 51) {
if (Rs < 1459) {
if (negBalance < 26) {
if (AFD < 7) {
if (Rs < 570) {
return 1;
}
else if (Rs >= 570) {
return 8;
}
}
else if (AFD >= 7) {
return 2;
}
}
else if (negBalance >= 26) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 4;
}
}
}
else if (Rs >= 1459) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 51) {
if (RsDiff < 2) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 855) {
return 2;
}
else if (Rs >= 855) {
return 1;
}
}
}
else if (RsDiff >= 2) {
if (RsCsDiff < 48) {
if (negBalance < 33) {
if (diffMVdiffVal < 4) {
if (diffRsCsdiff < 10) {
return 4;
}
else if (diffRsCsdiff >= 10) {
if (Rs < 874) {
return 1;
}
else if (Rs >= 874) {
return 4;
}
}
}
else if (diffMVdiffVal >= 4) {
if (ssDCval < 27) {
return 4;
}
else if (ssDCval >= 27) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (negBalance >= 33) {
if (Rs < 1393) {
return 8;
}
else if (Rs >= 1393) {
return 4;
}
}
}
else if (RsCsDiff >= 48) {
if (MVDiff < 9) {
if (Rs < 1411) {
return 8;
}
else if (Rs >= 1411) {
return 4;
}
}
else if (MVDiff >= 9) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
}
}
}
}
}
else if (refDCval >= 47) {
if (mu_mv_mag_sq < 73) {
if (refDCval < 93) {
if (negBalance < 63) {
if (Cs < 769) {
if (refDCval < 69) {
if (mu_mv_mag_sq < 48) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (negBalance < 36) {
if (SC < 2387) {
return 2;
}
else if (SC >= 2387) {
return 1;
}
}
else if (negBalance >= 36) {
if (RsCsDiff < 33) {
return 8;
}
else if (RsCsDiff >= 33) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 48) {
if (negBalance < 38) {
if (Cs < 506) {
return 4;
}
else if (Cs >= 506) {
return 1;
}
}
else if (negBalance >= 38) {
if (Rs < 585) {
return 1;
}
else if (Rs >= 585) {
return 8;
}
}
}
}
else if (refDCval >= 69) {
if (Cs < 751) {
return 1;
}
else if (Cs >= 751) {
return 2;
}
}
}
else if (Cs >= 769) {
if (tcor < 93) {
if (Rs < 1535) {
if (TSC < 6) {
if (AFD < 8) {
if (Rs < 812) {
if (Rs < 743) {
return 2;
}
else if (Rs >= 743) {
return 4;
}
}
else if (Rs >= 812) {
if (SC < 2218) {
return 8;
}
else if (SC >= 2218) {
return 1;
}
}
}
else if (AFD >= 8) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 2;
}
}
}
else if (TSC >= 6) {
return 8;
}
}
else if (Rs >= 1535) {
if (tcor < 82) {
return 4;
}
else if (tcor >= 82) {
if (Rs < 2110) {
return 1;
}
else if (Rs >= 2110) {
return 2;
}
}
}
}
else if (tcor >= 93) {
if (Cs < 960) {
if (Rs < 893) {
if (Cs < 818) {
return 2;
}
else if (Cs >= 818) {
return 4;
}
}
else if (Rs >= 893) {
return 1;
}
}
else if (Cs >= 960) {
return 4;
}
}
}
}
else if (negBalance >= 63) {
if (Rs < 462) {
if (SC < 949) {
return 8;
}
else if (SC >= 949) {
return 1;
}
}
else if (Rs >= 462) {
if (AFD < 9) {
if (ssDCval < 76) {
return 4;
}
else if (ssDCval >= 76) {
return 8;
}
}
else if (AFD >= 9) {
return 8;
}
}
}
}
else if (refDCval >= 93) {
if (ssDCval < 151) {
if (AFD < 11) {
if (SC < 1597) {
if (tcor < 91) {
if (Rs < 606) {
if (posBalance < 40) {
return 8;
}
else if (posBalance >= 40) {
return 2;
}
}
else if (Rs >= 606) {
return 2;
}
}
else if (tcor >= 91) {
return 8;
}
}
else if (SC >= 1597) {
if (AFD < 7) {
if (Cs < 1240) {
return 1;
}
else if (Cs >= 1240) {
return 8;
}
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (AFD >= 11) {
if (negBalance < 53) {
if (SC < 2566) {
return 8;
}
else if (SC >= 2566) {
return 1;
}
}
else if (negBalance >= 53) {
if (diffMVdiffVal < 14) {
return 2;
}
else if (diffMVdiffVal >= 14) {
if (Rs < 1486) {
return 8;
}
else if (Rs >= 1486) {
return 1;
}
}
}
}
}
else if (ssDCval >= 151) {
if (Rs < 1846) {
return 2;
}
else if (Rs >= 1846) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 73) {
if (refDCval < 58) {
if (TSC < 3) {
if (Rs < 589) {
if (Rs < 442) {
return 2;
}
else if (Rs >= 442) {
return 1;
}
}
else if (Rs >= 589) {
if (negBalance < 24) {
if (Cs < 930) {
return 8;
}
else if (Cs >= 930) {
if (Rs < 1404) {
return 1;
}
else if (Rs >= 1404) {
return 4;
}
}
}
else if (negBalance >= 24) {
return 4;
}
}
}
else if (TSC >= 3) {
if (Cs < 895) {
if (negBalance < 35) {
if (SC < 1526) {
return 4;
}
else if (SC >= 1526) {
return 1;
}
}
else if (negBalance >= 35) {
if (Rs < 939) {
return 8;
}
else if (Rs >= 939) {
return 4;
}
}
}
else if (Cs >= 895) {
if (CsDiff < 3) {
if (MVDiff < 1) {
if (Rs < 1750) {
return 4;
}
else if (Rs >= 1750) {
return 8;
}
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (CsDiff >= 3) {
if (Rs < 2103) {
return 8;
}
else if (Rs >= 2103) {
return 4;
}
}
}
}
}
else if (refDCval >= 58) {
if (mu_mv_mag_sq < 98) {
if (AFD < 5) {
if (tcor < 93) {
if (ssDCval < 111) {
return 2;
}
else if (ssDCval >= 111) {
return 8;
}
}
else if (tcor >= 93) {
if (SC < 3181) {
return 4;
}
else if (SC >= 3181) {
return 8;
}
}
}
else if (AFD >= 5) {
if (SC < 2825) {
if (Rs < 1684) {
if (Rs < 945) {
if (refDCval < 65) {
if (Cs < 579) {
return 8;
}
else if (Cs >= 579) {
if (Rs < 638) {
return 1;
}
else if (Rs >= 638) {
return 4;
}
}
}
else if (refDCval >= 65) {
if (Cs < 904) {
return 2;
}
else if (Cs >= 904) {
return 8;
}
}
}
else if (Rs >= 945) {
if (Cs < 1264) {
return 8;
}
else if (Cs >= 1264) {
return 4;
}
}
}
else if (Rs >= 1684) {
if (Cs < 770) {
return 1;
}
else if (Cs >= 770) {
return 2;
}
}
}
else if (SC >= 2825) {
if (Cs < 1045) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (Cs >= 1045) {
if (negBalance < 30) {
return 4;
}
else if (negBalance >= 30) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 98) {
if (RsDiff < 5) {
if (ssDCval < 64) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
if (Rs < 1479) {
return 1;
}
else if (Rs >= 1479) {
return 8;
}
}
}
else if (ssDCval >= 64) {
if (refDCval < 93) {
if (RsDiff < 4) {
if (negBalance < 37) {
return 4;
}
else if (negBalance >= 37) {
if (mu_mv_mag_sq < 108) {
if (Rs < 1872) {
return 4;
}
else if (Rs >= 1872) {
return 4;
}
}
else if (mu_mv_mag_sq >= 108) {
if (Rs < 1680) {
return 8;
}
else if (Rs >= 1680) {
return 8;
}
}
}
}
else if (RsDiff >= 4) {
return 8;
}
}
else if (refDCval >= 93) {
if (Rs < 520) {
return 4;
}
else if (Rs >= 520) {
return 8;
}
}
}
}
else if (RsDiff >= 5) {
if (Rs < 1795) {
return 8;
}
else if (Rs >= 1795) {
return 2;
}
}
}
}
}
}
}
else if (posBalance >= 42) {
if (mu_mv_mag_sq < 27) {
if (negBalance < 47) {
if (refDCval < 51) {
if (MVDiff < 15) {
if (refDCval < 45) {
if (Rs < 872) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (TSC < 5) {
if (Rs < 825) {
return 2;
}
else if (Rs >= 825) {
if (Rs < 856) {
return 8;
}
else if (Rs >= 856) {
return 2;
}
}
}
else if (TSC >= 5) {
return 8;
}
}
}
else if (Rs >= 872) {
return 8;
}
}
else if (refDCval >= 45) {
if (tcor < 87) {
if (negBalance < 35) {
if (SC < 1674) {
if (Rs < 548) {
return 2;
}
else if (Rs >= 548) {
return 4;
}
}
else if (SC >= 1674) {
return 8;
}
}
else if (negBalance >= 35) {
if (Cs < 1222) {
return 8;
}
else if (Cs >= 1222) {
return 4;
}
}
}
else if (tcor >= 87) {
if (diffRsCsdiff < -8) {
return 8;
}
else if (diffRsCsdiff >= -8) {
if (MVDiff < 4) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
return 8;
}
}
else if (MVDiff >= 4) {
return 4;
}
}
}
}
}
else if (MVDiff >= 15) {
if (negBalance < 38) {
if (ssDCval < 41) {
if (MVDiff < 32) {
if (Cs < 515) {
return 2;
}
else if (Cs >= 515) {
return 1;
}
}
else if (MVDiff >= 32) {
if (posBalance < 51) {
if (Rs < 233) {
return 1;
}
else if (Rs >= 233) {
return 8;
}
}
else if (posBalance >= 51) {
if (Rs < 540) {
return 2;
}
else if (Rs >= 540) {
return 1;
}
}
}
}
else if (ssDCval >= 41) {
if (TSC < 4) {
return 4;
}
else if (TSC >= 4) {
return 8;
}
}
}
else if (negBalance >= 38) {
if (Rs < 717) {
if (tcor < 83) {
return 2;
}
else if (tcor >= 83) {
if (Rs < 679) {
return 8;
}
else if (Rs >= 679) {
return 2;
}
}
}
else if (Rs >= 717) {
if (Cs < 699) {
return 2;
}
else if (Cs >= 699) {
if (diffAFD < 2) {
if (Cs < 1216) {
return 8;
}
else if (Cs >= 1216) {
return 1;
}
}
else if (diffAFD >= 2) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 51) {
if (Cs < 813) {
if (refDCval < 83) {
if (Rs < 2246) {
if (AFD < 8) {
if (refDCval < 72) {
if (Rs < 1810) {
if (diffRsCsdiff < -15) {
if (Rs < 1541) {
return 1;
}
else if (Rs >= 1541) {
return 4;
}
}
else if (diffRsCsdiff >= -15) {
if (tcor < 94) {
if (Cs < 505) {
return 1;
}
else if (Cs >= 505) {
if (tcor < 86) {
return 8;
}
else if (tcor >= 86) {
return 4;
}
}
}
else if (tcor >= 94) {
if (Cs < 733) {
return 4;
}
else if (Cs >= 733) {
return 8;
}
}
}
}
else if (Rs >= 1810) {
return 2;
}
}
else if (refDCval >= 72) {
if (Rs < 646) {
if (Rs < 574) {
return 8;
}
else if (Rs >= 574) {
return 2;
}
}
else if (Rs >= 646) {
return 1;
}
}
}
else if (AFD >= 8) {
if (RsDiff < 3) {
if (posBalance < 51) {
if (Rs < 664) {
return 2;
}
else if (Rs >= 664) {
return 1;
}
}
else if (posBalance >= 51) {
if (Rs < 390) {
return 1;
}
else if (Rs >= 390) {
if (MVDiff < 34) {
return 8;
}
else if (MVDiff >= 34) {
return 1;
}
}
}
}
else if (RsDiff >= 3) {
if (posBalance < 54) {
if (mcTcor < 94) {
if (ssDCval < 73) {
if (Rs < 1515) {
if (Rs < 1020) {
return 2;
}
else if (Rs >= 1020) {
return 1;
}
}
else if (Rs >= 1515) {
return 4;
}
}
else if (ssDCval >= 73) {
return 1;
}
}
else if (mcTcor >= 94) {
if (Rs < 664) {
return 8;
}
else if (Rs >= 664) {
if (RsCsDiff < 56) {
return 1;
}
else if (RsCsDiff >= 56) {
if (Cs < 604) {
return 1;
}
else if (Cs >= 604) {
return 2;
}
}
}
}
}
else if (posBalance >= 54) {
if (RsCsDiff < 23) {
if (Rs < 781) {
if (Cs < 626) {
return 1;
}
else if (Cs >= 626) {
return 2;
}
}
else if (Rs >= 781) {
if (Rs < 1169) {
if (diffTSC < -1) {
return 1;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (Rs >= 1169) {
return 1;
}
}
}
else if (RsCsDiff >= 23) {
if (Cs < 777) {
if (Cs < 588) {
if (diffMVdiffVal < -4) {
if (Rs < 1447) {
return 4;
}
else if (Rs >= 1447) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
return 1;
}
}
else if (Cs >= 588) {
if (tcor < 77) {
return 1;
}
else if (tcor >= 77) {
return 2;
}
}
}
else if (Cs >= 777) {
if (Rs < 1470) {
return 4;
}
else if (Rs >= 1470) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 2246) {
return 2;
}
}
else if (refDCval >= 83) {
if (mcTcor < 91) {
return 8;
}
else if (mcTcor >= 91) {
if (Rs < 770) {
if (ssDCval < 119) {
return 2;
}
else if (ssDCval >= 119) {
if (Rs < 767) {
return 8;
}
else if (Rs >= 767) {
return 2;
}
}
}
else if (Rs >= 770) {
if (negBalance < 45) {
if (negBalance < 34) {
if (RsDiff < 4) {
return 2;
}
else if (RsDiff >= 4) {
return 1;
}
}
else if (negBalance >= 34) {
if (negBalance < 41) {
return 2;
}
else if (negBalance >= 41) {
if (Rs < 1942) {
if (Rs < 1628) {
return 1;
}
else if (Rs >= 1628) {
return 8;
}
}
else if (Rs >= 1942) {
return 2;
}
}
}
}
else if (negBalance >= 45) {
if (Cs < 726) {
return 1;
}
else if (Cs >= 726) {
if (ssDCval < 90) {
return 4;
}
else if (ssDCval >= 90) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 813) {
if (Rs < 1389) {
if (MVDiff < 5) {
if (negBalance < 32) {
return 4;
}
else if (negBalance >= 32) {
if (Rs < 999) {
if (Cs < 1241) {
if (refDCval < 87) {
return 8;
}
else if (refDCval >= 87) {
if (Rs < 826) {
return 4;
}
else if (Rs >= 826) {
return 8;
}
}
}
else if (Cs >= 1241) {
return 2;
}
}
else if (Rs >= 999) {
if (negBalance < 43) {
if (Cs < 1058) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 1;
}
}
else if (Cs >= 1058) {
if (Rs < 1073) {
return 8;
}
else if (Rs >= 1073) {
return 4;
}
}
}
else if (negBalance >= 43) {
return 8;
}
}
}
}
else if (MVDiff >= 5) {
if (tcor < 80) {
if (Cs < 948) {
if (Rs < 885) {
return 1;
}
else if (Rs >= 885) {
return 8;
}
}
else if (Cs >= 948) {
if (MVDiff < 30) {
if (ssDCval < 138) {
if (gchDC < 2) {
return 1;
}
else if (gchDC >= 2) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
return 8;
}
}
}
else if (ssDCval >= 138) {
return 8;
}
}
else if (MVDiff >= 30) {
if (Rs < 963) {
return 4;
}
else if (Rs >= 963) {
return 2;
}
}
}
}
else if (tcor >= 80) {
if (mcTcor < 91) {
if (RsCsDiff < 33) {
if (Cs < 951) {
return 8;
}
else if (Cs >= 951) {
return 4;
}
}
else if (RsCsDiff >= 33) {
if (Rs < 1233) {
return 8;
}
else if (Rs >= 1233) {
return 2;
}
}
}
else if (mcTcor >= 91) {
if (posBalance < 45) {
return 1;
}
else if (posBalance >= 45) {
if (ssDCval < 66) {
if (TSC < 5) {
if (Rs < 1102) {
return 4;
}
else if (Rs >= 1102) {
return 8;
}
}
else if (TSC >= 5) {
if (Rs < 798) {
return 2;
}
else if (Rs >= 798) {
if (Rs < 1220) {
return 4;
}
else if (Rs >= 1220) {
return 2;
}
}
}
}
else if (ssDCval >= 66) {
if (negBalance < 36) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
return 2;
}
}
else if (negBalance >= 36) {
if (mu_mv_mag_sq < 7) {
if (mcTcor < 93) {
if (posBalance < 61) {
if (Rs < 1071) {
return 8;
}
else if (Rs >= 1071) {
return 2;
}
}
else if (posBalance >= 61) {
return 4;
}
}
else if (mcTcor >= 93) {
if (Rs < 1118) {
return 2;
}
else if (Rs >= 1118) {
if (Rs < 1235) {
return 4;
}
else if (Rs >= 1235) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (Cs < 1073) {
return 2;
}
else if (Cs >= 1073) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 1389) {
if (ssDCval < 61) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
if (ssDCval < 57) {
if (Rs < 1652) {
return 8;
}
else if (Rs >= 1652) {
return 4;
}
}
else if (ssDCval >= 57) {
return 8;
}
}
}
}
else if (ssDCval >= 61) {
if (ssDCval < 74) {
if (mcTcor < 94) {
if (negBalance < 40) {
return 2;
}
else if (negBalance >= 40) {
return 4;
}
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (ssDCval >= 74) {
if (Rs < 2102) {
if (ssDCval < 87) {
if (Cs < 1007) {
return 1;
}
else if (Cs >= 1007) {
if (Cs < 1103) {
return 4;
}
else if (Cs >= 1103) {
return 8;
}
}
}
else if (ssDCval >= 87) {
if (RsCsDiff < 85) {
if (SC < 2869) {
return 8;
}
else if (SC >= 2869) {
if (tcor < 89) {
return 2;
}
else if (tcor >= 89) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 2;
}
}
}
}
else if (RsCsDiff >= 85) {
return 2;
}
}
}
else if (Rs >= 2102) {
if (Cs < 1112) {
if (TSC < 8) {
return 4;
}
else if (TSC >= 8) {
return 8;
}
}
else if (Cs >= 1112) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 927) {
if (MVDiff < 5) {
if (RsCsDiff < 11) {
if (refDCval < 110) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
if (diffMVdiffVal < -100) {
return 8;
}
else if (diffMVdiffVal >= -100) {
return 1;
}
}
}
else if (refDCval >= 110) {
if (tcor < 87) {
if (Rs < 563) {
return 1;
}
else if (Rs >= 563) {
return 8;
}
}
else if (tcor >= 87) {
return 2;
}
}
}
else if (RsCsDiff >= 11) {
if (mcTcor < 86) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
if (diffMVdiffVal < -7) {
return 8;
}
else if (diffMVdiffVal >= -7) {
return 4;
}
}
}
else if (mcTcor >= 86) {
if (ssDCval < 96) {
if (ssDCval < 60) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
else if (ssDCval >= 60) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (ssDCval >= 96) {
return 1;
}
}
}
}
else if (MVDiff >= 5) {
if (Rs < 375) {
if (Rs < 219) {
return 8;
}
else if (Rs >= 219) {
return 1;
}
}
else if (Rs >= 375) {
if (TSC < 6) {
if (RsCsDiff < 29) {
if (ssDCval < 77) {
if (posBalance < 56) {
if (AFD < 12) {
if (diffTSC < -3) {
return 4;
}
else if (diffTSC >= -3) {
return 8;
}
}
else if (AFD >= 12) {
return 1;
}
}
else if (posBalance >= 56) {
if (Rs < 689) {
if (MVDiff < 23) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
return 1;
}
}
else if (MVDiff >= 23) {
return 2;
}
}
else if (Rs >= 689) {
if (Rs < 731) {
return 8;
}
else if (Rs >= 731) {
return 1;
}
}
}
}
else if (ssDCval >= 77) {
if (posBalance < 47) {
return 2;
}
else if (posBalance >= 47) {
if (MVDiff < 14) {
if (Rs < 887) {
return 8;
}
else if (Rs >= 887) {
return 1;
}
}
else if (MVDiff >= 14) {
if (Cs < 509) {
return 2;
}
else if (Cs >= 509) {
if (RsCsDiff < 11) {
return 8;
}
else if (RsCsDiff >= 11) {
if (Rs < 563) {
return 1;
}
else if (Rs >= 563) {
if (posBalance < 61) {
return 2;
}
else if (posBalance >= 61) {
return 8;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 29) {
if (Cs < 826) {
if (CsDiff < 4) {
if (Cs < 606) {
return 4;
}
else if (Cs >= 606) {
return 2;
}
}
else if (CsDiff >= 4) {
return 2;
}
}
else if (Cs >= 826) {
return 4;
}
}
}
else if (TSC >= 6) {
if (AFD < 11) {
if (refDCval < 48) {
if (Rs < 813) {
if (MVDiff < 6) {
if (Rs < 787) {
return 1;
}
else if (Rs >= 787) {
return 8;
}
}
else if (MVDiff >= 6) {
return 8;
}
}
else if (Rs >= 813) {
if (posBalance < 59) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Rs < 860) {
return 1;
}
else if (Rs >= 860) {
return 8;
}
}
}
else if (posBalance >= 59) {
if (Cs < 786) {
return 1;
}
else if (Cs >= 786) {
if (Rs < 902) {
return 8;
}
else if (Rs >= 902) {
return 1;
}
}
}
}
}
else if (refDCval >= 48) {
if (CsDiff < 6) {
if (diffAFD < -2) {
return 4;
}
else if (diffAFD >= -2) {
if (MVDiff < 7) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (MVDiff >= 7) {
return 8;
}
}
}
else if (CsDiff >= 6) {
return 4;
}
}
}
else if (AFD >= 11) {
if (Cs < 672) {
if (MVDiff < 8) {
if (Rs < 669) {
return 4;
}
else if (Rs >= 669) {
return 1;
}
}
else if (MVDiff >= 8) {
if (diffRsCsdiff < 3) {
if (Cs < 503) {
return 2;
}
else if (Cs >= 503) {
return 8;
}
}
else if (diffRsCsdiff >= 3) {
if (RsDiff < 4) {
if (AFD < 12) {
return 8;
}
else if (AFD >= 12) {
return 4;
}
}
else if (RsDiff >= 4) {
return 4;
}
}
}
}
else if (Cs >= 672) {
if (posBalance < 59) {
if (negBalance < 54) {
return 1;
}
else if (negBalance >= 54) {
if (AFD < 12) {
return 4;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (posBalance >= 59) {
if (ssDCval < 49) {
if (ssDCval < 45) {
return 4;
}
else if (ssDCval >= 45) {
return 8;
}
}
else if (ssDCval >= 49) {
return 4;
}
}
}
}
}
}
}
}
else if (Rs >= 927) {
if (MVDiff < 18) {
if (tcor < 82) {
if (Rs < 1304) {
if (mcTcor < 90) {
if (tcor < 70) {
if (MVDiff < 3) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
return 8;
}
}
else if (MVDiff >= 3) {
if (ssDCval < 112) {
return 8;
}
else if (ssDCval >= 112) {
return 1;
}
}
}
else if (tcor >= 70) {
return 8;
}
}
else if (mcTcor >= 90) {
if (negBalance < 63) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 2;
}
}
else if (negBalance >= 63) {
return 2;
}
}
}
else if (Rs >= 1304) {
if (CsDiff < 4) {
if (ssDCval < 99) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (MVDiff < 12) {
return 8;
}
else if (MVDiff >= 12) {
if (diffTSC < -1) {
if (Rs < 1566) {
return 8;
}
else if (Rs >= 1566) {
return 2;
}
}
else if (diffTSC >= -1) {
if (diffRsCsdiff < -13) {
return 8;
}
else if (diffRsCsdiff >= -13) {
return 2;
}
}
}
}
}
}
else if (ssDCval >= 99) {
return 8;
}
}
else if (CsDiff >= 4) {
if (AFD < 12) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (Rs < 1489) {
return 2;
}
else if (Rs >= 1489) {
return 8;
}
}
}
else if (AFD >= 12) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (ssDCval < 87) {
return 8;
}
else if (ssDCval >= 87) {
return 4;
}
}
}
}
}
}
else if (tcor >= 82) {
if (MVDiff < 8) {
if (ssDCval < 62) {
if (Cs < 1188) {
if (Rs < 1008) {
if (Rs < 987) {
return 8;
}
else if (Rs >= 987) {
return 4;
}
}
else if (Rs >= 1008) {
return 8;
}
}
else if (Cs >= 1188) {
if (Rs < 1655) {
return 4;
}
else if (Rs >= 1655) {
return 8;
}
}
}
else if (ssDCval >= 62) {
if (Cs < 832) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (ssDCval < 81) {
return 2;
}
else if (ssDCval >= 81) {
if (Rs < 2038) {
return 1;
}
else if (Rs >= 2038) {
return 8;
}
}
}
}
else if (Cs >= 832) {
if (negBalance < 53) {
if (Rs < 987) {
if (Rs < 954) {
if (Rs < 936) {
return 4;
}
else if (Rs >= 936) {
return 8;
}
}
else if (Rs >= 954) {
return 1;
}
}
else if (Rs >= 987) {
if (posBalance < 47) {
return 4;
}
else if (posBalance >= 47) {
if (TSC < 11) {
if (posBalance < 53) {
if (Cs < 1136) {
return 8;
}
else if (Cs >= 1136) {
return 1;
}
}
else if (posBalance >= 53) {
return 8;
}
}
else if (TSC >= 11) {
return 1;
}
}
}
}
else if (negBalance >= 53) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (CsDiff < 4) {
return 1;
}
else if (CsDiff >= 4) {
if (diffMVdiffVal < 2) {
if (TSC < 6) {
return 2;
}
else if (TSC >= 6) {
return 8;
}
}
else if (diffMVdiffVal >= 2) {
if (Rs < 1425) {
return 1;
}
else if (Rs >= 1425) {
if (Rs < 1875) {
return 8;
}
else if (Rs >= 1875) {
return 1;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 8) {
if (mu_mv_mag_sq < 3) {
if (AFD < 11) {
if (posBalance < 57) {
if (Rs < 942) {
return 8;
}
else if (Rs >= 942) {
if (refDCval < 91) {
return 4;
}
else if (refDCval >= 91) {
if (Rs < 1680) {
return 8;
}
else if (Rs >= 1680) {
return 4;
}
}
}
}
else if (posBalance >= 57) {
if (tcor < 89) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 4;
}
}
else if (tcor >= 89) {
return 8;
}
}
}
else if (AFD >= 11) {
if (Cs < 960) {
if (Cs < 852) {
if (Cs < 680) {
return 1;
}
else if (Cs >= 680) {
return 2;
}
}
else if (Cs >= 852) {
if (Rs < 1102) {
return 4;
}
else if (Rs >= 1102) {
return 8;
}
}
}
else if (Cs >= 960) {
if (ssDCval < 63) {
return 1;
}
else if (ssDCval >= 63) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (mcTcor < 93) {
if (Rs < 1526) {
return 8;
}
else if (Rs >= 1526) {
if (Rs < 1866) {
return 1;
}
else if (Rs >= 1866) {
return 8;
}
}
}
else if (mcTcor >= 93) {
return 8;
}
}
}
}
}
else if (MVDiff >= 18) {
if (ssDCval < 70) {
if (posBalance < 51) {
if (MVDiff < 40) {
return 4;
}
else if (MVDiff >= 40) {
if (Rs < 1460) {
return 4;
}
else if (Rs >= 1460) {
return 8;
}
}
}
else if (posBalance >= 51) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (negBalance < 51) {
return 4;
}
else if (negBalance >= 51) {
return 8;
}
}
}
}
else if (ssDCval >= 70) {
if (mcTcor < 93) {
if (ssDCval < 129) {
if (Cs < 832) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
if (CsDiff < 3) {
if (Rs < 1418) {
return 2;
}
else if (Rs >= 1418) {
return 8;
}
}
else if (CsDiff >= 3) {
if (TSC < 7) {
if (Rs < 1296) {
return 1;
}
else if (Rs >= 1296) {
return 8;
}
}
else if (TSC >= 7) {
return 1;
}
}
}
}
else if (Cs >= 832) {
if (ssDCval < 112) {
if (Cs < 993) {
return 8;
}
else if (Cs >= 993) {
return 2;
}
}
else if (ssDCval >= 112) {
return 4;
}
}
}
else if (ssDCval >= 129) {
if (Rs < 1364) {
return 2;
}
else if (Rs >= 1364) {
return 4;
}
}
}
else if (mcTcor >= 93) {
if (CsDiff < 3) {
if (tcor < 90) {
return 2;
}
else if (tcor >= 90) {
return 1;
}
}
else if (CsDiff >= 3) {
if (ssDCval < 106) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
if (Rs < 1179) {
return 4;
}
else if (Rs >= 1179) {
return 8;
}
}
}
else if (ssDCval >= 106) {
return 8;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (negBalance < 44) {
if (ssDCval < 78) {
if (ssDCval < 35) {
if (Rs < 581) {
if (Cs < 577) {
return 2;
}
else if (Cs >= 577) {
return 1;
}
}
else if (Rs >= 581) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
return 1;
}
}
}
else if (ssDCval >= 35) {
if (RsCsDiff < 15) {
if (Rs < 880) {
if (refDCval < 37) {
if (posBalance < 50) {
return 8;
}
else if (posBalance >= 50) {
return 1;
}
}
else if (refDCval >= 37) {
if (Rs < 497) {
if (ssDCval < 60) {
if (Cs < 721) {
if (Cs < 658) {
if (SC < 929) {
return 1;
}
else if (SC >= 929) {
if (Cs < 532) {
if (Rs < 463) {
return 8;
}
else if (Rs >= 463) {
return 1;
}
}
else if (Cs >= 532) {
return 2;
}
}
}
else if (Cs >= 658) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (Cs >= 721) {
if (negBalance < 23) {
if (Rs < 424) {
return 2;
}
else if (Rs >= 424) {
return 1;
}
}
else if (negBalance >= 23) {
if (MVDiff < 14) {
if (ssDCval < 46) {
return 4;
}
else if (ssDCval >= 46) {
return 1;
}
}
else if (MVDiff >= 14) {
return 2;
}
}
}
}
else if (ssDCval >= 60) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Rs < 469) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
else if (Rs >= 469) {
return 8;
}
}
}
}
else if (Rs >= 497) {
if (negBalance < 30) {
if (negBalance < 24) {
if (posBalance < 48) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
return 1;
}
}
else if (posBalance >= 48) {
return 1;
}
}
else if (negBalance >= 24) {
if (mu_mv_mag_sq < 105) {
if (tcor < 94) {
if (Rs < 801) {
return 4;
}
else if (Rs >= 801) {
return 2;
}
}
else if (tcor >= 94) {
if (Rs < 611) {
return 2;
}
else if (Rs >= 611) {
if (Rs < 746) {
return 1;
}
else if (Rs >= 746) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 105) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
return 4;
}
}
}
}
else if (negBalance >= 30) {
if (mu_mv_mag_sq < 119) {
if (AFD < 5) {
if (TSC < 3) {
if (mu_mv_mag_sq < 94) {
return 2;
}
else if (mu_mv_mag_sq >= 94) {
return 4;
}
}
else if (TSC >= 3) {
if (posBalance < 53) {
if (mu_mv_mag_sq < 85) {
if (Rs < 608) {
return 8;
}
else if (Rs >= 608) {
return 4;
}
}
else if (mu_mv_mag_sq >= 85) {
if (Cs < 1056) {
if (mu_mv_mag_sq < 114) {
return 8;
}
else if (mu_mv_mag_sq >= 114) {
if (Rs < 724) {
return 4;
}
else if (Rs >= 724) {
return 8;
}
}
}
else if (Cs >= 1056) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (MVDiff < 5) {
if (Rs < 767) {
return 4;
}
else if (Rs >= 767) {
if (Rs < 799) {
return 8;
}
else if (Rs >= 799) {
return 4;
}
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (AFD >= 5) {
if (negBalance < 37) {
if (mu_mv_mag_sq < 83) {
if (Rs < 827) {
if (Rs < 560) {
if (Cs < 924) {
return 8;
}
else if (Cs >= 924) {
return 1;
}
}
else if (Rs >= 560) {
if (Cs < 531) {
return 2;
}
else if (Cs >= 531) {
return 1;
}
}
}
else if (Rs >= 827) {
return 2;
}
}
else if (mu_mv_mag_sq >= 83) {
if (RsDiff < 3) {
if (Cs < 530) {
return 4;
}
else if (Cs >= 530) {
if (Rs < 876) {
return 8;
}
else if (Rs >= 876) {
return 4;
}
}
}
else if (RsDiff >= 3) {
return 1;
}
}
}
else if (negBalance >= 37) {
if (Rs < 795) {
if (refDCval < 55) {
if (mu_mv_mag_sq < 83) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 83) {
return 8;
}
}
else if (refDCval >= 55) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (Rs < 716) {
return 8;
}
else if (Rs >= 716) {
return 1;
}
}
}
}
else if (Rs >= 795) {
if (posBalance < 57) {
if (Cs < 766) {
return 8;
}
else if (Cs >= 766) {
if (Cs < 984) {
if (Rs < 801) {
return 4;
}
else if (Rs >= 801) {
return 2;
}
}
else if (Cs >= 984) {
if (Rs < 831) {
return 1;
}
else if (Rs >= 831) {
return 8;
}
}
}
}
else if (posBalance >= 57) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 119) {
return 8;
}
}
}
}
}
else if (Rs >= 880) {
if (RsCsDiff < 7) {
if (ssDCval < 65) {
if (Cs < 509) {
if (Rs < 1041) {
return 8;
}
else if (Rs >= 1041) {
return 2;
}
}
else if (Cs >= 509) {
if (negBalance < 15) {
return 1;
}
else if (negBalance >= 15) {
if (ssDCval < 60) {
if (negBalance < 32) {
if (AFD < 4) {
if (Cs < 557) {
if (Rs < 1026) {
return 8;
}
else if (Rs >= 1026) {
return 1;
}
}
else if (Cs >= 557) {
return 8;
}
}
else if (AFD >= 4) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
if (Rs < 1061) {
if (Rs < 939) {
return 1;
}
else if (Rs >= 939) {
return 8;
}
}
else if (Rs >= 1061) {
return 1;
}
}
}
}
else if (negBalance >= 32) {
return 8;
}
}
else if (ssDCval >= 60) {
if (Cs < 1068) {
if (Rs < 975) {
if (Cs < 601) {
return 8;
}
else if (Cs >= 601) {
return 1;
}
}
else if (Rs >= 975) {
return 8;
}
}
else if (Cs >= 1068) {
if (Rs < 1119) {
return 8;
}
else if (Rs >= 1119) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 65) {
if (SC < 1960) {
if (Rs < 1140) {
return 4;
}
else if (Rs >= 1140) {
return 1;
}
}
else if (SC >= 1960) {
return 8;
}
}
}
else if (RsCsDiff >= 7) {
if (Rs < 970) {
if (ssDCval < 55) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 8;
}
}
else if (ssDCval >= 55) {
if (mu_mv_mag_sq < 71) {
if (ssDCval < 62) {
if (Rs < 957) {
return 4;
}
else if (Rs >= 957) {
return 1;
}
}
else if (ssDCval >= 62) {
return 8;
}
}
else if (mu_mv_mag_sq >= 71) {
if (posBalance < 49) {
return 8;
}
else if (posBalance >= 49) {
return 4;
}
}
}
}
else if (Rs >= 970) {
if (mu_mv_mag_sq < 112) {
if (posBalance < 46) {
if (MVDiff < 9) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (ssDCval < 45) {
return 4;
}
else if (ssDCval >= 45) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
}
}
else if (MVDiff >= 9) {
if (AFD < 5) {
if (Rs < 1104) {
return 1;
}
else if (Rs >= 1104) {
return 4;
}
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (posBalance >= 46) {
if (refDCval < 60) {
if (tcor < 78) {
if (Cs < 575) {
return 8;
}
else if (Cs >= 575) {
return 1;
}
}
else if (tcor >= 78) {
if (Cs < 1279) {
if (Cs < 855) {
if (Rs < 1333) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (Rs >= 1333) {
return 2;
}
}
else if (Cs >= 855) {
if (negBalance < 26) {
return 4;
}
else if (negBalance >= 26) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (Rs < 1173) {
return 4;
}
else if (Rs >= 1173) {
return 8;
}
}
}
}
}
else if (Cs >= 1279) {
return 4;
}
}
}
else if (refDCval >= 60) {
if (refDCval < 70) {
if (Rs < 1312) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
if (MVDiff < 5) {
if (Cs < 722) {
return 8;
}
else if (Cs >= 722) {
return 1;
}
}
else if (MVDiff >= 5) {
if (mcTcor < 94) {
return 1;
}
else if (mcTcor >= 94) {
return 8;
}
}
}
}
else if (Rs >= 1312) {
if (SC < 2674) {
return 8;
}
else if (SC >= 2674) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 8;
}
}
}
}
else if (refDCval >= 70) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (Cs < 1044) {
return 8;
}
else if (Cs >= 1044) {
return 4;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 112) {
if (tcor < 96) {
if (tcor < 87) {
if (CsDiff < 3) {
return 8;
}
else if (CsDiff >= 3) {
return 1;
}
}
else if (tcor >= 87) {
if (Rs < 1684) {
if (diffRsCsdiff < -3) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Rs < 1320) {
if (Cs < 1040) {
return 8;
}
else if (Cs >= 1040) {
return 4;
}
}
else if (Rs >= 1320) {
return 8;
}
}
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
else if (Rs >= 1684) {
if (Rs < 1985) {
return 4;
}
else if (Rs >= 1985) {
return 8;
}
}
}
}
else if (tcor >= 96) {
if (TSC < 4) {
if (diffAFD < -1) {
if (SC < 2303) {
return 8;
}
else if (SC >= 2303) {
return 4;
}
}
else if (diffAFD >= -1) {
return 1;
}
}
else if (TSC >= 4) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 4;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 15) {
if (Cs < 913) {
if (mu_mv_mag_sq < 107) {
if (refDCval < 54) {
if (mcTcor < 95) {
if (Rs < 1602) {
if (Rs < 1009) {
if (ssDCval < 48) {
if (ssDCval < 43) {
if (MVDiff < 19) {
if (Rs < 569) {
return 1;
}
else if (Rs >= 569) {
return 8;
}
}
else if (MVDiff >= 19) {
return 2;
}
}
else if (ssDCval >= 43) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 8;
}
}
}
else if (ssDCval >= 48) {
if (refDCval < 48) {
return 1;
}
else if (refDCval >= 48) {
return 4;
}
}
}
else if (Rs >= 1009) {
if (negBalance < 35) {
if (Rs < 1185) {
return 8;
}
else if (Rs >= 1185) {
return 2;
}
}
else if (negBalance >= 35) {
return 8;
}
}
}
else if (Rs >= 1602) {
return 4;
}
}
else if (mcTcor >= 95) {
if (Cs < 588) {
if (ssDCval < 51) {
return 4;
}
else if (ssDCval >= 51) {
if (Rs < 1267) {
return 2;
}
else if (Rs >= 1267) {
return 1;
}
}
}
else if (Cs >= 588) {
return 1;
}
}
}
else if (refDCval >= 54) {
if (mu_mv_mag_sq < 97) {
if (negBalance < 32) {
if (mu_mv_mag_sq < 82) {
if (diffRsCsdiff < 14) {
if (Cs < 532) {
if (Rs < 655) {
return 8;
}
else if (Rs >= 655) {
return 2;
}
}
else if (Cs >= 532) {
if (diffRsCsdiff < -32) {
return 2;
}
else if (diffRsCsdiff >= -32) {
if (MVDiff < 3) {
if (Rs < 1855) {
return 2;
}
else if (Rs >= 1855) {
return 1;
}
}
else if (MVDiff >= 3) {
return 1;
}
}
}
}
else if (diffRsCsdiff >= 14) {
if (Cs < 510) {
return 1;
}
else if (Cs >= 510) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 82) {
return 2;
}
}
else if (negBalance >= 32) {
if (RsCsDiff < 33) {
if (tcor < 90) {
if (Rs < 893) {
return 1;
}
else if (Rs >= 893) {
if (ssDCval < 69) {
if (Rs < 1189) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (Rs >= 1189) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
return 2;
}
}
}
else if (ssDCval >= 69) {
if (Rs < 971) {
return 8;
}
else if (Rs >= 971) {
return 1;
}
}
}
}
else if (tcor >= 90) {
if (AFD < 7) {
return 1;
}
else if (AFD >= 7) {
if (Rs < 1415) {
return 4;
}
else if (Rs >= 1415) {
return 1;
}
}
}
}
else if (RsCsDiff >= 33) {
if (mu_mv_mag_sq < 43) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
if (negBalance < 37) {
return 1;
}
else if (negBalance >= 37) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (Rs < 1907) {
if (posBalance < 53) {
if (MVDiff < 13) {
if (Rs < 1285) {
return 8;
}
else if (Rs >= 1285) {
return 1;
}
}
else if (MVDiff >= 13) {
return 2;
}
}
else if (posBalance >= 53) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 8;
}
}
}
else if (Rs >= 1907) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (tcor < 92) {
if (Rs < 696) {
return 8;
}
else if (Rs >= 696) {
return 4;
}
}
else if (tcor >= 92) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 107) {
if (tcor < 89) {
return 1;
}
else if (tcor >= 89) {
if (Cs < 713) {
return 1;
}
else if (Cs >= 713) {
return 8;
}
}
}
}
else if (Cs >= 913) {
if (refDCval < 60) {
if (SC < 2385) {
if (refDCval < 47) {
if (Rs < 873) {
if (Rs < 535) {
return 1;
}
else if (Rs >= 535) {
return 2;
}
}
else if (Rs >= 873) {
if (AFD < 5) {
if (Rs < 1344) {
return 1;
}
else if (Rs >= 1344) {
return 8;
}
}
else if (AFD >= 5) {
return 8;
}
}
}
else if (refDCval >= 47) {
if (posBalance < 52) {
if (Rs < 762) {
return 2;
}
else if (Rs >= 762) {
return 4;
}
}
else if (posBalance >= 52) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
if (Rs < 1206) {
if (Rs < 481) {
return 1;
}
else if (Rs >= 481) {
if (diffAFD < 1) {
if (diffMVdiffVal < -3) {
return 8;
}
else if (diffMVdiffVal >= -3) {
if (negBalance < 36) {
if (Cs < 960) {
return 4;
}
else if (Cs >= 960) {
return 8;
}
}
else if (negBalance >= 36) {
return 4;
}
}
}
else if (diffAFD >= 1) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
return 8;
}
}
}
}
else if (Rs >= 1206) {
if (ssDCval < 56) {
if (Rs < 1355) {
return 8;
}
else if (Rs >= 1355) {
if (Rs < 1386) {
return 2;
}
else if (Rs >= 1386) {
return 8;
}
}
}
else if (ssDCval >= 56) {
return 2;
}
}
}
}
}
}
else if (SC >= 2385) {
if (posBalance < 45) {
if (MVDiff < 12) {
if (Rs < 1941) {
if (Rs < 1647) {
return 2;
}
else if (Rs >= 1647) {
return 8;
}
}
else if (Rs >= 1941) {
return 4;
}
}
else if (MVDiff >= 12) {
return 2;
}
}
else if (posBalance >= 45) {
if (RsCsDiff < 70) {
if (Rs < 1225) {
if (Rs < 1212) {
return 8;
}
else if (Rs >= 1212) {
return 4;
}
}
else if (Rs >= 1225) {
if (diffMVdiffVal < -4) {
if (mu_mv_mag_sq < 64) {
return 8;
}
else if (mu_mv_mag_sq >= 64) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
if (Cs < 1008) {
return 2;
}
else if (Cs >= 1008) {
return 8;
}
}
}
}
else if (RsCsDiff >= 70) {
if (mcTcor < 91) {
if (MVDiff < 14) {
return 8;
}
else if (MVDiff >= 14) {
if (Rs < 1922) {
return 1;
}
else if (Rs >= 1922) {
return 4;
}
}
}
else if (mcTcor >= 91) {
return 4;
}
}
}
}
}
else if (refDCval >= 60) {
if (ssDCval < 73) {
if (negBalance < 36) {
if (Rs < 1765) {
if (Rs < 923) {
if (Rs < 800) {
return 1;
}
else if (Rs >= 800) {
return 2;
}
}
else if (Rs >= 923) {
if (diffTSC < -1) {
if (Rs < 1395) {
return 4;
}
else if (Rs >= 1395) {
return 1;
}
}
else if (diffTSC >= -1) {
if (Cs < 1124) {
return 8;
}
else if (Cs >= 1124) {
if (SC < 2443) {
if (posBalance < 55) {
return 4;
}
else if (posBalance >= 55) {
if (Rs < 1040) {
return 8;
}
else if (Rs >= 1040) {
return 4;
}
}
}
else if (SC >= 2443) {
return 8;
}
}
}
}
}
else if (Rs >= 1765) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
else if (negBalance >= 36) {
if (Cs < 1262) {
if (MVDiff < 15) {
if (mcTcor < 95) {
if (MVDiff < 2) {
if (diffRsCsdiff < 9) {
if (posBalance < 57) {
return 8;
}
else if (posBalance >= 57) {
return 1;
}
}
else if (diffRsCsdiff >= 9) {
return 4;
}
}
else if (MVDiff >= 2) {
if (Cs < 1176) {
if (mu_mv_mag_sq < 43) {
return 8;
}
else if (mu_mv_mag_sq >= 43) {
return 4;
}
}
else if (Cs >= 1176) {
if (ssDCval < 61) {
return 1;
}
else if (ssDCval >= 61) {
if (diffRsCsdiff < -18) {
return 8;
}
else if (diffRsCsdiff >= -18) {
return 4;
}
}
}
}
}
else if (mcTcor >= 95) {
if (posBalance < 54) {
if (diffMVdiffVal < 1) {
if (diffRsCsdiff < 3) {
return 4;
}
else if (diffRsCsdiff >= 3) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (diffMVdiffVal >= 1) {
if (Cs < 936) {
return 8;
}
else if (Cs >= 936) {
if (Cs < 1234) {
return 2;
}
else if (Cs >= 1234) {
return 4;
}
}
}
}
else if (posBalance >= 54) {
if (Rs < 1263) {
return 4;
}
else if (Rs >= 1263) {
return 1;
}
}
}
}
else if (MVDiff >= 15) {
return 8;
}
}
else if (Cs >= 1262) {
if (negBalance < 42) {
if (Rs < 1468) {
return 4;
}
else if (Rs >= 1468) {
return 8;
}
}
else if (negBalance >= 42) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 4;
}
}
}
}
}
else if (ssDCval >= 73) {
if (Rs < 1514) {
if (Cs < 1231) {
return 8;
}
else if (Cs >= 1231) {
return 4;
}
}
else if (Rs >= 1514) {
if (mu_mv_mag_sq < 108) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
return 8;
}
}
else if (mu_mv_mag_sq >= 108) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 78) {
if (ssDCval < 155) {
if (refDCval < 100) {
if (Rs < 1748) {
if (TSC < 5) {
if (mu_mv_mag_sq < 70) {
if (tcor < 91) {
return 2;
}
else if (tcor >= 91) {
return 1;
}
}
else if (mu_mv_mag_sq >= 70) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (SC < 1341) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (SC >= 1341) {
if (ssDCval < 98) {
if (Cs < 979) {
if (Cs < 865) {
return 2;
}
else if (Cs >= 865) {
if (tcor < 93) {
return 2;
}
else if (tcor >= 93) {
return 1;
}
}
}
else if (Cs >= 979) {
if (Rs < 1718) {
return 8;
}
else if (Rs >= 1718) {
return 2;
}
}
}
else if (ssDCval >= 98) {
return 8;
}
}
}
}
}
else if (TSC >= 5) {
if (RsCsDiff < 23) {
if (Cs < 1032) {
if (ssDCval < 100) {
return 8;
}
else if (ssDCval >= 100) {
if (Rs < 411) {
return 2;
}
else if (Rs >= 411) {
if (Rs < 505) {
return 1;
}
else if (Rs >= 505) {
return 8;
}
}
}
}
else if (Cs >= 1032) {
if (Rs < 842) {
if (Rs < 546) {
return 8;
}
else if (Rs >= 546) {
return 4;
}
}
else if (Rs >= 842) {
return 8;
}
}
}
else if (RsCsDiff >= 23) {
if (AFD < 11) {
if (diffMVdiffVal < -3) {
if (SC < 2648) {
if (Rs < 525) {
return 2;
}
else if (Rs >= 525) {
return 8;
}
}
else if (SC >= 2648) {
return 2;
}
}
else if (diffMVdiffVal >= -3) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
if (Rs < 1447) {
return 8;
}
else if (Rs >= 1447) {
return 2;
}
}
}
}
else if (AFD >= 11) {
if (MVDiff < 13) {
if (Cs < 1025) {
return 8;
}
else if (Cs >= 1025) {
return 2;
}
}
else if (MVDiff >= 13) {
if (diffRsCsdiff < 6) {
return 1;
}
else if (diffRsCsdiff >= 6) {
return 8;
}
}
}
}
}
}
else if (Rs >= 1748) {
if (Cs < 643) {
return 1;
}
else if (Cs >= 643) {
if (diffRsCsdiff < -9) {
if (tcor < 92) {
if (Rs < 1783) {
return 2;
}
else if (Rs >= 1783) {
return 8;
}
}
else if (tcor >= 92) {
if (Cs < 1201) {
return 4;
}
else if (Cs >= 1201) {
return 8;
}
}
}
else if (diffRsCsdiff >= -9) {
if (RsCsDiff < 23) {
return 8;
}
else if (RsCsDiff >= 23) {
if (negBalance < 34) {
return 1;
}
else if (negBalance >= 34) {
if (negBalance < 41) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (diffMVdiffVal < 2) {
if (negBalance < 35) {
return 8;
}
else if (negBalance >= 35) {
if (ssDCval < 80) {
return 8;
}
else if (ssDCval >= 80) {
return 2;
}
}
}
else if (diffMVdiffVal >= 2) {
return 8;
}
}
}
else if (negBalance >= 41) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
}
}
else if (refDCval >= 100) {
if (tcor < 84) {
if (posBalance < 53) {
if (Rs < 1340) {
return 1;
}
else if (Rs >= 1340) {
return 8;
}
}
else if (posBalance >= 53) {
if (Rs < 940) {
return 8;
}
else if (Rs >= 940) {
return 4;
}
}
}
else if (tcor >= 84) {
if (MVDiff < 28) {
if (RsCsDiff < 19) {
if (RsCsDiff < 9) {
if (negBalance < 38) {
if (posBalance < 43) {
if (Rs < 568) {
return 4;
}
else if (Rs >= 568) {
return 8;
}
}
else if (posBalance >= 43) {
return 8;
}
}
else if (negBalance >= 38) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 1;
}
}
}
}
else if (RsCsDiff >= 9) {
if (tcor < 90) {
if (SC < 1981) {
return 8;
}
else if (SC >= 1981) {
if (Rs < 1758) {
return 4;
}
else if (Rs >= 1758) {
return 1;
}
}
}
else if (tcor >= 90) {
if (negBalance < 28) {
if (Rs < 950) {
if (Cs < 1176) {
if (Rs < 644) {
return 4;
}
else if (Rs >= 644) {
return 8;
}
}
else if (Cs >= 1176) {
return 4;
}
}
else if (Rs >= 950) {
return 8;
}
}
else if (negBalance >= 28) {
if (MVDiff < 20) {
if (diffAFD < 4) {
return 8;
}
else if (diffAFD >= 4) {
return 4;
}
}
else if (MVDiff >= 20) {
if (ssDCval < 130) {
return 8;
}
else if (ssDCval >= 130) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 19) {
if (Cs < 764) {
if (Rs < 1656) {
return 8;
}
else if (Rs >= 1656) {
return 2;
}
}
else if (Cs >= 764) {
return 8;
}
}
}
else if (MVDiff >= 28) {
if (CsDiff < 4) {
if (ssDCval < 138) {
return 8;
}
else if (ssDCval >= 138) {
return 4;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
}
}
else if (ssDCval >= 155) {
if (TSC < 7) {
if (Rs < 1408) {
return 2;
}
else if (Rs >= 1408) {
return 8;
}
}
else if (TSC >= 7) {
if (Rs < 1047) {
return 4;
}
else if (Rs >= 1047) {
return 8;
}
}
}
}
}
else if (negBalance >= 44) {
if (Rs < 468) {
if (MVDiff < 4) {
if (Rs < 403) {
if (SC < 1102) {
return 1;
}
else if (SC >= 1102) {
return 4;
}
}
else if (Rs >= 403) {
if (Cs < 829) {
if (AFD < 8) {
return 1;
}
else if (AFD >= 8) {
if (Rs < 413) {
return 4;
}
else if (Rs >= 413) {
return 1;
}
}
}
else if (Cs >= 829) {
return 4;
}
}
}
else if (MVDiff >= 4) {
if (Rs < 332) {
return 1;
}
else if (Rs >= 332) {
if (mu_mv_mag_sq < 96) {
if (AFD < 6) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
return 1;
}
}
else if (AFD >= 6) {
return 8;
}
}
else if (mu_mv_mag_sq >= 96) {
return 1;
}
}
}
}
else if (Rs >= 468) {
if (Rs < 1696) {
if (ssDCval < 125) {
if (posBalance < 56) {
if (ssDCval < 75) {
if (ssDCval < 59) {
if (Rs < 1177) {
if (SC < 2099) {
if (RsCsDiff < 19) {
if (AFD < 10) {
if (tcor < 95) {
return 8;
}
else if (tcor >= 95) {
if (Cs < 618) {
return 8;
}
else if (Cs >= 618) {
return 4;
}
}
}
else if (AFD >= 10) {
return 4;
}
}
else if (RsCsDiff >= 19) {
if (negBalance < 46) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
else if (SC >= 2099) {
if (ssDCval < 50) {
return 8;
}
else if (ssDCval >= 50) {
return 4;
}
}
}
else if (Rs >= 1177) {
if (tcor < 74) {
if (Rs < 1322) {
return 1;
}
else if (Rs >= 1322) {
return 8;
}
}
else if (tcor >= 74) {
return 8;
}
}
}
else if (ssDCval >= 59) {
if (CsDiff < 4) {
if (RsDiff < 3) {
if (Cs < 895) {
return 8;
}
else if (Cs >= 895) {
return 1;
}
}
else if (RsDiff >= 3) {
if (Rs < 1477) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (gchDC < 3) {
if (diffMVdiffVal < 1) {
if (ssDCval < 73) {
if (tcor < 93) {
if (diffRsCsdiff < -4) {
return 8;
}
else if (diffRsCsdiff >= -4) {
if (Rs < 838) {
return 8;
}
else if (Rs >= 838) {
return 4;
}
}
}
else if (tcor >= 93) {
return 4;
}
}
else if (ssDCval >= 73) {
return 8;
}
}
else if (diffMVdiffVal >= 1) {
if (posBalance < 51) {
if (Rs < 1030) {
return 8;
}
else if (Rs >= 1030) {
return 4;
}
}
else if (posBalance >= 51) {
return 4;
}
}
}
else if (gchDC >= 3) {
return 8;
}
}
}
else if (Rs >= 1477) {
if (Cs < 777) {
return 4;
}
else if (Cs >= 777) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
return 2;
}
}
}
}
}
else if (CsDiff >= 4) {
if (MVDiff < 9) {
if (TSC < 8) {
if (Cs < 1223) {
return 8;
}
else if (Cs >= 1223) {
if (Rs < 1447) {
return 1;
}
else if (Rs >= 1447) {
return 8;
}
}
}
else if (TSC >= 8) {
if (Rs < 1451) {
return 1;
}
else if (Rs >= 1451) {
return 2;
}
}
}
else if (MVDiff >= 9) {
if (Rs < 1298) {
return 1;
}
else if (Rs >= 1298) {
return 4;
}
}
}
}
}
else if (ssDCval >= 75) {
if (MVDiff < 7) {
if (TSC < 9) {
if (TSC < 5) {
if (Cs < 702) {
return 8;
}
else if (Cs >= 702) {
return 1;
}
}
else if (TSC >= 5) {
if (MVDiff < 2) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
return 4;
}
}
else if (MVDiff >= 2) {
if (diffAFD < -3) {
return 1;
}
else if (diffAFD >= -3) {
return 8;
}
}
}
}
else if (TSC >= 9) {
return 1;
}
}
else if (MVDiff >= 7) {
if (tcor < 87) {
if (Rs < 1345) {
if (diffRsCsdiff < 7) {
return 8;
}
else if (diffRsCsdiff >= 7) {
return 2;
}
}
else if (Rs >= 1345) {
return 2;
}
}
else if (tcor >= 87) {
return 8;
}
}
}
}
else if (posBalance >= 56) {
if (SC < 1701) {
if (ssDCval < 70) {
if (AFD < 11) {
if (mcTcor < 81) {
if (Rs < 858) {
return 8;
}
else if (Rs >= 858) {
if (Rs < 916) {
return 1;
}
else if (Rs >= 916) {
return 8;
}
}
}
else if (mcTcor >= 81) {
if (AFD < 8) {
if (Cs < 674) {
return 8;
}
else if (Cs >= 674) {
if (SC < 1530) {
return 4;
}
else if (SC >= 1530) {
if (negBalance < 48) {
return 8;
}
else if (negBalance >= 48) {
return 1;
}
}
}
}
else if (AFD >= 8) {
if (mu_mv_mag_sq < 71) {
if (MVDiff < 6) {
if (Rs < 905) {
return 4;
}
else if (Rs >= 905) {
return 8;
}
}
else if (MVDiff >= 6) {
if (diffRsCsdiff < 9) {
return 8;
}
else if (diffRsCsdiff >= 9) {
if (Rs < 1030) {
return 4;
}
else if (Rs >= 1030) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 71) {
return 4;
}
}
}
}
else if (AFD >= 11) {
if (ssDCval < 45) {
return 2;
}
else if (ssDCval >= 45) {
return 4;
}
}
}
else if (ssDCval >= 70) {
if (MVDiff < 2) {
if (ssDCval < 90) {
return 8;
}
else if (ssDCval >= 90) {
return 1;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (SC >= 1701) {
if (Rs < 1676) {
if (RsDiff < 5) {
if (negBalance < 47) {
if (posBalance < 60) {
if (tcor < 76) {
if (Rs < 1196) {
return 8;
}
else if (Rs >= 1196) {
return 1;
}
}
else if (tcor >= 76) {
if (SC < 2616) {
if (mu_mv_mag_sq < 39) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
else if (mu_mv_mag_sq >= 39) {
return 8;
}
}
else if (SC >= 2616) {
if (ssDCval < 70) {
return 8;
}
else if (ssDCval >= 70) {
if (Rs < 1576) {
return 4;
}
else if (Rs >= 1576) {
return 8;
}
}
}
}
}
else if (posBalance >= 60) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (ssDCval < 80) {
return 4;
}
else if (ssDCval >= 80) {
return 8;
}
}
}
}
}
else if (negBalance >= 47) {
if (Rs < 751) {
return 4;
}
else if (Rs >= 751) {
return 8;
}
}
}
else if (RsDiff >= 5) {
if (SC < 1873) {
if (SC < 1800) {
return 1;
}
else if (SC >= 1800) {
if (SC < 1851) {
return 8;
}
else if (SC >= 1851) {
return 4;
}
}
}
else if (SC >= 1873) {
if (tcor < 92) {
if (diffAFD < -1) {
if (Rs < 1619) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
if (Cs < 940) {
return 2;
}
else if (Cs >= 940) {
return 8;
}
}
}
else if (Rs >= 1619) {
return 2;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (tcor >= 92) {
if (ssDCval < 75) {
return 8;
}
else if (ssDCval >= 75) {
if (AFD < 10) {
return 1;
}
else if (AFD >= 10) {
if (Rs < 1499) {
return 2;
}
else if (Rs >= 1499) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1676) {
if (Cs < 803) {
return 8;
}
else if (Cs >= 803) {
if (Cs < 965) {
return 1;
}
else if (Cs >= 965) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 125) {
if (MVDiff < 24) {
if (negBalance < 50) {
if (MVDiff < 18) {
if (refDCval < 131) {
return 4;
}
else if (refDCval >= 131) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 1;
}
}
}
else if (MVDiff >= 18) {
if (Rs < 1418) {
if (Cs < 773) {
return 8;
}
else if (Cs >= 773) {
return 4;
}
}
else if (Rs >= 1418) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (TSC < 8) {
if (Rs < 855) {
return 2;
}
else if (Rs >= 855) {
return 1;
}
}
else if (TSC >= 8) {
if (Rs < 1327) {
return 4;
}
else if (Rs >= 1327) {
return 8;
}
}
}
}
else if (MVDiff >= 24) {
if (RsCsDiff < 33) {
if (SC < 1869) {
if (TSC < 5) {
return 4;
}
else if (TSC >= 5) {
return 1;
}
}
else if (SC >= 1869) {
return 4;
}
}
else if (RsCsDiff >= 33) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Rs < 1109) {
return 1;
}
else if (Rs >= 1109) {
return 4;
}
}
}
}
}
}
else if (Rs >= 1696) {
if (refDCval < 70) {
if (refDCval < 61) {
return 8;
}
else if (refDCval >= 61) {
if (CsDiff < 5) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
if (diffMVdiffVal < -6) {
return 8;
}
else if (diffMVdiffVal >= -6) {
if (SC < 3129) {
return 2;
}
else if (SC >= 3129) {
if (Rs < 2224) {
return 4;
}
else if (Rs >= 2224) {
return 2;
}
}
}
}
}
else if (CsDiff >= 5) {
return 8;
}
}
}
else if (refDCval >= 70) {
if (RsDiff < 9) {
if (MVDiff < 2) {
if (posBalance < 55) {
if (TSC < 7) {
if (Rs < 1775) {
return 8;
}
else if (Rs >= 1775) {
return 1;
}
}
else if (TSC >= 7) {
return 8;
}
}
else if (posBalance >= 55) {
if (tcor < 92) {
if (Rs < 1959) {
return 2;
}
else if (Rs >= 1959) {
return 8;
}
}
else if (tcor >= 92) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
}
}
else if (MVDiff >= 2) {
if (diffRsCsdiff < -12) {
if (Cs < 915) {
return 2;
}
else if (Cs >= 915) {
if (Cs < 1277) {
if (posBalance < 60) {
if (MVDiff < 61) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
return 8;
}
}
else if (MVDiff >= 61) {
return 2;
}
}
else if (posBalance >= 60) {
return 1;
}
}
else if (Cs >= 1277) {
return 2;
}
}
}
else if (diffRsCsdiff >= -12) {
if (diffMVdiffVal < -10) {
if (Rs < 2015) {
return 2;
}
else if (Rs >= 2015) {
return 1;
}
}
else if (diffMVdiffVal >= -10) {
if (TSC < 11) {
if (diffRsCsdiff < 12) {
if (MVDiff < 20) {
if (diffAFD < -2) {
if (Rs < 1994) {
return 1;
}
else if (Rs >= 1994) {
return 8;
}
}
else if (diffAFD >= -2) {
if (diffMVdiffVal < -4) {
if (diffTSC < -1) {
return 2;
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (diffMVdiffVal >= -4) {
return 8;
}
}
}
else if (MVDiff >= 20) {
if (Cs < 1278) {
return 8;
}
else if (Cs >= 1278) {
return 1;
}
}
}
else if (diffRsCsdiff >= 12) {
if (negBalance < 53) {
if (Cs < 1256) {
return 8;
}
else if (Cs >= 1256) {
return 2;
}
}
else if (negBalance >= 53) {
return 1;
}
}
}
else if (TSC >= 11) {
return 1;
}
}
}
}
}
else if (RsDiff >= 9) {
return 4;
}
}
}
}
}
}
}
}
else if (AFD >= 13) {
if (refDCval < 112) {
if (Rs < 1724) {
if (MVDiff < 17) {
if (ssDCval < 91) {
if (Rs < 1383) {
if (ssDCval < 57) {
if (RsDiff < 6) {
if (TSC < 5) {
return 1;
}
else if (TSC >= 5) {
return 4;
}
}
else if (RsDiff >= 6) {
if (mu_mv_mag_sq < 3) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (TSC < 7) {
if (Rs < 1081) {
return 4;
}
else if (Rs >= 1081) {
return 2;
}
}
else if (TSC >= 7) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
}
}
}
else if (ssDCval >= 57) {
if (tcor < 66) {
if (diffRsCsdiff < -30) {
if (Rs < 1227) {
return 4;
}
else if (Rs >= 1227) {
return 1;
}
}
else if (diffRsCsdiff >= -30) {
if (gchDC < 3) {
if (Rs < 327) {
if (Rs < 244) {
return 4;
}
else if (Rs >= 244) {
return 2;
}
}
else if (Rs >= 327) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (ssDCval < 88) {
if (Rs < 776) {
return 4;
}
else if (Rs >= 776) {
return 8;
}
}
else if (ssDCval >= 88) {
return 4;
}
}
}
}
else if (gchDC >= 3) {
if (Rs < 1255) {
return 4;
}
else if (Rs >= 1255) {
return 8;
}
}
}
}
else if (tcor >= 66) {
if (ssDCval < 76) {
if (tcor < 84) {
if (mcTcor < 94) {
if (Cs < 635) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
else if (Cs >= 635) {
if (Cs < 1238) {
if (ssDCval < 70) {
if (posBalance < 58) {
if (negBalance < 53) {
return 1;
}
else if (negBalance >= 53) {
if (Rs < 1079) {
if (Cs < 771) {
return 8;
}
else if (Cs >= 771) {
return 2;
}
}
else if (Rs >= 1079) {
return 8;
}
}
}
else if (posBalance >= 58) {
return 8;
}
}
else if (ssDCval >= 70) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 2;
}
}
}
else if (Cs >= 1238) {
if (ssDCval < 72) {
return 1;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 220) {
return 2;
}
else if (Rs >= 220) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
return 4;
}
}
}
}
else if (tcor >= 84) {
return 8;
}
}
else if (ssDCval >= 76) {
if (TSC < 8) {
if (posBalance < 62) {
if (negBalance < 44) {
return 4;
}
else if (negBalance >= 44) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
if (SC < 1987) {
return 8;
}
else if (SC >= 1987) {
return 2;
}
}
}
}
else if (posBalance >= 62) {
if (Cs < 775) {
if (Rs < 1033) {
return 4;
}
else if (Rs >= 1033) {
return 1;
}
}
else if (Cs >= 775) {
return 2;
}
}
}
else if (TSC >= 8) {
if (posBalance < 53) {
return 2;
}
else if (posBalance >= 53) {
if (diffMVdiffVal < -2) {
if (Rs < 1139) {
return 4;
}
else if (Rs >= 1139) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
return 4;
}
}
}
}
}
}
}
else if (Rs >= 1383) {
if (posBalance < 55) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Cs < 1049) {
return 2;
}
else if (Cs >= 1049) {
return 8;
}
}
}
else if (posBalance >= 55) {
if (AFD < 22) {
if (Cs < 1276) {
if (ssDCval < 64) {
if (ssDCval < 59) {
if (TSC < 9) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (TSC >= 9) {
return 8;
}
}
else if (ssDCval >= 59) {
return 2;
}
}
else if (ssDCval >= 64) {
if (Cs < 637) {
return 1;
}
else if (Cs >= 637) {
return 8;
}
}
}
else if (Cs >= 1276) {
if (Rs < 1483) {
return 4;
}
else if (Rs >= 1483) {
if (Cs < 1278) {
return 4;
}
else if (Cs >= 1278) {
return 8;
}
}
}
}
else if (AFD >= 22) {
if (Rs < 1544) {
return 4;
}
else if (Rs >= 1544) {
return 1;
}
}
}
}
}
else if (ssDCval >= 91) {
if (mcTcor < 91) {
if (negBalance < 55) {
if (Rs < 1591) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 4;
}
}
else if (Rs >= 1591) {
if (Rs < 1634) {
return 4;
}
else if (Rs >= 1634) {
return 2;
}
}
}
else if (negBalance >= 55) {
if (SC < 2725) {
if (AFD < 14) {
return 8;
}
else if (AFD >= 14) {
return 4;
}
}
else if (SC >= 2725) {
return 8;
}
}
}
else if (mcTcor >= 91) {
if (negBalance < 59) {
if (AFD < 18) {
if (Cs < 735) {
if (Rs < 882) {
return 1;
}
else if (Rs >= 882) {
return 2;
}
}
else if (Cs >= 735) {
if (mu_mv_mag_sq < 7) {
if (SC < 2064) {
if (Rs < 750) {
return 4;
}
else if (Rs >= 750) {
return 8;
}
}
else if (SC >= 2064) {
if (MVDiff < 11) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 8;
}
}
else if (MVDiff >= 11) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 7) {
return 8;
}
}
}
else if (AFD >= 18) {
if (tcor < 73) {
return 4;
}
else if (tcor >= 73) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
if (Cs < 1102) {
return 1;
}
else if (Cs >= 1102) {
return 8;
}
}
}
}
}
else if (negBalance >= 59) {
if (diffRsCsdiff < -16) {
if (Rs < 705) {
return 1;
}
else if (Rs >= 705) {
return 8;
}
}
else if (diffRsCsdiff >= -16) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 17) {
if (SC < 2464) {
if (mcTcor < 80) {
if (TSC < 10) {
if (ssDCval < 45) {
if (ssDCval < 41) {
if (Rs < 312) {
return 2;
}
else if (Rs >= 312) {
return 8;
}
}
else if (ssDCval >= 41) {
if (AFD < 22) {
return 2;
}
else if (AFD >= 22) {
return 4;
}
}
}
else if (ssDCval >= 45) {
if (diffMVdiffVal < 29) {
if (gchDC < 5) {
if (mu_mv_mag_sq < 5) {
if (refDCval < 51) {
return 4;
}
else if (refDCval >= 51) {
if (Rs < 303) {
return 4;
}
else if (Rs >= 303) {
if (Cs < 602) {
return 2;
}
else if (Cs >= 602) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 5) {
if (RsCsDiff < 22) {
return 2;
}
else if (RsCsDiff >= 22) {
return 4;
}
}
}
else if (gchDC >= 5) {
if (MVDiff < 68) {
if (Cs < 695) {
return 1;
}
else if (Cs >= 695) {
return 4;
}
}
else if (MVDiff >= 68) {
return 8;
}
}
}
else if (diffMVdiffVal >= 29) {
if (SC < 1054) {
return 4;
}
else if (SC >= 1054) {
return 8;
}
}
}
}
else if (TSC >= 10) {
if (posBalance < 50) {
if (posBalance < 30) {
return 8;
}
else if (posBalance >= 30) {
if (diffRsCsdiff < -23) {
if (Rs < 1044) {
return 8;
}
else if (Rs >= 1044) {
return 4;
}
}
else if (diffRsCsdiff >= -23) {
if (MVDiff < 88) {
return 8;
}
else if (MVDiff >= 88) {
return 1;
}
}
}
}
else if (posBalance >= 50) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (Rs < 483) {
return 2;
}
else if (Rs >= 483) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 2) {
if (Rs < 605) {
if (mu_mv_mag_sq < 54) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
return 4;
}
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 268) {
return 8;
}
else if (Rs >= 268) {
return 2;
}
}
}
else if (Rs >= 605) {
if (AFD < 25) {
if (MVDiff < 33) {
if (Cs < 652) {
return 8;
}
else if (Cs >= 652) {
if (posBalance < 65) {
if (Rs < 809) {
return 4;
}
else if (Rs >= 809) {
return 2;
}
}
else if (posBalance >= 65) {
return 4;
}
}
}
else if (MVDiff >= 33) {
if (posBalance < 62) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 4;
}
}
else if (posBalance >= 62) {
return 8;
}
}
}
else if (AFD >= 25) {
if (RsDiff < 10) {
if (AFD < 28) {
if (Rs < 820) {
return 1;
}
else if (Rs >= 820) {
return 4;
}
}
else if (AFD >= 28) {
return 2;
}
}
else if (RsDiff >= 10) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 80) {
if (tcor < 92) {
if (RsDiff < 10) {
if (refDCval < 89) {
if (mu_mv_mag_sq < 3) {
if (mcTcor < 87) {
if (ssDCval < 69) {
if (diffMVdiffVal < 12) {
if (negBalance < 54) {
if (ssDCval < 49) {
return 2;
}
else if (ssDCval >= 49) {
return 4;
}
}
else if (negBalance >= 54) {
if (Rs < 956) {
if (Rs < 718) {
return 4;
}
else if (Rs >= 718) {
return 1;
}
}
else if (Rs >= 956) {
return 2;
}
}
}
else if (diffMVdiffVal >= 12) {
return 8;
}
}
else if (ssDCval >= 69) {
return 8;
}
}
else if (mcTcor >= 87) {
if (gchDC < 3) {
if (Cs < 1036) {
if (Rs < 907) {
if (RsCsDiff < 33) {
if (posBalance < 61) {
if (Rs < 624) {
return 2;
}
else if (Rs >= 624) {
return 4;
}
}
else if (posBalance >= 61) {
if (Cs < 506) {
return 1;
}
else if (Cs >= 506) {
return 8;
}
}
}
else if (RsCsDiff >= 33) {
if (Cs < 846) {
return 1;
}
else if (Cs >= 846) {
if (MVDiff < 21) {
return 1;
}
else if (MVDiff >= 21) {
if (Rs < 814) {
return 2;
}
else if (Rs >= 814) {
if (Rs < 882) {
return 1;
}
else if (Rs >= 882) {
return 8;
}
}
}
}
}
}
else if (Rs >= 907) {
if (tcor < 78) {
if (diffRsCsdiff < 17) {
return 4;
}
else if (diffRsCsdiff >= 17) {
if (Rs < 1152) {
return 8;
}
else if (Rs >= 1152) {
return 1;
}
}
}
else if (tcor >= 78) {
if (Cs < 1010) {
return 2;
}
else if (Cs >= 1010) {
return 1;
}
}
}
}
else if (Cs >= 1036) {
if (MVDiff < 80) {
if (diffAFD < 2) {
return 2;
}
else if (diffAFD >= 2) {
if (Rs < 963) {
return 4;
}
else if (Rs >= 963) {
return 1;
}
}
}
else if (MVDiff >= 80) {
if (Rs < 876) {
return 4;
}
else if (Rs >= 876) {
return 8;
}
}
}
}
else if (gchDC >= 3) {
if (diffMVdiffVal < -1) {
if (RsDiff < 4) {
return 4;
}
else if (RsDiff >= 4) {
if (diffRsCsdiff < 15) {
return 2;
}
else if (diffRsCsdiff >= 15) {
if (Cs < 952) {
return 8;
}
else if (Cs >= 952) {
return 2;
}
}
}
}
else if (diffMVdiffVal >= -1) {
if (RsCsDiff < 123) {
if (Rs < 627) {
if (ssDCval < 60) {
if (posBalance < 64) {
return 2;
}
else if (posBalance >= 64) {
return 1;
}
}
else if (ssDCval >= 60) {
if (Cs < 504) {
return 2;
}
else if (Cs >= 504) {
return 4;
}
}
}
else if (Rs >= 627) {
if (diffTSC < 2) {
if (Rs < 738) {
return 8;
}
else if (Rs >= 738) {
return 4;
}
}
else if (diffTSC >= 2) {
if (Rs < 799) {
return 4;
}
else if (Rs >= 799) {
return 8;
}
}
}
}
else if (RsCsDiff >= 123) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (Rs < 255) {
if (AFD < 25) {
return 4;
}
else if (AFD >= 25) {
return 2;
}
}
else if (Rs >= 255) {
if (negBalance < 46) {
if (RsDiff < 6) {
if (tcor < 75) {
if (mcTcor < 93) {
if (TSC < 8) {
if (posBalance < 55) {
return 1;
}
else if (posBalance >= 55) {
if (Cs < 664) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (Cs >= 664) {
return 4;
}
}
}
else if (TSC >= 8) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (Rs < 545) {
if (Cs < 878) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
return 2;
}
}
else if (Cs >= 878) {
return 4;
}
}
else if (Rs >= 545) {
return 2;
}
}
}
else if (tcor >= 75) {
if (posBalance < 60) {
return 4;
}
else if (posBalance >= 60) {
return 8;
}
}
}
else if (RsDiff >= 6) {
if (mu_mv_mag_sq < 29) {
if (MVDiff < 47) {
if (gchDC < 1) {
if (Rs < 1016) {
return 1;
}
else if (Rs >= 1016) {
return 8;
}
}
else if (gchDC >= 1) {
return 1;
}
}
else if (MVDiff >= 47) {
if (ssDCval < 64) {
if (Rs < 938) {
return 8;
}
else if (Rs >= 938) {
return 1;
}
}
else if (ssDCval >= 64) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (Cs < 955) {
if (ssDCval < 46) {
return 1;
}
else if (ssDCval >= 46) {
return 2;
}
}
else if (Cs >= 955) {
return 4;
}
}
}
}
else if (negBalance >= 46) {
if (AFD < 16) {
if (CsDiff < 7) {
if (diffAFD < -4) {
return 4;
}
else if (diffAFD >= -4) {
if (SC < 1118) {
if (Rs < 511) {
return 1;
}
else if (Rs >= 511) {
return 2;
}
}
else if (SC >= 1118) {
if (TSC < 8) {
if (refDCval < 65) {
if (mu_mv_mag_sq < 7) {
if (Rs < 864) {
return 1;
}
else if (Rs >= 864) {
return 4;
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffMVdiffVal < 9) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
if (negBalance < 48) {
return 4;
}
else if (negBalance >= 48) {
return 8;
}
}
}
else if (diffMVdiffVal >= 9) {
return 4;
}
}
}
else if (refDCval >= 65) {
if (Cs < 533) {
return 2;
}
else if (Cs >= 533) {
return 8;
}
}
}
else if (TSC >= 8) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Rs < 1375) {
return 4;
}
else if (Rs >= 1375) {
return 1;
}
}
}
}
}
}
}
else if (CsDiff >= 7) {
return 4;
}
}
else if (AFD >= 16) {
if (posBalance < 43) {
if (TSC < 9) {
return 2;
}
else if (TSC >= 9) {
return 8;
}
}
else if (posBalance >= 43) {
if (Rs < 782) {
if (tcor < 66) {
if (RsCsDiff < 29) {
return 2;
}
else if (RsCsDiff >= 29) {
if (SC < 1231) {
if (AFD < 29) {
if (diffRsCsdiff < -33) {
return 8;
}
else if (diffRsCsdiff >= -33) {
return 4;
}
}
else if (AFD >= 29) {
return 8;
}
}
else if (SC >= 1231) {
if (Cs < 898) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (Rs < 601) {
return 4;
}
else if (Rs >= 601) {
return 8;
}
}
}
}
else if (Cs >= 898) {
if (CsDiff < 11) {
return 4;
}
else if (CsDiff >= 11) {
if (Cs < 1206) {
return 8;
}
else if (Cs >= 1206) {
return 4;
}
}
}
}
}
}
else if (tcor >= 66) {
if (Cs < 635) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
if (diffRsCsdiff < 2) {
if (Cs < 593) {
return 8;
}
else if (Cs >= 593) {
return 1;
}
}
else if (diffRsCsdiff >= 2) {
return 1;
}
}
}
else if (Cs >= 635) {
if (negBalance < 53) {
if (Cs < 863) {
return 4;
}
else if (Cs >= 863) {
return 8;
}
}
else if (negBalance >= 53) {
return 2;
}
}
}
}
else if (Rs >= 782) {
if (posBalance < 46) {
return 2;
}
else if (posBalance >= 46) {
if (Cs < 1086) {
if (refDCval < 65) {
if (tcor < 70) {
if (Cs < 767) {
return 8;
}
else if (Cs >= 767) {
if (Rs < 928) {
return 1;
}
else if (Rs >= 928) {
if (Rs < 1160) {
return 4;
}
else if (Rs >= 1160) {
return 8;
}
}
}
}
else if (tcor >= 70) {
return 1;
}
}
else if (refDCval >= 65) {
if (SC < 1726) {
if (tcor < 72) {
if (diffMVdiffVal < -6) {
if (SC < 1641) {
return 2;
}
else if (SC >= 1641) {
return 8;
}
}
else if (diffMVdiffVal >= -6) {
if (CsDiff < 7) {
return 4;
}
else if (CsDiff >= 7) {
return 1;
}
}
}
else if (tcor >= 72) {
return 1;
}
}
else if (SC >= 1726) {
if (AFD < 22) {
if (SC < 2054) {
if (gchDC < 2) {
return 4;
}
else if (gchDC >= 2) {
if (MVDiff < 43) {
return 4;
}
else if (MVDiff >= 43) {
return 8;
}
}
}
else if (SC >= 2054) {
if (Cs < 1069) {
return 8;
}
else if (Cs >= 1069) {
return 4;
}
}
}
else if (AFD >= 22) {
if (diffMVdiffVal < -20) {
return 2;
}
else if (diffMVdiffVal >= -20) {
return 4;
}
}
}
}
}
else if (Cs >= 1086) {
if (mu_mv_mag_sq < 7) {
if (posBalance < 59) {
if (Rs < 1240) {
return 1;
}
else if (Rs >= 1240) {
return 8;
}
}
else if (posBalance >= 59) {
return 2;
}
}
else if (mu_mv_mag_sq >= 7) {
if (CsDiff < 11) {
if (Cs < 1241) {
return 1;
}
else if (Cs >= 1241) {
if (Rs < 859) {
return 1;
}
else if (Rs >= 859) {
return 8;
}
}
}
else if (CsDiff >= 11) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (refDCval >= 89) {
if (AFD < 15) {
if (mcTcor < 92) {
if (Rs < 484) {
return 8;
}
else if (Rs >= 484) {
return 1;
}
}
else if (mcTcor >= 92) {
if (diffMVdiffVal < 9) {
if (TSC < 7) {
return 1;
}
else if (TSC >= 7) {
return 4;
}
}
else if (diffMVdiffVal >= 9) {
if (posBalance < 56) {
return 8;
}
else if (posBalance >= 56) {
if (Rs < 521) {
return 1;
}
else if (Rs >= 521) {
return 2;
}
}
}
}
}
else if (AFD >= 15) {
if (MVDiff < 36) {
if (tcor < 79) {
if (tcor < 69) {
if (posBalance < 61) {
if (Rs < 695) {
return 2;
}
else if (Rs >= 695) {
return 1;
}
}
else if (posBalance >= 61) {
if (Rs < 732) {
return 8;
}
else if (Rs >= 732) {
return 4;
}
}
}
else if (tcor >= 69) {
if (RsDiff < 6) {
if (Rs < 754) {
return 8;
}
else if (Rs >= 754) {
return 4;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (tcor >= 79) {
if (TSC < 9) {
if (ssDCval < 113) {
return 2;
}
else if (ssDCval >= 113) {
return 8;
}
}
else if (TSC >= 9) {
return 1;
}
}
}
else if (MVDiff >= 36) {
if (RsCsDiff < 48) {
if (gchDC < 10) {
if (tcor < 77) {
if (Cs < 806) {
return 2;
}
else if (Cs >= 806) {
return 4;
}
}
else if (tcor >= 77) {
if (Cs < 591) {
return 8;
}
else if (Cs >= 591) {
return 2;
}
}
}
else if (gchDC >= 10) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 8;
}
}
}
else if (RsCsDiff >= 48) {
if (mcTcor < 93) {
if (RsDiff < 9) {
if (ssDCval < 129) {
if (diffRsCsdiff < 26) {
if (CsDiff < 6) {
if (Cs < 642) {
return 2;
}
else if (Cs >= 642) {
if (Cs < 1092) {
return 1;
}
else if (Cs >= 1092) {
return 2;
}
}
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (diffRsCsdiff >= 26) {
return 1;
}
}
else if (ssDCval >= 129) {
return 4;
}
}
else if (RsDiff >= 9) {
return 1;
}
}
else if (mcTcor >= 93) {
if (Cs < 969) {
return 1;
}
else if (Cs >= 969) {
return 2;
}
}
}
}
}
}
}
else if (RsDiff >= 10) {
if (mu_mv_mag_sq < 24) {
if (RsDiff < 11) {
if (AFD < 23) {
if (SC < 2353) {
if (Cs < 968) {
if (Cs < 533) {
return 1;
}
else if (Cs >= 533) {
return 8;
}
}
else if (Cs >= 968) {
return 1;
}
}
else if (SC >= 2353) {
if (Rs < 1591) {
return 4;
}
else if (Rs >= 1591) {
return 2;
}
}
}
else if (AFD >= 23) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
if (Rs < 1247) {
return 8;
}
else if (Rs >= 1247) {
return 1;
}
}
}
}
else if (RsDiff >= 11) {
if (refDCval < 97) {
if (Cs < 900) {
if (Rs < 1609) {
if (mcTcor < 89) {
return 4;
}
else if (mcTcor >= 89) {
return 2;
}
}
else if (Rs >= 1609) {
return 4;
}
}
else if (Cs >= 900) {
if (mu_mv_mag_sq < 14) {
if (Rs < 964) {
return 2;
}
else if (Rs >= 964) {
return 8;
}
}
else if (mu_mv_mag_sq >= 14) {
if (Rs < 1399) {
return 4;
}
else if (Rs >= 1399) {
return 2;
}
}
}
}
else if (refDCval >= 97) {
if (Cs < 677) {
return 1;
}
else if (Cs >= 677) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 24) {
if (negBalance < 51) {
if (mcTcor < 88) {
if (Rs < 1496) {
return 4;
}
else if (Rs >= 1496) {
return 8;
}
}
else if (mcTcor >= 88) {
return 2;
}
}
else if (negBalance >= 51) {
if (AFD < 31) {
if (Cs < 695) {
return 4;
}
else if (Cs >= 695) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
if (Rs < 1547) {
return 8;
}
else if (Rs >= 1547) {
return 4;
}
}
}
}
}
else if (AFD >= 31) {
if (Rs < 1348) {
return 4;
}
else if (Rs >= 1348) {
return 2;
}
}
}
}
}
}
else if (tcor >= 92) {
if (Cs < 631) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 1;
}
}
else if (Cs >= 631) {
return 8;
}
}
}
}
else if (SC >= 2464) {
if (AFD < 18) {
if (TSC < 10) {
if (Rs < 1327) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (Rs >= 1327) {
if (RsCsDiff < 69) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 1;
}
}
else if (RsCsDiff >= 69) {
if (negBalance < 35) {
if (Rs < 1420) {
return 8;
}
else if (Rs >= 1420) {
return 2;
}
}
else if (negBalance >= 35) {
if (tcor < 80) {
if (MVDiff < 22) {
return 2;
}
else if (MVDiff >= 22) {
if (SC < 2698) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 8;
}
}
else if (SC >= 2698) {
return 4;
}
}
}
else if (tcor >= 80) {
if (diffTSC < 1) {
return 8;
}
else if (diffTSC >= 1) {
if (Cs < 1267) {
return 4;
}
else if (Cs >= 1267) {
return 8;
}
}
}
}
}
}
}
else if (TSC >= 10) {
if (diffRsCsdiff < -14) {
if (MVDiff < 26) {
if (Rs < 1667) {
return 8;
}
else if (Rs >= 1667) {
return 1;
}
}
else if (MVDiff >= 26) {
if (AFD < 14) {
return 1;
}
else if (AFD >= 14) {
return 4;
}
}
}
else if (diffRsCsdiff >= -14) {
if (Rs < 1399) {
if (Rs < 1349) {
return 8;
}
else if (Rs >= 1349) {
return 2;
}
}
else if (Rs >= 1399) {
return 8;
}
}
}
}
else if (AFD >= 18) {
if (gchDC < 5) {
if (ssDCval < 80) {
if (CsDiff < 7) {
if (Rs < 1685) {
return 8;
}
else if (Rs >= 1685) {
return 2;
}
}
else if (CsDiff >= 7) {
if (SC < 2512) {
if (SC < 2470) {
return 4;
}
else if (SC >= 2470) {
return 2;
}
}
else if (SC >= 2512) {
if (SC < 2582) {
if (diffTSC < -1) {
return 8;
}
else if (diffTSC >= -1) {
return 4;
}
}
else if (SC >= 2582) {
if (Cs < 1042) {
if (Rs < 1566) {
return 2;
}
else if (Rs >= 1566) {
return 4;
}
}
else if (Cs >= 1042) {
return 4;
}
}
}
}
}
else if (ssDCval >= 80) {
if (ssDCval < 98) {
if (RsCsDiff < 132) {
if (mcTcor < 90) {
if (Rs < 1675) {
return 8;
}
else if (Rs >= 1675) {
return 4;
}
}
else if (mcTcor >= 90) {
return 1;
}
}
else if (RsCsDiff >= 132) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
if (Rs < 1556) {
return 4;
}
else if (Rs >= 1556) {
return 1;
}
}
}
}
else if (ssDCval >= 98) {
return 4;
}
}
}
else if (gchDC >= 5) {
if (CsDiff < 8) {
return 8;
}
else if (CsDiff >= 8) {
if (MVDiff < 58) {
return 2;
}
else if (MVDiff >= 58) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1724) {
if (refDCval < 72) {
if (Cs < 1027) {
if (AFD < 22) {
if (Rs < 1897) {
if (mu_mv_mag_sq < 19) {
if (SC < 2593) {
return 2;
}
else if (SC >= 2593) {
return 8;
}
}
else if (mu_mv_mag_sq >= 19) {
return 2;
}
}
else if (Rs >= 1897) {
return 2;
}
}
else if (AFD >= 22) {
if (Rs < 2158) {
if (Rs < 2011) {
return 1;
}
else if (Rs >= 2011) {
return 4;
}
}
else if (Rs >= 2158) {
return 2;
}
}
}
else if (Cs >= 1027) {
if (posBalance < 50) {
if (refDCval < 41) {
return 1;
}
else if (refDCval >= 41) {
return 2;
}
}
else if (posBalance >= 50) {
if (MVDiff < 32) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (RsCsDiff < 175) {
return 8;
}
else if (RsCsDiff >= 175) {
if (Rs < 1805) {
return 2;
}
else if (Rs >= 1805) {
return 8;
}
}
}
}
else if (MVDiff >= 32) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (tcor < 66) {
if (Rs < 1779) {
return 4;
}
else if (Rs >= 1779) {
return 8;
}
}
else if (tcor >= 66) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 72) {
if (Cs < 782) {
if (MVDiff < 46) {
if (refDCval < 85) {
if (tcor < 68) {
return 8;
}
else if (tcor >= 68) {
if (RsCsDiff < 90) {
if (AFD < 18) {
return 1;
}
else if (AFD >= 18) {
return 2;
}
}
else if (RsCsDiff >= 90) {
return 2;
}
}
}
else if (refDCval >= 85) {
if (ssDCval < 101) {
if (ssDCval < 90) {
if (AFD < 21) {
if (Cs < 558) {
return 2;
}
else if (Cs >= 558) {
if (Rs < 2296) {
return 4;
}
else if (Rs >= 2296) {
return 2;
}
}
}
else if (AFD >= 21) {
return 1;
}
}
else if (ssDCval >= 90) {
if (Rs < 1759) {
return 2;
}
else if (Rs >= 1759) {
return 1;
}
}
}
else if (ssDCval >= 101) {
if (Rs < 1753) {
return 4;
}
else if (Rs >= 1753) {
return 2;
}
}
}
}
else if (MVDiff >= 46) {
if (Cs < 608) {
if (negBalance < 69) {
if (Cs < 595) {
return 8;
}
else if (Cs >= 595) {
return 1;
}
}
else if (negBalance >= 69) {
return 1;
}
}
else if (Cs >= 608) {
return 8;
}
}
}
else if (Cs >= 782) {
if (SC < 3303) {
if (ssDCval < 84) {
if (mcTcor < 94) {
if (AFD < 24) {
if (diffAFD < 1) {
if (diffMVdiffVal < -45) {
return 2;
}
else if (diffMVdiffVal >= -45) {
if (MVDiff < 84) {
if (mu_mv_mag_sq < 17) {
return 8;
}
else if (mu_mv_mag_sq >= 17) {
if (diffRsCsdiff < 4) {
return 8;
}
else if (diffRsCsdiff >= 4) {
return 1;
}
}
}
else if (MVDiff >= 84) {
return 2;
}
}
}
else if (diffAFD >= 1) {
if (mu_mv_mag_sq < 39) {
if (RsCsDiff < 148) {
if (refDCval < 80) {
return 8;
}
else if (refDCval >= 80) {
if (Rs < 1983) {
return 4;
}
else if (Rs >= 1983) {
return 8;
}
}
}
else if (RsCsDiff >= 148) {
return 4;
}
}
else if (mu_mv_mag_sq >= 39) {
if (Rs < 2066) {
return 2;
}
else if (Rs >= 2066) {
return 1;
}
}
}
}
else if (AFD >= 24) {
if (Rs < 1964) {
return 2;
}
else if (Rs >= 1964) {
if (Rs < 2227) {
return 4;
}
else if (Rs >= 2227) {
return 1;
}
}
}
}
else if (mcTcor >= 94) {
if (diffMVdiffVal < 3) {
if (Rs < 2168) {
return 4;
}
else if (Rs >= 2168) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
return 2;
}
}
}
else if (ssDCval >= 84) {
if (CsDiff < 6) {
if (diffTSC < -1) {
if (ssDCval < 93) {
if (Rs < 2183) {
return 4;
}
else if (Rs >= 2183) {
return 8;
}
}
else if (ssDCval >= 93) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (mcTcor < 87) {
if (ssDCval < 111) {
if (negBalance < 59) {
return 8;
}
else if (negBalance >= 59) {
return 2;
}
}
else if (ssDCval >= 111) {
return 4;
}
}
else if (mcTcor >= 87) {
if (MVDiff < 48) {
if (Cs < 1006) {
return 1;
}
else if (Cs >= 1006) {
if (Cs < 1092) {
if (RsDiff < 8) {
return 1;
}
else if (RsDiff >= 8) {
return 2;
}
}
else if (Cs >= 1092) {
if (Rs < 2076) {
return 8;
}
else if (Rs >= 2076) {
return 4;
}
}
}
}
else if (MVDiff >= 48) {
return 8;
}
}
}
}
else if (CsDiff >= 6) {
if (mcTcor < 83) {
if (ssDCval < 101) {
if (Rs < 1960) {
return 1;
}
else if (Rs >= 1960) {
return 8;
}
}
else if (ssDCval >= 101) {
return 4;
}
}
else if (mcTcor >= 83) {
if (tcor < 85) {
if (mu_mv_mag_sq < 17) {
if (diffMVdiffVal < -4) {
if (Cs < 1160) {
return 1;
}
else if (Cs >= 1160) {
if (Rs < 1790) {
return 2;
}
else if (Rs >= 1790) {
return 8;
}
}
}
else if (diffMVdiffVal >= -4) {
if (refDCval < 103) {
return 2;
}
else if (refDCval >= 103) {
if (mu_mv_mag_sq < 11) {
return 2;
}
else if (mu_mv_mag_sq >= 11) {
return 1;
}
}
}
}
else if (mu_mv_mag_sq >= 17) {
if (negBalance < 44) {
return 8;
}
else if (negBalance >= 44) {
if (RsCsDiff < 233) {
if (Rs < 1933) {
return 2;
}
else if (Rs >= 1933) {
return 4;
}
}
else if (RsCsDiff >= 233) {
return 1;
}
}
}
}
else if (tcor >= 85) {
if (ssDCval < 94) {
if (Cs < 1097) {
return 4;
}
else if (Cs >= 1097) {
return 8;
}
}
else if (ssDCval >= 94) {
return 1;
}
}
}
}
}
}
else if (SC >= 3303) {
if (AFD < 18) {
if (TSC < 10) {
if (refDCval < 92) {
if (Rs < 2300) {
return 8;
}
else if (Rs >= 2300) {
return 4;
}
}
else if (refDCval >= 92) {
return 4;
}
}
else if (TSC >= 10) {
return 8;
}
}
else if (AFD >= 18) {
if (ssDCval < 80) {
if (Cs < 1252) {
return 4;
}
else if (Cs >= 1252) {
return 2;
}
}
else if (ssDCval >= 80) {
if (ssDCval < 108) {
if (SC < 3333) {
return 4;
}
else if (SC >= 3333) {
return 2;
}
}
else if (ssDCval >= 108) {
return 4;
}
}
}
}
}
}
}
}
else if (refDCval >= 112) {
if (MVDiff < 16) {
if (negBalance < 59) {
if (tcor < 65) {
return 4;
}
else if (tcor >= 65) {
if (diffRsCsdiff < 69) {
if (MVDiff < 5) {
if (Rs < 990) {
return 1;
}
else if (Rs >= 990) {
return 4;
}
}
else if (MVDiff >= 5) {
if (posBalance < 67) {
if (diffRsCsdiff < -30) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
return 1;
}
}
else if (diffRsCsdiff >= -30) {
return 8;
}
}
else if (posBalance >= 67) {
if (ssDCval < 124) {
return 2;
}
else if (ssDCval >= 124) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= 69) {
if (ssDCval < 126) {
if (Rs < 1940) {
return 8;
}
else if (Rs >= 1940) {
return 1;
}
}
else if (ssDCval >= 126) {
return 2;
}
}
}
}
else if (negBalance >= 59) {
if (Cs < 821) {
if (tcor < 81) {
return 2;
}
else if (tcor >= 81) {
if (Rs < 955) {
return 8;
}
else if (Rs >= 955) {
return 2;
}
}
}
else if (Cs >= 821) {
if (RsDiff < 8) {
if (Cs < 921) {
return 8;
}
else if (Cs >= 921) {
return 2;
}
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
else if (MVDiff >= 16) {
if (SC < 2114) {
if (Rs < 517) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
if (ssDCval < 165) {
return 2;
}
else if (ssDCval >= 165) {
return 1;
}
}
}
else if (Rs >= 517) {
if (tcor < 41) {
if (ssDCval < 145) {
if (SC < 1969) {
return 1;
}
else if (SC >= 1969) {
return 2;
}
}
else if (ssDCval >= 145) {
if (AFD < 33) {
return 8;
}
else if (AFD >= 33) {
return 1;
}
}
}
else if (tcor >= 41) {
if (mu_mv_mag_sq < 6) {
if (TSC < 9) {
if (Cs < 654) {
if (Rs < 1033) {
return 2;
}
else if (Rs >= 1033) {
return 1;
}
}
else if (Cs >= 654) {
return 8;
}
}
else if (TSC >= 9) {
if (AFD < 24) {
return 2;
}
else if (AFD >= 24) {
if (Cs < 728) {
return 1;
}
else if (Cs >= 728) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 6) {
if (RsCsDiff < 60) {
if (refDCval < 145) {
if (CsDiff < 1) {
return 1;
}
else if (CsDiff >= 1) {
if (AFD < 14) {
if (Rs < 933) {
return 8;
}
else if (Rs >= 933) {
return 4;
}
}
else if (AFD >= 14) {
return 8;
}
}
}
else if (refDCval >= 145) {
if (MVDiff < 109) {
return 1;
}
else if (MVDiff >= 109) {
if (Rs < 687) {
return 8;
}
else if (Rs >= 687) {
return 4;
}
}
}
}
else if (RsCsDiff >= 60) {
if (MVDiff < 51) {
if (gchDC < 4) {
if (SC < 1436) {
return 8;
}
else if (SC >= 1436) {
return 2;
}
}
else if (gchDC >= 4) {
if (Rs < 1025) {
return 8;
}
else if (Rs >= 1025) {
if (Rs < 1164) {
return 2;
}
else if (Rs >= 1164) {
return 8;
}
}
}
}
else if (MVDiff >= 51) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
if (gchDC < 5) {
return 2;
}
else if (gchDC >= 5) {
return 8;
}
}
}
}
}
}
}
}
else if (SC >= 2114) {
if (tcor < 84) {
if (mu_mv_mag_sq < 189) {
if (AFD < 34) {
if (tcor < 68) {
if (mu_mv_mag_sq < 16) {
if (diffRsCsdiff < 53) {
if (RsCsDiff < 104) {
return 8;
}
else if (RsCsDiff >= 104) {
if (Rs < 2300) {
if (diffMVdiffVal < 16) {
return 4;
}
else if (diffMVdiffVal >= 16) {
return 8;
}
}
else if (Rs >= 2300) {
return 1;
}
}
}
else if (diffRsCsdiff >= 53) {
if (CsDiff < 8) {
return 1;
}
else if (CsDiff >= 8) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 16) {
if (negBalance < 53) {
if (Rs < 1663) {
return 8;
}
else if (Rs >= 1663) {
return 1;
}
}
else if (negBalance >= 53) {
return 8;
}
}
}
else if (tcor >= 68) {
if (RsCsDiff < 117) {
if (Cs < 1117) {
if (Cs < 858) {
return 8;
}
else if (Cs >= 858) {
if (Rs < 1679) {
if (Rs < 1111) {
return 8;
}
else if (Rs >= 1111) {
return 1;
}
}
else if (Rs >= 1679) {
return 8;
}
}
}
else if (Cs >= 1117) {
if (Cs < 1190) {
return 4;
}
else if (Cs >= 1190) {
return 8;
}
}
}
else if (RsCsDiff >= 117) {
if (negBalance < 47) {
if (AFD < 27) {
return 2;
}
else if (AFD >= 27) {
return 1;
}
}
else if (negBalance >= 47) {
if (posBalance < 55) {
if (Cs < 926) {
return 1;
}
else if (Cs >= 926) {
if (RsCsDiff < 187) {
if (Rs < 2336) {
return 8;
}
else if (Rs >= 2336) {
return 2;
}
}
else if (RsCsDiff >= 187) {
if (Rs < 1409) {
return 4;
}
else if (Rs >= 1409) {
return 2;
}
}
}
}
else if (posBalance >= 55) {
if (tcor < 78) {
if (RsCsDiff < 153) {
return 8;
}
else if (RsCsDiff >= 153) {
return 1;
}
}
else if (tcor >= 78) {
return 1;
}
}
}
}
}
}
else if (AFD >= 34) {
if (Cs < 611) {
return 1;
}
else if (Cs >= 611) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 189) {
return 1;
}
}
else if (tcor >= 84) {
if (RsCsDiff < 94) {
if (negBalance < 57) {
if (RsCsDiff < 55) {
if (Rs < 1478) {
return 1;
}
else if (Rs >= 1478) {
return 8;
}
}
else if (RsCsDiff >= 55) {
if (gchDC < 2) {
if (Rs < 2173) {
return 8;
}
else if (Rs >= 2173) {
return 1;
}
}
else if (gchDC >= 2) {
return 2;
}
}
}
else if (negBalance >= 57) {
return 1;
}
}
else if (RsCsDiff >= 94) {
if (posBalance < 41) {
return 8;
}
else if (posBalance >= 41) {
if (diffMVdiffVal < 6) {
return 1;
}
else if (diffMVdiffVal >= 6) {
if (diffMVdiffVal < 9) {
if (Cs < 1139) {
return 8;
}
else if (Cs >= 1139) {
return 1;
}
}
else if (diffMVdiffVal >= 9) {
if (CsDiff < 7) {
return 1;
}
else if (CsDiff >= 7) {
return 8;
}
}
}
}
}
}
}
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 88) {
if (SC < 2156) {
if (Rs < 1065) {
if (refDCval < 63) {
if (Rs < 540) {
if (Cs < 902) {
if (ssDCval < 46) {
if (posBalance < 18) {
return 8;
}
else if (posBalance >= 18) {
return 1;
}
}
else if (ssDCval >= 46) {
if (refDCval < 58) {
if (SC < 1067) {
if (mu_mv_mag_sq < 66) {
if (negBalance < 28) {
return 4;
}
else if (negBalance >= 28) {
if (posBalance < 42) {
if (Rs < 239) {
return 1;
}
else if (Rs >= 239) {
return 4;
}
}
else if (posBalance >= 42) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (diffRsCsdiff < 4) {
return 1;
}
else if (diffRsCsdiff >= 4) {
return 4;
}
}
}
else if (SC >= 1067) {
if (diffMVdiffVal < -3) {
if (Rs < 460) {
return 4;
}
else if (Rs >= 460) {
return 1;
}
}
else if (diffMVdiffVal >= -3) {
return 4;
}
}
}
else if (refDCval >= 58) {
if (SC < 760) {
return 4;
}
else if (SC >= 760) {
if (AFD < 4) {
if (mu_mv_mag_sq < 123) {
if (Cs < 726) {
if (Cs < 653) {
if (Rs < 411) {
return 4;
}
else if (Rs >= 411) {
return 2;
}
}
else if (Cs >= 653) {
return 4;
}
}
else if (Cs >= 726) {
return 8;
}
}
else if (mu_mv_mag_sq >= 123) {
return 1;
}
}
else if (AFD >= 4) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
if (Cs < 869) {
return 1;
}
else if (Cs >= 869) {
return 4;
}
}
}
}
}
}
}
else if (Cs >= 902) {
if (SC < 1443) {
if (refDCval < 52) {
return 2;
}
else if (refDCval >= 52) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 1;
}
}
}
else if (SC >= 1443) {
return 4;
}
}
}
else if (Rs >= 540) {
if (MVDiff < 1) {
if (SC < 1646) {
if (refDCval < 57) {
if (mu_mv_mag_sq < 120) {
if (negBalance < 26) {
if (Cs < 586) {
return 4;
}
else if (Cs >= 586) {
return 1;
}
}
else if (negBalance >= 26) {
if (refDCval < 54) {
return 4;
}
else if (refDCval >= 54) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 120) {
if (Rs < 613) {
return 4;
}
else if (Rs >= 613) {
return 8;
}
}
}
else if (refDCval >= 57) {
return 8;
}
}
else if (SC >= 1646) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
if (Rs < 648) {
if (Cs < 1067) {
return 1;
}
else if (Cs >= 1067) {
return 4;
}
}
else if (Rs >= 648) {
if (TSCindex < 3) {
if (SC < 1671) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 1;
}
}
else if (SC >= 1671) {
return 1;
}
}
else if (TSCindex >= 3) {
if (mu_mv_mag_sq < 122) {
if (SC < 2030) {
if (TSC < 4) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
return 8;
}
}
else if (TSC >= 4) {
if (Rs < 905) {
return 8;
}
else if (Rs >= 905) {
return 1;
}
}
}
else if (SC >= 2030) {
return 8;
}
}
else if (mu_mv_mag_sq >= 122) {
if (Cs < 1105) {
if (Rs < 751) {
return 1;
}
else if (Rs >= 751) {
return 8;
}
}
else if (Cs >= 1105) {
return 1;
}
}
}
}
}
}
}
else if (MVDiff >= 1) {
if (negBalance < 22) {
if (ssDCval < 51) {
if (Rs < 802) {
if (Rs < 691) {
if (Rs < 610) {
return 4;
}
else if (Rs >= 610) {
return 1;
}
}
else if (Rs >= 691) {
return 2;
}
}
else if (Rs >= 802) {
if (Cs < 562) {
return 1;
}
else if (Cs >= 562) {
return 8;
}
}
}
else if (ssDCval >= 51) {
if (diffRsCsdiff < -4) {
if (Rs < 776) {
return 8;
}
else if (Rs >= 776) {
if (Rs < 828) {
return 1;
}
else if (Rs >= 828) {
return 4;
}
}
}
else if (diffRsCsdiff >= -4) {
if (SC < 1210) {
if (Rs < 607) {
return 4;
}
else if (Rs >= 607) {
return 1;
}
}
else if (SC >= 1210) {
return 1;
}
}
}
}
else if (negBalance >= 22) {
if (Rs < 891) {
if (negBalance < 28) {
if (TSC < 3) {
if (ssDCval < 42) {
return 8;
}
else if (ssDCval >= 42) {
if (SC < 1368) {
if (ssDCval < 50) {
return 1;
}
else if (ssDCval >= 50) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (SC >= 1368) {
return 1;
}
}
}
else if (TSC >= 3) {
if (Rs < 610) {
return 8;
}
else if (Rs >= 610) {
if (posBalance < 37) {
return 1;
}
else if (posBalance >= 37) {
return 4;
}
}
}
}
else if (negBalance >= 28) {
if (MVDiff < 13) {
if (ssDCval < 59) {
if (tcor < 80) {
if (Cs < 573) {
return 8;
}
else if (Cs >= 573) {
return 1;
}
}
else if (tcor >= 80) {
if (Cs < 881) {
if (Rs < 696) {
if (Rs < 596) {
return 4;
}
else if (Rs >= 596) {
if (MVDiff < 5) {
if (TSC < 3) {
return 8;
}
else if (TSC >= 3) {
return 4;
}
}
else if (MVDiff >= 5) {
return 8;
}
}
}
else if (Rs >= 696) {
if (Rs < 880) {
if (SC < 1364) {
if (Rs < 745) {
return 4;
}
else if (Rs >= 745) {
return 8;
}
}
else if (SC >= 1364) {
return 4;
}
}
else if (Rs >= 880) {
return 8;
}
}
}
else if (Cs >= 881) {
if (negBalance < 38) {
if (Cs < 883) {
return 2;
}
else if (Cs >= 883) {
return 4;
}
}
else if (negBalance >= 38) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (Rs < 795) {
return 2;
}
else if (Rs >= 795) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 59) {
if (AFD < 4) {
if (ssDCval < 61) {
if (Rs < 597) {
return 8;
}
else if (Rs >= 597) {
if (Rs < 642) {
return 1;
}
else if (Rs >= 642) {
return 4;
}
}
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (AFD >= 4) {
if (MVDiff < 6) {
if (Cs < 1093) {
return 8;
}
else if (Cs >= 1093) {
return 1;
}
}
else if (MVDiff >= 6) {
if (negBalance < 39) {
if (MVDiff < 11) {
return 4;
}
else if (MVDiff >= 11) {
return 1;
}
}
else if (negBalance >= 39) {
return 8;
}
}
}
}
}
else if (MVDiff >= 13) {
if (ssDCval < 55) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (Rs < 635) {
return 8;
}
else if (Rs >= 635) {
return 1;
}
}
}
else if (ssDCval >= 55) {
if (Cs < 765) {
return 1;
}
else if (Cs >= 765) {
if (Rs < 799) {
return 8;
}
else if (Rs >= 799) {
return 4;
}
}
}
}
}
}
else if (Rs >= 891) {
if (ssDCval < 54) {
if (posBalance < 44) {
if (posBalance < 36) {
if (Rs < 926) {
return 2;
}
else if (Rs >= 926) {
return 8;
}
}
else if (posBalance >= 36) {
return 1;
}
}
else if (posBalance >= 44) {
if (Cs < 1107) {
return 8;
}
else if (Cs >= 1107) {
return 4;
}
}
}
else if (ssDCval >= 54) {
if (negBalance < 41) {
if (negBalance < 28) {
if (RsDiff < 3) {
if (Cs < 934) {
return 1;
}
else if (Cs >= 934) {
return 4;
}
}
else if (RsDiff >= 3) {
return 4;
}
}
else if (negBalance >= 28) {
return 1;
}
}
else if (negBalance >= 41) {
if (Rs < 964) {
return 8;
}
else if (Rs >= 964) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 63) {
if (Rs < 359) {
if (RsCsDiff < 19) {
if (ssDCval < 69) {
if (MVDiff < 12) {
if (mcTcor < 99) {
if (Rs < 278) {
return 4;
}
else if (Rs >= 278) {
return 1;
}
}
else if (mcTcor >= 99) {
if (Rs < 170) {
return 1;
}
else if (Rs >= 170) {
return 4;
}
}
}
else if (MVDiff >= 12) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 8;
}
}
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (RsCsDiff >= 19) {
if (Rs < 348) {
return 2;
}
else if (Rs >= 348) {
return 1;
}
}
}
else if (Rs >= 359) {
if (AFD < 3) {
if (Cs < 694) {
if (ssDCval < 72) {
if (ssDCval < 64) {
if (Rs < 934) {
return 8;
}
else if (Rs >= 934) {
return 1;
}
}
else if (ssDCval >= 64) {
if (SC < 1367) {
if (posBalance < 42) {
if (Rs < 493) {
return 4;
}
else if (Rs >= 493) {
return 1;
}
}
else if (posBalance >= 42) {
if (Rs < 687) {
if (Rs < 632) {
return 8;
}
else if (Rs >= 632) {
return 1;
}
}
else if (Rs >= 687) {
return 2;
}
}
}
else if (SC >= 1367) {
if (negBalance < 29) {
if (MVDiff < 6) {
if (AFD < 1) {
if (Rs < 966) {
return 2;
}
else if (Rs >= 966) {
return 8;
}
}
else if (AFD >= 1) {
return 2;
}
}
else if (MVDiff >= 6) {
return 1;
}
}
else if (negBalance >= 29) {
if (Rs < 790) {
return 1;
}
else if (Rs >= 790) {
if (Rs < 840) {
return 2;
}
else if (Rs >= 840) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 72) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
return 2;
}
}
}
else if (Cs >= 694) {
if (ssDCval < 84) {
if (Rs < 718) {
if (SC < 1180) {
return 4;
}
else if (SC >= 1180) {
if (ssDCval < 65) {
if (Rs < 674) {
return 8;
}
else if (Rs >= 674) {
return 4;
}
}
else if (ssDCval >= 65) {
if (diffRsCsdiff < -3) {
return 1;
}
else if (diffRsCsdiff >= -3) {
return 8;
}
}
}
}
else if (Rs >= 718) {
if (SC < 1874) {
if (Cs < 745) {
if (Rs < 859) {
return 1;
}
else if (Rs >= 859) {
return 2;
}
}
else if (Cs >= 745) {
if (diffAFD < -2) {
if (Cs < 1001) {
return 1;
}
else if (Cs >= 1001) {
return 4;
}
}
else if (diffAFD >= -2) {
if (Rs < 912) {
return 4;
}
else if (Rs >= 912) {
return 1;
}
}
}
}
else if (SC >= 1874) {
if (RsCsDiff < 2) {
return 1;
}
else if (RsCsDiff >= 2) {
if (negBalance < 28) {
if (Rs < 1031) {
return 8;
}
else if (Rs >= 1031) {
return 4;
}
}
else if (negBalance >= 28) {
if (Cs < 1012) {
return 1;
}
else if (Cs >= 1012) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 84) {
if (Cs < 768) {
if (Rs < 611) {
return 2;
}
else if (Rs >= 611) {
return 1;
}
}
else if (Cs >= 768) {
if (Rs < 696) {
return 2;
}
else if (Rs >= 696) {
return 4;
}
}
}
}
}
else if (AFD >= 3) {
if (MVDiff < 2) {
if (ssDCval < 81) {
if (Cs < 809) {
if (Rs < 604) {
if (ssDCval < 71) {
if (SC < 1045) {
if (Rs < 430) {
return 1;
}
else if (Rs >= 430) {
return 2;
}
}
else if (SC >= 1045) {
return 4;
}
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (Rs >= 604) {
if (negBalance < 39) {
if (refDCval < 73) {
if (Cs < 571) {
if (Rs < 706) {
return 1;
}
else if (Rs >= 706) {
return 2;
}
}
else if (Cs >= 571) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
return 1;
}
}
}
else if (refDCval >= 73) {
if (Rs < 783) {
if (Cs < 660) {
if (Rs < 698) {
return 1;
}
else if (Rs >= 698) {
return 4;
}
}
else if (Cs >= 660) {
return 1;
}
}
else if (Rs >= 783) {
if (Rs < 873) {
return 4;
}
else if (Rs >= 873) {
return 8;
}
}
}
}
else if (negBalance >= 39) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 8;
}
}
}
}
else if (Cs >= 809) {
if (mu_mv_mag_sq < 111) {
if (MVDiff < 1) {
if (SC < 1540) {
return 1;
}
else if (SC >= 1540) {
if (ssDCval < 65) {
if (Rs < 910) {
return 1;
}
else if (Rs >= 910) {
return 4;
}
}
else if (ssDCval >= 65) {
if (SC < 2141) {
return 8;
}
else if (SC >= 2141) {
if (Rs < 995) {
return 2;
}
else if (Rs >= 995) {
return 8;
}
}
}
}
}
else if (MVDiff >= 1) {
if (Rs < 809) {
if (refDCval < 71) {
return 8;
}
else if (refDCval >= 71) {
return 1;
}
}
else if (Rs >= 809) {
if (ssDCval < 73) {
return 1;
}
else if (ssDCval >= 73) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (RsCsDiff < 3) {
if (refDCval < 80) {
return 4;
}
else if (refDCval >= 80) {
return 8;
}
}
else if (RsCsDiff >= 3) {
if (ssDCval < 71) {
if (Cs < 1196) {
return 4;
}
else if (Cs >= 1196) {
return 2;
}
}
else if (ssDCval >= 71) {
if (SC < 2067) {
return 8;
}
else if (SC >= 2067) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 81) {
if (Cs < 728) {
return 2;
}
else if (Cs >= 728) {
return 4;
}
}
}
else if (MVDiff >= 2) {
if (Cs < 960) {
if (Rs < 899) {
if (ssDCval < 86) {
if (mu_mv_mag_sq < 63) {
if (negBalance < 41) {
if (tcor < 95) {
if (TSC < 5) {
if (Rs < 840) {
if (SC < 980) {
return 4;
}
else if (SC >= 980) {
return 2;
}
}
else if (Rs >= 840) {
return 1;
}
}
else if (TSC >= 5) {
return 1;
}
}
else if (tcor >= 95) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
if (Rs < 606) {
return 1;
}
else if (Rs >= 606) {
return 2;
}
}
}
}
else if (negBalance >= 41) {
if (Cs < 561) {
if (ssDCval < 78) {
if (Cs < 495) {
return 4;
}
else if (Cs >= 495) {
return 1;
}
}
else if (ssDCval >= 78) {
return 8;
}
}
else if (Cs >= 561) {
if (SC < 1415) {
if (Cs < 613) {
return 4;
}
else if (Cs >= 613) {
return 8;
}
}
else if (SC >= 1415) {
if (posBalance < 50) {
if (Rs < 831) {
return 2;
}
else if (Rs >= 831) {
return 8;
}
}
else if (posBalance >= 50) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 63) {
if (refDCval < 80) {
if (refDCval < 69) {
if (Rs < 742) {
if (Rs < 488) {
return 4;
}
else if (Rs >= 488) {
return 8;
}
}
else if (Rs >= 742) {
if (Rs < 841) {
return 4;
}
else if (Rs >= 841) {
return 1;
}
}
}
else if (refDCval >= 69) {
if (mu_mv_mag_sq < 89) {
if (Cs < 837) {
return 8;
}
else if (Cs >= 837) {
return 2;
}
}
else if (mu_mv_mag_sq >= 89) {
if (ssDCval < 69) {
return 4;
}
else if (ssDCval >= 69) {
return 2;
}
}
}
}
else if (refDCval >= 80) {
if (AFD < 7) {
if (Rs < 459) {
return 8;
}
else if (Rs >= 459) {
if (diffAFD < 1) {
return 1;
}
else if (diffAFD >= 1) {
if (Rs < 601) {
return 1;
}
else if (Rs >= 601) {
return 2;
}
}
}
}
else if (AFD >= 7) {
if (mu_mv_mag_sq < 73) {
return 8;
}
else if (mu_mv_mag_sq >= 73) {
return 4;
}
}
}
}
}
else if (ssDCval >= 86) {
if (SC < 1478) {
return 2;
}
else if (SC >= 1478) {
return 4;
}
}
}
else if (Rs >= 899) {
if (Cs < 907) {
if (AFD < 10) {
if (ssDCval < 65) {
return 1;
}
else if (ssDCval >= 65) {
if (MVDiff < 3) {
if (Rs < 982) {
return 1;
}
else if (Rs >= 982) {
return 8;
}
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (AFD >= 10) {
if (diffRsCsdiff < 12) {
return 1;
}
else if (diffRsCsdiff >= 12) {
if (Rs < 1030) {
return 8;
}
else if (Rs >= 1030) {
return 4;
}
}
}
}
else if (Cs >= 907) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
}
}
else if (Cs >= 960) {
if (RsDiff < 2) {
if (mu_mv_mag_sq < 88) {
return 1;
}
else if (mu_mv_mag_sq >= 88) {
return 8;
}
}
else if (RsDiff >= 2) {
if (Rs < 833) {
if (SC < 1758) {
if (SC < 1725) {
if (negBalance < 33) {
return 1;
}
else if (negBalance >= 33) {
return 2;
}
}
else if (SC >= 1725) {
return 1;
}
}
else if (SC >= 1758) {
return 2;
}
}
else if (Rs >= 833) {
if (TSC < 5) {
if (Cs < 1088) {
return 4;
}
else if (Cs >= 1088) {
return 2;
}
}
else if (TSC >= 5) {
if (Cs < 1084) {
if (Rs < 887) {
return 8;
}
else if (Rs >= 887) {
return 2;
}
}
else if (Cs >= 1084) {
if (Rs < 848) {
return 1;
}
else if (Rs >= 848) {
return 4;
}
}
}
}
}
}
}
}
}
}
}
else if (Rs >= 1065) {
if (mu_mv_mag_sq < 124) {
if (RsCsDiff < 18) {
if (posBalance < 45) {
if (refDCval < 49) {
if (Rs < 1223) {
return 8;
}
else if (Rs >= 1223) {
return 4;
}
}
else if (refDCval >= 49) {
if (posBalance < 32) {
if (mu_mv_mag_sq < 123) {
if (diffMVdiffVal < 1) {
if (refDCval < 51) {
return 4;
}
else if (refDCval >= 51) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
if (diffMVdiffVal < 3) {
if (Rs < 1120) {
return 1;
}
else if (Rs >= 1120) {
return 8;
}
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 123) {
return 2;
}
}
else if (posBalance >= 32) {
if (diffRsCsdiff < 5) {
if (posBalance < 39) {
return 1;
}
else if (posBalance >= 39) {
if (negBalance < 25) {
return 1;
}
else if (negBalance >= 25) {
if (Rs < 1158) {
return 1;
}
else if (Rs >= 1158) {
if (AFD < 5) {
if (gchDC < 1) {
return 8;
}
else if (gchDC >= 1) {
return 1;
}
}
else if (AFD >= 5) {
return 1;
}
}
}
}
}
else if (diffRsCsdiff >= 5) {
if (Rs < 1288) {
return 1;
}
else if (Rs >= 1288) {
if (Rs < 1411) {
return 8;
}
else if (Rs >= 1411) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 45) {
if (negBalance < 27) {
if (SC < 1812) {
if (Rs < 1153) {
return 2;
}
else if (Rs >= 1153) {
return 1;
}
}
else if (SC >= 1812) {
if (SC < 1967) {
if (Rs < 1078) {
return 8;
}
else if (Rs >= 1078) {
return 1;
}
}
else if (SC >= 1967) {
return 8;
}
}
}
else if (negBalance >= 27) {
if (MVDiff < 13) {
if (mu_mv_mag_sq < 118) {
if (posBalance < 55) {
if (negBalance < 28) {
if (RsDiff < 2) {
return 1;
}
else if (RsDiff >= 2) {
return 2;
}
}
else if (negBalance >= 28) {
return 1;
}
}
else if (posBalance >= 55) {
if (mu_mv_mag_sq < 76) {
return 8;
}
else if (mu_mv_mag_sq >= 76) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 118) {
if (posBalance < 47) {
return 8;
}
else if (posBalance >= 47) {
if (Cs < 611) {
return 8;
}
else if (Cs >= 611) {
return 1;
}
}
}
}
else if (MVDiff >= 13) {
if (TSC < 4) {
return 8;
}
else if (TSC >= 4) {
if (Rs < 1236) {
return 2;
}
else if (Rs >= 1236) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 18) {
if (Cs < 815) {
if (MVDiff < 12) {
if (ssDCval < 60) {
if (Rs < 1240) {
return 1;
}
else if (Rs >= 1240) {
return 8;
}
}
else if (ssDCval >= 60) {
if (Rs < 1094) {
return 8;
}
else if (Rs >= 1094) {
return 1;
}
}
}
else if (MVDiff >= 12) {
if (diffTSC < 1) {
if (Cs < 630) {
return 1;
}
else if (Cs >= 630) {
if (Rs < 1095) {
return 4;
}
else if (Rs >= 1095) {
return 2;
}
}
}
else if (diffTSC >= 1) {
return 2;
}
}
}
else if (Cs >= 815) {
if (Cs < 939) {
return 8;
}
else if (Cs >= 939) {
if (Rs < 1182) {
if (Rs < 1130) {
return 2;
}
else if (Rs >= 1130) {
return 8;
}
}
else if (Rs >= 1182) {
return 4;
}
}
}
}
}
else if (mu_mv_mag_sq >= 124) {
if (Rs < 1390) {
if (Cs < 748) {
if (mu_mv_mag_sq < 128) {
if (AFD < 3) {
if (Rs < 1092) {
if (Rs < 1082) {
if (Rs < 1077) {
return 2;
}
else if (Rs >= 1077) {
return 8;
}
}
else if (Rs >= 1082) {
return 4;
}
}
else if (Rs >= 1092) {
if (posBalance < 48) {
if (diffMVdiffVal < -1) {
return 2;
}
else if (diffMVdiffVal >= -1) {
if (negBalance < 17) {
return 1;
}
else if (negBalance >= 17) {
if (Cs < 550) {
if (Rs < 1337) {
return 8;
}
else if (Rs >= 1337) {
return 2;
}
}
else if (Cs >= 550) {
if (MVDiff < 1) {
return 1;
}
else if (MVDiff >= 1) {
return 2;
}
}
}
}
}
else if (posBalance >= 48) {
return 8;
}
}
}
else if (AFD >= 3) {
if (Cs < 624) {
return 8;
}
else if (Cs >= 624) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 615) {
if (AFD < 1) {
return 1;
}
else if (AFD >= 1) {
return 8;
}
}
else if (Cs >= 615) {
if (Cs < 713) {
return 1;
}
else if (Cs >= 713) {
return 4;
}
}
}
}
else if (Cs >= 748) {
if (Cs < 895) {
return 8;
}
else if (Cs >= 895) {
return 1;
}
}
}
else if (Rs >= 1390) {
if (negBalance < 19) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 2;
}
}
else if (negBalance >= 19) {
if (Cs < 667) {
return 8;
}
else if (Cs >= 667) {
return 1;
}
}
}
}
}
}
else if (SC >= 2156) {
if (TSC < 3) {
if (ssDCval < 81) {
if (MVDiff < 8) {
if (ssDCval < 52) {
if (CsDiff < 1) {
return 8;
}
else if (CsDiff >= 1) {
if (SC < 2981) {
if (posBalance < 47) {
if (ssDCval < 47) {
if (Rs < 1536) {
return 8;
}
else if (Rs >= 1536) {
if (posBalance < 30) {
return 4;
}
else if (posBalance >= 30) {
return 4;
}
}
}
else if (ssDCval >= 47) {
if (TSC < 2) {
if (Rs < 1699) {
return 8;
}
else if (Rs >= 1699) {
if (Rs < 1745) {
if (Cs < 1120) {
return 4;
}
else if (Cs >= 1120) {
return 8;
}
}
else if (Rs >= 1745) {
if (diffAFD < -1) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
return 8;
}
}
else if (diffAFD >= -1) {
if (Rs < 1778) {
if (Rs < 1762) {
if (Rs < 1761) {
return 4;
}
else if (Rs >= 1761) {
return 8;
}
}
else if (Rs >= 1762) {
if (Rs < 1769) {
return 4;
}
else if (Rs >= 1769) {
return 4;
}
}
}
else if (Rs >= 1778) {
if (negBalance < 18) {
if (Cs < 1148) {
if (Rs < 1813) {
return 4;
}
else if (Rs >= 1813) {
if (Rs < 1823) {
return 8;
}
else if (Rs >= 1823) {
return 4;
}
}
}
else if (Cs >= 1148) {
return 4;
}
}
else if (negBalance >= 18) {
if (Rs < 1781) {
return 8;
}
else if (Rs >= 1781) {
if (Cs < 1141) {
if (SC < 2856) {
return 8;
}
else if (SC >= 2856) {
if (Rs < 1821) {
if (Rs < 1793) {
return 8;
}
else if (Rs >= 1793) {
return 4;
}
}
else if (Rs >= 1821) {
return 8;
}
}
}
else if (Cs >= 1141) {
return 8;
}
}
}
}
}
}
}
}
else if (TSC >= 2) {
if (diffRsCsdiff < -5) {
return 4;
}
else if (diffRsCsdiff >= -5) {
if (refDCval < 50) {
if (Cs < 1155) {
if (refDCval < 48) {
if (Cs < 1056) {
return 8;
}
else if (Cs >= 1056) {
if (negBalance < 22) {
if (SC < 2852) {
if (tcor < 98) {
return 4;
}
else if (tcor >= 98) {
if (Rs < 1714) {
if (posBalance < 38) {
if (Rs < 1680) {
if (Rs < 1666) {
return 4;
}
else if (Rs >= 1666) {
if (Rs < 1677) {
return 8;
}
else if (Rs >= 1677) {
return 8;
}
}
}
else if (Rs >= 1680) {
return 4;
}
}
else if (posBalance >= 38) {
return 4;
}
}
else if (Rs >= 1714) {
if (Cs < 1079) {
return 8;
}
else if (Cs >= 1079) {
return 8;
}
}
}
}
else if (SC >= 2852) {
if (Rs < 1758) {
return 4;
}
else if (Rs >= 1758) {
if (Rs < 1783) {
if (negBalance < 21) {
if (Rs < 1766) {
return 4;
}
else if (Rs >= 1766) {
return 8;
}
}
else if (negBalance >= 21) {
return 4;
}
}
else if (Rs >= 1783) {
return 8;
}
}
}
}
else if (negBalance >= 22) {
if (Rs < 1746) {
return 4;
}
else if (Rs >= 1746) {
if (Rs < 1786) {
if (Rs < 1763) {
return 4;
}
else if (Rs >= 1763) {
return 8;
}
}
else if (Rs >= 1786) {
return 4;
}
}
}
}
}
else if (refDCval >= 48) {
return 4;
}
}
else if (Cs >= 1155) {
return 4;
}
}
else if (refDCval >= 50) {
if (Cs < 1253) {
if (Cs < 616) {
return 4;
}
else if (Cs >= 616) {
return 8;
}
}
else if (Cs >= 1253) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 47) {
return 8;
}
}
else if (SC >= 2981) {
if (SC < 3393) {
return 4;
}
else if (SC >= 3393) {
return 8;
}
}
}
}
else if (ssDCval >= 52) {
if (ssDCval < 69) {
if (SC < 2817) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 128) {
if (MVDiff < 3) {
if (diffTSC < -1) {
if (Rs < 1252) {
return 4;
}
else if (Rs >= 1252) {
if (Cs < 852) {
return 1;
}
else if (Cs >= 852) {
return 8;
}
}
}
else if (diffTSC >= -1) {
if (posBalance < 36) {
if (Rs < 1461) {
return 8;
}
else if (Rs >= 1461) {
return 1;
}
}
else if (posBalance >= 36) {
return 1;
}
}
}
else if (MVDiff >= 3) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 128) {
if (Cs < 1064) {
return 8;
}
else if (Cs >= 1064) {
return 4;
}
}
}
else if (mcTcor >= 99) {
if (RsDiff < 2) {
if (ssDCval < 56) {
if (SC < 2431) {
return 8;
}
else if (SC >= 2431) {
return 4;
}
}
else if (ssDCval >= 56) {
if (AFD < 3) {
if (mu_mv_mag_sq < 128) {
return 1;
}
else if (mu_mv_mag_sq >= 128) {
if (SC < 2220) {
if (Rs < 1299) {
return 1;
}
else if (Rs >= 1299) {
return 8;
}
}
else if (SC >= 2220) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 4;
}
}
}
}
else if (AFD >= 3) {
if (Cs < 794) {
return 1;
}
else if (Cs >= 794) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
return 4;
}
}
}
else if (SC >= 2817) {
if (posBalance < 37) {
if (mcTcor < 99) {
if (negBalance < 24) {
if (Rs < 1828) {
return 2;
}
else if (Rs >= 1828) {
return 1;
}
}
else if (negBalance >= 24) {
return 4;
}
}
else if (mcTcor >= 99) {
if (Rs < 1904) {
if (Rs < 1885) {
if (Cs < 1050) {
return 4;
}
else if (Cs >= 1050) {
if (Rs < 1840) {
return 4;
}
else if (Rs >= 1840) {
return 1;
}
}
}
else if (Rs >= 1885) {
return 1;
}
}
else if (Rs >= 1904) {
return 4;
}
}
}
else if (posBalance >= 37) {
if (RsDiff < 2) {
if (SC < 2884) {
return 4;
}
else if (SC >= 2884) {
return 1;
}
}
else if (RsDiff >= 2) {
if (Rs < 2126) {
return 4;
}
else if (Rs >= 2126) {
return 8;
}
}
}
}
}
else if (ssDCval >= 69) {
if (posBalance < 41) {
if (ssDCval < 72) {
if (Rs < 1617) {
return 8;
}
else if (Rs >= 1617) {
return 4;
}
}
else if (ssDCval >= 72) {
if (Rs < 1255) {
return 8;
}
else if (Rs >= 1255) {
return 1;
}
}
}
else if (posBalance >= 41) {
if (Cs < 1108) {
if (mu_mv_mag_sq < 98) {
return 1;
}
else if (mu_mv_mag_sq >= 98) {
if (Rs < 1174) {
return 4;
}
else if (Rs >= 1174) {
return 8;
}
}
}
else if (Cs >= 1108) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (refDCval < 70) {
if (Rs < 1276) {
return 4;
}
else if (Rs >= 1276) {
return 1;
}
}
else if (refDCval >= 70) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 8) {
if (mu_mv_mag_sq < 95) {
if (posBalance < 19) {
if (Cs < 970) {
return 4;
}
else if (Cs >= 970) {
return 8;
}
}
else if (posBalance >= 19) {
return 4;
}
}
else if (mu_mv_mag_sq >= 95) {
return 1;
}
}
}
else if (ssDCval >= 81) {
if (MVDiff < 1) {
if (SC < 2982) {
if (diffRsCsdiff < -5) {
return 8;
}
else if (diffRsCsdiff >= -5) {
if (Rs < 2195) {
return 4;
}
else if (Rs >= 2195) {
return 8;
}
}
}
else if (SC >= 2982) {
return 8;
}
}
else if (MVDiff >= 1) {
return 8;
}
}
}
else if (TSC >= 3) {
if (Cs < 864) {
if (RsCsDiff < 9) {
if (ssDCval < 83) {
if (AFD < 4) {
if (diffRsCsdiff < 4) {
if (tcor < 100) {
return 8;
}
else if (tcor >= 100) {
return 4;
}
}
else if (diffRsCsdiff >= 4) {
return 1;
}
}
else if (AFD >= 4) {
if (SC < 2162) {
return 8;
}
else if (SC >= 2162) {
return 1;
}
}
}
else if (ssDCval >= 83) {
if (SC < 2598) {
return 4;
}
else if (SC >= 2598) {
if (Rs < 2083) {
return 1;
}
else if (Rs >= 2083) {
return 8;
}
}
}
}
else if (RsCsDiff >= 9) {
if (TSC < 5) {
if (ssDCval < 54) {
return 4;
}
else if (ssDCval >= 54) {
if (ssDCval < 71) {
if (ssDCval < 63) {
if (negBalance < 41) {
return 1;
}
else if (negBalance >= 41) {
if (Rs < 2055) {
return 8;
}
else if (Rs >= 2055) {
return 2;
}
}
}
else if (ssDCval >= 63) {
if (refDCval < 70) {
return 2;
}
else if (refDCval >= 70) {
if (RsDiff < 5) {
return 1;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
else if (ssDCval >= 71) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
if (Rs < 1718) {
return 1;
}
else if (Rs >= 1718) {
return 4;
}
}
}
}
}
else if (TSC >= 5) {
if (SC < 2677) {
if (diffRsCsdiff < 11) {
if (ssDCval < 69) {
if (Rs < 1644) {
if (Rs < 1553) {
return 8;
}
else if (Rs >= 1553) {
return 1;
}
}
else if (Rs >= 1644) {
return 2;
}
}
else if (ssDCval >= 69) {
return 1;
}
}
else if (diffRsCsdiff >= 11) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (SC >= 2677) {
if (diffMVdiffVal < 3) {
if (RsDiff < 5) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
return 1;
}
}
else if (RsDiff >= 5) {
if (diffRsCsdiff < 61) {
return 2;
}
else if (diffRsCsdiff >= 61) {
return 4;
}
}
}
else if (diffMVdiffVal >= 3) {
return 1;
}
}
}
}
}
else if (Cs >= 864) {
if (mu_mv_mag_sq < 85) {
if (posBalance < 45) {
if (Cs < 982) {
if (Rs < 1818) {
return 1;
}
else if (Rs >= 1818) {
if (Cs < 947) {
return 4;
}
else if (Cs >= 947) {
return 1;
}
}
}
else if (Cs >= 982) {
if (diffRsCsdiff < -49) {
if (Cs < 1097) {
return 1;
}
else if (Cs >= 1097) {
return 8;
}
}
else if (diffRsCsdiff >= -49) {
if (tcor < 90) {
return 2;
}
else if (tcor >= 90) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 1038) {
if (diffAFD < 1) {
if (ssDCval < 60) {
return 4;
}
else if (ssDCval >= 60) {
return 2;
}
}
else if (diffAFD >= 1) {
if (Rs < 1025) {
return 8;
}
else if (Rs >= 1025) {
return 4;
}
}
}
else if (Rs >= 1038) {
if (Rs < 1277) {
if (MVDiff < 3) {
if (Rs < 1266) {
return 8;
}
else if (Rs >= 1266) {
return 4;
}
}
else if (MVDiff >= 3) {
if (SC < 2231) {
return 8;
}
else if (SC >= 2231) {
return 4;
}
}
}
else if (Rs >= 1277) {
if (negBalance < 42) {
if (diffRsCsdiff < 4) {
if (AFD < 5) {
return 1;
}
else if (AFD >= 5) {
if (Cs < 941) {
if (Rs < 1360) {
return 8;
}
else if (Rs >= 1360) {
return 1;
}
}
else if (Cs >= 941) {
return 8;
}
}
}
else if (diffRsCsdiff >= 4) {
return 2;
}
}
else if (negBalance >= 42) {
if (ssDCval < 55) {
return 4;
}
else if (ssDCval >= 55) {
return 8;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 1400) {
if (SC < 2329) {
if (Rs < 1163) {
if (ssDCval < 80) {
if (RsDiff < 2) {
return 8;
}
else if (RsDiff >= 2) {
if (Rs < 1084) {
if (Cs < 1206) {
return 8;
}
else if (Cs >= 1206) {
return 4;
}
}
else if (Rs >= 1084) {
return 4;
}
}
}
else if (ssDCval >= 80) {
return 4;
}
}
else if (Rs >= 1163) {
if (ssDCval < 67) {
if (Rs < 1184) {
return 1;
}
else if (Rs >= 1184) {
return 4;
}
}
else if (ssDCval >= 67) {
if (SC < 2189) {
if (SC < 2184) {
return 8;
}
else if (SC >= 2184) {
return 1;
}
}
else if (SC >= 2189) {
return 8;
}
}
}
}
else if (SC >= 2329) {
if (RsCsDiff < 7) {
if (mu_mv_mag_sq < 118) {
return 8;
}
else if (mu_mv_mag_sq >= 118) {
if (Cs < 1123) {
return 8;
}
else if (Cs >= 1123) {
return 1;
}
}
}
else if (RsCsDiff >= 7) {
if (diffRsCsdiff < 6) {
if (mu_mv_mag_sq < 120) {
if (Cs < 1195) {
return 8;
}
else if (Cs >= 1195) {
if (ssDCval < 56) {
return 8;
}
else if (ssDCval >= 56) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 120) {
return 4;
}
}
else if (diffRsCsdiff >= 6) {
if (Cs < 1247) {
if (Cs < 1195) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (Cs >= 1195) {
return 1;
}
}
else if (Cs >= 1247) {
return 4;
}
}
}
}
}
else if (Rs >= 1400) {
if (ssDCval < 49) {
if (Rs < 1829) {
if (Rs < 1547) {
return 4;
}
else if (Rs >= 1547) {
return 4;
}
}
else if (Rs >= 1829) {
return 4;
}
}
else if (ssDCval >= 49) {
if (Cs < 1282) {
if (diffRsCsdiff < 7) {
return 8;
}
else if (diffRsCsdiff >= 7) {
if (MVDiff < 4) {
return 8;
}
else if (MVDiff >= 4) {
if (Rs < 1750) {
return 1;
}
else if (Rs >= 1750) {
return 4;
}
}
}
}
else if (Cs >= 1282) {
if (Rs < 1997) {
return 1;
}
else if (Rs >= 1997) {
return 4;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 88) {
if (mu_mv_mag_sq < 103) {
if (SC < 1411) {
if (negBalance < 78) {
if (TSC < 4) {
if (Rs < 437) {
if (MVDiff < 7) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (ssDCval < 127) {
if (diffMVdiffVal < -2) {
if (Rs < 371) {
return 4;
}
else if (Rs >= 371) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
if (SC < 855) {
if (Rs < 288) {
return 4;
}
else if (Rs >= 288) {
return 1;
}
}
else if (SC >= 855) {
return 4;
}
}
}
else if (ssDCval >= 127) {
if (Rs < 351) {
return 2;
}
else if (Rs >= 351) {
if (Rs < 430) {
return 8;
}
else if (Rs >= 430) {
return 1;
}
}
}
}
}
else if (MVDiff >= 7) {
if (mu_mv_mag_sq < 8) {
if (Rs < 344) {
return 1;
}
else if (Rs >= 344) {
return 8;
}
}
else if (mu_mv_mag_sq >= 8) {
if (Rs < 402) {
return 2;
}
else if (Rs >= 402) {
if (MVDiff < 13) {
return 2;
}
else if (MVDiff >= 13) {
return 8;
}
}
}
}
}
else if (Rs >= 437) {
if (ssDCval < 121) {
if (MVDiff < 3) {
if (mcTcor < 99) {
if (Cs < 503) {
return 8;
}
else if (Cs >= 503) {
return 2;
}
}
else if (mcTcor >= 99) {
return 1;
}
}
else if (MVDiff >= 3) {
if (negBalance < 22) {
if (negBalance < 19) {
return 1;
}
else if (negBalance >= 19) {
if (Rs < 524) {
return 4;
}
else if (Rs >= 524) {
return 2;
}
}
}
else if (negBalance >= 22) {
if (posBalance < 37) {
if (Cs < 517) {
return 2;
}
else if (Cs >= 517) {
return 1;
}
}
else if (posBalance >= 37) {
if (diffMVdiffVal < 5) {
if (Cs < 827) {
return 2;
}
else if (Cs >= 827) {
return 1;
}
}
else if (diffMVdiffVal >= 5) {
if (Rs < 519) {
return 2;
}
else if (Rs >= 519) {
return 1;
}
}
}
}
}
}
else if (ssDCval >= 121) {
if (SC < 1222) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (SC >= 1222) {
if (MVDiff < 6) {
if (Rs < 577) {
return 1;
}
else if (Rs >= 577) {
return 4;
}
}
else if (MVDiff >= 6) {
if (Cs < 591) {
return 2;
}
else if (Cs >= 591) {
return 1;
}
}
}
}
}
}
else if (TSC >= 4) {
if (mu_mv_mag_sq < 75) {
if (ssDCval < 109) {
if (mcTcor < 98) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (Cs < 496) {
return 8;
}
else if (Cs >= 496) {
return 1;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 383) {
if (Rs < 329) {
return 8;
}
else if (Rs >= 329) {
return 4;
}
}
else if (Rs >= 383) {
if (posBalance < 62) {
if (Rs < 713) {
return 2;
}
else if (Rs >= 713) {
return 1;
}
}
else if (posBalance >= 62) {
if (Rs < 624) {
return 1;
}
else if (Rs >= 624) {
return 8;
}
}
}
}
}
else if (ssDCval >= 109) {
if (posBalance < 58) {
if (refDCval < 114) {
if (Rs < 439) {
return 4;
}
else if (Rs >= 439) {
return 8;
}
}
else if (refDCval >= 114) {
if (MVDiff < 11) {
if (mu_mv_mag_sq < 26) {
if (Cs < 514) {
return 2;
}
else if (Cs >= 514) {
return 8;
}
}
else if (mu_mv_mag_sq >= 26) {
if (Rs < 533) {
return 8;
}
else if (Rs >= 533) {
return 2;
}
}
}
else if (MVDiff >= 11) {
if (AFD < 6) {
return 4;
}
else if (AFD >= 6) {
return 2;
}
}
}
}
else if (posBalance >= 58) {
if (Rs < 723) {
if (diffMVdiffVal < 2) {
return 4;
}
else if (diffMVdiffVal >= 2) {
return 2;
}
}
else if (Rs >= 723) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 75) {
if (posBalance < 50) {
if (Rs < 565) {
return 8;
}
else if (Rs >= 565) {
return 2;
}
}
else if (posBalance >= 50) {
if (Rs < 358) {
return 4;
}
else if (Rs >= 358) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 2;
}
}
}
}
}
}
else if (negBalance >= 78) {
if (posBalance < 34) {
return 8;
}
else if (posBalance >= 34) {
if (Rs < 691) {
return 2;
}
else if (Rs >= 691) {
return 1;
}
}
}
}
else if (SC >= 1411) {
if (Cs < 609) {
if (RsDiff < 8) {
if (posBalance < 69) {
if (negBalance < 37) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
return 1;
}
}
else if (negBalance >= 37) {
if (MVDiff < 21) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (ssDCval < 130) {
return 1;
}
else if (ssDCval >= 130) {
return 8;
}
}
}
else if (MVDiff >= 21) {
if (Rs < 1249) {
return 2;
}
else if (Rs >= 1249) {
return 1;
}
}
}
}
else if (posBalance >= 69) {
return 2;
}
}
else if (RsDiff >= 8) {
if (mu_mv_mag_sq < 11) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (ssDCval < 94) {
return 8;
}
else if (ssDCval >= 94) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 11) {
if (mcTcor < 98) {
return 1;
}
else if (mcTcor >= 98) {
if (Cs < 530) {
return 1;
}
else if (Cs >= 530) {
return 2;
}
}
}
}
}
else if (Cs >= 609) {
if (MVDiff < 5) {
if (Rs < 659) {
if (mcTcor < 98) {
if (Cs < 919) {
return 1;
}
else if (Cs >= 919) {
return 8;
}
}
else if (mcTcor >= 98) {
if (ssDCval < 100) {
return 8;
}
else if (ssDCval >= 100) {
return 1;
}
}
}
else if (Rs >= 659) {
if (ssDCval < 94) {
if (RsDiff < 3) {
if (refDCval < 91) {
return 4;
}
else if (refDCval >= 91) {
if (Rs < 1380) {
if (Rs < 743) {
return 1;
}
else if (Rs >= 743) {
return 4;
}
}
else if (Rs >= 1380) {
return 8;
}
}
}
else if (RsDiff >= 3) {
if (Rs < 1355) {
if (Rs < 922) {
return 4;
}
else if (Rs >= 922) {
return 8;
}
}
else if (Rs >= 1355) {
if (ssDCval < 91) {
return 1;
}
else if (ssDCval >= 91) {
if (Rs < 1774) {
return 1;
}
else if (Rs >= 1774) {
return 8;
}
}
}
}
}
else if (ssDCval >= 94) {
if (ssDCval < 122) {
if (tcor < 91) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
return 8;
}
}
else if (tcor >= 91) {
if (Rs < 1051) {
if (ssDCval < 107) {
return 4;
}
else if (ssDCval >= 107) {
if (ssDCval < 115) {
return 1;
}
else if (ssDCval >= 115) {
return 8;
}
}
}
else if (Rs >= 1051) {
return 8;
}
}
}
else if (ssDCval >= 122) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
if (refDCval < 133) {
if (negBalance < 42) {
if (Cs < 1260) {
return 2;
}
else if (Cs >= 1260) {
return 8;
}
}
else if (negBalance >= 42) {
if (TSC < 5) {
if (Rs < 1364) {
return 1;
}
else if (Rs >= 1364) {
return 2;
}
}
else if (TSC >= 5) {
return 4;
}
}
}
else if (refDCval >= 133) {
if (posBalance < 40) {
if (Rs < 2042) {
return 4;
}
else if (Rs >= 2042) {
return 8;
}
}
else if (posBalance >= 40) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (ssDCval < 136) {
if (Rs < 1735) {
return 8;
}
else if (Rs >= 1735) {
return 2;
}
}
else if (ssDCval >= 136) {
return 8;
}
}
}
}
}
}
}
}
}
else if (MVDiff >= 5) {
if (tcor < 93) {
if (refDCval < 102) {
if (ssDCval < 96) {
if (MVDiff < 10) {
if (diffRsCsdiff < -7) {
return 1;
}
else if (diffRsCsdiff >= -7) {
if (Rs < 1719) {
return 8;
}
else if (Rs >= 1719) {
return 4;
}
}
}
else if (MVDiff >= 10) {
if (MVDiff < 18) {
return 2;
}
else if (MVDiff >= 18) {
if (Cs < 880) {
if (Rs < 1140) {
return 2;
}
else if (Rs >= 1140) {
return 1;
}
}
else if (Cs >= 880) {
return 8;
}
}
}
}
else if (ssDCval >= 96) {
if (TSC < 6) {
if (Rs < 1903) {
if (Cs < 1180) {
return 1;
}
else if (Cs >= 1180) {
return 2;
}
}
else if (Rs >= 1903) {
if (Rs < 2283) {
return 4;
}
else if (Rs >= 2283) {
return 1;
}
}
}
else if (TSC >= 6) {
if (refDCval < 95) {
return 2;
}
else if (refDCval >= 95) {
return 4;
}
}
}
}
else if (refDCval >= 102) {
if (RsCsDiff < 83) {
if (Rs < 885) {
if (MVDiff < 8) {
if (Rs < 693) {
return 1;
}
else if (Rs >= 693) {
return 4;
}
}
else if (MVDiff >= 8) {
if (Rs < 834) {
return 2;
}
else if (Rs >= 834) {
if (Rs < 880) {
return 1;
}
else if (Rs >= 880) {
return 2;
}
}
}
}
else if (Rs >= 885) {
if (mu_mv_mag_sq < 3) {
if (SC < 1924) {
if (AFD < 13) {
return 1;
}
else if (AFD >= 13) {
return 8;
}
}
else if (SC >= 1924) {
if (Cs < 640) {
return 1;
}
else if (Cs >= 640) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (negBalance < 59) {
return 8;
}
else if (negBalance >= 59) {
return 2;
}
}
}
}
else if (RsCsDiff >= 83) {
if (Rs < 1789) {
if (gchDC < 6) {
if (Cs < 656) {
return 8;
}
else if (Cs >= 656) {
if (Cs < 994) {
return 2;
}
else if (Cs >= 994) {
if (Rs < 1564) {
return 2;
}
else if (Rs >= 1564) {
return 8;
}
}
}
}
else if (gchDC >= 6) {
if (Cs < 694) {
return 8;
}
else if (Cs >= 694) {
return 1;
}
}
}
else if (Rs >= 1789) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
return 1;
}
}
}
}
}
else if (tcor >= 93) {
if (negBalance < 30) {
if (Rs < 1589) {
if (Rs < 647) {
return 1;
}
else if (Rs >= 647) {
return 8;
}
}
else if (Rs >= 1589) {
if (ssDCval < 118) {
return 1;
}
else if (ssDCval >= 118) {
return 4;
}
}
}
else if (negBalance >= 30) {
if (diffRsCsdiff < -10) {
if (Cs < 770) {
if (negBalance < 56) {
return 1;
}
else if (negBalance >= 56) {
if (Rs < 2191) {
return 1;
}
else if (Rs >= 2191) {
return 2;
}
}
}
else if (Cs >= 770) {
if (mu_mv_mag_sq < 6) {
if (ssDCval < 113) {
return 4;
}
else if (ssDCval >= 113) {
if (Rs < 1658) {
return 2;
}
else if (Rs >= 1658) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 6) {
if (ssDCval < 153) {
return 8;
}
else if (ssDCval >= 153) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= -10) {
if (mu_mv_mag_sq < 3) {
if (Cs < 680) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
if (Rs < 1619) {
return 8;
}
else if (Rs >= 1619) {
return 1;
}
}
}
else if (Cs >= 680) {
if (Rs < 2276) {
return 8;
}
else if (Rs >= 2276) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 161) {
if (negBalance < 35) {
if (posBalance < 48) {
if (AFD < 6) {
if (Rs < 943) {
return 1;
}
else if (Rs >= 943) {
return 8;
}
}
else if (AFD >= 6) {
return 1;
}
}
else if (posBalance >= 48) {
return 8;
}
}
else if (negBalance >= 35) {
if (Cs < 826) {
if (Rs < 2026) {
if (diffAFD < -2) {
if (Rs < 748) {
return 8;
}
else if (Rs >= 748) {
return 2;
}
}
else if (diffAFD >= -2) {
if (tcor < 94) {
if (Rs < 1211) {
return 1;
}
else if (Rs >= 1211) {
return 8;
}
}
else if (tcor >= 94) {
return 8;
}
}
}
else if (Rs >= 2026) {
return 2;
}
}
else if (Cs >= 826) {
return 8;
}
}
}
else if (ssDCval >= 161) {
if (Rs < 723) {
return 4;
}
else if (Rs >= 723) {
return 2;
}
}
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 103) {
if (ssDCval < 133) {
if (Cs < 1215) {
if (Rs < 342) {
if (ssDCval < 123) {
if (mu_mv_mag_sq < 118) {
return 4;
}
else if (mu_mv_mag_sq >= 118) {
return 8;
}
}
else if (ssDCval >= 123) {
return 2;
}
}
else if (Rs >= 342) {
if (refDCval < 97) {
if (MVDiff < 1) {
if (diffMVdiffVal < -1) {
if (Cs < 987) {
if (Rs < 647) {
return 1;
}
else if (Rs >= 647) {
return 4;
}
}
else if (Cs >= 987) {
if (RsDiff < 3) {
return 8;
}
else if (RsDiff >= 3) {
return 2;
}
}
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2147) {
if (posBalance < 24) {
return 4;
}
else if (posBalance >= 24) {
if (diffRsCsdiff < -1) {
if (ssDCval < 93) {
return 8;
}
else if (ssDCval >= 93) {
if (posBalance < 44) {
return 8;
}
else if (posBalance >= 44) {
if (Cs < 1014) {
return 4;
}
else if (Cs >= 1014) {
return 8;
}
}
}
}
else if (diffRsCsdiff >= -1) {
if (negBalance < 27) {
return 8;
}
else if (negBalance >= 27) {
if (CsDiff < 1) {
if (Rs < 888) {
return 1;
}
else if (Rs >= 888) {
return 4;
}
}
else if (CsDiff >= 1) {
return 8;
}
}
}
}
}
else if (Rs >= 2147) {
if (CsDiff < 1) {
if (diffRsCsdiff < -1) {
return 4;
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
else if (CsDiff >= 1) {
return 4;
}
}
}
}
else if (MVDiff >= 1) {
if (SC < 1744) {
if (Rs < 474) {
return 8;
}
else if (Rs >= 474) {
if (Rs < 790) {
return 1;
}
else if (Rs >= 790) {
return 8;
}
}
}
else if (SC >= 1744) {
if (AFD < 3) {
return 4;
}
else if (AFD >= 3) {
if (Rs < 1017) {
return 4;
}
else if (Rs >= 1017) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 97) {
if (Rs < 1673) {
if (Rs < 514) {
if (ssDCval < 102) {
if (Rs < 478) {
return 1;
}
else if (Rs >= 478) {
return 8;
}
}
else if (ssDCval >= 102) {
if (mu_mv_mag_sq < 108) {
if (Rs < 484) {
if (Cs < 839) {
return 8;
}
else if (Cs >= 839) {
return 4;
}
}
else if (Rs >= 484) {
return 1;
}
}
else if (mu_mv_mag_sq >= 108) {
if (negBalance < 29) {
if (Rs < 387) {
return 4;
}
else if (Rs >= 387) {
if (Cs < 567) {
return 1;
}
else if (Cs >= 567) {
return 8;
}
}
}
else if (negBalance >= 29) {
if (posBalance < 45) {
if (Rs < 371) {
return 4;
}
else if (Rs >= 371) {
return 8;
}
}
else if (posBalance >= 45) {
return 8;
}
}
}
}
}
else if (Rs >= 514) {
if (ssDCval < 121) {
if (ssDCval < 97) {
if (Rs < 986) {
return 1;
}
else if (Rs >= 986) {
return 8;
}
}
else if (ssDCval >= 97) {
return 8;
}
}
else if (ssDCval >= 121) {
if (Rs < 899) {
return 8;
}
else if (Rs >= 899) {
if (Rs < 1090) {
if (Rs < 1007) {
return 4;
}
else if (Rs >= 1007) {
return 8;
}
}
else if (Rs >= 1090) {
return 1;
}
}
}
}
}
else if (Rs >= 1673) {
if (mu_mv_mag_sq < 123) {
if (ssDCval < 121) {
return 8;
}
else if (ssDCval >= 121) {
if (Rs < 1865) {
return 2;
}
else if (Rs >= 1865) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Cs < 1014) {
return 1;
}
else if (Cs >= 1014) {
if (Rs < 1885) {
return 2;
}
else if (Rs >= 1885) {
return 8;
}
}
}
}
}
}
}
else if (Cs >= 1215) {
if (Rs < 634) {
return 1;
}
else if (Rs >= 634) {
if (Cs < 1217) {
return 1;
}
else if (Cs >= 1217) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
}
}
}
else if (ssDCval >= 133) {
if (mu_mv_mag_sq < 128) {
return 8;
}
else if (mu_mv_mag_sq >= 128) {
if (Rs < 1147) {
return 2;
}
else if (Rs >= 1147) {
return 1;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2354) {
if (MVDiff < 17) {
if (refDCval < 109) {
if (posBalance < 48) {
if (Rs < 3783) {
if (RsCsDiff < 53) {
if (mu_mv_mag_sq < 123) {
if (ssDCval < 104) {
if (Cs < 669) {
if (Rs < 2502) {
if (refDCval < 49) {
return 4;
}
else if (refDCval >= 49) {
if (posBalance < 32) {
return 4;
}
else if (posBalance >= 32) {
return 1;
}
}
}
else if (Rs >= 2502) {
if (mu_mv_mag_sq < 60) {
if (Rs < 3071) {
return 8;
}
else if (Rs >= 3071) {
return 2;
}
}
else if (mu_mv_mag_sq >= 60) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (Rs < 2922) {
return 4;
}
else if (Rs >= 2922) {
return 2;
}
}
}
}
}
else if (Cs >= 669) {
if (negBalance < 28) {
if (Cs < 768) {
if (MVDiff < 5) {
if (Cs < 712) {
return 1;
}
else if (Cs >= 712) {
return 8;
}
}
else if (MVDiff >= 5) {
return 4;
}
}
else if (Cs >= 768) {
if (RsCsDiff < 7) {
if (Rs < 3683) {
if (ssDCval < 97) {
return 8;
}
else if (ssDCval >= 97) {
if (Rs < 3237) {
return 8;
}
else if (Rs >= 3237) {
return 4;
}
}
}
else if (Rs >= 3683) {
return 4;
}
}
else if (RsCsDiff >= 7) {
if (RsDiff < 6) {
if (diffTSC < -2) {
return 2;
}
else if (diffTSC >= -2) {
if (negBalance < 23) {
return 4;
}
else if (negBalance >= 23) {
if (diffRsCsdiff < 6) {
if (mu_mv_mag_sq < 111) {
if (Rs < 3553) {
if (SC < 3622) {
return 8;
}
else if (SC >= 3622) {
if (ssDCval < 100) {
if (SC < 3663) {
return 4;
}
else if (SC >= 3663) {
return 4;
}
}
else if (ssDCval >= 100) {
return 4;
}
}
}
else if (Rs >= 3553) {
return 8;
}
}
else if (mu_mv_mag_sq >= 111) {
if (Rs < 3073) {
return 2;
}
else if (Rs >= 3073) {
return 4;
}
}
}
else if (diffRsCsdiff >= 6) {
return 8;
}
}
}
}
else if (RsDiff >= 6) {
if (Cs < 1151) {
if (Rs < 3459) {
if (Rs < 3410) {
return 8;
}
else if (Rs >= 3410) {
return 4;
}
}
else if (Rs >= 3459) {
return 8;
}
}
else if (Cs >= 1151) {
return 1;
}
}
}
}
}
else if (negBalance >= 28) {
if (Rs < 2990) {
if (ssDCval < 74) {
if (Rs < 2578) {
if (Cs < 1070) {
if (ssDCval < 73) {
return 2;
}
else if (ssDCval >= 73) {
if (Rs < 2500) {
if (Rs < 2436) {
return 4;
}
else if (Rs >= 2436) {
return 4;
}
}
else if (Rs >= 2500) {
return 2;
}
}
}
else if (Cs >= 1070) {
if (ssDCval < 50) {
return 4;
}
else if (ssDCval >= 50) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (Rs >= 2578) {
if (mcTcor < 94) {
return 4;
}
else if (mcTcor >= 94) {
if (Rs < 2944) {
if (Cs < 790) {
return 1;
}
else if (Cs >= 790) {
return 2;
}
}
else if (Rs >= 2944) {
if (Rs < 2957) {
return 4;
}
else if (Rs >= 2957) {
return 1;
}
}
}
}
}
else if (ssDCval >= 74) {
if (mu_mv_mag_sq < 55) {
return 1;
}
else if (mu_mv_mag_sq >= 55) {
if (diffMVdiffVal < -2) {
if (MVDiff < 9) {
return 1;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (diffMVdiffVal >= -2) {
if (Cs < 1002) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
if (Rs < 2606) {
return 4;
}
else if (Rs >= 2606) {
return 4;
}
}
}
else if (Cs >= 1002) {
return 8;
}
}
}
}
}
else if (Rs >= 2990) {
if (ssDCval < 79) {
if (diffRsCsdiff < -8) {
if (RsDiff < 6) {
if (diffAFD < -5) {
return 2;
}
else if (diffAFD >= -5) {
return 4;
}
}
else if (RsDiff >= 6) {
return 2;
}
}
else if (diffRsCsdiff >= -8) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
}
else if (ssDCval >= 79) {
if (mcTcor < 97) {
if (Rs < 3057) {
return 4;
}
else if (Rs >= 3057) {
if (Rs < 3392) {
return 8;
}
else if (Rs >= 3392) {
if (Cs < 1014) {
return 8;
}
else if (Cs >= 1014) {
return 8;
}
}
}
}
else if (mcTcor >= 97) {
if (ssDCval < 97) {
return 2;
}
else if (ssDCval >= 97) {
if (Cs < 913) {
return 2;
}
else if (Cs >= 913) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 104) {
if (posBalance < 36) {
if (Rs < 2864) {
return 8;
}
else if (Rs >= 2864) {
return 1;
}
}
else if (posBalance >= 36) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 123) {
if (Cs < 1072) {
if (TSCindex < 2) {
if (posBalance < 21) {
if (Rs < 3185) {
return 4;
}
else if (Rs >= 3185) {
return 8;
}
}
else if (posBalance >= 21) {
return 8;
}
}
else if (TSCindex >= 2) {
if (Rs < 3007) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
if (Rs < 2698) {
return 2;
}
else if (Rs >= 2698) {
return 1;
}
}
}
else if (Rs >= 3007) {
return 2;
}
}
}
else if (Cs >= 1072) {
if (SC < 4981) {
if (RsCsDiff < 5) {
if (negBalance < 20) {
return 8;
}
else if (negBalance >= 20) {
if (ssDCval < 92) {
return 8;
}
else if (ssDCval >= 92) {
return 2;
}
}
}
else if (RsCsDiff >= 5) {
if (diffTSC < -1) {
return 4;
}
else if (diffTSC >= -1) {
if (Rs < 3180) {
return 8;
}
else if (Rs >= 3180) {
if (Rs < 3618) {
return 4;
}
else if (Rs >= 3618) {
return 8;
}
}
}
}
}
else if (SC >= 4981) {
return 4;
}
}
}
}
else if (RsCsDiff >= 53) {
if (negBalance < 47) {
if (refDCval < 68) {
if (refDCval < 58) {
if (MVDiff < 3) {
return 2;
}
else if (MVDiff >= 3) {
if (SC < 4293) {
return 4;
}
else if (SC >= 4293) {
return 2;
}
}
}
else if (refDCval >= 58) {
if (diffRsCsdiff < -24) {
if (diffAFD < -2) {
if (Cs < 930) {
return 1;
}
else if (Cs >= 930) {
return 2;
}
}
else if (diffAFD >= -2) {
if (Rs < 3299) {
return 4;
}
else if (Rs >= 3299) {
if (Cs < 1079) {
return 4;
}
else if (Cs >= 1079) {
if (Rs < 3534) {
return 2;
}
else if (Rs >= 3534) {
return 1;
}
}
}
}
}
else if (diffRsCsdiff >= -24) {
if (diffRsCsdiff < 45) {
if (mu_mv_mag_sq < 74) {
if (posBalance < 36) {
return 1;
}
else if (posBalance >= 36) {
if (RsDiff < 14) {
if (AFD < 14) {
if (mcTcor < 93) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (mcTcor >= 93) {
if (refDCval < 67) {
return 2;
}
else if (refDCval >= 67) {
if (Rs < 2690) {
return 2;
}
else if (Rs >= 2690) {
return 4;
}
}
}
}
else if (AFD >= 14) {
return 2;
}
}
else if (RsDiff >= 14) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 74) {
if (Rs < 2379) {
return 2;
}
else if (Rs >= 2379) {
return 4;
}
}
}
else if (diffRsCsdiff >= 45) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
return 2;
}
}
}
}
}
else if (refDCval >= 68) {
if (ssDCval < 89) {
if (mu_mv_mag_sq < 80) {
if (SC < 3761) {
if (ssDCval < 68) {
if (Rs < 2847) {
return 2;
}
else if (Rs >= 2847) {
return 8;
}
}
else if (ssDCval >= 68) {
return 1;
}
}
else if (SC >= 3761) {
if (negBalance < 44) {
if (SC < 4657) {
if (posBalance < 40) {
return 1;
}
else if (posBalance >= 40) {
if (mu_mv_mag_sq < 24) {
if (MVDiff < 10) {
if (SC < 4544) {
if (MVDiff < 2) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 1;
}
}
else if (MVDiff >= 2) {
return 4;
}
}
else if (SC >= 4544) {
return 1;
}
}
else if (MVDiff >= 10) {
return 1;
}
}
else if (mu_mv_mag_sq >= 24) {
if (Rs < 2762) {
return 2;
}
else if (Rs >= 2762) {
return 4;
}
}
}
}
else if (SC >= 4657) {
if (Rs < 3561) {
return 4;
}
else if (Rs >= 3561) {
return 1;
}
}
}
else if (negBalance >= 44) {
if (Rs < 3100) {
return 2;
}
else if (Rs >= 3100) {
if (Rs < 3140) {
return 4;
}
else if (Rs >= 3140) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 80) {
if (posBalance < 38) {
return 1;
}
else if (posBalance >= 38) {
return 2;
}
}
}
else if (ssDCval >= 89) {
if (negBalance < 42) {
if (ssDCval < 97) {
if (Rs < 3654) {
return 4;
}
else if (Rs >= 3654) {
return 8;
}
}
else if (ssDCval >= 97) {
return 8;
}
}
else if (negBalance >= 42) {
return 2;
}
}
}
}
else if (negBalance >= 47) {
if (negBalance < 51) {
if (posBalance < 43) {
if (MVDiff < 8) {
if (Rs < 2793) {
return 2;
}
else if (Rs >= 2793) {
if (Rs < 3414) {
return 8;
}
else if (Rs >= 3414) {
return 4;
}
}
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (posBalance >= 43) {
return 2;
}
}
else if (negBalance >= 51) {
return 2;
}
}
}
}
else if (Rs >= 3783) {
if (Cs < 1093) {
if (refDCval < 91) {
if (Rs < 3818) {
if (Cs < 966) {
return 4;
}
else if (Cs >= 966) {
return 1;
}
}
else if (Rs >= 3818) {
return 4;
}
}
else if (refDCval >= 91) {
if (CsDiff < 3) {
return 2;
}
else if (CsDiff >= 3) {
if (Cs < 1065) {
if (Rs < 3856) {
return 8;
}
else if (Rs >= 3856) {
return 4;
}
}
else if (Cs >= 1065) {
return 4;
}
}
}
}
else if (Cs >= 1093) {
if (SC < 5336) {
if (negBalance < 34) {
if (Cs < 1275) {
if (tcor < 94) {
if (Rs < 4012) {
if (Rs < 3812) {
return 4;
}
else if (Rs >= 3812) {
return 4;
}
}
else if (Rs >= 4012) {
if (Cs < 1171) {
if (Rs < 4157) {
return 8;
}
else if (Rs >= 4157) {
return 8;
}
}
else if (Cs >= 1171) {
return 4;
}
}
}
else if (tcor >= 94) {
if (ssDCval < 68) {
return 1;
}
else if (ssDCval >= 68) {
return 4;
}
}
}
else if (Cs >= 1275) {
if (Rs < 3892) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 2;
}
}
else if (Rs >= 3892) {
if (mu_mv_mag_sq < 125) {
return 4;
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
}
}
else if (negBalance >= 34) {
if (tcor < 85) {
return 4;
}
else if (tcor >= 85) {
if (RsCsDiff < 42) {
return 4;
}
else if (RsCsDiff >= 42) {
return 2;
}
}
}
}
else if (SC >= 5336) {
if (ssDCval < 72) {
if (Rs < 4645) {
return 1;
}
else if (Rs >= 4645) {
return 4;
}
}
else if (ssDCval >= 72) {
if (Cs < 1206) {
if (Rs < 4219) {
return 8;
}
else if (Rs >= 4219) {
if (diffRsCsdiff < 5) {
return 4;
}
else if (diffRsCsdiff >= 5) {
return 4;
}
}
}
else if (Cs >= 1206) {
if (Rs < 4600) {
if (mu_mv_mag_sq < 56) {
return 8;
}
else if (mu_mv_mag_sq >= 56) {
return 8;
}
}
else if (Rs >= 4600) {
return 2;
}
}
}
}
}
}
}
else if (posBalance >= 48) {
if (ssDCval < 91) {
if (CsDiff < 3) {
if (Rs < 3707) {
if (Cs < 215) {
if (Rs < 2630) {
return 2;
}
else if (Rs >= 2630) {
return 1;
}
}
else if (Cs >= 215) {
if (negBalance < 22) {
if (SC < 3216) {
return 1;
}
else if (SC >= 3216) {
if (Cs < 1121) {
if (RsDiff < 4) {
if (Rs < 2709) {
return 8;
}
else if (Rs >= 2709) {
return 8;
}
}
else if (RsDiff >= 4) {
if (diffMVdiffVal < 1) {
return 8;
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
}
else if (Cs >= 1121) {
return 8;
}
}
}
else if (negBalance >= 22) {
if (gchDC < 2) {
if (ssDCval < 59) {
if (Rs < 2694) {
return 1;
}
else if (Rs >= 2694) {
return 4;
}
}
else if (ssDCval >= 59) {
if (posBalance < 54) {
if (MVDiff < 6) {
if (diffAFD < -2) {
if (Cs < 845) {
return 1;
}
else if (Cs >= 845) {
return 2;
}
}
else if (diffAFD >= -2) {
if (diffRsCsdiff < -3) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
if (Rs < 3504) {
return 2;
}
else if (Rs >= 3504) {
if (Rs < 3681) {
return 4;
}
else if (Rs >= 3681) {
return 2;
}
}
}
}
else if (diffRsCsdiff >= -3) {
return 2;
}
}
}
else if (MVDiff >= 6) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
if (Rs < 2467) {
return 1;
}
else if (Rs >= 2467) {
return 2;
}
}
}
}
else if (posBalance >= 54) {
return 2;
}
}
}
else if (gchDC >= 2) {
if (posBalance < 51) {
if (Rs < 2627) {
return 2;
}
else if (Rs >= 2627) {
return 8;
}
}
else if (posBalance >= 51) {
if (posBalance < 56) {
return 1;
}
else if (posBalance >= 56) {
if (SC < 2786) {
return 1;
}
else if (SC >= 2786) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 3707) {
return 4;
}
}
else if (CsDiff >= 3) {
if (refDCval < 66) {
if (SC < 3477) {
if (diffRsCsdiff < -160) {
return 4;
}
else if (diffRsCsdiff >= -160) {
if (mu_mv_mag_sq < 103) {
return 2;
}
else if (mu_mv_mag_sq >= 103) {
return 4;
}
}
}
else if (SC >= 3477) {
if (tcor < 84) {
if (diffMVdiffVal < 7) {
return 4;
}
else if (diffMVdiffVal >= 7) {
return 2;
}
}
else if (tcor >= 84) {
if (Cs < 993) {
if (tcor < 90) {
if (Cs < 890) {
return 2;
}
else if (Cs >= 890) {
return 4;
}
}
else if (tcor >= 90) {
return 2;
}
}
else if (Cs >= 993) {
if (refDCval < 63) {
if (mu_mv_mag_sq < 3) {
return 2;
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
else if (refDCval >= 63) {
if (negBalance < 43) {
if (RsCsDiff < 38) {
return 4;
}
else if (RsCsDiff >= 38) {
if (diffMVdiffVal < 2) {
return 2;
}
else if (diffMVdiffVal >= 2) {
if (AFD < 9) {
return 4;
}
else if (AFD >= 9) {
return 2;
}
}
}
}
else if (negBalance >= 43) {
return 4;
}
}
}
}
}
}
else if (refDCval >= 66) {
if (negBalance < 39) {
if (posBalance < 56) {
if (mu_mv_mag_sq < 66) {
if (tcor < 90) {
if (diffMVdiffVal < 2) {
if (Cs < 1233) {
return 4;
}
else if (Cs >= 1233) {
return 2;
}
}
else if (diffMVdiffVal >= 2) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
return 1;
}
}
}
else if (tcor >= 90) {
if (posBalance < 50) {
return 4;
}
else if (posBalance >= 50) {
if (negBalance < 37) {
if (ssDCval < 76) {
return 2;
}
else if (ssDCval >= 76) {
return 1;
}
}
else if (negBalance >= 37) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (Cs < 766) {
return 4;
}
else if (Cs >= 766) {
return 2;
}
}
}
else if (posBalance >= 56) {
if (SC < 3255) {
if (ssDCval < 74) {
return 2;
}
else if (ssDCval >= 74) {
if (diffAFD < -1) {
return 2;
}
else if (diffAFD >= -1) {
return 1;
}
}
}
else if (SC >= 3255) {
if (RsDiff < 5) {
return 4;
}
else if (RsDiff >= 5) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (ssDCval < 84) {
if (RsCsDiff < 375) {
if (diffRsCsdiff < -191) {
return 4;
}
else if (diffRsCsdiff >= -191) {
if (Rs < 2423) {
if (Rs < 2421) {
return 2;
}
else if (Rs >= 2421) {
return 4;
}
}
else if (Rs >= 2423) {
return 2;
}
}
}
else if (RsCsDiff >= 375) {
if (Cs < 692) {
return 2;
}
else if (Cs >= 692) {
if (Rs < 3506) {
return 4;
}
else if (Rs >= 3506) {
return 2;
}
}
}
}
else if (ssDCval >= 84) {
if (mcTcor < 93) {
if (Rs < 2585) {
if (RsDiff < 16) {
return 4;
}
else if (RsDiff >= 16) {
return 2;
}
}
else if (Rs >= 2585) {
if (TSC < 8) {
if (Rs < 3035) {
return 2;
}
else if (Rs >= 3035) {
return 1;
}
}
else if (TSC >= 8) {
return 2;
}
}
}
else if (mcTcor >= 93) {
if (RsDiff < 8) {
if (AFD < 10) {
return 2;
}
else if (AFD >= 10) {
return 1;
}
}
else if (RsDiff >= 8) {
if (tcor < 70) {
if (Rs < 3512) {
return 2;
}
else if (Rs >= 3512) {
return 1;
}
}
else if (tcor >= 70) {
return 2;
}
}
}
}
}
}
}
}
else if (ssDCval >= 91) {
if (RsDiff < 11) {
if (tcor < 96) {
if (negBalance < 53) {
if (RsCsDiff < 67) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
if (MVDiff < 5) {
return 8;
}
else if (MVDiff >= 5) {
if (MVDiff < 8) {
if (Rs < 3014) {
return 4;
}
else if (Rs >= 3014) {
return 4;
}
}
else if (MVDiff >= 8) {
if (diffAFD < 1) {
if (Cs < 692) {
return 8;
}
else if (Cs >= 692) {
return 8;
}
}
else if (diffAFD >= 1) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 67) {
if (AFD < 12) {
if (SC < 5317) {
if (AFD < 8) {
return 8;
}
else if (AFD >= 8) {
if (gchDC < 4) {
return 2;
}
else if (gchDC >= 4) {
return 8;
}
}
}
else if (SC >= 5317) {
if (Rs < 4258) {
return 4;
}
else if (Rs >= 4258) {
return 4;
}
}
}
else if (AFD >= 12) {
if (diffTSC < -2) {
return 4;
}
else if (diffTSC >= -2) {
if (Cs < 501) {
return 2;
}
else if (Cs >= 501) {
return 8;
}
}
}
}
}
else if (negBalance >= 53) {
return 8;
}
}
else if (tcor >= 96) {
if (posBalance < 56) {
if (Rs < 3131) {
if (Cs < 557) {
if (Cs < 517) {
return 8;
}
else if (Cs >= 517) {
return 2;
}
}
else if (Cs >= 557) {
return 8;
}
}
else if (Rs >= 3131) {
return 2;
}
}
else if (posBalance >= 56) {
if (Cs < 1126) {
return 2;
}
else if (Cs >= 1126) {
return 8;
}
}
}
}
else if (RsDiff >= 11) {
if (negBalance < 51) {
if (RsDiff < 13) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
if (Cs < 741) {
return 2;
}
else if (Cs >= 741) {
return 4;
}
}
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (negBalance >= 51) {
if (Cs < 826) {
if (RsCsDiff < 147) {
if (Rs < 2579) {
return 1;
}
else if (Rs >= 2579) {
return 2;
}
}
else if (RsCsDiff >= 147) {
return 2;
}
}
else if (Cs >= 826) {
if (posBalance < 56) {
return 2;
}
else if (posBalance >= 56) {
if (TSC < 14) {
if (refDCval < 102) {
if (refDCval < 99) {
if (negBalance < 52) {
return 2;
}
else if (negBalance >= 52) {
if (Cs < 1211) {
return 4;
}
else if (Cs >= 1211) {
return 2;
}
}
}
else if (refDCval >= 99) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
if (Rs < 2727) {
return 2;
}
else if (Rs >= 2727) {
return 4;
}
}
}
}
else if (refDCval >= 102) {
if (RsCsDiff < 287) {
return 4;
}
else if (RsCsDiff >= 287) {
if (Cs < 1152) {
return 2;
}
else if (Cs >= 1152) {
return 4;
}
}
}
}
else if (TSC >= 14) {
return 1;
}
}
}
}
}
}
}
}
else if (refDCval >= 109) {
if (Rs < 3365) {
if (mcTcor < 98) {
if (RsDiff < 8) {
if (tcor < 92) {
if (posBalance < 40) {
if (Rs < 2397) {
return 4;
}
else if (Rs >= 2397) {
return 8;
}
}
else if (posBalance >= 40) {
return 8;
}
}
else if (tcor >= 92) {
if (Rs < 2725) {
if (RsCsDiff < 36) {
if (ssDCval < 134) {
return 8;
}
else if (ssDCval >= 134) {
if (refDCval < 137) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
return 1;
}
}
else if (refDCval >= 137) {
return 8;
}
}
}
else if (RsCsDiff >= 36) {
if (negBalance < 65) {
return 1;
}
else if (negBalance >= 65) {
return 8;
}
}
}
else if (Rs >= 2725) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
if (Rs < 2757) {
return 2;
}
else if (Rs >= 2757) {
return 8;
}
}
}
}
}
else if (RsDiff >= 8) {
if (Cs < 969) {
if (refDCval < 136) {
if (tcor < 83) {
if (TSC < 6) {
return 8;
}
else if (TSC >= 6) {
return 2;
}
}
else if (tcor >= 83) {
if (diffMVdiffVal < -6) {
return 1;
}
else if (diffMVdiffVal >= -6) {
if (tcor < 89) {
if (Cs < 787) {
return 1;
}
else if (Cs >= 787) {
return 4;
}
}
else if (tcor >= 89) {
if (Rs < 2514) {
return 1;
}
else if (Rs >= 2514) {
return 2;
}
}
}
}
}
else if (refDCval >= 136) {
if (diffTSC < 2) {
return 1;
}
else if (diffTSC >= 2) {
if (Cs < 564) {
return 1;
}
else if (Cs >= 564) {
return 2;
}
}
}
}
else if (Cs >= 969) {
if (AFD < 13) {
if (posBalance < 57) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 2;
}
}
else if (posBalance >= 57) {
return 8;
}
}
else if (AFD >= 13) {
if (diffTSC < 1) {
return 1;
}
else if (diffTSC >= 1) {
if (Rs < 2861) {
return 8;
}
else if (Rs >= 2861) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 98) {
if (SC < 3190) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
if (AFD < 4) {
return 1;
}
else if (AFD >= 4) {
return 2;
}
}
}
else if (SC >= 3190) {
if (refDCval < 121) {
if (negBalance < 47) {
if (Rs < 3233) {
if (Cs < 915) {
if (Cs < 530) {
return 8;
}
else if (Cs >= 530) {
return 2;
}
}
else if (Cs >= 915) {
return 8;
}
}
else if (Rs >= 3233) {
if (Cs < 574) {
return 8;
}
else if (Cs >= 574) {
return 2;
}
}
}
else if (negBalance >= 47) {
if (AFD < 14) {
return 2;
}
else if (AFD >= 14) {
if (Rs < 3059) {
return 2;
}
else if (Rs >= 3059) {
return 4;
}
}
}
}
else if (refDCval >= 121) {
if (MVDiff < 9) {
return 2;
}
else if (MVDiff >= 9) {
if (Rs < 3236) {
if (diffRsCsdiff < -45) {
return 1;
}
else if (diffRsCsdiff >= -45) {
if (Cs < 459) {
return 8;
}
else if (Cs >= 459) {
return 2;
}
}
}
else if (Rs >= 3236) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3365) {
if (mcTcor < 94) {
if (ssDCval < 134) {
return 2;
}
else if (ssDCval >= 134) {
if (AFD < 17) {
if (Cs < 1141) {
return 4;
}
else if (Cs >= 1141) {
return 1;
}
}
else if (AFD >= 17) {
if (Rs < 3840) {
return 4;
}
else if (Rs >= 3840) {
return 2;
}
}
}
}
else if (mcTcor >= 94) {
if (CsDiff < 3) {
if (ssDCval < 145) {
if (refDCval < 110) {
if (SC < 4377) {
return 1;
}
else if (SC >= 4377) {
return 2;
}
}
else if (refDCval >= 110) {
return 2;
}
}
else if (ssDCval >= 145) {
if (Rs < 3830) {
if (negBalance < 26) {
return 1;
}
else if (negBalance >= 26) {
return 2;
}
}
else if (Rs >= 3830) {
return 2;
}
}
}
else if (CsDiff >= 3) {
if (refDCval < 134) {
if (Rs < 3889) {
if (MVDiff < 11) {
if (diffMVdiffVal < -34) {
return 1;
}
else if (diffMVdiffVal >= -34) {
if (mcTcor < 96) {
if (Cs < 842) {
return 2;
}
else if (Cs >= 842) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Cs < 1223) {
if (TSC < 4) {
if (Rs < 3543) {
return 1;
}
else if (Rs >= 3543) {
if (Rs < 3617) {
return 2;
}
else if (Rs >= 3617) {
return 8;
}
}
}
else if (TSC >= 4) {
return 2;
}
}
else if (Cs >= 1223) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 11) {
if (TSC < 7) {
if (TSC < 4) {
return 1;
}
else if (TSC >= 4) {
return 8;
}
}
else if (TSC >= 7) {
return 2;
}
}
}
else if (Rs >= 3889) {
if (MVDiff < 11) {
return 2;
}
else if (MVDiff >= 11) {
if (Cs < 1223) {
if (mu_mv_mag_sq < 2) {
if (ssDCval < 132) {
if (gchDC < 2) {
return 2;
}
else if (gchDC >= 2) {
if (Rs < 4635) {
return 2;
}
else if (Rs >= 4635) {
return 1;
}
}
}
else if (ssDCval >= 132) {
return 1;
}
}
else if (mu_mv_mag_sq >= 2) {
return 2;
}
}
else if (Cs >= 1223) {
return 4;
}
}
}
}
else if (refDCval >= 134) {
if (mcTcor < 98) {
if (RsDiff < 16) {
if (SC < 4545) {
if (Rs < 3780) {
return 2;
}
else if (Rs >= 3780) {
return 8;
}
}
else if (SC >= 4545) {
if (negBalance < 55) {
if (CsDiff < 4) {
if (Rs < 5683) {
if (mu_mv_mag_sq < 64) {
return 2;
}
else if (mu_mv_mag_sq >= 64) {
if (Rs < 4309) {
if (Cs < 966) {
return 2;
}
else if (Cs >= 966) {
return 1;
}
}
else if (Rs >= 4309) {
return 2;
}
}
}
else if (Rs >= 5683) {
return 1;
}
}
else if (CsDiff >= 4) {
if (negBalance < 32) {
return 2;
}
else if (negBalance >= 32) {
return 1;
}
}
}
else if (negBalance >= 55) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 4;
}
}
}
}
else if (RsDiff >= 16) {
if (refDCval < 135) {
if (Rs < 5233) {
return 2;
}
else if (Rs >= 5233) {
return 1;
}
}
else if (refDCval >= 135) {
return 2;
}
}
}
else if (mcTcor >= 98) {
if (Rs < 3887) {
if (Cs < 1135) {
if (Cs < 562) {
return 8;
}
else if (Cs >= 562) {
return 2;
}
}
else if (Cs >= 1135) {
if (Rs < 3560) {
return 2;
}
else if (Rs >= 3560) {
return 1;
}
}
}
else if (Rs >= 3887) {
return 2;
}
}
}
}
}
}
}
}
else if (MVDiff >= 17) {
if (mcTcor < 68) {
if (MVDiff < 43) {
return 1;
}
else if (MVDiff >= 43) {
if (gchDC < 7) {
if (AFD < 30) {
return 4;
}
else if (AFD >= 30) {
return 2;
}
}
else if (gchDC >= 7) {
return 8;
}
}
}
else if (mcTcor >= 68) {
if (ssDCval < 132) {
if (ssDCval < 105) {
if (RsCsDiff < 183) {
if (refDCval < 77) {
if (CsDiff < 3) {
if (Rs < 2856) {
return 8;
}
else if (Rs >= 2856) {
if (Cs < 873) {
return 4;
}
else if (Cs >= 873) {
return 8;
}
}
}
else if (CsDiff >= 3) {
if (mcTcor < 86) {
if (negBalance < 41) {
if (Rs < 2618) {
return 8;
}
else if (Rs >= 2618) {
return 2;
}
}
else if (negBalance >= 41) {
return 1;
}
}
else if (mcTcor >= 86) {
if (posBalance < 48) {
if (diffRsCsdiff < -15) {
if (gchDC < 5) {
return 4;
}
else if (gchDC >= 5) {
return 8;
}
}
else if (diffRsCsdiff >= -15) {
if (CsDiff < 5) {
if (Rs < 2428) {
return 1;
}
else if (Rs >= 2428) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
if (Cs < 417) {
return 8;
}
else if (Cs >= 417) {
return 4;
}
}
}
}
else if (CsDiff >= 5) {
if (Cs < 1252) {
if (Rs < 2406) {
return 1;
}
else if (Rs >= 2406) {
return 2;
}
}
else if (Cs >= 1252) {
return 4;
}
}
}
}
else if (posBalance >= 48) {
if (Rs < 2446) {
if (diffTSC < 1) {
if (Rs < 2400) {
return 4;
}
else if (Rs >= 2400) {
return 1;
}
}
else if (diffTSC >= 1) {
if (Cs < 1103) {
return 2;
}
else if (Cs >= 1103) {
return 8;
}
}
}
else if (Rs >= 2446) {
if (ssDCval < 56) {
return 4;
}
else if (ssDCval >= 56) {
if (posBalance < 54) {
if (tcor < 82) {
if (CsDiff < 6) {
return 4;
}
else if (CsDiff >= 6) {
return 2;
}
}
else if (tcor >= 82) {
if (Cs < 548) {
return 4;
}
else if (Cs >= 548) {
return 2;
}
}
}
else if (posBalance >= 54) {
return 2;
}
}
}
}
}
}
}
else if (refDCval >= 77) {
if (Cs < 1115) {
if (MVDiff < 30) {
if (RsDiff < 9) {
if (refDCval < 98) {
if (diffRsCsdiff < -49) {
return 2;
}
else if (diffRsCsdiff >= -49) {
if (RsCsDiff < 12) {
return 8;
}
else if (RsCsDiff >= 12) {
if (diffRsCsdiff < 3) {
return 1;
}
else if (diffRsCsdiff >= 3) {
if (AFD < 9) {
return 1;
}
else if (AFD >= 9) {
return 8;
}
}
}
}
}
else if (refDCval >= 98) {
return 4;
}
}
else if (RsDiff >= 9) {
if (Rs < 2400) {
return 1;
}
else if (Rs >= 2400) {
if (Rs < 3765) {
if (TSC < 13) {
return 2;
}
else if (TSC >= 13) {
return 1;
}
}
else if (Rs >= 3765) {
if (Rs < 4142) {
return 8;
}
else if (Rs >= 4142) {
if (Rs < 5231) {
return 4;
}
else if (Rs >= 5231) {
return 2;
}
}
}
}
}
}
else if (MVDiff >= 30) {
if (Cs < 383) {
if (negBalance < 48) {
if (Rs < 2405) {
return 1;
}
else if (Rs >= 2405) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 2;
}
}
}
else if (negBalance >= 48) {
if (Rs < 2656) {
return 2;
}
else if (Rs >= 2656) {
return 4;
}
}
}
else if (Cs >= 383) {
if (diffMVdiffVal < 31) {
if (Rs < 4104) {
if (AFD < 22) {
if (RsCsDiff < 94) {
if (negBalance < 49) {
return 1;
}
else if (negBalance >= 49) {
if (Rs < 2973) {
if (AFD < 15) {
if (mu_mv_mag_sq < 28) {
return 1;
}
else if (mu_mv_mag_sq >= 28) {
return 8;
}
}
else if (AFD >= 15) {
return 8;
}
}
else if (Rs >= 2973) {
return 8;
}
}
}
else if (RsCsDiff >= 94) {
if (ssDCval < 88) {
if (TSC < 9) {
return 1;
}
else if (TSC >= 9) {
return 8;
}
}
else if (ssDCval >= 88) {
if (diffRsCsdiff < 50) {
return 8;
}
else if (diffRsCsdiff >= 50) {
if (Cs < 713) {
return 8;
}
else if (Cs >= 713) {
return 1;
}
}
}
}
}
else if (AFD >= 22) {
if (SC < 3607) {
return 1;
}
else if (SC >= 3607) {
return 2;
}
}
}
else if (Rs >= 4104) {
if (Rs < 5666) {
return 4;
}
else if (Rs >= 5666) {
return 2;
}
}
}
else if (diffMVdiffVal >= 31) {
if (ssDCval < 91) {
return 8;
}
else if (ssDCval >= 91) {
return 2;
}
}
}
}
}
else if (Cs >= 1115) {
if (diffRsCsdiff < -97) {
if (diffTSC < -1) {
if (tcor < 93) {
return 8;
}
else if (tcor >= 93) {
return 2;
}
}
else if (diffTSC >= -1) {
return 2;
}
}
else if (diffRsCsdiff >= -97) {
if (CsDiff < 3) {
return 4;
}
else if (CsDiff >= 3) {
if (Cs < 1191) {
return 8;
}
else if (Cs >= 1191) {
if (CsDiff < 6) {
if (MVDiff < 58) {
return 8;
}
else if (MVDiff >= 58) {
return 4;
}
}
else if (CsDiff >= 6) {
if (MVDiff < 41) {
return 4;
}
else if (MVDiff >= 41) {
return 8;
}
}
}
}
}
}
}
}
else if (RsCsDiff >= 183) {
if (ssDCval < 73) {
if (Rs < 3290) {
if (AFD < 22) {
if (posBalance < 48) {
if (Rs < 2695) {
if (Rs < 2395) {
return 8;
}
else if (Rs >= 2395) {
return 2;
}
}
else if (Rs >= 2695) {
if (ssDCval < 58) {
return 2;
}
else if (ssDCval >= 58) {
return 4;
}
}
}
else if (posBalance >= 48) {
if (diffAFD < 6) {
if (MVDiff < 82) {
return 2;
}
else if (MVDiff >= 82) {
return 4;
}
}
else if (diffAFD >= 6) {
return 4;
}
}
}
else if (AFD >= 22) {
if (AFD < 26) {
if (tcor < 51) {
if (Rs < 2676) {
return 2;
}
else if (Rs >= 2676) {
if (Rs < 2960) {
return 4;
}
else if (Rs >= 2960) {
return 1;
}
}
}
else if (tcor >= 51) {
return 1;
}
}
else if (AFD >= 26) {
if (negBalance < 55) {
return 4;
}
else if (negBalance >= 55) {
if (Cs < 845) {
return 2;
}
else if (Cs >= 845) {
return 4;
}
}
}
}
}
else if (Rs >= 3290) {
if (gchDC < 3) {
return 4;
}
else if (gchDC >= 3) {
if (Rs < 4880) {
return 1;
}
else if (Rs >= 4880) {
return 8;
}
}
}
}
else if (ssDCval >= 73) {
if (mcTcor < 91) {
if (RsCsDiff < 357) {
if (ssDCval < 83) {
if (MVDiff < 25) {
if (posBalance < 52) {
if (Rs < 2603) {
return 2;
}
else if (Rs >= 2603) {
if (Rs < 3039) {
return 8;
}
else if (Rs >= 3039) {
return 4;
}
}
}
else if (posBalance >= 52) {
return 2;
}
}
else if (MVDiff >= 25) {
if (Rs < 2890) {
return 1;
}
else if (Rs >= 2890) {
return 2;
}
}
}
else if (ssDCval >= 83) {
if (TSC < 13) {
if (CsDiff < 3) {
if (Rs < 2480) {
return 2;
}
else if (Rs >= 2480) {
return 4;
}
}
else if (CsDiff >= 3) {
if (MVDiff < 22) {
if (Cs < 603) {
return 2;
}
else if (Cs >= 603) {
return 4;
}
}
else if (MVDiff >= 22) {
if (RsCsDiff < 278) {
if (Cs < 634) {
if (Rs < 3192) {
return 2;
}
else if (Rs >= 3192) {
return 8;
}
}
else if (Cs >= 634) {
return 8;
}
}
else if (RsCsDiff >= 278) {
return 2;
}
}
}
}
else if (TSC >= 13) {
if (Rs < 2810) {
if (Cs < 1054) {
return 1;
}
else if (Cs >= 1054) {
return 2;
}
}
else if (Rs >= 2810) {
if (AFD < 20) {
return 1;
}
else if (AFD >= 20) {
return 4;
}
}
}
}
}
else if (RsCsDiff >= 357) {
if (Cs < 932) {
if (TSC < 16) {
if (AFD < 17) {
return 4;
}
else if (AFD >= 17) {
return 2;
}
}
else if (TSC >= 16) {
return 1;
}
}
else if (Cs >= 932) {
if (AFD < 26) {
if (MVDiff < 70) {
return 8;
}
else if (MVDiff >= 70) {
if (Rs < 3253) {
return 4;
}
else if (Rs >= 3253) {
return 2;
}
}
}
else if (AFD >= 26) {
if (MVDiff < 54) {
return 2;
}
else if (MVDiff >= 54) {
if (RsCsDiff < 597) {
return 4;
}
else if (RsCsDiff >= 597) {
if (Rs < 3060) {
return 8;
}
else if (Rs >= 3060) {
return 2;
}
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (Rs < 3787) {
if (ssDCval < 92) {
if (AFD < 12) {
if (Rs < 2740) {
return 1;
}
else if (Rs >= 2740) {
return 2;
}
}
else if (AFD >= 12) {
if (mcTcor < 97) {
return 2;
}
else if (mcTcor >= 97) {
if (ssDCval < 79) {
if (Rs < 3260) {
return 4;
}
else if (Rs >= 3260) {
return 1;
}
}
else if (ssDCval >= 79) {
return 2;
}
}
}
}
else if (ssDCval >= 92) {
if (gchDC < 5) {
if (RsDiff < 17) {
if (diffRsCsdiff < 19) {
if (mcTcor < 92) {
return 4;
}
else if (mcTcor >= 92) {
return 2;
}
}
else if (diffRsCsdiff >= 19) {
if (CsDiff < 4) {
return 8;
}
else if (CsDiff >= 4) {
if (refDCval < 97) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
else if (refDCval >= 97) {
return 2;
}
}
}
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (gchDC >= 5) {
if (mcTcor < 96) {
if (Cs < 785) {
return 2;
}
else if (Cs >= 785) {
return 8;
}
}
else if (mcTcor >= 96) {
return 1;
}
}
}
}
else if (Rs >= 3787) {
if (ssDCval < 83) {
if (SC < 5881) {
if (posBalance < 56) {
return 4;
}
else if (posBalance >= 56) {
if (SC < 4582) {
return 4;
}
else if (SC >= 4582) {
return 2;
}
}
}
else if (SC >= 5881) {
return 1;
}
}
else if (ssDCval >= 83) {
if (Cs < 617) {
return 2;
}
else if (Cs >= 617) {
if (gchDC < 3) {
if (refDCval < 101) {
if (SC < 4769) {
return 8;
}
else if (SC >= 4769) {
return 2;
}
}
else if (refDCval >= 101) {
return 8;
}
}
else if (gchDC >= 3) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
return 1;
}
}
}
}
}
}
}
}
}
else if (ssDCval >= 105) {
if (Rs < 3113) {
if (tcor < 78) {
if (AFD < 26) {
if (CsDiff < 6) {
if (RsDiff < 17) {
if (diffAFD < 4) {
if (diffRsCsdiff < -53) {
if (Rs < 2956) {
return 2;
}
else if (Rs >= 2956) {
return 8;
}
}
else if (diffRsCsdiff >= -53) {
if (Rs < 2514) {
if (MVDiff < 43) {
return 8;
}
else if (MVDiff >= 43) {
if (Rs < 2489) {
return 2;
}
else if (Rs >= 2489) {
return 1;
}
}
}
else if (Rs >= 2514) {
return 8;
}
}
}
else if (diffAFD >= 4) {
return 1;
}
}
else if (RsDiff >= 17) {
return 2;
}
}
else if (CsDiff >= 6) {
if (TSC < 11) {
if (Rs < 2862) {
return 1;
}
else if (Rs >= 2862) {
return 8;
}
}
else if (TSC >= 11) {
return 4;
}
}
}
else if (AFD >= 26) {
if (gchDC < 3) {
if (Cs < 1012) {
if (mu_mv_mag_sq < 6) {
return 1;
}
else if (mu_mv_mag_sq >= 6) {
return 2;
}
}
else if (Cs >= 1012) {
return 1;
}
}
else if (gchDC >= 3) {
if (negBalance < 44) {
if (Rs < 3001) {
return 8;
}
else if (Rs >= 3001) {
return 1;
}
}
else if (negBalance >= 44) {
if (RsDiff < 23) {
if (RsCsDiff < 255) {
return 2;
}
else if (RsCsDiff >= 255) {
return 1;
}
}
else if (RsDiff >= 23) {
return 2;
}
}
}
}
}
else if (tcor >= 78) {
if (mu_mv_mag_sq < 27) {
if (diffAFD < -4) {
if (RsDiff < 17) {
return 1;
}
else if (RsDiff >= 17) {
if (Rs < 2889) {
return 2;
}
else if (Rs >= 2889) {
return 1;
}
}
}
else if (diffAFD >= -4) {
if (mcTcor < 93) {
if (SC < 3127) {
if (mcTcor < 92) {
return 1;
}
else if (mcTcor >= 92) {
return 8;
}
}
else if (SC >= 3127) {
if (TSC < 11) {
return 4;
}
else if (TSC >= 11) {
return 8;
}
}
}
else if (mcTcor >= 93) {
if (ssDCval < 121) {
if (posBalance < 66) {
if (MVDiff < 61) {
if (gchDC < 2) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 1;
}
}
else if (gchDC >= 2) {
if (CsDiff < 7) {
return 1;
}
else if (CsDiff >= 7) {
return 2;
}
}
}
else if (MVDiff >= 61) {
return 8;
}
}
else if (posBalance >= 66) {
return 1;
}
}
else if (ssDCval >= 121) {
if (CsDiff < 3) {
if (Rs < 2916) {
return 1;
}
else if (Rs >= 2916) {
return 2;
}
}
else if (CsDiff >= 3) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (SC < 3385) {
if (Cs < 348) {
if (negBalance < 22) {
return 1;
}
else if (negBalance >= 22) {
return 8;
}
}
else if (Cs >= 348) {
if (negBalance < 38) {
return 2;
}
else if (negBalance >= 38) {
if (MVDiff < 23) {
return 2;
}
else if (MVDiff >= 23) {
return 1;
}
}
}
}
else if (SC >= 3385) {
if (diffAFD < -11) {
if (Rs < 2482) {
return 8;
}
else if (Rs >= 2482) {
return 1;
}
}
else if (diffAFD >= -11) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
return 1;
}
}
}
}
}
}
else if (Rs >= 3113) {
if (RsCsDiff < 421) {
if (negBalance < 39) {
if (Rs < 4022) {
if (ssDCval < 127) {
return 8;
}
else if (ssDCval >= 127) {
return 2;
}
}
else if (Rs >= 4022) {
if (mu_mv_mag_sq < 3) {
if (Rs < 4876) {
return 2;
}
else if (Rs >= 4876) {
return 1;
}
}
else if (mu_mv_mag_sq >= 3) {
return 2;
}
}
}
else if (negBalance >= 39) {
if (CsDiff < 4) {
if (ssDCval < 123) {
if (Cs < 525) {
if (Rs < 3422) {
return 8;
}
else if (Rs >= 3422) {
if (SC < 4129) {
if (AFD < 19) {
return 4;
}
else if (AFD >= 19) {
if (Rs < 3491) {
return 2;
}
else if (Rs >= 3491) {
return 8;
}
}
}
else if (SC >= 4129) {
return 2;
}
}
}
else if (Cs >= 525) {
return 8;
}
}
else if (ssDCval >= 123) {
if (MVDiff < 31) {
return 2;
}
else if (MVDiff >= 31) {
if (Rs < 3409) {
return 2;
}
else if (Rs >= 3409) {
return 8;
}
}
}
}
else if (CsDiff >= 4) {
if (refDCval < 127) {
if (diffRsCsdiff < -151) {
return 2;
}
else if (diffRsCsdiff >= -151) {
if (ssDCval < 109) {
if (Rs < 3340) {
return 8;
}
else if (Rs >= 3340) {
if (Rs < 4785) {
return 2;
}
else if (Rs >= 4785) {
if (Rs < 5644) {
return 8;
}
else if (Rs >= 5644) {
return 2;
}
}
}
}
else if (ssDCval >= 109) {
if (Cs < 896) {
if (RsCsDiff < 251) {
if (diffMVdiffVal < -2) {
return 1;
}
else if (diffMVdiffVal >= -2) {
return 8;
}
}
else if (RsCsDiff >= 251) {
if (TSC < 12) {
return 2;
}
else if (TSC >= 12) {
return 8;
}
}
}
else if (Cs >= 896) {
if (mu_mv_mag_sq < 7) {
if (Rs < 3587) {
return 1;
}
else if (Rs >= 3587) {
return 8;
}
}
else if (mu_mv_mag_sq >= 7) {
if (Rs < 3430) {
return 8;
}
else if (Rs >= 3430) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 127) {
if (Rs < 5349) {
return 1;
}
else if (Rs >= 5349) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 421) {
if (Cs < 868) {
if (refDCval < 128) {
return 2;
}
else if (refDCval >= 128) {
if (Rs < 5532) {
if (MVDiff < 41) {
return 8;
}
else if (MVDiff >= 41) {
if (Rs < 3871) {
return 1;
}
else if (Rs >= 3871) {
return 2;
}
}
}
else if (Rs >= 5532) {
return 2;
}
}
}
else if (Cs >= 868) {
if (ssDCval < 121) {
if (diffAFD < -9) {
return 1;
}
else if (diffAFD >= -9) {
if (MVDiff < 103) {
if (RsCsDiff < 495) {
if (posBalance < 60) {
if (RsDiff < 21) {
return 2;
}
else if (RsDiff >= 21) {
return 8;
}
}
else if (posBalance >= 60) {
return 2;
}
}
else if (RsCsDiff >= 495) {
return 2;
}
}
else if (MVDiff >= 103) {
if (Rs < 3503) {
return 8;
}
else if (Rs >= 3503) {
return 4;
}
}
}
}
else if (ssDCval >= 121) {
if (negBalance < 59) {
if (RsCsDiff < 799) {
if (MVDiff < 41) {
return 8;
}
else if (MVDiff >= 41) {
if (refDCval < 124) {
return 8;
}
else if (refDCval >= 124) {
if (diffMVdiffVal < 15) {
return 2;
}
else if (diffMVdiffVal >= 15) {
if (MVDiff < 68) {
return 8;
}
else if (MVDiff >= 68) {
return 2;
}
}
}
}
}
else if (RsCsDiff >= 799) {
if (Cs < 1015) {
return 8;
}
else if (Cs >= 1015) {
return 2;
}
}
}
else if (negBalance >= 59) {
if (RsDiff < 22) {
if (AFD < 24) {
return 1;
}
else if (AFD >= 24) {
return 8;
}
}
else if (RsDiff >= 22) {
return 1;
}
}
}
}
}
}
}
}
else if (ssDCval >= 132) {
if (mcTcor < 92) {
if (AFD < 19) {
if (ssDCval < 147) {
if (tcor < 57) {
return 1;
}
else if (tcor >= 57) {
return 4;
}
}
else if (ssDCval >= 147) {
if (CsDiff < 4) {
if (Rs < 4497) {
return 1;
}
else if (Rs >= 4497) {
return 2;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
}
else if (AFD >= 19) {
if (posBalance < 53) {
if (diffAFD < -2) {
return 2;
}
else if (diffAFD >= -2) {
if (MVDiff < 61) {
return 1;
}
else if (MVDiff >= 61) {
return 2;
}
}
}
else if (posBalance >= 53) {
if (MVDiff < 66) {
if (ssDCval < 141) {
if (Rs < 2739) {
return 1;
}
else if (Rs >= 2739) {
return 8;
}
}
else if (ssDCval >= 141) {
return 1;
}
}
else if (MVDiff >= 66) {
if (AFD < 22) {
return 1;
}
else if (AFD >= 22) {
return 8;
}
}
}
}
}
else if (mcTcor >= 92) {
if (SC < 3584) {
if (mu_mv_mag_sq < 54) {
if (MVDiff < 154) {
return 1;
}
else if (MVDiff >= 154) {
if (MVDiff < 183) {
if (Rs < 2593) {
return 2;
}
else if (Rs >= 2593) {
return 8;
}
}
else if (MVDiff >= 183) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (Rs < 2512) {
return 2;
}
else if (Rs >= 2512) {
if (Rs < 2669) {
return 1;
}
else if (Rs >= 2669) {
return 8;
}
}
}
}
else if (SC >= 3584) {
if (RsDiff < 8) {
if (Rs < 3644) {
return 8;
}
else if (Rs >= 3644) {
return 2;
}
}
else if (RsDiff >= 8) {
if (CsDiff < 4) {
if (Cs < 1041) {
if (SC < 3864) {
if (diffRsCsdiff < -88) {
return 1;
}
else if (diffRsCsdiff >= -88) {
if (Rs < 3189) {
return 2;
}
else if (Rs >= 3189) {
return 8;
}
}
}
else if (SC >= 3864) {
if (mcTcor < 93) {
if (Rs < 6558) {
return 4;
}
else if (Rs >= 6558) {
return 2;
}
}
else if (mcTcor >= 93) {
if (MVDiff < 18) {
if (RsDiff < 14) {
return 2;
}
else if (RsDiff >= 14) {
return 4;
}
}
else if (MVDiff >= 18) {
return 2;
}
}
}
}
else if (Cs >= 1041) {
return 1;
}
}
else if (CsDiff >= 4) {
if (mu_mv_mag_sq < 34) {
if (Rs < 4165) {
if (diffMVdiffVal < 43) {
if (posBalance < 64) {
return 1;
}
else if (posBalance >= 64) {
if (MVDiff < 36) {
return 1;
}
else if (MVDiff >= 36) {
if (AFD < 24) {
return 8;
}
else if (AFD >= 24) {
return 1;
}
}
}
}
else if (diffMVdiffVal >= 43) {
return 8;
}
}
else if (Rs >= 4165) {
if (tcor < 75) {
return 8;
}
else if (tcor >= 75) {
if (negBalance < 50) {
if (Rs < 6405) {
return 1;
}
else if (Rs >= 6405) {
return 2;
}
}
else if (negBalance >= 50) {
if (CsDiff < 7) {
if (Rs < 4914) {
if (Rs < 4569) {
return 2;
}
else if (Rs >= 4569) {
return 1;
}
}
else if (Rs >= 4914) {
if (tcor < 79) {
if (Rs < 5619) {
return 2;
}
else if (Rs >= 5619) {
return 8;
}
}
else if (tcor >= 79) {
if (negBalance < 51) {
if (Rs < 6171) {
return 8;
}
else if (Rs >= 6171) {
return 2;
}
}
else if (negBalance >= 51) {
return 2;
}
}
}
}
else if (CsDiff >= 7) {
if (Cs < 676) {
return 2;
}
else if (Cs >= 676) {
return 1;
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 34) {
if (RsDiff < 14) {
if (Cs < 1198) {
return 1;
}
else if (Cs >= 1198) {
return 2;
}
}
else if (RsDiff >= 14) {
if (diffAFD < -4) {
return 1;
}
else if (diffAFD >= -4) {
return 2;
}
}
}
}
}
}
}
}
}
}
}
}
else if (Cs >= 1307) {
if (mu_mv_mag_sq < 111) {
if (RsCsDiff < 101) {
if (refDCval < 105) {
if (negBalance < 39) {
if (Rs < 743) {
if (Rs < 301) {
return 4;
}
else if (Rs >= 301) {
if (TSC < 5) {
if (ssDCval < 71) {
return 1;
}
else if (ssDCval >= 71) {
return 8;
}
}
else if (TSC >= 5) {
if (diffAFD < -1) {
return 8;
}
else if (diffAFD >= -1) {
if (Rs < 698) {
return 2;
}
else if (Rs >= 698) {
return 1;
}
}
}
}
}
else if (Rs >= 743) {
if (SC < 10031) {
if (posBalance < 45) {
if (ssDCval < 78) {
if (ssDCval < 64) {
if (Rs < 3382) {
if (mu_mv_mag_sq < 97) {
if (tcor < 96) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
if (posBalance < 43) {
return 2;
}
else if (posBalance >= 43) {
if (Rs < 2313) {
return 4;
}
else if (Rs >= 2313) {
return 8;
}
}
}
}
else if (tcor >= 96) {
if (Rs < 1245) {
return 1;
}
else if (Rs >= 1245) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 97) {
if (SC < 3147) {
if (Rs < 949) {
return 8;
}
else if (Rs >= 949) {
if (Rs < 1442) {
return 4;
}
else if (Rs >= 1442) {
return 1;
}
}
}
else if (SC >= 3147) {
return 8;
}
}
}
else if (Rs >= 3382) {
if (mu_mv_mag_sq < 84) {
return 4;
}
else if (mu_mv_mag_sq >= 84) {
if (diffAFD < 1) {
if (Cs < 2364) {
return 4;
}
else if (Cs >= 2364) {
if (Rs < 4592) {
return 8;
}
else if (Rs >= 4592) {
if (Rs < 4716) {
return 8;
}
else if (Rs >= 4716) {
if (Rs < 4768) {
return 4;
}
else if (Rs >= 4768) {
return 4;
}
}
}
}
}
else if (diffAFD >= 1) {
return 8;
}
}
}
}
else if (ssDCval >= 64) {
if (MVDiff < 7) {
if (Rs < 3467) {
if (RsDiff < 6) {
if (Rs < 2688) {
return 8;
}
else if (Rs >= 2688) {
return 4;
}
}
else if (RsDiff >= 6) {
if (Cs < 1861) {
return 2;
}
else if (Cs >= 1861) {
return 1;
}
}
}
else if (Rs >= 3467) {
if (Rs < 4374) {
return 4;
}
else if (Rs >= 4374) {
if (Rs < 4742) {
return 8;
}
else if (Rs >= 4742) {
return 4;
}
}
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
else if (ssDCval >= 78) {
if (MVDiff < 9) {
if (Rs < 3038) {
if (SC < 3539) {
return 8;
}
else if (SC >= 3539) {
if (RsDiff < 1) {
if (Rs < 2402) {
if (Rs < 2339) {
if (Rs < 2327) {
return 2;
}
else if (Rs >= 2327) {
return 1;
}
}
else if (Rs >= 2339) {
return 4;
}
}
else if (Rs >= 2402) {
if (ssDCval < 91) {
if (Cs < 2174) {
return 4;
}
else if (Cs >= 2174) {
return 1;
}
}
else if (ssDCval >= 91) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (AFD < 6) {
if (Cs < 2099) {
return 8;
}
else if (Cs >= 2099) {
if (Rs < 2643) {
return 1;
}
else if (Rs >= 2643) {
return 8;
}
}
}
else if (AFD >= 6) {
if (Cs < 2722) {
return 1;
}
else if (Cs >= 2722) {
return 4;
}
}
}
}
}
else if (Rs >= 3038) {
if (RsCsDiff < 55) {
if (ssDCval < 90) {
if (refDCval < 84) {
if (Cs < 2019) {
if (Rs < 4236) {
if (ssDCval < 81) {
if (SC < 4942) {
return 2;
}
else if (SC >= 4942) {
return 4;
}
}
else if (ssDCval >= 81) {
return 2;
}
}
else if (Rs >= 4236) {
return 8;
}
}
else if (Cs >= 2019) {
return 4;
}
}
else if (refDCval >= 84) {
return 4;
}
}
else if (ssDCval >= 90) {
if (ssDCval < 97) {
if (MVDiff < 3) {
if (Rs < 4099) {
return 2;
}
else if (Rs >= 4099) {
if (negBalance < 29) {
return 2;
}
else if (negBalance >= 29) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
if (posBalance < 39) {
if (Rs < 4538) {
return 2;
}
else if (Rs >= 4538) {
return 8;
}
}
else if (posBalance >= 39) {
return 8;
}
}
}
}
}
else if (MVDiff >= 3) {
if (Rs < 4119) {
return 8;
}
else if (Rs >= 4119) {
return 4;
}
}
}
else if (ssDCval >= 97) {
if (MVDiff < 4) {
if (Cs < 2010) {
return 4;
}
else if (Cs >= 2010) {
return 8;
}
}
else if (MVDiff >= 4) {
return 8;
}
}
}
}
else if (RsCsDiff >= 55) {
if (mu_mv_mag_sq < 58) {
if (Rs < 4554) {
return 4;
}
else if (Rs >= 4554) {
return 1;
}
}
else if (mu_mv_mag_sq >= 58) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
if (Rs < 4323) {
return 2;
}
else if (Rs >= 4323) {
if (Rs < 4589) {
return 8;
}
else if (Rs >= 4589) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 9) {
if (diffAFD < -2) {
if (posBalance < 17) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 1;
}
}
else if (posBalance >= 17) {
if (Rs < 2043) {
return 8;
}
else if (Rs >= 2043) {
return 4;
}
}
}
else if (diffAFD >= -2) {
if (ssDCval < 80) {
return 1;
}
else if (ssDCval >= 80) {
if (Rs < 4395) {
return 8;
}
else if (Rs >= 4395) {
return 4;
}
}
}
}
}
}
else if (posBalance >= 45) {
if (Rs < 2949) {
if (SC < 5072) {
if (mu_mv_mag_sq < 89) {
if (Rs < 1521) {
if (TSC < 6) {
if (SC < 2641) {
if (TSC < 4) {
if (SC < 2257) {
return 4;
}
else if (SC >= 2257) {
if (Rs < 1091) {
return 1;
}
else if (Rs >= 1091) {
return 4;
}
}
}
else if (TSC >= 4) {
if (Cs < 1403) {
if (diffAFD < -1) {
if (Cs < 1370) {
return 2;
}
else if (Cs >= 1370) {
return 4;
}
}
else if (diffAFD >= -1) {
return 4;
}
}
else if (Cs >= 1403) {
if (Rs < 822) {
return 4;
}
else if (Rs >= 822) {
return 2;
}
}
}
}
else if (SC >= 2641) {
if (ssDCval < 69) {
if (MVDiff < 3) {
if (Rs < 1087) {
return 2;
}
else if (Rs >= 1087) {
return 4;
}
}
else if (MVDiff >= 3) {
if (mu_mv_mag_sq < 78) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 78) {
if (ssDCval < 61) {
return 8;
}
else if (ssDCval >= 61) {
return 1;
}
}
}
}
else if (ssDCval >= 69) {
if (diffRsCsdiff < -24) {
if (Rs < 1319) {
return 2;
}
else if (Rs >= 1319) {
return 4;
}
}
else if (diffRsCsdiff >= -24) {
return 8;
}
}
}
}
else if (TSC >= 6) {
if (SC < 3236) {
if (SC < 2206) {
return 2;
}
else if (SC >= 2206) {
if (refDCval < 95) {
if (diffTSC < -1) {
if (Rs < 1290) {
return 8;
}
else if (Rs >= 1290) {
return 4;
}
}
else if (diffTSC >= -1) {
return 8;
}
}
else if (refDCval >= 95) {
if (Rs < 1177) {
return 8;
}
else if (Rs >= 1177) {
return 4;
}
}
}
}
else if (SC >= 3236) {
if (Rs < 1431) {
return 2;
}
else if (Rs >= 1431) {
return 4;
}
}
}
}
else if (Rs >= 1521) {
if (tcor < 93) {
if (Cs < 1375) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (AFD < 13) {
return 2;
}
else if (AFD >= 13) {
return 8;
}
}
}
else if (Cs >= 1375) {
if (tcor < 86) {
if (tcor < 82) {
if (Cs < 1382) {
return 4;
}
else if (Cs >= 1382) {
return 8;
}
}
else if (tcor >= 82) {
if (mu_mv_mag_sq < 78) {
return 8;
}
else if (mu_mv_mag_sq >= 78) {
return 4;
}
}
}
else if (tcor >= 86) {
if (diffMVdiffVal < -13) {
if (Rs < 1990) {
return 8;
}
else if (Rs >= 1990) {
return 2;
}
}
else if (diffMVdiffVal >= -13) {
return 8;
}
}
}
}
else if (tcor >= 93) {
if (diffTSC < -2) {
if (mu_mv_mag_sq < 77) {
return 8;
}
else if (mu_mv_mag_sq >= 77) {
return 2;
}
}
else if (diffTSC >= -2) {
if (Rs < 2919) {
if (ssDCval < 84) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
if (mcTcor < 98) {
if (diffMVdiffVal < -2) {
return 8;
}
else if (diffMVdiffVal >= -2) {
if (gchDC < 1) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
return 1;
}
}
else if (gchDC >= 1) {
return 1;
}
}
}
else if (mcTcor >= 98) {
return 8;
}
}
}
else if (ssDCval >= 84) {
if (TSC < 8) {
return 8;
}
else if (TSC >= 8) {
if (Rs < 2523) {
return 8;
}
else if (Rs >= 2523) {
return 1;
}
}
}
}
else if (Rs >= 2919) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (ssDCval < 87) {
if (Rs < 2657) {
if (MVDiff < 7) {
if (RsDiff < 5) {
if (refDCval < 57) {
if (ssDCval < 55) {
if (mcTcor < 98) {
if (Rs < 1124) {
return 4;
}
else if (Rs >= 1124) {
return 8;
}
}
else if (mcTcor >= 98) {
if (Rs < 1111) {
return 8;
}
else if (Rs >= 1111) {
return 4;
}
}
}
else if (ssDCval >= 55) {
return 4;
}
}
else if (refDCval >= 57) {
if (SC < 4981) {
return 8;
}
else if (SC >= 4981) {
return 4;
}
}
}
else if (RsDiff >= 5) {
if (diffMVdiffVal < -3) {
return 4;
}
else if (diffMVdiffVal >= -3) {
if (SC < 3697) {
return 8;
}
else if (SC >= 3697) {
if (Cs < 1609) {
return 4;
}
else if (Cs >= 1609) {
if (Cs < 2078) {
return 8;
}
else if (Cs >= 2078) {
return 4;
}
}
}
}
}
}
else if (MVDiff >= 7) {
if (Cs < 1671) {
if (Rs < 1830) {
return 4;
}
else if (Rs >= 1830) {
return 8;
}
}
else if (Cs >= 1671) {
if (gchDC < 1) {
return 1;
}
else if (gchDC >= 1) {
if (SC < 4318) {
if (TSC < 3) {
return 4;
}
else if (TSC >= 3) {
return 8;
}
}
else if (SC >= 4318) {
if (Rs < 2403) {
return 4;
}
else if (Rs >= 2403) {
return 8;
}
}
}
}
}
}
else if (Rs >= 2657) {
if (diffAFD < -1) {
if (Cs < 1505) {
return 2;
}
else if (Cs >= 1505) {
return 8;
}
}
else if (diffAFD >= -1) {
if (TSC < 6) {
return 4;
}
else if (TSC >= 6) {
if (MVDiff < 7) {
return 1;
}
else if (MVDiff >= 7) {
return 4;
}
}
}
}
}
else if (ssDCval >= 87) {
if (Cs < 1637) {
if (negBalance < 33) {
if (Rs < 1052) {
return 4;
}
else if (Rs >= 1052) {
return 8;
}
}
else if (negBalance >= 33) {
if (Rs < 2295) {
if (Rs < 1662) {
return 8;
}
else if (Rs >= 1662) {
return 2;
}
}
else if (Rs >= 2295) {
return 8;
}
}
}
else if (Cs >= 1637) {
return 8;
}
}
}
}
else if (SC >= 5072) {
if (negBalance < 29) {
if (Rs < 1769) {
return 1;
}
else if (Rs >= 1769) {
return 8;
}
}
else if (negBalance >= 29) {
if (Cs < 2577) {
return 8;
}
else if (Cs >= 2577) {
return 4;
}
}
}
}
else if (Rs >= 2949) {
if (Cs < 2162) {
if (ssDCval < 98) {
if (diffRsCsdiff < -35) {
if (ssDCval < 63) {
return 8;
}
else if (ssDCval >= 63) {
return 2;
}
}
else if (diffRsCsdiff >= -35) {
if (AFD < 9) {
if (RsCsDiff < 37) {
if (AFD < 7) {
if (SC < 4664) {
if (Rs < 3112) {
return 4;
}
else if (Rs >= 3112) {
return 2;
}
}
else if (SC >= 4664) {
if (ssDCval < 75) {
if (diffRsCsdiff < -6) {
if (Rs < 3579) {
return 4;
}
else if (Rs >= 3579) {
return 4;
}
}
else if (diffRsCsdiff >= -6) {
if (Rs < 3832) {
return 8;
}
else if (Rs >= 3832) {
return 8;
}
}
}
else if (ssDCval >= 75) {
return 4;
}
}
}
else if (AFD >= 7) {
return 8;
}
}
else if (RsCsDiff >= 37) {
if (MVDiff < 5) {
if (Rs < 3299) {
if (Cs < 1546) {
if (posBalance < 57) {
return 2;
}
else if (posBalance >= 57) {
return 4;
}
}
else if (Cs >= 1546) {
return 4;
}
}
else if (Rs >= 3299) {
if (Rs < 4836) {
if (Cs < 1323) {
return 8;
}
else if (Cs >= 1323) {
return 2;
}
}
else if (Rs >= 4836) {
return 4;
}
}
}
else if (MVDiff >= 5) {
if (Cs < 1755) {
if (posBalance < 52) {
if (Rs < 3735) {
return 4;
}
else if (Rs >= 3735) {
return 1;
}
}
else if (posBalance >= 52) {
return 2;
}
}
else if (Cs >= 1755) {
return 8;
}
}
}
}
else if (AFD >= 9) {
if (CsDiff < 6) {
if (ssDCval < 91) {
return 4;
}
else if (ssDCval >= 91) {
return 4;
}
}
else if (CsDiff >= 6) {
return 1;
}
}
}
}
else if (ssDCval >= 98) {
if (Rs < 3678) {
return 8;
}
else if (Rs >= 3678) {
return 2;
}
}
}
else if (Cs >= 2162) {
if (RsCsDiff < 94) {
if (mu_mv_mag_sq < 97) {
return 8;
}
else if (mu_mv_mag_sq >= 97) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (mu_mv_mag_sq < 105) {
return 4;
}
else if (mu_mv_mag_sq >= 105) {
if (Rs < 3349) {
return 8;
}
else if (Rs >= 3349) {
if (Rs < 4325) {
return 4;
}
else if (Rs >= 4325) {
return 8;
}
}
}
}
}
}
else if (RsCsDiff >= 94) {
if (Cs < 2301) {
return 8;
}
else if (Cs >= 2301) {
return 1;
}
}
}
}
}
}
else if (SC >= 10031) {
if (ssDCval < 49) {
return 8;
}
else if (ssDCval >= 49) {
return 2;
}
}
}
}
else if (negBalance >= 39) {
if (Rs < 1395) {
if (RsCsDiff < 51) {
if (ssDCval < 83) {
if (refDCval < 59) {
if (Rs < 968) {
if (RsCsDiff < 33) {
return 2;
}
else if (RsCsDiff >= 33) {
return 4;
}
}
else if (Rs >= 968) {
if (Cs < 1517) {
if (negBalance < 49) {
if (Rs < 1026) {
return 2;
}
else if (Rs >= 1026) {
if (Rs < 1295) {
return 4;
}
else if (Rs >= 1295) {
if (Cs < 1451) {
return 8;
}
else if (Cs >= 1451) {
return 4;
}
}
}
}
else if (negBalance >= 49) {
return 8;
}
}
else if (Cs >= 1517) {
return 8;
}
}
}
else if (refDCval >= 59) {
if (tcor < 81) {
if (ssDCval < 72) {
return 8;
}
else if (ssDCval >= 72) {
return 2;
}
}
else if (tcor >= 81) {
if (Rs < 1120) {
if (Rs < 526) {
return 1;
}
else if (Rs >= 526) {
if (TSC < 8) {
if (SC < 2402) {
if (Rs < 913) {
if (Rs < 643) {
return 2;
}
else if (Rs >= 643) {
if (posBalance < 64) {
return 8;
}
else if (posBalance >= 64) {
return 4;
}
}
}
else if (Rs >= 913) {
if (Rs < 965) {
return 4;
}
else if (Rs >= 965) {
return 1;
}
}
}
else if (SC >= 2402) {
if (gchDC < 1) {
if (Rs < 1066) {
return 1;
}
else if (Rs >= 1066) {
return 2;
}
}
else if (gchDC >= 1) {
if (diffRsCsdiff < 3) {
if (diffAFD < -2) {
return 1;
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (diffRsCsdiff >= 3) {
if (Cs < 1787) {
if (Rs < 907) {
return 8;
}
else if (Rs >= 907) {
if (Rs < 969) {
return 2;
}
else if (Rs >= 969) {
return 4;
}
}
}
else if (Cs >= 1787) {
return 1;
}
}
}
}
}
else if (TSC >= 8) {
if (AFD < 12) {
return 1;
}
else if (AFD >= 12) {
return 4;
}
}
}
}
else if (Rs >= 1120) {
if (SC < 2687) {
if (ssDCval < 64) {
return 1;
}
else if (ssDCval >= 64) {
return 4;
}
}
else if (SC >= 2687) {
if (refDCval < 69) {
return 1;
}
else if (refDCval >= 69) {
return 4;
}
}
}
}
}
}
else if (ssDCval >= 83) {
if (mu_mv_mag_sq < 3) {
if (Cs < 1400) {
if (posBalance < 66) {
return 4;
}
else if (posBalance >= 66) {
return 2;
}
}
else if (Cs >= 1400) {
if (RsDiff < 5) {
if (refDCval < 103) {
return 8;
}
else if (refDCval >= 103) {
return 4;
}
}
else if (RsDiff >= 5) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (ssDCval < 97) {
if (diffAFD < -3) {
return 4;
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (ssDCval >= 97) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
if (Rs < 759) {
return 8;
}
else if (Rs >= 759) {
return 4;
}
}
}
}
}
}
else if (RsCsDiff >= 51) {
if (TSC < 7) {
if (refDCval < 87) {
if (mcTcor < 98) {
if (AFD < 13) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (MVDiff < 8) {
return 4;
}
else if (MVDiff >= 8) {
return 2;
}
}
}
}
else if (AFD >= 13) {
if (Cs < 2228) {
if (RsCsDiff < 94) {
return 4;
}
else if (RsCsDiff >= 94) {
return 8;
}
}
else if (Cs >= 2228) {
return 2;
}
}
}
else if (mcTcor >= 98) {
return 1;
}
}
else if (refDCval >= 87) {
if (RsCsDiff < 79) {
return 8;
}
else if (RsCsDiff >= 79) {
if (Rs < 1163) {
if (Rs < 1125) {
return 8;
}
else if (Rs >= 1125) {
return 2;
}
}
else if (Rs >= 1163) {
return 1;
}
}
}
}
else if (TSC >= 7) {
if (negBalance < 56) {
if (MVDiff < 13) {
if (Cs < 1568) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
if (negBalance < 55) {
if (Rs < 960) {
if (diffAFD < -1) {
return 4;
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (Rs >= 960) {
return 8;
}
}
else if (negBalance >= 55) {
return 4;
}
}
}
else if (Cs >= 1568) {
if (ssDCval < 73) {
if (Rs < 932) {
return 2;
}
else if (Rs >= 932) {
return 8;
}
}
else if (ssDCval >= 73) {
if (Rs < 788) {
return 8;
}
else if (Rs >= 788) {
return 4;
}
}
}
}
else if (MVDiff >= 13) {
if (CsDiff < 8) {
if (TSC < 8) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
if (Rs < 845) {
return 4;
}
else if (Rs >= 845) {
return 8;
}
}
}
else if (TSC >= 8) {
if (AFD < 22) {
if (MVDiff < 15) {
if (AFD < 15) {
return 4;
}
else if (AFD >= 15) {
return 2;
}
}
else if (MVDiff >= 15) {
return 4;
}
}
else if (AFD >= 22) {
return 8;
}
}
}
else if (CsDiff >= 8) {
return 4;
}
}
}
else if (negBalance >= 56) {
if (RsCsDiff < 73) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
return 8;
}
}
else if (RsCsDiff >= 73) {
if (tcor < 62) {
return 4;
}
else if (tcor >= 62) {
if (diffMVdiffVal < 1) {
if (gchDC < 3) {
return 2;
}
else if (gchDC >= 3) {
if (Rs < 1105) {
return 4;
}
else if (Rs >= 1105) {
if (Rs < 1158) {
return 2;
}
else if (Rs >= 1158) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= 1) {
return 8;
}
}
}
}
}
}
}
else if (Rs >= 1395) {
if (CsDiff < 7) {
if (mcTcor < 94) {
if (ssDCval < 89) {
if (mu_mv_mag_sq < 91) {
if (RsDiff < 9) {
if (Cs < 1454) {
if (ssDCval < 88) {
if (negBalance < 47) {
if (diffAFD < -1) {
if (negBalance < 43) {
if (Rs < 1446) {
return 1;
}
else if (Rs >= 1446) {
return 8;
}
}
else if (negBalance >= 43) {
return 1;
}
}
else if (diffAFD >= -1) {
if (RsDiff < 8) {
if (Cs < 1448) {
if (posBalance < 71) {
return 8;
}
else if (posBalance >= 71) {
return 4;
}
}
else if (Cs >= 1448) {
return 1;
}
}
else if (RsDiff >= 8) {
if (Rs < 2371) {
return 4;
}
else if (Rs >= 2371) {
return 2;
}
}
}
}
else if (negBalance >= 47) {
if (MVDiff < 18) {
if (Rs < 2450) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (Rs < 1608) {
return 4;
}
else if (Rs >= 1608) {
return 8;
}
}
}
else if (Rs >= 2450) {
if (Rs < 2465) {
return 4;
}
else if (Rs >= 2465) {
return 8;
}
}
}
else if (MVDiff >= 18) {
if (mu_mv_mag_sq < 60) {
if (ssDCval < 58) {
if (Rs < 1775) {
return 4;
}
else if (Rs >= 1775) {
return 8;
}
}
else if (ssDCval >= 58) {
return 8;
}
}
else if (mu_mv_mag_sq >= 60) {
if (Cs < 1399) {
return 8;
}
else if (Cs >= 1399) {
return 1;
}
}
}
}
}
else if (ssDCval >= 88) {
if (Rs < 1558) {
if (Cs < 1329) {
return 8;
}
else if (Cs >= 1329) {
return 4;
}
}
else if (Rs >= 1558) {
return 8;
}
}
}
else if (Cs >= 1454) {
if (ssDCval < 78) {
if (posBalance < 49) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 4229) {
if (CsDiff < 6) {
return 8;
}
else if (CsDiff >= 6) {
return 4;
}
}
else if (Rs >= 4229) {
return 4;
}
}
}
else if (posBalance >= 49) {
if (SC < 2940) {
if (SC < 2932) {
return 8;
}
else if (SC >= 2932) {
return 4;
}
}
else if (SC >= 2940) {
return 8;
}
}
}
else if (ssDCval >= 78) {
if (AFD < 16) {
if (Cs < 2980) {
if (mu_mv_mag_sq < 46) {
if (refDCval < 80) {
if (posBalance < 54) {
if (Rs < 2190) {
return 1;
}
else if (Rs >= 2190) {
return 8;
}
}
else if (posBalance >= 54) {
return 4;
}
}
else if (refDCval >= 80) {
if (Cs < 2025) {
return 8;
}
else if (Cs >= 2025) {
if (Cs < 2051) {
return 4;
}
else if (Cs >= 2051) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 46) {
if (diffMVdiffVal < -5) {
if (TSC < 9) {
return 8;
}
else if (TSC >= 9) {
return 4;
}
}
else if (diffMVdiffVal >= -5) {
return 8;
}
}
}
else if (Cs >= 2980) {
if (Rs < 3141) {
return 4;
}
else if (Rs >= 3141) {
return 8;
}
}
}
else if (AFD >= 16) {
if (Rs < 1885) {
return 4;
}
else if (Rs >= 1885) {
return 8;
}
}
}
}
}
else if (RsDiff >= 9) {
return 4;
}
}
else if (mu_mv_mag_sq >= 91) {
if (Cs < 2431) {
if (MVDiff < 5) {
if (posBalance < 67) {
if (diffMVdiffVal < -5) {
if (Rs < 2252) {
return 8;
}
else if (Rs >= 2252) {
return 4;
}
}
else if (diffMVdiffVal >= -5) {
if (Cs < 1479) {
if (diffRsCsdiff < -1) {
if (Rs < 1778) {
return 8;
}
else if (Rs >= 1778) {
if (Rs < 2171) {
return 2;
}
else if (Rs >= 2171) {
return 8;
}
}
}
else if (diffRsCsdiff >= -1) {
return 8;
}
}
else if (Cs >= 1479) {
return 8;
}
}
}
else if (posBalance >= 67) {
return 4;
}
}
else if (MVDiff >= 5) {
if (diffRsCsdiff < -25) {
if (negBalance < 48) {
if (Cs < 1938) {
return 8;
}
else if (Cs >= 1938) {
return 4;
}
}
else if (negBalance >= 48) {
return 4;
}
}
else if (diffRsCsdiff >= -25) {
if (Rs < 1495) {
if (Rs < 1472) {
if (Rs < 1432) {
return 1;
}
else if (Rs >= 1432) {
return 8;
}
}
else if (Rs >= 1472) {
return 4;
}
}
else if (Rs >= 1495) {
if (RsDiff < 8) {
return 8;
}
else if (RsDiff >= 8) {
if (AFD < 10) {
return 4;
}
else if (AFD >= 10) {
return 8;
}
}
}
}
}
}
else if (Cs >= 2431) {
if (Rs < 2216) {
return 8;
}
else if (Rs >= 2216) {
return 4;
}
}
}
}
else if (ssDCval >= 89) {
if (SC < 4032) {
if (AFD < 17) {
if (mcTcor < 87) {
if (Rs < 1411) {
return 4;
}
else if (Rs >= 1411) {
return 8;
}
}
else if (mcTcor >= 87) {
if (RsCsDiff < 69) {
if (mu_mv_mag_sq < 64) {
if (Rs < 1439) {
if (AFD < 10) {
return 8;
}
else if (AFD >= 10) {
return 4;
}
}
else if (Rs >= 1439) {
return 8;
}
}
else if (mu_mv_mag_sq >= 64) {
if (diffAFD < 1) {
return 8;
}
else if (diffAFD >= 1) {
if (RsDiff < 4) {
return 8;
}
else if (RsDiff >= 4) {
return 2;
}
}
}
}
else if (RsCsDiff >= 69) {
if (AFD < 12) {
if (mu_mv_mag_sq < 64) {
return 2;
}
else if (mu_mv_mag_sq >= 64) {
return 8;
}
}
else if (AFD >= 12) {
return 8;
}
}
}
}
else if (AFD >= 17) {
if (SC < 3476) {
if (Rs < 1766) {
return 4;
}
else if (Rs >= 1766) {
return 8;
}
}
else if (SC >= 3476) {
return 1;
}
}
}
else if (SC >= 4032) {
if (Cs < 2706) {
if (diffAFD < -4) {
if (Cs < 1642) {
return 4;
}
else if (Cs >= 1642) {
return 8;
}
}
else if (diffAFD >= -4) {
return 8;
}
}
else if (Cs >= 2706) {
return 4;
}
}
}
}
else if (mcTcor >= 94) {
if (SC < 6661) {
if (RsDiff < 8) {
if (SC < 4412) {
if (refDCval < 89) {
if (mu_mv_mag_sq < 89) {
if (MVDiff < 7) {
if (AFD < 11) {
if (Cs < 1665) {
if (mu_mv_mag_sq < 85) {
if (ssDCval < 82) {
if (RsCsDiff < 57) {
if (RsDiff < 5) {
if (diffMVdiffVal < 1) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 1;
}
}
else if (diffMVdiffVal >= 1) {
return 1;
}
}
else if (RsDiff >= 5) {
if (MVDiff < 3) {
return 1;
}
else if (MVDiff >= 3) {
return 8;
}
}
}
else if (RsCsDiff >= 57) {
if (Rs < 2638) {
return 8;
}
else if (Rs >= 2638) {
return 2;
}
}
}
else if (ssDCval >= 82) {
if (refDCval < 88) {
if (diffMVdiffVal < -4) {
if (Rs < 2143) {
return 8;
}
else if (Rs >= 2143) {
return 2;
}
}
else if (diffMVdiffVal >= -4) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
else if (refDCval >= 88) {
if (Rs < 2216) {
return 2;
}
else if (Rs >= 2216) {
if (Cs < 1448) {
return 4;
}
else if (Cs >= 1448) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 85) {
if (SC < 3450) {
return 8;
}
else if (SC >= 3450) {
if (posBalance < 54) {
return 4;
}
else if (posBalance >= 54) {
return 1;
}
}
}
}
else if (Cs >= 1665) {
if (Cs < 2169) {
if (TSC < 8) {
if (mu_mv_mag_sq < 85) {
return 8;
}
else if (mu_mv_mag_sq >= 85) {
if (Rs < 2508) {
if (Rs < 1430) {
return 1;
}
else if (Rs >= 1430) {
return 8;
}
}
else if (Rs >= 2508) {
return 1;
}
}
}
else if (TSC >= 8) {
if (ssDCval < 82) {
return 4;
}
else if (ssDCval >= 82) {
return 8;
}
}
}
else if (Cs >= 2169) {
if (Rs < 2132) {
return 1;
}
else if (Rs >= 2132) {
return 8;
}
}
}
}
else if (AFD >= 11) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
if (RsCsDiff < 67) {
if (Rs < 1962) {
return 4;
}
else if (Rs >= 1962) {
return 1;
}
}
else if (RsCsDiff >= 67) {
return 8;
}
}
}
}
else if (MVDiff >= 7) {
if (Cs < 1341) {
if (Cs < 1320) {
return 8;
}
else if (Cs >= 1320) {
return 1;
}
}
else if (Cs >= 1341) {
if (tcor < 88) {
if (Cs < 1427) {
return 8;
}
else if (Cs >= 1427) {
if (MVDiff < 9) {
return 8;
}
else if (MVDiff >= 9) {
return 4;
}
}
}
else if (tcor >= 88) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 89) {
if (RsDiff < 6) {
if (Rs < 1468) {
if (Cs < 1797) {
return 4;
}
else if (Cs >= 1797) {
return 8;
}
}
else if (Rs >= 1468) {
if (AFD < 8) {
if (Cs < 1352) {
if (Rs < 2083) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 1;
}
}
else if (Rs >= 2083) {
return 1;
}
}
else if (Cs >= 1352) {
return 8;
}
}
else if (AFD >= 8) {
if (MVDiff < 1) {
if (Rs < 1826) {
return 4;
}
else if (Rs >= 1826) {
return 1;
}
}
else if (MVDiff >= 1) {
if (TSC < 7) {
return 4;
}
else if (TSC >= 7) {
if (negBalance < 41) {
if (Rs < 2255) {
return 4;
}
else if (Rs >= 2255) {
return 8;
}
}
else if (negBalance >= 41) {
return 8;
}
}
}
}
}
}
else if (RsDiff >= 6) {
if (negBalance < 43) {
if (Cs < 1731) {
return 2;
}
else if (Cs >= 1731) {
return 8;
}
}
else if (negBalance >= 43) {
if (SC < 4243) {
return 8;
}
else if (SC >= 4243) {
return 4;
}
}
}
}
}
else if (refDCval >= 89) {
if (Rs < 2052) {
if (posBalance < 59) {
if (RsDiff < 6) {
if (Cs < 1865) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
if (diffTSC < 2) {
return 8;
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (Cs >= 1865) {
if (tcor < 94) {
return 8;
}
else if (tcor >= 94) {
return 2;
}
}
}
else if (RsDiff >= 6) {
if (mcTcor < 96) {
if (posBalance < 49) {
return 1;
}
else if (posBalance >= 49) {
return 8;
}
}
else if (mcTcor >= 96) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
return 2;
}
}
}
}
else if (posBalance >= 59) {
if (tcor < 92) {
if (Rs < 1776) {
return 8;
}
else if (Rs >= 1776) {
return 2;
}
}
else if (tcor >= 92) {
return 2;
}
}
}
else if (Rs >= 2052) {
if (posBalance < 47) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (negBalance < 44) {
if (AFD < 3) {
return 2;
}
else if (AFD >= 3) {
return 4;
}
}
else if (negBalance >= 44) {
if (AFD < 5) {
if (mu_mv_mag_sq < 99) {
if (Rs < 2241) {
return 2;
}
else if (Rs >= 2241) {
return 4;
}
}
else if (mu_mv_mag_sq >= 99) {
return 1;
}
}
else if (AFD >= 5) {
return 8;
}
}
}
}
else if (posBalance >= 47) {
if (diffMVdiffVal < -43) {
return 4;
}
else if (diffMVdiffVal >= -43) {
if (diffRsCsdiff < -41) {
if (CsDiff < 5) {
return 8;
}
else if (CsDiff >= 5) {
if (Rs < 2362) {
return 1;
}
else if (Rs >= 2362) {
return 4;
}
}
}
else if (diffRsCsdiff >= -41) {
return 8;
}
}
}
}
}
}
else if (SC >= 4412) {
if (CsDiff < 4) {
if (tcor < 96) {
if (mu_mv_mag_sq < 5) {
if (Rs < 3030) {
return 8;
}
else if (Rs >= 3030) {
return 4;
}
}
else if (mu_mv_mag_sq >= 5) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
if (Rs < 2978) {
return 4;
}
else if (Rs >= 2978) {
return 8;
}
}
}
}
else if (tcor >= 96) {
if (Cs < 1884) {
return 2;
}
else if (Cs >= 1884) {
return 8;
}
}
}
else if (CsDiff >= 4) {
if (Cs < 2805) {
if (posBalance < 49) {
if (negBalance < 46) {
if (CsDiff < 6) {
if (mu_mv_mag_sq < 74) {
return 4;
}
else if (mu_mv_mag_sq >= 74) {
if (Cs < 2308) {
return 8;
}
else if (Cs >= 2308) {
return 4;
}
}
}
else if (CsDiff >= 6) {
if (mcTcor < 96) {
return 8;
}
else if (mcTcor >= 96) {
if (Rs < 3228) {
return 1;
}
else if (Rs >= 3228) {
return 2;
}
}
}
}
else if (negBalance >= 46) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
if (Rs < 2987) {
return 8;
}
else if (Rs >= 2987) {
return 2;
}
}
}
}
else if (posBalance >= 49) {
if (negBalance < 52) {
if (refDCval < 65) {
if (Cs < 1853) {
return 8;
}
else if (Cs >= 1853) {
return 4;
}
}
else if (refDCval >= 65) {
if (Cs < 1538) {
if (AFD < 9) {
if (Rs < 3091) {
return 4;
}
else if (Rs >= 3091) {
return 2;
}
}
else if (AFD >= 9) {
return 8;
}
}
else if (Cs >= 1538) {
if (Rs < 2153) {
return 1;
}
else if (Rs >= 2153) {
if (Cs < 2724) {
return 8;
}
else if (Cs >= 2724) {
if (Rs < 3073) {
return 4;
}
else if (Rs >= 3073) {
return 8;
}
}
}
}
}
}
else if (negBalance >= 52) {
if (AFD < 13) {
return 4;
}
else if (AFD >= 13) {
if (Rs < 2495) {
return 1;
}
else if (Rs >= 2495) {
return 8;
}
}
}
}
}
else if (Cs >= 2805) {
if (Rs < 3002) {
return 4;
}
else if (Rs >= 3002) {
if (Cs < 3184) {
return 8;
}
else if (Cs >= 3184) {
return 4;
}
}
}
}
}
}
else if (RsDiff >= 8) {
if (refDCval < 89) {
if (AFD < 10) {
if (Cs < 2108) {
return 2;
}
else if (Cs >= 2108) {
return 4;
}
}
else if (AFD >= 10) {
if (tcor < 88) {
if (TSC < 7) {
return 2;
}
else if (TSC >= 7) {
return 4;
}
}
else if (tcor >= 88) {
return 8;
}
}
}
else if (refDCval >= 89) {
if (negBalance < 55) {
return 8;
}
else if (negBalance >= 55) {
if (Rs < 2343) {
return 2;
}
else if (Rs >= 2343) {
if (Rs < 3036) {
return 8;
}
else if (Rs >= 3036) {
return 4;
}
}
}
}
}
}
else if (SC >= 6661) {
if (Cs < 2487) {
return 4;
}
else if (Cs >= 2487) {
if (ssDCval < 96) {
return 8;
}
else if (ssDCval >= 96) {
if (Cs < 3372) {
return 4;
}
else if (Cs >= 3372) {
return 8;
}
}
}
}
}
}
else if (CsDiff >= 7) {
if (ssDCval < 101) {
if (Cs < 2767) {
if (tcor < 87) {
if (MVDiff < 43) {
if (refDCval < 50) {
if (negBalance < 48) {
if (TSC < 6) {
return 1;
}
else if (TSC >= 6) {
return 8;
}
}
else if (negBalance >= 48) {
if (ssDCval < 44) {
return 8;
}
else if (ssDCval >= 44) {
return 4;
}
}
}
else if (refDCval >= 50) {
if (negBalance < 49) {
if (mu_mv_mag_sq < 3) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffRsCsdiff < 1) {
if (ssDCval < 52) {
return 4;
}
else if (ssDCval >= 52) {
return 8;
}
}
else if (diffRsCsdiff >= 1) {
if (Rs < 1837) {
return 4;
}
else if (Rs >= 1837) {
return 2;
}
}
}
}
else if (negBalance >= 49) {
if (AFD < 16) {
if (Cs < 1363) {
return 1;
}
else if (Cs >= 1363) {
if (RsDiff < 6) {
if (Cs < 2200) {
return 8;
}
else if (Cs >= 2200) {
return 4;
}
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (AFD >= 16) {
if (diffRsCsdiff < 2) {
if (ssDCval < 88) {
if (Rs < 2355) {
if (Cs < 1589) {
if (Rs < 1464) {
return 4;
}
else if (Rs >= 1464) {
return 8;
}
}
else if (Cs >= 1589) {
return 4;
}
}
else if (Rs >= 2355) {
return 8;
}
}
else if (ssDCval >= 88) {
return 8;
}
}
else if (diffRsCsdiff >= 2) {
return 8;
}
}
}
}
}
else if (MVDiff >= 43) {
if (Rs < 1599) {
return 8;
}
else if (Rs >= 1599) {
return 1;
}
}
}
else if (tcor >= 87) {
if (Cs < 2221) {
if (AFD < 16) {
if (diffRsCsdiff < 31) {
return 8;
}
else if (diffRsCsdiff >= 31) {
if (ssDCval < 79) {
return 4;
}
else if (ssDCval >= 79) {
if (ssDCval < 93) {
return 8;
}
else if (ssDCval >= 93) {
return 1;
}
}
}
}
else if (AFD >= 16) {
return 2;
}
}
else if (Cs >= 2221) {
if (Rs < 2193) {
if (Rs < 1883) {
if (AFD < 13) {
if (Rs < 1752) {
return 2;
}
else if (Rs >= 1752) {
return 8;
}
}
else if (AFD >= 13) {
return 4;
}
}
else if (Rs >= 1883) {
return 1;
}
}
else if (Rs >= 2193) {
if (posBalance < 51) {
if (Rs < 2953) {
return 1;
}
else if (Rs >= 2953) {
return 4;
}
}
else if (posBalance >= 51) {
if (ssDCval < 98) {
return 8;
}
else if (ssDCval >= 98) {
return 4;
}
}
}
}
}
}
else if (Cs >= 2767) {
if (Rs < 3027) {
if (Cs < 2907) {
if (Cs < 2803) {
if (Rs < 2489) {
return 4;
}
else if (Rs >= 2489) {
return 2;
}
}
else if (Cs >= 2803) {
return 8;
}
}
else if (Cs >= 2907) {
return 4;
}
}
else if (Rs >= 3027) {
if (MVDiff < 2) {
if (Rs < 3605) {
if (RsDiff < 7) {
return 1;
}
else if (RsDiff >= 7) {
return 8;
}
}
else if (Rs >= 3605) {
return 4;
}
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
else if (ssDCval >= 101) {
if (mu_mv_mag_sq < 87) {
if (mcTcor < 96) {
if (AFD < 16) {
if (AFD < 11) {
if (Rs < 3477) {
return 8;
}
else if (Rs >= 3477) {
return 1;
}
}
else if (AFD >= 11) {
return 8;
}
}
else if (AFD >= 16) {
if (Rs < 1731) {
return 4;
}
else if (Rs >= 1731) {
return 8;
}
}
}
else if (mcTcor >= 96) {
return 2;
}
}
else if (mu_mv_mag_sq >= 87) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 105) {
if (Rs < 4439) {
if (refDCval < 163) {
if (Rs < 1436) {
if (mu_mv_mag_sq < 29) {
if (ssDCval < 138) {
if (mcTcor < 88) {
if (Rs < 1134) {
if (Rs < 978) {
return 2;
}
else if (Rs >= 978) {
return 1;
}
}
else if (Rs >= 1134) {
return 4;
}
}
else if (mcTcor >= 88) {
if (AFD < 13) {
if (CsDiff < 4) {
if (Rs < 1161) {
return 8;
}
else if (Rs >= 1161) {
return 1;
}
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (AFD >= 13) {
if (Rs < 1120) {
if (Rs < 663) {
return 2;
}
else if (Rs >= 663) {
return 8;
}
}
else if (Rs >= 1120) {
if (mu_mv_mag_sq < 3) {
if (AFD < 19) {
return 8;
}
else if (AFD >= 19) {
return 4;
}
}
else if (mu_mv_mag_sq >= 3) {
return 4;
}
}
}
}
}
else if (ssDCval >= 138) {
if (SC < 2611) {
if (AFD < 11) {
return 2;
}
else if (AFD >= 11) {
return 1;
}
}
else if (SC >= 2611) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 29) {
if (Rs < 1247) {
if (Rs < 989) {
if (ssDCval < 119) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
if (AFD < 8) {
return 4;
}
else if (AFD >= 8) {
return 8;
}
}
}
else if (ssDCval >= 119) {
if (Cs < 2105) {
return 8;
}
else if (Cs >= 2105) {
return 1;
}
}
}
else if (Rs >= 989) {
if (CsDiff < 6) {
if (ssDCval < 127) {
if (negBalance < 40) {
if (ssDCval < 124) {
return 4;
}
else if (ssDCval >= 124) {
return 8;
}
}
else if (negBalance >= 40) {
return 8;
}
}
else if (ssDCval >= 127) {
if (ssDCval < 132) {
if (posBalance < 46) {
return 8;
}
else if (posBalance >= 46) {
if (SC < 2615) {
if (Cs < 1546) {
return 4;
}
else if (Cs >= 1546) {
return 8;
}
}
else if (SC >= 2615) {
return 4;
}
}
}
else if (ssDCval >= 132) {
if (ssDCval < 147) {
if (refDCval < 132) {
return 4;
}
else if (refDCval >= 132) {
return 8;
}
}
else if (ssDCval >= 147) {
return 4;
}
}
}
}
else if (CsDiff >= 6) {
if (posBalance < 119) {
return 8;
}
else if (posBalance >= 119) {
return 1;
}
}
}
}
else if (Rs >= 1247) {
if (Cs < 1535) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
if (Cs < 1351) {
return 2;
}
else if (Cs >= 1351) {
return 8;
}
}
}
else if (Cs >= 1535) {
if (posBalance < 52) {
if (refDCval < 125) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
else if (refDCval >= 125) {
return 4;
}
}
else if (posBalance >= 52) {
return 8;
}
}
}
}
}
else if (Rs >= 1436) {
if (ssDCval < 123) {
if (mu_mv_mag_sq < 43) {
if (mcTcor < 97) {
if (negBalance < 48) {
if (posBalance < 47) {
if (Rs < 3150) {
return 4;
}
else if (Rs >= 3150) {
return 8;
}
}
else if (posBalance >= 47) {
if (AFD < 14) {
if (mcTcor < 96) {
if (diffRsCsdiff < -32) {
return 4;
}
else if (diffRsCsdiff >= -32) {
return 8;
}
}
else if (mcTcor >= 96) {
if (Rs < 3043) {
if (Rs < 2327) {
return 8;
}
else if (Rs >= 2327) {
return 4;
}
}
else if (Rs >= 3043) {
return 8;
}
}
}
else if (AFD >= 14) {
if (Cs < 1901) {
return 4;
}
else if (Cs >= 1901) {
if (Rs < 1863) {
return 1;
}
else if (Rs >= 1863) {
return 8;
}
}
}
}
}
else if (negBalance >= 48) {
if (diffRsCsdiff < -64) {
if (Rs < 1866) {
return 1;
}
else if (Rs >= 1866) {
if (Rs < 2511) {
return 8;
}
else if (Rs >= 2511) {
return 4;
}
}
}
else if (diffRsCsdiff >= -64) {
if (Cs < 1364) {
if (Cs < 1351) {
return 8;
}
else if (Cs >= 1351) {
return 1;
}
}
else if (Cs >= 1364) {
return 8;
}
}
}
}
else if (mcTcor >= 97) {
if (RsCsDiff < 47) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
else if (RsCsDiff >= 47) {
if (Rs < 2130) {
if (MVDiff < 3) {
return 4;
}
else if (MVDiff >= 3) {
return 2;
}
}
else if (Rs >= 2130) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 43) {
if (TSC < 11) {
if (posBalance < 37) {
if (negBalance < 48) {
if (posBalance < 13) {
if (negBalance < 43) {
if (Cs < 2007) {
if (Rs < 2179) {
return 1;
}
else if (Rs >= 2179) {
return 8;
}
}
else if (Cs >= 2007) {
if (Rs < 2632) {
return 4;
}
else if (Rs >= 2632) {
return 8;
}
}
}
else if (negBalance >= 43) {
if (Rs < 2958) {
return 2;
}
else if (Rs >= 2958) {
return 8;
}
}
}
else if (posBalance >= 13) {
if (Cs < 1350) {
if (Rs < 4254) {
return 1;
}
else if (Rs >= 4254) {
return 2;
}
}
else if (Cs >= 1350) {
return 8;
}
}
}
else if (negBalance >= 48) {
return 8;
}
}
else if (posBalance >= 37) {
if (negBalance < 50) {
if (refDCval < 117) {
if (mu_mv_mag_sq < 54) {
if (mu_mv_mag_sq < 53) {
return 8;
}
else if (mu_mv_mag_sq >= 53) {
if (Rs < 3193) {
return 4;
}
else if (Rs >= 3193) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (diffRsCsdiff < -82) {
if (Rs < 3489) {
return 4;
}
else if (Rs >= 3489) {
return 8;
}
}
else if (diffRsCsdiff >= -82) {
return 8;
}
}
}
else if (refDCval >= 117) {
if (diffTSC < 2) {
if (posBalance < 40) {
if (Cs < 1505) {
return 2;
}
else if (Cs >= 1505) {
return 8;
}
}
else if (posBalance >= 40) {
return 8;
}
}
else if (diffTSC >= 2) {
return 2;
}
}
}
else if (negBalance >= 50) {
if (Rs < 3078) {
return 8;
}
else if (Rs >= 3078) {
if (Rs < 3442) {
return 4;
}
else if (Rs >= 3442) {
return 8;
}
}
}
}
}
else if (TSC >= 11) {
if (CsDiff < 5) {
if (Rs < 2529) {
return 1;
}
else if (Rs >= 2529) {
return 8;
}
}
else if (CsDiff >= 5) {
if (diffRsCsdiff < -37) {
if (Rs < 2954) {
return 4;
}
else if (Rs >= 2954) {
return 8;
}
}
else if (diffRsCsdiff >= -37) {
return 8;
}
}
}
}
}
else if (ssDCval >= 123) {
if (refDCval < 142) {
if (mcTcor < 92) {
if (Cs < 2121) {
if (ssDCval < 128) {
return 8;
}
else if (ssDCval >= 128) {
if (MVDiff < 13) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
if (diffTSC < 1) {
if (RsDiff < 5) {
return 8;
}
else if (RsDiff >= 5) {
return 1;
}
}
else if (diffTSC >= 1) {
if (MVDiff < 8) {
if (MVDiff < 3) {
return 8;
}
else if (MVDiff >= 3) {
return 1;
}
}
else if (MVDiff >= 8) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 8;
}
}
}
}
}
else if (MVDiff >= 13) {
if (Rs < 1623) {
if (diffRsCsdiff < -15) {
return 8;
}
else if (diffRsCsdiff >= -15) {
return 1;
}
}
else if (Rs >= 1623) {
return 8;
}
}
}
}
else if (Cs >= 2121) {
return 8;
}
}
else if (mcTcor >= 92) {
if (Cs < 1363) {
if (mu_mv_mag_sq < 80) {
if (Rs < 1558) {
return 2;
}
else if (Rs >= 1558) {
return 8;
}
}
else if (mu_mv_mag_sq >= 80) {
if (Rs < 2833) {
if (Cs < 1336) {
return 1;
}
else if (Cs >= 1336) {
if (Rs < 2182) {
return 2;
}
else if (Rs >= 2182) {
return 8;
}
}
}
else if (Rs >= 2833) {
return 2;
}
}
}
else if (Cs >= 1363) {
if (refDCval < 127) {
if (TSC < 8) {
if (Rs < 1496) {
return 1;
}
else if (Rs >= 1496) {
if (negBalance < 46) {
return 8;
}
else if (negBalance >= 46) {
if (Cs < 2702) {
return 1;
}
else if (Cs >= 2702) {
return 8;
}
}
}
}
else if (TSC >= 8) {
if (Rs < 3184) {
if (Rs < 2540) {
if (diffAFD < -7) {
return 1;
}
else if (diffAFD >= -7) {
return 8;
}
}
else if (Rs >= 2540) {
if (mcTcor < 93) {
return 8;
}
else if (mcTcor >= 93) {
if (posBalance < 48) {
return 8;
}
else if (posBalance >= 48) {
return 1;
}
}
}
}
else if (Rs >= 3184) {
if (Rs < 3697) {
return 8;
}
else if (Rs >= 3697) {
return 4;
}
}
}
}
else if (refDCval >= 127) {
if (mu_mv_mag_sq < 109) {
if (negBalance < 29) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 2118) {
return 2;
}
else if (Rs >= 2118) {
return 4;
}
}
}
else if (negBalance >= 29) {
if (mcTcor < 97) {
return 8;
}
else if (mcTcor >= 97) {
if (Cs < 1421) {
if (AFD < 6) {
return 1;
}
else if (AFD >= 6) {
return 8;
}
}
else if (Cs >= 1421) {
if (ssDCval < 126) {
return 4;
}
else if (ssDCval >= 126) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 109) {
if (ssDCval < 141) {
if (TSC < 3) {
return 2;
}
else if (TSC >= 3) {
return 8;
}
}
else if (ssDCval >= 141) {
return 1;
}
}
}
}
}
}
else if (refDCval >= 142) {
if (AFD < 18) {
if (diffMVdiffVal < 15) {
if (mu_mv_mag_sq < 3) {
if (Rs < 1686) {
return 2;
}
else if (Rs >= 1686) {
return 8;
}
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (diffMVdiffVal >= 15) {
if (Rs < 1676) {
return 2;
}
else if (Rs >= 1676) {
return 8;
}
}
}
else if (AFD >= 18) {
if (Cs < 1613) {
if (Rs < 2959) {
return 1;
}
else if (Rs >= 2959) {
return 8;
}
}
else if (Cs >= 1613) {
return 2;
}
}
}
}
}
}
else if (refDCval >= 163) {
if (Rs < 2283) {
if (Rs < 1017) {
return 1;
}
else if (Rs >= 1017) {
return 2;
}
}
else if (Rs >= 2283) {
return 8;
}
}
}
else if (Rs >= 4439) {
if (Cs < 1707) {
if (MVDiff < 8) {
return 2;
}
else if (MVDiff >= 8) {
return 1;
}
}
else if (Cs >= 1707) {
if (Cs < 3559) {
if (tcor < 96) {
if (Rs < 4628) {
return 8;
}
else if (Rs >= 4628) {
return 1;
}
}
else if (tcor >= 96) {
if (Rs < 6055) {
return 2;
}
else if (Rs >= 6055) {
return 4;
}
}
}
else if (Cs >= 3559) {
return 8;
}
}
}
}
}
else if (RsCsDiff >= 101) {
if (Cs < 1925) {
if (Rs < 3333) {
if (tcor < 78) {
if (refDCval < 86) {
if (AFD < 18) {
if (Rs < 1523) {
if (gchDC < 2) {
if (mcTcor < 76) {
return 4;
}
else if (mcTcor >= 76) {
if (Rs < 1193) {
return 4;
}
else if (Rs >= 1193) {
if (Rs < 1446) {
return 2;
}
else if (Rs >= 1446) {
return 8;
}
}
}
}
else if (gchDC >= 2) {
if (TSC < 10) {
if (CsDiff < 11) {
if (ssDCval < 71) {
if (Rs < 777) {
return 2;
}
else if (Rs >= 777) {
return 1;
}
}
else if (ssDCval >= 71) {
if (Rs < 1147) {
return 4;
}
else if (Rs >= 1147) {
return 8;
}
}
}
else if (CsDiff >= 11) {
return 4;
}
}
else if (TSC >= 10) {
return 2;
}
}
}
else if (Rs >= 1523) {
if (Cs < 1671) {
if (posBalance < 55) {
if (ssDCval < 38) {
return 8;
}
else if (ssDCval >= 38) {
if (MVDiff < 14) {
if (Rs < 2125) {
return 8;
}
else if (Rs >= 2125) {
if (RsDiff < 11) {
return 2;
}
else if (RsDiff >= 11) {
return 1;
}
}
}
else if (MVDiff >= 14) {
if (MVDiff < 51) {
return 1;
}
else if (MVDiff >= 51) {
return 8;
}
}
}
}
else if (posBalance >= 55) {
if (SC < 3262) {
if (mu_mv_mag_sq < 3) {
return 1;
}
else if (mu_mv_mag_sq >= 3) {
return 8;
}
}
else if (SC >= 3262) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
return 4;
}
}
}
}
}
else if (Cs >= 1671) {
if (negBalance < 43) {
if (posBalance < 48) {
return 1;
}
else if (posBalance >= 48) {
if (Cs < 1743) {
if (mcTcor < 65) {
return 1;
}
else if (mcTcor >= 65) {
return 8;
}
}
else if (Cs >= 1743) {
return 4;
}
}
}
else if (negBalance >= 43) {
if (refDCval < 50) {
if (Rs < 2412) {
return 4;
}
else if (Rs >= 2412) {
return 8;
}
}
else if (refDCval >= 50) {
if (ssDCval < 80) {
if (posBalance < 57) {
if (CsDiff < 9) {
if (Rs < 2099) {
return 4;
}
else if (Rs >= 2099) {
if (Rs < 2514) {
return 8;
}
else if (Rs >= 2514) {
return 4;
}
}
}
else if (CsDiff >= 9) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 1;
}
}
}
else if (posBalance >= 57) {
if (Rs < 1774) {
return 1;
}
else if (Rs >= 1774) {
return 8;
}
}
}
else if (ssDCval >= 80) {
return 4;
}
}
}
}
}
}
else if (AFD >= 18) {
if (TSC < 14) {
if (Rs < 2684) {
if (posBalance < 48) {
return 2;
}
else if (posBalance >= 48) {
if (diffAFD < 5) {
if (diffRsCsdiff < -142) {
return 2;
}
else if (diffRsCsdiff >= -142) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 22) {
if (diffAFD < -1) {
if (MVDiff < 13) {
return 8;
}
else if (MVDiff >= 13) {
if (diffAFD < -2) {
if (MVDiff < 16) {
if (Rs < 1942) {
return 1;
}
else if (Rs >= 1942) {
return 2;
}
}
else if (MVDiff >= 16) {
return 8;
}
}
else if (diffAFD >= -2) {
if (Cs < 1738) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
if (mcTcor < 81) {
return 2;
}
else if (mcTcor >= 81) {
return 4;
}
}
}
else if (Cs >= 1738) {
if (SC < 3718) {
return 2;
}
else if (SC >= 3718) {
return 1;
}
}
}
}
}
else if (diffAFD >= -1) {
if (MVDiff < 23) {
if (diffRsCsdiff < 18) {
if (CsDiff < 9) {
if (Rs < 1632) {
return 8;
}
else if (Rs >= 1632) {
return 4;
}
}
else if (CsDiff >= 9) {
return 4;
}
}
else if (diffRsCsdiff >= 18) {
if (MVDiff < 18) {
return 8;
}
else if (MVDiff >= 18) {
if (TSC < 10) {
return 4;
}
else if (TSC >= 10) {
if (Rs < 2045) {
return 8;
}
else if (Rs >= 2045) {
return 2;
}
}
}
}
}
else if (MVDiff >= 23) {
if (negBalance < 55) {
if (diffRsCsdiff < 39) {
if (ssDCval < 58) {
return 1;
}
else if (ssDCval >= 58) {
return 4;
}
}
else if (diffRsCsdiff >= 39) {
if (Cs < 1746) {
return 8;
}
else if (Cs >= 1746) {
if (Rs < 1808) {
return 2;
}
else if (Rs >= 1808) {
return 4;
}
}
}
}
else if (negBalance >= 55) {
if (posBalance < 57) {
if (RsCsDiff < 154) {
return 2;
}
else if (RsCsDiff >= 154) {
return 4;
}
}
else if (posBalance >= 57) {
return 2;
}
}
}
}
}
else if (mu_mv_mag_sq >= 22) {
if (mu_mv_mag_sq < 37) {
if (MVDiff < 71) {
if (Cs < 1654) {
return 8;
}
else if (Cs >= 1654) {
if (Rs < 1761) {
return 4;
}
else if (Rs >= 1761) {
return 8;
}
}
}
else if (MVDiff >= 71) {
return 4;
}
}
else if (mu_mv_mag_sq >= 37) {
if (diffRsCsdiff < 8) {
if (Rs < 820) {
return 8;
}
else if (Rs >= 820) {
return 1;
}
}
else if (diffRsCsdiff >= 8) {
if (Rs < 1875) {
return 4;
}
else if (Rs >= 1875) {
return 8;
}
}
}
}
}
else if (ssDCval >= 78) {
if (refDCval < 84) {
if (posBalance < 65) {
if (diffMVdiffVal < 5) {
if (Cs < 1430) {
return 8;
}
else if (Cs >= 1430) {
if (Cs < 1873) {
if (AFD < 27) {
return 4;
}
else if (AFD >= 27) {
if (Rs < 1521) {
return 8;
}
else if (Rs >= 1521) {
return 4;
}
}
}
else if (Cs >= 1873) {
return 8;
}
}
}
else if (diffMVdiffVal >= 5) {
if (MVDiff < 41) {
return 8;
}
else if (MVDiff >= 41) {
if (Rs < 2026) {
return 4;
}
else if (Rs >= 2026) {
return 2;
}
}
}
}
else if (posBalance >= 65) {
if (tcor < 72) {
return 8;
}
else if (tcor >= 72) {
if (Rs < 1224) {
return 4;
}
else if (Rs >= 1224) {
return 2;
}
}
}
}
else if (refDCval >= 84) {
if (CsDiff < 11) {
return 4;
}
else if (CsDiff >= 11) {
return 2;
}
}
}
}
}
else if (diffAFD >= 5) {
if (posBalance < 64) {
if (Cs < 1595) {
return 8;
}
else if (Cs >= 1595) {
if (Rs < 2116) {
return 1;
}
else if (Rs >= 2116) {
return 4;
}
}
}
else if (posBalance >= 64) {
return 2;
}
}
}
}
else if (Rs >= 2684) {
if (ssDCval < 80) {
return 2;
}
else if (ssDCval >= 80) {
return 4;
}
}
}
else if (TSC >= 14) {
if (RsCsDiff < 129) {
return 4;
}
else if (RsCsDiff >= 129) {
if (tcor < 59) {
if (mcTcor < 67) {
if (mcTcor < 54) {
return 8;
}
else if (mcTcor >= 54) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 4;
}
}
}
else if (mcTcor >= 67) {
if (Cs < 1598) {
if (tcor < 45) {
return 1;
}
else if (tcor >= 45) {
return 8;
}
}
else if (Cs >= 1598) {
return 1;
}
}
}
else if (tcor >= 59) {
return 8;
}
}
}
}
}
else if (refDCval >= 86) {
if (mcTcor < 87) {
if (RsDiff < 8) {
if (negBalance < 57) {
return 8;
}
else if (negBalance >= 57) {
return 2;
}
}
else if (RsDiff >= 8) {
if (ssDCval < 118) {
if (Cs < 1365) {
if (Rs < 2705) {
return 4;
}
else if (Rs >= 2705) {
return 2;
}
}
else if (Cs >= 1365) {
if (tcor < 64) {
if (TSC < 13) {
if (ssDCval < 102) {
return 8;
}
else if (ssDCval >= 102) {
if (Rs < 2103) {
return 8;
}
else if (Rs >= 2103) {
return 4;
}
}
}
else if (TSC >= 13) {
if (refDCval < 93) {
return 1;
}
else if (refDCval >= 93) {
if (AFD < 39) {
if (gchDC < 4) {
if (posBalance < 66) {
return 8;
}
else if (posBalance >= 66) {
return 4;
}
}
else if (gchDC >= 4) {
if (Rs < 2630) {
return 4;
}
else if (Rs >= 2630) {
return 2;
}
}
}
else if (AFD >= 39) {
if (Rs < 3066) {
return 2;
}
else if (Rs >= 3066) {
return 1;
}
}
}
}
}
else if (tcor >= 64) {
if (Cs < 1529) {
if (refDCval < 99) {
if (gchDC < 1) {
if (Rs < 1988) {
return 1;
}
else if (Rs >= 1988) {
return 2;
}
}
else if (gchDC >= 1) {
return 8;
}
}
else if (refDCval >= 99) {
return 4;
}
}
else if (Cs >= 1529) {
return 8;
}
}
}
}
else if (ssDCval >= 118) {
if (Rs < 2121) {
if (Cs < 1400) {
if (RsDiff < 13) {
return 1;
}
else if (RsDiff >= 13) {
return 8;
}
}
else if (Cs >= 1400) {
if (gchDC < 7) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
if (Rs < 1947) {
return 8;
}
else if (Rs >= 1947) {
return 1;
}
}
}
else if (gchDC >= 7) {
return 4;
}
}
}
else if (Rs >= 2121) {
if (refDCval < 140) {
if (gchDC < 3) {
return 1;
}
else if (gchDC >= 3) {
if (Rs < 2404) {
return 1;
}
else if (Rs >= 2404) {
return 8;
}
}
}
else if (refDCval >= 140) {
if (Cs < 1615) {
return 1;
}
else if (Cs >= 1615) {
if (mcTcor < 81) {
if (Rs < 2921) {
return 1;
}
else if (Rs >= 2921) {
return 8;
}
}
else if (mcTcor >= 81) {
return 8;
}
}
}
}
}
}
}
else if (mcTcor >= 87) {
if (mu_mv_mag_sq < 26) {
if (RsCsDiff < 383) {
if (Rs < 2972) {
if (Rs < 1879) {
if (AFD < 18) {
return 8;
}
else if (AFD >= 18) {
if (negBalance < 56) {
if (diffRsCsdiff < 51) {
if (Rs < 1833) {
return 4;
}
else if (Rs >= 1833) {
return 8;
}
}
else if (diffRsCsdiff >= 51) {
return 8;
}
}
else if (negBalance >= 56) {
if (Rs < 1866) {
if (RsDiff < 8) {
return 2;
}
else if (RsDiff >= 8) {
if (ssDCval < 94) {
if (Rs < 1646) {
return 2;
}
else if (Rs >= 1646) {
return 1;
}
}
else if (ssDCval >= 94) {
if (CsDiff < 11) {
return 8;
}
else if (CsDiff >= 11) {
return 2;
}
}
}
}
else if (Rs >= 1866) {
return 4;
}
}
}
}
else if (Rs >= 1879) {
if (TSC < 10) {
if (SC < 3928) {
if (tcor < 75) {
return 2;
}
else if (tcor >= 75) {
if (TSC < 8) {
return 2;
}
else if (TSC >= 8) {
return 8;
}
}
}
else if (SC >= 3928) {
if (SC < 4522) {
return 8;
}
else if (SC >= 4522) {
if (Rs < 2847) {
return 2;
}
else if (Rs >= 2847) {
if (Rs < 2945) {
return 4;
}
else if (Rs >= 2945) {
return 8;
}
}
}
}
}
else if (TSC >= 10) {
if (posBalance < 63) {
if (Cs < 1353) {
return 4;
}
else if (Cs >= 1353) {
return 8;
}
}
else if (posBalance >= 63) {
if (diffMVdiffVal < -1) {
return 1;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2520) {
return 8;
}
else if (Rs >= 2520) {
return 4;
}
}
}
}
}
}
else if (Rs >= 2972) {
if (tcor < 69) {
if (Cs < 1817) {
return 1;
}
else if (Cs >= 1817) {
return 4;
}
}
else if (tcor >= 69) {
if (Rs < 2992) {
return 8;
}
else if (Rs >= 2992) {
return 4;
}
}
}
}
else if (RsCsDiff >= 383) {
if (Cs < 1417) {
if (Rs < 2431) {
return 2;
}
else if (Rs >= 2431) {
return 1;
}
}
else if (Cs >= 1417) {
if (posBalance < 59) {
if (Rs < 2390) {
return 2;
}
else if (Rs >= 2390) {
return 8;
}
}
else if (posBalance >= 59) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 26) {
if (AFD < 19) {
if (posBalance < 55) {
if (CsDiff < 8) {
return 2;
}
else if (CsDiff >= 8) {
if (gchDC < 1) {
return 2;
}
else if (gchDC >= 1) {
return 8;
}
}
}
else if (posBalance >= 55) {
return 8;
}
}
else if (AFD >= 19) {
if (Cs < 1757) {
if (MVDiff < 130) {
return 2;
}
else if (MVDiff >= 130) {
return 1;
}
}
else if (Cs >= 1757) {
if (Rs < 1224) {
return 2;
}
else if (Rs >= 1224) {
return 8;
}
}
}
}
}
}
}
else if (tcor >= 78) {
if (Rs < 2930) {
if (Rs < 1537) {
if (refDCval < 104) {
if (mu_mv_mag_sq < 3) {
if (SC < 2566) {
if (MVDiff < 18) {
return 1;
}
else if (MVDiff >= 18) {
return 8;
}
}
else if (SC >= 2566) {
if (RsCsDiff < 127) {
if (Rs < 1024) {
return 4;
}
else if (Rs >= 1024) {
return 2;
}
}
else if (RsCsDiff >= 127) {
if (Rs < 1395) {
return 1;
}
else if (Rs >= 1395) {
return 2;
}
}
}
}
else if (mu_mv_mag_sq >= 3) {
if (diffAFD < 4) {
if (CsDiff < 12) {
if (ssDCval < 66) {
if (Rs < 1065) {
return 2;
}
else if (Rs >= 1065) {
if (Rs < 1473) {
return 1;
}
else if (Rs >= 1473) {
return 4;
}
}
}
else if (ssDCval >= 66) {
if (gchDC < 4) {
if (Cs < 1330) {
if (Rs < 1169) {
return 8;
}
else if (Rs >= 1169) {
return 1;
}
}
else if (Cs >= 1330) {
return 8;
}
}
else if (gchDC >= 4) {
return 4;
}
}
}
else if (CsDiff >= 12) {
return 1;
}
}
else if (diffAFD >= 4) {
if (AFD < 16) {
return 8;
}
else if (AFD >= 16) {
return 2;
}
}
}
}
else if (refDCval >= 104) {
if (posBalance < 64) {
if (Rs < 1275) {
if (Rs < 1137) {
return 8;
}
else if (Rs >= 1137) {
return 4;
}
}
else if (Rs >= 1275) {
if (Cs < 1413) {
return 8;
}
else if (Cs >= 1413) {
return 2;
}
}
}
else if (posBalance >= 64) {
return 4;
}
}
}
else if (Rs >= 1537) {
if (Cs < 1641) {
if (TSC < 11) {
if (RsDiff < 6) {
return 1;
}
else if (RsDiff >= 6) {
if (AFD < 19) {
if (refDCval < 97) {
if (MVDiff < 6) {
if (negBalance < 46) {
return 4;
}
else if (negBalance >= 46) {
if (AFD < 13) {
if (Rs < 2585) {
return 8;
}
else if (Rs >= 2585) {
return 4;
}
}
else if (AFD >= 13) {
return 2;
}
}
}
else if (MVDiff >= 6) {
if (mu_mv_mag_sq < 61) {
if (Rs < 2400) {
if (TSC < 9) {
if (CsDiff < 8) {
if (ssDCval < 77) {
return 8;
}
else if (ssDCval >= 77) {
if (SC < 3590) {
return 2;
}
else if (SC >= 3590) {
return 4;
}
}
}
else if (CsDiff >= 8) {
if (ssDCval < 63) {
if (Rs < 2147) {
return 1;
}
else if (Rs >= 2147) {
return 2;
}
}
else if (ssDCval >= 63) {
if (diffAFD < 3) {
return 8;
}
else if (diffAFD >= 3) {
return 1;
}
}
}
}
else if (TSC >= 9) {
if (tcor < 89) {
if (posBalance < 44) {
return 2;
}
else if (posBalance >= 44) {
if (Rs < 1571) {
return 4;
}
else if (Rs >= 1571) {
if (ssDCval < 72) {
return 4;
}
else if (ssDCval >= 72) {
return 8;
}
}
}
}
else if (tcor >= 89) {
if (MVDiff < 16) {
return 1;
}
else if (MVDiff >= 16) {
return 8;
}
}
}
}
else if (Rs >= 2400) {
if (mcTcor < 90) {
if (Rs < 2714) {
return 1;
}
else if (Rs >= 2714) {
return 8;
}
}
else if (mcTcor >= 90) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 61) {
return 8;
}
}
}
else if (refDCval >= 97) {
if (mu_mv_mag_sq < 66) {
if (Cs < 1573) {
return 8;
}
else if (Cs >= 1573) {
if (MVDiff < 19) {
if (Rs < 2230) {
return 2;
}
else if (Rs >= 2230) {
return 8;
}
}
else if (MVDiff >= 19) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 66) {
if (Rs < 2042) {
if (Rs < 1900) {
return 2;
}
else if (Rs >= 1900) {
return 8;
}
}
else if (Rs >= 2042) {
return 4;
}
}
}
}
else if (AFD >= 19) {
if (ssDCval < 102) {
if (MVDiff < 20) {
if (ssDCval < 83) {
return 4;
}
else if (ssDCval >= 83) {
return 2;
}
}
else if (MVDiff >= 20) {
if (Rs < 2038) {
return 4;
}
else if (Rs >= 2038) {
return 8;
}
}
}
else if (ssDCval >= 102) {
if (Cs < 1316) {
return 8;
}
else if (Cs >= 1316) {
return 1;
}
}
}
}
}
else if (TSC >= 11) {
if (MVDiff < 27) {
if (Cs < 1420) {
if (AFD < 21) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 1;
}
}
else if (AFD >= 21) {
return 1;
}
}
else if (Cs >= 1420) {
return 8;
}
}
else if (MVDiff >= 27) {
if (posBalance < 55) {
if (diffRsCsdiff < -46) {
if (Rs < 2308) {
return 2;
}
else if (Rs >= 2308) {
return 4;
}
}
else if (diffRsCsdiff >= -46) {
return 8;
}
}
else if (posBalance >= 55) {
if (Rs < 2028) {
return 8;
}
else if (Rs >= 2028) {
return 1;
}
}
}
}
}
else if (Cs >= 1641) {
if (refDCval < 84) {
if (posBalance < 53) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (tcor < 79) {
if (Rs < 2641) {
return 1;
}
else if (Rs >= 2641) {
return 2;
}
}
else if (tcor >= 79) {
if (Rs < 2789) {
return 8;
}
else if (Rs >= 2789) {
return 1;
}
}
}
}
else if (posBalance >= 53) {
if (negBalance < 52) {
if (diffAFD < -1) {
if (Cs < 1756) {
return 2;
}
else if (Cs >= 1756) {
if (Rs < 2550) {
return 4;
}
else if (Rs >= 2550) {
return 8;
}
}
}
else if (diffAFD >= -1) {
if (MVDiff < 11) {
return 8;
}
else if (MVDiff >= 11) {
if (Rs < 2256) {
return 8;
}
else if (Rs >= 2256) {
if (MVDiff < 20) {
if (Rs < 2849) {
return 4;
}
else if (Rs >= 2849) {
return 8;
}
}
else if (MVDiff >= 20) {
return 8;
}
}
}
}
}
else if (negBalance >= 52) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (refDCval >= 84) {
if (mu_mv_mag_sq < 7) {
if (diffRsCsdiff < -50) {
if (posBalance < 62) {
if (Rs < 2734) {
return 8;
}
else if (Rs >= 2734) {
return 2;
}
}
else if (posBalance >= 62) {
return 2;
}
}
else if (diffRsCsdiff >= -50) {
if (ssDCval < 96) {
if (MVDiff < 7) {
if (Cs < 1702) {
return 8;
}
else if (Cs >= 1702) {
return 2;
}
}
else if (MVDiff >= 7) {
if (negBalance < 51) {
return 8;
}
else if (negBalance >= 51) {
if (Rs < 1937) {
return 8;
}
else if (Rs >= 1937) {
return 2;
}
}
}
}
else if (ssDCval >= 96) {
if (MVDiff < 25) {
return 8;
}
else if (MVDiff >= 25) {
if (Rs < 2097) {
return 8;
}
else if (Rs >= 2097) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 7) {
if (diffMVdiffVal < -47) {
return 2;
}
else if (diffMVdiffVal >= -47) {
if (RsDiff < 7) {
if (Rs < 1912) {
return 1;
}
else if (Rs >= 1912) {
return 8;
}
}
else if (RsDiff >= 7) {
if (posBalance < 69) {
return 8;
}
else if (posBalance >= 69) {
if (Rs < 2752) {
return 4;
}
else if (Rs >= 2752) {
return 8;
}
}
}
}
}
}
}
}
}
else if (Rs >= 2930) {
if (ssDCval < 108) {
if (negBalance < 50) {
if (MVDiff < 8) {
if (mu_mv_mag_sq < 86) {
return 2;
}
else if (mu_mv_mag_sq >= 86) {
if (Rs < 3120) {
return 2;
}
else if (Rs >= 3120) {
return 8;
}
}
}
else if (MVDiff >= 8) {
if (mu_mv_mag_sq < 68) {
if (Cs < 1796) {
if (negBalance < 47) {
if (Cs < 1364) {
return 4;
}
else if (Cs >= 1364) {
return 2;
}
}
else if (negBalance >= 47) {
return 4;
}
}
else if (Cs >= 1796) {
return 8;
}
}
else if (mu_mv_mag_sq >= 68) {
return 8;
}
}
}
else if (negBalance >= 50) {
if (ssDCval < 103) {
if (mu_mv_mag_sq < 25) {
if (Cs < 1541) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (Cs >= 1541) {
if (Rs < 3242) {
return 4;
}
else if (Rs >= 3242) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 25) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
return 8;
}
}
}
else if (ssDCval >= 103) {
return 4;
}
}
}
else if (ssDCval >= 108) {
if (posBalance < 36) {
if (Rs < 3187) {
return 8;
}
else if (Rs >= 3187) {
return 4;
}
}
else if (posBalance >= 36) {
if (SC < 4379) {
return 4;
}
else if (SC >= 4379) {
return 8;
}
}
}
}
}
}
else if (Rs >= 3333) {
if (refDCval < 101) {
if (mcTcor < 91) {
if (RsDiff < 17) {
if (posBalance < 49) {
if (RsCsDiff < 226) {
if (ssDCval < 64) {
return 2;
}
else if (ssDCval >= 64) {
if (Rs < 3997) {
return 4;
}
else if (Rs >= 3997) {
return 8;
}
}
}
else if (RsCsDiff >= 226) {
return 1;
}
}
else if (posBalance >= 49) {
if (negBalance < 44) {
if (Cs < 1752) {
return 2;
}
else if (Cs >= 1752) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
return 2;
}
}
}
else if (negBalance >= 44) {
if (AFD < 27) {
return 8;
}
else if (AFD >= 27) {
return 4;
}
}
}
}
else if (RsDiff >= 17) {
if (mu_mv_mag_sq < 14) {
if (Rs < 3388) {
return 8;
}
else if (Rs >= 3388) {
return 2;
}
}
else if (mu_mv_mag_sq >= 14) {
if (mcTcor < 85) {
return 4;
}
else if (mcTcor >= 85) {
if (posBalance < 40) {
if (Rs < 4024) {
return 8;
}
else if (Rs >= 4024) {
return 4;
}
}
else if (posBalance >= 40) {
if (Cs < 1883) {
return 2;
}
else if (Cs >= 1883) {
return 4;
}
}
}
}
}
}
else if (mcTcor >= 91) {
if (posBalance < 42) {
if (negBalance < 44) {
if (RsDiff < 9) {
return 1;
}
else if (RsDiff >= 9) {
if (Rs < 4586) {
if (Rs < 4276) {
if (ssDCval < 67) {
if (Cs < 1430) {
return 4;
}
else if (Cs >= 1430) {
return 2;
}
}
else if (ssDCval >= 67) {
return 4;
}
}
else if (Rs >= 4276) {
return 8;
}
}
else if (Rs >= 4586) {
return 4;
}
}
}
else if (negBalance >= 44) {
if (AFD < 18) {
return 2;
}
else if (AFD >= 18) {
return 1;
}
}
}
else if (posBalance >= 42) {
if (tcor < 81) {
if (diffTSC < -1) {
if (Rs < 3432) {
return 8;
}
else if (Rs >= 3432) {
return 2;
}
}
else if (diffTSC >= -1) {
if (ssDCval < 97) {
if (RsDiff < 16) {
return 4;
}
else if (RsDiff >= 16) {
if (Rs < 5109) {
return 2;
}
else if (Rs >= 5109) {
return 1;
}
}
}
else if (ssDCval >= 97) {
return 8;
}
}
}
else if (tcor >= 81) {
if (refDCval < 73) {
if (Rs < 7893) {
if (Rs < 4479) {
if (SC < 4836) {
return 4;
}
else if (SC >= 4836) {
return 2;
}
}
else if (Rs >= 4479) {
return 4;
}
}
else if (Rs >= 7893) {
return 2;
}
}
else if (refDCval >= 73) {
return 2;
}
}
}
}
}
else if (refDCval >= 101) {
if (mu_mv_mag_sq < 86) {
if (RsDiff < 12) {
if (mcTcor < 97) {
if (ssDCval < 138) {
if (mu_mv_mag_sq < 79) {
if (AFD < 19) {
if (CsDiff < 9) {
if (tcor < 69) {
return 1;
}
else if (tcor >= 69) {
return 8;
}
}
else if (CsDiff >= 9) {
if (Rs < 4467) {
return 8;
}
else if (Rs >= 4467) {
return 1;
}
}
}
else if (AFD >= 19) {
return 1;
}
}
else if (mu_mv_mag_sq >= 79) {
if (Rs < 4183) {
return 8;
}
else if (Rs >= 4183) {
return 1;
}
}
}
else if (ssDCval >= 138) {
return 4;
}
}
else if (mcTcor >= 97) {
if (Rs < 3906) {
return 8;
}
else if (Rs >= 3906) {
return 2;
}
}
}
else if (RsDiff >= 12) {
if (tcor < 86) {
if (TSC < 13) {
if (mu_mv_mag_sq < 15) {
if (RsDiff < 17) {
if (diffAFD < -2) {
if (Rs < 5015) {
return 8;
}
else if (Rs >= 5015) {
return 1;
}
}
else if (diffAFD >= -2) {
if (AFD < 15) {
if (Rs < 3575) {
return 4;
}
else if (Rs >= 3575) {
return 8;
}
}
else if (AFD >= 15) {
return 8;
}
}
}
else if (RsDiff >= 17) {
if (Rs < 5106) {
if (ssDCval < 135) {
if (MVDiff < 46) {
if (refDCval < 102) {
return 8;
}
else if (refDCval >= 102) {
return 2;
}
}
else if (MVDiff >= 46) {
return 8;
}
}
else if (ssDCval >= 135) {
return 1;
}
}
else if (Rs >= 5106) {
if (MVDiff < 32) {
if (ssDCval < 126) {
return 8;
}
else if (ssDCval >= 126) {
if (Cs < 1710) {
if (Rs < 6608) {
if (Rs < 6338) {
return 2;
}
else if (Rs >= 6338) {
return 8;
}
}
else if (Rs >= 6608) {
return 2;
}
}
else if (Cs >= 1710) {
return 8;
}
}
}
else if (MVDiff >= 32) {
return 8;
}
}
}
}
else if (mu_mv_mag_sq >= 15) {
if (mcTcor < 85) {
return 8;
}
else if (mcTcor >= 85) {
if (Rs < 3988) {
if (posBalance < 58) {
return 8;
}
else if (posBalance >= 58) {
return 2;
}
}
else if (Rs >= 3988) {
if (tcor < 76) {
if (Cs < 1376) {
return 2;
}
else if (Cs >= 1376) {
if (diffAFD < 2) {
return 4;
}
else if (diffAFD >= 2) {
return 1;
}
}
}
else if (tcor >= 76) {
if (RsDiff < 16) {
return 8;
}
else if (RsDiff >= 16) {
if (Rs < 4962) {
if (Cs < 1359) {
return 1;
}
else if (Cs >= 1359) {
return 4;
}
}
else if (Rs >= 4962) {
return 2;
}
}
}
}
}
}
}
else if (TSC >= 13) {
if (MVDiff < 44) {
if (tcor < 63) {
if (negBalance < 57) {
return 4;
}
else if (negBalance >= 57) {
return 2;
}
}
else if (tcor >= 63) {
if (Rs < 6260) {
if (Cs < 1523) {
return 1;
}
else if (Cs >= 1523) {
if (AFD < 26) {
return 8;
}
else if (AFD >= 26) {
if (Rs < 4677) {
return 2;
}
else if (Rs >= 4677) {
return 1;
}
}
}
}
else if (Rs >= 6260) {
return 2;
}
}
}
else if (MVDiff >= 44) {
if (CsDiff < 9) {
if (negBalance < 64) {
if (Rs < 3436) {
return 2;
}
else if (Rs >= 3436) {
return 4;
}
}
else if (negBalance >= 64) {
if (Rs < 3991) {
return 1;
}
else if (Rs >= 3991) {
return 2;
}
}
}
else if (CsDiff >= 9) {
if (diffMVdiffVal < -55) {
if (Rs < 5864) {
return 2;
}
else if (Rs >= 5864) {
return 4;
}
}
else if (diffMVdiffVal >= -55) {
return 1;
}
}
}
}
}
else if (tcor >= 86) {
if (MVDiff < 23) {
if (Cs < 1317) {
if (Rs < 5915) {
return 1;
}
else if (Rs >= 5915) {
return 2;
}
}
else if (Cs >= 1317) {
if (Cs < 1744) {
if (posBalance < 60) {
return 2;
}
else if (posBalance >= 60) {
if (Rs < 4074) {
return 4;
}
else if (Rs >= 4074) {
return 2;
}
}
}
else if (Cs >= 1744) {
if (Rs < 4338) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 2;
}
}
else if (Rs >= 4338) {
return 2;
}
}
}
}
else if (MVDiff >= 23) {
if (negBalance < 54) {
return 8;
}
else if (negBalance >= 54) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 86) {
if (Rs < 5954) {
if (Rs < 3455) {
return 4;
}
else if (Rs >= 3455) {
return 1;
}
}
else if (Rs >= 5954) {
return 4;
}
}
}
}
}
else if (Cs >= 1925) {
if (refDCval < 101) {
if (negBalance < 46) {
if (ssDCval < 70) {
if (mu_mv_mag_sq < 54) {
if (AFD < 19) {
if (RsDiff < 10) {
if (TSC < 8) {
if (Rs < 4208) {
return 1;
}
else if (Rs >= 4208) {
return 8;
}
}
else if (TSC >= 8) {
if (Rs < 642) {
return 4;
}
else if (Rs >= 642) {
return 2;
}
}
}
else if (RsDiff >= 10) {
if (ssDCval < 49) {
if (Cs < 2032) {
return 1;
}
else if (Cs >= 2032) {
return 2;
}
}
else if (ssDCval >= 49) {
if (mu_mv_mag_sq < 52) {
return 4;
}
else if (mu_mv_mag_sq >= 52) {
if (Rs < 5288) {
return 2;
}
else if (Rs >= 5288) {
return 4;
}
}
}
}
}
else if (AFD >= 19) {
if (CsDiff < 17) {
return 2;
}
else if (CsDiff >= 17) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 54) {
if (posBalance < 29) {
if (CsDiff < 4) {
return 4;
}
else if (CsDiff >= 4) {
return 8;
}
}
else if (posBalance >= 29) {
if (SC < 6265) {
if (mcTcor < 90) {
if (posBalance < 59) {
if (Cs < 1952) {
return 1;
}
else if (Cs >= 1952) {
if (mu_mv_mag_sq < 95) {
return 4;
}
else if (mu_mv_mag_sq >= 95) {
if (Rs < 2647) {
return 4;
}
else if (Rs >= 2647) {
return 8;
}
}
}
}
else if (posBalance >= 59) {
if (Rs < 2417) {
if (Rs < 1781) {
return 1;
}
else if (Rs >= 1781) {
return 4;
}
}
else if (Rs >= 2417) {
return 8;
}
}
}
else if (mcTcor >= 90) {
if (TSC < 9) {
if (mu_mv_mag_sq < 81) {
if (SC < 4042) {
return 8;
}
else if (SC >= 4042) {
return 1;
}
}
else if (mu_mv_mag_sq >= 81) {
if (negBalance < 35) {
if (RsCsDiff < 133) {
return 1;
}
else if (RsCsDiff >= 133) {
return 4;
}
}
else if (negBalance >= 35) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 4;
}
}
}
}
else if (TSC >= 9) {
if (Cs < 2880) {
return 8;
}
else if (Cs >= 2880) {
if (Rs < 1454) {
return 1;
}
else if (Rs >= 1454) {
return 4;
}
}
}
}
}
else if (SC >= 6265) {
if (diffTSC < 2) {
return 4;
}
else if (diffTSC >= 2) {
if (Rs < 3950) {
return 4;
}
else if (Rs >= 3950) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 70) {
if (Cs < 2247) {
if (tcor < 91) {
if (RsCsDiff < 167) {
if (AFD < 17) {
if (posBalance < 47) {
if (Rs < 3329) {
if (Cs < 2114) {
return 2;
}
else if (Cs >= 2114) {
return 1;
}
}
else if (Rs >= 3329) {
return 8;
}
}
else if (posBalance >= 47) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 4;
}
}
}
else if (AFD >= 17) {
return 1;
}
}
else if (RsCsDiff >= 167) {
if (MVDiff < 28) {
if (SC < 4773) {
if (Rs < 1862) {
if (Cs < 2116) {
return 2;
}
else if (Cs >= 2116) {
if (Rs < 1572) {
return 1;
}
else if (Rs >= 1572) {
return 4;
}
}
}
else if (Rs >= 1862) {
return 8;
}
}
else if (SC >= 4773) {
if (negBalance < 37) {
if (MVDiff < 12) {
return 2;
}
else if (MVDiff >= 12) {
if (AFD < 7) {
return 4;
}
else if (AFD >= 7) {
return 8;
}
}
}
else if (negBalance >= 37) {
if (mcTcor < 87) {
return 8;
}
else if (mcTcor >= 87) {
if (diffTSC < 1) {
return 2;
}
else if (diffTSC >= 1) {
if (RsCsDiff < 194) {
return 8;
}
else if (RsCsDiff >= 194) {
if (Rs < 2755) {
return 8;
}
else if (Rs >= 2755) {
return 2;
}
}
}
}
}
}
}
else if (MVDiff >= 28) {
if (refDCval < 73) {
return 4;
}
else if (refDCval >= 73) {
if (Rs < 4340) {
return 8;
}
else if (Rs >= 4340) {
return 1;
}
}
}
}
}
else if (tcor >= 91) {
if (RsDiff < 9) {
if (posBalance < 51) {
return 1;
}
else if (posBalance >= 51) {
return 8;
}
}
else if (RsDiff >= 9) {
if (AFD < 12) {
if (MVDiff < 14) {
return 2;
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (AFD >= 12) {
return 1;
}
}
}
}
else if (Cs >= 2247) {
if (posBalance < 52) {
if (ssDCval < 73) {
if (Cs < 3607) {
if (TSC < 8) {
return 1;
}
else if (TSC >= 8) {
if (posBalance < 49) {
if (Rs < 3735) {
return 2;
}
else if (Rs >= 3735) {
return 4;
}
}
else if (posBalance >= 49) {
if (Rs < 7201) {
return 2;
}
else if (Rs >= 7201) {
return 1;
}
}
}
}
else if (Cs >= 3607) {
if (mcTcor < 80) {
return 4;
}
else if (mcTcor >= 80) {
if (AFD < 6) {
return 2;
}
else if (AFD >= 6) {
return 8;
}
}
}
}
else if (ssDCval >= 73) {
if (ssDCval < 88) {
if (CsDiff < 15) {
if (Rs < 3297) {
if (Rs < 2855) {
return 8;
}
else if (Rs >= 2855) {
return 4;
}
}
else if (Rs >= 3297) {
if (CsDiff < 8) {
if (AFD < 7) {
return 2;
}
else if (AFD >= 7) {
if (Rs < 5644) {
return 8;
}
else if (Rs >= 5644) {
return 4;
}
}
}
else if (CsDiff >= 8) {
if (Rs < 3768) {
if (Rs < 3455) {
return 2;
}
else if (Rs >= 3455) {
return 4;
}
}
else if (Rs >= 3768) {
return 2;
}
}
}
}
else if (CsDiff >= 15) {
if (AFD < 20) {
if (TSC < 14) {
return 8;
}
else if (TSC >= 14) {
return 2;
}
}
else if (AFD >= 20) {
return 4;
}
}
}
else if (ssDCval >= 88) {
if (AFD < 16) {
if (mcTcor < 90) {
return 8;
}
else if (mcTcor >= 90) {
if (RsCsDiff < 206) {
if (TSC < 7) {
return 8;
}
else if (TSC >= 7) {
return 2;
}
}
else if (RsCsDiff >= 206) {
return 2;
}
}
}
else if (AFD >= 16) {
if (tcor < 78) {
if (Rs < 5573) {
return 2;
}
else if (Rs >= 5573) {
return 8;
}
}
else if (tcor >= 78) {
if (MVDiff < 7) {
return 2;
}
else if (MVDiff >= 7) {
return 1;
}
}
}
}
}
}
else if (posBalance >= 52) {
if (Cs < 2676) {
if (refDCval < 88) {
if (MVDiff < 30) {
if (AFD < 11) {
return 4;
}
else if (AFD >= 11) {
if (Cs < 2464) {
if (tcor < 84) {
return 8;
}
else if (tcor >= 84) {
if (Rs < 2834) {
return 8;
}
else if (Rs >= 2834) {
return 4;
}
}
}
else if (Cs >= 2464) {
return 8;
}
}
}
else if (MVDiff >= 30) {
return 4;
}
}
else if (refDCval >= 88) {
if (MVDiff < 10) {
if (Rs < 2980) {
return 8;
}
else if (Rs >= 2980) {
if (Rs < 4391) {
return 2;
}
else if (Rs >= 4391) {
if (Rs < 5131) {
return 1;
}
else if (Rs >= 5131) {
return 2;
}
}
}
}
else if (MVDiff >= 10) {
if (Rs < 3240) {
if (Cs < 2474) {
return 2;
}
else if (Cs >= 2474) {
return 8;
}
}
else if (Rs >= 3240) {
return 8;
}
}
}
}
else if (Cs >= 2676) {
if (tcor < 73) {
if (Rs < 1042) {
return 2;
}
else if (Rs >= 1042) {
if (RsDiff < 23) {
if (diffMVdiffVal < 2) {
if (Rs < 5917) {
return 8;
}
else if (Rs >= 5917) {
return 1;
}
}
else if (diffMVdiffVal >= 2) {
return 1;
}
}
else if (RsDiff >= 23) {
return 8;
}
}
}
else if (tcor >= 73) {
if (Cs < 4660) {
if (refDCval < 77) {
if (RsDiff < 8) {
if (Rs < 1826) {
return 2;
}
else if (Rs >= 1826) {
return 1;
}
}
else if (RsDiff >= 8) {
if (RsDiff < 12) {
return 8;
}
else if (RsDiff >= 12) {
if (Rs < 4984) {
return 4;
}
else if (Rs >= 4984) {
return 8;
}
}
}
}
else if (refDCval >= 77) {
if (Cs < 2904) {
if (TSC < 10) {
if (RsDiff < 12) {
return 4;
}
else if (RsDiff >= 12) {
return 8;
}
}
else if (TSC >= 10) {
if (CsDiff < 11) {
if (Rs < 4499) {
return 8;
}
else if (Rs >= 4499) {
return 2;
}
}
else if (CsDiff >= 11) {
if (Rs < 3798) {
return 4;
}
else if (Rs >= 3798) {
return 8;
}
}
}
}
else if (Cs >= 2904) {
if (diffMVdiffVal < -5) {
return 8;
}
else if (diffMVdiffVal >= -5) {
if (refDCval < 99) {
return 4;
}
else if (refDCval >= 99) {
return 8;
}
}
}
}
}
else if (Cs >= 4660) {
return 8;
}
}
}
}
}
}
}
else if (negBalance >= 46) {
if (mu_mv_mag_sq < 27) {
if (ssDCval < 72) {
if (MVDiff < 6) {
return 1;
}
else if (MVDiff >= 6) {
if (CsDiff < 11) {
if (AFD < 18) {
if (negBalance < 58) {
if (Rs < 2697) {
return 8;
}
else if (Rs >= 2697) {
return 4;
}
}
else if (negBalance >= 58) {
return 2;
}
}
else if (AFD >= 18) {
if (refDCval < 69) {
if (Rs < 1969) {
return 4;
}
else if (Rs >= 1969) {
if (Rs < 2259) {
return 1;
}
else if (Rs >= 2259) {
return 4;
}
}
}
else if (refDCval >= 69) {
if (Cs < 1958) {
return 4;
}
else if (Cs >= 1958) {
return 2;
}
}
}
}
else if (CsDiff >= 11) {
if (refDCval < 68) {
if (negBalance < 49) {
if (Rs < 2710) {
return 2;
}
else if (Rs >= 2710) {
if (Rs < 3833) {
return 4;
}
else if (Rs >= 3833) {
return 2;
}
}
}
else if (negBalance >= 49) {
if (negBalance < 57) {
if (mcTcor < 87) {
if (diffMVdiffVal < 4) {
return 4;
}
else if (diffMVdiffVal >= 4) {
if (Rs < 1947) {
return 2;
}
else if (Rs >= 1947) {
return 4;
}
}
}
else if (mcTcor >= 87) {
if (posBalance < 56) {
if (AFD < 20) {
if (Cs < 3591) {
return 4;
}
else if (Cs >= 3591) {
return 2;
}
}
else if (AFD >= 20) {
return 2;
}
}
else if (posBalance >= 56) {
if (Rs < 1435) {
return 8;
}
else if (Rs >= 1435) {
return 4;
}
}
}
}
else if (negBalance >= 57) {
if (ssDCval < 61) {
return 2;
}
else if (ssDCval >= 61) {
if (Cs < 2438) {
return 4;
}
else if (Cs >= 2438) {
return 2;
}
}
}
}
}
else if (refDCval >= 68) {
if (Rs < 3518) {
if (RsDiff < 13) {
if (SC < 4536) {
if (Rs < 1077) {
return 4;
}
else if (Rs >= 1077) {
return 8;
}
}
else if (SC >= 4536) {
return 1;
}
}
else if (RsDiff >= 13) {
if (MVDiff < 24) {
return 4;
}
else if (MVDiff >= 24) {
if (Rs < 2859) {
return 1;
}
else if (Rs >= 2859) {
return 2;
}
}
}
}
else if (Rs >= 3518) {
if (CsDiff < 15) {
return 2;
}
else if (CsDiff >= 15) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 72) {
if (RsDiff < 17) {
if (Rs < 4684) {
if (refDCval < 79) {
if (MVDiff < 4) {
return 2;
}
else if (MVDiff >= 4) {
if (Rs < 1483) {
if (AFD < 15) {
return 8;
}
else if (AFD >= 15) {
if (refDCval < 75) {
if (diffRsCsdiff < 31) {
return 4;
}
else if (diffRsCsdiff >= 31) {
return 2;
}
}
else if (refDCval >= 75) {
return 4;
}
}
}
else if (Rs >= 1483) {
if (RsCsDiff < 255) {
if (posBalance < 56) {
if (Rs < 2566) {
return 8;
}
else if (Rs >= 2566) {
if (Rs < 3385) {
return 4;
}
else if (Rs >= 3385) {
return 2;
}
}
}
else if (posBalance >= 56) {
return 8;
}
}
else if (RsCsDiff >= 255) {
if (TSC < 16) {
if (Rs < 3317) {
if (AFD < 21) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
return 8;
}
}
else if (AFD >= 21) {
if (MVDiff < 21) {
if (Rs < 3014) {
return 1;
}
else if (Rs >= 3014) {
return 8;
}
}
else if (MVDiff >= 21) {
if (AFD < 23) {
if (Rs < 2395) {
return 4;
}
else if (Rs >= 2395) {
return 2;
}
}
else if (AFD >= 23) {
return 8;
}
}
}
}
else if (Rs >= 3317) {
if (Rs < 3379) {
return 4;
}
else if (Rs >= 3379) {
return 1;
}
}
}
else if (TSC >= 16) {
return 4;
}
}
}
}
}
else if (refDCval >= 79) {
if (Rs < 3245) {
if (TSC < 14) {
if (mcTcor < 94) {
if (tcor < 55) {
if (tcor < 49) {
if (Rs < 1537) {
if (Rs < 1019) {
return 2;
}
else if (Rs >= 1019) {
return 8;
}
}
else if (Rs >= 1537) {
return 1;
}
}
else if (tcor >= 49) {
return 4;
}
}
else if (tcor >= 55) {
if (SC < 5136) {
if (CsDiff < 10) {
if (posBalance < 64) {
if (diffMVdiffVal < 3) {
if (Rs < 2972) {
if (refDCval < 86) {
return 8;
}
else if (refDCval >= 86) {
if (ssDCval < 92) {
if (SC < 4885) {
if (SC < 4151) {
return 8;
}
else if (SC >= 4151) {
return 4;
}
}
else if (SC >= 4885) {
return 8;
}
}
else if (ssDCval >= 92) {
return 8;
}
}
}
else if (Rs >= 2972) {
return 4;
}
}
else if (diffMVdiffVal >= 3) {
if (diffRsCsdiff < -2) {
return 4;
}
else if (diffRsCsdiff >= -2) {
return 8;
}
}
}
else if (posBalance >= 64) {
if (Rs < 1531) {
return 8;
}
else if (Rs >= 1531) {
return 4;
}
}
}
else if (CsDiff >= 10) {
if (AFD < 25) {
if (diffAFD < 2) {
if (negBalance < 49) {
return 4;
}
else if (negBalance >= 49) {
if (negBalance < 61) {
if (mcTcor < 79) {
if (Rs < 2174) {
return 4;
}
else if (Rs >= 2174) {
return 8;
}
}
else if (mcTcor >= 79) {
return 8;
}
}
else if (negBalance >= 61) {
return 4;
}
}
}
else if (diffAFD >= 2) {
if (Cs < 2337) {
if (Rs < 1125) {
return 2;
}
else if (Rs >= 1125) {
return 1;
}
}
else if (Cs >= 2337) {
return 8;
}
}
}
else if (AFD >= 25) {
if (Rs < 1379) {
return 2;
}
else if (Rs >= 1379) {
return 4;
}
}
}
}
else if (SC >= 5136) {
if (SC < 6071) {
if (MVDiff < 3) {
if (Cs < 3119) {
if (Rs < 2356) {
return 4;
}
else if (Rs >= 2356) {
return 8;
}
}
else if (Cs >= 3119) {
return 4;
}
}
else if (MVDiff >= 3) {
if (SC < 5493) {
if (Rs < 3126) {
if (posBalance < 61) {
if (ssDCval < 87) {
return 4;
}
else if (ssDCval >= 87) {
return 8;
}
}
else if (posBalance >= 61) {
return 8;
}
}
else if (Rs >= 3126) {
return 4;
}
}
else if (SC >= 5493) {
if (posBalance < 66) {
return 4;
}
else if (posBalance >= 66) {
return 8;
}
}
}
}
else if (SC >= 6071) {
if (ssDCval < 96) {
if (Rs < 3205) {
return 8;
}
else if (Rs >= 3205) {
return 2;
}
}
else if (ssDCval >= 96) {
return 4;
}
}
}
}
}
else if (mcTcor >= 94) {
if (Rs < 2988) {
if (refDCval < 96) {
if (diffRsCsdiff < -22) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
if (Rs < 2579) {
return 8;
}
else if (Rs >= 2579) {
return 2;
}
}
}
else if (diffRsCsdiff >= -22) {
if (gchDC < 3) {
if (diffAFD < -1) {
if (Rs < 2122) {
return 8;
}
else if (Rs >= 2122) {
return 2;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
else if (gchDC >= 3) {
return 2;
}
}
}
else if (refDCval >= 96) {
return 2;
}
}
else if (Rs >= 2988) {
if (ssDCval < 87) {
return 1;
}
else if (ssDCval >= 87) {
return 8;
}
}
}
}
else if (TSC >= 14) {
if (AFD < 22) {
if (mcTcor < 81) {
return 4;
}
else if (mcTcor >= 81) {
if (Rs < 2211) {
return 4;
}
else if (Rs >= 2211) {
if (MVDiff < 5) {
return 4;
}
else if (MVDiff >= 5) {
return 8;
}
}
}
}
else if (AFD >= 22) {
if (Cs < 3080) {
if (diffAFD < 2) {
if (Rs < 1672) {
return 8;
}
else if (Rs >= 1672) {
if (ssDCval < 83) {
if (Rs < 2880) {
return 1;
}
else if (Rs >= 2880) {
return 8;
}
}
else if (ssDCval >= 83) {
return 1;
}
}
}
else if (diffAFD >= 2) {
if (Cs < 1989) {
return 4;
}
else if (Cs >= 1989) {
return 8;
}
}
}
else if (Cs >= 3080) {
if (RsDiff < 10) {
if (Rs < 2332) {
return 4;
}
else if (Rs >= 2332) {
return 8;
}
}
else if (RsDiff >= 10) {
if (MVDiff < 54) {
return 8;
}
else if (MVDiff >= 54) {
return 2;
}
}
}
}
}
}
else if (Rs >= 3245) {
if (CsDiff < 14) {
if (AFD < 18) {
if (Cs < 2234) {
if (RsDiff < 13) {
return 8;
}
else if (RsDiff >= 13) {
return 2;
}
}
else if (Cs >= 2234) {
if (ssDCval < 86) {
if (Cs < 2743) {
return 4;
}
else if (Cs >= 2743) {
return 2;
}
}
else if (ssDCval >= 86) {
if (posBalance < 60) {
if (MVDiff < 14) {
if (negBalance < 56) {
return 4;
}
else if (negBalance >= 56) {
return 8;
}
}
else if (MVDiff >= 14) {
return 8;
}
}
else if (posBalance >= 60) {
if (Rs < 3311) {
return 2;
}
else if (Rs >= 3311) {
return 8;
}
}
}
}
}
else if (AFD >= 18) {
if (MVDiff < 9) {
return 4;
}
else if (MVDiff >= 9) {
if (MVDiff < 19) {
if (Rs < 3805) {
return 8;
}
else if (Rs >= 3805) {
if (Rs < 3850) {
return 4;
}
else if (Rs >= 3850) {
return 8;
}
}
}
else if (MVDiff >= 19) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
if (Rs < 4033) {
return 8;
}
else if (Rs >= 4033) {
return 2;
}
}
}
}
}
}
else if (CsDiff >= 14) {
return 8;
}
}
}
}
else if (Rs >= 4684) {
if (AFD < 15) {
return 2;
}
else if (AFD >= 15) {
if (diffTSC < -3) {
return 2;
}
else if (diffTSC >= -3) {
return 8;
}
}
}
}
else if (RsDiff >= 17) {
if (Rs < 5977) {
if (RsCsDiff < 526) {
if (diffRsCsdiff < 67) {
if (refDCval < 97) {
if (Rs < 3309) {
return 8;
}
else if (Rs >= 3309) {
return 4;
}
}
else if (refDCval >= 97) {
if (Rs < 4165) {
return 1;
}
else if (Rs >= 4165) {
return 2;
}
}
}
else if (diffRsCsdiff >= 67) {
return 2;
}
}
else if (RsCsDiff >= 526) {
if (Rs < 4002) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
return 1;
}
}
else if (Rs >= 4002) {
if (posBalance < 62) {
if (tcor < 79) {
return 2;
}
else if (tcor >= 79) {
if (Rs < 4722) {
return 2;
}
else if (Rs >= 4722) {
return 8;
}
}
}
else if (posBalance >= 62) {
if (Cs < 3003) {
return 1;
}
else if (Cs >= 3003) {
if (Rs < 4899) {
return 8;
}
else if (Rs >= 4899) {
return 4;
}
}
}
}
}
}
else if (Rs >= 5977) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 27) {
if (SC < 6977) {
if (Cs < 2726) {
if (RsCsDiff < 153) {
if (ssDCval < 51) {
if (Rs < 1281) {
return 2;
}
else if (Rs >= 1281) {
return 4;
}
}
else if (ssDCval >= 51) {
if (ssDCval < 83) {
if (mcTcor < 89) {
if (SC < 3946) {
if (Rs < 1542) {
return 4;
}
else if (Rs >= 1542) {
return 1;
}
}
else if (SC >= 3946) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
if (Rs < 2517) {
return 1;
}
else if (Rs >= 2517) {
return 8;
}
}
}
}
else if (mcTcor >= 89) {
if (Rs < 2708) {
if (Rs < 1569) {
return 4;
}
else if (Rs >= 1569) {
return 8;
}
}
else if (Rs >= 2708) {
if (MVDiff < 7) {
if (Rs < 3286) {
return 8;
}
else if (Rs >= 3286) {
return 4;
}
}
else if (MVDiff >= 7) {
return 4;
}
}
}
}
else if (ssDCval >= 83) {
if (diffMVdiffVal < -19) {
return 1;
}
else if (diffMVdiffVal >= -19) {
if (Rs < 4436) {
return 8;
}
else if (Rs >= 4436) {
return 1;
}
}
}
}
}
else if (RsCsDiff >= 153) {
if (refDCval < 54) {
if (mu_mv_mag_sq < 51) {
if (Rs < 1620) {
return 1;
}
else if (Rs >= 1620) {
return 8;
}
}
else if (mu_mv_mag_sq >= 51) {
return 4;
}
}
else if (refDCval >= 54) {
if (SC < 5773) {
if (AFD < 26) {
if (ssDCval < 97) {
if (Rs < 2328) {
if (mcTcor < 77) {
if (TSC < 15) {
return 2;
}
else if (TSC >= 15) {
return 4;
}
}
else if (mcTcor >= 77) {
if (negBalance < 57) {
if (mu_mv_mag_sq < 55) {
return 8;
}
else if (mu_mv_mag_sq >= 55) {
if (Rs < 2009) {
return 2;
}
else if (Rs >= 2009) {
return 1;
}
}
}
else if (negBalance >= 57) {
if (TSC < 15) {
return 4;
}
else if (TSC >= 15) {
return 8;
}
}
}
}
else if (Rs >= 2328) {
if (posBalance < 42) {
if (Rs < 3569) {
if (Rs < 3080) {
return 1;
}
else if (Rs >= 3080) {
return 8;
}
}
else if (Rs >= 3569) {
return 2;
}
}
else if (posBalance >= 42) {
if (diffMVdiffVal < -3) {
if (SC < 5108) {
if (AFD < 17) {
return 8;
}
else if (AFD >= 17) {
if (Rs < 2582) {
return 2;
}
else if (Rs >= 2582) {
return 8;
}
}
}
else if (SC >= 5108) {
if (RsDiff < 10) {
return 4;
}
else if (RsDiff >= 10) {
if (tcor < 77) {
if (Rs < 2686) {
return 8;
}
else if (Rs >= 2686) {
return 4;
}
}
else if (tcor >= 77) {
return 8;
}
}
}
}
else if (diffMVdiffVal >= -3) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 97) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (Cs < 2110) {
return 8;
}
else if (Cs >= 2110) {
return 2;
}
}
}
}
else if (AFD >= 26) {
if (Rs < 2557) {
return 4;
}
else if (Rs >= 2557) {
return 1;
}
}
}
else if (SC >= 5773) {
if (tcor < 76) {
if (AFD < 29) {
return 8;
}
else if (AFD >= 29) {
return 1;
}
}
else if (tcor >= 76) {
if (CsDiff < 9) {
if (AFD < 15) {
if (diffRsCsdiff < -10) {
return 2;
}
else if (diffRsCsdiff >= -10) {
return 8;
}
}
else if (AFD >= 15) {
return 4;
}
}
else if (CsDiff >= 9) {
if (RsDiff < 12) {
return 4;
}
else if (RsDiff >= 12) {
if (mu_mv_mag_sq < 66) {
if (negBalance < 50) {
return 8;
}
else if (negBalance >= 50) {
if (TSC < 14) {
if (TSC < 12) {
return 8;
}
else if (TSC >= 12) {
return 4;
}
}
else if (TSC >= 14) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 66) {
return 8;
}
}
}
}
}
}
}
}
else if (Cs >= 2726) {
if (refDCval < 81) {
if (mcTcor < 71) {
return 4;
}
else if (mcTcor >= 71) {
if (negBalance < 49) {
if (tcor < 81) {
return 8;
}
else if (tcor >= 81) {
if (Rs < 2916) {
return 4;
}
else if (Rs >= 2916) {
return 1;
}
}
}
else if (negBalance >= 49) {
if (posBalance < 45) {
return 2;
}
else if (posBalance >= 45) {
return 8;
}
}
}
}
else if (refDCval >= 81) {
if (Rs < 2080) {
return 8;
}
else if (Rs >= 2080) {
if (diffAFD < -2) {
if (Rs < 3059) {
return 4;
}
else if (Rs >= 3059) {
if (Rs < 3744) {
return 1;
}
else if (Rs >= 3744) {
if (Rs < 3865) {
return 4;
}
else if (Rs >= 3865) {
return 8;
}
}
}
}
else if (diffAFD >= -2) {
if (ssDCval < 91) {
if (AFD < 16) {
return 4;
}
else if (AFD >= 16) {
if (Rs < 2711) {
return 4;
}
else if (Rs >= 2711) {
return 8;
}
}
}
else if (ssDCval >= 91) {
return 4;
}
}
}
}
}
}
else if (SC >= 6977) {
if (Rs < 6019) {
if (negBalance < 50) {
if (TSC < 16) {
if (Cs < 3872) {
if (mcTcor < 84) {
return 8;
}
else if (mcTcor >= 84) {
if (negBalance < 47) {
if (TSC < 11) {
return 2;
}
else if (TSC >= 11) {
if (Rs < 5220) {
return 8;
}
else if (Rs >= 5220) {
return 1;
}
}
}
else if (negBalance >= 47) {
if (diffAFD < 1) {
if (posBalance < 61) {
return 4;
}
else if (posBalance >= 61) {
if (Rs < 4949) {
return 2;
}
else if (Rs >= 4949) {
return 4;
}
}
}
else if (diffAFD >= 1) {
if (CsDiff < 13) {
if (Cs < 2873) {
if (Rs < 5024) {
return 8;
}
else if (Rs >= 5024) {
return 4;
}
}
else if (Cs >= 2873) {
return 4;
}
}
else if (CsDiff >= 13) {
return 8;
}
}
}
}
}
else if (Cs >= 3872) {
if (Cs < 5464) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 8;
}
}
else if (Cs >= 5464) {
return 4;
}
}
}
else if (TSC >= 16) {
if (Rs < 4243) {
return 8;
}
else if (Rs >= 4243) {
return 2;
}
}
}
else if (negBalance >= 50) {
if (Rs < 3179) {
if (ssDCval < 95) {
return 8;
}
else if (ssDCval >= 95) {
return 4;
}
}
else if (Rs >= 3179) {
if (RsDiff < 17) {
if (TSC < 12) {
if (ssDCval < 84) {
return 8;
}
else if (ssDCval >= 84) {
if (Rs < 4212) {
if (Rs < 4023) {
return 2;
}
else if (Rs >= 4023) {
return 4;
}
}
else if (Rs >= 4212) {
return 8;
}
}
}
else if (TSC >= 12) {
if (diffAFD < 5) {
return 8;
}
else if (diffAFD >= 5) {
return 1;
}
}
}
else if (RsDiff >= 17) {
if (TSC < 17) {
if (Cs < 2780) {
return 8;
}
else if (Cs >= 2780) {
if (Rs < 5163) {
return 4;
}
else if (Rs >= 5163) {
if (Cs < 3259) {
return 4;
}
else if (Cs >= 3259) {
return 2;
}
}
}
}
else if (TSC >= 17) {
if (RsDiff < 27) {
return 8;
}
else if (RsDiff >= 27) {
return 4;
}
}
}
}
}
}
else if (Rs >= 6019) {
if (refDCval < 100) {
if (RsDiff < 11) {
return 8;
}
else if (RsDiff >= 11) {
if (Rs < 6117) {
if (Rs < 6053) {
return 4;
}
else if (Rs >= 6053) {
if (Rs < 6082) {
return 8;
}
else if (Rs >= 6082) {
return 1;
}
}
}
else if (Rs >= 6117) {
return 4;
}
}
}
else if (refDCval >= 100) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 101) {
if (RsDiff < 9) {
if (mu_mv_mag_sq < 87) {
if (Rs < 5080) {
if (negBalance < 59) {
if (MVDiff < 2) {
if (Rs < 3601) {
return 1;
}
else if (Rs >= 3601) {
return 8;
}
}
else if (MVDiff >= 2) {
if (negBalance < 44) {
if (tcor < 89) {
if (mu_mv_mag_sq < 2) {
return 1;
}
else if (mu_mv_mag_sq >= 2) {
if (posBalance < 40) {
if (negBalance < 26) {
if (Cs < 6320) {
return 4;
}
else if (Cs >= 6320) {
return 8;
}
}
else if (negBalance >= 26) {
return 1;
}
}
else if (posBalance >= 40) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
return 1;
}
}
}
}
else if (tcor >= 89) {
return 8;
}
}
else if (negBalance >= 44) {
if (mu_mv_mag_sq < 44) {
if (tcor < 60) {
return 4;
}
else if (tcor >= 60) {
if (SC < 3955) {
if (RsCsDiff < 115) {
if (MVDiff < 5) {
if (Rs < 1648) {
return 8;
}
else if (Rs >= 1648) {
return 4;
}
}
else if (MVDiff >= 5) {
return 2;
}
}
else if (RsCsDiff >= 115) {
return 8;
}
}
else if (SC >= 3955) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 44) {
if (posBalance < 57) {
if (Cs < 4326) {
return 8;
}
else if (Cs >= 4326) {
return 4;
}
}
else if (posBalance >= 57) {
if (SC < 5678) {
if (Cs < 2166) {
return 8;
}
else if (Cs >= 2166) {
if (Rs < 1771) {
return 8;
}
else if (Rs >= 1771) {
return 4;
}
}
}
else if (SC >= 5678) {
if (Rs < 2670) {
return 4;
}
else if (Rs >= 2670) {
if (Rs < 4755) {
return 1;
}
else if (Rs >= 4755) {
return 8;
}
}
}
}
}
}
}
}
else if (negBalance >= 59) {
if (tcor < 76) {
if (diffTSC < -2) {
return 1;
}
else if (diffTSC >= -2) {
return 2;
}
}
else if (tcor >= 76) {
if (Cs < 2185) {
return 1;
}
else if (Cs >= 2185) {
return 8;
}
}
}
}
else if (Rs >= 5080) {
return 1;
}
}
else if (mu_mv_mag_sq >= 87) {
if (negBalance < 45) {
if (RsDiff < 7) {
if (Cs < 4481) {
return 1;
}
else if (Cs >= 4481) {
return 8;
}
}
else if (RsDiff >= 7) {
if (diffRsCsdiff < 16) {
return 8;
}
else if (diffRsCsdiff >= 16) {
if (MVDiff < 2) {
return 4;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
}
else if (negBalance >= 45) {
if (tcor < 94) {
return 1;
}
else if (tcor >= 94) {
if (RsDiff < 8) {
return 4;
}
else if (RsDiff >= 8) {
return 8;
}
}
}
}
}
else if (RsDiff >= 9) {
if (refDCval < 154) {
if (TSC < 8) {
if (SC < 9073) {
if (ssDCval < 113) {
if (negBalance < 45) {
return 2;
}
else if (negBalance >= 45) {
return 8;
}
}
else if (ssDCval >= 113) {
if (RsCsDiff < 243) {
if (TSC < 7) {
if (TSC < 5) {
return 8;
}
else if (TSC >= 5) {
return 4;
}
}
else if (TSC >= 7) {
if (diffAFD < -1) {
if (SC < 5153) {
return 8;
}
else if (SC >= 5153) {
return 4;
}
}
else if (diffAFD >= -1) {
return 8;
}
}
}
else if (RsCsDiff >= 243) {
if (tcor < 71) {
return 2;
}
else if (tcor >= 71) {
if (ssDCval < 118) {
return 2;
}
else if (ssDCval >= 118) {
if (ssDCval < 140) {
return 4;
}
else if (ssDCval >= 140) {
return 8;
}
}
}
}
}
}
else if (SC >= 9073) {
if (ssDCval < 129) {
if (Rs < 6468) {
if (mu_mv_mag_sq < 80) {
if (MVDiff < 4) {
return 4;
}
else if (MVDiff >= 4) {
if (Rs < 5553) {
return 8;
}
else if (Rs >= 5553) {
return 1;
}
}
}
else if (mu_mv_mag_sq >= 80) {
return 8;
}
}
else if (Rs >= 6468) {
if (ssDCval < 110) {
if (Rs < 8410) {
if (Cs < 3987) {
return 1;
}
else if (Cs >= 3987) {
return 8;
}
}
else if (Rs >= 8410) {
return 1;
}
}
else if (ssDCval >= 110) {
return 1;
}
}
}
else if (ssDCval >= 129) {
if (Cs < 3816) {
return 2;
}
else if (Cs >= 3816) {
return 8;
}
}
}
}
else if (TSC >= 8) {
if (Rs < 5323) {
if (AFD < 22) {
if (refDCval < 133) {
if (mcTcor < 94) {
if (Rs < 4385) {
if (mcTcor < 86) {
if (refDCval < 129) {
if (diffMVdiffVal < 6) {
return 8;
}
else if (diffMVdiffVal >= 6) {
if (Cs < 3307) {
if (Cs < 1949) {
return 4;
}
else if (Cs >= 1949) {
if (AFD < 21) {
return 8;
}
else if (AFD >= 21) {
if (Rs < 3185) {
return 8;
}
else if (Rs >= 3185) {
return 1;
}
}
}
}
else if (Cs >= 3307) {
return 4;
}
}
}
else if (refDCval >= 129) {
if (Cs < 3085) {
if (Rs < 2828) {
return 2;
}
else if (Rs >= 2828) {
return 8;
}
}
else if (Cs >= 3085) {
if (Rs < 3117) {
return 1;
}
else if (Rs >= 3117) {
return 4;
}
}
}
}
else if (mcTcor >= 86) {
if (Cs < 2691) {
if (Rs < 3466) {
if (diffAFD < 10) {
if (RsDiff < 12) {
if (SC < 4786) {
return 8;
}
else if (SC >= 4786) {
if (tcor < 85) {
if (ssDCval < 111) {
return 4;
}
else if (ssDCval >= 111) {
if (posBalance < 51) {
return 4;
}
else if (posBalance >= 51) {
return 8;
}
}
}
else if (tcor >= 85) {
if (diffRsCsdiff < -68) {
return 4;
}
else if (diffRsCsdiff >= -68) {
return 8;
}
}
}
}
else if (RsDiff >= 12) {
if (SC < 4947) {
if (TSC < 10) {
return 2;
}
else if (TSC >= 10) {
return 8;
}
}
else if (SC >= 4947) {
if (Cs < 2262) {
return 1;
}
else if (Cs >= 2262) {
return 2;
}
}
}
}
else if (diffAFD >= 10) {
if (Rs < 2891) {
return 4;
}
else if (Rs >= 2891) {
return 1;
}
}
}
else if (Rs >= 3466) {
if (mu_mv_mag_sq < 67) {
return 8;
}
else if (mu_mv_mag_sq >= 67) {
if (ssDCval < 104) {
if (Rs < 3985) {
if (MVDiff < 10) {
return 2;
}
else if (MVDiff >= 10) {
return 8;
}
}
else if (Rs >= 3985) {
return 1;
}
}
else if (ssDCval >= 104) {
return 8;
}
}
}
}
else if (Cs >= 2691) {
if (RsCsDiff < 255) {
if (mu_mv_mag_sq < 103) {
if (TSC < 11) {
if (Rs < 3728) {
if (CsDiff < 10) {
return 4;
}
else if (CsDiff >= 10) {
return 8;
}
}
else if (Rs >= 3728) {
if (Cs < 3627) {
return 8;
}
else if (Cs >= 3627) {
return 1;
}
}
}
else if (TSC >= 11) {
return 4;
}
}
else if (mu_mv_mag_sq >= 103) {
return 8;
}
}
else if (RsCsDiff >= 255) {
if (MVDiff < 15) {
if (diffAFD < 2) {
return 8;
}
else if (diffAFD >= 2) {
if (Rs < 3332) {
return 4;
}
else if (Rs >= 3332) {
return 8;
}
}
}
else if (MVDiff >= 15) {
if (Rs < 2820) {
return 4;
}
else if (Rs >= 2820) {
return 2;
}
}
}
}
}
}
else if (Rs >= 4385) {
if (RsDiff < 10) {
if (mu_mv_mag_sq < 77) {
return 8;
}
else if (mu_mv_mag_sq >= 77) {
if (Cs < 2841) {
return 1;
}
else if (Cs >= 2841) {
return 8;
}
}
}
else if (RsDiff >= 10) {
if (posBalance < 62) {
if (diffAFD < -5) {
return 4;
}
else if (diffAFD >= -5) {
return 8;
}
}
else if (posBalance >= 62) {
if (Rs < 4733) {
return 2;
}
else if (Rs >= 4733) {
return 1;
}
}
}
}
}
else if (mcTcor >= 94) {
if (refDCval < 109) {
if (Rs < 4809) {
if (RsCsDiff < 189) {
return 8;
}
else if (RsCsDiff >= 189) {
if (Cs < 3210) {
return 2;
}
else if (Cs >= 3210) {
return 8;
}
}
}
else if (Rs >= 4809) {
return 4;
}
}
else if (refDCval >= 109) {
if (posBalance < 52) {
if (SC < 6509) {
return 8;
}
else if (SC >= 6509) {
if (ssDCval < 119) {
if (SC < 6836) {
return 2;
}
else if (SC >= 6836) {
return 1;
}
}
else if (ssDCval >= 119) {
return 2;
}
}
}
else if (posBalance >= 52) {
if (Cs < 1943) {
return 1;
}
else if (Cs >= 1943) {
return 8;
}
}
}
}
}
else if (refDCval >= 133) {
if (tcor < 92) {
if (mu_mv_mag_sq < 83) {
if (Rs < 3847) {
if (refDCval < 134) {
if (mu_mv_mag_sq < 42) {
if (Rs < 2396) {
return 1;
}
else if (Rs >= 2396) {
if (Rs < 3183) {
return 2;
}
else if (Rs >= 3183) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 42) {
return 4;
}
}
else if (refDCval >= 134) {
if (CsDiff < 8) {
if (mu_mv_mag_sq < 41) {
return 1;
}
else if (mu_mv_mag_sq >= 41) {
if (Cs < 2463) {
return 8;
}
else if (Cs >= 2463) {
return 4;
}
}
}
else if (CsDiff >= 8) {
if (Rs < 2068) {
if (Rs < 1932) {
return 8;
}
else if (Rs >= 1932) {
return 2;
}
}
else if (Rs >= 2068) {
return 8;
}
}
}
}
else if (Rs >= 3847) {
if (tcor < 71) {
if (negBalance < 42) {
return 8;
}
else if (negBalance >= 42) {
if (mu_mv_mag_sq < 26) {
if (Cs < 2954) {
return 4;
}
else if (Cs >= 2954) {
return 8;
}
}
else if (mu_mv_mag_sq >= 26) {
if (negBalance < 45) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
return 8;
}
}
else if (negBalance >= 45) {
return 1;
}
}
}
}
else if (tcor >= 71) {
if (refDCval < 144) {
return 4;
}
else if (refDCval >= 144) {
if (Rs < 4189) {
return 8;
}
else if (Rs >= 4189) {
return 1;
}
}
}
}
}
else if (mu_mv_mag_sq >= 83) {
return 1;
}
}
else if (tcor >= 92) {
if (Cs < 1989) {
return 2;
}
else if (Cs >= 1989) {
if (RsCsDiff < 133) {
return 1;
}
else if (RsCsDiff >= 133) {
return 8;
}
}
}
}
}
else if (AFD >= 22) {
if (negBalance < 57) {
if (gchDC < 2) {
if (Rs < 3634) {
if (posBalance < 57) {
if (RsDiff < 18) {
return 8;
}
else if (RsDiff >= 18) {
return 4;
}
}
else if (posBalance >= 57) {
if (CsDiff < 12) {
if (Rs < 3436) {
return 4;
}
else if (Rs >= 3436) {
return 8;
}
}
else if (CsDiff >= 12) {
if (RsCsDiff < 587) {
return 2;
}
else if (RsCsDiff >= 587) {
if (Rs < 2881) {
return 4;
}
else if (Rs >= 2881) {
return 8;
}
}
}
}
}
else if (Rs >= 3634) {
if (RsCsDiff < 619) {
return 4;
}
else if (RsCsDiff >= 619) {
if (diffAFD < 2) {
if (Cs < 1979) {
return 2;
}
else if (Cs >= 1979) {
return 4;
}
}
else if (diffAFD >= 2) {
return 2;
}
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 107) {
if (ssDCval < 106) {
return 2;
}
else if (ssDCval >= 106) {
if (CsDiff < 12) {
return 4;
}
else if (CsDiff >= 12) {
return 1;
}
}
}
else if (ssDCval >= 107) {
if (SC < 6495) {
if (mcTcor < 90) {
if (RsDiff < 10) {
return 2;
}
else if (RsDiff >= 10) {
return 8;
}
}
else if (mcTcor >= 90) {
if (RsDiff < 14) {
return 8;
}
else if (RsDiff >= 14) {
return 2;
}
}
}
else if (SC >= 6495) {
if (SC < 6655) {
return 1;
}
else if (SC >= 6655) {
if (Rs < 5136) {
if (negBalance < 56) {
return 8;
}
else if (negBalance >= 56) {
return 4;
}
}
else if (Rs >= 5136) {
return 1;
}
}
}
}
}
}
else if (negBalance >= 57) {
if (posBalance < 57) {
if (RsDiff < 25) {
if (negBalance < 64) {
return 8;
}
else if (negBalance >= 64) {
if (negBalance < 72) {
if (Cs < 2040) {
return 1;
}
else if (Cs >= 2040) {
return 2;
}
}
else if (negBalance >= 72) {
return 8;
}
}
}
else if (RsDiff >= 25) {
if (CsDiff < 23) {
return 4;
}
else if (CsDiff >= 23) {
if (Rs < 3375) {
return 8;
}
else if (Rs >= 3375) {
return 2;
}
}
}
}
else if (posBalance >= 57) {
if (Rs < 4561) {
if (Rs < 3179) {
if (mu_mv_mag_sq < 8) {
if (Rs < 2760) {
return 2;
}
else if (Rs >= 2760) {
if (Rs < 2991) {
return 1;
}
else if (Rs >= 2991) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 8) {
if (MVDiff < 15) {
return 1;
}
else if (MVDiff >= 15) {
if (Rs < 2095) {
return 1;
}
else if (Rs >= 2095) {
if (TSC < 13) {
return 1;
}
else if (TSC >= 13) {
return 8;
}
}
}
}
}
else if (Rs >= 3179) {
if (tcor < 76) {
if (Rs < 3219) {
if (Cs < 2944) {
return 4;
}
else if (Cs >= 2944) {
return 1;
}
}
else if (Rs >= 3219) {
if (diffRsCsdiff < 148) {
if (ssDCval < 134) {
return 1;
}
else if (ssDCval >= 134) {
return 8;
}
}
else if (diffRsCsdiff >= 148) {
if (Rs < 3607) {
return 1;
}
else if (Rs >= 3607) {
return 4;
}
}
}
}
else if (tcor >= 76) {
return 8;
}
}
}
else if (Rs >= 4561) {
return 8;
}
}
}
}
}
else if (Rs >= 5323) {
if (TSC < 21) {
if (RsDiff < 17) {
if (mcTcor < 92) {
if (Cs < 2365) {
return 1;
}
else if (Cs >= 2365) {
return 8;
}
}
else if (mcTcor >= 92) {
if (tcor < 93) {
if (MVDiff < 10) {
return 4;
}
else if (MVDiff >= 10) {
if (RsDiff < 11) {
return 4;
}
else if (RsDiff >= 11) {
return 8;
}
}
}
else if (tcor >= 93) {
if (Rs < 6889) {
return 8;
}
else if (Rs >= 6889) {
return 2;
}
}
}
}
else if (RsDiff >= 17) {
if (negBalance < 46) {
if (Cs < 4472) {
if (Rs < 5692) {
return 8;
}
else if (Rs >= 5692) {
return 2;
}
}
else if (Cs >= 4472) {
if (mu_mv_mag_sq < 22) {
return 1;
}
else if (mu_mv_mag_sq >= 22) {
return 8;
}
}
}
else if (negBalance >= 46) {
if (gchDC < 2) {
if (RsCsDiff < 716) {
if (MVDiff < 17) {
if (diffRsCsdiff < 118) {
return 4;
}
else if (diffRsCsdiff >= 118) {
if (Rs < 5535) {
return 4;
}
else if (Rs >= 5535) {
return 8;
}
}
}
else if (MVDiff >= 17) {
return 8;
}
}
else if (RsCsDiff >= 716) {
if (refDCval < 150) {
return 4;
}
else if (refDCval >= 150) {
if (Rs < 7253) {
return 2;
}
else if (Rs >= 7253) {
return 4;
}
}
}
}
else if (gchDC >= 2) {
if (ssDCval < 103) {
return 4;
}
else if (ssDCval >= 103) {
return 8;
}
}
}
}
}
else if (TSC >= 21) {
if (Cs < 4873) {
return 4;
}
else if (Cs >= 4873) {
if (ssDCval < 110) {
return 4;
}
else if (ssDCval >= 110) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 154) {
if (posBalance < 53) {
if (negBalance < 30) {
return 8;
}
else if (negBalance >= 30) {
if (RsCsDiff < 592) {
if (Cs < 2798) {
if (ssDCval < 156) {
return 8;
}
else if (ssDCval >= 156) {
return 4;
}
}
else if (Cs >= 2798) {
return 1;
}
}
else if (RsCsDiff >= 592) {
if (Cs < 3154) {
return 2;
}
else if (Cs >= 3154) {
return 8;
}
}
}
}
else if (posBalance >= 53) {
if (ssDCval < 168) {
if (Rs < 3366) {
return 8;
}
else if (Rs >= 3366) {
if (diffTSC < 3) {
return 4;
}
else if (diffTSC >= 3) {
if (Rs < 7369) {
return 8;
}
else if (Rs >= 7369) {
return 4;
}
}
}
}
else if (ssDCval >= 168) {
return 2;
}
}
}
}
}
}
}
}
else if (mu_mv_mag_sq >= 111) {
if (Cs < 1664) {
if (Rs < 3143) {
if (Rs < 1488) {
if (ssDCval < 122) {
if (negBalance < 1) {
return 2;
}
else if (negBalance >= 1) {
if (ssDCval < 63) {
if (AFD < 3) {
if (mu_mv_mag_sq < 127) {
if (SC < 2543) {
if (Rs < 1091) {
return 1;
}
else if (Rs >= 1091) {
if (Rs < 1102) {
return 8;
}
else if (Rs >= 1102) {
return 1;
}
}
}
else if (SC >= 2543) {
if (posBalance < 42) {
if (ssDCval < 47) {
return 8;
}
else if (ssDCval >= 47) {
if (Cs < 1642) {
return 1;
}
else if (Cs >= 1642) {
if (Rs < 1027) {
return 8;
}
else if (Rs >= 1027) {
return 4;
}
}
}
}
else if (posBalance >= 42) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (Cs < 1374) {
if (negBalance < 18) {
return 8;
}
else if (negBalance >= 18) {
if (Rs < 1159) {
return 8;
}
else if (Rs >= 1159) {
return 4;
}
}
}
else if (Cs >= 1374) {
if (RsDiff < 1) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
return 1;
}
}
else if (RsDiff >= 1) {
return 8;
}
}
}
}
else if (AFD >= 3) {
if (MVDiff < 2) {
if (mcTcor < 98) {
return 8;
}
else if (mcTcor >= 98) {
if (Rs < 1227) {
return 8;
}
else if (Rs >= 1227) {
return 4;
}
}
}
else if (MVDiff >= 2) {
return 4;
}
}
}
else if (ssDCval >= 63) {
if (SC < 1938) {
if (SC < 1821) {
if (Rs < 322) {
return 4;
}
else if (Rs >= 322) {
return 8;
}
}
else if (SC >= 1821) {
return 1;
}
}
else if (SC >= 1938) {
if (Cs < 1402) {
if (RsDiff < 1) {
return 1;
}
else if (RsDiff >= 1) {
if (ssDCval < 73) {
if (Rs < 1186) {
if (Rs < 1028) {
return 2;
}
else if (Rs >= 1028) {
if (Rs < 1150) {
return 4;
}
else if (Rs >= 1150) {
return 8;
}
}
}
else if (Rs >= 1186) {
if (negBalance < 40) {
return 4;
}
else if (negBalance >= 40) {
return 1;
}
}
}
else if (ssDCval >= 73) {
if (Rs < 1181) {
return 4;
}
else if (Rs >= 1181) {
return 8;
}
}
}
}
else if (Cs >= 1402) {
if (negBalance < 36) {
if (diffRsCsdiff < -7) {
return 4;
}
else if (diffRsCsdiff >= -7) {
return 8;
}
}
else if (negBalance >= 36) {
if (Rs < 919) {
return 1;
}
else if (Rs >= 919) {
if (Rs < 1252) {
if (Rs < 1083) {
return 2;
}
else if (Rs >= 1083) {
return 4;
}
}
else if (Rs >= 1252) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 122) {
if (RsCsDiff < 2) {
if (ssDCval < 127) {
if (AFD < 1) {
return 8;
}
else if (AFD >= 1) {
return 4;
}
}
else if (ssDCval >= 127) {
return 8;
}
}
else if (RsCsDiff >= 2) {
if (refDCval < 133) {
if (Rs < 1251) {
if (Rs < 955) {
return 8;
}
else if (Rs >= 955) {
if (ssDCval < 129) {
if (mu_mv_mag_sq < 116) {
if (Rs < 1144) {
return 4;
}
else if (Rs >= 1144) {
return 8;
}
}
else if (mu_mv_mag_sq >= 116) {
return 4;
}
}
else if (ssDCval >= 129) {
if (SC < 2711) {
return 8;
}
else if (SC >= 2711) {
return 4;
}
}
}
}
else if (Rs >= 1251) {
return 8;
}
}
else if (refDCval >= 133) {
if (Rs < 1092) {
return 8;
}
else if (Rs >= 1092) {
return 1;
}
}
}
}
}
else if (Rs >= 1488) {
if (ssDCval < 86) {
if (tcor < 98) {
if (Rs < 2681) {
if (negBalance < 30) {
if (ssDCval < 46) {
if (Rs < 2189) {
return 8;
}
else if (Rs >= 2189) {
return 4;
}
}
else if (ssDCval >= 46) {
if (posBalance < 54) {
return 8;
}
else if (posBalance >= 54) {
if (posBalance < 59) {
return 4;
}
else if (posBalance >= 59) {
return 8;
}
}
}
}
else if (negBalance >= 30) {
if (gchDC < 2) {
return 8;
}
else if (gchDC >= 2) {
return 4;
}
}
}
else if (Rs >= 2681) {
if (TSC < 4) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
return 4;
}
}
else if (TSC >= 4) {
if (Rs < 2896) {
return 4;
}
else if (Rs >= 2896) {
return 2;
}
}
}
}
else if (tcor >= 98) {
if (negBalance < 18) {
if (SC < 4475) {
if (Cs < 1351) {
if (Rs < 2273) {
return 8;
}
else if (Rs >= 2273) {
if (SC < 3595) {
return 2;
}
else if (SC >= 3595) {
if (mu_mv_mag_sq < 125) {
if (Rs < 2319) {
if (Cs < 1330) {
if (Rs < 2307) {
return 2;
}
else if (Rs >= 2307) {
return 8;
}
}
else if (Cs >= 1330) {
return 8;
}
}
else if (Rs >= 2319) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
if (mu_mv_mag_sq < 126) {
return 8;
}
else if (mu_mv_mag_sq >= 126) {
if (Rs < 2306) {
if (Rs < 2295) {
if (Cs < 1329) {
if (Rs < 2287) {
return 8;
}
else if (Rs >= 2287) {
return 2;
}
}
else if (Cs >= 1329) {
return 8;
}
}
else if (Rs >= 2295) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (Rs >= 2306) {
return 8;
}
}
}
}
}
}
else if (Cs >= 1351) {
if (SC < 3339) {
if (Cs < 1434) {
return 8;
}
else if (Cs >= 1434) {
if (AFD < 2) {
return 1;
}
else if (AFD >= 2) {
return 8;
}
}
}
else if (SC >= 3339) {
return 8;
}
}
}
else if (SC >= 4475) {
if (ssDCval < 39) {
return 8;
}
else if (ssDCval >= 39) {
return 1;
}
}
}
else if (negBalance >= 18) {
if (negBalance < 22) {
if (ssDCval < 84) {
if (Rs < 1601) {
return 8;
}
else if (Rs >= 1601) {
if (Cs < 1647) {
return 4;
}
else if (Cs >= 1647) {
return 8;
}
}
}
else if (ssDCval >= 84) {
return 8;
}
}
else if (negBalance >= 22) {
if (Rs < 1828) {
return 4;
}
else if (Rs >= 1828) {
if (RsCsDiff < 11) {
if (Rs < 2393) {
return 8;
}
else if (Rs >= 2393) {
if (gchDC < 1) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 4;
}
}
else if (gchDC >= 1) {
if (Cs < 1431) {
return 4;
}
else if (Cs >= 1431) {
return 8;
}
}
}
}
else if (RsCsDiff >= 11) {
if (Cs < 1343) {
return 2;
}
else if (Cs >= 1343) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 86) {
if (ssDCval < 140) {
if (posBalance < 46) {
if (diffAFD < -20) {
if (Rs < 2540) {
if (Rs < 1802) {
return 1;
}
else if (Rs >= 1802) {
return 4;
}
}
else if (Rs >= 2540) {
return 8;
}
}
else if (diffAFD >= -20) {
if (Rs < 3048) {
return 8;
}
else if (Rs >= 3048) {
if (Rs < 3055) {
return 2;
}
else if (Rs >= 3055) {
return 8;
}
}
}
}
else if (posBalance >= 46) {
if (refDCval < 117) {
if (SC < 4186) {
if (diffAFD < -2) {
if (AFD < 4) {
return 8;
}
else if (AFD >= 4) {
if (ssDCval < 91) {
return 2;
}
else if (ssDCval >= 91) {
return 8;
}
}
}
else if (diffAFD >= -2) {
return 8;
}
}
else if (SC >= 4186) {
if (Rs < 2652) {
return 1;
}
else if (Rs >= 2652) {
return 8;
}
}
}
else if (refDCval >= 117) {
if (ssDCval < 119) {
if (mu_mv_mag_sq < 123) {
if (AFD < 5) {
return 8;
}
else if (AFD >= 5) {
return 1;
}
}
else if (mu_mv_mag_sq >= 123) {
return 1;
}
}
else if (ssDCval >= 119) {
if (MVDiff < 7) {
if (Rs < 2895) {
return 8;
}
else if (Rs >= 2895) {
if (Rs < 3038) {
return 2;
}
else if (Rs >= 3038) {
return 8;
}
}
}
else if (MVDiff >= 7) {
return 1;
}
}
}
}
}
else if (ssDCval >= 140) {
if (ssDCval < 142) {
return 1;
}
else if (ssDCval >= 142) {
if (MVDiff < 2) {
return 2;
}
else if (MVDiff >= 2) {
if (Rs < 2545) {
return 8;
}
else if (Rs >= 2545) {
return 2;
}
}
}
}
}
}
}
else if (Rs >= 3143) {
if (ssDCval < 110) {
if (ssDCval < 78) {
if (mu_mv_mag_sq < 125) {
if (SC < 4917) {
if (AFD < 3) {
return 1;
}
else if (AFD >= 3) {
return 2;
}
}
else if (SC >= 4917) {
if (RsDiff < 6) {
if (negBalance < 21) {
if (ssDCval < 61) {
if (Rs < 5656) {
return 4;
}
else if (Rs >= 5656) {
return 8;
}
}
else if (ssDCval >= 61) {
return 1;
}
}
else if (negBalance >= 21) {
if (ssDCval < 68) {
if (ssDCval < 64) {
return 8;
}
else if (ssDCval >= 64) {
return 1;
}
}
else if (ssDCval >= 68) {
if (Rs < 4455) {
return 4;
}
else if (Rs >= 4455) {
return 8;
}
}
}
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (mu_mv_mag_sq >= 125) {
if (Cs < 1568) {
if (Rs < 3915) {
if (ssDCval < 57) {
if (Rs < 3570) {
return 8;
}
else if (Rs >= 3570) {
return 4;
}
}
else if (ssDCval >= 57) {
if (Rs < 3827) {
return 8;
}
else if (Rs >= 3827) {
return 4;
}
}
}
else if (Rs >= 3915) {
if (negBalance < 25) {
return 8;
}
else if (negBalance >= 25) {
if (Rs < 4030) {
return 8;
}
else if (Rs >= 4030) {
return 4;
}
}
}
}
else if (Cs >= 1568) {
if (Rs < 3230) {
return 8;
}
else if (Rs >= 3230) {
return 4;
}
}
}
}
else if (ssDCval >= 78) {
if (posBalance < 42) {
if (Rs < 4511) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
if (ssDCval < 97) {
if (Rs < 3904) {
if (Rs < 3602) {
return 2;
}
else if (Rs >= 3602) {
return 4;
}
}
else if (Rs >= 3904) {
return 8;
}
}
else if (ssDCval >= 97) {
return 4;
}
}
}
else if (Rs >= 4511) {
if (AFD < 4) {
return 2;
}
else if (AFD >= 4) {
return 8;
}
}
}
else if (posBalance >= 42) {
if (refDCval < 90) {
return 2;
}
else if (refDCval >= 90) {
if (Cs < 1532) {
if (Rs < 3400) {
return 2;
}
else if (Rs >= 3400) {
if (Rs < 4506) {
return 4;
}
else if (Rs >= 4506) {
return 2;
}
}
}
else if (Cs >= 1532) {
if (Rs < 4117) {
if (Rs < 3733) {
return 2;
}
else if (Rs >= 3733) {
return 4;
}
}
else if (Rs >= 4117) {
return 8;
}
}
}
}
}
}
else if (ssDCval >= 110) {
if (mcTcor < 93) {
if (Rs < 3547) {
return 2;
}
else if (Rs >= 3547) {
return 4;
}
}
else if (mcTcor >= 93) {
return 2;
}
}
}
}
else if (Cs >= 1664) {
if (ssDCval < 98) {
if (SC < 4361) {
if (RsDiff < 2) {
if (ssDCval < 67) {
if (ssDCval < 59) {
if (Rs < 1617) {
if (Rs < 1016) {
return 8;
}
else if (Rs >= 1016) {
return 4;
}
}
else if (Rs >= 1617) {
if (Cs < 1850) {
return 8;
}
else if (Cs >= 1850) {
if (ssDCval < 52) {
if (ssDCval < 34) {
return 8;
}
else if (ssDCval >= 34) {
return 1;
}
}
else if (ssDCval >= 52) {
return 8;
}
}
}
}
else if (ssDCval >= 59) {
if (AFD < 3) {
if (Cs < 1772) {
if (CsDiff < 2) {
if (AFD < 2) {
if (Cs < 1760) {
return 8;
}
else if (Cs >= 1760) {
return 1;
}
}
else if (AFD >= 2) {
return 8;
}
}
else if (CsDiff >= 2) {
return 1;
}
}
else if (Cs >= 1772) {
if (Rs < 2093) {
return 1;
}
else if (Rs >= 2093) {
if (ssDCval < 65) {
return 8;
}
else if (ssDCval >= 65) {
return 1;
}
}
}
}
else if (AFD >= 3) {
return 8;
}
}
}
else if (ssDCval >= 67) {
if (negBalance < 34) {
if (ssDCval < 91) {
if (posBalance < 59) {
return 8;
}
else if (posBalance >= 59) {
return 4;
}
}
else if (ssDCval >= 91) {
if (Rs < 2439) {
if (Rs < 2247) {
return 4;
}
else if (Rs >= 2247) {
return 2;
}
}
else if (Rs >= 2439) {
return 8;
}
}
}
else if (negBalance >= 34) {
if (Rs < 1551) {
return 1;
}
else if (Rs >= 1551) {
return 8;
}
}
}
}
else if (RsDiff >= 2) {
if (Rs < 2527) {
if (MVDiff < 3) {
if (negBalance < 61) {
if (Rs < 1041) {
return 1;
}
else if (Rs >= 1041) {
return 8;
}
}
else if (negBalance >= 61) {
return 4;
}
}
else if (MVDiff >= 3) {
if (negBalance < 30) {
if (posBalance < 43) {
return 8;
}
else if (posBalance >= 43) {
return 4;
}
}
else if (negBalance >= 30) {
if (Rs < 1042) {
return 2;
}
else if (Rs >= 1042) {
return 8;
}
}
}
}
else if (Rs >= 2527) {
if (AFD < 3) {
return 8;
}
else if (AFD >= 3) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 8;
}
}
}
}
}
else if (SC >= 4361) {
if (posBalance < 3) {
if (diffRsCsdiff < -5) {
if (tcor < 64) {
return 4;
}
else if (tcor >= 64) {
return 8;
}
}
else if (diffRsCsdiff >= -5) {
if (refDCval < 65) {
if (Rs < 6888) {
return 8;
}
else if (Rs >= 6888) {
return 2;
}
}
else if (refDCval >= 65) {
if (Rs < 3850) {
if (Cs < 2052) {
return 8;
}
else if (Cs >= 2052) {
if (Rs < 3822) {
if (AFD < 1) {
return 4;
}
else if (AFD >= 1) {
return 8;
}
}
else if (Rs >= 3822) {
return 8;
}
}
}
else if (Rs >= 3850) {
if (ssDCval < 89) {
return 4;
}
else if (ssDCval >= 89) {
return 8;
}
}
}
}
}
else if (posBalance >= 3) {
if (Rs < 6650) {
if (SC < 8417) {
if (refDCval < 81) {
if (refDCval < 67) {
if (posBalance < 43) {
if (ssDCval < 51) {
if (Cs < 2682) {
return 8;
}
else if (Cs >= 2682) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
}
else if (ssDCval >= 51) {
if (Cs < 3308) {
if (RsCsDiff < 2) {
if (SC < 4891) {
return 1;
}
else if (SC >= 4891) {
return 4;
}
}
else if (RsCsDiff >= 2) {
if (Rs < 2894) {
return 8;
}
else if (Rs >= 2894) {
if (SC < 5438) {
if (SC < 4889) {
if (TSCindex < 3) {
if (Rs < 2910) {
return 4;
}
else if (Rs >= 2910) {
if (posBalance < 38) {
if (Rs < 3003) {
return 8;
}
else if (Rs >= 3003) {
return 4;
}
}
else if (posBalance >= 38) {
return 8;
}
}
}
else if (TSCindex >= 3) {
if (Cs < 1913) {
if (posBalance < 40) {
if (SC < 4829) {
if (posBalance < 39) {
return 4;
}
else if (posBalance >= 39) {
if (Rs < 2944) {
return 4;
}
else if (Rs >= 2944) {
return 4;
}
}
}
else if (SC >= 4829) {
if (Rs < 3014) {
if (Rs < 2970) {
return 4;
}
else if (Rs >= 2970) {
return 4;
}
}
else if (Rs >= 3014) {
return 8;
}
}
}
else if (posBalance >= 40) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (Cs < 1846) {
if (Rs < 3013) {
return 4;
}
else if (Rs >= 3013) {
return 4;
}
}
else if (Cs >= 1846) {
return 4;
}
}
}
}
else if (Cs >= 1913) {
return 8;
}
}
}
else if (SC >= 4889) {
if (diffRsCsdiff < -2) {
if (Rs < 2963) {
return 8;
}
else if (Rs >= 2963) {
if (Rs < 3056) {
return 4;
}
else if (Rs >= 3056) {
return 4;
}
}
}
else if (diffRsCsdiff >= -2) {
return 4;
}
}
}
else if (SC >= 5438) {
if (ssDCval < 58) {
if (Cs < 2100) {
if (posBalance < 29) {
return 8;
}
else if (posBalance >= 29) {
if (diffRsCsdiff < 5) {
if (Cs < 1978) {
return 8;
}
else if (Cs >= 1978) {
return 4;
}
}
else if (diffRsCsdiff >= 5) {
if (Rs < 4160) {
return 8;
}
else if (Rs >= 4160) {
return 8;
}
}
}
}
else if (Cs >= 2100) {
if (posBalance < 22) {
return 8;
}
else if (posBalance >= 22) {
if (mcTcor < 95) {
return 8;
}
else if (mcTcor >= 95) {
if (Cs < 2104) {
return 4;
}
else if (Cs >= 2104) {
if (Cs < 2124) {
if (Rs < 4334) {
return 8;
}
else if (Rs >= 4334) {
if (Rs < 4452) {
return 8;
}
else if (Rs >= 4452) {
return 4;
}
}
}
else if (Cs >= 2124) {
if (tcor < 97) {
if (Cs < 2587) {
if (posBalance < 42) {
if (Cs < 2373) {
if (tcor < 93) {
return 4;
}
else if (tcor >= 93) {
if (RsDiff < 7) {
if (diffMVdiffVal < 1) {
if (RsDiff < 4) {
if (Rs < 4341) {
return 4;
}
else if (Rs >= 4341) {
return 4;
}
}
else if (RsDiff >= 4) {
if (negBalance < 15) {
return 4;
}
else if (negBalance >= 15) {
return 4;
}
}
}
else if (diffMVdiffVal >= 1) {
return 4;
}
}
else if (RsDiff >= 7) {
return 4;
}
}
}
else if (Cs >= 2373) {
if (tcor < 95) {
if (diffRsCsdiff < 14) {
if (Cs < 2567) {
if (Cs < 2473) {
if (AFD < 4) {
return 4;
}
else if (AFD >= 4) {
return 4;
}
}
else if (Cs >= 2473) {
return 4;
}
}
else if (Cs >= 2567) {
return 4;
}
}
else if (diffRsCsdiff >= 14) {
return 8;
}
}
else if (tcor >= 95) {
return 8;
}
}
}
else if (posBalance >= 42) {
return 4;
}
}
else if (Cs >= 2587) {
if (diffTSC < -1) {
if (Rs < 5082) {
if (Rs < 4860) {
return 4;
}
else if (Rs >= 4860) {
return 4;
}
}
else if (Rs >= 5082) {
return 4;
}
}
else if (diffTSC >= -1) {
return 4;
}
}
}
else if (tcor >= 97) {
if (Rs < 5093) {
return 8;
}
else if (Rs >= 5093) {
return 8;
}
}
}
}
}
}
}
}
else if (ssDCval >= 58) {
return 8;
}
}
}
}
}
else if (Cs >= 3308) {
return 8;
}
}
}
else if (posBalance >= 43) {
if (diffRsCsdiff < 13) {
if (negBalance < 18) {
if (Cs < 3336) {
if (Rs < 4519) {
if (Cs < 1848) {
return 8;
}
else if (Cs >= 1848) {
return 4;
}
}
else if (Rs >= 4519) {
return 8;
}
}
else if (Cs >= 3336) {
return 8;
}
}
else if (negBalance >= 18) {
if (MVDiff < 6) {
if (Rs < 2012) {
return 4;
}
else if (Rs >= 2012) {
if (diffRsCsdiff < -32) {
return 4;
}
else if (diffRsCsdiff >= -32) {
if (SC < 4421) {
return 1;
}
else if (SC >= 4421) {
if (SC < 4978) {
if (Rs < 2919) {
if (negBalance < 19) {
return 4;
}
else if (negBalance >= 19) {
return 8;
}
}
else if (Rs >= 2919) {
if (Rs < 2923) {
return 4;
}
else if (Rs >= 2923) {
return 8;
}
}
}
else if (SC >= 4978) {
if (RsDiff < 7) {
return 8;
}
else if (RsDiff >= 7) {
if (Rs < 3836) {
return 8;
}
else if (Rs >= 3836) {
return 4;
}
}
}
}
}
}
}
else if (MVDiff >= 6) {
if (ssDCval < 63) {
if (mu_mv_mag_sq < 125) {
return 8;
}
else if (mu_mv_mag_sq >= 125) {
return 4;
}
}
else if (ssDCval >= 63) {
return 4;
}
}
}
}
else if (diffRsCsdiff >= 13) {
if (mu_mv_mag_sq < 122) {
if (Rs < 3783) {
return 4;
}
else if (Rs >= 3783) {
if (gchDC < 1) {
return 4;
}
else if (gchDC >= 1) {
if (Rs < 4485) {
return 4;
}
else if (Rs >= 4485) {
return 4;
}
}
}
}
else if (mu_mv_mag_sq >= 122) {
if (Cs < 2163) {
return 4;
}
else if (Cs >= 2163) {
return 8;
}
}
}
}
}
else if (refDCval >= 67) {
if (Rs < 2982) {
if (Rs < 2175) {
if (CsDiff < 5) {
return 4;
}
else if (CsDiff >= 5) {
return 8;
}
}
else if (Rs >= 2175) {
if (mu_mv_mag_sq < 121) {
if (negBalance < 52) {
if (negBalance < 45) {
if (MVDiff < 1) {
return 4;
}
else if (MVDiff >= 1) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (MVDiff < 2) {
return 8;
}
else if (MVDiff >= 2) {
return 4;
}
}
}
}
else if (negBalance >= 45) {
return 8;
}
}
else if (negBalance >= 52) {
return 4;
}
}
else if (mu_mv_mag_sq >= 121) {
if (Cs < 1777) {
if (posBalance < 26) {
return 8;
}
else if (posBalance >= 26) {
if (Cs < 1748) {
return 8;
}
else if (Cs >= 1748) {
return 4;
}
}
}
else if (Cs >= 1777) {
if (Cs < 3125) {
if (diffMVdiffVal < -5) {
return 4;
}
else if (diffMVdiffVal >= -5) {
if (tcor < 100) {
if (mu_mv_mag_sq < 124) {
if (tcor < 96) {
return 8;
}
else if (tcor >= 96) {
if (SC < 4576) {
return 8;
}
else if (SC >= 4576) {
return 4;
}
}
}
else if (mu_mv_mag_sq >= 124) {
return 8;
}
}
else if (tcor >= 100) {
return 4;
}
}
}
else if (Cs >= 3125) {
return 4;
}
}
}
}
}
else if (Rs >= 2982) {
if (Cs < 1904) {
if (ssDCval < 74) {
return 4;
}
else if (ssDCval >= 74) {
return 2;
}
}
else if (Cs >= 1904) {
if (posBalance < 5) {
return 4;
}
else if (posBalance >= 5) {
if (diffRsCsdiff < -24) {
if (Cs < 2049) {
return 4;
}
else if (Cs >= 2049) {
if (posBalance < 62) {
return 8;
}
else if (posBalance >= 62) {
return 4;
}
}
}
else if (diffRsCsdiff >= -24) {
return 8;
}
}
}
}
}
}
else if (refDCval >= 81) {
if (Cs < 2677) {
if (Rs < 4916) {
if (Cs < 1931) {
if (SC < 5109) {
if (tcor < 98) {
return 8;
}
else if (tcor >= 98) {
if (negBalance < 19) {
return 8;
}
else if (negBalance >= 19) {
if (Cs < 1827) {
return 4;
}
else if (Cs >= 1827) {
if (ssDCval < 84) {
if (posBalance < 49) {
return 4;
}
else if (posBalance >= 49) {
if (Rs < 2789) {
return 4;
}
else if (Rs >= 2789) {
return 8;
}
}
}
else if (ssDCval >= 84) {
return 8;
}
}
}
}
}
else if (SC >= 5109) {
if (MVDiff < 1) {
if (Cs < 1833) {
if (refDCval < 90) {
if (Rs < 3860) {
return 4;
}
else if (Rs >= 3860) {
if (Rs < 4590) {
return 2;
}
else if (Rs >= 4590) {
return 8;
}
}
}
else if (refDCval >= 90) {
return 8;
}
}
else if (Cs >= 1833) {
if (ssDCval < 91) {
if (Rs < 3521) {
return 8;
}
else if (Rs >= 3521) {
return 2;
}
}
else if (ssDCval >= 91) {
if (Cs < 1920) {
return 8;
}
else if (Cs >= 1920) {
return 2;
}
}
}
}
else if (MVDiff >= 1) {
return 2;
}
}
}
else if (Cs >= 1931) {
if (mu_mv_mag_sq < 127) {
if (tcor < 99) {
if (Rs < 3945) {
if (RsDiff < 1) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
if (Rs < 2977) {
return 4;
}
else if (Rs >= 2977) {
return 8;
}
}
}
else if (RsDiff >= 1) {
if (negBalance < 57) {
if (SC < 6355) {
if (Cs < 2619) {
return 8;
}
else if (Cs >= 2619) {
if (Rs < 2564) {
return 4;
}
else if (Rs >= 2564) {
return 8;
}
}
}
else if (SC >= 6355) {
return 4;
}
}
else if (negBalance >= 57) {
return 4;
}
}
}
else if (Rs >= 3945) {
if (posBalance < 42) {
if (Cs < 1969) {
return 2;
}
else if (Cs >= 1969) {
return 8;
}
}
else if (posBalance >= 42) {
return 2;
}
}
}
else if (tcor >= 99) {
if (negBalance < 11) {
if (Rs < 2370) {
if (refDCval < 92) {
return 1;
}
else if (refDCval >= 92) {
if (negBalance < 8) {
if (Rs < 2345) {
if (Cs < 2076) {
return 4;
}
else if (Cs >= 2076) {
return 2;
}
}
else if (Rs >= 2345) {
return 1;
}
}
else if (negBalance >= 8) {
return 4;
}
}
}
else if (Rs >= 2370) {
if (ssDCval < 85) {
if (Rs < 3100) {
return 1;
}
else if (Rs >= 3100) {
return 8;
}
}
else if (ssDCval >= 85) {
return 8;
}
}
}
else if (negBalance >= 11) {
if (Cs < 2275) {
return 2;
}
else if (Cs >= 2275) {
if (Rs < 2516) {
return 4;
}
else if (Rs >= 2516) {
return 8;
}
}
}
}
}
else if (mu_mv_mag_sq >= 127) {
if (AFD < 21) {
if (SC < 4460) {
if (Rs < 2255) {
return 8;
}
else if (Rs >= 2255) {
return 4;
}
}
else if (SC >= 4460) {
return 8;
}
}
else if (AFD >= 21) {
return 1;
}
}
}
}
else if (Rs >= 4916) {
return 4;
}
}
else if (Cs >= 2677) {
if (ssDCval < 89) {
if (Cs < 3292) {
if (SC < 6692) {
if (SC < 5751) {
return 4;
}
else if (SC >= 5751) {
if (SC < 6453) {
return 8;
}
else if (SC >= 6453) {
if (Cs < 3140) {
return 4;
}
else if (Cs >= 3140) {
return 8;
}
}
}
}
else if (SC >= 6692) {
return 4;
}
}
else if (Cs >= 3292) {
if (negBalance < 27) {
return 4;
}
else if (negBalance >= 27) {
if (posBalance < 63) {
if (diffRsCsdiff < 24) {
return 8;
}
else if (diffRsCsdiff >= 24) {
if (Cs < 3426) {
return 4;
}
else if (Cs >= 3426) {
return 8;
}
}
}
else if (posBalance >= 63) {
if (Rs < 3617) {
return 4;
}
else if (Rs >= 3617) {
return 8;
}
}
}
}
}
else if (ssDCval >= 89) {
if (Rs < 3962) {
if (negBalance < 50) {
if (Rs < 3328) {
if (tcor < 97) {
return 8;
}
else if (tcor >= 97) {
return 4;
}
}
else if (Rs >= 3328) {
if (mcTcor < 99) {
if (mu_mv_mag_sq < 123) {
return 4;
}
else if (mu_mv_mag_sq >= 123) {
if (Rs < 3418) {
return 8;
}
else if (Rs >= 3418) {
if (RsDiff < 8) {
if (Cs < 3140) {
if (Cs < 3002) {
return 4;
}
else if (Cs >= 3002) {
return 8;
}
}
else if (Cs >= 3140) {
return 4;
}
}
else if (RsDiff >= 8) {
if (Rs < 3437) {
return 4;
}
else if (Rs >= 3437) {
return 8;
}
}
}
}
}
else if (mcTcor >= 99) {
if (posBalance < 46) {
return 4;
}
else if (posBalance >= 46) {
return 8;
}
}
}
}
else if (negBalance >= 50) {
if (posBalance < 56) {
if (Cs < 3132) {
return 4;
}
else if (Cs >= 3132) {
return 8;
}
}
else if (posBalance >= 56) {
return 4;
}
}
}
else if (Rs >= 3962) {
if (Cs < 3683) {
return 4;
}
else if (Cs >= 3683) {
return 8;
}
}
}
}
}
}
else if (SC >= 8417) {
if (MVDiff < 3) {
if (negBalance < 29) {
if (Rs < 4583) {
if (Cs < 4213) {
return 8;
}
else if (Cs >= 4213) {
return 1;
}
}
else if (Rs >= 4583) {
if (RsDiff < 6) {
return 8;
}
else if (RsDiff >= 6) {
return 2;
}
}
}
else if (negBalance >= 29) {
if (Cs < 3683) {
if (Rs < 5208) {
return 4;
}
else if (Rs >= 5208) {
return 8;
}
}
else if (Cs >= 3683) {
return 8;
}
}
}
else if (MVDiff >= 3) {
if (Rs < 6022) {
return 8;
}
else if (Rs >= 6022) {
if (AFD < 2) {
return 8;
}
else if (AFD >= 2) {
return 2;
}
}
}
}
}
else if (Rs >= 6650) {
if (RsDiff < 44) {
return 4;
}
else if (RsDiff >= 44) {
return 8;
}
}
}
}
}
else if (ssDCval >= 98) {
if (posBalance < 53) {
if (diffRsCsdiff < -1161) {
if (Rs < 4261) {
return 1;
}
else if (Rs >= 4261) {
return 8;
}
}
else if (diffRsCsdiff >= -1161) {
if (tcor < 99) {
if (ssDCval < 200) {
if (Cs < 1809) {
if (Rs < 3179) {
if (MVDiff < 57) {
if (Cs < 1800) {
return 8;
}
else if (Cs >= 1800) {
if (Rs < 1848) {
return 8;
}
else if (Rs >= 1848) {
return 1;
}
}
}
else if (MVDiff >= 57) {
return 4;
}
}
else if (Rs >= 3179) {
if (AFD < 5) {
return 4;
}
else if (AFD >= 5) {
return 2;
}
}
}
else if (Cs >= 1809) {
if (tcor < 95) {
if (diffTSC < 7) {
return 8;
}
else if (diffTSC >= 7) {
if (Rs < 3321) {
return 1;
}
else if (Rs >= 3321) {
return 8;
}
}
}
else if (tcor >= 95) {
if (diffTSC < -2) {
if (negBalance < 32) {
if (mu_mv_mag_sq < 126) {
if (diffTSC < -7) {
return 1;
}
else if (diffTSC >= -7) {
if (AFD < 1) {
return 2;
}
else if (AFD >= 1) {
return 8;
}
}
}
else if (mu_mv_mag_sq >= 126) {
if (Cs < 2008) {
if (Rs < 2728) {
return 8;
}
else if (Rs >= 2728) {
return 4;
}
}
else if (Cs >= 2008) {
return 8;
}
}
}
else if (negBalance >= 32) {
if (Rs < 3803) {
if (AFD < 5) {
if (Rs < 3539) {
return 4;
}
else if (Rs >= 3539) {
return 8;
}
}
else if (AFD >= 5) {
return 8;
}
}
else if (Rs >= 3803) {
if (Cs < 3004) {
return 2;
}
else if (Cs >= 3004) {
return 4;
}
}
}
}
else if (diffTSC >= -2) {
if (CsDiff < 1) {
if (mu_mv_mag_sq < 125) {
if (Rs < 2768) {
if (SC < 5692) {
if (posBalance < 15) {
if (negBalance < 10) {
return 8;
}
else if (negBalance >= 10) {
return 4;
}
}
else if (posBalance >= 15) {
return 4;
}
}
else if (SC >= 5692) {
return 8;
}
}
else if (Rs >= 2768) {
return 2;
}
}
else if (mu_mv_mag_sq >= 125) {
return 8;
}
}
else if (CsDiff >= 1) {
if (negBalance < 33) {
if (Cs < 1881) {
if (Rs < 3546) {
return 8;
}
else if (Rs >= 3546) {
return 4;
}
}
else if (Cs >= 1881) {
if (Rs < 6526) {
return 8;
}
else if (Rs >= 6526) {
if (Cs < 4462) {
return 4;
}
else if (Cs >= 4462) {
return 8;
}
}
}
}
else if (negBalance >= 33) {
if (Rs < 7230) {
if (RsCsDiff < 4) {
if (Cs < 2390) {
return 8;
}
else if (Cs >= 2390) {
return 4;
}
}
else if (RsCsDiff >= 4) {
if (negBalance < 41) {
if (Rs < 6722) {
if (refDCval < 99) {
if (Cs < 2291) {
return 2;
}
else if (Cs >= 2291) {
return 8;
}
}
else if (refDCval >= 99) {
return 8;
}
}
else if (Rs >= 6722) {
if (Rs < 6795) {
return 4;
}
else if (Rs >= 6795) {
return 8;
}
}
}
else if (negBalance >= 41) {
if (RsCsDiff < 38) {
if (posBalance < 51) {
return 8;
}
else if (posBalance >= 51) {
if (Rs < 4250) {
return 4;
}
else if (Rs >= 4250) {
return 2;
}
}
}
else if (RsCsDiff >= 38) {
if (AFD < 9) {
return 8;
}
else if (AFD >= 9) {
if (Rs < 2901) {
return 8;
}
else if (Rs >= 2901) {
return 4;
}
}
}
}
}
}
else if (Rs >= 7230) {
return 2;
}
}
}
}
}
}
}
else if (ssDCval >= 200) {
if (Cs < 4084) {
return 8;
}
else if (Cs >= 4084) {
return 4;
}
}
}
else if (tcor >= 99) {
if (ssDCval < 123) {
if (MVDiff < 1) {
if (negBalance < 36) {
if (Cs < 1675) {
if (Rs < 4139) {
return 8;
}
else if (Rs >= 4139) {
return 2;
}
}
else if (Cs >= 1675) {
return 8;
}
}
else if (negBalance >= 36) {
if (refDCval < 100) {
return 4;
}
else if (refDCval >= 100) {
if (RsDiff < 1) {
return 2;
}
else if (RsDiff >= 1) {
return 8;
}
}
}
}
else if (MVDiff >= 1) {
if (AFD < 1) {
if (diffRsCsdiff < -7) {
return 8;
}
else if (diffRsCsdiff >= -7) {
if (posBalance < 13) {
return 2;
}
else if (posBalance >= 13) {
return 8;
}
}
}
else if (AFD >= 1) {
if (Cs < 2118) {
if (diffMVdiffVal < -1) {
return 4;
}
else if (diffMVdiffVal >= -1) {
return 8;
}
}
else if (Cs >= 2118) {
if (posBalance < 25) {
if (Rs < 2819) {
if (Cs < 2930) {
return 4;
}
else if (Cs >= 2930) {
return 8;
}
}
else if (Rs >= 2819) {
return 8;
}
}
else if (posBalance >= 25) {
return 8;
}
}
}
}
}
else if (ssDCval >= 123) {
if (ssDCval < 138) {
if (Cs < 3866) {
if (tcor < 100) {
if (refDCval < 137) {
return 8;
}
else if (refDCval >= 137) {
if (Rs < 2582) {
return 8;
}
else if (Rs >= 2582) {
return 1;
}
}
}
else if (tcor >= 100) {
if (posBalance < 44) {
if (posBalance < 3) {
if (MVDiff < 2) {
return 1;
}
else if (MVDiff >= 2) {
return 8;
}
}
else if (posBalance >= 3) {
return 8;
}
}
else if (posBalance >= 44) {
return 1;
}
}
}
else if (Cs >= 3866) {
if (Rs < 6512) {
return 1;
}
else if (Rs >= 6512) {
return 8;
}
}
}
else if (ssDCval >= 138) {
if (Cs < 1876) {
if (Rs < 3972) {
return 8;
}
else if (Rs >= 3972) {
return 2;
}
}
else if (Cs >= 1876) {
if (ssDCval < 202) {
return 8;
}
else if (ssDCval >= 202) {
return 4;
}
}
}
}
}
}
}
else if (posBalance >= 53) {
if (Rs < 1636) {
if (SC < 3081) {
if (ssDCval < 131) {
if (posBalance < 55) {
return 8;
}
else if (posBalance >= 55) {
if (Rs < 671) {
return 8;
}
else if (Rs >= 671) {
return 4;
}
}
}
else if (ssDCval >= 131) {
return 8;
}
}
else if (SC >= 3081) {
if (mcTcor < 92) {
if (TSC < 13) {
return 8;
}
else if (TSC >= 13) {
return 4;
}
}
else if (mcTcor >= 92) {
if (posBalance < 54) {
if (Rs < 1402) {
return 8;
}
else if (Rs >= 1402) {
return 2;
}
}
else if (posBalance >= 54) {
return 2;
}
}
}
}
else if (Rs >= 1636) {
if (ssDCval < 137) {
if (refDCval < 105) {
if (Rs < 3781) {
if (Cs < 2401) {
return 8;
}
else if (Cs >= 2401) {
if (Rs < 3305) {
if (SC < 6309) {
if (Rs < 3133) {
if (Cs < 3241) {
if (Rs < 2319) {
return 4;
}
else if (Rs >= 2319) {
return 8;
}
}
else if (Cs >= 3241) {
return 4;
}
}
else if (Rs >= 3133) {
return 4;
}
}
else if (SC >= 6309) {
if (Rs < 2652) {
return 4;
}
else if (Rs >= 2652) {
return 8;
}
}
}
else if (Rs >= 3305) {
return 8;
}
}
}
else if (Rs >= 3781) {
if (Cs < 3093) {
if (Rs < 4919) {
if (Rs < 4052) {
return 2;
}
else if (Rs >= 4052) {
return 1;
}
}
else if (Rs >= 4919) {
return 2;
}
}
else if (Cs >= 3093) {
return 8;
}
}
}
else if (refDCval >= 105) {
if (RsDiff < 2) {
if (Rs < 1901) {
if (Cs < 1815) {
return 8;
}
else if (Cs >= 1815) {
return 2;
}
}
else if (Rs >= 1901) {
return 8;
}
}
else if (RsDiff >= 2) {
if (gchDC < 1) {
if (diffAFD < -4) {
if (Rs < 5364) {
return 4;
}
else if (Rs >= 5364) {
return 8;
}
}
else if (diffAFD >= -4) {
if (diffAFD < 2) {
if (negBalance < 55) {
if (negBalance < 36) {
if (AFD < 6) {
return 8;
}
else if (AFD >= 6) {
return 4;
}
}
else if (negBalance >= 36) {
if (posBalance < 63) {
if (diffAFD < -3) {
if (MVDiff < 1) {
return 8;
}
else if (MVDiff >= 1) {
return 1;
}
}
else if (diffAFD >= -3) {
return 8;
}
}
else if (posBalance >= 63) {
if (AFD < 11) {
return 8;
}
else if (AFD >= 11) {
return 1;
}
}
}
}
else if (negBalance >= 55) {
if (RsDiff < 6) {
return 2;
}
else if (RsDiff >= 6) {
return 8;
}
}
}
else if (diffAFD >= 2) {
if (negBalance < 46) {
if (AFD < 7) {
return 8;
}
else if (AFD >= 7) {
if (Rs < 3192) {
return 8;
}
else if (Rs >= 3192) {
return 4;
}
}
}
else if (negBalance >= 46) {
return 8;
}
}
}
}
else if (gchDC >= 1) {
if (tcor < 87) {
if (Cs < 3503) {
if (Rs < 2706) {
return 8;
}
else if (Rs >= 2706) {
return 2;
}
}
else if (Cs >= 3503) {
return 8;
}
}
else if (tcor >= 87) {
if (diffAFD < -2) {
if (tcor < 95) {
if (Rs < 2470) {
return 8;
}
else if (Rs >= 2470) {
if (Rs < 3261) {
return 1;
}
else if (Rs >= 3261) {
return 2;
}
}
}
else if (tcor >= 95) {
return 8;
}
}
else if (diffAFD >= -2) {
if (diffAFD < 5) {
if (diffMVdiffVal < -2) {
if (posBalance < 61) {
return 8;
}
else if (posBalance >= 61) {
return 2;
}
}
else if (diffMVdiffVal >= -2) {
if (diffTSC < 3) {
return 8;
}
else if (diffTSC >= 3) {
if (Rs < 4329) {
return 8;
}
else if (Rs >= 4329) {
return 2;
}
}
}
}
else if (diffAFD >= 5) {
return 4;
}
}
}
}
}
}
}
else if (ssDCval >= 137) {
if (mcTcor < 96) {
if (Rs < 5912) {
return 1;
}
else if (Rs >= 5912) {
if (Rs < 8232) {
return 8;
}
else if (Rs >= 8232) {
return 4;
}
}
}
else if (mcTcor >= 96) {
if (mcTcor < 97) {
if (Cs < 3343) {
return 1;
}
else if (Cs >= 3343) {
return 8;
}
}
else if (mcTcor >= 97) {
return 8;
}
}
}
}
}
}
}
}
}
return 8;
}
typedef std::function<bool(
std::pair<mfxU16, mfxU16>,
std::pair<mfxU16, mfxU16>)>
Size_Comparator;

mfxU16 AGOPHEVCSelectRF(
mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 MVDiff, mfxU32 Rs, mfxU32 AFD,
mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 TSC, mfxU32 gchDC, mfxI32 diffRsCsdiff,
mfxU32 posBalance, mfxU32 SC, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 Cs,
mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxU32 refDCval, mfxU32 RsDiff,
mfxU32 mu_mv_mag_sq, mfxI16 mcTcor, mfxI16 tcor) {

(void)Scindex;
std::map<mfxU16, mfxU16> gop_freq;

Size_Comparator Comparison = [](
std::pair<mfxU16, mfxU16> size1,
std::pair<mfxU16, mfxU16> size2
)
{
return size1.second != size2.second ? size1.second > size2.second : size1.first < size2.first;
};

gop_freq[AGOPSelect1(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect2(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect3(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect4(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect5(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect6(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect7(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect8(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect9(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelectA(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelectB(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelectC(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelectD(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelectE(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelectF(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect10(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect11(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect12(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect13(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect14(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect15(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect16(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;
gop_freq[AGOPSelect17(Rs, CsDiff, TSC, diffAFD, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, Cs, mu_mv_mag_sq, negBalance, mcTcor, gchDC, SC, ssDCval)]++;

std::set<std::pair<mfxU16, mfxU16>, Size_Comparator>
size_sorted(
gop_freq.begin(),
gop_freq.end(),
Comparison);

return(size_sorted.begin()->first);
}
};//namespace aenc
#endif // MFX_ENABLE_ADAPTIVE_ENCODE
